% Casper Ti. Vector's biblatex style
%
% Copyright (c) 2012-2016,2018-2019 Casper Ti. Vector
%
% This work may be distributed and/or modified under the conditions of the
% LaTeX Project Public License, either version 1.3 of this license or (at
% your option) any later version.
% The latest version of this license is in
%   https://www.latex-project.org/lppl.txt
% and version 1.3 or later is part of all distributions of LaTeX version
% 2005/12/01 or later.
%
% This work has the LPPL maintenance status `maintained'.
% The current maintainer of this work is Casper Ti. Vector.
%
% This work consists of the following files:
%   caspervector.bbx
%   caspervector.cbx
%   caspervector-ay.bbx
%   caspervector-ay.cbx
%   blx-caspervector-base.def
%   blx-caspervector-gbk.def
%   blx-caspervector-utf8.def

\ProvidesFile{blx-caspervector-gbk.def}%
	[2024/03/07 v0.3.7 Chinese punctuations and captions in GBK]

\newcommand*\bbx@cnldquot{“}
\newcommand*\bbx@cnrdquot{”}
\newcommand*\bbx@cncomma{，}
\newcommand*\bbx@cnbcomma{、}
\newcommand*\bbx@cnscolon{；}
\newcommand*\bbx@cncolon{：}
\newcommand*\bbx@cnperiod{。}
\newcommand*\bbx@cnlparen{（}
\newcommand*\bbx@cnrparen{）}
\newcommand*\bbx@cnetal{等}
\newcommand*\bbx@cnnr{第}
\newcommand*\bbx@cnedition{版}
\newcommand*\bbx@cnin{出自}
\newcommand*\bbx@cnretr{访问于}

\DefineBibliographyStrings{english}{
	cnauthor = {著},
	cneditor = {编},
	cntranslator = {译},
	bibliography = {参考文献},
	references = {参考文献},
}

% vim:ft=tex:ts=2:sw=2
