---@diagnostic disable
-- This file is generated from the data in <https://github.com/JabRef/abbrv.jabref.org>
-- under the CC0-1.0 license.

abbrevs = {
  ["$k$-monographs in mathematics"] = "$K$-Monogr. Math.",
  ["(bulletin trimestriel de la) société de géographie et d’archéologie (de la province) d’oran"] = "SocGeoAOran",
  ["10th ieee international conference on big knowledge : proceedings : 10-11 november 2019, beijing, china. ieee international conference on big knowledge (10th : 2019 : beijing, china)"] = "10th IEEE Int Conf Big Knowl (2019)",
  ["1199 news. national union of hospital and health care employees. district 1199"] = "1199 News",
  ["19 : interdisciplinary studies in the long nineteenth century"] = "19 (Lond)",
  ["1999, zeitschrift für sozialgeschichte des 20. und 21. jahrhunderts"] = "1999 Z Sozial 20 21 Jhd",
  ["19th century music"] = "19th Century Music",
  ["20 century british history"] = "20 Century Br Hist",
  ["2013 acm conference on bioinformatics, computational biology and biomedical informatics : acm - bcb 2013 : washington, d.c., u.s.a., september 22 - 25, 2013. acm conference on bioinformatics, computational biology and biomedical informatics (4th : 2013 : washington, d.c.)"] = "ACM Conf Bioinform Comput Biol Biomed Inform (2013)",
  ["2013 ieee 26th international conference on micro electro mechanical systems (mems 2013) : taipei, taiwan, 20-24 january 2013. ieee international conference on micro electro mechanical systems (26th : 2013 : taipei, taiwan)"] = "2013 IEEE 26th Int Conf Micro Electro Mech Syst MEMS 2013 (2013)",
  ["2015 ieee signal processing and signal processing education workshop (sp/spe)"] = "2015 IEEE Signal Process Signal Process Educ Workshop SP SPE (2015)",
  ["2017 design of medical devices conference : april 10-13, 2017, minneapolis, minnesota, usa. design of medical devices conference (2017 : minneapolis, minn.)"] = "2017 Des Med Devices Conf (2017)",
  ["2017 first ieee mtt-s international microwave bio conference (imbioc) : 15-17 may, 2017, gothenburg, sweden. ieee mtt-s international microwave bio conference (1st : 2017 : göteborg, sweden)"] = "2017 First IEEE MTTS Int Microw Bio Conf (IMBioC) (2017)",
  ["2017 ieee sensors applications symposium (sas). ieee staff"] = "2017 IEEE Sens Appl Symp (SAS) (2017)",
  ["2017 ieee workshop on visual analytics in healthcare (vahc) : 1 october 2017. workshop on visual analytics in healthcare (8th : 2017 : phoenix, ariz.)"] = "2017 IEEE Workshop Vis Anal Healthc VAHC (2017)",
  ["2018 design of medical devices conference : april 9-12, 2018, minneapolis, minnesota, usa. design of medical devices conference (2018 : minneapolis, minn.)"] = "2018 Des Med Devices Conf (2018)",
  ["2018 ieee conference on cognitive and computational aspects of situation management (cogsima) : proceedings : boston, ma, usa, 11-14 june 2018. ieee conference on cognitive and computational aspects of situation management (8th : 2018 : boston, mass.)"] = "2018 IEEE Conf Cogn Comput Asp Situat Manag CogSIMA (2018)",
  ["2018 ieee international conference on healthcare informatics workshops : proceedings : 4-7 june 2018, new york, ny. international conference on healthcare informatics (6th : 2018 : new york, n.y.). workshops"] = "2018 IEEE Int Conf Healthc Inform Workshop (2018)",
  ["2018 international symposium on big data visual and immersive analytics (bdva) : konstanz, germany, october 17 -19, 2018. ieee international symposium on big data visual and immersive analytics (4th : 2018 : konstanz, germany)"] = "2018 Int Symp Big Data Vis Immers Analyt (BDVA) (2018)",
  ["2019 design of medical devices conference : april 15-18, 2019, minneapolis, minnesota, usa. design of medical devices conference (2019 : minneapolis, minn.)"] = "2019 Des Med Devices Conf (2019)",
  ["2019 ieee smartworld, ubiquitous intelligence & computing, advanced & trusted computing, scalable computing & communications, cloud & big data computing, internet of people and smart city innovation : smartworld/uic/atc/scalcom/iop/sci 2019"] = "2019 IEEE SmartWorld Ubiquitous Intell Comput Adv Trust Comput Scalable Comput Commun Cloud Big Data Comput Internet People Smart City Innov (2019)",
  ["21st century science & technology"] = "21st Century Sci Technol",
  ["24th international conference on 3d web technology"] = "24th Int Conf 3D Web Technol (2019)",
  ["2d materials"] = "2D Mater.",
  ["2nd international conference on bioinformatics and computational biology 2010, (bicob-2010), honolulu, hawaii, usa, 24-26 march 2010. international conference on bioinformatics and computational biology (2nd : 2010 : honolulu, hawaii)"] = "2nd Int Conf Bioinform Comput Biol (2010)",
  ["3 biotech"] = "3 Biotech",
  ["30$\\\\sp {\\\\rm o}$ coloquio brasileiro de matematica"] = "30$\\\\sp {\\\\rm o}$ Coloq. Bras. Mat.",
  ["31$\\\\sp {\\\\rm o}$ coloquio brasileiro de matematica"] = "31$\\\\sp {\\\\rm o}$ Coloq. Bras. Mat.",
  ["32 $\\\\sp {\\\\rm o}$ coloquio brasileiro de matematica"] = "32 $\\\\sp {\\\\rm o}$ Coloq. Bras. Mat.",
  ["33 $\\\\sp {\\\\rm o}$ coloquio brasileiro de matematica"] = "33 $\\\\sp {\\\\rm o}$ Coloq. Bras. Mat.",
  ["34 $\\\\sp {\\\\rm o}$ coloquio brasileiro de matematica"] = "34 $\\\\sp {\\\\rm o}$ Coloq. Bras. Mat.",
  ["35 vdi jahrestagung schadensanalyse:produktverbesserung durch schadensanalyse"] = "Vdi Bericht.",
  ["3d printing and additive manufacturing"] = "3D Print. Addit. Manuf.",
  ["3d printing in medicine"] = "3D Print Med",
  ["48th aiaa fluid dynamics conference 2018 : held at the aiaa aviation forum 2018 : atlanta, georgia, usa, 25-29 june 2018. aiaa fluid dynamics conference (48th : 2018 : atlanta, ga.)"] = "48th AIAA Fluid Dyn Conf 2018 (2018)",
  ["4or - a quarterly journal of operations research"] = "4OR - Q. J. Oper. Res.",
  ["4or. a quarterly journal of operations research"] = "4OR",
  ["50th anniversary cracow university of technology"] = "50th Anniv. Cracow Univ. Technol. Anniv. Issue",
  ["7th conference on retroviruses and opportunistic infections, january 30-february 2, 2000. conference on retroviruses and opportunistic infections (7th : 2000 : san francisco, calif.)"] = "7th Conf Retrovir Oppor Infect Jan 30 Feb 2 2000 Conf Retrovir Oppor Infect 7th 2000 San Franc Calif",
  ["8th annual symposium on nonhuman primate models for aids : november 28-30, 1990, the westin canal place hotel, new orleans, louisiana. symposium on nonhuman primate models for aids (8th : 1990 : new orleans, la.)"] = "8th Annu Symp Nonhum Primate Models AIDS Symp Nonhum Primate Models AIDS 8th 1990 New Orleans La",
  ["9th conference on retroviruses and opportunistic infections, february 24-28, 2002, washington state convention and trade center, seattle, wa. conference on retroviruses and opportunistic infections (9th : 2002 : seattle, wash.)"] = "9th Conf Retrovir Oppor Infect Feb 24 28 2002 Wash State Conv Trade Cent Seattle Wash Conf Retrovir Oppor Infect 9th 2002 Seattle Wash",
  ["<der >landkreis"] = "Landkreis",
  ["<le > bulletin trimestriel du planning familial"] = "Bull Trimest Plan Fam",
  ["[boei eisei] japanese defense forces medical journal"] = "Boei Eisei",
  ["[bōei eisei] japanese defense forces medical journal"] = "Boei Eisei",
  ["[ch'oesin ŭihak] the new medical journal"] = "Choesin Uihak New Med J",
  ["[chapchi] journal. taehan oekwa hakhoe"] = "Taehan Oekwa Hakhoe Chapchi",
  ["[chiryo] [therapy]"] = "Chiryo",
  ["[chiryō] [therapy]"] = "Chiryo",
  ["[choson uihak] [korean medicine]"] = "Choson Uihak",
  ["[chosŏn ŭihak] [korean medicine]"] = "Choson Uihak",
  ["[eisei dobutsu] japanese journal of sanitary zoology"] = "Eisei Dobutsu",
  ["[eisei dōbutsu] japanese journal of sanitary zoology"] = "Eisei Dobutsu",
  ["[eiyŏ to shokuryŏ] journal of the japanese society of food and nutrition"] = "Eiyo To Shokuryo",
  ["[fukushima igaku zasshi] fukushima medical journal"] = "Fukushima Igaku Zasshi",
  ["[gencho hiroshima igaku] the hiroshima medical journal"] = "Gencho Hiroshima Igaku",
  ["[gendai igaku] the current medicine"] = "Gendai Igaku",
  ["[hai] the lung"] = "Hai",
  ["[hanguk uiyak] korean journal of medicine"] = "Hanguk Uiyak",
  ["[hifu] skin research"] = "Hifu",
  ["[hokenfu zasshi] the japanese journal for public health nurse"] = "Hokenfu Zasshi",
  ["[hokkaido igaku zasshi] the hokkaido journal of medical science"] = "Hokkaido Igaku Zasshi",
  ["[igaku zasshi] medical journal. kagoshima daigaku. igakubu"] = "Kagoshima Daigaku Igaku Zasshi",
  ["[igaku] [medicine]"] = "Igaku",
  ["[igakushi kenkyu] [studies on history of medicine]"] = "Igakushi Kenkyu",
  ["[itan] [medical memoirs]"] = "Itan",
  ["[jika shinryō] journal for pediatric praxis"] = "Jika Shinryo",
  ["[kagaku] [science]"] = "Kagaku",
  ["[kango gijutsu] : [nursing technique]"] = "Kango Gijutsu",
  ["[kango kyoiku] japanese journal of nurses’ education"] = "Kango Kyoiku",
  ["[kango kyōiku] japanese journal of nurses' education"] = "Kango Kyoiku",
  ["[kango] japanese journal of nursing"] = "Kango",
  ["[kangogaku zasshi] the japanese journal of nursing"] = "Kangogaku Zasshi",
  ["[ke xue tong bao] [science bulletin]"] = "Ko Hsueh Tung Pao",
  ["[kokyuki shinryo] clinics of respiratory organs"] = "Kokyuki Shinryo",
  ["[kokyūki shinryō] clinics of respiratory organs"] = "Kokyuki Shinryo",
  ["[kōsei kagaku] the japanese journal of public health"] = "Kosei kagaku Jpn J Public Health",
  ["[ku chi chui tung wu yü ku jên lei] : vertebrata palasiatica"] = "Ku Chi Chui Tung Wu Yu Ku Jen Lei",
  ["[kusuri no chishiki] pharmaceutical review"] = "Kusuri No Chishiki",
  ["[lin chuang jian yan za zhi] [journal of clinical laboratory examination]"] = "Lin Chuang Jian Yan Za Zhi",
  ["[mediterranean studies]"] = "Mediterr Stud (Kirksv Mo)",
  ["[meiji seika kenkyū nenpō; yakuhin bumon] scientific reports of meiji seika kaisha; pharmaceutical division"] = "Meiji Seika Kenkyu Nenpo Yakuhin Bumon",
  ["[minzoku eisei] race hygiene"] = "Minzoku Eisei",
  ["[nenpo] the annual report of takamine laboratory. takamine kenkyusho"] = "Takamine Kenkyusho Nenpo",
  ["[nihon iji shinpō] = japanese medical journal"] = "Jpn Med J",
  ["[nihon kōshū eisei zasshi] japanese journal of public health"] = "Nihon Koshu Eisei Zasshi",
  ["[nihon rinshō kekkaku] the japanese journal of clinical tuberculosis"] = "Nihon Rinsho Kekkaku",
  ["[nihon shika hyōron] the nippon dental review"] = "Nihon Shika Hyoron",
  ["[nihon yakubutsugaku zasshi] [japanese journal of pharmacology]"] = "Nihon Yakubutsugaku Zasshi",
  ["[nippon iji shinpo] japanese medical journal"] = "Jpn Med J",
  ["[nippon koshu eisei zasshi] japanese journal of public health"] = "Nippon Koshu Eisei Zasshi",
  ["[nippon shika hyoron] the nippon dental review"] = "Nippon Shika Hyoron",
  ["[nisshin igaku] the japanese journal of medical progress"] = "Nisshin Igaku Jpn J Med Prog",
  ["[ochanomizu igaku zasshi] : the ochanomizu medical journal"] = "Ochanomizu Igaku Zasshi",
  ["[ochanomizu igaku zasshi] the ochanomizu medical journal"] = "Ochanomizu Igaku Zasshi",
  ["[onsen kagaku] journal of the balneological society of japan"] = "Onsen Kagaku",
  ["[osaka daigaku shigaku zasshi] the journal of osaka university dental society"] = "Osaka Daigaku Shigaku Zasshi",
  ["[preconference papers]. national dental health conference"] = "Preconf Papers Natl Dent Health Conf (U S)",
  ["[report no.] naec-acel-. united states. aerospace crew equipment laboratory, philadelphia"] = "Rep No NAEC ACEL United States Aerosp Crew Equip Lab Phila",
  ["[report no.] namc-acel-. united states. air crew equipment laboratory, philadelphia"] = "Rep No NAMC ACEL United States Air Crew Equip Lab Phila",
  ["[report]. civil aeromedical research institute (u.s.)"] = "Rep Civ Aeromed Res Inst US",
  ["[research reports]. naval medical field research laboratory (camp lejeune, n.c.)"] = "Res Rep U S Nav Med Field Res Lab",
  ["[rinsho eiyo] [clinical nutrition]"] = "Rinsho Eiyo",
  ["[rinsho fujinka sanka] clinical gynecology and obstetrics"] = "Rinsho Fujinka Sanka",
  ["[rinsho ketsueki] the japanese journal of clinical hematology"] = "Rinsho Ketsueki",
  ["[rinsho shika] [clinical dentistry]"] = "Rinsho Shika",
  ["[rinsho shokakibyogaku] clinical gastro-enterology"] = "Rinsho Shokakibyogaku",
  ["[rinshō eiyō] [clinical nutrition]"] = "Rinsho Eiyo",
  ["[rinshō fujinka sanka] clinical gynecology and obstetrics"] = "Rinsho Fujinka Sanka",
  ["[rinshō ketsueki] the japanese journal of clinical hematology"] = "Rinsho Ketsueki",
  ["[rinshō shika] [clinical dentistry]"] = "Rinsho Shika",
  ["[rinshō shōkakibyōgaku] clinical gastro-enterology"] = "Rinsho Shokakibyogaku",
  ["[rōdō kagaku kenkyū] the journal of science of labour"] = "Rodo Kagaku Kenkyu",
  ["[sanfujinka chiryo] obstetrical and gynecological therapy"] = "Sanfujinka Chiryo",
  ["[sanfujinka chiryō] obstetrical and gynecological therapy"] = "Sanfujinka Chiryo",
  ["[sanfujinka no shinpo] advances in obstetrics and gynecology"] = "Sanfujinka No Shinpo",
  ["[seitai no kagaku] [science of the living body]"] = "Seitai Kagaku",
  ["[sekai sanfujinka soran] [survey of world obstetrics and gynecology]"] = "Surv World Obstet Gynecol (Jpn)",
  ["[sekai sanfujinka sōran] [survey of world obstetrics and gynecology]"] = "Surv World Obstet Gynecol (Jpn)",
  ["[shandong yi kan] [shantung medical publication]"] = "Shandong Yi Kan",
  ["[shika igaku] [dental medicine]"] = "Shika Igaku",
  ["[shimane igaku] [shimane medicine]"] = "Shimane Igaku",
  ["[shinryo] [diagnosis and treatment]"] = "Shinryo",
  ["[shinryō] [diagnosis and treatment]"] = "Shinryo",
  ["[sogo rinsho] clinic all-round"] = "Sogo Rinsho",
  ["[sōgō rinshō] clinic all-round"] = "Sogo Rinsho",
  ["[technical report] sam-tr. usaf school of aerospace medicine"] = "Tech Rep SAM-TR",
  ["[tokyo iji shinshi] the tokyo medical journal"] = "Tokyo Iji Shinshi",
  ["[tsa chih] [journal of the] alumni of the university of takau. gaoxiong yi xue yuan. tong xue hui"] = "Tsa Chih Gaoxiong Yi Xue Yuan Tong Xue Hui",
  ["[wachira wēchasān] vajīra medical journal"] = "Wachira Wechasan",
  ["[xin zhong hua yi yao yue kan] [new china medical monthly]"] = "Xin Zhong Hua Yi Yao Yue Kan",
  ["[yakugaku kenkyu] japanese journal of pharmacy and chemistry"] = "Yakugaku Kenkyu",
  ["[yakugaku kenkyū] japanese journal of pharmacy and chemistry"] = "Yakugaku Kenkyu",
  ["[yan jiu bao gao] science reports of the national taiwan university. guo li taiwan da xue. xin li xue xi"] = "Yan Jiu Bao Gao",
  ["[yan ke xue shu hui kan] = ophthalmological series"] = "Yan ke xue shu hui kan",
  ["[ying yang xue bao] acta nutrimenta sinica"] = "Ying Yang Xue Bao",
  ["[zasshi] [journal]. fukuoka ika daigaku"] = "Zasshi Fukuoka Ika Daigaku",
  ["[zasshi] [journal]. nihon kyobu geka gakkai"] = "Nippon Kyobu Geka Gakkai Zasshi",
  ["[zasshi] [journal]. nihon kyōbu geka gakkai"] = "Nihon Kyobu Geka Gakkai Zasshi",
  ["[zasshi] journal. kanazawa daigaku. igakubu. juzen igakkai"] = "Zasshi Kanazawa Daigaku Igakubu Juzen Igakkai",
  ["[zasshi] journal. kanazawa daigaku. igakubu. jūzen igakkai"] = "Zasshi Kanazawa Daigaku Igakubu Juzen Igakkai",
  ["[zasshi] journal. kyoto furitsu ika daigaku"] = "Kyoto Furitsu Ika Daigaku Zasshi",
  ["[zasshi] tokyo jikeikai medical journal. tokyo jikeikai ika daigaku"] = "Tokyo Jikeikai Ika Daigaku Zasshi",
  ["[zasshi] tokyo jikeikai medical journal. tōkyō jikeikai ika daigaku"] = "Tokyo Jikeikai Ika Daigaku Zasshi",
  ["[zhonghua yan ke za zhi] chinese journal of ophthalmology"] = "Zhonghua Yan Ke Za Zhi",
  ["\\cyr biblioteka matematika"] = "Bibl. Mat.",
  ["\\cyr biblioteka studenta-matematika"] = "Bibl. Stud.-Mat.",
  ["\\cyr istoriki nauki rossii xx \\cyr veka"] = "Istor. Nauki Ross. XX Veka",
  ["\\cyr istoriko-astronomicheskie issledovaniya"] = "Istor.-Astronom. Issled.",
  ["\\cyr itogi nauki i tekhniki"] = "Itogi Nauki i Tekhniki",
  ["\\cyr kibernetika—neogranichennye vozmozhnosti i vozmozhnye ogranicheniya"] = "Kibern. Neogranich. Vozmozhn. Vozmozhn. Ogranich.",
  ["\\cyr matematika i ee prilozheniya"] = "Mat. Prilozh.",
  ["\\cyr metody matematicheskogo modelirovaniya"] = "Metody Mat. Model.",
  ["\\cyr nauchno-biograficheskaya literatura"] = "Nauchno-Biograf. Lit.",
  ["\\cyr novoe v global\\cprime nom analize"] = "Novoe Global. Anal.",
  ["\\cyr optimizatsiya i issledovanie operatsiy"] = "Optim. Issled. Oper.",
  ["\\cyr prikladnye matematicheskie modeli"] = "Prikl. Mat. Model.",
  ["\\cyr sibirskaya shkola algebry i logiki"] = "Sib. Shk. Algebry Log.",
  ["\\cyr soobshcheniya ob\\cdprime edinennogo instituta yadernykh issledovaniy"] = "Soobshch. Ob\\cdprime ed. Inst. Yadernykh Issled. Dubna",
  ["\\cyr soobshcheniya po prikladnoy matematike"] = "Soobshch. Prikl. Mat.",
  ["\\cyr sovremennaya matematika—studentam i aspirantam"] = "Sovrem. Mat.—Stud. Aspir.",
  ["\\cyr sovremennye problemy fiziki"] = "Sovrem. Probl. Fiz.",
  ["\\cyr sovremennye problemy matematiki"] = "Sovrem. Probl. Mat. Fund. Naprav.",
  ["\\cyr svyaz\\cprime vremen"] = "Svyaz\\cprime Vremen",
  ["\\cyr teoriya i metody sistemnogo analiza"] = "Teor. Metod. Sistem. Anal.",
  ["\\cyr teoriya veroyatnostey i matematicheskaya statistika"] = "Teor. Veroyatnost. Mat. Statist.",
  ["\\cyr trudy instituta matematiki"] = "Trudy Inst. Mat.",
  ["\\cyr trudy instituta obshchey fiziki"] = "Trudy IOFAN",
  ["\\cyr v pomoshch\\cprime abiturientu"] = "Pomoshch\\cprime Abitur.",
  ["\\cyr voprosy mekhaniki i protsessov upravleniya"] = "Voprosy Mekh. Protsess. Upravl.",
  ["\\cyr zhivoy yazyk"] = "Zhivoy Yazyk",
  ["a & a case reports"] = "A A Case Rep",
  ["a bradford book"] = "Bradford Book",
  ["a chartwell-bratt student text"] = "Chartwell-Bratt Student Text",
  ["a common place"] = "Common Place",
  ["a comprehensive course in analysis"] = "Compr. Course Anal.",
  ["a crianca portuguesa"] = "Crianca Port",
  ["a criança portuguesa"] = "Crianca Port",
  ["a current bibliography on african affairs"] = "Curr Bibliogr Afr Aff",
  ["a debreceni déri múzeum évkönyve"] = "DebrecMuzEvk",
  ["a gyógyszerész"] = "Gyogyszeresz",
  ["a henry holt reference book"] = "Henry Holt Ref. Book",
  ["a journal of church and state"] = "J Church State",
  ["a journal of jewish medical ethics and halacha"] = "J Jew Med Ethics Halacha",
  ["a listing of research in the cardiovascular field"] = "Listing Res Cardiovasc Field",
  ["a móra ferenc múzeum évkönyve"] = "MuzEvkSzeged",
  ["a national bureau of economic research monograph"] = "NBER Monogr.",
  ["a pediatria; revista mensal de clinica, patologia e higiene infantil"] = "Pediatria (Rio)",
  ["a reston book"] = "Reston Book",
  ["a series of student texts in contemporary biology"] = "Ser. Students Texts Contemp. Biology",
  ["a spectrum book"] = "Spectrum Book",
  ["a wiley press book"] = "Wiley Press Book",
  ["a wiley-interscience publication"] = "Wiley-Intersci. Publ.",
  ["a&a practice"] = "A A Pract",
  ["a&m : advertising and marketing"] = "A M",
  ["a. d. trendall, the red-figured vases of lucania, campania and sicily (oxford 1967–­1983)"] = "LCS",
  ["a. furtwängler – k. reichhold, griechische vasenmalerei (münchen 1900–1925)"] = "FR",
  ["a. m. a. archives of dermatology"] = "AMA Arch Derm",
  ["a. m. a. archives of dermatology and syphilology"] = "AMA Arch Derm Syphilol",
  ["a. m. a. archives of industrial hygiene and occupational medicine"] = "A M A Arch Ind Hyg Occup Med",
  ["a. m. a. archives of neurology"] = "AMA Arch Neurol",
  ["a. m. a. archives of neurology and psychiatry"] = "AMA Arch Neurol Psychiatry",
  ["a. m. a. archives of pathology"] = "AMA Arch Pathol",
  ["a.a.g. bijdragen"] = "A A G Bijdr",
  ["a.a.g. bijdragen / afdeling agrarische geschiedenis, landbouwhogeschool"] = "A A G Bijdr",
  ["a.i.d. evaluation news"] = "AID Eval News",
  ["a.m.a. american journal of diseases of children"] = "AMA Am J Dis Child",
  ["a.m.a. archives of dermatology"] = "AMA Arch Derm",
  ["a.m.a. archives of dermatology and syphilology"] = "AMA Arch Derm Syphilol",
  ["a.m.a. archives of general psychiatry"] = "AMA Arch Gen Psychiatry",
  ["a.m.a. archives of industrial health"] = "AMA Arch Ind Health",
  ["a.m.a. archives of industrial hygiene and occupational medicine"] = "AMA Arch Ind Hyg Occup Med",
  ["a.m.a. archives of internal medicine"] = "AMA Arch Intern Med",
  ["a.m.a. archives of neurology"] = "AMA Arch Neurol",
  ["a.m.a. archives of neurology and psychiatry"] = "AMA Arch Neurol Psychiatry",
  ["a.m.a. archives of ophthalmology"] = "AMA Arch Ophthalmol",
  ["a.m.a. archives of otolaryngology"] = "AMA Arch Otolaryngol",
  ["a.m.a. archives of pathology"] = "AMA Arch Pathol",
  ["a.m.a. archives of surgery"] = "AMA Arch Surg",
  ["a.m.a. journal of diseases of children"] = "AMA J Dis Child",
  ["a.r.a.g. national newsletter"] = "ARAG Natl Newsl",
  ["a.r.a.g. national newsletter / abortion reform action group"] = "ARAG Natl Newsl",
  ["a.w.r. bulletin. association for the study of the world refugee problem"] = "AWR Bull",
  ["a/z : itu journal of faculty of architecture"] = "AZ",
  ["aa : l'architecture d'aujourd'hui"] = "Archit Aujourdhui",
  ["aa : l’architecture d’aujourd’hui"] = "Archit Aujourdhui",
  ["aace clinical case reports"] = "AACE Clin Case Rep",
  ["aachener kunstblatter (dumont buchverlag)"] = "Aachen Kunstbl",
  ["aachener kunstblätter (dumont buchverlag)"] = "Aachen Kunstbl",
  ["aacn advanced critical care"] = "AACN Adv Crit Care",
  ["aacn clinical issues"] = "AACN Clin Issues",
  ["aacn clinical issues in critical care nursing"] = "AACN Clin Issues Crit Care Nurs",
  ["aacr meeting abstracts online. american association for cancer research. annual meeting"] = "AACR Meet Abstr Online",
  ["aade editors journal"] = "AADE Ed. J.",
  ["aade editors' journal"] = "AADE Ed J",
  ["aade editors’ journal"] = "AADE Ed J",
  ["aade in practice"] = "AADE Pract",
  ["aahs journal"] = "AAHS J",
  ["aahs journal / american aviation historical society"] = "AAHS J",
  ["aana journal"] = "AANA J",
  ["aannt journal"] = "AANNT J",
  ["aannt journal / the american association of nephrology nurses and technicians"] = "AANNT J",
  ["aaohn journal"] = "AAOHN J.",
  ["aaohn journal : official journal of the american association of occupational health nurses"] = "AAOHN J",
  ["aap news"] = "AAP news",
  ["aapg bulletin"] = "AAPG Bull.",
  ["aapg memoir"] = "AAPG Mem.",
  ["aapg memoirs"] = "Aapg Memoir.",
  ["aapi nexus : asian americans and pacific islanders, policy practice and community"] = "AAPI Nexus",
  ["aapl newsletter"] = "AAPL Newsl",
  ["aapm report"] = "AAPM Rep",
  ["aappo journal"] = "AAPPO J.",
  ["aappo journal : the journal of the american association of preferred provider organizations"] = "AAPPO J",
  ["aapps bulletin"] = "AAPPS Bull.",
  ["aaps journal"] = "AAPS J.",
  ["aaps open"] = "AAPS Open",
  ["aaps pharmsci"] = "AAPS PharmSci",
  ["aaps pharmscitech"] = "AAPS PharmSciTech",
  ["aarbøger for nordisk oldkyndighed og historie"] = "AarbKøb",
  ["aarn news letter"] = "AARN News Lett",
  ["aarsberetning - kobenhavns universitets medicinsk-historiske institut og museum"] = "Arsberet Kbh Univ Med Hist Inst Mus",
  ["aarsberetning - kobenhavns universitets medicinsk-historiske museum"] = "Arsberet Kobenhavns Univ Med Hist Mus",
  ["aarsberetning - københavns universitets medicinsk-historiske institut og museum"] = "Arsberet Kbh Univ Med Hist Inst Mus",
  ["aarsberetning - københavns universitets medicinsk-historiske museum"] = "Arsberet Kobenhavns Univ Med Hist Mus",
  ["aarsskrift - sydsvenska medicinhistoriska sällskapet"] = "Arsskr Sydven Medicinhist Sallsskapet",
  ["aas open research"] = "AAS Open Res",
  ["aastaraamat - eesti teadusliku seltsi rootsis. annales - societatis litterarum estonicae in svecia. eesti teaduslik selts rootsis"] = "Aastaraam Eesti Tead Seltsi Rootsis",
  ["aastaraamat. eesti matemaatika selts"] = "Aastaraam. Eesti Mat. Selts",
  ["aatcc journal of research"] = "AATCC J. Res.",
  ["aatcc review"] = "AATCC Rev",
  ["aaup bulletin : quarterly publication of the american association of university professors"] = "AAUP Bull",
  ["ab bookman's weekly : for the specialist book world"] = "AB Bookm Wkly",
  ["ab bookman’s weekly : for the specialist book world"] = "AB Bookm Wkly",
  ["ab-original : journal of indigenous studies and first nations' and first peoples' culture"] = "AbOrig",
  ["aba health esource"] = "ABA Health eSource",
  ["aba journal"] = "ABA J",
  ["abb review"] = "ABB Rev.",
  ["abc decor"] = "ABC Decor",
  ["abc décor"] = "ABC Decor",
  ["abdominal imaging"] = "Abdom Imaging",
  ["abdominal imaging : computation and clinical applications : 5th international workshop, held in conjunction with miccai 2013, nagoya, japan, september 22, 2013 : proceedings. abdominal imaging (workshop) (5th : 2013 : nagoya-shi, japan)"] = "Abdom Imaging (2013)",
  ["abdominal imaging : computational and clinical applications : 6th international workshop, abdi 2014, held in conjunction with miccai 2014, cambridge, ma, usa, september 14, 2014. abdi (workshop) (6th : 2014 : cambridge, mass.)"] = "Abdom Imaging (2014)",
  ["abdominal radiology (new york)"] = "Abdom Radiol (NY)",
  ["abdominal surgery"] = "Abdom Surg",
  ["abel symposia"] = "Abel Symp.",
  ["aberdeen university review"] = "Aberdeen Univ Rev",
  ["abhandlung zur erlangung der wurde eines doktors der naturwissenschaften der eidgenossischen technischen hochschule zurich"] = "Abhandlung zur Erlangung der Wurde eines Doktors der Naturwissenschaften der Eidgenossischen Technischen Hochschule",
  ["abhandlung zur erlangung der wurde eines doktors der technischen wissenschaften der eidgenossischen technischen hochschule zurich"] = "Abhandlung zur Erlangung der Wurde eines Doktors der technischen Wissenschaften der Eidgenossischen Technischen Hoch",
  ["abhandlungen aus dem mathematischen seminar der universitat hamburg"] = "Abh. Math. Semin. Univ. Hambg.",
  ["abhandlungen aus dem mathematischen seminar der universität hamburg"] = "Abh. Math. Semin. Univ. Hamburg",
  ["abhandlungen aus dem westfälischen museum für naturkunde"] = "Abh Westfal Mus Naturkunde",
  ["abhandlungen aus dem westfälischen museum für naturkunde"] = "Abh. Westfäl. Mus. Nat.kd.",
  ["abhandlungen der akademie der wissenschaften der ddr"] = "Abh. Akad. Wiss. DDR",
  ["abhandlungen der akademie der wissenschaften der ddr, abteilung mathematik, naturwissenschaften, technik"] = "Abh. Akad. Wiss. DDR, Abt. Math., Naturwiss., Tech.",
  ["abhandlungen der akademie der wissenschaften in hamburg"] = "Abh. Akad. Wiss. Hamburg",
  ["abhandlungen der akademie der wissenschaften in mainz, geistes und sozialwissenschaftliche klasse"] = "AAWM",
  ["abhandlungen der akademie der wissenschaften zu gottingen. neue folge"] = "Abh. Akad. Wiss. Gottingen Neue Folge",
  ["abhandlungen der akademie der wissenschaften zu göttingen. philologisch-historische klasse"] = "AbhGoettingen",
  ["abhandlungen der bayerischen akademie der wissenschaften, philos.-hist. klasse"] = "ABAW",
  ["abhandlungen der braunschweigischen wissenschaftlichen gesellschaft"] = "Abh. Braunschweig. Wiss. Ges.",
  ["abhandlungen der deutschen akademie der wissenschaften zu berlin"] = "AbhBerlin",
  ["abhandlungen der deutschen orient-gesellschaft"] = "ADOG",
  ["abhandlungen der geistes- und sozialwissenschaftlichen klasse"] = "Abh Geistes Sozwiss Kl  Akad Wiss Lit",
  ["abhandlungen der heidelberger akademie der wissenschaften, philos.- hist. klasse"] = "AHAW",
  ["abhandlungen der mathematisch-naturwissenschaftlichen klasse"] = "Abh. Math.-Naturwiss. Kl. Akad. Wiss. Lit. Mainz",
  ["abhandlungen der naturforschenden gesellschaft in zürich"] = "Abh. Nat.forsch. Ges. Zür.",
  ["abhandlungen der rheinisch-westfälischen akademie der wissenschaften"] = "AbhDüsseldorf",
  ["abhandlungen der sachsischen akademie der wissenschaften zu leipzig. mathematisch-naturwissenschaftliche klasse"] = "Abh. Sachs. Akad. Wiss. Leipzig Math.-Nat.wiss. Kl.",
  ["abhandlungen der senckenbergischen naturforschenden gesellschaft"] = "Abh. Senckb. Naturforsch. Ges.",
  ["abhandlungen der sächsischen akademie der wissenschaften zu leipzig"] = "ASAW",
  ["abhandlungen der sächsischen akademie der wissenschaften zu leipzig. philologisch-historische klasse"] = "AbhLeipzig",
  ["abhandlungen des deutschen archäologischen instituts, abteilung kairo"] = "ADAIK",
  ["abhandlungen des naturwissenschaftlichen vereins zu bremen"] = "Abh. Nat.wiss. Ver. Bremen",
  ["abhandlungen und berichte (deutsches museum (germany))"] = "Abh Ber Dtsch Mus",
  ["abhandlungen und berichte des staatlichen museums fur volkerkunde, dresden. staatliches museum fur volkerkunde dresden"] = "Abh Ber Staatl Mus Voelkerkd Dres",
  ["abhandlungen und berichte des staatlichen museums für völkerkunde, dresden. staatliches museum für völkerkunde dresden"] = "Abh Ber Staatl Mus Voelkerkd Dres",
  ["abhandlungen zur geschichte der medizin und der naturwissenschaften"] = "Abh Gesch Med Naturwiss",
  ["abhandlungen zur philosophie und ihrer geschichte"] = "Abh. Philos. Gesch.",
  ["abhb. annual bibliography of the history of the printed book and libraries"] = "Annu Bibliogr Hist Print Bool Libr",
  ["abi technik"] = "ABI Tech.",
  ["abnf journal"] = "ABNF J.",
  ["aboriginal and islander health worker journal"] = "Aborig Isl Health Work J",
  ["aboriginal health bulletin"] = "Aborig Health Bull",
  ["aboriginal history"] = "Aborig Hist",
  ["abortion research notes"] = "Abort Res Notes",
  ["abortion research notes / international reference center for abortion research"] = "Abort Res Notes",
  ["abortion review"] = "Abortion Rev",
  ["abr-nahrain : an annual under the auspices of the department of semitic studies, university of melbourne"] = "Abr Nahain",
  ["abr-nahrain. an annual published by the department of middle eastern studies, university of melbourne"] = "Abr-Nahrain",
  ["abs international guide to classical studies"] = "Int Guide Classical Stud",
  ["abs quarterly check-list of classical studies"] = "Q Check List Classical Stud",
  ["abstract and applied analysis"] = "Abstr. Appl. Anal.",
  ["abstract book, 2003 national hiv prevention conference, july 27-30, 2003, hyatt regency atlanta hotel, atlanta, georgia. national hiv prevention conference (2003 : atlanta, ga.)"] = "Abstr Book 2003 Natl HIV Prev Conf July 27 30 2003 Hyatt Regency Atlanta Hotel Atlanta Ga Natl HIV Prev Conf 2003 Atlanta Ga",
  ["abstract book, 2005 national hiv prevention conference, june 12-15, 2005, hyatt regency atlanta hotel, atlanta, georgia. national hiv prevention conference (2005 : atlanta, ga.)"] = "Abstr Book 2005 Natl HIV Prev Conf June 12 15 2005 Hyatt Regency Atlanta Hotel Atlanta Ga Natl HIV Prev Conf 2005 Atlanta Ga",
  ["abstract search tools (for the 1999 national hiv prevention conference). national hiv prevention conference (1999 : atlanta, ga.)"] = "Abstr Search Tools 1999 Natl HIV Prev Conf Natl HIV Prev Conf 1999 Atlanta Ga",
  ["abstracta botanica"] = "Abstr. Bot.",
  ["abstracts - society for neuroscience. society for neuroscience"] = "Abstr Soc Neurosci",
  ["abstracts 10th conference on retroviruses and opportunistic infections, february 10-14, 2003, hynes convention center, boston, massachusetts, usa. conference on retroviruses and opportunistic infections (10th : 2003: boston, mass.)"] = "Abstr 10th Conf Retrovir Oppor Infect Feb 10 14 2003 Hynes Conv Cent Boston Mass USA Conf Retrovir Oppor Infect 10th 2003 Boston Mass",
  ["abstracts and translations from the science library. institute of living"] = "Abstr Translat Sci Library",
  ["abstracts of papers - american chemical society"] = "Abstr Pap Am Chem Soc",
  ["abstracts of papers presented to the american mathematical society"] = "Abstracts Amer. Math. Soc.",
  ["abstracts of the annual meeting of the american society for microbiology"] = "Abstr Annu Meet Am Soc Microbiol",
  ["abstracts. international congress on tropical medicine and malaria (4th : 1948 : washington, d. c.)"] = "Abstr Int Congr Trop Med Malar",
  ["aca transactions"] = "ACA Trans.",
  ["academia brasileira de filosofia"] = "Acad. Bras. Filos.",
  ["academia colombiana de ciencias exactas, fisicas y naturales. revista"] = "Rev. Acad. Colombiana Cienc. Exact. Fis. Natur.",
  ["academia colombiana de ciencias exactas, físicas y naturales"] = "Rev. Acad. Colombiana Cienc. Exact. Fís. Natur.",
  ["academia economic papers"] = "Academia Econ. Pap.",
  ["academia journal of agricultural research"] = "Acad J Agric Res",
  ["academia peruana de cirugia"] = "Acad. Peru. Cir.",
  ["academia peruana de cirugía"] = "Acad Peru Cir",
  ["academia romana. memoriile sectiilor stiintifice. seria iv. memoirs of the scientific sections. romanian academy. series iv"] = "Mem. Sect. Stiint. Acad. Romana Ser. IV",
  ["academia română"] = "Mem. Secţ. Ştiinţ. Acad. Română Ser. IV",
  ["academiae scientiarum fennicae. annales. mathematica. dissertationes"] = "Ann. Acad. Sci. Fenn. Math. Diss.",
  ["academic emergency medicine"] = "Acad. Emerg. Med.",
  ["academic emergency medicine : official journal of the society for academic emergency medicine"] = "Acad Emerg Med",
  ["academic forensic pathology"] = "Acad Forensic Pathol",
  ["academic journal of pediatric and neonatology"] = "Acad J Pediatr Neonatol",
  ["academic journal of xia̕n jiaotong university : ajxjtu = xi'an jiao tong da xue xue bao (yi wen ban)"] = "Acad J Xian Jiaotong Univ",
  ["academic medicine"] = "Acad. Med.",
  ["academic medicine : journal of the association of american medical colleges"] = "Acad Med",
  ["academic nurse"] = "Acad. Nurse",
  ["academic pathology"] = "Acad Pathol",
  ["academic pediatrics"] = "Acad Pediatr",
  ["academic press series in cognition and perception"] = "Academic Press Ser. Cognition and Perception",
  ["academic psychiatry : the journal of the american association of directors of psychiatric residency training and the association for academic psychiatry"] = "Acad Psychiatry",
  ["academic psychology bulletin"] = "Academic Psychol Bull",
  ["academic radiology"] = "Acad Radiol",
  ["academiæ scientiarum fennicae. annales. mathematica. dissertationes"] = "Ann. Acad. Sci. Fenn. Math. Diss.",
  ["academiæ scientiarum fennicæ"] = "Ann. Acad. Sci. Fenn. Math. Diss.",
  ["academy news (milwaukee, wis.)"] = "Acad News",
  ["academy of management journal"] = "Acad. Manage. J.",
  ["academy of management journal. academy of management"] = "Acad Manage J",
  ["academy of management review"] = "Acad. Manage. Rev.",
  ["academy of management review. academy of management"] = "Acad Manage Rev",
  ["academy of science of the people’s democratic republic of korea"] = "Su-hak",
  ["academy of sciences of moldova"] = "Comput. Sci. J. Moldova",
  ["academy of sciences of moldova. institute of mathematics and computer science. computer science journal of moldova"] = "Comput. Sci. J. Moldova",
  ["academy of sciences of the czech republic"] = "Math. Bohem.",
  ["academy of sciences of the czech republic. mathematical institute. mathematica bohemica"] = "Math. Bohem.",
  ["academy review of the california academy of periodontology"] = "Acad. Rev. Calif. Acad. Periodontol.",
  ["academy review of the california academy of periodontology, united states section, arpa internationale"] = "Acad Rev Calif Acad Periodontol",
  ["académie des inscriptions et belles-lettres. comptes rendus des séances de l’académie"] = "CRAI",
  ["académie roumaine"] = "Mathematica (Cluj)",
  ["académie royale de belgique"] = "Acad. Roy. Belg. Bull. Cl. Sci. (6)",
  ["académie serbe des sciences et des arts"] = "Acad. Serbe Sci. Arts Glas",
  ["acarina : russkiǐ akarologicheskiǐ zhurnal + russian journal of acarology"] = "Acarina",
  ["acc current journal review"] = "ACC Curr J Rev",
  ["accademia delle scienze di torino"] = "Mem. Accad. Sci. Torino Cl. Sci. Fis. Mat. Natur. (5)",
  ["accademia ligure di scienze e lettere"] = "Atti Accad. Ligure Sci. Lett. (5)",
  ["accademia medica"] = "Accad Medica",
  ["accademia nazionale dei lincei. fondazione leone caetani"] = "Accad Naz Lincei Rome",
  ["accademia nazionale delle scienze detta dei xl"] = "Rend. Accad. Naz. Sci. XL Mem. Mat. Appl. (5)",
  ["accademia peloritana dei pericolanti"] = "Atti Accad. Peloritana Pericolanti Cl. Sci. Fis. Mat. Natur.",
  ["accademie e biblioteche d'italia"] = "Accad Bibl Ital",
  ["accademie e biblioteche d’italia"] = "AcBibl",
  ["accelerator physics"] = "Accel. Phys.",
  ["access microbiology"] = "Access Microbiol",
  ["accident analysis and prevention"] = "Accid. Anal. Prev.",
  ["accident and emergency nursing"] = "Accid Emerg Nurs",
  ["accident; analysis and prevention"] = "Accid Anal Prev",
  ["acción sindical. número científico"] = "Num Cient Accion Sind",
  ["accountability in research"] = "Account Res",
  ["accounting review"] = "Acc. Rev.",
  ["accounting, business and financial history"] = "Acc.",
  ["accounts of chemical research"] = "Acc Chem Res",
  ["accounts of materials research"] = "Acc Mater Res",
  ["accreditation and quality assurance"] = "Accredit. Qual. Assur.",
  ["accreditation and quality assurance : journal for quality, comparability and reliability in chemical measurement"] = "Accredit Qual Assur",
  ["accès sciences"] = "Accès Sci.",
  ["acção médica"] = "Accao Med",
  ["ace quarterly"] = "ACE Q",
  ["aces bulletin"] = "ACES Bull.",
  ["acetic acid bacteria"] = "Acetic Acid Bact",
  ["acg case reports journal"] = "ACG Case Rep J",
  ["ach - models in chemistry"] = "ACH - Models Chem.",
  ["achievements and solutions in mechanical engineering ii : selected, peer reviewed papers from the 5th international conference of mechanical engineering (icome) 2019, october 24-25, 2019, craiova, romania. icome (conference) (5th : 2019 : craiova, romania)"] = "Achiev Solut Mech Eng II (2019)",
  ["achievements in the life sciences"] = "Achiev. Life Sci.",
  ["achse, rad und wagen. beiträge zur geschichte der landfahrzeuge"] = "Achse",
  ["aci materials journal"] = "ACI Mater. J.",
  ["aci open"] = "ACI open",
  ["aci structural journal"] = "ACI Struct J",
  ["aci2017, improving relations : fourth international conference on animal-computer interaction : proceedings : 21-23 november 2017, milton keynes, united kingdom. international conference on animal-computer interaction (4th : 2017 : milton keynes, england)"] = "ACI 2017 Improv Relat (2017)",
  ["acm ai letters"] = "ACM AI Lett.",
  ["acm communications in computer algebra"] = "ACM Commun. Comput. Algebra",
  ["acm computing surveys"] = "ACM Comput Surv",
  ["acm digital threats: research and practice"] = "ACM Digital Threats: Res. Pract.",
  ["acm distinguished theses"] = "ACM Disting. Theses",
  ["acm inroads"] = "ACM Inroads",
  ["acm journal of data and information quality"] = "ACM J. Data Inf. Qual.",
  ["acm journal of experimental algorithmics"] = "ACM J. Exp. Algorithmics",
  ["acm journal on autonomous transportation systems"] = "ACM J. Auton. Transp. Syst.",
  ["acm journal on computing and cultural heritage"] = "ACM J. Comput. Cult. Heritage",
  ["acm journal on computing and sustainable societies"] = "ACM J. Comput. Sustainable Soc.",
  ["acm journal on emerging technologies in computing"] = "ACM J. Emerging Technol. Comput.",
  ["acm journal on emerging technologies in computing systems"] = "ACM J Emerg Technol Comput Syst",
  ["acm journal on responsible computing"] = "ACM J. Responsible Comput.",
  ["acm queue"] = "ACM Queue",
  ["acm queue : tomorrow's computing today"] = "ACM Queue",
  ["acm queue : tomorrow’s computing today"] = "ACM Queue",
  ["acm sigapp applied computing review : a publication of the special interest group on applied computing"] = "ACM SIGAPP Appl Comput Rev",
  ["acm transactions on accessible computing"] = "ACM Trans. Accessible Comput.",
  ["acm transactions on ai for science"] = "ACM Trans. AI Sci.",
  ["acm transactions on algorithms"] = "ACM Trans. Algorithms",
  ["acm transactions on applied perception"] = "ACM Trans Appl Percept",
  ["acm transactions on architecture and code optimization"] = "ACM Trans. Archit. Code Optim.",
  ["acm transactions on asian and low-resource language information processing"] = "ACM Trans. Asian Low-Resour. Lang. Inf. Process.",
  ["acm transactions on asian language information processing"] = "ACM Trans. Asian Lang. Inf. Process.",
  ["acm transactions on autonomous and adaptive systems"] = "ACM Trans. Auton. Adapt. Syst.",
  ["acm transactions on computation theory"] = "ACM Trans. Comput. Theory",
  ["acm transactions on computational logic"] = "ACM Trans. Comput. Log.",
  ["acm transactions on computer systems"] = "ACM Trans. Comput. Syst.",
  ["acm transactions on computer-human interaction"] = "ACM Trans. Comput.-Hum. Interact.",
  ["acm transactions on computer-human interaction : a publication of the association for computing machinery"] = "ACM Trans Comput Hum Interact",
  ["acm transactions on computing education"] = "ACM Trans. Comput. Educ.",
  ["acm transactions on computing for healthcare"] = "ACM Trans Comput Healthc",
  ["acm transactions on cyber-physical systems"] = "ACM Trans. Cyber-Phys. Syst.",
  ["acm transactions on database systems"] = "ACM Trans. Database Syst.",
  ["acm transactions on design automation of electronic systems"] = "ACM Transact Des Autom Electron Syst",
  ["acm transactions on economics and computation"] = "ACM Trans. Econ. Comput.",
  ["acm transactions on embedded computing systems"] = "ACM Trans. Embedded Comput. Syst.",
  ["acm transactions on embedded computing systems : tecs"] = "ACM Trans Embed Comput Syst",
  ["acm transactions on evolutionary learning"] = "ACM Trans. Evol. Learn.",
  ["acm transactions on evolutionary learning and optimization"] = "ACM Trans. Evol. Learn. Optim.",
  ["acm transactions on graphics"] = "ACM Trans Graph",
  ["acm transactions on human-robot interaction"] = "ACM Trans Hum Robot Interact",
  ["acm transactions on information and system security"] = "ACM Trans. Inf. Syst. Secur.",
  ["acm transactions on information systems"] = "ACM Trans Inf Syst",
  ["acm transactions on intelligent systems and technology"] = "ACM Trans. Intell. Syst. Technol.",
  ["acm transactions on interactive intelligent systems"] = "ACM Trans. Interact. Intell. Syst.",
  ["acm transactions on internet of things"] = "ACM Trans. Internet Things",
  ["acm transactions on internet technology"] = "ACM Trans Internet Technol",
  ["acm transactions on knowledge discovery from data"] = "ACM Trans. Knowl. Discovery Data",
  ["acm transactions on management information systems"] = "ACM Trans. Manage. Inf. Syst.",
  ["acm transactions on mathematical software"] = "ACM Trans. Math. Software",
  ["acm transactions on mathematical software. association for computing machinery"] = "ACM Trans Math Softw",
  ["acm transactions on modeling and computer simulation"] = "ACM Trans. Model. Comput. Simul.",
  ["acm transactions on modeling and computer simulation : a publication of the association for computing machinery"] = "ACM Trans Model Comput Simul",
  ["acm transactions on modeling and performance evaluation of computing systems"] = "ACM Trans. Model. Perform. Eval. Comput. Syst.",
  ["acm transactions on multimedia computing communications and applications"] = "ACM Trans. Multimedia Comput. Commun. Appl.",
  ["acm transactions on parallel computing"] = "ACM Trans. Parallel Comput.",
  ["acm transactions on privacy and security"] = "ACM Trans. Privacy Secur.",
  ["acm transactions on probabilistic machine learning"] = "ACM Trans. Probab. Mach. Learn.",
  ["acm transactions on programming languages and systems"] = "ACM Trans. Program. Lang. Syst.",
  ["acm transactions on quantum computing"] = "ACM Trans. Quantum Comput.",
  ["acm transactions on recommender systems"] = "ACM Trans. Recommender Syst.",
  ["acm transactions on reconfigurable technology and systems"] = "ACM Trans. Reconfigurable Technol. Syst.",
  ["acm transactions on sensor networks"] = "ACM Trans Sens Netw",
  ["acm transactions on social computing"] = "ACM Trans. Social Comput.",
  ["acm transactions on software engineering and methodology"] = "ACM Trans. Software Eng. Methodol.",
  ["acm transactions on spatial algorithms and systems"] = "ACM Trans. Spatial Algorithms Syst.",
  ["acm transactions on speech and language processing"] = "ACM Trans. Speech Lang. Process.",
  ["acm transactions on storage"] = "ACM Trans. Storage",
  ["acm transactions on the web"] = "ACM Trans. Web",
  ["acm/ims journal of data science"] = "ACM/IMS J. Data Sci.",
  ["acm/ims transactions on data science"] = "ACM/IMS Trans. Data Sci.",
  ["acme. annali della facoltà di lettere e filosofia dell’università degli studi di milano"] = "Acme",
  ["acog clinical review"] = "ACOG Clin Rev",
  ["acog committee opinion"] = "ACOG Comm Opin",
  ["acog educational bulletin"] = "ACOG Educ Bull",
  ["acog nurses bulletin"] = "Acog Nurse Bull",
  ["acog technical bulletin"] = "ACOG Tech Bull",
  ["acontia. revista de arqueología"] = "Acontia",
  ["acoustical imaging"] = "Acoust. Imaging",
  ["acoustical physics"] = "Acoust. Phys.",
  ["acoustical science and technology"] = "Acoust. Sci. Technol.",
  ["acoustics (basel, switzerland)"] = "Acoustics (Basel)",
  ["acoustics australia"] = "Acoust Aust",
  ["acoustics australia / australian acoustical society"] = "Acoust Aust",
  ["acoustics research letters online"] = "Acoust. Res. Lett. Online",
  ["acoustics research letters online : arlo"] = "Acoust Res Lett Online",
  ["acoustics today"] = "Acoust. Today",
  ["acp journal club"] = "ACP J Club",
  ["acquired immunodeficiency syndrome (aids) weekly surveillance report"] = "Acquir Immunodefic Syndr (AIDS) Wkly Surveill Rep",
  ["acquired immunodeficiency syndrome (aids) weekly surveillance report / aids activity, center for infectious diseases, centers for disease control"] = "Acquir Immunodefic Syndr (AIDS) Wkly Surveill Rep",
  ["acquisitions medicales recentes"] = "Acquis. Med. Recent.",
  ["acquisitions médicales récentes"] = "Acquis Med Recent",
  ["acr open rheumatology"] = "ACR Open Rheumatol",
  ["acrh [reports]. u.s. atomic energy commission"] = "ACRH Rep",
  ["acrh reports"] = "ACRH Rep.",
  ["across the board"] = "Across Board (NY)",
  ["acs agricultural science & technology"] = "ACS Agric. Sci. Technol.",
  ["acs applied bio materials"] = "ACS Appl Bio Mater",
  ["acs applied electronic materials"] = "ACS Appl Electron Mater",
  ["acs applied energy materials"] = "ACS Appl. Energy Mater.",
  ["acs applied engineering materials"] = "ACS Appl. Eng. Mater.",
  ["acs applied materials & interfaces"] = "ACS Appl. Mater. Interfaces",
  ["acs applied nano materials"] = "ACS Appl. Nano Mater.",
  ["acs applied optical materials"] = "ACS Appl. Opt. Mater.",
  ["acs applied polymer materials"] = "ACS Appl Polym Mater",
  ["acs bio & med chem au"] = "ACS Bio & Med Chem Au",
  ["acs biomaterials science & engineering"] = "ACS Biomater Sci Eng",
  ["acs case reviews in surgery"] = "ACS Case Rev Surg",
  ["acs catalysis"] = "ACS Catal.",
  ["acs central science"] = "ACS Cent Sci",
  ["acs chemical biology"] = "ACS Chem Biol",
  ["acs chemical health and safety"] = "ACS Chem. Health Saf.",
  ["acs chemical neuroscience"] = "ACS Chem Neurosci",
  ["acs combinatorial science"] = "ACS Comb Sci",
  ["acs earth & space chemistry"] = "ACS Earth Space Chem",
  ["acs earth and space chemistry"] = "ACS Earth Space Chem.",
  ["acs electrochemistry"] = "ACS Electrochem.",
  ["acs energy letters"] = "ACS Energy Lett.",
  ["acs engineering au"] = "ACS Eng. Au",
  ["acs environmental au"] = "ACS Environ. Au",
  ["acs es&t air"] = "ACS ES&T Air",
  ["acs es&t engineering"] = "ACS ES&T Eng.",
  ["acs es&t water"] = "ACS ES&T Water",
  ["acs food science & technology"] = "ACS Food Sci. Technol.",
  ["acs infectious diseases"] = "ACS Infect Dis",
  ["acs macro letters"] = "ACS Macro Lett.",
  ["acs materials au"] = "ACS Mater. Au",
  ["acs materials letters"] = "ACS Mater Lett",
  ["acs measurement science au"] = "ACS Meas. Sci. Au",
  ["acs medicinal chemistry letters"] = "ACS Med Chem Lett",
  ["acs nano"] = "ACS Nano",
  ["acs nanoscience au"] = "ACS Nanosci. Au",
  ["acs omega"] = "ACS Omega",
  ["acs organic & inorganic au"] = "ACS Org. Inorg. Au",
  ["acs pharmacology & translational science"] = "ACS Pharmacol Transl Sci",
  ["acs photonics"] = "ACS Photonics",
  ["acs physical chemistry au"] = "ACS Phys. Chem. Au",
  ["acs polymers au"] = "ACS Polym. Au",
  ["acs sensors"] = "ACS Sens",
  ["acs sustainable chemistry & engineering"] = "ACS Sustainable Chem. Eng.",
  ["acs sustainable chemistry and engineering"] = "ACS Sustainable Chem. Eng.",
  ["acs sustainable resource management"] = "ACS Sustainable Resour. Manage.",
  ["acs symposium series"] = "ACS Symp. Ser.",
  ["acs symposium series. american chemical society"] = "ACS Symp Ser Am Chem Soc",
  ["acs synthetic biology"] = "ACS Synth Biol",
  ["acsm's health & fitness journal"] = "ACSMs Health Fit J",
  ["acta - unio internationalis contra cancrum"] = "Acta Unio Int Contra Cancrum",
  ["acta academiae aboensis"] = "Acta Acad. Abo. Ser. B",
  ["acta academiae aboensis : ser b math et phys"] = "Act. Acad. Ab.",
  ["acta academiae medicinae wuhan"] = "Acta Acad. Med. Wuhan",
  ["acta academiae medicinae wuhan = wu-han i hsüeh yüan hsüeh pao"] = "Acta Acad Med Wuhan",
  ["acta academiae scientiarum polonae"] = "Acta Acad Sci Pol",
  ["acta academiae scientiarum polonae / polska akademia nauk"] = "Acta Acad Sci Pol",
  ["acta academiae scientiarum taurinensis"] = "Acta Acad. Sci. Taurinensis",
  ["acta acustica"] = "Acta Acust.",
  ["acta acustica united with acustica"] = "Acta Acust. united Acust.",
  ["acta acustica united with acustica : the journal of the european acoustics association (eeig)"] = "Acta Acust United Acust",
  ["acta ad archaeologiam et artium historiam pertinentia"] = "AAAH",
  ["acta ad archaeologiam et artium historiam pertinentia. series altera in 8"] = "ActaAArtHist SA",
  ["acta adriatica"] = "Acta Adriat",
  ["acta agralia fennica"] = "Acta Agral. Fenn.",
  ["acta agraria et silvestria, seria lesna, krakau"] = "Acta Agrar. Silv., Ser. Lesn. (Krak.)",
  ["acta agriculturae scandinavica section a: animal science"] = "Acta Agric. Scand. Sect A",
  ["acta agriculturae scandinavica section b: soil and plant science"] = "Acta Agric. Scand. Sect B",
  ["acta agriculturae slovenica"] = "Acta Agric Slov",
  ["acta agriculturæ scandinavica. section a, animal science"] = "Acta Agric Scand A Anim Sci",
  ["acta agriculturæ scandinavica. section b, soil and plant science"] = "Acta Agric Scand B Soil Plant Sci",
  ["acta agrobotanica"] = "Acta Agrobot",
  ["acta agronomica"] = "Acta Agron",
  ["acta agronomica hungarica"] = "Acta Agron. Hung.",
  ["acta agronomica sinica"] = "Acta Agron. Sin.",
  ["acta alimentaria"] = "Acta Aliment",
  ["acta allergologica"] = "Acta Allergol",
  ["acta allergologica. supplementum"] = "Acta Allergol Suppl (Copenh)",
  ["acta amazonica"] = "Acta Amazon",
  ["acta anaesthesiologica"] = "Acta Anaesthesiol",
  ["acta anaesthesiologica belgica"] = "Acta Anaesthesiol Belg",
  ["acta anaesthesiologica italica"] = "Acta Anaesthesiol Ital",
  ["acta anaesthesiologica scandinavica"] = "Acta Anaesthesiol Scand",
  ["acta anaesthesiologica scandinavica. supplementum"] = "Acta Anaesthesiol Scand Suppl",
  ["acta anaesthesiologica sinica"] = "Acta Anaesthesiol Sin",
  ["acta anaesthesiologica taiwanica"] = "Acta Anaesthesiol. Taiwan.",
  ["acta anaesthesiologica taiwanica : official journal of the taiwan society of anesthesiologists"] = "Acta Anaesthesiol Taiwan",
  ["acta analysis functionalis applicata. aafa. yingyong fanhanfenxi xuebao"] = "Acta Anal. Funct. Appl.",
  ["acta analytica"] = "Acta Anal.",
  ["acta analytica : philosophy and psychology"] = "Acta Anal",
  ["acta anatomica"] = "Acta Anat (Basel)",
  ["acta anatomica. supplement"] = "Acta Anat. Suppl. (Basel)",
  ["acta anatomica. supplementum"] = "Acta Anat Suppl (Basel)",
  ["acta anthropogenetica"] = "Acta Anthropogenet",
  ["acta antiqua academiae scientiarum hungaricae"] = "Acta Antiq Hung",
  ["acta applicandae mathematicae"] = "Acta Appl. Math.",
  ["acta arachnologica"] = "Acta Arachnol",
  ["acta archaeologica"] = "Acta Archaeol.",
  ["acta archaeologica (arheoloski vestnik)"] = "AArchSlov",
  ["acta archaeologica academiae scientiarum hungaricae"] = "AArchHung",
  ["acta archaeologica carpathica"] = "ActaACarp",
  ["acta archaeologica lovaniensia"] = "ActaALov",
  ["acta archaeologica lovaniensia. monographiae"] = "ActaALovMono",
  ["acta archaeologica. københavn"] = "ActaArch",
  ["acta argentina de fisiologia y fisiopatologia"] = "Acta Argent Fisiol Fisiopatol",
  ["acta argentina de fisiología y fisiopatología"] = "Acta Argent Fisiol Fisiopatol",
  ["acta arithmetica"] = "Acta Arith.",
  ["acta associationis internationalis »terra antiqua balcanica«"] = "TerraAntBalc",
  ["acta astronautica"] = "Acta Astronaut",
  ["acta astronomica"] = "Acta Astron.",
  ["acta astronomica sinica"] = "Acta Astronom. Sinica",
  ["acta astrophysica sinica"] = "Acta Astrophys. Sinica",
  ["acta automatica sinica"] = "Acta Autom. Sin.",
  ["acta baltica"] = "Acta Balt",
  ["acta belgica de arte medicinali et pharmaceutica militari"] = "Acta Belg Arte Med Pharm Mil",
  ["acta belgica historiae medicinae : official journal of the belgian association for the history of medicine"] = "Acta Belg Hist Med",
  ["acta belgica historiae medicinae : official journal of the belgian association for the history of medicine / societas belgica historiae medicinae"] = "Acta Belg Hist Med",
  ["acta belgica. medica physica"] = "Acta Belg. Med. Phys.",
  ["acta belgica. medica physica : organe officiel de la societe royale belge de medecine physique et de rehabilitation"] = "Acta Belg Med Phys",
  ["acta belgica. medica physica : organe officiel de la société royale belge de médecine physique et de réhabilitation"] = "Acta Belg Med Phys",
  ["acta bernensia"] = "Acta Bern.",
  ["acta bio-medica : atenei parmensis"] = "Acta Biomed",
  ["acta bio-medica de l ateneo parmense"] = "Acta Biomed Ateneo Parmense",
  ["acta bio-medica de l'ateneo parmense : organo della società di medicina e scienze naturali di parma"] = "Acta Biomed Ateneo Parmense",
  ["acta bio-medica de l’ateneo parmense"] = "Acta Biomed. Ateneo Parmense",
  ["acta bio-medica de l’ateneo parmense : organo della societa di medicina e scienze naturali di parma"] = "Acta Biomed Ateneo Parmense",
  ["acta biochimica et biophysica academiae scientiarum hungaricae"] = "Acta Biochim. Biophys. Acad. Sci. Hung.",
  ["acta biochimica et biophysica hungarica"] = "Acta Biochim Biophys Hung",
  ["acta biochimica et biophysica sinica"] = "Acta Biochim Biophys Sin (Shanghai)",
  ["acta biochimica et biophysica; academiae scientiarum hungaricae"] = "Acta Biochim Biophys Acad Sci Hung",
  ["acta biochimica polonica"] = "Acta Biochim Pol",
  ["acta bioethica"] = "Acta Bioeth",
  ["acta biologiae experimentalis"] = "Acta Biol Exp (Warsz)",
  ["acta biologica academiae scientiarum hungaricae"] = "Acta Biol Acad Sci Hung",
  ["acta biologica cracoviensia. series: botanica"] = "Acta Biol Crac Ser Bot",
  ["acta biologica et medica"] = "Acta Biol Med (Gdansk)",
  ["acta biologica et medica germanica"] = "Acta Biol. Med. Ger.",
  ["acta biologica hungarica"] = "Acta Biol Hung",
  ["acta biologica marisiensis"] = "Acta Biol Marisiensis",
  ["acta biologica montana"] = "Acta Biol. Mont.",
  ["acta biologica plantarum agriensis : a journal of plant biology"] = "Acta Biol Plant Agriensis",
  ["acta biologica sibirica"] = "Acta Biol Sib",
  ["acta biologica slovenica : abs"] = "Acta Biol Slov",
  ["acta biologica turcica"] = "Acta Biol Turc",
  ["acta biológica colombiana"] = "Acta Biolo Colomb",
  ["acta biomaterialia"] = "Acta Biomater",
  ["acta biomaterialia odontologica scandinavica"] = "Acta Biomater Odontol Scand",
  ["acta biomedica scientifica"] = "Acta Biomed Sci",
  ["acta bioquimica clinica latinoamericana"] = "Acta Bioquim. Clin. Latinoam.",
  ["acta biotechnologica"] = "Acta Biotechnol",
  ["acta biotheoretica"] = "Acta Biotheor",
  ["acta borealia"] = "Acta Borealia",
  ["acta botanica brasílica"] = "Acta Bot Brasilica",
  ["acta botanica croatica"] = "Acta Bot Croat",
  ["acta botanica fennica"] = "Acta Bot. Fenn.",
  ["acta botanica gallica : bulletin de la societe botanique de france"] = "Acta Bot Gallica",
  ["acta botanica gallica : bulletin de la société botanique de france"] = "Acta Bot Gall",
  ["acta botanica hungarica"] = "Acta Bot Hung",
  ["acta botanica mexicana"] = "Acta Bot. Mex.",
  ["acta botanica neerlandica"] = "Acta Bot. Neerl.",
  ["acta botanica sinica"] = "Acta Bot Sin",
  ["acta botánica mexicana"] = "Acta Bot Mex",
  ["acta botánica venezuélica"] = "Acta Bot Venez",
  ["acta brevia neerlandica de physiologia, pharmacologia, microbiologia e.a"] = "Acta Brevia Neerl Physiol Pharmacol Microbiol E A",
  ["acta cancerologica"] = "Acta Cancerol. (Lima)",
  ["acta cancerológica"] = "Acta Cancerol (Lima)",
  ["acta cardiologica"] = "Acta Cardiol",
  ["acta cardiologica sinica"] = "Acta Cardiol Sin",
  ["acta cardiologica. supplementum"] = "Acta Cardiol Suppl",
  ["acta chemica scandinavica"] = "Acta Chem. Scand.",
  ["acta chemica scandinavica (copenhagen, denmark : 1989)"] = "Acta Chem Scand",
  ["acta chemica scandinavica, series a: physical and inorganic chemistry"] = "Acta Chem. Scand. Ser. A",
  ["acta chemica scandinavica, series b: organic chemistry and biochemistry"] = "Acta Chem. Scand. Ser. B",
  ["acta chemica scandinavica. series a, physical and inorganic chemistry"] = "Acta Chem. Scand. A",
  ["acta chemica scandinavica. series a: physical and inorganic chemistry"] = "Acta Chem Scand A",
  ["acta chemica scandinavica. series b, organic chemistry and biochemistry"] = "Acta Chem. Scand. B",
  ["acta chemica scandinavica. series b: organic chemistry and biochemistry"] = "Acta Chem Scand B",
  ["acta chimica hungarica"] = "Acta Chim. Hung.",
  ["acta chimica sinica"] = "Acta Chimi Sin",
  ["acta chimica slovaca"] = "Acta Chim. Slovaca",
  ["acta chimica slovenica"] = "Acta Chim Slov",
  ["acta chiropterologica"] = "Acta Chiropt",
  ["acta chirurgiae orthopaedicae et traumatologiae cechoslovaca"] = "Acta Chir Orthop Traumatol Cech",
  ["acta chirurgiae plasticae"] = "Acta Chir Plast",
  ["acta chirurgica academiae scientiarum hungaricae"] = "Acta Chir Acad Sci Hung",
  ["acta chirurgica austriaca"] = "Acta Chir Austriaca",
  ["acta chirurgica belgica"] = "Acta Chir Belg",
  ["acta chirurgica hungarica"] = "Acta Chir Hung",
  ["acta chirurgica italica"] = "Acta Chir Ital",
  ["acta chirurgica iugoslavica"] = "Acta Chir Iugosl",
  ["acta chirurgica patavina"] = "Acta Chir Patav",
  ["acta chirurgica scandinavica"] = "Acta Chir Scand",
  ["acta chirurgica scandinavica. supplementum"] = "Acta Chir Scand Suppl",
  ["acta chromatographica"] = "Acta Chromatogr.",
  ["acta cienceia indica chemistry"] = "Acta Cienc. Indica, Chem.",
  ["acta ciencia indica"] = "Acta Cienc. Indica Math.",
  ["acta ciencia indica chemistry"] = "Acta Cienc. Indica, Chem.",
  ["acta ciencia indica phyics"] = "Acta Cienc. Indica, Phys.",
  ["acta cientifica venezolana"] = "Acta Cient Venez",
  ["acta cientifica venezolana. asociacion venezolana para el avance de la ciencia"] = "Acta Cient. Venezolana",
  ["acta científica venezolana"] = "Acta Cient Venez",
  ["acta científica venezolana"] = "Acta Cient. Venezolana",
  ["acta cirurgica brasileira / sociedade brasileira para desenvolvimento pesquisa em cirurgia"] = "Acta Cir Bras",
  ["acta cirúrgica brasileira"] = "Acta Cir Bras",
  ["acta classica"] = "Acta Cl",
  ["acta classica universitatis scientiarum debreceniensis"] = "ACD",
  ["acta classica. proceedings of the classical association of south africa"] = "ActaCl",
  ["acta classica: proceedings of the classical association of south africa"] = "AClass",
  ["acta clinica belgica"] = "Acta Clin Belg",
  ["acta clinica belgica. supplementum"] = "Acta Clin Belg Suppl",
  ["acta clinica croatica"] = "Acta Clin Croat",
  ["acta clinica odontologica"] = "Acta Clin. Odontol.",
  ["acta crystallographica"] = "Acta Crystallogr",
  ["acta crystallographica section a. foundations and advances"] = "Acta Crystallogr. Sect. A",
  ["acta crystallographica section a: foundations"] = "Acta Crystallogr., Sect. A: Found. Crystallogr.",
  ["acta crystallographica section b, structural science, crystal engineering and materials"] = "Acta Crystallogr B Struct Sci Cryst Eng Mater",
  ["acta crystallographica section b: structural science"] = "Acta Crystallogr., Sect. B: Struct. Sci",
  ["acta crystallographica section c: crystal structure communications"] = "Acta Crystallogr., Sect. C: Cryst. Struct. Commun.",
  ["acta crystallographica section d: biological crystallography"] = "Acta Crystallogr., Sect. D: Biol. Crystallogr.",
  ["acta crystallographica section e structure reports online"] = "Acta Crystallogr., Sect. E: Struct. Rep. Online",
  ["acta crystallographica section f: structural biology and crystallization communications"] = "Acta Crystallogr., Sect. F: Struct. Biol. Cryst. Commun.",
  ["acta crystallographica, section a: crystal physics, diffraction,theoretical and general crystallography"] = "Acta Crystallogr. Sec. A",
  ["acta crystallographica, section a: foundations and advances"] = "Acta Crystallogr., Sect. A: Found. Adv.",
  ["acta crystallographica, section a: foundations of crystallography"] = "Acta Crystallogr., Sect. A: Found. Crystallogr.",
  ["acta crystallographica, section b: structural crystallography and crystal chemistry"] = "Acta Crystallogr. Sec. B",
  ["acta crystallographica, section b: structural science"] = "Acta Crystallogr., Sect. B: Struct. Sci.",
  ["acta crystallographica, section b: structural science, crystal engineering and materials"] = "Acta Crystallogr., Sect. B: Struct. Sci., Cryst. Eng. Mater.",
  ["acta crystallographica, section c: crystal structure communications"] = "Acta Crystallogr., Sect. C: Cryst. Struct. Commun.",
  ["acta crystallographica, section c: structural chemistry"] = "Acta Crystallogr., Sect. C: Struct. Chem.",
  ["acta crystallographica, section d: biological crystallography"] = "Acta Crystallogr., Sect. D: Biol. Crystallogr.",
  ["acta crystallographica, section e: crystallographic communications"] = "Acta Crystallogr., Sect. E: Crystallogr. Commun.",
  ["acta crystallographica, section e: structure reports online"] = "Acta Crystallogr., Sect. E: Struct. Rep. Online",
  ["acta crystallographica, section f: structural biology and crystallization communications"] = "Acta Crystallogr., Sect. F: Struct. Biol. Cryst. Commun.",
  ["acta crystallographica, section f: structural biology communications"] = "Acta Crystallogr., Sect. F: Struct. Biol. Commun.",
  ["acta crystallographica. section a, crystal physics, diffraction, theoretical and general crystallography"] = "Acta Crystallogr A",
  ["acta crystallographica. section a, foundations and advances"] = "Acta Crystallogr A Found Adv",
  ["acta crystallographica. section a, foundations of crystallography"] = "Acta Crystallogr A",
  ["acta crystallographica. section b, structural crystallography and crystal chemistry"] = "Acta Crystallogr. B",
  ["acta crystallographica. section b, structural science"] = "Acta Crystallogr B",
  ["acta crystallographica. section b: structural crystallography and crystal chemistry"] = "Acta Crystallogr B",
  ["acta crystallographica. section c, crystal structure communications"] = "Acta Crystallogr C",
  ["acta crystallographica. section c, structural chemistry"] = "Acta Crystallogr C Struct Chem",
  ["acta crystallographica. section d, biological crystallography"] = "Acta Crystallogr D Biol Crystallogr",
  ["acta crystallographica. section d, structural biology"] = "Acta Crystallogr D Struct Biol",
  ["acta crystallographica. section e, crystallographic communications"] = "Acta Crystallogr E Crystallogr Commun",
  ["acta crystallographica. section e, structure reports online"] = "Acta Crystallogr Sect E Struct Rep Online",
  ["acta crystallographica. section f, structural biology and crystallization communications"] = "Acta Crystallogr Sect F Struct Biol Cryst Commun",
  ["acta crystallographica. section f, structural biology communications"] = "Acta Crystallogr F Struct Biol Commun",
  ["acta cybernetica"] = "Acta Cybernet.",
  ["acta cytologica"] = "Acta Cytol",
  ["acta de investigación psicológica"] = "Acta Investig Psicol",
  ["acta de odontologia pediatrica"] = "Acta Odontol. Pediatr.",
  ["acta de odontología pediátrica"] = "Acta Odontol Pediatr",
  ["acta demographica"] = "Acta Demogr",
  ["acta dermato-venereologica"] = "Acta Derm Venereol",
  ["acta dermato-venereologica. supplementum"] = "Acta Derm Venereol Suppl (Stockh)",
  ["acta dermatologica kyoto. english edition"] = "Acta Dermatol. Kyoto Engl. Ed.",
  ["acta dermatologica-kyoto. english edition"] = "Acta Dermatol Kyoto Engl Ed",
  ["acta dermatovenerologica alpina, pannonica, et adriatica"] = "Acta Dermatovenerol Alp Pannonica Adriat",
  ["acta dermatovenerologica alpina, panonica, et adriatica"] = "Acta Dermatovenerol Alp Panonica Adriat",
  ["acta dermatovenerologica croatica : adc"] = "Acta Dermatovenerol Croat",
  ["acta dermatovenerologica croatica : adc / hrvatsko dermatolosko drustvo"] = "Acta Dermatovenerol Croat",
  ["acta diabetologica"] = "Acta Diabetol",
  ["acta diabetologica latina"] = "Acta Diabetol Lat",
  ["acta ecologica sinica"] = "Acta Ecol. Sin.",
  ["acta embryologiae et morphologiae experimentalis"] = "Acta Embryol Morphol Exp",
  ["acta embryologiae experimentalis"] = "Acta Embryol Exp (Palermo)",
  ["acta endocrinologica"] = "Acta Endocrinol (Copenh)",
  ["acta endocrinologica (bucharest, romania : 2005)"] = "Acta Endocrinol (Buchar)",
  ["acta endocrinologica iberica"] = "Acta Endocrinol Iber",
  ["acta endocrinologica panamericana"] = "Acta Endocrinol Panam",
  ["acta endocrinologica. supplementum"] = "Acta Endocrinol Suppl (Copenh)",
  ["acta endoscopica"] = "Acta Endoscopica",
  ["acta entomologica musei nationalis pragae"] = "Acta Entomol Mus Natl Pragae",
  ["acta et commentationes universitatis tartuensis de mathematica"] = "Acta Comment. Univ. Tartu. Math.",
  ["acta ethnologica et linguistica"] = "Acta Ethnol Linguist",
  ["acta ethologica"] = "Acta Ethol",
  ["acta europaea fertilitatis"] = "Acta Eur Fertil",
  ["acta facultatis pharmaceuticae universitatis comenianae"] = "Acta Fac Pharm Univ Comen",
  ["acta farmaceutica bonaerense"] = "Acta Farm. Bonaerense",
  ["acta forestalia fennica"] = "Acta For. Fenn.",
  ["acta gastro-enterologica belgica"] = "Acta Gastroenterol Belg",
  ["acta gastroenterologica belgica"] = "Acta Gastroenterol. Belg.",
  ["acta gastroenterologica latinoamericana"] = "Acta Gastroenterol Latinoam",
  ["acta genetica et statistica medica"] = "Acta Genet Stat Med",
  ["acta genetica sinica"] = "Acta Genet. Sin.",
  ["acta geneticae medicae et gemellologiae"] = "Acta Genet Med Gemellol (Roma)",
  ["acta geochimica"] = "Acta Geochimica",
  ["acta geodaetica et geophysica"] = "Acta Geod. Geophys.",
  ["acta geodaetica et geophysica hungarica"] = "Acta Geod. Geophys. Hung.",
  ["acta geodynamics et geomaterialia"] = "Acta Geodyn. Geomater.",
  ["acta geographica"] = "Acta Geogr",
  ["acta geographica ac geologica et meteorologica debrecina. geológia, geomorfológia, természetföldrajz sorozat"] = "Acta Geogr Geol Meteorol Debr Geol Gemorfol Termeszfoldr Sor",
  ["acta geographica croatica : glasilo geografskog odjela pmf-a sveucilista u zagrebu"] = "Acta Geogr Croat",
  ["acta geographica croatica : glasilo geografskog odjela pmf-a sveučilišta u zagrebu"] = "Acta Geogr Croat",
  ["acta geographica sinica"] = "Acta Geogr. Sin.",
  ["acta geologica hungarica"] = "Acta Geol. Hung.",
  ["acta geologica polonica"] = "Acta Geol. Pol.",
  ["acta geologica sinica"] = "Acta Geol. Sin.",
  ["acta geologica slovaca : ageos"] = "Acta Geol Slovaca",
  ["acta geophysica"] = "Acta Geophys.",
  ["acta geotechnica"] = "Acta Geotech.",
  ["acta gerontologica"] = "Acta Gerontol (Milano)",
  ["acta ginecologica"] = "Acta Ginecol. (Madr.)",
  ["acta ginecológica"] = "Acta Ginecol (Madr)",
  ["acta gynaecologica et obstetrica hispano-lusitana"] = "Acta Gynaecol Obstet Hisp Lusit",
  ["acta haematologica"] = "Acta Haematol",
  ["acta haematologica polonica"] = "Acta Haematol Pol",
  ["acta hepato-gastroenterologica"] = "Acta Hepatogastroenterol (Stuttg)",
  ["acta hepato-splenologica"] = "Acta Hepatosplenol",
  ["acta hepatologica"] = "Act Hepato",
  ["acta herpetologica"] = "Acta Herpetol",
  ["acta histochemica"] = "Acta Histochem",
  ["acta histochemica et cytochemica"] = "Acta Histochem Cytochem",
  ["acta histochemica. supplementband"] = "Acta Histochem Suppl",
  ["acta historiae rerum naturalium necnon technicarum. special issue"] = "Acta Hist Rerum Nat Tech",
  ["acta historica academiae scientiarum hungaricae"] = "AHistHung",
  ["acta historica leopoldina"] = "Acta Hist Leopoldina",
  ["acta historica medicinae, pharmaciae, veterinae"] = "Acta Hist Med Pharm Vet (Beograd)",
  ["acta historica scientiarum naturalium et medicinalium"] = "Acta Hist Sci Nat Med",
  ["acta historica. societas academica dacoromana"] = "ActaHistDac",
  ["acta historico-medica vallisoletana. monografias"] = "Acta Hist Med Vallisoletana Monogr",
  ["acta histórico-médica vallisoletana. monografías"] = "Acta Hist Med Vallisoletana Monogr",
  ["acta horticulturae"] = "Acta Hortic",
  ["acta horticulturae et regiotectuare"] = "Acta Hortic Regiotect",
  ["acta hospitalia"] = "Acta Hosp",
  ["acta hydrochimica et hydrobiologica"] = "Acta Hydroch. Hydrob.",
  ["acta hydrophysica"] = "Acta Hydrophys.",
  ["acta hyperborea. danish studies in classical archaeology"] = "ActaHyp",
  ["acta iberica radiologica-cancerologica"] = "Acta Iber. Radiol. Cancerol.",
  ["acta ibérica radiológica-cancerológica"] = "Acta Iber Radiol Cancerol",
  ["acta ichthyologica et piscatoria"] = "Acta Ichthyol Piscat",
  ["acta imeko (2012)"] = "Acta IMEKO (2012)",
  ["acta informatica"] = "Acta Inform.",
  ["acta informatica medica : aim : journal of the society for medical informatics of bosnia & herzegovina : časopis društva za medicinsku informatiku bih"] = "Acta Inform Med",
  ["acta instituti atheniensis regni sueciae"] = "ActaAth",
  ["acta instituti forestalis zvolenensis"] = "Acta Inst. For. Zvolen.",
  ["acta instituti romani finlandiae"] = "ActaInstRomFin",
  ["acta iranica: encyclopédie permanente des études iraniennes"] = "AI",
  ["acta isotopica"] = "Acta Isot (Padova)",
  ["acta juridica (cape town, south africa)"] = "Acta Jurid (Cape Town)",
  ["acta leidensia"] = "Acta Leiden",
  ["acta leprologica"] = "Acta Leprol",
  ["acta limnologica brasiliensia"] = "Acta Limnol. Bras.",
  ["acta linguistica hungarica"] = "Acta Linguist Hung",
  ["acta manilana"] = "Acta Manila Ser A",
  ["acta materialia"] = "Acta Mater.",
  ["acta mathematica"] = "Acta Math.",
  ["acta mathematica academiae scientiarum hungaricae"] = "Acta Math. Acad. Sci. Hung.",
  ["acta mathematica et informatica universitatis ostraviensis"] = "Acta Math. Inform. Univ. Ostraviensis",
  ["acta mathematica hungarica"] = "Acta Math. Hungar.",
  ["acta mathematica scientia"] = "Acta Math. Sci.",
  ["acta mathematica scientia = shu xue wu li xue bao"] = "Acta Math Sci",
  ["acta mathematica scientia. series a. shuxue wuli xuebao. chinese edition"] = "Acta Math. Sci. Ser. A (Chin. Ed.)",
  ["acta mathematica scientia. series b. english edition"] = "Acta Math. Sci. Ser. B (Engl. Ed.)",
  ["acta mathematica sinica"] = "Acta Math. Sin.",
  ["acta mathematica sinica (english series)"] = "Acta Math. Sin. (Engl. Ser.)",
  ["acta mathematica sinica, english series"] = "Acta Math Sin Engl Ser",
  ["acta mathematica sinica. chinese series"] = "Acta Math. Sinica (Chin. Ser.)",
  ["acta mathematica universitatis comenianae"] = "Acta Math. Univ. Comenian. (N.S.)",
  ["acta mathematica universitatis comenianae. new series"] = "Acta Math. Univ. Comenian. (N.S.)",
  ["acta mathematica vietnamica"] = "Acta Math. Vietnam.",
  ["acta mathematica. academiae paedagogicae nyiregyhaziensis"] = "Acta Math. Acad. Paedagog. Nyhazi",
  ["acta mathematica. academiae paedagogicae nyiregyhaziensis. new series"] = "Acta Math. Acad. Paedagog. Nyhazi. (N.S.)",
  ["acta mathematicae applicatae sinica"] = "Acta Math. Appl. Sin.",
  ["acta mathematicae applicatae sinica (english series)"] = "Acta Math Appl Sin",
  ["acta mathematicae applicatae sinica. english series"] = "Acta Math. Appl. Sin. Engl. Ser.",
  ["acta mathematicae applicatae sinica. yingyong shuxue xuebao"] = "Acta Math. Appl. Sin.",
  ["acta mechanica"] = "Acta Mech.",
  ["acta mechanica sinica"] = "Acta Mech. Sin.",
  ["acta mechanica sinica = li xue xue bao"] = "Acta Mech Sin",
  ["acta mechanica solida sinica"] = "Acta Mech. Solida Sin.",
  ["acta medica"] = "Acta Med",
  ["acta medica (hradec kralove)"] = "Acta Medica (Hradec Kralove)",
  ["acta medica (hradec kralove) / universitas carolina, facultas medica hradec kralove"] = "Acta Medica (Hradec Kralove)",
  ["acta medica (hradec kralove). supplementum"] = "Acta Medica (Hradec Kralove) Suppl.",
  ["acta medica (hradec kralove). supplementum universitas carolina, facultas medica hradec kralove"] = "Acta Medica (Hradec Kralove) Suppl",
  ["acta medica (hradec králové)"] = "Acta Medica (Hradec Kralove)",
  ["acta medica (hradec králové). supplementum"] = "Acta Medica (Hradec Kralove) Suppl",
  ["acta medica academiae scientiarum hungaricae"] = "Acta Med Acad Sci Hung",
  ["acta medica academica"] = "Acta Med Acad",
  ["acta medica austriaca"] = "Acta Med Austriaca",
  ["acta medica austriaca. supplement"] = "Acta Med Austriaca Suppl",
  ["acta medica auxologica"] = "Acta Med Auxol (Milano)",
  ["acta medica catholica"] = "Acta Med Cathol",
  ["acta medica colombiana : amc : organo de la asociacion columbiana de medicina interna"] = "Acta Med Colomb",
  ["acta medica costarricense"] = "Acta Med Costarric",
  ["acta medica croatica"] = "Acta Med. Croatica",
  ["acta medica croatica : casopis hravatske akademije medicinskih znanosti"] = "Acta Med Croatica",
  ["acta medica croatica : c̆asopis hravatske akademije medicinskih znanosti"] = "Acta Med Croatica",
  ["acta medica de tenerife"] = "Acta Med Tenerife",
  ["acta medica del valle"] = "Acta Med Valle",
  ["acta medica dominicana : revista cientifica para medicos"] = "Acta Med Dom",
  ["acta medica et biologica"] = "Acta Med Biol (Niigata)",
  ["acta medica hidalguense"] = "Acta Med Hidalguense",
  ["acta medica hispanica"] = "Acta Med Hisp",
  ["acta medica hungarica"] = "Acta Med Hung",
  ["acta medica indonesiana"] = "Acta Med Indones",
  ["acta medica iranica"] = "Acta Med Iran",
  ["acta medica italica di malattie infettive e parassitarie"] = "Acta Med Ital Mal Infett Parassit",
  ["acta medica italica di medicina tropicale e subtropicale e di gastroenterologia"] = "Acta Med Ital Med Trop Subtrop Gastroenterol",
  ["acta medica iugoslavica"] = "Acta Med Iugosl",
  ["acta medica kinki university"] = "Acta Med Kinki Univ",
  ["acta medica lituanica"] = "Acta Med Litu",
  ["acta medica marisiensis"] = "Acta Med Marisiensis",
  ["acta medica nagasakiensia"] = "Acta Med Nagasaki",
  ["acta medica okayama"] = "Acta Med Okayama",
  ["acta medica orientalia"] = "Acta Med Orient",
  ["acta medica patavina"] = "Acta Med Patav",
  ["acta medica patavina. supplemento"] = "Acta Med Patavina Suppl",
  ["acta medica philippina"] = "Acta Med Philipp",
  ["acta medica polona"] = "Acta Med Pol",
  ["acta medica portuguesa"] = "Acta Med Port",
  ["acta medica romana; annali della facolta di medicina e chirurgia, universita cattolica del sacro cuore"] = "Acta Med Rom",
  ["acta medica romana; annali della facoltà di medicina e chirurgia, università cattolica del sacro cuore"] = "Acta Med Rom",
  ["acta medica scandinavica"] = "Acta Med Scand",
  ["acta medica scandinavica. supplementum"] = "Acta Med Scand Suppl",
  ["acta medica transilvanica"] = "Acta Medica Transilvanica",
  ["acta medica turcica"] = "Acta Med Turc",
  ["acta medica veterinaria"] = "Acta Med Vet (Napoli)",
  ["acta medica vietnamica"] = "Acta Med Vietnam",
  ["acta medicae historiae patavina"] = "Acta Med Hist Patav",
  ["acta medicinae legalis"] = "Acta Med Leg",
  ["acta medicinae legalis et socialis"] = "Acta Med Leg Soc (Liege)",
  ["acta medicinae okayama"] = "Acta Med Okayama",
  ["acta medico-historica adriatica : amha"] = "Acta Med Hist Adriat",
  ["acta metallurgica"] = "Acta Metall.",
  ["acta metallurgica et materialia"] = "Acta Metall. Mater.",
  ["acta metallurgica sinica"] = "Acta Metall. Sinica",
  ["acta metallurgica sinica- english letters"] = "Acta Metall. Sin. Engl. Lett.",
  ["acta metallurgica slovaca"] = "Acta Metall. Slovaca",
  ["acta médica colombiana : amc : organo de la asociación columbiana de medicina interna"] = "Acta Med Colomb",
  ["acta médica costarricense"] = "Acta Med Costarric",
  ["acta médica de tenerife"] = "Acta Med Tenerife",
  ["acta médica del valle"] = "Acta Med Valle",
  ["acta médica dominicana : revista científica para médicos"] = "Acta Med Dom",
  ["acta médica hidalguense"] = "Acta Med Hidalguense",
  ["acta médica hispánica"] = "Acta Med Hisp",
  ["acta médica portuguesa"] = "Acta Med Port",
  ["acta microbiologica academiae scientiarum hungaricae"] = "Acta Microbiol Acad Sci Hung",
  ["acta microbiologica bulgarica"] = "Acta Microbiol Bulg",
  ["acta microbiologica et immunologica hungarica"] = "Acta Microbiol Immunol Hung",
  ["acta microbiologica hellenica"] = "Acta Microbiol. Hell.",
  ["acta microbiologica hungarica"] = "Acta Microbiol Hung",
  ["acta microbiologica polonica"] = "Acta Microbiol Pol",
  ["acta microbiologica polonica (1952)"] = "Acta Microbiol Pol (1952)",
  ["acta microbiologica polonica. series a, microbiologia generalis"] = "Acta Microbiol. Pol. A",
  ["acta microbiologica polonica. series a: microbiologia generalis"] = "Acta Microbiol Pol A",
  ["acta microbiologica polonica. series b, microbiologia applicata"] = "Acta Microbiol. Pol. B",
  ["acta microbiologica polonica. series b: microbiologia applicata"] = "Acta Microbiol Pol B",
  ["acta microbiologica sinica"] = "Acta Microbiol Sin",
  ["acta microbiologica, virologica et immunologica"] = "Acta Microbiol Virol Immunol (Sofiia)",
  ["acta montanistica slovaca"] = "Acta Montan. Slovaca",
  ["acta morphologica academiae scientiarum hungaricae"] = "Acta Morphol Acad Sci Hung",
  ["acta morphologica academiae scientiarum hungaricae. supplementum"] = "Acta Morphol Acad Sci Hung Suppl",
  ["acta morphologica hungarica"] = "Acta Morphol Hung",
  ["acta morphologica neerlando-scandinavica"] = "Acta Morphol Neerl Scand",
  ["acta musei napocensis"] = "Acta Mus Napocensis",
  ["acta musei porolissensis"] = "ActaMusPorol",
  ["acta musei silesiae. scientiae naturales"] = "Acta Musei Sil Sci Natur",
  ["acta mycologica"] = "Acta Mycol",
  ["acta myologica : myopathies and cardiomyopathies : official journal of the mediterranean society of myology"] = "Acta Myol",
  ["acta myologica : myopathies and cardiomyopathies : official journal of the mediterranean society of myology / edited by the gaetano conte academy for the study of striated muscle diseases"] = "Acta Myol",
  ["acta médica costarricense"] = "Acta méd. costarric.",
  ["acta naturae"] = "Acta Naturae",
  ["acta neerlandica morphologiae normalis et pathologicae"] = "Acta Neerl Morphol Norm Pathol",
  ["acta neurobiologiae experimentalis"] = "Acta Neurobiol Exp (Wars)",
  ["acta neurochirurgica"] = "Acta Neurochir (Wien)",
  ["acta neurochirurgica. supplement"] = "Acta Neurochir Suppl",
  ["acta neurochirurgica. supplementum"] = "Acta Neurochir Suppl (Wien)",
  ["acta neurologica"] = "Acta Neurol (Napoli)",
  ["acta neurologica belgica"] = "Acta Neurol Belg",
  ["acta neurologica et psychiatrica belgica"] = "Acta Neurol Psychiatr Belg",
  ["acta neurologica latinoamericana"] = "Acta Neurol. Latinoam.",
  ["acta neurologica scandinavica"] = "Acta Neurol Scand",
  ["acta neurologica scandinavica. supplementum"] = "Acta Neurol Scand Suppl",
  ["acta neurologica taiwanica"] = "Acta Neurol Taiwan",
  ["acta neurologica. quaderni"] = "Acta Neurol Quad (Napoli)",
  ["acta neurológica latinoamericana"] = "Acta Neurol Latinoam",
  ["acta neuropathologica"] = "Acta Neuropathol",
  ["acta neuropathologica communications"] = "Acta Neuropathol Commun",
  ["acta neuropathologica. supplement"] = "Acta Neuropathol. Suppl. (Berl.)",
  ["acta neuropathologica. supplementum"] = "Acta Neuropathol Suppl",
  ["acta neuropsychiatrica"] = "Acta Neuropsychiatr",
  ["acta neuropsychologica : the official journal of the polish neuropsychological society"] = "Acta Neuropsychol",
  ["acta neurovegetativa"] = "Acta Neuroveg (Wien)",
  ["acta neurovegetativa. supplementum"] = "Acta Neuroveg Suppl",
  ["acta numerica"] = "Acta Numer.",
  ["acta numismatica"] = "Acta Numis",
  ["acta numismática"] = "Acta Numis",
  ["acta numismática (barcelona)"] = "ActaNum",
  ["acta obstetrica et gynaecologica japonica"] = "Acta Obstet Gynaecol Jpn",
  ["acta obstetrica y ginecologica hispano-lusitana"] = "Acta Obstet. Ginecol. Hisp. Lusit.",
  ["acta obstetrica y ginecologica hispano-lusitana. suplemento"] = "Acta Obstet Ginecol Hisp Lusit Suppl",
  ["acta obstetricia et gynecologica scandinavica"] = "Acta Obstet Gynecol Scand",
  ["acta obstetricia et gynecologica scandinavica. supplement"] = "Acta Obstet Gynecol Scand Suppl",
  ["acta obstétrica y ginecológica hispano-lusitana"] = "Acta Obstet Ginecol Hisp Lusit",
  ["acta oceanologica sinica"] = "Acta Oceanolog. Sin.",
  ["acta oceanologica sinica = hai yang xue bao"] = "Hai Yang Xue Bao",
  ["acta odontologica latinoamericana"] = "Acta Odontol. Latinoam.",
  ["acta odontologica latinoamericana : aol"] = "Acta Odontol Latinoam",
  ["acta odontologica scandinavica"] = "Acta Odontol Scand",
  ["acta odontologica scandinavica. supplementum"] = "Acta Odontol Scand Suppl",
  ["acta odontologica venezolana"] = "Acta Odontol. Venez.",
  ["acta odontológica latinoamericana : aol"] = "Acta Odontol Latinoam",
  ["acta odontológica venezolana"] = "Acta Odontol Venez",
  ["acta oecologica"] = "Acta Oecol.",
  ["acta oecologica (montrouge, france)"] = "Acta Oecol (Montrouge)",
  ["acta oeconomica"] = "Acta Oecon.",
  ["acta oeconomica : periodical of the hungarian academy of sciences"] = "Acta Oecon",
  ["acta of bioengineering and biomechanics"] = "Acta Bioeng Biomech",
  ["acta of bioengineering and biomechanics / wroclaw university of technology"] = "Acta Bioeng Biomech",
  ["acta oncologica"] = "Acta Oncol.",
  ["acta oncologica (stockholm, sweden)"] = "Acta Oncol",
  ["acta oncológica"] = "Acta Oncol (Madr)",
  ["acta ophthalmologica"] = "Acta Ophthalmol",
  ["acta ophthalmologica scandinavica"] = "Acta Ophthalmol Scand",
  ["acta ophthalmologica scandinavica. supplement"] = "Acta Ophthalmol Scand Suppl",
  ["acta ophthalmologica. supplement"] = "Acta Ophthalmol Suppl (Oxf )",
  ["acta ophthalmologica. supplementum"] = "Acta Ophthalmol Suppl",
  ["acta orientalia (kopenhagen)"] = "ActaOr",
  ["acta orientalia academiae scientiarum hungaricae"] = "AOrientHung",
  ["acta ornithologica"] = "Acta Ornithol",
  ["acta orthopaedica"] = "Acta Orthop",
  ["acta orthopaedica belgica"] = "Acta Orthop Belg",
  ["acta orthopaedica et traumatologica turcica"] = "Acta Orthop Traumatol Turc",
  ["acta orthopaedica scandinavica"] = "Acta Orthop Scand",
  ["acta orthopaedica scandinavica. supplementum"] = "Acta Orthop Scand Suppl",
  ["acta orthopaedica. supplementum"] = "Acta Orthop Suppl",
  ["acta ortopedica brasileira"] = "Acta Ortop Bras",
  ["acta ortopedica mexicana"] = "Acta Ortop Mex",
  ["acta ortopédica mexicana"] = "Acta Ortop Mex",
  ["acta oto-laryngologica"] = "Acta Otolaryngol",
  ["acta oto-laryngologica case reports"] = "Acta Otolaryngol Case Rep",
  ["acta oto-laryngologica orientalia"] = "Acta Otolaryngol Orient",
  ["acta oto-laryngologica. supplement"] = "Acta Otolaryngol. Suppl. (Stockh.)",
  ["acta oto-laryngologica. supplementum"] = "Acta Otolaryngol Suppl",
  ["acta oto-rhino-laryngologica belgica"] = "Acta Otorhinolaryngol Belg",
  ["acta oto-rino-laringologica ibero-americana"] = "Acta Otorinolaryngol. Iber. Am.",
  ["acta oto-rino-laringológica ibero-americana"] = "Acta Otorinolaryngol Iber Am",
  ["acta otorhinolaryngologica italica"] = "Acta Otorhinolaryngol. Ital.",
  ["acta otorhinolaryngologica italica : organo ufficiale della societa italiana di otorinolaringologia e chirurgia cervico-facciale"] = "Acta Otorhinolaryngol Ital",
  ["acta otorhinolaryngologica italica : organo ufficiale della società italiana di otorinolaringologia e chirurgia cervico-facciale"] = "Acta Otorhinolaryngol Ital",
  ["acta otorrinolaringologica espanola"] = "Acta Otorrinolaringol Esp",
  ["acta otorrinolaringológica española"] = "Acta Otorrinolaringol Esp",
  ["acta paediatrica"] = "Acta Paediatr",
  ["acta paediatrica (oslo, norway : 1992)"] = "Acta Paediatr",
  ["acta paediatrica (oslo, norway : 1992). supplement"] = "Acta Paediatr Suppl",
  ["acta paediatrica academiae scientiarum hungaricae"] = "Acta Paediatr Acad Sci Hung",
  ["acta paediatrica belgica"] = "Acta Paediatr Belg",
  ["acta paediatrica hungarica"] = "Acta Paediatr Hung",
  ["acta paediatrica japonica"] = "Acta Paediatr. Jpn.",
  ["acta paediatrica japonica : overseas edition"] = "Acta Paediatr Jpn",
  ["acta paediatrica japonica; overseas edition"] = "Acta Paediatr Jpn",
  ["acta paediatrica latina"] = "Acta Paediatr Lat",
  ["acta paediatrica scandinavica"] = "Acta Paediatr Scand",
  ["acta paediatrica scandinavica. supplement"] = "Acta Paediatr Scand Suppl",
  ["acta paediatrica taiwanica"] = "Acta Paediatr. Taiwan.",
  ["acta paediatrica taiwanica = taiwan er ke yi xue hui za zhi"] = "Acta Paediatr Taiwan",
  ["acta paediatrica. supplement"] = "Acta Paediatr. Suppl.",
  ["acta paediatrica. supplementum"] = "Acta Paediatr Suppl",
  ["acta paedologica"] = "Acta Paedolog",
  ["acta paedopsychiatrica"] = "Acta Paedopsychiatr",
  ["acta palaeobotanica"] = "Acta Palaeobot.",
  ["acta palaeontologica polonica"] = "Acta Palaeontol Pol",
  ["acta palaeontologica romaniae"] = "Acta Palaeontol. Rom.",
  ["acta parasitologica"] = "Acta Parasitol",
  ["acta parasitology"] = "Acta Parasitol.",
  ["acta pathologica et microbiologica scandinavica"] = "Acta Pathol Microbiol Scand",
  ["acta pathologica et microbiologica scandinavica. section a, pathology"] = "Acta Pathol Microbiol Scand A",
  ["acta pathologica et microbiologica scandinavica. section b, microbiology"] = "Acta Pathol Microbiol Scand B",
  ["acta pathologica et microbiologica scandinavica. section b, microbiology and immunology"] = "Acta Pathol. Microbiol. Scand. [B] Microbiol. Immunol.",
  ["acta pathologica et microbiologica scandinavica. section b: microbiology and immunology"] = "Acta Pathol Microbiol Scand B Microbiol Immunol",
  ["acta pathologica et microbiologica scandinavica. section c, immunology"] = "Acta Pathol Microbiol Scand C",
  ["acta pathologica et microbiologica scandinavica. supplement"] = "Acta Pathol Microbiol Scand Suppl",
  ["acta pathologica et microbiologica scandinavica. supplementum"] = "Acta Pathol Microbiol Scand Suppl",
  ["acta pathologica japonica"] = "Acta Pathol Jpn",
  ["acta pathologica, microbiologica, et immunologica scandinavica. section a, pathology"] = "Acta Pathol Microbiol Immunol Scand A",
  ["acta pathologica, microbiologica, et immunologica scandinavica. section b, microbiology"] = "Acta Pathol Microbiol Immunol Scand B",
  ["acta pathologica, microbiologica, et immunologica scandinavica. section c, immunology"] = "Acta Pathol Microbiol Immunol Scand C",
  ["acta pathologica, microbiologica, et immunologica scandinavica. supplement"] = "Acta Pathol Microbiol Immunol Scand Suppl",
  ["acta pediatrica espanola"] = "Acta Pediatr. Esp.",
  ["acta pediátrica española"] = "Acta Pediatr Esp",
  ["acta pharmaceutica (zagreb, croatia)"] = "Acta Pharm",
  ["acta pharmaceutica hungarica"] = "Acta Pharm Hung",
  ["acta pharmaceutica internationalia"] = "Acta Pharm Int",
  ["acta pharmaceutica jugoslavica"] = "Acta Pharm Jugosl",
  ["acta pharmaceutica nordica"] = "Acta Pharm Nord",
  ["acta pharmaceutica sinica. b"] = "Acta Pharm Sin B",
  ["acta pharmaceutica suecica"] = "Acta Pharm Suec",
  ["acta pharmacologica et toxicologica"] = "Acta Pharmacol Toxicol (Copenh)",
  ["acta pharmacologica sinica"] = "Acta Pharmacol Sin",
  ["acta philologica. societas academica dacoromana"] = "ActaPhilSocDac",
  ["acta physica"] = "Acta Phys.",
  ["acta physica academiae scientiarum hungaricae"] = "Acta Phys. Acad. Sci. Hung.",
  ["acta physica austriaca"] = "Acta Phys. Austriaca",
  ["acta physica debrecina"] = "Acta Phys. Debrecina",
  ["acta physica et chimica debrecina"] = "Acta Phys. Chim. Debrecina",
  ["acta physica hungarica"] = "Acta Phys. Hung.",
  ["acta physica hungarica b quantum electronics"] = "Acta Phys. Hung. B",
  ["acta physica hungarica series a, heavy ion physics"] = "Acta Phys. Hung. A",
  ["acta physica polonica"] = "Acta Phys. Pol.",
  ["acta physica polonica a"] = "Acta Phys. Pol. A",
  ["acta physica polonica b"] = "Acta Phys. Pol. B",
  ["acta physica polonica proceedings supplement"] = "Acta Phys. Pol. B. Proc. Suppl.",
  ["acta physica polonica, a"] = "Acta Phys. Pol., A",
  ["acta physica polonica, b"] = "Acta Phys. Pol., B",
  ["acta physica polonica-series a general physics"] = "Acta Phys. Pol.-Ser. Gen. Phys.",
  ["acta physica polonica. b, proceedings supplement"] = "Acta Phys Pol B Proc Suppl",
  ["acta physica polonica: a"] = "Acta Phys Pol A",
  ["acta physica sinica"] = "Acta Phys. Sinica",
  ["acta physico-chimica sinica"] = "Acta Phys. Chim. Sin.",
  ["acta physiologiae plantarum"] = "Acta Physiol Plant",
  ["acta physiologiae plantarum / polish academy of sciences, committee of plant physiology genetics and breeding"] = "Acta Physiol Plant",
  ["acta physiologica"] = "Acta Physiol.",
  ["acta physiologica (oxford, england)"] = "Acta Physiol (Oxf)",
  ["acta physiologica academiae scientiarum hungaricae"] = "Acta Physiol Acad Sci Hung",
  ["acta physiologica et pharmacologica bulgarica"] = "Acta Physiol Pharmacol Bulg",
  ["acta physiologica et pharmacologica latinoamericana"] = "Acta Physiol. Pharmacol. Latinoam.",
  ["acta physiologica et pharmacologica latinoamericana : organo de la asociacion latinoamericana de ciencias fisiologicas y de la asociacion latinoamericana de farmacologia"] = "Acta Physiol Pharmacol Latinoam",
  ["acta physiologica et pharmacologica latinoamericana : organo de la asociación latinoamericana de ciencias fisiológicas y de la asociación latinoamericana de farmacología"] = "Acta Physiol Pharmacol Latinoam",
  ["acta physiologica et pharmacologica neerlandica"] = "Acta Physiol Pharmacol Neerl",
  ["acta physiologica hungarica"] = "Acta Physiol Hung",
  ["acta physiologica latino americana"] = "Acta Physiol Lat Am",
  ["acta physiologica latinoamericana"] = "Acta Physiol. Lat. Am.",
  ["acta physiologica polonica"] = "Acta Physiol Pol",
  ["acta physiologica scandinavica"] = "Acta Physiol Scand",
  ["acta physiologica scandinavica. supplementum"] = "Acta Physiol Scand Suppl",
  ["acta physiologica, pharmacologica et therapeutica latinoamericana"] = "Acta Physiol. Pharmacol. Ther. Latinoam.",
  ["acta physiologica, pharmacologica et therapeutica latinoamericana : organo de la asociacion latinoamericana de ciencias fisiologicas y [de] la asociacion latinoamericana de farmacologia"] = "Acta Physiol Pharmacol Ther Latinoam",
  ["acta physiologica, pharmacologica et therapeutica latinoamericana : órgano de la asociación latinoamericana de ciencias fisiológicas y [de] la asociación latinoamericana de farmacología"] = "Acta Physiol Pharmacol Ther Latinoam",
  ["acta physiotherapica et rheumatologica belgica"] = "Acta Physiother Rheumatol Belg",
  ["acta phytogeographica suecica"] = "Acta Phytogeogr. Suec.",
  ["acta phytomedica"] = "Acta Phytomed.",
  ["acta phytopathologica et entomologica hungarica"] = "Acta Phytopathol Entomol Hung",
  ["acta phytotaxonomica et geobotanica"] = "Acta Phytotaxon. Geobotan.",
  ["acta poloniae historica"] = "Acta Pol Hist",
  ["acta poloniae pharmaceutica"] = "Acta Pol Pharm",
  ["acta polymerica"] = "Acta Polym.",
  ["acta polytechnica ctu proceedings"] = "Acta Polytech CTU Proc",
  ["acta polytechnica hungarica"] = "Acta Polytech. Hung.",
  ["acta polytechnica scandinavica"] = "Acta Polytech. Scand. Math. Manage. Comput. Eng. Ser.",
  ["acta polytechnica scandinavica - chemical technology series"] = "Acta Polytech. Scand., Chem. Technol. Ser",
  ["acta praehistorica et archaeologica"] = "ActaPraehistA",
  ["acta protozoologica"] = "Acta Protozool",
  ["acta psiquiatrica y psicologica de america latina"] = "Acta Psiquiatr. Psicol. Am. Lat.",
  ["acta psiquiátrica y psicológica de américa latina"] = "Acta Psiquiatr Psicol Am Lat",
  ["acta psychiatrica belgica"] = "Acta Psychiatr Belg",
  ["acta psychiatrica et neurologica"] = "Acta Psychiatr Neurol",
  ["acta psychiatrica et neurologica scandinavica"] = "Acta Psychiatr Neurol Scand",
  ["acta psychiatrica et neurologica scandinavica. supplementum"] = "Acta Psychiatr Neurol Scand Suppl",
  ["acta psychiatrica et neurologica. supplementum"] = "Acta Psychiatr Neurol Suppl",
  ["acta psychiatrica scandinavica"] = "Acta Psychiatr Scand",
  ["acta psychiatrica scandinavica. supplementum"] = "Acta Psychiatr Scand Suppl",
  ["acta psychologica"] = "Acta Psychol (Amst)",
  ["acta psychotherapeutica et psychosomatica"] = "Acta Psychother Psychosom",
  ["acta psychotherapeutica, psychosomatica et orthopaedagogica"] = "Acta Psychother Psychosom Orthopaedagog",
  ["acta radiologica"] = "Acta radiol",
  ["acta radiologica (stockholm, sweden : 1987)"] = "Acta Radiol",
  ["acta radiologica et cancerologica. bohemoslovencia"] = "Acta Radiol Cancerol",
  ["acta radiologica open"] = "Acta Radiol Open",
  ["acta radiologica short reports"] = "Acta Radiol Short Rep",
  ["acta radiologica. diagnosis"] = "Acta Radiol. Diagn. (Stockh.)",
  ["acta radiologica. oncology"] = "Acta Radiol Oncol",
  ["acta radiologica. oncology, radiation therapy, physics and biology"] = "Acta Radiol Oncol Radiat Ther Phys Biol",
  ["acta radiologica. oncology, radiation, physics, biology"] = "Acta Radiol. Oncol. Radiat. Phys. Biol.",
  ["acta radiologica. supplement"] = "Acta Radiol Suppl (Stockholm)",
  ["acta radiologica. supplementum"] = "Acta Radiol Suppl",
  ["acta radiologica. therapy, physics, biology"] = "Acta Radiol. Ther. Phys. Biol.",
  ["acta radiologica: diagnosis"] = "Acta Radiol Diagn (Stockh)",
  ["acta radiologica: oncology, radiation, physics, biology"] = "Acta Radiol Oncol Radiat Phys Biol",
  ["acta radiologica: therapy, physics, biology"] = "Acta Radiol Ther Phys Biol",
  ["acta reproductiva turcica"] = "Acta Reprod Turc",
  ["acta reumatologica portuguesa"] = "Acta Reumatol Port",
  ["acta reumatológica portuguesa"] = "Acta Reumatol Port",
  ["acta rheumatologica scandinavica"] = "Acta Rheumatol Scand",
  ["acta rheumatologica scandinavica. supplementum"] = "Acta Rheumatol Scand Suppl",
  ["acta rhumatologica"] = "Acta Rhumatol",
  ["acta rhumatologica belgica"] = "Acta Rhumatol Belg",
  ["acta salmanticensia. filosofia y letras"] = "Acta Salmanticensia Ser Filos Letra",
  ["acta salmanticensia. filosofía y letras"] = "Acta Salmanticensia Ser Filos Letra",
  ["acta salmanticensia. historia de la medicina"] = "Acta Salmant Hist Med",
  ["acta salmanticensia: coleccion de memorias y trabajos cientificos editados por la universidad de salamanca, serie de ciencias"] = "Acta Salmanticensia: Coleccion de Memorias y Trabajos Cientificos Editados por la Universidad de Salamanca, Serie",
  ["acta scholae medicinalis universitatis in kioto"] = "Acta Sch Med Univ Kioto",
  ["acta scientia sinica = chung-kuo k'o hsüeh"] = "Acta Sci Sin",
  ["acta scientiae veterinariae"] = "Acta Sci Vet",
  ["acta scientiarum mathematicarum"] = "Acta Sci. Math.",
  ["acta scientiarum naturalium academiae scientiarum bohemoslovaca"] = "Acta sci. nat. Acad. sci. Bohemoslov.",
  ["acta scientiarum naturalium universitatis jilinensis"] = "Acta Sci. Natur. Univ. Jilin.",
  ["acta scientiarum naturalium universitatis normalis hunanensis"] = "Acta Sci. Natur. Univ. Norm. Hunan.",
  ["acta scientiarum naturalium universitatis sunyatseni"] = "Acta Sci. Natur. Univ. Sunyatseni",
  ["acta scientiarum naturalium universitatis sunyatseni. zhongshan daxue xuebao. ziran kexue ban"] = "Acta Sci. Natur. Univ. Sunyatseni",
  ["acta scientiarum naturalium universitatis sunyatseni. zhongshan daxue xuebao. ziran kexue ban. (zheng ying wen)"] = "Acta Sci. Natur. Univ. Sunyatseni",
  ["acta scientiarum polonorum. technologia alimentaria"] = "Acta Sci Pol Technol Aliment",
  ["acta scientiarum-agronomy"] = "Acta Sci.-Agron.",
  ["acta scientiarum. agronomy"] = "Acta Sci Agron",
  ["acta scientiarum. animal sciences"] = "Acta Sci",
  ["acta scientiarum. biological sciences"] = "Acta Sci Biol Sci",
  ["acta scientific cancer biology"] = "Acta Sci Cancer Biol",
  ["acta scientific neurology"] = "Acta Sci Neurol",
  ["acta seismologica sinica"] = "Acta Seismol. Sin.",
  ["acta societatis botanicorum poloniae. polskie towarzystwo botaniczne"] = "Acta Soc Bot Pol Pol Tow Bot",
  ["acta societatis medicorum fennicae duodecim. series a. suomalainen lääkäriseura duodecim"] = "Acta Suom Laak Duodecim A",
  ["acta societatis medicorum fennicae duodecim. series b. suomalainen lääkäriseura duodecim"] = "Acta Suom laak Duodecim B",
  ["acta societatis medicorum upsaliensis"] = "Acta Soc Med Ups",
  ["acta societatis medicorum upsaliensis. supplementum"] = "Acta Soc Med Ups Suppl",
  ["acta societatis pro fauna et flora fennica"] = "Acta Soc. Fauna Flora Fenn.",
  ["acta societatis zoologicae bohemoslovacae"] = "Acta Soc Zool Bohem",
  ["acta socio-medica scandinavica"] = "Acta Sociomed Scand",
  ["acta socio-medica scandinavica. supplement"] = "Acta Sociomed Scand Suppl",
  ["acta sociologica"] = "Acta Sociol.",
  ["acta sociologica (copenhagen, denmark)"] = "Acta Sociol",
  ["acta stereologica"] = "Acta Stereol",
  ["acta stomatologica belgica"] = "Acta Stomatol Belg",
  ["acta stomatologica croatica"] = "Acta Stomatol Croat",
  ["acta stomatologica internationalia"] = "Acta Stomatol. Int.",
  ["acta stomatologica internationalia : revue scientifique et bulletin officiel de l'association stomatologique internationale"] = "Acta Stomatol Int",
  ["acta stomatologica internationalia : revue scientifique et bulletin officiel de l’association stomatologique internationale"] = "Acta Stomatol Int",
  ["acta technica"] = "Acta Tech.",
  ["acta technica csav"] = "Acta Tech. CSAV",
  ["acta therapeutica"] = "Acta Ther",
  ["acta theriologica"] = "Acta Theriol (Warsz)",
  ["acta tropica"] = "Acta Trop",
  ["acta tropica. supplement"] = "Acta Trop. Suppl.",
  ["acta tropica. supplementum"] = "Acta Trop Suppl",
  ["acta tuberculosea belgica"] = "Acta Tuberc Belg",
  ["acta tuberculosea et pneumologica belgica"] = "Acta Tuberc Pneumol Belg",
  ["acta tuberculosea et pneumologica scandinavica"] = "Acta Tuberc Pneumol Scand",
  ["acta tuberculosea et pneumologica scandinavica. supplementum"] = "Acta Tuberc Pneumol Scand Suppl",
  ["acta tuberculosea japonica"] = "Acta Tuberc Jpn",
  ["acta tuberculosea scandinavica"] = "Acta Tuberc Scand",
  ["acta tuberculosea scandinavica. supplementum"] = "Acta Tuberc Scand Suppl",
  ["acta univ. de attila jozsef nominatae, acta historica"] = "AHSzeged",
  ["acta univ. de attila józsef nominatae, acta antiqua et archaeologica."] = "AAASzeged",
  ["acta universitaria"] = "Acta Univ",
  ["acta universitatis agriculturae, brno. sborník. rada b: spisy fakulty veterinární"] = "Acta Univ Agric Brno Sb Spisy Fak Vet",
  ["acta universitatis agriculturae. facultas agronomica. vysoka skola zemedelska v brne. fakulta agronomicka"] = "Acta Univ Agric Fac Agron",
  ["acta universitatis agriculturae. facultas agronomica. vysoká škola zemědělská v brně. fakulta agronomická"] = "Acta Univ Agric Fac Agron",
  ["acta universitatis apulensis. mathematics. informatics"] = "Acta Univ. Apulensis Math. Inform.",
  ["acta universitatis carolinae"] = "Acta Univ. Carolin. Math. Phys.",
  ["acta universitatis carolinae. geographica. univerzita karlova"] = "Acta Univ Carol Geogr",
  ["acta universitatis carolinae. historia universitatis carolinae pragensis. univerzita karlova"] = "Acta Univ Carol Univerzita Karlov",
  ["acta universitatis carolinae. medica"] = "Acta Univ Carol [Med] (Praha)",
  ["acta universitatis carolinae. medica. monographia"] = "Acta Univ Carol Med Monogr",
  ["acta universitatis carolinae. philosophica et historica"] = "Acta Univ Carol Philos Hist",
  ["acta universitatis lodziensis"] = "Acta Univ. Lodz. Folia Math.",
  ["acta universitatis lundensis. sectio i, theologica, juridica, humaniora"] = "Acta Univ Lund",
  ["acta universitatis matthiae belii. series mathematics"] = "Acta Univ. M. Belii Ser. Math.",
  ["acta universitatis nicolai copernici. archaeologia"] = "ActaTorunA",
  ["acta universitatis nicolai copernici. historia"] = "ActaTorunHist",
  ["acta universitatis ouluensis"] = "Acta Univ. Oulu. Ser. A Sci. Rerum Natur.",
  ["acta universitatis palackianae olomucensis"] = "Acta Univ. Palack. Olomuc. Fac. Rerum Natur. Math.",
  ["acta universitatis palackianae olomucensis facultatis medicae"] = "Acta Univ Palacki Olomuc Fac Med",
  ["acta universitatis palackianae olomucensis. facultas rerum naturalium. mathematica"] = "Acta Univ. Palack. Olomuc. Fac. Rerum Natur. Math.",
  ["acta universitatis sapientiae. mathematica"] = "Acta Univ Sapientiae Matem",
  ["acta universitatis sapientiae. mathematica. an international scientific journal of sapientia university"] = "Acta Univ. Sapientiae Math.",
  ["acta universitatis stockholmiensis. stockholm studies in english"] = "Stockh Stud Engl",
  ["acta universitatis szegediensis"] = "Acta Sci. Math. (Szeged)",
  ["acta universitatis szegediensis de attila jozsef nominatae. acta juridica et politica"] = "Acta Univ Szeged Attila Jozsef Nominatae",
  ["acta universitatis szegediensis de attila józsef nominatae. acta juridica et politica"] = "Acta Univ Szeged Attila Jozsef Nominatae",
  ["acta universitatis szegediensis. acta scientiarum mathematicarum"] = "Acta Sci. Math. (Szeged)",
  ["acta universitatis tamperensis"] = "Acta Univ. Tamper. Ser. A",
  ["acta universitatis upsaliensis"] = "Acta Univ. Ups. Stud. Philos. Ups.",
  ["acta universitatis upsaliensis : studia germanistica upsaliensia"] = "A. U. U. St. Ge.",
  ["acta universitatis upsaliensis. studia sociologica upsaliensia"] = "Acta Univ Ups",
  ["acta universitatis upsaliensis. symbolae botanicae upsalienses : arbeten från botaniska institutionerna i uppsala"] = "Acta Univ Ups Symb Bot Ups",
  ["acta universitatis wratislaviensis"] = "Acta Univ. Wratislav.",
  ["acta universitatis wratislaviensis. classica wratislaviensia."] = "ClassWrat",
  ["acta urologica"] = "Acta Urol",
  ["acta urologica belgica"] = "Acta Urol Belg",
  ["acta veterinaria"] = "Acta Vet Brno",
  ["acta veterinaria academiae scientiarum hungaricae"] = "Acta Vet Acad Sci Hung",
  ["acta veterinaria hungarica"] = "Acta Vet Hung",
  ["acta veterinaria indonesiana"] = "Acta Vet Indones",
  ["acta veterinaria scandinavica"] = "Acta Vet Scand",
  ["acta veterinaria scandinavica. supplement"] = "Acta Vet. Scand. Suppl.",
  ["acta veterinaria scandinavica. supplementum"] = "Acta Vet Scand Suppl",
  ["acta virologica"] = "Acta Virol",
  ["acta vitaminologica"] = "Acta Vitaminol",
  ["acta vitaminologica et enzymologica"] = "Acta Vitaminol Enzymol",
  ["acta zoobot austria"] = "Acta Zoobot Austria",
  ["acta zoologica"] = "Acta Zool.",
  ["acta zoologica (stockholm, sweden)"] = "Acta Zool",
  ["acta zoologica academiae scientiarum hungaricae (budapest, hungary : 1994)"] = "Acta Zool Academ Sci Hung",
  ["acta zoologica bulgarica"] = "Acta Zool Bulg",
  ["acta zoologica et pathologica antverpiensia"] = "Acta Zool Pathol Antverp",
  ["acta zoologica lituanica"] = "Acta Zool Litu",
  ["acta zoologica mexicana"] = "Acta Zool Mex",
  ["acta zoologica taiwanica (1992)"] = "Acta Zool Taiwanica",
  ["actas ciba"] = "Actas Ciba (Rio J)",
  ["actas de dermatologia & dermatopatologia"] = "Actas Dermatol Dermatopatol",
  ["actas de dermatología & dermatopatología"] = "Actas Dermatol Dermatopatol",
  ["actas de las reuniones cientificas del cuerpo facultativo del instituto policlinico"] = "Actas Reun Cient Cuerfo Fac Inst Policlin",
  ["actas de las reuniones científicas del cuerpo facultativo del instituto policlínico"] = "Actas Reun Cient Cuerfo Fac Inst Policlin",
  ["actas del congreso “dr. antonio a. r. monteiro”"] = "Actas Congr. “Dr. Antonio A. R. Monteiro”",
  ["actas dermo-sifiliograficas"] = "Actas Dermosifiliogr.",
  ["actas dermo-sifiliográficas"] = "Actas Dermosifiliogr",
  ["actas espanolas de psiquiatria"] = "Actas Esp. Psiquiatr.",
  ["actas españolas de psiquiatría"] = "Actas Esp Psiquiatr",
  ["actas luso-espanolas de neurologia y psiquiatria"] = "Actas Luso. Esp. Neurol. Psiquiatr.",
  ["actas luso-espanolas de neurologia, psiquiatria y ciencias afines"] = "Actas Luso. Esp. Neurol. Psiquiatr. Cienc. Afines",
  ["actas luso-españolas de neurología y psiquiatría"] = "Actas Luso Esp Neurol Psiquiatr",
  ["actas luso-españolas de neurología, psiquiatría y ciencias afines"] = "Actas Luso Esp Neurol Psiquiatr Cienc Afines",
  ["actas urologicas espanolas"] = "Actas Urol Esp",
  ["actas urologicas españolas"] = "Actas Urol Esp",
  ["actes de la recherche en sciences sociales"] = "Actes Rech Sci Soc",
  ["actes de la société helvétique des sciences naturelles"] = "Actes Soc. helv. sci. nat.",
  ["actes de la société jurassienne d’émulation"] = "Actes Soc. jura. émul.",
  ["actes des rencontres du cirm"] = "Actes Rencontres CIRM",
  ["actex academic series"] = "ACTEX Acad. Ser.",
  ["actex professional series"] = "ACTEX Prof. Ser.",
  ["action (national fund for research into crippling diseases)"] = "Action Natl",
  ["action contre le sida"] = "Action Contre SIDA",
  ["action for children"] = "Action Child",
  ["action learning"] = "Action Learn",
  ["action research (london)"] = "Action Res (Lond)",
  ["actionaid disability news : the newsletter of disability division, actionaid, india"] = "Actionaid Disabil News",
  ["active and passive electronic components"] = "Act. Passive Electron. Compon.",
  ["activitas nervosa superior"] = "Act Nerv Super (Praha)",
  ["activitas nervosa superior (2007)"] = "Act Nerv Super (2007)",
  ["activities directors' quarterly for alzheimer's & other dementia patients"] = "Act Dir Q Alzheimers Other Dement Patients",
  ["activities, adaptation & aging"] = "Act Adapt Aging",
  ["actualidad juridica aranzadi"] = "Actual Jurid Aranzadi",
  ["actualidad jurídica aranzadi"] = "Actual Jurid Aranzadi",
  ["actualidad medica"] = "Actual Medica",
  ["actualidad médica"] = "Actual Medica",
  ["actualidad médica peruana"] = "Actual Med Peru",
  ["actualidad odontoestomatologica española"] = "Actual Odonto Estomatol Esp",
  ["actualidad pediatrica"] = "Actual. Pediatr. (Granada)",
  ["actualidad pediatrica; revista de bibliografia internacional"] = "Actual Pediatr (Granada)",
  ["actualidad pediátrica; revista de bibliografía internacional"] = "Actual Pediatr (Granada)",
  ["actualite chimique"] = "Actual. Chim.",
  ["actualites anatomo-pathologiques"] = "Actual Anatpathol",
  ["actualites biologiques"] = "Actual Biol",
  ["actualites cardiologiques et angeiologiques internationales"] = "Actual. Cardiol. Angeiol. Int. (Paris)",
  ["actualites de physiologie pathologique"] = "Actual. Physiol. Pathol. (Paris)",
  ["actualites endocrinologiques"] = "Actual. Endocrinol. (Paris)",
  ["actualites hepato-gastro-enterologiques"] = "Actual. Hepatogastroenterol. (Paris)",
  ["actualites neurophysiologiques"] = "Actual. Neurophysiol. (Paris)",
  ["actualites odonto-stomatologiques"] = "Actual. Odontostomatol. (Paris)",
  ["actualites pharmaceutiques"] = "Actual Pharm",
  ["actualites pharmacologiques"] = "Actual. Pharmacol. (Paris)",
  ["actualites psychiatriques"] = "Actual Psychiatr",
  ["actualites scientifiques et agronomiques de l’inra"] = "Actualites Sci. Agronom. INRA",
  ["actualités mathématiques"] = "Actualités Math.",
  ["actualités scientifiques et industrielles"] = "Actualités Sci. Indust.",
  ["actuelle urologie"] = "Actuel Urol",
  ["acupuncture & electro-therapeutics research"] = "Acupunct Electrother Res",
  ["acupuncture and electro-therapeutics research"] = "Acupunct. Electrother. Res.",
  ["acupuncture in medicine : journal of the british medical acupuncture society"] = "Acupunct Med",
  ["acute and critical care"] = "Acute Crit Care",
  ["acute cardiac care"] = "Acute Card. Care",
  ["acute care"] = "Acute Care",
  ["acute medicine"] = "Acute Med",
  ["acute medicine & surgery"] = "Acute Med Surg",
  ["ad astra (washington, d.c.)"] = "Ad Astra",
  ["ad hoc networks"] = "Ad Hoc Networks",
  ["ad nurse"] = "AD Nurse",
  ["ada news"] = "ADA News",
  ["adalya. annual of the suna & inan kiraç-research institute on mediterranean civilizations"] = "Adalya",
  ["adaptation, learning, and optimization"] = "Adapt. Learn. Optim.",
  ["adapted physical activity quarterly : apaq"] = "Adapt Phys Activ Q",
  ["adaptive and learning systems for signal processing, communications, and control"] = "Adapt. Learn. Syst. Signal Process. Commun. Control",
  ["adaptive behavior"] = "Adapt. Behav.",
  ["adaptive computation and machine learning"] = "Adapt. Comput. Mach. Learn.",
  ["adaptive human behavior and physiology"] = "Adapt. Hum. Behav. Physiol.",
  ["adcaij: advances in distributed computing and artificial intelligence journal"] = "ADCAIJ: Adv. Distrib. Comput. and Artif. Intell. J.",
  ["addiction & health"] = "Addict Health",
  ["addiction (abingdon, england)"] = "Addiction",
  ["addiction biology"] = "Addict Biol",
  ["addiction neuroscience"] = "Addict. Neurosci.",
  ["addiction research & theory"] = "Addict. Res. Theory",
  ["addiction science & clinical practice"] = "Addict Sci Clin Pract",
  ["addictive behaviors"] = "Addict Behav",
  ["addictive behaviors reports"] = "Addict Behav Rep",
  ["addictive diseases"] = "Addict Dis",
  ["addictive disorders & their treatment"] = "Addict Disord Their Treat",
  ["addison-wesley series in computer science and information processing"] = "Addison-Wesley Ser. Comput. Sci. Inform. Process.",
  ["addison-wesley series in mathematics"] = "Addison-Wesley Ser. Math.",
  ["addison-wesley series in statistics"] = "Addison-Wesley Ser. Statist.",
  ["additamenta ad folia medica neerlandica"] = "Additamenta Folia Med. Neerl.",
  ["additive manufacturing"] = "Addit. Manuf.",
  ["additive manufacturing letters"] = "Addit. Manuf. Lett.",
  ["additives for polymers"] = "Addit. Polym.",
  ["adhaesion–kleben & dichten"] = "Adhaes. Kleben Dicten",
  ["adhasion–kleben & dichten"] = "Adhas. Kleben Dicten",
  ["adhesives age"] = "Adhes. Age",
  ["adjacent government"] = "Adjac Gov",
  ["adler museum bulletin"] = "Adler Mus Bull",
  ["adm (asociacion dental mexicana : 1986)"] = "ADM",
  ["adm (asociación dental mexicana : 1986)"] = "ADM",
  ["adm; revista de la asociacion dental mexicana"] = "ADM",
  ["adm; revista de la asociación dental mexicana"] = "ADM",
  ["admet & dmpk"] = "ADMET DMPK",
  ["administration & society"] = "Adm Soc",
  ["administration and policy in mental health"] = "Adm Policy Ment Health",
  ["administration for development"] = "Adm Dev",
  ["administration in mental health"] = "Adm Ment Health",
  ["administration in social work"] = "Adm Soc Work",
  ["administrative change"] = "Adm Change",
  ["administrative law journal"] = "Adm Law J",
  ["administrative law review"] = "Adm Law Rev",
  ["administrative management"] = "Adm Manage",
  ["administrative radiology"] = "Adm. Radiol.",
  ["administrative radiology : ar"] = "Adm Radiol",
  ["administrative radiology journal"] = "Adm. Radiol. J.",
  ["administrative radiology journal : ar"] = "Adm Radiol J",
  ["administrative science quarterly"] = "Adm Sci Q",
  ["administrative sciences"] = "Adm Sci",
  ["admitting management journal"] = "Admit. Manage. J.",
  ["adolescence education newsletter"] = "Adolesc Educ Newsl",
  ["adolescent & family health : a & fh"] = "Adolesc Fam Health",
  ["adolescent and pediatric gynecology"] = "Adolesc Pediatr Gynecol",
  ["adolescent health, medicine and therapeutics"] = "Adolesc Health Med Ther",
  ["adolescent medicine"] = "Adolesc. Med.",
  ["adolescent medicine (philadelphia, pa.)"] = "Adolesc Med",
  ["adolescent medicine clinics"] = "Adolesc Med Clin",
  ["adolescent medicine: state of the art reviews"] = "Adolesc Med State Art Rev",
  ["adolescent psychiatry"] = "Adolesc Psychiatry",
  ["adolescent psychiatry (hilversum, netherlands)"] = "Adolesc Psychiatry (Hilversum)",
  ["adolescent research review"] = "Adolesc Res Rev",
  ["adolescência & saúde"] = "Adolesc Saude",
  ["adoption & fostering"] = "Adopt Foster",
  ["adoption quarterly"] = "Adopt Q",
  ["adsorption : journal of the international adsorption society"] = "Adsorption (Boston)",
  ["adsorption science & technology"] = "Adsorpt. Sci. Technol.",
  ["adsorption science & technology : interface science for advanced materials & technologies"] = "Adsorp Sci Technol",
  ["adsorption science and technology"] = "Adsorpt. Sci. Technol.",
  ["adult basic education and literacy journal"] = "Adult Basic Educ Lit J",
  ["adult education"] = "Adult Educ",
  ["adult education quarterly (american association for adult and continuing education)"] = "Adult Educ Q (Am Assoc Adult Contin Educ)",
  ["adult foster care journal : afcj"] = "Adult Foster Care J",
  ["adult residential care journal"] = "Adult Resid Care J",
  ["adumatu. a semi-annual archeological refereed journal on the arab world"] = "Adumatu",
  ["advance data"] = "Adv Data",
  ["advance for administrators of the laboratory"] = "Adv Adm Lab",
  ["advance for nps & pas"] = "Adv NPs PAs",
  ["advance for nurse practitioners"] = "Adv Nurse Pract",
  ["advance for occupational therapy practitioners"] = "Adv Occup Ther Pract",
  ["advance for physical therapy and rehab medicine"] = "Adv Phys Ther Rehab Med",
  ["advance journal of food science and technology : ajfst"] = "Adv J Food Sci Technol",
  ["advance research on endocrinology & metabolism"] = "Adv Res Endocrinol Metab",
  ["advance trends in soft computing : proceedings of wcsc 2013, december 16-18, san antonio, texas, usa. wcsc (conference : soft computing) (2013 : san antonio, tex.)"] = "Adv Trends Soft Comput (2013)",
  ["advanced agrochem"] = "Adv. Agrochem",
  ["advanced applications in pattern recognition"] = "Adv. Appl. Pattern Recogn.",
  ["advanced bio tech : india's biotechnology magazine"] = "Adv Bio Tech",
  ["advanced biology"] = "Adv Biol",
  ["advanced biomedical engineering"] = "Adv. Biomed. Eng.",
  ["advanced biomedical research"] = "Adv Biomed Res",
  ["advanced bionics, biokinetics and biocybernetics"] = "Adv. Bionics Biokinet. Biocybern.",
  ["advanced biosystems"] = "Adv Biosyst",
  ["advanced book classics"] = "Adv. Book Class.",
  ["advanced cement based materials"] = "Adv. Cem. Based Mater.",
  ["advanced composite materials"] = "Adv. Compos. Mater",
  ["advanced composites and hybrid materials"] = "Adv Compos Hybrid Mater",
  ["advanced composites bulletin"] = "Adv. Compos. Bull.",
  ["advanced composites letters"] = "Adv. Compos. Lett.",
  ["advanced control for applications"] = "Adv. Control Appl.",
  ["advanced courses in mathematics. crm barcelona"] = "Adv. Courses Math. CRM Barcelona",
  ["advanced device materials"] = "Adv. Device Mater.",
  ["advanced devices & instrumentation"] = "Adv. Devices Instrum.",
  ["advanced drug delivery reviews"] = "Adv Drug Deliv Rev",
  ["advanced electronic materials"] = "Adv. Electron. Mater.",
  ["advanced emergency nursing journal"] = "Adv Emerg Nurs J",
  ["advanced energy and sustainability research"] = "Adv. Energy Sustainability Res.",
  ["advanced energy conversion"] = "Adv. Energy Convers.",
  ["advanced energy materials"] = "Adv. Energy Mater.",
  ["advanced engineering informatics"] = "Adv. Eng. Inf.",
  ["advanced engineering materials"] = "Adv. Eng. Mater.",
  ["advanced fiber materials"] = "Adv. Fiber Mater.",
  ["advanced functional materials"] = "Adv. Funct. Mater.",
  ["advanced genetics"] = "Adv. Genet.",
  ["advanced genetics (hoboken, n.j.)"] = "Adv Genet (Hoboken)",
  ["advanced gut & microbiome research"] = "Adv. Gut Microbiome Res.",
  ["advanced health care technologies"] = "Adv Health Care Technol",
  ["advanced healthcare materials"] = "Adv Healthc Mater",
  ["advanced imaging (woodbury, n.y.)"] = "Adv Imaging",
  ["advanced industrial and engineering polymer research"] = "Adv. Ind. Eng. Polym. Res.",
  ["advanced information and knowledge processing"] = "Adv. Inf. Knowl. Process.",
  ["advanced intelligent discovery"] = "Adv. Intell. Discovery",
  ["advanced intelligent systems"] = "Adv. Intell. Syst.",
  ["advanced intelligent systems (weinheim an der bergstrasse, germany)"] = "Adv Intell Syst",
  ["advanced journal of emergency medicine"] = "Adv J Emerg Med",
  ["advanced lectures in mathematics"] = "Adv. Lectures Math.",
  ["advanced lectures in mathematics (alm)"] = "Adv. Lect. Math. (ALM)",
  ["advanced manufacturing: polymer & composites science"] = "Adv. Manuf. Polym. Compos. Sci.",
  ["advanced materials"] = "Adv. Mater.",
  ["advanced materials (deerfield beach, fla.)"] = "Adv Mater",
  ["advanced materials (weinheim, germany)"] = "Adv. Mater. (Weinheim, Ger.)",
  ["advanced materials and processes"] = "Adv. Mater. Processes",
  ["advanced materials for optics and electronics"] = "Adv. Mater. Opt. Electron.",
  ["advanced materials forum"] = "Adv. Mater. Forum",
  ["advanced materials interfaces"] = "Adv. Mater. Interfaces",
  ["advanced materials letters"] = "Adv Mater Lett",
  ["advanced materials research"] = "Adv Mat Res",
  ["advanced materials technologies"] = "Adv. Mater. Technol.",
  ["advanced membranes"] = "Adv. Membr.",
  ["advanced modeling and simulation in engineering sciences"] = "Adv. Model. Simul. Eng. Sci.",
  ["advanced nanobiomed research"] = "Adv Nanobiomed Res",
  ["advanced nonlinear studies"] = "Adv. Nonlinear Stud.",
  ["advanced optical materials"] = "Adv. Opt. Mater.",
  ["advanced optical technologies"] = "Adv. Opt. Technol.",
  ["advanced performance materials"] = "Adv. Perform. Mater.",
  ["advanced pharmaceutical bulletin"] = "Adv Pharm Bull",
  ["advanced photonics"] = "Adv. Photonics",
  ["advanced photonics nexus"] = "Adv. Photonics Nexus",
  ["advanced photonics research"] = "Adv Photonics Res",
  ["advanced physics research"] = "Adv. Phys. Res.",
  ["advanced powder materials"] = "Adv. Powder Mater.",
  ["advanced powder technology"] = "Adv. Powder Technol.",
  ["advanced powder technology : the international journal of the society of powder technology, japan"] = "Adv Powder Technol",
  ["advanced practice nurse : apn"] = "Adv Pract Nurse",
  ["advanced practice nursing quarterly"] = "Adv Pract Nurs Q",
  ["advanced quantitative techniques in the social sciences"] = "Adv. Quant. Tech. Soc. Sci.",
  ["advanced quantum technologies"] = "Adv. Quantum Technol.",
  ["advanced research in computing and software science"] = "Adv. Res. Comput. Softw. Sci.",
  ["advanced research in gastroenterology & hepatology"] = "Adv Res Gastroenterol Hepatol",
  ["advanced robotics"] = "Adv. Rob.",
  ["advanced robotics : the international journal of the robotics society of japan"] = "Adv Robot",
  ["advanced robotics research"] = "Adv. Rob. Res.",
  ["advanced science"] = "Adv. Sci.",
  ["advanced science (weinheim, baden-württemberg, germany)"] = "Adv Sci (Weinh)",
  ["advanced science letters"] = "Adv. Sci. Lett.",
  ["advanced science, engineering and medicine"] = "Adv Sci Eng Med",
  ["advanced sensor and energy materials"] = "Adv. Sens. Energy Mater.",
  ["advanced sensor research"] = "Adv. Sens. Res.",
  ["advanced series in applied physics"] = "Adv. Ser. Appl. Phys.",
  ["advanced series in dynamical systems"] = "Adv. Ser. Dynam. Systems",
  ["advanced series in electrical and computer engineering"] = "Adv. Ser. Electr. Comput. Eng.",
  ["advanced series in mathematical physics"] = "Adv. Ser. Math. Phys.",
  ["advanced series in mathematical science and engineering"] = "Adv. Ser. Math. Sci. Eng.",
  ["advanced series in neuroscience"] = "Adv. Ser. Neurosci.",
  ["advanced series in nonlinear dynamics"] = "Adv. Ser. Nonlinear Dynam.",
  ["advanced series on directions in high energy physics"] = "Adv. Ser. Dir. High Energy Phys.",
  ["advanced series on mathematical psychology"] = "Adv. Ser. Math. Psychol.",
  ["advanced series on statistical science & applied probability"] = "Adv. Ser. Stat. Sci. Appl. Probab.",
  ["advanced series on theoretical physical science"] = "Adv. Ser. Theoret. Phys. Sci.",
  ["advanced steel construction"] = "Adv. Steel Constr.",
  ["advanced structural and chemical imaging"] = "Adv Struct Chem Imaging",
  ["advanced structured materials"] = "Adv. Struct. Mater.",
  ["advanced studies in biology"] = "Adv Stud Biol",
  ["advanced studies in complex systems"] = "Adv. Stud. Complex Syst.",
  ["advanced studies in pure mathematics"] = "Adv. Stud. Pure Math.",
  ["advanced studies in theoretical and applied econometrics"] = "Adv. Stud. Theoret. Appl. Econometrics",
  ["advanced studies in theoretical physics"] = "Adv. Stud. Theor. Phys.",
  ["advanced studies. euro-tbilisi mathematical journal"] = "Adv. Stud. Euro-Tbil. Math. J.",
  ["advanced studies: euro-tbilisi mathematical journal"] = "Adv. Stud.: Euro-Tbilisi Math. J.",
  ["advanced sustainable systems"] = "Adv Sustain Syst",
  ["advanced synthesis & catalysis"] = "Adv Synth Catal",
  ["advanced synthesis and catalysis"] = "Adv. Synth. Catal.",
  ["advanced techniques in biology & medicine"] = "Adv Tech Biol Med",
  ["advanced textbooks in control and signal processing"] = "Adv. Textb. Control Signal Process.",
  ["advanced textbooks in mathematics"] = "Adv. Textb. Math.",
  ["advanced textbooks in physics"] = "Adv. Textb. Phys.",
  ["advanced texts in econometrics"] = "Adv. Texts Econometrics",
  ["advanced theory and simulations"] = "Adv Theory Simul",
  ["advanced therapeutics"] = "Adv Ther (Weinh)",
  ["advanced topics in mathematics"] = "Adv. Top. Math.",
  ["advanced topics in mechanical engineering series"] = "Adv. Topics Mech. Engrg. Ser.",
  ["advancement of science"] = "Adv Sci",
  ["advancements in genetic engineering"] = "Adv Genet Eng",
  ["advancements in journal of urology and nephrology"] = "Adv J Urol Nephrol",
  ["advancements in life sciences"] = "Adv Life Sci",
  ["advances and applications in bioinformatics and chemistry"] = "Adv. Appl Bioinf. Chem.",
  ["advances and applications in bioinformatics and chemistry : aabc"] = "Adv Appl Bioinform Chem",
  ["advances and applications in mathematical sciences"] = "Adv. Appl. Math. Sci.",
  ["advances and applications in statistical sciences"] = "Adv Appl Stat Sci",
  ["advances and applications in statistics"] = "Adv Appl Stat",
  ["advances and technical standards in neurosurgery"] = "Adv Tech Stand Neurosurg",
  ["advances in abortion care"] = "Adv Abort Care",
  ["advances in abortion care / ipas"] = "Adv Abort Care",
  ["advances in acoustics and vibration"] = "Adv Acoust Vib",
  ["advances in acoustics and vibrations"] = "Adv. Acoust. Vibr.",
  ["advances in adaptive data analysis"] = "Adv Adapt Data Anal",
  ["advances in adolescent mental health"] = "Adv Adolesc Mental Health",
  ["advances in aerodynamics"] = "Adv. Aerodyn.",
  ["advances in aging research"] = "Adv Aging Res",
  ["advances in agriculture"] = "Adv. Agric.",
  ["advances in agronomy"] = "Adv. Agron.",
  ["advances in air pollution series"] = "Adv. Air Pollut. Ser.",
  ["advances in aircraft & spacecraft science"] = "Adv Aircr Spacecr Sci",
  ["advances in alcohol & substance abuse"] = "Adv Alcohol Subst Abuse",
  ["advances in alcohol and substance abuse"] = "Adv. Alcohol Subst. Abuse",
  ["advances in alzheimer's disease"] = "Adv Alzheimer Dis",
  ["advances in analysis and geometry"] = "Adv. Anal. Geom.",
  ["advances in anatomic pathology"] = "Adv Anat Pathol",
  ["advances in anatomy, embryology and cell biology"] = "Adv. Anat. Embryol. Cell Biol.",
  ["advances in anatomy, embryology, and cell biology"] = "Adv Anat Embryol Cell Biol",
  ["advances in anesthesia"] = "Adv Anesth",
  ["advances in animal and veterinary sciences"] = "Adv Anim Vet Sci",
  ["advances in animal biosciences"] = "Adv. Anim. Biosci.",
  ["advances in antiviral drug design"] = "Adv Antivir Drug Des",
  ["advances in applied biodiversity science"] = "Adv. Appl. Biodivers. Sci.",
  ["advances in applied ceramics"] = "Adv. Appl. Ceram.",
  ["advances in applied clifford algebras"] = "Adv. Appl. Clifford Algebras",
  ["advances in applied energy"] = "Adv. Appl. Energy",
  ["advances in applied mathematics"] = "Adv. Appl. Math.",
  ["advances in applied mathematics and mechanics"] = "Adv. Appl. Math. Mech.",
  ["advances in applied mechanics"] = "Adv. Appl. Mech.",
  ["advances in applied microbiology"] = "Adv Appl Microbiol",
  ["advances in applied probability"] = "Adv. Appl. Probab.",
  ["advances in applied sociology"] = "Adv Appl Sociol",
  ["advances in applied statistics"] = "Adv. Appl. Statist.",
  ["advances in archaeological method and theory"] = "AMethTh",
  ["advances in archaeological practice"] = "Adv. Archaeol. Pract.",
  ["advances in archaeomaterials"] = "Adv. Archaeomater.",
  ["advances in artificial intelligence and machine learning"] = "Adv. Artif. Intell. Mach. Learn.",
  ["advances in artificial intelligence. canadian society for computational studies of intelligence. conference"] = "Adv Artif Intell",
  ["advances in astrobiology and biogeophysics"] = "Adv. Astrobiol. Biogeophys.",
  ["advances in astronautics science and technology"] = "Adv. Astronaut. Sci. Technol.",
  ["advances in astronomy"] = "Adv. Astron.",
  ["advances in atmospheric sciences"] = "Adv. Atmos. Sci.",
  ["advances in atomic and molecular physics"] = "Adv. At. Mol. Phys.",
  ["advances in atomic molecular and optical physics"] = "Adv. At. Mol. Opt. Phy.",
  ["advances in behavioral biology"] = "Adv Behav Biol",
  ["advances in biochemical engineering / biotechnology"] = "Adv. Biochem. Eng./Biotechnol.",
  ["advances in biochemical engineering/biotechnology"] = "Adv Biochem Eng Biotechnol",
  ["advances in biochemical psychopharmacology"] = "Adv Biochem Psychopharmacol",
  ["advances in biochemistry"] = "Adv Biochem",
  ["advances in bioethics"] = "Adv Bioeth",
  ["advances in bioinformatics"] = "Adv Bioinformatics",
  ["advances in biological and medical physics"] = "Adv Biol Med Phys",
  ["advances in biological chemistry"] = "Adv Biol Chem",
  ["advances in biological regulation"] = "Adv Biol Regul",
  ["advances in biological research"] = "Adv Biol Res (Rennes)",
  ["advances in biological sciences research"] = "Adv Biol Sci Res",
  ["advances in biology"] = "Adv Biol",
  ["advances in biology of skin"] = "Adv Biol Skin",
  ["advances in biomarker sciences and technology"] = "Adv Biomark Sci Technol",
  ["advances in biomedical engineering and medical physics"] = "Adv Biomed Eng Med Phys",
  ["advances in biomedical engineering research"] = "Adv Biomed Eng Res",
  ["advances in biophysics"] = "Adv Biophys",
  ["advances in bioresearch"] = "Adv Biores",
  ["advances in bioscience and bioengineering (new york, ny)"] = "Adv Biosci Bioeng (N Y)",
  ["advances in bioscience and biotechnology (print)"] = "Adv Biosci Biotechnol",
  ["advances in bioscience and clinical medicine"] = "Adv Biosci Clin Med",
  ["advances in biosensors and bioelectronics"] = "Adv Biosens Bioelectron",
  ["advances in biotechnological processes"] = "Adv Biotechnol Processes",
  ["advances in biotechnology & microbiology (newbury, calif.)"] = "Adv Biotechnol Microbiol",
  ["advances in botanical research"] = "Adv Bot Res",
  ["advances in boundary elements series"] = "Adv. Bound. Elem. Ser.",
  ["advances in breast cancer research"] = "Adv Breast Cancer Res",
  ["advances in bridge engineering"] = "Adv. Bridge Eng.",
  ["advances in building energy research"] = "Adv. Build. Energy Res.",
  ["advances in cad for vlsi"] = "Adv. CAD VLSI",
  ["advances in calculus of variations"] = "Adv. Calc. Var.",
  ["advances in cancer prevention"] = "Adv Cancer Prev",
  ["advances in cancer research"] = "Adv Cancer Res",
  ["advances in cancer: research & treatment"] = "Adv Cancer Res Treat",
  ["advances in carbohydrate chemistry"] = "Adv Carbohydr Chem",
  ["advances in carbohydrate chemistry and biochemistry"] = "Adv Carbohydr Chem Biochem",
  ["advances in cardiac surgery"] = "Adv Card Surg",
  ["advances in cardiology"] = "Adv Cardiol",
  ["advances in cardiopulmonary diseases"] = "Adv Cardiopulm Dis",
  ["advances in catalysis"] = "Adv. Catal.",
  ["advances in cell and gene therapy"] = "Adv Cell Gene Ther",
  ["advances in cell biology"] = "Adv Cell Biol",
  ["advances in cellular and molecular biology of plants"] = "Adv Cell Mol Biol Plants",
  ["advances in cement research"] = "Adv. Cem. Res.",
  ["advances in chemical physics"] = "Adv. Chem. Phys.",
  ["advances in chemistry research"] = "Adv Chem Res",
  ["advances in chemistry series"] = "Adv Chem Ser",
  ["advances in chemotherapy"] = "Adv Chemother",
  ["advances in child development and behavior"] = "Adv Child Dev Behav",
  ["advances in chromatography"] = "Adv Chromatogr",
  ["advances in chromatography (boca raton, fl, united states)"] = "Adv. Chromatogr. (Boca Raton, FL, U. S.)",
  ["advances in chromatography (new york, ny, united states)"] = "Adv. Chromatogr. (N. Y., NY, U. S.)",
  ["advances in chronic kidney disease"] = "Adv Chronic Kidney Dis",
  ["advances in civil engineering"] = "Adv. Civ. Eng.",
  ["advances in civil engineering materials"] = "Adv Civ Eng Mater",
  ["advances in climate change research"] = "Adv. Clim. Change Res.",
  ["advances in clinical and experimental medicine : official organ wroclaw medical university"] = "Adv Clin Exp Med",
  ["advances in clinical chemistry"] = "Adv Clin Chem",
  ["advances in clinical endocrinology and metabolism"] = "Adv Clin Endocrinol Metab",
  ["advances in clinical neuroscience & rehabilitation : acnr"] = "Adv Clin Neurosci Rehabil",
  ["advances in clinical pathology"] = "Adv. Clin. Path.",
  ["advances in clinical pathology : the official journal of adriatic society of pathology"] = "Adv Clin Path",
  ["advances in clinical pharmacology"] = "Adv Clin Pharmacol",
  ["advances in clinical radiology"] = "Adv. Clin. Radiol.",
  ["advances in clinical rehabilitation"] = "Adv Clin Rehabil",
  ["advances in cognitive neurodynamics"] = "Adv. Cogn. Neurodyn.",
  ["advances in cognitive psychology"] = "Adv Cogn Psychol",
  ["advances in colloid and interface science"] = "Adv Colloid Interface Sci",
  ["advances in combinatorics"] = "Adv. Comb.",
  ["advances in comparative physiology and biochemistry"] = "Adv Comp Physiol Biochem",
  ["advances in complex systems"] = "Adv. Complex Syst.",
  ["advances in complex systems. a multidisciplinary journal"] = "Adv. Complex Syst.",
  ["advances in computational fluid dynamics"] = "Adv. Comput. Fluid Dyn.",
  ["advances in computational intelligence"] = "Adv. Comput. Intell.",
  ["advances in computational intelligence and robotics (acir) book series"] = "Adv. Comput. Intell. Robot. ACIR Book Ser.",
  ["advances in computational management science"] = "Adv. Comput. Manag. Sci.",
  ["advances in computational mathematics"] = "Adv. Comput. Math.",
  ["advances in computational science and engineering"] = "Adv. Comput. Sci. Eng.",
  ["advances in computer and electrical engineering (acee) book series"] = "Adv. Comput. Electr. Eng. (ACEE) Book Ser.",
  ["advances in computer science and technology"] = "Adv. Comput. Sci. Technol.",
  ["advances in computer vision and pattern recognition"] = "Adv. Comput. Vis. Pattern Recognit.",
  ["advances in computing science"] = "Adv. Comput. Sci.",
  ["advances in concrete construction"] = "Adv. Concr. Constr.",
  ["advances in condensed matter physics"] = "Adv. Condens. Matter Phys.",
  ["advances in consumer research. association for consumer research (u.s.)"] = "Adv Consum Res",
  ["advances in continuous and discrete models"] = "Adv. Contin. Discrete Models",
  ["advances in continuous and discrete models. theory and modern applications"] = "Adv. Contin. Discrete Models",
  ["advances in continuum mechanics"] = "Adv. Contin. Mech.",
  ["advances in contraception"] = "Adv. Contracept.",
  ["advances in contraception : the official journal of the society for the advancement of contraception"] = "Adv Contracept",
  ["advances in contraceptive delivery systems : cds"] = "Adv Contracept Deliv Syst",
  ["advances in cryogenic engineering"] = "Adv Cryog Eng",
  ["advances in cyclic nucleotide and protein phosphorylation research"] = "Adv Cyclic Nucleotide Protein Phosphorylation Res",
  ["advances in cyclic nucleotide research"] = "Adv Cyclic Nucleotide Res",
  ["advances in cytopharmacology"] = "Adv Cytopharmacol",
  ["advances in damage mechanics"] = "Adv. Damage Mech.",
  ["advances in data analysis and classification"] = "Adv Data Anal Classif",
  ["advances in data analysis and classification. adac"] = "Adv. Data Anal. Classif.",
  ["advances in data analysis and classification. adac. theory, methods, and applications in data science"] = "Adv. Data Anal. Classif.",
  ["advances in data mining. industrial conference on data mining"] = "Adv Data Min",
  ["advances in data science and adaptive analysis"] = "Adv. Data Sci. Adapt. Anal.",
  ["advances in database technology : proceedings. international conference on extending database technology"] = "Adv Database Technol",
  ["advances in decision sciences"] = "Adv. Decis. Sci.",
  ["advances in delays and dynamics"] = "Adv. Delays Dyn.",
  ["advances in dental research"] = "Adv Dent Res",
  ["advances in dermatology"] = "Adv Dermatol",
  ["advances in design and control"] = "Adv. Des. Control",
  ["advances in developing human resources"] = "Adv Dev Hum Resour",
  ["advances in developmental biology"] = "Adv. Dev. Biol.",
  ["advances in developmental biology (amsterdam, netherlands)"] = "Adv Dev Biol",
  ["advances in difference equations"] = "Adv Differ Equ",
  ["advances in differential equations"] = "Adv. Differential Equations",
  ["advances in disease control and prevention"] = "Adv Dis Control Prev",
  ["advances in disease surveillance"] = "Adv Dis Surveill",
  ["advances in drug and alcohol research"] = "Adv. Drug Alcohol Res.",
  ["advances in drug research"] = "Adv Drug Res",
  ["advances in dual diagnosis"] = "Adv Dual Diagn",
  ["advances in dynamical systems and applications"] = "Adv. Dyn. Syst. Appl.",
  ["advances in eating disorders (abingdon, england )"] = "Adv Eat Disord",
  ["advances in ecological research"] = "Adv Ecol Res",
  ["advances in econometrics"] = "Adv. Econom.",
  ["advances in economics and business"] = "Adv Econ Bus",
  ["advances in electrochemical science and engineering"] = "Adv. Electrochem. Sci. Eng.",
  ["advances in endocrinology and metabolism"] = "Adv Endocrinol Metab",
  ["advances in engineering education"] = "Adv. Eng. Educ.",
  ["advances in engineering software"] = "Adv. Eng. Software",
  ["advances in engineering software (barking, london, england : 1992)"] = "Adv Eng Softw",
  ["advances in entomology (irvine, calif.)"] = "Adv Entomol",
  ["advances in environmental biology"] = "Adv Environ Biol",
  ["advances in environmental engineering and green technologies (aeegt) book series"] = "Adv. Environ. Eng. Green Technol. (AEEGT) Book Ser.",
  ["advances in environmental research"] = "Adv. Environ. Res.",
  ["advances in enzyme regulation"] = "Adv Enzyme Regul",
  ["advances in enzyme research"] = "Adv Enzyme Res",
  ["advances in enzymology and related areas of molecular biology"] = "Adv Enzymol Relat Areas Mol Biol",
  ["advances in enzymology and related subjects of biochemistry"] = "Adv Enzymol Relat Subj Biochem",
  ["advances in epidemiology"] = "Adv Epidemiol",
  ["advances in ethology"] = "Adv. Ethol.",
  ["advances in evolution equations"] = "Adv. Evol. Equ.",
  ["advances in evolutionary biology"] = "Adv Evol Biol",
  ["advances in experimental biology"] = "Adv Exp Biol",
  ["advances in experimental medicine and biology"] = "Adv Exp Med Biol",
  ["advances in experimental philosophy"] = "Adv. Exp. Philos.",
  ["advances in experimental social psychology"] = "Adv Exp Soc Psychol",
  ["advances in fertility control"] = "Adv Fertil Control",
  ["advances in filtration and separation technology"] = "Adv. Filtr. Sep. Technol.",
  ["advances in finance, accounting, and economics (afae) book series"] = "Adv. Finance Account. Econ. (AFAE) Book Ser.",
  ["advances in fluid mechanics"] = "Adv. Fluid Mech.",
  ["advances in fluorine research and dental caries prevention"] = "Adv Fluorine Res",
  ["advances in food and nutrition research"] = "Adv Food Nutr Res",
  ["advances in food research"] = "Adv Food Res",
  ["advances in food research. supplement"] = "Adv Food Res Suppl",
  ["advances in forensic haemogenetics"] = "Adv. Forensic Haemogenet.",
  ["advances in free radical biology and medicine"] = "Adv. Free Radical Biol. Med.",
  ["advances in fuzzy systems"] = "Adv. Fuzzy Syst.",
  ["advances in fuzzy systems—applications and theory"] = "Adv. Fuzzy Systems Appl. Theory",
  ["advances in genetics"] = "Adv Genet",
  ["advances in genetics research"] = "Adv Genet Res",
  ["advances in genomics and genetics"] = "Adv Genomics Genet",
  ["advances in geoecology"] = "Adv. Geoecol.",
  ["advances in geometry"] = "Adv. Geom.",
  ["advances in geophysical and environmental mechanics and mathematics"] = "Adv. Geophys. Environ. Mech. Math.",
  ["advances in geosciences"] = "Adv. Geosci.",
  ["advances in geriatric medicine and research"] = "Adv Geriatr Med Res",
  ["advances in geriatrics"] = "Adv Geriatr",
  ["advances in gerontological research"] = "Adv Gerontol Res",
  ["advances in gerontology = uspekhi gerontologii"] = "Adv Gerontol",
  ["advances in group decision and negotiation"] = "Adv. Group Decis. Negot.",
  ["advances in group processes"] = "Adv Group Process",
  ["advances in group theory and applications"] = "Adv. Group Theory Appl.",
  ["advances in health care management"] = "Adv Health Care Manag",
  ["advances in health economics and health services research"] = "Adv Health Econ Health Serv Res",
  ["advances in health economics and health services research. supplement"] = "Adv Health Econ Health Serv Res Suppl",
  ["advances in health sciences education : theory and practice"] = "Adv Health Sci Educ Theory Pract",
  ["advances in health sciences education theory and practice"] = "Adv. Health Sci. Educ. Theory Pract.",
  ["advances in heat transfer"] = "Adv Heat Transf",
  ["advances in hematology"] = "Adv Hematol",
  ["advances in hepatology"] = "Adv Hepatol",
  ["advances in heterocyclic chemistry"] = "Adv. Heterocycl. Chem.",
  ["advances in high energy physics"] = "Adv. High Energy Phys.",
  ["advances in horticultural science"] = "Adv Hortic Sci",
  ["advances in human genetics"] = "Adv Hum Genet",
  ["advances in human-computer interaction"] = "Adv. Hum. Comput. Interact.",
  ["advances in hydrogen energy"] = "Adv. Hydrogen Energy",
  ["advances in image communication"] = "Adv. Image Comm.",
  ["advances in immunity and cancer therapy"] = "Adv Immun Cancer Ther",
  ["advances in immunology"] = "Adv Immunol",
  ["advances in industrial and manufacturing engineering"] = "Adv. Ind. Manuf. Eng.",
  ["advances in industrial control"] = "Adv. Ind. Control",
  ["advances in infectious diseases"] = "Adv Infect Dis",
  ["advances in information security"] = "Adv. Inf. Secur.",
  ["advances in inorganic biochemistry"] = "Adv Inorg Biochem",
  ["advances in inorganic chemistry"] = "Adv. Inorg. Chem.",
  ["advances in insect physiology"] = "Adv In Insect Phys",
  ["advances in integrative medicine"] = "Adv Integr Med",
  ["advances in intelligent data analysis. international symposium on intelligent data analysis"] = "Adv Intell Data Anal",
  ["advances in intelligent systems and computing"] = "Adv. Intell. Syst. Comput.",
  ["advances in internal medicine"] = "Adv Intern Med",
  ["advances in japanese business and economics"] = "Adv. Jpn. Bus. Econ.",
  ["advances in knowledge acquisition, transfer and management (akatm)"] = "Adv. Knowl. Acquis. Transf. Manag. (AKATM)",
  ["advances in knowledge discovery and data mining : 21st pacific-asia conference, pakdd 2017, jeju, south korea, may 23-26, 2017, proceedings. part i. pacific-asia conference on knowledge discovery and data mining (21st : 2017 : cheju island, korea)"] = "Adv Knowl Discov Data Min (2017)",
  ["advances in legume systematics"] = "Adv Legume Syst",
  ["advances in life course research"] = "Adv Life Course Res",
  ["advances in limnology"] = "Adv Limnol",
  ["advances in lipid research"] = "Adv Lipid Res",
  ["advances in lung cancer (irvine)"] = "Adv Lung Cancer (Irvine)",
  ["advances in magnetic and optical resonance"] = "Adv. Magn. Opt. Reson.",
  ["advances in magnetic resonance"] = "Adv. Magn. Reson.",
  ["advances in manufacturing"] = "Adv Manuf",
  ["advances in marine biology"] = "Adv Mar Biol",
  ["advances in mass spectrometry"] = "Adv Mass Spectrom",
  ["advances in materials and processing technologies"] = "Adv. Mater. Process. Technol.",
  ["advances in materials research"] = "Adv. Mater. Res.",
  ["advances in materials science and applications"] = "Adv Mater Sci Appl",
  ["advances in materials science and engineering"] = "Adv. Mater. Sci. Eng.",
  ["advances in mathematical economics"] = "Adv. Math. Econ.",
  ["advances in mathematical fluid mechanics"] = "Adv. Math. Fluid Mech.",
  ["advances in mathematical physics"] = "Adv. Math. Phys.",
  ["advances in mathematical sciences and applications"] = "Adv. Math. Sci. Appl.",
  ["advances in mathematics"] = "Adv. Math.",
  ["advances in mathematics (china)"] = "Adv. Math. (China)",
  ["advances in mathematics (china). shuxue jinzhan"] = "Adv. Math. (China)",
  ["advances in mathematics education"] = "Adv. Math. Educ.",
  ["advances in mathematics of communications"] = "Adv. Math. Commun.",
  ["advances in mechanical engineering"] = "Adv. Mech. Eng.",
  ["advances in mechanics"] = "Adv. Mech.",
  ["advances in mechanics and mathematics"] = "Adv. Mech. Math.",
  ["advances in mechatronics and mechanical engineering (amme) book series"] = "Adv. Mechatron. Mech. Eng. (AMME) Book Ser.",
  ["advances in medical and dental sciences"] = "Adv. Med. Dent. Sci.",
  ["advances in medical education and practice"] = "Adv Med Educ Pract",
  ["advances in medical ethics"] = "Adv Med Ethics",
  ["advances in medical science"] = "Adv. Med. Sci.",
  ["advances in medical sciences"] = "Adv Med Sci",
  ["advances in medical sociology"] = "Adv Med Sociol",
  ["advances in medicine"] = "Adv Med",
  ["advances in medicine and biology"] = "Adv Med Biol",
  ["advances in mental health and intellectual disabilities"] = "Adv Ment Health Intellect Disabil",
  ["advances in metabolic disorders"] = "Adv Metab Disord",
  ["advances in metaheuristics"] = "Adv. Metaheuristics",
  ["advances in meteorology"] = "Adv. Meteorol.",
  ["advances in methods and practices in psychological science"] = "Adv Methods Pract Psychol Sci",
  ["advances in microbial ecology"] = "Adv Microb Ecol",
  ["advances in microbial physiology"] = "Adv Microb Physiol",
  ["advances in microbiology"] = "Adv Microbiol",
  ["advances in military technology"] = "Adv. Mil. Technol.",
  ["advances in mind-body medicine"] = "Adv Mind Body Med",
  ["advances in modelling and analysis a"] = "Adv. Modell. Anal. A",
  ["advances in modelling and analysis b"] = "Adv. Modell. Anal. B",
  ["advances in modern oncology research"] = "Adv Mod Oncol Res",
  ["advances in molecular imaging"] = "Adv J Mol Imaging",
  ["advances in molecular pathology"] = "Adv. Mol. Pathol.",
  ["advances in molecular relaxation and interaction processes"] = "Adv. Mol. Relax. Interact. Processes",
  ["advances in molten salt chemistry"] = "Adv. Molten Salt Chem.",
  ["advances in morphogenesis"] = "Adv Morphog",
  ["advances in motivation and achievement : a research annual"] = "Adv Motiv Achiev",
  ["advances in motivation science"] = "Adv Motiv Sci",
  ["advances in multimedia"] = "Adv. Multimedia",
  ["advances in multimedia information processing - pcm 2013 : 14th pacific-rim conference on multimedia, nanjing, china, december 13-16, 2013 : proceedings. ieee pacific rim conference on multimedia (14th : 2013 : nanjing, china)"] = "Adv Multimed Inf Process - PCM 2013 (2013)",
  ["advances in myocardiology"] = "Adv Myocardiol",
  ["advances in myochemistry"] = "Adv Myochem",
  ["advances in nano research"] = "Adv Nano Res",
  ["advances in nanomedicine and nanotechnology research"] = "Adv Nanomed Nanotechnol Res",
  ["advances in nanoparticles"] = "Adv Nanopart",
  ["advances in natural and applied science"] = "Adv. Nat. Appl. Sci.",
  ["advances in natural sciences: nanoscience and nanotechnology"] = "Adv. Nat. Sci.: Nanosci. Nanotechnol.",
  ["advances in neonatal care"] = "Adv. Neonatal Care",
  ["advances in neonatal care : official journal of the national association of neonatal nurses"] = "Adv Neonatal Care",
  ["advances in nephrology from the necker hospital"] = "Adv Nephrol Necker Hosp",
  ["advances in neural information processing systems"] = "Adv Neural Inf Process Syst",
  ["advances in neurobiology"] = "Adv Neurobiol",
  ["advances in neurodevelopmental disorders"] = "Adv Neurodev Disord",
  ["advances in neuroimmune biology"] = "Adv Neuroimmune Biol",
  ["advances in neuroimmunology"] = "Adv Neuroimmunol",
  ["advances in neurology"] = "Adv Neurol",
  ["advances in neuroscience (hindawi)"] = "Adv Neurosci (Hindawi)",
  ["advances in neurotoxicology"] = "Adv Neurotoxicol",
  ["advances in nonlinear analysis"] = "Adv. Nonlinear Anal.",
  ["advances in nonlinear variational inequalities"] = "Adv. Nonlinear Var. Inequal.",
  ["advances in nonlinear variational inequalities. an international journal"] = "Adv. Nonlinear Var. Inequal.",
  ["advances in nuclear physics"] = "Adv. Nucl. Phys.",
  ["advances in numerical analysis"] = "Adv. Numer. Anal.",
  ["advances in numerical mathematics"] = "Adv. Numer. Math.",
  ["advances in nursing"] = "Adv Nurs",
  ["advances in nursing doctoral education & research"] = "Adv Nurs Dr Educ Res",
  ["advances in nutrition"] = "Adv. Nutr.",
  ["advances in nutrition (bethesda, md.)"] = "Adv Nutr",
  ["advances in nutritional research"] = "Adv Nutr Res",
  ["advances in obesity, weight management & control"] = "Adv Obes Weight Manag Control",
  ["advances in oceanography and limnology"] = "Adv. Oceanogr. Limnol.",
  ["advances in operations research"] = "Adv. Oper. Res.",
  ["advances in operative orthopaedics"] = "Adv Oper Orthop",
  ["advances in operator theory"] = "Adv. Oper. Theory",
  ["advances in ophthalmic plastic and reconstructive surgery"] = "Adv Ophthalmic Plast Reconstr Surg",
  ["advances in ophthalmic, plastic, and reconstructive surgery"] = "Adv. Ophthalmic. Plast. Reconstr. Surg.",
  ["advances in ophthalmology"] = "Adv. Ophthalmol.",
  ["advances in ophthalmology & visual system"] = "Adv Ophthalmol Vis Syst",
  ["advances in ophthalmology = fortschritte der augenheilkunde = progrès en ophtalmologie"] = "Adv Ophthalmol",
  ["advances in ophthalmology and optometry"] = "Adv Ophthalmol Optom",
  ["advances in optical technologies"] = "Adv. Opt. Technol.",
  ["advances in optics and photonics"] = "Adv. Opt. Photonics",
  ["advances in optoelectronics"] = "Adv. OptoElectron.",
  ["advances in oral biology"] = "Adv Oral Biol",
  ["advances in organ biology"] = "Adv Organ Biol",
  ["advances in organic synthesis"] = "Adv. Org. Synth.",
  ["advances in organometallic chemistry"] = "Adv. Organomet. Chem.",
  ["advances in orthopedics"] = "Adv Orthop",
  ["advances in oto-rhino-laryngology"] = "Adv Otorhinolaryngol",
  ["advances in otolaryngology"] = "Adv Otolaryngol",
  ["advances in pain research and therapy"] = "Adv Pain Res Ther",
  ["advances in parallel computing"] = "Adv. Parallel Comput.",
  ["advances in parasitology"] = "Adv Parasitol",
  ["advances in partial differential equations"] = "Adv. Partial Differential Equations",
  ["advances in partial differential equations (basel)"] = "Adv. Partial Differ. Equ. (Basel)",
  ["advances in pathobiology"] = "Adv Pathobiol",
  ["advances in pediatric infectious diseases"] = "Adv Pediatr Infect Dis",
  ["advances in pediatric research"] = "Adv Pediatr Res",
  ["advances in pediatrics"] = "Adv Pediatr",
  ["advances in peritoneal dialysis"] = "Adv. Perit. Dial.",
  ["advances in peritoneal dialysis. conference on peritoneal dialysis"] = "Adv Perit Dial",
  ["advances in pest control research"] = "Adv Pest Control Res",
  ["advances in pharmaceutical sciences"] = "Adv Pharm Sci",
  ["advances in pharmacoepidemiology & drug safety"] = "Adv Pharmacoepidemiol Drug Saf",
  ["advances in pharmacological and pharmaceutial sciences"] = "Adv. Pharmacol. Pharm. Sci.",
  ["advances in pharmacological and pharmaceutical sciences"] = "Adv Pharmacol Pharm Sci",
  ["advances in pharmacological sciences"] = "Adv Pharmacol Sci",
  ["advances in pharmacology"] = "Adv Pharmacol",
  ["advances in pharmacology (san diego, calif.)"] = "Adv Pharmacol",
  ["advances in pharmacology and chemotherapy"] = "Adv Pharmacol Chemother",
  ["advances in pharmacology and clinical trials"] = "Adv Pharmacol Clin Trials",
  ["advances in photochemistry"] = "Adv. Photochem.",
  ["advances in physical organic chemistry"] = "Adv. Phys. Org. Chem.",
  ["advances in physics"] = "Adv. Phys.",
  ["advances in physics: x"] = "Adv Phys X",
  ["advances in physiology education"] = "Adv Physiol Educ",
  ["advances in physiotherapy"] = "Adv Physiother",
  ["advances in pineal research"] = "Adv Pineal Res",
  ["advances in planar lipid bilayers and liposomes"] = "Adv Planar Lipid Bilayers Liposomes",
  ["advances in planned parenthood"] = "Adv Plan Parent",
  ["advances in plant pathology"] = "Adv. Plant Pathol.",
  ["advances in polar science"] = "Adv Polar Sci",
  ["advances in polymer science"] = "Adv. Polym. Sci.",
  ["advances in polymer science = fortschritte der hochpolymeren-forschung"] = "Adv Polym Sci",
  ["advances in polymer technology"] = "Adv. Polym. Technol.",
  ["advances in population : psychosocial perspectives"] = "Adv Popul",
  ["advances in powder metallurgy and particulate materials"] = "Adv. Powder. Metall. Part. Mater.",
  ["advances in power electronics"] = "Adv. Power Electron.",
  ["advances in preventive medicine"] = "Adv Prev Med",
  ["advances in probability theory"] = "Adv. Probab. Theory",
  ["advances in production engineering & management"] = "Adv. Prod. Eng. Manage.",
  ["advances in prostaglandin and thromboxane research"] = "Adv Prostaglandin Thromboxane Res",
  ["advances in prostaglandin, thromboxane, and leukotriene research"] = "Adv Prostaglandin Thromboxane Leukot Res",
  ["advances in protein chemistry"] = "Adv Protein Chem",
  ["advances in protein chemistry and structural biology"] = "Adv Protein Chem Struct Biol",
  ["advances in protein phosphatases"] = "Adv Protein Phosphatases",
  ["advances in psychiatric treatment : the royal college of psychiatrists' journal of continuing professional development"] = "Adv Psychiatr Treat",
  ["advances in psychobiology"] = "Adv Psychobiol",
  ["advances in psychology"] = "Adv. Psych.",
  ["advances in psychology research"] = "Adv Psychol Res",
  ["advances in psychology study"] = "Adv Psychol Study",
  ["advances in psychosomatic medicine"] = "Adv Psychosom Med",
  ["advances in public health"] = "Adv Public Health",
  ["advances in pulmonary hypertension"] = "Adv Pulm Hypertens",
  ["advances in pure and applied mathematics"] = "Adv. Pure Appl. Math.",
  ["advances in quantitative analysis of finance and accounting"] = "Adv. Quant. Anal. Finance Account.",
  ["advances in quantum chemistry"] = "Adv. Quantum Chem.",
  ["advances in radiation biology"] = "Adv Radiat Biol",
  ["advances in radiation oncology"] = "Adv Radiat Oncol",
  ["advances in radio science"] = "Adv. Radio Sci.",
  ["advances in radiology"] = "Adv Radiol",
  ["advances in rare diseases"] = "Adv Rare Dis",
  ["advances in redox research"] = "Adv. Redox Res.",
  ["advances in regenerative biology"] = "Adv. Regener. Biol.",
  ["advances in renal replacement therapy"] = "Adv Ren Replace Ther",
  ["advances in reproductive physiology"] = "Adv Reprod Physiol",
  ["advances in research"] = "Adv Res",
  ["advances in respiratory medicine"] = "Adv Respir Med",
  ["advances in rheumatology (london, england)"] = "Adv Rheumatol",
  ["advances in sample preparation"] = "Adv. Sample Prep.",
  ["advances in school mental health promotion"] = "Adv Sch Ment Health Promot",
  ["advances in science and research"] = "Adv. Sci. Res.",
  ["advances in science and technology (owerri, nigeria)"] = "Adv Sci Technol",
  ["advances in science and technology in the ussr: chemistry series"] = "Adv. Sci. Tech. USSR Chem. Ser.",
  ["advances in second messenger and phosphoprotein research"] = "Adv Second Messenger Phosphoprotein Res",
  ["advances in sex hormone research"] = "Adv Sex Horm Res",
  ["advances in shock research"] = "Adv Shock Res",
  ["advances in simulation"] = "Adv. Simul.",
  ["advances in simulation (london, england)"] = "Adv Simul (Lond)",
  ["advances in skin & wound care"] = "Adv Skin Wound Care",
  ["advances in small animal care"] = "Adv. Small Anim. Care",
  ["advances in small animal medicine and surgery"] = "Adv Small Anim Med Surg",
  ["advances in social science and culture"] = "Adv Soc Sci Cult",
  ["advances in social sciences research journal"] = "Adv Soc Sci Res J",
  ["advances in social work"] = "Adv Soc Work",
  ["advances in socio-dental research"] = "Adv Sociodent Res",
  ["advances in space biology and medicine"] = "Adv Space Biol Med",
  ["advances in space research"] = "Adv. Space Res.",
  ["advances in space research : the official journal of the committee on space research (cospar)"] = "Adv Space Res",
  ["advances in spatial science"] = "Adv. Spat. Sci.",
  ["advances in spectroscopy"] = "Adv. Spectrosc.",
  ["advances in statistical analysis : asta : a journal of the german statistical society"] = "Adv Stat Anal",
  ["advances in statistical climatology, meteorology and oceanography"] = "Adv. Stat. Climatol. Meteorol. Oceanogr.",
  ["advances in statistics"] = "Adv Stat",
  ["advances in stem cells"] = "Adv Stem Cells",
  ["advances in steroid biochemistry and pharmacology"] = "Adv Steroid Biochem Pharmacol",
  ["advances in structural engineering"] = "Adv. Struct. Eng.",
  ["advances in surgery"] = "Adv Surg",
  ["advances in surgery (st. louis, mo"] = "Adv Surg",
  ["advances in synchrotron radiation"] = "Adv. Synchrotron Radiat.",
  ["advances in systems science and applications"] = "Adv Syst Sci Appl",
  ["advances in the biosciences"] = "Adv Biosci",
  ["advances in the mathematical sciences"] = "Adv. Math. Sci.",
  ["advances in the study of behavior"] = "Adv Study Behav",
  ["advances in theoretical and mathematical physics"] = "Adv. Theor. Math. Phys.",
  ["advances in therapy"] = "Adv Ther",
  ["advances in tracer methodology"] = "Adv Tracer Methodol",
  ["advances in transportation studies"] = "Adv. Transp. Stud.",
  ["advances in tribology"] = "Adv. Tribol.",
  ["advances in tuberculosis research"] = "Adv. Tuberc. Res.",
  ["advances in tuberculosis research. fortschritte der tuberkuloseforschung. progres de l’exploration de la tuberculose"] = "Adv Tuberc Res",
  ["advances in tuberculosis research. fortschritte der tuberkuloseforschung. progrès de l'exploration de la tuberculose"] = "Adv Tuberc Res",
  ["advances in tumor virology"] = "Adv Tumor Virol",
  ["advances in urology"] = "Adv Urol",
  ["advances in vascular medicine"] = "Adv Vasc Med",
  ["advances in veterinary medicine"] = "Adv Vet Med",
  ["advances in veterinary science"] = "Adv Vet Sci",
  ["advances in veterinary science and comparative medicine"] = "Adv Vet Sci Comp Med",
  ["advances in virology"] = "Adv Virol",
  ["advances in virus research"] = "Adv Virus Res",
  ["advances in water research"] = "Adv. Water. Res.",
  ["advances in water resources"] = "Adv. Water Res.",
  ["advances in wound care"] = "Adv Wound Care (New Rochelle)",
  ["advances in wound care : the journal for prevention and healing"] = "Adv Wound Care",
  ["advances in x-ray analysis"] = "Adv. X-Ray Anal.",
  ["advances materials technologies"] = "Adv. Mat. Technol.",
  ["advancing clinical care"] = "Adv. Clin. Care",
  ["advancing clinical care : official journal of noaadn"] = "Adv Clin Care",
  ["advancing the consumer interest"] = "Advanc. Cons. Interest",
  ["adventist heritage"] = "Advent Herit",
  ["adverse drug reaction bulletin"] = "Adverse Drug React Bull",
  ["adverse drug reactions and acute poisoning reviews"] = "Adverse Drug React Acute Poisoning Rev",
  ["adverse drug reactions and toxicological reviews"] = "Adverse Drug React Toxicol Rev",
  ["adversity and resilience science"] = "Advers Resil Sci",
  ["advertising age"] = "Advert Age",
  ["advocacy now"] = "Advocacy Now",
  ["advocate (boston, mass.)"] = "Advocate",
  ["aea papers and proceedings. american economic association"] = "AEA Pap Proc",
  ["aecu [reports]. u.s. atomic energy commission"] = "AECU Rep",
  ["aegaeum. annales d’archéologie égéenne de l’université de liège"] = "Aegaeum",
  ["aegean archaeology"] = "AeA",
  ["aem education and training"] = "AEM Educ Train",
  ["aeolian research"] = "Aeolian Res.",
  ["aequationes mathematicae"] = "Aequationes Math.",
  ["aer journal : research and practice in visual impairment and blindness"] = "AER J",
  ["aera open"] = "AERA Open",
  ["aere perennius"] = "Aere Perennius",
  ["aeromedica acta"] = "Aeromed Acta",
  ["aeromedical review"] = "Aeromed. Rev.",
  ["aeromedical reviews"] = "Aeromed Rev",
  ["aeronautical journal"] = "Aeronaut. J.",
  ["aeronautical quarterly"] = "Aeronaut. Q.",
  ["aerosol and air quality research"] = "Aerosol Air Qual Res",
  ["aerosol research"] = "Aerosol Res.",
  ["aerosol science and engineering"] = "Aerosol Sci. Eng.",
  ["aerosol science and technology"] = "Aerosol Sci. Technol.",
  ["aerosol science and technology : the journal of the american association for aerosol research"] = "Aerosol Sci Technol",
  ["aerospace (basel, switzerland)"] = "Aerospace (Basel)",
  ["aerospace america"] = "Aerosp Am",
  ["aerospace engineering (society of automotive engineers)"] = "Aerosp Eng",
  ["aerospace historian"] = "Aerosp Hist",
  ["aerospace medicine"] = "Aerosp Med",
  ["aerospace medicine and human performance"] = "Aerosp Med Hum Perform",
  ["aerospace science and technology"] = "Aerosp. Sci. Technol.",
  ["aerospace series"] = "Aerosp. Ser.",
  ["aerospace systems"] = "Aerosp. Syst.",
  ["aerotecnica missili & spazio"] = "Aerotec. Missili Spazio",
  ["aerzteblatt rheinland-pfalz"] = "Arztebl Rheinl Pfalz",
  ["aerztliche praxis"] = "Arztl Prax",
  ["aerztlicher feldpostbrief"] = "Aerztlicher Feldpostbrief",
  ["aesculapian (brooklyn, new york, n.y.)"] = "Aesculapian",
  ["aesthetic plastic surgery"] = "Aesthetic Plast Surg",
  ["aesthetic surgery journal"] = "Aesthet Surg J",
  ["aesthetic surgery journal. open forum"] = "Aesthet Surg J Open Forum",
  ["aeu international journal of electronics and communications"] = "AEU Int. J. Electron. Commun.",
  ["aevum antiquum"] = "Aevum(ant)",
  ["aevum. rassegna di scienze storiche linguistiche e filologiche"] = "Aevum",
  ["aevum: rassegna di scienze storiche, linguistiche e filologiche"] = "Aevum",
  ["afdeling informatica"] = "Afd. Inform.",
  ["afdeling numerieke wiskunde"] = "Afd. Numer. Wisk.",
  ["affari sociali internazionali"] = "Aff Soc Int",
  ["affective science"] = "Affect Sci",
  ["afghan medical journal"] = "Afghan Med J",
  ["afips conference proceedings"] = "AFIPS Conf. Proc.",
  ["afl-cio american federationist"] = "AFL. CIO Am. Fed.",
  ["afmdc tr. air force missile development center (u.s.)"] = "AFMDC TR",
  ["afosr tn. united states. air force. office of scientific research"] = "AFOSR TN United States Air Force Off Sci Res",
  ["afosr tr. united states. air force. office of scientific research"] = "AFOSR TR United States Air Force Off Sci Res",
  ["africa (roma), quaderni"] = "Africa",
  ["africa : journal of the international institute of african languages and cultures"] = "Africa (Lond)",
  ["africa : notiziario dell'associazione fra le imprese italiane in africa"] = "Africa",
  ["africa : notiziario dell’associazione fra le imprese italiane in africa"] = "Africa (Roma)",
  ["africa analysis"] = "Afr Anal",
  ["africa development. afrique et developpement"] = "Afr Dev",
  ["africa development. afrique et développement"] = "Afr Dev",
  ["africa health"] = "Afr Health",
  ["africa insight"] = "Afr Insight",
  ["africa journal of nursing and midwifery"] = "Afr J Nurs Midwifery",
  ["africa link : a publication of the africa region"] = "Afr Link",
  ["africa mathematics annals. afma. a pan african journal of mathematics"] = "Afr. Math. Ann. AFMA",
  ["africa media review"] = "Afr Media Rev",
  ["africa quarterly"] = "Afr Q",
  ["africa recovery"] = "Afr Recovery",
  ["africa report"] = "Afr Rep",
  ["africa south of the sahara"] = "Afr South",
  ["africa today"] = "Afr Today",
  ["africa woman"] = "Afr Woman",
  ["africa women & health : a safe motherhood magazine"] = "Afr Women Health",
  ["africa's population and development bulletin"] = "Afr Popul Dev Bull",
  ["africa. institut national d’archéologie et d’art, tunis"] = "Africa",
  ["african affairs"] = "Afr Aff (Lond)",
  ["african alternatives"] = "Afr Alternat",
  ["african alternatives / africa operations research and technical assistance project"] = "Afr Alternat",
  ["african anthropology : journal of the pan african anthropological association = anthropologie africaine : revue de l'association panafricaine de l'anthropologie"] = "Afr Anthropol",
  ["african archaeological review"] = "Afr. Archaeol. Rev.",
  ["african brain journal"] = "Afr. Brain J.",
  ["african concord"] = "Afr Concord",
  ["african crop science journal"] = "Afr Crop Sci J",
  ["african demography"] = "Afr Demogr",
  ["african dental journal"] = "Afr. Dent. J.",
  ["african dental journal : official publication of the federation of african dental associations = journal dentaire africain"] = "Afr Dent J",
  ["african development review"] = "African Devel. Rev.",
  ["african development review = revue africaine de développement"] = "Afr Dev Rev",
  ["african diaspora journal of mathematics"] = "Afr. Diaspora J. Math.",
  ["african diaspora journal of mathematics. new series"] = "Afr. Diaspora J. Math. New Ser.",
  ["african economic history"] = "Afr Econ Hist",
  ["african entomology"] = "Afr. Entomol.",
  ["african entomology : journal of the entomological society of southern africa"] = "Afr Entomol",
  ["african environment"] = "Afr Environ",
  ["african farmer : the key to africa's future"] = "Afr Farmer",
  ["african farmer : the key to africa’s future"] = "Afr Farmer",
  ["african finance journal"] = "African Finance J.",
  ["african geographical review"] = "Afr. Geogr. Rev.",
  ["african health sciences"] = "Afr Health Sci",
  ["african historical studies"] = "Afr Hist Stud",
  ["african institute of mathematics (aims) library series"] = "AIMS Libr. Ser.",
  ["african invertebrates"] = "Afr Invertebr",
  ["african journal for physical health education, recreation, and dance"] = "Afr J Phys Health Educ Recreat Dance",
  ["african journal for the psychological study of social issues : journal of the african society for the psychological study of social issues"] = "Afr J Psychol Study Soc Issues",
  ["african journal of agricultural research"] = "Afr J Agric Res",
  ["african journal of aids research : ajar"] = "Afr J AIDS Res",
  ["african journal of applied research"] = "Afr. J. Appl. Res.",
  ["african journal of applied statistics"] = "Afr. J. Appl. Stat.",
  ["african journal of aquatic science"] = "Afr. J. Aquat. Sci.",
  ["african journal of biochemistry research"] = "Afr. J. Biochem. Res.",
  ["african journal of biological sciences"] = "Afr. J. Biol. Sci.",
  ["african journal of biomedical research : ajbr"] = "Afr J Biomed Res",
  ["african journal of biotechnology"] = "Afr J Biotechnol",
  ["african journal of cellular pathology"] = "Afr J Cell Pathol",
  ["african journal of chemical education"] = "Afr. J. Chem. Educ.",
  ["african journal of clinical and experimental microbiology"] = "Afr. J. Clin. Exp. Microbiol.",
  ["african journal of computing and ict"] = "Afr J Comput Ict",
  ["african journal of disability"] = "Afr J Disabil",
  ["african journal of drug and alcohol studies"] = "Afr J Drug Alcohol Stud",
  ["african journal of ecology"] = "Afr J Ecol",
  ["african journal of economic and sustainable development"] = "Afr. J. Econ. Sustainable Dev.",
  ["african journal of emergency medicine : revue africaine de la médecine d'urgence"] = "Afr J Emerg Med",
  ["african journal of engineering & technology"] = "Afr. J. Eng. Technol.",
  ["african journal of environmental health sciences"] = "Afr J Environ Health Sci",
  ["african journal of environmental science and technology"] = "Afr. J. Environ. Sci. Technol.",
  ["african journal of fertility, sexuality, and reproductive health"] = "Afr J Fertil Sexual Reprod Heal",
  ["african journal of food science"] = "Afr. J. Food Sci.",
  ["african journal of food, agriculture, nutrition and development"] = "Afr. J. Food Agric. Nutr. Dev.",
  ["african journal of food, agriculture, nutrition, and development : ajfand"] = "Afr J Food Agric Nutr Dev",
  ["african journal of health professions education"] = "Afr J Health Prof Educ",
  ["african journal of health sciences"] = "Afr J Health Sci",
  ["african journal of herpetology"] = "Afr. J. Herpetol.",
  ["african journal of herpetology : journal of the herpetological association of africa"] = "Afr J Herpetol",
  ["african journal of infectious diseases"] = "Afr J Infect Dis",
  ["african journal of laboratory medicine"] = "Afr J Lab Med",
  ["african journal of marine science"] = "Afr. J. Mar. Sci.",
  ["african journal of medical and health sciences"] = "Afr J Med Health Sci",
  ["african journal of medical practice"] = "Afr J Med Pract",
  ["african journal of medical practice / african academy of sciences"] = "Afr J Med Pract",
  ["african journal of medical sciences"] = "Afr. J. Med. Sci.",
  ["african journal of medicine and medical sciences"] = "Afr J Med Med Sci",
  ["african journal of microbiology research"] = "Afr J Microbiol Res",
  ["african journal of midwifery and women's health"] = "Afr J Midwifery Womens Health",
  ["african journal of nephrology : official publication of the african association of nephrology. african association of nephrology"] = "Afr J Nephrol",
  ["african journal of paediatric surgery : ajps"] = "Afr J Paediatr Surg",
  ["african journal of pharmacy and pharmacology"] = "Afr J Pharm Pharmacol",
  ["african journal of plant science"] = "Afr. J. Plant Sci.",
  ["african journal of primary health care & family medicine"] = "Afr J Prim Health Care Fam Med",
  ["african journal of psychiatry"] = "Afr J Psychiatry (Johannesbg)",
  ["african journal of range & forage science"] = "Afr J Range Forage Sci",
  ["african journal of range and forage science"] = "Afr. J. Range Forage Sci.",
  ["african journal of reproductive health"] = "Afr J Reprod Health",
  ["african journal of research in mathematics, science and technology education"] = "Afr. J. Res. Math. Sci. Technol. Educ.",
  ["african journal of root and tuber crops"] = "Afr J Root Tuber Crops",
  ["african journal of science, technology, innovation and development"] = "Afr. J. Sci. Innovation Dev.",
  ["african journal of sociology"] = "Afr J Sociol",
  ["african journal of traditional, complementary, and alternative medicines : ajtcam"] = "Afr J Tradit Complement Altern Med",
  ["african journal of traumatic stress"] = "Afr J Trauma Stress",
  ["african journal of urology : the official journal of the pan african urological surgeons' association (pausa)"] = "Afr J Urol",
  ["african journal of wildlife research"] = "Afr J Wildl Res",
  ["african notes : bulletin of the institute of african studies, university of ibadan"] = "Afr Notes",
  ["african plant protection"] = "Afr Plant Prot",
  ["african population newsletter"] = "Afr Popul Newsl",
  ["african primates"] = "Afr Primates",
  ["african research & documentation"] = "Afr Res Doc",
  ["african review (dar es salaam, tanzania)"] = "Afr Rev",
  ["african social research"] = "Afr Soc Res",
  ["african statistical journal = journal statistique africain"] = "Afr Stat J",
  ["african studies"] = "Afr Stud",
  ["african studies review"] = "Afr Stud Rev",
  ["african study monographs"] = "Afr Study Monogr",
  ["african urban notes"] = "Afr Urban Notes",
  ["african urban notes. ser. b"] = "Afr Urban Notes B",
  ["african urban quarterly"] = "Afr Urban Q",
  ["african urban studies (michigan state university. african studies center)"] = "Afr Urban Stud",
  ["african zoology"] = "Afr Zool",
  ["africana bulletin (boston, mass.)"] = "Afr Bull",
  ["africana marburgensia"] = "Afr Marburgen",
  ["africana notes and news. africana aantekeninge en nuus"] = "Afr Notes News",
  ["africana research bulletin"] = "Afr Res Bull",
  ["africasia (paris, france : 1984)"] = "Africasia",
  ["africa’s population and development bulletin"] = "Afr Popul Dev Bull",
  ["africom / african council on communication education"] = "Africom",
  ["afrika matematika"] = "Afr. Mat.",
  ["afrika spectrum"] = "Afrika Spec.",
  ["afrika statistika"] = "Afr. Stat.",
  ["afrika und ubersee"] = "Afr Uebersee",
  ["afrika und übersee"] = "Afr Uebersee",
  ["afrique 2000 (geneva, switzerland)"] = "Afrique 2000",
  ["afrique contemporaine"] = "Afr Contemp",
  ["afrique documents"] = "Afr Doc",
  ["afrique francaise chirurgicale"] = "Afr Francaise Chir",
  ["afrique française chirurgicale"] = "Afr Fr Chir",
  ["afrique medicale"] = "Afr Med",
  ["afrique médicale"] = "Afr Med",
  ["afro technical papers"] = "Afro Tech Pap",
  ["afro-americans in new york life and history"] = "Afro Am N Y Life Hist",
  ["against the grain (charleston, s.c.)"] = "Against Grain (Charlest)",
  ["agal research"] = "Agal Res.",
  ["agarica : mykologisk tidsskrift"] = "Agarica",
  ["age (dordrecht, netherlands)"] = "Age (Dordr)",
  ["age (melbourne, vic.)"] = "Age",
  ["age and ageing"] = "Age Ageing",
  ["aged care & services review"] = "Aged Care Serv Rev",
  ["aged care and services review"] = "Aged Care Serv. Rev.",
  ["ageing and society"] = "Ageing Soc",
  ["ageing international"] = "Ageing Int",
  ["ageing research"] = "Ageing Res",
  ["ageing research reviews"] = "Ageing Res Rev",
  ["agenda (durban, south africa)"] = "Agenda",
  ["agenda salud"] = "Agenda Salud",
  ["agent-directed simulation symposium (ads). agent-directed simulation symposium"] = "Agent Dir Simul Symp",
  ["agents and actions"] = "Agents Actions",
  ["agents and actions. supplements"] = "Agents Actions Suppl",
  ["aggiornamenti clinicoterapeutici"] = "Aggiorn Clinico Ter",
  ["aggiornamenti di fisiologia"] = "Aggiorn Fisiol",
  ["aggiornamenti di terapia oftalmologica"] = "Aggiorn Ter Oftalmol",
  ["aggiornamenti in ostetricia e ginecologia"] = "Aggiorn Ostet Ginecol",
  ["aggiornamenti sulle malattie da infezione"] = "Aggiorn Sulle Mal Infez",
  ["aggiornamento pediatrico"] = "Aggiorn Pediatr",
  ["aggression and violent behavior"] = "Aggress Violent Behav",
  ["aggressive behavior"] = "Aggress Behav",
  ["agid report series : geosciences in international development"] = "Agid. Geo. In.",
  ["aging & leisure living"] = "Aging Leis Living",
  ["aging & mental health"] = "Aging Ment Health",
  ["aging (milan, italy)"] = "Aging (Milano)",
  ["aging and cancer"] = "Aging Cancer",
  ["aging and disease"] = "Aging Dis",
  ["aging and health research"] = "Aging Health Res",
  ["aging and leisure living"] = "Aging Leis. Living",
  ["aging and mental health"] = "Aging Ment. Health",
  ["aging and neurodegeneration"] = "Aging Neurodegener",
  ["aging and society : an interdisciplinary journal"] = "Aging Soc",
  ["aging and work"] = "Aging Work",
  ["aging brain"] = "Aging Brain",
  ["aging cell"] = "Aging Cell",
  ["aging clinical and experimental research"] = "Aging Clin Exp Res",
  ["aging health"] = "Aging health",
  ["aging male"] = "Aging Male",
  ["aging medicine (milton (n.s.w))"] = "Aging Med (Milton)",
  ["aging pathobiology and therapeutics"] = "Aging Pathobiol Ther",
  ["aging today : the bimonthly newspaper of the american society on aging"] = "Aging Today",
  ["aging trends"] = "Aging Trends",
  ["aging trends (hyattsville, md.)"] = "Aging Trends",
  ["agnes karll-schwester. der krankenpfleger"] = "Agnes Karll Schwest. Krankenpfleger",
  ["agora de heterodoxias"] = "Agora Heterodoxias",
  ["agora mathematica"] = "Agora Math.",
  ["agrarforschung schweiz"] = "Agrarforsch Schweiz",
  ["agrártörténeti szemle = historia rerum rusticarum"] = "Agratort Szle",
  ["agressologie: revue internationale de physio-biologie et de pharmacologie appliquees aux effets de l’agression"] = "Agressologie",
  ["agressologie: revue internationale de physio-biologie et de pharmacologie appliquées aux effets de l'agression"] = "Agressologie",
  ["agri gene"] = "Agri Gene",
  ["agribusiness (new york, n.y.)"] = "Agribusiness (N Y N Y)",
  ["agricoltore ticinese"] = "Agric. ticin.",
  ["agricoltura italiana"] = "Agric. Ital.",
  ["agricultural & environmental letters"] = "Agric. Environ. Lett.",
  ["agricultural and biological chemistry"] = "Agric Biol Chem",
  ["agricultural and food economics"] = "Agric. Food Econ.",
  ["agricultural and food science"] = "Agric. Food Sci.",
  ["agricultural and forest entomology"] = "Agric. For. Entomol.",
  ["agricultural and forest meteorology"] = "Agric. For. Meteorol.",
  ["agricultural and resource economics review"] = "Agric Resour Econ Rev",
  ["agricultural biology"] = "Agric Biol",
  ["agricultural economics"] = "Agric. Econ.",
  ["agricultural economics (amsterdam, netherlands)"] = "Agric Econ",
  ["agricultural economics research"] = "Agric Econ Res",
  ["agricultural engineering"] = "Agric. Eng.",
  ["agricultural finance review"] = "Agr. Finance Rev.",
  ["agricultural history"] = "Agric Hist",
  ["agricultural information bulletin usda forest service"] = "Agric. Inf. Bull. USDA For. Serv.",
  ["agricultural information development bulletin"] = "Agric Inf Dev Bull",
  ["agricultural meteorology"] = "Agric. Meteorol.",
  ["agricultural research"] = "Agric. Res.",
  ["agricultural research & technology : open access journal"] = "Agric Res Technol",
  ["agricultural science journal"] = "Agric Sci J",
  ["agricultural sciences in china"] = "Agric Sci China",
  ["agricultural systems"] = "Agric Syst",
  ["agricultural technology"] = "Agric Technol (Thail)",
  ["agricultural wastes"] = "Agric. Wastes",
  ["agricultural water management"] = "Agric. Water Manage.",
  ["agriculture & food security"] = "Agric Food Secur",
  ["agriculture (basel, switzerland)"] = "Agriculture (Basel)",
  ["agriculture (nitra, slovakia)"] = "Agriculture",
  ["agriculture and agricultural science procedia"] = "Agric. Agric. Sci. Procedia",
  ["agriculture and environment"] = "Agric. Environ.",
  ["agriculture and forestry bulletin"] = "Agric. For. Bull.",
  ["agriculture and human values"] = "Agric Human Values",
  ["agriculture and natural resources"] = "Agric. Nat. Resour.",
  ["agriculture and resources quarterly"] = "Agr. Resources Quart.",
  ["agriculture ecosystems and environment"] = "Agric. Ecosyst. Environ.",
  ["agriculture handbook"] = "Agric. Handb.",
  ["agriculture science developments"] = "Agric Sci Dev",
  ["agriculture, ecosystems & environment"] = "Agric Ecosyst Environ",
  ["agriculture, ecosystems and environment"] = "Agric. Ecosyst. Environ.",
  ["agro food industry hi-tech"] = "Agro Food Ind Hi Tech",
  ["agroecology and sustainable food systems"] = "Agroecol. Sustainable Food Syst.",
  ["agroforestry systems"] = "Agrofor. Syst.",
  ["agrolife scientific journal"] = "AgroLife Sci. J.",
  ["agronomia mesoamericana"] = "Agronomia Mesoamericana",
  ["agronomía & ambiente"] = "Agron Ambient",
  ["agronomía colombiana"] = "Agron Colomb",
  ["agronomía mesoamericana : órgano divulgativo del pccmca, programa cooperativo centroamericano de mejoramiento de cultivos y animales"] = "Agron Mesoam",
  ["agronomy (basel, switzerland)"] = "Agronomy (Basel)",
  ["agronomy for sustainable development"] = "Agron. Sustainable Dev.",
  ["agronomy journal"] = "Agron J",
  ["agrosystems, geosciences & environment"] = "Agrosyst. Geosci. Environ.",
  ["agrégation de mathématiques"] = "Agrég. Math.",
  ["agso journal of australian geology & geophysics"] = "AGSO J Aust Geol Geophys",
  ["agso journal of australian geology and geophysics"] = "AGSO J. Aust. Geol. Geophys.",
  ["agu advances"] = "AGU Adv.",
  ["aguiaine / societe d’etudes folkloriques du centre-ouest"] = "Aguiaine",
  ["aǧrı : ağrı (algoloji) derneği'nin yayın organıdır = the journal of the turkish society of algology"] = "Agri",
  ["aha hospital technology series"] = "AHA Hosp Technol Ser",
  ["ahapit : ukraïs'kyĭ istoryko-medychnyĭ zhurnal = agapit : the ukrainian historical and medical journal"] = "Agapit",
  ["ahip coverage"] = "AHIP Cover",
  ["ahme journal"] = "AHME J",
  ["ahot be-yisrael"] = "Ahot Beyisrael.",
  ["ahp journal"] = "AHP J",
  ["ahp journal / association for healthcare philanthropy"] = "AHP J",
  ["ahrq webm&m : morbidity & mortality rounds on the web"] = "AHRQ WebM&M",
  ["ai & society"] = "AI Soc",
  ["ai and ethics"] = "AI Ethics",
  ["ai and society"] = "AI Soc.",
  ["ai bian, ji bian, tu bian"] = "Ai Bian Ji Bian Tu Bian",
  ["ai communications"] = "AI Commun.",
  ["ai communications. the european journal on artificial intelligence"] = "AI Commun.",
  ["ai edam"] = "AI EDAM",
  ["ai in civil engineering"] = "AI Civ. Eng.",
  ["ai magazine"] = "AI Mag",
  ["ai matters"] = "AI Matters",
  ["ai open"] = "AI Open",
  ["ai switzerland"] = "AI Switz.",
  ["ai zheng = aizheng = chinese journal of cancer"] = "Ai Zheng",
  ["aia journal"] = "AIA J.",
  ["aia journal. american institute of architects"] = "AIA J",
  ["aiaa education series"] = "AIAA Educ. Ser.",
  ["aiaa journal"] = "AIAA J.",
  ["aiaa journal of air transportation"] = "AIAA J. Air Transp.",
  ["aiaa journal. american institute of aeronautics and astronautics"] = "AIAA J",
  ["aiaa student journal. american institute of aeronautics and astronautics"] = "AIAA Stud J",
  ["aicc economic review. indian national congress. all india congress committee"] = "AICC Econ Rev",
  ["aiccm bulletin"] = "AICCM Bull.",
  ["aiche journal"] = "AlChE J.",
  ["aiche journal. american institute of chemical engineers"] = "AIChE J",
  ["aiche symposium series"] = "AIChE Symp. Ser.",
  ["aichi gakuin daigaku shigakkai shi"] = "Aichi Gakuin Daigaku Shigakkai Shi",
  ["aichi gakuin daigaku shigakkai shi (aichi-gakuin journal of dental science)"] = "Aichi Gakuin Daigaku Shigakkai Shi",
  ["aichi-gakuin dental science"] = "Aichi Gakuin Dent Sci",
  ["aidos news : bimonthly bulletin of the italian association for women in development. associazione italiana donne per lo sviluppo"] = "Aidos News",
  ["aids & public policy journal"] = "AIDS Public Policy J",
  ["aids & society"] = "AIDS Soc",
  ["aids (london, england)"] = "AIDS",
  ["aids 2006 : time to deliver : abstract cd-rom. international conference on aids (16th : 2006 : toronto, canada)"] = "AIDS 2006",
  ["aids 2008 : abstract cd-rom, xvii international aids conference, mexico city, mexico, 3-8 august 2008. international conference on aids (17th : 2008 : mexico city)"] = "AIDS 2008",
  ["aids action"] = "AIDS Action",
  ["aids action policy brief"] = "AIDS Action Policy Brief",
  ["aids action update"] = "AIDS Action Update",
  ["aids action update / aids action foundation"] = "AIDS Action Update",
  ["aids alert"] = "AIDS Alert",
  ["aids analysis africa"] = "AIDS Anal Afr",
  ["aids analysis asia"] = "AIDS Anal Asia",
  ["aids and behavior"] = "AIDS Behav",
  ["aids asia : voice of the asian solidarity against aids"] = "AIDS Asia",
  ["aids care"] = "AIDS Care",
  ["aids clinical care"] = "AIDS Clin Care",
  ["aids clinical review"] = "AIDS Clin Rev",
  ["aids education and prevention"] = "AIDS Educ. Prev.",
  ["aids education and prevention : official publication of the international society for aids education"] = "AIDS Educ Prev",
  ["aids health promotion exchange"] = "AIDS Health Promot Exch",
  ["aids health promotion exchange / world health organization, global programme on aids, health promotion unit"] = "AIDS Health Promot Exch",
  ["aids illustrated"] = "AIDS Illus",
  ["aids information exchange"] = "AIDS Inf Exch",
  ["aids information exchange / the united states conference of mayors"] = "AIDS Inf Exch",
  ["aids institute newsletter. new york (state). aids institute"] = "AIDS Inst Newsl",
  ["aids literature & law review"] = "AIDS Lit Law Rev",
  ["aids newsletter (london, england)"] = "AIDS Newsl",
  ["aids patient care"] = "AIDS Patient Care",
  ["aids patient care and stds"] = "AIDS Patient Care STDS",
  ["aids policy & law"] = "AIDS Policy Law",
  ["aids research"] = "AIDS Res",
  ["aids research and human retroviruses"] = "AIDS Res Hum Retroviruses",
  ["aids research and therapy"] = "AIDS Res Ther",
  ["aids research and treatment"] = "AIDS Res Treat",
  ["aids reviews"] = "AIDS Rev",
  ["aids treatment news"] = "AIDS Treat News",
  ["aids updates"] = "AIDS Updates",
  ["aids vaccine 2001 : september 5-8, 2001, philadelphia marriott, philadelphia, pa, usa"] = "AIDS Vaccine 2001",
  ["aids watch"] = "AIDS Watch",
  ["aids weekly"] = "AIDS Wkly",
  ["aids weekly plus"] = "AIDS Wkly Plus",
  ["aids-forschung : aifo = acquired immune deficiency syndrome research"] = "AIDS Forsch",
  ["aids/std health promotion exchange"] = "AIDS STD Health Promot Exch",
  ["aidsed newsletter : newsletter of the who/unesco aids education and health promotion materials exchange centre for asia and the pacific"] = "AIDSED Newsl",
  ["aidslink : eastern, central & southern africa"] = "AIDSlink",
  ["aiha journal"] = "AIHA J.",
  ["aiha journal : a journal for the science of occupational and environmental health and safety"] = "AIHA J (Fairfax, Va)",
  ["aihaj : a journal for the science of occupational and environmental health and safety"] = "AIHAJ",
  ["aims agriculture and food"] = "AIMS Agric. Food",
  ["aims allergy and immunology"] = "AIMS Allergy Immunol",
  ["aims bioengineering"] = "AIMS Bioeng",
  ["aims biophysics"] = "AIMS Biophys",
  ["aims electronics and electrical engineering"] = "AIMS Electron. Electr. Eng.",
  ["aims energy"] = "AIMS Energy",
  ["aims environmental science"] = "AIMS Environ Sci",
  ["aims genetics"] = "AIMS Genet",
  ["aims geosciences"] = "AIMS Geosci",
  ["aims material science"] = "Aims Mater. Sci.",
  ["aims materials science"] = "AIMS Mater Sci",
  ["aims mathematics"] = "AIMS Math.",
  ["aims medical science"] = "AIMS Med Sci",
  ["aims microbiology"] = "AIMS Microbiol",
  ["aims molecular science"] = "AIMS Mol Sci",
  ["aims neuroscience"] = "AIMS Neurosci",
  ["aims public health"] = "AIMS Public Health",
  ["aims report"] = "AIMS Rep",
  ["aims series on applied mathematics"] = "AIMS Ser. Appl. Math.",
  ["aims series on differential equations & dynamical systems"] = "AIMS Ser. Differ. Equ. Dyn. Syst.",
  ["aims series on random and computational dynamics"] = "AIMS Ser. Random Comput. Dyn.",
  ["ain shams engineering journal"] = "Ain Shams Eng. J.",
  ["ain shams medical journal"] = "Ain Shams Med J",
  ["aip advances"] = "AIP Adv.",
  ["aip conference proceedings"] = "AIP Conf. Proc.",
  ["aip series in computational and applied mathematical physics"] = "AIP Ser. Comput. Appl. Math. Phys.",
  ["aip series in modern acoustics and signal processing"] = "AIP Ser. Modern Acoust. Signal Process.",
  ["aipla quarterly journal"] = "AIPLA Q J",
  ["aipla quarterly journal / american intellectual property law association"] = "AIPLA Q J",
  ["air & space smithsonian"] = "Air Space",
  ["air & waste : journal of the air & waste management association"] = "Air Waste",
  ["air and waste"] = "Air Waste",
  ["air and water pollution"] = "Air Water Pollut.",
  ["air line pilot"] = "Air Line Pilot",
  ["air medical journal"] = "Air Med J",
  ["air medical news-letter. great britain. admiralty. medical department"] = "Air Med News Lett",
  ["air quality atmosphere and health"] = "Air Qual. Atmos. Health",
  ["air quality, atmosphere and health"] = "Air Qual. Atmos. Health",
  ["air quality, atmosphere, & health"] = "Air Qual Atmos Health",
  ["air sea rescue bulletin"] = "Air Sea Rescue Bull",
  ["air university quarterly review"] = "Air Univ Rev",
  ["air university review (united states edition)"] = "Air Univ Rev (US ed)",
  ["air, soil and water research"] = "Air Soil Water Res.",
  ["aircraft design"] = "Aircr. Des.",
  ["aircraft engineering"] = "Aircr. Eng.",
  ["aircraft engineering and aerospace technology"] = "Aircr. Eng. Aerosp. Technol.",
  ["airo springer series"] = "AIRO Springer Ser.",
  ["aitna. quaderni di topografia antica"] = "Aitna",
  ["ajalooline ajakiri (tartu, estonia : 1998)"] = "Ajalooline Ajak",
  ["ajnr. american journal of neuroradiology"] = "AJNR Am J Neuroradiol",
  ["ajnr: american journal of neuroradiology"] = "AJNR Am. J. Neuroradiol.",
  ["ajob empirical bioethics"] = "AJOB Empir Bioeth",
  ["ajob neuroscience"] = "AJOB Neurosci",
  ["ajob primary research"] = "AJOB Prim Res",
  ["ajp reports"] = "AJP Rep",
  ["ajr. american journal of roentgenology"] = "AJR Am J Roentgenol",
  ["ajr: american journal of roentgenology"] = "AJR Am. J. Roentgenol.",
  ["ajs review. association for jewish studies"] = "AJS Rev",
  ["ajs: american journal of sociology"] = "AJS",
  ["ajs; american journal of sociology"] = "AJS",
  ["ajsp: reviews & reports"] = "AJSP Rev Rep",
  ["ak peters/crc recreational mathematics series"] = "AK Peters/CRC Recreat. Math. Ser.",
  ["akademaike iatrike"] = "Akad Iatr",
  ["akademie der wissenschaften und der literatur in mainz. abhandlungen der geistes- und sozialwissenschaftlichen klasse"] = "AbhMainz",
  ["akademija nauka i umjetnosti bosne i hercegovine"] = "Akad. Nauka Umjet. Bosne Hercegov. Rad. Odjelj. Prirod. Mat. Nauka",
  ["akademiska dzive"] = "Akad Dzive",
  ["akademiya nauk armenii"] = "Akad. Nauk Armenii Dokl.",
  ["akademiya nauk azerbay dzhana"] = "Dokl. Akad. Nauk Azerbay dzhana",
  ["akademiya nauk gruzii"] = "Soobshch. Akad. Nauk Gruzii",
  ["akademiya nauk latviiskoy ssr"] = "Magnit. Gidrodinamika",
  ["akademiya nauk respubliki armeniya"] = "Izv. Akad. Nauk Respub. Armeniya Mekh.",
  ["akademiya nauk respubliki kazakhstan"] = "Akad. Nauk Respub. Kazakhstan Trudy Astrofiz. Inst.",
  ["akademiya nauk respubliki kyrgyzstan"] = "Izv. Akad. Nauk Respub. Kyrgyzstan",
  ["akademiya nauk respubliki moldova"] = "Izv. Akad. Nauk Respub. Moldova Fiz. Tekhn.",
  ["akademiya nauk respubliki tadzhikistan"] = "Dokl. Akad. Nauk Respub. Tadzhikistan",
  ["akademiya nauk respubliki uzbekistan"] = "Uzbek. Mat. Zh.",
  ["akademiya nauk ukrainy"] = "Gibrid. Vychisl. Mashiny Kompleksy",
  ["akadēmiskā dzīve"] = "Akad Dzive",
  ["akce international journal of graphs and combinatorics"] = "AKCE Int. J. Graphs Comb.",
  ["akdeniz üniversitesi ziraat fakültesi dergisi = journal of faculty of agriculture akdeniz university"] = "Akdeniz Univ Ziraat Fak Derg",
  ["akp classics"] = "AKP Class.",
  ["akron business and economic review"] = "Akron Bus Econ Rev",
  ["akron law review"] = "Akron Law Rev",
  ["akroterion: quarterly for the classics in south africa"] = "Akroterion",
  ["akten der gesellschaft für griechische und hellenistische rechts- geschichte"] = "AGR",
  ["aktualʹnye voprosy veterinarnoj biologii"] = "Aktual Vopr Vet Biol",
  ["aktuelle dermatologie"] = "Aktuelle Derm",
  ["aktuelle endokrinologie und stoffwechsel"] = "Aktuel Endokrinol Stoffwechsel",
  ["aktuelle ernahrungsmedizin"] = "Aktuel Ernahrungsmed",
  ["aktuelle ernährungsmedizin"] = "Aktuel Ernahrungsmed",
  ["aktuelle gerontologie"] = "Aktuelle Gerontol",
  ["aktuelle oto-rhino-laryngologie"] = "Aktuelle Otorhinolaryngol",
  ["aktuelle probleme der dermatologie. current problems in dermatology. problèmes actuels de dermatologie"] = "Aktuelle Probl  Dermatol",
  ["aktuelle probleme der phoniatrie und logopadie. current problems in phoniatrics and logopedics. problemes actuels de phoniatrie et de logopedie"] = "Aktuel Probl Phoniatr Logop",
  ["aktuelle probleme der phoniatrie und logopädie. current problems in phoniatrics and logopedics. problèmes actuels de phoniatrie et de logopédie"] = "Aktuel Probl Phoniatr Logop",
  ["aktuelle probleme in chirurgie und orthopadie"] = "Aktuelle Probl. Chir. Orthop.",
  ["aktuelle probleme in chirurgie und orthopädie"] = "Aktuelle Probl Chir Orthop",
  ["aktuelle probleme in der chirurgie"] = "Aktuelle Probl Chir",
  ["aktuelle radiologie"] = "Aktuelle Radiol",
  ["aktuelle traumatologie"] = "Aktuelle Traumatol",
  ["aktuelle urologie"] = "Aktuelle Urol",
  ["akusherstvo i ginekologiia"] = "Akush Ginekol (Sofiia)",
  ["akusherstvo i ginekologii︠a︡"] = "Akush Ginekol (Mosk)",
  ["akusticheskii zhurnal [soviet physicsacoustics]"] = "Akust. Zh. [Sov. Phys. Acoust.]",
  ["akwesasne notes"] = "Akwesasne Notes",
  ["al azhar medical journal = majallat al-ṭibb al-azhar"] = "Al Azhar Med J",
  ["al darat"] = "Al Darat",
  ["al servizio del bosco e del paesaggio"] = "Al serv. bosco paesaggio",
  ["al-anbar journal of veterinary sciences"] = "Anbar J Vet Sci",
  ["al-maghrib al-tibbi. maroc medical"] = "Maroc Med",
  ["al-maghrib al-ṭibbī. maroc médical"] = "Maroc Med",
  ["al-mağallaẗ al-'irāqiyyaẗ li-buḥūt̲ al-sūq wa-ḥimāyaẗ al-mustahlik"] = "Almagallat Aliraqiyyat Libuhut Alsuq Wahimayat Almustahlik",
  ["al-maǧallaẗ al-ṭibbiyyaẗ al-bayṭariyyaẗ al-’irāqiyyaẗ = the iraqi journal of veterinary medicine"] = "Almagallat Altibbiyyat Albaytariyyat Aliraqiyyat",
  ["al-majallah al-tibbiyah al-'iraqiyah"] = "Al Majallah Al Tibbiyah Al Iraqiyah",
  ["al-majallah al-ʻarabīyah lil-bīʼāt al-jāffah"] = "Almagallat Alarabiyyat Lilbiat Algaffat",
  ["al-masāq : islam and the medieval mediterranean"] = "Masaq",
  ["al-mihan al-tibbiyah"] = "Al Mihan Al Tibbiyah",
  ["al-mustaqbal al-ʻarabī"] = "Almustaqbal Alarabi",
  ["al-mustaqbal al-’arabi"] = "Almustaqbal Alarabi",
  ["al-qanniš. boletín del taller de arqueología de alcañiz"] = "Al-Qannis",
  ["al-raida / institute for women's studies in the arab world, beirut university college"] = "Al Raida",
  ["al-raida / institute for women’s studies in the arab world, beirut university college"] = "Al Raida",
  ["al-usrah wa-al-ʻumrān al-basharī"] = "Al Usrah Wa Al Umran Al Bashari",
  ["al-usrah wa-al-’umran al-bashari"] = "Al Usrah Wa Al Umran Al Bashari",
  ["alabama dental review"] = "Ala Dent Rev",
  ["alabama heritage"] = "Ala Herit",
  ["alabama journal of mathematics"] = "Ala. J. Math.",
  ["alabama journal of medical sciences"] = "Ala. J. Med. Sci.",
  ["alabama law review"] = "Ala Law Rev",
  ["alabama medicine"] = "Ala. Med.",
  ["alabama medicine : journal of the medical association of the state of alabama"] = "Ala Med",
  ["alabama nurse"] = "Ala. Nurse",
  ["alabama social welfare"] = "Ala Soc Welf",
  ["alafo: revista de la asociacion latinoamericana de facultades de odontologia"] = "ALAFO",
  ["alafo; revista de la asociacion latinoamericana de facultades de odontologia"] = "ALAFO",
  ["alafo; revista de la asociación latinoamericana de facultades de odontología"] = "ALAFO",
  ["alaska health profile"] = "Alsk Health Profile",
  ["alaska history (anchorage, alaska)"] = "Alsk Hist",
  ["alaska journal (juneau, alaska : 1971)"] = "Alaska J",
  ["alaska medicine"] = "Alaska Med",
  ["alaska nurse"] = "Alaska Nurse",
  ["alaska statutes, 1962 : containing the general and permanent laws of the state of alaska / annotated, indexed, and printed by the michie company. alaska"] = "Alsk Statut 1962 Alsk",
  ["alaska statutes, 1962 : containing the general and permanent laws of the state of alaska. alaska"] = "Alsk Statut 1962 Alsk",
  ["alaska's health"] = "Alsk Health",
  ["alba regia. annales musei stephani regis"] = "AlbaRegia",
  ["alba regia: annales musei stephani regis"] = "Alba Regia",
  ["albanian journal of mathematics"] = "Albanian J. Math.",
  ["albany law journal of science & technology"] = "Albany Law J Sci Technol",
  ["albany law review"] = "Albany Law Rev",
  ["alber-reihe: philosophie"] = "Alber-Reihe Philos.",
  ["alberta history"] = "Alberta Hist",
  ["alberta law review"] = "Alta Law Rev",
  ["alberta medical bulletin"] = "Alberta Med Bull",
  ["alberta rn"] = "Alta RN",
  ["alberta rn / alberta association of registered nurses"] = "Alta RN",
  ["albertus magnus blatter"] = "Albertus Magnus Blaett",
  ["albertus magnus blätter"] = "Albertus Magnus Blaett",
  ["albion mathematics & applications series"] = "Albion Math. Appl. Ser.",
  ["albrecht von graefe's archiv für ophthalmologie"] = "Albrecht Von Graefes Arch Ophthalmol",
  ["albrecht von graefes archiv fur klinische und experimentelle ophthalmologie"] = "Albrecht von Graefes Arch. Klin. Exp. Ophthalmol.",
  ["albrecht von graefes archiv fur klinische und experimentelle ophthalmologie. albrecht von graefe’s archive for clinical and experimental ophthalmology"] = "Albrecht Von Graefes Arch Klin Exp Ophthalmol",
  ["albrecht von graefes archiv für klinische und experimentelle ophthalmologie. albrecht von graefe's archive for clinical and experimental ophthalmology"] = "Albrecht Von Graefes Arch Klin Exp Ophthalmol",
  ["albrecht von graefe’s archiv fur ophthalmologie"] = "Albrecht Von Graefes Arch Ophthalmol",
  ["alces (thunder bay, ont.)"] = "Alces",
  ["alcheringa: an australasian journal of palaeontology"] = "Alcheringa: Australas. J. Palaeontol.",
  ["alcohol (fayetteville, n.y.)"] = "Alcohol",
  ["alcohol (new york)"] = "Alcohol (N.Y.)",
  ["alcohol and alcoholism"] = "Alcohol Alcohol.",
  ["alcohol and alcoholism (oxford, oxfordshire)"] = "Alcohol Alcohol",
  ["alcohol and alcoholism (oxford, oxfordshire). supplement"] = "Alcohol Alcohol Suppl",
  ["alcohol and alcoholism (oxford, oxfordshire). supplement."] = "Alcohol Alcohol Suppl",
  ["alcohol and alcoholism. supplement"] = "Alcohol Alcohol. Suppl.",
  ["alcohol and drug research"] = "Alcohol Drug Res",
  ["alcohol health and research world"] = "Alcohol Health Res World",
  ["alcohol hygiene"] = "Alcohol Hyg",
  ["alcohol research & health : the journal of the national institute on alcohol abuse and alcoholism"] = "Alcohol Res Health",
  ["alcohol research : current reviews"] = "Alcohol Res",
  ["alcohol research and health"] = "Alcohol Res. Health",
  ["alcohol, drugs, and driving : abstracts and reviews"] = "Alcohol Drugs Driving",
  ["alcoholism and drug addiction"] = "Alcohol. Drug Addict.",
  ["alcoholism treatment quarterly"] = "Alcohol Treat Q",
  ["alcoholism, clinical and experimental research"] = "Alcohol Clin Exp Res",
  ["alcoholism: clinical and experimental research"] = "Alcohol.: Clin. Exp. Res.",
  ["aldrichimica acta"] = "Aldrichimica Acta",
  ["alea. latin american journal of probability and mathematical statistics"] = "ALEA Lat. Am. J. Probab. Math. Stat.",
  ["alembic club reprints"] = "Alembic Club Repr",
  ["aleph (jerusalem)"] = "Aleph (Jerus)",
  ["alexander blain hospital bulletin"] = "Alexander Blain Hosp Bull",
  ["alexandria dental journal"] = "Alex. Dent. J.",
  ["alexandria dental journal : adj"] = "Alex Dent J",
  ["alexandria engineering journal"] = "Alexandria Eng. J.",
  ["alexandria journal of agricultural research"] = "Alexandria J. Agric. Res.",
  ["alexandria journal of medicine"] = "Alexandria J. Med.",
  ["alexandria journal of pharmaceutical sciences"] = "Alex J Pharm Sci",
  ["alexandria journal of veterinary sciences"] = "Alex J Vet Sci",
  ["algae (korean phycological society)"] = "Algae",
  ["algal research"] = "Algal Res",
  ["algebra & number theory"] = "Algebra Number Theory",
  ["algebra and applications"] = "Algebr. Appl.",
  ["algebra and discrete mathematics"] = "Algebra Discrete Math.",
  ["algebra and logic"] = "Algebra Log.",
  ["algebra berichte"] = "Algebra Ber.",
  ["algebra colloquium"] = "Algebra Colloq.",
  ["algebra i logika. institut diskretnoi matematiki i informatiki"] = "Algebra Logika",
  ["algebra universalis"] = "Algebra universalis",
  ["algebra, logic and applications"] = "Algebra Logic Appl.",
  ["algebraic & geometric topology"] = "Algebraic Geom. Topol.",
  ["algebraic combinatorics"] = "Algebr. Comb.",
  ["algebraic geometry"] = "Algebraic Geom.",
  ["algebraic geometry and physics"] = "Algebraic Geom. Phys.",
  ["algebraic statistics"] = "Algebr. Stat.",
  ["algebras and representation theory"] = "Algebras Represent Theory",
  ["algebras, groups and geometries"] = "Algebras Groups Geom.",
  ["algerie medicale"] = "Alger Medicale",
  ["algérie médicale"] = "Alger Medicale",
  ["algological studies"] = "Arch Hydrobiol Suppl Algol Stud",
  ["algological studies (stuttgart, germany : 2007)"] = "Algol Stud (2007)",
  ["algologiia (kiev, ukraine)"] = "Algologiia (Kiev)",
  ["algorithmic finance"] = "Algorithmic Finance",
  ["algorithmica. an international journal in computer science"] = "Algorithmica",
  ["algorithms and combinatorics"] = "Algorithms Combin.",
  ["algorithms and computation in mathematics"] = "Algorithms Comput. Math.",
  ["algorithms for intelligent systems"] = "Algorithms Intell. Syst.",
  ["algorithms for molecular biology"] = "Algorithms Mol. Biol.",
  ["algorithms for molecular biology : amb"] = "Algorithms Mol Biol",
  ["alimentary pharmacology & therapeutics"] = "Aliment Pharmacol Ther",
  ["alimentary pharmacology and therapeutics"] = "Aliment. Pharmacol. Ther.",
  ["alimentation et la vie"] = "Aliment. Vie",
  ["alkalmazott matematikai lapok"] = "Alkalmaz. Mat. Lapok",
  ["alkaloids, chemistry and biology"] = "Alkaloids Chem. Biol.",
  ["alkoholizm i narkomania"] = "Alkohol Narkom",
  ["all earth"] = "All Earth",
  ["all life"] = "All Life",
  ["all life methods"] = "All Life Methods",
  ["allelopathy journal"] = "Allelopathy J.",
  ["allelopathy journal : official publication of international allelopathy foundation"] = "Allelopathy J",
  ["allemagne d'aujourd'hui : revue française d'information sur les deux l'allemagnes"] = "Allem Aujourdhui",
  ["allemagne d’aujourd’hui : revue francaise d’information sur les deux l’allemagnes"] = "Allem Aujourdhui",
  ["allergie (pointe-claire, quebec)."] = "Allerg Pointe Claire",
  ["allergie (pointe-claire, québec)"] = "Allerg Pointe Claire",
  ["allergie et immunologie"] = "Allerg Immunol (Paris)",
  ["allergie und asthma"] = "Allerg Asthma (Leipz)",
  ["allergie und asthmaforschung"] = "Allerg. Asthmaforsch.",
  ["allergie und immunologie"] = "Allerg Immunol (Leipz)",
  ["allergie- und asthmaforschung"] = "Allerg Asthmaforsch",
  ["allergo journal : interdisziplinare zeitschrift fur allergologie und umweltmedizin : organ der deutschen gesellschaft fur allergie- und immunitatsforschung"] = "Allergo J",
  ["allergo journal : interdisziplinäre zeitschrift für allergologie und umweltmedizin : organ der deutschen gesellschaft für allergie- und immunitätsforschung"] = "Allergo J",
  ["allergo journal international"] = "Allergo J Int",
  ["allergologia et immunopathologia"] = "Allergol Immunopathol (Madr)",
  ["allergologie select"] = "Allergol Select",
  ["allergology international : official journal of the japanese society of allergology"] = "Allergol Int",
  ["allergy & clinical immunology international : official organ of the international association of allergology and clinical immunology"] = "Allergy Clin Immunol Int",
  ["allergy & rhinology (providence, r.i.)"] = "Allergy Rhinol (Providence)",
  ["allergy and asthma proceedings"] = "Allergy Asthma Proc",
  ["allergy and asthma proceedings : the official journal of regional and state allergy societies"] = "Allergy Asthma Proc",
  ["allergy proceedings"] = "Allergy Proc.",
  ["allergy proceedings : the official journal of regional and state allergy societies"] = "Allergy Proc",
  ["allergy, asthma & immunology research"] = "Allergy Asthma Immunol Res",
  ["allergy, asthma, and clinical immunology : official journal of the canadian society of allergy and clinical immunology"] = "Allergy Asthma Clin Immunol",
  ["allgemeine forst und jagdzeitung"] = "Allg. Forst Jagdztg.",
  ["allgemeine forst zeitschrift für waldwirtschaft und umweltvorsorge [afz-der wald]"] = "Allg. Forst Z. Waldwirtsch. Umweltvorsorge",
  ["allgemeine forst- und jagdzeitung"] = "Allg. Forst- Jagdztg.",
  ["allgemeine homoopathische zeitung fur wissenschaftliche und praktische homoopathie"] = "Allg Homoopath Ztg",
  ["allgemeine homöopathische zeitung für wissenschaftliche und praktische homöopathie"] = "Allg Homoopath Ztg",
  ["allgemeine zeitschrift für philosophie"] = "AZP",
  ["allgemeines statistisches archiv"] = "All Stat Arch",
  ["allied health and behavioral sciences"] = "Allied Health Behav Sci",
  ["alma mater"] = "Alma Mater (Baltimore)",
  ["alma mater philippina"] = "Alma Mater Philipp",
  ["alma mater studiorum almanacco calabrese almoraima"] = "AlmaMaterSt",
  ["almanac (national academy of television arts and sciences (u.s.). international council)"] = "Almanac",
  ["alpen, die"] = "Alpen",
  ["alpha omega france"] = "Alpha Omega Fr",
  ["alpha omegan"] = "Alpha Omegan",
  ["alpine and mediterranean quaternary"] = "Alp. Mediterr. Quat.",
  ["alpine botany"] = "Alp Bot",
  ["alpine entomology"] = "Alp. Entomol.",
  ["altamura. bollettino dell’archivio-biblioteca-museo civico"] = "Altamura",
  ["alter orient und altes testament. veröffentlichungen zur kultur und geschichte des alten orients und des alten testaments"] = "AOAT",
  ["altern und entwicklung (aging and development)"] = "Altern Entwickl. Aging Dev.",
  ["altern und entwicklung. aging and development"] = "Altern Entwickl Aging Dev",
  ["alternativas (investigación operativa en planificación familiar y atención materno-infantil para américa latina y el caribe (project))"] = "Alternativas",
  ["alternative & complementary therapies : a new bimonthly publication for health care practitioners"] = "Altern Complement Ther",
  ["alternative & integrative medicine"] = "Altern Integr Med",
  ["alternative (auckland, n.z. : 2005)"] = "AlterNative (Nga Pae Maramatanga (Organ))",
  ["alternative lifestyles"] = "Alternative Lifestyles",
  ["alternative medicine review"] = "Altern. Med. Rev.",
  ["alternative medicine review : a journal of clinical therapeutic"] = "Altern Med Rev",
  ["alternative therapies in health and medicine"] = "Altern Ther Health Med",
  ["alternatives (boulder, colo.)"] = "Alternatives (Boulder)",
  ["alternatives to laboratory animals"] = "Altern. Lab. Anim.",
  ["alternatives to laboratory animals : atla"] = "Altern Lab Anim",
  ["altertümer von pergamon"] = "AvP",
  ["altex : alternativen zu tierexperimenten"] = "ALTEX",
  ["altérités (université de montréal. département d'anthropologie)"] = "Alterites (Univ Montr Dep Anthropol)",
  ["alto medioevo"] = "AltoMed",
  ["altorientalische forschungen"] = "AoF",
  ["aluminium (duesseldorf)"] = "Aluminium (Duesseldorf)",
  ["aluminum transactions"] = "Alum. Trans.",
  ["alumnae magazine (columbia university-presbyterian hospital school of nursing alumnae association)"] = "Alumnae Mag Columbia Univ Presbyt Hosp Sch Nurs Alumnae Assoc",
  ["alumnae magazine, columbia university-presbyterian hospital school of nursing alumnae association"] = "Alumnae Mag. Columbia Univ. Presbyt. Hosp. Sch. Nurs. Alumnae Assoc.",
  ["alumnae magazine, johns hopkins hospital school of nursing alumnae association"] = "Alumnae Mag. Johns Hopkins Hosp. Sch. Nurs. Alumnae Assoc.",
  ["alumni bulletin - school of dentistry, indiana university"] = "Alumni Bull Sch Dent Indiana Univ",
  ["alumni bulletin, school of dentistry, indiana university"] = "Alumni Bull. Sch. Dent. Indiana Univ.",
  ["alumni bulletin, university of michigan school of dentistry"] = "Alumni Bull. Univ. Mich. Sch. Dent.",
  ["alumni bulletin. long island college of medicine. alumni association"] = "Alumni Bull Long Island Coll Med Brooklyn",
  ["alumni bulletin. university of michigan. school of dentistry"] = "Alumni Bull Univ Mich Sch Dent",
  ["alumni bulletin. western reserve university. school of medicine"] = "Alumni Bull West Reserv Univ Sch Med",
  ["alumni magazine (columbia university-presbyterian hospital school of nursing alumni association)"] = "Alumni Mag Columbia Univ Presbyt Hosp Sch Nurs Alumni Assoc",
  ["alumni magazine, alumni association of the johns hopkins hospital school of nursing"] = "Alumni Mag.",
  ["alumni magazine, columbia university-presbyterian hospital school of nursing alumni association"] = "Alumni Mag. Columbia Univ. Presbyt. Hosp. Sch. Nurs. Alumni Assoc.",
  ["alzheimer disease and associated disorders"] = "Alzheimer Dis Assoc Disord",
  ["alzheimer's & dementia (amsterdam, netherlands)"] = "Alzheimers Dement (Amst)",
  ["alzheimer's & dementia (new york, n. y.)"] = "Alzheimers Dement (N Y)",
  ["alzheimer's & dementia : the journal of the alzheimer's association"] = "Alzheimers Dement",
  ["alzheimer's care quarterly"] = "Alzheimers Care Q",
  ["alzheimer's care today"] = "Alzheimers care today",
  ["alzheimer's disease research journal"] = "Alzheimers Dis Res J",
  ["alzheimer's reports"] = "Alzheimers Rep",
  ["alzheimer's research & therapy"] = "Alzheimers Res Ther",
  ["alzheimer's research & therapy open access"] = "Alzheimers Res Ther Open Access",
  ["alzheimer's, dementia & cognitive neurology"] = "Alzheimers Dement Cogn Neurol",
  ["alzheimers disease & dementia"] = "Alzheimers Dis Dement",
  ["alzheimer’s & dementia : the journal of the alzheimer’s association"] = "Alzheimers Dement",
  ["am [reports]. united states. office of aviation medicine"] = "AM Rep",
  ["am reports"] = "AM Rep.",
  ["ama journal of ethics"] = "AMA J Ethics",
  ["amatus lusitanus; revista de medicina e cirurgia"] = "Amatus Lusit",
  ["amb : revista da associação médica brasileira"] = "AMB Rev Assoc Med Bras",
  ["amb express"] = "AMB Express",
  ["amb: revista da associacao medica brasileira"] = "AMB Rev. Assoc. Med. Bras.",
  ["amb; revista da associacao medica brasileira"] = "AMB Rev Assoc Med Bras",
  ["ambio: a journal of the human environment"] = "AMBIO",
  ["ambulatory care"] = "Ambul Care",
  ["ambulatory outreach"] = "Ambul Outreach",
  ["ambulatory pediatrics"] = "Ambul. Pediatr.",
  ["ambulatory pediatrics : the official journal of the ambulatory pediatric association"] = "Ambul Pediatr",
  ["ambulatory surgery"] = "Ambul Surg",
  ["amd-tr [reports]. united states. aerospace medical division"] = "AMD TR Rep",
  ["amd-tr reports"] = "AMD. TR Rep.",
  ["amdi bollettino"] = "AMDI Boll",
  ["ame case reports"] = "AME Case Rep",
  ["ame medical journal"] = "AME Med J",
  ["amerasia journal"] = "Amerasia J",
  ["america clínica"] = "Am Clin",
  ["america indigena"] = "Am Indig",
  ["america indígena"] = "Am Indig",
  ["america latina"] = "Am Lat",
  ["america, history and life"] = "Am Hist Life",
  ["american annals of the deaf"] = "Am Ann Deaf",
  ["american anthropologist"] = "Am Anthropol",
  ["american antiquity"] = "Am Antiq",
  ["american archives of rehabilitation therapy"] = "Am Arch Rehabil Ther",
  ["american association for the advancement of science"] = "Science",
  ["american association for the advancement of science. science"] = "Science",
  ["american association of industrial nurses journal"] = "Am Assoc Ind Nurses J",
  ["american association of petroleum geologists bulletin"] = "Am. Assoc. Pet. Geol. Bull.",
  ["american astronautical society"] = "J. Astronaut. Sci.",
  ["american baptist quarterly"] = "Am Baptist Q",
  ["american bar association journal. american bar association"] = "Am Bar Assoc J",
  ["american bar foundation research journal. american bar foundation"] = "Am Bar Found Res J",
  ["american bee journal"] = "Am. Bee. J.",
  ["american behavioral scientist"] = "Am. Behav. Sci.",
  ["american biology teacher"] = "Am. Biol. Teach.",
  ["american biotechnology laboratory"] = "Am Biotechnol Lab",
  ["american book prices current"] = "Am Book Prices",
  ["american ceramic society bulletin"] = "Am. Ceram. Soc. Bull.",
  ["american chemical journal"] = "Am. Chem. J.",
  ["american chemical science journal"] = "Am. Chem. Sci. J.",
  ["american christmas tree journal"] = "Am. Christmas Tree J.",
  ["american clinical laboratory"] = "Am Clin Lab",
  ["american college of physicians observer"] = "Am Coll Physicians Obs",
  ["american concrete institute, sp"] = "Am. Concr. Inst., SP",
  ["american corrective therapy journal"] = "Am Correct Ther J",
  ["american demographics"] = "Am Demogr",
  ["american druggist"] = "Am Drug",
  ["american economic journal. applied economics"] = "Am Econ J Appl Econ",
  ["american economic journal. economic policy"] = "Am Econ J Econ Policy",
  ["american economic journal. microeconomics"] = "Am Econ J Microecon",
  ["american economic review"] = "Amer. Econ. Rev.",
  ["american economic review. insights"] = "Am Econ Rev Insights",
  ["american economist"] = "Amer. Economist",
  ["american education"] = "Am Educ",
  ["american educational research journal"] = "Am Educ Res J",
  ["american enterprise"] = "Amer. Enterprise",
  ["american enterprise (washington, d.c.)"] = "Am Enterp",
  ["american entomologist"] = "Am. Entomol.",
  ["american ethnologist"] = "Am Ethnol",
  ["american family physician"] = "Am Fam Physician",
  ["american family physician/gp"] = "Am Fam Physician GP",
  ["american fern journal"] = "Am Fern J",
  ["american health"] = "Am Health",
  ["american health & drug benefits"] = "Am Health Drug Benefits",
  ["american health (new york, n.y.)"] = "Am Health",
  ["american heart hospital journal"] = "Am. Heart Hosp. J.",
  ["american heart journal"] = "Am Heart J",
  ["american heritage"] = "Am Herit",
  ["american heritage of invention & technology"] = "Am Herit Invent Technol",
  ["american historical association, institutional services program"] = "Am Hist Assoc Inst Serv Program",
  ["american historical review"] = "AHR",
  ["american history illustrated"] = "Am Hist Illus",
  ["american horseman"] = "Am Horseman",
  ["american imago"] = "Am. Imago",
  ["american indian and alaska native mental health research"] = "Am. Indian Alsk. Native Ment. Health Res.",
  ["american indian and alaska native mental health research (monographic series)"] = "Am Indian Alsk Native Ment Health Res Monogr Ser",
  ["american indian and alaska native mental health research (online)"] = "Am Indian Alsk Native Ment Health Res",
  ["american indian and alaska native mental health research : journal of the national center"] = "Am Indian Alsk Native Ment Health Res",
  ["american indian and alaska native mental health research. monograph series"] = "Am. Indian Alsk. Native Ment. Health Res. Monogr. Ser.",
  ["american indian and alaska native mental health research. monographseries am indian alsk native ment health re"] = "Monogr Ser",
  ["american indian art magazine"] = "Am Indian Art Mag",
  ["american indian culture and research journal"] = "Am Indian Cult Res J",
  ["american indian quarterly"] = "Am Indian Q",
  ["american industrial hygiene association journal"] = "Am Ind Hyg Assoc J",
  ["american industrial hygiene association quarterly"] = "Am Ind Hyg Assoc Q",
  ["american international journal of contemporary research"] = "Am Int J Contemp Res",
  ["american international journal of research in formal, applied and natural sciences"] = "Am Int J Res Form Appl Nat Sci",
  ["american jewish archives"] = "Am Jew Arch",
  ["american jewish historical quarterly"] = "Am Jew Hist Q",
  ["american jewish history"] = "Am Jew Hist",
  ["american journal of accountable care"] = "Am J Accountable Care",
  ["american journal of acupuncture"] = "Am J Acupunct",
  ["american journal of advanced drug delivery"] = "Am J Adv Drug Deliv",
  ["american journal of aging science and research"] = "Am J Aging Sci Res",
  ["american journal of agricultural and biological sciences"] = "Am J Agric Biol Sci",
  ["american journal of agricultural economics"] = "Am J Agric Econ",
  ["american journal of alternative agriculture"] = "Am. J. Altern. Agric.",
  ["american journal of alzheimer's disease (columbia, mo.)"] = "Am J Alzheimers Dis (Columbia)",
  ["american journal of alzheimer's disease and other dementias"] = "Am J Alzheimers Dis Other Demen",
  ["american journal of alzheimer’s disease and other dementias"] = "Am. J. Alzheimers Dis. Other Demen.",
  ["american journal of analytical chemistry"] = "Am J Analyt Chem",
  ["american journal of anatomy"] = "Am. J. Anat.",
  ["american journal of ancient history"] = "AJAH",
  ["american journal of anesthesiology"] = "Am. J. Anesthesiol.",
  ["american journal of animal and veterinary sciences"] = "Am J Anim Vet Sci",
  ["american journal of applied mathematics and statistics"] = "Am J Appl Math Stat",
  ["american journal of applied sciences"] = "Am J Appl Sci",
  ["american journal of archaeology"] = "AJA",
  ["american journal of art therapy"] = "Am J Art Ther",
  ["american journal of audiology"] = "Am J Audiol",
  ["american journal of biochemistry & biotechnology"] = "Am J Biochem Biotechnol",
  ["american journal of biochemistry and biotechnology"] = "Am. J. Biochem. Biotechnol.",
  ["american journal of biochemistry and molecular biology"] = "Am J Biochem Mol Biol",
  ["american journal of bioethics"] = "Am. J. Bioethics",
  ["american journal of bioinformatics and computational biology"] = "Am J  Bioinform Comput Biol",
  ["american journal of biological anthropology"] = "Am. J. Biol. Anthropol.",
  ["american journal of biomedical engineering"] = "Am J Biomed Eng",
  ["american journal of biomedical science & research"] = "Am J Biomed Sci Res",
  ["american journal of biomedical sciences"] = "Am J Biomed Sci",
  ["american journal of biostatistics"] = "Am J Biostat",
  ["american journal of blood research"] = "Am J Blood Res",
  ["american journal of botany"] = "Am J Bot",
  ["american journal of cancer prevention"] = "Am J Cancer Prev",
  ["american journal of cancer research"] = "Am J Cancer Res",
  ["american journal of cancer science"] = "Am J Cancer Sci",
  ["american journal of cancer therapy and pharmacology"] = "Am J Cancer Ther Pharmacol",
  ["american journal of cardiac imaging"] = "Am J Card Imaging",
  ["american journal of cardiology"] = "Am. J. Cardiol.",
  ["american journal of cardiovascular disease"] = "Am J Cardiovasc Dis",
  ["american journal of cardiovascular drugs"] = "Am. J. Cardiovasc. Drugs",
  ["american journal of cardiovascular drugs : drugs, devices, and other interventions"] = "Am J Cardiovasc Drugs",
  ["american journal of cardiovascular pathology"] = "Am. J. Cardiovasc. Pathol.",
  ["american journal of chinese medicine"] = "Am. J. Chin. Med.",
  ["american journal of chinese studies"] = "Am J Chin Stud",
  ["american journal of climate change"] = "Am J Clim Change",
  ["american journal of clinical and experimental immunology"] = "Am J Clin Exp Immunol",
  ["american journal of clinical and experimental obstetrics and gynecology"] = "Am J Clin Exp Obstet Gynecol",
  ["american journal of clinical and experimental urology"] = "Am J Clin Exp Urol",
  ["american journal of clinical cancer research"] = "Am J Clin Cancer Res",
  ["american journal of clinical dermatology"] = "Am J Clin Dermatol",
  ["american journal of clinical hypnosis"] = "Am. J. Clin. Hypn.",
  ["american journal of clinical nutrition"] = "Am. J. Clin. Nutr.",
  ["american journal of clinical oncology"] = "Am J Clin Oncol",
  ["american journal of clinical pathology"] = "Am J Clin Pathol",
  ["american journal of community psychology"] = "Am J Community Psychol",
  ["american journal of contact dermatitis"] = "Am. J. Contact Dermat.",
  ["american journal of contact dermatitis : official journal of the american contact dermatitis society"] = "Am J Contact Dermat",
  ["american journal of correction"] = "Am J Correct",
  ["american journal of criminal justice : ajcj"] = "Am J Crim Justice",
  ["american journal of criminal law"] = "Am J Crim Law",
  ["american journal of critical care"] = "Am. J. Crit. Care",
  ["american journal of critical care : an official publication, american association of critical-care nurses"] = "Am J Crit Care",
  ["american journal of cultural sociology"] = "Am J Cult Sociol",
  ["american journal of dance therapy"] = "Am J Dance Ther",
  ["american journal of dentistry"] = "Am J Dent",
  ["american journal of dermatopathology"] = "Am. J. Dermatopathol.",
  ["american journal of digestive disease"] = "Am J Dig Dis (Madison)",
  ["american journal of digestive diseases"] = "Am. J. Dig. Dis.",
  ["american journal of disaster medicine"] = "Am J Disaster Med",
  ["american journal of diseases of children"] = "Am. J. Dis. Child.",
  ["american journal of diseases of children (1911)"] = "Am J Dis Child",
  ["american journal of diseases of children (1960)"] = "Am J Dis Child",
  ["american journal of drug and alcohol abuse"] = "Am. J. Drug Alcohol Abuse",
  ["american journal of drug delivery"] = "Am. J. Drug Delivery",
  ["american journal of economics and sociology"] = "Am J Econ Sociol",
  ["american journal of education (chicago, ill.)"] = "Am J Educ (Chic Ill)",
  ["american journal of educational research"] = "Am J Educ Res",
  ["american journal of eeg technology"] = "Am. J. EEG Technol.",
  ["american journal of electroneurodiagnostic technology"] = "Am J Electroneurodiagnostic Technol",
  ["american journal of emergency medicine"] = "Am. J. Emerg. Med.",
  ["american journal of engineering education"] = "Am J Eng Educ",
  ["american journal of engineering research"] = "Am. J. Eng. Res.",
  ["american journal of enology and viticulture"] = "Am J Enol Vitic",
  ["american journal of entrepreneurship"] = "Am J Entrep",
  ["american journal of environmental engineering"] = "Am J Environ Engineer",
  ["american journal of environmental sciences"] = "Am J Environ Sci",
  ["american journal of epidemiology"] = "Am J Epidemiol",
  ["american journal of ethics & medicine : a national medical student journal published by the university of pennsylvania school of medicine"] = "Am J Ethics Med",
  ["american journal of family law"] = "Am J Fam Law",
  ["american journal of food technology"] = "Am. J. Food Technol.",
  ["american journal of forensic medicine and pathology"] = "Am. J. Forensic Med. Pathol.",
  ["american journal of gastroenterology"] = "Am. J. Gastroenterol.",
  ["american journal of gastroenterology supplements (print)"] = "Am J Gastroenterol Suppl",
  ["american journal of geriatric pharmacotherapy"] = "Am. J. Geriatr. Pharmacother.",
  ["american journal of geriatric psychiatry"] = "Am. J. Geriatr. Psychiatry",
  ["american journal of health behavior"] = "Am J Health Behav",
  ["american journal of health economics"] = "Am J Health Econ",
  ["american journal of health education"] = "Am J Health Educ",
  ["american journal of health planning"] = "Am J Health Plann",
  ["american journal of health promotion"] = "Am. J. Health Promot.",
  ["american journal of health promotion : ajhp"] = "Am J Health Promot",
  ["american journal of health sciences"] = "Am J Health Sci",
  ["american journal of health studies"] = "Am J Health Stud",
  ["american journal of health system pharmacy"] = "Am. J. Health-Syst. Pharm.",
  ["american journal of health-system pharmacy"] = "Am. J. Health. Syst. Pharm.",
  ["american journal of health-system pharmacy : ajhp : official journal of the american society of health-system pharmacists"] = "Am J Health Syst Pharm",
  ["american journal of hematology"] = "Am J Hematol",
  ["american journal of hematology/oncology"] = "Am J Hematol Oncol",
  ["american journal of hospice and palliative care"] = "Am. J. Hosp. Palliat. Care",
  ["american journal of hospice care"] = "Am. J. Hosp. Care",
  ["american journal of hospital medicine"] = "Am J Hosp Med",
  ["american journal of hospital pharmacy"] = "Am J Hosp Pharm",
  ["american journal of human biology"] = "Am. J. Hum. Biol.",
  ["american journal of human biology : the official journal of the human biology council"] = "Am J Hum Biol",
  ["american journal of human genetics"] = "Am J Hum Genet",
  ["american journal of hygiene"] = "Am J Hyg",
  ["american journal of hypertension"] = "Am J Hypertens",
  ["american journal of hypertension : journal of the american society of hypertension"] = "Am J Hypertens",
  ["american journal of immunology"] = "Am J Immunol",
  ["american journal of industrial medicine"] = "Am J Ind Med",
  ["american journal of infection control"] = "Am J Infect Control",
  ["american journal of infectious diseases"] = "Am J Infect Dis",
  ["american journal of information management"] = "Am J Inf Manag",
  ["american journal of kidney diseases"] = "Am. J. Kidney Dis.",
  ["american journal of kidney diseases : the official journal of the national kidney foundation"] = "Am J Kidney Dis",
  ["american journal of knee surgery"] = "Am. J. Knee Surg.",
  ["american journal of laboratory medicine"] = "Am J Lab Med",
  ["american journal of law & medicine"] = "Am J Law Med",
  ["american journal of law and medicine"] = "Am. J. Law Med.",
  ["american journal of leukemia research"] = "Am J Leuk Res",
  ["american journal of lifestyle medicine"] = "Am J Lifestyle Med",
  ["american journal of managed care"] = "Am. J. Manag. Care",
  ["american journal of mathematical and management sciences"] = "Am. J. Math. Manage. Sci.",
  ["american journal of mathematics"] = "Amer. J. Math.",
  ["american journal of media psychology"] = "Am J Media Psychol",
  ["american journal of medical case reports"] = "Am J Med Case Rep",
  ["american journal of medical electronics"] = "Am. J. Med. Electron.",
  ["american journal of medical genetics"] = "Am J Med Genet",
  ["american journal of medical genetics part a"] = "Am. J. Med. Genet. Part A",
  ["american journal of medical genetics part b: neuropsychiatric genetics"] = "Am. J. Med. Genet. Part B: Neuropsychiatr. Genet.",
  ["american journal of medical genetics part c: seminars in medical genetics"] = "Am. J. Med. Genet. Part C: Semin. Med. Genet.",
  ["american journal of medical genetics. part a"] = "Am J Med Genet A",
  ["american journal of medical genetics. part b, neuropsychiatric genetics"] = "Am. J. Med. Genet. B Neuropsychiatr. Genet.",
  ["american journal of medical genetics. part b, neuropsychiatric genetics : the official publication of the international society of psychiatric genetics"] = "Am J Med Genet B Neuropsychiatr Genet",
  ["american journal of medical genetics. part c, seminars in medical genetics"] = "Am J Med Genet C Semin Med Genet",
  ["american journal of medical genetics. supplement"] = "Am J Med Genet Suppl",
  ["american journal of medical quality"] = "Am. J. Med. Qual.",
  ["american journal of medical quality : the official journal of the american college of medical quality"] = "Am J Med Qual",
  ["american journal of medical research (new york, n.y.)"] = "Am J Med Res (N Y)",
  ["american journal of medical technology"] = "Am. J. Med. Technol.",
  ["american journal of medicine"] = "Am. J. Med.",
  ["american journal of men's health"] = "Am J Mens Health",
  ["american journal of mental deficiency"] = "Am J Ment Defic",
  ["american journal of mental retardation"] = "Am. J. Ment. Retard.",
  ["american journal of mental retardation : ajmr"] = "Am J Ment Retard",
  ["american journal of microbiological research"] = "Am J Microbiol Res",
  ["american journal of microbiology and biotechnology"] = "Am J Microbiol Biotechnol",
  ["american journal of mining and metallurgy"] = "Am. J. Min. Metall.",
  ["american journal of modern chromatography"] = "Am J Mod Chromatogr",
  ["american journal of molecular biology"] = "Am J Mol Biol",
  ["american journal of nephrology"] = "Am J Nephrol",
  ["american journal of neurodegenerative disease"] = "Am J Neurodegener Dis",
  ["american journal of neuroprotection and neuroregeneration"] = "Am. J. Neuroprot. Neuroregener.",
  ["american journal of neuroscience"] = "Am J Neurosci",
  ["american journal of nuclear medicine and molecular imaging"] = "Am J Nucl Med Mol Imaging",
  ["american journal of numismatics"] = "AJN",
  ["american journal of nursing"] = "Am. J. Nurs.",
  ["american journal of obstetrics & gynecology mfm"] = "Am J Obstet Gynecol MFM",
  ["american journal of obstetrics and gynecology"] = "Am J Obstet Gynecol",
  ["american journal of occupational therapy"] = "Am. J. Occup. Ther.",
  ["american journal of ophthalmology"] = "Am J Ophthalmol",
  ["american journal of ophthalmology case reports"] = "Am J Ophthalmol Case Rep",
  ["american journal of optometry and archives of american academy of optometry"] = "Am J Optom Arch Am Acad Optom",
  ["american journal of optometry and physiological optics"] = "Am J Optom Physiol Opt",
  ["american journal of orthodontics"] = "Am J Orthod",
  ["american journal of orthodontics and dentofacial orthopedics"] = "Am. J. Orthod. Dentofacial Orthop.",
  ["american journal of orthodontics and dentofacial orthopedics : official publication of the american association of orthodontists, its constituent societies, and the american board of orthodontics"] = "Am J Orthod Dentofacial Orthop",
  ["american journal of orthodontics and oral surgery"] = "Am J Orthod Oral Surg",
  ["american journal of orthopedic surgery"] = "Am. J. Orthop. Surg.",
  ["american journal of orthopedics"] = "Am. J. Orthop.",
  ["american journal of orthopedics (belle mead, n.j.)"] = "Am J Orthop (Belle Mead NJ)",
  ["american journal of orthopsychiatry"] = "Am. J. Orthopsychiatry",
  ["american journal of otolaryngology"] = "Am J Otolaryngol",
  ["american journal of otology"] = "Am. J. Otol.",
  ["american journal of pathology"] = "Am. J. Pathol.",
  ["american journal of pediatric hematology/oncology"] = "Am. J. Pediatr. Hematol. Oncol.",
  ["american journal of perinatology"] = "Am J Perinatol",
  ["american journal of pharmaceutical education"] = "Am J Pharm Educ",
  ["american journal of pharmacogenomics"] = "Am. J. Pharmacogenomics",
  ["american journal of pharmacogenomics : genomics-related research in drug development and clinical practice"] = "Am J Pharmacogenomics",
  ["american journal of pharmacological sciences"] = "Am J Pharmacol Sci",
  ["american journal of pharmacology and toxicology"] = "Am. J. Pharmacol. Toxicol.",
  ["american journal of pharmacy and the sciences supporting public health"] = "Am J Pharm Sci Support Public Health",
  ["american journal of philology"] = "Am J Philol",
  ["american journal of physical anthropology"] = "Am J Phys Anthropol",
  ["american journal of physical medicine"] = "Am J Phys Med",
  ["american journal of physical medicine & rehabilitation"] = "Am J Phys Med Rehabil",
  ["american journal of physical medicine & rehabilitation / association of academic physiatrists"] = "Am J Phys Med Rehabil",
  ["american journal of physical medicine and rehabilitation"] = "Am. J. Phys. Med. Rehabil.",
  ["american journal of physics"] = "Am J Phys",
  ["american journal of physiologic imaging"] = "Am J Physiol Imaging",
  ["american journal of physiology"] = "Am. J. Physiol.",
  ["american journal of physiology cell physiology"] = "Am. J. Physiol. Cell Physiol.",
  ["american journal of physiology endocrinology and metabolism"] = "Am. J. Physiol. Endocrinol. Metab.",
  ["american journal of physiology heart and circulatory physiology"] = "Am. J. Physiol. Heart  Circ. Physiol.",
  ["american journal of physiology lung cellular and molecular physiology"] = "Am. J. Physiol. Lung Cell. Mol. Physiol.",
  ["american journal of physiology regulatory, integrative and comparative physiology"] = "Am. J. Physiol. Regul. Intgr. Comp. Physiol.",
  ["american journal of physiology renal physiology"] = "Am. J. Physiol. Renal Physiol.",
  ["american journal of physiology. cell physiology"] = "Am J Physiol Cell Physiol",
  ["american journal of physiology. endocrinology and metabolism"] = "Am J Physiol Endocrinol Metab",
  ["american journal of physiology. gastrointestinal and liver physiology"] = "Am J Physiol Gastrointest Liver Physiol",
  ["american journal of physiology. heart and circulatory physiology"] = "Am J Physiol Heart Circ Physiol",
  ["american journal of physiology. lung cellular and molecular physiology"] = "Am J Physiol Lung Cell Mol Physiol",
  ["american journal of physiology. regulatory, integrative and comparative physiology"] = "Am J Physiol Regul Integr Comp Physiol",
  ["american journal of physiology. renal physiology"] = "Am J Physiol Renal Physiol",
  ["american journal of phytomedicine and clinical therapeutics"] = "Am. J. Phytomed. Clin. Ther.",
  ["american journal of plant science"] = "Am. J. Plant Sci.",
  ["american journal of plant sciences"] = "Am J Plant Sci",
  ["american journal of play"] = "Am J Play",
  ["american journal of political science"] = "Am J Pol Sci",
  ["american journal of potato research"] = "Am. J. Potato Res.",
  ["american journal of potato research : an official publication of the potato association of america"] = "Am J Potato Res",
  ["american journal of practical nursing"] = "Am J Pract Nurs",
  ["american journal of preventive cardiology"] = "Am J Prev Cardiol",
  ["american journal of preventive medicine"] = "Am J Prev Med",
  ["american journal of primatology"] = "Am J Primatol",
  ["american journal of proctology"] = "Am J Proctol",
  ["american journal of proctology, gastroenterology & colon & rectal surgery"] = "Am J Proctol Gastroenterol Colon Rectal Surg",
  ["american journal of proctology, gastroenterology and colon and rectal surgery"] = "Am. J. Proctol. Gastroenterol. Colon Rectal Surg.",
  ["american journal of psychiatric rehabilitation"] = "Am J Psychiatr Rehabil",
  ["american journal of psychiatry"] = "Am. J. Psychiatry",
  ["american journal of psychoanalysis"] = "Am J Psychoanal",
  ["american journal of psychology"] = "Am. J. Psychol.",
  ["american journal of psychotherapy"] = "Am J Psychother",
  ["american journal of public health"] = "Am J Public Health",
  ["american journal of public health (new york, n.y. : 1912)"] = "Am J Public Health (N Y)",
  ["american journal of public health and the nation's health"] = "Am J Public Health Nations Health",
  ["american journal of public health and the nation’s health"] = "Am. J. Public Health Nations Health",
  ["american journal of public health research"] = "Am J Public Health Res",
  ["american journal of public hygiene"] = "Am J Public Hygiene",
  ["american journal of recreation therapy"] = "Am J Recreat Ther",
  ["american journal of reproductive immunology"] = "Am. J. Reprod. Immunol.",
  ["american journal of reproductive immunology (new york, n.y. : 1989)"] = "Am J Reprod Immunol",
  ["american journal of reproductive immunology : ajri : official journal of the american society for the immunology of reproduction and the international coordination committee for immunology of reproduction"] = "Am J Reprod Immunol",
  ["american journal of reproductive immunology and microbiology"] = "Am. J. Reprod. Immunol. Microbiol.",
  ["american journal of reproductive immunology and microbiology : ajrim"] = "Am J Reprod Immunol Microbiol",
  ["american journal of research communication"] = "Am J Res Commun",
  ["american journal of respiratory and critical care medicine"] = "Am J Respir Crit Care Med",
  ["american journal of respiratory cell and molecular biology"] = "Am J Respir Cell Mol Biol",
  ["american journal of respiratory medicine : drugs, devices, and other interventions"] = "Am J Respir Med",
  ["american journal of rhinology"] = "Am J Rhinol",
  ["american journal of rhinology & allergy"] = "Am J Rhinol Allergy",
  ["american journal of robotic surgery"] = "Am J Robot Surg",
  ["american journal of roentgenology"] = "Am. J. Roentgenol.",
  ["american journal of roentgenology and radium therapy"] = "Am. J. Roentgenol. Radium Ther.",
  ["american journal of roentgenology, radium therapy and nuclear medicine"] = "Am. J. Roentgenol. Radium Ther. Nucl. Med.",
  ["american journal of safe patient handling & movement"] = "Am J Safe Patient Handl Mov",
  ["american journal of science"] = "Am. J. Sci.",
  ["american journal of sexuality education"] = "Am J Sex Educ",
  ["american journal of sociology"] = "Am. J. Sociol.",
  ["american journal of speech-language pathology"] = "Am J Speech Lang Pathol",
  ["american journal of speech-language pathology / american speech-language-hearing association"] = "Am J Speech Lang Pathol",
  ["american journal of sports medicine"] = "Am. J. Sports Med.",
  ["american journal of stem cells"] = "Am J Stem Cells",
  ["american journal of surgery"] = "Am J Surg",
  ["american journal of surgical pathology"] = "Am. J. Surg. Pathol.",
  ["american journal of syphilis, gonorrhea, and venereal diseases"] = "Am J Syph Gonorrhea Vener Dis",
  ["american journal of the medical sciences"] = "Am. J. Med. Sci.",
  ["american journal of therapeutics"] = "Am J Ther",
  ["american journal of translational research"] = "Am J Transl Res",
  ["american journal of transplantation"] = "Am. J. Transplant.",
  ["american journal of transplantation : official journal of the american society of transplantation and the american society of transplant surgeons"] = "Am J Transplant",
  ["american journal of tropical medicine and hygiene"] = "Am. J. Trop. Med. Hyg.",
  ["american journal of undergraduate research"] = "Am J Undergrad Res",
  ["american journal of urology research"] = "Am J Urol Res",
  ["american journal of veterinary medicine"] = "Am J Vet Med",
  ["american journal of veterinary research"] = "Am J Vet Res",
  ["american journal of virology"] = "Am J Virol",
  ["american journal on addictions"] = "Am. J. Addict.",
  ["american journal on intellectual and developmental disabilities"] = "Am J Intellect Dev Disabil",
  ["american laboratory"] = "Am. Lab.",
  ["american laundry digest"] = "Am Laund Dig",
  ["american law and economics review"] = "Amer. Law Econ. Rev.",
  ["american law reports. alr 3d. cases and annotations"] = "Am Law Rep ALR 3rd Cases Annot",
  ["american libraries"] = "Am Libr",
  ["american literary history"] = "Am lit Hist",
  ["american literature: a journal of literary history, criticism and bibliography"] = "Am Lit",
  ["american literature; a journal of literary history, criticism and bibliography"] = "Am Lit",
  ["american lung association bulletin"] = "Am Lung Assoc Bull",
  ["american malacological bulletin"] = "Am Malacol Bull",
  ["american mathematical monthly"] = "Am. Math. Mon.",
  ["american mathematical society"] = "Bull. Amer. Math. Soc. (N.S.)",
  ["american mathematical society colloquium publications"] = "Amer. Math. Soc. Colloq. Publ.",
  ["american mathematical society translations, series 2"] = "Amer. Math. Soc. Transl. Ser. 2",
  ["american mathematical society. bulletin. new series"] = "Bull. Amer. Math. Soc. (N.S.)",
  ["american medical journal"] = "Am Med J",
  ["american medical news"] = "Am Med News",
  ["american medical student research journal"] = "Am Med Stud Res J",
  ["american midland naturalist"] = "Am. Midl. Nat.",
  ["american mineralogist"] = "Am. Mineral.",
  ["american museum novitates"] = "Am Mus Novit",
  ["american naturalist"] = "Am. Nat.",
  ["american numismatic society. numismatic notes and monographs"] = "NNM",
  ["american nurse"] = "Am. Nurse",
  ["american nurse today"] = "Am Nurse Today",
  ["american nurses association publications"] = "ANA Publ",
  ["american optical vision"] = "Am Opt Vis",
  ["american orthoptic journal"] = "Am. Orthopt. J.",
  ["american ose review"] = "Am OSE Rev",
  ["american pharmaceutical review"] = "Am Pharm Rev",
  ["american pharmacy"] = "Am Pharm",
  ["american philosophical quarterly"] = "Am Philos Q",
  ["american philosophical quarterly monograph series"] = "Amer. Philos. Quart. Monogr. Ser.",
  ["american political science review"] = "Amer. Polit. Sci. Rev.",
  ["american politics quarterly"] = "Am Polit Q",
  ["american potato journal"] = "Am Potato J",
  ["american practitioner and digest of treatment"] = "Am Pract Dig Treat",
  ["american presbyterian"] = "Am Presbyt",
  ["american professional pharmacist"] = "Am Prof Pharm",
  ["american prospect"] = "Amer. Prospect",
  ["american psychologist"] = "Am. Psychol.",
  ["american quarterly"] = "Am Q",
  ["american real estate and urban economics association journal"] = "Amer. Real Estate Urban Econ. Assoc. J.",
  ["american rehabilitation"] = "Am Rehabil",
  ["american review of public administration"] = "Am Rev Public Adm",
  ["american review of respiratory disease"] = "Am. Rev. Respir. Dis.",
  ["american review of soviet medicine"] = "Am Rev Sov Med",
  ["american review of tuberculosis"] = "Am Rev Tuberc",
  ["american scientist"] = "Am Sci",
  ["american society of clinical oncology educational book. american society of clinical oncology. annual meeting"] = "Am Soc Clin Oncol Educ Book",
  ["american society of mechanical engineers"] = "Trans. ASME J. Appl. Mech.",
  ["american sociological review"] = "Am Sociol Rev",
  ["american speech"] = "Am Speech",
  ["american statistician"] = "Am. Stat.",
  ["american studies (lawrence, kan.)"] = "Am Stud (Lawrence)",
  ["american studies in papyrology"] = "AmStP",
  ["american studies international"] = "Am Stud Int",
  ["american surgeon"] = "Am. Surg.",
  ["american universities field staff reports: africa"] = "Am Univ Field Staff Rep Afr",
  ["american universities field staff reports: asia"] = "Am Univ Field Staff Rep Asia",
  ["american universities field staff reports: north america"] = "Am Univ Field Staff Rep North Am",
  ["american universities field staff reports: south america"] = "Am Univ Field Staff Rep South Am",
  ["american university law review"] = "Am. Univ. Law Rev.",
  ["american university studies"] = "Am. Univ. Stud. Ser. IX Hist.",
  ["american zoologist"] = "Am. Zool.",
  ["american-eurasian journal of agricultural & environmental sciences"] = "Am Eurasian J Agric Environ Sci",
  ["americas (academy of american franciscan history)"] = "Americas (Acad Am Francisc Hist)",
  ["américa latina"] = "Am Lat",
  ["américa latina en la historia económica : boletín de fuentes"] = "Am Lat Hist Econ",
  ["amfar report"] = "AMFAR Rep",
  ["amfar report / american foundation for aids research"] = "AMFAR Rep",
  ["amhc forum"] = "AMHC Forum",
  ["amia annual symposium proceedings"] = "AMIA. Annu. Symp. Proc.",
  ["amia joint summits on translational science proceedings. amia joint summits on translational science"] = "AMIA Jt Summits Transl Sci Proc",
  ["amici molluscarum"] = "Amici Molluscarum",
  ["amino acids"] = "Amino Acids",
  ["ammonia plant safety and related facilities"] = "Ammonia Plant Saf. Relat. Facil",
  ["ampersand (oxford, uk)"] = "Ampersand (Oxford)",
  ["amphibia-reptilia : publication of the societas europaea herpetologica"] = "Amphib Reptil",
  ["amphibian & reptile conservation"] = "Amphib Reptile Conserv",
  ["amrc open research"] = "AMRC Open Res",
  ["amrl-tr. aerospace medical research laboratories (6570th)"] = "AMRL TR",
  ["amrl-tr. aerospace medical research laboratories (u.s.)"] = "AMRL TR",
  ["ams progress in mathematics lecture series"] = "AMS Progr. Math. Lecture Ser.",
  ["ams short course lecture notes"] = "AMS Short Course Lecture Notes",
  ["ams-maa joint lecture series"] = "AMS-MAA Joint Lecture Ser.",
  ["ams/ip studies in advanced mathematics"] = "AMS/IP Stud. Adv. Math.",
  ["ams/maa problem books series"] = "AMS/MAA Probl. Books Ser.",
  ["ams/maa spectrum"] = "AMS/MAA Spectr.",
  ["ams/maa textbooks"] = "AMS/MAA Textb.",
  ["amsterdam studies in the theory and history of linguistic science"] = "Amst. Stud. Theory Hist. Linguist. Sci. Ser. IV Curr. Issues Linguist. Theory",
  ["amsterdam studies in the theory and history of linguistic science. series iii: studies in the history of the language sciences"] = "Amst. Stud. Theory Hist. Linguist. Sci. Ser. III Stud. Hist. Lang. Sci.",
  ["amyloid : the international journal of experimental and clinical investigation : the official journal of the international society of amyloidosis"] = "Amyloid",
  ["amyotrophic lateral sclerosis"] = "Amyotroph. Lateral Scler.",
  ["amyotrophic lateral sclerosis & frontotemporal degeneration"] = "Amyotroph Lateral Scler Frontotemporal Degener",
  ["amyotrophic lateral sclerosis : official publication of the world federation of neurology research group on motor neuron diseases"] = "Amyotroph Lateral Scler",
  ["amyotrophic lateral sclerosis and other motor neuron disorders"] = "Amyotroph. Lateral Scler. Other Motor Neuron Disord.",
  ["amyotrophic lateral sclerosis and other motor neuron disorders : official publication of the world federation of neurology, research group on motor neuron diseases"] = "Amyotroph Lateral Scler Other Motor Neuron Disord",
  ["aménagement du territoire"] = "Aménage. territ.",
  ["an advanced course in nuclear engineering"] = "Adv. Course Nucl. Eng.",
  ["an hui ji shu shi fan xue yuan xue bao"] = "An Hui Ji Shu Shi Fan Xue Yuan Xue Bao",
  ["an ieee/oup classic reissue"] = "IEEE/OUP Class. Reissue",
  ["an international journal of optimization and control. theories & applications. ijocta"] = "Int. J. Optim. Control. Theor. Appl. IJOCTA",
  ["an international journal on information fusion"] = "Inf Fusion",
  ["an. arkansas nurse"] = "AN Ark Nurse",
  ["an: arkansas nurse"] = "AN Ark. Nurse",
  ["ana clinical conferences"] = "ANA Clin Conf",
  ["ana clinical sessions"] = "ANA Clin Sess",
  ["anadolu araştırmaları. jahrbuch für kleinasiatische forschung"] = "AnadoluAra",
  ["anadolu kardiyoloji dergisi"] = "Anadolu Kardiyol. Derg.",
  ["anadolu kardiyoloji dergisi : akd = the anatolian journal of cardiology"] = "Anadolu Kardiyol Derg",
  ["anadolu kliniği"] = "Anadolu Klinigi",
  ["anadolu medeniyetleri müzesi konferansları"] = "AnadoluKonf",
  ["anadolu medeniyetleri müzesi yıllığı"] = "AnadoluYil",
  ["anadolu psikiyatri dergisi"] = "Anadolu Psikiyatri Derg",
  ["anaesthesia & surgery open access journal"] = "Anaesth Surg Open Access J",
  ["anaesthesia and anaesthetics"] = "Anaesth Anaesth",
  ["anaesthesia and intensive care"] = "Anaesth Intensive Care",
  ["anaesthesia cases"] = "Anaesth Cases",
  ["anaesthesia reports"] = "Anaesth Rep",
  ["anaesthesia, critical care & pain medicine"] = "Anaesth Crit Care Pain Med",
  ["anaesthesia, resuscitation and intensive therapy"] = "Anaesth. Resusc. Intensive Ther.",
  ["anaesthesia, resuscitation, and intensive therapy"] = "Anaesth Resusc Intensive Ther",
  ["anaesthesiologie und reanimation"] = "Anaesthesiol Reanim",
  ["anaesthesiology intensive therapy"] = "Anaesthesiol Intensive Ther",
  ["anagennesis. a papyrological journal"] = "Anagennesis",
  ["anais azevedos"] = "An  Azevedos",
  ["anais brasileiros de dermatologia"] = "An Bras Dermatol",
  ["anais brasileiros de dermatologia e sifilografia"] = "An Bras Derm Sifilogr",
  ["anais brasileiros de ginecologia"] = "An Bras Ginecol",
  ["anais da academia brasileira de ciencias"] = "An Acad Bras Cienc",
  ["anais da academia brasileira de ciencias. annals of the brazilian academy of sciences"] = "An. Acad. Brasil. Cienc.",
  ["anais da academia brasileira de ciências"] = "An Acad Bras Cienc",
  ["anais da academia brasileira de ciências"] = "An. Acad. Brasil. Ciênc.",
  ["anais da escola nacional de saude publica e de medicina tropical"] = "An. Esc. Nacl. Saude Publica Med. Trop. (Lisb.)",
  ["anais da escola nacional de saúde pública e de medicina tropical"] = "An Esc Nacl Saude Publica Med Trop (Lisb)",
  ["anais da faculdade de farmacia do porto"] = "An Fac Farm Porto",
  ["anais da faculdade de farmácia do porto"] = "An Fac Farm Porto",
  ["anais da faculdade de medicina de porto alegre"] = "An Fac Med Porto Alegre",
  ["anais da faculdade de medicina de pôrto alegre"] = "An Fac Med Porto Alegre",
  ["anais da faculdade de medicina do recife"] = "An Fac Med Recife",
  ["anais da faculdade de odontologia da universidade federal de pernambuco"] = "An Fac Odontol Univ Fed Pernambuco",
  ["anais de enfermagem"] = "An Enferm",
  ["anais de historia : publicacao do departamento de historia da faculdade de filosofia, ciencias e letras de assis"] = "An Hist",
  ["anais de história : publicação do departamento de história da faculdade de filosofia, ciências e letras de assis"] = "An Hist",
  ["anais de história de além-mar"] = "Anais Hist Alem Mar",
  ["anais de microbiologia"] = "An Microbiol (Rio J)",
  ["anais do instituto de higiene e medicina tropical"] = "An Inst Hig Med Trop (Lisb)",
  ["anais do instituto de medicina tropical"] = "An Inst Med Trop (Lisb)",
  ["anais do municípo de faro"] = "AnMunFaro",
  ["anais nestle"] = "An Nestle (Rio De Janeiro)",
  ["anais nestlé"] = "An Nestle (Rio J)",
  ["anais paulistas de medicina e cirurgia"] = "An Paul Med Cir",
  ["anais. sao paulo, brazil (city) maternidade"] = "An Matern Sao Paulo",
  ["anais. são paulo (brazil). maternidade"] = "An Matern Sao Paulo",
  ["anais. universidade de sao paulo. faculdade de medicina"] = "An Fac Med Univ Sao Paulo",
  ["anais. universidade de são paulo. faculdade de medicina"] = "An Fac Med Univ Sao Paulo",
  ["anal. chem. indian journal of chemistry, section a: inorganic, bio-inorganic, physical, theoretical & analytical chemistry"] = "Indian J. Chem., Sect. A: Inorg., Bio-inorg., Phys., Theor.",
  ["analecta aegyptiaca"] = "AnAe",
  ["analecta bollandiana"] = "AB",
  ["analecta calasanctiana"] = "Analecta Calas",
  ["analecta malacitana: revista de la sección de filología de la facultad de filosofía y letras"] = "AMal",
  ["analecta medico-historica"] = "Analecta Med Hist",
  ["analecta médica"] = "Analecta Med",
  ["analecta orientalia. commentationes scientificae de rebus orientis antiqui"] = "AnOr",
  ["analecta papyrologica"] = "AnalP",
  ["analecta romana instituti danic"] = "ARID",
  ["analecta romana instituti danici"] = "AnalRom",
  ["analecta sacra tarraconensia; revista de ciencias historico-eclesiasticas"] = "Analecta Sacra Tarracon",
  ["analecta sacra tarraconensia; revista de ciencias histórico-eclesiásticas"] = "Analecta Sacra Tarracon",
  ["analecta veterinaria"] = "Analecta Vet",
  ["analectas farmacéutico-gerundenses"] = "Analectas Farm Gerundenses",
  ["analele româno-sovietice. seria medicină generală"] = "An Rom Sov Ser Med Gen",
  ["analele stiintifice ale universitatii “al. i. cuza” din iasi. serie noua. matematica. annals of the alexandru cuza university. mathematics"] = "An. Stiint. Univ. Al. I. Cuza Iasi. Mat. (N.S.)",
  ["analele stiintifice ale universitatii “ovidius” constanta. seria matematica. mathematical journal of the ovidius university of constantza"] = "An. Stiint. Univ. “Ovidius” Constanta Ser. Mat.",
  ["analele universitatii bucurersti. chimie"] = "Ann. Univ. Buc. Chim.",
  ["analele universitatii bucuresti. informatica"] = "An. Univ. Bucur. Inform.",
  ["analele universitatii bucuresti: istorie. universitatea din bucuresti"] = "An Univ Bucur Istor",
  ["analele universitatii din craiova. seria matematica-informatica. annals of the university of craiova. mathematics and computer science series"] = "An. Univ. Craiova Ser. Mat. Inform.",
  ["analele universitatii din oradea. fascicola matematica"] = "An. Univ. Oradea Fasc. Mat.",
  ["analele universitatii spiru haret. seria matematica-informatica. annals of spiru haret university. mathematics-informatics series"] = "An. Univ. Spiru Haret. Ser. Mat.-Inform.",
  ["analele universitatii “dunarea de jos” din galati. fascicula ii. matematica, fizica, mecanica teoretica. annals of the “dunarea de jos” university of galati. fascicle ii. mathematics, physics, theoretical mechanics"] = "An. Univ. Dunarea de Jos Galati Fasc. II Mat. Fiz. Mec. Teor.",
  ["analele universității bucurești: istorie. universitatea din bucurești"] = "An Univ Bucur Istor",
  ["analele universităţii bucurești"] = "An. Univ. București Mat.",
  ["analele universităţii din craiova"] = "An. Univ. Craiova Ser. Mat. Inform.",
  ["analele universităţii din galaţi"] = "An. Univ. Galaţi Fasc. II Mat. Fiz. Mec. Teoret.",
  ["analele universităţii din timișoara"] = "An. Univ. Timișoara Ser. Ştiinţ. Fiz.",
  ["analele universităţii din târgoviște"] = "An. Univ. Târgoviște Fasc. I",
  ["analele. universitatii de vest din timisoara. seria matematica-informatica. annals of west university of timisoara. mathematics and computer science"] = "An. Univ. Vest Timis. Ser. Mat.-Inform.",
  ["anales (ilustre colegio oficial de medicos (lerida, spain))"] = "An Ilustre Col Of Med",
  ["anales (ilustre colegio oficial de médicos (lerida, spain))"] = "An Ilustre Col Of Med",
  ["anales - instituto de investigaciones odontologicas, universidad del zulia"] = "An Inst Invest Odontol (Maracaibo)",
  ["anales - instituto de investigaciones odontológicas, universidad del zulia"] = "An Inst Invest Odontol (Maracaibo)",
  ["anales argentinos de oftalmología"] = "An Argent Oftalmol",
  ["anales chilenos de historia de la medicina"] = "An Chil Hist Med",
  ["anales de antropologia"] = "An Antropol",
  ["anales de antropología"] = "An Antropol",
  ["anales de arqueología cordobesa"] = "AnCord",
  ["anales de biología"] = "An Biol",
  ["anales de bromatologia"] = "An. Bromatol",
  ["anales de cirugia"] = "An. Cir. (Rosario)",
  ["anales de cirugía"] = "An Cir (Rosario)",
  ["anales de farmacia hospitalaria"] = "An Farm Hosp (Madr)",
  ["anales de filologia clásica"] = "AFC",
  ["anales de física"] = "An. Fís. Monogr.",
  ["anales de historia antigua y medieval"] = "An Hist Antigua Mediev",
  ["anales de historia contemporanea"] = "An Hist Contemp",
  ["anales de historia contemporánea"] = "An Hist Contemp",
  ["anales de la academia de geografia e historia de guatemala. academia de geografia e historia de guatemala"] = "An Acad Geogr Hist Guatem",
  ["anales de la academia de geografía e historia de guatemala. academia de geografía e historia de guatemala"] = "An Acad Geogr Hist Guatem",
  ["anales de la asociacion fisica argentina"] = "An. Asoc. Fis. Argent.",
  ["anales de la asociacion quimica argentina"] = "An. Asoc. Quim. Argent.",
  ["anales de la casa de salud valdecilla"] = "An Casa Salud Valdecilla",
  ["anales de la cátedra de patología y clínica de la tuberculosis"] = "An Catedra Patol Clin Tuberc",
  ["anales de la facultad de medicina de montevideo"] = "An Fac Med Montev",
  ["anales de la facultad de medicina, universidad de la republica, montevideo, uruguay"] = "An. Fac. Med. Univ. Repub. Montev. Urug.",
  ["anales de la facultad de medicina, universidad de la república, montevideo, uruguay"] = "An Fac Med Univ Repub Montev Urug",
  ["anales de la facultad de medicina, universidad nacional mayor de san marcos de lima"] = "An. Fac. Med. Lima",
  ["anales de la facultad de odontologia"] = "An. Fac. Odontol.",
  ["anales de la facultad de odontología"] = "An Fac Odontol",
  ["anales de la facultad de quimica y farmacia"] = "An Fac Quim Farm (Santiago)",
  ["anales de la facultad de química y farmacia"] = "An Fac Quim Farm (Santiago)",
  ["anales de la real academia de farmacia"] = "An Real Acad Farm",
  ["anales de la real academia de farmacia; instituto de espana"] = "An. Real Acad. Farm.",
  ["anales de la real academia de medicina de zaragoza. sesión inaugural"] = "An R Acad Med Zaragoza",
  ["anales de la real academia nacional de medicina"] = "An R Acad Nac Med (Madr)",
  ["anales de la sociedad de geografia e historia. sociedad de geografia e historia de guatemala"] = "An Soc Geogr Hist",
  ["anales de la sociedad de geografía e historia. sociedad de geografía e historia de guatemala"] = "An Soc Geogr Hist",
  ["anales de la sociedad de puericultura de buenos aires"] = "An Soc Pueric B Aires",
  ["anales de la sociedad mexicana de historia de la ciencia y de la tecnologia"] = "An Soc Mex Hist Cienc Tec",
  ["anales de la sociedad mexicana de historia de la ciencia y de la tecnología"] = "An Soc Mex Hist Cienc Tec",
  ["anales de la universidad de murcia, facultad de letras"] = "AUM",
  ["anales de medicina"] = "An Med (Barc)",
  ["anales de medicina (lima, peru)"] = "An Med (Lima)",
  ["anales de medicina interna"] = "An. Med. Interna",
  ["anales de medicina interna (madrid, spain : 1984)"] = "An Med Interna",
  ["anales de medicina publica"] = "An Med Publica",
  ["anales de medicina pública"] = "An Med Publica",
  ["anales de medicina y cirugia"] = "An Med Cir",
  ["anales de medicina y cirugía"] = "An Med Cir",
  ["anales de medicina. ateneo ramón y cajal, mexico"] = "An Med Ateneo Ramon Cajal Mex",
  ["anales de medicina: cirugia"] = "An. Med. Cir.",
  ["anales de medicina: cirugía"] = "An Med Cir",
  ["anales de medicina: especialidades"] = "An Med Espec",
  ["anales de medicina: medicina"] = "An Med Med",
  ["anales de pediatria"] = "An. Pediatr.",
  ["anales de pediatria (barcelona, spain : 2003)"] = "An Pediatr (Barc)",
  ["anales de pediatría"] = "An Pediatr (Engl Ed)",
  ["anales de pediatría (barcelona, spain : 2003)"] = "An Pediatr (Barc)",
  ["anales de prehistoria y arqueología. universidad de murcia"] = "AnMurcia",
  ["anales de quimica"] = "An. Quim.",
  ["anales del hospital de la santa cruz y san pablo"] = "An Hosp St Cruz San Pablo",
  ["anales del instituto barraquer"] = "An Inst Barraquer",
  ["anales del instituto de biología"] = "An Inst Bio",
  ["anales del instituto de estudios madrilenos. instituto de estudios madrilenos (consejo superior de investigaciones cientificas)"] = "An Inst Estud Madr",
  ["anales del instituto de estudios madrileños. instituto de estudios madrileños (consejo superior de investigaciones científicas)"] = "An Inst Estud Madr",
  ["anales del instituto de investigaciones veterinarias"] = "An Inst Invest Vet",
  ["anales del instituto de matemáticas"] = "An. Inst. Mat. Univ. Nac. Autónoma México",
  ["anales del instituto nacional de antropologia e historia"] = "An Inst Nac Antropol Hist",
  ["anales del jardín botánico de madrid (1979)"] = "An Jard Bot Madr (1979)",
  ["anales del programa academico de medicina"] = "An. Programa Acad. Med. (Lima)",
  ["anales del programa academico de medicina / universidad nacional mayor de san marcos de lima"] = "An Programa Acad Med (Lima)",
  ["anales del programa académico de medicina"] = "An Programa Acad Med (Lima)",
  ["anales del sistema sanitario de navarra"] = "An Sist Sanit Navar",
  ["anales espanoles de odontoestomatologia"] = "An. Esp. Odontoestomatol.",
  ["anales espanoles de pediatria"] = "An Esp Pediatr",
  ["anales españoles de odontoestomatología"] = "An Esp Odontoestomatol",
  ["anales españoles de pediatría"] = "An Esp Pediatr",
  ["anales medicos de concepcion"] = "An Med Concepc",
  ["anales médicos de concepción"] = "An Med Concepc",
  ["anales nestle"] = "An Nestle (B Aires)",
  ["anales nestlé"] = "An Nestle (B Aires)",
  ["anales otorrinolaringologicos ibero-americanos"] = "An Otorrinolaringol Ibero Am",
  ["anales otorrinolaringologicos iberoamericanos"] = "An. Otorrinolaringol. Ibero. Am.",
  ["anales otorrinolaringológicos ibero-americanos"] = "An Otorrinolaringol Ibero Am",
  ["anales, instituto de investigaciones odontologicas, universidad del zulia"] = "An. Inst. Invest. Odontol. (Maracaibo)",
  ["anales. academia de medicina, medellín"] = "An Academ Med Medellin",
  ["anales. asociación de química y farmacia del uruguay"] = "An Asoc Quim Farm Uruguay",
  ["anales. ateneo de clínica quirúrgica, montevideo"] = "An Ateneo Clin Quir Montev",
  ["anales. escuela nacional de ciencias biológicas (mexico)"] = "An Esc Nac Cienc Biol",
  ["anales. hospital torcuato de alvear (buenos aires, argentina). ateneo de patologia y clínica médica"] = "An Hosp Alvear B Aires",
  ["anales. hospital varsovia"] = "An Hosp Varsovia",
  ["anales. nuevo león, mexico. universidad, monterey. instituto de investigaciones científicas"] = "An Nuevo Leon Mex",
  ["anales. real academia de ciencias médicas, físicas y naturales de la habana"] = "An Acad Cienc Medicas Fisicas Nat Habana",
  ["anales. sociedad de biologia de bogotá"] = "Anal Soc Biol Bogota",
  ["anales. sociedad mexicana de oftalmología y oto-rino-laringología"] = "An Soc Mex Oftalmol Otorinolaringol",
  ["anales. sociedad médico-quirúrgica del guayas"] = "An Soc Med Quir Guayas",
  ["anales. universidad de santo domingo"] = "An Univ St. Domingo",
  ["anales. universidad nacional de asunción. facultad de ciencias médicas"] = "An Univ Nac Asuncion",
  ["anales. universidad nacional mayor de san marcos. facultad de farmacia y bioquimica"] = "An Lima",
  ["anales. universidad nacional mayor de san marcos. facultad de medicina"] = "An Fac Med Lima",
  ["analgesia & resuscitation : current research"] = "Analg Resusc",
  ["analgesia (elmsford, n.y.)"] = "Analgesia (Elmsford N Y)",
  ["anali gazi husrev-begove biblioteke. gazi husrevbegova biblioteka u sarajevu"] = "An Gazi Husrev Begove Bibl",
  ["analise social"] = "Anal Soc",
  ["analisis politico (bogota, colombia)"] = "Anal Polit",
  ["analiză modernă şi aplicaţii"] = "Anal. Modern. Apl.",
  ["analog circuits and signal processing"] = "Analog Circuits Signal Process.",
  ["analog dialogue"] = "Analog Dialogue",
  ["analog integrated circuits and signal processing"] = "Analog Integr Circuits Signal Process",
  ["analog science fiction/science fact"] = "Analog Sci Fict Sci Fact",
  ["analyse prévision"] = "Anal. Prévision",
  ["analyses of social issues and public policy : asap"] = "Anal Soc Issues Public Policy",
  ["analysis & pde"] = "Anal. PDE",
  ["analysis & sensing"] = "Anal. Sens.",
  ["analysis and applications"] = "Anal. Appl.",
  ["analysis and geometry in metric spaces"] = "Anal. Geom. Metr. Spaces",
  ["analysis and its applications"] = "Anal. Appl.",
  ["analysis and mathematical physics"] = "Anal. Math. Phys.",
  ["analysis for pdes set"] = "Anal. PDEs Set",
  ["analysis in theory and applications"] = "Anal. Theory Appl.",
  ["analysis mathematica"] = "Anal. Math.",
  ["analysis of gambling behavior"] = "Anal Gambl Behav",
  ["analysis, geometry and number theory. an international journal"] = "Anal. Geom. Number Theory",
  ["analysis. international mathematical journal of analysis and its applications"] = "Analysis (Berlin)",
  ["analyst (cambridge, united kingdom)"] = "Analyst (Cambridge, U. K.)",
  ["analytic methods in accident research"] = "Anal Methods Accid Res",
  ["analytic methods in accident research."] = "Anal. Methods Accid. Res",
  ["analytic philosophy"] = "Anal. Philos.",
  ["analytica chimica acta"] = "Anal Chim Acta",
  ["analytica chimica acta: x"] = "Anal Chim Acta X",
  ["analytical and bioanalytical chemistry"] = "Anal Bioanal Chem",
  ["analytical and quantitative cytology"] = "Anal Quant Cytol",
  ["analytical and quantitative cytology and histology"] = "Anal Quant Cytol Histol",
  ["analytical and quantitative cytology and histology / the international academy of cytology [and] american society of cytology"] = "Anal Quant Cytol Histol",
  ["analytical and quantitative cytopathology and histopathology"] = "Anal Quant Cytopathol Histpathol",
  ["analytical biochemistry"] = "Anal Biochem",
  ["analytical cellular pathology"] = "Anal. Cell. Pathol.",
  ["analytical cellular pathology (amsterdam)"] = "Anal Cell Pathol (Amst)",
  ["analytical cellular pathology : the journal of the european society for analytical cellular pathology"] = "Anal Cell Pathol",
  ["analytical chemistry"] = "Anal Chem",
  ["analytical chemistry (washington, dc, united states)"] = "Anal. Chem. (Washington, DC, U. S.)",
  ["analytical chemistry insights"] = "Anal Chem Insights",
  ["analytical chemistry letters"] = "Anal. Chem. Lett.",
  ["analytical chemistry research"] = "Anal. Chem. Res.",
  ["analytical chemistry symposia series"] = "Anal. Chem. Symp. Ser.",
  ["analytical communications"] = "Anal. Commun.",
  ["analytical letters"] = "Anal. Lett.",
  ["analytical methods"] = "Anal. Methods",
  ["analytical methods : advancing methods and applications"] = "Anal Methods",
  ["analytical methods and special functions"] = "Anal. Methods Spec. Funct.",
  ["analytical methods for social research"] = "Anal. Methods Soc. Res.",
  ["analytical proceedings"] = "Anal. Proc.",
  ["analytical proceedings including analytical communications"] = "Anal. Proc. incl. Anal. Commun.",
  ["analytical science advances"] = "Anal. Sci. Adv.",
  ["analytical sciences"] = "Anal. Sci.",
  ["analytical sciences : the international journal of the japan society for analytical chemistry"] = "Anal Sci",
  ["analytical sciences: x-ray structure analysis online"] = "Anal. Sci.: X-Ray Struct. Anal. Online",
  ["analytical spectroscopy library"] = "Anal. Spectrosc. Lib.",
  ["analytische psychologie"] = "Anal Psychol",
  ["anas. museo nacional de arte romano de mérida"] = "Anas",
  ["anasthesie, intensivtherapie, notfallmedizin"] = "Anasth. Intensivther. Notf. Med.",
  ["anasthesiologie, intensivmedizin, notfallmedizin, schmerztherapie"] = "Anasthesiol. Intensivmed. Notfallmed. Schmerzther.",
  ["anasthesiologie, intensivmedizin, notfallmedizin, schmerztherapie : ains"] = "Anasthesiol Intensivmed Notfallmed Schmerzther",
  ["anasthesiologie, intensivmedizin, notfallmedizin, schmerztherapieanasthesiol intensivmed notfallme"] = "Schmerzther",
  ["anasthesiologische und intensivmedizinische praxis"] = "Anasthesiol. Intensivmed. Prax.",
  ["anatolia sport research"] = "Anatol Sport Res",
  ["anatolia. revue annuelle de l’institut d’archéologie de l’université d’ankara"] = "Anatolia",
  ["anatolian archaeology. reports on research conducted in turkey"] = "AnatA",
  ["anatolian journal of cardiology"] = "Anatol J Cardiol",
  ["anatolian studies"] = "AnSt",
  ["anatolian studies: journal of the british inst. of archaeology at ankara"] = "AS",
  ["anatomia clinica"] = "Anat Clin",
  ["anatomia e chirurgia"] = "Anat Chir",
  ["anatomia, histologia, embryologia"] = "Anat Histol Embryol",
  ["anatomic pathology"] = "Anat. Pathol.",
  ["anatomic pathology (chicago, ill. : annual)"] = "Anat Pathol",
  ["anatomical record"] = "Anat. Rec.",
  ["anatomical record (hoboken, n.j. : 2007)"] = "Anat Rec (Hoboken)",
  ["anatomical record part a: discoveries in molecular, cellular, and evolutionary biology"] = "Anat. Rec. Part A",
  ["anatomical record part b: the anatomist"] = "Anat. Rec. Part B",
  ["anatomical record part b: the new anatomist"] = "Anat. Rec. Part B",
  ["anatomical record. part a, discoveries in molecular, cellular, and evolutionary biology"] = "Anat. Rec. A. Discov. Mol. Cell. Evol. Biol.",
  ["anatomical record. part b, new anatomist"] = "Anat Rec B New Anat",
  ["anatomical record. supplement"] = "Anat. Rec. Suppl.",
  ["anatomical science  international"] = "Anat. Sci. Int.",
  ["anatomical science international"] = "Anat Sci Int",
  ["anatomical science international / japanese association of anatomists"] = "Anat Sci Int",
  ["anatomical sciences education"] = "Anat Sci Educ",
  ["anatomischer anzeiger"] = "Anat Anz",
  ["anatomy & cell biology"] = "Anat Cell Biol",
  ["anatomy & physiology : current research"] = "Anat Physiol",
  ["anatomy and embryology"] = "Anat Embryol (Berl)",
  ["anatomy research international"] = "Anat Res Int",
  ["análisis político (bogotá, colombia)"] = "Anal Polit",
  ["anästhesie, intensivtherapie, notfallmedizin"] = "Anasth Intensivther Notfallmed",
  ["anästhesiologie, intensivmedizin, notfallmedizin, schmerztherapie : ains"] = "Anasthesiol Intensivmed Notfallmed Schmerzther",
  ["anästhesiologische und intensivmedizinische praxis"] = "Anasthesiol Intensivmed Prax",
  ["ancient biomolecules"] = "Anc Biomol",
  ["ancient civilizations from scythia to siberia. an international journal of comparative studies in history and archaeology"] = "AncCivScytSib",
  ["ancient commentators on aristotle"] = "Anc. Comment. Aristotle",
  ["ancient history"] = "AH",
  ["ancient mediterranean and medieval texts and contexts"] = "Anc. Mediterr. Mediev. Texts Contexts",
  ["ancient near eastern studies. an annual"] = "AncNearEastSt",
  ["ancient philosophy"] = "AncPhil",
  ["ancient philosophy (garland publishing)"] = "Anc Philos",
  ["ancient science"] = "Anc Sci",
  ["ancient science of life"] = "Anc Sci Life",
  ["ancient society"] = "AncSoc",
  ["ancient west and east"] = "AncWestEast",
  ["andean geology"] = "Andean Geol.",
  ["andover newton quarterly"] = "Andover Newton Q",
  ["androgens: clinical research and therapeutics"] = "Androg Clin Res Ther",
  ["andrology : open access"] = "Andrology (Los Angel)",
  ["anestesia e rianimazione"] = "Anest Rianim",
  ["anesteziologiia i reanimatologiia"] = "Anesteziol Reanimatol",
  ["anestezjologia intensywna terapia"] = "Anestezjol Intens Ter",
  ["anesthesia & pain control in dentistry"] = "Anesth Pain Control Dent",
  ["anesthesia and analgesia"] = "Anesth Analg",
  ["anesthesia and pain control in dentistry"] = "Anesth. Pain Control Dent.",
  ["anesthesia and pain medicine"] = "Anesth Pain Med (Seoul)",
  ["anesthesia progress"] = "Anesth Prog",
  ["anesthesia, essays and researches"] = "Anesth Essays Res",
  ["anesthesie et analgesie"] = "Anesth Anal",
  ["anesthesie, analgesie, reanimation"] = "Anesth. Analg. (Paris)",
  ["anesthesiology and pain medicine"] = "Anesth Pain Med",
  ["anesthesiology case reports"] = "Anesthesiol Case Rep",
  ["anesthesiology clinics"] = "Anesthesiol Clin",
  ["anesthesiology clinics of north america"] = "Anesthesiol Clin North Am",
  ["anesthesiology research and practice"] = "Anesthesiol Res Pract",
  ["anesthesiology review"] = "Anesthesiol Rev",
  ["anesthésie et analgésie"] = "Anesth Anal",
  ["anesthésie, analgésie, réanimation"] = "Anesth Analg (Paris)",
  ["angeiologie et annales de la societe francaise d’angeiologie et d’histopathologie"] = "Angeiol Ann Soc Fr Angeiol Histopathol",
  ["angelaki : journal of theoretical humanities"] = "Angelaki",
  ["angewandte carabidologie"] = "Angew Carabidol",
  ["angewandte chemie"] = "Angew. Chem.",
  ["angewandte chemie (international ed. in english)"] = "Angew Chem Int Ed Engl",
  ["angewandte chemie (weinheim an der bergstrasse, germany)"] = "Angew Chem Weinheim Bergstr Ger",
  ["angewandte chemie international edition"] = "Angew. Chem. Int. Ed.",
  ["angewandte chemie, international edition"] = "Angew. Chem., Int. Ed.",
  ["angewandte chemie, international edition in english"] = "Angew. Chem., Int. Ed. Engl.",
  ["angewandte chemie. international edition in english"] = "Angew. Chem. Int. Ed. Engl.",
  ["angewandte entomologie"] = "Angew. Entomol.",
  ["angewandte landschaftsökologie"] = "Angew. Landsch.ökol.",
  ["angewandte makromolekulare chemie"] = "Angew. Makromol. Chem.",
  ["angewandte mathematik"] = "Angew. Math.",
  ["angewandte parasitologie"] = "Angew Parasitol",
  ["angewandte pflanzensoziologie"] = "Angew. Pflanzensoziol.",
  ["angewandte statistik und ökonometrie"] = "Angew. Statist. Ökonom.",
  ["angéiologie et annales de la société française d'angéiologie et d'histopathologie"] = "Angeiol Ann Soc Fr Angeiol Histopathol",
  ["angiologii︠a︡ i sosudistai︠a︡ khirurgii︠a︡ = angiology and vascular surgery"] = "Angiol Sosud Khir",
  ["angiology: open access"] = "Angiol Open Access",
  ["angle orthodontist"] = "Angle Orthod.",
  ["anglia (halle an der saale, germany)"] = "Anglia",
  ["anglican and episcopal history"] = "Angl Episcop Hist",
  ["anglican theological review"] = "Anglican Theol Rev",
  ["anglo-german medical review"] = "Anglo. Ger. Med. Rev.",
  ["anglo-german medical review. deutsch-englische medizinische rundschau"] = "Anglo Ger Med Rev",
  ["anglo-saxon england"] = "Anglo Sax Engl",
  ["anhui nong ye da xue xue bao"] = "Anhui Nong Ye Da Xue Xue Bao",
  ["anhui nongxueyuan xuebao"] = "Journal of Anhui Agricultural College|Anhui Nongxueyuan Xuebao",
  ["anhui nongxueyuan xuebao = journal of anhui agricultural college"] = "Anhui Nongxueyuan Xuebao",
  ["anhui nongye kexue"] = "Anhui Nongye Kexue",
  ["anhui shida xuebao"] = "Anhui Shida Xuebao Ziran Kexue Ban",
  ["animal - open space"] = "Anim. Open Space",
  ["animal : an international journal of animal bioscience"] = "Animal",
  ["animal behavior and cognition"] = "Anim Behav Cogn",
  ["animal behaviour"] = "Anim Behav",
  ["animal behaviour monographs"] = "Anim. Behav. Monogr.",
  ["animal biodiversity and conservation"] = "Anim Biodivers Conserv",
  ["animal biology & animal husbandry"] = "Anim Biol Anim Husb",
  ["animal biology (leiden, netherlands)"] = "Anim Biol Leiden Neth",
  ["animal bioscience"] = "Anim Biosci",
  ["animal biotechnology"] = "Anim Biotechnol",
  ["animal biotechnology bulletin"] = "Anim Biotechnol Bull",
  ["animal biotelemetry"] = "Anim. Biotelem.",
  ["animal blood groups and biochemical genetics"] = "Anim Blood Groups Biochem Genet",
  ["animal blood groups and biochemical genetics. supplement"] = "Anim. Blood Groups Biochem. Genet. Suppl.",
  ["animal cells and systems"] = "Anim Cells Syst (Seoul)",
  ["animal cognition"] = "Anim Cogn",
  ["animal conservation"] = "Anim Conserv",
  ["animal diseases"] = "Anim. Dis.",
  ["animal feed science and technology"] = "Anim. Feed Sci. Technol.",
  ["animal frontiers"] = "Anim. Front.",
  ["animal frontiers : the review magazine of animal agriculture"] = "Anim Front",
  ["animal gene"] = "Anim. Gene",
  ["animal genetic resources = ressources génétiques animales = recursos genéticos animales"] = "Anim Genet Resour",
  ["animal genetics"] = "Anim Genet",
  ["animal health research reviews"] = "Anim Health Res Rev",
  ["animal health research reviews / conference of research workers in animal diseases"] = "Anim Health Res Rev",
  ["animal learning & behavior"] = "Anim Learn Behav",
  ["animal learning and behavior"] = "Anim. Learn. Behav.",
  ["animal microbiome"] = "Anim Microbiome",
  ["animal models and experimental medicine"] = "Animal Model Exp Med",
  ["animal nutrition"] = "Anim. Nutr.",
  ["animal nutrition (zhongguo xu mu shou yi xue hui)"] = "Anim Nutr",
  ["animal nutrition and feed technology"] = "Anim. Nutr. Feed Technol.",
  ["animal production"] = "Anim Prod",
  ["animal production science"] = "Anim Prod Sci",
  ["animal reproduction"] = "Anim Reprod",
  ["animal reproduction science"] = "Anim Reprod Sci",
  ["animal science"] = "Anim. Sci.",
  ["animal science (penicuik, scotland)"] = "Anim Sci",
  ["animal science and technology"] = "Anim. Sci. Technol.",
  ["animal science cases"] = "Anim. Sci. Cases",
  ["animal science journal"] = "Anim. Sci. J.",
  ["animal science journal = nihon chikusan gakkaihō"] = "Anim Sci J",
  ["animal science papers and reports"] = "Anim Sci Pap Rep",
  ["animal systematics, evolution and diversity"] = "Anim Syst Evol Diversity",
  ["animal taxonomy and ecology"] = "Anim. Taxon. Ecol.",
  ["animal technology : journal of the institute of animal technicians"] = "Anim Technol",
  ["animal welfare"] = "Anim. Welfare",
  ["animal welfare (south mimms, england)"] = "Anim Welf",
  ["animal welfare information center bulletin"] = "Anim. Welfare Inf. Cent. Bull.",
  ["animal welfare, etológia és tartástechnológia"] = "Anim Welf Etol Tartastechnol",
  ["animals : an open access journal from mdpi"] = "Animals (Basel)",
  ["anjisuan he shengwu ziyuan = amino acids & biotic resources"] = "Anjisuan He Shengwu Ziyuan",
  ["ankara universitesi dis hekimligi fakultesi dergisi (journal of the dental faculty of ankara university)"] = "Ankara Univ. Hekim. Fak. Derg.",
  ["ankara universitesi yilligi. annales de l’universite d’ankara. ankara universitesi"] = "Ann Univ Ankara",
  ["ankara üniversitesi diş hekimliği fakültesi dergisi = the journal of the dental faculty of ankara university"] = "Ankara Univ Hekim Fak Derg",
  ["ankara üniversitesi yilliği. annales de l'université d'ankara. ankara üniversitesi"] = "Ann Univ Ankara",
  ["anl reports"] = "ANL Rep.",
  ["anna journal"] = "ANNA J",
  ["anna journal / american nephrology nurses’ association"] = "ANNA J",
  ["annalen der hydrographie und maritimen meteorologie"] = "Ann. Hydrogr. marit. Meteorol",
  ["annalen der meteorologie"] = "Ann. Meteor.",
  ["annalen der pharmacie"] = "Ann. Pharm.",
  ["annalen der physik"] = "Ann. Phys.",
  ["annalen der physik (8)"] = "Ann. Physik (8)",
  ["annalen der physik (leipzig)"] = "Ann. Phys. (Leipzig)",
  ["annalen der schweizerischen meteorologischen anstalt"] = "Ann. Schweiz. meteorol. Anst.",
  ["annalen des naturhistorischen museums in wien. serie a, für mineralogie und petrographie, geologie und paläontologie, anthropologie und prähistorie"] = "Ann Naturhist Mus Wien Ser A Mineral Petrogr Geol Palaontol Anthropol Prahist",
  ["annalen des naturhistorischen museums in wien. serie b, für botanik und zoologie"] = "Ann Nat Hist Mus Wien Ser B Bot Zool",
  ["annales (paris, france : 1946)"] = "Ann Econ Soc Civilis",
  ["annales academiae medicae stetinensis"] = "Ann Acad Med Stetin",
  ["annales academiae medicae stetinensis. suplement"] = "Ann Acad Med Stetin Supl",
  ["annales academiae scientiarum fennicae. mathematica"] = "Ann. Acad. Sci. Fenn. Math.",
  ["annales academiae scientiarum fennicae. ser. a.5, medica"] = "Ann Acad Sci Fenn A",
  ["annales academiae scientiarum fennicae. series a iv, biologica"] = "Ann. Acad. Sci. Fenn. [Biol.]",
  ["annales academiae scientiarum fennicae. series a v, medica"] = "Ann. Acad. Sci. Fenn. A",
  ["annales academiae scientiarum fennicae. series a. 3, geologica--geographica"] = "Ann Acad Sci Fenn A 3 Geol Geogr",
  ["annales academiæ scientiarum fennicæ"] = "Ann. Acad. Sci. Fenn. Math.",
  ["annales archeologiques arabes syriennes"] = "Ann Archeol Syrie",
  ["annales archéologiques arabes syriennes"] = "Ann Archeol Syrie",
  ["annales belges de medecine militaire. belgisch tijdschrift voor militaire geneeskunde"] = "Ann Belg Med Mil",
  ["annales belges de médecine militaire. belgisch tijdschrift voor militaire geneeskunde"] = "Ann Belg Med Mil",
  ["annales bogorienses"] = "Ann Bogor",
  ["annales botanici fennici"] = "Ann Bot Fenn",
  ["annales chirurgiae et gynaecologiae"] = "Ann Chir Gynaecol",
  ["annales chirurgiae et gynaecologiae fenniae"] = "Ann Chir Gynaecol Fenn",
  ["annales chirurgiae et gynaecologiae fenniae. supplement"] = "Ann. Chir. Gynaecol. Fenn. Suppl.",
  ["annales chirurgiae et gynaecologiae fenniae. supplementum"] = "Ann Chir Gynaecol Fenn Suppl",
  ["annales chirurgiae et gynaecologiae. supplementum"] = "Ann Chir Gynaecol Suppl",
  ["annales cisalpines d'histoire sociale. série i"] = "Ann Cisalp Hist Soc",
  ["annales cisalpines d’histoire sociale. serie i"] = "Ann Cisalp Hist Soc",
  ["annales d endocrinologie"] = "Ann. Endocrinol. (Paris)",
  ["annales d histochimie"] = "Ann. Histochim.",
  ["annales d immunologie"] = "Ann. Immunol. (Paris)",
  ["annales d oculistique"] = "Ann. Ocul. (Paris)",
  ["annales d oto-laryngologie et de chirurgie cervico-faciale"] = "Ann. Otolaryngol. Chir. Cervicofac.",
  ["annales d urologie"] = "Ann. Urol. (Paris)",
  ["annales d'anatomie pathologique"] = "Ann Anat Pathol (Paris)",
  ["annales d'endocrinologie"] = "Ann Endocrinol (Paris)",
  ["annales d'économie et de statistique"] = "Ann Econ Stat (1986)",
  ["annales d'études internationales. annals of international studies"] = "Ann Etud Int",
  ["annales d'histochimie"] = "Ann Histochim",
  ["annales d'hygiène et de médecine coloniales"] = "Ann Hyg Med Colon",
  ["annales d'hygiène publique, industrielle et sociale"] = "Ann Hyg Publique Ind Soc (1923)",
  ["annales d'immunologie"] = "Ann Immunol (Paris)",
  ["annales d'oculistique"] = "Ann Ocul (Paris)",
  ["annales d'oto-laryngologie et de chirurgie cervico faciale : bulletin de la société d'oto-laryngologie des hôpitaux de paris"] = "Ann Otolaryngol Chir Cervicofac",
  ["annales d'urologie"] = "Ann Urol (Paris)",
  ["annales de biochimie clinque du quebec"] = "Ann. Biochim. Clin. Que.",
  ["annales de biologie animale, biochimie, biophysique"] = "Ann Biol Anim Biochim Biophys",
  ["annales de biologie clinique"] = "Ann Biol Clin (Paris)",
  ["annales de biologie clinique du quebec"] = "Ann. Biol. Clin. Que.",
  ["annales de bourgogne"] = "Ann Bourgogne",
  ["annales de bretagne et des pays de l'ouest"] = "Ann Bretagne Payes Ouest",
  ["annales de bretagne et des pays de l’ouest"] = "ABPO",
  ["annales de cardiologie et d angeiologie"] = "Ann Cardiol Angeiol (Paris)",
  ["annales de cardiologie et d'angéiologie"] = "Ann Cardiol Angeiol (Paris)",
  ["annales de cardiologie et d’angeiologie"] = "Ann. Cardiol. Angeiol. (Paris)",
  ["annales de chimie"] = "Ann. Chim.",
  ["annales de chimie (cachan, france)"] = "Ann. Chim. (Cachan, Fr.)",
  ["annales de chimie (paris, france)"] = "Ann. Chim. (Paris, Fr.)",
  ["annales de chimie - science des materiaux"] = "Ann. Chim. - Sci. Mater.",
  ["annales de chimie analytique et revue de chimie analytique reunies"] = "Ann. Chim. Anal. Rev. Chim. Anal. Reunies",
  ["annales de chimie et de physique"] = "Ann. Chim. Phys.",
  ["annales de chimie science des materiaux"] = "Ann. Chim. Sci. Mat.",
  ["annales de chimie--science des matériaux"] = "Ann Chimie Sci Materiaux",
  ["annales de chirurgie"] = "Ann Chir",
  ["annales de chirurgie de la main"] = "Ann. Chir. Main",
  ["annales de chirurgie de la main : organe officiel des societes de chirurgie de la main"] = "Ann Chir Main",
  ["annales de chirurgie de la main : organe officiel des sociétés de chirurgie de la main"] = "Ann Chir Main",
  ["annales de chirurgie de la main et du membre superieur"] = "Ann. Chir. Main Memb. Super.",
  ["annales de chirurgie de la main et du membre supérieur : organe officiel des sociétés de chirurgie de la main = annals of hand and upper limb surgery"] = "Ann Chir Main Memb Super",
  ["annales de chirurgie infantile"] = "Ann Chir Infant",
  ["annales de chirurgie plastique"] = "Ann Chir Plast",
  ["annales de chirurgie plastique et esthetique"] = "Ann Chir Plast Esthet",
  ["annales de chirurgie plastique et esthétique"] = "Ann Chir Plast Esthet",
  ["annales de chirurgie thoracique et cardio-vasculaire"] = "Ann Chir Thorac Cardiovasc",
  ["annales de demographie historique"] = "Ann Demogr Hist (Paris)",
  ["annales de dermatologie et de syphiligraphie"] = "Ann Dermatol Syphiligr (Paris)",
  ["annales de dermatologie et de venereologie"] = "Ann Dermatol Venereol",
  ["annales de dermatologie et de vénéréologie"] = "Ann Dermatol Venereol",
  ["annales de droit"] = "Annal Droit",
  ["annales de droit international medical"] = "Ann. Droit Int. Med.",
  ["annales de droit international médical"] = "Ann Droit Int Med",
  ["annales de démographie historique"] = "Ann Demogr Hist (Paris)",
  ["annales de gastroenterologie et d’hepatologie"] = "Ann. Gastroenterol. Hepatol. (Paris)",
  ["annales de gastroentérologie et d'hépatologie"] = "Ann Gastroenterol Hepatol (Paris)",
  ["annales de genetique"] = "Ann Genet",
  ["annales de genetique et de selection animale"] = "Ann. Genet. Sel. Anim.",
  ["annales de geographie"] = "Ann Georgr",
  ["annales de geophysique"] = "Ann. Geophys.",
  ["annales de génétique"] = "Ann Genet",
  ["annales de génétique et de sélection animale"] = "Ann Genet Sel Anim",
  ["annales de géographie"] = "Ann Georgr",
  ["annales de l institut pasteur"] = "Ann. Inst. Pasteur (Paris)",
  ["annales de l institut pasteur de lille"] = "Ann. Inst. Pasteur Lille",
  ["annales de l institut pasteur-virology"] = "Ann. Inst. Pasteur Vir.",
  ["annales de l institut pasteur. immunologie"] = "Ann. Inst. Pasteur. Immunol.",
  ["annales de l institut pasteur. immunology"] = "Ann. Inst. Pasteur. Immunol.",
  ["annales de l institut pasteur. microbiologie"] = "Ann. Inst. Pasteur. Microbiol.",
  ["annales de l institut pasteur. microbiology"] = "Ann. Inst. Pasteur. Microbiol.",
  ["annales de l institut pasteur. virology"] = "Ann. Inst. Pasteur. Virol.",
  ["annales de l'anesthésiologie française"] = "Ann Anesthesiol Fr",
  ["annales de l'est"] = "Ann Est",
  ["annales de l'i.h.p. probabilités et statistiques"] = "Ann I H P Probab Stat",
  ["annales de l'insee"] = "Ann Insee",
  ["annales de l'institut archéologique du luxembourg. institut archéologique du luxembourg (arlon, belgium)"] = "Ann Inst Archeol Luxemb",
  ["annales de l'institut d'hydrologie et de climatologie"] = "Ann Inst Hydrol Climatol",
  ["annales de l'institut pasteur"] = "Ann Inst Pasteur (Paris)",
  ["annales de l'institut pasteur de lille"] = "Ann Inst Pasteur Lille",
  ["annales de l'institut pasteur. actualités"] = "Ann Inst Pasteur Actual",
  ["annales de l'institut pasteur. immunologie"] = "Ann Inst Pasteur Immunol (1985)",
  ["annales de l'institut pasteur. immunology"] = "Ann Inst Pasteur Immunol",
  ["annales de l'institut pasteur. microbiologie"] = "Ann Inst Pasteur Microbiol (1985)",
  ["annales de l'institut pasteur. microbiology"] = "Ann Inst Pasteur Microbiol",
  ["annales de l'institut pasteur. virology"] = "Ann Inst Pasteur Virol",
  ["annales de l'université d'abidjan. série g: géographie. université d'abidjan"] = "Ann Univ Abidjan Ser G",
  ["annales de la faculte des lettres et sciences humaines (dakar, senegal)"] = "Ann Fac Lett Sci Hum",
  ["annales de la faculte des sciences de toulouse. mathematiques. serie 6"] = "Ann. Fac. Sci. Toulouse Math. (6)",
  ["annales de la faculté des lettres et sciences humaines (dakar, senegal)"] = "Ann Fac Lett Sci Hum",
  ["annales de la faculté des lettres et sciences humaines de l’université de dakar"] = "AFLD",
  ["annales de la faculté des sciences"] = "Ann. Fac. Sci. Univ. Kinshasa",
  ["annales de la nutrition et de l alimentation"] = "Ann. Nutr. Aliment.",
  ["annales de la nutrition et de l'alimentation"] = "Ann Nutr Aliment",
  ["annales de la nutrition et de l’alimentation"] = "Ann Nutr Aliment",
  ["annales de la societe belge de medecine tropicale"] = "Ann. Soc. Belg. Med. Trop.",
  ["annales de la societe belge d’histoire des hopitaux"] = "Ann Soc Belg Hist Hop",
  ["annales de la societe d’emulation de bruges"] = "Ann soc Emul Bruges",
  ["annales de la societe entomologique de france"] = "Ann. Soc. Entomol. France",
  ["annales de la societe entomologique de france. societe entomologique de france"] = "Ann Soc Entomol Fr",
  ["annales de la societe royale des sciences medicales et naturelles de bruxelles"] = "Ann. Soc. R. Sci. Med. Nat. Brux.",
  ["annales de la societe royale d’archeologie de bruxelles. societe royale d’archeologie de bruxelles (belgium)"] = "Ann Soc R Archeol Brux",
  ["annales de la société belge d'histoire des hôpitaux"] = "Ann Soc Belg Hist Hop",
  ["annales de la société belge de médecine tropicale"] = "Ann Soc Belg Med Trop",
  ["annales de la société belge de médecine tropicale (1920)"] = "Ann Soc Belg Med Trop (1920)",
  ["annales de la société d'émulation de bruges"] = "Ann soc Emul Bruges",
  ["annales de la société entomologique de france. société entomologique de france"] = "Ann Soc Entomol Fr",
  ["annales de la société royale d'archéologie de bruxelles. société royale d'archéologie de bruxelles (belgium)"] = "Ann Soc R Archeol Brux",
  ["annales de la société royale des sciences médicales et naturelles de bruxelles"] = "Ann Soc R Sci Med Nat Brux",
  ["annales de le institue fourier"] = "Ann. Inst. Fourier",
  ["annales de limnologie"] = "Ann Limnol",
  ["annales de l’anesthesiologie francaise"] = "Ann. Anesthesiol. Fr.",
  ["annales de l’est"] = "Ann Est",
  ["annales de l’i.s.u.p."] = "Ann. I.S.U.P.",
  ["annales de l’insee"] = "Ann. INSEE",
  ["annales de l’institut archeologique du luxembourg. institut archeologique du luxembourg (arlon, belgium)"] = "Ann Inst Archeol Luxemb",
  ["annales de l’institut de philosophie"] = "AIPh",
  ["annales de l’institut henri poincare c. analyse non lineaire"] = "Ann. Inst. H. Poincare C Anal. Non Lineaire",
  ["annales de l’institut henri poincare d. combinatorics, physics and their interactions"] = "Ann. Inst. Henri Poincare D",
  ["annales de l’institut henri poincare probabilites et statistiques"] = "Ann. Inst. Henri Poincare Probab. Stat.",
  ["annales de l’institut henri poincare. analyse non lineaire"] = "Ann. Inst. H. Poincare Anal. Non Lineaire",
  ["annales de l’institut henri poincaré"] = "Ann. Inst. Henri Poincaré",
  ["annales de l’institut henri poincaré, section a: physique theorique"] = "Ann. Inst. Henri Poincaré, A",
  ["annales de l’institut henri poincaré, section b: calculdes probabilités et statistique"] = "Ann. Inst. Henri Poincaré, B",
  ["annales de l’institut pasteur"] = "Ann Inst Pasteur (Paris)",
  ["annales de l’institut pasteur de lille"] = "Ann Inst Pasteur Lille",
  ["annales de l’institut pasteur microbiologie"] = "Ann. Inst. Pasteur Microbiol.",
  ["annales de l’institut pasteur. immunologie"] = "Ann Inst Pasteur Immunol",
  ["annales de l’institut pasteur. immunology"] = "Ann Inst Pasteur Immunol",
  ["annales de l’institut pasteur. microbiologie"] = "Ann Inst Pasteur Microbiol",
  ["annales de l’institut pasteur. microbiology"] = "Ann Inst Pasteur Microbiol",
  ["annales de l’institut pasteur. virology"] = "Ann Inst Pasteur Virol",
  ["annales de l’institute henri poincare"] = "Ann. Inst. Henri Poincare",
  ["annales de l’institute henri poincare c"] = "Ann. Inst. Henri Poincare C",
  ["annales de l’institute henri poincare d"] = "Ann. Inst. Henri Poincare D",
  ["annales de l’institute henri poincare, probabilites et statistiques"] = "Ann. Inst. Henri Poincare, Probab. Stat.",
  ["annales de l’universite d’abidjan. serie g: geographie. universite d’abidjan"] = "Ann Univ Abidjan Ser G",
  ["annales de l’école nationale des eaux et forêts et de la station de recherches et expériences forestières"] = "Ann. Éc. Natl. Eaux For.",
  ["annales de medecine interne"] = "Ann Med Interne (Paris)",
  ["annales de medecine legale, criminologie, police scientifique et toxicologie"] = "Ann. Med. Leg. Criminol. Police Sci. Toxicol.",
  ["annales de medecine physique"] = "Ann Med Phys (Lille)",
  ["annales de médecine interne"] = "Ann Med Interne (Paris)",
  ["annales de médecine légale, criminologie, police scientifique et toxicologie"] = "Ann Med Leg Criminol Police Sci Toxicol",
  ["annales de médecine physique"] = "Ann Med Phys (Lille)",
  ["annales de médecine vétérinaire"] = "Ann Med Vet",
  ["annales de microbiologie"] = "Ann Microbiol (Paris)",
  ["annales de normandie"] = "Ann Normandie",
  ["annales de paleontologie"] = "Ann. Paleontol.",
  ["annales de parasitologie humaine et comparee"] = "Ann. Parasitol. Hum. Comp.",
  ["annales de parasitologie humaine et comparée"] = "Ann Parasitol Hum Comp",
  ["annales de pathologie"] = "Ann Pathol",
  ["annales de pediatrie"] = "Ann. Pediatr. (Paris)",
  ["annales de pédiatrie"] = "Ann Pediatr (Paris)",
  ["annales de physique"] = "Ann Phys (Paris)",
  ["annales de physique (paris)"] = "Ann. Phys. (Paris)",
  ["annales de radiologie"] = "Ann Radiol (Paris)",
  ["annales de readaptation et de medecine physique : revue scientifique de la societe francaise de reeducation fonctionnelle de readaptation et de medecine physique"] = "Ann Readapt Med Phys",
  ["annales de recherches veterinaires"] = "Ann. Rech. Vet.",
  ["annales de recherches veterinaires. annals of veterinary research"] = "Ann Rech Vet",
  ["annales de recherches vétérinaires. annals of veterinary research"] = "Ann Rech Vet",
  ["annales de réadaptation et de médecine physique : revue scientifique de la société française de rééducation fonctionnelle de réadaptation et de médecine physique"] = "Ann Readapt Med Phys",
  ["annales de sciences economiques appliqués"] = "Ann. Sci. Econ. Appl.",
  ["annales de toxicologie analytique"] = "Ann. Toxicol. Anal.",
  ["annales de'l institut pasteur-immunology"] = "Ann. Inst. Pasteur. Imm.",
  ["annales de'l institut pasteur-microbiologie"] = "Ann. Inst. Pasteur. Mic.",
  ["annales de'l institut pasteur-virology"] = "Ann. Inst. Pasteur. Vir.",
  ["annales des composites"] = "Ann. Compos.",
  ["annales des falsifications et des fraudes"] = "Ann Falsif Fraud",
  ["annales des ponts et chausses"] = "Ann. Ponts Chaussees",
  ["annales des sciences forestières"] = "Ann. sci. for.",
  ["annales des sciences mathématiques du québec"] = "Ann. Sci. Math. Québec",
  ["annales des sciences naturelles - zoologie et biologie animale"] = "Ann. Sci. Nat. Zool. Biol. Anim.",
  ["annales des sociétés belges de médecine tropicale, de parasitologie, et de mycologie"] = "Ann Soc Belges Med Trop Parasitol Mycol",
  ["annales des telecommunications"] = "Ann Telecommun",
  ["annales du cercle archéologique d'enghien"] = "Ann Cercle Archeol Enghien",
  ["annales du cercle archéologique du canton de soignies"] = "Ann Cercle Archeol Canton Soignies",
  ["annales du cercle royal d'histoire et d'archéologie d'ath et de la région et musées athois"] = "Ann Cerc Roy Hist Archeol Ath",
  ["annales du cercle royal d’histoire et d’archeologie d’ath et de la region et musees athois"] = "Ann Cerc Roy Hist Archeol Ath",
  ["annales du midi"] = "Ann Midi",
  ["annales du service des antiquites de l’egypte / ministere des travaux publics"] = "Ann Serv Antiq Egypte",
  ["annales du service des antiquités de l'egypte"] = "Ann Serv Antiq Egypte",
  ["annales du service des antiquités de l’égypte"] = "ASAE",
  ["annales du service des antiquités d’égypte"] = "ASAE",
  ["annales d’anatomie pathologique"] = "Ann. Anat. Pathol. (Paris)",
  ["annales d’economie et de statistique"] = "Ann. Econ. Statist.",
  ["annales d’endocrinologie"] = "Ann Endocrinol (Paris)",
  ["annales d’histochimie"] = "Ann Histochim",
  ["annales d’histoire et d’archéologie"] = "AnnHistA",
  ["annales d’histoire et d’archéologie / université saint-joseph"] = "AHAUSJ",
  ["annales d’hygiene et de medecine coloniales"] = "Ann Hyg Med Colon",
  ["annales d’immunologie"] = "Ann Immunol (Paris)",
  ["annales d’oculistique"] = "Ann Ocul (Paris)",
  ["annales d’oto-laryngologie et de chirurgie cervico faciale : bulletin de la societe d’oto-laryngologie des hopitaux de paris"] = "Ann Otolaryngol Chir Cervicofac",
  ["annales d’urologie"] = "Ann Urol (Paris)",
  ["annales d’économie et de statistique"] = "Ann. Économ. Statist.",
  ["annales d’éthiopie"] = "AnnEth",
  ["annales entomologici fennici"] = "Ann. Entomol. Fenn.",
  ["annales fennici mathematici"] = "Ann. Fenn. Math.",
  ["annales francaises d anesthesie et de reanimation"] = "Ann Fr Anesth Reanim",
  ["annales francaises d’anesthesie et de reanimation"] = "Ann. Fr. Anesth. Reanim.",
  ["annales françaises d'anesthèsie et de rèanimation"] = "Ann Fr Anesth Reanim",
  ["annales geophysicae"] = "Ann. Geophys.",
  ["annales henri lebesgue"] = "Ann. H. Lebesgue",
  ["annales henri poincare"] = "Ann. Henri Poincare",
  ["annales henri poincare. a journal of theoretical and mathematical physics"] = "Ann. Henri Poincare",
  ["annales henri poincaré"] = "Ann Henri Poincare",
  ["annales henri poincaré"] = "Ann. Henri Poincaré",
  ["annales historico-naturales musei nationalis hungarici"] = "Ann. Hist.-Nat. Mus. Natl. Hung.",
  ["annales historiques de la revolution francaise"] = "Ann Hist Revolut Fr",
  ["annales historiques de la révolution française"] = "Ann Hist Revolut Fr",
  ["annales historiques du mantois"] = "Ann Hist Mantois",
  ["annales hydrographiques"] = "Ann. Hydrogr.",
  ["annales immunologiae hungaricae"] = "Ann Immunol Hung",
  ["annales latini montium arvernorum: bulletin du groupe d’études latines de l’université de clermont"] = "ALMArv",
  ["annales mathematicae et informaticae"] = "Ann. Math. Inform.",
  ["annales mathematicae silesianae"] = "Ann. Math. Sil.",
  ["annales mathematiques blaise pascal"] = "Ann. Math. Blaise Pascal",
  ["annales mathematiques du quebec"] = "Ann. Math. Que.",
  ["annales mathématiques blaise pascal"] = "Ann. Math. Blaise Pascal",
  ["annales medicales de nancy"] = "Ann. Med. Nancy",
  ["annales medicinae experimentalis et biologiae fenniae"] = "Ann Med Exp Biol Fenn",
  ["annales medicinae internae fenniae"] = "Ann Med Intern Fenn",
  ["annales medicinae internae fenniae. supplementum"] = "Ann Med Intern Fenn Suppl",
  ["annales medico-psychologiques"] = "Ann. Med. Psychol. (Paris)",
  ["annales médicales de nancy"] = "Ann Med Nancy",
  ["annales médico-psychologiques"] = "Ann Med Psychol (Paris)",
  ["annales nestle"] = "Ann. Nestle [Fr.]",
  ["annales nestle [english ed.]"] = "Ann Nestle [Eng]",
  ["annales nestle. [deutsche aufl.]"] = "Ann Nestle [Ger]",
  ["annales nestle. [ed. francaise]"] = "Ann Nestle [Fr]",
  ["annales nestlé [english ed.]"] = "Ann Nestle Eng",
  ["annales nestlé. [deutsche aufl.]"] = "Ann Nestle Ger",
  ["annales nestlé. [ed. française]"] = "Ann Nestle Fr",
  ["annales odonto-stomatologiques"] = "Ann Odontostomatol (Lyon)",
  ["annales paediatriae fenniae"] = "Ann Paediatr Fenn",
  ["annales paediatrici (international review of pediatrics)"] = "Ann. Paediatr. Int. Rev. Pediatr.",
  ["annales paediatrici. international review of pediatrics"] = "Ann Paediatr",
  ["annales pharmaceutiques francaises"] = "Ann. Pharm. Fr.",
  ["annales pharmaceutiques françaises"] = "Ann Pharm Fr",
  ["annales polonici mathematici"] = "Ann. Polon. Math.",
  ["annales scientifiques de l'université de besançon. médecine"] = "Ann Sci Univ Besancon Med",
  ["annales scientifiques de le ecole normale superieure"] = "Ann. Sci. Ec. Norm. Super.",
  ["annales scientifiques de l’ecole normale superieure. quatrieme serie"] = "Ann. Sci. Ec. Norm. Super. (4)",
  ["annales scientifiques de l’universite de besancon. medecine"] = "Ann Sci Univ Besancon Med",
  ["annales scientifiques de l’école normale supérieure"] = "Ann. Sci. École Norm. Sup. (4)",
  ["annales universitatis budapestinensis de rolando eötvös nominatae, sectio classica"] = "AUB(class)",
  ["annales universitatis budapestinensis de rolando eötvös nominatae, ê sectio philosophica et sociologica"] = "AUB(phil)",
  ["annales universitatis budapestinensis de rolando eötvös nominatae, ê sectio iuridica"] = "AUBFjur.)",
  ["annales universitatis mariae curie-sklodowska. sectio ai. informatica"] = "Ann. Univ. Mariae Curie-Sklodowska Sect. AI-Inform.",
  ["annales universitatis mariae curie-sklodowska. sectio d, medicina"] = "Ann. Univ. Mariae Curie Sklodowska [Med.]",
  ["annales universitatis mariae curie-sklodowska. sectio d: medicina"] = "Ann Univ Mariae Curie Sklodowska [Med]",
  ["annales universitatis mariae curie-skłodowska"] = "Ann. Univ. Mariae Curie-Skłodowska Sect. A",
  ["annales universitatis mariae curie-skłodowska. sectio d: medicina"] = "Ann Univ Mariae Curie Sklodowska Med",
  ["annales universitatis mariae curie-skłodowska. section ee, zootechnica"] = "Ann Univ Mariae Curie Sklodowska Sect EE Zootech",
  ["annales universitatis paedagogicae cracoviensis. studia mathematica"] = "Ann. Univ. Paedagog. Crac. Stud. Math.",
  ["annales universitatis saraviensis"] = "Ann. Univ. Sarav. Ser. Math.",
  ["annales universitatis saraviensis. medizin"] = "Ann Univ Sarav Med",
  ["annales universitatis scientiarum budapestinensis de rolando eotvos nominatae. sectio biologica"] = "Ann Univ Sci Budap Rolando Eotvos Nominatae [Biol]",
  ["annales universitatis scientiarum budapestinensis de rolando eotvos nominatae. sectio computatorica"] = "Ann. Univ. Sci. Budapest. Sect. Comput.",
  ["annales universitatis scientiarum budapestinensis de rolando eotvos nominatae. sectio geographica. eotvos lorand tudomanyegyetem"] = "Ann Univ Sci Budap Rolando Eotvos Nom Sect Geogr",
  ["annales universitatis scientiarum budapestinensis de rolando eotvos nominatae. sectio mathematica"] = "Ann. Univ. Sci. Budapest. Eotvos Sect. Math.",
  ["annales universitatis scientiarum budapestinensis de rolando eötvös nominatae. sectio biologica"] = "Ann Univ Sci Budap Rolando Eotvos Nominatae Biol",
  ["annales universitatis scientiarum budapestinensis de rolando eötvös nominatae. sectio geographica. eötvös loránd tudományegyetem"] = "Ann Univ Sci Budap Rolando Eotvos Nom Sect Geogr",
  ["annales universitatis scientiarum budapestinensis de rolando eötvös nominatae"] = "Ann. Univ. Sci. Budapest. Eötvös Sect. Math.",
  ["annales universitatis turkuensis"] = "Ann. Univ. Turku. Ser. A I",
  ["annales valaisannes"] = "AV",
  ["annales zoologici"] = "Annal Zool",
  ["annales zoologici fennici"] = "Ann Zool Fennici",
  ["annales. economies, sociétés, civilisations"] = "AnnEconSocCiv",
  ["annales. faculte des lettres et sciences humaines de nice"] = "Ann Fac Lett Sci Hum Nice",
  ["annales. faculté des lettres et sciences humaines de nice"] = "Ann Fac Lett Sci Hum Nice",
  ["annales. histoire, sciences sociales"] = "AnnHistScSoc",
  ["annales. societe francaise d’angeiologie et d’histopathologie"] = "Ann Soc Angeiol Histopathologie",
  ["annales. société française d'angéiologie et d'histopathologie"] = "Ann Soc Angeiol Histopathologie",
  ["annales. universitatis mariae curie-sklodowska. sectio a. mathematica"] = "Ann. Univ. Mariae Curie-Sklodowska Sect. A",
  ["annales: histoire, sciences sociales"] = "Annales (HSS)",
  ["annales: économies, sociétés, civilisations"] = "Annales (ESC)",
  ["annali - istituto giangiacomo feltrinelli"] = "Ann Ist Giangiacomo Feltrinelli",
  ["annali accademia italiana di scienze forestali"] = "Ann. Accad. Ital. Sci. For.",
  ["annali benacensi"] = "AnnBenac",
  ["annali d'igiene"] = "Ann Ig (Roma)",
  ["annali del dipartimento di filosofia dell’università di firenze"] = "ADFF",
  ["annali del liceo classico g. garibaldi di palermo"] = "ALGP",
  ["annali del museo civico di rovereto. sezione archeologia, storia, scienze naturali"] = "AnnMusRov",
  ["annali del seminario giuridico dell’univ. di palermo"] = "ASGP",
  ["annali dell istituto superiore di sanita"] = "Ann Ist Super Sanita",
  ["annali dell ospedale maria vittoria di torino"] = "Ann. Osp. Maria Vittoria Torino",
  ["annali dell'istituto e museo di storia della scienza di firenze"] = "Ann Ist Mus Stor Sci Fir",
  ["annali dell'istituto storico italo germanico"] = "Ann Ist Stor Italo Ger",
  ["annali dell'istituto storico italo-germanico in trento. contributi = jahrbuch des italienisch-deutschen historischen instituts in trient. beiträge"] = "Ann Ist Stor Italo Ger Trento",
  ["annali dell'istituto superiore di sanità"] = "Ann Ist Super Sanita",
  ["annali dell'ospedale maria vittoria di torino"] = "Ann Osp Maria Vittoria Torino",
  ["annali della facolta di economia e commercio (universita di messina)"] = "Ann Fac Econ Commer",
  ["annali della facolta di medicina e chirurgia della universita degli studi di perugia"] = "Ann Fac Med Chir Univ Studi Perugia",
  ["annali della facoltà di economia e commercio (università di messina)"] = "Ann Fac Econ Commer",
  ["annali della facoltà di medicina e chirurgia della università degli studi di perugia"] = "Ann Fac Med Chir Univ Studi Perugia",
  ["annali della facoltà di economia e commercio, università palermo"] = "Ann. Fac. Econ. Com.",
  ["annali della facoltà di giurisprudenza, università di genova"] = "AFGG",
  ["annali della facoltà di lettere di lecce"] = "AFLL",
  ["annali della facoltà di lettere e filosofia della università di cagliari"] = "AFLC",
  ["annali della facoltà di lettere e filosofia della università di napoli"] = "AFLN",
  ["annali della facoltà di lettere e filosofia dell’università di siena"] = "AFLS",
  ["annali della facoltà di lettere e filosofia di bari"] = "AFLB",
  ["annali della facoltà di lettere e filosofia di perugia"] = "AFLPer",
  ["annali della facoltà di lettere e filosofia, università degli studi di perugia, 1. studi classici"] = "AnnPerugia",
  ["annali della facoltà di lettere e filosofia, università degli studi, bari"] = "AnnBari",
  ["annali della facoltà di lettere e filosofia, università di macerata"] = "AFLM",
  ["annali della facoltà di lettere e filosofia, università di napoli"] = "AnnNap",
  ["annali della facoltà di lettere e filosofia, università di siena"] = "AnnSiena",
  ["annali della facoltà di magistero dell’università di cagliari"] = "AFMC",
  ["annali della facoltà di scienze della formazione dell’università di cagliari"] = "AFSFC",
  ["annali della fondazione luigi einaudi. fondazione luigi einaudi"] = "Ann Fond Luigi Einaudi",
  ["annali della fondazione per il museo claudio faina"] = "AnnFaina",
  ["annali della sanita pubblica"] = "Ann. Sanita Pubblica",
  ["annali della sanità pubblica"] = "Ann Sanita Pubblica",
  ["annali della scuola normale superiore di pisa"] = "Ann. Scuola Norm. Sup. Pisa Cl. Sci. (4)",
  ["annali della scuola normale superiore di pisa classe di scienze"] = "Ann. Sc. Norm. Super. Pisa Cl. Sci.",
  ["annali della scuola normale superiore di pisa, classe di lettere e filosofia"] = "ASNP",
  ["annali della scuola normale superiore di pisa. classe di lettere e filosofia. scuola normale superiore (italy)"] = "Ann Sc Norm Super Pisa",
  ["annali della scuola normale superiore di pisa. classe di scienze. serie v"] = "Ann. Sc. Norm. Super. Pisa Cl. Sci. (5)",
  ["annali della sperimentazione agraria"] = "Ann. Sper. Agrar.",
  ["annali dell’instituto di corrispondenza archeologica"] = "AdI",
  ["annali dell’istituto carlo forlanini"] = "Ann. Ist. Carlo Forlanini",
  ["annali dell’istituto di storia, univ. di firenze"] = "AISF",
  ["annali dell’istituto e museo di storia della scienza di firenze"] = "Ann Ist Mus Stor Sci Fir",
  ["annali dell’istituto italiano di numismatica"] = "AIIN",
  ["annali dell’istituto italiano per gli studi storici"] = "AIIS",
  ["annali dell’istituto sperimentale per la selvicoltura"] = "Ann. Ist. Sper. Selvic.",
  ["annali dell’istituto sperimentale per l’assestamento forestale e per l’alpicoltura"] = "Ann. Ist. Sper. Assest. For. Alpic.",
  ["annali dell’istituto superiore di sanita"] = "Ann. Ist. Super. Sanita",
  ["annali dell’istituto universitario orientale di napoli, dipartimento di studi del mondo classico e del mediterraneo antico, sezione di archeologia e storia antica"] = "AION(archeol)",
  ["annali dell’istituto universitario orientale di napoli, dipartimento di studi del mondo classico e del mediterraneo antico, sezione filologico-letteraria"] = "AION(filol)",
  ["annali dell’istituto universitario orientale di napoli, dipartimento di studi del mondo classico e del mediterraneo antico, sezione linguistica"] = "AION(ling)",
  ["annali dell’istituto universitario orientale di napoli. dipartimento di studi del mondo classico e del mediterraneo antico. sezione di archeologia e storia antica"] = "AIONArch",
  ["annali dell’istituto universitario orientale di napoli. dipartimento di studi del mondo classico e del mediterraneo antico. sezione filologicoletteraria"] = "AIONFil",
  ["annali dell’istituto universitario orientale di napoli. dipartimento di studi del mondo classico e del mediterraneo antico. sezione linguistica"] = "AIONLing",
  ["annali dell’ospedale maria vittoria di torino"] = "Ann Osp Maria Vittoria Torino",
  ["annali dell’universita di ferrara. sezione vii. scienze matematiche"] = "Ann. Univ. Ferrara Sez. VII Sci. Mat.",
  ["annali dell’università di ferrara"] = "Ann. Univ. Ferrara Sez. VII (N.S.)",
  ["annali dell’università di ferrara, sez. v, sc. giurid"] = "AUFG",
  ["annali dell’università di lecce. facoltà di lettere e filosofia e di magistero"] = "AnnLecce",
  ["annali di chimica"] = "Ann Chim",
  ["annali di chimica (rome, italy)"] = "Ann. Chim. (Rome, Italy)",
  ["annali di geofisica"] = "Ann. Geofis.",
  ["annali di igiene"] = "Ann. Ig.",
  ["annali di igiene : medicina preventiva e di comunita"] = "Ann Ig",
  ["annali di igiene : medicina preventiva e di comunità"] = "Ann Ig",
  ["annali di laringologia, otologia, rinologia, faringologia"] = "Ann Laringol Otol Rinol Faringol",
  ["annali di matematica pura ed applicata"] = "Ann Mat Pura Appl",
  ["annali di matematica pura ed applicata. series iv"] = "Ann. Mat. Pura Appl. (4)",
  ["annali di medicina navale"] = "Ann Med Nav (Roma)",
  ["annali di medicina navale e tropicale"] = "Ann Med Nav Trop",
  ["annali di microbiologia ed enzimologia"] = "Ann Microbiol Enzimol",
  ["annali di neuropsichiatria e psicoanalisi"] = "Ann Neuropsichiatr Psicoanal",
  ["annali di ostetricia e ginecologia"] = "Ann Ostet Ginecol",
  ["annali di ostetricia, ginecologia, medicina perinatale"] = "Ann Ostet Ginecol Med Perinat",
  ["annali di ottalmologia e clinica oculistica"] = "Ann Ottalmol Clin Ocul",
  ["annali di radiologia diagnostica"] = "Ann Radiol Diagn (Bologna)",
  ["annali di stomatologia"] = "Ann Stomatol (Roma)",
  ["annali di storia dell’esegesi"] = "AnnSE",
  ["annali d’igiene"] = "Ann Ig (Roma)",
  ["annali isnardi di auxologia normale e patologica"] = "Ann Isnardi Auxol Norm Patol",
  ["annali italiani di chirurgia"] = "Ann Ital Chir",
  ["annali italiani di dermatologia e sifilografia"] = "Ann Ital Dermatol Sifilogr",
  ["annali italiani di medicina interna"] = "Ann. Ital. Med. Int.",
  ["annali italiani di medicina interna : organo ufficiale della societa italiana di medicina interna"] = "Ann Ital Med Int",
  ["annali italiani di medicina interna : organo ufficiale della società italiana di medicina interna"] = "Ann Ital Med Int",
  ["annali italiani di pediatria"] = "Ann Ital Pediatr",
  ["annali medici"] = "Ann Med (Milano)",
  ["annali medici di sondalo"] = "Ann Med Sondalo",
  ["annali sclavo"] = "Ann. Sclavo",
  ["annali sclavo. collana monografica"] = "Ann Sclavo Collana Monogr",
  ["annali sclavo; rivista di microbiologia e di immunologia"] = "Ann Sclavo",
  ["annali zavoda za povijesne znanosti istraivakog centra jugoslavenske akademije znanosti i umjetnosti u dubrovniku"] = "AnDubr",
  ["annali. associazione nomentana di storia e archeologia"] = "AnnNoment",
  ["annali. istituto italiano di numismatica"] = "AnnIstItNum",
  ["annali. rivista del dipartimento di studi asiatici e del dipartimento di studi e ricerche su africa e paesi arabi, istituto universitario orientale di napoli"] = "AnnOrNap",
  ["annali. sezione di archeologia e storia antica. istituto universitario orientale di napoli. dipartimento di studi del mondo classico e del mediterraneo antico"] = "AnnAStorAnt",
  ["annals (royal college of physicians and surgeons of canada)"] = "Ann R Coll Physicians Surg Can",
  ["annals chimie analytique"] = "Ann. Chim. Anal.",
  ["annals of 3d printed medicine"] = "Ann. 3D Print. Med.",
  ["annals of 3d printing in medicine"] = "Ann. 3D Print. Med.",
  ["annals of advances in automotive medicine. association for the advancement of automotive medicine. annual scientific conference"] = "Ann Adv Automot Med",
  ["annals of african medicine"] = "Ann Afr Med",
  ["annals of agrarian science"] = "Ann Agrar Sci",
  ["annals of agricultural & crop sciences"] = "Ann Agric Crop Sci",
  ["annals of agricultural and environmental medicine"] = "Ann. Agric. Environ. Med.",
  ["annals of agricultural and environmental medicine : aaem"] = "Ann Agric Environ Med",
  ["annals of agricultural science"] = "Ann Agric Sci",
  ["annals of agricultural sciences"] = "Ann. Agric. Sci.",
  ["annals of allergy"] = "Ann Allergy",
  ["annals of allergy, asthma & immunology : official publication of the american college of allergy, asthma, & immunology"] = "Ann Allergy Asthma Immunol",
  ["annals of allergy, asthma, and immunology"] = "Ann. Allergy. Asthma. Immunol.",
  ["annals of anatomy = anatomischer anzeiger : official organ of the anatomische gesellschaft"] = "Ann Anat",
  ["annals of animal science"] = "Ann. Anim. Sci.",
  ["annals of anthropological practice"] = "Ann Anthropol Pract",
  ["annals of applied bio-sciences"] = "Ann Appl Biosci",
  ["annals of applied biology"] = "Ann. Appl. Biol.",
  ["annals of applied mathematics. yingyong shuxue niankan"] = "Ann. Appl. Math.",
  ["annals of applied probability"] = "Ann. Appl. Probab.",
  ["annals of applied sport science"] = "Ann Appl Sport Sci",
  ["annals of applied statistics"] = "Ann. Appl. Stat.",
  ["annals of arid zone"] = "Ann Arid Zone",
  ["annals of behavioral medicine"] = "Ann. Behav. Med.",
  ["annals of behavioral medicine : a publication of the society of behavioral medicine"] = "Ann Behav Med",
  ["annals of behavioral neuroscience"] = "Ann Behav Neurosci",
  ["annals of behavioral science and medical education : journal of the association for the behavioral sciences and medical education"] = "Ann Behav Sci Med Educ",
  ["annals of biochemistry and experimental medicine"] = "Ann Biochem Exp Med",
  ["annals of bioethics"] = "Ann Bioeth",
  ["annals of biological research"] = "Ann Biol Res",
  ["annals of biology"] = "Ann Biol",
  ["annals of biomedical engineering"] = "Ann Biomed Eng",
  ["annals of biometrics & biostatistics"] = "Ann Biom Biostat",
  ["annals of blood"] = "Ann Blood",
  ["annals of botany"] = "Ann Bot",
  ["annals of breast cancer and therapy"] = "Ann Breast Cancer Ther",
  ["annals of burns and fire disasters"] = "Ann Burns Fire Disasters",
  ["annals of cancer epidemiology"] = "Ann Cancer Epidemiol",
  ["annals of cancer research"] = "Ann Cancer Res",
  ["annals of cardiac anaesthesia"] = "Ann Card Anaesth",
  ["annals of cardiothoracic surgery"] = "Ann Cardiothorac Surg",
  ["annals of carnegie museum"] = "Ann. Carnegie Mus.",
  ["annals of clinical and experimental hypertension"] = "Ann Clin Exp Hypertens",
  ["annals of clinical and laboratory research"] = "Ann Clin Lab Res",
  ["annals of clinical and laboratory science"] = "Ann Clin Lab Sci",
  ["annals of clinical and translational neurology"] = "Ann Clin Transl Neurol",
  ["annals of clinical biochemistry"] = "Ann Clin Biochem",
  ["annals of clinical case reports"] = "Ann Clin Case Rep",
  ["annals of clinical cytology and pathology"] = "Ann Clin Cytol Pathol",
  ["annals of clinical laboratory science"] = "Ann Clin Lab Sci",
  ["annals of clinical microbiology and antimicrobials"] = "Ann Clin Microbiol Antimicrob",
  ["annals of clinical pathology"] = "Ann Clin Pathol",
  ["annals of clinical psychiatry"] = "Ann. Clin. Psychiatry",
  ["annals of clinical psychiatry : official journal of the american academy of clinical psychiatrists"] = "Ann Clin Psychiatry",
  ["annals of clinical research"] = "Ann Clin Res",
  ["annals of coloproctology"] = "Ann Coloproctol",
  ["annals of combinatorics"] = "Ann. Comb.",
  ["annals of data science"] = "Ann. Data Sci.",
  ["annals of dentistry"] = "Ann Dent",
  ["annals of dentistry and oral disorders"] = "Ann Dent Oral Disord",
  ["annals of depression and anxiety"] = "Ann Depress Anxiety",
  ["annals of dermatology"] = "Ann Dermatol",
  ["annals of diagnostic pathology"] = "Ann Diagn Pathol",
  ["annals of differential equations"] = "Ann. Differential Equations",
  ["annals of discrete mathematics"] = "Ann. Discrete Math.",
  ["annals of dyslexia"] = "Ann Dyslexia",
  ["annals of earth : a publication of ocean arks international and the lindisfarne association"] = "Ann Earth",
  ["annals of economic and social measurement"] = "Ann. Econ. Soc. Meas.",
  ["annals of economics and finance"] = "Ann. Econ. Finance",
  ["annals of economics and statistics"] = "Ann Econ Stat",
  ["annals of emergency dispatch & response"] = "Ann Emerg Dispatch Response",
  ["annals of emergency medicine"] = "Ann Emerg Med",
  ["annals of epidemiology"] = "Ann Epidemiol",
  ["annals of epidemiology and public health"] = "Ann Epidemiol Public Health",
  ["annals of eugenics"] = "Ann Eugen",
  ["annals of eye science"] = "Ann Eye Sci",
  ["annals of family medicine"] = "Ann Fam Med",
  ["annals of finance"] = "Ann. Finance",
  ["annals of fluid dynamics"] = "Ann. Fluid Dyn.",
  ["annals of forensic research and analysis"] = "Ann Forensic Res Anal",
  ["annals of forest research"] = "Ann For Res",
  ["annals of forest science"] = "Ann For Sci",
  ["annals of functional analysis"] = "Ann. Funct. Anal.",
  ["annals of fuzzy mathematics and informatics"] = "Ann. Fuzzy Math. Inform.",
  ["annals of gastroenterological surgery"] = "Ann Gastroenterol Surg",
  ["annals of gastroenterology"] = "Ann Gastroenterol",
  ["annals of gastroenterology & hepatology"] = "Ann Gastroentol Hepatol",
  ["annals of gastroenterology and digestive disorders"] = "Ann Gastroenterol Dig Disord",
  ["annals of general hospital psychiatry"] = "Ann Gen Hosp Psychiatry",
  ["annals of general psychiatry"] = "Ann Gen Psychiatry",
  ["annals of geophysics"] = "Ann. Geophys.",
  ["annals of geriatric medicine and research"] = "Ann Geriatr Med Res",
  ["annals of gerontology and geriatric research"] = "Ann Gerontol Geriatr Res",
  ["annals of gis"] = "Ann. Gis",
  ["annals of glaciology"] = "Ann. Glaciol.",
  ["annals of global analysis and geometry"] = "Ann. Global Anal. Geom.",
  ["annals of global health"] = "Ann Glob Health",
  ["annals of health law"] = "Ann Health Law",
  ["annals of health law / loyola university chicago, school of law, institute for health law"] = "Ann Health Law",
  ["annals of health law and life sciences"] = "Ann Health Law Life Sci",
  ["annals of hematology"] = "Ann Hematol",
  ["annals of hematology & oncology"] = "Ann Hematol Oncol",
  ["annals of hepato-biliary-pancreatic surgery"] = "Ann Hepatobiliary Pancreat Surg",
  ["annals of hepatology"] = "Ann Hepatol",
  ["annals of hepatology : official journal of the mexican association of hepatology"] = "Ann Hepatol",
  ["annals of human biology"] = "Ann Hum Biol",
  ["annals of human genetics"] = "Ann Hum Genet",
  ["annals of hygiene"] = "Ann Hyg",
  ["annals of ibadan postgraduate medicine"] = "Ann Ib Postgrad Med",
  ["annals of immunology"] = "Ann Immunol",
  ["annals of indian academy of neurology"] = "Ann Indian Acad Neurol",
  ["annals of infection"] = "Ann Infect",
  ["annals of integrative molecular medicine"] = "Ann Integr Mol Med",
  ["annals of intensive care"] = "Ann Intensive Care",
  ["annals of internal medicine"] = "Ann Intern Med",
  ["annals of iowa"] = "Ann Iowa",
  ["annals of joint"] = "Ann Jt",
  ["annals of k-theory"] = "Ann. K-Theory",
  ["annals of king edward medical university"] = "Ann King Edw Med Univ",
  ["annals of laboratory medicine"] = "Ann Lab Med",
  ["annals of laparoscopic and endoscopic surgery"] = "Ann Laparosc Endosc Surg",
  ["annals of lgbtq public and population health"] = "Ann LGBTQ Public Popul Health",
  ["annals of lung cancer"] = "Ann Lung Cancer",
  ["annals of lymphoma"] = "Ann Lymphoma",
  ["annals of mathematical logic"] = "Ann. Math. Log.",
  ["annals of mathematical sciences and applications"] = "Ann. Math. Sci. Appl",
  ["annals of mathematical statistics"] = "Ann. Math. Stat.",
  ["annals of mathematics"] = "Ann. Math.",
  ["annals of mathematics and artif. intell."] = "Ann. Math. Artif. Intell.",
  ["annals of mathematics and artificial intelligence"] = "Ann. Math. Artif. Intell.",
  ["annals of mathematics studies"] = "Ann. of Math. Stud.",
  ["annals of mathematics. second series"] = "Ann. of Math. (2)",
  ["annals of maxillofacial surgery"] = "Ann Maxillofac Surg",
  ["annals of medical and health sciences research"] = "Ann Med Health Sci Res",
  ["annals of medical history"] = "Ann Med Hist",
  ["annals of medical research"] = "Ann Med Res",
  ["annals of medicine"] = "Ann Med",
  ["annals of medicine and surgery (2012)"] = "Ann Med Surg (Lond)",
  ["annals of microbiology"] = "Ann Microbiol",
  ["annals of musculoskeletal medicine"] = "Ann Musculoskelet Med",
  ["annals of neurology"] = "Ann Neurol",
  ["annals of neuroscience and psychology"] = "Ann Neurosci Psychol",
  ["annals of neurosciences"] = "Ann Neurosci",
  ["annals of noninvasive electrocardiology"] = "Ann. Noninvasive Electrocardiol.",
  ["annals of noninvasive electrocardiology : the official journal of the international society for holter and noninvasive electrocardiology, inc"] = "Ann Noninvasive Electrocardiol",
  ["annals of nuclear cardiology"] = "Ann Nucl Cardiol",
  ["annals of nuclear energy"] = "Ann. Nucl. Energy",
  ["annals of nuclear engineering"] = "Ann. Nucl. Eng.",
  ["annals of nuclear medicine"] = "Ann Nucl Med",
  ["annals of nuclear science and engineering"] = "Ann. Nucl. Sci. Eng.",
  ["annals of numerical mathematics"] = "Ann. Numer. Math.",
  ["annals of nutrition & metabolism"] = "Ann Nutr Metab",
  ["annals of nutrition and metabolism"] = "Ann. Nutr. Metab.",
  ["annals of occupational and environmental medicine"] = "Ann Occup Environ Med",
  ["annals of occupational hygiene"] = "Ann. Occup. Hyg.",
  ["annals of oncology"] = "Ann. Oncol.",
  ["annals of oncology : official journal of the european society for medical oncology"] = "Ann Oncol",
  ["annals of oncology : official journal of the european society for medical oncology / esmo"] = "Ann Oncol",
  ["annals of operations research"] = "Ann. Oper. Res.",
  ["annals of ophthalmology"] = "Ann Ophthalmol",
  ["annals of ophthalmology (skokie, ill.)"] = "Ann Ophthalmol (Skokie)",
  ["annals of otology, rhinology and laryngology"] = "Ann. Otol. Rhinol. Laryngol.",
  ["annals of otology, rhinology, and laryngology. supplement"] = "Ann. Otol. Rhinol. Laryngol. Suppl.",
  ["annals of paediatric rheumatology"] = "Ann Paediatr Rheumatol",
  ["annals of palliative medicine"] = "Ann Palliat Med",
  ["annals of pancreatic cancer"] = "Ann Pancreat Cancer",
  ["annals of parasitology"] = "Ann Parasitol",
  ["annals of pde"] = "Ann PDE",
  ["annals of pde. journal dedicated to the analysis of problems from physical sciences"] = "Ann. PDE",
  ["annals of pediatric cardiology"] = "Ann Pediatr Cardiol",
  ["annals of pediatric endocrinology & metabolism"] = "Ann Pediatr Endocrinol Metab",
  ["annals of pediatric research"] = "Ann Pediatr Res",
  ["annals of pediatrics & child health"] = "Ann Pediatr Child Health",
  ["annals of periodontology"] = "Ann Periodontol",
  ["annals of periodontology / the american academy of periodontology"] = "Ann Periodontol",
  ["annals of pharmacology and pharmaceutics"] = "Ann Pharmacol Pharm",
  ["annals of pharmacotherapy"] = "Ann. Pharmacother.",
  ["annals of physical and rehabilitation medicine"] = "Ann Phys Rehabil Med",
  ["annals of physical medicine"] = "Ann Phys Med",
  ["annals of physics"] = "Ann. Phys.",
  ["annals of physics (amsterdam, netherlands)"] = "Ann. Phys. (Amsterdam, Neth.)",
  ["annals of physics (new york)"] = "Ann. Phys. (N.Y.)",
  ["annals of physics (oxford, united kingdom)"] = "Ann. Phys. (Oxford, U. K.)",
  ["annals of physics (san diego, ca, united states)"] = "Ann. Phys. (San Diego, CA, U. S.)",
  ["annals of physiological anthropology"] = "Ann. Physiol. Anthropol.",
  ["annals of phytomedicine"] = "Ann Phytomed",
  ["annals of plant protection sciences"] = "Ann. Plant Prot. Sci.",
  ["annals of plastic surgery"] = "Ann Plast Surg",
  ["annals of probability"] = "Ann Probab",
  ["annals of psychiatry and mental health"] = "Ann Psychiatry Ment Health",
  ["annals of public and cooperative economics"] = "Ann. Public Coop. Econ.",
  ["annals of public health and research"] = "Ann Public Health Res",
  ["annals of pulmonary and critical care medicine"] = "Ann Pulm Crit Care Med",
  ["annals of pure and applied logic"] = "Ann. Pure Appl. Logic",
  ["annals of radiation therapy and oncology"] = "Ann Radiat Ther Oncol",
  ["annals of regional science"] = "Ann. Reg. Sci.",
  ["annals of rehabilitation medicine"] = "Ann Rehabil Med",
  ["annals of representation theory"] = "Ann. Represent. Theory",
  ["annals of research hospitals"] = "Ann Res Hosp",
  ["annals of reviews and research"] = "Ann Rev Res",
  ["annals of saudi medicine"] = "Ann Saudi Med",
  ["annals of science"] = "Ann Sci",
  ["annals of science and technology policy"] = "Ann. Sci. Technol. Policy",
  ["annals of software engineering"] = "Ann. Software Eng.",
  ["annals of solid and structural mechanics"] = "Ann. Solid Struct. Mech.",
  ["annals of sports medicine and research"] = "Ann Sports Med Res",
  ["annals of statistics"] = "Ann Stat",
  ["annals of stem cell research"] = "Ann Stem Cell Res",
  ["annals of stem cell research & therapy"] = "Ann Stem Cell Res Ther",
  ["annals of surgery"] = "Ann Surg",
  ["annals of surgery open : perspectives of surgical history, education, and clinical approaches"] = "Ann Surg Open",
  ["annals of surgical innovation and research"] = "Ann Surg Innov Res",
  ["annals of surgical oncology"] = "Ann Surg Oncol",
  ["annals of surgical treatment and research"] = "Ann Surg Treat Res",
  ["annals of telecommunications"] = "Ann. Telecommun.",
  ["annals of the academy of medicine, singapore"] = "Ann Acad Med Singap",
  ["annals of the academy of romanian scientists"] = "Ann. Acad. Rom. Sci.",
  ["annals of the academy of romanian scientists. series on mathematics and its applications"] = "Ann. Acad. Rom. Sci. Ser. Math. Appl.",
  ["annals of the american academy of political and social science"] = "Ann. Amer. Acad. Polit. Soc. Sci.",
  ["annals of the american association of geographers"] = "Ann Am Assoc Geogr",
  ["annals of the american thoracic society"] = "Ann Am Thorac Soc",
  ["annals of the association of american geographers. association of american geographers"] = "Ann Assoc Am Geogr",
  ["annals of the australian college of dental surgeons"] = "Ann Aust Coll Dent Surg",
  ["annals of the canadian society for history and philosophy of mathematics/societe canadienne d’histoire et de philosophie des mathematiques"] = "Ann. Can. Soc. Hist. Philos. Math./Soc. Can. Hist. Philos. Math.",
  ["annals of the carnegie museum"] = "Ann Carnegie Mus",
  ["annals of the entomological society of america"] = "Ann Entomol Soc Am",
  ["annals of the history of computing"] = "Ann Hist Comput",
  ["annals of the icrp"] = "Ann ICRP",
  ["annals of the institute of statistical mathematics"] = "Ann. Inst. Statist. Math.",
  ["annals of the international communication association"] = "Ann Int Commun Assoc",
  ["annals of the international society of dynamic games"] = "Ann. Internat. Soc. Dynam. Games",
  ["annals of the israel physical society"] = "Ann. Israel Phys. Soc.",
  ["annals of the japan association for philosophy of science"] = "Ann. Japan Assoc. Philos. Sci.",
  ["annals of the medical section of the polish academy of sciences"] = "Ann Med Sect Pol Acad Sci",
  ["annals of the medical university, bialystok, poland"] = "Ann. Med. Univ. Bialyst. Pol.",
  ["annals of the medical university, białystok, poland = roczniki akademii medycznej w białymstoku"] = "Ann Med Univ Bialyst Pol",
  ["annals of the missouri botanical garden"] = "Ann. Mo. Bot. Gard.",
  ["annals of the missouri botanical garden. missouri botanical garden"] = "Ann Mo Bot Gard",
  ["annals of the national academy of medical sciences (india)"] = "Ann Natl Acad Med Sci",
  ["annals of the national association of geographers, india"] = "Ann Natl Assoc Geogr",
  ["annals of the new york academy of dentistry"] = "Annals. N. Y. Acad. Dent.",
  ["annals of the new york academy of sciences"] = "Ann N Y Acad Sci",
  ["annals of the rheumatic diseases"] = "Ann Rheum Dis",
  ["annals of the romanian society for cell biology"] = "Ann Rom Soc Cell Biol",
  ["annals of the royal australasian college of dental surgeons"] = "Ann R Australas Coll Dent Surg",
  ["annals of the royal college of surgeons of england"] = "Ann R Coll Surg Engl",
  ["annals of the social science council of nigeria"] = "Ann Soc Sci Counc Niger",
  ["annals of the tiberiu popoviciu seminar of functional equations, approximation and convexity"] = "Ann. Tiberiu Popoviciu Semin. Funct. Equ. Approx. Convexity",
  ["annals of the university of bucharest. mathematical series"] = "Ann. Univ. Buchar. Math. Ser.",
  ["annals of thoracic and cardiovascular surgery"] = "Ann. Thorac. Cardiovasc. Surg.",
  ["annals of thoracic and cardiovascular surgery : official journal of the association of thoracic and cardiovascular surgeons of asia"] = "Ann Thorac Cardiovasc Surg",
  ["annals of thoracic medicine"] = "Ann Thorac Med",
  ["annals of thoracic surgery"] = "Ann. Thorac. Surg.",
  ["annals of thyroid"] = "Ann Thyroid",
  ["annals of thyroid research"] = "Ann Thyroid Res",
  ["annals of tourism research"] = "Ann Tour Res",
  ["annals of translational medicine"] = "Ann Transl Med",
  ["annals of translational medicine & epidemiology"] = "Ann Transl Med Epidemiol",
  ["annals of transplantation"] = "Ann Transplant",
  ["annals of transplantation : quarterly of the polish transplantation society"] = "Ann Transplant",
  ["annals of trauma and acute care"] = "Ann Trauma Acute Care",
  ["annals of tropical medicine and parasitology"] = "Ann Trop Med Parasitol",
  ["annals of tropical medicine and public health"] = "Ann Trop Med Public Health",
  ["annals of tropical paediatrics"] = "Ann Trop Paediatr",
  ["annals of tuberculosis"] = "Ann Tuberc",
  ["annals of vascular diseases"] = "Ann Vasc Dis",
  ["annals of vascular medicine and research"] = "Ann Vasc Med Res",
  ["annals of vascular surgery"] = "Ann Vasc Surg",
  ["annals of virology and research"] = "Ann Virol Res",
  ["annals of western medicine and surgery"] = "Ann West Med Surg",
  ["annals of work exposures and health"] = "Ann Work Expo Health",
  ["annals of wyoming (cheyenne, wyo. : 1925)"] = "Ann Wyo",
  ["annals. orgone institute"] = "Ann Orgone Inst",
  ["annee biologique"] = "Annee Biol.",
  ["annee endocrinologique"] = "Annee Endocrinol.",
  ["annee psychologique"] = "Annee Psychol.",
  ["annee therapeutique et clinique en ophtalmologie"] = "Annee Ther. Clin. Ophtalmol.",
  ["anneli lax new mathematical library"] = "Anneli Lax New Math. Libr.",
  ["annnali della facoltà di lettere e filosofia dell’università di cagliari"] = "AnnCagl",
  ["annotated laws of massachusetts : containing all the laws of massachusetts of a general and permanent nature / completely annotated by the editorial staff of the publisher. massachusetts"] = "Annot Laws Mass Mass",
  ["annotated laws of massachusetts : containing all the laws of massachusetts of a general and permanent nature. massachusetts"] = "Annot Laws Mass",
  ["annotation : the newsletter of the national historical publications and records commission"] = "Annotation",
  ["annotationes zoologicae japonenses"] = "Annot Zool Jpn",
  ["annotazioni numismatiche"] = "AnnotNum",
  ["annuaire - facultes catholiques de lille (1972)"] = "Fac Cathol Lille",
  ["annuaire - facultés catholiques de lille (1972)"] = "Fac Cathol Lille",
  ["annuaire de l'afrique du nord"] = "Annu Afr Nord",
  ["annuaire de l’afrique du nord"] = "Annu Afr Nord",
  ["annuaire de l’institut de philologie et d’histoire orientales et slaves (université libre de bruxelles)"] = "AIPhOr",
  ["annuaire de l’institut de philologie et d’histoire orientales et slaves de l’université libre de bruxelles"] = "AIPhO",
  ["annuaire de l’université de sofia, faculté des lettres"] = "AUS",
  ["annuaire de l’école pratique des hautes études, iv’ sect., sciences hist. & philol"] = "AEHE IV",
  ["annuaire de l’école pratique des hautes études, ve sect., sciences religieuses"] = "AEHE V",
  ["annuaire des amis de la bibliothèque de sélestat"] = "AABS",
  ["annuaire des pays de l'océan indien"] = "Annu Pays Ocean Indien",
  ["annuaire des pays de l’ocean indien"] = "Annu Pays Ocean Indien",
  ["annuaire du collège de france"] = "ACF",
  ["annuaire statistique de la belgique et du congo belge"] = "Annu Stat Belg Congo Belge",
  ["annual (report). worcester art museum"] = "AnnWorcArtMus",
  ["annual book of astm standards. section 11, water and environmental technology. astm committee e-35 on pesticides"] = "Annu Book ASTM Stand Sect 11 Water Environ Technol",
  ["annual byzantine studies conference. abstracts of papers"] = "AnnByzConf",
  ["annual conference & exposition : final program and proceedings. american society for engineering education"] = "Annu Conf Expo (Am Soc Eng Educ)",
  ["annual conference / australasian society for hiv medicine"] = "Annu Conf Australas Soc HIV Med",
  ["annual conference on research in medical education"] = "Annu. Conf. Res. Med. Educ.",
  ["annual conference on research in medical education. conference on research in medical education"] = "Annu Conf Res Med Educ",
  ["annual egyptological bibliography"] = "AnnEgBibl",
  ["annual international conference of the ieee engineering in medicine and biology society. ieee engineering in medicine and biology society. annual international conference"] = "Annu Int Conf IEEE Eng Med Biol Soc",
  ["annual international conference syiah kuala university (aic-unsyiah)"] = "Annu Int Conf Syiah Kuala Univ",
  ["annual meeting & exhibition proceedings cd-rom. air & waste management association. meeting"] = "Proc Air Waste Manage Assoc Meet",
  ["annual meeting - american institute of oral biology"] = "Annu Meet Am Inst Oral Biol",
  ["annual meeting of the american climatological association. american climatological association. annual meeting"] = "Annu Meet Amer Climatol Assoc",
  ["annual meeting of the florida state horticultural society. florida state horticultural society. meeting"] = "Annu Meet Fla State Hort Soc",
  ["annual meeting, american institute of oral biology"] = "Annu. Meet. Am. Inst. Oral Biol.",
  ["annual of the british school at athens"] = "ABSA",
  ["annual of the department of antiquities of jordan"] = "AAJ",
  ["annual ornl biomedical science and engineering center conference. ornl biomedical science and engineering center conference"] = "Annu ORNL Biomed Sci Eng Cent Conf",
  ["annual proceedings / association for the advancement of automotive medicine. association for the advancement of automotive medicine"] = "Annu Proc Assoc Adv Automot Med",
  ["annual proceedings. association for the advancement of automotive medicine"] = "Annu Proc Assoc Adv Automot Med",
  ["annual progress in reproductive medicine"] = "Annu Prog Reprod Med",
  ["annual progress report - university of california, laboratory of nuclear medicine and radiation biology"] = "Annu Prog Rep Univ Calif Lab Nucl Radiat Biol",
  ["annual report - division of biological and medical research, argonne national laboratory. argonne national laboratory. division of biological and medical research"] = "Annu Rep Div Biol Med Res Argonne Natl Lab",
  ["annual report of council. british school of archaeology at athens"] = "AnnRepBSA",
  ["annual report of the bean improvement cooperative. bean improvement cooperative"] = "Annu Rep Bean Improv Coop",
  ["annual report of the board of regents of the smithsonian institution. smithsonian institution. board of regents"] = "Annu Rep Board Regents Smithson Inst",
  ["annual report of the council. royal society of health (great britain)"] = "Annu Rep R Soc Promot Health",
  ["annual report of the department of antiquities, republic of cyprus"] = "AnnRepCypr",
  ["annual report of the gorgas memorial laboratory and report of audit. gorgas memorial laboratory"] = "Annu Rep Gorgas Meml Lab Rep Audit",
  ["annual report of the research institute of environmental medicine, nagoya university"] = "Annu Rep Res Inst Environ Med Nagoya Univ",
  ["annual report. british guiana. mosquito control service"] = "Annu Rep British Guiana",
  ["annual report. louisiana. department of public welfare"] = "Annu Rep La Dep Public Welf",
  ["annual report. new york (state). department of health. division of laboratories and research"] = "Annu Rep N Y State  Dept Health Division Lab Res",
  ["annual report. rhode island. department of public welfare"] = "Annu Rep R I",
  ["annual report. state medical society of wisconsin"] = "State Med Soc Annual Rep",
  ["annual report. the society for libyan studies"] = "RepSocLibSt",
  ["annual report. united states national museum"] = "Annu Rep United States Natl Mus",
  ["annual reports - royal society of chemistry, section section a: inorganic chemistry"] = "Annu. Rep. R. Soc. Chem. Sect. A: Inorg. Chem.",
  ["annual reports - royal society of chemistry, section section b: organic chemistry"] = "Annu. Rep. R. Soc. Chem. Sect. B. Org. Chem.",
  ["annual reports - royal society of chemistry, section section c: physical chemistry"] = "Annu. Rep. R. Soc. Chem. Sect. C. Phys. Chem.",
  ["annual reports in computational chemistry"] = "Annu Rep Comput Chem",
  ["annual reports in medicinal chemistry"] = "Annu Rep Med Chem",
  ["annual reports on analytical atomic spectroscopy"] = "Annu. Rep. Anal. At. Spectrosc.",
  ["annual reports on nmr spectroscopy"] = "Annu Rep NMR Spectrosc",
  ["annual reports on the progress of chemistry, section a: inorganic chemistry"] = "Annu. Rep. Prog. Chem. Sect. A: Inorg. Chem.",
  ["annual reports on the progress of chemistry, section b: organic chemistry"] = "Annu. Rep. Prog. Chem. Sect. B: Org. Chem.",
  ["annual reports on the progress of chemistry, section c: physical chemistry"] = "Annu. Rep. Prog. Chem. Sect. C: Phys. Chem.",
  ["annual research & review in biology"] = "Annu Res Rev Biol",
  ["annual research briefs. center for turbulence research (u.s.)"] = "Annu Res Br",
  ["annual review in automatic programming"] = "Ann. Rev. Automat. Programming",
  ["annual review in control"] = "Annu. Rev. Control",
  ["annual review of analytical chemistry"] = "Annu. Rev. Anal. Chem.",
  ["annual review of analytical chemistry (palo alto, calif.)"] = "Annu Rev Anal Chem (Palo Alto Calif)",
  ["annual review of animal biosciences"] = "Annu Rev Anim Biosci",
  ["annual review of anthropology"] = "Annu Rev Anthropol",
  ["annual review of applied linguistics"] = "Annu Rev Appl Linguist",
  ["annual review of astronomy and astrophysics"] = "Annu Rev Astron Astrophys",
  ["annual review of atomic and molecular physics"] = "Annu. Rev. At. Mol. Phys.",
  ["annual review of biochemistry"] = "Annu Rev Biochem",
  ["annual review of biomedical data science"] = "Annu Rev Biomed Data Sci",
  ["annual review of biomedical engineering"] = "Annu Rev Biomed Eng",
  ["annual review of biophysics"] = "Annu Rev Biophys",
  ["annual review of biophysics and bioengineering"] = "Annu Rev Biophys Bioeng",
  ["annual review of biophysics and biomolecular structure"] = "Annu Rev Biophys Biomol Struct",
  ["annual review of biophysics and biophysical chemistry"] = "Annu Rev Biophys Biophys Chem",
  ["annual review of cancer biology"] = "Annu Rev Cancer Biol",
  ["annual review of cell and developmental biology"] = "Annu Rev Cell Dev Biol",
  ["annual review of cell biology"] = "Annu Rev Cell Biol",
  ["annual review of chemical and biomolecular engineering"] = "Annu Rev Chem Biomol Eng",
  ["annual review of clinical psychology"] = "Annu Rev Clin Psychol",
  ["annual review of computer science"] = "Annu. Rev. Comput. Sci.",
  ["annual review of condensed matter physics"] = "Annu. Rev. Condens. Matter Phys.",
  ["annual review of control, robotics, and autonomous systems"] = "Annu Rev Control Robot Auton Syst",
  ["annual review of criminology"] = "Annu Rev Criminol",
  ["annual review of developmental psychology"] = "Annu Rev Dev Psychol",
  ["annual review of earth and planetary science letters"] = "Annu. Rev. Earth Planet. Sci. Lett.",
  ["annual review of earth and planetary sciences"] = "Annu. Rev. Earth Planet. Sci.",
  ["annual review of ecology and systematics"] = "Annu Rev Ecol Syst",
  ["annual review of ecology, evolution, and systematics"] = "Annu. Rev. Ecol. Evol. Syst.",
  ["annual review of economics"] = "Annu Rev Econom",
  ["annual review of energy and the environment"] = "Annu. Rev. Energy Env.",
  ["annual review of entomology"] = "Annu Rev Entomol",
  ["annual review of environment and resources"] = "Annu Rev Environ Resour",
  ["annual review of fish diseases"] = "Annu Rev Fish Dis",
  ["annual review of fluid mechanics"] = "Annu. Rev. Fluid Mech.",
  ["annual review of food science and technology"] = "Annu Rev Food Sci Technol",
  ["annual review of genetics"] = "Annu Rev Genet",
  ["annual review of genomics and human genetics"] = "Annu Rev Genomics Hum Genet",
  ["annual review of gerontology & geriatrics"] = "Annu Rev Gerontol Geriatr",
  ["annual review of gerontology and geriatrics"] = "Annu. Rev. Gerontol. Geriatr.",
  ["annual review of heat transfer"] = "Annu. Rev. Heat Transfer",
  ["annual review of immunology"] = "Annu Rev Immunol",
  ["annual review of law and social science"] = "Annu Rev Law Soc Sci",
  ["annual review of linguistics"] = "Annu Rev Linguist",
  ["annual review of marine science"] = "Ann Rev Mar Sci",
  ["annual review of materials research"] = "Annu. Rev. Mater. Res.",
  ["annual review of materials science"] = "Annu. Rev. Mater. Sci.",
  ["annual review of medicine"] = "Annu Rev Med",
  ["annual review of microbiology"] = "Annu Rev Microbiol",
  ["annual review of neuroscience"] = "Annu Rev Neurosci",
  ["annual review of nuclear and particle science"] = "Annu. Rev. Nucl. Part. Sci.",
  ["annual review of nuclear science"] = "Annu Rev Nucl Sci",
  ["annual review of nursing research"] = "Annu Rev Nurs Res",
  ["annual review of nutrition"] = "Annu Rev Nutr",
  ["annual review of pathology"] = "Annu Rev Pathol",
  ["annual review of pathology: mechanisms of disease"] = "Annu. Rev. Pathol.: Pathol. Mech. Dis.",
  ["annual review of pharmacology"] = "Annu Rev Pharmacol",
  ["annual review of pharmacology and toxicology"] = "Annu Rev Pharmacol Toxicol",
  ["annual review of physical chemistry"] = "Annu Rev Phys Chem",
  ["annual review of physiology"] = "Annu Rev Physiol",
  ["annual review of phytopathology"] = "Annu Rev Phytopathol",
  ["annual review of plant biology"] = "Annu Rev Plant Biol",
  ["annual review of plant physiology"] = "Annu Rev Plant Physiol",
  ["annual review of plant physiology and plant molecular biology"] = "Annu Rev Plant Physiol Plant Mol Biol",
  ["annual review of political science (palo alto, calif.)"] = "Annu Rev Polit Sci (Palo Alto)",
  ["annual review of population law"] = "Annu Rev Popul Law",
  ["annual review of psychology"] = "Annu Rev Psychol",
  ["annual review of public health"] = "Annu Rev Public Health",
  ["annual review of rehabilitation"] = "Annu Rev Rehabil",
  ["annual review of resource economics"] = "Annu Rev Resour Economics",
  ["annual review of sex research"] = "Annu Rev Sex Res",
  ["annual review of sociology"] = "Annu Rev Sociol",
  ["annual review of statistics and its application"] = "Annu. Rev. Stat. Appl.",
  ["annual review of virology"] = "Annu Rev Virol",
  ["annual review of vision science"] = "Annu Rev Vis Sci",
  ["annual reviews in control"] = "Annu. Rev. Control",
  ["annual reviews in modern quantitative finance. including current aspects of fintech, risk and investments"] = "Annu. Rev. Mod. Quant. Financ.",
  ["annual reviews of computational physics"] = "Annual Rev. Comput. Phys.",
  ["annual survey of american law"] = "Annu Surv Am Law",
  ["annual symposium on nursing faculty practice"] = "Annu Symp Nurs Fac Pract",
  ["annuale mediaevale"] = "Annu Mediaev",
  ["annuario bibliografico di archeologia. modena"] = "AnnBiblAModena",
  ["annuario bibliografico di archeologia. nuove accessioni del \\ldots. biblioteca dell’istituto nazionale di archeologia e storia dell’arte, roma"] = "AnnBiblARom",
  ["annuario del ginnasio liceo a. volta di como"] = "AGLComo",
  ["annuario della accademia delle scienze di torino"] = "AnnAcTorino",
  ["annuario della accademia nazionale dei lincei"] = "AnnuarioAcLinc",
  ["annuario della pontificia accademia romana di archeologia"] = "AnnPontAcRom",
  ["annuario della scuola archeologica di atene a delle missioni italiane in oriente"] = "ASAA",
  ["annuario della scuola archeologica di atene e delle missioni italiane in oriente"] = "ASAtene",
  ["annuario dell’accademia etrusca di cortona"] = "AAEC",
  ["annuario statistico dell'assistenza e della previdenza sociale. istituto centrale di statistica (italy)"] = "Annu Stat Assist Previd Soc",
  ["annuario statistico dell’assistenza e della previdenza sociale. istituto centrale di statistica (italy)"] = "Annu Stat Assist Previd Soc",
  ["annuario. accademia etrusca di cortona"] = "AnnAcEtr",
  ["annuario. istituto giapponese di cultura in roma"] = "AnnIstGiapp",
  ["annuario. liceo-ginnasio statale g. palmieri, lecce"] = "AnnuarioLecce",
  ["annuarium historiae conciliorum"] = "AHC",
  ["anodos. studies of ancient world"] = "Anodos",
  ["anotaciones pediatricas"] = "Anot Pediatr",
  ["anotaciones pediátricas"] = "Anot Pediatr",
  ["anphi papers"] = "ANPHI Pap",
  ["ans. advances in nursing science"] = "ANS Adv Nurs Sci",
  ["ans: advances in nursing science"] = "ANS Adv. Nurs. Sci.",
  ["anshu keji"] = "Anshu Keji",
  ["antaeus. communicationes ex instituto archaeologico academiae scientiarum hungaricae"] = "Antaeus",
  ["antarctic journal of the united states"] = "Antarct J US",
  ["antarctic journal of the united states / national science foundation"] = "Antarct J US",
  ["antarctic record"] = "Antarct. Rec.",
  ["antarctic research series"] = "Antarct Res Ser",
  ["antarctic science"] = "Antarct. Sci.",
  ["antarctic science / blackwell scientific publications"] = "Antarct Sci",
  ["antenor (paris)"] = "Antenor (Paris)",
  ["anthem studies in wittgenstein"] = "Anthem Stud. Wittgenstein",
  ["anthropocene coasts"] = "Anthropocene Coasts",
  ["anthropocene review"] = "Anthropocene Rev.",
  ["anthropocene science"] = "Anthropocene Sci.",
  ["anthropologiai kozlemenyek"] = "Anthropol Kozl",
  ["anthropologiai közlemények"] = "Anthropol Kozl",
  ["anthropological forum"] = "Anthropol Forum",
  ["anthropological papers of the american museum of natural history"] = "Anthropol Pap Am Mus Nat Hist",
  ["anthropological quarterly"] = "Anthropol Q",
  ["anthropological science"] = "Anthropol. Sci.",
  ["anthropological science : journal of the anthropological society of nippon"] = "Jinruigaku zasshi|Anthropol. Sci.",
  ["anthropological science : journal of the anthropological society of nippon = jinruigaku zasshi"] = "Anthropol Sci",
  ["anthropologie et societes"] = "Anthropol Soc",
  ["anthropologie et sociétés"] = "Anthropol Soc",
  ["anthropologischer anzeiger"] = "Anthropol. Anz.",
  ["anthropologischer anzeiger; bericht uber die biologisch-anthropologische literatur"] = "Anthropol Anz",
  ["anthropologischer anzeiger; bericht über die biologisch-anthropologische literatur"] = "Anthropol Anz",
  ["anthropology & archeology of eurasia"] = "Anthropol Archeol Eurasia",
  ["anthropology & education quarterly"] = "Anthropol Educ Q",
  ["anthropology & medicine"] = "Anthropol Med",
  ["anthropology and philosophy"] = "Anthropol Philos",
  ["anthropology in action : newsletter of the british association for social anthropology in policy and practice (basapp)"] = "Anthropol Action",
  ["anthropology now"] = "Anthropol Now",
  ["anthropology of consciousness"] = "Anthropol Conscious",
  ["anthropology of work review : awr"] = "Anthropol Work Rev",
  ["anthropology today"] = "Anthropol Today",
  ["anthropology u.c.l.a"] = "Anthropol UCLA",
  ["anthropos: studie z oboru anthropologie, paleoethnologie, paleontologie a kvartérní geologie"] = "Anthropos",
  ["anthropotes / istituto giovanni paolo ii per studi su matrimonio e famiglia"] = "Anthropotes",
  ["anthroquest : the l.s.b. leakey foundation news"] = "Anthroquest",
  ["anti-cancer agents in medicinal chemistry"] = "Anticancer Agents Med Chem",
  ["anti-cancer drug design"] = "Anticancer Drug Des",
  ["anti-cancer drugs"] = "Anticancer Drugs",
  ["anti-corrosion methods and materials"] = "Anti-Corros. Methods Mater.",
  ["anti-infective agents"] = "Anti-Infect. Agents",
  ["anti-infective agents in medicinal chemistry"] = "Anti-Infect. Agents Med. Chem.",
  ["anti-inflammatory & anti-allergy agents in medicinal chemistry"] = "Antiinflamm Antiallergy Agents Med Chem",
  ["antibiotic medicine & clinical therapy"] = "Antibiotic Med Clin Ther",
  ["antibiotic medicine & clinical therapy (new york, ny)"] = "Antibiotic Med Clin Ther (New York)",
  ["antibiotica et chemotherapia. fortschritte. advances. progres"] = "Antibiot Chemother",
  ["antibiotica et chemotherapia. fortschritte. advances. progrès"] = "Antibiot Chemother",
  ["antibiotica. quaderni"] = "Antibiotica. [Quad.]",
  ["antibiotics & chemotherapy (northfield, ill.)"] = "Antibiot Chemother (Northfield)",
  ["antibiotics (basel, switzerland)"] = "Antibiotics (Basel)",
  ["antibiotics and chemotherapy"] = "Antibiot Chemother",
  ["antibiotics annual"] = "Antibiot Annu",
  ["antibiotiki i khimioterapiia"] = "Antibiot. Khimioter.",
  ["antibiotiki i khimioterapii︠a︡ = antibiotics and chemoterapy [sic]"] = "Antibiot Khimioter",
  ["antibiotiki i khimioterapiya"] = "Antibiot. Khimioter.",
  ["antibiotiki i meditsinskaia biotekhnologiia"] = "Antibiot. Med. Biotekhnol.",
  ["antibiotiki i meditsinskaia biotekhnologiia = antibiotics and medical biotechnology"] = "Antibiot Med Biotekhnol",
  ["antibiotiques (paris, france : 1999)"] = "Antibiotiques (Paris)",
  ["antibodies (basel, switzerland)"] = "Antibodies (Basel)",
  ["antibody therapeutics"] = "Antib Ther",
  ["anticancer research"] = "Anticancer Res.",
  ["antichità altoadriatiche"] = "AAAD",
  ["antichità pisane"] = "AntPisa",
  ["antichthon. journal of the australian society for classical studies"] = "Antichthon",
  ["anticipation science"] = "Anticip. Sci.",
  ["anticnaja drevnost’i srednije veka (l’antiquité et le moyen âge)"] = "ADSV",
  ["anticnyj mir i arkheologija (le monde antique et l’archéologie)"] = "AMA",
  ["antigüedad y cristianismo. monografías históricas sobre la antigüedad tardía"] = "AntigCr",
  ["antike denkmäler"] = "AD",
  ["antike gemmen in deutschen sammlungen"] = "AGD",
  ["antike kunst"] = "AntK",
  ["antike kunst, hrsg. von der vereiningung der freunde antiker kunst"] = "AK",
  ["antike münzen und geschnittene steine"] = "AMuGS",
  ["antike plastik"] = "AntPl",
  ["antike und abendland"] = "AuA",
  ["antike und abendland: beiträge zum verständnis der griechen und römer und ihre nachlebens"] = "A&A",
  ["antike und christentum"] = "AntChr",
  ["antike welt"] = "Antike Welt",
  ["antike welt. zeitschrift für archäologie und kulturgeschichte"] = "AW",
  ["antimicrobial agents and chemotherapy"] = "Antimicrob Agents Chemother",
  ["antimicrobial combination devices"] = "Antimicrob Comb Devices (2019)",
  ["antimicrobial resistance & infection control"] = "Antimicrob. Resist. Infect. Control",
  ["antimicrobial resistance and infection control"] = "Antimicrob Resist Infect Control",
  ["antimicrobial stewardship and healthcare epidemiology"] = "Antimicrob. Stewardship Healthcare Epidemiol.",
  ["antioquia medica"] = "Antioquia Med.",
  ["antioquia médica"] = "Antioquia Med",
  ["antioxidants & redox signaling"] = "Antioxid Redox Signal",
  ["antioxidants (basel, switzerland)"] = "Antioxidants (Basel)",
  ["antioxidants and redox signalling"] = "Antioxid. Redox. Signal.",
  ["antipolis. a journal of mediterranean archaeology"] = "Antipolis",
  ["antiqua. rivista dell’archeoclub d’italia"] = "Antiqua",
  ["antiquitates mathematicae. roczniki polskiego towarzystwa matematycznego. seria vi"] = "Antiq. Math.",
  ["antiquity and survival"] = "AntSurv",
  ["antiquity. a quarterly review of archaeology"] = "Antiquity",
  ["antiquité tardive. revue internationale d’histoire et d’archéologie"] = "AntTard",
  ["antiquités africaines"] = "AntAfr",
  ["antiquités nationales. saint-germain-en-laye"] = "AntNat",
  ["antisense & nucleic acid drug development"] = "Antisense Nucleic Acid Drug Dev",
  ["antisense and nucleic acid drug development"] = "Antisense Nucleic Acid Drug Dev.",
  ["antisense research and development"] = "Antisense Res Dev",
  ["antitrust bulletin"] = "Antitrust Bull.",
  ["antitrust law and economics review"] = "Antitrust Law Econ.",
  ["antiviral chemistry & chemotherapy"] = "Antivir Chem Chemother",
  ["antiviral chemistry and chemotherapy"] = "Antiviral Chem. Chemother.",
  ["antiviral research"] = "Antiviral Res",
  ["antiviral therapy"] = "Antivir Ther",
  ["antologia medica italiana. archivio"] = "Antol Med Ital Arch",
  ["antonie van leeuwenhoek"] = "Antonie Van Leeuwenhoek",
  ["antropologia (rome, italy)"] = "Antropologia (Rome)",
  ["anu historical journal"] = "ANU Hist J",
  ["anuari de filologia (universitat de barcelona)"] = "Anu. Filol. Univ. Barc.",
  ["anuari de filologia, secció d, studia graeca et latina"] = "AFB",
  ["anuario arqueológico de andalucía"] = "AnArqAnd",
  ["anuario brasileiro de odontologia"] = "Anu Bras Odontol",
  ["anuario de estudios americanos"] = "Anu Estud Am",
  ["anuario de estudios atlanticos"] = "Anu Estud Atl",
  ["anuario de estudios atlánticos"] = "Anu Estud Atl",
  ["anuario de estudios centroamericanos"] = "Anu Estud Centroam",
  ["anuario de estudios medievales"] = "Anu Estud Mediev",
  ["anuario de historia del derecho español"] = "AHDE",
  ["anuario de investigación en adicciones"] = "Anu Investig Adicciones",
  ["anuario estadístico del perú. peru. dirección nacional de estadística"] = "Anu Estad Peru",
  ["anuario iehs"] = "Anu IEHS",
  ["anuario indigenista"] = "Anu Indig",
  ["anwendungsorientierte mathematik"] = "Anwend.orientier. Math.",
  ["anwendungsorientierte statistik"] = "Anwend.orientier. Stat.",
  ["anxiety, stress, and coping"] = "Anxiety Stress Coping",
  ["anz journal of surgery"] = "ANZ J Surg",
  ["anzeiger (osterreichische akademie der wissenschaften. philosophisch-historische klasse)"] = "Anz Philol Hist Kl Osterr Akad Wiss",
  ["anzeiger (österreichische akademie der wissenschaften. philosophisch-historische klasse)"] = "Anz Philol Hist Kl Osterr Akad Wiss",
  ["anzeiger der österreichischen akademie der wissenschaften in wien, philos.-hist. klasse"] = "AAWW",
  ["anzeiger fur die altertumswissenschaft"] = "Anz Altertwiss",
  ["anzeiger für die altertumswissenschaft"] = "Anz Altertwiss",
  ["anzeiger für die altertumswissenschaft"] = "AnzAW",
  ["anzeiger für die altertumswissenschaft, hrsg. von der österreichischen humanistischen gesellschaft"] = "AAHG",
  ["anzeiger für schädlingskunde, pflanzenschutz, umweltschutz"] = "Anz. Schädl.kd. Pflanzenschutz Umweltschutz",
  ["anzeiger. osterreichische akademie der wissenschaften. philosophisch-historiche klasse"] = "Anz. Osterr. Akad. Wiss. Philos.-Hist. Kl.",
  ["anzeiger. österreichische akademie der wissenschaften, philosophisch-historische klasse"] = "AnzWien",
  ["anzhes journal. australian and new zealand history of education society"] = "Anzhes J",
  ["anziam journal"] = "ANZIAM J.",
  ["aob plants"] = "AoB Plants",
  ["aoha : a publication of the american osteopathic hospital association"] = "AOHA",
  ["aoha progress"] = "AOHA Prog.",
  ["aoha today"] = "AOHA Today",
  ["aorn journal"] = "AORN J",
  ["aorta (stamford, conn.)"] = "Aorta (Stamford)",
  ["apa newsletters"] = "APA Newsl",
  ["apacph-clearing house bulletin"] = "APACPH Clear House Bull",
  ["aparato respiratorio y tuberculosis"] = "Apar Respir Tuberc",
  ["apcbee procedia"] = "APCBEE Procedia",
  ["apdi newsletter. united nations asian and pacific development institute"] = "APDI Newsl",
  ["apeiron (clayton, vic.)"] = "Apeiron",
  ["apeiron (montreal, quebec)"] = "Apeiron",
  ["aperito journal of cellular and molecular biology"] = "Aperito J Cell Mol Biol",
  ["apl bioengineering"] = "APL Bioeng",
  ["apl computational physics"] = "APL Comput. Phys.",
  ["apl electronic devices"] = "APL Electron. Devices",
  ["apl energy"] = "APL Energy",
  ["apl machine learning"] = "APL Mach. Learn.",
  ["apl materials"] = "APL Mater.",
  ["apl photonics"] = "APL Photonics",
  ["apl quantum"] = "APL Quantum",
  ["apl: organic electronics and photonics"] = "APL: Org. Electron. Photonics",
  ["apmis : acta pathologica, microbiologica, et immunologica scandinavica"] = "APMIS",
  ["apmis. supplementum"] = "APMIS Suppl",
  ["apollo. bolletino dei musei provinciali del salernitano"] = "Apollo",
  ["apollo. the international magazine of the arts"] = "ApolloLond",
  ["apoptosis : an international journal on programmed cell death"] = "Apoptosis",
  ["aportaciones matematicas"] = "Aportaciones Mat.",
  ["aportaciones matematicas: comunicaciones"] = "Aportaciones Mat. Comun.",
  ["aportaciones matematicas: investigacion"] = "Aportaciones Mat. Investig.",
  ["aportaciones matematicas: textos"] = "Aportaciones Mat. Textos",
  ["aportaciones matemáticas: comunicaciones"] = "Aportaciones Mat. Comun.",
  ["aportaciones matemáticas: notas de investigación"] = "Aportaciones Mat. Notas Investigación",
  ["aportaciones matemáticas: textos"] = "Aportaciones Mat. Textos",
  ["apotheker und kunst"] = "Apoth Kunst",
  ["appalachian journal"] = "Appalach J",
  ["appea journal"] = "APPEA J.",
  ["appita journal"] = "Appita J.",
  ["appled entomology and zoology"] = "Appl. Entomol. Zool.",
  ["applicable algebra in engineering, communication and computing"] = "Appl. Algebra Eng. Commun. Comput.",
  ["applicable analysis"] = "Appl Anal",
  ["applicable analysis and discrete mathematics"] = "Appl. Anal. Discrete Math.",
  ["applicable analysis. an international journal"] = "Appl. Anal.",
  ["application of clinical genetics"] = "Appl. Clin. Genet.",
  ["application of mathematics"] = "Appl. Math.",
  ["applicationes mathematicae"] = "Appl. Math. (Warsaw)",
  ["applications and applied mathematics"] = "Appl. Appl. Math.",
  ["applications and applied mathematics : an international journal"] = "Appl Appl Math",
  ["applications and applied mathematics. an international journal (aam)"] = "Appl. Appl. Math.",
  ["applications in energy and combustion science"] = "Appl. Energy Combust. Sci.",
  ["applications in engineering science"] = "Appl. Eng. Sci.",
  ["applications in plant sciences"] = "Appl Plant Sci",
  ["applications of communications theory"] = "Appl. Commun. Theory",
  ["applications of evolutionary computation : 17th european conference, evoapplications 2014, granada, spain, april 23-25, 2014 : revised selected papers. evoapplications (conference) (17th : 2014 : granada, spain)"] = "Appl Evol Comput (2014)",
  ["applications of management science"] = "Appl. Manag. Sci.",
  ["applications of mathematics"] = "Appl. Math.",
  ["applications of mathematics (prague, czechoslovakia)"] = "Appl Math (Prague)",
  ["applications of modern technology in business"] = "Appl. Modern Tech. Business",
  ["applications of nmr spectroscopy"] = "Appl NMR Spectrosc",
  ["applications of surface science"] = "Appl. Surf. Sci.",
  ["applied & preventive psychology : journal of the american association of applied and preventive psychology"] = "Appl Prev Psychol",
  ["applied & translational genomics"] = "Appl Transl Genom",
  ["applied accoustics"] = "Appl. Acoust.",
  ["applied acoustics. acoustique appliqué. angewandte akustik"] = "Appl Acoust",
  ["applied adhesion science"] = "Appl. Adhes. Sci.",
  ["applied ai letters"] = "Appl. Ai Lett.",
  ["applied analysis and optimization"] = "Appl. Anal. Optim.",
  ["applied and computational harmonic analysis"] = "Appl. Comput. Harmon. Anal.",
  ["applied and computational harmonic analysis. time-frequency and time-scale analysis, wavelets, numerical algorithms, and applications"] = "Appl. Comput. Harmon. Anal.",
  ["applied and computational mathematics. an international journal"] = "Appl. Comput. Math.",
  ["applied and engineering mathematics series"] = "Appl. Eng. Math. Ser.",
  ["applied and environmental microbiology"] = "Appl Environ Microbiol",
  ["applied and environmental soil science"] = "Appl Environ Soil Sci",
  ["applied and numerical harmonic analysis"] = "Appl. Numer. Harmon. Anal.",
  ["applied and theoretical electrophoresis"] = "Appl. Theor. Electrophor.",
  ["applied and theoretical electrophoresis : the official journal of the international electrophoresis society"] = "Appl Theor Electrophor",
  ["applied and translational geonomics"] = "Appl. Transl. Geonomics",
  ["applied animal behaviour science"] = "Appl Anim Behav Sci",
  ["applied animal ethology"] = "Appl. Anim. Ethol.",
  ["applied animal science"] = "Appl. Anim. Sci.",
  ["applied anthropology"] = "Appl Anthropol",
  ["applied artifical intelligence"] = "Appl. Artif. Intell.",
  ["applied artificial intelligence : aai"] = "Appl Artif Intell",
  ["applied behavioral science review"] = "Appl Behav Sci Rev",
  ["applied biochemistry and biotechnology"] = "Appl. Biochem. Biotechnol.",
  ["applied biochemistry and microbiology"] = "Appl. Biochem. Microbiol.",
  ["applied biochemistry and microbiology (translation of prikladnaya biokhimiya i mikrobiologiya)"] = "Appl. Biochem. Microbiol.",
  ["applied bioinformatics"] = "Appl Bioinformatics",
  ["applied biological chemistry"] = "Appl Biol Chem",
  ["applied biological research"] = "Appl Biol Res",
  ["applied bionics and biomechanics"] = "Appl Bionics Biomech",
  ["applied biosafety"] = "Appl. Biosaf.",
  ["applied biosafety : journal of the american biological safety association"] = "Appl Biosaf",
  ["applied biotechnology, food science and policy"] = "Appl Biotechnol Food Sci Policy",
  ["applied cardiology"] = "Appl Cardiol",
  ["applied cardiopulmonary pathophysiology"] = "Appl. Cardiopulm. Pathophysiol.",
  ["applied cardiopulmonary pathophysiology : acp"] = "Appl Cardiopulm Pathophysiol",
  ["applied catalysis"] = "Appl Catal",
  ["applied catalysis a: general"] = "Appl. Catal., A",
  ["applied catalysis b: environment and energy"] = "Appl. Catal., B",
  ["applied catalysis b: environmental"] = "Appl. Catal., B",
  ["applied catalysis. a, general"] = "Appl Catal A Gen",
  ["applied catalysis. b, environmental"] = "Appl Catal B",
  ["applied categorical structures"] = "Appl. Categor. Struct.",
  ["applied categorical structures. a journal devoted to applications of categorical methods in algebra, analysis, computer science, logic, order and topology"] = "Appl. Categ. Structures",
  ["applied categorical structures. a journal devoted to applications of categorical methods in algebra, analysis, order, topology and computer science"] = "Appl. Categ. Structures",
  ["applied clay science"] = "Appl. Clay Sci.",
  ["applied clinical informatics"] = "Appl Clin Inform",
  ["applied clinical trials"] = "Appl Clin Trials",
  ["applied cognitive psychology"] = "Appl Cogn Psychol",
  ["applied composite materials"] = "Appl. Compos. Mater.",
  ["applied computational electromagnetics society journal"] = "Appl Comput Electromagn Soc J",
  ["applied computational intelligence and soft computing"] = "Appl. Comput. Intell. Soft Comput.",
  ["applied computer science. berichte zur praktischen informatik"] = "Appl. Comput. Sci.",
  ["applied computing and geosciences"] = "Appl. Comput. Geosci.",
  ["applied computing and informatics"] = "Appl. Comput. Inf.",
  ["applied cytogenetics : journal of the association of cytogenetic technologists"] = "Appl Cytogenet",
  ["applied demography"] = "Appl Demor",
  ["applied demography / population association of america, applied demography interest group"] = "Appl Demor",
  ["applied developmental science"] = "Appl Dev Sci",
  ["applied discrete mathematics and theoretical computer science"] = "Appl. Discrete Math. Theoret. Comput. Sci.",
  ["applied earth science"] = "Appl. Earth Sci.",
  ["applied ecology abstracts"] = "Appl. Ecol. Abstr.",
  ["applied ecology and environmental research"] = "Appl Ecol Environ Res",
  ["applied economic perspectives and policy"] = "Appl Econ Perspect Policy",
  ["applied economics"] = "Appl Econ",
  ["applied economics discussion paper series"] = "Appl Econ Discuss Pap Ser",
  ["applied economics discussion paper series / university of oxford, institute of economics and statistics"] = "Appl Econ Discuss Pap Ser",
  ["applied economics letters"] = "Appl Econ Lett",
  ["applied energy"] = "Appl. Energy",
  ["applied engineering in agriculture"] = "Appl Eng Agric",
  ["applied entomology and phytopathology"] = "Appl Entomol Phytopathol",
  ["applied entomology and zoology"] = "Appl Entomol Zool",
  ["applied environmental biotechnology"] = "Appl Environ Biotechnol",
  ["applied environmental education & communication"] = "Appl. Environ. Educ. Commun.",
  ["applied environmental education and communication (print)"] = "Appl Environ Educ Commun",
  ["applied ergonomics"] = "Appl Ergon",
  ["applied finance and accounting"] = "Appl Finance Account",
  ["applied financial economics"] = "Appl. Finan. Econ.",
  ["applied food biotechnology"] = "Appl. Food Biotechol.",
  ["applied food research"] = "Appl. Food Res.",
  ["applied general topology"] = "Appl. Gen. Topol.",
  ["applied genomics and proteomics"] = "Appl Genomics Proteomics",
  ["applied geochemistry"] = "Appl. Geochem.",
  ["applied geochemistry : journal of the international association of geochemistry and cosmochemistry"] = "Appl Geochem",
  ["applied geography"] = "Appl. Geogr.",
  ["applied geography (sevenoaks, england)"] = "Appl Geogr",
  ["applied geography and development"] = "Appl Geogr Dev",
  ["applied geomatics"] = "Appl. Geomatics",
  ["applied geophysics"] = "Appl. Geophys.",
  ["applied health economics and health policy"] = "Appl Health Econ Health Policy",
  ["applied human factors and ergonomics conference"] = "Appl Hum Factors Ergon Conf",
  ["applied human science"] = "Appl. Human Sci.",
  ["applied human science : journal of physiological anthropology"] = "Appl Human Sci",
  ["applied immunohistochemistry & molecular morphology : aimm"] = "Appl Immunohistochem Mol Morphol",
  ["applied immunohistochemistry & molecular morphology : aimm / official publication of the society for applied immunohistochemistry"] = "Appl Immunohistochem Mol Morphol",
  ["applied immunohistochemistry and molecular morphology"] = "Appl. Immunohistochem. Mol. Morphol.",
  ["applied in vitro toxicology"] = "Appl In Vitro Toxicol",
  ["applied informatics"] = "Appl Inform (Berl)",
  ["applied intelligence"] = "Appl. Intell.",
  ["applied linguistics"] = "Appl Linguist",
  ["applied logic series"] = "Appl. Log. Ser.",
  ["applied magnetic resonance"] = "Appl Magn Reson",
  ["applied materials today"] = "Appl. Mater. Today",
  ["applied mathematical finance"] = "Appl. Math. Finance",
  ["applied mathematical modelling"] = "Appl Math Model",
  ["applied mathematical modelling. simulation and computation for engineering and environmental systems"] = "Appl. Math. Model.",
  ["applied mathematical sciences"] = "Appl. Math. Sci.",
  ["applied mathematics"] = "Appl Math (Irvine)",
  ["applied mathematics & information sciences"] = "Appl. Math. Inf. Sci.",
  ["applied mathematics : a journal of chinese universities"] = "Appl Math",
  ["applied mathematics a journal of chinese universities"] = "Appl. Math. J. Chin. Univ.",
  ["applied mathematics and computation"] = "Appl. Math. Comput.",
  ["applied mathematics and computer science"] = "Appl. Math. Comput. Sci.",
  ["applied mathematics and engineering science texts"] = "Appl. Math. Eng. Sci. Texts",
  ["applied mathematics and mathematical computation"] = "Appl. Math. Math. Comput.",
  ["applied mathematics and mechanics"] = "Appl. Math. Mech.",
  ["applied mathematics and mechanics. english edition"] = "Appl. Math. Mech. (English Ed.)",
  ["applied mathematics and nonlinear sciences"] = "Appl. Math. Nonlinear Sci.",
  ["applied mathematics and optimization"] = "Appl. Math. Optim.",
  ["applied mathematics e-notes"] = "Appl. Math. E-Notes",
  ["applied mathematics for modern challenges"] = "Appl. Math. Mod. Challenges",
  ["applied mathematics in science and engineering"] = "Appl. Math. Sci. Eng.",
  ["applied mathematics letters"] = "Appl. Math. Lett.",
  ["applied mathematics letters. an international journal of rapid publication"] = "Appl. Math. Lett.",
  ["applied mathematics research express"] = "Appl. Math. Res. eXpress",
  ["applied mathematics research express : amrx"] = "Appl Math Res Express",
  ["applied mathematics, informatics and mechanics"] = "Appl. Math. Inform. Mech.",
  ["applied mathematics. a journal of chinese universities. ser. b"] = "Appl. Math. J. Chinese Univ. Ser. B",
  ["applied mathematics. a journal of chinese universities. series a. gaoxiao yingyong shuxue xuebao"] = "Appl. Math. J. Chinese Univ. Ser. A",
  ["applied mechanics"] = "Appl. Mech,",
  ["applied mechanics reviews"] = "Appl. Mech. Rev.",
  ["applied medical informatics"] = "Appl Med Inform",
  ["applied microbiology"] = "Appl Microbiol",
  ["applied microbiology (los angeles, ca)"] = "Appl Microbiol (Los Angel)",
  ["applied microbiology and biotechnology"] = "Appl. Microbiol. Biotechnol.",
  ["applied microgravity technology"] = "Appl Micrograv Technol",
  ["applied microscopy"] = "Appl Microsc",
  ["applied nanoscience"] = "Appl. Nanosci.",
  ["applied network science"] = "Appl Netw Sci",
  ["applied neurophysiology"] = "Appl Neurophysiol",
  ["applied neuropsychology"] = "Appl Neuropsychol",
  ["applied neuropsychology. adult"] = "Appl Neuropsychol Adult",
  ["applied neuropsychology. child"] = "Appl Neuropsychol Child",
  ["applied numerical  mathematics"] = "Appl. Numer. Math.",
  ["applied numerical analysis & computational mathematics"] = "Appl. Numer. Anal. Comput. Math.",
  ["applied numerical mathematics"] = "Appl. Numer. Math.",
  ["applied numerical mathematics : transactions of imacs"] = "Appl Numer Math",
  ["applied numerical mathematics. an imacs journal"] = "Appl. Numer. Math.",
  ["applied nursing research"] = "Appl. Nurs. Res.",
  ["applied nursing research : anr"] = "Appl Nurs Res",
  ["applied occupational and environmental hygiene"] = "Appl Occup Environ Hyg",
  ["applied ocean research"] = "Appl. Ocean Res.",
  ["applied ontology"] = "Appl Ontol",
  ["applied optics"] = "Appl. Opt.",
  ["applied optimization"] = "Appl. Optim.",
  ["applied organometallic chemistry"] = "Appl. Organomet. Chem.",
  ["applied parasitology"] = "Appl Parasitol",
  ["applied pathology"] = "Appl Pathol",
  ["applied petrochemical research"] = "Appl. Petrochem. Res.",
  ["applied philosophy (fort pierce, fla.)"] = "Appl Philos",
  ["applied phycology"] = "Appl. Phycol.",
  ["applied physics a"] = "Appl. Phys. A",
  ["applied physics a: materials science & processing"] = "Appl. Phys. Mater. Sci. Process.",
  ["applied physics a: solids and surfaces"] = "Appl. Phys. A",
  ["applied physics b"] = "Appl. Phys. B",
  ["applied physics b: lasers and optics"] = "Appl. Phys. B: Lasers Opt.",
  ["applied physics b: photophysics and laser chemistry"] = "Appl. Phys. B",
  ["applied physics express"] = "Appl. Phys. Express",
  ["applied physics letters"] = "Appl. Phys. Lett.",
  ["applied physics research"] = "Appl. Phys. Res.",
  ["applied physics reviews"] = "Appl. Phys. Rev.",
  ["applied physics. a, materials science & processing"] = "Appl Phys A Mater Sci Process",
  ["applied physics. b, lasers and optics"] = "Appl Phys B",
  ["applied physiology, nutrition and metabolism"] = "Appl. Physiol. Nutr. Metab.",
  ["applied physiology, nutrition, and metabolism"] = "Appl. Physiol. Nutr. Metab.",
  ["applied physiology, nutrition, and metabolism = physiologie appliquée, nutrition et métabolisme"] = "Appl Physiol Nutr Metab",
  ["applied polymer composites"] = "Appl. Polym. Compos.",
  ["applied polymer symposia"] = "Appl Polym Symp",
  ["applied probability"] = "Appl. Probab. Index",
  ["applied psycholinguistics"] = "Appl Psycholinguist",
  ["applied psychological measurement"] = "Appl Psychol Meas",
  ["applied psychology = psychologie appliquée"] = "Appl Psychol",
  ["applied psychology in criminal justice"] = "Appl Psychol Crim Justice",
  ["applied psychology. health and well-being"] = "Appl Psychol Health Well Being",
  ["applied psychophysiology and biofeedback"] = "Appl Psychophysiol Biofeedback",
  ["applied psychophysiology and biofeedback\tappl. psychophysiol. biofeedback"] = "Appl. Psychophysiol. Biofeedback",
  ["applied quantitative finance"] = "Appl. Quant. Finance",
  ["applied radiation and isotopes"] = "Appl. Radiat. Isot.",
  ["applied radiation and isotopes : including data, instrumentation and methods for use in agriculture, industry and medicine"] = "Appl Radiat Isot",
  ["applied radiation oncology"] = "Appl Radiat Oncol",
  ["applied radiology"] = "Appl Radiol",
  ["applied research"] = "Appl. Res.",
  ["applied research in mental retardation"] = "Appl Res Ment Retard",
  ["applied research in quality of life"] = "Appl Res Qual Life",
  ["applied rheology"] = "Appl. Rheol.",
  ["applied rheology (lappersdorf, germany : online)"] = "Appl Rheol",
  ["applied sciences (basel, switzerland)"] = "Appl Sci (Basel)",
  ["applied sciences (switzerland)"] = "Appl. Sci. Switz.",
  ["applied sciences and engineering"] = "Appl. Sci. Eng.",
  ["applied sciences-basel"] = "Appl. Sci.-Basel",
  ["applied sciences. monographs"] = "Appl. Sci. Monogr.",
  ["applied scientific research"] = "Appl. Sci. Res.",
  ["applied social research methods series"] = "Appl. Soc. Res. Methods Ser.",
  ["applied soft computing"] = "Appl. Soft Comput.",
  ["applied soil ecology"] = "Appl. Soil Ecol.",
  ["applied soil ecology : a section of agriculture, ecosystems & environment"] = "Appl Soil Ecol",
  ["applied solar energy"] = "Appl. Solar Energy",
  ["applied spatial analysis and policy"] = "Appl Spat Anal Policy",
  ["applied spectroscopy"] = "Appl. Spectrosc.",
  ["applied spectroscopy practica"] = "Appl. Spectrosc. Pract.",
  ["applied spectroscopy reviews"] = "Appl. Spectrosc. Rev.",
  ["applied statistics"] = "Appl Stat",
  ["applied stochastic methods series"] = "Appl. Stoch. Methods Ser.",
  ["applied stochastic models and data analysis"] = "Appl. Stochastic Models Data Anal.",
  ["applied stochastic models in business and industry"] = "Appl Stoch Models Bus Ind",
  ["applied superconductivity"] = "Appl. Supercond.",
  ["applied surface science"] = "Appl. Surf. Sci.",
  ["applied surface science advances"] = "Appl. Surf. Sci. Adv.",
  ["applied therapeutics"] = "Appl Ther",
  ["applied thermal engineering"] = "Appl. Therm. Eng.",
  ["applied turfgrass science"] = "Appl. Turfgrass Sci.",
  ["applied vegetation science"] = "Appl. Veg. Sci.",
  ["applied water science"] = "Appl. Water Sci.",
  ["approaches to semiotics"] = "Approaches Semiotics",
  ["approximation & optimization"] = "Approx. Optim.",
  ["approximation theory and its applications"] = "Approximation Theory Appl.",
  ["apps. applied sciences"] = "Appl. Sci.",
  ["appunti romani di filologia. studi e comunicazioni di filologia, linguistica e letteratura greca e latina"] = "AppRomFil",
  ["appunti. scuola normale superiore di pisa (nuova serie)"] = "Appunti. Sc. Norm. Super. Pisa (N. S.)",
  ["apres-demain / fonde par la ligue des droits de l’homme"] = "Apres Demain",
  ["apriori. estestvennye i tekhnicheskie nauki"] = "Apriori",
  ["aps observer"] = "APS Obs",
  ["apsipa transactions on signal and information processing"] = "APSIPA Trans Signal Inf Process",
  ["apsp journal of case reports"] = "APSP J Case Rep",
  ["aptamers (oxford, england)"] = "Aptamers (Oxf)",
  ["aptechnoe delo"] = "Aptechn Delo",
  ["apulum. acta musei apulensis"] = "Apulum",
  ["apuntes de psicología"] = "Apunt Psicol",
  ["apunts sports medicine"] = "Apunts Sports Med.",
  ["aq (balmain, n.s.w.)"] = "AQ",
  ["aqeic boletin tecnico"] = "AQEIC Bol. Tec.",
  ["aqua (neu-isenburg, germany)"] = "Aqua",
  ["aqua - water infrastructure, ecosystems and society"] = "AQUA Water Infrastruct. Ecosyst. Soc.",
  ["aqua planta"] = "Aqua Planta",
  ["aquacultural engineering"] = "Aquacult. Eng.",
  ["aquaculture (amsterdam, netherlands)"] = "Aquaculture",
  ["aquaculture and fisheries"] = "Aquac Fish",
  ["aquaculture association of canada special publication"] = "Aquac. Assoc. Can. Spec.",
  ["aquaculture economics & management"] = "Aquacult. Econ. Manage.",
  ["aquaculture economics and management"] = "Aquaculture Econ. Manage.",
  ["aquaculture environment interactions"] = "Aquac Environ Interact",
  ["aquaculture international"] = "Aquacult. Int.",
  ["aquaculture international : journal of the european aquaculture society"] = "Aquac Int",
  ["aquaculture nutrition"] = "Aquacult. Nutr.",
  ["aquaculture reports"] = "Aquacult. Rep.",
  ["aquaculture research"] = "Aquac Res",
  ["aquaculture, aquarium, conservation & legislation : international journal of the bioflux society"] = "Aquac Aquar Conserv Legis",
  ["aquaculture, fish and fisheries"] = "Aquacult. Fish Fish.",
  ["aquarium sciences and conservation"] = "Aquarium Sci. Conserv.",
  ["aquatic biology"] = "Aquat Biol",
  ["aquatic biosystems"] = "Aquat Biosyst",
  ["aquatic botany"] = "Aquat Bot",
  ["aquatic conservation : marine and freshwater ecosystems"] = "Aquat Conserv",
  ["aquatic conservation: marine and freshwater ecosystems"] = "Aquat. Conserv. Mar. Freshwater Ecosyst.",
  ["aquatic ecology"] = "Aquat Ecol",
  ["aquatic ecosystem health & management"] = "Aquat. Ecosyst. Health Manage.",
  ["aquatic geochemistry"] = "Aquat. Geochem.",
  ["aquatic insects"] = "Aquat Insects",
  ["aquatic invaders"] = "Aquat Invaders",
  ["aquatic invasions"] = "Aquat Invasions",
  ["aquatic living resources"] = "Aquat Living Resour",
  ["aquatic mammals"] = "Aquat Mamm",
  ["aquatic microbial ecology"] = "Aquat. Microb. Ecol.",
  ["aquatic microbial ecology : international journal"] = "Aquat Microb Ecol",
  ["aquatic procedia"] = "Aquat Procedia",
  ["aquatic sciences"] = "Aquat. Sci.",
  ["aquatic toxicology"] = "Aquat. Toxicol.",
  ["aquatic toxicology (amsterdam, netherlands)"] = "Aquat Toxicol",
  ["aquila legionis. cuadernos des estudios sobre el ejército romano"] = "AquiLeg",
  ["aquileia nostra. bollettino dell’associazione nazionale per aquileia"] = "AquilNost",
  ["aquileia nostra: bolletino dell’associazione nazionale per aquileia"] = "AN",
  ["aquilo. ser. botanica"] = "Aquilo Ser Bot",
  ["aquitania. une revue inter-régionale d’archéologie"] = "Aquitania",
  ["arab affairs (london, england)"] = "Arab Aff",
  ["arab gulf journal of scientific research"] = "Arab Gulf J. Sci. Res.",
  ["arab journal of basic and applied sciences"] = "Arab J Basic Appl Sci",
  ["arab journal of biotechnology"] = "Arab J Biotechnol",
  ["arab journal of gastroenterology : the official publication of the pan-arab association of gastroenterology"] = "Arab J Gastroenterol",
  ["arab journal of mathematical sciences"] = "Arab J. Math. Sci.",
  ["arab journal of nephrology and transplantation"] = "Arab J Nephrol Transplant",
  ["arab journal of urology"] = "Arab J Urol",
  ["arab studies quarterly"] = "Arab Stud Q",
  ["arabia (london, england)"] = "Arabia",
  ["arabian archaeology and epigraphy"] = "Arabian Archaeol. Epigr.",
  ["arabian journal for science and engineering"] = "Arab J Sci Eng",
  ["arabian journal of chemistry"] = "Arabian J. Chem.",
  ["arabian journal of geosciences"] = "Arabian J. Geosci.",
  ["arabian journal of mathematics"] = "Arabian J. Math.",
  ["arabic sciences and philosophy"] = "Arabic Sci. Philos.",
  ["arabic sciences and philosophy : a historical journal"] = "Arab Sci Philos",
  ["arabic sciences and philosophy. a historical journal"] = "Arabic Sci. Philos.",
  ["arachnologische mitteilungen"] = "Arachnol Mitt",
  ["arachnologische mitteilungen: arachnology letters"] = "Arachnol. Mitt.: Arachnol. Lett.",
  ["aramco world magazine"] = "Aramco World Mag",
  ["arbeit und soziales"] = "Arb Soz",
  ["arbeiten aus dem anatomischen institut der kaiserlich-japanischen universitaet zu sendai"] = "Arb. Anat. Inst. Kais.-Jpn. Univ. Sendai",
  ["arbeiten aus dem paul-ehrlich-institut (bundesamt fur sera und impfstoffe) zu frankfurt a. m."] = "Arb. Paul Ehrlich Inst. Bundesamt Sera Impfstoffe Frankf. a. M.",
  ["arbeiten aus dem paul-ehrlich-institut (bundesamt fur sera und impfstoffe) zu frankfurt a.m"] = "Arb Paul Ehrlich Inst Bundesamt Sera Impfstoffe Frankf A M",
  ["arbeiten aus dem paul-ehrlich-institut (bundesamt fur sera undimpfstoffe) zu frankfurt a.m. arb paul ehrlich inst bundesamt ser"] = "Impfstoffe Frankf A M",
  ["arbeiten aus dem paul-ehrlich-institut (bundesamt für sera und impfstoffe) zu frankfurt a.m"] = "Arb Paul Ehrlich Inst Bundesamt Sera Impfstoffe Frankf A M",
  ["arbeiten aus dem paul-ehrlich-institut (bundesinstitut für impfstoffe und biomedizinische arzneimittel) langen/hessen"] = "Arb Paul Ehrlich Inst Bundesinstitut Impfstoffe Biomed Arzneim Langen Hess",
  ["arbeiten aus dem paul-ehrlich-institut, dem georg-speyer-haus und dem ferdinand-blum-institut zu frankfurt a. m."] = "Arb. Paul Ehrlich Inst. Georg Speyer Haus Ferdinand Blum Inst. Frankf. a. M.",
  ["arbeiten aus dem paul-ehrlich-institut, dem georg-speyer-haus und dem ferdinand-blum-institut zu frankfurt a.m"] = "Arb Paul Ehrlich Inst Georg Speyer Haus Ferdinand Blum Inst Frankf A M",
  ["arbeiten der dlg"] = "Arb. DLG",
  ["arbeiten der kaiserlichen biologischen anstalt fur land- und forstwirtschaft"] = "Arb. Kais. Biol. Anst. Land Forstwirtsch.",
  ["arbeiten zur geschichte der medizin in giessen"] = "Arb Gesch Med Giessen",
  ["arbeits- und forschungsberichte zur sächsischen bodendenkmalpflege"] = "ArbFBerSaechs",
  ["arbeitsberichte des instituts fur mathematische maschinen und datenverarbeitung, band 15"] = "Arbeitsber. Inst. Math. Masch. Datenverarb., Band 15",
  ["arbeitsberichte des instituts für mathematische maschinen und datenverarbeitung (informatik)"] = "Arbeitsber. Inst. Math. Masch. Datenverarb. Inform.",
  ["arbeitsmedizin, sozialmedizin, praventivmedizin"] = "Arbeitsmed Sozialmed Praventivmed",
  ["arbeitsmedizin, sozialmedizin, präventivmedizin"] = "Arbeitsmed Sozialmed Praventivmed",
  ["arbeitspapiere des sonderforschungsbereichs 340, sprachtheoretische grundlagen für die computerlinguistik"] = "Arbeitspapiere Sonderforschungsbereichs 340 Sprachtheor. Grundlagen Computerlinguist. Ber.",
  ["arbeitsphysiologie; internationale zeitschrift fur angewandte physiologie"] = "Arbeitsphysiologie",
  ["arbeitsphysiologie; internationale zeitschrift für angewandte physiologie"] = "Arbeitsphysiologie",
  ["arbejderhistorie (1995)"] = "Arbejderhistorie",
  ["arbetarrorelsens arsbok"] = "Arbetarrorel Arsb",
  ["arbetarrörelsens årsbok"] = "Arbetarrorel Arsb",
  ["arbor scientiarum: beitrage zur wissenschaftsgeschichte, reihe a: abhandlungen"] = "Arbor Scientiarum: Beitrage Wissenschaftsgeschichte, Reihe A: Abhandlungen",
  ["arboricultural journal"] = "Arboric J",
  ["arboriculture & urban forestry"] = "Arboric Urban For",
  ["arc de cercle : an international journal of the history of the mind-sciences"] = "Arc Cercle",
  ["arc journal of addiction"] = "ARC J Addict",
  ["arc journal of diabetes and endocrinology"] = "ARC J Diabetes Endocrinol",
  ["arc journal of public health and community medicine"] = "ARC J Public Health Community Med",
  ["archaea (vancouver, b.c.)"] = "Archaea",
  ["archaeologia aeliana, or, miscellaneous tracts relating to antiquity"] = "Archaeol Aeliana",
  ["archaeologia austriaca"] = "AAustr",
  ["archaeologia bulgarica"] = "ABulg",
  ["archaeologia cantiana"] = "Archaeol Cantiana",
  ["archaeologia geographica"] = "AGeo",
  ["archaeologia homerica"] = "ArchHom",
  ["archaeologia jugoslavica"] = "AJug",
  ["archaeologia medievale. cultura materiale, insediamenti, territorio"] = "AMediev",
  ["archaeologia mosellana"] = "AMosel",
  ["archaeologia or miscellaneous tracts relating to antiquity published by the society of antiquaries of london"] = "Archaeologia",
  ["archaeologia polona"] = "APol",
  ["archaeologiai ertesitö"] = "AErt",
  ["archaeologiai értesitő"] = "AErt",
  ["archaeologica baltica"] = "Archaeol. Balt.",
  ["archaeological and anthropological sciences"] = "Archaeol Anthropol Sci",
  ["archaeological and historical studies. the archaeological society of alexandria"] = "AHistStAlex",
  ["archaeological dialogues"] = "Archaeol. Dialogues",
  ["archaeological journal"] = "Archaeol. J.",
  ["archaeological news"] = "ArchN",
  ["archaeological papers of the american anthropological association"] = "Archaeol. Pap. Am. Anthropol. Assoc.",
  ["archaeological prospection"] = "Archaeol Prospect",
  ["archaeological reports"] = "Archaeol. Rep.",
  ["archaeological research in asia"] = "Archaeol. Res. Asia",
  ["archaeologisches nachrichtenblatt"] = "Archaeol. Nachrichtenbl.",
  ["archaeology in oceania"] = "Archaeol. Oceania",
  ["archaeology. a magazine dealing with the antiquity of the world"] = "Archaeology",
  ["archaeometry. bulletin of the research laboratory for archaeology and history of art, oxford university"] = "Archaeometry",
  ["archaiologike ephemeris"] = "AEph",
  ["archaiologikon deltion"] = "AD",
  ["archaologischer anzeiger"] = "Archaeol Anz",
  ["archäologischer anzeiger"] = "Archaeol Anz",
  ["archeia iatrikōn hetaireōn = archives of the hellenic medical societies"] = "Arch Iatr Hetaireon",
  ["archeion iatrikon epistemon"] = "Arch Iatr Epistem",
  ["archeion iatrikōn epistēmōn"] = "Arch Iatr Epistem",
  ["archeo. attualità del passato"] = "Archeo",
  ["archeografo triestino"] = "ArcheogrTriest",
  ["archeologia (torun, poland)"] = "Archeologia",
  ["archeologia (toruń, poland)"] = "Archeologia",
  ["archeologia classica"] = "ArchCl",
  ["archeologia classica: rivista della scuola naz. di archeologia, pubbl. a cura degli ist. di archeologia e storia dell’arte greca e romana e di etruscologia e antichità italiche dell’univ. di roma"] = "ArchClass",
  ["archeologia corsa"] = "ACors",
  ["archeologia delle acque. rivista semestrale di antropologia, archeologia, etnografia, storia dell’acqua"] = "AAcque",
  ["archeologia delle alpi"] = "AAlpi",
  ["archeologia dell’architettura"] = "AArchit",
  ["archeologia dell’emilia-romagna"] = "AEmil",
  ["archeologia e calcolatori"] = "ACalc",
  ["archeologia e società"] = "ASoc",
  ["archeologia in campania. bolletino di informazioni a cura della soprintendenza archeologica delle province di napoli e caserta"] = "ACamp",
  ["archeologia subacquea. studi, ricerche e documenti"] = "ASubacq",
  ["archeologia veneta"] = "AVen",
  ["archeologia viva"] = "AViva",
  ["archeologia, paris. l’archéologie dans le monde et tout ce qui concerne les recherches historiques, artistiques et scientifiques sur terre et dans les mers"] = "ArcheologiaParis",
  ["archeologia, uomo, territorio. rivista dei gruppi archeologici nord italia"] = "AUTerr",
  ["archeologia. rivista bimestrale. roma"] = "ArcheologiaRoma",
  ["archeologia. rocznik instytutu archeologii i etnologii, polskiej akademii nauk"] = "ArcheologiaWarsz",
  ["archeological papers of the american anthropological association"] = "Arch. P. Amer. Ant. Asso.",
  ["archeologické rozhledy"] = "ARozhl",
  ["archeologičeskij sbornik. gosudarstvennyj ordena lenina ermitaž"] = "ASbor",
  ["archeologija. nacional’na akademija nauk ukraini. institut archeologii"] = "ArcheologijaKiiv",
  ["archeologija. organ na archeologičeskija institut i muzej (pri bălgarskata akademii nauk)"] = "ArcheologijaSof",
  ["archeologičeskie otkrytija"] = "AOtkryt",
  ["archeologični pamjatki ursr"] = "APamKiiv",
  ["archeosciences-revue d archeometrie"] = "ArcheoSci.-Rev. Archeom.",
  ["archief. mededelingen van het koninklijk zeeuwsch genootschap der wetenschappen. zeeuwsch genootschap der wetenschappen"] = "Archief Meded k Zeeuwsch Genoot Wet",
  ["archimedes: new studies in the history and philosophy of science and technology"] = "Archimedes",
  ["architectura. zeitschrift für geschichte der baukunst"] = "Architectura",
  ["architectural engineering and design management"] = "Archit. Eng. Des. Manage.",
  ["architectural intelligence"] = "Archit. Intell.",
  ["architectural record"] = "Archit Rec",
  ["architectural science review"] = "Archit Sci Rev",
  ["architectural theory review"] = "Archit. Theory Rev.",
  ["architecture (washington, d.c.)"] = "Architecture",
  ["architecture and culture"] = "Archit. Cult.",
  ["architecture, structures and construction"] = "Archit. Struct. Construct.",
  ["archiv der geschichte der naturwissenschaften"] = "Arch Gesch Naturwiss",
  ["archiv der julius klaus-stiftung fur vererbungsforschung, sozialanthropologie und rassenhygiene"] = "Arch. Julius Klaus Stift. Vererbungsforsch. Sozialanthropol. Rassenhyg.",
  ["archiv der julius klaus-stiftung für vererbungsforschung, sozialanthropologie und rassenhygiene"] = "Arch Julius Klaus Stift Vererbungsforsch Sozialanthropol Rassenhyg",
  ["archiv der mathematik"] = "Arch. Math. (Basel)",
  ["archiv der mathematik. archives of mathematics. archives mathématiques"] = "Arch Math",
  ["archiv der pharmazie"] = "Arch Pharm (Weinheim)",
  ["archiv der pharmazie (weinheim, germany)"] = "Arch. Pharm. (Weinheim, Ger.)",
  ["archiv der pharmazie und berichte der deutschen pharmazeutischen gesellschaft"] = "Arch Pharm Ber Dtsch Pharm Ges",
  ["archiv for pharmaci og chemi"] = "Arch Pharm Chem (Kbh)",
  ["archiv fur das eisenhuttenwesen"] = "Arch. Eisenhuttenwes.",
  ["archiv fur dermatologische forschung"] = "Arch Dermatol Forsch",
  ["archiv fur die gesamte psychologie"] = "Arch Gesamte Psychol",
  ["archiv fur die gesamte virusforschung"] = "Arch Gesamte Virusforsch",
  ["archiv fur die geschichte von oberfranken"] = "Arch Gesch Oberfranken",
  ["archiv fur elektrotechnik"] = "Arch. Elektrotech.",
  ["archiv fur experimentalle pathologie und pharmakologie"] = "Arch. Exp. Pathol. Pharmakol.",
  ["archiv fur experimentelle veterinarmedizin"] = "Arch Exp Veterinarmed",
  ["archiv fur genetik"] = "Arch Genet (Zur)",
  ["archiv fur geschichte des buchwesens"] = "Arch Gesch Buchwes",
  ["archiv fur geschwulstforschung"] = "Arch Geschwulstforsch",
  ["archiv fur gewerbepathologie und gewerbehygiene"] = "Arch Gewerbepathol Gewerbehyg",
  ["archiv fur gynakologie"] = "Arch Gynakol",
  ["archiv fur hydrobiologie, supplement"] = "Arch. Hydrobiol. Suppl.",
  ["archiv fur hygiene und bakteriologie"] = "Arch Hyg Bakteriol",
  ["archiv fur kinderheilkunde"] = "Arch Kinderheilkd",
  ["archiv fur kinderheilkunde. beihefte"] = "Arch Kinderheilkd Suppl",
  ["archiv fur klinische medizin"] = "Arch Klin Med",
  ["archiv fur klinische und experimentelle dermatologie"] = "Arch Klin Exp Dermatol",
  ["archiv fur klinische und experimentelle ohren- nasen- und kehlkopfheilkunde"] = "Arch Klin Exp Ohren Nasen Kehlkopfheilkd",
  ["archiv fur kreislaufforschung"] = "Arch Kreislaufforsch",
  ["archiv fur kriminologie"] = "Arch Kriminol",
  ["archiv fur kultur-geschichte"] = "Arch Kulturgesch",
  ["archiv fur lagerstattenforschung"] = "Arch. Lagerstattenforsch.",
  ["archiv fur lebensmittelhygiene"] = "Arch. Lebensmittelhyg.",
  ["archiv fur meteorologie, geophysik und bioklimatologie. serie b: klimatologie, umweltmeteorologie, strahlungsforschung"] = "Arch Meteorol Geophys Bioklimatol [B]",
  ["archiv fur meteorologie, geophysik, und bioklimatologie serie a meteorologie und geophysik"] = "Arch. Meteorol. Geophys. Bioklimatol. A",
  ["archiv fur mikrobiologie"] = "Arch Mikrobiol",
  ["archiv fur mikroskopische anatomie"] = "Arch. Mikrosk. Anat.",
  ["archiv fur molluskenkunde"] = "Arch. Molluskenkd.",
  ["archiv fur ohren-, nasen- und kehlkopfheilkunde, vereinigt mit zeitschrift fur hals-, nasen- und ohrenheilkunde"] = "Arch Ohren Nasen Kehlkopfheilkd",
  ["archiv fur orientforschung"] = "Arch Orientforsch",
  ["archiv fur orthopadische und unfall-chirurgie"] = "Arch Orthop Unfallchir",
  ["archiv fur papyrusforschung und verwandte gebiete"] = "Arch. Papyrusforsch. Verw. Geb.",
  ["archiv fur physikalische therapie"] = "Arch Phys Ther (Leipz)",
  ["archiv fur protistenkunde"] = "Arch. Protistenkd.",
  ["archiv fur psychiatrie und nervenkrankheiten"] = "Arch Psychiatr Nervenkr",
  ["archiv fur psychiatrie und nervenkrankheiten, vereinigt mit zeitschrift fur die gesamte neurologie und psychiatrie"] = "Arch Psychiatr Nervenkr Z Gesamte Neurol Psychiatr",
  ["archiv fur psychologie"] = "Arch Psychol (Frankf)",
  ["archiv fur reformationsgeschichte"] = "Arch Reformation Hist",
  ["archiv fur sozialgeschichte"] = "Arch Sozialgesch",
  ["archiv fur tierernahrung"] = "Arch Tierernahr",
  ["archiv fur toxikologie"] = "Arch Toxikol",
  ["archiv fur urkundenforschung"] = "Arch. Urkundenforsch.",
  ["archiv für acker- und pflanzenbau und bodenkunde"] = "Arch Acker Pflanzenbau Bodenkd",
  ["archiv für dermatologie und syphilis"] = "Arch Dermatol Syph",
  ["archiv für dermatologische forschung"] = "Arch Dermatol Forsch",
  ["archiv für die gesamte psychologie"] = "Arch Gesamte Psychol",
  ["archiv für die gesamte virusforschung"] = "Arch Gesamte Virusforsch",
  ["archiv für die geschichte von oberfranken"] = "Arch Gesch Oberfranken",
  ["archiv für entwicklungsmechanik der organismen"] = "Arch Entwicklmech Org",
  ["archiv für experimentelle veterinärmedizin"] = "Arch Exp Veterinarmed",
  ["archiv für genetik"] = "Arch Genet (Zur)",
  ["archiv für geschichte des buchwesens"] = "Arch Gesch Buchwes",
  ["archiv für geschwulstforschung"] = "Arch Geschwulstforsch",
  ["archiv für gewerbepathologie und gewerbehygiene"] = "Arch Gewerbepathol Gewerbehyg",
  ["archiv für gynäkologie"] = "Arch Gynakol",
  ["archiv für hydrobiologie"] = "Arch Hydrobiol",
  ["archiv für hygiene und bakteriologie"] = "Arch Hyg Bakteriol",
  ["archiv für innere medizin"] = "Arch Inn Med",
  ["archiv für kinderheilkunde"] = "Arch Kinderheilkd",
  ["archiv für kinderheilkunde. beihefte"] = "Arch Kinderheilkd Suppl",
  ["archiv für klinische medizin"] = "Arch Klin Med",
  ["archiv für klinische und experimentelle dermatologie"] = "Arch Klin Exp Dermatol",
  ["archiv für klinische und experimentelle ohren- nasen- und kehlkopfheilkunde"] = "Arch Klin Exp Ohren Nasen Kehlkopfheilkd",
  ["archiv für kreislaufforschung"] = "Arch Kreislaufforsch",
  ["archiv für kriminologie"] = "Arch Kriminol",
  ["archiv für kultur-geschichte"] = "Arch Kulturgesch",
  ["archiv für lebensmittelhygiene"] = "Arch Lebensmittelhyg",
  ["archiv für meteorologie, geophysik und bioklimatologie. serie b: klimatologie, umweltmeteorologie, strahlungsforschung"] = "Arch Meteorol Geophys Bioklimatol B",
  ["archiv für mikrobiologie"] = "Arch Mikrobiol",
  ["archiv für mikroskopische anatomie und entwicklungsmechanik"] = "Arch Mikrosk Anat Enwicklmech",
  ["archiv für molluskenkunde"] = "Arch Molluskenkd 2004",
  ["archiv für ohren-, nasen- und kehlkopfheilkunde"] = "Arch Ohren Nasen Kehlkopfheilkd",
  ["archiv für orientforschung"] = "Arch Orientforsch",
  ["archiv für orthopädische und unfall-chirurgie"] = "Arch Orthop Unfallchir",
  ["archiv für papyrusforschung und verwandte gebiete"] = "Arch Papyrusforsch",
  ["archiv für physikalische therapie"] = "Arch Phys Ther (Leipz)",
  ["archiv für phytopathologie und pflanzenschutz"] = "Arch Phytopathol Pflanzenschutz",
  ["archiv für protistenkunde"] = "Arch Protistenkunde",
  ["archiv für psychiatrie und nervenkrankheiten"] = "Arch Psychiatr Nervenkr",
  ["archiv für psychiatrie und nervenkrankheiten, vereinigt mit zeitschrift für die gesamte neurologie und psychiatrie"] = "Arch Psychiatr Nervenkr Z Gesamte Neurol Psychiatr",
  ["archiv für psychologie"] = "Arch Psychol (Frankf)",
  ["archiv für reformationsgeschichte"] = "Arch Reformation Hist",
  ["archiv für sozialgeschichte"] = "Arch Sozialgesch",
  ["archiv für tierernährung"] = "Arch Tierernahr",
  ["archiv für tierzucht"] = "Arch Tierz",
  ["archiv für toxikologie"] = "Arch Toxikol",
  ["archiv für begriffsgeschichte"] = "ArchBegriffsGesch",
  ["archiv für begriffsgeschichte: bausteine zu einem historischen wörterbuch der philosophie"] = "ABG",
  ["archiv für dermatologische forschung"] = "Arch. Dermatol. Forsch.",
  ["archiv für die gesamte psychologie"] = "Arch. Gesamte Psychol.",
  ["archiv für die gesamte virusforschung"] = "Arch. Gesamte Virusforsch.",
  ["archiv für elektrotechnik"] = "Arch. Elektrotech.",
  ["archiv für experimentalle pathologie und pharmakologie"] = "Arch. Exp. Pathol. Pharmakol.",
  ["archiv für experimentelle veterinarmedizin"] = "Arch. Exp. Veterinarmed.",
  ["archiv für forstwesen"] = "Arch. Forstwes.",
  ["archiv für genetik"] = "Arch. Genet. (Zur.)",
  ["archiv für geschichte der philosophie"] = "AGPh",
  ["archiv für geschwulstforschung"] = "Arch. Geschwulstforsch.",
  ["archiv für gynakologie"] = "Arch. Gynakol.",
  ["archiv für hydrobiologie"] = "Arch. Hydrobiol.",
  ["archiv für hydrobiologie, supplement"] = "Arch. Hydrobiol. Suppl.",
  ["archiv für hygiene und bakteriologie"] = "Arch. Hyg. Bakteriol.",
  ["archiv für kinderheilkunde"] = "Arch. Kinderheilkd.",
  ["archiv für klinische medizin"] = "Arch. Klin. Med.",
  ["archiv für klinische und experimentelle dermatologie"] = "Arch. Klin. Exp. Dermatol.",
  ["archiv für klinische und experimentelle ohren-, nasen- und kehlkopfheilkunde"] = "Arch. Klin. Exp. Ohren. Nasen. Kehlkopfheilkd.",
  ["archiv für kreislaufforschung"] = "Arch. Kreislaufforsch.",
  ["archiv für kriminologie"] = "Arch. Kriminol.",
  ["archiv für kulturgeschichte"] = "AKG",
  ["archiv für lagerstattenforschung"] = "Arch. Lagerstattenforsch.",
  ["archiv für lebensmittelhygiene"] = "Arch. Lebensmittelhyg.",
  ["archiv für meteorologie, geophysik und bioklimatologie"] = "Arch. Meteor. Geophys. Bioklimatol.",
  ["archiv für meteorologie, geophysik und bioklimatologie, serie b"] = "Arch. Meteorol. Geophys. Bioklimatol., Ser. B",
  ["archiv für meteorologie, geophysik, und bioklimatologie serie a meteorologie und geophysik"] = "Arch. Meteorol. Geophys. Bioklimatol. A",
  ["archiv für mikrobiologie"] = "Arch. Mikrobiol.",
  ["archiv für musikwissenschaft"] = "AMW",
  ["archiv für naturschutz und landschaftsforschung"] = "Arch. Nat.schutz Landsch.forsch.",
  ["archiv für ohren-, nasen- und kehlkopfheilkunde"] = "Arch. Ohren Nasen Kehlkopfheilkd.",
  ["archiv für orientforschung"] = "AfO",
  ["archiv für orientforschung: internationale zeitschrift für die wissenschaft vom vorderen orient"] = "AOF",
  ["archiv für orthopadische und unfall-chirurgie"] = "Arch. Orthop. Unfallchir.",
  ["archiv für papyrusforschung und verwandte gebiete"] = "APF",
  ["archiv für physikalische therapie"] = "Arch. Phys. Ther. (Leipz.)",
  ["archiv für psychiatrie und nervenkrankheiten"] = "Arch. Psychiatr. Nervenkr.",
  ["archiv für psychologie"] = "Arch. Psychol. (Frankf.)",
  ["archiv für religionsgeschichte"] = "ArchRel",
  ["archiv für tierernahrung"] = "Arch. Tierernahr.",
  ["archiv für toxikologie"] = "Arch. Toxikol.",
  ["archiv orientalni"] = "Arch Orient",
  ["archiv orientální"] = "Arch Orient",
  ["archiv orientální"] = "ArchOrient",
  ["archiv und wissenschaft; schriftenreihe der archivalischen zeitschrift"] = "Arch Wiss",
  ["archiva moldaviæ"] = "Arch Mold",
  ["archivalia médica"] = "Arch Med",
  ["archivalische zeitschrift"] = "Arch Z",
  ["archive for history of exact sciences"] = "Arch. Hist. Exact Sci.",
  ["archive for mathematical logic"] = "Arch. Math. Logic",
  ["archive for rational mechanics and analysis"] = "Arch. Ration. Mech. Anal.",
  ["archive for the psychology of religion = archiv für religionspsychologie"] = "Arch Psychol Relig",
  ["archive of applied mechanics"] = "Arch. Appl. Mech.",
  ["archive of applied mechanics = ingenieur-archiv"] = "Arch Appl Mech",
  ["archive of oncology"] = "Arch Oncol",
  ["archive of urological research"] = "Arch Urol Res",
  ["archive reporter. data archive on adolescent pregnancy and pregnancy prevention"] = "Arch Report",
  ["archives and records (abingdon, england)"] = "Arch Rec (Abingdon)",
  ["archives animal breeding"] = "Arch Anim Breed",
  ["archives belges"] = "Arch. Belg.",
  ["archives belges = belgisch archief"] = "Arch Belg",
  ["archives belges de dermatologie"] = "Arch Belg Dermatol",
  ["archives belges de dermatologie et de syphiligraphie"] = "Arch Belg Dermatol Syphiligr",
  ["archives belges de medecine sociale, hygiene, medecine du travail et medecine legale"] = "Arch. Belg. Med. Soc.",
  ["archives belges de medecine sociale, hygiene, medecine du travail et medecine legale. belgisch archief van sociale geneeskunde, hygiene, arbeidsgeneeskunde en gerechtelijke geneeskunde"] = "Arch Belg Med Soc",
  ["archives belges de médecine sociale, hygiène, médecine du travail et médecine légale. belgisch archief van sociale geneeskunde, hygiëne, arbeidsgeneeskunde en gerechtelijke geneeskunde"] = "Arch Belg Med Soc",
  ["archives d anatomie et de cytologie pathologiques"] = "Arch Anat Cytol Pathol",
  ["archives d'anatomie et de cytologie pathologiques"] = "Arch Anat Cytol Pathol",
  ["archives d'anatomie microscopique et de morphologie expérimentale"] = "Arch Anat Microsc Morphol Exp",
  ["archives d'anatomie pathologique"] = "Arch Anat Pathol (Paris)",
  ["archives d'anatomie, d'histologie et d'embryologie normales et expérimentales"] = "Arch Anat Histol Embryol",
  ["archives d'histoire doctrinale et littéraire du moyen âge"] = "Arch Hist Doctrin Litt Moyen Age",
  ["archives d'ophtalmologie"] = "Arch Ophtalmol (Paris)",
  ["archives d'ophtalmologie et revue générale d'ophtalmologie"] = "Arch Ophtalmol Rev Gen Ophtalmol",
  ["archives de biologie"] = "Arch Biol (Liege)",
  ["archives de l institut pasteur d algerie"] = "Arch Inst Pasteur Alger",
  ["archives de l institut pasteur de madagascar"] = "Arch Inst Pasteur Madagascar",
  ["archives de l institut pasteur de tunis"] = "Arch Inst Pasteur Tunis",
  ["archives de l'institut d'hessarek"] = "Arch Inst Hessarek",
  ["archives de l'institut pasteur d'algérie. institut pasteur d'algérie"] = "Arch Inst Pasteur Alger",
  ["archives de l'institut pasteur de la martinique"] = "Arch Inst Pasteur Martinique",
  ["archives de l'institut pasteur de madagascar"] = "Arch Inst Pasteur Madagascar",
  ["archives de l'institut pasteur de tunis"] = "Arch Inst Pasteur Tunis",
  ["archives de l'institut pasteur hellénique"] = "Arch Inst Pasteur Hell",
  ["archives de l’institut pasteur d algerie"] = "Arch. Inst. Pasteur Alger.",
  ["archives de l’institut pasteur de la martinique"] = "Arch Inst Pasteur Martinique",
  ["archives de l’institut pasteur de madagascar"] = "Arch. Inst. Pasteur Madagascar",
  ["archives de l’institut pasteur de tunis"] = "Arch. Inst. Pasteur Tunis",
  ["archives de l’institut pasteur d’algerie institut pasteur d’algerie"] = "Arch Inst Pasteur Alger",
  ["archives de l’institut pasteur hellenique"] = "Arch Inst Pasteur Hell",
  ["archives de medecine generale et tropicale"] = "Arch. Med. Gen. Trop.",
  ["archives de medecine sociale"] = "Arch Med Soc",
  ["archives de médecine générale et tropicale"] = "Arch Med Gen Trop",
  ["archives de médecine sociale"] = "Arch Med Soc",
  ["archives de pediatrie"] = "Arch. Pediatr.",
  ["archives de pediatrie : organe officiel de la societe francaise de pediatrie"] = "Arch Pediatr",
  ["archives de pédiatrie : organe officiel de la sociéte française de pédiatrie"] = "Arch Pediatr",
  ["archives de philosophie"] = "Arch Philos",
  ["archives de philosophie du droit"] = "APhD",
  ["archives de philosophie: recherches et documentation"] = "ArchPhilos",
  ["archives de psychologie"] = "Arch Psychol (Geneve)",
  ["archives de rhumatologie"] = "Arch Rhumatol",
  ["archives de sciences sociales des religions"] = "Arch Sci Soc Relig",
  ["archives de stomatologie"] = "Arch Stomatol",
  ["archives des maladies de l'appareil digestif et des maladies de la nutrition"] = "Arch Mal Appar Dig Mal Nutr",
  ["archives des maladies de l’appareil digestif et des maladies de la nutrition"] = "Arch. Mal. Appar. Dig. Mal. Nutr.",
  ["archives des maladies du coeur et des vaisseaux"] = "Arch Mal Coeur Vaiss",
  ["archives des maladies du coeur et des vaisseaux. pratique"] = "Arch Mal Coeur Vaiss Pratique",
  ["archives des maladies professionnelles de medecine du travail et de securite sociale"] = "Arch. Mal. Prof.",
  ["archives des maladies professionnelles de médecine du travail et de sécurité sociale"] = "Arch Mal Prof",
  ["archives des maladies professionnelles et de l’environment"] = "Arch. Mal. Prof. Environ.",
  ["archives des sciences"] = "Arch Sci",
  ["archives des sciences / editees par la societe de physique et d’histoire naturelle de geneve"] = "Arch Sci",
  ["archives des sciences et compte rendu des seances de la societe / edites par la societe de physique et d’histoire naturelle de geneve"] = "Arch Sci Compte Rendu Seances Soc",
  ["archives des sciences et compte rendu des séances de la société"] = "Arch Sci Compte Rendu Seances Soc",
  ["archives des sciences physiologiques"] = "Arch Sci Physiol (Paris)",
  ["archives d’anatomie et de cytologie pathologiques"] = "Arch. Anat. Cytol. Pathol.",
  ["archives d’anatomie microscopique et de morphologie experimentale"] = "Arch. Anat. Microsc. Morphol. Exp.",
  ["archives d’anatomie pathologique"] = "Arch. Anat. Pathol. (Paris)",
  ["archives d’anatomie, d’histologie et d’embryologie"] = "Arch. Anat. Histol. Embryol.",
  ["archives d’anatomie, d’histologie et d’embryologie normales et experimentales"] = "Arch Anat Histol Embryol",
  ["archives d’histoire doctrinale et littéraire du moyen âge"] = "AHMA",
  ["archives d’ophtalmologie"] = "Arch. Ophtalmol. (Paris)",
  ["archives d’ophtalmologie et revue generale d’ophtalmologie"] = "Arch. Ophtalmol. Rev. Gen. Ophtalmol.",
  ["archives europeennes de sociologie"] = "Arch. Eur. Sociol.",
  ["archives europeennes de sociologie. european journal of sociology. europaisches archiv fur soziologie"] = "Arch Eur Sociol",
  ["archives européennes de sociologie. european journal of sociology. europäisches archiv für soziologie"] = "Arch Eur Sociol",
  ["archives for dermatological research (archiv fur dermatologische forschung)"] = "Arch. Dermatol. Res.",
  ["archives for dermatological research = archiv für dermatologische forschung"] = "Arch Dermatol Res",
  ["archives for dermatological research. archiv fur dermatologische forschung"] = "Arch Dermatol Res",
  ["archives francaises de pediatrie"] = "Arch. Fr. Pediatr.",
  ["archives francaises des maladies de l’appareil digestif"] = "Arch. Fr. Mal. App. Dig.",
  ["archives françaises de pédiatrie"] = "Arch Fr Pediatr",
  ["archives françaises des maladies de l'appareil digestif"] = "Arch Fr Mal App Dig",
  ["archives hospitalieres"] = "Arch Hosp (Paris)",
  ["archives hospitalières"] = "Arch Hosp (Paris)",
  ["archives in biomedical engineering & biotechnology"] = "Arch Biomed Eng Biotechnol",
  ["archives in cancer research"] = "Arch Cancer Res",
  ["archives in neurology & neuroscience"] = "Arch Neurol Neurosci",
  ["archives internationales claude bernard"] = "Arch Int Claude Bernard",
  ["archives internationales d'histoire des idées = international archives of the history of ideas"] = "Arch Int Hist Idees",
  ["archives internationales d'histoire des sciences"] = "Arch Int Hist Sci (Paris)",
  ["archives internationales de neurologie, des maladies hereditaires, de medecine mentale et psychosomatique"] = "Arch Int Neurol",
  ["archives internationales de neurologie, des maladies héréditaires, de médecine mentale et psychosomatique"] = "Arch Int Neurol",
  ["archives internationales de pharmacodynamie et de therapie"] = "Arch. Int. Pharmacodyn. Ther.",
  ["archives internationales de pharmacodynamie et de thérapie"] = "Arch Int Pharmacodyn Ther",
  ["archives internationales de physiologie"] = "Arch Int Physiol",
  ["archives internationales de physiologie et de biochimie"] = "Arch Int Physiol Biochim",
  ["archives internationales de physiologie, de biochimie et de biophysique"] = "Arch Int Physiol Biochim Biophys",
  ["archives internationales d’histoire des sciences"] = "Arch. Int. Hist. Sci.",
  ["archives italiennes de biologie"] = "Arch Ital Biol",
  ["archives medicales de l’ouest: angers/nantes/rennes"] = "Arch Med Ouest",
  ["archives mediterraneennes de medecine"] = "Arch. Mediterr. Med.",
  ["archives médicales de l'ouest: angers/nantes/rennes"] = "Arch Med Ouest",
  ["archives méditerranéennes de médecine"] = "Arch Mediterr Med",
  ["archives neerlandaises des sciences exactes et naturelles"] = "Arch. Neerl. Sci. Exactes Nat.",
  ["archives neerlandaises des sciences exactes et naturelles. serie ii"] = "Arch. Neerl. Sci. Exactes Nat. Ser. II",
  ["archives neerlandaises sci exactes et naturelles"] = "Arch. Neerl. Sci. Exactes Nat.",
  ["archives néerlandaises de physiologie de l'homme et des animaux"] = "Arch Neerl Physiol Homme Anim",
  ["archives of academic emergency medicine"] = "Arch Acad Emerg Med",
  ["archives of agronomy and soil science"] = "Arch. Agron. Soil Sci.",
  ["archives of aids research"] = "Arch AIDS Res",
  ["archives of andrology"] = "Arch Androl",
  ["archives of anesthesiology"] = "Arch Anesthesiol",
  ["archives of animal nutrition"] = "Arch Anim Nutr",
  ["archives of applied science research"] = "Arch Appl Sci Res",
  ["archives of autoimmune diseases"] = "Arch Autoimmune Dis",
  ["archives of automotive engineering"] = "Arch. Automot. Eng.",
  ["archives of basic and applied medicine"] = "Arch Basic Appl Med",
  ["archives of biochemistry"] = "Arch Biochem",
  ["archives of biochemistry and biophysics"] = "Arch. Biochem. Biophys.",
  ["archives of biological sciences"] = "Arch Biol Sci",
  ["archives of cancer biology and therapy"] = "Arch Cancer Biol Ther",
  ["archives of cardiovascular diseases"] = "Arch Cardiovasc Dis",
  ["archives of child health"] = "Arch Child Health",
  ["archives of civil and mechanical engineering"] = "Arch. Civ. Mech. Eng.",
  ["archives of clinical and biomedical research"] = "Arch Clin Biomed Res",
  ["archives of clinical and medical case reports"] = "Arch Clin Med Case Rep",
  ["archives of clinical case reports"] = "Arch Clin Case Rep",
  ["archives of clinical infectious diseases"] = "Arch Clin Infect Dis",
  ["archives of clinical microbiology"] = "Arch Clin Microbiol",
  ["archives of clinical neuropsychology : the official journal of the national academy of neuropsychologists"] = "Arch Clin Neuropsychol",
  ["archives of clinical toxicology"] = "Arch Clin Toxicol (Middlet)",
  ["archives of computational methods in engineering"] = "Arch. Comput. Methods Eng.",
  ["archives of computational methods in engineering : state of the art reviews"] = "Arch Comput Methods Eng",
  ["archives of computational methods in engineering. state of the art reviews"] = "Arch. Comput. Methods Eng.",
  ["archives of control sciences"] = "Arch. Control Sci.",
  ["archives of control sciences. polish academy of sciences. committee of automatic control and robotics"] = "Arch. Control Sci.",
  ["archives of craniofacial surgery"] = "Arch Craniofac Surg",
  ["archives of depression and anxiety"] = "Arch Depress Anxiety",
  ["archives of dermatological research"] = "Arch Dermatol Res",
  ["archives of dermatology"] = "Arch Dermatol",
  ["archives of dermatology and syphilology"] = "Arch Derm Syphilol",
  ["archives of disease in childhood"] = "Arch Dis Child",
  ["archives of disease in childhood. education and practice edition"] = "Arch Dis Child Educ Pract Ed",
  ["archives of disease in childhood. fetal and neonatal edition"] = "Arch Dis Child Fetal Neonatal Ed",
  ["archives of drug information"] = "Arch Drug Inf",
  ["archives of economic history"] = "Archives Econ. Hist.",
  ["archives of electrical engineering"] = "Arch. Electr. Eng.",
  ["archives of emergency medicine"] = "Arch Emerg Med",
  ["archives of emergency medicine and critical care"] = "Arch Emerg Med Crit Care",
  ["archives of endocrinology and metabolism"] = "Arch Endocrinol Metab",
  ["archives of environment contamination and toxicology"] = "Arch. Environ. Contam. Toxicol.",
  ["archives of environment health"] = "Arch. Environ. Health",
  ["archives of environmental & occupational health"] = "Arch Environ Occup Health",
  ["archives of environmental and occupational health"] = "Arch. Environ. Occup. Health",
  ["archives of environmental contamination and toxicology"] = "Arch Environ Contam Toxicol",
  ["archives of environmental health"] = "Arch Environ Health",
  ["archives of epidemiology"] = "Arch Epidemiol",
  ["archives of experimental medicine and biology"] = "Arch. Exp. Med. Biol.",
  ["archives of facial plastic surgery"] = "Arch Facial Plast Surg",
  ["archives of facial plastic surgery : official publication for the american academy of facial plastic and reconstructive surgery, inc. and the international federation of facial plastic surgery societies"] = "Arch Facial Plast Surg",
  ["archives of family medicine"] = "Arch Fam Med",
  ["archives of foundry engineering"] = "Arch. Foundry Eng.",
  ["archives of general internal medicine"] = "Arch Gen Intern Med",
  ["archives of general psychiatry"] = "Arch Gen Psychiatry",
  ["archives of gerontology and geriatrics"] = "Arch Gerontol Geriatr",
  ["archives of gerontology and geriatrics. supplement"] = "Arch Gerontol Geriatr Suppl",
  ["archives of gynecology"] = "Arch Gynecol",
  ["archives of gynecology and obstetrics"] = "Arch Gynecol Obstet",
  ["archives of hematology and blood diseases"] = "Arch Hematol Blood Dis",
  ["archives of histology and cytology"] = "Arch Histol Cytol",
  ["archives of industrial hygiene and occupational medicine"] = "Arch Ind Hyg Occup Med",
  ["archives of insect biochemistry and physiology"] = "Arch Insect Biochem Physiol",
  ["archives of interamerican rheumatology"] = "Arch. Interam. Rheumatol.",
  ["archives of interamerican rheumatology : a.i.r"] = "Arch Interam Rheumatol",
  ["archives of internal medicine"] = "Arch Intern Med",
  ["archives of internal medicine (chicago, ill. : 1908)"] = "Arch Intern Med (Chic)",
  ["archives of iranian medicine"] = "Arch Iran Med",
  ["archives of materials science and engineering"] = "Arch. Mater. Sci. Eng.",
  ["archives of mechanics"] = "Arch. Mech.",
  ["archives of medical case reports"] = "Arch Med Case Rep",
  ["archives of medical research"] = "Arch Med Res",
  ["archives of medical science : ams"] = "Arch Med Sci",
  ["archives of medical sciences. atherosclerotic diseases"] = "Arch Med Sci Atheroscler Dis",
  ["archives of medicine"] = "Arch Med (Oviedo)",
  ["archives of metallurgy and materials"] = "Arch. Metall. Mater.",
  ["archives of microbiology"] = "Arch Microbiol",
  ["archives of mining sciences"] = "Arch. Min. Sci.",
  ["archives of natural history"] = "Arch Nat Hist",
  ["archives of nature conservation and landscape research"] = "Arch. Nat. Conserv. Landscape Res.",
  ["archives of neurology"] = "Arch Neurol",
  ["archives of neurology and neuro disorders"] = "Arch Neurol Neuro Disord",
  ["archives of neurology and psychiatry"] = "Arch Neurol Psychiatry",
  ["archives of neuroscience"] = "Arch Neurosci",
  ["archives of obstetrics and gynaecology"] = "Arch Obstet Gynaecol",
  ["archives of ophthalmology"] = "Arch. Ophthalmol.",
  ["archives of ophthalmology (chicago, ill. : 1929)"] = "Arch Ophthal",
  ["archives of ophthalmology (chicago, ill. : 1960)"] = "Arch Ophthalmol",
  ["archives of oral biology"] = "Arch Oral Biol",
  ["archives of orthopaedic and trauma surgery"] = "Arch Orthop Trauma Surg",
  ["archives of orthopaedic and traumatic surgery"] = "Arch. Orthop. Trauma. Surg.",
  ["archives of orthopaedic and traumatic surgery. archiv fur orthopadische und unfall-chirurgie"] = "Arch Orthop Trauma Surg",
  ["archives of orthopaedic and traumatic surgery. archiv für orthopädische und unfall-chirurgie"] = "Arch Orthop Trauma Surg",
  ["archives of orthopedics and rheumatology"] = "Arch Orthop Rheumatol",
  ["archives of osteoporosis"] = "Arch Osteoporos",
  ["archives of oto-rhino-laryngology"] = "Arch Otorhinolaryngol",
  ["archives of oto-rhino-laryngology. supplement"] = "Arch. Otorhinolaryngol. Suppl.",
  ["archives of oto-rhino-laryngology. supplement = archiv für ohren-, nasen- und kehlkopfheilkunde. supplement"] = "Arch Otorhinolaryngol Suppl",
  ["archives of otolaryngology"] = "Arch Otolaryngol",
  ["archives of otolaryngology (chicago, ill. : 1960)"] = "Arch Otolaryngol",
  ["archives of otolaryngology – head and neck surgery"] = "Arch. Otolaryngol. Head Neck Surg.",
  ["archives of otolaryngology--head & neck surgery"] = "Arch Otolaryngol Head Neck Surg",
  ["archives of palliative care and medicine"] = "Arch Palliat Care Med",
  ["archives of pathology"] = "Arch Pathol (Chic)",
  ["archives of pathology & laboratory medicine"] = "Arch Pathol Lab Med",
  ["archives of pathology and laboratory medicine"] = "Arch. Pathol. Lab. Med.",
  ["archives of pediatric infectious diseases"] = "Arch Pediatr Infect Dis",
  ["archives of pediatrics"] = "Arch Pediatr",
  ["archives of pediatrics & adolescent medicine"] = "Arch Pediatr Adolesc Med",
  ["archives of pediatrics and adolescent medicine"] = "Arch. Pediatr. Adolesc. Med.",
  ["archives of pharmacal research"] = "Arch Pharm Res",
  ["archives of pharmacology and therapeutics"] = "Arch Pharmacol Ther",
  ["archives of pharmacy practice"] = "Arch Pharm Pract",
  ["archives of physical medicine"] = "Arch Phys Med",
  ["archives of physical medicine and rehabilitation"] = "Arch Phys Med Rehabil",
  ["archives of physical therapy"] = "Arch Phys Ther",
  ["archives of physiology and biochemistry"] = "Arch Physiol Biochem",
  ["archives of physiotherapy"] = "Arch Physiother",
  ["archives of phytopathology and plant protection"] = "Arch. Phytopathol. Plant Prot.",
  ["archives of plastic surgery"] = "Arch Plast Surg",
  ["archives of proteomics and bioinformatics"] = "Arch Proteom Bioinform",
  ["archives of psychiatric nursing"] = "Arch Psychiatr Nurs",
  ["archives of psychology (chicago, ill.)"] = "Arch Psychol (Chic)",
  ["archives of public health = archives belges de santé publique"] = "Arch Public Health",
  ["archives of razi institute"] = "Arch Razi Inst",
  ["archives of rehabilitation research and clinical translation"] = "Arch Rehabil Res Clin Transl",
  ["archives of rheumatology"] = "Arch Rheumatol",
  ["archives of scientific psychology"] = "Arch Sci Psychol",
  ["archives of sexual behavior"] = "Arch Sex Behav",
  ["archives of std/hiv research"] = "Arch STD HIV Res",
  ["archives of stem cell and therapy"] = "Arch Stem Cell Ther",
  ["archives of stem cell research"] = "Arch Stem Cell Res",
  ["archives of suicide research : official journal of the international academy for suicide research"] = "Arch Suicide Res",
  ["archives of surgery"] = "Arch Surg",
  ["archives of surgery (chicago, ill. : 1920)"] = "Arch Surg",
  ["archives of surgery (chicago, ill. : 1960)"] = "Arch Surg",
  ["archives of toxicology"] = "Arch Toxicol",
  ["archives of toxicology. supplement"] = "Arch. Toxicol. Suppl.",
  ["archives of toxicology. supplement. = archiv für toxikologie. supplement"] = "Arch Toxicol Suppl",
  ["archives of trauma research"] = "Arch Trauma Res",
  ["archives of virology"] = "Arch Virol",
  ["archives of virology. supplementum"] = "Arch Virol Suppl",
  ["archives of women's mental health"] = "Arch Womens Ment Health",
  ["archives of women’s mental health"] = "Arch Womens Ment Health",
  ["archives roumaines de pathologie experimentale et de microbiologie"] = "Arch. Roum. Pathol. Exp. Microbiol.",
  ["archives roumaines de pathologie experimentales et de microbiologie"] = "Arch Roum Pathol Exp Microbiol",
  ["archives roumaines de pathologie expérimentales et de microbiologie"] = "Arch Roum Pathol Exp Microbiol",
  ["archives suisses d'anthropologie générale"] = "Arch Suisses Anthropol Gen",
  ["archives. institut pasteur du maroc"] = "Arch Inst Pasteur Maroc",
  ["archives. middlesex hospital"] = "Arch Middx Hosp",
  ["archivi. archivi d’italia e rassegna internazionale degli archivi"] = "Archivi",
  ["archivii italiani di laringologia"] = "Arch Ital Laringol",
  ["archiving : final program and proceedings. is & t's archiving conference"] = "Archiving",
  ["archivio - ospedale al mare"] = "Arch Osp Mare",
  ["archivio de vecchi per l’anatomia patologica e la medicina clinica"] = "Arch. de Vecchi Anat. Patol.",
  ["archivio della corrispondenza degli scienziati italiani"] = "Arch. Corrisp. Sci. Italiani",
  ["archivio della societa romana di storia patria"] = "Arch Soc Rom Stor Patria",
  ["archivio della società romana di storia patria"] = "Arch Soc Rom Stor Patria",
  ["archivio della società romana di storia patria"] = "ArchStorRom",
  ["archivio di antropologia criminale, psichiatria e medicina legale"] = "Arch Antropol Crim Psichiatr Med Leg",
  ["archivio di chirurgia del torace"] = "Archivio Chir Torace",
  ["archivio di chirurgia toracica e cardiovascolare"] = "Arch Chir Torac Cardiovasc",
  ["archivio di farmacologia sperimentale e scienze affini"] = "Arch Farmacol Sper Sci Affin",
  ["archivio di fisiologia"] = "Arch Fisiol",
  ["archivio di medicina interna"] = "Arch Med Interna",
  ["archivio di medicina mutualistica"] = "Arch Med Mutual",
  ["archivio di ortopedia"] = "Arch Ortop",
  ["archivio di ostetricia e ginecologia"] = "Arch Ostet Ginecol",
  ["archivio di ottalmologia"] = "Arch Ottamol",
  ["archivio di patologia e clinica medica"] = "Arch Patol Clin Med",
  ["archivio di psicologia, neurologia e psichiatria"] = "Arch Psicol Neurol Psichiatr",
  ["archivio di radiologia"] = "Arch Radiol",
  ["archivio di scienze biologiche"] = "Arch Sci Biol (Bologna)",
  ["archivio di tisiologia"] = "Arch Tisiol",
  ["archivio di tisiologia e delle malattie dell apparato respiratorio"] = "Arch. Tisiol. Mal. Appar. Respir.",
  ["archivio di tisiologia e delle malattie dell'apparato respiratorio"] = "Arch Tisiol Mal Appar Respir",
  ["archivio di tisiologia e delle malattie dell’apparato respiratorio"] = "Arch Tisiol Mal Appar Respir",
  ["archivio ed atti. società italiana di chirurgia"] = "Arch Atti",
  ["archivio glottologico italiano"] = "AGI",
  ["archivio internazionale di studi neurologici"] = "Arch Internazionale Studi Neurol",
  ["archivio italiano delle malattie dell apparato digerente"] = "Arch. Ital. Mal. Appar. Dig.",
  ["archivio italiano delle malattie dell'apparato digerente"] = "Arch Ital Mal Appar Dig",
  ["archivio italiano delle malattie dell’apparato digerente"] = "Arch Ital Mal Appar Dig",
  ["archivio italiano di anatomia e di embriologia"] = "Arch. Ital. Anat. Embriol.",
  ["archivio italiano di anatomia e di embriologia. italian journal of anatomy and embryology"] = "Arch Ital Anat Embriol",
  ["archivio italiano di anatomia e istologia patologica"] = "Arch Ital Anat Istol Patol",
  ["archivio italiano di chirurgia"] = "Arch Ital Chir",
  ["archivio italiano di dermatologia, sifilografia, e venereologia"] = "Arch Ital Dermatol Sifilogr Venereol",
  ["archivio italiano di dermatologia, venereologia e sessuologia"] = "Arch. Ital. Dermatol. Venereol. Sessuol.",
  ["archivio italiano di dermatologia, venereologia, e sessuologia"] = "Arch Ital Dermatol Venereol Sessuol",
  ["archivio italiano di otologia, rinologia e laringologia"] = "Arch Ital Otol Rinol Laringol",
  ["archivio italiano di otologia, rinologia e laringologia. supplemento"] = "Arch Ital Otol Rinol Laringol Suppl",
  ["archivio italiano di otologia, rinologia, laringologia e patologia cervico-facciale"] = "Arch. Ital. Otol. Rinol. Laringol. Patol. Cervicofacc.",
  ["archivio italiano di otologia, rinologia, laringologia e patologia cervico-facciale. supplemento"] = "Arch Ital Otol Rinol Laringol Patol Cervicofacc Suppl",
  ["archivio italiano di otologia, rinologia, laringologia, e patologia cervico-facciale"] = "Arch Ital Otol Rinol Laringol Patol Cervicofacc",
  ["archivio italiano di patologia e clinica dei tumori"] = "Arch Ital Patol Clin Tumori",
  ["archivio italiano di pediatria e puericoltura"] = "Arch Ital Pediatr Pueric",
  ["archivio italiano di scienze farmacologiche"] = "Arch Ital Sci Farmacol",
  ["archivio italiano di scienze mediche tropicali e di parassitologia"] = "Arch Ital Sci Med Trop Parassitol",
  ["archivio italiano di urologia"] = "Arch Ital Urol",
  ["archivio italiano di urologia e nefrologia"] = "Arch Ital Urol Nefrol",
  ["archivio italiano di urologia, andrologia"] = "Arch. Ital. Urol. Androl.",
  ["archivio italiano di urologia, andrologia : organo ufficiale [di] societa italiana di ecografia urologica e nefrologica / associazione ricerche in urologia"] = "Arch Ital Urol Androl",
  ["archivio italiano di urologia, andrologia : organo ufficiale [di] società italiana di ecografia urologica e nefrologica"] = "Arch Ital Urol Androl",
  ["archivio italiano di urologia, nefrologia, andrologia"] = "Arch. Ital. Urol. Nefrol. Androl.",
  ["archivio italiano di urologia, nefrologia, andrologia : organo ufficiale dell'associazione per la ricerca in urologia = urological, nephrological, and andrological sciences"] = "Arch Ital Urol Nefrol Androl",
  ["archivio monaldi per la tisiologia e le malattie dell apparato respiratorio"] = "Arch. Monaldi",
  ["archivio monaldi per la tisiologia e le malattie dell'apparato respiratorio"] = "Arch Monaldi",
  ["archivio monaldi per la tisiologia e le malattie dell’apparato respiratorio"] = "Arch Monaldi",
  ["archivio monaldi per le malattie del torace"] = "Arch Monaldi Mal Torace",
  ["archivio per le scienze mediche"] = "Arch Sci Med (Torino)",
  ["archivio putti di chirurgia degli organi di movimento"] = "Arch. Putti Chir. Organi Mov.",
  ["archivio siciliano di medicina e chirurgia. 4, acta medica mediterranea"] = "Arch Sicil Med Chir 4 Acta Med Mediterr",
  ["archivio stomatologico"] = "Arch Stomatol (Napoli)",
  ["archivio storico italiano"] = "Arch Stor Ital",
  ["archivio storico lombardo"] = "Arch Stor Lomb",
  ["archivio storico per la calabria e la lucania"] = "ASCL",
  ["archivio storico per la sicilia orientale"] = "ArchStorSicOr",
  ["archivio storico per le province napoletane"] = "Arch Stor Prov Napol",
  ["archivio storico per le province parmensi"] = "Arch Stor Prov Parm",
  ["archivio storico pugliese"] = "ASP",
  ["archivio storico sardo"] = "ASSARD",
  ["archivio storico siciliano (palermo, italy : 1946)"] = "Arch Stor Sicil",
  ["archivio storico siracusano"] = "ArchStorSir",
  ["archivio veneto (1927)"] = "Arch Veneto",
  ["archivio veterinario italiano"] = "Arch Vet Ital",
  ["archivium hibernicum : or, irish historical records"] = "Archiv Hibernicum",
  ["archivni casopis"] = "Arch Cas",
  ["archivní časopis"] = "Arch Cas",
  ["archivo de prehistoria levantina"] = "ArchPrehistLev",
  ["archivo espanol de morfologia"] = "Arch. Esp. Morfol.",
  ["archivo español de arte"] = "Arch Esp Arte",
  ["archivo español de morfología"] = "Arch Esp Morfol",
  ["archivo español de arqueología"] = "AEA",
  ["archivo hispalense"] = "Arch Hisp",
  ["archivo ibero-americano"] = "Arch Ibero Am",
  ["archivos americanos de medicina"] = "Arch Am Med",
  ["archivos argentinos de dermatologia"] = "Arch. Argent. Dermatol.",
  ["archivos argentinos de dermatología"] = "Arch Argent Dermatol",
  ["archivos argentinos de enfermedades del aparato digestivo"] = "Arch Argent Enferm Apar Dig",
  ["archivos argentinos de kinesiología; kinesiterapia y kinefilaxia"] = "Arch Argent Kinesiol Kinesiter Kinefilax",
  ["archivos argentinos de pediatria"] = "Arch. Argent. Pediatr.",
  ["archivos argentinos de pediatría"] = "Arch Argent Pediatr",
  ["archivos argentinos de tisiologia"] = "Arch Argent Tisiol",
  ["archivos argentinos de tisiologia y neumonologia"] = "Arch Argent Tisiol Neumonol",
  ["archivos argentinos de tisiología"] = "Arch Argent Tisiol",
  ["archivos argentinos de tisiología y neumonología"] = "Arch Argent Tisiol Neumonol",
  ["archivos brasileiros de medicina"] = "Arch Bras Med",
  ["archivos cubanos de cancerologia"] = "Arch Cuba Cancerol",
  ["archivos cubanos de cancerología"] = "Arch Cuba Cancerol",
  ["archivos de biologia andina"] = "Arch. Biol. Andina",
  ["archivos de biologia y medicina experimentales"] = "Arch. Biol. Med. Exp. (Santiago)",
  ["archivos de biología andina"] = "Arch Biol Andina",
  ["archivos de biología y medicina experimentales"] = "Arch Biol Med Exp",
  ["archivos de bioquímica, química y farmacia, tucumán"] = "Arch Bioquim Quim Farm Tucuman",
  ["archivos de bronconeumologia"] = "Arch Bronconeumol",
  ["archivos de bronconeumología"] = "Arch Bronconeumol",
  ["archivos de cardiologia de mexico"] = "Arch. Cardiol. Mex.",
  ["archivos de cardiología de méxico"] = "Arch Cardiol Mex",
  ["archivos de enfermedades del corazon y vasos"] = "Arch. Enferm. Coraz. Vasos",
  ["archivos de enfermedades del corazón y vasos"] = "Arch Enferm Coraz Vasos",
  ["archivos de farmacologia y toxicologia"] = "Arch. Farmacol. Toxicol.",
  ["archivos de farmacología y toxicología"] = "Arch Farmacol Toxicol",
  ["archivos de ginecología y obstetricia"] = "Arch Ginecol Obstet",
  ["archivos de histologia normal y patologica"] = "Arch Histol Norm Patol",
  ["archivos de histologia normal y patologica."] = "Arch. Histol. Norm. Patol.",
  ["archivos de histología normal y patológica"] = "Arch Histol Norm Patol",
  ["archivos de historia da medicina portugueza"] = "Arch Hist Med Port",
  ["archivos de investigacion medica"] = "Arch. Invest. Med. (Mex.)",
  ["archivos de investigación médica"] = "Arch Invest Med (Mex)",
  ["archivos de la fundacion roux-ocefa"] = "Arch. Fund. Roux. Ocefa",
  ["archivos de la fundación roux-ocefa"] = "Arch Fund Roux Ocefa",
  ["archivos de la sociedad de cirujanos de chile"] = "Arch Soc Cir Chile",
  ["archivos de la sociedad espanola de oftalmologia"] = "Arch Soc Esp Oftalmol",
  ["archivos de la sociedad española de oftalmología"] = "Arch Soc Esp Oftalmol",
  ["archivos de la sociedad oftalmologica hispano-americana"] = "Archivos Soc Oftalmol Hisp Am",
  ["archivos de la sociedad oftalmológica hispano-americana"] = "Archivos Soc Oftalmol Hisp Am",
  ["archivos de medicina del deporte : publicación de la federación española de medicina del deporte"] = "Arch Med Deporte",
  ["archivos de medicina experimental; trabajos del instituto nacional de ciencias medicas"] = "Arch Med Exp",
  ["archivos de medicina experimental; trabajos del instituto nacional de ciencias médicas"] = "Arch Med Exp",
  ["archivos de medicina infantil"] = "Arch Med Infant",
  ["archivos de medicina interna (montevideo, uruguay)"] = "Arch Med Intern",
  ["archivos de medicina veterinaria"] = "Arch Med Vet",
  ["archivos de neurobiologia"] = "Arch. Neurobiol. (Madr.)",
  ["archivos de neurobiologiá"] = "Arch Neurobiol (Madr)",
  ["archivos de neurocirugía"] = "Arch Neurocir",
  ["archivos de neurología y psiquiatría de méxico : organo de la sociedad mexicana de neurología y psiquiatría"] = "Arch Neurol Psiquiatr Mex",
  ["archivos de odonto estomatologia"] = "Arch Odonto Estomatol",
  ["archivos de oftalmologia de buenos aires"] = "Arch. Oftalmol. B. Aires",
  ["archivos de oftalmología de buenos aires"] = "Arch Oftalmol B Aires",
  ["archivos de pediatria"] = "Arch Pediatr (Barc)",
  ["archivos de pediatria del uruguay"] = "Arch. Pediatr. Urug.",
  ["archivos de pediatría"] = "Arch Pediatr (Barc)",
  ["archivos de pediatría del uruguay"] = "Arch Pediatr Urug",
  ["archivos de prevención de riesgos laborales"] = "Arch Prev Riesgos Labor",
  ["archivos del colegio medico de el salvador"] = "Arch. Col. Med. El Salv.",
  ["archivos del colegio médico de el salvador"] = "Arch Col Med El Salv",
  ["archivos del instituto de biologia andina"] = "Arch. Inst. Biol. Andina",
  ["archivos del instituto de biología andina"] = "Arch Inst Biol Andina",
  ["archivos del instituto de cardiologia de mexico"] = "Arch Inst Cardiol Mex",
  ["archivos del instituto de cardiología de méxico"] = "Arch Inst Cardiol Mex",
  ["archivos del instituto de farmacologia experimental"] = "Arch. Inst. Farmacol. Exp. (Madr.)",
  ["archivos del instituto de farmacologia experimental (medicina)"] = "Arch Inst Farmacol Exp (Madr)",
  ["archivos del instituto de farmacología experimental (medicina)"] = "Arch Inst Farmacol Exp (Madr)",
  ["archivos dominicanos de pediatria"] = "Arch Domin Pediatr",
  ["archivos dominicanos de pediatría"] = "Arch Domin Pediatr",
  ["archivos espanoles de urologia"] = "Arch Esp Urol",
  ["archivos españoles de urología"] = "Arch Esp Urol",
  ["archivos latinoamericanos de nutricion"] = "Arch Latinoam Nutr",
  ["archivos latinoamericanos de nutrición"] = "Arch Latinoam Nutr",
  ["archivos medicos de cuba"] = "Arch Med Cuba",
  ["archivos medicos mexicanos"] = "Arch Medicos Mex",
  ["archivos medicos panamenos"] = "Arch. Med. Panamenos",
  ["archivos mexicanos de venereología y dermatología"] = "Arch Mex Venereol Dermatol",
  ["archivos médicos de cuba"] = "Arch Med Cuba",
  ["archivos médicos mexicanos"] = "Arch Medicos Mex",
  ["archivos médicos panameños"] = "Arch Med Panamenos",
  ["archivos peruanos de patología y clínica"] = "Arch Peru Patol Clin",
  ["archivos uruguayos de medicina, cirugia y especialidades"] = "Arch Urug Med Cir Espec",
  ["archivos uruguayos de medicina, cirugía y especialidades"] = "Arch Urug Med Cir Espec",
  ["archivos venezolanos de medicina tropical y parasitologia medica"] = "Arch Venez Med Trop Parasitol Med",
  ["archivos venezolanos de medicina tropical y parasitología médica"] = "Arch Venez Med Trop Parasitol Med",
  ["archivos venezolanos de puericultura y pediatria"] = "Arch Venez Pueric Pediatr",
  ["archivos venezolanos de puericultura y pediatría"] = "Arch Venez Pueric Pediatr",
  ["archivos. argentina. secretaría de salud pública"] = "Arch Argent Repub Secr Salud Publica",
  ["archivos. asociación para evitar la ceguera en méxico"] = "Arch Asoc Evit Ceguera Mex",
  ["archivos. hospital santo tomás (panama, panama)"] = "Arch Panama City",
  ["archivos. hospital universitario general calixto garcía"] = "Arch Hosp Univ",
  ["archivos. instituto nacional de hidrología y climatología médicas (cuba)"] = "Arch Cuba",
  ["archivos. sociedad argentina de anatomía normal y patológica"] = "Arch Soc Agent Anat Norm Patol",
  ["archivos. sociedad de biología de montevideo"] = "Arch Soc Biol Montev",
  ["archivos. sociedad de cirujanos de hospital, santiago de chile"] = "Arch Soc Cir Hosp",
  ["archivos. sociedad de estudios clínicos de la habana"] = "Arch Soc Estud Clin Habana",
  ["archivum chirurgicum neerlandicum"] = "Arch Chir Neerl",
  ["archivum combustionis"] = "Arch. Combust.",
  ["archivum franciscanum historicum : periodica publicatio trimestris cura pp. collegii d. bonaventurae"] = "Arch Franciscanum Hist",
  ["archivum histologicum japonicum"] = "Arch. Histol. Jpn.",
  ["archivum histologicum japonicum = nihon soshikigaku kiroku"] = "Arch Histol Jpn",
  ["archivum histologicum japonicum. nippon soshikigaku kiroku"] = "Arch Histol Jpn",
  ["archivum historiae pontificiae"] = "Arch Hist Pontif",
  ["archivum immunolgiae et therapiae experimentalis"] = "Arch. Immunol. Ther. Exp.",
  ["archivum immunologiae et therapiae experimentalis"] = "Arch Immunol Ther Exp (Warsz)",
  ["archivum latinitatis medii aevi (bulletin du cange)"] = "ALMA",
  ["archivum ottomanicum"] = "Arch Ottoman",
  ["archivum societatis zoologicae botanicae fennicae vanamo"] = "Arch. Soc. Zool. Bot. Fenn. Vanamo",
  ["archivum veterinarium polonicum"] = "Arch Vet Pol",
  ["archivum veterinarium polonicum / polish academy of sciences, committee of veterinary sciences"] = "Arch Vet Pol",
  ["archiwum historii i filozofii medycyny"] = "Arch Hist Filoz Med",
  ["archiwum historii i filozofii medycyny / polskii towarzystwo historii medycyny i farmacji"] = "Arch Hist Filoz Med",
  ["archiwum historii medycyny"] = "Arch Hist Med (Warsz)",
  ["archiwum medycyny sadowej i kryminologii"] = "Arch Med Sadowej Kryminol",
  ["archiwum medycyny sa̧dowej i kryminologii"] = "Arch Med Sadowej Kryminol",
  ["archív orientální. quarterly journal of african, asian and latin american studies. monografie archívu orientálního"] = "ArOrMono",
  ["archív orientální. quarterly journal of african, asian and latin american studies. supplementa"] = "ArOrSuppl",
  ["archäographie. archäologie und elektronische datenverarbeitung"] = "Archäographie",
  ["archäologie der schweiz"] = "Archäol. Schweiz",
  ["archäologie der schweiz. mitteilungsblatt der schweizerischen gesellschaft für ur- und frühgeschichte"] = "ASchw",
  ["archäologie der schweiz: mitteilungsblatt der schweizerischen gesellschaft für ur- und frühgeschichte"] = "ArchS",
  ["archäologie des kantons solothurn"] = "ASoloth",
  ["archäologie im kanton bern: fundberichte und aufsätze"] = "AKBern",
  ["archäologie in deutschland"] = "AiD",
  ["archäologie in sachsen-anhalt"] = "ASachs",
  ["archäologie und naturwissenschaften"] = "ANaturwiss",
  ["archäologische ausgrabungen in baden-württemberg"] = "AausgrBadWürt",
  ["archäologische berichte aus dem yemen"] = "ABADY",
  ["archäologische forschungen"] = "AF",
  ["archäologische informationen. mitteilungen zur ur- und frühgeschichte"] = "AInf",
  ["archäologische mitteilungen aus iran"] = "AMI",
  ["archäologische mitteilungen aus iran und turan"] = "AMIT",
  ["archäologische nachrichten aus baden"] = "ANachrBad",
  ["archäologische sammlung der universität zürich"] = "ASammlUnZuerch",
  ["archäologische veröffentlichungen. deutsches archäologisches institut, abteilung kairo"] = "AV",
  ["archäologische zeitung"] = "AZ",
  ["archäologischer anzeiger"] = "AA",
  ["archäologisches korrespondenzblatt"] = "AKorrBl",
  ["archäologisches korrespondenzblatt: urgeschichte, römerzeit, frühmittelalter"] = "AKB",
  ["archäologisches nachrichtenblatt"] = "ANachr",
  ["archéologie du nil moyen"] = "ANilMoy",
  ["archéologie en bretagne. bulletin d’information"] = "ABret",
  ["archéologie neuchâteloise"] = "Archéol. neuchâtel.",
  ["archív orientální. quarterly journal of african and asian studies"] = "ArOr",
  ["arcispedale s. anna di ferrara"] = "Arcisp. S. Anna Ferrara",
  ["arctic and alpine research"] = "Arct. Alp. Res.",
  ["arctic anthropology"] = "Arctic Anthropol",
  ["arctic environmental research"] = "Arct Environ Res",
  ["arctic medical research"] = "Arctic Med Res",
  ["arctic science"] = "Arct. Sci.",
  ["arctic, antarctic, and alpine research"] = "Arct Antarct Alp Res",
  ["arctic, antartic, and alpine research"] = "Arct. Antarct. Alp. Res.",
  ["arctos. acta philologica fennica"] = "Arctos",
  ["area (oxford, england)"] = "Area (Oxf)",
  ["arealstatistik der schweiz"] = "Arealstat. Schweiz",
  ["arerugi (japanese journal of allergology)"] = "Arerugi",
  ["arerugi. japanese journal of allergology"] = "Arerugi",
  ["arerugī = [allergy]"] = "Arerugi",
  ["argeion oikonomikes istorias (archives of economic history, athens)"] = "AEH",
  ["argentine republic. laws, statutes, etc"] = "Argent Repub Laws Statut",
  ["argo. časopis slovenskih muzejev, narodni muzej slovenije"] = "Argo",
  ["argos (utrecht, netherlands)"] = "Argos",
  ["argumentation library"] = "Argum. Libr.",
  ["arheologia moldovei"] = "AMold",
  ["arheologiâ, publ. par l’inst. d’archéol. de l’acad. des sciences d’ukraine et par la soc. ukrainienne pour la préservation des monuments culturels et historiques"] = "ARH",
  ["arheolokiradovi i rasprave"] = "ARadRaspr",
  ["arheološki pregled. arheološko društvo jugoslavije"] = "APregl",
  ["arheološki vestnik (ljubljana)"] = "AVes",
  ["arhiv za farmaciju"] = "Arh Farm (Belgr)",
  ["arhiv za higijenu rada"] = "Arh Hig Rada",
  ["arhiv za higijenu rada i toksikologiju"] = "Arh Hig Rada Toksikol",
  ["arhiv za zastitu majke i djeteta. archives of mother and child welfare. archives pour la protection de la mere et de l’enfant"] = "Arh Zast",
  ["arhiv za zas̆titu majke i djeteta. archives of mother and child welfare. archives pour la protection de la mère et de l'enfant"] = "Arh Zast",
  ["arid ecosystems"] = "Arid Ecosyst.",
  ["arid land research and management"] = "Arid Land Res. Manage.",
  ["ariel. ariʼel"] = "Ariel (Jerus)",
  ["ariel. ari’el"] = "Ariel (Jerus)",
  ["arima. revue africaine de la recherche en informatique et mathematiques appliquees"] = "ARIMA Rev. Afr. Rech. Inform. Math. Appl.",
  ["aristote. traductions et etudes"] = "Aristote. Trad. Etudes",
  ["arizona and the west"] = "Ariz West",
  ["arizona dental journal"] = "Ariz. Dent. J.",
  ["arizona geriatrics society journal"] = "Ariz Geriatr Soc J",
  ["arizona journal of environmental law & policy"] = "Arizona J Environ Law Policy",
  ["arizona journal of international and comparative law"] = "Ariz J Int Comp Law",
  ["arizona law review"] = "Ariz Law Rev",
  ["arizona medicine"] = "Ariz Med",
  ["arizona nurse"] = "Ariz. Nurse",
  ["arizona revised statutes, annotated : prepared under legislative authority, laws 1956, chapter 129. arizona"] = "Ariz Revis Statut Annot Ariz",
  ["arizona revised statutes. arizona"] = "Ariz Revis Statut",
  ["arizona state law journal"] = "Ariz State Law J",
  ["ark-light newsletter"] = "Ark. Light Newsl.",
  ["arkansas archeology"] = "Ark Archeol",
  ["arkansas dental journal"] = "Ark. Dent. J.",
  ["arkansas dentistry"] = "Ark Dent",
  ["arkansas law review"] = "Ark Law Rev",
  ["arkeoloji dergisi. ege üniversitesi edebiyat fakültesi"] = "ADerg",
  ["arkhiv anatomii, gistologii i embriologii"] = "Arkh. Anat. Gistol. Embriol.",
  ["arkhiv anatomii, gistologii i émbriologii"] = "Arkh Anat Gistol Embriol",
  ["arkhiv patologii"] = "Arkh Patol",
  ["arkiv foer fysik"] = "Ark. Fys.",
  ["arkiv foer geohysik"] = "Ark. Geofys.",
  ["arkiv foer kemi, mineralogi och geologi"] = "Ark. Kemi Mineral. Geol.",
  ["arkiv foer matematik"] = "Ark. Mat.",
  ["arkiv foer matematik, astronomi, och fysik"] = "Ark. Mat. Astron. Fys.",
  ["arkiv foer physik"] = "Ark. Fys.",
  ["arkiv for matematik"] = "Ark. Mat.",
  ["arkiv for matematik, astronomi, och fysik"] = "Ark. Mat. Astron. Fys.",
  ["arkiv für matematik"] = "Ark. Mat.",
  ["arkivoc (gainesville, fl, united states)"] = "ARKIVOC (Gainesville, FL, U. S.)",
  ["arkivoc : free online journal of organic chemistry"] = "ARKIVOC",
  ["arkivoc : free online journal of organic chemistry / arkat-usa, inc"] = "ARKIVOC",
  ["arktos : the journal of arctic geosciences"] = "Arktos",
  ["armed forces & society"] = "Armed Forces Soc.",
  ["armed forces and society"] = "Armed Forces Soc",
  ["armed forces medical journal, india"] = "Armed Forces Med J India",
  ["armenian journal of mathematics"] = "Armen. J. Math.",
  ["army materials technology conference series"] = "Army Materials Tech. Conf. Ser.",
  ["army medical corps journal, india"] = "Army Med Corps J India",
  ["army medical department bulletin"] = "Army Med Dep Bull",
  ["arn journal"] = "ARN J.",
  ["arn journal : the official journal of the association of rehabilitation nurses"] = "ARN J",
  ["arne ryde memorial lectures series"] = "Arne Ryde Meml. Lect. Ser.",
  ["arnold applications of statistics"] = "Arnold Appl. Stat.",
  ["arnold mathematical journal"] = "Arnold Math. J.",
  ["arqueologia e história"] = "AHist",
  ["arqueologia mexicana"] = "Arqueol. Mex.",
  ["arquipélago : boletim da universidade dos açores. ciências biológicas e marinhas = bulletin of the university of the azores. life and marine sciences"] = "Arquipelago",
  ["arquivo brasileiro de medicina veterinária e zootecnia"] = "Arq Bras Med Vet Zootec",
  ["arquivo de beja. boletim, estudos, arquivo"] = "ArqBeja",
  ["arquivo de patologia"] = "Arq Patol",
  ["arquivo do instituto gulbenkian de ciência"] = "Arquivo Inst.",
  ["arquivo: orgao da sociedade brasileira de implantologia bucal"] = "Arquivo",
  ["arquivo; orgao da sociedade brasileira de implantologia bucal"] = "Arquivo",
  ["arquivos brasileiros de cardiologia"] = "Arq Bras Cardiol",
  ["arquivos brasileiros de cirurgia digestiva : abcd = brazilian archives of digestive surgery"] = "Arq Bras Cir Dig",
  ["arquivos brasileiros de endocrinologia e metabologia"] = "Arq Bras Endocrinol Metabol",
  ["arquivos brasileiros de medicina"] = "Arq Bras Med",
  ["arquivos brasileiros de medicina naval"] = "Arq Bras Med Nav",
  ["arquivos brasileiros de oftalmologia"] = "Arq Bras Oftalmol",
  ["arquivos brasileiros de psicologia"] = "Arq Bras Psicol",
  ["arquivos da faculdade de higiene e saude publica da universidade de sao paulo"] = "Arq. Fac. Hig. Saude Publica Univ. Sao Paulo",
  ["arquivos da faculdade de higiene e saúde pública da universidade de são paulo"] = "Arq Fac Hig Saude Publica Univ Sao Paulo",
  ["arquivos da polícia civil de são paulo"] = "Arq Polic Civ Sao Paul",
  ["arquivos de cirurgia clinica e experimental"] = "Arq Cir Clin Exp",
  ["arquivos de clinica"] = "Arq Clin (Rio De J)",
  ["arquivos de clínica"] = "Arq Clin (Rio De J)",
  ["arquivos de gastroenterologia"] = "Arq Gastroenterol",
  ["arquivos de higiene e saude publica"] = "Arq. Hig. Saude Publica",
  ["arquivos de higiene e saúde pública"] = "Arq Hig Saude Publica",
  ["arquivos de medicina : revista de ciência e arte médicas"] = "Arq Med",
  ["arquivos de medicina e cirurgia de pernambuco"] = "Arq Med Cirur Pernamb",
  ["arquivos de neuro-psiquiatria"] = "Arq Neuropsiquiatr",
  ["arquivos de oncologia"] = "Arq Oncol",
  ["arquivos de oncología"] = "Arq Oncol",
  ["arquivos de zoologia"] = "Arq Zool",
  ["arquivos do centro cultural portugues. centre culturel portugais"] = "Arq Cent Cult Port",
  ["arquivos do centro cultural português. centre culturel portugais"] = "Arq Cent Cult Port",
  ["arquivos do centro de estudos da faculdade de odontologia da u. f. m. g"] = "Arq Cent Estud Fac Odontol UFMG (Belo Horiz)",
  ["arquivos do centro de estudos da faculdade de odontologia da u.f.m.g."] = "Arq. Cent. Estud. Fac. Odontol. UFMG. (Belo. Horiz.)",
  ["arquivos do centro de estudos do curso de odontologia"] = "Arq Cent Estud Curso Odontol",
  ["arquivos do centro de estudos, curso de odontologia universidade federal de minas gerais"] = "Arq. Cent. Estud. Curso Odontol. Univ. Fed. Minas Gerais",
  ["arquivos do instituto biologico"] = "Arq. Inst. Biol. (Sao Paulo)",
  ["arquivos do instituto biológico"] = "Arq Inst Biol (Sao Paulo)",
  ["arquivos do museu nacional. museu nacional (brazil)"] = "Arq Mus Nac",
  ["arquivos mineiros de leprologia"] = "Arq Min Leprol",
  ["arquivos portugueses de bioquimica"] = "Arq Port Bioquim",
  ["arquivos portugueses de bioquímica"] = "Arq Port Bioquim",
  ["arquivos portugueses de oftalmologia"] = "Arq Port Oftalmol",
  ["arquivos. coimbra. universidade. instituto de farmacologia e terapeutica experimental"] = "Arq Inst Farm",
  ["arquivos. coimbra. universidade. instituto de farmacologia e terapêutica experimental"] = "Arq Inst Farm",
  ["arquivos. instituto brasileiro para investigacao da tuberculose"] = "Arq Inst Bras Invest Tubers",
  ["arquivos. instituto brasileiro para investigação da tuberculose"] = "Arq Inst Bras Invest Tubers",
  ["arquivos. rio grande do sul (brazil : state). departamento estadual de saúde"] = "Arq Rio Gd Sul Braz State",
  ["arrhythmia & electrophysiology review"] = "Arrhythm Electrophysiol Rev",
  ["arrows for change"] = "Arrows Change",
  ["ars combinatoria"] = "Ars Combin.",
  ["ars combinatoria. a canadian journal of combinatorics"] = "Ars Combin.",
  ["ars curandi em odontologia"] = "Ars Curandi Odontol",
  ["ars georgica"] = "ArsGeorg",
  ["ars inveniendi analytica"] = "Ars Inven. Anal.",
  ["ars mathematica contemporanea"] = "Ars Math. Contemp.",
  ["ars medici"] = "Ars Med",
  ["ars medici; revue internationale pour le médecin praticien"] = "Ars Med Rev Int Med Prat",
  ["arsbok. goteborgs tandlakare-sallskap"] = "Arsb Goteb Tandlak Sallsk",
  ["arsbok. odontologiska samfundet i finland"] = "Arsb Odontol Samf Finl",
  ["arse. boletín del centro arqueológico saguntino"] = "Arse",
  ["art et archéologie en rhône-alpes"] = "ArtARhone",
  ["art history"] = "Art Hist",
  ["art in america"] = "Art Am",
  ["art journal"] = "ArtJ",
  ["art therapy : journal of the american art therapy association"] = "Art Ther (Alex)",
  ["arte antica e moderna"] = "ArtAntMod",
  ["arte lombarda"] = "ArtLomb",
  ["arte medievale"] = "ArtMediev",
  ["artech house antenna library"] = "Artech House Antenna Lib.",
  ["artech house telecommunications library"] = "Artech House Telecommun. Libr.",
  ["arteres et veines"] = "Arteres Veines",
  ["arteriosclerosis (dallas, tex.)"] = "Arteriosclerosis",
  ["arteriosclerosis and thrombosis"] = "Arterioscler. Thromb.",
  ["arteriosclerosis and thrombosis : a journal of vascular biology"] = "Arterioscler Thromb",
  ["arteriosclerosis and thrombosis : a journal of vascular biology / american heart association"] = "Arterioscler Thromb",
  ["arteriosclerosis, thrombosis, and vascular biology"] = "Arterioscler Thromb Vasc Biol",
  ["artery research"] = "Artery Res",
  ["artes de mexico"] = "Artes Mex",
  ["artes de méxico"] = "Artes Mex",
  ["artères et veines"] = "Arteres Veines",
  ["artha vijñāna : journal of the gokhale institute of politics and economics, poona (india) = artha vijñāna : gokhale artha śāstra samsthā dvārā prakāśita traimēsika pātrikā"] = "Artha Vijnana",
  ["arthritis & rheumatology (hoboken, n.j.)"] = "Arthritis Rheumatol",
  ["arthritis + rheuma"] = "Arthritis Rheum (Munch)",
  ["arthritis and rheumatism"] = "Arthritis Rheum",
  ["arthritis care & research"] = "Arthritis Care Res (Hoboken)",
  ["arthritis care and research"] = "Arthritis Care Res.",
  ["arthritis care and research : the official journal of the arthritis health professions association"] = "Arthritis Care Res",
  ["arthritis research"] = "Arthritis Res",
  ["arthritis research & therapy"] = "Arthritis Res Ther",
  ["arthritis research and therapy"] = "Arthritis Res. Ther.",
  ["arthroplasty today"] = "Arthroplast Today",
  ["arthropod management tests"] = "Arthropod Manage. Tests",
  ["arthropod structure"] = "Arthropod Struct. Dev.",
  ["arthropod structure & development"] = "Arthropod Struct Dev",
  ["arthropod systematics & phylogeny"] = "Arthropod Syst Phylogeny",
  ["arthropod-plant interactions"] = "Arthropod Plant Interact",
  ["arthropoda selecta"] = "Arthropoda Sel",
  ["arthroscopy : the journal of arthroscopic & related surgery : official publication of the arthroscopy association of north america and the international arthroscopy association"] = "Arthroscopy",
  ["arthroscopy techniques"] = "Arthrosc Tech",
  ["arthroscopy, sports medicine, and rehabilitation"] = "Arthrosc Sports Med Rehabil",
  ["artibus asiae"] = "ArtAs",
  ["artif. intell."] = "Artif. Intell.",
  ["artifical satellites"] = "Artif. Satell.",
  ["artificial cells blood substitutes and immobilization biotechnology"] = "Artif. Cells, Blood Substitues, Immobilization Biotechnol.",
  ["artificial cells, blood substitutes, and immobilization biotechnology"] = "Artif Cells Blood Substit Immobil Biotechnol",
  ["artificial cells, blood substitutes, and immobilization biotechnologyartif cells blood substit immobi"] = "Biotechnol",
  ["artificial cells, nanomedicine, and biotechnology"] = "Artif Cells Nanomed Biotechnol",
  ["artificial dna, pna & xna"] = "Artif DNA PNA XNA",
  ["artificial dna: pna & xna"] = "Artif: DNA: PNA XNA",
  ["artificial intelligence"] = "Artif Intell",
  ["artificial intelligence and applications (commerce, calif.)"] = "Artif Intell Appl",
  ["artificial intelligence and law"] = "Artif Intell Law (Dordr)",
  ["artificial intelligence and soft computing for industrial transformation"] = "Artif. Intell. Soft Comp. Ind. Transform.",
  ["artificial intelligence for engineering design, analysis and manufacturing"] = "Artif. Intell. Eng. Des. Anal. Manuf.",
  ["artificial intelligence for engineering design, analysis and manufacturing : ai edam"] = "Artif Intell Eng Des Anal Manuf",
  ["artificial intelligence in agriculture"] = "Artif. Intell. Agric.",
  ["artificial intelligence in geosciences"] = "Artif. Intell. Geosci.",
  ["artificial intelligence in health : first international workshop, aih 2018, stockholm, sweden, july 13-14, 2018, revised selected papers. aih (workshop) (1st : 2018 : stockholm, sweden)"] = "Artif Intell Health (2018)",
  ["artificial intelligence in medicine"] = "Artif Intell Med",
  ["artificial intelligence in medicine : 15th conference on artificial intelligence in medicine, aime 2015, pavia, italy, june 17-20, 2015 : proceedings. conference on artificial intelligence in medicine (2005-) (15th : 2015 : pavia, italy)"] = "Artif Intell Med (2015)",
  ["artificial intelligence in medicine : 16th conference on artificial intelligence in medicine, aime 2017, vienna, austria, june 21-24, 2017 : proceedings. conference on artificial intelligence in medicine (2005-) (16th : 2017 : vienna, austria)"] = "Artif Intell Med (2017)",
  ["artificial intelligence in medicine. conference on artificial intelligence in medicine (2005-)"] = "Artif Intell Med Conf Artif Intell Med (2005-)",
  ["artificial intelligence in the life sciences"] = "Artif. Intell. Life Sci.",
  ["artificial intelligence research"] = "Artif Intell Res",
  ["artificial intelligence review"] = "Artif Intell Rev",
  ["artificial intelligence surgery"] = "Artif. Intell. Surg.",
  ["artificial intelligence: foundations, theory, and algorithms"] = "Artif. Intell. Found. Theory Algorithms",
  ["artificial life"] = "Artif Life",
  ["artificial life and robotics"] = "Artif Life Robot",
  ["artificial limbs"] = "Artif Limbs",
  ["artificial organs"] = "Artif Organs",
  ["arts & health"] = "Arts Health",
  ["arts in virginia"] = "ArtVirg",
  ["arukoru kenkyu-to yakubutsu ison (japanese journal of alcohol studies and drug dependence)"] = "Arukoru Kenkyuto Yakubutsu Ison",
  ["arukōru kenkyū to yakubutsu izon = japanese journal of alcohol studies & drug dependence"] = "Arukoru Kenkyuto Yakubutsu Ison",
  ["arxius de miscel·lània zoològica"] = "Arx Misc Zool",
  ["arxiv.org, e-print archive astrophysics"] = "arXiv.org, e-Print Arch. Astrophys.",
  ["arxiv.org, e-print archive condensed matter"] = "arXiv.org, e-Print Arch. Condens. Matter",
  ["arxiv.org, e-print archive general relativity and quantum cosmology"] = "arXiv.org, e-Print Arch. Gen. Relativ. Quantum Cosmol.",
  ["arxiv.org, e-print archive high energy physics - experimental"] = "arXiv.org, e-Print Arch. High Energy Phys. Exp.",
  ["arxiv.org, e-print archive high energy physics - lattice"] = "arXiv.org, e-Print Arch. High Energy Phys. Lattice",
  ["arxiv.org, e-print archive high energy physics - phenomenology"] = "arXiv.org, e-Print Arch. High Energy Phys. Phenomenol.",
  ["arxiv.org, e-print archive high energy physics - theory"] = "arXiv.org, e-Print Arch. High Energy Phys.Theor.",
  ["arxiv.org, e-print archive nuclear experiment"] = "arXiv.org, e-Print Arch. Nucl. Exp.",
  ["arxiv.org, e-print archive nuclear theory"] = "arXiv.org, e-Print Arch. Nucl. Theory",
  ["arxiv.org, e-print archive physics"] = "arXiv.org, e-Print Arch. Phys.",
  ["arxiv.org, e-print archive quantitative biology"] = "arXiv.org, e-Print Arch. Quant. Biol.",
  ["arxiv.org, e-print archive quantum physics"] = "arXiv.org, e-Print Arch. Quantum Phys.",
  ["arya atherosclerosis"] = "ARYA Atheroscler",
  ["arzneimittel forschung"] = "Arzneim. Forsch.",
  ["arzneimittel-forschung/drug research"] = "Arzneim.-Forsch.",
  ["arzt und christ"] = "Arzt Christ",
  ["arztliche forschung"] = "Arztl. Forsch.",
  ["arztliche jugendkunde"] = "Arztl. Jugendkd.",
  ["arztliche sammelblatter"] = "Arztl Sammelbl",
  ["arztliche wochenschrift"] = "Arztl Wochensch",
  ["asa newsletter"] = "ASA Newsl",
  ["asa-crc series on statistical reasoning in science and society"] = "ASA-CRC Ser. Stat. Reason. Sci. Soc.",
  ["asa-siam series on statistics and applied probability"] = "ASA-SIAM Ser. Stat. Appl. Probab.",
  ["asaio journal"] = "ASAIO J.",
  ["asaio journal (american society for artificial internal organs : 1992)"] = "ASAIO J",
  ["asaio transactions"] = "ASAIO Trans",
  ["asaio transactions / american society for artificial internal organs"] = "ASAIO Trans",
  ["asce open"] = "ASCE OPEN",
  ["asce-asme journal of risk and uncertainty in engineering systems, part a: civil engineering"] = "ASCE-ASME J. Risk Uncertainty Eng. Syst. Part A: Civ. Eng.",
  ["asce-asme journal of risk and uncertainty in engineering systems, part b: mechanical engineering"] = "ASCE-ASME J. Risk Uncertainty Eng. Syst. Part B: Mech. Eng.",
  ["asce-asme journal of risk and uncertainty in engineering systems. part a, civil engineering"] = "ASCE ASME J Risk Uncertain Eng Syst A Civ Eng",
  ["asclepio; archivo iberoamericano de historia de la medicina y antropologia medica"] = "Asclepio",
  ["asclepio; archivo iberoamericano de historia de la medicina y antropología médica"] = "Asclepio",
  ["ascomycete.org : revue internationale pour la taxinomie des ascomycota"] = "Ascomyceteorg",
  ["asd technical report. united states. air force. systems command. aeronautical systems division"] = "ASD Tech Rep",
  ["asda news"] = "ASDA News",
  ["asda news (american student dental association : 1981)"] = "ASDA News",
  ["asdc journal of dentistry for children"] = "ASDC J Dent Child",
  ["asean economic bulletin"] = "ASEAN Econ Bull",
  ["asean heart journal : official journal of the asean federation of cardiology"] = "ASEAN Heart J",
  ["asean journal of open and distance learning"] = "ASEAN J Open Distance Learn",
  ["asgsb bulletin : publication of the american society for gravitational and space biology"] = "ASGSB Bull",
  ["asha leader"] = "ASHA Lead",
  ["asha monographs"] = "ASHA Monogr",
  ["asha. supplement"] = "ASHA Suppl",
  ["ashrae journal"] = "ASHRAE J",
  ["ashrae transactions"] = "ASHRAE Trans",
  ["ashrae winter conference papers. american society of heating, refrigeration and air-conditioning engineers"] = "ASHRAE Winter Conf Pap",
  ["ashuu koukogaku"] = "Ashuu Koukogaku",
  ["asia & the pacific policy studies"] = "Asia Pac Policy Stud",
  ["asia europe journal"] = "Asia Eur J",
  ["asia life sciences"] = "Asia Life Sci",
  ["asia major"] = "Asia Major",
  ["asia minor studien"] = "AMS",
  ["asia oceania journal of nuclear medicine & biology"] = "Asia Ocean J Nucl Med Biol",
  ["asia pacific allergy"] = "Asia Pac Allergy",
  ["asia pacific business review"] = "Asia Pacific Bus. Rev.",
  ["asia pacific family medicine"] = "Asia Pac Fam Med",
  ["asia pacific journal of clinical nutrition"] = "Asia Pac J Clin Nutr",
  ["asia pacific journal of counselling and psychotherapy"] = "Asia Pac J Couns Psychother",
  ["asia pacific journal of economics and business"] = "Asia Pacific J. Econ. Bus.",
  ["asia pacific journal of health law & ethics"] = "Asia Pac J Health Law Ethics",
  ["asia pacific journal of medical toxicology"] = "Asia Pac. J. Med. Toxicol.",
  ["asia pacific journal of molecular biology and biotechnology"] = "Asia Pac. J. Mol. Biol. Biotechnol.",
  ["asia pacific journal of social work"] = "Asia Pac J Soc Work",
  ["asia pacific journal of social work and development"] = "Asia Pac J Soc Work Dev",
  ["asia pacific journal of speech, language, and hearing"] = "Asia Pac J Speech Lang Hear",
  ["asia pacific mathematics newsletter"] = "Asia Pac. Math. Newsl.",
  ["asia pacific mathematics newsletters"] = "Asia Pac. Math. Newsl.",
  ["asia pacific observer : a quarterly newsletter from the east-west center"] = "Asia Pac Obs",
  ["asia pacific physics newsletter"] = "Asia Pac. Phys. Newsl.",
  ["asia pacific viewpoint"] = "Asia Pac Viewp",
  ["asia quarterly"] = "Asia Q",
  ["asia t'aep'yŏngyang sangdam yŏn'gu"] = "Asia Taepyongyang Sangdam Yongu",
  ["asia-oceania journal of obstetrics and gynaecology"] = "Asia Oceania J Obstet Gynaecol",
  ["asia-oceania journal of obstetrics and gynaecology / aofog"] = "Asia Oceania J Obstet Gynaecol",
  ["asia-pacific biotech news"] = "Asia Pac Biotech News",
  ["asia-pacific development journal"] = "Asia-Pacific Devel. J.",
  ["asia-pacific economic review"] = "Asia-Pacific Econ. Rev.",
  ["asia-pacific financial markets"] = "Asia-Pacific Finan. Markets",
  ["asia-pacific journal of atmospheric sciences"] = "Asia-Pac. J. Atmos. Sci.",
  ["asia-pacific journal of blood types and genes"] = "Asia Pac J Blood Types Genes",
  ["asia-pacific journal of cardiology"] = "Asia Pac J Cardiol",
  ["asia-pacific journal of chemical engineering"] = "Asia-Pac. J. Chem. Eng.",
  ["asia-pacific journal of clinical oncology"] = "Asia Pac J Clin Oncol",
  ["asia-pacific journal of molecular biology and biotechnology"] = "Asia Pac J Mol Biol Biotechnol",
  ["asia-pacific journal of oncology nursing"] = "Asia Pac J Oncol Nurs",
  ["asia-pacific journal of operational research"] = "Asia-Pac. J. Oper. Res.",
  ["asia-pacific journal of ophthalmology (philadelphia, pa.)"] = "Asia Pac J Ophthalmol (Phila)",
  ["asia-pacific journal of public health"] = "Asia Pac J Public Health",
  ["asia-pacific journal of public health / asia-pacific academic consortium for public health"] = "Asia Pac J Public Health",
  ["asia-pacific journal of rural development"] = "Asia Pac J Rural Dev",
  ["asia-pacific journal of sports medicine, arthroscopy, rehabilitation and technology"] = "Asia Pac J Sports Med Arthrosc Rehabil Technol",
  ["asia-pacific popin bulletin"] = "Asia Pac POPIN Bull",
  ["asia-pacific popin newsletter"] = "Asia-Pac POPIN Newsl",
  ["asia-pacific population & policy"] = "Asia Pac Pop Policy",
  ["asia-pacific population journal"] = "Asia Pac Popul J",
  ["asia-pacific population journal / united nations"] = "Asia Pac Popul J",
  ["asia-pacific population research abstracts"] = "Asia Pac Popul Res Abstr",
  ["asia-pacific population research abstracts / east-west center, program on population"] = "Asia Pac Popul Res Abstr",
  ["asia-pacific psychiatry : official journal of the pacific rim college of psychiatrists"] = "Asia Pac Psychiatry",
  ["asian & pacific population programme news"] = "Asian Pac Popul Programme News",
  ["asian affairs"] = "Asian Aff (Lond)",
  ["asian agri-history"] = "Asian Agrihist",
  ["asian american and pacific islander journal of health"] = "Asian Am Pac Isl J Health",
  ["asian american journal of psychology"] = "Asian Am J Psychol",
  ["asian and african studies"] = "Asian Afr Stud",
  ["asian and pacific census forum"] = "Asian Pac Cens Forum",
  ["asian and pacific census forum / east-west population institute"] = "Asian Pac Cens Forum",
  ["asian and pacific census newsletter"] = "Asian Pac Census Newsl",
  ["asian and pacific migration journal : apmj"] = "Asian Pac Migr J",
  ["asian and pacific population forum"] = "Asian Pac Popul Forum",
  ["asian and pacific population forum / east-west population institute, east-west center"] = "Asian Pac Popul Forum",
  ["asian anthropology"] = "Asian Anthropol",
  ["asian bioethics review"] = "Asian Bioeth Rev",
  ["asian biomedicine"] = "Asian Biomed.",
  ["asian biomedicine : research, reviews and news"] = "Asian Biomed (Res Rev News)",
  ["asian cardiovascular & thoracic annals"] = "Asian Cardiovasc Thorac Ann",
  ["asian control conference. asian control conference"] = "Asian Control Conf",
  ["asian development review"] = "Asian Dev Rev",
  ["asian economic journal"] = "Asian Econ. J.",
  ["asian economic review"] = "Asian Econ. Rev.",
  ["asian economies"] = "Asian Econ",
  ["asian european journal of mathematics"] = "Asian Eur. J. Math.",
  ["asian fisheries science"] = "Asian Fish Sci",
  ["asian folklore studies"] = "Asian Folkl Stud",
  ["asian forum newsletter : official organ of the asian forum of parliamentarians on population and development"] = "Asian Forum Newsl",
  ["asian geographer"] = "Asian Geogr",
  ["asian herpetological research"] = "Asian Herpetol Res",
  ["asian journal for mathematics education"] = "Asian J. Math. Educ.",
  ["asian journal of aesthetic dentistry"] = "Asian J Aesthet Dent",
  ["asian journal of aesthetic dentistry / the asian academy of aesthetic dentistry"] = "Asian J Aesthet Dent",
  ["asian journal of agricultural economics"] = "Asian J. Agr. Econ.",
  ["asian journal of agriculture and food science"] = "Asian J Agric Food Sci",
  ["asian journal of andrology"] = "Asian J Androl",
  ["asian journal of anesthesiology"] = "Asian J Anesthesiol",
  ["asian journal of animal and veterinary advances"] = "Asian J Anim Vet Adv",
  ["asian journal of animal sciences"] = "Asian J Anim Sci",
  ["asian journal of atmospheric environment"] = "Asian J. Atmos. Environ.",
  ["asian journal of biochemistry"] = "Asian J. Biochem.",
  ["asian journal of biological and life sciences"] = "Asian J Biol Life Sci",
  ["asian journal of biotechnology"] = "Asian J. Biotechnol.",
  ["asian journal of biotechnology and bioresource technology"] = "Asian J Biotechnol Bioresour Technol",
  ["asian journal of chemistry"] = "Asian J. Chem.",
  ["asian journal of civil engineering"] = "Asian J. Civ. Eng.",
  ["asian journal of communication"] = "Asian J Commun",
  ["asian journal of conservation biology"] = "Asian J Conserv Biol",
  ["asian journal of control"] = "Asian J. Control",
  ["asian journal of criminology"] = "Asian J Criminol",
  ["asian journal of economics and social studies"] = "Asian J Econ Soc Stud",
  ["asian journal of endoscopic surgery"] = "Asian J Endosc Surg",
  ["asian journal of experimental biological sciences"] = "Asian J. Exp. Biol. Sci.",
  ["asian journal of experimental sciences"] = "Asian J. Exp. Sci.",
  ["asian journal of gambling issues and public health"] = "Asian J Gambl Issues Public Health",
  ["asian journal of infectious diseases"] = "Asian J Infect Dis",
  ["asian journal of materials science"] = "Asian J. Mater. Sci.",
  ["asian journal of mathematics"] = "Asian J. Math.",
  ["asian journal of medical sciences"] = "Asian J Med Sci",
  ["asian journal of medicine"] = "Asian Med J",
  ["asian journal of microbiology, biotechnology & environmental sciences"] = "Asian J. Microbiol. Biotechnol. Environ. Exp. Sci.",
  ["asian journal of microbiology, biotechnology & environmental sciences : ajmbes"] = "Asian J Microbiol Biotechnol Environ Sci",
  ["asian journal of mycology"] = "Asian J Mycol",
  ["asian journal of neurosurgery"] = "Asian J Neurosurg",
  ["asian journal of occupational therapy"] = "Asian J Occup Ther",
  ["asian journal of organic chemistry"] = "Asian J. Org. Chem.",
  ["asian journal of pharmaceutical and clinical research"] = "Asian J. Pharm. Clin. Res.",
  ["asian journal of pharmaceutical and clinical research : ajpcr"] = "Asian J Pharm Clin Res",
  ["asian journal of pharmaceutical sciences"] = "Asian J Pharm Sci",
  ["asian journal of pharmaceutics"] = "Asian J Pharm",
  ["asian journal of pharmacy and pharmacology"] = "Asian J Pharm Pharmacol",
  ["asian journal of physics"] = "Asian J. Phys.",
  ["asian journal of physics : an international quarterly research journal"] = "Asian J Phys",
  ["asian journal of plant pathology"] = "Asian J Plant Pathol",
  ["asian journal of plant science"] = "Asian J Plant Sci",
  ["asian journal of plant sciences"] = "Asian J Plant Sci",
  ["asian journal of psychiatry"] = "Asian J Psychiatr",
  ["asian journal of research in animal and veterinary sciences"] = "Asian J Res Anim Vet Sci",
  ["asian journal of research in crop science"] = "Asian J Res Crop Sci",
  ["asian journal of research in social sciences and humanities"] = "Asian J Res Soc Sci Humanit",
  ["asian journal of scientific research"] = "Asian J Sci Res",
  ["asian journal of social psychology"] = "Asian J Soc Psychol",
  ["asian journal of social science"] = "Asian J Soc Sci",
  ["asian journal of spectroscopy"] = "Asian J. Spectro.",
  ["asian journal of sport and exercise psychology"] = "Asian J. Sport Exercise Psychol.",
  ["asian journal of sports medicine"] = "Asian J Sports Med",
  ["asian journal of surgery"] = "Asian J Surg",
  ["asian journal of surgery / asian surgical association"] = "Asian J Surg",
  ["asian journal of technology innovation"] = "Asian J. Technol. Innovation",
  ["asian journal of traditional medicines"] = "Asian J Tradit Med",
  ["asian journal of transfusion science"] = "Asian J Transfus Sci",
  ["asian journal of urology"] = "Asian J Urol",
  ["asian journal of water, environment and pollution"] = "Asian J. Water Environ. Pollut.",
  ["asian journal of women's studies"] = "Asian J Women Stud",
  ["asian medical journal"] = "Asian Med J",
  ["asian medicine (leiden, netherlands)"] = "Asian Med (Leiden)",
  ["asian migrant"] = "Asian Migr",
  ["asian nursing research"] = "Asian Nurs Res (Korean Soc Nurs Sci)",
  ["asian pacific journal of allergy and immunology"] = "Asian Pac J Allergy Immunol",
  ["asian pacific journal of allergy and immunology / launched by the allergy and immunology society of thailand"] = "Asian Pac J Allergy Immunol",
  ["asian pacific journal of cancer prevention"] = "Asian Pac. J. Cancer Prev.",
  ["asian pacific journal of cancer prevention : apjcp"] = "Asian Pac J Cancer Prev",
  ["asian pacific journal of tropical biomedicine"] = "Asian Pac J Trop Biomed",
  ["asian pacific journal of tropical disease"] = "Asian Pac J Trop Dis",
  ["asian pacific journal of tropical medicine"] = "Asian Pac J Trop Med",
  ["asian perspective"] = "Asian Perspect",
  ["asian population programme news"] = "Asian Popul Programme News",
  ["asian population studies"] = "Asian Popul Stud",
  ["asian population studies series"] = "Asian Popul Stud Ser",
  ["asian population studies series / economic commission for asia and the far east"] = "Asian Popul Stud Ser",
  ["asian primates journal : a journal of the southeast asia, south asia and china of the iucn/ssc primate specialist group"] = "Asian Primates J",
  ["asian profile"] = "Asian Profile",
  ["asian social science"] = "Asian Soc Sci",
  ["asian spine journal"] = "Asian Spine J",
  ["asian studies (quezon city, philippines)"] = "Asian Stud",
  ["asian studies review"] = "Asian Stud Rev",
  ["asian survey"] = "Asian Surv",
  ["asian textile journal"] = "Asian Text. J.",
  ["asian thought & society"] = "Asian Thought Soc",
  ["asian transport studies"] = "Asian Transport Stud.",
  ["asian-australasian journal of animal sciences"] = "Asian-Australas J Anim Sci",
  ["asian-european journal of mathematics"] = "Asian-Eur. J. Math.",
  ["asian-pacific economic literature"] = "Asia Pac Econ Lit",
  ["asian-pacific population programme news"] = "Asian Pac Popul Programme News",
  ["asian/pacific island nursing journal"] = "Asian Pac Isl Nurs J",
  ["asiatic herpetological research"] = "Asiat Herpetol Res",
  ["asiatic journal of biotechnology resources"] = "Asiat J Biotechnol Resour",
  ["asiatische forschungen"] = "Asiat Forsch",
  ["asien, afrika, lateinamerika (berlin, germany : 1973)"] = "Asien Afr Lateinam",
  ["askeri sihhiye dergisi"] = "Ask Sihhiye Derg",
  ["askerî sihhiye dergisi"] = "Ask Sihhiye Derg",
  ["askerı̂ veteriner mecmuasi"] = "Ask Vet Mecm",
  ["asklepii : bolgaro-sovetskii ezhegodnik istorii i teorii meditsiny"] = "Asklepii",
  ["asklepios; ärztliches mitteilungsblatt"] = "Asklep",
  ["aslib proceedings"] = "Aslib Proc",
  ["asm animal microbiology"] = "ASM Anim. Microbiol.",
  ["asm case reports"] = "ASM Case Rep.",
  ["asm food microbiology"] = "ASM Food Microbiol.",
  ["asm news"] = "ASM News",
  ["asme applied mechanics reviews"] = "ASME Appl. Mech. Rev.",
  ["asme journal of applied mechanics"] = "J. Appl. Mech. Trans. ASME",
  ["asme journal of autonomous vehicles and systems"] = "ASME J. Auton. Veh. Syst.",
  ["asme journal of biomechanical engineering"] = "ASME J. Biomech. Eng.",
  ["asme journal of computational and nonlinear dynamics"] = "ASME J. Comput. Nonlinear Dyn.",
  ["asme journal of computing and information science in engineering"] = "ASME J. Comput. Inf. Sci. Eng.",
  ["asme journal of convergence in biomedical engineering & biotechnology"] = "ASME J. Convergence Biomed. Eng. Biotechnol.",
  ["asme journal of dynamic systems measurement & control"] = "ASME J. Dyn. Syst. Meas. Control",
  ["asme journal of electrochemical energy conversion and storage"] = "ASME J. Electrochem. Energy Convers. Storage",
  ["asme journal of electronic packaging"] = "ASME J. Electron. Packag.",
  ["asme journal of energy resources technology"] = "ASME J. Energy Res. Technol.",
  ["asme journal of engineering and science in medical diagnostics and therapy"] = "ASME J. Eng. Sci. Med. Diagn. Ther.",
  ["asme journal of engineering for gas turbines and power"] = "ASME J. Eng. Gas Turbines Power",
  ["asme journal of engineering for sustainable buildings and cities"] = "ASME J. Eng. Sustainable Build. Cities",
  ["asme journal of engineering materials and technology"] = "ASME J. Eng. Mater. Technol.",
  ["asme journal of fluids engineering"] = "ASME J. Fluids Eng.",
  ["asme journal of fuel cell science and technology"] = "ASME J. Fuel Cell Sci. Technol.",
  ["asme journal of heat transfer"] = "ASME J. Heat Transfer",
  ["asme journal of manufacturing science and engineering"] = "ASME J. Manuf. Sci. Eng.",
  ["asme journal of mechanical design"] = "ASME J. Mech. Des.",
  ["asme journal of mechanisms and robotics"] = "ASME J. Mech. Rob.",
  ["asme journal of medical devices"] = "ASME J. Med. Devices",
  ["asme journal of medical diagnostics"] = "ASME J. Med. Diagn.",
  ["asme journal of micro and nano-manufacturing"] = "ASME J. Micro Nano-Manuf.",
  ["asme journal of nanotechnology in engineering and medicine"] = "ASME J. Nanotechnol. Eng. Med.",
  ["asme journal of nondestructive evaluation, diagnostics & prognostics of engineering systems"] = "ASME J. Nondestr. Eval. Diagn. Progn. Eng. Syst.",
  ["asme journal of nuclear engineering and radiation science"] = "ASME J. Nucl. Eng. Radiat. Sci.",
  ["asme journal of offshore mechanics and arctic engineering"] = "ASME J. Offshore Mech. Arct. Eng.",
  ["asme journal of pressure vessel technology"] = "ASME J. Pressure Vessel Technol.",
  ["asme journal of solar energy engineering"] = "ASME J. Sol. Energy Eng.",
  ["asme journal of thermal science and engineering applications"] = "ASME J. Therm. Sci. Eng. Appl.",
  ["asme journal of tribology"] = "ASME J. Tribol.",
  ["asme journal of turbomachinery"] = "ASME J. Turbomach.",
  ["asme journal of verification, validation, and uncertainty quantification"] = "ASME J. Verif. Validation Uncertainty Quantif.",
  ["asme journal of vibration and acoustics"] = "ASME J. Vib. Acoust.",
  ["asme letters in dynamic systems and control"] = "ASME Lett. Dyn. Syst. Control",
  ["asme mechanical engineering"] = "ASME Mech. Eng.",
  ["asn neuro"] = "ASN Neuro",
  ["asna reporter"] = "ASNA Rep",
  ["aspects of applied biology"] = "Asp Appl Biol",
  ["aspects of homogeneous catalysis"] = "Aspects Homogeneous Catal.",
  ["aspects of mathematics"] = "Aspects Math.",
  ["aspects statistiques"] = "Asp Stat Serv Cant Stat Geneve SCS",
  ["aspekte komplexer systeme"] = "Asp. Komplexer Systeme",
  ["aspen emphysema conference"] = "Aspen Emphysema Conf",
  ["aspen's advisor for nurse executives"] = "Aspens Advis Nurse Exec",
  ["aspen’s advisor for nurse executives"] = "Aspens Advis. Nurse Exec.",
  ["assaph. studies in art history"] = "Assaph",
  ["assay and drug development technologies"] = "Assay Drug Dev Technol",
  ["assembly automation"] = "Assem. Autom.",
  ["assertive nurse"] = "Assertive Nurse",
  ["assessment & development matters"] = "Assess Dev Matters",
  ["assessment and evaluation in higher education"] = "Assess Eval High Educ",
  ["assessment for effective intervention : official journal of the council for educational diagnostic services"] = "Assess Eff Interv",
  ["assessment in education : principles, policy & practice"] = "Assess Educ",
  ["asset analytics"] = "Asset Anal.",
  ["assets. acm conference on assistive technologies"] = "ASSETS",
  ["assia--jewish medical ethics"] = "Assia Jew Med Ethics",
  ["assignment children"] = "Assignment Child",
  ["assisted reproduction reviews"] = "Assist Reprod Rev",
  ["assistenza infermieristica e ricerca : air"] = "Assist Inferm Ric",
  ["assistive technology"] = "Assistive Technol.",
  ["assistive technology : the official journal of resna"] = "Assist Technol",
  ["assistive technology outcomes and benefits"] = "Assist Technol Outcomes Benefits",
  ["assistive technology research series"] = "Assist technol Res Ser",
  ["assiut veterinary medical journal"] = "Assiut Vet Med J",
  ["association & society manager"] = "Assoc Soc Manager",
  ["association and society manager"] = "Assoc. Soc. Manager",
  ["association for computing machinery"] = "ACM Trans. Math. Software",
  ["association for computing machinery. transactions on mathematical software"] = "ACM Trans. Math. Software",
  ["association for mathematical research monographs"] = "Assoc. Math. Res. Monogr.",
  ["association for women in mathematics series"] = "Assoc. Women Math. Ser.",
  ["association management"] = "Assoc Manage",
  ["association medical journal"] = "Assoc Med J",
  ["association of american publishers electronic manuscript series"] = "Assoc. American Publ. Electron. Manuscript Ser.",
  ["association of canadian map libraries and archives bulletin"] = "Assoc. Can. Map Lib. Arch. Bull.",
  ["assyriological miscellanies"] = "AssyrMisc",
  ["asta advances in statistical analysis"] = "AStA Adv. Stat. Anal.",
  ["asta. advances in statistical analysis. a journal of the german statistical society"] = "AStA Adv. Stat. Anal.",
  ["asthetische medizin"] = "Asthet. Med. (Berl.)",
  ["asthma in general practice : journal of the gps in asthma group"] = "Asthma Gen Pract",
  ["asthma research and practice"] = "Asthma Res Pract",
  ["astin bulletin"] = "Astin Bull.",
  ["astin bulletin. the journal of the international actuarial association"] = "Astin Bull.",
  ["astm special technical publication"] = "ASTM Spec. Tech. Publ.",
  ["astm specical technical publication"] = "ASTM Spec. Tech. Publ.",
  ["astm standardization news"] = "ASTM Stand. News",
  ["astra proceedings"] = "ASTRA Proc.",
  ["astronomical & astrophysical transactions"] = "Astron. Astrophys. Trans.",
  ["astronomical journal"] = "Astron. J.",
  ["astronomical review"] = "Astron. Rev.",
  ["astronomicheskii zhurnal [soviet astronomy]"] = "Astron. Zh. [Sov. Astron.]",
  ["astronomische nachrichten"] = "Astron. Nachr.",
  ["astronomy & astrophysics, supplement series"] = "Astron. Astrophys., Suppl. Ser.",
  ["astronomy & astrophysics. supplement series"] = "Astron Astrophys Suppl Ser",
  ["astronomy & geophysics"] = "Astron. Geophys.",
  ["astronomy and astrophysics"] = "Astron. Astrophys.",
  ["astronomy and astrophysics library"] = "Astronom. Astrophys. Lib.",
  ["astronomy and astrophysics review"] = "Astron. Astrophys. Rev.",
  ["astronomy and computing"] = "Astron. Comput.",
  ["astronomy education review"] = "Astron. Educ. Rev.",
  ["astronomy journal"] = "Astron. J.",
  ["astronomy letters"] = "Astron. Lett.",
  ["astronomy quarterly"] = "Astron. Q.",
  ["astronomy reports"] = "Astron. Rep.",
  ["astroparticle physics"] = "Astropart. Phys.",
  ["astrophysical bulletin"] = "Astrophys. Bull.",
  ["astrophysical journal"] = "Astrophys. J.",
  ["astrophysical journal letters"] = "Astrophys. J. Lett.",
  ["astrophysical journal supplement series"] = "Astrophys. J. Suppl. Ser.",
  ["astrophysical journal, letters to the editor"] = "Astrophys. J. Lett.",
  ["astrophysical journal, supplement series"] = "Astrophys. J. Suppl. Ser.",
  ["astrophysical letters"] = "Astrophys. Lett.",
  ["astrophysical letters & communications"] = "Astrophys Lett Commun",
  ["astrophysics and space science"] = "Astrophys. Space Sci.",
  ["astrophysics and space science library"] = "Astrophys. Space Sci. Libr.",
  ["astrophysics and space science transactions"] = "Astrophys. Space Sci. Trans",
  ["astrophysics and space science. an international journal of astronomy, astrophysics and space science"] = "Astrophys. Space Sci.",
  ["astrophysics journal"] = "Astrophys. J.",
  ["astrophysics journal supplement series"] = "Astrophys. J. Suppl. Ser.",
  ["aswan heart centre science & practice series"] = "Aswan Heart Cent Sci Pract Ser",
  ["asymptotic analysis"] = "Asymptot. Anal.",
  ["atalanta (munich, germany)"] = "Atalanta (Munch)",
  ["atcc connection : the newsletter of the american type culture collection"] = "ATCC Connect",
  ["atencion primaria"] = "Aten Primaria",
  ["atencion primaria / sociedad espanola de medicina de familia y comunitaria"] = "Aten Primaria",
  ["atene e roma"] = "Atene Roma",
  ["atene e roma: rassegna trimestrale dell’associazione italiana di cultura classica"] = "A&R",
  ["ateneo parmense. acta bio-medica"] = "Ateneo Parmense. Acta Biomed.",
  ["ateneo parmense. sezione i, acta bio-medica"] = "Ateneo Parmense. [1]",
  ["ateneo veneto : revista di scienze, lettere ed arti"] = "Ateneo Veneto",
  ["athena scientific optimization and computation series"] = "Athena Sci. Optim. Comput. Ser.",
  ["athena; rassegna mensile di biologia, clinica e terapia"] = "Athena",
  ["athenaeum economics"] = "Athenaeum Econom.",
  ["athenaeum. studi di letteratura e storia dell’antichità"] = "Athenaeum",
  ["athenaum monografien. philosophie"] = "Athenaum Monogr. Philos.",
  ["athens journal of sciences"] = "Athens J Sci",
  ["atherosclerosis (amsterdam, netherlands)"] = "Atherosclerosis (Amsterdam, Neth.)",
  ["atherosclerosis (shannon, ireland)"] = "Atherosclerosis (Shannon, Irel.)",
  ["atherosclerosis. supplements"] = "Atheroscler Suppl",
  ["athletic therapy today : the journal for sports health care professionals"] = "Athl Ther Today",
  ["athletic training & sports health care"] = "Athl Train Sports Health Care",
  ["atla abstracts"] = "ATLA Abstr",
  ["atla, alternatives to laboratory animals"] = "ATLA, Altern. Lab. Anim.",
  ["atlal. the journal of saudi arabian archaeology"] = "Atlal",
  ["atlanta history"] = "Atlanta Hist",
  ["atlanta medicine : bulletin of the medical association of atlanta"] = "Atlanta Med",
  ["atlanta weekly"] = "Atlanta Wkly",
  ["atlantic (boston, mass. : 1932)"] = "Atlantic",
  ["atlantic (boston, mass. : 1981)"] = "Atlantic",
  ["atlantic economic journal"] = "Atlantic Econ. J.",
  ["atlantic economic journal : aej"] = "Atl Econ J",
  ["atlantic geology"] = "Atl. Geol.",
  ["atlantic journal of communication"] = "Atl J Commun",
  ["atlantic monthly (boston, mass. : 1971)"] = "Atl Mon",
  ["atlantic monthly (boston, mass. : 1993)"] = "Atl Mon",
  ["atlantic reporter"] = "Atl Report",
  ["atlantic studies (abingdon, england)"] = "Atl Stud (Abingdon)",
  ["atlantida (lisbon, portugal)"] = "Atlantida",
  ["atlantis (montreal, quebec)"] = "Atlantis (Montr)",
  ["atlantis (montréal, quebec)"] = "Atlantis (Montr)",
  ["atlantis (wolfville, n.s.)"] = "Atlantis (Wolfv)",
  ["atlantis briefs in differential equations"] = "Atlantis Briefs Differ. Equ.",
  ["atlantis computational intelligence systems"] = "Atlantis Comput. Intell. Syst.",
  ["atlantis studies in astroparticle physics and cosmology"] = "Atlantis Stud. Astropart. Phys. Cosmol.",
  ["atlantis studies in computational finance and financial engineering"] = "Atlantis Stud. Comput. Finance Financ. Eng.",
  ["atlantis studies in computing"] = "Atlantis Stud. Comput.",
  ["atlantis studies in differential equations"] = "Atlantis Stud. Differ. Equ.",
  ["atlantis studies in dynamical systems"] = "Atlantis Stud. Dyn. Syst.",
  ["atlantis studies in mathematical physics: theory and applications"] = "Atlantis Stud. Math. Phys. Theory Appl.",
  ["atlantis studies in mathematics"] = "Atlantis Stud. Math.",
  ["atlantis studies in mathematics for engineering and science"] = "Atlantis Stud. Math. Eng. Sci.",
  ["atlantis studies in probability and statistics"] = "Atlantis Stud. Probab. Stat.",
  ["atlantis studies in scientific computing in electromagnetics"] = "Atlantis Stud. Sci. Comput. Electromagn.",
  ["atlantis studies in variational geometry"] = "Atlantis Stud. Var. Geom.",
  ["atlantis transactions in geometry"] = "Atlantis Trans. Geom.",
  ["atlas de radiologie clinique de la presse medicale"] = "Atlas Radiol. Clin. Presse Med.",
  ["atlas de radiologie clinique de la presse médicale"] = "Atlas Radiol Clin Presse Med",
  ["atlas journal of biology"] = "Atlas J Biol",
  ["atlas journal of plant biology"] = "Atlas J Plant Biol",
  ["atlas of genetics and cytogenetics in oncology and haematology"] = "Atlas Genet Cytogenet Oncol Haematol",
  ["atlas of the oral and maxillofacial surgery clinics of north america"] = "Atlas Oral Maxillofac Surg Clin North Am",
  ["atlas world press review"] = "Atlas World Press Rev",
  ["atmosphere: atmosphere—ocean"] = "Atmos.—Ocean",
  ["atmospheric and oceanic optics"] = "Atmos. Oceanic Opt.",
  ["atmospheric and oceanic science letters"] = "Atmos. Oceanic Sci. Lett.",
  ["atmospheric and oceanographic sciences library"] = "Atmos. Oceanogr. Sci. Libr.",
  ["atmospheric chemistry and physics"] = "Atmos. Chem. Phys.",
  ["atmospheric chemistry and physics discussions"] = "Atmos. Chem. Phys. Discuss.",
  ["atmospheric chemistry and physics discussions : acpd"] = "Atmos Chem Phys Discuss",
  ["atmospheric environment"] = "Atmos. Environ.",
  ["atmospheric environment\tpart a general topics"] = "Atmos. Environ. Part A",
  ["atmospheric environment\tpart b urban atmosphere"] = "Atmos. Environ. Part B",
  ["atmospheric environment (oxford, england : 1994)"] = "Atmos Environ (1994)",
  ["atmospheric environment part a general topics"] = "Atmos. Environ. Part A",
  ["atmospheric environment part b urban atmosphere"] = "Atmos. Environ. Part B",
  ["atmospheric environment, part a: general topics"] = "Atmos. Environ., Part A",
  ["atmospheric environment: x"] = "Atmos Environ X",
  ["atmospheric environmentpart a general topics"] = "Atmos. Environ. Part A",
  ["atmospheric environmentpart b urban atmosphere"] = "Atmos. Environ. Part B",
  ["atmospheric measurement techniques"] = "Atmos. Meas. Tech.",
  ["atmospheric measurement techniques discussions"] = "Atmos. Meas. Tech. Discuss.",
  ["atmospheric pollution research"] = "Atmos Pollut Res",
  ["atmospheric research"] = "Atmos. Res.",
  ["atmospheric research letters"] = "Atmos. Res. Lett.",
  ["atmospheric science letters"] = "Atmos. Sci. Lett.",
  ["atmospheric science letters : asl"] = "Atmos Sci Lett",
  ["atoll research bulletin"] = "Atoll Res Bull",
  ["atomic data and nuclear data tables"] = "At. Data Nucl. Data Tables",
  ["atomic energy"] = "At. Energy",
  ["atomic energy (new york, ny, united states)"] = "At. Energy (N. Y., NY, U. S.)",
  ["atomic energy in biophysics, biology and medicine"] = "At Energy Biophys Biol Med",
  ["atomic energy law journal"] = "At Energy Law J",
  ["atomic energy review"] = "At Energy Rev",
  ["atomic spectroscopy"] = "At. Spectrosc.",
  ["atomization and sprays"] = "Atomization Sprays",
  ["atomnaya energiya [soviet journal of atomic energy]"] = "At. Energ. [Sov. J. At. Energy]",
  ["ats scholar"] = "ATS Sch",
  ["attachment & human development"] = "Attach Hum Dev",
  ["attachment (london, england)"] = "Attachment (Lond)",
  ["attachment and human development"] = "Attach. Hum. Dev.",
  ["attention deficit and hyperactivity disorders"] = "Atten Defic Hyperact Disord",
  ["attention, perception & psychophysics"] = "Atten Percept Psychophys",
  ["atti de la accademia delle scienze dell’istituto di bologna. memorie"] = "MemBologna",
  ["atti dei civici musei di storia ed arte di trieste"] = "AttiMusTrieste",
  ["atti dei convegni dell’a"] = "Atti Convegni A.I.C.M.",
  ["atti del centro ricerche e documentazione sull’antichità classica"] = "CRDAC",
  ["atti del congresso nazionale. societa italiana di medicina legale e delle assicurazioni"] = "Atti Congr Naz Soc Ital Med Leg",
  ["atti del congresso nazionale. società italiana di medicina legale e delle assicurazioni"] = "Atti Congr Naz Soc Ital Med Leg",
  ["atti del seminario matematico e fisico dell’università di modena"] = "Atti Sem. Mat. Fis. Univ. Modena",
  ["atti del sodalizio glottologico milanese"] = "ASGM",
  ["atti della accademia dei fisiocritici in siena. sezione medico-fisica"] = "Atti Accad Fisiocrit Siena Med Fis",
  ["atti della accademia delle scienze dell'istituto di bologna, classe di scienze fisiche. memorie"] = "Atti Accad Sci Ist Bologna Classe Sci Fis Rend",
  ["atti della accademia delle scienze dell’istituto di bologna"] = "Atti Accad. Sci. Istit. Bologna Cl. Sci. Fis. Rend. (14)",
  ["atti della accademia delle scienze dell’istituto di bologna, classe di scienze fisiche. memorie"] = "Atti Accad Sci Ist Bologna Classe Sci Fis Rend",
  ["atti della accademia delle scienze dell’istituto di bologna. rendiconti"] = "RendBologna",
  ["atti della accademia delle scienze di torino"] = "Atti Accad. Sci. Torino Cl. Sci. Fis. Mat. Natur.",
  ["atti della accademia delle scienze di torino, 2. classe di scienze morali, storiche e filologiche"] = "AttiAcTorino",
  ["atti della accademia delle scienze di torino, classe di scienze morali, storiche e filologiche"] = "AAT",
  ["atti della accademia di scienze morali e politiche della società nazionale di scienze, lettere ed arti di napoli"] = "AAN",
  ["atti della accademia di scienze, lettere e arti di palermo"] = "AttiPalermo",
  ["atti della accademia medica lombarda"] = "Atti Accad Med Lomb",
  ["atti della accademia mediterranea delle scienze"] = "AAMed",
  ["atti della accademia nazionale dei lincei"] = "Atti Accad Naz Lincei",
  ["atti della accademia nazionale dei lincei, classe di scienze fisiche, matematiche e naturali, rendiconti"] = "Atti Accad. Naz. Lincei Cl. Sci. Fis. Mat. Nat. Rend.",
  ["atti della accademia nazionale dei lincei, classe di scienze morali, storiche e filologiche. rendiconti"] = "RAL",
  ["atti della accademia nazionale dei lincei. rendiconti lincei. matematica e applicazioni"] = "Atti Accad. Naz. Lincei Rend. Lincei Mat. Appl.",
  ["atti della accademia peloritana dei pericolanti, classe di lettere, filosofia e belle arti"] = "AAPel",
  ["atti della accademia peloritana dei pericolanti. classe di scienze, fisiche, matematiche e naturali. aapp. physical, mathematical, and natural sciences"] = "Atti Accad. Peloritana Pericolanti Cl. Sci. Fis. Mat. Natur.",
  ["atti della accademia pontaniana"] = "Atti Accad. Pontaniana (N.S.)",
  ["atti della accademia roveretana degli agiati, classe di scienze umane, lettere ed arti"] = "AARov",
  ["atti della accademia roveretana degli agiati. contributi della classe di scienze umane, di lettere ed arti"] = "AttiAcRov",
  ["atti della pontificia accademia romana di archeologia, ser. iiia, memorie (in-4¡)"] = "MPAA",
  ["atti della pontificia accademia romana di archeologia. memorie"] = "MemPontAc",
  ["atti della r. accademia dei fisiocritici in siena"] = "Atti Accad Fisiocrit Siena",
  ["atti della societa italiana di cardiologia"] = "Atti Soc. Ital. Cardiol.",
  ["atti della societa italiana di dermatologia e sifilografia e delle sezioni interprovinciali. societa italiana di dermatologia e sifilografia"] = "Soc Ital Dermatol Sifilogr Sezioni Interprov Soc Ital Dermatol Sifilogr",
  ["atti della società italiana di cardiologia"] = "Atti Soc Ital Cardiol",
  ["atti della società italiana di dermatologia e sifilografia e delle sezioni interprovinciali. società italiana di dermatologia e sifilografia"] = "Soc Ital Dermatol Sifilogr Sezioni Interprov Soc Ital Dermatol Sifilogr",
  ["atti della società per la preistoria e protostoria della regione friuli – venezia giulia"] = "AttiSocFriuli",
  ["atti dell’accademia dei fisiocritici in siena"] = "Atti Accad. Fisiocrit. Siena",
  ["atti dell’accademia di scienze, lettere e arti di palermo"] = "AAPal",
  ["atti dell’accademia ligure di scienze e lettere"] = "AALIG",
  ["atti dell’accademia nazionale dei lincei, classe di scienze morali, storiche e filologiche. memorie"] = "MemLinc",
  ["atti dell’accademia pontaniana"] = "AAP",
  ["atti dell’istituto veneto di scienze, lettere ed arti, classe di scienze morali e lettere"] = "AIV",
  ["atti e memorie - accademia nazionale di scienze, lettere e arti, modena. accademia di scienze, lettere e arti (modena, italy)"] = "Atti Mem Accad Naz Sci Lett Arti Modena",
  ["atti e memorie - deputazione di storia patria per le antiche provincie modenesi. deputazione di storia patria per le antiche provincie modenesi"] = "Atti Mem Deput Stor Patria Antiche Prov Modenesi",
  ["atti e memorie dell'accademia toscana di scienze e lettere la colombaria"] = "Atti Mem Accad Toscana Sci Lett Colombaria",
  ["atti e memorie della accademia di storia dell'arte sanitaria"] = "Atti Mem Accad Stor Arte Sanit",
  ["atti e memorie della accademia di storia dell’arte sanitaria"] = "Atti Mem. Accad. Stor. Arte Sanit.",
  ["atti e memorie della accademia italiana di storia della farmacia"] = "Atti Mem Accad Ital Stor Farm",
  ["atti e memorie della accademia virgiliana di mantova"] = "AVM",
  ["atti e memorie della reale accademia petrarca di lettere, arti e scienze"] = "Atti Mem R Accad Petrarca Lett Arti Sci",
  ["atti e memorie della società dalmata di storia patria"] = "AttiMemDal",
  ["atti e memorie della società istriana di archeologia e storia patria"] = "AMSI",
  ["atti e memorie della società magna grecia"] = "ASMG",
  ["atti e memorie della società tiburtina di storia e d’arte"] = "AttiMemTivoli",
  ["atti e memorie delle accademie di agricoltura, scienze e lettere di verona"] = "AMAV",
  ["atti e memorie dell’academia toscana di scienze e lettere »la columbaria«"] = "AttiMemFirenze",
  ["atti e memorie dell’accademia patavina di scienze, lettere ed arti, classe di sc. mor., lett. ed arti"] = "AAPat",
  ["atti e memorie dell’accademia toscana di scienze e lettere la colombaria"] = "Atti Mem Accad Toscana Sci Lett Colombaria",
  ["atti e memorie dell’accademia toscana la colombaria"] = "AATC",
  ["atti e memorie dell’arcadia"] = "AMArc",
  ["atti e memorie. deputazione di storia patria per le antiche provincie modenesi"] = "AttiMemModena",
  ["atti e memorie. deputazione di storia patria per le province di romagna"] = "AttiMemBologna",
  ["atti e memorie. memorie scientifiche, giuridiche, letterarie. serie viii"] = "Atti Mem. Mem. Sci. Giuridiche Lett. (8)",
  ["atti. centro di ricerche storiche, rovigno"] = "AttiRovigno",
  ["atti. centro ricerche e documentazione sull’antichità classica"] = "AttiCAntCl",
  ["atti. centro studi e documentazione sull’italia romana"] = "AttiCItRom",
  ["atti. congresso nazionale di storia della medicina"] = "Atti Congr Naz Stor Med",
  ["atti. istituto veneto di scienze, lettere ed arti"] = "AttiVenezia",
  ["atti. universita di palermo. clinica oto-rino-laringoiatrica"] = "Atti Clin Otorinolaringoitr Univ Palermo",
  ["atti. università di palermo. clinica oto-rino-laringoiatrica"] = "Atti Clin Otorinolaringoitr Univ Palermo",
  ["attività didattico-scientifica nell'anno accademico. università di roma. istituto di storia della medicina"] = "Attiv Didat",
  ["attualita dentale"] = "Attual. Dent.",
  ["attualita di ostetricia e ginecologia"] = "Attual. Ostet. Ginecol.",
  ["attualita in ematologia"] = "Attual Ematol",
  ["attualità dentale"] = "Attual Dent",
  ["attualità di ostetricia e ginecologia"] = "Attual Ostet Ginecol",
  ["atw internationale zeitschrift fur kernenergie"] = "ATW Int. Z. Kernerg.",
  ["aua news"] = "AUA News",
  ["aua update series"] = "AUA Update Ser",
  ["auaa journal"] = "AUAA J.",
  ["auaa journal : official journal of the american urological association allied"] = "AUAA J",
  ["auckland university law review"] = "Auckl Univ Law Rev",
  ["auckland-waikato historical journal"] = "Auckl Waikate Hist J",
  ["audiol. neurotol."] = "Audiology and Neurotology",
  ["audiological medicine"] = "Audiol Med",
  ["audiology & neuro-otology"] = "Audiol Neurootol",
  ["audiology : official organ of the international society of audiology"] = "Audiology",
  ["audiology and neuro-otology"] = "Audiol. Neurootol.",
  ["audiology and neurotology. extra"] = "Audiol Neurotol Extra",
  ["audiology research"] = "Audiol Res",
  ["audiology today : bulletin of the american academy of audiology"] = "Audiol Today",
  ["audit unit news"] = "Audit Unit News",
  ["auditory neuroscience"] = "Audit Neurosci",
  ["auditory perception & cognition"] = "Audit Percept Cogn",
  ["aufbaukurs mathematik"] = "Aufbaukurs Math.",
  ["aufstieg und niedergang der römischen welt"] = "ANRW",
  ["aufstieg und niedergang der römischen welt: geschichte und kultur roms im spiegel der neueren forschung"] = "ANRW",
  ["augmentative and alternative communication (baltimore, md. : 1985)"] = "Augment Altern Commun",
  ["augmented environments for computer-assisted interventions : 9th international workshop, ae-cai 2014, held in conjunction with miccai 2014, boston, ma, usa, september 14, 2014 : proceedings. ae-cai (workshop) (9th : 2014 : boston, mass.)"] = "Augment Environ Comput Assist Interv (2014)",
  ["augmented human research"] = "Augment. Hum. Res.",
  ["augsburger mathematisch-naturwissenschaftliche schriften"] = "Augsbg. Math.-Naturwiss. Schrift.",
  ["augsburger ökologische schriften"] = "Augsbg. Ökol. Schr.",
  ["augustan age"] = "AugAge",
  ["augustinian studies"] = "AugStud",
  ["auk, the"] = "Auk",
  ["aula orientalis. revista de estudios del próximo oriente antiguo"] = "AulaOr",
  ["auris, nasus, larynx"] = "Auris Nasus Larynx",
  ["aurora dover modern math originals"] = "Aurora Dover Mod. Math Orig.",
  ["aus politik und zeitgeschichte"] = "Polit Zeitgesch",
  ["ausgrabungen der deutschen forschungsgemeinschaft in uruk-warka"] = "ADFU",
  ["ausgrabungen in alt-paphos auf cypern"] = "Alt-Paphos",
  ["ausgrabungen in uruk-warka. endberichte"] = "AUWE",
  ["ausgrabungen und funde"] = "Ausgrab Funde",
  ["ausgrabungen und funde in westfalen-lippe"] = "AusgrFuWestf",
  ["ausgrabungen und funde. nachrichtenblatt der landesarchäologie"] = "AusgrFu",
  ["aussenwirtschaft : zeitschrift fur internationale wirtschaftsbeziehungen / herausgegeben vom schweizerischen institut fur aussenwirtschafts- und marktforschung an der handels-hochschule st. gallen"] = "Aussenwirtschaft",
  ["aussenwirtschaft : zeitschrift für internationale wirtschaftsbeziehungen"] = "Aussenwirtschaft",
  ["aust-agder arv"] = "Aust Agder Arv",
  ["austin addiction sciences"] = "Austin Addict Sci",
  ["austin alzheimer's and parkinson's disease"] = "Austin Alzheimers Parkinsons Dis",
  ["austin biochemistry"] = "Austin Biochem",
  ["austin biometrics and biostatistics"] = "Austin Biom Biostat",
  ["austin journal of analytical and pharmaceutical chemistry"] = "Austin J Anal Pharm Chem",
  ["austin journal of anesthesia and analgesia"] = "Austin J Anesth Analg",
  ["austin journal of biomedical engineering"] = "Austin J Biomed Eng",
  ["austin journal of cancer and clinical research"] = "Austin J Cancer Clin Res",
  ["austin journal of cerebrovascular disease & stroke"] = "Austin J Cerebrovasc Dis Stroke",
  ["austin journal of clinical cardiology"] = "Austin J Clin Cardiol",
  ["austin journal of clinical immunology"] = "Austin J Clin  Immunol",
  ["austin journal of clinical ophthalmology"] = "Austin J Clin Ophthalmol",
  ["austin journal of clinical pathology"] = "Austin J Clin Pathol",
  ["austin journal of drug abuse and addiction"] = "Austin J Drug Abuse Addict",
  ["austin journal of environmental toxicology"] = "Austin J Environ Toxicol",
  ["austin journal of hiv/aids research"] = "Austin J HIV AIDS Res",
  ["austin journal of medical oncology"] = "Austin J Med Oncol",
  ["austin journal of nanomedicine & nanotechnology"] = "Austin J Nanomed Nanotechnol",
  ["austin journal of nephrology and hypertension"] = "Austin J Nephrol Hypertens",
  ["austin journal of nutrition and food sciences"] = "Austin J Nutr Food Sci",
  ["austin journal of pediatrics"] = "Austin J Pediatr",
  ["austin journal of pharmacology and therapeutics"] = "Austin J Pharmacol Ther",
  ["austin journal of pulmonary and respiratory medicine"] = "Austin J Pulm Respir Med",
  ["austin journal of surgery"] = "Austin J Surg",
  ["austin journal of urology"] = "Austin J Urol",
  ["austin journal of vascular medicine"] = "Austin J Vasc Med",
  ["austin liver"] = "Austin Liver",
  ["austin neurology & neurosciences"] = "Austin Neurol Neurosci",
  ["austin pharmacology & pharmaceutics"] = "Austin Pharmacol Pharm",
  ["austin virology and retro virology"] = "Austin Virol Retro Virol",
  ["austral ecology"] = "Austral Ecol",
  ["austral entomology"] = "Aust Entomol",
  ["austral journal of veterinary sciences"] = "Austral J Vet Sci",
  ["australasian annals of medicine"] = "Australas Ann Med",
  ["australasian biotechnology"] = "Australas Biotechnol",
  ["australasian chiropractic & osteopathy : journal of the chiropractic & osteopathic college of australasia"] = "Australas Chiropr Osteopathy",
  ["australasian emergency care"] = "Australas Emerg Care",
  ["australasian emergency nursing journal : aenj"] = "Australas Emerg Nurs J",
  ["australasian epidemiologist"] = "Australas epidemiol",
  ["australasian historical archaeology : journal of the australasian society for historical archaeology"] = "Australas Hist Archaeol",
  ["australasian journal of dermatology"] = "Australas. J. Dermatol.",
  ["australasian journal of engineering education"] = "Australas. J. Eng. Educ.",
  ["australasian journal of environmental education"] = "Australas. J. Environ. Educ.",
  ["australasian journal of environmental management"] = "Australas. J. Environ. Manage.",
  ["australasian journal of philosophy"] = "Australas J Philos",
  ["australasian journal of regional studies"] = "Australasian J. Reg. Stud.",
  ["australasian journal of ultrasound in medicine"] = "Australas J Ultrasound Med",
  ["australasian journal of water resources"] = "Australas. J. Water Resour.",
  ["australasian journal on ageing"] = "Australas J Ageing",
  ["australasian lichenology"] = "Australas Lichenol",
  ["australasian mycologist : journal of the australasian mycological society, inc"] = "Aust Mycol",
  ["australasian nurses journal"] = "Australas. Nurses J.",
  ["australasian palaeontological memoirs"] = "Australas Palaeontol Mem",
  ["australasian philosophical review"] = "Australas. Philos. Rev.",
  ["australasian physical & engineering sciences in medicine"] = "Australas Phys Eng Sci Med",
  ["australasian physical & engineering sciences in medicine / supported by the australasian college of physical scientists in medicine and the australasian association of physical sciences in medicine"] = "Australas Phys Eng Sci Med",
  ["australasian physical and engineering sciences in medicine"] = "Australas. Phys. Eng. Sci. Med.",
  ["australasian plant disease notes"] = "Australas Plant Dis Notes",
  ["australasian plant pathology"] = "Australas. Plant Pathol.",
  ["australasian plant pathology : app"] = "Australas Plant Pathol",
  ["australasian psychiatry : bulletin of royal australian and new zealand college of psychiatrists"] = "Australas Psychiatry",
  ["australasian radiology"] = "Australas Radiol",
  ["australia and new zealand health policy"] = "Aust New Zealand Health Policy",
  ["australian & new zealand journal of statistics"] = "Aust. N. Z. J. Stat.",
  ["australian and new zealand journal of art"] = "Aust N Z J Art",
  ["australian and new zealand journal of medicine"] = "Aust N Z J Med",
  ["australian and new zealand journal of mental health nursing"] = "Aust. N. Z. J. Ment. Health Nurs.",
  ["australian and new zealand journal of obstetrics and gynaecology"] = "Aust. N. Z. J. Obstet. Gynaecol.",
  ["australian and new zealand journal of obstetrics and gynaecology. supplement"] = "Aust. N. Z. J. Obstet. Gynaecol. Suppl.",
  ["australian and new zealand journal of ophthalmology"] = "Aust N Z J Ophthalmol",
  ["australian and new zealand journal of psychiatry"] = "Aust. N. Z. J. Psychiatry",
  ["australian and new zealand journal of public health"] = "Aust N Z J Public Health",
  ["australian and new zealand journal of surgery"] = "Aust. N. Z. J. Surg.",
  ["australian archaeology"] = "Aust. Archaeol.",
  ["australian bulletin of labour"] = "Aust Bull Labour",
  ["australian clinical review"] = "Aust Clin Rev",
  ["australian clinical review / australian medical association [and] the australian council on hospital standards"] = "Aust Clin Rev",
  ["australian college of midwives incorporated journal"] = "Aust Coll Midwives Inc J",
  ["australian commodities forecasts and issues"] = "Australian Commodities Forecasts Issues",
  ["australian community psychologist (online)"] = "Aust Community Psychol",
  ["australian computer science communications"] = "Austral. Comput. Sci. Commun.",
  ["australian critical care"] = "Aust. Crit. Care",
  ["australian critical care : official journal of the confederation of australian critical care nurses"] = "Aust Crit Care",
  ["australian cultural history"] = "Aust Cult Hist",
  ["australian dental journal"] = "Aust Dent J",
  ["australian dental practice"] = "Aust Dent Pract",
  ["australian economic history review"] = "Aust Econ Hist Rev",
  ["australian economic papers"] = "Aust Econ Pap",
  ["australian economic review"] = "Australian Econ. Rev.",
  ["australian educational researcher"] = "Aust Educ Res",
  ["australian endodontic journal : the journal of the australian society of endodontology inc"] = "Aust Endod J",
  ["australian family physician"] = "Aust Fam Physician",
  ["australian feminist studies"] = "Aust Fem Stud",
  ["australian field ornithology"] = "Aust. Field Ornith.",
  ["australian forestry"] = "Aust For",
  ["australian gemmologist"] = "Aust. Gemmol.",
  ["australian geographer"] = "Aust. Geogr.",
  ["australian geographical studies"] = "Aust Geogr Stud",
  ["australian geomechanics journal"] = "Aust. Geomech. J.",
  ["australian health review"] = "Aust. Health Rev.",
  ["australian health review : a publication of the australian hospital association"] = "Aust Health Rev",
  ["australian historical studies"] = "Aust Hist Stud",
  ["australian hospital"] = "Aust Hosp",
  ["australian infection control : official journal of the australian infection control association inc"] = "Aust Infect Control",
  ["australian intellectual property journal"] = "Aust Intellect Prop J",
  ["australian journal of advanced nursing"] = "Aust. J. Adv. Nurs.",
  ["australian journal of agricultural and resource economics"] = "Aust. J. Agric. Resour. Econ.",
  ["australian journal of agricultural economics"] = "Aust. J. Agric. Econ.",
  ["australian journal of agricultural research"] = "Aust J Agric Res",
  ["australian journal of anthropology"] = "Aust. J. Anthropol.",
  ["australian journal of applied science"] = "Aust. J. Appl. Sci.",
  ["australian journal of basic and applied sciences"] = "Aust. J. Basic Appl. Sci.",
  ["australian journal of biological sciences"] = "Aust J Biol Sci",
  ["australian journal of biotechnology"] = "Aust J Biotechnol",
  ["australian journal of botany"] = "Aust J Bot",
  ["australian journal of chemical engineers"] = "Aust. J. Chem. Eng.",
  ["australian journal of chemistry"] = "Aust. J. Chem.",
  ["australian journal of civil engineering"] = "Aust. J. Civ. Eng.",
  ["australian journal of crop science"] = "Aust. J. Crop Sci.",
  ["australian journal of dermatology"] = "Aust J Dermatol",
  ["australian journal of early childhood"] = "Aust J Early Child",
  ["australian journal of earth sciences"] = "Aust. J. Earth Sci.",
  ["australian journal of ecology"] = "Aust. J. Ecol.",
  ["australian journal of education"] = "Aust J Educ",
  ["australian journal of education in chemistry"] = "Aust. J. Edu. Chem.",
  ["australian journal of electrical and electronics engineering"] = "Aust. J. Electr. Electron. Eng.",
  ["australian journal of entomology"] = "Aust J Entomol",
  ["australian journal of environmental education"] = "Aust. J. Environ. Educ.",
  ["australian journal of experimental agriculture"] = "Aust J Exp Agric",
  ["australian journal of experimental biology and medical science"] = "Aust. J. Exp. Biol. Med. Sci.",
  ["australian journal of experimental research"] = "Aust. J. Exp. Agric.",
  ["australian journal of family law"] = "Aust J Fam Law",
  ["australian journal of forensic sciences"] = "Aust. J. Forensic Sci.",
  ["australian journal of french studies"] = "Aust J Fr Stud",
  ["australian journal of general practice"] = "Aust J Gen Pract",
  ["australian journal of geodesy, photogrammetry, and surveying"] = "Aust J Geod Photogramm Surv",
  ["australian journal of grape and wine research"] = "Aust. J. Grape Wine Res.",
  ["australian journal of labour economics"] = "Australian J. Lab. Econ.",
  ["australian journal of learning difficulties"] = "Aust J  Learn Diffic",
  ["australian journal of management"] = "Australian J. Manage.",
  ["australian journal of marine & freshwater research"] = "Aust. J. Mar. Freshwater Res.",
  ["australian journal of maritime & ocean affairs"] = "Aust. J. Marit. Ocean Aff.",
  ["australian journal of marriage & family"] = "Aust J Marriage Fam",
  ["australian journal of mathematical analysis and applications"] = "Aust. J. Math. Anal. Appl.",
  ["australian journal of mechanical engineering"] = "Aust. J. Mech. Eng.",
  ["australian journal of midwifery : professional journal of the australian college of midwives incorporated"] = "Aust J Midwifery",
  ["australian journal of multi-disciplinary engineering"] = "Aust. J. Multi-Discip. Eng.",
  ["australian journal of ophthalmology"] = "Aust J Ophthalmol",
  ["australian journal of physics"] = "Aust. J. Phys.",
  ["australian journal of plant physiology"] = "Aust J Plant Physiol",
  ["australian journal of political science"] = "Aust J Polit Sci",
  ["australian journal of primary health"] = "Aust J Prim Health",
  ["australian journal of professional and applied ethics"] = "Aust J Prof Appl Ethics",
  ["australian journal of psychology"] = "Aust J Psychol",
  ["australian journal of public health"] = "Aust J Public Health",
  ["australian journal of rural health"] = "Aust. J. Rural Health",
  ["australian journal of science"] = "Aust. J. Sci",
  ["australian journal of science and medicine in sport"] = "Aust J Sci Med Sport",
  ["australian journal of scientific research. ser. b: biological sciences"] = "Aust J Sci Res B",
  ["australian journal of sex, marriage & family"] = "Aust J Sex Marriage And Fam",
  ["australian journal of soil research"] = "Aust. J. Soil Res.",
  ["australian journal of structural engineering"] = "Aust. J. Struct. Eng.",
  ["australian journal of water resources"] = "Aust. J. Water Resour.",
  ["australian journal of zoology"] = "Aust J Zool",
  ["australian journal on ageing"] = "Aust J Ageing",
  ["australian mammalogy"] = "Aust Mammal",
  ["australian mathematical society"] = "J. Austral. Math. Soc. Ser. A",
  ["australian mathematical society lecture series"] = "Austral. Math. Soc. Lect. Ser.",
  ["australian medical record journal"] = "Aust Med Rec J",
  ["australian medical record journal / medical record association of australia"] = "Aust Med Rec J",
  ["australian meteorological and oceanographic journal"] = "Aust. Meteorol. Oceanogr. J.",
  ["australian meteorological magazine"] = "Aust. Meteorol. Mag.",
  ["australian nurses’ journal"] = "Aust. Nurses J.",
  ["australian nursing & midwifery journal"] = "Aust Nurs Midwifery J",
  ["australian nursing journal"] = "Aust. Nurs. J.",
  ["australian nursing journal (july 1993)"] = "Aust Nurs J",
  ["australian nursing journal (june 1993)"] = "Aust Nurs J",
  ["australian occupational therapy journal"] = "Aust Occup Ther J",
  ["australian orchid review"] = "Aust Orchid Rev",
  ["australian orthodontic journal"] = "Aust Orthod J",
  ["australian paediatric journal"] = "Aust Paediatr J",
  ["australian prescriber"] = "Aust Prescr",
  ["australian prosthodontic journal"] = "Aust Prosthodont J",
  ["australian prosthodontic journal / australian prosthodontic society"] = "Aust Prosthodont J",
  ["australian prosthodontic society bulletin"] = "Aust Prosthodont Soc Bull",
  ["australian psychologist"] = "Aust Psychol",
  ["australian society"] = "Aust Soc",
  ["australian society of prosthodontists bulletin"] = "Aust Soc Prosthodontists Bull",
  ["australian systematic botany"] = "Aust Syst Bot",
  ["australian tax forum"] = "Australian Tax Forum",
  ["australian veterinary journal"] = "Aust Vet J",
  ["austrian history yearbook"] = "Austrian Hist Yearb",
  ["austrian journal of earth sciences"] = "Austrian J. Earth Sci.",
  ["austrian journal of earth sciences : an international journal of the austrian geological society"] = "Austrian J Earth Sci",
  ["austrian journal of forest science"] = "Austrian J. For. Sci.",
  ["austrian medical bulletin"] = "Austrian Med Bull",
  ["autism & developmental language impairments"] = "Autism Dev Lang Impair",
  ["autism : the international journal of research and practice"] = "Autism",
  ["autism in adulthood : challenges and management"] = "Autism Adulthood",
  ["autism policy & practice"] = "Autism Policy Pract",
  ["autism research : official journal of the international society for autism research"] = "Autism Res",
  ["autism research and treatment"] = "Autism Res Treat",
  ["autism-open access"] = "Autism Open Access",
  ["auto- immunity highlights"] = "Auto Immun Highlights",
  ["auto/biography studies : a/b"] = "Autobiogr Stud",
  ["autocarto research symposium"] = "Autocarto Res Symp",
  ["autoimmune and infectious diseases : open access"] = "Autoimmune Infect Dis",
  ["autoimmune diseases"] = "Autoimmune Dis",
  ["autoimmune diseases and therapeutic approaches : open access"] = "Autoimmune Dis Ther Approaches",
  ["autoimmunity (yverdon-les-bains"] = "Autoimmunity",
  ["autoimmunity reviews"] = "Autoimmun Rev",
  ["automated experimentation"] = "Autom Exp",
  ["automated reasoning series"] = "Automat. Reason. Ser.",
  ["automated software engineering"] = "Autom. Software Eng.",
  ["automatic control and computer sciences"] = "Autom. Control Comput. Sci.",
  ["automatic documentation and mathematical linguistics"] = "Autom. Doc. Math. Ling.",
  ["automatica : the journal of ifac, the international federation of automatic control"] = "Automatica (Oxf)",
  ["automatica. a journal of ifac, the international federation of automatic control"] = "Automatica J. IFAC",
  ["automation and remote control"] = "Autom. Remote Control",
  ["automation computers applied mathematics"] = "Automat. Comput. Appl. Math.",
  ["automation computers applied mathematics. scientific journal"] = "Automat. Comput. Appl. Math.",
  ["automation in construction"] = "Autom. Constr.",
  ["automation-control and industrial engineering series"] = "Autom.-Control Ind. Eng. Ser.",
  ["automotive and engine technology"] = "Automot. Engine Technol.",
  ["automotive industries ai"] = "Automot. Ind. AI",
  ["automotive innovation"] = "Automot. Innovation",
  ["autonomic & autacoid pharmacology"] = "Auton Autacoid Pharmacol",
  ["autonomic and autacoid pharmacology"] = "Auton. Autacoid Pharmacol.",
  ["autonomic neuroscience"] = "Auton. Neurosci.",
  ["autonomic neuroscience : basic & clinical"] = "Auton Neurosci",
  ["autonomic systems"] = "Auton. Syst.",
  ["autonomous agents and multi-agent systems"] = "Auton Agent Multi Agent Syst",
  ["autonomous intelligent systems"] = "Auton. Intell. Syst.",
  ["autonomous robots"] = "Auton Robots",
  ["autonomy (birmingham, england)"] = "Autonomy (Birm)",
  ["autophagy reports"] = "Autophagy Rep.",
  ["autopsy & case reports"] = "Autops Case Rep",
  ["aux editions belin"] = "Ed. Belin",
  ["aux plaines news"] = "Aux Plaines News",
  ["av communication review"] = "Av Commun Rev",
  ["ava research reviews"] = "AVA Res Rev",
  ["avance de investigación"] = "Av Invest",
  ["avances en investigación agropecuaria"] = "Av Investig Agropecu",
  ["avances en odontoestomatologia"] = "Av Odontoestomatol",
  ["avances en periodoncia"] = "Av Periodoncia",
  ["avances en psicologia clinica latinoamericana"] = "Av Piscol Clin Latinonot",
  ["avances en psicología clínica latinoamericana"] = "Av Piscol Clin Latinonot",
  ["avances en quimica"] = "Av. Quim.",
  ["avi : proceedings of the workshop on advanced visual interfaces. avi (conference)"] = "AVI",
  ["aviakosmicheskaia i ekologicheskaia meditsina"] = "Aviakosm. Ekolog. Med.",
  ["aviakosmicheskaia i ekologicheskaia meditsina = aerospace and environmental medicine"] = "Aviakosm Ekolog Med",
  ["avian biology research"] = "Avian Biol Res",
  ["avian conservation and ecology"] = "Avian Conserv. Ecol.",
  ["avian conservation and ecology = écologie et conservation des oiseaux"] = "Avian Conserv Ecol",
  ["avian diseases"] = "Avian Dis",
  ["avian diseases digest"] = "Avian Dis Dig",
  ["avian pathology"] = "Avian Pathol.",
  ["avian pathology : journal of the w.v.p.a"] = "Avian Pathol",
  ["avian research"] = "Avian Res",
  ["aviation space and environmental medicine"] = "Aviat. Space Environ. Med.",
  ["aviation week & space technology"] = "Aviat Week Space Technol",
  ["aviation, space and environmental medicine"] = "Aviat. Space Environ. Med.",
  ["aviation, space, and environmental medicine"] = "Aviat Space Environ Med",
  ["avicenna journal of clinical microbiology and infection : ajcmi"] = "Avicenna J Clin Microbiol Infect",
  ["avicenna journal of medical biotechnology"] = "Avicenna J Med Biotechnol",
  ["avicenna journal of medicine"] = "Avicenna J Med",
  ["avicenna journal of neuropsychophysiology"] = "Avicenna J Neuropsychophysiol",
  ["avicenna journal of phytomedicine"] = "Avicenna J Phytomed",
  ["avicultural magazine"] = "Avic. Mag.",
  ["aviso (ottawa, ont.)"] = "AVISO",
  ["avocetta (centro italiano studi ornitologici)"] = "Avocetta",
  ["avs news"] = "AVS News",
  ["avs news / association for voluntary sterilization"] = "AVS News",
  ["avs quantum science"] = "AVS Quantum Sci.",
  ["avsc news (association for voluntary surgical contraception (u.s.))"] = "AVSC News",
  ["avtomaticheskaya svarka"] = "Avtom. Svarka",
  ["awhonn lifelines"] = "AWHONN Lifelines",
  ["awhonn lifelines / association of women’s health, obstetric and neonatal nurses"] = "AWHONN Lifelines",
  ["awhonn voice"] = "AWHONN Voice",
  ["awhonn voice / association of women’s health, obstetric, and neonatal nurses (awhonn)"] = "AWHONN Voice",
  ["awhonn's clinical issues in perinatal and women's health nursing"] = "AWHONNS Clin Issues Perinat Womens Health Nurs",
  ["awhonn’s clinical issues in perinatal and women’s health nursing"] = "AWHONNS Clin. Issues Perinat. Womens Health Nurs.",
  ["awwa water science"] = "AWWA Water Sci",
  ["axone (dartmouth, n.s.)"] = "Axone",
  ["ayasofia müzesi yıllığı. annual of ayasofya museum"] = "AyasofyaMuezYil",
  ["azania: archaeological research in africa"] = "Azania: Archaeol. Res. Afr.",
  ["azerbaidzhanskii khimicheskii zhurnal"] = "Azerb. Khim. Zh.",
  ["azerbaijan journal of mathematics"] = "Azerb. J. Math.",
  ["azerbaĭdzhanskiĭ medit︠s︡inskiĭ zhurnal = aserbaidschanische medizinische zeitschrift"] = "Azerbaidzhanskii Meditsinskii Zhurnal",
  ["aziia i afrika segodnia"] = "Aziia Afr Segodnia",
  ["azii︠a︡ i afrika segodni︠a︡"] = "Aziia Afr Segodnia",
  ["azotea. revista de cultura del ayuntamiento de coria del río"] = "Azotea",
  ["ámbito de encuentros"] = "Ambito Encuentros",
  ["ärztliche forschung"] = "Arztl Forsch",
  ["ärztliche jugendkunde"] = "Arztl Jugendkd",
  ["ärztliche kosmetologie = medical cosmetology"] = "Arztl Kosmetol",
  ["ärztliche mitteilungen : nebst anzeiger und wissenschaftlicher beilage"] = "Arztl Mitt",
  ["ärztliche monatshefte für berufliche fortbildung. cahiers mensuels de médecine"] = "Arztl Monatshefte Berufl Fortbild",
  ["ärztliche sammelblätter"] = "Arztl Sammelbl",
  ["ärztliche wochenschrift"] = "Arztl Wochensch",
  ["ästhetische medizin"] = "Asthet Med (Berl)",
  ["årsbok. göteborgs tandläkare-sällskap"] = "Arsb Goteb Tandlak Sallsk",
  ["årsbok. odontologiska samfundet i finland"] = "Arsb Odontol Samf Finl",
  ["b. v. head, historia numorum. a manual of greek numismatics (oxford 1887"] = "Head",
  ["b.t.t.a. review"] = "BTTA Rev",
  ["baalbek. ergebnisse der ausgrabungen und untersuchungen in den jahren 1898 bis 1905"] = "Baalbek",
  ["babis̆ki vestnik"] = "Babiski Vestn",
  ["background paper (united states. physician payment review commission)"] = "Backgr Pap",
  ["backgrounder (washington, d.c.)"] = "Backgrounder",
  ["bacteriologia, virusologia, parazitologia, epidemiologia"] = "Bacteriol Virusol Parazitol Epidemiol (Bucur)",
  ["bacteriologia, virusologia, parazitologia, epidemiologia (bucharest, romania : 1990)"] = "Bacteriol Virusol Parazitol Epidemiol",
  ["bacteriological reviews"] = "Bacteriol Rev",
  ["badener neujahrsblätter"] = "Baden. Neujahrsbl.",
  ["badische fundberichte"] = "BadFuBer",
  ["baessler-archiv; beiträge zur völkerkunde"] = "Baessler Arch Beitr Volkerkd",
  ["baetica. estudios de arte, geografía e historia"] = "Baetica",
  ["bag. journal of basic and applied genetics"] = "BAG J Basic Appl Genet",
  ["baghdader forschungen"] = "BaF",
  ["baghdader mitteilungen"] = "BaM",
  ["bahia médica"] = "Bahia Med",
  ["bahrain medical bulletin : bmb"] = "Bahrain Med Bull",
  ["baillieres clinical endocrinology and metabolism"] = "Baillieres Clin Endocrinol Metab",
  ["baillieres clinical gastroenterology"] = "Baillieres Clin Gastroenterol",
  ["baillieres clinical haematology"] = "Baillieres Clin Haematol",
  ["baillieres clinical neurology"] = "Baillieres Clin Neurol",
  ["baillieres clinical obstetrics and gynaecology"] = "Baillieres Clin Obstet Gynaecol",
  ["baillieres clinical oncology"] = "Baillieres Clin Oncol",
  ["baillieres clinical rheumatology"] = "Baillieres Clin Rheumatol",
  ["bailliere’s best practice & research. clinical endocrinology & metabolism"] = "Baillieres Best Pract Res Clin Endocrinol Metab",
  ["bailliere’s best practice & research. clinical gastroenterology"] = "Baillieres Best Pract Res Clin Gastroenterol",
  ["bailliere’s best practice & research. clinical haematology"] = "Baillieres Best Pract Res Clin Haematol",
  ["bailliere’s best practice & research. clinical obstetrics & gynaecology"] = "Baillieres Best Pract Res Clin Obstet Gynaecol",
  ["bailliere’s best practice & research. clinical rheumatology"] = "Baillieres Best Pract Res Clin Rheumatol",
  ["bailliere’s clinical anaesthesiology"] = "Baillieres Clin Anaesthesiol",
  ["bailliere’s clinical endocrinology and metabolism"] = "Baillieres Clin Endocrinol Metab",
  ["bailliere’s clinical gastroenterology"] = "Baillieres Clin Gastroenterol",
  ["bailliere’s clinical haematology"] = "Baillieres Clin Haematol",
  ["bailliere’s clinical immunology and allergy"] = "Baillieres Clin Immunol Allergy",
  ["bailliere’s clinical neurology"] = "Baillieres Clin Neurol",
  ["bailliere’s clinical obstetrics and gynaecology"] = "Baillieres Clin Obstet Gynaecol",
  ["bailliere’s clinical rheumatology"] = "Baillieres Clin Rheumatol",
  ["baillière's best practice & research. clinical anaesthesiology"] = "Baillieres Best Pract Res Clin Anaesthesiol",
  ["baillière's best practice & research. clinical endocrinology & metabolism"] = "Baillieres Best Pract Res Clin Endocrinol Metab",
  ["baillière's best practice & research. clinical gastroenterology"] = "Baillieres Best Pract Res Clin Gastroenterol",
  ["baillière's best practice & research. clinical haematology"] = "Baillieres Best Pract Res Clin Haematol",
  ["baillière's best practice & research. clinical obstetrics & gynaecology"] = "Baillieres Best Pract Res Clin Obstet Gynaecol",
  ["baillière's best practice & research. clinical rheumatology"] = "Baillieres Best Pract Res Clin Rheumatol",
  ["baillière's clinical anaesthesiology"] = "Baillieres Clin Anaesthesiol",
  ["baillière's clinical endocrinology and metabolism"] = "Baillieres Clin Endocrinol Metab",
  ["baillière's clinical gastroenterology"] = "Baillieres Clin Gastroenterol",
  ["baillière's clinical haematology"] = "Baillieres Clin Haematol",
  ["baillière's clinical immunology and allergy"] = "Baillieres Clin Immunol Allergy",
  ["baillière's clinical neurology"] = "Baillieres Clin Neurol",
  ["baillière's clinical obstetrics and gynaecology"] = "Baillieres Clin Obstet Gynaecol",
  ["baillière's clinical rheumatology"] = "Baillieres Clin Rheumatol",
  ["baillière’s clinical endocrinology and metabolism"] = "Baillieres Clin. Endocrinol. Metab.",
  ["baillière’s clinical gastroenterology"] = "Baillieres Clin. Gastroenterol.",
  ["baillière’s clinical haematology"] = "Baillieres Clin. Haematol.",
  ["baillière’s clinical neurology"] = "Baillieres Clin. Neurol.",
  ["baillière’s clinical obstetrics and gynaecology"] = "Baillieres Clin. Obstet. Gynaecol.",
  ["baillière’s clinical rheumatology"] = "Baillieres Clin. Rheumatol.",
  ["baker-cederberg notebook. baker-cederberg museum & archives"] = "Baker Cederberg Noteb",
  ["balance (alexandria, va.)"] = "Balance",
  ["balcanica (rome, italy)"] = "Balcanica",
  ["baldwin's kentucky revised statutes annotated : krs. kentucky"] = "Baldwins Ky Revis Statut Annot Ky",
  ["baldwin’s kentucky revised statutes annotated : krs. kentucky"] = "Baldwins Ky Revis Statut Annot Ky",
  ["balisage series on markup technologies"] = "Balisage Ser Markup Technol",
  ["balkan journal of geometry and its applications"] = "Balkan J. Geom. Appl.",
  ["balkan journal of medical genetics : bjmg"] = "Balkan J Med Genet",
  ["balkan medical journal"] = "Balkan Med J",
  ["balkan society of geometers monographs and textbooks"] = "Balkan Soc. Geom. Monogr. Textb.",
  ["balkan studies"] = "BalkSt",
  ["baltic forestry"] = "Balt For",
  ["baltic journal of health and physical activity"] = "Balt J Health Phys Act",
  ["baltic journal of road and bridge engineering"] = "Balt. J. Road Bridge Eng.",
  ["baltic journal of sport & health sciences"] = "Balt J Sport Health Sci",
  ["baltimore health news"] = "Baltim Health News",
  ["baltimore magazine"] = "Baltim Mag",
  ["balácai közlemények"] = "BalacaiKoez",
  ["banach center publications"] = "Banach Center Publ.",
  ["banach journal of mathematical analysis"] = "Banach J. Math. Anal.",
  ["banat's journal of biotechnology"] = "Banats J Biotechnol",
  ["banbury report"] = "Banbury Rep",
  ["banca nazionale del lavoro quarterly review"] = "Banca Naz. Lavoro Quart. Rev.",
  ["banco de españa economic bulletin"] = "Banco España Econ. Bull.",
  ["bandaoti xuebao"] = "Bandaoti Xuebao",
  ["bangladesh development studies"] = "Bangladesh Dev Stud",
  ["bangladesh economic review"] = "Bangladesh Econ. Rev.",
  ["bangladesh journal of botany"] = "Bangladesh J Bot",
  ["bangladesh journal of microbiology"] = "Banglad J Microbiol",
  ["bangladesh journal of microbiology / bangladesh society of microbiologists"] = "Banglad J Microbiol",
  ["bangladesh journal of pharmacology"] = "Bangladesh J Pharmacol",
  ["bangladesh journal of plant taxonomy"] = "Bangladesh J Plant Taxon",
  ["bangladesh journal of zoology"] = "Bangladesh J Zool",
  ["bangladesh medical journal"] = "Bangladesh Med J",
  ["bangladesh medical research council bulletin"] = "Bangladesh Med Res Counc Bull",
  ["bangsaseon san-eob haghoeji"] = "Bangsaseon San Eob Haghoeji",
  ["bank of england quarterly bulletin"] = "Bank Engl. Quart. Bull.",
  ["bank of israel banking review"] = "Bank Israel Banking Rev.",
  ["bank of israel economic review"] = "Bank Israel Econ. Rev.",
  ["bank of japan monetary and economic studies"] = "Bank Japan Monet. Econ. Stud.",
  ["bank of korea economic papers"] = "Bank Korea Econ. Pap.",
  ["bank of valletta review"] = "Bank Valletta Rev.",
  ["banko janakari"] = "Banko Janakari",
  ["baoj cancer research & therapy"] = "BAOJ Cancer Res Ther",
  ["baoj obesity & weight loss management"] = "BAOJ Obes Weight Loss Manag",
  ["baoj pharmaceutical sciences"] = "BAOJ Pharm Sci",
  ["baoji college of arts and science"] = "J. Baoji College Arts Sci. Nat. Sci.",
  ["baptist history and heritage"] = "Baptist Hist Herit",
  ["barbados nursing journal"] = "Barbados Nurs J",
  ["barbastella (alcala de henares)"] = "Barbastella (Alcala Hen)",
  ["barcelona quirurgica"] = "Barc Quir",
  ["barcelona quirúrgica"] = "Barc Quir",
  ["barcelona respiratory network reviews"] = "Barc Respir Netw Rev",
  ["bariatric nursing and surgical patient care : official journal of the national association of bariatric nurses"] = "Bariatr Nurs Surg Patient Care",
  ["bariatric surgical practice and patient care"] = "Bariatr Surg Pract Patient Care",
  ["barley genetics newsletter"] = "Barley Genet Newsl",
  ["barrister (chicago, ill.)"] = "Barrister",
  ["barron's national business and financial weekly"] = "Barrons",
  ["barron’s national business and financial weekly"] = "Barrons",
  ["basal facts"] = "Basal Facts",
  ["basal facts (chicago, ill. : 1974)"] = "Basal Facts",
  ["basal ganglia"] = "Basal Ganglia",
  ["basic & clinical pharmacology & toxicology"] = "Basic Clin Pharmacol Toxicol",
  ["basic and applied dryland research"] = "Basic Appl Dryland Res",
  ["basic and applied ecology"] = "Basic Appl. Ecol.",
  ["basic and applied herpetology"] = "Basic Appl Herpetol",
  ["basic and applied histochemistry"] = "Basic Appl Histochem",
  ["basic and applied myology : bam"] = "Basic Appl Myol",
  ["basic and applied pathology"] = "Basic Appl Pathol",
  ["basic and applied social psychology"] = "Basic Appl Soc Psych",
  ["basic and clinical andrology"] = "Basic Clin Androl",
  ["basic and clinical neuroscience"] = "Basic Clin Neurosci",
  ["basic and clinical pharmacology and toxicology"] = "Basic Clin. Pharmacol. Toxicol.",
  ["basic life sciences"] = "Basic Life Sci",
  ["basic research in cardiology"] = "Basic Res Cardiol",
  ["basics of"] = "Basics",
  ["basin research"] = "Basin Res.",
  ["basiswissen statistik fur wirtschaftswissenschaftler"] = "Basiswissen Statist. Wirtschaftswiss.",
  ["basler beiträge zur geographie"] = "Basl. Beitr. Geogr.",
  ["basler beiträge zur geschichtswissenschaft"] = "Basl. Beitr. Gesch.wiss.",
  ["basler veröffentlichungen zur geschichte der medizin und der biologie"] = "Basler Veroff Gesch Med Biol",
  ["basler zeitschrift fur geschichte und altertumskunde / herausgegeben von der historischen und antiquarischen gesellechaft zu basel"] = "Basl Z Gesch Altertumskd",
  ["basler zeitschrift für geschichte und altertumskunde"] = "Basl Z Gesch Altertumskd",
  ["basler zeitschrift für geschichte und altertumskunde"] = "BZG",
  ["batteries & supercaps"] = "Batter Supercaps",
  ["batteries and electrochemistry"] = "Batteries Electrochem.",
  ["battery energy"] = "Battery Energy",
  ["bausch & lomb magazine"] = "Bausch Lomb Mag",
  ["bayer cropscience journal"] = "Bayer CropSci. J.",
  ["bayerische akademie der wissenschaften"] = "Bayer. Akad. Wiss. Math.-Natur. Kl. Abh. (N.F.)",
  ["bayerische akademie der wissenschaften. jahrbuch"] = "JbMuench",
  ["bayerische akademie der wissenschaften. philosophisch-historische klasse. abhandlungen"] = "AbhMünchen",
  ["bayerische akademie der wissenschaften. philosophisch-historische klasse. sitzungsberichte"] = "SBMuenchen",
  ["bayerische vorgeschichtsblätter"] = "BVBl",
  ["bayerisches ärzteblatt : amtliches organ der bayerischen landesärztekammer und ihrer bezirksvereine"] = "Bayer Aztebl",
  ["bayesian analysis"] = "Bayesian Anal.",
  ["bayesian and graphical models for biomedical imaging : first international workshop, bambi 2014, cambridge, ma, usa, september 18, 2014 ; revised selected papers. bambi (workshop) (1st : 2014 : cambridge, mass.)"] = "Bayesian Graph Models Biomed Imaging (2014)",
  ["baylor dental journal"] = "Baylor Dent. J.",
  ["baylor law review"] = "Bayl Law Rev",
  ["baylor nursing educator"] = "Baylor Nurs Educ",
  ["bayreuther mathematische schriften"] = "Bayreuth. Math. Schr.",
  ["bba clinical"] = "BBA Clin",
  ["bbgnt. berliner beitrage zur geschichte der naturwissenschaften und der technik"] = "BBGNT. Berliner Beitr. Gesch. Naturwiss. Tech.",
  ["bc studies"] = "B C Stud",
  ["bdd : be-khol derakhekha daʻehu : ketav-ʻet le-ʻinyene torah u-madaʻ"] = "BDD",
  ["bdd : be-khol derakhekha da’ehu : ketav-’et le-’inyene torah u-mada’"] = "BDD",
  ["bdj open"] = "BDJ Open",
  ["beagle : (darwin)"] = "Beagle (Darwin)",
  ["bedi kartlisa. revue de kartvélologie"] = "BediKart",
  ["bedside nurse"] = "Bedside Nurse",
  ["bee world"] = "Bee World",
  ["begg journal of orthodontic theory and treatment"] = "Begg J Orthod Theory Treat",
  ["beginnings (american holistic nurses' association)"] = "Beginnings",
  ["beginnings (american holistic nurses’ association)"] = "Beginnings",
  ["behavior analysis (washington, d.c.)"] = "Behav Anal (Wash D C)",
  ["behavior analysis in practice"] = "Behav Anal Pract",
  ["behavior genetics"] = "Behav Genet",
  ["behavior modification"] = "Behav Modif",
  ["behavior research methods"] = "Behav Res Methods",
  ["behavior research methods, instruments, & computers : a journal of the psychonomic society, inc"] = "Behav Res Methods Instrum Comput",
  ["behavior science research"] = "Behav Sci Res",
  ["behavior therapy"] = "Behav Ther",
  ["behavioral & social sciences librarian"] = "Behav Soc Sci Librar",
  ["behavioral and brain functions"] = "Behav. Brain Funct.",
  ["behavioral and brain functions : bbf"] = "Behav Brain Funct",
  ["behavioral and brain sciences"] = "Behav. Brain Sci.",
  ["behavioral and cognitive neuroscience reviews"] = "Behav Cogn Neurosci Rev",
  ["behavioral and neural biology"] = "Behav Neural Biol",
  ["behavioral assessment"] = "Behav Assess",
  ["behavioral biology"] = "Behav Biol",
  ["behavioral development bulletin"] = "Behav Dev Bull",
  ["behavioral disorders"] = "Behav Disord",
  ["behavioral ecology"] = "Behav. Ecol.",
  ["behavioral ecology : official journal of the international society for behavioral ecology"] = "Behav Ecol",
  ["behavioral ecology and sociobiology"] = "Behav Ecol Sociobiol",
  ["behavioral engineering"] = "Behav Eng",
  ["behavioral healthcare"] = "Behav Healthc",
  ["behavioral healthcare tomorrow"] = "Behav. Healthc. Tomorrow",
  ["behavioral medicine"] = "Behav. Med.",
  ["behavioral medicine (washington, d.c.)"] = "Behav Med",
  ["behavioral medicine update : a publication of the society of behavioral medicine"] = "Behav Med Update",
  ["behavioral neuropsychiatry"] = "Behav. Neuropsychiatry",
  ["behavioral neuroscience"] = "Behav Neurosci",
  ["behavioral research in accounting"] = "Behavioral Res. Acc.",
  ["behavioral science"] = "Behav. Sci.",
  ["behavioral sciences & the law"] = "Behav Sci Law",
  ["behavioral sciences (basel, switzerland)"] = "Behav Sci (Basel)",
  ["behavioral sciences and the law"] = "Behav. Sci. Law",
  ["behavioral sleep medicine"] = "Behav Sleep Med",
  ["behaviorists for social action journal"] = "Behav Soc Action J",
  ["behaviormetrics: quantitative approaches to human behavior"] = "Behaviormetr. Quant. Approaches Hum. Behav.",
  ["behaviour & information technology"] = "Behav Inf Technol",
  ["behaviour change : journal of the australian behaviour modification association"] = "Behav Change",
  ["behaviour research and therapy"] = "Behav Res Ther",
  ["behavioural and cognitive psychotherapy"] = "Behav Cogn Psychother",
  ["behavioural brain research"] = "Behav Brain Res",
  ["behavioural ecology"] = "Behav. Ecol.",
  ["behavioural ecology and sociobiology"] = "Behav. Ecol. Sociobiol.",
  ["behavioural neurology"] = "Behav Neurol",
  ["behavioural pharmacology"] = "Behav Pharmacol",
  ["behavioural processes"] = "Behav Processes",
  ["behring institute mitteilungen"] = "Behring Inst. Mitt.",
  ["behring institute mitteilungen (marburg/lahn"] = "Behring Inst Mitt",
  ["beiheft zu den annalen der schweizerischen meteorologischen anstalt"] = "Beih. Ann. Schweiz. meteorol. Anst.",
  ["beiheft zu den zeitschriften des schweiz. forstvereins"] = "Beih. Z. Schweiz. Forstver.",
  ["beiheft zu nova hedwigia"] = "Beih. Nova Hedwigia",
  ["beiheft zum jahrbuch der schweizerischen naturforschenden gesellschaft"] = "Beih. Jahrb. Schweiz. nat.forsch. Ges.",
  ["beiheft zur schweizerischen zeitschrift für forstwesen"] = "Beih. Schweiz. Z. Forst.wes.",
  ["beiheft zur zeitschrift für mykologie"] = "Beih. Z. Mykol.",
  ["beihefte zum botanischen centralblatt"] = "Beih. Bot. Cent.bl.",
  ["beijing da xue xue bao. yi xue ban (journal of peking university. health sciences)"] = "Beijing Da Xue Xue Bao",
  ["beijing da xue xue bao. yi xue ban = journal of peking university. health sciences"] = "Beijing Da Xue Xue Bao Yi Xue Ban",
  ["beijing daxue xuebao"] = "Beijing Daxue Xuebao Ziran Kexue Ban",
  ["beijing daxue xuebao. ziran kexue ban. acta scientiarum naturalium universitatis pekinensis"] = "Beijing Daxue Xuebao Ziran Kexue Ban",
  ["beijing institute of technology"] = "J. Beijing Inst. Tech.",
  ["beijing journal of pure and applied mathematics"] = "Beijing J. Pure Appl. Math.",
  ["beijing shifan daxue xuebao"] = "Beijing Shifan Daxue Xuebao",
  ["beijing shifan daxue xuebao. ziran kexue ban. journal of beijing normal university (natural science)"] = "Beijing Shifan Daxue Xuebao",
  ["beijing yi ke da xue xue bao"] = "Journal of Beijing Medical University|Beijing Yi Ke Da Xue Xue Bao",
  ["beilstein journal of nanotechnology"] = "Beilstein J. Nanotechnol.",
  ["beilstein journal of organic chemistry"] = "Beilstein J Org Chem",
  ["beitraege zur chemischen physiologie und pathologie"] = "Beitr. Chem. Physiol. Pathol.",
  ["beitraege zur entomologie"] = "Beitr. Entomol.",
  ["beitraege zur gerichtlichen medizin"] = "Beitr. Gerichtl. Med.",
  ["beitraege zur physik der atmosphaere"] = "Beitr. Phys. Atmos.",
  ["beitrage aus der plasmaphysik"] = "Beitr. Plasmaphys.",
  ["beitrage datenverarbeitung und unternehmensforschung"] = "Beitr. Datenverarb. Unternehmensforsch.",
  ["beitrage zu infusionstherapie und klinische ernahrung"] = "Beitr. Infusionther. Klin. Ernahr.",
  ["beitrage zur algebra und geometrie"] = "Beitr. Algebra Geom.",
  ["beitrage zur algebra und geometrie. contributions to algebra and geometry"] = "Beitr. Algebra Geom.",
  ["beitrage zur bolzano-forschung"] = "Beitr. Bolzano-Forsch.",
  ["beitrage zur deutschen philologie"] = "Beitr Dtsch Philol",
  ["beitrage zur forschungstechnologie"] = "Beitr. Forschungstech.",
  ["beitrage zur gerichtlichen medizin"] = "Beitr Gerichtl Med",
  ["beitrage zur geschichte der arbeiterbewegung : bzg"] = "Beitr Gesch Arb",
  ["beitrage zur geschichte der medizin und ihrer nebengebiete"] = "Beitr Gesch Med Nebengeb",
  ["beitrage zur geschichte der pharmazie und ihrer nachbargebiete"] = "Beitr Gesch Pharm Ihrer Nachbargeb",
  ["beitrage zur hochschul- und wissenschaftsgeschichte erfurts / herausgegeben vom rektor der medizinischen akademie erfurt"] = "Beitr Hochsch Wissenschaftsgesch Erfurts",
  ["beitrage zur hygiene und epidemiologie"] = "Beitr. Hyg. Epidemiol.",
  ["beitrage zur infusionstherapie"] = "Beitr. Infusionsther.",
  ["beitrage zur infusionstherapie und transfusionsmedizin"] = "Beitr. Infusionsther. Transfusionsmed.",
  ["beitrage zur klinik der tuberkulose und spezifischen tuberkulose-forschung"] = "Beitr Klin Tuberk Spezif Tuberkuloseforsch",
  ["beitrage zur klinik und erforschung der tuberkulose und der lungenkrankheiten"] = "Beitr. Klin. Erforsch. Tuberk. Lungenkr.",
  ["beitrage zur klinischen chirurgie"] = "Beitr Klin Chir",
  ["beitrage zur klinischen neurologie und psychiatrie"] = "Beitr. Klin. Neurol. Psychiatr.",
  ["beitrage zur krebsforschung"] = "Beitr. Krebsforsch.",
  ["beitrage zur krystallographie und mineralogie"] = "Beitr. Kryst. Mineral.",
  ["beitrage zur mathematik"] = "Beitr. Math.",
  ["beitrage zur neurochirurgie"] = "Beitr. Neurochir.",
  ["beitrage zur orthopadie und traumatologie"] = "Beitr. Orthop. Traumatol.",
  ["beitrage zur pathologie"] = "Beitr. Pathol.",
  ["beitrage zur pathologischen anatomie und zur allgemeinen pathologie"] = "Beitr. Pathol. Anat.",
  ["beitrage zur rheumatologie"] = "Beitr. Rheumatol.",
  ["beitrage zur sexualforschung"] = "Beitr. Sexualforsch.",
  ["beitrage zur silikose-forschung"] = "Beitr. Silikoseforsch.",
  ["beitrage zur silikose-forschung (pneumokoniose)"] = "Beitr Silikoseforsch Pneumokoniose",
  ["beitrage zur silikose-forschung. pneumokoniose"] = "Beitr. Silikoseforsch. Pneumokoniose",
  ["beitrage zur tropischen landwirtschaft und veterinarmedizin"] = "Beitr. Trop. Landwirtsch. Veterinarmed.",
  ["beitrage zur wurttembergischen apothekengeschichte"] = "Beitr Wurttemb Apothekengesch",
  ["beiträge zum römischen oberwinterthur – vitudurum. ausgrabungen im unteren bühl"] = "Vitudurum",
  ["beiträge zur neurochirurgie"] = "Beitr Neurochir",
  ["beiträge zur orthopädie und traumatologie"] = "Beitr Orthop Traumatol",
  ["beiträge zur pathologie"] = "Beitr Pathol",
  ["beiträge zur pathologischen anatomie und zur allgemeinen pathologie"] = "Beitr Pathol Anat",
  ["beiträge zur rheumatologie"] = "Beitr Rheumatol",
  ["beiträge zur sexualforschung"] = "Beitr Sexualforsch",
  ["beiträge zur silikose-forschung"] = "Beitr Silikoseforsch",
  ["beiträge zur silikose-forschung (pneumokoniose)"] = "Beitr Silikoseforsch Pneumokoniose",
  ["beiträge zur tabakforschung international"] = "Beitr Tab Int",
  ["beiträge zur tropischen landwirtschaft und veterinärmedizin"] = "Beitr Trop Landwirtsch Veterinarmed",
  ["beiträge zur württembergischen apothekengeschichte"] = "Beitr Wurttemb Apothekengesch",
  ["beiträge für die forstwirtschaft"] = "Beitr. Forstwirtsch.",
  ["beiträge für forstwirtschaft und landschaftsökologie"] = "Beitr. Forstwirtsch. Landsch.ökol.",
  ["beiträge zum naturschutz in der schweiz"] = "Beitr. Nat.schutz Schweiz",
  ["beiträge zur algebra und geometrie"] = "Beiträge Algebra Geom.",
  ["beiträge zur allgemeinen und vergleichenden archäologie"] = "BAVA",
  ["beiträge zur avifauna des rheinlandes"] = "Beitr. Avifauna Rheinl.",
  ["beiträge zur biologie der pflanzen"] = "Beitr. Biol. Pflanzen",
  ["beiträge zur erschliessung hellenistischer und kaiserzeitlicher skulptur und architektur"] = "BeitrESkAr",
  ["beiträge zur geobotanischen landesaufnahme der schweiz"] = "Beitr. geobot. Landesaufn. Schweiz",
  ["beiträge zur geologie der schweiz: hydrologie"] = "Beitr. Geol. Schweiz: Hydrol.",
  ["beiträge zur geschichte der medizin und ihrer nebengebiete"] = "Beitr Gesch Med Nebengeb",
  ["beiträge zur hochschul- und wissenschaftsgeschichte erfurts"] = "Beitr Hochsch Wissenschaftsgesch Erfurts",
  ["beiträge zur hydrologie der schweiz"] = "Beitr. Hydrol. Schweiz",
  ["beiträge zur jagd- und wildforschung"] = "Beitr. Jagd- Wildforsch.",
  ["beiträge zur kryptogamenflora der schweiz"] = "Beitr. Kryptogamenflora Schweiz",
  ["beiträge zur namenforschung"] = "BN",
  ["beiträge zur naturkundlichen forschung in südwestdeutschland"] = "Beitr. nat.kdl. Forsch. Südwestdtschl.",
  ["beiträge zur physik der atmosphäre"] = "Beitr. Zur Phys. Atmos.",
  ["beiträge zur sudanforschung"] = "BeitrSudanF",
  ["belcast journal of belizean affairs"] = "Belcast J Belizean Aff",
  ["beleid & maatschappij"] = "Beleid Maatsch",
  ["belgian journal of botany"] = "Belg. J. Bot.",
  ["belgian journal of botany : bulletin de la société royale de botanique de belgique"] = "Belg J Bot",
  ["belgian journal of operations research statistics and computer science (jorbel)"] = "Belg. J. Oper. Res. Statist. Comput. Sci.",
  ["belgian journal of zoology"] = "Belg J Zool",
  ["belgisch tijdschrift voor geneeskunde"] = "Belg Tijdschr Geneesk",
  ["belitung nursing journal"] = "Belitung Nurs J",
  ["belizean studies"] = "Belizean Stud",
  ["bell journal of economics"] = "Bell J. Econ.",
  ["bell laboratories record"] = "Bell Lab. Rec.",
  ["bell labs technical journal"] = "Bell Labs Tech. J.",
  ["bell system technical journal"] = "Bell Syst. Tech. J.",
  ["bellas artes (madrid)"] = "BelArt",
  ["belleten (turk tarih kurumu)"] = "Belleten",
  ["belleten (türk tarih kurumu)"] = "Belleten",
  ["belleten. türk tarih kurumu"] = "Belleten",
  ["belorusskiy politekhnicheskiy institut"] = "Teoret. i Prikl. Mekh. (Minsk)",
  ["benchcouncil transactions on benchmarks, standards and evaluations"] = "BenchCouncil Trans. Benchmarks Stand. Eval.",
  ["benchmark papers in human physiology"] = "Benchmark Pap Hum Physiol",
  ["benchmarking for quality management & technology"] = "Benchmarking Qual. Manage. Technol.",
  ["benchmarking: an international journal"] = "Benchmarking: Int. J.",
  ["bender's health care law monthly"] = "Benders Health Care Law Mon",
  ["bender’s health care law monthly"] = "Benders Health Care Law Mon.",
  ["beneficial microbes"] = "Benef Microbes",
  ["benefits quarterly"] = "Benefits Q",
  ["bengal, past & present : journal of the calcutta historical society"] = "Bengal Past Present",
  ["benha veterinary medical journal"] = "Benha Vet Med J",
  ["beni culturali e ambientali. sicilia"] = "BCASic",
  ["beni-suef university journal of basic and applied sciences"] = "Beni Suef Univ J Basic Appl Sci",
  ["benàcus. museo archeologico della val tenesi"] = "Benacus",
  ["beonline : journal of the west african bioethics training program"] = "BEOnline",
  ["berc bulletin"] = "BERC Bull",
  ["berc bulletin / basic education resource centre"] = "BERC Bull",
  ["bereavement care : for all those who help the bereaved"] = "Bereave Care",
  ["beretning. oslo. helserådet"] = "Beretning Oslo",
  ["berg- und huettenmaennische monatshefte"] = "Berg- Huettenmaenn. Monatsh.",
  ["berg- und hüttenmännische monatshefte"] = "Berg Huttenmannische Monatshefte",
  ["bericht atwd"] = "Ber. ATWD",
  ["bericht der bayerischen bodendenkmalpflege"] = "BerBayDenkmPfl",
  ["bericht der deutschen forschungsgemeinschaft"] = "BerDFG",
  ["bericht der römisch-germanischen kommission"] = "BerRGK",
  ["bericht der römisch-germanischen kommission des deutschen archäologischen instituts"] = "BRGK",
  ["bericht uber die zusammenkunft der deutschen ophthalmologischen gesellschaft"] = "Ber. Zusammenkunft Dtsch. Ophthalmol. Ges.",
  ["bericht uber die zusammenkunft. deutsche ophthalmologische gesellschaft"] = "Ber Zusammenkunft Dtsch Ophthalmol Ges",
  ["bericht über die zusammenkunft. deutsche ophthalmologische gesellschaft"] = "Ber Zusammenkunft Dtsch Ophthalmol Ges",
  ["bericht über die ausgrabungen in olympia"] = "OlBer",
  ["berichte aus dem institut fur statistik und versicherungsmathematik und aus dem institut fur angewandte statistik der freien universitat berlin"] = "Berichte aus dem Institut fur Statistik und Versicherungsmathematik und aus dem Institut fur Angewandte Statistik de",
  ["berichte aus dem institut fur statistik und wirtschaftsmathematik und aus dem institut fur angewandte statistik der freien universitat berlin"] = "Berichte aus dem Institut fur Statistik und Wirtschaftsmathematik und aus dem Institut fur Angewandte Statistik der",
  ["berichte aus der bayerischen landesanstalt für wald und forstwirtschaft"] = "Ber. Bayer. Landesanst. Wald Forstwirtsch.",
  ["berichte aus der bonner universitatsklinik und poliklinik fur mund-, zahn- und kieferkrankheiten"] = "Ber Bonn Univ Poliklin Mund Zahn Kieferkr",
  ["berichte aus der mathematik"] = "Ber. Math.",
  ["berichte der akademie für naturschutz und landschaftspflege"] = "Ber. Akad. Nat.schutz Landsch.pfl.",
  ["berichte der botanisch-zoologischen gesellschaft liechtenstein- sargans-werdenberg"] = "Ber. Bot.-Zool. Ges. Liecht.-Sargans-Werdenberg",
  ["berichte der bunsen-gesellschaft"] = "Ber. Bunsen Ges.",
  ["berichte der bunsen-gesellschaft fur physikalische chemie"] = "Ber. Bunsen Ges. Phys. Chem.",
  ["berichte der bunsen-gesellschaft für physikalische chemie"] = "Ber. Bunsen-Ges. Phys. Chem.",
  ["berichte der bunsen-gesellschaft physical chemistry chemical physics"] = "Ber. Bunsen-Ges. Phys. Chem",
  ["berichte der bunsengesellschaft für physikalische chemie"] = "Ber. Bunsenges. Phys. Chem.",
  ["berichte der deutschen botanischen gesellschaft"] = "Ber Dtsch Bot Ges",
  ["berichte der deutschen chemischen gesellschaft"] = "Ber. Dtsch. Chem. Ges.",
  ["berichte der deutschen gesellschaft fur geologische wissenschaft reihe a: geologie und palaeontologie"] = "Ber. Dtsch. Ges. Geol. Wiss. Reihe A",
  ["berichte der deutschen gesellschaft fur geologische wissenschaft reihe b: mineralogie und lagerstattenforschung"] = "Ber. Dtsch. Ges. Geol. Wiss. Reihe B",
  ["berichte der deutschen keramischen gesellschaft"] = "Ber. Dtsch. Keram. Ges.",
  ["berichte der deutschen mineralogischen gesellschaft"] = "Ber. Dtsch. Mineral. Ges.",
  ["berichte der deutschen pharmazeutischen gesellschaft"] = "Ber. Dtsch. Pharm. Ges.",
  ["berichte der deutschen physikalischen gesellschaft"] = "Ber. Dtsch. Phys. Ges.",
  ["berichte der deutschen wetterdienstes"] = "Ber. Dtsch. Wetterdienstes",
  ["berichte der eidg. forschungsanstalt für wald, schnee und landschaft"] = "Ber. Eidgenöss. Forsch.anst. Wald Schnee Landsch.",
  ["berichte der gesellschaft für mathematik und datenverarbeitung"] = "Ber. Gesellsch. Math. Datenverarb.",
  ["berichte der sankt gallischen naturwissenschaftlichen gesellschaft"] = "Ber. St. Gallen Nat.wiss. Ges.",
  ["berichte der schwyzerischen naturforschenden gesellschaft"] = "Ber. Schwyzerische Nat.forsch. Ges.",
  ["berichte der versuchsanstalt für wasserbau, hydrologie und glaziologie an der eidg. technischen hochschule zürich"] = "Ber. Vers.anst. Wasserbau Hydrol. Glaziol. Eidgenöss. Tech. Hochsch. Zür.",
  ["berichte des bundesamtes für wasser und geologie"] = "Ber. Bundesamt Wasser Geol.",
  ["berichte des deutschen wetterdienstes"] = "Ber. Dtsch. Wetterd.",
  ["berichte des forschungszentrums waldökosysteme"] = "Ber. Forsch.zent. Waldökosyst.",
  ["berichte des geobotanischen institutes der eidg. technischen hochschule, stiftung rübel"] = "Ber. Geobot. Inst. Eidgenöss. Tech. Hochsch., Stift. Rübel",
  ["berichte des german chapter of the acm"] = "Berichte German Chapter ACM",
  ["berichte eidg. anstalt für das forstliche versuchswesen"] = "Ber. Eidgenöss. Anst. forstl. Vers.wes.",
  ["berichte freiburger forstliche forschung"] = "Ber. Freibg. Forstl. Forsch.",
  ["berichte uber die gesamte physiologie und experimentalle pharmakologie"] = "Ber. Gesamte Physiol. Exp. Pharmakol.",
  ["berichte uber landwirtschaft"] = "Ber.. Landwirtsch.",
  ["berichte zur deutschen landeskunde"] = "Ber Dtsch Landeskd",
  ["berichte zur raumforschung und raumplanung (vienna, austria)"] = "Ber Raumforsch Raumplan",
  ["berichte zur wissenschaftsgeschichte"] = "Ber Wiss",
  ["berichte über die verhandlungen der sächsischen akademie der wissenschaften zu leipzig"] = "BerVerhLeipz",
  ["berichte. naturforschende gesellschaft zu freiburg i. b"] = "Ber Naturforsch Ges Freibg Br",
  ["berichte. physikalisch-medicinische gesellschaft, wurzburg"] = "Ber Phys Med Ges Wurzbg",
  ["berichte. physikalisch-medicinische gesellschaft, würzburg"] = "Ber Phys Med Ges Wurzbg",
  ["berichten van de rijksdienst voor het oudheidkundig bodemonderzoek"] = "BerOudhBod",
  ["berita biologi"] = "Ber Biol",
  ["berita jururawat"] = "Berita Jururawat",
  ["berkala ilmu kedokteran"] = "Berkala Ilmu Kedokteran",
  ["berkeley journal of sociology"] = "Berkeley J Sociol",
  ["berkeley municipal code, 1976 : a codification of the general ordinances of the city of berkeley, california / codified, indexed by book publishing company. berkeley (calif.)"] = "Berkeley Munic Code 1976 Berkeley Calif",
  ["berkeley municipal code, 1976 : a codification of the general ordinances of the city of berkeley, california. berkeley (calif.)"] = "Berkeley Munic Code 1976 Berkeley Calif",
  ["berkeley papers in history of science"] = "Berkeley Papers Hist. Sci.",
  ["berkeley symposium on mathematical statistics and probability"] = "Berkeley Symp. Math. Stat. Probab.",
  ["berkeley technology law journal"] = "Berkeley Technol Law J",
  ["berkeley technology law journal / boalt hall school of law, university of california, berkeley"] = "Berkeley Technol Law J",
  ["berkeley women's law journal"] = "Berkeley Womens Law J",
  ["berkeley women’s law journal"] = "Berkeley Womens Law J",
  ["berlin studies of the ancient world"] = "Berlin Stud. Anc. World",
  ["berlin-brandenburgische akademie der wissenschaften"] = "Berlin-Brandenburgische Akad. Wiss. Ber. Abh.",
  ["berliner beitraege zur archaeometrie"] = "Berl. Beitr. Archaeom.",
  ["berliner beiträge zur archäometrie"] = "BerlBeitrArchaeom",
  ["berliner blätter für vor- und frühgeschichte"] = "BerlBlVFruehGesch",
  ["berliner entomologische zeitschrift"] = "Berliner entomol. Z.",
  ["berliner geographische studien"] = "Berl Geogr Stud",
  ["berliner geowissenschaftliche abhandlungen, reihe a: geologie und palaeontologie"] = "Berl. Geowiss. Abh., Reihe A",
  ["berliner historische studien"] = "Berl Hist Stud",
  ["berliner jahrbuch für vor- und frühgeschichte"] = "BerlJbVFruehGesch",
  ["berliner journal fur soziologie"] = "Berl J Soziol",
  ["berliner journal für soziologie"] = "Berl J Soziol",
  ["berliner journal für soziologie"] = "Berliner J. Soz.",
  ["berliner medizin"] = "Berl Med",
  ["berliner museen"] = "BerlMus",
  ["berliner numismatische zeitschrift"] = "BerlNumZ",
  ["berliner paläobiologische abhandlungen"] = "Berl Palaobiologische Abh",
  ["berliner statistik"] = "Berl Stat",
  ["berliner statistik. monatsschrift"] = "Berl Stat Mon Schr",
  ["berliner statistik. monatsschrift / herausgeber, statistisches landesamt berlin"] = "Berl Stat Mon Schr",
  ["berliner studienreihe zur mathematik"] = "Berliner Studienreihe Math.",
  ["berliner tierarztliche wochenschrift"] = "Berl Tierarztl Wochenschr",
  ["berliner tierärztliche wochenschrift"] = "Berl Tierarztl Wochenschr",
  ["berliner und munchener tierarztliche wochenschrift"] = "Berl Munch Tierarztl Wochenschr",
  ["berliner und münchener tierärztliche wochenschrift"] = "Berl Munch Tierarztl Wochenschr",
  ["berner beiträge zur geschichte der medizin und der naturwissenschaften"] = "Berner Beitr Gesch Med Naturwiss Neue Folge",
  ["berner wald"] = "Bern. Wald",
  ["berner zeitschrift für geschichte und heimatkunde"] = "Bern. Z. Gesch. Heim.kd.",
  ["berner zeitung"] = "Bern. Ztg.",
  ["bernoulli : official journal of the bernoulli society for mathematical statistics and probability"] = "Bernoulli (Andover)",
  ["bernoulli. official journal of the bernoulli society for mathematical statistics and probability"] = "Bernoulli",
  ["bernstein series in computational neuroscience"] = "Bernstein Ser. Comput. Neurosci.",
  ["berufsdermatosen. monographien"] = "Berufsdermatosen (Monogr)",
  ["berytus. archaeological studies"] = "Berytus",
  ["best practice & research, clinical endocrinology & metabolism"] = "Best Pract. Res. Clin. Endocrinol. Metab.",
  ["best practice & research, clinical haematology"] = "Best Pract. Res. Clin. Haematol.",
  ["best practice & research. clinical anaesthesiology"] = "Best Pract Res Clin Anaesthesiol",
  ["best practice & research. clinical endocrinology & metabolism"] = "Best Pract Res Clin Endocrinol Metab",
  ["best practice & research. clinical gastroenterology"] = "Best Pract Res Clin Gastroenterol",
  ["best practice & research. clinical haematology"] = "Best Pract Res Clin Haematol",
  ["best practice & research. clinical obstetrics & gynaecology"] = "Best Pract Res Clin Obstet Gynaecol",
  ["best practice & research. clinical rheumatology"] = "Best Pract Res Clin Rheumatol",
  ["best practice & research: clinical anaesthesiology"] = "Best Pract. Res. Clin. Anaesthesiol.",
  ["best practice & research: clinical endocrinology and metabolism"] = "Best Pract. Res. Clin. Endocrinol. Metab.",
  ["best practice & research: clinical gastroenterology"] = "Best Pract. Res. Clin. Gastroenterol.",
  ["best practice & research: clinical haematology"] = "Best Pract. Res. Clin. Haematol.",
  ["best practice & research: clinical obstetrics and gynaecology"] = "Best Pract. Res. Clin. Obstet. Gynaecol.",
  ["best practice & research: clinical rheumatology"] = "Best Pract. Res. Clin. Rheumatol.",
  ["best practices & benchmarking in healthcare"] = "Best Pract. Benchmarking Healthc.",
  ["best practices and benchmarking in healthcare : a practical journal for clinical and management application"] = "Best Pract Benchmarking Healthc",
  ["best practices in mental health"] = "Best Pract Ment Health",
  ["best's review. life-health insurance edition"] = "Bests Rev Life Health Insur Ed",
  ["best’s review. life-health insurance edition"] = "Bests Rev Life Health Insur Ed",
  ["best’s review. life/health insurance edition"] = "Bests Rev. Life. Health Insur. Ed.",
  ["beta : bulletin of experimental treatments for aids : a publication of the san francisco aids foundation"] = "BETA",
  ["beta bulletin of experimental treatments for aids : a publication of the san francisco aids foundation"] = "BETA",
  ["beton- und stahlbetonbau"] = "Beton- Stahlbetonbau",
  ["betonwerk und fertigteil technik"] = "Betonwerk Fertigteil Tech.",
  ["better health"] = "Better Health (Lond)",
  ["better homes and gardens"] = "Better Homes Gard",
  ["between the species : a journal of ethics"] = "Between Species",
  ["beverage plant research"] = "Beverage Plant Res.",
  ["bevolking en gezin"] = "Bevolking Gezin",
  ["beyond relief"] = "Beyond Relief",
  ["bfpp : bulletin français de la pêche et de la protection des milieux aquatiques : bulletin français de la pêche et de la pisciculture"] = "BFPP Bull Fr Peche Prot Milieux Aquat",
  ["bhm berg- und huttenmannische monatshefte"] = "BHM Berg- Huttenmann. Monatsh.",
  ["bi-monthly research notes"] = "Bi-mon. Res. Notes",
  ["bib-mitteilungen : informationen aus dem bundesinstitut fur bevolkerungsforschung"] = "Bib Mitt",
  ["bib-mitteilungen : informationen aus dem bundesinstitut für bevölkerungsforschung"] = "Bib Mitt",
  ["bibe 2004 : proceedings : 19-21 may, 2004, taichung, taiwan, roc. ieee international symposium on bioinformatics and bioengineering (4th : 2004 : taichung, taiwan)"] = "BIBE 2004",
  ["bibliografia italiana di storia della scienza"] = "Bibliogr Ital Stor Sci",
  ["bibliografia italiana di storia della scienza / istituto e museo di storia della scienza"] = "Bibliogr Ital Stor Sci",
  ["bibliografía hispánica"] = "Bibliogr Hisp",
  ["bibliographia scientiae naturalis helvetica"] = "Bibliogr. Sci. Nat. Helv.",
  ["bibliographie internationale de la démographie historique"] = "Bibliogr Int Demogr Hist",
  ["bibliographie zur symbolik, ikonographie und mythologie. internationales referateorgan"] = "BiblSymb",
  ["bibliographies in the history of psychology and psychiatry"] = "Bibliogr Hist Psychol Psychiatry",
  ["bibliography of reproduction; a classified monthly list of references compiled from the research literature, vertebrates including man"] = "Bibliographr Reprod",
  ["biblioteca de la ciencia española"] = "Bibl. Cienc. Esp.",
  ["biblioteca de la revista matemática iberoamericana"] = "Bibl. Rev. Mat. Iberoamericana",
  ["biblioteca di bibliografia italiana"] = "Bibl. Bibliogr. Ital.",
  ["biblioteca di nuncius: studi e testi"] = "Bibl. Nuncius Studi Testi",
  ["biblioteca di physis"] = "Bibl. Physis",
  ["biblioteca di storia della scienza"] = "Bibl Stor Sci",
  ["biblioteca storica toscana : sezione di storia del risorgimento"] = "Bibl. Stor. T.",
  ["bibliotek for laeger"] = "Bibl Laeger",
  ["biblioteka ekonometryczna"] = "Bibl. Ekonom.",
  ["biblioteka higijenskog instituta nr srbije"] = "Bibl Hig Instituta NR Srb",
  ["biblioteka inzhenera"] = "Bibl. Inzh.",
  ["biblioteka po avtomatike"] = "Bibl. Avtomat.",
  ["biblioteka uchitelya matematiki"] = "Bibl. Uchitelya Mat.",
  ["biblioteka “znanie”"] = "Bibl. “Znanie”",
  ["bibliotheca anatomica"] = "Bibl Anat",
  ["bibliotheca botanica"] = "Bibl. Bot.",
  ["bibliotheca cardiologica"] = "Bibl Cardiol",
  ["bibliotheca classica orientalis"] = "BiblClOr",
  ["bibliotheca gastroenterologica"] = "Bibl Gastroenterol",
  ["bibliotheca gynaecologica"] = "Bibl Gynaecol",
  ["bibliotheca haematologica"] = "Bibl Haematol",
  ["bibliotheca lichenologica"] = "Bibl Lichenol",
  ["bibliotheca microbiologica"] = "Bibl Microbiol",
  ["bibliotheca mycologica"] = "Bibl Mycol",
  ["bibliotheca nutritio et dieta"] = "Bibl Nutr Dieta",
  ["bibliotheca ophthalmologica"] = "Bibl. Ophthalmol.",
  ["bibliotheca ophthalmologica : supplementa ad ophthalmologica"] = "Bibl Ophthalmol",
  ["bibliotheca orientalis"] = "BiOr",
  ["bibliotheca orientalis, uitg. van het nederlandsch instituut voor het nabije oosten"] = "BO",
  ["bibliotheca paediatrica"] = "Bibl Paediatr",
  ["bibliotheca phycologica"] = "Bibl Phycol",
  ["bibliotheca psychiatrica"] = "Bibl Psychiatr",
  ["bibliotheca psychiatrica et neurologica"] = "Bibl Psychiatr Neurol",
  ["bibliotheca radiologica"] = "Bibl Radiol",
  ["bibliotheca tuberculosea"] = "Bibl Tuberc",
  ["bibliothek des technischen allgemeinwissens"] = "Bibl. Techn. Allgemeinwiss.",
  ["bibliothek und wissenschaft"] = "Bibl Wiss",
  ["bibliotheque de litterature moderne"] = "Bib. Lit. Modern.",
  ["bibliotheque de l’ecole des chartes"] = "Bibl Ec Chartes",
  ["bibliotheque d’humanisme et renaissance; travaux et documents"] = "Bibl Humanisme Renaiss",
  ["bibliotheque scientifique"] = "Bibl. Sci.",
  ["bibliothèque d'humanisme et renaissance; travaux et documents"] = "Bibl Humanisme Renaiss",
  ["bibliothèque de l'école des chartes"] = "Bibl Ec Chartes",
  ["bibliothèque de l’école des chartes"] = "BECh",
  ["bibliothèque des écoles françaises d’athènes et de rome"] = "BEFAR",
  ["bibliothèque d’histoire des sciences"] = "Bibl. Hist. Sci.",
  ["bibliothèque d’humanisme et renaissance"] = "BiblH&R",
  ["bibliothèque d’études coptes, institut français d’archéologie orientale, kairo"] = "BdEC",
  ["bibliothèque d’études, institut français d’archéologie orientale, kairo"] = "BdE",
  ["biblische zeitschrift"] = "BZ",
  ["biblische zeitschrift (neue folge)"] = "BiZ",
  ["biedermanns zentralblatt"] = "Biedermanns Zentralbl.",
  ["biedermanns zentralblatt abteilung a: allgemeiner und referierender teil"] = "Biedermanns Zentralbl. Abt. A",
  ["biedermanns zentralblatt abteilung b: tierernahrung"] = "Biedermanns Zentralbl. Abt. B",
  ["bielefelder schriften zur didaktik der mathematik"] = "Bielefelder Schr. Didakt. Math.",
  ["bien-naitre / association beninoise pour la promotion de la famille. association beninoise pour la promotion de la famille"] = "Bien Naitre",
  ["bien-naître. association béninoise pour la promotion de la famille"] = "Bien Naitre",
  ["biennial review of counseling psychology"] = "Bienn Rev Couns Psychol",
  ["big and integrated artificial intelligence"] = "Big Integr. Artif. Intell.",
  ["big data"] = "Big Data",
  ["big data & society"] = "Big Data Soc",
  ["big data analytics"] = "Big Data Anal",
  ["big data analytics and knowledge discovery : 19th international conference, dawak 2017, lyon, france, august 28-31, 2017, proceedings. dawak (conference) (19th : 2017 : lyon, france)"] = "Big Data Anal Knowl Discov (2017)",
  ["big data and information analytics"] = "Big Data Inf. Anal.",
  ["big data management"] = "Big Data Manag.",
  ["big data mining and analytics"] = "Big Data Min. Anal.",
  ["big data research"] = "Big Data Res.",
  ["big earth data"] = "Big Earth Data",
  ["bihar mathematical society"] = "J. Bihar Math. Soc.",
  ["bihar mathematical society. journal"] = "J. Bihar Math. Soc.",
  ["biharean biologist"] = "Biharean Biol",
  ["bijblijven (amsterdam, netherlands)"] = "Bijblijven (Amst)",
  ["bijdragen en mededelingen (gelre, vereeniging tot beoefening van geldersche geschiedenis, oudheidkunde, en recht)"] = "Bijdr Meded",
  ["bijdragen en mededelingen betreffende de geschiedenis der nederlanden"] = "Bijdr Meded Geschied Ned",
  ["bijdragen tijdschrift voor filosofie en theologie"] = "Bijdragen",
  ["bijdragen tot de geschiedenis"] = "Bijdr Geschied",
  ["bijdragen tot de taal-, land- en volkenkunde"] = "Bijdr Taal Land Volkenkd",
  ["biken journal"] = "Biken J",
  ["bild der wissenschaft"] = "Bild Wiss",
  ["bildgebung = imaging"] = "Bildgebung",
  ["bildmessung und luftbildwesen"] = "Bildmess. Luftbildwes.",
  ["bilingual research journal"] = "Biling Res J",
  ["bilingualism (cambridge, england)"] = "Biling (Camb Engl)",
  ["bilten mednarodne federacije zobotehnikov (bulletin of the international federation of dental technicians)"] = "Bilt. Mednar. Fed. Zob. Teh.",
  ["bilten mednarodne federacije zobotehnikov = bulletin of the international federation of dental technicians. bulletin de la federation internationale prothesistes dentaires. bulletin der internationalen foderation der zahntechniker"] = "Bilt Mednar Fed Zob Teh",
  ["bilten udruzenja ortodonata jugoslavije"] = "Bilt. Udruz. Ortodonata Jugosl.",
  ["bilten udruz̆enja ortodonata jugoslavije = bulletin of orthodontic society of yugoslavia"] = "Bilt Udruz Ortodonata Jugosl",
  ["bilten za hematologiju i transfuziju"] = "Bilt Hematol Transfuz",
  ["bing du xue bao = chinese journal of virology"] = "Bing Du Xue Bao",
  ["bing du xue za zhi"] = "Virologica Sinica|Ping Tu Hsueh Tsa Chih",
  ["bing du xue za zhi = virologica sinica"] = "Ping Tu Hsueh Tsa Chih",
  ["binocular vision & strabismus quarterly"] = "Binocul Vis Strabismus Q",
  ["binocular vision & strabology quarterly, simms-romano's"] = "Binocul Vis Strabolog Q Simms Romano",
  ["binocular vision and strabismus quarterly"] = "Binocul. Vis. Strabismus Q.",
  ["bio systems"] = "Biosystems",
  ["bio web of conferences"] = "BIO Web Conf",
  ["bio-design and manufacturing"] = "Biodes Manuf",
  ["bio-medical instrumentation"] = "Biomed Instrum",
  ["bio-medical materials and engineering"] = "Biomed Mater Eng",
  ["bio-medical purview"] = "Biomed Purv",
  ["bio/medical instrumentation"] = "Biomed. Instrum.",
  ["bio/technology (nature publishing company)"] = "Biotechnology (N Y)",
  ["bioactive carbohydrates and dietary fibre"] = "Bioact. Carbohydr. Dietary Fibre",
  ["bioactive compounds in health and disease"] = "Bioact Compd Health Dis",
  ["bioactive materials"] = "Bioact Mater",
  ["bioanalytical reviews"] = "Bioanal Rev",
  ["bioarchaeology international"] = "Bioarchaeol Int",
  ["biocatalysis and agricultural biotechnology"] = "Biocatal Agric Biotechnol",
  ["biocatalysis and biotransformation"] = "Biocatal. Biotransform.",
  ["bioceramics development and applications"] = "Bioceram Dev Appl",
  ["biochemia medica"] = "Biochem Med (Zagreb)",
  ["biochemica (indianapolis, ind.)"] = "Biochemica (Indianapolis)",
  ["biochemical and biophysical research communications"] = "Biochem Biophys Res Commun",
  ["biochemical and cellular archives"] = "Biochem Cell Arch",
  ["biochemical and molecular medicine"] = "Biochem. Mol. Med.",
  ["biochemical archives"] = "Biochem Arch",
  ["biochemical clinics"] = "Biochem Clin",
  ["biochemical education"] = "Biochem Educ",
  ["biochemical engineering journal"] = "Biochem Eng J",
  ["biochemical genetics"] = "Biochem Genet",
  ["biochemical journal"] = "Biochem. J.",
  ["biochemical medicine"] = "Biochem Med",
  ["biochemical medicine and metabolic biology"] = "Biochem. Med. Metab. Biol.",
  ["biochemical pharmacology"] = "Biochem Pharmacol",
  ["biochemical pharmacology (amsterdam, netherlands)"] = "Biochem. Pharmacol. (Amsterdam, Neth.)",
  ["biochemical preparations"] = "Biochem. Prep.",
  ["biochemical society symposia"] = "Biochem. Soc. Symp.",
  ["biochemical society symposium"] = "Biochem Soc Symp",
  ["biochemical society transactions"] = "Biochem Soc Trans",
  ["biochemical systematics and ecology"] = "Biochem. Syst. Ecol.",
  ["biochemie und physiologie der pflanzen"] = "Biochem. Physiol. Pflanz.",
  ["biochemie und physiologie der pflanzen : bpp"] = "Biochem Physiol Pflanz",
  ["biochemische zeitschrift"] = "Biochem Z",
  ["biochemistry & molecular biology journal"] = "Biochem Mol Biol J",
  ["biochemistry & pharmacology : open access"] = "Biochem Pharmacol (Los Angel)",
  ["biochemistry & physiology"] = "Biochem Physiol",
  ["biochemistry (moscow)"] = "Biochemistry (Mosc.)",
  ["biochemistry (moscow) supplement. series a, membrane and cell biology"] = "Biochem (Mosc) Suppl Ser A Membr Cell Biol",
  ["biochemistry (moscow) supplement. series b, biomedical chemistry"] = "Biochem Mosc Suppl B Biomed Chem",
  ["biochemistry (rajkot, india)"] = "BioChemistry (Rajkot, India)",
  ["biochemistry and analytical biochemistry : current research"] = "Biochem Anal Biochem",
  ["biochemistry and biophysics reports"] = "Biochem Biophys Rep",
  ["biochemistry and biotechnology research (delta)"] = "Biochem Biotechnol Res",
  ["biochemistry and cell biology"] = "Biochem. Cell Biol.",
  ["biochemistry and cell biology = biochimie et biologie cellulaire"] = "Biochem Cell Biol",
  ["biochemistry and cell biology-biochimie et biologie cellulaire"] = "Biochem. Cell Biol.",
  ["biochemistry and experimental biology"] = "Biochem Exp Biol",
  ["biochemistry and molecular biology (new york, n.y.)"] = "Biochem Mol Biol (N Y)",
  ["biochemistry and molecular biology education"] = "Biochem. Mol. Biol. Educ.",
  ["biochemistry and molecular biology education : a bimonthly publication of the international union of biochemistry and molecular biology"] = "Biochem Mol Biol Educ",
  ["biochemistry and molecular biology international"] = "Biochem Mol Biol Int",
  ["biochemistry insights"] = "Biochem Insights",
  ["biochemistry international"] = "Biochem Int",
  ["biochemistry research international"] = "Biochem Res Int",
  ["biochemistry. biokhimiia"] = "Biochemistry (Mosc)",
  ["biochemistry. biokhimii︠a︡"] = "Biochemistry (Mosc)",
  ["biochemistry: an indian journal"] = "BioChem.: Indian J.",
  ["biochimica clinica"] = "Biochim Clin",
  ["biochimica et biophysica acta"] = "Biochim Biophys Acta",
  ["biochimica et biophysica acta (bba) - biophysics including photosynthesis"] = "Biochim. Biophys. Acta, Biophys. Incl. Photsynth.",
  ["biochimica et biophysica acta (bba) - specialized section on biophysical subjects"] = "Biochim. Biophys. Acta, Spec. Sect. Biophys. Subj.",
  ["biochimica et biophysica acta (bba) - specialized section on enzymological subjects"] = "Biochim. Biophys. Acta, Spec. Sect. Enzymol. Subj.",
  ["biochimica et biophysica acta (bba) - specialized section on lipids and related subjects"] = "Biochim. Biophys. Acta, Spec. Sect. Lipids Relat. Subj.",
  ["biochimica et biophysica acta (bba) - specialized section on mucoproteins and mucopolysaccharides"] = "Biochim. Biophys. Acta, Spec. Sect. Mucoproteins Mucopolysaccharides",
  ["biochimica et biophysica acta (bba) - specialized section on nucleic acids and related subjects"] = "Biochim. Biophys. Acta, Spec. Sect. Nucleic Acids Relat. Subj.",
  ["biochimica et biophysica acta, advances"] = "Biochim. Biophys. Acta, Adv.",
  ["biochimica et biophysica acta, bioenergetics"] = "Biochim. Biophys. Acta, Bioenerg.",
  ["biochimica et biophysica acta, biomembranes"] = "Biochim. Biophys. Acta, Biomembr.",
  ["biochimica et biophysica acta, biophysics including photosynthesis"] = "Biochim. Biophys. Acta, Biophys. Incl. Photosynth.",
  ["biochimica et biophysica acta, enzymology"] = "Biochim. Biophys. Acta, Enzymol.",
  ["biochimica et biophysica acta, enzymology and biological oxidation"] = "Biochim. Biophys. Acta, Enzymol. Biol. Oxid.",
  ["biochimica et biophysica acta, gene regulatory mechanisms"] = "Biochim. Biophys. Acta, Gene Regul. Mech.",
  ["biochimica et biophysica acta, gene structure and expression"] = "Biochim. Biophys. Acta, Gene Struct. Expression",
  ["biochimica et biophysica acta, general subjects"] = "Biochim. Biophys. Acta, Gen. Subj.",
  ["biochimica et biophysica acta, lipids and lipid metabolism"] = "Biochim. Biophys. Acta, Lipids Lipid Metab.",
  ["biochimica et biophysica acta, molecular and cell biology of lipids"] = "Biochim. Biophys. Acta, Mol. Cell. Biol. Lipids",
  ["biochimica et biophysica acta, molecular basis of disease"] = "Biochim. Biophys. Acta, Mol. Basis Dis.",
  ["biochimica et biophysica acta, molecular cell research"] = "Biochim. Biophys. Acta, Mol. Cell. Res.",
  ["biochimica et biophysica acta, mucoproteins and mucopolysaccharides"] = "Biochim. Biophys. Acta, Mucroproteins Mucopolysaccharides",
  ["biochimica et biophysica acta, nucleic acids and protein synthesis"] = "Biochim. Biophys. Acta, Nucleic Acids Protein Synth.",
  ["biochimica et biophysica acta, protein structure"] = "Biochim. Biophys. Acta, Protein Struct.",
  ["biochimica et biophysica acta, protein structure and molecular enzymology"] = "Biochim. Biophys. Acta, Protein Struct. Mol. Enzymol.",
  ["biochimica et biophysica acta, proteins and proteomics"] = "Biochim. Biophys. Acta, Proteins Proteomics",
  ["biochimica et biophysica acta, reviews on bioenergetics"] = "Biochim. Biophys. Acta, Rev. Bioenerg.",
  ["biochimica et biophysica acta, reviews on biomembranes"] = "Biochim. Biophys. Acta, Rev. Biomembr.",
  ["biochimica et biophysica acta, reviews on cancer"] = "Biochim. Biophys. Acta, Rev. Cancer",
  ["biochimica et biophysica acta, specialized section on biophysical subjects"] = "Biochim. Biophys. Acta, Spec. Sect. Biophys. Subj.",
  ["biochimica et biophysica acta, specialized section on enzymological subjects"] = "Biochim. Biophys. Acta, Spec. Sect. Enzymol. Subj.",
  ["biochimica et biophysica acta, specialized section on lipids and related subjects"] = "Biochim. Biophys. Acta, Spec. Sect. Lipids Relat. Subj.",
  ["biochimica et biophysica acta. bioenergetics"] = "Biochim Biophys Acta Bioenerg",
  ["biochimica et biophysica acta. biomembranes"] = "Biochim Biophys Acta Biomembr",
  ["biochimica et biophysica acta. gene regulatory mechanisms"] = "Biochim Biophys Acta Gene Regul Mech",
  ["biochimica et biophysica acta. general subjects"] = "Biochim Biophys Acta Gen Subj",
  ["biochimica et biophysica acta. molecular and cell biology of lipids"] = "Biochim Biophys Acta Mol Cell Biol Lipids",
  ["biochimica et biophysica acta. molecular basis of disease"] = "Biochim Biophys Acta Mol Basis Dis",
  ["biochimica et biophysica acta. molecular cell research"] = "Biochim Biophys Acta Mol Cell Res",
  ["biochimica et biophysica acta. proteins and proteomics"] = "Biochim Biophys Acta Proteins Proteom",
  ["biochimica et biophysica acta. reviews on cancer"] = "Biochim Biophys Acta Rev Cancer",
  ["biochimie open"] = "Biochim Open",
  ["biochip journal"] = "Biochip J",
  ["bioconjugate chemistry"] = "Bioconjug Chem",
  ["biocontrol (dordrecht, netherlands)"] = "Biocontrol (Dordr)",
  ["biocontrol science"] = "Biocontrol Sci",
  ["biocontrol science and technology"] = "Biocontrol Sci. Technol.",
  ["biocybernetics and biomedical engineering"] = "Biocybern Biomed Eng",
  ["biodata mining"] = "BioData Min",
  ["biodemography and social biology"] = "Biodemography Soc Biol",
  ["biodesign research"] = "Biodes. Res.",
  ["biodiversitas : journal of biological diversity"] = "Biodiversitas",
  ["biodiversity (nepean, ont.)"] = "Biodiversity (Nepean)",
  ["biodiversity and conservation"] = "Biodivers Conserv",
  ["biodiversity and ecology = biodiversität und ökologie"] = "Biodivers Ecol",
  ["biodiversity data journal"] = "Biodivers Data J",
  ["biodiversity information science and standards"] = "Biodivers. Inf. Sci. Stand.",
  ["biodiversity journal"] = "Biodivers J",
  ["biodiversity letters"] = "Biodivers. Lett.",
  ["biodiversity research and conservation"] = "Biodivers Res Conserv",
  ["biodiversity science"] = "Biodivers. Sci.",
  ["biodrugs : clinical immunotherapeutics, biopharmaceuticals and gene therapy"] = "BioDrugs",
  ["bioelectrochemistry (amsterdam, netherlands)"] = "Bioelectrochemistry",
  ["bioelectrochemistry and bioenergetics"] = "Bioelectrochem. Bioenerg.",
  ["bioelectrochemistry and bioenergetics (lausanne, switzerland)"] = "Bioelectrochem Bioenerg",
  ["bioelectronic medicine"] = "Bioelectron Med",
  ["bioelectronics in medicine"] = "Bioelectron Med (Lond)",
  ["bioenergetics : open access"] = "Bioenergetics",
  ["bioenergy research"] = "Bioenergy Res",
  ["bioengineered bugs"] = "Bioeng Bugs",
  ["bioengineering & translational medicine"] = "Bioeng Transl Med",
  ["bioengineering (basel, switzerland)"] = "Bioengineering (Basel)",
  ["bioengineering today"] = "Bioeng. Today",
  ["bioessays : news and reviews in molecular, cellular and developmental biology"] = "Bioessays",
  ["bioethica belgica"] = "Bioethica Belg",
  ["bioethics bulletin (washington, d.c.)"] = "Bioeth Bull",
  ["bioethics digest"] = "Bioethics Dig",
  ["bioethics examiner"] = "Bioeth Exam",
  ["bioethics forum"] = "Bioethics Forum",
  ["bioethics news"] = "Bioeth News",
  ["bioethics northwest"] = "Bioethics Northwest",
  ["bioethics quarterly"] = "Bioethics Q",
  ["bioethics research notes"] = "Bioeth Res Notes",
  ["bioetica y debat"] = "Bioet Debat",
  ["bioètica & debat"] = "Bioet Debat",
  ["bioética (brasília, brazil)"] = "Bioetica",
  ["biofactors (oxford, england)"] = "Biofactors",
  ["biofeedback and self regulation"] = "Biofeedback Self Regul.",
  ["biofeedback and self-regulation"] = "Biofeedback Self Regul",
  ["bioformosa = sheng wu xue bao"] = "Bioformosa",
  ["bioforum : biologische chemie, mikrobiologie, genetik, immunologie, virologie, biotechnologie"] = "Bioforum",
  ["bioforum europe : trends and techniques in life science research"] = "Bioforum Eur",
  ["biofuel research journal"] = "Biofuel Res. J.",
  ["biofuels, bioproducts & biorefining"] = "Biofuels, Bioprod. Biorefin.",
  ["biofuels, bioproducts & biorefining : biofpr"] = "Biofuel Bioprod Biorefin",
  ["biofuels, bioproducts and biorefining"] = "Biofuels, Bioprod. Biorefin.",
  ["biogenesis : jurnal ilmiah biologia"] = "Biogenesis",
  ["biogenic amines"] = "Biog. Amines",
  ["biogeographia : the journal of integrative biogeography"] = "Biogeographia",
  ["biogeosciences (online)"] = "Biogeosciences",
  ["biogeosciences discussions"] = "Biogeosci. Discuss.",
  ["biogeosciences discussions : (bgd)"] = "Biogeosci Discuss",
  ["biogéographica : compte-rendu des séances de la société de biogéographie"] = "Biogeographica",
  ["biogradski zbornik"] = "BiogrZbor",
  ["biographical memoirs of fellows of the royal society. royal society (great britain)"] = "Biogr Mem Fellows R Soc",
  ["biographical memoirs. national academy of sciences (u.s.)"] = "Biogr Mem Natl Acad Sci",
  ["biography and genealogy master index"] = "Biogr Geneal Master Index",
  ["bioimpacts : bi"] = "Bioimpacts",
  ["bioinformatics (oxford, england)"] = "Bioinformatics",
  ["bioinformatics advances"] = "Bioinf. Adv",
  ["bioinformatics and biology insights"] = "Bioinform Biol Insights",
  ["bioinformatics and biomedical engineering : 7th international work-conference, iwbbio 2019, granada, spain, may 8-10, 2019, proceedings, parts i and ii. iwbbio (conference) (7th : 2019 : granada, spain)"] = "Bioinform Biomed Eng (2019)",
  ["bioinformatics research and applications : 14th international symposium, isbra 2018, beijing, china, june 8-11, 2018, proceedings. isbra (conference) (14th : 2018 : beijing, china)"] = "Bioinform Res Appl (2018)",
  ["bioinorganic chemistry"] = "Bioinorg Chem",
  ["bioinorganic chemistry and applications"] = "Bioinorg Chem Appl",
  ["bioinorganic reaction mechanisms"] = "Bioinorg React Mech",
  ["bioinspiration & biomimetics"] = "Bioinspir Biomim",
  ["bioinspired, biomimetic and nanobiomaterials"] = "Bioinspired Biomimetic Nanobiomater.",
  ["biointerface research in applied chemistry"] = "Biointerface Res Appl Chem",
  ["bioinvasions records"] = "Bioinvasions Rec",
  ["biokhimiia (moscow, russia)"] = "Biokhimiia",
  ["biokhimii︠a︡ (moscow, russia)"] = "Biokhimiia",
  ["biokhimiya (moscow)"] = "Biokhimiya (Moscow)",
  ["biolife : an international quarterly journal of biology and life sciences"] = "Biolife",
  ["biologia (lahore, pakistan)"] = "Biologia",
  ["biologia ambientale"] = "Biol Ambient",
  ["biologia futura"] = "Biol Futur",
  ["biologia gallo-hellenica"] = "Biol. Gallo-hellenica",
  ["biologia marina mediterranea"] = "Biol Mar Mediterr",
  ["biologia medica"] = "Biol Med (Niteroi Braz)",
  ["biologia neonatorum"] = "Biol. Neonat.",
  ["biologia neonatorum. neo-natal studies"] = "Biol Neonat",
  ["biologia plantarum"] = "Biol Plant",
  ["biologica (santiago, chile)"] = "Biologica (Santiago)",
  ["biologica latina"] = "Biol Lat",
  ["biologica nyssana"] = "Biol Nyssana",
  ["biological & pharmaceutical bulletin"] = "Biol Pharm Bull",
  ["biological agriculture & horticulture"] = "Biol. Agric. Hortic.",
  ["biological and biomedical reports"] = "Biol Biomed Rep",
  ["biological and medical physics, biomedical engineering"] = "Biol. Med. Phys. Biomed. Eng.",
  ["biological and medical research division semiannual report. argonne national laboratory. division of biological and medical research"] = "Biol Med Res Div Semiannu Rep (Argonne National Laboratory)",
  ["biological and pharmaceutical bulletin"] = "Biol. Pharm. Bull.",
  ["biological bulletin"] = "Biol. Bull.",
  ["biological chemistry"] = "Biol Chem",
  ["biological chemistry hoppe-seyler"] = "Biol. Chem. Hoppe-Seyler",
  ["biological chemistry hoppe-seyler. supplement"] = "Biol Chem Hoppe Seyler Suppl",
  ["biological conservation"] = "Biol Conserv",
  ["biological control"] = "Biol. Control",
  ["biological control : theory and applications in pest management"] = "Biol Control",
  ["biological cybernetics"] = "Biol Cybern",
  ["biological diversity and conservation"] = "Biol Divers Conserv",
  ["biological engineering"] = "Biol. Eng.",
  ["biological engineering transactions"] = "Biol Eng Trans",
  ["biological forum : an international journal"] = "Biol Forum",
  ["biological imaging"] = "Biol. Imaging",
  ["biological invasions"] = "Biol Invasions",
  ["biological journal of the linnean society"] = "Biol. J. Linn. Soc.",
  ["biological journal of the linnean society of london"] = "Biol. J. Linn. Soc. Lond.",
  ["biological journal of the linnean society. linnean society of london"] = "Biol J Linn Soc Lond",
  ["biological magnetic resonance"] = "Biol Magn Reson",
  ["biological mass spectrometry"] = "Biol Mass Spectrom",
  ["biological membranes"] = "Biol. Membr.",
  ["biological memoirs"] = "Biol Mem",
  ["biological oceanography"] = "Biol. Oceanogr.",
  ["biological procedures online"] = "Biol Proced Online",
  ["biological psychiatry"] = "Biol Psychiatry",
  ["biological psychiatry. cognitive neuroscience and neuroimaging"] = "Biol Psychiatry Cogn Neurosci Neuroimaging",
  ["biological psychiatry: cognitive neuroscience and neuroimaging"] = "Biol. Psychiatry: Cognit. Neurosci. Neuroimaging",
  ["biological psychology"] = "Biol Psychol",
  ["biological psychology bulletin"] = "Biol Psychol Bull",
  ["biological research"] = "Biol Res",
  ["biological research for nursing"] = "Biol Res Nurs",
  ["biological research in pregnancy and perinatology"] = "Biol Res Pregnancy Perinatol",
  ["biological reviews"] = "Biol. Rev.",
  ["biological reviews of the cambridge philosophical society"] = "Biol Rev Camb Philos Soc",
  ["biological rhythm research"] = "Biol Rhythm Res",
  ["biological sciences"] = "Prog Nucl Energy 6 Biol Sci",
  ["biological sciences in space 1986 : proceedings of the international symposium on biological sciences in space, nagoya (japan), november 10-12, 1986"] = "Biol Sci Space 1986",
  ["biological signals"] = "Biol Signals",
  ["biological signals and receptors"] = "Biol Signals Recept",
  ["biological structures and morphogenesis"] = "Biol Struct Morphog",
  ["biological systems, open access"] = "Biol Syst Open Access",
  ["biological systems: open access"] = "Biol. Syst. Open Access",
  ["biological theory"] = "Biol Theory",
  ["biological trace element research"] = "Biol Trace Elem Res",
  ["biological wastes"] = "Biol. Wastes",
  ["biologically inspired cognitive architectures"] = "Biol. Inspired Cognit. Archit.",
  ["biologically-inspired systems"] = "Biol.-Inspired Syst.",
  ["biologicals : journal of the international association of biological standardization"] = "Biologicals",
  ["biologicas : revista de la des ciencias biológico agropecuarias universidad michoacana de san nicolás de hidalgo"] = "Biologicas",
  ["biologicheskie membrany"] = "Biol Membrany",
  ["biologicheskie membrany / akademiia nauk sssr"] = "Biol Membr",
  ["biologicheskie nauki"] = "Biol. Nauki",
  ["biologicheskie nauki (moscow)"] = "Biol. Nauki (Moscow)",
  ["biologicheskiĭ zhurnal armenii = hayastani kensabanakan handes"] = "Biol Z Armen",
  ["biologicke listy"] = "Biol Listy",
  ["biologické listy"] = "Biol Listy",
  ["biologické listy. supplementum"] = "Biol Listy Suppl",
  ["biologics : targets & therapy"] = "Biologics",
  ["biologics in therapy"] = "Biol Ther",
  ["biologics: targets and therapy"] = "Biol.: Targets Ther.",
  ["biologie aujourd'hui"] = "Biol Aujourdhui",
  ["biologie cellulaire"] = "Biol Cell",
  ["biologie et gastro-enterologie"] = "Biol. Gastroenterol. (Paris)",
  ["biologie et gastro-entérologie"] = "Biol Gastroenterol (Paris)",
  ["biologie et médecine"] = "Biol Med",
  ["biologie in unserer zeit"] = "Biol. unserer Zeit",
  ["biologie medicale"] = "Biol. Med. (Paris)",
  ["biologie médicale"] = "Biol Med (Paris)",
  ["biologija (vilnius, lithuania)"] = "Biologija (Vilnius)",
  ["biologinė psichiatrija ir psichofarmakologija = biological psychiatry and psychopharmacology"] = "Biol Psichiatr Psichofarmakol",
  ["biologische rundschau"] = "Biol. Rundsch.",
  ["biologisches zentralblatt"] = "Biol Zent Bl",
  ["biologist (columbus, ohio)"] = "Biologist",
  ["biologist (london, england)"] = "Biologist (London)",
  ["biology & philosophy"] = "Biol Philos",
  ["biology and environment: proceedings of the royal irish academy"] = "Biol. Environ.: Proc. R. Ir. Acad.",
  ["biology and fertility of soils"] = "Biol. Fertil. Soils",
  ["biology and human affairs : a british social hygiene council publication"] = "Biol Hum Aff",
  ["biology and medicine (aligarh)"] = "Biol Med (Aligarh)",
  ["biology and philosophy"] = "Biol. Philos.",
  ["biology and society : the journal of the eugenics society"] = "Biol Soc",
  ["biology bulletin"] = "Biol. Bull.",
  ["biology bulletin of the academy of sciences of the ussr"] = "Biol Bull Acad Sci USSR",
  ["biology bulletin of the russian academy of sciences"] = "Biol Bull Russ Acad Sci",
  ["biology bulletin reviews"] = "Biol. Bull. Rev.",
  ["biology direct"] = "Biol Direct",
  ["biology letters"] = "Biol Lett",
  ["biology methods & protocols"] = "Biol Methods Protoc",
  ["biology of blood and marrow transplantation"] = "Biol. Blood Marrow Transplant.",
  ["biology of blood and marrow transplantation : journal of the american society for blood and marrow transplantation"] = "Biol Blood Marrow Transplant",
  ["biology of metals"] = "Biol Met",
  ["biology of mood & anxiety disorders"] = "Biol Mood Anxiety Disord",
  ["biology of reproduction"] = "Biol Reprod",
  ["biology of reproduction. supplement"] = "Biol Reprod Suppl",
  ["biology of sex differences"] = "Biol Sex Differ",
  ["biology of sport"] = "Biol Sport",
  ["biology of sport / institute of sport"] = "Biol Sport",
  ["biology of the cell"] = "Biol Cell",
  ["biology of the cell / under the auspices of the european cell biology organization"] = "Biol Cell",
  ["biology of the neonate"] = "Biol Neonate",
  ["biology open"] = "Biol Open",
  ["biology, engineering and medicine"] = "Biol Eng Med",
  ["biolohichnyĭ visnyk melitopolʹsʹkoho derz︠h︡avnoho pedahohichnoho universytetu imeni bohdana khmelʹnyt︠s︡koho"] = "Biol Visnyk Melitopolskoho Derzhavnoho Pedahohichnoho Universytetu Im Bohdana Khmelnytskoho",
  ["biomacromolecular mass spectrometry"] = "Biomacromol. Mass Spectrom.",
  ["biomagnetic research and technology"] = "Biomagn Res Technol",
  ["biomanufacturing reviews"] = "Biomanuf. Rev.",
  ["biomarker insights"] = "Biomark Insights",
  ["biomarker research"] = "Biomark Res",
  ["biomarkers : biochemical indicators of exposure, response, and susceptibility to chemicals"] = "Biomarkers",
  ["biomarkers and applications"] = "Biomark Appl",
  ["biomarkers and genes"] = "Biomark Genes",
  ["biomarkers and genomic medicine"] = "Biomarkers Genomic Med.",
  ["biomarkers in cancer"] = "Biomark Cancer",
  ["biomarkers in medicine"] = "Biomark Med",
  ["biomarkers in neuropsychiatry"] = "Biomark Neuropsychiatry",
  ["biomarkers journal"] = "Biomark J",
  ["biomass & bioenergy"] = "Biomass Bioenergy",
  ["biomass and bioenergy"] = "Biomass Bioenergy",
  ["biomass conversion and biorefinery"] = "Biomass Convers. Biorefin.",
  ["biomaterial investigations in dentistry"] = "Biomater Investig Dent",
  ["biomaterials advances"] = "Biomater. Adv.",
  ["biomaterials and biosystems"] = "Biomater. Biosyst.",
  ["biomaterials artificial cells and artificial organs"] = "Biomater. Artif. Cells Artif. Organs",
  ["biomaterials medical devices and artificial organs"] = "Biomater. Med. Devices Artif. Organs",
  ["biomaterials research"] = "Biomater Res",
  ["biomaterials science"] = "Biomater Sci",
  ["biomaterials, artificial cells, and artificial organs"] = "Biomater Artif Cells Artif Organs",
  ["biomaterials, artificial cells, and immobilization biotechnology"] = "Biomater. Artif. Cells. Immobilization Biotechnol.",
  ["biomaterials, artificial cells, and immobilization biotechnology : official journal of the international society for artificial cells and immobilization biotechnology"] = "Biomater Artif Cells Immobilization Biotechnol",
  ["biomaterials, medical devices and artificial organs"] = "Biomater. Med. Devices Artif. Organs",
  ["biomaterials, medical devices, and artificial organs"] = "Biomater Med Devices Artif Organs",
  ["biomath (sofia, bulgaria)"] = "Biomath (Sofia)",
  ["biomath. international journal on mathematical methods and models in biosciences"] = "BIOMATH",
  ["biomechanics and modeling in mechanobiology"] = "Biomech Model Mechanobiol",
  ["biomed research international"] = "Biomed Res Int",
  ["biomedica : revista del instituto nacional de salud"] = "Biomedica",
  ["biomedica biochimica acta"] = "Biomed. Biochim. Acta",
  ["biomedical & environmental mass spectrometry"] = "Biomed Environ Mass Spectrom",
  ["biomedical and biotechnology research journal"] = "Biomed. Biotechnol. Res. J.",
  ["biomedical and environmental mass spectrometry"] = "Biomed. Environ. Mass Spectrom.",
  ["biomedical and environmental sciences"] = "Biomed. Environ. Sci.",
  ["biomedical and environmental sciences : bes"] = "Biomed Environ Sci",
  ["biomedical bulletin"] = "Biomed Bull",
  ["biomedical bulletin / association for voluntary sterilization, inc"] = "Biomed Bull",
  ["biomedical chromatography"] = "Biomed. Chromatogr.",
  ["biomedical chromatography : bmc"] = "Biomed Chromatogr",
  ["biomedical communications"] = "Biomed Commun",
  ["biomedical data management and graph online querying : vldb 2015 workshops, big-o(q) and dmah, waikoloa, hi, usa, august 31-september 4, 2015, revised selected papers. international conference on very large data bases (41st : 2015 : waikoloa, hawaii)"] = "Biomed Data Manag Graph Online Querying (2015)",
  ["biomedical digital libraries"] = "Biomed Digit Libr",
  ["biomedical engineering"] = "Biomed. Eng.",
  ["biomedical engineering : applications, basis, and communications"] = "Biomed Eng (Singapore)",
  ["biomedical engineering advances"] = "Biomed. Eng. Adv.",
  ["biomedical engineering and computational biology"] = "Biomed Eng Comput Biol",
  ["biomedical engineering education"] = "Biomed Eng Educ",
  ["biomedical engineering letters"] = "Biomed Eng Lett",
  ["biomedical engineering online"] = "Biomed Eng Online",
  ["biomedical engineering research"] = "Biomed Eng Res",
  ["biomedical engineering: applications, basis and communications"] = "Biomed. Eng.: Appl., Basis Commun.",
  ["biomedical ethics (tubingen, germany)"] = "Biomed Ethics",
  ["biomedical ethics (tübingen, germany)"] = "Biomed Ethics",
  ["biomedical genetics and genomics"] = "Biomed Genet Genom",
  ["biomedical human kinetics"] = "Biomed Hum Kinet",
  ["biomedical image registration : second international workshop, wbir 2003, philadelphia, pa, usa, june 23-24, 2003 : revised papers. international workshop on biomedical image registration (2nd : 2003 : philadelphia, pa.)"] = "Biomed Image Registration",
  ["biomedical imaging and intervention journal"] = "Biomed Imaging Interv J",
  ["biomedical informatics insights"] = "Biomed Inform Insights",
  ["biomedical instrumentation & technology"] = "Biomed Instrum Technol",
  ["biomedical instrumentation & technology / association for the advancement of medical instrumentation"] = "Biomed Instrum Technol",
  ["biomedical instrumentation and technology"] = "Biomed. Instrum. Technol.",
  ["biomedical journal"] = "Biomed J",
  ["biomedical journal of scientific & technical research"] = "Biomed J Sci Tech Res",
  ["biomedical library bulletin (la jolla, san diego, calif.)"] = "Biomed Lib Bull",
  ["biomedical mass spectrometry"] = "Biomed Mass Spectrom",
  ["biomedical materials"] = "Biomed. Mater.",
  ["biomedical materials & devices"] = "Biomed. Mater. Devices",
  ["biomedical materials (bristol, england)"] = "Biomed Mater",
  ["biomedical materials (bristol, united kingdom)"] = "Biomed. Mater. (Bristol, U. K.)",
  ["biomedical microdevices"] = "Biomed Microdevices",
  ["biomedical news; the newspaper for the life scientist"] = "Biomed News",
  ["biomedical optics (washington, d.c.)"] = "Biomed Opt",
  ["biomedical optics express"] = "Biomed Opt Express",
  ["biomedical papers of the medical faculty of the university palacky, olomouc, czechoslovakia"] = "Biomed Pap Med Fac Univ Palacky Olomouc Czech Repub",
  ["biomedical papers of the medical faculty of the university palacký, olomouc, czechoslovakia"] = "Biomed Pap Med Fac Univ Palacky Olomouc Czech Repub",
  ["biomedical peptides, proteins & nucleic acids : structure, synthesis & biological activity"] = "Biomed Pept Proteins Nucleic Acids",
  ["biomedical peptides, proteins and nucleic acids"] = "Biomed. Pept. Proteins Nucleic Acids",
  ["biomedical physics & engineering express"] = "Biomed Phys Eng Express",
  ["biomedical reports"] = "Biomed Rep",
  ["biomedical research"] = "Biomed. Res.",
  ["biomedical research (aligarh, india)"] = "Biomed Res (Aligarh)",
  ["biomedical research (tokyo, japan)"] = "Biomed Res",
  ["biomedical research and clinical practice"] = "Biomed Res Clin Pract",
  ["biomedical research and reviews"] = "Biomed Res Rev",
  ["biomedical research on trace elements"] = "Biomed. Res. Trace Elem.",
  ["biomedical safety & standards"] = "Biomed. Saf. Stand.",
  ["biomedical science"] = "Biomed Sci",
  ["biomedical science and technology"] = "Biomed Sci Technol",
  ["biomedical sciences instrumentation"] = "Biomed Sci Instrum",
  ["biomedical signal processing and control"] = "Biomed. Signal Process. Control",
  ["biomedical spectroscopy and imaging"] = "Biomed Spectrosc Imaging",
  ["biomedical technology"] = "Biomed. Technol.",
  ["biomedicine & pharmacotherapy"] = "Biomed. Pharmacother.",
  ["biomedicine & pharmacotherapy = biomédecine & pharmacothérapie"] = "Biomed Pharmacother",
  ["biomedicine / [publiee pour l’a.a.i.c.i.g.]"] = "Biomedicine",
  ["biomedicine / [publiée pour l'a.a.i.c.i.g.]"] = "Biomedicine",
  ["biomedicine and pharmacotherapy"] = "Biomed. Pharmacother.",
  ["biomedicine hub"] = "Biomed Hub",
  ["biomeditsinskaia khimiia"] = "Biomed Khim",
  ["biomeditsinskaya khimiya"] = "Biomed. Khim.",
  ["biomedit︠s︡inskai︠a︡ khimii︠a︡"] = "Biomed Khim",
  ["biomedizinisch technik"] = "Biomed. Tech.",
  ["biomedizinische technik"] = "Biomed. Tech.",
  ["biomedizinische technik (biomedical engineering)"] = "Biomed. Tech. (Berl)",
  ["biomedizinische technik. biomedical engineering"] = "Biomed Tech (Berl)",
  ["biometals : an international journal on the role of metal ions in biology, biochemistry, and medicine"] = "Biometals",
  ["biometric technology today"] = "Biom. Technol. Today",
  ["biometrical journal"] = "Biom. J.",
  ["biometrical journal. biometrische zeitschrift"] = "Biom J",
  ["biometrics & biostatistics international journal"] = "Biom Biostat Int J",
  ["biometrics. journal of the international biometric society"] = "Biometrics",
  ["biometrie humaine"] = "Biom Hum",
  ["biometrie und informatik in der biologie und medizin"] = "Biom. Inform. Biol. Med.",
  ["biometrische zeitschrift"] = "Biom Z",
  ["biomédica : revista del instituto nacional de salud"] = "Biomedica",
  ["biométrie humaine"] = "Biom Hum",
  ["biomimetic intelligence and robotics"] = "Biomimetic Intell. Rob.",
  ["biomimetics (basel, switzerland)"] = "Biomimetics (Basel)",
  ["biomolecular concepts"] = "Biomol Concepts",
  ["biomolecular detection and quantification"] = "Biomol. Detect. Quantif.",
  ["biomolecular engineering"] = "Biomol Eng",
  ["biomolecular nmr assignments"] = "Biomol NMR Assign",
  ["biomolecules & therapeutics"] = "Biomol Ther (Seoul)",
  ["biooptics world"] = "BioOpt World",
  ["bioorganic & medicinal chemistry"] = "Bioorg Med Chem",
  ["bioorganic & medicinal chemistry letters"] = "Bioorg Med Chem Lett",
  ["bioorganic and medicinal chemistry"] = "Bioorg. Med. Chem.",
  ["bioorganic and medicinal chemistry letters"] = "Bioorg. Med. Chem. Lett.",
  ["bioorganic chemistry"] = "Bioorg Chem",
  ["bioorganic chemistry and applications"] = "Bioorg. Chem. App.",
  ["bioorganicheskaia khimiia"] = "Bioorg Khim",
  ["bioorganicheskaya khimiya (russian journal of bioorganic chemistry)"] = "Bioorg. Khim",
  ["biopesticides international"] = "Biopestic Int",
  ["biopharm international"] = "Biopharm Int",
  ["biopharmaceutical report"] = "Biopharm Rep",
  ["biopharmaceutics & drug disposition"] = "Biopharm Drug Dispos",
  ["biopharmaceutics and drug disposition"] = "Biopharm. Drug Dispos.",
  ["biophotonics discovery"] = "Biophotonics Discovery",
  ["biophysical chemistry"] = "Biophys Chem",
  ["biophysical economics and sustainability"] = "Biophys. Econ. Sustainability",
  ["biophysical journal"] = "Biophys J",
  ["biophysical reports"] = "Biophys. Rep.",
  ["biophysical reviews"] = "Biophys Rev",
  ["biophysical reviews and letters"] = "Biophys. Rev. Lett.",
  ["biophysicist (rockville, md.)"] = "Biophysicist (Rockv)",
  ["biophysics (english translation)"] = "Biophysics (Engl. Transl.)",
  ["biophysics (nagoya-shi, japan)"] = "Biophysics (Nagoya-shi)",
  ["biophysics and physicobiology"] = "Biophys Physicobiol",
  ["biophysics of structure and mechanism"] = "Biophys Struct Mech",
  ["biophysics reports"] = "Biophys Rep",
  ["biophysics reviews"] = "Biophys Rev",
  ["biophysics reviews and letters"] = "Biophys. Rev. Lett.",
  ["biopolimery i kletka"] = "Biopolim Kletka",
  ["biopolymers & cell"] = "Biopolym Cell",
  ["biopolymers symposia"] = "Biopolym Symp",
  ["biopreservation and biobanking"] = "Biopreserv Biobank",
  ["bioprinting (amsterdam, netherlands)"] = "Bioprinting",
  ["bioprocess and biosystems engineering"] = "Bioprocess Biosyst Eng",
  ["bioprocess engineering"] = "Bioprocess. Eng.",
  ["bioprocess international"] = "Bioprocess Int.",
  ["bioprocess technology"] = "Bioprocess Technol",
  ["bioprocessing journal"] = "Bioprocess. J.",
  ["biopsychosocial medicine"] = "Biopsychosoc Med",
  ["bioquímica y patología clínica : bypc : revista de la asociación bioquímica argentina"] = "Bioquim Patol Clin",
  ["bioremediation journal"] = "Biorem. J.",
  ["bioresearch open access"] = "Biores Open Access",
  ["bioresource technology"] = "Bioresour Technol",
  ["bioresource technology reports"] = "Bioresour Technol Rep",
  ["bioresources and bioprocessing"] = "Bioresour Bioprocess",
  ["biorheology. supplement"] = "Biorheology. Suppl.",
  ["biorheology. supplement : the official journal of the international society of biorheology"] = "Biorheology Suppl",
  ["biorisk : biodiversity & ecosystem risk assessment"] = "BioRisk",
  ["biorxiv : the preprint server for biology"] = "bioRxiv",
  ["bios (leverkusen, germany)"] = "Bios (Leverk)",
  ["biosafety and health"] = "Biosaf Health",
  ["bioscience and bioengineering (boston, mass.)"] = "Biosci Bioeng",
  ["bioscience and microflora"] = "Biosci Microflora",
  ["bioscience biotechnology and biochemistry"] = "Biosci. Biotechnol., Biochem.",
  ["bioscience biotechnology research communications"] = "Biosci Biotechnol Res Commun",
  ["bioscience horizons"] = "Biosci. Horiz.",
  ["bioscience hypotheses"] = "Biosci Hypotheses",
  ["bioscience journal"] = "Biosci. J.",
  ["bioscience journal : bj"] = "Biosci J",
  ["bioscience of microbiota, food and health"] = "Biosci Microbiota Food Health",
  ["bioscience reports"] = "Biosci Rep",
  ["bioscience trends"] = "Biosci Trends",
  ["bioscience, biotechnology, and biochemistry"] = "Biosci Biotechnol Biochem",
  ["biosciences, biotechnology research asia"] = "Biosci. Biotechnol. Res. Asia",
  ["bioscientifica proceedings"] = "Biosci Proc",
  ["biosecurity and bioterrorism"] = "Biosecur. Bioterror.",
  ["biosecurity and bioterrorism : biodefense strategy, practice, and science"] = "Biosecur Bioterror",
  ["biosensors & bioelectronics"] = "Biosens Bioelectron",
  ["biosensors & bioelectronics: x"] = "Biosens Bioelectron X",
  ["biosensors and bioelectronics"] = "Biosens. Bioelectron.",
  ["biosensors and bioelectronics: x"] = "Biosens. Bioelectron.: X",
  ["biosimilars (auckland, new zealand)"] = "Biosimilars",
  ["biospektrum : zeitschrift der gesellschaft für biologishe chemie (gbch) und der vereinigung für allgemeine und angewandte mikrobiologie (vaam)"] = "Biospektrum (Heidelb)",
  ["biostatistics & epidemiology"] = "Biostat Epidemiol",
  ["biostatistics (oxford, england)"] = "Biostatistics",
  ["biostatistics and biometrics open access journal"] = "Biostat Biom Open Access J",
  ["biostatistics, bioinformatics and biomathematics"] = "Biostat Bioinforma Biomath",
  ["biosurface and biotribology"] = "Biosurf Biotribol",
  ["biosystems & biorobotics"] = "Biosyst. Biorobot.",
  ["biosystems diversity"] = "Biosyst Divers",
  ["biosystems engineering"] = "Biosyst Eng",
  ["biota colombiana"] = "Biota Colomb",
  ["biota neotropica"] = "Biota Neotrop",
  ["biotechnic & histochemistry : official publication of the biological stain commission"] = "Biotech Histochem",
  ["biotechnic and histochemistry"] = "Biotech. Histochem.",
  ["biotechniques. rapid dispatches"] = "Biotech Rapid Dispatches",
  ["biotechnologia acta"] = "Biotechnol Acta",
  ["biotechnological research"] = "Biotechnol Res",
  ["biotechnologie, agronomie, societe et environnement"] = "Biotechnol. Agron. Soc. Environ.",
  ["biotechnology & genetic engineering reviews"] = "Biotechnol Genet Eng Rev",
  ["biotechnology (faisalābād, pakistan)"] = "Biotechnology (Faisalabad)",
  ["biotechnology (rajkot, india)"] = "BioTechnology (Rajkot, India)",
  ["biotechnology (reading, mass.)"] = "Biotechnology",
  ["biotechnology : an indian journal"] = "Biotechnol (Rajkot)",
  ["biotechnology advances"] = "Biotechnol Adv",
  ["biotechnology and applied biochemistry"] = "Biotechnol Appl Biochem",
  ["biotechnology and bioengineering"] = "Biotechnol Bioeng",
  ["biotechnology and bioengineering symposium"] = "Biotechnol Bioeng Symp",
  ["biotechnology and bioprocess engineering"] = "Biotechnol. Bioprocess Eng.",
  ["biotechnology and bioprocess engineering : bbe"] = "Biotechnol Bioprocess Eng",
  ["biotechnology and biotechnological equipment"] = "Biotechnol. Biotechnol. Equip.",
  ["biotechnology and development monitor"] = "Biotechnol. Dev. Monit.",
  ["biotechnology and genetic engineering reviews"] = "Biotechnol. Genet. Eng. Rev.",
  ["biotechnology annual review"] = "Biotechnol Annu Rev",
  ["biotechnology et alia"] = "Biotechnol Alia",
  ["biotechnology focus"] = "Biotechnol Focus",
  ["biotechnology for biofuels"] = "Biotechnol Biofuels",
  ["biotechnology for biofuels and bioproducts"] = "Biotechnol. Biofuels Bioprod.",
  ["biotechnology healthcare"] = "Biotechnol Healthc",
  ["biotechnology journal"] = "Biotechnol J",
  ["biotechnology law report"] = "Biotechnol Law Rep",
  ["biotechnology letters"] = "Biotechnol Lett",
  ["biotechnology notes"] = "Biotechnol. Notes",
  ["biotechnology progress"] = "Biotechnol Prog",
  ["biotechnology reports"] = "Biotechnol. Rep,",
  ["biotechnology reports (amsterdam, netherlands)"] = "Biotechnol Rep (Amst)",
  ["biotechnology research and innovation"] = "Biotechnol. Res. Innovation",
  ["biotechnology research international"] = "Biotechnol Res Int",
  ["biotechnology techniques"] = "Biotechnol. Tech.",
  ["biotechnology therapeutics"] = "Biotechnol Ther",
  ["biotechnology, biotechnological equipment"] = "Biotechnol Biotechnol Equip",
  ["biotecnologia ciência & desenvolvimento"] = "Biotecnolog Cienc Desenvolv",
  ["biotecnología aplicada : revista de la sociedad iberolatinoamericana para investigaciones sobre interferón y biotecnología en salud"] = "Biotecnol Apl",
  ["biotecnología vegetal"] = "Biotecnol Veg",
  ["biotelemetry and patient monitoring"] = "Biotelem Patient Monit",
  ["biotherapy (dordrecht, netherlands)"] = "Biotherapy",
  ["biotribology (oxford)"] = "Biotribology (Oxf)",
  ["bioworld : the business of biotechnology"] = "Bioworld",
  ["bipolar disorders"] = "Bipolar Disord",
  ["bird conservation international"] = "Bird Conserv Int",
  ["bird study"] = "Bird Study",
  ["bird study : the journal of the british trust for ornithology"] = "Bird Study",
  ["birkhauser advanced texts: basler lehrbucher"] = "Birkhauser Adv. Texts Basler Lehrbucher",
  ["birkhauser skripten"] = "Birkhauser Skr.",
  ["birperht publication"] = "BIRPERHT Publ",
  ["birth (berkeley, calif.)"] = "Birth",
  ["birth and the family journal"] = "Birth Fam J",
  ["birth defects original article series"] = "Birth Defects Orig Artic Ser",
  ["birth defects research"] = "Birth Defects Res",
  ["birth defects research, part a: clinical and molecular teratology"] = "Birth Defects Res., Part A",
  ["birth defects research, part b: developmental and reproductive toxicology"] = "Birth Defects Res., Part B",
  ["birth defects research, part c: embryo today–reviews"] = "Birth Defects Res., Part C",
  ["birth defects research. part a, clinical and molecular teratology"] = "Birth Defects Res A Clin Mol Teratol",
  ["birth defects research. part b, developmental and reproductive toxicology"] = "Birth Defects Res B Dev Reprod Toxicol",
  ["birth defects research. part c, embryo today"] = "Birth Defects Res. C. Embryo Today",
  ["birth defects research. part c, embryo today : reviews"] = "Birth Defects Res C Embryo Today",
  ["birth gazette"] = "Birth Gaz.",
  ["birth psychology bulletin"] = "Birth Psychol Bull",
  ["bit numerical mathematics"] = "BIT Numer. Math.",
  ["bit. numerical mathematics"] = "BIT Numer Math",
  ["biuletyn historii sztuki"] = "Biul Hist Sztuki",
  ["biuletyn igs"] = "Biul IGS",
  ["biuletyn igs / szkola glowna handlowa, instytut gospodarstwa spolecznego"] = "Biul IGS",
  ["biuletyn instytutu medycyny morskiej w gdansku"] = "Biul. Inst. Med. Morsk. Gdansk.",
  ["biuletyn instytutu medycyny morskiej w gdańsku"] = "Biul Inst Med Morsk Gdansk",
  ["biuletyn instytutu roślin leczniczych. instytut przemysłu zielarskiego (poznań, poland)"] = "Biul Inst Rosl Lecz Inst Przem Zielar Pozn Pol",
  ["biuletyn państwowego instytutu medycyny morskiej i tropikalnej w gdańsku. biulleten' gosudarstvennogo instituta morskoĭ i tropicheskoĭ meditsiny v gdan'ske. bulletin of the state institute of marine and tropical medicine in gdańsk, poland"] = "Biul Panstw Inst Med Morsk Trop J W Gdansku",
  ["biuletyn zydowskiego instytutu historycznego"] = "Biul Zydowskiego Inst Hist Pol",
  ["biuletyn żydowskiego instytutu historycznego"] = "Biul Zydowskiego Inst Hist Pol",
  ["biuletyn-biblioteki jagiellonskiej"] = "Biul Bibl Jagiellon",
  ["biuletyn-biblioteki jagiellońskiej"] = "Biul Bibl Jagiellon",
  ["biulleten eksperimentalnoi biologii i meditsiny"] = "Biull. Eksp. Biol. Med.",
  ["biulleten vsesoiuznogo kardiologicheskogo nauchnogo tsentra amn sssr"] = "Biull. Vsesoiuznogo Kardiol. Nauchn. Tsentra AMN SSSR",
  ["biulleten' eksperimental'noĭ biologii i meditsiny"] = "Biull Eksp Biol Med",
  ["biulleten' vsesoiuznogo kardiologicheskogo nauchnogo tsentra amn sssr"] = "Biull Vsesoiuznogo Kardiol Nauchn Tsentra AMN SSSR",
  ["biulleten'. russian s.f.s.r. ministerstvo zdravookhraneniia. uchenyĭ meditsinskiĭ sovet"] = "Biulleten Russ SFSR Minist Zdravookhraneniia Uchenyi Meditsinskii sov",
  ["biulleten’ eksperimental’noi biologii i meditsiny"] = "Biull Eksp Biol Med",
  ["biulleten’ moskovskogo obshchestva ispytatelei prirody. otdel biologicheskii"] = "Biull Mosk Ova Ispyt Prir (Biol)",
  ["biulleten’ vsesoiuznogo kardiologicheskogo nauchnogo tsentra amn sssr"] = "Biull Vsesoiuznogo Kardiol Nauchn Tsentra AMN SSSR",
  ["biulleten’. russian s.f.s.r. ministerstvo zdravookhraneniia. uchenyi meditsinskii sovet"] = "Biulleten Russ SFSR Minist Zdravookhraneniia Uchenyi Meditsinskii sov",
  ["biyutiknuluzhī-i kishāvarzī"] = "Biyutiknuluzhii Kishavarzi",
  ["bìologìčnì studìï"] = "Biol Stud",
  ["bi︠u︡lletenʹ dalʹnevostochnogo malakologicheskogo obshchestva"] = "Bull Dalnevost Malakol Obs",
  ["bi︠u︡lletenʹ moskovskogo obshchestva ispytateleĭ prirody. otdel biologicheskiĭ"] = "Biull Mosk Ova Ispyt Prir (Biol)",
  ["bi︠u︡lletenʹ sibirskogo otdelenii︠a︡ rossiĭskoĭ akademii medit︠s︡inskikh nauk"] = "Bull Sib Otd Ross Akad Med Nauk",
  ["bi︠u︡lletenʹ sibirskoĭ medit︠s︡iny"] = "Biulleten Sib Meditsiny",
  ["bi︠u︡lletenʹ vostochno-sibirskogo nauchnogo t︠s︡entra"] = "Bull Vost Sib Naucn Sent",
  ["bja education"] = "BJA Educ",
  ["bjgp open"] = "BJGP Open",
  ["bjhs themes"] = "BJHS Themes",
  ["bjog : an international journal of obstetrics and gynaecology"] = "BJOG",
  ["bjp review of books"] = "BJP Rev Books",
  ["bjpsych advances"] = "BJPsych Adv",
  ["bjpsych bulletin"] = "BJPsych Bull",
  ["bjpsych international"] = "BJPsych Int",
  ["bjpsych open"] = "BJPsych Open",
  ["bjr case reports"] = "BJR Case Rep",
  ["bjr open"] = "BJR Open",
  ["bjr supplement"] = "BJR Suppl",
  ["bjr supplement / bir"] = "BJR Suppl",
  ["bjs open"] = "BJS Open",
  ["bju international"] = "BJU Int",
  ["bjui compass"] = "BJUI Compass",
  ["black aging"] = "Black Aging",
  ["bladder (san francisco, calif.)"] = "Bladder (San Franc)",
  ["bladder cancer (amsterdam, netherlands)"] = "Bladder Cancer",
  ["blake studies"] = "Blake Stud",
  ["blasting and fragmentation journal"] = "Blasting Fragm. J.",
  ["blatter fur deutsche und internationale politik"] = "Bl Dtsch Int Polit",
  ["blatter fur wurttembergische kirchengeschichte"] = "BL Wueritemb Kirchengesch",
  ["blatter fur zahnheilkunde"] = "Bl. Zahnheilkd.",
  ["blatter fur zahnheilkunde. bulletin dentaire"] = "Bl Zahnheilkd",
  ["blaue, die"] = "Blaue",
  ["blätter für deutsche landesgeschichte"] = "Bl Dtsch Landesgesch",
  ["blätter für deutsche und internationale politik"] = "Bl Dtsch Int Polit",
  ["blätter für württembergische kirchengeschichte"] = "BL Wueritemb Kirchengesch",
  ["blätter für zahnheilkunde. bulletin dentaire"] = "Bl Zahnheilkd",
  ["bleter far geshikhte"] = "BL Gesz",
  ["bleṭer far geshikhṭe"] = "BL Gesz",
  ["blockchain in healthcare today"] = "Blockchain Healthc Today",
  ["blockchain: research and applications"] = "Blockchain: Res. Appl.",
  ["blood advances"] = "Blood Adv",
  ["blood and lymphatic cancer : targets and therapy"] = "Blood Lymphat Cancer",
  ["blood cancer discovery"] = "Blood Cancer Discov",
  ["blood cancer journal"] = "Blood Cancer J",
  ["blood cells"] = "Blood Cells",
  ["blood cells, molecules & diseases"] = "Blood Cells Mol Dis",
  ["blood cells, molecules, and diseases"] = "Blood Cells Mol. Dis.",
  ["blood coagulation & fibrinolysis : an international journal in haemostasis and thrombosis"] = "Blood Coagul Fibrinolysis",
  ["blood coagulation and fibrinolysis"] = "Blood Coagul. Fibrinolysis",
  ["blood donor. bloedskenker"] = "Blood Donor",
  ["blood immunology & cellular therapy"] = "Blood Immunol. Cell. Ther.",
  ["blood pressure"] = "Blood Press",
  ["blood pressure monitoring"] = "Blood Press Monit",
  ["blood pressure. supplement"] = "Blood Press Suppl",
  ["blood purification"] = "Blood Purif",
  ["blood red cells & iron"] = "Blood Red Cells Iron",
  ["blood research"] = "Blood Res",
  ["blood reviews"] = "Blood Rev",
  ["blood science"] = "Blood Sci.",
  ["blood transfusion = trasfusione del sangue"] = "Blood Transfus",
  ["blood vessels"] = "Blood Vessels",
  ["bloomsbury critical introductions to contemporary epistemology"] = "Bloomsbury Crit. Introd. Contemp. Epistemol.",
  ["bloomsbury geographer"] = "Bloomsbury Geogr",
  ["bloomsbury studies in the aristotelian tradition"] = "Bloomsbury Stud. Aristot. Tradit.",
  ["blue cross association research series"] = "Blue Cross Assoc Res Ser",
  ["blue jay"] = "Blue Jay",
  ["blue-green systems"] = "Blue-Green Syst.",
  ["blumea : tijdschrift voor de systematiek en de geografie der planten : a journal of plant taxonomy and plant geography"] = "Blumea",
  ["blätter der vereinigung pro sihltal"] = "Bl. Ver. Pro Sihltal",
  ["blätter für münzfreunde und münzforschung"] = "BlMueFreundeF",
  ["bmb reports"] = "BMB Rep",
  ["bmc anesthesiology"] = "BMC Anesthesiol",
  ["bmc biochemistry"] = "BMC Biochem",
  ["bmc bioinformatics"] = "BMC Bioinformatics",
  ["bmc biology"] = "BMC Biol",
  ["bmc biomedical engineering"] = "BMC Biomed Eng",
  ["bmc biophysics"] = "BMC Biophys",
  ["bmc biotechnology"] = "BMC Biotechnol",
  ["bmc blood disorders"] = "BMC Blood Disord",
  ["bmc cancer"] = "BMC Cancer",
  ["bmc cardiovascular disorders"] = "BMC Cardiovasc Disord",
  ["bmc cell biology"] = "BMC Cell Biol",
  ["bmc chemical biology"] = "BMC Chem Biol",
  ["bmc chemical engineering"] = "BMC Chem. Eng.",
  ["bmc chemistry"] = "BMC Chem",
  ["bmc clinical pathology"] = "BMC Clin Pathol",
  ["bmc clinical pharmacology"] = "BMC Clin Pharmacol",
  ["bmc complementary and alternative medicine"] = "BMC Complement Altern Med",
  ["bmc complementary medicine and therapies"] = "BMC Complement Med Ther",
  ["bmc dermatology"] = "BMC Dermatol",
  ["bmc developmental biology"] = "BMC Dev Biol",
  ["bmc digital health"] = "BMC Digital Health",
  ["bmc ear, nose, and throat disorders"] = "BMC Ear Nose Throat Disord",
  ["bmc ecology"] = "BMC Ecol",
  ["bmc ecology and evolution"] = "BMC Ecol Evol",
  ["bmc emergency medicine"] = "BMC Emerg Med",
  ["bmc endocrine disorders"] = "BMC Endocr Disord",
  ["bmc energy"] = "BMC Energy",
  ["bmc evolutionary biology"] = "BMC Evol Biol",
  ["bmc family practice"] = "BMC Fam Pract",
  ["bmc gastroenterology"] = "BMC Gastroenterol",
  ["bmc genetics"] = "BMC Genet",
  ["bmc genomic data"] = "BMC Genom Data",
  ["bmc genomics"] = "BMC Genomics",
  ["bmc geriatrics"] = "BMC Geriatr",
  ["bmc health services research"] = "BMC Health Serv Res",
  ["bmc hematology"] = "BMC Hematol",
  ["bmc immunology"] = "BMC Immunol",
  ["bmc infectious diseases"] = "BMC Infect Dis",
  ["bmc international health and human rights"] = "BMC Int Health Hum Rights",
  ["bmc materials"] = "BMC Mater",
  ["bmc mechanical engineering"] = "BMC Mech. Eng.",
  ["bmc medical education"] = "BMC Med Educ",
  ["bmc medical ethics"] = "BMC Med Ethics",
  ["bmc medical genetics"] = "BMC Med Genet",
  ["bmc medical genomics"] = "BMC Med Genomics",
  ["bmc medical imaging"] = "BMC Med Imaging",
  ["bmc medical informatics and decision making"] = "BMC Med Inform Decis Mak",
  ["bmc medical physics"] = "BMC Med Phys",
  ["bmc medical research methodology"] = "BMC Med Res Methodol",
  ["bmc medicine"] = "BMC Med",
  ["bmc methods"] = "BMC Methods",
  ["bmc microbiology"] = "BMC Microbiol",
  ["bmc molecular and cell biology"] = "BMC Mol Cell Biol",
  ["bmc molecular biology"] = "BMC Mol Biol",
  ["bmc musculoskeletal disorders"] = "BMC Musculoskelet Disord",
  ["bmc nephrology"] = "BMC Nephrol",
  ["bmc neurology"] = "BMC Neurol",
  ["bmc neuroscience"] = "BMC Neurosci",
  ["bmc nuclear medicine"] = "BMC Nucl Med",
  ["bmc nursing"] = "BMC Nurs",
  ["bmc nutrition"] = "BMC Nutr",
  ["bmc obesity"] = "BMC Obes",
  ["bmc ophthalmology"] = "BMC Ophthalmol",
  ["bmc oral health"] = "BMC Oral Health",
  ["bmc palliative care"] = "BMC Palliat Care",
  ["bmc pediatrics"] = "BMC Pediatr",
  ["bmc pharmacology"] = "BMC Pharmacol",
  ["bmc pharmacology & toxicology"] = "BMC Pharmacol Toxicol",
  ["bmc pharmacology and toxicology"] = "BMC Pharmacol. Toxicol.",
  ["bmc physiology"] = "BMC Physiol",
  ["bmc plant biology"] = "BMC Plant Biol",
  ["bmc pregnancy and childbirth"] = "BMC Pregnancy Childbirth",
  ["bmc proceedings"] = "BMC Proc",
  ["bmc psychiatry"] = "BMC Psychiatry",
  ["bmc psychology"] = "BMC Psychol",
  ["bmc public health"] = "BMC Public Health",
  ["bmc pulmonary medicine"] = "BMC Pulm Med",
  ["bmc research notes"] = "BMC Res Notes",
  ["bmc rheumatology"] = "BMC Rheumatol",
  ["bmc sports science, medicine & rehabilitation"] = "BMC Sports Sci Med Rehabil",
  ["bmc sports science, medicine and rehabilitation"] = "BMC Sports Sci. Med. Rehabil.",
  ["bmc structural biology"] = "BMC Struct Biol",
  ["bmc surgery"] = "BMC Surg",
  ["bmc systems biology"] = "BMC Syst Biol",
  ["bmc urology"] = "BMC Urol",
  ["bmc veterinary research"] = "BMC Vet Res",
  ["bmc women's health"] = "BMC Womens Health",
  ["bmc women’s health"] = "BMC Womens Health",
  ["bmc zoology"] = "BMC Zool",
  ["bme = bio medical engineering"] = "BME",
  ["bme frontiers"] = "BME Front",
  ["bmj (clinical research ed.)"] = "BMJ",
  ["bmj (international edition)"] = "BMJ (Int Ed)",
  ["bmj case reports"] = "BMJ Case Rep",
  ["bmj clinical evidence"] = "BMJ Clin Evid",
  ["bmj connections clinical genetics and genomics"] = "BMJ Connect. Clin Genet. Genomics",
  ["bmj connections oncology"] = "BMJ Connect. Oncol.",
  ["bmj digital health & ai"] = "BMJ Digital Health AI",
  ["bmj evidence-based medicine"] = "BMJ Evid Based Med",
  ["bmj global health"] = "BMJ Glob Health",
  ["bmj health & care informatics"] = "BMJ Health Care Inform",
  ["bmj immunology"] = "BMJ Immunol.",
  ["bmj innovations"] = "BMJ Innov",
  ["bmj medicine"] = "BMJ Med.",
  ["bmj military health"] = "BMJ Mil Health",
  ["bmj neurology open"] = "BMJ Neurol Open",
  ["bmj nutrition, prevention & health"] = "BMJ Nutr Prev Health",
  ["bmj open"] = "BMJ Open",
  ["bmj open diabetes research & care"] = "BMJ Open Diabetes Res Care",
  ["bmj open gastroenterology"] = "BMJ Open Gastroenterol",
  ["bmj open ophthalmology"] = "BMJ Open Ophthalmol",
  ["bmj open quality"] = "BMJ Open Qual",
  ["bmj open respiratory research"] = "BMJ Open Respir Res",
  ["bmj open science"] = "BMJ Open Sci.",
  ["bmj open sport & exercise medicine"] = "BMJ Open Sport Exerc Med",
  ["bmj outcomes"] = "BMJ Outcomes",
  ["bmj paediatrics open"] = "BMJ Paediatr Open",
  ["bmj quality & safety"] = "BMJ Qual Saf",
  ["bmj quality improvement reports"] = "BMJ Qual Improv Rep",
  ["bmj sexual & reproductive health"] = "BMJ Sex Reprod Health",
  ["bmj simulation & technology enhanced learning"] = "BMJ Simul Technol Enhanc Learn",
  ["bmj supportive & palliative care"] = "BMJ Support Palliat Care",
  ["bmj surgery, interventions, & health technologies"] = "BMJ Surg Interv Health Technol",
  ["bmj: british medical journal"] = "BMJ",
  ["bmq: the boston medical quarterly"] = "BMQ.",
  ["bmq; the boston medical quarterly"] = "BMQ",
  ["bmr journal of geology and geophysics"] = "BMR J. Aust. Geol. Geophys.",
  ["bmvc : proceedings of the british machine vision conference. british machine vision conference"] = "BMVC",
  ["bni quarterly"] = "BNI Q",
  ["bnwl [reports]. u.s. atomic energy commission"] = "BNWL Rep",
  ["bnwl reports"] = "BNWL Rep.",
  ["bo pu xue za zhi = chinese journal of microwave & radio-frequency spectroscopy"] = "Bo Pu Xue Za Zhi",
  ["boardroom reports"] = "Boardr Rep",
  ["bocconi & springer series"] = "Bocconi Springer Ser.",
  ["bochumer studien zur philosophie"] = "Boch. Stud. Philos.",
  ["bodenkultur, die"] = "Bodenkultur",
  ["bodenkunde und pflanzenernahrung"] = "Bodenkd. Pflanzenernahr.",
  ["body & society"] = "Body Soc",
  ["body image"] = "Body Image",
  ["body, movement, and dance in psychotherapy"] = "Body Mov Dance Psychother",
  ["boethius: texte und abhandlungen zur geschichte der mathematik und der naturwissenschaften"] = "Boethius Texte Abh. Gesch. Math. Naturwiss.",
  ["bogens verden"] = "Bogens Verden",
  ["boğazköy-hattuša. ergebnisse der ausgrabungen"] = "Bogazkoey-Hattusa",
  ["bohemia (munich, germany)"] = "Bohemia Jahrb Coll Carolinum",
  ["bois et forêts des tropiques"] = "Bois For. Trop.",
  ["boj o zdravie"] = "Boj Zdr",
  ["boj proti tuberkulose"] = "Boj Tuberk",
  ["bold : quarterly journal of the international institute on aging (united nations - malta)"] = "Bold",
  ["boleettino d'informazioni della consociazione nazionale infermiere professionali e assistenti sanitarie visitatrici"] = "Boll Inf Consoc Naz (Rome)",
  ["boleettino d’informazioni della consociazione nazionale infermiere professionali e assistenti sanitarie visitatrici"] = "Boll Inf Consoc Naz (Rome)",
  ["bolet'in de la agrupación odontológica de la zona central de la capital federal"] = "Bol Agrup Odontol Zona Cent Cap Fed B Aires",
  ["boletim (academia nacional de medicina (brazil))"] = "Bol Acad Nac Med",
  ["boletim - faculdade de odontologia de piracicaba"] = "Bol Fac Odontol Piracicaba",
  ["boletim bibliográfico"] = "Bol Bibliogr (Sao Paulo)",
  ["boletim clinico dos hospitais civis de lisboa"] = "Bol Clin Hosp Civis Lisb",
  ["boletim clínico dos hospitais civis de lisboa"] = "Bol Clin Hosp Civis Lisb",
  ["boletim clínico e de estatística. lisbon. hospital colonial"] = "Bol Clin Estat",
  ["boletim da academia nacional de medicina"] = "Bol Acad Nac Med",
  ["boletim da divisao nacional de dermatologia sanitaria"] = "Bol. Div. Nac. Dermatol. Sanit.",
  ["boletim da divisão nacional de dermatologia sanitária"] = "Bol Div Nac Dermatol Sanit",
  ["boletim da equipe de odontologia sanitaria"] = "Bol Equipe Odontol Sanit",
  ["boletim da faculdade de farmacia e odontologia de ribeirao preto"] = "Bol. Fac. Farm. Odontol. Ribeirao Preto",
  ["boletim da faculdade de farmácia e odontologia de ribeirão prêto"] = "Bol Fac Farm Odontol Ribeirao Preto",
  ["boletim da sociedade brasileira de matemática"] = "Bol. Soc. Brasil. Mat. (N.S.)",
  ["boletim da sociedade de geografia de lisboa"] = "Bol Soc Geogr Lisb",
  ["boletim da sociedade paranaense de matematica. 3rd serie"] = "Bol. Soc. Parana. Mat. (3)",
  ["boletim da sociedade portuguesa de hemorreologia e microcirculação"] = "Bol Soc Port Hemorreol Microcirc",
  ["boletim da sociedade portuguesa de matematica"] = "Bol. Soc. Port. Mat.",
  ["boletim da sociedade portuguesa de matemática"] = "Bol. Soc. Port. Mat.",
  ["boletim de biotecnologia"] = "Bol Biotecnol",
  ["boletim de ciencias geodesicas"] = "Bol. Cienc. Geod.",
  ["boletim de dentistica operatoria"] = "Bol. Dent. Oper.",
  ["boletim de dentística operatória"] = "Bol Dent Oper",
  ["boletim de geografia teoretica"] = "Bol Geogr Teor",
  ["boletim de geografia teorética"] = "Bol Geogr Teor",
  ["boletim de higiene mental"] = "Bol Hig Ment",
  ["boletim de informacao socio-economica / gabinete de estudos economicos, ministerio do plano e cooperacao internacional"] = "Bol Inf Socio Econ",
  ["boletim de informação sócio-económica"] = "Bol Inf Socio Econ",
  ["boletim de materiais dentarios"] = "Bol Mat Dent",
  ["boletim de materiais dentários"] = "Bol Mat Dent",
  ["boletim de materias dentarios"] = "Bol. Mat. Dent.",
  ["boletim de pesquisa da embrapa soja"] = "Bol Pesqui (Embrapa Soja)",
  ["boletim demografico / secretaria de planejamento da presidencia da republica, fundacao instituto brasileiro de geografia e estatistica, diretoria de populacao e social, superintendencia de estudos de populacao e sociais"] = "Bol Demogr",
  ["boletim demográfico"] = "Bol Demogr",
  ["boletim do c.b.r"] = "Bol Cent Biol Reprod",
  ["boletim do c.b.r. / ministerio da educacao e cultura, universidade federal de juiz de fora, centro de biologia da reproducao"] = "Bol Cent Biol Reprod",
  ["boletim do centro de estudos, hospital dos servidores do estado"] = "Bol Cent Estud Hosp Serv Estado",
  ["boletim do instituto de angola"] = "Bol Inst Angola",
  ["boletim do instituto de botânica"] = "Bol Inst Bot",
  ["boletim do instituto de pesca. instituto de pesca (são paulo, brazil : state)"] = "Bol Inst Pesca",
  ["boletim do instituto de puericultura"] = "Bol Inst Pueric",
  ["boletim do instituto de puericultura e pediatria martagao gesteira"] = "Bol. Inst. Pueric. Martagao Gesteira",
  ["boletim do instituto de puericultura e pediatria martagão gesteira"] = "Bol Inst Pueric Martagao Gesteira",
  ["boletim do instituto superior de higiene doutor ricardo jorge"] = "Bol Inst Super Hig Doutor Ricardo Jorge",
  ["boletim do sanatorio sao lucas"] = "Bol. Sanat. Sao Lucas",
  ["boletim do sanatorio são lucas. são paulo, brazil (city) sanatorio são lucas"] = "Bol Sanat Sao Lucas",
  ["boletim do servico de odontologia sanitaria da secretaria da saude do rio grande do sul"] = "Bol Serv Odontol Sanit (Porto Alegre)",
  ["boletim epidemiologico. aids / ministerio da saude"] = "Bol Epidemiol AIDS",
  ["boletim epidemiológico. aids"] = "Bol Epidemiol AIDS",
  ["boletim geografico"] = "Bol Geogr",
  ["boletim geográfico"] = "Bol Geogr",
  ["boletim informativo do centro de documentacao e pesquisa / cedope unisinos"] = "Bol Inf Cent Doc Pesqui",
  ["boletim informativo do centro de documentação e pesquisa"] = "Bol Inf Cent Doc Pesqui",
  ["boletim sobre populacao, emprego e renda no nordeste / superintendencia do desenvolvimento do nordeste, fundacao joaquim nabuco"] = "Bol Popul Emprego Renda Nordeste",
  ["boletim sobre população, emprego e renda no nordeste"] = "Bol Popul Emprego Renda Nordeste",
  ["boletim tecnico da producao de petroleo"] = "Bol. Tec. Prod. Pet.",
  ["boletim, faculdade de odontologia de piracicaba"] = "Bol. Fac. Odontol. Piracicaba",
  ["boletim. colégio brasileiro de cirurgiões"] = "Bol Colegio Bras Cir",
  ["boletim. sao paulo, brazil (city) sanatorio sao lucas"] = "Bol Sanat Sao Lucas",
  ["boletim. são paulo, brazil (state) repartição de aguas e esgotos"] = "Bol Sao Paulo Braz State Reparticao Aguas Esgotos",
  ["boletim. sociedade brasileira de medicina veterinaria"] = "Bol Soc Bras Med Vet",
  ["boletim. sociedade de medicina e cirurgia de campinas"] = "Bol Soc Med Cir Camp",
  ["boletim. sociedade paulista de medicina veterinária"] = "Bol Soc Paul Med Vet",
  ["boletin (consejo nacional de poblacion (peru))"] = "Bol Cons Nac Poblac",
  ["boletin (instituto de estudios de poblacion y desarrollo (dominican republic))"] = "Boletin",
  ["boletin (sociedad espanola de hidrologia medica)"] = "Bol Soc Esp Hidrol Med",
  ["boletin - asociacion medica de puerto rico"] = "Bol Asoc Med P R",
  ["boletin - circulo argentino de odontologia"] = "Bol Circ Argent Odontol",
  ["boletin - colegio de profesionales de la enfermeria de puerto rico"] = "Bol Col Prof Enferm P R",
  ["boletin - colegio de profesionales de la enfermería de puerto rico"] = "Bol Col Prof Enferm P R",
  ["boletin bibliografico de la secretaria de hacienda y credito publico"] = "Bol Bibliogr Secr Hacienda Credito Publico",
  ["boletin biológico"] = "Bol Biol",
  ["boletin chileno de parasitologia"] = "Bol Chil Parasitol",
  ["boletin cultural e informativo - consejo general de colegios medicos de espana"] = "Bol Cult Inf Cons Gen Col Med Esp",
  ["boletin cultural e informativo - consejo general de colegios medicos de españa"] = "Bol Cult Inf Cons Gen Col Med Esp",
  ["boletin cultural y bibliografico"] = "Bol Cult Bibliogr",
  ["boletin de analisis demografico"] = "Bol Anal Demogr",
  ["boletin de antropologia"] = "Bol Antropol",
  ["boletin de antropologia americana"] = "Bol Antropol Am",
  ["boletin de educación física"] = "Bol Educ Fis (Santiago)",
  ["boletin de estudios latinoamericanos y del caribe"] = "Bol Estud Latinoam Caribe",
  ["boletin de estudios medicos y biologicos"] = "Bol Estud Med Biol",
  ["boletin de historia y antiguedades"] = "Bol Hist Antig",
  ["boletin de indicadores de coyuntura / republica de venezuela, presidencia de la republica, oficina central de estadistica e informatica"] = "Bol Indic Coyunt",
  ["boletin de informacion - ilustre colegio oficial de odontologos y estomatologos 2. a region (cataluna)"] = "Bol Inf Col Odontol Estomatol (Barc)",
  ["boletin de informacion dental"] = "Bol. Inf. Dent. (Madr.)",
  ["boletin de informacion, ilustre colegio oficial de odontologos y estomatologos, 2a region, cataluna"] = "Bol. Inf. Col. Odontol. Estomatol. (Barc.)",
  ["boletin de informaciones parasitarias chilenas"] = "Bol Inf Parasit Chil",
  ["boletin de información - ilustre colegio oficial de odontólogos y estomatólogos 2. a región (cataluña)"] = "Bol Inf Col Odontol Estomatol (Barc)",
  ["boletin de la aaon"] = "Bol. AAON",
  ["boletin de la academia nacional de la historia. academia nacional de la historia (argentina)"] = "Bol Acad Nac Hist (Argent)",
  ["boletin de la academia nacional de la historia. academia nacional de la historia (venezuela)"] = "Bol Acad Nac Hist (Caracas)",
  ["boletin de la asociacion argentina de odontologia para ninos"] = "Bol Asoc Argent Odontol Ninos",
  ["boletin de la asociacion de demografia historica"] = "Bol Asoc Demogr Hist",
  ["boletin de la asociacion matematica venezolana"] = "Bol. Asoc. Mat. Venez.",
  ["boletin de la asociacion medica de puerto rico"] = "Bol Asoc Med P R",
  ["boletin de la asociacion venezolana de enfermeras profesionales"] = "Bol. Asoc. Venez. Enferm. Prof.",
  ["boletin de la asociación española de entomologia"] = "Bol Asoc Esp Entomol",
  ["boletin de la biblioteca de menendez pelayo / sociedad de menendez pelayo"] = "Bol Bibl Menendez Pelayo",
  ["boletin de la direccion general de odontologia"] = "Bol. Dir. Gen. Odontol. (Santa Fe)",
  ["boletin de la liga contra el cancer. liga contra el cancer (cuba)"] = "Bol Liga Contra Cancer Havana",
  ["boletin de la oficina sanitaria panamericana"] = "Bol. Oficina Sanit. Panam.",
  ["boletin de la oficina sanitaria panamericana. english ed"] = "Bol Of Sanit Panam (Engl)",
  ["boletin de la oficina sanitaria panamericana. pan american sanitary bureau"] = "Bol Oficina Sanit Panam",
  ["boletin de la real sociedad espanola de historia natural. seccion biologica : organo del instituto de ciencias naturales jose de acosta"] = "Bol R Soc Esp Hist Nat Secc Biol",
  ["boletin de la sociedad argentina de botanica"] = "Bol. Soc. Argent. Bot.",
  ["boletin de la sociedad castellonense de cultura. sociedad castellonense de cultura"] = "Bol Soc Castell Cult",
  ["boletin de la sociedad catalana de pediatria"] = "Bol Soc Catalana Pediatr",
  ["boletin de la sociedad chilena de quimica"] = "Bol. Soc. Chil. Quim.",
  ["boletin de la sociedad dental de guatemala"] = "Bol. Soc. Dent. Guatem.",
  ["boletin de la sociedad espanola de ceramica y vidrio"] = "Bol. Soc. Esp. Ceram. Vidrio",
  ["boletin de la sociedad espanola de historia de la farmacia"] = "Bol Soc Esp Hist Farm",
  ["boletin de la sociedad espanola de historia de la medicina"] = "Bol Soc Esp Hist Med",
  ["boletin de la sociedad estomatologica argentina"] = "Bol Soc Estomatol Argent",
  ["boletin de la sociedad geologica mexicana"] = "Bol. Soc. Geol. Mex.",
  ["boletin de la sociedad matematica mexicana"] = "Bol. Soc. Mat. Mex.",
  ["boletin de la sociedad matematica mexicana. third series"] = "Bol. Soc. Mat. Mex. (3)",
  ["boletin de la sociedad mexicana de historia y filosofia de la medicina"] = "Bol Soc Mex Hist Filos Med",
  ["boletin de la sociedad quimica del peru"] = "Bol Soc Quim Peru",
  ["boletin de la sociedad valenciana de pediatria"] = "Bol Soc Valencia Pediatr",
  ["boletin de la universidad de granada. universidad de granada"] = "Bol Univ Granada",
  ["boletin de matematicas. nueva serie"] = "Bol. Mat.",
  ["boletin de odontologia"] = "Bol. Odontol. (Bogota)",
  ["boletin de poblacion"] = "Bol Poblac",
  ["boletin de protesis"] = "Bol. Protes.",
  ["boletin de psicologia / republica de cuba, ministerio de salud publica, hospital psiquiatrico de la habana"] = "Bol Psicol",
  ["boletin de sanidad militar"] = "Bol Sanid Mil",
  ["boletin del centro de estudios del siglo xviii. universidad de oviedo. centro de estudios del siglo xviii"] = "Bol Cent Estud Siglo XVIII",
  ["boletin del circulo argentina de odontologia"] = "Bol. Circ. Argent. Odontol.",
  ["boletin del hospital militar. cuba. ejercito. hospital militar dr. carlos j. finlay"] = "Bol Hosp Mil Cuba Ejercito Hosp Mil Dr Carlos J Finlay",
  ["boletin del hospital san juan de dios"] = "Bol Hosp San Juan Dios",
  ["boletin del instituto de estudios asturianos. instituto de estudios asturianos (oviedo, spain)"] = "Bol Inst Estud Asturianos",
  ["boletin del instituto de estudios medicos y biologicos, universidad nacional autonoma de mexico"] = "Bol. Inst. Estud. Med. Biol. Univ. Nac. Auton. Mex.",
  ["boletin del instituto de investigaciones bibliograficas"] = "Bol Inst Invest Bibliogr (Mexico)",
  ["boletin del instituto interamericano del nino"] = "Bol Inst Interam Nino",
  ["boletin del musea social argentino"] = "Bol. Mus. Soc. Argent.",
  ["boletin del prolap"] = "Bol PROLAP",
  ["boletin del sidema : publicacion periodico del centro de estudios de poblacion, cenep"] = "Bol SIDEMA",
  ["boletin dental uruguayo"] = "Bol Dent Urug",
  ["boletin editorial / el colegio de mexico, departamento de publicaciones"] = "Bol Ed Col Mex",
  ["boletin epidemiologico de chile / republica de chile, ministerio de salud"] = "Bol Epidemiol Chile",
  ["boletin geologico y minero"] = "Bol. Geol. Min.",
  ["boletin informativo - asociación médica argentina"] = "Bol Inf Asoc Med Argent",
  ["boletin latinoamericano y del caribe de plantas medicinales y aromaticas"] = "Bol. Latinoam. Caribe Plant. Med. Aromat.",
  ["boletin medico del hospital infantil de mexico"] = "Bol. Med. Hosp. Infant. Mex.",
  ["boletin mensual de estadisticas (colombia. departamento administrativo nacional de estadisticas)"] = "Bol Mens Estad DANE",
  ["boletin mexicano de derecho comparado / instituto de investigaciones juridicas, unam"] = "Bol Mex Derecho Comp",
  ["boletin médico científico"] = "Bol Med Cient",
  ["boletin odontologico"] = "Bol. Odontol. (B. Aires)",
  ["boletin odontologico mexicano"] = "Bol. Odontol. Mex.",
  ["boletin oficial del estado: gaceta de madrid. spain"] = "Bol Of Estado Gac Madr Spain",
  ["boletin socioeconomico (cali, colombia : 1985)"] = "Bol Socioecon",
  ["boletin, asociacion medica de puerto rico"] = "Bol. Asoc. Med. P. R.",
  ["boletin, colegio de profesionales de la enfermeria de puerto rico"] = "Bol. Col. Prof. Enferm. P. R.",
  ["boletin. academia peruana de cirugia"] = "Bol Acad Peru Cir",
  ["boletin. asociacion medica nacional de la republica panama"] = "Bol Asoc Medica Nac Repub Panama",
  ["boletin. asociacion venezolana de enfermeras profesionales"] = "Bol Asoc Venez Enferm Prof",
  ["boletin. buenos aires. universidad nacional. instituto de clinica quirurgica"] = "Bol B Aires Univ Nac Inst Clin Quir",
  ["boletin. mexico (city) hospital oftalmologico de nuestra senora de la luz"] = "Bol Hosp Oftalmol",
  ["boletin. santa fe, argentine republic (province) direccion general de odontologia"] = "Bol Dir Gen Odontol (Santa Fe)",
  ["boletin. sociedad chilena de obstetricia y ginecologia"] = "Bol Soc Chil Obstet Ginecol",
  ["boletin. sociedad chilena de obstetricia y ginecología"] = "Bol Soc Chil Obstet Ginecol",
  ["boletin. sociedad cubana de dermatologia y sifilografia"] = "Bol Soc Cubana Dermatol Sifilogr",
  ["boletin. sociedad de cirugia del uruguay"] = "Bol Soc Cir Urug",
  ["boletines de la sociedad de cirugia de rosario"] = "Bol Soc Cir Rosario",
  ["boletines de la sociedad de cirugía de rosario"] = "Bol Soc Cir Rosario",
  ["boletines y trabajos - sociedad argentina de cirujanos"] = "Bol Tr Soc Argent Cir",
  ["boletines y trabajos - sociedad de cirugia de buenos aires"] = "Bol Trab Soc Cir B Aires",
  ["boletines y trabajos / academia argentina de cirugia. academia argentina de cirugia"] = "Bol Trab Acad Argent Cir",
  ["boletines y trabajos de la academia argentina de cirugía. academia argentina de cirugía"] = "Bol Trab Acad Argent Cir",
  ["boletines y trabajos, sociedad de cirugia de buenos aires"] = "Bol. Trab. Soc. Cir. B. Aires",
  ["boletines y trabajos. sociedad de cirugía de buenos aires (1953)"] = "Bol Trab Soc Cir B Aires",
  ["boletiń de información dental"] = "Bol Inf Dent (Madr)",
  ["boletín (consejo nacional de población (peru))"] = "Bol Cons Nac Poblac",
  ["boletín (instituto de estudios de población y desarrollo (dominican republic))"] = "Boletin",
  ["boletín (sociedad española de hidrología médica)"] = "Bol Soc Esp Hidrol Med",
  ["boletín - círculo argentino de odontología"] = "Bol Circ Argent Odontol",
  ["boletín bibliográfico de la secretaría de hacienda y crédito público"] = "Bol Bibliogr Secr Hacienda Credito Publico",
  ["boletín chileno de parasitología"] = "Bol Chil Parasitol",
  ["boletín clínico"] = "Bol Clin",
  ["boletín cultural y bibliográfico"] = "Bol Cult Bibliogr",
  ["boletín de análisis demográfico"] = "Bol Anal Demogr",
  ["boletín de antropología"] = "Bol Antropol",
  ["boletín de antropología americana"] = "Bol Antropol Am",
  ["boletín de estadística. colombia. ministerio de higiene. división técnica de bio-estadística"] = "Bol Estad",
  ["boletín de estudios latinoamericanos y del caribe"] = "Bol Estud Latinoam Caribe",
  ["boletín de estudios médicos y biológicos"] = "Bol Estud Med Biol",
  ["boletín de historia y antigüedades"] = "Bol Hist Antig",
  ["boletín de indicadores de coyuntura"] = "Bol Indic Coyunt",
  ["boletín de la academia nacional de la historia. academia nacional de la historia (argentina)"] = "Bol Acad Nac Hist (Argent)",
  ["boletín de la academia nacional de la historia. academia nacional de la historia (venezuela)"] = "Bol Acad Nac Hist (Caracas)",
  ["boletín de la asociación argentina de odontología para niños"] = "Bol Asoc Argent Odontol Ninos",
  ["boletín de la asociación de demografía histórica"] = "Bol Asoc Demogr Hist",
  ["boletín de la asociación herpetológica española"] = "Bol Asoc Herpetol Esp",
  ["boletín de la asociación médica de puerto rico"] = "Bol Asoc Med P R",
  ["boletín de la biblioteca de menéndez pelayo"] = "Bol Bibl Menendez Pelayo",
  ["boletín de la liga contra el cáncer. liga contra el cáncer (cuba)"] = "Bol Liga Contra Cancer Havana",
  ["boletín de la oficina sanitaria panamericana. english ed"] = "Bol Of Sanit Panam (Engl)",
  ["boletín de la oficina sanitaria panamericana. pan american sanitary bureau"] = "Bol Oficina Sanit Panam",
  ["boletín de la real sociedad española de historia natural. sección biológica : órgano del instituto de ciencias naturales josé de acosta"] = "Bol R Soc Esp Hist Nat Secc Biol",
  ["boletín de la sea"] = "Bol SEA",
  ["boletín de la sociedad argentina de botánica. sociedad argentina de botánica"] = "Bol Soc Argent Bot",
  ["boletín de la sociedad castellonense de cultura. sociedad castellonense de cultura"] = "Bol Soc Castell Cult",
  ["boletín de la sociedad catalana de pediatría"] = "Bol Soc Catalana Pediatr",
  ["boletín de la sociedad de biología de concepción"] = "Bol Soc Biol Concepc",
  ["boletín de la sociedad dental de guatemala"] = "Bol Soc Dent Guatem",
  ["boletín de la sociedad española de historia de la farmacia"] = "Bol Soc Esp Hist Farm",
  ["boletín de la sociedad española de historia de la medicina"] = "Bol Soc Esp Hist Med",
  ["boletín de la sociedad matemática mexicana"] = "Bol Soc Mat Mex",
  ["boletín de la sociedad micológica de madrid"] = "Bol Soc Micol Madr",
  ["boletín de la sociedad mutualista médico-farmacéutica de guadalajara (1905)"] = "Bol Soc Mutual Med Farm Guadalaj",
  ["boletín de la sociedad química del péru"] = "Bol Soc Quim Peru",
  ["boletín de la sociedad valenciana de pediatría"] = "Bol Soc Valencia Pediatr",
  ["boletín de la sociedad zoológica del uruguay"] = "Bol Soc Zool Urug",
  ["boletín de la universidad de granada. universidad de granada"] = "Bol Univ Granada",
  ["boletín de los hospitales"] = "Bol Hosp",
  ["boletín de malariología y salud ambiental"] = "Bol Malariol Salud Ambient",
  ["boletín de odontología"] = "Bol Odontol (Bogota)",
  ["boletín de población"] = "Bol Poblac",
  ["boletín de prótesis"] = "Bol Protes",
  ["boletín de psicología"] = "Bol Psicol",
  ["boletín de sanidad vegetal, plagas"] = "Bol Sanid Veg Plagas",
  ["boletín de sanidad y asistencia pública. dominican republic. secretaría de sanidad y asistencia pública"] = "Bol Sanid Asist Publica",
  ["boletín del centro de estudios del siglo xviii. universidad de oviedo. centro de estudios del siglo xviii"] = "Bol Cent Estud Siglo XVIII",
  ["boletín del colegio médico de la habana"] = "Bol Col Med La Habana",
  ["boletín del hospital militar. cuba. ejército. hospital militar dr. carlos j. finlay"] = "Bol Hosp Mil Cuba Ejercito Hosp Mil Dr Carlos J Finlay",
  ["boletín del hospital san juan de dios"] = "Bol Hosp San Juan Dios",
  ["boletín del hospital universitario de caracas"] = "Bol Hosp Univ Caracas",
  ["boletín del instituto de estudios asturianos. instituto de estudios asturianos (oviedo, spain)"] = "Bol Inst Estud Asturianos",
  ["boletín del instituto de estudios médicos y biológicos, universidad nacional autónoma de méxico"] = "Bol Inst Estud Med Biol Univ Nac Auton Mex",
  ["boletín del instituto interamericano del niño"] = "Bol Inst Interam Nino",
  ["boletín del instituto internacional americano de protección a la infancia. interamerican children's institute"] = "Bol Inst Int Am Prot Infanc",
  ["boletín del prolap"] = "Bol PROLAP",
  ["boletín del servicio médico nacional de empleados. chile. servicio médico nacional de empleados"] = "Bol Servic Med Nac Empl",
  ["boletín del sidema : publicación periódico del centro de estudios de población, cenep"] = "Bol SIDEMA",
  ["boletín editorial"] = "Bol Ed Col Mex",
  ["boletín epidemiológico (mexico city, mexico)"] = "Bol Epidemiol (Mex City Mex)",
  ["boletín epidemiológico de chile"] = "Bol Epidemiol Chile",
  ["boletín informativo (sociedad micológica extremeña)"] = "Bol Inf Soc Micol Extremena",
  ["boletín latinoamericano y del caribe de plantas medicinales y aromáticas"] = "Bol Latinoam Caribe Plantas Med Aromat",
  ["boletín mensual de estadística demográfico-sanitaria de la península é islas adyacentes / ministerio de la gobernación del reino, dirección general de beneficencia y sanidad"] = "Bol Mens Estad Demogr Sanit Penins Islas Adyac",
  ["boletín mensual de estadísticas (colombia. departamento administrativo nacional de estadísticas)"] = "Bol Mens Estad DANE",
  ["boletín mexicano de derecho comparado"] = "Bol Mex Derecho Comp",
  ["boletín médico"] = "Bol Med",
  ["boletín médico del hospital infantil de méxico"] = "Bol Med Hosp Infant Mex",
  ["boletín médico-social"] = "Bol Med Soc",
  ["boletín micológico de famcal"] = "Bol Micolog FAMCAL",
  ["boletín odontológico"] = "Bol Odontol (B Aires)",
  ["boletín odontológico mexicano"] = "Bol Odontol Mex",
  ["boletín oficial del estado: gaceta de madrid. spain"] = "Bol Of Estado Gac Madr Spain",
  ["boletín socioeconómico (cali, colombia : 1985)"] = "Bol Socioecon",
  ["boletín veterinario"] = "Bol Vet",
  ["boletín. academia nacional de ciencias (córdoba, argentina)"] = "Bol Acad Nac Cienc",
  ["boletín. academia peruana de cirugía"] = "Bol Acad Peru Cir",
  ["boletín. asociación médica de santiago de los caballeros"] = "Bol Asoc Med Santiago Caballer",
  ["boletín. asociación médica nacional de la república panamá"] = "Bol Asoc Medica Nac Repub Panama",
  ["boletín. asociación venezolana de enfermeras profesionales"] = "Bol Asoc Venez Enferm Prof",
  ["boletín. bogota (colombia). clínica de marly"] = "Boletin Bogota",
  ["boletín. ciudad trujillo. hospital general padre billini"] = "Bol Ciudad Trujillo",
  ["boletín. clínica luis razetti, caracas. laboratorio"] = "Boletin Clin Luis Razetti Caracas",
  ["boletín. colegio médico de camagüey, cuba"] = "Bol Col Med Camaguey Cuba",
  ["boletín. colegio médico de holguín"] = "Bol Col Med Holguin",
  ["boletín. colegio médico de la habana (havana, cuba)"] = "Bol Col Med La Habana (La Habana)",
  ["boletín. consejo nacional de tuberculosis (cuba)"] = "Bol Cons Nac Tuberc Cuba",
  ["boletín. ecuador. departamento médico del seguro social ecuatoriano"] = "Bol Dep Med Seguro Social Ecuat",
  ["boletín. instituto psicopedagógico nacional (peru)"] = "Bol Inst Psicopedag Nac",
  ["boletín. liga argentina contra el reumatismo, buenos aires"] = "Bol Liga Argent Contra Reum B Aires",
  ["boletín. madrid. hospital provincial. instituto de patología médica"] = "Bol Patol Med",
  ["boletín. mexico city (mexico). hospital oftalmológico de nuestra señora de la luz"] = "Bol Hosp Oftalmol",
  ["boletín. peru. servicio nacional de protección materno-infantil"] = "Bol Serv Nac Prot Matern Infant Peru",
  ["boletín. santa fe (argentina : province). dirección general de odontología"] = "Bol Dir Gen Odontol (Santa Fe)",
  ["boletín. sociedad cubana de dermatología y sifilografía"] = "Bol Soc Cubana Dermatol Sifilogr",
  ["boletín. sociedad cubana de pediatría"] = "Bol Soc Cuba Pediatr",
  ["boletín. sociedad de cirugía del uruguay"] = "Bol Soc Cir Urug",
  ["boletín. sociedad médica de mendoza"] = "Bol Soc Med Mendoza",
  ["boletín. universidad de buenos aires. instituto de clínica quirúrgica"] = "Bol Univ B Aires",
  ["boletín. universidad nacional autónoma de méxico. instituto de química"] = "Bol Mex City Univ Nac",
  ["boletn asociacin chilena de proteccin de la familia"] = "Bol Asoc Chil Prot Fam",
  ["boletń trimestral - hospital de vin̂a del mar"] = "Bol Trimest Hosp Vina del Mar",
  ["boletn̕ asociacin̤ chilena de proteccin̤ de la familia"] = "Bol Asoc Chil Prot Fam",
  ["boletín auriense"] = "BAur",
  ["boletín chileno de parasitología"] = "Bol. chil. parasitol.",
  ["boletín de la asociación española de orientalistas"] = "BEspOr",
  ["boletín de la comisión provincial de monumentos históricos y artísticos de lugo"] = "BLugo",
  ["boletín de la institución »sancho el sabio«"] = "BVitoria",
  ["boletín de la real academia de la historia"] = "BAcRHist",
  ["boletín de la sociedad mexicana de historia y filosofía de la medicina"] = "Bol Soc Mex Hist Filos Med",
  ["boletín del instituto de estudios giennenses"] = "BJaen",
  ["boletín del museo arqueológico nacional"] = "BMAN",
  ["boletín del museo arqueológico nacional, madrid"] = "BMusMadr",
  ["boletín del museo de cádiz"] = "BMusCadiz",
  ["boletín del museo provincial de bellas artes"] = "BMusPBelArt",
  ["boletín del seminario de estudios de arte y arqueología"] = "BSEAA",
  ["boletín del seminario de estudios de arte y arqueología, universidad de valladolid"] = "BSAA",
  ["boletín informativo hispanoamericano de historia de la medicina"] = "Bol Inf Hisp Am Hist Med",
  ["boletín. asociación española de amigos de la arqueología"] = "BAsEspA",
  ["bolet’in de la agrupacion odontologica de la zona central de la capital federal."] = "Bol Agrup Odontol Zona Cent Cap Fed B Aires",
  ["bolletino del centro inernazionale per lo studio dei papyri ercolanesi (cronache ercolanesi)"] = "BCPE",
  ["bolletino della società italiana della scienza del suolo"] = "Boll. Soc. Italiana Sci. Suolo",
  ["bolletino della società ticinese di scienze naturali"] = "Boll. Soc. ticin. Sci. nat.",
  ["bollettino - societa italiana biologia sperimentale"] = "Boll Soc Ital Biol Sper",
  ["bollettino - societa italiana di biologia sperimentale"] = "Boll.-Soc. Ital. Biol. Sper.",
  ["bollettino - societa medico chirurgica cremona"] = "Boll Soc Med Chir Cremona",
  ["bollettino - societa medico-chirurgica di pisa"] = "Boll Soc Med Chir Pisa",
  ["bollettino - società medico chirurgica cremona"] = "Boll Soc Med Chir Cremona",
  ["bollettino - società medico-chirurgica di modena"] = "Boll Soc Med Chir Modena",
  ["bollettino - società medico-chirurgica di pisa"] = "Boll Soc Med Chir Pisa",
  ["bollettino - unione italiana chimici igienisti"] = "Boll. Unione Ital Chim. Ig.",
  ["bollettino annuale. musei ferraresi"] = "BAnnMusFerr",
  ["bollettino chimico farmaceutico"] = "Boll Chim Farm",
  ["bollettino d'oculistica"] = "Boll Ocul",
  ["bollettino dei chimici igienisti"] = "Boll. Chim. Ig.",
  ["bollettino dei classici"] = "BollClass",
  ["bollettino dei monumenti, musei e gallerie pontificie"] = "BMMP",
  ["bollettino dei musei comunali di roma"] = "BMusRom",
  ["bollettino dei musei e degli istituti biologici dell'università di genova"] = "Boll Mus Ist Biol Univ Genova",
  ["bollettino del centro di studi per la storia dell’architettura"] = "BArchit",
  ["bollettino del centro internazionale a. beltrame di storia dello spazio e del tempo"] = "Boll Cent Int Beltrame Stor Spazio Tempo",
  ["bollettino del centro micologico friulano"] = "Boll Cent Micol Friul",
  ["bollettino del circolo numismatico napoletano"] = "BCircNumNap",
  ["bollettino del gruppo micologico g. bresadola"] = "Boll Gruppo Micol G Bresadola",
  ["bollettino del museo civico di padova"] = "BMusPadova",
  ["bollettino del museo civico di storia naturale di verona. botanica, zoologia"] = "Boll Mus Civ Stor Nat Verona Bot Zool",
  ["bollettino dell istituto sieroterapico milanese"] = "Boll. Ist. Sieroter. Milan.",
  ["bollettino dell'archivio per la storia del movimento sociale cattolico in italia"] = "Boll Arch Stor Mov Soc Cattol Ital",
  ["bollettino dell'istituto sieroterapico milanese"] = "Boll Ist Sieroter Milan",
  ["bollettino dell'ospedale di varese"] = "Boll Osp Varese",
  ["bollettino della associazione micologica ed ecologica romana"] = "Boll Assoc Micol Ecol Romana",
  ["bollettino della badia greca di grottaferrata"] = "BBGG",
  ["bollettino della societa italiana di biologia sperimentale"] = "Boll Soc Ital Biol Sper",
  ["bollettino della societa italiana di cardiologia"] = "Boll. Soc. Ital. Cardiol.",
  ["bollettino della societa medico chirurgica bresciana"] = "Boll Soc Med Chir Bresciana",
  ["bollettino della societa paleontologica italiana"] = "Boll. Soc. Paleontol. Ital.",
  ["bollettino della società italiana di biologia sperimentale"] = "Boll Soc Ital Biol Sper",
  ["bollettino della società italiana di cardiologia"] = "Boll Soc Ital Cardiol",
  ["bollettino della società medico chirurgica bresciana"] = "Boll Soc Med Chir Bresciana",
  ["bollettino della società paleontologica italiana. società paleontologica italiana"] = "Boll Soc Paleontol Ital",
  ["bollettino della società piemontese di archeologia e belle arti"] = "BTorino",
  ["bollettino della unione matematica italiana"] = "Boll. Unione Mat. Ital. Sez. B Artic. Ric. Mat. (8)",
  ["bollettino della unione matematica italiana (2008)"] = "Boll Unione Mat Ital (2008)",
  ["bollettino della unione storia ed arte"] = "BStorArt",
  ["bollettino delle malattie dell orecchio, della gola, del naso"] = "Boll. Mal. Orecch. Gola. Naso",
  ["bollettino delle malattie dell'orecchio, della gola e del naso"] = "Boll Mal Orecch Gola Naso",
  ["bollettino delle malattie dell'orecchio, della gola, del naso"] = "Boll Mal Orecch Gola Naso",
  ["bollettino delle malattie dell’orecchio, della gola, del naso"] = "Boll Mal Orecch Gola Naso",
  ["bollettino delle sezioni regionali"] = "Boll Sez Reg Soc Ital Dermatol Sifilogr",
  ["bollettino dell’archivio per la storia del movimento sociale cattolico in italia"] = "Boll Arch Stor Mov Soc Cattol Ital",
  ["bollettino dell’associazione iasos di caria"] = "BIasos",
  ["bollettino dell’istituto centrale del restauro"] = "BRest",
  ["bollettino dell’istituto di filologia greca, università di padova"] = "BFilGrPadova",
  ["bollettino dell’istituto di storia e di arte del lazio meridionale"] = "BLazioMerid",
  ["bollettino dell’istituto sieroterapico milanese"] = "Boll Ist Sieroter Milan",
  ["bollettino dell’istituto storico artistico orvietano"] = "BIstOrvieto",
  ["bollettino dell’ospedale di varese"] = "Boll Osp Varese",
  ["bollettino dell’unione matematica italiana"] = "Boll. Unione Mat. Ital.",
  ["bollettino di archeologia"] = "BA",
  ["bollettino di archeologia subacquea"] = "BASub",
  ["bollettino di demografia storica"] = "Boll Demogr Stor",
  ["bollettino di informazioni del centro di studi e documentazione sull’area elima"] = "BInfCESDAE",
  ["bollettino di informazioni. associazione archeologica centumcellae"] = "BInfCentumcellae",
  ["bollettino di numismatica"] = "BNum",
  ["bollettino di oculistica"] = "Boll Ocul",
  ["bollettino di storia delle scienze matematiche"] = "Boll. Stor. Sci. Mat.",
  ["bollettino di studi e ricerche. biblioteca comunale di bolsena"] = "BBolsena",
  ["bollettino di studi latini"] = "BStLat",
  ["bollettino di studi latini: periodico quadrimestrale d’informazione bibliografica"] = "BStudLat",
  ["bollettino di zoologia"] = "Boll Zool",
  ["bollettino d’arte"] = "BdA",
  ["bollettino d’arte del ministero per i beni culturali e ambientali"] = "BA",
  ["bollettino d’informazioni della consociazione nazionale infermiere professionali e assistenti sanitaire visitatrici"] = "Boll. Inf. Consoc. Naz. (Rome)",
  ["bollettino d’oculistica"] = "Boll. Ocul.",
  ["bollettino e memorie della societa piemontese di chirurgia"] = "Boll Mem Soc Piemont Chir",
  ["bollettino e memorie della societa tosco-umbra di chirurgia"] = "Boll Mem Soc Tosco Umbra Chir",
  ["bollettino e memorie della società piemontese di chirurgia"] = "Boll Mem Soc Piemont Chir",
  ["bollettino e memorie della società tosco-umbra di chirurgia"] = "Boll Mem Soc Tosco Umbra Chir",
  ["bollettino flegreo. rivista di storia, arte e scienze"] = "BFlegr",
  ["bollettino malacologico"] = "Boll Malacol",
  ["bollettino mensile di statistica (rome, italy)"] = "Boll Mens Stat (Rome)",
  ["bollettino metallografico"] = "Boll. Metallogr.",
  ["bollettino metallografico. metallurgical bulletin"] = "Boll Metallogr",
  ["bollettino schermografico"] = "Boll Schermogr",
  ["bollettino storico della basilicata"] = "BBasil",
  ["bollettino storico della città di foligno"] = "BFoligno",
  ["bollettino storico-bibliografico subalpino"] = "Boll Stor Bibliogr Subalp",
  ["bollettino, societa italiana biologia sperimentale"] = "Boll. Soc. Ital. Biol. Sper.",
  ["bollettino, societa medico chirurgica cremona"] = "Boll. Soc. Med. Chir. Cremona",
  ["bollettino. centro di studi filologici e linguistici siciliani"] = "BFilLingSic",
  ["bollettino. monumenti, musei e gallerie pontificie"] = "BMonMusPont",
  ["bollettino. societa italiana di medicina e igiene tropicale. sezione eritrea"] = "Bull Soc Ital Med Ig Trop Sez Eritrea",
  ["bologna incontri"] = "Bologna Incontri",
  ["bologna medica"] = "Bologna Medica",
  ["bolskan. revista de arqueología del instituto de estudios altoaragoneses, revista de arqueología oscense"] = "Bolskan",
  ["bolyai society mathematical studies"] = "Bolyai Soc. Math. Stud.",
  ["bone & joint open"] = "Bone Jt Open",
  ["bone & joint research"] = "Bone Joint Res",
  ["bone and mineral"] = "Bone Miner",
  ["bone and tissue regeneration insights"] = "Bone Tissue Regen Insights",
  ["bone marrow research"] = "Bone Marrow Res",
  ["bone marrow transplantation"] = "Bone Marrow Transplant",
  ["bone reports"] = "Bone Rep",
  ["bone research"] = "Bone Res",
  ["bonekey osteovision"] = "Bonekey Osteovision",
  ["bonekey reports"] = "Bonekey Rep",
  ["bonn zoological bulletin"] = "Bonn Zool Bull",
  ["bonner geschichtsblatter / im auftrage des vereins alt-bonn herausgegeben von stadtarchivar"] = "Bonn Geschichtsbl",
  ["bonner geschichtsblätter"] = "Bonn Geschichtsbl",
  ["bonner hefte zur vorgeschichte"] = "BonnHVg",
  ["bonner jahrbücher des rheinischen landesmuseums in bonn"] = "BJb",
  ["bonner jahrbücher des rheinischen landesmuseums in bonn und des vereins von altertumsfreunden im rheinlande"] = "BJ",
  ["bonner mathematische schriften"] = "Bonner Math. Schriften",
  ["bonner zoologische beiträge : herausgeber: zoologisches forschungsinstitut und museum alexander koenig, bonn"] = "Bonn Zool Beitr",
  ["bonner zoologische beiträge"] = "Bonn. Zool. Beitr.",
  ["bonner zoologische monographien"] = "Bonn Zool Monogr",
  ["book history"] = "Book Hist",
  ["book of abstracts : iv annual meeting health technology assessment international, barcelona, catalonia, spain 17-20 june 2007. health technology assessment international. meeting (4th : 2007 : barcelona, spain)"] = "Book Abstr Health Technol Assess Int Meet 4th 2007 Barc Spain",
  ["book supplement to the journal of child psychology and psychiatry"] = "Book Suppl J Child Psychol Psychiatr",
  ["bookman's price index : an annual guide to the values of rare and other out-of-print books and sets of periodicals"] = "Bookmans Price Index (Det)",
  ["books at iowa"] = "Books Iowa",
  ["books in print"] = "Books Print",
  ["bopuxue zazhi"] = "Bopuxue Zazhi",
  ["bordeaux chirurgical"] = "Bord Chir",
  ["bordeaux medical"] = "Bord. Med.",
  ["bordeaux médical"] = "Bord Med",
  ["borden's review of nutrition research"] = "Bordens Rev Nutr Res",
  ["borden’s review of nutrition research"] = "Bordens Rev. Nutr. Res.",
  ["borderline personality disorder and emotion dysregulation"] = "Borderline Personal Disord Emot Dysregul",
  ["boreal environment research"] = "Boreal Environ. Res.",
  ["boreal environment research : an international interdisciplinary journal"] = "Boreal Environ Res",
  ["boreas (oslo)"] = "Boreas (Oslo)",
  ["boreas. münstersche beiträge zur archäologie"] = "Boreas",
  ["boreas. uppsala studies in ancient mediterranean and near eastern civilization"] = "BoreasUpps",
  ["borgyogyaszati es venerologiaia szemle"] = "Borgyogy Venerol Sz",
  ["borsuye : zeitschrift fur medizin u. kultur"] = "Borsuye",
  ["borsuye : zeitschrift für medizin u. kultur"] = "Borsuye",
  ["bosnian journal of basic medical sciences"] = "Bosn J Basic Med Sci",
  ["bosque (valdivia)"] = "Bosque (Valdivia)",
  ["boston bar journal"] = "Boston Bar J",
  ["boston college environmental affairs law review. boston college. law school"] = "Boston Coll Environ Aff Law Rev",
  ["boston college industrial and commercial law review"] = "Boston Coll Ind Commer Law Rev",
  ["boston college international and comparative law review"] = "Boston College Int Comp Law Rev",
  ["boston college law review. boston college. law school"] = "Boston Coll Law Rev",
  ["boston college third world law journal"] = "Boston Coll Third World Law J",
  ["boston studies in the philosophy and history of science"] = "Boston Stud. Philos. Hist. Sci.",
  ["boston studies in the philosophy of science"] = "Boston Stud. Philos. Sci.",
  ["boston university international law journal"] = "Boston Univ Int Law J",
  ["boston university journal"] = "Boston Univ J",
  ["boston university law review. boston university. school of law"] = "Boston Univ Law Rev",
  ["bostonia (boston, mass. : 1986)"] = "Bostonia",
  ["botanica (vilnius, lithuania)"] = "Botanica (Vilnius)",
  ["botanica acta"] = "Botan. Acta",
  ["botanica acta : berichte der deutschen botanischen gesellschaft = journal of the german botanical society"] = "Bot Acta",
  ["botanica lithuanica"] = "Bot Lith",
  ["botanica marina"] = "Botanica Marina",
  ["botanica pacifica : journal of plant science and conservation"] = "Bot Pac",
  ["botanica serbica"] = "Bot Serb",
  ["botanical bulletin of academia sinica. new series. zhong yang yan jiu yuan. zhi wu yan jiu suo"] = "Bot Bull Acad Sinica (Taiwan)",
  ["botanical gazette (chicago, ill.)"] = "Bot Gaz",
  ["botanical journal of scotland"] = "Bot J Scotl",
  ["botanical journal of the linnean society"] = "Bot. J. Linn. Soc.",
  ["botanical journal of the linnean society. linnean society of london"] = "Bot J Linn Soc",
  ["botanical magazine tokyo"] = "Bot. Mag. Tokyo",
  ["botanical museum leaflets, harvard university"] = "Bot Mus Lealf Harv Univ",
  ["botanical review"] = "Bot. Rev.",
  ["botanical sciences"] = "Bot Sci",
  ["botanical studies"] = "Bot Stud",
  ["botanicheskiĭ zhurnal"] = "Bot Z",
  ["botanics : targets and therapy"] = "Botanics",
  ["botanische jahrbucher fur systematik, pflanzengeschichte und pflanzengeographie"] = "Bot Jahrb Syst Pflanzengesch Pflanzengeogr",
  ["botanische jahrbücher für systematik, pflanzengeschichte und pflanzengeographie"] = "Bot. Jahrb. Syst. Pflanzengesch. Pflanzengeogr.",
  ["botanische zeitung"] = "Bot. Ztg.",
  ["botanisches centralblatt"] = "Bot. Cent.bl.",
  ["botany letters"] = "Bot Lett",
  ["botánica complutensis"] = "Bot Complut",
  ["botswana national health bulletin"] = "Botsw Natl Health Bull",
  ["botswana national health bulletin / issued by the national health institute"] = "Botsw Natl Health Bull",
  ["botswana notes and records"] = "Botsw Notes Rec",
  ["boundary elements xxiv: incorporating meshless solutions"] = "Adv. Bound. Elem. Ser.",
  ["boundary value problems"] = "Boundary Value Probl.",
  ["boundary-layer meteorology"] = "Boundary Layer Meteorol.",
  ["bōei ika daigakkō zasshi = journal of the national defense medical college"] = "Boei Ika Daigakko Zasshi",
  ["börgyógyászati és venerologiaia szemle"] = "Borgyogy Venerol Sz",
  ["bracara augusta. revista cultural da câmara municipal de braga"] = "BracAug",
  ["bragantia : boletim técnico do instituto agronômico do estado de são paulo"] = "Bragantia",
  ["brain & development"] = "Brain Dev",
  ["brain : a journal of neurology"] = "Brain",
  ["brain : broad research in artificial intelligence and neuroscience"] = "Brain (Bacau)",
  ["brain and behavior"] = "Brain Behav",
  ["brain and cognition"] = "Brain Cogn",
  ["brain and development"] = "Brain Dev.",
  ["brain and language"] = "Brain Lang",
  ["brain and nerve = shinkei kenkyū no shinpo"] = "Brain Nerve",
  ["brain and neuroscience advances"] = "Brain Neurosci Adv",
  ["brain cell biology"] = "Brain Cell Biol",
  ["brain circulation"] = "Brain Circ",
  ["brain communications"] = "Brain Commun",
  ["brain computer interfaces (abingdon, england)"] = "Brain Comput Interfaces (Abingdon)",
  ["brain connectivity"] = "Brain Connect",
  ["brain disorders"] = "Brain Disord",
  ["brain disorders & therapy"] = "Brain Disord Ther",
  ["brain hemorrhages"] = "Brain Hemorrhages",
  ["brain imaging and behavior"] = "Brain Imaging Behav",
  ["brain impairment"] = "Brain Impairment",
  ["brain impairment : a multidisciplinary journal of the australian society for the study of brain impairment"] = "Brain Impair",
  ["brain informatics"] = "Brain Inform",
  ["brain informatics : international conference, bi 2018, arlington, tx, usa, december 7-9, 2018, proceedings. international conference on brain informatics (2018 : arlington, tex.)"] = "Brain Inform (2018)",
  ["brain informatics and health : 8th international conference, bih 2015, london, uk, august 30-september 2, 2015 : proceedings. bih (conference) (8th : 2015 : london, england)"] = "Brain Inform Health (2015)",
  ["brain injury"] = "Brain Inj",
  ["brain injury : [bi]"] = "Brain Inj",
  ["brain multiphysics"] = "Brain Multiphys",
  ["brain pathology"] = "Brain Pathol.",
  ["brain pathology (zurich, switzerland)"] = "Brain Pathol",
  ["brain plasticity"] = "Brain Plast.",
  ["brain plasticity (amsterdam, netherlands)"] = "Brain Plast",
  ["brain research"] = "Brain Res",
  ["brain research bulletin"] = "Brain Res Bull",
  ["brain research protocols"] = "Brain Res. Protoc.",
  ["brain research reviews"] = "Brain Res Rev",
  ["brain research. brain research protocols"] = "Brain Res Brain Res Protoc",
  ["brain research. brain research reviews"] = "Brain Res Brain Res Rev",
  ["brain research. cognitive brain research"] = "Brain Res Cogn Brain Res",
  ["brain research. developmental brain research"] = "Brain Res Dev Brain Res",
  ["brain research. gene expression patterns"] = "Brain Res Gene Expr Patterns",
  ["brain research. molecular brain research"] = "Brain Res Mol Brain Res",
  ["brain science advances"] = "Brain Sci. Adv.",
  ["brain sciences"] = "Brain Sci",
  ["brain stimulation"] = "Brain Stimul",
  ["brain structure & function"] = "Brain Struct Funct",
  ["brain structure and function"] = "Brain Struct. Funct.",
  ["brain topography"] = "Brain Topogr",
  ["brain tumor pathology"] = "Brain Tumor Pathol",
  ["brain tumor research and treatment"] = "Brain Tumor Res Treat",
  ["brain, behavior and evolution"] = "Brain Behav Evol",
  ["brain, behavior and immunity"] = "Brain. Behav. Immun.",
  ["brain, behavior, & immunity - health"] = "Brain Behav Immun Health",
  ["brain, behavior, and immunity"] = "Brain Behav Immun",
  ["brain-apparatus communication: a journal of bacomics"] = "Brain-Appar. Commun.: J. Bacomics",
  ["brain-computer interfaces"] = "Brain-Comput. Interfaces",
  ["brainlesion : glioma, multiple sclerosis, stroke and traumatic brain injuries. brainles (workshop)"] = "Brainlesion",
  ["brains, minds & media : journal of new media in neural and cognitive science and education"] = "Brains Minds Media",
  ["brandeis journal of family law"] = "Brandeis J Fam Law",
  ["brandeis journal of family law / louis d. brandeis school of law, university of louisville"] = "Brandeis J Fam Law",
  ["brandeis law journal"] = "Brandeis Law J",
  ["brandeis review"] = "Brandeis Rev",
  ["brasil médico-cirúrgico"] = "Bras Med Cir",
  ["braspen journal"] = "BRASPEN J",
  ["bratislavske lekarske listy"] = "Bratisl Lek Listy",
  ["bratislavské lekárske listy"] = "Bratisl Lek Listy",
  ["braunschweiger veroffentlichungen zur geschichte der pharmazie und der naturwissenschaften"] = "Braunschw Veroff Gesch Pharm Naturwiss",
  ["braunschweiger veröffentlichungen zur geschichte der pharmazie und der naturwissenschaften"] = "Braunschw Veroff Gesch Pharm Naturwiss",
  ["brazilian archives of biology and technology"] = "Braz. Arch. Biol. Technol.",
  ["brazilian archives of biology and technology = (archivos de biologia e tecnologia)"] = "Braz Arch Biol Technol",
  ["brazilian dental journal"] = "Braz Dent J",
  ["brazilian dental science"] = "Braz Dent Sci",
  ["brazilian economic studies"] = "Braz Econ Stud",
  ["brazilian electronic journal of economics"] = "Brazilian Electronic J. Econ.",
  ["brazilian journal of analytical chemistry"] = "Braz. J. Anal. Chem.",
  ["brazilian journal of anesthesiology (elsevier)"] = "Braz J Anesthesiol",
  ["brazilian journal of animal science"] = "Braz. J. Anim. Sci.",
  ["brazilian journal of applied economics"] = "Brazilian J. Appl. Econ.",
  ["brazilian journal of aquatic sciences and technology"] = "Braz. J. Aquat. Sci. Technol.",
  ["brazilian journal of biology"] = "Braz. J. Biol.",
  ["brazilian journal of biology = revista brasleira de biologia"] = "Braz J Biol",
  ["brazilian journal of botany"] = "Braz. J. Bot.",
  ["brazilian journal of cardiovascular surgery"] = "Braz J Cardiovasc Surg",
  ["brazilian journal of chemical engineering"] = "Braz. J. Chem. Eng.",
  ["brazilian journal of chemistry"] = "Braz. J. Chem.",
  ["brazilian journal of food technology"] = "Braz. J. Food Technol.",
  ["brazilian journal of food technology = revista brasileira de tecnologia de alimentos"] = "Braz J Food Technol",
  ["brazilian journal of geology"] = "Braz. J. Geol.",
  ["brazilian journal of infectious diseases"] = "Braz. J. Infect. Dis.",
  ["brazilian journal of materials science and engineering"] = "Braz. J. Mater. Sci. Eng.",
  ["brazilian journal of medical and biological research"] = "Braz. J. Med. Biol. Res.",
  ["brazilian journal of medical and biological research = revista brasileira de pesquisas médicas e biológicas"] = "Braz J Med Biol Res",
  ["brazilian journal of microbiology"] = "Braz. J. Microbiol.",
  ["brazilian journal of microbiology : [publication of the brazilian society for microbiology]"] = "Braz J Microbiol",
  ["brazilian journal of oceanography"] = "Braz. J. Oceanogr.",
  ["brazilian journal of operations and production management"] = "Braz. J. Oper. Prod. Manage.",
  ["brazilian journal of oral sciences"] = "Braz J Oral Sci",
  ["brazilian journal of otorhinolaryngology"] = "Braz J Otorhinolaryngol",
  ["brazilian journal of pharmaceutical sciences"] = "Braz. J. Pharm. Sci.",
  ["brazilian journal of physical therapy"] = "Braz J Phys Ther",
  ["brazilian journal of physics"] = "Braz. J. Phys.",
  ["brazilian journal of political economy"] = "Brazilian J. Polit. Economy",
  ["brazilian journal of population studies"] = "Braz J Popul Stud",
  ["brazilian journal of poultry science"] = "Braz. J. Poult. Sci.",
  ["brazilian journal of probability and statistics"] = "Braz. J. Probab. Stat.",
  ["brazilian journal of science"] = "Braz. J. of Sci.",
  ["brazilian journal of science and technology"] = "Braz. J. of Sci. Technol.",
  ["brazilian journal of veterinary pathology"] = "Braz. J. Vet. Pathol.",
  ["brazilian journal of veterinary research and animal science"] = "Braz J Vet Res Anim Sci",
  ["brazilian oral research"] = "Braz Oral Res",
  ["brazilian review of econometrics"] = "Brazilian Rev. Econometrics",
  ["breast (edinburgh, scotland)"] = "Breast",
  ["breast cancer (dove medical press)"] = "Breast Cancer (Dove Med Press)",
  ["breast cancer (tokyo, japan)"] = "Breast Cancer",
  ["breast cancer : basic and clinical research"] = "Breast Cancer (Auckl)",
  ["breast cancer management"] = "Breast Cancer Manag",
  ["breast cancer online : bco"] = "Breast Cancer Online",
  ["breast cancer research"] = "Breast Cancer Res.",
  ["breast cancer research : bcr"] = "Breast Cancer Res",
  ["breast cancer research and treatment"] = "Breast Cancer Res Treat",
  ["breast care (basel, switzerland)"] = "Breast Care (Basel)",
  ["breast disease"] = "Breast Dis",
  ["breast imaging : 11th international workshop, iwdm 2012, philadelphia, pa, usa, july 8-11, 2012 : proceedings. international workshop on breast imaging (11th : 2012 : philadelphia, pa.)"] = "Breast Imaging (2012)",
  ["breastfeeding medicine : the official journal of the academy of breastfeeding medicine"] = "Breastfeed Med",
  ["breastfeeding review"] = "Breastfeed. Rev.",
  ["breastfeeding review : professional publication of the nursing mothers' association of australia"] = "Breastfeed Rev",
  ["breastfeeding review : professional publication of the nursing mothers’ association of australia"] = "Breastfeed Rev",
  ["breathe (sheffield, england)"] = "Breathe (Sheff)",
  ["breeding science"] = "Breed Sci",
  ["bremer archäologische blätter"] = "BremABl",
  ["brethren life and thought"] = "Brethr Life Thought",
  ["brewer and distiller international"] = "Brew. Distill. Int.",
  ["brewing science (nürnberg)"] = "Brew Sci",
  ["brewingscience - monatsschrift fur brauwissenschaft"] = "BrewingSci. Monatsschr. Brauwiss.",
  ["bridge (washington, d.c. : 1969)"] = "Bridge (Wash D C)",
  ["bridge (washington, d.c. : 1992)"] = "Bridg Wash D C",
  ["bridge structures"] = "Bridge Struct.",
  ["brief treatment and crisis intervention"] = "Brief Treat Crisis Interv",
  ["briefings in bioinformatics"] = "Brief Bioinform",
  ["briefings in functional genomics"] = "Brief Funct Genomics",
  ["briefings in functional genomics & proteomics"] = "Brief Funct Genomic Proteomic",
  ["briefings in functional genomics and proteomics"] = "Brief. Funct. Genomic. Proteomic.",
  ["briefings in medical ethics"] = "Brief Med Ethics",
  ["brigantium. museo arqueolóxico e histórico"] = "Brigantium",
  ["brigham young university law review"] = "Brigh Young Univ Law Rev",
  ["brigham young university studies. brigham young university"] = "Brigham Young Univ Stud",
  ["brill’s companions to classical reception"] = "Brill’s Companions Class. Recept.",
  ["bristol medico-chirurgical journal"] = "Bristol Med. Chir. J.",
  ["bristol medico-chirurgical journal (1883)"] = "Bristol Med Chir J (1883)",
  ["bristol medico-chirurgical journal (1963)"] = "Bristol Med Chir J",
  ["britannia (society for the promotion of roman studies)"] = "Britannia",
  ["britannica yearbook of science and the future"] = "Br Yearb Sci",
  ["british & irish botany"] = "Br Ir Bot",
  ["british archaeological abstracts"] = "BritAA",
  ["british archaeological bibliography"] = "BritAB",
  ["british archaeological reports, british series"] = "Br. Archaeol. Rep., Br. Ser.",
  ["british archaeological reports. british series"] = "BAR",
  ["british archaeological reports. international series"] = "BARIntSer",
  ["british biotechnology journal"] = "Br Biotechnol J",
  ["british birds; an illustrated magazine devoted to the birds on the british list"] = "Br Birds",
  ["british book news"] = "Br Book News",
  ["british ceramic transactions"] = "Br. Ceram. Trans.",
  ["british columbia medical journal"] = "B C Med J",
  ["british corrosion journal"] = "Br. Corros. J.",
  ["british dental journal"] = "Br Dent J",
  ["british dental nurses' journal (fleetwood, england : 1994)"] = "Br Dent Nurs J",
  ["british dental nurses’ journal"] = "Br. Dent. Nurs. J.",
  ["british dental nurses’ journal (fleetwood, england : 1994)"] = "Br Dent Nurs J",
  ["british dental surgery assistant"] = "Br. Dent. Surg. Assist.",
  ["british educational research journal"] = "Br Educ Res J",
  ["british food journal"] = "Br. Food J.",
  ["british food journal (croydon, england)"] = "Br Food J",
  ["british heart journal"] = "Br Heart J",
  ["british heritage"] = "Br Herit",
  ["british history illustrated"] = "Br Hist Illus",
  ["british homoeopathic journal"] = "Br. Homeopath. J.",
  ["british journal for military history"] = "Br J Mil Hist",
  ["british journal for the history of mathematics"] = "Br. J. Hist. Math.",
  ["british journal for the history of philosophy"] = "BJHP",
  ["british journal for the history of philosophy : bjhp : the journal of the british society for the history of philosophy"] = "Br J Hist Philos",
  ["british journal for the history of science"] = "Br J Hist Sci",
  ["british journal for the philosophy of science"] = "Br. J. Philos. Sci.",
  ["british journal of addiction"] = "Br J Addict",
  ["british journal of addiction to alcohol & other drugs"] = "Br. J. Addict. Alcohol Other Drugs",
  ["british journal of addiction to alcohol and other drugs"] = "Br. J. Addict. Alcohol Other Drugs",
  ["british journal of anaesthesia"] = "Br J Anaesth",
  ["british journal of animal behaviour"] = "Br. J. Anim. Behav.",
  ["british journal of applied physics"] = "Br. J. Appl. Phys.",
  ["british journal of applied science & technology"] = "Br J Appl Sci Technol",
  ["british journal of audiology"] = "Br J Audiol",
  ["british journal of audiology. supplement"] = "Br J Audiol Suppl",
  ["british journal of biomedical science"] = "Br J Biomed Sci",
  ["british journal of canadian studies"] = "Br. J. Can. Stud.",
  ["british journal of cancer"] = "Br J Cancer",
  ["british journal of cancer. supplement"] = "Br. J. Cancer. Suppl.",
  ["british journal of clinical pharmacology"] = "Br J Clin Pharmacol",
  ["british journal of clinical practice"] = "Br. J. Clin. Pract.",
  ["british journal of clinical practice. supplement"] = "Br J Clin Pract Suppl",
  ["british journal of clinical practice. symposium supplement"] = "Br. J. Clin. Pract. Symp. Suppl.",
  ["british journal of clinical psychology"] = "Br. J. Clin. Psychol.",
  ["british journal of community nursing"] = "Br J Community Nurs",
  ["british journal of dermatology"] = "Br. J. Dermatol.",
  ["british journal of diseases of the chest"] = "Br J Dis Chest",
  ["british journal of disorders of communication"] = "Br. J. Disord. Commun.",
  ["british journal of educational psychology"] = "Br. J. Educ. Psychol.",
  ["british journal of educational technology : journal of the council for educational technology"] = "Br J Educ Technol",
  ["british journal of experimental pathology"] = "Br J Exp Pathol",
  ["british journal of family planning"] = "Br. J. Fam. Plann.",
  ["british journal of forensic practice"] = "Br. J. Forensic Pract.",
  ["british journal of general practice"] = "Br. J. Gen. Pract.",
  ["british journal of guidance & counselling"] = "Br J Guid Counc",
  ["british journal of haematology"] = "Br J Haematol",
  ["british journal of health psychology"] = "Br J Health Psychol",
  ["british journal of hospital medicine"] = "Br J Hosp Med",
  ["british journal of hospital medicine (london, england : 2005)"] = "Br J Hosp Med (Lond)",
  ["british journal of industrial medicine"] = "Br J Ind Med",
  ["british journal of industrial relations"] = "Br J Ind Relat",
  ["british journal of international studies"] = "Br J Int Stud",
  ["british journal of law and society"] = "Br J Law Soc",
  ["british journal of learning disabilities"] = "Br J Learn Disabil",
  ["british journal of mathematical and statistical psychology"] = "Br. J. Math. Stat. Psychol.",
  ["british journal of medical & surgical urology"] = "Br J Med Surg Urol",
  ["british journal of medical education"] = "Br J Med Educ",
  ["british journal of medical hypnotism"] = "Br. J. Med. Hypn.",
  ["british journal of medical hypnotism : official organ of the british society of medical hypnotists"] = "Br J Med Hypn",
  ["british journal of medical practitioners"] = "Br. J. Med. Pract.",
  ["british journal of medical psychology"] = "Br. J. Med. Psychol.",
  ["british journal of medicine and medical research"] = "Br J Med Med Res",
  ["british journal of midwifery"] = "Br J Midwifery",
  ["british journal of neurosurgery"] = "Br J Neurosurg",
  ["british journal of nursing"] = "Br. J. Nurs.",
  ["british journal of nursing (mark allen publishing)"] = "Br J Nurs",
  ["british journal of nutrition"] = "Br. J. Nutr.",
  ["british journal of obstetrics and gynaecology"] = "Br J Obstet Gynaecol",
  ["british journal of ophthalmology"] = "Br. J. Ophthalmol.",
  ["british journal of oral and maxillofacial surgery"] = "Br. J. Oral Maxillofac. Surg.",
  ["british journal of oral surgery"] = "Br. J. Oral Surg.",
  ["british journal of orthodontics"] = "Br J Orthod",
  ["british journal of pain"] = "Br J Pain",
  ["british journal of perioperative nursing : the journal of the national association of theatre nurses"] = "Br J Perioper Nurs",
  ["british journal of pharmaceutical research"] = "Br J Pharm Res",
  ["british journal of pharmacology"] = "Br J Pharmacol",
  ["british journal of pharmacology and chemotherapy"] = "Br J Pharmacol Chemother",
  ["british journal of physiological optics"] = "Br. J. Physiol. Opt.",
  ["british journal of physiotherapy"] = "Br J Physiother",
  ["british journal of plastic surgery"] = "Br J Plast Surg",
  ["british journal of political science"] = "Br J Polit Sci",
  ["british journal of politics & international relations"] = "Br J Polit Int Relat",
  ["british journal of preventive & social medicine"] = "Br J Prev Soc Med",
  ["british journal of preventive and social medicine"] = "Br. J. Prev. Soc. Med.",
  ["british journal of psychiatric social work"] = "Br J Psychiatr Soc Work",
  ["british journal of psychiatry"] = "Br. J. Psychiatry",
  ["british journal of psychiatry. supplement"] = "Br. J. Psychiatry. Suppl.",
  ["british journal of psychology"] = "Br. J. Psychol.",
  ["british journal of psychology (london, england : 1953)"] = "Br J Psychol",
  ["british journal of psychotherapy"] = "Br J Psychother",
  ["british journal of radiology"] = "Br. J. Radiol.",
  ["british journal of radiology. supplement"] = "Br J Radiol Suppl",
  ["british journal of rheumatology"] = "Br J Rheumatol",
  ["british journal of sexual medicine"] = "Br J Sex Med",
  ["british journal of social and clinical psychology"] = "Br. J. Soc. Clin. Psychol.",
  ["british journal of social medicine"] = "Br J Soc Med",
  ["british journal of social psychology"] = "Br. J. Soc. Psychol.",
  ["british journal of social work"] = "Br J Soc Work",
  ["british journal of sociology"] = "Br. J. Sociol.",
  ["british journal of sociology of education"] = "Br J Sociol Educ",
  ["british journal of sports medicine"] = "Br J Sports Med",
  ["british journal of surgery"] = "Br. J. Surg.",
  ["british journal of theatre nursing"] = "Br. J. Theatre Nurs.",
  ["british journal of urology"] = "Br J Urol",
  ["british journal of venereal diseases"] = "Br. J. Vener. Dis.",
  ["british journal of virology"] = "Br J Virol",
  ["british medical bulletin"] = "Br Med Bull",
  ["british medical journal"] = "Br Med J",
  ["british medical journal (clinical research ed.)"] = "Br Med J (Clin Res Ed)",
  ["british medical journal (clinical research edition)"] = "Br. Med. J. (Clin. Res. Ed.)",
  ["british microbiology research journal"] = "Br Microbiol Res J",
  ["british paramedic journal"] = "Br Paramed J",
  ["british polymer journa"] = "Br. Polym. J.",
  ["british poultry science"] = "Br Poult Sci",
  ["british review of economic issues"] = "Br Rev Econ Issues",
  ["british science news"] = "Br Sci News",
  ["british society for the history of science"] = "British J. Hist. Sci.",
  ["british society for the history of science. british journal for the history of science"] = "British J. Hist. Sci.",
  ["british veterinary journal"] = "Br. Vet. J.",
  ["brjac–brazilian journal of analytical chemistry"] = "BrJAC–Braz. J. Anal. Chem.",
  ["bromley local history"] = "Bromley Local Hist",
  ["bronte society transactions. bronte society"] = "Bronte Soc Trans",
  ["brontë society transactions. brontë society"] = "Bronte Soc Trans",
  ["brookhaven symposia in biology"] = "Brookhaven Symp Biol",
  ["brookings bulletin"] = "Brookings Bull.",
  ["brookings bulletin (washington, d.c. : 1962)"] = "Brookings Bull",
  ["brookings papers on economic activity"] = "Brookings Pap Econ Act",
  ["brookings review"] = "Brookings Rev.",
  ["brookings-wharton papers on financial services"] = "Brookings-Wharton Pap. Finan. Services",
  ["brookings-wharton papers on urban affairs"] = "Brookings-Wharton Pap. Urban Aff.",
  ["brooklyn journal of international law"] = "Brooklyn J Int Law",
  ["brooklyn law review"] = "Brooklyn Law Rev",
  ["broward review"] = "Broward Rev",
  ["brown university digest of addiction theory and application : data"] = "Brown Univ Dig Addict Theory Appl",
  ["brown university psychopharmacology update"] = "Brown Univ. Psychopharmacol. Update",
  ["brucken zur mathematik"] = "Brucken Math.",
  ["bruno pini mathematical analysis seminar"] = "Bruno Pini Math. Anal. Semin.",
  ["bruns' beiträge für klinische chirurgie"] = "Bruns Beitr Klin Chir (1971)",
  ["bruns' beiträge zur klinischen chirurgie"] = "Bruns Beitr Klin Chir",
  ["bruns’ beitrage fur klinische chirurgie"] = "Bruns Beitr. Klin. Chir.",
  ["bruns’ beitrage zur klinischen chirurgie"] = "Bruns Beitr Klinischen Chir",
  ["bruxelles medical"] = "Brux. Med.",
  ["bruxelles médical"] = "Brux Med",
  ["bryn mawr classical review"] = "BMCRev",
  ["bryonora : zpravodaj bryologicko-lichenologické sekce čbs"] = "Bryonora (Praha)",
  ["bryophyte diversity and evolution"] = "Bryophyt Divers Evol",
  ["bryophytorum bibliotheca"] = "Bryophyt Bibl",
  ["bscs pamphlets"] = "BSCS Pam",
  ["bsg proceedings"] = "BSG Proc.",
  ["bsgf earth sciences bulletin"] = "BSGF Earth Sci. Bull.",
  ["bshm bulletin. journal of the british society for the history of mathematics"] = "BSHM Bull.",
  ["bshp new texts in the history of philosophy"] = "BSHP New Texts Hist. Philos.",
  ["bt technology journal"] = "BT Technol. J.",
  ["bts comptable-iut de gestion. enseignement technique superieur"] = "BTS Compt.-IUT Gestion Enseign. Tech. Super.",
  ["btta review"] = "BTTA Rev.",
  ["bubble science engineering and technology"] = "Bubble Sci Eng Technol",
  ["bubble science, engineering & technology"] = "Bubble Sci. Eng. Technol.",
  ["bucherei des augenarztes"] = "Buch. Augenarzt.",
  ["budapest régiségei"] = "BudReg",
  ["budapest statisztikai zsebkönyve"] = "Bp Stat Zsebkv",
  ["budapest székesföváros közegészségügye"] = "Bp Szekesfov Kozeu",
  ["buffalo law review"] = "Buffalo Law Rev",
  ["building acoustics"] = "Build. Acoust.",
  ["building and environment"] = "Build. Environ.",
  ["building operating management"] = "Build Oper Manage",
  ["building research and information"] = "Build. Res. Inf.",
  ["building science"] = "Build. Sci.",
  ["building services engineering research & technology : bser & t"] = "Build Serv Eng Res Technol",
  ["building services engineering research and technology"] = "Build. Serv. Eng. Res. Technol.",
  ["building simulation"] = "Build. Simul.",
  ["building systems design"] = "Build Syst Des",
  ["buildings and cities"] = "Build. Cities",
  ["buildsys'15 : proceedings of the 2nd acm international conference on embedded systems for energy-efficient buildings : november 4-5, 2015, seoul, south korea. acm conference on embedded systems for energy-efficient buildings (2nd : 2015 : seoul, korea)"] = "BuildSys15 (2015)",
  ["built environment"] = "Built Environ",
  ["built environment project and asset management"] = "Built Environ. Project Asset Manage.",
  ["buletin i universitetit shteteror te tiranes. seria shkencat mjekesore"] = "Bul Univ Shtet Tiranes Ser Shk Mjekesore",
  ["buletin i universitetit shtetëror të tiranës. seria shkencat mjekësore"] = "Bul Univ Shtet Tiranes Ser Shk Mjekesore",
  ["buletin keluarga"] = "Bul Keluarga",
  ["buletin lab veteriner"] = "Bul Lab Veteriner",
  ["buletin stiintific. sectiunea de stiinte medicale. academia republicii populare romane"] = "Bul Stiint Sect Stiint Medicale Acad Republicii Pop Romane",
  ["buletin ştiinţific. secţiunea de ştiinţe medicale. academia republicii populare romîne"] = "Bul Stiint Sect Stiint Medicale Acad Republicii Pop Romane",
  ["buletinul academiei de stiinte a republicii moldova. matematica. izvestiya akademii nauk respubliki moldova. matematika"] = "Bul. Acad. Stiinte Repub. Mold. Mat.",
  ["buletinul institutului politehnic din iaşi : secția științe socio-umane"] = "Bul Inst Politeh Lasi Sect Stiinte Socio Umane",
  ["buletinul institutului politehnic din iaşi"] = "Bul. Inst. Politeh. Iaşi. Secţ. I. Mat. Mec. Teor. Fiz.",
  ["buletinul societăţii numismatice române"] = "BSocNumRom",
  ["buletinul stiintific al universitatii “politehnica” din timisoara. seria matematica-fizica. scientific bulletin of the politehnica university of timisoara. transactions on mathematics & physics"] = "Bul. Stiint. Univ. Politeh. Timis. Ser. Mat. Fiz.",
  ["buletinul ştiinţific al institutului politehnic cluj-napoca"] = "Bul. Ştiinţ. Inst. Politehn. Cluj-Napoca Ser. Mat. Apl. Mec.",
  ["buletinul ştiinţific al universităţii din baia mare"] = "Bul. Ştiinţ. Univ. Baia Mare Ser. B",
  ["bulgarian astronomical journal"] = "Bulg. Astron. J.",
  ["bulgarian chemical communications"] = "Bulg. Chem. Commun.",
  ["bulgarian historical review"] = "Bulg Hist Rev",
  ["bulgarian journal of agricultural science"] = "Bulg. J. Agric. Sci.",
  ["bulgarian journal of physics"] = "Bulg. J. Phys.",
  ["bulgarian journal of veterinary medicine"] = "Bulg J Vet Med",
  ["bulletin (alabama museum of natural history)"] = "Bulletin (Ala Mus Nat Hist)",
  ["bulletin (american association for the history of nursing)"] = "Bull Am Assoc Hist Nurs",
  ["bulletin (annuel) de l’école antique de nîmes"] = "BEcAntNimes",
  ["bulletin (british arachnological society)"] = "Bull Br Arachnol Soc",
  ["bulletin (british psychological society)"] = "Bull Br Psychol Soc",
  ["bulletin (eugenics society (london, england))"] = "Bull Eugen Soc",
  ["bulletin (gandhigram institute of rural health and family welfare trust)"] = "Bull Gandhigram Inst Rural Health Fam Welf Trust",
  ["bulletin (hospital for joint diseases (new york, n.y.))"] = "Bull Hosp Jt Dis",
  ["bulletin (institute of medical ethics (great britain))"] = "IME bull",
  ["bulletin (jāmiʻat al-qāhirah. kullīyat al-zirāʻah)"] = "Bull Cairo Univ Fac Agric",
  ["bulletin (medical sciences historical society (great britain))"] = "Bull Med Sci Hist Soc",
  ["bulletin (national clearinghouse for poison control centers (u.s.))"] = "Bull Natl Clgh Poison Control Cent",
  ["bulletin (national medical and dental association (u.s))"] = "Bull Natl Med Dent Assoc Natl Advocates Soc",
  ["bulletin (national medical and dental association (u.s))."] = "Bull Natl Med Dent Assoc Natl Advocates Soc",
  ["bulletin (new series) of the american mathematical society"] = "Bull New Ser Am Math Soc",
  ["bulletin (new york state archeological association : 1987)"] = "Bull J N Y State Archeol Assoc",
  ["bulletin (peabody museum of natural history)"] = "Bull Peabody Mus Nat Hist",
  ["bulletin (societe belge d’etudes napoleoniennes)"] = "Bull Soc Belge Etud Napoleon",
  ["bulletin (société belge d'études napoléoniennes)"] = "Bull Soc Belge Etud Napoleon",
  ["bulletin (société des naturalistes luxembourgeois)"] = "Bull Soc Nat Luxemb",
  ["bulletin (southern california academy of sciences)"] = "Bull South Calif Acad Sci",
  ["bulletin (st. louis park medical center)"] = "Bull St Louis Park Med Cen",
  ["bulletin (university of minnesota. economic development center)"] = "Bull (Univ Minn Econ Dev Cent)",
  ["bulletin (western dental society)"] = "West Dent Soc Bull",
  ["bulletin (western dental society)."] = "West Dent Soc Bull",
  ["bulletin - academy of medicine of cleveland"] = "Bull Acad Med Clevel",
  ["bulletin - academy of medicine of new jersey"] = "Bull Acad Med N J",
  ["bulletin - académie serbe des sciences. classe des sciences mathématiques et naturelles. sciences naturelles"] = "Bull Acad Serbe Sci Cl Sci Math Nat Sci Nat",
  ["bulletin - alameda county dental society"] = "Bull Alameda Cty Dent Soc",
  ["bulletin - american academy of arts and sciences. american academy of arts and sciences"] = "Bull Am Acad Arts Sci",
  ["bulletin - american association of hospital dentists"] = "Bull Am Assoc Hosp Dent",
  ["bulletin - american protestant hospital association"] = "Bull Am Protestant Hosp Assoc",
  ["bulletin - bibliotheque royale albert ier. bibliotheque royale albert ier"] = "Bull Bibl R Albert",
  ["bulletin - bibliothèque royale albert ier. bibliotheque royale albert ier"] = "Bull Bibl R Albert",
  ["bulletin - british association of sport and medicine"] = "Bull Br Assoc Sport Med",
  ["bulletin - british society for middle eastern studies. british society for middle eastern studies"] = "Bull Br Soc Middle East Stud",
  ["bulletin - canadian tuberculosis association"] = "Bull Can Tuberc Assoc",
  ["bulletin - cercle benelux d'histoire de la pharmacie"] = "Bull Cercle Benelux Hist Pharm",
  ["bulletin - cercle benelux d’histoire de la pharmacie"] = "Bull Cercle Benelux Hist Pharm",
  ["bulletin - cincinnati dental society"] = "Bull Cincinnati Dent Soc",
  ["bulletin - cosmos club. cosmos club (washington, d.c.)"] = "Bull Cosm Club",
  ["bulletin - dade county medical association"] = "Bull Dade Cty Med Assoc Inc",
  ["bulletin - dight institute of the university of minnesota"] = "Bull Dight Inst Univ Minn",
  ["bulletin - fulton county medical society"] = "Bull Fulton Cty Med Soc",
  ["bulletin - geisinger medical center"] = "Bull Geisinger",
  ["bulletin - genesee county medical society"] = "Bull Genesee Cty Med Soc",
  ["bulletin - hennepin county medical society"] = "Bull Hennepin Cty Med Soc",
  ["bulletin - institute for medical research, kuala lumpur"] = "Bull Inst Med Res Kuala Lumpur",
  ["bulletin - jackson park branch, chicago medical society"] = "Bull Jackson Park Branch Chic Med Soc",
  ["bulletin - king county medical society"] = "Bull King Cty Med Soc",
  ["bulletin - los angeles county medical association"] = "Bull Los Angel Cty Med Assoc",
  ["bulletin - mount desert island biological laboratory, salisbury cove, maine"] = "Bull Mt Desert Isl Biol Lab Salisb Cove Maine",
  ["bulletin - musees royaux des beaux-arts de belgique. musees royaux des beaux-arts de belgique"] = "Bull Mus Rov Beaux Arts Belg",
  ["bulletin - musées royaux des beaux-arts de belgique. musées royaux des beaux-arts de belgique"] = "Bull Mus Rov Beaux Arts Belg",
  ["bulletin - n. z. society of periodontology"] = "Bull N Z Soc Periodontol",
  ["bulletin - national tuberculosis and respiratory disease association"] = "Bull Natl Tuberc Respir Dis Assoc",
  ["bulletin - national tuberculosis association"] = "Bull Natl Tuberc Assoc",
  ["bulletin - new jersey society of dentistry for children"] = "Bull N J Soc Dent Child",
  ["bulletin - north shore branch, chicago medical society"] = "Bull North Shore Branch Chic Med Soc",
  ["bulletin - northern district dental society"] = "Bull North Dist Dent Soc",
  ["bulletin - office international des epizooties"] = "Bull Off Int Epizoot",
  ["bulletin - office international des épizooties"] = "Bull Off Int Epizoot",
  ["bulletin - pacific coast society of orthodontists"] = "Bull Pac Coast Soc Orthod",
  ["bulletin - san diego county dental society"] = "Bull San Diego Cty Dent Soc",
  ["bulletin - san diego county medical society"] = "Bull S Diego Cty Med Soc",
  ["bulletin - san francisco county medical society"] = "Bull San Franc Cty Med Soc",
  ["bulletin - san mateo county dental society"] = "Bull San Mateo Cty Dent Soc",
  ["bulletin - sangamon county medical society"] = "Bull Sangamon Cty Med Soc",
  ["bulletin - second district dental society"] = "Bull Second Dist Dent Soc",
  ["bulletin - sinai hospital of detroit"] = "Bull Sinai Hosp Detroit",
  ["bulletin - societe francaise d’histoire des hopitaux"] = "Bull Soc Fr Hist Hop",
  ["bulletin - society of vacuum coaters"] = "Bull. Soc. Vac. Coaters",
  ["bulletin - société française d'histoire des hôpitaux"] = "Bull Soc Fr Hist Hop",
  ["bulletin - spokane county medical society"] = "Bull Spokane Cty Med Soc",
  ["bulletin - st. louis dental society"] = "Bull St Louis Dent Soc",
  ["bulletin - suffolk county medical society"] = "Bull Suffolk Cty Med Soc Suffolk Acad Med",
  ["bulletin - tennessee nurses association"] = "Bull Tenn Nurses Assoc",
  ["bulletin - texas nurses association"] = "Bull Tex Nurses Assoc",
  ["bulletin - utica academy of medicine"] = "Bull Utica Acad Med",
  ["bulletin / atlantic-cape may county dental society"] = "Bull Atl Cape May Cty Dent Soc",
  ["bulletin / hospital for joint diseases"] = "Bull Hosp Jt Dis",
  ["bulletin / midtown dental society"] = "Bull Midtown Dent Soc",
  ["bulletin / monmouth-ocean county dental society"] = "Bull Monmouth Ocean Cty Dent Soc",
  ["bulletin [of the] medical society, county of erie [and] the buffalo academy of medicine. medical society of the county of erie (n.y.)"] = "Bull Med Soc Cty Erie Buffalo Acad Med",
  ["bulletin algérien de carcinologie"] = "Bull Alger Carcinol",
  ["bulletin angewandte geologie"] = "Bull. angew. Geol.",
  ["bulletin annuel d'épidemiologie"] = "Bull Annu Epidemiol",
  ["bulletin antieke beschaving"] = "BABesch",
  ["bulletin antieke beschaving. annual papers on classical archaeology"] = "BABesch",
  ["bulletin archéologique de provence"] = "BAProv",
  ["bulletin archéologique du comité des travaux historique"] = "BCTH",
  ["bulletin archéologique du comité des travaux historiques et scientifiques. antiquités nationales"] = "BAParis",
  ["bulletin biologique de la france et de la belgique"] = "Bull Biol Fr Belg",
  ["bulletin bodenkundliche gesellschaft der schweiz"] = "Bull. Bodenkd. Ges. Schweiz",
  ["bulletin canadien d'histoire de la médecine = canadian bulletin of medical history"] = "Bull Can Hist Med",
  ["bulletin d'audiophonologie"] = "Bull Audiophonol",
  ["bulletin d'études orientales"] = "Bull Etud Orient",
  ["bulletin d'histoire de l'électricité"] = "Bull Hist Electr",
  ["bulletin d'histoire politique"] = "Bull Hist Polit",
  ["bulletin d'histologie appliquée à la physiologie et à la pathologie et de technique microscopique"] = "Bull Histol Appl Physiol Pathol Tech Microsc",
  ["bulletin d'histologie appliquée et de technique microscopique"] = "Bull Histol Appl Tech Microsc",
  ["bulletin d'hygiène. health bulletin. montréal (québec). department of health"] = "Bull Hyg",
  ["bulletin d'information - cenaddom. centre national de documentation des départements d'outre-mer"] = "Bull Inf Cenaddom",
  ["bulletin d'information. federation pharmaceutique méditerranéenne"] = "Bull Inform Fed Pharm Mediterr",
  ["bulletin de correspondance hellénique"] = "BCH",
  ["bulletin de documentation sanitaire"] = "Bull Doc Sanit",
  ["bulletin de groupement europeen pour la recher che scientifique en stomatologie & odontologie"] = "Bull Group Eur Rech Sci Stomatol Odontol",
  ["bulletin de groupement européen pour la recher che scientifique en stomatologie & odontologie"] = "Bull Group Eur Rech Sci Stomatol Odontol",
  ["bulletin de l academie nationale de medecine"] = "Bull Acad Natl Med",
  ["bulletin de l association des anatomistes"] = "Bull Assoc Anat (Nancy)",
  ["bulletin de l'academie de médecine"] = "Bull Acad Med",
  ["bulletin de l'académie de chirurgie dentaire"] = "Bull Acad Chir Dent (Paris)",
  ["bulletin de l'académie dentaire"] = "Bull Acad Dent (Paris)",
  ["bulletin de l'académie lorraine des sciences"] = "Bull Acad Soc Lorraines Sci",
  ["bulletin de l'académie nationale de chirurgie dentaire (1983)"] = "Bull Acad Natl Chir Dent",
  ["bulletin de l'académie nationale de médecine"] = "Bull Acad Natl Med",
  ["bulletin de l'académie polonaise des sciences. série des sciences biologiques"] = "Bull Acad Pol Sci Biol",
  ["bulletin de l'académie royale de médecine de belgique"] = "Bull Acad R Med Belg",
  ["bulletin de l'académie vétérinaire de france"] = "Bull Acad Vet Fr",
  ["bulletin de l'association de géographes français"] = "Bull Assoc Geogr Fr",
  ["bulletin de l'association des anatomistes"] = "Bull Assoc Anat (Nancy)",
  ["bulletin de l'association des diplômés de microbiologie de la faculté de pharmacie de nancy"] = "Bull Assoc Diplomes Microbiol Fac Pharm Nancy",
  ["bulletin de l'association des mycologues francophones de belgique"] = "Bull Assoc Mycol Francoph Belg",
  ["bulletin de l'association française pour l'étude du cancer"] = "Bull Assoc Fr Etud Cancer",
  ["bulletin de l'association guillaume budé. association guillaume budé"] = "Bull Assoc Guillaume Bude",
  ["bulletin de l'association tunisienne du planning familial : publication interne"] = "Bull Assoc Tunis Plan Fam",
  ["bulletin de l'ecole française d'extrême-orient"] = "Bull Ec Fr Extr Orient",
  ["bulletin de l'institut d'egypte"] = "Bull Inst Egypte",
  ["bulletin de l'institut fondamental d'afrique noire. série b: sciences humaines"] = "Bull Inst Fond Afr Noire Ser B",
  ["bulletin de l'institut français d'archéologie orientale. institut français d'archéologie orientale du caire"] = "Bull Inst Fr Archeol Orient",
  ["bulletin de l'institut national de la santé et de la recherche médicale"] = "Bull Inst Natl Sante Rech Med",
  ["bulletin de l'institut national des sciences et technologies de la mer"] = "Bull Inst Natl Sci Technol Mer",
  ["bulletin de l'institut océanographique. institut océanographique"] = "Bull Inst Oceanogr",
  ["bulletin de l'institut pasteur"] = "Bull Inst Pasteur",
  ["bulletin de l'ordre des médecins"] = "Bull Ordre Med",
  ["bulletin de la bibliotheque nationale. bibliotheque nationale (france)"] = "Bull Bibl Nat",
  ["bulletin de la bibliothèque nationale. bibliothèque nationale (france)"] = "Bull Bibl Nat",
  ["bulletin de la classe des lettres de l’académie royale de belgiquen"] = "BAB",
  ["bulletin de la classe des lettres et des sciences morales et politiques. academie royale des sciences, des lettres et des beaux-arts de belgique. classe des lettres et des sciences morales et politiques"] = "Bull Cl Lett Sci Morales Polit",
  ["bulletin de la classe des lettres et des sciences morales et politiques. académie royale des sciences, des lettres et des beaux-arts de belgique. classe des lettres et des sciences morales et politiques"] = "Bull Cl Lett Sci Morales Polit",
  ["bulletin de la classe des sciences, academie royale de belgique"] = "Bull. Cl. Sci. Acad. R. Belg.",
  ["bulletin de la faculté des lettres de mulhouse"] = "BFLM",
  ["bulletin de la federation des societes de gynecologie et d obstetrique de langue francaise"] = "Bull. Fed. Soc. Gynecol. Obstet. Lang. Fr.",
  ["bulletin de la federation des societes de gynecologie et dobstetrique de langue francaise"] = "Bull Fed Soc Gynecol Obstet Lang Fr",
  ["bulletin de la fédération des associations mycologiques méditerranéennes"] = "Bull Fed Assoc Mycol Mediterr",
  ["bulletin de la fédération des sociétés de gynécologie et dóbstétrique de langue française"] = "Bull Fed Soc Gynecol Obstet Lang Fr",
  ["bulletin de la murithienne"] = "Bull. Murithienne",
  ["bulletin de la santé publique. belgium. ministère de la santé publique et de la famille"] = "Bull Sante publique (Brux )",
  ["bulletin de la section d'histoire moderne et contemporaine. france. comité des travaux historiques et scientifiques. section d'histoire moderne et contemporaine"] = "Bull Sect Hist Mod Contemp",
  ["bulletin de la section d’histoire moderne et contemporaine. france. comite des travaux historiques et scientifiques. section d’histoire moderne et contemporaine"] = "Bull Sect Hist Mod Contemp",
  ["bulletin de la societe archeologique du finistere. societe archeologique du finistere"] = "Bull Soc Archeol Finistere",
  ["bulletin de la societe archeologique, historique et artistique le vieux papier pour l’etude de la vie et des moeurs d’autrefois"] = "Bull Soc Archeol Hist Artist Vieux Pap",
  ["bulletin de la societe archeologique, historique, litteraire & scientifique du gers. societe archeologique, historique, litteraire et scientifique du gers"] = "Bull Soc Archeol Hist Litt Sci Gers",
  ["bulletin de la societe belge d ophtalmologie"] = "Bull Soc Belge Ophtalmol",
  ["bulletin de la societe belge d’etudes geographiques. tijdschrift van de belgische vereniging voor aardrijkskundige studies. societe belge d’etudes geographiques"] = "Bull Soc Belg Etud Geogr",
  ["bulletin de la societe belge d’ophtalmologie"] = "Bull. Soc. Belge Ophtalmol.",
  ["bulletin de la societe chimique de france"] = "Bull. Soc. Chim. Fr.",
  ["bulletin de la societe de borda (dax, france : 1876)"] = "Bull Soc Borda",
  ["bulletin de la societe de chimie biologique"] = "Bull. Soc. Chim. Biol. (Paris)",
  ["bulletin de la societe de l’histoire du protestantisme francais (1852)"] = "Bull Soc Hist Protestant Fr",
  ["bulletin de la societe de pathologie exotique"] = "Bull. Soc. Pathol. Exot.",
  ["bulletin de la societe de pathologie exotique (1990)"] = "Bull Soc Pathol Exot",
  ["bulletin de la societe de pathologie exotique et de ses filiales"] = "Bull. Soc. Pathol. Exot. Filiales",
  ["bulletin de la societe de pharmacie de bordeaux"] = "Bull Soc Pharm Bord",
  ["bulletin de la societe de pharmacie de strasbourg"] = "Bull Soc Pharm Strasb",
  ["bulletin de la societe des amis des sciences et des lettres de poznan. seria c: medecine"] = "Bull Soc Amis Sci (Med) (Poznan)",
  ["bulletin de la societe des amis du vieux toulon"] = "Bull Soc Amis Vieux Toulon",
  ["bulletin de la societe des etudes oceaniennes (polynesie orientale)"] = "Bull Soc Etud Oceaniennes",
  ["bulletin de la societe des lettres, sciences et arts de la correze. societe des lettres, sciences et arts de la correze"] = "Bull Soc Lett Sci Arts Correze",
  ["bulletin de la societe des sciences et des lettres de lodz. serie: recherches sur les deformations"] = "Bull. Soc. Sci. Lett. Lodz Ser. Rech. Deform.",
  ["bulletin de la societe des sciences medicales du grand-duche de luxembourg"] = "Bull Soc Sci Med Grand Duche Luxemb",
  ["bulletin de la societe d’agriculture, sciences et arts de la sarthe. societe d’agriculture, sciences et arts de la sarthe"] = "Bull Soc Agric Sci Arts Sarthe",
  ["bulletin de la societe d’etudes des hautes-alpes. societe d’etudes des hautes-alpes"] = "Bull Soc Etud His Alpes",
  ["bulletin de la societe d’histoire moderne. societe d’histoire moderne (paris, france)"] = "Bull Soc Hist Mod",
  ["bulletin de la societe francaise de dermatologie et de syphiligraphie"] = "Bull. Soc. Fr. Dermatol. Syphiligr.",
  ["bulletin de la societe francaise de physique"] = "Bull. Soc. Fr. Phys.",
  ["bulletin de la societe geologique de france"] = "Bull. Soc. Geol. Fr.",
  ["bulletin de la societe historique et archeologique du perigord. societe historique et archeologique du perigord"] = "Bull Soc Hist Archeol Perigord",
  ["bulletin de la societe industrielle de mulhouse"] = "Bull Soc Ind Mulhouse",
  ["bulletin de la societe internationale de chirurgie"] = "Bull. Soc. Int. Chir.",
  ["bulletin de la societe languedocienne de geographie"] = "Bull Soc Languedoc Geogr",
  ["bulletin de la societe libanaise d’histoire de la medecine"] = "Bull Soc Liban Hist Med",
  ["bulletin de la societe mathematique de france"] = "Bull. Soc. Math. Fr.",
  ["bulletin de la societe mathematique de france. supplement memoire"] = "Bull. Soc. Math. France Suppl. Mem.",
  ["bulletin de la societe medicale d’afrique noire de langue francaise"] = "Bull. Soc. Med. Afr. Noire Lang. Fr.",
  ["bulletin de la societe nationale des antiquaires de france. societe nationale des antiquaires de france"] = "Bull Soc Nat Antiq Fr",
  ["bulletin de la societe royale belge de gynecologie et d’obstetrique"] = "Bull. Soc. R. Belge Gynecol. Obstet.",
  ["bulletin de la societé linnéenne de provence"] = "Bull Soc Linn Provence",
  ["bulletin de la société archéologique du finistère. société archéologique du finistère"] = "Bull Soc Archeol Finistere",
  ["bulletin de la société archéologique et historique du limousin"] = "Bull Soc Archeol Hist Limousin",
  ["bulletin de la société archéologique, historique et artistique le vieux papier pour l'étude de la vie et des mœurs d'autrefois"] = "Bull Soc Archeol Hist Artist Vieux Pap",
  ["bulletin de la société archéologique, historique, littéraire & scientifique du gers. société archéologique, historique, littéraire et scientifique du gers"] = "Bull Soc Archeol Hist Litt Sci Gers",
  ["bulletin de la société belge d'études géographiques. tijdschrift van de belgische vereniging voor aardrijkskundige studies. société belge d'études géographiques"] = "Bull Soc Belg Etud Geogr",
  ["bulletin de la société belge d'ophtalmologie"] = "Bull Soc Belge Ophtalmol",
  ["bulletin de la société chimique de france"] = "Bull Soc Chim Fr",
  ["bulletin de la société d'agriculture, sciences et arts de la sarthe. société d'agriculture, sciences et arts de la sarthe"] = "Bull Soc Agric Sci Arts Sarthe",
  ["bulletin de la société d'études des hautes-alpes. société d'études des hautes-alpes"] = "Bull Soc Etud His Alpes",
  ["bulletin de la société d'histoire moderne. société d'histoire moderne (paris, france)"] = "Bull Soc Hist Mod",
  ["bulletin de la société d'ophtalmologie de paris et des sociétés d'ophtalmologie de l'est, de lyon et de l'ouest. société d'ophtalmologie de paris"] = "Bull Soc Ophtalmol Paris Soc Ophtalmol Est Lyon Ouest",
  ["bulletin de la société de borda (dax, france : 1876)"] = "Bull Soc Borda",
  ["bulletin de la société de chimie biologique"] = "Bull Soc Chim Biol (Paris)",
  ["bulletin de la société de l'histoire du protestantisme français (1852)"] = "Bull Soc Hist Protestant Fr",
  ["bulletin de la société de pathologie exotique (1990)"] = "Bull Soc Pathol Exot",
  ["bulletin de la société de pathologie exotique et de ses filiales"] = "Bull Soc Pathol Exot Filiales",
  ["bulletin de la société de pharmacie de bordeaux"] = "Bull Soc Pharm Bord",
  ["bulletin de la société de pharmacie de strasbourg"] = "Bull Soc Pharm Strasb",
  ["bulletin de la société des amis des sciences et des lettres de poznań. séria c: médecine"] = "Bull Soc Amis Sci (Med) (Poznan)",
  ["bulletin de la société des amis du vieux toulon"] = "Bull Soc Amis Vieux Toulon",
  ["bulletin de la société des études océaniennes (polynésie orientale)"] = "Bull Soc Etud Oceaniennes",
  ["bulletin de la société des lettres, sciences et arts de la corrèze. société des lettres, sciences et arts de la corrèze"] = "Bull Soc Lett Sci Arts Correze",
  ["bulletin de la société des sciences médicales du grand-duché de luxembourg"] = "Bull Soc Sci Med Grand Duche Luxemb",
  ["bulletin de la société des sciences naturelles de l'ouest de la france (1983)"] = "Bull Soc Sci Nat Ouest Fr",
  ["bulletin de la société des sciences naturelles de l'ouest de la france. société des sciences naturelles de l'ouest de la france"] = "Bull Soc Sci Nat Ouest Fr",
  ["bulletin de la société entomologique de france"] = "Bull Soc Entomol France",
  ["bulletin de la société entomologique de mulhouse"] = "Bull Soc Entomol Mulhouse",
  ["bulletin de la société française de dermatologie et de syphiligraphie"] = "Bull Soc Fr Dermatol Syphiligr",
  ["bulletin de la société historique et archéologique du périgord. société historique et archéologique du périgord"] = "Bull Soc Hist Archeol Perigord",
  ["bulletin de la société historique et scientifique des deux-sévres"] = "Bull Soc hist sci Deux Sevres",
  ["bulletin de la société industrielle de mulhouse"] = "Bull Soc Ind Mulhouse",
  ["bulletin de la société internationale de chirurgie"] = "Bull Soc Int Chir",
  ["bulletin de la société languedocienne de géographie"] = "Bull Soc Languedoc Geogr",
  ["bulletin de la société médicale d'afrique noire de langue française"] = "Bull Soc Med Afr Noire Lang Fr",
  ["bulletin de la société médicale de pau"] = "Bull Soc Med Pau",
  ["bulletin de la société nationale des antiquaires de france. société nationale des antiquaires de france"] = "Bull Soc Nat Antiq Fr",
  ["bulletin de la société royale belge de gynécologie et d'obstétrique"] = "Bull Soc R Belge Gynecol Obstet",
  ["bulletin de la société archéologique champenoise"] = "BSocAChamp",
  ["bulletin de la société botanique de france"] = "Bull. Soc. Bot. Fr.",
  ["bulletin de la société de linguistique de paris"] = "BSL",
  ["bulletin de la société des sciences et des lettres de łódź"] = "Bull. Soc. Sci. Lett. Łódź Sér. Rech. Déform.",
  ["bulletin de la société d’archéologie copte"] = "BACopt",
  ["bulletin de la société entomologique de france"] = "Bull. Soc. entomol. Fr.",
  ["bulletin de la société entomologique suisse"] = "Bull. soc. entomol. Suisse",
  ["bulletin de la société française de numismatique"] = "BSFN",
  ["bulletin de la société française de philosophie"] = "BSPh",
  ["bulletin de la société française de photogrammétrie et télédétection"] = "Bull. Soc. fr. photogramm. télédétect.",
  ["bulletin de la société française d’égyptologie"] = "BSFE",
  ["bulletin de la société fribourgeoise des sciences naturelles"] = "Bull. soc. fribg. sci. nat.",
  ["bulletin de la société géologique de france"] = "Bull. Soc. Géol. Fr.",
  ["bulletin de la société internationale pour l’étude de la philosophie médiévale"] = "BPHM",
  ["bulletin de la société libanaise d'histoire de la médecine"] = "Bull Soc Liban Hist Med",
  ["bulletin de la société mathématique de france"] = "Bull. Soc. Math. France",
  ["bulletin de la société mycologique de france"] = "Bull. Soc. mycol. Fr.",
  ["bulletin de la société nationale des antiquaires de france"] = "BSAF",
  ["bulletin de la société neuchâteloise des sciences naturelles"] = "Bull. soc. neuchâtel. sci. nat.",
  ["bulletin de la société royale des sciences de liège"] = "Bull. Soc. Roy. Sci. Liège",
  ["bulletin de la société royale forestière de belgique"] = "Bull. Soc. r. for. Belg.",
  ["bulletin de la société toulousaine d’études classiques"] = "BSTEC",
  ["bulletin de la société vaudoise des sciences naturelles"] = "Bull. Soc. Vaud. sci. nat.",
  ["bulletin de liaison de demographie africaine"] = "Bull Liaison Demogr Afr",
  ["bulletin de liaison de démographie africaine"] = "Bull Liaison Demogr Afr",
  ["bulletin de liaison de la société des amis de la bibliothèque salomon reinach"] = "BSocBiblReinach",
  ["bulletin de liaison. centre d’études des peintures murales romaines"] = "BPeintRom",
  ["bulletin de littérature ecclésiastique"] = "BLE",
  ["bulletin de l’academie de chirurgie dentaire"] = "Bull. Acad. Chir. Dent. (Paris)",
  ["bulletin de l’academie dentaire"] = "Bull. Acad. Dent. (Paris)",
  ["bulletin de l’academie lorraine des sciences"] = "Bull Acad Soc Lorraines Sci",
  ["bulletin de l’academie nationale de chirurgie dentaire"] = "Bull. Acad. Natl. Chir. Dent.",
  ["bulletin de l’academie nationale de chirurgie dentaire (1983)"] = "Bull Acad Natl Chir Dent",
  ["bulletin de l’academie nationale de medecine"] = "Bull. Acad. Natl. Med.",
  ["bulletin de l’academie polonaise des sciences. serie des sciences biologiques"] = "Bull. Acad. Pol. Sci. [Biol.]",
  ["bulletin de l’academie royale de medecine de belgique"] = "Bull. Acad. R. Med. Belg.",
  ["bulletin de l’academie veterinaire de france"] = "Bull. Acad. Vet. Fr.",
  ["bulletin de l’association de geographes francais"] = "Bull Assoc Geogr Fr",
  ["bulletin de l’association des anatomistes"] = "Bull. Assoc. Anat. (Nancy)",
  ["bulletin de l’association des diplomes de microbiologie de la faculte de pharmacie de nancy"] = "Bull Assoc Diplomes Microbiol Fac Pharm Nancy",
  ["bulletin de l’association francaise pour l’etude du cancer"] = "Bull. Assoc. Fr. Etud. Cancer",
  ["bulletin de l’association guillaume bude. association guillaume bude"] = "Bull Assoc Guillaume Bude",
  ["bulletin de l’association guillaume budé"] = "BAGB",
  ["bulletin de l’association pour la défense des intérêts du jura"] = "Bull. Assoc. Déf. Intérêts Jura",
  ["bulletin de l’association pro aventico"] = "BProAvent",
  ["bulletin de l’association tunisienne du planning familial : publication interne"] = "Bull Assoc Tunis Plan Fam",
  ["bulletin de l’ecole francaise d’extreme-orient"] = "Bull Ec Fr Extr Orient",
  ["bulletin de l’institut d’egypte"] = "Bull Inst Egypte",
  ["bulletin de l’institut fondamental d’afrique noire. serie b: sciences humaines"] = "Bull Inst Fond Afr Noire Ser B",
  ["bulletin de l’institut francais d’archeologie orientale. institut francais d’archeologie orientale du caire"] = "Bull Inst Fr Archeol Orient",
  ["bulletin de l’institut français d’archéologie orientale"] = "BIAO",
  ["bulletin de l’institut historique belge de rome"] = "BIBR",
  ["bulletin de l’institut national de la sante et de la recherche medicale"] = "Bull. Inst. Natl. Sante Rech. Med.",
  ["bulletin de l’institut pasteur"] = "Bull. Inst. Pasteur",
  ["bulletin de l’office international des epizooties"] = "Bull. Off. Int. Epizoot.",
  ["bulletin de medecine legale, urgence medicale, centre anti-poisons"] = "Bull Med Leg Urgence Med Cent Antipoisons",
  ["bulletin de methodologie sociologique : bms"] = "Bull Methodol Sociol",
  ["bulletin de médecine légale, urgence médicale, centre anti-poisons"] = "Bull Med Leg Urgence Med Cent Antipoisons",
  ["bulletin de méthodologie sociologique : bms"] = "Bull Methodol Sociol",
  ["bulletin de microscopie appliquee"] = "Bull Micr Appl",
  ["bulletin de microscopie appliquée"] = "Bull Micr Appl",
  ["bulletin de philosophie medievale"] = "Bull Philos Mediev",
  ["bulletin de philosophie médiévale"] = "Bull Philos Mediev",
  ["bulletin de physio-pathologie respiratoire"] = "Bull Physiopathol Respir (Nancy)",
  ["bulletin de presse - santé de la france"] = "Bull Presse Sante Fr",
  ["bulletin de statistique / belgium institut national de statistique"] = "Bull Stat Inst Natl Stat",
  ["bulletin de théologie ancienne et médiévale"] = "BTh",
  ["bulletin de vulgarisation forestière"] = "Bull. vulg. for.",
  ["bulletin der schweizerischen akademie der medizinischen wissenschaften"] = "Bull Schweiz Akad Med Wiss",
  ["bulletin des actualités médicales"] = "Bull Actual Med",
  ["bulletin des antiquités luxembourgeoises"] = "BAL",
  ["bulletin des etudes portugaises"] = "Bull Etud Port Bres",
  ["bulletin des etudes valeryennes"] = "Bull Etud Valeryennes",
  ["bulletin des études portugaises"] = "Bull Etud Port Bres",
  ["bulletin des études valéryennes"] = "Bull Etud Valeryennes",
  ["bulletin des infirmieres catholiques du canada"] = "Bull. Infirm. Cathol. Can.",
  ["bulletin des infirmières catholiques du canada"] = "Bull Infirm Cathol Can",
  ["bulletin des leo baeck instituts. leo baeck institute"] = "Bull Leo Baeck Inst",
  ["bulletin des musées royaux d’art et d’histoire"] = "BMAH",
  ["bulletin des musées royaux d’art et d’histoire, bruxelles"] = "BMusBrux",
  ["bulletin des recherches agronomiques de gembloux"] = "Bull. rech. agron. Gembloux",
  ["bulletin des sciences mathematiques"] = "Bull. Sci. Math.",
  ["bulletin des sciences mathématiques"] = "Bull. Sci. Math.",
  ["bulletin des seances- academie royale des sciences d’outre-mer"] = "Bull Seances Acad R Sci Outre Mer",
  ["bulletin des séances- académie royale des sciences d'outre-mer"] = "Bull Seances Acad R Sci Outre Mer",
  ["bulletin des societes chimiques belges"] = "Bull. Soc. Chim. Belg.",
  ["bulletin des societes d’ophtalmologie de france"] = "Bull. Soc. Ophtalmol. Fr.",
  ["bulletin des sociétés d'ophtalmologie de france"] = "Bull Soc Ophtalmol Fr",
  ["bulletin des travaux de l’institut national du patrimoine. comptes rendus"] = "BTravTun",
  ["bulletin du c. e. r. p"] = "Bull Cerp",
  ["bulletin du cancer"] = "Bull Cancer",
  ["bulletin du cancer. radiotherapie"] = "Bull. Cancer. Radiother.",
  ["bulletin du cancer. radiotherapie : journal de la societe francaise du cancer : organe de la societe francaise de radiotherapie oncologique"] = "Bull Cancer Radiother",
  ["bulletin du cancer. radiothérapie : journal de la société française du cancer : organe de la société française de radiothérapie oncologique"] = "Bull Cancer Radiother",
  ["bulletin du centre d'histoire économique et sociale de la région lyonnaise. centre d'histoire économique et sociale de la région lyonnaise"] = "Bull Cent Hist Econ Soc Reg Lyon",
  ["bulletin du centre de recherches et d’enseignement de l’antiquité, angers"] = "BAngers",
  ["bulletin du centre d’histoire economique et sociale de la region lyonnaise. centre d’histoire economique et sociale de la region lyonnaise"] = "Bull Cent Hist Econ Soc Reg Lyon",
  ["bulletin du centre international d’étude des textiles anciens"] = "BTextilAnc",
  ["bulletin du cercle d’études numismatiques"] = "BCercleNum",
  ["bulletin du cercle vaudois de botanique"] = "Bull Cercle Vaud Bot",
  ["bulletin du groupement europeen pour la recherche scientifique en stomatologie et odontologie"] = "Bull. Group. Eur. Rech. Sci. Stomatol. Odontol.",
  ["bulletin du groupement international pour la recherche scientifique en stomatologie"] = "Bull Group Int Rech Sci Stomatol",
  ["bulletin du groupement international pour la recherche scientifique en stomatologie & odontologie"] = "Bull Group Int Rech Sci Stomatol Odontol",
  ["bulletin du groupement international pour la recherche scientifique en stomatologie et odontologie"] = "Bull. Group. Int. Rech. Sci. Stomatol. Odontol.",
  ["bulletin du groupèment international pour la recherche scientifique en stomatologie & odontologie"] = "Bull Group Int Rech Sci Stomatol Odontol",
  ["bulletin du laboratoire du musée du louvre"] = "BLaborMusLouvre",
  ["bulletin du musée hongrois des beaux-arts = o.m. szépművészeti múzeum közleményei"] = "Bull Mus Hong B Arts",
  ["bulletin du musée de beyrouth"] = "BMB",
  ["bulletin du musée d’anthropologie préhistorique de monaco"] = "BMusMonaco",
  ["bulletin du musée hongrois des beaux-arts"] = "BMusHongr",
  ["bulletin du musée national de varsovie"] = "BMusVars",
  ["bulletin du statec. luxembourg. service central de la statistique et des etudes economiques"] = "Bull Statec",
  ["bulletin du statec. luxembourg. service central de la statistique et des études économiques"] = "Bull Statec",
  ["bulletin d’archéologie algérienne"] = "BAA",
  ["bulletin d’archéologie luxembourgeoise"] = "BALux",
  ["bulletin d’archéologie marocaine"] = "BAM",
  ["bulletin d’archéologie sud-est européenne"] = "BASudEstEur",
  ["bulletin d’audiophonologie"] = "Bull Audiophonol",
  ["bulletin d’etudes orientales / institut francais de damas"] = "Bull Etud Orient",
  ["bulletin d’histoire de l’electricite"] = "Bull Hist Electr",
  ["bulletin d’histoire politique"] = "Bull Hist Polit",
  ["bulletin d’hygiene. health bulletin. montreal (quebec). dept. of health"] = "Bull Hyg",
  ["bulletin d’information de l’association internationale pour l’étude de la mosaïque antique"] = "BullAIEMA",
  ["bulletin d’information office national des forêts"] = "Bull. inf. Off. natl. for.",
  ["bulletin d’information. federation pharmaceutique mediterraneenne"] = "Bull Inform Fed Pharm Mediterr",
  ["bulletin d’informations techniques"] = "Bull. inf. tech.",
  ["bulletin d’études orientales"] = "Bull. études Orient.",
  ["bulletin d’études orientales, publié par l’institut français de damas"] = "BEO",
  ["bulletin d’études préhistoriques alpines"] = "BPrehistAlp",
  ["bulletin economique et social du maroc"] = "Bull Econ Soc Maroc",
  ["bulletin epidémiologique hebdomadaire (paris, france)"] = "Bull Epidemiol Hebd (Paris)",
  ["bulletin et memoires de l academie royale de medecine de belgique"] = "Bull Mem Acad R Med Belg",
  ["bulletin et memoires de la societe des chirurgiens de paris"] = "Bull. Mem. Soc. Chir. Paris",
  ["bulletin et memoires de l’academie royale de medecine de belgique"] = "Bull. Mem. Acad. R. Med. Belg.",
  ["bulletin et mémoires de l'académie royale de médecine de belgique"] = "Bull Mem Acad R Med Belg",
  ["bulletin et mémoires de la société des chirurgiens de paris"] = "Bull Mem Soc Chir Paris",
  ["bulletin et mémoires de la société médicale de passy (haute-savoie)"] = "Bull Mem Soc Med Passy Ht.-Savoie",
  ["bulletin et mémoires. international society of the history of medicine"] = "Bull Soc Int His Med",
  ["bulletin et mémoires. société de médecine de paris"] = "Bull Mem Soc Med Paris (1907)",
  ["bulletin europeen de physiopathologie respiratoire"] = "Bull. Eur. Physiopathol. Respir.",
  ["bulletin européen de physiopathologie respiratoire"] = "Bull Eur Physiopathol Respir",
  ["bulletin économique et social du maroc"] = "Bull Econ Soc Maroc",
  ["bulletin for international fiscal documentation"] = "Bull. Int. Fiscal Doc.",
  ["bulletin for the council for british research in the levant\tbull. counc. br. res. levant"] = "Bull. Counc. Br. Res. Levant",
  ["bulletin for the history of chemistry"] = "Bull. Hist. Chem.",
  ["bulletin forestry commission"] = "Bull. For. Comm.",
  ["bulletin geodesique"] = "Bull. Geodesique",
  ["bulletin geological society of america"] = "Bull. Geol. Soc. Am.",
  ["bulletin geological survey of canada"] = "Bull. Geol. Surv. Can.",
  ["bulletin hispanique"] = "Bull Hisp",
  ["bulletin historique et scientifique de l'auvergne"] = "Bull His Sci Auvergne",
  ["bulletin historique et scientifique de l’auvergne"] = "Bull His Sci Auvergne",
  ["bulletin institute of mining and metallurgy"] = "Bull. Inst. Min. Metall.",
  ["bulletin international de l academie polonaise des sciences et des lettres, classe de medecine"] = "Bull Int Acad Pol Sci Let Cl Med",
  ["bulletin international de l'académie polonaise des sciences et des lettres. classe de médecine"] = "Bull Int Acad Pol Sci Let Cl Med",
  ["bulletin international des services de sante des armees de terre, de mer et de l’air"] = "Bull Int Serv Sante Armees Terre Mer Air",
  ["bulletin international des services de santé des armées de terre, de mer et de l'air"] = "Bull Int Serv Sante Armees Terre Mer Air",
  ["bulletin mathematical association of india"] = "Bull. Math. Assoc. India",
  ["bulletin mathematique de la societe des sciences mathematiques de roumanie. nouvelle serie"] = "Bull. Math. Soc. Sci. Math. Roumanie (N.S.)",
  ["bulletin medecine legale, toxicologie"] = "Bull Med Leg Toxicol",
  ["bulletin medical"] = "Bull Med",
  ["bulletin medical. revue medicale francaise. guide du praticien"] = "Bull Med Rev Med Fr Guide Prat",
  ["bulletin mensuel - societe de medecine militaire francaise"] = "Bull Mens Soc Med Mil Fr",
  ["bulletin mensuel - société de médecine militaire française"] = "Bull Mens Soc Med Mil Fr",
  ["bulletin mensuel de la societe veterinaire pratique de france"] = "Bull Mens Soc Vet Prat Fr",
  ["bulletin mensuel de la société linnéenne de lyon"] = "Bull Mens Soc Linn Lyon",
  ["bulletin mensuel de la société vétérinaire pratique de france"] = "Bull Mens Soc Vet Prat Fr",
  ["bulletin mensuel de la société linnéenne de lyon"] = "Bull. Mens. Soc. Linn. Lyon",
  ["bulletin mensuel, societe de medecine militaire francaise"] = "Bull. Mens. Soc. Med. Mil. Fr.",
  ["bulletin mensuel. office international d'hygiène publique"] = "Bull Mens Off Int Hyg Publique",
  ["bulletin médecine légale, toxicologie"] = "Bull Med Leg Toxicol",
  ["bulletin médical"] = "Bull Med",
  ["bulletin médical. revue médicale française. guide du praticien"] = "Bull Med Rev Med Fr Guide Prat",
  ["bulletin médical. shanghai. université l'aurore"] = "Bull Med Univ Aurore",
  ["bulletin monumental"] = "BMon",
  ["bulletin municipal officiel de la ville de marseille"] = "Bull Munic Off Ville Mars",
  ["bulletin mycologique et botanique dauphiné-savoie"] = "Bull Mycol Bot Dauphine Savoie",
  ["bulletin norges geologiske undersoekels"] = "Bull. Nor. Geol. Unders.",
  ["bulletin oepp. eppo bulletin. european and mediterranean plant protection organisation"] = "Bull OEPP",
  ["bulletin of academy of dentistry for the handicapped"] = "Bull Acad Dent Handicap",
  ["bulletin of allergy"] = "Bull Allergy",
  ["bulletin of alloy phase diagrams"] = "Bull. Alloy Phase Diagr.",
  ["bulletin of anesthesia history"] = "Bull Anesth Hist",
  ["bulletin of animal behaviour"] = "Bull. Anim. Behav.",
  ["bulletin of animal health and production in africa"] = "Bull. Anim. Health Prod. Afr.",
  ["bulletin of animal health and production in africa. bulletin des sante et production animales en afrique"] = "Bull Anim Health Prod Afr",
  ["bulletin of animal health and production in africa. bulletin des santé et production animales en afrique"] = "Bull Anim Health Prod Afr",
  ["bulletin of atmospheric science and technology"] = "Bull. Atmos. Sci. Technol.",
  ["bulletin of bibliography (westwood, mass. : 1979)"] = "Bull Bibliogr",
  ["bulletin of biomathematics"] = "Bull. Biomath.",
  ["bulletin of brewing science"] = "Bull. Brew. Sci.",
  ["bulletin of british byzantine studies"] = "BBrByzSt",
  ["bulletin of canadian energy geoscience"] = "Bull. Can. Energy Geosci.",
  ["bulletin of canadian petroleum geology"] = "Bull. Can. Petrol. Geol.",
  ["bulletin of carnegie museum of natural history"] = "Bull. Carnegie Mus. Nat. Hist.",
  ["bulletin of central research institute fukuoka university"] = "Bull. Central Res. Inst. Fukuoka Univ.",
  ["bulletin of clinical neurosciences"] = "Bull Clin Neurosci",
  ["bulletin of communicable diseases and medical notes. united nations relief and rehabilitation administration"] = "Bull Commun Dis Med Notes",
  ["bulletin of computational applied mathematics"] = "Bull. Comput. Appl. Math.",
  ["bulletin of concerned asian scholars"] = "Bull Concern Asian Sch",
  ["bulletin of czechoslovak law"] = "Bull Czechoslovak Law",
  ["bulletin of department of general education"] = "Bull. Dept. Gen. Ed. College Sci. Tech. Nihon Univ.",
  ["bulletin of earthquake engineering"] = "Bull. Earthquake Eng.",
  ["bulletin of eastern-siberian scientific center of siberian branch of russian academy of medical sciences"] = "Bull East Sib Sci Cent Sib Branch Russ Acad Med Sci",
  ["bulletin of ecers"] = "Bull. ECerS",
  ["bulletin of economic research"] = "Bull Econ Res",
  ["bulletin of electrochemistry"] = "Bull. Electrochem.",
  ["bulletin of emergency and trauma"] = "Bull Emerg Trauma",
  ["bulletin of endemic diseases"] = "Bull Endem Dis (Baghdad)",
  ["bulletin of engineering geology and the environment"] = "Bull. Eng. Geol. Environ.",
  ["bulletin of entomological research"] = "Bull Entomol Res",
  ["bulletin of environment contamination and toxicology"] = "Bull. Environ. Contam. Toxicol.",
  ["bulletin of environmental contamination and toxicology"] = "Bull Environ Contam Toxicol",
  ["bulletin of epizootic diseases of africa"] = "Bull. Epizoot. Dis. Afr.",
  ["bulletin of epizootic diseases of africa. bulletin des epizooties en afrique"] = "Bull Epizoot Dis Afr",
  ["bulletin of epizootic diseases of africa. bulletin des épizooties en afrique"] = "Bull Epizoot Dis Afr",
  ["bulletin of experimental biology and medicine"] = "Bull Exp Biol Med",
  ["bulletin of faculty of pharmacy, cairo university"] = "Bull. Fac. Pharm. Cairo Univ.",
  ["bulletin of fish biology"] = "Bull Fish Biol",
  ["bulletin of francophone africa"] = "Bull Francoph Afr",
  ["bulletin of francophone africa / maghreb research group"] = "Bull Francoph Afr",
  ["bulletin of fukuoka university of education"] = "Bull. Fukuoka Univ. Ed. III",
  ["bulletin of gastrointestinal endoscopy"] = "Bull Gastrointest Endosc",
  ["bulletin of general education"] = "Bull. Gen. Ed. Dokkyo Univ. School Medicine",
  ["bulletin of geosciences"] = "Bull Geosci",
  ["bulletin of glaciological research"] = "Bull. Glaciol. Res.",
  ["bulletin of gulistan state university"] = "Bull Gulistan State Univ",
  ["bulletin of haffkine institute"] = "Bull Haffkine",
  ["bulletin of hygiene"] = "Bull Hyg (Lond)",
  ["bulletin of indonesian economic studies"] = "Bull Indones Econ Stud",
  ["bulletin of informatics and cybernetics"] = "Bull. Inform. Cybernet.",
  ["bulletin of insectology"] = "Bull Insectology",
  ["bulletin of international mathematical virtual institute"] = "Bull. Int. Math. Virtual Inst.",
  ["bulletin of islamic medicine"] = "Bull Islam Med",
  ["bulletin of islamic medicine / kuwait ministry of public health [and] national council of culture, arts and letters"] = "Bull Islam Med",
  ["bulletin of japan society of coordination chemistry"] = "Bull Jpn Soc Coord Chem",
  ["bulletin of kerala mathematics association"] = "Bull. Kerala Math. Assoc.",
  ["bulletin of koshien university"] = "Bull. Koshien Univ. B",
  ["bulletin of latin american research"] = "Bull Lat Am Res",
  ["bulletin of marine science"] = "Bull. Mar. Sci.",
  ["bulletin of materials science"] = "Bull. Mater. Sci.",
  ["bulletin of mathematical analysis and applications"] = "Bull. Math. Anal. Appl.",
  ["bulletin of mathematical biology"] = "Bull Math Biol",
  ["bulletin of mathematical biology. a journal devoted to research at the interface of the life and mathematical sciences"] = "Bull. Math. Biol.",
  ["bulletin of mathematical biophysics"] = "Bull. Math. Biophys.",
  ["bulletin of mathematical sciences"] = "Bull Math Sci",
  ["bulletin of medical ethics"] = "Bull Med Ethics",
  ["bulletin of military clinical psychologists"] = "Bull Mil Clin Psychol",
  ["bulletin of molecular biology and medicine"] = "Bull Mol Biol Med",
  ["bulletin of nagoya institute of technology"] = "Bull. Nagoya Inst. Tech.",
  ["bulletin of nara university of education"] = "Bull. Nara Univ. Ed. Natur. Sci.",
  ["bulletin of nara university of education. natural science"] = "Bull. Nara Univ. Ed. Natur. Sci.",
  ["bulletin of osaka prefectural college of technology"] = "Bull. Osaka Prefect. College Tech.",
  ["bulletin of osaka prefectural university college of technology"] = "Bull. Osaka Prefect. Univ. Coll. Technol.",
  ["bulletin of osaka prefecture university"] = "Bull. Osaka Prefect. Univ. Ser. A Eng. Nat. Sci.",
  ["bulletin of osaka prefecture university. series b, agriculture and life sciences"] = "Bull Osaka Prefect Univ Ser B Agric Life Sci",
  ["bulletin of peace proposals"] = "Bull Peace Propos",
  ["bulletin of pharmaceutical research institute"] = "Bull Pharm Res Inst",
  ["bulletin of polish medical science and history"] = "Bull. Pol. Med. Sci. Hist.",
  ["bulletin of postgraduate institute of medical education and research, chandigarh"] = "Bull Postgrad Inst Med Educ Res Chandigarh",
  ["bulletin of practical ophthalmology"] = "Bull Pract Ophthalmol",
  ["bulletin of prosthetics research"] = "Bull Prosthet Res",
  ["bulletin of pure & applied sciences"] = "Bull. Pure Appl. Sci. Sect. E Math. Stat.",
  ["bulletin of pure and applied mathematics"] = "Bull. Pure Appl. Math.",
  ["bulletin of reprints"] = "Bull Repr",
  ["bulletin of research in the humanities"] = "Bull Res Humanit",
  ["bulletin of science and engineering research laboratory"] = "Bull. Sci. Eng. Res. Lab. Waseda Univ.",
  ["bulletin of science, technology & society"] = "Bull Sci Technol Soc",
  ["bulletin of society of mathematicians"] = "Bull. Soc. Math. Banja Luka",
  ["bulletin of symbolic logic"] = "Bull. Symb. Log.",
  ["bulletin of the 8th district dental society"] = "Bull Eighth Dist Dent Soc",
  ["bulletin of the aas"] = "Bull. AAS",
  ["bulletin of the academy of general dentistry"] = "Bull. Acad. Gen. Dent.",
  ["bulletin of the academy of medicine, toronto"] = "Bull Acad Med Tor",
  ["bulletin of the academy of sciences of the ussr, physical series [translation of izvestiya akademii nauk sssr, seriya fizicheskaya]"] = "Bull. Acad. Sci. USSR, Phys. Ser.",
  ["bulletin of the akita research institute of food and brewing"] = "Bull Akita Res Inst Food Brew",
  ["bulletin of the akron dental society"] = "Bull Akron Dent Soc",
  ["bulletin of the allahabad mathematical society"] = "Bull. Allahabad Math. Soc.",
  ["bulletin of the allyn museum"] = "Bull Allyn Mus",
  ["bulletin of the american academy of psychiatry and the law"] = "Bull. Am. Acad. Psychiatry Law",
  ["bulletin of the american astronomical society"] = "Bull. Am. Astron. Soc.",
  ["bulletin of the american college of nurse-midwifery"] = "Bull Am Coll Nurse Midwifery",
  ["bulletin of the american college of nurse-midwives"] = "Bull Am Coll Nurse Midwives",
  ["bulletin of the american college of physicians"] = "Bull. Am. Coll. Physicians",
  ["bulletin of the american college of surgeons"] = "Bull Am Coll Surg",
  ["bulletin of the american mathematical society"] = "Bull. Am. Math. Soc.",
  ["bulletin of the american meteorological society"] = "Bull. Am. Meteorol. Soc.",
  ["bulletin of the american museum of natural history"] = "Bull Am Mus Nat Hist",
  ["bulletin of the american physical society"] = "Bull. Am. Phys. Soc.",
  ["bulletin of the american schools of oriental research"] = "BASOR",
  ["bulletin of the american schools of oriental research in jerusalem and baghdad"] = "BASO",
  ["bulletin of the american schools of oriental research. american schools of oriental research"] = "Bull Am Schools Orient Res",
  ["bulletin of the american society for information science"] = "Bull Am Soc Inf Sci",
  ["bulletin of the american society for information science / asis"] = "Bull Am Soc Inf Sci",
  ["bulletin of the american society of overseas research"] = "Bull. Am. Soc. Overseas Res.",
  ["bulletin of the american society of papyrologists"] = "BASP",
  ["bulletin of the american society of veterinary clinical pathologists"] = "Bull. Am. Soc. Vet. Clin. Pathol.",
  ["bulletin of the ancient orient museum (tokyo)"] = "BAncOrMus",
  ["bulletin of the anglo-israel archaeological society"] = "BAngloIsrASoc",
  ["bulletin of the asia institute"] = "BAsInst",
  ["bulletin of the association of american physicians and surgeons"] = "Bull Assoc Am Physicians Surg",
  ["bulletin of the association of medical librarians"] = "Bull Assoc Med Libr",
  ["bulletin of the astronomical institutes of the netherlands"] = "Bull. Astron. Inst. Neth.",
  ["bulletin of the astronomical society of india"] = "Bull. Astron. Soc. India",
  ["bulletin of the atomic scientists"] = "Bull. At. Sci.",
  ["bulletin of the australian mathematical society"] = "Bull. Aust. Math. Soc.",
  ["bulletin of the belgian mathematical society"] = "Bull. Belg. Math. Soc. Simon Stevin",
  ["bulletin of the belgian mathematical society - simon stevin"] = "Bull. Belg. Math. Soc. Simon Stevin",
  ["bulletin of the belgian mathematical society. simon stevin"] = "Bull. Belg. Math. Soc. Simon Stevin",
  ["bulletin of the bergen county dental society"] = "Bull. Bergen Cty. Dent. Soc.",
  ["bulletin of the bio-geographical society of japan = nihon seibutsuchiri gakkai kaiho"] = "Bull Biogeogr Soc Jpn",
  ["bulletin of the biogeographical society of japan"] = "Bull. Biogeogr. Soc. Jpn.",
  ["bulletin of the biological society of washington"] = "Bull Biol Soc Wash",
  ["bulletin of the brazilian mathematical society"] = "Bull. Braz. Math. Soc.",
  ["bulletin of the brazilian mathematical society. new series. boletim da sociedade brasileira de matematica"] = "Bull. Braz. Math. Soc. (N.S.)",
  ["bulletin of the british association of orientalists. british association of orientalists"] = "Bull Br Assoc Orient",
  ["bulletin of the british museum (natural history). historical series"] = "Bull Br Mus",
  ["bulletin of the british museum of natural history, botany"] = "Bull. Br. Mus. Nat. Hist., Bot.",
  ["bulletin of the british mycological society"] = "Bull. Br. Mycol. Soc.",
  ["bulletin of the british myriapod and isopod group"] = "Bull Br Myriap Isopod Group",
  ["bulletin of the british ornithologists' club"] = "Bull Br Ornithol Club",
  ["bulletin of the british ornithologists’ club"] = "Bull. Br. Ornithol. Club",
  ["bulletin of the british psychological society"] = "Bull Br Psychol Soc",
  ["bulletin of the bronx county dental society"] = "Bull. Bronx Cty. Dent. Soc.",
  ["bulletin of the calcutta mathematical society"] = "Bull. Calcutta Math. Soc.",
  ["bulletin of the calcutta school of tropical medicine"] = "Bull Calcutta Sch Trop Med",
  ["bulletin of the california agricultural experiment station"] = "Bull. Calif. Agric. Exp. Stn.",
  ["bulletin of the california dietetic association"] = "Bull Calif Diet Assoc",
  ["bulletin of the canadian biochemical society"] = "Bull. Can. Biochem. Soc.",
  ["bulletin of the chemical society of ethiopia"] = "Bull. Chem. Soc. Ethiop.",
  ["bulletin of the chemical society of japan"] = "Bull. Chem. Soc. Jpn.",
  ["bulletin of the chest disease research institute, kyoto university"] = "Bull Chest Dis Res Inst Kyoto Univ",
  ["bulletin of the cincinnati dental society"] = "Bull. Cincinnati Dent. Soc.",
  ["bulletin of the cleveland dental society"] = "Bull. Cleve. Dent. Soc.",
  ["bulletin of the contra costa dental society"] = "Bull Contra Costa Dent Soc",
  ["bulletin of the dental guidance council for cerebral palsy"] = "Bull Dent Guid Counc Cereb Palsy",
  ["bulletin of the department of health, commonwealth of kentucky. kentucky. department of health"] = "Bull Dep Health Commonw Ky",
  ["bulletin of the ecological society of america"] = "Bull. Ecol. Soc. Am.",
  ["bulletin of the egyptological seminar"] = "BES",
  ["bulletin of the eleventh district dental society"] = "Bull Eleventh Dist Dent Soc",
  ["bulletin of the entomological society of america"] = "Bull. Entomol. Soc. Am.",
  ["bulletin of the entomological society of egypt : economic series"] = "Bull Entomol Soc Egypt",
  ["bulletin of the european association for theoretical computer science"] = "Bull. Eur. Assoc. Theor. Comput. Sci. EATCS",
  ["bulletin of the european association for theoretical computer science. eatcs"] = "Bull. Eur. Assoc. Theor. Comput. Sci. EATCS",
  ["bulletin of the european association of fish pathologists"] = "Bull Eur Assoc Fish Pathol",
  ["bulletin of the european ceramic society"] = "Bull. Eur. Ceram. Soc.",
  ["bulletin of the faculty of education"] = "Bull. Fac. Ed. Kagoshima Univ. Natur. Sci.",
  ["bulletin of the faculty of education. kagoshima university. natural science"] = "Bull. Fac. Ed. Kagoshima Univ. Natur. Sci.",
  ["bulletin of the faculty of engineering"] = "Bull. Fac. Eng. Hiroshima Univ.",
  ["bulletin of the faculty of liberal arts"] = "Bull. Fac. Liberal Arts Nagasaki Univ.",
  ["bulletin of the faculty of school education"] = "Bull. Fac. School Ed. Hiroshima Univ. Part II",
  ["bulletin of the faculty of school education hiroshima university"] = "Bull. Fac. Sch. Educ. Hiroshima Univ. Part II",
  ["bulletin of the faculty of science"] = "Bull. Fac. Sci. Assiut Univ. A",
  ["bulletin of the faculty of science and engineering"] = "Bull. Fac. Sci. Eng. Chuo Univ. Ser. I Math.",
  ["bulletin of the faculty of science. university of the ryukyus"] = "Bull. Fac. Sci. Univ. Ryukyus",
  ["bulletin of the faculty of the liberal arts"] = "Bull. Fac. Liberal Arts Chukyo Univ.",
  ["bulletin of the far eastern branch of the russian academy of sciences"] = "Bull Far East Branch Russ Acad Sci",
  ["bulletin of the fifth district dental society"] = "Bull. Fifth Dist. Dent. Soc. (Fresno)",
  ["bulletin of the fifth district dental society of the state of new york"] = "Bull Fifth Dist Dent Soc State N Y",
  ["bulletin of the florida museum of natural history"] = "Bull Fla Mus Nat Hist",
  ["bulletin of the geobotanical institute eth"] = "Bull. Geobot. Inst. ETH",
  ["bulletin of the geological society of china"] = "Bull. Geol. Soc. China",
  ["bulletin of the geological society of denmark"] = "Bull. Geol. Soc. Den.",
  ["bulletin of the geological society of finland"] = "Bull. Geol. Soc. Finl.",
  ["bulletin of the geological survey of finland"] = "Bull. Geol. Surv. Finl.",
  ["bulletin of the georgia academy of science"] = "Bull Ga Acad Sci",
  ["bulletin of the georgian national academy of sciences. new series"] = "Bull. Georgian Natl. Acad. Sci. (N.S.)",
  ["bulletin of the ghana geographical association"] = "Bull Ghana Geogr Assoc",
  ["bulletin of the gloucester county historical society"] = "Bull Gloucester Cty Hist Soc",
  ["bulletin of the greek mathematical society"] = "Bull. Greek Math. Soc.",
  ["bulletin of the hellenic mathematical society"] = "Bull. Hellenic Math. Soc.",
  ["bulletin of the history of dentistry"] = "Bull Hist Dent",
  ["bulletin of the history of medicine"] = "Bull Hist Med",
  ["bulletin of the history of medicine. supplements"] = "Bull Hist Med Suppl",
  ["bulletin of the hong kong mathematical society"] = "Bull. Hong Kong Math. Soc.",
  ["bulletin of the hospital for joint disease (2013)"] = "Bull Hosp Jt Dis  (2013)",
  ["bulletin of the hospital for joint diseases"] = "Bull Hosp Joint Dis",
  ["bulletin of the hospital for joint diseases orthopaedic institute"] = "Bull Hosp Jt Dis Orthop Inst",
  ["bulletin of the hudson county dental society"] = "Bull. Hudson Cty. Dent. Soc.",
  ["bulletin of the hunt institute for botanical documentation. hunt institute for botanical documentation"] = "Bull Hunt Inst Bot Doc",
  ["bulletin of the iacfs/me"] = "Bull IACFS ME",
  ["bulletin of the illinois dental hygienists association"] = "Bull. Ill. Dent. Hyg. Assoc.",
  ["bulletin of the indian institute of history of medicine (hyderabad)"] = "Bull Indian Inst Hist Med Hyderabad",
  ["bulletin of the indian vacuum society"] = "Bull. Indian Vac. Soc.",
  ["bulletin of the institute for chemical research, kyoto university"] = "Bull Inst Chem Res Kyoto Univ",
  ["bulletin of the institute for medical research, university of madrid"] = "Bull Inst Med Res Univ Madr",
  ["bulletin of the institute of archaeology of the univ. of london"] = "BIAL",
  ["bulletin of the institute of archaeology, university of london"] = "BALond",
  ["bulletin of the institute of classical studies of the university of london"] = "BICS",
  ["bulletin of the institute of combinatorics and its applications"] = "Bull. Inst. Combin. Appl.",
  ["bulletin of the institute of historical research"] = "Bull Inst Hist Res",
  ["bulletin of the institute of history of medicine (hyderabad)"] = "Bull Inst Hist Med Hyderabad",
  ["bulletin of the institute of maritime and tropical medicine in gdynia"] = "Bull Inst Marit Trop Med Gdynia",
  ["bulletin of the institute of mathematics"] = "Bull. Inst. Math. Acad. Sinica",
  ["bulletin of the institute of mathematics. academia sinica. new series"] = "Bull. Inst. Math. Acad. Sin. (N.S.)",
  ["bulletin of the institute of medicine (hyderabad)"] = "Bull Inst Hist Med Hyderabad",
  ["bulletin of the international association for paleodontology"] = "Bull. Int. Assoc. Paleondontol.",
  ["bulletin of the international association of engineering geology"] = "Bull. Int. Assoc. Eng. Geol.",
  ["bulletin of the international association of scientific hydrology"] = "Bull. Int. Assoc. Sci. Hydrol.",
  ["bulletin of the international committee on urgent anthropological and ethnological research. international committee on urgent anthropological and ethnological research"] = "Bull Int Comm Urgent Anthropol Ethnol Res",
  ["bulletin of the international dairy association"] = "Bull. Int. Dairy Assoc.",
  ["bulletin of the international pediatric association. bulletin de l'association internationale de pédiatrie. boletin de la asociacion internacional de pediatria"] = "Bull Int Pediatr Assoc",
  ["bulletin of the international pediatric association. bulletin de l’association internationale de pediatrie. boletin de la asociacion internacional de pediatria"] = "Bull Int Pediatr Assoc",
  ["bulletin of the international union against tuberculosis"] = "Bull Int Union Tuberc",
  ["bulletin of the international union against tuberculosis and lung disease"] = "Bull Int Union Tuberc Lung Dis",
  ["bulletin of the iranian mathematical society"] = "Bull. Iran. Math. Soc.",
  ["bulletin of the john rylands library"] = "BRL",
  ["bulletin of the johns hopkins hospital"] = "Bull Johns Hopkins Hosp",
  ["bulletin of the kanagawa dental college"] = "Bull. Kanagawa Dent. Coll.",
  ["bulletin of the karaganda university. mathematics series"] = "Bull. Karaganda Univ. Math. Ser.",
  ["bulletin of the korean chemical society"] = "Bull. Korean Chem. Soc.",
  ["bulletin of the korean mathematical society"] = "Bull. Korean Math. Soc.",
  ["bulletin of the kyoto university of education"] = "Bull. Kyoto Univ. Ed. Ser. B",
  ["bulletin of the kyushu institute of technology"] = "Bull. Kyushu Inst. Tech. Math. Natur. Sci.",
  ["bulletin of the kyushu institute of technology. pure and applied mathematics"] = "Bull. Kyushu Inst. Technol. Pure Appl. Math.",
  ["bulletin of the london mathematical society"] = "Bull. London Math. Soc.",
  ["bulletin of the los angeles dental society"] = "Bull Los Angeles Dent Soc",
  ["bulletin of the los angeles neurological societies"] = "Bull Los Angeles Neurol Soc",
  ["bulletin of the los angeles neurological society"] = "Bull Los Angel Neuro Soc",
  ["bulletin of the malaysian mathematical sciences society"] = "Bull. Malays. Math. Sci. Soc.",
  ["bulletin of the mason clinic"] = "Bull Mason Clin",
  ["bulletin of the massachusetts nurses association"] = "Bull Mass Nurses Assoc",
  ["bulletin of the medical library association"] = "Bull Med Libr Assoc",
  ["bulletin of the menninger clinic"] = "Bull Menninger Clin",
  ["bulletin of the metropolitan museum of art"] = "BMM",
  ["bulletin of the michigan dental hygienists association"] = "Bull. Mich. Dent. Hyg. Assoc.",
  ["bulletin of the missouri historical society. missouri historical society"] = "Bull Miss Hist Soc",
  ["bulletin of the montgomery-bucks dental society"] = "Bull. Montg. Bucks Dent. Soc.",
  ["bulletin of the museum of comparative zoology"] = "Bull. Mus. Comp. Zool.",
  ["bulletin of the museum of comparative zoology at harvard college"] = "Bull Mus Comp Zool",
  ["bulletin of the museum of mediterranean and near eastern antiquities"] = "BMNE",
  ["bulletin of the museums of art and archaeology of the university of michigan"] = "BMUSUM",
  ["bulletin of the national guild of catholic psychiatrists"] = "Bull Natl Guild Cathol Psychiatr",
  ["bulletin of the national museum of nature and science. series a, zoology"] = "Bull Natl Mus Nat Sci Ser A Zool",
  ["bulletin of the national museum of nature and science. series b, botany"] = "Bull Natl Mus Nat Sci Ser B Bot",
  ["bulletin of the national research centre"] = "Bull Natl Res Cent",
  ["bulletin of the national science museum. series b, botany"] = "Bull Natl Sci Mus Ser B Bot",
  ["bulletin of the national society for medical research"] = "Bull Natl Soc Med Res",
  ["bulletin of the natural family planning council of victoria"] = "Bull Nat Fam Plan Counc Vic",
  ["bulletin of the new jersey college of medicine and dentistry"] = "Bull N J Coll Med Dent",
  ["bulletin of the new jersey society of dentistry for children"] = "Bull. N. J. Soc. Dent. Child.",
  ["bulletin of the new york academy of medicine"] = "Bull N Y Acad Med",
  ["bulletin of the new york public library, astor, lenox, and tilden foundations. new york public library"] = "Bull N Y Public Libr",
  ["bulletin of the new york state dental society of anesthesiology"] = "Bull N Y State Dent Soc Anesthesiol",
  ["bulletin of the new york state society of dentistry for children"] = "Bull N Y State Soc Dent Child",
  ["bulletin of the newark dental club"] = "Bull Newark Dent Club",
  ["bulletin of the ninth district dental society"] = "Bull. Ninth Dist. Dent. Soc.",
  ["bulletin of the nyu hospital for joint diseases"] = "Bull NYU Hosp Jt Dis",
  ["bulletin of the ophthalmological society of egypt"] = "Bull Ophthalmol Soc Egypt",
  ["bulletin of the osaka medical college"] = "Bull Osaka Med Coll",
  ["bulletin of the osaka medical school"] = "Bull Osaka Med Sch",
  ["bulletin of the osaka medical school. supplement"] = "Bull Osaka Med Sch Suppl",
  ["bulletin of the oxford university institute of economics and statistics"] = "Bull. Oxford Univ. Inst. Econ. Statist.",
  ["bulletin of the pan american health organization"] = "Bull Pan Am Health Organ",
  ["bulletin of the parenteral drug association"] = "Bull Parenter Drug Assoc",
  ["bulletin of the park ridge center"] = "Bull Park Ridge Cent",
  ["bulletin of the passaic county dental society"] = "Bull Passaic Cty Dent Soc",
  ["bulletin of the peabody museum of natural history"] = "Bull. Peabody Mus. Nat. Hist.",
  ["bulletin of the philadelphia association for psychoanalysis"] = "Bull Phila Assoc Psychoanal",
  ["bulletin of the philadelphia county dental society"] = "Bull. Phila. Cty. Dent. Soc.",
  ["bulletin of the plainfield dental society"] = "Bull Plainfield Dent Soc",
  ["bulletin of the polish academy of sciences chemistry"] = "Bull. Pol. Acad. Sci., Chem.",
  ["bulletin of the polish academy of sciences technical sciences"] = "Bull. Pol. Acad. Sci. Tech. Sci.",
  ["bulletin of the polish academy of sciences, chemistry"] = "Bull. Pol. Acad. Sci., Chem.",
  ["bulletin of the polish academy of sciences. chemistry"] = "Bull Pol Acad Sci Chem",
  ["bulletin of the polish academy of sciences. mathematics"] = "Bull. Pol. Acad. Sci. Math.",
  ["bulletin of the population and development studies center. soul taehakkyo. in’gu mit palchon munje yon’guso"] = "Bull Popul Dev Stud Cent",
  ["bulletin of the population and development studies center. sŏul taehakkyo. inʼgu mit palchŏn munje yŏnʼguso"] = "Bull Popul Dev Stud Cent",
  ["bulletin of the post-graduate committee in medicine, university of sydney"] = "Bull Postgrad Comm Med Univ Syd",
  ["bulletin of the psychonomic society"] = "Bull Psychon Soc",
  ["bulletin of the public health society"] = "Bull Publ Health Soc (Kuala Lumpur)",
  ["bulletin of the research council of israel"] = "Bull Res Counc Isr",
  ["bulletin of the research council of israel. section e: experimental medicine"] = "Bull Res Counc Isr Sect E Exp Med",
  ["bulletin of the rhode island school of design. museum notes"] = "BProvidence",
  ["bulletin of the royal institute for inter-faith studies"] = "Bull R Inst Inter Faith Stud",
  ["bulletin of the royal ontario museum of archaeology, university of toronto"] = "BOntMus",
  ["bulletin of the russian academy of sciences. physics"] = "Bull. Russ. Acad. Sci. Phys.",
  ["bulletin of the russian academy of sciences: physics"] = "Bull. Russ. Acad. Sci.: Phys.",
  ["bulletin of the school of medicine"] = "Bull Sch Med",
  ["bulletin of the school of medicine (baltimore, md.)"] = "Bull Sch Med Univ Md",
  ["bulletin of the school of oriental and african studies (london)"] = "BSOAS",
  ["bulletin of the school of oriental and african studies. university of london. school of oriental and african studies"] = "Bull Sch Orient Afr Stud",
  ["bulletin of the scientific instrument society"] = "Bull Sci Instrum Soc",
  ["bulletin of the seismological society of america"] = "Bull. Seismol. Soc. Am.",
  ["bulletin of the sloane hospital for women in the columbia-presbyterian medical center, new york city"] = "Bull Sloane Hosp Women Columbia Presbyt Med",
  ["bulletin of the southern california academy of sciences"] = "Bull. South. Calif. Acad. Sci.",
  ["bulletin of the suffolk county dental society"] = "Bull Suffolk Cty Dent Soc",
  ["bulletin of the technical committee on data engineering"] = "Bull Tech Comm Data Eng",
  ["bulletin of the tenth district dental society of the state of new york"] = "Bull Tenth Dist Dent Soc (Rockville Centre)",
  ["bulletin of the toledo dental society"] = "Bull. Toledo Dent. Soc.",
  ["bulletin of the toronto east medical association"] = "Bull Tor East Med Assoc",
  ["bulletin of the transilvania university of brasov. series iii. mathematics and computer science"] = "Bull. Transilv. Univ. Brasov Ser. III. Math. Comput. Sci.",
  ["bulletin of the transilvania university of brasov. series iii. mathematics, informatics, physics"] = "Bull. Transilv. Univ. Brasov Ser. III",
  ["bulletin of the tri-county dental society"] = "Bull. Tri. Cty. Dent. Soc.",
  ["bulletin of the u.s. army medical department. united states. army. medical department"] = "Bull U S Army Med Dep",
  ["bulletin of the unesco regional office for education in asia and the pacific"] = "Bull Unesco Reg Off Educ Asia Pac",
  ["bulletin of the union county dental society"] = "Bull. Union Cty. Dent. Soc.",
  ["bulletin of the university of electro-communications"] = "Bull. Univ. Electro-Comm.",
  ["bulletin of the university of minnesota hospitals and minnesota medical foundation. university of minnesota. hospitals"] = "Bull Univ Minn Hosp Minn Med Found",
  ["bulletin of the university of nebraska state museum"] = "Bull Univ Nebr State Mus",
  ["bulletin of the valley dental society"] = "Bull Val Dent Soc",
  ["bulletin of the veterinary institute in puławy = biuletyn instytutu weterynarii w puławach"] = "Bull Vet Inst Pulawy",
  ["bulletin of the world health organization"] = "Bull World Health Organ",
  ["bulletin of the world health organization. supplement"] = "Bull World Health Org Suppl",
  ["bulletin of the yamagata university"] = "Bull. Yamagata Univ. Natur. Sci.",
  ["bulletin of tokyo dental college"] = "Bull. Tokyo Dent. Coll.",
  ["bulletin of tokyo gakugei university"] = "Bull. Tokyo Gakugei Univ. (4)",
  ["bulletin of tokyo gakugei university. natural sciences"] = "Bull. Tokyo Gakugei Univ. Nat. Sci.",
  ["bulletin of tokyo medical and dental university"] = "Bull. Tokyo Med. Dent. Univ.",
  ["bulletin of university of agricultural sciences and veterinary medicine cluj-napoca"] = "Bull Univ Agric Sci Vet Med Cluj Napoca",
  ["bulletin of university of osaka prefecture"] = "Bull. Univ. Osaka Prefect.",
  ["bulletin of university of teacher education fukuoka. part iii. mathematics, natural sciences and technology"] = "Bull. Univ. Teach. Educ. Fukuoka. III",
  ["bulletin of venereal disease control. ontario. department of health"] = "Bull Vener Dis Control",
  ["bulletin of venereal diseases"] = "Bull Vener Dis",
  ["bulletin of volcanology"] = "Bull. Volcanol.",
  ["bulletin of war medicine"] = "Bull War Med",
  ["bulletin of zoological nomenclature"] = "Bull. Zool. Nomencl.",
  ["bulletin officiel du conseil national de l'ordre"] = "Ordre Natl Chir Dent Cons Natl Bull Off",
  ["bulletin officiel du conseil national de l’ordre / ordre national des chirurgiens-dentistes"] = "Ordre Natl Chir Dent Cons Natl Bull Off",
  ["bulletin officiel, conseil national, ordre national des chirurgiens-dentistes"] = "Ordre Natl. Chir. Dent. Cons. Natl. Bull. Off.",
  ["bulletin officiel. chambre syndicale des medecins de la seine"] = "Bull Off Chambre Synd Med Seine",
  ["bulletin officiel. chambre syndicale des médecins de la seine"] = "Bull Off Chambre Synd Med Seine",
  ["bulletin on narcotics"] = "Bull Narc",
  ["bulletin on the rheumatic diseases"] = "Bull Rheum Dis",
  ["bulletin on training"] = "Bull. Train.",
  ["bulletin on training, trends, techniques, topics"] = "Bull Train",
  ["bulletin peabody museum of natural history, yale university"] = "Bull. Peabody Mus. Nat. Hist., Yale Univ.",
  ["bulletin philologique et historique (jusqu'à 1610) du comité des travaux historiques et scientifiques"] = "Bull Philol Hist",
  ["bulletin philologique et historique (jusqu’a 1610) du comite des travaux historiques et scientifiques"] = "Bull Philol Hist",
  ["bulletin razreda za likovne umjetnosti hrvatske akademije znanosti i umjetnosti"] = "BLikUm",
  ["bulletin romand d’entomologie"] = "Bull. romand entomol.",
  ["bulletin sanitaire"] = "Bull Sanit",
  ["bulletin suisse de mycologie"] = "Bull. Suisse Mycol.",
  ["bulletin technique de la suisse romande"] = "Bull. tech. Suisse romande",
  ["bulletin trimestriel - crédit communal de belgique"] = "Bull Trimest Credit Communal Belg",
  ["bulletin trimestriel de l'institut des actuaires français. institut des actuaires français, paris"] = "Bull Trimest Inst Actuaires Francais",
  ["bulletin trimestriel de la societe mycologique de france"] = "Bull Trimest Soc Mycol Fr",
  ["bulletin trimestriel de la société mycologique de france"] = "Bull Trimest Soc Mycol Fr",
  ["bulletin trimestriel de l’institut des actuaires francais. institut des actuaires francais, paris"] = "Bull Trimest Inst Actuaires Francais",
  ["bulletin, alameda county dental society"] = "Bull. Alameda Cty. Dent. Soc.",
  ["bulletin, american association for the history of nursing"] = "Bull. Am. Assoc. Hist. Nurs.",
  ["bulletin, american association of dental editors"] = "Bull. Am. Assoc. Dent. Ed.",
  ["bulletin, american association of hospital dentists"] = "Bull. Am. Assoc. Hosp. Dent.",
  ["bulletin, american protestant hospital association"] = "Bull. Am. Protestant Hosp. Assoc.",
  ["bulletin, atlantic-cape may county dental society"] = "Bull. Atl. Cape May Cty. Dent. Soc.",
  ["bulletin, cincinnati dental society"] = "Bull. Cincinnati Dent. Soc.",
  ["bulletin, dayton dental society"] = "Bull. Dayton Dent. Soc.",
  ["bulletin, essex county dental society"] = "Bull. Essex Cty. Dent. Soc.",
  ["bulletin, hospital for joint diseases"] = "Bull. Hosp. Jt. Dis.",
  ["bulletin, mcguire clinic and st. luke's hospital. st. luke's hospital (richmond, va.)"] = "Bull McGuire Clin St Lukes Hosp",
  ["bulletin, mcguire clinic and st. luke’s hospital. richmond. st. luke’s hospital. mcguire clinic"] = "Bull McGuire Clin St Lukes Hosp Richmond St Lukes Hosp McGuire Clin",
  ["bulletin, midtown dental society"] = "Bull. Midtown Dent. Soc.",
  ["bulletin, monmouth-ocean county dental society"] = "Bull. Monmouth. Ocean Cty. Dent. Soc.",
  ["bulletin, national medical and dental association and national advocates society"] = "Bull. Natl. Med. Dent. Assoc. Natl. Advocates Soc.",
  ["bulletin, national tuberculosis and respiratory disease association"] = "Bull. Natl. Tuberc. Respir. Dis. Assoc.",
  ["bulletin, national tuberculosis association"] = "Bull. Natl. Tuberc. Assoc.",
  ["bulletin, northern district dental society"] = "Bull. North. Dist. Dent. Soc.",
  ["bulletin, nz society of periodontology"] = "Bull. N. Z. Soc. Periodontol.",
  ["bulletin, pacific coast society of orthodontists"] = "Bull. Pac. Coast Soc. Orthod.",
  ["bulletin, san diego county dental society"] = "Bull. San Diego Cty. Dent. Soc.",
  ["bulletin, san mateo county dental society"] = "Bull. San Mateo Cty. Dent. Soc.",
  ["bulletin, second district dental society"] = "Bull. Second Dist. Dent. Soc.",
  ["bulletin, tennessee nurses association"] = "Bull. Tenn. Nurses Assoc.",
  ["bulletin, texas nurses association"] = "Bull. Tex. Nurses Assoc.",
  ["bulletin. académie royale des sciences, des lettres et des beaux-arts de belgique. classe des sciences"] = "Bull Cl Sci Acad R Belg",
  ["bulletin. acuff clinic, knoxville, tenn"] = "Bull Acuff Clin Knoxv Tenn",
  ["bulletin. adelphi hospital (brooklyn, new york, n.y.)"] = "Bull Adelphi Hosp Brooklyn",
  ["bulletin. american association of medical social workers"] = "Bull Am Assoc Med Soc Workers",
  ["bulletin. american cancer society"] = "Bull Am Cancer Soc",
  ["bulletin. american pharmaceutical association. committee on national formulary"] = "Bull Am Pharm Assoc",
  ["bulletin. american society of hospital pharmacists"] = "Bull Am Soc Hosp Pharm",
  ["bulletin. association medicale haitienne"] = "Bull Assoc Med Haiti",
  ["bulletin. association médicale haïtienne"] = "Bull Assoc Med Haiti",
  ["bulletin. augusta, ga. university hospital"] = "Bull Univ Hosp",
  ["bulletin. bronx county dental society"] = "Bull Bronx Cty Dent Soc",
  ["bulletin. brooklyn. jewish hospital"] = "Bull Brooklyn",
  ["bulletin. buffalo (n.y.). millard fillmore hospital"] = "Bull Millard Fillmore Hosp Buffalo",
  ["bulletin. buffalo. millard fillmore hospital"] = "Bull Millard Fillmore Hosp Buffalo",
  ["bulletin. california state nurses' association"] = "Bull Calif State Nurses Assoc",
  ["bulletin. charlotte, n. c. charlotte memorial hospital"] = "Bull Charlotte Meml Hosp",
  ["bulletin. chicago heart association"] = "Bull Chic Heart Assoc",
  ["bulletin. chicago medical society"] = "Bull Chic Med Soc",
  ["bulletin. classe des sciences mathematiques et naturelles. sciences mathematiques"] = "Bull. Cl. Sci. Math. Nat. Sci. Math.",
  ["bulletin. college of general practice of canada"] = "Bull Coll Gen Pract Can",
  ["bulletin. colorado. state board of health"] = "Bull Colo",
  ["bulletin. columbus dental society"] = "Columbus Dent Soc Bull",
  ["bulletin. connecticut state dental association"] = "Bull Conn State Dent Assoc",
  ["bulletin. creighton university. school of medicine"] = "Bull Creighton Univ",
  ["bulletin. dayton (ohio) dental society"] = "Bull Dayton Dent Soc",
  ["bulletin. georgetown university. medical center"] = "Bull Georgetown Univ Med Cent",
  ["bulletin. health organisation"] = "Health Organ Bull",
  ["bulletin. hospital for special surgery"] = "Bull Hosp Spec Surg",
  ["bulletin. hudson county dental society, jersey city"] = "Bull Hudson Cty Dent Soc",
  ["bulletin. institut national d'hygiène (france)"] = "Bull Inst Natl Hyg",
  ["bulletin. international association of medical museums"] = "Bull Int Assoc Med Mus",
  ["bulletin. j. paul getty museum of art"] = "BulletinGetty",
  ["bulletin. jami’at al-iskandariyah. kulliyat al-tibb"] = "Bull Alexandria Fac",
  ["bulletin. jersey city. margaret hague maternity hospital"] = "Bull Jersey City Margaret Hague Mat Hosp",
  ["bulletin. john rylands library, manchester"] = "Bull John Rylands Univ Lib",
  ["bulletin. john rylands university library of manchester"] = "Bull J Rylands Univ Libr Manchester",
  ["bulletin. kresge eye institute"] = "Bull Kresge Eye Inst",
  ["bulletin. kullīyat ṭibb al-iskandarīyah"] = "Bull Alexandria Fac",
  ["bulletin. league of red cross societies"] = "Bull Leag Red Cross Soc",
  ["bulletin. linn county (iowa) medical society"] = "Bull Linn Cty Iowa",
  ["bulletin. los angeles surgical society"] = "Bull Los Angel Surg Soc",
  ["bulletin. mahoning county medical society"] = "Bull Mahoning Cty Med Soc",
  ["bulletin. margaret hague maternity hospital (jersey city, n.j.)"] = "Bull Jersey City Margaret Hague Mat Hosp",
  ["bulletin. massachusetts dental society"] = "Bull Mass Dent Soc",
  ["bulletin. medical society of the county of kings (kings county, n.y.)"] = "Bull Med Soc Cty Kings Kings Cty N Y",
  ["bulletin. medical society of the county of monroe"] = "Bull Med Soc Cty Monroe",
  ["bulletin. meriden, conn. hospital"] = "Bull Meriden Conn Hosp",
  ["bulletin. minnesota medical foundation"] = "Bull Minn Med Found",
  ["bulletin. missouri society for crippled children"] = "Bull Mo Soc Crippl Child",
  ["bulletin. montgomery-bucks dental society"] = "Bull Montg Bucks Dent Soc",
  ["bulletin. moore-white medical foundation, los angeles"] = "Bull Moore White Med Found Los Angel",
  ["bulletin. multnomah county (or.) medical society"] = "Bull Multnomah Cty",
  ["bulletin. museum of fine arts, boston"] = "BMusFA",
  ["bulletin. museums of art and archaeology, university of michigan"] = "BMusMich",
  ["bulletin. national institutes of health (u.s.)"] = "Bull Natl Inst Health",
  ["bulletin. new england medical center"] = "Bull New Engl Med Cent",
  ["bulletin. new york medical college"] = "Bulletin NY Med Coll",
  ["bulletin. ninth district dental society of the state of new york"] = "Bull Ninth Dist Dent Soc",
  ["bulletin. north carolina dental society"] = "Bull N C Dent Soc",
  ["bulletin. oak park, ill. west suburban hospital"] = "Bull Oak Park Ill",
  ["bulletin. okayama daigaku. igakubu. gangen kenkyu shisetsu"] = "Bull Cancer Inst Okayama Univ Med Sch",
  ["bulletin. okayama daigaku. igakubu. gangen kenkyū shisetsu"] = "Bull Cancer Inst Okayama Univ Med Sch",
  ["bulletin. oklahoma state dental association"] = "Bull Okla State Dent Assoc",
  ["bulletin. pennsylvania hospital (philadelphia, pa.). ayer clinical laboratory"] = "Bull Phila Pa Hosp Ayer Clin Lab",
  ["bulletin. philadelphia. pennsylvania hospital. ayer clinical laboratory"] = "Bull Phila Pa Hosp Ayer Clin Lab",
  ["bulletin. portland, or. clinic"] = "Portland Clin Bull",
  ["bulletin. pottawatomie county (okla.) medical society"] = "Bull Pottawatomie Cty Okla",
  ["bulletin. roslyn, n. y. st. francis hospital and sanatorium for cardiac children"] = "Heart Cent Bull (Roslyn)",
  ["bulletin. roslyn, n. y. st. francis sanatorium for cardiac children"] = "Bull Rosl N Y St. Francis Sanat Card Child",
  ["bulletin. roswell park memorial institute"] = "Bull Roswell Park Meml Inst",
  ["bulletin. scientific issue. huron road hospital"] = "Bull Sci Issue",
  ["bulletin. smith college museum of art"] = "BulletinNorthampton",
  ["bulletin. societe de chirurgie de paris"] = "Bull Soc Chir Paris",
  ["bulletin. societe des medecins hygienistes d’algerie"] = "Bull Soc Med Hyg Alger",
  ["bulletin. societe des sciences, lettres et arts de pau"] = "Bull Soc Sci Lettr Arts Pau",
  ["bulletin. society of medical history of chicago"] = "Bull Soc Med History Chic",
  ["bulletin. société de chirurgie de paris"] = "Bull Soc Chir Paris",
  ["bulletin. société des médecins hygiénistes d'algérie"] = "Bull Soc Med Hyg Alger",
  ["bulletin. société des sciences, lettres et arts de pau"] = "Bull Soc Sci Lettr Arts Pau",
  ["bulletin. société scientifique d'hygiène alimentaire et d'alimentation rationnelle"] = "Bull Soc Sci Hyg Aliment Aliment Ration Homme",
  ["bulletin. société archéologique d’alexandrie"] = "BArchAlex",
  ["bulletin. srpska akademija nauka i umetnosti, belgrad. odeljenje medicinskih nauka"] = "Bull Acad Serbe Sci Arts",
  ["bulletin. srpska akademija nauka i umetnosti. odeljenje medicinskih nauka"] = "Bull Acad Serbe Sci Arts",
  ["bulletin. st. louis medical society"] = "Bull St Louis Med Society",
  ["bulletin. switzerland. eidgenössisches gesundheitsamt"] = "Bull Switz",
  ["bulletin. tarrant county medical society"] = "Bull Tarrant Cty Med Soc",
  ["bulletin. tbilisi international centre of mathematics and informatics"] = "Bull. TICMI",
  ["bulletin. the hebrew university, jerusalem, rabinowitz fund"] = "BJerus",
  ["bulletin. tufts-new england medical center"] = "Bull Tufts N Engl Med Cent",
  ["bulletin. union county (n. j.) dental society"] = "Bull Union Cty Dent Soc",
  ["bulletin. united states. army medical dept"] = "Bull U S Army Med Dep",
  ["bulletin. university of london. institute of classical studies"] = "Bull Inst Class Stud Univ Lond",
  ["bulletin. vancouver medical association"] = "Bull Vanc Med Assoc",
  ["bulletin. world medical association"] = "Bull World Med Assoc",
  ["bulletins de la société des antiquaires de l’ouest"] = "BSAO",
  ["bulletins et memoires de la faculte mixte de medecine et de pharmacie de dakar"] = "Bull Mem Fac Mixte Med Pharm Dakar",
  ["bulletins et memoires de la societe d’anthropologie de paris"] = "Bull Mem Soc Anthropol Paris",
  ["bulletins et memoires de la societe francaise d’ophtalmologie"] = "Bull. Mem. Soc. Fr. Ophtalmol.",
  ["bulletins et memoires de la societe medicale des hopitaux de paris"] = "Bull. Mem. Soc. Med. Hop. Paris",
  ["bulletins et mémoires de la faculté mixte de médecine et de pharmacie de dakar"] = "Bull Mem Fac Mixte Med Pharm Dakar",
  ["bulletins et mémoires de la société d'anthropologie de paris"] = "Bull Mem Soc Anthropol Paris",
  ["bulletins et mémoires de la société française d'ophtalmologie"] = "Bull Mem Soc Fr Ophtalmol",
  ["bulletins et mémoires de la société médicale des hôpitaux de paris"] = "Bull Mem Soc Med Hop Paris",
  ["bulletins et mémoires. société de chirurgie de marseille"] = "Bull Mem Chir Marseille",
  ["bulletins of the public health. united states. marine hospital service"] = "Bull Public Health",
  ["bulletins trimestriels de la société des antiquaires de picardie"] = "BSAP",
  ["bulletin’s science series"] = "Bull. Sci. Ser.",
  ["bulletí arqueològic. reial societat arqueològica tarraconense. boletín arqueológico. real sociedad arqueológica tarragonese"] = "BATarr",
  ["bullettino del centro camuno di studi preistorici"] = "BCamuno",
  ["bullettino del museo della civiltà romana"] = "BMusCivRom",
  ["bullettino della commissione archeologica comunale di roma"] = "BCom",
  ["bullettino della commissione archeologica comunale in roma"] = "BCAR",
  ["bullettino delle scienze mediche"] = "Bull Sci Med (Bologna)",
  ["bullettino dell’instituto di corrispondenza archeologica"] = "BdI",
  ["bullettino dell’istituto di diritto romano"] = "BIDR",
  ["bullettino dell’istituto di diritto romano »vittorio scialoja«"] = "BDirRom",
  ["bullettino di paletnologia italiana"] = "BPI",
  ["bullettino senese di storia patria"] = "BSiena",
  ["bumed news letter"] = "BuMed News Lett",
  ["bumed news letter. aviation supplement"] = "BuMed News Lett Aviat Suppl",
  ["bundesgesundheitsblatt, gesundheitsforschung, gesundheitsschutz"] = "Bundesgesundheitsblatt Gesundheitsforschung Gesundheitsschutz",
  ["bunseki kagaku"] = "Bunseki Kagaku",
  ["bureau of justice statistics special report"] = "Bur Justice Stat Spec Rep",
  ["bureau of mines information circular"] = "Bur. Mines Inf. Circ.",
  ["bureau of mines report of investigations"] = "Bur. Mines Rep. Invest.",
  ["burg monographs in science"] = "Burg Monogr. in Sci.",
  ["burgenlandische heimatblatter"] = "Burgenl Heimatbl",
  ["burgenländische heimatblätter"] = "Burgenl Heimatbl",
  ["burma medical journal"] = "Burma Med J",
  ["burnout research"] = "Burnout Res.",
  ["burns & trauma"] = "Burns Trauma",
  ["burns : journal of the international society for burn injuries"] = "Burns",
  ["burns open : an international open access journal for burn injuries"] = "Burns Open",
  ["burns, including thermal injury"] = "Burns Incl Therm Inj",
  ["business & professional ethics"] = "Bus Prof Ethics",
  ["business & professional ethics journal"] = "Bus Prof Ethics J",
  ["business and economic history"] = "Bus. Econ. Hist.",
  ["business and economic review (columbia, s.c.)"] = "Bus Econ Rev",
  ["business and economics journal"] = "Bus Econ J",
  ["business and health"] = "Bus Health",
  ["business and politics"] = "Bus Polit",
  ["business and society"] = "Bus Soc",
  ["business cycles, indicators and forecasting"] = "Stud. Bus. Cycles.",
  ["business economics"] = "Bus. Econ.",
  ["business economics (cleveland, ohio)"] = "Bus Econ",
  ["business ethics (oxford, england)"] = "Bus Ethics",
  ["business ethics quarterly : the journal of the society for business ethics"] = "Bus Ethics Q",
  ["business history"] = "Bus Hist",
  ["business history review"] = "Bus Hist Rev",
  ["business horizons"] = "Bus Horiz",
  ["business in thailand"] = "Bus Thail",
  ["business india"] = "BusinessIndia",
  ["business insurance"] = "Bus Insur",
  ["business review (federal reserve bank of philadelphia)"] = "Bus Rev",
  ["business strategy and the environment"] = "Bus Strategy Environ",
  ["business week"] = "Bus Week",
  ["butlleti de la societat catalana de matematiques"] = "Butl. Soc. Catalana Mat.",
  ["butlleti de la societat d’amics de la historia i de la ciencia farmaceutica catalana : bsahcfc"] = "Butll Soc Amics Hist Cienc Farm Catalana",
  ["butlletí de la institució catalana d'història natural"] = "Butll Inst Catalana Hist Nat",
  ["butlletí de la societat d'amics de la història i de la ciència farmacèutica catalana : bsahcfc"] = "Butll Soc Amics Hist Cienc Farm Catalana",
  ["butlletí de la societat catalana de matemàtiques"] = "Butl. Soc. Catalana Mat.",
  ["butlletí informatiu de l’institut de prehistòria i arqueologia de la diputació provincial de barcelona"] = "BABarcel",
  ["bŭlgarska klinika"] = "Bulg Klin",
  ["bücherei des augenarztes"] = "Buch Augenarzt",
  ["by og bygd"] = "By Bygd",
  ["bydesign : a technical publication of asse's engineering practice specialty. american society of safety engineers"] = "ByDesign",
  ["bygone kent"] = "Bygone Kent",
  ["byogen biseibutsu kenshutsu joho, geppo"] = "Infectious agents surveillance report (IASR)|Byogen Biseibutsu Kenshutsu Joho Geppo",
  ["byōgen biseibutsu kenshutsu jōhō, geppo = infectious agents surveillance report (iasr)"] = "Byogen Biseibutsu Kenshutsu Joho Geppo",
  ["byulleten eksperimental’noi biologii i meditsiny"] = "Byull. Eksp. Biol. Med.",
  ["byzantinisch-neugriechische jahrbücher"] = "ByzJb",
  ["byzantinische forschungen"] = "Byzantinische Forsch",
  ["byzantinische forschungen. internationale zeitschrift für byzantinistik"] = "ByzF",
  ["byzantinische forschungen: internationale zeitschrift für byzantinistik"] = "ByzF",
  ["byzantinische zeitschrift"] = "Byzantinische Z",
  ["byzantinoslavica: revue internationale des études byzantines"] = "ByzSlav",
  ["byzantion: revue internationale des études byzantines"] = "Byzantion",
  ["byzantion; revue internationale des etudes byzantines"] = "Byzantion",
  ["byzantion; revue internationale des études byzantines"] = "Byzantion",
  ["bzb. bayerisches zahnarzteblatt"] = "BZB Bayer Zahnarztebl",
  ["bzb. bayerisches zahnärzteblatt"] = "BZB Bayer Zahnarztebl",
  ["bzb: bayerisches zahnarzteblatt, mit mitteilungen d. kassenzahnarztlichen vereinigung bayerns"] = "BZB Bayer. Zahnarztebl. Mitt. Kassenzahnarztl. Ver. Bayerns",
  ["bündner monatsblatt"] = "Bündner Mon.bl.",
  ["bündnerwald, beiheft"] = "Bündnerwald, Beih.",
  ["bürger im staat, der"] = "Bürg. Staat",
  ["c&en global enterprise"] = "C&EN Global Enterp.",
  ["c.h.a.c. review"] = "CHAC Rev",
  ["c.i.c.i.a.m.s. nouvelles"] = "CICIAMS Nouv",
  ["ca: a cancer journal for clinicians"] = "CA Cancer J Clin",
  ["caai transactions on intelligence technology"] = "CAAI Trans Intell Technol",
  ["cabi agriculture and bioscience"] = "CABI Agric. Biosci.",
  ["cabi one health"] = "CABI One Health",
  ["cabi reviews"] = "CABI Rev.",
  ["caderno de terapeutica labor"] = "Cad Ter Labor",
  ["caderno de terapêutica labor"] = "Cad Ter Labor",
  ["cadernos da casa de oswaldo cruz"] = "Cad Casa Oswaldo Cruz",
  ["cadernos de arqueologia"] = "CadA",
  ["cadernos de ciência e tecnologia"] = "Cad. Ciênc. Tecnol.",
  ["cadernos de historia e saude / casa de oswaldo cruz"] = "Cad Hist Saude",
  ["cadernos de história e filosofia da ciência"] = "Cad. Hist. Filos. Ciênc. (3)",
  ["cadernos de história e saúde"] = "Cad Hist Saude",
  ["cadernos de logica e computacao"] = "Cad. Log. Comput.",
  ["cadernos de logica e filosofia"] = "Cad. Log. Filos.",
  ["cadernos de pesquisas"] = "Cad Pesqui",
  ["cadernos de saude publica / ministerio da saude, fundacao oswaldo cruz, escola nacional de saude publica"] = "Cad Saude Publica",
  ["cadernos de saúde pública"] = "Cad Saude Publica",
  ["cadernos de saúde pública"] = "Cad. Saúde Pública",
  ["cadernos saúde coletiva"] = "Cad Saude Colet",
  ["cadth technology overviews"] = "CADTH Technol Overv",
  ["caduceus (springfield, ill.)"] = "Caduceus",
  ["caesaraugusta. publicaciones del seminario de arqueología y numismática aragonesas"] = "Caesaraugusta",
  ["caesarodunum. bulletin de l’institut d’études latines et du centre de recherches a. piganiol"] = "Caesarodunum",
  ["cafs news : a centre for african family studies (cafs) newsletter"] = "CAFS News",
  ["cahier du centre technique du bois"] = "Cah. Cent. tech. bois",
  ["cahier. centre d’études chypriotes"] = "CahCEC",
  ["cahiers africains d'administration publique = african administrative studies"] = "Cah Afr Adm Publique",
  ["cahiers agricultures"] = "Cah. Agric.",
  ["cahiers alsaciens d'archéologie, d'art et d'histoire"] = "Cah Alsac Archeol Art Hist",
  ["cahiers alsaciens d’archeologie, d’art et d’histoire"] = "Cah Alsac Archeol Art Hist",
  ["cahiers alsaciens d’archéologie, d’art et d’histoire"] = "CAAH",
  ["cahiers archéologiques"] = "CArch",
  ["cahiers archéologiques: fin de l’antiquité et moyen âge"] = "CArch",
  ["cahiers art et science"] = "Cahiers Art Sci.",
  ["cahiers bruxellois"] = "Cah Brux",
  ["cahiers confrontation"] = "Cah Confront",
  ["cahiers d anesthesiologie"] = "Cah Anesthesiol",
  ["cahiers d'alexandrie"] = "Cah Alex",
  ["cahiers d'anesthésiologie"] = "Cah Anesthesiol",
  ["cahiers d'archéologie et d'histoire du berry"] = "Cah Archeol Hist Berry",
  ["cahiers d'études africaines"] = "Cah Etud Afr",
  ["cahiers d'études hongroises et finlandaises"] = "Cah Etud Hong Finl",
  ["cahiers d'études médiévales (montréal, québec)"] = "Cah Etud Mediev",
  ["cahiers d'études sur la méditerranée orientale et le monde turco-iranien"] = "Cah Mediterr Orient Monde Turco Iran",
  ["cahiers d'histoire"] = "Cah Hist",
  ["cahiers d'histoire (espaces marx (association))"] = "Cah Hist Espaces Marx Assoc",
  ["cahiers d'histoire de la seconde guerre mondiale"] = "Cah Hist Second Guerr Mond",
  ["cahiers d'histoire mondiale. journal of world history"] = "Cah Hist Mond",
  ["cahiers d'homéopathie et de thérapeutique comparée"] = "Cah  Homeopath Ther Comp",
  ["cahiers d'odonto-stomatologie"] = "Cah Odontostomatol Touraine",
  ["cahiers danthropologie"] = "Cah Anthropol",
  ["cahiers dánthropologie"] = "Cah Anthropol",
  ["cahiers de biologie marine"] = "Cah Biol Mar",
  ["cahiers de byrsa"] = "CahByrsa",
  ["cahiers de civilisation alpine"] = "Cah Civilis Alp",
  ["cahiers de clio"] = "Cah Clio",
  ["cahiers de fanjeaux"] = "Cah Fanjeaux",
  ["cahiers de geographie de quebec"] = "Cah Geogr Que",
  ["cahiers de geographie du quebec"] = "Cah Geogr Que",
  ["cahiers de géographie de québec"] = "Cah Geogr Que",
  ["cahiers de géographie du québec"] = "Cah Geogr Que",
  ["cahiers de karnak"] = "CahKarnak",
  ["cahiers de la céramique égyptienne"] = "CahCerEg",
  ["cahiers de la délégation française en iran"] = "CahDelFrIran",
  ["cahiers de lacademie de bretagne"] = "Cah Acad Bretagne",
  ["cahiers de lácadémie de bretagne"] = "Cah Acad Bretagne",
  ["cahiers de logique et d’epistemologie"] = "Cah. Log. Epistemol.",
  ["cahiers de l’institut du moyen âge grec et latin"] = "CIMA",
  ["cahiers de medecine"] = "Cah. Med.",
  ["cahiers de medecine inter professionnelle"] = "Cah Med Inter Prof",
  ["cahiers de médecine interprofessionnelle"] = "Cah Med Inter Prof",
  ["cahiers de numismatique: bulletin de la société d’études numismatiques et archéologiques"] = "CahNum",
  ["cahiers de nutrition et de dietetique"] = "Cah. Nutr. Diet.",
  ["cahiers de pédiatrie sociale"] = "Cah Pediatr Soc",
  ["cahiers de philosophie politique et juridique"] = "CPhPJ",
  ["cahiers de physique"] = "Cah. Phys.",
  ["cahiers de prothese"] = "Cah. Prothese",
  ["cahiers de psychiatrie"] = "Cah Psychiatr",
  ["cahiers de recherche sociologique"] = "Cah Rech Sociol",
  ["cahiers de recherches de l’institut de papyrologie et d’égyptologie de lille iii"] = "CRIPEL",
  ["cahiers de recherches de l’institut de papyrologie et d’égyptologie de lille. sociétés urbaines en égypte et au soudan"] = "CahPEg",
  ["cahiers de sexologie clinique"] = "Cah Sexol Clin",
  ["cahiers de sociologie et de demographie medicales"] = "Cah Sociol Demogr Med",
  ["cahiers de sociologie et de démographie médicales"] = "Cah Sociol Demogr Med",
  ["cahiers de sociologie économique et culturelle, ethnopsychologie"] = "Cah Sociol Econ Cult",
  ["cahiers de topologie et geometrie differentielle categoriques"] = "Cah. Topol. Geom. Differ. Categ.",
  ["cahiers de topologie et géométrie différentielle catégoriques"] = "Cahiers Topologie Géom. Différentielle Catég.",
  ["cahiers de tunisie"] = "CahTun",
  ["cahiers des ameriques latines (paris, france : 1985)"] = "Cah Am Lat",
  ["cahiers des amériques latines (paris, france : 1985)"] = "Cah Am Lat",
  ["cahiers des annales de normandie. annales de normandie"] = "Cah Ann Normandie",
  ["cahiers des etudes anciennes"] = "Cah Etud Anc",
  ["cahiers des études anciennes"] = "Cah Etud Anc",
  ["cahiers des religions africaines"] = "Cah Relig Afr",
  ["cahiers des sciences humaines"] = "Cah Sci Hum",
  ["cahiers des études anciennes"] = "CEA",
  ["cahiers du centre de logique"] = "Cahiers Centre Logique",
  ["cahiers du centre de recherches et d'études sur paris et l'ile-de-france (crepif)"] = "Cah Cent Rech Etud Paris Lle de France",
  ["cahiers du centre de recherches et d’etudes sur paris et l’ile-de-france (crepif)"] = "Cah Cent Rech Etud Paris Lle de France",
  ["cahiers du centre d’études chypriotes"] = "CCEC",
  ["cahiers du centre d’études de recherche opérationnelle"] = "Cahiers Centre études Rech. Opér.",
  ["cahiers du centre georges radet"] = "CCGR",
  ["cahiers du centre gustave glotz"] = "CCG",
  ["cahiers du centre gustave-glotz. revue reconnue par le cnrs"] = "CahGlotz",
  ["cahiers du college de medecine des hopitaux de paris"] = "Cah. Coll. Med. Hop. Paris",
  ["cahiers du groupe de recherches sur l’armée romaine et les provinces"] = "CahArmeeRom",
  ["cahiers du groupe interdisciplinaire du théâtre antique (cahiers du gita)"] = "CGITA",
  ["cahiers du monde russe"] = "Cah Monde Russe",
  ["cahiers du monde russe et sovietique"] = "Cah Monde Russe Sov",
  ["cahiers du monde russe et soviétique"] = "Cah Monde Russe Sov",
  ["cahiers du musée champollion. histoire et archéologie"] = "CahMusChampollion",
  ["cahiers du nursing"] = "Cah. Nurs.",
  ["cahiers du seminaire de philosophie"] = "Cah Semin Philos",
  ["cahiers du séminaire de philosophie"] = "Cah Semin Philos",
  ["cahiers d’alexandrie"] = "Cah Alex",
  ["cahiers d’anesthesiologie"] = "Cah. Anesthesiol.",
  ["cahiers d’archeologie et d’histoire du berry"] = "Cah Archeol Hist Berry",
  ["cahiers d’archéologie subaquatique"] = "CahASubaqu",
  ["cahiers d’economie et sociologie rurales"] = "Cah. Econ. Sociologie Rurales",
  ["cahiers d’economie politique"] = "Cah. Econ. Politique",
  ["cahiers d’etudes africaines"] = "Cah Etud Afr",
  ["cahiers d’etudes medievales (montreal, quebec)"] = "Cah Etud Mediev",
  ["cahiers d’etudes sur la mediterranee orientale et le monde turco-iranien / fondation nationale des science politiques, centre d’etudes et de recherches internationales"] = "Cah Mediterr Orient Monde Turco Iran",
  ["cahiers d’histoire"] = "Cah Hist",
  ["cahiers d’histoire (espaces marx (association))"] = "Cah Hist Espaces Marx Assoc",
  ["cahiers d’histoire de la seconde guerre mondiale / ministere de l’education nationale, archives generales du royaume, centre de recherches et d’etudes historiques de la seconde guerre mondiale"] = "Cah Hist Second Guerr Mond",
  ["cahiers d’histoire mondiale. journal of world history"] = "Cah Hist Mond",
  ["cahiers d’histoire publ. par les univ. de clermont-lyon-grenoble"] = "CH",
  ["cahiers d’informations techniques / revue de metallurgie"] = "Cah. Inf. Tech./Rev Metall",
  ["cahiers d’odonto-stomatologie"] = "Cah. Odontostomatol. (Touraine)",
  ["cahiers economiques de bruxelles"] = "Cah. Econ. Bruxelles",
  ["cahiers economiques et sociaux"] = "Cah. Econ. Soc.",
  ["cahiers économiques de bruxelles"] = "Cah Econ Brux",
  ["cahiers économiques et sociaux"] = "Cah Econ Soc",
  ["cahiers ferdinand de saussure: review suisse de linguistique générale"] = "CFS",
  ["cahiers geographiques de rouen"] = "Cah Geogr Rouen",
  ["cahiers geologiques"] = "Cah Geol",
  ["cahiers géologiques"] = "Cah Geol",
  ["cahiers internationaux d'histoire économique et sociale"] = "Cah Int Hist Econ Soc",
  ["cahiers internationaux d’histoire economique et sociale"] = "Cah Int Hist Econ Soc",
  ["cahiers laennec"] = "Cah Laennec",
  ["cahiers laënnec"] = "Cah Laennec",
  ["cahiers leopold delisle"] = "Cah Leopold Delisle",
  ["cahiers léopold delisle"] = "Cah Leopold Delisle",
  ["cahiers ligures de préhistoire et de protohistoire"] = "CahLig",
  ["cahiers mathématiques de l’école polytechnique fédérale de lausanne"] = "Cahiers Math. École Polytech. Féd. Lausanne",
  ["cahiers medicaux de l’union francaise; revue medico-chirurgicale de la france d’outre-mer"] = "Cah Medicaux Union Fr",
  ["cahiers medicaux lyonnais"] = "Cah Med Lyon",
  ["cahiers médicaux d'auvergne"] = "Cah Med Auvergne",
  ["cahiers médicaux de l'union française; revue médico-chirurgicale de la france d'outre-mer"] = "Cah Med Union Fr",
  ["cahiers médicaux lyonnais"] = "Cah Med Lyon",
  ["cahiers o. r. s. t. o. m. serie sciences humaines"] = "Cah Orstom (Sci Hum)",
  ["cahiers o. r. s. t. o. m. série sciences humaines"] = "Cah Orstom (Sci Hum)",
  ["cahiers odonto-stomatologiques"] = "Cah Odontostomatol",
  ["cahiers quebecois de demographie / association des demographes du quebec"] = "Cah Que Demogr",
  ["cahiers québécois de démographie"] = "Cah Que Demogr",
  ["cahiers r. m. f.: bulletin medical-guide du praticien-revue medicale francaise"] = "Cah Rmf Bull Med Guide Pract Rev",
  ["cahiers r. m. f.: bulletin médical-guide du praticien-revue médicale française"] = "Cah Rmf Bull Med Guide Pract Rev",
  ["cahiers rhodaniens"] = "CahRhod",
  ["cahiers techniques afro"] = "Cah Tech AFRO",
  ["cahiers victoriens & edouardiens"] = "Cah Victor Edouardiens",
  ["cahiers victoriens & édouardiens"] = "Cah Victor Edouardiens",
  ["cahiers vilfredo pareto"] = "Cah Vilfredo Pareto",
  ["cailiao baohu"] = "Cailiao Baohu",
  ["cailiao rechuli xuebao"] = "Cailiao Rechuli Xuebao",
  ["cairo today"] = "Cairo Today",
  ["cais faculty newsletter"] = "CAIS Fac Newsl",
  ["cal [magazine] certified akers laboratories"] = "CAL",
  ["calabar journal of health sciences"] = "Calabar J Health Sci",
  ["calcified tissue international"] = "Calcif Tissue Int",
  ["calcified tissue research"] = "Calcif Tissue Res",
  ["calcium binding proteins"] = "Calcium Bind Proteins",
  ["calcium signaling (santa clara, calif.)"] = "Calcium Signal (St Clara)",
  ["calcoli sublimi—collana di analisi matematica"] = "Calcoli Sublimi",
  ["calcolo. a quarterly on numerical analysis and theory of computation"] = "Calcolo",
  ["calculus of variations and partial differential equations"] = "Calc. Var. Partial Differential Equations",
  ["calcutta mathematical society"] = "Rev. Bull. Calcutta Math. Soc.",
  ["calcutta mathematical society. review bulletin"] = "Rev. Bull. Calcutta Math. Soc.",
  ["calcutta medical journal"] = "Calcutta Med J",
  ["calcutta medical review"] = "Calcutta Med Rev",
  ["calcutta statistical association bulletin"] = "Calcutta Stat. Assoc. Bull.",
  ["california agriculture"] = "Calif Agric (Berkeley)",
  ["california and western medicine"] = "Cal West Med",
  ["california anthropologist"] = "Calif Anthropol",
  ["california archaeology"] = "Calif. Archaeol.",
  ["california code of regulations"] = "Calif Code Regul",
  ["california counts"] = "Calif Counts",
  ["california counts / public policy institute of california"] = "Calif Counts",
  ["california fish and game"] = "Calif Fish Game",
  ["california historical quarterly"] = "Calif Hist Q",
  ["california historical society quarterly"] = "Calif Hist Soc Q",
  ["california hospitals"] = "Calif Hosp",
  ["california journal"] = "Calif J",
  ["california journal of politics and policy"] = "Calif J Politics Policy",
  ["california law review"] = "Calif Law Rev",
  ["california magazine (beverly hills, calif.)"] = "Calif Mag",
  ["california management review"] = "Calif Manage Rev",
  ["california medicine"] = "Calif Med",
  ["california nurse"] = "Calif Nurse",
  ["california public health report. california. department of public health"] = "Calif Public Health Rep",
  ["california series in public anthropology"] = "Calif Ser Public Anthropol",
  ["california sociologist"] = "Calif Sociol",
  ["california state bar journal"] = "Calif State Bar J",
  ["california state journal of medicine"] = "Cal State J Med",
  ["california studies in classical antiquity"] = "CalifStClAnt",
  ["california trial lawyers journal"] = "Calif Trial Lawyers J",
  ["california western international law journal"] = "Calif West Int Law J",
  ["california western law review"] = "Calif West Law Rev",
  ["california's health"] = "Calif Health",
  ["californian journal of health promotion"] = "Calif J Health Promot",
  ["california’s health"] = "Calif Health",
  ["calorimetrie et analyse thermique"] = "Calorim. Anal. Therm.",
  ["calphad - computer coupling of phase diagrams and thermochemistry"] = "Calphad",
  ["calphad ; computer coupling of phase diagrams and thermochemistry"] = "CALPHAD",
  ["calphad: computer coupling of phase diagrams and thermochemistry"] = "CALPHAD: Comput. Coupling Phase Diagrams Thermochem.",
  ["calyx (toronto, ont.)"] = "Calyx",
  ["calyx (toronto, ont.)."] = "Calyx",
  ["cambios revista médica"] = "Cambios Rev Med",
  ["cambodge soir"] = "Cambodge Soir",
  ["cambridge aerospace series"] = "Camb. Aerosp. Ser.",
  ["cambridge anthropology : a journal of the department of social anthropology, cambridge university"] = "Camb Anthropol",
  ["cambridge archaeological journal"] = "Cambridge Archaeol. J.",
  ["cambridge companions to philosophy"] = "Camb. Companions Philos.",
  ["cambridge economic policy review"] = "Cambridge Econ. Pol. Rev.",
  ["cambridge elements. elements in applied category theory"] = "Cambridge Elem. Elem. Appl. Categ. Theory",
  ["cambridge international series on parallel computation"] = "Cambridge Int. Ser. Parallel Comput.",
  ["cambridge introductions to philosophy"] = "Camb. Introd. Philos.",
  ["cambridge journal for junior scientists"] = "Cambridge J. Jr. Scientists",
  ["cambridge journal of economics"] = "Cambridge J Econ",
  ["cambridge journal of mathematics"] = "Cambridge J. Math.",
  ["cambridge lecture notes in physics"] = "Cambridge Lecture Notes Phys.",
  ["cambridge library collection"] = "Camb. Libr. Collect.",
  ["cambridge mathematical library"] = "Cambridge Math. Lib.",
  ["cambridge mathematical textbooks"] = "Cambridge Math. Textbooks",
  ["cambridge monographs on applied and computational mathematics"] = "Cambridge Monogr. Appl. Comput. Math.",
  ["cambridge monographs on atomic, molecular and chemical physics"] = "Camb. Monogr. At. Mol. Chem. Phys.",
  ["cambridge monographs on mathematical physics"] = "Cambridge Monogr. Math. Phys.",
  ["cambridge monographs on mechanics"] = "Cambridge Monogr. Mech.",
  ["cambridge monographs on particle physics, nuclear physics and cosmology"] = "Camb. Monogr. Part. Phys. Nucl. Phys. Cosmol.",
  ["cambridge monographs on physics"] = "Cambridge Monogr. Phys.",
  ["cambridge municipal code 1988 : a codification of the general ordinances of the city of cambridge, massachusetts / codified, indexed and published by book publishing company. cambridge (mass.)"] = "Camb Munic Code 1988 Camb Mass",
  ["cambridge municipal code 1988 : a codification of the general ordinances of the city of cambridge, massachusetts. cambridge (mass.)"] = "Camb Munic Code 1988 Camb Mass",
  ["cambridge nonlinear science series"] = "Cambridge Nonlinear Sci. Ser.",
  ["cambridge observing handbooks for research astronomers"] = "Camb. Obs. Handb. Res. Astron.",
  ["cambridge philosophy classics"] = "Camb. Philos. Class.",
  ["cambridge prisms: coastal futures"] = "Cambridge Prisms: Coastal Futures",
  ["cambridge prisms: extinction"] = "Cambridge Prisms: Extinction",
  ["cambridge prisms: global mental health"] = "Cambridge Prisms: Global Mental Health",
  ["cambridge prisms: plastics"] = "Cambridge Prisms: Plast.",
  ["cambridge prisms: precision medicine"] = "Cambridge Prisms: Precis. Med.",
  ["cambridge prisms: water"] = "Cambridge Prisms: Water",
  ["cambridge quarterly of healthcare ethics"] = "Cambridge Q. Healthcare Ethics",
  ["cambridge quarterly of healthcare ethics : cq : the international journal of healthcare ethics committees"] = "Camb Q Healthc Ethics",
  ["cambridge series in statistical and probabilistic mathematics"] = "Camb. Ser. Stat. Probab. Math.",
  ["cambridge series on human-computer interaction"] = "Camb. Ser. Hum. Comput. Interact.",
  ["cambridge series on statistical and probabilistic mathematics"] = "Camb. Ser. Stat. Probab. Math.",
  ["cambridge solid state science series"] = "Cambridge Solid State Sci. Ser.",
  ["cambridge studies in advanced mathematics"] = "Cambridge Stud. Adv. Math.",
  ["cambridge studies in applied econometrics"] = "Cambridge Stud. Appl. Econom.",
  ["cambridge studies in mathematical biology"] = "Cambridge Stud. Math. Biol.",
  ["cambridge studies in modern optics"] = "Cambridge Stud. Modern Optics",
  ["cambridge studies in philosophy"] = "Cambridge Stud. Philos.",
  ["cambridge studies in probability, induction, and decision theory"] = "Cambridge Stud. Probab. Induc. Decis. Theory",
  ["cambridge texts in applied mathematics"] = "Cambridge Texts Appl. Math.",
  ["cambridge texts in the history of philosophy"] = "Camb. Texts Hist. Philos.",
  ["cambridge tracts in mathematics"] = "Cambridge Tracts in Math.",
  ["cambridge tracts in theoretical computer science"] = "Cambridge Tracts Theoret. Comput. Sci.",
  ["cambridge university medical society magazine"] = "Cambridge Univ. Med. Soc. Mag.",
  ["cambridge-iisc series"] = "Camb. IISc Ser.",
  ["campbell law review"] = "Campbell Law Rev",
  ["campbell systematic reviews"] = "Campbell Syst. Rev.",
  ["campi del sapere/filosofia"] = "Campi Sapere Filos.",
  ["camsi journal. journal acemi"] = "CAMSI J",
  ["can ye ke xue = canye kexue = acta sericologica sinica"] = "Can Ye Ke Xue",
  ["canada communicable disease report"] = "Can. Commun. Dis. Rep.",
  ["canada communicable disease report = relevé des maladies transmissibles au canada"] = "Can Commun Dis Rep",
  ["canada communicable disease report weekly"] = "Can Commun Dis Rep Wkly",
  ["canada diseases weekly report"] = "Can. Dis. Wkly. Rep.",
  ["canada diseases weekly report = rapport hebdomadaire des maladies au canada"] = "Can Dis Wkly Rep",
  ["canada supreme court reports. recueil des arrets de la cour supreme du canada. canada. supreme court"] = "Can Supreme Court Rep Can Supreme Court",
  ["canada supreme court reports. recueil des arrêts de la cour suprême du canada. canada. supreme court"] = "Can Supreme Court Rep Can Supreme Court",
  ["canada's health & welfare"] = "Can Health Welf",
  ["canada's mental health"] = "Can Ment Health",
  ["canada’s mental health"] = "Can. Ment. Health",
  ["canadian acoustics"] = "Can. Acoust.",
  ["canadian acoustics = acoustique canadienne"] = "Can Acoust",
  ["canadian aeronautics and space journal"] = "Can. Aeronaut. Space J.",
  ["canadian aeronautics and space journal. le journal aeronautique et spatial du canada"] = "Can Aeronaut Space J",
  ["canadian aeronautics and space journal. le journal aéronautique et spatial du canada"] = "Can Aeronaut Space J",
  ["canadian agricultural engineering"] = "Can. Agric. Eng.",
  ["canadian aids news"] = "Can AIDS News",
  ["canadian air force journal"] = "Can. Air Force J.",
  ["canadian alpine journal"] = "Can. Alp. J.",
  ["canadian anaesthetists society journal"] = "Can. Anaesth. Soc. J.",
  ["canadian anaesthetists' society journal"] = "Can Anaesth Soc J",
  ["canadian anaesthetists’ society journal"] = "Can Anaesth Soc J",
  ["canadian and international education. education canadienne et internationale"] = "Can Int Educ",
  ["canadian and international education. éducation canadienne et internationale"] = "Can Int Educ",
  ["canadian applied mathematics quarterly"] = "Can. Appl. Math. Q.",
  ["canadian association of radiologists journal"] = "Can. Assoc. Radiol. J.",
  ["canadian association of radiologists journal = journal l'association canadienne des radiologistes"] = "Can Assoc Radiol J",
  ["canadian bar journal"] = "Can Bar J",
  ["canadian bee journal"] = "Can. Bee J.",
  ["canadian biosystems engineering"] = "Can. Biosyst. Eng.",
  ["canadian bulletin of cardiovascular nursing"] = "Can Bull Cardiovasc Nurs",
  ["canadian bulletin of medical history"] = "Can. Bull. Med. Hist.",
  ["canadian bulletin of medical history = bulletin canadien d'histoire de la médecine"] = "Can Bull Med Hist",
  ["canadian business economics"] = "Can. Bus. Econ.",
  ["canadian cancer conference"] = "Can. Cancer Conf.",
  ["canadian ceramics quarterly"] = "Can. Ceram. Q.",
  ["canadian chemical news"] = "Can. Chem. News",
  ["canadian chemical processing"] = "Can. Chem. Process.",
  ["canadian chemical transactions"] = "Can Chem Trans",
  ["canadian collection"] = "Can. Collect.",
  ["canadian critical care nursing journal"] = "Can Crit Care Nurs J",
  ["canadian dental hygienist"] = "Can. Dent. Hyg.",
  ["canadian dimension"] = "Can Dimens",
  ["canadian doctor"] = "Can Doct",
  ["canadian electrical engineering journal"] = "Can. Electr. Eng. J.",
  ["canadian entomologist"] = "Can. Entomol.",
  ["canadian ethnic studies"] = "Can Ethn Stud",
  ["canadian family law quarterly"] = "Can Fam Law Q",
  ["canadian family physician"] = "Can. Fam. Physician",
  ["canadian family physician medecin de famille canadien"] = "Can Fam Physician",
  ["canadian family physician médecin de famille canadien"] = "Can Fam Physician",
  ["canadian field-naturalist"] = "Can. Field-Nat.",
  ["canadian forces dental services bulletin"] = "Can Forces Dent Serv Bull",
  ["canadian forces dental services quarterly"] = "Can Forces Dent Serv Q",
  ["canadian forestry journal"] = "Can. For. J.",
  ["canadian geographer"] = "Can. Geogr.",
  ["canadian geographic"] = "Can Geogr (Ottowa)",
  ["canadian geographical journal"] = "Can. Geogr. J.",
  ["canadian geotechnical journal"] = "Can. Geotech. J.",
  ["canadian geriatrics journal : cgj"] = "Can Geriatr J",
  ["canadian hiv-aids policy & law newsletter"] = "Can HIV AIDS Policy Law Newsl",
  ["canadian hiv/aids policy & law review"] = "Can HIV AIDS Policy Law Rev",
  ["canadian hiv/aids policy & law review / canadian hiv/aids legal network"] = "Can HIV AIDS Policy Law Rev",
  ["canadian home economics journal"] = "Can. Home Econ. J.",
  ["canadian hospital"] = "Can Hosp",
  ["canadian human rights advocate"] = "Can Hum Rights Advocate",
  ["canadian human rights yearbook = annuaire canadien des droits de la personne"] = "Can Hum Rights Yearb",
  ["canadian institute of food science and technology journal"] = "Can. Inst. Food Sci. Technol. J.",
  ["canadian institute of food technology journal"] = "Can. Inst. Food Technol. J.",
  ["canadian journal of aboriginal community-based hiv/aids research"] = "Canadi J Aborig Community Based HIV/AIDS Res",
  ["canadian journal of action research"] = "Can. J. Action Res.",
  ["canadian journal of addiction"] = "Can. J. Addict.",
  ["canadian journal of administrative sciences"] = "Can. J. Admin. Sci.",
  ["canadian journal of adminstrative sciences"] = "Can. J. Adm. Sci.",
  ["canadian journal of african studies"] = "Can J Afr Stud",
  ["canadian journal of agricultural economics"] = "Can. J. Agric. Econ.",
  ["canadian journal of agricultural science"] = "Can. J. Agr. Sci.",
  ["canadian journal of anaesthesia"] = "Can. J. Anaesth.",
  ["canadian journal of anaesthesia = journal canadien d'anesthésie"] = "Can J Anaesth",
  ["canadian journal of analytical sciences and spectroscopy"] = "Can. J. Anal. Sci. Spectros.",
  ["canadian journal of anesthesia"] = "Can. J. Anesth.",
  ["canadian journal of animal science"] = "Can. J. Anim. Sci.",
  ["canadian journal of anthropology"] = "Can. J. Anthropol.",
  ["canadian journal of applied linguistics"] = "Can. J. Appl. Ling.",
  ["canadian journal of applied physiology"] = "Can. J. Appl. Physiol.",
  ["canadian journal of applied physiology = revue canadienne de physiologie appliquée"] = "Can J Appl Physiol",
  ["canadian journal of applied sciences"] = "Can. J. Appl. Sci.",
  ["canadian journal of applied spectroscopy"] = "Can. J. Appl. Spectrosc.",
  ["canadian journal of applied sport sciences"] = "Can. J. Appl. Sport Sci.",
  ["canadian journal of applied sport sciences. journal canadien des sciences appliquees au sport"] = "Can J Appl Sport Sci",
  ["canadian journal of applied sport sciences. journal canadien des sciences appliquées au sport"] = "Can J Appl Sport Sci",
  ["canadian journal of archaeology"] = "Can. J. Archaeol.",
  ["canadian journal of archaeology = journal canadien d'archéologie"] = "Can J Archaeol",
  ["canadian journal of art therapy"] = "Can. J. Art Ther.",
  ["canadian journal of arthropod identification"] = "Can J Arthropod Identif",
  ["canadian journal of behavioural science"] = "Can. J. Behav. Sci.",
  ["canadian journal of behavioural science. revue canadienne des sciences du comportement"] = "Can J Behav Sci",
  ["canadian journal of biochemistry"] = "Can. J. Biochem.",
  ["canadian journal of biochemistry and cell biology"] = "Can. J. Biochem. Cell Biol.",
  ["canadian journal of biochemistry and cell biology = revue canadienne de biochimie et biologie cellulaire"] = "Can J Biochem Cell Biol",
  ["canadian journal of biochemistry and physiology"] = "Can J Biochem Physiol",
  ["canadian journal of bioethics"] = "Can. J. Bioethics",
  ["canadian journal of bioethics = revue canadienne de bioéthique"] = "Can J Bioeth",
  ["canadian journal of botany"] = "Can. J. Bot.",
  ["canadian journal of botany. journal canadien de botanique"] = "Can J Bot",
  ["canadian journal of cardiology"] = "Can. J. Cardiol",
  ["canadian journal of cardiovascular nursing"] = "Can. J. Cardiovasc. Nurs.",
  ["canadian journal of cardiovascular nursing = journal canadien en soins infirmiers cardio-vasculaires"] = "Can J Cardiovasc Nurs",
  ["canadian journal of career development"] = "Can. J. Career Dev.",
  ["canadian journal of catholic history"] = "Can. J. Cathol. Hist.",
  ["canadian journal of chemical engineering"] = "Can. J. Chem. Eng.",
  ["canadian journal of chemistry"] = "Can. J. Chem.",
  ["canadian journal of civil engineering"] = "Can. J. Civ. Eng.",
  ["canadian journal of clinical pharmacology"] = "Can. J. Clin. Pharmacol.",
  ["canadian journal of communication"] = "Can. J. Commun.",
  ["canadian journal of community dentistry"] = "Can. J. Community Dent.",
  ["canadian journal of community mental health"] = "Can. J. Community Mental Health",
  ["canadian journal of community mental health = revue canadienne de santé mentale communautaire"] = "Can J Commun Ment Health",
  ["canadian journal of comparative medicine"] = "Can. J. Comp. Med.",
  ["canadian journal of comparative medicine (gardenvale, quebec)"] = "Can J Comp Med",
  ["canadian journal of comparative medicine (gardenvale, québec)"] = "Can J Comp Med",
  ["canadian journal of comparative medicine : revue canadienne de médecine comparée"] = "Can J Comp Med",
  ["canadian journal of comparative medicine and veterinary science"] = "Can J Comp Med Vet Sci",
  ["canadian journal of comparative medicine. revue canadienne de medecine comparee"] = "Can J Comp Med",
  ["canadian journal of criminology"] = "Can. J. Criminol.",
  ["canadian journal of criminology and criminal justice"] = "Can. J. Criminol. Crim. Justice",
  ["canadian journal of criminology. revue canadienne de criminologie"] = "Can J. Criminol",
  ["canadian journal of dental hygiene"] = "Can. J. Dent. Hyg.",
  ["canadian journal of dental hygiene : cjdh = journal canadien de l'hygiène dentaire : jchd"] = "Can J Dent Hyg",
  ["canadian journal of dermatology"] = "Can. J. Dermatol.",
  ["canadian journal of development studies"] = "Can. J. Dev. Stud.",
  ["canadian journal of diabetes"] = "Can J Diabetes",
  ["canadian journal of dietetic practice and research"] = "Can. J. Diet. Pract. Res.",
  ["canadian journal of dietetic practice and research : a publication of dietitians of canada = revue canadienne de la pratique et de la recherche en diététique : une publication des diététistes du canada"] = "Can J Diet Pract Res",
  ["canadian journal of earth sciences"] = "Can. J. Earth Sci.",
  ["canadian journal of economics"] = "Can. J.  Econ.",
  ["canadian journal of education"] = "Can. J. Educ.",
  ["canadian journal of educational administration and policy"] = "Can. J. Educ. Adm. Policy",
  ["canadian journal of electrical and computer engineering"] = "Can. J. Electr. Comput. Eng.",
  ["canadian journal of emergency management"] = "Can. J. Emerg. Manage.",
  ["canadian journal of emergency medicine"] = "Can. J. Emerg. Med.",
  ["canadian journal of european and russian studies"] = "Can. J. Eur. Russ. Stud.",
  ["canadian journal of experimental psychology"] = "Can. J. Exp. Psychol.",
  ["canadian journal of experimental psychology = revue canadienne de psychologie expérimentale"] = "Can J Exp Psychol",
  ["canadian journal of exploration geophysics"] = "Can. J. Explor. Geophys.",
  ["canadian journal of family and youth"] = "Can. J. Fam. Youth",
  ["canadian journal of family law. revue canadienne de droit familial"] = "Can J Fam Law",
  ["canadian journal of film studies"] = "Can. J. Film Stud.",
  ["canadian journal of fisheries and aquatic sciences"] = "Can. J. Fish. Aquat. Sci.",
  ["canadian journal of fisheries and aquatic sciences. journal canadien des sciences halieutiques et aquatiques"] = "Can J Fish Aquat Sci",
  ["canadian journal of forest research"] = "Can. J. For. Res.",
  ["canadian journal of forest research. journal canadien de la recherche forestiere"] = "Can J For Res",
  ["canadian journal of forest research. journal canadien de la recherche forestière"] = "Can J For Res",
  ["canadian journal of gastroenterology"] = "Can. J. Gastroenterol.",
  ["canadian journal of gastroenterology & hepatology"] = "Can J Gastroenterol Hepatol",
  ["canadian journal of gastroenterology = journal canadien de gastroenterologie"] = "Can J Gastroenterol",
  ["canadian journal of gastroenterology and hepatology"] = "Can. J. Gastroenterol. Hepatol.",
  ["canadian journal of genetics and cytology"] = "Can. J. Genet. Cytol.",
  ["canadian journal of genetics and cytology. journal canadien de genetique et de cytologie"] = "Can J Genet Cytol",
  ["canadian journal of genetics and cytology. journal canadien de génétique et de cytologie"] = "Can J Genet Cytol",
  ["canadian journal of health history"] = "Can. J. Heath Hist.",
  ["canadian journal of health technologies"] = "Can. J. Health Technol.",
  ["canadian journal of higher education"] = "Can. J. Higher Educ.",
  ["canadian journal of history"] = "Can J Hist",
  ["canadian journal of history of sport"] = "Can. J. Hist. Sport",
  ["canadian journal of history of sport = revue canadienne de l'histoire des sports"] = "Can J Hist Sport",
  ["canadian journal of history of sport and physical education"] = "Can J Hist Sport Phys Educ",
  ["canadian journal of history. annales canadiennes d'histoire"] = "Can J Hist",
  ["canadian journal of hospital pharmacy"] = "Can. J. Hosp. Pharm.",
  ["canadian journal of human sexuality"] = "Can. J. Hum. Sex.",
  ["canadian journal of infection control"] = "Can. J. Infect. Control",
  ["canadian journal of infectious diseases and medical microbiology"] = "Can. J. Infect. Dis. Med. Microbiol.",
  ["canadian journal of information and library science"] = "Can. J. Inf. Lib. Sci.",
  ["canadian journal of irish studies"] = "Can. J. Ir. Stud.",
  ["canadian journal of italian studies"] = "Can J Ital Stud",
  ["canadian journal of kidney health and disease"] = "Can J Kidney Health Dis",
  ["canadian journal of latin american and caribbean studies"] = "Can. J. Lat. Am. Caribb. Stud.",
  ["canadian journal of latin american and caribbean studies = revue canadienne des études latino-américaines et caraïbes"] = "Can J Lat Am Caribb Stud",
  ["canadian journal of law and jurisprudence"] = "Can. J. Law Jurisprud.",
  ["canadian journal of law and society"] = "Can. J. Law Soc.",
  ["canadian journal of law and society = revue canadienne de droit et société"] = "Can J Law Soc",
  ["canadian journal of learning and technology"] = "Can. J. Learn. Technol.",
  ["canadian journal of linguistics"] = "Can. J. Ling.",
  ["canadian journal of linguistics. la revue canadienne de linguistique"] = "Can J Linguist",
  ["canadian journal of mathematics"] = "Can. J. Math.",
  ["canadian journal of mathematics. journal canadien de mathematiques"] = "Canad. J. Math.",
  ["canadian journal of medical laboratory science"] = "Can. J. Med. Lab. Sci.",
  ["canadian journal of medical radiation technology"] = "Can. J. Med. Radiat. Technol.",
  ["canadian journal of medical sciences"] = "Can J Med Sci",
  ["canadian journal of medical technology"] = "Can J Med Technol",
  ["canadian journal of medicine and surgery"] = "Can. J. Med. Surg.",
  ["canadian journal of microbiology"] = "Can J Microbiol",
  ["canadian journal of mineralogy and petrology"] = "Can. J. Mineral. Petrol.",
  ["canadian journal of music"] = "Can. J. Music",
  ["canadian journal of neurological sciences"] = "Can. J. Neurol. Sci.",
  ["canadian journal of neurology"] = "Can. J. Neurol.",
  ["canadian journal of neuroscience nursing"] = "Can J Neurosci Nurs",
  ["canadian journal of nonprofit and social economy research"] = "Can. J. Nonprofit Social Econ. Res.",
  ["canadian journal of nursing administration"] = "Can J Nurs Adm",
  ["canadian journal of nursing informatics"] = "Can. J. Nurs. Inf.",
  ["canadian journal of nursing leadership"] = "Can J Nurs Leadersh",
  ["canadian journal of nursing research"] = "Can. J. Nurs. Res.",
  ["canadian journal of occupational therapy"] = "Can. J. Occup. Ther.",
  ["canadian journal of occupational therapy. revue canadienne d'ergothérapie"] = "Can J Occup Ther",
  ["canadian journal of occupational therapy. revue canadienne d’ergotherapie"] = "Can J Occup Ther",
  ["canadian journal of oncology"] = "Can. J. Oncol.",
  ["canadian journal of ophthalmology"] = "Can. J. Ophthalmol.",
  ["canadian journal of ophthalmology. journal canadien d'ophtalmologie"] = "Can J Ophthalmol",
  ["canadian journal of ophthalmology. journal canadien d’ophtalmologie"] = "Can J Ophthalmol",
  ["canadian journal of optometry"] = "Can. J. Optom.",
  ["canadian journal of otolaryngology"] = "Can. J. Otolaryngol.",
  ["canadian journal of otolaryngology. journal canadien d’otolaryngologie"] = "Can J Otolaryngol",
  ["canadian journal of otolaryngology. journal canadien d’otolaryngologie. supplement"] = "Can J Otolaryngol Suppl",
  ["canadian journal of otolaryngology. supplement"] = "Can J Otolaryngol Suppl",
  ["canadian journal of pain"] = "Can. J. Pain",
  ["canadian journal of pain = revue canadienne de la douleur"] = "Can J Pain",
  ["canadian journal of pathology"] = "Can. J. Pathol.",
  ["canadian journal of pharmaceutical sciences"] = "Can. J. Pharm. Sci.",
  ["canadian journal of philosophy"] = "Can J Philos",
  ["canadian journal of physics"] = "Can. J. Phys.",
  ["canadian journal of physiology and pharmacology"] = "Can J Physiol Pharmacol",
  ["canadian journal of plant pathology"] = "Can. J. Plant Pathol.",
  ["canadian journal of plant pathology. revue canadienne de phytopathologie"] = "Can J Plant Pathol",
  ["canadian journal of plant science"] = "Can. J. Plant Sci.",
  ["canadian journal of plant science. revue canadienne de phytotechnie"] = "Can J Plant Sci",
  ["canadian journal of plastic surgery"] = "Can. J. Plast. Surg.",
  ["canadian journal of political science"] = "Can. J. Polit. Sci.",
  ["canadian journal of political science. revue canadienne de science politique"] = "Can J Polit Sci",
  ["canadian journal of program evaluation"] = "Can. J. Program Eval.",
  ["canadian journal of psychiatric nursing"] = "Can. J. Psychiatr. Nurs.",
  ["canadian journal of psychiatry"] = "Can. J. Psychiatry",
  ["canadian journal of psychiatry. revue canadienne de psychiatrie"] = "Can J Psychiatry",
  ["canadian journal of psychology"] = "Can J Psychol",
  ["canadian journal of public health"] = "Can. J. Public Health",
  ["canadian journal of public health = revue canadienne de santé publique"] = "Can J Public Health",
  ["canadian journal of public health. revue canadienne de sante publique"] = "Can J Public Health",
  ["canadian journal of pure & applied science"] = "Can. J. Pure Appl. Sci.",
  ["canadian journal of radiography, radiotherapy, nuclear medicine"] = "Can. J. Radiogr. Radiother. Nucl. Med.",
  ["canadian journal of regional science"] = "Can. J. Reg. Sci.",
  ["canadian journal of remote sensing"] = "Can. J. Remote Sens.",
  ["canadian journal of remote sensing = journal canadien de télédétection"] = "Can J Remote Sens",
  ["canadian journal of research"] = "Can. J. Res.",
  ["canadian journal of research. section e, medical sciences"] = "Can J Res E Med Sci",
  ["canadian journal of respiratory therapy : cjrt = revue canadienne de la thérapie respiratoire : rctr"] = "Can J Respir Ther",
  ["canadian journal of respiratory, critical care, and sleep medicine"] = "Can. J. Respir. Crit. Care Sleep Med.",
  ["canadian journal of rural medicine"] = "Can. J. Rural Med.",
  ["canadian journal of rural medicine : the official journal of the society of rural physicians of canada = journal canadien de la médecine rurale : le journal officiel de la société de médecine rurale du canada"] = "Can J Rural Med",
  ["canadian journal of school psychology"] = "Can J Sch Psychol",
  ["canadian journal of science, mathematics and technology education"] = "Can. J. Sci. Math. Technol. Educ.",
  ["canadian journal of sociology"] = "Can. J. Sociol.",
  ["canadian journal of sociology = cahiers canadiens de sociologie"] = "Can J Sociol",
  ["canadian journal of sociology-cahiers canadiens de sociologie"] = "Can. J. Sociol.-Cahiers Can. Sociol.",
  ["canadian journal of soil science"] = "Can. J. Soil Sci.",
  ["canadian journal of spectroscopy"] = "Can. J. Spectrosc.",
  ["canadian journal of speech language pathology and audiology"] = "Can. J. Speech Lang. Pathol. Audiol.",
  ["canadian journal of speech-language pathology and audiology : cjslpa = revue canadienne d'orthophonie et d'audiologie : rcoa"] = "Can J Speech Lang Pathol Audiol",
  ["canadian journal of sport science"] = "Can. J. Sport Sci.",
  ["canadian journal of sport sciences"] = "Can. J. Sport Sci.",
  ["canadian journal of sport sciences = journal canadien des sciences du sport"] = "Can J Sport Sci",
  ["canadian journal of statistics"] = "Can. J. Stat.",
  ["canadian journal of surgery"] = "Can. J. Surg.",
  ["canadian journal of surgery. journal canadien de chirurgie"] = "Can J Surg",
  ["canadian journal of transportation"] = "Can. J. Transp.",
  ["canadian journal of undergraduate research"] = "Can. J. Res. Undergrad. Res.",
  ["canadian journal of urban research"] = "Can J Urban Res",
  ["canadian journal of urology"] = "Can. J. Urol.",
  ["canadian journal of veterinary research"] = "Can. J. Vet. Res.",
  ["canadian journal of veterinary research = revue canadienne de recherche vétérinaire"] = "Can J Vet Res",
  ["canadian journal of women and the law"] = "Can. J. Women Law",
  ["canadian journal of women and the law = revue juridique la femme et le droit"] = "Can J Women Law",
  ["canadian journal of zombie science"] = "Can. J. Zombie Sci.",
  ["canadian journal of zoology"] = "Can J Zool",
  ["canadian journal on aging"] = "Can. J. Aging",
  ["canadian journal on aging = la revue canadienne du vieillissement"] = "Can J Aging",
  ["canadian labour"] = "Can Labour",
  ["canadian liver journal"] = "Can. Liver J.",
  ["canadian mathematical bulletin"] = "Can. Math. Bull.",
  ["canadian mathematical bulletin. bulletin canadien de mathematiques"] = "Canad. Math. Bull.",
  ["canadian mathematical society series of monographs and advanced texts"] = "Canad. Math. Soc. Ser. Monogr. Adv. Texts",
  ["canadian medical association journal"] = "Can Med Assoc J",
  ["canadian medical education journal"] = "Can Med Educ J",
  ["canadian medical journal"] = "Can. Med. J.",
  ["canadian metallurgical quarterly"] = "Can. Metall. Q.",
  ["canadian mineralogist"] = "Can. Mineral.",
  ["canadian mining journal"] = "Can. Min. J.",
  ["canadian modern language review"] = "Can. Mod. Lang. Rev.",
  ["canadian nurse"] = "Can. Nurse",
  ["canadian oncology nursing journal"] = "Can. Oncol. Nurs. J.",
  ["canadian oncology nursing journal = revue canadienne de nursing oncologique"] = "Can Oncol Nurs J",
  ["canadian operating room nursing journal"] = "Can Oper Room Nurs J",
  ["canadian packaging"] = "Can. Packag.",
  ["canadian pharmaceutical journal"] = "Can Pharm J",
  ["canadian pharmacists journal"] = "Can. Pharm. J.",
  ["canadian pharmacists journal : cpj = revue des pharmaciens du canada : rpc"] = "Can Pharm J (Ott)",
  ["canadian philosophical reviews"] = "CPhRev",
  ["canadian prosthetics and orthotics journal"] = "Can. Prosthet. Orthot. J.",
  ["canadian psychiatric association journal"] = "Can Psychiatr Assoc J",
  ["canadian psychological review. psychologie canadienne"] = "Can Psychol Rev",
  ["canadian psychologist"] = "Can Psychol",
  ["canadian psychology"] = "Can. Psychol.",
  ["canadian psychology = psychologie canadienne"] = "Can Psychol",
  ["canadian public administration"] = "Can. Public Adm.",
  ["canadian public administration : administration publique du canada"] = "Can Public Adm",
  ["canadian public administration. administration publique du canada"] = "Can Public Adm",
  ["canadian public policy"] = "Can. Public Pol.",
  ["canadian public policy. analyse de politiques"] = "Can Public Policy",
  ["canadian respiratory journal"] = "Can Respir J",
  ["canadian respiratory journal : journal of the canadian thoracic society"] = "Can Respir J",
  ["canadian review of american studies"] = "Can Rev Am Stud",
  ["canadian review of comparative literature"] = "CRCL",
  ["canadian review of sociology = revue canadienne de sociologie"] = "Can Rev Sociol",
  ["canadian review of sociology and anthropology-revue canadienne de sociologie et d anthropologie"] = "Can. Rev. Sociol. Anthropol.-Rev. Can. Sociol. Anthrol.",
  ["canadian services medical journal"] = "Can Serv Med J",
  ["canadian slavonic papers"] = "Can Slavon Pap",
  ["canadian social trends"] = "Can Soc Trends",
  ["canadian social work review = revue canadienne de service social"] = "Can Soc Work Rev",
  ["canadian society of forensic science journal"] = "Can. Soc. Forensic Sci. J.",
  ["canadian statistical review (ottawa, ont.)"] = "Can Stat Rev",
  ["canadian statistical review (ottawa, ont.)."] = "Can Stat Rev",
  ["canadian studies in population"] = "Can Stud Popul",
  ["canadian tax journal"] = "Can. Tax J.",
  ["canadian technical report of fisheries and aquatic sciences"] = "Can Tech Rep Fish Aquat Sci",
  ["canadian urological association journal"] = "Can. Urol. Assoc. J.",
  ["canadian urological association journal = journal de l'association des urologues du canada"] = "Can Urol Assoc J",
  ["canadian veterinary journal"] = "Can. Vet. J.",
  ["canadian water resources journal"] = "Can. Water Resour. J.",
  ["canadian women's studies = les cahiers de la femme"] = "Can Womens Stud",
  ["canadian young scientist journal"] = "Can. Young Sci. J.",
  ["canadian-american slavic studies. revue canadienne-americaine d’etudes slaves"] = "Can Am Slav Stud",
  ["canadian-american slavic studies. revue canadienne-américaine d'études slaves"] = "Can Am Slav Stud",
  ["cancer & metabolism"] = "Cancer Metab",
  ["cancer and clinical oncology"] = "Cancer Clin Oncol",
  ["cancer and metastasis reviews"] = "Cancer Metastasis Rev.",
  ["cancer biochemistry biophysics"] = "Cancer Biochem Biophys",
  ["cancer biology & medicine"] = "Cancer Biol Med",
  ["cancer biology & therapy"] = "Cancer Biol Ther",
  ["cancer biology and therapy"] = "Cancer Biol. Ther.",
  ["cancer biomarkers"] = "Cancer Biomarkers",
  ["cancer biomarkers : section a of disease markers"] = "Cancer Biomark",
  ["cancer biotherapy"] = "Cancer Biother",
  ["cancer biotherapy & radiopharmaceuticals"] = "Cancer Biother Radiopharm",
  ["cancer biotherapy and radiopharmaceuticals"] = "Cancer Biother. Radiopharm.",
  ["cancer bulletin (houston, tex.)"] = "Cancer Bull",
  ["cancer causes & control : ccc"] = "Cancer Causes Control",
  ["cancer causes and control"] = "Cancer Causes Control",
  ["cancer cell"] = "Cancer Cell",
  ["cancer cell & microenvironment"] = "Cancer Cell Microenviron",
  ["cancer cell international"] = "Cancer Cell Int",
  ["cancer cells"] = "Cancer Cells",
  ["cancer cells (cold spring harbor, n.y. : 1989)"] = "Cancer Cells",
  ["cancer chemotheraphy and pharmacology"] = "Cancer Chemother. Pharmacol.",
  ["cancer chemotherapy and biological response modifiers"] = "Cancer Chemother Biol Response Modif",
  ["cancer chemotherapy and pharmacology"] = "Cancer Chemother Pharmacol",
  ["cancer chemotherapy reports"] = "Cancer Chemother Rep",
  ["cancer chemotherapy reports. part 1"] = "Cancer Chemother. Rep.",
  ["cancer chemotherapy reports. part 2"] = "Cancer Chemother Rep 2",
  ["cancer chemotherapy reports. part 3"] = "Cancer Chemother Rep 3",
  ["cancer clinical trials"] = "Cancer Clin Trials",
  ["cancer communications"] = "Cancer Commun",
  ["cancer communications (london, england)"] = "Cancer Commun (Lond)",
  ["cancer control : journal of the moffitt cancer center"] = "Cancer Control",
  ["cancer convergence"] = "Cancer Converg",
  ["cancer cytology"] = "Cancer Cytol",
  ["cancer cytopathology"] = "Cancer Cytopathol",
  ["cancer detection and prevention"] = "Cancer Detect Prev",
  ["cancer detection and prevention. supplement"] = "Cancer Detect. Prev. Suppl.",
  ["cancer detection and prevention. supplement : official publication of the international society for preventive oncology, inc"] = "Cancer Detect Prev Suppl",
  ["cancer discovery"] = "Cancer Discov",
  ["cancer drug delivery"] = "Cancer Drug Deliv",
  ["cancer drug resistance (alhambra, calif.)"] = "Cancer Drug Resist",
  ["cancer epidemiology"] = "Cancer Epidemiol",
  ["cancer epidemiology, biomarkers & prevention : a publication of the american association for cancer research, cosponsored by the american society of preventive oncology"] = "Cancer Epidemiol Biomarkers Prev",
  ["cancer epidemiology, biomarkers and prevention"] = "Cancer Epidemiol. Biomarkers Prev.",
  ["cancer forum"] = "Cancer Forum",
  ["cancer gene therapy"] = "Cancer Gene Ther",
  ["cancer genetics"] = "Cancer Genet",
  ["cancer genetics and cytogenetics"] = "Cancer Genet Cytogenet",
  ["cancer genomics & proteomics"] = "Cancer Genomics Proteomics",
  ["cancer growth and metastasis"] = "Cancer Growth Metastasis",
  ["cancer hallmarks"] = "Cancer Hallm",
  ["cancer health disparities"] = "Cancer Health Disparities",
  ["cancer imaging"] = "Cancer Imaging",
  ["cancer imaging : the official publication of the international cancer imaging society"] = "Cancer Imaging",
  ["cancer immunity"] = "Cancer Immun",
  ["cancer immunity : a journal of the academy of cancer immunology"] = "Cancer Immun",
  ["cancer immunology research"] = "Cancer Immunol Res",
  ["cancer immunology, immunotherapy"] = "Cancer Immunol., Immunother.",
  ["cancer immunology, immunotherapy : cii"] = "Cancer Immunol Immunother",
  ["cancer informatics"] = "Cancer Inform",
  ["cancer investigation"] = "Cancer Invest",
  ["cancer journal"] = "Cancer J.",
  ["cancer journal (sudbury, mass.)"] = "Cancer J",
  ["cancer journal from scientific american"] = "Cancer J. Sci. Am.",
  ["cancer letters"] = "Cancer Lett",
  ["cancer letters (amsterdam, netherlands)"] = "Cancer Lett. (Amsterdam, Neth.)",
  ["cancer letters (new york, ny, united states)"] = "Cancer Lett. (N. Y., NY, U. S.)",
  ["cancer letters (shannon, ireland)"] = "Cancer Lett. (Shannon, Irel.)",
  ["cancer management and research"] = "Cancer Manag Res",
  ["cancer medicine"] = "Cancer Med",
  ["cancer medicine journal"] = "Cancer Med J",
  ["cancer metastasis reviews"] = "Cancer Metastasis Rev",
  ["cancer microenvironment : official journal of the international cancer microenvironment society"] = "Cancer Microenviron",
  ["cancer nanotechnology"] = "Cancer Nanotechnol.",
  ["cancer news"] = "Cancer News",
  ["cancer nursing"] = "Cancer Nurs",
  ["cancer practice"] = "Cancer Pract",
  ["cancer prevention & control : cpc = prévention & contrôle en cancérologie : pcc"] = "Cancer Prev Control",
  ["cancer prevention (philadelphia, pa.)"] = "Cancer Prev",
  ["cancer prevention and control"] = "Cancer Prev. Control",
  ["cancer prevention research (philadelphia, pa.)"] = "Cancer Prev Res (Phila)",
  ["cancer progress"] = "Cancer Prog",
  ["cancer radiotherapie"] = "Cancer Radiother.",
  ["cancer radiotherapie : journal de la societe francaise de radiotherapie oncologique"] = "Cancer Radiother",
  ["cancer radiothérapie : journal de la société française de radiothérapie oncologique"] = "Cancer Radiother",
  ["cancer reports"] = "Cancer Rep",
  ["cancer reports (hoboken, n.j.)"] = "Cancer Rep (Hoboken)",
  ["cancer reports and reviews"] = "Cancer Rep Rev",
  ["cancer research"] = "Cancer Res",
  ["cancer research and treatment"] = "Cancer Res Treat",
  ["cancer research communications"] = "Cancer Res. Commun.",
  ["cancer research frontiers"] = "Cancer Res Front",
  ["cancer research journal"] = "Cancer Res J",
  ["cancer research. supplement"] = "Can Res Suppl",
  ["cancer science"] = "Cancer Sci",
  ["cancer studies and molecular medicine : open journal"] = "Cancer Stud Mol Med",
  ["cancer studies and therapeutics"] = "Cancer Stud Ther",
  ["cancer surveys"] = "Cancer Surv",
  ["cancer therapy"] = "Cancer Ther",
  ["cancer therapy & oncology international journal"] = "Cancer Ther Oncol Int J",
  ["cancer translational medicine"] = "Cancer Transl Med",
  ["cancer treatment and research"] = "Cancer Treat Res",
  ["cancer treatment and research communications"] = "Cancer Treat Res Commun",
  ["cancer treatment communications"] = "Cancer Treat Commun",
  ["cancer treatment reports"] = "Cancer Treat Rep",
  ["cancer treatment reviews"] = "Cancer Treat Rev",
  ["cancers of the head & neck"] = "Cancers Head Neck",
  ["canine genetics and epidemiology"] = "Canine Genet Epidemiol",
  ["canine medicine and genetics"] = "Canine Med Genet",
  ["canine practice"] = "Canine Pract",
  ["cannabis (research society on marijuana)"] = "Cannabis",
  ["cannabis and cannabinoid research"] = "Cannabis Cannabinoid Res",
  ["cannt journal = journal acitn"] = "CANNT J",
  ["cao ye ke xue = pratacultural science = caoye kexue"] = "Cao Ye Ke Xue",
  ["cao ye xue bao = acta prataculturae sinica = caoye xuebao"] = "Cao Ye Xue Bao",
  ["cap today"] = "CAP Today",
  ["cap today / college of american pathologists"] = "CAP Today",
  ["cape breton post"] = "Cape Breton Post",
  ["capital nursing"] = "Cap Nurs",
  ["capital university law review"] = "Cap Univ Law Rev",
  ["capitation management report"] = "Capitation Manag Rep",
  ["capitation rates & data"] = "Capitation Rates Data",
  ["capitation rates and data"] = "Capitation Rates Data",
  ["caraka tani: journal of sustainable agriculture"] = "Caraka Tani: J. Sustainable Agric.",
  ["caravelle (toulouse, france)"] = "Caravelle",
  ["carbohydrate chemistry"] = "Carbohydr. Chem.",
  ["carbohydrate letters"] = "Carbohydr Lett",
  ["carbohydrate polymer technologies and applications"] = "Carbohydr. Polym. Technol. Appl.",
  ["carbohydrate polymers"] = "Carbohydr Polym",
  ["carbohydrate research"] = "Carbohydr Res",
  ["carbon and climate law review"] = "Carbon Clim. Law Rev.",
  ["carbon balance and management"] = "Carbon Balance Manag",
  ["carbon capture science & technology"] = "Carbon Capture Sci. Technol.",
  ["carbon energy"] = "Carbon Energy",
  ["carbon letters"] = "Carbon Lett.",
  ["carbon management"] = "Carbon Manage.",
  ["carbon materials: chemistry and physics"] = "Carbon Mater.: Chem. Phys.",
  ["carbon neutrality"] = "Carbon Neutrality",
  ["carbon neutralization"] = "Carbon Neutralization",
  ["carbon research"] = "Carbon Res.",
  ["carbon resources conversion"] = "Carbon Resour. Convers.",
  ["carbon trends"] = "Carbon Trends",
  ["carbonates and evaporites"] = "Carbonates Evaporites",
  ["carcinogenesis: a comprehensive survey"] = "Carcinog. Compr. Surv.",
  ["carcinogenesis; a comprehensive survey"] = "Carcinog Compr Surv",
  ["cardiac electrophysiology clinics"] = "Card Electrophysiol Clin",
  ["cardiac electrophysiology review"] = "Card Electrophysiol Rev",
  ["cardiac failure review"] = "Card Fail Rev",
  ["cardio-oncology (london, england)"] = "Cardiooncology",
  ["cardio-vascular nursing"] = "Cardiovasc Nurs",
  ["cardiologia (rome, italy)"] = "Cardiologia",
  ["cardiologia pratica"] = "Cardiol Prat",
  ["cardiologie tunisienne"] = "Cardiol Tunis",
  ["cardiology & vascular research (wilmington, del.)"] = "Cardiol Vasc Res (Wilmington)",
  ["cardiology and angiology"] = "Cardiol Angiol",
  ["cardiology and cardiovascular medicine"] = "Cardiol Cardiovasc Med",
  ["cardiology and therapy"] = "Cardiol Ther",
  ["cardiology clinics"] = "Cardiol Clin",
  ["cardiology in review"] = "Cardiol Rev",
  ["cardiology in the young"] = "Cardiol Young",
  ["cardiology journal"] = "Cardiol J",
  ["cardiology management"] = "Cardiol Manage",
  ["cardiology plus"] = "Cardiol Plus",
  ["cardiology research"] = "Cardiol Res",
  ["cardiology research and cardiovascular medicine"] = "Cardiol Res Cardiovasc Med",
  ["cardiology research and practice"] = "Cardiol Res Pract",
  ["cardiopulmonary physical therapy journal"] = "Cardiopulm Phys Ther J",
  ["cardiorenal medicine"] = "Cardiorenal Med",
  ["cardiovascular & hematological agents"] = "Cardiovasc. Hematol. Agents",
  ["cardiovascular & hematological agents in medicinal chemistry"] = "Cardiovasc Hematol Agents Med Chem",
  ["cardiovascular & hematological disorders drug targets"] = "Cardiovasc Hematol Disord Drug Targets",
  ["cardiovascular & hematological disorders: drug targets"] = "Cardiovasc. Hematol. Disord.: Drug Targets",
  ["cardiovascular and hematological agents in medicinal chemistry"] = "Cardiovasc. Hematol. Agents Med. Chem.",
  ["cardiovascular and hematological disorders drug targets"] = "Cardiovasc. Hematol. Disord. Drug Targets.",
  ["cardiovascular and interventional radiology"] = "Cardiovasc Intervent Radiol",
  ["cardiovascular and thoracic open"] = "Cardiovasc Thorac Open",
  ["cardiovascular clinics"] = "Cardiovasc Clin",
  ["cardiovascular diabetology"] = "Cardiovasc Diabetol",
  ["cardiovascular diagnosis and therapy"] = "Cardiovasc Diagn Ther",
  ["cardiovascular digital health journal"] = "Cardiovasc Digit Health J",
  ["cardiovascular disease and medicine"] = "Cardiovasc Dis Med",
  ["cardiovascular diseases"] = "Cardiovasc Dis",
  ["cardiovascular disorder and medicine"] = "Cardiovasc Disord Med",
  ["cardiovascular drug reviews"] = "Cardiovasc Drug Rev",
  ["cardiovascular drugs and therapy"] = "Cardiovasc Drugs Ther",
  ["cardiovascular drugs and therapy / sponsored by the international society of cardiovascular pharmacotherapy"] = "Cardiovasc Drugs Ther",
  ["cardiovascular endocrinology"] = "Cardiovasc Endocrinol",
  ["cardiovascular endocrinology & metabolism"] = "Cardiovasc Endocrinol Metab",
  ["cardiovascular engineering"] = "Cardiovasc. Eng.",
  ["cardiovascular engineering (dordrecht, netherlands)"] = "Cardiovasc Eng",
  ["cardiovascular engineering and technology"] = "Cardiovasc Eng Technol",
  ["cardiovascular innovations and applications"] = "Cardiovasc Innov Appl",
  ["cardiovascular intervention and therapeutics"] = "Cardiovasc Interv Ther",
  ["cardiovascular journal of africa"] = "Cardiovasc J Afr",
  ["cardiovascular journal of south africa"] = "Cardiovasc. J. S. Afr.",
  ["cardiovascular journal of south africa : official journal for southern africa cardiac society [and] south african society of cardiac practitioners"] = "Cardiovasc J S Afr",
  ["cardiovascular medicine"] = "Cardiovasc Med",
  ["cardiovascular nursing"] = "Cardiovasc. Nurs.",
  ["cardiovascular pathobiology"] = "Cardiovasc Pathobiol",
  ["cardiovascular pathology"] = "Cardiovasc. Pathol.",
  ["cardiovascular pathology : the official journal of the society for cardiovascular pathology"] = "Cardiovasc Pathol",
  ["cardiovascular pharmacology: open access"] = "Cardiovasc Pharm Open Access",
  ["cardiovascular psychiatry and neurology"] = "Cardiovasc Psychiatry Neurol",
  ["cardiovascular radiation medicine"] = "Cardiovasc Radiat Med",
  ["cardiovascular radiology"] = "Cardiovasc Radiol",
  ["cardiovascular regenerative medicine"] = "Cardiovasc Regen Med",
  ["cardiovascular research"] = "Cardiovasc Res",
  ["cardiovascular research center bulletin"] = "Cardiovasc Res Cent Bull",
  ["cardiovascular revascularization medicine"] = "Cardiovasc. Revasc. Med.",
  ["cardiovascular revascularization medicine : including molecular interventions"] = "Cardiovasc Revasc Med",
  ["cardiovascular reviews & reports"] = "Cardiovasc Rev Rep",
  ["cardiovascular surgery"] = "Cardiovasc. Surg.",
  ["cardiovascular surgery (london, england)"] = "Cardiovasc Surg",
  ["cardiovascular therapeutics"] = "Cardiovasc Ther",
  ["cardiovascular toxicology"] = "Cardiovasc Toxicol",
  ["cardiovascular ultrasound"] = "Cardiovasc Ultrasound",
  ["cardozo law review"] = "Cardozo Law Rev",
  ["care briefs on development issues"] = "CARE Briefs Develop Isssues",
  ["care giver"] = "Care Giver",
  ["care giver : journal of the college of chaplains"] = "Care Giver",
  ["care management journals"] = "Care Manag. J.",
  ["care management journals : journal of case management ; the journal of long term home health care"] = "Care Manag J",
  ["care weekly"] = "Care Wkly",
  ["career development and transition for exceptional individuals"] = "Career Dev Transit Except Individ",
  ["caribbean affairs (port of spain, trinidad and tobago)"] = "Caribb Aff",
  ["caribbean geography"] = "Caribb. Geogr.",
  ["caribbean health"] = "Caribb Health",
  ["caribbean herpetology"] = "Caribb Herpetol",
  ["caribbean journal of mathematical and computing sciences"] = "Caribbean J. Math. Comput. Sci.",
  ["caribbean journal of science"] = "Caribb J Sci",
  ["caribbean medical journal"] = "Caribb Med J",
  ["caribbean quarterly"] = "Caribb Q",
  ["caribbean review"] = "Caribb Rev",
  ["caribbean studies (rio piedras, san juan, p.r.)"] = "Caribb Stud",
  ["caribbean studies (río piedras, san juan, p.r.)"] = "Caribb Stud",
  ["caridad, ciencia y arte"] = "Caridad Cienc Arte",
  ["caries research"] = "Caries Res",
  ["caring : national association for home care magazine"] = "Caring",
  ["carinthia i : mittheilungen des geschichtsvereins fur karnten"] = "Carinthia I",
  ["carinthia i : mittheilungen des geschichtsvereins für kärnten"] = "Carinthia I",
  ["carinthia i. geschichtliche und volkskundliche beiträge zur heimatkunde kärntens"] = "CarinthiaI",
  ["carle selected papers"] = "Carle Sel Pap",
  ["carlsberg research communications"] = "Carlsberg Res Commun",
  ["carnegie quarterly"] = "Carnegie Q",
  ["carnegie-rochester conference series on public policy"] = "Carnegie-Rochester Conf. Ser. Public Pol.",
  ["carnet de l'économe"] = "Carnet Econ",
  ["carnets de l’enfance"] = "Carnets Enfance",
  ["carney hospital journal"] = "Carney Hosp J",
  ["carnuntum-jahrbuch. zeitschrift für archäologie und kulturgeschichte des donauraumes"] = "CarnuntumJb",
  ["carolina fire rescue ems journal"] = "Carol Fire Rescue EMS J",
  ["carolina lecture series"] = "Carolina Lecture Ser.",
  ["carolina planning"] = "Carol Plann",
  ["carolinea. beiheft"] = "Carolinea Beih (Karlsr)",
  ["carpathian journal of food science and technology"] = "Carpathian J. Food Sci. Technol.",
  ["carpathian journal of mathematics"] = "Carpathian J. Math.",
  ["carpathian mathematical publications. karpatski matematichni publikatsii"] = "Carpathian Math. Publ.",
  ["carpica. muzeul judeţean de istorie şi artă, bacău"] = "Carpica",
  ["carrefour africain"] = "Carrefour Afr",
  ["carta economica regional : cer"] = "Carta Econ Reg",
  ["carta económica regional : cer"] = "Carta Econ Reg",
  ["carta informativa (asociacion demografica salvadorena)"] = "Carta Inf",
  ["carta informativa (asociación demográfica salvadoreña)"] = "Carta Inf",
  ["cartographic journal"] = "Cartogr. J.",
  ["cartography and geographic information science"] = "Cartogr Geogr Inf Sci",
  ["carus mathematical monographs"] = "Carus Math. Monogr.",
  ["caryologia; giornale di citologia, citosistematica, e citogenetica"] = "Caryologia G Citol Citosistematica Citogenet",
  ["casa de la mujer"] = "Casa Mujer",
  ["case (philadelphia, pa.)"] = "CASE (Phila)",
  ["case and comment"] = "Case Comment",
  ["case orthopaedic journal"] = "Case Orthop J",
  ["case reports in anesthesiology"] = "Case Rep Anesthesiol",
  ["case reports in cardiology"] = "Case Rep Cardiol",
  ["case reports in clinical pathology"] = "Case Rep Clin Pathol",
  ["case reports in critical care"] = "Case Rep Crit Care",
  ["case reports in dentistry"] = "Case Rep Dent",
  ["case reports in dermatological medicine"] = "Case Rep Dermatol Med",
  ["case reports in dermatology"] = "Case Rep Dermatol",
  ["case reports in emergency medicine"] = "Case Rep Emerg Med",
  ["case reports in endocrinology"] = "Case Rep Endocrinol",
  ["case reports in gastroenterology"] = "Case Rep Gastroenterol",
  ["case reports in gastrointestinal medicine"] = "Case Rep Gastrointest Med",
  ["case reports in genetics"] = "Case Rep Genet",
  ["case reports in hematology"] = "Case Rep Hematol",
  ["case reports in hepatology"] = "Case Reports Hepatol",
  ["case reports in immunology"] = "Case Reports Immunol",
  ["case reports in infectious diseases"] = "Case Rep Infect Dis",
  ["case reports in internal medicine"] = "Case Rep Intern Med",
  ["case reports in medicine"] = "Case Rep Med",
  ["case reports in nephrology"] = "Case Rep Nephrol",
  ["case reports in nephrology and dialysis"] = "Case Rep Nephrol Dial",
  ["case reports in nephrology and urology"] = "Case Rep Nephrol Urol",
  ["case reports in neurological medicine"] = "Case Rep Neurol Med",
  ["case reports in neurology"] = "Case Rep Neurol",
  ["case reports in obstetrics and gynecology"] = "Case Rep Obstet Gynecol",
  ["case reports in oncological medicine"] = "Case Rep Oncol Med",
  ["case reports in oncology"] = "Case Rep Oncol",
  ["case reports in ophthalmological medicine"] = "Case Rep Ophthalmol Med",
  ["case reports in ophthalmology"] = "Case Rep Ophthalmol",
  ["case reports in orthopedics"] = "Case Rep Orthop",
  ["case reports in otolaryngology"] = "Case Rep Otolaryngol",
  ["case reports in pancreatic cancer"] = "Case Rep Pancreat Cancer",
  ["case reports in pathology"] = "Case Rep Pathol",
  ["case reports in pediatrics"] = "Case Rep Pediatr",
  ["case reports in plastic surgery & hand surgery"] = "Case Reports Plast Surg Hand Surg",
  ["case reports in psychiatry"] = "Case Rep Psychiatry",
  ["case reports in pulmonology"] = "Case Rep Pulmonol",
  ["case reports in radiology"] = "Case Rep Radiol",
  ["case reports in rheumatology"] = "Case Rep Rheumatol",
  ["case reports in surgery"] = "Case Rep Surg",
  ["case reports in transplantation"] = "Case Rep Transplant",
  ["case reports in urology"] = "Case Rep Urol",
  ["case reports in vascular medicine"] = "Case Rep Vasc Med",
  ["case reports in veterinary medicine"] = "Case Rep Vet Med",
  ["case reports in women's health"] = "Case Rep Womens Health",
  ["case reports. children's memorial hospital (chicago, ill.)"] = "Case Rep Child Meml Hosp Chic",
  ["case studies in business, industry and government statistics : csbigs"] = "Case Studies Bus Ind Gov Stat",
  ["case studies in chemical and environmental engineering"] = "Case Stud. Chem. Environ. Eng.",
  ["case studies in construction materials"] = "Case Stud. Constr. Mater.",
  ["case studies in engineering failure analysis"] = "Case Stud. Eng. Fail. Anal.",
  ["case studies in fire safety"] = "Case Stud. Fire Saf.",
  ["case studies in health administration"] = "Case Stud Health Adm",
  ["case studies in mechanical systems and signal processing"] = "Case Stud. Mech. Syst. Signal Process.",
  ["case studies in nondestructive testing and evaluation"] = "Case Stud. Nondestr.Test. Eval.",
  ["case studies in structural engineering"] = "Case Stud. Struct. Eng.",
  ["case studies in surgery"] = "Case Stud Surg",
  ["case studies in the environment"] = "Case Stud. Environ.",
  ["case studies in thermal engineering"] = "Case Stud. Therm. Eng.",
  ["case studies in transport policy"] = "Case Stud. Transp. Policy",
  ["case studies on transport policy"] = "Case Stud Transp Policy",
  ["case western reserve journal of international law"] = "Case West Reserve J Int Law",
  ["case western reserve law review"] = "Case West Reserve Law Rev",
  ["cases in public health communication and marketing"] = "Cases Public Health Commun Mark",
  ["cases journal"] = "Cases J",
  ["casopis ceskeho lekarnictva"] = "Cas Cesk Lek",
  ["casopis ceskoslovenskych veterinaru"] = "Cas Cesk Vet",
  ["casopis c̆eského lékárnictva"] = "Cas Cesk Lek",
  ["casopis c̆eskoslovenských veterinár̆ů"] = "Cas Cesk Vet",
  ["casopis lekaru ceskych"] = "Cas Lek Cesk",
  ["casopis matice moravske (1968)"] = "Cas Matice Moravske",
  ["casopis narodniho musea"] = "Cas Nar Muz",
  ["casopis za zgodovino in narodopisje. review for history and ethnography"] = "Cas Zgodovino Narodop",
  ["casopís lékar̆ů c̆eských"] = "Cas Lek Cesk",
  ["caspian journal of internal medicine"] = "Caspian J Intern Med",
  ["caspian journal of mathematical sciences"] = "Casp. J. Math. Sci.",
  ["caspian journal of neurological sciences"] = "Casp J Neurol Sci",
  ["cataloging & classification quarterly"] = "Cat Classif Q",
  ["catalogue of oriental coins in the british museum i–x"] = "BMCOR",
  ["catalogue of the greek coins in the british museums"] = "BMCGreekCoins",
  ["catalysis communications"] = "Catal. Commun.",
  ["catalysis for sustainable energy"] = "Catal. Sustainable Energy",
  ["catalysis in green chemistry and engineering"] = "Catal. Green Chem. Eng.",
  ["catalysis in industry"] = "Catal. Ind.",
  ["catalysis letters"] = "Catal. Lett.",
  ["catalysis reviews - science and engineering"] = "Catal. Rev. Sci. Eng.",
  ["catalysis reviews, science and engineering"] = "Catal Rev Sci Eng",
  ["catalysis reviews: science and engineering"] = "Catal. Rev.: Sci. Eng.",
  ["catalysis science & technology"] = "Catal. Sci. Technol.",
  ["catalysis surveys from asia"] = "Catal. Surv. Asia",
  ["catalysis surveys from japan"] = "Catal. Surv. Jpn.",
  ["catalysis today"] = "Catal. Today",
  ["catalysis, structure & reactivity"] = "Catal. Struct. React.",
  ["catalysts (basel, switzerland)"] = "Catalysts",
  ["categories and general algebraic structures with applications"] = "Categ. Gen. Algebr. Struct. Appl.",
  ["catena supplement"] = "Catena Suppl",
  ["catering & health"] = "Cater Health",
  ["catering and health"] = "Cater. Health",
  ["catheterization and cardiovascular diagnosis"] = "Cathet Cardiovasc Diagn",
  ["catheterization and cardiovascular interventions"] = "Catheter. Cardiovasc. Interv.",
  ["catheterization and cardiovascular interventions : official journal of the society for cardiac angiography & interventions"] = "Catheter Cardiovasc Interv",
  ["catholic historical review"] = "CHR",
  ["catholic hospital"] = "Cathol Hosp",
  ["catholic nurse"] = "Cathol. Nurse",
  ["catholic standard (washington, d.c.)"] = "Cathol Stand",
  ["catholic university law review"] = "Cathol Univers Law Rev",
  ["catholic update"] = "Cathol Update",
  ["catholic woman"] = "Cathol Woman",
  ["catholic world"] = "Cathol World",
  ["cato journal"] = "Cato J.",
  ["catrina : the international journal of environmental sciences"] = "Catrina Int J Environ Sci",
  ["cattle practice : journal of the british cattle veterinary association"] = "Cattle Pract",
  ["cave research"] = "Cave Res",
  ["cátedra y clínica"] = "Catedra Clin",
  ["cbe life sciences education"] = "CBE Life Sci Educ",
  ["cbe views : official publication of the council of biology editors"] = "CBE Views",
  ["cbms issues in mathematics education"] = "CBMS Issues Math. Ed.",
  ["cbms regional conference series in mathematics"] = "CBMS Reg. Conf. Ser. Math.",
  ["cbms-nsf regional conference series in applied mathematics"] = "CBMS-NSF Regional Conf. Ser. in Appl. Math.",
  ["cc ai"] = "CC AI",
  ["ccar journal (new york, n.y. : 1991)"] = "CCAR J",
  ["ccf transactions on high performance computing"] = "CCF Trans. High Perform. Comput.",
  ["ccf transactions on networking"] = "CCF Trans. Networking",
  ["ccf transactions on pervasive computing and interaction"] = "CCF Trans. Pervasive Comput. Interact.",
  ["ccica annual : publication of the catholic commission on intellectual and cultural affairs"] = "CCICA Annual",
  ["ccl family foundations"] = "CCL Family Found",
  ["ccq. critical care quarterly"] = "CCQ",
  ["ccq: critical care quarterly"] = "CCQ.",
  ["ccs chemistry"] = "CCS Chem.",
  ["cda journal"] = "CDA J",
  ["cda journal california dental association"] = "CDA J",
  ["cdc aids weekly"] = "CDC AIDS Wkly",
  ["cdc bulletin. communicable disease center (u.s.)"] = "CDC Bull",
  ["cdc nchstp daily news update"] = "CDC NCHSTP Dly News Update",
  ["cde working paper. university of wisconsin--madison. center for demography and ecology"] = "CDE Work Pap",
  ["cdr (london, england : review)"] = "CDR (Lond Engl Rev)",
  ["cdr (london, england : weekly)"] = "CDR (Lond Engl Wkly)",
  ["cdr (london, review)"] = "CDR (Lond. Engl. Rev.)",
  ["cdr (london, weekly)"] = "CDR (Lond. Engl. Wkly.)",
  ["cdr weekly : communicable disease report"] = "CDR Wkly (Online)",
  ["cds review"] = "CDS Rev",
  ["cdt digest"] = "CDT Dig",
  ["ce focus"] = "CE Focus",
  ["cea'13 : proceedings of the 5th international workshop on multimedia for cooking & eating activities : october 21, 2013, barcelona, spain. workshop on multimedia for cooking and eating activities (5th : 2013 : barcelona, spain)"] = "CEA13 (2013)",
  ["ceará médico : orgão do centro médico cearense"] = "Ceara Med",
  ["ceas aeronautical journal"] = "CEAS Aeronaut J",
  ["ceas space journal"] = "CEAS Space J.",
  ["cedac. bulletin. centre d’études et de documentation archéologique de la conservation de cartage"] = "CEDAC",
  ["cedpa network (online)"] = "CEDPA Netw",
  ["cedpa world wide. centre for development and population activities"] = "CEDPA World Wide",
  ["celebrity studies"] = "Celebr Stud",
  ["celestial mechanics & dynamical astronomy"] = "Celestial Mech. Dynam. Astronom.",
  ["celestial mechanics & dynamical astronomy. an international journal of space dynamics"] = "Celestial Mech. Dynam. Astronom.",
  ["celestial mechanics and dynamic astronomy"] = "Celestial Mech. Dyn. Astron.",
  ["celestial mechanics and dynamical astronomy"] = "Celest Mech Dyn Astron",
  ["cell & bioscience"] = "Cell Biosci",
  ["cell & chromosome"] = "Cell Chromosome",
  ["cell & developmental biology"] = "Cell Dev Biol",
  ["cell & gene therapy insights"] = "Cell Gene Ther Insights",
  ["cell (cambridge, ma, united states)"] = "Cell (Cambridge, MA, U. S.)",
  ["cell adhesion & migration"] = "Cell Adh Migr",
  ["cell adhesion and communication"] = "Cell Adhes Commun",
  ["cell and molecular biology reports"] = "Cell Mol. Biol. Rep.",
  ["cell and muscle motility"] = "Cell Muscle Motil",
  ["cell and tissue banking"] = "Cell Tissue Bank",
  ["cell and tissue biology"] = "Cell tissue biol",
  ["cell and tissue kinetics"] = "Cell Tissue Kinet",
  ["cell and tissue research"] = "Cell Tissue Res",
  ["cell biochemistry and biophysics"] = "Cell Biochem Biophys",
  ["cell biochemistry and function"] = "Cell Biochem Funct",
  ["cell biology : research & therapy"] = "Cell Biol (Henderson, NV)",
  ["cell biology and toxicology"] = "Cell Biol Toxicol",
  ["cell biology education"] = "Cell Biol Educ",
  ["cell biology international"] = "Cell Biol Int",
  ["cell biology international reports"] = "Cell Biol Int Rep",
  ["cell biology reviews"] = "Cell Biol. Rev.",
  ["cell biology reviews : cbr"] = "Cell Biol Rev",
  ["cell biophysics"] = "Cell Biophys",
  ["cell calcium"] = "Cell Calcium",
  ["cell chemical biology"] = "Cell Chem Biol",
  ["cell communication & adhesion"] = "Cell Commun Adhes",
  ["cell communication and adhesion"] = "Cell Commun. Adhes.",
  ["cell communication and signaling"] = "Cell Commun. Signaling",
  ["cell communication and signaling : ccs"] = "Cell Commun Signal",
  ["cell communication insights"] = "Cell Commun Insights",
  ["cell cycle"] = "Cell Cycle",
  ["cell cycle (georgetown, tex.)"] = "Cell Cycle",
  ["cell death & disease"] = "Cell Death Dis",
  ["cell death and differentiation"] = "Cell Death Differ",
  ["cell death and disease"] = "Cell Death Dis.",
  ["cell death discovery"] = "Cell Death Discov",
  ["cell differentiation"] = "Cell Differ",
  ["cell differentiation and development"] = "Cell Differ. Dev.",
  ["cell differentiation and development : the official journal of the international society of developmental biologists"] = "Cell Differ Dev",
  ["cell discovery"] = "Cell Discov",
  ["cell division"] = "Cell Div",
  ["cell genomics"] = "Cell Genomics",
  ["cell growth & differentiation : the molecular biology journal of the american association for cancer research"] = "Cell Growth Differ",
  ["cell growth and differentiation"] = "Cell Growth Differ.",
  ["cell health and cytoskeleton"] = "Cell Health Cytoskelet",
  ["cell host & microbe"] = "Cell Host Microbe",
  ["cell journal"] = "Cell J",
  ["cell medicine"] = "Cell Med",
  ["cell metabolism"] = "Cell Metab",
  ["cell motility"] = "Cell Motil",
  ["cell motility and the cytoskeleton"] = "Cell Motil Cytoskeleton",
  ["cell preservation and technology"] = "Cell Preserv. Technol.",
  ["cell preservation technology"] = "Cell Preserv Technol",
  ["cell proliferation"] = "Cell Prolif",
  ["cell regeneration"] = "Cell Regener.",
  ["cell regeneration (london, england)"] = "Cell Regen",
  ["cell regulation"] = "Cell Regul",
  ["cell reports"] = "Cell Rep",
  ["cell reports medicine"] = "Cell Rep. Med.",
  ["cell reports methods"] = "Cell Rep. Methods",
  ["cell reports physical science"] = "Cell Rep. Phys. Sci.",
  ["cell reports. medicine"] = "Cell Rep Med",
  ["cell reports. physical science"] = "Cell Rep Phys Sci",
  ["cell research"] = "Cell Res",
  ["cell stem cell"] = "Cell Stem Cell",
  ["cell stress"] = "Cell Stress",
  ["cell stress & chaperones"] = "Cell Stress Chaperones",
  ["cell stress and chaperones"] = "Cell Stress Chaperones",
  ["cell structure and function"] = "Cell Struct Funct",
  ["cell surface"] = "Cell Surf.",
  ["cell surface (amsterdam, netherlands)"] = "Cell Surf",
  ["cell systems"] = "Cell Syst",
  ["cell transplantation"] = "Cell Transplant",
  ["cell vision"] = "Cell Vis.",
  ["cell vision : the journal of analytical morphology"] = "Cell Vis",
  ["cell, stem cells and regenerative medicine"] = "Cell Stem Cells Regen Med",
  ["cellr4-- repair, replacement, regeneration, & reprogramming"] = "CellR4 Repair Replace Regen Reprogram",
  ["cells and development"] = "Cells Dev.",
  ["cells, tissues, organs"] = "Cells Tissues Organs",
  ["cellular & molecular biology letters"] = "Cell Mol Biol Lett",
  ["cellular & molecular biology research"] = "Cell Mol Biol Res",
  ["cellular & molecular immunology"] = "Cell Mol Immunol",
  ["cellular & molecular medicine: open access"] = "Cell Mol Med Open Access",
  ["cellular and molecular bioengineering"] = "Cell Mol Bioeng",
  ["cellular and molecular biology"] = "Cell Mol Biol",
  ["cellular and molecular biology (noisy-le-grand, france)"] = "Cell Mol Biol (Noisy-le-grand)",
  ["cellular and molecular biology (omics)"] = "Cell Mol Biol (OMICS)",
  ["cellular and molecular biology (paris, france, online)"] = "Cell. Mol. Biol. (Paris, Fr., Online)",
  ["cellular and molecular biology (paris, france, print)"] = "Cell. Mol. Biol. (Paris, Fr., Print)",
  ["cellular and molecular biology (sarreguemines, france, online)"] = "Cell. Mol. Biol. (Sarreguemines, Fr., Online)",
  ["cellular and molecular biology (sarreguemines, france, print)"] = "Cell. Mol. Biol. (Sarreguemines, Fr., Print)",
  ["cellular and molecular biology letters"] = "Cell. Mol. Biol. Lett.",
  ["cellular and molecular biology research"] = "Cell. Mol. Biol. Res.",
  ["cellular and molecular biology, including cyto-enzymology"] = "Cell Mol Biol Incl Cyto Enzymol",
  ["cellular and molecular gastroenterology and hepatology"] = "Cell Mol Gastroenterol Hepatol",
  ["cellular and molecular immunology"] = "Cell. Mol. Immunol.",
  ["cellular and molecular life sciences"] = "Cell. Mol. Life Sci.",
  ["cellular and molecular life sciences : cmls"] = "Cell Mol Life Sci",
  ["cellular and molecular neurobiology"] = "Cell Mol Neurobiol",
  ["cellular immunology"] = "Cell Immunol",
  ["cellular logistics"] = "Cell Logist",
  ["cellular microbiology"] = "Cell Microbiol",
  ["cellular oncology"] = "Cell. Oncol.",
  ["cellular oncology (dordrecht)"] = "Cell Oncol (Dordr)",
  ["cellular oncology : the official journal of the international society for cellular oncology"] = "Cell Oncol",
  ["cellular pharmacology"] = "Cell Pharmacol",
  ["cellular physiology and biochemistry"] = "Cell. Physiol. Biochem.",
  ["cellular physiology and biochemistry : international journal of experimental cellular physiology, biochemistry, and pharmacology"] = "Cell Physiol Biochem",
  ["cellular polymers"] = "Cell. Polym.",
  ["cellular reprogramming"] = "Cell Reprogram",
  ["cellular signalling"] = "Cell Signal",
  ["cellular therapy and transplantation"] = "Cell Ther Transplant",
  ["cellulose (london, england)"] = "Cellulose (Lond)",
  ["cellulose chemistry and technology"] = "Cellul. Chem. Technol.",
  ["celticum. supplément à ogam"] = "Celticum",
  ["cement & concrete composites"] = "Cem Concr Compos",
  ["cement and concrete composites"] = "Cem. Concr. Compos.",
  ["cement and concrete research"] = "Cem. Concr. Res.",
  ["cemla boletin mensual"] = "CEMLA Bol. Mens.",
  ["cempi series"] = "CEMPI Ser.",
  ["cen case reports"] = "CEN Case Rep",
  ["centaur (london, england)"] = "Centaur Lond Engl",
  ["centaurus. an international journal of the history of science and its cultural aspects"] = "Centaurus",
  ["centaurus. journal of the european society for the history of science"] = "Centaurus",
  ["centaurus; international magazine of the history of science and medicine"] = "Centaurus",
  ["center view : a publication of the center for applied and professional ethics, university of tennessee, knoxville"] = "Cent View",
  ["centerviews (honolulu, hawaii)"] = "Centerviews",
  ["central african journal of medicine"] = "Cent. Afr. J. Med.",
  ["central asia monitor"] = "Cent Asia Monit",
  ["central asian journal of global health"] = "Cent Asian J Glob Health",
  ["central asian survey"] = "Centr Asian Surv",
  ["central european forests"] = "Cent. Eur. For.",
  ["central european geology"] = "Cent. Eur. Geol.",
  ["central european history"] = "Cent Eur Hist",
  ["central european journal for operations research and economics"] = "Central Europ. J. Operations Res. Econ.",
  ["central european journal of biology"] = "Cent Eur J Biol",
  ["central european journal of chemistry"] = "Cent. Eur. J. Chem.",
  ["central european journal of computer science"] = "Cent. Eur. J. Comput. Sci.",
  ["central european journal of energetic materials"] = "Cent. Eur. J. Energetic Mater.",
  ["central european journal of engineering"] = "Cent. Eur. J. Eng.",
  ["central european journal of geosciences"] = "Cent. Eur. J. Geosci.",
  ["central european journal of mathematics"] = "Cent. Eur. J. Math.",
  ["central european journal of medicine"] = "Cent Eur J Med",
  ["central european journal of operations research"] = "Cent Eur J Oper Res",
  ["central european journal of operations research (cejor)"] = "CEJOR Cent. Eur. J. Oper. Res.",
  ["central european journal of physics"] = "Cent. Eur. J. Phys.",
  ["central european journal of public health"] = "Cent Eur J Public Health",
  ["central european journal of sport sciences and medicine"] = "Cent. Eur. J. Sport Sci. Med.",
  ["central european journal of urology"] = "Cent European J Urol",
  ["central european neurosurgery"] = "Cent Eur Neurosurg",
  ["central issues in anthropology : a journal of the central states anthropological society"] = "Cent Issues Anthropol",
  ["central nervous system agents in medicinal chemistry"] = "Cent Nerv Syst Agents Med Chem",
  ["central nervous system trauma"] = "Cent. Nerv. Syst. Trauma",
  ["central nervous system trauma : journal of the american paralysis association"] = "Cent Nerv Syst Trauma",
  ["central states speech journal"] = "Cent States Speech J",
  ["central-european journal of immunology"] = "Cent Eur J Immunol",
  ["centralblatt für das gesamte forstwesen"] = "Cent.bl. gesamte Forstwes.",
  ["centre calling"] = "Cent Call",
  ["centre de physique les houches"] = "Cent. Phys. Les Houches",
  ["centre de recherches mathematiques proceedings"] = "Centre Rech. Math. Proc.",
  ["centre d’études et de documentation archéologique de la conservation de carthage. bulletin"] = "CEDAC",
  ["centre for pure and applied differential geometry (padge)"] = "Centre Pure Appl. Differential Geom. (PADGE)",
  ["centre medical"] = "Cent Med",
  ["centre médical"] = "Cent Med",
  ["centro de estudios de recursos odontologicos para el nino"] = "Cent. Estud. Recur. Odontol. Nino",
  ["centro de estudios de recursos odontologicos para el nino : ceron"] = "Cent Estud Recur Odontol Nino",
  ["centro de estudios de recursos odontológicos para el niño : ceron"] = "Cent Estud Recur Odontol Nino",
  ["centro di ricerca matematica ennio de giorgi (crm) series"] = "CRM Series",
  ["centro internazionale matematico estivo (c.i.m.e.) summer schools"] = "C.I.M.E. Summer Sch.",
  ["centropa : a journal of central european architecture and related arts"] = "Centropa (N Y N Y)",
  ["centrum voor wiskunde en informatica"] = "CWI Quarterly",
  ["cepal review"] = "CEPAL Rev",
  ["cephalalgia : an international journal of headache"] = "Cephalalgia",
  ["cephalalgia reports"] = "Cephalalgia Rep",
  ["ceramic engineering and science proceedings"] = "Ceram. Eng. Sci. Proc.",
  ["ceramic transactions"] = "Ceram. Trans.",
  ["ceramics - silikaty"] = "Ceram. Silik.",
  ["ceramics international"] = "Ceram. Int.",
  ["ceramurgia international"] = "Ceramurgia Int.",
  ["cercetări numismatice. muzeul naţional de istorie"] = "CercNum",
  ["cereal chemistry"] = "Cereal Chem.",
  ["cereal research communications"] = "Cereal Res. Commun.",
  ["cerebellum & ataxias"] = "Cerebellum Ataxias",
  ["cerebellum (london, england)"] = "Cerebellum",
  ["cerebral circulation - cognition and behavior"] = "Cereb Circ Cogn Behav",
  ["cerebral cortex"] = "Cereb. Cortex",
  ["cerebral cortex (new york, n.y. : 1991)"] = "Cereb Cortex",
  ["cerebral cortex communications"] = "Cereb Cortex Commun",
  ["cerebral palsy bulletin"] = "Cereb Palsy Bull",
  ["cerebral palsy journal"] = "Cereb. Palsy J.",
  ["cerebral palsy review"] = "Cereb Palsy Rev",
  ["cerebrospinal fluid research"] = "Cerebrospinal Fluid Res",
  ["cerebrovascular and brain metabolism reviews"] = "Cerebrovasc Brain Metab Rev",
  ["cerebrovascular diseases"] = "Cerebrovasc. Dis.",
  ["cerebrovascular diseases (basel, switzerland)"] = "Cerebrovasc Dis",
  ["cerebrovascular diseases extra"] = "Cerebrovasc Dis Extra",
  ["cerebrum : the dana forum on brain science"] = "Cerebrum",
  ["cern ideasquare journal of experimental innovation"] = "CERN Ideasq J Exp Innov",
  ["cerrahpasa tip fakultesi dergisi"] = "Cerrahpasa Tip Fak Derg",
  ["cerrahpaşa tip fakültesi dergisi"] = "Cerrahpasa Tip Fak Derg",
  ["certified dental technician"] = "Certif Dent Tech",
  ["certified milk"] = "Certif Milk",
  ["ceræ : an australasian journal of medieval and early modern studies"] = "Cerae",
  ["ces odontologia"] = "CES Odontol.",
  ["ces odontologia / instituto de ciencias de la salud"] = "CES Odontol",
  ["ces odontología"] = "CES Odontol",
  ["ces transactions on electrical machines and systems"] = "CES Trans. Electr. Mach. Syst.",
  ["cesifo economic studies"] = "CESifo Econ Stud",
  ["ceska a slovenska farmacie"] = "Ceska Slov. Farm.",
  ["ceska a slovenska farmacie : casopis ceske farmaceuticke spolecnosti a slovenske farmaceuticke spolecnosti"] = "Ceska Slov Farm",
  ["ceska a slovenska oftalmologie"] = "Cesk. Slov. Oftalmol.",
  ["ceska a slovenska oftalmologie : casopis ceske oftalmologicke spolecnosti a slovenske oftalmologicke spolecnosti"] = "Cesk Slov Oftalmol",
  ["ceska a slovenska psychiatrie"] = "Ceska Slov. Psychiatr.",
  ["ceska a slovenska psychiatrie / ceska lekarska spolecnost j.e. purkyne"] = "Ceska Slov Psychiatr",
  ["ceska gynekologie"] = "Ceska Gynekol.",
  ["ceska gynekologie / ceska lekarska spolecnost j. ev. purkyne"] = "Ceska Gynekol",
  ["ceská a slovenská farmacie : casopis ceské farmaceutické spolecnosti a slovenské farmaceutické spolecnosti"] = "Ceska Slov Farm",
  ["ceská a slovenská oftalmologie : casopis ceské oftalmologické spolecnosti a slovenské oftalmologické spolecnosti"] = "Cesk Slov Oftalmol",
  ["ceská a slovenská psychiatrie"] = "Ceska Slov Psychiatr",
  ["ceská gynekologie"] = "Ceska Gynekol",
  ["ceskoslovenska biologie"] = "Ceskoslov Biol",
  ["ceskoslovenska dermatologie"] = "Cesk. Dermatol.",
  ["ceskoslovenska epidemiologie, mikrobiologie, imunologie"] = "Cesk. Epidemiol. Mikrobiol. Imunol.",
  ["ceskoslovenska farmacie"] = "Cesk. Farm.",
  ["ceskoslovenska fysiologie"] = "Cesk. Fysiol.",
  ["ceskoslovenska fysiologie / ustredni ustav biologicky"] = "Cesk Fysiol",
  ["ceskoslovenska gastroenterologie a vyziva"] = "Cesk. Gastroenterol. Vyz.",
  ["ceskoslovenska gynekologie"] = "Cesk. Gynekol.",
  ["ceskoslovenska hygiena"] = "Cesk. Hyg.",
  ["ceskoslovenska morfologie"] = "Cesk Morfol",
  ["ceskoslovenska nemocnice"] = "Ceskoslov Nemocnice",
  ["ceskoslovenska neurologie"] = "Cesk. Neurol.",
  ["ceskoslovenska neurologie a neurochirurgie"] = "Cesk. Neurol. Neurochir.",
  ["ceskoslovenska oftalmologie"] = "Cesk. Oftalmol.",
  ["ceskoslovenska onkologia"] = "Cesk Onkol",
  ["ceskoslovenska otolaryngologie"] = "Cesk. Otolaryngol.",
  ["ceskoslovenska patologie"] = "Cesk Patol",
  ["ceskoslovenska pediatrie"] = "Cesk. Pediatr.",
  ["ceskoslovenska psychiatrie"] = "Cesk. Psychiatr.",
  ["ceskoslovenska psychologie"] = "Cesk Psychol",
  ["ceskoslovenska radiologie"] = "Cesk. Radiol.",
  ["ceskoslovenska rentgenologie"] = "Cesk Rentgenol",
  ["ceskoslovenska stomatologie"] = "Cesk. Stomatol.",
  ["ceskoslovenská biologie"] = "Ceskoslov Biol",
  ["ceskoslovenská dermatologie"] = "Cesk Dermatol",
  ["ceskoslovenská epidemiologie, mikrobiologie, imunologie"] = "Cesk Epidemiol Mikrobiol Imunol",
  ["ceskoslovenská farmacie"] = "Cesk Farm",
  ["ceskoslovenská fysiologie"] = "Cesk Fysiol",
  ["ceskoslovenská gastroenterologie a výz̆iva"] = "Cesk Gastroenterol Vyz",
  ["ceskoslovenská gynekologie"] = "Cesk Gynekol",
  ["ceskoslovenská hygiena"] = "Cesk Hyg",
  ["ceskoslovenská morfologie"] = "Cesk Morfol",
  ["ceskoslovenská nemocnice"] = "Ceskoslov Nemocnice",
  ["ceskoslovenská neurologie"] = "Cesk Neurol",
  ["ceskoslovenská neurologie a neurochirurgie"] = "Cesk Neurol Neurochir",
  ["ceskoslovenská oftalmologie"] = "Cesk Oftalmol",
  ["ceskoslovenská onkológia"] = "Cesk Onkol",
  ["ceskoslovenská otolaryngologie"] = "Cesk Otolaryngol",
  ["ceskoslovenská pediatrie"] = "Cesk Pediatr",
  ["ceskoslovenská psychiatrie"] = "Cesk Psychiatr",
  ["ceskoslovenská psychologie"] = "Cesk Psychol",
  ["ceskoslovenská radiologie"] = "Cesk Radiol",
  ["ceskoslovenská rentgenologie"] = "Cesk Rentgenol",
  ["ceskoslovenská stomatologie"] = "Cesk Stomatol",
  ["ceskoslovenske zdravotnictvi"] = "Cesk. Zdrav.",
  ["ceskoslovenské zdravotnictví"] = "Cesk Zdrav",
  ["ceskoslovensky casopis historicky / ceskoslovenska akademie ved"] = "Cesk Cas Hist",
  ["ceskoslovenský dentista"] = "Ceskoslov Dent",
  ["cesra saule"] = "Cesra Saule",
  ["cesra säule"] = "Cesra Saule",
  ["ceur workshop proceedings"] = "CEUR Workshop Proc",
  ["cex [reports]; civil effects exercise. u.s. atomic energy commission"] = "CEX Rep Civ Eff Exerc",
  ["cex reports, civil effects exercise"] = "CEX Rep. Civ. Eff. Exerc.",
  ["ceylon dental journal"] = "Ceylon Dent J",
  ["ceylon forester"] = "Ceylon For.",
  ["ceylon journal of science. biological sciences"] = "Ceylon J Sci Biol Sci",
  ["ceylon medical journal"] = "Ceylon Med. J.",
  ["cfbs newsletter = bulletin de nouvelles de la fc-sb"] = "CFBS Newsl",
  ["cfw plexus"] = "CFW Plex",
  ["ch'ikwa kijae hakhoe chi. the journal of the korea research society for dental materials"] = "Chilwa Kijae Hakhoe Chi",
  ["ch'ŏngsonyŏnhak yŏngu"] = "Chongsonyonhak Yongu",
  ["ch'ŏnmun hakhoe chi. the journal of the korean astronomical society. han'guk ch'ŏnmun hakhoe"] = "J Korean Astron Soc",
  ["ch. w. clairmont, classical attic tombstones (kilchberg 1993–1995)"] = "CAT",
  ["cha insight"] = "CHA Insight",
  ["cha ye ke xue = journal of tea science"] = "Cha Ye Ke Xue",
  ["chac review"] = "CHAC Rev.",
  ["chaeoe hanin yŏn'gu = studies of koreans abroad"] = "Chaeoe Hanin Yon Gu",
  ["chain drug review"] = "Chain Drug Rev",
  ["challenge (atlanta, ga.)"] = "Challenge (Atlanta Ga)",
  ["challenge; the magazine of economic affairs"] = "Challenge (White Plains)",
  ["challenges and advances in computational chemistry and physics"] = "Chall. Adv. Comput. Chem. Phys.",
  ["challenges in sustainability"] = "Challenges Sustainability",
  ["chamchuri journal of mathematics"] = "Chamchuri J. Math.",
  ["chance (new york, n.y.)"] = "Chance (N Y)",
  ["chang gung medical journal"] = "Chang Gung Med J",
  ["chang jiang da xue xue bao. zi ran ke xue ban = journal of yangtze university. natural science edition"] = "Chang Jiang Da Xue Xue Bao Zi Ran Ke Xue Ban",
  ["chang keng i hsueh chang gung medical journal"] = "Chang Keng I Hsueh",
  ["change and adaptation in socio-ecological systems"] = "Change Adapt. Socio-Ecol. Syst.",
  ["changes (hove, england)"] = "Changes",
  ["changgeng yi xue za zhi"] = "Changgeng Yi Xue Za Zhi",
  ["changgeng yi xue za zhi (chang gung medical journal)"] = "Changgeng Yi Xue Za Zhi",
  ["changing men (madison, wis.)"] = "Chang Men",
  ["changing times"] = "Changing Times",
  ["changsha communications university"] = "J. Changsha Comm. Univ.",
  ["changsha university of electric power"] = "J. Changsha Univ. Electr. Power Nat. Sci. Ed.",
  ["changshu gao-zhuan xuebao = journal of changshu college"] = "Chang Gaozhuan Xuebao",
  ["channels (austin, tex.)"] = "Channels (Austin)",
  ["chaos (woodbury, n.y.)"] = "Chaos",
  ["chaos and complexity letters"] = "Chaos Complex Lett",
  ["chaos, solitons & fractals"] = "Chaos Solitons Fractals",
  ["chaos, solitons & fractals:x"] = "Chaos, Solitons  Fractals:X",
  ["chaos, solitons, and fractals"] = "Chaos Solitons Fractals",
  ["chaos. an interdisciplinary journal of nonlinear science"] = "Chaos",
  ["chapman & hall probability series"] = "Chapman & Hall Probab. Ser.",
  ["chapman & hall/crc big data series"] = "Chapman & Hall/CRC Big Data Ser.",
  ["chapman & hall/crc biostatistics series"] = "Chapman & Hall/CRC Biostat. Ser.",
  ["chapman & hall/crc computational biology series"] = "Chapman & Hall/CRC Comput. Biol. Ser.",
  ["chapman & hall/crc computational science series"] = "Chapman & Hall/CRC Comput. Sci. Ser.",
  ["chapman & hall/crc computer and information science series"] = "Chapman & Hall/CRC Comput. Inf. Sci. Ser.",
  ["chapman & hall/crc computer science and data analysis series"] = "Chapman & Hall/CRC Comput. Sci. Data Anal. Ser.",
  ["chapman & hall/crc cryptography and network security"] = "Chapman & Hall/CRC Cryptogr. Netw. Secur.",
  ["chapman & hall/crc data mining and knowledge discovery series"] = "Chapman & Hall/CRC Data Min. Knowl. Discov. Ser.",
  ["chapman & hall/crc financial mathematics series"] = "Chapman & Hall/CRC Financ. Math. Ser.",
  ["chapman & hall/crc handbooks of modern statistical methods"] = "Chapman Hall CRC Handb Mod Stat Methods",
  ["chapman & hall/crc interdisciplinary statistics series"] = "Chapman & Hall/CRC Interdiscip. Stat. Ser.",
  ["chapman & hall/crc machine learning & pattern recognition series"] = "Chapman & Hall/CRC Mach. Learn. Pattern Recogn. Ser.",
  ["chapman & hall/crc mathematical & computational biology series"] = "Chapman Hall CRC Math Comput Biol Ser",
  ["chapman & hall/crc mathematical and computational biology series"] = "Chapman & Hall/CRC Math. Comput. Biol. Ser.",
  ["chapman & hall/crc mathematical and computational imaging sciences"] = "Chapman & Hall/CRC Math. Comput. Imaging Sci.",
  ["chapman & hall/crc mathematical biology series"] = "Chapman Hall/CRC Math. Biol. Ser.",
  ["chapman & hall/crc mathematics"] = "Chapman & Hall/CRC Math.",
  ["chapman & hall/crc numerical analysis and scientific computing"] = "Chapman & Hall/CRC Numer. Anal. Sci. Comput.",
  ["chapman & hall/crc series in actuarial science"] = "Chapman & Hall/CRC Ser. Actuar. Sci.",
  ["chapman & hall/crc series in operations research"] = "Chapman & Hall/CRC Ser. Oper. Res.",
  ["chapman & hall/crc texts in statistical science series"] = "Chapman & Hall/CRC Texts Stat. Sci. Ser.",
  ["chapman and hall mathematics series"] = "Chapman and Hall Math. Ser.",
  ["chapman mathematical notes"] = "Chapman Math. Notes",
  ["character and personality"] = "Character Pers",
  ["charadrius : zeitschrift für vogelkunde, vogelschutz und naturschutz in rheinland und in westfalen"] = "Charadrius (Bonn)",
  ["chatelaine (toronto, ont. : 1928)"] = "Chatelaine",
  ["chebyshevskii sbornik"] = "Chebyshevskii Sb.",
  ["check list"] = "Check List",
  ["check list (luís felipe toledo)"] = "Check List",
  ["chekhoslovatskaia biologiia"] = "Chekhoslovatskaia Biol",
  ["chekhoslovatskaia fiziologiia"] = "Chekh Fiziol",
  ["chelonian conservation and biology"] = "Chelonian Conserv. Biol.",
  ["chelonian conservation and biology : journal of the iucn/ssc tortoise and freshwater turtle specialist group and international bulletin of chelonian research"] = "Chelonian Conserv Biol",
  ["chelyabinskii fiziko-matematicheskii zhurnal"] = "Chelyab. Fiz.-Mat. Zh.",
  ["chelyabinskiy universitet"] = "Vestnik Chelyabinsk. Univ. Ser. 3 Mat. Mekh.",
  ["chem catalysis"] = "Chem Catal.",
  ["chem-bio informatics journal"] = "Chem-Bio Inf. J.",
  ["chembiochem : a european journal of chemical biology"] = "Chembiochem",
  ["chembioeng reviews"] = "ChemBioEng Rev.",
  ["chemia analityczna"] = "Chem Analityczna",
  ["chemia analityczna (warsaw, poland)"] = "Chem. Anal. (Warsaw, Pol.)",
  ["chemica scripta"] = "Chem Scr",
  ["chemica sinica"] = "Chem. Sin.",
  ["chemical & biomedical imaging"] = "Chem. Biomed. Imaging",
  ["chemical & pharmaceutical bulletin"] = "Chem Pharm Bull (Tokyo)",
  ["chemical analysis"] = "Chem. Anal.",
  ["chemical and biochemical engineering quarterly"] = "Chem. Biochem. Eng. Q.",
  ["chemical and biological technologies in agriculture"] = "Chem. Biol. Technol. Agric.",
  ["chemical and engineering news"] = "Chem. Eng. News",
  ["chemical and petroleum engineering"] = "Chem. Pet. Eng.",
  ["chemical and pharmaceutical bulletin"] = "Chem. Pharm. Bull.",
  ["chemical and physical processes in combustion"] = "Chem. Phys. Processes Combust.",
  ["chemical biology & drug design"] = "Chem Biol Drug Des",
  ["chemical biology and drug design"] = "Chem. Biol. Drug Des.",
  ["chemical communications"] = "Chem. Commun.",
  ["chemical communications (cambridge, england)"] = "Chem Commun (Camb)",
  ["chemical communications (cambridge, united kingdom)"] = "Chem. Commun. (Cambridge, U. K.)",
  ["chemical data collections"] = "Chem. Data Collect.",
  ["chemical dependencies"] = "Chem Depend",
  ["chemical education research and practice"] = "Chem. Educ. Res. Pract.",
  ["chemical education research and practice in europe"] = "CHem. Educ. Res. Pract. Eur.",
  ["chemical educator"] = "Chem. Educ.",
  ["chemical educator (boise, idaho)"] = "Chem Educ",
  ["chemical engineer (london)"] = "Chem. Eng. (London)",
  ["chemical engineering & technology"] = "Chem Eng Technol",
  ["chemical engineering (new york)"] = "Chem. Eng. (N. Y.)",
  ["chemical engineering and processing"] = "Chem. Eng. Process.",
  ["chemical engineering and processing = génie des procédés = verfahrenstechnik"] = "Chem Eng Process",
  ["chemical engineering and processing: process intensification"] = "Chem. Eng. Process. Process Intensif.",
  ["chemical engineering and technology"] = "Chem. Eng. Technol.",
  ["chemical engineering communications"] = "Chem. Eng. Commun.",
  ["chemical engineering education"] = "Chem. Eng. Educ.",
  ["chemical engineering journal"] = "Chem. Eng. J.",
  ["chemical engineering journal (amsterdam, netherlands)"] = "Chem. Eng. J. (Amsterdam, Neth.)",
  ["chemical engineering journal (lausanne)"] = "Chem. Eng. J. (Lausanne)",
  ["chemical engineering journal (lausanne, switzerland : 1996)"] = "Chem Eng J",
  ["chemical engineering journal advances"] = "Chem. Eng. J. Adv.",
  ["chemical engineering journal and the biochemical engineering journal"] = "Chem. Eng. J.",
  ["chemical engineering progress"] = "Chem. Eng. Prog.",
  ["chemical engineering research & design : transactions of the institution of chemical engineers"] = "Chem Eng Res Des",
  ["chemical engineering research and design"] = "Chem. Eng. Res. Des.",
  ["chemical engineering science"] = "Chem. Eng. Sci.",
  ["chemical engineering science: x"] = "Chem. Eng. Sci.: X",
  ["chemical engineering series"] = "Chem. Eng. Ser.",
  ["chemical engineering transactions"] = "Chem Eng Trans",
  ["chemical fibers international"] = "Chem. Fibers Int.",
  ["chemical geology"] = "Chem. Geol.",
  ["chemical health & safety"] = "Chem Health Saf",
  ["chemical health and safety"] = "Chem. Health Saf.",
  ["chemical heritage"] = "Chem Herit",
  ["chemical heritage / chf"] = "Chem Herit",
  ["chemical immunology"] = "Chem Immunol",
  ["chemical immunology and allergy"] = "Chem Immunol Allergy",
  ["chemical industry & chemical engineering quarterly"] = "Chem. Ind. Chem. Eng. Q.",
  ["chemical informatics (wilmington, del.)"] = "Chem Inform",
  ["chemical journal of armenia"] = "Chem J Armen",
  ["chemical papers"] = "Chem. Pap.",
  ["chemical papers - chemicke zvesti"] = "Chem. Pap. - Chem. Zvesti",
  ["chemical physics"] = "Chem. Phys.",
  ["chemical physics impact"] = "Chem. Phys. Impact",
  ["chemical physics letters"] = "Chem. Phys. Lett.",
  ["chemical physics letters: x"] = "Chem. Phys. Lett.: X",
  ["chemical physics research journal"] = "Chem. Phys. Res. J.",
  ["chemical physics reviews"] = "Chem. Phys. Rev.",
  ["chemical processing"] = "Chem. Process.",
  ["chemical product and process modeling"] = "Chem. Prod. Process Model.",
  ["chemical record"] = "Chem. Rec.",
  ["chemical record (new york, n.y.)"] = "Chem Rec",
  ["chemical research in chinese universities"] = "Chem. Res. Chin. Univ.",
  ["chemical research in toxicology"] = "Chem Res Toxicol",
  ["chemical reviews"] = "Chem Rev",
  ["chemical reviews (washington, dc, united states)"] = "Chem. Rev. (Washington, DC, U. S.)",
  ["chemical science"] = "Chem Sci",
  ["chemical science transactions"] = "Chem. Sci. Trans.",
  ["chemical sciences journal"] = "Chem Sci J",
  ["chemical senses"] = "Chem Senses",
  ["chemical society reviews"] = "Chem Soc Rev",
  ["chemical speciation and bioavailability"] = "Chem. Speciation Bioavailability",
  ["chemical thermodynamics and thermal analysis"] = "Chem. Thermodyn. Therm. Anal.",
  ["chemical vapor deposition"] = "Chem. Vap. Deposition",
  ["chemical week"] = "Chem Week",
  ["chemically modified surfaces : recent developments. symposium on chemically modified surfaces (6th : 1995 : san jose, calif.)"] = "Chem Modif Surf Symp Chem Modif Surf",
  ["chemicke listy"] = "Chem. Listy",
  ["chemicke zvesti"] = "Chem Zvesti",
  ["chemické zvesti"] = "Chem Zvesti",
  ["chemicky prumysl"] = "Chem. Prum.",
  ["chemico-biological interactions"] = "Chem Biol Interact",
  ["chemie anlagen und verfahren"] = "Chem. Anlagen Verfahren",
  ["chemie der erde"] = "Chem. Erde",
  ["chemie der erde - geochemistry"] = "Chem. Erde",
  ["chemie der erde : beitrage zur chemischen mineralogie, petrographie und geologie"] = "Chem Erde",
  ["chemie der erde : beiträge zur chemischen mineralogie, petrographie und geologie"] = "Chem Erde",
  ["chemie in unserer zeit"] = "Chem. unserer Zeit",
  ["chemie ingenieur technik"] = "Chem. Ing. Tech.",
  ["chemika chronika"] = "Chem. Chron.",
  ["chemine technologija (kaunas, lithuania)"] = "Chem. Technol. (Kaunas, Lith.)",
  ["chemioterapia : international journal of the mediterranean society of chemotherapy"] = "Chemioterapia",
  ["chemisch weekblad"] = "Chem Weekbl",
  ["chemische berichte"] = "Chem. Ber.",
  ["chemische berichte-recueil"] = "Chem. Ber. Recl.",
  ["chemische industrie (duesseldorf)"] = "Chem. Ind. (Duesseldorf)",
  ["chemische revue uber die fett- und harz-industrie"] = "Chem. Rev. Fett- Harz-Ind.",
  ["chemische technik"] = "Chem. Tech. (Leipzig)",
  ["chemische technik (leipzig)"] = "Chem. Tech. (Leipzig)",
  ["chemische umschau auf dem gebiete der fette, oele, wachse und harze"] = "Chem. Umsch. Geb. Fette, Oele, Wachse Harze",
  ["chemisches zentralblatt"] = "Chem. Zentralbl.",
  ["chemistry & biodiversity"] = "Chem Biodivers",
  ["chemistry & biology"] = "Chem Biol",
  ["chemistry & biology (cambridge, ma, united states)"] = "Chem. Biol. (Cambridge, MA, U. S.)",
  ["chemistry & biology (oxford, united kingdom)"] = "Chem. Biol. (Oxford, U. K.)",
  ["chemistry & biology interface"] = "Chem. Biol. Interface",
  ["chemistry & industry"] = "Chem Ind",
  ["chemistry & industry (chichester, united kingdom)"] = "Chem. Ind. (Chichester, U. K.)",
  ["chemistry & industry (london, united kingdom)"] = "Chem. Ind. (London, U. K.)",
  ["chemistry (rajkot, india)"] = "Chemistry (Rajkot, India)",
  ["chemistry (weinheim an der bergstrasse, germany)"] = "Chemistry",
  ["chemistry - a european journal"] = "Chem. - Eur. J.",
  ["chemistry - an asian journal"] = "Chem. Asian J.",
  ["chemistry - methods"] = "Chem. - Methods",
  ["chemistry a european journal"] = "Chem. Eur. J.",
  ["chemistry africa"] = "Chem. Afr.",
  ["chemistry an asian journal"] = "Chem. Asian J.",
  ["chemistry and biodiversity"] = "Chem. Biodivers.",
  ["chemistry and biology"] = "Chem. Biol.",
  ["chemistry and ecology"] = "Chem. Ecol.",
  ["chemistry and industry"] = "Chem. Ind. (London)",
  ["chemistry and physics of carbon"] = "Chem. Phys. Carbon",
  ["chemistry and physics of lipids"] = "Chem Phys Lipids",
  ["chemistry and technology of fuels and oils"] = "Chem. Technol. Fuels Oils",
  ["chemistry central journal"] = "Chem Cent J",
  ["chemistry education research and practice"] = "Chem. Educ. Res. Pract.",
  ["chemistry express"] = "Chem. Express",
  ["chemistry in australia"] = "Chem. Aust.",
  ["chemistry in britain"] = "Chem. Br.",
  ["chemistry in canada"] = "Chem Can",
  ["chemistry in ecology"] = "Chem Ecol",
  ["chemistry international"] = "Chem. Int.",
  ["chemistry journal of moldova"] = "Chem. J. Mold.",
  ["chemistry letters"] = "Chem. Lett.",
  ["chemistry methods"] = "Chem. Methods",
  ["chemistry of heterocyclic compounds"] = "Chem. Heterocycl. Compd.",
  ["chemistry of heterocyclic compounds (new york, ny, united states)"] = "Chem. Heterocycl. Compd. (N. Y., NY, U. S.)",
  ["chemistry of materials"] = "Chem. Mater.",
  ["chemistry of materials : a publication of the american chemical society"] = "Chem Mater",
  ["chemistry of natural compounds"] = "Chem. Nat. Compd.",
  ["chemistry squared"] = "Chem Sq",
  ["chemistry teacher international"] = "Chem. Teach. Int.",
  ["chemistry teacher international : best practices in chemistry education"] = "Chem Teach Int",
  ["chemistry teacher international: best practices in chemistry education"] = "Chem. Teach. Int. Best Pract. Chem. Educ.",
  ["chemistry, an asian journal"] = "Chem Asian J",
  ["chemistry–a european journal"] = "Chem.–Eur. J.",
  ["chemistry–an asian journal"] = "Chem.–Asian J.",
  ["chemnanomat : chemistry of nanomaterials for energy, biology and more"] = "ChemNanoMat",
  ["chemometrics and intelligent laborary systems"] = "Chemom. Intell. Lab. Syst.",
  ["chemometrics and intelligent laboratory systems"] = "Chemom. Intell. Lab. Syst.",
  ["chemometrics and intelligent laboratory systems : an international journal sponsored by the chemometrics society"] = "Chemometr Intell Lab Syst",
  ["chemometrics research studies series"] = "Chemometrics Res. Stud. Ser.",
  ["chemosensors (basel, switzerland)"] = "Chemosensors (Basel)",
  ["chemosensory perception"] = "Chemosens. Percept.",
  ["chemotherapy (basel, switzerland)"] = "Chemotherapy (Basel, Switz.)",
  ["chemotherapy (tokyo)"] = "Chemotherapy (Tokyo)",
  ["chemotherapy research and practice"] = "Chemother Res Pract",
  ["chemphyschem : a european journal of chemical physics and physical chemistry"] = "Chemphyschem",
  ["chemrxiv : the preprint server for chemistry"] = "ChemRxiv",
  ["chemtracts: inorganic chemistry"] = "Chemtracts: Inorg. Chem.",
  ["chemtracts: macromolecular chemistry"] = "Chemtracts: Macromol. Chem.",
  ["chemtracts: organic chemistry"] = "Chemtracts: Org. Chem.",
  ["chen tzu yen chiu (acupuncture research)"] = "Chen Tzu Yen Chiu",
  ["chen tzu yen chiu acupuncture research"] = "Chen Tzu Yen Chiu",
  ["cheng shi she ji (2015)"] = "Cheng Shi She Ji (2015)",
  ["chest surgery clinics of north america"] = "Chest Surg Clin N Am",
  ["chiba igakkai zasshi. the journal of chiba medical society"] = "J Chiba Med Soc",
  ["chiba kenritsu chūō hakubutsukan shizenshi kenkyū hōkoku = journal of the natural history museum and institute, chiba"] = "J Nat Hist Mus Inst Chiba",
  ["chicago bar record"] = "Chicago Bar Rec",
  ["chicago history"] = "Chic Hist",
  ["chicago journal of international law"] = "Chic J Int Law",
  ["chicago journal of theoretical computer science"] = "Chic. J. Theoret. Comput. Sci.",
  ["chicago lectures in mathematics"] = "Chicago Lectures in Math.",
  ["chicago lectures in physics"] = "Chicago Lectures Phys.",
  ["chicago medical school quarterly"] = "Chic. Med. Sch. Q.",
  ["chicago medicine"] = "Chic Med",
  ["chicago policy review"] = "Chic Pol Rev",
  ["chicago studies"] = "Chic Stud",
  ["chicago today (university of chicago)"] = "Chic Today",
  ["chicago-kent law review"] = "Chic Kent Law Rev",
  ["chicano law review"] = "Chicano Law Rev",
  ["chief information officer journal"] = "Chief Inf Off J",
  ["chikusan sōchi kenkyūjo kenkyū hōkoku"] = "Chikusan Sochi Kenkyujo Kenkyu Hokoku",
  ["chikwa kijae hakhoe chi (journal of the korea research society for dental materials)"] = "Chikwa Kijae Hakhoe Chi",
  ["child & adolescent psychopharmacology news"] = "Child Adolesc Psychopharmacol  News",
  ["child & adolescent social work journal : c & a"] = "Child Adolesc Social Work J",
  ["child & family behavior therapy"] = "Child Fam Behav Ther",
  ["child & family social work"] = "Child Fam Soc Work",
  ["child & youth care forum"] = "Child Youth Care Forum",
  ["child & youth services"] = "Child Youth Serv",
  ["child abuse & neglect"] = "Child Abuse Negl",
  ["child abuse and neglect"] = "Child Abuse Negl.",
  ["child abuse review (chichester, england : 1992)"] = "Child Abuse Rev",
  ["child and adolescent mental health"] = "Child Adolesc Ment Health",
  ["child and adolescent obesity (abingdon, england)"] = "Child Adolesc Obes",
  ["child and adolescent psychiatric clinics of north america"] = "Child Adolesc Psychiatr Clin N Am",
  ["child and adolescent psychiatry and mental health"] = "Child Adolesc Psychiatry Ment Health",
  ["child and family"] = "Child Fam",
  ["child and family law quarterly"] = "Child Fam Law Q",
  ["child and family policy and practice review"] = "Child Fam Policy Pract Rev",
  ["child care in practice : northern ireland journal of multi-disciplinary child care practice"] = "Child Care Pract",
  ["child care quarterly"] = "Child Care Q",
  ["child development"] = "Child Dev",
  ["child development perspectives"] = "Child Dev Perspect",
  ["child development research"] = "Child Dev Res",
  ["child health alert"] = "Child Health Alert",
  ["child health dialogue"] = "Child Health Dialogue",
  ["child indicators research"] = "Child Indic Res",
  ["child language teaching and therapy"] = "Child Lang Teach Ther",
  ["child maltreatment"] = "Child Maltreat",
  ["child nephrology and urology"] = "Child Nephrol Urol",
  ["child neurology open"] = "Child Neurol Open",
  ["child neuropsychology : a journal on normal and abnormal development in childhood and adolescence"] = "Child Neuropsychol",
  ["child psychiatry and human development"] = "Child Psychiatry Hum Dev",
  ["child psychiatry quarterly"] = "Child Psychiatry Q",
  ["child psychology & psychiatry review"] = "Child Psychol Psychiatry Review",
  ["child studies in asia-pacific context"] = "Child Stud Asia Pac Context",
  ["child study; a journal of parent education"] = "Child Study",
  ["child survival action news"] = "Child Surviv Action News",
  ["child survival action news /"] = "Child Surviv Action News",
  ["child welfare"] = "Child Welfare",
  ["child's brain"] = "Childs Brain",
  ["child's nervous system : chns : official journal of the international society for pediatric neurosurgery"] = "Childs Nerv Syst",
  ["child: care, health and development"] = "Child Care Health Dev",
  ["childhood (copenhagen, denmark)"] = "Childhood",
  ["childhood education"] = "Child Educ",
  ["childhood obesity (print)"] = "Child Obes",
  ["children & libraries"] = "Child Libr",
  ["children & schools"] = "Child Sch",
  ["children & society"] = "Child Soc",
  ["children (basel, switzerland)"] = "Children (Basel)",
  ["children (washington, d.c.)"] = "Children",
  ["children and teenagers"] = "Child Teenagers",
  ["children and youth services review"] = "Child Youth Serv Rev",
  ["children in the tropics"] = "Child Trop",
  ["children today"] = "Child Today",
  ["children worldwide"] = "Child Worldw",
  ["children's geographies"] = "Child Geogr",
  ["children's health care : journal of the association for the care of children's health"] = "Child Health Care",
  ["children's legal rights journal"] = "Child Leg Rights J",
  ["children, youth and environments"] = "Child Youth Environ",
  ["children’s health care"] = "Child. Health Care",
  ["children’s health care : journal of the association for the care of children’s health"] = "Child Health Care",
  ["children’s legal rights journal"] = "Child Leg Rights J",
  ["childs nervous system"] = "Childs Nerv Syst",
  ["child’s brain"] = "Childs Brain",
  ["child’s nervous system"] = "Childs Nerv. Syst.",
  ["child’s nervous system : chns : official journal of the international society for pediatric neurosurgery"] = "Childs Nerv Syst",
  ["chilean journal of agricultural & animal sciences : ex agro-ciencia"] = "Chil J Agric Anim Sci",
  ["chilean journal of agricultural and animal sciences"] = "Chil. J. Agric Anim. Sci",
  ["chilean journal of agricultural research"] = "Chil J Agric Res",
  ["chilean journal of statistics"] = "Chil. J. Stat.",
  ["chilton’s food eng."] = "Chilton’s Food Eng.",
  ["chimica acta turcica"] = "Chim. Acta Turc.",
  ["chimica e l’industria"] = "Chim. Ind.",
  ["chimica e l’industria (milan, italy)"] = "Chim. Ind. (Milan, Italy)",
  ["chimica oggi"] = "Chim. Oggi",
  ["chimica therapeutica"] = "Chim Ther",
  ["chimie actualites"] = "Chim. Actual.",
  ["chimie et industrie, genie chimique"] = "Chim. Ind. Genie Chim.",
  ["chimie nouvelle"] = "Chim. Nouv.",
  ["china (national university of singapore. east asian institute)"] = "China (Singap)",
  ["china aktuell"] = "China Aktuell",
  ["china aktuell / institut fur asienkunde"] = "China Aktuell",
  ["china cdc weekly"] = "China CDC Wkly",
  ["china center of advanced science and technology (world laboratory) symposium/workshop proceedings"] = "China Center Adv. Sci. Tech. World Lab. Sympos. Workshop Proc.",
  ["china communications"] = "China Commun.",
  ["china economic journal"] = "China Economic J",
  ["china economic review"] = "China Econ. Rev.",
  ["china environmental science"] = "China Environ Sci",
  ["china foundry"] = "China Foundry",
  ["china geology"] = "China Geol.",
  ["china health review"] = "China Health Rev",
  ["china journal (canberra, a.c.t.)"] = "China J",
  ["china law and practice"] = "China Law Pract",
  ["china national journal of new gastroenterology"] = "Zhongguo xin xiao hua bing xue za zhi|China Natl. J. New Gastroenterol.",
  ["china national journal of new gastroenterology = zhongguo xin xiao hua bing xue za zhi"] = "China Natl J New Gastroenterol",
  ["china newsletter"] = "China Newsl",
  ["china ocean engineering"] = "China Ocean Eng.",
  ["china population and development studies"] = "China Popul Dev Stud",
  ["china population headliners"] = "China Popul Headl",
  ["china population newsletter"] = "China Popul Newsl",
  ["china population research leads"] = "China Popul Res Leads",
  ["china population research newsletter"] = "China Popul Res Newsl",
  ["china population today"] = "China Popul Today",
  ["china quarterly"] = "China Quart.",
  ["china reconstructs"] = "China Reconstr",
  ["china report (new delhi, india)"] = "China Rep",
  ["china's medicine"] = "Chinas Med",
  ["china’s medicine"] = "Chinas Med.",
  ["chinese annals of history of science and technology"] = "Chinese Ann. Hist. Sci. Tech.",
  ["chinese annals of mathematics"] = "Chin. Ann. Math.",
  ["chinese annals of mathematics, series a"] = "Chinese Ann. Math. Ser. A",
  ["chinese annals of mathematics, series b"] = "Chin. Ann. Math. Ser. B",
  ["chinese annals of mathematics. series a. shuxue niankan. a ji"] = "Chinese Ann. Math. Ser. A",
  ["chinese annals of mathematics. series b"] = "Chin. Ann. Math. Ser. B",
  ["chinese astronomy and astrophysics"] = "Chin. Astron. Astrophy",
  ["chinese birds : the international journal of ornithology"] = "Chin Birds",
  ["chinese chemical letters"] = "Chin. Chem. Lett.",
  ["chinese chemical letters = zhongguo hua xue kuai bao"] = "Chin Chem Lett",
  ["chinese clinical oncology"] = "Chin Clin Oncol",
  ["chinese culture"] = "Chin Cult",
  ["chinese economic studies"] = "Chin Econ Stud",
  ["chinese economy"] = "Chinese Economy",
  ["chinese environment & development (armonk, n.y.)"] = "Chin Environ Dev",
  ["chinese geographical science"] = "Chin Geogr Sci",
  ["chinese geographical science / sponsored by changchun institute of geography, chinese academy of sciences"] = "Chin Geogr Sci",
  ["chinese herbal medicines"] = "Chin Herb Med",
  ["chinese institute of engineers"] = "J. Chinese Inst. Engrs.",
  ["chinese journal of academic radiology"] = "Chin J Acad Radiol",
  ["chinese journal of aeronautics"] = "Chin. J. Aeronaut.",
  ["chinese journal of agricultural biotechnology"] = "Chin. J. Agric. Biotechnol.",
  ["chinese journal of analytical chemistry"] = "Chin. J. Anal. Chem.",
  ["chinese journal of applied mechanics"] = "Chinese J. Appl. Mech.",
  ["chinese journal of applied probability and statistics"] = "Chinese J. Appl. Probab. Statist.",
  ["chinese journal of applied probability and statistics. yingyong gailu tongji"] = "Chinese J. Appl. Probab. Statist.",
  ["chinese journal of artificial intelligence"] = "Chin. J. Artif. Intell.",
  ["chinese journal of astronomy and astrophysics"] = "Chin. J. Astron. Astrophys.",
  ["chinese journal of atmospheric sciences"] = "Chin. J. Atmos. Sci.",
  ["chinese journal of biochemistry and biophysics = acta biochimica et biophysica sinica = sheng wu hua xue yu sheng wu wu li xue bao"] = "Chin J Biochem Biophys",
  ["chinese journal of biotechnology"] = "Chin J Biotechnol",
  ["chinese journal of botany"] = "Chin J Bot",
  ["chinese journal of cancer"] = "Chin J Cancer",
  ["chinese journal of cancer research"] = "Chung-kuo yen cheng yen chiu|Chin. J. Cancer Res.",
  ["chinese journal of cancer research = chung-kuo yen cheng yen chiu"] = "Chin J Cancer Res",
  ["chinese journal of catalysis"] = "Chin. J. Catal.",
  ["chinese journal of chemical engineering"] = "Chin. J. Chem. Eng.",
  ["chinese journal of chemical physics"] = "Chin. J. Chem. Phys.",
  ["chinese journal of chemistry"] = "Chin. J. Chem.",
  ["chinese journal of chromatography"] = "Chin. J. Chromatogr.",
  ["chinese journal of computers. jisuanji xuebao"] = "Chinese J. Comput.",
  ["chinese journal of contemporary mathematics"] = "Chinese J. Contemp. Math.",
  ["chinese journal of digestive diseases"] = "Chin J Dig Dis",
  ["chinese journal of electrical engineering"] = "Chin. J.  Electr. Eng.",
  ["chinese journal of electronics"] = "Chin. J. Electron.",
  ["chinese journal of genetics"] = "Acta genetica Sinica|Chin. J. Genet.",
  ["chinese journal of genetics = acta genetica sinica"] = "Chin J Genet",
  ["chinese journal of geochemistry"] = "Chin. J. Geochem.",
  ["chinese journal of geochemistry = di qiu hua xue"] = "Chin J Geochem",
  ["chinese journal of geophysics"] = "Chin. J. Geophys.",
  ["chinese journal of inorganic chemistry"] = "Chin. J. Inorg. Chem.",
  ["chinese journal of integrative medicine"] = "Chin J Integr Med",
  ["chinese journal of mechanical engineering"] = "Chin. J. Mech. Eng.",
  ["chinese journal of mechanical engineering: additive manufacturing frontiers"] = "Chin. J. Mech. Eng.: Addit. Manuf. Front.",
  ["chinese journal of metal science & technology"] = "Chin. J. Met. Sci. Technol.",
  ["chinese journal of natural medicines"] = "Chin J Nat Med",
  ["chinese journal of nuclear physics"] = "Chin. J. Nucl. Phys.",
  ["chinese journal of numerical mathematics and applications"] = "Chinese J. Numer. Math. Appl.",
  ["chinese journal of oceanology and limnology"] = "Chin. J. Oceanol. Limnol.",
  ["chinese journal of oceanology and limnology = zhongguo hai yang hu zhao xue bao"] = "Chin J Oceanol Limnol",
  ["chinese journal of organic chemistry"] = "Chin. J. Org. Chem.",
  ["chinese journal of physics"] = "Chin. J. Phys.",
  ["chinese journal of physics [translation of wuli xuebao (acta physica sinica)]"] = "Chin. J. Phys.",
  ["chinese journal of physiology"] = "Chin. J. Physiol.",
  ["chinese journal of polymer science"] = "Chin. J. Polym. Sci.",
  ["chinese journal of population science"] = "Chin J Popul Sci",
  ["chinese journal of population, resources and environment"] = "Chin. J. Popul. Resour. Environ.",
  ["chinese journal of preventive veterinary medicine"] = "Chin J Prev Vet Med",
  ["chinese journal of sociology"] = "Chin J Sociol",
  ["chinese journal of structural chemistry"] = "Chin. J. Struct. Chem.",
  ["chinese journal of traumatology"] = "Chin. J. Traumatol.",
  ["chinese journal of traumatology = zhonghua chuang shang za zhi"] = "Chin J Traumatol",
  ["chinese journal of urban and environmental studies"] = "Chin. J. Urban Environ. Stud.",
  ["chinese journal of wood science and technology"] = "Chin. J. Wood Sci. Technol.",
  ["chinese medical journal"] = "Chin Med J (Engl)",
  ["chinese medical journal (peking, china : 1932)"] = "Chin Med J",
  ["chinese medical sciences journal"] = "Chin. Med. Sci. J.",
  ["chinese medical sciences journal = chung-kuo i hsüeh k'o hsüeh tsa chih"] = "Chin Med Sci J",
  ["chinese medicine"] = "Chin Med",
  ["chinese neurosurgical journal"] = "Chin Neurosurg J",
  ["chinese optics letters"] = "Chin. Opt. Lett.",
  ["chinese optics letters : col"] = "Chin Opt Lett",
  ["chinese physics"] = "Chin. Phys.",
  ["chinese physics b"] = "Chin. Phys. B",
  ["chinese physics b = zhongguo wu li b"] = "Chin Phys B",
  ["chinese physics c"] = "Chin. Phys. C",
  ["chinese physics letters"] = "Chin. Phys. Lett.",
  ["chinese quarterly journal of mathematics"] = "Chinese Quart. J. Math.",
  ["chinese quarterly journal of mathematics. shuxue jikan"] = "Chinese Quart. J. Math.",
  ["chinese science"] = "Chin Sci",
  ["chinese science bulletin"] = "Chin. Sci. Bull.",
  ["chinese science bulletin = kexue tongbao"] = "Chin Sci Bull",
  ["chinese sociological review"] = "Chin Sociol Rev",
  ["chinese sociology and anthropology"] = "Chin Sociol Anthropol",
  ["chinese tobacco science"] = "Chin Tob Sci",
  ["chinese-german journal of clinical oncology"] = "Chin. Ger. J. Clin. Oncol.",
  ["chirigaku hyoron"] = "Chirigaku Hyoron",
  ["chirigaku hyōron"] = "Chirigaku Hyoron",
  ["chiron. mitteilungen der kommission für alte geschichte und epigraphik des deutschen archäologischen instituts"] = "Chiron",
  ["chiropodist (society of chiropodists)"] = "Chiropodist (Lond)",
  ["chiropody record"] = "Chirop Rec",
  ["chiropractic & manual therapies"] = "Chiropr Man Therap",
  ["chiropractic & osteopathy"] = "Chiropr Osteopat",
  ["chiropractic history : the archives and journal of the association for the history of chiropractic"] = "Chiropr Hist",
  ["chirurgia (bucharest, romania : 1990)"] = "Chirurgia (Bucur)",
  ["chirurgia degli organi di movimento"] = "Chir. Organi Mov.",
  ["chirurgia e patologia sperimentale"] = "Chir Patol Sper",
  ["chirurgia italiana"] = "Chir Ital",
  ["chirurgia maxillofacialis & plastica"] = "Chir Maxillofac Plast",
  ["chirurgia maxillofacialis et plastica"] = "Chir. Maxillofac. Plast.",
  ["chirurgia narzadow ruchu i ortopedia polska"] = "Chir Narzadow Ruchu Ortop Pol",
  ["chirurgia narzadów ruchu i ortopedia polska"] = "Chir Narzadow Ruchu Ortop Pol",
  ["chirurgia urologica"] = "Chir Urol",
  ["chirurgie de la main"] = "Chir Main",
  ["chirurgie pediatrique"] = "Chir. Pediatr.",
  ["chirurgie pédiatrique"] = "Chir Pediatr",
  ["chirurgie; memoires de l’academie de chirurgie"] = "Chirurgie",
  ["chirurgie; mémoires de l'académie de chirurgie"] = "Chirurgie",
  ["chirurgien-dentiste de france"] = "Chir. Dent. Fr.",
  ["chirurgisches forum fur experimentelle und klinische forschung"] = "Chir. Forum Exp. Klin. Forsch.",
  ["chirurgisches forum für experimentelle und klinische forschung"] = "Chir Forum Exp Klin Forsch",
  ["chittagong university studies"] = "Chittagong Univ. Stud. Part II Sci.",
  ["chitty's law journal"] = "Chittys Law J",
  ["chitty’s law journal"] = "Chittys Law J",
  ["cho-sŏn min-ju-ju-ŭi in-min kong-hwa-kuk kwa-hak-wŏn"] = "Cho-sŏn In-min Kong-hwa-kuk Kwa-hak-wŏn T’ong-bo",
  ["choice (chicago, ill.)"] = "Choice (Middletown)",
  ["choices (new york, n.y.)"] = "Choices",
  ["choices : the magazine of food, farm, and resource issues"] = "Choices (Ames)",
  ["choices in respiratory management"] = "Choices Respir Manage",
  ["choix de travaux étrangers"] = "Choix Trav Etrang",
  ["chongqing shi fan da xue xue bao. zi ran ke xue ban"] = "Chongqing Shifan Daxue Xuebao Zi Ran Ke Xue Ban",
  ["chongqing yi xue"] = "Chongqing Yi Xue",
  ["chonnam medical journal"] = "Chonnam Med J",
  ["chonnam uidae chapchi"] = "Chonnam Med J",
  ["chornomorsʹkyĭ botanichnyĭ z︠h︡urnal = chornomorski botanical journal"] = "Chornomorskyi Bot Zhurnal",
  ["chosen i-bo. the korean medical journal"] = "Chosen Ibo",
  ["chot mai het kan phayaban"] = "Chot Mai Het Kan Phayaban",
  ["chot mai hēt kān phayābān"] = "Chot Mai Het Kan Phayaban",
  ["chŏngsin pogŏn kwa sahoe saŏp"] = "Chongsin Pogon Kwa Sahoe Saop",
  ["chŏnnam uidae chapchi"] = "Chonnam Uidae Chapchi",
  ["christ hospital medical bulletin"] = "Christ Hosp Med Bull",
  ["christ to the world"] = "Christ World",
  ["christian bioethics"] = "Christ Bioeth",
  ["christian century (chicago, ill. : 1902)"] = "Christ Century",
  ["christian history (worcester, pa. : township)"] = "Christ Hist (Worcest Pa)",
  ["christian medical dental society journal"] = "Christ Med Dent Soc J",
  ["christian medical society journal"] = "Christ Med Soc J",
  ["christian nurse"] = "Christ Nurse (Mysore)",
  ["christian nurse international"] = "Christ Nurse Int",
  ["christian scholar's review"] = "Christ Sch Rev",
  ["christian scholar’s review"] = "Christ Sch Rev",
  ["christian science monitor (boston, mass. : 1983)"] = "Christ Sci Monitor",
  ["christian science monitor (boston, mass. : eastern ed.)"] = "Christ Sci Monitor (East Ed)",
  ["christian social action (washington, d.c.)"] = "Christ Soc Action",
  ["christiana albertina; kieler universitats-zeitschrift"] = "Christiana Albertina Kiel Univ Z",
  ["christiana albertina; kieler universitaẗs-zeitschrift"] = "Christiana Albertina Kiel Univ Z",
  ["christianity and crisis"] = "Christ Crisis",
  ["christianity today"] = "Christ Today",
  ["chromatographic reviews"] = "Chromatogr Rev",
  ["chromatographic science series"] = "Chromatogr. Sci. Ser.",
  ["chromatography (basel)"] = "Chromatography (Basel)",
  ["chromatography research international"] = "Chromatogr. Res. Int.",
  ["chromosome botany"] = "Chromosom Bot",
  ["chromosome research"] = "Chromosome Res.",
  ["chromosome research : an international journal on the molecular, supramolecular and evolutionary aspects of chromosome biology"] = "Chromosome Res",
  ["chronic diseases and injuries in canada"] = "Chronic Dis Inj Can",
  ["chronic diseases and translational medicine"] = "Chronic Dis Transl Med",
  ["chronic diseases in canada"] = "Chronic Dis Can",
  ["chronic illness"] = "Chronic Illn",
  ["chronic obstructive pulmonary diseases (miami, fla.)"] = "Chronic Obstr Pulm Dis",
  ["chronic pain & management"] = "Chronic Pain Manag",
  ["chronic respiratory disease"] = "Chron Respir Dis",
  ["chronic stress (thousand oaks, calif.)"] = "Chronic Stress (Thousand Oaks)",
  ["chronica botanica"] = "Chron Bot",
  ["chronica horticulturae"] = "Chron Horticult",
  ["chronicle (galveston, tex.)"] = "Chronicle",
  ["chronicle (philadelphia, pa. : 1990)"] = "Chronicle (Phila)",
  ["chronicle (royal college of physicians of edinburgh)"] = "Chron R Coll Physicians Edinb",
  ["chronicle of the new researcher"] = "Chron. New Res.",
  ["chronicle of the world health organization"] = "Chron World Health Organ",
  ["chronicles of health impact assessment"] = "Chron Health Impact Assess",
  ["chronicles of oklahoma"] = "Chron Okla",
  ["chronique d'égypte; bulletin périodique de la fondation égyptologique reine élisabeth"] = "Chron Egypte",
  ["chronique d’egypte"] = "CE",
  ["chronique d’egypte; bulletin periodique de la fondation egyptologique reine elisabeth"] = "Chron Egypte",
  ["chronique d’égypte"] = "ChronEg",
  ["chroniques d'actualité de la s.e.d.e.i.s"] = "Chron Actual SEDEIS",
  ["chroniques d’actualite de la s.e.d.e.i.s"] = "Chron Actual SEDEIS",
  ["chronobiology international"] = "Chronobiol Int",
  ["chulālongkōn wētchasān = medical journal of chulalongkorn hospital medical school (thailand)"] = "Chulalongkorn Wetchasan",
  ["chuncui shuxue yu yingyong shuxue zhuanzhu"] = "Chuncui Shuxue yu Yingyong Shuxue Zhuanzhu",
  ["chung chien wai ko ts"] = "Chih",
  ["chung yao tsai journal of chinese medicinal materials (kuang-chou shih"] = "Chung Yao Tsai",
  ["chung-hua cheng hsing shao shang wai ko tsa chih [chinese journal ofplastic surgery and burns] chung hua cheng hsing shao shang wai ko ts"] = "Chih",
  ["chung-hua chieh ho ho hu hsi tsa chih chinese journal of tuberculosis and respiratory diseases"] = "Chung Hua Chieh Ho Ho Hu Hsi Tsa Chih",
  ["chung-hua chung liu tsa chih [chinese journal of oncology]"] = "Chung Hua Chung Liu Tsa Chih",
  ["chung-hua erh pi yen hou ko tsa chih chinese journal of otorhinolaryngology"] = "Chung Hua Erh Pi Yen Hou Ko Tsa Chih",
  ["chung-hua fu chan ko tsa chih [chinese journal of obstetrics andgynecology]"] = "Chung Hua Fu Chan Ko Tsa Chih",
  ["chung-hua hsueh yeh hsueh tsa chih chinese journal of hematology"] = "Chung Hua Hsueh Yeh Hsueh Tsa Chi",
  ["chung-hua i hsueh tsa chih [chinese medical journal]"] = "Chung Hua I Hsueh Tsa Chih",
  ["chung-hua kou chiang i hsueh tsa chih chinese journal of stomatology"] = "Chung Hua Kou Chiang Hsueh Tsa Chih",
  ["chung-hua liu hsing ping hsueh tsa chih chinese journal of epidemiology (pei-ching"] = "Chung Hua Liu Hsing Ping Hsueh Tsa Chih",
  ["chung-hua min kuo hsiao erh ko i hsueh hui tsa chih chung hua min kuohsiao erh ko i hsueh hu"] = "Tsa Chih",
  ["chung-hua min kuo wei sheng wu chi mien i hsueh tsa chih chinese journalof microbiology and immunology chung hua min kuo wei sheng wu chi mien"] = "Hsueh Tsa Chih",
  ["chung-hua nei ko tsa chih chinese journal of internal medicine"] = "Chung Hua Nei Ko Tsa Chih",
  ["chung-hua ping li hsueh tsa chih [chinese journal of pathology]"] = "Chung Hua Ping Li Hsueh Tsa Chih",
  ["chung-hua wai ko tsa chih [chinese journal of surgery]"] = "Chung Hua Wai Ko Tsa Chih",
  ["chung-hua yen ko tsa chih [chinese journal of ophthalmology]"] = "Chung Hua Yen Ko Tsa Chih",
  ["chung-hua yu fang i hsueh tsa chih [chinese journal of preventive medicine]"] = "Chung Hua Yu Fang I Hsueh Tsa Chih",
  ["chung-kuo chung hsi i chieh ho tsa chih (pei-ching"] = "Chung Kuo Chung Hsi I Chieh Ho Tsa Chih",
  ["chung-kuo chung yao tsa chih china journal of chinese materia medica (pei-ching"] = "Chung Kuo Chung Yao Tsa Chih",
  ["chung-kuo i hsueh ko hsueh yuan hsueh pao acta academiae medicinae sinicae"] = "Chung Kuo I Hsueh Ko Hsueh Yuan Hsueh Pao",
  ["chung-kuo i liao chi hsieh tsa chih chinese journal of medical instrumentation (shang-hai shih"] = "Chung Kuo I Liao Chi Hsieh Tsa Chih",
  ["chung-kuo yao li hsueh pao [acta pharmacologica sinica]"] = "Chung Kuo Yao Li Hsueh Pao",
  ["chung-kuo ying yung sheng li hsueh tsa chih chinese journal of applied physiology"] = "Chung Kuo Ying Yung Sheng Li Hsueh Tsa Chih",
  ["chungang ŭidae chi = the chung-ang journal of medicine"] = "Chung Ang Ui Dai Chi",
  ["chungang ŭihak = the korean central journal of medicine"] = "Chungang Uihak",
  ["chunlasan phantusart"] = "Chunlasan Phantusart",
  ["church and society"] = "Church Soc",
  ["church history"] = "Church Hist",
  ["churchill/brown series"] = "Churchill/Brown Ser.",
  ["chuzhou xue yuan xue bao"] = "Chuzhou Xue Yuan Xue Bao",
  ["chūdoku kenkyū : chūdoku kenkyūkai jun kikanshi = the japanese journal of toxicology"] = "Chudoku Kenkyu",
  ["ch’ikwa kijae hakhoe chi. the journal of the korea research society for dental materials"] = "Chikwa Kijae Hakhoe Chi",
  ["ch’onmun hakhoe chi. the journal of the korean astronomical society. han’guk ch’onmun hakhoe"] = "J Korean Astron Soc",
  ["ci ji yi xue za zhi = tzu-chi medical journal"] = "Ci Ji Yi Xue Za Zhi",
  ["ciba clinical symposia"] = "Ciba Clin Symp",
  ["ciba foundation study group"] = "Ciba Found Study Group",
  ["ciba foundation symposium"] = "Ciba Found Symp",
  ["ciba symposia"] = "CIBA Symp",
  ["ciba symposium. [english ed.]"] = "Ciba Symp",
  ["ciciams nouvelles"] = "CICIAMS Nouv.",
  ["ciclos en la historia, la economía y la sociedad"] = "Ciclos Hist Econom Soc",
  ["ciel et terre"] = "Ciel Terre",
  ["ciencia & saude coletiva"] = "Cien Saude Colet",
  ["ciencia & trabajo : c&t"] = "Cienc Trab",
  ["ciencia (mexico city, mexico)"] = "Cienc Cienc (Mex City Mex)",
  ["ciencia biologia molecular and cellular biology"] = "Cienc. Biol. Mol. Cell. Biol.",
  ["ciencia e agrotecnologia"] = "Cienc. Agrotecnol.",
  ["ciencia e cultura"] = "Cienc Cult",
  ["ciencia e innovación en salud"] = "Cienc Innov Salud",
  ["ciencia e investigacion"] = "Cienc Invest",
  ["ciencia e investigación"] = "Cienc Invest",
  ["ciencia e investigación agraria"] = "Cienc Investig Agrar",
  ["ciencia e tecnica vitivinicola"] = "Cienc. Tec. Vitivinic.",
  ["ciencia florestal"] = "Cienc. Florestal",
  ["ciencia y desarrollo"] = "Cienc Desarro",
  ["ciencia y sociedad"] = "Cienc Soc",
  ["ciencia y tecnologia alimentaria"] = "Cienc. Tecnol. Aliment.",
  ["ciencia y tecnología"] = "Cienc. Tecn.",
  ["ciencia y técnica"] = "Cienc. Téc.",
  ["ciencias de la conducta : revista del centro caribeño de estudios post-graduados"] = "Cienc Conducta",
  ["ciencias económicas (san josé, costa rica)"] = "Cienc Econ (San Jose)",
  ["ciencias marinas"] = "Cienc Mar",
  ["ciencias médicas hispano-americanas"] = "Cienc Med Hisp Am",
  ["ciencias sociales y religión = ciências sociais e religião"] = "Cienc Soc Relig",
  ["ciencias veterinarias (heredia, costa rica)"] = "Cienc Vet (Heredia)",
  ["ciência & saúde coletiva"] = "Cien Saude Colet",
  ["ciência agronômica"] = "Cienc Agron",
  ["ciência e cultura"] = "Cienc Cult",
  ["ciência rural (online)"] = "Cienc Rural",
  ["cim bulletin"] = "CIM Bull.",
  ["cim journal"] = "CIM J.",
  ["cim series in mathematical sciences"] = "CIM Ser. Math. Sci.",
  ["cim14, conference on interdisciplinary musicology : proceedings. conference on interdisciplinary musicology (9th : 2014 : berlin, germany)"] = "CIM14 Conf Interdiscip Musicol (2014)",
  ["cime foundation subseries"] = "CIME Found. Subser.",
  ["cin plus"] = "CIN Plus",
  ["cincinnati dental society bulletin"] = "Cincinnati Dent. Soc. Bull.",
  ["cincinnati dental society bulletin (1979)"] = "Cincinnati Dent Soc Bull",
  ["cincinnati journal of medicine"] = "Cinci J Med",
  ["cinema journal"] = "Cine J",
  ["circuit world"] = "Circuit World",
  ["circuits, systems, and signal processing"] = "Circuits Systems Signal Process.",
  ["circular economy and sustainability"] = "Circ. Econ. Sustainability",
  ["circular farmaceutica"] = "Circ Farm",
  ["circular farmacéutica"] = "Circ Farm",
  ["circular letter. united states. army. forces, pacific. office of the chief surgeon"] = "Cir Lett U S Army",
  ["circular letter. united states. far east command. medical section"] = "Circ Lett U S Far East Command",
  ["circulation journal"] = "Circ. J.",
  ["circulation journal : official journal of the japanese circulation society"] = "Circ J",
  ["circulation reports"] = "Circ Rep",
  ["circulation research"] = "Circ Res",
  ["circulation. arrhythmia and electrophysiology"] = "Circ Arrhythm Electrophysiol",
  ["circulation. cardiovascular genetics"] = "Circ Cardiovasc Genet",
  ["circulation. cardiovascular imaging"] = "Circ Cardiovasc Imaging",
  ["circulation. cardiovascular interventions"] = "Circ Cardiovasc Interv",
  ["circulation. cardiovascular quality and outcomes"] = "Circ Cardiovasc Qual Outcomes",
  ["circulation. genomic and precision medicine"] = "Circ Genom Precis Med",
  ["circulation. heart failure"] = "Circ Heart Fail",
  ["circulation: arrhythmia and electrophysiology"] = "Circ.: Arrhythmia Electrophysiol.",
  ["circulation: cardiovascular imaging"] = "Circ.: Cardiovasc. Imaging",
  ["circulation: cardiovascular interventions"] = "Circ.: Cardiovasc. Interventions",
  ["circulation: cardiovascular quality & outcomes"] = "Circ.: Cardiovasc. Qual. Outcomes",
  ["circulation: genomic and precision medicine"] = "Circ.: Genomic Precis. Med.",
  ["circulation: heart failure"] = "Circ.: Heart Failure",
  ["circulatory shock"] = "Circ Shock",
  ["circulatory shock. supplement"] = "Circ Shock Suppl",
  ["circulo odontologico de san martin y tres de febrero"] = "Circ Odontol San Martin Tres Febr",
  ["cirdap development digest"] = "CIRDAP Dev Dig",
  ["cired open access proceedings journal"] = "CIRED Open Access Proc. J.",
  ["cires, cahiers ivoiriens de recherche economique et sociale. centre ivoirien de recherches economiques et sociales"] = "CIRES Cah Ivoir Rech Econ Soc",
  ["cires, cahiers ivoiriens de recherche économique et sociale. centre ivoirien de recherches économiques et sociales"] = "CIRES Cah Ivoir Rech Econ Soc",
  ["ciriec-españa, revista de economía pública, social y cooperativa"] = "CIRIEC-España",
  ["cirm jean-morlet series"] = "CIRM Jean-Morlet Ser.",
  ["cirp annals"] = "CIRP Ann.",
  ["cirp journal of manufacturing science and technology"] = "CIRP J. Manuf. Sci. Technol.",
  ["cirugia bucal"] = "Cir Bucal",
  ["cirugia espanola"] = "Cir Esp",
  ["cirugia pediatrica"] = "Cir. Pediatr.",
  ["cirugia pediatrica : organo oficial de la sociedad espanola de cirugia pediatrica"] = "Cir Pediatr",
  ["cirugia y cirujanos"] = "Cir. Cir.",
  ["cirugia, ginecologia y urologia"] = "Cir. Ginecol. Urol.",
  ["cirugía española"] = "Cir Esp",
  ["cirugía ortopédica y traumatología"] = "Cir Ortop Traumatol",
  ["cirugía pediátrica : organo oficial de la sociedad española de cirugía pediátrica"] = "Cir Pediatr",
  ["cirugía y cirujanos"] = "Cir Cir",
  ["cirugía, ginecología y urología"] = "Cir Ginecol Urol",
  ["cism courses and lectures"] = "CISM Courses and Lectures",
  ["cism international centre for mechanical sciences. courses and lectures"] = "CISM Courses and Lect.",
  ["cities & health"] = "Cities Health",
  ["cities (london, england)"] = "Cities",
  ["cities and the environment"] = "Cities Environ",
  ["citizen science : theory and practice"] = "Citiz Sci",
  ["citizen science: theory and practice"] = "Citiz. Sci.: Theory Pract.",
  ["citizenship studies"] = "Citizensh Stud",
  ["city & community"] = "City Community",
  ["city & society (washington, d.c.)"] = "City Soc (Wash)",
  ["city (london, england)"] = "City (Lond)",
  ["city and built environment"] = "City Built Environ.",
  ["city and environment interactions"] = "City Environ. Interact.",
  ["city journal (new york, n.y.)"] = "City J",
  ["cityscape (washington, d.c.)"] = "Cityscape",
  ["cityscape: a journal of policy development and research"] = "Cityscape: J. Pol. Devel. Res.",
  ["civil code annotated of the state of california / annotated and indexed by the publisher’s editorial staff. california"] = "Civ Code Annot State Calif Calif",
  ["civil code annotated of the state of california. california"] = "Civ Code Annot State Calif Calif",
  ["civil engineering"] = "Civ. Eng.",
  ["civil engineering (new york, n.y. : 1983)"] = "Civ Eng",
  ["civil engineering and environmental systems"] = "Civ. Eng. Environ. Syst.",
  ["civil engineering design"] = "Civ. Eng. Des.",
  ["civil engineering magazine archive"] = "Civ. Eng. Mag. Arch.",
  ["civil engineering science"] = "Civ. Eng. Sci.",
  ["civil liberties"] = "Civil Lib",
  ["civil war history"] = "Civ War Hist",
  ["civil war times illustrated"] = "Civ War Times Illus",
  ["civiltà classica e cristiana"] = "CCC",
  ["civiltà padana. archeologia e storia del territorio"] = "CivPad",
  ["cjc open"] = "CJC Open",
  ["cladistics : the international journal of the willi hennig society"] = "Cladistics",
  ["clao journal"] = "CLAO J.",
  ["clara rhodos"] = "ClRh",
  ["clarendon aristotle series"] = "Clarendon Aristotle Ser.",
  ["clarendon library of logic and philosophy"] = "Clarendon Lib. Logic Philos.",
  ["classic philosophical arguments"] = "Class. Philos. Argum.",
  ["classic problem collections"] = "Class. Probl. Collect.",
  ["classic texts in the sciences"] = "Class. Texts Sci.",
  ["classica et mediaevalia. revue danoise de philologie et d’histoire"] = "ClMediaev",
  ["classica et mediaevalia: revue danoise d’histoire et de philologie publiée par la société danoise pour les études anciennes et médiévales"] = "C&M",
  ["classica: boletim de pedagogia e cultura"] = "Classica(Lisboa)",
  ["classica: revista brasileira de estudos clássicos"] = "Classica(Brasil)",
  ["classical and modern literature"] = "CML",
  ["classical and quantum gravity"] = "Classical Quantum Gravity",
  ["classical antiquity"] = "Class Antiq",
  ["classical philology"] = "Class Philol",
  ["classical quarterly"] = "Class Q",
  ["classical review"] = "CR",
  ["classical studies"] = "ClassStud",
  ["classical world"] = "Class World",
  ["classici dell’economia"] = "Classici Econom.",
  ["classics in applied mathematics"] = "Classics Appl. Math.",
  ["classics in mathematics"] = "Classics Math.",
  ["classics ireland"] = "ClIre",
  ["classics of soviet mathematics"] = "Classics Soviet Math.",
  ["classroom companion. economics"] = "Classr. Companion. Econ.",
  ["classroom resource materials"] = "Classr. Resour. Mater.",
  ["classroom resource materials series"] = "Classr. Res. Mater. Ser.",
  ["clay lecture notes"] = "Clay Lect. Notes",
  ["clay mathematics proceedings"] = "Clay Math. Proc.",
  ["clay minerals"] = "Clay Miner.",
  ["clay research"] = "Clay Res.",
  ["clay technology"] = "Clay Technol.",
  ["clays and clay minerals"] = "Clays Clay Miner.",
  ["clean - soil, air, water"] = "CLEAN - Soil Air Water",
  ["clean : soil, air, water"] = "Clean (Weinh)",
  ["clean air"] = "Clean Air",
  ["clean energy"] = "Clean Energy",
  ["clean technologies"] = "Clean Technol.",
  ["clean technologies and environmental policy"] = "Clean Technol. Environ. Policy",
  ["clean-soil air water"] = "Clean-Soil Air Water",
  ["clean: soil, air, water"] = "Clean: Soil, Air, Water",
  ["cleaner and circular bioeconomy"] = "Cleaner Circ. Bioecon.",
  ["cleaner and responsible consumption"] = "Cleaner Responsible Consumption",
  ["cleaner chemical engineering"] = "Cleaner Chem. Eng.",
  ["cleaner energy systems"] = "Cleaner Energy Syst.",
  ["cleaner engineering and technology"] = "Clean Eng Technol",
  ["cleaner environmental systems"] = "Cleaner Environ. Syst.",
  ["cleaner logistics and supply chain"] = "Cleaner Logist. Supply Chain",
  ["cleaner materials"] = "Cleaner Mater.",
  ["cleaner production letters"] = "Cleaner Prod. Lett.",
  ["cleaner waste systems"] = "Cleaner Waste Syst.",
  ["clearing house (menasha, wis.)"] = "Clearing House",
  ["clearinghouse review"] = "Clgh Rev",
  ["cleft palate journal"] = "Cleft Palate J.",
  ["cleft palate-craniofacial journal"] = "Cleft Palate. Craniofac. J.",
  ["cleveland clinic journal of medicine"] = "Cleve Clin J Med",
  ["cleveland clinic quarterly"] = "Cleve Clin Q",
  ["cleveland state law review"] = "Clevel State Law Rev",
  ["cleveland studies in the history of art"] = "ClevStHistArt",
  ["climacteric : the journal of the international menopause society"] = "Climacteric",
  ["climate (basel, switzerland)"] = "Climate (Basel)",
  ["climate action"] = "Clim. Action",
  ["climate and development"] = "Clim Dev",
  ["climate and energy"] = "Clim. Energy",
  ["climate change ecology"] = "Clim. Change Ecol.",
  ["climate change economics"] = "Clim Chang Econ (Singap)",
  ["climate change research"] = "Clim. Change Res.",
  ["climate dynamics"] = "Clim. Dyn.",
  ["climate law"] = "Clim. Law",
  ["climate of the past"] = "Clim. Past",
  ["climate of the past discussion"] = "Clim. Past Discuss.",
  ["climate policy"] = "Clim. Policy",
  ["climate research"] = "Clim Res",
  ["climate resilience and sustainability"] = "Clim. Resilience Sustainability",
  ["climate risk management"] = "Clim. Risk Manage.",
  ["climate services"] = "Clim Serv",
  ["climatic change"] = "Clim. Change",
  ["climatological bulletin"] = "Climatol. Bull.",
  ["clinic (paris, france)"] = "Clinic (Paris)",
  ["clinic odontologia"] = "Clin. Odontol.",
  ["clinic odontologia : la revue de l'omnipraticien"] = "Clin Odontol",
  ["clinic odontologia : la revue de l’omnipraticien"] = "Clin Odontol",
  ["clinica (bologna, italy)"] = "Clinica",
  ["clinica chimica acta"] = "Clin. Chim. Acta",
  ["clinica chimica acta; international journal of clinical chemistry"] = "Clin Chim Acta",
  ["clinica del torax"] = "Clin Torax",
  ["clinica del tórax"] = "Clin Torax",
  ["clinica e investigacion en ginecologia y obstetricia"] = "Clin Invest Ginecol Obstet",
  ["clinica europea"] = "Clin Eur",
  ["clinica hispánica"] = "Clin. Hisp",
  ["clinica latina"] = "Clin Lat",
  ["clinica nuova; rassegna del progresso medico internazionale"] = "Clin Nuova Rass Prog Med Int",
  ["clinica odontoiatrica; revista mensile di operativa dentaria, chirurgia orale, ortodontia e protesi"] = "Clin Odontoiatr Rev Mens Oper Dent Chir Orale Ortod Protesi",
  ["clinica ortopedica"] = "Clin. Ortop.",
  ["clinica ostetrica e ginecologica"] = "Clin. Ostet. Ginecol.",
  ["clinica otorinolaringoiatrica"] = "Clin. Otorinolaringoiatr.",
  ["clinica pediatrica"] = "Clin. Pediatr. (Bologna)",
  ["clinica terapeutica"] = "Clin. Ter.",
  ["clinica tisiológica"] = "Clin Tisiol",
  ["clinica y laboratorio"] = "Clin Lab (Zaragoza)",
  ["clinica, higiene e hidrologia"] = "Clin Hig Hidrol",
  ["clinical & developmental immunology"] = "Clin Dev Immunol",
  ["clinical & experimental gastroenterology & hepatology"] = "Clin Exp Gastroenterol Hepatol",
  ["clinical & experimental immunology"] = "Clin. Exp. Immunol.",
  ["clinical & experimental metastasis"] = "Clin Exp Metastasis",
  ["clinical & experimental neuroimmunology"] = "Clin Exp Neuroimmunol",
  ["clinical & experimental ophthalmology"] = "Clin Exp Ophthalmol",
  ["clinical & experimental optometry"] = "Clin Exp Optom",
  ["clinical & experimental optometry : journal of the australian optometrical association"] = "Clin Exp Optom",
  ["clinical & experimental pharmacology"] = "Clin Exp Pharmacol",
  ["clinical & translational immunology"] = "Clin Transl Immunology",
  ["clinical & translational oncology : official publication of the federation of spanish oncology societies and of the national cancer institute of mexico"] = "Clin Transl Oncol",
  ["clinical advances in hematology & oncology : h&o"] = "Clin Adv Hematol Oncol",
  ["clinical advances in hematology and oncology"] = "Clin. Adv. Hematol. Oncol.",
  ["clinical advances in periodontics"] = "Clin Adv Periodontics",
  ["clinical allergy"] = "Clin Allergy",
  ["clinical allergy and immunology"] = "Clin Allergy Immunol",
  ["clinical anatomy"] = "Clin. Anat.",
  ["clinical anatomy (new york, n.y.)"] = "Clin Anat",
  ["clinical and applied immunology reviews"] = "Clin Appl Immunol Rev",
  ["clinical and applied thrombosis/hemostasis : official journal of the international academy of clinical and applied thrombosis/hemostasis"] = "Clin Appl Thromb Hemost",
  ["clinical and developmental immunology"] = "Clin. Dev. Immunol.",
  ["clinical and diagnostic laboratory immunology"] = "Clin Diagn Lab Immunol",
  ["clinical and diagnostic pathology"] = "Clin Diagn Pathol",
  ["clinical and diagnostic virology"] = "Clin Diagn Virol",
  ["clinical and experimental allergy"] = "Clin. Exp. Allergy",
  ["clinical and experimental allergy : journal of the british society for allergy and clinical immunology"] = "Clin Exp Allergy",
  ["clinical and experimental dental research"] = "Clin Exp Dent Res",
  ["clinical and experimental dermatology"] = "Clin Exp Dermatol",
  ["clinical and experimental dialysis and apheresis"] = "Clin Exp Dial Apheresis",
  ["clinical and experimental emergency medicine"] = "Clin Exp Emerg Med",
  ["clinical and experimental gastroenterology"] = "Clin Exp Gastroenterol",
  ["clinical and experimental hepatology"] = "Clin Exp Hepatol",
  ["clinical and experimental hypertension"] = "Clin Exp Hypertens",
  ["clinical and experimental hypertension (new york, n.y. : 1993)"] = "Clin Exp Hypertens",
  ["clinical and experimental hypertension. part a, theory and practice"] = "Clin Exp Hypertens A",
  ["clinical and experimental hypertension. part b, hypertension in pregnancy"] = "Clin Exp Hypertens B",
  ["clinical and experimental immunology"] = "Clin Exp Immunol",
  ["clinical and experimental medical sciences"] = "Clin Exp Med Sci",
  ["clinical and experimental medicine"] = "Clin Exp Med",
  ["clinical and experimental metastasis"] = "Clin. Exp. Metastasis",
  ["clinical and experimental nephrology"] = "Clin Exp Nephrol",
  ["clinical and experimental neuroimmunology"] = "Clin. Exp. Neuroimmunol.",
  ["clinical and experimental neurology"] = "Clin Exp Neurol",
  ["clinical and experimental obstetrics & gynecology"] = "Clin Exp Obstet Gynecol",
  ["clinical and experimental obstetrics and gynecology"] = "Clin. Exp. Obstet. Gynecol.",
  ["clinical and experimental ophthalmology"] = "Clin. Experiment. Ophthalmol.",
  ["clinical and experimental otorhinolaryngology"] = "Clin Exp Otorhinolaryngol",
  ["clinical and experimental pathology"] = "Clin Exp Pathol",
  ["clinical and experimental pediatrics"] = "Clin Exp Pediatr",
  ["clinical and experimental pharmacology & physiology"] = "Clin Exp Pharmacol Physiol",
  ["clinical and experimental pharmacology & physiology. supplement"] = "Clin Exp Pharmacol Physiol Suppl",
  ["clinical and experimental pharmacology and physiology"] = "Clin. Exp. Pharmacol. Physiol.",
  ["clinical and experimental pharmacology and physiology. supplement"] = "Clin. Exp. Pharmacol. Physiol. Suppl.",
  ["clinical and experimental psychology"] = "Clin Exp Psychol",
  ["clinical and experimental reproductive medicine"] = "Clin Exp Reprod Med",
  ["clinical and experimental rheumatology"] = "Clin Exp Rheumatol",
  ["clinical and experimental vaccine research"] = "Clin Exp Vaccine Res",
  ["clinical and investigative medicine"] = "Clin. Invest. Med.",
  ["clinical and investigative medicine. medecine clinique et experimentale"] = "Clin Invest Med",
  ["clinical and investigative medicine. médecine clinique et experimentale"] = "Clin Invest Med",
  ["clinical and laboratory haematology"] = "Clin Lab Haematol",
  ["clinical and molecular allergy : cma"] = "Clin Mol Allergy",
  ["clinical and molecular hepatology"] = "Clin Mol Hepatol",
  ["clinical and translational allergy"] = "Clin Transl Allergy",
  ["clinical and translational gastroenterology"] = "Clin Transl Gastroenterol",
  ["clinical and translational imaging"] = "Clin Transl Imaging",
  ["clinical and translational medicine"] = "Clin Transl Med",
  ["clinical and translational neurocience"] = "Clin. Transl. Neurosci.",
  ["clinical and translational oncology"] = "Clin. Transl. Oncol.",
  ["clinical and translational radiation oncology"] = "Clin Transl Radiat Oncol",
  ["clinical and translational science"] = "Clin Transl Sci",
  ["clinical and vaccine immunology"] = "Clin. Vaccine Immunol.",
  ["clinical and vaccine immunology : cvi"] = "Clin Vaccine Immunol",
  ["clinical anesthesia"] = "Clin Anesth",
  ["clinical archives of bone and joint diseases"] = "Clin Arch Bone Jt Dis",
  ["clinical archives of communication disorders"] = "Clin Arch Commun Disord",
  ["clinical autonomic research"] = "Clin. Auton. Res.",
  ["clinical autonomic research : official journal of the clinical autonomic research society"] = "Clin Auton Res",
  ["clinical biochemistry"] = "Clin Biochem",
  ["clinical biomechanics"] = "Clin. Biomech.",
  ["clinical biomechanics (bristol, avon)"] = "Clin Biomech (Bristol, Avon)",
  ["clinical breast cancer"] = "Clin Breast Cancer",
  ["clinical bulletin"] = "Clin Bull",
  ["clinical calcium"] = "Clin Calcium",
  ["clinical cancer drugs"] = "Clin Cancer Drugs",
  ["clinical cancer investigation journal"] = "Clin Cancer Investig J",
  ["clinical cancer research"] = "Clin. Cancer Res.",
  ["clinical cancer research : an official journal of the american association for cancer research"] = "Clin Cancer Res",
  ["clinical cancer research: an official journal of the american association for cancer research"] = "Clin. Cancer Res.",
  ["clinical cardiology"] = "Clin Cardiol",
  ["clinical case reports"] = "Clin Case Rep",
  ["clinical case reports and reviews"] = "Clin Case Rep Rev",
  ["clinical case studies"] = "Clin Case Stud",
  ["clinical cases in mineral and bone metabolism : the official journal of the italian society of osteoporosis, mineral metabolism, and skeletal diseases"] = "Clin Cases Miner Bone Metab",
  ["clinical chemistry"] = "Clin Chem",
  ["clinical chemistry (washington, dc, united states)"] = "Clin. Chem. (Washington, DC, U. S.)",
  ["clinical chemistry and laboratory medicine"] = "Clin Chem Lab Med",
  ["clinical chemistry and laboratory medicine : cclm / fescc"] = "Clin Chem Lab Med",
  ["clinical child and family psychology review"] = "Clin Child Fam Psychol Rev",
  ["clinical child psychology and psychiatry"] = "Clin Child Psychol Psychiatry",
  ["clinical colorectal cancer"] = "Clin Colorectal Cancer",
  ["clinical cornerstone"] = "Clin Cornerstone",
  ["clinical cytology"] = "Clin Cytol",
  ["clinical dermatology (wilmington, del.)"] = "Clin Dermatol (Wilmington)",
  ["clinical diabetes : a publication of the american diabetes association"] = "Clin Diabetes",
  ["clinical diabetes and endocrinology"] = "Clin Diabetes Endocrinol",
  ["clinical drug investigation"] = "Clin Drug Investig",
  ["clinical dysmorphology"] = "Clin Dysmorphol",
  ["clinical eeg (electroencephalography)"] = "Clin Electroencephalogr",
  ["clinical eeg and neuroscience"] = "Clin EEG Neurosci",
  ["clinical eeg and neuroscience : official journal of the eeg and clinical neuroscience society (encs)"] = "Clin EEG Neurosci",
  ["clinical effectiveness in nursing"] = "Clin Eff Nurs",
  ["clinical electroencephalography"] = "Clin. Electroencephalogr.",
  ["clinical endocrinology"] = "Clin Endocrinol (Oxf)",
  ["clinical endoscopy"] = "Clin Endosc",
  ["clinical engineering"] = "Clin Eng",
  ["clinical engineering information service"] = "Clin Eng Inf Serv",
  ["clinical engineering news"] = "Clin Eng News",
  ["clinical epidemiology"] = "Clin Epidemiol",
  ["clinical epidemiology and global health"] = "Clin Epidemiol Glob Health",
  ["clinical epigenetics"] = "Clin Epigenetics",
  ["clinical epigenetics: epigenetics communications"] = "Clin. Epigenet.: Epigenet. Commun.",
  ["clinical ethics"] = "Clin Ethics",
  ["clinical ethics report"] = "Clin Ethics Rep",
  ["clinical evidence"] = "Clin Evid",
  ["clinical excellence for nurse practitioners : the international journal of npace"] = "Clin Excell Nurse Pract",
  ["clinical excerpts; a journal devoted to therapeutics"] = "Clin Excerpts J Devoted Ther",
  ["clinical exercise physiology"] = "Clin Exerc Physiol",
  ["clinical eye and vision care"] = "Clin Eye Vis Care",
  ["clinical gastroenterology and hepatology"] = "Clin. Gastroenterol. Hepatol.",
  ["clinical gastroenterology and hepatology : the official clinical practice journal of the american gastroenterological association"] = "Clin Gastroenterol Hepatol",
  ["clinical genetics"] = "Clin Genet",
  ["clinical genitourinary cancer"] = "Clin Genitourin Cancer",
  ["clinical geriatrics"] = "Clin Geriatr",
  ["clinical gerontologist"] = "Clin Gerontol",
  ["clinical governance"] = "Clin Gov",
  ["clinical hematology international"] = "Clin Hematol Int",
  ["clinical hemorheology"] = "Clin Hemorheol",
  ["clinical hemorheology and microcirculation"] = "Clin Hemorheol Microcirc",
  ["clinical hypertension"] = "Clin Hypertens",
  ["clinical imaging"] = "Clin Imaging",
  ["clinical immunology"] = "Clin. Immunol.",
  ["clinical immunology (amsterdam, netherlands)"] = "Clin. Immunol. (Amsterdam, Neth.)",
  ["clinical immunology (orlando, fla.)"] = "Clin Immunol",
  ["clinical immunology (san diego, ca, united states)"] = "Clin. Immunol. (San Diego, CA, U. S.)",
  ["clinical immunology and immunopathology"] = "Clin Immunol Immunopathol",
  ["clinical immunology reviews"] = "Clin Immunol Rev",
  ["clinical immunotherapeutics"] = "Clin Immunother",
  ["clinical implant dentistry and related research"] = "Clin Implant Dent Relat Res",
  ["clinical infection in practice"] = "Clin Infect Pract",
  ["clinical infectious diseases"] = "Clin. Infect. Dis.",
  ["clinical infectious diseases : an official publication of the infectious diseases society of america"] = "Clin Infect Dis",
  ["clinical intensive care"] = "Clin. Intensive Care",
  ["clinical intensive care : international journal of critical & coronary care medicine"] = "Clin Intensive Care",
  ["clinical interventions in aging"] = "Clin Interv Aging",
  ["clinical investigation"] = "Clin Investig (Lond)",
  ["clinical investigator"] = "Clin. Investig.",
  ["clinical journal of gastroenterology"] = "Clin J Gastroenterol",
  ["clinical journal of oncology nursing"] = "Clin J Oncol Nurs",
  ["clinical journal of pain"] = "Clin. J. Pain",
  ["clinical journal of sport medicine"] = "Clin. J. Sport Med.",
  ["clinical journal of sport medicine : official journal of the canadian academy of sport medicine"] = "Clin J Sport Med",
  ["clinical journal of the american society of nephrology : cjasn"] = "Clin J Am Soc Nephrol",
  ["clinical journal of women's health"] = "Clin J Womens Health",
  ["clinical kidney journal"] = "Clin Kidney J",
  ["clinical laboratory"] = "Clin Lab",
  ["clinical laboratory international"] = "Clin Lab Int",
  ["clinical laboratory management review"] = "Clin. Lab. Manage. Rev.",
  ["clinical laboratory management review : official publication of the clinical laboratory management association"] = "Clin Lab Manage Rev",
  ["clinical laboratory management review : official publication of the clinical laboratory management association / clma"] = "Clin Lab Manage Rev",
  ["clinical laboratory science"] = "Clin. Lab. Sci.",
  ["clinical laboratory science : journal of the american society for medical technology"] = "Clin Lab Sci",
  ["clinical lactation (amarillo, tex.)"] = "Clin Lact (Amarillo)",
  ["clinical laser monthly"] = "Clin Laser Mon",
  ["clinical leadership & management review : the journal of clma"] = "Clin Leadersh Manag Rev",
  ["clinical linguistics & phonetics"] = "Clin Linguist Phon",
  ["clinical lipidology"] = "Clin Lipidol",
  ["clinical lipidology and metabolic disorders"] = "Clin. Lipidol. Metab. Disord.",
  ["clinical liver disease"] = "Clin Liver Dis (Hoboken)",
  ["clinical lung cancer"] = "Clin Lung Cancer",
  ["clinical lymphoma"] = "Clin Lymphoma",
  ["clinical lymphoma & myeloma"] = "Clin Lymphoma Myeloma",
  ["clinical lymphoma and myeloma"] = "Clin. Lymphoma Myeloma",
  ["clinical lymphoma, myeloma & leukemia"] = "Clin Lymphoma Myeloma Leuk",
  ["clinical mass spectrometry"] = "Clin. Mass Spectrom.",
  ["clinical mass spectrometry (del mar, calif.)"] = "Clin Mass Spectrom",
  ["clinical materials"] = "Clin. Mater.",
  ["clinical medical reviews and case reports"] = "Clin Med Rev Case Rep",
  ["clinical medicine"] = "Clin. Med.",
  ["clinical medicine & research"] = "Clin Med Res",
  ["clinical medicine (london, england)"] = "Clin Med (Lond)",
  ["clinical medicine (northfield, ill.)"] = "Clin Med (Northfield)",
  ["clinical medicine and research"] = "Clin. Med. Res.",
  ["clinical medicine insights. arthritis and musculoskeletal disorders"] = "Clin Med Insights Arthritis Musculoskelet Disord",
  ["clinical medicine insights. blood disorders"] = "Clin Med Insights Blood Disord",
  ["clinical medicine insights. cardiology"] = "Clin Med Insights Cardiol",
  ["clinical medicine insights. case reports"] = "Clin Med Insights Case Rep",
  ["clinical medicine insights. circulatory, respiratory and pulmonary medicine"] = "Clin Med Insights Circ Respir Pulm Med",
  ["clinical medicine insights. ear, nose and throat"] = "Clin Med Insights Ear Nose Throat",
  ["clinical medicine insights. endocrinology and diabetes"] = "Clin Med Insights Endocrinol Diabetes",
  ["clinical medicine insights. gastroenterology"] = "Clin Med Insights Gastroenterol",
  ["clinical medicine insights. oncology"] = "Clin Med Insights Oncol",
  ["clinical medicine insights. pathology"] = "Clin Med Insights Pathol",
  ["clinical medicine insights. pediatrics"] = "Clin Med Insights Pediatr",
  ["clinical medicine insights. reproductive health"] = "Clin Med Insights Reprod Health",
  ["clinical medicine insights. therapeutics"] = "Clin Med Insights Ther",
  ["clinical medicine insights. urology"] = "Clin Med Insights Urol",
  ["clinical medicine insights. women's health"] = "Clin Med Insights Womens Health",
  ["clinical medicine research"] = "Clin Med Res (N Y)",
  ["clinical medicine reviews in oncology"] = "Clin Med Rev Oncol",
  ["clinical medicine reviews in therapeutics"] = "Clin Med Rev Ther",
  ["clinical medicine reviews in vascular health"] = "Clin Med Rev Vasc Health",
  ["clinical medicine. arthritis and musculoskeletal disorders"] = "Clin Med Arthritis Musculoskelet Disord",
  ["clinical medicine. cardiology"] = "Clin Med Cardiol",
  ["clinical medicine. case reports"] = "Clin Med Case Rep",
  ["clinical medicine. circulatory, respiratory and pulmonary medicine"] = "Clin Med Circ Respirat Pulm Med",
  ["clinical medicine. dermatology"] = "Clin Med Dermatol",
  ["clinical medicine. gastroenterology"] = "Clin Med Gastroenterol",
  ["clinical medicine. geriatrics"] = "Clin Med Geriatr",
  ["clinical medicine. oncology"] = "Clin Med Oncol",
  ["clinical medicine. pathology"] = "Clin Med Pathol",
  ["clinical medicine. pediatrics"] = "Clin Med Pediatr",
  ["clinical medicine. reproductive health"] = "Clin Med Reprod Health",
  ["clinical medicine. therapeutics"] = "Clin Med Ther",
  ["clinical medicine. urology [electronic resource]"] = "Clin Med Urol",
  ["clinical microbiology & case reports"] = "Clin Microbiol Case Rep",
  ["clinical microbiology (los angeles, calif.)"] = "Clin Microbiol",
  ["clinical microbiology and infection : the official publication of the european society of clinical microbiology and infectious diseases"] = "Clin Microbiol Infect",
  ["clinical microbiology and infectious diseases"] = "Clin Microbiol Infect Dis",
  ["clinical microbiology newsletter"] = "Clin Microbiol Newsl",
  ["clinical microbiology reviews"] = "Clin Microbiol Rev",
  ["clinical molecular medicine"] = "Clin Mol Med",
  ["clinical molecular pathology"] = "Clin Mol Pathol",
  ["clinical nephrology"] = "Clin Nephrol",
  ["clinical nephrology and research"] = "Clin Nephrol Res",
  ["clinical nephrology. case studies"] = "Clin Nephrol Case Stud",
  ["clinical neurology and neurosurgery"] = "Clin Neurol Neurosurg",
  ["clinical neuropathology"] = "Clin Neuropathol",
  ["clinical neuropharmacology"] = "Clin Neuropharmacol",
  ["clinical neurophysiology"] = "Clin. Neurophysiol.",
  ["clinical neurophysiology : official journal of the international federation of clinical neurophysiology"] = "Clin Neurophysiol",
  ["clinical neurophysiology practice"] = "Clin Neurophysiol Pract",
  ["clinical neuropsychiatry"] = "Clin Neuropsychiatry",
  ["clinical neuroradiology"] = "Clin Neuroradiol",
  ["clinical neuroscience"] = "Clin. Neurosci.",
  ["clinical neuroscience (new york, n.y.)"] = "Clin Neurosci",
  ["clinical neuroscience research"] = "Clin Neurosci Res",
  ["clinical neurosurgery"] = "Clin Neurosurg",
  ["clinical notes on respiratory diseases"] = "Clin Notes Respir Dis",
  ["clinical nuclear medicine"] = "Clin Nucl Med",
  ["clinical nurse specialist"] = "Clin. Nurse Spec.",
  ["clinical nurse specialist cns"] = "Clin Nurse Spec",
  ["clinical nursing practice in epilepsy"] = "Clin Nurs Pract Epilepsy",
  ["clinical nursing research"] = "Clin Nurs Res",
  ["clinical nursing studies"] = "Clin Nurs Stud",
  ["clinical nutrition"] = "Clin. Nutr.",
  ["clinical nutrition (edinburgh, scotland)"] = "Clin Nutr",
  ["clinical nutrition espen"] = "Clin Nutr ESPEN",
  ["clinical nutrition experimental"] = "Clin Nutr Exp",
  ["clinical nutrition open science"] = "Clin. Nutr. Open Sci.",
  ["clinical nutrition research"] = "Clin Nutr Res",
  ["clinical obesity"] = "Clin Obes",
  ["clinical obstetrics and gynecology"] = "Clin Obstet Gynecol",
  ["clinical obstetrics, gynecology and reproductive medicine"] = "Clin Obstet Gynecol Reprod Med",
  ["clinical oncology"] = "Clin Oncol",
  ["clinical oncology (royal college of radiologists (great britain))"] = "Clin Oncol (R Coll Radiol)",
  ["clinical oncology (royal college of radiologists)"] = "Clin. Oncol. (R. Coll. Radiol.)",
  ["clinical oncology and cancer research"] = "Clin. Oncol. Cancer Res.",
  ["clinical oncology and research"] = "Clin Oncol Res",
  ["clinical oncology in adolescents and young adults"] = "Clin Oncol Adolesc Young Adults",
  ["clinical ophthalmology (auckland, n.z.)"] = "Clin Ophthalmol",
  ["clinical optometry"] = "Clin Optom (Auckl)",
  ["clinical oral implants research"] = "Clin Oral Implants Res",
  ["clinical oral investigations"] = "Clin Oral Investig",
  ["clinical orthodontics and research"] = "Clin Orthod Res",
  ["clinical orthopaedics"] = "Clin Orthop",
  ["clinical orthopaedics and related research"] = "Clin Orthop Relat Res",
  ["clinical otolaryngology"] = "Clin. Otolaryngol.",
  ["clinical otolaryngology : official journal of ent-uk ; official journal of netherlands society for oto-rhino-laryngology & cervico-facial surgery"] = "Clin Otolaryngol",
  ["clinical otolaryngology and allied sciences"] = "Clin Otolaryngol Allied Sci",
  ["clinical ovarian cancer and other gynecologic malignancies"] = "Clin Ovarian Cancer Other Gynecol Malig",
  ["clinical parkinsonism & related disorders"] = "Clin Park Relat Disord",
  ["clinical pathology"] = "Clin. Pathol.",
  ["clinical pathology (thousand oaks, ventura county, calif.)"] = "Clin Pathol",
  ["clinical pediatric dermatology (wilmington. del.)"] = "Clin Pediatr Dermatol",
  ["clinical pediatric emergency medicine"] = "Clin Pediatr Emerg Med",
  ["clinical pediatric endocrinology : case reports and clinical investigations : official journal of the japanese society for pediatric endocrinology"] = "Clin Pediatr Endocrinol",
  ["clinical pediatrics"] = "Clin Pediatr (Phila)",
  ["clinical performance and quality health care"] = "Clin Perform Qual Health Care",
  ["clinical pharmacokinetics"] = "Clin Pharmacokinet",
  ["clinical pharmacology & biopharmaceutics"] = "Clin Pharmacol Biopharm",
  ["clinical pharmacology & therapeutics"] = "Clin. Pharmacol. Ther.",
  ["clinical pharmacology : advances and applications"] = "Clin Pharmacol",
  ["clinical pharmacology and therapeutics"] = "Clin Pharmacol Ther",
  ["clinical pharmacology and therapeutics (st. louis)"] = "Clin. Pharmacol. Ther. (St. Louis)",
  ["clinical pharmacology and translational medicine"] = "Clin Pharmacol Transl Med",
  ["clinical pharmacology in drug development"] = "Clin Pharmacol Drug Dev",
  ["clinical pharmacology: advances and applications"] = "Clin. Pharmacol.: Adv. Appl.",
  ["clinical pharmacy"] = "Clin Pharm",
  ["clinical physics and physiological measurement"] = "Clin. Phys. Physiol. Meas.",
  ["clinical physics and physiological measurement : an official journal of the hospital physicists' association, deutsche gesellschaft für medizinische physik and the european federation of organisations for medical physics"] = "Clin Phys Physiol Meas",
  ["clinical physics and physiological measurement : an official journal of the hospital physicists’ association, deutsche gesellschaft fur medizinische physik and the european federation of organisations for medical physics"] = "Clin Phys Physiol Meas",
  ["clinical physiology"] = "Clin. Physiol.",
  ["clinical physiology (oxford, england)"] = "Clin Physiol",
  ["clinical physiology and biochemistry"] = "Clin Physiol Biochem",
  ["clinical physiology and functional imaging"] = "Clin Physiol Funct Imaging",
  ["clinical physiology series"] = "Clin. Phys.",
  ["clinical phytoscience"] = "Clin. Phytosci.",
  ["clinical plasma medicine"] = "Clin Plasma Med",
  ["clinical positron imaging : official journal of the institute for clinical p.e.t"] = "Clin Positron Imaging",
  ["clinical practice (london, england)"] = "Clin Pract (Lond)",
  ["clinical practice and cases in emergency medicine"] = "Clin Pract Cases Emerg Med",
  ["clinical practice and epidemiology in mental health : cp & emh"] = "Clin Pract Epidemiol Ment Health",
  ["clinical practice guideline. quick reference guide for clinicians"] = "Clin Pract Guidel Quick Ref Guide Clin",
  ["clinical practice in pediatric psychology"] = "Clin Pract Pediatr Psychol",
  ["clinical preventive dentistry"] = "Clin Prev Dent",
  ["clinical privilege white paper"] = "Clin Privil White Pap",
  ["clinical proceedings"] = "Clin Proc",
  ["clinical proceedings - children's hospital national medical center"] = "Clin Proc Child Hosp Natl Med Cent",
  ["clinical proceedings - children's hospital of the district of columbia"] = "Clin Proc Child Hosp Dist Columbia",
  ["clinical proceedings - children’s hospital national medical center"] = "Clin Proc Child Hosp Natl Med Cent",
  ["clinical proceedings - children’s hospital of the district of columbia"] = "Clin Proc Child Hosp Dist Columbia",
  ["clinical proceedings, childrens hospital of the district of columbia"] = "Clin. Proc. Child. Hosp. Dist. Columbia",
  ["clinical prostate and kidney cancer"] = "Clin. Prostate Kidney Cancer",
  ["clinical prostate cancer"] = "Clin Prostate Cancer",
  ["clinical proteomics"] = "Clin Proteomics",
  ["clinical psychiatry (wilmington, del.)"] = "Clin Psychiatry (Wilmington)",
  ["clinical psychological science : a journal of the association for psychological science"] = "Clin Psychol Sci",
  ["clinical psychologist (australian psychological society)"] = "Clin Psychol (Aust Psychol Soc)",
  ["clinical psychology & psychotherapy"] = "Clin Psychol Psychother",
  ["clinical psychology : a publication of the division of clinical psychology of the american psychological association"] = "Clin Psychol (New York)",
  ["clinical psychology review"] = "Clin Psychol Rev",
  ["clinical psychopharmacology and neuroscience"] = "Clin. Psychopharmacol. Neurosci.",
  ["clinical psychopharmacology and neuroscience : the official scientific journal of the korean college of neuropsychopharmacology"] = "Clin Psychopharmacol Neurosci",
  ["clinical pulmonary medicine"] = "Clin Pulm Med",
  ["clinical radiology"] = "Clin Radiol",
  ["clinical radiology & imaging journal"] = "Clin Radiol Imaging J",
  ["clinical rehabilitation"] = "Clin Rehabil",
  ["clinical report. university of cape town. department of obstetrics"] = "Clin Rep",
  ["clinical reports"] = "Clin Rep",
  ["clinical reproduction and fertility"] = "Clin Reprod Fertil",
  ["clinical research"] = "Clin Res",
  ["clinical research and regulatory affairs"] = "Clin Res Regul Aff",
  ["clinical research and trials"] = "Clin Res Trials",
  ["clinical research in cardiology"] = "Clin. Res. Cardiol.",
  ["clinical research in cardiology : official journal of the german cardiac society"] = "Clin Res Cardiol",
  ["clinical research in cardiology supplements"] = "Clin Res Cardiol Suppl",
  ["clinical research in diabetes and endocrinology"] = "Clin Res Diabetes Endocrinol",
  ["clinical research in hiv/aids"] = "Clin Res HIV AIDS",
  ["clinical research in infectious diseases"] = "Clin Res Infect Dis",
  ["clinical research in pulmonology"] = "Clin Res Pulmonol",
  ["clinical research on foot & ankle"] = "Clin Res Foot Ankle",
  ["clinical research practices and drug regulatory affairs"] = "Clin Res Pr Drug Regul Aff",
  ["clinical researcher (alexandria, va.)"] = "Clin Res (Alex)",
  ["clinical resource management"] = "Clin Resour Manag",
  ["clinical reviews in allergy"] = "Clin Rev Allergy",
  ["clinical reviews in allergy & immunology"] = "Clin Rev Allergy Immunol",
  ["clinical reviews in allergy and immunology"] = "Clin. Rev. Allergy Immunol.",
  ["clinical reviews in bone and mineral metabolism"] = "Clin Rev Bone Miner Metab",
  ["clinical rheumatology"] = "Clin Rheumatol",
  ["clinical risk"] = "Clin Risk",
  ["clinical sarcoma research"] = "Clin Sarcoma Res",
  ["clinical schizophrenia & related psychoses"] = "Clin Schizophr Relat Psychoses",
  ["clinical scholars review"] = "Clin Sch Rev",
  ["clinical science"] = "Clin Sci",
  ["clinical science (london, england : 1979)"] = "Clin Sci (Lond)",
  ["clinical science and molecular medicine"] = "Clin Sci Mol Med",
  ["clinical science and molecular medicine. supplement"] = "Clin Sci Mol Med Suppl",
  ["clinical simulation in nursing"] = "Clin Simul Nurs",
  ["clinical skin cancer"] = "Clin Skin Cancer",
  ["clinical social work journal"] = "Clin Soc Work J",
  ["clinical society journal"] = "Clin Soc J",
  ["clinical spectroscopy"] = "Clin. Spectrosc.",
  ["clinical spine surgery"] = "Clin Spine Surg",
  ["clinical symposia"] = "Clin. Symp.",
  ["clinical symposia (summit, n.j. : 1948)"] = "Clin Symp (Summit)",
  ["clinical symposia (summit, n.j. : 1957)"] = "Clin Symp",
  ["clinical techniques in small animal practice"] = "Clin Tech Small Anim Pract",
  ["clinical therapeutics"] = "Clin Ther",
  ["clinical thyroidology"] = "Clin Thyroidol",
  ["clinical toxicology"] = "Clin Toxicol",
  ["clinical toxicology (philadelphia, pa.)"] = "Clin Toxicol (Phila)",
  ["clinical transplantation"] = "Clin Transplant",
  ["clinical transplants"] = "Clin Transpl",
  ["clinical trials"] = "Clin. Trials",
  ["clinical trials (london, england)"] = "Clin Trials",
  ["clinical trials and meta-analysis"] = "Clin Trials Metaanal",
  ["clinical trials and regulatory science in cardiology"] = "Clin Trials Regul Sci Cardiol",
  ["clinical trials in degenerative diseases"] = "Clin Trials Degener Dis",
  ["clinical trials journal"] = "Clin Trials J",
  ["clinical, cosmetic and investigational dentistry"] = "Clin Cosmet Investig Dent",
  ["clinical, cosmetic and investigational dermatology"] = "Clin Cosmet Investig Dermatol",
  ["clinicoeconomics and outcomes research"] = "ClinicoEcon. Outcomes Res.",
  ["clinicoeconomics and outcomes research : ceor"] = "Clinicoecon Outcomes Res",
  ["clinics (sao paulo, brazil)"] = "Clinics",
  ["clinics (são paulo, brazil)"] = "Clinics (Sao Paulo)",
  ["clinics and practice"] = "Clin Pract",
  ["clinics and research in hepatology and gastroenterology"] = "Clin Res Hepatol Gastroenterol",
  ["clinics in chest medicine"] = "Clin Chest Med",
  ["clinics in colon and rectal surgery"] = "Clin Colon Rectal Surg",
  ["clinics in communication disorders"] = "Clin Commun Disord",
  ["clinics in dermatology"] = "Clin Dermatol",
  ["clinics in developmental medicine"] = "Clin Dev Med",
  ["clinics in diagnostic ultrasound"] = "Clin Diagn Ultrasound",
  ["clinics in endocrinology and metabolism"] = "Clin Endocrinol Metab",
  ["clinics in gastroenterology"] = "Clin Gastroenterol",
  ["clinics in geriatric medicine"] = "Clin Geriatr Med",
  ["clinics in haematology"] = "Clin Haematol",
  ["clinics in laboratory medicine"] = "Clin Lab Med",
  ["clinics in liver disease"] = "Clin Liver Dis",
  ["clinics in mother and child health"] = "Clin Mother Child Health",
  ["clinics in obstetrics and gynaecology"] = "Clin Obstet Gynaecol",
  ["clinics in occupational and environmental medicine"] = "Clin Occup Environ Med",
  ["clinics in oncology (belmont, calif.)"] = "Clin Oncol (Belmont)",
  ["clinics in orthopedic surgery"] = "Clin Orthop Surg",
  ["clinics in perinatology"] = "Clin Perinatol",
  ["clinics in plastic surgery"] = "Clin Plast Surg",
  ["clinics in podiatric medicine and surgery"] = "Clin Podiatr Med Surg",
  ["clinics in podiatry"] = "Clin Podiatry",
  ["clinics in rheumatic diseases"] = "Clin Rheum Dis",
  ["clinics in shoulder and elbow"] = "Clin Shoulder Elb",
  ["clinics in sports medicine"] = "Clin Sports Med",
  ["clinics in surgery"] = "Clin Surg",
  ["clinics of oncology"] = "Clin Oncol (Las Vegas)",
  ["clinics of surgery"] = "Clin Surg (Las Vegas)",
  ["clinics of the virginia mason hospital"] = "Clin Va Mason Hosp",
  ["clinique (paris, france)"] = "Clinique (Paris)",
  ["clio (roma)"] = "Clio (Roma)",
  ["clio (toulouse, france)"] = "Clio (Toulouse)",
  ["clio medica"] = "Clio Med.",
  ["clio medica (amsterdam, netherlands)"] = "Clio Med",
  ["clio medica: acta academiae internationalis historiae medicinae"] = "CM",
  ["clio. revista do centro de história da universidade lisboa"] = "Clio",
  ["clínica (valladolid, spain)"] = "Clinica (Valladolid)",
  ["clínica contemporânea; revista de medicina e cirurgia"] = "Clin Contemp",
  ["clínica e investigación en arteriosclerosis : publicación oficial de la sociedad española de arteriosclerosis"] = "Clin Investig Arterioscler",
  ["clínica e investigación en ginecología y obstetricia"] = "Clin Invest Ginecol Obstet",
  ["clínica y salud"] = "Clin Salud",
  ["cllinical and experimental dental research"] = "Clin. Exp. Dent. Res.",
  ["clocks & sleep"] = "Clocks Sleep",
  ["cloning & transgenesis"] = "Cloning Transgenes",
  ["cloning and stem cells"] = "Cloning Stem Cells",
  ["clothing and textiles research journal"] = "Clothing Text. Res. J.",
  ["clujul medical (1957)"] = "Clujul Med",
  ["cluster computing"] = "Cluster Comput",
  ["clustering phenomena in nuclei"] = "Clustering Phenom. Nuclei",
  ["cmaj : canadian medical association journal = journal de l'association medicale canadienne"] = "CMAJ",
  ["cmaj open"] = "CMAJ Open",
  ["cmc-computers materials & continua"] = "CMC-Comput. Mater. Continua",
  ["cmd. current medical dialog"] = "Cmd",
  ["cmd; current medical digest"] = "Cmd Curr Med Dig",
  ["cme (berlin, germany)"] = "CME (Berl)",
  ["cme journal. geriatric medicine"] = "CME J Geriatr Med",
  ["cmes computer modeling in engineering and sciences"] = "CMES-Comp. Model. Eng. Sci.",
  ["cmi lecture series in mathematics"] = "CMI Lect. Ser. Math.",
  ["cms books in mathematics/ouvrages de mathematiques de la smc"] = "CMS Books Math./Ouvrages Math. SMC",
  ["cms conference proceedings"] = "CMS Conf. Proc.",
  ["cms/caims books in mathematics"] = "CMS/CAIMS Books Math.",
  ["cna bulletin"] = "CNA Bull",
  ["cnl nuclear review"] = "CNL Nucl. Rev.",
  ["cns & neurological disorders - drug targets"] = "CNS Neurol. Disord. Drug Targets",
  ["cns & neurological disorders drug targets"] = "CNS Neurol Disord Drug Targets",
  ["cns & neurological disorders: drug targets"] = "CNS Neurol. Disord.: Drug Targets",
  ["cns and neurological disorders drug targets"] = "CNS Neurol. Disord. Drug Targets",
  ["cns drug reviews"] = "CNS Drug Rev",
  ["cns drugs"] = "CNS Drugs",
  ["cns neuroscience & therapeutics"] = "CNS Neurosci Ther",
  ["cns oncology"] = "CNS Oncol",
  ["cns spectrums"] = "CNS Spectr",
  ["co-existence (glasgow, scotland)"] = "Coexistence",
  ["coal preparation (philadelphia, pa, united states)"] = "Coal Prep. (Philadelphia, PA, U. S.)",
  ["coastal and estuarine studies"] = "Coastal Estuarine Stud.",
  ["coastal engineering"] = "Coastal Eng.",
  ["coastal engineering journal"] = "Coastal Eng. J.",
  ["coastal management"] = "Coastal Manage.",
  ["coastal management : an international journal of marine environment, resources, law, and society"] = "Coast Manage",
  ["coastal marine science"] = "Coast Mar Sci",
  ["coatings (basel, switzerland)"] = "Coatings (Basel)",
  ["cochlear implants international"] = "Cochlear Implants Int",
  ["cochrane database of systematic reviews (online)"] = "Cochrane Database Syst Rev",
  ["cochrane evidence synthesis and methods"] = "Cochrane Evidence Synth. Methods",
  ["code of alabama, 1975 : with provision for subsequent pocket parts. alabama"] = "Code Ala 1975 Ala",
  ["code of federal regulations"] = "Code Fed Regul",
  ["code of federal regulations. 45: public welfare"] = "Code Fed Regul Public Welfare",
  ["code of federal regulations. 46, shipping"] = "Code Fed Regul Shipping",
  ["code of georgia annotated : including code of 1981; including the code of 1933, the code of 1981 and all laws of general application subsequently enacted, with editorial notes and complete annotations of the decisions of the supreme court and court of appeals of georgia and of all the federal courts construing georgia laws. georgia"] = "Code Ga Annot Ga",
  ["code of laws of south carolina, 1976, annotated : containing permanent public statutes of general application to the end of the 1975 legislative session : prepared under the supervision of the legislative council and code commissioner and adopted as the official code of the state by the 1977 session of the general assembly. south carolina"] = "Code Laws S C 1976 Annot S C",
  ["code of virginia, 1950 : with provisions for subsequent pocket parts. annotated / prepared under the supervision of the virginia code commission by the editorial staff of the publishers. virginia"] = "Code Va 1950 Va",
  ["code of virginia, 1950 : with provisions for subsequent pocket parts. annotated. virginia"] = "Code Va 1950 Va",
  ["codesign : international journal of cocreation in design and the arts"] = "CoDesign",
  ["codices manuscripti: zeitschrift für handschriftenkunde"] = "CodMan",
  ["coeur et medecine interne"] = "Coeur Med. Interne",
  ["coeur et médecine interne"] = "Coeur Med Interne",
  ["coeur; revue de cardiologie medico-chirurgicale"] = "Coeur",
  ["coeur; revue de cardiologie médico-chirurgicale"] = "Coeur",
  ["coffee science"] = "Coffee Sci",
  ["cogeneration & distributed generation journal"] = "Cogener. Distrib. Gener. J.",
  ["cogent arts & humanities"] = "Cogent Arts Humanit",
  ["cogent biology"] = "Cogent Biol",
  ["cogent chemistry"] = "Cogent Chem",
  ["cogent engineering"] = "Cogent Eng",
  ["cogent environmental science"] = "Cogent Environ Sci",
  ["cogent food & agriculture"] = "Cogent Food Agric",
  ["cogent geoscience"] = "Cogent Geosci",
  ["cogent geosciences"] = "Cogent Geosci.",
  ["cogent mathematics & statisics"] = "Cogent Math. Stat.",
  ["cogent mathematics & statistics"] = "Cogent Math Stat",
  ["cogent medicine"] = "Cogent Med",
  ["cogent physics"] = "Cogent Phys.",
  ["cogent psychology"] = "Cogent Psychol",
  ["cogent social science"] = "Cogent Social Sci.",
  ["cogent social sciences"] = "Cogent Soc Sci",
  ["cognition & emotion"] = "Cogn Emot",
  ["cognition and instruction"] = "Cogn Instr",
  ["cognition, brain, behavior : an interdisciplinary journal"] = "Cogn Brain Behav",
  ["cognition, technology & work"] = "Cognit. Technol. Work",
  ["cognition, technology & work (online)"] = "Cogn Technol Work",
  ["cognitive and behavioral neurology"] = "Cogn. Behav. Neurol.",
  ["cognitive and behavioral neurology : official journal of the society for behavioral and cognitive neurology"] = "Cogn Behav Neurol",
  ["cognitive and behavioral practice"] = "Cogn Behav Pract",
  ["cognitive behaviour therapist"] = "Cogn Behav Therap",
  ["cognitive behaviour therapy"] = "Cogn Behav Ther",
  ["cognitive computation"] = "Cognit. Comput.",
  ["cognitive computation and systems"] = "Cognit. Comput. Syst.",
  ["cognitive development"] = "Cogn Dev",
  ["cognitive intelligence and robotics"] = "Cogn. Intell. Robot.",
  ["cognitive linguistics"] = "Cogn Linguist",
  ["cognitive neurodynamics"] = "Cogn Neurodyn",
  ["cognitive neuropsychiatry"] = "Cogn Neuropsychiatry",
  ["cognitive neuropsychology"] = "Cogn Neuropsychol",
  ["cognitive neuroscience"] = "Cogn Neurosci",
  ["cognitive processing"] = "Cogn Process",
  ["cognitive psychology"] = "Cogn Psychol",
  ["cognitive research: principles and implications"] = "Cogn Res Princ Implic",
  ["cognitive robotics"] = "Cognit. Rob.",
  ["cognitive science"] = "Cogn Sci",
  ["cognitive science and technology"] = "Cogn. Sci. Technol.",
  ["cognitive sciences"] = "Cogn Sci (Hauppauge)",
  ["cognitive systems monographs"] = "Cogn. Syst. Monogr.",
  ["cognitive systems research"] = "Cogn Syst Res",
  ["cognitive therapy and research"] = "Cognit Ther Res",
  ["cognitive, affective & behavioral neuroscience"] = "Cogn Affect Behav Neurosci",
  ["cognitive, affective and behavioral neuroscience"] = "Cogn. Affect. Behav. Neurosci.",
  ["coimbra mathematical texts"] = "Coimbra Math. Texts",
  ["coimbra medica"] = "Coimbra Med",
  ["coimbra médica"] = "Coimbra Med",
  ["coin hoards. the royal numismatic society, london"] = "CoinHoards",
  ["coin world"] = "Coin World",
  ["coke and chemistry"] = "Coke Chem.",
  ["colby quarterly"] = "ColbyQ",
  ["cold regions science and technology"] = "Cold Reg. Sci. Technol.",
  ["cold spring harbor molecular case studies"] = "Cold Spring Harb Mol Case Stud",
  ["cold spring harbor perspectives in biology"] = "Cold Spring Harb Perspect Biol",
  ["cold spring harbor perspectives in medicine"] = "Cold Spring Harb Perspect Med",
  ["cold spring harbor protocols"] = "Cold Spring Harb Protoc",
  ["cold spring harbor symposia on quantitative biology"] = "Cold Spring Harb Symp Quant Biol",
  ["cold war history"] = "Cold War Hist",
  ["colecao matematica e aplicacoes"] = "Colec. Mat. Apl.",
  ["colecao matematica universitaria"] = "Colec. Mat. Univ.",
  ["coleccion estudio"] = "Colecc. Estud.",
  ["coleccion estudios cieplan"] = "Colecc Estud CIEPLAN",
  ["coleccion historica de ciencias de la salud"] = "Colecc Hist Cienc Salud",
  ["coleccion julio carrizosa valenzuela"] = "Colecc. Julio Carrizosa Valenzuela",
  ["coleccion obra selecta"] = "Colecc. Obra Sel.",
  ["colección estudios cieplan"] = "Colecc Estud CIEPLAN",
  ["colección enrique pérez arbeláez"] = "Colec. Enrique Pérez Arbeláez",
  ["colección estudios cieplan"] = "Colección Estud. CIEPLAN",
  ["colección histórica de ciencias de la salud"] = "Colecc Hist Cienc Salud",
  ["colección julio carrizosa valenzuela"] = "Colecc. Julio Carrizosa Valenzuela",
  ["colección memorias"] = "Colec. Mem.",
  ["colección: congresos"] = "Colecc. Congr.",
  ["colectia universitaria (cluj-napoca)"] = "Colect. Univ. (Cluj-Napoca)",
  ["colecţia universitaria"] = "Colecţ. Univ.",
  ["coleopterists bulletin"] = "Coleopterists Bull.",
  ["coleção cle"] = "Coleç. CLE",
  ["collaborations (coral gables, fla.)"] = "Collaborations (Coral Gables)",
  ["collaborative anthropologies"] = "Collab Anthropol",
  ["collabra. psychology"] = "Collabra Psychol",
  ["collagen and related research"] = "Coll Relat Res",
  ["collana di scienze matematiche"] = "Coll. Sci. Mat.",
  ["collana di studi"] = "Coll. Studi",
  ["collana monografica. gazzetta medica siciliana"] = "Collana Monogr Gazz Med Sicil",
  ["collectanea botánica"] = "Collect Bot",
  ["collectanea mathematica"] = "Collect. Math.",
  ["collected papers of the mayo clinic and the mayo foundation. mayo clinic"] = "Collect Papers Mayo Clinic Mayo Found",
  ["collected studies series"] = "Collect. Studies Ser.",
  ["collected works"] = "Collect. Works",
  ["collected works on cardio-pulmonary disease"] = "Coll Works Cardiopulm Dis",
  ["collected works on cardiopulmonary disease"] = "Coll. Works Cardiopulm. Dis.",
  ["collection aléa-saclay: monographs and texts in statistical physics"] = "Collect. Aléa-Saclay Monogr. Texts Statist. Phys.",
  ["collection automatique de base"] = "Collect. Autom. Base",
  ["collection blain faye martin"] = "Collect. Blain Faye Martin",
  ["collection comptes rendus. géoscience"] = "Collect C R Geosci",
  ["collection d'études classiques"] = "Collect Etud Class",
  ["collection de l’i.m.e."] = "Collect. I.M.E.",
  ["collection de l’école française de rome"] = "CEFR",
  ["collection de notes internes de la direction des etudes et recherches: mathematiques, informatique, telecommunications"] = "Collect. Notes Internes Dir. Etudes Rech. Math. Inform. Telecomm.",
  ["collection de travaux de l’académie internationale d’histoire des sciences"] = "Collect. Trav. Acad. Int. Hist. Sci.",
  ["collection des industries mecaniques"] = "Collect. Indust. Mec.",
  ["collection des universités de france"] = "Collect. Univ. France",
  ["collection droit de l’administration locale"] = "Collect. Droit Adm. Locale",
  ["collection essais"] = "Collect. Essais",
  ["collection histoire de la pensée"] = "Collect. Hist. Pensée",
  ["collection idees"] = "Collect. Idees",
  ["collection info sup"] = "Collect. Info Sup",
  ["collection informatique"] = "Collect. Inform.",
  ["collection informatique prepas-universite"] = "Collect. Inform. Prepas-Univ.",
  ["collection meta"] = "Collect. META",
  ["collection méthodes"] = "Collect. Méthodes",
  ["collection nabla"] = "Coll. NABLA",
  ["collection of czechoslovak chemical communications"] = "Collect. Czech. Chem. Commun.",
  ["collection points: serie sciences"] = "Collect. Points Ser. Sci.",
  ["collection regards sur la science"] = "Collect. Regards Sci.",
  ["collection sciences dans l’histoire"] = "Collect. Sci. Hist.",
  ["collection see"] = "Collect. SEE",
  ["collection seminaires inria"] = "Collect. Sem. INRIA",
  ["collection visions des sciences"] = "Collect. Vis. Sci.",
  ["collection études"] = "Collect. études",
  ["collection “didactiques”. mathematiques"] = "Collect. Didact. Math.",
  ["collections : the newsletter of the archives and special collections on women in medicine, the medical college of pennsylvania"] = "Collections",
  ["collections of the nova scotia historical society"] = "Collect N S Hist Soc",
  ["collective dynamics"] = "Collect Dyn",
  ["college and research libraries"] = "Coll Res Libr",
  ["college and university : the journal of the american association of collegiate registrars"] = "Coll Univ",
  ["college mathematics journal"] = "Coll. Math. J.",
  ["college review"] = "Coll. Rev.",
  ["college review (denver, colo.)"] = "Coll Rev",
  ["college student journal"] = "Coll Stud J",
  ["collegian (royal college of nursing, australia)"] = "Collegian",
  ["collegium antropologicum"] = "Coll Antropol",
  ["collegium logicum"] = "Coll. Logicum Ann. Kurt-Gödel-Soc.",
  ["colloid and interface science communications"] = "Colloid Interface Sci. Commun.",
  ["colloid and polymer science"] = "Colloid. Polym. Sci.",
  ["colloid journal"] = "Colloid J.",
  ["colloids and interface science communications"] = "Colloids Interface Sci. Commun.",
  ["colloids and surfaces"] = "Colloids Surf.",
  ["colloids and surfaces a: physicochemical and engineering aspects"] = "Colloids Surf., A",
  ["colloids and surfaces b: biointerfaces"] = "Colloids Surf., B",
  ["colloids and surfaces, a: physicochemical and engineering aspects"] = "Colloids Surf., A",
  ["colloids and surfaces, b: biointerfaces"] = "Colloids Surf., B",
  ["colloids and surfaces. a, physicochemical and engineering aspects"] = "Colloids Surf A Physicochem Eng Asp",
  ["colloids and surfaces. b, biointerfaces"] = "Colloids Surf B Biointerfaces",
  ["colloques et sʹeminaires - institut franc̦ais de recherche scientifique pour le dʹeveloppement en coopʹeration"] = "Colloq Semin Inst Fr Rech Sci Dev Coop",
  ["colloques internationaux. centre national de la recherche scientifique (france)"] = "Colloq Int CNRS",
  ["colloques nationaux du centre national de la recherche scientifique"] = "Colloq. Nat. CNRS",
  ["colloques nationaux. centre national de la recherche scientifique (france)"] = "Colloq Natx Cent Natl Rech Sci",
  ["colloqui del sodalizio"] = "ColloquiSod",
  ["colloquia mathematica societatis jános bolyai"] = "Colloq. Math. Soc. János Bolyai",
  ["colloquium mathematicum"] = "Colloq. Math.",
  ["colombia medica"] = "Colomb Med",
  ["colombia medica (cali, colombia)"] = "Colomb Med",
  ["colombia médica"] = "Colomb Med",
  ["colombia médica (cali, colombia)"] = "Colomb Med (Cali)",
  ["colombian applied linguistics journal"] = "Colomb Appl Linguist J",
  ["colonial latin american historical review : clahr"] = "Coln Latin Am Hist Rev",
  ["colonial latin american review"] = "Colon latin Am Rev",
  ["colonial plant and animal products"] = "Colon. Plant Anim. Prod.",
  ["color research & application"] = "Color Res. Appl.",
  ["color research and application"] = "Color Res Appl",
  ["colorado heritage : the journal of the colorado historical society"] = "Colo Herit",
  ["colorado journal of international environmental law and policy"] = "Colo J Int Environ Law Policy",
  ["colorado lawyer"] = "Colo Lawyer",
  ["colorado medicine"] = "Colo Med",
  ["colorado nurse"] = "Colo Nurse",
  ["colorado nurse (1985)"] = "Colo Nurse",
  ["colorado quarterly"] = "Colo Q",
  ["colorado revised statutes, 1973 / completely annotated and indexed by the editorial staff of the michie company. colorado"] = "Colo Revis Statut 1973 Colo",
  ["colorado revised statutes, 1973. colorado"] = "Colo Revis Statut 1973 Colo",
  ["coloration technology"] = "Color. Technol.",
  ["colorectal cancer"] = "Colorectal Cancer",
  ["colorectal disease : the official journal of the association of coloproctology of great britain and ireland"] = "Colorectal Dis",
  ["columbia forum"] = "Columbia Forum",
  ["columbia human rights law review"] = "Columbia Human Rights Law Rev",
  ["columbia journal of gender and law"] = "Columbia J Gend Law",
  ["columbia journal of law and social problems"] = "Columbia J Law Soc Probl",
  ["columbia journal of world business"] = "Columbia J. World Bus.",
  ["columbia journalism review"] = "Columbia J Rev",
  ["columbia law review"] = "Columbia Law Rev",
  ["columbus dental society bulletin"] = "Columbus Dent. Soc. Bull.",
  ["combinatorial chemistry  -  an online journal"] = "Comb. Chem.",
  ["combinatorial chemistry & high throughput screening"] = "Comb Chem High Throughput Screen",
  ["combinatorial chemistry - an online journal"] = "Comb. Chem.",
  ["combinatorial chemistry and high throughput screening"] = "Comb. Chem. High Throughput Screening",
  ["combinatorial optimization"] = "Comb. Optim.",
  ["combinatorial theory"] = "Comb. Theory",
  ["combinatorica. an international journal on combinatorics and the theory of computing"] = "Combinatorica",
  ["combinatorics and number theory"] = "Comb. Number Theory",
  ["combinatorics, probability & computing : cpc"] = "Comb Probab Comput",
  ["combinatorics, probability and computing"] = "Comb. Probab. Comput.",
  ["combustion and flame"] = "Combust. Flame",
  ["combustion explosion and shock waves"] = "Combust. Explos. Shock Waves",
  ["combustion science and technology"] = "Combust. Sci. Technol.",
  ["combustion science and technology : cst"] = "Combust Sci Technol",
  ["combustion theory and modelling"] = "Combust. Theor. Model.",
  ["combustion, explosion, and shock waves"] = "Combust Explos Shock Waves",
  ["comercio exterior"] = "Comercio Exterior",
  ["comertul modern"] = "Comertul Mod.",
  ["commagene journal of biology"] = "Commagene J Biol",
  ["command historical report. naval medical research institute (u.s.)"] = "Command Hist Rep",
  ["comment, a research/action report on wo/men"] = "Comment",
  ["commentarii mathematici helvetici"] = "Comment. Math. Helv.",
  ["commentarii mathematici helvetici. a journal of the swiss mathematical society"] = "Comment. Math. Helv.",
  ["commentarii mathematici universitatis sancti pauli"] = "Comment. Math. Univ. St. Pauli",
  ["commentarii. pontificia accademia delle scienze. plenary session"] = "Commentarii",
  ["commentary (new york, n.y.)"] = "Commentary",
  ["commentationes mathematicae universitatis carolinae"] = "Comment. Math. Univ. Carolin.",
  ["commentationes mathematicae. annals of the polish mathematical society. series i"] = "Comment. Math.",
  ["comments on astrophysics"] = "Comments Astrophys.",
  ["comments on astrophysics and space physics"] = "Comments Astrophys. Space Phys.",
  ["comments on atomic and molecular physics"] = "Comments At. Mol. Phys.",
  ["comments on condensed matter physics"] = "Comments Condens. Matter Phys.",
  ["comments on inorganic chemistry"] = "Comments Inorg. Chem.",
  ["comments on modern chemistry. part a, comments on inorganic chemistry : a journal of critical discussion of the current literature"] = "Comments Mod Chem A Comments Inorg Chem",
  ["comments on modern physics"] = "Comments Mod. Phys.",
  ["comments on nuclear and particle physics"] = "Comments Nucl. Part. Phys.",
  ["comments on plasma physics and controlled fusion"] = "Comments Plasma Phys. Controlled Fusion",
  ["comments on solid state physics"] = "Comments Solid State Phys.",
  ["commercial law journal"] = "Commer Law J",
  ["commercial laws of europe"] = "Commer Law Eur",
  ["common factor (stoughton, mass.)"] = "Common Factor",
  ["common law world review"] = "Common Law World Rev",
  ["common sense pest control quarterly"] = "Common Sense Pest Control Q.",
  ["commonhealth (philadelphia, pa.)"] = "Commonhealth (Phila)",
  ["commonweal (new york, n.y.)"] = "Commonweal",
  ["commonwealth & comparative politics"] = "Commonw Comp Polit",
  ["commonwealth forestry review, the"] = "Commonw. For. Rev.",
  ["communal societies"] = "Communal Soc",
  ["communicable disease and public health"] = "Commun Dis Public Health",
  ["communicable disease and public health / phls"] = "Commun Dis Public Health",
  ["communicable disease report. cdr review"] = "Commun Dis Rep CDR Rev",
  ["communicable disease report. cdr supplement"] = "Commun Dis Rep CDR Suppl",
  ["communicable disease report. cdr weekly"] = "Commun Dis Rep CDR Wkly",
  ["communicable diseases intelligence"] = "Commun Dis Intell",
  ["communicable diseases intelligence (2018)"] = "Commun Dis Intell (2018)",
  ["communicable diseases intelligence quarterly report"] = "Commun Dis Intell Q Rep",
  ["communicating nursing research"] = "Commun Nurs Res",
  ["communication & cognition monographies"] = "Comm. Cognition Monogr.",
  ["communication & medicine"] = "Commun Med",
  ["communication and cognition"] = "Comm. Cogn.",
  ["communication and medicine"] = "Commun. Med.",
  ["communication disorders quarterly"] = "Commun Disord Q",
  ["communication education"] = "Commun Educ",
  ["communication methods and measures"] = "Commun Methods Meas",
  ["communication monographs"] = "Commun Monogr",
  ["communication quarterly"] = "Commun Q",
  ["communication research"] = "Communic Res",
  ["communication research reports : crr"] = "Commun Res Rep",
  ["communication sciences & disorders (seoul, korea)"] = "Commun Sci Disord",
  ["communication studies"] = "Commun Stud",
  ["communication theory : ct : a journal of the international communication association"] = "Commun Theory",
  ["communication yearbook"] = "Commun Yearb",
  ["communication, culture & critique"] = "Commun Cult Crit",
  ["communicationes archaeologicae hungaricae"] = "CommunicAHung",
  ["communicationes de historia artis medicinae. suppl"] = "Commun Hist Artis Med Suppl",
  ["communicationes instituti forestalis fenniae"] = "Commun. Inst. For. Fenn.",
  ["communications and control engineering series"] = "Comm. Control Engrg. Ser.",
  ["communications and strategies"] = "Communications Strategies",
  ["communications biology"] = "Commun Biol",
  ["communications chemistry"] = "Commun Chem",
  ["communications earth & environment"] = "Commun Earth Environ",
  ["communications engineering"] = "Commun. Eng.",
  ["communications for statistical applications and methods"] = "Commun Stat Appl Methods",
  ["communications in agricultural and applied biological sciences"] = "Commun. Agric. Appl. Biol. Sci.",
  ["communications in algebra"] = "Comm. Algebra",
  ["communications in analysis and geometry"] = "Comm. Anal. Geom.",
  ["communications in analysis and geomtry"] = "Commun. Anal. Geom.",
  ["communications in analysis and mechanics"] = "Commun. Anal. Mech.",
  ["communications in applied analysis"] = "Commun Appl Anal",
  ["communications in applied and industrial mathematics"] = "Commun. Appl. Ind. Math.",
  ["communications in applied mathematics and computational science"] = "Commun. Appl. Math. Comput. Sci.",
  ["communications in behavioral biology. part a, original articles"] = "Commun. Behav. Biol.",
  ["communications in behavioral biology. part a: [original articles]"] = "Commun Behav Biol",
  ["communications in combinatorics and optimization"] = "Commun. Comb. Optim.",
  ["communications in computational physics"] = "Commun. Comput. Phys.",
  ["communications in computer and information science"] = "Commun. Comput. Inf. Sci.",
  ["communications in contemporary mathematics"] = "Commun. Contemp. Math.",
  ["communications in cybernetics, systems science and engineering"] = "Commun. Cybern. Syst. Sci. Eng.",
  ["communications in information & systems"] = "Commun. Inf. Syst.",
  ["communications in information and systems"] = "Commun Inf Syst",
  ["communications in mathematical analysis"] = "Commun. Math. Anal.",
  ["communications in mathematical analysis and applications"] = "Commun. Math. Anal. Appl.",
  ["communications in mathematical physics"] = "Commun. Math. Phys.",
  ["communications in mathematical research"] = "Commun. Math. Res.",
  ["communications in mathematical sciences"] = "Commun. Math. Sci.",
  ["communications in mathematics"] = "Commun. Math.",
  ["communications in mathematics and statistics"] = "Commun. Math. Stat.",
  ["communications in nonlinear science & numerical simulation"] = "Commun Nonlinear Sci Numer Simul",
  ["communications in nonlinear science and numerical simulation"] = "Commun. Nonlinear Sci. Numer. Simul.",
  ["communications in number theory and physics"] = "Commun. Number Theory Phys.",
  ["communications in numerical analysis"] = "Commun. Numer. Anal.",
  ["communications in numerical methods in engineering"] = "Commun. Numer. Methods Eng.",
  ["communications in partial differential equations"] = "Comm. Partial Differential Equations",
  ["communications in psychopharmacology"] = "Commun Psychopharmacol",
  ["communications in soil science and plant analysis"] = "Commun. Soil Sci. Plant Anal.",
  ["communications in statistics"] = "Comm. Statist. Simulation Comput.",
  ["communications in statistics - simulation and computation"] = "Commun. Stat.- Simul. Comput.",
  ["communications in statistics - theory and methods"] = "Commun. Stat.- Theory Methods",
  ["communications in statistics. case studies, data analysis and applications"] = "Commun Stat Case Stud Data Anal Appl",
  ["communications in statistics. simulation and computation"] = "Comm. Statist. Simulation Comput.",
  ["communications in statistics. theory and methods"] = "Comm. Statist. Theory Methods",
  ["communications in statistics: case studies, data analysis and applications"] = "Commun. Stat.: Case Stud. Data Anal. Appl.",
  ["communications in statistics: simulation and computation"] = "Commun Stat Simul Comput",
  ["communications in statistics: theory and methods"] = "Commun Stat Theory Methods",
  ["communications in the analytic theory of continued fractions"] = "Comm. Anal. Theory Contin. Fractions",
  ["communications in theoretical physics"] = "Commun. Theor. Phys.",
  ["communications in theoretical physics (allahabad)"] = "Comm. Theoret. Phys. (Allahabad)",
  ["communications in transportation research"] = "Commun. Transp. Res.",
  ["communications materials"] = "Commun Mater",
  ["communications medicine"] = "Commun. Med.",
  ["communications of the acm"] = "Commun ACM",
  ["communications of the american mathematical society"] = "Comm. Amer. Math. Soc.",
  ["communications on analysis and computation"] = "Commun. Anal. Comput.",
  ["communications on applied mathematics and computation"] = "Commun. Appl Math. Comput.",
  ["communications on applied nonlinear analysis"] = "Comm. Appl. Nonlinear Anal.",
  ["communications on pure and applied analysis"] = "Commun. Pure Appl. Anal.",
  ["communications on pure and applied mathematics"] = "Commun. Pure Appl. Math.",
  ["communications on stochastic analysis"] = "Commun. Stoch. Anal.",
  ["communications physics"] = "Commun Phys",
  ["communications surveys and tutorials"] = "Commun. Surveys Tuts.",
  ["communications. faculty of sciences. university of ankara. series a1. mathematics and statistics"] = "Commun. Fac. Sci. Univ. Ank. Ser. A1. Math. Stat.",
  ["communicative & integrative biology"] = "Commun Integr Biol",
  ["communicative disorders review"] = "Commun Disord Rev",
  ["communio (spokane, wash.)"] = "Communio",
  ["communique (milwaukee, wis.)"] = "Communique",
  ["communique / american nurses association, center for ethics and human rights"] = "Communique (Wash DC)",
  ["communique newsletter"] = "Commun Newsl",
  ["communiqué (milwaukee, wis.)"] = "Communique",
  ["communist and post-communist studies"] = "Communist Post-Communist Stud",
  ["communist economies"] = "Communist Econ",
  ["communist economies and economic transformation"] = "Communist Econ Econ Transform",
  ["community and junior college journal"] = "Community Jr Coll J",
  ["community college journal of research and practice"] = "Community Coll J Res Pract",
  ["community college review"] = "Community Coll Rev",
  ["community dental health"] = "Community Dent Health",
  ["community dental health (houndsmills, hampshire"] = "Community Dent Health",
  ["community dentistry and oral epidemiology"] = "Community Dent Oral Epidemiol",
  ["community development (columbus, ohio)"] = "Community Dev (Columb)",
  ["community development journal"] = "Community Dev J",
  ["community ecology"] = "Community Ecol.",
  ["community ecology : ce"] = "Community Ecol",
  ["community eye health"] = "Community Eye Health",
  ["community eye health / international centre for eye health"] = "Community Eye Health",
  ["community genetics"] = "Community Genet",
  ["community health"] = "Community Health (Bristol)",
  ["community health (bristol, england)"] = "Community Health (Bristol)",
  ["community health studies"] = "Community Health Stud",
  ["community literacy journal"] = "Community Lit J",
  ["community medicine"] = "Community Med",
  ["community mental health journal"] = "Community Ment Health J",
  ["community mental health review"] = "Community Ment Health Rev",
  ["community nurse"] = "Community Nurse",
  ["community nursing"] = "Community Nurs",
  ["community oncology"] = "Community Oncol",
  ["community outlook"] = "Community Outlook",
  ["community practitioner : the journal of the community practitioners' & health visitors' association"] = "Community Pract",
  ["community practitioner : the journal of the community practitioners’ & health visitors’ association"] = "Community Pract",
  ["community psychologist"] = "Community Psychol",
  ["community psychology in global perspective"] = "Community Psychol Glob Perspect",
  ["community, work & family"] = "Community Work Fam",
  ["community-based public health policy & practice"] = "Community Based Public Health Policy Pract",
  ["community-based public health policy & practice / partnership for the public’s health"] = "Community Based Public Health Policy Pract",
  ["compact textbooks in mathematics"] = "Compact Textb. Math.",
  ["companion animal"] = "Companion Anim",
  ["comparativ : leipziger beiträge zur universalgeschichte und vergleichenden gesellschaftsforschung"] = "Comparativ (Leipz)",
  ["comparative and functional genomics"] = "Comp Funct Genomics",
  ["comparative and general pharmacology"] = "Comp Gen Pharmacol",
  ["comparative biochemistry and physiology"] = "Comp Biochem Physiol",
  ["comparative biochemistry and physiology part a: molecular & integrative physiology"] = "Comp. Biochem. Physiol. A: Mol. Integr. Physiol.",
  ["comparative biochemistry and physiology part a: physiology"] = "Comp. Biochem. Physiol. A: Physiol.",
  ["comparative biochemistry and physiology part b: biochemistry and molecular biology"] = "Comp. Biochem. Physiol. B: Biochem. Mol. Biol.",
  ["comparative biochemistry and physiology part b: comparative biochemistry"] = "Comp. Biochem. Physiol. B: Comp. Biochem.",
  ["comparative biochemistry and physiology part c: pharmacology toxicology and endocrinology"] = "Comp. Biochem. Physiol. C: Pharmacol. Toxicol. Endocrinol.",
  ["comparative biochemistry and physiology part c: toxicology & pharmacology"] = "Comp. Biochem. Physiol. C: Toxicol. Pharmacol.",
  ["comparative biochemistry and physiology part d: genomics and proteomics"] = "Comp. Biochem. Physiol. D: Genomics Proteomics",
  ["comparative biochemistry and physiology, a: comparative physiology"] = "Comp. Biochem. Physiol., A: Comp. Physiol.",
  ["comparative biochemistry and physiology, b: comparative biochemistry"] = "Comp. Biochem. Physiol., B: Comp. Biochem.",
  ["comparative biochemistry and physiology, c: comparative pharmacology and toxicology"] = "Comp. Biochem. Physiol., C: Comp. Pharmacol. Toxicol.",
  ["comparative biochemistry and physiology, part a: molecular & integrative physiology"] = "Comp. Biochem. Physiol., Part A: Mol. Integr. Physiol.",
  ["comparative biochemistry and physiology, part b: biochemistry & molecular biology"] = "Comp. Biochem. Physiol., Part B: Biochem. Mol. Biol.",
  ["comparative biochemistry and physiology, part c: toxicology & pharmacology"] = "Comp. Biochem. Physiol., Part C: Toxicol. Pharmacol.",
  ["comparative biochemistry and physiology, part d: genomics & proteomics"] = "Comp. Biochem. Physiol., Part D: Genomics Proteomics",
  ["comparative biochemistry and physiology. a, comparative physiology"] = "Comp Biochem Physiol A Comp Physiol",
  ["comparative biochemistry and physiology. b, comparative biochemistry"] = "Comp Biochem Physiol B",
  ["comparative biochemistry and physiology. biochemistry and molecular biology"] = "Comp Biochem Physiol Biochem Mol Biol",
  ["comparative biochemistry and physiology. c, comparative pharmacology"] = "Comp. Biochem. Physiol. C",
  ["comparative biochemistry and physiology. c, comparative pharmacology and toxicology"] = "Comp Biochem Physiol C Comp Pharmacol Toxicol",
  ["comparative biochemistry and physiology. c, comparative pharmacology and toxicology."] = "Comp. Biochem. Physiol. C",
  ["comparative biochemistry and physiology. c: comparative pharmacology"] = "Comp Biochem Physiol C Comp Pharmacol",
  ["comparative biochemistry and physiology. comparative physiology"] = "Comp Biochem Physiol Comp Physiol",
  ["comparative biochemistry and physiology. part a, molecular & integrative physiology"] = "Comp Biochem Physiol A Mol Integr Physiol",
  ["comparative biochemistry and physiology. part a, molecular and integrative physiology"] = "Comp. Biochem. Physiol. A. Mol. Integr. Physiol.",
  ["comparative biochemistry and physiology. part a, physiology"] = "Comp Biochem Physiol A Physiol",
  ["comparative biochemistry and physiology. part b, biochemistry & molecular biology"] = "Comp Biochem Physiol B Biochem Mol Biol",
  ["comparative biochemistry and physiology. part b, biochemistry and molecular biology"] = "Comp. Biochem. Physiol. B Biochem. Mol. Biol.",
  ["comparative biochemistry and physiology. part c, pharmacology, toxicology & endocrinology"] = "Comp Biochem Physiol C Pharmacol Toxicol Endocrinol",
  ["comparative biochemistry and physiology. part c, pharmacology, toxicology and endocrinology"] = "Comp. Biochem. Physiol. C Pharmacol. Toxicol. Endocrinol.",
  ["comparative biochemistry and physiology. part d, genomics & proteomics"] = "Comp Biochem Physiol Part D Genomics Proteomics",
  ["comparative biochemistry and physiology. pharmacology, toxicology and endocrinology"] = "Comp Biochem Physiol Pharmacol Toxicol Endocrinol",
  ["comparative biochemistry and physiology. physiology"] = "Comp Biochem Physiol Physiol",
  ["comparative biochemistry and physiology. toxicology & pharmacology : cbp"] = "Comp Biochem Physiol C Toxicol Pharmacol",
  ["comparative biochemistry and physiology. toxicology and pharmacology"] = "Comp. Biochem. Physiol. C Toxicol. Pharmacol.",
  ["comparative clinical pathology"] = "Comp Clin Path",
  ["comparative cognition & behavior reviews"] = "Comp Cogn Behav Rev",
  ["comparative cognition and neuroscience"] = "Comp. Cogn. Neurosci.",
  ["comparative cytogenetics"] = "Comp Cytogenet",
  ["comparative economic studies"] = "Comp Econ Stud",
  ["comparative education"] = "Comp Educ",
  ["comparative education review"] = "Comp Educ Rev",
  ["comparative effectiveness research (auckland, n.z.)"] = "Comp Eff Res (Auckl)",
  ["comparative exercise physiology"] = "Comp Exerc Physiol",
  ["comparative hepatology"] = "Comp Hepatol",
  ["comparative immunology, microbiology and infectious diseases"] = "Comp Immunol Microbiol Infect Dis",
  ["comparative labor law journal : a publication of the u.s. national branch of the international society for labor law and social security [and] the wharton school, and the law school of the university of pennsylvania"] = "Comp Labor Law J",
  ["comparative law"] = "Nihon Univ Comp Law",
  ["comparative literature"] = "CompLit",
  ["comparative literature studies"] = "CLS",
  ["comparative medicine"] = "Comp Med",
  ["comparative medicine east and west"] = "Comp Med East West",
  ["comparative migration studies"] = "Comp Migr Stud",
  ["comparative parasitology"] = "Comp Parasitol",
  ["comparative political studies"] = "Comp Polit Stud",
  ["comparative politics"] = "Comp Polit",
  ["comparative social research"] = "Comp Soc Res",
  ["comparative studies in society and history"] = "Comp Stud Soc Hist",
  ["comparative urban research"] = "Comp Urban Res",
  ["compendium (newtown, pa.)"] = "Compendium",
  ["compendium (newtown, pa.). supplement"] = "Compend Suppl",
  ["compendium (yardley, pa)"] = "Compend Contin Educ Vet",
  ["compendium of continuing education in dentistry"] = "Compend. Contin. Educ. Dent.",
  ["compendium of continuing education in dentistry (jamesburg, n.j. : 1995)"] = "Compend Contin Educ Dent",
  ["compendium of continuing education in dentistry. (jamesburg, n.j. : 1995). supplement"] = "Compend Contin Educ Dent Suppl",
  ["compendium. supplement"] = "Compend. Suppl.",
  ["compensation and benefits review"] = "Compens Benefits Rev",
  ["compensation medicine"] = "Compens Med",
  ["compensation review"] = "Compens Rev",
  ["competition & change"] = "Compet Change",
  ["complement (basel, switzerland)"] = "Complement",
  ["complement and inflammation"] = "Complement Inflamm",
  ["complementary health practice review"] = "Complement Health Pract Rev",
  ["complementary medicine research"] = "Complement Med Res",
  ["complementary therapies in clinical practice"] = "Complement Ther Clin Pract",
  ["complementary therapies in medicine"] = "Complement Ther Med",
  ["complementary therapies in nursing & midwifery"] = "Complement Ther Nurs Midwifery",
  ["complementary therapies in nursing and midwifery"] = "Complement. Ther. Nurs. Midwifery",
  ["complementi di matematica per l’indirizzo didàttico"] = "Complementi Mat. Indirizzo Didàtt.",
  ["complex & intelligent systems"] = "Complex Intell. Syst.",
  ["complex adaptive systems"] = "Complex Adapt. Systems",
  ["complex adaptive systems modeling"] = "Complex Adapt. Syst. Model.",
  ["complex analysis and its synergies"] = "Complex Anal. Synergies",
  ["complex analysis and operator theory"] = "Complex Anal. Oper. Theory",
  ["complex metals"] = "Complex Met.",
  ["complex networks and dynamic systems"] = "Complex Netw. Dyn. Syst.",
  ["complex networks and their applications vii : proceedings the 7th international conference on complex networks and their applications complex networks 2018. international conference on complex networks and their applications (7th : 2018 : cambridge, england)"] = "Complex Netw Appl VII (2018)",
  ["complex networks xi : proceedings of the 11th conference on complex networks complenet 2020. complenet (workshop) (11th : 2020 : exeter, england)"] = "Complex Netw XI (2020)",
  ["complex psychiatry"] = "Complex Psychiatry",
  ["complex systems"] = "Complex Syst.",
  ["complex variables"] = "Complex Variables Theory Appl.",
  ["complex variables and elliptic equations"] = "Complex Var. Elliptic Equations",
  ["complex variables and elliptic equations. an international journal"] = "Complex Var. Elliptic Equ.",
  ["complexity international"] = "Complexity Int.",
  ["complicated cardiac patient"] = "Complicat Card Patient",
  ["complutum. publicaciones del departamento de prehistoria de la universidad complutense de madrid"] = "Complutum",
  ["compositae newsletter"] = "Compos Newsl",
  ["composite interfaces"] = "Compos. Interfaces",
  ["composite structures"] = "Compos. Struct.",
  ["composites and advanced materials"] = "Compos. Adv. Mater.",
  ["composites business analyst"] = "Compos. Bus. Anal.",
  ["composites communications"] = "Compos. Commun.",
  ["composites engineering"] = "Compos. Eng.",
  ["composites manufacturing"] = "Compos. Manuf.",
  ["composites part a applied science and manufacturing"] = "Composites Part A",
  ["composites part a: applied science and manufacturing"] = "Composites, Part A",
  ["composites part b engineering"] = "Composites Part B",
  ["composites part b: engineering"] = "Composites, Part B",
  ["composites part c: open access"] = "Composites, Part C: Open Access",
  ["composites science and technology"] = "Compos. Sci. Technol.",
  ["composites theory and practice"] = "Compos. Theory Pract.",
  ["composites, part a applied science and manufacturing"] = "Composites, Part A",
  ["composites, part b engineering"] = "Composites, Part B",
  ["composites. part a, applied science and manufacturing"] = "Compos Part A Appl Sci Manuf",
  ["composites. part b, engineering"] = "Compos B Eng",
  ["composites: mechanics, computations, applications: an international journal"] = "Compos.: Mech. Comput. Appl.: Int. J.",
  ["compositio mathematica"] = "Compos. Math.",
  ["compost science & utilization"] = "Compost Sci Util",
  ["comprehensive child and adolescent nursing"] = "Compr Child Adolesc Nurs",
  ["comprehensive gerontology. section a, clinical and laboratory sciences"] = "Compr Gerontol A",
  ["comprehensive gerontology. section b, behavioural, social and applied sciences"] = "Compr. Gerontol. [B]",
  ["comprehensive gerontology. section b, behavioural, social, and applied sciences"] = "Compr Gerontol B",
  ["comprehensive gerontology. section c, interdisciplinary topics"] = "Compr Gerontol C",
  ["comprehensive immunology"] = "Compr Immunol",
  ["comprehensive ophthalmology update"] = "Compr Ophthalmol Update",
  ["comprehensive physiology"] = "Compr Physiol",
  ["comprehensive psychiatry"] = "Compr Psychiatry",
  ["comprehensive psychoneuroendocrinology"] = "Compr Psychoneuroendocrinol",
  ["comprehensive results in social psychology"] = "Compr Results Soc Psychol",
  ["comprehensive reviews in food science and food safety"] = "Compr Rev Food Sci Food Saf",
  ["comprehensive series in photochemistry & photobiology"] = "Compre. Ser. Photochem. Photobiol.",
  ["comprehensive therapy"] = "Compr Ther",
  ["compte-rendu de la commission impériale archéologique, st. pétersbourg"] = "CRPetersbourg",
  ["comptes rendus biologies"] = "C R Biol",
  ["comptes rendus chimie"] = "C. R. Chim.",
  ["comptes rendus de l academie des sciences. serie iii, sciences de la vie"] = "C R Acad Sci III",
  ["comptes rendus de l association des anatomistes"] = "C. R. Assoc. Anat.",
  ["comptes rendus de l'académie bulgare des sciences : sciences mathématiques et naturelles"] = "C R Acad Bulg Sci",
  ["comptes rendus de l'académie des sciences. la vie des sciences"] = "C R Acad Sci Gen",
  ["comptes rendus de l'académie des sciences. série ii, mécanique, physique, chimie, sciences de l'univers, sciences de la terre"] = "C R Acad Sci II",
  ["comptes rendus de l'académie des sciences. série ii. fascicule b, mécanique = mechanics"] = "C R Acad Sci IIb Mec",
  ["comptes rendus de l'académie des sciences. série iii, sciences de la vie"] = "C R Acad Sci III",
  ["comptes rendus de l'association des anatomistes"] = "C R Assoc Anat",
  ["comptes rendus de la societe francaise de gynecologie"] = "C R Soc Fr Gyncol",
  ["comptes rendus de la société française de gynécologie"] = "C R Soc Fr Gyncol",
  ["comptes rendus de l’ academie des sciences serie iia:sciences de la terre et des planets"] = "C. R. Acad. Sci., Ser. IIa: Sci. Terre Planets",
  ["comptes rendus de l’ academie des sciences serie iib:mecanique physique chimie astronomie"] = "C. R. Acad. Sci., Ser. IIb: Mec., Phys., Chim., Astron.",
  ["comptes rendus de l’ academie des sciences serie iic: chemie"] = "C. R. Acad. Sci., Ser. IIc: Chim.",
  ["comptes rendus de l’ academie des sciences serie iic:chemie"] = "C. R. Acad. Sci., Ser. IIc: Chim.",
  ["comptes rendus de l’ academie des sciences serie iii: sciences de la vie"] = "C. R. Acad. Sci., Ser. III",
  ["comptes rendus de l’academie des sciences, serie iia: sciences de la terre et des planets"] = "C. R. Acad. Sci., Ser. IIa: Sci. Terre Planets",
  ["comptes rendus de l’academie des sciences, serie iib: mecanique physique chimie astronomie"] = "C. R. Acad. Sci., Ser. IIb: Mec., Phys., Chim., Astron.",
  ["comptes rendus de l’academie des sciences. la vie des sciences"] = "C R Acad Sci Gen",
  ["comptes rendus de l’academie des sciences. serie ii, mecanique, physique, chimie, sciences de l’univers, sciences de la terre"] = "C R Acad Sci II",
  ["comptes rendus de l’academie des sciences. serie iii, sciences de la vie"] = "C. R. Acad. Sci. III",
  ["comptes rendus de l’académie des inscriptions et belles-lettres"] = "CRAI",
  ["comptes rendus de l’académie des sciences"] = "C. R. Acad. Sci.",
  ["comptes rendus de l’académie des sciences - series i - mathematics"] = "Comptes Rendus Académie Sci. - Ser. I",
  ["comptes rendus de l’académie des sciences - series iv - physique et astrophysique"] = "Comptes Rendus Académie Sci. - Ser. IV",
  ["comptes rendus de l’académie des sciences, serie a: sciences mathématiques"] = "C. R. Acad. Sci. Ser. A",
  ["comptes rendus de l’académie des sciences, serie b: sciences physiques"] = "C. R. Acad. Sci. Ser. B",
  ["comptes rendus de l’association des anatomistes"] = "C R Assoc Anat",
  ["comptes rendus de therapeutique et de pharmacologie clinique"] = "C R Ther Pharmacol Clin",
  ["comptes rendus de thérapeutique et de pharmacologie clinique"] = "C R Ther Pharmacol Clin",
  ["comptes rendus des seances de l academie des sciences. serie iii, sciences de la vie"] = "C. R. Seances Acad. Sci. III",
  ["comptes rendus des seances de la societe de biologie et de ses filiales"] = "C R Seances Soc Biol Fil",
  ["comptes rendus des seances de l’academie des sciences. serie d, sciences naturelles"] = "C. R. Seances Acad. Sci. D.",
  ["comptes rendus des seances de l’academie des sciences. serie iii, sciences de la vie"] = "C R Seances Acad Sci III",
  ["comptes rendus des séances de l'académie des sciences. série d, sciences naturelles"] = "C R Seances Acad Sci D",
  ["comptes rendus des séances de l'académie des sciences. série iii, sciences de la vie"] = "C R Seances Acad Sci III",
  ["comptes rendus des séances de la société de biologie et de ses filiales"] = "C R Seances Soc Biol Fil",
  ["comptes rendus des séances de l’académie d’agriculture de france"] = "Comptes rendus séance Acad. Agric. Fr.",
  ["comptes rendus des travaux du laboratoire carlsberg"] = "C. R. Trav. Lab. Carlsberg",
  ["comptes rendus des travaux du laboratoire carlsberg. serie chimique"] = "C R Trav Lab Carlsberg [Chim]",
  ["comptes rendus des travaux du laboratoire carlsberg. serie physiologique"] = "Cr Trav Lab Carlsberg Ser Physiol",
  ["comptes rendus des travaux du laboratoire carlsberg. série chimique"] = "C R Trav Lab Carlsberg Chim",
  ["comptes rendus des travaux du laboratoire carlsberg. série physiologique"] = "Cr Trav Lab Carlsberg Ser Physiol",
  ["comptes rendus geoscience"] = "C. R. Geosci.",
  ["comptes rendus hebdomadaires des seances de l’academie des sciences"] = "C. R. Hebd. Seances Acad. Sci.",
  ["comptes rendus hebdomadaires des seances de l’academie des sciences. d, sciences naturelles"] = "C. R. Acad. Sci. Hebd. Seances Acad. Sci. D",
  ["comptes rendus hebdomadaires des seances de l’academie des sciences. serie d: sciences naturelles"] = "C R Acad Sci Hebd Seances Acad Sci D",
  ["comptes rendus hebdomadaires des séances de l'académie des sciences"] = "C R Hebd Seances Acad Sci",
  ["comptes rendus hebdomadaires des séances de l'académie des sciences. série d: sciences naturelles"] = "C R Acad Hebd Seances Acad Sci D",
  ["comptes rendus hebdomadaires des séances de l’académie des sciences"] = "C. R. Hebd. Séances Acad. Sci.",
  ["comptes rendus hebdomadaires des séances de l’académie des sciences, paris"] = "Comptes rendus hebd. séances acad. sci. Paris",
  ["comptes rendus hebdomadaires des séances de l’académie des sciences, serie a: sciences mathématiques"] = "C. R. Acad. Sci. Ser. A",
  ["comptes rendus hebdomadaires des séances de l’académie des sciences, serie b: sciences physiques"] = "C. R. Acad. Sci. Ser. B",
  ["comptes rendus mathematique"] = "C R Math",
  ["comptes rendus mathematique. academie des sciences. paris"] = "C. R. Math. Acad. Sci. Paris",
  ["comptes rendus mathematiques de l’academie des sciences. la societe royale du canada. mathematical reports of the academy of science. the royal society of canada"] = "C. R. Math. Acad. Sci. Soc. R. Can.",
  ["comptes rendus mathématiques des l’académie des sciences"] = "C. R. Math. Acad. Sci. Soc. R. Can.",
  ["comptes rendus mecanique"] = "C. R. Mec.",
  ["comptes rendus palevol"] = "C. R. Palevol",
  ["comptes rendus physique"] = "C. R. Phys.",
  ["comptes rendus. chimie (print)"] = "C R Chim",
  ["comptes rendus. palévol"] = "C R Palevol",
  ["comptes rendus. physique"] = "C R Phys",
  ["comptes-rendus des travaux du laboratoire carlsberg"] = "C R Trav Lab Carlsberg",
  ["computability. the journal of the association cie"] = "Computability",
  ["computation (basel, switzerland)"] = "Computation (Basel)",
  ["computational & applied mathematics"] = "Comput. Appl. Math.",
  ["computational & mathematical organization theory"] = "Comput. Math. Organ. Theory",
  ["computational & theoretical chemistry"] = "Comput Theor Chem",
  ["computational and applied mathematics"] = "Comput. Appl. Math.",
  ["computational and experimental methods in structures"] = "Comput. Exp. Methods Struct.",
  ["computational and mathematical biophysics"] = "Comput Math Biophys",
  ["computational and mathematical methods"] = "Comput. Math. Methods",
  ["computational and mathematical methods in medicine"] = "Comput Math Methods Med",
  ["computational and mathematical organization theory"] = "Comput Math Organ Theory",
  ["computational and structural biotechnology journal"] = "Comput Struct Biotechnol J",
  ["computational and systems oncology"] = "Comput Syst Oncol",
  ["computational and theoretical chemistry"] = "Comput. Theor. Chem.",
  ["computational and theoretical polymer science"] = "Comput. Theor. Polym. Sci.",
  ["computational astrophysics and cosmology"] = "Comput Astrophys Cosmol",
  ["computational biology"] = "Comput. Biol.",
  ["computational biology and chemistry"] = "Comput. Biol. Chem.",
  ["computational biology journal"] = "Comput Biol J",
  ["computational biomechanics for medicine : measurements, models, and predictions"] = "Comput Biomech Med (2017)",
  ["computational biomechanics for medicine. from algorithms to models and applications"] = "Comput Biomech Med Algorithms Models Appl (2017)",
  ["computational biomechanics for medicine. personalisation, validation and therapy"] = "Comput Biomech Med Personal Valid Ther (2019)",
  ["computational biomechanics for medicine. soft tissues and the musculoskeletal system"] = "Comput Biomech Med Soft Tiss Musculoskelet Syst",
  ["computational brain & behavior"] = "Comput Brain Behav",
  ["computational cognitive science"] = "Comput Cogn Sci",
  ["computational complexity"] = "Comput. Complexity",
  ["computational condensed matter"] = "Comput. Condens. Matter",
  ["computational diffusion mri : miccai workshop, athens, greece, october 2016. cdmri (workshop) (8th : 2016 : athens, greece)"] = "Comput Diffus MRI (2016)",
  ["computational diffusion mri : miccai workshop, boston, ma, usa, september 2014. cdmri (workshop) (6th : 2014 : boston, mass.)"] = "Comput Diffus MRI (2014)",
  ["computational diffusion mri : miccai workshop, munich, germany, october 9th, 2015. cdmri (workshop) (7th : 2015 : munich, germany)"] = "Comput Diffus MRI (2015)",
  ["computational diffusion mri and brain connectivity : miccai workshops, nagoya, japan, september 22nd, 2013. miccai workshop on computation diffusion mri (5th : 2013 : nagoya-shi, japan)"] = "Comput Diffus MRI Brain Connect (2013)",
  ["computational economics"] = "Comput Econ",
  ["computational geometry"] = "Comput. Geom.",
  ["computational geometry : theory and applications"] = "Comput Geom",
  ["computational geometry. theory and applications"] = "Comput. Geom.",
  ["computational geometry: theory and applications"] = "Comput. Geom. Theory Appl.",
  ["computational geosciences"] = "Comput. Geosci.",
  ["computational geosciences. modeling, simulation and data analysis"] = "Comput. Geosci.",
  ["computational imaging and vision"] = "Comput. Imaging Vision",
  ["computational intelligence"] = "Comput. Intell.",
  ["computational intelligence and neuroscience"] = "Comput Intell Neurosci",
  ["computational intelligence in engineering problem solving"] = "Comput. Intell. Eng. Probl. Solving",
  ["computational intelligence methods and applications"] = "Comput. Intell. Methods Appl.",
  ["computational intelligence methods for bioinformatics and biostatistics : 13th international meeting, cibb 2016, stirling, uk, september 1-3, 2016, revised selected papers. cibb (meeting) (13th : 2016 : stirling, england)"] = "Comput Intell Methods Bioinform Biostat (2016)",
  ["computational intelligence. an international journal"] = "Comput. Intell.",
  ["computational linguistics"] = "Comput. Linguist.",
  ["computational linguistics (association for computational linguistics)"] = "Comput Linguist Assoc Comput Linguist",
  ["computational management science"] = "Comput. Manag. Sci.",
  ["computational materials science"] = "Comput. Mater. Sci.",
  ["computational mathematics and analysis series"] = "Comput. Math. Anal. Ser.",
  ["computational mathematics and mathematical physics"] = "Comput. Math. Math. Phys.",
  ["computational mathematics and modeling"] = "Comput. Math. Model.",
  ["computational mechanics"] = "Comput. Mech.",
  ["computational medicine, public health, and biotechnology : building a man in the machine"] = "Comput Med Public Health Biotechnol (1994)",
  ["computational methods"] = "Comput. Methods",
  ["computational methods and clinical applications in musculoskeletal imaging : 5th international workshop, mski 2017, held in conjunction with miccai 2017, quebec city, qc, canada, september 10, 2017, revised selected papers. mski (workshop) (5th : 2017 : québec, québec)"] = "Comput Methods Clin Appl Musculoskelet Imaging (2017)",
  ["computational methods and function theory"] = "Comput. Methods Funct. Theory",
  ["computational methods for differential equations"] = "Comput. Methods Differ. Equ.",
  ["computational methods in applied mathematics"] = "Comput. Methods Appl. Math.",
  ["computational methods in applied sciences"] = "Comput. Methods Appl. Sci.",
  ["computational methods in applied sciences (springer)"] = "Comput Methods Appl Sci",
  ["computational methods in engineering & the sciences"] = "Comput. Methods Eng. Sci.",
  ["computational modeling of objects presented in images : fundamentals, methods, and applications : 4th international conference, compimage 2014, pittsburgh, pa, usa, september 3-5, 2014 : proceedings. compimage (conference) (4th : 2014 : pittsburgh, pa.)"] = "Comput Model Objects Present Images (2014)",
  ["computational molecular bioscience"] = "Comput Mol Biosci",
  ["computational music science"] = "Comput. Music Sci.",
  ["computational optimization and applications"] = "Comput. Optim. Appl.",
  ["computational optimization and applications. an international journal"] = "Comput. Optim. Appl.",
  ["computational particle mechanics"] = "Comput Part Mech",
  ["computational pathology and ophthalmic medical image analysis : first international workshop, compay 2018, and 5th international workshop, omia 2018, held in conjunction with miccai 2018, granada, spain, september 16-20, 2018, proceedings. compay (workshop) (1st : 2018 : granada, spain)"] = "Comput Pathol Ophthalmic Med Image Anal (2018)",
  ["computational psychiatry"] = "Comput. Psychiatry",
  ["computational psychiatry (cambridge, mass.)"] = "Comput Psychiatr",
  ["computational science & discovery"] = "Comput. Sci. Discovery",
  ["computational science & engineering"] = "Comput. Sci. Eng.",
  ["computational social networks"] = "Comput Soc Netw",
  ["computational social sciences"] = "Comput. Soc. Sci.",
  ["computational statistics"] = "Comput Stat",
  ["computational statistics & data analysis"] = "Comput Stat Data Anal",
  ["computational statistics and data analysis"] = "Comput. Stat. Data Anal.",
  ["computational synthesis and creative systems"] = "Comp. Synth. Creat. Sys.",
  ["computational systems bioinformatics / life sciences society. computational systems bioinformatics conference"] = "Comput Syst Bioinformatics Conf",
  ["computational systems bioinformatics. computational systems bioinformatics conference"] = "Comput Syst Bioinformatics Conf",
  ["computational thermal sciences: an international journal"] = "Comput. Therm. Sci.: Int. J.",
  ["computational topology in image context : 6th international workshop, ctic 2016, marseille, france, june 15-17, 2016, proceedings. ctic (conference) (6th : 2016 : marseille, france)"] = "Comput Topol Image Context (2016)",
  ["computational toxicology"] = "Comput. Toxicol.",
  ["computational toxicology (amsterdam, netherlands)"] = "Comput Toxicol",
  ["computational urban science"] = "Comput. Urban Sci.",
  ["computational visual media"] = "Comput Vis Media (Beijing)",
  ["computer aided design"] = "Comput Aided Des",
  ["computer aided geometric design"] = "Comput. Aided Geom. Des.",
  ["computer aided surgery"] = "Comput. Aided Surg.",
  ["computer aided surgery : official journal of the international society for computer aided surgery"] = "Comput Aided Surg",
  ["computer animation and virtual worlds"] = "Comput. Anim. Virtual Worlds",
  ["computer animation. acm siggraph symposium on computer animation"] = "Comput Animat",
  ["computer applications in engineering education"] = "Comput. Appl. Eng. Educ.",
  ["computer applications in the biosciences"] = "Comput. Appl. Biosci.",
  ["computer applications in the biosciences : cabios"] = "Comput Appl Biosci",
  ["computer architecture and design methodologies"] = "Comput. Archit. Des. Methodol.",
  ["computer architecture news"] = "Comput Archit News",
  ["computer assisted and robotic endoscopy and clinical image-based procedures : 4th international workshop, care 2017, and 6th international workshop, clip 2017, held in conjunction with miccai 2017 québec city, qc, canada, september 14, 2017, proceedings. care (workshop) (4th : 2017 : québec, québec)"] = "Comput Assist Robot Endosc Clin Image Based Proced (2017)",
  ["computer assisted language learning"] = "Comput Assist Lang Learn",
  ["computer assisted surgery"] = "Comput. Assisted Surg.",
  ["computer assisted surgery (abingdon, england)"] = "Comput Assist Surg (Abingdon)",
  ["computer communications"] = "Comput. Commun.",
  ["computer communications and networks"] = "Comput. Commun. Netw.",
  ["computer engineering series"] = "Comput. Eng. Ser.",
  ["computer frontiers conference : [proceedings]. computer frontiers conference"] = "Comut Front Conf",
  ["computer games journal"] = "Comput. Games J.",
  ["computer graphics"] = "Comput Graph (ACM)",
  ["computer graphics forum"] = "Comput. Graphics Forum",
  ["computer graphics forum : journal of the european association for computer graphics"] = "Comput Graph Forum",
  ["computer integrated manufacturing systems"] = "Comput. Integr. Manuf. Syst.",
  ["computer journal"] = "Comput. J.",
  ["computer languages, systems & structures"] = "Comput Lang Syst Struct",
  ["computer life"] = "Comput. Life",
  ["computer methods and programs in biomedicine"] = "Comput Methods Programs Biomed",
  ["computer methods in applied mechanics and engineering"] = "Comput. Methods Appl. Mech. Eng.",
  ["computer methods in biomechanics and biomedical engineering"] = "Comput Methods Biomech Biomed Engin",
  ["computer methods in biomechanics and biomedical engineering. imaging & visualization"] = "Comput Methods Biomech Biomed Eng Imaging Vis",
  ["computer methods in biomechanics and biomedical engineering: imaging and visualization"] = "Comput. Methods Biomech. Biomed. Eng.: Imaging Visualization",
  ["computer modeling and simulation in engineering"] = "Comput. Model. Simul. Eng.",
  ["computer modeling in engineering & sciences : cmes"] = "Comput Model Eng Sci",
  ["computer modeling in engineering and sciences"] = "Comp. Model. Eng. Sci.",
  ["computer music journal"] = "Comput. Music J.",
  ["computer networks"] = "Comput. Networks",
  ["computer physics communications"] = "Comput. Phys. Commun.",
  ["computer programs in biomedicine"] = "Comput Programs Biomed",
  ["computer science"] = "Comput. Sci.",
  ["computer science (berlin, germany)"] = "Comput Sci (Berl)",
  ["computer science and scientific computing"] = "Comput. Sci. Sci. Comput.",
  ["computer science education"] = "Comput. Sci. Educ.",
  ["computer science foundations and applied logic"] = "Comput. Sci. Found. Appl. Log.",
  ["computer science in economics and management"] = "Computer Sci. Econ. Manage.",
  ["computer science monographs"] = "Comput. Sci. Monogr.",
  ["computer science review"] = "Comput. Sci. Rev.",
  ["computer science reviews and trends"] = "Comput. Sci. Rev. Trends",
  ["computer science: computer architecture and design"] = "Comput. Sci.: Comput. Archit. Design",
  ["computer science: systems programming"] = "Comput. Sci.: Systems Programming",
  ["computer speech & language"] = "Comput. Speech Lang.",
  ["computer speech and language"] = "Comput. Speech Lang.",
  ["computer standards & interfaces"] = "Comput Stand Interfaces",
  ["computer supported cooperative work"] = "Comput. Support. Coop. Work",
  ["computer supported cooperative work : cscw : an international journal"] = "Comput Support Coop Work",
  ["computer systems science and engineering"] = "Comput. Syst. Sci. Eng.",
  ["computer technology and application"] = "Comput Technol Appl",
  ["computer vision and image understanding"] = "Comput. Vision Image Understanding",
  ["computer vision and image understanding : cviu"] = "Comput Vis Image Underst",
  ["computer vision for biomedical image applications : first international workshop, cvbia 2005, beijing, china, october 21, 2005 : proceedings. cvbia 2005 (2005 : beijing, china)"] = "Comput Vis Biomed Image Appl (2005)",
  ["computer vision, graphics, and image processing"] = "Comput Vis Graph Image Process",
  ["computer-aided civil and infrastructure engineering"] = "Comput.-Aided Civ. Infrastruct. Eng.",
  ["computer-aided design"] = "Comput.-Aided Des.",
  ["computer-aided design and applications"] = "Comput.-Aided Des. Applic.",
  ["computer-aided verification : proceedings. cav (conference)"] = "Comput Aided Verif",
  ["computer-assisted and robotic endoscopy : first international workshop, care 2014, held in conjunction with miccai 2014, boston, ma, usa, september 18, 2014 : revised selected papers. care (workshop) (1st : 2014 : boston, mass.)"] = "Comput Assist Robot Endosc (2014)",
  ["computer/law journal"] = "Comput Law J",
  ["computerized medical imaging and graphics"] = "Comput. Med. Imaging Graphics",
  ["computerized medical imaging and graphics : the official journal of the computerized medical imaging society"] = "Comput Med Imaging Graph",
  ["computerized radiology"] = "Comput. Radiol.",
  ["computerized radiology : official journal of the computerized tomography society"] = "Comput Radiol",
  ["computerized tomography"] = "Comput Tomogr",
  ["computers & chemical engineering"] = "Comput. Chem. Eng.",
  ["computers & chemistry"] = "Comput Chem",
  ["computers & education"] = "Comput Educ",
  ["computers & electrical engineering : an international journal"] = "Comput Electr Eng",
  ["computers & fluids"] = "Comput. Fluids",
  ["computers & fluids. an international journal"] = "Comput. & Fluids",
  ["computers & geosciences"] = "Comput Geosci",
  ["computers & graphics"] = "Comput Graph",
  ["computers & graphics: x"] = "Comput Graph X",
  ["computers & industrial engineering"] = "Comput. Ind. Eng.",
  ["computers & mathematcs with applications"] = "Comput. Math. Appl.",
  ["computers & mathematics with applications"] = "Comput. Math. Appl.",
  ["computers & mathematics with applications (oxford, england : 1987)"] = "Comput Math Appl",
  ["computers & mathematics with applications. an international journal"] = "Comput. Math. Appl.",
  ["computers & operations research"] = "Comput. Oper. Res.",
  ["computers & operations research and their application to problems of world concern"] = "Comput. Oper. Res.",
  ["computers & security"] = "Comput Secur",
  ["computers & structures"] = "Comput Struct",
  ["computers and artif. intell."] = "Comput. Artif. Intell.",
  ["computers and artificial intelligence"] = "Comput. Artif. Intell.",
  ["computers and biomedical research"] = "Comput. Biomed. Res.",
  ["computers and biomedical research, an international journal"] = "Comput Biomed Res",
  ["computers and chemical engineering"] = "Comput. Chem. Eng.",
  ["computers and chemistry"] = "Comput. Chem. (Oxford)",
  ["computers and composition"] = "Comput Compos",
  ["computers and concrete"] = "Comput. Concr.",
  ["computers and education open"] = "Comput. Educ. Open",
  ["computers and education: artificial intelligence"] = "Comput. Educ.: Artif. Intell.",
  ["computers and electrical engineering"] = "Comput. Electr. Eng.",
  ["computers and electronics in agriculture"] = "Comput. Electron. Agric.",
  ["computers and geosciences"] = "Comput. Geosci.",
  ["computers and geotechnics"] = "Comput. Geotech.",
  ["computers and people series"] = "Comput. People Ser.",
  ["computers and structures"] = "Comput. Struct.",
  ["computers and the humanitie"] = "CHum",
  ["computers and the humanities"] = "Comput Hum",
  ["computers in biology and medicine"] = "Comput Biol Med",
  ["computers in cardiology"] = "Comput Cardiol",
  ["computers in healthcare"] = "Comput Healthc",
  ["computers in hospitals"] = "Comput Hosp",
  ["computers in human behavior"] = "Comput Human Behav",
  ["computers in human behavior reports"] = "Comput. Hum. Behav. Rep.",
  ["computers in industry"] = "Comput. Ind.",
  ["computers in nursing"] = "Comput Nurs",
  ["computers in physics"] = "Comput. Phys.",
  ["computers in the schools : interdisciplinary journal of practice, theory, and applied research"] = "Comput Sch",
  ["computers, environment and urban systems"] = "Comput Environ Urban Syst",
  ["computers, informatics, nursing"] = "Comput. Inform. Nurs.",
  ["computers, informatics, nursing : cin"] = "Comput Inform Nurs",
  ["computers, materials & continua"] = "Comput. Mater. Continua",
  ["computing and informatics"] = "Comput. Inform.",
  ["computing and software for big science"] = "Comput Softw Big Sci",
  ["computing and visualization in science"] = "Comput Vis Sci",
  ["computing in cardiology"] = "Comput Cardiol (2010)",
  ["computing in geometry and topology (cgt)"] = "Comput. Geom. Topol.",
  ["computing in science & engineering"] = "Comput. Sci. Eng.",
  ["computing letters"] = "Comput. Lett.",
  ["computing open"] = "Comput. Open",
  ["computing supplementum"] = "Comput. Suppl.",
  ["computing. archives for scientific computing"] = "Computing",
  ["comsig review"] = "COMSIG Rev",
  ["comsig review / comsig, chiropractors and osteopaths musculo-skeletal interest group"] = "COMSIG Rev",
  ["comunicaciones presentadas a las jornadas del comite espanol de la detergencia"] = "Comun. Jorn. Com. Esp. Deterg.",
  ["concepts in immunopathology"] = "Concepts Immunopathol",
  ["concepts in magenetic resonance"] = "Concepts Magn. Reson.",
  ["concepts in magnetic resonance"] = "Concepts Magn Reson",
  ["concepts in magnetic resonance part a"] = "Concepts Magn. Reson. Part A",
  ["concepts in magnetic resonance part b"] = "Concepts Magn. Reson. Part B",
  ["concepts in magnetic resonance. part a, bridging education and research"] = "Concepts Magn Reson Part A Bridg Educ Res",
  ["concepts in magnetic resonance. part b, magnetic resonance engineering"] = "Concepts Magn Reson Part B Magn Reson Eng",
  ["concern (regina, sask.)"] = "Concern",
  ["concern for dying"] = "Concern Dying",
  ["concern in care of the aging"] = "Concern Care Aging",
  ["concerned demography"] = "Concerned Demogr",
  ["conciencia latinoamericana : boletin periodico de opiniones de catolicas por el derecho a decidir"] = "Concienc Latinoam",
  ["conciencia latinoamericana : boletín periódico de opiniones de católicas por el derecho a decidir"] = "Concienc Latinoam",
  ["concordia historical institute quarterly. concordia historical institute"] = "Concordia Hist Inst Q",
  ["concordia journal"] = "Concordia J",
  ["concordia law review"] = "Concordia Law Rev",
  ["concours medical"] = "Concours Med.",
  ["concours médical"] = "Concours Med",
  ["concrete engineering international"] = "Concr. Eng. Int.",
  ["concrete international"] = "Concr. Int.",
  ["concrete international : design & construction"] = "Concr Int",
  ["concrete operators"] = "Concr. Oper.",
  ["concrete science and engineering"] = "Concr. Sci. Eng.",
  ["concurrency and computation : practice & experience"] = "Concurr Comput",
  ["concurrency and computation: practice and experience"] = "Concurrency Comput. Pract. Exper.",
  ["concurrent engineering"] = "Concurrent Eng.",
  ["concurrent engineering, research, and applications"] = "Concurr Eng Res Appl",
  ["concurrent engineering: research and applications"] = "Concurrent Eng.: Res. Appl.",
  ["concussion (london, england)"] = "Concussion",
  ["condensed matter"] = "Condens Matter",
  ["condensed matter physics"] = "Condens. Matter Phys.",
  ["condensed matter theories"] = "Condens. Matter Theor.",
  ["conditio judaica"] = "Conditio Jud",
  ["conditional reflex"] = "Cond Reflex",
  ["conditioning medicine"] = "Cond Med",
  ["condor, the"] = "Condor",
  ["condorcet studies"] = "Condorcet Stud",
  ["conductual : revista internacional de interconductismo y análisis de conducta"] = "Conductual",
  ["confederate states medical and surgical journal"] = "Confed State Med Surg J",
  ["confederation of australian critical care nurses journal"] = "Confed Aust Crit Care Nurses J",
  ["confederation of the australian critical care nurses journal"] = "Confed. Aust. Crit. Care Nurses J.",
  ["conference on advances in aids vaccine development : program book. national cooperative vaccine development groups for aids. meeting (8th : 1996 : bethesda, md.)"] = "Conf Adv AIDS Vaccine Dev Natl Coop Vaccine Dev Groups AIDS Meet 8th 1996 Bethesda Md",
  ["conference on advances in aids vaccine development, 1997 : ninth annual meeting of the national cooperative vaccine development groups for aids, may 4-7, 1997 : program book. conference on advances in aids vaccine development (1997 : bethesda, md.)"] = "Conf Adv AIDS Vaccine Dev 1997 Conf Adv AIDS Vaccine Dev 1997 Bethesda Md",
  ["conference on computer vision and pattern recognition workshops. ieee computer society conference on computer vision and pattern recognition. workshops"] = "Conf Comput Vis Pattern Recognit Workshops",
  ["conference on human system interactions : [proceedings]. international conference on human system interaction"] = "Conf Human Syst Interact",
  ["conference on lasers and electro-optics : (cleo). conference on lasers and electro-optics"] = "Conf Lasers Electro Optics",
  ["conference proceedings and lecture notes in applied physics"] = "Conf. Proc. Lecture Notes Appl. Phys.",
  ["conference proceedings and lecture notes in geometry and topology"] = "Conf. Proc. Lecture Notes Geom. Topology",
  ["conference proceedings and lecture notes in physics"] = "Conf. Proc. Lecture Notes Phys.",
  ["conference proceedings of the society for experimental mechanics series"] = "Conf. Proc. Soc. Exp. Mech. Ser.",
  ["conference proceedings on theory of computing"] = "Conf. Proc. Theory Comput.",
  ["conference proceedings. ethnographic praxis in industry conference"] = "Conf Proc Ethnogr Prax Ind Conf",
  ["conference proceedings. frontiers in education conference"] = "Conf Proc Front Educ Conf",
  ["conference proceedings. ieee international conference on signal and image processing applications"] = "Conf Proc IEEE Int Conf Signal Image Process Appl",
  ["conference proceedings. ieee international conference on systems, man, and cybernetics"] = "Conf Proc IEEE Int Conf Syst Man Cybern",
  ["conference proceedings. international conference on image formation in x-ray computed tomography"] = "Conf Proc Int Conf Image Form Xray Comput Tomogr",
  ["conference proceedings. international research council on biomechanics of injury"] = "Conf Proc Int Res Counc Biomech Inj",
  ["conference proceedings. society of plastics engineers. technical conference"] = "Conf Proc Soc Plast Eng",
  ["conference record of the ieee photovoltaic specialists conference"] = "Conf. Rec. IEEE Photovoltaic Spec. Conf.",
  ["conference record. asilomar conference on signals, systems & computers"] = "Conf Rec Asilomar Conf Signals Syst Comput",
  ["conferences d’histoire de la medecine : [fascicule] / institut d’histoire de la medecine, musee d’histoire de la medecine. conferences d’histoire de la medecine"] = "Conf Hist Med",
  ["conferences lyonnaises d’ophtalmologie"] = "Conf Lyon Ophtalmol",
  ["conferencias y estudios de historia y organizacion de la ciencia / academia de ciencias de cuba"] = "Conf Estud Hist Organ Cienc",
  ["conferencias y estudios de historia y organización de la ciencia"] = "Conf Estud Hist Organ Cienc",
  ["conferenze del seminario di matematica dell’università di bari"] = "Conf. Semin. Mat. Univ. Bari",
  ["conférences lyonnaises d'ophtalmologie"] = "Conf Lyon Ophtalmol",
  ["confinia neurologica"] = "Confin Neurol",
  ["confinia psychiatrica"] = "Confin. Psychiatr.",
  ["confinia psychiatrica. borderland of psychiatry. grenzgebiete der psychiatrie. les confins de la psychiatrie"] = "Confin Psychiatr",
  ["conflict and health"] = "Confl Health",
  ["conflict management and peace science"] = "Conflict Manage. Peace Sci.",
  ["confluentes mathematici"] = "Confluentes Math.",
  ["conformal geometry and dynamics"] = "Conform. Geom. Dyn.",
  ["conformal geometry and dynamics. an electronic journal of the american mathematical society"] = "Conform. Geom. Dyn.",
  ["confrontations radio-anatomo-cliniques"] = "Confront. Radio. Anat. Clin.",
  ["conférences d'histoire de la médecine : [fascicule]. conférences d'histoire de la médecine"] = "Conf Hist Med",
  ["congenital anomalies"] = "Congenit Anom (Kyoto)",
  ["congenital heart disease"] = "Congenit Heart Dis",
  ["congestive heart failure"] = "Congest. Heart Fail.",
  ["congestive heart failure (greenwich, conn.)"] = "Congest Heart Fail",
  ["congres international de stomatologie"] = "Congr. Int. Stomatol.",
  ["congreso. sociedad de cirujanos de chile"] = "Congr Soc Cir Chile",
  ["congress monthly"] = "Congr. Mon",
  ["congressional quarterly weekly report"] = "Congr Q Wkly Rep",
  ["congressus numerantium"] = "Congr. Numer.",
  ["congressus numerantium. a conference journal on numerical themes"] = "Congr. Numer.",
  ["congrès international de stomatologie"] = "Congr Int Stomatol",
  ["connaissance de l’egypte ancienne"] = "Connaiss. Egypte Anc.",
  ["connaissance des arts"] = "Connaiss Arts",
  ["connaissance hellénique"] = "ConnHell",
  ["connecticut dental student journal"] = "Conn Dent Stud J",
  ["connecticut dental student journal / the university of connecticut health center, school of dental medicine"] = "Conn Dent Stud J",
  ["connecticut general statutes annotated / under arrangement of the official general statutes of connecticut, revision of 1958. connecticut"] = "Conn Gen Statut Annot Conn",
  ["connecticut general statutes annotated. connecticut"] = "Conn Gen Statut Annot Conn",
  ["connecticut health bulletin"] = "Conn Health Bull",
  ["connecticut journal of international law"] = "Conn J Int Law",
  ["connecticut law review"] = "Conn Law Rev",
  ["connecticut medicine"] = "Conn Med",
  ["connecticut nursing news"] = "Conn. Nurs. News",
  ["connecticut nursing news (meriden, conn. : 1980)"] = "Conn Nurs News",
  ["connecticut state medical journal"] = "Conn State Med J",
  ["connection science"] = "Conn Sci",
  ["connections (toronto, ont.)"] = "Connect (Tor)",
  ["connective tissue research"] = "Connect Tissue Res",
  ["connectomics in neuroimaging : first international workshop, cni 2017, held in conjunction with miccai 2017, quebec city, qc, canada, september 14, 2017, proceedings. cni (workshop) (1st : 2017 : québec, québec)"] = "Connectomics Neuroimaging (2017)",
  ["connectomics in neuroimaging : second international workshop, cni 2018, held in conjunction with miccai 2018, granada, spain, september 20, 2018 : proceedings. cni (workshop) (2nd : 2018 : granada, spain)"] = "Connect Neuroimaging (2018)",
  ["connectomics in neuroimaging : third international workshop, cni 2019, held in conjunction with miccai 2019, shenzhen, china, october 13, 2019, proceedings. cni (workshop) (3rd : 2019 : shenzhen shi, china)"] = "Connect Neuroimaging (2019)",
  ["conoscenze. rivista annuale della soprintendenza archeologica e per i beni ambientali, architettonici, artistici e storici del molise"] = "Conoscenze",
  ["conscience (washington, d.c.)"] = "Conscience",
  ["consciousness and cognition"] = "Conscious Cogn",
  ["consensus development conference summaries"] = "Consens Dev Conf Summ Natl Inst Health",
  ["consensus development conference summaries / national institutes of health"] = "Consens Dev Conf Summ Natl Inst Health",
  ["consensus development conference summaries, national institutes of health"] = "Consens. Dev. Conf. Summ. Natl. Inst. Health",
  ["consensus statement"] = "Consens. Statement",
  ["consensus statement / nih consensus development conference. national institutes of health consensus development conference"] = "Consens Statement",
  ["consensus statement. national institutes of health consensus development conference"] = "Consens Statement",
  ["conservation and management of archaeological sites"] = "Conserv. Manage. Archaeol. Sites",
  ["conservation biology"] = "Conserv. Biol.",
  ["conservation biology : the journal of the society for conservation biology"] = "Conserv Biol",
  ["conservation biology series"] = "Conserv Biol Ser",
  ["conservation genetics"] = "Conserv. Genet.",
  ["conservation genetics (print)"] = "Conserv Genet",
  ["conservation genetics resources"] = "Conserv Genet Resour",
  ["conservation letters"] = "Conserv Lett",
  ["conservation physiology"] = "Conserv Physiol",
  ["conservation science and practice"] = "Conserv Sci Pract",
  ["conservative judaism"] = "Conserv Jud",
  ["consommation, revue de socio-economie"] = "Consommation",
  ["consommation, revue de socio-économie"] = "Consommation",
  ["constellations (oxford, england)"] = "Constellations",
  ["constitution (foundation for the united states constitution)"] = "Constitution",
  ["constitutional commentary"] = "Const Comment",
  ["constitutional law journal (newark, n.j. : 1990)"] = "Const Law J",
  ["constitutional political economy"] = "Constit. Polit. Economy",
  ["constraints : an international journal"] = "Constraints",
  ["constraints. an international journal"] = "Constraints",
  ["construction and building materials"] = "Constr. Build. Mater.",
  ["construction history : journal of the construction history group"] = "Constr Hist",
  ["construction innovation"] = "Constr. Innovation",
  ["construction management and economics"] = "Construct. Manage. Econ.",
  ["construction research and innovation"] = "Construct. Res. Innovation",
  ["construction robotics"] = "Construct. Rob.",
  ["constructive approximation"] = "Constr Approx",
  ["constructive approximation. an international journal for approximations and expansions"] = "Constr. Approx.",
  ["constructive mathematical analysis"] = "Constr. Math. Anal.",
  ["consulting psychology journal"] = "Consult Psychol J",
  ["consulting-specifying engineer"] = "Consult Specif Eng",
  ["consumer health perspectives"] = "Consum Health Perspect",
  ["consumer markets abroad"] = "Consum Mark Abroad",
  ["consumer reports"] = "Consum Rep",
  ["consumer reports on health"] = "Consum Rep Health",
  ["contact (geneva, switzerland)"] = "Contact",
  ["contact (thousand oaks (ventura county, calif.))"] = "Contact (Thousand Oaks)",
  ["contact and intraocular lens medical journal"] = "Contact Intraocul Lens Med J",
  ["contact dermatitis"] = "Contact Dermatitis",
  ["contact in context"] = "Contact Context",
  ["contact lens & anterior eye : the journal of the british contact lens association"] = "Cont Lens Anterior Eye",
  ["contact point"] = "Contact Point",
  ["contamination control"] = "Contam Control",
  ["contamination control-biomedical environments"] = "Contam Control Biomed Environ",
  ["contemporanea (bologna, italy : 1998)"] = "Contemporanea",
  ["contemporary administrator"] = "Contemp Adm",
  ["contemporary administrator for long-term care"] = "Contemp Adm Long Term Care",
  ["contemporary advances in science and technology"] = "Contemp. Adv. Sci. Technol.",
  ["contemporary analysis and applied mathematics"] = "Contemp. Anal. Appl. Math.",
  ["contemporary anesthesia practice"] = "Contemp Anesth Pract",
  ["contemporary behavioral health care"] = "Contemp Behav Health Care",
  ["contemporary british history"] = "Contemp Br Hist",
  ["contemporary clinical dentistry"] = "Contemp Clin Dent",
  ["contemporary clinical trials"] = "Contemp Clin Trials",
  ["contemporary clinical trials communications"] = "Contemp Clin Trials Commun",
  ["contemporary crises"] = "Contemp Crisis",
  ["contemporary drug problems"] = "Contemp Drug Probl",
  ["contemporary economic policy"] = "Contemp Econ Policy",
  ["contemporary educational psychology"] = "Contemp Educ Psychol",
  ["contemporary educational technology"] = "Contemp Educ Technol",
  ["contemporary european history"] = "Contemp Eur Hist",
  ["contemporary evaluation research"] = "Contemp. Evaluation Res.",
  ["contemporary family therapy"] = "Contemp Fam Ther",
  ["contemporary french civilization"] = "Contemp Fr Civiliz",
  ["contemporary gerontology"] = "Contemp Gerontol",
  ["contemporary history in context series"] = "Contemp. Hist. Con. Ser.",
  ["contemporary hypnosis & integrative therapy"] = "Contemp Hypn Integr Ther",
  ["contemporary hypnosis : the journal of the british society of experimental and clinical hypnosis"] = "Contemp Hypn",
  ["contemporary internal medicine"] = "Contemp Intern Med",
  ["contemporary issues in clinical biochemistry"] = "Contemp Issues Clin Biochem",
  ["contemporary issues in communication science and disorders : cicsd"] = "Contemp Issues Commun Sci Disord",
  ["contemporary issues in education research (littleton, colo.)"] = "Contemp Issues Educ Res (Littleton)",
  ["contemporary jewry"] = "Contemp Jew",
  ["contemporary longterm care"] = "Contemp Longterm Care",
  ["contemporary marxism"] = "Contemp Marx",
  ["contemporary materials"] = "Contemp Mater",
  ["contemporary mathematicians"] = "Contemp. Mathematicians",
  ["contemporary mathematics"] = "Contemp. Math.",
  ["contemporary mathematics and its applications: monographs, expositions and lecture notes"] = "Contemp. Math. Appl. Monogr. Expo. Lect. Notes",
  ["contemporary neurology series"] = "Contemp Neurol Ser",
  ["contemporary neurosurgery"] = "Contemp Neurosurg",
  ["contemporary nurse"] = "Contemp Nurse",
  ["contemporary nurse : a journal for the australian nursing profession"] = "Contemp Nurse",
  ["contemporary ob/gyn"] = "Contemp Ob Gyn",
  ["contemporary oncology"] = "Contemp Oncol",
  ["contemporary oncology (poznań, poland)"] = "Contemp Oncol (Pozn)",
  ["contemporary organic synthesis"] = "Contemp. Org. Synth.",
  ["contemporary orthopaedics"] = "Contemp Orthop",
  ["contemporary pediatrics"] = "Contemp Pediatr",
  ["contemporary pharmacy practice"] = "Contemp Pharm Pract",
  ["contemporary philosophy"] = "Contemp Philos",
  ["contemporary philosophy-new survery"] = "Contemp. Philos. New. S.",
  ["contemporary physicists"] = "Contemp. Phys.",
  ["contemporary physics"] = "Contemp. Phys.",
  ["contemporary policy issues"] = "Contemp Policy Issues",
  ["contemporary problems of ecology"] = "Contemp Probl Ecol",
  ["contemporary psychiatry (hagerstown, md.)"] = "Contemp Psychiatry (Hagerstown Md)",
  ["contemporary psychoanalysis"] = "Contemp Psychoanal",
  ["contemporary review (london, england)"] = "Contemp Rev",
  ["contemporary reviews in obstetrics and gynaecology"] = "Contemp Rev Obstet Gynaecol",
  ["contemporary school psychology"] = "Contemp Sch Psychol",
  ["contemporary security policy"] = "Contemp Secur Policy",
  ["contemporary social science"] = "Contemp Soc Sci",
  ["contemporary sociology"] = "Contemp Sociol",
  ["contemporary sociology-a journal of reviews"] = "Contemp. Sociol.-J. Rev.",
  ["contemporary south asia"] = "Contemp South Asia",
  ["contemporary southeast asia"] = "Contemp Southeast Asia",
  ["contemporary surgery"] = "Contemp Surg",
  ["contemporary systems thinking"] = "Contemp. Syst. Think.",
  ["contemporary therapy"] = "Contemp Ther",
  ["contemporary topics in immunobiology"] = "Contemp Top Immunobiol",
  ["contemporary topics in laboratory animal science"] = "Contemp Top Lab Anim Sci",
  ["contemporary topics in laboratory animal science / american association for laboratory animal science"] = "Contemp Top Lab Anim Sci",
  ["contemporary topics in molecular immunology"] = "Contemp Top Mol Immunol",
  ["contemporary urology"] = "Contemp Urol",
  ["contests in mathematics"] = "Contests Math.",
  ["contexts (berkeley, calif.)"] = "Contexts (Berkeley Calif)",
  ["continental philosophy review"] = "Cont Philos Rev",
  ["continental shelf research"] = "Cont. Shelf Res.",
  ["continuing care"] = "Contin Care",
  ["continuity and change"] = "Contin Chang",
  ["continuum (minneapolis, minn.)"] = "Continuum (Minneap Minn)",
  ["continuum (mount lawley, w.a.)"] = "Continuum (Mount Lawley)",
  ["continuum (society for social work administrators in health care)"] = "Continuum",
  ["continuum (society for social work leadership in health care)"] = "Continuum Soc Soc Work Leadersh Health Care",
  ["continuum mechanics and thermodynamics"] = "Continuum Mech. Thermodyn.",
  ["contra costa dental bulletin"] = "Contra Costa Dent Bull",
  ["contraception and reproductive medicine"] = "Contracept Reprod Med",
  ["contraception report"] = "Contracept Rep",
  ["contraception, fertilite, sexualite"] = "Contracept. Fertil. Sex. (Paris)",
  ["contraception, fertilite, sexualite (1992)"] = "Contracept Fertil Sex",
  ["contraception, fertilité, sexualité"] = "Contracept Fertil Sex (Paris)",
  ["contraception, fertilité, sexualité (1992)"] = "Contracept Fertil Sex",
  ["contraception: x"] = "Contracept X",
  ["contraceptive delivery systems"] = "Contracept Deliv Syst",
  ["contraceptive technology"] = "Contracept Technol",
  ["contraceptive technology update"] = "Contracept Technol Update",
  ["contract (new york, n.y. : 1960)"] = "Contract",
  ["contract design"] = "Contract Des",
  ["contract healthcare"] = "Contract Healthc",
  ["contract interiors"] = "Contract Inter",
  ["contrast media & molecular imaging"] = "Contrast Media Mol Imaging",
  ["contrastes (murcia, spain)"] = "Contrastes",
  ["contree / [raad vir geesteswetenskaplike navorsing, instituut vir geskiedenisnavorsing, afdeling streekgeskiedenis]"] = "Contree",
  ["contributi (biblioteca antonio panizzi)"] = "Contributi",
  ["contributi del centro linceo interdisciplinare di scienze matematiche e loro applicazioni"] = "Contrib. Centro Linceo Interdiscip. Sci. Mat. Appl.",
  ["contributi della scuola di specializzazione in archeologia dell’università degli studi di pisa"] = "CSSpecPisa",
  ["contributi dell’istituto di archeologia. pubblicazioni dell’università cattolica del sacro cuore, milano"] = "CIstAMilano",
  ["contributi dell’istituto di storia antica dell’univ. del sacro cuore"] = "CISA",
  ["contributi e materiali di archeologia orientale"] = "CMatAOr",
  ["contributions a l’histoire economique et sociale"] = "Contrib Hist Econ Soc",
  ["contributions à l'histoire économique et sociale"] = "Contrib Hist Econ Soc",
  ["contributions from the central research institute for agriculture bogor"] = "Contrib. Cent. Res. Inst. Agric. Bogor",
  ["contributions from the united states national herbarium"] = "Contrib. U. S. Natl. Herb.",
  ["contributions from the university of michigan herbarium"] = "Contrib Univ Mich Herb",
  ["contributions in american history"] = "Contr. Am. Hi.",
  ["contributions in biology and geology"] = "Contrib Biol Geol",
  ["contributions in biology and geology / milwaukee public museum"] = "Contrib Biol Geol",
  ["contributions in marine science"] = "Contrib Mar Sci",
  ["contributions in mathematical and computational sciences"] = "Contrib. Math. Comput. Sci.",
  ["contributions in medical history"] = "Contr Med Hist",
  ["contributions in medical studies"] = "Contrib Med Stud",
  ["contributions in philosophy"] = "Contrib. Philos.",
  ["contributions in psychology"] = "Contrib Psychol",
  ["contributions of the astronomical observatory skalnate pleso"] = "Contrib. Astron. Obs. Skalnate Pleso",
  ["contributions of the meteoritical society"] = "Contrib. Meteorit. Soc.",
  ["contributions of the society for research on meteorites"] = "Contrib. Soc. Res. Meteorites",
  ["contributions to algebra and geometry"] = "Contrib. Algebra Geom.",
  ["contributions to applied and mathematical statistics"] = "Contrib. Appl. Math. Stat.",
  ["contributions to asian studies"] = "Contrib Asian Stud",
  ["contributions to atmospheric physics"] = "Contrib. Atmos. Phys.",
  ["contributions to discrete mathematics"] = "Contrib. Discrete Math.",
  ["contributions to economic analysis"] = "Contrib. Econom. Anal.",
  ["contributions to economic analysis & policy"] = "Contrib Econ Analysis Policy",
  ["contributions to economics"] = "Contrib. Econ.",
  ["contributions to embryology"] = "Contrib Embryol",
  ["contributions to entomology"] = "Contrib. Entomol.",
  ["contributions to epidemiology and biostatistics"] = "Contrib Epidemiol Biostat",
  ["contributions to geophysics & geodesy"] = "Contrib. Geophys. Geod.",
  ["contributions to gynecology and obstetrics"] = "Contrib Gynecol Obstet",
  ["contributions to human development"] = "Contrib Hum Dev",
  ["contributions to indian sociology"] = "Contrib Indian Sociol",
  ["contributions to indian sociology / ecole pratique des hautes etudes (vio section)-paris [and] institute of social anthropology-oxford"] = "Contrib Indian Sociol",
  ["contributions to management science"] = "Contrib. Manag. Sci.",
  ["contributions to microbiology"] = "Contrib Microbiol",
  ["contributions to microbiology and immunology"] = "Contrib Microbiol Immunol",
  ["contributions to mineralogy and petrology"] = "Contrib. Mineral. Petrol.",
  ["contributions to mineralogy and petrology. beitrage zur mineralogie und petrologie"] = "Contrib Mineral Petrol",
  ["contributions to mineralogy and petrology. beiträge zur mineralogie und petrologie"] = "Contrib Mineral Petrol",
  ["contributions to natural history"] = "Contrib Nat Hist",
  ["contributions to nepalese studies"] = "Contrib Nepalese Stud",
  ["contributions to nephrology"] = "Contrib Nephrol",
  ["contributions to phenomenology"] = "Contrib. Phenomenol.",
  ["contributions to plasma physics"] = "Contrib. Plasma Phys.",
  ["contributions to political economy"] = "Contrib Polit Econ",
  ["contributions to primatology"] = "Contrib Primatol",
  ["contributions to sensory physiology"] = "Contrib Sens Physiol",
  ["contributions to statistics"] = "Contrib. Statist.",
  ["contributions to zoology"] = "Contrib. Zool.",
  ["contributions to zoology (amsterdam, netherlands : 1995)"] = "Contrib Zool",
  ["control & automation"] = "Control Autom.",
  ["control and cybernetics"] = "Control Cybernet.",
  ["control and dynamic systems"] = "Control Dyn. Syst.",
  ["control and intelligent systems"] = "Control Intell. Syst.",
  ["control de plagas"] = "Control Plagas",
  ["control engineering"] = "Control Eng.",
  ["control engineering practice"] = "Control Eng Pract",
  ["control theory & applications"] = "Control Theory Appl.",
  ["control theory and technology"] = "Control Theory Technol.",
  ["control, robotics and sensors series"] = "Control Robot. Sens. Ser.",
  ["controlled clinical trials"] = "Control Clin Trials",
  ["controlled release newsletter"] = "Controll Release Newsl",
  ["convegni di studi sulla magna grecia"] = "CMGr",
  ["convergence (london, england)"] = "Convergence (Lond)",
  ["convergent science physical oncology"] = "Converg Sci Phys Oncol",
  ["convivium (são paulo, brazil)"] = "Convivium",
  ["convulsive therapy"] = "Convuls Ther",
  ["coo [reports]. u.s. atomic energy commission"] = "COO Rep",
  ["coo reports"] = "COO Rep.",
  ["cooley law review"] = "Cooley Law Rev",
  ["cooley law review / the thomas m. cooley law school"] = "Cooley Law Rev",
  ["cooper rowan medical journal"] = "Cooper Rowan Med J",
  ["cooperador dental"] = "Coop. Dent. (B. Aires)",
  ["cooperation and conflict"] = "Coop Confl",
  ["coordination chemistry reviews"] = "Coord. Chem. Rev.",
  ["coordinators' notebook : an international resource for early childhood development"] = "Coord Noteb",
  ["coordinators’ notebook : an international resource for early childhood development"] = "Coord Noteb",
  ["copernicus books"] = "Copernic. Books",
  ["cor et vasa"] = "Cor Vasa",
  ["coral reefs"] = "Coral Reefs",
  ["coral reefs (online)"] = "Coral Reefs",
  ["corduba archaeologica"] = "Corduba",
  ["core books in advanced mathematics"] = "Core Books in Adv. Math.",
  ["core evidence"] = "Core Evid",
  ["core journals in pediatrics"] = "Core J Pediatr",
  ["corinth. results of excavations conducted by the american school of classical studies at athens"] = "Corinth",
  ["cornell east asia series"] = "Cornell East Asia Ser.",
  ["cornell hotel and restaurant administration quarterly"] = "Cornell Hotel Restaur. Adm. Q.",
  ["cornell international law journal"] = "Cornell Int Law J",
  ["cornell journal of law and public policy"] = "Cornell J Law Public Policy",
  ["cornell law review"] = "Cornell Law Rev",
  ["cornell magazine (ithaca, n.y. : 1993)"] = "Cornell Mag",
  ["cornell medical journal"] = "Cornell Med J",
  ["cornell plantations"] = "Cornell Plant.",
  ["cornell veterinarian"] = "Cornell Vet.",
  ["corolla londiniensis"] = "CL",
  ["coronary artery disease"] = "Coron Artery Dis",
  ["corporate commentary"] = "Corp Comment",
  ["corporate social responsibility and environmental management"] = "Corp Soc Responsib Environ Manag",
  ["corps medical"] = "Corps Med. (Ettelbruck)",
  ["corpus antiquitatum aegyptiacarum"] = "CAA",
  ["corpus der minoischen und mykenischen siegel"] = "CMS",
  ["corpus inscriptionum atticarum"] = "CIA",
  ["corpus inscriptionum etruscarum"] = "CIE",
  ["corpus inscriptionum graecarum"] = "CIG",
  ["corpus inscriptionum latinarum"] = "CIL",
  ["corpus inscriptionum semiticarum"] = "CIS",
  ["corpus inscriptionum semiticarum. pars quarta. inscriptiones himyariticas et sabaeas continens"] = "CIH",
  ["corpus medicorum graecorum"] = "Corpus Med Graecorum",
  ["corpus medicorum graecorum. supplementum"] = "Corpus Med Graecorum Suppl Lipsiae",
  ["corpus pragmatics : international journal of corpus linguistics and pragmatics"] = "Corpus Pragmat",
  ["corpus signorum imperii romani"] = "CSIR",
  ["corpus speculorum etruscorum"] = "CSE",
  ["corpus vasorum antiquorum"] = "CVA",
  ["corrections : policy, practice and research"] = "Corrections",
  ["corrections compendium"] = "Correct Compend",
  ["corrections today"] = "Correct Today",
  ["corrective and social psychiatry and journal of behavior technology methods and therapy"] = "Correct Soc Psych J Behav Tech Methods Ther",
  ["correo poblacional y de la salud"] = "Correo Poblac Salud",
  ["correo poblacional y de la salud / centro de estudios de poblacion y paternidad responsable"] = "Correo Poblac Salud",
  ["corrosion & materials"] = "Corros. Mater.",
  ["corrosion (houston, tx, united states)"] = "Corrosion (Houston, TX, U. S.)",
  ["corrosion communications"] = "Corros. Commun.",
  ["corrosion engineering science and technology"] = "Corros. Eng. Sci. Technol.",
  ["corrosion engineering, science and technology"] = "Corros. Eng., Sci. Technol.",
  ["corrosion reviews"] = "Corros. Rev.",
  ["corrosion science"] = "Corros. Sci.",
  ["cortex; a journal devoted to the study of the nervous system and behavior"] = "Cortex",
  ["cosmic research"] = "Cosmic Res.",
  ["cospar colloquia series"] = "COSPAR Colloq. Ser.",
  ["cospar information bulletin"] = "COSPAR Inf. Bull.",
  ["cost & quality : cq"] = "Cost Qual",
  ["cost & quality quarterly journal : cq"] = "Cost Qual Q J",
  ["cost and quality quarterly journal"] = "Cost Qual. Q. J.",
  ["cost containment"] = "Cost Containment",
  ["cost effectiveness and resource allocation : c/e"] = "Cost Eff Resour Alloc",
  ["cota zero. revista d’arqueologia i ciencia"] = "CZero",
  ["cough (london, england)"] = "Cough",
  ["council on undergraduate research quarterly"] = "Counc Undergrad Res Q",
  ["counseling and values"] = "Couns Values",
  ["counseling et spiritualité = counselling and spirituality"] = "Couns Spiritual",
  ["counselling and psychotherapy research"] = "Couns Psychother Res",
  ["counselling psychology quarterly"] = "Couns Psychol Q",
  ["counselor (deerfield beach, fla.)"] = "Counselor (Deerfield Beach)",
  ["countdown (athens, ohio)"] = "Countdown (Athens, Ohio)",
  ["countdown to istanbul"] = "Countdown Istanb",
  ["country demographic profiles"] = "Ctry Demogr Profiles",
  ["country life (london, england)"] = "Ctry Life",
  ["country profiles"] = "Ctry Profiles",
  ["couple & family psychology"] = "Couple Family Psychol",
  ["courant lecture notes in mathematics"] = "Courant Lect. Notes Math.",
  ["courier (paris, france : 1984)"] = "Cour UNESCO",
  ["courrier de l’environnement de l’inra, le"] = "Courr. environ. INRA",
  ["courrier de l’exploitant et du scieur"] = "Courr. exploit. scieur",
  ["cours de géométrie de la faculté des sciences"] = "Cours Géom. Fac. Sci.",
  ["cours specialises"] = "Cours Spec.",
  ["cours spécialisés"] = "Cours Spéc.",
  ["courts, health science & the law"] = "Courts Health Sci Law",
  ["courts, health science & the law / georgetown university medical & law centers"] = "Courts Health Sci Law",
  ["covertaction quarterly"] = "Covertaction Q",
  ["coyuntura demográfica"] = "Coyunt Demogr",
  ["coyuntura economica (bogota, colombia)"] = "Coyunt Econ",
  ["coyuntura económica (bogotá, colombia)"] = "Coyunt Econ",
  ["cpc research review"] = "CPC Res Rev",
  ["cphc/bcs distinguished dissertations"] = "CPHC/BCS Disting. Diss.",
  ["cpj : canadian pharmaceutical journal = rpc : la revue pharmaceutique canadienne"] = "CPJ",
  ["cpp chemical plants and processing"] = "CPP Chem. Plants and Process.",
  ["cpq cancer"] = "CPQ Cancer",
  ["cpq neurology and psychology"] = "CPQ Neurol Psychol",
  ["cpr population research"] = "Cpr Popul Res",
  ["cpss power electronics series"] = "CPSS Power Electron. Ser.",
  ["cpss transactions on power electronics and applications"] = "CPSS Trans. Power Electron. Appl.",
  ["cpt: pharmacometrics & systems pharmacology"] = "CPT Pharmacometrics Syst Pharmacol",
  ["cranio : the journal of craniomandibular practice"] = "Cranio",
  ["cranio clinics international"] = "Cranio Clin Int",
  ["craniofacial growth series"] = "Craniofac Growth Ser",
  ["craniomaxillofacial trauma & reconstruction"] = "Craniomaxillofac Trauma Reconstr",
  ["crc critical reviews in biochemistry"] = "CRC Crit Rev Biochem",
  ["crc critical reviews in bioengineering"] = "CRC Crit Rev Bioeng",
  ["crc critical reviews in clinical laboratory sciences"] = "CRC Crit Rev Clin Lab Sci",
  ["crc critical reviews in clinical neurobiology"] = "CRC Crit Rev Clin Neurobiol",
  ["crc critical reviews in clinical radiology and nuclear medicine"] = "CRC Crit Rev Clin Radiol Nucl Med",
  ["crc critical reviews in diagnostic imaging"] = "CRC Crit Rev Diagn Imaging",
  ["crc critical reviews in food science and nutrition"] = "CRC Crit Rev Food Sci Nutr",
  ["crc critical reviews in immunology"] = "CRC Crit Rev Immunol",
  ["crc critical reviews in microbiology"] = "CRC Crit Rev Microbiol",
  ["crc critical reviews in radiological sciences"] = "CRC Crit Rev Radiol Sci",
  ["crc critical reviews in toxicology"] = "CRC Crit Rev Toxicol",
  ["crc mathematical modelling series"] = "CRC Math. Model. Ser.",
  ["crc press series on discrete mathematics and its applications"] = "CRC Press Ser. Discrete Math. Appl.",
  ["crc press/chapman and hall handbooks in mathematics series"] = "CRC Press/Chapman Hall Handb. Math. Ser.",
  ["crc series in applied and computational mechanics"] = "CRC Ser. Applied Comput. Mech.",
  ["crc series in computational mechanics and applied analysis"] = "CRC Ser. Comput. Mech. Appl. Anal.",
  ["creative education"] = "Creat Educ",
  ["creative mathematics and informatics"] = "Creat. Math. Inform.",
  ["creative nursing"] = "Creat Nurs",
  ["creativity research journal"] = "Creat Res J",
  ["creces / conin"] = "Creces",
  ["creighton law review"] = "Creighton Law Rev",
  ["creta antica. rivista annuale di studi archeologici, storici ed epigrafici"] = "CretAnt",
  ["cretaceous research"] = "Cretaceous Res.",
  ["cretan studies"] = "CretSt",
  ["crhcs news : the newsletter of the commonwealth regional health community secretariat for east, central and southern africa. commonwealth regional health community for east, central, and southern africa. secretariat"] = "CRHCS news",
  ["crime and delinquency"] = "Crime Delinq",
  ["crime and justice (chicago, ill.)"] = "Crime Justice",
  ["crime and social justice"] = "Crime Soc Justice",
  ["crime science"] = "Crime Sci",
  ["crime, histoire & sociétés = crime, history & societies"] = "Crime Hist Soc",
  ["crime, law, and social change"] = "Crime Law Soc Change",
  ["crime, media, culture"] = "Crime Media Cult",
  ["criminal behaviour and mental health : cbmh"] = "Crim Behav Ment Health",
  ["criminal justice and behavior"] = "Crim Justice Behav",
  ["criminal justice ethics"] = "Crim Justice Ethics",
  ["criminal justice history"] = "Crim Justice Hist",
  ["criminal justice journal"] = "Crim Justice J",
  ["criminal justice policy review"] = "Crim Justice Policy Rev",
  ["criminal justice review"] = "Crim Justice Rev",
  ["criminal justice studies (abingdon, england)"] = "Crim Justice Stud (Abingdon)",
  ["criminal law and philosophy"] = "Crim Law Philos",
  ["criminal law bulletin"] = "Crim Law Bull",
  ["criminal law forum"] = "Crim Law Forum",
  ["criminal law journal"] = "Crim Law J",
  ["criminal law review (london, england)"] = "Crim Law Rev",
  ["criminal reports [canada] new series, annotated; a series of reports with annotations and practice notes on criminal cases arising in the courts of the various provinces in canada"] = "Crim Rep Can New Ser",
  ["criminology & criminal justice : the international journal of policy and practice"] = "Criminol Crim Justice",
  ["criminology & public policy"] = "Criminol Public Policy",
  ["criminology : an interdisciplinary journal"] = "Criminology",
  ["crispr journal"] = "CRISPR J.",
  ["cristianesimo nella storia: richerche storiche, esegetiche, teologiche"] = "CrSt",
  ["critica storica"] = "Crit Stor",
  ["critica; revista hispanoamericana de filosofia"] = "Critica",
  ["critical african studies"] = "Crit Afr Stud",
  ["critical and radical social work"] = "Crit Radic Soc Work",
  ["critical arts"] = "Crit Arts",
  ["critical asian studies"] = "Crit Asian Stud",
  ["critical care"] = "Crit Care",
  ["critical care (houten, netherlands)"] = "Crit Care (Houten)",
  ["critical care (london, england)"] = "Crit Care",
  ["critical care and resuscitation : journal of the australasian academy of critical care medicine"] = "Crit Care Resusc",
  ["critical care and shock"] = "Crit Care Shock",
  ["critical care clinics"] = "Crit Care Clin",
  ["critical care explorations"] = "Crit Care Explor",
  ["critical care medicine"] = "Crit Care Med",
  ["critical care nurse"] = "Crit Care Nurse",
  ["critical care nursing clinics of north america"] = "Crit Care Nurs Clin North Am",
  ["critical care nursing quarterly"] = "Crit Care Nurs Q",
  ["critical care research and practice"] = "Crit Care Res Pract",
  ["critical care update"] = "Crit Care Update",
  ["critical connections : the complete news source for critical care professionals"] = "Crit Connect",
  ["critical criminology"] = "Crit Criminol",
  ["critical inquiry"] = "Crit Inq",
  ["critical insights in aquaculture"] = "Crit. Insights Aquacult.",
  ["critical insights in climate change"] = "Crit. Insights Clim. Change",
  ["critical insights in environmental science and technology"] = "Crit. Insights Environ. Sci. Technol.",
  ["critical insights in geochemistry and geophysics"] = "Crit. Insights Geochem. Geophys.",
  ["critical military studies"] = "Crit Mil Stud",
  ["critical path aids project"] = "Crit Path AIDS Proj",
  ["critical pathways in cardiology"] = "Crit Pathw Cardiol",
  ["critical policy studies"] = "Crit Policy Stud",
  ["critical public health"] = "Crit Public Health",
  ["critical review"] = "Critic. Rev.",
  ["critical review (new york, n.y.)"] = "Crit Rev (N Y)",
  ["critical review of international social and political philosophy"] = "Crit Rev Int Soc Political Philos",
  ["critical reviews in analytical chemistry"] = "Crit Rev Anal Chem",
  ["critical reviews in biochemistry and molecular biology"] = "Crit Rev Biochem Mol Biol",
  ["critical reviews in bioengineering"] = "Crit Rev Bioeng",
  ["critical reviews in biomedical engineering"] = "Crit Rev Biomed Eng",
  ["critical reviews in biotechnology"] = "Crit Rev Biotechnol",
  ["critical reviews in bone and mineral metabolism"] = "Crit. Rev. Bone Miner. Metab.",
  ["critical reviews in clinical laboratory sciences"] = "Crit Rev Clin Lab Sci",
  ["critical reviews in computed tomography"] = "Crit Rev Comput Tomogr",
  ["critical reviews in computed tomography7"] = "Crit. Rev. Comput. Tomogr.",
  ["critical reviews in critical laboratory sciences"] = "Crit. Rev. Clin. Lab. Sci.",
  ["critical reviews in diagnostic imaging"] = "Crit Rev Diagn Imaging",
  ["critical reviews in environment science and technology"] = "Crit. Rev. Env. Sci. Technol.",
  ["critical reviews in environmental science and technology"] = "Crit Rev Environ Sci Technol",
  ["critical reviews in eukaryotic gene expression"] = "Crit Rev Eukaryot Gene Expr",
  ["critical reviews in food science and nutrition"] = "Crit Rev Food Sci Nutr",
  ["critical reviews in immunology"] = "Crit Rev Immunol",
  ["critical reviews in medical informatics"] = "Crit Rev Med Inform",
  ["critical reviews in microbiology"] = "Crit Rev Microbiol",
  ["critical reviews in neurobiology"] = "Crit Rev Neurobiol",
  ["critical reviews in neurosurgery : cr"] = "Crit Rev Neurosurg",
  ["critical reviews in oncogenesis"] = "Crit Rev Oncog",
  ["critical reviews in oncology/hematology"] = "Crit Rev Oncol Hematol",
  ["critical reviews in oral biology & medicine"] = "Crit. Rev. Oral Biol. Med.",
  ["critical reviews in oral biology and medicine"] = "Crit. Rev. Oral Biol. Med.",
  ["critical reviews in oral biology and medicine : an official publication of the american association of oral biologists"] = "Crit Rev Oral Biol Med",
  ["critical reviews in physical and rehabilitation medicine"] = "Crit Rev Phys Rehabil Med",
  ["critical reviews in plant sciences"] = "CRC Crit Rev Plant Sci",
  ["critical reviews in solid state and materials sciences"] = "Crit. Rev. Solid State Mater. Sci.",
  ["critical reviews in therapeutic drug carrier systems"] = "Crit Rev Ther Drug Carrier Syst",
  ["critical reviews in toxicology"] = "Crit Rev Toxicol",
  ["critical social policy"] = "Crit Soc Policy",
  ["critical social research"] = "Crit Soc Res",
  ["critical sociology"] = "Crit Sociol (Eugene)",
  ["critical studies in mass communication : csmc : a publication of the speech communication association"] = "Crit Stud Mass Commun",
  ["critical studies in media communication"] = "Crit Stud Media Commun",
  ["critical studies on security"] = "Crit Stud Secur",
  ["critical studies on terrorism"] = "Crit Stud Terror",
  ["critical ultrasound journal"] = "Crit Ultrasound J",
  ["critique (clandeboye, man.)"] = "Critique",
  ["critique of anthropology"] = "Crit Anthropol",
  ["crítica; revista hispanoamericana de filosofía"] = "Critica",
  ["crm advocate"] = "CRM Advocate",
  ["crm monograph series"] = "CRM Monogr. Ser.",
  ["crm proceedings & lecture notes"] = "CRM Proc. Lecture Notes",
  ["crm series in mathematical physics"] = "CRM Ser. Math. Phys.",
  ["crna : the clinical forum for nurse anesthetists"] = "CRNA",
  ["croatian medical journal"] = "Croat Med J",
  ["croatian operational research review. crorr"] = "Croat. Oper. Res. Rev. CRORR",
  ["croatica chemica acta"] = "Croat. Chem. Acta",
  ["croatica chemica acta. arhiv za kemiju"] = "Croat Chem Acta",
  ["crohn's & colitis 360"] = "Crohns Colitis 360",
  ["cronache dell'idi"] = "Cron IDI",
  ["cronache dell’idi"] = "Cron IDI",
  ["cronache di archeologia"] = "CronA",
  ["cronache ercolanesi"] = "CronErc",
  ["cronache ercolanesi. bollettino del centro internazionale per lo studio dei papiri ercolanesi"] = "CronErcol",
  ["cronache pompeiane"] = "Cron Pompeiane",
  ["cronos (valencia, spain)"] = "Cronos",
  ["cronos (valencia, spain)."] = "Cronos",
  ["croom helm series in geography and environment"] = "Croom Helm Ser. Geogr. Environment",
  ["crop & pasture science"] = "Crop Pasture Sci",
  ["crop and environment"] = "Crop Environ.",
  ["crop and pasture science"] = "Crop Pasture Sci.",
  ["crop breeding and applied biotechnology"] = "Crop Breed. Appl. Biotechnol.",
  ["crop breeding journal"] = "Crop Breed J",
  ["crop breeding, genetics and genomics"] = "Crop Breed Genet Genom",
  ["crop design"] = "Crop Des.",
  ["crop journal"] = "Crop J.",
  ["crop protection"] = "Crop Prot.",
  ["crop protection (guildford, surrey)"] = "Crop Prot",
  ["crop science"] = "Crop Sci",
  ["crop, forage & turfgrass management"] = "Crop Forage Turfgrass Manage.",
  ["crops & soils"] = "Crops Soils",
  ["cross currents"] = "Cross Curr",
  ["cross currents (ann arbor, mich.)"] = "Cross Curr (Ann Arbor MI)",
  ["cross-cultural research : official journal of the society for cross-cultural research"] = "Cross Cult Res",
  ["cross-reference on human resources management"] = "Crossref Hum Resour Manage",
  ["crossroads (de kalb, ill.)"] = "Crossroads (De Kalb)",
  ["crossroads. history of interactions across the silk routes"] = "Crossroads",
  ["crustacean issues"] = "Crustac Issues",
  ["cruz roja cubana"] = "Cruz Roja Cuba",
  ["cryo letters"] = "Cryo Letters",
  ["cryogenic materials series"] = "Cryogenic Materials Ser.",
  ["cryosphere discussions"] = "Cryosphere Discuss.",
  ["cryptogam biodiversity and assessment"] = "Cryptogr Biodivers Assess",
  ["cryptogamic botany"] = "Cryptogam. Bot.",
  ["cryptogamica helvetica"] = "Cryptogam Helv",
  ["cryptogamie, algologie"] = "Cryptogam. Algol.",
  ["cryptogamie, bryologie"] = "Cryptogam. Bryol.",
  ["cryptogamie, mycologie"] = "Cryptogam. Mycol.",
  ["cryptogamie. algologie"] = "Cryptogam Algol",
  ["cryptogamie. bryologie"] = "Cryptogam Bryol",
  ["cryptogamie. bryologie, lichénologie"] = "Cryptogam Bryol Lichenol",
  ["cryptogamie. mycologie"] = "Cryptogam Mycol",
  ["cryptography and communications"] = "Cryptogr. Commun.",
  ["cryptography and communications : discrete structures, boolean functions and sequences"] = "Cryptogr Commun",
  ["cryptography and communications. discrete structures, boolean functions and sequences"] = "Cryptogr. Commun.",
  ["crystal engineering"] = "Cryst. Eng.",
  ["crystal growth & design"] = "Cryst Growth Des",
  ["crystal growth and design"] = "Cryst. Growth Des.",
  ["crystal research and technology"] = "Cryst. Res. Technol.",
  ["crystallography reports"] = "Crystallogr. Rep.",
  ["crystallography reviews"] = "Crystallogr. Rev.",
  ["csa journal (denver, colo.)"] = "CSA J (Denver)",
  ["cscw : proceedings of the conference on computer-supported cooperative work. conference on computer-supported cooperative work"] = "CSCW Conf Comput Support Coop Work",
  ["csd communications"] = "CSD Commun.",
  ["csee journal of power and energy systems"] = "CSEE J. Power Energy Syst",
  ["csh protocols"] = "CSH Protoc",
  ["csiam transactions on applied mathematics"] = "CSIAM Trans. Appl. Math.",
  ["csli lecture notes"] = "CSLI Lecture Notes",
  ["ct li lun yu ying yong yan jiu"] = "CT Li Lun Yu Ying Yong Yan Jiu",
  ["ctm. classical topics in mathematics"] = "CTM. Class. Top. Math.",
  ["ctwatch quarterly : cyberinfrastructure technology watch"] = "CTWatch Q",
  ["cuaderni catanesi di studi classici e medievali"] = "CuadCat",
  ["cuaderno de investigación en la educación : serie especial de temas especiales. university of puerto rico. recinto de río piedras. centro de investigaciones educativas"] = "Cuad Investig Educ",
  ["cuadernos americanos"] = "Cuad Am",
  ["cuadernos aragoneses de historia de la medicina y de la ciencia. serie a, monografías"] = "Cuad Aragon Hist Med Cienc A",
  ["cuadernos complutenses de historia de la medicina y de la ciencia"] = "Cuad Complut Hist Med Cienc",
  ["cuadernos de arqueología de la universidad de navarra"] = "CuadNavarra",
  ["cuadernos de arqueología mediterránea"] = "CuadAMed",
  ["cuadernos de arquitectura romana"] = "CuadArquitRom",
  ["cuadernos de arquitectura y urbanismo"] = "Cuad Arquit Urban",
  ["cuadernos de bioética : revista oficial de la asociación española de bioética y ética médica"] = "Cuad Bioet",
  ["cuadernos de comunicacion amidep"] = "Cuad Comun AMIDEP",
  ["cuadernos de comunicación amidep"] = "Cuad Comun AMIDEP",
  ["cuadernos de economia"] = "Cuadernos Econ.",
  ["cuadernos de economia (santiago, chile)"] = "Cuad Econ",
  ["cuadernos de economia social"] = "Cuad Econ Soc",
  ["cuadernos de economía (santiago, chile)"] = "Cuad Econ",
  ["cuadernos de economía social"] = "Cuad Econ Soc",
  ["cuadernos de estudios gallegos"] = "Cuad Estud Gallegos",
  ["cuadernos de estudios gallegos / instituto padre sarmiento de estudios gallegos"] = "Cuad Estud Gallegos",
  ["cuadernos de farmacología"] = "Cuad Farmacol",
  ["cuadernos de filologia clasica"] = "Cuad Filol Cl",
  ["cuadernos de filología clásica"] = "Cuad Filol Cl",
  ["cuadernos de filología clásica. estudios griegos e indoeuropeos."] = "CFC(G)",
  ["cuadernos de filología clásica. estudios latinos."] = "CFC(L)",
  ["cuadernos de filología clásica. facultad de letras y filosofía, universidad de madrid"] = "CuadFilCl",
  ["cuadernos de historia (santiago, chile)"] = "Cuad Hist",
  ["cuadernos de historia contemporánea"] = "Cuad Hist Contemp",
  ["cuadernos de historia de espana / instituto de historia de la cultura espanola, medioeval y moderna"] = "Cuad Hist Esp",
  ["cuadernos de historia de españa"] = "Cuad Hist Esp",
  ["cuadernos de historia de la medicina espanola"] = "Cuad Hist Med Esp",
  ["cuadernos de historia de la medicina española"] = "Cuad Hist Med Esp",
  ["cuadernos de historia de la medicina española. monografías"] = "Cuad Hist Med Esp Monogr",
  ["cuadernos de historia de la salud publica"] = "Cuad Hist Salud Publica",
  ["cuadernos de historia de la salud pública"] = "Cuad Hist Salud Publica",
  ["cuadernos de historia del derecho (madrid, spain)"] = "Cuad Hist Derecho",
  ["cuadernos de historia moderna"] = "Cuad Hist Mod",
  ["cuadernos de investigacion historica"] = "Cuad Invest Hist",
  ["cuadernos de investigación histórica"] = "Cuad Invest Hist",
  ["cuadernos de investigación filológica"] = "CIF",
  ["cuadernos de la c.v.f."] = "Cuadernos de la C.V.F.",
  ["cuadernos de la uned"] = "Cuad. UNED",
  ["cuadernos de logica, epistemologia y lenguaje"] = "Cuad. Log. Epistemol. Leng.",
  ["cuadernos de nutricion"] = "Cuad Nutr",
  ["cuadernos de nutrición"] = "Cuad Nutr",
  ["cuadernos de prehistoria de la universidad de granada"] = "CuadGranada",
  ["cuadernos de prehistoria i arqueología"] = "CPAM",
  ["cuadernos de prehistoria y arqueología castellonense"] = "CuadCastellon",
  ["cuadernos de prehistoria y arqueología. universidad autónoma de madrid"] = "CuadPrehistA",
  ["cuadernos de trabajos de la escuela española de historia y arqueología en roma"] = "CuadRom",
  ["cuadernos del cendes"] = "Cuad CENDES",
  ["cuadernos del claeh"] = "Cuad CLAEH",
  ["cuadernos economicos de i.c.e."] = "Cuadernos Econ. I.C.E.",
  ["cuadernos emeritenses"] = "CE",
  ["cuadernos hispanicos de historia de la medicina y de la ciencia"] = "Cuad Hisp Hist Med Cienc",
  ["cuadernos hispanoamericanos"] = "Cuad Hispanoam",
  ["cuadernos hispánicos de historia de la medicina y de la ciencia"] = "Cuad Hisp Hist Med Cienc",
  ["cuadernos medicos-sociales"] = "Cuad Med Soc",
  ["cuadernos médicos y de divulgación científica"] = "Cuad Med Divulg Cient",
  ["cuadernos médicos-sociales"] = "Cuad Med Soc",
  ["cuadernos valencianos de historia de la medicina y de la ciencia"] = "Cuad Valencia Hist Med Cienc",
  ["cuaderns de prehistòria i arqueologia de castelló"] = "CPAC",
  ["cuban studies"] = "Cuban Stud",
  ["cubo. a mathematical journal"] = "Cubo",
  ["cuihua xuebao"] = "Cuihua Xuebao",
  ["culta bononia"] = "Culta Bononia",
  ["cultura (iași, romania)"] = "Cultura (Iași)",
  ["cultura e scuola"] = "Cult Sc",
  ["cultura medica"] = "Cult Med (Rome)",
  ["cultura médica"] = "Cult Med",
  ["cultural and social history : the journal of the social history society"] = "Cult Soc Hist",
  ["cultural anthropology : journal of the society for cultural anthropology"] = "Cult Anthropol",
  ["cultural approaches to parenting"] = "Cross. Cont.",
  ["cultural diversity & ethnic minority psychology"] = "Cultur Divers Ethnic Minor Psychol",
  ["cultural diversity and ethnic minority psychology"] = "Cultur. Divers. Ethnic Minor. Psychol.",
  ["cultural diversity and mental health"] = "Cult Divers Ment Health",
  ["cultural geographies"] = "Cult Geogr",
  ["cultural sociology"] = "Cult Sociol",
  ["cultural studies <=> critical methodologies"] = "Cult Stud Crit Methodol",
  ["cultural studies of science education"] = "Cult Stud Sci Educ",
  ["cultural survival quarterly"] = "Cult Surv Q",
  ["culture & psychology"] = "Cult Psychol",
  ["culture (canadian ethnology society)"] = "Culture",
  ["culture and brain"] = "Cult Brain",
  ["culture and civilization in the middle east"] = "Cult. Civiliz. Middle East",
  ["culture and religion"] = "Cult Relig",
  ["culture technique"] = "Cult Tech",
  ["culture, agriculture, food and environment"] = "Cult Agric Food Environ",
  ["culture, health & sexuality"] = "Cult Health Sex",
  ["culture, medicine and psychiatry"] = "Cult Med Psychiatry",
  ["culture, medicine, and psychiatry"] = "Cult. Med. Psychiatry",
  ["cultures et conflits = culturas y conflictos = cultures and conflicts"] = "Cult Confl",
  ["cultures et developpement / universite catholique de louvain"] = "Cult Dev",
  ["cultures et développement"] = "Cult Dev",
  ["cumberland county history. cumberland county historical society and hamilton library association"] = "Cumberl Cty Hist",
  ["cumberland law review"] = "Cumberland Law Rev",
  ["cumberland-samford law review"] = "Cumberland Samford Law Rev",
  ["cuore e circolazione"] = "Cuore Circ",
  ["curator : a quarterly publication of the american museum of natural history"] = "Curator (N Y)",
  ["curity research notes"] = "Curity Res Notes",
  ["curq on the web"] = "CURQ Web",
  ["current addiction reports"] = "Curr Addict Rep",
  ["current advances in ophthalmology"] = "Curr Adv Ophthalmol",
  ["current affairs bulletin"] = "Curr Aff Bull",
  ["current aging science"] = "Curr Aging Sci",
  ["current allergy and asthma reports"] = "Curr Allergy Asthma Rep",
  ["current allergy reports"] = "Curr Allergy Rep",
  ["current alternative energy"] = "Curr. Altern. Energy",
  ["current alzheimer research"] = "Curr Alzheimer Res",
  ["current anaesthesia and critical care"] = "Curr Anaesth Crit Care",
  ["current analytical chemistry"] = "Curr. Anal. Chem.",
  ["current anesthesiology reports"] = "Curr Anesthesiol Rep",
  ["current angiogenesis"] = "Curr Angiogenes",
  ["current anthropology"] = "Curr Anthropol",
  ["current applied materials"] = "Curr. Appl Mater.",
  ["current applied physics"] = "Curr. Appl Phys.",
  ["current applied physics : the official journal of the korean physical society"] = "Curr Appl Phys",
  ["current applied polymer science"] = "Curr. Appl. Polym. Sci.",
  ["current applied science and technology"] = "Curr Appl Sci Technol",
  ["current artificial intelligence"] = "Curr. Artif. Intell.",
  ["current atherosclerosis reports"] = "Curr Atheroscler Rep",
  ["current behavioral neuroscience reports"] = "Curr Behav Neurosci Rep",
  ["current bioactive compounds"] = "Curr. Bioact. Compd.",
  ["current bioinformatics"] = "Curr. Bioinf.",
  ["current biology"] = "Curr. Biol.",
  ["current biology : cb"] = "Curr Biol",
  ["current biomarker findings"] = "Curr Biomark Find",
  ["current biotechnology"] = "Curr Biotechnol",
  ["current bladder dysfunction reports"] = "Curr Bladder Dysfunct Rep",
  ["current breast cancer reports"] = "Curr Breast Cancer Rep",
  ["current cancer drug targets"] = "Curr Cancer Drug Targets",
  ["current cancer reports"] = "Curr Cancer Rep",
  ["current cancer therapy reviews"] = "Curr Cancer Ther Rev",
  ["current cardiology reports"] = "Curr Cardiol Rep",
  ["current cardiology reviews"] = "Curr Cardiol Rev",
  ["current cardiovascular imaging reports"] = "Curr Cardiovasc Imaging Rep",
  ["current cardiovascular risk reports"] = "Curr Cardiovasc Risk Rep",
  ["current catalysis"] = "Curr. Catal.",
  ["current cellular biochemistry"] = "Curr Cell Biochem",
  ["current chemical biology"] = "Curr Chem Biol",
  ["current chemical genomics"] = "Curr Chem Genomics",
  ["current chemical genomics and translational medicine"] = "Curr Chem Genom Transl Med",
  ["current chromatography"] = "Curr Chromatogr",
  ["current climate change reports"] = "Curr Clim Change Rep",
  ["current clinical microbiology reports"] = "Curr Clin Microbiol Rep",
  ["current clinical pharmacology"] = "Curr Clin Pharmacol",
  ["current clinical topics in infectious diseases"] = "Curr Clin Top Infect Dis",
  ["current colorectal cancer reports"] = "Curr Colorectal Cancer Rep",
  ["current communications in cell & molecular biology"] = "Curr Commun Cell Mol Biol",
  ["current computer-aided drug design"] = "Curr Comput Aided Drug Des",
  ["current concepts in hospital pharmacy management"] = "Curr Concepts Hosp Pharm Manage",
  ["current concepts in nutrition"] = "Curr Concepts Nutr",
  ["current concepts of cerebrovascular disease: stroke"] = "Curr Conc Cerebrovasc Dis Stroke",
  ["current contents. clinical medicine"] = "Curr Contents Clin Med",
  ["current controlled trials in cardiovascular medicine"] = "Curr Control Trials Cardiovasc Med",
  ["current dermatology reports"] = "Curr Dermatol Rep",
  ["current developmental disorders reports"] = "Curr Dev Disord Rep",
  ["current developments in mathematical sciences"] = "Curr. Dev. Math. Sci.",
  ["current developments in mathematics"] = "Curr. Dev. Math.",
  ["current developments in nutrition"] = "Curr Dev Nutr",
  ["current developments in psychopharmacology"] = "Curr Dev Psychopharmacol",
  ["current diabetes reports"] = "Curr Diab Rep",
  ["current diabetes reviews"] = "Curr Diabetes Rev",
  ["current diagnostic pathology"] = "Curr Diagn Pathol",
  ["current directions in autoimmunity"] = "Curr Dir Autoimmun",
  ["current directions in psychological science"] = "Curr Dir Psychol Sci",
  ["current directions in psychological science : a journal of the american psychological society"] = "Curr Dir Psychol Sci",
  ["current drug abuse reviews"] = "Curr. Drug Abuse Rev.",
  ["current drug delivery"] = "Curr Drug Deliv",
  ["current drug discovery technologies"] = "Curr Drug Discov Technol",
  ["current drug metabolism"] = "Curr Drug Metab",
  ["current drug research reviews"] = "Curr Drug Res Rev",
  ["current drug safety"] = "Curr Drug Saf",
  ["current drug targets"] = "Curr Drug Targets",
  ["current drug targets. cardiovascular & haematological disorders"] = "Curr Drug Targets Cardiovasc Haematol Disord",
  ["current drug targets. cns and neurological disorders"] = "Curr Drug Targets CNS Neurol Disord",
  ["current drug targets. immune, endocrine and metabolic disorders"] = "Curr Drug Targets Immune Endocr Metabol Disord",
  ["current drug targets. infectious disorders"] = "Curr Drug Targets Infect Disord",
  ["current drug targets. inflammation and allergy"] = "Curr Drug Targets Inflamm Allergy",
  ["current drug targets: cardiovascular & haematological disorders"] = "Curr. Drug Targets: Cardiovasc. & Haematol. Disord.",
  ["current drug targets: cns & neurological disorders"] = "Curr. Drug Targets: CNS Neurol. Disord.",
  ["current drug targets: immune, endocrine and metabolic disorders"] = "Curr. Drug Targets: Immune, Endocr. Metab. Disord.",
  ["current drug targets: infectious disorders"] = "Curr. Drug Targets: Infect. Disord.",
  ["current drug targets: inflammation & allergy"] = "Curr. Drug Targets: Inflammation Allergy",
  ["current drug therapy"] = "Curr Drug ther",
  ["current emergency and hospital medicine reports"] = "Curr Emerg Hosp Med Rep",
  ["current engineering letters"] = "Curr. Eng. Lett.",
  ["current engineering letters and reviews"] = "Curr. Eng. Lett. Rev.",
  ["current environmental health reports"] = "Curr Environ Health Rep",
  ["current enzyme inhibition"] = "Curr Enzym Inhib",
  ["current epidemiology reports"] = "Curr Epidemiol Rep",
  ["current eye research"] = "Curr Eye Res",
  ["current findings of infectious diseases"] = "Curr Find Infect Dis",
  ["current food science and technology reports"] = "Curr. Food Sci. Technol. Rep.",
  ["current forensic science"] = "Curr. Forensic Sci.",
  ["current forestry reports"] = "Curr. For. Rep.",
  ["current functional foods"] = "Curr. Funct. Foods",
  ["current fungal infection reports"] = "Curr Fungal Infect Rep",
  ["current gastroenterology reports"] = "Curr Gastroenterol Rep",
  ["current gene therapy"] = "Curr Gene Ther",
  ["current genetic medicine reports"] = "Curr Genet Med Rep",
  ["current genetics"] = "Curr Genet",
  ["current genomics"] = "Curr Genomics",
  ["current geriatrics reports"] = "Curr Geriatr Rep",
  ["current gerontology and geriatrics research"] = "Curr Gerontol Geriatr Res",
  ["current green chemistry"] = "Curr. Green Chem.",
  ["current health sciences journal"] = "Curr Health Sci J",
  ["current heart failure reports"] = "Curr Heart Fail Rep",
  ["current hematologic malignancy reports"] = "Curr Hematol Malig Rep",
  ["current hematology reports"] = "Curr Hematol Rep",
  ["current hepatitis reports"] = "Curr Hepat Rep",
  ["current hepatology reports"] = "Curr Hepatol Rep",
  ["current herpetology"] = "Curr Herpetol",
  ["current history (new york, n.y. : 1941)"] = "Curr Hist",
  ["current hiv research"] = "Curr HIV Res",
  ["current hiv/aids reports"] = "Curr HIV/AIDS Rep",
  ["current hypertension reports"] = "Curr Hypertens Rep",
  ["current hypertension reviews"] = "Curr Hypertens Rev",
  ["current immunology reviews"] = "Curr Immunol Rev",
  ["current infectious disease reports"] = "Curr Infect Dis Rep",
  ["current inorganic chemistry"] = "Curr. Inorg. Chem.",
  ["current interventional cardiology reports"] = "Curr Interv Cardiol Rep",
  ["current issues in education (tempe, ariz.)"] = "Curr Issues Educ (Tempe)",
  ["current issues in emerging elearning"] = "Curr Issues Emerg Elearn",
  ["current issues in intestinal microbiology"] = "Curr Issues Intest Microbiol",
  ["current issues in molecular biology"] = "Curr Issues Mol Biol",
  ["current issues in public health"] = "Curr Issues Public Health",
  ["current journal of neurology"] = "Curr J Neurol",
  ["current landscape ecology reports"] = "Curr. Landscape Ecol. Rep.",
  ["current law statutes annotated. great britain"] = "Curr Law Statut Annot GB",
  ["current legal problems"] = "Curr Leg Probl",
  ["current legal thought; the lawyers' digest of law reviews"] = "Curr Leg Thought Lawyers Dig Law Rev",
  ["current materials science"] = "Curr. Mater. Sci.",
  ["current mathematical publications"] = "CMP",
  ["current medical digest"] = "Curr Med Dig",
  ["current medical imaging"] = "Curr Med Imaging",
  ["current medical imaging reviews"] = "Curr Med Imaging Rev",
  ["current medical literature. cardiology"] = "Curr Med Lit Cardiol",
  ["current medical literature. dermatology"] = "Curr Med Lit Dermatol",
  ["current medical literature. rheumatology"] = "Curr Med Lit Rheumatol",
  ["current medical mycology"] = "Curr Med Mycol",
  ["current medical practice"] = "Curr Med Pract",
  ["current medical research and opinion"] = "Curr Med Res Opin",
  ["current medical science"] = "Curr Med Sci",
  ["current medicinal chemistry"] = "Curr Med Chem",
  ["current medicinal chemistry. anti-cancer agents"] = "Curr Med Chem Anticancer Agents",
  ["current medicinal chemistry. anti-inflammatory & anti-allergy agents"] = "Curr Med Chem Anti Inflamm Anti Allergy Agents",
  ["current medicinal chemistry. cardiovascular and hematological agents"] = "Curr Med Chem Cardiovasc Hematol Agents",
  ["current medicinal chemistry. immunology, endocrine & metabolic agents"] = "Curr Med Chem Immunol Endocr Metab Agents",
  ["current medicinal chemistry: anti-cancer agents"] = "Curr. Med. Chem.: Anti-Cancer Agents",
  ["current medicinal chemistry: anti-infective agents"] = "Curr. Med. Chem.: Anti-Infect. Agents",
  ["current medicinal chemistry: anti-inflammatory & anti-allergy agents"] = "Curr. Med. Chem.: Anti-Inflammatory Anti-Allergy Agents",
  ["current medicinal chemistry: cardiovascular & hematological agents"] = "Curr. Med. Chem.: Cardiovasc. Hematol. Agents",
  ["current medicinal chemistry: central nervous system agents"] = "Curr. Med. Chem.: Cent. Nerv. Syst. Agents",
  ["current medicinal chemistry: immunology, endocrine & metabolic agents"] = "Curr. Med. Chem.: Immunol., Endocr. Metab. Agents",
  ["current medicine and drugs"] = "Curr Med Drugs",
  ["current medicine for attorneys"] = "Curr Med Atty",
  ["current medicine research and practice"] = "Curr Med Res Pract",
  ["current metabolomics"] = "Curr Metabolomics",
  ["current microbiology"] = "Curr Microbiol",
  ["current microscopy reports"] = "Curr. Microsc. Rep.",
  ["current microwave chemistry"] = "Curr. Microwave Chem.",
  ["current molecular biology reports"] = "Curr Mol Biol Rep",
  ["current molecular imaging"] = "Curr Mol Imaging",
  ["current molecular medicine"] = "Curr Mol Med",
  ["current molecular pharmacology"] = "Curr Mol Pharmacol",
  ["current nanomaterials"] = "Curr. Nanomater.",
  ["current nanomedicine"] = "Curr. Nanomed.",
  ["current nanoscience"] = "Curr Nanosci",
  ["current nanotoxicity and prevention"] = "Curr. Nanotoxic. Prev.",
  ["current natural sciences"] = "Curr. Nat. Sci.",
  ["current neurobiology"] = "Curr Neurobiol",
  ["current neurology and neuroscience reports"] = "Curr Neurol Neurosci Rep",
  ["current neuropharmacology"] = "Curr Neuropharmacol",
  ["current neurovascular research"] = "Curr Neurovasc Res",
  ["current nutrition & food science"] = "Curr. Nutr. Food Sci.",
  ["current nutrition and food science"] = "Curr Nutr Food Sci",
  ["current nutrition reports"] = "Curr Nutr Rep",
  ["current obesity reports"] = "Curr Obes Rep",
  ["current obstetrics & gynaecology"] = "Curr Obstet Gynaecol",
  ["current obstetrics and gynecology reports"] = "Curr Obstet Gynecol Rep",
  ["current oncology (toronto, ont.)"] = "Curr Oncol",
  ["current oncology reports"] = "Curr Oncol Rep",
  ["current ophthalmology reports"] = "Curr Ophthalmol Rep",
  ["current opinion in allergy and clinical immunology"] = "Curr Opin Allergy Clin Immunol",
  ["current opinion in anaesthesiology"] = "Curr Opin Anaesthesiol",
  ["current opinion in behavioral sciences"] = "Curr Opin Behav Sci",
  ["current opinion in biomedical engineering"] = "Curr Opin Biomed Eng",
  ["current opinion in biotechnology"] = "Curr Opin Biotechnol",
  ["current opinion in cardiology"] = "Curr Opin Cardiol",
  ["current opinion in cell biology"] = "Curr Opin Cell Biol",
  ["current opinion in chemical biology"] = "Curr Opin Chem Biol",
  ["current opinion in chemical engineering"] = "Curr Opin Chem Eng",
  ["current opinion in clinical nutrition & metabolic care"] = "Curr. Opin. Clin. Nutr. Metab. Care",
  ["current opinion in clinical nutrition and metabolic care"] = "Curr Opin Clin Nutr Metab Care",
  ["current opinion in colloid & interface science"] = "Curr Opin Colloid Interface Sci",
  ["current opinion in colloid and interface science"] = "Curr. Opin. Colloid Interface Sci.",
  ["current opinion in cosmetic dentistry"] = "Curr Opin Cosmet Dent",
  ["current opinion in critical care"] = "Curr Opin Crit Care",
  ["current opinion in dentistry"] = "Curr Opin Dent",
  ["current opinion in drug discovery & development"] = "Curr Opin Drug Discov Devel",
  ["current opinion in drug discovery and development"] = "Curr. Opin. Drug Discov. Devel.",
  ["current opinion in electrochemistry"] = "Curr Opin Electrochem",
  ["current opinion in endocrine and metabolic research"] = "Curr Opin Endocr Metab Res",
  ["current opinion in endocrinology & diabetes"] = "Curr Opin Endocrinol Diabetes",
  ["current opinion in endocrinology, diabetes, and obesity"] = "Curr Opin Endocrinol Diabetes Obes",
  ["current opinion in environment sustainability"] = "Curr. Opin. Environ. Sustainability",
  ["current opinion in environmental science & health"] = "Curr Opin Environ Sci Health",
  ["current opinion in environmental sustainability"] = "Curr Opin Environ Sustain",
  ["current opinion in food science"] = "Curr. Opin. Food Sci.",
  ["current opinion in gastroenterology"] = "Curr Opin Gastroenterol",
  ["current opinion in general surgery"] = "Curr Opin Gen Surg",
  ["current opinion in genetics & development"] = "Curr Opin Genet Dev",
  ["current opinion in genetics and development"] = "Curr. Opin. Genet. Dev.",
  ["current opinion in green and sustainable chemistry"] = "Curr Opin Green Sustain Chem",
  ["current opinion in gynecology and obstetrics"] = "Curr Opin Gynecol Obstet",
  ["current opinion in hematology"] = "Curr Opin Hematol",
  ["current opinion in hiv and aids"] = "Curr Opin HIV AIDS",
  ["current opinion in immunology"] = "Curr Opin Immunol",
  ["current opinion in infectious diseases"] = "Curr Opin Infect Dis",
  ["current opinion in insect science"] = "Curr Opin Insect Sci",
  ["current opinion in investigational drugs"] = "Curr. Opin. Investig. Drugs",
  ["current opinion in investigational drugs (london, england : 2000)"] = "Curr Opin Investig Drugs",
  ["current opinion in lipidology"] = "Curr Opin Lipidol",
  ["current opinion in microbiology"] = "Curr Opin Microbiol",
  ["current opinion in molecular therapeutics"] = "Curr. Opin. Mol. Ther.",
  ["current opinion in nephrology and hypertension"] = "Curr Opin Nephrol Hypertens",
  ["current opinion in neurobiology"] = "Curr Opin Neurobiol",
  ["current opinion in neurology"] = "Curr Opin Neurol",
  ["current opinion in neurology and neurosurgery"] = "Curr Opin Neurol Neurosurg",
  ["current opinion in obstetrics & gynecology"] = "Curr Opin Obstet Gynecol",
  ["current opinion in obstetrics and gynecology"] = "Curr. Opin. Obstet. Gynecol.",
  ["current opinion in oncology"] = "Curr Opin Oncol",
  ["current opinion in ophthalmology"] = "Curr Opin Ophthalmol",
  ["current opinion in organ transplantation"] = "Curr Opin Organ Transplant",
  ["current opinion in orthopaedics"] = "Curr Opin Orthop",
  ["current opinion in otolaryngology & head and neck surgery"] = "Curr Opin Otolaryngol Head Neck Surg",
  ["current opinion in pediatrics"] = "Curr Opin Pediatr",
  ["current opinion in periodontology"] = "Curr Opin Periodontol",
  ["current opinion in pharmacology"] = "Curr Opin Pharmacol",
  ["current opinion in physiology"] = "Curr Opin Physiol",
  ["current opinion in plant biology"] = "Curr Opin Plant Biol",
  ["current opinion in psychiatry"] = "Curr Opin Psychiatry",
  ["current opinion in psychology"] = "Curr Opin Psychol",
  ["current opinion in pulmonary medicine"] = "Curr Opin Pulm Med",
  ["current opinion in radiology"] = "Curr Opin Radiol",
  ["current opinion in rheumatology"] = "Curr Opin Rheumatol",
  ["current opinion in solid state & materials science"] = "Curr Opin Solid State Mater Sci",
  ["current opinion in solid state and materials science"] = "Curr. Opin. Solid State Mater. Sci.",
  ["current opinion in structural biology"] = "Curr Opin Struct Biol",
  ["current opinion in supportive and palliative care"] = "Curr Opin Support Palliat Care",
  ["current opinion in systems biology"] = "Curr Opin Syst Biol",
  ["current opinion in toxicology"] = "Curr Opin Toxicol",
  ["current opinion in urology"] = "Curr Opin Urol",
  ["current opinion in virology"] = "Curr Opin Virol",
  ["current optics and photonics"] = "Curr. Opt. Photonics",
  ["current oral health reports"] = "Curr Oral Health Rep",
  ["current organic chemistry"] = "Curr. Org. Chem.",
  ["current organic synthesis"] = "Curr Org Synth",
  ["current organocatalysis"] = "Curr. Organocatal.",
  ["current orthopaedic practice"] = "Curr Orthop Pract",
  ["current orthopaedics"] = "Curr Orthop",
  ["current osteoporosis reports"] = "Curr Osteoporos Rep",
  ["current otorhinolaryngology reports"] = "Curr Otorhinolaryngol Rep",
  ["current pain and headache reports"] = "Curr Pain Headache Rep",
  ["current pathobiology reports"] = "Curr Pathobiol Rep",
  ["current pediatric reviews"] = "Curr Pediatr Rev",
  ["current pediatrics reports"] = "Curr Pediatr Rep",
  ["current pharmaceutical analysis"] = "Curr Pharm Anal",
  ["current pharmaceutical biotechnology"] = "Curr Pharm Biotechnol",
  ["current pharmaceutical design"] = "Curr Pharm Des",
  ["current pharmacogenomics"] = "Curr. Pharmacogenomics",
  ["current pharmacogenomics and personalized medicine"] = "Curr. Pharmacogenomics Pers. Med.",
  ["current pharmacology reports"] = "Curr Pharmacol Rep",
  ["current physical chemistry"] = "Curr Phys Chem",
  ["current physical medicine and rehabilitation reports"] = "Curr Phys Med Rehabil Rep",
  ["current physics"] = "Curr. Phys.",
  ["current plant biology"] = "Curr. Plant Biol.",
  ["current plant science and biotechnology in agriculture"] = "Curr Plant Sci Biotechnol Agric",
  ["current plant science in biotechnology and agriculture"] = "Curr. Plant Sci. Biotechnol. Agric.",
  ["current pollution reports"] = "Curr Pollut Rep",
  ["current population reports, series p-25. population estimates and projections"] = "Curr Popul Rep Popul Estim Proj",
  ["current population reports. series p-20, population characteristics"] = "Curr Popul Rep Popul Charact",
  ["current population reports. series p-23, technical studies"] = "Curr Popul Rep [23] Tech Stud",
  ["current population reports. series p-23. special studies"] = "Curr Popul Rep 23 Spec Stud",
  ["current population reports. series p-25, population estimates and projections"] = "Curr Popul Rep Popul Estim Proj",
  ["current population reports. series p-26, federal-state cooperative program for population estimates"] = "Curr Popul Rep [P-26]",
  ["current population reports. series p-26, local population estimates"] = "Curr Popul Rep P-26",
  ["current population reports. series p-26, population estimates and projections"] = "Curr Popul Rep P-26",
  ["current population reports. series p-27, farm population"] = "Cur Popul Rep Farm Popul",
  ["current population reports. series p-28, special censuses"] = "Curr Popul Rep Spec Censuses",
  ["current population reports. series p-60, consumer income"] = "Curr Popul Rep Consum Income",
  ["current population reports. series p-70, household economic studies"] = "Curr Popul Rep Ser.P-70 Househ Econ Stud",
  ["current population reports. special censuses. series p-28"] = "Curr. Popul. Rep. [Spec. Censuses]",
  ["current practice in gerontological nursing"] = "Curr Pract Gerontol Nurs",
  ["current practice in obstetric and gynecologic nursing"] = "Curr Pract Obstet Gynecol Nurs",
  ["current practice in orthopaedic surgery"] = "Curr Pract Orthop Surg",
  ["current practice in pediatric nursing"] = "Curr Pract Pediatr Nurs",
  ["current prescribing"] = "Curr Prescr",
  ["current probiotics"] = "Curr. Probiotics",
  ["current problems in cancer"] = "Curr Probl Cancer",
  ["current problems in cardiology"] = "Curr Probl Cardiol",
  ["current problems in clinical biochemistry"] = "Curr Probl Clin Biochem",
  ["current problems in dermatology"] = "Curr Probl Dermatol",
  ["current problems in diagnostic radiology"] = "Curr Probl Diagn Radiol",
  ["current problems in obstetrics and gynecology"] = "Curr Probl Obstet Gynecol",
  ["current problems in obstetrics, gynecology and fertility"] = "Curr Probl Obstet Gynecol Fertil",
  ["current problems in pediatric and adolescent health care"] = "Curr Probl Pediatr Adolesc Health Care",
  ["current problems in pediatrics"] = "Curr Probl Pediatr",
  ["current problems in surgery"] = "Curr Probl Surg",
  ["current problems of soviet medicine"] = "Curr Probl Sov Med",
  ["current prostate reports"] = "Curr Prostate Rep",
  ["current protein & peptide science"] = "Curr Protein Pept Sci",
  ["current protein and peptide science"] = "Curr. Protein Pept. Sci.",
  ["current proteomics"] = "Curr. Proteomics",
  ["current protocols"] = "Curr Protoc",
  ["current protocols essential laboratory techniques"] = "Curr Protoc Essent Lab Tech",
  ["current protocols in bioinformatics"] = "Curr Protoc Bioinformatics",
  ["current protocols in cell biology"] = "Curr Protoc Cell Biol",
  ["current protocols in chemical biology"] = "Curr Protoc Chem Biol",
  ["current protocols in cytometry"] = "Curr Protoc Cytom",
  ["current protocols in human genetics"] = "Curr Protoc Hum Genet",
  ["current protocols in immunology"] = "Curr Protoc Immunol",
  ["current protocols in microbiology"] = "Curr Protoc Microbiol",
  ["current protocols in molecular biology"] = "Curr Protoc Mol Biol",
  ["current protocols in mouse biology"] = "Curr Protoc Mouse Biol",
  ["current protocols in neuroscience"] = "Curr Protoc Neurosci",
  ["current protocols in nucleic acid chemistry"] = "Curr Protoc Nucleic Acid Chem",
  ["current protocols in pharmacology"] = "Curr Protoc Pharmacol",
  ["current protocols in plant biology"] = "Curr Protoc Plant Biol",
  ["current protocols in protein science"] = "Curr Protoc Protein Sci",
  ["current protocols in stem cell biology"] = "Curr Protoc Stem Cell Biol",
  ["current protocols in toxicology"] = "Curr Protoc Toxicol",
  ["current psychiatric therapies"] = "Curr Psychiatr Ther",
  ["current psychiatry"] = "Curr Psychiatr",
  ["current psychiatry reports"] = "Curr Psychiatry Rep",
  ["current psychiatry reviews"] = "Curr Psychiatry Rev",
  ["current psychology (new brunswick, n.j.)"] = "Curr Psychol",
  ["current psychology letters : behaviour, brain & cognition : cpl"] = "Curr Psychol Lett",
  ["current psychology of cognition = cahiers de psychologie cognitive : cpc"] = "Curr Psychol Cogn",
  ["current psychopharmacology"] = "Curr Psychopharmacol",
  ["current psychosis & therapeutics reports"] = "Curr Psychos Ther Rep",
  ["current pulmonology reports"] = "Curr Pulmonol Rep",
  ["current radiology reports"] = "Curr Radiol Rep",
  ["current radiopharmaceuticals"] = "Curr Radiopharm",
  ["current research in bacteriology"] = "Curr Res Bacteriol",
  ["current research in biotechnology"] = "Curr Res Biotechnol",
  ["current research in cell biology"] = "Curr Res Cell Biol",
  ["current research in chemical biology"] = "Curr. Res. Chem. Biol.",
  ["current research in diabetes & obesity journal"] = "Curr Res Diabetes Obes J",
  ["current research in ecological and social psychology"] = "Curr. Res. Ecol. Social Psychol.",
  ["current research in environmental & applied mycology. journal of fungal biology"] = "Curr Res Environ Appl Mycol J Fungal Biol",
  ["current research in environmental sustainability"] = "Curr. Res. Environ. Sustainability",
  ["current research in food science"] = "Curr Res Food Sci",
  ["current research in green and sustainable chemistry"] = "Curr. Res. Green Sustainable Chem.",
  ["current research in immunology"] = "Curr. Res. Immunol.",
  ["current research in insect science"] = "Curr. Res. Insect Sci.",
  ["current research in microbial sciences"] = "Curr Res Microb Sci",
  ["current research in microbiology and biotechnology"] = "Curr Res Microbiol Biotechnol",
  ["current research in neurobiology"] = "Curr. Res. Neurobiol.",
  ["current research in pharmacology and drug discovery"] = "Curr. Res. Pharmacol. Drug Discovery",
  ["current research in physiology"] = "Curr Res Physiol",
  ["current research in psychology"] = "Curr Res Psychol",
  ["current research in structural biology"] = "Curr Res Struct Biol",
  ["current research in toxicology"] = "Curr Res Toxicol",
  ["current research in translational medicine"] = "Curr Res Transl Med",
  ["current research in virological science"] = "Curr Res Virol Sci",
  ["current research on occupations and professions"] = "Curr Res Occup Prof",
  ["current researches in anesthesia & analgesia"] = "Curr Res Anesth Analg",
  ["current respiratory care reports"] = "Curr Respir Care Rep",
  ["current respiratory medicine reviews"] = "Curr Respir Med Rev",
  ["current review of pain"] = "Curr Rev Pain",
  ["current reviews in clinical and experimental pharmacology"] = "Curr. Rev. Clin. Exp. Pharmacol.",
  ["current reviews in musculoskeletal medicine"] = "Curr Rev Musculoskelet Med",
  ["current rheumatology reports"] = "Curr Rheumatol Rep",
  ["current rheumatology reviews"] = "Curr Rheumatol Rev",
  ["current robotics reports"] = "Curr. Rob. Rep.",
  ["current science"] = "Curr Sci",
  ["current science international"] = "Curr Sci Int",
  ["current separations"] = "Curr Sep",
  ["current separations and drug development"] = "Curr. Sep Drug Dev.",
  ["current sexual health reports"] = "Curr Sex Health Rep",
  ["current signal transduction therapy"] = "Curr Signal Transduct Ther",
  ["current sleep medicine reports"] = "Curr Sleep Med Rep",
  ["current sociology. la sociologie contemporaine"] = "Curr Sociol",
  ["current sociology. la sociologie contemporaine."] = "Curr Sociol",
  ["current sports medicine reports"] = "Curr Sports Med Rep",
  ["current stem cell reports"] = "Curr Stem Cell Rep",
  ["current stem cell research & therapy"] = "Curr Stem Cell Res Ther",
  ["current studies in hematology and blood transfusion"] = "Curr Stud Hematol Blood Transfus",
  ["current studies on the nature of brain function"] = "Curr Stud Nat Brain Funct",
  ["current surgery"] = "Curr Surg",
  ["current surgery reports"] = "Curr Surg Rep",
  ["current sustainable/renewable energy reports"] = "Curr. Sustainable/Renewable Energy Rep.",
  ["current swedish archaeology"] = "Curr. Swed. Archaeol.",
  ["current therapeutic research, clinical and experimental"] = "Curr Ther Res Clin Exp",
  ["current therapeutics"] = "Curr Ther (Seaforth)",
  ["current therapy in endocrinology and metabolism"] = "Curr Ther Endocrinol Metab",
  ["current tissue engineering"] = "Curr Tissue Eng",
  ["current tissue microenvironment reports"] = "Curr Tissue Microenviron Rep",
  ["current topics in behavioral neurosciences"] = "Curr Top Behav Neurosci",
  ["current topics in biochemical research"] = "Curr Top Biochem Res",
  ["current topics in cellular regulation"] = "Curr Top Cell Regul",
  ["current topics in comparative pathobiology"] = "Curr Top Comp Pathobiol",
  ["current topics in developmental biology"] = "Curr Top Dev Biol",
  ["current topics in electrochemistry"] = "Curr Top Electrochem",
  ["current topics in experimental endocrinology"] = "Curr Top Exp Endocrinol",
  ["current topics in eye research"] = "Curr Top Eye Res",
  ["current topics in genetics"] = "Curr Top Genet",
  ["current topics in hematology"] = "Curr Top Hematol",
  ["current topics in medical mycology"] = "Curr Top Med Mycol",
  ["current topics in medicinal chemistry"] = "Curr Top Med Chem",
  ["current topics in medicinal chemistry (sharjah, united arab emirates)"] = "Curr. Top. Med. Chem. (Sharjah, United Arab Emirates)",
  ["current topics in membranes"] = "Curr Top Membr",
  ["current topics in membranes and transport"] = "Curr Top Membr Transp",
  ["current topics in microbiology and immunology"] = "Curr Top Microbiol Immunol",
  ["current topics in molecular endocrinology"] = "Curr Top Mol Endocrinol",
  ["current topics in neuroendocrinology"] = "Curr Top Neuroendocrinol",
  ["current topics in nutraceutical research"] = "Curr Top Nutraceutical Res",
  ["current topics in pathology"] = "Curr. Top. Pathol.",
  ["current topics in pathology. ergebnisse der pathologie"] = "Curr Top Pathol",
  ["current topics in peptide & protein research"] = "Curr Top Pept Protein Res",
  ["current topics in pharmacology"] = "Curr Top Pharmacol",
  ["current topics in phytochemistry"] = "Curr Top Phytochem",
  ["current topics in plant biology"] = "Curr Top Plant Biol",
  ["current topics in plant physiology"] = "Curr Topics Plant Physiol",
  ["current topics in radiation research quarterly"] = "Curr Top Radiat Res Q",
  ["current topics in surgical research"] = "Curr Top Surg Res",
  ["current topics in toxicology"] = "Curr Top Toxicol",
  ["current topics in virology"] = "Curr Top Virol",
  ["current translational geriatrics and experimental gerontology reports"] = "Curr Transl Geriatr Exp Gerontol Rep",
  ["current transplantation reports"] = "Curr Transplant Rep",
  ["current trauma reports"] = "Curr Trauma Rep",
  ["current treatment options in allergy"] = "Curr Treat Options Allergy",
  ["current treatment options in cardiovascular medicine"] = "Curr Treat Options Cardiovasc Med",
  ["current treatment options in gastroenterology"] = "Curr Treat Options Gastroenterol",
  ["current treatment options in infectious diseases"] = "Curr Treat Options Infect Dis",
  ["current treatment options in neurology"] = "Curr Treat Options Neurol",
  ["current treatment options in oncology"] = "Curr Treat Options Oncol",
  ["current treatment options in pediatrics"] = "Curr Treat Options Pediatr",
  ["current treatment options in psychiatry"] = "Curr Treat Options Psychiatry",
  ["current treatment options in rheumatology"] = "Curr Treatm Opt Rheumatol",
  ["current trends in biomedical engineering & biosciences"] = "Curr Trends Biomed Eng Biosci",
  ["current trends in biotechnology and pharmacy"] = "Curr Trends Biotechnol Pharm",
  ["current trends in clinical & medical imaging"] = "Curr Trends Clin Med Imaging",
  ["current trends in endocrinology"] = "Curr Trends Endocinol",
  ["current trends in immunology"] = "Curr Trends Immunol",
  ["current trends in medical diagnostic methods"] = "Curr Trends Med Diagn Methods",
  ["current trends in medicinal chemistry"] = "Curr Trends Med Chem",
  ["current trends in microbiology"] = "Curr Trends Microbiol",
  ["current trends in neurology"] = "Curr Trends Neurol",
  ["current trends in polymer science"] = "Curr Trends Polym Sci",
  ["current tropical medicine reports"] = "Curr Trop Med Rep",
  ["current urology"] = "Curr Urol",
  ["current urology reports"] = "Curr Urol Rep",
  ["current vascular pharmacology"] = "Curr Vasc Pharmacol",
  ["current women's health reports"] = "Curr Womens Health Rep",
  ["current women's health reviews"] = "Curr Womens Health Rev",
  ["current women’s health reports"] = "Curr. Womens Health Rep.",
  ["current work in the history of medicine"] = "Curr Work Hist Med",
  ["current world leaders"] = "Curr World Lead",
  ["current zoology"] = "Curr Zool",
  ["currents in alcoholism"] = "Curr Alcohol",
  ["currents in modern biology"] = "Curr Mod Biol",
  ["currents in pharmacy teaching & learning"] = "Curr Pharm Teach Learn",
  ["currents in theology and mission"] = "Curr Theol Mission",
  ["curso internacional de oftalmologia. curso internacional de oftalmologia"] = "Curso Int Oftalmol",
  ["curso internacional de oftalmología. curso internacional de oftalmología"] = "Curso Int Oftalmol",
  ["cursos e congresos da universidade de santiago de compostela"] = "Cursos Congr. Univ. Santiago de Compostela",
  ["curtis’s botanical magazine"] = "Curtis’s Bot. Mag.",
  ["curved and layered structures"] = "Curved Layered Struct.",
  ["cutaneous and ocular toxicology"] = "Cutan Ocul Toxicol",
  ["cutis; cutaneous medicine for the practitioner"] = "Cutis",
  ["cutting edge psychiatry in practice"] = "Cut Edge Psychiatry Pract",
  ["cvd prevention : the journal of the international society and federation of cardiology"] = "CVD Prev",
  ["cvd prevention and control"] = "CVD Prev Control",
  ["cvi forum"] = "CVI Forum",
  ["cvi newswatch"] = "CVI Newswatch",
  ["cvir endovascular"] = "CVIR Endovasc",
  ["cw, canadian welfare"] = "Can Welf",
  ["cwi syllabi"] = "CWI Syllabi",
  ["cwi tract"] = "CWI Tract",
  ["cyber security and applications"] = "Cyber Secur. Appl.",
  ["cyber-physical systems"] = "Cyber-Phys. Syst.",
  ["cybernetics and information technologies"] = "Cybern. Inf. Technol.",
  ["cybernetics and systems"] = "Cybern Syst",
  ["cybernetics and systems analysis"] = "Cybernet. Systems Anal.",
  ["cyberpsychology & behavior : the impact of the internet, multimedia and virtual reality on behavior and society"] = "Cyberpsychol Behav",
  ["cyberpsychology and behavior"] = "Cyberpsychol. Behav.",
  ["cyberpsychology, behavior and social networking"] = "Cyberpsychol Behav Soc Netw",
  ["cyberpsychology, behavior, and social networking"] = "Cyberpsychol. Behav. Social Networking",
  ["cyborg and bionic systems"] = "Cyborg Bionic Syst.",
  ["cylchgrawn llyfrgell genedlaethol cymru. the national library of wales journal. national library of wales"] = "Cylchgrawn Llyfrgell Genedlaethol Cymru",
  ["cymbella : revista de investigación y difusió́n sobre algas"] = "Cymbella",
  ["cyprus journal of economics"] = "Cyprus J. Econ.",
  ["cyprus medical journal"] = "Cyprus Med J",
  ["cyprus review"] = "Cyprus Rev.",
  ["cyta journal of food"] = "CyTA J. Food",
  ["cyta: journal of food (print)"] = "CYTA J Food",
  ["cytogenetic and genome research"] = "Cytogenet Genome Res",
  ["cytogenetics and cell genetics"] = "Cytogenet Cell Genet",
  ["cytokine & growth factor reviews"] = "Cytokine Growth Factor Rev",
  ["cytokine (philadelphia)"] = "Cytokine (Philadelphia)",
  ["cytokine and growth factor reviews"] = "Cytokine Growth Factor Rev.",
  ["cytokine: x"] = "Cytokine X",
  ["cytokines and molecular therapy"] = "Cytokines Mol Ther",
  ["cytokines, cellular & molecular therapy"] = "Cytokines Cell Mol Ther",
  ["cytokines, cellular and molecular therapy"] = "Cytokines Cell. Mol. Ther.",
  ["cytology and genetics"] = "Cytol Genet",
  ["cytometry, part a"] = "Cytometry, Part A",
  ["cytometry, part b: clinical cytometry"] = "Cytometry, Part B",
  ["cytometry. part a"] = "Cytometry A",
  ["cytometry. part a : the journal of the international society for analytical cytology"] = "Cytometry A",
  ["cytometry. part b, clinical cytometry"] = "Cytometry B Clin Cytom",
  ["cytometry. supplement"] = "Cytometry. Suppl.",
  ["cytometry. supplement : the journal of the society for analytical cytology"] = "Cytometry Suppl",
  ["cytopathology : official journal of the british society for clinical cytology"] = "Cytopathology",
  ["cytoskeleton (hoboken, n.j.)"] = "Cytoskeleton (Hoboken)",
  ["czasopismo geograficzne : kwartalnik zrzeszenia pol. nauczycieli geografji, towarzystwa geograficznego we lwowie i towarzystwa geograficznego w poznaniu"] = "Czas Geogr",
  ["czasopismo prawno-historyczne"] = "Czas Praw Hist",
  ["czasopismo stomatologiczne"] = "Czas Stomatol",
  ["czech journal of animal science"] = "Czech J. Anim. Sci.",
  ["czech journal of animal science = zivočišná výroba"] = "Czech J Anim Sci",
  ["czech journal of food science"] = "Czech J. Food Sci.",
  ["czech journal of food science = potravinářské vědy"] = "Czech J Food Sci",
  ["czech journal of food sciences"] = "Czech J. Food Sci.",
  ["czech journal of genetics and plant breeding"] = "Czech J. Genet. Plant Breed.",
  ["czech journal of genetics and plant breeding = genetika a éslechtéenâi"] = "Czech J Genet Plant Breed",
  ["czech mycology"] = "Czech Mycol",
  ["czech polar reports"] = "Czech Polar Rep",
  ["czech sociological review"] = "Czech Sociol Rev",
  ["czechoslovak economic digest"] = "Czech Econ Dig",
  ["czechoslovak economic papers"] = "Czech. Econ. Pap.",
  ["czechoslovak journal for operations research"] = "Czech. J. Operations Res.",
  ["czechoslovak journal of food sciences"] = "Czech. J. Food Sci.",
  ["czechoslovak journal of physics"] = "Czech. J. Phys.",
  ["czechoslovak mathematical journal"] = "Czechoslovak Math. J.",
  ["czechoslovak medicine"] = "Czech Med",
  ["czechoslovak sociological review"] = "Czechoslov Sociol Rev",
  ["časopis matice moravské (1968)"] = "Cas Matice Moravske",
  ["časopis národního musea"] = "Cas Nar Muz",
  ["časopis za suvremenu povijest"] = "Cas Suvrem Povij",
  ["časopis za zgodovino in narodopisje. review for history and ethnography"] = "Cas Zgodovino Narodop",
  ["československá patologie"] = "Cesk Patol",
  ["československý časopis historický"] = "Cesk Cas Hist",
  ["d-lib magazine : the magazine of the digital library forum"] = "Dlib Mag",
  ["da zhong ri bao"] = "Da zhong Ri Bao",
  ["da, asia report"] = "DA Asia Rep",
  ["da, west asia report"] = "DA West Asia Rep",
  ["dacia. revue d’archéologie et d’histoire ancienne"] = "Dacia",
  ["dacoromania. jahrbuch für östliche latinität"] = "Dacoromania",
  ["dados (reunião, análise e difusão de informação sobre saúde)"] = "Dados",
  ["daehan hwan'gyeong gonghag hoeji"] = "Daehan Hwangyeong Gonghag Hoeji",
  ["dagstuhl reports"] = "Dagstuhl Rep",
  ["daidalos. studi e ricerche del dipartimento di scienze del mondo antico"] = "Daidalos",
  ["daily (statistics canada)"] = "Dly Stat Can",
  ["daily mail and guardian"] = "Dly Mail Guard",
  ["daily report (united states. foreign broadcast information service)"] = "Dly Rep (U S Foreign Broadcast Inf Serv)",
  ["daily report. china"] = "Dly Rep China",
  ["daily report. people's republic of china"] = "Dly Rep Peoples Repub China",
  ["daily report. people’s republic of china"] = "Dly Rep Peoples Repub China",
  ["daily times (lagos, nigeria)"] = "Dly Times",
  ["dairy science & technology"] = "Dairy Sci Technol",
  ["dakar medical"] = "Dakar Med",
  ["dakar médical"] = "Dakar Med",
  ["dalhousie dental journal"] = "Dalhousie Dent J",
  ["dalhousie law journal"] = "Dalhous Law J",
  ["dalian hai yang da xue xue bao"] = "Dalian Hai Yang Da Xue Xue Bao",
  ["dalnevostochnyi matematicheskii zhurnal. far eastern mathematical journal"] = "Dalnevost. Mat. Zh.",
  ["dalton transactions"] = "Dalton Trans.",
  ["dalton transactions (cambridge, england : 2003)"] = "Dalton Trans",
  ["dalʹnevostochnyĭ zhurnal infekt︠s︡ionnoĭ patologii"] = "Dalnevost Z Infekc Patol",
  ["damaszener forschungen"] = "DaF",
  ["damaszener mitteilungen"] = "DaM",
  ["damaszener mitteilungen, hrsg. von dem deutschen archäologischen institut (station damaskus)"] = "MDAI(D)",
  ["dang dai zhongguo yan jiu = modern china studies"] = "Dang Dai Zhongguo Yan Jiu",
  ["dangerous properties of industrial materials report"] = "Dangerous Prop. Ind. Mater. Rep.",
  ["danish journal of archaeology"] = "Dan. J. Archaeol.",
  ["danish medical bulletin"] = "Dan Med Bull",
  ["danish medical journal"] = "Dan Med J",
  ["danmarks amtsrad"] = "Dan Amtsrad",
  ["dansk medicinhistorisk arbog"] = "Dan Medicinhist Arbog",
  ["dansk medicinhistorisk årbog"] = "Dan Medicinhist Arbog",
  ["dansk tidsskrift for farmaci"] = "Dan Tidsskr Farm",
  ["danski kemi"] = "Dan. Kemi",
  ["dapim refuiim"] = "Dapim Refuiim",
  ["dapim refuiim. folia medica"] = "Dapim Refuiim",
  ["daru : journal of faculty of pharmacy, tehran university of medical sciences"] = "Daru",
  ["daru journal of pharmaceutical sciences"] = "DARU J. Pharm. Sci.",
  ["das altertum"] = "Altertum",
  ["das altertum, hrsg. vom zentralinst. für alte gesch. und archäol. der dt. akad. der ddr"] = "Altertum",
  ["das archäologische jahr in bayern"] = "AJahrBay",
  ["das argument"] = "Argument",
  ["das band; monatsblatt für kranke und gesunde"] = "Band",
  ["das dental-labor. le laboratoire dentaire. the dental laboratory"] = "Dent Labor (Munch)",
  ["das deutsche archäologische institut. geschichte und dokumente"] = "DAIGeschDok",
  ["das deutsche gesundheitswesen"] = "Dtsch Gesundheitsw",
  ["das logbuch"] = "Logbuch",
  ["das medizinische laboratorium"] = "Med Lab (Stuttg)",
  ["das medizinische periskop ingelheim"] = "Med Periskop Ingelheim",
  ["das offentliche gesundheitswesen"] = "Offentl Gesundheitswes",
  ["das orchester"] = "Orchester",
  ["das osterreichische zahntechniker handwerk : offizielles organ der bundesinnung der zahntechniker"] = "Osterr Zahntech Handw",
  ["das österreichische zahntechniker handwerk"] = "Osterr Zahntech Handwerk",
  ["das österreichische zahntechniker handwerk : offizielles organ der bundesinnung der zahntechniker"] = "Osterr Zahntech Handw",
  ["das rheinische landesmuseum bonn. berichte aus der arbeit des museums"] = "RheinMusBonn",
  ["das römische herrscherbild"] = "Herrscherbild",
  ["das weltall. bildgeschmuckte zeitschrift fur astronomie und verwandte gebiete"] = "Weltall. Bildgeschmuckte Z. Astron. verw. Geb.",
  ["dasc-picom-datacom-cyberscitech 2017 : 2017 ieee 15th international conference on dependable, autonomic and secure computing ; 2017 ieee 15th international conference on pervasive intelligence and computing ; 2017 ieee 3rd international conference on big data intelligence and computing ; 2017 ieee cyber science and technology congress : 6-11 november 2017, orlando, florida. ieee international symposium on dependable, autonomic and secure computing (15th : 2017 : orlando, fla.)"] = "DASC PICom DataCom CyberSciTech 2017 (2017)",
  ["data & knowledge engineering"] = "Data Knowl Eng",
  ["data & policy"] = "Data Policy",
  ["data and applications security and privacy xxxii : 32nd annual ifip wg 11.3 conference, dbsec 2018, bergamo, italy, july 16-18, 2018, proceedings. annual ifip wg 11.3 working conference on data and applications security (32nd : 2018 : bergamo, italy)"] = "Data Appl Secur Priv XXXII (2018)",
  ["data and information management"] = "Data Inf Manag",
  ["data and knowledge engineering"] = "Data Knowl. Eng.",
  ["data asia"] = "Data Asia",
  ["data basics : a publication supported by and for the members of the society for clinical data management, inc"] = "Data Basics",
  ["data bulletin (center for studying health system change)"] = "Data Bull (Cent Stud Health Syst Change)",
  ["data in brief"] = "Data Brief",
  ["data intelligence"] = "Data Intell",
  ["data mining and big data : second international conference, dmbd 2017, fukuoka, japan, july 27-august 1, 2017. proceedings. dmbd (conference) (2nd : 2017 : fukuoka, japan)"] = "Data Min Big Data (2017)",
  ["data mining and knowledge discovery"] = "Data Min Knowl Discov",
  ["data science"] = "Data Sci.",
  ["data science and digital engineering"] = "Data Sci. Digital Eng.",
  ["data science and engineering"] = "Data Sci. Eng.",
  ["data science and management"] = "Data Sci. Manage.",
  ["data science in science"] = "Data Sci. Sci.",
  ["data science journal"] = "Data Sci. J.",
  ["data strategies & benchmarks : the monthly advisory for health care executives"] = "Data Strateg Benchmarks",
  ["data strategies and benchmarks"] = "Data Strategies Benchmarks",
  ["data technologies and applications"] = "Data Technol. Appl.",
  ["data user news"] = "Data User News",
  ["data-centric engineering"] = "Data-Centric Eng.",
  ["data-driven modelling"] = "Data-Driven Modell.",
  ["database : the journal of biological databases and curation"] = "Database (Oxford)",
  ["dataset papers in biology"] = "Dataset Pap Biol",
  ["datenbank-spektrum : zeitschrift für datenbanktechnologie : organ der fachgruppe datenbanken der gesellschaft für informatik e.v"] = "Datenbank Spektrum",
  ["daughters of the american revolution magazine"] = "DAR Mag",
  ["dauphin county reports"] = "Dauphin Cty Rep",
  ["david hilbert’s lectures on the foundations of mathematics and physics 1891–1933"] = "David Hilbert’s Lect. Found. Math. Phys. 1891–1933",
  ["davoser revue"] = "Davos. Rev.",
  ["daxue huaxue"] = "Daxue Huaxue",
  ["ddz. das deutsche zahnärzteblatt"] = "DDZ",
  ["ddz: das deutsche zahnarzteblatt"] = "DDZ.",
  ["ddz; das deutsche zahnarzteblatt"] = "DDZ",
  ["de economia"] = "De Economia",
  ["de economist"] = "Economist (Leiden)",
  ["de economía"] = "De Economia",
  ["de economía"] = "De Economía",
  ["de gruyter expositions in mathematics"] = "De Gruyter Exp. Math.",
  ["de gruyter graduate"] = "De Gruyter Grad.",
  ["de gruyter proceedings"] = "De Gruyter Proc.",
  ["de gruyter proceedings in mathematics"] = "De Gruyter Proc. Math.",
  ["de gruyter series in applied and numerical mathematics"] = "De Gruyter Ser. Appl. Numer. Math.",
  ["de gruyter series in computational science and engineering"] = "De Gruyter Ser. Comput. Sci. Eng.",
  ["de gruyter series in discrete mathematics and applications"] = "De Gruyter Ser. Discrete Math. Appl.",
  ["de gruyter series in logic and its applications"] = "De Gruyter Ser. Log. Appl.",
  ["de gruyter series in mathematics and life sciences"] = "De Gruyter Ser. Math. Life Sci.",
  ["de gruyter series in nonlinear analysis and applications"] = "De Gruyter Ser. Nonlinear Anal. Appl.",
  ["de gruyter series in probability and stochastics"] = "De Gruyter Ser. Probab. Stoch.",
  ["de gruyter series on the applications of mathematics in engineering and information sciences"] = "De Gruyter Ser. Appl. Math. Eng. Inf. Sci.",
  ["de gruyter studies in mathematical physics"] = "De Gruyter Stud. Math. Phys.",
  ["de gruyter studies in mathematics"] = "De Gruyter Stud. Math.",
  ["de gruyter studium"] = "De Gruyter Stud.",
  ["de gruyter textbook"] = "de Gruyter Textbook",
  ["de halve maen"] = "Halve Maen",
  ["de museus. quaderns de museologia i museografia"] = "DeMuseus",
  ["de paul law review"] = "De Paul Law Rev",
  ["de stethoscoop"] = "Stethosc",
  ["de ziekenfondsgids"] = "Ziekenfondsgids",
  ["de: journal of dental engineering"] = "DE J. Dent. Eng.",
  ["de; the journal of dental engeering"] = "DE J Dent Eng",
  ["deaconess hospital medical bulletin"] = "Deaconess Hosp Med Bull",
  ["deafness & education international : the journal of the british association of teachers of the deaf"] = "Deafness Educ Int",
  ["deans list"] = "Deans List",
  ["deans notes"] = "Deans Notes",
  ["death education"] = "Death Educ",
  ["death studies"] = "Death Stud",
  ["debates do ner : publicação do núcleo de estudos da religião do programa de pós-graduação em antropologia social da universidade federal do rio grande do sul"] = "Debates NER",
  ["dechema biotechnology conferences"] = "DECHEMA Biotechnol. Conf.",
  ["decision (washington, d.c.)"] = "Decision (Wash D C )",
  ["decision analysis"] = "Decis. Anal.",
  ["decision analysis : a journal of the institute for operations research and the management sciences"] = "Decis Anal",
  ["decision engineering"] = "Decis. Eng.",
  ["decision making in manufacturing and services"] = "Decis. Mak. Manuf. Serv.",
  ["decision science"] = "Decision Sci.",
  ["decision sciences"] = "Decis. Sci.",
  ["decision support systems"] = "Decis Support Syst",
  ["decisions in economics and finance. a journal of applied mathematics"] = "Decis. Econ. Finance",
  ["deep learning and data labeling for medical applications : first international workshop, labels 2016, and second international workshop, dlmia 2016, held in conjunction with miccai 2016, athens, greece, october 21, 2016, proceedings"] = "Deep Learn Data Label Med Appl (2016)",
  ["deep learning in medical image analysis and multimodal learning for clinical decision support : 4th international workshop, dlmia 2018, and 8th international workshop, ml-cds 2018, held in conjunction with miccai 2018, granada, spain, september 20, 2018 : proceedings. dlmia (workshop) (4th : 2018 : granada, spain)"] = "Deep Learn Med Image Anal Multimodal Learn Clin Decis Support (2018)",
  ["deep learning in medical image analysis and multimodal learning for clinical decision support : third international workshop, dlmia 2017, and 7th international workshop, ml-cds 2017, held in conjunction with miccai 2017 québec city, qc, canada, september 14, 2017, proceedings"] = "Deep Learn Med Image Anal Multimodal Learn Clin Decis Support (2017)",
  ["deep sea research part a: oceanographic research papers"] = "Deep Sea Res. Part A",
  ["deep sea research part b: oceanographic literature review"] = "Deep Sea Res. Part B",
  ["deep sea research part i: oceanographic research papers"] = "Deep Sea Res. Part I",
  ["deep sea research part ii: topical studies in oceanography"] = "Deep Sea Res. Part II",
  ["deep underground science & engineering"] = "Deep Underground Sci. Eng.",
  ["deep-sea research"] = "Deep-Sea Res.",
  ["deep-sea research. part a, oceanographic research papers"] = "Deep Sea Res A",
  ["deep-sea research. part i, oceanographic research papers"] = "Deep Sea Res 1 Oceanogr Res Pap",
  ["deep-sea research. part ii, topical studies in oceanography"] = "Deep Sea Res 2 Top Stud Oceanogr",
  ["def : development education forum"] = "DEF",
  ["defence and peace economics"] = "Defence Peace Econ.",
  ["defence economics"] = "Defence Econ.",
  ["defence life science journal"] = "Def Life Sci J",
  ["defence science journal"] = "Def Sci J",
  ["defence studies"] = "Def. Stud.",
  ["defence technology"] = "Def. Technol.",
  ["defensa odontológica"] = "Def Odontol",
  ["defense & security analysis"] = "Def. Secur. Anal.",
  ["defense acquisition research journal"] = "Def Acquis Res J",
  ["defense counsel journal"] = "Def Couns J",
  ["defense counsel journal / international association of defense counsel"] = "Def Couns J",
  ["defense law journal"] = "Def Law J",
  ["defense nationale"] = "Def Natl",
  ["degenerative neurological and neuromuscular disease"] = "Degener Neurol Neuromuscul Dis",
  ["dei journal of science & engineering research"] = "DEI J Sci Eng Res",
  ["dejiny matematiky/history of mathematics"] = "Dej. Mat./Hist. Math.",
  ["delaware code annotated. delaware"] = "Del Code Annot Del",
  ["delaware history"] = "Del Hist",
  ["delaware journal of public health"] = "Dela J Public Health",
  ["delaware medical journal"] = "Del Med J",
  ["delaware nurse"] = "Del Nurse",
  ["delft progress report"] = "Delft Prog Rep",
  ["delpinoa : nuova serie del bullettino dell'orto botanico della università di napoli"] = "Delpinoa",
  ["delta journal of science"] = "Delta J. Sci.",
  ["deltion hellenikes mikrobiologikes kai hygieinologikes hetaireias"] = "Delt Hell Mikrobiol Hygieinol Hetair",
  ["deltion tes paidiatrikes klinikes tou panepistemiou athenon"] = "Delt Paidiatr Klin Panepistem Athenon",
  ["deltion. iatrocheirourgike hetaireia athenon"] = "Deltion Iatrocheirourgike Hetaireia Athenon",
  ["deltion. iatrocheirourgikē hetaireia athēnōn"] = "Deltion Iatrocheirourgike Hetaireia Athenon",
  ["dementia & neuropsychologia"] = "Dement Neuropsychol",
  ["dementia (basel, switzerland)"] = "Dementia",
  ["dementia (london, england)"] = "Dementia (London)",
  ["dementia and geriatric cognitive disorders"] = "Dement Geriatr Cogn Disord",
  ["dementia and geriatric cognitive disorders extra"] = "Dement Geriatr Cogn Dis Extra",
  ["dementia and neurocognitive disorders"] = "Dement Neurocogn Disord",
  ["demircihüyük. die ergebnisse der ausgrabungen 1975–1978"] = "Demircihueyuek",
  ["democracy (new york, n.y.)"] = "Democracy",
  ["demografia y economia"] = "Demogr Econ",
  ["demograficheskie issledovaniia (kiev, ukraine)"] = "Demogr Issled",
  ["demograficheskie issledovanii︠a︡ (kiev, ukraine)"] = "Demogr Issled",
  ["demografía y economía"] = "Demogr Econ",
  ["demografske sveske"] = "Demogr Sveske",
  ["demografske sveske / centar za demografska istrazivanja instituta drustvenih nauka univerziteta u beogradu"] = "Demogr Sveske",
  ["demografía y economía"] = "Demogr. Econ.",
  ["demographic bulletin (wellington, n.z.)"] = "Demogr Bull",
  ["demographic research"] = "Demogr Res",
  ["demographic research monographs"] = "Demogr. Res. Monogr.",
  ["demographie africaine"] = "Demogr Afr",
  ["demographische informationen"] = "Demogr Inf",
  ["demographische informationen / hrsg. vom institut fur demographie, osterreichische akademie der wissenschaften"] = "Demogr Inf",
  ["demography india"] = "Demogr India",
  ["demohrafichni doslidzhennia"] = "Demohrafichni Doslidzhennia",
  ["demohrafichni doslidz︠h︡enni︠a︡"] = "Demohrafichni Doslidzhennia",
  ["demonstratio mathematica"] = "Demonstr. Math.",
  ["demos (mexico city, mexico)"] = "Demos",
  ["den jyske historiker"] = "Den Jyske Hist",
  ["den norske tannlaegeforenings tidende"] = "Nor Tannlaegeforen Tid",
  ["dendrome : forest tree genome research updates"] = "Dendrome",
  ["dengue bulletin"] = "Dengue Bull",
  ["denki kagaku oyobi kogyo butsuri kagaku"] = "Denki Kagaku oyobi Kogyo Butsuri Kagaku",
  ["denkmalpflege in baden-württemberg"] = "DenkmPflBadWuert",
  ["denkmäler antiker architektur"] = "DAA",
  ["denkschriften der schweizerischen naturforschenden gesellschaft"] = "Denkschr. Schweiz. nat.forsch. Ges.",
  ["denkschriften osterreichische akademie der wissenschaften, philosophisch-historische klasse"] = "Denkschr. Osterr. Akad. Wiss. Philos.-Hist. Kl.",
  ["dens sapiens"] = "Dens Sapiens",
  ["dent, contemporary dentistry"] = "Dent",
  ["dent: contemporary dentistry"] = "Dent",
  ["dental abstracts"] = "Dent. Abstr.",
  ["dental abstracts; a selection of world dental literature"] = "Dent Abstr",
  ["dental anaesthesia and sedation"] = "Dent Anaesth Sedat",
  ["dental and medical problems"] = "Dent Med Probl",
  ["dental angles"] = "Dent Angles",
  ["dental assistant"] = "Dent. Assist.",
  ["dental assistant (chicago, ill. : 1994)"] = "Dent Assist",
  ["dental assistant journal"] = "Dent. Assist. J.",
  ["dental assisting"] = "Dent Assist (Waco Tx)",
  ["dental cadmos"] = "Dent Cadmos",
  ["dental caries and fluorine. american association for the advancement of science. section on medical sciences. subsection on dentistry"] = "Dent Caries Fluor",
  ["dental clinics of north america"] = "Dent Clin North Am",
  ["dental concepts"] = "Dent Concepts",
  ["dental delineator"] = "Dent Delin",
  ["dental dialogue"] = "Dent. Dialogue",
  ["dental dialogue (bombay, india)"] = "Dent Dialogue",
  ["dental dienst"] = "Dent. Dienst",
  ["dental dienst; fachzeitschrift fur den dental-markt; technisches fachblatt fur prothetik"] = "Dent Dienst",
  ["dental dienst; fachzeitschrift für den dental-markt; technisches fachblatt für prothetik"] = "Dent Dienst",
  ["dental digest"] = "Dent Dig",
  ["dental dimensions"] = "Dent Dimens",
  ["dental discourse"] = "Dent Discourse",
  ["dental echo"] = "Dent Echo (Heidelb)",
  ["dental economics"] = "Dent. Econ.",
  ["dental economics - oral hygiene"] = "Dent Econ",
  ["dental health"] = "Dent Health",
  ["dental historian"] = "Dent. Hist.",
  ["dental historian : lindsay club newsletter"] = "Dent Hist",
  ["dental hygiene"] = "Dent Hyg (Chic)",
  ["dental hygienist"] = "Dent. Hyg. (San Franc.)",
  ["dental hygienist (san francisco, calif.)"] = "Dent Hyg (San Franc)",
  ["dental hypotheses"] = "Dent Hypotheses",
  ["dental images"] = "Dent Images",
  ["dental implantology update"] = "Dent Implantol Update",
  ["dental items of interest"] = "Dent Items Interest",
  ["dental journal"] = "Dent J",
  ["dental journal of malaysia"] = "Dent J Malays",
  ["dental journal of malaysia and singapore"] = "Dent. J. Malaysia Singapore",
  ["dental journal of zambia"] = "Dent. J. Zamb.",
  ["dental journal of zambia : official publication of the zambia dental association"] = "Dent J Zamb",
  ["dental lab management today"] = "Dent Lab Manage Today",
  ["dental labor"] = "Dent. Labor (Munch.)",
  ["dental laboratorie bladet"] = "Dent Lab Bl",
  ["dental laboratory review"] = "Dent Lab Rev",
  ["dental magazine and oral topics"] = "Dent. Mag. Oral Top.",
  ["dental management"] = "Dent Manage",
  ["dental materials"] = "Dent. Mater.",
  ["dental materials : official publication of the academy of dental materials"] = "Dent Mater",
  ["dental materials journal"] = "Dent Mater J",
  ["dental mirror"] = "Dent Mirror (Quezon City)",
  ["dental news"] = "Dent News (Lond)",
  ["dental office"] = "Dent Off",
  ["dental outlook"] = "Dent. Outlook",
  ["dental practice"] = "Dent Pract",
  ["dental practice management"] = "Dent Pract Manage",
  ["dental practice; views, trends and news of dentistry"] = "Dent Pract (Cincinnati)",
  ["dental practitioner and dental record"] = "Dent. Pract. Dent. Rec.",
  ["dental press"] = "Dent Press",
  ["dental press journal of orthodontics"] = "Dental Press J Orthod",
  ["dental progress"] = "Dent Prog (Chic)",
  ["dental radiography and photography"] = "Dent Radiogr Photogr",
  ["dental research and graduate study"] = "Dent Res Grad Study",
  ["dental research journal"] = "Dent Res J (Isfahan)",
  ["dental school quarterly"] = "Dent Sch Q",
  ["dental school quarterly / the university of texas health science center at san antonio"] = "Dent Sch Q",
  ["dental school quarterly, university of texas health science center at san antonio"] = "Dent. Sch. Q.",
  ["dental student"] = "Dent Stud",
  ["dental survey"] = "Dent Surv",
  ["dental team"] = "Dent Team",
  ["dental teamwork"] = "Dent Teamwork",
  ["dental technician"] = "Dent. Tech.",
  ["dental therapeutics newsletter"] = "Dent Ther Newsl",
  ["dental traumatology : official publication of international association for dental traumatology"] = "Dent Traumatol",
  ["dental update"] = "Dent Update",
  ["dental world"] = "Dent. World",
  ["dental world (london, england)"] = "Dent World",
  ["dental, oral, and craniofacial research"] = "Dent Oral Craniofac Res",
  ["dental-revue (solothurn, switzerland)"] = "Dent Rev",
  ["dentist (waco, tex.)"] = "Dentist",
  ["dentist news"] = "Dent News",
  ["dentistische rundschau"] = "Dent Rundsch",
  ["dentistry (american student dental association)"] = "Dentistry",
  ["dentistry (lisle, ill.)"] = "Dentistry (Lisle)",
  ["dentistry (loma linda, calif.)"] = "Dentistry (Loma Linda)",
  ["dentistry (sunnyvale, calif.)"] = "Dentistry (Sunnyvale)",
  ["dentistry 3000"] = "Dent 3000",
  ["dentistry in japan"] = "Dent Jpn (Tokyo)",
  ["dentistry journal"] = "Dent J (Basel)",
  ["dentistry review"] = "Dent. Rev.",
  ["dentistry today"] = "Dent Today",
  ["dentisuto (dentist)"] = "Dentisuto.",
  ["dentisuto = dentist"] = "Dentisuto",
  ["dento maxillo facial radiology"] = "Dentomaxillofac Radiol",
  ["dento-maxillo-facial radiology"] = "Dentomaxillofac. Radiol.",
  ["dento-maxillo-facial radiology. supplement"] = "Dentomaxillofac Radiol Suppl",
  ["denver journal of international law and policy"] = "Denver J Int Law Policy",
  ["denver law journal"] = "Denver Law J",
  ["denver university law review"] = "Denver Univ Law Rev",
  ["department of health and social services quarterly magazine"] = "Dep Health Soc Serv Q Mag",
  ["department of state publication. background notes series"] = "Backgr Notes Ser",
  ["department of the army technical bulletin. tb med"] = "Tech Bull Dep Army",
  ["depaul journal of health care law"] = "DePaul J Health Care Law",
  ["dependence modeling"] = "Depend. Model.",
  ["depositional record"] = "Depositional Rec.",
  ["depression and anxiety"] = "Depress Anxiety",
  ["depression research and treatment"] = "Depress Res Treat",
  ["depression, mind and body"] = "Depress Mind Body",
  ["depression-causes diagnosis and treatment"] = "Depress. Causes. Diagn.",
  ["der altsprachliche unterricht"] = "Altsprachi Unterr",
  ["der altsprachliche unterricht: arbeitshefte zu seiner wissenschaftlichen begründung und praktischen gestalt"] = "AU",
  ["der anaesthesist"] = "Anaesthesist",
  ["der anschnitt. mitteilungsblatt der vereinigung der freunde von kunst und kultur im bergbau"] = "Anschnitt",
  ["der archivar"] = "Archivar",
  ["der arztliche dienst"] = "Arztl Dienst",
  ["der ärztliche dienst"] = "Arztl Dienst",
  ["der chirurg; zeitschrift fur alle gebiete der operativen medizen"] = "Chirurg",
  ["der chirurg; zeitschrift für alle gebiete der operativen medizen"] = "Chirurg",
  ["der deutsche apotheker"] = "Dtsch Apoth",
  ["der donauraum"] = "Donauraum",
  ["der donaurum"] = "Der Donaurum",
  ["der freie zahnarzt"] = "Freie Zahnarzt",
  ["der gastroenterologe : zeitschrift für gastroenterologie und hepatologie"] = "Gastroenterologe",
  ["der gynakologe"] = "Gynakologe",
  ["der gynäkologe"] = "Gynakologe",
  ["der hautarzt; zeitschrift fur dermatologie, venerologie, und verwandte gebiete"] = "Hautarzt",
  ["der hautarzt; zeitschrift für dermatologie, venerologie, und verwandte gebiete"] = "Hautarzt",
  ["der internist"] = "Internist (Berl)",
  ["der kassenzahnarzt; colloquium med. dent"] = "Kassenzahnarzt Colloq Med Dent",
  ["der knappschaftsarzt"] = "Knappschaftsarzt",
  ["der krebsarzt"] = "Krebsarzt",
  ["der landarzt"] = "Landarzt",
  ["der medizinische sachverstandige"] = "Med Sachverstand",
  ["der medizinische sachverständige"] = "Med Sachverstand",
  ["der nephrologe"] = "Nephrologe",
  ["der nervenarzt"] = "Nervenarzt",
  ["der neue pauly. enzyklopädie der antike"] = "DNP",
  ["der offentliche gesundheitsdienst"] = "Offentl Gesundheitsdienst",
  ["der onkologe : organ der deutschen krebsgesellschaft e.v"] = "Onkologe (Berl)",
  ["der ophthalmologe : zeitschrift der deutschen ophthalmologischen gesellschaft"] = "Ophthalmologe",
  ["der orthopade"] = "Orthopade",
  ["der orthopäde"] = "Orthopade",
  ["der öffentliche gesundheitsdienst"] = "Offentl Gesundheitsdienst",
  ["der pathologe"] = "Pathologe",
  ["der pharmacia lettre"] = "Pharm Lett",
  ["der pneumologe"] = "Pneumologe (Berl)",
  ["der praktische arzt"] = "Prakt Arzt",
  ["der praktische tierarzt"] = "Prakt Tierarzt",
  ["der radiologe"] = "Radiologe",
  ["der schweizer optiker. l'opticien suisse"] = "Schweiz Opt Opt Suisse",
  ["der sportarzt vereinigt mit sportmedizin"] = "Sportarzt Ver Sportmed",
  ["der stahlbau"] = "Stahlbau",
  ["der tuberkulosearzt"] = "Tuberkulosearzt",
  ["der unfallchirurg"] = "Unfallchirurg",
  ["der urologe"] = "Urologe",
  ["der urologe. ausg. a"] = "Urologe A",
  ["der urologe. ausg. b"] = "Urologe B",
  ["der wendepunkt im leben und im leiden"] = "Wendepunkt Leben Leiden",
  ["der zahnarzt; colloquium med. dent"] = "Zahnarzt",
  ["der zahntechniker. le mecanicien-dentiste"] = "Zahntechniker (Basel)",
  ["der zahntechniker. le mécanicien-dentiste"] = "Zahntech (Basel)",
  ["der zoologische garten"] = "Zool Gart",
  ["der züchter : genetics and breeding research"] = "Zuchter",
  ["dermatitis : contact, atopic, occupational, drug"] = "Dermatitis",
  ["dermatitis : contact, atopic, occupational, drug : official journal of the american contact dermatitis society, north american contact dermatitis group"] = "Dermatitis",
  ["dermatologia ibero latino-americana"] = "Dermatol Iber Lat Am",
  ["dermatologia internationalis"] = "Dermatol Int",
  ["dermatologia tropica et ecologica geographica"] = "Dermatol Trop Ecol Geogr",
  ["dermatologia; revista mexicana"] = "Dermatol Rev Mex",
  ["dermatologic clinics"] = "Dermatol Clin",
  ["dermatologic surgery"] = "Dermatol. Surg.",
  ["dermatologic surgery : official publication for american society for dermatologic surgery [et al.]"] = "Dermatol Surg",
  ["dermatologic therapy"] = "Dermatol Ther",
  ["dermatological nursing : the journal of the british dermatological nursing group"] = "Dermatol Nurs (Lond)",
  ["dermatologische monatschrift"] = "Dermatol Monatsschr",
  ["dermatologische monatsschrift"] = "Dermatol. Monatsschr.",
  ["dermatologische wochenschrift"] = "Dermatol Wochenschr",
  ["dermatology (basel, switzerland)"] = "Dermatology",
  ["dermatology and therapy"] = "Dermatol Ther (Heidelb)",
  ["dermatology case reports"] = "Dermatol Case Rep",
  ["dermatology nursing"] = "Dermatol Nurs",
  ["dermatology nursing / dermatology nurses’ association"] = "Dermatol Nurs",
  ["dermatology online journal"] = "Dermatol Online J",
  ["dermatology practical & conceptual"] = "Dermatol Pract Concept",
  ["dermatology reports"] = "Dermatol Reports",
  ["dermatology research and practice"] = "Dermatol Res Pract",
  ["dermatology times"] = "Dermatol. Times",
  ["dermatopathology (basel, switzerland)"] = "Dermatopathology (Basel)",
  ["dermatosen in beruf und umwelt (occupational and environmental dermatoses)"] = "Derm. Beruf Umwelt",
  ["dermatosen in beruf und umwelt. occupation and environment"] = "Derm Beruf Umwelt",
  ["desalination and water treatment"] = "Desalin. Water Treat.",
  ["desarrollo de base : revista de la fundacion interamericana"] = "Desarro Base",
  ["desarrollo de base : revista de la fundación interamericana"] = "Desarro Base",
  ["desarrollo economico"] = "Desarrollo Econ",
  ["desarrollo económico"] = "Desarrollo Econ",
  ["desarrollo económico"] = "Desarrollo Econ.",
  ["desarrollo rural en las americas"] = "Desarro Rural Amer",
  ["desarrollo rural en las américas"] = "Desarro Rural Amer",
  ["desarrollo y sociedad"] = "Desarro Soc",
  ["descriptions of ectomycorrhizae"] = "Descr. Ectomycorrhizae",
  ["deseret news (salt lake city, utah : 1964)"] = "Deseret News",
  ["desertification control"] = "Desertif Control",
  ["design and technology education : an international journal"] = "Des Technol Educ",
  ["design automation for embedded systems"] = "Des. Autom. Embedded Syst.",
  ["design engineering canada"] = "Des. Eng. Can.",
  ["design for health (abingdon, england)"] = "Design Health (Abingdon)",
  ["design issues"] = "Des. Issues",
  ["design management journal"] = "Des. Manage. J.",
  ["design management review"] = "Des Manage Rev",
  ["design science"] = "Des. Sci.",
  ["design studies"] = "Des Stud",
  ["design, user experience, and usability : design philosophy, methods, and tools : second international conference, duxu 2013, held as part of hci international 2013, las vegas, nv, usa, july 21-26, 2013, proceedings. part i. duxu (conference) (2nd : 2013 : las vegas, nev.)"] = "Des User Exp Usability",
  ["designed monomers and polymers"] = "Des Monomers Polym",
  ["designs codes and cryptography"] = "Des. Codes Cryptogr.",
  ["designs, codes and cryptography"] = "Des. Codes Cryptogr.",
  ["designs, codes and cryptography. an international journal"] = "Des. Codes Cryptogr.",
  ["designs, codes, and cryptography"] = "Des Codes Cryptogr",
  ["detroit college of law review"] = "Detroit Coll Law Rev",
  ["detroit dental bulletin"] = "Detroit Dent Bull",
  ["detroit in perspective"] = "Detroit Perspect",
  ["detroit medical news"] = "Detroit Med News",
  ["deutsche akademie der naturforscher leopoldina"] = "Nova Acta Leopoldina (N.F.)",
  ["deutsche apotheker zeitung"] = "Dtsch. Apoth. Ztg.",
  ["deutsche apotheker-zeitung"] = "Dtsch Apoth Ztg",
  ["deutsche aussenpolitik"] = "Dtsch Aussenpolit",
  ["deutsche baumschule"] = "Dtsch. Baumsch.",
  ["deutsche dentistische zeitschrift"] = "Dtsch Dent Z",
  ["deutsche drogistenzeitung"] = "Dtsch Drog Ztg",
  ["deutsche entomologische zeitschrift"] = "Dtsch. Entomol. Z.",
  ["deutsche gesundheitspolitik"] = "Dtsch Gesundheitspolit",
  ["deutsche gesundheitswesen"] = "Dtsch. Gesundheitswes.",
  ["deutsche gewässerkundliche mitteilungen"] = "Dtsch. gewässerkdl. Mitt.",
  ["deutsche hochschulschriften"] = "Deutsche Hochschulschrift.",
  ["deutsche hydrographische zeitschrift"] = "Dtsch. Hydrogr. Z.",
  ["deutsche krankenpflegezeitschrift"] = "Dtsch Krankenpflegez",
  ["deutsche kunst und denkmalpflege"] = "DKuDenkmPfl",
  ["deutsche lebensmittel rundschau"] = "Dtsch. Lebensm.- Rundsch.",
  ["deutsche lebensmittel-rundschau : zeitschrift fur lebensmittelkunde und lebensmittelrecht"] = "Dtsch Lebensmitt Rundsch",
  ["deutsche lebensmittel-rundschau : zeitschrift für lebensmittelkunde und lebensmittelrecht"] = "Dtsch Lebensmitt Rundsch",
  ["deutsche literaturzeitung für kritik der internationalen wissenschaft"] = "DLZ",
  ["deutsche medizinische rundschau; monatsschrift mit ärztlicher akademie"] = "Dtsch Med Rundsch",
  ["deutsche medizinische wochenschrift"] = "Dtsch. Med. Wochenschr.",
  ["deutsche medizinische wochenschrift (1946)"] = "Dtsch Med Wochenschr",
  ["deutsche rentenversicherung"] = "Dtsch Rentenversicher",
  ["deutsche schwesternzeitung"] = "Dtsch Schwesternztg",
  ["deutsche stomatologie"] = "Dtsch Stomatol",
  ["deutsche stomatologie (berlin, germany : 1990)"] = "Dtsch Stomatol",
  ["deutsche tierarztliche wochenschrift"] = "Dtsch. Tierarztl. Wochenschr.",
  ["deutsche tierärztliche wochenschrift"] = "Dtsch Tierarztl Wochenschr",
  ["deutsche vierteljahrsschrift fur literaturwissenschaft und geistesgeschichte"] = "Dtsch Vierteljahresschr Litt Wiss Geistesgesch",
  ["deutsche vierteljahrsschrift für literaturwissenschaft und geistesgeschichte"] = "Dtsch Vierteljahresschr Litt Wiss Geistesgesch",
  ["deutsche zahn-, mund- und kieferheilkunde mit zentralblatt fur die gesamte zahn-, mund- und kieferheilkunde"] = "Dtsch. Zahn Mund Kieferheilkd. Zentralbl. Gesamte",
  ["deutsche zahn-, mund-, und kieferheilkunde mit zentralblatt"] = "Dtsch Zahn Mund Kieferheilkd Zentralbl",
  ["deutsche zahn-, mund-, und kieferheilkunde mit zentralblatt fur die gesamte zahn-, mund-, und kieferheilkunde"] = "Dtsch Zahn Mund Kieferheilkd Zentralbl Gesamte",
  ["deutsche zahn-, mund-, und kieferheilkunde mit zentralblatt für die gesamte zahn-, mund-, und kieferheilkunde"] = "Dtsch Zahn Mund Kieferheilkd Zentralbl Gesamte",
  ["deutsche zahnarztliche zeitschrift"] = "Dtsch. Zahnarztl. Z.",
  ["deutsche zahnärztliche zeitschrift"] = "Dtsch Zahnarztl Z",
  ["deutsche zeitschrift fur die gesamte gerichtliche medizin"] = "Dtsch. Z. Gesamte Gerichtl. Med.",
  ["deutsche zeitschrift fur mund-, kiefer- und gesichts-chirurgie"] = "Dtsch Z Mund Kiefer Gesichtschir",
  ["deutsche zeitschrift fur mund-, kiefer-, und gesichts-chirurgie"] = "Dtsch. Z. Mund Kiefer Gesichtschir.",
  ["deutsche zeitschrift fur nervenheilkunde"] = "Dtsch. Z. Nervenheilkd.",
  ["deutsche zeitschrift fur philosophie"] = "Dtsch Z Philos",
  ["deutsche zeitschrift fur sportmedizin"] = "Dtsch. Z. Sportmed.",
  ["deutsche zeitschrift fur verdauungs- und stoffwechselkrankheiten"] = "Dtsch. Z. Verdau. Stoffwechselkr.",
  ["deutsche zeitschrift für die gesamte gerichtliche medizin"] = "Dtsch Z Gesamte Gerichtl Med",
  ["deutsche zeitschrift für mund-, kiefer- und gesichts-chirurgie"] = "Dtsch Z Mund Kiefer Gesichtschir",
  ["deutsche zeitschrift für nervenheilkunde"] = "Dtsch Z Nervenheilkd",
  ["deutsche zeitschrift für philosophie"] = "Dtsch Z Philos",
  ["deutsche zeitschrift für sportmedizin"] = "Dtsch Z Sportmed",
  ["deutsche zeitschrift für verdauungs- und stoffwechselkrankheiten"] = "Dtsch Z Verdau Stoffwechselkr",
  ["deutsche zeitschrift für zahnärztliche hypnose"] = "Dtsch Z Zahnarztl Hypn",
  ["deutscher gesundheitskalender"] = "Dtsch Gesundheit",
  ["deutscher zahnärztekalender"] = "Dtsch Zahnarztekal",
  ["deutsches archiv fur klinische medizin"] = "Dtsch. Arch. Klin. Med.",
  ["deutsches archiv für klinische medizin"] = "Dtsch Arch Klin Med",
  ["deutsches arzteblatt"] = "Dtsch Arztebl",
  ["deutsches ärzteblatt"] = "Dtsch Arztebl",
  ["deutsches ärzteblatt (ausg. a)"] = "Dtsch Arztebl Ausg A",
  ["deutsches ärzteblatt international"] = "Dtsch Arztebl Int",
  ["deutsches dante-jahrbuch"] = "DDJ",
  ["deutsches jahrbuch fur volkskunde"] = "Dtsch Jahrb Volkskd",
  ["deutsches jahrbuch für volkskunde"] = "Dtsch Jahrb Volkskd",
  ["deutsches medizinisches journal"] = "Dtsch Med J",
  ["deutsches schiffahrtsarchiv"] = "Dtsch Schiffahrtsarch",
  ["deutsches zentralblatt fur krankenpflege"] = "Dtsch. Zentralbl. Krankenpfl.",
  ["deutsches zentralblatt für krankenpflege"] = "Dtsch Zentralbl Krankenpfl",
  ["deutschland archiv"] = "Dtschl Arch",
  ["developing economies"] = "Developing Economies",
  ["developing world bioethics"] = "Dev World Bioeth",
  ["development & reproduction"] = "Dev Reprod",
  ["development (cambridge, england)"] = "Development",
  ["development (cambridge, england). supplement"] = "Dev Suppl",
  ["development (cambridge, united kingdom)"] = "Development (Cambridge, U. K.)",
  ["development (society for international development)"] = "Development (Rome)",
  ["development and change"] = "Dev Change",
  ["development and gender in brief : a quarterly update from bridge, raising gende awareness among policy-makers and practitioners"] = "Dev Gend Brief",
  ["development and peace"] = "Dev Peace",
  ["development and psychopathology"] = "Dev Psychopathol",
  ["development and society (soul taehakkyo. institute for social devdelopment and policy research)"] = "Dev Soc",
  ["development and society (sŏul taehakkyo. institute for social devdelopment and policy research)"] = "Dev Soc",
  ["development bank of japan research series"] = "Dev. Bank Jpn. Res. Ser.",
  ["development brief"] = "Dev Brief",
  ["development bulletin (australian development studies network)"] = "Dev Bull",
  ["development communication report"] = "Dev Commun Rep",
  ["development dialogue"] = "Dev Dialogue",
  ["development digest"] = "Dev Dig",
  ["development directions"] = "Dev Dir",
  ["development education exchange papers : deep"] = "DEEP Dev Educ Exch Pap",
  ["development engineering"] = "Dev Eng",
  ["development forum"] = "Dev Forum",
  ["development genes and evolution"] = "Dev Genes Evol",
  ["development growth and differentiation"] = "Dev. Growth Differ.",
  ["development in practice"] = "Dev Pract",
  ["development international"] = "Dev Int",
  ["development network"] = "Dev Network",
  ["development outreach : putting knowledge to work for development"] = "Dev Outreach",
  ["development policy review : the journal of the overseas development institute"] = "Dev Policy Rev",
  ["development southern africa"] = "Dev South Afr",
  ["development, growth & differentiation"] = "Dev Growth Differ",
  ["development. supplement"] = "Dev. Suppl.",
  ["developmental & comparative immunology"] = "Dev. Comp. Immunol.",
  ["developmental and comparative immunology"] = "Dev Comp Immunol",
  ["developmental biology"] = "Dev Biol",
  ["developmental biology (amsterdam, netherlands)"] = "Dev. Biol. (Amsterdam, Neth.)",
  ["developmental biology (new york, n.y. : 1985)"] = "Dev Biol (N Y 1985)",
  ["developmental biology (san diego, ca, united states)"] = "Dev. Biol. (San Diego, CA, U. S.)",
  ["developmental cell"] = "Dev Cell",
  ["developmental child welfare"] = "Dev Child Welf",
  ["developmental cognitive neuroscience"] = "Dev Cogn Neurosci",
  ["developmental disabilities research reviews"] = "Dev Disabil Res Rev",
  ["developmental dynamics"] = "Dev. Dyn.",
  ["developmental dynamics : an official publication of the american association of anatomists"] = "Dev Dyn",
  ["developmental genetics"] = "Dev Genet",
  ["developmental immunology"] = "Dev Immunol",
  ["developmental medicine and child neurology"] = "Dev Med Child Neurol",
  ["developmental medicine and child neurology. supplement"] = "Dev Med Child Neurol Suppl",
  ["developmental neurobiology"] = "Dev Neurobiol",
  ["developmental neuropsychology"] = "Dev Neuropsychol",
  ["developmental neurorehabilitation"] = "Dev Neurorehabil",
  ["developmental neuroscience"] = "Dev Neurosci",
  ["developmental neuroscience (basel, switzerland)"] = "Dev. Neurosci. (Basel, Switz.)",
  ["developmental period medicine"] = "Dev Period Med",
  ["developmental pharmacology and therapeutics"] = "Dev Pharmacol Ther",
  ["developmental psychobiology"] = "Dev Psychobiol",
  ["developmental psychology"] = "Dev Psychol",
  ["developmental review"] = "Dev. Rev.",
  ["developmental review : dr"] = "Dev Rev",
  ["developmental science"] = "Dev Sci",
  ["developments in biological standardization"] = "Dev Biol Stand",
  ["developments in biologicals"] = "Dev Biol (Basel)",
  ["developments in biologicals (basel, switzerland)"] = "Dev. Biol. (Basel, Switz.)",
  ["developments in clay science"] = "Dev Clay Sci",
  ["developments in endocrinology"] = "Dev Endocrinol",
  ["developments in health economics and public policy"] = "Dev Health Econ Public Policy",
  ["developments in heat transfer"] = "Dev. Heat Transf.",
  ["developments in mathematics"] = "Dev. Math.",
  ["developments in ophthalmology"] = "Dev Ophthalmol",
  ["developments in petrology"] = "Develop. Petrol.",
  ["developments in the built environment"] = "Dev. Built Environ.",
  ["developments in toxicology and environmental science"] = "Dev Toxicol Environ Sci",
  ["developnet news"] = "DevelopNet News",
  ["developpement & civilisations"] = "Dev Civilis",
  ["developpement et sante : revue de perfectionnement medical et sanitaire en pays tropical"] = "Dev Sante",
  ["deviant behavior"] = "Deviant Behav",
  ["défense nationale"] = "Def Natl",
  ["démographie africaine"] = "Demogr Afr",
  ["développement & civilisations"] = "Dev Civilis",
  ["développement et santé : revue de perfectionnement médical et sanitaire en pays tropical"] = "Dev Sante",
  ["dfi journal"] = "DFI J.",
  ["dgds. differential geometry—dynamical systems. monographs"] = "DGDS Differ. Geom. Dyn. Syst. Monogr.",
  ["dghs chronicle"] = "DGHS Chron",
  ["dh : bulletin d'information"] = "Demogr Hist Bull Inf Soc",
  ["dh : bulletin d’information / societe de demographie historique"] = "Demogr Hist Bull Inf Soc",
  ["dh'15: proceedings of the 5th international conference on digital health 2015 : may 18-20, 2015, florence, italy. international conference on digital health (5th : 2015 : florence, italy)"] = "DH15 (2015)",
  ["dhs+ dimensions : a biannual newsletter of the demographic and health surveys project"] = "DHS Dimens",
  ["di 1 jun yi da xue xue bao = academic journal of the first medical college of pla"] = "Di Yi Jun Yi Da Xue Xue Bao",
  ["di 2 jun yi da xue xue bao = dier junyi daxue xuebao = academic journal of second military medical college"] = "Ti Erh Chun I Ta Hsueh Hsueh Pao",
  ["di 3 jun yi da xue xue bao = disanjunyidaxue xuebao = acta acadiemiae [i.e. academiae] medicinae militaris tertiae"] = "Di 3 Jun Yi Da Xue Xue Bao",
  ["di 4 jun yi da xue xue bao = journal of the fourth military medical university = disi junyi daxue xuebao"] = "Di 4 Jun Yi Da Xue Xue Bao",
  ["di fang bing tong bao = endemic disease bulletin"] = "Di Fang Bing Tong Bao",
  ["di li xue bao"] = "Acta Geogr Sin",
  ["di li xue bao / chung-kuo ti li hsueh hui pien chi"] = "Acta Geogr Sin",
  ["dia medico uruguayo"] = "Dia Med Urug",
  ["diabete & metabolisme"] = "Diabete Metab",
  ["diabetes & metabolic syndrome"] = "Diabetes Metab Syndr",
  ["diabetes & metabolism"] = "Diabetes Metab",
  ["diabetes & metabolism journal"] = "Diabetes Metab J",
  ["diabetes & obesity international journal"] = "Diabetes Obes Int J",
  ["diabetes & vascular disease research"] = "Diab Vasc Dis Res",
  ["diabetes & vascular disease research : official journal of the international society of diabetes and vascular disease"] = "Diab Vasc Dis Res",
  ["diabetes and metabolism"] = "Diabetes Metab.",
  ["diabetes and vascular disease research"] = "Diab. Vasc. Dis. Res.",
  ["diabetes care"] = "Diabetes Care",
  ["diabetes care & education newsletter"] = "Diabetes Care Educ Newsl",
  ["diabetes educator"] = "Diabetes Educ.",
  ["diabetes forecast"] = "Diabetes Forecast",
  ["diabetes management (london, england)"] = "Diabetes Manag (Lond)",
  ["diabetes research"] = "Diabetes Res.",
  ["diabetes research (edinburgh, scotland)"] = "Diabetes Res",
  ["diabetes research (fairfax, va.)"] = "Diabetes Res (Fairfax)",
  ["diabetes research and clinical practice"] = "Diabetes Res Clin Pract",
  ["diabetes research and clinical practice. supplement"] = "Diabetes Res Clin Pract Suppl",
  ["diabetes reviews (alexandria, va.)"] = "Diabetes Rev (Alex)",
  ["diabetes self-management"] = "Diabetes Self Manag",
  ["diabetes spectrum : a publication of the american diabetes association"] = "Diabetes Spectr",
  ["diabetes technology & therapeutics"] = "Diabetes Technol Ther",
  ["diabetes technology and therapeutics"] = "Diabetes Technol. Ther.",
  ["diabetes therapy : research, treatment and education of diabetes and related disorders"] = "Diabetes Ther",
  ["diabetes, metabolic syndrome and obesity : targets and therapy"] = "Diabetes Metab Syndr Obes",
  ["diabetes, nutrition & metabolism"] = "Diabetes Nutr Metab",
  ["diabetes, obesity & metabolism"] = "Diabetes Obes Metab",
  ["diabetes, obesity and metabolism"] = "Diabetes Obes. Metab.",
  ["diabetes/metabolism research and reviews"] = "Diabetes Metab Res Rev",
  ["diabetes/metabolism reviews"] = "Diabetes Metab Rev",
  ["diabetic foot & ankle"] = "Diabet Foot Ankle",
  ["diabetic medicine"] = "Diabet. Med.",
  ["diabetic medicine : a journal of the british diabetic association"] = "Diabet Med",
  ["diabetic nephropathy : dn"] = "Diabet Nephrop",
  ["diabetologia croatica"] = "Diabetol Croat",
  ["diabetology & metabolic syndrome"] = "Diabetol Metab Syndr",
  ["diabetology international"] = "Diabetol Int",
  ["diabète & métabolisme"] = "Diabete Metab",
  ["diadora. glasilo arheoloskog muzeja u zadru"] = "Diadora",
  ["diagnosis (berlin, germany)"] = "Diagnosis (Berl)",
  ["diagnostic and clinical immunology"] = "Diagn. Clin. Immunol.",
  ["diagnostic and interventional imaging"] = "Diagn Interv Imaging",
  ["diagnostic and interventional radiology"] = "Diagn. Interv. Radiol.",
  ["diagnostic and interventional radiology (ankara, turkey)"] = "Diagn Interv Radiol",
  ["diagnostic and prognostic research"] = "Diagn Progn Res",
  ["diagnostic and therapeutic endoscopy"] = "Diagn Ther Endosc",
  ["diagnostic cytopathology"] = "Diagn Cytopathol",
  ["diagnostic gynecology and obstetrics"] = "Diagn Gynecol Obstet",
  ["diagnostic histopathology"] = "Diagn Histopathol",
  ["diagnostic histopathology (oxford, england)"] = "Diagn Histopathol (Oxf)",
  ["diagnostic histopathology / published in association with the pathological society of great britain and ireland"] = "Diagn Histopathol",
  ["diagnostic imaging"] = "Diagn Imaging",
  ["diagnostic imaging europe"] = "Diagn Imaging Eur",
  ["diagnostic imaging in clinical medicine"] = "Diagn Imaging Clin Med",
  ["diagnostic immunology"] = "Diagn Immunol",
  ["diagnostic microbiology and infectious disease"] = "Diagn Microbiol Infect Dis",
  ["diagnostic molecular pathology"] = "Diagn. Mol. Pathol.",
  ["diagnostic molecular pathology : the american journal of surgical pathology, part b"] = "Diagn Mol Pathol",
  ["diagnostic pathology"] = "Diagn Pathol",
  ["diagnostic pathology: open access"] = "Diagn Pathol Open Access",
  ["diagnostics & traitements"] = "Diagn Traitements",
  ["diagnostics (basel, switzerland)"] = "Diagnostics (Basel)",
  ["diagnostics in africa"] = "Diagn Afr",
  ["diakonia (mainz, germany : 1972)"] = "Diakonia",
  ["dialectica (bern, switzerland)"] = "Dialectica (Bern)",
  ["dialectical anthropology"] = "Dialect Anthropol",
  ["dialectics and humanism"] = "Dialect Hum",
  ["dializ, transplantasyon ve yanik : türkiye organ nakli ve yanik tedavi vakfi, tip bilimleri dergisi = dialysis, transplantation & burn : medical journal of the turkish transplantation and burn foundation"] = "Dializ Transplant Yanik",
  ["dialog: fairleigh dickinson university school of dentistry"] = "Dialog. Fairleigh Dickinson Univ. Sch. Dent.",
  ["dialoghi di archeologia"] = "DArch",
  ["dialogue (los angeles, calif. : 1986)"] = "Dialogue",
  ["dialogue (saint-etienne, loire, france)"] = "Dialogue (Lyon)",
  ["dialogue on diarrhoea"] = "Dialogue Diarrhoea",
  ["dialogues and games of logic"] = "Dialog. Games Log.",
  ["dialogues d’histoire ancienne"] = "DHA",
  ["dialogues in cardiovascular medicine : dcm"] = "Dialog Cardiovasc Med",
  ["dialogues in clinical neuroscience"] = "Dialogues Clin Neurosci",
  ["dialogues in contraception"] = "Dialogues Contracept",
  ["dialogues in human geography"] = "Dialogues Hum Geogr",
  ["dialogues on climate change"] = "Dialogues Clim. Change",
  ["dialysis & transplantation"] = "Dial Transplant",
  ["diamond (philadelphia, pa.)"] = "Diamond",
  ["diamond and related materials"] = "Diamond Relat. Mater.",
  ["diamond films and technology"] = "Diamond Films Technol.",
  ["dian hua xue"] = "Dian Hua Xue",
  ["dian zi xian wei xue bao = journal of chinese electron microscopy society"] = "Dian Zi Xian Wei Xue Bao",
  ["dianzi keji daxue xuebao"] = "Dianzi Keji Daxue Xuebao",
  ["diario de centro america (guatemala, guatemala : 1880)"] = "D Cent Am",
  ["diario de centro américa (guatemala, guatemala : 1880)"] = "D Cent Am",
  ["diario de yucatan"] = "D Yucatan",
  ["diario de yucatán"] = "D Yucatan",
  ["diarrhoea dialogue"] = "Diarrhoea Dialogue",
  ["diatom research"] = "Diatom Res.",
  ["diatom research : the journal of the international society for diatom research"] = "Diatom Res",
  ["dickinson law review"] = "Dickinson Law Rev",
  ["dicle university tip fakultesi dergisitesi"] = "Dicle Univ. Tip Fakul. Derg.",
  ["dicle üniversitesi tip fakültesi dergisi = the journal of the faculty of medicine, university of dicle"] = "Dicle Univ Tip Fakul Derg",
  ["dicp : the annals of pharmacotherapy"] = "DICP",
  ["dicta : digital image computing : techniques and applications. australian pattern recognition society. conference"] = "DICTA",
  ["dictionnaire des antiquités grecques et romaines d’après les textes et les monuments. ouvrage rédigé sous la direction de ch. daremberg et e. saglio"] = "Daremberg-Saglio",
  ["dictionnaire d’archéologie chrétienne et de liturgie"] = "DACL",
  ["didattica: proposte ed esperienze"] = "Didatt. Propos. Esper.",
  ["die agnes karll-schwester, der krankenpfleger"] = "Agnes Karll Schwest Krankenpfleger",
  ["die alte stadt: vierteljahreszeitschrift für stadtgeschichte, stadtsoziologie und denkmalpflege"] = "ASTADT",
  ["die antiken sarkophagreliefs"] = "ASR",
  ["die arztliche fortbildung"] = "Arztl Fortbild",
  ["die ärztliche fortbildung"] = "Arztl Fortbild",
  ["die dritte welt"] = "Dritte Welt",
  ["die erde; zeitschrift der gesellschaft fur erdkunde zu berlin"] = "Erde",
  ["die erde; zeitschrift der gesellschaft für erdkunde zu berlin"] = "Erde",
  ["die fundmünzen der römischen zeit in deutschland"] = "FMRD",
  ["die grundlehren der mathematischen wissenschaften in einzeldarstellungen mit besonderer berucksichtigung der anwendungsgebiete"] = "Die Grundlehren der mathematischen Wissenschaften in Einzeldarstellungen mit besonderer Berucksichtigung der Anwendung",
  ["die grundlehren der mathematischen wissenschaften in einzeldarstellungen mit besonderer berucksightigung der anwendungsgebiete"] = "Die Grundlehren der mathematischen Wissenschaften in Einzeldarstellungen mit besonderer Berucksightigung der Anwendung",
  ["die heilberufe"] = "Heilberufe",
  ["die heilkunst"] = "Heilkunst",
  ["die infusionstherapie"] = "Infusionstherapie",
  ["die krankenschwester"] = "Krankenschwester",
  ["die kulturpflanze : berichte und mitteilungen aus dem institut fur kulturpflanzenforschung der deutschen akademie der wissenschaften zu berlin in gatersleben krs. aschersleben"] = "Kulturpflanze",
  ["die kulturpflanze : berichte und mitteilungen aus dem institut für kulturpflanzenforschung der deutschen akademie der wissenschaften zu berlin in gatersleben krs. aschersleben"] = "Kulturpflanze",
  ["die medizinische"] = "Medizinische",
  ["die medizinische welt"] = "Med Welt",
  ["die nahrung"] = "Nahrung",
  ["die naturheilkunde"] = "Naturheikunde",
  ["die naturwissenschaften"] = "Naturwissenschaften",
  ["die neue rundschau"] = "Neue Rundsch",
  ["die pharmazeutische industrie"] = "Pharm Ind",
  ["die pharmazie"] = "Pharmazie",
  ["die pharmazie. beihefte"] = "Pharm Beih",
  ["die politische meinung"] = "Polit Mein",
  ["die quintessenz"] = "Quintessenz",
  ["die quintessenz der zahntechnik"] = "Quintessenz Zahntech",
  ["die rehabilitation"] = "Rehabilitation (Stuttg)",
  ["die schweizer hebamme"] = "Schweiz Hebamme",
  ["die sonde"] = "Sonde",
  ["die stärke = starch"] = "Starke",
  ["die suid-afrikaanse tydskrif vir natuurwetenskap en tegnologie"] = "Suid-Afrikaanse Tydskr. Natuurwetenskap Tegnol.",
  ["die suid-afrikaanse tydskrif vir sosiologie. the south african journal of sociology"] = "S Afr Tydskr Sosiol",
  ["die therapiewoche"] = "Therapiewoche",
  ["die tuberkulose und ihre grenzgebiete in einzeldarstellungen"] = "Tuberk Grenzgeb Einzeldarst",
  ["die unterrichtspraxis. teaching german"] = "Unterrichtspraxis",
  ["die waage"] = "Waage",
  ["die welt als geschichte"] = "WeltGesch",
  ["die welt des orients. wissenschaftliche beiträge zur kunde des morgenlandes"] = "WO",
  ["die welt des orients: wissenschaftliche beiträge zur kunde des morgenlandes"] = "WO",
  ["difesa sociale"] = "Dif Soc",
  ["differential and integral equations"] = "Differ. Integr. Equations",
  ["differential and integral equations. an international journal for theory & applications"] = "Differential Integral Equations",
  ["differential equations"] = "Differ. Equations",
  ["differential equations & applications"] = "Differ. Equ. Appl.",
  ["differential equations and dynamical systems"] = "Differ Equ Dyn Syst",
  ["differential equations and dynamical systems. international journal for theory, real world modelling and simulations"] = "Differ. Equ. Dyn. Syst.",
  ["differential geometry and its applications"] = "Differential Geom. Appl.",
  ["differential geometry—dynamical systems"] = "Differ. Geom. Dyn. Syst.",
  ["differential-algebraic equations forum"] = "Differ.-Algebr. Equ. Forum",
  ["differentiation (malden, ma, united states)"] = "Differentiation (Malden, MA, U. S.)",
  ["differentiation (oxford, united kingdom)"] = "Differentiation (Oxford, U. K.)",
  ["differentiation; research in biological diversity"] = "Differentiation",
  ["differentsial\\cprime naya geometriya mnogoobraziy figur"] = "Differentsial\\cprime naya Geom. Mnogoobraz. Figur",
  ["differentsial\\cprime nye uravneniya"] = "Differ. Uravn.",
  ["differentsialnaya geometriya mnogoobrazii figur"] = "Differentsialnaya Geom. Mnogoobraz. Figur",
  ["differentsialnye uravneniya i protsessy upravleniya. differential equations and control processes"] = "Differ. Uravn. Protsessy Upr.",
  ["diffusion and defect data–solid state data, pt. a: defect and diffusion forum"] = "Diffus. Defect Data, Pt. A",
  ["diffusion and defect data–solid state data, pt. b: solid state phenomena"] = "Diffus. Defect Data, Pt. B",
  ["digest journal of nanomaterials and biostructures"] = "Dig J Nanomater Biostruct",
  ["digest of neurology and psychiatry"] = "Dig Neurol Psychiatr",
  ["digest of public general bills and selected resolutions, with index"] = "Digest Public Gen Bills",
  ["digest of technical papers. ieee international solid-state circuits conference"] = "Dig Tech Pap IEEE Int Solid State Circuits Conf",
  ["digest of technical papers. international conference on solid-state sensors, actuators, and microsystems"] = "Dig Tech Papers",
  ["digest of technical papers. sid international symposium"] = "Dig Tech Pap",
  ["digesta antibiotica"] = "Dig Antibiot",
  ["digestive and liver disease"] = "Dig. Liver Dis.",
  ["digestive and liver disease : official journal of the italian society of gastroenterology and the italian association for the study of the liver"] = "Dig Liver Dis",
  ["digestive disease interventions"] = "Dig Dis Interv",
  ["digestive diseases"] = "Dig. Dis.",
  ["digestive diseases (basel, switzerland)"] = "Dig Dis",
  ["digestive diseases and sciences"] = "Dig Dis Sci",
  ["digestive endoscopy : official journal of the japan gastroenterological endoscopy society"] = "Dig Endosc",
  ["digestive medicine research"] = "Dig Med Res",
  ["digestive surgery"] = "Dig Surg",
  ["digital applications in archaeology and cultural heritage"] = "Digital Appl. Archaeol. Cult. Heritage",
  ["digital biomarkers"] = "Digit Biomark",
  ["digital chemical engineering"] = "Digital Chem. Eng.",
  ["digital communications and networks"] = "Digital Commun. Networks",
  ["digital culture & education"] = "Digit Cult Educ",
  ["digital discovery"] = "Digital Discovery",
  ["digital engineering and digital twin"] = "Digital Eng. Digital Twin",
  ["digital experiences in mathematics education"] = "Digital Exper. Math. Educ.",
  ["digital finance"] = "Digit Finance",
  ["digital geography and society"] = "Digital Geogr. Soc.",
  ["digital government: research and practice"] = "Digital Gov.: Res. Pract.",
  ["digital health"] = "Digit Health",
  ["digital investigation"] = "Digit Investig",
  ["digital journal of ophthalmology : djo"] = "Digit J Ophthalmol",
  ["digital journalism (abingdon, england)"] = "Digit Journal (Abingdon)",
  ["digital libraries for cultural heritage, knowledge dissemination, and future creation : 13th international conference on asia-pacific digital libraries, icadl 2011, beijing, china, october 24-27, 2011. international conference on asia-pacific digital libraries (13th : 2011 : beijing, china)"] = "Digit Libraries Cult Herit Knowl Dissem Future Creat (2011)",
  ["digital library perspectives"] = "Digit Libr Perspect",
  ["digital medicine"] = "Digit Med",
  ["digital pathology : 15th european congress, ecdp 2019, warwick, uk, april 10-13, 2019, proceedings. european congress on digital pathology (15th : 2019 : warwick, united kingdom)"] = "Digit Pathol (2019)",
  ["digital signal processing"] = "Digital Signal Process.",
  ["digital threats: research and practice"] = "Digital Threats: Res. Pract.",
  ["digital twins and applications"] = "Digital Twins Appl.",
  ["digital war"] = "Digital War",
  ["digitalbiomarkers'17 : proceedings of the 1st workshop on digital biomarkers : june 23, 2017, niagara falls, ny, usa. workshop on digital biomarkers (1st : 2017 : niagara falls, n.y.)"] = "DigitalBiomarkers 17 (2017)",
  ["digitale bilddiagnostik"] = "Digitale Bilddiagn",
  ["dike. rivista di storia del diritto greco ed ellenistico. università degli studi di milano. facoltà di giurisprudenza"] = "Dike",
  ["dimacs series in discrete mathematics and theoretical computer science"] = "DIMACS Ser. Discrete Math. Theoret. Comput. Sci.",
  ["dimensions (new york, n.y.)"] = "Dimensions (N Y N Y)",
  ["dimensions in health service"] = "Dimens Health Serv",
  ["dimensions in oncology nursing"] = "Dimens. Oncol. Nurs.",
  ["dimensions in oncology nursing : journal of the division of nursing"] = "Dimens Oncol Nurs",
  ["dimensions of critical care nursing"] = "Dimens. Crit. Care Nurs.",
  ["dimensions of critical care nursing : dccn"] = "Dimens Crit Care Nurs",
  ["dimensions of dental hygiene"] = "Dimens Dent Hyg",
  ["dinamika i prochnost\\cprime mashin"] = "Dinamika i Prochnost\\cprime Mashin",
  ["dine yisrael : shanaton le-mishpat ’ivri ule-mishpahah be-yisra’el"] = "Dyny Ysrl",
  ["dine yiśraʼel : shanaton le-mishpaṭ ʻivri ule-dine mishpaḥah be-yiśraʼel"] = "Dyny Ysrl",
  ["dioniso. annale della fondazione inda, istituto nazionale del dramma antico"] = "Dioniso",
  ["dipartimento di metodi e modelli matematici per le scienze applicate"] = "Dip. Metodi Modelli Mat. Sci. Appl.",
  ["dipendenze patologiche"] = "Dipend Patologiche",
  ["diplomacy and statecraft"] = "Dipl Statecraft",
  ["diqiu huaxue"] = "Diqiu Huaxue",
  ["dirasat sukkaniyah"] = "Dirasat Sukkaniyah",
  ["dirasat. medical and biological sciences"] = "Dirasat Med Biol Sci",
  ["dirasat. series b, pure and applied sciences"] = "Dirasat B Pure Appl Sci",
  ["dirāsāt sukkānīyah"] = "Dirasat Sukkaniyah",
  ["directions in chaos"] = "Dir. Chaos",
  ["directions in psychiatry"] = "Dir Psychiatry",
  ["director (cincinnati, ohio)"] = "Director",
  ["directors & boards"] = "Dir Boards",
  ["directors and boards"] = "Dir. Boards",
  ["directory of archives and manuscript repositories"] = "Dir Arch Manuscr Repos",
  ["dis. designing interactive systems (conference)"] = "DIS (Des Interact Syst Conf)",
  ["disability & society"] = "Disabil Soc",
  ["disability and health journal"] = "Disabil Health J",
  ["disability and rehabilitation"] = "Disabil Rehabil",
  ["disability and rehabilitation. assistive technology"] = "Disabil Rehabil Assist Technol",
  ["disability studies quarterly"] = "Disabil Stud Q",
  ["disability, handicap & society"] = "Disabil Handicap Soc",
  ["disaster and military medicine"] = "Disaster Mil Med",
  ["disaster health"] = "Disaster Health",
  ["disaster management & response : dmr : an official publication of the emergency nurses association"] = "Disaster Manag Response",
  ["disaster medicine and public health preparedness"] = "Disaster Med Public Health Prep",
  ["disaster prevention and management"] = "Disaster Prev Manag",
  ["discharge planning update"] = "Disch Plann Update",
  ["discourse & society"] = "niscl. Soc.",
  ["discourse processes"] = "Discourse Process",
  ["discourse studies"] = "Discourse Stud",
  ["discourses in mathematics and its applications"] = "Discourses Math. Appl.",
  ["discover artificial intelligence"] = "Discover Artif. Intell.",
  ["discover chemical engineering"] = "Discover Chem. Eng.",
  ["discover energy"] = "Discover Energy",
  ["discover food"] = "Discover Food",
  ["discover internet of things"] = "Discover Internet Things",
  ["discover materials"] = "Discov Mater",
  ["discover mechanical engineering"] = "Discover Mech. Eng.",
  ["discover sustainability"] = "Discover Sustainability",
  ["discover water"] = "Discover Water",
  ["discoveries (craiova, romania)"] = "Discoveries (Craiova)",
  ["discovering the subtleties of sugars : proceedings of the 3rd beilstein glyco-bioinformatics symposium : june 10th - 14th, 2013, potsdam, germany. international beilstein symposium on glyco-bioinformatics (3rd : 2013 : potsdam, germany)"] = "Discov Subtleties Sugars (2013)",
  ["discovery (austin, tex.)"] = "Discovery",
  ["discovery and innovation"] = "Discov Innov",
  ["discovery immunotherapy"] = "Discovery Immunother.",
  ["discovery medicine"] = "Discov Med",
  ["discrete & computational geometry"] = "Discrete Comput. Geom.",
  ["discrete & computational geometry. an international journal of mathematics and computer science"] = "Discrete Comput. Geom.",
  ["discrete analysis"] = "Discrete Anal.",
  ["discrete and continuous dynamical systems"] = "Discrete Contin. Dyn. Syst.",
  ["discrete and continuous dynamical systems - series s"] = "Discrete Contin. Dyn. Syst. - Ser. S",
  ["discrete and continuous dynamical systems series b"] = "Discrete Contin. Dyn. Syst. - Ser. B",
  ["discrete and continuous dynamical systems. series a"] = "Discrete Contin Dyn Syst Ser A",
  ["discrete and continuous dynamical systems. series b"] = "Discrete Continuous Dyn Syst Ser B",
  ["discrete and continuous dynamical systems. series b. a journal bridging mathematics and sciences"] = "Discrete Contin. Dyn. Syst. Ser. B",
  ["discrete and continuous dynamical systems. series s"] = "Discrete Contin. Dyn. Syst. Ser. S",
  ["discrete applied mathematics"] = "Discrete Appl. Math.",
  ["discrete applied mathematics (amsterdam, netherlands : 1988)"] = "Discrete Appl Math",
  ["discrete applied mathematics. the journal of combinatorial algorithms, informatics and computational sciences"] = "Discrete Appl. Math.",
  ["discrete dynamics in nature and society"] = "Discrete Dyn Nat Soc",
  ["discrete dynamics in nature and society. an international multidisciplinary research and review journal"] = "Discrete Dyn. Nat. Soc.",
  ["discrete event dynamic systems"] = "Discret Event Dyn Syst",
  ["discrete event dynamic systems theory and applications"] = "Discrete Event Dyn. Syst. Theory Appl.",
  ["discrete event dynamic systems: theory and applications"] = "Discrete Event Dyn. Syst.",
  ["discrete mathematics"] = "Discrete Math.",
  ["discrete mathematics & theoretical computer science proceedings"] = "Discrete Math. Theor. Comput. Sci. Proc.",
  ["discrete mathematics & theoretical computer science. dmtcs."] = "Discrete Math. Theor. Comput. Sci.",
  ["discrete mathematics and applications"] = "Discrete Math. Appl.",
  ["discrete mathematics and its applications (boca raton)"] = "Discrete Math. Appl. (Boca Raton)",
  ["discrete mathematics and theoretical computer science"] = "Discrete Math. Theor. Comput. Sci.",
  ["discrete mathematics letters"] = "Discrete Math. Lett.",
  ["discrete mathematics, algorithms and applications"] = "Discrete Math., Algorithms Appl.",
  ["discrete mathematics, algorithms, and applications"] = "Discrete Math Algorithms Appl",
  ["discrete optimization"] = "Discrete Optim.",
  ["discussion paper (university of toronto. dept. of geography)"] = "Discuss Pap Univ Tor Dep Geogr",
  ["discussiones mathematicae"] = "Discuss. Math. Algebra Stochastic Methods",
  ["discussiones mathematicae. differential inclusions, control and optimization"] = "Discuss. Math. Differ. Incl. Control Optim.",
  ["discussiones mathematicae. general algebra and applications"] = "Discuss. Math. Gen. Algebra Appl.",
  ["discussiones mathematicae. graph theory"] = "Discuss. Math. Graph Theory",
  ["discussiones mathematicae. probability and statistics"] = "Discuss. Math. Probab. Stat.",
  ["discussions of the faraday society"] = "Discuss Faraday Soc",
  ["disease management : dm"] = "Dis Manag",
  ["disease management advisor"] = "Dis Manag Advis",
  ["disease management and clinical outcomes"] = "Dis Manag Clin Outcomes",
  ["disease markers"] = "Dis Markers",
  ["disease models & mechanisms"] = "Dis Model Mech",
  ["disease-a-month : dm"] = "Dis Mon",
  ["diseases (basel, switzerland)"] = "Diseases",
  ["diseases in asian aquaculture"] = "Dis Asian Aquac",
  ["diseases of aquatic organisms"] = "Dis Aquat Organ",
  ["diseases of the chest"] = "Dis Chest",
  ["diseases of the colon and rectum"] = "Dis Colon Rectum",
  ["diseases of the esophagus"] = "Dis. Esophagus",
  ["diseases of the esophagus : official journal of the international society for diseases of the esophagus"] = "Dis Esophagus",
  ["diseases of the esophagus : official journal of the international society for diseases of the esophagus / i.s.d.e"] = "Dis Esophagus",
  ["diseases of the nervous system"] = "Dis Nerv Syst",
  ["diski. dissertationen zur kunstlichen intelligenz"] = "DISKI",
  ["diskretnaya matematika"] = "Diskret. Mat.",
  ["diskretnyi analiz i issledovanie operatsii"] = "Diskretn. Anal. Issled. Oper.",
  ["diskretnyy analiz i issledovanie operatsiy"] = "Diskret. Anal. Issled. Oper.",
  ["diskussionen zur archäologischen bauforschung"] = "DiskAB",
  ["diskussionsforum medizinische ethik"] = "Diskussionsforum Med Ethik",
  ["disruptive science and technology"] = "Disrupt Sci Technol",
  ["dissertation abstracts"] = "Diss Abstr",
  ["dissertation abstracts international"] = "Diss Abstr Int",
  ["dissertation eth"] = "Diss. ETH",
  ["dissertation summaries in mathematics"] = "Diss. Summ. Math.",
  ["dissertationes botanicae"] = "Diss Bot",
  ["dissertationes mathematicae"] = "Dissertationes Math.",
  ["dissertationes mathematicae (rozprawy matematyczne)"] = "Dissertationes Math. (Rozprawy Mat.)",
  ["dissertationes mathematicae universitatis tartuensis"] = "Diss. Math. Univ. Tartu.",
  ["dissertations abstracts: international abstracts of dissertation available in microfilm or as xerographic reproductions"] = "DA",
  ["dissolution technologies"] = "Dissolut Technol",
  ["distinguished dissertations in computer science"] = "Disting. Diss. Comput. Sci.",
  ["distributed and parallel databases"] = "Distrib Parallel Databases",
  ["distributed computing"] = "Distrib. Comput.",
  ["distributed generation & alternative energy journal"] = "Distrib. Gener. Altern. Energy J.",
  ["distributed ledger technologies: research and practice"] = "Distrib. Ledger Technol.: Res. Pract.",
  ["distributed systems engineering"] = "Distrib. Sys. Eng.",
  ["district nursing"] = "Dist Nurs",
  ["district of columbia dental society journal. district of columbia dental society"] = "J Dist Columbia Dent Soc",
  ["diversity & distributions"] = "Divers Distrib",
  ["diversity and distributions"] = "Divers. Distrib.",
  ["diversity and equality in health and care"] = "Divers Equal Health Care",
  ["diving and hyperbaric medicine"] = "Diving Hyperb Med",
  ["divulgacion cultural odontologica"] = "Divulg. Cult. Odontol.",
  ["divulgacion cultural odontológica"] = "Divulg Cult Odontol",
  ["divulgación médico-dental; boletín de educación médico-dental y especialidades"] = "Divulg Med Dent",
  ["dix-huitieme siecle"] = "Dix Huit Siecle",
  ["dix-huitième siècle"] = "Dix Huit Siecle",
  ["dlr-nachrichten : mitteilungsblatt der deutschen forschungsanstalt fur luft- und raumfahrt"] = "DLR Nachr",
  ["dlr-nachrichten : mitteilungsblatt der deutschen forschungsanstalt für luft- und raumfahrt"] = "DLR Nachr",
  ["dmv seminar"] = "DMV Sem.",
  ["dna (mary ann liebert, inc.)"] = "DNA",
  ["dna and cell biology"] = "DNA Cell Biol",
  ["dna barcodes (berlin)"] = "DNA Barcodes (Berlin)",
  ["dna repair"] = "DNA Repair (Amst)",
  ["dna research"] = "DNA Res.",
  ["dna research : an international journal for rapid publication of reports on genes and genomes"] = "DNA Res",
  ["dna sequence"] = "DNA Seq.",
  ["dna sequence : the journal of dna sequencing and mapping"] = "DNA Seq",
  ["dobutsu iden ikushu kenkyu"] = "Journal of animal genetics|Dobutsu Iden Ikushu Kenkyu",
  ["doctoral thesis"] = "Doctor. Thesis",
  ["documenta (centro de documentación e información del ipss)"] = "Documenta",
  ["documenta albana"] = "DocAlb",
  ["documenta de medicina geographica et tropica"] = "Doc Med Geogr Trop",
  ["documenta dell'istituto italiano di idrobiologia. istituto italiano di idrobiologia (pallanza, italy)"] = "Doc Ist Ital Idrobiol Dott Marco De Marchi",
  ["documenta faunistica helvetiae"] = "Doc. Faun. Helv.",
  ["documenta geigy. acta psychosomatica. [deutsche ausg]"] = "Doc Geigy Acta Psychosom (Dtsch Ausg)",
  ["documenta geigy. acta rheumatologica. [english ed.]"] = "Doc Geigy Acta Rheumatol",
  ["documenta geigy. series chirurgica"] = "Doc Geigy Ser Chir",
  ["documenta mathematica"] = "Doc. Math.",
  ["documenta neerlandica et indonesica de morbis tropicis; quarterly journal of tropical medicine and hygiene"] = "Doc Neerl Indones Morbis Trop",
  ["documenta ophthalmologica"] = "Doc. Ophthalmol.",
  ["documenta ophthalmologica. advances in ophthalmology"] = "Doc Ophthalmol",
  ["documenta ophthalmologica. proceedings series"] = "Doc Ophthalmol Proc Ser",
  ["documenta rheumatologica. [deutsche ausg.]"] = "Doc Rheumatol",
  ["documenta rheumatologica. [english ed.]"] = "Doc Rheumatol (1956)",
  ["documentation in medical ethics"] = "Doc Med Ethics",
  ["documentation sur l'europe centrale"] = "Doc Eur Cent",
  ["documenti. istituto per i beni artistici, culturali, naturali della regione emilia-romagna"] = "DocEmRom",
  ["documents (paris, france : 1967)"] = "Documents",
  ["documents d'anàlisi geogràfica"] = "Doc Anal Geogr",
  ["documents d’analisi geografica / [publicacions del departament de geografia, universitat autonoma de barcelona]"] = "Doc Anal Geogr",
  ["documents d’archéologie méridionale"] = "DocAMerid",
  ["documents d’archéologie régionale. université catholique de louvain"] = "DocALouv",
  ["documents mathematiques (paris)"] = "Doc. Math. (Paris)",
  ["documents mycologiques"] = "Doc Mycol",
  ["documents phytosociologiques, nouvelle série"] = "Doc. phytosociol., nouv. sér.",
  ["documents pour l'histoire du vocabulaire scientifique"] = "Doc Hist Vocab Sci",
  ["documents pour l’histoire du vocabulaire scientifique / centre national de la recherche scientifique, greco, histoire du vocabulaire scientifique"] = "Doc Hist Vocab Sci",
  ["documents. international chamber of commerce"] = "Documents",
  ["dokladi na b\\cdprime lgarskata akademiya na naukite"] = "C. R. Acad. Bulgare Sci.",
  ["dokladi na blgarskata akademiya na naukite. comptes rendus de l’academie bulgare des sciences"] = "C. R. Acad. Bulgare Sci.",
  ["dokladi na bulgarskata akademiya na naukite"] = "Dokl. Bulg. Akad. Nauk",
  ["doklady - akademiya nauk azerbaidzhana"] = "Dokl. Akad. Nauk Az. SSR",
  ["doklady akademii nauk"] = "Dokl. Akad. Nauk",
  ["doklady akademii nauk / [rossiiskaia akademii nauk]"] = "Dokl Akad Nauk",
  ["doklady akademii nauk belarusi"] = "Dokl. Akad. Nauk Belarusi",
  ["doklady akademii nauk bssr"] = "Dokl Akad Nauk",
  ["doklady akademii nauk soiuza sovetskikh sotsialisticheskikh respublik"] = "Dokl Akad Nauk Soiuza Sov Sotsialisticheskikh Resp",
  ["doklady akademii nauk sssr"] = "Dokl Akad Nauk SSSR",
  ["doklady akademii nauk sssr [soviet physics–doklady]"] = "Dokl. Akad. Nauk SSSR [Sov. Phys. Dokl.]",
  ["doklady akademii nauk tadzhikskoi ssr"] = "Dokl. Akad. Nauk Tadzh. SSR",
  ["doklady akademii nauk ukrainy"] = "Dokl. Akad. Nauk Ukr. SSR",
  ["doklady akademii nauk uzssr"] = "Dokl. Akad. Nauk UzSSR",
  ["doklady biochemistry : proceedings of the academy of sciences of the ussr, biochemistry section"] = "Dokl Biochem",
  ["doklady biochemistry : proceedings of the academy of sciences of the ussr, biochemistry section / translated from russian"] = "Dokl Biochem",
  ["doklady biochemistry and biophysics"] = "Dokl. Biochem. Biophys.",
  ["doklady biological sciences"] = "Dokl. Biol. Sci.",
  ["doklady biological sciences : proceedings of the academy of sciences of the ussr, biological sciences sections"] = "Dokl Biol Sci",
  ["doklady biological sciences : proceedings of the academy of sciences of the ussr, biological sciences sections / translated from russian"] = "Dokl Biol Sci",
  ["doklady biophysics"] = "Dokl. Biophys.",
  ["doklady biophysics : proceedings of the academy of sciences of the ussr, biophysics section"] = "Dokl Biophys",
  ["doklady biophysics : proceedings of the academy of sciences of the ussr, biophysics section / translated from russian"] = "Dokl Biophys",
  ["doklady bolgarskoi akademii nauk"] = "Dokl. Bolg. Akad. Nauk",
  ["doklady botanical sciences"] = "Dokl. Bot. Sci.",
  ["doklady chemical technology"] = "Dokl. Chem. Technol.",
  ["doklady chemistry"] = "Dokl. Chem.",
  ["doklady earth sciences"] = "Dokl. Earth Sci.",
  ["doklady mathematics"] = "Dokl. Math.",
  ["doklady natsionalnoi akademii nauk belarusi"] = "Dokl. Nats. Akad. Nauk Belarusi",
  ["doklady natsionalʹnoĭ akademii nauk belarusi"] = "Dokl Natsionalnoi Akad Nauk Belarusi",
  ["doklady physical chemistry"] = "Dokl. Phys. Chem.",
  ["doklady physics"] = "Dokl. Phys.",
  ["doklady rossiĭskoĭ akademii selʹskokhozi︠a︡ĭstvennykh nauk"] = "Dokl Ross Akademii Selskokhoziaistvennykh Nauk",
  ["doklady. biochemistry and biophysics"] = "Dokl Biochem Biophys",
  ["doklady. natsionalnaya akademiya nauk armenii. reports. national academy of sciences of armenia"] = "Dokl. Nats. Akad. Nauk Armen.",
  ["doklady. natsionalnaya akademiya nauk azerbaidzhana. reports. national academy of sciences of azerbaijan"] = "Dokl. Nats. Akad. Nauk Azerb.",
  ["dokument bodenkundliche gesellschaft der schweiz"] = "Dok. Bodenkd. Ges. Schweiz",
  ["dokumente aus hoechster archiven"] = "Dok Hoechst Arch",
  ["dokumente und informationen zur schweizerischen orts-, regional- und landesplanung, disp"] = "Dok. Inf. schweiz. Orts- Reg.- Landesplan., DISP",
  ["dokumente zur geschichte der mathematik"] = "Dokumente Gesch. Math.",
  ["dolentium hominum"] = "Dolentium Hominum",
  ["dolomites research notes on approximation"] = "Dolomites Res. Notes Approx.",
  ["domain adaptation and representation transfer and medical image learning with less labels and imperfect data : first miccai workshop, dart 2019, and first international workshop, mil3id 2019, shenzhen, held in conjunction with miccai 2019, shenzhen, china, october 13 and 17, 2019, proceedings. dart (workshop) (1st : 2019 : shenzhen shi, china)"] = "Domain Adapt Represent Transf Med Image Learn Less Labels Imperfect Data (2019)",
  ["domain decomposition methods in science and engineering xx. international conference on domain decomposition methods for partial differential equations (20th : 2011 : san diego, calif.)"] = "Domain Decompos Method Sci Eng XX (2011)",
  ["domes : digest of middle east studies"] = "Domes",
  ["domestic animal endocrinology"] = "Domest Anim Endocrinol",
  ["dominion law reports"] = "Dom Law Rep",
  ["domus medici"] = "Domus Med",
  ["donetskiy gosudarstvennyy universitet"] = "Teoret. i Prikl. Mekh. (Donetsk)",
  ["dong bei nong ye da xue xue bao : journal of northeast agricultural university"] = "Dong Bei Nong Ye Da Xue Xue Bao",
  ["dong wu fen lei xue bao"] = "Acta zootaxonomica Sinica|Dong Wu Fen Lei Xue Bao",
  ["dong wu fen lei xue bao = acta zootaxonomica sinica"] = "Dong Wu Fen Lei Xue Bao",
  ["dong wu xue bao. [acta zoologica sinica]"] = "Dong Wu Xue Bao",
  ["dong wu xue yan jiu"] = 'Zoological research / "Dong wu xue yan jiu" bian ji we yuan hui bian ji|Zool. Res.',
  ["dong wu xue yan jiu = zoological research"] = "Dongwuxue Yanjiu",
  ["dong wu ying yang xue bao = acta zoonutrimenta sinica"] = "Dong Wu Ying Yang Xue Bao",
  ["dongbei lin ye da xue xue bao = journal of north-east forestry university"] = "Dongbei Lin Ye Da Xue Xue Bao",
  ["dongbei shida xuebao"] = "Dongbei Shida Xuebao",
  ["dongmul jawon yeon-gu"] = "Dongmul Jawon Yeongu",
  ["dongwu yixue jinzhan"] = "Dongwu Yixue Jinzhan",
  ["dopovidi akademiī nauk ukraīns'koī rsr. seriia b: heolohichni, khimichni ta biolohichni nauky"] = "Dopov Akad Nauk Ukr RSR B",
  ["dopovidi natsionalnoi akademii nauk ukraini. matematika. prirodoznavstvo. tekhnichni nauki"] = "Dopov. Nats. Akad. Nauk Ukr. Mat. Prirodozn. Tekh. Nauki",
  ["dopovidi natsional’noi akademii nauk ukraini"] = "Dopov. Nats. Akad. Nauk Ukr.",
  ["dopovidi nat︠s︡ionalʹnoï akademiï nauk ukraïny. matematyka, pryrodoznavstvo, tekhnichni nauky"] = "Dopov Natsionalnoi Akad Nauk Ukr Matemat Pryrodoznavstvo Tekhnichni Nauky",
  ["dose-response : a publication of international hormesis society"] = "Dose Response",
  ["douleurs : evaluation - diagnostic - traitement"] = "Douleurs: Eval. Diagn. Trait.",
  ["down to earth"] = "Down Earth",
  ["down's syndrome, research and practice : the journal of the sarah duffen centre"] = "Downs Syndr Res Pract",
  ["downs syndrome research and practice"] = "Downs Syndr Res Pract",
  ["downside review"] = "DR",
  ["down’s syndrome, research and practice : the journal of the sarah duffen centre / university of portsmouth"] = "Downs Syndr Res Pract",
  ["dòngwùyuán xuébào"] = "Dongwuyuan Xuebao",
  ["dōbutsu iden ikushu kenkyū = journal of animal genetics"] = "Dobutsu Iden Ikushu Kenkyu",
  ["dp [reports]. u.s. atomic energy commission"] = "DP Rep",
  ["dp reports"] = "DP Rep.",
  ["dp. dental practice"] = "Dent Pract (Ewell)",
  ["dp: dental practice"] = "Dent. Pract. (Ewell)",
  ["draco pro medico"] = "Draco med",
  ["drake law review"] = "Drake Law Rev",
  ["draper fund report"] = "Draper Fund Rep",
  ["draper world population fund report"] = "Draper World Popul Fund Rep",
  ["dreaming : journal of the association for the study of dreams"] = "Dreaming",
  ["drevnejsije gosudarstva na territorii sssr (les états les plus anciens sur le territoire de l’urss: matériaux et recherches)"] = "DGT",
  ["drg monitor"] = "DRG Monit",
  ["driemaandelijks tijdschrift - gemeentekrediet van belgie"] = "Driem Tijdschr Gemeentekrediet Belg",
  ["drinking water engineering and science"] = "Drinking Water Eng. Sci.",
  ["drinking water engineering and science discussions"] = "Drinking Water Eng. Sci. Discuss.",
  ["droit social"] = "Droit Soc",
  ["drone systems and applications"] = "Drone Syst. Appl.",
  ["drosophila information service"] = "Drosoph Inf Serv",
  ["drug and alcohol dependence"] = "Drug Alcohol Depend",
  ["drug and alcohol dependence reports"] = "Drug Alcohol Depend. Rep.",
  ["drug and alcohol review"] = "Drug Alcohol Rev",
  ["drug and chemical toxicology"] = "Drug Chem Toxicol",
  ["drug and chemical toxicology (1977)"] = "Drug Chem. Toxicol. (1977)",
  ["drug and cosmetic industry"] = "Drug Cosmet Ind",
  ["drug and therapeutics bulletin"] = "Drug Ther Bull",
  ["drug court review"] = "Drug Court Rev",
  ["drug delivery"] = "Drug Deliv",
  ["drug delivery (london, england. 2007)"] = "Drug Deliv (Lond)",
  ["drug delivery and translation research"] = "Drug Delivery Transl. Res.",
  ["drug delivery and translational research"] = "Drug Deliv Transl Res",
  ["drug delivery letters"] = "Drug Deliv Lett",
  ["drug design and delivery"] = "Drug Des Deliv",
  ["drug design and discovery"] = "Drug Des Discov",
  ["drug design and discovery (yverdon-les-bains"] = "Drug Des Discov",
  ["drug design development and therapy"] = "Drug Des. Dev. Ther.",
  ["drug design, development and therapy"] = "Drug Des Devel Ther",
  ["drug designing : open access"] = "Drug Des",
  ["drug development and industrial pharmacy"] = "Drug Dev Ind Pharm",
  ["drug development research"] = "Drug Dev Res",
  ["drug discoveries & therapeutics"] = "Drug Discov Ther",
  ["drug discovery and development"] = "Drug Discovery Dev.",
  ["drug discovery today"] = "Drug Discov Today",
  ["drug discovery today. biosilico"] = "Drug Discov Today Biosilico",
  ["drug discovery today. disease mechanisms"] = "Drug Discov Today Dis Mech",
  ["drug discovery today. disease models"] = "Drug Discov Today Dis Models",
  ["drug discovery today. technologies"] = "Drug Discov Today Technol",
  ["drug discovery today. therapeutic strategies"] = "Drug Discov Today Ther Strateg",
  ["drug discovery today: technologies"] = "Drug Discovery Today: Technol.",
  ["drug forum"] = "Drug Forum",
  ["drug information journal"] = "Drug Inf J",
  ["drug intelligence & clinical pharmacy"] = "Drug Intell Clin Pharm",
  ["drug intelligence and clinical pharmacy"] = "Drug Intell. Clin. Pharm.",
  ["drug interactions newsletter"] = "Dr Interact Newsl",
  ["drug investigation"] = "Drug Investig",
  ["drug merchandising"] = "Drug Merch",
  ["drug metabolism & disposition"] = "Drug Metab. Dispos.",
  ["drug metabolism and disposition"] = "Drug Metab. Dispos.",
  ["drug metabolism and disposition: the biological fate of chemicals"] = "Drug Metab Dispos",
  ["drug metabolism and drug interactions"] = "Drug Metabol Drug Interact",
  ["drug metabolism and personalized therapy"] = "Drug Metab Pers Ther",
  ["drug metabolism and pharmacokinetics"] = "Drug Metab Pharmacokinet",
  ["drug metabolism letters"] = "Drug Metab Lett",
  ["drug metabolism reviews"] = "Drug Metab Rev",
  ["drug news & perspectives"] = "Drug News Perspect",
  ["drug repurposing, rescue, and repositioning"] = "Drug Repurposing Rescue Repositioning",
  ["drug research"] = "Drug Res (Stuttg)",
  ["drug research (stuttgart, germany)"] = "Drug Res. (Stuttgart, Ger.)",
  ["drug research reports"] = "Drug Res Rep",
  ["drug resistance updates"] = "Drug Resist. Updat.",
  ["drug resistance updates : reviews and commentaries in antimicrobial and anticancer chemotherapy"] = "Drug Resist Updat",
  ["drug safety"] = "Drug Saf",
  ["drug safety - case reports"] = "Drug Saf Case Rep",
  ["drug safety : an international journal of medical toxicology and drug experience"] = "Drug Saf",
  ["drug science, policy and law"] = "Drug Sci Policy Law",
  ["drug standards"] = "Drug Stand",
  ["drug target insights"] = "Drug Target Insights",
  ["drug target review"] = "Drug Target Rev",
  ["drug testing and analysis"] = "Drug Test  Anal",
  ["drug therapeutics"] = "Drug Ther",
  ["drug therapy"] = "Drug Ther (NY)",
  ["drug therapy. [hospital edition]"] = "Drug Ther Hosp Ed",
  ["drug topics"] = "Drug Topics",
  ["drug trade news"] = "Drug Trade News",
  ["drug, healthcare and patient safety"] = "Drug Healthc Patient Saf",
  ["drug-nutrient interactions"] = "Drug Nutr Interact",
  ["drugs & aging"] = "Drugs Aging",
  ["drugs & society (new york, n.y.)"] = "Drugs Soc (New York)",
  ["drugs & therapy perspectives"] = "Drugs Ther. Perspect.",
  ["drugs & therapy perspectives : for rational drug selection and use"] = "Drugs Ther Perspect",
  ["drugs (abingdon, england)"] = "Drugs (Abingdon Engl)",
  ["drugs - real world outcome"] = "Drugs Real World Outcome",
  ["drugs - real world outcomes"] = "Drugs Real World Outcomes",
  ["drugs and aging"] = "Drugs Aging",
  ["drugs and alcohol today"] = "Drugs Alcohol Today",
  ["drugs in context"] = "Drugs Context",
  ["drugs in r and d"] = "Drugs R. D.",
  ["drugs in r&d"] = "Drugs R D",
  ["drugs made in germany"] = "Drugs Made Ger",
  ["drugs of the future"] = "Drugs Future",
  ["drugs of today (barcelona, spain : 1998)"] = "Drugs Today (Barc)",
  ["drugs under experimental and clinical research"] = "Drugs Exp Clin Res",
  ["drustvena istrazivanja"] = "Drus. Istraz.",
  ["drustvo matematikov, fizikov in astronomov slovenije. obzornik za matematiko in fiziko. izdaja"] = "Obzornik Mat. Fiz.",
  ["društvo matematikov, fizikov in astronomov slovenije"] = "Obzornik Mat. Fiz.",
  ["drying technology"] = "Drying Technol.",
  ["dtw. deutsche tierarztliche wochenschrift"] = "DTW Dtsch Tierarztl Wochenschr",
  ["dtw. deutsche tierärztliche wochenschrift"] = "Dtsch Tierarztl Wochenschr",
  ["dtw: deutsche tierarztliche wochenschrift"] = "DTW Dtsch. Tierarztl. Wochenschr.",
  ["du bois review : social science research on race"] = "Du Bois Rev",
  ["dual diagnosis (foster city)"] = "Dual Diagn (Foster City)",
  ["dublin historical record"] = "Dublin Hist Rec",
  ["dublin journal of medical and chemical science"] = "Dublin J. Med. Chem. Sci.",
  ["dublin journal of medical science"] = "Dublin J. Med. Sci.",
  ["dublin university law journal"] = "Dubl Univ Law J",
  ["duke journal of gender law & policy"] = "Duke J Gend Law Policy",
  ["duke law and technology review"] = "Duke Law Technol Rev",
  ["duke law journal"] = "Duke Law J",
  ["duke mathematical journal"] = "Duke Math. J.",
  ["dukemedicine healthnews"] = "Duke Med Health News",
  ["dumbarton oaks papers"] = "DOP",
  ["duneland notes"] = "Dunel Notes",
  ["duodecim; laaketieteellinen aikakauskirja"] = "Duodecim",
  ["duodecim; laaketieteellinen aikakauskirja. supplementum"] = "Duodecim Suppl",
  ["duodecim; lääketieteellinen aikakauskirja"] = "Duodecim",
  ["duodecim; lääketieteellinen aikakauskirja. supplementum"] = "Duodecim Suppl",
  ["duquesne law review"] = "Duquesne Law Rev",
  ["durham university journal"] = "DUJ",
  ["dusseldorfer arbeiten zur geschichte der medizin"] = "Dusseld Arb Gesch Med",
  ["dusseldorfer arbeiten zur geschichte der medizin. beiheft"] = "Dusseld Arb Gesch Med Beih",
  ["dusseldorfer geographische schriften"] = "Dusseld Geogr Schr",
  ["dutch birding"] = "Dutch Bird",
  ["duv mathematik"] = "DUV Math.",
  ["duxbury advanced series"] = "Duxbury Adv. Ser.",
  ["düsseldorfer arbeiten zur geschichte der medizin"] = "Dusseld Arb Gesch Med",
  ["düsseldorfer geographische schriften"] = "Dusseld Geogr Schr",
  ["düşünen adam : bakırköy ruh ve sinir hastalıkları hastanesi yayın organı"] = "Dusunen Adam",
  ["dvs berichte"] = "DVS Ber.",
  ["dvt, dejiny ved a techniky"] = "Dejiny Ved Tech",
  ["dvt, dějiny věd a techniky"] = "Dejiny Ved Tech",
  ["dyes and pigments"] = "Dyes Pigm.",
  ["dyes and pigments : an international journal"] = "Dyes Pigm",
  ["dynamic biochemistry, process biotechnology & molecular biology"] = "Dyn Biochem Process Biotechnol Mol Biol",
  ["dynamic economics: theory and applications"] = "Dynam. Econom. Theory and Appl.",
  ["dynamic games and applications"] = "Dyn. Games Appl.",
  ["dynamic medicine : dm"] = "Dyn Med",
  ["dynamic modeling and econometrics in economics and finance"] = "Dyn. Model. Econom. Econ. Finance",
  ["dynamic systems and applications"] = "Dynam. Systems Appl.",
  ["dynamical systems an international journal"] = "Dyn. Syst. Int. J.",
  ["dynamical systems. an international journal"] = "Dyn. Syst.",
  ["dynamics (pembroke, ont.)"] = "Dynamics",
  ["dynamics and control"] = "Dyn. Control",
  ["dynamics and stability of systems"] = "Dynam. Stability Systems",
  ["dynamics and statistics of the climate system"] = "Dyn. Stat. Clim. Syst.",
  ["dynamics of atmospheres and oceans"] = "Dyn. Atmos. Oceans",
  ["dynamics of continuous, discrete & impulsive systems. series a. mathematical analysis"] = "Dyn. Contin. Discrete Impuls. Syst. Ser. A Math. Anal.",
  ["dynamics of continuous, discrete & impulsive systems. series b. applications & algorithms"] = "Dyn. Contin. Discrete Impuls. Syst. Ser. B Appl. Algorithms",
  ["dynamics of continuous, discrete and impulsive systems"] = "Dynam. Contin. Discrete Impuls. Systems",
  ["dynamics of partial differential equations"] = "Dyn. Partial Differ. Equations",
  ["dynamics reported: expositions in dynamical systems (new series)"] = "Dynam. Report. Expositions Dynam. Systems (N.S.)",
  ["dynamis (granada, spain)"] = "Dynamis",
  ["dynamische psychiatrie. dynamic psychiatry"] = "Dyn Psychiatr",
  ["dyslexia (chichester, england)"] = "Dyslexia",
  ["dysmorphology and clinical genetics : official publication of the center for birth defects information services, inc"] = "Dysmorphol Clin Genet",
  ["dysphrenia : a peer-reviewed biannual academic journal of psychiatry"] = "Dysphrenia",
  ["dédalo. revista de arte e arqueologia"] = "Dedalo",
  ["düsseldorfer arbeiten zur geschichte der medizin. beiheft"] = "Dusseld Arb Gesch Med Beih",
  ["dějiny věd a techniky"] = "DVT—Dějiny Věd Tech.",
  ["e (norwalk, conn.)"] = "E (Norwalk)",
  ["e law : murdoch university electronic journal of law"] = "E Law",
  ["e n e"] = "E N E",
  ["e&g quaternary science journal"] = "E&G Quat. Sci. J.",
  ["e-book of abstracts : v annual meeting htai 2008, july 6-9, 2008, montréal, canada. health technology assessment international. meeting (5th : 2008 : montréal, canada)"] = "Ebook Abstr Health Technol Assess Int Meet 5th 2008 Montr Can",
  ["e-forschungsberichte des deutschen archäologischen instituts"] = "eDAI-F",
  ["e-health telecommunication systems and networks"] = "Ehealth Telecommun Syst Netw",
  ["e-itv: educational and industrial television"] = "EITV",
  ["e-jahresberichte des deutschen archäologischen instituts"] = "eDAI-J",
  ["e-journal of applied psychology : clinical and social issues"] = "E J Appl Psychol",
  ["e-journal of chemistry"] = "E-J. Chem.",
  ["e-journal of reservoir engineering"] = "E-J. Reservoir Eng.",
  ["e-journal of surface science and nanotechnology"] = "e-J. Surf. Sci. Nanotechnol.",
  ["e-prime advances in electrical engineering, electronics and energy"] = "e-Prime Adv. Electr. Eng. Electron. Energy",
  ["e-spen journal"] = "ESPEN J",
  ["e-spen, the european e-journal of clinical nutrition and metabolism"] = "E Spen Eur E J Clin Nutr Metab",
  ["e. kalinka (hrsg.), tituli asiae minoris (wien 1901–­1941)"] = "TAM",
  ["eai endorsed transactions on pervasive health and technology"] = "EAI Endorsed Trans Pervasive Health Technol",
  ["eai/springer innovations in communication and computing"] = "EAI/Springer Innov. Commun. Comput.",
  ["ear and hearing"] = "Ear Hear",
  ["ear, nose, & throat journal"] = "Ear Nose Throat J",
  ["ear, nose, and throat journal"] = "Ear. Nose. Throat J.",
  ["early american literature"] = "Early Am Lit",
  ["early child development and care"] = "Early Child Dev Care",
  ["early childhood education journal"] = "Early Child Educ J",
  ["early childhood research quarterly"] = "Early Child Res Q",
  ["early childhood services (san diego, calif.)"] = "Early Child Serv (San Diego)",
  ["early development & parenting"] = "Early Dev Parent",
  ["early education and development"] = "Early Educ Dev",
  ["early human development"] = "Early Hum Dev",
  ["early intervention in psychiatry"] = "Early Interv Psychiatry",
  ["early medieval europe"] = "Early Medieval Eur.",
  ["early pregnancy"] = "Early Pregnancy",
  ["early pregnancy (online)"] = "Early Pregnancy",
  ["early pregnancy : biology and medicine : the official journal of the society for the investigation of early pregnancy"] = "Early Pregnancy",
  ["early science and medicine"] = "Early Sci Med",
  ["early science and medicine. a journal for the study of science, technology and medicine in the pre-modern period"] = "Early Sci. Med.",
  ["early sources in aphasia and related disorders"] = "Early Sources Aphasia Relat Disord",
  ["early years : an international journal of research and development"] = "Early Years (Stoke-on-Trent)",
  ["earth (waukesha, wis.)"] = "Earth (Waukesha)",
  ["earth and environmental science transactions"] = "Earth Environ. Sci. Trans.",
  ["earth and environmental science transactions of the royal society of edinburgh"] = "Earth Environ Sci Trans R Soc Edinb",
  ["earth and mineral sciences"] = "Earth Miner Sci",
  ["earth and planetary physics"] = "Earth Planet. Phys.",
  ["earth and planetary science letters"] = "Earth Planet. Sci. Lett.",
  ["earth and space science"] = "Earth Space Sci.",
  ["earth and space science (hoboken, n.j.)"] = "Earth Space Sci",
  ["earth in space"] = "Earth Space",
  ["earth interactions"] = "Earth Interact",
  ["earth moon and planets"] = "Earth Moon Planets",
  ["earth negotiations bulletin"] = "Earth Negot Bull",
  ["earth planets space [formerly journal of geomagnetism and geoelectricity]"] = "Earth Planets Space",
  ["earth science frontiers"] = "Earth Sci. Front.",
  ["earth science informatics"] = "Earth Sci. Inf.",
  ["earth science, systems and society"] = "Earth Sci. Syst. Soc.",
  ["earth sciences history : journal of the history of the earth sciences society"] = "Earth Sci Hist",
  ["earth sciences research journal"] = "Earth Sci. Res. J.",
  ["earth space review"] = "Earth Space Rev",
  ["earth surface dynamics"] = "Earth Surf. Dyn.",
  ["earth surface dynamics discussions"] = "Earth Surf. Dyn. Discuss.",
  ["earth surface processes"] = "Earth Surf. Processes",
  ["earth surface processes and landforms"] = "Earth Surf Process Landf",
  ["earth system dynamics"] = "Earth Syst. Dyn.",
  ["earth system dynamics : esd"] = "Earth Syst Dyn",
  ["earth system dynamics discussions"] = "Earth Syst. Dyn. Discuss.",
  ["earth system governance"] = "Earth Syst. Governance",
  ["earth system science data"] = "Earth Syst. Sci. Data",
  ["earth system science data discussions"] = "Earth Syst. Sci. Data Discuss.",
  ["earth systems and environment"] = "Earth Syst. Environ.",
  ["earth's future"] = "Earths Future",
  ["earth, moon, and planets"] = "Earth Moon Planets",
  ["earth, planets and space"] = "Earth Planets Space",
  ["earth, planets, and space : eps"] = "Earth Planets Space",
  ["earth-science reviews"] = "Earth Sci. Rev.",
  ["earthquake engineering & structural dynamics"] = "Earthq Eng Struct Dyn",
  ["earthquake engineering and engineering vibration"] = "Earthquake Eng. Eng. Vibr.",
  ["earthquake engineering and resilience"] = "Earthquake Eng. Resilience",
  ["earthquake engineering and structural dynamics"] = "Earthquake Eng. Struct. Dyn.",
  ["earthquake research advances"] = "Earthquake Res. Adv.",
  ["earthquake science"] = "Earthquake Sci.",
  ["earthquake spectra"] = "Earthquake Spectra",
  ["earthquake spectra : the professional journal of the earthquake engineering research institute"] = "Earthq Spectra",
  ["earthquakes and structures"] = "Earthquakes Struct.",
  ["earth’s future"] = "Earth’s Future",
  ["east african (online)"] = "East Afr",
  ["east african agricultural and forestry journal"] = "East Afri Agric For J",
  ["east african journal of applied health monitoring and evaluation"] = "East Afr J Appl Health Monitor Eval",
  ["east african journal of public health"] = "East Afr J Public Health",
  ["east african journal of rural development"] = "East African J. Rural Devel.",
  ["east african journal of science, technology and innovation"] = "East Afr. J. Sci. Technol. Innovation",
  ["east african medical journal"] = "East Afr Med J",
  ["east african standard (nairobi, kenya : daily : 1995)"] = "East Afr Stand",
  ["east and central african journal of surgery"] = "East Cent Afr J Surg",
  ["east and west"] = "E&W",
  ["east asia (piscataway, n.j.)"] = "East Asia (Piscataway)",
  ["east asian archives of psychiatry : official journal of the hong kong college of psychiatrists = dong ya jing shen ke xue zhi : xianggang jing shen ke yi xue yuan qi kan"] = "East Asian Arch Psychiatry",
  ["east asian history"] = "East Asian Hist (Canberra)",
  ["east asian journal on applied mathematics"] = "East Asian J. Appl. Math.",
  ["east asian science, technology and society"] = "East Asian Sci Technol Soc",
  ["east asian science, technology and society: an international journal"] = "East Asian Sci. Technol. Soc.: Int. J.",
  ["east asian science, technology, and medicine"] = "East Asian Sci Technol Med",
  ["east central europe. l'europe du centre-est"] = "East Cent Eur",
  ["east central europe. l’europe du centre-est"] = "East Cent Eur",
  ["east china normal university scientific reports"] = "East China Norm. Univ. Sci. Rep.",
  ["east european jewish affairs"] = "East Eur Jew Aff",
  ["east european quarterly"] = "East Eur Q",
  ["east journal on approximations"] = "East J. Approx.",
  ["east-west journal of economics and business"] = "East-West J. Econ. Bus.",
  ["east-west journal of mathematics"] = "East-West J. Math.",
  ["east-west journal of numerical mathematics"] = "East-West J. Numer. Math.",
  ["east-west perspectives : journal of the east-west center"] = "East West Perspect",
  ["east-west series in economics, business, and the environment"] = "East-West Ser. Econ.",
  ["eastern africa economic review"] = "East Afr Econ Rev",
  ["eastern africa journal of rural development"] = "East Afr J Rural Dev",
  ["eastern africa social science research review"] = "East Afr Soc Sci Res Rev",
  ["eastern and southern africa geographical journal"] = "East South Afr Geogr J",
  ["eastern biologist"] = "East Biol",
  ["eastern economic journal"] = "East Econ J",
  ["eastern european economics"] = "East Europ Econ",
  ["eastern european politics and societies : eeps"] = "East Eur Polit Soc",
  ["eastern horizon"] = "East Horiz",
  ["eastern journal of medicine : ejm"] = "East J Med",
  ["eastern mediterranean health journal = la revue de santé de la méditerranée orientale = al-majallah al-ṣiḥḥīyah li-sharq al-mutawassiṭ"] = "East Mediterr Health J",
  ["eating and weight disorders"] = "Eat. Weight Disord.",
  ["eating and weight disorders : ewd"] = "Eat Weight Disord",
  ["eating behaviors"] = "Eat Behav",
  ["eating disorders"] = "Eat Disord",
  ["ebp briefs"] = "EBP Briefs",
  ["ebri issue brief"] = "EBRI Issue Brief",
  ["ebri issue brief / employee benefit research institute"] = "EBRI Issue Brief",
  ["ec anaesthesia"] = "EC anaesth",
  ["ec clinical and experimental anatomy"] = "EC Clin Exp Anat",
  ["ec clinical and medical case reports"] = "EC Clin Med Case Rep",
  ["ec dental science"] = "EC Dent Sci",
  ["ec gastroenterology and digestive system"] = "EC Gastroenterol Dig Syst",
  ["ec gynaecology"] = "EC Gynaecol",
  ["ec microbiology"] = "EC Microbiol",
  ["ec nutrition"] = "EC Nutr",
  ["ec ophthalmology"] = "EC Ophthalmol",
  ["ec orthopaedics"] = "EC Orthop",
  ["ec pharmacology and toxicology"] = "EC Pharmacol Toxicol",
  ["ec psychology and psychiatry"] = "EC Psychol Psychiatr",
  ["ec pulmonology and respiratory medicine"] = "EC Pulmonol Respir Med",
  ["eca : estudios centro americanos"] = "Estud Cent Am",
  ["echo research and practice"] = "Echo Res Pract",
  ["echocardiography (mount kisco, n.y.)"] = "Echocardiography",
  ["echos du monde classique. classical views"] = "EchosCl",
  ["ecletica quimica"] = "Ecletica Quim.",
  ["eclogae geologicae helvetiae"] = "Eclogae Geol. Helv.",
  ["eco-efficiency in industry and science"] = "Eco-Effic. Ind. Sci.",
  ["eco-environment & health"] = "Eco-Environ. Health",
  ["ecohydrology & hydrobiology"] = "Ecohydrol. Hydrobiol.",
  ["ecohydrology : ecosystems, land and water process interactions, ecohydrogeomorphology"] = "Ecohydrology",
  ["ecole d’ete de probabilites de saint-flour"] = "Ec. Ete Probab. St.-Flour",
  ["ecologia mediterranea"] = "Ecol. Mediterr.",
  ["ecologia mediterranea : revue d'écologie terrestre et limnique"] = "Ecol Mediterr",
  ["ecologic science"] = "Ecol Sci",
  ["ecologica montenegrina"] = "Ecol Montenegrina",
  ["ecological applications"] = "Ecol. Appl.",
  ["ecological applications : a publication of the ecological society of america"] = "Ecol Appl",
  ["ecological chemistry and engineering"] = "Ecol. Chem. Eng.",
  ["ecological chemistry and engineering a"] = "Ecol. Chem. Eng. A",
  ["ecological chemistry and engineering s"] = "Ecol. Chem. Eng. S",
  ["ecological complexity"] = "Ecol. Complexity",
  ["ecological conservation"] = "Ecol. Conserv.",
  ["ecological economics"] = "Ecol. Econ.",
  ["ecological economics : the journal of the international society for ecological economics"] = "Ecol Econ",
  ["ecological engineering"] = "Ecol Eng",
  ["ecological engineering: x"] = "Ecol. Eng.: X",
  ["ecological entomology"] = "Ecol Entomol",
  ["ecological genetics and genomics"] = "Ecol Genet Genom",
  ["ecological indicators"] = "Ecol. Indic.",
  ["ecological informatics"] = "Ecol Inform",
  ["ecological management and restoration"] = "Ecol. Manage. Restor.",
  ["ecological modelling"] = "Ecol Modell",
  ["ecological monographs"] = "Ecol Monogr",
  ["ecological processes"] = "Ecol Process",
  ["ecological psychology"] = "Ecol. Psychol.",
  ["ecological psychology : a publication of the international society for ecological psychology"] = "Ecol Psychol",
  ["ecological research"] = "Ecol. Res.",
  ["ecological research monographs"] = "Ecol. Res. Monogr.",
  ["ecological solutions and evidence"] = "Ecol. Solutions Evidence",
  ["ecological studies"] = "Ecol. Stud.",
  ["ecology and evolution"] = "Ecol Evol",
  ["ecology and society"] = "Ecol. Soc.",
  ["ecology and society : a journal of integrative science for resilience and sustainability"] = "Ecol Soc",
  ["ecology law quarterly"] = "Ecol Law Q",
  ["ecology letters"] = "Ecol Lett",
  ["ecology of disease"] = "Ecol Dis",
  ["ecology of food and nutrition"] = "Ecol Food Nutr",
  ["ecology of freshwater fish"] = "Ecol Freshw Fish",
  ["econ journal watch"] = "Econ J Watch",
  ["econometric and tinbergen institutes lectures"] = "Econom. Tinbergen Inst. Lect.",
  ["econometric reviews"] = "Econometric Rev.",
  ["econometric society monographs"] = "Econom. Soc. Monogr.",
  ["econometric theory"] = "Econometric Theory",
  ["econometrica : journal of the econometric society"] = "Econometrica",
  ["econometrica. journal of the econometric society"] = "Econometrica",
  ["econometrics and statistics"] = "Econom Stat",
  ["econometrics in the information age: theory and practice of measurement"] = "Econom. Inf. Age Theory Pract. Meas.",
  ["econometrics journal"] = "Econometrics J.",
  ["economia & diritto agroalimentare"] = "Econ Dirit Agroaliment",
  ["economia & lavoro"] = "Econ Lav",
  ["economia (pontifical catholic university of peru)"] = "Economia (Peru)",
  ["economia (pontificia universidad catolica del peru. departamento de economia)"] = "Economia",
  ["economia (portuguese catholic university)"] = "Economia (Portugal)",
  ["economia aplicada"] = "Econ. Aplicada",
  ["economia delle fonti di energia"] = "Econ Fonti Energ Ambiente",
  ["economia delle scelte pubbliche"] = "Econ. Scelte Pubbliche",
  ["economia e credito"] = "Econ Credito",
  ["economia e lavoro"] = "Econ. Lavoro",
  ["economia e socialismo"] = "Econ Social",
  ["economia e sociologia (evora, portugal)"] = "Econ Sociol",
  ["economia e storia"] = "Econ Stor",
  ["economia industrial"] = "Econ. Ind.",
  ["economia internazionale"] = "Econ. Int.",
  ["economia mexicana, nueva epoca"] = "Econ. Mexicana",
  ["economia politica"] = "Econ. Politica",
  ["economia pubblica : mensile di studi e d'informazione del ciriec"] = "Econ Pubblica",
  ["economia pubblica : mensile di studi e d’informazione del ciriec"] = "Econ Pubblica",
  ["economia y desarrollo"] = "Econ Desarro",
  ["economic affairs"] = "Econ. Aff.",
  ["economic analysis"] = "Econ. Anal.",
  ["economic analysis and policy"] = "Econ Anal Policy",
  ["economic analysis and policy, new series"] = "Econ. Anal. Pol.",
  ["economic analysis and workers’ management"] = "Econ. Anal. Workers’ Manage.",
  ["economic and business bulletin"] = "Econ. Bus. Bull.",
  ["economic and business review"] = "Econ. Bus. Rev.",
  ["economic and financial modelling"] = "Econ. Finan. Modelling",
  ["economic and financial review"] = "Econ. Finan. Rev.",
  ["economic and industrial democracy"] = "Econ. Ind. Democracy",
  ["economic and labour relations review"] = "Econ. Lab. Relat. Rev.",
  ["economic and political studies"] = "Econ Polit Stud",
  ["economic and political weekly"] = "Econ Polit Wkly",
  ["economic and social affairs"] = "Econ Soc Aff",
  ["economic and social review"] = "Econ. Soc. Rev.",
  ["economic and social studies. national institute of economic and social research"] = "Econ. Soc. Stud. Natl. Inst. Econ. Soc. Res.",
  ["economic anthropology (hoboken, n.j.)"] = "Econ Anthropol",
  ["economic botany"] = "Econ Bot",
  ["economic bulletin - national bank of egypt"] = "Econ Bull Natl Bank Egypt.",
  ["economic bulletin for asia and the pacific"] = "Econ Bull Asia Pac",
  ["economic bulletin for europe"] = "Econ Bull Eur",
  ["economic bulletin of ghana, second series"] = "Econ. Bull. Ghana",
  ["economic computation and economic cybernetics studies and research"] = "Econ Comput Econ Cybern Stud Res",
  ["economic computation and economic cybernetics studies and research / academy of economic studies"] = "Econ Comput Econ Cybern Stud Res",
  ["economic design"] = "Econ. Design",
  ["economic development and cultural change"] = "Econ Dev Cult Change",
  ["economic development quarterly"] = "Econ. Devel. Quart.",
  ["economic eye"] = "Econ Eye",
  ["economic forum"] = "Econ Forum",
  ["economic geography"] = "Econ Geogr",
  ["economic geology"] = "Econ. Geol.",
  ["economic geology and the bulletin of the society of economic geologists"] = "Econ Geol",
  ["economic history review"] = "Econ. Hist. Rev.",
  ["economic history review, second series"] = "Econ. Hist. Rev.",
  ["economic inquiry"] = "Econ Inq",
  ["economic issues"] = "Econ. Issues",
  ["economic issues, problems and perspectives"] = "Econ. Issues Probl. Perspect.",
  ["economic journal"] = "Econ. J.",
  ["economic journal (london, england)"] = "Econ J (London)",
  ["economic modelling"] = "Econ Model",
  ["economic notes"] = "Econ. Notes",
  ["economic outlook usa"] = "Econ Outlook USA",
  ["economic perspectives"] = "Econ. Perspect.",
  ["economic policy"] = "Econ Policy",
  ["economic policy review"] = "Econ. Pol. Rev.",
  ["economic policy: a european forum"] = "Econ. Pol.: A Europ. Forum",
  ["economic record"] = "Econ. Rec.",
  ["economic research"] = "Econ. Res.",
  ["economic research journal"] = "Econ Res J",
  ["economic review (federal reserve bank of atlanta)"] = "Econ Rev",
  ["economic review (federal reserve bank of dallas)"] = "Econ Rev (Dallas)",
  ["economic review (keizai kenkyu)"] = "Econ. Rev. (Keizai Kenkyu)",
  ["economic studies (calcutta, india)"] = "Econ Stud",
  ["economic studies in inequality, social exclusion and well-being"] = "Econ. Stud. Inequal. Soc. Exclusion Well-Being",
  ["economic studies quarterly"] = "Econ. Stud. Quart.",
  ["economic systems"] = "Econ. Systems",
  ["economic systems research"] = "Econ. Systems Res.",
  ["economic systems research : journal of the international input-output association"] = "Econ Syst Res",
  ["economic theory"] = "Econom. Theory",
  ["economic theory bulletin"] = "Econ. Theory Bull.",
  ["economic theory, econometrics, and mathematical economics"] = "Econom. Theory Econometrics Math. Econom.",
  ["economic trends (new delhi, india)"] = "Econ Trends",
  ["economica, n.s."] = "Economica",
  ["economica. new series"] = "Economica (N.S.)",
  ["economics affairs (calcutta)"] = "Econ Aff",
  ["economics and business education"] = "Econ. Bus. Educ.",
  ["economics and human biology"] = "Econ Hum Biol",
  ["economics and philosophy"] = "Econ Philos",
  ["economics and politics"] = "Econ. Politics",
  ["economics letters"] = "Econ Lett",
  ["economics of disasters and climate change"] = "Econ Disaster Clim Chang",
  ["economics of education review"] = "Econ Educ Rev",
  ["economics of governance"] = "Econ. Governance",
  ["economics of innovation and new technology"] = "Econ. Innovation New Tech.",
  ["economics of planning"] = "Econ Plann",
  ["economics of science, technology and innovation"] = "Econ. Sci. Technol. Innov.",
  ["economics of transition"] = "Econ. Transition",
  ["economics: a biannual collection of recent german studies"] = "Econ.: Biann. Collec. Recent Ger. Stud.",
  ["economics: the journal of the economics association"] = "Econ.: J. Econ. Assoc.",
  ["economie & prévision. statistiques & études financières. série économique"] = "Econ Previs",
  ["economie & statistique"] = "Econ Stat",
  ["economie appliquée"] = "Econ. Appl.",
  ["economie et humanisme"] = "Econ Humanisme",
  ["economie et medecine animales"] = "Econ Med Anim",
  ["economie et médecine animales"] = "Econ Med Anim",
  ["economie et prévision"] = "Econ. Prévision",
  ["economie internationale"] = "Econ. Int.",
  ["economie meridionale"] = "Econ Merid",
  ["economie méridionale"] = "Econ Merid",
  ["economie prospective internationale"] = "Econ Prospect Int",
  ["economie prospective internationale / centre d’etudes prospectives et d’informations internationales"] = "Econ Prospect Int",
  ["economie rurale"] = "Econ. Rurale",
  ["economies et societes"] = "Econ Soc",
  ["economies et sociétés"] = "Econ. Soc.",
  ["economisch en sociaal tijdschrift"] = "Econ. Soc. Tijdschr.",
  ["economisch- en sociaal-historisch jaarboek"] = "Econ Soc Hist Jaarb",
  ["economisch- en sociaal-historisch jaarboek / uitgegeven door de vereniging het nederlandsch economisch-historisch archief"] = "Econ Soc Hist Jaarb",
  ["economisch-statistische berichten"] = "Econ Stat Ber",
  ["economist (london, england : 1843)"] = "Economist",
  ["economía (pontificia universidad católica del perú. departamento de economía)"] = "Economia",
  ["economía colombiana : revista de la contraloría general de la república"] = "Econ Colomb",
  ["economía y desarrollo"] = "Econ Desarro",
  ["economía y sociología del trabajo"] = "Econ Sociol Trab",
  ["economy and history"] = "Econ Hist (Sweden)",
  ["economy and society"] = "Econ Soc",
  ["economía chilena"] = "Econ. Chilena",
  ["económica (national university of la plata)"] = "Económica (La Plata)",
  ["ecosal plus"] = "EcoSal Plus",
  ["ecosistemas : revista científica y técnica de ecología y medio ambiente"] = "Ecosistemas",
  ["ecosphere (washington, d.c)"] = "Ecosphere",
  ["ecosystem health and sustainability"] = "Ecosyst. Health Sustainability",
  ["ecosystem services"] = "Ecosyst Serv",
  ["ecosystems (new york, n.y.)"] = "Ecosystems",
  ["ecosystems and people"] = "Ecosyst. People",
  ["ecosystems and people (abingdon, england)"] = "Ecosyst People (Abingdon)",
  ["ecotoxicology (london, england)"] = "Ecotoxicology",
  ["ecotoxicology and environment safety"] = "Ecotoxicol. Environ. Saf.",
  ["ecotoxicology and environmental safety"] = "Ecotoxicol Environ Saf",
  ["ecrits de paris: revue des questions actuelles"] = "Ecrits Paris",
  ["ecs advances"] = "ECS Adv.",
  ["ecs electrochemistry letters"] = "ECS Electrochem. Lett.",
  ["ecs journal of solid state science and technology"] = "ECS J. Solid State Sci. Technol.",
  ["ecs journal of solid state science and technology : jss"] = "ECS J Solid State Sci Technol",
  ["ecs sensors plus"] = "ECS Sens. Plus",
  ["ecs solid state letters"] = "ECS Solid State Lett.",
  ["ecs transactions"] = "ECS Trans.",
  ["ecuador debate"] = "Ecuad Debate",
  ["ed management"] = "ED Manag.",
  ["ed management : the monthly update on emergency department management"] = "ED Manag",
  ["ed-media. ed-media (conference)"] = "Ed Media",
  ["edinburgh advanced textbooks in linguistics"] = "Edinb. Adv. Textb. Linguist.",
  ["edinburgh dental hospital gazette"] = "Edinb Dent Hosp Gaz",
  ["edinburgh journal of botany"] = "Edinb J Bot",
  ["edinburgh medical and surgical journal"] = "Edinb Med Surg J",
  ["edinburgh medical journal"] = "Edinb Med J",
  ["edition am gutenbergplatz leipzig (eagle)"] = "Ed. Gutenbergplatz Leipz. (EAGLE)",
  ["edition angewandte"] = "Ed. Angew.",
  ["edition open sources"] = "Ed. Open Sources",
  ["editorial research reports"] = "Editor Res Rep",
  ["edizione nazionale dell’opera matematica di francesco maurolico"] = "Ed. Naz. Opera Mat. Francesco Maurolico",
  ["edorium journal of disability and rehabilitation"] = "Edorium J Disabil Rehabil",
  ["edorium journal of surgery"] = "Edorium J Surg",
  ["eds magazine"] = "EDS Mag.",
  ["eds magazine : official journal of the e.e.c. dental students committee"] = "EDS Mag",
  ["edtna/erca journal"] = "EDNTNA ERCA J.",
  ["edtna/erca journal (english ed.)"] = "EDTNA ERCA J",
  ["educacion dental"] = "Educ. Dent. (Ica)",
  ["educacion medica y salud"] = "Educ. Med. Salud",
  ["educacion quimica"] = "Educ. Quim.",
  ["educación dental"] = "Educ Dent (Ica)",
  ["educación médica y salud"] = "Educ Med Salud",
  ["educação & sociedade"] = "Educ Soc",
  ["educação e realidade"] = "Educ Real",
  ["educafrica (english version)"] = "Educafrica",
  ["educafrica (version française)"] = "Educafrica (Ed Fr)",
  ["education & treatment of children"] = "Educ Treat Children",
  ["education 3-13"] = "Educ 3 13",
  ["education and information technologies"] = "Educ Inf Technol (Dordr)",
  ["education and training in autism and developmental disabilities"] = "Educ Train Autism Dev Disabil",
  ["education and training in developmental disabilities"] = "Educ Train Dev Disabil",
  ["education and training of the mentally retarded"] = "Educ Train Ment Retard",
  ["education and urban society"] = "Educ Urban Soc",
  ["education economics"] = "Educ Econ",
  ["education et formation"] = "Educ. Form.",
  ["education finance and policy"] = "Educ Finance Policy",
  ["education for chemical engineers"] = "Educ. Chem. Eng.",
  ["education for health (abingdon, england)"] = "Educ Health (Abingdon)",
  ["education for information"] = "Educ. Inf.",
  ["education for primary care : an official publication of the association of course organisers, national association of gp tutors, world organisation of family doctors"] = "Educ Prim Care",
  ["education in chemistry"] = "Educ. Chem.",
  ["education in science : the bulletin of the association for science education"] = "Educ Sci",
  ["education in the north"] = "Educ North",
  ["education libraries bulletin"] = "Libr Bull Univ Lond",
  ["education policy analysis archives"] = "Educ Policy Anal Arch",
  ["education research international"] = "Educ Res Int",
  ["education sciences"] = "Educ Sci (Basel)",
  ["education week"] = "Educ Week",
  ["education, citizenship and social justice"] = "Educ Citizsh Soc Justice",
  ["educational & industrial television"] = "Educ Ind Telev",
  ["educational action research"] = "Educ Action Res",
  ["educational administration quarterly"] = "Educ. Adm. Q.",
  ["educational administration quarterly : eaq"] = "Educ Adm Q",
  ["educational and industrial television"] = "Educ. Ind. Telev.",
  ["educational and psychological measurement"] = "Educ Psychol Meas",
  ["educational assessment, evaluation and accountability"] = "Educ Assess Eval Account",
  ["educational broadcasting international"] = "Educ Broadcast Int",
  ["educational directions for dental auxiliaries"] = "Educ Dir Dent Aux",
  ["educational directions in dental hygiene"] = "Educ Dir Dent Hyg",
  ["educational documentation and information : bulletin of the international bureau of education"] = "Educ Doc Inf",
  ["educational evaluation and policy analysis"] = "Educ Eval Policy Anal",
  ["educational focus"] = "Educ Focus",
  ["educational gerontology"] = "Educ Gerontol",
  ["educational horizons"] = "Educ Horiz",
  ["educational leadership : journal of the department of supervision and curriculum development, n.e.a"] = "Educ Leadersh",
  ["educational media international"] = "EMI Educ Media Int",
  ["educational neuroscience"] = "Educ. Neurosci.",
  ["educational policy (los altos, calif.)"] = "Educ Policy (Los Altos Calif)",
  ["educational psychologist"] = "Educ Psychol",
  ["educational psychology"] = "Educ Psychol (Lond)",
  ["educational psychology in practice"] = "Educ Psychol Pract",
  ["educational psychology review"] = "Educ Psychol Rev",
  ["educational research and evaluation : an international journal on theory and practice"] = "Educ Res Eval",
  ["educational research review"] = "Educ Res Rev",
  ["educational research; a review for teachers and all concerned with progress in education"] = "Educ Res (Windsor)",
  ["educational researcher (washington, d.c. : 1972)"] = "Educ Res",
  ["educational review"] = "Educ Rev (Birm)",
  ["educational studies"] = "Educ Stud",
  ["educational studies in mathematics"] = "Educ. Stud. Math.",
  ["educational technology research and development : etr & d"] = "Educ Technol Res Dev",
  ["educational theory"] = "Educ Theory",
  ["educator's update"] = "Educ Update",
  ["educator’s update"] = "Educ Update",
  ["eecs technical report series"] = "EECS Tech Rep Ser",
  ["eeg-emg zeitschrift fur elektroenzephalographie, elektromyographie und verwandte gebiete"] = "EEG EMG Z Elektroenzephalogr Elektromyogr Verwandte Geb",
  ["eeg-emg zeitschrift für elektroenzephalographie, elektromyographie und verwandte gebiete"] = "EEG EMG Z Elektroenzephalogr Elektromyogr Verwandte Geb",
  ["eeg-emg: zeitschrift fur elektroenzephalographie, elektromyographie und verwandte gebiete"] = "EEG EMG Z. Elektroenzephalogr. Elektromyogr. Verwandte Geb.",
  ["ees catalysis"] = "EES Catal.",
  ["eesti nsv teaduste akadeemia toimetised. bioloogia = izvestiia akademii nauk estonskoĭ ssr. biologiia"] = "Eesti Nsv Tead Akad TOIM Biol",
  ["effective clinical practice"] = "Eff. Clin. Pract.",
  ["effective clinical practice : ecp"] = "Eff Clin Pract",
  ["effective health care"] = "Eff Health Care",
  ["effluent and water treatment journal"] = "Effluent Water Treat J",
  ["efort open reviews"] = "EFORT Open Rev",
  ["efsa journal"] = "EFSA J.",
  ["efsa journal. european food safety authority"] = "EFSA J",
  ["efsa supporting publications"] = "EFSA Supporting Publ.",
  ["egems (washington, dc)"] = "EGEMS (Wash DC)",
  ["egitto e vicino oriente"] = "EVO",
  ["egypt today (international business associates (cairo, egypt))"] = "Egypt Today",
  ["egypte/monde arabe"] = "Egypte Monde Arabe",
  ["egyptian academic journal of biological sciences. b, zoology"] = "Egypt Acad J Biol Sci B Zool",
  ["egyptian academic journal of biological sciences. c, physiology & molecular biology"] = "Egypt Acad J Biol Sci C Physiol Mol Biol",
  ["egyptian archaeology. the bulletin of the egypt exploration society"] = "EgA",
  ["egyptian dental journal"] = "Egypt Dent J",
  ["egyptian informatics journal"] = "Egypt. Inf. J.",
  ["egyptian journal of agronomy"] = "Egypt. J. Agron.",
  ["egyptian journal of anaesthesia"] = "Egypt J Anaesth",
  ["egyptian journal of analyical chemistry"] = "Egypt. J. Anal. Chem.",
  ["egyptian journal of analytical chemistry"] = "Egypt. J. Anal. Chem.",
  ["egyptian journal of aquatic biology and fisheries"] = "Egypt J Aquat Biol Fish",
  ["egyptian journal of aquatic research"] = "Egypt. J. Aquat. Res.",
  ["egyptian journal of archaeological and restoration studies"] = "Egypt. J. Archaeol. Restor. Stud.",
  ["egyptian journal of basic and applied sciences"] = "Egypt. J. Basic Appl. Sci.",
  ["egyptian journal of basic and clinical pharmacology"] = "Egypt J Basic Clin Pharmacol",
  ["egyptian journal of bilharziasis"] = "Egypt J Bilharz",
  ["egyptian journal of biochemistry and molecular biology"] = "Egypt. J. Biochem. Mol. Biol.",
  ["egyptian journal of biological pest control"] = "Egypt J Biol Pest Control",
  ["egyptian journal of biotechnology"] = "Egypt. J. Biotechnol.",
  ["egyptian journal of botany"] = "Egypt. J. Bot.",
  ["egyptian journal of chemistry"] = "Egypt J Chem",
  ["egyptian journal of experimental biology. zoology"] = "Egypt J Exp Biol Zool",
  ["egyptian journal of forensic sciences"] = "Egypt J Forensic Sci",
  ["egyptian journal of genetics and cytology"] = "Egypt J Genet Cytol",
  ["egyptian journal of geology"] = "Egypt. J. Geol.",
  ["egyptian journal of hospital medicine"] = "Egypt. J. Hosp. Med.",
  ["egyptian journal of medical human genetics"] = "Egypt. J. Med. Hum. Genet.",
  ["egyptian journal of medical microbiology"] = "Egypt J Med Microbiol",
  ["egyptian journal of microbiology"] = "Egypt J Microbiol",
  ["egyptian journal of occupational medicine"] = "Egypt J Occup Med",
  ["egyptian journal of petroleum"] = "Egypt. J. Pet.",
  ["egyptian journal of phytopathology"] = "Egypt J Phytopathol",
  ["egyptian journal of psychiatry"] = "Egypt. J. Psychiatry",
  ["egyptian journal of radiology and nuclear medicine"] = "Egypt. J. Radiol. Nucl. Med.",
  ["egyptian journal of remote sensing and space science"] = "Egypt. J. Remote Sens. Space. Sci.",
  ["egyptian journal of veterinary science"] = "Egypt J Vet Sci",
  ["egyptian journal of virology"] = "Egypt J Virol",
  ["egyptian medical journal"] = "Egypt Med J",
  ["eğitimde ve psikolojide ölçme ve değerlendirme dergisi"] = "Egit Psikol Olcme Deger Derg",
  ["ehealth international"] = "eHealth Int",
  ["ehp toxicogenomics : journal of the national institute of environmental health sciences"] = "EHP Toxicogenomics",
  ["ehs law bulletin series, japan. japan"] = "EHS Law Bull Ser Japan",
  ["eib papers"] = "EIB Pap.",
  ["eichsfelder heimathefte"] = "Eichsfelder Heimath",
  ["eighteenth century (lubbock, tex.)"] = "Eighteenth Century (Lubbock)",
  ["eighteenth century life"] = "Eighteenth Century Life",
  ["eighteenth century: a current bibliography"] = "Eighteenth Century Curr Bibliogr",
  ["eighteenth-century ireland"] = "Eighteenth Century Irel",
  ["eighteenth-century studies"] = "Eighteenth Century Stud",
  ["eikasmos: quaderni bolognesi di filologia classica"] = "Eikasmos",
  ["einblicke in die wissenschaft"] = "Einblicke Wiss.",
  ["einstein (sao paulo)"] = "einstein (Sao Paulo)",
  ["einstein (são paulo, brazil)"] = "Einstein (Sao Paulo)",
  ["einstein journal of biology and medicine"] = "Einstein J. Biol. Med.",
  ["einstein quarterly journal of biology and medicine"] = "Einstein Q. J. Biol. Med.",
  ["einstein studies"] = "Einstein Stud.",
  ["eire-ireland; a journal of irish studies"] = "Eire Irel",
  ["eirene. studia graeca et latina"] = "Eirene",
  ["eirene: studia graeca et latina"] = "Eirene",
  ["eisei shikenjo hokoku (bulletin of national institute of hygienic sciences)"] = "Eisei Shikenjo Hokoku",
  ["eisei shikenjo hokoku. bulletin of national institute of hygienic sciences"] = "Eisei Shikenjo Hokoku",
  ["eisei shikenjo hōkoku. bulletin of national institute of hygienic sciences"] = "Eisei Shikenjo Hokoku",
  ["eiszeitalter und gegenwart"] = "Eiszeitalter Ggw.",
  ["ejnmmi physics"] = "EJNMMI Phys",
  ["ejnmmi radiopharmacy and chemistry"] = "EJNMMI Radiopharm Chem",
  ["ejnmmi research"] = "EJNMMI Res",
  ["ejournal of biological sciences"] = "Ejournal Biol Sci",
  ["ejves short reports"] = "EJVES Short Rep",
  ["ejves vascular forum"] = "EJVES Vasc Forum",
  ["ekistics; reviews on the problems and science of human settlements"] = "Ekistics",
  ["eklem hastalıkları ve cerrahisi = joint diseases & related surgery"] = "Eklem Hastalik Cerrahisi",
  ["ekologia (cssr)"] = "Ekol CSSR",
  ["ekologija (vilnius, lithuania)"] = "Ekologija (Liet Moksl Akad (Spausd))",
  ["ekonomi dan keuangan indonesia"] = "Ekon Keuangan Indones",
  ["ekonomicko-matematicky obzor"] = "Ekon.-Mat. Obzor",
  ["ekonomicky casopis"] = "Ekon Cas",
  ["ekonomický c̆asopis"] = "Ekon Cas",
  ["ekonomika i upravlenie"] = "Ekonom. Upravl.",
  ["ekonomika radvanskoyi ukhayeny"] = "Ekonomika Radvanskoyi Ukhayeny",
  ["ekonomika sovetskoi ukrainy"] = "Ekon Sov Ukr",
  ["ekonomiska samfundets tidskrift"] = "Ekon Samf Tidskr",
  ["ekonomska istrazivanja"] = "Ekon. Istrazivanja",
  ["ekonomska misao"] = "Ekon Misao",
  ["ekonomska misao i praksa"] = "Ekon. Misao Praksa",
  ["ekonomska revija"] = "Ekon Revija",
  ["ekonomski pregled"] = "Ekon Pregl",
  ["ekotekhnologii i resursosberezhenie"] = "Ekotekhnol. Resursosberezhenie",
  ["eksperimental'naia i klinicheskaia farmakologiia"] = "Eksp Klin Farmakol",
  ["eksperimental'naia khirurgiia"] = "Eksp Khirurgiia",
  ["eksperimental'naia khirurgiia i anesteziologiia"] = "Eksp Khir Anesteziol",
  ["eksperimentalna meditsina i morfologiia"] = "Eksp Med Morfol",
  ["eksperimentalnaia i klinicheskaia farmakologiia"] = "Eksp. Klin. Farmakol.",
  ["eksperimentalnaia i klinicheskaia gastroenterologiia"] = "Eksp. Klin. Gastroenterol.",
  ["eksperimentalnaia khirurgiia i anesteziologiia"] = "Eksp. Khir. Anesteziol.",
  ["eksperimentalnaia onkologiia"] = "Eksp. Onkol.",
  ["eksperimentalʹnai︠a︡ onkologii︠a︡"] = "Eksp Onkol",
  ["eksperimental’naia i klinicheskaia farmakologiia"] = "Eksp Klin Farmakol",
  ["eksperimental’naia khirurgiia"] = "Eksp Khirurgiia",
  ["eksperimental’naia khirurgiia i anesteziologiia"] = "Eksp Khir Anesteziol",
  ["eksperimental’naia onkologiia"] = "Eksp Onkol",
  ["eksperimental’naya i klinicheskaya farmakologiya"] = "Eksp. Klin. Farmakol.",
  ["el bisturí"] = "Bisturi (S Juan)",
  ["el caribe contemporaneo"] = "Caribe Contemp",
  ["el cooperador dental; cooperativismo, informacion y ciencia odontologica"] = "Coop Dent (B Aires)",
  ["el cooperador dental; cooperativismo, información y ciencia odontológica"] = "Coop Dent (B Aires)",
  ["el demográfico"] = "Demografico",
  ["el dia medico"] = "Dia Med",
  ["el día médico"] = "Dia Med",
  ["el farmaceutico"] = "Farmaceutico",
  ["el hospital : la revista interamericana de hospitales"] = "Hospital",
  ["el mecanico dental"] = "Mec Dent",
  ["el medico"] = "Medico",
  ["el medico practico"] = "Med Pract",
  ["el medico práctico"] = "Med Pract",
  ["el médico"] = "Medico",
  ["el monitor médico-farmacéutico"] = "Monit Med Farm",
  ["el museo de pontevedra"] = "MusPontevedra",
  ["el odontologo"] = "Odontologo",
  ["el odontólogo"] = "Odontologo",
  ["el palacio"] = "Palacio",
  ["el peruano : normas legales :. peru"] = "Peruano Normas Legales",
  ["el peruano. diario oficial. peru"] = "El Peruano D Of",
  ["el progreso médico; revista de la asociación médica latinoamericana"] = "Prog Med Rev Asoc Med Latinoam",
  ["el protesista dental"] = "Protes Dent",
  ["el torax"] = "Torax",
  ["el trimestre economico"] = "Trim. Econ.",
  ["el trimestre económico"] = "Trimest Econ",
  ["elder law review (penrith south, n.s.w.)"] = "Elder Law Rev (Penrith South)",
  ["elderly care"] = "Elder Care",
  ["election law journal"] = "Elect Law J",
  ["electoral studies"] = "Elect Stud",
  ["electric power components & systems"] = "Electr. Power Compon. Syst.",
  ["electric power systems research"] = "Electr. Power Syst. Res.",
  ["electric power systems reserch"] = "Electr. Power Syst. Res.",
  ["electrical engineering"] = "Electr. Eng.",
  ["electrical engineering and applied signal processing series"] = "Electr. Eng. Appl. Signal Process. Ser.",
  ["electrical engineering and electronics"] = "Elec. Eng. Electron.",
  ["electrical engineering communications and signal processing series"] = "Electr. Engrg. Commun. Signal Process. Ser.",
  ["electrical engineering in japan"] = "Electr. Eng. Jpn.",
  ["electrical materials and applications"] = "Electr. Mater. Appl.",
  ["electricity journal"] = "Electr. J.",
  ["electro- and magnetobiology"] = "Electro- Magnetobiol.",
  ["electroanalysis (new york)"] = "Electroanalysis",
  ["electrocatalysis (new york, n.y.)"] = "Electrocatalysis (N Y)",
  ["electrochemical and solid-state letters"] = "Electrochem. Solid-State Lett.",
  ["electrochemical energy reviews"] = "Electrochem. Energy Rev.",
  ["electrochemical energy technology"] = "Electrochem. Energy Technol.",
  ["electrochemical science advances"] = "Electrochem. Sci. Adv.",
  ["electrochemical society interface"] = "Electrochem. Soc. Interface",
  ["electrochemistry (tokyo, japan)"] = "Electrochemistry (Tokyo)",
  ["electrochemistry communications"] = "Electrochem. Commun.",
  ["electrochimica acta"] = "Electrochim. Acta",
  ["electrocomponent science and technology"] = "Electrocomponent Sci. Technol.",
  ["electrodeposition and surface treatment"] = "Electrodeposition Surf. Treat.",
  ["electroencephalography and clinical neurophysiology"] = "Electroencephalogr Clin Neurophysiol",
  ["electroencephalography and clinical neurophysiology. supplement"] = "Electroencephalogr Clin Neurophysiol Suppl",
  ["electrofisiología & arritmias"] = "Electrofisiol Arritm",
  ["electrolyte & blood pressure : e & bp"] = "Electrolyte Blood Press",
  ["electromagnetic biology and medicine"] = "Electromagn Biol Med",
  ["electromagnetic waves (cambridge, mass.)"] = "Electromagn Waves (Camb)",
  ["electromagnetics, wireless, radar, and microwaves"] = "Electromagn. Wirel. Radar Microw.",
  ["electromyography and clinical neurophysiology"] = "Electromyogr Clin Neurophysiol",
  ["electron backscatter diffraction in materials science"] = "Electron Backscatter Diffr. Mater. Sci.",
  ["electron microscopy"] = "Electron Microsc.",
  ["electron microscopy reviews"] = "Electron Microsc Rev",
  ["electronic & electrical engineering research studies: electromagnetic applications series"] = "Electron. Electric. Engrg. Res. Stud. Electromagnet. Appl. Ser.",
  ["electronic commerce research and applications"] = "Electron Commer Res Appl",
  ["electronic communications in probability"] = "Electron. Commun. Probab.",
  ["electronic engineering systems series"] = "Electron. Eng. Syst. Ser.",
  ["electronic international journal of time use research"] = "Electron Int J Time Use Res",
  ["electronic journal of applied statistical analysis"] = "Electron. J. Appl. Stat. Anal.",
  ["electronic journal of biotechnology"] = "Electron. J. Biotechnol.",
  ["electronic journal of biotechnology : ejb"] = "Electron J Biotechnol",
  ["electronic journal of boundary elements"] = "Electron. J. Boundary Elem.",
  ["electronic journal of combinatorics"] = "Electron. J. Combin.",
  ["electronic journal of differential equations"] = "Electron. J. Differential Equations",
  ["electronic journal of differential equations conference"] = "Electron. J. Differ. Equ. Conf.",
  ["electronic journal of geotechnical engineering"] = "Electron. J. Geotech. Eng.",
  ["electronic journal of graph theory and applications. ejgta"] = "Electron. J. Graph Theory Appl. (EJGTA)",
  ["electronic journal of information systems in developing countries"] = "Electron. J. Inf. Syst. Dev. Countries",
  ["electronic journal of linear algebra"] = "Electron. J. Linear Algebra",
  ["electronic journal of mathematical analysis and applications"] = "Electron J Math Anal Appl",
  ["electronic journal of mathematical analysis and applications. ejmaa"] = "Electron. J. Math. Anal. Appl.",
  ["electronic journal of operational meteorology"] = "Electron. J. Oper. Meteorol.",
  ["electronic journal of probability"] = "Electron. J. Probab.",
  ["electronic journal of qualitative theory of differential equations"] = "Electron. J. Qual. Theory Differ. Equ.",
  ["electronic journal of severe storms metereology"] = "Electron. J. Severe Storms Metereol.",
  ["electronic journal of statistics"] = "Electron. J. Stat.",
  ["electronic journal of theoretical chemistry"] = "Electron. J. Theor. Chem.",
  ["electronic materials letters"] = "Electron. Mater. Lett.",
  ["electronic medical digest"] = "Electron Med Dig",
  ["electronic news"] = "Electron. News",
  ["electronic notes in discrete mathematics"] = "Electron. Notes Discrete Math.",
  ["electronic notes in theoretical computer science"] = "Electron. Notes Theor. Comput. Sci.",
  ["electronic notes in theoretical informatics and computer science"] = "Electron. Notes Theor. Inform. Comput. Sci.",
  ["electronic physician"] = "Electron Physician",
  ["electronic proceedings in theoretical computer science"] = "Electron Proc Theor Comput Sci",
  ["electronic proceedings in theoretical computer science (eptcs)"] = "Electron. Proc. Theor. Comput. Sci. (EPTCS)",
  ["electronic research announcements in mathematical sciences"] = "Electron. Res. Announc. Math. Sci.",
  ["electronic research announcements of the american mathematical society"] = "Electron. Res. Announc. Amer. Math. Soc.",
  ["electronic research archive"] = "Electron. Res. Arch.",
  ["electronic structure"] = "Electron. Struct.",
  ["electronic systems news"] = "Electron. Syst. News",
  ["electronic technology"] = "Electron. Technol.",
  ["electronic transactions on artificial intelligence"] = "Electron. Trans. Artif. Intell.",
  ["electronic transactions on numerical analysis"] = "Electron. Trans. Numer. Anal.",
  ["electronic workshops in computing"] = "Electron. Workshops Comput.",
  ["electronic, magnetic and photonic materials division (empmd) monograph series"] = "Electron. Magn. Photonic Mater. Div. (EMPMD) Monogr. Ser.",
  ["electronics and communications in japan"] = "Electron. Commun. Jpn.",
  ["electronics and communications in japan = denki gakkai ronbunshi"] = "Electron Commun Jpn",
  ["electronics and power"] = "Electron. Power",
  ["electronics education"] = "Electron. Educ.",
  ["electronics letters"] = "Electron. Lett.",
  ["electronics texts for engineers and scientists"] = "Electron. Texts Engrs. Sci.",
  ["electrophoresis (weinheim, federal republic of germany)"] = "Electrophoresis (Weinheim, Fed. Repub. Ger.)",
  ["elektro medizin; biomedizin und technik"] = "Elektromed Biomed Tech",
  ["elektromedizin, biomedizin und technik"] = "Elektromed. Biomed. Tech.",
  ["elektronika ir elektrotechnika"] = "Elektron. Elektrotech.",
  ["elektrotechnik und informationstechnik"] = "Elektrotech. Inf.",
  ["elektrotechnik und maschinenbau"] = "Elektrotech. Maschinenbau",
  ["elektrotechnische zeitschrift"] = "Elektrotech. Z.",
  ["elementa (washington, d.c.)"] = "Elementa (Wash D C)",
  ["elementa: science of the anthropocene"] = "Elem. Sci. Anth.",
  ["elemente der mathematik"] = "Elem. Math.",
  ["elements (québec, québec)"] = "Elements (Que)",
  ["elements in applied category theory"] = "Elem. Appl. Categ. Theory",
  ["elements in decision theory and philosophy"] = "Elem. Decis. Theory Philos.",
  ["elements in non-local data interactions: foundations and applications"] = "Elem. Non-local Data Interact.: Found. Appl.",
  ["elements in the philosophy of biology"] = "Elem. Philos. Biol.",
  ["elements in the philosophy of mathematics"] = "Elem. Philos. Math.",
  ["elenchos. rivista di studi sul pensiero antico"] = "Elenchos",
  ["ellis horwood series in physics and its applications"] = "Ellis Horwood Ser. Phys. Appl.",
  ["elsevier oceanography series"] = "Elsevier Oceanogr. Ser.",
  ["eltex studientexte elektrotechnik"] = "ELTEX Studientexte Elektrotech.",
  ["em (pittsburgh, pa.)"] = "EM (Pittsburgh Pa)",
  ["embo journal"] = "EMBO J.",
  ["embo molecular medicine"] = "EMBO Mol Med",
  ["embo reports"] = "EMBO Rep",
  ["emc. pediatría"] = "EMC Pediatr",
  ["eme eme; estudios dominicanos"] = "Eme Eme Estud Domin",
  ["emedical research"] = "Emed Res",
  ["emergence, complexity and computation"] = "Emerg. Complex. Comput.",
  ["emergencias : revista de la sociedad española de medicina de emergencias"] = "Emergencias",
  ["emergency (tehran, iran)"] = "Emerg (Tehran)",
  ["emergency department news"] = "Emerg. Dep. News",
  ["emergency department news : edn"] = "Emerg Dep News",
  ["emergency health services quarterly"] = "Emerg Health Serv Q",
  ["emergency health services review"] = "Emerg Health Serv Rev",
  ["emergency medical services"] = "Emerg Med Serv",
  ["emergency medicine"] = "Emerg Med (N Y)",
  ["emergency medicine (fremantle, w.a.)"] = "Emerg Med (Fremantle)",
  ["emergency medicine (glendale, calif.) [sound recording]"] = "Emerg Med",
  ["emergency medicine : open access"] = "Emerg Med (Los Angel)",
  ["emergency medicine australasia"] = "Emerg. Med. Australas.",
  ["emergency medicine australasia : ema"] = "Emerg Med Australas",
  ["emergency medicine clinics of north america"] = "Emerg Med Clin North Am",
  ["emergency medicine international"] = "Emerg Med Int",
  ["emergency medicine investigations"] = "Emerg Med Investig",
  ["emergency medicine journal"] = "Emerg. Med. J.",
  ["emergency medicine journal : emj"] = "Emerg Med J",
  ["emergency medicine practice"] = "Emerg Med Pract",
  ["emergency medicine reports"] = "Emerg Med Rep",
  ["emergency nurse"] = "Emerg. Nurse",
  ["emergency nurse : the journal of the rcn accident and emergency nursing association"] = "Emerg Nurse",
  ["emergency planning digest"] = "Emerg Plann Dig",
  ["emergency radiology"] = "Emerg Radiol",
  ["emergent materials"] = "Emergent Mater",
  ["emergent scientist"] = "Emergent Sci.",
  ["emerging adulthood (print)"] = "Emerg Adulthood",
  ["emerging animal species"] = "Emerging Anim. Species",
  ["emerging contaminants"] = "Emerg Contam",
  ["emerging health threats journal"] = "Emerg Health Threats J",
  ["emerging infectious diseases"] = "Emerg Infect Dis",
  ["emerging materials research"] = "Emerging Mater. Res.",
  ["emerging methodologies and applications in modelling, identification and control"] = "Emerg. Methodol. Appl. Model. Identif. Control",
  ["emerging microbes & infections"] = "Emerg Microbes Infect",
  ["emerging themes in epidemiology"] = "Emerg Themes Epidemiol",
  ["emerging topics in life sciences"] = "Emerg Top Life Sci",
  ["emerging topics in statistics and biostatistics"] = "Emerg. Top. Stat. Biostat.",
  ["emerging trends in mechatronics"] = "Emerg. Trends Mechatron.",
  ["emerita / junta para ampliacion de estudios, centro de estudios historicos"] = "Emerita",
  ["emerita. revista de linguistica y filología clasica"] = "Emerita",
  ["emilia preromana"] = "EmPrerom",
  ["emirates journal of food and agriculture"] = "Emir J Food Agric",
  ["emisor demografico"] = "Emisor Demogr",
  ["emisor demográfico"] = "Emisor Demogr",
  ["emission control science and technology"] = "Emiss. Control Sci. Technol.",
  ["emj. gastroenterology"] = "EMJ Gastroenterol",
  ["emory international law review"] = "Emory Int Law Rev",
  ["emory law journal"] = "Emory Law J",
  ["emory magazine"] = "Emory Mag",
  ["emotion (washington, d.c.)"] = "Emotion",
  ["emotion review : journal of the international society for research on emotion"] = "Emot Rev",
  ["emotion, space and society"] = "Emot Space Soc",
  ["emotional and behavioural difficulties : the journal of the association of workers for children with emotional and behavioural difficulties"] = "Emot Behav Diffic",
  ["emphasis, nursing"] = "Emphasis Nurs",
  ["emphasis, nursing / los angeles county, harbor-ucla medical center, nursing department"] = "Emphasis Nurs",
  ["empire state report"] = "Emp State Rep",
  ["empirical economics"] = "Empir Econ",
  ["empirical musicology review : emr"] = "Empir Musicol Rev",
  ["empirical software engineering"] = "Empir Softw Eng",
  ["empleo y desempleo"] = "Empl Desempl",
  ["employee benefit plan review"] = "Employee Benefit Plan Rev",
  ["employee benefits journal"] = "Empl Benefits J",
  ["employee health & fitness"] = "Empl Health Fit",
  ["employee health and fitness newsletter"] = "Employee Health Fitness Newsl.",
  ["employee relations law journal"] = "Employee Relat Law J",
  ["employment relations today"] = "Employ Relat Today",
  ["empúries. revista de prehistòria, arqueologia i etnologia"] = "Empuries",
  ["ems cancer science journal"] = "EMS Cancer Sci J",
  ["ems magazine"] = "EMS Mag",
  ["ems monographs in mathematics"] = "EMS Monogr. Math.",
  ["ems series in industrial and applied mathematics"] = "EMS Ser. Ind. Appl. Math.",
  ["ems series of congress reports"] = "EMS Ser. Congr. Rep.",
  ["ems series of lectures in mathematics"] = "EMS Ser. Lect. Math.",
  ["ems surveys in mathematical sciences"] = "EMS Surv. Math. Sci.",
  ["ems textbooks in mathematics"] = "EMS Textbk. Math.",
  ["ems tracts in mathematics"] = "EMS Tracts Math.",
  ["ems world"] = "EMS World",
  ["emt journal"] = "EMT J.",
  ["emu - austral ornithology"] = "Emu - Austral Ornithol.",
  ["en familia (santo domingo, dominican republic)"] = "Familia",
  ["en familia. venezuela. ministerio de agricultura y cría. dirección de extensión. división de divulgación"] = "En familia",
  ["enalia. annual. english edition of the hellenic institute of marine archaeology"] = "EnaliaAnn",
  ["enchoria. zeitschrift für demotistik und koptologie"] = "Enchoria",
  ["enciclopedia dell’arte antica classica e orientale"] = "EAA",
  ["encyclopaedia of mathematical sciences"] = "Encyclopaedia Math. Sci.",
  ["encyclopedia of complexity and systems science"] = "Encycl. Complex. Syst. Sci.",
  ["encyclopedia of mathematics and its applications"] = "Encyclopedia Math. Appl.",
  ["encyclopedia of social work"] = "Encycl Soc Work",
  ["encyclopedia with semantic computing and robotic intelligence"] = "Encycl. Semant. Comput. Rob. Intell.",
  ["endangered species research"] = "Endanger Species Res",
  ["endo : revue française d'endodontie : publication officielle de la société française d'endodontie"] = "ENDO",
  ["endocrine and metabolic science"] = "Endocr. Metab. Sci.",
  ["endocrine connections"] = "Endocr Connect",
  ["endocrine development"] = "Endocr Dev",
  ["endocrine disruptors"] = "Endocr. Disruptors",
  ["endocrine disruptors (austin, tex.)"] = "Endocr Disruptors (Austin)",
  ["endocrine journal"] = "Endocr J",
  ["endocrine metabolic & immune disorders drug targets"] = "Endocr. Metab. Immune Disord. Drug Targets",
  ["endocrine oncology"] = "Endocr. Oncol.",
  ["endocrine pathology"] = "Endocr Pathol",
  ["endocrine practice : official journal of the american college of endocrinology and the american association of clinical endocrinologists"] = "Endocr Pract",
  ["endocrine regulations"] = "Endocr Regul",
  ["endocrine research"] = "Endocr Res",
  ["endocrine research communications"] = "Endocr Res Commun",
  ["endocrine reviews"] = "Endocr Rev",
  ["endocrine, metabolic & immune disorders drug targets"] = "Endocr Metab Immune Disord Drug Targets",
  ["endocrine, metabolic & immune disorders: drug targets"] = "Endocr., Metab. Immune Disord.: Drug Targets",
  ["endocrine, metabolic and immune disorders drug targets"] = "Endocr. Metab. Immune Disord. Drug Targets",
  ["endocrine-related cancer"] = "Endocr Relat Cancer",
  ["endocrinologia e scienza della costituzione"] = "Endocrinol Sci Cost",
  ["endocrinologia experimentalis"] = "Endocrinol Exp",
  ["endocrinologia japonica"] = "Endocrinol Jpn",
  ["endocrinologia y terapeutica"] = "Endocr Terap",
  ["endocrinologia y terapéutica"] = "Endocr Terap",
  ["endocrinología y nutrición : órgano de la sociedad española de endocrinología y nutrición"] = "Endocrinol Nutr",
  ["endocrinología, diabetes y nutrición"] = "Endocrinol Diabetes Nutr",
  ["endocrinology & metabolic syndrome : current research"] = "Endocrinol Metab Syndr",
  ["endocrinology (baltimore)"] = "Endocrinology (Baltimore)",
  ["endocrinology and metabolism"] = "Endocrinol. Metab.",
  ["endocrinology and metabolism (seoul, korea)"] = "Endocrinol Metab (Seoul)",
  ["endocrinology and metabolism clinics of north america"] = "Endocrinol Metab Clin North Am",
  ["endocrinology, diabetes & metabolism"] = "Endocrinol Diabetes Metab",
  ["endocrinology, diabetes & metabolism case reports"] = "Endocrinol Diabetes Metab Case Rep",
  ["endocrinology, diabetes and metabolism journal"] = "Endocrinol Diabetes Metab J",
  ["endocytobiosis and cell research"] = "Endocytobiosis Cell Res",
  ["endodoncia (madrid, spain)"] = "Endodoncia",
  ["endodontic practice"] = "Endod Prac",
  ["endodontic report"] = "Endod. Rep.",
  ["endodontic topics"] = "Endod Topics",
  ["endodontics & dental traumatology"] = "Endod Dent Traumatol",
  ["endodontics and dental traumatology"] = "Endod. Dent. Traumatol.",
  ["endokrynologia polska"] = "Endokrynol Pol",
  ["endokrynologia, diabetologia i choroby przemiany materii wieku rozwojowego : organ polskiego towarzystwa endokrynologow dzieciecych"] = "Endokrynol Diabetol Chor Przemiany Materii Wieku Rozw",
  ["endokrynologia, diabetologia i choroby przemiany materii wieku rozwojowego : organ polskiego towarzystwa endokrynologów dziecięcych"] = "Endokrynol Diabetol Chor Przemiany Materii Wieku Rozw",
  ["endoplasmic reticulum stress in diseases"] = "Endoplasmic Reticulum Stress Dis",
  ["endoscopic surgery and allied technologies"] = "Endosc Surg Allied Technol",
  ["endoscopic ultrasound"] = "Endosc Ultrasound",
  ["endoscopy international open"] = "Endosc Int Open",
  ["endothelium : journal of endothelial cell research"] = "Endothelium",
  ["enegie wasser praxis"] = "Energ. Wasser Prax.",
  ["energetic materials frontiers"] = "Energetic Mater. Front.",
  ["energie wasser praxis"] = "Energy Wasser Prax.",
  ["energotekhnologii i resursosberezhenie"] = "Energotekhnol. Resursosberezhenie",
  ["energy & environment"] = "Energy Environ.",
  ["energy & environmental materials"] = "Energy Environ. Mater.",
  ["energy & environmental science"] = "Energy Environ. Sci.",
  ["energy & fuels"] = "Energy Fuels",
  ["energy & fuels : an american chemical society journal"] = "Energy Fuels",
  ["energy (oxford, england)"] = "Energy (Oxf)",
  ["energy advances"] = "Energy Adv.",
  ["energy and ai"] = "Energy AI",
  ["energy and buildings"] = "Energy Build.",
  ["energy and built environment"] = "Energy Built Environ.",
  ["energy and climate change"] = "Energy Clim. Change",
  ["energy and emission control technologies"] = "Energy Emiss Control Technol",
  ["energy and environment"] = "Energy Environ.",
  ["energy and environmental materials"] = "Energy Environ. Mater.",
  ["energy and fuels"] = "Energy Fuels",
  ["energy and policy research"] = "Energy Policy Res.",
  ["energy and power engineering"] = "Energy Power Eng",
  ["energy conversion and economics"] = "Energy Convers. Econ.",
  ["energy conversion and management"] = "Energy Convers. Manage.",
  ["energy conversion and management: x"] = "Energy Convers. Manage.: X",
  ["energy economics"] = "Energy Econ.",
  ["energy efficiency"] = "Energy Effic.",
  ["energy engineering"] = "Energy Eng.",
  ["energy exploration & exploitation"] = "Energy Explor. Exploit.",
  ["energy for sustainable development"] = "Energy Sustainable Dev.",
  ["energy for sustainable development : the journal of the international energy initiative"] = "Energy Sustain Dev",
  ["energy geoscience"] = "Energy Geosci.",
  ["energy harvesting and systems"] = "Energy Harvesting Syst.",
  ["energy harvesting systems"] = "Energy Harvesting Syst.",
  ["energy informatics"] = "Energy Inf.",
  ["energy internet"] = "Energy Internet",
  ["energy journal"] = "Energy J.",
  ["energy material advances"] = "Energy Mater. Adv.",
  ["energy materials"] = "Energy Mater.",
  ["energy minimization methods in computer vision and pattern recognition. international workshop on energy minimization methods in computer vision and pattern recognition"] = "Energy Minimization Methods Comput Vis Pattern Recognit",
  ["energy policy"] = "Energy Policy",
  ["energy procedia"] = "Energy Procedia",
  ["energy reports"] = "Energy Rep.",
  ["energy research & social science"] = "Energy Res. Social Sci.",
  ["energy reviews"] = "Energy Rev.",
  ["energy science"] = "Energy Sci.",
  ["energy science & engineering"] = "Energy Sci Eng",
  ["energy sources, part a: recovery, utilization, and environmental effects"] = "Energy Sources Part A",
  ["energy sources, part b: economics, planning and policy"] = "Energy Sources Part B",
  ["energy storage"] = "Energy Storage",
  ["energy storage and saving"] = "Energy Storage Sav.",
  ["energy storage materials"] = "Energy Storage Mater",
  ["energy strategy reviews"] = "Energy Strategy Rev.",
  ["energy studies review"] = "Energy Stud. Rev.",
  ["energy systems"] = "Energy Syst.",
  ["energy systems in electrical engineering"] = "Energy Syst. Electr. Eng.",
  ["energy technology"] = "Energy Technol.",
  ["energy technology (weinheim, germany)"] = "Energy Technol (Weinh)",
  ["energy, ecology & environment"] = "Energy Ecol Environ",
  ["energy, ecology and environment"] = "Energy Ecol. Environ.",
  ["energy, sustainability and society"] = "Energy Sustain Soc",
  ["enfance; psychologie, pedagogie, neuropsychiatrie, sociologie"] = "Enfance",
  ["enfance; psychologie, pédagogie, neuropsychiatrie, sociologie"] = "Enfance",
  ["enfermedades del torax"] = "Enferm Torax",
  ["enfermedades del tórax"] = "Enferm Torax",
  ["enfermedades infecciosas y microbiologia clinica"] = "Enferm Infecc Microbiol Clin",
  ["enfermedades infecciosas y microbiología clínica"] = "Enferm Infecc Microbiol Clin",
  ["enfermeria clinica"] = "Enferm Clin",
  ["enfermeria intensiva"] = "Enferm. Intensiva",
  ["enfermeria intensiva / sociedad espanola de enfermeria intensiva y unidades coronarias"] = "Enferm Intensiva",
  ["enfermería clínica"] = "Enferm Clin",
  ["enfermería intensiva"] = "Enferm Intensiva",
  ["enfoque feminista"] = "Enfoque Fem",
  ["enfoques en atencion primaria"] = "Enfoques Aten Prim",
  ["enfoques en atención primaria"] = "Enfoques Aten Prim",
  ["engage/social action"] = "Engage Soc Action",
  ["engaging science, technology, and society"] = "Engag Sci Technol Soc",
  ["engei gakkai zasshi. engei gakkai (japan)"] = "Engei Gakkai Zasshi",
  ["engineered casting solutions"] = "Eng. Cast. Solutions",
  ["engineered regeneration"] = "Eng. Regener.",
  ["engineering & technology"] = "Eng. Technol.",
  ["engineering (beijing, china)"] = "Engineering (Beijing)",
  ["engineering analysis with boundary elements"] = "Eng. Anal. Boundary Elem.",
  ["engineering and mining journal"] = "Eng. Min. J.",
  ["engineering and mining journal press"] = "Eng. Min. J. Press",
  ["engineering and process economics"] = "Eng. Process Econ.",
  ["engineering and science"] = "Eng Sci",
  ["engineering applications of artificial intelligence"] = "Eng. Appl. Artif. Intell.",
  ["engineering applications of computational fluid mechanics"] = "Eng. Appl. Comput. Fluid Mech.",
  ["engineering applications of computational methods"] = "Eng. Appl. Comput. Methods",
  ["engineering biology"] = "Eng. Biol.",
  ["engineering computations"] = "Eng. Comput.",
  ["engineering construction and architectural management"] = "Eng. Constr. Archit. Manage.",
  ["engineering costs and production economics"] = "Eng. Costs Prod. Econ.",
  ["engineering economist"] = "Eng. Econ.",
  ["engineering failure analysis"] = "Eng. Fail. Anal.",
  ["engineering fracture mechanics"] = "Eng. Fract. Mech.",
  ["engineering geology"] = "Eng. Geol.",
  ["engineering in agriculture, environment and food"] = "Eng. Agric. Environ. Food",
  ["engineering in life sciences"] = "Eng Life Sci",
  ["engineering in medicine"] = "Eng Med",
  ["engineering journal american institute of steel construction"] = "Eng. J. Am. Inst. steel Constr.",
  ["engineering management journal"] = "Eng. Manage. J.",
  ["engineering materials"] = "Eng. Mater.",
  ["engineering materials and processes"] = "Eng. Mater. Process.",
  ["engineering microbiology"] = "Eng. Microbiol.",
  ["engineering optimization"] = "Eng. Optim.",
  ["engineering optimization. methods and applications"] = "Eng. Optim. Methods App.",
  ["engineering project organization journal"] = "Eng. Project Organ. J.",
  ["engineering reports"] = "Eng. Rep.",
  ["engineering reports : open access"] = "Eng Rep",
  ["engineering research express"] = "Eng. Res. Express",
  ["engineering science and education journal"] = "Eng. Sci. Educ. J.",
  ["engineering science and technology"] = "Eng. Sci. Technol.",
  ["engineering science and technology, an international journal"] = "Eng. Sci. Technol. Int. J.",
  ["engineering sciences"] = "Eng. Sci.",
  ["engineering structures"] = "Eng. Struct.",
  ["engineering structures and technologies"] = "Eng. Struct. Technol.",
  ["engineering studies"] = "Eng. Stud.",
  ["engineering systems and sustainability series"] = "Eng. Syst. Sustain. Ser.",
  ["engineering transactions"] = "Eng. Trans.",
  ["engineering with computers"] = "Eng Comput",
  ["english books, 1475-1640; consolidated cross index by stc numbers"] = "Early Engl Books (Cross Index)",
  ["english historical review"] = "EHR",
  ["english language notes"] = "Engl Lang Notes",
  ["english literary renaissance"] = "Engl Lit Renaiss",
  ["english studies (amsterdam, netherlands)"] = "Engl Stud (Amst)",
  ["enke reihe zur ao[a]"] = "Enke Reihe AO[A]",
  ["enlace. inter-american commission of women"] = "Enlace",
  ["enlb. emergency nurse legal bulletin"] = "ENLB Emerg Nurse Leg Bull",
  ["enlb: emergency nurse legal bulletin"] = "ENLB Emerg. Nurse Leg. Bull.",
  ["enliven. challenges in cancer detection and therapy"] = "Enliven Chall Cancer Detect Ther",
  ["enliven. journal of stem cell research & regenerative medicine"] = "Enliven J Stem Cell Res Regen Med",
  ["eno fo"] = "ENO FO",
  ["enquetes du musee de la vie wallonne"] = "Enquetes Mus Vie Wallonne",
  ["enquêtes du musée de la vie wallonne"] = "Enquetes Mus Vie Wallonne",
  ["ensaios de literatura e filologia, publ. do dept. de letras clássicas, fac. de letras univ. federal de minas gerais"] = "ELF",
  ["ensaios matematicos"] = "Ensaios Mat.",
  ["enseignement assiste par ordinateur"] = "Enseign. Assiste Ordinateur",
  ["enseignement des mathematiques"] = "Enseign. Math.",
  ["enseignement des mathématiques"] = "Enseign. Math.",
  ["enseignement sup-maths"] = "Enseign. SUP-Maths",
  ["ent & audiology news"] = "ENT Audiol News",
  ["enteric pathogens monitoring group report"] = "Enteric Pathogens Monit Group Rep",
  ["enterprise & society"] = "Enterp Soc",
  ["enterprise and innovation management studies"] = "Enterpr. Innovation Manage. Stud.",
  ["enterprise and society"] = "Enterprise Society",
  ["enterprise information systems"] = "Enterp Inf Syst",
  ["entertainment computing"] = "Entertain Comput",
  ["entomologia experimentalis et applicata"] = "Entomol Exp Appl",
  ["entomologia generalis"] = "Entomol. Gen.",
  ["entomologia hellenica"] = "Entomol Hell",
  ["entomologia sinica"] = "Zhongguo kun chong xue|Entomol. Sin.",
  ["entomologia sinica = zhongguo kun chong xue"] = "Entomol Sin",
  ["entomologia y vectores"] = "Entomol. Vect.",
  ["entomologica (bari, italy)"] = "Entomologica (Bari)",
  ["entomologica americana"] = "Entomol. Am.",
  ["entomologica fennica"] = "Entomol Fenn",
  ["entomologica romanica"] = "Entomol Rom",
  ["entomological news"] = "Entomol News",
  ["entomological research"] = "Entomol. Res.",
  ["entomological review"] = "Entomol Rev",
  ["entomological science"] = "Entomol. Sci.",
  ["entomologische abhandlungen"] = "Entomol Abh",
  ["entomologische berichte luzern"] = "Entomol. Ber. Luzern",
  ["entomologische berichten"] = "Entomolog Ber",
  ["entomologische blätter für biologie und systematik der käfer"] = "Entomol Bl Biol Syst Kafer",
  ["entomologische nachrichten und berichte"] = "Entomol Nachr Ber",
  ["entomologische zeitschrift"] = "Entomol. Z.",
  ["entomologische zeitschrift mit insektenbörse"] = "Entomol Z Insektenborse",
  ["entomologisches nachrichtenblatt (vienna, austria : 1994)"] = "Entomol Nachr Bl",
  ["entomologisk tidskrift"] = "Entomol Tidskr",
  ["entomologist's gazette"] = "Entomol Gaz",
  ["entomology, ornithology, & herpetology : current research"] = "Entomol Ornithol Herpetol",
  ["entomon; internationale zeitschrift für die gesamte insektenkunde"] = "Entomon Int Z Gesamte Insektenkunde",
  ["entre nous (copenhagen, denmark)"] = "Entre Nous Cph Den",
  ["entrepreneurship and regional development"] = "Entrepreneurship Reg. Devel.",
  ["entreprises et histoire"] = "Entrep Hist",
  ["entropy (basel, switzerland)"] = "Entropy (Basel)",
  ["entropy production in everyday live"] = "Entropy Prod. Everyday Live",
  ["entropy. an international and interdisciplinary journal of entropy and information studies"] = "Entropy",
  ["entscheidungen des bundesgerichtshofes in zivilsachen. germany (west). bundesgerichtshof"] = "Entscheid Bundesgerichtshofes Zivilsachen",
  ["enumerative combinatorics and applications"] = "Enumer. Comb. Appl.",
  ["environ. chem. preprints of extended abstracts presented at the acs national meeting, american chemical society, division of environmental chemistry"] = "Prepr. Ext. Abstr. ACS Natl. Meet., Am. Chem. Soc., Div.",
  ["environ. eng. journal of environmental science and health, part a: toxic/hazardous substances & environmental engineering"] = "J. Environ. Sci. Health: Toxic/Hazard. Subst.",
  ["environment & health"] = "Environ. Health",
  ["environment & planning a"] = "Environ Plan A",
  ["environment and behavior"] = "Environ Behav",
  ["environment and development economics"] = "Environ Dev Econ",
  ["environment and history"] = "Environ Hist Camb",
  ["environment and natural resources journal"] = "Environ Nat Resour J",
  ["environment and planning a"] = "Environ. Planning A",
  ["environment and planning a: economy and space"] = "Environ. Plann. A: Econ. Space",
  ["environment and planning b: urban analytics and city science"] = "Environ. Plann. B: Urban Anal. City Sci.",
  ["environment and planning c: government and policy"] = "Environ. Planning C: Govt. Pol.",
  ["environment and planning c: politics and spaces"] = "Environ. Plann. C: Polit. Spaces",
  ["environment and planning d: society and space"] = "Environ. Plann. D: Soc. Space",
  ["environment and planning e: nature and space"] = "Environ. Plann. E: Nat. Space",
  ["environment and planning f"] = "Environ. Plann. F",
  ["environment and planning. b, planning & design"] = "Environ Plann B Plann Des",
  ["environment and planning. b, urban analytics and city science"] = "Environ Plan B Urban Anal City Sci",
  ["environment and planning. c, government & policy"] = "Environ Plann C Gov Policy",
  ["environment and planning. d, society & space"] = "Environ Plan D",
  ["environment and planning. e, nature and space"] = "Environ Plan E Nat Space",
  ["environment and pollution (toronto, ont.)"] = "Environ Pollut (Tor)",
  ["environment and security"] = "Environ. Secur.",
  ["environment and urbanization"] = "Environ Urban",
  ["environment carcinogenesis and ecotoxicology reviews"] = "Environ. Carcinog. Ecotoxicol. Rev.",
  ["environment conservation journal"] = "Environ Conserv J",
  ["environment folio"] = "Environ Folio",
  ["environment international"] = "Environ Int",
  ["environment protection engineering"] = "Environ. Prot. Eng.",
  ["environment research journal"] = "Environ Res J",
  ["environment systems & decisions"] = "Environ Syst Decis",
  ["environment systems and decisions"] = "Environ. Syst. Dec.",
  ["environment, development and sustainability"] = "Environ Dev Sustain",
  ["environment, innovation and management"] = "Environ. Innovation Manage.",
  ["environment: science and policy for sustainable development"] = "Environ.: Sci. Policy Sustainable Dev.",
  ["environmental & resource economics"] = "Environ Resour Econ (Dordr)",
  ["environmental action"] = "Environ Aciton",
  ["environmental advances"] = "Environ. Adv.",
  ["environmental affairs"] = "Environ Aff",
  ["environmental analysis health and toxicology"] = "Environ. Anal. Health Toxicol.",
  ["environmental analysis, health and toxicology"] = "Environ Anal Health Toxicol",
  ["environmental and climate technologies"] = "Environ. Clim. Technol.",
  ["environmental and ecological statistics"] = "Environ. Ecol. Stat.",
  ["environmental and energy policy and the economy"] = "Environ Energy Policy Econ",
  ["environmental and engineering geoscience"] = "Environ. Eng. Geosci.",
  ["environmental and experimental biology"] = "Environ. Exp. Biol.",
  ["environmental and experimental botany"] = "Environ Exp Bot",
  ["environmental and molecular mutagenesis"] = "Environ Mol Mutagen",
  ["environmental and resource economics"] = "Environ. Resour. Econ.",
  ["environmental and sustainability indicators"] = "Environ. Sustainability Indic.",
  ["environmental archaeology"] = "Environ. Archaeol.",
  ["environmental bioindicators"] = "Environ. Bioindic.",
  ["environmental biology and medicine"] = "Environ Biol Med",
  ["environmental biology of fishes"] = "Environ. Biol. Fishes",
  ["environmental biosafety research"] = "Environ. Biosaf. Res.",
  ["environmental biotechnology"] = "Environ Biotechnol",
  ["environmental challenges"] = "Environ. Challenges",
  ["environmental change and security project report"] = "Environ Change Secur Proj Rep",
  ["environmental chemistry"] = "Environ. Chem.",
  ["environmental chemistry (collingwood, vic.)"] = "Environ Chem",
  ["environmental chemistry and ecotoxicology"] = "Environ. Chem. Ecotoxicol.",
  ["environmental chemistry letters"] = "Environ. Chem. Lett.",
  ["environmental claims journal"] = "Environ. Claims J.",
  ["environmental communication"] = "Environ Commun",
  ["environmental conservation"] = "Environ Conserv",
  ["environmental control in biology"] = "Environ. Control. Biol.",
  ["environmental data science"] = "Environ. Data Sci.",
  ["environmental development"] = "Environ Dev",
  ["environmental disease"] = "Environ Dis",
  ["environmental dna"] = "Environ. DNA",
  ["environmental dna (hoboken, n.j.)"] = "Environ DNA",
  ["environmental earth sciences"] = "Environ Earth Sci",
  ["environmental economics and policy studies"] = "Environ. Econ. Policy Stud.",
  ["environmental education and information"] = "Environ Educ Inf",
  ["environmental education research"] = "Environ Educ Res",
  ["environmental endocrinology"] = "Environ. Endocrinol.",
  ["environmental engineering and management journal"] = "Environ. Eng. Manage. J.",
  ["environmental engineering and policy"] = "Environ. Eng. Policy",
  ["environmental engineering research"] = "Environ.  Eng. Res.",
  ["environmental engineering science"] = "Environ. Eng. Sci.",
  ["environmental entomology"] = "Environ Entomol",
  ["environmental epidemiology"] = "Environ. Epidemiol.",
  ["environmental epidemiology (philadelphia, pa.)"] = "Environ Epidemiol",
  ["environmental epigenetics"] = "Environ Epigenet",
  ["environmental ethics"] = "Environ Ethics",
  ["environmental evidence"] = "Environ Evid",
  ["environmental fluid mechanics"] = "Environ. Fluid Mech.",
  ["environmental fluid mechanics (dordrecht, netherlands : 2001)"] = "Environ Fluid Mech (Dordr)",
  ["environmental forensics"] = "Environ. Forensics",
  ["environmental geochemistry and health"] = "Environ Geochem Health",
  ["environmental geology"] = "Environ. Geol.",
  ["environmental geology (new york, n.y. : 1993)"] = "Environ Geol",
  ["environmental geosciences"] = "Environ. Geosci.",
  ["environmental hazards"] = "Environ. Hazards",
  ["environmental health"] = "Environ. Health",
  ["environmental health : a global access science source"] = "Environ Health",
  ["environmental health and preventive medicine"] = "Environ Health Prev Med",
  ["environmental health and toxicology"] = "Environ Health Toxicol",
  ["environmental health insights"] = "Environ Health Insights",
  ["environmental health perspectives"] = "Environ Health Perspect",
  ["environmental health series. [rh] radiological health"] = "Environ Health Ser [Radiol Health]",
  ["environmental health series. radiological health"] = "Environ. Health Ser. [Radiol. Health]",
  ["environmental health series. rh, radiological health"] = "Environ Health Ser Radiol Health",
  ["environmental history"] = "Environ Hist Durh N C",
  ["environmental history newsletter"] = "Environ. Hist. Newsl.",
  ["environmental history review"] = "Environ. Hist. Rev.",
  ["environmental history review : ehr : a publication of the american society for environmental history"] = "Environ Hist Rev",
  ["environmental humanities"] = "Environ Humanit",
  ["environmental impact assessment review"] = "Environ Impact Assess Rev",
  ["environmental innovation and societal transitions"] = "Environ Innov Soc Transit",
  ["environmental justice"] = "Environ. Justice",
  ["environmental justice (print)"] = "Environ Justice",
  ["environmental law & management"] = "Environ. Law Manage.",
  ["environmental law (northwestern school of law)"] = "Environ Law",
  ["environmental letters"] = "Environ Lett",
  ["environmental management"] = "Environ Manage",
  ["environmental management and health"] = "Environ. Manage. Health",
  ["environmental medicine : annual report of the research institute of environmental medicine, nagoya university"] = "Environ Med",
  ["environmental microbiology"] = "Environ Microbiol",
  ["environmental microbiology reports"] = "Environ Microbiol Rep",
  ["environmental microbiome"] = "Environ Microbiome",
  ["environmental microbiomes"] = "Environ. Microbiomes",
  ["environmental modeling and assessment"] = "Environ. Model. Assess.",
  ["environmental modelling & software : with environment data news"] = "Environ Model Softw",
  ["environmental modelling and software"] = "Environ. Model. Softw.",
  ["environmental monitoring and assessment"] = "Environ Monit Assess",
  ["environmental monographs & symposia"] = "Environmental Monogr. Sympos.",
  ["environmental mutagenesis"] = "Environ Mutagen",
  ["environmental nanotechnology, monitoring & management"] = "Environ. Nanotechnol. Monit. Manage.",
  ["environmental physiology & biochemistry"] = "Environ Physiol Biochem",
  ["environmental physiology and biochemistry"] = "Environ. Physiol. Biochem.",
  ["environmental policy and governance"] = "Environ. Policy Governance",
  ["environmental policy and law"] = "Environ. Policy Law",
  ["environmental politics"] = "Env Polit",
  ["environmental pollutants and bioavailability"] = "Environ. Pollut. Bioavailability",
  ["environmental pollution"] = "Environ. Pollut.",
  ["environmental pollution (amsterdam, netherlands)"] = "Environ. Pollut. (Amsterdam, Neth.)",
  ["environmental pollution (barking, essex : 1987)"] = "Environ Pollut",
  ["environmental pollution (oxford, united kingdom)"] = "Environ. Pollut. (Oxford, U. K.)",
  ["environmental practice"] = "Environ. Pract.",
  ["environmental practice : journal of the national association of environmental professionals"] = "Environ Pract",
  ["environmental processes"] = "Environ. Processes",
  ["environmental progress"] = "Environ. Prog.",
  ["environmental progress & sustainable energy"] = "Environ. Prog. Sustainable Energy",
  ["environmental psychology research"] = "Environ. Psychol. Res.",
  ["environmental quality and safety"] = "Environ Qual Saf",
  ["environmental quality and safety. supplement"] = "Environ Qual Saf Suppl",
  ["environmental quality management"] = "Environ. Qual. Manage.",
  ["environmental quality; annual report of the council on environmental quality"] = "Environ Qual Annu Rep Counc Environ Qual",
  ["environmental research"] = "Environ Res",
  ["environmental research and technology"] = "Environ. Res. Technol.",
  ["environmental research communications"] = "Environ. Res. Commun.",
  ["environmental research letters"] = "Environ. Res. Lett.",
  ["environmental research letters : erl [web site]"] = "Environ Res Lett",
  ["environmental research: climate"] = "Environ. Res.: Clim.",
  ["environmental research: ecology"] = "Environ. Res.: Ecol.",
  ["environmental research: energy"] = "Environ. Res.: Energy",
  ["environmental research: food systems"] = "Environ. Res.: Food Syst.",
  ["environmental research: health"] = "Environ. Res.: Health",
  ["environmental research: infrastructure and sustainability"] = "Environ. Res.: Infrastruct. Sustainability",
  ["environmental review"] = "Environ Rev",
  ["environmental reviews"] = "Environ. Rev.",
  ["environmental science & policy"] = "Environ. Sci. Policy",
  ["environmental science & technology"] = "Environ Sci Technol",
  ["environmental science & technology letters"] = "Environ. Sci. Technol. Lett.",
  ["environmental science : water research & technology"] = "Environ Sci (Camb)",
  ["environmental science and ecotechnology"] = "Environ. Sci. Ecotechnol.",
  ["environmental science and pollution research"] = "Environ. Sci. Pollut. Res.",
  ["environmental science and pollution research international"] = "Environ Sci Pollut Res Int",
  ["environmental science and technology"] = "Environ. Sci. Technol.",
  ["environmental science and technology letters"] = "Environ. Sci. Technol. Lett.",
  ["environmental science nano"] = "Environ. Sci. Nano",
  ["environmental science. nano"] = "Environ Sci Nano",
  ["environmental science. processes & impacts"] = "Environ Sci Process Impacts",
  ["environmental science: advances"] = "Environ. Sci.: Adv.",
  ["environmental science: atmospheres"] = "Environ. Sci.: Atmos.",
  ["environmental science: nano"] = "Environ. Sci.: Nano",
  ["environmental science: processes & impacts"] = "Environ. Sci. Processes Impacts",
  ["environmental science: water research & technology"] = "Environ. Sci. Water Res. Technol.",
  ["environmental sciences"] = "Environ Sci (Ruse)",
  ["environmental sciences : an international journal of environmental physiology and toxicology"] = "Environ Sci",
  ["environmental sciences europe"] = "Environ. Sci. Eur.",
  ["environmental sociology"] = "Environ Sociol",
  ["environmental sustainability"] = "Environ. Sustainability",
  ["environmental systems research"] = "Environ. Syst. Res.",
  ["environmental technology"] = "Environ Technol",
  ["environmental technology & innovation"] = "Environ. Technol. Innovation",
  ["environmental technology reviews"] = "Environ. Technol. Rev.",
  ["environmental toxicology"] = "Environ Toxicol",
  ["environmental toxicology and chemistry"] = "Environ. Toxicol. Chem.",
  ["environmental toxicology and chemistry / setac"] = "Environ Toxicol Chem",
  ["environmental toxicology and pharmacology"] = "Environ Toxicol Pharmacol",
  ["environmental toxicology and water quality"] = "Environ. Toxicol. Water Qual.",
  ["environmental values"] = "Environ Values",
  ["environments (basel, switzerland)"] = "Environments (Basel)",
  ["environmetrics (london, ont.)"] = "Environmetrics",
  ["environnement africain"] = "Environ Afr",
  ["enzyme & protein"] = "Enzyme Protein",
  ["enzyme and microbial technology"] = "Enzyme Microb Technol",
  ["enzyme and protein"] = "Enzyme Protein",
  ["enzyme research"] = "Enzyme Res",
  ["enzymologia biologica et clinica"] = "Enzymol Biol Clin (Basel)",
  ["enzymology and molecular biology of carbonyl metabolism"] = "Enzymol Mol Biol Carbonyl Metab",
  ["eos (rome, italy)"] = "EOS",
  ["eos, transactions, american geophysical union"] = "EOS Trans. Trans. Am. Geophys. Union",
  ["eos, transactions, american geophysical union (eos)"] = "Eos, Trans. Amer. Geophys. Union",
  ["eos. commentarii societatis philologae polonorum"] = "Eos",
  ["ep lab digest"] = "EP Lab Dig",
  ["epa journal"] = "Epa J",
  ["epa journal / united states environmental protection agency, office of public awareness"] = "Epa J",
  ["epatologia (roma)"] = "Epatologia",
  ["epe journal"] = "EPE J.",
  ["epeirotike hestia"] = "Epeirotike Hestia",
  ["epeteris hetaireias stereoelladikon meleton"] = "Epet Etaireias Stereoelladikon Meleton",
  ["epeteris tes hetaireias eleiakon meleton"] = "EHEM",
  ["epeteris tes kentrou ereunes tes historias tou hellenikou dikaiou"] = "EHHD",
  ["epetēris hetaireias stereoelladikōn meletōn"] = "Epet Etaireias Stereoelladikon Meleton",
  ["ephemerides theologicae lovanienses"] = "EThL",
  ["ephemeris dacoromana. annuario della scuola romena di roma"] = "EphemDac",
  ["ephemeris napocensis"] = "EphemNapoc",
  ["epheta: revista de enfermeria y accion social"] = "Epheta",
  ["epheta; revista de enfermeria y accion social"] = "Epheta",
  ["epheta; revista de enfermería y acción social"] = "Epheta",
  ["epi newsletter"] = "EPI Newsl",
  ["epi newsletter / c expanded program on immunization in the americas"] = "EPI Newsl",
  ["epic series in computing"] = "EPiC Ser. Comput.",
  ["epidemiologia e prevenzione"] = "Epidemiol Prev",
  ["epidemiologia e psichiatria sociale"] = "Epidemiol Psichiatr Soc",
  ["epidemiologia e serviços de saúde : revista do sistema unico de saúde do brasil"] = "Epidemiol Serv Saude",
  ["epidemiologic methods"] = "Epidemiol Methods",
  ["epidemiologic perspectives & innovations : ep+i"] = "Epidemiol Perspect Innov",
  ["epidemiologic reviews"] = "Epidemiol Rev",
  ["epidemiological and vital statistics report. rapport épidémiologique et démographique"] = "Epidemiol Vital Stat Rep",
  ["epidemiological bulletin"] = "Epidemiol Bull",
  ["epidemiologie, mikrobiologie, imunologie"] = "Epidemiol. Mikrobiol. Imunol.",
  ["epidemiologie, mikrobiologie, imunologie : casopis spolecnosti pro epidemiologii a mikrobiologii ceske lekarske spolecnosti j.e. purkyne"] = "Epidemiol Mikrobiol Imunol",
  ["epidemiologie, mikrobiologie, imunologie : casopis spolecnosti pro epidemiologii a mikrobiologii ceské lékarské spolecnosti j.e. purkyne"] = "Epidemiol Mikrobiol Imunol",
  ["epidemiology (cambridge, mass.)"] = "Epidemiology",
  ["epidemiology (sunnyvale, calif.)"] = "Epidemiology (Sunnyvale)",
  ["epidemiology and community health"] = "Epidemiol Community Health",
  ["epidemiology and health"] = "Epidemiol Health",
  ["epidemiology and infection"] = "Epidemiol Infect",
  ["epidemiology and psychiatric sciences"] = "Epidemiol Psychiatr Sci",
  ["epidemiology bulletin (taipei, taiwan)"] = "Epidemiol Bull (Taipei Taiwan)",
  ["epidemiology research international"] = "Epidemiol Res Int",
  ["epidemiology, biostatistics and public health"] = "Epidemiol Biostat Public Health",
  ["epigenetic diagnosis & therapy"] = "Epigenet Diagn Ther",
  ["epigenetics & chromatin"] = "Epigenetics Chromatin",
  ["epigenetics : official journal of the dna methylation society"] = "Epigenetics",
  ["epigenetics communications"] = "Epigenet. Commun.",
  ["epigenetics insights"] = "Epigenet Insights",
  ["epigraphica anatolica. zeitschrift für epigraphik und historische geographie anatoliens"] = "EpigrAnat",
  ["epigraphica anatolica: zeitschrift für epigraphik und historische geographie anatoliens"] = "EA",
  ["epigraphische studien"] = "Epigraph Stud",
  ["epijournal de geometrie algebrique. epiga"] = "Epijournal Geom. Algebrique",
  ["epilepsia open"] = "Epilepsia Open",
  ["epilepsy & behavior : e&b"] = "Epilepsy Behav",
  ["epilepsy & behavior case reports"] = "Epilepsy Behav Case Rep",
  ["epilepsy & behavior reports"] = "Epilepsy Behav Rep",
  ["epilepsy currents"] = "Epilepsy Curr",
  ["epilepsy currents / american epilepsy society"] = "Epilepsy Curr",
  ["epilepsy journal"] = "Epilepsy J",
  ["epilepsy research"] = "Epilepsy Res",
  ["epilepsy research and treatment"] = "Epilepsy Res Treat",
  ["epilepsy research. supplement"] = "Epilepsy Res Suppl",
  ["epileptic disorders"] = "Epileptic. Disord.",
  ["epileptic disorders : international epilepsy journal with videotape"] = "Epileptic Disord",
  ["epinfo : epidemiológiai információs hetilap"] = "Epinfo",
  ["episteme (edinburgh)"] = "Episteme (Edinb)",
  ["episteme. a series in the foundational, methodological, philosophical, psychological, sociological, and political aspects of the sciences, pure and applied"] = "Episteme",
  ["episteme; rivista critica di storia delle scienze mediche e biologiche"] = "Episteme",
  ["epistemonike epeterida tes philosophikes scholes tou aristoteleiou panepistemiou thessalonikes, tmema philosophias"] = "EEThess(philos)",
  ["epistemonike epeteris tes philosophikes scholes tou panepistemiou athenon"] = "EEAth",
  ["epithelial cell biology"] = "Epithelial Cell Biol",
  ["epitheorese klinikes farmakologias kai farmakokinetikes, international edition"] = "Epitheor. Klin. Farmakol. Farmakokinet. Int. Ed.",
  ["epitheorese koinonikon ereunon"] = "Epitheor Koin Ereun",
  ["epizootiology and animal health in west africa"] = "Epizoot Anim Health West Afr",
  ["epj applied metamaterials"] = "EPJ Appl. Metamater.",
  ["epj applied physics"] = "EPJ Appl. Phys.",
  ["epj data science"] = "EPJ Data Sci.",
  ["epj n nuclear sciences & technologies"] = "EPJ N Nucl. Sci. Technol.",
  ["epj nonlinear biomedical physics"] = "EPJ Nonlinear Biomed. Phys.",
  ["epj nuclear sciences & technologies"] = "EPJ Nucl. Sci. Technol.",
  ["epj nuclear sciences and technologies"] = "EPJ Nucl. Sci. Technol.",
  ["epj photovoltaics"] = "EPJ Photovoltaics",
  ["epj plus"] = "EPJ Plus",
  ["epj quantum technnology"] = "EPJ Quantum Technol.",
  ["epj quantum technology"] = "EPJ Quantum Technol",
  ["epj special topics"] = "EPJ Special Topics",
  ["epj techniques and instrumentation"] = "EPJ Tech. Instrum.",
  ["epj web of conferences"] = "EPJ Web Conf",
  ["epl (europhysics letters)"] = "Europhys. Lett.",
  ["eppo bulletin"] = "EPPO Bull.",
  ["equine and comparative exercise physiology"] = "Equine Comp. Exercise Physiol.",
  ["equine veterinary education"] = "Equine Vet Educ",
  ["equine veterinary journal"] = "Equine Vet J",
  ["equine veterinary journal. supplement"] = "Equine Vet J Suppl",
  ["equinoxe (cayenne, french guiana)"] = "Equinoxe",
  ["equity & excellence in education : university of massachusetts school of education journal"] = "Equity Excell Educ",
  ["era socialista"] = "Era Social",
  ["era socialistă"] = "Era Social",
  ["eranos. acta philologica suecana"] = "Eranos",
  ["eranos: acta philologica suecana"] = "Eranos",
  ["ercoftac series"] = "ERCOFTAC Ser.",
  ["erdkundliches wissen"] = "Erdkd Wissen",
  ["erdoel & kohle erdgas, petrochemie"] = "Erdoel, Kohle, Erdgas, Petrochem.",
  ["erdoel, erdgas, kohle"] = "Erdoel, Erdgas, Kohle",
  ["eretria. fouilles et recherches"] = "Eretria",
  ["eretz-israel. archaeological, historical and geographical studies"] = "Eretz-Israel",
  ["erevanskiy gosudarstvennyy universitet"] = "Erevan. Gos. Univ. Uchen. Zap. Estestv. Nauki",
  ["ergebnisse der allgemeinen pathologie und pathologischen anatomie"] = "Ergeb Allg Pathol Pathol Anat",
  ["ergebnisse der anatomie und entwicklungsgeschichte"] = "Ergeb Anat Entwicklungsgesch",
  ["ergebnisse der biologie"] = "Ergeb Biol",
  ["ergebnisse der chirurgie und orthopadie"] = "Ergeb. Chir. Orthop.",
  ["ergebnisse der chirurgie und orthopädie"] = "Ergeb Chir Orthop",
  ["ergebnisse der enzymforschung"] = "Ergeb Enzymforsch",
  ["ergebnisse der gesamten tuberkuloseforschung"] = "Ergeb Gesamten Tuberkuloseforsch",
  ["ergebnisse der hygiene, bakteriologie, immunitatsforschung und experimentellen therapie"] = "Ergeb Hyg Bakteriol Immunitatsforsch Exp Ther",
  ["ergebnisse der hygiene, bakteriologie, immunitätsforschung und experimentellen therapie"] = "Ergeb Hyg Bakteriol Immunitatsforsch Exp Ther",
  ["ergebnisse der inneren medizin und kinderheilkunde"] = "Ergeb Inn Med Kinderheilkd",
  ["ergebnisse der mathematik und ihrer grenzgebiete"] = "Ergeb. Math. Grenzgeb. (3)",
  ["ergebnisse der mathematik und ihrer grenzgebiete. 3. folge. a series of modern surveys in mathematics"] = "Ergeb. Math. Grenzgeb. (3)",
  ["ergebnisse der mikrobiologie, immunitatsforschung und experimentellen therapie"] = "Ergeb. Mikrobiol. Immunitatsforsch. Exp. Ther.",
  ["ergebnisse der mikrobiologie, immunitätsforschung und experimentellen therapie"] = "Ergeb Mikrobiol Immunitatsforsch Exp Ther",
  ["ergebnisse der physiologie, biologischen chemie und experimentellen pharmakologie"] = "Ergeb Physiol",
  ["ergebnisse der wissenschaftlichen untersuchungen im schweizerischen nationalpark"] = "Ergeb. wiss. Unters. schweiz. Natl.park",
  ["ergo (ann arbor, mich.)"] = "Ergo (Ann Arbor)",
  ["ergodic theory and dynamical systems"] = "Ergodic Theory Dynam. Systems",
  ["ergonomics in design"] = "Ergon. Des.",
  ["ergonomics in design : the magazine of human factors applications"] = "Ergon Des",
  ["ergonomics international journal"] = "Ergon Int J",
  ["erj open research"] = "ERJ Open Res",
  ["erkenntnis. an international journal of analytic philosophy"] = "Erkenntnis",
  ["erkenntnis. an international journal of scientific philosophy"] = "Erkenntnis",
  ["ernahrungsforschung. berichte und mitteilungen"] = "Ernahrungsforsch Ber Mitt",
  ["ernährungsforschung. berichte und mitteilungen"] = "Ernahrungsforsch Ber Mitt",
  ["ernst schering foundation symposium proceedings"] = "Ernst Schering Found Symp Proc",
  ["ernst schering research foundation workshop"] = "Ernst Schering Res Found Workshop",
  ["ernährung - wissenschaft und praxis"] = "Ernahrung Wiss. Prax.",
  ["ers spectrum"] = "ERS Spectr",
  ["ertel’s collected papers"] = "Ertel’s Collect. Pap.",
  ["erziehung und unterricht; osterreichische padagogische zeitschrift"] = "Erzieh Unterr",
  ["erziehung und unterricht; österreichische pädagogische zeitschrift"] = "Erzieh Unterr",
  ["es + t"] = "Es +T",
  ["esa bulletin"] = "ESA Bull.",
  ["esa bulletin. bulletin ase. european space agency"] = "ESA Bull",
  ["esa journal"] = "ESA J",
  ["esa scientific and technical review"] = "ESA Sci. Tech. Rev.",
  ["esaim probability and statistics"] = "ESAIM Probab. Stat.",
  ["esaim proceedings"] = "ESAIM Proc.",
  ["esaim. control, optimisation and calculus of variations"] = "ESAIM. Control. Optim. Calc. Var.",
  ["esaim. mathematical modelling and numerical analysis"] = "ESAIM. Math. Model. Numer. Anal.",
  ["esaim. mathematical modelling and numerical analysis = esaim. modélisation mathématique et analyse numérique : m=2an"] = "Esaim Math Model Numer Anal",
  ["esaim. probability and statistics"] = "ESAIM Probab. Stat.",
  ["esaim. probability and statistics = probabilités et statistique : p & s"] = "ESAIM Probab Stat",
  ["esaim. proceedings and surveys"] = "ESAIM Proc Surv",
  ["esaim: proceedings and surveys"] = "ESAIM: Proc. Surv.",
  ["esakia : occasional papers of the hikosan biological laboratory in entomology"] = "Esakia",
  ["esc heart failure"] = "ESC Heart Fail",
  ["escape. european symposium on computer aided process engineering"] = "ESCAPE",
  ["escritos de psicología"] = "Escr Psicol",
  ["esencia odontologica"] = "Esencia Odontol",
  ["esercitazioni di matematica"] = "Esercit. Mat.",
  ["esi lectures in mathematics and physics"] = "ESI Lect. Math. Phys.",
  ["esis publication"] = "ESIS Publ.",
  ["esmo open"] = "ESMO Open",
  ["esope : bi-mensuel social, politique, economique / publie avec la collaboration de la s.e.d.d.e.s"] = "Esope",
  ["esope : bi-mensuel social, politique, économique"] = "Esope",
  ["esophagus : official journal of the japan esophageal society"] = "Esophagus",
  ["espace, populations, societes"] = "Espace Popul Soc",
  ["espace, populations, sociétés"] = "Espace Popul Soc",
  ["espaces et societes (paris, france)"] = "Espaces Soc",
  ["espaces et sociétés (paris, france)"] = "Espaces Soc",
  ["espacio abierto"] = "Espac Abierto",
  ["espacio, tiempo y forma. revista de la facultad de geografia e historia. serie 2, historia antigua"] = "EspacioHist",
  ["espacio, tiempo y forma: revista de la facultad de geografia e historia, ser. 2"] = "ETF(hist)",
  ["espacio, tiempo y forma: revista de la facultad de geografía e historia, ser. 1"] = "ETF(arqueol)",
  ["esperienze dermatologiche"] = "Esper Dermatol",
  ["esperienze e ricerche"] = "Esper. ric.",
  ["espiral (guadalajara, mexico)"] = "Espiral",
  ["esprit (paris, france : 1932)"] = "Esprit",
  ["esquire (new york, n.y.)"] = "Esquire",
  ["esquisses mathematiques"] = "Esquisses Math.",
  ["esrc data archive bulletin"] = "ESRC Data Arch Bull",
  ["essays and studies (london, england : 1950)"] = "Essays Stud (Lond)",
  ["essays in arts and sciences"] = "Essays Arts Sci",
  ["essays in biochemistry"] = "Essays Biochem",
  ["essays in cognitive psychology"] = "Essays Cogn. Psych.",
  ["essays in fundamental immunology"] = "Essays Fundam Immunol",
  ["essays in neurochemistry and neuropharmacology"] = "Essays Neurochem Neuropharmacol",
  ["essential drugs monitor"] = "Essent Drugs Monit",
  ["essential number theory"] = "Essent. Number Theory",
  ["essential psychopharmacology"] = "Essent Psychopharmacol",
  ["essential textbooks in mathematics"] = "Essent. Textb. Math.",
  ["essex archaeology and history : the transactions of the essex archaeological society"] = "Trans Essex Archaeol Soc",
  ["essex institute historical collections"] = "Essex Inst Hist Collect",
  ["essex journal"] = "Essex J",
  ["estadistica panamena. boletin"] = "Estad Panamena",
  ["estadistica panamena. situacion demografica, seccion 231: migracion internacional. panama. direccion de estadistica y censo"] = "Estad Panamena Situac Demogr Secc 231 Migr Migr Int",
  ["estadistica venezolana"] = "Estad Venez",
  ["estadistica. revista semestral del instituto interamericano de estadistica"] = "Estadistica",
  ["estadística panameña. boletín"] = "Estad Panamena",
  ["estadística panameña. situación demográfica, sección 231: migración internacional. panama. dirección de estadística y censo"] = "Estad Panamena Situac Demogr Secc 231 Migr Migr Int",
  ["estadística venezolana"] = "Estad Venez",
  ["este pais (mexico city, mexico)"] = "Este Pais",
  ["este país (mexico city, mexico)"] = "Este Pais",
  ["estodont press"] = "Estodont Press",
  ["estomatologia e cultura"] = "Estomatol Cult",
  ["estonian academy of sciences"] = "Proc. Estonian Acad. Sci. Phys. Math.",
  ["estonian journal of archaeology"] = "Est. J. Archaeol.",
  ["estonian journal of earth sciences"] = "Est. J. Earth Sci.",
  ["estonian journal of engineering"] = "Est. J. Eng.",
  ["estuaries and coasts"] = "Estuaries Coasts",
  ["estuaries and coasts : journal of the estuarine research federation"] = "Estuaries Coast",
  ["estuarine and coastal marine science"] = "Estuarine Coastal Mar. Sci.",
  ["estuarine, coastal and shelf science"] = "Estuarine Coastal Shelf Sci.",
  ["estudios (madrid, spain)"] = "Estudios",
  ["estudios clasicos"] = "Estud Cl",
  ["estudios clásicos"] = "Estud Cl",
  ["estudios clásicos: organo de la sociedad espanola de estudios clásicos"] = "EClás",
  ["estudios de asia y africa"] = "Estud Asia Afr",
  ["estudios de cultura nahuatl"] = "Estud Cult Nahuatl",
  ["estudios de cultura náhuatl"] = "Estud Cult Nahuatl",
  ["estudios de economia"] = "Estud. de Econ.",
  ["estudios de economía"] = "Estud Econ",
  ["estudios de historia de la medicina vasca"] = "Estud Hist Med Vaca",
  ["estudios de historia novohispana"] = "Estud Hist Novohisp",
  ["estudios de historia social (madrid, spain)"] = "Estud Hist Soc",
  ["estudios de historia social y economica de america"] = "Estud Hist Soc Econ Am",
  ["estudios de historia social y económica de américa"] = "Estud Hist Soc Econ Am",
  ["estudios de la antigüedad"] = "EstAnt",
  ["estudios de la cátedra de historia de la farmacia y legislación farmacéutica de la universidad de granada"] = "Estud Catedra Hist Farm Legis Farm Univ Granada",
  ["estudios de poblacion"] = "Estud Poblac",
  ["estudios de población"] = "Estud Poblac",
  ["estudios de prehistoria y arqueología madrileñas"] = "EstMadr",
  ["estudios de psicología"] = "Estud Psicol",
  ["estudios del departamento de historia moderna. universidad de zaragoza. departamento de historia moderna"] = "Estud Dep Hist Mod",
  ["estudios del seminario de prehistoria, arqueología e historia antigua de la facultad de filosofía y letras de zaragoza"] = "EstZaragoza",
  ["estudios demograficos y urbanos"] = "Estud Demogr Urbanos Col Mex",
  ["estudios demográficos y urbanos"] = "Estud Demogr Urbanos Col Mex",
  ["estudios económicos (méxico city, mexico)"] = "Estud Econ (Mex)",
  ["estudios económicos"] = "Estud. Econ.",
  ["estudios empresariales"] = "Estud. Empresariales",
  ["estudios geograficos"] = "Estud Geogr",
  ["estudios geográficos"] = "Estud Geogr",
  ["estudios humanisticos"] = "EHum",
  ["estudios internacionales"] = "Estud. Int.",
  ["estudios latinoamericanos (mexico city, mexico)"] = "Estud Latinoam",
  ["estudios migratorios"] = "Estud Migr",
  ["estudios migratorios latinoamericanos"] = "Estud Migr Latinoam",
  ["estudios paraguayos"] = "Estud Parag",
  ["estudios publicos / centro de estudios publicos"] = "Estud Publicos",
  ["estudios públicos"] = "Estud Publicos",
  ["estudios rurales latinoamericanos"] = "Estud Rurales Latinoam",
  ["estudios segovianos"] = "Estud Segov",
  ["estudios sociales (santo domingo, dominican republic)"] = "Estud Soc",
  ["estudios sociales centroamericanos"] = "Estud Soc Centroam",
  ["estudios sociologicos (mexico city, mexico)"] = "Estud Sociol",
  ["estudios sociológicos (mexico city, mexico)"] = "Estud Sociol",
  ["estudis : revista de historia moderna"] = "Estudis",
  ["estudis historics i documents dels arxius de protocols / col.legi notarial de barcelona"] = "Estud Hist Doc Arxius Protoc",
  ["estudis històrics i documents dels arxius de protocols"] = "Estud Hist Doc Arxius Protoc",
  ["estudis romanics"] = "Estud Rom (Barcelona)",
  ["estudis romànics"] = "Estud Rom (Barcelona)",
  ["estudos cebrap. centro brasileiro de analise e planejamento"] = "Estud CEBRAP",
  ["estudos cebrap. centro brasileiro de análise e planejamento"] = "Estud CEBRAP",
  ["estudos de historia e saude"] = "Estud Hist Saude",
  ["estudos de história e saúde"] = "Estud Hist Saude",
  ["estudos demograficos : revista do gabinete de estudos demograficos do instituto nacional de estatistica"] = "Estud Demogr",
  ["estudos demográficos : revista do gabinete de estudos demográficos do instituto nacional de estatística"] = "Estud Demogr",
  ["estudos economicos"] = "Estudos Econ.",
  ["estudos econômicos"] = "Estud Econ (Sao Paulo)",
  ["estudos feministas"] = "Estud Fem",
  ["estudos históricos (rio de janeiro, brazil)"] = "Estud Hist",
  ["estudos psíquicos"] = "Estud Psiq",
  ["et journal"] = "ET J",
  ["etc.; a review of general semantics"] = "Etc",
  ["etfrn (european tropical forest research network) newsletter"] = "ETFRN Newsletter",
  ["ethical currents : a publication of the center for healthcare ethics, st. joseph health system, and california association of catholic hospitals"] = "Ethical Currents",
  ["ethical human psychology and psychiatry"] = "Ethical Hum Psychol Psychiatry",
  ["ethical human sciences and services : an international journal of critical inquiry"] = "Ethical Hum Sci Serv",
  ["ethical perspectives"] = "Ethical Perspect",
  ["ethical perspectives / catholic university of leuven ; european centre for christian ethics"] = "Ethical Perspect",
  ["ethical theory and moral practice : an international forum"] = "Ethical Theory Moral Pract",
  ["ethics & behavior"] = "Ethics Behav",
  ["ethics & human research"] = "Ethics Hum Res",
  ["ethics & international affairs"] = "Ethics Int Aff",
  ["ethics & medicine : a christian perspective on issues in bioethics"] = "Ethics Med",
  ["ethics & the environment"] = "Ethics Environ.",
  ["ethics and information technology"] = "Ethics Inf Technol",
  ["ethics and intellectual disability"] = "Ethics Intellect Disabil",
  ["ethics and medics"] = "Ethics Medics",
  ["ethics and social welfare"] = "Ethics Soc Welf",
  ["ethics committee newsletter"] = "Ethics Comm Newsl",
  ["ethics in biology, engineering & medicine"] = "Ethics Biol Eng Med",
  ["ethics in biology, engineering and medicine: an international journal"] = "Ethics Biol. Eng. Med.: Int. J.",
  ["ethics in science & medicine"] = "Ethics Sci Med",
  ["ethics in science and environmental politics"] = "Ethics Sci Environ Polit",
  ["ethics in science and medicine"] = "Ethics Sci. Med.",
  ["ethics, law, and aging review"] = "Ethics Law Aging Rev",
  ["ethics, medicine, and public health"] = "Ethics Med Public Health",
  ["ethics, policy & environment"] = "Ethics Policy Environ",
  ["ethik in der medizin"] = "Ethik Med.",
  ["ethik in der medizin : organ der akademie fur ethik in der medizin"] = "Ethik Med",
  ["ethik in der medizin : organ der akademie für ethik in der medizin"] = "Ethik Med",
  ["ethiopia observer"] = "Ethiop Obs",
  ["ethiopian journal of development research"] = "Ethiop J Dev Res",
  ["ethiopian journal of health sciences"] = "Ethiop J Health Sci",
  ["ethiopian journal of public health and nutrition"] = "Ethiop J Public Health Nutr",
  ["ethiopian medical journal"] = "Ethiop Med J",
  ["ethnic and racial studies"] = "Ethn Racial Stud",
  ["ethnic forum : bulletin of ethnic studies and ethnic bibliography"] = "Ethn Forum",
  ["ethnic groups"] = "Ethnic Groups",
  ["ethnicity & disease"] = "Ethn Dis",
  ["ethnicity & health"] = "Ethn Health",
  ["ethnicity and disease"] = "Ethn. Dis.",
  ["ethnicity and health"] = "Ethn. Health",
  ["ethnicity and inequalities in health and social care"] = "Ethn Inequal Health Soc Care",
  ["ethnographisch-archäologische zeitschrift"] = "EAZ",
  ["ethnohistory (columbus, ohio)"] = "Ethnohistory",
  ["ethnologia europaea"] = "Ethnol Eur",
  ["ethnologia scandinavica"] = "Ethnol Scand",
  ["ethnologie francaise"] = "Ethnol Fr",
  ["ethnologie française"] = "Ethnol Fr",
  ["ethnos. revista do instituto português de arqueología, história e etnografia"] = "Ethnos",
  ["ethology : formerly zeitschrift fur tierpsychologie"] = "Ethology",
  ["ethology : formerly zeitschrift für tierpsychologie"] = "Ethology",
  ["ethology and sociobiology"] = "Ethol Sociobiol",
  ["ethology ecology & evolution"] = "Ethol. Ecol. Evol.",
  ["ethology, ecology & evolution"] = "Ethol Ecol Evol",
  ["ethos (berkeley, calif.)"] = "Ethos",
  ["etnografia polska"] = "Etnogr Pol",
  ["etnográfica : revista do centro de estudos de antropologia social"] = "Etnografica",
  ["etnologiska studier"] = "Etnol Stud",
  ["etri journal"] = "ETRI J.",
  ["etruscan studies. journal of the etruscan foundation"] = "EtrSt",
  ["etruscans. bulletin of the etruscan foundation"] = "Etruscans",
  ["ettore majorana international science series: physical sciences"] = "Ettore Majorana Int. Sci. Ser.: Phys. Sci.",
  ["etude de la population africaine = african population studies"] = "Etude Popul Afr",
  ["etude et gestion des sols"] = "Etud. Gest. Sols",
  ["etudes & documents (groupe de demographie africaine (france))"] = "Etud Doc Groupe Demogr Afr",
  ["etudes & documents (groupe de démographie africaine (france))"] = "Etud Doc Groupe Demogr Afr",
  ["etudes & expansion"] = "Etud Expans",
  ["etudes anglaises (periodical)"] = "Etud Angl",
  ["etudes balkaniques"] = "Etud Balk",
  ["etudes balkaniques / academie bulgare des sciences, institut d’etudes balkaniques"] = "Etud Balk",
  ["etudes canadiennes. canadian studies"] = "Etud Can",
  ["etudes classiques (namur, belgium)"] = "Etud Class",
  ["etudes d'histoire de la médecine"] = "Etud Hist Med",
  ["etudes d’histoire africaine"] = "Etud Hist Afr",
  ["etudes et conjoncture"] = "Etud. Conjoncture",
  ["etudes françaises (montréal, québec)"] = "Etud Fr",
  ["etudes freudiennes"] = "Etudes Freud",
  ["etudes germaniques"] = "Etud Ger",
  ["etudes internationales"] = "Etud Int",
  ["etudes inuit. inuit studies"] = "Etud Inuit",
  ["etudes irlandaises"] = "Etud Irl",
  ["etudes limousines"] = "Etud Limousines",
  ["etudes maliennes"] = "Etudes Mali",
  ["etudes napoleoniennes : bulletin historique de la societe de sauvegarde du chateau imperial de pont-de-briques"] = "Etud Napoleon",
  ["etudes napoléoniennes : bulletin historique de la société de sauvegarde du château impérial de pont-de-briques"] = "Etud Napoleon",
  ["etudes neo-natales. neo-natal studies"] = "Etudes Neonatales",
  ["etudes néo-natales. neo-natal studies"] = "Etudes Neonatales",
  ["etudes polemologiques"] = "Etud Polemol",
  ["etudes polémologiques"] = "Etud Polemol",
  ["etudes rurales"] = "Etud Rurales",
  ["etudes rwandaises"] = "Etudes Rwandaises",
  ["etudes statistiques"] = "Etudes Stat Inst Natl Stat",
  ["etudes sur les soins et le service infirmier"] = "Etud Soins Serv Infirm",
  ["etudes togolaises de population"] = "Etud Togol Popul",
  ["etudes vauclusiennes"] = "Etud Vauclus",
  ["etudes zairoises"] = "Etud Zair",
  ["etudes zaïroises"] = "Etud Zair",
  ["eubios journal of asian and international bioethics : ejaib"] = "Eubios J Asian Int Bioeth",
  ["eugenics quarterly"] = "Eugen Q",
  ["eugenics review"] = "Eugen. Rev.",
  ["eugenics society symposia"] = "Eugen Soc Symp",
  ["eukaryotic cell"] = "Eukaryot Cell",
  ["eunomia. ephemeridis listy filologické supplementum"] = "Eunomia",
  ["eupa open proteomics"] = "EuPA Open Proteom",
  ["euphrosyne. revista de filologia clássica"] = "Euphrosyne",
  ["euphytica: netherlands journal of plant breeding"] = "Euphytica",
  ["eurasia antiqua"] = "EurAnt",
  ["eurasia septentrionalis antiqua"] = "ESA",
  ["eurasian chemico-technological journal"] = "Eurasian Chem. Technol. J.",
  ["eurasian geography and economics"] = "Eurasian Geogr Econ",
  ["eurasian journal of analytical chemistry"] = "Eurasian J. Anal. Chem.",
  ["eurasian journal of biosciences"] = "Eurasian J Biosci",
  ["eurasian journal of medicine and oncology"] = "Eurasian J Med Oncol",
  ["eurasian mathematical journal"] = "Eurasian Math. J.",
  ["eurasian soil science"] = "Eurasian Soil Sci.",
  ["eurasip journal on advances in signal processing"] = "EURASIP J Adv Signal Process",
  ["eurasip journal on applied signal processing"] = "EURASIP J Appl Signal Processing",
  ["eurasip journal on audio, speech, and music processing"] = "EURASIP J Audio Speech Music Process",
  ["eurasip journal on bioinformatics & systems biology"] = "EURASIP J Bioinform Syst Biol",
  ["eurasip journal on embedded systems"] = "EURASIP J. Embedded Syst.",
  ["eurasip journal on image and video processing"] = "EURASIP J Image Video Process",
  ["eurasip journal on information security"] = "EURASIP J Inf Secur",
  ["eurasip journal on wireless communications and networking"] = "EURASIP J Wirel Commun Netw",
  ["eure. revista latinoamericana de estudios urbano regionales"] = "EURE",
  ["euro advanced tutorials on operational research"] = "EURO Adv. Tutor. Oper. Res.",
  ["euro journal on computational optimization"] = "EURO J. Comput. Optim.",
  ["euro journal on decision processes"] = "EURO J. Decis. Process.",
  ["euro journal on transportation and logistics"] = "EURO J. Transp. Logist.",
  ["euro reports and studies"] = "EURO Rep Stud",
  ["euro surveillance : bulletin européen sur les maladies transmissibles = european communicable disease bulletin"] = "Euro Surveill",
  ["euro-mediterranean journal for environmental integration"] = "EuroMediterr J Environ Integr",
  ["euroasian entomological journal"] = "Eurasian Entomol. J.",
  ["euroasian journal of hepato-gastroenterology"] = "Euroasian J Hepatogastroenterol",
  ["euroasian mining"] = "Eurasian Min.",
  ["eurocourses. computer and information science"] = "Eurocourses Comput. Inform. Sci.",
  ["eurographics symposium on parallel graphics and visualization : eg pgv : [proceedings]. eurographics symposium on parallel graphics and visualization"] = "Eurographics Symp Parallel Graph Vis",
  ["eurographics workshop on visual computing for biomedicine"] = "Eurographics Workshop Vis Comput Biomed",
  ["eurographics/ieee vgtc symposium on visualization : eurovis : [proceedings]. eurographics/ieee vgtc symposium on visualization"] = "Eurograph IEEE VGTC Symp Vis",
  ["eurointervention : journal of europcr in collaboration with the working group on interventional cardiology of the european society of cardiology"] = "EuroIntervention",
  ["euromath bulletin"] = "Euromath Bull.",
  ["europa ethnica"] = "Eur Ethn",
  ["europa medica (edicion espanola)"] = "Eur. Med. (Paris)",
  ["europa medica. [edicion espanola]"] = "Eur Med (Paris)",
  ["europa medica. [edición española]"] = "Eur Med (Paris)",
  ["europa medicophysica"] = "Eura Medicophys",
  ["europa regional"] = "Eur Reg",
  ["europace : european pacing, arrhythmias, and cardiac electrophysiology : journal of the working groups on cardiac pacing, arrhythmias, and cardiac cellular electrophysiology of the european society of cardiology"] = "Europace",
  ["europaea memoria. reihe i. studien"] = "Eur. Mem. Reihe I Stud.",
  ["europe's journal of psychology"] = "Eur J Psychol",
  ["europe, outremer"] = "Eur Om",
  ["europe-asia studies"] = "Eur Asia Stud",
  ["europe; revue litteraire mensuelle"] = "Europe",
  ["europe; revue littéraire mensuelle"] = "Europe",
  ["european academic research"] = "Eur Acad Res",
  ["european actuarial academy (eaa) series"] = "EAA Ser.",
  ["european actuarial journal"] = "Eur Actuar J",
  ["european addiction research"] = "Eur Addict Res",
  ["european annals of allergy and clinical immunology"] = "Eur Ann Allergy Clin Immunol",
  ["european annals of otorhinolaryngology, head and neck diseases"] = "Eur Ann Otorhinolaryngol Head Neck Dis",
  ["european archives of oto-rhino-laryngology"] = "Eur. Arch. Otorhinolaryngol.",
  ["european archives of oto-rhino-laryngology : official journal of the european federation of oto-rhino-laryngological societies (eufos) : affiliated with the german society for oto-rhino-laryngology - head and neck surgery"] = "Eur Arch Otorhinolaryngol",
  ["european archives of oto-rhino-laryngology. supplement"] = "Eur Arch Otorhinolaryngol Suppl",
  ["european archives of paediatric dentistry"] = "Eur. Arch. Paediatr. Dent.",
  ["european archives of paediatric dentistry : official journal of the european academy of paediatric dentistry"] = "Eur Arch Paediatr Dent",
  ["european archives of psychiatry and clinical neuroscience"] = "Eur Arch Psychiatry Clin Neurosci",
  ["european archives of psychiatry and neurological sciences"] = "Eur Arch Psychiatry Neurol Sci",
  ["european biophysics journal"] = "Eur. Biophys. J.",
  ["european biophysics journal : ebj"] = "Eur Biophys J",
  ["european cardiology"] = "Eur Cardiol",
  ["european cells & materials"] = "Eur Cell Mater",
  ["european cells and materials"] = "Eur. Cells Mater.",
  ["european chemical bulletin"] = "Eur. Chem. Bull.",
  ["european child & adolescent psychiatry"] = "Eur Child Adolesc Psychiatry",
  ["european child and adolescent psychiatry"] = "Eur. Child Adolesc. Psychiatry",
  ["european clinical respiratory journal"] = "Eur Clin Respir J",
  ["european consortium for mathematics in industry"] = "European Consort. Math. Indust.",
  ["european consortium for mathematics in industry (berlin)"] = "Eur. Consort. Math. Ind. (Berl.)",
  ["european cytokine network"] = "Eur Cytokine Netw",
  ["european demographic information bulletin"] = "Europ Demogr Inf Bull",
  ["european demographic information bulletin / european centre for population studies"] = "Europ Demogr Inf Bull",
  ["european eating disorders review : the journal of the eating disorders association"] = "Eur Eat Disord Rev",
  ["european economic review"] = "Eur Econ Rev",
  ["european endocrinology"] = "Eur Endocrinol",
  ["european endodontic journal"] = "Eur Endod J",
  ["european finance review"] = "Europ. Finance Rev.",
  ["european financial management"] = "Europ. Finan. Manage.",
  ["european food research and technology"] = "Eur. Food Res. Technol.",
  ["european food research and technology = zeitschrift für lebensmittel-untersuchung und -forschung. a"] = "Eur Food Res Technol",
  ["european foreign affairs review"] = "Eur Foreign Aff Rev",
  ["european gastroenterology & hepatology review"] = "Eur Gastroenterol Hepatol Rev",
  ["european geriatric medicine"] = "Eur Geriatr Med",
  ["european heart journal"] = "Eur Heart J",
  ["european heart journal - acute cardiovascular care"] = "Eur. Heart J. Acute Cardiovasc. Care",
  ["european heart journal - cardiovascular imaging"] = "Eur. Heart J. Cardiovasc. Imaging",
  ["european heart journal - cardiovascular pharmacotherapy"] = "Eur. Heart J. Cardiovasc. Pharmacother.",
  ["european heart journal - case reports"] = "Eur. Heart J. Case Rep.",
  ["european heart journal - digital health"] = "Eur. Heart J. Digital Health",
  ["european heart journal - imaging methods and practice"] = "Eur. Heart J. Imaging Methods Pract.",
  ["european heart journal - quality of care and clinical outcomes"] = "Eur. Heart J. Qual. Care Clin. Outcomes",
  ["european heart journal cardiovascular imaging"] = "Eur Heart J Cardiovasc Imaging",
  ["european heart journal open"] = "Eur. Heart J. Open",
  ["european heart journal supplements"] = "Eur. Heart J. Suppl.",
  ["european heart journal supplements : journal of the european society of cardiology"] = "Eur Heart J Suppl",
  ["european heart journal. acute cardiovascular care"] = "Eur Heart J Acute Cardiovasc Care",
  ["european heart journal. cardiovascular pharmacotherapy"] = "Eur Heart J Cardiovasc Pharmacother",
  ["european heart journal. case reports"] = "Eur Heart J Case Rep",
  ["european heart journal. quality of care & clinical outcomes"] = "Eur Heart J Qual Care Clin Outcomes",
  ["european history quarterly"] = "Eur Hist Q",
  ["european human rights law review"] = "Eur Hum Rights Law Rev",
  ["european infectious disease"] = "Eur Infect Dis",
  ["european intellectual property review"] = "Eur Intellect Prop Rev",
  ["european journal for biomedical informatics"] = "Eur J Biomed Inform (Praha)",
  ["european journal for person centered healthcare"] = "Eur J Pers Cent Healthc",
  ["european journal for philosophy of science"] = "Eur. J. Philos. Sci.",
  ["european journal for security research"] = "Eur. J. Secur. Res.",
  ["european journal for sport and society"] = "Eur. J. Sport Soc.",
  ["european journal of advanced research in biological and life sciences"] = "Eur J Adv Res Biol Life Sci",
  ["european journal of ageing"] = "Eur J Ageing",
  ["european journal of agronomy"] = "Eur. J. Agron.",
  ["european journal of agronomy : the journal of the european society for agronomy"] = "Eur J Agron",
  ["european journal of anaesthesiology"] = "Eur J Anaesthesiol",
  ["european journal of anaesthesiology. supplement"] = "Eur J Anaesthesiol Suppl",
  ["european journal of anatomy : official journal of the spanish society of anatomy"] = "Eur J Anat",
  ["european journal of applied mathematics"] = "Eur. J. Appl. Math.",
  ["european journal of applied microbiology"] = "Eur. J. Appl. Microbiol.",
  ["european journal of applied microbiology and biotechnology"] = "Eur. J. Appl. Microbiol. Biotechnol.",
  ["european journal of applied physiology"] = "Eur J Appl Physiol",
  ["european journal of applied physiology and occupational physiology"] = "Eur J Appl Physiol Occup Physiol",
  ["european journal of archaeology"] = "Eur J Archaeol",
  ["european journal of arrhythmia & electrophysiology"] = "Eur J Arrhythm Electrophysiol",
  ["european journal of basic and applied histochemistry"] = "Eur J Basic Appl Histochem",
  ["european journal of behavior analysis"] = "Eur J Behav Anal",
  ["european journal of biochemistry"] = "Eur. J. Biochem.",
  ["european journal of biochemistry / febs"] = "Eur J Biochem",
  ["european journal of biological research"] = "Eur J Biol Res",
  ["european journal of biological sciences"] = "Eur J Biol Sci",
  ["european journal of breast health"] = "Eur J Breast Health",
  ["european journal of cancer"] = "Eur J Cancer",
  ["european journal of cancer & clinical oncology"] = "Eur J Cancer Clin Oncol",
  ["european journal of cancer (oxford, england : 1990)"] = "Eur J Cancer",
  ["european journal of cancer and clinical oncology"] = "Eur. J. Cancer Clin. Oncol.",
  ["european journal of cancer care"] = "Eur J Cancer Care (Engl)",
  ["european journal of cancer care (english language edition)"] = "Eur. J. Cancer Care (Engl.)",
  ["european journal of cancer prevention"] = "Eur. J. Cancer Prev.",
  ["european journal of cancer prevention : the official journal of the european cancer prevention organisation (ecp)"] = "Eur J Cancer Prev",
  ["european journal of cancer. part b, oral oncology"] = "Eur J Cancer B Oral Oncol",
  ["european journal of cardio-thoracic surgery"] = "Eur. J. Cardiothorac. Surg.",
  ["european journal of cardio-thoracic surgery : official journal of the european association for cardio-thoracic surgery"] = "Eur J Cardiothorac Surg",
  ["european journal of cardiology"] = "Eur J Cardiol",
  ["european journal of cardiovascular nursing"] = "Eur. J. Cardiovasc. Nurs.",
  ["european journal of cardiovascular nursing : journal of the working group on cardiovascular nursing of the european society of cardiology"] = "Eur J Cardiovasc Nurs",
  ["european journal of cardiovascular prevention and rehabilitation"] = "Eur. J. Cardiovasc. Prev. Rehabil.",
  ["european journal of cardiovascular prevention and rehabilitation : official journal of the european society of cardiology, working groups on epidemiology & prevention and cardiac rehabilitation and exercise physiology"] = "Eur J Cardiovasc Prev Rehabil",
  ["european journal of cardiovascular research"] = "Eur j Cardiovasc Res",
  ["european journal of case reports in internal medicine"] = "Eur J Case Rep Intern Med",
  ["european journal of cell biology"] = "Eur J Cell Biol",
  ["european journal of cell biology. supplement"] = "Eur J Cell Biol Suppl",
  ["european journal of chemistry"] = "Eur. J. Chem.",
  ["european journal of chemistry (print)"] = "Eur J Chem",
  ["european journal of clinical & medical oncology"] = "European J Clin Med Oncol",
  ["european journal of clinical chemistry and clinical biochemistry"] = "Eur. J. Clin. Chem. Clin. Biochem.",
  ["european journal of clinical chemistry and clinical biochemistry : journal of the forum of european clinical chemistry societies"] = "Eur J Clin Chem Clin Biochem",
  ["european journal of clinical investigation"] = "Eur J Clin Invest",
  ["european journal of clinical microbiology"] = "Eur J Clin Microbiol",
  ["european journal of clinical microbiology & infectious diseases"] = "Eur. J. Clin. Microbiol. Infect. Dis.",
  ["european journal of clinical microbiology & infectious diseases : official publication of the european society of clinical microbiology"] = "Eur J Clin Microbiol Infect Dis",
  ["european journal of clinical microbiology and infectious diseases"] = "Eur. J. Clin. Microbiol. Infect. Dis.",
  ["european journal of clinical nutrition"] = "Eur J Clin Nutr",
  ["european journal of clinical pharmacology"] = "Eur J Clin Pharmacol",
  ["european journal of combinatorics"] = "European J. Combin.",
  ["european journal of combinatorics = journal européen de combinatoire = europäische zeitschrift für kombinatorik"] = "Eur J Comb",
  ["european journal of communication"] = "Eur J Commun",
  ["european journal of computational mechanics"] = "Eur. J. Comput. Mech.",
  ["european journal of contraception and reproductive health care"] = "Eur. J. Contracept. Reprod. Health Care",
  ["european journal of control"] = "Eur. J. Control",
  ["european journal of criminology"] = "Eur J Criminol",
  ["european journal of cultural and political sociology"] = "Eur J Cult Polit Sociol",
  ["european journal of dental education"] = "Eur. J. Dent. Educ.",
  ["european journal of dental education : official journal of the association for dental education in europe"] = "Eur J Dent Educ",
  ["european journal of dentistry"] = "Eur J Dent",
  ["european journal of dermatology"] = "Eur. J. Dermatol.",
  ["european journal of dermatology : ejd"] = "Eur J Dermatol",
  ["european journal of development research"] = "Europ. J. Devel. Res.",
  ["european journal of developmental science"] = "Eur J Dev Sci",
  ["european journal of disorders of communication"] = "Eur. J. Disord. Commun.",
  ["european journal of disorders of communication : the journal of the college of speech and language therapists, london"] = "Eur J Disord Commun",
  ["european journal of drug metabolism and pharmacokinetics"] = "Eur J Drug Metab Pharmacokinet",
  ["european journal of east asian studies"] = "Eur J East Asian Stud",
  ["european journal of echocardiography"] = "Eur. J. Echocardiogr.",
  ["european journal of echocardiography : the journal of the working group on echocardiography of the european society of cardiology"] = "Eur J Echocardiogr",
  ["european journal of ecology"] = "Eur J Ecol",
  ["european journal of education"] = "Eur J Educ",
  ["european journal of emergency medicine"] = "Eur. J. Emerg. Med.",
  ["european journal of emergency medicine : official journal of the european society for emergency medicine"] = "Eur J Emerg Med",
  ["european journal of endocrinology"] = "Eur J Endocrinol",
  ["european journal of endocrinology / european federation of endocrine societies"] = "Eur J Endocrinol",
  ["european journal of engineering educatino"] = "Eur. J. Eng. Educ.",
  ["european journal of engineering education"] = "Eur. J. Eng. Educ.",
  ["european journal of english studies"] = "Eur J Engl Stud",
  ["european journal of entomology"] = "Eur. J. Entomol.",
  ["european journal of environment and public health"] = "Eur J Environ Public Health",
  ["european journal of environmental and civil engineering"] = "Eur. J. Environ. Civ. Eng.",
  ["european journal of epidemiology"] = "Eur J Epidemiol",
  ["european journal of experimental biology"] = "Eur J Exp Biol",
  ["european journal of finance"] = "Europ. J. Finance",
  ["european journal of forest pathology"] = "Eur. J. For. Pathol.",
  ["european journal of forest pathology. journal européen de pathologie forestière. europäische zeitschrift für forstpathologie"] = "Eur J Forest Pathol",
  ["european journal of forest research"] = "Eur J For Res",
  ["european journal of forestry research"] = "Eur. J. For. Res.",
  ["european journal of futures research"] = "Eur. J. Futures Res.",
  ["european journal of gastroenterology & hepatology"] = "Eur J Gastroenterol Hepatol",
  ["european journal of gastroenterology and hepatology"] = "Eur. J. Gastroenterol. Hepatol.",
  ["european journal of general dentistry"] = "European J Gen Dent",
  ["european journal of genetics in society : an ethical approach to genetics"] = "Eur J Genet Soc",
  ["european journal of gynaecological oncology"] = "Eur J Gynaecol Oncol",
  ["european journal of haematology"] = "Eur J Haematol",
  ["european journal of haematology. supplementum"] = "Eur J Haematol Suppl",
  ["european journal of health economics"] = "Eur. J. Health Econ.",
  ["european journal of health law"] = "Eur J Health Law",
  ["european journal of health psychology"] = "Eur J Health Psychol",
  ["european journal of heart failure"] = "Eur J Heart Fail",
  ["european journal of heart failure : journal of the working group on heart failure of the european society of cardiology"] = "Eur J Heart Fail",
  ["european journal of histochemistry"] = "Eur. J. Histochem.",
  ["european journal of histochemistry : ejh"] = "Eur J Histochem",
  ["european journal of homelessness"] = "Eur J Homelessness",
  ["european journal of horticultural science"] = "Eur J Hortic Sci",
  ["european journal of hospital pharmacy"] = "Eur. J. Hosp. Pharm.",
  ["european journal of hospital pharmacy : science and practice"] = "Eur J Hosp Pharm",
  ["european journal of housing policy"] = "Europ. J. Housing Pol.",
  ["european journal of human genetics"] = "Eur. J. Hum. Genet.",
  ["european journal of human genetics : ejhg"] = "Eur J Hum Genet",
  ["european journal of hybrid imaging"] = "Eur J Hybrid Imaging",
  ["european journal of immunogenetics"] = "Eur. J. Immunogenet.",
  ["european journal of immunogenetics : official journal of the british society for histocompatibility and immunogenetics"] = "Eur J Immunogenet",
  ["european journal of immunology"] = "Eur J Immunol",
  ["european journal of industrial engineering"] = "Eur. J. Ind. Eng.",
  ["european journal of industrial relations"] = "Europ. J. Ind. Relat.",
  ["european journal of inflammation"] = "Eur J Inflamm",
  ["european journal of information systems"] = "Eur. J. Inf. Syst.",
  ["european journal of information systems : an official journal of the operational research society"] = "Eur J Inf Syst",
  ["european journal of innovation management"] = "Eur. J. Innovation Manage.",
  ["european journal of inorganic chemistry"] = "Eur. J. Inorg. Chem.",
  ["european journal of integrative medicine"] = "Eur J Integr Med",
  ["european journal of intensive care medicine"] = "Eur J Intensive Care Med",
  ["european journal of internal medicine"] = "Eur J Intern Med",
  ["european journal of international relations"] = "Eur J Int Relat",
  ["european journal of investigation in health, psychology and education"] = "Eur J Investig Health Psychol Educ",
  ["european journal of law and economics"] = "Europ. J. Law Econ.",
  ["european journal of lipid science and technology"] = "Eur. J. Lipid Sci. Technol.",
  ["european journal of lipid science and technology : ejlst"] = "Eur J Lipid Sci Technol",
  ["european journal of marketing"] = "Eur J Mark",
  ["european journal of mass spectrometry"] = "Eur. J. Mass Spectrom.",
  ["european journal of mass spectrometry (chichester, england)"] = "Eur J Mass Spectrom (Chichester)",
  ["european journal of materials"] = "Eur. J. Mater.",
  ["european journal of mathematics"] = "Eur. J. Math.",
  ["european journal of mechanics"] = "Eur. J. Mech. A Solids",
  ["european journal of mechanics - a/solids"] = "Eur. J. Mech. A. Solids",
  ["european journal of mechanics - b/fluids"] = "Eur. J. Mech. B. Fluids",
  ["european journal of mechanics. a, solids"] = "Eur J Mech A Solids",
  ["european journal of mechanics. a. solids"] = "Eur. J. Mech. A Solids",
  ["european journal of mechanics. b, fluids"] = "Eur J Mech B Fluids",
  ["european journal of mechanics. b. fluids"] = "Eur. J. Mech. B Fluids",
  ["european journal of medical genetics"] = "Eur J Med Genet",
  ["european journal of medical research"] = "Eur J Med Res",
  ["european journal of medicinal chemistry"] = "Eur J Med Chem",
  ["european journal of medicinal chemistry reports"] = "Eur. J. Med. Chem. Rep.",
  ["european journal of medicinal plants"] = "European J Med Plants",
  ["european journal of medicine"] = "Eur. J. Med.",
  ["european journal of microbiology & immunology"] = "Eur J Microbiol Immunol (Bp)",
  ["european journal of microbiology and immunology"] = "Eur. J. Microbiol. Immunol.",
  ["european journal of midwifery"] = "Eur J Midwifery",
  ["european journal of mineralogy"] = "Eur. J. Mineral.",
  ["european journal of morphology"] = "Eur J Morphol",
  ["european journal of nanomedicine"] = "Eur J Nanomed",
  ["european journal of neurodegenerative disease"] = "Eur J Neurodegener Dis",
  ["european journal of neurology"] = "Eur J Neurol",
  ["european journal of neurology : the official journal of the european federation of neurological societies"] = "Eur J Neurol",
  ["european journal of neuroscience"] = "Eur. J. Neurosci.",
  ["european journal of nuclear medicine"] = "Eur J Nucl Med",
  ["european journal of nuclear medicine and molecular imaging"] = "Eur J Nucl Med Mol Imaging",
  ["european journal of nutrition"] = "Eur J Nutr",
  ["european journal of nutrition & food safety"] = "Eur J Nutr Food Saf",
  ["european journal of obstetrics & gynecology and reproductive biology: x"] = "Eur J Obstet Gynecol Reprod Biol X",
  ["european journal of obstetrics and gynecology"] = "Eur J Obstet Gynecol",
  ["european journal of obstetrics, gynecology, and reproductive biology"] = "Eur J Obstet Gynecol Reprod Biol",
  ["european journal of oncology nursing : the official journal of european oncology nursing society"] = "Eur J Oncol Nurs",
  ["european journal of oncology pharmacy"] = "Eur. J. Oncol. Pharm.",
  ["european journal of operational research"] = "Eur. J. Oper. Res.",
  ["european journal of ophthalmology"] = "Eur J Ophthalmol",
  ["european journal of oral implantology"] = "Eur J Oral Implantol",
  ["european journal of oral sciences"] = "Eur J Oral Sci",
  ["european journal of organic chemistry"] = "Eur. J. Org. Chem.",
  ["european journal of orthodontics"] = "Eur J Orthod",
  ["european journal of orthopaedic surgery & traumatology : orthopédie traumatologie"] = "Eur J Orthop Surg Traumatol",
  ["european journal of paediatric dentistry"] = "Eur J Paediatr Dent",
  ["european journal of paediatric dentistry : official journal of european academy of paediatric dentistry"] = "Eur J Paediatr Dent",
  ["european journal of paediatric neurology"] = "Eur. J. Paediatr. Neurol.",
  ["european journal of paediatric neurology : ejpn : official journal of the european paediatric neurology society"] = "Eur J Paediatr Neurol",
  ["european journal of pain"] = "Eur. J. Pain",
  ["european journal of pain (london, england)"] = "Eur J Pain",
  ["european journal of pain supplements"] = "Eur J Pain Suppl",
  ["european journal of palliative care : the journal of the european association for palliative care"] = "Eur J Palliat Care",
  ["european journal of pediatric surgery"] = "Eur. J. Pediatr. Surg.",
  ["european journal of pediatric surgery reports"] = "European J Pediatr Surg Rep",
  ["european journal of pediatrics"] = "Eur J Pediatr",
  ["european journal of personality"] = "Eur J Pers",
  ["european journal of pharmaceutical sciences"] = "Eur. J. Pharm. Sci.",
  ["european journal of pharmaceutical sciences : official journal of the european federation for pharmaceutical sciences"] = "Eur J Pharm Sci",
  ["european journal of pharmaceutics and biopharmaceutics"] = "Eur. J. Pharm. Biopharm.",
  ["european journal of pharmaceutics and biopharmaceutics : official journal of arbeitsgemeinschaft fur pharmazeutische verfahrenstechnik e.v"] = "Eur J Pharm Biopharm",
  ["european journal of pharmaceutics and biopharmaceutics : official journal of arbeitsgemeinschaft für pharmazeutische verfahrenstechnik e.v"] = "Eur J Pharm Biopharm",
  ["european journal of pharmacology"] = "Eur J Pharmacol",
  ["european journal of pharmacology, molecular pharmacology section"] = "Eur. J. Pharmacol., Mol. Pharmacol. Sect.",
  ["european journal of pharmacology: environmental toxicology and pharmacology"] = "Eur. J. Pharmacol. Environ. Toxicol. Pharm.",
  ["european journal of pharmacology: molecular pharmacology"] = "Eur. J. Pharmacol. Mol. Pharmacol.",
  ["european journal of philosophy"] = "Eur J Philos",
  ["european journal of phycology"] = "Eur. J. Phycol.",
  ["european journal of physical and rehabilitation medicine"] = "Eur J Phys Rehabil Med",
  ["european journal of physics"] = "Eur. J. Phys.",
  ["european journal of physiotherapy"] = "Eur J Physiother",
  ["european journal of plant pathology"] = "Eur J Plant Pathol",
  ["european journal of plastic surgery"] = "Eur J Plast Surg",
  ["european journal of political economy"] = "Eur J Polit Econ",
  ["european journal of political research"] = "Eur J Polit Res",
  ["european journal of political theory : ejpt"] = "Eur J Political Theory",
  ["european journal of population"] = "Eur. J. Popul.",
  ["european journal of population = revue européenne de démographie"] = "Eur J Popul",
  ["european journal of preventive cardiology"] = "Eur J Prev Cardiol",
  ["european journal of prosthodontics and restorative dentistry"] = "Eur. J. Prosthodont. Restor. Dent.",
  ["european journal of protistology"] = "Eur J Protistol",
  ["european journal of psychological assessment : official organ of the european association of psychological assessment"] = "Eur J Psychol Assess",
  ["european journal of psychotherapy & counselling"] = "Eur J Psychother Couns",
  ["european journal of psychotraumatology"] = "Eur J Psychotraumatol",
  ["european journal of public health"] = "Eur J Public Health",
  ["european journal of pure and applied mathematics"] = "Eur. J. Pure Appl. Math.",
  ["european journal of radiology"] = "Eur J Radiol",
  ["european journal of radiology open"] = "Eur J Radiol Open",
  ["european journal of remote sensing"] = "Eur J Remote Sens",
  ["european journal of respiratory diseases"] = "Eur J Respir Dis",
  ["european journal of respiratory diseases. supplement"] = "Eur J Respir Dis Suppl",
  ["european journal of rheumatology"] = "Eur J Rheumatol",
  ["european journal of rheumatology and inflammation"] = "Eur J Rheumatol Inflamm",
  ["european journal of risk regulation"] = "Eur. J. Risk Regul.",
  ["european journal of risk regulation : ejrr"] = "Eur J Risk Regul",
  ["european journal of scientific research = revue européenne de la recherche scientifique"] = "Eur J Sci Res",
  ["european journal of sexually transmitted diseases"] = "Eur J Sex Transm Dis",
  ["european journal of social psychology"] = "Eur J Soc Psychol",
  ["european journal of social security"] = "Eur J Soc Secur",
  ["european journal of soil biology"] = "Eur. J. Soil Biol.",
  ["european journal of soil science"] = "Eur. J. Soil Sci.",
  ["european journal of solid state and inorganic chemistry"] = "Eur. J. Solid State Inorg. Chem.",
  ["european journal of special needs education"] = "Eur J Spec Needs Educ",
  ["european journal of sport science"] = "Eur J Sport Sci",
  ["european journal of sports & exercise science"] = "Eur J Sports Exerc Sci",
  ["european journal of surgery"] = "Eur. J. Surg.",
  ["european journal of surgery. supplement"] = "Eur. J. Surg. Suppl.",
  ["european journal of surgical oncology"] = "Eur. J. Surg. Oncol.",
  ["european journal of surgical oncology : the journal of the european society of surgical oncology and the british association of surgical oncology"] = "Eur J Surg Oncol",
  ["european journal of taxonomy"] = "Eur J Taxon",
  ["european journal of the history of economic thought"] = "Europ. J. Hist. Econ. Thought",
  ["european journal of toxicology"] = "Eur J Toxicol",
  ["european journal of toxicology and environmental hygiene"] = "Eur. J. Toxicol. Environ. Hyg.",
  ["european journal of toxicology and environmental hygiene. journal europeen de toxicologie"] = "Eur J Toxicol Environ Hyg",
  ["european journal of toxicology and environmental hygiene. journal européen de toxicologie"] = "Eur J Toxicol Environ Hyg",
  ["european journal of toxicology and hygiene of environment"] = "Eur J Toxicol Hyg Environ",
  ["european journal of translational and clinical medicine"] = "Eur J Transl Clin Med",
  ["european journal of translational myology"] = "Eur J Transl Myol",
  ["european journal of trauma and emergency surgery : official publication of the european trauma society"] = "Eur J Trauma Emerg Surg",
  ["european journal of ultrasound"] = "Eur. J. Ultrasound",
  ["european journal of ultrasound : official journal of the european federation of societies for ultrasound in medicine and biology"] = "Eur J Ultrasound",
  ["european journal of vascular and endovascular surgery"] = "Eur. J. Vasc. Endovasc. Surg.",
  ["european journal of vascular and endovascular surgery : the official journal of the european society for vascular surgery"] = "Eur J Vasc Endovasc Surg",
  ["european journal of vascular surgery"] = "Eur J Vasc Surg",
  ["european journal of wildlife research"] = "Eur J Wildl Res",
  ["european journal of women's studies"] = "Eur J Womens Stud",
  ["european journal of wood and wood products"] = "Eur. J. Wood Wood Prod.",
  ["european journal of zoological research"] = "Eur J Zool Res",
  ["european journal on criminal policy and research"] = "Eur J Crim Pol Res",
  ["european journal pharmaceutical and medical research"] = "Eur J Pharm Med Res",
  ["european law review"] = "Eur Law Rev",
  ["european mass spectrometry"] = "Eur. Mass Spectrom.",
  ["european mathematical society magazine"] = "Eur. Math. Soc. Mag.",
  ["european mathematical society. newsletter"] = "Eur. Math. Soc. Newsl.",
  ["european medical journal (chelmsford, england)"] = "Eur Med J (Chelmsf)",
  ["european medical journal. diabetes"] = "Eur Med J Diabetes",
  ["european medical journal. hepatology"] = "Eur Med J Hepatol",
  ["european medical journal. innovations"] = "Euro Med J Innov",
  ["european medical journal. oncology"] = "Eur Med J Oncol",
  ["european medical journal. reproductive health"] = "Eur Med J Reprod Health",
  ["european medical journal. respiratory"] = "Eur Med J Respir",
  ["european medical journal. rheumatology"] = "Eur Med J Rheumatol",
  ["european medical journal. urology"] = "Eur Med J Urol",
  ["european molecular biology organization journal"] = "Eur. Mol. Biol. Organ. J.",
  ["european musculoskeletal review"] = "Eur Musculoskelet Rev",
  ["european neurological journal"] = "Eur Neurol J",
  ["european neurological review"] = "Eur Neurol Rev",
  ["european neurology"] = "Eur Neurol",
  ["european neuropsychopharmacology"] = "Eur. Neuropsychopharmacol.",
  ["european neuropsychopharmacology : the journal of the european college of neuropsychopharmacology"] = "Eur Neuropsychopharmacol",
  ["european oncology"] = "Eur Oncol",
  ["european oncology & haematology"] = "Eur Oncol Haematol",
  ["european oral research"] = "Eur Oral Res",
  ["european orthopaedics and traumatology"] = "Eur Orthop Traumatol",
  ["european paediatrics review"] = "Eur Paediatr Rev",
  ["european physical education review"] = "Eur Phy Educ Rev",
  ["european physical journal"] = "Eur. Phys. J. B",
  ["european physical journal a: hadrons and nuclei"] = "Eur. Phys. J. A",
  ["european physical journal am: applied metamaterials"] = "Eur. Phys. J. AM",
  ["european physical journal ap: applied physics"] = "Eur. Phys. J. AP",
  ["european physical journal applied physics"] = "Eur. Phys. J. Appl. Phys.",
  ["european physical journal b: condensed matter and complex"] = "Eur. Phys. J. B",
  ["european physical journal b: condensed matter and complex systems"] = "Eur. Phys. J. B",
  ["european physical journal b: condensed matter physics"] = "Eur. Phys. J. B",
  ["european physical journal c: particles and fields"] = "Eur. Phys. J. C",
  ["european physical journal d: atomic, molecular and optical physics"] = "Eur. Phys. J. D",
  ["european physical journal d: atomic, molecular, optical and plasma physics"] = "Eur. Phys. J. D",
  ["european physical journal ds: data science"] = "Eur. Phys. J. DS",
  ["european physical journal e: soft matter"] = "Eur. Phys. J. E",
  ["european physical journal e: soft matter and biological physics"] = "Eur. Phys. J. E: Soft Matter Biol. Phys.",
  ["european physical journal h: historical perspectives on contemporary physics"] = "Eur. Phys. J. H",
  ["european physical journal n: nuclear sciences and technologies"] = "Eur. Phys. J. N",
  ["european physical journal nbp: nonlinear biomedical physics"] = "Eur. Phys. J. NBP",
  ["european physical journal plus"] = "Eur. Phys. J. Plus",
  ["european physical journal pv: photovoltaics"] = "Eur. Phys. J. PV",
  ["european physical journal qt: quantum technology"] = "Eur. Phys. J. QT",
  ["european physical journal special topics"] = "Eur. Phys. J. Spec. Top.",
  ["european physical journal st: special topics"] = "Eur. Phys. J. ST",
  ["european physical journal ti: techniques and instrumentation"] = "Eur. Phys. J. TI",
  ["european physical journal woc: web of conferences"] = "Eur. Phys. J. WOC",
  ["european physics journal b: condensed matter"] = "Eur. Phys. J. B",
  ["european physics journal c: particles and fields"] = "Eur. Phys. J. C",
  ["european physics journal d: atomic, molecular, and optical physics"] = "Eur. Phys. J. D",
  ["european physics journal e: soft matter"] = "Eur. Phys. J. E",
  ["european polymer journal"] = "Eur. Polym. J.",
  ["european potato journal"] = "Eur. Potato J.",
  ["european poultry science"] = "Eur. Poult. Sci.",
  ["european psychiatry : the journal of the association of european psychiatrists"] = "Eur Psychiatry",
  ["european psychologist"] = "Eur Psychol",
  ["european quality assurance network newsletter"] = "Eur Qual Assur Netw Newsl",
  ["european quality assurance network newsletter / euroquan"] = "Eur Qual Assur Netw Newsl",
  ["european radiology"] = "Eur Radiol",
  ["european radiology experimental"] = "Eur Radiol Exp",
  ["european radiology supplements"] = "Eur. Radio. Suppl.",
  ["european respiratory journal"] = "Eur. Respir. J.",
  ["european respiratory journal. supplement"] = "Eur. Respir. J. Suppl.",
  ["european respiratory review"] = "Eur. Respir. Rev.",
  ["european respiratory review : an official journal of the european respiratory society"] = "Eur Respir Rev",
  ["european review (chichester, england)"] = "Eur Rev",
  ["european review for medical and pharmacological sciences"] = "Eur Rev Med Pharmacol Sci",
  ["european review of aging and physical activity"] = "Eur. Rev. Aging Phys. Act.",
  ["european review of aging and physical activity : official journal of the european group for research into elderly and physical activity"] = "Eur Rev Aging Phys Act",
  ["european review of agricultural economics"] = "Eur. Rev. Agric. Econ.",
  ["european review of applied psychology = revue européenne de psychologie appliquée"] = "Eur Rev Appl Psychol",
  ["european review of economic history"] = "Eur Rev Econ Hist",
  ["european review of history = revue européene d'histoire"] = "Eur Rev Hist",
  ["european review of history. revue européenne d’histoire"] = "EurRHist",
  ["european review of social psychology"] = "Eur Rev Soc Psychol",
  ["european scientific journal"] = "Eur Sci J",
  ["european series in applied and industrial mathematics"] = "ESAIM Probab. Statist.",
  ["european societies"] = "Eur. Soc.",
  ["european sociological review"] = "Eur Sociol Rev",
  ["european space agency special publication sp"] = "Eur. Space Agency Spec. Publ. SP",
  ["european spine journal"] = "Eur. Spine J.",
  ["european spine journal : official publication of the european spine society, the european spinal deformity society, and the european section of the cervical spine research society"] = "Eur Spine J",
  ["european stroke journal"] = "Eur Stroke J",
  ["european studies in philosophy of science"] = "Eur. Stud. Philos. Sci.",
  ["european studies review"] = "Eur Stud Rev",
  ["european surgery : aca : acta chirurgica austriaca"] = "Eur Surg",
  ["european surgical research"] = "Eur. Surg. Res.",
  ["european surgical research. europaische chirurgische forschung. recherches chirurgicales europeennes"] = "Eur Surg Res",
  ["european surgical research. europäische chirurgische forschung. recherches chirurgicales européennes"] = "Eur Surg Res",
  ["european thyroid journal"] = "Eur Thyroid J",
  ["european transactions on electrical power"] = "Eur. Trans. Electr. Power",
  ["european transactions on telecommunications"] = "Eur. Trans. Telecommun.",
  ["european transport research review"] = "Eur. Transport Res. Rev.",
  ["european union politics"] = "Eur Union Polit",
  ["european university institute. series d: economics"] = "European Univ. Inst. Ser. D: Econom.",
  ["european urban and regional studies"] = "Eur Urban Reg Stud",
  ["european urological review"] = "Eur Urol Rev",
  ["european urology"] = "Eur Urol",
  ["european urology focus"] = "Eur Urol Focus",
  ["european urology oncology"] = "Eur Urol Oncol",
  ["european urology open science"] = "Eur Urol Open Sci",
  ["european water pollution control : official publication of the european water pollution control association (ewpca)"] = "Eur Water Pollut Control",
  ["european zoological journal"] = "Eur. Zool. J.",
  ["europhysics letters"] = "Europhys. Lett.",
  ["europhysics letters (epl)"] = "Europhys. Lett.",
  ["europhysics news"] = "Europhys. News",
  ["europäische hochschulschriften"] = "Europäische Hochschulschrift. Reihe V Volks- Betriebswirtsch.",
  ["europäische hochschulschriften. reihe vii, medizin. abt. b, geschichte der medizin = publications universitaires européennes. série vii, médicine. div. b, histoire de la médecine = european university papers. series vii, health sciences. div. b, history of medicine"] = "Eur Hochschulschr VII Med B Gesch Med",
  ["euscorpius (online)"] = "Euscorpius",
  ["eut report-wsk"] = "EUT Report-WSK",
  ["euthanasia news"] = "Euthanasia News",
  ["eutopia. commentarii novi de antiquitatibus totius europae"] = "Eutopia",
  ["evaluation & the health professions"] = "Eval Health Prof",
  ["evaluation (london, england : 1995)"] = "Evaluation (Lond)",
  ["evaluation and program planning"] = "Eval Program Plann",
  ["evaluation and the health professions"] = "Eval. Health Prof.",
  ["evaluation journal of australasia"] = "Eval J Australas",
  ["evaluation practice"] = "Eval Pract",
  ["evaluation quarterly"] = "Eval Q",
  ["evaluation review"] = "Eval Rev",
  ["evening sun (baltimore, md.)"] = "Evening Sun",
  ["everybody's health"] = "Everybody's Health",
  ["evidence & policy : a journal of research, debate and practice"] = "Evid Policy",
  ["evidence based library and information practice"] = "Evid Based Libr Inf Pract",
  ["evidence for action : challenges for the cochrane collaboration in the 21st century : abstract book. international cochrane colloquium (8th : 2000 : cape town, south africa)"] = "Evid Action Int Cochrane Colloq 8th 2000 Cape Town S Afr",
  ["evidence report/technology assessment"] = "Evid Rep Technol Assess (Full Rep)",
  ["evidence report/technology assessment (summary)"] = "Evid Rep Technol Assess (Summ)",
  ["evidence-based cardiovascular medicine"] = "Evid Based Cardiovasc Med",
  ["evidence-based child health : a cochrane review journal"] = "Evid Based Child Health",
  ["evidence-based communication assessment and intervention"] = "Evid Based Commun Assess Interv",
  ["evidence-based complementary and alternative medicine : ecam"] = "Evid Based Complement Alternat Med",
  ["evidence-based dentistry"] = "Evid Based Dent",
  ["evidence-based integrative medicine"] = "Evidence-Based Integr. Med.",
  ["evidence-based medicine"] = "Evid Based Med",
  ["evidence-based medicine & public health"] = "Evid Based Med Public Health",
  ["evidence-based mental health"] = "Evid Based Ment Health",
  ["evidence-based nursing"] = "Evid Based Nurs",
  ["evidence-based practice in child and adolescent mental health"] = "Evid Based Pract Child Adolesc Ment Health",
  ["evidence-based preclinical medicine"] = "Evid Based Preclin Med",
  ["evidence-based spine-care journal"] = "Evid Based Spine Care J",
  ["evidence-based toxicology"] = "Evidence-Based Toxicol.",
  ["evolution & development"] = "Evol Dev",
  ["evolution and development"] = "Evol. Dev.",
  ["evolution and human behavior : official journal of the human behavior and evolution society"] = "Evol Hum Behav",
  ["evolution equations and control theory"] = "Evol. Equations Control Theory",
  ["evolution letters"] = "Evol Lett",
  ["evolution psychiatrique"] = "Evol. Psychiatr. (Paris)",
  ["evolution, medicine, and public health"] = "Evol Med Public Health",
  ["evolution: education and outreach"] = "Evol.: Educ. Outreach",
  ["evolution; international journal of organic evolution"] = "Evolution",
  ["evolutionary anthropology"] = "Evol Anthropol",
  ["evolutionary anthropology: issues, news, and reviews"] = "Evol. Anthropol.: News Issues Rev.",
  ["evolutionary applications"] = "Evol Appl",
  ["evolutionary behavioral sciences"] = "Evol Behav Sci",
  ["evolutionary bioinformatics"] = "Evol. Bioinf.",
  ["evolutionary bioinformatics online"] = "Evol Bioinform Online",
  ["evolutionary biology"] = "Evol Biol",
  ["evolutionary computation"] = "Evol Comput",
  ["evolutionary computation, machine learning and data mining in bioinformatics. evobio (conference)"] = "Evol Comput Mach Learn Data Min Bioinform",
  ["evolutionary ecology"] = "Evol Ecol",
  ["evolutionary ecology research"] = "Evol Ecol Res",
  ["evolutionary economics and social complexity science"] = "Evol. Econ. Soc. Complex. Sci.",
  ["evolutionary human sciences"] = "Evol Hum Sci",
  ["evolutionary intelligence"] = "Evol Intell",
  ["evolutionary journal of the linnean society"] = "Evol. J. Linn. Soc.",
  ["evolutionary psychological science"] = "Evol Psychol Sci",
  ["evolutionary psychology"] = "Evol. Psychol.",
  ["evolutionary psychology : an international journal of evolutionary approaches to psychology and behavior"] = "Evol Psychol",
  ["evolutionary studies in imaginative culture"] = "Evol Stud Imaginative Cult",
  ["evolutionary systematics"] = "Evol Syst",
  ["evolving systems"] = "Evol. Syst.",
  ["evraziatskiĭ ėntomologicheskiĭ zhurnal = euroasian entomological journal"] = "Evraziat Entomol Z",
  ["evraziĭskiĭ soi︠u︡z uchenykh"] = "Evraziiskii Soiuz Uchenykh",
  ["examples and counterexamples"] = "Examples Counterexamples",
  ["excavaciones arqueológicas en españa"] = "EAE",
  ["excavations and surveys in israel"] = "ExcIsr",
  ["excavations at gözlü kule, tarsus"] = "Tarsus",
  ["excavations at olynthus"] = "Olynthus",
  ["excavations of the athenian agora. picture book"] = "AgoraPB",
  ["exceptional children"] = "Except Child",
  ["excerpta medica"] = "Excerpta Med",
  ["excerpta medica. section 12, ophthalmology"] = "Excerpta Med (Ophthalmol)",
  ["excerpta medica. section 13, dermatology and venereology"] = "Excerpta Med (Dermatol)",
  ["excerpta medica. section 15, chest diseases"] = "Excerpta Medica 15 Chest Dis",
  ["excerpta medica. section 20, gerontology and geriatrics"] = "Excerpta Medica 20 Gerontol Geriatr",
  ["excerpta medica. section 28, urology and nephrology"] = "Excerpta Med (Urol)",
  ["excerpta medica. section 9, surgery"] = "Excerpta Med 9 Surg",
  ["excerpta philologica"] = "ExcPhilol",
  ["excli journal"] = "EXCLI J",
  ["excursions in recreational mathematics series"] = "Excursions Rec. Math. Ser.",
  ["executive housekeeper"] = "Exec Housekeeper",
  ["executive housekeeping today"] = "Exec Housekeep Today",
  ["executive housekeeping today / the official magazine of the national executive housekeepers association"] = "Exec Housekeep Today",
  ["executive intelligence review"] = "Exec Intell Rev",
  ["executive solutions for healthcare management"] = "Exec Solut Healthc Manag",
  ["exemplaria classica"] = "ExClass",
  ["exemplaria: revista de literatura comparada"] = "Exemplaria",
  ["exercise and sport sciences reviews"] = "Exerc Sport Sci Rev",
  ["exercise immunology review"] = "Exerc Immunol Rev",
  ["exercise, sport, and movement"] = "Exercise, Sport, Movement",
  ["exeter papers in economic history"] = "Exeter Papers Econ Hist",
  ["exhibition guide. wellcome historical medical library"] = "Exhib Guide Wellcome Hist Med Libr",
  ["exobiology and future mars missions"] = "Exobiol Future Mars Missions",
  ["exosomes and microvesicles"] = "Exosomes Microvesicles",
  ["expedition. the magazine of archaeology, anthropology"] = "Expedition",
  ["experientia supplementum (2012)"] = "Exp Suppl",
  ["experientia. supplementum"] = "Experientia Suppl",
  ["experimental & applied acarology"] = "Exp Appl Acarol",
  ["experimental & molecular medicine"] = "Exp Mol Med",
  ["experimental & translational stroke medicine"] = "Exp Transl Stroke Med",
  ["experimental aging research"] = "Exp Aging Res",
  ["experimental agriculture"] = "Exp. Agric.",
  ["experimental analysis of human behavior bulletin [electronic resource]"] = "Exp Anal Hum Behav Bull",
  ["experimental and applied acarology"] = "Exp. Appl. Acarol.",
  ["experimental and clinical cardiology"] = "Exp Clin Cardiol",
  ["experimental and clinical endocrinology"] = "Exp Clin Endocrinol",
  ["experimental and clinical endocrinology & diabetes"] = "Exp. Clin. Endocrinol. Diabetes",
  ["experimental and clinical endocrinology & diabetes : official journal, german society of endocrinology [and] german diabetes association"] = "Exp Clin Endocrinol Diabetes",
  ["experimental and clinical endocrinology and diabetes"] = "Exp. Clin. Endocrinol. Diabetes",
  ["experimental and clinical immunogenetics"] = "Exp Clin Immunogenet",
  ["experimental and clinical psychopharacology"] = "Exp. Clin. Psychopharmacol.",
  ["experimental and clinical psychopharmacology"] = "Exp Clin Psychopharmacol",
  ["experimental and clinical transplantation"] = "Exp. Clin. Transplant.",
  ["experimental and clinical transplantation : official journal of the middle east society for organ transplantation"] = "Exp Clin Transplant",
  ["experimental and computational multiphase flow"] = "Exp Comput Multiph Flow",
  ["experimental and molecular medicine"] = "Exp. Mol. Med.",
  ["experimental and molecular pathology"] = "Exp Mol Pathol",
  ["experimental and molecular pathology. supplement"] = "Exp Mol Pathol Suppl",
  ["experimental and therapeutic medicine"] = "Exp Ther Med",
  ["experimental and toxicologic pathology"] = "Exp. Toxicol. Pathol.",
  ["experimental and toxicologic pathology : official journal of the gesellschaft fur toxikologische pathologie"] = "Exp Toxicol Pathol",
  ["experimental and toxicologic pathology : official journal of the gesellschaft für toxikologische pathologie"] = "Exp Toxicol Pathol",
  ["experimental animals"] = "Exp Anim",
  ["experimental animals / japanese association for laboratory animal science"] = "Exp Anim",
  ["experimental astronomy"] = "Exp. Astron.",
  ["experimental biology"] = "Exp Biol",
  ["experimental biology and medicine"] = "Exp Biol Med",
  ["experimental biology and medicine (london, united kingdom)"] = "Exp. Biol. Med. (London, U. K.)",
  ["experimental biology and medicine (maywood, n.j.)"] = "Exp Biol Med (Maywood)",
  ["experimental biology and medicine (maywood, nj, united states)"] = "Exp. Biol. Med. (Maywood, NJ, U. S.)",
  ["experimental biology online"] = "Exp. Biol. Online",
  ["experimental botany"] = "Exp. Bot.",
  ["experimental brain research"] = "Exp Brain Res",
  ["experimental brain research. experimentelle hirnforschung. experimentation cerebrale"] = "Exp Brain Res",
  ["experimental cell biology"] = "Exp Cell Biol",
  ["experimental cell research"] = "Exp Cell Res",
  ["experimental cell research. supplement"] = "Exp Cell Res Suppl",
  ["experimental dermatology"] = "Exp Dermatol",
  ["experimental diabesity research"] = "Exp Diabesity Res",
  ["experimental diabetes research"] = "Exp Diabetes Res",
  ["experimental economics"] = "Exp Econ",
  ["experimental embryology and teratology"] = "Exp Embryol Teratol",
  ["experimental eye research"] = "Exp Eye Res",
  ["experimental gerontology"] = "Exp Gerontol",
  ["experimental heat transfer"] = "Exp. Heat Transfer",
  ["experimental hematology"] = "Exp Hematol",
  ["experimental hematology & oncology"] = "Exp Hematol Oncol",
  ["experimental ir meets multilinguality, multimodality, and interaction : 8th international conference of the clef association, clef 2017, dublin, ireland, september 11-14, 2017, proceedings. cross-language evaluation forum. conference (8th : 2017 : dublin, ireland)"] = "Exp IR Meets Multilinguality Multimodality Interact (2017)",
  ["experimental lung research"] = "Exp Lung Res",
  ["experimental mathematics"] = "Exp. Math.",
  ["experimental mechanics"] = "Exp. Mech.",
  ["experimental medicine"] = "Exp Med",
  ["experimental medicine and surgery"] = "Exp Med Surg",
  ["experimental mycology"] = "Exp. Mycol.",
  ["experimental nephrology"] = "Exp Nephrol",
  ["experimental neurobiology"] = "Exp Neurobiol",
  ["experimental neurology"] = "Exp Neurol",
  ["experimental neurology. supplement"] = "Exp Neurol Suppl",
  ["experimental oncology"] = "Exp Oncol",
  ["experimental parasitology"] = "Exp Parasitol",
  ["experimental pathology"] = "Exp Pathol",
  ["experimental pathology. supplement"] = "Exp Pathol Suppl (1981)",
  ["experimental physiology"] = "Exp Physiol",
  ["experimental psychology"] = "Exp Psychol",
  ["experimental results"] = "Exp Results",
  ["experimental reults"] = "Exp. Results",
  ["experimental studies in psychology and pedagogy"] = "Exp Stud Psychol Pedagog",
  ["experimental techniques"] = "Exp. Tech.",
  ["experimental thermal and fluid science"] = "Exp Therm Fluid Sci",
  ["experimentelle medizin, pathologie und klinik"] = "Exp Med Pathol Klin",
  ["experimentelle pathologie"] = "Exp Pathol (Jena)",
  ["experimentelle pathologie. supplement = experimental pathology. supplement"] = "Exp Pathol Suppl",
  ["experimentelle strahlentherapie und klinische strahlenbiologie"] = "Exp. Strahlenther. Klin. Strahlenbiol.",
  ["experimentelle technik der physik"] = "Exp. Tech. Phys.",
  ["experiments in fluids"] = "Exp. Fluids",
  ["expert opinion on biological therapy"] = "Expert Opin Biol Ther",
  ["expert opinion on drug delivery"] = "Expert Opin Drug Deliv",
  ["expert opinion on drug discovery"] = "Expert Opin Drug Discov",
  ["expert opinion on drug metabolism & toxicology"] = "Expert Opin Drug Metab Toxicol",
  ["expert opinion on drug metabolism and toxicology"] = "Expert Opin. Drug Metab. Toxicol.",
  ["expert opinion on drug metabolsim & toxicology"] = "Expert Opin. Drug Metabol. Toxicol.",
  ["expert opinion on drug safety"] = "Expert Opin Drug Saf",
  ["expert opinion on emerging drugs"] = "Expert Opin Emerg Drugs",
  ["expert opinion on environmental biology"] = "Expert Opin Environ Biol",
  ["expert opinion on investigational drugs"] = "Expert Opin Investig Drugs",
  ["expert opinion on medical diagnostics"] = "Expert Opin Med Diagn",
  ["expert opinion on orphan drugs"] = "Expert Opin. Orphan Drugs",
  ["expert opinion on pharmacotherapy"] = "Expert Opin Pharmacother",
  ["expert opinion on therapeutic patents"] = "Expert Opin Ther Pat",
  ["expert opinion on therapeutic targets"] = "Expert Opin Ther Targets",
  ["expert review of anti-infective therapy"] = "Expert Rev Anti Infect Ther",
  ["expert review of anticancer therapy"] = "Expert Rev Anticancer Ther",
  ["expert review of cardiovascular therapy"] = "Expert Rev Cardiovasc Ther",
  ["expert review of clinical immunology"] = "Expert Rev Clin Immunol",
  ["expert review of clinical pharmacology"] = "Expert Rev Clin Pharmacol",
  ["expert review of dermatology"] = "Expert Rev Dermatol",
  ["expert review of endocrinology & metabolism"] = "Expert Rev Endocrinol Metab",
  ["expert review of gastroenterology & hepatology"] = "Expert Rev Gastroenterol Hepatol",
  ["expert review of hematology"] = "Expert Rev Hematol",
  ["expert review of medical devices"] = "Expert Rev Med Devices",
  ["expert review of molecular diagnostics"] = "Expert Rev Mol Diagn",
  ["expert review of neurotherapeutics"] = "Expert Rev Neurother",
  ["expert review of obstetrics & gynecology"] = "Expert Rev Obstet Gynecol",
  ["expert review of ophthalmology"] = "Expert Rev Ophthalmol",
  ["expert review of pharmacoeconomics & outcomes research"] = "Expert Rev Pharmacoecon Outcomes Res",
  ["expert review of precision medicine and drug development"] = "Expert Rev Precis Med Drug Dev",
  ["expert review of proteomics"] = "Expert Rev Proteomics",
  ["expert review of quality of life in cancer care"] = "Expert Rev Qual Life Cancer Care",
  ["expert review of respiratory medicine"] = "Expert Rev Respir Med",
  ["expert review of vaccines"] = "Expert Rev Vaccines",
  ["expert reviews in molecular medicine"] = "Expert Rev Mol Med",
  ["expert systems"] = "Expert Syst",
  ["expert systems with applications"] = "Expert Syst. Appl.",
  ["expert systems with applications: x"] = "Expert Syst. Appl.: X",
  ["explicit description of language and automatic processing of text"] = "Explicit Description Language Automat. Process. Text",
  ["exploration and mining geology"] = "Explor. Min. Geol.",
  ["exploration archéologique de délos faite par l’école française d’athènes"] = "Delos",
  ["exploration geophysics"] = "Explor. Geophys.",
  ["exploration of medicine"] = "Explor Med",
  ["exploration of targeted anti-tumor therapy"] = "Explor Target Antitumor Ther",
  ["exploration: cours et contributions pour les sciences de l’education"] = "Explor.: Cours Contrib. Sci. Ed.",
  ["explorations in early american culture"] = "Explor Early Am Culture",
  ["explorations in economic history"] = "Explor Econ Hist",
  ["explorations in ethnic studies : the journal of the national association of interdisciplinary ethnic studies"] = "Explor Ethn Stud",
  ["explorations in knowledge"] = "Explor Knowl",
  ["explorations in renaissance culture"] = "Explor Renaiss Cult",
  ["explorations in sights and sounds"] = "Explor Sights Sounds",
  ["explorations pyreneennes : bulletin trimestriel de la societe ramond"] = "Bull Soc Ramond",
  ["explorations pyrénéennes : bulletin trimestriel de la société ramond"] = "Bull Soc Ramond",
  ["exploratory research and hypothesis in medicine"] = "Explor Res Hypothesis Med",
  ["explore (new york, n.y.)"] = "Explore (NY)",
  ["explorer (kansas city)"] = "Explorer (Kansas City)",
  ["explorers journal"] = "Explor J",
  ["exploring complexity"] = "Explor. Complex.",
  ["exposes annuels de biochimie medicale"] = "Expos. Annu. Biochim. Med.",
  ["exposés annuels de biochimie médicale"] = "Expos Annu Biochim Med",
  ["expositiones mathematicae"] = "Expo Math",
  ["exposure and health"] = "Expo Health",
  ["express polymer letters"] = "Express Polym Lett",
  ["extended abstracts on human factors in computing systems. chi conference"] = "Ext Abstr Hum Factors Computing Syst",
  ["extending schumacher's concept of total accounting and accountability into the 21st century"] = "Adv. Public. Inter. Acc.",
  ["extracellular vesicles and circulating nucleic acids"] = "Extracell Vesicles Circ Nucl Acids",
  ["extracta mathematicae"] = "Extracta Math.",
  ["extractive industries and society"] = "Extr. Ind. Soc.",
  ["extremadura arqueológica"] = "ExtremA",
  ["extreme life, biospeology & astrobiology"] = "Extreme Life Biospeology Astrobiol",
  ["extreme mechanics letters"] = "Extreme Mech. Lett.",
  ["extreme physiology & medicine"] = "Extrem Physiol Med",
  ["extremes. statistical theory and applications in science, engineering and economics"] = "Extremes",
  ["extremophiles : life under extreme conditions"] = "Extremophiles",
  ["exxon monographs"] = "Exxon Monogr.",
  ["eye & contact lens"] = "Eye Contact Lens",
  ["eye (london, england)"] = "Eye (Lond)",
  ["eye and brain"] = "Eye Brain",
  ["eye and contact lens"] = "Eye Contact Lens",
  ["eye and vision"] = "Eye Vision",
  ["eye and vision (london, england)"] = "Eye Vis (Lond)",
  ["eye reports"] = "Eye Rep",
  ["eye science"] = "Eye Sci",
  ["eye, ear, nose & throat monthly"] = "Eye Ear Nose Throat Mon",
  ["eye, ear, nose and throat monthly"] = "Eye. Ear. Nose Throat Mon.",
  ["économie appliquée"] = "Econ Appl",
  ["économies et sociétés"] = "Econ Soc (Paris)",
  ["écrits de paris: revue des questions actuelles"] = "Ecrits Paris",
  ["éducation et francophonie"] = "Educ Francoph",
  ["éire-ireland; a journal of irish studies"] = "Eire Irel",
  ["éthique & santé"] = "Ethique Sante",
  ["études canadiennes. canadian studies"] = "Etud Can",
  ["études classiques (namur, belgium)"] = "Etud Class",
  ["études d'histoire africaine"] = "Etud Hist Afr",
  ["études du centre de recherches et d'études démographiques. centre de recherche et d'études démographiques"] = "Etudes Cent Rech Etudes Demogr",
  ["études inuit. inuit studies"] = "Etud Inuit",
  ["études limousines"] = "Etud Limousines",
  ["études maliennes"] = "Etudes Mali",
  ["ēpeirōtikē hestia"] = "Epeirotike Hestia",
  ["ėkologicheskai︠a︡ genetika"] = "Ekol Genet",
  ["ėkonomicheskai︠a︡ gazeta"] = "Ekon Gaz",
  ["ėkonomika selʹskogo khozi︠a︡ĭstva : ėkonomicheskiĭ zhurnal ministerstva selʹskogo khozi︠a︡ĭstva sssr"] = "Ekon Selsk Khoz",
  ["ėkonomika sovetskoĭ ukrainy"] = "Ekon Sov Ukr",
  ["ėksperimental'nai︠a︡ i klinicheskai︠a︡ gastroėnterologii︠a︡ = experimental & clinical gastroenterology"] = "Eksp Klin Gastroenterol",
  ["ėntomologicheskie i parazitologicheskie issledovanii︠a︡ v povolzh'e"] = "Entomol I Parazitol Issled V Povolzhe",
  ["ėpidemiologii︠a︡ i infekt︠s︡ionnye bolezni (moscow, russia : 2011)"] = "Epidemiol Infekc Bolezni (Mosc : 2011)",
  ["ėpidemiologii︠a︡ i vakt︠s︡inoprofilaktika"] = "Epidemiol Vakcinoprofil",
  ["f&s reports"] = "F S Rep",
  ["f&s reviews"] = "F S Rev",
  ["f&s science"] = "F S Sci",
  ["f. jacoby, die fragmente der griechischen historiker"] = "FGrHist",
  ["f.a.s. professional bulletin"] = "FAS Prof Bull",
  ["f.a.s. public interest report"] = "FAS Public Interest Rep",
  ["f.n.i.b.; organe de la federation nationale des infirmier(e)s belges"] = "FNIB",
  ["f.n.i.b.; organe de la fédération nationale des infirmier(e)s belges"] = "FNIB",
  ["f1000 biology reports"] = "F1000 Biol Rep",
  ["f1000 faculty reviews"] = "F1000 Fac Rev",
  ["f1000 medicine reports"] = "F1000 Med Rep",
  ["f1000prime reports"] = "F1000Prime Rep",
  ["fa i hsueh tsa chih journal of forensic medicine (shang-hai"] = "Fa I Hsueh Tsa Chih",
  ["fa yi xue za zhi"] = "Fa Yi Xue Za Zhi",
  ["fa yi xue za zhi (journal of forensic medicine)"] = "Fa Yi Xue Za Zhi",
  ["facets (ottawa)"] = "Facets (Ott)",
  ["fachberichte simulation"] = "Fachber. Simul.",
  ["fachblatt für schweizerische heime und anstalten. revue suisse des établissements hospitaliers et d'éducation"] = "Fachbl Schweiz Heime Anst Rev Suisse Etabl Hosp Educ",
  ["fachblatt für schweizerisches anstaltswesen. revue suisse des établissements hospitaliers"] = "Fachbl Schweiz Anst",
  ["fachbuchreihe fur studium fortbildung praxis"] = "Fachbuchreihe Stud. Fortbild. Prax.",
  ["facial orthopedics and temporomandibular arthrology"] = "Facial Orthop Temporomandibular Arthrol",
  ["facial plastic surgery"] = "Facial Plast. Surg.",
  ["facial plastic surgery & aesthetic medicine"] = "Facial Plast Surg Aesthet Med",
  ["facial plastic surgery : fps"] = "Facial Plast Surg",
  ["facial plastic surgery clinics of north america"] = "Facial Plast Surg Clin North Am",
  ["fact sheet (center for home care policy and research (u.s.))"] = "Fact Sheet (Cent Home Care Policy Res)",
  ["fact sheets on sweden"] = "Fact Sheets Swed",
  ["facta universitatis"] = "Facta Univ. Ser. Math. Inform.",
  ["facta universitatis - series: electronics and energetics"] = "Facta Univ. - Ser. Electron. Energ.",
  ["facta universitatis. series automatic control and robotics"] = "Facta Univ. Ser. Autom. Control Robot.",
  ["facta universitatis. series, mechanics, automatic control and robotics"] = "Facta Univ Ser Mech Autom Control Robot",
  ["facta universitatis. series: mathematics and informatics"] = "Facta Univ. Ser. Math. Inform.",
  ["factor odontologico"] = "Factor Odontol.",
  ["factor odontológico"] = "Factor Odontol",
  ["facts about infant feeding"] = "Facts Infant Feed",
  ["facts, views & vision in obgyn"] = "Facts Views Vis Obgyn",
  ["faculty notes (new orleans, la.)"] = "Fac Notes (New Orleans La)",
  ["faculty reviews"] = "Fac Rev",
  ["faenza. bollettino del museo internazionale delle ceramiche in faenza. rivista bimestrale di studi storici e di tecnica dell’arte ceramica"] = "Faenza",
  ["fag tidsskriftet sykepleien"] = "Fag Tidsskr Sykepleien",
  ["fagopyrum : novosti o ajdi"] = "buckwheat newsletter|Fagopyrum",
  ["fagopyrum : novosti o ajdi = buckwheat newsletter"] = "Fagopyrum",
  ["faith and philosophy : journal of the society of christian philosophers"] = "Faith Philos",
  ["faits de langues"] = "Faits Lang",
  ["faits et tendances"] = "Fait Tend",
  ["fako de l’funkcialaj ekvacioj japana matematika societo"] = "Funkcial. Ekvac.",
  ["familial cancer"] = "Fam Cancer",
  ["families and child health"] = "Families Child Health",
  ["families in society : the journal of contemporary human services"] = "Fam Soc",
  ["families, relationships and societies : in international journal of research and debate"] = "Fam Relatsh Soc",
  ["families, systems & health : the journal of collaborative family healthcare"] = "Fam Syst Health",
  ["famille et developpement"] = "Fam Dev",
  ["famille et développement"] = "Fam Dev",
  ["family & community health"] = "Fam Community Health",
  ["family & community history : journal of the family and community historical research society"] = "Fam Community Hist",
  ["family and community health"] = "Fam. Community Health",
  ["family and conciliation courts review"] = "Fam Concil Courts Rev",
  ["family and consumer sciences research journal"] = "Fam Consum Sci Res J",
  ["family and consumer sciences research journal / american association of family and consumer sciences"] = "Fam Consum Sci Res J",
  ["family circle (mount morris, ill.)"] = "Fam Circle",
  ["family coordinator"] = "Fam Coord",
  ["family court review"] = "Fam Court Rev",
  ["family economics and nutrition review"] = "Family Econ. Nutrition Rev.",
  ["family economics review"] = "Fam Econ Rev",
  ["family health"] = "Fam Health",
  ["family journal (alexandria, va.)"] = "Fam J Alex Va",
  ["family law (chichester)"] = "Family Law",
  ["family law quarterly"] = "Fam Law Q",
  ["family law reports"] = "Fam Law Rep",
  ["family law review"] = "Fam Law Rev",
  ["family life educator"] = "Fam Life Educ",
  ["family life matters"] = "Fam Life Matters",
  ["family matters (melbourne, vic.)"] = "Fam Matters",
  ["family medicine"] = "Fam Med",
  ["family medicine & medical science research"] = "Fam Med Med Sci Res",
  ["family medicine and community health"] = "Fam Med Community Health",
  ["family medicine and primary care -- open access"] = "Fam Med Prim Care Open Access",
  ["family planning"] = "Fam. Plann.",
  ["family planning digest"] = "Fam Plann Dig",
  ["family planning information service"] = "Fam Plann Inf Serv",
  ["family planning manager"] = "Fam Plan Manag",
  ["family planning news"] = "Fam Plann News (Tokyo)",
  ["family planning perspectives"] = "Fam Plann Perspect",
  ["family planning quarterly"] = "Fam Plann Q",
  ["family planning resume"] = "Fam Plann Resume",
  ["family planning résumé"] = "Fam Plann Resume",
  ["family planning today"] = "Fam Plann Today",
  ["family planning world"] = "Family Plan World",
  ["family planning/population reporter; a review of state laws and policies"] = "Fam Plann Popul Rep",
  ["family practice"] = "Fam Pract",
  ["family practice management"] = "Fam Pract Manag",
  ["family practice news"] = "Fam Pract News",
  ["family practice research journal"] = "Fam Pract Res J",
  ["family process"] = "Fam Process",
  ["family relations"] = "Fam Relat",
  ["family science"] = "Fam Sci",
  ["family systems medicine"] = "Fam Syst Med",
  ["family therapy"] = "Fam Ther",
  ["family urology"] = "Fam Urol",
  ["family violence prevention and health practice : an e-journal of the family violence prevention fund"] = "Fam Viol Prev Health Pract",
  ["famplanco news"] = "Famplanco News",
  ["fao fisheries circular = fao circulaire sur les pêches = fao circular de pesca"] = "FAO Fish Circ",
  ["fao food and nutrition paper"] = "FAO Food Nutr Pap",
  ["fao food and nutrition series"] = "FAO Food Nutr Ser",
  ["fao forestry development paper"] = "FAO For. Dev. Pap.",
  ["fao nutrition meetings report series"] = "FAO Nutr Meet Rep Ser",
  ["fao nutritional studies"] = "FAO Nutr Stud",
  ["far east journal of applied mathematics"] = "Far East J Appl Math",
  ["far east journal of mathematical sciences"] = "Far East J. Math. Sci.",
  ["far eastern affairs"] = "Far East Aff",
  ["far eastern economic review"] = "Far East Econ Rev",
  ["far eastern entomologist = dalʹnevostochnyĭ ėntomolog"] = "Far East Entomol",
  ["far, the french-american review"] = "Fr Am Rev",
  ["faraday discussions"] = "Faraday Discuss",
  ["faraday discussions of the chemical society"] = "Faraday Discuss. Chem. Soc.",
  ["faraday discussions of the royal society of chemistry"] = "Faraday Discuss. R. Soc. Chem.",
  ["faraday special discussions of the chemical society"] = "Farday Spec. Discuss. Chem. Soc.",
  ["faraday symposium of the chemical society"] = "Faraday Symp. Chem. Soc.",
  ["faraday symposium of the royal society of chemistry"] = "Faraday Symp. R. Soc. Chem.",
  ["faraday transactions"] = "Faraday Trans.",
  ["farmaceuticky obzor"] = "Farm Obz",
  ["farmaceutický obzor"] = "Farm Obz",
  ["farmaceutisk tidende"] = "Farm Tid",
  ["farmaceutski glasnik"] = "Farm Glas",
  ["farmacevtisk revy"] = "Farm Revy",
  ["farmacevtski vestnik"] = "Farm Vestn",
  ["farmacia clinica"] = "Farm Clin",
  ["farmacia clínica"] = "Farm Clin",
  ["farmacia hospitalaria : organo oficial de expresion cientifica de la sociedad espanola de farmacia hospitalaria"] = "Farm Hosp",
  ["farmacia hospitalaria : órgano oficial de expresión científica de la sociedad española de farmacia hospitalaria"] = "Farm Hosp",
  ["farmacia nueva"] = "Farm Nueva",
  ["farmacia y química"] = "Farm Quim",
  ["farmacihistoriska sallskapets degrees arsskrift"] = "Farmacihist Sallsk Ars",
  ["farmacihistoriska sallskapets °arsskrift"] = "Farmacihist Sallsk Ars",
  ["farmacja polska"] = "Farm Pol",
  ["farmaco (societa chimica italiana : 1989)"] = "Farmaco",
  ["farmaco (società chimica italiana : 1989)"] = "Farmaco",
  ["farmaco. edizione pratica"] = "Farmaco. [Prat.]",
  ["farmaco. edizione scientifica"] = "Farmaco. [Sci.]",
  ["farmacognosia; anales del instituto jose celestino mutis"] = "Farmacognosia",
  ["farmacognosia; anales del instituto josé celestino mutis"] = "Farmacognosia",
  ["farmacoterapia actual"] = "Farmacoter Actual",
  ["farmakologiia i toksikologiia"] = "Farmakol Toksikol",
  ["farmatsevtychnyi zhurnal"] = "Farm. Zh.",
  ["farmatsevtychnyĭ zhurnal"] = "Farm Zh",
  ["farmatsiya (moscow)"] = "Farmatsiya (Moscow)",
  ["faseb bioadvances"] = "FASEB Bioadv",
  ["faseb journal"] = "FASEB J.",
  ["faseb journal : official publication of the federation of american societies for experimental biology"] = "FASEB J",
  ["fashion and textiles"] = "Fashion Text.",
  ["fasti archaeologici"] = "FA",
  ["fasttimes : the eegs newsletter"] = "FastTIMES",
  ["faṣlnāmah-i kūmish"] = "Faslnamahi Kumish",
  ["faṣlnāmah-i takhaṣṣuṣī-í ulūm-i zīstī"] = "Faslnamahi Takhassusii Ulumi Zisti",
  ["fatigue & fracture of engineering materials & structures"] = "Fatigue Fract Eng Mater Struct",
  ["fatigue : biomedicine, health & behavior"] = "Fatigue",
  ["fatigue and fracture of engineering materials and structures"] = "Fatigue Fract. Eng. Mater. Struct.",
  ["faulkner & gray's medicine & health"] = "Faulkner Grays Med Health",
  ["faulkner & gray’s medicine & health"] = "Faulkner Grays Med Health",
  ["faulkner and gray’s medicine and health"] = "Faulkner Grays Med. Health",
  ["fauna norvegica"] = "Fauna Norv",
  ["fauna och flora"] = "Fauna Flora",
  ["fauna of new zealand"] = "Fauna N Z",
  ["fauna ryukyuana"] = "Fauna Ryukyuana",
  ["fave. seccion ciencias veterinarias"] = "FAVE Secc Cienc Vet",
  ["faventia. departement de clássiques, facultat de letres, universitat autónoma de barcelona"] = "Faventia",
  ["fda consumer"] = "FDA Consum",
  ["fda drug bulletin"] = "FDA Drug Bull",
  ["fda medical bulletin : important information for health professionals from the u.s. food & drug administration"] = "FDA Med Bull",
  ["fdc reports. drugs and cosmetics"] = "FDC Rep Drugs Cosmet",
  ["fdg : proceedings of the international conference on foundations of digital games. international conference on the foundations of digital games"] = "FDG",
  ["fdi world"] = "FDI World",
  ["fdic banking review"] = "FDIC Banking Rev.",
  ["fdli's food and drug policy forum"] = "FDLIs Food Drug Policy Forum",
  ["featured reviews in mathematical reviews"] = "Featur. Rev. Math. Rev.",
  ["febs journal"] = "FEBS J.",
  ["febs letters"] = "FEBS Lett",
  ["febs open bio"] = "FEBS Open Bio",
  ["feddersen wierde. die ergebnisse der ausgrabung der vorgeschichtlichen wurt feddersen wierde bei bremerhaven in den jahren 1955 bis 1963"] = "FeddersenWierde",
  ["feddes repertorium"] = "Feddes Repert",
  ["federal bar journal : the official publication of the federal bar association"] = "Fed Bar J",
  ["federal personnel manual system. fpm bulletin / office of personnel management. united states. office of personnel management"] = "FPM Bull",
  ["federal personnel manual system. fpm bulletin. united states. office of personnel management"] = "FPM Bull",
  ["federal practitioner : for the health care professionals of the va, dod, and phs"] = "Fed Pract",
  ["federal probation"] = "Fed Probat",
  ["federal register"] = "Fed. Regist.",
  ["federal reserve bank business review of philadelphia"] = "Fed. Reserve Bank Bus. Rev. Phila.",
  ["federal reserve bank of atlanta economic review"] = "Fed. Reserve Bank Atlanta Econ. Rev.",
  ["federal reserve bank of atlanta monthly review"] = "Fed. Reserve Bank Atlanta Rev.",
  ["federal reserve bank of boston new england economic review"] = "Fed. Reserve Bank Boston New Eng. Econ. Rev.",
  ["federal reserve bank of chicago economic perspectives"] = "Fed. Reserve Bank Chicago Econ. Perspect.",
  ["federal reserve bank of cleveland economic review"] = "Fed. Reserve Bank Cleveland Econ. Rev.",
  ["federal reserve bank of dallas economic and financial review"] = "Fed. Reserve Bank Dallas Econ. Finan. Rev.",
  ["federal reserve bank of dallas economic review"] = "Fed. Reserve Bank Dallas Econ. Rev.",
  ["federal reserve bank of kansas city economic review"] = "Fed. Reserve Bank Kansas City Econ. Rev.",
  ["federal reserve bank of kansas city monthly review"] = "Fed. Reserve Bank Kansas City Rev.",
  ["federal reserve bank of minneapolis quarterly review"] = "Fed. Reserve Bank Minneapolis Quart. Rev.",
  ["federal reserve bank of new york economic policy review"] = "Fed. Reserve Bank New York Econ. Pol. Rev.",
  ["federal reserve bank of new york monthly review"] = "Fed. Reserve Bank New York Rev.",
  ["federal reserve bank of new york quarterly review"] = "Fed. Reserve Bank New York Quart. Rev.",
  ["federal reserve bank of philadelphia business review"] = "Fed. Reserve Bank Philadelphia Bus. Rev.",
  ["federal reserve bank of richmond economic quarterly"] = "Fed. Reserve Bank Richmond Econ. Quart.",
  ["federal reserve bank of richmond economic review"] = "Fed. Reserve Bank Richmond Econ. Rev.",
  ["federal reserve bank of san francisco economic review"] = "Fed. Reserve Bank San Francisco Econ. Rev.",
  ["federal reserve bank of san francisco monthly review"] = "Fed. Reserve Bank San Francisco Rev.",
  ["federal reserve bank of st. louis review"] = "Fed. Reserve Bank St. Louis Rev.",
  ["federal reserve bulletin"] = "Fed. Reserve Bull.",
  ["federal rules decisions"] = "Fed Rules Decis",
  ["federal supplement"] = "Fed Suppl",
  ["federalist (washington, d.c. : 1980)"] = "Federalist",
  ["federation bulletin"] = "Fed Bull",
  ["federation bulletin / federation of state medical boards of the united states"] = "Fed Bull",
  ["federation of european biochemical societies letters"] = "Fed. Eur. Biochem. Soc. Lett.",
  ["federation of insurance & corporate counsel quarterly"] = "Fed Insur Corp Couns Q",
  ["federation of insurance counsel quarterly"] = "Fed Insur Couns Q",
  ["federation of operative dentistry"] = "Fed Oper Dent",
  ["federation proceedings"] = "Fed Proc",
  ["federation proceedings. translation supplement"] = "Fed. Proc. Transl. Suppl.",
  ["federation proceedings. translation supplement; selected translations from medical-related science"] = "Fed Proc Transl Suppl",
  ["feedback : icomp newsletter on management of population programmes"] = "Feedback",
  ["feit & fictie : tijdschrift voor de geschiedenis van de representatie"] = "Feit Fict",
  ["feldsher i akusherka"] = "Feldsher Akush.",
  ["feline practice"] = "Feline Pract",
  ["felix ravenna"] = "FelRav",
  ["felʹdsher i akusherka"] = "Feldsher Akush",
  ["fel’dsher i akusherka"] = "Feldsher Akush",
  ["female health topics & diagnostic reporter"] = "Female Health Top Diagn Report",
  ["female patient"] = "Female Patient",
  ["female patient (primary care ed.)"] = "Female Patient (Prim Care Ed)",
  ["female pelvic medicine & reconstructive surgery"] = "Female Pelvic Med Reconstr Surg",
  ["femina : revista da federacao brasileira das sociedades de ginecologia e obstetricia"] = "Femina",
  ["femina : revista da federação brasileira das sociedades de ginecologia e obstetrícia"] = "Femina",
  ["feminism & psychology"] = "Fem Psychol",
  ["feminist criminology"] = "Fem Criminol",
  ["feminist economics"] = "Fem Econ",
  ["feminist issues"] = "Fem Issues",
  ["feminist legal studies"] = "Fem Leg Stud",
  ["feminist media studies"] = "Fem Media Stud",
  ["feminist review"] = "Fem Rev",
  ["feminist studies : fs"] = "Fem Stud",
  ["femnet news"] = "FEMNET News",
  ["fems immunology and medical microbiology"] = "FEMS Immunol Med Microbiol",
  ["fems microbes"] = "FEMS Microbes",
  ["fems microbiology ecology"] = "FEMS Microbiol Ecol",
  ["fems microbiology immunology"] = "FEMS Microbiol Immunol",
  ["fems microbiology letters"] = "FEMS Microbiol Lett",
  ["fems microbiology review"] = "FEMS Microbiol. Rev.",
  ["fems microbiology reviews"] = "FEMS Microbiol Rev",
  ["fems yeast research"] = "FEMS Yeast Res",
  ["fen dergisi. journal of science"] = "Fen Derg.",
  ["fen zi xi bao sheng wu xue bao (journal of molecular cell biology)"] = "Fen Zi Xi Bao Sheng Wu Xue Bao",
  ["fen zi xi bao sheng wu xue bao = journal of molecular cell biology"] = "Fen Zi Xi Bao Sheng Wu Xue Bao",
  ["fenetre sur cours"] = "Fenetre Cours",
  ["fenxi ceshi tongbao"] = "Fenxi Ceshi Tongbao",
  ["fenxi huaxue"] = "Fenxi Huaxue",
  ["fenxi shiyanshi"] = "Fenxi Shiyanshi",
  ["fenzi cuihua"] = "Fenzi Cuihua",
  ["fermentation (basel, switzerland)"] = "Fermentation (Basel)",
  ["fermentation technology"] = "Ferment Technol",
  ["fernerkundung und photogrammetrie"] = "Fernerkund. Photogramm.",
  ["ferroelectric letters section"] = "Ferroelectr. Lett. Sect.",
  ["ferroelectrics letters"] = "Ferroelectr. Lett.",
  ["ferroelectrics letters section"] = "Ferroelectr. Lett. Sect.",
  ["ferroelectrics review"] = "Ferroelectr. Rev.",
  ["ferroelectrics, letters section"] = "Ferroelectr., Lett. Sect.",
  ["fertilite, contraception, sexualite"] = "Fertil Contracept Sex",
  ["fertilite, orthogenie"] = "Fertil Orthog",
  ["fertilité, contraception, sexualité"] = "Fertil Contracept Sex",
  ["fertilité, orthogénie"] = "Fertil Orthog",
  ["fertility and contraception"] = "Fertil Contracept",
  ["fertility and sterility"] = "Fertil Steril",
  ["fertility determinants research notes"] = "Fertil Determ Res Notes",
  ["fertility determinants research notes / the population council"] = "Fertil Determ Res Notes",
  ["fertility research and practice"] = "Fertil Res Pract",
  ["fetal and maternal medicine review"] = "Fetal Matern Med Rev",
  ["fetal and pediatric pathology"] = "Fetal Pediatr Pathol",
  ["fetal diagnosis and therapy"] = "Fetal Diagn Ther",
  ["fetal therapy"] = "Fetal Ther",
  ["fetal, infant and ophthalmic medical image analysis : international workshop, fifi 2017 and 4th international workshop, omia 2017, held in conjunction with miccai 2017, québec city, qc, canada, september 14, 2017. fifi (workshop) (2017 : québec, québec)"] = "Fetal Infant Ophthalmic Med Image Anal (2017)",
  ["fett - lipid"] = "Fett - Lipid",
  ["fett wissenschaft technologie : organ der deutschen gesellschaft fur fettwissenschaft e.v."] = "Fat science technology|Fett Wiss. Technol.",
  ["fett wissenschaft technologie : organ der deutschen gesellschaft für fettwissenschaft e.v. = fat science technology"] = "Fett Wiss Technol",
  ["fettchemische umschau"] = "Fettchem. Umsch.",
  ["fette und seifen"] = "Fette Seifen",
  ["feuillets de biologie"] = "Feuill Biol",
  ["feuillets du praticien"] = "Feuill Prat",
  ["few-body systems"] = "Few-Body Syst.",
  ["few-body systems, supplementum"] = "Few-Body Syst., Suppl.",
  ["fias interdisciplinary science series"] = "FIAS Interdiscip. Sci. Ser.",
  ["fiber and integrated optics"] = "Fiber Integr. Opt.",
  ["fibers (basel, switzerland)"] = "Fibers (Basel)",
  ["fibers and polymers"] = "Fibers Polym.",
  ["fibre chemistry"] = "Fibre Chem.",
  ["fibre science and technology"] = "Fibre Sci. Technol.",
  ["fibres & textiles in eastern europe"] = "Fibres Text. East. Eur",
  ["fibrogenesis & tissue repair"] = "Fibrogenesis Tissue Repair",
  ["ficheiro epigráfico. suplemento de »conimbriga«"] = "FichEpigr",
  ["fichero médico terapéutico"] = "Fich Med Ter Puriss",
  ["fides et historia : official publication of the conference on faith and history"] = "Fides Et Hist",
  ["field actions science reports"] = "Field Actions Sci Rep",
  ["field analytical chemistry and technology."] = "Field Anal. Chem. Technol.",
  ["field crops research"] = "Field Crops Res",
  ["field exchange"] = "Field Exch",
  ["field methods"] = "Field methods",
  ["field mycology"] = "Field Mycol.",
  ["field mycology : a magazine for the study and identification of wild fungi"] = "Field Mycol",
  ["field staff reports"] = "Field Staff Rep",
  ["field staff reports / ufsi"] = "Field Staff Rep",
  ["field studies"] = "Field Stud.",
  ["fieldiana anthropology"] = "Fieldiana Anthropol.",
  ["fieldiana botany"] = "Fieldiana Botan.",
  ["fieldiana geology"] = "Fieldiana Geol.",
  ["fieldiana life and earth sciences"] = "Fieldiana Life Earth Sci.",
  ["fieldiana zoology"] = "Fieldiana Zool.",
  ["fieldiana. botany"] = "Fieldiana Bot",
  ["fieldiana. life and earth sciences"] = "Fieldiana Life Earth Sci",
  ["fields and quanta"] = "Fields Quanta",
  ["fields institute communications"] = "Fields Inst. Commun.",
  ["fields institute monographs"] = "Fields Inst. Monogr.",
  ["fields mathematics education journal"] = "Fields Math. Educ. J.",
  ["fieldstaff reports. africa. northeast africa series"] = "Fieldstaff Rep Northeast Afr Ser",
  ["fieldstaff reports. asia. southeast asia series"] = "Fieldstaff Rep",
  ["fieldstaff reports. southeast asia series"] = "Fieldstaff Rep",
  ["fieldstaff reports. west europe series"] = "Fieldstaff Rep West Eur Ser",
  ["figlina. documents du laboratoire de céramologie de lyon"] = "Figlina",
  ["fiji medical journal"] = "Fiji Med J",
  ["filaria journal"] = "Filaria J",
  ["film history"] = "Film Hist",
  ["film international"] = "Film Int",
  ["filosofia—filosofia contemporanea e historia da filosofia (london)"] = "Filosofia (London)",
  ["filtration & separation"] = "Filtr. Sep.",
  ["filtration and separation"] = "Filtr. Sep.",
  ["filtration industry analyst"] = "Filtr. Ind. Anal.",
  ["final programme & abstracts : aids vaccine 04, lausanne, switzerland, august 30th-september 1st, 2004. aids vaccine international conference (2004 : lausanne, switzerland)"] = "Programme Abstr AIDS Vaccine 2004",
  ["finance & development"] = "Finance Dev",
  ["finance a úver"] = "Finance Úver",
  ["finance and development"] = "Finance Devel.",
  ["finance and stochastics"] = "Finance Stochastics",
  ["finance and trade review"] = "Finance Trade Rev.",
  ["finance for professionals"] = "Finance Prof.",
  ["finance india"] = "Finance India",
  ["finance research letters"] = "Financ Res Lett",
  ["financial engineering and the japanese markets"] = "Finan. Eng. Japanese Markets",
  ["financial engineering explained"] = "Financ. Eng. Explain.",
  ["financial history review"] = "Finan. Hist. Rev.",
  ["financial industry studies"] = "Finan. Industry Stud.",
  ["financial management"] = "Financ Manage",
  ["financial markets, institutions and instruments"] = "Finan. Markets",
  ["financial mathematics"] = "Financ. Math.",
  ["financial mathematics and applications"] = "Financ. Math. Appl.",
  ["financial mathematics and fintech"] = "Financ. Math. Fintech",
  ["financial practice and education"] = "Finan. Practice Educ.",
  ["financial review"] = "Finan. Rev.",
  ["financial statistics"] = "Financ. Stat.",
  ["financial times (north american edition)"] = "Financ Times",
  ["finanz-archiv : zeitschrift fur das gesamte finanzwesen"] = "Finanzarchiv",
  ["finanz-archiv : zeitschrift für das gesamte finanzwesen"] = "Finanzarchiv",
  ["findings brief : health care financing & organization"] = "Find Brief",
  ["fine focus"] = "Fine Focus",
  ["finite elements in analysis and design"] = "Finite Elem. Anal. Des.",
  ["finite elements in analysis and design : the international journal of applied finite elements and computer aided engineering"] = "Finite Elem Anal Des",
  ["finite fields and their applications"] = "Finite Fields Appl.",
  ["finnish economic papers"] = "Finnish Econ. Pap.",
  ["finska lakaresallskapets handlingar"] = "Fin. Lakaresallsk. Handl.",
  ["finska läkaresällskapets handlingar"] = "Fin Lakaresallsk Handl",
  ["fire (basel, switzerland)"] = "Fire (Basel)",
  ["fire and materials"] = "Fire Mater.",
  ["fire ecology"] = "Fire Ecol.",
  ["fire journal"] = "Fire J.",
  ["fire journal (boston, mass.)"] = "Fire J",
  ["fire safety journal"] = "Fire Saf. J.",
  ["fire safety science"] = "Fire Saf. Sci.",
  ["fire science reviews"] = "Fire Sci. Rev.",
  ["fire technology"] = "Fire Technol.",
  ["first annual ieee conference on control technology and applications : ccta 2017 : kohala coast, hawai'i, august 27-30, 2017. ieee conference on control technology and applications (1st : 2017 : waimea, hawaii island, hawaii)"] = "First Annu IEEE Conf Control Technol Appl (2017)",
  ["first international press lecture series"] = "First Int. Press Lect. Ser.",
  ["first language"] = "First Lang",
  ["first peoples child & family review"] = "First Peoples Child Fam Rev",
  ["first things (new york, n.y.)"] = "First Things",
  ["fiscal studies"] = "Fisc Stud",
  ["fish & shellfish immunology"] = "Fish Shellfish Immunol",
  ["fish and fisheries"] = "Fish Fish.",
  ["fish and fisheries (oxford, england)"] = "Fish Fish (Oxf)",
  ["fish and shellfish immunology reports"] = "Fish Shellfish Immunol. Rep.",
  ["fish pathology"] = "Fish Pathol",
  ["fish physiology and biochemistry"] = "Fish Physiol Biochem",
  ["fisheries and aquatic science"] = "Fish. Aquat. Sci.",
  ["fisheries and aquatic sciences"] = "Fish Aquatic Sci",
  ["fisheries management and ecology"] = "Fish Manag Ecol",
  ["fisheries oceanography"] = "Fish Oceanogr",
  ["fisheries research"] = "Fish Res",
  ["fisheries science"] = "Fish. Sci.",
  ["fisheries science (carlton, australia)"] = "Fish. Sci. (Carlton, Aust.)",
  ["fisheries science (tokyo, japan)"] = "Fish. Sci. (Tokyo, Jpn.)",
  ["fisheries science : fs"] = "Fish Sci",
  ["fishery bulletin (washington, d.c. : 1971)"] = "Fish Bull (Wash D C)",
  ["fishes of sahul : journal of the australia new guinea fishes association"] = "Fish Sahul",
  ["fishtaxa : journal of fish taxonomy"] = "FishTaxa",
  ["fisiologia e medicina"] = "Fisiol Med",
  ["fission product inhalation project"] = "Fission Prod. Inhal. Proj.",
  ["fission product inhalation project [technical progress report]. lovelace foundation for medical education and research"] = "Fission Prod Inhal Proj",
  ["fitopatologia brasileira"] = "Fitopatol Bras",
  ["fixed point theory"] = "Fixed Point Theory",
  ["fixed point theory and algorithms for sciences and engineering"] = "Fixed Point Theory Algorithms Sci. Eng.",
  ["fixed point theory and applications"] = "Fixed Point Theory Appl.",
  ["fixed point theory. an international journal on fixed point theory, computation and applications"] = "Fixed Point Theory",
  ["fizika elementarnykh chastits i atomnogo yadra [soviet journal of particles and nuclei]"] = "Fiz. Elem. Chastits At. Yadra [Sov. J. Part. Nucl.]",
  ["fizika goreniya i vzryva"] = "Fiz. Goreniya Vzryva",
  ["fizika i khimiya obrabotki materialov"] = "Fiz. Khim. Obrab. Mater.",
  ["fizika i khimiya stekla"] = "Fiz. Khim. Stekla",
  ["fizika i tekhnika poluprovodnikov (s. -peterburg)"] = "Fiz. Tekh. Poluprovodn. (Leningrad)",
  ["fizika i tekhnika poluprovodnikov [soviet physics–semiconductors]"] = "Fiz. Tekh. Poluprovodn. [Sov. Phys. Semicond.]",
  ["fizika metallov i metallovedenie"] = "Fiz. Met. Metalloved.",
  ["fizika metallov i metallovedenie [physics of metals and metallography (ussr)]"] = "Fiz. Met. Metalloved. [Phys. Met. Metallogr. (USSR)]",
  ["fizika nizkikh temperatur (kiev)"] = "Fiz. Nizk. Temp. (Kiev)",
  ["fizika nizkikh temperatur [soviet journal of low temperature physics]"] = "Fiz. Nizk. Temp. [Sov. J. Low Temp. Phys.]",
  ["fizika plazmy"] = "Fiz. Plazmy",
  ["fizika plazmy (moscow)"] = "Fiz. Plazmy (Moscow)",
  ["fizika plazmy [soviet journal of plasma physics]"] = "Fiz. Plazmy [Sov. J. Plasma Phys.]",
  ["fizika tverdogo tela (leningrad) [soviet physics–solid state]"] = "Fiz. Tverd. Tela (Leningrad) [Sov. Phys. Solid State]",
  ["fizika tverdogo tela (s. -peterburg)"] = "Fiz. Tverd. Tela (Leningrad)",
  ["fizika èlementarnykh chastits i atomnogo yadra"] = "Fiz. Èlementar. Chastits i Atom. Yadra",
  ["fiziko-khimicheskaya mekhanika materialov"] = "Fiz.-Khim. Mekh. Mater.",
  ["fiziko-matematichesko spisanie"] = "Fiz.-Mat. Spis.",
  ["fiziko-matematicheskoe nasledie"] = "Fiz.-Mat. Nasled.",
  ["fiziologia normala si patologica"] = "Fiziol. Norm. Patol.",
  ["fiziologia normală şi patologică"] = "Fiziol Norm Patol",
  ["fiziologicheskii zhurnal"] = "Fiziol. Zh.",
  ["fiziologicheskii zhurnal imeni i. m. sechenova"] = "Fiziol. Zh. Im. I. M. Sechenova",
  ["fiziologicheskii zhurnal imeni i.m. sechenova / rossiiskaia akademiia nauk"] = "Fiziol Zh Im I M Sechenova",
  ["fiziologicheskii zhurnal sssr imeni i. m. sechenova"] = "Fiziol. Zh. SSSR Im. I. M. Sechenova",
  ["fiziologicheskiĭ zhurnal"] = "Fiziol Zh",
  ["fiziologicheskiĭ zhurnal imeni i.m. sechenova"] = "Fiziol Zh Im I M Sechenova",
  ["fiziologicheskiĭ zhurnal sssr imeni i. m. sechenova"] = "Fiziol Zh SSSR Im I M Sechenova",
  ["fiziologichno aktivni rechovini"] = "Fiziol. Akt. Rechovini",
  ["fiziologichnyi zhurnal"] = "Fiziol. Zh.",
  ["fiziologiia cheloveka"] = "Fiziol Cheloveka",
  ["fiziologiya i biokhimiya kul’turnykh rastenii"] = "Fiziol. Biokhim. Kul’t. Rast.",
  ["fiziologiya rastenii (moscow)"] = "Fiziol. Rast. (Moscow)",
  ["fiziolohichnyi zhurnal"] = "Fiziol Zh",
  ["fiziolohichnyi zhurnal (kiev, ukraine : 1994)"] = "Fiziol Zh",
  ["fiziolohichnyĭ zhurnal"] = "Fiziol Zh",
  ["fiziolohichnyĭ zhurnal (kiev, ukraine : 1994)"] = "Fiziol Zh",
  ["fj-lmi subseries"] = "FJ-LMI Subser.",
  ["flambeau (bakwanga, congo)"] = "Flambeau",
  ["flame retardancy and thermal stability of materials"] = "Flame Retard. Therm. Stab. Mater.",
  ["flatchem chemistry of flat materials"] = "FlatChem Chem. Flat Mater.",
  ["flavins and flavoproteins"] = "Flavins Flavoproteins",
  ["flavour and fragance journal"] = "Flavour Fragance J.",
  ["flavour and fragrance journal"] = "Flavour Fragr J",
  ["flexible and printed electronics"] = "Flex. Print. Electron.",
  ["flexible services and manufacturing journal"] = "Flex Serv Manuf J",
  ["flight safety digest (arlington, va. : 1988)"] = "Flight Saf Dig",
  ["flinders journal of history and politics"] = "Flinders J Hist Polit",
  ["flintknappers' exchange"] = "Flintknappers Exch",
  ["flintknappers’ exchange"] = "Flintknappers Exch",
  ["flora (ankara, turkey)"] = "Flora",
  ["flora oder allgemeine botanische zeitung"] = "Flora Allg. Bot. Ztg.",
  ["flora oder allgemeine botanische zeitung. abt. a, physiologie und biochemie"] = "Flora Allg. Bot. Ztg. Abt. A, Physiol. Biochem.",
  ["flora oder allgemeine botanische zeitung. abt. b, morphologie und geobotanik"] = "Flora Allg. Bot. Ztg. Abt. B, Morphol. Geobot.",
  ["florence nightingale journal of nursing"] = "Florence Nightingale J Nurs",
  ["florentia iliberritana. revista de estudios de antigüedad clásica"] = "FlorIl",
  ["florentia. studi di archeologia"] = "Florentia",
  ["florida dental journal"] = "Fla Dent J",
  ["florida entomoloigist"] = "Fla. Entomolog.",
  ["florida health notes"] = "Fla Health Notes",
  ["florida journal of environmental health"] = "Fla J Environ Health",
  ["florida journal of international law"] = "Fla J Int Law",
  ["florida law review"] = "Fla Law Rev",
  ["florida nurse"] = "Fla. Nurse",
  ["florida optometrist"] = "Fla Optom",
  ["florida public health review"] = "Fla Public Health Rev",
  ["florida scientist"] = "Fla. Sci.",
  ["florida state university law review. florida state university. college of law"] = "Fla State Univ Law Rev",
  ["florida supplement : decisions of lower courts of record and state commissions"] = "Fla Suppl",
  ["floristische rundbriefe"] = "Florist. Rd.br.",
  ["flow measurement and instrumentation"] = "Flow Meas. Instrum.",
  ["flow measurement and instrumentation : fmi"] = "Flow Meas Instrum",
  ["flow, turbulence and combustion"] = "Flow Turbul. Combust.",
  ["flow: applications of fluid mechanics"] = "Flow: Appl. Fluid Mech.",
  ["fluctuation and noise letters"] = "Fluctuation Noise Lett.",
  ["fluctuations and noise letters"] = "Fluctuation Noise Lett.",
  ["fluegers archiv fur die gesamte physiologie des menschen und der tiere"] = "Pfluegers Arch. Gesame Physiol. Menschen Tiere",
  ["fluid dynamics"] = "Fluid Dyn.",
  ["fluid dynamics & materials processing"] = "Fluid Dyn. Mater. Process.",
  ["fluid dynamics research"] = "Fluid Dyn. Res.",
  ["fluid dynamics research. an international journal"] = "Fluid Dyn. Res.",
  ["fluid mechanics and its applications"] = "Fluid Mech. Appl.",
  ["fluid mechanics research"] = "Fluid Mech. Res.",
  ["fluid phase equilibria"] = "Fluid Phase Equilib.",
  ["fluid/particle separation journal"] = "Fluid/Part. Sep. J.",
  ["fluids and barriers of the cns"] = "Fluids Barriers CNS",
  ["flussiges obst"] = "Fluss. Obst",
  ["flying safety (washington, d.c. : 1981)"] = "Flying Saf",
  ["fmc : formación médica continuada en atención primaria"] = "FMC",
  ["fnib info"] = "FNIB Info",
  ["fnib: organe de la federation nationale des infirmier(e)s belges"] = "FNIB",
  ["fo. facultad de odontologia"] = "FO",
  ["fo: facultad de odontologia"] = "FO",
  ["focus (american psychiatric publishing)"] = "Focus (Am Psychiatr Publ)",
  ["focus (gaithersburg, md.)"] = "Focus",
  ["focus (san francisco, calif.)"] = "Focus",
  ["focus mda : the official publication of the missouri dental association"] = "Focus MDA",
  ["focus on aacn"] = "Focus AACN",
  ["focus on alternative and complementary therapies : fact"] = "Focus Altern Complement Ther",
  ["focus on autism and other developmental disabilities"] = "Focus Autism Other Dev Disabl",
  ["focus on catalysts"] = "Focus Catal.",
  ["focus on computer graphics"] = "Focus Comput. Graph.",
  ["focus on critical care"] = "Focus Crit Care",
  ["focus on critical care / american association of critical-care nurses"] = "Focus Crit Care",
  ["focus on exceptional children"] = "Focus Except Child",
  ["focus on gender"] = "Focus Gend",
  ["focus on ohio dentistry"] = "Focus Ohio Dent",
  ["focus on pigments"] = "Focus Pigm.",
  ["focus on polyvinyl chloride"] = "Focus Polyvinyl Chloride",
  ["focus on powder coatings"] = "Focus Powder Coat.",
  ["focus on surfactants"] = "Focus Surfactants",
  ["focus series in digital signal and image processing"] = "Focus Ser. Digit. Signal Image Process.",
  ["focus series in finance, business and management"] = "Focus Ser. Finance Bus. Manag.",
  ["focus. automation and control series"] = "Focus Autom. Control Ser.",
  ["focus: technical cooperation. focus. cooperacion tecnica. focus. cooperation technique"] = "Focus Tech Coop",
  ["focus: technical cooperation. focus. cooperación técnica. focus. coopération technique"] = "Focus Tech Coop",
  ["fogg art museum. acquisitions"] = "FoggArtMusAcqu",
  ["fogorvosi szemle"] = "Fogorv Sz",
  ["foi et langage"] = "Foi Lang",
  ["folding & design"] = "Fold Des",
  ["folding and design"] = "Fold Des.",
  ["foldrajzi ertesito"] = "Foldr Ert",
  ["folha medica"] = "Folha Med",
  ["folia allergologica"] = "Folia Allergol (Roma)",
  ["folia archaeologica"] = "FolA",
  ["folia biologica"] = "Folia Biol (Praha)",
  ["folia biotheoretica"] = "Folia Biotheor",
  ["folia cardiologica"] = "Folia Cardiol",
  ["folia clinica et biologica"] = "Folia Clin Biol (Sao Paulo)",
  ["folia clinica internacional"] = "Folia Clin. Int. (Barc.)",
  ["folia clínica internacional"] = "Folia Clin Int (Barc)",
  ["folia cryptogamica estonica"] = "Folia Cryptogam Est",
  ["folia endocrinologica"] = "Folia Endocrinol",
  ["folia endocrinologica; mensile di incretologia e incretoterapia"] = "Folia Endocrinol Mens Incretologia Incretoterapia",
  ["folia entomologica hungarica (budapest, hungary : 1970)"] = "Folia Entomol Hung (1970)",
  ["folia facultatis scientiarium naturalium universitatis masarykianae brunensis"] = "Folia Fac. Sci. Natur. Univ. Masaryk. Brun. Math.",
  ["folia forestalia polonica"] = "Folia For. Pol.",
  ["folia forestalia polonica. series a, forestry"] = "Folia For Pol Ser A For",
  ["folia geobotanica"] = "Folia Geobot",
  ["folia geobotanica et phytotaxonomica"] = "Folia Geobot. Phytotaxon.",
  ["folia geographica. series geographica-oeconomica"] = "Folia Geogr Ser Geogr Oecon",
  ["folia geographica. series geographica-oeconomica / polska akademia nauk, oddzial w krakowie, komisja nauk geograficznych"] = "Folia Geogr Ser Geogr Oecon",
  ["folia gynaecologica"] = "Folia Gynaecol (1908)",
  ["folia haematologica (leipzig, germany : 1928)"] = "Folia Haematol Int Mag Klin Morphol Blutforsch",
  ["folia haematologica: internationales magazin fur blutforschung"] = "Folia Haematol. (Frankf.)",
  ["folia haematologica: internationales magazin fur klinische und morphologische blutforschung"] = "Folia Haematol. Int. Mag. Klin. Morphol. Blutforsch.",
  ["folia haematologica; internationales magazin fur blutforschung"] = "Folia Haematol (Frankf)",
  ["folia haematologica; internationales magazin für blutforschung"] = "Folia Haematol (Frankf)",
  ["folia hereditaria et pathologica"] = "Folia Hered Pathol (Milano)",
  ["folia histochemica et cytobiologica"] = "Folia Histochem Cytobiol",
  ["folia histochemica et cytobiologica / polish academy of sciences, polish histochemical and cytochemical society"] = "Folia Histochem Cytobiol",
  ["folia histochemica et cytochemica"] = "Folia Histochem Cytochem (Krakow)",
  ["folia humanistica"] = "Folia Humanist",
  ["folia malacologica"] = "Folia Malacol",
  ["folia mathematica"] = "Folia Math.",
  ["folia medica"] = "Folia Med (Plovdiv)",
  ["folia medica cracoviensia"] = "Folia Med Cracov",
  ["folia medica neerlandica"] = "Folia Med Neerl",
  ["folia medica neerlandica. additamenta"] = "Additamenta Folia Med Neerl",
  ["folia medica. folia medica (naples, italy)"] = "Folia Med (Napoli)",
  ["folia mendeliana"] = "Folia Mendeliana",
  ["folia microbiologica"] = "Folia Microbiol (Praha)",
  ["folia microbiologica (dordrecht, netherlands)"] = "Folia Microbiol. (Dordrecht, Neth.)",
  ["folia microbiologica (prague, czech republic)"] = "Folia Microbiol. (Prague, Czech Repub.)",
  ["folia morphologica"] = "Folia Morphol (Warsz)",
  ["folia morphologica (poland)"] = "Folia Morphol. (Warsz.)",
  ["folia morphologica (prague)"] = "Folia Morphol. (Praha)",
  ["folia neuropathologica"] = "Folia Neuropathol",
  ["folia neuropathologica / association of polish neuropathologists and medical research centre, polish academy of sciences"] = "Folia Neuropathol",
  ["folia oecologica"] = "Folia Oecol.",
  ["folia ophthalmologica"] = "Fol Ophthalmol",
  ["folia orientalia"] = "Folia Orient",
  ["folia parasitologica"] = "Folia Parasitol (Praha)",
  ["folia phoniatrica"] = "Folia Phoniatr (Basel)",
  ["folia phoniatrica et logopaedica : official organ of the international association of logopedics and phoniatrics (ialp)"] = "Folia Phoniatr Logop",
  ["folia phoniatrica et logopedica"] = "Folia Phoniatr. Logop.",
  ["folia primatologica"] = "Folia Primatol.",
  ["folia primatologica; international journal of primatology"] = "Folia Primatol (Basel)",
  ["folia psychiatrica"] = "Folia Psychiatr",
  ["folia psychiatrica et neurologica japonica"] = "Folia Psychiatr Neurol Jpn",
  ["folia psychiatrica, neurologica et neurochirurgica neerlandica"] = "Folia Psychiatr Neurol Neurochir Neerl",
  ["folia stomatologica"] = "Folia Stomatol",
  ["folia veterinaria"] = "Folia Vet",
  ["folia veterinaria latina"] = "Folia Vet Lat",
  ["folia zoologica"] = "Folia Zool Brno",
  ["folk (kobenhavn)"] = "Folk",
  ["folk (københavn)"] = "Folk",
  ["folklore americano"] = "Folkl Am",
  ["folklore brabancon"] = "Folkl Brabancon",
  ["folklore brabançon"] = "Folkl Brabancon",
  ["folklore de champagne"] = "Folkl Champagne",
  ["folklore suisse"] = "Folkl Suisse",
  ["folli publications on logic, language and information"] = "FoLLI Publ. Log. Lang. Inf.",
  ["fonaments. prehistòria i mon antic als paisos catalans"] = "Fonaments",
  ["fondamenti. rivista quadrimestrale di cultura"] = "Fondamenti",
  ["fondation louis de broglie"] = "Ann. Fond. Louis de Broglie",
  ["fondation louis de broglie. annales"] = "Ann. Fond. Louis De Broglie",
  ["fondazione c"] = "Fond. C.I.M.E.",
  ["fondazione cime/cime foundation subseries"] = "Fond. CIME/CIME Found. Subser.",
  ["fonderie fondeur d’aujourd’hui"] = "Fonderie Fondeur Aujourd’hui",
  ["fontes archaeologici posnanienses"] = "FontAPos",
  ["fontes. rivista di filologia, iconografia e storia della tradizione classica"] = "Fontes",
  ["fonti per la storia della sanita : collana del ciso veneto patrocinata dalla regione veneto"] = "Fonti Stor Sanita",
  ["fonti per la storia della sanità : collana del ciso veneto patrocinata dalla regione veneto"] = "Fonti Stor Sanita",
  ["food & foods ingredients journal of japan"] = "Food Foods Ingredients J. Jpn.",
  ["food & foodways"] = "Food Foodways",
  ["food & function"] = "Food Funct",
  ["food & nutrition research"] = "Food Nutr Res",
  ["food additives & contaminants"] = "Food Addit. Contam.",
  ["food additives & contaminants, part a"] = "Food Addit. Contam., Part A",
  ["food additives & contaminants, part b: surveillance"] = "Food Addit. Contam., Part B",
  ["food additives & contaminants. part a, chemistry, analysis, control, exposure & risk assessment"] = "Food Addit Contam Part A Chem Anal Control Expo Risk Assess",
  ["food additives & contaminants. part b, surveillance"] = "Food Addit Contam Part B Surveill",
  ["food additives and contaminants"] = "Food Addit Contam",
  ["food additives and contaminants, part a: chemistry analysis control exposure & risk assessment"] = "Food Addit. Contam., Part A",
  ["food additives and contaminants, part b: surveillance"] = "Food Addit. Contam., Part B",
  ["food analytical methods"] = "Food Anal. Methods",
  ["food and agricultural immunology"] = "Food Agric Immunol",
  ["food and bioprocess technology"] = "Food Bioprocess Technol.",
  ["food and bioproducts processing"] = "Food Bioprod. Process.",
  ["food and chemical toxicology"] = "Food Chem. Toxicol.",
  ["food and chemical toxicology : an international journal published for the british industrial biological research association"] = "Food Chem Toxicol",
  ["food and cosmetics toxicology"] = "Food Cosmet. Toxicol.",
  ["food and drug law journal"] = "Food Drug Law J",
  ["food and energy security"] = "Food Energy Secur.",
  ["food and environmental virology"] = "Food Environ Virol",
  ["food and foodways"] = "Food Foodways",
  ["food and nutrition"] = "Food Nutr (Roma)",
  ["food and nutrition bulletin"] = "Food Nutr Bull",
  ["food and nutrition sciences"] = "Food Nutr Sci",
  ["food and public health"] = "Food Public Health",
  ["food and waterborne parasitology"] = "Food Waterborne Parasitol",
  ["food bioengineering"] = "Food Bioeng.",
  ["food biophysics"] = "Food Biophys.",
  ["food bioscience"] = "Food Biosci",
  ["food biotechnology"] = "Food Biotechnol",
  ["food chemistry"] = "Food Chem",
  ["food chemistry advances"] = "Food Chem. Adv.",
  ["food chemistry: molecular sciences"] = "Food Chem.: Mol. Sci.",
  ["food chemistry: x"] = "Food Chem X",
  ["food control"] = "Food Control",
  ["food digestion"] = "Food Dig",
  ["food digestion: research and current opinion"] = "Food Digest.: Res. Curr. Opin.",
  ["food engineering reviews"] = "Food Eng. Rev.",
  ["food ethics"] = "Food Ethics",
  ["food frontiers"] = "Food Front",
  ["food hydrocolloids"] = "Food Hydrocolloids",
  ["food industries"] = "Food Ind",
  ["food management"] = "Food Manage",
  ["food materials and equipment"] = "Food Mater Equip",
  ["food microbiology"] = "Food Microbiol",
  ["food modelling journal"] = "Food Modell. J.",
  ["food packaging and shelf life"] = "Food Packag Shelf Life",
  ["food policy"] = "Food Policy",
  ["food production, processing and nutrition"] = "Food Prod. Process. Nutr.",
  ["food protection trends"] = "Food Prot Trends",
  ["food quality"] = "Food Qual.",
  ["food quality and preference"] = "Food Qual. Preference",
  ["food quality and safety"] = "Food Qual. Saf.",
  ["food research"] = "Food Res",
  ["food research institute studies"] = "Food Res. Inst. Stud.",
  ["food research institute studies (stanford, calif. : 1975)"] = "Food Res Inst Stud",
  ["food research institute studies in agricultural economics, trade, and development"] = "Food Res Inst Stud Agric Econ Trade Dev",
  ["food research international"] = "Food Res. Int.",
  ["food research international (ottawa, ont.)"] = "Food Res Int",
  ["food reviews international"] = "Food Rev. Int.",
  ["food safety (tokyo, japan)"] = "Food Saf (Tokyo)",
  ["food science & nutrition"] = "Food Sci Nutr",
  ["food science and agricultural chemistry"] = "Food Sci Agric Chem",
  ["food science and applied biotechnology"] = "Food Sci. Appl. Biotechnol.",
  ["food science and biotechnology"] = "Food Sci Biotechnol",
  ["food science and human wellness"] = "Food Sci. Hum. Wellness",
  ["food science and nutrition cases"] = "Food Sci. Nutr. Cases",
  ["food science and technology"] = "Food Sci. Technol.",
  ["food science and technology brazil"] = "Food Sci. Technol. Braz.",
  ["food science and technology bulletin"] = "Food Sci Technol Bull",
  ["food science and technology international"] = "Food Sci. Technol. Int.",
  ["food science and technology international = ciencia y tecnología de los alimentos internacional"] = "Food Sci Technol Int",
  ["food science and technology research"] = "Food Sci. Technol. Res.",
  ["food science of animal resources"] = "Food Sci Anim Resour",
  ["food security"] = "Food Secur",
  ["food structure"] = "Food Struct.",
  ["food studies"] = "Food Stud",
  ["food technology"] = "Food Technol",
  ["food technology and biotechnology"] = "Food Technol Biotechnol",
  ["food webs"] = "Food Webs",
  ["food, culture, & society"] = "Food Cult Soc",
  ["food, drug, cosmetic law journal"] = "Food Drug Cosmet Law J",
  ["food, drug, cosmetic law quarterly"] = "Food Drug Cosmet Law Q",
  ["food, drug, cosmetic, and medical device law digest : a publication of the food, drug, and cosmetic law section"] = "Food Drug Cosmet Med Device Law Dig",
  ["foodborne pathogens and disease"] = "Foodborne Pathog Dis",
  ["foods (basel, switzerland)"] = "Foods",
  ["foot & ankle"] = "Foot Ankle",
  ["foot & ankle international"] = "Foot Ankle Int",
  ["foot & ankle international / american orthopaedic foot and ankle society [and] swiss foot and ankle society"] = "Foot Ankle Int",
  ["foot & ankle orthopaedics"] = "Foot Ankle Orthop",
  ["foot & ankle specialist"] = "Foot Ankle Spec",
  ["foot (edinburgh, scotland)"] = "Foot (Edinb)",
  ["foot and ankle"] = "Foot Ankle",
  ["foot and ankle clinics"] = "Foot Ankle Clin",
  ["foot and ankle international"] = "Foot Ankle Int.",
  ["foot and ankle surgery : official journal of the european society of foot and ankle surgeons"] = "Foot Ankle Surg",
  ["footwear science"] = "Footwear Sci.",
  ["for your information (institute of health record & information management)"] = "For Your Inf (Inst Health Rec Inf Manag)",
  ["forage & grazinglands"] = "Forage Grazinglands",
  ["forced migration review"] = "Forced Migr Rev",
  ["forces in mechanics"] = "Forces Mech.",
  ["ford foundation letter. ford foundation"] = "Lett Ford Found",
  ["ford foundation report (new york, n.y. : 1992)"] = "Ford Found Rep",
  ["fordham international law journal"] = "Fordham Int Law J",
  ["fordham law review"] = "Fordham Law Rev",
  ["fordham law review / edited by fordham law students"] = "Fordham Law Rev",
  ["foreign affairs"] = "Foreign Aff.",
  ["foreign affairs (council on foreign relations)"] = "Foreign Aff",
  ["foreign language annals"] = "Foreign Lang Ann",
  ["foreign policy"] = "Foreign Policy",
  ["forensic chemistry"] = "Forensic Chem.",
  ["forensic chemistry (amsterdam, netherlands)"] = "Forensic Chem",
  ["forensic engineering"] = "Forensic Eng.",
  ["forensic genomics"] = "Forensic Genomics",
  ["forensic imaging"] = "Forensic Imaging",
  ["forensic medicine and anatomy research"] = "Forensic Med Anat Res",
  ["forensic science"] = "Forensic Sci",
  ["forensic science & addiction research"] = "Forensic Sci Addict Res",
  ["forensic science and criminology"] = "Forensic Sci Criminol",
  ["forensic science international"] = "Forensic Sci Int",
  ["forensic science international supplement series"] = "Forensic Sci. Int. Suppl. Ser.",
  ["forensic science international. genetics"] = "Forensic Sci Int Genet",
  ["forensic science international. genetics supplement series"] = "Forensic Sci Int Genet Suppl Ser",
  ["forensic science international. mind and law"] = "Forensic Sci Int Mind Law",
  ["forensic science international. synergy"] = "Forensic Sci Int",
  ["forensic science international: digital investigation"] = "Forensic Sci. Int.: Digital Invest.",
  ["forensic science international: genetics"] = "Forensic Sci. Int. Genet.",
  ["forensic science international: genetics supplement series"] = "Forensic Sci. Int.: Genet. Suppl. Ser.",
  ["forensic science international: mind and law"] = "Forensic Sci. Int.: Mind Law",
  ["forensic science international: reports"] = "Forensic Sci. Int.: Rep.",
  ["forensic science international: synergy"] = "Forensic Sci. Int.: Synergy",
  ["forensic science policy and management"] = "Forensic Sci. Policy Manage.",
  ["forensic science progress"] = "Forensic Sci. Prog.",
  ["forensic science review"] = "Forensic Sci Rev",
  ["forensic science, medicine and pathology"] = "Forensic Sci. Med. Pathol.",
  ["forensic science, medicine, and pathology"] = "Forensic Sci Med Pathol",
  ["forensic sciences research"] = "Forensic Sci Res",
  ["forensic toxicology"] = "Forensic Toxicol.",
  ["forensische psychiatrie, psychologie, kriminologie"] = "Forensische Psychiatr. Psychol. Kriminol.",
  ["foresight (colchester, vt.)"] = "Foresight (Colch)",
  ["foresight (los angeles, calif.)"] = "Foresight",
  ["forest & conservation history"] = "For. Conserv. Hist.",
  ["forest and landscape research"] = "For. Landsc. Res.",
  ["forest ecology and management"] = "For Ecol Manage",
  ["forest ecosystems"] = "For Ecosyst",
  ["forest genetics"] = "For. Genet.",
  ["forest genetics : fg : international journal of forest genetics"] = "For Genet",
  ["forest industries"] = "For. Ind.",
  ["forest insect & disease leaflet"] = "For. Insect Dis. Leafl.",
  ["forest pathology"] = "For Pathol",
  ["forest policy and economics"] = "For Policy Econ",
  ["forest products journal"] = "For Prod J",
  ["forest science"] = "For. Sci.",
  ["forest science and practice"] = "For. Sci. Pract.",
  ["forest science and technology"] = "Forest Sci Technol",
  ["forest snow and landscape research"] = "For. Snow Landsc. Res.",
  ["forest systems"] = "For Syst",
  ["forestry (london, england)"] = "Forestry (Lond)",
  ["forestry abstracts"] = "For. Abstr.",
  ["forestry canda modeling working group : proceedings of the fifth annual workshop"] = "For. Can. Nor.",
  ["forestry chronicle"] = "For. Chron.",
  ["forestry chronicle, the"] = "For. Chron.",
  ["forestry ideas"] = "For Ideas",
  ["forestry research"] = "For. Res.",
  ["forestry studies"] = "For. Stud.",
  ["forestry studies in china"] = "For Stud China",
  ["forestry systems"] = "For. Syst.",
  ["forestry technical report"] = "For. Tech. Rep.",
  ["forests, trees and livelihoods"] = "Forests Trees Livelihoods",
  ["forlimpopoli. documenti e studi"] = "Forlimpopoli",
  ["formal aspects of computing"] = "Formal Aspects Comput.",
  ["formal aspects of computing: applicable formal methods"] = "Formal Aspects Comput.: Appl. Formal Methods",
  ["formal methods in system design"] = "Form Methods Syst Des",
  ["formazione e aggiornamento in matematica degli insegnanti"] = "Form. Aggiorn. Mat. Insegn.",
  ["formosan entomologist"] = "Formos Entomol",
  ["formosan journal of medical humanities"] = "Formos J Med Humanit",
  ["formulary (cleveland, ohio)"] = "Formulary",
  ["fornvännen. tidskrift för svensk antikvarisk forskning"] = "Fornvaennen",
  ["foro internacional"] = "Foro Int",
  ["foro mundial de la salud"] = "Foro Mund Salud",
  ["forschende komplementarmedizin"] = "Forsch. Komplementarmed.",
  ["forschende komplementarmedizin (2006)"] = "Forsch Komplement Med (2006)",
  ["forschende komplementarmedizin und klassische naturheilkunde"] = "Forsch. Komplementarmed. Klass. Naturheilkd.",
  ["forschende komplementärmedizin"] = "Forsch Komplementarmed",
  ["forschende komplementärmedizin (2006)"] = "Forsch Komplementmed",
  ["forschende komplementärmedizin und klassische naturheilkunde = research in complementary and natural classical medicine"] = "Forsch Komplementarmed Klass Naturheilkd",
  ["forschung auf dem gebiete des ingenieurwesens"] = "Forsch. Geb. Ingenieurwes.",
  ["forschung im ingenieurwesen"] = "Forsch. Ingenieurwes.",
  ["forschung und fortbildung in der chirurgie des bewegungsapparates"] = "Forsch Fortbild Chir Bewegungsappar",
  ["forschungen des instituts fur realienkunde des mittelalters und der fruhen neuzeit. diskussionen und materialien"] = "Forsch. Inst. Realienkd. Mittelalt. Fruhen Neuzeit Diskuss. Mater.",
  ["forschungen in augst"] = "FiA",
  ["forschungen in ephesos"] = "FiE",
  ["forschungen und berichte zur vor- und frühgeschichte in baden-württemberg"] = "FBerBadWuert",
  ["forschungen und berichte. staatliche museen zu berlin"] = "FuB",
  ["forschungen und fortschritte"] = "Forsch. Fortschr.",
  ["forschungen und fortschritte; nachrichtenblatt der deutschen wissenschaft und technik"] = "Forsch Fortschr",
  ["forschungen und materialien zur deutschen aufklarung (fmda). abteilung ii: monographien"] = "FMDA Abt. II: Monogr.",
  ["forschungen und materialien zur deutschen aufklärung (fmda)"] = "FMDA Abt. I: Texte",
  ["forschungen zur allgemeinen und vergleichenden archäologie"] = "FAVA",
  ["forschungen zur archäologie aussereuropäischer kulturen"] = "FAAK",
  ["forschungen zur volks- und landeskunde"] = "Forsch Volks Landeskd",
  ["forschungsbericht forstliche forschungsanstalt münchen"] = "Forsch.ber. forstl. Forsch.anst. Münch.",
  ["forschungsberichte des atb"] = "Research reports of ATB|Forschungsberichte ATB",
  ["forschungsberichte des atb = research reports of atb"] = "Forschungsberichte ATB",
  ["forschungsergebnisse zur informatik"] = "Forsch.ergeb. Inform.",
  ["forschungsjournal neue soziale bewegungen"] = "Forsch J Neue Soz Beweg",
  ["forst und holz"] = "Forst Holz",
  ["forst- und holzwirt, der"] = "Forst- Holzwirt",
  ["forstlich-naturwissenschaftliche zeitschrift"] = "Forstl.-nat.wiss. Z.",
  ["forstliche forschungsberichte münchen"] = "Forstl. Forsch.ber. Münch.",
  ["forstliche rundschau"] = "Forstl. Rundsch.",
  ["forstliche umschau"] = "Forstl. Umsch.",
  ["forstlige forsøgsvaesen i danmark, det"] = "Forstl. forsøgsvaes. Dan.",
  ["forstschutz bulletin"] = "Forstschutz Bull.",
  ["forsttechnische informationen"] = "Forsttech. Inf.",
  ["forstwissenschaftliches centralblatt"] = "Forstwiss. Centralbl.",
  ["fort concho report"] = "Fort Concho Rep",
  ["fortnightly review of the chicago dental society"] = "Fortn. Rev. Chic. Dent. Soc.",
  ["fortpflanzung, besamung, und aufzucht der haustiere"] = "Fortpflanz Besamung Aufzucht Haustiere",
  ["fortran numerical recipes"] = "Fortran Numer. Recipes",
  ["fortschriftsberiche der deutschen keramischen gesellschaft"] = "Fortschrittsber. Dtschen Keram. Ges.",
  ["fortschriftsberichte der deutschen keramischen gesellschaft"] = "Fortschrittsber. Dtschen Keram. Ges.",
  ["fortschritte auf dem gebiete der rontgenstrahlen"] = "Fortschr. Geb. Rontgenstr.",
  ["fortschritte auf dem gebiete der rontgenstrahlen und der neuen bildgebenden verfahren. erganzungsband"] = "Fortschr. Geb. Rontgenstrahlen Neuen Bildgeb. Verfahr. Erganzungsbd.",
  ["fortschritte auf dem gebiete der rontgenstrahlen und der nuklearmedizin"] = "Fortschr. Geb. Rontgenstr. Nuklearmed.",
  ["fortschritte auf dem gebiete der rontgenstrahlen und der nuklearmedizin. erganzungsband"] = "Fortschr. Geb. Rontgenstrahlen Nuklearmed. Erganzungsbd.",
  ["fortschritte auf dem gebiete der rontgenstrahlen und nuklearmedizine"] = "Fortschr. Geb. Rontgenstr. Nuklearmed.",
  ["fortschritte auf dem gebiete der röntgenstrahlen"] = "Fortschr Geb Rontgenstr",
  ["fortschritte auf dem gebiete der röntgenstrahlen und der neuen bildgebenden verfahren. ergänzungsband"] = "Fortschr Geb Rontgenstrahlen Neuen Bildgeb Verfahr Erganzungsbd",
  ["fortschritte auf dem gebiete der röntgenstrahlen und der nuklearmedizin"] = "Fortschr Geb Rontgenstr Nuklearmed",
  ["fortschritte auf dem gebiete der röntgenstrahlen und der nuklearmedizin. ergänzungsband"] = "Fortschr Geb Rontgenstrahlen Nuklearmed Erganzungsbd",
  ["fortschritte auf dem gebiete der röntgenstrahlen. ergänzungsband"] = "Fortschr Geb Rontgenstrahlen Erganzungsbd",
  ["fortschritte der allergielehre"] = "Fortschr Allerg",
  ["fortschritte der andrologie"] = "Fortschr Androl",
  ["fortschritte der angewandten radioisotopie und grenzgebiete"] = "Fortschr Angew Radioisot Grenzgeb",
  ["fortschritte der arzneimittelforschung (progress in drug research)"] = "Fortschr. Arzneimittelforsch.",
  ["fortschritte der arzneimittelforschung. progress in drug research. progres des recherches pharmaceutiques"] = "Fortschr Arzneimittelforsch",
  ["fortschritte der arzneimittelforschung. progress in drug research. progrès des recherches pharmaceutiques"] = "Fortschr Arzneimittelforsch",
  ["fortschritte der chemie organischer naturstoffe"] = "Fortschr. Chem. Org. Naturst.",
  ["fortschritte der chemie organischer naturstoffe = progress in the chemistry of organic natural products. progrès dans la chimie des substances organiques naturelles"] = "Fortschr Chem Org Naturst",
  ["fortschritte der chemie organischer naturstoffe. progress in the chemistry of organic natural products. progres dans la chimie des substances organiques naturelles"] = "Fortschr Chem Org Naturst",
  ["fortschritte der chemie, physik und physikalischen chemie"] = "Fortschr. Chem. Phys. Phys. Chem.",
  ["fortschritte der chemischen forschung"] = "Fortschr. Chem. Forsch.",
  ["fortschritte der geburtshilfe und gynakologie"] = "Fortschr. Geburtshilfe Gynakol.",
  ["fortschritte der geburtshilfe und gynäkologie"] = "Fortschr Geburtshilfe Gynakol",
  ["fortschritte der hals-, nasen-, ohrenheilkunde"] = "Fortschr. Hals. Nasen. Ohrenheilkd.",
  ["fortschritte der hals-nasen-ohrenheilkunde"] = "Fortschr Hals Nasen Ohrenheilkd",
  ["fortschritte der hochpolymoren-forschung"] = "Fortschr. Hochpolym. Forsch.",
  ["fortschritte der immunitatsforschung"] = "Fortschr. Immunitatsforsch.",
  ["fortschritte der immunitätsforschung"] = "Fortschr Immunitatsforsch",
  ["fortschritte der kardiologie"] = "Fortschr Kardiol",
  ["fortschritte der kiefer- und gesichts-chirurgie"] = "Fortschr Kiefer Gesichtschir",
  ["fortschritte der kieferorthopadie"] = "Fortschr. Kieferorthop.",
  ["fortschritte der kieferorthopädie"] = "Fortschr Kieferorthop",
  ["fortschritte der medizin"] = "Fortschr Med",
  ["fortschritte der medizin. monographie"] = "Fortschr Med Monogr",
  ["fortschritte der medizin. originalien"] = "Fortschr Med Orig",
  ["fortschritte der medizin. supplement"] = "Fortschr. Med. Suppl.",
  ["fortschritte der medizin. supplement : die kongressinformation fur die praxis"] = "Fortschr Med Suppl",
  ["fortschritte der medizin. supplement : die kongressinformation für die praxis"] = "Fortschr Med Suppl",
  ["fortschritte der mineralogie"] = "Fortschr. Mineral.",
  ["fortschritte der mineralogie, kristallograhie, und petrographie"] = "Fortschr. Mineral. Kristallog. Petrogr.",
  ["fortschritte der neurologie, psychiatrie und ihrer grenzgebiete"] = "Fortschr. Neurol. Psychiatr. Grenzgeb.",
  ["fortschritte der neurologie, psychiatrie, und ihrer grenzgebiete"] = "Fortschr Neurol Psychiatr Grenzgeb",
  ["fortschritte der neurologie-psychiatrie"] = "Fortschr Neurol Psychiatr",
  ["fortschritte der ophthalmologie"] = "Fortschr. Ophthalmol.",
  ["fortschritte der ophthalmologie : zeitschrift der deutschen ophthalmologischen gesellschaft"] = "Fortschr Ophthalmol",
  ["fortschritte der physik"] = "Fortschr. Phys.",
  ["fortschritte der physik. progress of physics"] = "Fortschr. Phys.",
  ["fortschritte der physikalischen chemie"] = "Fortschr. Phys. Chem.",
  ["fortschritte der psychosomatischen medizin. advances in psychosomatic medicine"] = "Fortschr Psychosom Med",
  ["fortschritte der teerfarben fabrikation und verwandter industriezweige"] = "Fortschr. Teerfarbenfabr. Verw. Industriezweige",
  ["fortschritte der verfahrenstechnik"] = "Fortschr. Verfahrenstech.",
  ["fortschritte der verfahrenstechnik abteilung a: grundlagen der verfahrenstechnik"] = "Fortschr. Verfahrenstech. Abt. A",
  ["fortschritte der verfahrenstechnik abteilung b: mechanische verfahrenstechnik"] = "Fortschr. Verfahrenstech. Abt. B",
  ["fortschritte der verfahrenstechnik abteilung c: thermische verfahrenstechnik"] = "Fortschr. Verfahrenstech. Abt. C",
  ["fortschritte der verfahrenstechnik abteilung d: reaktionstechnik"] = "Fortschr. Verfahrenstech. Abt. D",
  ["fortschritte der verfahrenstechnik abteilung e: planung und betrieb von anlagen"] = "Fortschr. Verfahrenstech. Abt. E",
  ["fortschritte der verfahrenstechnik abteilung f: spezielle technologien"] = "Fortschr. Verfahrenstech. Abt. F",
  ["fortschritte der verhaltensforschung"] = "Fortschr Verhaltensforsch",
  ["fortschritte der zoologie"] = "Fortschr Zool",
  ["fortschritte im strahlenschutz"] = "Fortschr. Strahlenschutz",
  ["fortschritte in der tierphysiologie und tierernahrung"] = "Fortschr. Tierphysiol. Tierernahr.",
  ["fortschritte in der tierphysiologie und tierernährung"] = "Fortschr Tierphysiol Tierernahr",
  ["fortschrittsberichte fur die landwirtschaft und nahrungsguterwirtschaft"] = "Fortschrittsber. Landwirtsch. Nahrungsguterwirtsch.",
  ["fortschrittsberichte uber kolloide und polymere"] = "Fortschrittsber. Kolloide Polym.",
  ["fortuna vitrea"] = "Fortuna Vitrea",
  ["forum (chicago, ill.)"] = "Forum",
  ["forum (genoa, italy)"] = "Forum (Genova)",
  ["forum (washington, d.c. : 1977)"] = "Forum (Wash)",
  ["forum for applied research and public policy"] = "Forum Appl Res Public Policy",
  ["forum for development studies"] = "Forum Dev Stud",
  ["forum for health economics & policy"] = "Forum Health Econ Policy",
  ["forum for interdisciplinary mathematics"] = "Forum Interdiscip. Math.",
  ["forum for social economics"] = "Forum Soc. Econ.",
  ["forum geometricorum. a journal on classical euclidean geometry and related areas"] = "Forum Geom.",
  ["forum implantologicum"] = "Forum Implantol",
  ["forum mathematicum"] = "Forum Math.",
  ["forum mondial de la sante"] = "Forum Mond Sante",
  ["forum mondial de la santé"] = "Forum Mond Sante",
  ["forum of clinical oncology"] = "Forum Clin Oncol",
  ["forum of mathematics pi"] = "Forum Math. Pi",
  ["forum of mathematics sigma"] = "Forum Math. Sigma",
  ["forum of mathematics. pi"] = "Forum Math. Pi",
  ["forum of mathematics. sigma"] = "Forum Math. Sigma",
  ["forum of nutrition"] = "Forum Nutr",
  ["forum on immunopathological diseases and therapeutics"] = "For Immunopathol Dis Therap",
  ["forum on medicine"] = "Forum Med",
  ["forum on public policy"] = "Forum Public Policy",
  ["forum statisticum"] = "Forum Stat",
  ["forum statisticum / verband schweizerischer statistischer amter"] = "Forum Stat",
  ["forum, qualitative social research"] = "Forum Qual Soc Res",
  ["forum. revue du groupe d’archéologie antique"] = "Forum",
  ["forum; revista invatamintului superior"] = "Forum Rev Invat Super",
  ["forêt méditerranéenne"] = "For. méditerr.",
  ["forêt suisse"] = "Forêt Suisse",
  ["forêt, la"] = "Forêt",
  ["forêts de france"] = "For. Fr.",
  ["forêts de france et action forestière"] = "For. Fr. action for.",
  ["fossil record"] = "Fossil Rec.",
  ["fottea (praha)"] = "Fottea (Praha)",
  ["fouilles de conimbriga. publiées sous la direction de j. alarcão et r. etienne"] = "FdC",
  ["fouilles de delphes"] = "FdD",
  ["fouilles de l’institut français d’archéologie orientale du caire"] = "FIFAO",
  ["fouilles de xanthos"] = "FdX",
  ["foundation news"] = "Found News",
  ["foundation news & commentary"] = "Found News Comment",
  ["foundation news and commentary"] = "Found. News Comment.",
  ["foundation of physics letters"] = "Found. Phys. Lett.",
  ["foundational questions in science"] = "Found. Quest. Sci.",
  ["foundations and trends in machine learning"] = "Found. Trends Mach. Learn.",
  ["foundations and trends in systems and control"] = "Found. Trends Syst. Control",
  ["foundations and trends${}^\\\\circledr$ in signal processing"] = "Found. Trends Signal Process.",
  ["foundations and trends${}^\\\\circledr$ in theoretical computer science"] = "Found. Trends Theor. Comput. Sci.",
  ["foundations and trends${}^circledr$ in signal processing"] = "Found. Trends Signal Process.",
  ["foundations and trends${}^circledr$ in theoretical computer science"] = "Found. Trends Theor. Comput. Sci.",
  ["foundations for undergraduate research in mathematics"] = "Found. Undergrad. Res. Math.",
  ["foundations in signal processing, communications and networking"] = "Found. Signal. Process. Commun. Netw.",
  ["foundations of chemistry"] = "Found. Chem.",
  ["foundations of communication and cognition"] = "Found. Comm. Cogn.",
  ["foundations of computational mathematics"] = "Found. Comput. Math.",
  ["foundations of computational mathematics (new york, n.y.)"] = "Found Comut Math",
  ["foundations of computational mathematics. the journal of the society for the foundations of computational mathematics"] = "Found. Comput. Math.",
  ["foundations of computing series"] = "Found. Comput. Ser.",
  ["foundations of data science"] = "Found. Data Sci.",
  ["foundations of engineering mechanics"] = "Found. Eng. Mech.",
  ["foundations of molecular modeling and simulation : select papers from fomms 2015. international conference on foundations of molecular modeling and simulation (6th : 2015 : mount hood, or.)"] = "Found Mol Model Simul (2015)",
  ["foundations of physics"] = "Found. Phys.",
  ["foundations of physics letters"] = "Found. Phys. Lett.",
  ["foundations of physics. an international journal devoted to the conceptual bases and fundamental theories of modern physics"] = "Found. Phys.",
  ["foundations of science"] = "Found. Sci.",
  ["foundations of science. the official journal of the association for foundations of science, language and cognition"] = "Found. Sci.",
  ["földrajzi értesítő"] = "Foldr Ert",
  ["fp essentials"] = "FP Essent",
  ["fpga. acm international symposium on field-programmable gate arrays"] = "FPGA",
  ["fpop bulletin"] = "Fpop Bull",
  ["fprc [reports]. great britain. flying personnel research committee"] = "Fprc",
  ["fra sundhedsstyrelsen"] = "Fra Sundhedsstyr",
  ["fractal : revista de psicologia"] = "Fractal (Niterói)",
  ["fractals and dynamics in mathematics, science, and the arts: theory and applications"] = "Fractals Dyn. Math. Sci. Arts Theory Appl.",
  ["fractals. complex geometry, patterns, and scaling in nature and society"] = "Fractals",
  ["fractional calculus & applied analysis"] = "Fract Calc Appl Anal",
  ["fractional calculus and applied analysis"] = "Fract. Calc. Appl. Anal.",
  ["fractional calculus and applied analysis. an international journal for theory and applications"] = "Fract. Calc. Appl. Anal.",
  ["fractional calculus in applied sciences and engineering"] = "Fract. Calc. Appl. Sci. Eng.",
  ["fractional differential calculus"] = "Fract. Differ. Calc.",
  ["fragmenta entomologica"] = "Fragm Entomol",
  ["fragmenta floristica et geobotanica polonica"] = "Fragm Florist Geobot Pol",
  ["fragmenta historicorum graecorum"] = "FHG",
  ["fragrance journal"] = "Fragance J.",
  ["france medecine"] = "Fr Med",
  ["frankfurter beitrage zur geschichte, theorie und ethik der medizin"] = "Frankf Beitr Gesch Theor Ethik Med",
  ["frankfurter hefte"] = "Frankf Hefte",
  ["frankfurter zeitschrift fur pathologie"] = "Frankf. Z. Pathol.",
  ["frankfurter zeitschrift für pathologie"] = "Frankf Z Pathol",
  ["franklin open"] = "Franklin Open",
  ["frater of psi omega"] = "Frater Psi Omega",
  ["free china review"] = "Free China Rev",
  ["free clinic research collective"] = "Free Clin Res Collect",
  ["free inquiry (buffalo, n.y.)"] = "Free Inq",
  ["free inquiry in creative sociology"] = "Free Inq Creat Sociol",
  ["free neuropathology"] = "Free Neuropathol",
  ["free radical biology & medicine"] = "Free Radic Biol Med",
  ["free radical biology and medicine"] = "Free Radical Biol. Med.",
  ["free radical research"] = "Free Radic Res",
  ["free radical research communications"] = "Free Radic Res Commun",
  ["free radicals and antioxidants"] = "Free Radicals Antioxid.",
  ["free world horizons"] = "Free World Horiz",
  ["freedom at issue"] = "Freedom Issue",
  ["freedom review"] = "Freedom Rev",
  ["freiberger forschungshefte a."] = "Freiberg. Forschungsh. A.",
  ["freiburger bodenkundliche abhandlungen"] = "Freibg. bodenkd. Abh.",
  ["freiburger forschungen zur medizingeschichte"] = "Freib Forsch Medizingesch",
  ["freiburger geographische hefte"] = "Freibg. Geogr. Hefte",
  ["freiburger schriften zur hydrologie"] = "Freibg. Schr. Hydrol.",
  ["freiburger studien zur frühen neuzeit"] = "Freibg. Stud. Frühen Neuzeit",
  ["freie zahnarzt"] = "Freie Zahnarzt",
  ["french colonial history"] = "Fr Colon Hist",
  ["french historical studies"] = "Fr Hist Stud",
  ["french history"] = "Fr Hist",
  ["french studies"] = "Fr Stud",
  ["fresenius environment bulletin"] = "Fresenius Environ. Bull.",
  ["fresenius environmental bulletin"] = "Fresenius Environ Bull",
  ["fresenius journal of analytical chemistry"] = "Fresenius J. Anal. Chem.",
  ["fresenius' journal of analytical chemistry"] = "Fresenius J Anal Chem",
  ["fresenius’ journal of analytical chemistry"] = "Fresenius J Anal Chem",
  ["freshwater biology"] = "Freshw Biol",
  ["freshwater crayfish"] = "Freshw Crayfish",
  ["freshwater mollusk biology and conservation"] = "Freshwater Mollusk Biol. Conserv.",
  ["freshwater mollusk biology and conservation : the journal of the freshwater mollusk conservation society"] = "Freshw Mollusk Biol Conserv",
  ["freshwater reviews"] = "Freshwater Rev.",
  ["freshwater reviews : a journal of the freshwater biological association"] = "Freshw Rev",
  ["freshwater science"] = "Freshwater Sci.",
  ["freshwater science (print)"] = "Freshw Sci",
  ["friends of the p.i. nixon medical historical library"] = "Friends P I Nixon Med Hist Libr",
  ["friends of women newsletter"] = "Friends Women Newsl",
  ["fritschiana : veröffentlichungen aus dem herbarium des instituts für botanik der karl-franzens-universität graz (gzu)"] = "Fritschiana (Graz)",
  ["friuli medico"] = "Friuli Med.",
  ["front lines"] = "Front Lines",
  ["front lines research"] = "Front Lines Res",
  ["fronteras (bogotá, colombia)"] = "Fronteras",
  ["frontier nursing service quarterly bulletin"] = "Front Nurs Serv Q Bull",
  ["frontier research in computation and mechanics of materials and biology"] = "Front. Res. Comput. Mech. Mater. Biol.",
  ["frontier science series"] = "Front. Sci. Ser.",
  ["frontiers for young minds"] = "Front Young Minds",
  ["frontiers in aerospace enginering"] = "Front. Aerosp. Eng.",
  ["frontiers in aging"] = "Front. Aging",
  ["frontiers in aging neuroscience"] = "Front Aging Neurosci",
  ["frontiers in aging series"] = "Front Aging Ser",
  ["frontiers in agronomy"] = "Front. Agron.",
  ["frontiers in allergy"] = "Front. Allergy",
  ["frontiers in amphibian and reptile science"] = "Front. Amphib. Reptile Sci.",
  ["frontiers in analytical science"] = "Front. Anal. Sci.",
  ["frontiers in anesthesiology"] = "Front. Anesthesiol.",
  ["frontiers in animal science"] = "Front. Anim. Sci.",
  ["frontiers in antibiotics"] = "Front. Antibiot.",
  ["frontiers in applied dynamical systems: reviews and tutorials"] = "Front. Appl. Dyn. Syst. Rev. Tutor.",
  ["frontiers in applied mathematics"] = "Frontiers Appl. Math.",
  ["frontiers in applied mathematics and statistics"] = "Front Appl Math Stat",
  ["frontiers in aquacult."] = "Front. Aquacult.",
  ["frontiers in aquatic physiology"] = "Front. Aquat. Physiol.",
  ["frontiers in arachnid science"] = "Front. Arachnid Sci.",
  ["frontiers in artif. intell. and applications"] = "Frontiers Artif. Intell. Appl.",
  ["frontiers in artificial intelligence"] = "Front Artif Intell",
  ["frontiers in artificial intelligence and applications"] = "Front. Artif. Intell. Appl.",
  ["frontiers in astronomy and space sciences"] = "Front. Astron. Space Sci.",
  ["frontiers in behavioral neuroscience"] = "Front Behav Neurosci",
  ["frontiers in big data"] = "Front Big Data",
  ["frontiers in bioengineering and biotechnology"] = "Front Bioeng Biotechnol",
  ["frontiers in bioinformatics"] = "Front. Bioinf.",
  ["frontiers in biology"] = "Front Biol (Beijing)",
  ["frontiers in biomaterials science"] = "Front. Biomater. Sci.",
  ["frontiers in bioscience"] = "Front. Biosci.",
  ["frontiers in bioscience (elite edition)"] = "Front Biosci (Elite Ed)",
  ["frontiers in bioscience (landmark edition)"] = "Front Biosci (Landmark Ed)",
  ["frontiers in bioscience (scholar edition)"] = "Front Biosci (Schol Ed)",
  ["frontiers in bioscience : a journal and virtual library"] = "Front Biosci",
  ["frontiers in bird science"] = "Front. Bird Sci.",
  ["frontiers in blockchain"] = "Front. Blockchain",
  ["frontiers in built environment"] = "Front Built Environ",
  ["frontiers in carbon"] = "Front. Carbon",
  ["frontiers in cardiovascular medicine"] = "Front Cardiovasc Med",
  ["frontiers in catalysis"] = "Front. Catal.",
  ["frontiers in cell and developmental biology"] = "Front Cell Dev Biol",
  ["frontiers in cell death"] = "Front. Cell Death",
  ["frontiers in cellular and infection microbiology"] = "Front Cell Infect Microbiol",
  ["frontiers in cellular neuroscience"] = "Front Cell Neurosci",
  ["frontiers in chemical engineering"] = "Front. Chem. Eng.",
  ["frontiers in chemistry"] = "Front Chem",
  ["frontiers in climate"] = "Front. Clim.",
  ["frontiers in cognition"] = "Front. Cognit.",
  ["frontiers in communication"] = "Front Commun (Lausanne)",
  ["frontiers in communications and networks"] = "Front. Commun. Networks",
  ["frontiers in computational neuroscience"] = "Front Comput Neurosci",
  ["frontiers in computer science"] = "Front Comput Sci",
  ["frontiers in conservation science"] = "Front. Conserv. Sci.",
  ["frontiers in control engineering"] = "Front. Control Eng.",
  ["frontiers in dental medicine"] = "Front. Dent. Med.",
  ["frontiers in dentistry"] = "Front Dent",
  ["frontiers in digital health"] = "Front Digit Health",
  ["frontiers in drug delivery"] = "Front. Drug Delivery",
  ["frontiers in drug design and discovery"] = "Front. Drug Des. Discovery",
  ["frontiers in drug discovery"] = "Front. Drug Discovery",
  ["frontiers in drug safety and regulation"] = "Front. Drug Saf. Regul.",
  ["frontiers in earth science"] = "Front. Earth Sci.",
  ["frontiers in ecology and evolution"] = "Front Ecol Evol",
  ["frontiers in ecology and the environment"] = "Front Ecol Environ",
  ["frontiers in education"] = "Front Educ (Lausanne)",
  ["frontiers in electronic materials"] = "Front. Electron. Mater.",
  ["frontiers in electronics"] = "Front. Electron.",
  ["frontiers in elliptic and parabolic problems"] = "Front. Elliptic Parabol. Probl.",
  ["frontiers in endocrinology"] = "Front Endocrinol (Lausanne)",
  ["frontiers in energy"] = "Front. Energy",
  ["frontiers in energy research"] = "Front. Energy Res.",
  ["frontiers in engineering and built environment"] = "Front. Eng. Built Environ.",
  ["frontiers in environmental archaeology"] = "Front. Environ. Archaeol.",
  ["frontiers in environmental chemistry"] = "Front. Environ. Chem.",
  ["frontiers in environmental economics"] = "Front. Environ. Econ.",
  ["frontiers in environmental science"] = "Front. Environ. Sci.",
  ["frontiers in epidemiology"] = "Front. Epidemiol.",
  ["frontiers in evolutionary neuroscience"] = "Front Evol Neurosci",
  ["frontiers in food science and technology"] = "Front. Food Sci. Technol.",
  ["frontiers in forests and global change"] = "Front. For. Global Change",
  ["frontiers in fuels"] = "Front. Fuels",
  ["frontiers in fungal biology"] = "Front. Fungal Biol.",
  ["frontiers in future transportation"] = "Front. Future Transp.",
  ["frontiers in gastroenterology"] = "Front. Gastroenterol.",
  ["frontiers in genetics"] = "Front Genet",
  ["frontiers in genome editing"] = "Front Genome Ed",
  ["frontiers in geochemistry"] = "Front. Geochem.",
  ["frontiers in global women's health"] = "Front Glob Womens Health",
  ["frontiers in global women’s health"] = "Front. Global Women’s Health",
  ["frontiers in health policy research"] = "Front Health Policy Res",
  ["frontiers in health policy research / national bureau of economic research"] = "Front Health Policy Res",
  ["frontiers in health services"] = "Front. Health Serv.",
  ["frontiers in heat and mass transfer"] = "Front. Heat Mass Transfer",
  ["frontiers in hematology"] = "Front. Hematol.",
  ["frontiers in high performance computing"] = "Front. High Perform. Comput.",
  ["frontiers in horticulture"] = "Front. Hortic.",
  ["frontiers in human dynamics"] = "Front. Hum. Dyn.",
  ["frontiers in human neuroscience"] = "Front Hum Neurosci",
  ["frontiers in ict"] = "Front. ICT",
  ["frontiers in ict (lausanne, switzerland)"] = "Front ICT",
  ["frontiers in imaging"] = "Front. Imaging",
  ["frontiers in immunology"] = "Front Immunol",
  ["frontiers in insect science"] = "Front. Insect Sci.",
  ["frontiers in integrative neuroscience"] = "Front Integr Neurosci",
  ["frontiers in lab on a chip technologies"] = "Front. Lab Chip Technol.",
  ["frontiers in language sciences"] = "Front. Lang. Sci.",
  ["frontiers in life science"] = "Front Life Sci",
  ["frontiers in malaria"] = "Front. Malar.",
  ["frontiers in manufacturing technology"] = "Front. Manuf. Technol.",
  ["frontiers in marine science"] = "Front. Mar. Sci.",
  ["frontiers in materials"] = "Front. Mater.",
  ["frontiers in materials science"] = "Front. Mater. Sci.",
  ["frontiers in mathematics"] = "Front. Math.",
  ["frontiers in mechanical engineering"] = "Front Mech Eng",
  ["frontiers in medical technology"] = "Front Med Technol",
  ["frontiers in medicine"] = "Front Med (Lausanne)",
  ["frontiers in membrane science and technology"] = "Front. Membr. Sci. Technol.",
  ["frontiers in metals and alloys"] = "Front. Met. Alloys",
  ["frontiers in microbiology"] = "Front Microbiol",
  ["frontiers in microbiomes"] = "Front. Microbiomes",
  ["frontiers in molecular biosciences"] = "Front Mol Biosci",
  ["frontiers in molecular medicine"] = "Front. Mol. Med.",
  ["frontiers in molecular neuroscience"] = "Front Mol Neurosci",
  ["frontiers in nanotechnology"] = "Front. Nanotechnol.",
  ["frontiers in natural products"] = "Front. Nat. Prod.",
  ["frontiers in nephrology"] = "Front. Nephrol.",
  ["frontiers in network physiology"] = "Front. Network Physiol.",
  ["frontiers in neural circuits"] = "Front Neural Circuits",
  ["frontiers in neuroanatomy"] = "Front Neuroanat",
  ["frontiers in neuroendocrine science"] = "Front. Neuroendocr. Sci.",
  ["frontiers in neuroendocrinology"] = "Front Neuroendocrinol",
  ["frontiers in neuroenergetics"] = "Front Neuroenergetics",
  ["frontiers in neuroengineering"] = "Front Neuroeng",
  ["frontiers in neuroergonomics"] = "Front. Neuroergon.",
  ["frontiers in neuroimaging"] = "Front. Neuroimaging",
  ["frontiers in neuroinformatics"] = "Front Neuroinform",
  ["frontiers in neurology"] = "Front Neurol",
  ["frontiers in neurology and neuroscience research"] = "Front Neurol Neurosci Res",
  ["frontiers in neurorobotics"] = "Front Neurorobot",
  ["frontiers in neuroscience"] = "Front Neurosci",
  ["frontiers in nuclear engineering"] = "Front. Nucl. Eng.",
  ["frontiers in nuclear medicine"] = "Front. Nucl. Med.",
  ["frontiers in nutrition"] = "Front Nutr",
  ["frontiers in oncology"] = "Front Oncol",
  ["frontiers in optics. annual meeting of the optical society of america"] = "Front Opt",
  ["frontiers in oral health"] = "Front. Oral Health",
  ["frontiers in pain research"] = "Front. Pain Res.",
  ["frontiers in parasitology"] = "Front. Parasitol.",
  ["frontiers in pediatrics"] = "Front Pediatr",
  ["frontiers in pharmacology"] = "Front Pharmacol",
  ["frontiers in photonics"] = "Front. Photonics",
  ["frontiers in physics"] = "Front Phys",
  ["frontiers in physiology"] = "Front Physiol",
  ["frontiers in plant science"] = "Front Plant Sci",
  ["frontiers in political science"] = "Front Polit Sci",
  ["frontiers in probability and the statistical sciences"] = "Front. Probab. Stat. Sci.",
  ["frontiers in psychiatry"] = "Front Psychiatry",
  ["frontiers in psychological and behavioral science"] = "Front Psychol Behav Sci",
  ["frontiers in psychology"] = "Front Psychol",
  ["frontiers in public health"] = "Front Public Health",
  ["frontiers in public health services & systems research"] = "Front Public Health Serv Syst Res",
  ["frontiers in quantum science and technology"] = "Front. Quantum Sci. Technol.",
  ["frontiers in radiology"] = "Front. Radiol.",
  ["frontiers in rehabilitation sciences"] = "Front. Rehabil. Sci.",
  ["frontiers in remote sensing"] = "Front. Remote Sens.",
  ["frontiers in research metrics and analytics"] = "Front Res Metr Anal",
  ["frontiers in robotics and ai"] = "Front Robot AI",
  ["frontiers in sensors"] = "Front. Sens.",
  ["frontiers in signal processing"] = "Front. Signal Process.",
  ["frontiers in sleep"] = "Front. Sleep",
  ["frontiers in smart grids"] = "Front. Smart Grids",
  ["frontiers in soft matter"] = "Front. Soft Matter",
  ["frontiers in soil science"] = "Front. Soil Sci.",
  ["frontiers in space technologies"] = "Front. Space Technol.",
  ["frontiers in sports and active living"] = "Front Sports Act Living",
  ["frontiers in statistical quality control"] = "Front. Stat. Qual. Control",
  ["frontiers in surgery"] = "Front Surg",
  ["frontiers in sustainability"] = "Front. Sustainability",
  ["frontiers in sustainable cities"] = "Front. Sustainable Cities",
  ["frontiers in sustainable energy policy"] = "Front. Sustainable Energy Policy",
  ["frontiers in sustainable food systems"] = "Front Sustain Food Syst",
  ["frontiers in sustainable resource management"] = "Front. Sustainable Resour. Manage.",
  ["frontiers in sustainable tourism"] = "Front. Sustainable Tourism",
  ["frontiers in synaptic neuroscience"] = "Front Synaptic Neurosci",
  ["frontiers in systems biology"] = "Front. Syst. Biol.",
  ["frontiers in systems neuroscience"] = "Front Syst Neurosci",
  ["frontiers in the history of science"] = "Front. Hist. Sci.",
  ["frontiers in the internet of things"] = "Front. Internet Things",
  ["frontiers in theoretical physics"] = "Front. Theor. Phys.",
  ["frontiers in thermal engineering"] = "Front. Therm. Eng.",
  ["frontiers in toxicology"] = "Front. Toxicol.",
  ["frontiers in transplantation"] = "Front. Transplant.",
  ["frontiers in tropical diseases"] = "Front.Trop. Dis.",
  ["frontiers in urology"] = "Front. Urol.",
  ["frontiers in veterinary science"] = "Front Vet Sci",
  ["frontiers in virology"] = "Front. Virol.",
  ["frontiers in virtual reality"] = "Front Virtual Real",
  ["frontiers in water"] = "Front. Water",
  ["frontiers in women's health"] = "Front Womens Health",
  ["frontiers in zoology"] = "Front Zool",
  ["frontiers of agricultural science and engineering"] = "Front Agric Sci Eng",
  ["frontiers of agriculture in china"] = "Front Agric China",
  ["frontiers of architectural research"] = "Front. Archit. Res.",
  ["frontiers of biogeography"] = "Front Biogeogr",
  ["frontiers of biology"] = "Front Biol",
  ["frontiers of biology in china : selected publications from chinese universities"] = "Front Biol China",
  ["frontiers of biomechanics"] = "Front. Biomech.",
  ["frontiers of chemical engineering in china"] = "Front. Chem. Eng. China",
  ["frontiers of chemical science and engineering"] = "Front Chem Sci Eng",
  ["frontiers of chemistry in china"] = "Front. Chem. China",
  ["frontiers of chemistry in china : selected publications from chinese universities"] = "Front Chem China",
  ["frontiers of computer science"] = "Front Comput Sci",
  ["frontiers of computer science in china"] = "Front Comput Sci China",
  ["frontiers of earth science"] = "Front Earth Sci",
  ["frontiers of earth science in china"] = "Front Earth Sci. China",
  ["frontiers of economic research"] = "Frontiers Econom. Res.",
  ["frontiers of electrical and electronic engineering"] = "Front. Electr. Electron. Eng.",
  ["frontiers of electrical and electronic engineering in china"] = "Front. Electr. Electron. Eng. China",
  ["frontiers of energy and power engineering in china"] = "Front. Energy Power Eng. China",
  ["frontiers of engineering management"] = "Front. Eng. Manage.",
  ["frontiers of environmental science & engineering"] = "Front Environ Sci Eng",
  ["frontiers of environmental science & engineering in china"] = "Front. Environ. Sci. Eng. China",
  ["frontiers of gastrointestinal research"] = "Front Gastrointest Res",
  ["frontiers of health services management"] = "Front Health Serv Manage",
  ["frontiers of history in china"] = "Front Hist China",
  ["frontiers of hormone research"] = "Front Horm Res",
  ["frontiers of information technology & electronic engineering"] = "Front. Inf. Technol. Electron. Eng.",
  ["frontiers of materials science"] = "Front Mater Sci",
  ["frontiers of materials science in china"] = "Front. Mater. Sci. China",
  ["frontiers of mathematical finance"] = "Front. Math. Finance",
  ["frontiers of mathematics"] = "Front. Math.",
  ["frontiers of mathematics in china"] = "Front. Math. China",
  ["frontiers of mathematics in china : selected papers from chinese universities"] = "Front Math China",
  ["frontiers of mechanical engineering"] = "Front. Mech. Eng.",
  ["frontiers of mechanical engineering in china"] = "Front. Mech. Eng. China",
  ["frontiers of medical and biological engineering"] = "Front. Med. Biol. Eng.",
  ["frontiers of medical and biological engineering : the international journal of the japan society of medical electronics and biological engineering"] = "Front Med Biol Eng",
  ["frontiers of medicine"] = "Front Med",
  ["frontiers of medicine in china"] = "Front Med China",
  ["frontiers of neurology and neuroscience"] = "Front Neurol Neurosci",
  ["frontiers of operational research and applied systems analysis"] = "Frontiers Oper. Res. Appl. Systems Anal.",
  ["frontiers of optoelectronics"] = "Front. Optoelectron.",
  ["frontiers of oral and maxillofacial medicine"] = "Front Oral Maxillofac Med",
  ["frontiers of oral biology"] = "Front Oral Biol",
  ["frontiers of oral physiology"] = "Front Oral Physiol",
  ["frontiers of physics"] = "Front Phys (Beijing)",
  ["frontiers of physics in china"] = "Front Phys China",
  ["frontiers of radiation therapy and oncology"] = "Front Radiat Ther Oncol",
  ["frontiers of structural and civil engineering"] = "Front. Struct. Civ. Eng.",
  ["frontiers of stuctural and civil engineering"] = "Front. Struct. Civ. Eng.",
  ["frontline gastroenterology"] = "Frontline Gastroenterol",
  ["frontline learning research"] = "Frontline Learn Res",
  ["fruhmittelalterliche studien"] = "Fruhmittelalt Stud",
  ["frühmittelalterliche studien"] = "Fruhmittelalt Stud",
  ["frühmittelalterliche studien. jahrbuch des instituts für frühmittelalterforschung der universität münster"] = "FruehMitAltSt",
  ["fu dan xue bao. zi ran ke xue ban"] = "Journal of Fudan University. Natural science|Fu Dan Xue Bao Zi Ran Ke Xue Ban",
  ["fu dan xue bao. zi ran ke xue ban = journal of fudan university. natural science"] = "Fu Dan Xue Bao Zi Ran Ke Xue Ban",
  ["fu jian shui chan = fujian fisheries"] = "Fu Jian Shui Chan",
  ["fudan university"] = "J. Fudan Univ. Nat. Sci.",
  ["fudan university. journal. natural science. fudan xuebao. ziran kexue ban"] = "J. Fudan Univ. Nat. Sci.",
  ["fuel (london, england)"] = "Fuel (Lond)",
  ["fuel cells"] = "Fuel Cells",
  ["fuel cells (weinheim)"] = "Fuel Cells (Weinh)",
  ["fuel cells (weinheim, germany)"] = "Fuel Cells (Weinheim, Ger.)",
  ["fuel cells bulletin"] = "Fuel Cells Bull.",
  ["fuel communications"] = "Fuel Commun.",
  ["fuel processing technology"] = "Fuel Process. Technol.",
  ["fuel science and technology international"] = "Fuel Sci. Technol. Int.",
  ["fuerzas armadas de venezuela (caracas, venezuela : 1971)"] = "Rev Fuerzas Armadas Venez",
  ["fugitive leaves from the historical collections / library of the college of physicians of philadelphia. college of physicians of philadelphia. library"] = "Fugitive Leaves Hist Collect",
  ["fugitive leaves from the historical collections. college of physicians of philadelphia. library"] = "Fugitive Leaves Hist Collect",
  ["fujian linxueyuan xuebao"] = "Fujian Linxueyuan Xuebao",
  ["fujian nong ye ke ji"] = "Fujian Nong Ye Ke Ji",
  ["fujian nong ye xue bao = fujian journal of agricultural sciences"] = "Fujian Nong Ye Xue Bao",
  ["fujian shifan daxue xuebao"] = "Fujian Shifan Daxue Xuebao Ziran Kexue Ban",
  ["fujian yi yao za zhi"] = "Fujian medical journal|Fu Jian Yi Yao Za Zhi",
  ["fujian yi yao za zhi = fujian medical journal"] = "Fu Jian Yi Yao Za Zhi",
  ["fujian yike daxue xuebao. shehui kexue ban"] = "Fujian Yike Daxue Xuebao Shehui Kexue Ban",
  ["fujitsu scientific & technical journal"] = "Fujitsu Sci. Tech. J.",
  ["fukui-ken nogyo shikenjo kenkyu hokoku = bulletin of the fukui agricultural experiment station"] = "Fukui Ken Nogyo Shikenjo kenkyu hokoku",
  ["fukuoka igaku zasshi (fukuoka acta medica)"] = "Fukuoka Igaku Zasshi",
  ["fukuoka igaku zasshi = hukuoka acta medica"] = "Fukuoka Igaku Zasshi",
  ["fukuoka igaku zasshi. fukuoka acta medica"] = "Fukuoka Igaku Zasshi",
  ["fukuoka shika daigaku gakkai zasshi"] = "Fukuoka Shika Daigaku Gakkai Zasshi",
  ["fukuoka university science reports"] = "Fukuoka Univ. Sci. Rep.",
  ["fukushima journal of medical science"] = "Fukushima J Med Sci",
  ["fukushima university"] = "Sci. Rep. Fac. Ed. Fukushima Univ.",
  ["ful-, orr-, gegegyogyazat"] = "Fulorrgegegyogyaszat",
  ["full circle"] = "Full Circ.",
  ["fullerence, nanotubes, and carbon nanostructures"] = "Fullerenes Nanotubes Carbon Nanostruct.",
  ["fullerene science and technology"] = "Fullerene Sci. Technol.",
  ["fullerenes, nanotubes, and carbon nanostructures"] = "Fullerenes, Nanotubes, Carbon Nanostruct.",
  ["function (oxford, england)"] = "Function (Oxf)",
  ["functional & integrative genomics"] = "Funct Integr Genomics",
  ["functional analysis and its applications"] = "Funct. Anal. Appl.",
  ["functional analysis and other mathematics"] = "Funct. Anal. Other Math.",
  ["functional analysis, approximation and computation"] = "Funct. Anal. Approx. Comput.",
  ["functional and developmental morphology"] = "Funct Dev Morphol",
  ["functional and functionally structured materials : selected, peer reviewed papers from the chinese materials conference 2015, july 10-14, 2015, guiyang, china. chinese materials conference (2015 : guiyang, china)"] = "Funct Funct Struct Mater",
  ["functional and integrative genomics"] = "Funct. Integr. Genomics",
  ["functional composite materials"] = "Funct. Compos. Mater.",
  ["functional composites and structures"] = "Funct. Compos. Struct.",
  ["functional diamond"] = "Funct. Diamond",
  ["functional differential equations"] = "Funct. Differ. Equ.",
  ["functional ecology"] = "Funct Ecol",
  ["functional food reviews (print)"] = "Funct Food Rev",
  ["functional materials"] = "Funct. Mater.",
  ["functional materials letters"] = "Funct. Mater. Lett.",
  ["functional neurology"] = "Funct Neurol",
  ["functional orthodontist"] = "Funct. Orthod.",
  ["functional plant biology"] = "Funct. Plant Biol.",
  ["functional plant biology : fpb"] = "Funct Plant Biol",
  ["functional plant science & biotechnology"] = "Funct Plant Sci Biotechnol",
  ["functiones et approximatio commentarii mathematici"] = "Funct. Approximatio Comment. Math.",
  ["fund raising management"] = "Fund Raising Manage",
  ["fundamenta informaticae"] = "Fund. Inform.",
  ["fundamenta mathematicae"] = "Fund. Math.",
  ["fundamenta scientiae"] = "Fundam Sci",
  ["fundamental & clinical pharmacology"] = "Fundam Clin Pharmacol",
  ["fundamental and applied limnology"] = "Fundam. Appl. Limnol.",
  ["fundamental and applied limnology : official journal of the international association of theoretical and applied limnology"] = "Fundam Appl Limnol",
  ["fundamental and applied nematology"] = "Fundam Appl Nematol",
  ["fundamental and applied nuclear physics series"] = "Fundam. Appl. Nucl. Phys. Ser.",
  ["fundamental and applied toxicology"] = "Fundam. Appl. Toxicol.",
  ["fundamental and applied toxicology : official journal of the society of toxicology"] = "Fundam Appl Toxicol",
  ["fundamental and clinical pharmacology"] = "Fundam. Clin. Pharmacol.",
  ["fundamental plasma physics"] = "Fundam. Plasma Phys.",
  ["fundamental research"] = "Fundam. Res.",
  ["fundamental studies in computer science"] = "Fund. Stud. Comput. Sci.",
  ["fundamental theories of physics"] = "Fundam. Theor. Phys.",
  ["fundamental toxicological sciences"] = "Fundam Toxicol Sci",
  ["fundamentalnaya i prikladnaya matematika"] = "Fundam. Prikl. Mat.",
  ["fundamentals of algorithms"] = "Fundam. Algorithms",
  ["fundberichte aus baden-württemberg"] = "FuBerBadWuert",
  ["fundberichte aus hessen"] = "FuBerHessen",
  ["fundberichte aus schwaben"] = "FuBerSchwab",
  ["fundberichte aus österreich"] = "FuBerOe",
  ["funde und ausgrabungen im bezirk trier"] = "FuAusgrTrier",
  ["fundmünzen der römischen zeit in österreich"] = "FMROe",
  ["fundort wien. berichte zur archäologie"] = "FuWien",
  ["fungal biology"] = "Fungal Biol",
  ["fungal biology and biotechnology"] = "Fungal Biol Biotechnol",
  ["fungal biology reviews"] = "Fungal Biol Rev",
  ["fungal diversity"] = "Fungal Divers",
  ["fungal ecology"] = "Fungal Ecol",
  ["fungal genetics and biology"] = "Fungal Genet. Biol.",
  ["fungal genetics and biology : fg & b"] = "Fungal Genet Biol",
  ["fungal genetics and ecology"] = "Fungal Genet. Ecol.",
  ["fungal genetics newsletter"] = "Fungal Genet Newsl",
  ["fungal genetics reports"] = "Fungal Genet Rep",
  ["fungal genomics & biology"] = "Fungal Genom Biol",
  ["fungal planet"] = "Fungal Planet",
  ["fungal science"] = "Fungal Sci",
  ["fungal systematics and evolution"] = "Fungal Syst Evol",
  ["fungi iberici"] = "Fungi Iber",
  ["fungi non delineati : raro vel haud perspecte et explorate descripti aut definite picti"] = "Fungi Non Delineati",
  ["funkcialaj ekvacioj. serio internacia"] = "Funkcial. Ekvac.",
  ["funktsionalnyi analiz i ego prilozheniya"] = "Funktsional. Anal. i Prilozhen.",
  ["funtai oyobi funmatsu yakin"] = "Funtai oyobi Funmatsu Yakin",
  ["fushe fanghu"] = "Fushe Fanghu",
  ["fusio : the bentley undergraduate research journal"] = "Fusio",
  ["fusion engineering and design"] = "Fusion Eng. Des.",
  ["fusion science and technology"] = "Fusion Sci. Technol.",
  ["fusion technology"] = "Fusion Technol.",
  ["futur anterieur"] = "Futur Anter",
  ["futur antérieur"] = "Futur Anter",
  ["future cardiology"] = "Future Cardiol",
  ["future cities and environment"] = "Future Cities Environ.",
  ["future computing and informatics journal"] = "Future Comput. Inf. J.",
  ["future dentistry"] = "Future Dent",
  ["future drug discovery"] = "Future Drug Discov",
  ["future foods"] = "Future Foods",
  ["future generation computer systems"] = "Future Gener. Comput. Syst.",
  ["future generations computer systems : fgcs"] = "Future Gener Comput Syst",
  ["future healthcare journal"] = "Future Healthc J",
  ["future hiv therapy"] = "Futur HIV Ther",
  ["future hospital journal"] = "Future Hosp J",
  ["future internet"] = "Future Internet",
  ["future journal of pharmaceutical sciences"] = "Futur J Pharm Sci",
  ["future lipidology"] = "Future Lipidol",
  ["future medicinal chemistry"] = "Future Med Chem",
  ["future medicine ai"] = "Future Med. AI",
  ["future microbiology"] = "Future Microbiol",
  ["future neurology"] = "Future Neurol",
  ["future of anthropological knowledge"] = "Asa. Decen. Conf. Ser.",
  ["future of business and finance"] = "Future Bus. Finance",
  ["future of children"] = "Future Child.",
  ["future oncology"] = "Future Oncol.",
  ["future oncology (london, england)"] = "Future Oncol",
  ["future rheumatology"] = "Fut Rheumatol",
  ["future science oa"] = "Future Sci OA",
  ["future virology"] = "Future Virol",
  ["futuribles (paris, france : 1981)"] = "Futuribles",
  ["futuribles 2000 [i.e. deux mille]"] = "Futruibles 2000",
  ["fuzzy information and engineering"] = "Fuzzy Inf. Eng.",
  ["fuzzy management methods"] = "Fuzzy Manag. Methods",
  ["fuzzy optimization and decision making"] = "Fuzzy Optim. Dec. Making",
  ["fuzzy optimization and decision making. a journal of modeling and computation under uncertainty"] = "Fuzzy Optim. Decis. Mak.",
  ["fuzzy sets and systems"] = "Fuzzy Sets Syst.",
  ["fuzzy sets and systems. an international journal in information science and engineering"] = "Fuzzy Sets and Systems",
  ["fül-, orr-, gégegyógyázat"] = "Fulorrgegegyogyaszat",
  ["fysiatricky a reumatologicky vestnik"] = "Fysiatr Revmatol Vestn",
  ["fysiatricky a revmatologicky vestnik"] = "Fysiatr. Revmatol. Vestn.",
  ["fysiatricky vestnik. ceskoslovenska fysiatricka spolecnost"] = "Fysiatr Vestn Cesk Fysiatr Spol",
  ["fysiatrický a reumatologický vestník"] = "Fysiatr Revmatol Vestn",
  ["fysiatrický vĕstník. československá fysiatrická společnost v praze"] = "Fysiatr Vestn Cesk Fysiatr Spol",
  ["g. liddell – r. scott – h. s. jones, a greek-english lexikon \\textsuperscript{9}(1996)"] = "Liddell-­Scott-Jones",
  ["g.i.t. laboratory journal europe"] = "GIT Lab J Eur",
  ["g3 (bethesda, md.)"] = "G3 (Bethesda)",
  ["g3-genes genomes genetics"] = '"G3-Genes Genomes Genet."',
  ["g3: genes, genomes, genetics"] = "G3: Genes, Genomes, Genet.",
  ["gabi journal"] = "GaBI J",
  ["gaceta conasida"] = "Gac CONASIDA",
  ["gaceta conasida / consejo nacional para la prevencion y control del sida"] = "Gac CONASIDA",
  ["gaceta medica"] = "Gac Med (Guayaquil)",
  ["gaceta medica de caracas"] = "Gac. Med. Caracas",
  ["gaceta medica de mexico"] = "Gac Med Mex",
  ["gaceta medica del norte"] = "Gac Med Norte",
  ["gaceta medica espanola"] = "Gac Med Esp",
  ["gaceta médica"] = "Gac Med Villahermosa",
  ["gaceta médica centroamericana"] = "Gac Med Centroam",
  ["gaceta médica de caracas"] = "Gac Med Caracas",
  ["gaceta médica de lima"] = "Gac Med Lima",
  ["gaceta médica de méxico"] = "Gac Med Mex",
  ["gaceta médica de occidente"] = "Gac Med Occidente",
  ["gaceta médica del norte"] = "Gac Med Norte",
  ["gaceta médica española"] = "Gac Med Esp",
  ["gaceta médico-quirúrgica de bolivia"] = "Gac Med Quir Boliv",
  ["gaceta médica de méxico"] = "Gac. Méd. Méx.",
  ["gaceta numismática"] = "GacNum",
  ["gaceta peruana de cirugía y medicina"] = "Gac Peru Cir Med",
  ["gaceta sanitaria"] = "Gac Sanit",
  ["gaceta sanitaria / s.e.s.p.a.s"] = "Gac Sanit",
  ["gades. revista del colegio universitario de filosofía y letras"] = "Gades",
  ["gaea; anales de la sociedad argentina de estudios geograficos"] = "GAEA An Soc Argent Estud Geogr",
  ["gaea; anales de la sociedad argentina de estudios geográficos"] = "GAEA An Soc Argent Estud Geogr",
  ["gait & posture"] = "Gait Posture",
  ["gait and posture"] = "Gait Posture",
  ["gakuto international series"] = "GAKUTO Int. Ser. Math. Sci. Appl.",
  ["gakuto international series. mathematical sciences and applications"] = "GAKUTO Internat. Ser. Math. Sci. Appl.",
  ["galapagos research"] = "Galapagos Res.",
  ["galenika; casopis za medicinu, farmaciju, hemiju i srodne nauke"] = "Galenika Cas Med Farm Hem Srod Nauke",
  ["galenika; c̆asopis za medicinu, farmaciju, hemiju i srodne nauke"] = "Galen Cas Med Farm Hem Srod Nauke",
  ["galénica acta"] = "Galen Acta",
  ["gallaecia. publicación del departamento de prehistoria y arqueología"] = "Gallaecia",
  ["gallia informations. l’archéologie des régions"] = "GalliaInfAReg",
  ["gallia informations. préhistoire et histoire"] = "GalliaInf",
  ["gallia préhistoire. archéologie de la france préhistorique"] = "GalliaPrehist",
  ["gallia. fouilles et monuments archéologiques en france metropolitaine"] = "Gallia",
  ["gallica biologica acta"] = "Gall Biol Acta",
  ["gallup report (princeton, n.j. : 1981)"] = "Gallup Rep",
  ["game theory and applications"] = "Game Theory Appl.",
  ["games and culture"] = "Games Cult",
  ["games and economic behavior"] = "Games Econ Behav",
  ["games for health journal"] = "Games Health J",
  ["games: research and practice"] = "Games: Res. Pract.",
  ["gamete research"] = "Gamete Res",
  ["gaming law review and economics"] = "Gaming Law Rev Econ",
  ["gams journal of mathematics and mathemaical biosciences"] = "GAMS J. Math. Math. Biosci.",
  ["gams journal of mathematics and mathematical biosciences"] = "GAMS J. Math. Math. Biosci.",
  ["gan no rinsho (japanese journal of cancer clinics)"] = "Gan No Rinsho",
  ["gan no rinsho. japan journal of cancer clinics"] = "Gan No Rinsho",
  ["gan to kagaku ryoho (cancer and chemotherapy)"] = "Gan To Kagaku Ryoho",
  ["gan to kagaku ryoho [japanese journal of cancer and chemotherapy]"] = "Gan To Kagaku Ryoho",
  ["gan to kagaku ryoho. cancer & chemotherapy"] = "Gan To Kagaku Ryoho",
  ["gan. gann; the japanese journal of cancer research"] = "Gan",
  ["ganguang kexue yu guang huaxue"] = "Ganguang Kexue Yu Guang Huaxue",
  ["ganit. journal of bangladesh mathematical society"] = "Ganit",
  ["ganita bharati. bulletin of the indian society for history of mathematics"] = "Ganita Bharati",
  ["ganita sandesh"] = "Ganita Sandesh",
  ["ganita. a journal devoted to mathematical sciences"] = "Ganita",
  ["ganka (ophthalmology)"] = "Ganka",
  ["ganka. ophthalmology"] = "Ganka",
  ["gansu nong ye da xue xue bao"] = "Gansu nongye daxue xuebao|Gansu Nong Ye Da Xue Xue Bao",
  ["gansu nong ye da xue xue bao = gansu nongye daxue xuebao"] = "Gansu Nong Ye Da Xue Xue Bao",
  ["gansu nong ye ke ji = gansu nongye keji"] = "Gansu Nong Ye Ke Ji",
  ["gao fen zi cai liao ke xue yu gong cheng = polymeric materials science & engineering"] = "Gao Fen Zi Cai Liao Ke Xue Yu Gong Cheng",
  ["gaodeng xuexiao huaxue xuebao"] = "Gaodeng Xuexiao Huaxue Xuebao",
  ["gaofenzi cailiao kexue yu gongcheng"] = "Gaofenzi Cailiao Kexue Yu Gongcheng",
  ["gaofenzi xuebao"] = "Gaofenzi Xuebao",
  ["gaojishu tongxun"] = "Gaojishu Tongxun",
  ["gaoneng wuli yu hewuli"] = "Gaoneng Wuli Yu Hewuli",
  ["gaoxiao huaxue gongcheng xuebao"] = "Gaoxiao Huaxue Gongcheng Xuebao",
  ["gaoxiao yingyong shuxue xuebao"] = "Gaoxiao Yingyong Shuxue Xuebao Ser. A",
  ["gaoxiong yi xue ke xue za zhi (kaohsiung journal of medical sciences)"] = "Gaoxiong Yi Xue Ke Xue Za Zhi",
  ["gaoxiong yi xue ke xue za zhi = the kaohsiung journal of medical sciences"] = "Gaoxiong Yi Xue Ke Xue Za Zhi",
  ["garden history"] = "Gard Hist",
  ["garland reference library of social science"] = "Garland Ref. Libr. Soc. Sci.",
  ["gartenamt, das"] = "Gartenamt",
  ["gartenbau, der"] = "Gartenbau",
  ["gas science and engineering"] = "Gas Sci. Eng.",
  ["gas separation and purification"] = "Gas Sep. Purif.",
  ["gas, wasser, abwasser"] = "Gas Wasser Abwasser",
  ["gastric cancer"] = "Gastric Cancer",
  ["gastric cancer : official journal of the international gastric cancer association and the japanese gastric cancer association"] = "Gastric Cancer",
  ["gastroenterologia bohema"] = "Gastroenterol Bohema",
  ["gastroenterologia japonica"] = "Gastroenterol Jpn",
  ["gastroenterologia polska : organ polskiego towarzystwa gastroenterologii"] = "Gastroenterol Pol",
  ["gastroenterologia y hepatologia"] = "Gastroenterol Hepatol",
  ["gastroenterologie clinique et biologique"] = "Gastroenterol Clin Biol",
  ["gastroenterologische fortbildungskurse fur die praxis"] = "Gastroenterol. Fortbildungskurse Prax.",
  ["gastroenterologische fortbildungskurse für die praxis"] = "Gastroenterol Fortbildungskurse Prax",
  ["gastroenterologisches journal"] = "Gastroenterol. J.",
  ["gastroenterologisches journal : organ der gesellschaft fur gastroenterologie der ddr"] = "Gastroenterol J",
  ["gastroenterologisches journal : organ der gesellschaft für gastroenterologie der ddr"] = "Gastroenterol J",
  ["gastroenterología y hepatología"] = "Gastroenterol Hepatol",
  ["gastroenterology & hepatology"] = "Gastroenterol Hepatol (N Y)",
  ["gastroenterology & hepatology (bartlesville, okla.)"] = "Gastroenterol Hepatol (Bartlesville)",
  ["gastroenterology (glendale, calif.) [sound recording]"] = "Gastroenterol",
  ["gastroenterology and hepatology"] = "Gastroenterol Hepatol (Qué)",
  ["gastroenterology and hepatology from bed to bench"] = "Gastroenterol Hepatol Bed Bench",
  ["gastroenterology clinics of north america"] = "Gastroenterol Clin North Am",
  ["gastroenterology insights"] = "Gastroenterol Insights",
  ["gastroenterology nursing"] = "Gastroenterol. Nurs.",
  ["gastroenterology nursing : the official journal of the society of gastroenterology nurses and associates"] = "Gastroenterol Nurs",
  ["gastroenterology report"] = "Gastroenterol Rep (Oxf)",
  ["gastroenterology research"] = "Gastroenterology Res",
  ["gastroenterology research and practice"] = "Gastroenterol Res Pract",
  ["gastroentérologie clinique et biologique"] = "Gastroenterol Clin Biol",
  ["gastrointestinal cancer : research & therapy"] = "Gastrointest Cancer (Jersey City)",
  ["gastrointestinal cancer : targets and therapy"] = "Gastrointest Cancer",
  ["gastrointestinal cancer research : gcr"] = "Gastrointest Cancer Res",
  ["gastrointestinal disorders (basel, switzerland)"] = "Gastrointest Disord (Basel)",
  ["gastrointestinal endoscopy"] = "Gastrointest Endosc",
  ["gastrointestinal endoscopy clinics of north america"] = "Gastrointest Endosc Clin N Am",
  ["gastrointestinal radiology"] = "Gastrointest Radiol",
  ["gastrointestinal tumors"] = "Gastrointest Tumors",
  ["gastronomica : the journal of food and culture"] = "Gastronomica (Berkeley Calif)",
  ["gaswaerme international"] = "Gaswaerme Int.",
  ["gates open research"] = "Gates Open Res",
  ["gateway heritage : quarterly journal of the missouri historical society"] = "Gateway Herit",
  ["gateways : international journal of community research & engagement"] = "Gateways",
  ["gauss-gesellschaft e"] = "Gauss-Ges. Göttingen Mitt.",
  ["gayana. botánica"] = "Gayana Bot",
  ["gaz, woda; technika sanitarna"] = "Gaz Woda Tech",
  ["gazeta medica portuguesa"] = "Gaz Med Port",
  ["gazeta médica portuguesa"] = "Gaz Med Port",
  ["gazette - the law society of upper canada. law society of upper canada"] = "Gaz Law Soc Upper Can",
  ["gazette des beaux-arts"] = "GazBA",
  ["gazette des hopitaux civils et militaire de l empire ottoman"] = "Gaz Hop Civ Mil Empire Ottoman",
  ["gazette des hôpitaux civils et militaires de l'empire ottoman"] = "Gaz Hop Civ Mil Empire Ottoman",
  ["gazette des mathematiciens"] = "Gaz. Math.",
  ["gazette des mathématiciens"] = "Gaz. Math.",
  ["gazette medicale de france"] = "Gaz. Med. Fr.",
  ["gazette médicale (paris, france)"] = "Gaz Med (Paris)",
  ["gazette médicale de france"] = "Gaz Med Fr",
  ["gazette médicale limousine"] = "Gaz Med Limousine (1908)",
  ["gazette numismatique suisse"] = "GNS",
  ["gazette of the egyptian paediatric association"] = "Gaz. Egypt. Paediatr. Assoc.",
  ["gazette of the grolier club. grolier club"] = "Gaz Grolier Club",
  ["gazodinamika i teploobmen"] = "Gazodinam. i Teploobmen",
  ["gazzetta chimica italiana"] = "Gazz Chim Ital",
  ["gazzetta degli ospedali e delle cliniche"] = "Gazz Osp Clin",
  ["gazzetta internazionale di medicina e chirurgia"] = "Gazz Int Med Chir",
  ["gazzetta medica italiana"] = "Gazz Med Ital",
  ["gazzetta medica lombarda"] = "Gaz Med Lomb",
  ["gazzetta sanitaria"] = "Gazz Sanit",
  ["gbf monographs"] = "GBF Monogr.",
  ["gcb bioenergy"] = "GCB Bioenergy",
  ["gcb bioenergy: bioproducts for a sustainable bioeconomy"] = "GCB Bioenergy",
  ["ge portuguese journal of gastroenterology"] = "GE Port J Gastroenterol",
  ["geburtshilfe und frauenheilkunde"] = "Geburtshilfe Frauenheilkd",
  ["gedrag & gezondheid"] = "Gedrag Gezond",
  ["gefahrstoffe reinhaltung der luft"] = "Gefahrstoffe - Reinhalt. Luft",
  ["gefahrstoffe, reinhaltung der luft = air quality control"] = "Gefahrst Reinhalt Luft",
  ["gefässchirurgie : zeitschrift für vaskuläre und endovaskuläre chirurgie : organ der deutschen und der österreichischen gesellschaft für gefässchirurgie unter mitarbeit der schweizerischen gesellschaft für gefässchirurgie"] = "Gefasschirurgie",
  ["gefiederte welt, die"] = "Gefied.  Welt",
  ["gegenbaurs morphologisches jahrbuch"] = "Gegenbaurs Morphol Jahrb",
  ["geka chiryo (surgical therapy)"] = "Geka Chiryo",
  ["geka chiryo. surgical therapy"] = "Geka Chiryo",
  ["gelişme dergisi = studies in development"] = "Gelis Derg Orta dogu Tek Univ",
  ["geloof en wetenschap"] = "Geloof Wet",
  ["gels (basel, switzerland)"] = "Gels",
  ["gem - international journal on geomathematics"] = "GEM Int. J. Geomath.",
  ["gem : international journal on geomathematics"] = "GEM",
  ["gem state rn news letter"] = "Gem State RN News Lett.",
  ["gem. international journal on geomathematics"] = "GEM Int. J. Geomath.",
  ["gematologiia i transfuziologiia"] = "Gematol Transfuziol",
  ["gems & gemology"] = "Gems Gemol.",
  ["genava. revue d’histoire de l’art et d’archéologie"] = "Genava",
  ["gendaishi kenkyū"] = "Gendaishi Kenkyu",
  ["gender & history"] = "Gend Hist",
  ["gender & society"] = "Gend. Soc.",
  ["gender & society : official publication of sociologists for women in society"] = "Gend Soc",
  ["gender action : a newsletter of the usaid office of women in development"] = "Gend Action",
  ["gender and development"] = "Gend Dev",
  ["gender and education"] = "Gend Educ",
  ["gender and the genome"] = "Gend Genome",
  ["gender in management"] = "Gend Manag",
  ["gender issues"] = "Gender Issues",
  ["gender medicine"] = "Gend Med",
  ["gender medicine : official journal of the partnership for gender-specific medicine at columbia university"] = "Gend Med",
  ["gender, place and culture : a journal of feminist geography"] = "Gend Place Cult",
  ["gender, technology and development"] = "Gend Technol Dev",
  ["gender, work, and organization"] = "Gend Work Organ",
  ["gene & translational bioinformatics"] = "Gene Transl Bioinform",
  ["gene amplification and analysis"] = "Gene Amplif Anal",
  ["gene analysis techniques"] = "Gene Anal Tech",
  ["gene and genome editing"] = "Gene Genome Ed.",
  ["gene expression"] = "Gene Expr",
  ["gene expression patterns"] = "Gene Expr. Patterns",
  ["gene expression patterns : gep"] = "Gene Expr Patterns",
  ["gene function & disease"] = "Gene Funct Dis",
  ["gene geography"] = "Gene Geogr.",
  ["gene geography : a computerized bulletin on human gene frequencies"] = "Gene Geogr",
  ["gene regulation and systems biology"] = "Gene Regul Syst Bio",
  ["gene reports"] = "Gene Rep",
  ["gene technology"] = "Gene Technol",
  ["gene therapy"] = "Gene Ther",
  ["gene therapy & molecular biology"] = "Gene Ther Mol Biol",
  ["gene therapy and regulation"] = "Gene Ther. Regul.",
  ["gene x environment"] = "Gene x Environ.",
  ["gene, cell and tissue"] = "Gene Cell Tissue",
  ["gene: x"] = "Gene X",
  ["geneeskunde en sport"] = "Geneeskd Sport",
  ["geneeskundige bladen uit kliniek en laboratorium voor de praktijk"] = "Geneeskd Bl",
  ["geneeskundige gids"] = "Geneeskd Gids",
  ["general & diagnostic pathology"] = "Gen Diagn Pathol",
  ["general and comparative endocrinology"] = "Gen Comp Endocrinol",
  ["general and diagnostic pathology"] = "Gen. Diagn. Pathol.",
  ["general cytochemical methods"] = "Gen Cytochem Methods",
  ["general dentistry"] = "Gen Dent",
  ["general hospital psychiatry"] = "Gen Hosp Psychiatry",
  ["general information programme, unisist newsletter"] = "Gen Inf Programme UNISISTNewsl",
  ["general laws of rhode island, 1956 : completely annotated. rhode island"] = "Gen Laws R I 1956 R I",
  ["general medicine (los angeles, calif.)"] = "Gen Med (Los Angel)",
  ["general pharmacology"] = "Gen Pharmacol",
  ["general physiology and biophysics"] = "Gen Physiol Biophys",
  ["general practice"] = "Gen Pract",
  ["general practice clinics"] = "Gen Pract Clin",
  ["general psychiatry"] = "Gen Psychiatr",
  ["general relativity and gravitation"] = "Gen. Relativ. Gravitation",
  ["general surgery & laparoscopy news"] = "Gen Surg Laparosc News",
  ["general systems : yearbook of the society for the advancement of general systems theory"] = "Gen Syst",
  ["general technical report usda forest service"] = "Gen. Tech. Rep. USDA For. Serv.",
  ["general thoracic and cardiovascular surgery"] = "Gen Thorac Cardiovasc Surg",
  ["general topology and its applications"] = "Gen. Topol. Appl.",
  ["generations (san francisco, calif.)"] = "Generations",
  ["genes & cancer"] = "Genes Cancer",
  ["genes & development"] = "Genes Dev",
  ["genes & diseases"] = "Genes Dis",
  ["genes & genetic systems"] = "Genes Genet Syst",
  ["genes & genomics"] = "Genes Genomics",
  ["genes & immunity"] = "Genes Immun.",
  ["genes & nutrition"] = "Genes Nutr",
  ["genes and development"] = "Genes Dev.",
  ["genes and environment"] = "Genes Environ.",
  ["genes and environment : the official journal of the japanese environmental mutagen society"] = "Genes Environ",
  ["genes and function"] = "Genes Funct",
  ["genes and genetic systems"] = "Genes Genet. Syst.",
  ["genes and immunity"] = "Genes Immun",
  ["genes to cells"] = "Genes Cells",
  ["genes to cells : devoted to molecular & cellular mechanisms"] = "Genes Cells",
  ["genes, brain and behavior"] = "Genes, Brain Behav.",
  ["genes, brain, and behavior"] = "Genes Brain Behav",
  ["genes, chromosomes & cancer"] = "Genes Chromosomes Cancer",
  ["genes, chromosomes and cancer"] = "Genes, Chromosomes Cancer",
  ["genesis (new york, n.y. : 2000)"] = "Genesis",
  ["genethics news : genetic engineering, ethics and the environment"] = "Genethics News",
  ["genetic analysis"] = "Genet. Anal.",
  ["genetic analysis - biomolecular engineering"] = "Genet. Anal. - Biomol. Eng.",
  ["genetic analysis : biomolecular engineering"] = "Genet Anal",
  ["genetic analysis, techniques and applications"] = "Genet Anal Tech Appl",
  ["genetic and evolutionary computation conference : [proceedings]. genetic and evolutionary computation conference"] = "Genet Evol Comput Conf",
  ["genetic and evolutionary computation series"] = "Genet. Evol. Comput. Ser.",
  ["genetic counseling"] = "Genet. Couns.",
  ["genetic counseling (geneva, switzerland)"] = "Genet Couns",
  ["genetic engineering"] = "Genet Eng (N Y)",
  ["genetic engineering (academic press)"] = "Genet Eng",
  ["genetic engineering and biosafety journal"] = "Genet Eng Biosaf J",
  ["genetic engineering and biotechnology journal"] = "Genet Eng Biotechnol J",
  ["genetic engineering news : gen"] = "Genet Eng News",
  ["genetic epidemiology"] = "Genet Epidemiol",
  ["genetic epidemiology. supplement"] = "Genet Epidemiol Suppl",
  ["genetic programming and evolvable machines"] = "Genet Program Evolvable Mach",
  ["genetic psychology monographs"] = "Genet Psychol Monogr",
  ["genetic resources and crop evolution"] = "Genet Resour Crop Evol",
  ["genetic testing"] = "Genet Test",
  ["genetic testing and molecular biomarkers"] = "Genet Test Mol Biomarkers",
  ["genetic vaccines and therapy"] = "Genet Vaccines Ther",
  ["genetic, social, and general psychology monographs"] = "Genet Soc Gen Psychol Monogr",
  ["genetica iberica"] = "Genet Iber",
  ["genetical research"] = "Genet Res",
  ["genetics & epigenetics"] = "Genet Epigenet",
  ["genetics and genetic engineering"] = "Genet Genet Eng",
  ["genetics and molecular biology"] = "Genet Mol Biol",
  ["genetics and molecular research"] = "Genet. Mol. Res.",
  ["genetics and molecular research : gmr"] = "Genet Mol Res",
  ["genetics and plant physiology"] = "Genet Plant Physiol",
  ["genetics in medicine"] = "Genet. Med.",
  ["genetics in medicine : official journal of the american college of medical genetics"] = "Genet Med",
  ["genetics of aquatic organisms"] = "Genet Aquat Org",
  ["genetics research"] = "Genet Res (Camb)",
  ["genetics research international"] = "Genet Res Int",
  ["genetics selection evolution"] = "Genet. Sel. Evol.",
  ["genetics, selection, evolution : gse"] = "Genet Sel Evol",
  ["genetics, selection, evolution. : gse"] = "Genet Sel Evol",
  ["genetika (moscow)"] = "Genetika (Moscow)",
  ["geneva papers on risk and insurance theory"] = "Geneva Pap. Risk Ins. Theory",
  ["geneva papers on risk and insurance: issues and practice"] = "Geneva Pap. Risk Ins.: Issues Practice",
  ["geneve-afrique. geneva-africa"] = "Geneva Afr",
  ["genewatch : a bulletin of the committee for responsible genetics"] = "Genewatch",
  ["genève-afrique. geneva-africa"] = "Geneva Afr",
  ["genitourinary medicine"] = "Genitourin Med",
  ["genome announcements"] = "Genome Announc",
  ["genome biology"] = "Genome Biol",
  ["genome biology and evolution"] = "Genome Biol Evol",
  ["genome dynamics"] = "Genome Dyn",
  ["genome dynamics and stability"] = "Genome Dyn Stab",
  ["genome informatics"] = "Genome Inform.",
  ["genome informatics. international conference on genome informatics"] = "Genome Inform",
  ["genome informatics. workshop on genome informatics"] = "Genome Inform Ser Workshop Genome Inform",
  ["genome instability & disease"] = "Genome Instab Dis",
  ["genome integrity"] = "Genome Integr",
  ["genome letters"] = "Genome Lett",
  ["genome medicine"] = "Genome Med",
  ["genome research"] = "Genome Res",
  ["genome science & technology"] = "Genome Sci Technol",
  ["genomic medicine"] = "Genomic Med",
  ["genomic medicine, biomarkers, and health science"] = "Genomic Med., Biomarkers, Heath Sci.",
  ["genomics & informatics"] = "Genomics Inform",
  ["genomics and applied biology"] = "Genomics Appl Biol",
  ["genomics and computational biology"] = "Genom Comput Biol",
  ["genomics and genetics : the official journal of genetics society of thailand"] = "Genom Genet",
  ["genomics data"] = "Genom Data",
  ["genomics discovery"] = "Genom Discov",
  ["genomics insights"] = "Genomics Insights",
  ["genomics proteomics & bioinformatcs"] = "Genomics Proteomics Bioinf.",
  ["genomics, proteomics & bioinformatics"] = "Genomics Proteomics Bioinformatics",
  ["genomics, proteomics & bioinformatics / beijing genomics institute"] = "Genomics Proteomics Bioinformatics",
  ["genomics, proteomics and bioinformatics"] = "Genomics Proteomics Bioinformatics",
  ["genomics, society, and policy"] = "Genom Soc Policy",
  ["genomome instability & disease"] = "Genome Instab. Dis.",
  ["genre (los angeles, calif.)"] = "Genre",
  ["gensei dobutsugaku zasshi"] = "Gensei Dobutsugaku Zasshi",
  ["genshikaku kenkyu"] = "Genshikaku Kenkyu",
  ["geo : geography and environment"] = "Geo",
  ["geo-marine letters"] = "Geo-Mar. Lett.",
  ["geo-spatial information science"] = "Geo-spatial Inf. Sci.",
  ["geo-spatial information science = diqui kongjian xinxi kexue xuebao"] = "Geo Spat Inf Sci",
  ["geobios (lyon, france). memoire special"] = "Geobios Mem Spec",
  ["geobios (lyon, france). mémoire spécial"] = "Geobios Mem Spec",
  ["geobotanica selecta"] = "Geobot. sel.",
  ["geocarto international"] = "Geocarto Int",
  ["geochemical journal"] = "Geochem. J.",
  ["geochemical perspectives"] = "Geochem. Perspect.",
  ["geochemical perspectives letters"] = "Geochem Perspect Lett",
  ["geochemical transactions"] = "Geochem Trans",
  ["geochemistry international"] = "Geochem. Int.",
  ["geochemistry, geophysics, geosystems"] = "Geochem. Geophys. Geosyst.",
  ["geochemistry, geophysics, geosystems : g³"] = "Geochem Geophys Geosyst",
  ["geochemistry: exploration, environment, analysis"] = "Geochem. Explor. Environ. Anal.",
  ["geochimica et cosmochimica acta"] = "Geochim. Cosmochim. Acta",
  ["geodermal regional"] = "Geodermal Reg.",
  ["geodesy and cartography"] = "Geod. Cartogr.",
  ["geodesy and geodynamics"] = "Geod Geodyn",
  ["geodinamica acta"] = "Geodin. Acta",
  ["geoenergy science and engineering"] = "Geoenergy Sci. Eng.",
  ["geoenvironmental disasters"] = "Geoenvironmental Disasters",
  ["geofisica internacional"] = "Geofis. Int.",
  ["geofisica internationale"] = "Geofis. Int.",
  ["geofisica pura e applicata"] = "Geofis. pura appl.",
  ["geoforum; journal of physical, human, and regional geosciences"] = "Geoforum",
  ["geofysiske publikasjoner"] = "Geofys. Publ.",
  ["geofãsica internacional"] = "GeofÃs. Int.",
  ["geofã­sica internacional"] = "GeofÃ­s. Int.",
  ["geografický časopis = geograficheskiĭ zhurnal = geographical review"] = "Geogr Cas",
  ["geografiia v shkole (moscow, russia)"] = "Geogr Shkole",
  ["geografii︠a︡ v shkole (moscow, russia)"] = "Geogr Shkole",
  ["geografisch tijdschrift"] = "Geogr Tydschr",
  ["geografisk tidskrift"] = "Geogr Tidsskr",
  ["geografisk tidskrift / udgivet af bestyrelsen for det kongelige danske geografiske selskab"] = "Geogr Tidsskr",
  ["geografisk tidsskrift - danish journal of geography"] = "Geogr. Tidsskr. - Dan. J. Geogr.",
  ["geografiska annaler"] = "Geogr. Ann.",
  ["geografiska annaler. series b, human geography"] = "Geogr Ann Ser B",
  ["geografiska annaler: series a, physical geography"] = "Geogr. Ann. Ser. A  Phys. Geogr.",
  ["geografiska annaler: series b, human geography"] = "Geogr. Ann. Ser. B  Hum. Geogr.",
  ["geografski glasnik = bulletin de géographie"] = "Geog Glas",
  ["geografski pregled. revue de geographie"] = "Geogr Pregl",
  ["geografski pregled. revue de géographie"] = "Geogr Pregl",
  ["geografski vestnik = bulletin de la société de géographie de ljubljana"] = "Geogr Vestn",
  ["geographia antiqua"] = "GeorgAnt",
  ["geographia antiqua. rivista di geografia storica del mondo antico e di storia della geografia"] = "GeoAnt",
  ["geographia medica"] = "Geogr Med",
  ["geographia medica. supplement"] = "Geogr. Med. Suppl.",
  ["geographia medica. supplement = geographia medica. sonderband"] = "Geogr Med Suppl",
  ["geographia polonica"] = "Geogr Pol",
  ["geographica (madrid, spain)"] = "Geographica",
  ["geographica bernensia"] = "Geogr. Bern.",
  ["geographica helvetica"] = "Geogr Helv",
  ["geographical analysis"] = "Geogr Anal",
  ["geographical bulletin"] = "Geogr. Bull.",
  ["geographical journal"] = "Geogr. J.",
  ["geographical journal, the"] = "Geogr. J.",
  ["geographical perspectives"] = "Geogr Perspect",
  ["geographical research"] = "Geogr. Res.",
  ["geographical review"] = "Geogr Rev",
  ["geographical review of india"] = "Geogr Rev India",
  ["geographical review of japan"] = "Geogr Rev Jpn",
  ["geographical review, the"] = "Geogr. Rev.",
  ["geographical survey"] = "Geogr Surv",
  ["geographie et recherche"] = "Geogr Rech",
  ["geographie physique et quaternaire"] = "Geogr. Phys. Quat.",
  ["geographische berichte"] = "Geogr Ber",
  ["geographische rundschau"] = "Geogr Rundsch",
  ["geographische zeitschrift"] = "Geogr Z",
  ["geography (sheffield, england)"] = "Geography",
  ["geography and environment"] = "Geogr. Environ.",
  ["geography and natural resources"] = "Geogr. Nat. Resour.",
  ["geography and sustainability"] = "Geogr. Sustainability",
  ["geography bulletin"] = "Geogr Bull",
  ["geography compass"] = "Geogr Compass",
  ["geography review"] = "Geogr Rev (Oxf)",
  ["geography, environment, sustainability"] = "Geogr. Environ. Sustainability",
  ["geojournal of tourism and geosites"] = "Geoj. Tourism Geosites",
  ["geologica belgica"] = "Geol. Belg.",
  ["geologica carpathica : the journal of geological institute of slovak academy of sciences"] = "Geol Carpath",
  ["geologica saxonica"] = "Geol. Saxonica",
  ["geologica ultraiectina (gedrukt). rijksuniversiteit faculteit aardwetenschappen"] = "Geol Ultraiectina (Gedrukt)",
  ["geological journal"] = "Geol. J.",
  ["geological magazine"] = "Geol. Mag.",
  ["geological society of america bulletin"] = "Geol. Soc. Am. Bull.",
  ["geological society of america special papers"] = "Geol. Soc. Spec. Pap.",
  ["geological society special publication"] = "Geol. Soc. Spec. Publ.",
  ["geological society special publications"] = "Geol. Soc. Spec.Publ.",
  ["geological society, london, special publications"] = "Geol. Soc. London Spec. Publ.",
  ["geological survey of finland special paper"] = "Geol. Surv. Finl. Spec. Pap.",
  ["geologie en mijnbouw"] = "Geol. Mijnbouw",
  ["geologische rundschau"] = "Geol. Rundsch.",
  ["geologische rundschau : zeitschrift fur allgemeine geologie"] = "Geol Rundsch",
  ["geologische rundschau : zeitschrift für allgemeine geologie"] = "Geol Rundsch",
  ["geologisches jahrbuch"] = "Geol. Jahrb.",
  ["geologiska foereningens i stockholm foerhandlingar"] = "GFF",
  ["geologiska foreningen i stockholm forhandlingar"] = "GFF",
  ["geologiya i geofizika"] = "Geol. Geofiz.",
  ["geology of ore deposits"] = "Geol. Ore Deposits",
  ["geology today"] = "Geol. Today",
  ["geology, ecology, and landscapes"] = "Geol. Ecol. Landscapes",
  ["geomagnetism and aeronomy"] = "Geomag. Aeron.",
  ["geomatics, natural hazards and risk"] = "Geomatics Nat. Hazards Risk",
  ["geomechanics and engineering"] = "Geomech. Eng.",
  ["geomechanics and geoengineering"] = "Geomech. Geoeng.",
  ["geomechanics and geophysics for geo-energy and geo-resources"] = "Geomech. Geophys. Geo-Energy Geo-Resour.",
  ["geomechanics and tunnelling"] = "Geomech. Tunnelling",
  ["geomechanics for energy and the environment"] = "Geomech. Energy Environ.",
  ["geometriae dedicata"] = "Geom. Dedicata",
  ["geometric and functional analysis"] = "Geom. Funct. Anal.",
  ["geometric and functional analysis : gafa"] = "Geom Funct Anal",
  ["geometric design publications"] = "Geom. Des. Publ.",
  ["geometric flows"] = "Geom. Flows",
  ["geometric mechanics"] = "Geom. Mech.",
  ["geometric science of information : first international conference, gsi 2013, paris, france, august 28-30, 2013 : proceedings. geometric science of information (1st : 2013 : paris, france)"] = "Geom Sci Inf (2013)",
  ["geometry & topology"] = "Geom. Topol",
  ["geometry & topology monographs"] = "Geom. Topol. Monogr.",
  ["geometry and computing"] = "Geom. Comput.",
  ["geometry and topology"] = "Geom. Topol.",
  ["geometry, imaging and computing"] = "Geom. Imaging Comput.",
  ["geometry, integrability and quantization"] = "Geom. Integrability Quantization",
  ["geometry, integrability and quantization, proceedings series"] = "Geom. Integrability Quantization Proc. Ser.",
  ["geometry, topology & physics"] = "Conf. Proc. Lect. Not. G.",
  ["geomicrobiology journal"] = "Geomicrobiol J",
  ["geomorphologie relief processus environnement"] = "Geomorphol. Relef Processus Environ.",
  ["geomorphology (amsterdam, netherlands)"] = "Geomorphology (Amst)",
  ["geophysical and astrophysical fluid dynamics"] = "Geophys. Astrophys. Fluid Dyn.",
  ["geophysical fluid dynamics"] = "Geophys. Fluid Dyn.",
  ["geophysical journal"] = "Geophys. J.",
  ["geophysical journal international"] = "Geophys. J. Int.",
  ["geophysical journal of the royal astronomical society"] = "Geophys. J. R. Astron. Soc.",
  ["geophysical magazine"] = "Geophys. Mag.",
  ["geophysical monograph"] = "Geophys. Monogr.",
  ["geophysical prospecting"] = "Geophys. Prospect.",
  ["geophysical research letters"] = "Geophys. Res. Lett.",
  ["geophysical supplements to the monthly notices of the royal astronomical society"] = "Geophys. Suppl. Mon. Not. R. Astron. Soc.",
  ["geoplanet: earth and planetary sciences"] = "GeoPlanet Earth Planet. Sci.",
  ["george mason university civil rights law journal"] = "Geoge Mason Univ Civ Rights Law J",
  ["george mason university law review"] = "George Mason Univ Law Rev",
  ["georgetown dental journal"] = "Georgetown Dent J",
  ["georgetown immigration law journal"] = "Georget Immgr Law J",
  ["georgetown law journal"] = "Georgetown Law J.",
  ["georgetown magazine"] = "Georget Mag",
  ["georgetown medical bulletin"] = "Georgetown Med Bull",
  ["georgetown university school of dentistry mirror"] = "Georgetown Univ. Sch. Dent. Mirror",
  ["georgia epidemiology report"] = "Ga Epidemiol Rep",
  ["georgia epidemiology report / georgia dept. of human resources, office of epidemiology"] = "Ga Epidemiol Rep",
  ["georgia hospitals today"] = "Ga Hosp Today",
  ["georgia hospitals today / gha"] = "Ga Hosp Today",
  ["georgia journal of science : official publication of the georgia academy of science"] = "Ga J Sci",
  ["georgia law review (athens, ga. : 1966)"] = "Georgia Law Rev",
  ["georgia nursing"] = "Ga Nurse",
  ["georgia pharmacist quarterly"] = "Ga Pharm Q",
  ["georgia state bar journal"] = "Georgia State Bar J",
  ["georgia state university law review"] = "Ga State Univ Law Rev",
  ["georgian academy of sciences"] = "Mem. Differential Equations Math. Phys.",
  ["georgian mathematical journal"] = "Georgian Math. J.",
  ["georgian medical news"] = "Georgian Med News",
  ["georisk: assessment and management of risk for engineered systems and geohazards"] = "Georisk: Assess. Manage. Risk Eng. Syst. Geohazards",
  ["geoscience canada"] = "Geosci. Can.",
  ["geoscience communication"] = "Geosci. Comm.",
  ["geoscience data journal"] = "Geosci. Data J.",
  ["geoscience frontiers"] = "Geosci. Front.",
  ["geoscience letters"] = "Geosci Lett",
  ["geosciences journal"] = "Geosci. J.",
  ["geoscientific instrumentation, methods and data systems"] = "Geosci. Instrum. Methods Data Syst.",
  ["geoscientific instrumentation, methods and data systems discussions"] = "Geosci. Instrum. Methods Data Syst. Discuss.",
  ["geoscientific model development"] = "Geosci Model Dev",
  ["geoscientific model development discussions"] = "Geosci. Model Dev. Discuss.",
  ["geospatial health"] = "Geospat Health",
  ["geosphere (boulder, colo.)"] = "Geosphere (Boulder)",
  ["geostandards and geoanalytical research"] = "Geostand. Geoanal. Res.",
  ["geosynthetics international"] = "Geosynth Int",
  ["geosystem engineering"] = "Geosyst. Eng.",
  ["geosystems and geoenvironment"] = "Geosyst. Geoenviron.",
  ["geosystems mathematics"] = "Geosyst. Math.",
  ["geotechnical and geological engineering"] = "Geotech. Geol. Eng.",
  ["geotechnical testing journal"] = "Geotech. Test. J.",
  ["geotextiles and geomembranes"] = "Geotext. Geomembr.",
  ["geothermal energy"] = "Geotherm. Energy",
  ["geothermal energy science"] = "Geotherm. Energy Sci.",
  ["gep news"] = "GEP News",
  ["geriatric nephrology and urology"] = "Geriatr Nephrol Urol",
  ["geriatric nursing"] = "Geriatr Nurs (Minneap)",
  ["geriatric nursing (london, england)"] = "Geriatr Nurs (Lond)",
  ["geriatric nursing (new york, n.y.)"] = "Geriatr Nurs",
  ["geriatric nursing and home care"] = "Geriatr Nurs Home Care",
  ["geriatric orthopaedic surgery & rehabilitation"] = "Geriatr Orthop Surg Rehabil",
  ["geriatrics & aging"] = "Geriatr Aging",
  ["geriatrics & gerontology international"] = "Geriatr Gerontol Int",
  ["geriatrics (basel, switzerland)"] = "Geriatrics (Basel)",
  ["gerlands beitrage zur geophysik"] = "Gerlands Beitr. Geophys.",
  ["german comments"] = "Ger Comments",
  ["german economic review"] = "Ger Econ Rev",
  ["german history : the journal of the german history society"] = "Ger Hist",
  ["german journal of agricultural economics"] = "Ger. J. Agric. Econ.",
  ["german journal of exercise and sport research"] = "Ger. J. Exercise Sport Res.",
  ["german journal of ophthalmology"] = "Ger J Ophthalmol",
  ["german journal of veterinary research"] = "Ger. J. Vet. Res.",
  ["german life and letters"] = "Ger Life Lett",
  ["german medical monthly"] = "Ger Med Mon",
  ["german medical science : gms e-journal"] = "Ger Med Sci",
  ["german medicine"] = "Ger Med",
  ["german politics"] = "Ger Polit",
  ["german research"] = "Ger. Res.",
  ["german studies review"] = "Ger Stud Rev",
  ["german studies: a review of german-language research contributions. sect. l: philosophy and history"] = "GS",
  ["germania. anzeiger der römisch-germanischen kommission des deutschen archäologischen instituts"] = "Germania",
  ["germanisch-romanische monatsschrift"] = "Ger Rom Monatsschr",
  ["germantowne crier"] = "Germant Crier",
  ["gerontechnology : international journal on the fundamental aspects of technology to serve the ageing society"] = "Gerontechnology",
  ["gerontologia clinica"] = "Gerontol Clin (Basel)",
  ["gerontologie et societe : cahiers de la fondation nationale de gerontologie"] = "Gerontol Soc",
  ["gerontology & geriatric medicine"] = "Gerontol Geriatr Med",
  ["gerontology & geriatrics education"] = "Gerontol Geriatr Educ",
  ["gerontology and geriatrics education"] = "Gerontol. Geriatr. Educ.",
  ["geschichte des hospitals"] = "Gesch Hosp",
  ["geschichte in wissenschaft und unterricht"] = "Gesch Wiss Unterr",
  ["geschichte und gegenwart"] = "Gesch Ggw (Graz)",
  ["geschichte und gesellschaft (vandenhoeck & ruprecht)"] = "Gesch Ges",
  ["geschichte und region = storia e regione"] = "Gesch Reg",
  ["gesetzgebung des bundes und der kantone"] = "Gesetzgeb. Bundes Kantone",
  ["gesnerus. supplement"] = "Gesnerus Suppl",
  ["gesprächsforschung : online-zeitschrift zur verbalen interaktion"] = "Gesprachsforschung",
  ["gesta. international center of medieval art"] = "Gesta",
  ["gestalt theory"] = "Gestalt Theory",
  ["gesture (amsterdam, netherlands)"] = "Gesture (Amst)",
  ["gesunde pflanz."] = "Gesunde Pflanzen",
  ["gesundheit (derendingen, switzerland)"] = "Gesundheit Derending",
  ["gesundheit und wohlfahrt. revue suisse d'hygiène"] = "Gesund Wohlfahrt",
  ["gesundheits- und umwelttechnik"] = "Gesund. Umwelttech.",
  ["gesundheitswesen (bundesverband der arzte des offentlichen gesundheitsdienstes (germany))"] = "Gesundheitswesen",
  ["gesundheitswesen (bundesverband der ärzte des öffentlichen gesundheitsdienstes (germany))"] = "Gesundheitswesen",
  ["gesundsheitsschutz und umwelttechnik"] = "Gesundheitsschutz Umwelttech.",
  ["getmobile (new york, n.y.)"] = "GetMobile",
  ["geus bulletin"] = "GEUS Bull.",
  ["gewaesserschutz, wasser, abwasser"] = "Gewaesserschutz, Wasser, Abwasser",
  ["génétique, sélection, évolution"] = "Genet Sel Evol",
  ["géographie et recherche"] = "Geogr Rech",
  ["gériatrie et psychologie neuropsychiatrie du vieillissement"] = "Geriatr Psychol Neuropsychiatr Vieil",
  ["gérontologie et société : cahiers de la fondation nationale de gérontologie"] = "Gerontol Soc",
  ["gha today"] = "GHA Today",
  ["ghaa journal"] = "GHAA J",
  ["ghana journal of sociology"] = "Ghana J Sociol",
  ["ghana medical journal"] = "Ghana Med J",
  ["ghana official news bulletin"] = "Ghana Off News Bull",
  ["ghana social science journal"] = "Ghana Soc Sci J",
  ["ghanaian nurse"] = "Ghana. Nurse",
  ["gi-edition. proceedings"] = "GI Ed Proc",
  ["gibier, faune sauvage"] = "Gibier Faune Sauvage",
  ["gidrokhimicheskie materialy"] = "Gidrokhim. Mater.",
  ["giessener geographische schriften"] = "Giess. Geogr. Schr.",
  ["gifted and talented international"] = "Gifted Talent Int",
  ["gifu shika gakkai zasshi (journal of gifu dental society)"] = "Gifu Shika Gakkai Zasshi",
  ["gifu shika gakkai zasshi = the journal of gifu dental society"] = "Gifu Shika Gakkai Zasshi",
  ["gifu university"] = "Bull. Fac. Gen. Ed. Gifu Univ.",
  ["gigiena i sanitariia"] = "Gig Sanit",
  ["gigiena i sanitariya"] = "Gig. Sanit.",
  ["gigiena truda i professional'nye zabolevaniia"] = "Gig Tr Prof Zabol",
  ["gigiena truda i professionalnye zabolevaniia"] = "Gig. Tr. Prof. Zabol.",
  ["gigiena truda i professional’nye zabolevaniia"] = "Gig Tr Prof Zabol",
  ["gigiena truda i professional’nye zabolevaniya"] = "Gig. Tr. Prof. Zabol.",
  ["giglio di roccia"] = "GiRoccPalermo",
  ["ginecologia clinica"] = "Ginecol Clin",
  ["ginecologia y obstetricia"] = "Ginecol Obstet (Lima)",
  ["ginecologia y obstetricia de mexico"] = "Ginecol Obstet Mex",
  ["ginecología y obstetricia de méxico"] = "Ginecol Obstet Mex",
  ["ginekologia polska"] = "Ginekol Pol",
  ["giornale botanico italiano (florence, italy : 1962)"] = "G Bot Ital (Florence Italy)",
  ["giornale critico della filosofia italiana"] = "G Crit Filos Ital",
  ["giornale de techniche nefrologiche & dialitiche"] = "G Tec Nefrol Dial",
  ["giornale degli economisti e annali di economia"] = "G Econ Ann Econ",
  ["giornale degli economisti e annali di economia / universita commerciale luigi bocconi"] = "G Econ Ann Econ",
  ["giornale del popolo"] = "G. Pop.",
  ["giornale della accademia di medicina di torino"] = "G Accad Med Torino",
  ["giornale di anestesia stomatologica"] = "G. Anest. Stomatol.",
  ["giornale di anestesia stomatologica = journal of dental anaesthesia"] = "G Anest Stomatol",
  ["giornale di batteriologia e immunologia"] = "G Batteriol Immunol",
  ["giornale di batteriologia, virologia ed immunologia"] = "G Batteriol Virol Immunol",
  ["giornale di batteriologia, virologia ed immunologia ed annali dell ospedale maria vittoria di torino"] = "G. Batteriol. Virol. Immunol.",
  ["giornale di batteriologia, virologia ed immunologia ed annali dell ospedale maria vittoria di torino. parte 2, sezione clinica"] = "G. Batteriol. Virol. Immunol. [Clin.]",
  ["giornale di batteriologia, virologia ed immunologia ed annali dell'ospedale maria vittoria di torino. parte 1: sezione microbiologica"] = "G Batteriol Virol Immunol Microbiol",
  ["giornale di batteriologia, virologia ed immunologia ed annali dell'ospedale maria vittoria di torino. parte 2: sezione clinica"] = "G Batteriol Virol Immunol Clin",
  ["giornale di batteriologia, virologia ed immunologia ed annali dell’ospedale maria vittoria di torino. parte 1: sezione microbiologica"] = "G Batteriol Virol Immunol Microbiol",
  ["giornale di batteriologia, virologia ed immunologia ed annali dell’ospedale maria vittoria di torino. parte 2: sezione clinica"] = "G Batteriol Virol Immunol Clin",
  ["giornale di batteriologia, virologia, ed immunologia ed annali dell'ospedale maria vittoria di torino"] = "G Batteriol Virol Immunol",
  ["giornale di batteriologia, virologia, ed immunologia ed annali dell’ospedale maria vittoria di torino"] = "G Batteriol Virol Immunol",
  ["giornale di biochimica"] = "G Biochim",
  ["giornale di chirurgia"] = "G. Chir.",
  ["giornale di clinica medica"] = "G Clin Med",
  ["giornale di endodonzia"] = "G Endodonzia",
  ["giornale di fisica sanitaria e protezione contro le radiazioni"] = "G. Fis. Sanit. Prot. Radiaz.",
  ["giornale di fisica sanitaria e protezione contro le radiazioni. the journal of health physics and radiation protection"] = "G Fis Sanit Prot Radiaz",
  ["giornale di gerontologia"] = "G Gerontol",
  ["giornale di gerontologia. supplemento"] = "G Gerontol Suppl",
  ["giornale di igiene e medicina preventiva"] = "G Ig Med Prev",
  ["giornale di malattie infettive e parassitarie"] = "G Mal Infett Parassit",
  ["giornale di medicina militare"] = "G Med Mil",
  ["giornale di metafisica"] = "GM",
  ["giornale di pneumologia"] = "G Pneumol",
  ["giornale di psichiatria e di neuropatologia"] = "G Psichiatr Neuropatol",
  ["giornale di scienze mediche"] = "G Sci Mediche",
  ["giornale di stomatologia e di ortognatodonzia"] = "G Stomatol Ortognatodonzia",
  ["giornale ferrarese di retorica e filologia"] = "GFRF",
  ["giornale filologico ferrarese"] = "GiornFilFerr",
  ["giornale italiano dell'arteriosclerosi"] = "G Ital Arterioscler",
  ["giornale italiano della tubercolosi"] = "G Ital Della Tuberc",
  ["giornale italiano della tubercolosi e delle malattie del torace"] = "G Ital Tuberc Mal Torace",
  ["giornale italiano delle malattie del torace"] = "G Ital Mal Torace",
  ["giornale italiano di anestesiologia"] = "G Ital Anestesiol",
  ["giornale italiano di cardiologia"] = "G Ital Cardiol",
  ["giornale italiano di cardiologia (2006)"] = "G Ital Cardiol (Rome)",
  ["giornale italiano di chemioterapia"] = "G Ital Chemioter",
  ["giornale italiano di chirurgia"] = "G Ital Chir",
  ["giornale italiano di dermatologia"] = "G Ital Dermatol",
  ["giornale italiano di dermatologia e venereologia"] = "G. Ital. Dermatol. Venereol.",
  ["giornale italiano di dermatologia e venereologia : organo ufficiale, societa italiana di dermatologia e sifilografia"] = "G Ital Dermatol Venereol",
  ["giornale italiano di dermatologia e venereologia : organo ufficiale, società italiana di dermatologia e sifilografia"] = "G Ital Dermatol Venereol",
  ["giornale italiano di dermatologia. minerva dermatologica"] = "G. Ital. Dermatol. Minerva Dermatol.",
  ["giornale italiano di dermatolotia. minerva dermatologica"] = "G Ital Dermatol Minerva Dermatol",
  ["giornale italiano di endodonzia"] = "G Ital Endod",
  ["giornale italiano di filologia"] = "G Ital Filol",
  ["giornale italiano di filologia: rivista trimestrale di cultura"] = "GIF",
  ["giornale italiano di medicina del lavoro"] = "G Ital Med Lav",
  ["giornale italiano di medicina del lavoro ed ergonomia"] = "G Ital Med Lav Ergon",
  ["giornale italiano di nefrologia : organo ufficiale della societa italiana di nefrologia"] = "G Ital Nefrol",
  ["giornale italiano di nefrologia : organo ufficiale della società italiana di nefrologia"] = "G Ital Nefrol",
  ["giornale italiano di oftalmologia"] = "G Ital Oftalmol",
  ["giornale italiano di oncologia"] = "G Ital Oncol",
  ["giornale italiano di patologia e di scienze affini"] = "G Ital Patol",
  ["giornale storico della lunigiana e del territorio lucense"] = "GiornStorLun",
  ["giornale veneto di scienze mediche"] = "G Veneto Sci Med",
  ["gire. grupo de informacion en reproduccion"] = "GIRE",
  ["gire. grupo de información en reproducción"] = "GIRE",
  ["giscience & remote sensing"] = "GIsci Remote Sens",
  ["git labor-fachzeitschrift"] = "GIT Labor Fachz.",
  ["git, fachzeitschrift für das laboratorium"] = "GIT, Fachz. Lab.",
  ["giyāh/pizishkī (ahvāz, iran)"] = "Giyah Pizishki (Ahvaz)",
  ["gladius. estudios sobre armas antiguas, armamento, arte militar y vida cultural en oriente y occidente"] = "Gladius",
  ["gland surgery"] = "Gland Surg",
  ["glas. srpska akademija nauka i umetnosti, odeljenje medicinskih nauka"] = "Glas Srp Akad Nauka [Med]",
  ["glas. srpska akademija nauka i umetnosti. odeljenje medicinskih nauka"] = "Glas Srp Akad Nauka Med",
  ["glasgow archaeological journal"] = "GlasAJ",
  ["glasgow dental journal"] = "Glasg Dent J",
  ["glasgow mathematical journal"] = "Glasgow Math. J.",
  ["glasgow medical journal"] = "Glasgow Med J",
  ["glasnik (srpska akademija nauka)"] = "Glasnik",
  ["glasnik hemijskog drustva beograd = bulletin de la société chimique beograd"] = "Glas Hem Drus Beogr",
  ["glasnik matematicki. serija iii"] = "Glas. Mat. Ser. Iii",
  ["glasnik matematički"] = "Glas Mat",
  ["glasnik matematički"] = "Glas. Mat. Ser. III",
  ["glasnik zemaljskog muzeja bosne i hercegovie u sarajevu. arheologija"] = "GlasSarajevo",
  ["glasnik, srpska akademija nauka i umetnosti, odeljenje medicinskih nauka"] = "Glas. Srp. Akad. Nauka [Med.]",
  ["glasnik. belgrad. higijenski institut nr srbije"] = "Glas Belgrad Hig Inst NR Srb",
  ["glasnik. republicki zavod za zdravstvenu zastitu sr srbije"] = "Glas Zavodzdrav",
  ["glass and ceramics"] = "Glass Ceram.",
  ["glass physics and chemistry"] = "Glass Phys. Chem",
  ["glass research"] = "Glass Res.",
  ["glass science and technology"] = "Glass Sci. Technol.",
  ["glass science and technology (offenbach, germany)"] = "Glass Sci. Technol. (Offenbach, Ger.)",
  ["glass structures & engineering"] = "Glass Struct. Eng.",
  ["glass technology"] = "Glass Technol.",
  ["glass technology: european journal of glass science and technology, part a"] = "Glass Technol.: Eur. J. Glass Sci. Technol., Part A",
  ["glaucoma update"] = "Glaucoma Update",
  ["glendale law review"] = "Glendale Law Rev",
  ["glimpse (dhaka, bangladesh)"] = "Glimpse",
  ["glioma (mumbai, india)"] = "Glioma",
  ["global access to std diagnostics"] = "Glob Access STD Diagn",
  ["global advances in health and medicine"] = "Glob Adv Health Med",
  ["global aidsnews : the newsletter of the world health organization global programme on aids"] = "Glob AIDSnews",
  ["global and planetary change"] = "Glob Planet Change",
  ["global anesthesia and perioperative medicine"] = "Glob Anesth Perioper Med",
  ["global bioethics = problemi di bioetica"] = "Glob Bioeth",
  ["global biogeochemical cycles"] = "Global Biogeochem Cycles",
  ["global business and economics review"] = "Global Bus. Econ. Rev.",
  ["global cardiology science & practice"] = "Glob Cardiol Sci Pract",
  ["global center news"] = "Glob Cent News",
  ["global center news / center for women’s global leadership"] = "Glob Cent News",
  ["global challenges"] = "Global Challenges",
  ["global challenges (hoboken, nj)"] = "Glob Chall",
  ["global change biology"] = "Glob Chang Biol",
  ["global change biology. bioenergy"] = "Glob Change Biol Bioenergy",
  ["global child health news & review"] = "Glob Child Health New Rev",
  ["global dermatology"] = "Glob Dermatol",
  ["global drugs and therapeutics"] = "Glob Drugs Ther",
  ["global ecology and biogeography"] = "Glob. Ecol. Biogeogr.",
  ["global ecology and biogeography : a journal of macroecology"] = "Glob Ecol Biogeogr",
  ["global ecology and conservation"] = "Global Ecol. Conserv.",
  ["global economic review"] = "Global Econ. Rev.",
  ["global economy quarterly"] = "Global Economy Quart.",
  ["global energy interconnection"] = "Global Energy Interconnect.",
  ["global environmental change"] = "Glob. Environ. Change",
  ["global environmental change : human and policy dimensions"] = "Glob Environ Change",
  ["global environmental change part b: environmental hazards"] = "Global Environ. Change Part B: Environ. Hazards",
  ["global environmental politics"] = "Glob Environ Polit",
  ["global epidemiology"] = "Glob Epidemiol",
  ["global finance journal"] = "Global Finance J.",
  ["global food history"] = "Glob Food Hist",
  ["global food security"] = "Glob Food Sec",
  ["global futures digest"] = "Global Futures Dig",
  ["global health & medicine"] = "Glob Health Med",
  ["global health action"] = "Glob Health Action",
  ["global health communication"] = "Glob Health Commun",
  ["global health governance : the scholarly journal for the new health security paradigm"] = "Glob Health Gov",
  ["global health innovation"] = "Glob Health Innov",
  ["global health journal (amsterdam, netherlands)"] = "Glob Health J",
  ["global health promotion"] = "Glob Health Promot",
  ["global health research and policy"] = "Glob Health Res Policy",
  ["global health, epidemiology and genomics"] = "Glob Health Epidemiol Genom",
  ["global health, science and practice"] = "Glob Health Sci Pract",
  ["global heart"] = "Glob Heart",
  ["global imaging insights"] = "Glob Imaging Insights",
  ["global impacts : research for development review"] = "Glob Impacts",
  ["global implementation research and applications"] = "Global Implementation Res. Appl.",
  ["global issues (washington, d.c.)"] = "Glob Issues",
  ["global issues in water policy"] = "Glob. Issues Water Policy",
  ["global journal for research analysis"] = "Glob J Res Anal",
  ["global journal of addiction & rehabilitation medicine"] = "Glob J Addict Rehabil Med",
  ["global journal of advanced engineering technologies and sciences"] = "Glob J Adv Eng Technol Sci",
  ["global journal of advanced research on classical and modern geometries"] = "Glob. J. Adv. Res. Class. Mod. Geom.",
  ["global journal of animal breeding and genetics"] = "Glob J Anim Breed Genet",
  ["global journal of biochemistry"] = "Glob J Biochem",
  ["global journal of breast cancer research"] = "Global J Breast Cancer Res",
  ["global journal of chemistry"] = "Glob J Chem",
  ["global journal of community psychology practice"] = "Glob J Community Psychol Pract",
  ["global journal of environmental sciences"] = "Glob. J. Environ. Sci.",
  ["global journal of flexible systems management"] = "Global J. Flexible Syst. Manage.",
  ["global journal of geological sciences"] = "Global J. Geol. Sci.",
  ["global journal of health education and promotion"] = "Glob J Health Educ Promot",
  ["global journal of health science"] = "Glob J Health Sci",
  ["global journal of medical research"] = "Glob J Med Res",
  ["global journal of obesity, diabetes and metabolic syndrome"] = "Glob J Obes Diabetes Metab Syndr",
  ["global journal of orthopedics research"] = "Glob J Orthop Res",
  ["global journal of pediatrics & neonatal care"] = "Glob J Pediatr Neonatal Care",
  ["global journal of perioperative medicine"] = "Glob J Perioper Med",
  ["global journal of pure and applied sciences"] = "Global J. Pure Appl. Sci.",
  ["global journal of reproductive medicine"] = "Glob J Reprod Med",
  ["global journal of surgery"] = "Glob j Surg",
  ["global medical genetics"] = "Glob Med Genet",
  ["global medicine and therapeutics"] = "Glob Med Ther",
  ["global mental health (cambridge, england)"] = "Glob Ment Health (Camb)",
  ["global networks (oxford, england)"] = "Glob Netw (Oxf)",
  ["global networks-a journal of transnational affairs"] = "Glob. Netw.",
  ["global pediatric health"] = "Glob Pediatr Health",
  ["global policy"] = "Glob Policy",
  ["global psychiatry"] = "Glob Psychiatry",
  ["global public health"] = "Glob Public Health",
  ["global qualitative nursing research"] = "Glob Qual Nurs Res",
  ["global reproductive health"] = "Glob Reprod Health",
  ["global security: health, science and policy"] = "Global Secur.: Health, Sci. Policy",
  ["global social policy"] = "Glob Soc Policy",
  ["global social welfare : research, policy & practice"] = "Glob Soc Welf",
  ["global social work"] = "Glob Soc Work",
  ["global society : journal of interdisciplinary international relations"] = "Glob Soc",
  ["global spine journal"] = "Global Spine J",
  ["global studies quarterly"] = "Global Stud. Q.",
  ["global surgery (london)"] = "Glob Surg",
  ["global sustainability"] = "Global Sustainability",
  ["global transitions"] = "Glob Transit",
  ["global transitions proceedings"] = "Global Transitions Proc.",
  ["global vaccines and immunology"] = "Glob Vaccines Immunol",
  ["global veterinaria"] = "Glob Vet",
  ["globalization and health"] = "Global Health",
  ["globe. international college of dentists"] = "Globe",
  ["glomerular diseases"] = "Glomerular Dis",
  ["glossa (london)"] = "Glossa",
  ["glotta. zeitschrift für griechische und lateinische sprache"] = "Glotta",
  ["glycobiology insights"] = "Glycobiol Insights",
  ["glycoconjugate journal"] = "Glycoconj J",
  ["gm crops"] = "GM Crops",
  ["gm crops & food"] = "GM Crops Food",
  ["gmd research series"] = "GMD Res. Ser.",
  ["gmda bulletin"] = "GMDA Bull",
  ["gmhc treatment issues : the gay men's health crisis newsletter of experimental aids therapies"] = "GMHC Treat Issues",
  ["gmhc treatment issues : the gay men’s health crisis newsletter of experimental aids therapies"] = "GMHC Treat Issues",
  ["gms current topics in computer and robot assisted surgery"] = "GMS Curr Top Comput Robot Assist Surg",
  ["gms current topics in otorhinolaryngology, head and neck surgery"] = "GMS Curr Top Otorhinolaryngol Head Neck Surg",
  ["gms health innovation and technologies"] = "GMS Health Innov Technol",
  ["gms health technology assessment"] = "GMS Health Technol Assess",
  ["gms hygiene and infection control"] = "GMS Hyg Infect Control",
  ["gms infectious diseases"] = "GMS Infect Dis",
  ["gms interdisciplinary plastic and reconstructive surgery dgpw"] = "GMS Interdiscip Plast Reconstr Surg DGPW",
  ["gms journal for medical education"] = "GMS J Med Educ",
  ["gms krankenhaushygiene interdisziplinär"] = "GMS Krankenhhyg Interdiszip",
  ["gms ophthalmology cases"] = "GMS Ophthalmol Cases",
  ["gms zeitschrift für medizinische ausbildung"] = "GMS Z Med Ausbild",
  ["gmu law review"] = "GMU Law Rev",
  ["gnomon. kritische zeitschrift für die gesamte klassische altertumswissenschaft"] = "Gnomon",
  ["gnomon: kritische zeitschrift für die gesamte klassische altertumswissenschaft"] = "Gnomon",
  ["gnu radio"] = "GNU Radio",
  ["godisen zbornik na medicinskiot fakultet vo skopje"] = "God. Zb. Med. Fak. Skopje",
  ["godishen zbornik na institutot za matematika. prirodno-matematichki. fakultet universitet “sv. kiril i metodij”. skopje. annuaire de l’institut des mathematiques. faculte des sciences de l’universite “sts. kiril et metodij”. skopje"] = "God. Zb. Inst. Mat. Prir.-Mat. Fak. Univ. Kiril Metodij Skopje",
  ["godishnik na visshite uchebni zavedeniya"] = "Godishnik Vissh. Uchebn. Zaved. Prilozhna Mat.",
  ["godishnik na visshiya khimiko-tekhnologicheski institut, sofiya"] = "God. Vissh. Khim.-Tekhnol. Inst., Sofia",
  ["godisnjak vojnomedicinske akademije"] = "God. Vojnomed. Akad.",
  ["godisnjak vojnomedicinske akademije = annual of the military medical academy"] = "God Vojnomed Akad",
  ["godis̆en zbornik na medicinskiot fakultet vo skopje"] = "God Zb Med Fak Skopje",
  ["godišen zbornik na filozofskiot fakultet na universitetot vo skopje"] = "GodZborSkopje",
  ["godišnik na archeologičeski muzej plovdiv. annuaire du musée archéologique plovdiv"] = "GodMuzPlov",
  ["godišnik na departament archeologija"] = "GodDepA",
  ["godišnik na nacionalnija archeologičeski muzej. annuaire du musée national archéologique (sofia)"] = "GodMuzSof",
  ["gold bulletin"] = "Gold Bull",
  ["golden gate law review"] = "Gold Gate Law Rev",
  ["golden gate university law review. golden gate university. school of law"] = "Gold Gate Univ Law Rev",
  ["gondwana research"] = "Gondwana Res.",
  ["gondwana research : international geoscience journal"] = "Gondwana Res",
  ["gong ye wei sheng wu"] = "Industrial microbiology|Gong Ye Wei Sheng Wu",
  ["gongcheng shuxue xuebao"] = "Gongcheng Shuxue Xuebao",
  ["gongcheng shuxue xuebao. chinese journal of engineering mathematics"] = "Gongcheng Shuxue Xuebao",
  ["gongneng gaofenzi xuebao"] = "Gongneng Gaofenzi Xuebao",
  ["gonzaga law review"] = "Gonzaga Law Rev",
  ["good clinical practice journal : gcp"] = "Good Clin Pract J",
  ["good housekeeping (new york, n.y.)"] = "Good Housekeeping",
  ["gor\\cprime kovskiy gosudarstvennyy universitet imeni n"] = "Prikl. Problemy Proch. i Plast.",
  ["goriški letnik. zbornik goriškega muzeja"] = "GorLet",
  ["gorman lectures in economics"] = "Gorman Lect. Econ.",
  ["gospital'noe delo"] = "Gosp Delo",
  ["gospodarstvo istre"] = "Gospodarstvo Istre",
  ["goteborg university - department of sociology monograph"] = "Goteb. Univ. Dep. Socio.",
  ["goteborgs tandlakare-sallskaps arsbok"] = "Goteb. Tandlak. Sallsk. Arsb.",
  ["gothenburg papers in theoretical linguistics"] = "Gothenbg Pap Theor Linguist",
  ["gottinger jahrbuch"] = "Gott Jahrb",
  ["governance (oxford, england)"] = "Governance (Oxf)",
  ["government information quarterly"] = "Gov Inf Q",
  ["government publications review"] = "Gov Publ Rev",
  ["government publications review (new york, n.y. : 1982)"] = "Gov Publ Rev",
  ["government relations note"] = "Gov Relat Note",
  ["göttinger jahrbuch"] = "Gott Jahrb",
  ["gps solutions"] = "GPS Solutions",
  ["graduate faculty philosophy journal"] = "Grad Fac Philos J",
  ["graduate student series in physics"] = "Grad. Student Ser. Phys.",
  ["graduate studies in mathematics"] = "Grad. Stud. Math.",
  ["graduate texts in computer science"] = "Grad. Texts Comput. Sci.",
  ["graduate texts in contemporary physics"] = "Grad. Texts Contemp. Phys.",
  ["graduate texts in mathematics"] = "Grad. Texts in Math.",
  ["graduate texts in operations research"] = "Grad. Texts Oper. Res.",
  ["graduate texts in physics"] = "Grad. Texts Phys.",
  ["graecolatina et orientalia. zborník filozofickej fakulty univerzity komenského"] = "GrLatOr",
  ["graecolatina et orientalia: zborník filoz. fak. univerz. komenského"] = "GLO",
  ["graecolatina pragensia. acta universitatis carolinae. philologica"] = "GrLatPrag",
  ["graecolatina pragensia: acta universitatis carolinae"] = "GLP",
  ["graefe's archive for clinical and experimental ophthalmology = albrecht von graefes archiv für klinische und experimentelle ophthalmologie"] = "Graefes Arch Clin Exp Ophthalmol",
  ["graefes archive for clinical and experimental ophthalmology"] = "Graefes Arch Clin Exp Ophthalmol",
  ["graefe’s archive for clinical and experimental ophthalmology"] = "Graefes Arch. Clin. Exp. Ophthalmol.",
  ["graft (georgetown, tex.)"] = "Graft (Georget Tex)",
  ["grain & oil science and technology"] = "Grain Oil Sci. Technol.",
  ["grand rounds in urology"] = "Grand rounds Urol",
  ["grandfamilies : the contemporary journal of research, practice and policy"] = "Grandfamilies",
  ["grants magazine"] = "Grants Mag",
  ["grantsmanship center news"] = "Grantsmanship Cent News",
  ["granular computing"] = "Granular Comput.",
  ["granular matter"] = "Granular Matter",
  ["graph learning in medical imaging : first international workshop, glmi 2019, held in conjunction with miccai 2019, shenzhen, china, october 17, 2019, proceedings"] = "Graph Learn Med Imaging (2019)",
  ["graphene and 2d materials"] = "Graphene 2D Mater.",
  ["graphene technology"] = "Graphene Technol",
  ["graphical models"] = "Graph Models",
  ["graphics and visual computing"] = "Graphics Visual Comput.",
  ["graphics gems series"] = "Graph. Gems Ser.",
  ["graphis scripta"] = "Graph Scr",
  ["graphs and combinatorics"] = "Graphs Combin.",
  ["graphs in biomedical image analysis and integrating medical imaging and non-imaging modalities : second international workshop, grail 2018 and first international workshop, beyond mic 2018, held in conjunction with miccai 2018, granada, spain, september 20, 2018, proceedings. grail (workshop) (2nd : 2018 : granada, spain)"] = "Graphs Biomed Image Anal Integr Med Imaging Nonimaging Modalities (2018)",
  ["graphs in biomedical image analysis, computational anatomy and imaging genetics : first international workshop, grail 2017, 6th international workshop, mfca 2017, and third international workshop, micgen 2017, held in conjunction with miccai 2017, québec city, qc, canada, september 10-14, 2017, proceedings. grail (workshop) (1st : 2017 : québec, québec)"] = "Graphs Biomed Image Anal Comput Anat Imaging Genet (2017)",
  ["grasas y aceites"] = "Grasas Aceites",
  ["grasas y aceites (seville)"] = "Grasas Aceites (Seville)",
  ["grass and forage science"] = "Grass Forage Sci.",
  ["grass and forage science : the journal of the british grassland society"] = "Grass Forage Sci",
  ["grassland research"] = "Grassland Res.",
  ["grassland science"] = "Grassland Sci.",
  ["grassroots development"] = "Grassroots Dev",
  ["gratz college annual of jewish studies. gratz college"] = "Gratz Coll Annual Jew Stud",
  ["gravitation and cosmology"] = "Gravitation Cosmol.",
  ["gravitational and space biology bulletin : publication of the american society for gravitational and space biology"] = "Gravit Space Biol Bull",
  ["gravitational and space research : publication of the american society for gravitational and space research"] = "Gravit Space Res",
  ["grazer beiträge. zeitschrift für die klassische altertumswissenschaft"] = "GrazBeitr",
  ["grazer beiträge: zeitschrift für die klassische altertumswissenschaft"] = "GB",
  ["grazer mathematische berichte"] = "Grazer Math. Ber.",
  ["građa i rasprave. arheološki muzej istre, pula"] = "GraRaspr",
  ["grb coordinates network"] = "NASA-GCN",
  ["great lakes entomologist"] = "Great Lakes Entomol.",
  ["great ormond street journal"] = "Great Ormond St J",
  ["great plains research"] = "Great Plains Res.",
  ["greater milwaukee dental bulletin"] = "Greater Milw. Dent. Bull.",
  ["greater st louis dental society bulletin"] = "Greater St Louis Dent. Soc. Bull.",
  ["greece and rome"] = "G&R",
  ["greek economic review"] = "Greek Econ Rev",
  ["greek, roman and byzantine studies"] = "Greek Roman Byz Stud (Cambridge Mass)",
  ["green analytical chemistry"] = "Green Anal. Chem.",
  ["green and sustainable chemistry"] = "Green Sustainable Chem.",
  ["green chemical engineering"] = "Green Chem. Eng.",
  ["green chemical processing"] = "Green Chem. Process.",
  ["green chemistry"] = "Green Chem.",
  ["green chemistry : an international journal and green chemistry resource : gc"] = "Green Chem",
  ["green chemistry letters and reviews"] = "Green Chem. Lett. Rev.",
  ["green energy & environment"] = "Green Energy Environ.",
  ["green energy and environment technology"] = "Green Energy Environ. Technol.",
  ["green energy and intelligent transportation"] = "Green Energy Intell Transp.",
  ["green energy and resources"] = "Green Energy Resour.",
  ["green energy and technology"] = "Green Energy Technol.",
  ["green materials"] = "Green Mater.",
  ["green processing and synthesis"] = "Green Process. Synth,",
  ["green schools catalyst quarterly"] = "Green Schools Catal Q",
  ["green synthesis and catalysis"] = "Green Synth. Catal.",
  ["green technologies and sustainability"] = "Green Technol. Sustainability",
  ["green technology, resilience, and sustainability"] = "Green Technol. Resilience Sustainability",
  ["greenhouse gases: science and technology"] = "Greenhouse Gases Sci. Technol.",
  ["greenland geoscience"] = "Greenl. Geosci.",
  ["grenoble sciences"] = "Grenoble Sci.",
  ["grenzgebiete der medizin"] = "Grenzgeb Med",
  ["grenzgebiete der wissenschaft"] = "Grenzgeb Wiss",
  ["grey systems theory and application"] = "Grey Syst. Theory Appl.",
  ["grief matters"] = "Grief Matters",
  ["griffin’s statistical monograph series"] = "Griffin’s Statist. Monogr. Ser.",
  ["grma news"] = "GRMA News",
  ["groningen oriental studies"] = "Groningen Orient. Stud.",
  ["ground water"] = "Ground Water",
  ["ground water monitoring & remediation"] = "Ground Water Monit Remediat",
  ["ground water monitoring and remediation"] = "Ground Water Monit. Rem.",
  ["ground water monitoring review"] = "Ground Water Monit. Rev.",
  ["groundwater for sustainable development"] = "Groundw Sustain Dev",
  ["groundwater monitoring and remediation"] = "Groundwater Monit. Rem.",
  ["group & organization management"] = "Group Organ Manag",
  ["group decision and negotiation"] = "Group Decis Negot",
  ["group dynamics : theory, research, and practice : the official journal of division 49, group psychology and group psychotherapy of the american psychological association"] = "Group Dyn",
  ["group facilitation : a research & applications journal"] = "Group Facil",
  ["group health journal"] = "Group Health J.",
  ["group practice"] = "Group Pract",
  ["group practice journal"] = "Group Pract J",
  ["group processes & intergroup relations : gpir"] = "Group Process Intergroup Relat",
  ["groups, complexity, cryptology"] = "Groups Complexity Cryptology",
  ["groups, geometry, and dynamics"] = "Groups Geom. Dyn.",
  ["groups. complexity. cryptology"] = "Groups Complex. Cryptol.",
  ["groupwork : an interdisciplinary journal for working with groups"] = "Groupwork",
  ["growth and change"] = "Growth Change",
  ["growth and form"] = "Growth Form",
  ["growth factors"] = "Growth Factors",
  ["growth factors (chur, switzerland)"] = "Growth Factors",
  ["growth factors (yverdon-les-bains"] = "Growth Factors",
  ["growth horm. igf res."] = "Growth Hormone & IGF Research",
  ["growth hormone & igf research : official journal of the growth hormone research society and the international igf research society"] = "Growth Horm IGF Res",
  ["growth hormone and igf research"] = "Growth Horm. IGF Res.",
  ["growth regulation"] = "Growth Regul",
  ["growth, development and aging"] = "Growth. Dev. Aging",
  ["growth, development, and aging"] = "Growth Dev Aging",
  ["growth, development, and aging : gda"] = "Growth Dev Aging",
  ["growth, genetics & hormones"] = "Growth Genet Horm",
  ["grudnaia i serdechno-sosudistaia khirurgiia"] = "Grud Serdechnososudistaia Khir",
  ["grudnaia i serdechno-sosudistaia khirurgiia / ministerstvo zdravookhraneniia sssr [i] vsesoiuznoe nauchnoe obshchestvo khirurgov"] = "Grud Serdechnososudistaia Khir",
  ["grudnaia khirurgiia"] = "Grudn. Khir.",
  ["grudnaia khirurgiia (moscow, russia)"] = "Grudn Khir",
  ["grundfragen aus der silikoseforschung"] = "Grundfragen Silikoseforsch",
  ["grundkurs mathematik"] = "Grundkurs Math.",
  ["grundkurs physik"] = "Grundkurs Phys.",
  ["grundlagen der mathematik und informatik"] = "Grundlagen Math. Inform.",
  ["grundlehren der mathematischen wissenschaften"] = "Grundlehren Math. Wiss.",
  ["grundlehren text editions"] = "Grundlehren Text Ed.",
  ["grundwasser (berlin)"] = "Grundwasser (Berl)",
  ["grundwissen mathematik"] = "Grundwissen Math.",
  ["gruzlica (warsaw, poland : 1926)"] = "Gruzlica",
  ["gruzlica i choroby pluc"] = "Gruzlica",
  ["gruzlica i choroby pluc; tuberculosis et pneumonologia"] = "Gruzlica",
  ["gruźlica (warsaw, poland : 1926)"] = "Gruzlica",
  ["gruźlica i choroby płuc; tuberculosis et pneumonologia"] = "Gruzlica",
  ["grüne, die"] = "Grüne",
  ["gsa today : a publication of the geological society of america"] = "GSA Today",
  ["gstf international journal on bioinformatics & biotechnology"] = "GSTF Int J Bioinforma Biotechnol",
  ["gstf international journal on computing"] = "GSTF Int J Comput",
  ["gstf journal of advances in medical research"] = "GSTF J Adv Med Res",
  ["gstf journal of chemical sciences"] = "GSTF J. Chem. Sci.",
  ["gstf journal of nursing and health care"] = "GSTF J Nurs Health Care",
  ["gstf journal on media & communications"] = "GSTF J Media Commun",
  ["gu sheng wu xue bao"] = "Acta palaeontologica Sinica|Gu Sheng Wu Xue Bao",
  ["gu sheng wu xue bao = acta palaeontologica sinica"] = "Gu Sheng Wu Xue Bao",
  ["guam recorder"] = "Guam Rec",
  ["guang ming ri bao"] = "Guang Ming Ri Bao",
  ["guang pu xue yu guang pu fen xi = guang pu"] = "Guang Pu Xue Yu Guang Pu Fen Xi",
  ["guangdong hai yang da xue xue bao = journal of guangdong ocean university"] = "Guangdong Hai Yang Da Xue Xue Bao",
  ["guangdong nong ye ke xue = guangdong agricultural sciences"] = "Guangdong Nong Ye Ke Xue",
  ["guangdong yi xue = guangdong medical journal"] = "Guangdong Yi Xue",
  ["guangpuxue yu guangpu fenxi"] = "Guangpuxue Yu Guangpu Fenxi",
  ["guangxi nongye shengwu kexue"] = "Journal of agricultural and biological science|Guangxi Nongye Shengwu Kexue",
  ["guangxi nongye shengwu kexue = journal of agricultural and biological science"] = "Guangxi Nongye Shengwu Kexue",
  ["guangxi sciences"] = "Guangxi Sci.",
  ["guangxi xu mu shou yi"] = "Guangxi Xu Mu Shou Yi",
  ["guangxi yi ke da xue xue bao = guangxi yike daxue xuebao = acta academiae medicinae guangxi"] = "Guangxi Yi Ke Da Xue Xue Bao",
  ["guangxi zhi wu"] = "Guangxi Zhi Wu",
  ["guangxue xuebao"] = "Guangxue Xuebao",
  ["guatemala indigena"] = "Guatem Indig",
  ["guatemala indígena"] = "Guatem Indig",
  ["guatemala médica"] = "Guatem Med",
  ["guelph ichthyology reviews"] = "Guelph Ichthyol Rev",
  ["guerres mondiales et conflits contemporains"] = "Guerr Mond Conflits Contemp",
  ["guidance, navigation, and control"] = "Guid., Navig., Control",
  ["guide du praticien; edition du district parisien"] = "Guide Prat",
  ["guide du praticien; édition du district parisien"] = "Guide Prat",
  ["guide to microforms in print. author, title"] = "Guide Microforms Print Author Title",
  ["guide to reprints"] = "Guide Repr",
  ["guildhall studies in london history"] = "Guildhall Stud Lond Hist",
  ["guisuanyan xuebao"] = "Guisuanyan Xuebao",
  ["guizhou nong ye ke xue = journal of guizhou agricultural science"] = "Guizhou Nong Ye Ke Xue",
  ["guizhou shifan daxue xuebao"] = "Guizhou Shifan Daxue Xuebao Ziran Kexue Ban",
  ["gujarat statistical review"] = "Gujarat Stat Rev",
  ["gulf and caribbean research"] = "Gulf Caribb Res",
  ["gulf coast historical review"] = "Gulf Coast Hist Rev",
  ["gulf journal of mathematics"] = "Gulf J. Math.",
  ["gulf of mexico science"] = "Gulf Mex Sci",
  ["gunma journal of medical sciences"] = "Gunma J. Med. Sci.",
  ["guo ji bing du xue za zhi"] = "Guo Ji Bing Du Xue Za Zhi",
  ["guo ji fang she yi xue he yi xue za zhi = international journal of radiation medicine and nuclear medicine"] = "Guo Ji Fang She Yi Xue He Yi Xue Za Zhi",
  ["guo ji yi xue ji sheng chong bing za zhi"] = "Guo Ji Yi Xue Ji Sheng Chong Bing Za Zhi",
  ["guo li taiwan shi fan da xue di li yan jiu suo di li yan jiu bao gao. kuo li tʻai-wan shih fan ta hsüeh. ti li yen chiu so"] = "Guo Li Taiwan Shi Fan Da Xue Di li Yan Jiu Suo Di Li Yan Jiu Bao Gaob",
  ["guo li taiwan shi fan da xue di li yan jiu suo di li yan jiu bao gao. kuo li t’ai-wan shih fan ta hsueh. ti li yen chiu so"] = "Guo Li Taiwan Shi Fan Da Xue Di li Yan Jiu Suo Di Li Yan Jiu Bao Gaob",
  ["guocheng gongcheng xuebao"] = "Guocheng Gongcheng Xuebao",
  ["guoshu xuebao"] = "Journal of fruit science|Guoshu xuebao",
  ["guoshu xuebao = journal of fruit science"] = "Guoshu Xuebao",
  ["guru nanak journal of sociology"] = "Guru Nanak J Sociol",
  ["gut and liver"] = "Gut Liver",
  ["gut microbes"] = "Gut Microbes",
  ["gut microbiome"] = "Gut Microbiome",
  ["gut pathogens"] = "Gut Pathog",
  ["gutenberg jahrbuch"] = "Gutenberg Jahrb",
  ["guthrie clinic bulletin"] = "Guthrie Clin Bull",
  ["guy's hospital gazette"] = "Guys Hosp Gaz",
  ["guy's hospital reports"] = "Guys Hosp Rep",
  ["guy’s hospital gazette"] = "Guys Hosp Gaz",
  ["guy’s hospital reports"] = "Guys Hosp. Rep.",
  ["guʹoli̕ tʹaiwan da̕xuʹe zhʹiwu̕ bi̕ng cʹongha̕i xʹeukan = ntu phytopathologist & entomologist"] = "Guoli Taiwan Daxue Zhiwu Bing Chonghai Xeukan",
  ["gyermekgyógyászat : az orvosegészségügyi szakszervezet gyermekorvos szakcsoportjának folyóirata = pediatriia"] = "Gyermekgyogyaszat",
  ["gymnasium. zeitschrift für kultur der antike und humanistische bildung"] = "Gymnasium",
  ["gynaecologia. international monthly review of obstetrics and gynecology. revue internationale mensuelle d'obstétrique et de gynécologie. monatsschrift für geburtshilfe und gynäkologie"] = "Gynaecologia",
  ["gynaecologia. international monthly review of obstetrics and gynecology. revue internationale mensuelle d’obstetrique et de gynecologie. monatsschrift fur geburtshilfe und gynakologie"] = "Gynaecologia",
  ["gynakologisch-geburtshilfliche rundschau"] = "Gynakol Geburtshilfliche Rundsch",
  ["gynakologische praxis"] = "Gynakol Prax",
  ["gynakologische rundschau"] = "Gynakol. Rundsch.",
  ["gynäkologisch-geburtshilfliche rundschau"] = "Gynakol Geburtshilfliche Rundsch",
  ["gynäkologische endokrinologie"] = "Gynakol Endokrinol",
  ["gynäkologische praxis"] = "Gynakol Prax",
  ["gynäkologische rundschau"] = "Gynakol Rundsch",
  ["gynecologic and obstetric investigation"] = "Gynecol Obstet Invest",
  ["gynecologic investigation"] = "Gynecol Invest",
  ["gynecologic oncology"] = "Gynecol Oncol",
  ["gynecologic oncology case reports"] = "Gynecol Oncol Case Rep",
  ["gynecologic oncology reports"] = "Gynecol Oncol Rep",
  ["gynecologic oncology research and practice"] = "Gynecol Oncol Res Pract",
  ["gynecological endocrinology"] = "Gynecol. Endocrinol.",
  ["gynecological endocrinology : the official journal of the international society of gynecological endocrinology"] = "Gynecol Endocrinol",
  ["gynecological surgery"] = "Gynecol Surg",
  ["gynecologie et obstetrique"] = "Gynecol. Obstet. (Paris)",
  ["gynecologie pratique"] = "Gynecol. Prat.",
  ["gynecologie, obstetrique & fertilite"] = "Gynecol Obstet Fertil",
  ["gynecologie, obstetrique et fertilite"] = "Gynecol. Obstet. Fertil.",
  ["gynecology & obstetrics (sunnyvale, calif.)"] = "Gynecol Obstet (Sunnyvale)",
  ["gynecology & reproductive health"] = "Gynecol Reprod Health",
  ["gynecology and minimally invasive therapy"] = "Gynecol Minim Invasive Ther",
  ["gynecology and obstetrics research : open journal"] = "Gynecol Obstet Res",
  ["gynecology and pelvic medicine"] = "Gynecol Pelvic Med",
  ["gynecology and reproductive endocrinology"] = "Gynecol Reprod Endocrinol",
  ["gynécologie (paris, france : 1993)"] = "Gynecol Rev Gynecol",
  ["gynécologie et obstétrique"] = "Gynecol Obstet (Paris)",
  ["gynécologie pratique"] = "Gynecol Prat",
  ["gynécologie, obstétrique & fertilité"] = "Gynecol Obstet Fertil",
  ["gynécologie, obstétrique, fertilité & sénologie"] = "Gynecol Obstet Fertil Senol",
  ["gyoruigaku zasshi"] = "Gyoruigaku Zasshi",
  ["géographie physique et quaternaire"] = "Géogr. Phys. Quat.",
  ["göttinger forum für altertumswissenschaft"] = "GFA",
  ["göttinger miszellen. beiträge zur ägyptologischen diskussion"] = "GoettMisz",
  ["göttinger universitätsschriften"] = "Göttinger Universitätsschr. Ser. A Schr.",
  ["göttingische gelehrte anzeigen"] = "GGA",
  ["h. a. grueber, coins of the roman republic in the british museum i–iii (london 1910)"] = "BMCRRI-III",
  ["h. dessau, inscriptiones latinae selectae (berlin 1892–1916)"] = "ILS",
  ["h. mattingly (u. a.), coins of the roman empire in the british museum (london 1923–1950"] = "BMCRE",
  ["h. mattingly – e. a. sydenham, the roman imperial coinage"] = "RIC",
  ["h2open journal"] = "H2Open J.",
  ["ha'erbin yi ke da xue xue bao = journal of harbin medical university = harbin yike daxue xuebao"] = "Haerbin Yi Ke Da Xue Xue Bao",
  ["ha-ahot be-yisrael"] = "Ahot Beyisrael",
  ["habis. universidad de sevilla. arqueología, filología clásica"] = "Habis",
  ["habitat debate"] = "Habitat Debate",
  ["habitat debate / unchs (habitat), the united nations centre for human settlements"] = "Habitat Debate",
  ["habitat international"] = "Habitat Int",
  ["habitation (elmsford, n.y.)"] = "Habitation (Elmsford)",
  ["hacettepe bulletin of medicine/surgery"] = "Hacettepe Bull Med Surg",
  ["hacettepe bulletin of natural sciences and engineering"] = "Hacet. Bull. Nat. Sci. Eng. Ser. B",
  ["hacettepe journal of mathematics and statistics"] = "Hacet. J. Math. Stat.",
  ["hacettepe sosyal ve beseri bilimler dergisi"] = "Hacettepe Sosyal Bilimler Derg",
  ["hacettepe sosyal ve beşerî bilimler dergisi"] = "Hacettepe Sosyal Bilimler Derg",
  ["hadassah magazine"] = "Hadassah Mag",
  ["hadronic journal"] = "Hadronic J.",
  ["hadronic journal supplement"] = "Hadronic J. Suppl.",
  ["hadronic press monographs in mathematics"] = "Hadronic Press Monogr. Math.",
  ["hadtortenelmi kozlemenyek"] = "Hadtort Kozl",
  ["hadtörténelmi közlemények"] = "Hadtort Kozl",
  ["haematologica latina"] = "Haematol Lat",
  ["haematologica polonica"] = "Haematol Pol",
  ["haematologica reports"] = "Haematol Rep",
  ["haematology and blood transfusion"] = "Haematol Blood Transfus",
  ["haematology meeting reports"] = "Haematol Meet Rep",
  ["haemophilia : the official journal of the world federation of hemophilia"] = "Haemophilia",
  ["hahnemann hospital tidings"] = "Hahnemann Hosp Tidings",
  ["hahnemann monthly"] = "Hahnemann Mon",
  ["hai yang hu zhao tong bao = transactions of oceanology and limnology"] = "Hai Yang Hu Zhao Tong Bao",
  ["haiyang kexue = marine sciences"] = "Haiyang Kexue",
  ["haiyang yuye = marine fisheries"] = "Haiyang Yuye",
  ["hall's journal of health"] = "Halls J Health",
  ["hallisches winckelmannsprogramm"] = "HallWPr",
  ["hallym international journal of aging : hija"] = "Hallym Int J Aging HIJA",
  ["halsbury's statutes of england. great britain"] = "Halsburys Statut Engl G B",
  ["halsbury’s statutes of england. great britain"] = "Halsburys Statut Engl G B",
  ["hama. fouilles et recherches de la fondation carlsberg"] = "Hama",
  ["hamatologie und bluttransfusion"] = "Hamatol. Bluttransfus.",
  ["hamburger arzteblatt"] = "Hamb Arztebl",
  ["hamburger ärzteblatt"] = "Hamb Arztebl",
  ["hamburger beiträge zur archäologie"] = "HBA",
  ["hamburger beiträge zur numismatik"] = "HBN",
  ["hamdard islamicus : quarterly journal of the hamdard national foundation, pakistan"] = "Hamdard Islam",
  ["hamdard medicus"] = "Hamdard Med",
  ["hamline law review"] = "Hamline Law Rev",
  ["hampshire : the county magazine"] = "Hampshire",
  ["han xue yan jiu = chinese studies"] = "Han Xue Yan Jiu",
  ["han'gug haeyang haghoeji. korean society of oceanography"] = "Hangug Haeyang Haghoeji",
  ["han'gug hwan'gyeong saengtae haghoeji = korean journal of environment and ecology"] = "Hangug hwangyeong saengtae haghoeji",
  ["han'gug jaweon sig'mul haghoeji = korean journal of plant resources"] = "Hangug Jaweon Sigmul Haghoeji",
  ["han'gug sigpum wi'saeng anjeonseong haghoeji"] = "Hangug Sigpum Wisaeng Anjeonseong Haghoeji",
  ["han'gug yangbong haghoeji"] = "Hangug Yangbong Haghoeji",
  ["han'gug yugsu haghoeji"] = "Hangug Yugsu Haghoeji",
  ["han'guk ch'uksan hakhoe chi = korean journal of animal science"] = "Hanguk Chuksan Hakhoe Chi",
  ["han'guk haeyang paio hakhoe"] = "Hanguk Haeyang Paio Hakhoe",
  ["han'guk hwan'gyŏng nonghakhoe chi = korean journal of environmental agriculture"] = "Hanguk Hwangyong Nonghakhoe Chi",
  ["han'guk in'gu hakhoe chi = the journal of the population association of korea"] = "J Popul Assoc Korea",
  ["han'guk kach'uk wisaeng hakhoe chi"] = "Hanguk Kachuk Wisaeng Hakhoe Chi",
  ["han'guk kwahaksa hakhoeji"] = "Hanguk Kwahaksa Hakhoeji",
  ["han'guk misaengmul, saengmyŏng konghakhoe chi = korean journal of microbiology and biotechnology"] = "Hanguk Misaengmul Saengmyong Konghakhoe Chi",
  ["han'guk nonghwa hakhoe chi"] = "Hangug Nonghwahag Hoeji",
  ["han'guk nonyŏnhak = journal of korea gerontological society"] = "Hanguk Nonyonhak",
  ["han'guk sikp'um yŏngyang kwahakhoe chi"] = "Hanguk Sikpum Yongyang Kwahakhoe Chi",
  ["han'guk simni hakhoe chi. kŏn'gang = the korean journal of health psychology"] = "Hanguk Simni Hakhoe Chi Kongang",
  ["han'guk susan hakhoe chi = bulletin of the korean fisheries society"] = "Hangug Susan Haghoi Ji",
  ["han'guk t'oyang piryo hakhoe chi. han'guk t'oyang piryo hakhoe"] = "Hanguk Toyang Piryo Hakhoe Chi",
  ["han'guk tongmul pullyu hakhoe chi"] = "Hanguk Tongmul Pullyu Hakhoe Chi",
  ["han'guk ŭngyong konchʻung hakhoe chi = korean journal of applied entomology"] = "Hanguk Ungyong Konchung Hakhoe Chi",
  ["han'guk ŭngyong saengmyŏng hwahakhoe chi = journal of the korean society for applied biological chemistry"] = "Hanguk Ungyong Saengmyong Hwahakhoe Chi",
  ["han'guk wŏnye hakhoe chi. han'guk wŏnye hakhoe"] = "Hanguk Wonye Hakhoe Chi",
  ["han'guk yakyong changmul hakhoe chi = korean journal of medicinal crop science"] = "Hanguk Yakyong Changmul Hakhoe Chi",
  ["han'guk yukchong hakhoe chi"] = "Nyngjon Hag Hoi Ji",
  ["han-gug eoryu hag-hoeji = korean journal of ichthyology"] = "Han Gug Eoryu Hag Hoeji",
  ["han-gug yangsig hag-hoeji"] = "Hangug Yangsig Haghoeji",
  ["han-guk sikpum jeojang yutong hakoeji"] = "Hanguk Sikpum Jeojang Yutong Hakoeji",
  ["han-guk yang-seo pachungnyu hakoeji"] = "Hanguk Yangseo Pachungnyu Hakoeji",
  ["hand & microsurgery"] = "Hand Microsurg",
  ["hand (new york, n.y.)"] = "Hand (N Y)",
  ["hand clinics"] = "Hand Clin",
  ["hand surgery"] = "Hand Surg.",
  ["hand surgery & rehabilitation"] = "Hand Surg Rehabil",
  ["hand surgery : an international journal devoted to hand and upper limb surgery and related research : journal of the asia-pacific federation of societies for surgery of the hand"] = "Hand Surg",
  ["hand therapy"] = "Hand Ther",
  ["handbook : 3rd annual meeting. health technology assessment international. meeting (3rd : 2006 : adelaide, s. aust.)"] = "Handb Health Technol Assess Int",
  ["handbook of behavioral neuroscience"] = "Handb Behav Neurosci",
  ["handbook of clinical neurology"] = "Handb Clin Neurol",
  ["handbook of experimental pharmacology"] = "Handb Exp Pharmacol",
  ["handbook of graph grammars and computing by graph transformation"] = "Handb. Graph Gramm. Comput. Graph Transform.",
  ["handbook of international documentation and information"] = "Handbook Internat. Doc. Info.",
  ["handbook of logic in computer science"] = "Handb. Log. Comput. Sci.",
  ["handbook of numerical analysis"] = "Handb. Numer. Anal.",
  ["handbook of philosophical logic"] = "Handb. Philos. Log.",
  ["handbook of statistics"] = "Handbook of Statist.",
  ["handbook of the history of logic"] = "Handb. Hist. Log.",
  ["handbook of the philosophy of science"] = "Handb. Philos. Sci.",
  ["handbook of vegetation science"] = "Handb. Veg. Sci.",
  ["handbooks in economics"] = "Handbooks in Econom.",
  ["handbooks in operations research and management science"] = "Handbooks Oper. Res. Management Sci.",
  ["handbuch der altertumswissenschaften"] = "HAW",
  ["handbuch der archäologie"] = "HdArch",
  ["handbuch der physik"] = "Handbuch Phys.",
  ["handchirurgie, mikrochirurgie, plastische chirurgie"] = "Handchir. Mikrochir. Plast. Chir.",
  ["handchirurgie, mikrochirurgie, plastische chirurgie : organ der deutschsprachigen arbeitsgemeinschaft für handchirurgie : organ der deutschsprachigen arbeitsgemeinschaft für mikrochirurgie der peripheren nerven und gefässe : organ der vereinigung der deutschen plastischen chirurgen"] = "Handchir Mikrochir Plast Chir",
  ["handelingen der maatschappij voor geschiedenis en oudheidkunde te gent. societe d’histoire et d’archeologie de gand"] = "Handel Maatsch Geschied Oudheidk Gent",
  ["handelingen der maatschappij voor geschiedenis en oudheidkunde te gent. société d'histoire et d'archéologie de gand"] = "Handel Maatsch Geschied Oudheidk Gent",
  ["handeln und entscheiden in komplexen ökonomischen situationen"] = "Handeln Entscheid. Komplexen Ökon. Situat.",
  ["handicap; the news magazine for the nation's handicapped"] = "Handicap (Washington)",
  ["hang tian yi xue yu yi xue gong cheng = space medicine & medical engineering"] = "Space Med Med Eng (Beijing)",
  ["hang'uk jakmul hakhoe chi"] = "Hanguk Jakmul Hakhoe Chi",
  ["hangug eobyeong haghoeji"] = "Hangug Eobyeong Haghoeji",
  ["hangug n̊imhag hoi ji"] = "Hangug Nimhag Hoi Ji",
  ["hang̕ug hwang̕yeong saengmul haghoeji"] = "Hangug Hwangyeong Saengmul Haghoeji",
  ["hannoversche abhandlungen zur geschichte der medizin und der naturwissenschaften"] = "Hann Abh Gesch Med Natwiss",
  ["hansenologia internationalis"] = "Hansenol Int",
  ["hanʼguk kyunhakhoe chi"] = "Hanguk Kyun Hakoe Chi",
  ["hanʼguk saenghwahakhoe chi = the korean biochemical journal"] = "Hanguk Saenghwahakhoe Chi",
  ["han’guk ch’uksan hakhoe chi"] = "Korean journal of animal science|Hanguk Chuksan Hakhoe Chi",
  ["han’guk kyunhakhoe chi"] = "Hanguk Kyun Hakoe Chi",
  ["han’guk misaengmul, saengmyong konghakhoe chi"] = "Korean journal of microbiology and biotechnology|Hanguk Misaengmul Saengmyong Konghakhoe Chi",
  ["han’guk nonghwa hakhoe chi"] = "Hanguk Nonghwa Hakhoe Chi",
  ["han’guk saenghwa hakhoechi"] = "Han’guk Saenghwa Hakhoechi",
  ["han’guk sikp’um chori kwahakhoe chi"] = "Hanguk Sikpum Chori Kwahakhoe Chi",
  ["han’guk t’oyang piryo hakhoechi"] = "Han’guk T’oyang Piryo Hakhoechi",
  ["han’guk ungyong konch’ung hakhoe chi"] = "Korean journal of applied entomology|Hanguk Ungyong Konchung Hakhoe Chi",
  ["han’guk yakyong changmul hakhoe chi"] = "Korean journal of medicinal crop science|Hanguk Yakyong Changmul Hakhoe Chi",
  ["haptics : generating and perceiving tangible sensations : international conference, eurohaptics 2010, amsterdam, july 8-10, 2010 : proceedings. eurohaptics 2010 (2010 : amsterdam, netherlands)"] = "Haptics (2010)",
  ["haptics : neuroscience, devices, modeling, and applications : 9th international conference, eurohaptics 2014, versailles, france, june 24-26, 2014 : proceedings"] = "Haptics (2014)",
  ["haptics : science, technology, and applications : 11th international conference, eurohaptics 2018, pisa, italy, june 13-16, 2018, proceedings. eurohaptics conference (11th : 2018 : pisa, italy)"] = "Haptics (2018)",
  ["harbin institute of technology"] = "J. Harbin Inst. Tech.",
  ["harbor dental log"] = "Harb Dent Log",
  ["hard tissue"] = "Hard Tissue",
  ["hardy-ramanujan journal"] = "Hardy-Ramanujan J.",
  ["harefuah. foreign ed"] = "Harefuah (Tel Aviv)",
  ["harm reduction journal"] = "Harm Reduct J",
  ["harmful algae"] = "Harmful Algae",
  ["harmful algae 2002 : proceedings of the xth international conference on harmful algae, st. pete beach, florida, usa, october 21-25, 2002. international conference on harmful algae (10th : 2002 : st. pete beach, florida)"] = "Harmful Algae 2002 (2002)",
  ["harmful algae 2012 : proceedings of the 15th international conference on harmful algae : october 29 - november 2, 2012, ceco, changwon, gyeongnam, korea. international conference on harmful algae (15th : 2012 : changwon, gyeongnam, korea)"] = "Harmful Algae 2012 (2012)",
  ["harmful algae news"] = "Harmful Algae News",
  ["harmonia : coccinellidae du monde"] = "Harmonia (Etagnac)",
  ["harofe haivri. the hebrew medical journal"] = "Harofe Haivri Heb Med J",
  ["harofé haivri. the hebrew medical journal"] = "Harofe Haivri Heb Med J",
  ["harper hospital bulletin"] = "Harper Hosp Bull",
  ["hartford hospital bulletin"] = "Hartford Hosp Bull",
  ["harvard aids review"] = "Harv AIDS Rev",
  ["harvard blackletter law journal"] = "Harv Blacklett Law J",
  ["harvard business review"] = "Harv Bus Rev",
  ["harvard civil rights-civil liberties law review"] = "Harv Civ Rights-Civil Lib Law Rev",
  ["harvard data science review"] = "Harv Data Sci Rev",
  ["harvard dental alumni bulletin"] = "Harv Dent Alumni Bull",
  ["harvard dental bulletin"] = "Harv Dent Bull",
  ["harvard educational review"] = "Harv Educ Rev",
  ["harvard health letter"] = "Harv Health Lett",
  ["harvard health letter / from harvard medical school"] = "Harv Health Lett",
  ["harvard health policy review : a student publication of the harvard interfaculty initiative in health policy"] = "Harvard Health Policy Rev",
  ["harvard heart letter"] = "Harv. Heart Lett.",
  ["harvard heart letter : from harvard medical school"] = "Harv Heart Lett",
  ["harvard human rights journal"] = "Harv Hum Rights J",
  ["harvard international law journal"] = "Harvard Int Law J",
  ["harvard international review"] = "Harvard Int Rev",
  ["harvard journal of asiatic studies"] = "Harv J Asiat Stud",
  ["harvard journal of law & public policy"] = "Harv J Law Public Policy",
  ["harvard journal of law & technology"] = "Harv J Law Technol",
  ["harvard journal on legislation"] = "Harvard J Legis",
  ["harvard law review"] = "Harv Law Rev",
  ["harvard library bulletin"] = "Harv Libr Bull",
  ["harvard magazine"] = "Harv Mag",
  ["harvard medical alumni bulletin"] = "Harv Med Alumni Bull",
  ["harvard medical student review"] = "Harv Med Stud Rev",
  ["harvard men's health watch"] = "Harv Mens Health Watch",
  ["harvard mental health letter"] = "Harv. Ment. Health Lett.",
  ["harvard men’s health watch"] = "Harv. Mens Health Watch",
  ["harvard papers in botany"] = "Harv Pap Bot",
  ["harvard public health alumni bulletin"] = "Harv Public Health Alumni Bull",
  ["harvard public health review (cambridge, mass.)"] = "Harv Public Health Rev (Camb)",
  ["harvard review of psychiatry"] = "Harv Rev Psychiatry",
  ["harvard semitic series"] = "HSS",
  ["harvard series in ukrainian studies"] = "Harv. S. Ukrain. St.",
  ["harvard studies in classical philology"] = "Harv Stud Classic Philol",
  ["harvard theological review"] = "HThR",
  ["harvard ukrainian studies"] = "Harv Ukr Stud",
  ["harvard university art museums bulletin"] = "BHarvMus",
  ["harvard university center of mathematical sciences and applications (cmsa) series in mathematics"] = "Harv. Univ. Cent. Math. Sci. Appl. Ser. Math.",
  ["harvard women's health watch"] = "Harv Womens Health Watch",
  ["harvard women's law journal"] = "Harv Women's Law J",
  ["harvard women’s health watch"] = "Harv. Womens Health Watch",
  ["harvard women’s law journal"] = "Harv Women’s Law J",
  ["harvey lectures"] = "Harvey Lect",
  ["hasl [reports]. u.s. atomic energy commission"] = "HASL Rep",
  ["hasl reports"] = "HASL Rep.",
  ["hastings center report"] = "Hastings Cent. Rep.",
  ["hastings center studies"] = "Hastings Cent. Stud.",
  ["hastings constitutional law quarterly"] = "Hastings Constit Law Q",
  ["hastings international and comparative law review"] = "Hastings Int Comp Law Rev",
  ["hastings women's law journal"] = "Hastings Womens Law J",
  ["hastings women’s law journal"] = "Hastings Womens Law J",
  ["hau : journal of ethnographic theory"] = "HAU",
  ["havemeyer foundation monograph series. dorothy russell havemeyer foundation"] = "Havemeyer Found Monogr Ser",
  ["hawai'i journal of health & social welfare"] = "Hawaii J Health Soc Welf",
  ["hawai'i journal of medicine & public health : a journal of asia pacific medicine & public health"] = "Hawaii J Med Public Health",
  ["hawaii bar journal (honolulu, hawaii : 1963)"] = "Haw Bar J",
  ["hawaii dental association journal"] = "Hawaii Dent Assoc J",
  ["hawaii dental journal"] = "Hawaii Dent J",
  ["hawaii journal of public health"] = "Hawaii J Public Health",
  ["hawaii medical journal"] = "Hawaii Med J",
  ["hawaii nurse"] = "Hawaii Nurse",
  ["hawaii nurse (honolulu, hawaii : 1994)"] = "Hawaii Nurse (Honol)",
  ["hawaii nurses pipeline"] = "Hawaii Nurses Pipeline",
  ["hawaii revised statutes : comprising the statutes of the state of hawaii, consolidated, revised, and annotated / published by authority. hawaii"] = "Hawaii Revis Statut Hawaii",
  ["hawaii revised statutes : comprising the statutes of the state of hawaii, consolidated, revised, and annotated. hawaii"] = "Hawaii Revis Statut Hawaii",
  ["hayati (bogor, indonesia)"] = "Hayati",
  ["hayati : journal of biosciences"] = "Hayati",
  ["hayati journal of biosciences"] = "HAYATI J. Biosci.",
  ["hayes historical journal"] = "Hayes Hist J",
  ["hazardous waste and hazardous materials"] = "Hazard. Waste Hazard. Mater.",
  ["hazards control quarterly report"] = "Hazard Control Qual Rep",
  ["hämatologie und bluttransfusion"] = "Hamatol Bluttransfus",
  ["hǎiyáng yǔ húzhǎo = oceanologia et limnologia sinica"] = "Haiyang Yu Huzhao",
  ["hba lecture notes in mathematics"] = "HBA Lect. Notes Math.",
  ["hbi report"] = "HBI Rep",
  ["hci international 2018 - posters' extended abstracts : 20th international conference, hci international 2018, las vegas, nv, usa, july 15-20, 2018, proceedings, part i. international conference on human-computer interaction (20th : 2018 : las vegas, nev.)"] = "HCI Int 2018 Posters Ext Abstr (2018)",
  ["hci international 2019 - posters : 21st international conference"] = "HCI Int 2019 Posters (2019)",
  ["hci international 2020 - late breaking posters : 22nd international conference, hcii 2020, copenhagen, denmark, july 19-24, 2020, proceedings, part i"] = "HCI Int 2020 Late Break Posters (2020)",
  ["hda now"] = "HDA Now",
  ["he nong xue bao = acta agriculturae nucleatae sinica"] = "He Nong Xue Bao",
  ["head & face medicine"] = "Head Face Med",
  ["head & neck"] = "Head Neck",
  ["head & neck oncology"] = "Head Neck Oncol",
  ["head & neck surgery"] = "Head Neck Surg",
  ["head and neck"] = "Head Neck",
  ["head and neck pathology"] = "Head Neck Pathol",
  ["head and neck surgery"] = "Head Neck Surg.",
  ["head and neck tumor segmentation : first challenge, hecktor 2020, held in conjunction with miccai 2020, lima, peru, october 4, 2020, proceedings"] = "Head Neck Tumor Segm (2020)",
  ["headache: the journal of head and face pain"] = "Headache: J. Head Face Pain",
  ["healing light"] = "Heal Light",
  ["health & justice"] = "Health Justice",
  ["health & medical care services review"] = "Health Med Care Serv Rev",
  ["health & medicine : journal of the health and medicine policy research group"] = "Health Med",
  ["health & place"] = "Health Place",
  ["health & social care in the community"] = "Health Soc Care Community",
  ["health & social work"] = "Health Soc Work",
  ["health (family media, inc.)"] = "Health (N Y)",
  ["health (london, england : 1997)"] = "Health (London)",
  ["health affairs"] = "Health Aff",
  ["health affairs (project hope)"] = "Health Aff (Millwood)",
  ["health after 50 with scientific american consumer health"] = "Health After 50 Sci Am Consum Health",
  ["health and canadian society (winnipeg, man.)"] = "Health Can Soc",
  ["health and history"] = "Health History",
  ["health and human rights"] = "Health Hum Rights",
  ["health and medical care services review"] = "Health Med. Care Serv. Rev.",
  ["health and medicine"] = "Health Med.",
  ["health and population: perspectives and issues"] = "Health Popul. Perspect. Issues",
  ["health and population; perspectives and issues"] = "Health Popul Perspect Issues",
  ["health and quality of life outcomes"] = "Health Qual Life Outcomes",
  ["health and safety code annotated of the state of california, adopted april 7, 1939 / annotated and indexed by the publisher’s editorial staff. california"] = "Health Saf Code Annot State Calif Adopt April 7 1939 Calif",
  ["health and safety code annotated of the state of california, adopted april 7, 1939. california"] = "Health Saf Code Annot State Calif Adopt April 7 1939 Calif",
  ["health and sexuality"] = "Health Sex",
  ["health and social service journal"] = "Health Soc Serv J",
  ["health and social welfare"] = "Health Soc Welf",
  ["health and social work"] = "Health Soc. Work",
  ["health and technology"] = "Health Technol (Berl)",
  ["health beat (manila, philippines)"] = "Health Beat",
  ["health behavior and policy review"] = "Health Behav Policy Rev",
  ["health behavior research"] = "Health Behav Res",
  ["health bulletin"] = "Health Bull (Melb)",
  ["health bulletin for teachers"] = "Health Bull Teach",
  ["health bulletin. kern co., calif. department of public health"] = "Health Bull",
  ["health care"] = "Health Care (Don Mills)",
  ["health care analysis"] = "Health Care Anal.",
  ["health care analysis : hca : journal of health philosophy and policy"] = "Health Care Anal",
  ["health care cost reengineering report"] = "Health Care Cost Reengineering Rep",
  ["health care dimensions"] = "Health Care Dimen",
  ["health care education"] = "Health Care Educ",
  ["health care ethics usa : a publication of the center for health care ethics"] = "Health Care Ethics USA",
  ["health care financing review"] = "Health Care Financ Rev",
  ["health care financing review. annual supplement"] = "Health Care Financ Rev Annu Suppl",
  ["health care financing review. statistical supplement"] = "Health Care Financ Rev Stat Suppl",
  ["health care financing trends"] = "Health Care Financ Trends",
  ["health care food & nutrition focus"] = "Health Care Food Nutr Focus",
  ["health care food and nutrition focus"] = "Health Care Food Nutr. Focus",
  ["health care for women international"] = "Health Care Women Int",
  ["health care in canada"] = "Health Care Can",
  ["health care innovations"] = "Health Care Innov.",
  ["health care innovations : the journal of the american association of preferred provider organizations"] = "Health Care Innov",
  ["health care law monthly"] = "Health Care Law Mon",
  ["health care law newsletter"] = "Health Care Law Newsl",
  ["health care law newsletter / weissburg and aronson, inc"] = "Health Care Law Newsl",
  ["health care management"] = "Health Care Manag.",
  ["health care management (philadelphia, pa.)"] = "Health Care Manag",
  ["health care management review"] = "Health Care Manage Rev",
  ["health care management science"] = "Health Care Manag Sci",
  ["health care marketer"] = "Health Care Mark",
  ["health care marketer & target market"] = "Health Care Mark Target Market",
  ["health care marketer and target market"] = "Health Care Mark. Target Market",
  ["health care newsletter"] = "Health Care Newsl",
  ["health care newsletter / national safety council"] = "Health Care Newsl",
  ["health care planning & marketing : hcp&m"] = "Health Care Plann Mark",
  ["health care planning and marketing"] = "Health Care Plann. Mark.",
  ["health care reform week"] = "Health Care Reform Week",
  ["health care security and safety management"] = "Health Care Secur Saf Manage",
  ["health care strategic management"] = "Health Care Strateg Manage",
  ["health care supervisor"] = "Health Care Superv.",
  ["health care systems"] = "Health Care Syst",
  ["health care week"] = "Health Care Week",
  ["health care. current reviews"] = "Health Care Curr Rev",
  ["health communication"] = "Health Commun",
  ["health communications and informatics"] = "Health Commun Informatics",
  ["health cost management"] = "Health Cost Manage",
  ["health data management"] = "Health Data Manag",
  ["health data science"] = "Health Data Sci.",
  ["health devices"] = "Health Devices",
  ["health economics"] = "Health Econ",
  ["health economics & outcome research : open access"] = "Health Econ Outcome Res",
  ["health economics review"] = "Health Econ Rev",
  ["health economics, policy, and law"] = "Health Econ Policy Law",
  ["health education"] = "Health Educ",
  ["health education & behavior : the official publication of the society for public health education"] = "Health Educ Behav",
  ["health education (bradford, west yorkshire, england)"] = "Health Educ (Lond)",
  ["health education and behavior"] = "Health Educ. Behav.",
  ["health education and behavior. supplement"] = "Health Educ Behav Suppl",
  ["health education and care"] = "Health Educ Care",
  ["health education and public health"] = "Health Educ Public Health",
  ["health education bulletin"] = "Health Educ Bull",
  ["health education journal"] = "Health Educ J",
  ["health education monographs"] = "Health Educ Monogr",
  ["health education quarterly"] = "Health Educ Q",
  ["health education quarterly. supplement"] = "Health Educ Q Suppl",
  ["health education reports"] = "Health Educ Rep",
  ["health education research"] = "Health Educ Res",
  ["health educator : journal of eta sigma gamma"] = "Health Educ (Muncie)",
  ["health educators at work"] = "Health Educ Work",
  ["health equity"] = "Health Equity",
  ["health estate"] = "Health Estate",
  ["health estate journal"] = "Health Estate J.",
  ["health estate journal : journal of the institute of hospital engineering"] = "Health Estate J",
  ["health expectations"] = "Health Expectations",
  ["health expectations : an international journal of public participation in health care and health policy"] = "Health Expect",
  ["health facilities management"] = "Health Facil Manage",
  ["health for the millions"] = "Health Millions",
  ["health forum journal"] = "Health Forum J",
  ["health horizon"] = "Health Horiz",
  ["health industry today"] = "Health Ind Today",
  ["health informatics journal"] = "Health Informatics J",
  ["health information and libraries journal"] = "Health Info Libr J",
  ["health information management"] = "Health Inf. Manag.",
  ["health information management : journal of the health information management association of australia"] = "Health Inf Manag",
  ["health information management journal"] = "Health Inf. Manage. J.",
  ["health information science"] = "Health Inf. Sci.",
  ["health information science and systems"] = "Health Inf Sci Syst",
  ["health information systems and telemedicine"] = "Health Inf Syst Telemed",
  ["health insurance statistics. cms; current medicare survey report"] = "Health Insur Stat Cms",
  ["health laboratory science"] = "Health Lab Sci",
  ["health law in canada"] = "Health Law Can",
  ["health law journal"] = "Health Law J",
  ["health law news. university of houston. health law and policy institute"] = "Health Law News",
  ["health law project library bulletin"] = "Health Law Proj Libr Bull",
  ["health law review"] = "Health Law Rev",
  ["health law vigil"] = "Health Law Vigil",
  ["health libraries review"] = "Health Libr Rev",
  ["health literacy research and practice"] = "Health Lit Res Pract",
  ["health management"] = "Health Manage",
  ["health management forum"] = "Health Manage Forum",
  ["health management quarterly"] = "Health Manage. Q.",
  ["health management quarterly : hmq"] = "Health Manage Q",
  ["health management technology"] = "Health Manag Technol",
  ["health manpower literature"] = "Health Manpow Lit",
  ["health manpower management"] = "Health Manpow Manage",
  ["health manpower report"] = "Health Manpow Rep",
  ["health marketing quarterly"] = "Health Mark Q",
  ["health matrix"] = "Health Matrix",
  ["health matrix (cleveland, ohio : 1991)"] = "Health Matrix Clevel",
  ["health news"] = "Health News (Rochester)",
  ["health news (waltham, mass.)"] = "Health News",
  ["health news and views (gaborone, botswana)"] = "Health News Views",
  ["health of the army. united states. surgeon-general's office"] = "Health Army",
  ["health of the people"] = "Health People",
  ["health outcomes research in medicine"] = "Health Outcomes Res Med",
  ["health pac bulletin"] = "Health PAC Bull",
  ["health perspectives"] = "Health Perspect",
  ["health physics"] = "Health Phys",
  ["health planning & manpower report"] = "Health Plann Manpow Rep",
  ["health planning & manpower reports"] = "Health Plann Manpow Rep (1976)",
  ["health planning and manpower report"] = "Health Plann. Manpow. Rep.",
  ["health planning and manpower reports"] = "Health Plann. Manpow. Rep.",
  ["health policy"] = "Health Policy (New York)",
  ["health policy (amsterdam, netherlands)"] = "Health Policy",
  ["health policy and education"] = "Health Policy Educ",
  ["health policy and planning"] = "Health Policy Plan",
  ["health policy and technology"] = "Health Policy Technol",
  ["health policy open"] = "Health Policy Open",
  ["health policy quarterly"] = "Health Policy Q",
  ["health policy week"] = "Health Policy Week",
  ["health policy week : hpw"] = "Health Policy Week",
  ["health practitioner. physician assistant"] = "Health Pract Physician Assist",
  ["health progress"] = "Health Prog.",
  ["health progress (saint louis, mo.)"] = "Health Prog",
  ["health promotion"] = "Health Promot.",
  ["health promotion (oxford, england)"] = "Health Promot",
  ["health promotion and chronic disease prevention in canada"] = "Health Promot. Chronic Dis. Prev. Can.",
  ["health promotion and chronic disease prevention in canada : research, policy and practice"] = "Health Promot Chronic Dis Prev Can",
  ["health promotion international"] = "Health Promot Int",
  ["health promotion journal of australia : official journal of australian association of health promotion professionals"] = "Health Promot J Austr",
  ["health promotion perspectives"] = "Health Promot Perspect",
  ["health promotion practice"] = "Health Promot Pract",
  ["health psychology"] = "Health Psychol.",
  ["health psychology : official journal of the division of health psychology, american psychological association"] = "Health Psychol",
  ["health psychology and behavioral medicine"] = "Health Psychol Behav Med",
  ["health psychology bulletin"] = "Health Psychol Bull",
  ["health psychology open"] = "Health Psychol Open",
  ["health psychology report"] = "Health Psychol Rep",
  ["health psychology research"] = "Health Psychol Res",
  ["health psychology review"] = "Health Psychol Rev",
  ["health rays"] = "Health Rays",
  ["health reform and priority services"] = "Health Reform Prior Serv",
  ["health reform and priority services / phr"] = "Health Reform Prior Serv",
  ["health reports"] = "Health Rep",
  ["health research policy and systems"] = "Health Res Policy Syst",
  ["health research policy and systems / biomed central"] = "Health Res Policy Syst",
  ["health sa = sa gesondheid"] = "Health SA",
  ["health science reports"] = "Health Sci Rep",
  ["health security"] = "Health Secur",
  ["health service journal"] = "Health Serv. J.",
  ["health services & outcomes research methodology"] = "Health Serv Outcomes Res Methodol",
  ["health services and outcomes research methodology"] = "Health Serv. Outcomes Res. Methodol.",
  ["health services insights"] = "Health Serv Insights",
  ["health services journal"] = "Health Serv J",
  ["health services management"] = "Health Serv Manage",
  ["health services management research"] = "Health Serv Manage Res",
  ["health services management research : an official journal of the association of university programs in health administration / hsmc, aupha"] = "Health Serv Manage Res",
  ["health services manager"] = "Health Serv Manager",
  ["health services manpower review"] = "Health Serv Manpow Rev",
  ["health services reports"] = "Health Serv Rep",
  ["health services research"] = "Health Serv Res",
  ["health services research and managerial epidemiology"] = "Health Serv Res Manag Epidemiol",
  ["health sociology review : the journal of the health section of the australian sociological association"] = "Health Sociol Rev",
  ["health statistics in the nordic countries = helsestatistikk i de nordiske land"] = "Health Stat Nord Ctries",
  ["health statistics quarterly"] = "Health Stat Q",
  ["health statistics quarterly / office for national statistics"] = "Health Stat Q",
  ["health system leader"] = "Health Syst Lead",
  ["health systems (basingstoke, england)"] = "Health Syst (Basingstoke)",
  ["health systems and policy research"] = "Health Syst Policy Res",
  ["health systems and reform"] = "Health Syst Reform",
  ["health systems in transition"] = "Health Syst Transit",
  ["health systems management"] = "Health Syst Manage",
  ["health systems review"] = "Health Syst Rev",
  ["health technology"] = "Health Technol",
  ["health technology assessment"] = "Health Technol Assess (Rockv)",
  ["health technology assessment (winchester, england)"] = "Health Technol Assess",
  ["health technology assessment reports"] = "Health Technol Assess Rep",
  ["health technology directions"] = "Health Technol Dir",
  ["health transition review"] = "Health Transit. Rev.",
  ["health transition review : the cultural, social, and behavioural determinants of health"] = "Health Transit Rev",
  ["health trends"] = "Health Trends",
  ["health values"] = "Health Values",
  ["health visitor"] = "Health Visit",
  ["health, education, and welfare indicators"] = "Health Educ Welf Indic",
  ["health, risk & society"] = "Health Risk Soc",
  ["healthcare (amsterdam, netherlands)"] = "Healthc (Amst)",
  ["healthcare (basel, switzerland)"] = "Healthcare (Basel)",
  ["healthcare alabama"] = "Healthc Ala",
  ["healthcare alabama / alabama hospital association"] = "Healthc Ala",
  ["healthcare analytics"] = "Healthcare Anal.",
  ["healthcare benchmarks"] = "Healthc Benchmarks",
  ["healthcare benchmarks and quality improvement"] = "Healthcare Benchmarks Qual Improv",
  ["healthcare bottom line"] = "Healthc Bottom Line",
  ["healthcare computing & communications"] = "Healthc Comput Commun",
  ["healthcare computing and communications"] = "Healthc. Comput. Commun.",
  ["healthcare demand & disease management"] = "Healthc Demand Dis Manag",
  ["healthcare demand and disease management"] = "Healthc. Demand Dis. Manag.",
  ["healthcare executive"] = "Healthc Exec",
  ["healthcare executive currents"] = "Healthc Exec Curr",
  ["healthcare facilities management series"] = "Healthc Facil Manag Ser",
  ["healthcare financial management"] = "Healthc. Financ. Manage.",
  ["healthcare financial management : journal of the healthcare financial management association"] = "Healthc Financ Manage",
  ["healthcare foodservice"] = "Healthc Foodserv",
  ["healthcare foodservice magazine : the international trade publication for the healthcare foodservice industry"] = "Healthc Foodserv Mag",
  ["healthcare forum"] = "Healthc Forum",
  ["healthcare forum journal"] = "Healthc. Forum J.",
  ["healthcare hazard management monitor : hhmm : the newsletter of the center for healthcare environmental management"] = "Healthc Hazard Manage Monit",
  ["healthcare hazardous materials management"] = "Healthc. Hazard. Mater. Manage.",
  ["healthcare hazardous materials management : hhmm"] = "Healthc Hazard Mater Manage",
  ["healthcare human resources"] = "Healthc Hum Resour",
  ["healthcare human resources. special report"] = "Healthc Hum Resour Spec Rep",
  ["healthcare in low-resource settings"] = "Healthc Low Resour Settings",
  ["healthcare infection"] = "Healthc Infect",
  ["healthcare informatics"] = "Healthc. Inform.",
  ["healthcare informatics : the business magazine for information and communication systems"] = "Healthc Inform",
  ["healthcare informatics research"] = "Healthc Inform Res",
  ["healthcare information management"] = "Healthc. Inf. Manage.",
  ["healthcare information management : journal of the healthcare information and management systems society of the american hospital association"] = "Healthc Inf Manage",
  ["healthcare leadership & management report"] = "Healthc Leadersh Manag Rep",
  ["healthcare leadership report"] = "Healthc Leadersh Rep",
  ["healthcare management forum"] = "Healthc Manage Forum",
  ["healthcare philanthropy"] = "Healthc Philanthr",
  ["healthcare policy = politiques de santé"] = "Healthc Policy",
  ["healthcare protection management"] = "Healthc Prot Manage",
  ["healthcare quarterly (toronto, ont.)"] = "Healthc Q",
  ["healthcare strategist"] = "Healthc. Strateg.",
  ["healthcare systems strategy report"] = "Healthc Syst Strategy Rep",
  ["healthcare technology letters"] = "Healthc Technol Lett",
  ["healthcare transformation"] = "Healthc Transform",
  ["healthcare transformation: artificial intelligence, automation, and robotics"] = "Healthcare Transform.: Artif. Intell. Autom. Rob.",
  ["healthcare trends & transition"] = "Healthc Trends Transit",
  ["healthcare trends and transition"] = "Healthc. Trends Transit.",
  ["healthcom. international conference on e-health networking, applications and services"] = "Healthcom",
  ["healthlink (new york, n.y.)"] = "Healthlink",
  ["healthtexas / texas hospital association"] = "Healthtexas",
  ["healthy aging & clinical care in the elderly"] = "Healthy Aging Clin Care Elder",
  ["healthy aging research"] = "Healthy Aging Res",
  ["healthy people 2000 statistical notes"] = "Healthy People 2000 Stat Notes",
  ["healthy people 2000 statistical notes / national center for health statistics"] = "Healthy People 2000 Stat Notes",
  ["healthy people 2000 statistics and surveillance"] = "Healthy People 2000 Stat Surveill",
  ["healthy people 2000 statistics and surveillance / national center for health statistics"] = "Healthy People 2000 Stat Surveill",
  ["healthy people 2010 statistical notes : from the centers for disease control and prevention/national center for health statistics"] = "Healthy People 2010 Stat Notes",
  ["healthy people statistical notes"] = "Healthy People Stat Notes",
  ["hearing news"] = "Hear News (Washington)",
  ["hearing research"] = "Hear Res",
  ["hearing, balance and communication"] = "Hearing Balance Commun",
  ["heart & lung : the journal of critical care"] = "Heart Lung",
  ["heart (british cardiac society)"] = "Heart",
  ["heart advisor"] = "Heart Advis",
  ["heart advisor / the cleveland clinic"] = "Heart Advis",
  ["heart and lung"] = "Heart Lung",
  ["heart and metabolism : management of the coronary patient"] = "Heart Metab",
  ["heart and mind (mumbai, india)"] = "Heart Mind (Mumbai)",
  ["heart and vessels"] = "Heart Vessels",
  ["heart and vessels. supplement"] = "Heart Vessels Suppl",
  ["heart asia"] = "Heart Asia",
  ["heart disease"] = "Heart Dis.",
  ["heart disease (hagerstown, md.)"] = "Heart Dis",
  ["heart disease and stroke"] = "Heart Dis. Stroke",
  ["heart disease and stroke : a journal for primary care physicians"] = "Heart Dis Stroke",
  ["heart failure"] = "Heart Fail",
  ["heart failure clinics"] = "Heart Fail Clin",
  ["heart failure monitor"] = "Heart Fail Monit",
  ["heart failure reviews"] = "Heart Fail Rev",
  ["heart international"] = "Heart Int",
  ["heart rhythm"] = "Heart Rhythm",
  ["heart rhythm : the official journal of the heart rhythm society"] = "Heart Rhythm",
  ["heart rhythm o2"] = "Heart Rhythm O2",
  ["heart surgery forum"] = "Heart Surg. Forum",
  ["heart transplantation"] = "Heart Transplant",
  ["heart transplantation / international society for heart transplantation"] = "Heart Transplant",
  ["heart views : the official journal of the gulf heart association"] = "Heart Views",
  ["heart, lung & circulation"] = "Heart Lung Circ",
  ["heart, lung and vessels"] = "Heart Lung Vessel",
  ["heartbeat (san francisco, calif.)"] = "Heartbeat",
  ["heartdrug : excellence in cardiovascular trials"] = "Heartdrug",
  ["heartrhythm case reports"] = "HeartRhythm Case Rep",
  ["heat and mass transfer"] = "Heat Mass Transfer",
  ["heat and mass transfer = wärme- und stoffübertragung"] = "Heat Mass Transf",
  ["heat exchanger design updates"] = "Heat Exch. Des. Updates",
  ["heat pipe science and technology, an international journal"] = "Heat Pipe Sci. Technol. Int. J.",
  ["heat recovery systems and chp"] = "Heat Recovery Syst. CHP",
  ["heat transfer"] = "Heat Transfer",
  ["heat transfer - asian research"] = "Heat Transfer Asian Res.",
  ["heat transfer engineering"] = "Heat Transfer Eng.",
  ["heat transfer research"] = "Heat Transf Res",
  ["heat transfer: soviet reviews"] = "Heat Transfer: Soviet Rev.",
  ["heat treatment and surface engineering"] = "Heat Treat. Surf. Eng.",
  ["heat treatment of metals"] = "Heat Treat. Met.",
  ["heating, piping and air conditioning"] = "Heat. Piping Air Cond.",
  ["heating, piping, and air conditioning"] = "Heat Piping Air Cond",
  ["hebammen wissen"] = "Hebammen Wissen",
  ["hebei nong ye da xue xue bao = hebei nongye daxue xuebao"] = "Hebei Nong Ye Da Xue Xue Bao",
  ["hebrew union college annual"] = "HebrUCA",
  ["hebrew union college annual. hebrew union college"] = "Heb Union Coll Annu",
  ["hec forum"] = "HEC Forum",
  ["hec forum : an interdisciplinary journal on hospitals' ethical and legal issues"] = "HEC Forum",
  ["hec forum : an interdisciplinary journal on hospitals’ ethical and legal issues"] = "HEC Forum",
  ["hecheng xiangjiao gongye"] = "Hecheng Xiangjiao Gongye",
  ["hefte des archäologischen seminars der universität bern"] = "HASB",
  ["hefte zur unfallheilkunde"] = "Hefte Unfallheilkd",
  ["heidelberger jahrbucher"] = "Heidelb Jahrb",
  ["heidelberger jahrbücher"] = "Heidelb Jahrb",
  ["heiko shinkei kagaku"] = "Equilib Res",
  ["heikō shinkei kagaku"] = "Equilib Res",
  ["heilkunde - heilwege"] = "Heilkd Heilwege",
  ["heilongjiang bayi nongken daxue xuebao"] = "Heilongjiang Bayi Nongken Daxue Xuebao",
  ["heilongjiang daxue ziran kexue xuebao"] = "Heilongjiang Daxue Ziran Kexue Xuebao",
  ["heilongjiang xumu shouyi"] = "Heilongjiang Xumu Shouyi",
  ["heilpadagogische werkblatter"] = "Heilpadagog Werkbl",
  ["heilpädagogische werkblätter"] = "Heilpadagog Werkbl",
  ["hektoen international : a journal of medical humanities"] = "Hektoen Int",
  ["helgoland marine research"] = "Helgol Mar Res",
  ["helike. universidad nacional de educación a distancia, centro regional de elche"] = "Helike",
  ["helikon. rivista di tradizione e cultura classica"] = "Helikon",
  ["helinium. revue consacrée à l’archéologie des pays-bas, de la belgique et du grand-duché de luxembourg"] = "Helinium",
  ["helios. a journal devoted to critical and methodological studies of classical culture, literature and society"] = "Helios",
  ["hellenic journal of cardiology"] = "Hellenic J. Cardiol.",
  ["hellenic journal of cardiology : hjc = hellēnikē kardiologikē epitheōrēsē"] = "Hellenic J Cardiol",
  ["hellenic journal of nuclear medicine"] = "Hell J Nucl Med",
  ["hellenic mathematical society"] = "Comput. Math. Appl. (Athens)",
  ["hellenika stomatologika chronika"] = "Hell. Stomatol. Chron.",
  ["hellenika stomatologika chronika. hellenic stomatological annals"] = "Hell Stomatol Chron",
  ["hellenika. jahrbuch für die freunde griechenlands"] = "HellenikaJb",
  ["hellenike cheirourgike. acta chirurgica hellenica"] = "Hell Cheirourgike",
  ["hellenike iatrike"] = "Hell Iatr",
  ["helleniko periodiko gia stomatike kai gnathoprosopike cheirourgike"] = "Hell. Period. Stomat. Gnathopathoprosopike. Cheir.",
  ["hellenis adelphe"] = "Hell Adelphe",
  ["hellēnikē cheirourgikē. acta chirurgica hellenica"] = "Hell Cheirourgike",
  ["helsinki university of technology"] = "Helsinki Univ. Tech. Digit. Systems Lab. Ser. A Res. Rep.",
  ["helvetia archaeologica"] = "HA",
  ["helvetica chimica acta"] = "Helv. Chim. Acta",
  ["helvetica chirurgica acta"] = "Helv Chir Acta",
  ["helvetica chirurgica acta. supplementum"] = "Helv Chir Acta Suppl",
  ["helvetica medica acta"] = "Helv Med Acta",
  ["helvetica medica acta. supplement"] = "Helv. Med. Acta. Suppl.",
  ["helvetica medica acta. supplementum"] = "Helv Med Acta Suppl",
  ["helvetica odontologica acta"] = "Helv Odontol Acta",
  ["helvetica paediatrica acta"] = "Helv Paediatr Acta",
  ["helvetica paediatrica acta. supplementum"] = "Helv Paediatr Acta Suppl",
  ["helvetica physica acta"] = "Helv. Phys. Acta",
  ["helvetica physiologica et pharmacologica acta"] = "Helv Physiol Pharmacol Acta",
  ["helvetica physiologica et pharmacologica acta. supplementum"] = "Helv Physiol Pharmacol Acta Suppl 1",
  ["hematologic pathology"] = "Hematol Pathol",
  ["hematological oncology"] = "Hematol Oncol",
  ["hematology & medical oncology"] = "Hematol Med Oncol",
  ["hematology (amsterdam, netherlands)"] = "Hematology",
  ["hematology / the education program of the american society of hematology. american society of hematology. education program"] = "Hematology Am Soc Hematol Educ Program",
  ["hematology and cell therapy"] = "Hematol Cell Ther",
  ["hematology education. european hematology association. congress. education program"] = "Hematol Educ",
  ["hematology journal"] = "Hematol. J.",
  ["hematology reports"] = "Hematol Rep",
  ["hematology reviews"] = "Hematol Rev",
  ["hematology, transfusion and cell therapy"] = "Hematol Transfus Cell Ther",
  ["hematology. american society of hematology. education program"] = "Hematology Am Soc Hematol Educ Program",
  ["hematology/oncology and stem cell therapy"] = "Hematol Oncol Stem Cell Ther",
  ["hematology/oncology clinics of north america"] = "Hematol Oncol Clin North Am",
  ["hematopathology and molecular hematology"] = "Hematopathol Mol Hematol",
  ["hemijska industrija"] = "Hem Ind",
  ["hemispheres (paris, france)"] = "Hemispheres",
  ["hemodialysis international"] = "Hemodial. Int.",
  ["hemodialysis international. international symposium on home hemodialysis"] = "Hemodial Int",
  ["hemonc today"] = "HemOnc Today",
  ["henan nongye daxue xuebao"] = "Henan Nongye Daxue Xuebao",
  ["henan nongye kexue"] = "Henan Nongye Kexue",
  ["henan yi ke da xue xue bao = journal of henan medical university = henan yikedaxue xuebao"] = "Honan I Ko Ta Hsueh Hsueh Pao",
  ["hengyang shi fan xue yuan xue bao"] = "Journal of Heng Yang Teachers’ College|Hengyang Shi Fan Xue Yuan Xue Bao",
  ["hengyang shi fan xue yuan xue bao = journal of heng yang teachers' college"] = "Hengyang Shi Fan Xue Yuan Xue Bao",
  ["henry e. sigerist supplements to the bulletin of the history of medicine"] = "Henry E. Sigerist Suppl. Bull. Hist. Med.",
  ["henry ford hospital medical bulletin"] = "Henry Ford Hosp Med Bull",
  ["henry ford hospital medical journal"] = "Henry Ford Hosp Med J",
  ["henry j. kaiser family foundation: reproductive health report"] = "Henry J Kaiser Fam Found Reprod Health Rep",
  ["hepatic medicine : evidence and research"] = "Hepat Med",
  ["hepatic oncology"] = "Hepat Oncol",
  ["hepatitis monthly"] = "Hepat Mon",
  ["hepatitis research and treatment"] = "Hepat Res Treat",
  ["hepatitis weekly"] = "Hepatitis Wkly",
  ["hepatobiliary & pancreatic diseases international : hbpd int"] = "Hepatobiliary Pancreat Dis Int",
  ["hepatobiliary and pancreatic diseases international"] = "Hepatobiliary Pancreat. Dis. Int.",
  ["hepatobiliary surgery and nutrition"] = "Hepatobiliary Surg Nutr",
  ["hepatology (baltimore, md.)"] = "Hepatology",
  ["hepatology communications"] = "Hepatol Commun",
  ["hepatology international"] = "Hepatol Int",
  ["hepatology research : the official journal of the japan society of hepatology"] = "Hepatol Res",
  ["hepatology, medicine and policy"] = "Hepatol Med Policy",
  ["hepatoma research"] = "Hepatoma Res",
  ["hephaistos. kritische zeitschrift zur theorie und praxis der archäologie und angrenzender wissenschaften"] = "Hephaistos",
  ["herald (harare, zimbabwe)"] = "Herald",
  ["herald of the russian academy of sciences"] = "Her. Russ. Acad. Sci.",
  ["heraldique et genealogie"] = "Her Geneal",
  ["heraldique et généalogie"] = "Her Geneal",
  ["herd: health environments research & design journal"] = "HERD: Health Environ. Res. Des. J.",
  ["hereditary cancer in clinical practice"] = "Hered Cancer Clin Pract",
  ["hereditary genetics : current research"] = "Hereditary Genet",
  ["heritage of european mathematics"] = "Herit. Eur. Math.",
  ["heritage of zimbabwe"] = "Herit Zimb",
  ["heritage review (bismarck, n.d.)"] = "Herit Rev",
  ["heritage science"] = "Herit Sci",
  ["hermes. zeitschrift für klassische philologie"] = "Hermes",
  ["hermes: zeitschrift für klassische philologie"] = "Hermes",
  ["hernia : the journal of hernias and abdominal wall surgery"] = "Hernia",
  ["heroin addiction and related clinical problems"] = "Heroin Addict Relat Clin Probl",
  ["herpes : the journal of the ihmf"] = "Herpes",
  ["herpetolgical conservation and biology"] = "Herpetol. Conser. Biol.",
  ["herpetological conservation and biology"] = "Herpetol Conserv Biol",
  ["herpetological journal"] = "Herpetol. J.",
  ["herpetological monograph"] = "Herpetological Monographs",
  ["herpetological monographs"] = "Herpetol. Monogr.",
  ["herpetological review"] = "Herpetol Rev",
  ["herpetology notes"] = "Herpetol Notes",
  ["herzschrittmachertherapie & elektrophysiologie"] = "Herzschrittmacherther Elektrophysiol",
  ["herzschrittmachertherapie + elektrophysiologie"] = "Herzschr. Elektrophysiol.",
  ["hespa mitteilungen"] = "HESPA Mitt.",
  ["hesperia. journal of the american school of classical studies at athens"] = "Hesperia",
  ["hesperian foundation news"] = "Hesperian Found News",
  ["hesperis tamuda / universite mohammed v., faculte des lettres et des sciences humaines"] = "Hesperis Tamuda",
  ["hespéris tamuda"] = "Hesperis Tamuda",
  ["hessische familienkunde"] = "Hess Familienkd",
  ["hessische floristische briefe"] = "Hess. Florist. Briefe",
  ["hessisches arzteblatt"] = "Hess Arztebl",
  ["hessisches ärzteblatt"] = "Hess Arztebl",
  ["het groene en het witte kruis"] = "Groene Witte Kruis",
  ["het hormoon"] = "Hormoon",
  ["het land van aalst"] = "Land Aalst",
  ["het ziekenhuiswezen"] = "Ziekenhuiswezen",
  ["heteroatom chemistry"] = "Heteroat. Chem",
  ["heterocera sumatrana"] = "Heteroc Sumatr",
  ["heterocyclic communications"] = "Heterocycl. Commun.",
  ["heterocyclic letters"] = "Heterocycl Lett",
  ["heterogeneous chemistry reviews"] = "Heterogen. Chem. Rev.",
  ["heterogeneous data management, polystores, and analytics for healthcare : vldb 2018 workshops, poly and dmah, rio de janeiro, brazil, august 31, 2018, revised selected papers"] = "Heterog Data Manag Polystores Anal Healthc (2018)",
  ["heuresis v"] = "Heuresis V Ling. Log. Sci.",
  ["hexagon roche"] = "Hexagon Roche Engl",
  ["hémisphères (paris, france)"] = "Hemispheres",
  ["hfsp journal"] = "HFSP J",
  ["hgg advances"] = "HGG Adv",
  ["hidrobiológica : [revista del departamento de hidrobiología]"] = "Hidrobiologica",
  ["hifu to hinyo. the dermatology and urology"] = "Hifu To Hinyo Dermatol Urol",
  ["hifuka kiyo (acta dermatologica)"] = "Hifuka Kiyo",
  ["hifuka kiyo. acta dermatologica"] = "Hifuka Kiyo",
  ["hifuka no rinsho. rinsho derma (tokyo)"] = "Hifuka No Rinsho",
  ["higashi nihon shigaku zasshi"] = "Higashi Nihon Shigaku Zasshi",
  ["higashi nippon shigaku zasshi"] = "Higashi Nippon Shigaku Zasshi",
  ["high altitude medicine & biology"] = "High Alt Med Biol",
  ["high altitude medicine and biology"] = "High Alt. Med. Biol.",
  ["high blood pressure & cardiovascular prevention : the official journal of the italian society of hypertension"] = "High Blood Press Cardiovasc Prev",
  ["high dimensional probability"] = "High Dimens Probab",
  ["high energy chemistry"] = "High Energy Chem.",
  ["high energy density physics"] = "High Energy Density Phys.",
  ["high energy physics and nuclear physics"] = "High Energy Phys. Nuclear Phys.",
  ["high entropy alloys & materials"] = "High Entropy Alloys Mater.",
  ["high frequency"] = "High Freq.",
  ["high performance computing : 31st international conference, isc high performance 2016, frankfurt, germany, june 19-23, 2016, proceedings. isc high performance (conference) (31st : 2016 : frankfurt, germany)"] = "High Perform Comput (2016)",
  ["high performance polymers"] = "High Perform. Polym.",
  ["high power laser science and engineering"] = "High Power Laser Sci. Eng.",
  ["high pressure research"] = "High Pressure Res.",
  ["high school journal (chapel hill, n.c.)"] = "High Sch J",
  ["high technology (boston, mass.)"] = "High Technol (Boston)",
  ["high technology law journal"] = "High Technol Law J",
  ["high temperature"] = "High Temp.",
  ["high temperature (ussr) [translation of teplofizika vysokikh temperatur]"] = "High Temp. (USSR)",
  ["high temperature corrosion of materials"] = "High Temp. Corros. Mater.",
  ["high temperature material processes"] = "High Temp. Mater. Processes (New York)",
  ["high temperature materials and processes"] = "High Temp. Mater. Processes (London)",
  ["high temperature materials and processes (berlin, germany)"] = "High Temp. Mater. Processes (Berlin, Ger.)",
  ["high temperature materials and processes (london, united kingdom)"] = "High Temp. Mater. Processes (London, U. K.)",
  ["high temperature materials and processes (tel aviv, israel)"] = "High Temp. Mater. Processes (Tel Aviv, Israel)",
  ["high temperature materials processes: an international quarterly of high-technology plasma processes"] = "High Temp. Mater. Processes: Int. Q. High-Technol. Plasma Processes",
  ["high temperatures - high pressures"] = "High Temp High Press",
  ["high temperatures-high pressures"] = "High Temp.-High Pressures",
  ["high voltage"] = "High Voltage",
  ["high-confidence computing"] = "High-Confid. Comput.",
  ["higher education"] = "High Educ (Dordr)",
  ["higher education pedagogies"] = "High Educ Pedagog",
  ["higher structures"] = "High. Struct.",
  ["higher-order and symbolic computation"] = "Higher-Order Symb. Comput.",
  ["highlights of astronomy"] = "Highlights Astron.",
  ["highlights of ophthalmology"] = "Highl Ophthalmol",
  ["higiene sanitaria"] = "Hig Sanit",
  ["higiene y salubridad"] = "Hig Salubr",
  ["higiene y seguridad"] = "Higiene Y Seguridad (Mexico 1936)",
  ["higijena: casopis za higijenu, mikrobiologiju, epidemiologiju i sanitarnu tehniku"] = "Hig. Cas. Hig. Mikrobiol. Epidemiol. Sanit. Teh.",
  ["higijena; casopis za higijenu, mikrobiologiju, epidemiologiju i sanitarnu tehniku"] = "Hig Cas Hig Mikrobiol Epidemiol Sanit Teh",
  ["higijena; c̆asopis za higijenu, mikrobiologiju, epidemiologiju i sanitarnu tehniku"] = "Hig Cas Hig Mikrobiol Epidemiol Sanit Teh",
  ["hildesheimer informatik-bericht"] = "Hildesheimer Inform.-Ber.",
  ["hillside journal of clinical psychiatry"] = "Hillside J. Clin. Psychiatry",
  ["himalayan chemical and pharmaceutical bulletin"] = "Himalayan Chem. Pharm. Bull.",
  ["himalayan geology"] = "Himalayan Geol.",
  ["himalayan journal of environment and zoology"] = "Himalayan J. Environ. Zool.",
  ["hims 2016 : proceedings of the 2016 international conference on health informatics and medical systems. hims (conference) (2016 : las vegas, nev.)"] = "HIMS 2016 (2016)",
  ["hindsight (saint louis, mo.)"] = "Hindsight",
  ["hindu (madras, india : daily)"] = "Hindu",
  ["hindu astronomical and mathematical text series"] = "Hindu Astronom. Math. Text Ser.",
  ["hindustan antibiotics bulletin"] = "Hindustan Antibiot Bull",
  ["hinyokika kiyo (acta urologica japonica)"] = "Hinyokika Kiyo",
  ["hinyokika kiyo. acta urologica japonica"] = "Hinyokika Kiyo",
  ["hip & pelvis"] = "Hip Pelvis",
  ["hip international : the journal of clinical and experimental research on hip pathology and therapy"] = "Hip Int",
  ["hipertensión y riesgo vascular"] = "Hipertens Riesgo Vasc",
  ["hippocrates (sausalito, calif.)"] = "Hippocrates (Sausalito)",
  ["hippokrates (helsinki, finland)"] = "Hippokrates (Helsinki)",
  ["hirosaki daigaku nōgakubu gakujutsu hōkoku = bulletin of the faculty of agriculture, hirosaki university"] = "Hirosaki Daigaku Nogakubu Gakujutsu Hokoku",
  ["hirosaki igaku = hirosaki medical journal"] = "Hirosaki Igaku",
  ["hiroshima daigaku shigaku zasshi (journal of hiroshima university dental society)"] = "Hiroshima Daigaku Shigaku Zasshi",
  ["hiroshima daigaku shigaku zasshi. the journal of hiroshima university dental society"] = "Hiroshima Daigaku Shigaku Zasshi",
  ["hiroshima journal of mathematics education"] = "Hiroshima J. Math. Ed.",
  ["hiroshima journal of medical sciences"] = "Hiroshima J Med Sci",
  ["hiroshima mathematical journal"] = "Hiroshima Math. J.",
  ["hispalis medica"] = "Hisp. Med.",
  ["hispalis medica; revista sevillana de medicina y cirugia"] = "Hisp Med",
  ["hispalis médica; revista sevillana de medicina y cirugia"] = "Hisp Med",
  ["hispania (madrid, spain : 1940)"] = "Hispania",
  ["hispania antiqua epigraphica"] = "HispAntEpigr",
  ["hispania antiqua. revista de historia antigua"] = "HispAnt",
  ["hispania antiqua: revista de historia antigua"] = "HAnt",
  ["hispania epigraphica"] = "HispEpigr",
  ["hispania sacra"] = "Hisp Sacra",
  ["hispania. revista española de historia"] = "Hispania",
  ["hispanic health care international : the official journal of the national association of hispanic nurses"] = "Hisp Health Care Int",
  ["hispanic journal of behavioral sciences"] = "Hisp J Behav Sci",
  ["hispanic review"] = "Hisp Rev",
  ["histochemical journal"] = "Histochem. J.",
  ["histochemie. histochemistry. histochimie"] = "Histochemie",
  ["histochemistry and cell biology"] = "Histochem Cell Biol",
  ["histoire & mesure"] = "Hist Mes",
  ["histoire & sociétés rurales"] = "Hist Soc Rurales",
  ["histoire (paris, france)"] = "Histoire",
  ["histoire de l'éducation"] = "Hist Educ (Paris)",
  ["histoire de l’art. bulletin d’information de l’institut national d’histoire de l’art"] = "HistArt",
  ["histoire des accidents du travail"] = "Hist Accid Trav",
  ["histoire des sciences"] = "Hist. Sci.",
  ["histoire des sciences (paris)"] = "Hist. Sci. (Paris)",
  ["histoire des sciences medicales"] = "Hist Sci Med",
  ["histoire des sciences médicales"] = "Hist Sci Med",
  ["histoire et nature"] = "Hist Nat",
  ["histoire et philosophie des sciences (paris)"] = "Hist. Philos. Sci. (Paris)",
  ["histoire magazine"] = "Hist Mag",
  ["histoire sociale. social history"] = "Histoire Soc",
  ["histoire urbaine"] = "Hist Urbaine",
  ["histoire, economie et societe"] = "Hist Econ Soc",
  ["histoire, économie et société"] = "Hist Econ Soc",
  ["histoire, médecine et santé"] = "Hist Med Sante",
  ["histoire, épistémologie, langage"] = "HEL",
  ["histoires de geometries"] = "Hist. Geom.",
  ["histoires des institutions scientifiques"] = "Hist. Inst. Sci.",
  ["histology and histopathology"] = "Histol Histopathol",
  ["historein : a review of the past and other stories"] = "Historein",
  ["historia (historical association of south africa)"] = "Hist (Hist Assoc S Afr)",
  ["historia (wiesbaden, germany)"] = "Historia",
  ["historia agraria"] = "Hist Agrar",
  ["historia antropología y fuentes orales"] = "Historia Antropol Fuentes Orales",
  ["historia contemporánea (series) (universidad del país vasco)"] = "Hist Contemp Ser Univers Pais Vasco",
  ["historia crítica (bogotá, colombia)"] = "Hist Crit",
  ["historia hospitalium"] = "Hist Hosp",
  ["historia hospitalium. sonderheft"] = "Hist Hosp",
  ["historia industrial"] = "Hist Ind",
  ["historia mathematica"] = "Historia Math.",
  ["historia mathematica: international journal of the history of mathematics"] = "HM",
  ["historia medicinae veterinariae"] = "Hist Med Vet",
  ["historia mexicana"] = "Hist Mex",
  ["historia natural (mendoza)"] = "Hist Nat (Mendoza)",
  ["historia naturalis bulgarica"] = "Hist Nat Bulg",
  ["historia nauk biologicznych i medycznych"] = "Hist Nauk Biol Med",
  ["historia ophthalmologica internationalis"] = "Hist Ophthalmol Int",
  ["historia scientiarum"] = "Historia Sci. (2)",
  ["historia scientiarum : international journal of the history of science society of japan"] = "Hist Sci (Tokyo)",
  ["historia y comunicación social"] = "Hist Comun Soc",
  ["historia y cultura"] = "Hist Cult",
  ["historia y grafía"] = "Hist Graf",
  ["historia y política (madrid, spain : 1999)"] = "Hist Polit",
  ["historia y sociedad (rio piedras, san juan, p.r.)"] = "Hist Soc",
  ["historia y sociedad (río piedras, san juan, p.r.)"] = "Hist Soc",
  ["historia y vida"] = "Hist Vida",
  ["historia. zeitschrift für alte geschichte"] = "Historia",
  ["historiallinen aikakauskirja"] = "Hist Aikak",
  ["historiallinen arkisto. suomen historiallinen seura"] = "Hist Ark",
  ["historian (london, england)"] = "Historian (Lond)",
  ["historias : revista de la dirección de estudios históricos del instituto nacional de antiopología e historia"] = "Historias (Mex)",
  ["historic preservation (washington, d.c.)"] = "Hist Preserv",
  ["historic steps"] = "Hist Steps",
  ["historica (historický ústav (akademie věd česke republiky))"] = "Hist Hist Ustav Akad Ved Ces Repub",
  ["historica. academia rsr. centrul de istorie, filologie şi etnografie din craiova"] = "Historica",
  ["historical & cultural astronomy"] = "Hist. Cult. Astron.",
  ["historical abstracts"] = "Hist. Abstr.",
  ["historical abstracts. part a, modern history abstracts 1450-1914"] = "Hist Abstr [A]",
  ["historical archaeology"] = "Hist Archaeol",
  ["historical biology"] = "Hist Biol",
  ["historical bulletin; notes and abstracts dealing with medical history"] = "Hist Bull (Calgary)",
  ["historical geography"] = "Hist Geogr",
  ["historical journal (cambridge, england)"] = "Hist J",
  ["historical journal of film, radio, and television"] = "Hist J Film Radio Telev",
  ["historical journal of massachusetts"] = "Hist J Mass",
  ["historical journal of western massachusetts"] = "Hist J West Mass",
  ["historical journal: auckland-waikato"] = "Hist J Auckl Waikato",
  ["historical life course studies"] = "Hist Life Course Stud",
  ["historical magazine of the protestant episcopal church"] = "Hist Mag Protestant Episcop Church",
  ["historical methods"] = "Hist Methods",
  ["historical methods newsletter"] = "Hist Methods Newsl",
  ["historical new hampshire"] = "Hist N H",
  ["historical news"] = "Hist News",
  ["historical papers. canadian historical association. meeting"] = "Hist Pap Can Hist Assoc",
  ["historical records of australian science"] = "Hist Rec Aust Sci",
  ["historical reflections. reflexions historiques"] = "Hist Reflect",
  ["historical reflections. réflexions historiques"] = "Hist Reflect",
  ["historical research : the bulletin of the institute of historical research"] = "Hist Res",
  ["historical studies (melbourne, vic.)"] = "Hist Stud",
  ["historical studies in the natural sciences"] = "Hist Stud Nat Sci",
  ["historical studies in the physical and biological sciences : hsps"] = "Hist Stud Phys Biol Sci",
  ["historical studies in the physical and biological sciences : hsps / office of history of science and technology, university of california, berkeley"] = "Hist Stud Phys Biol Sci",
  ["historical-analytical studies on nature, mind and action"] = "Hist.-Anal. Stud. Nat. Mind Action",
  ["historicka demografie"] = "Hist Demogr",
  ["historická demografie"] = "Hist Demogr",
  ["historicky casopis"] = "Hist Casopis",
  ["historický časopis"] = "Hist Casopis",
  ["historie (arhus, denmark)"] = "Historie (Arhus)",
  ["historie (århus, denmark)"] = "Historie (Arhus)",
  ["historielararnas forenings arsskrift"] = "Historiel Foren Arsskr",
  ["historielärarnas förenings årsskrift"] = "Historiel Foren Arsskr",
  ["historiens et géographes"] = "Historiens Geogr",
  ["historiographia linguistica"] = "Historiogr Linguist",
  ["historisch-demographische mitteilungen. communication de demographie historique"] = "Hist Demogr Mitt",
  ["historisch-demographische mitteilungen. communicatión de démographie historique"] = "Hist Demogr Mitt",
  ["historische anthropologie. kultur, gesellschaft, alltag"] = "HistAnthr",
  ["historische sozialforschung = historical social research"] = "Hist Soz Forsch",
  ["historische sprachforschung"] = "HistSprF",
  ["historische sprachforschung (historical linguistics)"] = "HSF",
  ["historische zeitschrift"] = "Hist Z",
  ["historisches jahrbuch"] = "Hist Jahrb",
  ["historisches jahrbuch der stadt linz"] = "Hist Jahrb Stadt Linz",
  ["historisk tidskrift (stockholm, sweden)"] = "Hist Tidskr",
  ["historisk tidskrift for finland"] = "Hist Tidskr Finl",
  ["historisk tidskrift för finland"] = "Hist Tidskr Finl",
  ["historisk tidsskrift : udgivet af den norske historiske forening"] = "Hist Tidsskr",
  ["historiska och litteraturhistoriska studier"] = "Hist Litteraturhist Stud",
  ["history & computing"] = "Hist Comput",
  ["history & memory"] = "Hist Mem",
  ["history & philosophy of psychology"] = "Hist Philos Psychol",
  ["history (historical association (great britain))"] = "History (Lond)",
  ["history and anthropology"] = "Hist Anthropol Chur",
  ["history and philosophy of logic"] = "Hist. Philos. Logic",
  ["history and philosophy of science—heresy, crossroads, and intersections"] = "Hist. Philos. Sci.",
  ["history and philosophy of the life sciences"] = "Hist Philos Life Sci",
  ["history and technology"] = "Hist Technol",
  ["history and theory"] = "Hist Theory",
  ["history and theory: studies in the philosophy of history"] = "H&T",
  ["history compass"] = "Hist Compass",
  ["history in africa"] = "Hist Afr",
  ["history ireland"] = "Hist Irel",
  ["history now (christchurch, n.z.)"] = "Hist Now (Christch)",
  ["history of agriculture"] = "Hist Agric",
  ["history of analytic philosophy"] = "Hist. Anal. Philos.",
  ["history of anthropology"] = "Hist Anthropol",
  ["history of chemistry"] = "Hist. Chem.",
  ["history of childhood quarterly"] = "Hist Child Q",
  ["history of computing"] = "Hist. Comput.",
  ["history of economic ideas"] = "Hist. Econ. Ideas",
  ["history of economics review"] = "Hist. Econ. Rev.",
  ["history of education"] = "Hist Educ",
  ["history of education quarterly"] = "Hist Educ Q",
  ["history of european ideas"] = "Hist Eur Ideas",
  ["history of geo- and space science"] = "Hist. Geo Space Sci.",
  ["history of geo- and space sciences"] = "Hist Geo Space Sci",
  ["history of international public health"] = "Hist Int Public Health",
  ["history of mathematics"] = "Hist. Math.",
  ["history of mathematics education"] = "Hist. Math. Educ.",
  ["history of mathematics subseries"] = "Hist. Math. Subser.",
  ["history of mechanism and machine science"] = "Hist. Mech. Mach. Sci.",
  ["history of medicine"] = "Hist Med",
  ["history of medicine (mobile, ala.)"] = "Hist Med Univ South Ala Coll Med",
  ["history of neuroscience"] = "Hist Neurosci",
  ["history of nursing bulletin"] = "Hist Nurs Bull",
  ["history of nursing society journal"] = "Hist Nurs Soc J",
  ["history of philosophy quarterly"] = "HPhQ",
  ["history of photography"] = "Hist Photogr",
  ["history of physics"] = "Hist. Phys.",
  ["history of political economy"] = "Hist Polit Econ",
  ["history of political thought"] = "Hist Polit Thought",
  ["history of psychiatry"] = "Hist Psychiatry",
  ["history of psychoanalysis"] = "Hist Psychoanal",
  ["history of psychology"] = "Hist Psychol",
  ["history of religions"] = "Hist Relig",
  ["history of science"] = "Hist Sci",
  ["history of science and technology in china"] = "Hist. Sci. Tech. China",
  ["history of science and technology reprint series"] = "Hist. Sci. Tech. Reprint Ser.",
  ["history of science; an annual review of literature, research and teaching"] = "Hist Sci",
  ["history of technology"] = "HTechn",
  ["history of the human sciences"] = "Hist Human Sci",
  ["history of the present (champaign, ill.)"] = "Hist Present (Champaign)",
  ["history of universities"] = "Hist Univ",
  ["history scotland"] = "Hist Scotl",
  ["history today"] = "Hist Today",
  ["history workshop"] = "Hist Workshop",
  ["history workshop journal : hwj"] = "Hist Workshop J",
  ["história unisinos : revista do programa de pós-graduação em historía da universidade do vale do rio dos sinos"] = "Historia Unisinos",
  ["história, questões & debates : revista da associação paranaense de história"] = "Hist Questoes Debates",
  ["histria antiqua. casopis meunarodnog istraivakog centra za arheologiju. journal of the international research centre for archeology"] = "HistriaAnt",
  ["histria archaeologica"] = "HistriaA",
  ["histria. les résultats des fouilles"] = "Histria",
  ["história, ciências, saúde--manguinhos"] = "Hist Cienc Saude Manguinhos",
  ["hitotsubashi journal of arts & sciences"] = "Hitotsubashi J. Arts Sci.",
  ["hitotsubashi journal of commerce and management"] = "Hitotsubashi J. Com. Manage.",
  ["hitotsubashi journal of economics"] = "Hitotsubashi J. Econ.",
  ["hitotsubashi journal of law & politics"] = "Hitotsubashi J Law Polit",
  ["hiv advance research and development : open access"] = "HIV Adv Res Dev",
  ["hiv capsule report"] = "HIV Capsule Rep",
  ["hiv clinical trials"] = "HIV Clin Trials",
  ["hiv clinician"] = "HIV Clin",
  ["hiv clinician / delta region aids education & training center"] = "HIV Clin",
  ["hiv hotline"] = "HIV Hotline",
  ["hiv inside : a newsletter for correctional professionals"] = "HIV Inside",
  ["hiv medicine"] = "HIV Med",
  ["hiv prevention plus!"] = "HIV Prev Plus",
  ["hiv research & clinical practice"] = "HIV Res Clin Pract",
  ["hiv sequence compendium"] = "HIV Seq Compend",
  ["hiv specialist"] = "HIV Spec",
  ["hiv therapy"] = "HIV Ther",
  ["hiv/aids (auckland, n.z.)"] = "HIV AIDS (Auckl)",
  ["hiv/aids policy & law review"] = "HIV AIDS Policy Law Rev",
  ["hiv/aids policy & law review / canadian hiv/aids legal network"] = "HIV AIDS Policy Law Rev",
  ["hiv/aids research and treatment : open journal"] = "HIV/AIDS Res Treat",
  ["hiv/aids surveillance report (atlanta, ga.)"] = "HIV AIDS Surveill Rep",
  ["hkie transactions"] = "HKIE Trans.",
  ["hmo practice"] = "HMO Pract",
  ["hmo practice / hmo group"] = "HMO Pract",
  ["hmong studies journal"] = "Hmong Stud J",
  ["ho. kyushu byogaichu kenkyukai"] = "Ho Kyushu Byogaichu Kenkyukai",
  ["hoaj biology"] = "HOAJ Biol",
  ["hochschuldidaktische materialien"] = "Hochschuldidakt. Material.",
  ["hofstra law review"] = "Hofstra Law Rev",
  ["hohenheimer volkswirtschaftliche schriften"] = "Hohenh. Volkswirtsch. Schr.",
  ["hoja tisiologica"] = "Hoja Tisiol.",
  ["hoja tisiológica"] = "Hoja Tisiol",
  ["hokenfu zasshi (japanese journal for public health nurse)"] = "Hokenfu Zasshi",
  ["hokkaido igaku zasshi (hokkaido journal of medical science)"] = "Hokkaido Igaku Zasshi",
  ["hokkaido igaku zasshi. hokkaido journal of medical science"] = "Hokkaido Igaku Zasshi",
  ["hokkaido mathematical journal"] = "Hokkaido Math. J.",
  ["hokkaido shika ishikai shi"] = "Hokkaido Shika Ishikai Shi",
  ["holdsworth law review"] = "Holdsworth Law Rev",
  ["holdsworth law review / university of birmingham"] = "Holdsworth Law Rev",
  ["holistic assertive nurse"] = "Holistic Assertive Nurse",
  ["holistic nursing practice"] = "Holist Nurs Pract",
  ["holocaust and genocide studies"] = "Holocaust Genocide Stud",
  ["holocene, the"] = "Holocene",
  ["holz als roh und werkstoff"] = "Holz Roh Werkst.",
  ["holz als roh- und werkstoff"] = "Holz Roh- Werkst.",
  ["holz, schweizerische holzzeitung"] = "Holz, Schweiz. Holzztg.",
  ["holzforschung schweiz"] = "Holzforsch. Schweiz",
  ["holzpreisstatistik – rohholz"] = "Holzpreisstat. – Rohholz",
  ["holzzucht, die"] = "Holzzucht",
  ["home care economics"] = "Home Care Econ",
  ["home care manager"] = "Home Care Manag",
  ["home care provider"] = "Home Care Provid",
  ["home health care management & practice"] = "Home Health Care Manag Pract",
  ["home health care services quarterly"] = "Home Health Care Serv Q",
  ["home health journal"] = "Home Health J",
  ["home health review"] = "Home Health Rev",
  ["home healthcare now"] = "Home Healthc Now",
  ["home healthcare nurse"] = "Home Healthc Nurse",
  ["home healthcare nurse manager"] = "Home Healthc Nurse Manag",
  ["home hemodialysis international. international symposium on daily home hemodialysis"] = "Home Hemodial Int (1997)",
  ["home hemodialysis international. international symposium on home hemodialysis"] = "Home Hemodial Int",
  ["homemaker's magazine"] = "Homemak Mag",
  ["homemaker’s magazine"] = "Homemak Mag",
  ["homeopathy : the journal of the faculty of homeopathy"] = "Homeopathy",
  ["homeostasis in health and disease"] = "Homeost. Health Dis.",
  ["homeostasis in health and disease : international journal devoted to integrative brain functions and homeostatic systems"] = "Homeost Health Dis",
  ["homicide studies"] = "Homicide Stud",
  ["homme (vienna, austria)"] = "Homme (Vienna Aust)",
  ["hommes & migrations"] = "Hommes Migr",
  ["hommes et migrations. documents"] = "Hommes Migr Doc",
  ["hommes et terres du nord"] = "Hommes Terres Nord",
  ["homo : internationale zeitschrift fur die vergleichende forschung am menschen"] = "Homo",
  ["homo : internationale zeitschrift für die vergleichende forschung am menschen"] = "Homo",
  ["homoeopathic bulletin"] = "Homeopath Bull",
  ["homoeopathic herald"] = "Homeopath Her",
  ["homology homotopy and applications\t."] = "Homol. Homotopy Appl.",
  ["homology, homotopy and applications"] = "Homology Homotopy Appl.",
  ["honam mathematical journal"] = "Honam Math. J.",
  ["honam mathematical society"] = "Bull. Honam Math. Soc.",
  ["hong kong economic papers"] = "Hong Kong Econ. Pap.",
  ["hong kong journal of occupational therapy : hkjot"] = "Hong Kong J Occup Ther",
  ["hong kong journal of ophthalmology : the official publication of the college of ophthalmologists of hong kong = xianggang yan ke xue kan : xianggang yan ke yi xue yuan"] = "Hong Kong J Ophthalmol",
  ["hong kong law journal"] = "Hong Kong Law J",
  ["hong kong medical journal = xianggang yi xue za zhi"] = "Hong Kong Med J",
  ["hong kong monthly digest of statistics"] = "Hong Kong Mon Dig Stat",
  ["hong kong pharmaceutical journal"] = "Hong Kong Pharm J",
  ["hong kong physiotherapy journal : official publication of the hong kong physiotherapy association limited = wu li chih liao"] = "Hong Kong Physiother J",
  ["hong qi"] = "Hong Qi",
  ["honismereti módszertani füzetek"] = "Szabolcs Szatmari Szle",
  ["honyūrui kagaku"] = "Honyurui Kagaku",
  ["hoosier health herald"] = "Hoosier Health Her",
  ["hoppe-seyler's zeitschrift für physiologische chemie"] = "Hoppe Seylers Z Physiol Chem",
  ["hoppe-seylers zeitschrift fur physiologische chemie"] = "Hoppe. Seylers Z. Physiol. Chem.",
  ["hoppe-seyler’s zeitschrift fur physiologische chemie"] = "Hoppe-Seyler’s Z. Physiol. Chem.",
  ["hoppe-seyler’s zeitschrift für physiologische chemie"] = "Hoppe-Seyler’s Z. Physiol. Chem.",
  ["hora médica"] = "Hora Med",
  ["horizon : a magazine of the arts"] = "Horizon",
  ["horizons in biochemistry and biophysics"] = "Horiz Biochem Biophys",
  ["horizons in cancer research"] = "Horiz Cancer Res",
  ["horizons medicaux"] = "Horiz Med",
  ["horizons médicaux"] = "Horiz Med",
  ["horizonte de enfermería"] = "Horiz Enferm",
  ["horizonte médico"] = "Horiz Med (Barcelona)",
  ["hormone and metabolic research"] = "Horm. Metab. Res.",
  ["hormone and metabolic research = hormon- und stoffwechselforschung = hormones et métabolisme"] = "Horm Metab Res",
  ["hormone and metabolic research. hormon- und stoffwechselforschung. hormones et metabolisme"] = "Horm Metab Res",
  ["hormone and metabolic research. supplement"] = "Horm. Metab. Res. Suppl.",
  ["hormone and metabolic research. supplement series"] = "Horm Metab Res Suppl",
  ["hormone molecular biology and clinical investigation"] = "Horm Mol Biol Clin Investig",
  ["hormone research"] = "Horm Res",
  ["hormone research in paediatrics"] = "Horm. Res. Paediatr.",
  ["hormone research in pædiatrics"] = "Horm Res Paediatr",
  ["hormones & cancer"] = "Horm Cancer",
  ["hormones (athens, greece)"] = "Hormones (Athens)",
  ["hormones and behavior"] = "Horm Behav",
  ["horticultura brasileira"] = "Hortic Bras",
  ["horticultural plant journal"] = "Hortic Plant J",
  ["horticultural reviews"] = "Hortic Rev (Am Soc Hortic Sci)",
  ["horticulture research"] = "Hortic Res",
  ["horticulture, environment and biotechnology"] = "Hortic. Environ. Biotechnol.",
  ["hortscience : a publication of the american society for horticultural science"] = "HortScience",
  ["horumon to rinsho"] = "Horumon to Rinsho",
  ["horumon to rinsho (clinical endocrinology)"] = "Horumon To Rinsho",
  ["horumon to rinsho. clinical endocrinology"] = "Horumon To Rinsho",
  ["horumon to rinshō. clinical endocrinology"] = "Horumon To Rinsho",
  ["horwood series in mathematics & applications"] = "Horwood Ser. Math. Appl.",
  ["hospice journal"] = "Hosp. J.",
  ["hospital & community psychiatry"] = "Hosp Community Psychiatry",
  ["hospital & health services administration"] = "Hosp Health Serv Adm",
  ["hospital (rio de janeiro, brazil)"] = "Hospital (Rio J)",
  ["hospital administration"] = "Hosp Adm (New Delhi)",
  ["hospital administration currents"] = "Hosp Admin Curr",
  ["hospital administration in canada"] = "Hosp Adm Can",
  ["hospital admitting monthly"] = "Hosp Admitting Mon",
  ["hospital and community psychiatry"] = "Hosp. Community Psychiatry",
  ["hospital and health management"] = "Hosp Health Manag",
  ["hospital and health services administration"] = "Hosp. Health Serv. Adm.",
  ["hospital and health services review"] = "Hosp. Health Serv. Rev.",
  ["hospital aviation"] = "Hosp Aviat",
  ["hospital bond review"] = "Hosp Bond Rev",
  ["hospital capital finance"] = "Hosp. Cap. Finance",
  ["hospital capital finance american hospital association, division of hospital planning and capital finance"] = "Hosp Cap Finance",
  ["hospital case management"] = "Hosp. Case Manag.",
  ["hospital case management : the monthly update on hospital-based care planning and critical paths"] = "Hosp Case Manag",
  ["hospital chronicles = nosokomeiaka chronika"] = "Hosp Chron",
  ["hospital corps quarterly"] = "Hosp Corps Q",
  ["hospital cost management and accounting"] = "Hosp Cost Manag Account",
  ["hospital council bulletin"] = "Hosp Counc Bull",
  ["hospital development"] = "Hosp Dev",
  ["hospital employee health"] = "Hosp Employee Health",
  ["hospital engineering"] = "Hosp Eng",
  ["hospital entrepreneurs newsletter"] = "Hosp. Entrep. Newsl.",
  ["hospital entrepreneurs' newsletter"] = "Hosp Entrep Newsl",
  ["hospital entrepreneurs’ newsletter"] = "Hosp Entrep Newsl",
  ["hospital equipment & supplies"] = "Hosp Equip Supplies",
  ["hospital equipment and supplies"] = "Hosp. Equip. Supplies",
  ["hospital ethics"] = "Hosp Ethics",
  ["hospital ethics / american hospital association"] = "Hosp Ethics",
  ["hospital financial management"] = "Hosp Financ Manage",
  ["hospital food & nutrition focus"] = "Hosp Food Nutr Focus",
  ["hospital food and nutrition focus"] = "Hosp. Food Nutr. Focus",
  ["hospital formulary"] = "Hosp Formul",
  ["hospital forum"] = "Hosp Forum",
  ["hospital gauzette"] = "Hosp Gauzette",
  ["hospital general"] = "Hosp Gen (Madr)",
  ["hospital gift shop management"] = "Hosp Gift Shop Manage",
  ["hospital guest relations report"] = "Hosp Guest Relations Rep",
  ["hospital hazardous materials management"] = "Hosp Hazard Mater Manage",
  ["hospital health care newsletter"] = "Hosp. Health Care Newsl.",
  ["hospital infection control"] = "Hosp Infect Control",
  ["hospital law newsletter"] = "Hosp Law Newsl",
  ["hospital libraries"] = "Hosp Libr",
  ["hospital management"] = "Hosp Manage",
  ["hospital management communications"] = "Hosp Manage Commun",
  ["hospital management quarterly"] = "Hosp. Manage. Q.",
  ["hospital management quarterly : hmq"] = "Hosp Manage Q",
  ["hospital manager"] = "Hosp. Manager",
  ["hospital material[dollar sign] management"] = "Hosp Mater Manage",
  ["hospital materials management"] = "Hosp. Mater. Manage.",
  ["hospital materiel management quarterly"] = "Hosp Mater Manage Q",
  ["hospital medical staff"] = "Hosp. Med. Staff",
  ["hospital medicine"] = "Hosp. Med.",
  ["hospital medicine (london, england : 1998)"] = "Hosp Med",
  ["hospital medicine clinics"] = "Hosp Med Clin",
  ["hospital outlook"] = "Hosp Outlook",
  ["hospital patient relations report"] = "Hosp Patient Relat Rep",
  ["hospital pediatrics"] = "Hosp Pediatr",
  ["hospital peer review"] = "Hosp Peer Rev",
  ["hospital pharmacy"] = "Hosp Pharm",
  ["hospital physician"] = "Hosp Physician",
  ["hospital practice"] = "Hosp Pract",
  ["hospital practice (1995)"] = "Hosp Pract (1995)",
  ["hospital practice (hospital ed.)"] = "Hosp Pract (Hosp Ed)",
  ["hospital practice (hospital edition)"] = "Hosp. Pract. (Hosp. Ed.)",
  ["hospital practice (office ed.)"] = "Hosp Pract (Off Ed)",
  ["hospital practice (office edition)"] = "Hosp. Pract. (Off. Ed.)",
  ["hospital practices and research"] = "Hosp Pract Res",
  ["hospital progress"] = "Hosp Prog",
  ["hospital progress. directory issue"] = "Hosp Prog Directory Issue",
  ["hospital purchasing management"] = "Hosp Purch Manage",
  ["hospital purchasing news : hpn"] = "HPN Hosp Purch News",
  ["hospital quarterly"] = "Hosp Q",
  ["hospital record study : a joint study by cpha and ims america ltd"] = "Hosp Rec Study",
  ["hospital record study : a joint study by cpha and ims america ltd. / ims america"] = "Hosp Rec Study",
  ["hospital revenue report"] = "Hosp Revenue Rep",
  ["hospital risk management"] = "Hosp Risk Manage",
  ["hospital security and safety management"] = "Hosp Secur Saf Manage",
  ["hospital strategy report"] = "Hosp Strategy Rep",
  ["hospital supervision"] = "Hosp Superv",
  ["hospital supervisor's bulletin"] = "Hosp Superv Bull",
  ["hospital supervisors bulletin"] = "Hosp. Superv. Bull.",
  ["hospital supervisor’s bulletin"] = "Hosp Superv Bull",
  ["hospital technology series"] = "Hosp Technol Ser",
  ["hospital topics"] = "Hosp Top",
  ["hospital tribune"] = "Hosp Trib",
  ["hospital trustee"] = "Hosp Trustee",
  ["hospitals & health networks"] = "Hosp Health Netw",
  ["hospitals & health networks / aha"] = "Hosp Health Netw",
  ["hospitals & health networks 360 degrees : h&hn 360 degrees"] = "Hosp Health Netw 360",
  ["hospitals & health networks 360° : h&hn 360°"] = "Hosp Health Netw 360",
  ["hospitals and health networks"] = "Hosp. Health Netw.",
  ["hotel + touristik revue"] = "Hotel Tour. Rev.",
  ["hotel journal"] = "Hotel J.",
  ["hotetsu rinsho (practice in prosthodontics)"] = "Hotetsu Rinsho",
  ["hotetsu rinsho. practice in prosthodontics"] = "Hotetsu Rinsho",
  ["houille blanche"] = "Houille Blanche",
  ["household & personal products industry"] = "Household Pers. Prod. Ind.",
  ["housework: craft production and domestic economy in ancient mesoamerica"] = "Archeol. Pap. Am. Anthr.",
  ["housing and society"] = "Hous Soc",
  ["housing finance review"] = "Housing Finance Rev.",
  ["housing policy debate"] = "Hous Policy Debate",
  ["housing studies"] = "Hous Stud",
  ["houston journal of health law & policy"] = "Houst J Health Law Policy",
  ["houston journal of international law"] = "Houst J Int Law",
  ["houston journal of mathematics"] = "Houston J. Math.",
  ["houston law review"] = "Houst Law Rev",
  ["houston law review / university of houston"] = "Houst Law Rev",
  ["houston lawyer : official monthly publication of the houston bar association"] = "Houst Lawyer",
  ["how to evaluate health programs"] = "How Eval. Health Programs",
  ["how to: evaluate health programs"] = "How Eval Health Programs",
  ["howard law journal"] = "Howard Law J",
  ["hpa resource"] = "HPA Resour",
  ["hpb : the official journal of the international hepato pancreato biliary association"] = "HPB (Oxford)",
  ["hpb surgery"] = "HPB Surg.",
  ["hpb surgery (yverdon-les-bains"] = "HPB Surg",
  ["hpb surgery : a world journal of hepatic, pancreatic and biliary surgery"] = "HPB Surg",
  ["hpn: hospital purchasing news"] = "HPN Hosp. Purch. News",
  ["hrb open research"] = "HRB Open Res",
  ["hrc & cc. journal of high resolution chromatography & chromatography communications"] = "HRC CC J High Resolut Chromatogr Chromatogr Commun",
  ["hrc journal of high resolution chromatography"] = "HRC J. High Resolut. Chromatogr.",
  ["hrmagazine : on human resource management"] = "HRMAGAZINE",
  ["hrsa careaction"] = "HRSA Careaction",
  ["hsmha health reports"] = "HSMHA Health Rep",
  ["hsoa journal of addiction & addictive disorders"] = "HSOA J Addict Addict Disord",
  ["hsoa journal of addiction & neuropharmacology"] = "HSOA J Addict Neuropharmacol",
  ["hsoa journal of aids clinical research & stds"] = "HSOA J AIDS Clin Res STDs",
  ["hsoa journal of biotech research & biochemistry"] = "HSOA J Biotech Res Biochem",
  ["hsoa journal of cell biology & cell metabolism"] = "HSOA J Cell Biol Cell Metabol",
  ["hsoa journal of community medicine & public health care"] = "HSOA J Community Med Public Health Care",
  ["hsoa journal of food science & nutrition"] = "HSOA J Food Sci Nutr",
  ["hsoa journal of gastroenterology & hepatology research"] = "HSOA J Gastroenterol Hepatol Res",
  ["hsoa journal of human endocrinology"] = "HSOA J Hum Endocrinol",
  ["hsoa journal of infectious & non infectious diseases"] = "HSOA J Infect Non Infect Dis",
  ["hsoa journal of nephrology & renal therapy"] = "HSOA J Nephrol Ren Ther",
  ["hsoa journal of ophthalmology & clinical research"] = "HSOA J Ophthalmol Clin Res",
  ["hsoa journal of psychiatry, depression & anxiety"] = "HSOA J Psychiatry Depress Anxiety",
  ["hsoa journal of reproductive medicine gynaecology & obstetrics"] = "HSOA J Reprod Med Gynaecol Obstet",
  ["hsoa journal of stem cells research, development & therapy"] = "HSOA J Stem Cells Res Dev Ther",
  ["hsoa journal of toxicology : current research"] = "HSOA J Toxicol",
  ["hsoa trends in anatomy and physiology"] = "HSOA Trends Anat Physiol",
  ["hsr proceedings in intensive care & cardiovascular anesthesia"] = "HSR Proc Intensive Care Cardiovasc Anesth",
  ["hss journal : the musculoskeletal journal of hospital for special surgery"] = "HSS J",
  ["hstc bulletin"] = "Hstc Bull",
  ["hsueh yu ch"] = "Sheng Chung Ping Tsa Chih",
  ["htd (american society of mechanical engineers)"] = "HTD (Am. Soc. Mech. Eng.)",
  ["htm journal of heat treatment and materials"] = "HTM J. Heat Treat. Mater.",
  ["htm, haerterei technische mitteilungen"] = "HTM Haerteri Tech. Mitt.",
  ["htm, haerterei-technische mitteilungen"] = "HTM, Haerterei Tech. Mit.",
  ["hu li yan jiu = nursing research"] = "Hu Li Yan Jiu",
  ["hu li za zhi (journal of nursing)"] = "Hu Li Za Zhi",
  ["hu li za zhi the journal of nursing"] = "Hu Li Za Zhi",
  ["hu po ke xue = journal of lake sciences"] = "Hu Po Ke Xue",
  ["hu'nan agricultural science & technology"] = "Hunan Agric Sci Technol",
  ["hu'nan keji xueyuan xuebao"] = "Hunan Keji Xueyuan Xuebao",
  ["hu-nan i ko ta hsueh hsueh pao bulletin of hunan medical university (chang-sha"] = "Hunan I Ko Ta Hsueh Hsueh Pao",
  ["hua bei nong xue bao = acta agriculturae boreali-sinica"] = "Hua Bei Nong Xue Bao",
  ["hua dong li gong da xue xue bao. zi ran ke xue ban = journal of east china university of science and technology. natural science edition"] = "Hua Dong Li Gong Da Xue Xue Bao Zi Ran Ke Xue Ban",
  ["hua gong jin zhan = chemical industry and engineering progress : ciep"] = "Huagong Jinzhan",
  ["hua gong xue bao = chemical industry and engineering"] = "Hua Gong Xue Bao",
  ["hua hsi kou chiang i hsueh tsa chih west china journal of stomatology (cheng-tu"] = "Hua Hsi Kou Chiang I Hsueh Tsa Chih",
  ["hua nan nong ye da xue xue bao"] = "Journal of South China Agricultural University|Hua Nan Nong Ye Da Xue Xue Bao",
  ["hua nan nong ye da xue xue bao = journal of south china agricultural university"] = "Hua Nan Nong Ye Da Xue Xue Bao",
  ["hua sheng xue bao"] = "Journal of peanut science|Hua Sheng Xue Bao",
  ["hua sheng xue bao = journal of peanut science"] = "Hua Sheng Xue Bao",
  ["hua xi kou qiang yi xue za zhi = huaxi kouqiang yixue zazhi = west china journal of stomatology"] = "Hua Xi Kou Qiang Yi Xue Za Zhi",
  ["hua xi yi ke da xue xue bao (journal of west china university of medical sciences)"] = "Hua Xi Yi Ke Da Xue Xue Bao",
  ["hua xi yi ke da xue xue bao = journal of west china university of medical sciences = huaxi yike daxue xuebao"] = "Hua Xi Yi Ke Da Xue Xue Bao",
  ["hua xue xue bao. acta chimica sinica"] = "Huaxue Xuebao",
  ["hua zhong nong ye da xue xue bao"] = "Journal Huazhong (Central China) Agricultural University|Hua Zhong Nong Ye Da Xue Xue Bao",
  ["hua zhong nong ye da xue xue bao = journal huazhong (central china) agricultural university"] = "Hua Zhong Nong Ye Da Xue Xue Bao",
  ["hua zhong shi fan da xue xue bao. zi ran ke xue ban = journal of central china normal university. natural sciences edition"] = "Hua Zhong Shi Fan Da Xue Xue Bao Zi Ran Ke Xue Ban",
  ["hua-hsi i ko ta hsueh hsueh pao [journal of west china university of medical sciences]"] = "Hua Hsi I Ko Ta Hsueh Hsueh Pao",
  ["huadong huagong xueyuan xuebao"] = "Huadong Huagong Xueyuan Xuebao",
  ["huadong ligong daxue xuebao"] = "Huadong Ligong Daxue Xuebao",
  ["huadong ligong daxue xuebao, ziran kexueban"] = "Huadong Ligong Daxue Xuebao, Ziran Kexueban",
  ["huadong shifan daxue xuebao"] = "Journal of East China Normal University|Huadong Shifan Daxue Xuebao",
  ["huadong shifan daxue xuebao = journal of east china normal university"] = "Huadong Shifan Daxue Xuebao",
  ["huagong xuebao (chinese edition)"] = "Huagong Xuebao (Chin. Ed.)",
  ["huaihai gongxueyuan xuebao : ziran kexue ban"] = "Huaihai Gongxueyuan Xuebao",
  ["huaihua shizhuan xuebao"] = "Huaihua Shizhuan Xuebao",
  ["huan jing hua xue = environmental chemistry"] = "Huanjing Huaxue",
  ["huan jing ke xue xue bao = acta scientiae circumstantiae"] = "Huanjing Kexue Xuebao",
  ["huan jing ke xue= huanjing kexue"] = "Huan Jing Ke Xue",
  ["huan jing kun chong xue bao"] = "Huan Jing Kun Chong Xue Bao",
  ["huan jing wu ran zhi li ji shu yu she bei"] = "Techniques and equipment for environmental pollution control / Zhongguo ke xue yuan sheng tai huan jing yan jiu zhong xin zhu ban|Huan Jing Wu Ran Zhi Li Ji Shu Yu She Bei",
  ["huan jing wu ran zhi li ji shu yu she bei = techniques and equipment for environmental pollution control"] = "Huan Jing Wu Ran Zhi Li Ji Shu Yu She Bei",
  ["huan jing yu jian kang za zhi = journal of environment and health"] = "Huan Jing Yu Jian Kang Za Zhi",
  ["huanan shifan daxue xuebao. ziran kexue ban. huanan shifan daxue"] = "Huanan Shifan Daxue Xuebao",
  ["huanjing huaxue"] = "Huanjing Huaxue",
  ["huanjing kexue"] = "Huanjing Kexue",
  ["huanjing kexue xuebao"] = "Huanjing Kexue Xuebao",
  ["huaxue fanying gongcheng yu gongyi"] = "Huaxue Fanying Gongcheng Yu Gongyi",
  ["huaxue shiji"] = "Huaxue Shiji",
  ["huaxue shijie"] = "Huaxue Shijie",
  ["huaxue tongbao"] = "Huaxue Tongbao",
  ["huaxue wuli xuebao"] = "Huaxue Wuli Xuebao",
  ["huaxue xuebao"] = "Huaxue Xuebao",
  ["huazhong (central china) university of science and technology"] = "J. Huazhong Univ. Sci. Tech.",
  ["hubei nong ye ke xue"] = "Hubei Nong Ye Ke Xue",
  ["hudson county health record [monthly]"] = "Hudson Cty Health Rec",
  ["huelva arqueológica"] = "HuelvaA",
  ["hugoye : journal of syriac studies"] = "Hugoye (Wash D C)",
  ["huisarts en wetenschap"] = "Huisarts Wet",
  ["human & ecological risk assessment"] = "Hum. Ecol. Risk Assess.",
  ["human & experimental toxicology"] = "Hum Exp Toxicol",
  ["human affairs (bratislava, slovakia)"] = "Hum Affairs (Bratisl)",
  ["human and ecological risk assessment"] = "Hum. Ecol. Risk Assess.",
  ["human and ecological risk assessment : hera"] = "Hum Ecol Risk Assess",
  ["human and ecological risk assessment: an international journal"] = "Hum. Ecol. Risk Assess.: Int. J.",
  ["human and experimental toxicology"] = "Hum. Exp. Toxicol.",
  ["human antibodies"] = "Hum Antibodies",
  ["human antibodies and hybridomas"] = "Hum Antibodies Hybridomas",
  ["human aspects of it for the aged population : design for aging : first international conference, itap 2015, held as part of hci international 2015, los angeles, ca, usa, august 2-7, 2015. proceedings. part i. itap (conference) (1st : 2015 : los angeles, calif.)"] = "Hum Asp IT Aged Popul (2015)",
  ["human aspects of it for the aged population : healthy and active aging : second international conference, itap 2016, held as part of hci international 2016 toronto, on, canada, july 17-22, 2016, proceedings. part ii. itap (conference) (2nd : 2016 : toronto, ont.)"] = "Human Asp IT Aged Popul (2016)",
  ["human behavior and emerging technologies"] = "Hum Behav Emerg Technol",
  ["human behavior; the newsmagazine of the social sciences"] = "Hum Behav",
  ["human biology"] = "Hum Biol",
  ["human biology in oceania"] = "Hum Biol Oceania",
  ["human biology; an international record of research"] = "Hum Biol",
  ["human brain and artificial intelligence : first international workshop, hbai 2019, held in conjunction with ijcai 2019, macao, china, august 12, 2019, revised selected papers. hbai (conference) (1st : 2019 : macau, china)"] = "Hum Brain Artif Intell (2019)",
  ["human brain mapping"] = "Hum Brain Mapp",
  ["human cell"] = "Hum Cell",
  ["human cell : official journal of human cell research society"] = "Hum Cell",
  ["human communication research"] = "Hum Commun Res",
  ["human communication research series"] = "Human Comm. Res. Ser.",
  ["human computation (fairfax, va.)"] = "Hum Comput (Fairfax)",
  ["human development"] = "Hum Dev",
  ["human dimensions of wildlife"] = "Hum. Dimens. Wildl.",
  ["human ecology"] = "Hum Ecol",
  ["human ecology: an interdisciplinary journal"] = "Hum Ecol Interdiscip J",
  ["human ethology bulletin"] = "Hum Ethol Bull",
  ["human events"] = "Hum Events",
  ["human evolution"] = "Hum Evol",
  ["human factors"] = "Hum Factors",
  ["human factors and ergonomics in manufacturing"] = "Hum Factors Ergon Manuf",
  ["human factors and ergonomics in manufacturing & service industries"] = "Hum. Factors Ergon. Manuf. Serv. Ind.",
  ["human factors and mechanical engineering for defense and safety"] = "Hum. Factors Mech. Eng. Def. Saf.",
  ["human factors in healthcare"] = "Hum. Factors Healthcare",
  ["human fertility"] = "Hum Fertil",
  ["human fertility (cambridge, england)"] = "Hum Fertil (Camb)",
  ["human gene therapy"] = "Hum Gene Ther",
  ["human gene therapy clinical development"] = "Hum. Gene Ther. Clin. Dev.",
  ["human gene therapy methods"] = "Hum. Gene Ther. Methods",
  ["human gene therapy. clinical development"] = "Hum Gene Ther Clin Dev",
  ["human gene therapy: clinical development"] = "Hum. Gene Ther.: Clin. Dev.",
  ["human gene therapy: methods"] = "Hum. Gene Ther: Methods",
  ["human genetics"] = "Hum Genet",
  ["human genetics & embryology : current research"] = "Hum Genet Embryol",
  ["human genetics and genomics advances"] = "Hum. Genet. Genomics Adv.",
  ["human genetics. supplement"] = "Hum Genet Suppl",
  ["human genome news"] = "Hum Genome News",
  ["human genome news / national center for human genome research, national institutes of health"] = "Hum Genome News",
  ["human genome variation"] = "Hum Genome Var",
  ["human genomics"] = "Hum Genomics",
  ["human genomics and proteomics : hgp"] = "Hum Genomics Proteomics",
  ["human geographies"] = "Human Geogr",
  ["human heredity"] = "Hum Hered",
  ["human immunology"] = "Hum Immunol",
  ["human microbiome journal"] = "Hum Microb J",
  ["human molecular genetics"] = "Hum Mol Genet",
  ["human movement science"] = "Hum Mov Sci",
  ["human mutation"] = "Hum Mutat",
  ["human nature"] = "Hum. Nat.",
  ["human nature (hawthorne, n.y.)"] = "Hum Nat",
  ["human neurobiology"] = "Hum Neurobiol",
  ["human nutrition & metabolism"] = "Hum. Nutr. Metab.",
  ["human nutrition. applied nutrition"] = "Hum Nutr Appl Nutr",
  ["human nutrition. clinical nutrition"] = "Hum Nutr Clin Nutr",
  ["human organization"] = "Hum Organ",
  ["human parasitic diseases (auckland, n.z.)"] = "Hum Parasit Dis (Auckl)",
  ["human pathology"] = "Hum Pathol",
  ["human pathology (new york)"] = "Hum Pathol (N Y)",
  ["human pathology: case reports"] = "Hum. Pathol.: Case Rep.",
  ["human performance"] = "Hum Perform",
  ["human performance in extreme environments : the journal of the society for human performance in extreme environments"] = "Hum Perf Extrem Environ",
  ["human perspectives in health sciences and technology"] = "Hum. Perspect. Health Sci. Tech.",
  ["human physiology"] = "Hum Physiol",
  ["human psychopharmacology"] = "Hum Psychopharmacol",
  ["human psychopharmacology - clinical and experimental"] = "Hum. Psychopharmacol. - Clin. Exp.",
  ["human psychopharmacology: clinical and experimental"] = "Hum. Psychopharmacol. Clin. Exp.",
  ["human relations; studies towards the integration of the social sciences"] = "Hum Relat",
  ["human reproduction"] = "Hum. Reprod.",
  ["human reproduction (oxford, england)"] = "Hum Reprod",
  ["human reproduction and genetic ethics"] = "Hum Reprod Genet Ethics",
  ["human reproduction open"] = "Hum Reprod Open",
  ["human reproduction update"] = "Hum Reprod Update",
  ["human research report"] = "Hum Res Rep",
  ["human resource development quarterly"] = "Hum Resour Dev Q",
  ["human resource management"] = "Hum Resour Manage",
  ["human resources for health"] = "Hum Resour Health",
  ["human rights (chicago, ill.)"] = "Human Rights",
  ["human rights quarterly"] = "Hum Rights Q",
  ["human service organizations, management, leadership & governance"] = "Hum Serv Organ Manag Leadersh Gov",
  ["human studies"] = "Hum Stud",
  ["human systems management"] = "Hum. Syst. Manage.",
  ["human toxicology"] = "Hum Toxicol",
  ["human vaccines"] = "Hum Vaccin",
  ["human vaccines & immunotherapeutics"] = "Hum Vaccin Immunother",
  ["human-animal interaction bulletin"] = "Hum Anim Interact Bull",
  ["human-centric computing and information sciences"] = "Hum.-centric Comput. Inf. Sci.",
  ["human-centric intelligent systems"] = "Hum.-Centric Intell. Syst.",
  ["human-computer interaction"] = "Hum Comput Interact",
  ["human-computer interaction series"] = "Hum.-Comput. Interact. Ser.",
  ["human-intelligent systems integration"] = "Hum.-Intell. Syst. Integr.",
  ["human-wildlife interactions"] = "Hum.-Wildl. Interact.",
  ["humane health care"] = "Hum Health Care",
  ["humane health care international"] = "Hum Health Care Int",
  ["humane medicine"] = "Humane Med",
  ["humanismus und technik jahrbuch"] = "Humanismus Tech",
  ["humanistica lovaniensia"] = "HumLov",
  ["humanistische bildung"] = "HumBild",
  ["humanities & social sciences communications"] = "Humanit Soc Sci Commun",
  ["humanities report"] = "Humanit Rep",
  ["humanity & society"] = "Humanity Soc",
  ["humanâ€”animal interactions"] = "Hum.-Anim. Interact.",
  ["humboldt journal of social relations"] = "Humboldt J Soc Relat",
  ["humic substances in the environment"] = "Humic Subst. Environ.",
  ["hunan annals of mathematics"] = "Hunan Ann. Math.",
  ["hunan daxue xuebao"] = "Hunan Daxue Xuebao",
  ["hunan jiaoyu xueyuan"] = "Hunan Jiaoyu Xueyuan Xuebao (Ziran Kexue)",
  ["hunan nong ye da xue xue bao"] = "Journal of Hunan Agricultural University|Hunan Nong Ye Da Xue Xue Bao",
  ["hunan nong ye da xue xue bao = journal of hunan agricultural university"] = "Hunan Nong Ye Da Xue Xue Bao",
  ["hunan yi ke da xue xue bao (bulletin of hunan medical university)"] = "Hunan Yi Ke Da Xue Xue Bao",
  ["hunan yi ke da xue xue bao = hunan yike daxue xuebao = bulletin of hunan medical university"] = "Hunan Yi Ke Da Xue Xue Bao",
  ["hungarian journal of industrial chemistry"] = "Hung. J. Ind. Chem.",
  ["hungarian studies review : hsr"] = "Hung Stud Rev",
  ["hungarica acta physica"] = "Hung. Acta Phys.",
  ["hungarica acta physiologica"] = "Hung Acta Physiol",
  ["huntia; a yearbook of botanical and horticultural bibliography"] = "Huntia",
  ["husserl studies"] = "Husserl Stud",
  ["hutnicke listy"] = "Hutn. Listy",
  ["hvac&r research"] = "HVAC&R Res",
  ["hw-sa [reports]. u.s. atomic energy commission"] = "HW SA US At Energy Comm",
  ["hwahak konghak"] = "Hwahak Konghak",
  ["hwahak kwa kongop ui chinbo"] = "Hwahak Kwa Kongop Ui Chinbo",
  ["hybrid advances"] = "Hybrid Adv.",
  ["hybrid methods in engineering"] = "Hybrid Methods Eng.",
  ["hybridoma (2005)"] = "Hybridoma (Larchmt)",
  ["hybridoma and hybridomics"] = "Hybrid Hybridomics",
  ["hydrobiological journal"] = "Hydrobiol. J.",
  ["hydrobiological journal. american fisheries society"] = "Hydrobiol J",
  ["hydrocarbon engineering"] = "Hydrocarbon Eng.",
  ["hydrocarbon processing, international edition"] = "Hydrocarbon Process., Int. Ed.",
  ["hydroecologie appliquee"] = "Hydroecol. Appl.",
  ["hydrogeology journal"] = "Hydrol. J.",
  ["hydrological processes"] = "Hydrol. Process.",
  ["hydrological research letters"] = "Hydrol. Res. Lett.",
  ["hydrological sciences bulletin"] = "Hydrol. Sci. Bull.",
  ["hydrological sciences journal"] = "Hydrol. Sci. J.",
  ["hydrological sciences journal = journal des sciences hydrologiques"] = "Hydrol Sci J",
  ["hydrologie und wasserbewirtschaftung"] = "Hydrol. Wasserbewirtsch.",
  ["hydrology and earth system sciences"] = "Hydrol. Earth Syst. Sci.",
  ["hydrology and earth system sciences discussions"] = "Hydrol. Earth Syst. Sci. Discuss.",
  ["hydrology journal"] = "Hydrol. J.",
  ["hydrology research"] = "Hydrol. Res.",
  ["hygiea internationalis"] = "Hygiea Int",
  ["hygiene mentale"] = "Hyg. Ment.",
  ["hygienisk revy"] = "Hyg Revy",
  ["hyogo university journal"] = "Hyogo Univ. J.",
  ["hyomen gijutsu"] = "Hyomen Gijutsu",
  ["hyperboreus. studia classica"] = "Hyp",
  ["hyperboreus: studia classica"] = "Hyperboreus",
  ["hyperfine interactions"] = "Hyperfine Interact.",
  ["hypertension (dallas)"] = "Hypertension (Dallas)",
  ["hypertension (dallas, tex. : 1979)"] = "Hypertension",
  ["hypertension in pregnancy"] = "Hypertens Pregnancy",
  ["hypertension in pregnancy : official journal of the international society for the study of hypertension in pregnancy"] = "Hypertens Pregnancy",
  ["hypertension research"] = "Hypertens. Res.",
  ["hypertension research : official journal of the japanese society of hypertension"] = "Hypertens Res",
  ["hypothesis (university of toronto. dept. of medical biophysics)"] = "Hypothesis (Tor)",
  ["hypothesis : the newsletter of the research section of mla"] = "Hypothesis (Macon)",
  ["hypoxia (auckland, n.z.)"] = "Hypoxia (Auckl)",
  ["hystrix (online)"] = "Hystrix",
  ["i blu—pagine di scienza"] = "Blu Pagine Sci.",
  ["i chuan hsueh pao. acta genetica sinica"] = "I Chuan Hsueh Pao",
  ["i-manager's journal on nursing"] = "Imanagers J Nurs",
  ["i. c. r. s. medical reports"] = "ICRS Med Rep",
  ["i.d.a.a. communique"] = "IDAA Commun",
  ["i.e. (wheeling, ill.)"] = "I E",
  ["iads newsletter"] = "IADS Newsl",
  ["iaea bulletin"] = "IAEA Bull.",
  ["iaeng international journal of applied mathematics"] = "IAENG Int. J. Appl. Math.",
  ["iaeng international journal of computer science"] = "IAENG Int J Comput Sci",
  ["iaes international journal on artificial intelligence"] = "IAES Int. J. Artif. Intell.",
  ["iahs (international association of hydrological sciences) publication"] = "IAHS Publ.",
  ["iahs publication"] = "IAHS Publ.",
  ["ial news"] = "IAL News",
  ["iale (international association for landscape ecology) bulletin"] = "IALE Bull.",
  ["iapac monthly"] = "IAPAC Mon",
  ["iapqr transactions"] = "IAPQR Trans.",
  ["iarc monographs on the evaluation of carcinogenic risk of chemicals to man"] = "IARC Monogr. Eval. Carcinog. Risk Chem. Man",
  ["iarc monographs on the evaluation of carcinogenic risks to humans"] = "IARC Monogr Eval Carcinog Risks Hum",
  ["iarc monographs on the evaluation of carcinogenic risks to humans / world health organization, international agency for research on cancer"] = "IARC Monogr Eval Carcinog Risks Hum",
  ["iarc monographs on the evaluation of carcinogenic risks to humans. supplement"] = "IARC Monogr Eval Carcinog Risks Hum Suppl",
  ["iarc monographs on the evaluation of carcinogenic risks to humans. supplement / world health organization, international agency for research on cancer"] = "IARC Monogr Eval Carcinog Risks Hum Suppl",
  ["iarc monographs on the evaluation of the carcinogenic risk of chemicals to humans"] = "IARC Monogr Eval Carcinog Risk Chem Hum",
  ["iarc monographs on the evaluation of the carcinogenic risk of chemicals to humans. supplement"] = "IARC Monogr Eval Carcinog Risk Chem Hum Suppl",
  ["iarc monographs on the evaluation of the carcinogenic risk of chemicals to man"] = "IARC Monogr Eval Carcinog Risk Chem Man",
  ["iarc publications"] = "IARC Publ",
  ["iarc scientific publications"] = "IARC Sci Publ",
  ["ias/park city mathematical subseries"] = "IAS/Park City Math. Subser.",
  ["ias/park city mathematics series"] = "IAS/Park City Math. Ser.",
  ["ias/pcmi—the teacher program series"] = "IAS/PCMI Teach. Program Ser.",
  ["iassist quarterly"] = "IASSIST Q",
  ["iassist quarterly / international association for social science information service and technology"] = "IASSIST Q",
  ["iasted international conference on computational and systems biology : november 13-14, 2006, dallas, texas, usa (casb 2006)"] = "IASTED Int Conf Comput Syst Biol (2006)",
  ["iatrike epitheoresis enoplon dynameon. hellenic armed forces medical review"] = "Iatr Epitheor Enoplon Dyn",
  ["iatss research"] = "IATSS Res.",
  ["iau symposium"] = "IAU Symp.",
  ["iavi report : newsletter on international aids vaccine research"] = "IAVI Rep",
  ["iawa (international association of wood anatomists) journal"] = "IAWA J.",
  ["iawa journal"] = "IAWA J",
  ["iawa journal / international association of wood anatomists"] = "IAWA J",
  ["ib scientific journal of science"] = "IB Sci. J. Sci.",
  ["iberia archaeologica"] = "IA",
  ["iberia. revista della antigüedad"] = "Iberia",
  ["ibero-americana (stockholm, sweden : 1971)"] = "Ibero Americana (Stockholm)",
  ["iberoamericana (madrid, spain)"] = "Iberoamericana",
  ["iberus : revista de la sociedad española de malacología"] = "Iberus (Madr)",
  ["ibm journal of research and development"] = "IBM J. Res. Dev.",
  ["ibm systems journal"] = "IBM Syst. J.",
  ["ibm technical disclosure bulletin"] = "IBM Tech. Discl. Bull.",
  ["ibms bonekey"] = "IBMS Bonekey",
  ["ibro neuroscience reports"] = "IBRO Neurosci Rep",
  ["ibro reports"] = "IBRO Rep",
  ["ibs journal of science"] = "IBS J Sci",
  ["ibscientific journal of science"] = "IBSci. J. Sci.",
  ["icao journal"] = "ICAO J",
  ["icase/larc interdisciplinary series in science and engineering"] = "ICASE/LaRC Interdiscip. Ser. Sci. Eng.",
  ["iccad. ieee/acm international conference on computer-aided design"] = "ICCAD IEEE ACM Int Conf Comput Aided Des",
  ["iccm notices. notices of the international congress of chinese mathematicians"] = "ICCM Not.",
  ["iccm notices. notices of the international consortium of chinese mathematicians"] = "ICCM Not.",
  ["iccw journal"] = "ICCW J",
  ["iccw news bulletin"] = "ICCW News Bull",
  ["icd scientific and educational journal"] = "ICD Sci Educ J",
  ["icea news"] = "Icea News",
  ["iceland agricultural sciences"] = "Icel. Agric. Sci.",
  ["icelandic agricultural sciences"] = "Icel Agric Sci",
  ["ices journal of marine science"] = "ICES J. Mar. Sci.",
  ["ices journal of marine science : journal du conseil"] = "ICES J Mar Sci",
  ["icfai university journal of genetics and evolution"] = "ICFAI Univ J Genet Evol",
  ["ichthyological exploration of freshwaters"] = "Ichthyol Explor Freshw",
  ["ichthyological research"] = "Ichthyol Res",
  ["ichthyology & herpetology"] = "Ichthyol. Herpetol.",
  ["ici journal"] = "ICI J.",
  ["iciam 2019 sema simai springer series"] = "ICIAM 2019 SEMA SIMAI Springer Ser.",
  ["icis chemical business"] = "ICIS Chem. Bus.",
  ["icis chemical business americas"] = "ICIS Chem. Bus. Am.",
  ["icm series"] = "ICM Ser.",
  ["icme-13 monographs"] = "ICME-13 Monogr.",
  ["icmpc : proceedings. international conference on music perception and cognition"] = "ICMPC",
  ["icmr annals : annual reports of the international center for medical research"] = "ICMR Ann",
  ["icmr bulletin"] = "Icmr Bull",
  ["icmr'17 : proceedings of the 2017 acm international conference on multimedia retrieval : june 6-9, 2017, bucharest, romania. acm international conference on multimedia retrieval (2017 : bucharest, romania)"] = "ICMR 17 (2017)",
  ["icon (london, england)"] = "Icon (Lond)",
  ["iconographia mycologica"] = "Iconogr. Mycol.",
  ["icp selected papers"] = "ICP Sel. Pap.",
  ["icpd 94 : newsletter of the international conference on population and development"] = "ICPD 94",
  ["icsa book series in statistics"] = "ICSA Book Ser. Stat.",
  ["icssr newsletter. indian council of social science research"] = "ICSSR Newsl",
  ["icsu review"] = "Icsu Rev",
  ["icsu short reports"] = "ICSU Short Rep.",
  ["ict express"] = "ICT Express",
  ["ict systems security and privacy protection : 32nd ifip tc 11 international conference, sec 2017, rome, italy, may 29-31, 2017, proceedings. ifip tc11 international information security conference (32nd : 2017 : rome, italy)"] = "ICT Syst Secur Priv Prot (2017)",
  ["icu director"] = "ICU Dir",
  ["icu management & practice"] = "ICU Manag Pract",
  ["ida journal of desalination and water reuse"] = "IDA J. Desalin. Water Reuse",
  ["idaa communique"] = "IDAA Commun.",
  ["idaho code : containing the general laws of idaho annotated, compiled under the supervision of the idaho code commission. idaho"] = "Ida Code Ida",
  ["idaho code : containing the general laws of idaho annotated,. idaho"] = "Ida Code Ida",
  ["idaho law review"] = "Ida Law Rev",
  ["idea the journal of law and technology"] = "IDEA J Law Technol",
  ["ideas & production"] = "Ideas Prod",
  ["ideas in context"] = "Ideas Context",
  ["ideggyogyaszati szemle"] = "Ideggyogy Sz",
  ["ideggyógyászati szemle"] = "Ideggyogy Sz",
  ["idengaku zasshi"] = "Jpn J Genet",
  ["identities (yverdon, switzerland)"] = "Identities (Yverdon)",
  ["identity (mahwah, n.j.)"] = "Identity (Mahwah, N J)",
  ["idm materialien und studien"] = "IDM Materialien Studien",
  ["ido [reports]. u.s. atomic energy commission"] = "IDO Rep",
  ["ido reports"] = "IDO Rep.",
  ["idrugs : the investigational drugs journal"] = "IDrugs",
  ["ids bulletin"] = "IDS Bull",
  ["ids bulletin (university of sussex. institute of development studies : 1985)"] = "IDS Bull",
  ["iee electromagnetic waves series"] = "IEE Electromagnet. Waves Ser.",
  ["iee history of technology series"] = "IEE Hist. Tech. Ser.",
  ["iee journal on computers and digital techniques"] = "IEE J. Comput. Digital Tech.",
  ["iee journal on electric power applications"] = "IEE J. Electr. Power Appl.",
  ["iee journal on microwaves, antennas & propagation"] = "IEE J. Microwaves Antennas Propag.",
  ["iee proceedings. nanobiotechnology"] = "IEE Proc Nanobiotechnol",
  ["iee proceedings. systems biology"] = "IEE Proc Syst Biol",
  ["iee proceedings: electric power applications"] = "IEE Proc.: Electr. Power Appl.",
  ["iee proceedings: nanobiotechnology"] = "IEE Proc.: Nanobiotechnol.",
  ["iee proceedings: optolectronics"] = "IEE Proc.: Optoelectron.",
  ["iee proceedings: science, measurement & technology"] = "IEE Proc.: Sci. Meas. Technol.",
  ["iee-iere proceedings - india"] = "IEE-IERE Proc. - India",
  ["ieee access"] = "IEEE Access",
  ["ieee access : practical innovations, open solutions"] = "IEEE Access",
  ["ieee aerospace and electronic systems magazine"] = "IEEE Aerosp. Electron. Syst. Mag.",
  ["ieee aerospace and electronics systems magazine"] = "IEEE Aerosp. Electron. Syst. Mag.",
  ["ieee annals of the history of computing"] = "IEEE Ann. Hist. Comput.",
  ["ieee antennas & propagation magazine"] = "IEEE Antennas Propag Mag",
  ["ieee antennas and propagation magazine"] = "IEEE Antennas Propag. Mag.",
  ["ieee antennas and wireless propagation letters"] = "IEEE Antennas Wirel. Propag. Lett.",
  ["ieee applied imagery pattern recognition workshop : [proceedings]. ieee applied imagery pattern recognition workshop"] = "IEEE Appl Imag Pattern Recognit Workshop",
  ["ieee assp magazine"] = "IEEE ASSP Mag.",
  ["ieee assp magazine (1984-1990)"] = "IEEE ASSP Mag.",
  ["ieee biomedical circuits and systems conference : healthcare technology : [proceedings]. ieee biomedical circuits and systems conference"] = "IEEE Biomed Circuits Syst Conf",
  ["ieee biometrics compendium"] = "IEEE Biometrics Compend.",
  ["ieee canadian journal of electrical and computer engineering"] = "IEEE Can. J. Electr. Comput. Eng.",
  ["ieee china summit & international conference on signal and information processing. ieee china summit & international conference on signal and information processing"] = "IEEE China Summit Int Conf Signal Inf Process",
  ["ieee circuits and devices magazine"] = "IEEE Circuits Devices Mag.",
  ["ieee circuits and devices magazine (1985-present)"] = "IEEE Circuits Devices Mag.",
  ["ieee circuits and systems magazine"] = "IEEE Circuits Syst. Mag.",
  ["ieee circuits and systems magazine (1979-1984)"] = "IEEE Circuits Syst. Mag.",
  ["ieee cloud computing"] = "IEEE Cloud Comput.",
  ["ieee cogsima : 2020 ieee international conference on cognitive and computational aspects of situation management (cogsima) : proceedings : virtual conference, 24-28 august 2020. ieee conference on cognitive and computational aspects of situation management (2020 : online)"] = "IEEE CogSIMA (2020)",
  ["ieee communication surveys and tutorials"] = "IEEE Commun. Surv. Tutorials",
  ["ieee communications letters"] = "IEEE Commun. Lett.",
  ["ieee communications letters : a publication of the ieee communications society"] = "IEEE Commun Lett",
  ["ieee communications magazine"] = "IEEE Commun. Mag.",
  ["ieee communications magazine (1979-present)"] = "IEEE Commun. Mag.",
  ["ieee communications magazine. ieee communications society"] = "IEEE Commun Mag",
  ["ieee communications society magazine"] = "IEEE Commun. Soc. Mag.",
  ["ieee communications society magazine (through 1978)"] = "IEEE Commun. Soc. Mag.",
  ["ieee communications standards magazine"] = "IEEE Commun. Stand. Mag.",
  ["ieee communications surveys and tutorials"] = "IEEE Commun. Surv. Tutorials",
  ["ieee computation in science and engineering magazine"] = "IEEE Comput. Sci. Eng. Mag.",
  ["ieee computational intelligence magazine"] = "IEEE Comput. Intell. Mag.",
  ["ieee computational science and engineering magazine"] = "IEEE Comput. Sci. Eng. Mag.",
  ["ieee computer"] = "IEEE Computer",
  ["ieee computer applications in power"] = "IEEE Comput. Appl. Power",
  ["ieee computer architecture letters"] = "IEEE Comput. Archit. Lett.",
  ["ieee computer graphics and applications"] = "IEEE Comput Graph Appl",
  ["ieee computing in science and engineering"] = "IEEE Comput. Sci. Eng.",
  ["ieee concurrency"] = "IEEE Concurrency",
  ["ieee conference on games 2019 : london, united kingdom, 20-23 august 2019"] = "IEEE Conf Games 2019 (2019)",
  ["ieee consumer electronics magazine"] = "IEEE Consum. Electron. Mag.",
  ["ieee control systems"] = "IEEE Control Syst.",
  ["ieee control systems letters"] = "IEEE Control Syst. Lett.",
  ["ieee control systems magazine"] = "IEEE Control Syst. Mag.",
  ["ieee control systems magazine (new york, n.y. : 1985)"] = "IEEE Control Syst Mag N Y",
  ["ieee data descriptions"] = "IEEE Data Descr.",
  ["ieee design & test"] = "IEEE Des. Test",
  ["ieee design and test of computers"] = "IEEE Des. Test Comput.",
  ["ieee distributed systems online"] = "IEEE Distrib. Syst. Online",
  ["ieee electrical insulation magazine"] = "IEEE Electr. Insul. Mag.",
  ["ieee electrification magazine"] = "IEEE Electrif. Mag.",
  ["ieee electromagnetic compatibility magazine"] = "IEEE Electromagn Compat Mag",
  ["ieee electron device letters"] = "IEEE Electron Device Lett.",
  ["ieee electron device letters : a publication of the ieee electron devices society"] = "IEEE Electron Device Lett",
  ["ieee electron devices magazine"] = "IEEE Electron Devices Mag.",
  ["ieee electrotechnology review"] = "IEEE ElectroTechnol. Rev.",
  ["ieee embedded systems letters"] = "IEEE Embedded Sys. Lett.",
  ["ieee engineering in medicine and biology magazine"] = "IEEE Eng. Med. Biol. Mag.",
  ["ieee engineering in medicine and biology magazine : the quarterly magazine of the engineering in medicine & biology society"] = "IEEE Eng Med Biol Mag",
  ["ieee engineering management review"] = "IEEE Eng. Manage. Rev.",
  ["ieee expert"] = "IEEE Expert",
  ["ieee expert (through 1997)"] = "IEEE Expert",
  ["ieee geoscience and remote sensing letters"] = "IEEE Geosci. Remote Sens. Lett.",
  ["ieee geoscience and remote sensing letters : a publication of the ieee geoscience and remote sensing society"] = "IEEE Geosci Remote Sens Lett",
  ["ieee geoscience and remote sensing magazine"] = "IEEE Geosci. Remote Sens. Mag.",
  ["ieee haptics symposium : [proceedings]. ieee haptics symposium"] = "IEEE Haptics Symp",
  ["ieee industrial electronics magazine"] = "IEEE Ind. Electron. Mag.",
  ["ieee industry applications magazine"] = "IEEE Ind. Appl. Mag.",
  ["ieee instrumentation & measurement magazine"] = "IEEE Instrum Meas Mag",
  ["ieee instrumentation and measurement magazine"] = "IEEE Instrum. Meas. Mag.",
  ["ieee intelligent systems"] = "IEEE Intell. Syst.",
  ["ieee intelligent transportation systems magazine"] = "IEEE Intell. Transp. Syst. Mag.",
  ["ieee international conference on automatic face & gesture recognition and workshops"] = "IEEE Int Conf Autom Face Gesture Recognit Workshops",
  ["ieee international conference on automation science and engineering (case) : [proceedings]. ieee conference on automation science and engineering"] = "IEEE Int Conf Automation Sci Eng (CASE)",
  ["ieee international conference on bioinformatics and biomedicine workshops. ieee international conference on bioinformatics and biomedicine"] = "IEEE Int Conf Bioinform Biomed Workshops",
  ["ieee international conference on control and automation : [proceedings]. international conference on control and automation"] = "IEEE Int Conf Control Autom",
  ["ieee international conference on development and learning : [proceedings]. ieee international conference on development and learning"] = "IEEE Int Conf Dev Learn",
  ["ieee international conference on development and learning and epigenetic robotics : [proceedings]. ieee international conference on development and learning and epigenetic robotics"] = "IEEE Int Conf Dev Learn Epigenetic Robot",
  ["ieee international conference on electro information technology : [proceedings]. ieee international conference on electro information technology"] = "IEEE Int Conf Electro Inf Technol",
  ["ieee international conference on healthcare informatics. ieee international conference on healthcare informatics"] = "IEEE Int Conf Healthc Inform",
  ["ieee international conference on rfid. ieee international conference on rfid"] = "IEEE Int Conf RFID",
  ["ieee international conference on robotics and automation : icra : [proceedings]. ieee international conference on robotics and automation"] = "IEEE Int Conf Robot Autom",
  ["ieee international conference on software quality, reliability and security : proceedings. ieee international conference on software quality, reliability and security"] = "IEEE Int Conf Softw Qual Reliab Secur",
  ["ieee international conference on systems biology : [proceedings]. ieee international conference on systems biology"] = "IEEE Int Conf Systems Biol",
  ["ieee international geoscience and remote sensing symposium proceedings. international geoscience and remote sensing symposium"] = "IEEE Int Geosci Remote Sens Symp Proc",
  ["ieee international symposium on circuits and systems proceedings. ieee international symposium on circuits and systems"] = "IEEE Int Symp Circuits Syst Proc",
  ["ieee international symposium on parallel & distributed processing, workshops and phd forum : [proceedings]. ieee international symposium on parallel & distributed processing, workshops and phd forum"] = "IEEE Int Symp Parallel Distrib Process Workshops Phd Forum",
  ["ieee international ultrasonics symposium : [proceedings]. ieee international ultrasonics symposium"] = "IEEE Int Ultrason Symp",
  ["ieee international vacuum electronics conference. international vacuum electronics conference"] = "IEEE Int Vac Electron Conf",
  ["ieee international workshop on biomedical circuits & systems"] = "IEEE Int Workshop Biomed Circuits Syst",
  ["ieee international workshop on genomic signal processing and statistics : [proceedings]. ieee international workshop on genomic signal processing and statistics"] = "IEEE Int Workshop Genomic Signal Process Stat",
  ["ieee international workshop on machine learning for signal processing : [proceedings]. ieee international workshop on machine learning for signal processing"] = "IEEE Int Workshop Mach Learn Signal Process",
  ["ieee internet computing"] = "IEEE Internet Comput.",
  ["ieee internet of things journal"] = "IEEE Internet Things J.",
  ["ieee internet of things magazine"] = "IEEE Internet Things Mag.",
  ["ieee it professional"] = "IEEE IT Prof.",
  ["ieee journal of biomedical and health informatics"] = "IEEE J Biomed Health Inform",
  ["ieee journal of electromagnetics"] = "IEEE J. Electromagn.",
  ["ieee journal of electromagnetics, rf and microwaves in medicine and biology"] = "IEEE J Electromagn RF Microw Med Biol",
  ["ieee journal of electron devices society"] = "IEEE J. Electron Devices Soc.",
  ["ieee journal of emerging and selected topics in industrial electronics"] = "IEEE J. Emerging Sel. Top. Ind. Electron.",
  ["ieee journal of emerging and selected topics in power electronics"] = "IEEE J Emerg Sel Top Power Electron",
  ["ieee journal of microelectromechanical systems"] = "J. Microelectromech. Syst.",
  ["ieee journal of microwaves"] = "IEEE J. Microwaves",
  ["ieee journal of oceanic engineering"] = "IEEE J. Oceanic Eng.",
  ["ieee journal of optical communications and networking"] = "IEEE J. Opt. Commun. Netw.",
  ["ieee journal of photovoltaics"] = "IEEE J Photovolt",
  ["ieee journal of product safety engineering"] = "IEEE J. Product Safety Eng.",
  ["ieee journal of quantum electronics"] = "IEEE J. Quantum Electron.",
  ["ieee journal of radio frequency identification (rfid)"] = "IEEE J. Radio Freq. Identif. (RFID)",
  ["ieee journal of robotics and automation"] = "IEEE J. Robot. Autom.",
  ["ieee journal of selected topics in applied earth observations and remote sensing"] = "IEEE J Sel Top Appl Earth Obs Remote Sens",
  ["ieee journal of selected topics in quantum electronics"] = "IEEE J. Sel. Top. Quantum Electron.",
  ["ieee journal of selected topics in quantum electronics : a publication of the ieee lasers and electro-optics society"] = "IEEE J Sel Top Quantum Electron",
  ["ieee journal of selected topics in signal processing"] = "IEEE J Sel Top Signal Process",
  ["ieee journal of solid-state circuits"] = "IEEE J. Solid-State Circuits",
  ["ieee journal of the electron devices society"] = "IEEE J. Electron Devices Soc.",
  ["ieee journal of translational engineering in health and medicine"] = "IEEE J Transl Eng Health Med",
  ["ieee journal on electronic circuits and systems"] = "IEE J. Electron. Circuits Syst.",
  ["ieee journal on emerging and selected topics in circuits and systems"] = "IEEE J Emerg Sel Top Circuits Syst",
  ["ieee journal on exploratory solid-state computational devices and circuits"] = "IEEE J. Explor. Solid-State Comput. Devices Circuits",
  ["ieee journal on indoor and seamless positioning and navigation"] = "IEEE J. Indoor Seamless Positioning Navig.",
  ["ieee journal on miniaturization for air and space systems"] = "IEEE J. Miniaturization Air Space Syst.",
  ["ieee journal on multiscale and multiphysics computational techniques"] = "IEEE J Multiscale Multiphys Comput Tech",
  ["ieee journal on selected areas in communications"] = "IEEE J. Sel. Areas Commun.",
  ["ieee journal on selected areas in communications : a publication of the ieee communications society"] = "IEEE J Sel Areas Commun",
  ["ieee journal on selected areas in information theory"] = "IEEE J. Sel. Areas Inf. Theory",
  ["ieee journal on selected areas in sensors"] = "IEEE J. Sel. Areas Sens.",
  ["ieee journal on selected topics in quantum electronics"] = "IEEE J. Sel. Top. Quantum Electron.",
  ["ieee journal on selected topics in signal processing"] = "IEEE J. Sel. Top. Sign. Proces.",
  ["ieee journal on technology in computer aided design"] = "IEEE J. Technol. Comput. Aided Design",
  ["ieee latin america transactions"] = "IEEE Lat. Am. Trans.",
  ["ieee learning technology"] = "IEEE Learn. Technol.",
  ["ieee letters of the computer society"] = "IEEE Lett Comput Soc",
  ["ieee letters on electromagnetic compatibility practice and applications"] = "IEEE Lett. Electromagn. Compat. Pract. Appl.",
  ["ieee life sciences conference. life sciences conference"] = "IEEE Life Sci Conf",
  ["ieee life sciences letters"] = "IEEE Life Sci. Lett.",
  ["ieee magnetics letters"] = "IEEE Magn. Lett.",
  ["ieee micro"] = "IEEE Micro",
  ["ieee microwave and guided wave letters"] = "IEEE Microw. Guided Wave Lett.",
  ["ieee microwave and wireless components letters"] = "IEEE Microwave Wireless Compon. Lett.",
  ["ieee microwave and wireless components letters : a publication of the ieee microwave theory and techniques society"] = "IEEE Microw Wirel Compon Lett",
  ["ieee microwave magazine"] = "IEEE Microwave Mag.",
  ["ieee mtt-s international microwave symposium digest. ieee mtt-s international microwave symposium"] = "IEEE MTTS Int Microw Symp",
  ["ieee multimedia"] = "IEEE Multimedia",
  ["ieee nanotechnology express"] = "IEEE Nanotechnol. Express",
  ["ieee nanotechnology magazine"] = "IEEE Nanatechnol. Mag.",
  ["ieee network"] = "IEEE Netw",
  ["ieee networking letters"] = "IEEE Networking Lett.",
  ["ieee npss real time conference : conference record. conference on real-time computer applications in nuclear, particle, and plasma physics"] = "IEEE NPSS Real Time Conf",
  ["ieee nuclear science symposium conference record. nuclear science symposium"] = "IEEE Nucl Sci Symp Conf Rec (1997)",
  ["ieee open access journal of power and energy"] = "IEEE Open Access J. Power Energy",
  ["ieee open journal of antennas and propagation"] = "IEEE Open J. Antennas Propag.",
  ["ieee open journal of circuits and systems"] = "IEEE Open J. Circuits Syst.",
  ["ieee open journal of control systems"] = "IEEE Open J. Control Syst.",
  ["ieee open journal of engineering in medicine and biology"] = "IEEE Open J Eng Med Biol",
  ["ieee open journal of industry applications"] = "IEEE Open J. Ind. Appl.",
  ["ieee open journal of instrumentation and measurement"] = "IEEE Open J. Instrum. Meas.",
  ["ieee open journal of intelligent transportation systems"] = "IEEE Open J. Intell. Transp. Syst.",
  ["ieee open journal of nanotechnology"] = "IEEE Open J. Nanotechnol.",
  ["ieee open journal of power electronics"] = "IEEE Open J. Power Electron.",
  ["ieee open journal of signal processing"] = "IEEE Open J. Signal Process.",
  ["ieee open journal of solid-state circuits"] = "IEEE Open J. Solid-State Circuits",
  ["ieee open journal of systems engineering"] = "IEEE Open J. Syst. Eng.",
  ["ieee open journal of the communications society"] = "IEEE Open J. Commun. Soc.",
  ["ieee open journal of the computer society"] = "IEEE Open J. Comput. Soc.",
  ["ieee open journal of the industrial electronics society"] = "IEEE Open J. Ind. Electron. Soc.",
  ["ieee open journal of the microwave theory and techniques society"] = "IEEE Open J. Microwave Theory Tech. Soc.",
  ["ieee open journal of vehicular technology"] = "IEEE Open J. Veh. Technol.",
  ["ieee open journal on immersive displays"] = "IEEE Open J. Immersive Disp.",
  ["ieee open journal on ultrasonics, ferroelectrics, and frequency control"] = "IEEE Open J. Ultrason. Ferroelectr. Freq. Control",
  ["ieee pacific visualization symposium : [proceedings]. ieee pacific visualisation symposium"] = "IEEE Pac Vis Symp",
  ["ieee personal communications"] = "IEEE Pers. Commun.",
  ["ieee personal communications magazine"] = "IEEE Personal Commun. Mag.",
  ["ieee pervasive computing"] = "IEEE Pervasive Comput.",
  ["ieee pervasive computing / ieee computer society [and] ieee communications society"] = "IEEE Pervasive Comput",
  ["ieee photonics conference : [proceedings]. ieee photonics conference"] = "IEEE Photonics Conf",
  ["ieee photonics journal"] = "IEEE Photonics J.",
  ["ieee photonics technology letters"] = "IEEE Photonics Technol. Lett.",
  ["ieee photonics technology letters : a publication of the ieee laser and electro-optics society"] = "IEEE Photonics Technol Lett",
  ["ieee photonicsglobal[at]singapore"] = "IEEE PhotonicsGlob Singap",
  ["ieee potentials"] = "IEEE Potentials",
  ["ieee power and energy magazine"] = "IEEE Power Energy Mag.",
  ["ieee power and energy technology systems journal"] = "IEEE Power Energy Technol. Syst. J.",
  ["ieee power electronics letters"] = "IEEE Power Electron. Lett.",
  ["ieee power electronics magazine"] = "IEEE Power Electron. Mag.",
  ["ieee power engineering review"] = "IEEE Power Eng. Rev.",
  ["ieee press series on computational intelligence"] = "IEEE Press Ser. Comput. Intell.",
  ["ieee press series on power engineering"] = "IEEE Press Ser. Power Eng.",
  ["ieee press series on rf and microwave technology"] = "IEEE Press Ser. RF Microw. Technol.",
  ["ieee press understanding science & technology series"] = "IEEE Press Underst. Sci. Technol. Ser.",
  ["ieee proceedings i - solid-state and electron devices"] = "IEEE Proc. Solid State Electron Devices",
  ["ieee pulse"] = "IEEE Pulse",
  ["ieee reliability magazine"] = "IEEE Reliab. Mag.",
  ["ieee reviews in biomedical engineering"] = "IEEE Rev Biomed Eng",
  ["ieee revista iberoamericana de technolog’ias del aprendizaje"] = "IEEE Revista Iberoamericana de Technolog’ias del Aprendizaje",
  ["ieee revista iberoamericana de tecnologias del aprendizaje"] = "IEEE Rev. Iberoam. Tecnol. Aprendizaje",
  ["ieee rfic journal"] = "IEEE RFIC J.",
  ["ieee rfic virtual journal"] = "IEEE RFIC Virtual J.",
  ["ieee rfid journal"] = "IEEE RFID J.",
  ["ieee rfid virtual journal"] = "IEEE RFID Virtual J.",
  ["ieee robotics & automation magazine"] = "IEEE Robot Autom Mag",
  ["ieee robotics and automation letters"] = "IEEE Robot Autom Lett",
  ["ieee robotics and automation magazine"] = "IEEE Rob. Autom Mag.",
  ["ieee robotics and automation practice"] = "IEEE Rob. Autom. Pract.",
  ["ieee security & privacy"] = "IEEE Secur Priv",
  ["ieee security and privacy"] = "IEEE Secur. Privacy",
  ["ieee sensors journal"] = "IEEE Sens J",
  ["ieee sensors letters"] = "IEEE Sens Lett",
  ["ieee sensors reviews"] = "IEEE Sens. Rev.",
  ["ieee signal processing letters"] = "IEEE Signal Process Lett.",
  ["ieee signal processing magazine"] = "IEEE Signal Process Mag.",
  ["ieee signal processing magazine (1991-present)"] = "IEEE Signal Process. Mag.",
  ["ieee software"] = "IEEE Softw.",
  ["ieee solid-state circuits letters"] = "IEEE Solid State Circuits Lett",
  ["ieee solid-state circuits magazine"] = "IEEE Solid-State Circuits Mag.",
  ["ieee spectrum"] = "IEEE Spectr.",
  ["ieee sustainable computing"] = "IEEE Sustain. Comput.",
  ["ieee symposium on computational intelligence and data mining. ieee symposium on computational intelligence and data mining"] = "IEEE Symp Comput Intell Data Min",
  ["ieee symposium on computational intelligence in bioinformatics and computational biology proceedings. ieee symposium on computational intelligence in bioinformatics and computational biology"] = "IEEE Symp Comput Intell Bioinforma Comput Biol Proc",
  ["ieee systems journal"] = "IEEE Syst. J.",
  ["ieee systems, man, and cybernetics letters"] = "IEEE Syst. Man Cybern. Lett.",
  ["ieee systems, man, and cybernetics magazine"] = "IEEE Syst Man Cybern Mag",
  ["ieee technology and society magazine"] = "IEEE Technol. Soc. Mag.",
  ["ieee transactions on acoustics, speech, and signal processing"] = "IEEE Trans Acoust",
  ["ieee transactions on advanced packaging"] = "IEEE Trans. Adv. Packag.",
  ["ieee transactions on aeronautical and navigational electronics"] = "IEEE Trans. Aeronaut. Navig. Electron.",
  ["ieee transactions on aerospace"] = "IEEE Trans. Aerosp.",
  ["ieee transactions on aerospace and electronic systems"] = "IEEE Trans. Aerosp. Electron. Syst.",
  ["ieee transactions on aerospace and navigation electronics"] = "IEEE Trans. Aerosp. Navig. Electron.",
  ["ieee transactions on aerospace and navigational electronics"] = "IEEE Trans. Aerosp. Navig. Electron.",
  ["ieee transactions on affective computing"] = "IEEE Trans. Affective Comput.",
  ["ieee transactions on agrifood electronics"] = "IEEE Trans. Agrifood Electron.",
  ["ieee transactions on airborne electronics"] = "IEEE Trans. Airborne Electron.",
  ["ieee transactions on antennas and propagation"] = "IEEE Trans. Antennas Propag.",
  ["ieee transactions on applications and industry"] = "IEEE Trans. Appl. Indus.",
  ["ieee transactions on applied superconductivity"] = "IEEE Trans. Appl. Supercond.",
  ["ieee transactions on applied superconductivity : a publication of the ieee superconductivity committee"] = "IEEE Trans Appl Supercond",
  ["ieee transactions on artificial intelligence"] = "IEEE Trans. Artif. Intell.",
  ["ieee transactions on audio"] = "IEEE Trans. Audio",
  ["ieee transactions on audio and electroacoustics"] = "IEEE Trans. Audio Electroacoust.",
  ["ieee transactions on audio, speech and language processing"] = "IEEE Trans. Audio Speech Lang. Process.",
  ["ieee transactions on audio, speech, and language processing"] = "IEEE Trans Audio Speech Lang Process",
  ["ieee transactions on automatic control"] = "IEEE Trans. Autom. Control",
  ["ieee transactions on automation science and engineering"] = "IEEE Trans. Autom. Sci. Eng.",
  ["ieee transactions on automation science and engineering : a publication of the ieee robotics and automation society"] = "IEEE Trans Autom Sci Eng",
  ["ieee transactions on autonomous mental development"] = "IEEE Trans. Auton. Ment. Dev.",
  ["ieee transactions on big data"] = "IEEE Trans. Big Data",
  ["ieee transactions on bio-medical electronics"] = "IEEE Trans. Bio-med. Electron.",
  ["ieee transactions on bio-medical engineering"] = "IEEE Trans Biomed Eng",
  ["ieee transactions on biomedical circuits and systems"] = "IEEE Trans Biomed Circuits Syst",
  ["ieee transactions on biomedical engineering"] = "IEEE Trans. Biomed. Eng.",
  ["ieee transactions on biometrics, behavior, and identity science"] = "IEEE Trans Biom Behav Identity Sci",
  ["ieee transactions on broadcast and television receivers"] = "IEEE Trans. Broadcast Telev. Receivers",
  ["ieee transactions on broadcasting"] = "IEEE Trans. Broadcast.",
  ["ieee transactions on cable television"] = "IEEE Trans. Cable Telev.",
  ["ieee transactions on circuit theory"] = "IEEE Trans. Circuit Theory",
  ["ieee transactions on circuits and systems"] = "IEEE Trans Circuits Syst",
  ["ieee transactions on circuits and systems for artificial intelligence"] = "IEEE Trans. Circuits Syst. Artif. Intell.",
  ["ieee transactions on circuits and systems for video technology"] = "IEEE Trans. Circuits Syst. Video Technol.",
  ["ieee transactions on circuits and systems for video technology : a publication of the circuits and systems society"] = "IEEE Trans Circuits Syst Video Technol",
  ["ieee transactions on circuits and systems i: fundamental theory and applications"] = "IEEE Trans. Circuits Syst. I",
  ["ieee transactions on circuits and systems i: regular papers"] = "IEEE Trans. Circuits Syst. Regul. Pap.",
  ["ieee transactions on circuits and systems ii: analog and digital signal processing"] = "IEEE Trans. Circuits Syst. II",
  ["ieee transactions on circuits and systems ii: express briefs"] = "IEEE Trans. Circuits Syst. Express Briefs",
  ["ieee transactions on circuits and systems. i, fundamental theory and applications : a publication of the ieee circuits and systems society"] = "IEEE Trans Circuits Syst I Fundam Theory Appl",
  ["ieee transactions on circuits and systems. i, regular papers : a publication of the ieee circuits and systems society"] = "IEEE Trans Circuits Syst I Regul Pap",
  ["ieee transactions on circuits and systems. i. regular papers"] = "IEEE Trans. Circuits Syst. I. Regul. Pap.",
  ["ieee transactions on circuits and systems. ii, express briefs : a publication of the ieee circuits and systems society"] = "IEEE Trans Circuits Syst II Express Briefs",
  ["ieee transactions on circuits and systems—part i: fundamental theory and applications"] = "IEEE Trans. Circuits Syst. I",
  ["ieee transactions on circuits and systems—part ii: analog and digital signal processing"] = "IEEE Trans. Circuits Syst. II",
  ["ieee transactions on cloud computing"] = "IEEE Trans. Cloud Comput.",
  ["ieee transactions on cognitive and developmental systems"] = "IEEE Trans Cogn Dev Syst",
  ["ieee transactions on cognitive communications and networking"] = "IEEE Trans Cogn Commun Netw",
  ["ieee transactions on communication technology"] = "IEEE Trans. Commun. Technol.",
  ["ieee transactions on communications"] = "IEEE Trans. Commun.",
  ["ieee transactions on communications systems"] = "IEEE Trans. Commun. Syst.",
  ["ieee transactions on communications. ieee communications society"] = "IEEE Trans Commun",
  ["ieee transactions on component parts"] = "IEEE Trans. Compon. Parts",
  ["ieee transactions on components and packaging technologies"] = "IEEE Trans. Compon. Packag. Technol.",
  ["ieee transactions on components and packaging technology"] = "IEEE Trans. Compon. Packag. Technol.",
  ["ieee transactions on components, hybrids and manufacturing technology"] = "IEEE Trans. Compon., Hybrids, Manuf. Technol.",
  ["ieee transactions on components, packaging and manufacturing technology"] = "IEEE Trans. Compon. Packag. Manuf. Technol.",
  ["ieee transactions on components, packaging and manufacturing technology a"] = "IEEE Trans. Compon., Packag., Manuf. Technol. A",
  ["ieee transactions on components, packaging and manufacturing technology b: advanced packaging"] = "IEEE Trans. Compon., Packag., Manuf. Technol. B",
  ["ieee transactions on components, packaging and manufacturing technology c: manufacturing"] = "IEEE Trans. Compon., Packag., Manuf. Technol. C",
  ["ieee transactions on components, packaging and manufacturing technology—part a"] = "IEEE Trans. Comp., Packag., Manufact. Technol. A",
  ["ieee transactions on components, packaging and manufacturing technology—part b: advanced packaging"] = "IEEE Trans. Comp., Packag., Manufact. Technol. B",
  ["ieee transactions on components, packaging and manufacturing technology—part c: manufacturing"] = "IEEE Trans. Comp., Packag., Manufact. Technol. C",
  ["ieee transactions on components, packaging, and manufacturing technology"] = "IEEE Trans Compon Packaging Manuf Technol",
  ["ieee transactions on components, packaging, and manufacturing technology, part a:"] = "IEEE Trans. Compon. Packag. Manuf. Technol. Part A:",
  ["ieee transactions on components, packaging, and manufacturing technology, part b:"] = "IEEE Trans. Compon. Packag. Manuf. Technol. Part B:",
  ["ieee transactions on components, packaging, and manufacturing technology, part c:"] = "IEEE Trans. Compon. Packag. Manuf. Technol. Part C:",
  ["ieee transactions on computational imaging"] = "IEEE Trans. Comput. Imaging",
  ["ieee transactions on computational intelligence and ai in games"] = "IEEE Trans Comput Intell AI Games",
  ["ieee transactions on computational intelligence in ai and games"] = "IEEE Trans. Comput. Intell. AI Games",
  ["ieee transactions on computational social systems"] = "IEEE Trans. Comput. Social Syst.",
  ["ieee transactions on computer-aided design of integrated circuits and systems"] = "IEEE Trans. Comput. Aided Des. Integr. Circuits Syst.",
  ["ieee transactions on computers"] = "IEEE Trans. Comput.",
  ["ieee transactions on computers. institute of electrical and electronics engineers"] = "IEEE Trans Comput",
  ["ieee transactions on consumer electronics"] = "IEEE Trans. Consum. Electron.",
  ["ieee transactions on control of network systems"] = "IEEE Trans. Control Netw. Syst.",
  ["ieee transactions on control systems technology"] = "IEEE Trans. Control Syst. Technol.",
  ["ieee transactions on control systems technology : a publication of the ieee control systems society"] = "IEEE Trans Control Syst Technol",
  ["ieee transactions on cybernetics"] = "IEEE Trans Cybern",
  ["ieee transactions on dependable and secure computing"] = "IEEE Trans. Dependable Secure Comput.",
  ["ieee transactions on device and materials reliability"] = "IEEE Trans. Device Mater. Reliab.",
  ["ieee transactions on device and materials reliability : a publication of the ieee electron devices society and the ieee reliability society"] = "IEEE Trans Device Mater Reliab",
  ["ieee transactions on dielectrics and electrical insulation"] = "IEEE Trans. Dielectr. Electr. Insul.",
  ["ieee transactions on dielectrics and electrical insulation : a publication of the ieee dielectrics and electrical insulation society"] = "IEEE Trans Dielectr Electr Insul",
  ["ieee transactions on education"] = "IEEE Trans. Educ.",
  ["ieee transactions on electrical insulation"] = "IEEE Trans. Electr. Insul.",
  ["ieee transactions on electromagnetic compatibility"] = "IEEE Trans. Electromagn. Compat.",
  ["ieee transactions on electron devices"] = "IEEE Trans. Electron Devices",
  ["ieee transactions on electronic computers"] = "IEEE Trans. Electron. Comput.",
  ["ieee transactions on electronics packaging manufacturing"] = "IEEE Trans. Electron. Packag. Manuf.",
  ["ieee transactions on emerging and selected topics in circuits and systems"] = "IEEE Trans. Emerg. Sel. Topics Circuits Syst.",
  ["ieee transactions on emerging and selected topics in power electronics"] = "IEEE Trans. Emerg. Sel. Topics Power Electron.",
  ["ieee transactions on emerging topics in computational intelligence"] = "IEEE Trans Emerg Top Comput Intell",
  ["ieee transactions on emerging topics in computing"] = "IEEE Trans Emerg Top Comput",
  ["ieee transactions on energy conversion"] = "IEEE Trans. Energy Convers.",
  ["ieee transactions on energy markets, policy and regulation"] = "IEEE Trans. Energy Mark. Policy Regul.",
  ["ieee transactions on engineering management"] = "IEEE Trans. Eng. Manage.",
  ["ieee transactions on engineering writing and speech"] = "IEEE Trans. Eng. Writing Speech",
  ["ieee transactions on evolutionary computation"] = "IEEE Trans. Evol. Comput.",
  ["ieee transactions on evolutionary computation : a publication of the ieee neural networks council"] = "IEEE Trans Evol Comput",
  ["ieee transactions on field robotics"] = "IEEE Trans. Field Rob.",
  ["ieee transactions on fuzzy systems"] = "IEEE Trans. Fuzzy Syst.",
  ["ieee transactions on fuzzy systems : a publication of the ieee neural networks council"] = "IEEE Trans Fuzzy Syst",
  ["ieee transactions on games"] = "IEEE Trans Games",
  ["ieee transactions on geoscience and remote sensing"] = "IEEE Trans. Geosci. Remote Sens.",
  ["ieee transactions on geoscience and remote sensing : a publication of the ieee geoscience and remote sensing society"] = "IEEE Trans Geosci Remote Sens",
  ["ieee transactions on geoscience electronics"] = "IEEE Trans. Geosci. Electron.",
  ["ieee transactions on green communications and networking"] = "IEEE Trans. Green Commun. Networking",
  ["ieee transactions on haptics"] = "IEEE Trans Haptics",
  ["ieee transactions on human factors in electronics"] = "IEEE Trans. Hum. Factors Electron.",
  ["ieee transactions on human-machine systems"] = "IEEE Trans Hum Mach Syst",
  ["ieee transactions on image processing"] = "IEEE Trans. Image Process.",
  ["ieee transactions on image processing : a publication of the ieee signal processing society"] = "IEEE Trans Image Process",
  ["ieee transactions on industrial cyber-physical systems"] = "IEEE Trans. Ind. Cyber-Phys. Syst.",
  ["ieee transactions on industrial electronics"] = "IEEE Trans. Ind. Electron.",
  ["ieee transactions on industrial electronics (1982)"] = "IEEE Trans Ind Electron",
  ["ieee transactions on industrial electronics and control instrumentation"] = "IEEE Trans. Ind. Electron. Contr. Instrum.",
  ["ieee transactions on industrial informatics"] = "IEEE Trans. Ind. Inf.",
  ["ieee transactions on industry and general applications"] = "IEEE Trans. Ind. Gen. Appl.",
  ["ieee transactions on industry applications"] = "IEEE Trans. Ind. Appl.",
  ["ieee transactions on information forensics and security"] = "IEEE Trans. Inf. Forensics Secur.",
  ["ieee transactions on information technology in biomedicine"] = "IEEE Trans. Inf. Technol. Biomed.",
  ["ieee transactions on information technology in biomedicine : a publication of the ieee engineering in medicine and biology society"] = "IEEE Trans Inf Technol Biomed",
  ["ieee transactions on information theory"] = "IEEE Trans. Inf. Theory",
  ["ieee transactions on instrumentation and measurement"] = "IEEE Trans. Instrum. Meas.",
  ["ieee transactions on intelligent transportation systems"] = "IEEE Trans. Intell. Transp. Syst.",
  ["ieee transactions on intelligent transportation systems : a publication of the ieee intelligent transportation systems council"] = "IEEE trans Intell Transp Syst",
  ["ieee transactions on intelligent vehicles"] = "IEEE Trans. Intell. Veh.",
  ["ieee transactions on knowledge and data engineering"] = "IEEE Trans. Knowl. Data Eng.",
  ["ieee transactions on learning technologies"] = "IEEE Trans. Learn. Technol.",
  ["ieee transactions on machine learning in communications and networking"] = "IEEE Trans. Mach. Learn. Commun. Networking",
  ["ieee transactions on magnetics"] = "IEEE Trans. Magn.",
  ["ieee transactions on man-machine systems"] = "IEEE Trans. Man-Mach. Syst.",
  ["ieee transactions on manufacturing technology"] = "IEEE Trans. Manuf. Technol.",
  ["ieee transactions on materials for electron devices"] = "IEEE Trans. Mater. Electron Devices",
  ["ieee transactions on medical electronics"] = "IEEE Trans. Med. Electron.",
  ["ieee transactions on medical imaging"] = "IEEE Trans Med Imaging",
  ["ieee transactions on medical robotics and bionics"] = "IEEE Trans Med Robot Bionics",
  ["ieee transactions on microwave theory and techniques"] = "IEEE Trans. Microwave Theory Tech.",
  ["ieee transactions on military electronics"] = "IEEE Trans. Mil. Electron.",
  ["ieee transactions on mobile computing"] = "IEEE Trans. Mob. Comput.",
  ["ieee transactions on molecular, biological and multi-scale communications"] = "IEEE Trans. Mol. Biol. Multi-Scale Commun.",
  ["ieee transactions on molecular, biological, and multi-scale communications"] = "IEEE Trans Mol Biol Multiscale Commun",
  ["ieee transactions on multi-scale computing systems"] = "IEEE Trans. Multi-Scale Comput. Syst.",
  ["ieee transactions on multimedia"] = "IEEE Trans. Multimedia",
  ["ieee transactions on nanobioscience"] = "IEEE Trans Nanobioscience",
  ["ieee transactions on nanotechnology"] = "IEEE Trans. Nanotechnol.",
  ["ieee transactions on network and service management"] = "IEEE Trans. Netw. Serv. Manage.",
  ["ieee transactions on network science and engineering"] = "IEEE Trans. Network Sci. Eng.",
  ["ieee transactions on neural networks"] = "IEEE Trans. Neural Networks",
  ["ieee transactions on neural networks / a publication of the ieee neural networks council"] = "IEEE Trans Neural Netw",
  ["ieee transactions on neural networks and learning systems"] = "IEEE Trans Neural Netw Learn Syst",
  ["ieee transactions on neural systems and rehabilitation engineering"] = "IEEE Trans. Neural Syst. Rehabil. Eng.",
  ["ieee transactions on neural systems and rehabilitation engineering : a publication of the ieee engineering in medicine and biology society"] = "IEEE Trans Neural Syst Rehabil Eng",
  ["ieee transactions on nuclear science"] = "IEEE Trans. Nucl. Sci.",
  ["ieee transactions on parallel and distributed systems"] = "IEEE Trans. Parallel Distrib. Syst.",
  ["ieee transactions on parallel and distributed systems : a publication of the ieee computer society"] = "IEEE Trans Parallel Distrib Syst",
  ["ieee transactions on parts, hybrids and packaging"] = "IEEE Trans. Parts, Hybrids, Packag.",
  ["ieee transactions on parts, materials and packaging"] = "IEEE Trans. Parts, Mater., Packag.",
  ["ieee transactions on pattern analysis and machine intelligence"] = "IEEE Trans Pattern Anal Mach Intell",
  ["ieee transactions on plasma science"] = "IEEE Trans. Plasma Sci.",
  ["ieee transactions on plasma science. ieee nuclear and plasma sciences society"] = "IEEE Trans Plasma Sci IEEE Nucl Plasma Sci Soc",
  ["ieee transactions on power apparatus and systems"] = "IEEE Trans. Power App. Syst.",
  ["ieee transactions on power delivery"] = "IEEE Trans. Power Delivery",
  ["ieee transactions on power electronics"] = "IEEE Trans. Power Electron.",
  ["ieee transactions on power systems"] = "IEEE Trans. Power Syst.",
  ["ieee transactions on power systems : a publication of the power engineering society"] = "IEEE Trans Power Syst",
  ["ieee transactions on privacy"] = "IEEE Trans. Privacy",
  ["ieee transactions on professional communication"] = "IEEE Trans. Prof. Commun.",
  ["ieee transactions on quantum engineering"] = "IEEE Trans. Quantum Eng.",
  ["ieee transactions on radar systems"] = "IEEE Trans. Radar Syst.",
  ["ieee transactions on radiation and plasma medical sciences"] = "IEEE Trans Radiat Plasma Med Sci",
  ["ieee transactions on radio frequency interference"] = "IEEE Trans. Radio Freq. Interference",
  ["ieee transactions on rehabilitation engineering"] = "IEEE Trans. Rehabil. Eng.",
  ["ieee transactions on rehabilitation engineering : a publication of the ieee engineering in medicine and biology society"] = "IEEE Trans Rehabil Eng",
  ["ieee transactions on reliability"] = "IEEE Trans. Reliab.",
  ["ieee transactions on robotics"] = "IEEE Trans. Rob.",
  ["ieee transactions on robotics : a publication of the ieee robotics and automation society"] = "IEEE Trans Robot",
  ["ieee transactions on robotics and automation"] = "IEEE Trans. Robot. Autom.",
  ["ieee transactions on robotics and automation : a publication of the ieee robotics and automation society"] = "IEEE Trans Rob Autom",
  ["ieee transactions on semiconductor manufacturing"] = "IEEE Trans. Semicond. Manuf.",
  ["ieee transactions on semiconductor technology modeling and simulation"] = "IEEE Trans. Semicond. Technol. Model. Simul.",
  ["ieee transactions on services computing"] = "IEEE Trans. Serv. Comput.",
  ["ieee transactions on signal and information processing over networks"] = "IEEE Trans Signal Inf Process Netw",
  ["ieee transactions on signal and power integrity"] = "IEEE Trans. Signal Power Integr.",
  ["ieee transactions on signal processing"] = "IEEE Trans. Signal Process.",
  ["ieee transactions on signal processing : a publication of the ieee signal processing society"] = "IEEE Trans Signal Process",
  ["ieee transactions on smart grid"] = "IEEE Trans. Smart Grid",
  ["ieee transactions on software engineering"] = "IEEE Trans. Software Eng.",
  ["ieee transactions on sonics and ultrasonics"] = "IEEE Trans. Sonics Ultrason.",
  ["ieee transactions on space electronics and telemetry"] = "IEEE Trans. Space Electron. Telem.",
  ["ieee transactions on speech and audio processing"] = "IEEE Trans. Speech Audio Process.",
  ["ieee transactions on speech and audio processing : a publication of the ieee signal processing society"] = "IEEE Trans Speech Audio Process",
  ["ieee transactions on sustainable computing"] = "IEEE Trans. Sustainable Comput.",
  ["ieee transactions on sustainable energy"] = "IEEE Trans. Sustainable Energy",
  ["ieee transactions on systems science and cybernetics"] = "IEEE Trans. Syst. Sci. Cybern.",
  ["ieee transactions on systems, man and cybernetics"] = "IEEE Trans. Syst. Man Cybern.",
  ["ieee transactions on systems, man and cybernetics. part c, applications and reviews : a publication of the ieee systems, man, and cybernetics society"] = "IEEE Trans Syst Man Cybern C Appl Rev",
  ["ieee transactions on systems, man, and cybernetics"] = "IEEE Trans Syst Man Cybern",
  ["ieee transactions on systems, man, and cybernetics a: systems and humans"] = "IEEE Trans. Syst., Man, Cybern. A",
  ["ieee transactions on systems, man, and cybernetics b: cybernetics"] = "IEEE Trans. Syst., Man, Cybern. B",
  ["ieee transactions on systems, man, and cybernetics c: applications and reviews"] = "IEEE Trans. Syst., Man, Cybern. C",
  ["ieee transactions on systems, man, and cybernetics part a: systems and humans"] = "IEEE Trans. Syst. Man Cybern. Part A Syst. Humans",
  ["ieee transactions on systems, man, and cybernetics part b: cybernetics"] = "IEEE Trans. Syst. Man Cybern. Part B Cybern.",
  ["ieee transactions on systems, man, and cybernetics part c: applications and reviews"] = "IEEE Trans. Syst. Man Cybern. Part C Appl. Rev.",
  ["ieee transactions on systems, man, and cybernetics. part a, systems and humans : a publication of the ieee systems, man, and cybernetics society"] = "IEEE Trans Syst Man Cybern A Syst Hum",
  ["ieee transactions on systems, man, and cybernetics. part b, cybernetics : a publication of the ieee systems, man, and cybernetics society"] = "IEEE Trans Syst Man Cybern B Cybern",
  ["ieee transactions on systems, man, and cybernetics. systems"] = "IEEE Trans Syst Man Cybern Syst",
  ["ieee transactions on systems, man, and cybernetics: systems"] = "IEEE Trans. Syst. Man Cybern.: Syst.",
  ["ieee transactions on systems, man, and cybernetics—part a: systems and humans"] = "IEEE Trans. Syst., Man, Cybern. A",
  ["ieee transactions on systems, man, and cybernetics—part b: cybernetics"] = "IEEE Trans. Syst., Man, Cybern. B",
  ["ieee transactions on systems, man, and cybernetics—part c: applications and reviews"] = "IEEE Trans. Syst., Man, Cybern. C",
  ["ieee transactions on technology and society"] = "IEEE Trans. Technol. Soc.",
  ["ieee transactions on terahertz science and technology"] = "IEEE Trans. Terahertz Sci. Technol.",
  ["ieee transactions on transportation electrification"] = "IEEE Trans. Transp. Electrif.",
  ["ieee transactions on ultrasonics engineering"] = "IEEE Trans. Ultrason. Eng.",
  ["ieee transactions on ultrasonics, ferroelectrics, and frequency control"] = "IEEE Trans Ultrason Ferroelectr Freq Control",
  ["ieee transactions on vehicular communications"] = "IEEE Trans. Veh. Commun.",
  ["ieee transactions on vehicular technology"] = "IEEE Trans. Veh. Technol.",
  ["ieee transactions on very large scale integration (vlsi) systems"] = "IEEE Trans. Very Large Scale Integr. VLSI Syst.",
  ["ieee transactions on visualization and computer graphics"] = "IEEE Trans Vis Comput Graph",
  ["ieee transactions on wireless communications"] = "IEEE Trans. Wireless Commun.",
  ["ieee translation journal on magnetics in japan"] = "IEEE Transl. J. Magn. Jpn.",
  ["ieee vehicular technology magazine"] = "IEEE Veh. Technol. Mag.",
  ["ieee winter conference on applications of computer vision. ieee winter conference on applications of computer vision"] = "IEEE Winter Conf Appl Comput Vis",
  ["ieee wireless communications"] = "IEEE Wireless Commun.",
  ["ieee wireless communications letters"] = "IEEE Wireless Commun. Lett.",
  ["ieee wireless communications magazine"] = "IEEE Wireless Commun. Mag.",
  ["ieee women in engineering magazine"] = "IEEE Women Eng. Mag.",
  ["ieee workshop on microelectronics and electron devices : [proceedings]. ieee workshop on microelectronics and electron devices"] = "IEEE Workshop Microelectron Electron Devices",
  ["ieee/acm international conference on automated software engineering : [proceedings]. ieee/acm international conference on automated software engineering"] = "IEEE/ACM Int Conf Autom Softw Eng",
  ["ieee/acm international conference on automated software engineering workshops. ieee/acm international conference on automated software engineering"] = "IEEE/ACM Int Conf Autom Softw Eng Workshops",
  ["ieee/acm international symposium on cluster, cloud and grid computing. ieee/acm international symposium on cluster, cloud, and grid computing"] = "IEEE ACM Int Symp Clust Cloud Grid Comput",
  ["ieee/acm transactions on audio, speech and language processing"] = "IEEE/ACM Trans. Audio Speech Lang. Process.",
  ["ieee/acm transactions on audio, speech, and language processing"] = "IEEE/ACM Trans Audio Speech Lang Process",
  ["ieee/acm transactions on computational biology and bioinformatics"] = "IEEE/ACM Trans Comput Biol Bioinform",
  ["ieee/acm transactions on computational biology and bioinformatics / ieee, acm"] = "IEEE/ACM Trans Comput Biol Bioinform",
  ["ieee/acm transactions on networking"] = "IEEE/ACM Trans. Networking",
  ["ieee/acm transactions on networking : a joint publication of the ieee communications society, the ieee computer society, and the acm with its special interest group on data communication"] = "IEEE ACM Trans Netw",
  ["ieee/asme international conference on advanced intelligent mechatronics : [proceedings]. ieee/asme international conference on advanced intelligent mechatronics"] = "IEEE ASME Int Conf Adv Intell Mechatron",
  ["ieee/asme journal of microelectromechanical systems"] = "J. Microelectromech. Syst.",
  ["ieee/asme transactions on mechatronics"] = "IEEE/ASME Trans. Mechatron.",
  ["ieee/asme transactions on mechatronics : a joint publication of the ieee industrial electronics society and the asme dynamic systems and control division"] = "IEEE ASME Trans Mechatron",
  ["ieee/caa journal of automatica sinica"] = "IEEE/CAA J. Autom. Sin.",
  ["ieee/ecs electrochemical and solid-state letters"] = "IEEE/ECS Electrochem. Solid-State Lett.",
  ["ieee/nih life science systems and applications workshop. ieee/nih life science systems and applications workshop"] = "IEEE NIH Life Sci Syst Appl Workshop",
  ["ieee/osa journal of display technology"] = "IEEE/OSA J. Disp. Technol.",
  ["ieee/osa journal of lightwave technology"] = "J. Lightw. Technol.",
  ["ieee/osa journal of optical communications and networking"] = "IEEE/OSA J. Opt. Commun. Networking",
  ["ieee/oup series on electromagnetic wave theory"] = "IEEE/OUP Ser. Electromagn. Wave Theory",
  ["ieee/sice/rsj international conference on multisensor fusion and integration for intelligent systems. ieee/sice/rsj international conference on multisensor fusion and integration for intelligent systems"] = "IEEE SICE RSJ Int Conf Multisens Fusion Integr Intell Syst",
  ["ieee/tms journal of electronic materials"] = "IEEE/TMS J. Electron. Mater.",
  ["ieej journal of industry applications"] = "IEEJ J. Ind. Appl.",
  ["ieej transactions on electrical and electronic engineering"] = "IEEJ Trans. Electr. Electron. Eng.",
  ["ieej transactions on fundamentals and materials"] = "EEJ Trans. Fundam. Mater.",
  ["ieej transactions on sensors and micromachines"] = "IEEJ Trans. Sens. Micromachines",
  ["ieice electronics express"] = "IEICE Electron. Express",
  ["ieice transactions on communications"] = "IEICE Trans. Commun.",
  ["ieice transactions on electronics"] = "IEICE Trans. Electron.",
  ["ieice transactions on fundamentals of electronics, communications and computer science"] = "IEICE Trans. Fundam. Electron. Commun. Comput. Sci.",
  ["ieice transactions on information and systems"] = "IEICE Trans. Inf. Syst.",
  ["ieri procedia"] = "IERI Procedia",
  ["ies journal part a: civil & structural engineering"] = "IES J. Part A: Civ. Struct. Eng.",
  ["iet biometrics"] = "IET Biom.",
  ["iet blockchain"] = "IET Blockchain",
  ["iet circuits, devices and systems"] = "IET Circuits Devices Syst.",
  ["iet collaborative intelligent manufacturing"] = "IET Collab. Intell. Manuf.",
  ["iet communications"] = "IET Commun.",
  ["iet communications engineer"] = "IET Commun. Eng.",
  ["iet computer vision"] = "IET Comput. Vision",
  ["iet computers & digital techniques"] = "IET Comput Digit Tech",
  ["iet computers and digital techniques"] = "IET Comput. Digital Tech.",
  ["iet control engineering series"] = "IET Control Eng. Ser.",
  ["iet control theory & applications"] = "IET Control Theory Appl.",
  ["iet control theory and applications"] = "IET Control Theory Appl.",
  ["iet cyber-physical systems: theory & applications"] = "IET Cyber-Phys. Syst.: Theor. Appl.",
  ["iet cyber-systems and robotics"] = "IET Cyber-Syst. Robot.",
  ["iet electric power applications"] = "IET Electr. Power Appl.",
  ["iet electrical systems in transportation"] = "IET Electr. Syst. Transp.",
  ["iet electronics systems and software"] = "IET Electron. Syst. Softw.",
  ["iet energy systems integration"] = "IET Energy Syst. Integr.",
  ["iet engineering management"] = "IET Eng. Manage.",
  ["iet generation, transmission and distribution"] = "IET Gener. Transm. Distrib.",
  ["iet image processing"] = "IET Image Proc.",
  ["iet information security"] = "IET Inf. Secur.",
  ["iet intelligent transport systems"] = "IET Intel. Transport Syst.",
  ["iet manufacturing"] = "IET Manuf.",
  ["iet micro & nano letters"] = "IET Micro Nano Lett.",
  ["iet microwaves, antennas and propagation"] = "IET Microwaves Antennas Propag.",
  ["iet nanobiotechnology"] = "IET Nanobiotechnol",
  ["iet nanobiotechnology / iet"] = "IET Nanobiotechnol",
  ["iet nanodielectrics"] = "IET Nanodielectr.",
  ["iet networks"] = "IET Networks",
  ["iet optoelectronics"] = "IET Optoelectron",
  ["iet power and energy series"] = "IET Power Energy Ser.",
  ["iet power electronics"] = "IET Power Electron.",
  ["iet power engineer"] = "IET Power Eng.",
  ["iet proceedings: science, measurement & technology"] = "IET Proc.: Sci. Meas. Technol.",
  ["iet quantum communication"] = "IET Quantum Commun.",
  ["iet radar, sonar and navigation"] = "IET Radar Sonar Navig.",
  ["iet renewable power generation"] = "IET Renew. Power Gener.",
  ["iet science, measurement and technology"] = "IET Sci. Meas. Technol.",
  ["iet signal processing"] = "IET Signal Proc.",
  ["iet smart cities"] = "IET Smart Cities",
  ["iet smart grid"] = "IET Smart Grid",
  ["iet software"] = "IET Software",
  ["iet synthetic biology"] = "IET Synth. Biol.",
  ["iet systems biology"] = "IET Syst Biol",
  ["iet wireless sensor systems"] = "IET Wireless Sens. Syst.",
  ["iete journal of education"] = "IETE J. Educ.",
  ["iete journal of research"] = "IETE J Res",
  ["iete technical review"] = "IETE Tech. Rev.",
  ["ifac journal of systems and control"] = "IFAC J. Syst. Control",
  ["ifac proceedings volumes"] = "IFAC Proc. Vol.",
  ["ifac symposia series"] = "IFAC Sympos. Ser.",
  ["ifac workshop series"] = "IFAC Workshop Ser.",
  ["ifip advances in information and communication technology"] = "IFIP Adv Inf Commun Technol",
  ["ifip congress series"] = "IFIP Congress Ser.",
  ["ifip transactions"] = "IFIP Trans.",
  ["ifip transactions a: computer science and technology"] = "IFIP Trans. A Comput. Sci. Tech.",
  ["ifip transactions c: communication systems"] = "IFIP Trans. C Comm. Systems",
  ["ifmbe proceedings"] = "IFMBE Proc",
  ["ifo beitrage zur quantitativen wirtschaftsforschung"] = "Ifo Beitr. Quant. Wirtschaftsforsch.",
  ["ifo studien"] = "Ifo Studien",
  ["iforest - biogeosciences and forestry"] = "iForest - Biogeosc. For.",
  ["iforest : biogeosciences and forestry"] = "IForest",
  ["ifsr international series in systems science and systems engineering"] = "IFSR Int. Ser. Syst. Sci. Syst. Eng.",
  ["ifsr international series on systems science and engineering"] = "IFSR Internat. Ser. Systems Sci. Engrg.",
  ["igaku butsuri : nihon igaku butsuri gakkai kikanshi = japanese journal of medical physics : an official journal of japan society of medical physics"] = "Igaku Butsuri",
  ["igaku kenkyu (acta medica)"] = "Igaku Kenkyu",
  ["igaku kenkyu. acta medica"] = "Igaku Kenkyu",
  ["igaku to seibutsugaku (medicine and biology)"] = "Igaku To Seibutsugaku",
  ["igaku to seibutsugaku. medicine and biology"] = "Igaku To Seibutsugaku",
  ["igcc news"] = "IGCC News",
  ["igiene e sanita pubblica"] = "Ig Sanita Pubbl",
  ["igiene e sanita pubblica. collana di monografie"] = "Ig Sanita Pubblica Collana Monogr",
  ["igiene e sanità pubblica"] = "Ig Sanita Pubbl",
  ["igiene e sanitá pubblica. collana di monografie"] = "Ig Sanita Pubblica Collana Monogr",
  ["igiene moderna"] = "Ig. Mod.",
  ["ihag nonjib"] = "Ihag Nonjib",
  ["iheringia. série zoologia"] = "Iheringia Ser Zool",
  ["ihrim : the journal of the institute of health record information and management"] = "IHRIM",
  ["iiasa proceedings series"] = "IIASA Proc. Ser.",
  ["iiasa reports"] = "IIASA Rep",
  ["iic; international review of industrial property and copyright law"] = "IIC Int Rev Ind Prop Copyr Law",
  ["iie transactions"] = "IIE Trans.",
  ["iie transactions : industrial engineering research & development"] = "IIE Trans",
  ["iie transactions on healthcare systems engineering"] = "IIE Trans Healthc Syst Eng",
  ["iie transactions on occupational ergonomics & human factors"] = "IIE Trans. Occup. Ergon. Hum. Factors",
  ["iie transactions on occupational ergonomics and human factors"] = "IIE Trans Occup",
  ["iioab journal"] = "IIOAB J",
  ["iips newsletter"] = "IIPS Newsl",
  ["iisc lecture notes series"] = "IISc Lect. Notes Ser.",
  ["iise transactions"] = "IISE Trans",
  ["iise transactions on healthcare systems engineering"] = "IISE Trans Healthc Syst Eng",
  ["iise transactions on occupational ergonomics and human factors"] = "IISE Trans Occup Ergon Hum Factors",
  ["iium journal of economics and management"] = "IIUM J. Econ. Manage.",
  ["ijc metabolic & endocrine"] = "IJC Metab Endocr",
  ["ijcai : proceedings of the conference"] = "IJCAI (U S)",
  ["iju case reports"] = "IJU Case Rep",
  ["ikushugaku kenkyu"] = "Ikushugaku Kenkyu",
  ["il cancro"] = "Cancro",
  ["il carrobbio. rivista di studi bolognesi"] = "Carrobbio",
  ["il cervello; giornale di nevrologia"] = "Cervello G Nevrol",
  ["il corriere dei farmacisti"] = "Corr Farm",
  ["il dermosifilografo"] = "Dermosifilografo (Torino)",
  ["il diritto del lavoro"] = "Dir Lav",
  ["il farmaco, scienza e tecnica"] = "Farmaco Sci Tec",
  ["il farmaco; edizione pratica"] = "Farmaco Prat",
  ["il farmaco; edizione scientifica"] = "Farmaco Sci",
  ["il fegato"] = "Fegato",
  ["il filarete: pubblicazioni della facolta di lettere e filosofia dell’universita degli studi di milano"] = "Il Filarete",
  ["il fracastoro"] = "Fracastoro",
  ["il friuli medico"] = "Friuli Med",
  ["il giornale di chirurgia"] = "G Chir",
  ["il lattante"] = "Lattante",
  ["il mar nero. annali di archeologia e storia"] = "MarNero",
  ["il nuovo cimento della societa italiana di fisica. c"] = "Nuovo Cimento C",
  ["il nuovo cimento della società italiana di fisica. c"] = "Nuovo Cimento C",
  ["il nuovo cimento della società italiana di fisica"] = "Nuovo Cimento Soc. Ital. Fis. A (12)",
  ["il pensiero economico italiano"] = "Il Pensiero Econ. Italiano",
  ["il pensiero politico"] = "Pensiero Polit",
  ["il pensiero politico: rivista di storia delle idee politiche e sociali"] = "PPol",
  ["il pisani : organo del manicomio di palermo"] = "Pisani",
  ["il policlinico. sezione chirurgica"] = "Policlinico Chir",
  ["il policlinico. sezione medica"] = "Policlinico Med",
  ["il policlinico. sezione pratica"] = "Policlinico Prat",
  ["il politico"] = "Il Politico",
  ["il ponte"] = "Ponte",
  ["il progresso medico"] = "Prog Med (Napoli)",
  ["il progresso terapeutico"] = "Prog Ter",
  ["il quaderno dell'ostetrica"] = "Quad Ostet",
  ["il risparmio"] = "Il Risparmio",
  ["il sangue"] = "Sangue",
  ["il torace"] = "Torace",
  ["il valsalva"] = "Valsalva",
  ["il veltro"] = "Veltro",
  ["ilar journal"] = "ILAR J",
  ["ilar journal / national research council, institute of laboratory animal resources"] = "ILAR J",
  ["ilar news"] = "Ilar News",
  ["ilerda. instituto de estudios ilerdenses"] = "Ilerda",
  ["iliria. revistë arkeologjike"] = "Iliria",
  ["illinois agricultural economics"] = "Ill. Agr. Econ.",
  ["illinois bar journal"] = "Ill Bar J",
  ["illinois classical studies"] = "Ill Classic Stud",
  ["illinois dental journal"] = "Ill Dent J",
  ["illinois health messenger"] = "Ill Health Messenger",
  ["illinois journal of mathematics"] = "Illinois J. Math.",
  ["illinois research"] = "Ill Res",
  ["illinois revised statutes / compiled and edited under smith-hurd classification. illinois"] = "Ill Revis Statut Ill Ill",
  ["illinois revised statutes. illinois"] = "Ill Revis Statut",
  ["illness, crises, and loss"] = "Illn Crises Loss",
  ["illuminating engineering"] = "Illum Eng",
  ["ilmu kelautan : indonesian journal of marine sciences"] = "Ilmu Kelaut",
  ["ilsa journal of international law"] = "ILSA J Int Law",
  ["im dienste von wald und landschaft"] = "Im Dienste Wald Landsch.",
  ["ima fungus"] = "IMA Fungus",
  ["ima journal of applied mathematics"] = "IMA J. Appl. Math.",
  ["ima journal of management mathematics"] = "IMA J. Manage. Math.",
  ["ima journal of mathematical control and information"] = "IMA J. Math. Control Inf.",
  ["ima journal of mathematics applied in business and industry"] = "IMA J. Math. Appl. Bus. Indust.",
  ["ima journal of mathematics applied in medicine and biology"] = "IMA J Math Appl Med Biol",
  ["ima journal of numerical analysis"] = "IMA J. Numer. Anal.",
  ["image - the journal of nursing scholarship"] = "Image J Nurs Sch",
  ["image analysis & stereology. quantitative methods and applications"] = "Image Anal. Stereol.",
  ["image analysis and stereology"] = "Image Anal. Stereol.",
  ["image analysis for moving organ, breast, and thoracic images : third international workshop, rambo 2018, fourth international workshop, bia 2018, and first international workshop, tia 2018, held in conjunction with miccai 2018, granada, spain, september 16 and 20, 2018, proceedings. rambo (workshop) (3rd : 2018 : granada, spain)"] = "Image Anal Mov Organ Breast Thorac Images (2018)",
  ["image and vision computing"] = "Image Vision Comput.",
  ["image--the journal of nursing scholarship"] = "Image J Nurs Sch",
  ["image; the journal of nursing scholarship"] = "Image J. Nurs. Sch.",
  ["images in paediatric cardiology"] = "Images Paediatr Cardiol",
  ["images; marquette university dental reflections"] = "Images Marquette Univ Dent Reflections",
  ["imagination, cognition and personality"] = "Imagin Cogn Pers",
  ["imaging decisions (berlin, germany)"] = "Imaging Decis (Berl)",
  ["imaging for patient-customized simulations and systems for point-of-care ultrasound : international workshops, bivpcs 2017 and pocus 2017, held in conjunction with miccai 2017, québec city, qc, canada, september 14, 2017, proceedings. bivpcs (workshop) (2017 : québec, québec)"] = "Imaging Patient Cust Simul Syst Point Care Ultrasound (2017)",
  ["imaging in medicine"] = "Imaging Med",
  ["imaging neuroscience"] = "Imaging Neurosci.",
  ["imaging science in dentistry"] = "Imaging Sci Dent",
  ["imaging science journal"] = "Imaging Sci. J.",
  ["imago mundi"] = "Imago Mundi",
  ["imbonezamuryango = famille, santé, développement"] = "Imbonezamuryango",
  ["imch newsletter"] = "ICMH Newsl",
  ["imch newsletter /"] = "ICMH Newsl",
  ["imf staff papers"] = "IMF Staff Pap.",
  ["imf survey. international monetary fund"] = "IMF Surv",
  ["imhotep mathematical proceedings"] = "Imhotep Math. Proc.",
  ["imhotep. journal africain de mathematiques pures et appliquees. african journal of pure and applied mathematics"] = "IMHOTEP J. Afr. Math. Pures Appl.",
  ["imj, illinois medical journal"] = "IMJ Ill. Med. J.",
  ["imj. illinois medical journal"] = "IMJ Ill Med J",
  ["immigrants & minorities"] = "Immigr Minor",
  ["immune network"] = "Immune Netw",
  ["immune system"] = "Immune Syst.",
  ["immunitat und infektion"] = "Immun. Infekt.",
  ["immunität und infektion"] = "Immun Infekt",
  ["immunity & ageing : i & a"] = "Immun Ageing",
  ["immunity, inflammation and disease"] = "Immun Inflamm Dis",
  ["immunobiology. supplement"] = "Immunobiol Suppl",
  ["immunodeficiency reviews"] = "Immunodefic Rev",
  ["immunoendocrinology (houston, tex.)"] = "Immunoendocrinology (Houst)",
  ["immunohematology / american red cross"] = "Immunohematol",
  ["immunologia clinica"] = "Immunol Clin",
  ["immunologic research"] = "Immunol Res",
  ["immunological communications"] = "Immunol Commun",
  ["immunological disorders & immunotherapy"] = "Immunol Disord Immunother",
  ["immunological investigations"] = "Immunol Invest",
  ["immunological medicine"] = "Immunol Med",
  ["immunological reviews"] = "Immunol Rev",
  ["immunology & cell biology"] = "Immunol. Cell Biol.",
  ["immunology & infectious diseases"] = "Immunol Infect Dis",
  ["immunology and allergy clinics of north america"] = "Immunol Allergy Clin North Am",
  ["immunology and cell biology"] = "Immunol Cell Biol",
  ["immunology and immunogenetics insights"] = "Immunol Immunogenet  Insights",
  ["immunology and infectious diseases"] = "Immunol Infect Dis",
  ["immunology innovation"] = "Immunol Innov",
  ["immunology letters"] = "Immunol Lett",
  ["immunology series"] = "Immunol Ser",
  ["immunology today"] = "Immunol Today",
  ["immunology, endocrine & metabolic agents in medicinal chemistry"] = "Immunol Endocr Metab Agents Med Chem",
  ["immunology. supplement"] = "Immunol Suppl",
  ["immunome research"] = "Immunome Res",
  ["immunopharmacology and immunotoxicology"] = "Immunopharmacol Immunotoxicol",
  ["immunotargets and therapy"] = "Immunotargets Ther",
  ["immunotechnology : an international journal of immunological engineering"] = "Immunotechnology",
  ["immunotherapy (los angeles, calif.)"] = "Immunotherapy (Los Angel)",
  ["immunotherapy advances"] = "Immunother. Adv.",
  ["impa monographs"] = "IMPA Monogr.",
  ["impact (american college of physicians)"] = "Impact (Am Coll Physicians)",
  ["impact assessment and project appraisal"] = "Impact Assess. Project Appraisal",
  ["impact magazine"] = "IMPACT  Mag",
  ["impact of computing in science and engineering"] = "IMPACT Comput. Sci. Eng.",
  ["impact of science on society"] = "Impact Sci Soc",
  ["impact on hiv"] = "Impact HIV",
  ["impan lecture notes"] = "IMPAN Lect. Notes",
  ["impartial, l’"] = "Impartial",
  ["implant dentistry"] = "Implant Dent",
  ["implant society"] = "Implant Soc.",
  ["implementation research and practice"] = "Implement Res Pract",
  ["implementation science : is"] = "Implement Sci",
  ["implementation science communications"] = "Implement Sci Commun",
  ["implicit religion : journal of the centre for the study of implicit religion and contemporary spirituality"] = "Implicit Relig",
  ["important advances in oncology"] = "Important Adv Oncol",
  ["imprensa medica"] = "Imprensa Medica",
  ["imprensa médica"] = "Imprensa Medica",
  ["imprint (american historical print collectors society)"] = "Imprint (N Y N Y 1976)",
  ["improving human performance quarterly"] = "Improv Hum Performance Q",
  ["impulse (columbia, s.c.)"] = "Impulse (Columbia)",
  ["ims report"] = "IMS Rep.",
  ["ims, industrial medicine and surgery"] = "IMS Ind Med Surg",
  ["imsc lecture notes in mathematics"] = "IMSc Lect. Notes Math.",
  ["in health"] = "In Health",
  ["in mind : the inquisitive mind, social psychology for you"] = "In Mind",
  ["in point of fact"] = "In Point Fact",
  ["in practice"] = "In Pract",
  ["in silico biology"] = "In Silico Biol",
  ["in silico cell and tissue science"] = "In Silico Cell Tissue Sci",
  ["in silico pharmacology"] = "In Silico Pharmacol",
  ["in silico plants"] = "In Silico Plants",
  ["in vitro"] = "In Vitro",
  ["in vitro & molecular toxicology"] = "In Vitr Mol Toxicol",
  ["in vitro and molecular toxicology"] = "In Vitr. Mol. Toxicol.",
  ["in vitro cellular & developmental biology : journal of the tissue culture association"] = "In Vitro Cell Dev Biol",
  ["in vitro cellular & developmental biology. animal"] = "In Vitro Cell Dev Biol Anim",
  ["in vitro cellular & developmental biology. plant : journal of the tissue culture association"] = "In Vitro Cell Dev Biol Plant",
  ["in vitro cellular & developmental biology: animal"] = "In Vitro Cell. Dev. Biol. Anim.",
  ["in vitro cellular & developmental biology: plant"] = "In Vitro Cell. Dev. Biol. Plant",
  ["in vitro cellular and developmental biology"] = "In Vitro Cell. Dev. Biol.",
  ["in vitro cellular and developmental biology. animal"] = "In Vitro Cell. Dev. Biol. Anim.",
  ["in vitro models"] = "In vitro models",
  ["in vitro toxicology"] = "In Vitro Toxicol",
  ["in vitro. monograph"] = "In Vitro Monogr",
  ["in vivo"] = "In Vivo",
  ["in vivo (athens, greece)"] = "In Vivo",
  ["in'gan kwahak. human science"] = "Ingan Kwahak",
  ["in'gu munje nonjip = journal of population studies"] = "Ingu munje nonjip",
  ["in'gu pogŏn nonjip = journal of population and health studies"] = "Ingu Pogon Nonjip",
  ["in/fire ethics : newsletter of the international network of feminists interested in reproductive health"] = "In Fire Ethics",
  ["inclusion (washington, d.c.)"] = "Inclusion (Wash)",
  ["incontri linguistici"] = "ILing",
  ["indagationes mathematicae"] = "Indagationes Math.",
  ["indent; journal of international dentistry; english edition"] = "Indent Engl",
  ["independent practitioner (lutterworth, england)"] = "Indep Pract (Lutterworth)",
  ["independent review"] = "Independent Rev.",
  ["index fungorum : published numbers"] = "Index Fungorum",
  ["index. quaderni camerti di studi romanistici"] = "IndexQuad",
  ["india policy forum : [papers]. india policy forum. conference"] = "India Policy Forum",
  ["india quarterly"] = "India Q",
  ["indian academy of mathematics"] = "J. Indian Acad. Math.",
  ["indian academy of sciences"] = "Proc. Indian Acad. Sci. Math. Sci.",
  ["indian academy of sciences. proceedings. mathematical sciences"] = "Proc. Indian Acad. Sci. Math. Sci.",
  ["indian archives"] = "Indian Arch",
  ["indian chemical engineer"] = "Indian Chem. Eng.",
  ["indian concrete journal"] = "Indian Concr. J.",
  ["indian dermatology online journal"] = "Indian Dermatol Online J",
  ["indian drugs"] = "Indian Drugs",
  ["indian economic and social history review"] = "Indian Econ. Soc. Hist. Rev.",
  ["indian economic journal"] = "Indian Econ. J.",
  ["indian economic review"] = "Indian Econ Rev",
  ["indian fern journal : international journal of pteridology published by the indian fern society"] = "Indian Fern J",
  ["indian foundry journal"] = "Indian Foundry J.",
  ["indian geotechnical journal"] = "Indian Geotech. J.",
  ["indian heart journal"] = "Indian Heart J",
  ["indian heart journal. teaching series"] = "Indian Heart J. Teach. Ser.",
  ["indian heart journal: teaching series"] = "Indian Heart J Teach Ser",
  ["indian institute of science"] = "J. Indian Inst. Sci.",
  ["indian journal of (geo) marine sciences"] = "Indian J Geomarine Sci",
  ["indian journal of agricultural biochemistry"] = "Indian J.  Agric. Biochem.",
  ["indian journal of agricultural chemistry"] = "Indian J. Agric. Chem.",
  ["indian journal of agricultural economics"] = "Indian J.  Agric. Econ.",
  ["indian journal of agricultural research"] = "Indian J Agric Res",
  ["indian journal of agricultural sciences"] = "Indian J.  Agric. Sci.",
  ["indian journal of agriculture, environment and bio-technology"] = "Indian J. Agric. Environ. Bio-Technol.",
  ["indian journal of agronomy"] = "Indian J. Agron.",
  ["indian journal of american studies"] = "Indian J Am Stud",
  ["indian journal of anaesthesia"] = "Indian J Anaesth",
  ["indian journal of animal research"] = "Indian J Anim Res",
  ["indian journal of animal sciences"] = "Indian J. Anim. Sci.",
  ["indian journal of applied economics"] = "Indian J. Appl. Econ.",
  ["indian journal of applied research"] = "Indian J Appl Res",
  ["indian journal of behaviour"] = "Indian J Behav",
  ["indian journal of biochemistry"] = "Indian J. Biochem.",
  ["indian journal of biochemistry & biophysics"] = "Indian J Biochem Biophys",
  ["indian journal of biochemistry and biophysics"] = "Indian J. Biochem. Biophys.",
  ["indian journal of biotechnology"] = "Indian J Biotechnol",
  ["indian journal of cancer"] = "Indian J Cancer",
  ["indian journal of cardiovascular disease in women wincars"] = "Indian J Cardiovasc Dis Women WINCARS",
  ["indian journal of chemical technology"] = "Indian J. Chem. Technol.",
  ["indian journal of chemistry"] = "Indian J Chem",
  ["indian journal of chemistry section a: inorganic, bio-inorganic, physical, theoretical and analytical chemistry"] = "Indian J. Chem., Sect A",
  ["indian journal of chemistry section b: organic chemistry including medicinal chemistry"] = "Indian J. Chem., Sect B",
  ["indian journal of chemistry, section a: inorganic, bio-inorganic, physical, theoretical & analytical chemistry"] = "Indian J. Chem., Sect. A: Inorg., Bio-inorg., Phys., Theor. Anal. Chem.",
  ["indian journal of chemistry, section b: organic chemistry including medicinal chemistry"] = "Indian J. Chem., Sect. B: Org. Chem. Incl. Med. Chem.",
  ["indian journal of chest diseases"] = "Indian J. Chest Dis.",
  ["indian journal of chest diseases and allied sciences"] = "Indian J. Chest Dis. Allied Sci.",
  ["indian journal of child health"] = "Indian J Child Health (Bhopal)",
  ["indian journal of clinical anatomy and physiology"] = "Indian J Clin Anat Physiol",
  ["indian journal of clinical biochemistry"] = "Indian J. Clin. Biochem.",
  ["indian journal of clinical biochemistry : ijcb"] = "Indian J Clin Biochem",
  ["indian journal of clinical cardiology"] = "Indian J. Clin. Cardiol.",
  ["indian journal of clinical psychology"] = "Indian J Clin Psychol",
  ["indian journal of community health"] = "Indian J Community Health",
  ["indian journal of community medicine : official publication of indian association of preventive & social medicine"] = "Indian J Community Med",
  ["indian journal of computer science and engineering"] = "Indian J. Comput. Sci. Eng.",
  ["indian journal of critical care medicine : peer-reviewed, official publication of indian society of critical care medicine"] = "Indian J Crit Care Med",
  ["indian journal of dental research"] = "Indian J. Dent. Res.",
  ["indian journal of dental research : official publication of indian society for dental research"] = "Indian J Dent Res",
  ["indian journal of dentistry"] = "Indian J Dent",
  ["indian journal of dermatology"] = "Indian J Dermatol",
  ["indian journal of dermatology and venereology"] = "Indian J Dermatol Venereol",
  ["indian journal of dermatology, venereology and leprology"] = "Indian J Dermatol Venereol Leprol",
  ["indian journal of discrete mathematics"] = "Indian J. Discrete Math.",
  ["indian journal of ecology"] = "Indian J. Ecol.",
  ["indian journal of economics"] = "Indian J Econ",
  ["indian journal of economics and development"] = "Indian J Econ Dev",
  ["indian journal of endocrinology and metabolism"] = "Indian J Endocrinol Metab",
  ["indian journal of engineering and materials science"] = "Indian J. Eng. Mater. Sci.",
  ["indian journal of entomology"] = "Indian J. Entomol.",
  ["indian journal of environment and ecoplanning"] = "Indian J. Environ. Ecoplann.",
  ["indian journal of environmental health"] = "Indian J Environ Health",
  ["indian journal of environmental protection"] = "Indian J. Environ. Prot.",
  ["indian journal of experimental biology"] = "Indian J Exp Biol",
  ["indian journal of fibre & textile research"] = "Indian J Fibre Text Res",
  ["indian journal of fibre and textile research"] = "Indian J. Fibre Text. Res.",
  ["indian journal of fisheries"] = "Indian J. Fish.",
  ["indian journal of forensic medicine and pathology"] = "Indian J. Forensic Med. Pathol.",
  ["indian journal of forensic medicine and toxicology"] = "Indian J. Forensic Med. Toxicol.",
  ["indian journal of forensic sciences"] = "Indian J. Forensic Sci.",
  ["indian journal of gastroenterology"] = "Indian J. Gastroenterol.",
  ["indian journal of gastroenterology : official journal of the indian society of gastroenterology"] = "Indian J Gastroenterol",
  ["indian journal of gender studies"] = "Indian J Gend Stud",
  ["indian journal of genetics and plant breeding"] = "Indian J. Genet. Plant Breed.",
  ["indian journal of geo marine sciences"] = "Indian J. Geo Mar. Sci.",
  ["indian journal of geochemistry"] = "Indian J. Geochem.",
  ["indian journal of geosciences"] = "Indian J. Geosci.",
  ["indian journal of gynecologic oncology"] = "Indian J Gynecol Oncol",
  ["indian journal of helminthology"] = "Indian J Helminthol",
  ["indian journal of hematology & blood transfusion : an official journal of indian society of hematology and blood transfusion"] = "Indian J Hematol Blood Transfus",
  ["indian journal of heterocyclic chemistry"] = "Indian J. Heterocycl. Chem.",
  ["indian journal of history of science"] = "Indian J. Hist. Sci.",
  ["indian journal of horticulture"] = "Indian J. Hortic.",
  ["indian journal of human genetics"] = "Indian J Hum Genet",
  ["indian journal of industrial and applied mathematics"] = "Indian J. Ind. Appl. Math.",
  ["indian journal of industrial relations"] = "Indian J Ind Relat",
  ["indian journal of labour economics"] = "Indian J. Lab. Econ.",
  ["indian journal of leprosy"] = "Indian J Lepr",
  ["indian journal of malariology"] = "Indian J Malariol",
  ["indian journal of marine sciences"] = "Indian J Mar Sci",
  ["indian journal of maternal and child health : official publication of indian maternal and child health association"] = "Indian J Matern Child Health",
  ["indian journal of maternal and child health : official publication of indian maternal and child health association,"] = "Indian J Matern Child Health",
  ["indian journal of mathematics"] = "Indian J. Math.",
  ["indian journal of medical and paediatric oncology : official journal of indian society of medical & paediatric oncology"] = "Indian J Med Paediatr Oncol",
  ["indian journal of medical education"] = "Indian J Med Educ",
  ["indian journal of medical ethics"] = "Indian J Med Ethics",
  ["indian journal of medical informatics"] = "Indian J Med Inform",
  ["indian journal of medical microbiology"] = "Indian J Med Microbiol",
  ["indian journal of medical research"] = "Indian J. Med. Res.",
  ["indian journal of medical sciences"] = "Indian J Med Sci",
  ["indian journal of microbiology"] = "Indian J Microbiol",
  ["indian journal of natural products"] = "Indian J. Nat. Prod.",
  ["indian journal of natural products and resources"] = "Indian J Nat Prod Resour",
  ["indian journal of nematology"] = "Indian J. Nematol.",
  ["indian journal of nematology : official publication of the nematological society of india"] = "Indian J Nematol",
  ["indian journal of nephrology"] = "Indian J Nephrol",
  ["indian journal of nuclear medicine"] = "Indian J. Nucl. Med.",
  ["indian journal of nuclear medicine : ijnm : the official journal of the society of nuclear medicine, india"] = "Indian J Nucl Med",
  ["indian journal of occupational and environmental medicine"] = "Indian J Occup Environ Med",
  ["indian journal of ophthalmology"] = "Indian J Ophthalmol",
  ["indian journal of orthopaedics"] = "Indian J Orthop",
  ["indian journal of otolaryngology"] = "Indian J Otolaryngol",
  ["indian journal of otolaryngology and head and neck surgery : official publication of the association of otolaryngologists of india"] = "Indian J Otolaryngol Head Neck Surg",
  ["indian journal of palliative care"] = "Indian J Palliat Care",
  ["indian journal of pathology & bacteriology"] = "Indian J Pathol Bacteriol",
  ["indian journal of pathology & microbiology"] = "Indian J Pathol Microbiol",
  ["indian journal of pathology and bacteriology"] = "Indian J. Pathol. Bacteriol.",
  ["indian journal of pathology and microbiology"] = "Indian J. Pathol. Microbiol.",
  ["indian journal of pediatrics"] = "Indian J Pediatr",
  ["indian journal of pharmaceutical education and research"] = "Indian J. Pharm. Educ. Res.",
  ["indian journal of pharmaceutical science"] = "Indian J. Pharm. Sci.",
  ["indian journal of pharmaceutical sciences"] = "Indian J Pharm Sci",
  ["indian journal of pharmacology"] = "Indian J Pharmacol",
  ["indian journal of physical anthropology and human genetics"] = "Indian J Phys Anthropol Hum Genet",
  ["indian journal of physics"] = "Indian J. Phys.",
  ["indian journal of physics and proceedings of the indian association for the cultivation of science (2004)"] = "Indian J Phys Proc Indian Assoc Cultiv Sci (2004)",
  ["indian journal of physics, a"] = "Indian J. Phys., A",
  ["indian journal of physics, b"] = "Indian J. Phys., B",
  ["indian journal of physiology and pharmacology"] = "Indian J Physiol Pharmacol",
  ["indian journal of physiotherapy and occupational therapy"] = "Indian J Physiother Occup Ther",
  ["indian journal of plant physiology"] = "Indian J Plant Physiol",
  ["indian journal of plant protection"] = "Indian J Plant Prot",
  ["indian journal of plastic surgery : official publication of the association of plastic surgeons of india"] = "Indian J Plast Surg",
  ["indian journal of preventive and social medicine"] = "Indian J Prev Soc Med",
  ["indian journal of psychiatry"] = "Indian J Psychiatry",
  ["indian journal of psychological medicine"] = "Indian J Psychol Med",
  ["indian journal of psychology"] = "Indian J Psychol",
  ["indian journal of public health"] = "Indian J Public Health",
  ["indian journal of public health research & development"] = "Indian J Public Health Res Dev",
  ["indian journal of pure and applied mathematics"] = "Indian J. Pure Appl. Math.",
  ["indian journal of pure and applied physics"] = "Indian J. Pure Appl. Phys.",
  ["indian journal of quantitative economics"] = "Indian J Quant Econ",
  ["indian journal of radio & space physics"] = "Indian J. Radio Space Phys.",
  ["indian journal of radio and space physics"] = "Indian J. Radio Space Phys.",
  ["indian journal of radiology and imaging"] = "Indian J. Radiol. Imaging",
  ["indian journal of rheumatology"] = "Indian J Rheumatol",
  ["indian journal of science"] = "Indian J Sci",
  ["indian journal of science and technology"] = "Indian J Sci Technol",
  ["indian journal of scientific research"] = "Indian J Sci Res",
  ["indian journal of sexually transmitted diseases"] = "Indian J Sex Transm Dis",
  ["indian journal of sexually transmitted diseases and aids"] = "Indian J Sex Transm Dis AIDS",
  ["indian journal of sleep medicine"] = "Indian J Sleep Med",
  ["indian journal of small ruminants"] = "Indian J. Small Ruminants",
  ["indian journal of social psychiatry"] = "Indian J Soc Psychiatry",
  ["indian journal of social research"] = "Indian J Soc Res",
  ["indian journal of social science"] = "Indian J Soc Sci",
  ["indian journal of surgical oncology"] = "Indian J Surg Oncol",
  ["indian journal of technology"] = "Indian J. Technol.",
  ["indian journal of the history of medicine"] = "Indian J Hist Med",
  ["indian journal of thoracic and cardiovascular surgery"] = "Indian J Thorac Cardiovasc Surg",
  ["indian journal of traditional knowledge"] = "Indian J. Tradit. Knowl.",
  ["indian journal of urology : iju : journal of the urological society of india"] = "Indian J Urol",
  ["indian journal of venereal diseases and dermatology"] = "Indian J Vener Dis Dermatol",
  ["indian journal of veterinary pathology"] = "Indian J Vet Pathol",
  ["indian journal of virology : an official organ of indian virological society"] = "Indian J Virol",
  ["indian medical gazette"] = "Ind Med Gaz",
  ["indian medical journal"] = "Indian Med J",
  ["indian medical record"] = "Indian Med Rec",
  ["indian medical tribune"] = "Indian Med Trib",
  ["indian obstetrics & gynaecology"] = "Indian Obstet Gynaecol",
  ["indian pacing and electrophysiology journal"] = "Indian Pacing Electrophysiol J",
  ["indian pediatrics"] = "Indian Pediatr",
  ["indian phytopathology"] = "Indian Phytopathol",
  ["indian practitioner"] = "Indian Pract.",
  ["indian psychological review"] = "Indian Psychol Rev",
  ["indian science cruiser"] = "Indian Sci. Cruiser",
  ["indian sociological bulletin"] = "Indian Sociol Bull",
  ["indian statistical institute series"] = "Indian Stat. Inst. Ser.",
  ["indian veterinary journal"] = "Indian Vet. J.",
  ["indiana business review"] = "Indiana Bus. Rev.",
  ["indiana code : comprising all laws of a general and permanent nature including statutes enacted through the 1982 regular session of the indiana general assembly. indiana"] = "Indiana Code Indiana",
  ["indiana folklore"] = "Indiana Folkl",
  ["indiana health law review"] = "Indiana Health Law Rev",
  ["indiana health law review / [indiana university school of law-indianapolis]"] = "Indiana Health Law Rev",
  ["indiana international & comparative law review"] = "Indiana Int Comp Law Rev",
  ["indiana law journal (indianapolis, ind. : 1926)"] = "Indiana Law J",
  ["indiana law review"] = "Indiana Law Rev",
  ["indiana magazine of history"] = "Indiana Mag Hist",
  ["indiana medical history quarterly"] = "Indiana Med Hist Q",
  ["indiana medical history quarterly / indiana historical society"] = "Indiana Med Hist Q",
  ["indiana medicine"] = "Indiana Med.",
  ["indiana medicine : the journal of the indiana state medical association"] = "Indiana Med",
  ["indiana nurse"] = "Indiana Nurse",
  ["indiana social studies quarterly"] = "Indiana Soc Stud Q",
  ["indiana university art museum bulletin"] = "IndUnArtB",
  ["indiana university journal of undergraduate research"] = "Indiana Univ J Undergrad Res",
  ["indiana university mathematics journal"] = "Indiana Univ. Math. J.",
  ["indicator (minnesota mining and manfuacturing company)"] = "Indicator",
  ["indices naturwissenschaftlich-medizinischer periodica bis 1850"] = "Indices Naturwiss Med Period 1850",
  ["individual differences research : idr"] = "Individ Differ Res",
  ["individual psychology"] = "Individ Psychol",
  ["individual psychology bulletin"] = "Individ Psychol Bull",
  ["indo koten kenkyū = acta indologica"] = "Indo Koten Kenkyu",
  ["indo-iranian journal"] = "Indo Iran J",
  ["indogermanische forschungen"] = "IF",
  ["indonesia and the malay world"] = "Indones Malay World",
  ["indonesian fisheries research journal"] = "Indones Fish Res J",
  ["indonesian journal of biotechnology"] = "Indones J Biotechnol",
  ["indonesian journal of chemistry"] = "Indones. J. Chem.",
  ["indoor + built environment : the journal of the international society of the built environment"] = "Indoor Built Environ",
  ["indoor air"] = "Indoor Air",
  ["indoor and built environment"] = "Indoor Built Environ.",
  ["inductive logic programming. ilp"] = "Inductive Log Program",
  ["industrial & engineering chemistry chemical & engineering data series"] = "Ind. Eng. Chem. Chem. Eng. Data Series",
  ["industrial & engineering chemistry process design and development"] = "Ind. Eng. Chem. Process Des. Dev.",
  ["industrial & engineering chemistry product research and development"] = "Ind. Eng. Chem. Prod. Res. Dev.",
  ["industrial & engineering chemistry research"] = "Ind Eng Chem Res",
  ["industrial & labor relations review"] = "Ind Labor Relat Rev",
  ["industrial and applied mathematics"] = "Ind. Appl. Math.",
  ["industrial and corporate change"] = "Ind. Corp. Change",
  ["industrial and engineering chemistry"] = "Ind Eng Chem",
  ["industrial and engineering chemistry fundamentals"] = "Ind. Eng. Chem. Fundam.",
  ["industrial and engineering chemistry research"] = "Ind. Eng. Chem. Res.",
  ["industrial and engineering chemistry, analytical edition"] = "Ind. Eng. Chem. Anal. Ed.",
  ["industrial and labor relations review"] = "Ind. Lab. Relat. Rev.",
  ["industrial and organizational psychology"] = "Ind Organ Psychol",
  ["industrial archaeology review"] = "Ind Archaeol Rev",
  ["industrial biotechnology"] = "Ind. Biotechnol.",
  ["industrial biotechnology (new rochelle, n.y.)"] = "Ind Biotechnol (New Rochelle N Y)",
  ["industrial chemistry & materials"] = "Ind. Chem. Mater.",
  ["industrial crops and products"] = "Ind. Crops Prod.",
  ["industrial crops research journal"] = "Ind Crops Res J",
  ["industrial diamond review"] = "Ind. Diamond Rev.",
  ["industrial engineering"] = "Ind. Eng.",
  ["industrial engineering (american institute of industrial engineers)"] = "Ind Eng",
  ["industrial health"] = "Ind Health",
  ["industrial health bulletin"] = "Ind Health Bull",
  ["industrial health monthly"] = "Ind Health Mon",
  ["industrial health review"] = "Ind Health Rev",
  ["industrial hygiene newsletter"] = "Ind Hyg Newsl",
  ["industrial lubrication and tribology"] = "Ind. Lubr. Tribol.",
  ["industrial management"] = "Ind. Manage.",
  ["industrial mathematics"] = "Indust. Math.",
  ["industrial medicine & surgery"] = "Ind Med Surg",
  ["industrial medicine and surgery"] = "Ind. Med. Surg.",
  ["industrial metrology"] = "Ind. Metrol.",
  ["industrial nursing"] = "Ind Nurs",
  ["industrial organization review"] = "Ind. Organ. Rev.",
  ["industrial photography"] = "Ind. Photogr.",
  ["industrial psychiatry journal"] = "Ind Psychiatry J",
  ["industrial relations"] = "Ind Relat (Berkeley)",
  ["industrial robot"] = "Ind. Robot",
  ["industrielle organisation"] = "Ind. Organ.",
  ["industry and development"] = "Industry Devel.",
  ["industry and health care"] = "Ind Health Care",
  ["industry and health care (ballinger publishing co.)"] = "Ind Health Care (Cambridge Ma)",
  ["industry and innovation"] = "Ind Innov",
  ["industry week"] = "Ind Week",
  ["inf & ing: vorlesungen zum informatik- und ingenieurstudium"] = "Inf & Ing Vorles. Inform. Ingenieurstud.",
  ["infancy : the official journal of the international society on infant studies"] = "Infancy",
  ["infant and child development"] = "Infant Child Dev",
  ["infant behavior & development"] = "Infant Behav Dev",
  ["infant mental health journal"] = "Infant Ment Health J",
  ["infant, child & adolescent nutrition"] = "Infant Child Adolesc Nutr",
  ["infants and young children"] = "Infants Young Child",
  ["infanzia anormale"] = "Infanz Anorm",
  ["infectio : revista de la asociación colombiana de infectología"] = "Infectio",
  ["infection & chemotherapy"] = "Infect Chemother",
  ["infection and drug resistance"] = "Infect Drug Resist",
  ["infection and immunity"] = "Infect Immun",
  ["infection control"] = "Infect. Control",
  ["infection control & urological care"] = "Infect Control Urol Care",
  ["infection control : ic"] = "Infect Control",
  ["infection control and hospital epidemiology"] = "Infect Control Hosp Epidemiol",
  ["infection control and hospital epidemiology : the official journal of the society of hospital epidemiologists of america"] = "Infect Control Hosp Epidemiol",
  ["infection control and urological care"] = "Infect. Control Urol. Care",
  ["infection control canada"] = "Infect Control Can",
  ["infection control digest"] = "Infect Control Dig",
  ["infection control rounds"] = "Infect Control Rounds",
  ["infection control weekly"] = "Infect Control Wkly",
  ["infection ecology & epidemiology"] = "Infect Ecol Epidemiol",
  ["infection epidemiology and medicine"] = "Infect Epidemiol Med",
  ["infection, disease & health"] = "Infect Dis Health",
  ["infection, epidemiology and microbiology"] = "Infect Epidemiol Microbiol",
  ["infection, genetics and evolution"] = "Infect., Genet. Evol.",
  ["infection, genetics and evolution : journal of molecular epidemiology and evolutionary genetics in infectious diseases"] = "Infect Genet Evol",
  ["infection, immunity, and pharmacology"] = "Infect Immun Pharmacol",
  ["infections in medicine"] = "Infect Med",
  ["infectious agents and cancer"] = "Infect Agent Cancer",
  ["infectious agents and disease"] = "Infect Agents Dis",
  ["infectious disease clinics of north america"] = "Infect Dis Clin North Am",
  ["infectious disease modelling"] = "Infect Dis Model",
  ["infectious disease reports"] = "Infect Dis Rep",
  ["infectious diseases"] = "Infect Dis",
  ["infectious diseases (london, england)"] = "Infect Dis (Lond)",
  ["infectious diseases and therapeutics"] = "Infect Dis Ther",
  ["infectious diseases and therapy"] = "Infect Dis Ther",
  ["infectious diseases diagnosis & treatment"] = "Infect Dis Diagn Treat",
  ["infectious diseases in clinical practice (baltimore, md.)"] = "Infect Dis Clin Pract (Baltim Md)",
  ["infectious diseases in obstetrics and gynecology"] = "Infect Dis Obstet Gynecol",
  ["infectious diseases journal : idj"] = "Infect Dis J",
  ["infectious diseases newsletter (new york, n.y.)"] = "Infect Dis Newsl (N Y)",
  ["infectious diseases now"] = "Infect Dis Now",
  ["infectious diseases of poverty"] = "Infect Dis Poverty",
  ["infectious disorders drug targets"] = "Infect Disord Drug Targets",
  ["infectious disorders: drug targets"] = "Infect. Disord.: Drug Targets",
  ["infectious microbes & diseases"] = "Infect. Microbes Dis.",
  ["infekt︠s︡ii︠a︡ i immunitet"] = "Infektsiia Immun",
  ["infinite dimensional analysis quantum probability and related topics"] = "Infin. Dimension. Anal. Quantum Probab. Relat. Top.",
  ["infinite dimensional analysis, quantum probability and related topics"] = "Infin. Dimens. Anal. Quantum Probab. Relat. Top.",
  ["infirmiere auxiliaire"] = "Infirm. Aux.",
  ["infirmiere canadienne"] = "Infirm. Can.",
  ["infirmiere du quebec"] = "Infirm. Que.",
  ["infirmiere francaise"] = "Infirm. Fr.",
  ["infirmiere haitienne"] = "Infirm. Haiti.",
  ["inflammation & allergy drug targets"] = "Inflamm Allergy Drug Targets",
  ["inflammation & allergy: drug targets"] = "Inflamm. Allergy: Drug Targets",
  ["inflammation and allergy drug targets"] = "Inflamm. Allergy Drug Targets",
  ["inflammation and cell signaling"] = "Inflamm Cell Signal",
  ["inflammation and regeneration"] = "Inflamm Regen",
  ["inflammation research"] = "Inflamm. Res.",
  ["inflammatory bowel diseases"] = "Inflamm Bowel Dis",
  ["inflammatory intestinal diseases"] = "Inflamm Intest Dis",
  ["influenza and other respiratory viruses"] = "Influenza Other Respir Viruses",
  ["influenza research and treatment"] = "Influenza Res Treat",
  ["info trends : medicine, law & ethics"] = "Info Trends",
  ["info trends : medicine, law & ethics / university of medicine and dentistry of new jersey-robert wood johnson medical school"] = "Info Trends",
  ["infocare : information strategies for healthcare networks"] = "Infocare",
  ["infor. information systems and operational research"] = "INFOR Inf. Syst. Oper. Res.",
  ["infor: information systems and operational research"] = "INFOR: Inf. Syst. Oper. Res.",
  ["inform (champaign, ill.)"] = "Inform (Champaign)",
  ["inform (silver spring, md.)"] = "Inform",
  ["informacion comercial española revista de economia"] = "Info. Comercial Española Revista Econ.",
  ["informaciones medicas"] = "Inf Medicas",
  ["informaciones médicas"] = "Inf Medicas",
  ["información sobre enfermedades venéreas"] = "Inf Sobre Enferm Vener",
  ["informação psiquiátrica : ip : órgão oficial do serviço de psiquiatria e psicologia médica do hospital de clínicas da uerj"] = "Inf Psiquiatr",
  ["informatica. an international journal of computing and informatics"] = "Informatica (Ljubl.)",
  ["informatics (mdpi)"] = "Informatics (MDPI)",
  ["informatics for health & social care"] = "Inform Health Soc Care",
  ["informatics for health and social care"] = "Inf. Health Social Care",
  ["informatics in medicine unlocked"] = "Inform Med Unlocked",
  ["informatics in primary care"] = "Inform Prim Care",
  ["informatik spektrum"] = "Inf. Spektrum",
  ["information & management"] = "Inf. Manage.",
  ["information (basel)"] = "Information (Basel)",
  ["information and behavior"] = "Inf Behav",
  ["information and computation"] = "Inform. and Comput.",
  ["information and control"] = "Inform. and Control (Shenyang)",
  ["information and inference : a journal of the ima"] = "Inf inference",
  ["information and inference. a journal of the ima"] = "Inf. Inference",
  ["information and inference: a journal of the ima"] = "Inf. Inference J. IMA",
  ["information and organization"] = "Inf. Organ.",
  ["information and software technology"] = "Inf Softw Technol",
  ["information bulletin"] = "Inform. Bull.",
  ["information dentaire"] = "Inf. Dent.",
  ["information discovery and delivery"] = "Inf Discov Deliv",
  ["information display"] = "Inf Disp (1975)",
  ["information economics and policy"] = "Info. Econ. Pol.",
  ["information fusion"] = "Inf. Fusion",
  ["information fusion and data science"] = "Inf. Fusion Data Sci.",
  ["information geometry"] = "Inf. Geom.",
  ["information in contemporary society : 14th international conference, iconference 2019, washington, dc, usa, march 31-april 3, 2019, proceedings. iconference (conference) (14th : 2019 : washington, d.c.)"] = "Inf Contemp Soc (2019)",
  ["information management"] = "Inf. Manage.",
  ["information management (ptn publishing corporation)"] = "Inf Manage",
  ["information privacy"] = "Inf Priv",
  ["information processing & management"] = "Inf Process Manag",
  ["information processing in computer-assisted interventions : third international conference, ipcai 2012, pisa, italy, june 27, 2012 proceedings. ipcai (conference) (3rd : 2012 : pisa, italy)"] = "Inf Process Comput Assist Interv (2012)",
  ["information processing letters"] = "Inform. Process. Lett.",
  ["information processing society of japan"] = "Trans. Inform. Process. Soc. Japan",
  ["information professional"] = "Inf. Prof.",
  ["information psychiatrique"] = "Inf. Psychiatr.",
  ["information report northern forest research centre"] = "Inf. Rep. North. For. Res. Cent.",
  ["information retrieval"] = "Inf Retr Boston",
  ["information retrieval journal"] = "Inf. Retr. J.",
  ["information reuse and integration (iri), ieee international conference on"] = "Inf Reuse Integr IRI IEEE Int Conf",
  ["information sciences"] = "Inf Sci (N Y)",
  ["information sciences. an international journal"] = "Inform. Sci.",
  ["information security and computer fraud"] = "Inf Secur Comput Fraud",
  ["information security and cryptography"] = "Inf. Secur. Cryptography",
  ["information security applications : 14th international workshop, wisa 2013, jeju island, korea, august 19-21, 2013 : revised selected papers. wisa (workshop : information security) (14th : 2013 : cheju-do, korea)"] = "Inform Secur Appl (2013)",
  ["information security journal: a global perspective"] = "Inf. Secur. J.: Global Perspect.",
  ["information services & use"] = "Inf Serv Use",
  ["information summary (international clearinghouse on adolescent fertility)"] = "Inf Summ",
  ["information systems"] = "Inf Syst",
  ["information systems frontiers"] = "Inf. Syst. Front.",
  ["information systems frontiers : a journal of research and innovation"] = "Inf Syst Front",
  ["information systems journal"] = "Inf. Syst. J.",
  ["information systems management"] = "Inf. Syst. Manage.",
  ["information systems research"] = "Inf. Syst. Res.",
  ["information systems research : isr"] = "Inf Syst Res",
  ["information technology and management"] = "Inf. Technol. Manage.",
  ["information technology for development"] = "Inf Technol Dev",
  ["information visualization"] = "Inf Vis",
  ["information zürcher wald"] = "Inf. Zür. Wald",
  ["information, communication & society"] = "Inf., Commun. Soc.",
  ["information, communication and society"] = "Inf Commun Soc",
  ["information, knowledge, systems management"] = "Inf Knowl Syst Manage",
  ["informationen aus orthodontie und kieferorthopadie : mit beitragen aus der internationalen literatur"] = "Inf Orthod Kieferorthop",
  ["informationen aus orthodontie und kieferorthopadie mit beitragen aus der internationalen literatur"] = "Inf. Orthod. Kieferorthop.",
  ["informationen aus orthodontie und kieferorthopädie : mit beiträgen aus der internationalen literatur"] = "Inf Orthod Kieferorthop",
  ["informationen zu naturschutz und landschaftspflege in nordwestdeutschland"] = "Inf. Nat.schutz Landsch.pfl. Nordwestdtschl.",
  ["informationen zur raumentwicklung"] = "Inf Raumentwickl",
  ["informationsberichte des bayerischen landesamtes für wasserwirtschaft"] = "Inf.ber.  Bayer.  Landesamtes  Wasserwirtsch.",
  ["informationsblatt des forschungsbereiches landschaft"] = "Inf.bl. Forsch.bereiches Landsch.",
  ["informationsblatt des forschungsbereiches landschaftsökologie"] = "Inf.bl. Forsch.bereiches Landsch.ökol.",
  ["informationsdienst naturschutz niedersachsen"] = "Inf.dienst Nat.schutz Niedersachs.",
  ["informationstechnik und datenverarbeitung"] = "Informationstech. Datenverarb.",
  ["informatore botanico italiano"] = "Inf. bot. ital.",
  ["informatore fitopatologico"] = "Inf Fitopatol",
  ["informatore odonto-stomatologico"] = "Inf Odontostomatol",
  ["informe demografico (sao paulo (brazil : state). fundacao sistema estadual de analise de dados)"] = "Inf Demogr",
  ["informe demográfico (são paulo (brazil : state). fundação sistema estadual de análise de dados)"] = "Inf Demogr",
  ["informing science"] = "Inf Sci",
  ["informing science and it education conference : proceedings and schedule. informing science and it education conference"] = "Inf Sci IT Educ Conf",
  ["informs journal on computing"] = "INFORMS J. Comput.",
  ["informs journal on optimization"] = "INFORMS J. Optim.",
  ["infortunistica e traumatologia del lavoro"] = "Infort Traumatol Lav",
  ["inforum (adelaide, s. aust.)"] = "Inforum",
  ["infosys science foundation series"] = "Infosys Sci. Found. Ser.",
  ["infosys science foundation series in mathematical sciences"] = "Infosys Sci. Found. Ser. Math. Sci.",
  ["infrared physics"] = "Infrared Phys.",
  ["infrared physics & technology"] = "Infrared Phys Technol",
  ["infrared physics and technology"] = "Infrared Phys. Technol.",
  ["infrastructure complexity"] = "Infrastruct. Complexity",
  ["infusionstherapie (basel, switzerland)"] = "Infusionstherapie",
  ["infusionstherapie und klinische ernahrung"] = "Infusionsther. Klin. Ernahr.",
  ["infusionstherapie und klinische ernahrung. sonderheft"] = "Infusionsther. Klin. Ernahr. Sonderh.",
  ["infusionstherapie und klinische ernährung"] = "Infusionsther Klin Ernahr",
  ["infusionstherapie und klinische ernährung. sonderheft"] = "Infusionsther Klin Ernahr Sonderh",
  ["infusionstherapie und transfusionsmedizin"] = "Infusionsther Transfusionsmed",
  ["ingeniera quimica (madrid)"] = "Ing. Quim.",
  ["ingenieria quimica (madrid)"] = "Ing. Quim.",
  ["ingenieria y competitividad"] = "Ing. Compet.",
  ["ingeniería y ciencia"] = "Ing Cienc",
  ["ingenierã­a del agua"] = "Ing. agua",
  ["ingenierã­a, investigaciã³n y tecnologã­a"] = "Ing. Invest. Tecnol.",
  ["ingenieur archiv"] = "Ing. Arch.",
  ["ingenieurbiologie mitteilungsblatt"] = "Ing.biol. Mitt.bl.",
  ["ingénierie et recherche biomédicale : irbm = biomedical engineering and research"] = "Ing Rech Biomed",
  ["inhalation toxicology"] = "Inhal Toxicol",
  ["inhaled particles"] = "Inhaled Part",
  ["initiatives in population"] = "Initiatives Popul",
  ["initiatives in reproductive health policy"] = "Initiat Reprod Health Policy",
  ["injury control and safety promotion"] = "Inj Control Saf Promot",
  ["injury epidemiology"] = "Inj Epidemiol",
  ["injury prevention"] = "Inj. Prev.",
  ["injury prevention : journal of the international society for child and adolescent injury prevention"] = "Inj Prev",
  ["inland seas"] = "Inland Seas",
  ["inland water biology"] = "Inland Water Biol",
  ["inland waters"] = "Inland Waters",
  ["inland waters : journal of the international society of limnology"] = "Inland Waters",
  ["inmunología (barcelona, spain : 1987)"] = "Inmunologia",
  ["innate immunity"] = "Innate Immun",
  ["innere medizin"] = "Inn Med",
  ["innovate (north miami beach, fla.)"] = "Innovate (North Miami Beach)",
  ["innovation (abingdon, england)"] = "Innovation (Abingdon)",
  ["innovation (new york, n.y.)"] = "Innovation (N Y)",
  ["innovation (north sydney, n.s.w.)"] = "Innovation (North Syd)",
  ["innovation and emerging technologies"] = "Innovation Emerging Technol.",
  ["innovation and entrepreneurship in health"] = "Innov Entrep Health",
  ["innovation and green development"] = "Innovation Green Dev.",
  ["innovation and research in clinical services, community support, and rehabilitation"] = "Innov Res Clin Serv Community Support Rehabil",
  ["innovation in aging"] = "Innov Aging",
  ["innovations (philadelphia, pa.)"] = "Innovations (Phila)",
  ["innovations in applied mathematics"] = "Innov. Appl. Math.",
  ["innovations in clinical neuroscience"] = "Innov Clin Neurosci",
  ["innovations in graph theory"] = "Innov. Graph Theory",
  ["innovations in incidence geometry â€” algebraic, topological and combinatorial"] = "Innovations Incidence Geom.- Algebraic Topol. Comb.",
  ["innovations in incidence geometry. algebraic, topological and combinatorial"] = "Innov. Incidence Geom.",
  ["innovations in oncology nursing"] = "Innov Oncol Nurs",
  ["innovations in pharmacy"] = "Innov Pharm",
  ["innovations in physics: x"] = "Innovations Phys: X",
  ["innovations in systems and software engineering"] = "Innov Syst Softw Eng",
  ["innovative biosystems and bioengineering"] = "Innovative Biosyst. Bioeng.",
  ["innovative food science & emerging technologies"] = "Innovative Food Sci. Emerg. Technol.",
  ["innovative food science & emerging technologies : ifset : the official scientific journal of the european federation of food science and technology"] = "Innov Food Sci Emerg Technol",
  ["innovative higher education"] = "Innov High Educ",
  ["innovative infrastructure solutions"] = "Innovative Infrastruct. Solutions",
  ["innovative romanian food biotechnology"] = "Innov Rom Food Biotechnol",
  ["innovative surgical sciences"] = "Innov Surg Sci",
  ["inorganic and nano-metal chemistry"] = "Inorg. Nano-Metal Chem.",
  ["inorganic and nuclear chemistry letters"] = "Inorg. Nucl. Chem. Lett.",
  ["inorganic chemistry"] = "Inorg Chem",
  ["inorganic chemistry (washington, dc, united states)"] = "Inorg. Chem. (Washington, DC, U. S.)",
  ["inorganic chemistry communications"] = "Inorg. Chem. Commun.",
  ["inorganic chemistry frontiers"] = "Inorg. Chem. Front.",
  ["inorganic materials"] = "Inorg. Mater.",
  ["inorganic materials (ussr) [translation of izvestiya akademii nauk sssr, neorganicheskie materialy]"] = "Inorg. Mater. (USSR)",
  ["inorganic materials: applied research"] = "Inorg. Mater. Appl. Res.",
  ["inorganic reaction mechanisms"] = "Inorg. React. Mech.",
  ["inorganic reaction mechanisms (philadelphia, pa, united states)"] = "Inorg. React. Mech. (Philadelphia, PA, U. S.)",
  ["inorganic syntheses"] = "Inorg. Synth.",
  ["inorganica chimica acta"] = "Inorg. Chim. Acta",
  ["inorganica chimica acta reviews"] = "Inorg. Chim. Acta Rev.",
  ["inozemna filologìâ (philologie étrangère)"] = "InFil",
  ["inquiry (oslo, norway)"] = "Inquiry (Oslo)",
  ["inquiry : a journal of medical care organization, provision and financing"] = "Inquiry",
  ["ins reporter"] = "INS Rep",
  ["ins reporter / immigration and naturalization service, u.s. department of justice"] = "INS Rep",
  ["inschriften griechischer städte aus kleinasien"] = "IK",
  ["inscriptiones graecae"] = "IG",
  ["inscriptiones graecae ad res romanas pertinentes"] = "IGR",
  ["insect biochemistry"] = "Insect Biochem",
  ["insect biochemistry and molecular biology"] = "Insect Biochem Mol Biol",
  ["insect conservation and diversity"] = "Insect Conserv Divers",
  ["insect molecular biology"] = "Insect Mol Biol",
  ["insect sci."] = "Insect Sci.",
  ["insect science"] = "Insect Sci",
  ["insect systematics & evolution"] = "Insect Syst Evol",
  ["insect systematics and diversity"] = "Insect Syst Divers",
  ["insecta matsumurana"] = "Insecta Matsumurana",
  ["insecta mundi"] = "Insecta mundi",
  ["insectes sociaux"] = "Insectes Soc.",
  ["inside the cell"] = "Inside Cell",
  ["insight (american society of ophthalmic registered nurses)"] = "Insight",
  ["insight (lawrence, kan.)"] = "Insight (Lawrence)",
  ["insight bioinformatics"] = "Insight Bioinforma",
  ["insights in allergy, asthma & bronchitis"] = "Insights Allergy Asthma Bronchitis",
  ["insights in biomedicine"] = "Insights Biomed",
  ["insights in cancer research"] = "Insights Cancer Res",
  ["insights in enzyme research"] = "Insights Enzym Res",
  ["insights in neurosurgery"] = "Insights Neurosurg",
  ["insights in nutrition and metabolism"] = "Insights Nutr Metab",
  ["insights into imaging"] = "Insights Imaging",
  ["insights of biomedical research"] = "Insights Biomed Res",
  ["institut des belles lettres arabes : ibla"] = "IBLA",
  ["institut des hautes études scientifiques"] = "Inst. Hautes études Sci. Publ. Math.",
  ["institut fur angewandte wirtschaftsforschung tubingen schriftenreihe"] = "Inst. Angew. Wirtschaftsforsch. Tubingen Schriftenreihe",
  ["institut mathematique. publications. nouvelle serie"] = "Publ. Inst. Math. (Beograd) (N.S.)",
  ["institut mathématique"] = "Publ. Inst. Math. (Beograd) (N.S.)",
  ["institut élie cartan"] = "Inst. Élie Cartan",
  ["institute for nonlinear science"] = "Inst. Nonlinear Sci.",
  ["institute of electrical and electronics engineers"] = "IEEE Trans. Antennas and Propagation",
  ["institute of electrical and electronics engineers. transactions on antennas and propagation"] = "IEEE Trans. Antennas and Propagation",
  ["institute of electrical and electronics engineers. transactions on automatic control"] = "IEEE Trans. Automat. Control",
  ["institute of electrical and electronics engineers. transactions on computers"] = "IEEE Trans. Comput.",
  ["institute of electrical and electronics engineers. transactions on information theory"] = "IEEE Trans. Inform. Theory",
  ["institute of mathematical sciences lecture notes"] = "Inst. Math. Sci. Lect. Notes",
  ["institute of mathematical statistics (ims) collections"] = "Inst. Math. Stat. (IMS) Collect.",
  ["institute of mathematical statistics (ims) monographs"] = "Inst. Math. Stat. Monogr.",
  ["institute of mathematical statistics collections"] = "Inst Math Stat Collect",
  ["institute of mathematical statistics lecture notes—monograph series"] = "IMS Lecture Notes Monogr. Ser.",
  ["institute of mathematical statistics textbooks"] = "IMS Textb.",
  ["institute of mathematics & computer sciences"] = "J. Inst. Math. Comput. Sci. Comput. Sci. Ser.",
  ["institute of mathematics & computer sciences. journal. (computer science series)"] = "J. Inst. Math. Comput. Sci. Comput. Sci. Ser.",
  ["institute of systems, control and information engineers"] = "Trans. Inst. Systems Control Inform. Engrs.",
  ["institute of transportation engineers. ite journal"] = "Inst. Transp. Eng. ITE J.",
  ["institution of civil engineers construction law quarterly"] = "Inst. Civ. Eng. Constr. Law Q.",
  ["institutionalised children exploration and beyond"] = "Inst Child Explor Beyond",
  ["institutionen für internationell pedagogik"] = "Inst. Int. Pedagog. Report",
  ["institutions/volume feeding"] = "Inst Vol Feed",
  ["instituto nacional de estadística"] = "Estadíst. Española",
  ["instraw news : women and development"] = "INSTRAW News",
  ["instructional course lectures"] = "Instr Course Lect",
  ["instructional science"] = "Instr Sci",
  ["instrumentation science & technology"] = "Instrum Sci Technol",
  ["instrumentation science and technology"] = "Instrum Sci. Technol.",
  ["instruments and experiment techniques"] = "Instrum. Exp. Tech.",
  ["instruments and experimental techniques"] = "Instrum. Exp. Tech.",
  ["instruments and experimental techniques (ussr) [translation of pribory i tekhnika eksperimenta]"] = "Instrum. Exp. Tech. (USSR)",
  ["insula fulcheria"] = "InsFulc",
  ["insurance counsel journal"] = "Insur Couns J",
  ["insurance law journal (chicago, ill.)"] = "Insur Law J",
  ["insurance mathematics and economics"] = "Insur. Math. Econ.",
  ["insurance, mathematics & economics"] = "Insur Math Econ",
  ["insurance: mathematics & economics"] = "Insurance Math. Econom.",
  ["insurance: mathematics and economics"] = "Ins.: Mathematics Econ.",
  ["inted proceedings"] = "INTED Proc",
  ["integers : electronic journal of combinatorial number theory"] = "Integers",
  ["integers. electronic journal of combinatorial number theory"] = "Integers",
  ["integral equations and operator theory"] = "Integral Equations Operator Theory",
  ["integral transforms and special functions"] = "Integral Transform Spec Funct",
  ["integral transforms and special functions. an international journal"] = "Integral Transforms Spec. Funct.",
  ["integrated assessment"] = "Integr. Assess.",
  ["integrated blood pressure control"] = "Integr Blood Press Control",
  ["integrated computer-aided engineering"] = "Integr Comput Aided Eng",
  ["integrated environmental assessment and manaement"] = "Integr. Environ. Assess. Manage.",
  ["integrated environmental assessment and management"] = "Integr Environ Assess Manag",
  ["integrated ferroelectrics"] = "Integr. Ferroelectr.",
  ["integrated ferroelectronics"] = "Integr. Ferroelectr.",
  ["integrated healthcare report"] = "Integr Healthc Rep",
  ["integrated manufacturing systems"] = "Integr. Manuf. Syst.",
  ["integrated pest management reviews"] = "Integr. Pest Manage. Rev.",
  ["integrated pharmacy research & practice"] = "Integr Pharm Res Pract",
  ["integrated pharmacy research and practice"] = "Integr. Pharm. Res. Pract.",
  ["integrated series in information systems"] = "Integr. Ser. Inf. Syst.",
  ["integrating materials and manufacturing innovation"] = "Integr Mater Manuf Innov",
  ["integration (amsterdam, netherlands)"] = "Integration (Amst)",
  ["integration (tokyo, japan)"] = "Integration",
  ["integration and trade"] = "Integration Trade",
  ["integrative and comparative biology"] = "Integr Comp Biol",
  ["integrative biology"] = "Integr. Biol.",
  ["integrative biology : quantitative biosciences from nano to macro"] = "Integr Biol (Camb)",
  ["integrative cancer biology & research"] = "Integr Cancer Biol Res",
  ["integrative cancer science and therapeutics"] = "Integr Cancer Sci Ther",
  ["integrative cancer therapies"] = "Integr Cancer Ther",
  ["integrative clinical medicine"] = "Integr Clin Med",
  ["integrative conservation"] = "Integr. Converv.",
  ["integrative food, nutrition and metabolism"] = "Integr Food Nutr Metab",
  ["integrative gynecology and obstetrics journal"] = "Integr Gynecol Obstet J",
  ["integrative medicine (encinitas, calif.)"] = "Integr Med (Encinitas)",
  ["integrative medicine : integrating conventional and alternative medicine"] = "Integr Med",
  ["integrative medicine insights"] = "Integr Med Insights",
  ["integrative medicine international"] = "Integr Med Int",
  ["integrative medicine research"] = "Integr Med Res",
  ["integrative molecular medicine"] = "Integr Mol Med",
  ["integrative obesity and diabetes"] = "Integr Obes Diabetes",
  ["integrative organismal biology"] = "Integr. Org. Biol.",
  ["integrative organismal biology (oxford, england)"] = "Integr Org Biol",
  ["integrative physiological and behavioral science"] = "Integr. Physiol. Behav. Sci.",
  ["integrative physiological and behavioral science : the official journal of the pavlovian society"] = "Integr Physiol Behav Sci",
  ["integrative psychiatry : ip"] = "Integr Psychiatry",
  ["integrative psychological & behavioral science"] = "Integr Psychol Behav Sci",
  ["integrative systematics : stuttgart contributions to natural history"] = "Integr Syst",
  ["integrative systematics: stuttgart contributions to natural history"] = "Integr. Syst.: Stuttgart Contrib. Nat. Hist.",
  ["integrative zoology"] = "Integr Zool",
  ["intellectual and developmental disabilities"] = "Intellect Dev Disabil",
  ["intellectual digest"] = "Intellect Dig",
  ["intellectual history of the islamicate world"] = "Intellect Hist Islam World",
  ["intellectual history review"] = "Intellect Hist Rev",
  ["intellectual property journal"] = "Intellect Prop J",
  ["intelligence science and big data engineering : 4th international conference, iscide 2013, beijing, china, july 31-august 2, 2013, revised selected papers. iscide (4th : 2013 : beijing, china)"] = "Intell Sci Big Data Eng (2013)",
  ["intelligence-based medicine"] = "Intell Based Med",
  ["intelligent and converged networks"] = "Intell. Converged Networks",
  ["intelligent automation & soft computing"] = "Intell. Autom. Soft Comput.",
  ["intelligent buildings international"] = "Intell. Build. Int.",
  ["intelligent computing"] = "Intell. Comput.",
  ["intelligent control and learning systems"] = "Intell. Control Learn. Syst.",
  ["intelligent data analysis"] = "Intell. Data Anal.",
  ["intelligent decision technologies"] = "Intell. Decis. Technol.",
  ["intelligent industrial systems"] = "Intell. Ind. Syst.",
  ["intelligent information management"] = "Intell Inf Manag",
  ["intelligent medicine"] = "Intell. Med.",
  ["intelligent robotics and autonomous agents"] = "Intel. Robotics Auton. Agents",
  ["intelligent robotics atonomous agents"] = "Intel. Robotics Aton. Agents",
  ["intelligent robotics autonomous agents"] = "Intel. Robotics Auton. Agent",
  ["intelligent service robotics"] = "Intell Serv Robot",
  ["intelligent surgery"] = "Intell. Surg.",
  ["intelligent systems engineering"] = "Intell. Syst. Eng.",
  ["intelligent systems reference library"] = "Intell. Syst. Ref. Libr.",
  ["intelligent systems with applications"] = "Intell. Syst. Appl.",
  ["intelligent systems, control and automation: science and engineering"] = "Intell. Syst. Control Autom. Sci. Eng.",
  ["intensive & critical care nursing"] = "Intensive Crit Care Nurs",
  ["intensive & critical care nursing : the official journal of the british association of critical care nurses"] = "Intensive Crit Care Nurs",
  ["intensive and critical care nursing"] = "Intensive Crit. Care Nurs.",
  ["intensive care medicine"] = "Intensive Care Med",
  ["intensive care medicine experimental"] = "Intensive Care Med Exp",
  ["intensive care nursing"] = "Intensive Care Nurs",
  ["intensive care world"] = "Intensive Care World",
  ["intensivmedizin + notfallmedizin : organ der deutschen und der österreichischen gesellschaft für internistische intensivmedizin, der sektion neurologie der dgim und der sektion intensivmedizin im berufsverband deutscher internisten e.v"] = "Intensivmed Notfallmed",
  ["intensivmedizinische praxis"] = "Intensivmed Prax",
  ["inter economics"] = "Inter Econ",
  ["inter-american economic affairs"] = "Inter Am Econ Aff",
  ["inter-university faculty work conference"] = "Interuniv Fac Work Conf",
  ["interacting with computers"] = "Interact Comput",
  ["interaction of mechanics and mathematics"] = "Interact. Mech. Math.",
  ["interaction studies"] = "Interact Stud",
  ["interactions (new york, n.y.)"] = "Interactions (N Y)",
  ["interactive cardiovascular and thoracic surgery"] = "Interact Cardiovasc Thorac Surg",
  ["interactive journal of medical research"] = "Interact J Med Res",
  ["intercambio urológico por correspondencia"] = "Intercamb Urol Corresp",
  ["interceram - international ceramic review"] = "Interceram - Int. Ceram. Rev.",
  ["interchange (toronto, ont. : 1984)"] = "Interchange (Tor : 1984)",
  ["intercom; the international newsletter on population and family planning"] = "Intercom",
  ["interdisciplinaria archaeologica"] = "Interdiscip Archaeol",
  ["interdisciplinary applied mathematics"] = "Interdiscip. Appl. Math.",
  ["interdisciplinary contributions to archaeology"] = "Interdiscip. Contrib. Archaeol.",
  ["interdisciplinary environmental review"] = "Interdiscip. Environ. Rev.",
  ["interdisciplinary information sciences"] = "Interdiscip. Inform. Sci.",
  ["interdisciplinary journal of best practices in global development"] = "Interdiscip J Best Pract Glob Dev",
  ["interdisciplinary journal of partnership studies"] = "Interdiscip J Partnersh Stud",
  ["interdisciplinary journal of research on religion"] = "Interdiscip J Res Relig",
  ["interdisciplinary journal of sustainable oceans and coasts"] = "Interdiscip. J. Sustainable Oceans Coasts",
  ["interdisciplinary materials"] = "Interdiscip. Mater.",
  ["interdisciplinary mathematical sciences"] = "Interdiscip. Math. Sci.",
  ["interdisciplinary mathematics"] = "Interdiscip. Math.",
  ["interdisciplinary neurosurgery : advanced techniques and case management"] = "Interdiscip Neurosurg",
  ["interdisciplinary neurosurgery: advanced techniques and case management"] = "Interdiscip. Neurosurg. Adv. Tech. Case Manage.",
  ["interdisciplinary perspectives on infectious diseases"] = "Interdiscip Perspect Infect Dis",
  ["interdisciplinary science reviews"] = "Interdiscip. Sci. Rev.",
  ["interdisciplinary science reviews : isr"] = "Interdiscip Sci Rev",
  ["interdisciplinary sciences, computational life sciences"] = "Interdiscip Sci",
  ["interdisciplinary sciences-computational life sciences"] = "Interdiscip. Sci.-Comput. Life Sci.",
  ["interdisciplinary sciences: computational life sciences"] = "Interdiscip. Sci.: Comput. Life Sci.",
  ["interdisciplinary statistics"] = "Interdiscip. Statist.",
  ["interdisciplinary studies in economics and management"] = "Interdiscip. Stud. Econ. Manag.",
  ["interdisciplinary studies on the nature of mathematics"] = "Interdiscip. Stud. Nat. Math.",
  ["interdisciplinary systems research"] = "Interdiscip. Systems Res.",
  ["interdisciplinary topics in gerontology"] = "Interdiscip Top Gerontol",
  ["interdisciplinary topics in gerontology and geriatrics"] = "Interdiscip Top Gerontol Geriatr",
  ["interdisciplinary toxicology"] = "Interdiscip Toxicol",
  ["interface : a journal for and about social movements"] = "Interface (Maynooth)",
  ["interface focus"] = "Interface Focus",
  ["interface science"] = "Interface Sci.",
  ["interfaces and free boundaries"] = "Interfaces Free Boundaries",
  ["interfaces and free boundaries. mathematical analysis, computation and applications"] = "Interfaces Free Bound.",
  ["interfacial phenomena and heat transfer"] = "Interfacial Phenom Heat Transf",
  ["interior design"] = "Inter. Des.",
  ["interior design (new york, n.y.)"] = "Inter Des",
  ["interiors (new york, n.y. : 1978)"] = "Interiors",
  ["intermountain economic review"] = "Intermountain Econ. Rev.",
  ["intermountain journal of sciences : ijs"] = "Intermt J Sci",
  ["internal and emergency medicine"] = "Intern Emerg Med",
  ["internal medicine"] = "Intern. Med.",
  ["internal medicine (tokyo, japan)"] = "Intern Med",
  ["internal medicine journal"] = "Intern Med J",
  ["internal medicine review (washington, d.c. : online)"] = "Intern Med Rev (Wash D C)",
  ["internal medicine: open access"] = "Intern Med Open Access",
  ["international abstracts of surgery"] = "Int Abstr Surg",
  ["international advances in economic research"] = "Int. Advances Econ. Res.",
  ["international advances in surgical oncology"] = "Int Adv Surg Oncol",
  ["international affairs"] = "Int Aff",
  ["international agrophysics"] = "Int. Agrophys.",
  ["international anesthesiology clinics"] = "Int Anesthesiol Clin",
  ["international angiology"] = "Int. Angiol.",
  ["international angiology : a journal of the international union of angiology"] = "Int Angiol",
  ["international annual conference of ict"] = "Int. Annu. Conf. ICT",
  ["international applied mechanics"] = "Int. Appl. Mech.",
  ["international aquatic research"] = "Int. Aquat. Res,",
  ["international archives of allergy and applied immunology"] = "Int Arch Allergy Appl Immunol",
  ["international archives of allergy and immunology"] = "Int Arch Allergy Immunol",
  ["international archives of medicine"] = "Int Arch Med",
  ["international archives of nursing and health care"] = "Int Arch Nurs Health Care",
  ["international archives of occupational and environmental health"] = "Int Arch Occup Environ Health",
  ["international archives of otorhinolaryngology"] = "Int Arch Otorhinolaryngol",
  ["international archives of photogrammetry and remote sensing"] = "Int. Arch. Photogramm. Remote Sens.",
  ["international archives of the history of ideas/archives internationales d’histoire des idees"] = "Int. Arch. Hist. Ideas",
  ["international archives of translational medicine"] = "Int Arch Transl Med",
  ["international association for mathematical geosciences. studies in mathematical geosciences"] = "Int. Assoc. Math. Geosci. Stud. Math. Geosci.",
  ["international astronomical union colloquium"] = "Int. Astron. Union Colloq.",
  ["international bibliography of historical sciences"] = "Int Bibliogr Hist Sci",
  ["international biodeterioration & biodegradation"] = "Int Biodeterior Biodegradation",
  ["international biodeterioration and biodegradation"] = "Int. Biodeterior. Biodegrad.",
  ["international biological and biomedical journal"] = "Int Biol Biomed J",
  ["international biology review"] = "Int Biol Rev",
  ["international biomechanics"] = "Int Biomech",
  ["international braz j urol"] = "Int. Braz. J. Urol.",
  ["international braz j urol : official journal of the brazilian society of urology"] = "Int Braz J Urol",
  ["international breastfeeding journal"] = "Int Breastfeed J",
  ["international business review (oxford, england)"] = "Int Bus Rev",
  ["international cancer conference journal"] = "Int Cancer Conf J",
  ["international cardiovascular research journal"] = "Int Cardiovasc Res J",
  ["international child health"] = "Int Child Health",
  ["international clinical pathology journal"] = "Int Clin Pathol J",
  ["international clinical psychopharmacology"] = "Int Clin Psychopharmacol",
  ["international communication gazette"] = "Int Commun Gaz",
  ["international communications in heat and mass transfer"] = "Int. Commun. Heat Mass Transfer",
  ["international conference on adaptive structures and technologies : [proceedings]. international conference on adaptive structures and technologies"] = "Int Conf Adapt Struct Technol",
  ["international conference on affective computing and intelligent interaction and workshops : [proceedings]. acii (conference)"] = "Int Conf Affect Comput Intell Interact Workshops",
  ["international conference on aids"] = "Int Conf AIDS",
  ["international conference on auditory-visual speech processing"] = "Int Conf Audit Vis Speech Process",
  ["international conference on bio-inspired computing, theories and applications : [proceedings]. international conference on bio-inspired computing, theories and applications"] = "Int Conf Bioinspired Comput Theor Appl",
  ["international conference on bioinformatics and biomedical engineering : [proceedings]. international conference on bioinformatics and biomedical engineering"] = "Int Conf Bioinform Biomed Eng",
  ["international conference on collaborative computing : networking, applications and worksharing (collaboratecom). international conference on collaborative computing: networking, applications, and worksharing"] = "Int Conf Collab Comput",
  ["international conference on communication systems and networks : [proceedings]. international conference on communication systems and networks"] = "Int Conf Commun Syst Netw",
  ["international conference on computational & mathematical biomedical engineering"] = "Int Conf Comput Math Biomed Eng",
  ["international conference on computing, networking, and communications : [proceedings]. international conference on computing, networking and communications"] = "Int Conf Comput Netw Commun",
  ["international conference on dependable systems and networks workshops : [proceedings]. international conference on dependable systems and networks"] = "Int Conf Dependable Syst Netw Workshops",
  ["international conference on digital signal processing proceedings : dsp. international conference on digital signal processing"] = "Int Conf Digit Signal Process Proc",
  ["international conference on electromagnetics in advanced applications : proceedings : iceaa. international conference on electromagnetics in advanced applications"] = "Int Conf Electromagn Adv Appl",
  ["international conference on environmental systems. american institute of aeronautics and astronautics"] = "Int Conf Environ Syst",
  ["international conference on geoinformatics : [proceedings]. international conference on geoinformatics"] = "Int Conf Geoinform",
  ["international conference on giscience short paper proceedings"] = "Int Conf GISci Short Pap Proc",
  ["international conference on infrared, millimeter, and terahertz waves : [proceedings]. international conference on infrared, millimeter, and terahertz waves"] = "Int Conf Infrared Millim Terahertz Waves",
  ["international conference on pattern recognition"] = "Int Conf Pattern Recognit",
  ["international conference on pervasive computing technologies for healthcare : [proceedings]. international conference on pervasive computing technologies for healthcare"] = "Int Conf Pervasive Comput Technol Healthc",
  ["international conference on scientific and statistical database management : [proceedings]. international conference on scientific and statistical database management"] = "Int Conf Sci Stat Database Manag",
  ["international conference on signal processing & communications (spcom). conference on signal processing, and communications"] = "Int Conf Signal Process Commun",
  ["international conference on signal processing proceedings. international conference on signal processing"] = "Int Conf Signal Process Proc",
  ["international congress series"] = "Int Congr Ser",
  ["international congress series - excerpta medica"] = "Int. Congr. Ser. - Excerpta Med.",
  ["international congress series / excerpta medica"] = "Int Congr Ser",
  ["international contact lens clinic (new york, n.y.)"] = "Int Contact Lens Clin",
  ["international contributions to labour studies"] = "Int. Contrib. to Lab. Stud.",
  ["international corrosion conference series"] = "Int. Corros. Conf. Ser.",
  ["international criminal justice review"] = "Int Crim Justice Rev",
  ["international dairy journal"] = "Int. Dairy J.",
  ["international data series, selected data on mixtures, series a"] = "Int. DATA Ser., Sel. Data Mixtures, Ser. A",
  ["international data series, selected data on mixtures, series a: thermodynamic properties of non-reacting binary systems of organic substances"] = "Int. DATA Ser., Sel. Data Mixtures, Ser. A",
  ["international demographics"] = "Int Demogr",
  ["international dental journal"] = "Int Dent J",
  ["international desalination & water reuse quarterly"] = "Int. Desalin. Water Reuse Q.",
  ["international development review"] = "Int Dev Rev",
  ["international digest of health legislation"] = "Int Dig Health Legis",
  ["international disability studies"] = "Int Disabil Stud",
  ["international drug discovery"] = "Int Drug Discov",
  ["international drug therapy newsletter"] = "Int Drug Ther Newsl",
  ["international economic insights"] = "Int Econ Insights",
  ["international economic journal"] = "Int Econ J",
  ["international economic review"] = "Int Econ Rev (Philadelphia)",
  ["international economics"] = "Int. Econ.",
  ["international economy"] = "Int. Economy",
  ["international education and research journal"] = "Int Educ Res J",
  ["international electronic journal of algebra"] = "Int. Electron. J. Algebra",
  ["international electronic journal of geometry"] = "Int. Electron. J. Geom.",
  ["international electronic journal of pure and applied mathematics"] = "Int. Electron. J. Pure Appl. Math.",
  ["international emergency nursing"] = "Int Emerg Nurs",
  ["international endodontic journal"] = "Int Endod J",
  ["international energy journal"] = "Int. Energy J.",
  ["international environmental agreements : politics, law and economics"] = "Int Environ Agreem",
  ["international family planning digest"] = "Int Fam Plann Dig",
  ["international family planning perspectives"] = "Int Fam Plan Perspect",
  ["international family planning perspectives and digest"] = "Int Fam Plann Persp",
  ["international feminist journal of politics"] = "Int Fem J Polit",
  ["international finance"] = "Int. Finance",
  ["international fire service journal of leadership and management"] = "Int Fire Serv J Leadersh Manag",
  ["international food and agribusiness management review"] = "Int. Food Agribus. Manage. Rev.",
  ["international food research journal"] = "Int Food Res J",
  ["international forest fire news"] = "Int. For. Fire News",
  ["international forestry review"] = "Int. For. Rev.",
  ["international forum of allergy & rhinology"] = "Int Forum Allergy Rhinol",
  ["international gambling studies"] = "Int Gambl Stud",
  ["international game theory review"] = "Int. Game Theory Rev.",
  ["international gazette (forest hill, md.)"] = "Int Gaz (For Hill)",
  ["international geology review"] = "Int. Geol. Rev.",
  ["international geophysics series"] = "Internat. Geophys. Ser.",
  ["international handbooks on information systems"] = "Int. Handb. Inf. Syst.",
  ["international health"] = "Int Health",
  ["international health and development"] = "Int Health Dev",
  ["international health history newsletter"] = "Int Health Hist Newsl",
  ["international health news"] = "Int Health News",
  ["international health news / national council for international health"] = "Int Health News",
  ["international heart journal"] = "Int Heart J",
  ["international heat treatment and surface engineering"] = "Int. Heat Treat. Surf. Eng.",
  ["international history of nursing journal : ihnj"] = "Int Hist Nurs J",
  ["international ieee/embs conference on neural engineering : [proceedings]. international ieee embs conference on neural engineering"] = "Int IEEE EMBS Conf Neural Eng",
  ["international immunology"] = "Int Immunol",
  ["international immunopharmacology"] = "Int Immunopharmacol",
  ["international indigenous policy journal"] = "Int Indig Policy J",
  ["international innovation : disseminating science, research and technology"] = "Int Innov",
  ["international journal (toronto, ont.)"] = "Int J",
  ["international journal for computational methods in engineering science and mechanics"] = "Int. J. Comput. Methods Eng. Sci. Mech.",
  ["international journal for computational vision and biomechanics"] = "Int J Comput Vis Biomech",
  ["international journal for educational and vocational guidance"] = "Int J Educ Vocat Guid",
  ["international journal for engineering modelling"] = "Int. J. Eng. Model.",
  ["international journal for equity in health"] = "Int J Equity Health",
  ["international journal for housing science and its applications"] = "Int. J. Hous. Sci. Appl.",
  ["international journal for human caring"] = "Int J Hum Caring",
  ["international journal for innovation education and research"] = "Int J Innov Educ Res",
  ["international journal for ion mobility spectrometry"] = "Int. J. Ion Mobility Spectrom.",
  ["international journal for ion mobility spectrometry : official publication of the international society for ion mobility spectrometry"] = "Int J Ion Mobil Spectrom",
  ["international journal for multiscale computational engineering"] = "Int. J. Multiscale Comput. Eng.",
  ["international journal for numerical and analytical methods in geomechanics"] = "Int. J. Numer. Anal. Methods Geomech.",
  ["international journal for numerical methods in biomedical engineering"] = "Int J Numer Method Biomed Eng",
  ["international journal for numerical methods in engineering"] = "Int. J. Numer. Methods Eng.",
  ["international journal for numerical methods in fluids"] = "Int. J. Numer. Methods Fluids",
  ["international journal for parasitology"] = "Int J Parasitol",
  ["international journal for parasitology. drugs and drug resistance"] = "Int J Parasitol Drugs Drug Resist",
  ["international journal for parasitology. parasites and wildlife"] = "Int J Parasitol Parasites Wildl",
  ["international journal for quality in health care"] = "Int. J. Qual. Health Care",
  ["international journal for quality in health care : journal of the international society for quality in health care"] = "Int J Qual Health Care",
  ["international journal for quality in health care : journal of the international society for quality in health care / isqua"] = "Int J Qual Health Care",
  ["international journal for radiation physics and chemistry"] = "Int. J. Radiat. Phys. Chem.",
  ["international journal for research in applied science and engineering technology"] = "Int J Res Appl Sci Eng Technol",
  ["international journal for research in learning disabilities"] = "Int J Res Learn Disabil",
  ["international journal for science and advance research in technology"] = "Int J Sci Adv Res Technol",
  ["international journal for simulation and multidisciplinary design optimization"] = "Int. J. Simul. Multi. Des. Optim.",
  ["international journal for the advancement of counseling"] = "Int J Adv Couns",
  ["international journal for the history of engineering & technology"] = "Int. J. Hist. Eng. Technol.",
  ["international journal for the semiotics of law = revue internationale de sémiotique juridique"] = "Int J Semiot Law",
  ["international journal for the study of animal problems"] = "Int J Study Anim Probl",
  ["international journal for uncertainty quantification"] = "Int J Uncertain Quantif",
  ["international journal for vitamin and nutrition research"] = "Int. J. Vitam. Nutr. Res.",
  ["international journal for vitamin and nutrition research. internationale zeitschrift fur vitamin- und ernahrungsforschung. journal international de vitaminologie et de nutrition"] = "Int J Vitam Nutr Res",
  ["international journal for vitamin and nutrition research. internationale zeitschrift für vitamin- und ernährungsforschung. journal international de vitaminologie et de nutrition"] = "Int J Vitam Nutr Res",
  ["international journal for vitamin and nutrition research. supplement"] = "Int. J. Vitam. Nutr. Res. Suppl.",
  ["international journal for vitamin and nutrition research. supplement = internationale zeitschrift für vitamin- und ernährungsforschung. supplement"] = "Int J Vitam Nutr Res Suppl",
  ["international journal of  environmental sustainability"] = "Int. J. Environ. Sustainability",
  ["international journal of abrasive technology"] = "Int. J. Abras. Technol.",
  ["international journal of academic medicine"] = "Int J Acad Med",
  ["international journal of academic research"] = "Int J Acad Res",
  ["international journal of academic research and development"] = "Int J Acad Res Dev",
  ["international journal of acarology"] = "Int J Acarol",
  ["international journal of acoustics and vibrations"] = "Int. J. Acoust. Vibr.",
  ["international journal of ad hoc and ubiquitous computing"] = "Int. J. Ad Hoc Ubiquitous Comput.",
  ["international journal of adaptive and innovative systems"] = "Int. J. Adapt. Innovative Syst.",
  ["international journal of adaptive control and signal processing"] = "Int. J. Adapt. Control Signal Process.",
  ["international journal of additive and subtractive materials manufacturing"] = "Int. J. Addit. Subtract. Manuf.",
  ["international journal of adhesion & adhesives"] = "Int J Adhes Adhes",
  ["international journal of adhesion and adhesives"] = "Int. J. Adhes. Adhes.",
  ["international journal of adolescence and youth"] = "Int J Adolesc Youth",
  ["international journal of adolescent medicine and health"] = "Int J Adolesc Med Health",
  ["international journal of adult orthodontics and orthognathic surgery"] = "Int. J. Adult Orthodon. Orthognath. Surg.",
  ["international journal of advance & innovative research"] = "Int J Adv Innov Res",
  ["international journal of advanced biotechnology and research"] = "Int J Adv Biotechnol Res",
  ["international journal of advanced computer science"] = "Int J  Adv Comput Sci",
  ["international journal of advanced computer science and applications : ijacsa"] = "Int J Adv Comput Sci Appl",
  ["international journal of advanced intelligence paradigms"] = "Int. J. Adv. Intell. Paradigms",
  ["international journal of advanced life sciences"] = "Int J Adv Life Sci (Coimbatore)",
  ["international journal of advanced logistics"] = "Int. J. Adv. Logist.",
  ["international journal of advanced manufacturing systems"] = "Int J Adv Manuf Syst",
  ["international journal of advanced manufacturing technology"] = "Int. J. Adv. Manuf. Technol.",
  ["international journal of advanced mechatronic systems"] = "Int. J. Adv. Mechatron. Syst.",
  ["international journal of advanced nuclear reactor design and technology"] = "Int. J. Adv. Nucl. Reactor Des. Technol.",
  ["international journal of advanced research"] = "Int J Adv Res (Indore)",
  ["international journal of advanced research in science, engineering and technology"] = "Int J Adv Res Sci Eng Technol",
  ["international journal of advanced robotic systems"] = "Int. J. Adv. Rob. Syst.",
  ["international journal of advanced structural engineering"] = "Int. J. Adv. Struct. Eng.",
  ["international journal of advancements in computing technology"] = "Int J Adv Comput Technol",
  ["international journal of advances in applied mathematics and mechanics"] = "Int. J. Adv. Appl. Math. Mech.",
  ["international journal of advances in engineering and technology"] = "Int J Adv Eng Technol",
  ["international journal of advances in engineering sciences and applied mathematics"] = "Int. J. Adv. Eng. Sci. Appl. Math.",
  ["international journal of advances in science, engineering and technology"] = "Int J Adv Sci Eng Technol",
  ["international journal of advertising"] = "Int J Advert",
  ["international journal of aeroacoustics"] = "Int J Aeroacoust",
  ["international journal of aerodynamics"] = "Int. J. Aerodyn.",
  ["international journal of aeronautical and space sciences"] = "Int. J. Aeronaut. Space Sci.",
  ["international journal of aerospace engineering"] = "Int. J. Aerosp. Eng.",
  ["international journal of aerospace psychology"] = "Int. J. Aerosp. Psychol.",
  ["international journal of aerospace system science and engineering"] = "Int. J. Aerosp. Syst. Sci. Eng.",
  ["international journal of africa nursing sciences"] = "Int J Afr Nurs Sci",
  ["international journal of agent technologies and systems"] = "Int J Agent Technol Syst",
  ["international journal of agent-oriented software engineering"] = "Int. J. Agent-Oriented Softw. Eng.",
  ["international journal of agile and extreme software development"] = "Int. J. Agile Extreme Software Dev.",
  ["international journal of agile manufacturing"] = "Int. J. Agile Manuf.",
  ["international journal of agile systems and management"] = "Int. J. Agile Syst. Manage.",
  ["international journal of aging & human development"] = "Int J Aging Hum Dev",
  ["international journal of aging and human development"] = "Int. J. Aging Hum. Dev.",
  ["international journal of agricultural and biological engineering"] = "Int. J. Agric. Biol. Eng.",
  ["international journal of agricultural resources, governance and ecology"] = "Int. J. Agric. Resour. Governance Ecol.",
  ["international journal of agricultural science and research (ijasr)"] = "Int J Agric Sci Res (Chennai)",
  ["international journal of agricultural sustainability"] = "Int J Agric Sustain",
  ["international journal of agriculture and biology"] = "Int J Agric Biol",
  ["international journal of agriculture and forestry (print)"] = "Int J Agric For",
  ["international journal of agriculture innovation, technology and globalisation"] = "Int. J. Agric. Innovation Technol. Globalisation",
  ["international journal of agriculture innovations and research"] = "Int J Agric Innov Res",
  ["international journal of agriscience"] = "Int J Agrisci",
  ["international journal of agronomy"] = "Int. J. Agron.",
  ["international journal of air and water pollution"] = "Int. J. Air Water Pollut.",
  ["international journal of air pollution"] = "Int J Air Pollut",
  ["international journal of air-conditioning and refrigeration"] = "Int. J. Air-Cond. Refrig.",
  ["international journal of alcohol and drug research"] = "Int J Alcohol Drug Res",
  ["international journal of algebra and computation"] = "Int. J. Algebra Comput.",
  ["international journal of alternative propulsion"] = "Int. J. Altern. Propul.",
  ["international journal of alzheimer's disease"] = "Int J Alzheimers Dis",
  ["international journal of ambient energy"] = "Int. J. Ambient Energy",
  ["international journal of analysis"] = "Int. J. Anal.",
  ["international journal of analytical chemistry"] = "Int J Anal Chem",
  ["international journal of analytical mass spectrometry and chromatography"] = "Int J Anal Mass Spectrom Cromatogr",
  ["international journal of analytical, pharmaceutical and biomedical sciences"] = "Int J Anal Pharm Biomed Sci",
  ["international journal of anatomical variations"] = "Int J Anat Var",
  ["international journal of andrology"] = "Int J Androl",
  ["international journal of anesthesia"] = "Int J Anesth",
  ["international journal of anesthesiology & pain medicine"] = "Int J Anesthesiol Pain Med",
  ["international journal of anesthesiology & research"] = "Int J Anesthesiol Res",
  ["international journal of anesthetics and anesthesiology"] = "Int J Anesth Anesth",
  ["international journal of animal and veterinary advances"] = "Int J Anim Vet Adv",
  ["international journal of antennas and propagation"] = "Int J Antennas Propag",
  ["international journal of anthropology"] = "Int J Anthropol",
  ["international journal of anthropology and ethnology"] = "Int J Anthropol Ethnol",
  ["international journal of antimicrobial agents"] = "Int. J. Antimicrob. Agents",
  ["international journal of applied & basic medical research"] = "Int J Appl Basic Med Res",
  ["international journal of applied and computational mathematics"] = "Int. J. Appl. Comput. Math.",
  ["international journal of applied and pure science and agriculture"] = "Int J Appl Pure Sci Agric",
  ["international journal of applied biology and pharmaceutical technology"] = "Int J Appl Biol Pharm",
  ["international journal of applied biotechnology and biochemistry"] = "Int J Appl Biotechnol Biochem",
  ["international journal of applied ceramic technology"] = "Int. J. Appl. Ceram. Technol.",
  ["international journal of applied chemistry"] = "Int. J. Appl. Chem.",
  ["international journal of applied cryptography"] = "Int. J. Appl. Cryptogr.",
  ["international journal of applied cryptography. ijact"] = "Int. J. Appl. Cryptogr.",
  ["international journal of applied decision sciences"] = "Int. J. Appl. Decis. Sci.",
  ["international journal of applied earth observation and geoinformation"] = "Int. J. Appl. Earth Obs. Geoinf.",
  ["international journal of applied earth observation and geoinformation : itc journal"] = "Int J Appl Earth Obs Geoinf",
  ["international journal of applied economics and econometrics"] = "Int. J. Appl. Econ. Econometrics",
  ["international journal of applied electromagnetics and mechanics"] = "Int. J. Appl. Electromagnet Mech",
  ["international journal of applied engineering research : ijaer"] = "Int J Appl Eng Res",
  ["international journal of applied glass science"] = "Int. J. Appl. Glass Sci.",
  ["international journal of applied information systems"] = "Int J Appl Inf Syst",
  ["international journal of applied mathematical analysis and applications"] = "Int. J. Appl. Math. Anal. Appl.",
  ["international journal of applied mathematical sciences"] = "Int. J. Appl. Math. Sci.",
  ["international journal of applied mathematics"] = "Int. J. Appl. Math.",
  ["international journal of applied mathematics and computation"] = "Int. J. Appl. Math. Comput.",
  ["international journal of applied mathematics and computer science"] = "Int. J. Appl. Math. Comput. Sci.",
  ["international journal of applied mechanics"] = "Int J Appl Mech",
  ["international journal of applied nonlinear science"] = "Int. J. Appl. Nonlinear Sci.",
  ["international journal of applied pattern recognition"] = "Int. J. Appl. Pattern Recognit.",
  ["international journal of applied positive psychology"] = "Int J Appl Posit Psychol",
  ["international journal of applied psychology"] = "Int J Appl Psychol",
  ["international journal of applied radiation and isotopes"] = "Int. J. Appl. Radiat. Isot.",
  ["international journal of applied research in veterinary medicine"] = "Int J Appl Res Vet Med",
  ["international journal of applied science and computations"] = "Internat. J. Appl. Sci. Comput.",
  ["international journal of applied science and technology"] = "Int. J. Appl. Sci. Technol.",
  ["international journal of applied sciences and biotechnology"] = "Int J Appl Sci Biotechnol",
  ["international journal of applied systemic studies"] = "Int. J. Appl. Syst. Stud.",
  ["international journal of approximate reasoning"] = "Int. J. Approximate Reasoning",
  ["international journal of approximate reasoning : official publication of the north american fuzzy information processing society"] = "Int J Approx Reason",
  ["international journal of aquaculture and fishery sciences"] = "Int J Aquac Fish Sci",
  ["international journal of aquatic research and environmental studies"] = "Int. J. Aquat. Res. Environ. Stud.",
  ["international journal of architectural computing"] = "Int. J. Archit. Comput.",
  ["international journal of architectural heritage"] = "Int. J. Archit. Heritage",
  ["international journal of aromatherapy"] = "Int. J. Aromather.",
  ["international journal of artificial intelligence and robotics research"] = "Int. J. Artif. Intell. Rob. Res.",
  ["international journal of artificial intelligence and soft computing"] = "Int. J. Artif. Intell. Soft Comput.",
  ["international journal of artificial intelligence in education"] = "Int J Artif Intell Educ",
  ["international journal of artificial intelligence tools : architectures, languages, algorithms"] = "Int J Artif Intell Tools",
  ["international journal of artificial organs"] = "Int. J. Artif. Organs",
  ["international journal of asian language processing"] = "Int. J. Asian Lang. Process.",
  ["international journal of astrobiology"] = "Int. J. Astrobiol.",
  ["international journal of audiology"] = "Int J Audiol",
  ["international journal of automation and computing"] = "Int. J. Autom. Comput.",
  ["international journal of automation and control"] = "Int. J. Autom. Control",
  ["international journal of automotive composites"] = "Int. J. Automot. Compos.",
  ["international journal of automotive engineering"] = "Int J Automot Eng",
  ["international journal of automotive technology"] = "Int. J. Automot. Technol.",
  ["international journal of automotive technology and management"] = "Int. J. Automot. Technol. Manage.",
  ["international journal of autonomic computing"] = "Int. J. Auton. Comput.",
  ["international journal of autonomous and adaptive communications systems"] = "Int J Auton Adapt Commun Syst",
  ["international journal of ayurveda research"] = "Int J Ayurveda Res",
  ["international journal of bacteriology"] = "Int J Bacteriol",
  ["international journal of basic and applied physiology"] = "Int J Basic Appl Physiol",
  ["international journal of basic and clinical immunology"] = "Int J Basic Clin Immunol",
  ["international journal of basic and clinical pharmacology"] = "Int J Basic Clin Pharmacol",
  ["international journal of behavioral and consultation therapy"] = "Int J Behav Consult Ther",
  ["international journal of behavioral development"] = "Int J Behav Dev",
  ["international journal of behavioral medicine"] = "Int J Behav Med",
  ["international journal of behavioral nutrition and physical activity"] = "Int. J. Behav. Nutr. Phys. Act.",
  ["international journal of behavioural & healthcare research"] = "Int J Behav Healthc Res",
  ["international journal of bifurcation and chaos"] = "Int. J. Bifurcation Chaos",
  ["international journal of bifurcation and chaos in applied sciences and engineering"] = "Int J Bifurcat Chaos",
  ["international journal of big data"] = "Int J Big Data",
  ["international journal of big data intelligence"] = "Int. J. Big Data Intel.",
  ["international journal of big data management"] = "Int. J. Big Data Manage.",
  ["international journal of big data mining for global warming"] = "Int. J. Big Data Min. Global Warming",
  ["international journal of bilingual education and bilingualism"] = "Int J Biling Educ Biling",
  ["international journal of bio-chromatography"] = "Int J Biochromatogr",
  ["international journal of bio-inspired computation"] = "Int. J. Bio-Inspired Comput.",
  ["international journal of bio-medical computing"] = "Int J Biomed Comput",
  ["international journal of bio-technology and research (chennai)"] = "Int J Biotechnol Res (Chennai)",
  ["international journal of bioassays"] = "Int J Bioassays",
  ["international journal of biobased plastics"] = "Int. J. Biobased Plast.",
  ["international journal of biochemiphysics"] = "Int. J. BioChemiPhysics",
  ["international journal of biochemistry"] = "Int. J. Biochem.",
  ["international journal of biochemistry & cell biology"] = "Int. J. Biochem. Cell Biol.",
  ["international journal of biochemistry and biotechnology"] = "Int J Biochem Biotechnol",
  ["international journal of biochemistry and cell biology"] = "Int. J. Biochem. Cell Biol.",
  ["international journal of biochemistry and molecular biology"] = "Int J Biochem Mol Biol",
  ["international journal of biochemistry research & review"] = "Int J Biochem Res Rev",
  ["international journal of biochromatography"] = "Int. J. Bio-Chromatogr.",
  ["international journal of bioclimatology and biometeorology"] = "Int. J. Bioclimatol. Biometeorol.",
  ["international journal of biodiversity and conservation"] = "Int J Biodivers Conserv",
  ["international journal of biodiversity science & management"] = "Int. J. Biodivers. Sci. Manage.",
  ["international journal of biodiversity science, ecosystem services & management"] = "Int J Biodivers Sci Ecosyst Serv Manag",
  ["international journal of bioelectromagnetism"] = "Int J Bioelectromagn",
  ["international journal of bioinformatics research and applications"] = "Int. J. Bioinf. Res. Appl.",
  ["international journal of biological and chemical sciences"] = "Int J Biol Chem Sci",
  ["international journal of biological and medical research"] = "Int J Biol Med Res",
  ["international journal of biological chemistry"] = "Int. J. Biol. Chem.",
  ["international journal of biological macromolecules"] = "Int J Biol Macromol",
  ["international journal of biological markers"] = "Int. J. Biol. Markers",
  ["international journal of biological research in pregnancy"] = "Int J Biol Res Pregnancy",
  ["international journal of biological sciences"] = "Int J Biol Sci",
  ["international journal of biology"] = "Int J Biol",
  ["international journal of biology and biotechnology"] = "Int. J. Biol. Biotech.",
  ["international journal of biology, pharmacy and allied sciences"] = "Int J Biol Pharm Allied Sci",
  ["international journal of biomaterials"] = "Int J Biomater",
  ["international journal of biomaterials research and engineering"] = "Int J Biomater Res Eng",
  ["international journal of biomathematics"] = "Int. J. Biomath.",
  ["international journal of biomechatronics and biomedical robotics"] = "Int. J. Biomechatron. Biomed. Rob.",
  ["international journal of biomedical data mining"] = "Int J Biomed Data Min",
  ["international journal of biomedical engineering"] = "Int. J. Biomed. Eng.",
  ["international journal of biomedical engineering and clinical science"] = "Int J Biomed Eng Clin Sci",
  ["international journal of biomedical engineering and technology"] = "Int. J. Biomed. Eng. Technol.",
  ["international journal of biomedical imaging"] = "Int J Biomed Imaging",
  ["international journal of biomedical investigation"] = "Int J Biomed Investig",
  ["international journal of biomedical nanoscience and"] = "Int. J. Biomed. Nanosci. Nanotechnol.",
  ["international journal of biomedical nanoscience and nanotechnology"] = "Int J Biomed Nanosci Nanotechnol",
  ["international journal of biomedical research"] = "Int J Biomed Res",
  ["international journal of biomedical science : ijbs"] = "Int J Biomed Sci",
  ["international journal of biomedical sciences"] = "Int. J. Biomed. Sci.",
  ["international journal of biometeorologie"] = "Int. J. Biometeorol.",
  ["international journal of biometeorology"] = "Int J Biometeorol",
  ["international journal of biometrics"] = "Int J Biom",
  ["international journal of biometrics and bioinformatics : ijbb"] = "Int J Biom Bioinformatics",
  ["international journal of bioorganic chemistry & molecular biology"] = "Int J Bioorganic Chem Mol Biol",
  ["international journal of biopharmaceutical sciences"] = "Int J Biopharm Sci",
  ["international journal of bioprinting"] = "Int J Bioprint",
  ["international journal of bioscience, biochemistry and bioinformatics"] = "Int J Biosci Biochem Bioinforma",
  ["international journal of biosciences"] = "Int J Biosci",
  ["international journal of biosciences, psychiatry, and technology (ijbspt)"] = "Int J Biosci Psychiatr Technol IJBSPT",
  ["international journal of biosensors & bioelectronics"] = "Int J Biosens Bioelectron",
  ["international journal of biostatistics"] = "Int. J. Biostat.",
  ["international journal of biotech trends and technology"] = "Int J Biotech Trends Technol",
  ["international journal of biotechnology"] = "Int. J. Biotechnol.",
  ["international journal of biotechnology for wellness industries"] = "Int J Biotechnol Wellness Ind",
  ["international journal of bipolar disorders"] = "Int J Bipolar Disord",
  ["international journal of birth and parent education"] = "Int J Birth Parent Educ",
  ["international journal of blood research and disorders"] = "Int J Blood Res Disord",
  ["international journal of body composition research"] = "Int J Body Compos Res",
  ["international journal of bone marrow research"] = "Int J Bone Marrow Res",
  ["international journal of botany : ijb"] = "Int J Botany",
  ["international journal of botany studies"] = "Int J Bot Stud",
  ["international journal of brain and cognitive sciences"] = "Int J Brain Cogn Sci",
  ["international journal of brain disorders and treatment"] = "Int J Brain Disord Treat",
  ["international journal of breast cancer"] = "Int J Breast Cancer",
  ["international journal of building pathology and adaptation"] = "Int. J. Build. Pathol. Adapt.",
  ["international journal of bullying prevention : an official publication of the international bullying prevention association"] = "Int J Bullying Prev",
  ["international journal of burns and trauma"] = "Int J Burns Trauma",
  ["international journal of business"] = "Int. J. Bus.",
  ["international journal of business information systems"] = "Int J Bus Inf Syst",
  ["international journal of cancer"] = "Int J Cancer",
  ["international journal of cancer and clinical research"] = "Int J Cancer Clin Res",
  ["international journal of cancer and oncology"] = "Int J Cancer Oncol",
  ["international journal of cancer management"] = "Int J Cancer Manag",
  ["international journal of cancer prevention"] = "Int J Canc Prev",
  ["international journal of cancer research"] = "Int J Cancer Res",
  ["international journal of cancer research & therapy"] = "Int J Cancer Res Ther",
  ["international journal of cancer research and molecular mechanisms"] = "Int J Cancer Res Mol Mech",
  ["international journal of cancer research and prevention"] = "Int J Cancer Res Prev",
  ["international journal of cancer science & therapy"] = "Int J Cancer Sci Ther",
  ["international journal of cancer. journal international du cancer"] = "Int J Cancer",
  ["international journal of cancer. supplement"] = "Int. J. Cancer. Suppl.",
  ["international journal of cancer. supplement = journal international du cancer. supplement"] = "Int J Cancer Suppl",
  ["international journal of cardiac imaging"] = "Int J Card Imaging",
  ["international journal of cardiology"] = "Int J Cardiol",
  ["international journal of cardiology-heart & vessels"] = "Int. J. Cardiol. Heart Vessels",
  ["international journal of cardiology-metabolic & endocrine"] = "Int. J. Cardol. Metab. Endocr.",
  ["international journal of cardiology. heart & vasculature"] = "Int J Cardiol Heart Vasc",
  ["international journal of cardiology. heart & vessels"] = "Int J Cardiol Heart Vessel",
  ["international journal of cardiology. hypertension"] = "Int J Cardiol Hypertens",
  ["international journal of cardiovascular imaging"] = "Int. J. Cardiovasc. Imaging",
  ["international journal of cardiovascular interventions"] = "Int J Cardiovasc Intervent",
  ["international journal of cardiovascular research"] = "Int J Cardiovasc Res",
  ["international journal of care and caring"] = "Int J Care Caring",
  ["international journal of care coordination"] = "Int J Care Coord",
  ["international journal of care pathways"] = "Int J Care Pathw",
  ["international journal of caring sciences"] = "Int J Caring Sci",
  ["international journal of cartography"] = "Int. J. Cartogr.",
  ["international journal of case reports and images"] = "Int J Case Rep Imag",
  ["international journal of cast metals research"] = "Int. J. Cast Met. Res.",
  ["international journal of cell and tissue engineering, artificial cells and regenerative medicine"] = "Int. J. Cell Tissue Eng. Artif. Cells Regener. Med.",
  ["international journal of cell biology"] = "Int J Cell Biol",
  ["international journal of cell biology and physiology"] = "Int J Cell Biol Physiol",
  ["international journal of cell cloning"] = "Int J Cell Cloning",
  ["international journal of cell science & molecular biology"] = "Int J Cell Sci Mol Biol",
  ["international journal of cement composites and lightweight concrete"] = "Int. J. Cem. Compos. Lightweight Concrete",
  ["international journal of central banking"] = "Int J Cent Bank",
  ["international journal of ceramic engineering and science"] = "Int. J. Ceram. Eng. Sci.",
  ["international journal of cerebrovascular disease and stroke"] = "Int J Cerebrovasc Dis Stroke",
  ["international journal of chemical and analytical science"] = "Int. J. Chem. Anal. Sci.",
  ["international journal of chemical and biochemical sciences"] = "Int. J. Chem. Biochem. Sci.",
  ["international journal of chemical engineering"] = "Int. J. Chem. Eng.",
  ["international journal of chemical engineering and"] = "Int. J. Chem. Eng. Appl.",
  ["international journal of chemical engineering and applications (ijcea)"] = "Int J Chem Eng Appl",
  ["international journal of chemical kinetics"] = "Int. J. Chem. Kinet.",
  ["international journal of chemical modeling"] = "Int. J. Chem. Model.",
  ["international journal of chemical reactor engineering"] = "Int. J. Chem. Reactor Eng.",
  ["international journal of chemical sciences"] = "Int. J. Chem. Sc.",
  ["international journal of chemical studies"] = "Int J Chem Stud",
  ["international journal of chemistry"] = "Int J Chem",
  ["international journal of chemtech research"] = "Int J Chemtech Res",
  ["international journal of child and adolescent health"] = "Int J Child Adolesc health",
  ["international journal of child and adolescent resilience"] = "Int J Child Adolesc Resil",
  ["international journal of child development and mental health"] = "Int J Child Dev Ment Health",
  ["international journal of child health and human development : ijchd"] = "Int J Child Health Hum Dev",
  ["international journal of child health and nutrition"] = "Int J Child Health Nutr",
  ["international journal of child psychotherapy"] = "Int J Child Psychother",
  ["international journal of child, youth & family studies : ijcyfs"] = "Int J Child Youth Family Stud",
  ["international journal of child-computer interaction"] = "Int. J. Child-Comput. Interact.",
  ["international journal of childbirth"] = "Int J Childbirth",
  ["international journal of chinese culture and management"] = "Int J Chin Cult Manag",
  ["international journal of chronic diseases"] = "Int J Chronic Dis",
  ["international journal of chronic diseases & therapy"] = "Int J Chronic Dis Ther",
  ["international journal of chronic obstructive pulmonary disease"] = "Int J Chron Obstruct Pulmon Dis",
  ["international journal of chronobiology"] = "Int J Chronobiol",
  ["international journal of circuit theory and applications"] = "Int. J. Circuit Theory Appl.",
  ["international journal of circumpolar health"] = "Int J Circumpolar Health",
  ["international journal of circumpolar health. supplement"] = "Int J Circumpolar Health Suppl",
  ["international journal of civic engagement and social change"] = "Int J Civ Engagem Soc Change",
  ["international journal of civil engineering"] = "Int. J. Civ. Eng.",
  ["international journal of climate change strategies and management"] = "Int J Clim Chang Strateg Manag",
  ["international journal of climate change: impacts and responses"] = "Int. J. Clim. Change: Impacts Responses",
  ["international journal of climatology"] = "Int. J. Climatol.",
  ["international journal of climatology : a journal of the royal meteorological society"] = "Int J Climatol",
  ["international journal of clinical & laboratory research"] = "Int J Clin Lab Res",
  ["international journal of clinical and experimental hypnosis"] = "Int. J. Clin. Exp. Hypn.",
  ["international journal of clinical and experimental medicine"] = "Int. J. Clin. Exp. Med.",
  ["international journal of clinical and experimental pathology"] = "Int J Clin Exp Pathol",
  ["international journal of clinical and experimental physiology"] = "Int J Clin Exp Physiol",
  ["international journal of clinical and health psychology : ijchp"] = "Int J Clin Health Psychol",
  ["international journal of clinical and laboratory research"] = "Int. J. Clin. Lab. Res.",
  ["international journal of clinical anesthesiology"] = "Int J Clin Anesthesiol",
  ["international journal of clinical biostatistics and biometrics"] = "Int J Clin Biostat Biom",
  ["international journal of clinical case reports and reviews : open access"] = "Int J Clin Case Rep Rev",
  ["international journal of clinical endocrinology and metabolism"] = "Int J Clin Endocrinol Metab",
  ["international journal of clinical medicine"] = "Int J Clin Med",
  ["international journal of clinical monitoring and computing"] = "Int J Clin Monit Comput",
  ["international journal of clinical oncology"] = "Int J Clin Oncol",
  ["international journal of clinical oncology / japan society of clinical oncology"] = "Int J Clin Oncol",
  ["international journal of clinical oral and maxillofacial surgery"] = "Int J Clin Oral Maxillofac Surg",
  ["international journal of clinical pediatric dentistry"] = "Int J Clin Pediatr Dent",
  ["international journal of clinical pharmacology & toxicology"] = "Int J Clin Pharmacol Toxicol",
  ["international journal of clinical pharmacology and biopharmacy"] = "Int J Clin Pharmacol Biopharm",
  ["international journal of clinical pharmacology and therapeutics"] = "Int J Clin Pharmacol Ther",
  ["international journal of clinical pharmacology research"] = "Int J Clin Pharmacol Res",
  ["international journal of clinical pharmacology, therapy and toxicology"] = "Int J Clin Pharmacol",
  ["international journal of clinical pharmacology, therapy, and toxicology"] = "Int J Clin Pharmacol Ther Toxicol",
  ["international journal of clinical pharmacy"] = "Int J Clin Pharm",
  ["international journal of clinical practice"] = "Int J Clin Pract",
  ["international journal of clinical practice. supplement"] = "Int J Clin Pract Suppl",
  ["international journal of clinical psychiatry and mental health"] = "Int J Clin Psychiatry Ment Health",
  ["international journal of clinical research & trials"] = "Int J Clin Res Trials",
  ["international journal of clinical rheumatology"] = "Int J Clin Rheumtol",
  ["international journal of clinical transfusion medicine"] = "Int. J. Clin. Tranfus. Med.",
  ["international journal of clinical trials"] = "Int J Clin Trials",
  ["international journal of clothing science and technology"] = "Int. J. Clothing Sci. Technol.",
  ["international journal of coal geology"] = "Int. J. Coal Geol.",
  ["international journal of coal preparation and utilization"] = "Int. J. Coal Prep. Util.",
  ["international journal of coal science & technology"] = "Int J Coal Sci Technol",
  ["international journal of cognitive biometrics"] = "Int. J. Cognit. Biom.",
  ["international journal of cognitive computing in engineering"] = "Int. J. Cognit. Comput. Eng.",
  ["international journal of cognitive ergonomics"] = "Int J Cogn Ergon",
  ["international journal of cognitive linguistics"] = "Int J Cogn Linguist",
  ["international journal of cognitive therapy"] = "Int J Cogn Ther",
  ["international journal of collaborative research on internal medicine & public health"] = "Int J Collab Res Intern Med Public Health",
  ["international journal of colorectal disease"] = "Int J Colorectal Dis",
  ["international journal of comadem"] = "Int. J. COMADEM",
  ["international journal of communication"] = "Int J Commun",
  ["international journal of communication and health"] = "Int J Commun Health",
  ["international journal of communication networks and distributed systems"] = "Int. J. Commun. Networks Distrib. Syst.",
  ["international journal of communication systems"] = "Int. J. Commun. Syst.",
  ["international journal of communications, network and systems sciences"] = "Int J Commun Netw Syst Sci",
  ["international journal of community based nursing and midwifery"] = "Int J Community Based Nurs Midwifery",
  ["international journal of community currency research"] = "Int J Community Curr Res",
  ["international journal of community medicine and public health"] = "Int J Community Med Public Health",
  ["international journal of comparative and applied criminal justice"] = "Int J Comp Appl Crim Justice",
  ["international journal of comparative psychology"] = "Int J Comp Psychol",
  ["international journal of comparative psychology / iscp ; sponsored by the international society for comparative psychology and the university of calabria"] = "Int J Comp Psychol",
  ["international journal of comparative sociology"] = "Int J Comp Sociol",
  ["international journal of complementary & alternative medicine"] = "Int J Complement Altern Med",
  ["international journal of computational biology and drug design"] = "Int J Comput Biol Drug Des",
  ["international journal of computational economics and econometrics"] = "Int. J. Comput. Econ. Econom.",
  ["international journal of computational engineering science"] = "Int. J. Comput. Eng. Sci.",
  ["international journal of computational fluid dynamics"] = "Int. J. Comput. Fluid Dyn.",
  ["international journal of computational geometry & applications"] = "Int J Comput Geom Appl",
  ["international journal of computational geometry and applications"] = "Int. J. Comput. Geom. Appl.",
  ["international journal of computational intelligence and applications"] = "Int. J. Comput. Intell. Appl.",
  ["international journal of computational intelligence systems"] = "Int. J. Comput. Intell. Syst.",
  ["international journal of computational materials science and engineering"] = "Int J Comput Mater Sci Eng",
  ["international journal of computational materials science and surface engineering"] = "Int. J. Comput. Mater. Sci. Surf. Eng.",
  ["international journal of computational methods"] = "Int. J. Comput. Methods",
  ["international journal of computational methods and experimental measurements"] = "Int. J. Comput. Methods Exp. Meas.",
  ["international journal of computational models and algorithms in medicine"] = "Int J Comput Models Algorithms Med",
  ["international journal of computational science"] = "Int J Comput Sci",
  ["international journal of computational science and engineering"] = "Int. J. Comput. Sci. Eng.",
  ["international journal of computational vision and robotics"] = "Int J Comput Vis Robot",
  ["international journal of computer applications"] = "Int J Comput Appl",
  ["international journal of computer applications in technology"] = "Int. J. Comput. Appl. Technol.",
  ["international journal of computer assisted radiology and surgery"] = "Int J Comput Assist Radiol Surg",
  ["international journal of computer games technology"] = "Int. J. Comput. Games Technol.",
  ["international journal of computer integrated manufacturing"] = "Int J Comput Integr Manuf",
  ["international journal of computer mathematics"] = "Int. J. Comput. Math.",
  ["international journal of computer mathematics. computer systems theory"] = "Int. J. Comput. Math. Comput. Syst. Theory",
  ["international journal of computer mathematics: computer systems theory"] = "Int. J. Comput. Math.: Comput. Syst. Theory",
  ["international journal of computer processing of languages"] = "Int. J. Comput. Process. Lang.",
  ["international journal of computer science"] = "Int J Comp Sci",
  ["international journal of computer science (rabat)"] = "Int J Comput Sci (Rabat)",
  ["international journal of computer science in sport"] = "Int J Comput Sci Sport",
  ["international journal of computer vision"] = "Int. J. Comput. Vision",
  ["international journal of computer-supported collaborative learning"] = "Int J Comput Support Collab Learn",
  ["international journal of computerized dentistry"] = "Int J Comput Dent",
  ["international journal of computers and applications"] = "Int. J. Comput. Appl.",
  ["international journal of computers for mathematical learning"] = "Int. J. Comput. Math. Learn.",
  ["international journal of computers in healthcare"] = "Int J Comput Healthc",
  ["international journal of computing science and mathematics"] = "Int. J. Comput. Sci. Math.",
  ["international journal of concrete structures & materials"] = "Int J Concr Struct Mater",
  ["international journal of concrete structures and materials"] = "Int. J. Concr. Struct. Mater.",
  ["international journal of condensed matter research"] = "Int. J. Condens. Matter Res.",
  ["international journal of conflict and violence"] = "Int J Conf Violence",
  ["international journal of constitutional law"] = "Int J Const Law",
  ["international journal of construction education and research"] = "Int J Constr Educ Res",
  ["international journal of construction management"] = "Int. J. Construct. Manage.",
  ["international journal of construction project management"] = "Int. J. Construct. Project Manage.",
  ["international journal of consumer studies"] = "Int J Consum Stud",
  ["international journal of contemporary pediatrics"] = "Int J Contemp Pediatrics",
  ["international journal of contemporary sociology"] = "Int J Contemp Sociol",
  ["international journal of continuing engineering education and life-long learning"] = "Int J Contin Eng Educ Life Long Learn",
  ["international journal of control"] = "Int. J. Control",
  ["international journal of control, automation and systems"] = "Int. J. Control Autom. Syst.",
  ["international journal of control, automation, and systems"] = "Int J Control Autom Syst",
  ["international journal of cooperative information systems"] = "Int J Coop Inf Syst",
  ["international journal of corrosion"] = "Int. J. Corros.",
  ["international journal of cosmetic science"] = "Int J Cosmet Sci",
  ["international journal of crashworthiness"] = "Int. J. Crashworthiness",
  ["international journal of criminology and sociology"] = "Int J Criminol Sociol",
  ["international journal of critical illness and injury science"] = "Int J Crit Illn Inj Sci",
  ["international journal of critical infrastructure protection"] = "Int. J. Crit. Infrastruct. Prot.",
  ["international journal of critical infrastructures"] = "Int. J. Crit. Infrastruct.",
  ["international journal of crowd science"] = "Int. J. Crowd Sci.",
  ["international journal of crude drug research"] = "Int J Crude Drug Res",
  ["international journal of culture and mental health"] = "Int J Cult Ment Health",
  ["international journal of current advanced research"] = "Int J Curr Adv Res",
  ["international journal of current chemistry"] = "Int. J. Curr. Chem.",
  ["international journal of current engineering and scientific research"] = "Int. J. Curr. Eng. Sci. Res.",
  ["international journal of current microbiology and applied sciences"] = "Int J Curr Microbiol Appl Sci",
  ["international journal of current multidisciplinary studies"] = "Int J Curr Multidiscip Stud",
  ["international journal of current pharmaceutical research"] = "Int J Curr Pharm Res",
  ["international journal of current research"] = "Int J Curr Res",
  ["international journal of current research and academic review"] = "Int J Curr Res Acad Rev",
  ["international journal of current research and review"] = "Int J Curr Res Rev",
  ["international journal of current research in biosciences and plant biology"] = "Int J Curr Res Biosci Plant Biol",
  ["international journal of current trends in science and technology"] = "Int. J. Curr. Trends Sci. Technol.",
  ["international journal of dairy technology"] = "Int. J. Dairy Technol.",
  ["international journal of damage mechanics"] = "Int. J. Damage Mech.",
  ["international journal of data mining and bioinformatics"] = "Int J Data Min Bioinform",
  ["international journal of data mining modelling and management"] = "Int. J. Data Min. Model. Manage.",
  ["international journal of data science and analytics"] = "Int J Data Sci Anal",
  ["international journal of data science in mathematical sciences"] = "Int. J. Data Sci. Math. Sci.",
  ["international journal of data science in the mathematical sciences"] = "Int. J. Data Sci. Math. Sci.",
  ["international journal of data warehousing and mining"] = "Int. J. Data Warehouse. Min.",
  ["international journal of dental hygiene"] = "Int J Dent Hyg",
  ["international journal of dental science and research"] = "Int. J. Dent. Sci. Res.",
  ["international journal of dental symposia"] = "Int J Dent Symp",
  ["international journal of dentistry"] = "Int J Dent",
  ["international journal of dentistry and oral health"] = "Int J Dent Oral Health",
  ["international journal of dentistry and oral science"] = "Int J Dent Oral Sci",
  ["international journal of dermatology"] = "Int J Dermatol",
  ["international journal of dermatology and venereology"] = "Int J Dermatol Venereol",
  ["international journal of design & nature and ecodynamics"] = "Int. J. Des. Nat. Ecodyn.",
  ["international journal of design & nature and ecodynamics : a transdisciplinary journal relating to nature, science and the humanities"] = "Int J  Des Nat Ecodyn",
  ["international journal of design and nature"] = "Int. J. Des. Nat.",
  ["international journal of design creativity and innovation"] = "Int. J. Des. Creativity Innovation",
  ["international journal of designs for learning"] = "Int J Des Learn",
  ["international journal of developmental biology"] = "Int. J. Dev. Biol.",
  ["international journal of developmental disabilities"] = "Int J Dev Disabil",
  ["international journal of developmental neuroscience"] = "Int. J. Dev. Neurosci.",
  ["international journal of developmental neuroscience : the official journal of the international society for developmental neuroscience"] = "Int J Dev Neurosci",
  ["international journal of developmental research"] = "Int J Dev Res",
  ["international journal of developmental science"] = "Int J Dev Sci",
  ["international journal of diabetes & metabolic syndrome"] = "Int J Diabetes Metab Syndr",
  ["international journal of diabetes and clinical research"] = "Int J Diabetes Clin Res",
  ["international journal of diabetes in developing countries"] = "Int J Diabetes Dev Ctries",
  ["international journal of diabetes mellitus"] = "Int J Diabetes Mellit",
  ["international journal of diabetes research"] = "Int J Diabetes Res",
  ["international journal of diabetology & vascular disease research"] = "Int J Diabetol Vasc Dis Res",
  ["international journal of difference equations"] = "Int. J. Difference Equ.",
  ["international journal of differential equations"] = "Int. J. Differ. Equations",
  ["international journal of digital curation"] = "Int J Digit Curation",
  ["international journal of digital earth"] = "Int. J. Digital Earth",
  ["international journal of digital multimedia broadcasting"] = "Int. J. Digital Multimedia Broadcast.",
  ["international journal of disability, development, and education"] = "Intl J Disabil Dev Educ",
  ["international journal of disaster resilience in the built environment"] = "Int J Disaster Resil Built Environ",
  ["international journal of disaster risk reduction"] = "Int. J. Disaster Risk Reduct.",
  ["international journal of disaster risk reduction : ijdrr"] = "Int J Disaster Risk Reduct",
  ["international journal of disaster risk science"] = "Int. J. Disaster Risk Sci.",
  ["international journal of distributed sensor networks"] = "Int. J. Distrib. Sens. Netw.",
  ["international journal of dynamical systems and differential equations"] = "Int. J. Dyn. Syst. Differ. Equ.",
  ["international journal of dynamics and control"] = "Int. J. Dyn. Control",
  ["international journal of e-navigation and maritime economy"] = "Int. J. e-Navig. Marit. Econ.",
  ["international journal of early childhood = revue internationale de l'enfance préscolaire = revista internacional de la infancia pre-escolar"] = "Int J Early Child",
  ["international journal of early years education"] = "Int J Early Years Educ",
  ["international journal of earth sciences"] = "Int. J. Earth Sci.",
  ["international journal of earth sciences : geologische rundschau"] = "Int J Earth Sci",
  ["international journal of eating disorders"] = "Int. J. Eat. Disord.",
  ["international journal of ecodynamics"] = "Int. J. Ecodyn.",
  ["international journal of ecological economics and statistics"] = "Int J Ecol Econ Stat",
  ["international journal of ecology"] = "Int. J. Ecol.",
  ["international journal of economic research"] = "Int J Econ Res",
  ["international journal of economic theory"] = "Int. J. Econ. Theory",
  ["international journal of economics and finance"] = "Int J Econ Finance",
  ["international journal of ecotoxicology and ecobiology"] = "Int J Ecotoxicol Ecobiol",
  ["international journal of education and development using information and communication technology"] = "Int J Educ Dev Using Inf Commun Technol",
  ["international journal of education and social science"] = "Int J Educ Soc Sci",
  ["international journal of educational development"] = "Int J Educ Dev",
  ["international journal of educational methodology"] = "Int J Educ Method",
  ["international journal of educational research"] = "Int J Educ Res",
  ["international journal of electric and hybrid vehicles"] = "Int. J. Electr. Hybrid Veh.",
  ["international journal of electrical engineering"] = "Int. J. Electr. Eng.",
  ["international journal of electrical engineering & education"] = "Int. J. Electr. Eng. Educ.",
  ["international journal of electrical engineering and education"] = "Int. J. Electr. Eng. Educ.",
  ["international journal of electrical engineering education"] = "Int. J. Electr. Eng. Educ.",
  ["international journal of electrical power & energy systems"] = "Int. J. Electr. Power Energy Syst.",
  ["international journal of electrical, computer, and systems engineering"] = "Int. J. Electr. Comput. Syst. Eng.",
  ["international journal of electrochemical science"] = "Int. J. Electrochem. Sci.",
  ["international journal of electrochemistry"] = "Int. J. Electrochem.",
  ["international journal of electronic healthcare"] = "Int J Electron Healthc",
  ["international journal of electronic security and digital forensics"] = "Int. J. Electron. Secur. Digit. Forensics",
  ["international journal of electronics"] = "Int. J. Electron.",
  ["international journal of electronics and communications = archiv fur elektronik und ubertragungstechnik : aeu"] = "Int J Electron Commun",
  ["international journal of electronics letters"] = "Int. J. Electron. Lett.",
  ["international journal of embedded systems"] = "Int. J. Embedded Syst.",
  ["international journal of emergency management"] = "Int. J. Emergency Manage.",
  ["international journal of emergency medicine"] = "Int J Emerg Med",
  ["international journal of emergency mental health"] = "Int J Emerg Ment Health",
  ["international journal of emerging electric power systems"] = "Int. J. Emerg. Electr. Power Syst.",
  ["international journal of emerging multidisciplinary fluid sciences"] = "Int J Emerg Multidiscip Fluid Sci",
  ["international journal of empirical economics"] = "Int. J. Empirical Econ.",
  ["international journal of endocrine oncology"] = "Int J Endocr Oncol",
  ["international journal of endocrinology"] = "Int J Endocrinol",
  ["international journal of endocrinology and metabolic disorders"] = "Int J Endocrinol Metab Disord",
  ["international journal of endocrinology and metabolism"] = "Int J Endocrinol Metab",
  ["international journal of energetic materials and chemical propulsion"] = "Int. J. Energetic Mater. Chem. Propul.",
  ["international journal of energy and environment"] = "Int. J. Energy Environ.",
  ["international journal of energy and environmental engineering"] = "Int. J. Energy Environ. Eng.",
  ["international journal of energy and water resources"] = "Int. J. Energy Water Resour.",
  ["international journal of energy for a clean environment"] = "Int. J. Energy Clean Environ.",
  ["international journal of energy production and management"] = "Int. J. Energy Prod. Manage.",
  ["international journal of energy research"] = "Int. J. Energy Res.",
  ["international journal of energy sector management"] = "Int. J. Energy Sect. Manage.",
  ["international journal of engine research"] = "Int. J. Engine Res.",
  ["international journal of engineering"] = "Int. J. Eng.",
  ["international journal of engineering and advanced technology"] = "Int J Eng Adv Technol",
  ["international journal of engineering and scientific research"] = "Int J Eng Sci Res",
  ["international journal of engineering and technology : ijet"] = "Int J Eng Technol",
  ["international journal of engineering business management"] = "Int. J. Eng. Bus. Manage.",
  ["international journal of engineering education"] = "Int. J. Eng. Educ.",
  ["international journal of engineering mathematics"] = "Int. J. Eng. Math.",
  ["international journal of engineering research & technology (ahmedabad)"] = "Int J Eng Res Technol (Ahmedabad)",
  ["international journal of engineering research and applications"] = "Int J Eng Res Appl",
  ["international journal of engineering research in africa"] = "Int. J. Eng. Res. Afr.",
  ["international journal of engineering science"] = "Int. J. Eng. Sci.",
  ["international journal of engineering sciences & research technology"] = "Int J Eng Sci Res Technol",
  ["international journal of engineering systems modelling and simulation"] = "Int. J. Eng. Syst. Modell. Simul.",
  ["international journal of engineering technology and scientific innovation"] = "Int J Eng Technol Sci Innov",
  ["international journal of engineering, transactions a:"] = "Int. J. Eng., Trans. A:",
  ["international journal of engineering, transactions a: basics"] = "Int. J. Eng. Trans. A",
  ["international journal of engineering, transactions b: applications"] = "Int. J. Eng. Trans. B",
  ["international journal of english linguistics"] = "Int J Engl Linguist",
  ["international journal of enhanced research in science technology and engineering"] = "Int J Enhanc Res Sci Technol Eng",
  ["international journal of enteric pathogens"] = "Int J Enteric Pathog",
  ["international journal of enterprise information systems"] = "Int. J. Enterp. Inf. Syst.",
  ["international journal of entomological research"] = "Int J Entomol Res",
  ["international journal of entrepreneurship and innovation management"] = "Int J Entrep Innov Manag",
  ["international journal of environment and pollution"] = "Int. J. Environ. Pollut.",
  ["international journal of environment and pollution research"] = "Int J Environ Pollut Res",
  ["international journal of environment and sustainable development"] = "Int. J. Environ. Sustainable Dev.",
  ["international journal of environment and waste management"] = "Int. J. Environ. Waste Manage.",
  ["international journal of environmental & agriculture research"] = "Int J Environ Agric Res",
  ["international journal of environmental analytical chemistry"] = "Int. J. Environ. Anal. Chem.",
  ["international journal of environmental and analytical chemistry"] = "Int. J. Environ. Anal. Chem.",
  ["international journal of environmental and science education : ijese"] = "Int J Environ Sci Educ",
  ["international journal of environmental bioremediation & biodegradation"] = "Int J Environ Bioremediat Biodegrad",
  ["international journal of environmental health engineering"] = "Int J Environ Health Eng",
  ["international journal of environmental health research"] = "Int J Environ Health Res",
  ["international journal of environmental impacts"] = "Int. J. Environ. Impacts",
  ["international journal of environmental research"] = "Int J Environ Res",
  ["international journal of environmental research and public health"] = "Int J Environ Res Public Health",
  ["international journal of environmental resources research"] = "Int J Environ Resour Res",
  ["international journal of environmental science and engineering (pairie view, tex.)"] = "Int J Environ Sci Eng",
  ["international journal of environmental science and technology"] = "Int. J. Environ. Sci. Technol.",
  ["international journal of environmental science and technology : ijest"] = "Int J Environ Sci Technol (Tehran)",
  ["international journal of environmental sciences"] = "Int J Environ Sci",
  ["international journal of environmental studies"] = "Int. J. Environ. Stud.",
  ["international journal of environmental technology and management"] = "Int. J. Environ. Technol. Manage.",
  ["international journal of epidemiologic research"] = "Int J Epidemiol Res",
  ["international journal of epidemiology"] = "Int J Epidemiol",
  ["international journal of epilepsy"] = "Int J Epilepsy",
  ["international journal of equilibrium research"] = "Int J Equilib Res",
  ["international journal of essential oil therapeutics"] = "Int. J. Essent. Oil Ther.",
  ["international journal of evidence-based healthcare"] = "Int J Evid Based Healthc",
  ["international journal of evolution"] = "Int J Evol",
  ["international journal of evolution equations"] = "Int J Evol Equ",
  ["international journal of evolutionary biology"] = "Int J Evol Biol",
  ["international journal of exercise science"] = "Int J Exerc Sci",
  ["international journal of exergy"] = "Int. J. Exergy",
  ["international journal of experimental and computational biomechanics"] = "Int J Exp Comput Biomech",
  ["international journal of experimental diabesity research"] = "Int J Exp Diabesity Res",
  ["international journal of experimental diabetes research"] = "Int J Exp Diabetes Res",
  ["international journal of experimental pathology"] = "Int J Exp Pathol",
  ["international journal of extensive research"] = "Int J Extensive Res",
  ["international journal of extreme manufacturing"] = "Int. J. Extrem. Manuf.",
  ["international journal of family medicine"] = "Int J Family Med",
  ["international journal of fatigue"] = "Int. J. Fatigue",
  ["international journal of feminist approaches to bioethics"] = "Int J Fem Approaches Bioeth",
  ["international journal of fermented foods"] = "Int J Fermented Foods",
  ["international journal of fertility"] = "Int J Fertil",
  ["international journal of fertility & sterility"] = "Int J Fertil Steril",
  ["international journal of fertility and menopausal studies"] = "Int J Fertil Menopausal Stud",
  ["international journal of fertility and women's medicine"] = "Int J Fertil Womens Med",
  ["international journal of fertility and women’s medicine"] = "Int. J. Fertil. Womens Med.",
  ["international journal of finance and economics"] = "Int. J. Finance Econ.",
  ["international journal of financial engineering"] = "Int. J. Financ. Eng.",
  ["international journal of fisheries and aquatic studies"] = "Int J Fish Aquat Stud",
  ["international journal of fitness"] = "Int J Fit",
  ["international journal of flexible manufacturing systems"] = "Int. J. Flexible Manuf. Syst.",
  ["international journal of fluid dynamics"] = "Int. J. Comput. Fluid Dyn.",
  ["international journal of fluid engineering"] = "Int. J. Fluid Eng.",
  ["international journal of fluid mechanics research"] = "Int. J. Fluid Mech. Res.",
  ["international journal of fluid power"] = "Int. J. Fluid Power",
  ["international journal of food and agricultural economics"] = "Int J Food Agric Econ",
  ["international journal of food contamination"] = "Int. J. Food Contam.",
  ["international journal of food engineering"] = "Int. J. Food Eng.",
  ["international journal of food microbiology"] = "Int J Food Microbiol",
  ["international journal of food properties"] = "Int. J. Food Prop.",
  ["international journal of food science"] = "Int J Food Sci",
  ["international journal of food science & technology"] = "Int J Food Sci Technol",
  ["international journal of food science and nutrition"] = "Int. J. Food Sci. Nutr.",
  ["international journal of food science and technology"] = "Int. J. Food Sci. Technol.",
  ["international journal of food science, nutrition and dietetics"] = "Int J Food Sci Nutr Diet",
  ["international journal of food sciences and nutrition"] = "Int J Food Sci Nutr",
  ["international journal of food studies"] = "Int. J. Food Stud.",
  ["international journal of forecasting"] = "Int J Forecast",
  ["international journal of forensic dentistry"] = "Int. J. Forensic Dent.",
  ["international journal of forensic mental health"] = "Int J Forensic Ment Health",
  ["international journal of forensic science & pathology"] = "Int J Forensic Sci Pathol",
  ["international journal of forest engineering"] = "Int. J. For. Eng.",
  ["international journal of forestry research"] = "Int. J. For. Res.",
  ["international journal of forming processes"] = "Int. J. Form. Processes",
  ["international journal of foundations of computer science"] = "Int. J. Found. Comput. Sci.",
  ["international journal of fracture"] = "Int. J. Fract.",
  ["international journal of fracture mechanics"] = "Int. J. Fract.Mech.",
  ["international journal of fruit science"] = "Int. J. Fruit Sci.",
  ["international journal of functional informatics and personalised medicine"] = "Int J Funct Inform Personal Med",
  ["international journal of functional nutrition"] = "Int J Funct Nutr",
  ["international journal of fuzzy systems"] = "Int. J. Fuzzy Syst.",
  ["international journal of game theory"] = "Int. J. Game Theory",
  ["international journal of gaming and computer-mediated simulations"] = "Int J Gaming Comput Mediat Simul",
  ["international journal of gastrointestinal cancer"] = "Int J Gastrointest Cancer",
  ["international journal of gastronomy and food science"] = "Int J Gastron Food Sci",
  ["international journal of gender, science and technology"] = "Int J Gend Sci Technol",
  ["international journal of general medicine"] = "Int J Gen Med",
  ["international journal of general systems"] = "Int. J. Gen. Syst.",
  ["international journal of genetics and molecular biology"] = "Int J Genet Mol Biol",
  ["international journal of genomics"] = "Int J Genomics",
  ["international journal of genomics and data mining"] = "Int J Genom Data Min",
  ["international journal of genomics and proteomics"] = "Int J Genomics Proteomics",
  ["international journal of geo-engineering"] = "Int. J. Geo-Eng.",
  ["international journal of geographical information science"] = "Int. J. Geogr. Inf. Sci.",
  ["international journal of geographical information science : ijgis"] = "Int J Geogr Inf Sci",
  ["international journal of geographical sciences"] = "Int. J. Geogr. Inf. Sci.",
  ["international journal of geoheritage and parks"] = "Int. J. Geoheritage Parks",
  ["international journal of geomagnetism and aeronomy"] = "Int. J. Geomag. Aeron.",
  ["international journal of geomatics"] = "Int. J. Geomatics",
  ["international journal of geomechanics"] = "Int. J. Geomech.",
  ["international journal of geometric methods in modern physics"] = "Int. J. Geom. Methods Mod. Phys.",
  ["international journal of geometry"] = "Int. J. Geom.",
  ["international journal of geophysics"] = "Int. J. Geophys.",
  ["international journal of geosynthetics and ground engineering"] = "Int. J. Geosynth. Ground Eng.",
  ["international journal of geotechnical engineering"] = "Int. J. Geotech. Eng.",
  ["international journal of geriatric psychiatry"] = "Int J Geriatr Psychiatry",
  ["international journal of gerontology"] = "Int J Gerontol",
  ["international journal of global energy issues"] = "Int. J. Global Energy Issues",
  ["international journal of global warming"] = "Int. J. Global Warming",
  ["international journal of green energy"] = "Int. J. Green Energy",
  ["international journal of green nanotechnology"] = "Int. J. Green Nanotechnol.",
  ["international journal of green nanotechnology. biomedicine"] = "Int J Green Nanotechnol Biomed",
  ["international journal of green nanotechnology: biomedicine"] = "Int. J. Green Nanotechnol. Biomed.",
  ["international journal of green nanotechnology: materials science and engineering"] = "Int. J. Green Nanotechnol. Mater. Sci. Eng.",
  ["international journal of green nanotechnology: physics and chemistry"] = "Int. J. Green Nanotechnol. Phys. Chem.",
  ["international journal of greenhouse gas control"] = "Int. J. Greenhouse Gas Control",
  ["international journal of grid and utility computing"] = "Int. J. Grid Util. Comput.",
  ["international journal of group psychotherapy"] = "Int J Group Psychother",
  ["international journal of group theory"] = "Int. J. Group Theory",
  ["international journal of gynaecology and obstetrics"] = "Int. J. Gynaecol. Obstet.",
  ["international journal of gynaecology and obstetrics: the official organ of the international federation of gynaecology and obstetrics"] = "Int J Gynaecol Obstet",
  ["international journal of gynecological cancer : official journal of the international gynecological cancer society"] = "Int J Gynecol Cancer",
  ["international journal of gynecological obstetrical and reproductive medicine research"] = "Int J Gynecol Obstet Reprod Med Res",
  ["international journal of gynecological pathology"] = "Int. J. Gynecol. Pathol.",
  ["international journal of gynecological pathology : official journal of the international society of gynecological pathologists"] = "Int J Gynecol Pathol",
  ["international journal of head and neck surgery"] = "Int J Head Neck Surg",
  ["international journal of health and nutrition"] = "Int J Health Nutr",
  ["international journal of health care finance and economics"] = "Int J Health Care Finance Econ",
  ["international journal of health care quality assurance"] = "Int J Health Care Qual Assur",
  ["international journal of health care quality assurance incorporating leadership in health services"] = "Int J Health Care Qual Assur Inc Leadersh Health Serv",
  ["international journal of health economics and management"] = "Int J Health Econ Manag",
  ["international journal of health education"] = "Int J Health Educ",
  ["international journal of health geographics"] = "Int J Health Geogr",
  ["international journal of health planning and management"] = "Int. J. Health Plann. Manage.",
  ["international journal of health policy and management"] = "Int J Health Policy Manag",
  ["international journal of health professions"] = "Int J Health Prof",
  ["international journal of health promotion and education"] = "Int J Health Promot Educ",
  ["international journal of health sciences"] = "Int J Health Sci (Qassim)",
  ["international journal of health sciences and research"] = "Int J Health Sci Res",
  ["international journal of health sciences education"] = "Int J Health Sci Educ",
  ["international journal of health services"] = "Int. J. Health Serv.",
  ["international journal of health services : planning, administration, evaluation"] = "Int J Health Serv",
  ["international journal of healthcare"] = "Int J Healthc",
  ["international journal of healthcare and medical sciences"] = "Int J Healthc Med Sci",
  ["international journal of healthcare information systems and informatics : official publication of the information resources management association"] = "Int J Healthc Inf Syst Inform",
  ["international journal of healthcare management"] = "Int J Healthc Manag",
  ["international journal of heat and fluid flow"] = "Int. J. Heat Fluid Flow",
  ["international journal of heat and mass transfer"] = "Int. J. Heat Mass Transfer",
  ["international journal of heat and technology"] = "Int. J. Heat Technol.",
  ["international journal of heat exchangers"] = "Int. J. Heat Exch.",
  ["international journal of heavy vehicle systems"] = "Int. J. Heavy Veh. Syst.",
  ["international journal of hematologic oncology"] = "Int J Hematol Oncol",
  ["international journal of hematology"] = "Int J Hematol",
  ["international journal of hematology & therapy"] = "Int J Hematol Ther",
  ["international journal of hematology research"] = "Int J Hematol Res",
  ["international journal of hematology-oncology and stem cell research"] = "Int J Hematol Oncol Stem Cell Res",
  ["international journal of hepatology"] = "Int J Hepatol",
  ["international journal of herbal medicine"] = "Int J Herb Med",
  ["international journal of high performance computing and networking"] = "Int. J. High Perform. Comput. Networking",
  ["international journal of high performance computing applications"] = "Int. J. High Perform. Comput. Appl.",
  ["international journal of high performance systems architecture"] = "Int. J. High Perform. Syst. Archit.",
  ["international journal of high risk behaviors & addiction"] = "Int J High Risk Behav Addict",
  ["international journal of high speed computing"] = "Int. J. High Speed Comput.",
  ["international journal of high speed electronics and systems"] = "Int. J. High Speed Electron. Syst.",
  ["international journal of high technology ceramics"] = "Int. J. Hig Technol. Ceram.",
  ["international journal of high throughput screening"] = "Int J High Throughput Screen",
  ["international journal of hispanic psychology"] = "Int J Hisp Psychol",
  ["international journal of historical archaeology"] = "Int J Hist Archaeol",
  ["international journal of hiv/aids and research"] = "Int J HIV AIDS Res",
  ["international journal of horticultural science"] = "Int J Hortic Sci",
  ["international journal of horticultural science and technology"] = "Int J Hortic Sci Technol",
  ["international journal of hospitality management"] = "Int J Hosp Manag",
  ["international journal of hr : humanoid robotics"] = "Int J HR",
  ["international journal of human ecology"] = "Int J Hum Ecol",
  ["international journal of human factors and ergonomics"] = "Int J Hum Factors Ergon",
  ["international journal of human genetics"] = "Int J Hum Genet",
  ["international journal of human genetics and genetic disorders"] = "Int J Hum Genet Genet Disord",
  ["international journal of human rights in healthcare"] = "Int J Hum Rights Healthc",
  ["international journal of human-computer interaction"] = "Int J Hum Comput Interact",
  ["international journal of human-computer studies"] = "Int J Hum Comput Stud",
  ["international journal of humanities and social science"] = "Int J Humanit Soc Sci",
  ["international journal of humanoid robotics"] = "Int. J. Humanoid Rob.",
  ["international journal of hybrid intelligent systems"] = "Int J Hybrid Intell Syst",
  ["international journal of hybrid microelectronics"] = "Int. J. Hybrid Microelectron.",
  ["international journal of hydrocarbon engineering"] = "Int. J. Hydrocarbon Eng.",
  ["international journal of hydrogen energy"] = "Int. J. Hydrogen Energy",
  ["international journal of hydrology science and technology"] = "Int. J. Hydrol. Sci. Technol.",
  ["international journal of hydromechatronics"] = "Int. J. Hydromechatron.",
  ["international journal of hygiene and environmental health"] = "Int J Hyg Environ Health",
  ["international journal of hypertension"] = "Int J Hypertens",
  ["international journal of hyperthermia"] = "Int. J. Hyperthermia",
  ["international journal of hyperthermia : the official journal of european society for hyperthermic oncology, north american hyperthermia group"] = "Int J Hyperthermia",
  ["international journal of image and data fusion"] = "Int J Image Data Fusion",
  ["international journal of image and graphics"] = "Int J Image Graph",
  ["international journal of imaging systems and technology"] = "Int. J. Imaging Syst. Technol.",
  ["international journal of immunogenetics"] = "Int J Immunogenet",
  ["international journal of immunological studies"] = "Int J Immunol Stud",
  ["international journal of immunology and immunotherapy"] = "Int J Immunol Immunother",
  ["international journal of immunopathology and pharmacology"] = "Int J Immunopathol Pharmacol",
  ["international journal of immunopharmacology"] = "Int J Immunopharmacol",
  ["international journal of immunotherapy and cancer research"] = "Int J Immunother Cancer Res",
  ["international journal of impact engineering"] = "Int J Impact Eng",
  ["international journal of implant dentistry"] = "Int J Implant Dent",
  ["international journal of impotence research"] = "Int J Impot Res",
  ["international journal of indigenous health"] = "Int J Indig Health",
  ["international journal of industrial and systems engineering"] = "Int. J. Ind. Syst. Eng.",
  ["international journal of industrial chemistry"] = "Int. J. Ind. Chem.",
  ["international journal of industrial engineering and operations management"] = "Int. J. Ind. Eng. Oper. Manage.",
  ["international journal of industrial engineering: theory applications and practice"] = "Int. J. Ind. Eng.Theory Appl. Pract.",
  ["international journal of industrial entomology"] = "Int J Indust Entomol",
  ["international journal of industrial ergonomics"] = "Int J Ind Ergon",
  ["international journal of industrial organization"] = "Int. J. Ind. Organ.",
  ["international journal of infection control"] = "Int J Infect Control",
  ["international journal of infectious diseases"] = "Int. J. Infect. Dis.",
  ["international journal of infectious diseases : ijid : official publication of the international society for infectious diseases"] = "Int J Infect Dis",
  ["international journal of inflammation"] = "Int J Inflam",
  ["international journal of information & systems sciences"] = "Int. J. Inf. Syst. Sci.",
  ["international journal of information acquisition"] = "Int. J. Inf. Acquis.",
  ["international journal of information and coding theory. ijicot"] = "Int. J. Inf. Coding Theory",
  ["international journal of information and communication technology"] = "Int. J. Inf. Commun. Technol.",
  ["international journal of information and communication technology education : an official publication of the information resources management association"] = "Int J Inf Commun Technol Educ",
  ["international journal of information and computer security"] = "Int. J. Inf. Comput. Secur.",
  ["international journal of information and education technology (ijiet)"] = "Int J Inf Educ Technol",
  ["international journal of information and management sciences"] = "Int. J. Inf. Manage. Sci.",
  ["international journal of information management"] = "Int. J. Inf. Manage.",
  ["international journal of information management data insights"] = "Int. J. Inf. Manage. Data Insights",
  ["international journal of information quality"] = "Int. J. Inf. Qual.",
  ["international journal of information security"] = "Int. J. Inf. Secur.",
  ["international journal of information systems and change management"] = "Int. J. Inf. Syst. Change Mange.",
  ["international journal of information technology"] = "Int. J. Inf. Technol.",
  ["international journal of information technology & decision making"] = "Int J Inf Technol Decis Mak",
  ["international journal of information technology : an official journal of bharati vidyapeeth's institute of computer applications and management"] = "Int J Inf Technol",
  ["international journal of information technology and management"] = "Int. J. Inf. Technol. Manage.",
  ["international journal of information technology and web engineering"] = "Int. J. Inf. Technol. Web. Eng.",
  ["international journal of infrared and millimeter waves"] = "Int J Infrared Millimeter Waves",
  ["international journal of injury control and safety promotion"] = "Int J Inj Contr Saf Promot",
  ["international journal of innovation and applied studies"] = "Int J Innov Appl Stud",
  ["international journal of innovation and sustainable development"] = "Int. J. Innovation Sustainable Dev.",
  ["international journal of innovation and technology management"] = "Int. J. Innovation Technol. Manage.",
  ["international journal of innovation management"] = "Int. J. Innovation Manage.",
  ["international journal of innovation science"] = "Int. J. Innovation Sci.",
  ["international journal of innovation studies"] = "Int. J. Innovation Stud.",
  ["international journal of innovative computing and applications"] = "Int. J. Innovative Comput. Appl.",
  ["international journal of innovative medicine and health science"] = "Int J Innov Med Health Sci",
  ["international journal of innovative research"] = "Int J Innov Res (Patuakhali)",
  ["international journal of innovative research and studies"] = "Int J Innov Res Stud",
  ["international journal of innovative research in science, engineering and technology"] = "Int J Innov Res Sci Eng Technol",
  ["international journal of innovative science and research technology"] = "Int J Innov Sci Res Technol",
  ["international journal of inorganic chemistry"] = "Int J Inorg Chem",
  ["international journal of inorganic materials"] = "Int. J. Inorg. Mater.",
  ["international journal of insect morphology & embryology"] = "Int J Insect Morphol Embryol",
  ["international journal of insect morphology and embryology"] = "Int. J. Insect Morphol. Embryol.",
  ["international journal of insect science"] = "Int J Insect Sci",
  ["international journal of instructional media"] = "Int J Instr Media",
  ["international journal of integrated care"] = "Int J Integr Care",
  ["international journal of integrative biology"] = "Int J Integr Biol",
  ["international journal of integrative medicine"] = "Int J Integr Med",
  ["international journal of integrative pediatrics and environmental medicine"] = "Int J Integr Pediatr Environ Med",
  ["international journal of intelligence and counterintelligence"] = "Int. J. Intell. CounterIntell.",
  ["international journal of intelligence science"] = "Int J Intell Sci",
  ["international journal of intelligent computing and cybernetics"] = "Int. J. Intell. Comput. Cybern.",
  ["international journal of intelligent engineering informatics"] = "Int. J. Intell. Eng. Inf.",
  ["international journal of intelligent information and database systems"] = "Int. J. Intell. Inf. Database Syst.",
  ["international journal of intelligent information technologies"] = "Int. J. Intell. Inf. Technol.",
  ["international journal of intelligent networks"] = "Int. J. Intell. Networks",
  ["international journal of intelligent robotics and applications"] = "Int J Intell Robot Appl",
  ["international journal of intelligent systems"] = "Int. J. Intell. Syst.",
  ["international journal of intelligent transportation systems research"] = "Int. J. Intell. Transp. Syst. Res.",
  ["international journal of interactive multimedia and artificial intelligence"] = "Int. J. Interact. Multimedia Artif. Intell.",
  ["international journal of intercultural relations"] = "Int. J. Intercult. Relat.",
  ["international journal of intercultural relations : ijir"] = "Int J Intercult Relat",
  ["international journal of interferon, cytokine and mediator research"] = "Int  J Interferon Cytokine Mediat Res",
  ["international journal of internet manufacturing and services"] = "Int. J. Internet Manuf. Serv.",
  ["international journal of internet protocol technology"] = "Int. J. Internet Protoc. Technol.",
  ["international journal of ion mobility spectrometry"] = "Int. J. Ion Mobility Spectrom.",
  ["international journal of keratoconus and ectatic corneal diseases"] = "Int J Keratoconus Ectatic Corneal Dis",
  ["international journal of knowledge engineering and soft data paradigms"] = "Int J Knowl Eng Soft Data Paradig",
  ["international journal of knowledge-based and intelligent engineering systems"] = "Int. J. Knowledge-Based Intell. Eng. Syst.",
  ["international journal of laboratory hematology"] = "Int J Lab Hematol",
  ["international journal of language & communication disorders"] = "Int J Lang Commun Disord",
  ["international journal of language & communication disorders / royal college of speech & language therapists"] = "Int J Lang Commun Disord",
  ["international journal of language and communication disorders"] = "Int. J. Lang. Commun. Disord.",
  ["international journal of latest trends in engineering & technology : ijltet"] = "Int J Latest Trends Eng Technol",
  ["international journal of law and psychiatry"] = "Int J Law Psychiatry",
  ["international journal of law and the family"] = "Int J Law Fam",
  ["international journal of law in context"] = "Int J Law Context",
  ["international journal of law in the built environment"] = "Int. J. Law Built Environ.",
  ["international journal of law, crime and justice"] = "Int J Law Crime Justice",
  ["international journal of law, policy, and the family"] = "Int J Law Policy Family",
  ["international journal of learning and change"] = "Int J Learn Chang",
  ["international journal of legal medicine"] = "Int J Legal Med",
  ["international journal of leprosy"] = "Int J Lepr",
  ["international journal of leprosy and other mycobacterial diseases"] = "Int. J. Lepr. Other Mycobact. Dis.",
  ["international journal of leprosy and other mycobacterial diseases : official organ of the international leprosy association"] = "Int J Lepr Other Mycobact Dis",
  ["international journal of liberal arts and social science"] = "Int J Lib Arts Soc Sci",
  ["international journal of life cycle assessment"] = "Int. J. Life Cycle Assess.",
  ["international journal of life science & pharma research"] = "Int J Life Sci Pharma Res",
  ["international journal of life sciences (amravati)"] = "Int J Life Sci",
  ["international journal of life sciences and technology"] = "Int J Life Sci Technol",
  ["international journal of life sciences biotechnology and pharma research"] = "Int J Life Sci Biotechnol Pharma Res",
  ["international journal of life sciences research"] = "Int J Life Sci Res",
  ["international journal of lightweight materials and manufacture"] = "Int. J. Lightweight Mater. Manuf.",
  ["international journal of limnology"] = "Int. J. Limnol.",
  ["international journal of logistics research and applications"] = "Int. J. Logist. Res. Appl.",
  ["international journal of low carbon technologies"] = "Int. J. Low Carbon Technol.",
  ["international journal of low radiation"] = "Int J Low Radiat",
  ["international journal of low-carbon technologies"] = "Int. J. Low-Carbon Technol.",
  ["international journal of lower extremity wounds"] = "Int. J. Low. Extrem. Wounds",
  ["international journal of machine consciousness"] = "Int. J. Mach. Conscious.",
  ["international journal of machine learning and computing"] = "Int J Mach Learn Comput",
  ["international journal of machine learning and cybernetics"] = "Int. J. Mach. Learn. Cybern.",
  ["international journal of machine tool design and research"] = "Int. J. Mach. Tool Des. Res.",
  ["international journal of machine tools & manufacture"] = "Int J Mach Tools Manuf",
  ["international journal of machine tools and manufacture"] = "Int. J. Mach. Tools Manuf",
  ["international journal of machining and machinability of materials"] = "Int. J. Mach. Mach. Mater.",
  ["international journal of magnetism"] = "Int. J. Magn.",
  ["international journal of malacology"] = "Int. J. Malacol.",
  ["international journal of man-machine studies"] = "Int. J. Man Mach. Stud.",
  ["international journal of management and business"] = "Int J Manag Bus",
  ["international journal of management and sustainability"] = "Int. J. Manage. Sustainability",
  ["international journal of management science and engineering management"] = "Int. J. Manage. Sci. Eng. Manage.",
  ["international journal of manpower"] = "Int J Manpow",
  ["international journal of manufacturing research"] = "Int. J. Manuf. Res.",
  ["international journal of manufacturing technology and management"] = "Int. J. Manuf. Technol. Manage.",
  ["international journal of manufacturing, materials, and mechanical engineering"] = "Int. J. Manuf. Mater. Mech. Eng.",
  ["international journal of maps in mathematics"] = "Int. J. Maps Math.",
  ["international journal of marine energy"] = "Int. J. Mar. Energy",
  ["international journal of marine science"] = "Int. J. Mar. Sci.",
  ["international journal of maritime economics"] = "Int. J. Maritime Econ.",
  ["international journal of maritime engineering"] = "Int. J. Marit. Eng.",
  ["international journal of maritime history"] = "Int J Marit Hist",
  ["international journal of marketing studies"] = "Int J Mark Stud",
  ["international journal of masonry research and innovation"] = "Int. J. Masonry Res. Innov.",
  ["international journal of mass emergencies and disasters"] = "Int J Mass Emerg Disasters",
  ["international journal of mass spectrometry"] = "Int J Mass Spectrom",
  ["international journal of mass spectrometry and ion physics"] = "Int. J. Mass Spectrom. Ion Phys.",
  ["international journal of mass spectrometry and ion processes"] = "Int J Mass Spectrom Ion Process",
  ["international journal of material forming"] = "Int. J. Mater. Form.",
  ["international journal of material science"] = "Int  J Mater Sci",
  ["international journal of materials and product technology"] = "Int. J. Mater. Prod. Technol.",
  ["international journal of materials and structural integrity"] = "Int. J. Mater. Struct. Integrity",
  ["international journal of materials engineering innovation"] = "Int. J. Mater. Eng. Innovation",
  ["international journal of materials in engineering applications"] = "Int. J. Mater. Eng. Appl.",
  ["international journal of materials research"] = "Int. J. Mater. Res.",
  ["international journal of mathematical and statistical sciences"] = "Int. J. Math. Stat. Sci.",
  ["international journal of mathematical education in science and technology"] = "Int. J. Math. Educ. Sci. Technol.",
  ["international journal of mathematical, computational, physical, electrical and computer engineering"] = "Int. J. Math. Comput. Phys. Electr. Comput. Eng.",
  ["international journal of mathematics"] = "Int. J. Math.",
  ["international journal of mathematics and analysis. new series"] = "Int. J. Math. Anal. (N.S.)",
  ["international journal of mathematics and computer science"] = "Int J Math Comput Sci",
  ["international journal of mathematics and mathematical sciences"] = "Int. J. Math. Math. Sci.",
  ["international journal of mathematics and statistics"] = "Int. J. Math. Stat.",
  ["international journal of mathematics for industry"] = "Int. J. Math. Ind.",
  ["international journal of mathematics in operational research"] = "Int. J. Math. Oper. Res.",
  ["international journal of mathematics, game theory, and algebra"] = "Int. J. Math. Game Theory Algebra",
  ["international journal of mch and aids"] = "Int J MCH AIDS",
  ["international journal of mechanical and materials engineering"] = "Int. J. Mech. Mater. Eng.",
  ["international journal of mechanical engineering education"] = "Int. J. Mech. Eng. Educ.",
  ["international journal of mechanical sciences"] = "Int. J. Mech. Sci.",
  ["international journal of mechanical systems dynamics"] = "Int. J. Mech. Syst. Dyn.",
  ["international journal of mechanical systems science and engineering"] = "Int. J. Mech. Syst. Sci. Eng.",
  ["international journal of mechanics and materials in design"] = "Int. J. Mech. Mater. Des.",
  ["international journal of mechatronics and manufacturing systems"] = "Int. J. Mechatron. Manuf. Syst.",
  ["international journal of medical and biological frontiers"] = "Int J Med Biol Front",
  ["international journal of medical and biomedical sciences"] = "Int J Med Biomed Sci",
  ["international journal of medical and biomedical studies"] = "Int J Med Biomed Stud",
  ["international journal of medical and pharmaceutical case reports"] = "Int J Med Pharm Case Reports",
  ["international journal of medical biochemistry"] = "Int. J. Med. Biochem.",
  ["international journal of medical education"] = "Int J Med Educ",
  ["international journal of medical engineering and informatics"] = "Int. J. Med. Eng. Inf.",
  ["international journal of medical informatics"] = "Int J Med Inform",
  ["international journal of medical microbiology"] = "Int. J. Med. Microbiol.",
  ["international journal of medical microbiology : ijmm"] = "Int J Med Microbiol",
  ["international journal of medical microbiology, virology, parasitology and infectious diseases"] = "Int. J. Med. Microbiol. Virol. Parasitol. Infect. Dis.",
  ["international journal of medical nano research"] = "Int J Med Nano Res",
  ["international journal of medical physics, clinical engineering and radiation oncology"] = "Int J Med Phys Clin Eng Radiat Oncol",
  ["international journal of medical robotics and computer assisted surgery"] = "Int. J. Med. Rob. Comput. Assisted Surg.",
  ["international journal of medical science and clinical invention"] = "Int J Med Sci Clin Invent",
  ["international journal of medical science and public health"] = "Int J Med Sci Public Health",
  ["international journal of medical sciences"] = "Int J Med Sci",
  ["international journal of medical toxicology and forensic medicine"] = "Int. J. Med. Toxicol. Forensic Med.",
  ["international journal of medical toxicology and legal medicine"] = "Int. J. Med. Toxicol. Legal Med.",
  ["international journal of medicinal chemistry"] = "Int J Med Chem",
  ["international journal of medicinal mushrooms"] = "Int J Med Mushrooms",
  ["international journal of medicine"] = "Int J Med (Dubai)",
  ["international journal of medicine and public health"] = "Int J Med Public Health",
  ["international journal of men's health"] = "Int J Mens Health",
  ["international journal of mental health"] = "Int J Ment Health",
  ["international journal of mental health & psychiatry"] = "Int J Ment Health Psychiatry",
  ["international journal of mental health and addiction"] = "Int J Ment Health Addict",
  ["international journal of mental health nursing"] = "Int J Ment Health Nurs",
  ["international journal of mental health systems"] = "Int J Ment Health Syst",
  ["international journal of metadata, semantics and ontologies"] = "Int. J. Metadata Semant. Ontol.",
  ["international journal of metalcasting"] = "Int. J. Metalcast.",
  ["international journal of methods in psychiatric research"] = "Int J Methods Psychiatr Res",
  ["international journal of metrology and quality engineering"] = "Int. J. Metrol. Qual. Eng.",
  ["international journal of micro air vehicles"] = "Int. J. Micro Air Veh.",
  ["international journal of micro-nano scale transport"] = "Int J Micronano Scale Transp",
  ["international journal of microbiology"] = "Int J Microbiol",
  ["international journal of microbiology and biotechnology"] = "Int J Microbiol Biotechnol",
  ["international journal of microbiology and current research"] = "Int J Microbiol Curr Res",
  ["international journal of microbiology and immunology research"] = "Int J Microbiol Immunol Res",
  ["international journal of microbiology research"] = "Int J Microbiol Res",
  ["international journal of microcircuits and electronic packaging"] = "Int. J. Microcircuits Electron Packag.",
  ["international journal of microcirculation, clinical and experimental"] = "Int J Microcirc Clin Exp",
  ["international journal of microcirculation, clinical and experimental / sponsored by the european society for microcirculation"] = "Int J Microcirc Clin Exp",
  ["international journal of microcirculation; clinical and experimental"] = "Int. J. Microcirc. Clin. Exp.",
  ["international journal of microsimulation"] = "Int J Microsimul",
  ["international journal of microstructure and materials properties"] = "Int. J. Microstruct. Mater. Prop.",
  ["international journal of microwave and wireless technologies"] = "Int. J. Microwave Wireless Technolog.",
  ["international journal of middle east studies"] = "Int J Middle East Stud",
  ["international journal of migration and border studies"] = "Int J Migr Bord Stud",
  ["international journal of migration, health, and social care"] = "Int J Migr Health Soc Care",
  ["international journal of mine water"] = "Int. J. Mine Water",
  ["international journal of mineral processing"] = "Int. J. Miner. Process.",
  ["international journal of minerals, metallurgy, and materials"] = "Int. J. Miner. Metall. Mater.",
  ["international journal of mining and geo-engineering"] = "Int. J. Min. Geo-Eng.",
  ["international journal of mining and mineral engineering"] = "Int J Min Miner Eng",
  ["international journal of mining and mineral processing engineering"] = "Int. J. Min. Miner. Eng.",
  ["international journal of mining science and technology"] = "Int J Min Sci Technol",
  ["international journal of mining, reclamation and environment"] = "Int. J. Min. Reclam. Environ.",
  ["international journal of mobile communications"] = "Int. J. Mobile Commun.",
  ["international journal of mobile network design and innovation"] = "Int. J. Mobile Network Des. Innov.",
  ["international journal of modeling, simulation, and scientific computing"] = "Int. J. Model., Simul., Sci. Comput.",
  ["international journal of modelling and simulation"] = "Int. J. Model. Simul.",
  ["international journal of modelling identification and control"] = "Int J Model Identif Control",
  ["international journal of modelling, identification and control"] = "Int. J. Model. Ident. Control",
  ["international journal of modern chemistry"] = "Int. J. Mod. Chem.",
  ["international journal of modern organic chemistry"] = "Int. J. Mod. Org. Chem.",
  ["international journal of modern physics"] = "Int. J. Modern Phys. D",
  ["international journal of modern physics a"] = "Int. J. Mod. Phys. A",
  ["international journal of modern physics a. particles and fields. gravitation. cosmology"] = "Internat. J. Modern Phys. A",
  ["international journal of modern physics a. particles and fields. gravitation. cosmology. astrophysics. accelerator physics"] = "Internat. J. Modern Phys. A",
  ["international journal of modern physics a: particles and fields; gravitation; cosmology; nuclear physics"] = "Int. J. Mod. Phys. A",
  ["international journal of modern physics b"] = "Int. J. Mod. Phys. B",
  ["international journal of modern physics b: condensed matter physics, statistical physics, applied physics"] = "Int. J. Mod. Phys. B",
  ["international journal of modern physics c"] = "Int. J. Mod. Phys. C",
  ["international journal of modern physics c. computational physics and physical computation"] = "Internat. J. Modern Phys. C",
  ["international journal of modern physics c: computational physics and physical computation"] = "Int. J. Mod. Phys. C",
  ["international journal of modern physics d"] = "Int. J. Mod. Phys. D",
  ["international journal of modern physics d: gravitation; astrophysics and cosmology"] = "Int. J. Mod. Phys. D",
  ["international journal of modern physics e"] = "Int. J. Mod. Phys. E",
  ["international journal of modern physics e: nuclear physics"] = "Int. J. Mod. Phys. E",
  ["international journal of modern physics. a, particles and fields, gravitation, cosmology"] = "Int J Mod Phys A",
  ["international journal of modern physics. b"] = "Int J Mod Phys B",
  ["international journal of modern physics. c, physics and computers"] = "Int J Mod Phys C",
  ["international journal of modern physics. conference series"] = "Int J Mod Phys Conf Ser",
  ["international journal of modern physics. d, gravitation, astrophysics and cosmology"] = "Int J Mod Phys D Gravit Astrophys Cosmol",
  ["international journal of modern physics. d. gravitation, astrophysics, cosmology"] = "Internat. J. Modern Phys. D",
  ["international journal of modern physics: conference series"] = "Int. J. Mod. Phys. Conf. Ser.",
  ["international journal of modern science and technology (kumbakonam)"] = "Int J Mod Sci Technol (Kumbakonam)",
  ["international journal of molecular and cellular medicine"] = "Int J Mol Cell Med",
  ["international journal of molecular and clinical microbiology"] = "Int J Mol Clin Microbiol",
  ["international journal of molecular biology (edmond, okla.)"] = "Int J Mol Biol (Edmond)",
  ["international journal of molecular epidemiology and genetics"] = "Int J Mol Epidemiol Genet",
  ["international journal of molecular imaging"] = "Int J Mol Imaging",
  ["international journal of molecular medical science"] = "Int J Mol Med Sci",
  ["international journal of molecular medicine"] = "Int J Mol Med",
  ["international journal of molecular medicine and advance sciences"] = "Int. J. Mol. Med. Adv. Sci.",
  ["international journal of molecular sciences"] = "Int J Mol Sci",
  ["international journal of monitoring and surveillance technologies research"] = "Int J Monit Surveill Technol Res",
  ["international journal of morphology"] = "Int. J. Morphol.",
  ["international journal of morphology = revista internacional de morfología"] = "Int J Morphol",
  ["international journal of mosquito research"] = "Int J Mosq Res",
  ["international journal of ms care"] = "Int J MS Care",
  ["international journal of multilingualism"] = "Int J Multiling",
  ["international journal of multimedia information retrieval"] = "Int J Multimed Inf Retr",
  ["international journal of multiphase flow"] = "Int. J. Multiphase Flow",
  ["international journal of multiphysics"] = "Int. J. Multiphys.",
  ["international journal of multiple research approaches"] = "Int J Mult Res Approaches",
  ["international journal of mycobacteriology"] = "Int J Mycobacteriol",
  ["international journal of myriapodology"] = "Int J Myriap",
  ["international journal of nano and biomaterials"] = "Int. J. Nano Biomater.",
  ["international journal of nano studies & technology"] = "Int J Nano Stud Technol",
  ["international journal of nanodimension"] = "Int J Nanodimens",
  ["international journal of nanoelectronics and materials"] = "Int. J. Nanoelectron. Mater.",
  ["international journal of nanomanufacturing"] = "Int. J. Nanomanuf.",
  ["international journal of nanomedicine"] = "Int J Nanomedicine",
  ["international journal of nanomedicine and nanosurgery"] = "Int J Nanomed Nanosurg",
  ["international journal of nanoparticles"] = "Int. J. Nanopart.",
  ["international journal of nanoscience"] = "Int. J. Nanosci.",
  ["international journal of nanoscience and nanotechnology"] = "Int. J. Nanosci. Nanotechnol.",
  ["international journal of nanotechnology"] = "Int. J. Nanotechnol.",
  ["international journal of nanotechnology and applications"] = "Int J Nanotechnol Appl",
  ["international journal of nanotechnology and nanomedicine"] = "Int J Nanotechnol Nanomed",
  ["international journal of nanotechnology in medicine & engineering"] = "Int J Nanotechnol Eng Med",
  ["international journal of natural and engineering sciences"] = "Int. J. Nat. Eng. Sci.",
  ["international journal of nautical archaeology"] = "Int. J. Naut. Archaeol.",
  ["international journal of naval architecture and ocean engineering"] = "Int. J. Nav. Archit. Ocean Eng.",
  ["international journal of nematology"] = "Int J Nematol",
  ["international journal of neonatal screening"] = "Int J Neonatal Screen",
  ["international journal of nephrology"] = "Int J Nephrol",
  ["international journal of nephrology & urology"] = "Int J Nephrol Urol",
  ["international journal of nephrology and kidney failure"] = "Int J Nephrol Kidney Fail",
  ["international journal of nephrology and renovascular disease"] = "Int J Nephrol Renovasc Dis",
  ["international journal of network management"] = "Int. J. Network Manage.",
  ["international journal of networked and distributed computing"] = "Int. J. Networked Distrib. Comput.",
  ["international journal of networking and virtual organisations"] = "Int. J. Networking Virtual Organ.",
  ["international journal of neural systems"] = "Int J Neural Syst",
  ["international journal of neurodegenerative disorders"] = "Int J Neurodegener Dis",
  ["international journal of neurology"] = "Int J Neurol",
  ["international journal of neurology & brain disorders"] = "Int J Neurol Brain Disord",
  ["international journal of neurology and neurotherapy"] = "Int J Neurol Neurother",
  ["international journal of neuropharmacology"] = "Int J Neuropharmacol",
  ["international journal of neuroprotection and neuroregeneration"] = "Int J Neuroprot Neuroregener",
  ["international journal of neuropsychiatry"] = "Int J Neuropsychiatry",
  ["international journal of neuropsychopharmacology"] = "Int. J. Neuropsychopharmacol.",
  ["international journal of neurorehabilitation"] = "Int J Neurorehabil",
  ["international journal of neuroscience"] = "Int. J. Neurosci.",
  ["international journal of neuroscience and neuroinformatics"] = "Int. J. Neurosci. Neuroinf.",
  ["international journal of non-equilibrium processing"] = "Int. J. Non Equilibr. Process",
  ["international journal of non-linear mechanics"] = "Int. J. Non Linear Mech.",
  ["international journal of noncommunicable diseases"] = "Int J Noncommun Dis",
  ["international journal of nondestructive testing"] = "Int. J. Nondestr. Test.",
  ["international journal of nonferrous metallurgy"] = "Int. J. Nonferrous Metall.",
  ["international journal of nonlinear modelling in science and engineering"] = "Int. J. Nonlinear Modell. Sci. Eng.",
  ["international journal of nonlinear science"] = "Int. J. Nonlinear Sci.",
  ["international journal of nonlinear sciences and numerical simulation"] = "Int. J. Nonlinear Sci. Numer. Simul.",
  ["international journal of nonlineare modelling in science and engineering"] = "Int. J. Nonlinear Model. Sci. Eng.",
  ["international journal of nuclear desalination"] = "Int. J. Nucl. Desalin.",
  ["international journal of nuclear energy science and technology"] = "Int. J. Nucl. Energy Sci. Technol.",
  ["international journal of nuclear medicine and biology"] = "Int. J. Nucl. Med. Biol.",
  ["international journal of number theory"] = "Int. J. Number Theory",
  ["international journal of numerical analysis & modeling. series b"] = "Int J Numer Anal Model B",
  ["international journal of numerical analysis and modeling"] = "Int. J. Numer. Anal. Model.",
  ["international journal of numerical analysis and modeling. series b"] = "Int. J. Numer. Anal. Model. Ser. B",
  ["international journal of numerical methods for heat & fluid flow"] = "Int J Numer Methods Heat Fluid Flow",
  ["international journal of numerical methods for heat and fluid flow"] = "Int. J. Numer. Methods Heat Fluid Flow",
  ["international journal of numerical methods in biomedical engineering"] = "Int. J. Numer. Methods Biomed. Eng.",
  ["international journal of numerical modelling: electronic networks, devices and fields"] = "Int. J. Numer. Model. Electron. Networks Devices Fields",
  ["international journal of nursing & clinical practices"] = "Int J Nurs Clin Pract",
  ["international journal of nursing (new york, n.y.)"] = "Int J Nurs (N Y)",
  ["international journal of nursing education scholarship"] = "Int J Nurs Educ Scholarsh",
  ["international journal of nursing knowledge"] = "Int J Nurs Knowl",
  ["international journal of nursing practice"] = "Int J Nurs Pract",
  ["international journal of nursing research"] = "Int J Nurs Res",
  ["international journal of nursing sciences"] = "Int J Nurs Sci",
  ["international journal of nursing studies"] = "Int J Nurs Stud",
  ["international journal of nursing studies advances"] = "Int J Nurs Stud Adv",
  ["international journal of nursing terminologies and classifications : the official journal of nanda international"] = "Int J Nurs Terminol Classif",
  ["international journal of nutrition"] = "Int J Nutr",
  ["international journal of nutrition and metabolism"] = "Int J Nutr Metab",
  ["international journal of nutrition, pharmacology, neurological diseases"] = "Int J Nutr Pharmacol Neurol Dis",
  ["international journal of obesity"] = "Int J Obes",
  ["international journal of obesity (2005)"] = "Int J Obes (Lond)",
  ["international journal of obesity and nutritional science"] = "Int J Obes Nutr Sci",
  ["international journal of obesity and related metabolic disorders"] = "Int. J. Obes. Relat. Metab. Disord.",
  ["international journal of obesity and related metabolic disorders : journal of the international association for the study of obesity"] = "Int J Obes Relat Metab Disord",
  ["international journal of obesity supplements"] = "Int J Obes Suppl",
  ["international journal of obstetric anesthesia"] = "Int J Obstet Anesth",
  ["international journal of occupational and environmental health"] = "Int J Occup Environ Health",
  ["international journal of occupational health and safety"] = "Int. J. Occup. Health Saf.",
  ["international journal of occupational medicine and environmental health"] = "Int J Occup Med Environ Health",
  ["international journal of occupational safety and ergonomics"] = "Int. J. Occup. Saf. Ergon.",
  ["international journal of occupational safety and ergonomics : jose"] = "Int J Occup Saf Ergon",
  ["international journal of occupational safety and health"] = "Int J Occup Saf Health",
  ["international journal of ocean and coastal engineering"] = "Int. J. Ocean Coastal Eng.",
  ["international journal of oceanology and limnology"] = "Int. J. Oceanol. Limnol.",
  ["international journal of odonatology"] = "Int. J. Odonatol.",
  ["international journal of odonatology : official organ of the worldwide dragonfly association"] = "Int J Odonatol",
  ["international journal of offender therapy and comparative criminology"] = "Int J Offender Ther Comp Criminol",
  ["international journal of offshore and polar engineering"] = "Int. J. Offshore Polar Eng.",
  ["international journal of oil, gas and coal technology"] = "Int. J. Oil Gas Coal Technol.",
  ["international journal of older people nursing"] = "Int J Older People Nurs",
  ["international journal of oncology"] = "Int J Oncol",
  ["international journal of one health"] = "Int J One Health",
  ["international journal of operational research"] = "Int. J. Oper. Res.",
  ["international journal of operational research nepal. ijorn"] = "Int. J. Oper. Res. Nepal",
  ["international journal of operations & production management"] = "Int. J. Oper. Prod. Manage.",
  ["international journal of operations research"] = "Int. J. Oper. Res. (Taichung)",
  ["international journal of ophthalmic pathology"] = "Int J Ophthalmic Pathol",
  ["international journal of ophthalmic research"] = "Int J Ophthalmic Res",
  ["international journal of ophthalmology"] = "Int J Ophthalmol",
  ["international journal of ophthalmology & eye science"] = "Int J Ophthalmol Eye Sci",
  ["international journal of ophthalmology and clinical research"] = "Int J Ophthalmol Clin Res",
  ["international journal of optics"] = "Int J Opt",
  ["international journal of optimization. theory methods and applications"] = "Int. J. Optim. Theory Methods Appl.",
  ["international journal of optomechatronics"] = "Int. J. Optomechatronics",
  ["international journal of oral and maxillofacial implants"] = "Int. J. Oral Maxillofac. Implants",
  ["international journal of oral and maxillofacial surgery"] = "Int J Oral Maxillofac Surg",
  ["international journal of oral biology : official journal of the korean academy of oral biology and the ucla dental research institute"] = "Int J Oral Biol",
  ["international journal of oral implantology"] = "Int. J. Oral Implantol.",
  ["international journal of oral implantology (berlin, germany)"] = "Int J Oral Implantol (Berl)",
  ["international journal of oral myology"] = "Int. J. Oral Myol.",
  ["international journal of oral science"] = "Int J Oral Sci",
  ["international journal of oral surgery"] = "Int J Oral Surg",
  ["international journal of organ transplantation medicine"] = "Int J Organ Transplant Med",
  ["international journal of organic chemistry"] = "Int J Org Chem (Irvine)",
  ["international journal of orofacial myology"] = "Int. J. Orofacial Myology",
  ["international journal of orthodontics"] = "Int J Orthod",
  ["international journal of orthodontics (milwaukee, wis.)"] = "Int J Orthod Milwaukee",
  ["international journal of orthopaedic and trauma nursing"] = "Int J Orthop Trauma Nurs",
  ["international journal of orthopaedics (hong kong)"] = "Int J Orthop (Hong Kong)",
  ["international journal of orthopedics"] = "Int J Orthop",
  ["international journal of osteoarchaeology"] = "Int J Osteoarchaeol",
  ["international journal of osteopathic medicine : ijom"] = "Int J Osteopath Med",
  ["international journal of otolaryngology"] = "Int J Otolaryngol",
  ["international journal of otorhinolaryngology"] = "Int J Otorhinolaryngol",
  ["international journal of paediatric dentistry"] = "Int J Paediatr Dent",
  ["international journal of paediatric dentistry / the british paedodontic society [and] the international association of dentistry for children"] = "Int J Paediatr Dent",
  ["international journal of paleopathology"] = "Int J Paleopathol",
  ["international journal of palliative nursing"] = "Int J Palliat Nurs",
  ["international journal of pancreatology"] = "Int. J. Pancreatol.",
  ["international journal of pancreatology : official journal of the international association of pancreatology"] = "Int J Pancreatol",
  ["international journal of parallel programming"] = "Int. J. Parallel Program.",
  ["international journal of parallel, emergent and distributed systems"] = "Int. J. Parallel Emergent Distrib. Syst.",
  ["international journal of parallel, emergent and distributed systems : ijpeds"] = "Int J Parallel Emergent Distrib Syst",
  ["international journal of parasitology: drugs and drug resistance"] = "Int. J. Parasitol.: Drugs Drug Resist.",
  ["international journal of parasitology: parasites and wildlife"] = "Int. J. Parasitol.: Parasites Wildl.",
  ["international journal of partial hospitalization"] = "Int J Partial Hosp",
  ["international journal of particle therapy"] = "Int J Part Ther",
  ["international journal of pattern recognition and artificial intelligence"] = "Int. J. Pattern Recognit. Artif. Intell.",
  ["international journal of pavement engineering"] = "Int. J. Pavement Eng.",
  ["international journal of pavement research and technology"] = "Int. J. Pavement Res. Technol.",
  ["international journal of pediatric endocrinology"] = "Int J Pediatr Endocrinol",
  ["international journal of pediatric nephrology"] = "Int. J. Pediatr. Nephrol.",
  ["international journal of pediatric obesity : ijpo : an official journal of the international association for the study of obesity"] = "Int J Pediatr Obes",
  ["international journal of pediatric otorhinolaryngology"] = "Int J Pediatr Otorhinolaryngol",
  ["international journal of pediatric otorhinolaryngology extra"] = "Int J Pediatr Otorhinolaryngol Extra",
  ["international journal of pediatrics"] = "Int J Pediatr",
  ["international journal of pediatrics & adolescent medicine"] = "Int J Pediatr Adolesc Med",
  ["international journal of pediatrics and child health"] = "Int J Pediatr Child Health",
  ["international journal of peptide and protein research"] = "Int. J. Pept. Protein Res.",
  ["international journal of peptide research and therapeutics"] = "Int. J. Pept. Res. Ther.",
  ["international journal of peptides"] = "Int J Pept",
  ["international journal of performance analysis in sport"] = "Int J Perform Anal Sport",
  ["international journal of performance arts and digital media"] = "Int J Perform Arts Digit Media",
  ["international journal of periodontics and restorative dentistry"] = "Int. J. Periodontics Restorative Dent.",
  ["international journal of person centered medicine"] = "Int J Pers Cent Med",
  ["international journal of personality psychology"] = "Int J Personal Psychol",
  ["international journal of pervasive computing and communications"] = "Int. J. Pervasive Comput. Commun.",
  ["international journal of pest management"] = "Int J Pest Manag",
  ["international journal of petroleum science and technology"] = "Int J Pet Sci Technol",
  ["international journal of pharma and bio sciences"] = "Int J Pharma Bio Sci",
  ["international journal of pharma informa : journal of biotechnology and biotherapeutics"] = "Int J Pharma Inf J Biotechnol Biother",
  ["international journal of pharma research & review"] = "Int J Pharma Res Rev",
  ["international journal of pharma sciences"] = "Int J Pharma Sci",
  ["international journal of pharmaceutical and biological archive"] = "Int J Pharm Biol Arch",
  ["international journal of pharmaceutical and healthcare marketing"] = "Int J Pharm Healthc Mark",
  ["international journal of pharmaceutical compounding"] = "Int J Pharm Compd",
  ["international journal of pharmaceutical investigation"] = "Int J Pharm Investig",
  ["international journal of pharmaceutical science invention"] = "Int J Pharm Sci Invent",
  ["international journal of pharmaceutical sciences and nanotechnology"] = "Int. J. Pharm. Sci. Nanotechnol.",
  ["international journal of pharmaceutical sciences and research"] = "Int J Pharm Sci Res",
  ["international journal of pharmaceutical sciences review and research"] = "Int. J Pharm Sci Rev Res",
  ["international journal of pharmaceutical technology and product manufacture"] = "Int. J. Pharm. Technol. Prod. Manuf",
  ["international journal of pharmaceutics"] = "Int J Pharm",
  ["international journal of pharmaceutics (amsterdam, netherlands)"] = "Int. J. Pharm. (Amsterdam, Neth.)",
  ["international journal of pharmaceutics: x"] = "Int J Pharm X",
  ["international journal of pharmacokinetics"] = "Int J Pharmacokinet",
  ["international journal of pharmacology : ijp"] = "Int J Pharmacol",
  ["international journal of pharmacy and biological sciences"] = "Int J Pharm Biol Sci",
  ["international journal of pharmacy and pharmaceutical research"] = "Int J Pharm Pharm Res",
  ["international journal of pharmacy and pharmaceutical sciences"] = "Int. J. Pharm. Pharm. Sci.",
  ["international journal of pharmacy practice"] = "Int. J. Pharm. Pract.",
  ["international journal of pharmtech research"] = "Int J Pharmtech Res",
  ["international journal of philippine science and technology"] = "Int J Philipp Sci Technol",
  ["international journal of philosophical studies"] = "Int J Philos Stud",
  ["international journal of photoenergy"] = "Int. J. Photoenergy",
  ["international journal of phramcology and toxicology"] = "Int J Pharmacol Toxicol",
  ["international journal of phycology and phycochemistry"] = "Int. J. Phycol. Phycochem.",
  ["international journal of physical medicine & rehabilitation"] = "Int J Phys Med Rehabil",
  ["international journal of physiology"] = "Int J Physiol",
  ["international journal of physiology and pathophysiology"] = "Int. J. Physiol. Pathophysiol.",
  ["international journal of physiology, pathophysiology and pharmacology"] = "Int J Physiol Pathophysiol Pharmacol",
  ["international journal of phytoremediation"] = "Int J Phytoremediation",
  ["international journal of pixe"] = "Int. J. PIXE",
  ["international journal of plant & soil science"] = "Int J Plant Soil Sci",
  ["international journal of plant biology & research"] = "Int J Plant Biol Res",
  ["international journal of plant genomics"] = "Int J Plant Genomics",
  ["international journal of plant pathology"] = "Int J Plant Pathol",
  ["international journal of plant production"] = "Int J Plant Prod",
  ["international journal of plant protection"] = "Int J Plant Prot",
  ["international journal of plant sciences"] = "Int J Plant Sci",
  ["international journal of plasma science and engineering"] = "Int. J. Plasma Sci. Eng.",
  ["international journal of plasticity"] = "Int. J. Plast.",
  ["international journal of plastics technology"] = "Int. J. Plast. Technol.",
  ["international journal of play"] = "Int J Play",
  ["international journal of politics"] = "Int J Polit",
  ["international journal of politics, culture, and society"] = "Int J Polit Cult Soc",
  ["international journal of pollution research"] = "Int J Pollut Res",
  ["international journal of polymer analysis and characterization"] = "Int. J. Polym. Anal. Charact.",
  ["international journal of polymer science"] = "Int J Polym Sci",
  ["international journal of polymeric materials"] = "Int. J. Polym. Mater.",
  ["international journal of polymeric materials and polymeric biomaterials"] = "Int. J. Polym. Mater. Polym. Biomater.",
  ["international journal of polymeric science"] = "Int. J. Polym. Sci.",
  ["international journal of population data science"] = "Int J Popul Data Sci",
  ["international journal of population geography : ijpg"] = "Int J Popul Geogr",
  ["international journal of population research"] = "Int J Popul Res",
  ["international journal of population studies"] = "Int J Popul Stud",
  ["international journal of poultry science"] = "Int J Poult Sci",
  ["international journal of powder metallurgy"] = "Int. J. Powder Metall",
  ["international journal of powder metallurgy (princeton, new jersey)"] = "Int. J. Powder Metall. (Princeton, N. J.)",
  ["international journal of power and energy systems"] = "Int. J. Power Energy Syst",
  ["international journal of precision engineering"] = "Int. J. Precis. Eng.",
  ["international journal of precision engineering and manufacturing"] = "Int. J. Precis. Eng. Manuf.",
  ["international journal of precision engineering and manufacturing - green technology"] = "Int. J. Precis. Eng. Manuf. Green Technol.",
  ["international journal of pressure vessels and piping"] = "Int. J. Press. Vessels Pip.",
  ["international journal of preventive medicine"] = "Int J Prev Med",
  ["international journal of primatology"] = "Int J Primatol",
  ["international journal of prisoner health"] = "Int J Prison Health",
  ["international journal of privacy and health information management"] = "Int J Priv Health Inf Manag",
  ["international journal of probiotics & prebiotics"] = "Int J Probiotics Prebiotics",
  ["international journal of product lifecycle management"] = "Int J Prod Lifecycle Manag",
  ["international journal of production economics"] = "Int. J. Prod. Econ.",
  ["international journal of production research"] = "Int. J. Prod. Res.",
  ["international journal of prognostics and health management"] = "Int J Progn Health Manag",
  ["international journal of project management"] = "Int. J. Project Manage.",
  ["international journal of prosthodontics"] = "Int. J. Prosthodont.",
  ["international journal of protective structures"] = "Int. J. Prot. Struct.",
  ["international journal of protein research"] = "Int J Protein Res",
  ["international journal of proteomics"] = "Int J Proteomics",
  ["international journal of psychiatry"] = "Int J Psychiatry",
  ["international journal of psychiatry (overland park)"] = "Int J Psychiatry (Overl Park)",
  ["international journal of psychiatry in clinical practice"] = "Int J Psychiatry Clin Pract",
  ["international journal of psychiatry in medicine"] = "Int J Psychiatry Med",
  ["international journal of psycho-analysis"] = "Int. J. Psychoanal.",
  ["international journal of psychoanalytic psychotherapy"] = "Int J Psychoanal Psychother",
  ["international journal of psychoanalytic self psychology"] = "Int J Psychoanal Self Psychol",
  ["international journal of psychological research"] = "Int J Psychol Res (Medellin)",
  ["international journal of psychological studies"] = "Int J Psychol Stud",
  ["international journal of psychology : journal international de psychologie"] = "Int J Psychol",
  ["international journal of psychology and behavioral sciences"] = "Int J Psychol Behav Sci",
  ["international journal of psychology and neuroscience"] = "Int J Psychol Neurosci",
  ["international journal of psychology and psychoanalysis"] = "Int J Psychol Psychoanal",
  ["international journal of psychology. journal international de psychologie"] = "Int J Psychol",
  ["international journal of psychophysiology"] = "Int. J. Psychophysiol.",
  ["international journal of psychophysiology : official journal of the international organization of psychophysiology"] = "Int J Psychophysiol",
  ["international journal of psychosomatics"] = "Int. J. Psychosom.",
  ["international journal of psychosomatics : official publication of the international psychosomatics institute"] = "Int J Psychosom",
  ["international journal of public health"] = "Int J Public Health",
  ["international journal of public health research"] = "Int J Public Health Res",
  ["international journal of public health science"] = "Int J Publ Health Sci",
  ["international journal of public mental health and neurosciences"] = "Int J Public Ment Health Neurosci",
  ["international journal of public opinion research"] = "Int J Public Opin Res",
  ["international journal of public policy"] = "Int J Public Pol",
  ["international journal of punjab studies"] = "Int J Punjab Stud",
  ["international journal of pure and applied bioscience"] = "Int J Pure Appl Biosci",
  ["international journal of pure and applied mathematics : ijpam"] = "Int J Pure Appl Math",
  ["international journal of pure and applied sciences and technology"] = "Int J Pure Appl Sci Technol",
  ["international journal of qualitative methods"] = "Int J Qual Methods",
  ["international journal of qualitative studies in education : qse"] = "Int J Qual Stud Educ",
  ["international journal of qualitative studies on health and well-being"] = "Int J Qual Stud Health Well-being",
  ["international journal of quality & reliability management"] = "Int. J. Qual. Reliab. Manage.",
  ["international journal of quality science"] = "Int. J. Qual. Sci.",
  ["international journal of quantitative research in education"] = "Int J Quant Res Educ",
  ["international journal of quantum chemistry"] = "Int. J. Quantum Chem.",
  ["international journal of quantum chemistry, part 1"] = "Int. J. Quantum Chem. 1",
  ["international journal of quantum chemistry, part 2"] = "Int. J. Quantum Chem. 2",
  ["international journal of quantum chemistry. quantum biology symposium : proceedings of the international symposium on quantum biology and quantum pharmacology. international symposium on quantum biology and quantum pharmacology"] = "Int J Quantum Chem Quantum Biol Symp",
  ["international journal of quantum information"] = "Int. J. Quantum Inf.",
  ["international journal of radiation applications and instrumentation. part a, applied radiation and isotopes"] = "Int J Rad Appl Instrum A",
  ["international journal of radiation applications and instrumentation. part a. applied radiation and isotopes"] = "Int. J. Radiat. Appl. Instrum. Part A",
  ["international journal of radiation applications and instrumentation. part b, nuclear medicine and biology"] = "Int J Rad Appl Instrum B",
  ["international journal of radiation applications and instrumentation. part b. nuclear medicine and biology"] = "Int. J. Radiat. Appl. Instrum. Part B",
  ["international journal of radiation applications and instrumentation. part c. radiation physics and chemistry"] = "Int. J. Radiat. Appl. Instrum. Part C",
  ["international journal of radiation applications and instrumentation. part d, nuclear tracks and radiation measurements"] = "Int J Rad Appl Instrum D",
  ["international journal of radiation applications and instrumentation. part d. nuclear tracks and radiation measurements"] = "Int. J. Radiat. Appl. Instrum. Part D",
  ["international journal of radiation applications and instrumentation. part e. nuclear geophysics"] = "Int. J. Radiat. Appl. Instrum. Part E",
  ["international journal of radiation biology"] = "Int J Radiat Biol",
  ["international journal of radiation biology and related studies in physics, chemistry and medicine"] = "Int. J. Radiat. Biol. Relat. Stud. Phys. Chem. Med.",
  ["international journal of radiation biology and related studies in physics, chemistry, and medicine"] = "Int J Radiat Biol Relat Stud Phys Chem Med",
  ["international journal of radiation medicine"] = "Int. J. Radiat. Med.",
  ["international journal of radiation oncology biology physics"] = "Int. J. Radiat. Oncol. Biol. Phys.",
  ["international journal of radiation oncology, biology, physics"] = "Int J Radiat Oncol Biol Phys",
  ["international journal of radiation physics"] = "Int. J. Radiat. Phys.",
  ["international journal of radio frequency identification technology and applications"] = "Int. J. Radio Freq. Identif. Technol. Appl.",
  ["international journal of rail transportation"] = "Int. J. Rail Transp.",
  ["international journal of recent scientific research"] = "Int J Recent Sci Res",
  ["international journal of reconfigurable computing"] = "Int. J. Reconfigurable Comput.",
  ["international journal of recycling of organic waste in agriculture"] = "Int. J. Recyl. Org. Waste Agric.",
  ["international journal of refractory & hard metals"] = "Int J Refract Hard Met",
  ["international journal of refractory metals & hard materials"] = "Int J Refract Metals Hard Mater",
  ["international journal of refractory metals and hard materials"] = "Int. J. Refract. Met. Hard Mater",
  ["international journal of refrigeration"] = "Int. J. Refrig",
  ["international journal of rehabilitation research"] = "Int. J. Rehabil. Res.",
  ["international journal of rehabilitation research. internationale zeitschrift fur rehabilitationsforschung. revue internationale de recherches de readaptation"] = "Int J Rehabil Res",
  ["international journal of rehabilitation research. internationale zeitschrift für rehabilitationsforschung. revue internationale de recherches de réadaptation"] = "Int J Rehabil Res",
  ["international journal of reliability and safety"] = "Int. J. Reliab. Saf.",
  ["international journal of reliability, quality and safety engineering"] = "Int. J. Reliab. Qual. Saf. Eng.",
  ["international journal of remote sensing"] = "Int. J. Remote Sens.",
  ["international journal of renewable energy development"] = "Int. J. Renewable Energy Dev.",
  ["international journal of renewable energy research"] = "Int. J. Renewable Energy Res.",
  ["international journal of reproduction, contraception, obstetrics and gynecology"] = "Int J Reprod Contracept Obstet Gynecol",
  ["international journal of reproductive biomedicine"] = "Int J Reprod Biomed",
  ["international journal of reproductive medicine"] = "Int J Reprod Med",
  ["international journal of research in ayurveda and pharmacy"] = "Int J Res Ayurveda Pharm",
  ["international journal of research in engineering and technology"] = "Int J Res Eng Technol",
  ["international journal of research in medical sciences"] = "Int J Res Med Sci",
  ["international journal of research in nursing"] = "Int J Res Nurs",
  ["international journal of research in orthopaedics"] = "Int J Res Orthop",
  ["international journal of research in undergraduate mathematics education"] = "Int. J. Res. Undergrad. Math. Educ.",
  ["international journal of research in zoology"] = "Int J Res Zool",
  ["international journal of research studies in biosciences"] = "Int J Res Stud Biosci",
  ["international journal of researches in biosciences, agriculture & technology"] = "Int J Res Biosci Agric Technol",
  ["international journal of respiratory and pulmonary medicine"] = "Int J Respir Pulm Med",
  ["international journal of retina and vitreous"] = "Int J Retina Vitreous",
  ["international journal of rf and microwave computer-aided engineering"] = "Int. J. RF Microwave Comput. Aided Eng.",
  ["international journal of rf technologies: research and aoplications"] = "Int. J. RF Technol. Res. Appl.",
  ["international journal of rf technologies: research and applications"] = "Int. J. RF Technol. Res. Appl.",
  ["international journal of rheumatic diseases"] = "Int J Rheum Dis",
  ["international journal of rheumatology"] = "Int J Rheumatol",
  ["international journal of risk & safety in medicine"] = "Int. J. Risk Saf. Med.",
  ["international journal of risk and contingency management"] = "Int J Risk Conting Manag",
  ["international journal of risk assessment and management"] = "Int J Risk Assess Manag",
  ["international journal of river basin management"] = "Int. J. River Basin Manage.",
  ["international journal of robotics and automation"] = "Int. J. Rob. Autom.",
  ["international journal of robotics research"] = "Int. J. Rob. Res.",
  ["international journal of robust and nonlinear control"] = "Int. J. Robust Nonlinear Control",
  ["international journal of rock mechanics and mining sciences"] = "Int. J. Rock Mech. Min. Sci.",
  ["international journal of rock mechanics and mining sciences (oxford, england : 1997)"] = "Int J Rock Mech Min Sci (1997)",
  ["international journal of rotating machinery"] = "Int. J. Rotating Mach.",
  ["international journal of salt lake research"] = "Int J Salt Lake Res",
  ["international journal of satellite communications and networking"] = "Int. J. Satell. Commun. Networking",
  ["international journal of school & educational psychology"] = "Int J Sch Educ Psychol",
  ["international journal of science and applied technology"] = "Int J Sci Appl Technol",
  ["international journal of science and mathematics education"] = "Int J Sci Math Educ",
  ["international journal of science and nature"] = "Int J Sci Nat",
  ["international journal of science and research (raipur, india)"] = "Int J Sci Res (Raipur)",
  ["international journal of science and research methodology"] = "Int J Sci Res Methodol",
  ["international journal of science education"] = "Int J Sci Educ",
  ["international journal of science education, part b"] = "Int. J. Sci. Educ., Part B",
  ["international journal of science education. part b. communication and public engagement"] = "Int J Sci Educ B Commun Public Engagem",
  ["international journal of science, commerce and humanities"] = "Int J Sci Commer Humanit",
  ["international journal of science, environment and technology"] = "Int J Sci Environ Technol",
  ["international journal of sciences"] = "Int J Sci",
  ["international journal of sciences and applied research"] = "Int J Sci Appl Res",
  ["international journal of sciences, basic and applied research"] = "Int J Sci Basic Appl Res",
  ["international journal of scientific and engineering research"] = "Int J Sci Eng Res",
  ["international journal of scientific computing"] = "Int J Sci Comput",
  ["international journal of scientific research"] = "Int J Sci Res",
  ["international journal of scientific research (ahmedabad, india)"] = "Int J Sci Res (Ahmedabad)",
  ["international journal of scientific research in computer science applications and management studies"] = "Int J Sci Res Comput Sci Appl Manag Stud",
  ["international journal of scientific research in science and technology"] = "Int J Sci Res Sci Technol",
  ["international journal of scientific research in science, engineering and technology"] = "Int J Sci Res Sci Eng Technol",
  ["international journal of scientific study"] = "Int J Sci Study",
  ["international journal of secure software engineering"] = "Int J Secur Softw Eng",
  ["international journal of security and networks"] = "Int. J. Secur. Netw.",
  ["international journal of sediment research"] = "Int. J. Sediment Res.",
  ["international journal of self help & self care"] = "Int J Self Help Self Care",
  ["international journal of self-propagating high-temperature synthesis"] = "Int. J. Self-Propag. High-Temp. Synth.",
  ["international journal of semantic computing"] = "Int J Semant Comput",
  ["international journal of sensor networks"] = "Int. J. Sens. Netw.",
  ["international journal of sensor networks and data communications"] = "Int J Sens Netw Data Commun",
  ["international journal of services operations and informatics"] = "Int. J. Serv. Oper. Inf.",
  ["international journal of services, technology and management"] = "Int. J. Serv. Technol. Manage.",
  ["international journal of sex-economy and orgone-research"] = "Int J Sex Econ Orgone Res",
  ["international journal of sexology"] = "Int J Sexol",
  ["international journal of sexual health : official journal of the world association for sexual health"] = "Int J Sex Health",
  ["international journal of shape modeling"] = "Int. J. Shape Model.",
  ["international journal of shoulder surgery"] = "Int J Shoulder Surg",
  ["international journal of simulation and process modelling"] = "Int. J. Simul. Process Model.",
  ["international journal of simulation: systems, science and technology"] = "Int. J. Simul. Syst. Sci. Technol.",
  ["international journal of smart and nano materials"] = "Int J Smart Nano Mater",
  ["international journal of smart sensing and intelligent systems"] = "Int. J. Smart Sens. Intell. Syst.",
  ["international journal of smart technologies"] = "Int. J. Smart Technol.",
  ["international journal of social ecology and sustainable development"] = "Int. J. Social Ecol. Sustainable Dev.",
  ["international journal of social economics"] = "Int J Soc Econ",
  ["international journal of social psychiatry"] = "Int. J. Soc. Psychiatry",
  ["international journal of social research methodology"] = "Int J Soc Res Methodol",
  ["international journal of social robotics"] = "Int J Soc Robot",
  ["international journal of social science and humanity : ijssh"] = "Int J Soc Sci Humanit",
  ["international journal of social science research"] = "Int J Soc Sci Res",
  ["international journal of social science studies"] = "Int J Soc Sci Stud",
  ["international journal of social welfare"] = "Int J Soc Welf",
  ["international journal of society systems science"] = "Int J Soc Syst Sci",
  ["international journal of sociology"] = "Int J Sociol",
  ["international journal of sociology of the family"] = "Int J Sociol Fam",
  ["international journal of soft computing"] = "Int. J. Soft Comput.",
  ["international journal of software engineering and knowledge engineering"] = "Int. J. Software Eng. Knowledge Eng.",
  ["international journal of soil dynamics and earthquake engineering"] = "Int. J. Soil Dyn. Earthquake Eng.",
  ["international journal of soil science"] = "Int. J. Soil Sci.",
  ["international journal of solids and structures"] = "Int. J. Solids Struct.",
  ["international journal of south asian studies"] = "Int J South Asian Stud (New Delhi)",
  ["international journal of south asian studies (pondicherry, india)"] = "Int J South Asian Stud (Pondicherry)",
  ["international journal of space structures"] = "Int. J. Space Struct.",
  ["international journal of special education"] = "Int J Spec Educ",
  ["international journal of spectroscopy"] = "Int J Spectrosc",
  ["international journal of speech & language pathology and audiology"] = "Int J Speech Lang Pathol Audiol",
  ["international journal of speech technology"] = "Int. J. Speech Technol.",
  ["international journal of speech-language pathology"] = "Int J Speech Lang Pathol",
  ["international journal of speleology"] = "Int J Speleol",
  ["international journal of spine surgery"] = "Int J Spine Surg",
  ["international journal of sport and exercise psychology"] = "Int J Sport Exerc Psychol",
  ["international journal of sport nutrition"] = "Int J Sport Nutr",
  ["international journal of sport nutrition and exercise metabolism"] = "Int J Sport Nutr Exerc Metab",
  ["international journal of sport psychology"] = "Int J Sport Psychol",
  ["international journal of sports and exercise medicine"] = "Int J Sports Exerc Med",
  ["international journal of sports medicine"] = "Int J Sports Med",
  ["international journal of sports physical therapy"] = "Int J Sports Phys Ther",
  ["international journal of sports physiology and performance"] = "Int J Sports Physiol Perform",
  ["international journal of sports science & coaching"] = "Int J Sports Sci Coach",
  ["international journal of spray and combustion dynamics"] = "Int. J. Spray Combust. Dyn.",
  ["international journal of statistics and applications"] = "Int J Stat Appl",
  ["international journal of statistics and management system"] = "Int J Stat Manag Syst",
  ["international journal of statistics and probability"] = "Int J Stat Probab",
  ["international journal of statistics in medical research"] = "Int J Stat Med Res",
  ["international journal of std & aids"] = "Int J STD AIDS",
  ["international journal of std and aids"] = "Int. J. STD AIDS",
  ["international journal of steel structures"] = "Int. J. Steel Struct.",
  ["international journal of stem cell research and therapy"] = "Int J Stem Cell Res Ther",
  ["international journal of stem cells"] = "Int J Stem Cells",
  ["international journal of stem education"] = "Int J STEM Educ",
  ["international journal of stem education research"] = "Int. J. STEM Educ. Res.",
  ["international journal of stochastic analysis"] = "Int. J. Stoch. Anal.",
  ["international journal of stomatology & occlusion medicine"] = "Int J Stomatol Occlusion Med",
  ["international journal of streel structures"] = "Int. J. Steel Struct.",
  ["international journal of stress management"] = "Int J Stress Manag",
  ["international journal of stroke : official journal of the international stroke society"] = "Int J Stroke",
  ["international journal of structural engineering"] = "Int. J. Struct. Eng.",
  ["international journal of structural integrity"] = "Int. J. Struct. Integrity",
  ["international journal of structural stability and dynamics"] = "Int. J. Struct. Stab. Dyn.",
  ["international journal of student scholarship in physical therapy"] = "Int J Stud Scholarsh Phys Ther",
  ["international journal of sulfur chemistry part a: original experimental and theoretical studies"] = "Int. J. Sulfur Chem. Part A:",
  ["international journal of sulfur chemistry part b: quartler reports on sulfur chemistry"] = "Int. J. Sulfur Chem. Part B:",
  ["international journal of surface engineering and interdisciplinary materials science"] = "Int. J. Surf. Eng. Interdiscip. Mater. Sci.",
  ["international journal of surface mining, reclamation, and environment"] = "Int. J. Surf. Min. Reclam. Environ.",
  ["international journal of surface science and engineering"] = "Int. J. Surf. Sci. Eng.",
  ["international journal of surgery (london, england)"] = "Int J Surg",
  ["international journal of surgery and research"] = "Int J Surg Res",
  ["international journal of surgery case reports"] = "Int J Surg Case Rep",
  ["international journal of surgery protocols"] = "Int J Surg Protoc",
  ["international journal of surgery. oncology"] = "Int J Surg Oncol (N Y)",
  ["international journal of surgical investigation"] = "Int J Surg Investig",
  ["international journal of surgical oncology"] = "Int J Surg Oncol",
  ["international journal of surgical pathology"] = "Int J Surg Pathol",
  ["international journal of sustainability in economic social and cultural context"] = "Int. J. Sustainability Econ. Social Cult. Context",
  ["international journal of sustainability in higher education"] = "Int. J. Sustainability Higher Educ.",
  ["international journal of sustainable agricultural management and informatics"] = "Int. J. Sustainable Agric. Manage. Inf.",
  ["international journal of sustainable aviation"] = "Int. J. Sustainable Aviat.",
  ["international journal of sustainable build environment"] = "Int. J. Sustainable Built Environ.",
  ["international journal of sustainable building technology and urban development"] = "Int. J. Sustainable Build. Technol. Urban Dev.",
  ["international journal of sustainable built environment"] = "Int. J. Sustainable Built Environ.",
  ["international journal of sustainable construction engineering and technology"] = "Int. J. Sustainable Constr. Eng. Technol.",
  ["international journal of sustainable development"] = "Int. J. Sustainable Dev.",
  ["international journal of sustainable development and planning"] = "Int. J. Sustainable Dev. Plann.",
  ["international journal of sustainable development and world ecology"] = "Int. J. Sustainable Dev. World Ecol.",
  ["international journal of sustainable economy"] = "Int. J.  Sustainable Econ.",
  ["international journal of sustainable energy"] = "Int. J. Sustainable Energy",
  ["international journal of sustainable energy planning and management"] = "Int. J. Sustainable Energy Plann. Manage.",
  ["international journal of sustainable engineering"] = "Int. J. Sustainable Eng.",
  ["international journal of sustainable oceans and coasts"] = "Int. J. Sustainable Oceans Coasts",
  ["international journal of sustainable society"] = "Int. J. Sustainable Soc.",
  ["international journal of sustainable transportation"] = "Int J Sustain Transp",
  ["international journal of systematic and evolutionary microbiology"] = "Int J Syst Evol Microbiol",
  ["international journal of systematic bacteriology"] = "Int J Syst Bacteriol",
  ["international journal of systems and synthetic biology"] = "Int J Syst Synth Biol",
  ["international journal of systems assurance engineering and management"] = "Int. J. Syst. Assur. Eng. Manage.",
  ["international journal of systems science"] = "Int. J. Syst. Sci.",
  ["international journal of systems science. principles and applications of systems and integration"] = "Internat. J. Systems Sci.",
  ["international journal of systems science: operations and logistics"] = "Int. J. Syst. Sci.: Oper. Logist.",
  ["international journal of technoethics"] = "Int J Technoethics",
  ["international journal of technology & aging"] = "Int J Technol Aging",
  ["international journal of technology and design education"] = "Int. J. Technol. Des. Educ.",
  ["international journal of technology and human interaction"] = "Int. J. Technol. Hum. Interact.",
  ["international journal of technology assessment in health care"] = "Int J Technol Assess Health Care",
  ["international journal of technology intelligence and planning"] = "Int. J. Technol. Intell. Plann.",
  ["international journal of technology management"] = "Int. J. Technol. Manage.",
  ["international journal of technology management = journal international de la gestion technologique"] = "Int J Technol Manag",
  ["international journal of technology management and sustainable development"] = "Int. J. Technol. Manage. Sustainable Dev.",
  ["international journal of technology, knowledge and society"] = "Int J Technol Knowl Soc",
  ["international journal of technology, policy and management"] = "Int. J. Technol. Policy Manage.",
  ["international journal of telemedicine and applications"] = "Int J Telemed Appl",
  ["international journal of telerehabilitation"] = "Int J Telerehabil",
  ["international journal of testing"] = "Int J Test",
  ["international journal of the addictions"] = "Int. J. Addict.",
  ["international journal of the classical tradition"] = "IJCT",
  ["international journal of the commons"] = "Int J Commons",
  ["international journal of the economics of business"] = "Int J Econ Bus",
  ["international journal of the japan society for precision engineering"] = "Int. J. Jpn Soc. Precis. Eng.",
  ["international journal of the society of material engineering for resources"] = "Int. J. Soc. Mater. Eng. Resour.",
  ["international journal of the sociology of language"] = "Int J Soc Lang",
  ["international journal of the sociology of law"] = "Int. J. Sociol. Law",
  ["international journal of theoretical and applied finance"] = "Int. J. Theor. Appl. Finance",
  ["international journal of theoretical physics"] = "Int. J. Theor. Phys.",
  ["international journal of theoretical physics, group theory, and nonlinear optics"] = "Int. J. Theor. Phys. Group Theory Nonlinear Opt.",
  ["international journal of therapeutic massage & bodywork"] = "Int J Ther Massage Bodywork",
  ["international journal of therapy and rehabilitation"] = "Int J Ther Rehabil",
  ["international journal of thermal sciences"] = "Int. J. Therm. Sci.",
  ["international journal of thermal sciences = revue générale de thermique"] = "Int J Therm Sci",
  ["international journal of thermodynamics"] = "Int. J. Thermodyn.",
  ["international journal of thermofluids"] = "Int. J. Thermofluids",
  ["international journal of thermophysics"] = "Int. J. Thermophys.",
  ["international journal of tissue engineering"] = "Int J Tissue Eng",
  ["international journal of tissue reactions"] = "Int J Tissue React",
  ["international journal of toxicology"] = "Int J Toxicol",
  ["international journal of training and development"] = "Int J Train Dev",
  ["international journal of transgender health"] = "Int J Transgend Health",
  ["international journal of translational science"] = "Int J Transl Sci",
  ["international journal of transport development and integration"] = "Int. J. Transp. Dev. Integr.",
  ["international journal of transport economics"] = "Int. J. Transport Econ.",
  ["international journal of transport phenomena"] = "Int. J. Transp. Phenom.",
  ["international journal of transportation science and technology"] = "Int. J. Transp. Sci. Technol.",
  ["international journal of trauma nursing"] = "Int J Trauma Nurs",
  ["international journal of travel medicine and global health"] = "Int J Travel Med Glob Health",
  ["international journal of trichology"] = "Int J Trichology",
  ["international journal of tropical agriculture = antarrāshtriya ūshnakatibandhīya kr̥shi-śodha patrikā"] = "Int J Trop Agric",
  ["international journal of tropical and subtropical horticulture"] = "Int. J. Trop. Subtrop. Hortic.",
  ["international journal of tropical disease & health"] = "Int J Trop Dis Health",
  ["international journal of tropical insect science"] = "Int. J. Trop. Insect Sci.",
  ["international journal of tryptophan research"] = "Int. J. Tryptophan Res.",
  ["international journal of tryptophan research : ijtr"] = "Int J Tryptophan Res",
  ["international journal of tuberculosis and lung disease"] = "Int. J. Tuberc. Lung Dis.",
  ["international journal of turbo and jet engines"] = "Int. J. Turbo Jet Engines",
  ["international journal of uncertainty, fuzziness and knowledge-based systems"] = "Int. J. Uncertainty Fuzziness Knowledge Based Syst.",
  ["international journal of uncertainty, fuzziness and knowlege-based systems"] = "Int. J. Uncertainty Fuzziness Knowledge Based Syst.",
  ["international journal of universal pharmacy and life sciences"] = "Int J Univers Pharm Life Sci",
  ["international journal of university teaching and faculty development"] = "Int J Univ Teach Fac Dev",
  ["international journal of unmanned systems"] = "Int. J. Unmanned Syst.",
  ["international journal of urban and regional research"] = "Int J Urban Reg Res",
  ["international journal of urban sciences"] = "Int. J. Urban Sci.",
  ["international journal of urban sustainable development"] = "Int. J. Urban Sustainable Dev.",
  ["international journal of urological nursing : the journal of the baun"] = "Int J Urol Nurs",
  ["international journal of urology"] = "Int. J. Urol.",
  ["international journal of urology : official journal of the japanese urological association"] = "Int J Urol",
  ["international journal of vascular medicine"] = "Int J Vasc Med",
  ["international journal of vegetable science"] = "Int. J. Veg. Sci.",
  ["international journal of vehicle autonomous systems"] = "Int. J. Veh. Auton. Syst.",
  ["international journal of vehicle design"] = "Int. J. Veh. Des.",
  ["international journal of vehicle information and communication systems"] = "Int. J. Veh. Inf. Commun. Syst.",
  ["international journal of vehicle noise and vibration"] = "Int. J. Veh. Noise Vib.",
  ["international journal of vehicle safety"] = "Int. J. Veh. Saf.",
  ["international journal of vehicle systems modelling and testing"] = "Int. J. Veh. Syst. Model. Test.",
  ["international journal of ventilation"] = "Int. J. Vent.",
  ["international journal of veterinary science"] = "Int J Vet Sci",
  ["international journal of veterinary science and medicine"] = "Int J Vet Sci Med",
  ["international journal of virology"] = "Int J Virol",
  ["international journal of virology and molecular biology"] = "Int J Virol Mol Biol",
  ["international journal of virology studies & research"] = "Int J Virol Stud Res",
  ["international journal of virtual communities and social networking"] = "Int J Virtual Communities Soc Netw",
  ["international journal of waste resources"] = "Int J Waste Resour",
  ["international journal of water"] = "Int. J. Water",
  ["international journal of water resources development"] = "Int. J. Water Resour. Dev.",
  ["international journal of wavelets, multiresolution and information processing"] = "Int. J. Wavelets Multiresolution Inf. Process.",
  ["international journal of web and grid services"] = "Int. J. Web Grid Serv.",
  ["international journal of web based communities"] = "Int. J. Web Based Communities",
  ["international journal of web engineering and technology"] = "Int. J. Web Eng. Technol.",
  ["international journal of web services research"] = "Int. J. Web Serv. Res.",
  ["international journal of wild silkmoth & silk"] = "Int J Wild Silkmoth Silk",
  ["international journal of wildland fire"] = "Int J Wildland Fire",
  ["international journal of wine research"] = "Int. J. Wine Res.",
  ["international journal of wireless and mobile computing"] = "Int. J. Wireless Mobile Comput.",
  ["international journal of wireless information networks"] = "Int. J. Wireless Inf. Networks",
  ["international journal of women's dermatology"] = "Int J Womens Dermatol",
  ["international journal of women's health"] = "Int J Womens Health",
  ["international journal of women's health and wellness"] = "Int J Womens Health Wellness",
  ["international journal of women's studies"] = "Int J Womens Stud",
  ["international journal of women’s studies"] = "Int J Womens Stud",
  ["international journal of workplace health management"] = "Int J Workplace Health Manag",
  ["international journal of wrestling science"] = "Int. J. Wrestling Sci.",
  ["international journal of yoga"] = "Int J Yoga",
  ["international journal of yoga therapy"] = "Int J Yoga Therap",
  ["international journal of zoology"] = "Int J Zool",
  ["international journal of zoology research"] = "Int J Zool Res",
  ["international journal of zoonoses"] = "Int J Zoonoses",
  ["international journal on advanced science, engineering and information technology"] = "Int J Adv Sci Eng Inf Technol",
  ["international journal on advances in life sciences"] = "Int J Adv Life Sci",
  ["international journal on alcohol and alcoholism"] = "Int J Alcohol Alcohol",
  ["international journal on algae"] = "Int. J. Algae",
  ["international journal on artificial intelligence tools"] = "Int. J. Artif. Intell. Tools",
  ["international journal on child maltreatment : research, policy and practice"] = "Int J Child Maltreat",
  ["international journal on communications antenna and propagation"] = "Int J Commun Antenna Propag",
  ["international journal on disability and human development : ijdhd"] = "Int J Disabil Hum Dev",
  ["international journal on document analysis and recognition"] = "Int. J. Doc. Anal. Recogn.",
  ["international journal on document analysis and recognition (online)"] = "Int J Doc Anal Recognit",
  ["international journal on drug policy"] = "Int. J. Drug Policy",
  ["international journal on energy conversion"] = "Int. J. Energy Convers.",
  ["international journal on finite volumes"] = "Int. J. Finite Vol.",
  ["international journal on hydropower and dams"] = "Int. J. Hydropower Dams",
  ["international journal on interactive design and manufacturing"] = "Int. J. Interact. Des. Manuf.",
  ["international journal on magnetic particle imaging"] = "Int J Magn Part Imaging",
  ["international journal on semantic web and information systems"] = "Int J Semant Web Inf Syst",
  ["international journal on smart and sustainable cities"] = "Int. J. Smart Sustainable Cities",
  ["international journal on software tools for technology transfer"] = "Int. J. Softw. Tools Technol. Trans.",
  ["international journal on software tools for technology transfer : sttt"] = "Int J Softw Tools Technol Transf",
  ["international journal on uncertainty quantification"] = "Int. J. Uncertainty Quantif.",
  ["international journal on wireless & optical communications"] = "Int. J. Wireless Opt. Commun.",
  ["international journal precision engineering"] = "Int. J. Precis. Eng.",
  ["international labor and working class history"] = "Int Labor Work Class Hist",
  ["international labour review"] = "Int Labour Rev",
  ["international labour review / international labour office"] = "Int Labour Rev",
  ["international law and economics"] = "Int. Law Econ.",
  ["international library of diabetes & metabolism"] = "Int Libr Diabetes Metab",
  ["international library review"] = "Int Libr Rev",
  ["international lichenological newsletter"] = "Int. Lichenol. Newsl.",
  ["international marine energy journal"] = "Int. Mar. Energy J.",
  ["international maritime health"] = "Int Marit Health",
  ["international materials reviews"] = "Int. Mater. Rev.",
  ["international mathematical news"] = "Int. Math. News",
  ["international mathematics research notices"] = "Int. Math. Res. Not.",
  ["international mathematics research notices. imrn"] = "Int. Math. Res. Not. IMRN",
  ["international mathematics research papers"] = "Int. Math. Res. Pap.",
  ["international mathematics research surveys"] = "Int. Math. Res. Surv.",
  ["international mechanical engineering congress and exposition : [proceedings]. international mechanical engineering congress and exposition"] = "Int Mech Eng Congress Expo",
  ["international medical abstracts & reviews"] = "Int Med Abstr Rev",
  ["international medical case reports journal"] = "Int Med Case Rep J",
  ["international microbiology"] = "Int. Microbiol.",
  ["international microbiology : the official journal of the spanish society for microbiology"] = "Int Microbiol",
  ["international migration (geneva, switzerland)"] = "Int Migr",
  ["international monetary fund staff papers"] = "Int. Monet. Fund Staff Pap.",
  ["international ms journal"] = "Int MS J",
  ["international ms journal / ms forum"] = "Int MS J",
  ["international multidisciplinary research journal"] = "Int Multidiscip Res J",
  ["international multilingual research journal"] = "Int multiling res j",
  ["international musculoskeletal medicine"] = "Int Musculoskelet Med",
  ["international nano letters"] = "Int Nano Lett",
  ["international neuropsychiatric disease journal"] = "Int Neuropsychiatr Dis J",
  ["international neurotrauma letter"] = "Int Neurotrauma Lett",
  ["international neurourology journal"] = "Int Neurourol J",
  ["international nonwovens journal"] = "Int. Nonwovens J.",
  ["international nursing review"] = "Int Nurs Rev",
  ["international online journal of education & teaching"] = "Int Online J Educ Teach",
  ["international ophthalmology"] = "Int Ophthalmol",
  ["international ophthalmology clinics"] = "Int Ophthalmol Clin",
  ["international organization"] = "Int Organ",
  ["international orthodontics"] = "Int Orthod",
  ["international orthopaedics"] = "Int Orthop",
  ["international peacekeeping (london, england)"] = "Int Peacekeep",
  ["international peat journal"] = "Int. Peat J.",
  ["international pediatrics : the journal of the miami children's hospital"] = "Int Pediatr",
  ["international perspectives in psychology : research, practice, consultation"] = "Int Perspect Psychol",
  ["international perspectives in victimology"] = "Int Perspect Vict",
  ["international perspectives on sexual and reproductive health"] = "Int Perspect Sex Reprod Health",
  ["international pharmacopsychiatry"] = "Int Pharmacopsychiatry",
  ["international philosophical quarterly"] = "IPQ",
  ["international philosophical quarterly : ipq"] = "Int Philos Q",
  ["international physiology journal"] = "Int Physiol J",
  ["international planned parenthood news"] = "Int Plann Parenthood News",
  ["international planning studies"] = "Int Plan Stud",
  ["international policy review"] = "Int. Pol. Rev.",
  ["international political sociology : ips"] = "Int Political Sociol",
  ["international polymer processing"] = "Int. Polym. Proc.",
  ["international polymer processing : the journal of the polymer processing society"] = "Int Polym Process",
  ["international proceedings of chemical, biological & environmental engineering"] = "Int Proc Chem Biol Environ Eng",
  ["international project newsletter"] = "Int Proj Newsl",
  ["international psychiatry : bulletin of the board of international affairs of the royal college of psychiatrists"] = "Int Psychiatry",
  ["international psychiatry clinics"] = "Int Psychiatry Clin",
  ["international psychogeriatrics"] = "Int Psychogeriatr",
  ["international psychogeriatrics / ipa"] = "Int Psychogeriatr",
  ["international public health journal"] = "Int Public Health J",
  ["international public management journal : ipmj"] = "Int Public Manag J",
  ["international quarterly of community health education"] = "Int Q Community Health Educ",
  ["international record of medicine"] = "Int Rec Med",
  ["international record of medicine and general practice clinics"] = "Int Rec Med Gen Pract Clin",
  ["international regional science review"] = "Int Reg Sci Rev",
  ["international rehabilitation medicine"] = "Int Rehabil Med",
  ["international relations (david davies memorial institute of international studies)"] = "Int Relat (David Davies Mem Inst Int Stud)",
  ["international rescuer"] = "Int Rescuer",
  ["international research in geographical and environmental education"] = "Int. Res. Geogr. Environ. Educ.",
  ["international research journal of engineering and technology"] = "Int. Res. J. Eng. Technol.",
  ["international research journal of environmental sciences"] = "Int Res J Environ Sci",
  ["international research journal of natural and applied sciences"] = "Int Res J Nat Appl Sci",
  ["international research journal of pure and applied chemistry"] = "Int Res J Pure Appl Chem",
  ["international research on medical sciences"] = "Int Res Med Sci",
  ["international review (steubenville, ohio)"] = "Int Rev",
  ["international review for spatial planning and sustainable development"] = "Int. Rev. Spatial Plann. Sustainable Dev.",
  ["international review for the sociology of sport"] = "Int Rev Sociol Sport",
  ["international review of applied economics"] = "Int Rev Appl Econ",
  ["international review of applied sciences and engineering"] = "Int. Rev. Appl. Sci. Eng.",
  ["international review of cell and molecular biology"] = "Int Rev Cell Mol Biol",
  ["international review of connective tissue research"] = "Int Rev Connect Tissue Res",
  ["international review of cytology"] = "Int Rev Cytol",
  ["international review of cytology. supplement"] = "Int Rev Cytol Suppl",
  ["international review of economics"] = "Int Rev Econ",
  ["international review of economics and finance"] = "Int. Rev. Econ. Finance",
  ["international review of education. internationale zeitschrift fur erziehungswissenschaft. revue internationale de pedagogie"] = "Int Rev Educ",
  ["international review of education. internationale zeitschrift für erziehungswissenschaft. revue internationale de pédagogie"] = "Int Rev Educ",
  ["international review of electrical engineering"] = "Int. Rev. Electr. Eng.",
  ["international review of experimental pathology"] = "Int Rev Exp Pathol",
  ["international review of financial analysis"] = "Int. Rev. Finan. Anal.",
  ["international review of hydrobiology"] = "Int. Rev. Hydrobiol.",
  ["international review of law and economics"] = "Int Rev Law Econ",
  ["international review of missions"] = "Int Rev Missions",
  ["international review of modern sociology"] = "Int Rev Mod Sociol",
  ["international review of natural family planning"] = "Int Rev Nat Fam Plann",
  ["international review of neurobiology"] = "Int Rev Neurobiol",
  ["international review of neurobiology. supplement"] = "Int Rev Neurobiol Suppl",
  ["international review of physiology"] = "Int Rev Physiol",
  ["international review of poultry science"] = "Int. Rev. Poult. Sci.",
  ["international review of psychiatry (abingdon, england)"] = "Int Rev Psychiatry",
  ["international review of qualitative research : irqr"] = "Int Rev Qual Res",
  ["international review of research in developmental disabilities"] = "Int Rev Res Dev Disabil",
  ["international review of research in mental retardation"] = "Int Rev Res Ment Retard",
  ["international review of social history"] = "Int Rev Soc Hist",
  ["international review of social history / international instituut voor sociale geschiedenis, amsterdam"] = "Int Rev Soc Hist",
  ["international review of social research"] = "Int Rev Soc Res",
  ["international review of sport and exercise psychology"] = "Int Rev Sport Exerc Psychol",
  ["international review of thrombosis"] = "Int Rev Thromb",
  ["international review of tropical medicine"] = "Int Rev Trop Med",
  ["international review of victimology"] = "Int Rev Vict",
  ["international reviews in physical chemistry"] = "Int. Rev. Phys. Chem.",
  ["international reviews of immunology"] = "Int Rev Immunol",
  ["international rice research notes"] = "Int Rice Res Notes",
  ["international sampe symposium and exhibition"] = "Int. SAMPE Symp. Exhib.",
  ["international sampe symposium and exhibition : [proceedings]. international sampe symposium and exhibition"] = "Int SAMPE Symp Exhib",
  ["international scholarly research notices"] = "Int Sch Res Notices",
  ["international schools for computer scientists"] = "Int. Schools Comput. Sci.",
  ["international science review series"] = "Int Sci Rev Ser",
  ["international security"] = "Int Secur",
  ["international seminars in surgical oncology : isso"] = "Int Semin Surg Oncol",
  ["international series in intelligent technologies"] = "Int. Ser. Intell. Tech.",
  ["international series in operations research & management science"] = "Internat. Ser. Oper. Res. Management Sci.",
  ["international series in pure and applied physics"] = "Internat. Ser. Pure Appl. Phys.",
  ["international series in quantitative marketing"] = "Int. Ser. Quant. Mark.",
  ["international series of monographs on chemistry"] = "Int. Ser. Monogr. Chem.",
  ["international series of monographs on oral biology"] = "Int Ser Monogr Oral Biol",
  ["international series of monographs on physics"] = "Internat. Ser. Monogr. Phys.",
  ["international series of monographs on pure and applied mathematics"] = "Int. Ser. Monogr. Pure Appl. Math.",
  ["international series of numerical mathematics"] = "Internat. Ser. Numer. Math.",
  ["international series on actuarial science"] = "Int. Ser. Actuar. Sci.",
  ["international series on applied systems analysis"] = "Internat. Ser. Appl. Systems Anal.",
  ["international series on computational engineering"] = "Int. Ser. Comput. Eng.",
  ["international series on systems and control"] = "Internat. Ser. Systems Control",
  ["international shipbuilding progress"] = "Int. Shipbuild. Prog.",
  ["international social development review"] = "Int Soc Dev Rev",
  ["international social science journal"] = "Int Soc Sci J",
  ["international social science review"] = "Int Soc Sci Rev",
  ["international social security review"] = "Int Soc Secur Rev",
  ["international social work"] = "Int Soc Work",
  ["international socialist review"] = "Int Social Rev",
  ["international society for analysis, applications and computation"] = "Int. Soc. Anal. Appl. Comput.",
  ["international sociology"] = "Int. Sociol.",
  ["international sociology : journal of the international sociological association"] = "Int Sociol",
  ["international soil and water conservation research"] = "Int. Soil Water Conserv. Res.",
  ["international solid-state sensors, actuators and microsystems conference : [proceedings]. international conference on solid-state sensors, actuators, and microsystems"] = "Int Solid State Sens Actuators Microsyst Conf",
  ["international sport coaching journal"] = "Int Sport Coach J",
  ["international statistical review"] = "Int. Stat. Rev.",
  ["international statistical review = revue internationale de statistique"] = "Int Stat Rev",
  ["international statistical review. revue internationale de statistique"] = "Int. Stat. Rev.",
  ["international std research & reviews"] = "Int STD Res Rev",
  ["international studies in economic modelling"] = "Internat. Stud. Econom. Model.",
  ["international studies in philosophy"] = "ISPh",
  ["international studies in the history of mathematics and its teaching"] = "Int. Stud. Hist. Math. Teach.",
  ["international studies in the philosophy of science"] = "Int. Stud. Philos. Sci.",
  ["international studies quarterly : a publication of the international studies association"] = "Int Stud Q",
  ["international sugar journal"] = "Int. Sugar J.",
  ["international surgery"] = "Int Surg",
  ["international symposia in economic theory and econometrics"] = "Int. Sympos. Econom. Theory Econometrics",
  ["international symposium on fluoridation and preventive dentistry"] = "Int Symp Fluorid Prev Dent",
  ["international symposium on information theory and its applications. international symposium on information theory and its applications"] = "Int Symp Inf Theory Appl",
  ["international symposium on mixed and augmented reality : (ismar) [proceedings]. ieee and acm international symposium on mixed and augmented reality"] = "Int Symp Mix Augment Real",
  ["international symposium on process systems engineering"] = "Int Symp Process Syst Eng",
  ["international symposium on reproductive physiology of fish"] = "Int Symp Reprod Physiol Fish",
  ["international tax and public finance"] = "Int Tax Public Financ",
  ["international trade journal"] = "Int. Trade J.",
  ["international transaction journal of engineering, management, & applied sciences & technologies"] = "Int Trans J Eng Manag Appl Sci Technol",
  ["international transactions in operational research"] = "Int. Tran. Oper. Res.",
  ["international transactions in operational research : a journal of the international federation of operational research societies"] = "Int Trans Oper Res",
  ["international transactions on electrical energy systems"] = "Int. Trans. Electr. Energy Syst.",
  ["international tree crops journal"] = "Int. Tree Crops J.",
  ["international trends in immunity"] = "Int Trends Immun",
  ["international trends in manufacturing technology"] = "Internat. Trends Manufacturing Tech.",
  ["international turfgrass society research journal"] = "Int Turfgrass Soc Res J",
  ["international union of crystallography crystallographic symposia"] = "Internat. Union Cryst. Cryst. Sympos.",
  ["international urogynecology journal"] = "Int Urogynecol J",
  ["international urogynecology journal and pelvic floor dysfunction"] = "Int Urogynecol J Pelvic Floor Dysfunct",
  ["international urology and nephrology"] = "Int Urol Nephrol",
  ["international wildlife"] = "Int Wildl",
  ["international wood products journal"] = "Int. Wood Prod. J.",
  ["international work-conference on the interplay between natural and artificial computation"] = "Int Work Conf Interp Nat Artif Comput",
  ["international workshop on databases and expert systems applications : proceedings"] = "Int Workshop databases Expert Syst Appl",
  ["international wound journal"] = "Int Wound J",
  ["international year-book of game theory and applications"] = "Int. Year-Book Game Theory Appl.",
  ["international zoo yearbook"] = "Int. Zoo Yearb.",
  ["internationale mitteilungen für bodenkunde"] = "Int. Mitt. Bodenkd.",
  ["internationale revue der gesamten hydrobiologie"] = "Int. Rev. Gesamten Hydrobiol.",
  ["internationale revue der gesamten hydrobiologie und hydrographie"] = "Int. Rev. Gesamten Hydrobiol. Hydrogr.",
  ["internationale spectator"] = "Int Spect",
  ["internationale wissenschaftliche korrespondenz zur geschichte der deutschen arbeiterbewegung"] = "IWK Int Wiss Korresp Gesch Dtsch Arb Beweg",
  ["internationale zeitschrift fur angewandte physiologie einschliesslich arbeitsphysiologie"] = "Int. Z. Angew. Physiol.",
  ["internationale zeitschrift fur angewandte physiologie, einschliesslich arbeitsphysiologie"] = "Int Z Angew Physiol",
  ["internationale zeitschrift fur klinische pharmakologie, therapie und toxikologie"] = "Int. Z. Klin. Pharmakol. Ther. Toxikol.",
  ["internationale zeitschrift fur klinische pharmakologie, therapie, und toxikologie. international journal of clinical pharmacology, therapy, and toxicology"] = "Int Z Klin Pharmakol Ther Toxikol",
  ["internationale zeitschrift fur physikalische medizin und rehabilitation, mit zentralblatt fur die gesamte rehabilitation"] = "Int Z Phys Med Rehabil",
  ["internationale zeitschrift fur vitamin- und ernahrungsforschung. beiheft"] = "Int. Z. Vitam. Ernahrungsforsch. Beih.",
  ["internationale zeitschrift fur vitaminforschung (international journal of vitamin research)"] = "Int. Z. Vitaminforsch.",
  ["internationale zeitschrift fur vitaminforschung. beiheft"] = "Int Z Vitaminforsch Beih",
  ["internationale zeitschrift fur vitaminforschung. international journal of vitamin research. journal international de vitaminologie"] = "Int Z Vitaminforsch",
  ["internationale zeitschrift für angewandte physiologie, einschliesslich arbeitsphysiologie"] = "Int Z Angew Physiol",
  ["internationale zeitschrift für klinische pharmakologie, therapie, und toxikologie. international journal of clinical pharmacology, therapy, and toxicology"] = "Int Z Klin Pharmakol Ther Toxikol",
  ["internationale zeitschrift für physikalische medizin und rehabilitation, mit zentralblatt für die gesamte rehabilitation"] = "Int Z Phys Med Rehabil",
  ["internationale zeitschrift für vitamin- und ernährungsforschung. beiheft"] = "Int Z Vitam Ernahrungsforsch Beih",
  ["internationale zeitschrift für vitaminforschung. beiheft"] = "Int Z Vitaminforsch Beih",
  ["internationale zeitschrift für vitaminforschung. internătional journal of vitamin research. journal international de vitaminologie"] = "Int Z Vitaminforsch",
  ["internationale zeitschriftenschau für bibelwissenschaft und grenzgebiete"] = "IntZSchauBibelWiss",
  ["internationales archiv fur arbeitsmedizin"] = "Int. Arch. Arbeitsmed.",
  ["internationales archiv fur gewerbepathologie und gewerbehygiene"] = "Int Arch Gewerbepathol Gewerbehyg",
  ["internationales archiv für arbeitsmedizin"] = "Int Arch Arbeitsmed",
  ["internationales archiv für ethnographie"] = "Arch Int Ethnogr",
  ["internationales archiv für gewerbepathologie und gewerbehygiene"] = "Int Arch Gewerbepathol Gewerbehyg",
  ["internationales asienforum"] = "Int Asianforum",
  ["internet computing"] = "Internet Comput.",
  ["internet healthcare strategies"] = "Internet Healthc Strateg",
  ["internet histories"] = "Internet Hist.",
  ["internet interventions"] = "Internet Interv",
  ["internet journal of chemistry"] = "Internet J. Chem.",
  ["internet journal of vibrational spectroscopy"] = "Internet J. Vib. Spectro.",
  ["internet mathematics"] = "Internet Math.",
  ["internet of things"] = "Internet Things",
  ["internet of things and cyber-physical systems"] = "Internet Things Cyber-Phys. Syst.",
  ["internet technology letters"] = "Internet Technol. Lett.",
  ["internistische praxis"] = "Internist Prax",
  ["interpersona : an international journal of personal relationships"] = "Interpers Int J Pers Relatsh",
  ["interpreting : international journal of research and practice in interpreting"] = "Interpreting (Amst)",
  ["intersezioni (bologna, italy : periodical)"] = "Intersezioni",
  ["intervención psicosocial"] = "Interv Psicosoc",
  ["intervention (amstelveen, netherlands)"] = "Intervention (Amstelveen)",
  ["intervention in school and clinic"] = "Interv Sch Clin",
  ["interventional cardiology"] = "Interv Cardiol (Lond)",
  ["interventional cardiology (london, england)"] = "Interv Cardiol",
  ["interventional cardiology clinics"] = "Interv Cardiol Clin",
  ["interventional medicine & applied science"] = "Interv Med Appl Sci",
  ["interventional neurology"] = "Interv Neurol",
  ["interventional neuroradiology"] = "Interventional Neuroradiol.",
  ["interventional neuroradiology : journal of peritherapeutic neuroradiology, surgical procedures and related neurosciences"] = "Interv Neuroradiol",
  ["interventional oncology 360"] = "Interv Oncol 360",
  ["interventions (london, england)"] = "Interventions (Lond)",
  ["intestinal research"] = "Intest Res",
  ["intouch (melville, n.y.)"] = "InTouch",
  ["intractable & rare diseases research"] = "Intractable Rare Dis Res",
  ["intrinsically disordered proteins"] = "Intrinsically Disord Proteins",
  ["intus : revista de las unidades docentes de psicologia medica, clinica psiquiatrica e historia de la medicina"] = "Intus",
  ["intus : revista de las unidades docentes de psicología médica, clínica psiquíatrica e historia de la medicina"] = "Intus",
  ["invariant theory and algebraic transformation groups"] = "Invariant Theory Algebr. Transform. Groups",
  ["invasion & metastasis"] = "Invasion Metastasis",
  ["invasion and metastasis"] = "Invasion Metastasis",
  ["invasive plant science and management"] = "Invasive Plant Sci Manag",
  ["inventiones mathematica"] = "Invent. Math.",
  ["inventiones mathematicae"] = "Invent. Math.",
  ["inventions (basel, switzerland)"] = "Inventions (Basel)",
  ["inventory optimization"] = "Inventory Optim.",
  ["inverse and ill-posed problems series"] = "Inverse Ill-posed Probl. Ser.",
  ["inverse problems"] = "Inverse Probl",
  ["inverse problems and imaging"] = "Inverse Probl. Imaging",
  ["inverse problems and imaging (springfield, mo.)"] = "Inverse Probl Imaging (Springfield)",
  ["inverse problems in engineerign"] = "Inverse Prob. Eng.",
  ["inverse problems in engineering"] = "Inverse Probl. Eng.",
  ["inverse problems in science and engineering"] = "Inverse Probl. Sci. Eng.",
  ["inverse problems. an international journal on the theory and practice of inverse problems, inverse methods and computerized inversion of data"] = "Inverse Problems",
  ["invertebrate biology"] = "Invertebr. Biol.",
  ["invertebrate biology : a quarterly journal of the american microscopical society and the division of invertebrate zoology/asz"] = "Invertebr Biol",
  ["invertebrate neuroscience"] = "Invertebr. Neurosci.",
  ["invertebrate neuroscience : in"] = "Invert Neurosci",
  ["invertebrate reproduction & development"] = "Invertebr Reprod Dev",
  ["invertebrate survival journal : isj"] = "Invertebrate Surviv J",
  ["invertebrate systematics"] = "Invertebr Syst",
  ["investigacion clinica"] = "Invest Clin",
  ["investigacion economica / escuela nacional de economia, universidad nacional autonoma de mexico"] = "Invest Econ",
  ["investigacion medica internacional"] = "Invest Med Int",
  ["investigacion operacional"] = "Investigacion Oper.",
  ["investigacion y ciencia"] = "Investig Cienc",
  ["investigaciones economicas"] = "Investigaciones Econ.",
  ["investigaciones historicas (universidad de valladolid. departamento de historia moderna)"] = "Invest Hist",
  ["investigaciones históricas (universidad de valladolid. departamento de historia moderna)"] = "Invest Hist",
  ["investigaciones marinas cicimar"] = "Investig Mar CICIMAR",
  ["investigación clínica"] = "Invest Clin",
  ["investigación económica"] = "Invest Econ",
  ["investigación médica internacional"] = "Invest Med Int",
  ["investigación y educación en enfermería"] = "Invest Educ Enferm",
  ["investigación económica"] = "Investigación Econ.",
  ["investigación operacional"] = "Investigación Oper.",
  ["investigation and practice in pediatric pathology"] = "Investig. Pract. Pediatr. Pathol.",
  ["investigational new drugs"] = "Invest New Drugs",
  ["investigations in mathematics learning"] = "Invest. Math. Learn.",
  ["investigative & cell pathology"] = "Invest Cell Pathol",
  ["investigative and cell pathology"] = "Invest. Cell Pathol.",
  ["investigative and clinical urology"] = "Investig Clin Urol",
  ["investigative genetics"] = "Investig Genet",
  ["investigative magnetic resonance imaging"] = "Investig Magn Reson Imaging",
  ["investigative ophthalmology"] = "Invest Ophthalmol",
  ["investigative ophthalmology & visual science"] = "Invest Ophthalmol Vis Sci",
  ["investigative ophthalmology and visual science"] = "Invest. Ophthalmol. Vis. Sci.",
  ["investigative radiology"] = "Invest Radiol",
  ["investigative urology"] = "Invest Urol",
  ["investigative urology (berlin, germany)"] = "Investig Urol (Berl)",
  ["investment policy"] = "Invest. Pol.",
  ["investor-owned hospital review"] = "Investor. Owned Hosp. Rev.",
  ["invigilata lucernis"] = "InvLuc",
  ["invigilata lucernis: rivista dell’istituto di latino"] = "InvLuc",
  ["involve : a journal of mathematics"] = "Involve",
  ["involve. a journal of mathematics"] = "Involve",
  ["involve: a journal of mathematics"] = "Involve",
  ["inzhenerno-fizicheskii zhurnal"] = "Inzh.-Fiz. Zh.",
  ["inzynieria chemiczna i procesowa"] = "Inz. Chem. Procesowa",
  ["iobc/wprs bulletin"] = "Bulletin OILB/SROP|IOBC/WPRS Bull.",
  ["iobc/wprs bulletin = bulletin oilb/srop"] = "IOBC WPRS Bull",
  ["iochimica et biophysica acta, biomembranes"] = "Biochim. Biophys. Acta, Biomembr.",
  ["iom letters"] = "IOM Lett",
  ["ion channels"] = "Ion Channels",
  ["ion exchange and membranes"] = "Ion Exch Membr",
  ["ion exchange and solvent extraction"] = "Ion Exch. Solvent Extr.",
  ["iop concise physics"] = "IOP Concise Phys.",
  ["iop conference series. earth and environmental science"] = "IOP Conf Ser Earth Environ Sci",
  ["iop conference series. materials science and engineering"] = "IOP Conf Ser Mater Sci Eng",
  ["iop conference series: earth and environmental science"] = "IOP Conf. Ser.: Earth Environ. Sci.",
  ["iop conference series: materials science and engineering"] = "IOP Conf. Ser.: Mater. Sci. Eng.",
  ["iop expanding physics"] = "IOP Expand. Phys.",
  ["iop scinotes"] = "IOP SciNotes",
  ["iop short meetings series"] = "IOP Short Meet. Ser.",
  ["iosr journal of agriculture and veterinary science"] = "IOSR J Agric Vet Sci",
  ["iosr journal of biotechnology and biochemistry"] = "IOSR J Biotechnol Biochem",
  ["iosr journal of computer engineering"] = "IOSR J Comput Eng",
  ["iosr journal of environmental science, toxicology and food technology"] = "IOSR J Environ Sci Toxicol Food Technol",
  ["iosr journal of pharmacy"] = "IOSR J Pharm",
  ["iosr journal of pharmacy and biological sciences"] = "IOSR J Pharm Biol Sci",
  ["iotdi 2017 : 2017 ieee/acm second international conference on internet-of-things design and implementation : proceedings : 18-20 april 2017, pittsburgh, pennsylvania, usa. iotdi (conference) (2nd : 2017 : pittsburgh, pa.)"] = "IoTDI 2017 (2017)",
  ["iowa agricultural experiment station research bulletin"] = "IOWA Agric. Exp. Stn. Res. Bull.",
  ["iowa code annotated. iowa"] = "Iowa Code Annot Iowa",
  ["iowa dental bulletin"] = "Iowa Dent Bull",
  ["iowa dental journal"] = "Iowa Dent. J.",
  ["iowa law review"] = "Iowa Law Rev",
  ["iowa medicine"] = "Iowa Med.",
  ["iowa medicine : journal of the iowa medical society"] = "Iowa Med",
  ["iowa orthopaedic journal"] = "Iowa Orthop. J.",
  ["iowa state college journal of science"] = "Iowa State Coll J Sci",
  ["iowa state university veterinarian"] = "Iowa State Univ Vet",
  ["ipa dissertation series"] = "IPA Diss. Ser.",
  ["ipol journal. image processing online"] = "IPOL J. Image Process. Online",
  ["ippa - newsletter"] = "IPPA Newsl",
  ["ippf medical bulletin"] = "IPPF Med Bull",
  ["ippf news"] = "IPPF News",
  ["ippf open file"] = "IPPF Open File",
  ["ippf/whr news service. international planned parenthood federation. western hemisphere region"] = "IPPF WHR News Serv",
  ["ippologia : rivista ufficiale sive"] = "Ippologia (Cremona)",
  ["ippta journal"] = "Ippta J.",
  ["ipsn : [proceedings]. ipsn (conference)"] = "IPSN",
  ["iral, international review of applied linguistics in language teaching : revue internationale de linguistique appliquée enseignement des langues. internationale zeitschrift für angewandte linguistik in der spracherziehung"] = "IRAL Int Rev Appl Linguist Lang Teach",
  ["iran biomedical journal"] = "Iran. Biomed. J.",
  ["iran journal of computer science"] = "Iran J.Comput. Sci.",
  ["iran journal of science"] = "Iran J. Sci.",
  ["iran journal of toxicology"] = "Iran J. Toxicol.",
  ["iran. journal of the british institute of persian studies"] = "Iran",
  ["iranian biomedical journal"] = "Iran Biomed J",
  ["iranian endodontic journal"] = "Iran Endod J",
  ["iranian journal of allergy, asthma, and immunology"] = "Iran J Allergy Asthma Immunol",
  ["iranian journal of animal biosystematics"] = "Iran J Anim Biosyst",
  ["iranian journal of applied animal science"] = "Iran J Appl Anim Sci",
  ["iranian journal of arthropod-borne diseases"] = "Iran J Arthropod Borne Dis",
  ["iranian journal of basic medical sciences"] = "Iran J Basic Med Sci",
  ["iranian journal of biotechnology"] = "Iran J Biotechnol",
  ["iranian journal of cancer prevention"] = "Iran J Cancer Prev",
  ["iranian journal of chemistry & chemical engineering"] = "Iran. J. Chem. Chem. Eng.",
  ["iranian journal of child neurology"] = "Iran J Child Neurol",
  ["iranian journal of environmental health science & engineering"] = "Iranian J Environ Health Sci Eng",
  ["iranian journal of fisheries sciences"] = "Iran J Fish Sci",
  ["iranian journal of fuzzy systems"] = "Iran. J. Fuzzy Syst.",
  ["iranian journal of genetics and plant breeding"] = "Iran J Genet Plant Breed",
  ["iranian journal of ichthyology"] = "Iran J Ichthyol",
  ["iranian journal of immunology : iji"] = "Iran J Immunol",
  ["iranian journal of kidney diseases"] = "Iran J Kidney Dis",
  ["iranian journal of materials science and engineering"] = "Iran. J. Mater. Sci. Eng.",
  ["iranian journal of mathematical sciences and informatics"] = "Iran. J. Math. Sci. Inform.",
  ["iranian journal of medical sciences"] = "Iran J Med Sci",
  ["iranian journal of microbiology"] = "Iran J Microbiol",
  ["iranian journal of neurology"] = "Iran J Neurol",
  ["iranian journal of nursing and midwifery research"] = "Iran J Nurs Midwifery Res",
  ["iranian journal of ophthalmology"] = "Iran J Ophthalmol",
  ["iranian journal of otorhinolaryngology"] = "Iran J Otorhinolaryngol",
  ["iranian journal of parasitology"] = "Iran J Parasitol",
  ["iranian journal of pathology"] = "Iran J Pathol",
  ["iranian journal of pediatric hematology and oncology"] = "Iran J Ped Hematol Oncol",
  ["iranian journal of pediatrics"] = "Iran J Pediatr",
  ["iranian journal of pharmaceutical research : ijpr"] = "Iran J Pharm Res",
  ["iranian journal of plant pathology"] = "Iran J Plant Pathol",
  ["iranian journal of plant physiology"] = "Iran. J. Plant Physiol.",
  ["iranian journal of polymer science & technology"] = "Iran. J. Polym Sci. Technol.",
  ["iranian journal of psychiatry"] = "Iran J Psychiatry",
  ["iranian journal of psychiatry and behavioral sciences"] = "Iran J Psychiatry Behav Sci",
  ["iranian journal of public health"] = "Iran J Public Health",
  ["iranian journal of radiology : a quarterly journal published by the iranian radiological society"] = "Iran J Radiol",
  ["iranian journal of reproductive medicine"] = "Iran J Reprod Med",
  ["iranian journal of science"] = "Iran. J. Sci.",
  ["iranian journal of science & technology transactions of civil engineering"] = "Iran. J. Sci. Technol. Trans. Civ. Eng.",
  ["iranian journal of science & technology transactions of electrical engineering"] = "Iran. J. Sci. Technol. Trans. Electr. Eng.",
  ["iranian journal of science & technology transactions of mechanical engineering"] = "Iran. J. Sci. Technol. Trans. Mech. Eng.",
  ["iranian journal of science and technology"] = "Iranian J. Sci. Tech.",
  ["iranian journal of science and technology. transaction a, science"] = "Iran J Sci Technol Trans A Sci",
  ["iranian journal of science and technology. transaction a. science"] = "Iran. J. Sci. Technol. Trans. A Sci.",
  ["iranian journal of veterinary medicine"] = "Iran J Vet Med",
  ["iranian journal of veterinary research"] = "Iran J Vet Res",
  ["iranian journal of veterinary science and technology"] = "Iran J Vet Sci Technol",
  ["iranian journal of veterinary surgery"] = "Iran J Vet Surg",
  ["iranian journal of virology"] = "Iran J Virol",
  ["iranian mathematical society"] = "Bull. Iranian Math. Soc.",
  ["iranian polymer journal"] = "Iran. Polym. J.",
  ["iranian red crescent medical journal"] = "Iran Red Crescent Med J",
  ["iranian studies : bulletin of the society for iranian cultural and social studies"] = "Iran Stud",
  ["iranica antiqua"] = "IA",
  ["iraqi dental journal"] = "Iraqi Dent J",
  ["iraqi journal of medical sciences"] = "Iraqi J Med Sci",
  ["irb; a review of human subjects research"] = "IRB.",
  ["ircs journal of international research communications"] = "ICRS J Int Res Commun",
  ["ircs journal of medical science"] = "IRCS J Med Sci",
  ["ircs medical science"] = "IRCS Med Sci",
  ["ircs medical science. clinical medicine"] = "IRCS Med Sci Clin Med.",
  ["ircs medical science: biomedical technology"] = "IRCS Med Sci Biomed Technol",
  ["ircs medical science: cardiovascular system"] = "IRCS Med Sci Cardiovasc Syst",
  ["ircs medical science: reproduction, obstetrics and gynecology"] = "Irc Med Sci Reprod Obstet Gynecol",
  ["ire transactions on  industrial electronics"] = "IRE Trans. Ind. Electron.",
  ["ire transactions on  information theory"] = "IRE Trans. Inf. Theory",
  ["ire transactions on aeronautical and navigational electronics"] = "IRE Trans. Aeronaut. Navig. Electron.",
  ["ire transactions on aerospace and navigational electronics"] = "IRE Trans. Aerosp. Navig. Electron.",
  ["ire transactions on antennas and propagation"] = "IRE Trans. Anntenas Propag.",
  ["ire transactions on audio"] = "IRE Trans. Audio",
  ["ire transactions on automatic control"] = "IRE Trans. Autom. Control",
  ["ire transactions on bio-medical electronics"] = "Ire Trans Biomed Electron",
  ["ire transactions on broadcast and television receivers"] = "IRE Trans. Broadcast Telev. Receivers",
  ["ire transactions on broadcast transmission systems"] = "IRE Trans. Broadcast Transm. Syst.",
  ["ire transactions on broadcasting"] = "IRE Trans. Broadcast.",
  ["ire transactions on circuit theory"] = "IRE Trans. Circuit Theory",
  ["ire transactions on communication systems"] = "IRE Trans. Commun. Syst.",
  ["ire transactions on component parts"] = "IRE Trans. Compon. Parts",
  ["ire transactions on electron devices"] = "IRE Trans. Electron Devices",
  ["ire transactions on electronic computers"] = "IRE Trans. Electron. Comput.",
  ["ire transactions on engineering management"] = "IRE Trans. Eng. Manage.",
  ["ire transactions on human factors in electronics"] = "IRE Trans. Hum. Factors Electron.",
  ["ire transactions on industrial electronics"] = "IRE Trans. Ind. Electron.",
  ["ire transactions on information theory"] = "IRE Trans. Inf. Theory",
  ["ire transactions on instrumentation"] = "IRE Trans. Instrum.",
  ["ire transactions on medical electronics"] = "IRE Trans Med Electron",
  ["ire transactions on microwave theory and techniques"] = "IRE Trans. Microwave Theory Tech.",
  ["ire transactions on military electronics"] = "IRE Trans. Mil. Electron.",
  ["ire transactions on nuclear science"] = "IRE Trans. Nucl. Sci.",
  ["ire transactions on product engineering and production"] = "IRE Trans. Prod. Eng. Prod.",
  ["ire transactions on production techniques"] = "IRE Trans. Prod. Tech.",
  ["ire transactions on radio frequency interference"] = "IRE Trans. Radio Freq. Interference",
  ["ire transactions on reliability and quality control"] = "IRE Trans. Reliab. Qual. Control",
  ["ire transactions on space electronics and telemetry"] = "IRE Trans. Space Electron. Telem.",
  ["ire transactions on telemetry and remote control"] = "IRE Trans. Telem. Remote Control",
  ["ire transactions on ultrasonic engineering"] = "IRE Trans. Ultrason. Eng.",
  ["ire transactions on vehicular communications"] = "IRE Trans. Veh. Commun.",
  ["irish banking review"] = "Irish Banking Rev.",
  ["irish economic and social history : journal of the economic and social history society of ireland"] = "Ir Econ Soc Hist",
  ["irish geography : bulletin of the geographical society of ireland"] = "Ir Geogr",
  ["irish historical studies : joint journal of the irish historical society and the ulster society for irish historical studies"] = "Ir Hist Stud",
  ["irish journal of agricultural and food research"] = "Ir. J. Agric. Food Res.",
  ["irish journal of agricultural economics and rural sociology"] = "Irish J. Agr. Econ. Rural Sociology",
  ["irish journal of earth sciences"] = "Ir. J. Earth Sci.",
  ["irish journal of medical science"] = "Ir J Med Sci",
  ["irish journal of occupational therapy"] = "Ir. J. Occup. Ther.",
  ["irish journal of psychological medicine"] = "Ir J Psychol Med",
  ["irish law reports monthly"] = "Ir Law Rep Mon",
  ["irish law reports monthly / [editor, bart d. daly]"] = "Ir Law Rep Mon",
  ["irish mathematical society bulletin"] = "Irish Math. Soc. Bull.",
  ["irish medical journal"] = "Ir Med J",
  ["irish nurses journal"] = "Ir. Nurses J.",
  ["irish nurses' journal"] = "Ir Nurses J",
  ["irish nurses’ journal"] = "Ir Nurses J",
  ["irish nursing and hospital world"] = "Ir Nurs Hosp World",
  ["irish nursing news"] = "Ir. Nurs. News",
  ["irish studies review"] = "Ir Stud Rev",
  ["irish university review"] = "Ir. Univ. Rev.",
  ["irish veterinary journal"] = "Ir Vet J",
  ["irm; information and records management"] = "IRM.",
  ["irma lectures in mathematics and theoretical physics"] = "IRMA Lect. Math. Theor. Phys.",
  ["iron & steel technology"] = "Iron Steel Technol.",
  ["ironmaking & steelmaking"] = "Ironmaking Steelmaking",
  ["ironmaking and steelmaking"] = "Ironmaking Steelmaking",
  ["irrigation and drainage"] = "Irrig. Drain.",
  ["irrigation and drainage (international commission on irrigation and drainage)"] = "Irrig Drain",
  ["irrigation and drainage systens"] = "Irrig. Drain. Syst.",
  ["irrigation science"] = "Irrig Sci",
  ["is&t international symposium on electronic imaging"] = "IS&T Int Symp Electron Imaging",
  ["isa transactions"] = "ISA Trans",
  ["isbt science series"] = "ISBT Sci Ser",
  ["iscb community journal"] = "ISCB Community J.",
  ["iscience notes"] = "Isci Notes",
  ["ise review"] = "ISE Rev.",
  ["ish journal of hydraulic engineering"] = "ISH J. Hydraul. Eng.",
  ["ishikawa-ken hakusan shizen hogo sentā kenkyū hōkoku"] = "Ishikawaken Hakusan Shizen Hogo Senta Kenkyu Hokoku",
  ["isij international"] = "ISIJ Int.",
  ["isis. an international review devoted to the history of science and its cultural influences"] = "Isis",
  ["isis; an international review devoted to the history of science and its cultural influences"] = "Isis",
  ["isl law review. international school of law"] = "ISL Law Rev",
  ["isla (mangilao, guam)"] = "Isla",
  ["islamic economic studies"] = "Islamic Econ. Stud.",
  ["islamic philosophy, theology, and science"] = "Islam. Philos. Theol. Sci.",
  ["islamic republic of iran"] = "J. Sci. Islam. Repub. Iran",
  ["islamic republic of iran. journal of sciences"] = "J. Sci. Islam. Repub. Iran",
  ["islamic studies"] = "Islam Stud",
  ["island arc"] = "Isl. Arc",
  ["isme communications"] = "ISME Commun.",
  ["isme journal"] = "ISME J.",
  ["isme/glomis electronic journal"] = "ISME GLOMIS Electron J",
  ["isokinetics and exercise science"] = "Isokinet Exerc Sci",
  ["isotope geoscience"] = "Isot. Geosci.",
  ["isotopes and radiation technology"] = "Isot. Radiat. Technol.",
  ["isotopes in environmental and health studies"] = "Isotopes Environ Health Stud",
  ["isozyme bulletin"] = "Isozyme Bull",
  ["isozymes; current topics in biological medical research"] = "Isozymes Curr. Top. Biol. Med. Res.",
  ["ispra courses on materials, engineering and mechanical science"] = "Ispra Courses Materials, Engrg. Mech. Sci.",
  ["ispra courses on nuclear engineering and technology"] = "Ispra Courses Nuclear Engrg. Tech.",
  ["ispra courses on reliability and risk analysis"] = "Ispra Courses Reliab. Risk Anal.",
  ["isprs annals of the photogrammetry, remote sensing and spatial information sciences"] = "ISPRS Ann. Photogramm. Remote Sens. Spatial Inf. Sci.",
  ["isprs international journal of geo-information"] = "ISPRS Int J Geoinf",
  ["isprs journal of photogrammetry and remote sensing"] = "ISPRS J. Photogramm. Remote Sens.",
  ["isprs journal of photogrammetry and remote sensing : official publication of the international society for photogrammetry and remote sensing (isprs)"] = "ISPRS J Photogramm Remote Sens",
  ["isprs open journal of photogrammetry and remote sensing"] = "ISPRS Open J. Photogramm. Remote Sens.",
  ["isra 2013"] = "ISRA 2013 (2013)",
  ["israel annals of psychiatry and related disciplines"] = "Isr. Ann. Psychiatr. Relat. Discip.",
  ["israel exploration journal"] = "IEJ",
  ["israel journal of chemistry"] = "Isr. J. Chem.",
  ["israel journal of dental medicine"] = "Isr. J. Dent. Med.",
  ["israel journal of dental sciences"] = "Isr J Dent Sci",
  ["israel journal of ecology & evolution"] = "Isr J Ecol Evol",
  ["israel journal of entomology"] = "Isr J Entomol",
  ["israel journal of experimental medicine"] = "Isr J Exp Med",
  ["israel journal of health policy research"] = "Isr J Health Policy Res",
  ["israel journal of mathematics"] = "Israel J. Math.",
  ["israel journal of medical sciences"] = "Isr J Med Sci",
  ["israel journal of plant sciences"] = "Isr J Plant Sci",
  ["israel journal of psychiatry and related sciences"] = "Isr. J. Psychiatry Relat. Sci.",
  ["israel journal of zoology"] = "Isr J Zool",
  ["israel law reports. israel. bet ha-mishpaṭ ha-ʻelyon"] = "Israel Law Rep",
  ["israel law review"] = "Isr Law Rev",
  ["israel mathematical conference proceedings"] = "Israel Math. Conf. Proc.",
  ["israel medical association journal"] = "Isr. Med. Assoc. J.",
  ["israel medical journal"] = "Isr Med J",
  ["israel museum studies in archaeology. an annual publication by the samuel bronfman biblical and archaeological museum of the israel museum, jerusalem"] = "IsrMusStA",
  ["israel numismatic journal"] = "IsrNumJ",
  ["israel oriental studies"] = "Isr Orient Stud",
  ["israel social science research"] = "Isr Soc Sci Res",
  ["israel studies (bloomington, ind.)"] = "Israel Stud (Bloomingt Ind)",
  ["isrn addiction"] = "ISRN Addict",
  ["isrn aids"] = "ISRN AIDS",
  ["isrn allergy"] = "ISRN Allergy",
  ["isrn anatomy"] = "ISRN Anat",
  ["isrn anesthesiology"] = "ISRN Anesthesiol",
  ["isrn biochemistry"] = "ISRN Biochem",
  ["isrn bioinformatics"] = "ISRN Bioinform",
  ["isrn biomathematics"] = "ISRN Biomath",
  ["isrn biotechnology"] = "ISRN Biotechnol",
  ["isrn cardiology"] = "ISRN Cardiol",
  ["isrn cell biology"] = "ISRN Cell Biol",
  ["isrn computational biology"] = "ISRN Comput Biol",
  ["isrn dentistry"] = "ISRN Dent",
  ["isrn dermatology"] = "ISRN Dermatol",
  ["isrn ecology"] = "ISRN Ecol",
  ["isrn endocrinology"] = "ISRN Endocrinol",
  ["isrn entomology"] = "ISRN Entomol",
  ["isrn family medicine"] = "ISRN Family Med",
  ["isrn gastroenterology"] = "ISRN Gastroenterol",
  ["isrn genetics"] = "ISRN Genetics",
  ["isrn geriatrics"] = "ISRN Geriatr",
  ["isrn hematology"] = "ISRN Hematol",
  ["isrn hepatology"] = "ISRN Hepatol",
  ["isrn infectious diseases"] = "ISRN Infect Dis",
  ["isrn inflammation"] = "ISRN Inflamm",
  ["isrn microbiology"] = "ISRN Microbiol",
  ["isrn minimally invasive surgery"] = "ISRN Minim Invasive Surg",
  ["isrn molecular biology"] = "ISRN Mol Biol",
  ["isrn molecular imaging"] = "ISRN Mol Imaging",
  ["isrn nanomaterials"] = "ISRN Nanomater",
  ["isrn nephrology"] = "ISRN Nephrol",
  ["isrn neurology"] = "ISRN Neurol",
  ["isrn neuroscience"] = "ISRN Neurosci",
  ["isrn nursing"] = "ISRN Nurs",
  ["isrn nutrition"] = "ISRN Nutr",
  ["isrn obesity"] = "ISRN Obes",
  ["isrn obstetrics and gynecology"] = "ISRN Obstet Gynecol",
  ["isrn oncology"] = "ISRN Oncol",
  ["isrn ophthalmology"] = "ISRN Ophthalmol",
  ["isrn optics"] = "ISRN Opt",
  ["isrn organic chemistry"] = "ISRN Org Chem",
  ["isrn orthopedics"] = "ISRN Orthop",
  ["isrn otolaryngology"] = "ISRN Otolaryngol",
  ["isrn oxidative medicine"] = "ISRN Oxidative Med",
  ["isrn pain"] = "ISRN Pain",
  ["isrn parasitology"] = "ISRN Parasitol",
  ["isrn pathology"] = "ISRN Pathol",
  ["isrn pediatrics"] = "ISRN Pediatr",
  ["isrn pharmaceutics"] = "ISRN Pharm",
  ["isrn pharmacology"] = "ISRN Pharmacol",
  ["isrn physiology"] = "ISRN Physiol",
  ["isrn preventive medicine"] = "ISRN Prev Med",
  ["isrn psychiatry"] = "ISRN Psychiatry",
  ["isrn public health"] = "ISRN Public Health",
  ["isrn pulmonology"] = "ISRN Pulmonol",
  ["isrn radiology"] = "ISRN Radiol",
  ["isrn rehabilitation"] = "ISRN Rehabil",
  ["isrn rheumatology"] = "ISRN Rheumatol",
  ["isrn stroke"] = "ISRN Stroke",
  ["isrn structural biology"] = "ISRN Struct Biol",
  ["isrn surgery"] = "ISRN Surg",
  ["isrn tissue engineering"] = "ISRN Tissue Eng",
  ["isrn toxicology"] = "ISRN Toxicol",
  ["isrn tropical medicine"] = "ISRN Trop Med",
  ["isrn urology"] = "ISRN Urol",
  ["isrn veterinary science"] = "ISRN Vet Sci",
  ["isrn virology"] = "ISRN Virol",
  ["isrn zoology"] = "ISRN Zool",
  ["issbd bulletin"] = "ISSBD Bull",
  ["isscs 2013 : international symposium on signals, circuits and systems : 11-12 july, 2013, iasi, romania : program. international symposium on signals, circuits, and systems (12th : 2013 : iași, romania)"] = "ISSCS 2013 (2013)",
  ["issledovaniia po genetike"] = "Issled Genet",
  ["issledovaniya po prikladnoi matematike i informatike"] = "Issled. Prikl. Mat. Inform.",
  ["isss journal of micro and smart systems"] = "ISSS J Micro Smart Syst",
  ["issue brief (center for medicare education)"] = "Issue Brief Cent Medicare Educ",
  ["issue brief (center for studying health system change)"] = "Issue Brief Cent Stud Health Syst Change",
  ["issue brief (commonwealth fund)"] = "Issue Brief (Commonw Fund)",
  ["issue brief (george washington university. center for health services research and policy)"] = "Issue Brief George Wash Univ Cent Health Serv Res Policy",
  ["issue brief (george washington university. medical center. ensuring solutions to alcohol problems)"] = "Issue Brief (George Wash Univ Med Cent Ensuring Solut Alcohol Probl)",
  ["issue brief (george washington university. national health policy forum : 2005)"] = "Issue Brief George Wash Univ Natl Health Policy Forum",
  ["issue brief (george washington university. national health policy forum)"] = "Issue Brief Natl Health Policy Forum",
  ["issue brief (grantmakers in health)"] = "Issue Brief (Grantmakers Health)",
  ["issue brief (health policy tracking service)"] = "Issue Brief Health Policy Track Serv",
  ["issue brief (institute on health care costs and solutions)"] = "Issue Brief (Inst Health Care Costs Solut)",
  ["issue brief (massachusetts health policy forum)"] = "Issue Brief (Mass Health Policy Forum)",
  ["issue brief (public policy institute (american association of retired persons))"] = "Issue Brief (Public Policy Inst (Am Assoc Retired Pers))",
  ["issue brief, national health policy forum"] = "Issue Brief Natl. Health Policy Forum",
  ["issues & studies"] = "Issues Stud",
  ["issues (national council of state boards of nursing (u.s.))"] = "Issues",
  ["issues (saint louis, mo.)"] = "Issues (St Louis Mo)",
  ["issues in brief (alan guttmacher institute)"] = "Issues Brief (Alan Guttmacher Inst)",
  ["issues in comprehensive pediatric nursing"] = "Issues Compr Pediatr Nurs",
  ["issues in criminology"] = "Issues Ciminol",
  ["issues in emerging health technologies"] = "Issues Emerg Health Technol",
  ["issues in ethics"] = "Issues Ethics",
  ["issues in ethics / center for applied ethics"] = "Issues Ethics",
  ["issues in health care"] = "Issues Health Care",
  ["issues in health care (philadelphia, pa.)"] = "Issues Health Care",
  ["issues in health care of women"] = "Issues Health Care Women",
  ["issues in law & medicine"] = "Issues Law Med",
  ["issues in law and medicine"] = "Issues Law Med.",
  ["issues in mathematics education"] = "Issues Math. Ed.",
  ["issues in medical ethics"] = "Issues Med Ethics",
  ["issues in mental health nursing"] = "Issues Ment Health Nurs",
  ["issues in race & society : an interdisciplinary global journal"] = "Issues Race Soc",
  ["issues in reproductive and genetic engineering : journal of international feminist analysis"] = "Issues Reprod Genet Eng",
  ["issues in science and technology"] = "Issues Sci Technol",
  ["issues in subject teaching series"] = "Issues Subj. Teach. Ser.",
  ["ist international surface technology"] = "IST Int. Surf. Technol.",
  ["istanbul arkeoloji müzeleri yıllığı"] = "IstanbAMuezYil",
  ["istanbul tip fakultesi mecmuasi"] = "Istanbul Tip Fak Mecmuasi",
  ["istanbul tıp fakültesi mecmuasi"] = "Istanbul Tip Fak Mecmuasi",
  ["istanbul universitesi dishekimligi fakultesi dergisi"] = "Istanbul Univ. Dishekim. Fak. Derg.",
  ["istanbul university. science faculty. the journal of mathematics, physics and astronomy"] = "Istanb. Univ. Sci. Fac. J. Math. Phys. Astronom.",
  ["istanbuler forschungen"] = "IstForsch",
  ["istanbuler mitteilungen"] = "IstMitt",
  ["istatistik. journal of the turkish statistical association"] = "Istatistik",
  ["isthmia. excavations by the university of chicago under the auspices of the american school of classical studies at athens"] = "Isthmia",
  ["istituto e museo di storia della scienza firenze"] = "Nuncius Ann. Storia Sci.",
  ["istituto lombardo"] = "Istit. Lombardo Accad. Sci. Lett. Rend. A",
  ["istituto per la ricerca di base"] = "Istit. Ric. Base Ser. Monogr. Adv. Math.",
  ["istituto veneto di scienze, lettere ed arti"] = "Istit. Veneto Sci. Lett. Arti Atti Cl. Sci. Fis. Mat. Natur.",
  ["istoricheski pregled"] = "Istor Pregl",
  ["istoricheskiĭ arkhiv (moscow, russia : 1992)"] = "Istor Arh (Mosk 1992)",
  ["istoriia sssr (moscow, russia : 1957)"] = "Istor SSSR",
  ["istorii︠a︡ sssr (moscow, russia : 1957)"] = "Istor SSSR",
  ["istorija (skopje, macedonia)"] = "Istorija",
  ["istorijski casopis : organ istorijskog instituta srpske akademije nauka i umetnosti"] = "Istor Cas",
  ["istorijski časopis : organ istorijskog instituta srpske akademije nauka i umetnosti"] = "Istor Cas",
  ["istorijski zapisi : organ istoriskog instituta i drustva istoricara sr crne gore"] = "Istor Zapisi",
  ["istorijski zapisi : organ istoriskog instituta i društva istoričara sr crne gore"] = "Istor Zapisi",
  ["istoriko-biologicheskie issledovaniia"] = "Istor Biol Issled",
  ["istoriko-filosofsky ezegodnik (history of philosophy yearbook), ed. by the inst. of philosophy of the acad. of sciences of urss"] = "IFE",
  ["istoriko-matematicheskie issledovaniya"] = "Istor.-Mat. Issled.",
  ["istoriya i metodologiya estestvennykh nauk"] = "Istor. Metodol. Estestv. Nauk",
  ["it professional"] = "IT Prof.",
  ["itab 2010 corfu, greece : 10th international conference on information technology and applications in biomedicine : emerging technologies for patient specific healthcare : 2-5 november 2010, aquis corfu holiday palace hotel, greece. international conference on information technology and applications in biomedicine (10th : 2010 : corfu, greece)"] = "ITAB Corfu Greece (2010)",
  ["italia contemporanea"] = "Ital Contemp",
  ["italia forestale e montana, l’"] = "Ital. for. mont.",
  ["italia medioevale e umanistica"] = "Ital Medioev Um",
  ["italia nostra"] = "ItNostr",
  ["italian botanist"] = "Ital. Botan.",
  ["italian general review of dermatology"] = "Ital Gen Rev Dermatol",
  ["italian general review of oto-rhino-laryngology. revue generale italienne d’oto-rhino-laryngologie"] = "Ital Gen Rev Otorhinolaryngol",
  ["italian general review of oto-rhino-laryngology. revue générale italienne d'oto-rhino-laryngologie"] = "Ital Gen Rev Otorhinolaryngol",
  ["italian heart journal"] = "Ital. Heart J.",
  ["italian heart journal : official journal of the italian federation of cardiology"] = "Ital Heart J",
  ["italian heart journal. supplement : official journal of the italian federation of cardiology"] = "Ital Heart J Suppl",
  ["italian journal of aerospace medicine"] = "Ital J Aerosp Med",
  ["italian journal of agrometeorology"] = "Ital. J. Agrometeorol.",
  ["italian journal of agronomy"] = "Ital. J. Agron.",
  ["italian journal of anatomy and embryology"] = "Ital. J. Anat. Embryol.",
  ["italian journal of anatomy and embryology = archivio italiano di anatomia ed embriologia"] = "Ital J Anat Embryol",
  ["italian journal of animal science"] = "Ital J Anim Sci",
  ["italian journal of biochemistry"] = "Ital. J. Biochem.",
  ["italian journal of engineering geology and environment"] = "Ital. J. Eng. Geol. Environ.",
  ["italian journal of food safety"] = "Ital J Food Saf",
  ["italian journal of food science"] = "Ital. J. Food Sci.",
  ["italian journal of food science : ijfs = rivista italiana di scienza degli alimenti"] = "Ital J Food Sci",
  ["italian journal of gastroenterology"] = "Ital. J. Gastroenterol.",
  ["italian journal of gastroenterology and hepatology"] = "Ital J Gastroenterol Hepatol",
  ["italian journal of geoscience"] = "Ital. J. Geosci.",
  ["italian journal of gynaecology & obstetrics : official publication of the società italiana di ginecologia e ostetricia (sigo)"] = "Ital J Gynaecol Obstet",
  ["italian journal of laboratory medicine"] = "Ital. J. Lab. Med.",
  ["italian journal of mycology"] = "Ital J Mycol",
  ["italian journal of neurological sciences"] = "Ital J Neurol Sci",
  ["italian journal of orthopaedics and traumatology"] = "Ital J Orthop Traumatol",
  ["italian journal of orthopaedics and traumatology. supplementum"] = "Ital J Orthop Traumatol Suppl",
  ["italian journal of pediatrics"] = "Ital J Pediatr",
  ["italian journal of pure and applied mathematics"] = "Ital. J. Pure Appl. Math.",
  ["italian journal of surgical sciences"] = "Ital. J. Surg. Sci.",
  ["italian review of agricultural economics"] = "Ital. Rev. Agric. Econ.",
  ["italian society for logic and philosophy of science silfs series"] = "SILFS Ser.",
  ["italian studies"] = "Ital Stud",
  ["italica. cuadernos de trabajos de la escuela española de historia y arqueología en roma"] = "Italica",
  ["italy; documents and notes"] = "Italy Doc Notes",
  ["itch (philadelphia, pa.)"] = "Itch (Phila)",
  ["ite letters on batteries new technologies & medicine"] = "ITE Lett. Batteries New Technol. Med.",
  ["ite-iba letters new technologies & medicine"] = "ITE-IBA Lett. Batteries New Technol. Med.",
  ["items & issues"] = "Items Issues",
  ["itinerario (leiden, netherlands)"] = "Itinerario",
  ["itogi nauki"] = "Itogi Nauk",
  ["itogi nauki i tekhniki. seriya sovremennaya matematika i ee prilozheniya. tematicheskie obzory"] = "Itogi Nauki Tekh. Ser. Sovrem. Mat. Prilozh. Temat. Obz.",
  ["itogi nauki. yug rossii. matematicheskaya monografiya"] = "Itogi Nauki. Yug Ross. Ser. Mat. Monogr.",
  ["itu journal : ict discoveries"] = "ITU J (Geneva)",
  ["iubmb life"] = "IUBMB Life",
  ["iucr monographs on crystallography"] = "IUCr Monogr. Crystallogr.",
  ["iucr texts on crystallography"] = "IUCr Texts Crystallogr.",
  ["iugoslavica physiologica et pharmacologica acta"] = "Iugosl Physiol Pharmacol Acta",
  ["iui. international conference on intelligent user interfaces"] = "IUI",
  ["iup journal of science and technology"] = "IUP J Sci Technol",
  ["iura. rivista internazionale di diritto romano e antico"] = "Jura",
  ["iussp papers"] = "IUSSP Pap",
  ["iustitia (bloomington, ind.)"] = "Iustitia",
  ["iutam bookseries"] = "IUTAM Bookser.",
  ["ivd technology : for in vitro diagnostics development and manufacturing"] = "IVD Technol",
  ["ivra; rivista internazionale di diritto romano e antico"] = "Iura",
  ["ivy journal of ethics"] = "Ivy J Ethics",
  ["iwanami series in modern mathematics"] = "Iwanami Ser. Mod. Math.",
  ["iwate university"] = "Annual Rep. Fac. Ed. Iwate Univ.",
  ["iyakuhin kenkyu"] = "Iyakuhin Kenkyu",
  ["iyo denshi to seitai kogaku"] = "Iyodenshi To Seitai Kogaku",
  ["iyo denshi to seitai kogaku. japanese journal of medical electronics and biological engineering"] = "Iyodenshi To Seitai Kogaku",
  ["iyo kizai kenkyujo hokoku. reports of the institute for medical and dental engineering, tokyo medical and dental university"] = "Tokyo Ika Shika Daigaku Iyo Kizai Kenkyusho Hokoku",
  ["iyō denshi to seitai kōgaku. japanese journal of medical electronics and biological engineering"] = "Iyodenshi To Seitai Kogaku",
  ["iyō kizai kenkyūjo hōkoku. reports of the institute for medical and dental engineering, tokyo medical and dental university"] = "Tokyo Ika Shika Daigaku Iyo Kizai Kenkyusho Hokoku",
  ["iza journal of labor economics"] = "IZA J Labor Econ",
  ["iza journal of labor policy"] = "IZA J Labor Policy",
  ["iza world of labor : evidence-based policy making"] = "IZA World Labor",
  ["izdanja biblioteke"] = "Izdan. Bibl.",
  ["izmeritel’naya tekhnika"] = "Izmer. Tekh.",
  ["izvestiia akademii nauk gruzinskoi ssr. seriia biologicheskaia"] = "Izv Akad Gruz Ssr. Ser Biol",
  ["izvestiia akademii nauk gruzinskoĭ ssr. seriia biologicheskaia"] = "Izv Akad Gruz Ssr. Ser Biol",
  ["izvestiia akademii nauk kirgizskoi ssr. seriia biologicheskikh nauk"] = "Izv Akad Nauk Kirg Ssr [Biol]",
  ["izvestiia akademii nauk kirgizskoĭ ssr. seriia biologicheskikh nauk"] = "Izv Akad Nauk Kirg Ssr Biol",
  ["izvestiia akademii nauk sssr, otdelenie khimicheskikh nauk"] = "Izv Akad Nauk Sssr Otdelenie Khim Nauk",
  ["izvestiia akademii nauk sssr. seriia biologicheskaia"] = "Izv Akad Nauk SSSR Biol",
  ["izvestiia akademii nauk sssr. seriia geograficheskaia"] = "Izv Akad Nauk SSSR Ser Geogr",
  ["izvestiia akademii nauk sssr. seriia khimicheskaia"] = "Izv Akad Nauk SSSR Ser Khim",
  ["izvestiia akademii nauk. seriia biologicheskaia"] = "Izv Akad Nauk Ser Biol",
  ["izvestiia akademii nauk. seriia biologicheskaia / rossiiskaia akademiia nauk"] = "Izv Akad Nauk Ser Biol",
  ["izvestiia na instituta po fiziologiia"] = "Izv Inst Fiziol (Sofiia)",
  ["izvestiia na instituta po istoriia na bkp"] = "Izv Inst Istor BKP",
  ["izvestiia na meditsinskite instituti. [bulletin des instituts de medecine]. bulgarska akademiia na naukite, sofia. otdelenie za biologicheski i meditsinski nauki"] = "Izv Meditsinskite Inst Bulg Akad Naukite Sofia Otd Biol Meditsinski Nauki",
  ["izvestiia na meditsinskite instituti. [bulletin des instituts de médecine]. bŭlgarska akademii︠a︡ na naukite. otdelenie za biologicheski i medit︠s︡inski nauki"] = "Izv Meditsinskite Inst Bulg Akad Naukite Sofia Otd Biol Meditsinski Nauki",
  ["izvestiia na mikrobiologicheskiia institut"] = "Izv Mikrobiol Inst (Sofiia)",
  ["izvestiia russkogo geograficheskogo obshchestva (1992)"] = "Izv Russ Geogr Obshchestva",
  ["izvestiia vsesoiuznogo geograficheskogo obshchestva. geograficheskoe obshchestvo sssr"] = "Izv Vses Geogr Obshchestva",
  ["izvestiia. bulgarska akademiia na naukite, sofia. nauchnoizsledovatelski institut za protivorakovi antibiotitsi"] = "Izv Bulg Akad Naukite Sofia Nauchnoizsledovatelski Inst protivorak antibiot",
  ["izvestiia. bŭlgarska akademiia na naukite, sofia. institut za sotsialna meditsina"] = "Izv Bulg Akad Naukite Sofia",
  ["izvestiia. nauchnoizsledovatelski institut za protivorakovi antibiotit︠s︡i (bulgarska akademii︠a︡ na naukite)"] = "Izv Bulg Akad Naukite Sofia Nauchnoizsledovatelski Inst protivorak antibiot",
  ["izvestiia. seriia fiziologii i meditsiny. qazaq ssr ghylym akademiiasy"] = "Izv Seriia Fiziol Meditsiny Qazaq SSR Ghylum Akad",
  ["izvestii︠a︡ akademii nauk sssr. serii︠a︡ geograficheskai︠a︡"] = "Izv Akad Nauk SSSR Ser Geogr",
  ["izvestii︠a︡ na instituta po istorii︠a︡ na bkp"] = "Izv Inst Istor BKP",
  ["izvestii︠a︡ russkogo geograficheskogo obshchestva (1992)"] = "Izv Russ Geogr Obshchestva",
  ["izvestii︠a︡ samarskogo nauchnogo t︠s︡entra rossiĭskoĭ akademii nauk"] = "Izv Samar Naucn Centra Ross Akad Nauk",
  ["izvestii︠a︡ saratovskogo universiteta. novai︠a︡ serii︠a︡. serii︠a︡ khimii︠a︡. biologii︠a︡. ėkologii︠a︡"] = "Isv Sarat Univ Nov Ser Ser Him Biol Ekol",
  ["izvestii︠a︡ timiri︠a︡zevskoĭ selʹskokhozi︠a︡ĭstvennoĭ akademii"] = "Izv Timirjazevsk Selskokhoziaistvennoi Akad",
  ["izvestii︠a︡ vsesoi︠u︡znogo geograficheskogo obshchestva. geograficheskoe obshchestvo soi︠u︡za ssr"] = "Izv Vses Geogr Obshchestva",
  ["izvestija na archeologičeskija institut. bulletin de l’institut d’archéologie"] = "BIBulg",
  ["izvestija na muzeite ot južna bălgarija. bulletin des musées de la bulgarie du sud"] = "IzvMuzJuzBalg",
  ["izvestija na narodnija muzej burgas. bulletin du musée national de bourgas"] = "IzvBurgas",
  ["izvestija na narodnija muzej varna"] = "IzvVarna",
  ["izvestiya akademii nauk gruzii, seriya khimicheskaya"] = "Izv. Akad. Nauk Gruz. SSR, Ser. Khim.",
  ["izvestiya akademii nauk respubliki kazakhstan, seriya khimicheskaya"] = "Izv. Akad. Nauk Kaz. SSR, Ser. Khim.",
  ["izvestiya akademii nauk sssr, fizika atmosfery i okeana [izvestiya, academy of sciences, ussr, atmospheric and oceanic physics]"] = "Izv. Acad. Nauk SSSR, Fiz. Atmos. Okeana [Izv. Acad. Sci. USSR, Atmos. Oceanic Phys.]",
  ["izvestiya akademii nauk sssr, fizika zemli [izvestiya, academy of sciences, ussr, physics of the solid earth]"] = "Izv. Acad. Nauk SSSR, Fiz. Zemli. [Izv. Acad. Sci. USSR, Phys. Solid Earth]",
  ["izvestiya akademii nauk sssr, neorganicheskie materialy [inorganic materials (ussr)]"] = "Izv. Akad. Nauk SSSR, Neorg. Mater. [Inorg. Mater. (USSR)]",
  ["izvestiya akademii nauk sssr, seriya fizicheskaya [bulletin of the academy of sciences of the ussr, physical series]"] = "Izv. Acad. Nauk SSSR, Ser. Fiz. [Bull. Acad. Sci. USSR, Phys. Ser.]",
  ["izvestiya akademii nauk, seriya fizicheskaya"] = "Izv. Akad. Nauk, Ser. Fiz.",
  ["izvestiya akademii nauk, seriya geologicheskaya"] = "Izv. Akad. Nauk SSSR, Ser. Geol.",
  ["izvestiya atmospheric and oceanic physics"] = "Izv. Atmos. Oceanic Phys.",
  ["izvestiya chelyabinskogo nauchnogo tsentra. proceedings of the chelyabinsk scientific center"] = "Izv. Chelyabinsk. Nauchn. Tsentra",
  ["izvestiya instituta matematiki i informatiki"] = "Izv. Inst. Mat. Inform.",
  ["izvestiya irkutskogo gosudarstvennogo universiteta. seriya matematika"] = "Izv. Irkutsk. Gos. Univ. Ser. Mat.",
  ["izvestiya ministerstva nauki–akademii nauk respubliki kazakhstan"] = "Izv. Minister. Nauki Akad. Nauk Resp. Kaz. Ser. Fiz.-Mat.",
  ["izvestiya natsional\\cprime noy akademii nauk armenii"] = "Izv. Nats. Akad. Nauk Armen. Mekh.",
  ["izvestiya natsionalnoi akademii nauk armenii. mekhanika. proceedings of national academy of sciences of armenia. mechanics"] = "Izv. Nats. Akad. Nauk Armen. Mekh.",
  ["izvestiya natsionalnoi akademii nauk respubliki kazakhstan. seriya fiziko-matematicheskaya"] = "Izv. Nats. Akad. Nauk Resp. Kaz. Ser. Fiz.-Mat.",
  ["izvestiya natsional’noi akademii nauk respubliki kazakhstan, seriya khimicheskaya"] = "Izv. Nats. Akad. Nauk Resp. Kaz., Ser. Khim.",
  ["izvestiya natsional’noi akademii nauk respubliki kazakhstan, seriya khimii i tekhnologii"] = "Izv. Nats. Akad. Nauk Resp. Kaz., Ser. Khim. Tekhnol.",
  ["izvestiya physics of the solid earth"] = "Izv. Phys. Solid Earth",
  ["izvestiya rossiiskoi akademii nauk, seriya fizicheskaya"] = "Izv. Ross. Akad. Nauk, Ser. Fiz.",
  ["izvestiya rossiiskoi akademii nauk. seriya matematicheskaya"] = "Izv. Ross. Akad. Nauk Ser. Mat.",
  ["izvestiya rossiiskoi akademii nauk. teoriya i sistemy upravleniya"] = "Izv. Ross. Akad. Nauk Teor. Sist. Upr.",
  ["izvestiya saratovskogo universiteta. novaya seriya. seriya. matematika. mekhanika. informatika"] = "Izv. Sarat. Univ. (N.S.) Ser. Mat. Mekh. Inform.",
  ["izvestiya vysshikh uchebnykh zavedenii, chernaya metallurgiya"] = "Izv. Vyssh. Uchebn. Zaved., Chern. Metall.",
  ["izvestiya vysshikh uchebnykh zavedenii, fizika"] = "Izv. Vyssh. Uchebn. Zaved., Fiz.",
  ["izvestiya vysshikh uchebnykh zavedenii, fizika [soviet physics journal]"] = "Izv. Vyssh. Uchebn. Zavad. Fiz. [Sov. Phys. J.]",
  ["izvestiya vysshikh uchebnykh zavedenii, khimiya i khimicheskaya tekhnologiya"] = "Izv. Vyssh. Uchebn. Zaved., Khim. Khim. Tekhnol.",
  ["izvestiya vysshikh uchebnykh zavedenii, radiofizika [soviet radiophysics]"] = "Izv. Vyssh. Uchebn. Zaved. Radiofiz. [Sov. Radiophys.]",
  ["izvestiya vysshikh uchebnykh zavedenii, tsvetnaya metallurgiya"] = "Izv. Vyssh. Uchebn. Zaved., Tsvetn. Metall.",
  ["izvestiya vysshikh uchebnykh zavedenii. matematika. kazanskii gosudarstvennyi universitet"] = "Izv. Vyssh. Uchebn. Zaved. Mat.",
  ["izvestiya vysshikh uchebnykh zavedeniy"] = "Izv. Vyssh. Uchebn. Zaved. Fiz.",
  ["izvestiya, academy of sciences, ussr, atmospheric"] = "Izv. Acad. Sci. USSR, Atmos. Oceanic Phys.",
  ["izvestiya, academy of sciences, ussr, atmospheric and oceanic physics [translation of izvestiya akademii nauk sssr, fizika atmosfery i okeana]"] = "Izv. Acad. Sci. USSR, Atmos. Oceanic Phys.",
  ["izvestiya, academy of sciences, ussr, physics of the solid earth [translation of izvestiya akademii nauk sssr, fizika zemli]"] = "Izv. Acad. Sci. USSR, Phys. Solid Earth",
  ["izvestiya, atmospheric and oceanic physics"] = "Izv. Atmos. Ocean. Phys.",
  ["izvestiya. mathematics"] = "Izv. Math.",
  ["izvestiya: mathematics"] = "Izv. Math.",
  ["i̇stanbul üniversitesi dişhekimliği fakültesi dergisi = the journal of the dental faculty of istanbul"] = "Istanbul Univ Dishekim Fak Derg",
  ["i︠u︡g rossii - ėkologii︠a︡, razvitie"] = "Ug Ross Ekol Razvit",
  ["j journal of agricultural safety and health"] = "J. Agric. Saf. Health",
  ["j-for journal of science & technology for forest products and processes"] = "J-FOR",
  ["j. d. beazley, attic black-figure vase-painters (oxford 1956)"] = "ABV",
  ["j. d. beazley, attic red-figure vase-painters \\textsuperscript{2}(oxford 1963)"] = "ARV2",
  ["j. d. beazley, etruscan vase painting (oxford 1947)"] = "EVP",
  ["j. d. beazley, paralipomena. additions to attic black-figure vase-painters and to attic red-figure vase-painters (oxford 1971)"] = "BeazleyPara.",
  ["j.ucs. journal of universal computer science"] = "J.UCS",
  ["ja clinical reports"] = "JA Clin Rep",
  ["jaad case reports"] = "JAAD Case Rep",
  ["jaad international"] = "JAAD Int",
  ["jaakko hintikka selected papers"] = "Jaakko Hintikka Sel. Pap.",
  ["jaami : journal of the association for the advancement of medical instrumentation"] = "J Assoc Adv Med Instrum",
  ["jaami; journal of the association for the advancement of medical instrumentation"] = "J. Assoc. Adv. Med. Instrum.",
  ["jaapa : official journal of the american academy of physician assistants"] = "JAAPA",
  ["jaarbericht van het voor-aziatisch-egyptisch genootschap ex oriente lux"] = "JVEG",
  ["jaarbericht van het vooraziatisch-egyptisch genootschap ex oriente lux"] = "JEOL",
  ["jaarboek numaga : gewijd aan heden en verleden van nijmegen en omgeving"] = "Jaarb Numaga",
  ["jaarboek van kankeronderzoek en kankerbestrijding in nederland"] = "Jaarb. Kankeronderz. Kankerbestrijd. Ned.",
  ["jaarboek van kankeronderzoek en kankerbestrijding in nederland. yearbook for cancer research and fight against cancer in the netherlands"] = "Jaarb Kankeronderz Kankerbestrijd Ned",
  ["jabega : revista de la diputacion provincial de malaga"] = "Jabega",
  ["jabs : journal of applied biological sciences"] = "JABS",
  ["jac-antimicrobial resistance"] = "JAC Antimicrob Resist",
  ["jacc. basic to translational science"] = "JACC Basic Transl Sci",
  ["jacc. cardiooncology"] = "JACC CardioOncol",
  ["jacc. cardiovascular imaging"] = "JACC Cardiovasc Imaging",
  ["jacc. cardiovascular interventions"] = "JACC Cardiovasc Interv",
  ["jacc. case reports"] = "JACC Case Rep",
  ["jacc. clinical electrophysiology"] = "JACC Clin Electrophysiol",
  ["jacc. heart failure"] = "JACC Heart Fail",
  ["jacc: advances"] = "JACC: Adv.",
  ["jacc: asia"] = "JACC: Asia",
  ["jacc: basic to translational science"] = "JACC: Basic Transl. Sci.",
  ["jacc: cardiooncology"] = "JACC: CardioOncol.",
  ["jacc: cardiovascular imaging"] = "JACC: Cardiovasc. Imaging",
  ["jacc: cardiovascular interventions"] = "JACC: Cardiovasc. Interventions",
  ["jacc: case reports"] = "JACC: Case Rep.",
  ["jacc: clinical electrophysiology"] = "JACC: Clin. Electrophysiol.",
  ["jacc: heart failure"] = "JACC: Heart Failure",
  ["jacobs journal of addiction and therapy"] = "Jacobs J Addict Ther",
  ["jacobs journal of aids/hiv"] = "Jacobs J AIDS HIV",
  ["jacobs journal of allergy and immunology"] = "Jacobs J Allergy Immunol",
  ["jacobs journal of biotechnology and bioengineering"] = "Jacobs J Biotechnol Bioeng",
  ["jacobs journal of community medicine"] = "Jacobs J Community Med",
  ["jacobs journal of diabetes and endocrinology"] = "Jacobs J Diabetes Endocrinol",
  ["jacobs journal of environmental sciences"] = "Jacobs J Environ Sci",
  ["jacobs journal of food and nutrition"] = "Jacobs J Food Nutr",
  ["jacobs journal of genetics"] = "Jacobs J Genet",
  ["jacobs journal of gerontology"] = "Jacobs J Gerontol",
  ["jacobs journal of medical diagnosis and medical imaging"] = "Jacobs J Med Diagn Med Imaging",
  ["jacobs journal of microbiology and pathology"] = "Jacobs J Microbiol Pathol",
  ["jacobs journal of molecular and translational medicine"] = "Jacobs J Mol Transl Med",
  ["jacobs journal of ophthalmology"] = "Jacobs J Ophthalmol",
  ["jacobs journal of pediatrics"] = "Jacobs J Pediatr",
  ["jacobs journal of psychiatry and behavioral science"] = "Jacobs J Psychiatry Behav Sci",
  ["jacobs journal of pulmonology"] = "Jacobs J Pulmonol",
  ["jacobs journal of radiation oncology"] = "Jacobs J Radiat Oncol",
  ["jacs au"] = "JACS Au",
  ["jadara : a journal for professionals networking for excellence in service delivery with individuals who are deaf or hard of hearing"] = "JADARA",
  ["jadranski zbornik. prilozi za povijest istre, rijeke i hrvatskog primorja"] = "JadrZbor",
  ["jaen journal on approximation"] = "Jaen J. Approx.",
  ["jagd und natur"] = "Jagd Nat.",
  ["jagellonian university"] = "Acta Phys. Polon. B",
  ["jagellonian university. institute of physics. acta physica polonica b"] = "Acta Phys. Polon. B",
  ["jahrbuch der akademie der wissenschaften in göttingen"] = "JbGoett",
  ["jahrbuch der albertus-universitat zu konigsberg/pr"] = "Jahrb Albertus Univ Koenigsb",
  ["jahrbuch der albertus-universität zu königsberg/pr"] = "Jahrb Albertus Univ Koenigsb",
  ["jahrbuch der auktionspreise für bücher, handschriften und autographen; ergebnisse der auktionen in deutschland, holland, oesterreich und der schweiz"] = "Jahrb Auktionspreise Bucher",
  ["jahrbuch der berliner museen"] = "JBerlM",
  ["jahrbuch der coburger landesstiftung. coburger landesstiftung"] = "Jahrb Coburg Landesstift",
  ["jahrbuch der deutschen luftfahrtforschung"] = "Jahrb. Dtsch. Luftfahrtforsch.",
  ["jahrbuch der geologischen bundesanstalt"] = "Jahrb. Geol. Bundesanst.",
  ["jahrbuch der geologischen bundesanstalt (vienna, austria : 1945)"] = "Jahrb Geol Bundesanst (1945)",
  ["jahrbuch der hamburger kunstsammlungen"] = "JbHambKuSamml",
  ["jahrbuch der heidelberger akad. der wissenschaften"] = "JHAW",
  ["jahrbuch der historischen forschung in der bundesrepublik deutschland"] = "Jahrb Hist Forsch Bundesrepub Dtschl",
  ["jahrbuch der kunsthistorischen sammlungen in wien"] = "JbKHSWien",
  ["jahrbuch der preussischen geologischen landesanstalt"] = "Jahrb. Preuss. Geol. Landesanst.",
  ["jahrbuch der schlesischen friedrich-wilhelms-universitat zu breslau"] = "Jahrb Schlesisch Friedrich Wilhelms Univ Breslau",
  ["jahrbuch der schlesischen friedrich-wilhelms-universität zu breslau"] = "Jahrb Schlesisch Friedrich Wilhelms Univ Breslau",
  ["jahrbuch der schweizerischen gesellschaft für ur- und frühgeschichte"] = "JSGU",
  ["jahrbuch der schweizerischen naturforschenden gesellschaft"] = "Jahrb. Schweiz. nat.forsch. Ges.",
  ["jahrbuch der schweizerischen wald- und holzwirtschaft"] = "Jahrb. schweiz. Wald- Holzwirtsch.",
  ["jahrbuch der st. gallischen naturwissenschaftlichen gesellschaft"] = "Jahrb. St. Gallische Nat.wiss. Ges.",
  ["jahrbuch der staatlichen kunstsammlungen in baden-württemberg"] = "JbBadWuert",
  ["jahrbuch der universitat dusseldorf"] = "Jahrb Univ Duesseld",
  ["jahrbuch der universität düsseldorf"] = "Jahrb Univ Duesseld",
  ["jahrbuch der versuchs und lehranstalt fur brauerei in berlin"] = "Jahrb. Vers. Lehranst. Brau. Berlin",
  ["jahrbuch der wirtschaft osteuropas"] = "Jahr. Wirtsch. Osteuropas",
  ["jahrbuch der österreichischen byzantinistik"] = "JöByz",
  ["jahrbuch des bernischen historischen museums in bern"] = "JbBernHistMus",
  ["jahrbuch des deutschen archäologischen instituts"] = "JDAI",
  ["jahrbuch des historischen vereins fur das furstentum liechtenstein. historischer verein fur das furstentum liechtenstein"] = "Jahrb Hist Ver Furst Liechtenstein",
  ["jahrbuch des historischen vereins für das fürstentum liechtenstein. historischer verein für das fürstentum liechtenstein"] = "Jahrb Hist Ver Furst Liechtenstein",
  ["jahrbuch des instituts fur deutsche geschichte. universitat tel-aviv. makhon le-historyah germanit"] = "Jahrb Inst Dtsch Gesch",
  ["jahrbuch des instituts fur geschichte der medizin der robert bosch stiftung"] = "Jahrb Inst Gesch Med Robert Bosch Stift",
  ["jahrbuch des instituts für deutsche geschichte. universiṭat tel-aviv. makhon le-hisṭoryah germanit"] = "Jahrb Inst Dtsch Gesch",
  ["jahrbuch des instituts für geschichte der medizin der robert bosch stiftung"] = "Jahrb Inst Gesch Med Robert Bosch Stift",
  ["jahrbuch des kunsthistorischen museums wien"] = "JbKHMWien",
  ["jahrbuch des museums für kunst und gewerbe, hamburg"] = "JbMusKGHamb",
  ["jahrbuch des oberaargaus"] = "Jahrb. Oberaargau",
  ["jahrbuch des oberosterreichischen musealvereines. oberosterreichischer musealverein"] = "Jahrb Oberoesterr Musealver",
  ["jahrbuch des oberösterreichischen musealvereines. oberösterreichischer musealverein"] = "Jahrb Oberoesterr Musealver",
  ["jahrbuch des oberösterreichischen musealvereins"] = "JbMusLinz",
  ["jahrbuch des römisch-germanischen zentralmuseums"] = "JRGZ",
  ["jahrbuch des römisch-germanischen zentralmuseums mainz"] = "JbRGZM",
  ["jahrbuch des schweizerischen wasserwirtschaftsverbandes"] = "Jahrb. Schweiz. Wasserwirtschafts-verb.",
  ["jahrbuch des vereins fur geschichte der stadt wien. verein fur geschichte der stadt wien"] = "Jahrb Vereins Gesch Stadt Wien",
  ["jahrbuch des vereins für geschichte der stadt wien. verein für geschichte der stadt wien"] = "Jahrb Vereins Gesch Stadt Wien",
  ["jahrbuch des vereins zum schutze der alpenpflanzen und -tiere"] = "Jahrb. Ver. Schutz Alp.pflanzen -Tiere",
  ["jahrbuch fur antike und christentum"] = "Jahrb Antike Christentum",
  ["jahrbuch fur brandenburgische landesgeschichte"] = "Jahrb Brandenbg Landesgesch",
  ["jahrbuch fur geschichte"] = "Jahrb Gesch",
  ["jahrbuch fur geschichte von staat, wirtschaft und gesellschaft lateinamerikas"] = "Jahrb Gesch Staat Wirtsch Ges Lateinam",
  ["jahrbuch fur internationales recht. german yearbook of international law"] = "Ger Yearb Int Law",
  ["jahrbuch fur landeskunde von niederosterreich"] = "Jahrb Landeskd Niederosterr",
  ["jahrbuch fur sozialwissenschaft"] = "Jahrb Sozialwiss",
  ["jahrbuch fur wirtschaftsgeschichte"] = "Jahrb Wirtschaftsgesch",
  ["jahrbuch für antike und christentum"] = "Jahrb Antike Christentum",
  ["jahrbuch für brandenburgische landesgeschichte"] = "Jahrb Brandenbg Landesgesch",
  ["jahrbuch für geschichte"] = "Jahrb Gesch",
  ["jahrbuch für geschichte lateinamerikas"] = "Jahrb Gesch Lat Am",
  ["jahrbuch für geschichte von staat, wirtschaft und gesellschaft lateinamerikas"] = "Jahrb Gesch Staat Wirtsch Ges Lateinam",
  ["jahrbuch für internationales recht. german yearbook of international law"] = "Ger Yearb Int Law",
  ["jahrbuch für landeskunde von niederösterreich"] = "Jahrb Landeskd Niederosterr",
  ["jahrbuch für sozialwissenschaft"] = "Jahrb Sozialwiss",
  ["jahrbuch für wirtschaftsgeschichte"] = "Jahrb Wirtschaftsgesch",
  ["jahrbuch für wissenschaft und ethik"] = "Jahr Wiss Ethik",
  ["jahrbuch für antike und christentum"] = "JbAC",
  ["jahrbuch für kleinasiatische forschung"] = "JbKleinasF",
  ["jahrbuch für naturschutz und landschaftspflege"] = "Jahrb. Nat.schutz Landsch.pfl.",
  ["jahrbuch für numismatik und geldgeschichte"] = "JNG",
  ["jahrbuch für prähistorische und ethnographische kunst"] = "Ipek",
  ["jahrbuch für regionalwissenschaft"] = "Jahr. Regionalwissen.",
  ["jahrbuch für wirtschaftsgeschichte"] = "JWG",
  ["jahrbuch max planck gesellschaft"] = "Jahrb. Max Planck Gesell.",
  ["jahrbuch oberflachentechnik"] = "Jahrb. Oberflachentech.",
  ["jahrbuch preussischer kulturbesitz"] = "JbPreussKul",
  ["jahrbuch schweizerische akademie der naturwissenschaften"] = "Jahrb. Schweiz. Akad. Nat.wiss.",
  ["jahrbuch volksgesundheit"] = "Jahrb Volksgesundh",
  ["jahrbuch vom thuner- und brienzersee"] = "Jahrb. Thuner- Brienzersee",
  ["jahrbuch vom zürichsee"] = "Jahrb. Zür.see",
  ["jahrbuch. akademie der wissenschaften und der literatur, mainz"] = "JbAkMainz",
  ["jahrbucher fur geschichte osteuropas"] = "Jahrb Gesch Osteur",
  ["jahrbucher fur nationalokonomie und statistik"] = "Jahrb Natl Okon Stat",
  ["jahrbucher fur wissenschaftliche botanik"] = "Jahrb. Wiss. Bot.",
  ["jahrbücher für geschichte osteuropas"] = "Jahrb Gesch Osteur",
  ["jahrbücher für nationalökonomie und statistik"] = "Jahrb Natl Okon Stat",
  ["jahrbücher für nationalökonomie und statistik"] = "Jahr. Nationalökon. Statist.",
  ["jahresbericht (schweizerische akademie der medizinischen wissenschaften : 1988)"] = "Jahresber Schweiz Akad Med Wiss",
  ["jahresbericht / schweizerische akademie der medizinischen wissenschaften"] = "Jahresber Schweiz Akad Med Wiss",
  ["jahresbericht der archäologischen bodenforschung des kantons basel-stadt"] = "JberBasel",
  ["jahresbericht der bayerischen bodendenkmalpflege"] = "JberBayDenkmPfl",
  ["jahresbericht der deutschen mathematiker-vereinigung"] = "Jahresber. Dtsch. Math.-Ver.",
  ["jahresbericht der naturforschenden gesellschaft graubünden"] = "Jahresber. Nat.forsch. Ges. Graubünden",
  ["jahresbericht des instituts für vorgeschichte der universität frankfurt a. m."] = "JberVgFrankf",
  ["jahresbericht lignum"] = "Jahresber. Lignum",
  ["jahresbericht, schweizerische akademie der medizinischen wissenschaften"] = "Jahresber. Schweiz. Akad. Med. Wiss.",
  ["jahresbericht. gesellschaft pro vindonissa"] = "JberProVindon",
  ["jahresbericht. schweizerisches landesmuseum zürich"] = "JberZuerich",
  ["jahresberichte aus augst und kaiseraugst"] = "JAK",
  ["jahresberichte des literarischen zentralblattes fur deutschland mathematik astronomie"] = "Jahresber. Lit. Zentralbl. Dtschl. Math. Astron.",
  ["jahreshefte der gesellschaft für naturkunde in württemberg. gesellschaft für naturkunde in württemberg"] = "Jahresh Ges Naturkd Wurtt",
  ["jahreshefte des vereins für vaterländische naturkunde in württemberg"] = "Jahresh. Ver. vaterl. Nat.kd. Württ.",
  ["jahreshefte des österreichischen archäologischen institutes in wien"] = "OeJh",
  ["jahreshefte des österreichischen archäologischen instituts"] = "JöAI",
  ["jahresschrift für mitteldeutsche vorgeschichte"] = "JSchrVgHalle",
  ["jala (charlottesville, va.)"] = "JALA Charlottesv Va",
  ["jama : the journal of the american medical association"] = "JAMA",
  ["jama cardiology"] = "JAMA Cardiol",
  ["jama dermatology"] = "JAMA Dermatol",
  ["jama facial plastic surgery"] = "JAMA Facial Plast Surg",
  ["jama health forum"] = "JAMA Health Forum",
  ["jama internal medicine"] = "JAMA Intern Med",
  ["jama network open"] = "JAMA Netw Open",
  ["jama neurology"] = "JAMA Neurol",
  ["jama oncology"] = "JAMA Oncol",
  ["jama ophthalmology"] = "JAMA Ophthalmol",
  ["jama otolaryngology-- head & neck surgery"] = "JAMA Otolaryngol Head Neck Surg",
  ["jama pediatrics"] = "JAMA Pediatr",
  ["jama psychiatry"] = "JAMA Psychiatry",
  ["jama surgery"] = "JAMA Surg",
  ["jama the journal of the american medical association"] = "JAMA J. Am. Med. Assoc.",
  ["jama, the journal of the american medical association"] = "JAMA, J. Am. Med. Assoc.",
  ["jamaica journal"] = "Jam J",
  ["jamaica public health"] = "Jam Public Health",
  ["jamaican nurse"] = "Jamaican Nurse",
  ["james arthur lecture on the evolution of the human brain"] = "James Arthur Lect",
  ["james joyce quarterly"] = "James Joyce Q",
  ["jamia open"] = "JAMIA Open",
  ["janasaṃkhyā śikshā mukhapatra"] = "Janasamkhya Siksha Mukhapatra",
  ["janus; revue internationale de l'histoire des sciences, de la médecine, de la pharmacie, et de la technique"] = "Janus",
  ["janus; revue internationale de l’histoire des sciences, de la medecine, de la pharmacie, et de la technique"] = "Janus",
  ["japan academy. proceedings. series a. mathematical sciences"] = "Proc. Japan Acad. Ser. A Math. Sci.",
  ["japan agricultural research quarterly"] = "Jpn Agric Res Q",
  ["japan and the world economy"] = "Japan World Econ",
  ["japan annual reviews in electronics, computers & telecommunications 1982"] = "Japan Annual Rev. Electron. Comput. Telecom. 1982",
  ["japan journal of industral and applied mathematics"] = "Jpn. J. Ind. Appl. Math.",
  ["japan journal of industrial and applied mathematics"] = "Jpn. J. Ind. Appl. Math.",
  ["japan journal of medicine"] = "Jpn J Med (Lond)",
  ["japan journal of nursing science : jjns"] = "Jpn J Nurs Sci",
  ["japan medical association journal : jmaj"] = "Japan Med Assoc J",
  ["japan quarterly (asahi shinbunsha)"] = "Jpn Q",
  ["japan research quarterly"] = "Japan Res. Quart.",
  ["japan society of fluid mechanics"] = "Fluid Dynam. Res.",
  ["japan society of mechanical engineers international journal series a: solid mechanics and material engineering"] = "J. Soc. Mech. Eng. Int. J. Ser.A:",
  ["japan society of mechanical engineers international journal series b: fluids and thermal engineering"] = "J. Soc. Mech. Eng. Int. J. Ser.B:",
  ["japan society of mechanical engineers international journal series c: mechanical systems, machine elements and manufacturing"] = "J. Soc. Mech. Eng. Int. J. Ser.C:",
  ["japan-hospitals : the journal of the japan hospital association"] = "Jpn Hosp",
  ["japanese circulation journal"] = "Jpn Circ J",
  ["japanese clinical medicine"] = "Jpn Clin Med",
  ["japanese dental journal"] = "Jpn Dent J",
  ["japanese economic review"] = "Japanese Econ. Rev.",
  ["japanese economic review (oxford, england)"] = "Jpn Econ Rev (Oxf)",
  ["japanese economic studies"] = "Japanese Econ Stud",
  ["japanese economy"] = "Japanese Economy",
  ["japanese heart journal"] = "Jpn Heart J",
  ["japanese journal of antibiotics"] = "Jpn. J. Antibiot.",
  ["japanese journal of applied entomology and zoology"] = "Jpn. J. Appl. Entomol. Zool.",
  ["japanese journal of applied physics"] = "Jpn. J. Appl. Phys.",
  ["japanese journal of applied physics (2008)"] = "Jpn J Appl Phys (2008)",
  ["japanese journal of applied physics part 1"] = "Jpn. J. Appl. Phys., Part 1",
  ["japanese journal of applied physics part 2"] = "Jpn. J. Appl. Phys., Part 2",
  ["japanese journal of applied physics, part 1"] = "Jpn. J. Appl. Phys., Part 1",
  ["japanese journal of applied physics, part 1: regular papers, brief communications & review papers"] = "Jpn. J. Appl. Phys., Part 1",
  ["japanese journal of applied physics, part 2"] = "Jpn. J. Appl. Phys., Part 2",
  ["japanese journal of applied physics, part 2: letters & express letters"] = "Jpn. J. Appl. Phys., Part 2",
  ["japanese journal of applied physics, part 2: letters and express letters"] = "Jpn. J. Appl. Phys. Part 2 Lett. Express Lett.",
  ["japanese journal of applied physics, part i: regular papers and short notes"] = "Jpn. J. Appl. Phys., Part 1",
  ["japanese journal of applied physics, part ii: letters"] = "Jpn. J. Appl. Phys., Part 2",
  ["japanese journal of applied physics. part 1, regular papers & short notes"] = "Jpn J Appl Phys Pt 1",
  ["japanese journal of cancer and chemotherapy"] = "Jpn. J. Cancer Chemother.",
  ["japanese journal of cancer research"] = "Jpn. J. Cancer Res.",
  ["japanese journal of cancer research : gann"] = "Jpn J Cancer Res",
  ["japanese journal of chemotherapy"] = "Jpn. J. Chemother.",
  ["japanese journal of clinical chemistry"] = "Jpn. J. Clin. Chem.",
  ["japanese journal of clinical oncology"] = "Jpn J Clin Oncol",
  ["japanese journal of clinical pharmacology and therapeutics"] = "Jpn. J. Clin. Pharmacol. Ther.",
  ["japanese journal of crop science"] = "Jpn. J. Crop Sci.",
  ["japanese journal of dermatology: series b"] = "Jpn. J. Dermatol. B",
  ["japanese journal of ecology"] = "Jpn. J. Ecol.",
  ["japanese journal of electrophoresis"] = "Jpn J Electrophor",
  ["japanese journal of electrophoresis : the official journal of the japanese electrophoresis society"] = "Jpn J Electrophor",
  ["japanese journal of experimental medicine"] = "Jpn. J. Exp. Med.",
  ["japanese journal of forensic toxicology"] = "Jpn. J. Forensic Toxicol.",
  ["japanese journal of fuzzy theory and systems"] = "Japanese J. Fuzzy Theory Systems",
  ["japanese journal of gastroenterology"] = "Jpn J Gastroenterol",
  ["japanese journal of genetics"] = "Jpn. J. Genet.",
  ["japanese journal of health physics"] = "Jpn. J. Health Phys.",
  ["japanese journal of helicobacter research"] = "Jpn J Helicobacter Res",
  ["japanese journal of human genetics"] = "Jpn. J. Hum. Genet.",
  ["japanese journal of human geography"] = "Jpn. J. Hum. Geogr.",
  ["japanese journal of infectious diseases"] = "Jpn J Infect Dis",
  ["japanese journal of legal medicine"] = "Jpn. J. Legal Med.",
  ["japanese journal of limnnology"] = "Jpn. J. Limnol.",
  ["japanese journal of mathematics"] = "Jpn. J. Math.",
  ["japanese journal of medical science & biology"] = "Jpn J Med Sci Biol",
  ["japanese journal of medical science and biology"] = "Jpn. J. Med. Sci. Biol.",
  ["japanese journal of medicine"] = "Jpn J Med",
  ["japanese journal of microbiology"] = "Jpn J Microbiol",
  ["japanese journal of ophthalmology"] = "Jpn J Ophthalmol",
  ["japanese journal of pharmacology"] = "Jpn J Pharmacol",
  ["japanese journal of phycology"] = "Jpn. J. Phycol.",
  ["japanese journal of physical fitness and sports medicine"] = "Jpn. J. Sports Med. Phys. Fitness",
  ["japanese journal of physics"] = "Jpn. J. Phys.",
  ["japanese journal of physiology"] = "Jpn. J. Physiol.",
  ["japanese journal of psychiatry and neurology"] = "Jpn. J. Psychiatry Neurol.",
  ["japanese journal of radiology"] = "Jpn J Radiol",
  ["japanese journal of statistics and data science"] = "Jpn. J. Stat. Data Sci.",
  ["japanese journal of surgery"] = "Jpn. J. Surg.",
  ["japanese journal of systematic entomology = nihon konchū bunrui gakkai kaihō"] = "Jpn J Syst Entomol",
  ["japanese journal of thoracic and cardiovascular surgery"] = "Jpn. J. Thorac. Cardiovasc. Surg.",
  ["japanese journal of toxicology and environment health"] = "Jpn. J. Toxicol. Environ. Health",
  ["japanese journal of tuberculosis"] = "Jpn. J. Tuberc.",
  ["japanese journal of tuberculosis and chest diseases"] = "Jpn. J. Tuberc. Chest Dis.",
  ["japanese journal of veterinary research"] = "Jpn. J. Vet. Res.",
  ["japanese medical journal"] = "Jpn Med J (Natl Inst Health Jpn)",
  ["japanese studies in the history of science"] = "Jpn Stud Hist Sci",
  ["jasa express letters"] = "JASA Express Lett",
  ["jawra journal of the american water resources association"] = "JAWRA J. Am. Water Resour. Assoc.",
  ["jàmbá (potchefstroom, south africa)"] = "Jamba",
  ["jábega : revista de la diputación provincial de málaga"] = "Jabega",
  ["jb & js open access"] = "JB JS Open Access",
  ["jbi database of systematic reviews and implementation reports"] = "JBI Database System Rev Implement Rep",
  ["jbi evidence implementation"] = "JBI Evid Implement",
  ["jbi evidence synthesis"] = "JBI Evid Synth",
  ["jbi library of systematic reviews"] = "JBI Libr Syst Rev",
  ["jbic, journal of biological inorganic chemistry"] = "JBIC, J. Biol. Inorg. Chem.",
  ["jbjs case connector"] = "JBJS Case Connect",
  ["jbjs essential surgical techniques"] = "JBJS Essent Surg Tech",
  ["jbjs reviews"] = "JBJS Rev",
  ["jbmr plus"] = "JBMR Plus",
  ["jbr-btr : organe de la société royale belge de radiologie (srbr) = orgaan van de koninklijke belgische vereniging voor radiologie (kbvr)"] = "JBR-BTR",
  ["jbra assisted reproduction"] = "JBRA Assist Reprod",
  ["jbsp : the journal of the british society for phenomenology. british society for phenomenology"] = "JBSP",
  ["jcah perspectives"] = "JCAH Perspect.",
  ["jcah perspectives. joint commission on accreditation of hospitals"] = "JCAH Perspect",
  ["jcca. journal of the canadian chiropractic association. journal de l’association chiropratique canadienne"] = "JCCA J Can Chiropr Assoc",
  ["jcem case reports"] = "JCEM Case Rep.",
  ["jci insight"] = "JCI Insight",
  ["jcis open"] = "JCIS Open",
  ["jco clinical cancer informatics"] = "JCO Clin Cancer Inform",
  ["jco global oncology"] = "JCO Glob Oncol",
  ["jco oncology practice"] = "JCO Oncol Pract",
  ["jco precision oncology"] = "JCO Precis Oncol",
  ["jcp: biochemical physics"] = "JCP: Biochem. Phys.",
  ["jcrs online case reports"] = "JCRS Online Case Rep",
  ["jcsm clinical reports"] = "JCSM Clin Rep",
  ["jcsm rapid communications"] = "JCSM Rapid Commun",
  ["jdr clinical & translational research"] = "JDR Clin. Transl. Res.",
  ["jdr clinical and translational research"] = "JDR Clin Trans Res",
  ["jednota annual furdek"] = "Jednota Annu Furdek",
  ["jems : a journal of emergency medical services"] = "JEMS",
  ["jen min jih pao"] = "Ren Min Ri Bao",
  ["jenaische zeitschrift für medicin und naturwissenschaft"] = "Jena Z Med Naturwiss",
  ["jerusalem studies in philosophy and history of science"] = "Jerus. Stud. Philos. Hist. Sci.",
  ["jesuit studies. modernity through the prism of jesuit history"] = "Jesuit Stud.",
  ["jetp letters"] = "JETP Lett.",
  ["jetp letters [translation of pisma v zhurnal eksperimentalnoi i teoreticheskoi fiziki]"] = "JETP Lett.",
  ["jeune afrique (paris, france : 1980)"] = "Jeune Afr",
  ["jewellery studies"] = "JewelSt",
  ["jewish affairs"] = "Jew Aff",
  ["jewish culture and history"] = "Jew Cult Hist",
  ["jewish historical studies : transactions of the jewish historical society of england"] = "Trans Jew Hist Soc Engl",
  ["jewish journal of sociology"] = "Jew J Sociol",
  ["jewish life (union of orthodox jewish congregations of america : 1975)"] = "Jew Life",
  ["jewish observer and middle east review"] = "Jew Obs Middle East Rev",
  ["jewish quarterly review (philadelphia, pa.)"] = "Jew Q Rev",
  ["jewish social studies"] = "Jew Soc Stud",
  ["jfe journal du four electrique"] = "JFE J. Four Electr.",
  ["jfe journal du four electrique & des industries electrochimiques"] = "JFE J. Four Electr. Ind. Electrochim.",
  ["jfma, the journal of the florida medical association. florida medical association"] = "JFMA",
  ["jfma; journal of the florida medical association"] = "JFMA.",
  ["jfms open reports"] = "JFMS Open Rep",
  ["jforl. journal francais d’oto-rhino-laryngologie; audiophonologie et chirurgie maxillo-faciale"] = "JFORL J Fr Otorhinolaryngol Audiophonol Chir Maxillofac",
  ["jforl. journal français d'oto-rhino-laryngologie; audiophonologie et chirurgie maxillo-faciale"] = "JFORL J Fr Otorhinolaryngol Audiophonol Chir Maxillofac",
  ["jforl; journal francais d’oto-rhino laryngologie, audiophonologie et chirurgie maxillo-faciale"] = "JFORL J. Fr. Otorhinolaryngol. Audiophonol. Chir. Maxillofac.",
  ["jgh open : an open access journal of gastroenterology and hepatology"] = "JGH Open",
  ["jhep reports : innovation in hepatology"] = "JHEP Rep",
  ["ji bing jian ce"] = "Ji Bing Jian Ce",
  ["ji chu yi xue yu lin chuang = jichu yixue yu linchuang = basic medical sciences and clinics"] = "Ji Chu Yi Xue Yu Lin Chuang",
  ["ji di yan jiu = chinese journal of polar research"] = "Ji Di Yan Jiu",
  ["ji guang sheng wu xue bao = acta laser biology sinica = jiguang shengwu xuebao"] = "Ji Guang Sheng Wu Xue Bao",
  ["ji sheng chong xue yu ji sheng chong bing za zhi (journal of parasitology and parasitic diseases)"] = "Ji Sheng Chong Xue Yu Ji Sheng Chong Bing Za Zhi",
  ["ji sheng chong xue yu ji sheng chong bing za zhi = journal of parasitology & parasitic diseases"] = "Ji Sheng Chong Xue Yu Ji Sheng Chong Bing Za Zhi",
  ["ji sheng chong yu yi xue kun chong xue bao"] = "Ji Sheng Chong Yu Yi Xue Kun Chong Xue Bao",
  ["ji shou da xue xue bao. zi ran ke xue ban"] = "Ji Shou Da Xue Xue Bao Zi Ran Ke Xue Ban",
  ["ji yin zu xue yu ying yong sheng wu xue"] = "Ji Yin Zu Xue Yu Ying Yong Sheng Wu Xue",
  ["jian kang bao"] = "Jian Kang Bao",
  ["jiangsu gongye xueyuan xuebao = journal of jiangsu polytechnic university"] = "Jiangsu Gongye Xueyuan Xuebao",
  ["jiangsu nong ye ke xue"] = "Jiangsu nongye kexue / Jiangsu sheng nong ye ke xue yuan bian|Jiangsu Nong Ye Ke Xue",
  ["jiangsu nong ye ke xue = jiangsu nongye kexue"] = "Jiangsu Nong Ye Ke Xue",
  ["jiangsu nong ye xue bao = journal of agricultural sciences"] = "Jiangsu nong ye xue bao",
  ["jiangxi nong ye xue bao"] = "Jiangxi Nong Ye Xue Bao",
  ["jiangxi nongye daxue xuebao = acta agriculturae universitatis jiangxiensis"] = "Jiangxi Nongye Daxue Xuebao",
  ["jiangxi yixueyuan xuebao"] = "Acta academiae medicinae jiangxi|Jiangxi Yixueyuan Xuebao",
  ["jiangxi yixueyuan xuebao = acta academiae medicinae jiangxi"] = "Jiangxi Yixueyuan Xuebao",
  ["jibi inkoka otolaryngology"] = "Jibiinkoka",
  ["jibi inkōka otolaryngology"] = "Jibiinkoka",
  ["jica newsletter"] = "JICA Newsl",
  ["jid innovations"] = "JID Innov",
  ["jie fang jun yi xue za zhi"] = "Jie Fang Jun Yi Xue Za Zhi",
  ["jie fang jun yi xue za zhi. [liberation army medical journal]"] = "Chieh Fang Chun I Hsueh Tsa Chih",
  ["jiegou huaxue"] = "Jiegou Huaxue",
  ["jikeikai medical journal"] = "Jikeikai Med J",
  ["jikken dobutsu (experimental animals)"] = "Jikken Dobutsu",
  ["jikken dobutsu. experimental animals"] = "Jikken Dobutsu",
  ["jikoketsu yuketsu : nihon jikoketsu yuketsu gakkai kaishi = journal of japanese society of autologous blood transfusion"] = "Jikoketsu Yuketsu",
  ["jilin da xue xue bao. li xue ban = journal of jilin university. science edition"] = "Jilin Da Xue Xue Bao",
  ["jilin nong ye da xue xue bao"] = "Acta agriculturae universitatis jilinensis|Jilin Nong Ye Da Xue Xue Bao",
  ["jilin nong ye da xue xue bao = acta agriculturae universitatis jilinensis"] = "Jilin Nong Ye Da Xue Xue Bao",
  ["jilin xu mu shou yi"] = "Jilin Xu Mu Shou Yi",
  ["jimd reports"] = "JIMD Rep",
  ["jimlar mutane"] = "Jimlar Mutane",
  ["jin dai shi yan jiu"] = "Jin Dai Shi Yan Jiu",
  ["jin dai zhongguo fu nü shi yan jiu"] = "Jin Dai Zhongguo Fu Nu Shi Yan Jiu",
  ["jin dai zhongguo shi yan jiu tong xun : newsletter for modern chinese history"] = "Jin Dai Zhong Shi Yan Jiu Tong Xun",
  ["jinan da xue xue bao. zi ran ke xue yu yi xue ban"] = "Jinan Da Xue Xue Bao",
  ["jinbun (nihon jinbun kagakkai)"] = "Zinbun",
  ["jing ji lin yan jiu = economic forest researches"] = "Jing Ji Lin Yan Jiu",
  ["jing ji lun wen"] = "Jing Ji Lun Wen",
  ["jinko mondai kenkyu. [journal of population problems]"] = "Jinko Mondai Kenkyu",
  ["jinkogaku kenkyu / nihon jinko gakkai henshu"] = "Jinkogaku Kenkyu",
  ["jinkōgaku kenkyū"] = "Jinkogaku Kenkyu",
  ["jinrui idengaku zasshi"] = "Jinrui Idengaku Zasshi",
  ["jinrui idengaku zasshi. the japanese journal of human genetics"] = "Jinrui Idengaku Zasshi",
  ["jinruigaku zasshi = the journal of the anthropological society of nihon"] = "Jinruigaku Zasshi",
  ["jinruigaku zasshi. the journal of the anthropological society of nippon"] = "Jinruigaku Zasshi",
  ["jinshu rechuli"] = "Jinshu Rechuli",
  ["jinshu xuebao"] = "Jinshu Xuebao",
  ["jircas journal for scientific papers"] = "JIRCAS J Sci Pap",
  ["jirss : journal of the iranian statistical society"] = "JIRSS",
  ["jise. journal of information science and engineering"] = "JISE J. Inf. Sci. Eng.",
  ["jishou university"] = "J. Jishou Univ. Nat. Sci. Ed.",
  ["jisuanji yu yingyong huaxue"] = "Jisuanji Yu Yingyong Huaxue",
  ["jjournal of agricultural safety and health"] = "J. Agric. Saf. Health",
  ["jk practitioner : a journal of current clinical medicine & surgery"] = "JK Pract",
  ["jma journal"] = "JMA J",
  ["jme practical bioethics"] = "JME Pract. Bioethics",
  ["jmir aging"] = "JMIR Aging",
  ["jmir ai"] = "JMIR AI",
  ["jmir bioinformatics and biotechnology"] = "JMIR Bioinform Biotech",
  ["jmir biomedical engineering"] = "JMIR Biomed Eng",
  ["jmir cancer"] = "JMIR Cancer",
  ["jmir cardio"] = "JMIR Cardio",
  ["jmir data"] = "JMIR Data",
  ["jmir dermatology"] = "JMIR Dermatol",
  ["jmir diabetes"] = "JMIR Diabetes",
  ["jmir formative research"] = "JMIR Form Res",
  ["jmir human factors"] = "JMIR Hum Factors",
  ["jmir infodemiology"] = "JMIR Infodemiol.",
  ["jmir medical education"] = "JMIR Med Educ",
  ["jmir medical informatics"] = "JMIR Med Inform",
  ["jmir mental health"] = "JMIR Ment Health",
  ["jmir mhealth and uhealth"] = "JMIR Mhealth Uhealth",
  ["jmir nursing"] = "JMIR Nurs",
  ["jmir pediatrics and parenting"] = "JMIR Pediatr Parent",
  ["jmir perioperative medicine"] = "JMIR Perioper Med",
  ["jmir public health and surveillance"] = "JMIR Public Health Surveill",
  ["jmir rehabilitation and assistive technologies"] = "JMIR Rehabil Assist Technol",
  ["jmir research protocols"] = "JMIR Res Protoc",
  ["jmir serious games"] = "JMIR Serious Games",
  ["jmlr workshop and conference proceedings"] = "JMLR Workshop Conf Proc",
  ["jmlr: workshop and conference proceedings"] = "JMLR Workshop Conf. Proc.",
  ["jmm case reports"] = "JMM Case Rep",
  ["jmr, journal of marketing research"] = "J Mark Res",
  ["jmst advances"] = "JMST Adv.",
  ["jnaiam. journal of numerical analysis, industrial and applied mathematics"] = "JNAIAM. J. Numer. Anal. Ind. Appl. Math.",
  ["jnci cancer spectrum"] = "JNCI Cancer Spectr",
  ["jnma; journal of the nepal medical association"] = "JNMA J Nepal Med Assoc",
  ["jnul publications"] = "JNUL Publ.",
  ["joannea--geologie und paläontologie"] = "Joannea Geol Palaontol",
  ["joensuun yliopiston luonnontieteellisiä julkaisuja"] = "Joensuun Yliop. Luonnont. Julk.",
  ["jogn nursing"] = "JOGN Nurs.",
  ["jogn nursing; journal of obstetric, gynecologic, and neonatal nursing"] = "JOGN Nurs",
  ["johns hopkins apl technical digest"] = "Johns Hopkins APL Tech Dig",
  ["johns hopkins magazine"] = "Johns Hopkins Mag",
  ["johns hopkins medical journal"] = "Johns Hopkins Med. J.",
  ["johns hopkins medical journal. supplement"] = "Johns Hopkins Med J Suppl",
  ["johns hopkins medical letter, health after 50"] = "Johns Hopkins Med. Lett. Health After 50",
  ["johns hopkins nurses alumni magazine"] = "Johns Hopkins Nurses Alumni Mag.",
  ["johns hopkins studies in the history of mathematics"] = "Johns Hopkins Stud. Hist. Math.",
  ["johns hopkins studies in the mathematical sciences"] = "Johns Hopkins Stud. Math. Sci.",
  ["johnson matthey technology review"] = "Johnson Matthey Technol. Rev.",
  ["joicfp news"] = "JOICFP News",
  ["joicfp review"] = "JOICFP Rev",
  ["joint bone spine"] = "Joint Bone Spine",
  ["joint commission journal on quality and patient safety"] = "Jt Comm J Qual Patient Saf",
  ["joint commission journal on quality and patient safety / joint commission resources"] = "Jt Comm J Qual Patient Saf",
  ["joint commission journal on quality and safety"] = "Jt Comm J Qual Saf",
  ["joint commission journal on quality improvement"] = "Jt. Comm. J. Qual. Improv.",
  ["joint commission perspectives"] = "Jt. Comm. Perspect.",
  ["joint commission perspectives. joint commission on accreditation of healthcare organizations"] = "Jt Comm Perspect",
  ["joint diseases and related surgery"] = "Jt Dis Relat Surg",
  ["joint eurohaptics conference and symposium on haptic interfaces for virtual environment and teleoperator systems : world haptics conference. world haptics conference"] = "Joint Eurohaptics Conf Symp Haptic Interfaces Virtual Environ Teleoper Syst",
  ["joint meeting abstracts. american dairy science association. meeting"] = "Jt Meet Abstr Am Dairy Sci Assoc",
  ["joint, bone, spine"] = "Joint Bone Spine",
  ["joint, bone, spine : revue du rhumatisme"] = "Joint Bone Spine",
  ["joj ophthalmology"] = "JOJ Ophthalmol",
  ["jom (warrendale, pa. : 1989)"] = "JOM (1989)",
  ["jom journal of the minerals metals and materials society"] = "JOM",
  ["jomec journal : journalism, media and cultural studies"] = "JOMEC J",
  ["jona's healthcare law, ethics and regulation"] = "JONAS Healthc Law Ethics Regul",
  ["jona’s healthcare law, ethics and regulation"] = "JONAS Healthc Law Ethics Regul",
  ["jones and bartlett books in mathematics and computer science"] = "Jones Bartlett Books Math. Comput. Sci.",
  ["jones and bartlett series in computer science"] = "Jones Bartlett Ser. Comput. Sci.",
  ["jop : journal of the pancreas"] = "JOP",
  ["jor spine"] = "JOR Spine",
  ["jordan dental journal"] = "Jordan Dent. J.",
  ["jordan journal of biological sciences"] = "Jordan J Biol Sci",
  ["jordan journal of chemistry"] = "Jordan J. Chem.",
  ["jordan journal of mathematics and statistics"] = "Jordan J. Math. Stat.",
  ["jordan medical journal"] = "Jordan Med J",
  ["jordmorbladet : bilag til tidsskriftet sykepleien"] = "Jordmorbladet",
  ["jornada médica"] = "Jorn Med",
  ["jornal brasileiro de doencas toracicas"] = "J. Bras. Doencas Torac.",
  ["jornal brasileiro de doenças torácicas"] = "J Bras Doencas Torac",
  ["jornal brasileiro de ginecologia"] = "J Bras Ginecol",
  ["jornal brasileiro de medicina"] = "J Bras Med",
  ["jornal brasileiro de nefrologia : ʹorgão oficial de sociedades brasileira e latino-americana de nefrologia"] = "J Bras Nefrol",
  ["jornal brasileiro de neurologia"] = "J Bras Neurol",
  ["jornal brasileiro de patologia e medicina laboratorial"] = "J Bras Patol Med Lab",
  ["jornal brasileiro de pneumologia : publicacao oficial da sociedade brasileira de pneumologia e tisilogia"] = "J Bras Pneumol",
  ["jornal brasileiro de pneumologia : publicaça̋o oficial da sociedade brasileira de pneumologia e tisilogia"] = "J Bras Pneumol",
  ["jornal brasileiro de psiquiatria"] = "J Bras Psiquiatr",
  ["jornal brasileiro de reprodução assistida"] = "J Bras Reprod Assist",
  ["jornal da sociedade brasileira de fonoaudiologia"] = "J Soc Bras Fonoaudiol",
  ["jornal da sociedade das ciencias medicas de lisboa"] = "J. Soc. Cienc. Med. Lisb.",
  ["jornal da sociedade das ciências médicas de lisboa"] = "J Soc Cienc Med Lisb",
  ["jornal de historia da medicina"] = "J Hist Med",
  ["jornal de medicina de pernambuco"] = "J Med Pernamb",
  ["jornal de pediatria"] = "J Pediatr (Rio J)",
  ["jornal de psicologia"] = "J Psicol",
  ["jornal do medico"] = "J. Med. (Oporto)",
  ["jornal do médico"] = "J Med (Oporto)",
  ["jornal dos clinicos"] = "J Clin",
  ["jornal dos farmacêuticos"] = "J Farm",
  ["jornal vascular brasileiro"] = "J Vasc Bras",
  ["josai shika daigaku kiyo (bulletin of the josai dental university)"] = "Josai Shika Daigaku Kiyo",
  ["josai shika daigaku kiyo. the bulletin of the josai dental university"] = "Josai Shika Daigaku Kiyo",
  ["josanpu zasshi (japanese journal for midwives)"] = "Josanpu Zasshi",
  ["josanpu zasshi = the japanese journal for midwife"] = "Josanpu Zasshi",
  ["jot journal fur oberflaechentechnik"] = "JOT, J. Oberflaechentech.",
  ["jot journal für oberflaechentechnik"] = "JOT, J. Oberflaechentech.",
  ["jounal of agriculture and food technology"] = "J Agric Food Technol",
  ["jounal of wildlife research"] = "J. Wildl. Res.",
  ["journal (academy of hospital administration (india))"] = "J Acad Hosp Adm",
  ["journal (american academy of gnathologic orthopedics)"] = "J Am Acad Gnathol Orthop",
  ["journal (american association for medical transcription)"] = "J Am Assoc Med Transcr",
  ["journal (american medical record association)"] = "J Am Med Rec Assoc",
  ["journal (anglo-continental dental society)"] = "J Anglocont Dent Soc",
  ["journal (association for healthcare philanthropy (u.s.))"] = "J Assoc Healthc Philanthr",
  ["journal (canadian dental association)"] = "J Can Dent Assoc",
  ["journal (canadian society of forensic science)"] = "Journal (Can Soc Forensic Sci)",
  ["journal (indiana dental association)"] = "J Indiana Dent Assoc",
  ["journal (institute of health record & information management : 2008)"] = "J (Inst Health Rec Inf Manag)",
  ["journal (institute of health record information and management)"] = "Journal (Inst Health Rec Inf Manag)",
  ["journal (institute of sterile services management)"] = "J Inst Sterile Serv Manage",
  ["journal (national association for hospital development (u.s.))"] = "J Natl Assoc Hosp Dev",
  ["journal (new hampshire dental society)"] = "J N H Dent Soc",
  ["journal (north louisiana historical association)"] = "J North La Hist Assoc",
  ["journal (southern california dental assistants association)"] = "J South Calif Dent Assistants Assoc",
  ["journal - academy of general dentistry"] = "J Acad Gen Dent",
  ["journal - alabama dental association"] = "J Ala Dent Assoc",
  ["journal - american health care association"] = "J Am Health Care Assoc",
  ["journal - american intra-ocular implant society"] = "J Am Intraocul Implant Soc",
  ["journal - american society of agronomy"] = "J. Am. Soc. Agron.",
  ["journal - american water works association"] = "J Am Water Works Assoc",
  ["journal - association for hospital medical education"] = "J Assoc Hosp Med Educ",
  ["journal - association of official analytical chemists"] = "J Assoc Off Anal Chem",
  ["journal - california dental association"] = "J Calif Dent Assoc",
  ["journal - college of general practice of canada"] = "J Coll Gen Pract Can",
  ["journal - connecticut state dental association"] = "J Conn State Dent Assoc",
  ["journal - forensic science society"] = "J Forensic Sci Soc",
  ["journal - michigan state medical society"] = "J Mich State Med Soc",
  ["journal - mississippi dental association"] = "J Miss Dent Assoc",
  ["journal - national association of private psychiatric hospitals"] = "J Natl Assoc Priv Psychiatr Hosp",
  ["journal - newark beth israel hospital"] = "J Newark Beth Isr Hosp",
  ["journal - oklahoma dental association"] = "J Okla Dent Assoc",
  ["journal - oklahoma state dental association"] = "J Okla State Dent Assoc",
  ["journal - oklahoma state medical association"] = "J Okla State Med Assoc",
  ["journal - scottish labour history society. scottish labour history society"] = "J Scott Labour Hist Soc",
  ["journal - seattle-king county dental society"] = "J Seattle King Cty Dent Soc",
  ["journal - south carolina medical association"] = "J S C Med Assoc",
  ["journal - southern california dental association"] = "J South Calif Dent Assoc",
  ["journal - southern california state dental association"] = "J South Calif State Dent Assoc",
  ["journal - tennessee state dental association"] = "J Tenn State Dent Assoc",
  ["journal - united reformed church history society. united reformed church history society"] = "J United Reform Church Hist Soc",
  ["journal - water pollution control federation"] = "J Water Pollut Control Fed",
  ["journal / canadian dental association. journal de l association dentaire canadienne"] = "J Can Dent Assoc",
  ["journal / contra costa dental society"] = "J Contra Costa Dent Soc",
  ["journal / international association for the history of psychoanalysis"] = "J Int Assoc Hist Psychoanal",
  ["journal / maine dental association"] = "J Maine Dent Assoc",
  ["journal about women in higher education"] = "J Women High Educ",
  ["journal american water works association"] = "J. Am. Water Works Assn.",
  ["journal and proceedings of the royal institute of chemistry"] = "J. Proc. R. Inst. Chem.",
  ["journal and proceedings of the royal institute of chemistry of great britain and ireland"] = "J. Proc. R. Inst. G.B. Irel.",
  ["journal and proceedins of the institute of chemistry of great britain and ireland"] = "J. Proc. Inst. Chem. G.B. Irel.",
  ["journal asiatique"] = "J Asiat",
  ["journal awwa"] = "J. AWWA",
  ["journal belge d'urologie"] = "J Belge Urol",
  ["journal belge de medecine physique"] = "J. Belge Med. Phys.",
  ["journal belge de medecine physique et de rehabilitation"] = "J. Belge Med. Phys. Rehabil.",
  ["journal belge de medecine physique et de rehabilitation. belgisch tijdschrift voor fysische geneeskunde en rehabilitatie"] = "J Belge Med Phys Rehabil",
  ["journal belge de medecine physique et de rhumatologie. belgisch tijdschrift voor fysische geneeskunde en reumatologie"] = "J Belge Med Phys Rhumatol",
  ["journal belge de medecine physique. belgisch tijdschrift voor fysische geneeskunde"] = "J Belge Med Phys",
  ["journal belge de médecine physique = belgisch tijdschrift voor fysische geneeskunde"] = "J Belge Med Phys",
  ["journal belge de médecine physique et de réhabilitation. = belgisch tijdschrift voor fysische geneeskunde en rehabilitatie"] = "J Belge Med Phys Rehabil",
  ["journal belge de médecine physique et de rhumatologie. belgisch tijdschrift voor fysische geneeskunde en reumatologie"] = "J Belge Med Phys Rhumatol",
  ["journal belge de neurologie et de psychiatrie"] = "J Belge Neurol Psychiatr",
  ["journal belge de radiologie"] = "J Belge Radiol",
  ["journal belge de rhumatologie et de medecine physique"] = "J. Belge Rhumatol. Med. Phys.",
  ["journal belge de rhumatologie et de medecine physique. belgisch tijdschrift voor reumatologie en fysische geneeskunde"] = "J Belge Rhumatol Med Phys",
  ["journal belge de rhumatologie et de médecine physique = belgisch tijdschrift voor reumatologie en fysische geneeskunde"] = "J Belge Rhumatol Med Phys",
  ["journal belge d’urologie"] = "J Belge Urol",
  ["journal canadien des sciences de la terre"] = "J. Can. Sci. Terre",
  ["journal cannt"] = "J. CANNT",
  ["journal d'agriculture traditionnelle et de botanique appliquée : jatba"] = "J Agric Tradit Bot Appl",
  ["journal d'agriculture tropicale et de botanique appliquée"] = "J Agric Trop Bot Appl",
  ["journal d'odontologie conservatrice"] = "J Odontol Conserv",
  ["journal d'urologie"] = "J Urol (Paris)",
  ["journal d'urologie et de néphrologie"] = "J Urol Nephrol (Paris)",
  ["journal d'urologie médicale et chirurgicale"] = "J Urol Medicale Chir",
  ["journal das auxiliares odontologicas"] = "J Aux Odontol",
  ["journal de biologie buccale"] = "J Biol Buccale",
  ["journal de biomateriaux dentaires"] = "J. Biomater. Dent.",
  ["journal de biomateriaux dentaires : [publication du college francais de biomateriaux dentaires]"] = "J Biomater Dent",
  ["journal de biomatériaux dentaires : [publication du collège français de biomatériaux dentaires]"] = "J Biomater Dent",
  ["journal de chimie physique"] = "J. Chim. Phys.",
  ["journal de chimie physique et de physico-chimie biologique"] = "J. Chim. Phys. Phys.- Chim. Biol.",
  ["journal de chimie physique et de physicochimie biologique"] = "J Chim Phys Physicochim Biol",
  ["journal de chirurgie"] = "J Chir (Paris)",
  ["journal de chirurgie viscérale"] = "J Chir Visc",
  ["journal de genetique humaine"] = "J. Genet. Hum.",
  ["journal de génétique humaine"] = "J Genet Hum",
  ["journal de gynecologie, obstetrique et biologie de la reproduction"] = "J Gynecol Obstet Biol Reprod (Paris)",
  ["journal de gynécologie, obstétrique et biologie de la reproduction"] = "J Gynecol Obstet Biol Reprod (Paris)",
  ["journal de la sfds. journal de la societe francaise de statistique"] = "J. SFdS",
  ["journal de la societa de biologie"] = "J. Soc. Biol.",
  ["journal de la societe algerienne de chimie"] = "J. Soc. Alg. Chim.",
  ["journal de la societe chimique de tunisie"] = "J. Soc. Chim. Tunis.",
  ["journal de la societe de biologie"] = "J. Soc. Biol.",
  ["journal de la societe de statistique de paris"] = "J Soc Stat Paris",
  ["journal de la societe des americanistes"] = "J Soc Am",
  ["journal de la societe des oceanistes"] = "J Soc Ocean",
  ["journal de la societe ouest africaine de chimie"] = "J. Soc. Ouest Afr. Chim.",
  ["journal de la société de biologie"] = "J Soc Biol",
  ["journal de la société de statistique de paris"] = "J Soc Stat Paris",
  ["journal de la société des américanistes"] = "J Soc Am",
  ["journal de la société des océanistes"] = "J Soc Ocean",
  ["journal de la société française de statistique (2009)"] = "J Soc Fr Statistique (2009)",
  ["journal de l’ecole polytechnique. mathematiques"] = "J. Ec. polytech. Math.",
  ["journal de mathematiques pures et appliquees"] = "J. Math. Pures Appl.",
  ["journal de mathematiques pures et appliquees. neuvieme serie"] = "J. Math. Pures Appl. (9)",
  ["journal de mathématiques pures et appliquées"] = "J Math Pures Appl",
  ["journal de mathématiques pures et appliquées"] = "J. Math. Pures Appl. (9)",
  ["journal de medecine de besancon"] = "J Med Besancon",
  ["journal de medecine de bordeaux et du sud-ouest"] = "J. Med. Bord.",
  ["journal de medecine de caen"] = "J Med Caen",
  ["journal de medecine de lyon"] = "J. Med. Lyon",
  ["journal de medecine de nantes"] = "J Med Nantes",
  ["journal de medecine de paris"] = "J Med Paris",
  ["journal de medecine de strasbourg"] = "J Med Strasb",
  ["journal de medecine et de chirurgie pratiques"] = "J Med Chir Prat",
  ["journal de medecine legale droit medical"] = "J Med Leg Droit Med",
  ["journal de medecine nucleaire et biophysique"] = "J. Med. Nucl. Biophys.",
  ["journal de médecine de besançon"] = "J Med Besancon",
  ["journal de médecine de bordeaux et du sud-ouest"] = "J Med Bord",
  ["journal de médecine de caen"] = "J Med Caen",
  ["journal de médecine de lyon"] = "J Med Lyon",
  ["journal de médecine de nantes"] = "J Med Nantes",
  ["journal de médecine de paris"] = "J Med Paris",
  ["journal de médecine de strasbourg"] = "J Med Strasb",
  ["journal de médecine et de chirurgie pratiques"] = "J Med Chir Prat",
  ["journal de médecine légale droit médical"] = "J Med Leg Droit Med",
  ["journal de médecine vasculaire"] = "J Med Vasc",
  ["journal de microscopie"] = "J Microsc (Paris)",
  ["journal de microscopie et de biologie cellulaire"] = "J Microsc Biol Cell",
  ["journal de mycologie médicale"] = "J Mycol Med",
  ["journal de parodontologie"] = "J Parodontol",
  ["journal de pédiatrie et de puériculture"] = "J Pediatr Pueric",
  ["journal de pharmacie de belgique"] = "J Pharm Belg",
  ["journal de pharmacologie"] = "J Pharmacol",
  ["journal de physiologie"] = "J Physiol (Paris)",
  ["journal de physiologie et de pathologie générale"] = "J Physiol Pathol Gen",
  ["journal de physiologie. supplement"] = "J. Physiol. Suppl. (Paris)",
  ["journal de physiologie. supplément"] = "J Physiol Suppl (Paris)",
  ["journal de physique"] = "J. Phys.",
  ["journal de physique (paris)"] = "J. Phys. (Paris)",
  ["journal de physique colloques"] = "J. Phys. Colloq.",
  ["journal de physique et le radium"] = "J. Phys. Radium",
  ["journal de physique i"] = "J. Phys. I",
  ["journal de physique ii"] = "J. Phys. II",
  ["journal de physique iii"] = "J. Phys. III",
  ["journal de physique iv"] = "J. Phys. IV",
  ["journal de physique iv (proceedings)"] = "J. Phys. IV Proc.",
  ["journal de psychologie normale et pathologique"] = "J Psychol Norm Pathol (Paris)",
  ["journal de radiologie"] = "J Radiol",
  ["journal de radiologie, d'électrologie & archives d'électricité médicale"] = "J Radiol Electrol Arch Electr Medicale",
  ["journal de radiologie, d'électrologie, et de médecine nucléaire"] = "J Radiol Electrol Med Nucl",
  ["journal de radiologie, d’electrologie & archives d’electricite medicale"] = "J Radiol Electrol Arch Electr Medicale",
  ["journal de radiologie, d’electrologie, et de medecine nucleaire"] = "J. Radiol. Electrol. Med. Nucl.",
  ["journal de recherches atmospheriques"] = "J. Rech. Atmos.",
  ["journal de stomatologie de belgique"] = "J Stomatol Belg",
  ["journal de theorie des nombres de bordeaux"] = "J. Theor. Nombres Bordeaux",
  ["journal de théorie des nombres de bordeaux"] = "J. Théor. Nombres Bordeaux",
  ["journal de toxicologie clinique et experimentale"] = "J. Toxicol. Clin. Exp.",
  ["journal de toxicologie clinique et expérimentale"] = "J Toxicol Clin Exp",
  ["journal de toxicologie medicale"] = "J. Toxicol. Med.",
  ["journal dentaire du quebec"] = "J. Dent. Que.",
  ["journal der deutschen dermatologischen gesellschaft (journal of the german society of dermatology)"] = "J. Dtsch. Dermatol. Ges.",
  ["journal der deutschen dermatologischen gesellschaft = journal of the german society of dermatology : jddg"] = "J Dtsch Dermatol Ges",
  ["journal des jec"] = "J JEC",
  ["journal des maladies vasculaires"] = "J Mal Vasc",
  ["journal des praticiens; revue generale de clinique et de therapeutique"] = "J Prat Rev Gen Clin Ther",
  ["journal des praticiens; revue générale de clinique et de thérapeutique"] = "J Prat Rev Gen Clin Ther",
  ["journal des savants"] = "JS",
  ["journal des sciences medicales de lille"] = "J. Sci. Med. Lille",
  ["journal des sciences médicales de lille"] = "J Sci Med Lille",
  ["journal d’agriculture traditionnelle et de botanique appliquee : jatba"] = "J Agric Tradit Bot Appl",
  ["journal d’agriculture tropicale et de botanique appliquee"] = "J Agric Trop Bot Appl",
  ["journal d’analyse mathematique"] = "J. Anal. Math.",
  ["journal d’analyse mathématique"] = "J. Anal. Math.",
  ["journal d’odontologie conservatrice"] = "J. Odontol. Conserv.",
  ["journal d’odontologie conservatrice / cneoc"] = "J Odontol Conserv",
  ["journal d’urologie"] = "J. Urol. (Paris)",
  ["journal d’urologie et de nephrologie"] = "J. Urol. Nephrol. (Paris)",
  ["journal d’urologie medicale et chirurgicale"] = "J Urol Medicale Chir",
  ["journal europeen d’hydrologie"] = "J. Eur. Hydrol.",
  ["journal for development and leadership. nelson mandela metropolitan university"] = "J Dev Leadersh",
  ["journal for eighteenth-century studies"] = "J Eighteenth Century Stud",
  ["journal for general philosophy of science"] = "J. Gen. Philos. Sci.",
  ["journal for general philosophy of science = zeitschrift für allgemeine wissenschaftstheorie"] = "J Gen Philos Sci",
  ["journal for general philosophy of science. zeitschrift fur allgemeine wissenschaftstheorie"] = "J. Gen. Philos. Sci.",
  ["journal for geometry and graphics"] = "J. Geom. Graph.",
  ["journal for healthcare quality"] = "J. Healthc. Qual.",
  ["journal for healthcare quality : official publication of the national association for healthcare quality"] = "J Healthc Qual",
  ["journal for hospital admitting management"] = "J. Hosp. Admit. Manage.",
  ["journal for immunotherapy of cancer"] = "J Immunother Cancer",
  ["journal for labour market research"] = "J Labour Mark Res",
  ["journal for maritime research"] = "J Marit Res",
  ["journal for medicaid management"] = "J Medicaid Manage",
  ["journal for modeling in ophthalmology"] = "J Model Ophthalmol",
  ["journal for nature conservation"] = "J Nat Conserv",
  ["journal for nurses in professional development"] = "J Nurses Prof Dev",
  ["journal for nurses in staff development"] = "J. Nurses Staff Dev.",
  ["journal for nurses in staff development : jnsd : official journal of the national nursing staff development organization"] = "J Nurses Staff Dev",
  ["journal for person-oriented research"] = "J Pers Oriented Res",
  ["journal for research in mathematics education"] = "J Res Math Educ",
  ["journal for scientific research. medical sciences"] = "J Sci Res Med Sci",
  ["journal for social action in counseling and psychology"] = "J Soc Action Couns Psychol",
  ["journal for specialists in pediatric nursing"] = "J. Spec. Pediatr. Nurs.",
  ["journal for specialists in pediatric nursing : jspn"] = "J Spec Pediatr Nurs",
  ["journal for stem education research"] = "J STEM Educ Res",
  ["journal for studies in economics and econometrics"] = "J. Stud. Econ. Econometrics",
  ["journal for the anthropology of north america"] = "J Anthropol N Am",
  ["journal for the history of arabic science"] = "J Hist Arabic Sci",
  ["journal for the history of astronomy"] = "J. Hist. Astron.",
  ["journal for the measurement of physical behaviour"] = "J Meas Phys Behav",
  ["journal for the scientific study of religion"] = "J Sci Study Relig",
  ["journal for the study of judaism: (in the persian, hellenistic and roman period)"] = "JSJ",
  ["journal for the study of religious experience"] = "J Study Relig Exp",
  ["journal for the study of the new testament"] = "J Study New Testam",
  ["journal for the theory of social behaviour"] = "J Theory Soc Behav",
  ["journal for vascular ultrasound"] = "J. Vasc. Ultrasound",
  ["journal forestier suisse"] = "J. for. suisse",
  ["journal francais d ophtalmologie"] = "J Fr Ophtalmol",
  ["journal francais de medecine et chirurgie thoraciques"] = "J. Fr. Med. Chir. Thorac.",
  ["journal francais d’ophtalmologie"] = "J. Fr. Ophtalmol.",
  ["journal francais d’oto-rhino-laryngologie et chirurgie maxillo-faciale"] = "J. Fr. Otorhinolaryngol. Chir. Maxillofac.",
  ["journal francais d’oto-rhino-laryngologie, audio-phonologie et chirurgie maxillo-faciale"] = "J. Fr. Otorhinolaryngol. Audiophonol. Chir. Maxillofac.",
  ["journal francais d’oto-rhino-laryngologie, audiophonologie, chirurgie maxillo-faciale"] = "J. Fr. Otorhinolaryngol. Audiophonol. Chir. Maxillofac.",
  ["journal francais d’oto-rhino-laryngologie; audiophonologie, chirurgie maxillo-faciale"] = "J Fr Otorhinolaryngol Audiophonol Chir Maxillofac",
  ["journal français d'ophtalmologie"] = "J Fr Ophtalmol",
  ["journal français d'oto-rhino-laryngologie et chirurgie maxillo-faciale"] = "J Fr Otorhinolaryngol Chir Maxillofac",
  ["journal français d'oto-rhino-laryngologie, audio-phonologie et chirurgie maxillo-faciale"] = "J Fr Otorhinolaryngol Audiophonol Chir Maxillofac",
  ["journal français d'oto-rhino-laryngologie; audiophonologie, chirurgie maxillo-faciale"] = "J Fr Otorhinolaryngol Audiophonol Chir Maxillofac",
  ["journal français de médecine et chirurgie thoraciques"] = "J Fr Med Chir Thorac",
  ["journal fur die reine und angewandte mathematik"] = "J. Reine Angew. Math.",
  ["journal fur die reine und angewandte mathematik. [crelle’s journal]"] = "J. Reine Angew. Math.",
  ["journal fur entwicklungspolitik : jep"] = "J Entwickl Polit",
  ["journal fur gasbeleuchtung und verwandte beleuchtungsarten sowie fur wasserversorgung"] = "J. Gasbeleucht. Verw. Beleuchtungsarten Wasserversorg.",
  ["journal fur geschichte"] = "J Gesch",
  ["journal fur hirnforschung"] = "J Hirnforsch",
  ["journal fur klinische endokrinologie und stoffwechsel"] = "J. Klin. Endokrinolog. Stoffwechsel",
  ["journal fur kulturpflanzen"] = "J. Kulturpflanz.",
  ["journal fur landwirtschaft"] = "J. Landwirtschaft",
  ["journal fur mathematik-didaktik"] = "J. Math.-Didakt.",
  ["journal fur ornithologie"] = "J. Ornithol.",
  ["journal fur praktische chemie"] = "J. Prakt. Chem.",
  ["journal fur sozialforschung"] = "J Soz Forsch",
  ["journal fur verbraucherschutz und lebensmittelsicherheit"] = "J. Verbraucherschutz Lebensmittelsicherh.",
  ["journal für ästhetische chirurgie"] = "J Asthet Chir",
  ["journal für entwicklungspolitik : jep"] = "J Entwickl Polit",
  ["journal für gastroenterologische und hepatologische erkrankungen"] = "J Gastroenterol Hepatol Erkrank",
  ["journal für geschichte"] = "J Gesch",
  ["journal für gynäkologische endokrinologie (österreichische ausg.)"] = "J Gynakol Endokrinol",
  ["journal für hirnforschung"] = "J Hirnforsch",
  ["journal für klinische endokrinologie und stoffwechsel"] = "J Klin Endokrinol Stoffwechs",
  ["journal für kulturpflanzen = journal of cultivated plants"] = "J Kulturpflanzen",
  ["journal für sozialforschung"] = "J Soz Forsch",
  ["journal für urologie und urogynäkologie : zeitschrift für urologie und urogynäkologie in klinik und praxis"] = "J Urol Urogynakologie",
  ["journal für verbraucherschutz und lebensmittelsicherheit = journal of consumer protection and food safety"] = "J Verbrauch Lebensm",
  ["journal für die reine und angewandte mathematik"] = "J. Reine Angew. Math.",
  ["journal für kunstgeschichte"] = "JKuGesch",
  ["journal für ornithologie"] = "J. Ornithol.",
  ["journal für praktische chemie"] = "J. Prakt. Chem.",
  ["journal für verbraucherschutz und lebensmittelsicherheit"] = "J. Verbraucherschutz Lebensmittelsicherh.",
  ["journal international de bioéthique = international journal of bioethics"] = "J Int Bioethique",
  ["journal international de bioéthique et d'éthique des sciences"] = "J Int Bioethique Ethique Sci",
  ["journal international de chirurgie"] = "J Int Chir",
  ["journal international de victimologie"] = "J Int Vict",
  ["journal medical libanais (lebanese medical journal)"] = "J. Med. Liban.",
  ["journal medical libanais. lebanese medical journal"] = "J Med Liban",
  ["journal méditerranée médical"] = "J Mediterr Med",
  ["journal of  circadian rhythms"] = "J. Circadian Rhythms",
  ["journal of 3d printing in medicine"] = "J 3D Print Med",
  ["journal of aapos"] = "J. AAPOS",
  ["journal of aapos : the official publication of the american association for pediatric ophthalmology and strabismus"] = "J AAPOS",
  ["journal of aapos : the official publication of the american association for pediatric ophthalmology and strabismus / american association for pediatric ophthalmology and strabismus"] = "J AAPOS",
  ["journal of abbasid studies"] = "J Abbasid Stud",
  ["journal of abnormal and social psychology"] = "J Abnorm Soc Psychol",
  ["journal of abnormal child psychology"] = "J Abnorm Child Psychol",
  ["journal of abnormal psychology"] = "J Abnorm Psychol",
  ["journal of aboriginal health"] = "J Aborig Health",
  ["journal of abstract differential equations and applications"] = "J. Abstr. Differ. Equ. Appl.",
  ["journal of academic ethics"] = "J Acad Ethics",
  ["journal of academic ophthalmology (2017)"] = "J Acad Ophthalmol",
  ["journal of access services"] = "J Access Serv",
  ["journal of accident & emergency medicine"] = "J Accid Emerg Med",
  ["journal of accident and emergency medicine"] = "J. Accid. Emerg. Med.",
  ["journal of accounting and economics"] = "J. Acc. Econ.",
  ["journal of accounting research"] = "J. Acc. Res.",
  ["journal of accounting, auditing and finance"] = "J. Acc.",
  ["journal of achievements in materials and manufacturing engineering"] = "J. Achiev. Mater. Manuf. Eng.",
  ["journal of acquired immune deficiency syndromes"] = "J Acquir Immune Defic Syndr (1988)",
  ["journal of acquired immune deficiency syndromes (1999)"] = "J Acquir Immune Defic Syndr",
  ["journal of acquired immune deficiency syndromes and human retrovirology"] = "J. Acquir. Immune Defic. Syndr. Hum. Retrovirol.",
  ["journal of acquired immune deficiency syndromes and human retrovirology : official publication of the international retrovirology association"] = "J Acquir Immune Defic Syndr Hum Retrovirol",
  ["journal of activity, sedentary and sleep behaviors"] = "J. Act. Sedentary Sleep Behav",
  ["journal of acupuncture and meridian studies"] = "J Acupunct Meridian Stud",
  ["journal of acute care physical therapy"] = "J Acute Care Phys Ther",
  ["journal of acute medicine"] = "J Acute Med",
  ["journal of addiction"] = "J Addict",
  ["journal of addiction & prevention"] = "J Addict Prev",
  ["journal of addiction and dependence"] = "J Addict Depend",
  ["journal of addiction and preventive medicine"] = "J Addict Prev Med",
  ["journal of addiction medicine"] = "J Addict Med",
  ["journal of addiction medicine and therapeutic science"] = "J Addict Med Ther Sci",
  ["journal of addiction research & therapy"] = "J. Addict. Res. Ther.",
  ["journal of addiction research (opast group)"] = "J Addict Res (OPAST Group)",
  ["journal of addiction science"] = "J Addict Sci",
  ["journal of addictions & offender counseling"] = "J Addict Offender Couns",
  ["journal of addictions nursing"] = "J Addict Nurs",
  ["journal of addictive behaviors and therapy"] = "J Addict Behav Ther",
  ["journal of addictive behaviors, therapy & rehabilitation"] = "J Addict Behav Ther Rehabil",
  ["journal of addictive diseases"] = "J Addict Dis",
  ["journal of addictive diseases : the official journal of the asam, american society of addiction medicine"] = "J Addict Dis",
  ["journal of adenocarcinoma"] = "J Adenocarcinoma",
  ["journal of adhd & related disorders"] = "J ADHD Relat Disord",
  ["journal of adhesion"] = "J. Adhes.",
  ["journal of adhesion science and technology"] = "J. Adhes. Sci. Technol.",
  ["journal of adhesive dentistry"] = "J. Adhes. Dent.",
  ["journal of adolescence"] = "J Adolesc",
  ["journal of adolescent & adult literacy : a journal from the international reading association"] = "J Adolesc Adult Lit",
  ["journal of adolescent and young adult oncology"] = "J Adolesc Young Adult Oncol",
  ["journal of adolescent health"] = "J. Adolesc. Health",
  ["journal of adolescent health care"] = "J. Adolesc. Health Care",
  ["journal of adolescent health care : official publication of the society for adolescent medicine"] = "J Adolesc Health Care",
  ["journal of adolescent research"] = "J Adolesc Res",
  ["journal of adult development"] = "J Adult Dev",
  ["journal of advanced academics"] = "J Adv Acad",
  ["journal of advanced biotechnology and experimental therapeutics"] = "J. Adv. Biotechnol. Exp. Ther.",
  ["journal of advanced ceramics"] = "J. Adv. Ceram.",
  ["journal of advanced concrete technology"] = "J. Adv. Concr. Technol.",
  ["journal of advanced dielectrics"] = "J. Adv. Dielectr.",
  ["journal of advanced joining processes"] = "J. Adv. Joining Processes",
  ["journal of advanced manufacturing and processing"] = "J Adv Manuf Process",
  ["journal of advanced manufacturing systems"] = "J. Adv. Manuf. Syst.",
  ["journal of advanced materials"] = "J. Adv. Mater.",
  ["journal of advanced mathematical studies"] = "J. Adv. Math. Stud.",
  ["journal of advanced medical-surgical nursing"] = "J Adv Med Surg Nurs",
  ["journal of advanced microscopy research"] = "J. Adv. Microsc. Res.",
  ["journal of advanced nursing"] = "J Adv Nurs",
  ["journal of advanced nutritional and human metabolism"] = "J Adv Nutr Hum Metab",
  ["journal of advanced oral research"] = "J Adv Oral Res",
  ["journal of advanced oxidation technologies"] = "J. Adv. Oxid. Technol.",
  ["journal of advanced pharmaceutical technology & research"] = "J Adv Pharm Technol Res",
  ["journal of advanced research"] = "J Adv Res",
  ["journal of advanced scientific research"] = "J Adv Sci Res",
  ["journal of advanced smart convergence"] = "J Adv Smart Convergence",
  ["journal of advanced studies in topology"] = "J. Adv. Stud. Topol.",
  ["journal of advanced transportation"] = "J Adv Transp",
  ["journal of advanced veterinary and animal research"] = "J Adv Vet Anim Res",
  ["journal of advanced veterinary research"] = "J Adv Vet Res",
  ["journal of advancement in medical and life sciences"] = "J Adv Med Life Sci",
  ["journal of advances in biology"] = "J Adv Biol",
  ["journal of advances in biology & biotechnology"] = "J Adv Biol Biotechnol",
  ["journal of advances in chemical physics"] = "J. Adv. Chem. Phys.",
  ["journal of advances in dairy research"] = "J Adv Dairy Res",
  ["journal of advances in medical and pharmaceutical sciences"] = "J Adv Med Pharm Sci",
  ["journal of advances in medical education & professionalism"] = "J Adv Med Educ Prof",
  ["journal of advances in medicine and medical research"] = "J Adv Med Med Res",
  ["journal of advances in microbiology"] = "J Adv Microbiol",
  ["journal of advances in modeling earth systems"] = "J Adv Model Earth Syst",
  ["journal of advertising"] = "J Advert",
  ["journal of advertising research"] = "J Advert Res",
  ["journal of aerosol medicine"] = "J. Aerosol Med.",
  ["journal of aerosol medicine : the official journal of the international society for aerosols in medicine"] = "J Aerosol Med",
  ["journal of aerosol medicine and pulmonary drug delivery"] = "J Aerosol Med Pulm Drug Deliv",
  ["journal of aerosol science"] = "J. Aerosol Sci.",
  ["journal of aerospace engineering"] = "J Aerosp Eng",
  ["journal of aerospace operations"] = "J. Aerosp. Oper.",
  ["journal of affective disorders"] = "J Affect Disord",
  ["journal of affective disorders reports"] = "J Affect Disord Rep",
  ["journal of african american studies (new brunswick, n.j.)"] = "J Afr Am Stud (New Brunsw)",
  ["journal of african association of physiological sciences"] = "J Afr Assoc Physiol Sci",
  ["journal of african business"] = "J. African Bus.",
  ["journal of african earth sciences"] = "J. Afr. Earth. Sci.",
  ["journal of african earth sciences (oxford, england : 1994)"] = "J Afr Earth Sci",
  ["journal of african economies"] = "J Afr Econ",
  ["journal of african finance and economic development"] = "J. African Finance Econ. Devel.",
  ["journal of african history"] = "J Afr Hist",
  ["journal of african law"] = "J Afr Law",
  ["journal of african studies"] = "J Afr Stud",
  ["journal of ageing research and healthcare"] = "J Ageing Res Healthc",
  ["journal of aggression, conflict and peace research"] = "J Aggress Confl Peace Res",
  ["journal of aggression, maltreatment & trauma"] = "J Aggress Maltreat Trauma",
  ["journal of aging & social policy"] = "J Aging Soc Policy",
  ["journal of aging and gerontology"] = "J Aging Gerontol",
  ["journal of aging and health"] = "J Aging Health",
  ["journal of aging and judaism"] = "J Aging Jud",
  ["journal of aging and physical activity"] = "J Aging Phys Act",
  ["journal of aging and social policy"] = "J. Aging Soc. Policy",
  ["journal of aging research"] = "J Aging Res",
  ["journal of aging science"] = "J Aging Sci",
  ["journal of aging studies"] = "J Aging Stud",
  ["journal of agribusiness in developing and emerging economies"] = "J Agribus Dev Emerg Econ",
  ["journal of agricultural  and urban entomology"] = "J. Agric. Urban Entomol.",
  ["journal of agricultural & environmental ethics"] = "J Agric Environ Ethics",
  ["journal of agricultural & food industrial organization"] = "J Agric Food Ind Organ",
  ["journal of agricultural & food information"] = "J. Agric. Food Inf.",
  ["journal of agricultural and applied economics"] = "J. Agric. Appl. Econ.",
  ["journal of agricultural and applied economics association"] = "J. Agric. Appl. Econ. Assoc.",
  ["journal of agricultural and biological science"] = "J Agric Biol Sci",
  ["journal of agricultural and environmental ethics"] = "J. Agric. Environ. Ethics",
  ["journal of agricultural and food chemistry"] = "J Agric Food Chem",
  ["journal of agricultural and resource economics"] = "J. Agr. Resource Econ.",
  ["journal of agricultural and urban entomology"] = "J Agric Urban Entomol",
  ["journal of agricultural biological and environmental statistics"] = "J. Agric. Biol. Environ. Stat.",
  ["journal of agricultural chemistry and environment"] = "J Agric Chem Environ",
  ["journal of agricultural economics"] = "J Agric Econ",
  ["journal of agricultural economics research"] = "J. Agr. Econ. Res.",
  ["journal of agricultural education"] = "J Agric Educ",
  ["journal of agricultural education and extension"] = "J. Agric. Educ. Ext.",
  ["journal of agricultural engineering research"] = "J. Agric. Eng. Res.",
  ["journal of agricultural entomology"] = "J. Agric. Entomol.",
  ["journal of agricultural ethics"] = "J Agric Ethics",
  ["journal of agricultural meteorology"] = "J. Agric. Meteorol.",
  ["journal of agricultural research"] = "J Agric Res",
  ["journal of agricultural safety and health"] = "J Agric Saf Health",
  ["journal of agricultural science"] = "J. Agric. Sci.",
  ["journal of agricultural science (toronto, ont.)"] = "J Agric Sci (Tor)",
  ["journal of agricultural science and technology"] = "J. Agric. Sci. Technol.",
  ["journal of agricultural science and technology : jast"] = "J Agric Sci Technol",
  ["journal of agricultural science and technology. a"] = "J Agric Sci Technol A",
  ["journal of agricultural science and technology. b"] = "J Agric Sci Technol B",
  ["journal of agricultural sciences"] = "J Agric Sci (Belihuloya)",
  ["journal of agricultural, biological and environmental statistics"] = "J. Agric. Biol. Environ. Stat.",
  ["journal of agricultural, biological, and environmental statistics"] = "J Agric Biol Environ Stat",
  ["journal of agriculture and food research"] = "J Agric Food Res",
  ["journal of agriculture and forestry (new york, n.y.)"] = "J Agric For",
  ["journal of agriculture, food systems, and community development"] = "J. Agric. Food Sys. Community Dev.",
  ["journal of agroalimentary processes and technologies"] = "J. Agroaliment. Proc. Technol.",
  ["journal of agrobiology"] = "J Agrobiol",
  ["journal of agrobiotechnology"] = "J Agrobiotechnol",
  ["journal of agromedicine"] = "J Agromedicine",
  ["journal of agrometeorology"] = "J. Agrometeorol.",
  ["journal of agronomic education"] = "J. Agron. Educ.",
  ["journal of agronomy and crop science"] = "J Agron Crop Sci",
  ["journal of ahima"] = "J AHIMA",
  ["journal of ahima / american health information management association"] = "J AHIMA",
  ["journal of aids & clinical research"] = "J AIDS Clin Res",
  ["journal of aids and hiv infections"] = "J AIDS HIV Infect",
  ["journal of aids and hiv research (online)"] = "J AIDS HIV Res",
  ["journal of aids and hiv treatment"] = "J AIDS HIV Treat",
  ["journal of aids and immune research"] = "J AIDS Immune Res",
  ["journal of air medical transport"] = "J. Air Med. Transp.",
  ["journal of air transport management"] = "J Air Transp Manag",
  ["journal of air transportation (reston, va.)"] = "J Air Transp (Rest)",
  ["journal of aircraft"] = "J Aircr",
  ["journal of alcohol and drug education"] = "J Alcohol Drug Educ",
  ["journal of alcoholism and drug dependence"] = "J Alcohol Drug Depend",
  ["journal of algal biomass utilization"] = "J Algal Biomass Util",
  ["journal of algebra"] = "J. Algebra",
  ["journal of algebra and computational applications"] = "J. Algebra Comput. Appl.",
  ["journal of algebra and its applications"] = "J. Algebra Appl.",
  ["journal of algebra and related topics"] = "J. Algebra Relat. Topics",
  ["journal of algebra combinatorics discrete structures and applications"] = "J. Algebra Comb. Discrete Struct. Appl.",
  ["journal of algebraic combinatorics"] = "J. Algebraic Comb.",
  ["journal of algebraic combinatorics. an international journal"] = "J. Algebraic Combin.",
  ["journal of algebraic geometry"] = "J. Algebraic Geom.",
  ["journal of algebraic hyperstructures and logical algebras"] = "J. Algebr. Hyperstruct. Log. Algebras",
  ["journal of algebraic statistics"] = "J Algebr Stat",
  ["journal of algebraic systems"] = "J. Algebr. Syst.",
  ["journal of algorithms"] = "J. Algorithms",
  ["journal of algorithms & computational technology"] = "J. Algorithms Comput. Technol.",
  ["journal of allergy"] = "J Allergy (Cairo)",
  ["journal of allergy & therapy"] = "J Allergy Ther",
  ["journal of allergy and clinical immunology"] = "J. Allergy Clin. Immunol.",
  ["journal of allergy and infectious diseases"] = "J Allergy Infect Dis",
  ["journal of allied health"] = "J Allied Health",
  ["journal of alloys and compounds"] = "J. Alloys Compd.",
  ["journal of alloys and compounds communications"] = "J. Alloys Compd. Commun.",
  ["journal of alternative and complementary medicine"] = "J. Altern. Complement. Med.",
  ["journal of alternative and complementary medicine (new york, n.y.)"] = "J Altern Complement Med",
  ["journal of alzheimer's disease & parkinsonism"] = "J Alzheimers Dis Parkinsonism",
  ["journal of alzheimer's disease : jad"] = "J Alzheimers Dis",
  ["journal of alzheimer's disease reports"] = "J Alzheimers Dis Rep",
  ["journal of alzheimer's parkinsonism & dementia"] = "J Alzheimers Parkinsonism Dement",
  ["journal of alzheimer’s disease"] = "J. Alzheimers Dis.",
  ["journal of alzheimer’s disease : jad"] = "J Alzheimers Dis",
  ["journal of amateur sport"] = "J Amat Sport",
  ["journal of ambient intelligence and humanized computing"] = "J Ambient Intell Humaniz Comput",
  ["journal of ambient intelligence and smart environments"] = "J. Ambient Intell. Smart Environ.",
  ["journal of ambulatory care management"] = "J. Ambulatory Care Manage.",
  ["journal of ambulatory care marketing"] = "J Ambul Care Mark",
  ["journal of american college health"] = "J. Am. Coll. Health",
  ["journal of american college health : j of ach"] = "J Am Coll Health",
  ["journal of american culture"] = "J Am Cult",
  ["journal of american culture (malden, mass. : 2003)"] = "J Am Cult (Malden)",
  ["journal of american ethnic history"] = "J Am Ethn Hist",
  ["journal of american health care"] = "J. Am. Health Care",
  ["journal of american health policy"] = "J. Am. Health Policy",
  ["journal of american history (bloomington, ind.)"] = "J Am Hist",
  ["journal of american indian education"] = "J Am Indian Educ",
  ["journal of american insurance"] = "J Am Insur",
  ["journal of american studies"] = "J Am Stud",
  ["journal of amino acids"] = "J Amino Acids",
  ["journal of anaesthesiology, clinical pharmacology"] = "J Anaesthesiol Clin Pharmacol",
  ["journal of analysis"] = "J. Anal.",
  ["journal of analysis and applications"] = "J. Anal. Appl.",
  ["journal of analysis and testing"] = "J Anal Test",
  ["journal of analytical & bioanalytical techniques"] = "J Anal Bioanal Tech",
  ["journal of analytical & molecular techniques"] = "J Anal Mol Tech",
  ["journal of analytical & pharmaceutical research"] = "J Anal Pharm Res",
  ["journal of analytical and applied pyrolysis"] = "J. Anal. Appl. Pyrolysis",
  ["journal of analytical atomic spectrometry"] = "J. Anal. At. Spectrom.",
  ["journal of analytical chemistry"] = "J. Anal. Chem.",
  ["journal of analytical methods in chemistry"] = "J Anal Methods Chem",
  ["journal of analytical oncology"] = "J Anal Oncol",
  ["journal of analytical psychology"] = "J. Anal. Psychol.",
  ["journal of analytical science and technology"] = "J. Anal. Sci. Technol.",
  ["journal of analytical toxicology"] = "J Anal Toxicol",
  ["journal of anatomy"] = "J Anat",
  ["journal of anatomy and physiology"] = "J Anat Physiol",
  ["journal of ancient civilizations"] = "JAC",
  ["journal of andrology"] = "J Androl",
  ["journal of anesthesia"] = "J Anesth",
  ["journal of anesthesia & clinical research"] = "J Anesth Clin Res",
  ["journal of anesthesia & critical care : open access"] = "J Anesth Crit Care",
  ["journal of anesthesia and perioperative medicine"] = "J Anesth Perioper Med",
  ["journal of anesthesia history"] = "J Anesth Hist",
  ["journal of anesthetic research and pain medicine"] = "J Anesth Res Pain Med",
  ["journal of angiogenesis research"] = "J Angiogenes Res",
  ["journal of anhui university. natural science edition. anhui daxue xuebao. ziran kexue ban"] = "J. Anhui Univ. Nat. Sci.",
  ["journal of animal and feed sciences"] = "J Anim Feed Sci",
  ["journal of animal and plant sciences"] = "J Anim Plant Sci",
  ["journal of animal and veterinary advances : java"] = "J Anim Vet Adv",
  ["journal of animal breeding and genetics"] = "J. Anim. Breed. Genet.",
  ["journal of animal breeding and genetics = zeitschrift für tierzüchtung und züchtungsbiologie"] = "J Anim Breed Genet",
  ["journal of animal diversity"] = "J Anim Divers",
  ["journal of animal ecology"] = "J. Anim. Ecol.",
  ["journal of animal physiology and animal nutrition"] = "J Anim Physiol Anim Nutr (Berl)",
  ["journal of animal physiology and nutrition"] = "J. Anim. Physiol. Nutr.",
  ["journal of animal production"] = "J Anim Prod",
  ["journal of animal science"] = "J Anim Sci",
  ["journal of animal science (champaign, il, united states)"] = "J. Anim. Sci. (Champaign, IL, U. S.)",
  ["journal of animal science (savoy, il, united states)"] = "J. Anim. Sci. (Savoy, IL, U. S.)",
  ["journal of animal science and biotechnology"] = "J Anim Sci Biotechnol",
  ["journal of animal science and technology"] = "J Anim Sci Technol",
  ["journal of anthropological archaeology"] = "J Anthropol Archaeol",
  ["journal of anthropological research"] = "J Anthropol Res",
  ["journal of anthropological sciences"] = "J. Anthropol. Sci.",
  ["journal of anthropological sciences = rivista di antropologia : jass"] = "J Anthropol Sci",
  ["journal of anti-aging medicine"] = "J Anti Aging Med",
  ["journal of antibiotics"] = "J. Antibiot.",
  ["journal of antibiotics. ser. b."] = "J. Antibiot. [B]",
  ["journal of antimicrobial agents"] = "J Antimicrob Agents",
  ["journal of antimicrobial chemotherapy"] = "J. Antimicrob. Chemother.",
  ["journal of antivirals & antiretrovirals"] = "J Antivir Antiretrovir",
  ["journal of anxiety disorders"] = "J Anxiety Disord",
  ["journal of aoac international"] = "J AOAC Int",
  ["journal of apdsa; asian pacific dental student association"] = "J APDSA (Tokyo)",
  ["journal of apicultural research"] = "J. Apic. Res.",
  ["journal of apicultural science"] = "J Apic Sci",
  ["journal of appalachian health"] = "J Appalach Health",
  ["journal of appalachian studies"] = "J Appalach Stud",
  ["journal of appled electrochemistry"] = "J. Appl. Electrochem.",
  ["journal of applied & environmental microbiology"] = "J Appl Environ Microbiol",
  ["journal of applied analysis"] = "J. Appl. Anal.",
  ["journal of applied analysis and computation"] = "J. Appl. Anal. Comput.",
  ["journal of applied and computational topology"] = "J Appl Comput Topol",
  ["journal of applied and industrial mathematics"] = "J. Appl. Ind. Math.",
  ["journal of applied and numerical optimization"] = "J. Appl. Numer. Optim.",
  ["journal of applied animal nutrition"] = "J. Appl. Anim. Nutr.",
  ["journal of applied animal research"] = "J Appl Anim Res",
  ["journal of applied animal welfare research"] = "J. Appl. Anim. Welfare Res.",
  ["journal of applied animal welfare science"] = "J. Appl. Anim. Welfare Sci.",
  ["journal of applied animal welfare science : jaaws"] = "J Appl Anim Welf Sci",
  ["journal of applied aquaculture"] = "J. Appl. Aquacult.",
  ["journal of applied arts & health"] = "J Appl Arts Health",
  ["journal of applied bacteriology"] = "J. Appl. Bacteriol.",
  ["journal of applied behavior analysis"] = "J Appl Behav Anal",
  ["journal of applied behavioral science"] = "J. Appl. Behav. Sci.",
  ["journal of applied bioanalysis"] = "J Appl Bioanal",
  ["journal of applied biobehavioral research"] = "J Appl Biobehav Res",
  ["journal of applied biochemistry"] = "J Appl Biochem",
  ["journal of applied bioinformatics & computational biology"] = "J Appl Bioinforma Comput Biol",
  ["journal of applied biological chemistry"] = "J. Appl. Biol. Chem.",
  ["journal of applied biological sciences"] = "J. Appl. Biol. Sci.",
  ["journal of applied biology & biotechnology"] = "J Appl Biol Biotechnol",
  ["journal of applied biomaterials"] = "J. Appl. Biomater.",
  ["journal of applied biomaterials & biomechanics : jabb"] = "J Appl Biomater Biomech",
  ["journal of applied biomaterials & functional materials"] = "J Appl Biomater Funct Mater",
  ["journal of applied biomaterials : an official journal of the society for biomaterials"] = "J Appl Biomater",
  ["journal of applied biomechanics"] = "J Appl Biomech",
  ["journal of applied biomedicine"] = "J Appl Biomed",
  ["journal of applied biosciences"] = "J Appl Biosci",
  ["journal of applied biotechnology reports"] = "J. Appl. Biotechnol. Rep.",
  ["journal of applied botany and food quality"] = "Angewandte Botanik / Vereinigung fur Angewandte Botanik [und] Deutsche Gesellschaft fur Qualitatsforschung|J. Appl. Bot. Food Qual.",
  ["journal of applied botany and food quality = angewandte botanik"] = "J Appl Bot Food Qual",
  ["journal of applied business research"] = "J. Appl. Bus. Res.",
  ["journal of applied cardiology"] = "J Appl Cardiol",
  ["journal of applied chemistry"] = "J. Appl. Chem.",
  ["journal of applied chemistry & biotechnology"] = "J. Appl. Chem. Biotech.",
  ["journal of applied clinical medical physics"] = "J Appl Clin Med Phys",
  ["journal of applied clinical medical physics / american college of medical physics"] = "J Appl Clin Med Phys",
  ["journal of applied communication research : jacr"] = "J Appl Commun Res",
  ["journal of applied computer science methods"] = "J Appl Comput Sci Methods",
  ["journal of applied crystallography"] = "J. Appl. Crystallogr.",
  ["journal of applied developmental psychology"] = "J Appl Dev Psychol",
  ["journal of applied ecology"] = "J. Appl. Ecol.",
  ["journal of applied econometrics"] = "J. Appl. Econometrics",
  ["journal of applied econometrics (chichester, england)"] = "J Appl Econ (Chichester Engl)",
  ["journal of applied economics"] = "J Appl Econ",
  ["journal of applied electrochemistry"] = "J. Appl. Electrochem.",
  ["journal of applied entomology"] = "J. Appl. Entomol.",
  ["journal of applied entomology = zeitschrift für angewandte entomologie"] = "J Appl Entomol",
  ["journal of applied functional analysis"] = "J. Appl. Funct. Anal.",
  ["journal of applied genetics"] = "J Appl Genet",
  ["journal of applied geochemistry"] = "J. Appl. Geochem.",
  ["journal of applied geodesy"] = "J. Appl. Geod.",
  ["journal of applied geophysics"] = "J. Appl. Geophys.",
  ["journal of applied gerontology"] = "J. Appl. Gerontol.",
  ["journal of applied gerontology : the official journal of the southern gerontological society"] = "J Appl Gerontol",
  ["journal of applied glycoscience"] = "J. Appl. Glycosci.",
  ["journal of applied history"] = "J Appl Hist",
  ["journal of applied ichthyology"] = "J. Appl. Ichthyol.",
  ["journal of applied juvenile justice services"] = "J Appl Juv Justice Serv",
  ["journal of applied laboratory medicine"] = "J. Appl. Lab. Med.",
  ["journal of applied learning technology"] = "J Appl Learn Technol",
  ["journal of applied logic"] = "J. Appl. Logic",
  ["journal of applied logics. the ifcolog journal of logics and their applications"] = "J. Appl. Logics",
  ["journal of applied mathematical analysis and applications"] = "J. Appl. Math. Anal. Appl.",
  ["journal of applied mathematics"] = "J. Appl. Math.",
  ["journal of applied mathematics & computing"] = "J Appl Math Comput",
  ["journal of applied mathematics & decision sciences"] = "J. Appl. Math. Decis. Sci.",
  ["journal of applied mathematics & informatics"] = "J. Appl. Math. Inform.",
  ["journal of applied mathematics and computational mechanics"] = "J. Appl. Math. Comput. Mech.",
  ["journal of applied mathematics and computing"] = "J. Appl. Math. Comput.",
  ["journal of applied mathematics and informatics"] = "J. Appl. Math. Inform.",
  ["journal of applied mathematics and mechanics"] = "J. Appl. Math. Mech.",
  ["journal of applied mathematics and mechanics /  zeitschrift fur angewandte mathematik und mechanik"] = "ZAMM",
  ["journal of applied mathematics, statistics and informatics"] = "J. Appl. Math. Stat. Inform.",
  ["journal of applied measurement"] = "J Appl Meas",
  ["journal of applied mechanical engineering"] = "J Appl Mech Eng",
  ["journal of applied mechanics"] = "J. Appl. Mech.",
  ["journal of applied mechanics and technical physics"] = "J. Appl. Mech. Tech. Phys.",
  ["journal of applied medical sciences"] = "J Appl Med Sci",
  ["journal of applied metalworking"] = "J. Appl. Metalwork.",
  ["journal of applied meteorology"] = "J. Appl. Meteorol.",
  ["journal of applied meteorology and climatology"] = "J Appl Meteorol Climatol",
  ["journal of applied microbiology"] = "J Appl Microbiol",
  ["journal of applied microbiology and biochemistry"] = "J Appl Microbiol Biochem",
  ["journal of applied non-classical logics"] = "J. Appl. Non-Classical Logics",
  ["journal of applied nonlinear dynamics"] = "J. Appl. Nonlinear Dyn.",
  ["journal of applied nutrition"] = "J. Appl. Nutr.",
  ["journal of applied oral science : revista fob"] = "J Appl Oral Sci",
  ["journal of applied pharmaceutical science"] = "J Appl Pharm Sci",
  ["journal of applied philosophy"] = "J Appl Philos",
  ["journal of applied photographic engineering"] = "J. Appl. Photogr. Eng.",
  ["journal of applied phycology"] = "J. Appl. Phycol.",
  ["journal of applied physics"] = "J. Appl. Phys.",
  ["journal of applied physics (melville, ny, united states)"] = "J. Appl. Phys. (Melville, NY, U. S.)",
  ["journal of applied physiology"] = "J Appl Physiol",
  ["journal of applied physiology (bethesda, md. : 1985)"] = "J Appl Physiol (1985)",
  ["journal of applied physiology: respiratory, environmental and exercise physiology"] = "J Appl Physiol Respir Environ Exerc Physiol",
  ["journal of applied polymer science"] = "J. Appl. Polym. Sci.",
  ["journal of applied polymer science. applied polymer symposium"] = "J Appl Polym Sci Symp",
  ["journal of applied polymer science: applied polymer symposium"] = "J. Appl. Polym. Sci.: Appl. Polym. Symp.",
  ["journal of applied poultry research"] = "J. Appl. Poult. Res.",
  ["journal of applied probability"] = "J. Appl. Probab.",
  ["journal of applied probability and statistics"] = "J Appl Probab Stat",
  ["journal of applied psychology"] = "J. Appl. Psychol.",
  ["journal of applied rehabilitation counseling"] = "J Appl Rehabil Couns",
  ["journal of applied remote sensing"] = "J. Appl. Remote Sens.",
  ["journal of applied research and technology"] = "J. Appl. Res. Technol.",
  ["journal of applied research in intellectual disabilities : jarid"] = "J Appl Res Intellect Disabil",
  ["journal of applied research in memory and cognition"] = "J Appl Res Mem Cogn",
  ["journal of applied research in the community college"] = "J. Appl Res. Community Coll.",
  ["journal of applied research on medicinal and aromatic plants"] = "J. Appl. Res. Med. Aromat. Plants",
  ["journal of applied school psychology"] = "J Appl Sch Psychol",
  ["journal of applied science in southern africa : jassa : the journal of the university of zimbabwe"] = "J Appl Sci South Afr",
  ["journal of applied sciences (faisalabad, pakistan)"] = "J Appl Sci (Faisalabad)",
  ["journal of applied sciences research"] = "J Appl Sci Res",
  ["journal of applied social psychology"] = "J Appl Soc Psychol",
  ["journal of applied social science"] = "J Appl Soc Sci (Boulder)",
  ["journal of applied social science studies"] = "J. Appl. Soc. Sci. Stud.",
  ["journal of applied solution chemistry and modeling"] = "J. Appl. Solution Chem. Model.",
  ["journal of applied spectroscopy"] = "J. Appl. Spectrosc.",
  ["journal of applied spectroscopy (ussr) [translation of zhurnal prikladnoi spektroskopii]"] = "J. Appl. Spectrosc. (USSR)",
  ["journal of applied sport psychology"] = "J Appl Sport Psychol",
  ["journal of applied sports sciences"] = "J. Appl. Sports Sci.",
  ["journal of applied statistical science"] = "J Appl Statist Sci",
  ["journal of applied statistics"] = "J Appl Stat",
  ["journal of applied systems analysis"] = "J. Appl. Syst. Anal.",
  ["journal of applied toxicology"] = "J. Appl. Toxicol.",
  ["journal of applied toxicology : jat"] = "J Appl Toxicol",
  ["journal of applied volcanology"] = "J. Appl. Volcanol.",
  ["journal of applied water engineering and research"] = "J. Appl. Water Eng. Res.",
  ["journal of approximation theory"] = "J. Approx. Theory",
  ["journal of aquaculture research & development"] = "J Aquac Res Dev",
  ["journal of aquatic animal health"] = "J Aquat Anim Health",
  ["journal of aquatic ecosystem stress and recovery"] = "J. Aquat. Ecosyst. Stress Recovery",
  ["journal of aquatic food product technology"] = "J. Aquat. Food Prod. Technol.",
  ["journal of aquatic plant management"] = "J Aquat Plant Manag",
  ["journal of aquatic research and marine sciences"] = "J Aquat Res Mar Sci",
  ["journal of arab affairs"] = "J Arab Aff",
  ["journal of arachnology"] = "J. Arachnol.",
  ["journal of archaeological method and theory"] = "J Archaeol Method Theory",
  ["journal of archaeological research"] = "J. Archaeol. Res.",
  ["journal of archaeological science"] = "J Archaeol Sci",
  ["journal of archaeological science, reports"] = "J Archaeol Sci Rep",
  ["journal of archaeological science: reports"] = "J. Archaeol. Sci.: Rep.",
  ["journal of architectural and planning research"] = "J Archit Plann Res",
  ["journal of architectural conservation"] = "J. Archit. Conserv.",
  ["journal of architectural engineering"] = "J. Archit. Eng.",
  ["journal of arid environments"] = "J Arid Environ",
  ["journal of arid land"] = "J Arid Land",
  ["journal of arrhythmia"] = "J Arrhythm",
  ["journal of arthritis"] = "J Arthritis",
  ["journal of arthroplasty"] = "J. Arthroplasty",
  ["journal of arthropod-borne diseases"] = "J Arthropod Borne Dis",
  ["journal of artif. intell. research"] = "J. Artif. Intell. Res.",
  ["journal of artificial intelligence and consciousness"] = "J. Artif. Intell. Conscious",
  ["journal of artificial intelligence research"] = "J. Artif. Intell. Res.",
  ["journal of artificial organs : the official journal of the japanese society for artificial organs"] = "J Artif Organs",
  ["journal of artificial societies and social simulation : jasss"] = "J Artif Soc Soc Simul",
  ["journal of asia-pacific biodiversity"] = "J. Asia-Pac. Biodivers.",
  ["journal of asia-pacific business"] = "J. Asia-Pacific Bus.",
  ["journal of asia-pacific entomology"] = "J. Asia-Pac. Entomol.",
  ["journal of asian and african studies"] = "J Asian Afr Stud",
  ["journal of asian architecture and building engineering"] = "J. Asian Archit. Build. Eng.",
  ["journal of asian ceramic societies"] = "J. Asian Ceram. Soc.",
  ["journal of asian earth sciences"] = "J. Asian Earth Sci.",
  ["journal of asian earth sciences: x"] = "J. Asian Earth Sci.: X",
  ["journal of asian economics"] = "J Asian Econ",
  ["journal of asian natural products research"] = "J Asian Nat Prod Res",
  ["journal of asian public policy"] = "J Asian Public Policy",
  ["journal of asian studies"] = "J. Asian Stud.",
  ["journal of assisted reproduction and genetics"] = "J Assist Reprod Genet",
  ["journal of assistive technologies"] = "J Assist Technol",
  ["journal of assistive, rehabilitative and therapeutic technologies"] = "J. Assistive Rehabil. Ther. Technol.",
  ["journal of asthma"] = "J. Asthma",
  ["journal of asthma & allergy educators"] = "J Asthma Allergy Educ",
  ["journal of asthma and allergy"] = "J Asthma Allergy",
  ["journal of asthma research"] = "J. Asthma Res.",
  ["journal of astm international"] = "J. ASTM Int.",
  ["journal of astronautical sciences"] = "J. Astronaut. Sci.",
  ["journal of astronomical history and heritage"] = "J. Astron. Hist. Heritage",
  ["journal of astronomical instrumentation"] = "J. Astron. Instrum.",
  ["journal of astronomical telescopes, instruments, and systems"] = "J. Astron. Telesc. Instrum. Syst.",
  ["journal of astrophysics and astronomy"] = "J. Astrophys. Astron.",
  ["journal of atherosclerosis and thrombosis"] = "J Atheroscler Thromb",
  ["journal of atherosclerosis research"] = "J Atheroscler Res",
  ["journal of athletic enhancement"] = "J Athl Enhanc",
  ["journal of athletic training"] = "J Athl Train",
  ["journal of atmospheric & ocean science"] = "J. Atmos. Ocean Sci.",
  ["journal of atmospheric and oceanic technology"] = "J. Atmos. Oceanic Technol.",
  ["journal of atmospheric and solar-terrestrial physics"] = "J. Atmos. Sol. Terr. Phys.",
  ["journal of atmospheric and terrestrial physics"] = "J. Atmos. Terr. Phys.",
  ["journal of atmospheric chemistry"] = "J. Atmos. Chem.",
  ["journal of atmospheric science"] = "J. Atmos. Sci.",
  ["journal of atmospheric sciences"] = "J. Atmos. Sci.",
  ["journal of atomic and molecular physics"] = "J At Mol Phys",
  ["journal of atomic, molecular, and optical physics"] = "J At Mol Opt Phys",
  ["journal of atrial fibrillation"] = "J Atr Fibrillation",
  ["journal of attention disorders"] = "J Atten Disord",
  ["journal of audiology & otology"] = "J Audiol Otol",
  ["journal of audiovisual media in medicine"] = "J. Audiov. Media Med.",
  ["journal of auditory research"] = "J. Aud. Res.",
  ["journal of auditory research. supplement"] = "J Aud Res Suppl",
  ["journal of australian political economy"] = "J. Australian Polit. Economy",
  ["journal of autism and childhood schizophrenia"] = "J Autism Child Schizophr",
  ["journal of autism and developmental disorders"] = "J Autism Dev Disord",
  ["journal of autoimmune diseases"] = "J Autoimmune Dis",
  ["journal of autoimmune diseases and rheumatology"] = "J Autoimmune Dis Rheumatol",
  ["journal of autoimmune disorders"] = "J Autoimmune Disord",
  ["journal of autoimmunity"] = "J Autoimmun",
  ["journal of automata, languages and combinatorics"] = "J. Autom. Lang. Comb.",
  ["journal of automated methods & management in chemistry"] = "J Autom Methods Manag Chem",
  ["journal of automated reasoning"] = "J. Automat. Reason.",
  ["journal of automatic chemistry"] = "J. Autom. Chem.",
  ["journal of automation and information sciences"] = "J. Autom. Inf. Sci.",
  ["journal of automation and intelligence"] = "J. Autom. Intell.",
  ["journal of autonomic pharmacology"] = "J Auton Pharmacol",
  ["journal of autonomous intelligence"] = "J. Auton. Intell.",
  ["journal of autonomous vehicles and systems"] = "J. Auton. Veh. Syst.",
  ["journal of avian biology"] = "J Avian Biol",
  ["journal of avian medicine and surgery"] = "J Avian Med Surg",
  ["journal of ayub medical college, abbottabad : jamc"] = "J Ayub Med Coll Abbottabad",
  ["journal of ayurveda and integrative medicine"] = "J Ayurveda Integr Med",
  ["journal of b.u.on. : official journal of the balkan union of oncology"] = "J BUON",
  ["journal of back and musculoskeletal rehabilitation"] = "J Back Musculoskelet Rehabil",
  ["journal of bacteriology"] = "J Bacteriol",
  ["journal of bacteriology & parasitology"] = "J Bacteriol Parasitol",
  ["journal of bacteriology and mycology (monroe (middlesex county, n.j. : township))"] = "J Bacteriol Mycol (Monroe Township)",
  ["journal of bacteriology and virology : jbv"] = "J Bacteriol Virol",
  ["journal of baltic studies"] = "J Balt Stud",
  ["journal of bangladesh academy of sciences. bangladesh academy of sciences"] = "J Bangladesh Acad Sci",
  ["journal of bangladesh college of physicians & surgeons"] = "J Bangladesh Coll Phys Surg",
  ["journal of bank research"] = "J. Bank Res.",
  ["journal of banking & finance"] = "J Bank Financ",
  ["journal of banking and finance"] = "J. Banking Finance",
  ["journal of basic & applied zoology"] = "J Basic Appl Zool",
  ["journal of basic and applied mycology, egypt"] = "J Basic Appl Mycol Egypt",
  ["journal of basic and applied zoology"] = "J. Basic Appl. Zool.",
  ["journal of basic and clinical medicine"] = "J Basic Clin Med",
  ["journal of basic and clinical pharmacy"] = "J Basic Clin Pharm",
  ["journal of basic and clinical physiology and pharmacology"] = "J Basic Clin Physiol Pharmacol",
  ["journal of basic microbiology"] = "J Basic Microbiol",
  ["journal of behavior analysis in health, sports, fitness and medicine"] = "J Behav Anal Health Sports Fit Med",
  ["journal of behavior therapy and experimental psychiatry"] = "J Behav Ther Exp Psychiatry",
  ["journal of behavioral addictions"] = "J Behav Addict",
  ["journal of behavioral and brain science"] = "J Behav Brain Sci",
  ["journal of behavioral and cognitive therapy"] = "J Behav Cogn Ther",
  ["journal of behavioral and experimental economics"] = "J Behav Exp Econ",
  ["journal of behavioral and experimental finance"] = "J Behav Exp Finance",
  ["journal of behavioral and neuroscience research"] = "J Behav Neurosci Res",
  ["journal of behavioral and social sciences (cedarville, ohio)"] = "J Behav Soc Sci (Cedarv)",
  ["journal of behavioral assessment"] = "J Behav Assess",
  ["journal of behavioral decision making"] = "J Behav Decis Mak",
  ["journal of behavioral economics"] = "J. Behavioral Econ.",
  ["journal of behavioral education"] = "J Behav Educ",
  ["journal of behavioral health"] = "J Behav Health",
  ["journal of behavioral health services and research"] = "J. Behav. Health Serv. Res.",
  ["journal of behavioral medicine"] = "J Behav Med",
  ["journal of beijing institute of technology (chinese edition)"] = "J. Beijing Inst. Technol. (Chin. Ed.)",
  ["journal of beijing university of technology. beijing gongye daxue xuebao"] = "J. Beijing Univ. Technol.",
  ["journal of benefit-cost analysis"] = "J Benefit Cost Anal",
  ["journal of berry research"] = "J Berry Res",
  ["journal of best practices in health professions diversity : research, education and policy"] = "J Best Pract Health Prof Divers",
  ["journal of biblical ethics in medicine"] = "J Biblic Ethics Med",
  ["journal of biblical literature"] = "JBL",
  ["journal of big data"] = "J Big Data",
  ["journal of big data analytics in transportation"] = "J. Big Data Anal. Transp.",
  ["journal of binocular vision and ocular motility"] = "J Binocul Vis Ocul Motil",
  ["journal of bio- and tribo-corrosion"] = "J Bio Tribocorros",
  ["journal of bio-science"] = "J Biosci (Rajshari)",
  ["journal of bio-x research"] = "J BioX Res",
  ["journal of bioactive and bocompatible polymers"] = "J. Bioact. Compat. Polym.",
  ["journal of bioactive and compatable polymers"] = "J. Bioact. Compat. Polym.",
  ["journal of bioactive and compatible polymers"] = "J Bioact Compat Polym",
  ["journal of bioanalysis & biomedicine"] = "J Bioanal Biomed",
  ["journal of biobased materials and bioenergy"] = "J. Biobased Mater. Bioenergy",
  ["journal of biochemical and biophysical methods"] = "J Biochem Biophys Methods",
  ["journal of biochemical and microbiological technology and engineering"] = "J. Biochem. Microbiol. Technol. Eng.",
  ["journal of biochemical and molecular toxicology"] = "J Biochem Mol Toxicol",
  ["journal of biochemical and pharmacological research"] = "J Biochem Pharmacol Res",
  ["journal of biochemical technology"] = "J Biochem Technol",
  ["journal of biochemical toxicology"] = "J. Biochem. Toxicol.",
  ["journal of biochemistry"] = "J. Biochem.",
  ["journal of biochemistry (tokyo, japan)"] = "J. Biochem. (Tokyo, Jpn.)",
  ["journal of biochemistry and analytical studies"] = "J Biochem Anal Stud",
  ["journal of biochemistry and molecular biology"] = "J Biochem Mol Biol",
  ["journal of biochemistry and physiology"] = "J Biochem Physiol",
  ["journal of biochemistry, molecular biology, and biophysics"] = "J. Biochem. Mol. Biol. Biophys.",
  ["journal of biochemistry, molecular biology, and biophysics : jbmbb : the official journal of the federation of asian and oceanian biochemists and molecular biologists (faobmb)"] = "J Biochem Mol Biol Biophys",
  ["journal of biocommunication"] = "J. Biocommun.",
  ["journal of biodiversity"] = "J. Biodivers.",
  ["journal of biodiversity and environmental sciences"] = "J Biodivers Environ Sci",
  ["journal of bioeconomics"] = "J. Bioecon.",
  ["journal of bioenergetics"] = "J Bioenerg",
  ["journal of bioenergetics and biomembranes"] = "J Bioenerg Biomembr",
  ["journal of bioengineering"] = "J. Bioeng.",
  ["journal of bioengineering & biomedical science"] = "J Bioeng Biomed Sci",
  ["journal of bioequivalence & bioavailability"] = "J. Bioequivalence Bioavailability",
  ["journal of bioethical inquiry"] = "J Bioeth Inq",
  ["journal of bioethics"] = "J Bioeth",
  ["journal of biofertilizers & biopesticides"] = "J Biofertil Biopestic",
  ["journal of biogeography"] = "J Biogeogr",
  ["journal of bioinformatics"] = "J Bioinform",
  ["journal of bioinformatics and computational biology"] = "J Bioinform Comput Biol",
  ["journal of bioinformatics and diabetes"] = "J Bioinform Diabetes",
  ["journal of bioinformatics and genomics"] = "J Bioinform Genom",
  ["journal of bioinformatics and proteomics review"] = "J Bioinform Proteom Rev",
  ["journal of biological and chemical sciences"] = "J Biol Chem Sci",
  ["journal of biological and medical sciences"] = "J Biol Med Sci",
  ["journal of biological chemistry"] = "J. Biol. Chem.",
  ["journal of biological dynamics"] = "J Biol Dyn",
  ["journal of biological education"] = "J Biol Educ",
  ["journal of biological engineering"] = "J Biol Eng",
  ["journal of biological inorganic chemistry"] = "J. Biol. Inorg. Chem.",
  ["journal of biological inorganic chemistry : jbic : a publication of the society of biological inorganic chemistry"] = "J Biol Inorg Chem",
  ["journal of biological methods"] = "J Biol Methods",
  ["journal of biological photography"] = "J Biol Photogr",
  ["journal of biological physics"] = "J Biol Phys",
  ["journal of biological physics and chemistry : jbpc"] = "J Biol Phys Chem",
  ["journal of biological regulators and homeostatic agents"] = "J Biol Regul Homeost Agents",
  ["journal of biological research (thessalonikē, greece)"] = "J Biol Res (Thessalon)",
  ["journal of biological research italy"] = "J. Biol. Res. Italy",
  ["journal of biological research-thessaloniki"] = "J. Biol. Res. Thessaloniki",
  ["journal of biological response modifiers"] = "J Biol Response Mod",
  ["journal of biological rhythms"] = "J Biol Rhythms",
  ["journal of biological sciences (faisalābād, pakistan)"] = "J Biol Sci (Faisalabad)",
  ["journal of biological standardization"] = "J Biol Stand",
  ["journal of biological systems"] = "J Biol Syst",
  ["journal of biological, pharmaceutical and chemical research"] = "J Biol Pharm Chem Res",
  ["journal of biologically active products from nature"] = "J. Biol. Act. Prod. Nat.",
  ["journal of biology"] = "J Biol",
  ["journal of biology and nature"] = "J Biol Nat",
  ["journal of biology, agriculture and healthcare"] = "J Biol Agric Healthc",
  ["journal of bioluminescence and chemiluminescence"] = "J Biolumin Chemilumin",
  ["journal of biomarkers"] = "J Biomark",
  ["journal of biomarkers in drug development"] = "J Biomark Drug Dev",
  ["journal of biomaterials and nanobiotechnology"] = "J Biomater Nanobiotechnol",
  ["journal of biomaterials and tissue engineering"] = "J Biomater Tissue Eng",
  ["journal of biomaterials applications"] = "J Biomater Appl",
  ["journal of biomaterials science polymer edition"] = "J. Biomater. Sci., Polym. Ed.",
  ["journal of biomaterials science, polymer edition"] = "J. Biomater. Sci., Polym. Ed.",
  ["journal of biomaterials science. polymer edition"] = "J Biomater Sci Polym Ed",
  ["journal of biomathematics"] = "J. Biomath.",
  ["journal of biomathematics. shengwu shuxue xuebao"] = "J. Biomath.",
  ["journal of biomechanical engineering"] = "J Biomech Eng",
  ["journal of biomechanics"] = "J Biomech",
  ["journal of biomedial materials research"] = "J. Biomed. Mater. Res.",
  ["journal of biomedial materials research part a"] = "J. Biomed. Mater. Res. Part A",
  ["journal of biomedial materials research part b applied biomaterials"] = "J. Biomed. Mater. Res. Part B",
  ["journal of biomedial materials researchpart a"] = "J. Biomed. Mater. Res. Part A",
  ["journal of biomedial optics"] = "J. Biomed. Opt.",
  ["journal of biomedical and allied research"] = "J Biomed Allied Res",
  ["journal of biomedical and laboratory sciences"] = "J Biomed Lab Sci",
  ["journal of biomedical discovery and collaboration"] = "J Biomed Discov Collab",
  ["journal of biomedical education"] = "J Biomed Educ",
  ["journal of biomedical engineering"] = "J. Biomed. Eng.",
  ["journal of biomedical engineering and informatics"] = "J Biomed Eng Inform",
  ["journal of biomedical engineering and medical imaging"] = "J Biomed Eng Med Imaging",
  ["journal of biomedical graphics and computing"] = "J Biomed Graph Comput",
  ["journal of biomedical informatics"] = "J Biomed Inform",
  ["journal of biomedical informatics: x"] = "J Biomed Inform X",
  ["journal of biomedical materials research"] = "J Biomed Mater Res",
  ["journal of biomedical materials research part a"] = "J. Biomed. Mater. Res. Part A",
  ["journal of biomedical materials research part b"] = "J. Biomed. Mater. Res. Part B",
  ["journal of biomedical materials research part b applied biomaterials"] = "J. Biomed. Mater. Res. Part B",
  ["journal of biomedical materials research, part a"] = "J. Biomed. Mater. Res., Part A",
  ["journal of biomedical materials research, part b: applied biomaterials"] = "J. Biomed. Mater. Res., Part B",
  ["journal of biomedical materials research. part a"] = "J Biomed Mater Res A",
  ["journal of biomedical materials research. part b, applied biomaterials"] = "J Biomed Mater Res B Appl Biomater",
  ["journal of biomedical nanotechnology"] = "J Biomed Nanotechnol",
  ["journal of biomedical optics"] = "J Biomed Opt",
  ["journal of biomedical photonics & engineering"] = "J Biomed Photonics Eng",
  ["journal of biomedical physics & engineering"] = "J Biomed Phys Eng",
  ["journal of biomedical research"] = "J Biomed Res",
  ["journal of biomedical science"] = "J Biomed Sci",
  ["journal of biomedical science and engineering"] = "J. Biomed. Sci. Eng.",
  ["journal of biomedical semantics"] = "J Biomed Semantics",
  ["journal of biomedical technology and research"] = "J Biomed Technol Res",
  ["journal of biomedicine & biotechnology"] = "J Biomed Biotechnol",
  ["journal of biomedicine (sydney, nsw)"] = "J Biomed (Syd)",
  ["journal of biomedicine and biotechnology"] = "J. Biomed. Biotechnol.",
  ["journal of biometrics & biostatistics"] = "J Biom Biostat",
  ["journal of biomimetics biomaterials and tissue engineering"] = "J. Biomimetics Biomater. Tissue Eng.",
  ["journal of biomimetics, biomaterials, and tissue engineering"] = "J Biomim Biomater Tissue Eng",
  ["journal of biomolecular nmr"] = "J Biomol NMR",
  ["journal of biomolecular research & therapeutics"] = "J Biomol Res Ther",
  ["journal of biomolecular screening"] = "J. Biomol. Screening",
  ["journal of biomolecular screening : the official journal of the society for biomolecular screening"] = "J Biomol Screen",
  ["journal of biomolecular structure & dynamics"] = "J Biomol Struct Dyn",
  ["journal of biomolecular structure and dynamics"] = "J. Biomol. Struct. Dyn.",
  ["journal of biomolecular techniques : jbt"] = "J Biomol Tech",
  ["journal of bionanoscience"] = "J. Bionanosci.",
  ["journal of bionic engineering"] = "J. Bionic Eng.",
  ["journal of biopharmaceutical statistics"] = "J Biopharm Stat",
  ["journal of biophotonics"] = "J Biophotonics",
  ["journal of biophysics (hindawi publishing corporation : online)"] = "J Biophys",
  ["journal of bioprocessing & biotechniques"] = "J Bioprocess Biotech",
  ["journal of bioremediation & biodegradation"] = "J Bioremediat Biodegrad",
  ["journal of bioresources and bioproducts"] = "J. Bioresour. Bioprod.",
  ["journal of biorheology"] = "J. Biorheol.",
  ["journal of biosafety and biosecurity"] = "J Biosaf Biosecur",
  ["journal of bioscience and bioengineering"] = "J Biosci Bioeng",
  ["journal of bioscience and biotechnology discovery"] = "J Biosci Biotechnol Discov",
  ["journal of biosciences"] = "J Biosci",
  ["journal of biosciences and medicines"] = "J Biosci Med (Irvine)",
  ["journal of biosensors & bioelectronics"] = "J Biosens Bioelectron",
  ["journal of biosensors, biomarkers and diagnostics"] = "J Biosens Biomark Diagn",
  ["journal of biosocial science"] = "J Biosoc Sci",
  ["journal of biosocial science. supplement"] = "J Biosoc Sci Suppl",
  ["journal of biosystems engineering"] = "J. Biosyst. Eng.",
  ["journal of biotech research"] = "J Biotech Res",
  ["journal of biotechnology"] = "J Biotechnol",
  ["journal of biotechnology & biomaterials"] = "J Biotechnol Biomater",
  ["journal of biotechnology: x"] = "J. Biotechnol.: X",
  ["journal of bioterrorism & biodefense"] = "J Bioterror Biodef",
  ["journal of bisexuality"] = "J Bisex",
  ["journal of black health perspectives"] = "J Black Health Perspect",
  ["journal of black sexuality and relationships"] = "J Black Sex Relatsh",
  ["journal of black studies"] = "J Black Stud",
  ["journal of blockchain and intelligent computing"] = "J. Blockchain Intell. Comput.",
  ["journal of blockchain research"] = "J. Blockchain Res.",
  ["journal of blood & lymph"] = "J Blood Lymph",
  ["journal of blood disorders & transfusion"] = "J Blood Disord Transfus",
  ["journal of blood medicine"] = "J Blood Med",
  ["journal of blood transfusion"] = "J Blood Transfus",
  ["journal of bodywork and movement therapies"] = "J Bodyw Mov Ther",
  ["journal of bone and joint infection"] = "J Bone Jt Infect",
  ["journal of bone and joint surgery. american volume"] = "J. Bone Joint Surg. Am.",
  ["journal of bone and joint surgery. british volume"] = "J. Bone Joint Surg. Br.",
  ["journal of bone and mineral metabolism"] = "J Bone Miner Metab",
  ["journal of bone and mineral research"] = "J. Bone Miner. Res.",
  ["journal of bone and mineral research : the official journal of the american society for bone and mineral research"] = "J Bone Miner Res",
  ["journal of bone marrow research"] = "J Bone Marrow Res",
  ["journal of bone metabolism"] = "J Bone Metab",
  ["journal of bone oncology"] = "J Bone Oncol",
  ["journal of bone reports & recommendations"] = "J Bone Rep Recomm",
  ["journal of bone research"] = "J Bone Res",
  ["journal of borderlands studies"] = "J Borderl Stud",
  ["journal of botany"] = "J Bot",
  ["journal of botany (faisalabad)"] = "J Bot (Faisalabad)",
  ["journal of brachial plexus and peripheral nerve injury"] = "J Brachial Plex Peripher Nerve Inj",
  ["journal of brain disease"] = "J Brain Dis",
  ["journal of breast cancer"] = "J Breast Cancer",
  ["journal of breast imaging"] = "J Breast Imaging",
  ["journal of breath research"] = "J Breath Res",
  ["journal of bridge engineering"] = "J. Bridge Eng.",
  ["journal of broadcasting & electronic media"] = "J Broadcast Electron Media",
  ["journal of bronchology & interventional pulmonology"] = "J Bronchology Interv Pulmonol",
  ["journal of bryology"] = "J Bryol",
  ["journal of buddhist ethics"] = "J Buddh Ethics",
  ["journal of buffalo science"] = "J Buffalo Sci",
  ["journal of building engineering"] = "J. Build. Eng.",
  ["journal of building pathology and rehabilitation"] = "J. Build. Pathol. Rehabil.",
  ["journal of building performance simulation"] = "J. Build. Perform. Simul.",
  ["journal of building physics"] = "J. Build. Phys.",
  ["journal of buon"] = "J. Buon",
  ["journal of burn care & research : official publication of the american burn association"] = "J Burn Care Res",
  ["journal of burn care and rehabilitation"] = "J. Burn Care Rehabil.",
  ["journal of burn care and research"] = "J. Burn Care Res.",
  ["journal of burns and wounds"] = "J Burns Wounds",
  ["journal of business"] = "J. Bus.",
  ["journal of business & economic statistics"] = "J. Bus. Econom. Statist.",
  ["journal of business & economic statistics : a publication of the american statistical association"] = "J Bus Econ Stat",
  ["journal of business (chicago, ill.)"] = "J Bus",
  ["journal of business and economic statistics"] = "J. Bus. Econ. Statist.",
  ["journal of business and psychology"] = "J Bus Psychol",
  ["journal of business and technical communication"] = "J Bus Tech Commun",
  ["journal of business continuity & emergency planning"] = "J Bus Contin Emer Plan",
  ["journal of business ethics : jbe"] = "J Bus Ethics",
  ["journal of business finance & accounting"] = "J Bus Finance Account",
  ["journal of business research"] = "J Bus Res",
  ["journal of business strategy"] = "J. Bus. Strategy",
  ["journal of business venturing"] = "J Bus Ventur",
  ["journal of cachexia, sarcopenia and muscle"] = "J Cachexia Sarcopenia Muscle",
  ["journal of caffeine and adenosine research"] = "J Caffeine Adenosine Res",
  ["journal of caffeine research"] = "J Caffeine Res",
  ["journal of california and great basin anthropology"] = "J Calif Gt Basin Anthropol",
  ["journal of camel practice and research"] = "J. Camel Pract. Res.",
  ["journal of canadian petroleum technology"] = "J. Can. Pet. Technol.",
  ["journal of canadian studies. revue d'études canadiennes"] = "J Can Stud",
  ["journal of canadian studies. revue d’etudes canadiennes"] = "J Can Stud",
  ["journal of cancer"] = "J Cancer",
  ["journal of cancer and clinical oncology"] = "J Cancer Clin Oncol",
  ["journal of cancer biology"] = "J Cancer Biol",
  ["journal of cancer biology & research"] = "J Cancer Biol Res",
  ["journal of cancer clinical trials"] = "J Cancer Clin Trials",
  ["journal of cancer education"] = "J. Cancer Educ.",
  ["journal of cancer education : the official journal of the american association for cancer education"] = "J Cancer Educ",
  ["journal of cancer epidemiology"] = "J Cancer Epidemiol",
  ["journal of cancer epidemiology & treatment"] = "J Cancer Epidemiol Treat",
  ["journal of cancer epidemiology and prevention"] = "J Cancer Epidemiol Prev",
  ["journal of cancer epidemiology and prevention (imedpub)"] = "J Cancer Epidemiol Prev (iMedPub)",
  ["journal of cancer immunology"] = "J Cancer Immunol (Wilmington)",
  ["journal of cancer metastasis and treatment"] = "J Cancer Metastasis Treat",
  ["journal of cancer molecules"] = "J. Cancer Mol.",
  ["journal of cancer policy"] = "J Cancer Policy",
  ["journal of cancer prevention"] = "J Cancer Prev",
  ["journal of cancer prevention & current research"] = "J Cancer Prev Curr Res",
  ["journal of cancer rehabilitation"] = "J Cancer Rehabil",
  ["journal of cancer research & therapy"] = "J Cancer Res Ther (Manch)",
  ["journal of cancer research and cellular therapeutics"] = "J Cancer Res Cell Ther",
  ["journal of cancer research and clinical oncology"] = "J Cancer Res Clin Oncol",
  ["journal of cancer research and oncobiology"] = "J Cancer Res Oncobiol",
  ["journal of cancer research and therapeutic oncology"] = "J Cancer Res Ther Oncol",
  ["journal of cancer research and therapeutics"] = "J Cancer Res Ther",
  ["journal of cancer research forecast"] = "J Cancer Res Forecast",
  ["journal of cancer research updates"] = "J Can Res Updates",
  ["journal of cancer science & therapy"] = "J Cancer Sci Ther",
  ["journal of cancer science and clinical therapeutics"] = "J Cancer Sci Clin Ther",
  ["journal of cancer sciences"] = "J Cancer Sci",
  ["journal of cancer stem cell research"] = "J Cancer Stem Cell Res",
  ["journal of cancer survivorship : research and practice"] = "J Cancer Surviv",
  ["journal of cancer therapeutics & research"] = "J Cancer Ther Res",
  ["journal of cancer therapy"] = "J Cancer Ther",
  ["journal of cancer treatment & diagnosis"] = "J Cancer Treatment Diagn",
  ["journal of cancer treatment and research"] = "J Cancer Treat Res",
  ["journal of cannabis research"] = "J Cannabis Res",
  ["journal of cannabis therapeutics"] = "J. Cannabis Ther.",
  ["journal of capillary electrophoresis"] = "J Capillary Electrophor",
  ["journal of capillary electrophoresis and microchip technology"] = "J Capill Electrophor Microchip Technol",
  ["journal of carbohydrate chemistry"] = "J. Carbohydr. Chem.",
  ["journal of carcinogenesis"] = "J Carcinog",
  ["journal of carcinogenesis & mutagenesis"] = "J Carcinog Mutagen",
  ["journal of cardiac failure"] = "J Card Fail",
  ["journal of cardiac surgery"] = "J Card Surg",
  ["journal of cardio-respiratory diseases"] = "J Cardiorespir Dis",
  ["journal of cardiobiology"] = "J Cardiobiol",
  ["journal of cardiography"] = "J Cardiogr",
  ["journal of cardiography. supplement"] = "J Cardiogr Suppl",
  ["journal of cardiology"] = "J Cardiol",
  ["journal of cardiology & cardiovascular therapy"] = "J Cardiol Cardiovasc Ther",
  ["journal of cardiology & clinical research"] = "J Cardiol Clin Res",
  ["journal of cardiology & current research"] = "J Cardiol Curr Res",
  ["journal of cardiology and cardiovascular medicine"] = "J Cardiol Cardiovasc Med",
  ["journal of cardiology and cardiovascular sciences"] = "J Cardiol Cardiovasc Sci",
  ["journal of cardiology and therapeutics"] = "J Cardiol Ther",
  ["journal of cardiology and therapy"] = "J Cardiol Ther (Hong Kong)",
  ["journal of cardiology cases"] = "J Cardiol Cases",
  ["journal of cardiology. supplement"] = "J Cardiol Suppl",
  ["journal of cardiopulmonary rehabilitation"] = "J Cardiopulm Rehabil",
  ["journal of cardiopulmonary rehabilitation and prevention"] = "J Cardiopulm Rehabil Prev",
  ["journal of cardiothoracic and vascular anesthesia"] = "J Cardiothorac Vasc Anesth",
  ["journal of cardiothoracic anesthesia"] = "J Cardiothorac Anesth",
  ["journal of cardiothoracic surgery"] = "J Cardiothorac Surg",
  ["journal of cardiovascular and thoracic research"] = "J Cardiovasc Thorac Res",
  ["journal of cardiovascular computed tomography"] = "J Cardiovasc Comput Tomogr",
  ["journal of cardiovascular development and disease"] = "J Cardiovasc Dev Dis",
  ["journal of cardiovascular diagnosis and procedures"] = "J Cardiovasc Diagn Proced",
  ["journal of cardiovascular disease"] = "J Cardiovasc Dis",
  ["journal of cardiovascular disease research"] = "J Cardiovasc Dis Res",
  ["journal of cardiovascular diseases & diagnosis"] = "J Cardiovasc Dis Diagn",
  ["journal of cardiovascular echography"] = "J Cardiovasc Echogr",
  ["journal of cardiovascular electrophysiology"] = "J Cardiovasc Electrophysiol",
  ["journal of cardiovascular imaging"] = "J Cardiovasc Imaging",
  ["journal of cardiovascular magnetic resonance"] = "J. Cardiovasc. Magn. Reson.",
  ["journal of cardiovascular magnetic resonance : official journal of the society for cardiovascular magnetic resonance"] = "J Cardiovasc Magn Reson",
  ["journal of cardiovascular management"] = "J. Cardiovasc. Manag.",
  ["journal of cardiovascular medicine"] = "J. Cardiovasc. Med.",
  ["journal of cardiovascular medicine (hagerstown, md.)"] = "J Cardiovasc Med (Hagerstown)",
  ["journal of cardiovascular nursing"] = "J. Cardiovasc. Nurs.",
  ["journal of cardiovascular pharmacology"] = "J Cardiovasc Pharmacol",
  ["journal of cardiovascular pharmacology and therapeutics"] = "J Cardiovasc Pharmacol Ther",
  ["journal of cardiovascular risk"] = "J Cardiovasc Risk",
  ["journal of cardiovascular surgery"] = "J. Cardiovasc. Surg. (Torino)",
  ["journal of cardiovascular translational research"] = "J Cardiovasc Transl Res",
  ["journal of cardiovascular ultrasound"] = "J Cardiovasc Ultrasound",
  ["journal of care services management"] = "J Care Serv Manag",
  ["journal of career assessment"] = "J Career Assess",
  ["journal of career development"] = "J Career Dev",
  ["journal of caring sciences"] = "J Caring Sci",
  ["journal of case management"] = "J Case Manag",
  ["journal of case reports in medicine"] = "J Case Rep Med",
  ["journal of catalysis"] = "J. Catal.",
  ["journal of catalysts"] = "J. Catal.",
  ["journal of cataract and refractive surgery"] = "J Cataract Refract Surg",
  ["journal of causal inference"] = "J Causal Inference",
  ["journal of cave and karst studies : the national speleological society bulletin"] = "J Caves Karst Stud",
  ["journal of cell biology"] = "J. Cell Biol.",
  ["journal of cell biology and histology"] = "J Cell Biol Histol",
  ["journal of cell communication and signaling"] = "J Cell Commun Signal",
  ["journal of cell death"] = "J Cell Death",
  ["journal of cell science"] = "J Cell Sci",
  ["journal of cell science & therapy"] = "J Cell Sci Ther",
  ["journal of cell science. supplement"] = "J Cell Sci Suppl",
  ["journal of cell signaling"] = "J Cell Signal",
  ["journal of cellular and comparative physiology"] = "J Cell Comp Physiol",
  ["journal of cellular and molecular medicine"] = "J Cell Mol Med",
  ["journal of cellular automata"] = "J. Cell. Autom.",
  ["journal of cellular biochemistry"] = "J Cell Biochem",
  ["journal of cellular biochemistry. supplement"] = "J Cell Biochem Suppl",
  ["journal of cellular biology & molecular sciences"] = "J Cell Biol Mol Sci",
  ["journal of cellular biotechnology"] = "J Cell Biotechnol",
  ["journal of cellular immunology"] = "J Cell Immunol",
  ["journal of cellular immunotherapy"] = "J. Cell. Immunother.",
  ["journal of cellular physiology"] = "J Cell Physiol",
  ["journal of cellular physiology. supplement"] = "J Cell Physiol Suppl",
  ["journal of cellular plastics"] = "J. Cell. Plast.",
  ["journal of cellular signaling"] = "J Cell Signal",
  ["journal of central china normal university"] = "J. Central China Normal Univ. Natur. Sci.",
  ["journal of central china normal university. natural sciences"] = "J. Cent. China Norm. Univ. Nat. Sci.",
  ["journal of central nervous system disease"] = "J Cent Nerv Syst Dis",
  ["journal of central south university"] = "J Cent South Univ",
  ["journal of ceramic processing research"] = "J. Ceram. Process. Res.",
  ["journal of ceramic science and technology"] = "J. Ceram. Sci. Technol.",
  ["journal of ceramics"] = "J. Ceram.",
  ["journal of cereal science"] = "J. Cereal Sci.",
  ["journal of cerebral blood flow & metabolism"] = "J. Cereb. Blood Flow Metab.",
  ["journal of cerebral blood flow and metabolism"] = "J. Cereb. Blood Flow Metab.",
  ["journal of cerebral blood flow and metabolism : official journal of the international society of cerebral blood flow and metabolism"] = "J Cereb Blood Flow Metab",
  ["journal of cerebrovascular and endovascular neurosurgery"] = "J Cerebrovasc Endovasc Neurosurg",
  ["journal of character education"] = "J Character Educ",
  ["journal of chemical & engineering data"] = "J. Chem. Eng. Data",
  ["journal of chemical and engineering data"] = "J. Chem. Eng. Data",
  ["journal of chemical and pharmaceutical research"] = "J Chem Pharm Res",
  ["journal of chemical biology"] = "J Chem Biol",
  ["journal of chemical crystallography"] = "J. Chem. Crystallogr.",
  ["journal of chemical documentation"] = "J. Chem. Doc.",
  ["journal of chemical ecology"] = "J Chem Ecol",
  ["journal of chemical education"] = "J. Chem. Educ.",
  ["journal of chemical engineering and chemistry research"] = "J Chem Eng Chem Res",
  ["journal of chemical engineering of japan"] = "J. Chem. Eng. Jpn.",
  ["journal of chemical health & safety"] = "J Chem Health Saf",
  ["journal of chemical heath and safetyu"] = "J. Chem. Health Saf.",
  ["journal of chemical information and computer science"] = "J. Chem. Inf. Comput. Sci.",
  ["journal of chemical information and computer sciences"] = "J Chem Inf Comput Sci",
  ["journal of chemical information and modeling"] = "J Chem Inf Model",
  ["journal of chemical metrology"] = "J. Chem. Metrol.",
  ["journal of chemical neuroanatomy"] = "J Chem Neuroanat",
  ["journal of chemical physics"] = "J. Chem. Phys.",
  ["journal of chemical research"] = "J. Chem. Res.",
  ["journal of chemical research synopsis"] = "J. Chem. Res., Synop .",
  ["journal of chemical research. synopses"] = "J Chem Res Synop",
  ["journal of chemical sciences"] = "J. Chem. Sci.",
  ["journal of chemical sciences (bangalore, india)"] = "J Chem Sci (Bangalore)",
  ["journal of chemical sciences (berlin, germany)"] = "J. Chem. Sci. (Berlin, Ger.)",
  ["journal of chemical software"] = "J. Chem. Software",
  ["journal of chemical technology & biotechnology"] = "J. Chem. Technol. Biotechnol.",
  ["journal of chemical technology and biotechnology"] = "J. Chem. Technol. Biotechnol.",
  ["journal of chemical technology and biotechnology (oxford, oxfordshire : 1986)"] = "J Chem Technol Biotechnol",
  ["journal of chemical technology and biotechnology. biotechnology"] = "J. Chem. Technol. Biotechnol. Biotechnol.",
  ["journal of chemical technology and biotechnology. chemical technology"] = "J. Chem. Technol. Biotechnol. Chem. Technol.",
  ["journal of chemical technology and metallurgy"] = "J. Chem. Technol. Metall.",
  ["journal of chemical theory and computation"] = "J Chem Theory Comput",
  ["journal of chemical thermodynamics"] = "J. Chem. Thermodyn.",
  ["journal of chemical, biological, and physical sciences"] = "J Chem Biol Phys Sci",
  ["journal of cheminformatics"] = "J Cheminform",
  ["journal of chemistry"] = "J Chem",
  ["journal of chemistry and chemical engineering"] = "J Chem Chem Eng",
  ["journal of chemometrics"] = "J. Chemom.",
  ["journal of chemotherapy"] = "J. Chemother.",
  ["journal of chemotherapy (florence, italy)"] = "J Chemother",
  ["journal of chengdu university"] = "J. Chengdu Univ. Natur. Sci.",
  ["journal of chengdu university of science and technology"] = "J. Chengdu Univ. Sci. Tech.",
  ["journal of cherokee studies"] = "J Cherokee Stud",
  ["journal of chest surgery"] = "J Chest Surg",
  ["journal of child & adolescent substance abuse"] = "J Child Adolesc Subst Abuse",
  ["journal of child & adolescent trauma"] = "J Child Adolesc Trauma",
  ["journal of child and adolescent behavior"] = "J Child Adolesc Behav",
  ["journal of child and adolescent counseling"] = "J Child Adolesc Couns",
  ["journal of child and adolescent mental health"] = "J Child Adolesc Ment Health",
  ["journal of child and adolescent psychiatric and mental health nursing"] = "J Child Adolesc Psychiatr Ment Health Nurs",
  ["journal of child and adolescent psychiatric nursing"] = "J. Child Adolesc. Psychiatr. Nurs.",
  ["journal of child and adolescent psychiatric nursing : official publication of the association of child and adolescent psychiatric nurses, inc"] = "J Child Adolesc Psychiatr Nurs",
  ["journal of child and adolescent psychopharmacology"] = "J Child Adolesc Psychopharmacol",
  ["journal of child and family nursing"] = "J Child Fam Nurs",
  ["journal of child and family studies"] = "J Child Fam Stud",
  ["journal of child custody"] = "J Child Custody",
  ["journal of child health care : for professionals working with children in the hospital and community"] = "J Child Health Care",
  ["journal of child language"] = "J Child Lang",
  ["journal of child language acquisition and development"] = "J Child Lang Acquis Dev",
  ["journal of child neurology"] = "J Child Neurol",
  ["journal of child nutrition & management (online)"] = "J Child Nutr Manag",
  ["journal of child psychology and psychiatry and allied disciplines"] = "J. Child Psychol. Psychiatry",
  ["journal of child psychology and psychiatry, and allied disciplines"] = "J Child Psychol Psychiatry",
  ["journal of child science"] = "J. Child Sci.",
  ["journal of child sexual abuse"] = "J Child Sex Abus",
  ["journal of childhood & developmental disorders"] = "J Child Dev Disord",
  ["journal of childhood obesity"] = "J Child Obes",
  ["journal of children & poverty"] = "J Child Poverty",
  ["journal of children and media"] = "J Child Media",
  ["journal of children's orthopaedics"] = "J Child Orthop",
  ["journal of children's services"] = "J Child Serv",
  ["journal of china university of mining and technology"] = "J. China Univ. Min. Technol.",
  ["journal of china university of science and technology"] = "J. China Univ. Sci. Tech.",
  ["journal of chinese political science"] = "J Chin Polit Sci",
  ["journal of chiropractic humanities"] = "J Chiropr Humanit",
  ["journal of chiropractic medicine"] = "J Chiropr Med",
  ["journal of christian nursing"] = "J. Christ. Nurs.",
  ["journal of christian nursing : a quarterly publication of nurses christian fellowship"] = "J Christ Nurs",
  ["journal of chromatographic science"] = "J Chromatogr Sci",
  ["journal of chromatography"] = "J. Chromatogr.",
  ["journal of chromatography & separation techniques"] = "J Chromatogr Sep Tech",
  ["journal of chromatography a"] = "J. Chromatogr. A",
  ["journal of chromatography and separation techniques"] = "J. Chromatogr. Sep. Tech.",
  ["journal of chromatography b"] = "J. Chromatogr. B",
  ["journal of chromatography b, biomedical applications"] = "J. Chromatogr. B Biomed. Appl.",
  ["journal of chromatography b: analytical technologies in the biomedical and life sciences"] = "J. Chromatogr. B: Anal. Technol. Biomed. Life Sci.",
  ["journal of chromatography open"] = "J. Chromatogr. Open",
  ["journal of chromatography, a"] = "J. Chromatogr., A",
  ["journal of chromatography, b: analytical technologies in the biomedical and life sciences"] = "J. Chromatogr., B: Anal. Technol. Biomed. Life Sci.",
  ["journal of chromatography. a"] = "J Chromatogr A",
  ["journal of chromatography. b, analytical technologies in the biomedical and life sciences"] = "J Chromatogr B Analyt Technol Biomed Life Sci",
  ["journal of chromatography. b, biomedical applications"] = "J Chromatogr B Biomed Appl",
  ["journal of chromatography. b, biomedical sciences and applications"] = "J Chromatogr B Biomed Sci Appl",
  ["journal of chromatography. supplementary volume"] = "J Chromatogr Suppl",
  ["journal of chronic diseases"] = "J Chronic Dis",
  ["journal of chronic diseases and management"] = "J Chronic Dis Manag",
  ["journal of chronic fatigue syndrome"] = "J Chronic Fatigue Syndr",
  ["journal of circadian rhythms"] = "J Circadian Rhythms",
  ["journal of circuits, systems and computers"] = "J. Circuits, Syst. Comput.",
  ["journal of circulating biomarkers"] = "J Circ Biomark",
  ["journal of citrus pathology"] = "J Citrus Pathol",
  ["journal of civil engineering and management"] = "J. Civ. Eng. Manage.",
  ["journal of civil structural health monitoring"] = "J Civ Struct Health Monit",
  ["journal of civl engineering education"] = "J. Civ. Eng. Educ.",
  ["journal of classical analysis"] = "J. Class. Anal.",
  ["journal of classical studies: the journal of the class. soc. of japan"] = "JCS",
  ["journal of classification"] = "J. Classification",
  ["journal of clean energy and energy storage"] = "J. Clean Energy Energy Storage",
  ["journal of cleaner production"] = "J. Cleaner Prod.",
  ["journal of cleft lip palate and craniofacial anomalies"] = "J Cleft Lip Palate Craniofacial Anomalies",
  ["journal of climate"] = "J Clim",
  ["journal of climate action, research, and policy"] = "J. Clim. Action Res. Policy",
  ["journal of climate and applied meteorology"] = "J. Clim. Appl. Meteorol.",
  ["journal of climate change"] = "J. Clim. Change",
  ["journal of climate change and health"] = "J. Clim. Change Health",
  ["journal of climate meteorology"] = "J. Clim. Meteorol.",
  ["journal of climate resilience & climate justice"] = "J. Clim. Resilience Clim. Justice",
  ["journal of climatology"] = "J. Climatol.",
  ["journal of clinical & anatomic pathology (jcap)"] = "J Clin Anat Pathol (JCAP)",
  ["journal of clinical & cellular immunology"] = "J Clin Cell Immunol",
  ["journal of clinical & experimental cardiology"] = "J Clin Exp Cardiolog",
  ["journal of clinical & experimental dermatology research"] = "J Clin Exp Dermatol Res",
  ["journal of clinical & experimental immunology"] = "J Clin Exp Immunol",
  ["journal of clinical & experimental nephrology"] = "J Clin Exp Nephrol",
  ["journal of clinical & experimental oncology"] = "J Clin Exp Oncol",
  ["journal of clinical & experimental ophthalmology"] = "J Clin Exp Ophthalmol",
  ["journal of clinical & experimental pathology"] = "J Clin Exp Pathol",
  ["journal of clinical & laboratory immunology"] = "J Clin Lab Immunol",
  ["journal of clinical & translational endocrinology"] = "J Clin Transl Endocrinol",
  ["journal of clinical & translational endocrinology: case reports"] = "J. Clin. Transl. Endocrinol.: Case Rep.",
  ["journal of clinical and diagnostic research : jcdr"] = "J Clin Diagn Res",
  ["journal of clinical and experimental dentistry"] = "J Clin Exp Dent",
  ["journal of clinical and experimental hematopathology"] = "J. Clin. Exp. Hematop.",
  ["journal of clinical and experimental hematopathology : jceh"] = "J Clin Exp Hematop",
  ["journal of clinical and experimental hepatology"] = "J Clin Exp Hepatol",
  ["journal of clinical and experimental neuropsychology"] = "J Clin Exp Neuropsychol",
  ["journal of clinical and experimental neuropsychology : official journal of the international neuropsychological society"] = "J Clin Exp Neuropsychol",
  ["journal of clinical and experimental psychopathology"] = "J Clin Exp Psychopathol",
  ["journal of clinical and experimental psychopathology & quarterly review of psychiatry and neurology"] = "J Clin Exp Psychopathol Q Rev Psychiatry Neurol",
  ["journal of clinical and experimental research in cardiology"] = "J Clin Exp Res Cardiol",
  ["journal of clinical and hospital pharmacy"] = "J Clin Hosp Pharm",
  ["journal of clinical and investigative dermatology"] = "J Clin Investig Dermatol",
  ["journal of clinical and laboratory immunology"] = "J. Clin. Lab. Immunol.",
  ["journal of clinical and laboratory medicine"] = "J Clin Lab Med",
  ["journal of clinical and molecular endocrinology"] = "J Clin Mol Endocrinol",
  ["journal of clinical and preventive cardiology"] = "J Clin Prev Cardiol",
  ["journal of clinical and translational endocrinology case reports"] = "J Clin Transl Endocrinol Case Rep",
  ["journal of clinical and translational hepatology"] = "J Clin Transl Hepatol",
  ["journal of clinical and translational research"] = "J Clin Transl Res",
  ["journal of clinical and translational science"] = "J Clin Transl Sci",
  ["journal of clinical anesthesia"] = "J Clin Anesth",
  ["journal of clinical anesthesia and management"] = "J Clin Anesth Manag",
  ["journal of clinical anesthesia and pain management"] = "J Clin Anesth Pain Manag",
  ["journal of clinical apheresis"] = "J Clin Apher",
  ["journal of clinical biochemistry and nutrition"] = "J Clin Biochem Nutr",
  ["journal of clinical bioinformatics"] = "J Clin Bioinforma",
  ["journal of clinical cardiology and cardiovascular therapy"] = "J Clin Cardiol Cardiovasc Ther",
  ["journal of clinical case reports"] = "J Clin Case Rep",
  ["journal of clinical cases & reports"] = "J Clin Cases Rep",
  ["journal of clinical chemistry and clinical biochemistry"] = "J. Clin. Chem. Clin. Biochem.",
  ["journal of clinical chemistry and clinical biochemistry. zeitschrift fur klinische chemie und klinische biochemie"] = "J Clin Chem Clin Biochem",
  ["journal of clinical chemistry and clinical biochemistry. zeitschrift für klinische chemie und klinische biochemie"] = "J Clin Chem Clin Biochem",
  ["journal of clinical child and adolescent psychology"] = "J. Clin. Child Adolesc. Psychol.",
  ["journal of clinical child and adolescent psychology : the official journal for the society of clinical child and adolescent psychology, american psychological association, division 53"] = "J Clin Child Adolesc Psychol",
  ["journal of clinical child psychology"] = "J Clin Child Psychol",
  ["journal of clinical computing"] = "J Clin Comput",
  ["journal of clinical densitometry"] = "J. Clin. Densitom.",
  ["journal of clinical densitometry : the official journal of the international society for clinical densitometry"] = "J Clin Densitom",
  ["journal of clinical dentistry"] = "J. Clin. Dent.",
  ["journal of clinical dermatology"] = "J Clin Dermatol",
  ["journal of clinical dysmorphology"] = "J. Clin. Dysmorphol.",
  ["journal of clinical endocrinology and diabetes"] = "J Clin Endocrinol Diabetes",
  ["journal of clinical endocrinology and metabolism"] = "J. Clin. Endocrinol. Metab.",
  ["journal of clinical engineering"] = "J Clin Eng",
  ["journal of clinical epidemiology"] = "J Clin Epidemiol",
  ["journal of clinical epigenetics"] = "J Clin Epigenet",
  ["journal of clinical ethics"] = "J. Clin. Ethics",
  ["journal of clinical exercise physiology"] = "J Clin Exerc Physiol",
  ["journal of clinical forensic medicine"] = "J Clin Forensic Med",
  ["journal of clinical gastroenterology"] = "J Clin Gastroenterol",
  ["journal of clinical gastroenterology and hepatology"] = "J Clin Gastroenterol Hepatol",
  ["journal of clinical genomics"] = "J Clin Genom",
  ["journal of clinical gynecology and obstetrics"] = "J Clin Gynecol Obstet",
  ["journal of clinical haematology"] = "J Clin Haematol",
  ["journal of clinical hypertension"] = "J Clin Hypertens",
  ["journal of clinical hypertension (greenwich, conn.)"] = "J Clin Hypertens (Greenwich)",
  ["journal of clinical imaging science"] = "J Clin Imaging Sci",
  ["journal of clinical immunology"] = "J Clin Immunol",
  ["journal of clinical insights in neonatology"] = "J. Clin. Insights Neonatol.",
  ["journal of clinical investigation"] = "J. Clin. Invest.",
  ["journal of clinical issues in psychology"] = "J Clin Issues Psychol",
  ["journal of clinical laboratory analysis"] = "J Clin Lab Anal",
  ["journal of clinical laser medicine & surgery"] = "J Clin Laser Med Surg",
  ["journal of clinical laser medicine and surgery"] = "J. Clin. Laser Med. Surg.",
  ["journal of clinical lipidology"] = "J Clin Lipidol",
  ["journal of clinical medicine"] = "J Clin Med",
  ["journal of clinical medicine and therapeutics"] = "J Clin Med Ther",
  ["journal of clinical medicine research"] = "J Clin Med Res",
  ["journal of clinical metabolism & diabetes"] = "J Clin Metab Diabetes",
  ["journal of clinical microbiology"] = "J Clin Microbiol",
  ["journal of clinical monitoring"] = "J Clin Monit",
  ["journal of clinical monitoring and computing"] = "J Clin Monit Comput",
  ["journal of clinical movement disorders"] = "J Clin Mov Disord",
  ["journal of clinical neonatology"] = "J Clin Neonatol",
  ["journal of clinical nephrology and renal care"] = "J Clin Nephrol Ren Care",
  ["journal of clinical nephrology and research"] = "J Clin Nephrol Res",
  ["journal of clinical neuro-ophthalmology"] = "J Clin Neuroophthalmol",
  ["journal of clinical neurology (seoul, korea)"] = "J Clin Neurol",
  ["journal of clinical neuromuscular disease"] = "J Clin Neuromuscul Dis",
  ["journal of clinical neurophysiology"] = "J. Clin. Neurophysiol.",
  ["journal of clinical neurophysiology : official publication of the american electroencephalographic society"] = "J Clin Neurophysiol",
  ["journal of clinical neuropsychology"] = "J Clin Neuropsychol",
  ["journal of clinical neuroscience : official journal of the neurosurgical society of australasia"] = "J Clin Neurosci",
  ["journal of clinical nursing"] = "J Clin Nurs",
  ["journal of clinical nutrition and food science"] = "J Clin Nutr Food Sci",
  ["journal of clinical nutrition and metabolism"] = "J Clin Nutr Metab",
  ["journal of clinical oncology"] = "J. Clin. Oncol.",
  ["journal of clinical oncology : official journal of the american society of clinical oncology"] = "J Clin Oncol",
  ["journal of clinical oncology and research"] = "J Clin Oncol Res",
  ["journal of clinical oncology: clinical cancer informatics"] = "J. Clin. Oncol.: Clin. Cancer Inf.",
  ["journal of clinical oncology: precision oncology"] = "J. Clin. Oncol.: Precis. Oncol.",
  ["journal of clinical ophthalmology & eye disorders"] = "J Clin Ophthalmol Eye Disord",
  ["journal of clinical orthodontics"] = "J. Clin. Orthod.",
  ["journal of clinical orthodontics : jco"] = "J Clin Orthod",
  ["journal of clinical orthopaedics and trauma"] = "J Clin Orthop Trauma",
  ["journal of clinical outcomes management : jcom"] = "J Clin Outcomes Manag",
  ["journal of clinical pathology"] = "J Clin Pathol",
  ["journal of clinical pathology. supplement (association of clinical pathologists)"] = "J Clin Pathol Suppl (Assoc Clin Pathol)",
  ["journal of clinical pathology. supplement (college of pathologists)"] = "J Clin Pathol Suppl Coll Pathol",
  ["journal of clinical pathology. supplement (royal college of pathologists)"] = "J Clin Pathol Suppl (R Coll Pathol)",
  ["journal of clinical pathways : the foundation of value-based care"] = "J Clin Pathw",
  ["journal of clinical pediatric dentistry"] = "J. Clin. Pediatr. Dent.",
  ["journal of clinical periodontology"] = "J Clin Periodontol",
  ["journal of clinical pharmacology"] = "J Clin Pharmacol",
  ["journal of clinical pharmacology and new drugs"] = "J. Clin. Pharmacol. New Drugs",
  ["journal of clinical pharmacology and the journal of new drugs"] = "J. Clin. Pharmacol. J. New Drugs",
  ["journal of clinical pharmacy and therapeutics"] = "J Clin Pharm Ther",
  ["journal of clinical psychiatry"] = "J. Clin. Psychiatry",
  ["journal of clinical psychology"] = "J Clin Psychol",
  ["journal of clinical psychology in medical settings"] = "J Clin Psychol Med Settings",
  ["journal of clinical psychopharmacology"] = "J Clin Psychopharmacol",
  ["journal of clinical research & bioethics"] = "J Clin Res Bioeth",
  ["journal of clinical research and drug development"] = "J Clin Res Drug Dev",
  ["journal of clinical research and drug development."] = "J Clin Res Drug Dev",
  ["journal of clinical research and reports"] = "J Clin Res Rep",
  ["journal of clinical research best practices"] = "J Clin Res Best Pract",
  ["journal of clinical research in hiv aids and prevention"] = "J Clin Res HIV AIDS Prev",
  ["journal of clinical research in oncology"] = "J Clin Res Oncol",
  ["journal of clinical research in pediatric endocrinology"] = "J Clin Res Pediatr Endocrinol",
  ["journal of clinical respiratory medicine"] = "J Clin Respir Med",
  ["journal of clinical rheumatology & musculoskeletal medicine"] = "J Clin Rheumatol Musculoskelet Med",
  ["journal of clinical rheumatology : practical reports on rheumatic & musculoskeletal diseases"] = "J Clin Rheumatol",
  ["journal of clinical sleep medicine"] = "J. Clin. Sleep Med.",
  ["journal of clinical sleep medicine : jcsm : official publication of the american academy of sleep medicine"] = "J Clin Sleep Med",
  ["journal of clinical sport psychology"] = "J Clin Sport Psychol",
  ["journal of clinical stomatology conferences"] = "J Clin Stomatol Conf",
  ["journal of clinical toxicology"] = "J Clin Toxicol",
  ["journal of clinical trials"] = "J Clin Trials",
  ["journal of clinical tuberculosis and other mycobacterial diseases"] = "J Clin Tuberc Other Mycobact Dis",
  ["journal of clinical ultrasound"] = "J. Clin. Ultrasound",
  ["journal of clinical ultrasound : jcu"] = "J Clin Ultrasound",
  ["journal of clinical urology"] = "J Clin Urol",
  ["journal of clinical virology"] = "J. Clin. Virol.",
  ["journal of clinical virology : the official publication of the pan american society for clinical virology"] = "J Clin Virol",
  ["journal of cloud computing"] = "J. Cloud Comput.",
  ["journal of cloud computing (heidelberg, germany)"] = "J Cloud Comput (Heidelb)",
  ["journal of cluster science"] = "J. Cluster Sci.",
  ["journal of co2 utilization"] = "J. CO2 Util.",
  ["journal of coagulation disorders"] = "J Coagul Disord",
  ["journal of coal science and engineering"] = "J. Coal Sci. Eng.",
  ["journal of coastal conservation"] = "J Coast Conserv",
  ["journal of coastal research"] = "J Coast Res",
  ["journal of coatings technology"] = "J. Coat. Technol.",
  ["journal of coatings technology and research"] = "J. Coat. Technol. Res.",
  ["journal of cognition"] = "J Cogn",
  ["journal of cognition and culture"] = "J Cogn Cult",
  ["journal of cognition and development : official journal of the cognitive development society"] = "J Cogn Dev",
  ["journal of cognitive and behavioral psychotherapies : the official journal of the international institute for the advanced studies of psychotherapy and applied mental health"] = "J Cogn Behav Psychother",
  ["journal of cognitive education and psychology : jcep"] = "J Cogn Educ Psychol",
  ["journal of cognitive engineering and decision making"] = "J Cogn Eng Decis Mak",
  ["journal of cognitive enhancement"] = "J. Cognit. Enhancement",
  ["journal of cognitive enhancement : towards the integration of theory and practice"] = "J Cogn Enhanc",
  ["journal of cognitive neuroscience"] = "J Cogn Neurosci",
  ["journal of cognitive psychology (hove, england)"] = "J Cogn Psychol (Hove)",
  ["journal of cognitive psychotherapy"] = "J Cogn Psychother",
  ["journal of cognitive science"] = "J Cogn Sci (Seoul)",
  ["journal of cold regions engineering"] = "J. Cold Reg. Eng.",
  ["journal of collective negotiations in the public sector"] = "J Collect Negotiations Public Sect",
  ["journal of college and character"] = "J Coll Character",
  ["journal of college science teaching"] = "J Coll Sci Teach",
  ["journal of college student development"] = "J Coll Stud Dev",
  ["journal of college student psychotherapy"] = "J College Stud Psychother",
  ["journal of college student retention : research, theory & practice"] = "J Coll Stud Ret",
  ["journal of college teaching and learning"] = "J Coll Teach Learn",
  ["journal of colloid and interface science"] = "J. Colloid Interface Sci.",
  ["journal of colloid science"] = "J Colloid Sci",
  ["journal of colonialism & colonial history"] = "J Colon Colon Hist",
  ["journal of combinatorial algebra"] = "J. Comb. Algebra",
  ["journal of combinatorial chemistry"] = "J. Comb. Chem.",
  ["journal of combinatorial designs"] = "J. Comb. Des.",
  ["journal of combinatorial mathematics and combinatorial computing"] = "J. Combin. Math. Combin. Comput.",
  ["journal of combinatorial optimization"] = "J. Comb. Optim.",
  ["journal of combinatorial theory"] = "J. Combin. Theory Ser. A",
  ["journal of combinatorial theory series a"] = "J. Comb. Theory Ser. A",
  ["journal of combinatorial theory series b"] = "J. Comb. Theory Ser. B",
  ["journal of combinatorial theory. series a"] = "J Comb Theory Ser A",
  ["journal of combinatorial theory. series b"] = "J. Combin. Theory Ser. B",
  ["journal of combinatorics"] = "J. Comb.",
  ["journal of combinatorics and number theory"] = "J. Comb. Number Theory",
  ["journal of combinatorics, information & system sciences"] = "J. Comb. Inf. Syst. Sci.",
  ["journal of combustion"] = "J. Combust.",
  ["journal of commercial biotechnology"] = "J Commer Biotechnol",
  ["journal of commodity science, technology and quality"] = "J. Commod. Sci. Technol. Qual.",
  ["journal of common market studies"] = "J Common Mark Stud",
  ["journal of commonwealth political studies"] = "J Commonw Polit Stud",
  ["journal of communicable diseases"] = "J. Commun. Dis.",
  ["journal of communication"] = "J. Commun.",
  ["journal of communication disorders"] = "J Commun Disord",
  ["journal of communication in healthcare"] = "J Commun Healthc",
  ["journal of communications and information networks"] = "J. Commun. Inf. Networks",
  ["journal of communications and networks"] = "J. Commun. Networks",
  ["journal of communications technology and electronics"] = "J. Commun. Technol. Electron.",
  ["journal of community & applied social psychology"] = "J Community Appl Soc Psychol",
  ["journal of community & public health nursing"] = "J Community Public Health Nurs",
  ["journal of community engagement and scholarship"] = "J Community Engagem Scholarsh",
  ["journal of community genetics"] = "J Community Genet",
  ["journal of community health"] = "J Community Health",
  ["journal of community health nursing"] = "J Community Health Nurs",
  ["journal of community health research"] = "J Community Health Res",
  ["journal of community hospital internal medicine perspectives"] = "J Community Hosp Intern Med Perspect",
  ["journal of community medicine & health education"] = "J Community Med Health Educ",
  ["journal of community medicine & public health"] = "J Community Med Public Health",
  ["journal of community medicine (reno, nev.)"] = "J Community Med (Reno)",
  ["journal of community practice"] = "J Community Pract",
  ["journal of community psychology"] = "J Community Psychol",
  ["journal of commutative algebra"] = "J. Commut. Algebra",
  ["journal of comorbidity"] = "J Comorb",
  ["journal of comparative and physiological psychology"] = "J Comp Physiol Psychol",
  ["journal of comparative economics"] = "J Comp Econ",
  ["journal of comparative effectiveness research"] = "J Comp Eff Res",
  ["journal of comparative family studies"] = "J Comp Fam Stud",
  ["journal of comparative neurology"] = "J. Comp. Neurol.",
  ["journal of comparative pathology"] = "J Comp Pathol",
  ["journal of comparative physiology"] = "J Comp Physiol",
  ["journal of comparative physiology a neuroethology, sensory, neural, and behavioral physiology"] = "J. Comp. Physiol. A",
  ["journal of comparative physiology b biochemical, systemic, and environmental physiology"] = "J. Comp. Physiol. B",
  ["journal of comparative physiology. a, neuroethology, sensory, neural, and behavioral physiology"] = "J Comp Physiol A Neuroethol Sens Neural Behav Physiol",
  ["journal of comparative physiology. a, sensory, neural, and behavioral physiology"] = "J Comp Physiol [A]",
  ["journal of comparative physiology. b, biochemical, systemic, and environmental physiology"] = "J Comp Physiol B",
  ["journal of comparative psychology"] = "J Comp Psychol",
  ["journal of comparative psychology (washington, d.c. : 1983)"] = "J Comp Psychol",
  ["journal of compassionate health care"] = "J Compassionate Health Care",
  ["journal of complementary & integrative medicine"] = "J Complement Integr Med",
  ["journal of complex analysis"] = "J. Complex Anal.",
  ["journal of complex networks"] = "J Complex Netw",
  ["journal of complexity"] = "J. Complexity",
  ["journal of compliance in health care"] = "J. Compliance Health Care",
  ["journal of composite materials"] = "J. Compos. Mater.",
  ["journal of composites"] = "J Compos",
  ["journal of composites for construction"] = "J. Compos. Constr.",
  ["journal of composites science"] = "J. Compos. Sci.",
  ["journal of composites technology and research"] = "J. Compos. Tech. Res.",
  ["journal of computational acoustics"] = "J. Comput. Acoust.",
  ["journal of computational algebra"] = "J. Comput. Algebra",
  ["journal of computational analysis and applications"] = "J. Comput. Anal. Appl.",
  ["journal of computational and applied mathematics"] = "J. Comput. Appl. Math.",
  ["journal of computational and engineering mathematics"] = "J. Comput. Eng. Math.",
  ["journal of computational and graphical statistics"] = "J. Comput. Graph. Statist.",
  ["journal of computational and graphical statistics : a joint publication of american statistical association, institute of mathematical statistics, interface foundation of north america"] = "J Comput Graph Stat",
  ["journal of computational and nonlinear dynamics"] = "J. Comput. Nonlinear Dyn.",
  ["journal of computational and theoretical nanoscience"] = "J. Comput. Theor. Nanosci.",
  ["journal of computational and theoretical transport"] = "J. Comput. Theor. Transp.",
  ["journal of computational biology"] = "J. Comput. Biol.",
  ["journal of computational biology : a journal of computational molecular cell biology"] = "J Comput Biol",
  ["journal of computational biology and bioinformatics research"] = "J Comput Biol Bioinform Res",
  ["journal of computational biophysics and chemistry"] = "J. Comput. Biophys. Chem.",
  ["journal of computational chemistry"] = "J Comput Chem",
  ["journal of computational design and engineering"] = "J. Comput. Des. Eng.",
  ["journal of computational dynamics"] = "J. Comput. Dyn.",
  ["journal of computational electronics"] = "J Comput Electron",
  ["journal of computational finance"] = "J. Comput. Finance",
  ["journal of computational geometry"] = "J. Comput. Geom.",
  ["journal of computational interdisciplinary sciences"] = "J Comput Interdiscip Sci",
  ["journal of computational mathematics"] = "J. Comput. Math.",
  ["journal of computational mathematics : an international journal on numerical methods, analysis and applications"] = "J Comput Math",
  ["journal of computational mathematics and data science"] = "J. Comput. Math. Data Sci.",
  ["journal of computational methods in applied mathematics"] = "J Comput Methods Appl Math",
  ["journal of computational methods in molecular design"] = "J. Comput. Methods Mol. Des.",
  ["journal of computational methods in science and engineering"] = "J. Comput. Methods Sci. Eng.",
  ["journal of computational methods in sciences and engineering"] = "J. Comput. Methods Sci. Eng.",
  ["journal of computational multiphase flows"] = "J. Comput. Multiphase Flows",
  ["journal of computational neuroscience"] = "J Comput Neurosci",
  ["journal of computational physics"] = "J. Comput. Phys.",
  ["journal of computational physics x"] = "J. Comput. Phys.: X",
  ["journal of computational physics: x"] = "J. Comput. Phys. X",
  ["journal of computational science"] = "J. Comput. Sci.",
  ["journal of computational social science"] = "J Comput Soc Sci",
  ["journal of computational surgery"] = "J Comput Surg",
  ["journal of computed tomography"] = "J. Comput. Tomogr.",
  ["journal of computer and system sciences"] = "J. Comput. System Sci.",
  ["journal of computer and system sciences international"] = "J. Comput. Syst. Sci. Int.",
  ["journal of computer and systems sciences international"] = "J. Comput. Syst. Sci. Int.",
  ["journal of computer applications in archaeology"] = "J. Comput. Appl. Archaeol.",
  ["journal of computer assisted learning"] = "J Comput Assist Learn",
  ["journal of computer assisted tomography"] = "J Comput Assist Tomogr",
  ["journal of computer chemistry, japan"] = "J. COmput. Chem. Jpn.",
  ["journal of computer information systems"] = "J. Comput. Inf. Syst.",
  ["journal of computer languages"] = "J Comput Lang",
  ["journal of computer networks and communications"] = "J. Comput. Networks Commun.",
  ["journal of computer science and applied mathematics"] = "J. Comput. Sci. Appl. Math.",
  ["journal of computer science and systems biology"] = "J Comput Sci Syst Biol",
  ["journal of computer science and technology"] = "J. Comput. Sci. Tech.",
  ["journal of computer science and technology argentina"] = "J. Comput. Sci. Technol. Argent.",
  ["journal of computer security"] = "J Comput Secur",
  ["journal of computer virology and hacking techniques"] = "J. Comput. Virol Hacking Tech.",
  ["journal of computer vision and image processing"] = "J Comput Vis Image Process",
  ["journal of computer-aided materials design"] = "J. Comput.-Aided Mater. Des.",
  ["journal of computer-aided molecular design"] = "J Comput Aided Mol Des",
  ["journal of computer-assisted microscopy"] = "J Comput Assist Microsc",
  ["journal of computer-mediated communication"] = "J. Comput.-Mediated Commun.",
  ["journal of computer-mediated communication : jcmc"] = "J Comput Mediat Commun",
  ["journal of computers"] = "J Comput (Taipei)",
  ["journal of computing"] = "J Comput",
  ["journal of computing and information"] = "J. Comput. Inf.",
  ["journal of computing and information science in engineering"] = "J. Comput. Inf. Sci. Eng.",
  ["journal of computing in civil engineering"] = "J. Comput. Civil Eng.",
  ["journal of computing in higher education"] = "J Comput High Educ",
  ["journal of computing science and engineering"] = "J. Comput. Sci. Eng.",
  ["journal of computing science and engineering : jcse"] = "J Comput Sci Eng",
  ["journal of conchology"] = "J Conchol",
  ["journal of concrete and applicable mathematics"] = "J. Concr. Appl. Math.",
  ["journal of concussion"] = "J Concussion",
  ["journal of conflict resolution"] = "J. Conflict Resolution",
  ["journal of consciousness studies : controversies in science & the humanities"] = "J Conscious Stud",
  ["journal of conservation & museum studies"] = "J Conserv Mus Stud",
  ["journal of conservative dentistry : jcd"] = "J Conserv Dent",
  ["journal of construction engineering"] = "J. Constr. Eng.",
  ["journal of construction engineering and management"] = "J Constr Eng Manag",
  ["journal of construction engineering management"] = "J. Constr. Eng. Manage.",
  ["journal of construction research"] = "J. Constr. Res.",
  ["journal of constructional steel research"] = "J. Constr. Steel Res.",
  ["journal of constructivist psychology"] = "J Constr Psychol",
  ["journal of consulting and clinical psychology"] = "J Consult Clin Psychol",
  ["journal of consulting psychology"] = "J Consult Psychol",
  ["journal of consumer affairs"] = "J. Cons. Aff.",
  ["journal of consumer health on the internet"] = "J Consum Health Internet",
  ["journal of consumer policy"] = "J Consum Policy (Dordr)",
  ["journal of consumer protection and food safety"] = "J. Consum. Prot, Food Saf.",
  ["journal of consumer psychology : the official journal of the society for consumer psychology"] = "J Consum Psychol",
  ["journal of consumer research"] = "J. Cons. Res.",
  ["journal of contaminant hydrology"] = "J Contam Hydrol",
  ["journal of contemporary african studies : jcas"] = "J Contemp Afr Stud",
  ["journal of contemporary asia"] = "J Contemp Asia",
  ["journal of contemporary brachytherapy"] = "J Contemp Brachytherapy",
  ["journal of contemporary criminal justice"] = "J Contemp Crim Justice",
  ["journal of contemporary dental practice"] = "J. Contemp. Dent. Pract.",
  ["journal of contemporary diabetes research"] = "J Contemp Diabetes Res",
  ["journal of contemporary ethnography"] = "J Contemp Ethnogr",
  ["journal of contemporary health law and policy"] = "J. Contemp. Health Law Policy",
  ["journal of contemporary history"] = "J Contemp Hist",
  ["journal of contemporary law"] = "J Contemp Law",
  ["journal of contemporary mathematical analysis"] = "J. Contemp. Math. Anal.",
  ["journal of contemporary mathematical analysis. armenian academy of sciences"] = "J. Contemp. Math. Anal.",
  ["journal of contemporary medical education"] = "J Contemp Med Educ",
  ["journal of contemporary physics"] = "J. Contemp. Phys.",
  ["journal of contemporary psychotherapy"] = "J Contemp Psychother",
  ["journal of contemporary water research & education"] = "J Contemp Water Res Educ",
  ["journal of contextual behavioral science"] = "J Contextual Behav Sci",
  ["journal of continuing education in nursing"] = "J Contin Educ Nurs",
  ["journal of continuing education in the health professions"] = "J. Contin. Educ. Health Prof.",
  ["journal of contraceptive studies"] = "J Contracept Stud",
  ["journal of control and decision"] = "J. Control Decis.",
  ["journal of control science and engineering"] = "J. Control Sci. Eng.",
  ["journal of control, automation and electrical systems"] = "J. Control Autom. Electr. Syst.",
  ["journal of controlled release"] = "J. Controlled Release",
  ["journal of controlled release : official journal of the controlled release society"] = "J Control Release",
  ["journal of convex analysis"] = "J. Convex Anal.",
  ["journal of coordination chemistry"] = "J. Coord. Chem.",
  ["journal of corporate finance: contracting, governance and organization"] = "J. Corp. Finan.: Contracting",
  ["journal of correctional education (glen mills, pa.)"] = "J Correct Educ (Glen Mills)",
  ["journal of correctional health care : the official journal of the national commission on correctional health care"] = "J Correct Health Care",
  ["journal of corrosion science and engineering"] = "J. Corros. Sci. Eng.",
  ["journal of cosmetic and laser therapy"] = "J. Costmet Laser Ther.",
  ["journal of cosmetic and laser therapy : official publication of the european society for laser dermatology"] = "J Cosmet Laser Ther",
  ["journal of cosmetic dermatology"] = "J Cosmet Dermatol",
  ["journal of cosmetic science"] = "J Cosmet Sci",
  ["journal of cosmology and astro-particle physics"] = "J. Cosmology & Astro-Part. Phys.",
  ["journal of cosmology and astroparticle physics"] = "J. Cosmol. Astropart. Phys.",
  ["journal of cotton research"] = "J. Cotton Res.",
  ["journal of cotton science"] = "J Cotton Sci",
  ["journal of counseling and development : jcd"] = "J Couns Dev",
  ["journal of counseling psychology"] = "J Couns Psychol",
  ["journal of counseling sexology & sexual wellness : research, practice, and education"] = "J Couns Sexol Sex Wellness",
  ["journal of couple & relationship therapy"] = "J Couple Relatsh Ther",
  ["journal of coupled systems and multiscale dynamics"] = "J Coupled Syst Multiscale Dyn",
  ["journal of cranio-mandibular practice"] = "J. Craniomandibular. Pract.",
  ["journal of cranio-maxillo-facial surgery"] = "J. Craniomaxillofac. Surg.",
  ["journal of cranio-maxillo-facial surgery : official publication of the european association for cranio-maxillo-facial surgery"] = "J Craniomaxillofac Surg",
  ["journal of craniofacial genetics and developmental biology"] = "J Craniofac Genet Dev Biol",
  ["journal of craniofacial genetics and developmental biology. supplement"] = "J Craniofac Genet Dev Biol Suppl",
  ["journal of craniofacial surgery"] = "J. Craniofac. Surg.",
  ["journal of craniomandibular disorders"] = "J. Craniomandib. Disord.",
  ["journal of craniomandibular disorders : facial & oral pain"] = "J Craniomandib Disord",
  ["journal of craniovertebral junction & spine"] = "J Craniovertebr Junction Spine",
  ["journal of crash prevention and injury control"] = "J. Crash Prev. Inj. Control",
  ["journal of creativity in mental health"] = "J Creat Ment Health",
  ["journal of crime and justice"] = "J Crime Justice",
  ["journal of criminal justice"] = "J Crim Justice",
  ["journal of criminal psychology"] = "J Crim Psychol",
  ["journal of critical care"] = "J Crit Care",
  ["journal of critical care medicine"] = "J Crit Care Med",
  ["journal of critical care medicine (universitatea de medicină și farmacie din târgu-mureș)"] = "J Crit Care Med (Targu Mures)",
  ["journal of critical illness"] = "J. Crit. Illn.",
  ["journal of critical realism"] = "J Crit Realism",
  ["journal of crohn's & colitis"] = "J Crohns Colitis",
  ["journal of crop improvement"] = "J Crop Improv",
  ["journal of crop protection"] = "J Crop Prot",
  ["journal of crop science and biotechnology"] = "J. Crop Sci. Biotechnol.",
  ["journal of cross-cultural gerontology"] = "J Cross Cult Gerontol",
  ["journal of cross-cultural psychology"] = "J Cross Cult Psychol",
  ["journal of crustacean biology"] = "J. Crustacean Biol.",
  ["journal of crustacean biology : a quarterly of the crustacean society for the publication of research on any aspect of the biology of crustacea"] = "J Crustac Biol",
  ["journal of cryptographic engineering"] = "J Cryptogr Eng",
  ["journal of cryptography"] = "J. Cryptogr.",
  ["journal of cryptology"] = "J. Cryptology",
  ["journal of cryptology. the journal of the international association for cryptologic research"] = "J. Cryptology",
  ["journal of crystal growth"] = "J. Cryst. Growth",
  ["journal of cultural and evolutionary psychology"] = "J Cult Evol Psychol",
  ["journal of cultural cognitive science"] = "J Cult Cogn Sci",
  ["journal of cultural diversity"] = "J Cult Divers",
  ["journal of cultural economics"] = "J. Cult. Econ.",
  ["journal of cultural economy"] = "J Cult Econ",
  ["journal of cultural heritage"] = "J Cult Herit",
  ["journal of cultural heritage management and sustainable development"] = "J. Cult. Heritage Manage. Sustainable Dev.",
  ["journal of cuneiform studies"] = "JCS",
  ["journal of current chinese affairs"] = "J Curr Chin Aff",
  ["journal of current glaucoma practice"] = "J Curr Glaucoma Pract",
  ["journal of current ophthalmology"] = "J Curr Ophthalmol",
  ["journal of current science and technology"] = "J Curr Sci Technol",
  ["journal of current social issues"] = "J Curr Soc Issues",
  ["journal of current surgery"] = "J Curr Surg",
  ["journal of cutaneous and aesthetic surgery"] = "J Cutan Aesthet Surg",
  ["journal of cutaneous laser therapy"] = "J Cutan Laser Ther",
  ["journal of cutaneous medicine and surgery"] = "J Cutan Med Surg",
  ["journal of cutaneous pathology"] = "J Cutan Pathol",
  ["journal of cyber security"] = "J. Cyber Secur.",
  ["journal of cyber security technology"] = "J. Cyber Secur. Technol.",
  ["journal of cyber therapy and rehabilitation"] = "J Cyber Ther Rehabil",
  ["journal of cybersecurity"] = "J Cybersecur",
  ["journal of cyclic nucleotide and protein phosphorylation research"] = "J Cyclic Nucleotide Protein Phosphor Res",
  ["journal of cyclic nucleotide research"] = "J Cyclic Nucleotide Res",
  ["journal of cystic fibrosis"] = "J. Cyst. Fibros.",
  ["journal of cystic fibrosis : official journal of the european cystic fibrosis society"] = "J Cyst Fibros",
  ["journal of cytokine biology"] = "J Cytokine Biol",
  ["journal of cytology"] = "J Cytol",
  ["journal of cytology & histology"] = "J Cytol Histol",
  ["journal of cytology & molecular biology"] = "J Cytol Mol Biol",
  ["journal of dairy research"] = "J. Dairy Res.",
  ["journal of dairy science"] = "J Dairy Sci",
  ["journal of dairy science and biotechnology"] = "J Dairy Sci Biotechnol",
  ["journal of dairying, foods & home sciences"] = "J Dairy Foods Home Sci",
  ["journal of dalian university of technology"] = "J. Dalian Univ. Tech.",
  ["journal of dalian university of technology. dalian ligon daxue xuebao"] = "J. Dalian Univ. Technol.",
  ["journal of dance & somatic practices"] = "J Dance Somat Pract",
  ["journal of dance medicine & science : official publication of the international association for dance medicine & science"] = "J Dance Med Sci",
  ["journal of data and dynamic systems"] = "J. Data Dyn. Syst.",
  ["journal of data and information quality"] = "J. Data Inf. Qual.",
  ["journal of data and information science (warsaw, poland)"] = "J Data Inf Sci",
  ["journal of data mining in genomics & proteomics"] = "J Data Mining Genomics Proteomics",
  ["journal of data science : jds"] = "J Data Sci",
  ["journal of data, information and management"] = "J. Data Inf. Manage.",
  ["journal of deaf studies and deaf education"] = "J Deaf Stud Deaf Educ",
  ["journal of decision systems"] = "J Decis Syst",
  ["journal of defense analytics and logistics"] = "J. Def. Anal. Logist.",
  ["journal of defense modeling and simulation"] = "J. Def. Model. Simul.",
  ["journal of degraded and mining lands management"] = "J. Degraded Min. Lands Manage.",
  ["journal of demographic economics"] = "J Demogr Economics",
  ["journal of dental and craniofacial research"] = "J Dent Craniofac Res",
  ["journal of dental and maxillofacial surgery"] = "J Dent Maxillofac Surg",
  ["journal of dental anesthesia and pain medicine"] = "J Dent Anesth Pain Med",
  ["journal of dental biomaterials"] = "J Dent Biomater",
  ["journal of dental biomechanics"] = "J Dent Biomech",
  ["journal of dental education"] = "J Dent Educ",
  ["journal of dental health, oral disorders & therapy"] = "J Dent Health Oral Disord Ther",
  ["journal of dental hygiene"] = "J. Dent. Hyg.",
  ["journal of dental hygiene : jdh"] = "J Dent Hyg",
  ["journal of dental hygiene : jdh / american dental hygienists’ association"] = "J Dent Hyg",
  ["journal of dental medicine"] = "J. Dent. Med.",
  ["journal of dental practice administration"] = "J. Dent. Pract. Adm.",
  ["journal of dental practice administration : jdpa : official publication of american academy of dental practice administration, organization of teachers of dental practice administration, american academy of dental group practice"] = "J Dent Pract Adm",
  ["journal of dental research"] = "J Dent Res",
  ["journal of dental research, dental clinics, dental prospects"] = "J Dent Res Dent Clin Dent Prospects",
  ["journal of dental sciences"] = "J Dent Sci",
  ["journal of dental sleep medicine"] = "J Dent Sleep Med",
  ["journal of dental symposia"] = "J. Dent. Symp.",
  ["journal of dental technology"] = "J. Dent. Technol.",
  ["journal of dental technology : the peer-reviewed publication of the national association of dental laboratories"] = "J Dent Technol",
  ["journal of dental, oral and craniofacial epidemiology"] = "J Dent Oral Craniofac Epidemiol",
  ["journal of dentistry"] = "J Dent",
  ["journal of dentistry (shīrāz, iran)"] = "J Dent (Shiraz)",
  ["journal of dentistry (tehran, iran)"] = "J Dent (Tehran)",
  ["journal of dentistry and dental medicine"] = "J Dent Dent Med",
  ["journal of dentistry and oral biology"] = "J Dent Oral Biol",
  ["journal of dentistry and oral hygiene"] = "J Dent Oral Hyg",
  ["journal of dentistry and oral sciences"] = "J Dent Oral Sci",
  ["journal of dentistry for children"] = "J Dent Child",
  ["journal of dentistry for children (chicago, ill.)"] = "J Dent Child (Chic)",
  ["journal of dentistry for the handicapped"] = "J Dent Handicap",
  ["journal of depression & anxiety"] = "J Depress Anxiety",
  ["journal of derivatives"] = "J. Derivatives",
  ["journal of dermatologic surgery"] = "J. Dermatol. Surg.",
  ["journal of dermatologic surgery and oncology"] = "J. Dermatol. Surg. Oncol.",
  ["journal of dermatological case reports"] = "J Dermatol Case Rep",
  ["journal of dermatological science"] = "J Dermatol Sci",
  ["journal of dermatology"] = "J. Dermatol.",
  ["journal of design and manufacturing automation"] = "J. Des. Manuf. Autom.",
  ["journal of design and science"] = "J. Des. Sci.",
  ["journal of design history"] = "J Des Hist",
  ["journal of developing areas"] = "J Dev Areas",
  ["journal of developing drugs"] = "J Dev Drugs",
  ["journal of developing societies"] = "J Dev Soc",
  ["journal of development and agricultural economics"] = "J Dev Agric Econ",
  ["journal of development and economic policies"] = "J. Devel. Econ. Pol.",
  ["journal of development economics"] = "J Dev Econ",
  ["journal of development effectiveness"] = "J Dev Effect",
  ["journal of development studies"] = "J. Devel. Stud.",
  ["journal of developmental and behavioral pediatrics"] = "J. Dev. Behav. Pediatr.",
  ["journal of developmental and behavioral pediatrics : jdbp"] = "J Dev Behav Pediatr",
  ["journal of developmental and life-course criminology"] = "J Dev Life Course Criminol",
  ["journal of developmental and physical disabilities"] = "J Dev Phys Disabil",
  ["journal of developmental biology"] = "J Dev Biol",
  ["journal of developmental biology & regenerative medicine"] = "J Dev Biol Regen Med",
  ["journal of developmental biology and tissue engineering"] = "J Dev Biol Tissue Eng",
  ["journal of developmental entrepreneurship : jde"] = "J Dev Entrep",
  ["journal of developmental origins of health and disease"] = "J Dev Orig Health Dis",
  ["journal of developmental physiology"] = "J Dev Physiol",
  ["journal of diabetes"] = "J Diabetes",
  ["journal of diabetes & metabolic disorders"] = "J. Diabetes Metab. Disord.",
  ["journal of diabetes & metabolism"] = "J Diabetes Metab",
  ["journal of diabetes and clinical research"] = "J Diabetes Clin Res",
  ["journal of diabetes and its complications"] = "J Diabetes Complications",
  ["journal of diabetes and metabolic disorders"] = "J Diabetes Metab Disord",
  ["journal of diabetes and obesity"] = "J Diabetes Obes",
  ["journal of diabetes and related disorders"] = "J Diabetes Relat Disord",
  ["journal of diabetes and treatment"] = "J Diabetes Treat",
  ["journal of diabetes investigation"] = "J Diabetes Investig",
  ["journal of diabetes mellitus"] = "J Diabetes Mellitus",
  ["journal of diabetes mellitus and metabolic syndrome"] = "J Diabetes Mellit Metab Syndr",
  ["journal of diabetes nursing"] = "J Diabetes Nurs",
  ["journal of diabetes research"] = "J Diabetes Res",
  ["journal of diabetes research & clinical metabolism"] = "J Diabetes Res Clin Metab",
  ["journal of diabetes research and therapy"] = "J Diabetes Res Ther",
  ["journal of diabetes science and technology"] = "J Diabetes Sci Technol",
  ["journal of diabetes, metabolic disorders & control"] = "J Diabetes Metab Disord Control",
  ["journal of diabetic complications"] = "J. Diabet. Complications",
  ["journal of diagnostic imaging in therapy"] = "J Diagn Imaging Ther",
  ["journal of diagnostic medical sonography"] = "J. Diagn. Med. Sonogr.",
  ["journal of diagnostic medical sonography : jdms"] = "J Diagn Med Sonogr",
  ["journal of diagnostic radiography and imaging"] = "J. Diagn. Radiogr. Imaging",
  ["journal of diagnostic techniques and biomedical analysis"] = "J Diagn Tech Biomed Anal",
  ["journal of dialysis"] = "J Dial",
  ["journal of diarrhoeal diseases research"] = "J Diarrhoeal Dis Res",
  ["journal of dietary supplements"] = "J Diet Suppl",
  ["journal of difference equations and applications"] = "J. Differ. Equations Appl.",
  ["journal of differential equations"] = "J. Differential Equations",
  ["journal of differential geometry"] = "J. Differential Geom.",
  ["journal of digestive cancer reports"] = "J Dig Cancer Rep",
  ["journal of digestive diseases"] = "J Dig Dis",
  ["journal of digestive diseases and hepatology"] = "J Dig Dis Hepatol",
  ["journal of digital imaging"] = "J Digit Imaging",
  ["journal of digital imaging : the official journal of the society for computer applications in radiology"] = "J Digit Imaging",
  ["journal of digital landscape architecture : jodla"] = "J Digit Landsc Archit",
  ["journal of digital media management"] = "J Digit Media Manag",
  ["journal of disability policy studies"] = "J Disabil Policy Stud",
  ["journal of discrete algorithms"] = "J. Discrete Algoritms",
  ["journal of discrete algorithms (amsterdam, netherlands)"] = "J Discrete Algorithms (Amst)",
  ["journal of discrete mathematical sciences & cryptography"] = "J. Discrete Math. Sci. Cryptography",
  ["journal of discrete mathematical sciences and cryptography"] = "J. Discrete Math. Sci. Cryptogr.",
  ["journal of disease and global health"] = "J Dis Glob Health",
  ["journal of dispersion science and technology"] = "J. Dispersion Sci. Technol.",
  ["journal of display technology"] = "J. Disp. Technol.",
  ["journal of diversity in higher education"] = "J Divers High Educ",
  ["journal of divorce"] = "J Divorce",
  ["journal of divorce & remarriage"] = "J Divorce Remarriage",
  ["journal of doctoral nursing practice"] = "J Dr Nurs Pract",
  ["journal of down syndrome & chromosome abnormalities"] = "J Down Syndr Chromosom Abnorm",
  ["journal of dr. ntr university of health sciences"] = "J Dr NTR Univ Health Sci",
  ["journal of drug abuse"] = "J Drug Abuse",
  ["journal of drug addiction, education, and eradication"] = "J Drug Addict Educ Erad",
  ["journal of drug and alcohol research"] = "J Drug Alcohol Res",
  ["journal of drug assessment"] = "J Drug Assess",
  ["journal of drug delivery"] = "J Drug Deliv",
  ["journal of drug delivery science and technology"] = "J. Drug Delivery Sci. Technol.",
  ["journal of drug design and research"] = "J Drug Des Res",
  ["journal of drug education"] = "J Drug Educ",
  ["journal of drug issues"] = "J Drug Issues",
  ["journal of drug metabolism & toxicology"] = "J Drug Metab Toxicol",
  ["journal of drug policy analysis"] = "J Drug Policy Anal",
  ["journal of drug research"] = "J Drug Res",
  ["journal of drug targeting"] = "J Drug Target",
  ["journal of drugs in dermatology"] = "J. Drugs Dermatol.",
  ["journal of drugs in dermatology : jdd"] = "J Drugs Dermatol",
  ["journal of dual diagnosis"] = "J Dual Diagn",
  ["journal of dynamic behavior of materials"] = "J. Dyn. Behav. Mater.",
  ["journal of dynamic systems measurement & control"] = "J. Dyn. Syst. Meas. Control",
  ["journal of dynamic systems, measurement, and control"] = "J Dyn Syst Meas Control",
  ["journal of dynamical and control systems"] = "J. Dyn. Control Syst.",
  ["journal of dynamical systems and geometric theories"] = "J. Dyn. Syst. Geom. Theories",
  ["journal of dynamics and differential equations"] = "J. Dynam. Differential Equations",
  ["journal of dynamics and games"] = "J. Dyn. Games",
  ["journal of early and intensive behavior intervention : jeibi"] = "J Early Intensive Behav Interv",
  ["journal of early childhood and infant psychology"] = "J Early Child Infant Psychol",
  ["journal of early childhood research : ecr"] = "J Early Child Res",
  ["journal of early childhood teacher education"] = "J Early Child Teach Educ",
  ["journal of early christian studies"] = "JECS",
  ["journal of early christian studies. journal of the north american patristics society"] = "JEChrSt",
  ["journal of early hearing detection and intervention"] = "J Early Hear Detect Interv",
  ["journal of early intervention"] = "J Early Interv",
  ["journal of early modern history"] = "J Early Mod Hist",
  ["journal of early modern studies"] = "J Early Mod Stud (Bucur)",
  ["journal of earth and environmental sciences"] = "J Earth Environ Sci",
  ["journal of earth science"] = "J. Earth Sci.",
  ["journal of earth science & climatic change"] = "J Earth Sci Clim Change",
  ["journal of earth science (wuhan, china)"] = "J Earth Sci",
  ["journal of earth system science"] = "J. Earth Syst. Sci.",
  ["journal of earthquake and tsunami"] = "J. Earthquake Tsunami",
  ["journal of earthquake engineering"] = "J. Earthquake Eng.",
  ["journal of east african natural history"] = "J East Afr Nat Hist",
  ["journal of east and west studies"] = "J East West Stud",
  ["journal of east asian linguistics"] = "J East Asian Ling",
  ["journal of east china normal university"] = "J. East China Norm. Univ. Natur. Sci. Ed.",
  ["journal of east china normal university. natural science edition. huadong shifan daxue xuebao. ziran kexue ban"] = "J. East China Norm. Univ. Natur. Sci. Ed.",
  ["journal of east-west business"] = "J. East-West Bus.",
  ["journal of eastern african research & development"] = "J East Afr Res Dev",
  ["journal of eastern african studies : the journal of the british institute in eastern africa"] = "J East Afr Stud",
  ["journal of eating disorders"] = "J Eat Disord",
  ["journal of ecclesiastical history"] = "JEH",
  ["journal of echocardiography"] = "J Echocardiogr",
  ["journal of ecohydraulics"] = "J. Ecohydraul.",
  ["journal of ecology"] = "J. Ecol.",
  ["journal of ecology and environment"] = "J Ecol Environ",
  ["journal of ecology of health & environment"] = "J Ecol Health Environ",
  ["journal of econometric methods"] = "J Econom Method",
  ["journal of econometrics"] = "J. Econometrics",
  ["journal of economic and social history of the orient"] = "J Econ Soc Hist Orient",
  ["journal of economic and social measurement"] = "J Econ Soc Meas",
  ["journal of economic and social policy"] = "J. Econ. Soc. Pol.",
  ["journal of economic and social research"] = "J. Econ. Soc. Res.",
  ["journal of economic asymmetries"] = "J Econ Asymmetries",
  ["journal of economic behavior & organization"] = "J Econ Behav Organ",
  ["journal of economic behavior and organization"] = "J. Econ. Behav. Organ.",
  ["journal of economic cooperation among islamic countries"] = "J. Econ. Cooperation Islamic Countries",
  ["journal of economic development"] = "J Econ Dev",
  ["journal of economic dynamics & control"] = "J Econ Dyn Control",
  ["journal of economic dynamics and control"] = "J. Econ. Dynam. Control",
  ["journal of economic education"] = "J. Econ. Educ.",
  ["journal of economic entomology"] = "J Econ Entomol",
  ["journal of economic geography"] = "J Econ Geogr",
  ["journal of economic growth"] = "J. Econ. Growth",
  ["journal of economic growth (boston, mass.)"] = "J Econ Growth (Boston)",
  ["journal of economic history"] = "J. Econ. Hist.",
  ["journal of economic inequality"] = "J Econ Inequal",
  ["journal of economic integration"] = "J. Econ. Integration",
  ["journal of economic interaction and coordination"] = "J Econ Interact Coord",
  ["journal of economic issues"] = "J Econ Issues",
  ["journal of economic literature"] = "J Econ Lit",
  ["journal of economic methodology"] = "J. Econ. Methodology",
  ["journal of economic perspectives"] = "J. Econ. Perspect.",
  ["journal of economic psychology"] = "J Econ Psychol",
  ["journal of economic research"] = "J. Econ. Res.",
  ["journal of economic structures"] = "J Econ Struct",
  ["journal of economic studies"] = "J. Econ. Stud.",
  ["journal of economic studies (glasgow, scotland)"] = "J Econ Stud",
  ["journal of economic surveys"] = "J Econ Surv",
  ["journal of economic theory"] = "J Econ Theory",
  ["journal of economic theory and econometrics"] = "J. Econ. Theory Econometrics",
  ["journal of economics & management strategy"] = "J Econ Manag Strategy",
  ["journal of economics (mvea)"] = "J. Econ. (MVEA)",
  ["journal of economics (vienna, austria : online)"] = "J Econ (Vienna)",
  ["journal of economics (zeitschrift für nationalökonomie)"] = "J. Econ. (Z. Nationalökon.)",
  ["journal of economics and business"] = "J Econ Bus",
  ["journal of economics and finance"] = "J. Econ. Finance",
  ["journal of economics and management strategy"] = "J. Econ. Manage. Strategy",
  ["journal of economics and public finance"] = "J Econ Public Financ",
  ["journal of economics, trade and marketing management"] = "J Econ Trade Mark Manag",
  ["journal of ecosystem & ecography"] = "J Ecosyst Ecography",
  ["journal of ecotourism"] = "J. Ecotourism",
  ["journal of ect"] = "J. ECT",
  ["journal of ecumenical studies"] = "J Ecumenical Stud",
  ["journal of education (boston, mass.)"] = "J Educ (Boston)",
  ["journal of education and culture studies"] = "J Educ Cult Stud",
  ["journal of education and e-learning research"] = "J Educ Elearn Res",
  ["journal of education and health promotion"] = "J Educ Health Promot",
  ["journal of education and training studies"] = "J Educ Train Stud",
  ["journal of education finance"] = "J. Educ. Finance",
  ["journal of education for social work"] = "J Educ Soc Work",
  ["journal of education for students placed at risk"] = "J Educ Stud Placed Risk",
  ["journal of education for sustainable development"] = "J Educ Sustain Dev",
  ["journal of education for teaching : international research and pedagogy"] = "J Educ Teach",
  ["journal of education in science, environment and health"] = "J Educ Environ Sci Health",
  ["journal of education, teaching and social studies"] = "J Educ Teach Soc Stud",
  ["journal of educational administration and history"] = "J Educ Adm Hist",
  ["journal of educational and behavioral statistics : a quarterly publication sponsored by the american educational research association and the american statistical association"] = "J Educ Behav Stat",
  ["journal of educational and developmental psychology"] = "J Educ Develop Psychol",
  ["journal of educational and psychological consultation : the official journal of the association for educational and psychological consultants"] = "J Educ Psychol Consult",
  ["journal of educational audiology : official journal of the educational audiology association"] = "J Educ Audiol",
  ["journal of educational computing research"] = "J. Educ. Comput. Res.",
  ["journal of educational evaluation for health professions"] = "J Educ Eval Health Prof",
  ["journal of educational measurement"] = "J Educ Meas",
  ["journal of educational psychology"] = "J Educ Psychol",
  ["journal of educational technology & society"] = "J Educ Techno Soc",
  ["journal of educational technology systems"] = "J. Educ. Technol. Syst.",
  ["journal of egyptian archaeology"] = "J. Egypt. Archaeol.",
  ["journal of ehealth technology and application"] = "J Ehealth Technol Appl",
  ["journal of elasticity"] = "J. Elast.",
  ["journal of elasticity. the physical and mathematical science of solids"] = "J. Elasticity",
  ["journal of elastomers and plastics"] = "J. Elastomers Plast.",
  ["journal of elder abuse & neglect"] = "J Elder Abuse Negl",
  ["journal of elections, public opinion and parties"] = "J Elect Public Opin Parties",
  ["journal of electric propulsion"] = "J. Electr. Propul.",
  ["journal of electrical and computer engineering"] = "J. Electr. Comput. Eng.",
  ["journal of electrical and electronics engineering, australia"] = "J Electr Electron Eng Aust",
  ["journal of electrical bioimpedance"] = "J Electr Bioimpedance",
  ["journal of electrical engineering and technology"] = "J. Electr. Eng. Technol.",
  ["journal of electrical systems and information technology"] = "J. Electr. Syst. Inf. Technol.",
  ["journal of electroanalytical chemistry"] = "J. Electroanal. Chem.",
  ["journal of electroanalytical chemistry (lausanne, switzerland)"] = "J Electroanal Chem (Lausanne)",
  ["journal of electroanalytical chemistry and interfacial electrochemistry"] = "J. Electroanal. Chem. Interfacial Electrochem.",
  ["journal of electrocardiology"] = "J Electrocardiol",
  ["journal of electroceramics"] = "J. Electroceram.",
  ["journal of electrochemical energy conversion and storage"] = "J. Electrochem. Energy Convers. Storage",
  ["journal of electrochemical science and technology"] = "J. Electrochem. Sci. Technol.",
  ["journal of electromagnetic analysis and applications"] = "J. Electromagn. Anal. Appl.",
  ["journal of electromagnetic waves and applications"] = "J Electromagn Waves Appl",
  ["journal of electromyography and kinesiology"] = "J. Electromyogr. Kinesiol.",
  ["journal of electromyography and kinesiology : official journal of the international society of electrophysiological kinesiology"] = "J Electromyogr Kinesiol",
  ["journal of electron microscopy"] = "J Electron Microsc (Tokyo)",
  ["journal of electron microscopy technique"] = "J Electron Microsc Tech",
  ["journal of electron spectroscopy and related phenomena"] = "J. Electron. Spectrosc. Relat. Phenom.",
  ["journal of electron spectroscopy and related phenomenon"] = "J. Electron Spectrosc. Relat. Phenom.",
  ["journal of electronic imaging"] = "J Electron Imaging",
  ["journal of electronic materials"] = "J. Electron. Mater.",
  ["journal of electronic packaging"] = "J. Electron. Packag.",
  ["journal of electronic resources librarianship"] = "J Electron Resour Librariansh",
  ["journal of electronic science and technology"] = "J. Electron. Sci. Technol.",
  ["journal of electronic testing"] = "J. Electron. Test.",
  ["journal of electronics (china)"] = "J. Electron.",
  ["journal of electronics manufacturing"] = "J. Electron. Manuf.",
  ["journal of electrostatics"] = "J. Electrostat.",
  ["journal of electrotechnics and mathematics"] = "J. Electrotech. Math.",
  ["journal of elementology"] = "J Elem",
  ["journal of elliptic and parabolic equations"] = "J Elliptic Parabol Equ",
  ["journal of embryology and experimental morphology"] = "J Embryol Exp Morphol",
  ["journal of emergencies, trauma, and shock"] = "J Emerg Trauma Shock",
  ["journal of emergency and critical care medicine (hong kong, china)"] = "J Emerg Crit Care Med",
  ["journal of emergency management (weston, mass.)"] = "J Emerg Manag",
  ["journal of emergency management and disaster communications"] = "J. Emerg. Manage. Disaster Commun.",
  ["journal of emergency medical services"] = "J. Emerg. Med. Serv. JEMS",
  ["journal of emergency medicine"] = "J. Emerg. Med.",
  ["journal of emergency medicine forecast"] = "J Emerg Med Forecast",
  ["journal of emergency nursing"] = "J Emerg Nurs",
  ["journal of emergency nursing: jen : official publication of the emergency department nurses association"] = "J Emerg Nurs",
  ["journal of emerging diseases and virology"] = "J Emerg Dis Virol",
  ["journal of emerging infectious diseases"] = "J Emerg Infect Dis",
  ["journal of emerging investigators"] = "J Emerg Investig",
  ["journal of emerging markets"] = "J. Emerging Markets",
  ["journal of emerging technologies and innovative research"] = "J Emerg Technol Innov Res",
  ["journal of emotional and behavioral disorders"] = "J Emot Behav Disord",
  ["journal of empirical finance"] = "J Empir Finance",
  ["journal of empirical legal studies"] = "J Empir Leg Stud",
  ["journal of empirical research on human research ethics"] = "J. Empirical Res. Hum. Res. Ethics",
  ["journal of empirical research on human research ethics : jerhre"] = "J Empir Res Hum Res Ethics",
  ["journal of employment counseling"] = "J Employ Couns",
  ["journal of enabling technologies"] = "J Enabling Technol",
  ["journal of endocrinological investigation"] = "J Endocrinol Invest",
  ["journal of endocrinology"] = "J. Endocrinol.",
  ["journal of endocrinology and diabetes"] = "J Endocrinol Diabetes",
  ["journal of endocrinology and diabetes mellitus"] = "J Endocrinol Diabetes Mellit",
  ["journal of endocrinology and metabolism"] = "J Endocrinol Metab",
  ["journal of endocrinology and reproduction : jer"] = "J Endocrinol Reprod",
  ["journal of endocrinology and thyroid research"] = "J Endocrinol Thyroid Res",
  ["journal of endocrinology, diabetes & obesity"] = "J Endocrinol Diabetes Obes",
  ["journal of endodontia"] = "J Endod",
  ["journal of endodontics"] = "J Endod",
  ["journal of endometriosis"] = "J Endometr",
  ["journal of endometriosis and pelvic pain disorders"] = "J Endometr Pelvic Pain Disord",
  ["journal of endotoxin research"] = "J Endotoxin Res",
  ["journal of endourology"] = "J Endourol",
  ["journal of endourology / endourological society"] = "J Endourol",
  ["journal of endourology case reports"] = "J Endourol Case Rep",
  ["journal of endourology. part b, videourology"] = "J Endourol B Videourol",
  ["journal of endovascular resuscitation and trauma management"] = "J Endovasc Resusc Trauma Manag",
  ["journal of endovascular surgery"] = "J. Endovasc. Surg.",
  ["journal of endovascular surgery : the official journal of the international society for endovascular surgery"] = "J Endovasc Surg",
  ["journal of endovascular therapy"] = "J. Endovasc. Ther.",
  ["journal of endovascular therapy : an official journal of the international society of endovascular specialists"] = "J Endovasc Ther",
  ["journal of energetic materials"] = "J. Energetic Mater.",
  ["journal of energy"] = "J. Energy",
  ["journal of energy and development"] = "J. Energy Devel.",
  ["journal of energy chemistry"] = "J. Energy Chem.",
  ["journal of energy engineering"] = "J. Energy Eng.",
  ["journal of energy finance and development"] = "J. Energy Finance Devel.",
  ["journal of energy in southern africa"] = "J. Energy South. Afr.",
  ["journal of energy literature"] = "J. Energy Lit.",
  ["journal of energy resources technology"] = "J. Energy Res. Technol.",
  ["journal of energy storage"] = "J Energy Storage",
  ["journal of engineered fibers and fabrics"] = "J. Eng. Fibers Fabr.",
  ["journal of engineering"] = "J. Eng.",
  ["journal of engineering (stevenage, england)"] = "J Eng (Stevenage)",
  ["journal of engineering and applied science"] = "J. Eng. Appl. Sci.",
  ["journal of engineering and science in medical diagnostics and therapy"] = "J Eng Sci Med Diagn Ther",
  ["journal of engineering and technology management"] = "J. Eng. Tech. Manage.",
  ["journal of engineering design"] = "J. Eng. Des.",
  ["journal of engineering education"] = "J. Eng. Educ.",
  ["journal of engineering for gas turbines and power"] = "J. Eng. Gas Turbines Power",
  ["journal of engineering for sustainable buildings and cities"] = "J. Eng. Sustainable Build. Cities",
  ["journal of engineering materials and technology"] = "J. Eng. Mater. Technol.",
  ["journal of engineering mathematics"] = "J. Eng. Math.",
  ["journal of engineering mechanics"] = "J. Eng. Mech.",
  ["journal of engineering physics and thermophysics"] = "J. Eng. Phys. Thermophys.",
  ["journal of engineering psychology"] = "J Eng Psychol",
  ["journal of engineering thermophysics"] = "J. Eng. Thermophys.",
  ["journal of engineering, design and technology"] = "J. Eng., Des. Technol.",
  ["journal of english for academic purposes"] = "J Engl Acad Purp",
  ["journal of english linguistics"] = "J Eng Linguist",
  ["journal of enhanced heat transfer"] = "J. Enhanced Heat Transfer",
  ["journal of enterostomal therapy"] = "J Enterostomal Ther",
  ["journal of enterprise transformation"] = "J. Enterp. Transform.",
  ["journal of entomological and acarological research"] = "J. Entomol. Acarol. Res.",
  ["journal of entomological research"] = "J. Entomol. Res.",
  ["journal of entomological science"] = "J Entomol Sci",
  ["journal of entomology (online)"] = "J Entomol",
  ["journal of entomology and nematology"] = "J Entomol Nematol",
  ["journal of entomology and zoology studies"] = "J Entomol Zool Stud",
  ["journal of entomology series a, general entomology"] = "J. Entomol. Ser. A. Gen. Entomol.",
  ["journal of entomology series b, taxonomy"] = "J. Entomol. Ser. B. Taxon.",
  ["journal of entrepreneurship education"] = "J Entrep Educ",
  ["journal of environment and development"] = "J. Environ. Dev.",
  ["journal of environment and health sciences"] = "J Environ Health Sci",
  ["journal of environment biology"] = "J. Environ. Biol.",
  ["journal of environment health"] = "J. Environ. Health",
  ["journal of environment management"] = "J. Environ. Manage.",
  ["journal of environment quality"] = "J. Environ. Qual.",
  ["journal of environment radioactivity"] = "J. Environ. Radioact.",
  ["journal of environment science and health, part a environmental science"] = "J. Environ. Sci. Health., Part A",
  ["journal of environment science and health, part b pesticides"] = "J. Environ. Sci. Health., Part B",
  ["journal of environmental & analytical toxicology"] = "J Environ Anal Toxicol",
  ["journal of environmental & engineering geophysics"] = "J Environ Eng Geophys",
  ["journal of environmental and engineering geophysics"] = "J. Environ. Eng. Geophys.",
  ["journal of environmental and occupational science"] = "J Environ Occup Sci",
  ["journal of environmental and public health"] = "J Environ Public Health",
  ["journal of environmental and toxicological studies"] = "J Environ Toxicol Stud",
  ["journal of environmental assessment policy and management"] = "J. Environ. Assess. Policy Manage.",
  ["journal of environmental biology"] = "J. Environ. Biol.",
  ["journal of environmental biology / academy of environmental biology, india"] = "J Environ Biol",
  ["journal of environmental chemical engineering"] = "J. Environ. Chem. Eng.",
  ["journal of environmental economics and management"] = "J Environ Econ Manage",
  ["journal of environmental engineering"] = "J. Environ. Eng.",
  ["journal of environmental engineering & ecological science"] = "J Environ Eng Ecol Sci",
  ["journal of environmental engineering (new york, n.y.)"] = "J Environ Eng (New York)",
  ["journal of environmental engineering (reston, va, united states)"] = "J. Environ. Eng. (Reston, VA, U.S.)",
  ["journal of environmental engineering and landscape management"] = "J. Environ. Eng. Landscape Manage.",
  ["journal of environmental engineering and science"] = "J. Environ. Eng. Sci.",
  ["journal of environmental health"] = "J. Environ. Health",
  ["journal of environmental health and sustainable development"] = "J. Environ. Health Sustainable Dev.",
  ["journal of environmental health science & engineering"] = "J Environ Health Sci Eng",
  ["journal of environmental health science and engineering"] = "J. Environ. Health Sci. Eng.",
  ["journal of environmental horticulture"] = "J Environ Hortic",
  ["journal of environmental immunology and toxicology"] = "J Environ Immunol Toxicol",
  ["journal of environmental informatics"] = "J. Environ. Inf.",
  ["journal of environmental management"] = "J Environ Manage",
  ["journal of environmental materials and sustainable energy"] = "J. Environ. Mater. Sustainable Energy",
  ["journal of environmental monitoring"] = "J. Environ. Monit.",
  ["journal of environmental monitoring : jem"] = "J Environ Monit",
  ["journal of environmental nanotechnology"] = "J. Environ. Nanotechnol.",
  ["journal of environmental pathology and toxicology"] = "J. Environ. Pathol. Toxicol.",
  ["journal of environmental pathology, toxicology and oncology"] = "J. Environ. Pathol. Toxicol. Oncol.",
  ["journal of environmental pathology, toxicology and oncology : official organ of the international society for environmental toxicology and cancer"] = "J Environ Pathol Toxicol Oncol",
  ["journal of environmental planning and management"] = "J. Environ. Plann. Manage.",
  ["journal of environmental policy & planning"] = "J. Environ. Plann. Policy Manage.",
  ["journal of environmental polymer degradation"] = "J Environ Polym Degrad",
  ["journal of environmental protection"] = "J Environ Prot (Irvine, Calif)",
  ["journal of environmental protection and ecology"] = "J. Environ. Prot. Ecol.",
  ["journal of environmental protection science"] = "J Environ Prot Sci",
  ["journal of environmental psychology"] = "J Environ Psychol",
  ["journal of environmental quality"] = "J Environ Qual",
  ["journal of environmental radioactivity"] = "J Environ Radioact",
  ["journal of environmental science & engineering"] = "J Environ Sci Eng",
  ["journal of environmental science & engineering (nagpur, india)"] = "J. Environ. Sci. Eng. (Nagpur, India)",
  ["journal of environmental science and engineering"] = "J. Environ. Sci. Eng.",
  ["journal of environmental science and health, part a environmental science"] = "J. Environ. Sci. Health. A",
  ["journal of environmental science and health, part a: environmental science and engineering"] = "J. Environ. Sci. Health A Environ. Sci. Health Part A: Environ. Sci. Eng",
  ["journal of environmental science and health, part a: environmental science and engineering & toxic and hazardous substance control"] = "J. Environ. Sci. Health. A: Environ. Sci. Eng.Toxic Hazard. Subst. Control",
  ["journal of environmental science and health, part a: toxic/hazardous substances & environmental engineering"] = "J. Environ. Sci. Health A: Toxic/Hazard. Subst. Environ. Eng.",
  ["journal of environmental science and health, part b pesticides"] = "J. Environ. Sci. Health B",
  ["journal of environmental science and health, part b: pesticides, food contaminants, and agricultural wastes"] = "J. Environ. Sci. Health B",
  ["journal of environmental science and health, part c: environmental carcinogenesis & ecotoxicology reviews"] = "J. Environ. Sci. Health C: Environ. Carcinog. Ecotoxicol. Rev.",
  ["journal of environmental science and health, part c: environmental carcinogenesis reviews"] = "J. Environ. Sci. Health., Part C Environ. Carcinog. Rev.",
  ["journal of environmental science and health, part c: environmental health sciences"] = "J. Environ. Sci. Health., Part C Environ. Health Sci.",
  ["journal of environmental science and health. part a, environmental science and engineering"] = "J Environ Sci Health A Environ Sci Eng",
  ["journal of environmental science and health. part a, environmental science and engineering & toxic and hazardous substance control"] = "J Environ Sci Health A Environ Sci Eng Toxic Hazard Subst Control",
  ["journal of environmental science and health. part a, toxic/hazardous substances & environmental engineering"] = "J Environ Sci Health A Tox Hazard Subst Environ Eng",
  ["journal of environmental science and health. part a, toxic/hazardous substances and environmental engineering"] = "J. Environ. Sci. Health Part A Tox. Hazard. Subst. Environ. Eng.",
  ["journal of environmental science and health. part b, pesticides, food contaminants, and agricultural wastes"] = "J. Environ. Sci. Health B",
  ["journal of environmental science and health. part b: pesticides, food contaminants, and agricultural wastes"] = "J Environ Sci Health B",
  ["journal of environmental science and health. part c, environmental carcinogenesis & ecotoxicology reviews"] = "J Environ Sci Health C Environ Carcinog Ecotoxicol Rev",
  ["journal of environmental science and health. part c, environmental health sciences"] = "J. Environ. Sci. Health [C]",
  ["journal of environmental science and health. part c, toxicology and carcinogenesis"] = "J Environ Sci Health C Toxicol Carcinog",
  ["journal of environmental science and health. part c: environmental health sciences"] = "J Environ Sci Health C",
  ["journal of environmental science and health. part. b, pesticides, food contaminants, and agricultural wastes"] = "J Environ Sci Health B",
  ["journal of environmental science and technology"] = "J. Environ. Sci. Technol.",
  ["journal of environmental sciences"] = "J. Environ. Sci.",
  ["journal of environmental sciences (china)"] = "J Environ Sci (China)",
  ["journal of environmental solutions for oil, gas, and mining"] = "J. Environ. Solutions Oil Gas Min.",
  ["journal of environmental studies (northborough, mass.)"] = "J Environ Stud (Northborough)",
  ["journal of environmental studies and sciences"] = "J Environ Stud Sci",
  ["journal of environmental toxiology and public health"] = "J Environ Toxicol Public Health",
  ["journal of enzyme inhibition"] = "J Enzyme Inhib",
  ["journal of enzyme inhibition and medicinal chemistry"] = "J Enzyme Inhib Med Chem",
  ["journal of enzymology and metabolism"] = "J Enzymol Metab",
  ["journal of epidemiological research"] = "J Epidemiol Res",
  ["journal of epidemiology"] = "J Epidemiol",
  ["journal of epidemiology / japan epidemiological association"] = "J Epidemiol",
  ["journal of epidemiology and biostatistics"] = "J Epidemiol Biostat",
  ["journal of epidemiology and community health"] = "J Epidemiol Community Health",
  ["journal of epidemiology and global health"] = "J Epidemiol Glob Health",
  ["journal of epidemiology and preventive medicine"] = "J Epidemiol Prev Med",
  ["journal of epidemiology and public health reviews"] = "J Epidemiol Public Health Rev",
  ["journal of epilepsy"] = "J Epilepsy",
  ["journal of epilepsy research"] = "J Epilepsy Res",
  ["journal of epithelial biology & pharmacology"] = "J Epithel Biol Pharmacol",
  ["journal of equine science"] = "J Equine Sci",
  ["journal of equine veterinary science"] = "J Equine Vet Sci",
  ["journal of escience librarianship"] = "J Escience Librariansh",
  ["journal of essential oil bearing plants"] = "J. Essent. Oil Bear. Plants",
  ["journal of essential oil research"] = "J. Essent. Oil Res.",
  ["journal of essential oil-bearing plants"] = "J. Essent. Oil Bear. Plants",
  ["journal of esthetic dentistry"] = "J Esthet Dent",
  ["journal of et nursing"] = "J. ET Nurs.",
  ["journal of et nursing : official publication, international association for enterostomal therapy"] = "J ET Nurs",
  ["journal of ethics in mental health"] = "J Ethics Ment Health",
  ["journal of ethics, law, and aging"] = "J Ethics Law Aging",
  ["journal of ethiopian studies"] = "J Ethiop Stud",
  ["journal of ethiopian studies / haile sellassie i university, institute of ethiopian studies"] = "J Ethiop Stud",
  ["journal of ethnic & cultural diversity in social work"] = "J Ethn Cult Divers Soc Work",
  ["journal of ethnic and migration studies"] = "J Ethn Migr Stud",
  ["journal of ethnic foods"] = "J. Ethnic Foods",
  ["journal of ethnicity in criminal justice"] = "J Ethn Crim Justice",
  ["journal of ethnicity in substance abuse"] = "J Ethn Subst Abuse",
  ["journal of ethnobiology"] = "J Ethnobiol",
  ["journal of ethnobiology and ethnomedicine"] = "J Ethnobiol Ethnomed",
  ["journal of ethnographic and qualitative research"] = "J Ethnogr Qual Res",
  ["journal of ethnopharmacology"] = "J Ethnopharmacol",
  ["journal of ethology"] = "J Ethol",
  ["journal of eukaryotic microbiology"] = "J. Eukaryotic Microbiol.",
  ["journal of european archaeology"] = "JEurArch",
  ["journal of european cme"] = "J Eur CME",
  ["journal of european economic history"] = "J. Europ. Econ. Hist.",
  ["journal of european integration"] = "J Eur Integr",
  ["journal of european landscape"] = "J. Eur. Landscape",
  ["journal of european public policy"] = "J Eur Public Policy",
  ["journal of european social policy"] = "J Eur Soc Policy",
  ["journal of european studies"] = "J Eur Stud",
  ["journal of evaluation in clinical practice"] = "J Eval Clin Pract",
  ["journal of evidence and quality in health care"] = "J. Evidence Qual. Health Care",
  ["journal of evidence-based complementary & alternative medicine"] = "J Evid Based Complementary Altern Med",
  ["journal of evidence-based integrative medicine"] = "J Evid Based Integr Med",
  ["journal of evidence-based medicine"] = "J Evid Based Med",
  ["journal of evidence-based psychotherapies"] = "J Evid Based Psychother",
  ["journal of evidence-based social work"] = "J Evid Based Soc Work",
  ["journal of evidence-based social work (2019)"] = "J Evid Based Soc Work (2019)",
  ["journal of evidence-informed social work"] = "J Evid Inf Soc Work",
  ["journal of evolution and health"] = "J Evol Health",
  ["journal of evolution and technology"] = "J Evol Technol",
  ["journal of evolution equations"] = "J Evol Equ",
  ["journal of evolution of medical and dental sciences"] = "J Evol Med Dent Sci",
  ["journal of evolutionary biochemistry and physiology"] = "J Evol Biochem Physiol",
  ["journal of evolutionary biology"] = "J Evol Biol",
  ["journal of evolutionary economics"] = "J Evol Econ",
  ["journal of evolving stem cell research"] = "J Evol Stem Cell Res",
  ["journal of exceptional children"] = "J Except Child",
  ["journal of exercise nutrition & biochemistry"] = "J Exerc Nutrition Biochem",
  ["journal of exercise physiology"] = "J. Exercise Physiol.",
  ["journal of exercise physiology online"] = "J Exerc Physiol Online",
  ["journal of exercise rehabilitation"] = "J Exerc Rehabil",
  ["journal of exercise science and fitness"] = "J Exerc Sci Fit",
  ["journal of exercise, sports & orthopedics"] = "J Exerc Sports Orthop",
  ["journal of existentialism"] = "J Existent",
  ["journal of exotic pet medicine"] = "J Exot Pet Med",
  ["journal of experimental & clinical assisted reproduction"] = "J Exp Clin Assist Reprod",
  ["journal of experimental & clinical cancer research"] = "J. Exp. Clin. Cancer Res.",
  ["journal of experimental & clinical cancer research : cr"] = "J Exp Clin Cancer Res",
  ["journal of experimental & theoretical artificial intelligence"] = "J. Exp. Theor. Artif. Intell.",
  ["journal of experimental & theoretical artificial intelligence : jetai"] = "J Exp Theor Artif Intell",
  ["journal of experimental and applied animal sciences"] = "J Exp Appl Anim Sci",
  ["journal of experimental and clinical cancer research"] = "J. Exp. Clin. Cancer Res.",
  ["journal of experimental and clinical medicine"] = "J Exp Clin Med",
  ["journal of experimental and integrative medicine"] = "J Exp Integr Med",
  ["journal of experimental and theoretical physics"] = "J. Exp. Theor. Phys.",
  ["journal of experimental animal science"] = "J Exp Anim Sci",
  ["journal of experimental biology"] = "J. Exp. Biol.",
  ["journal of experimental botany"] = "J Exp Bot",
  ["journal of experimental child psychology"] = "J Exp Child Psychol",
  ["journal of experimental criminology"] = "J Exp Criminol",
  ["journal of experimental education"] = "J Exp Educ",
  ["journal of experimental life science"] = "J Exp Life Sci",
  ["journal of experimental marine biology and ecology"] = "J. Exp. Mar. Biol. Ecol.",
  ["journal of experimental medical sciences"] = "J Exp Med Sci",
  ["journal of experimental medicine"] = "J. Exp. Med.",
  ["journal of experimental nanoscience"] = "J. Exp. Nanosci.",
  ["journal of experimental nephrology"] = "J Exp Nephrol",
  ["journal of experimental neurology"] = "J Exp Neurol",
  ["journal of experimental neuroscience"] = "J Exp Neurosci",
  ["journal of experimental orthopaedics"] = "J Exp Orthop",
  ["journal of experimental pathology"] = "J Exp Pathol",
  ["journal of experimental pathology (oxford, england)"] = "J Exp Pathol (Oxford)",
  ["journal of experimental pharmacology"] = "J Exp Pharmacol",
  ["journal of experimental psychology"] = "J Exp Psychol",
  ["journal of experimental psychology. animal behavior processes"] = "J Exp Psychol Anim Behav Process",
  ["journal of experimental psychology. animal learning and cognition"] = "J Exp Psychol Anim Learn Cogn",
  ["journal of experimental psychology. applied"] = "J Exp Psychol Appl",
  ["journal of experimental psychology. general"] = "J Exp Psychol Gen",
  ["journal of experimental psychology. human learning and memory"] = "J Exp Psychol Hum Learn",
  ["journal of experimental psychology. human perception and performance"] = "J Exp Psychol Hum Percept Perform",
  ["journal of experimental psychology. learning, memory, and cognition"] = "J Exp Psychol Learn Mem Cogn",
  ["journal of experimental psychology: animal behavior processes"] = "J. Exp. Psychol. Anim. Behav. Process.",
  ["journal of experimental psychology: general"] = "J. Exp. Psychol. Gen.",
  ["journal of experimental psychology: human learning and memory"] = "J. Exp. Psychol. [Hum. Learn.]",
  ["journal of experimental psychology: human perception and performance"] = "J. Exp. Psychol. Hum. Percept. Perform.",
  ["journal of experimental psychology: learning, memory, and cognition"] = "J. Exp. Psychol. Learn. Mem. Cogn.",
  ["journal of experimental psychopathology"] = "J Exp Psychopathol",
  ["journal of experimental research on human growth & aging"] = "J Exp Res Hum Growth Aging",
  ["journal of experimental social psychology"] = "J Exp Soc Psychol",
  ["journal of experimental stroke & translational medicine"] = "J Exp Stroke Transl Med",
  ["journal of experimental therapeutics & oncology"] = "J Exp Ther Oncol",
  ["journal of experimental therapeutics and oncology"] = "J. Exp. Ther. Oncol.",
  ["journal of experimental zoology"] = "J. Exp. Zool.",
  ["journal of experimental zoology india"] = "J Exp Zoology India",
  ["journal of experimental zoology part a, comparative experimental biology"] = "J. Exp. Zool. Part A: Comp. Exp. Biol.",
  ["journal of experimental zoology part a, ecological and integrative physiology"] = "J. Exp. Zool. Part A: Ecol. Integr. Physiol.",
  ["journal of experimental zoology part a: ecological genetics and physiology"] = "J. Exp. Zool. Part A: Ecol. Genet. Physiol.",
  ["journal of experimental zoology part b molecular and developmental evolution"] = "J. Exp. Zool. Part B",
  ["journal of experimental zoology. part a, comparative experimental biology"] = "J Exp Zool A Comp Exp Biol",
  ["journal of experimental zoology. part a, ecological and integrative physiology"] = "J Exp Zool A Ecol Integr Physiol",
  ["journal of experimental zoology. part a, ecological genetics and physiology"] = "J Exp Zool A Ecol Genet Physiol",
  ["journal of experimental zoology. part b, molecular and developmental evolution"] = "J Exp Zool B Mol Dev Evol",
  ["journal of experimental zoology. part b. molecular and developmental evolution"] = "J Exp Zoolog B Mol Dev Evol",
  ["journal of experimental zoology. supplement"] = "J. Exp. Zool. Suppl.",
  ["journal of exploratory research in pharmacology"] = "J Explor Res Pharmacol",
  ["journal of explosives engineering"] = "J. Explos. Eng.",
  ["journal of exposure analysis and environment epidemiology"] = "J. Exposure Anal. Environ. Epidemiol.",
  ["journal of exposure analysis and environmental epidemiology"] = "J Expo Anal Environ Epidemiol",
  ["journal of exposure science & environmental epidemiology"] = "J Expo Sci Environ Epidemiol",
  ["journal of exposure science and environmental epidemiology"] = "J. Expo. Sci. Environ. Epidemiol.",
  ["journal of extension"] = "J Ext",
  ["journal of extra-corporeal technology"] = "J. Extra. Corpor. Technol.",
  ["journal of extracellular biology"] = "J. Extracell. Biol.",
  ["journal of extracellular vesicles"] = "J Extracell Vesicles",
  ["journal of extreme events"] = "J. Extreme Events",
  ["journal of eye movement research"] = "J Eye Mov Res",
  ["journal of facade design and engineering"] = "J. Facade Des. Eng.",
  ["journal of failure analysis and prevention"] = "J. Fail. Anal. Prev.",
  ["journal of family & community medicine"] = "J Family Community Med",
  ["journal of family & reproductive health"] = "J Family Reprod Health",
  ["journal of family and consumer sciences"] = "J Fam Consum Sci",
  ["journal of family and economic issues"] = "J Fam Econ Issues",
  ["journal of family communication"] = "J Fam Commun",
  ["journal of family health"] = "J Fam Health",
  ["journal of family health training"] = "J Famil Health Train",
  ["journal of family history"] = "J Fam Hist",
  ["journal of family issues"] = "J Fam Issues",
  ["journal of family law"] = "J Fam Law",
  ["journal of family law / university of louisville school of law"] = "J Fam Law",
  ["journal of family medicine"] = "J Fam Med",
  ["journal of family medicine & community health"] = "J Fam Med Community Health",
  ["journal of family medicine and disease prevention"] = "J Fam Med Dis Prev",
  ["journal of family medicine and primary care"] = "J Family Med Prim Care",
  ["journal of family nursing"] = "J Fam Nurs",
  ["journal of family planning and reproductive health care"] = "J. Fam. Plann. Reprod. Health Care",
  ["journal of family practice"] = "J. Fam. Pract.",
  ["journal of family psychology : jfp : journal of the division of family psychology of the american psychological association (division 43)"] = "J Fam Psychol",
  ["journal of family psychotherapy"] = "J Fam Psychother",
  ["journal of family social work"] = "J Fam Soc Work",
  ["journal of family strengths"] = "J Family Strengths",
  ["journal of family studies"] = "J Fam Stud",
  ["journal of family theory & review"] = "J Fam Theory Rev",
  ["journal of family therapy"] = "J Fam Ther",
  ["journal of family violence"] = "J Fam Violence",
  ["journal of far eastern business"] = "J. Far Eastern Bus.",
  ["journal of farm economics"] = "J. Farm Econ.",
  ["journal of fasting and health"] = "J Fasting Health",
  ["journal of feline medicine and surgery"] = "J Feline Med Surg",
  ["journal of feline medicine and surgery open reports"] = "J. Feline Med. Surg. Open Rep.",
  ["journal of feminist family therapy"] = "J Fem Fam Ther",
  ["journal of fermentation and bioengineering"] = "J. Ferment. Bioeng.",
  ["journal of fermentation technology"] = "J. Ferment. Technol.",
  ["journal of fertilization in vitro"] = "J Fertili In Vitro",
  ["journal of fertilization. in vitro ivf worldwide reproductive medicine genetics & stem cell biology"] = "J Fertil In Vitro IVF Worldw Reprod Med Genet Stem Cell Biol",
  ["journal of fiber bioengineering and informatics"] = "J. Fiber Bioeng. Inf.",
  ["journal of fiber science and technology"] = "J. Fiber Sci. Technol.",
  ["journal of field archaeology"] = "J Field Archaeol",
  ["journal of field ornithology"] = "J Field Ornithol",
  ["journal of field robotics"] = "J. Field Rob.",
  ["journal of finance"] = "J. Finance",
  ["journal of financial and quantitative analysis"] = "J. Finan. Quant. Anal.",
  ["journal of financial crime"] = "J Financ Crime",
  ["journal of financial economics"] = "J financ econ",
  ["journal of financial intermediation"] = "J. Finan. Intermediation",
  ["journal of financial management and analysis"] = "J. Finan. Manage. Anal.",
  ["journal of financial research"] = "J. Finan. Res.",
  ["journal of financial services research"] = "J. Finan. Services Res.",
  ["journal of finsler geometry and its applications"] = "J. Finsler Geom. Appl.",
  ["journal of fintech"] = "J. FinTech",
  ["journal of fire protection engineering"] = "J. Fire. Prot. Eng.",
  ["journal of fire sciences"] = "J. Fire Sci.",
  ["journal of fish and wildlife management"] = "J Fish Wildl Manag",
  ["journal of fish biology"] = "J Fish Biol",
  ["journal of fish diseases"] = "J Fish Dis",
  ["journal of fisheries and aquaculture"] = "J Fish Aquac",
  ["journal of fisheries and aquatic science"] = "J. Fish. Aquat. Sci.",
  ["journal of fisheries and environment"] = "J Fish Environ",
  ["journal of fisheries science and technology"] = "J Fish Sci Technol",
  ["journal of fisheriessciences.com"] = "J Fishscicom",
  ["journal of fixed point theory and applications"] = "J. Fixed Point Theory Appl.",
  ["journal of flood risk management"] = "J. Flood Risk Manage.",
  ["journal of flow chemistry"] = "J Flow Chem",
  ["journal of flow visualization and image processing"] = "J. Flow Visualization Image Process.",
  ["journal of fluency disorders"] = "J Fluency Disord",
  ["journal of fluid flow, heat and mass transfer"] = "J Fluid Flow Heat Mass Transf",
  ["journal of fluid mechanics"] = "J. Fluid Mech.",
  ["journal of fluids and structures"] = "J. Fluids Struct.",
  ["journal of fluids engineering"] = "J. Fluids Eng.",
  ["journal of fluorescence"] = "J Fluoresc",
  ["journal of fluorine chemistry"] = "J. Fluorine Chem.",
  ["journal of foetal medicine"] = "J Foetal Med",
  ["journal of food & industrial microbiology"] = "J Food Ind Microbiol",
  ["journal of food & nutritional disorders"] = "J Food Nutr Disord",
  ["journal of food and drug analysis"] = "J Food Drug Anal",
  ["journal of food and nutrition research (newark, del.)"] = "J Food Nutr Res (Newark)",
  ["journal of food bioactives : an official scientific publication of the international society of nutraceuticals and functional foods (isnff)"] = "J Food Bioact",
  ["journal of food biochemistry"] = "J Food Biochem",
  ["journal of food chemistry and nanotechnology"] = "J. Food Chem. Nanotechnol.",
  ["journal of food composition and analysis"] = "J. Food Compos. Anal.",
  ["journal of food composition and analysis : an official publication of the united nations university, international network of food data systems"] = "J Food Compost Anal",
  ["journal of food engineering"] = "J. Food Eng.",
  ["journal of food lipids"] = "J. Food Lipids",
  ["journal of food measurement and characterization"] = "J. Food Meas. Charact.",
  ["journal of food process engineering"] = "J Food Process Eng",
  ["journal of food processing & technology"] = "J Food Process Technol",
  ["journal of food processing and preservation"] = "J. Food Process. Preserv.",
  ["journal of food products marketing"] = "J. Food Prod. Mark.",
  ["journal of food protection"] = "J Food Prot",
  ["journal of food quality"] = "J Food Qual",
  ["journal of food quality & hazards control"] = "J Food Qual Hazards Control",
  ["journal of food research"] = "J Food Res",
  ["journal of food safety"] = "J. Food Saf.",
  ["journal of food safety and food quality"] = "J. Food Saf. Food Qual.",
  ["journal of food science"] = "J Food Sci",
  ["journal of food science and engineering"] = "J Food Sci Eng",
  ["journal of food science and nutrition research"] = "J Food Sci Nutr Res",
  ["journal of food science and technology"] = "J Food Sci Technol",
  ["journal of food science education"] = "J. Food Sci. Educ.",
  ["journal of food technology"] = "J. Food Technol.",
  ["journal of food, agriculture & environment"] = "J. Food Agric. Environ.",
  ["journal of food, agriculture and environment"] = "J. Food Agric. Environ.",
  ["journal of foot and ankle research"] = "J Foot Ankle Res",
  ["journal of foot and ankle surgery"] = "J. Foot Ankle Surg.",
  ["journal of foot surgery"] = "J. Foot Surg.",
  ["journal of foramineferal research"] = "J. Foramineferal Res.",
  ["journal of foraminiferal research"] = "J Foraminifer Res",
  ["journal of forecasting"] = "J Forecast",
  ["journal of forensic and legal medicine"] = "J Forensic Leg Med",
  ["journal of forensic dental sciences"] = "J Forensic Dent Sci",
  ["journal of forensic economics"] = "J. Forensic Econ.",
  ["journal of forensic identification"] = "J Forensic Identif",
  ["journal of forensic medicine"] = "J. Forensic Med.",
  ["journal of forensic medicine and toxicology"] = "J. Forensic Med. Toxicol.",
  ["journal of forensic neuropsychology"] = "J. Forensic Neuropsychol.",
  ["journal of forensic nursing"] = "J Forensic Nurs",
  ["journal of forensic odonto-stomatology"] = "J. Forensic Odontostomatol.",
  ["journal of forensic practice"] = "J. Forensic Pract.",
  ["journal of forensic psychiatry & psychology"] = "J. Forensic Psychiatry Psychol.",
  ["journal of forensic psychology practice"] = "J Forensic Psychol Pract",
  ["journal of forensic psychology research and practice"] = "J Forensic Psychol Res Pract",
  ["journal of forensic radiology and imaging"] = "J. Forensic Radiol. Imaging",
  ["journal of forensic science"] = "J. Forensic Sci.",
  ["journal of forensic science & criminology"] = "J Forensic Sci Criminol",
  ["journal of forensic sciences"] = "J Forensic Sci",
  ["journal of forensic sciences & criminal investigation"] = "J Forensic Sci Crim Investig",
  ["journal of forensic social work"] = "J Forensic Soc Work",
  ["journal of forensics research"] = "J Forensics Res",
  ["journal of forest economics"] = "J For Econ",
  ["journal of forest research"] = "J. For. Res.",
  ["journal of forest research : open access"] = "J For Res",
  ["journal of forest science"] = "J For Sci (Prague)",
  ["journal of forestry"] = "J For",
  ["journal of forestry research"] = "J For Res (Harbin)",
  ["journal of formalized reasoning"] = "J Formaliz Reason",
  ["journal of formative design in learning"] = "J Form Des Learn",
  ["journal of fourier analysis and applications"] = "J.Fourier Anal. Appl.",
  ["journal of fractal geometry"] = "J. Fractal Geom.",
  ["journal of fractal geometry. mathematics of fractals and related topics"] = "J. Fractal Geom.",
  ["journal of fractional calculus"] = "J. Fract. Calc.",
  ["journal of fractional calculus and applications"] = "J. Fract. Calc. Appl.",
  ["journal of frailty, sarcopenia and falls"] = "J Frailty Sarcopenia Falls",
  ["journal of free radicals in biology & medicine"] = "J Free Radic Biol Med",
  ["journal of free radicals in biology and medicine"] = "J. Free Radic. Biol. Med.",
  ["journal of freshwater ecology"] = "J Freshw Ecol",
  ["journal of freshwater fish"] = "J. Freshwater Fish",
  ["journal of friction and wear"] = "J. Frict. Wear",
  ["journal of fuel cell science and technology"] = "J. Fuel Cell Sci. Technol.",
  ["journal of fuel chemistry and technology"] = "J. Fuel Chem. Technol.",
  ["journal of function spaces"] = "J. Funct. Spaces",
  ["journal of functional analysis"] = "J. Funct. Anal.",
  ["journal of functional biomaterials"] = "J Funct Biomater",
  ["journal of functional foods"] = "J. Funct. Foods",
  ["journal of functional morphology and kinesiology"] = "J Funct Morphol Kinesiol",
  ["journal of functional programming"] = "J. Funct. Program.",
  ["journal of fungi"] = "J. Fungi",
  ["journal of fungi (basel, switzerland)"] = "J Fungi (Basel)",
  ["journal of further and higher education"] = "J Furth High Educ",
  ["journal of fusion energy"] = "J. Fusion Energy",
  ["journal of future foods"] = "J. Future Foods",
  ["journal of future robot life"] = "J. Future Robot LIfe",
  ["journal of futures markets"] = "J. Futures Markets",
  ["journal of fuzhou university"] = "J. Fuzhou Univ. Nat. Sci. Ed.",
  ["journal of fuzhou university. natural science edition. fuzhou daxue xuebao. ziran kexue ban"] = "J. Fuzhou Univ. Nat. Sci. Ed.",
  ["journal of fuzzy mathematics"] = "J. Fuzzy Math.",
  ["journal of gambling issues"] = "J Gambl Issues",
  ["journal of gambling studies"] = "J Gambl Stud",
  ["journal of gambling studies / co-sponsored by the national council on problem gambling and institute for the study of gambling and commercial gaming"] = "J Gambl Stud",
  ["journal of gansu university of technology"] = "J. Gansu Univ. Technol.",
  ["journal of garden history"] = "J Gard Hist",
  ["journal of gastric cancer"] = "J Gastric Cancer",
  ["journal of gastroenterology"] = "J Gastroenterol",
  ["journal of gastroenterology and hepatology"] = "J Gastroenterol Hepatol",
  ["journal of gastroenterology and hepatology research"] = "J Gastroenterol Hepatol Res",
  ["journal of gastroenterology, pancreatology & liver disorders"] = "J Gastroenterol Pancreatol Liver Disord",
  ["journal of gastrointestinal & digestive system"] = "J Gastrointest Dig Syst",
  ["journal of gastrointestinal and liver diseases"] = "J. Gastrointestin. Liver Dis.",
  ["journal of gastrointestinal and liver diseases : jgld"] = "J Gastrointestin Liver Dis",
  ["journal of gastrointestinal cancer"] = "J Gastrointest Cancer",
  ["journal of gastrointestinal oncology"] = "J Gastrointest Oncol",
  ["journal of gastrointestinal surgery"] = "J. Gastrointest. Surg.",
  ["journal of gastrointestinal surgery : official journal of the society for surgery of the alimentary tract"] = "J Gastrointest Surg",
  ["journal of gay & lesbian mental health"] = "J Gay Lesbian Ment Health",
  ["journal of gay & lesbian social services"] = "J Gay Lesbian Soc Serv",
  ["journal of gemmology"] = "J. Gemmol.",
  ["journal of gender studies"] = "J Gend Stud",
  ["journal of gender-specific medicine"] = "J. Gend. Specif. Med.",
  ["journal of gene medicine"] = "J. Gene Med.",
  ["journal of gene therapy"] = "J Gene Ther",
  ["journal of general and applied microbiology"] = "J. Gen. Appl. Microbiol.",
  ["journal of general and family medicine"] = "J Gen Fam Med",
  ["journal of general and molecular virology"] = "J Gen Mol Virol",
  ["journal of general internal medicine"] = "J Gen Intern Med",
  ["journal of general internal medicine : official journal of the society for research and education in primary care internal medicine"] = "J Gen Intern Med",
  ["journal of general medicine (dover, del.)"] = "J Gen Med (Dover)",
  ["journal of general microbiology"] = "J. Gen. Microbiol.",
  ["journal of general orthodontics"] = "J Gen Orthod",
  ["journal of general physiology"] = "J. Gen. Physiol.",
  ["journal of general plant pathology"] = "J. Gen. Plant Pathol.",
  ["journal of general plant pathology : jgpp"] = "J Gen Plant Pathol",
  ["journal of general practice (los angeles, calif.)"] = "J Gen Pract (Los Angel)",
  ["journal of general psychology"] = "J. Gen. Psychol.",
  ["journal of general virology"] = "J. Gen. Virol.",
  ["journal of generalized lie theory and applications"] = "J. Gen. Lie Theory Appl.",
  ["journal of generic medicines"] = "J Generic Med",
  ["journal of genes and proteins"] = "J Genes Proteins",
  ["journal of genetic counseling"] = "J Genet Couns",
  ["journal of genetic disorders & disease information"] = "J Genet Disord Dis Inf",
  ["journal of genetic disorders & genetic reports"] = "J Genet Disord Genet Rep",
  ["journal of genetic engineering and biotechnology"] = "J. Genet. Eng. Biotechnol.",
  ["journal of genetic psychology"] = "J. Genet. Psychol.",
  ["journal of genetic syndromes & gene therapy"] = "J Genet Syndr Gene Ther",
  ["journal of genetics"] = "J Genet",
  ["journal of genetics & breeding"] = "J Genet Breed",
  ["journal of genetics and genome research"] = "J Genet Genome Res",
  ["journal of genetics and genomics"] = "J. Genet. Genomics",
  ["journal of genetics and genomics = yi chuan xue bao"] = "J Genet Genomics",
  ["journal of genetics study"] = "J Genet Study",
  ["journal of genital system & disorders"] = "J Genit Syst Disord",
  ["journal of genitourinary disorders"] = "J Genitourin Disord",
  ["journal of genocide research"] = "J Genocide Res",
  ["journal of genome science and technology"] = "J. Genome Sci. Technol.",
  ["journal of genomics"] = "J Genomics",
  ["journal of geo-engineering sciences"] = "J. Geo-Eng. Sci.",
  ["journal of geochemical exploration"] = "J. Geochem. Explor.",
  ["journal of geodesy"] = "J. Geod.",
  ["journal of geodetic science"] = "J. Geodetic Sci.",
  ["journal of geodynamics"] = "J. Geodyn.",
  ["journal of geographical sciences"] = "J. Geog. Sci.",
  ["journal of geographical systems"] = "J Geogr Syst",
  ["journal of geography"] = "J. Geogr.",
  ["journal of geography, environment and earth science international"] = "J. Geogr. Environ. Earth Sci. Int.",
  ["journal of geological education"] = "J. Geol. Educ.",
  ["journal of geological research"] = "J. Geol. Res.",
  ["journal of geology"] = "J. Geol.",
  ["journal of geomagnetism and geoelectricity"] = "J. Geomagn. Geoelec.",
  ["journal of geometric analysis"] = "J. Geom. Anal.",
  ["journal of geometric mechanics"] = "J. Geom. Mech.",
  ["journal of geometry"] = "J. Geom.",
  ["journal of geometry and physics"] = "J. Geom. Phys.",
  ["journal of geometry and symmetry in physics"] = "J. Geom. Symmetry Phys.",
  ["journal of geophysical research"] = "J. Geophys. Res.",
  ["journal of geophysical research, [atmospheres]"] = "J. Geophys. Res. [Atmos.]",
  ["journal of geophysical research, [biogeosciences]"] = "J. Geophys. Res. [Biogeosci.]",
  ["journal of geophysical research, [earth surface]"] = "J. Geophys. Res. [Earth Surface]",
  ["journal of geophysical research, [oceans]"] = "J. Geophys. Res. [Oceans]",
  ["journal of geophysical research, [planets]"] = "J. Geophys. Res. [Planets]",
  ["journal of geophysical research, [solid earth]"] = "J. Geophys. Res. [Solid Earth]",
  ["journal of geophysical research, [space physics]"] = "J. Geophys. Res. [Space Phys.]",
  ["journal of geophysical research, a: space physics"] = "J. Geophys. Res. A: Space Phys.",
  ["journal of geophysical research, b: solid earth"] = "J. Geophys. Res. B: Solid Earth",
  ["journal of geophysical research, biogeosciences"] = "J. Geophys. Res. Biogeosci.",
  ["journal of geophysical research, c: oceans"] = "J. Geophys. Res. C: Oceans",
  ["journal of geophysical research, d: atmospheres"] = "J. Geophys. Res. D: Atmos.",
  ["journal of geophysical research, e: planets"] = "J. Geophys. Res. E: Planets",
  ["journal of geophysical research, earth surface"] = "J. Geophys. Res. Earth Surf.",
  ["journal of geophysical research, f: earth surface"] = "J. Geophys. Res. F: Earth Surf.",
  ["journal of geophysical research, g: biogeosciences"] = "J. Geophys. Res. G: Biogeosci.",
  ["journal of geophysical research, oceans"] = "J. Geophys. Res. Oceans",
  ["journal of geophysical research, planets"] = "J. Geophys. Res. Planets",
  ["journal of geophysical research, solid earth"] = "J. Geophys. Res. Solid Earth",
  ["journal of geophysical research, space physics"] = "J. Geophys. Res. Space Phys.",
  ["journal of geophysical research. atmospheres : jgr"] = "J Geophys Res Atmos",
  ["journal of geophysical research. biogeosciences"] = "J Geophys Res Biogeosci",
  ["journal of geophysical research. earth surface"] = "J Geophys Res Earth Surf",
  ["journal of geophysical research. oceans"] = "J Geophys Res Oceans",
  ["journal of geophysical research. planets"] = "J Geophys Res Planets",
  ["journal of geophysical research. solid earth"] = "J Geophys Res Solid Earth",
  ["journal of geophysical research. space physics"] = "J Geophys Res Space Phys",
  ["journal of geophysical research: atmospheres"] = "J. Geophys. Res.: Atmos.",
  ["journal of geophysical research: biogeosciences"] = "J. Geophys. Res.: Biogeosci.",
  ["journal of geophysical research: earth surface"] = "J. Geophys. Res.: Earth Surf.",
  ["journal of geophysical research: machine learning and computation"] = "J. Geophys. Res.: Mach. Learn. Comput.",
  ["journal of geophysical research: oceans"] = "J. Geophys. Res.: Oceans",
  ["journal of geophysical research: planets"] = "J. Geophys. Res.: Planets",
  ["journal of geophysical research: solid earth"] = "J. Geophys. Res.: Solid Earth",
  ["journal of geophysical research: space physics"] = "J. Geophys. Res.: Space Phys.",
  ["journal of geophysics"] = "J. Geophys.",
  ["journal of geophysics and engineering"] = "J. Geophys. Eng.",
  ["journal of geoscience education"] = "J. Geosci. Educ.",
  ["journal of geosciences"] = "J. Geosci.",
  ["journal of geotechnical and geoenvironmental engineering"] = "J. Geotech. Geoenviron. Eng.",
  ["journal of geovisualization and spatial analysis"] = "J. Geovisualization Spatial Anal.",
  ["journal of geriatric cardiology : jgc"] = "J Geriatr Cardiol",
  ["journal of geriatric care and research"] = "J Geriatr Care Res",
  ["journal of geriatric medicine and gerontology"] = "J Geriatr Med Gerontol",
  ["journal of geriatric oncology"] = "J Geriatr Oncol",
  ["journal of geriatric physical therapy"] = "J. Geriatr. Phys. Ther.",
  ["journal of geriatric physical therapy (2001)"] = "J Geriatr Phys Ther",
  ["journal of geriatric psychiatry"] = "J Geriatr Psychiatry",
  ["journal of geriatric psychiatry and neurology"] = "J Geriatr Psychiatry Neurol",
  ["journal of geriatrics"] = "J Geriatr",
  ["journal of geriatrics and palliative care"] = "J Geriatr Palliat Care",
  ["journal of gerontological nursing"] = "J Gerontol Nurs",
  ["journal of gerontological social work"] = "J Gerontol Soc Work",
  ["journal of gerontology"] = "J Gerontol",
  ["journal of gerontology & geriatric research"] = "J Gerontol Geriatr Res",
  ["journal of ginseng research"] = "J Ginseng Res",
  ["journal of glaciology"] = "J. Glaciol.",
  ["journal of glaciology and geocryology"] = "J Glaciol Geocryol",
  ["journal of glass studies"] = "J Glass Stud",
  ["journal of glaucoma"] = "J Glaucoma",
  ["journal of glbt family studies"] = "J GLBT Fam Stud",
  ["journal of global antimicrobial resistance"] = "J Glob Antimicrob Resist",
  ["journal of global change data & discovery"] = "J. Global Change Data Discovery",
  ["journal of global economic analysis"] = "J Glob Econ Anal",
  ["journal of global ethics"] = "J Glob Ethics",
  ["journal of global health"] = "J Glob Health",
  ["journal of global health perspectives"] = "J Glob Health Perspect",
  ["journal of global health reports"] = "J Glob Health Rep",
  ["journal of global history"] = "J Glob Hist",
  ["journal of global infectious diseases"] = "J Glob Infect Dis",
  ["journal of global information technology management"] = "J. Global Inf. Technol. Manage.",
  ["journal of global oncology"] = "J Glob Oncol",
  ["journal of global optimization"] = "J. Global Optim.",
  ["journal of global optimization : an international journal dealing with theoretical and computational aspects of seeking global optima and their applications in science, management and engineering"] = "J Glob Optim",
  ["journal of global optimization. an international journal dealing with theoretical and computational aspects of seeking global optima and their applications in science, management and engineering"] = "J. Global Optim.",
  ["journal of glycobiology"] = "J Glycobiol",
  ["journal of glycomics & lipidomics"] = "J Glycomics Lipidomics",
  ["journal of gnathology"] = "J. Gnathol.",
  ["journal of gokova geometry topology. ggt"] = "J. Gokova Geom. Topol. GGT",
  ["journal of government information : an international review of policy, issues and resources"] = "J Gov Inf",
  ["journal of graduate medical education"] = "J Grad Med Educ",
  ["journal of graph algorithms and applications"] = "J. Graph Algorithms Appl.",
  ["journal of graph labeling. a graph theoretical research journal"] = "J. Graph Label.",
  ["journal of graph theory"] = "J. Graph Theory",
  ["journal of gravitational physiology : a journal of the international society for gravitational physiology"] = "J Gravit Physiol",
  ["journal of great lakes research"] = "J. Great Lakes Res.",
  ["journal of greco-roman studies"] = "JGRS",
  ["journal of grid computing"] = "J Grid Comput",
  ["journal of group psychotherapy, psychodrama and sociometry"] = "J Group Psychother Psychodrama Sociom",
  ["journal of group theory"] = "J. Group Theory",
  ["journal of groups in addiction & recovery"] = "J Groups Addict Recover",
  ["journal of groups, complexity, cryptology"] = "J. Groups Complex. Cryptol.",
  ["journal of guidance, control, and dynamics"] = "J. Guid. Control Dyn.",
  ["journal of guidance, control, and dynamics : a publication of the american institute of aeronautics and astronautics devoted to the technology of dynamics and control"] = "J Guid Control Dyn",
  ["journal of gynaecological endocrinology"] = "J Gynaecol Endocrinol",
  ["journal of gynecologic oncology"] = "J Gynecol Oncol",
  ["journal of gynecologic surgery"] = "J Gynecol Surg",
  ["journal of gynecological endoscopy and surgery"] = "J Gynecol Endosc Surg",
  ["journal of gynecology obstetrics and human reproduction"] = "J Gynecol Obstet Hum Reprod",
  ["journal of gynecology research"] = "J Gynecol Res",
  ["journal of halacha and contemporary society"] = "J Halacha Contemporary Society",
  ["journal of hand and microsurgery"] = "J Hand Microsurg",
  ["journal of hand surgery (edinburgh, scotland)"] = "J Hand Surg Br",
  ["journal of hand surgery global online"] = "J Hand Surg Glob Online",
  ["journal of hand surgery. american volume"] = "J. Hand Surg. [Am.]",
  ["journal of hand surgery. british volume"] = "J. Hand Surg. [Br.]",
  ["journal of hand therapy"] = "J. Hand Ther.",
  ["journal of hand therapy : official journal of the american society of hand therapists"] = "J Hand Ther",
  ["journal of hangzhou university"] = "J. Hangzhou Univ. Natur. Sci. Ed.",
  ["journal of happiness studies"] = "J Happiness Stud",
  ["journal of harbin institute of technology. harbin gongye daxue xuebao"] = "J. Harbin Inst. Tech.",
  ["journal of hard tissue biology"] = "J Hard Tissue Biol",
  ["journal of hardware and systems security"] = "J. Hardware Syst. Secur.",
  ["journal of harmonized research in medical and health sciences"] = "J Harmon Res Med Health Sci",
  ["journal of hazardous materials"] = "J. Hazard. Mater.",
  ["journal of hazardous materials advances"] = "J. Hazard. Mater. Adv.",
  ["journal of hazardous materials letters"] = "J. Hazard. Mater. Lett.",
  ["journal of hazardous, toxic and radioactive waste"] = "J Hazard Toxic Radioact Waste",
  ["journal of hazardous, toxic, and radioactive waste"] = "J. Hazard. Toxic Radioact. Waste",
  ["journal of head trauma rehabilitation"] = "J. Head Trauma Rehabil.",
  ["journal of head, neck & spine surgery"] = "J Head Neck Spine Surg",
  ["journal of health & biomedical law"] = "J Health Biomed Law",
  ["journal of health & life sciences law"] = "J Health Life Sci Law",
  ["journal of health & mass communication"] = "J Health Mass Commun",
  ["journal of health & medical economics"] = "J Health Med Econ",
  ["journal of health & medical informatics"] = "J Health Med Inform",
  ["journal of health & pollution"] = "J Health Pollut",
  ["journal of health & population in developing countries"] = "J Health Popul Dev Ctries",
  ["journal of health & social policy"] = "J Health Soc Policy",
  ["journal of health administration education"] = "J. Health Adm. Educ.",
  ["journal of health and hospital law"] = "J. Health Hosp. Law",
  ["journal of health and hospital law : a publication of the american academy of hospital attorneys of the american hospital association"] = "J Health Hosp Law",
  ["journal of health and human behavior"] = "J Health Hum Behav",
  ["journal of health and human resources administration"] = "J Health Hum Resour Adm",
  ["journal of health and human services administration"] = "J Health Hum Serv Adm",
  ["journal of health and social behavior"] = "J Health Soc Behav",
  ["journal of health and social policy"] = "J. Health Soc. Policy",
  ["journal of health and social sciences"] = "J Health Soc Sci",
  ["journal of health behavior and public health"] = "J Health Behav Public Health",
  ["journal of health care benefits"] = "J Health Care Benefits",
  ["journal of health care chaplaincy"] = "J Health Care Chaplain",
  ["journal of health care finance"] = "J Health Care Finance",
  ["journal of health care for the poor and underserved"] = "J Health Care Poor Underserved",
  ["journal of health care interior design"] = "J. Health Care Inter. Des.",
  ["journal of health care law & policy"] = "J Health Care Law Policy",
  ["journal of health care marketing"] = "J Health Care Mark",
  ["journal of health care technology"] = "J Health Care Technol",
  ["journal of health communication"] = "J Health Commun",
  ["journal of health disparities research and practice"] = "J Health Dispar Res Pract",
  ["journal of health economics"] = "J Health Econ",
  ["journal of health economics and development"] = "J Health Econ Dev",
  ["journal of health economics and outcomes research"] = "J Health Econ Outcomes Res",
  ["journal of health education"] = "J Health Educ",
  ["journal of health education / association for the advancement of health education"] = "J Health Educ",
  ["journal of health education research & development"] = "J Health Educ Res Dev",
  ["journal of health education teaching"] = "J Health Educ Teach",
  ["journal of health informatics"] = "J Health Inform",
  ["journal of health informatics in developing countries"] = "J Health Inform Dev Ctries",
  ["journal of health law"] = "J Health Law",
  ["journal of health management"] = "J Health Manag",
  ["journal of health organization and management"] = "J Health Organ Manag",
  ["journal of health politics, policy and law"] = "J Health Polit Policy Law",
  ["journal of health psychology"] = "J Health Psychol",
  ["journal of health research"] = "J Health Res",
  ["journal of health research and reviews in developing countries"] = "J Health Res Rev Dev Ctries",
  ["journal of health science & education"] = "J Health Sci Educ",
  ["journal of health science (el monte)"] = "J Health Sci (El Monte)",
  ["journal of health science (tokyo, japan)"] = "J Health Sci",
  ["journal of health sciences and surveillance system"] = "J Health Sci Surveill Syst",
  ["journal of health service psychology"] = "J Health Serv Psychol",
  ["journal of health services research & policy"] = "J Health Serv Res Policy",
  ["journal of health services research and policy"] = "J. Health Serv. Res. Policy",
  ["journal of health, population and nutrition"] = "J. Health Popul. Nutr.",
  ["journal of health, population, and nutrition"] = "J Health Popul Nutr",
  ["journal of healthcare communications"] = "J Healthc Commun",
  ["journal of healthcare design"] = "J. Healthc. Des.",
  ["journal of healthcare education and training"] = "J. Healthc. Educ. Train.",
  ["journal of healthcare education and training : the journal of the american society for healthcare education and training"] = "J Healthc Educ Train",
  ["journal of healthcare education and training : the journal of the american society for healthcare education and training / american hospital association"] = "J Healthc Educ Train",
  ["journal of healthcare engineering"] = "J Healthc Eng",
  ["journal of healthcare informatics research"] = "J Healthc Inform Res",
  ["journal of healthcare information management"] = "J. Healthc. Inf. Manag.",
  ["journal of healthcare information management : jhim"] = "J Healthc Inf Manag",
  ["journal of healthcare leadership"] = "J Healthc Leadersh",
  ["journal of healthcare management"] = "J. Healthc. Manag.",
  ["journal of healthcare management / american college of healthcare executives"] = "J Healthc Manag",
  ["journal of healthcare materiel management"] = "J Healthc Mater Manage",
  ["journal of healthcare protection management"] = "J. Healthc. Prot. Manage.",
  ["journal of healthcare protection management : publication of the international association for hospital security"] = "J Healthc Prot Manage",
  ["journal of healthcare quality research"] = "J Healthc Qual Res",
  ["journal of healthcare resource management"] = "J Healthc Resour Manag",
  ["journal of healthcare risk management"] = "J. Healthc. Risk Manag.",
  ["journal of healthcare risk management : the journal of the american society for healthcare risk management"] = "J Healthc Risk Manag",
  ["journal of healthcare, science and the humanities"] = "J Healthc Sci Humanit",
  ["journal of hearing science"] = "J Hear Sci",
  ["journal of heart and cardiology"] = "J Heart Cardiol",
  ["journal of heart and lung transplantation"] = "J. Heart Lung Transplant.",
  ["journal of heart health"] = "J Heart Health",
  ["journal of heart transplantation"] = "J. Heart Transplant.",
  ["journal of heart valve disease"] = "J. Heart Valve Dis.",
  ["journal of heat recovery systems"] = "J. Heat Recovery Syst.",
  ["journal of heat transfer"] = "J. Heat Transfer",
  ["journal of heat treating"] = "J. Heat. Treat.",
  ["journal of hefei university of technology. natural science. hefei gongye daxue xuebao. ziran kexue ban"] = "J. Hefei Univ. Technol. Nat. Sci.",
  ["journal of hellenic studies"] = "JHS",
  ["journal of helminthology"] = "J Helminthol",
  ["journal of hematological malignancies"] = "J Hematol Malig",
  ["journal of hematology"] = "J Hematol",
  ["journal of hematology & oncology"] = "J Hematol Oncol",
  ["journal of hematology & thrombosis"] = "J Hematol Thromb",
  ["journal of hematology & transfusion"] = "J Hematol Transfus",
  ["journal of hematology and oncology"] = "J. Hematol. Oncol.",
  ["journal of hematology oncology pharmacy"] = "J Hematol Oncol Pharm",
  ["journal of hematology research"] = "J Hematol Res",
  ["journal of hematopathology"] = "J Hematop",
  ["journal of hematotherapy"] = "J Hematother",
  ["journal of hematotherapy & stem cell research"] = "J Hematother Stem Cell Res",
  ["journal of hematotherapy and stem cell research"] = "J. Hematother. Stem Cell Res.",
  ["journal of hepato-biliary-pancreatic sciences"] = "J Hepatobiliary Pancreat Sci",
  ["journal of hepato-biliary-pancreatic surgery"] = "J Hepatobiliary Pancreat Surg",
  ["journal of hepatocellular carcinoma"] = "J Hepatocell Carcinoma",
  ["journal of hepatology"] = "J Hepatol",
  ["journal of hepatology. supplement"] = "J Hepatol Suppl",
  ["journal of hepatology. supplement / easl"] = "J Hepatol Suppl",
  ["journal of herbal medicine"] = "J Herb Med",
  ["journal of herbal pharmacotherapy"] = "J Herb Pharmacother",
  ["journal of herbs spices & medicinal plants"] = "J. Herbs Spices Med Plants",
  ["journal of herbs, spices & medicinal plants"] = "J Herbs Spices Med Plants",
  ["journal of heredity"] = "J. Hered.",
  ["journal of heritage management"] = "J. Heritage Manage.",
  ["journal of herpetological medicine and surgery"] = "J Herpetol Med Surg",
  ["journal of herpetology"] = "J Herpetol",
  ["journal of heterocyclic chemistry"] = "J. Heterocycl. Chem.",
  ["journal of heuristics"] = "J. Heuristics",
  ["journal of high energy astrophysics"] = "J. High Energy Astrophys.",
  ["journal of high energy physics"] = "J. High Energy Phys.",
  ["journal of high energy physics : jhep"] = "J High Energy Phys",
  ["journal of high resolution chromatography"] = "J. High. Resolut. Chromatogr.",
  ["journal of high resolution chromatography : hrc"] = "J High Resolut Chromatogr",
  ["journal of high school research in the clinical sciences"] = "J High School Res Clin Sci",
  ["journal of high speed networks"] = "J. High Speed Networks",
  ["journal of high technology management research"] = "J. High Technol. Manage. Res.",
  ["journal of higher education outreach and engagement"] = "J High Educ Outreach Engagem",
  ["journal of higher education policy and management"] = "J. Higher Educ. Pol. Manage.",
  ["journal of highway and transportation research and development (english edition)"] = "J. Highway Transp. Res. Dev.",
  ["journal of hill research"] = "J Hill Res",
  ["journal of hip preservation surgery"] = "J Hip Preserv Surg",
  ["journal of hispanic higher education"] = "J Hispanic High Educ",
  ["journal of histochemistry and cytochemistry"] = "J. Histochem. Cytochem.",
  ["journal of historical geography"] = "J Hist Geogr",
  ["journal of historical sociology"] = "J Hist Sociol",
  ["journal of histotechnology"] = "J Histotechnol",
  ["journal of hiv & retro virus"] = "J HIV Retro Virus",
  ["journal of hiv aids surveillance & epidemiology"] = "J HIV AIDS Surveill Epidemiol",
  ["journal of hiv and aids"] = "J HIV AIDS",
  ["journal of hiv for clinical and scientific research"] = "J HIV Clin Sci Res",
  ["journal of hiv therapy"] = "J HIV Ther",
  ["journal of hiv/aids & social services"] = "J HIV AIDS Soc Serv",
  ["journal of hiv/aids prevention & education for adolescents & children"] = "J HIV/AIDS Prev Educ Adolesc Child",
  ["journal of hiv/aids prevention in children & youth"] = "J HIV AIDS Prev Child Youth",
  ["journal of hohai university"] = "J. Hohai Univ.",
  ["journal of hokkaido university of education"] = "J. Hokkaido Univ. Ed. Sect. II A",
  ["journal of holistic nursing"] = "J. Holist. Nurs.",
  ["journal of holistic nursing : official journal of the american holistic nurses' association"] = "J Holist Nurs",
  ["journal of holistic nursing : official journal of the american holistic nurses’ association"] = "J Holist Nurs",
  ["journal of holography and speckle"] = "J. Hologr. Speckle",
  ["journal of home economics"] = "J Home Econ",
  ["journal of homeland security and emergency management"] = "J Homel Secur Emerg Manag",
  ["journal of homosexuality"] = "J Homosex",
  ["journal of homotopy and related structures"] = "J. Homotopy Relat. Struct.",
  ["journal of horticultural science and biotechnology"] = "J. Hortic. Sci. Biotechnol.",
  ["journal of hospice and palliative nursing : jhpn : the official journal of the hospice and palliative nurses association"] = "J Hosp Palliat Nurs",
  ["journal of hospital administration"] = "J Hosp Adm",
  ["journal of hospital dental practice"] = "J Hosp Dent Pract",
  ["journal of hospital infection"] = "J. Hosp. Infect.",
  ["journal of hospital librarianship"] = "J Hosp Librariansh",
  ["journal of hospital management and health policy"] = "J Hosp Manag Health Policy",
  ["journal of hospital marketing"] = "J Hosp Mark",
  ["journal of hospital marketing & public relations"] = "J Hosp Mark Public Relations",
  ["journal of hospital medicine"] = "J Hosp Med",
  ["journal of hospital medicine (online)"] = "J Hosp Med",
  ["journal of hospital supply, processing and distribution"] = "J. Hosp. Supply. Process. Distrib.",
  ["journal of hospital supply, processing, and distribution"] = "J Hosp Supply Process Distrib",
  ["journal of hospitality, leisure, sport & tourism education"] = "J Hosp Leis Sport Tour Educ",
  ["journal of housing and the built environment"] = "J. Hous. Built Environ.",
  ["journal of housing and the built environment : hbe"] = "J Hous Built Environ",
  ["journal of housing economics"] = "J Hous Econ",
  ["journal of housing for the elderly"] = "J Hous Elderly",
  ["journal of housing research"] = "J. Hous. Res.",
  ["journal of huaqiao university"] = "J. Huaqiao Univ. Nat. Sci. Ed.",
  ["journal of huazhong university of science and technology. medical sciences"] = "J. Huazhong Univ. Sci. Technolog. Med. Sci.",
  ["journal of huazhong university of science and technology. medical sciences = hua zhong ke ji da xue xue bao. yi xue ying de wen ban = huazhong keji daxue xuebao. yixue yingdewen ban"] = "J Huazhong Univ Sci Technolog Med Sci",
  ["journal of huazhong university of science and technology. natural science edition. huazhong keji daxue xuebao. ziran kexue ban"] = "J. Huazhong Univ. Sci. Technol. Nat. Sci.",
  ["journal of human & clinical genetics"] = "J Hum Clin Genet",
  ["journal of human behavior in the social environment"] = "J Hum Behav Soc Environ",
  ["journal of human capital"] = "J Hum Cap",
  ["journal of human development and capabilities"] = "J Human Dev Capabil",
  ["journal of human ecology"] = "J. Hum. Ecol.",
  ["journal of human ecology (delhi, india)"] = "J Hum Ecol",
  ["journal of human ergology"] = "J Hum Ergol (Tokyo)",
  ["journal of human evolution"] = "J Hum Evol",
  ["journal of human genetics"] = "J Hum Genet",
  ["journal of human hypertension"] = "J Hum Hypertens",
  ["journal of human immunity"] = "J. Hum. Immun.",
  ["journal of human kinetics"] = "J Hum Kinet",
  ["journal of human lactation"] = "J. Hum. Lact.",
  ["journal of human lactation : official journal of international lactation consultant association"] = "J Hum Lact",
  ["journal of human nutrition"] = "J Hum Nutr",
  ["journal of human nutrition (carson city, nev.)"] = "J Hum Nutr (Carson City)",
  ["journal of human nutrition and dietetics"] = "J. Hum. Nutr. Diet.",
  ["journal of human nutrition and dietetics : the official journal of the british dietetic association"] = "J Hum Nutr Diet",
  ["journal of human nutrition and food science"] = "J Hum Nutr Food Sci",
  ["journal of human reproductive sciences"] = "J Hum Reprod Sci",
  ["journal of human resources"] = "J. Human Res.",
  ["journal of human rights"] = "J Hum Rights",
  ["journal of human rights and social work"] = "J Hum Rights Soc Work",
  ["journal of human rights and the environment"] = "J Hum Rights Environ",
  ["journal of human rights practice"] = "J Hum Rights Pract",
  ["journal of human security"] = "J Hum Secur",
  ["journal of human stress"] = "J Human Stress",
  ["journal of human trafficking"] = "J Hum Traffick",
  ["journal of human transcriptome"] = "J Hum Transcr",
  ["journal of human virology"] = "J Hum Virol",
  ["journal of human virology & retrovirology"] = "J Hum Virol Retrovirol",
  ["journal of human-robot interaction"] = "J Hum Robot Interact",
  ["journal of humanistic mathematics"] = "J. Humanist. Math.",
  ["journal of humanistic psychology"] = "J Humanist Psychol",
  ["journal of hunan university. natural sciences. hunan daxue xuebao. ziran kexue ban"] = "J. Hunan Univ. Nat. Sci.",
  ["journal of hunger & environmental nutrition"] = "J Hunger Environ Nutr",
  ["journal of huntington's disease"] = "J Huntingtons Dis",
  ["journal of huntington’s disease"] = "J. Huntington’s Dis.",
  ["journal of hydraulic engineering"] = "J. Hydraul. Eng.",
  ["journal of hydraulic research"] = "J. Hydraul. Res.",
  ["journal of hydro-environment research"] = "J. Hydro-environ. Res.",
  ["journal of hydrodynamics"] = "J. Hydrodyn.",
  ["journal of hydrodynamics, series b"] = "J. Hydrodyn. Ser. B",
  ["journal of hydrodynamics. ser. b"] = "J Hydrodynam B",
  ["journal of hydroinformatics"] = "J. Hydroinf.",
  ["journal of hydrologic engineering"] = "J. Hydrol. Eng.",
  ["journal of hydrology"] = "J. Hydrol.",
  ["journal of hydrology and hydromechanics"] = "J. Hydrol. Hydromech.",
  ["journal of hydrology new zealand"] = "J. Hydrol. N.Z.",
  ["journal of hydrology x"] = "J. Hydrol. X",
  ["journal of hydrology. regional studies"] = "J Hydrol Reg Stud",
  ["journal of hydrology: regional studies"] = "J. Hydrol.: Reg. Stud.",
  ["journal of hydrology: x"] = "J Hydrol X",
  ["journal of hydrometeorology"] = "J Hydrometeorol",
  ["journal of hygiene"] = "J. Hyg. (Lond.)",
  ["journal of hygiene, epidemiology, microbiology and immunology"] = "J. Hyg. Epidemiol. Microbiol. Immunol.",
  ["journal of hygiene, epidemiology, microbiology, and immunology"] = "J Hyg Epidemiol Microbiol Immunol",
  ["journal of hymenoptera research"] = "J Hymenopt Res",
  ["journal of hyperbolic differential equations"] = "J. Hyperbolic Differ. Equations",
  ["journal of hyperstructures"] = "J. Hyperstruct.",
  ["journal of hypertension"] = "J Hypertens",
  ["journal of hypertension : open access"] = "J Hypertens (Los Angel)",
  ["journal of hypertension and management"] = "J Hypertens Manag",
  ["journal of hypertension. supplement"] = "J. Hypertens. Suppl.",
  ["journal of hypertension. supplement : official journal of the international society of hypertension"] = "J Hypertens Suppl",
  ["journal of iberian and latin american studies"] = "J Iber Lat Am Stud",
  ["journal of iberian archaeology"] = "JIbA",
  ["journal of iberian geology"] = "J. Iber. Geol.",
  ["journal of ichthyology"] = "J Ichthyol",
  ["journal of image guided surgery"] = "J Image Guid Surg",
  ["journal of imaging"] = "J Imaging",
  ["journal of imaging science"] = "J Imaging Sci",
  ["journal of imaging science and technology"] = "J. Imaging Sci. Technol.",
  ["journal of immigrant & refugee studies"] = "J Immigr Refug Stud",
  ["journal of immigrant and minority health"] = "J Immigr Minor Health",
  ["journal of immigrant and minority health / center for minority public health"] = "J Immigr Minor Health",
  ["journal of immigrant health"] = "J Immigr Health",
  ["journal of immune based therapies and vaccines"] = "J Immune Based Ther Vaccines",
  ["journal of immune based therapies, vaccines and antimicrobials"] = "J Immune Based Ther Vaccines Antimicrob",
  ["journal of immune disorders & therapy"] = "J Immune Disord Ther",
  ["journal of immunoassay"] = "J Immunoassay",
  ["journal of immunoassay & immunochemistry"] = "J Immunoassay Immunochem",
  ["journal of immunoassay and immunochemistry"] = "J. Immunoassay Immunochem.",
  ["journal of immunobiology"] = "J Immunobiol",
  ["journal of immunodeficiency & disorders"] = "J immunodefic Disord",
  ["journal of immunogenetics"] = "J Immunogenet",
  ["journal of immunological methods"] = "J Immunol Methods",
  ["journal of immunological sciences"] = "J Immunol Sci",
  ["journal of immunological techniques in infectious diseases"] = "J Immunol Tech Infect Dis",
  ["journal of immunology"] = "J. Immunol.",
  ["journal of immunology (baltimore, md. : 1950)"] = "J Immunol",
  ["journal of immunology and clinical research"] = "J Immunol Clin Res",
  ["journal of immunology and immuno-techniques"] = "J Immunol Immunotech",
  ["journal of immunology and infectious diseases"] = "J Immunol Infect Dis",
  ["journal of immunology and regenerative medicine"] = "J Immunol Regen Med",
  ["journal of immunology research"] = "J Immunol Res",
  ["journal of immunology research and therapy"] = "J Immunol Res Ther",
  ["journal of immunopharmacology"] = "J Immunopharmacol",
  ["journal of immunotherapy"] = "J. Immunother.",
  ["journal of immunotherapy (hagerstown, md. : 1997)"] = "J Immunother",
  ["journal of immunotherapy : official journal of the society for biological therapy"] = "J Immunother (1991)",
  ["journal of immunotherapy and precision oncology"] = "J Immunother Precis Oncol",
  ["journal of immunotherapy applications"] = "J Immunother Appl",
  ["journal of immunotherapy with emphasis on tumor immunology"] = "J. Immunother. Emphasis Tumor Immunol.",
  ["journal of immunotherapy with emphasis on tumor immunology : official journal of the society for biological therapy"] = "J Immunother Emphasis Tumor Immunol",
  ["journal of immunotoxicology"] = "J Immunotoxicol",
  ["journal of in vitro fertilization and embryo transfer"] = "J. In Vitro Fert. Embryo Transf.",
  ["journal of in vitro fertilization and embryo transfer : ivf"] = "J In Vitro Fert Embryo Transf",
  ["journal of inborn errors of metabolism and screening"] = "J Inborn Errors Metab Screen",
  ["journal of inclusion phenomena and macrocyclic chemistry"] = "J. Inclusion Phenom. Macrocyclic Chem.",
  ["journal of inclusion phenomena and molecular recognition in chemistry"] = "J. Inclusion Phenom. Mol. Recognit. Chem.",
  ["journal of income distribution"] = "J. Income Distribution",
  ["journal of indian academy of dental specialist researchers"] = "J Indian Acad Dent Spec Res",
  ["journal of indian academy of forensic medicine"] = "J. Indian Acad. Forensic Med.",
  ["journal of indian association for communicable diseases"] = "J Indian Assoc Commun Dis",
  ["journal of indian association of pediatric surgeons"] = "J Indian Assoc Pediatr Surg",
  ["journal of indian council of chemists"] = "J. Indian Counc. Chem.",
  ["journal of indian philosophy"] = "J Indian Philos",
  ["journal of indian prosthodontic society"] = "J Indian Prosthodont Soc",
  ["journal of indian society of periodontology"] = "J Indian Soc Periodontol",
  ["journal of indian society of statistics and operations research"] = "J. Indian Soc. Statist. Oper. Res.",
  ["journal of indigenous social development"] = "J Indig Soc Dev",
  ["journal of indigenous wellbeing : te mauri-pimatisiwin"] = "J Indig Wellbeing",
  ["journal of individual differences"] = "J Individ Differ",
  ["journal of individual psychology"] = "J Individ Psychol",
  ["journal of individual psychology (1998)"] = "J Individ Psychol (1998)",
  ["journal of indo-european studies"] = "JIES",
  ["journal of indonesian natural history"] = "J Indones Nat Hist",
  ["journal of industrial and engineering chemistry"] = "J. Ind. Eng. Chem.",
  ["journal of industrial and engineering chemistry (amsterdam, netherlands)"] = "J. Ind. Eng. Chem. (Amsterdam, Neth.)",
  ["journal of industrial and engineering chemistry (seoul, korea)"] = "J Ind Eng Chem",
  ["journal of industrial and engineering chemistry (seoul, republic of korea)"] = "J. Ind. Eng. Chem. (Seoul, Repub. Korea)",
  ["journal of industrial and management optimization"] = "J. Ind. Manage. Optim.",
  ["journal of industrial and production engineering"] = "J. Ind. Prod. Eng.",
  ["journal of industrial ecology"] = "J. Ind. Ecol.",
  ["journal of industrial economics"] = "J. Ind. Econ.",
  ["journal of industrial engineering international"] = "J. Ind. Eng. Int.",
  ["journal of industrial information integration"] = "J Ind Inf Integr",
  ["journal of industrial integration and management"] = "J. Ind. Integr. Manage.",
  ["journal of industrial microbiology"] = "J. Ind. Microbiol.",
  ["journal of industrial microbiology & biotechnology"] = "J Ind Microbiol Biotechnol",
  ["journal of industrial microbiology and biotechnology"] = "J. Ind. Microbiol. Biotechnol.",
  ["journal of industrial pollution control"] = "J. Ind. Pollut. Control",
  ["journal of industrial textiles"] = "J. Ind. Text.",
  ["journal of industry studies"] = "J. Industry Stud.",
  ["journal of inequalities and applications"] = "J. Inequal. Appl.",
  ["journal of inequalities and special functions"] = "J. Inequal. Spec. Funct.",
  ["journal of infant, child, and adolescent psychotherapy : jicap"] = "J Infant Child Adolesc Psychother",
  ["journal of infectiology"] = "J Infect (Grand Rapids)",
  ["journal of infection"] = "J. Infect.",
  ["journal of infection and chemotherapy"] = "J. Infect. Chemother.",
  ["journal of infection and chemotherapy : official journal of the japan society of chemotherapy"] = "J Infect Chemother",
  ["journal of infection and public health"] = "J Infect Public Health",
  ["journal of infection in developing countries"] = "J Infect Dev Ctries",
  ["journal of infection prevention"] = "J Infect Prev",
  ["journal of infectious disease and therapy"] = "J Infect Dis Ther",
  ["journal of infectious diseases"] = "J. Infect. Dis.",
  ["journal of infectious diseases and medicine"] = "J Infect Dis Med",
  ["journal of infectious pulmonary diseases"] = "J Infect Pulm Dis",
  ["journal of inflammation"] = "J Inflamm",
  ["journal of inflammation (london, england)"] = "J Inflamm (Lond)",
  ["journal of inflammation research"] = "J Inflamm Res",
  ["journal of inflammatory bowel diseases & disorders"] = "J Inflamm Bowel Dis Disord",
  ["journal of informatics"] = "J. Inf.",
  ["journal of informatics nursing"] = "J Inform Nurs",
  ["journal of information & knowledge management"] = "J. Inf. Knowl. Manage.",
  ["journal of information & optimization sciences"] = "J. Inform. Optim. Sci.",
  ["journal of information and image management"] = "J. Inf. Image Manage.",
  ["journal of information and intelligence"] = "J. Inf. Intell.",
  ["journal of information and optimization sciences"] = "J. Inf. Optim. Sci.",
  ["journal of information and telecommunication"] = "J. Inf. Telecommun.",
  ["journal of information display"] = "J. Inf. Disp.",
  ["journal of information ethics"] = "J Infor Ethics",
  ["journal of information processing systems"] = "J. Inf. Process. Syst.",
  ["journal of information science"] = "J Inf Sci",
  ["journal of information security and applications"] = "J. Inf. Secur. Appl.",
  ["journal of information technology"] = "J. Inf. Technol.",
  ["journal of information technology & software engineering"] = "J Inf Technol Softw Eng",
  ["journal of information technology case and application research"] = "J. Inf. Technol. Case Appl. Res.",
  ["journal of information technology in construction"] = "J. Inf. Technol. Construct.",
  ["journal of informetrics"] = "J Informetr",
  ["journal of infrared and millimeter waves"] = "J. Infrared Millimeter Waves",
  ["journal of infrared, millimeter and terahertz waves"] = "J Infrared Millim Terahertz Waves",
  ["journal of infrared, millimeter, and terahertz waves"] = "J. Infrared Millim. Terahertz Waves",
  ["journal of infrastructure development"] = "J. Infrastruct. Dev.",
  ["journal of infrastructure intelligence and resilience"] = "J. Infrastruct. Intell. Resilience",
  ["journal of infrastructure preservation and resilience"] = "J. Infrastruct. Preserv. Resilience",
  ["journal of infrastructure systems"] = "J. Infrastruct. Syst.",
  ["journal of infusion nursing : the official publication of the infusion nurses society"] = "J Infus Nurs",
  ["journal of infusional chemotherapy"] = "J. Infus. Chemother.",
  ["journal of inherited metabolic disease"] = "J Inherit Metab Dis",
  ["journal of injury & violence research"] = "J Inj Violence Res",
  ["journal of innate immunity"] = "J Innate Immun",
  ["journal of inner mongolia normal university. nei mongol shifan daxue xuebao. ziran kexue. hanwen ban"] = "J. Inn. Mong. Norm. Univ. Nat. Sci.",
  ["journal of inner mongolia university. natural science edition"] = "J. Inn. Mong. Univ. Nat. Sci.",
  ["journal of innovation and entrepreneurship"] = "J Innov Entrep",
  ["journal of innovation and knowledge"] = "J. Innovation Knowl.",
  ["journal of innovation in digital ecosystems"] = "J. Innovation Digital Ecosyst.",
  ["journal of innovation in health informatics"] = "J Innov Health Inform",
  ["journal of innovations in pharmaceutical and biological sciences"] = "J Innov Pharm Biol Sci",
  ["journal of innovative biology"] = "J Innov Biol",
  ["journal of innovative optical health sciences"] = "J Innov Opt Health Sci",
  ["journal of inorganic and nuclear chemistry"] = "J. Inorg. Nucl. Chem.",
  ["journal of inorganic and organometallic polymers"] = "J. Inorg. Organomet. Polym.",
  ["journal of inorganic and organometallic polymers and materials"] = "J. Inorg. Organomet. Polym. Mater.",
  ["journal of inorganic biochemistry"] = "J Inorg Biochem",
  ["journal of insect behavior"] = "J Insect Behav",
  ["journal of insect biodiversity"] = "J Insect Biodivers",
  ["journal of insect biodiversity and systematics"] = "J Insect Biodivers Syst",
  ["journal of insect biotechnology and sericology"] = "J Insect Biotechnol Sericology",
  ["journal of insect conservation"] = "J Insect Conserv",
  ["journal of insect physiology"] = "J Insect Physiol",
  ["journal of insect science"] = "J. Insect Sci.",
  ["journal of insect science (online)"] = "J Insect Sci",
  ["journal of insects"] = "J Insects",
  ["journal of insects as food and feed"] = "J. Insects Food Feed",
  ["journal of institute of chemists (india)"] = "J. Inst. Chem.",
  ["journal of institute of economic research. j.s.s. institute of economic research"] = "J Inst Econ Res",
  ["journal of institutional and theoretical economics"] = "J. Inst. Theoretical Econ.",
  ["journal of institutional and theoretical economics : jite = zeitschrift für die gesamte staatswissenschaft"] = "J Inst Theor Econ",
  ["journal of instrumentation"] = "J. Instrum.",
  ["journal of instrumentation : an iop and sissa journal"] = "J Instrum",
  ["journal of insurance medicine"] = "J Insur Med",
  ["journal of insurance medicine (new york, n.y.)"] = "J Insur Med",
  ["journal of integer sequences"] = "J Integer Seq",
  ["journal of integrable systems"] = "J. Integrable Syst.",
  ["journal of integral equations and applications"] = "J. Integral Equations Appl.",
  ["journal of integral equations and mathematical physics"] = "J. Integral Equations Math. Phys.",
  ["journal of integrated care (brighton, england)"] = "J Integr Care (Brighton)",
  ["journal of integrated design and process science"] = "J. Integr. Des. Process Sci.",
  ["journal of integrated omics"] = "J Integr OMICS",
  ["journal of integrated pest management"] = "J Integr Pest Manag",
  ["journal of integrative agriculture"] = "J Integr Agric",
  ["journal of integrative bioinformatics"] = "J Integr Bioinform",
  ["journal of integrative cardiology"] = "J Integr Cardiol",
  ["journal of integrative cardiology open access"] = "J Integr Cardiol Open Access",
  ["journal of integrative environmental sciences"] = "J. Integr. Environ. Sci.",
  ["journal of integrative medicine"] = "J Integr Med",
  ["journal of integrative neuroscience"] = "J Integr Neurosci",
  ["journal of integrative oncology"] = "J Integr Oncol",
  ["journal of integrative plant biology"] = "J Integr Plant Biol",
  ["journal of intellectual & developmental disability"] = "J Intellect Dev Disabil",
  ["journal of intellectual disabilities"] = "J. Intellect. Disabil.",
  ["journal of intellectual disabilities : joid"] = "J Intellect Disabil",
  ["journal of intellectual disabilities and offending behaviour"] = "J Intellect Disabil Offending Behav",
  ["journal of intellectual disability research"] = "J. Intellect. Disabil. Res.",
  ["journal of intellectual disability research : jidr"] = "J Intellect Disabil Res",
  ["journal of intelligence"] = "J Intell",
  ["journal of intelligence history"] = "J. Intell. Hist.",
  ["journal of intelligent & fuzzy systems"] = "J. Intell. Fuzzy Syst.",
  ["journal of intelligent & robotic systems"] = "J Intell Robot Syst",
  ["journal of intelligent and connected vehicles"] = "J. Intell. Connected Veh.",
  ["journal of intelligent and robotic systems"] = "J. Intell. Rob. Syst.",
  ["journal of intelligent information systems"] = "J Intell Inf Syst",
  ["journal of intelligent manufacturing"] = "J Intell Manuf",
  ["journal of intelligent manufacturing and special equipment"] = "J. Intell. Manuf. Spec. Equip.",
  ["journal of intelligent material systems and structures"] = "J. Intell. Mater. Syst. Struct.",
  ["journal of intelligent medicine and healthcare"] = "J. Intell. Med. Healthcare",
  ["journal of intelligent tranportation systems: technology, planning, and operations"] = "J. Intell. Transp. Syst. Technol. Plann. Oper.",
  ["journal of intelligent transportation systems"] = "J Intell Transp Syst",
  ["journal of intelligent transportation systems: technology, planning, and operations"] = "J. Intell. Transp. Syst. Technol. Plann. Oper.",
  ["journal of intensive and critical care"] = "J Intensive Crit Care",
  ["journal of intensive care"] = "J Intensive Care",
  ["journal of intensive care medicine"] = "J Intensive Care Med",
  ["journal of interaction science"] = "J Interact Sci",
  ["journal of interactional research in communication disorders"] = "J Interact Res Commun Disord",
  ["journal of interamerican studies and world affairs"] = "J Inter Am Stud World Aff",
  ["journal of interconnection networks"] = "J. Interconnection Networks",
  ["journal of intercultural communication research"] = "J Intercult Commun Res",
  ["journal of intercultural ethnopharmacology"] = "J Intercult Ethnopharmacol",
  ["journal of intercultural studies (melbourne, vic.)"] = "J Intercult Stud",
  ["journal of interdisciplinary biology"] = "J Interdiscip Biol",
  ["journal of interdisciplinary cycle research"] = "J Interdiscipl Cycle Res",
  ["journal of interdisciplinary economics"] = "J. Interdisciplinary Econ.",
  ["journal of interdisciplinary histopathology"] = "J Interdiscip Histopathol",
  ["journal of interdisciplinary mathematics"] = "J. Interdiscip. Math.",
  ["journal of interdisciplinary nanomedicine"] = "J Interdiscip Nanomed",
  ["journal of interferon & cytokine research"] = "J. Interferon Cytokine Res.",
  ["journal of interferon & cytokine research : the official journal of the international society for interferon and cytokine research"] = "J Interferon Cytokine Res",
  ["journal of interferon and cytokine research"] = "J. Interferon Cytokine Res.",
  ["journal of interferon research"] = "J. Interferon Res.",
  ["journal of intergenerational relationships"] = "J Intergener Relatsh",
  ["journal of interior design"] = "J Inter Des",
  ["journal of internal medicine"] = "J Intern Med",
  ["journal of internal medicine. supplement"] = "J Intern Med Suppl",
  ["journal of international academy of physical sciences"] = "J. Int. Acad. Phys. Sci.",
  ["journal of international affairs"] = "J Int Aff",
  ["journal of international and comparative economics"] = "J. Int. Compar. Econ.",
  ["journal of international business studies"] = "J Int Bus Stud",
  ["journal of international consumer marketing"] = "J Int Consum Mark",
  ["journal of international cooperation"] = "J Int Coop",
  ["journal of international council on electrical engineering"] = "J. Int. Counc. Electr. Eng.",
  ["journal of international development"] = "J Int Dev",
  ["journal of international economic law"] = "J. Int. Econ. Law",
  ["journal of international economics"] = "J Int Econ",
  ["journal of international financial markets, institutions and money"] = "J. Int. Finan. Markets",
  ["journal of international life sciences research"] = "J. Int. Life Sci. Res.",
  ["journal of international maritime safety, environmental affairs, and shipping"] = "J. Int. Marit. Saf. Environ. Aff. Ship.",
  ["journal of international mathematical virtual institute"] = "J. Int. Math. Virtual Inst.",
  ["journal of international medical research"] = "J. Int. Med. Res.",
  ["journal of international migration and integration"] = "J Int Migr Integr",
  ["journal of international money and finance"] = "J Int Money Finance",
  ["journal of international oral health : jioh"] = "J Int Oral Health",
  ["journal of international relations and development"] = "J Int Relat Dev (Ljubl)",
  ["journal of international research communications"] = "J. Int. Res. Commun.",
  ["journal of international society of preventive & community dentistry"] = "J Int Soc Prev Community Dent",
  ["journal of international studies and development"] = "J Int Stud Dev",
  ["journal of international trade and economic development"] = "J. Int. Trade Econ. Devel.",
  ["journal of international wildlife law & policy"] = "J. Int. Wildl. Law Policy",
  ["journal of international women's studies"] = "J Int Womens Stud",
  ["journal of internet services and applications"] = "J. Internet Serv. Appl.",
  ["journal of interpersonal violence"] = "J Interpers Violence",
  ["journal of interpolation and approximation in scientific computing"] = "J. Interpolat. Approx. Sci. Comput.",
  ["journal of interprofessional care"] = "J Interprof Care",
  ["journal of interprofessional education & practice"] = "J Interprof Educ Pract",
  ["journal of interventional cardiac electrophysiology"] = "J. Interventional Card. Electrophysiol.",
  ["journal of interventional cardiac electrophysiology : an international journal of arrhythmias and pacing"] = "J Interv Card Electrophysiol",
  ["journal of interventional cardiology"] = "J Interv Cardiol",
  ["journal of interventional gastroenterology"] = "J Interv Gastroenterol",
  ["journal of interventional oncology"] = "J Interv Oncol",
  ["journal of intravenous nursing"] = "J. Intraven. Nurs.",
  ["journal of intravenous nursing : the official publication of the intravenous nurses society"] = "J Intraven Nurs",
  ["journal of invasive cardiology"] = "J. Invasive Cardiol.",
  ["journal of inverse and ill-posed problems"] = "J. Inverse Ill-Posed Probl.",
  ["journal of invertebrate pathology"] = "J Invertebr Pathol",
  ["journal of investigational allergology & clinical immunology"] = "J Investig Allergol Clin Immunol",
  ["journal of investigational allergology & clinical immunology : official organ of the international association of asthmology (interasma) and sociedad latinoamericana de alergia e inmunologia"] = "J Investig Allergol Clin Immunol",
  ["journal of investigational allergology and clinical immunology"] = "J. Investig. Allergol. Clin. Immunol.",
  ["journal of investigative and clinical dentistry"] = "J. Invest. Clin. Dent.",
  ["journal of investigative dermatology"] = "J. Invest. Dermatol.",
  ["journal of investigative dermatology. symposium proceedings"] = "J. Investig. Dermatol. Symp. Proc.",
  ["journal of investigative medicine"] = "J. Invest. Med.",
  ["journal of investigative medicine : the official publication of the american federation for clinical research"] = "J Investig Med",
  ["journal of investigative medicine high impact case reports"] = "J Investig Med High Impact Case Rep",
  ["journal of investigative surgery"] = "J. Invest. Surg.",
  ["journal of investigative surgery : the official journal of the academy of surgical research"] = "J Invest Surg",
  ["journal of ionic liquids"] = "J. Ionic Liq.",
  ["journal of iron and steel research international"] = "J. Iron. Steel Res. Int.",
  ["journal of irrigation and drainage engineering"] = "J. Irrig. Drain. Eng.",
  ["journal of isakos : joint disorders & orthopaedic sports medicine"] = "J ISAKOS",
  ["journal of islamic dental association of iran"] = "J Islam Dent Assoc Iran",
  ["journal of island and coastal archaeology"] = "J. Isl. Coastal Archaeol.",
  ["journal of issues in intercollegiate athletics"] = "J Issues Intercoll Athl",
  ["journal of istanbul university faculty of dentistry"] = "J Istanb Univ Fac Dent",
  ["journal of j. j. group of hospitals and grant medical college"] = "J Jj Group Hosp Grant Med Coll",
  ["journal of japanese botany"] = "J. Jpn. Bot.",
  ["journal of jewish communal service"] = "J Jew Communal Serv",
  ["journal of jiangxi normal university"] = "J. Jiangxi Norm. Univ. Nat. Sci. Ed.",
  ["journal of jilin university. science edition. jilin daxue xuebao. lixue ban"] = "J. Jilin Univ. Sci.",
  ["journal of jinan university. natural science & medicine edition. jinan daxue xuebao. ziran kexue yu yixue ban"] = "J. Jinan Univ. Nat. Sci. Med.",
  ["journal of juristic papyrology"] = "JJP",
  ["journal of juvenile justice"] = "J Juv Justice",
  ["journal of kidney"] = "J Kidney",
  ["journal of kidney cancer and vhl"] = "J Kidney Cancer VHL",
  ["journal of king abdulaziz university: islamic economics"] = "J. King Abdulaziz Univ.: Islamic Econ.",
  ["journal of king saud university, agricultural sciences"] = "J. King Saud Univ. Agric. Sci.",
  ["journal of king saud university, computer and information sciences"] = "J. King Saud Univ. Comput. Inf. Sci.",
  ["journal of king saud university, engineering sciences"] = "J. King Saud Univ. Eng. Sci.",
  ["journal of king saud university, science"] = "J. King Saud Univ. Sci.",
  ["journal of king saud university. science"] = "J King Saud Univ Sci",
  ["journal of knee surgery"] = "J. Knee Surg.",
  ["journal of knot theory and its ramifications"] = "J. Knot Theory Ramifications",
  ["journal of korean academy of nursing"] = "J Korean Acad Nurs",
  ["journal of korean american medical association"] = "J Korean Am Med Assoc",
  ["journal of korean institute of metals and materials"] = "J. Korean Inst. Met. Mater.",
  ["journal of korean medical science"] = "J Korean Med Sci",
  ["journal of korean nature"] = "J. Korean Nat.",
  ["journal of korean neurosurgical society"] = "J Korean Neurosurg Soc",
  ["journal of korean society for atmospheric environment"] = "J. Korean Soc. Atmos. Environ.",
  ["journal of labelled compounds"] = "J Labelled Compd",
  ["journal of labelled compounds & radiopharmaceuticals"] = "J Labelled Comp Radiopharm",
  ["journal of labelled compounds and radiopharmaceuticals"] = "J. Labelled Compd. Radiopharm.",
  ["journal of labor economics"] = "J Labor Econ",
  ["journal of labor research"] = "J Labor Res",
  ["journal of laboratory and clinical medicine"] = "J. Lab. Clin. Med.",
  ["journal of laboratory and precision medicine"] = "J Lab Precis Med",
  ["journal of laboratory automation"] = "J Lab Autom",
  ["journal of laboratory chemical education"] = "J Lab Chem Educ",
  ["journal of laboratory medicine"] = "J. Lab. Med.",
  ["journal of laboratory physicians"] = "J Lab Physicians",
  ["journal of land use science"] = "J Land Use Sci",
  ["journal of landscape architecture"] = "J. Landscape Archit.",
  ["journal of landscape ecology"] = "J. Landscape Ecol.",
  ["journal of language and social psychology"] = "J Lang Soc Psychol",
  ["journal of language evolution"] = "J Lang Evol",
  ["journal of lanzhou railway institute"] = "J. Lanzhou Railway Inst.",
  ["journal of lanzhou university"] = "J. Lanzhou Univ. Nat. Sci.",
  ["journal of lao studies"] = "J Lao Stud",
  ["journal of laparoendoscopic & advanced surgical techniques. part a"] = "J Laparoendosc Adv Surg Tech A",
  ["journal of laparoendoscopic & advanced surgical techniques. part b, videoscopy"] = "J Laparoendosc Adv Surg Tech Part B Videoscop",
  ["journal of laparoendoscopic and advanced surgical techniques. part a"] = "J. Laparoendosc. Adv. Surg. Tech. A",
  ["journal of laparoendoscopic surgery"] = "J Laparoendosc Surg",
  ["journal of laryngology and otology"] = "J. Laryngol. Otol.",
  ["journal of laryngology and otology. supplement"] = "J. Laryngol. Otol. Suppl.",
  ["journal of laser applications"] = "J. Laser Appl.",
  ["journal of laser micro/nanoengineering"] = "J. Laser Micro/Nanoeng.",
  ["journal of lasers in medical sciences"] = "J Lasers Med Sci",
  ["journal of lasers, optics & photonics"] = "J Lasers Opt Photonics",
  ["journal of latin american geriatric medicine"] = "J Lat Am Geriatr Med",
  ["journal of latin american lore"] = "J Lat Am Lore",
  ["journal of latin american studies"] = "J Lat Am Stud",
  ["journal of latina/o psychology"] = "J Lat Psychol",
  ["journal of latinos and education"] = "J Lat Educ",
  ["journal of law and economics"] = "J. Law Econ.",
  ["journal of law and ethics in dentistry"] = "J Law Ethics Dent",
  ["journal of law and health"] = "J Law Health",
  ["journal of law and medicine"] = "J Law Med",
  ["journal of law and policy"] = "J Law Policy",
  ["journal of law and society"] = "J Law Soc",
  ["journal of law and the biosciences"] = "J Law Biosci",
  ["journal of law enforcement leadership and ethics"] = "J Law Enforc Leadersh Ethics",
  ["journal of law, economics, & organization"] = "J Law Econ Organ",
  ["journal of law, economics, and organization"] = "J. Law",
  ["journal of law, medicine and ethics"] = "J. Law Med. Ethics",
  ["journal of leadership & organizational studies"] = "J Leadersh Organ Stud",
  ["journal of learning disabilities"] = "J Learn Disabil",
  ["journal of leather and chemical technology"] = "J. Leather Chem. Technol.",
  ["journal of leather science and engineering"] = "J. Leather Sci. Eng.",
  ["journal of legal affairs and dispute resolution in engineering and construction"] = "J. Leg. Aff. Dispute Resolut. Eng. Constr.",
  ["journal of legal economics"] = "J. Legal Econ.",
  ["journal of legal education"] = "J Legal Educ",
  ["journal of legal medicine"] = "J. Legal Med.",
  ["journal of legal studies"] = "J. Legal Stud.",
  ["journal of legislation"] = "J Legis",
  ["journal of leisure research"] = "J Leis Res",
  ["journal of lesbian studies"] = "J Lesbian Stud",
  ["journal of less-common metals"] = "J. Less-Common Met.",
  ["journal of leukemia (los angeles, calif.)"] = "J Leuk (Los Angel)",
  ["journal of leukocyte biology"] = "J Leukoc Biol",
  ["journal of leukocyte biology. supplement"] = "J Leukoc Biol Suppl",
  ["journal of lgbt health research"] = "J LGBT Health Res",
  ["journal of lgbt issues in counseling"] = "J LGBT Issues Couns",
  ["journal of lgbt youth"] = "J LGBT Youth",
  ["journal of librarianship"] = "J Libr",
  ["journal of librarianship and scholarly communication"] = "J Libr Sch Commun",
  ["journal of library administration"] = "J Libr Adm",
  ["journal of library history (tallahassee, fla. : 1974)"] = "J Libr Hist",
  ["journal of library metadata"] = "J Libr Metadata",
  ["journal of lie theory"] = "J. Lie Theory",
  ["journal of life care planning"] = "J Life Care Plan",
  ["journal of life science"] = "J Life Sci",
  ["journal of life sciences"] = "J. Life Sci.",
  ["journal of life sciences (libertyville, ill.)"] = "J Life Sci (Libertyville)",
  ["journal of life sciences (westlake village, calif.)"] = "J Life Sci (Westlake Village)",
  ["journal of lifestyle medicine"] = "J Lifestyle Med",
  ["journal of light & visual environment"] = "J. Light Visual Environ.",
  ["journal of light metals"] = "J. Light Met.",
  ["journal of lightwave technology"] = "J. Lightwave Technol.",
  ["journal of lightwave technology : a joint ieee/osa publication"] = "J Lightwave Technol",
  ["journal of limnology"] = "J Limnol",
  ["journal of linear and topological algebra"] = "J. Linear Topol. Algebra",
  ["journal of linguistic geography"] = "J. Ling. Geogr.",
  ["journal of linguistics"] = "J Linguist",
  ["journal of lipid and atherosclerosis"] = "J Lipid Atheroscler",
  ["journal of lipid mediators"] = "J Lipid Mediat",
  ["journal of lipid mediators and cell signalling"] = "J Lipid Mediat Cell Signal",
  ["journal of lipid research"] = "J Lipid Res",
  ["journal of lipid science and technology"] = "J. Lipid Sci. Technol.",
  ["journal of lipids"] = "J Lipids",
  ["journal of liposome research"] = "J Liposome Res",
  ["journal of liquid chromatography"] = "J Liq Chromatogr",
  ["journal of liquid chromatography & related technologies"] = "J Liq Chromatogr Relat Technol",
  ["journal of liquid chromatography and related technologies"] = "J. Liq. Chromatogr. Related Technol.",
  ["journal of literacy research : jlr"] = "J Lit Res",
  ["journal of literature and science"] = "J Lit Sci",
  ["journal of lithotripsy and stone disease"] = "J. Lithotr. Stone Dis.",
  ["journal of liver"] = "J Liver",
  ["journal of liver research, disorders & therapy"] = "J Liver Res Disord Ther",
  ["journal of liver: disease & transplantation"] = "J Liver Disease Transplant",
  ["journal of location based services"] = "J. Locat. Based Serv.",
  ["journal of logic and algebraic programming"] = "J. Log. Algebraic Program.",
  ["journal of logic and analysis"] = "J. Log. Anal.",
  ["journal of logic and computation"] = "J. Logic Comput.",
  ["journal of logic programming"] = "J. Log. Program.",
  ["journal of logic, language and information"] = "J. Logic Lang. Inf.",
  ["journal of logic, language, and information"] = "J Logic Lang Inf",
  ["journal of logical and algebraic methods in programming"] = "J. Logical Algebraic Methods Program.",
  ["journal of long term care administration"] = "J. Long Term Care Adm.",
  ["journal of long term home health care"] = "J. Long Term Home Health Care",
  ["journal of long-term effects of medical implants"] = "J Long Term Eff Med Implants",
  ["journal of loss & trauma"] = "J Loss Trauma",
  ["journal of loss prevention in the process industries"] = "J. Loss Prev. Process Ind.",
  ["journal of low frequency noise, vibration and active control"] = "J. Low Freq. Noise Vibr. Act. Control",
  ["journal of low frequency noise, vibration, and active control"] = "J. Low Freq. Noise Vib. Act. Control",
  ["journal of low power electronics"] = "J. Low Power Electron.",
  ["journal of low temperature physics"] = "J. Low Temp. Phys.",
  ["journal of lower genital tract disease"] = "J Low Genit Tract Dis",
  ["journal of lubrication technology"] = "J. Lubr. Technol.",
  ["journal of luminescence"] = "J. Lumin.",
  ["journal of lung cancer"] = "J Lung Cancer",
  ["journal of lung health and diseases"] = "J Lung Health Dis",
  ["journal of lung, pulmonary & respiratory research"] = "J Lung Pulm Respir Res",
  ["journal of lymphoedema"] = "J Lymphoedema",
  ["journal of machine learning"] = "J. Mach. Learn.",
  ["journal of machine learning for modeling and computing"] = "J. Mach. Learn. Model. Comput.",
  ["journal of machine learning research"] = "J. Mach. Learn. Res.",
  ["journal of machine learning research (jmlr)"] = "J. Mach. Learn. Res.",
  ["journal of machine learning research : jmlr"] = "J Mach Learn Res",
  ["journal of machining and forming technologies"] = "J. Mach. Forming Technol.",
  ["journal of macroeconomics"] = "J Macroecon",
  ["journal of macromolecular science - physics"] = "J. Macromol. Sci., Phys.",
  ["journal of macromolecular science - reviews in macromolecular chemistry and physics"] = "J. Macromol. Sci., Rev. Macromol. Chem. Phys.",
  ["journal of macromolecular science polymer reviews"] = "J. Macromol. Sci., Polym. Rev.",
  ["journal of macromolecular science pure and applied chemistry"] = "J. Macromol. Sci., Pure Appl. Chem.",
  ["journal of macromolecular science, [part b] physics"] = "J. Macromol. Sci. Phys.",
  ["journal of macromolecular science, part a pure and applied chemistry"] = "J. Macromol. Sci. A Pure Appl. Chem.",
  ["journal of macromolecular science, part a: pure and applied chemistry"] = "J. Macromol. Sci., Part A: Pure Appl. Chem.",
  ["journal of macromolecular science, part b physics"] = "J. Macromol. Sci. B Phys.",
  ["journal of macromolecular science, part b: physics"] = "J. Macromol. Sci., Part B: Phys.",
  ["journal of macromolecular science, part c: polymer reviews"] = "J. Macromol. Sci., Part C: Polym. Rev.",
  ["journal of macromolecular science, physics"] = "J. Macromol. Sci., Phys.",
  ["journal of macromolecular science, polymer reviews"] = "J. Macromol. Sci., Polym. Rev.",
  ["journal of macromolecular science, pure and applied chemistry"] = "J. Macromol. Sci., Pure Appl. Chem.",
  ["journal of macromolecular science. physics"] = "J Macromol Sci Phys",
  ["journal of magnesium and alloys"] = "J. Magnesium Alloys",
  ["journal of magnetic resonance"] = "J. Magn. Reson.",
  ["journal of magnetic resonance (san diego, calif. : 1997)"] = "J Magn Reson",
  ["journal of magnetic resonance imaging"] = "J. Magn. Reson. Imaging",
  ["journal of magnetic resonance imaging : jmri"] = "J Magn Reson Imaging",
  ["journal of magnetic resonance open"] = "J Magn Reson Open",
  ["journal of magnetic resonance series a"] = "J. Magn. Reson. A",
  ["journal of magnetic resonance series b"] = "J. Magn. Reson. B",
  ["journal of magnetic resonance. series a"] = "J Magn Reson A",
  ["journal of magnetic resonance. series b"] = "J Magn Reson B",
  ["journal of magnetics"] = "J. Magn.",
  ["journal of magnetism and magnetic materials"] = "J. Magn. Magn. Mater.",
  ["journal of maintenance in the addictions"] = "J Maint Addict",
  ["journal of mammalian evolution"] = "J Mamm Evol",
  ["journal of mammalian ova research"] = "J Mamm Ova Res",
  ["journal of mammalogy"] = "J Mammal",
  ["journal of mammary gland biology and neoplasia"] = "J Mammary Gland Biol Neoplasia",
  ["journal of managed care & specialty pharmacy"] = "J Manag Care Spec Pharm",
  ["journal of managed care medicine : jmcm"] = "J Manag Care Med",
  ["journal of managed care pharmacy : jmcp"] = "J Manag Care Pharm",
  ["journal of management"] = "J Manage",
  ["journal of management & marketing in healthcare"] = "J Manag Marketing Healthc",
  ["journal of management accounting research"] = "J. Manage. Acc. Res.",
  ["journal of management and governance"] = "J. Manage. Governance",
  ["journal of management in engineering"] = "J. Manage. Eng.",
  ["journal of management in medicine"] = "J Manag Med",
  ["journal of management information systems : jmis"] = "J Manag Inf Syst",
  ["journal of management science and engineering"] = "J. Manage. Sci. Eng.",
  ["journal of management scientific reports"] = "J. Manage. Sci. Rep.",
  ["journal of manipulative and physiological therapeutics"] = "J Manipulative Physiol Ther",
  ["journal of manufacturing and materials processing"] = "J. Manuf. Mater. Process.",
  ["journal of manufacturing processes"] = "J. Manuf. Processes",
  ["journal of manufacturing science and engineering"] = "J. Manuf. Sci. Eng.",
  ["journal of manufacturing science and production"] = "J. Manuf. Sci. Prod.",
  ["journal of manufacturing systems"] = "J. Manuf. Syst.",
  ["journal of manufacturing technology management"] = "J. Manuf. Technol. Manage.",
  ["journal of manufacturing technology research"] = "J. Manuf. Technol. Res.",
  ["journal of map & geography libraries"] = "J Map Geogr Libr",
  ["journal of maps"] = "J Maps",
  ["journal of marine and island cultures"] = "J. Mar. Isl. Cult.",
  ["journal of marine biology"] = "J Mar Biol",
  ["journal of marine biotechnology"] = "J Mar Biotechnol",
  ["journal of marine engineering & technology"] = "J. Mar. Eng. Technol.",
  ["journal of marine research"] = "J. Mar. Res.",
  ["journal of marine science and application"] = "J. Mar. Sci. Appl.",
  ["journal of marine science and engineering"] = "J Mar Sci Eng",
  ["journal of marine science and technology"] = "J. Mar. Sci. Technol.",
  ["journal of marine sciences"] = "J. Mar. Sci.",
  ["journal of marine systems"] = "J. Mar. Syst.",
  ["journal of marine systems : journal of the european association of marine sciences and techniques"] = "J Mar Syst",
  ["journal of marital and family therapy"] = "J Marital Fam Ther",
  ["journal of maritime archaeology"] = "J. Marit. Archaeol.",
  ["journal of maritime research"] = "J. Marit. Res.",
  ["journal of market access & health policy"] = "J Mark Access Health Policy",
  ["journal of marketing"] = "J Mark",
  ["journal of marketing for mental health"] = "J Mark Ment Health",
  ["journal of marketing for professions"] = "J Mark Prof",
  ["journal of marketing research"] = "J. Marketing Res.",
  ["journal of markets and morality"] = "J. Markets Morality",
  ["journal of marmara university dental faculty"] = "J Marmara Univ Dent Fac",
  ["journal of marriage and family counseling"] = "J Marriage Fam Couns",
  ["journal of marriage and the family"] = "J Marriage Fam",
  ["journal of mass spectrometry"] = "J. Mass Spectrom.",
  ["journal of mass spectrometry : jms"] = "J Mass Spectrom",
  ["journal of mass spectrometry and advances in the clinical lab"] = "J. Mass Spectrom. Adv. Clin. Lab",
  ["journal of material cycles and waste management"] = "J. Mater. Cycles Waste Manage.",
  ["journal of materials"] = "J Mater",
  ["journal of materials (new york, ny, united states)"] = "J. Mater. (N. Y., NY, U. S.)",
  ["journal of materials and engineering structures"] = "J. Mater. Eng. Struct.",
  ["journal of materials and environmental science"] = "J. Mater. Environ. Sci.",
  ["journal of materials chemistry"] = "J. Mater. Chem.",
  ["journal of materials chemistry a"] = "J. Mater. Chem.\tA",
  ["journal of materials chemistry a: materials for energy and sustainability"] = "J. Mater. Chem. A",
  ["journal of materials chemistry b"] = "J. Mater. Chem. B",
  ["journal of materials chemistry b: materials for biology and medicine"] = "J. Mater. Chem. B",
  ["journal of materials chemistry c"] = "J. Mater. Chem. C",
  ["journal of materials chemistry c: materials for optical and electronic devices"] = "J. Mater. Chem. C",
  ["journal of materials chemistry. a"] = "J Mater Chem A Mater",
  ["journal of materials chemistry. b"] = "J Mater Chem B",
  ["journal of materials chemistry. c"] = "J Mater Chem C Mater",
  ["journal of materials education"] = "J. Mater. Educ.",
  ["journal of materials engineering"] = "J. Mater. Eng.",
  ["journal of materials engineering and performance"] = "J. Mater. Eng. Perform.",
  ["journal of materials for energy systems"] = "J. Mater. Energy Syst.",
  ["journal of materials in civil engineering"] = "J. Mater. Civ. Eng.",
  ["journal of materials processing and manufacturing science"] = "J. Mater. Process. Manuf. Sci.",
  ["journal of materials processing technology"] = "J. Mater. Process. Technol.",
  ["journal of materials research"] = "J. Mater. Res.",
  ["journal of materials research and technology"] = "J. Mater. Res. Technol.",
  ["journal of materials science"] = "J. Mater. Sci.",
  ["journal of materials science & technology"] = "J Mater Sci Technol",
  ["journal of materials science - materials in electronics"] = "J. Mater. Sci. - Mater. Electron.",
  ["journal of materials science - materials in medicine"] = "J. Mater. Sci. - Mater. Med.",
  ["journal of materials science and engineering"] = "J. Mater. Sci. Eng",
  ["journal of materials science and engineering. a"] = "J Mater Sci Eng A",
  ["journal of materials science and technology"] = "J. Mater. Sci. Technol.",
  ["journal of materials science letters"] = "J Mater Sci Lett",
  ["journal of materials science materials in medicine"] = "J. Mater. Sci. - Mater. Med.",
  ["journal of materials science. materials in medicine"] = "J Mater Sci Mater Med",
  ["journal of materials science: materials in electronics"] = "J. Mater. Sci.: Mater. Electron.",
  ["journal of materials science: materials in medicine"] = "J. Mater. Sci.: Mater. Med.",
  ["journal of materials shaping technology"] = "J. Mater. Shaping Technol.",
  ["journal of materials synthesis and processing"] = "J. Mater. Synth. Process.",
  ["journal of materiomics"] = "J. Materiomics",
  ["journal of maternal-fetal and neonatal medicine"] = "J. Matern. Fetal Neonatal Med.",
  ["journal of maternal-fetal medicine"] = "J. Matern. Fetal Med.",
  ["journal of mathematical analysis"] = "J. Math. Anal.",
  ["journal of mathematical analysis and applications"] = "J. Math. Anal. Appl.",
  ["journal of mathematical and fundamental sciences"] = "J. Math. Fundam. Sci.",
  ["journal of mathematical and physical sciences"] = "J. Math. Phys. Sci.",
  ["journal of mathematical behavior"] = "J. Math. Behav.",
  ["journal of mathematical biology"] = "J Math Biol",
  ["journal of mathematical chemistry"] = "J. Math. Chem.",
  ["journal of mathematical cryptology"] = "J. Math. Cryptol.",
  ["journal of mathematical economics"] = "J. Math. Econ.",
  ["journal of mathematical fluid mechanics"] = "J. Math. Fluid Mech.",
  ["journal of mathematical imaging and vision"] = "J. Math. Imaging Vision",
  ["journal of mathematical inequalities"] = "J. Math. Inequal.",
  ["journal of mathematical logic"] = "J. Math. Logic",
  ["journal of mathematical modeling"] = "J. Math. Model.",
  ["journal of mathematical modelling and algorithms in operations research"] = "J. Math. Modell. Algorithms Oper. Res.",
  ["journal of mathematical neuroscience"] = "J Math Neurosci",
  ["journal of mathematical physics"] = "J. Math. Phys.",
  ["journal of mathematical physics (new york)"] = "J. Math. Phys. (N.Y.)",
  ["journal of mathematical physics analysis geometry"] = "J. Math. Phys. Anal. Geom.",
  ["journal of mathematical physics, analysis, geometry"] = "J. Math. Phys. Anal. Geom.",
  ["journal of mathematical psycholog y"] = "J. Math. Psych.",
  ["journal of mathematical psychology"] = "J Math Psychol",
  ["journal of mathematical research and exposition"] = "J. Math. Res. Exposition",
  ["journal of mathematical research with applications"] = "J. Math. Res. Appl.",
  ["journal of mathematical sciences"] = "J. Math. Sci.",
  ["journal of mathematical sciences (calcutta)"] = "J. Math. Sci. (Calcutta)",
  ["journal of mathematical sciences (dattapukur)"] = "J. Math. Sci. (Dattapukur)",
  ["journal of mathematical sciences (new york)"] = "J. Math. Sci. (N.Y.)",
  ["journal of mathematical sciences. advances and applications"] = "J. Math. Sci. Adv. Appl.",
  ["journal of mathematical sociology"] = "J. Math. Sociol.",
  ["journal of mathematical study"] = "J. Math. Study",
  ["journal of mathematical study. shuxue yanjiu"] = "J. Math. Study",
  ["journal of mathematical systems, estimation, and control"] = "J. Math. Systems Estim. Control",
  ["journal of mathematics"] = "J. Math.",
  ["journal of mathematics & music. mathematical and computational approaches to music theory, analysis, composition and performance"] = "J. Math. Music",
  ["journal of mathematics and applications"] = "J. Math. Appl.",
  ["journal of mathematics and music"] = "J. Math. Music",
  ["journal of mathematics and physics"] = "J. Math. Phys.",
  ["journal of mathematics and statistics"] = "J Math Stat",
  ["journal of mathematics and the arts"] = "J. Math. Arts",
  ["journal of mathematics in industry"] = "J. Math. Ind.",
  ["journal of mathematics of kyoto university"] = "J. Math. Kyoto Univ.",
  ["journal of mathematics teacher education"] = "J. Math. Teach. Educ.",
  ["journal of mathematics. tokushima university"] = "J. Math. Tokushima Univ.",
  ["journal of maxillofacial and oral surgery"] = "J Maxillofac Oral Surg",
  ["journal of maxillofacial orthopedics"] = "J Maxillofac Orthop",
  ["journal of maxillofacial surgery"] = "J Maxillofac Surg",
  ["journal of maxillofacial trauma"] = "J Maxillofac Trauma",
  ["journal of mechanical design"] = "J. Mech. Des.",
  ["journal of mechanical design (new york, n.y. : 1990)"] = "J Mech Des N Y",
  ["journal of mechanical engineering science"] = "J. Mech. Eng. Sci.",
  ["journal of mechanical science and technology"] = "J. Mech. Sci. Technol.",
  ["journal of mechanical working technology"] = "J. Mech. Work. Technol.",
  ["journal of mechanics"] = "J. Mech.",
  ["journal of mechanics in medicine and biology"] = "J Mech Med Biol",
  ["journal of mechanics of materials and structures"] = "J. Mech. Mater. Struct.",
  ["journal of mechanisms and robotics"] = "J Mech Robot",
  ["journal of mechanochemistry & cell motility"] = "J Mechanochem Cell Motil",
  ["journal of mechanochemistry and cell motility"] = "J. Mechanochem. Cell Motil.",
  ["journal of media business studies"] = "J Media Bus Stud",
  ["journal of media psychology"] = "J Media Psychol",
  ["journal of medical & surgical pathology"] = "J Med Surg Pathol",
  ["journal of medical and bioengineering"] = "J Med Bioeng",
  ["journal of medical and biological engineering"] = "J Med Biol Eng",
  ["journal of medical and dental sciences"] = "J Med Dent Sci",
  ["journal of medical and veterinary mycology"] = "J. Med. Vet. Mycol.",
  ["journal of medical and veterinary mycology : bi-monthly publication of the international society for human and animal mycology"] = "J Med Vet Mycol",
  ["journal of medical artificial intelligence"] = "J Med Artif Intell",
  ["journal of medical bacteriology"] = "J Med Bacteriol",
  ["journal of medical biochemistry"] = "J Med Biochem",
  ["journal of medical biography"] = "J Med Biogr",
  ["journal of medical case reports"] = "J Med Case Rep",
  ["journal of medical case reports and reviews"] = "J Med Case Rep Rev",
  ["journal of medical cases"] = "J Med Cases",
  ["journal of medical devices"] = "J. Med. Devices",
  ["journal of medical diagnostic methods"] = "J Med Diagn Methods",
  ["journal of medical diagnostics"] = "J. Med. Diagn.",
  ["journal of medical economics"] = "J Med Econ",
  ["journal of medical education"] = "J Med Educ",
  ["journal of medical education and curricular development"] = "J Med Educ Curric Dev",
  ["journal of medical education and training"] = "J Med Educ Train",
  ["journal of medical education research"] = "J. Med. Educ. Res.",
  ["journal of medical engineering"] = "J Med Eng",
  ["journal of medical engineering & technology"] = "J Med Eng Technol",
  ["journal of medical engineering and technology"] = "J. Med. Eng. Technol.",
  ["journal of medical entomology"] = "J Med Entomol",
  ["journal of medical entomology. supplement"] = "J Med Entomol Suppl",
  ["journal of medical ethics"] = "J Med Ethics",
  ["journal of medical ethics and history of medicine"] = "J Med Ethics Hist Med",
  ["journal of medical extended reality"] = "J. Med. Ext. Reality",
  ["journal of medical genetics"] = "J Med Genet",
  ["journal of medical humanities"] = "J. Med. Humanit.",
  ["journal of medical humanities and bioethics"] = "J. Med. Humanit. Bioeth.",
  ["journal of medical hypotheses & ideas"] = "J. Med. Hypotheses Ideas",
  ["journal of medical imaging"] = "J. Med. Imaging",
  ["journal of medical imaging (bellingham, wash.)"] = "J Med Imaging (Bellingham)",
  ["journal of medical imaging and health informatics"] = "J Med Imaging Health Inform",
  ["journal of medical imaging and radiation oncology"] = "J Med Imaging Radiat Oncol",
  ["journal of medical imaging and radiation sciences"] = "J Med Imaging Radiat Sci",
  ["journal of medical internet research"] = "J Med Internet Res",
  ["journal of medical investigation"] = "J. Med. Invest.",
  ["journal of medical laboratory technology"] = "J. Med. Lab. Technol.",
  ["journal of medical law and ethics"] = "J Med Law Ethics",
  ["journal of medical microbiology"] = "J Med Microbiol",
  ["journal of medical microbiology & diagnosis"] = "J Med Microbiol Diagn",
  ["journal of medical oncology and therapeutics"] = "J Med Oncol Ther",
  ["journal of medical pest control"] = "J. Med. Pest Control",
  ["journal of medical physics"] = "J Med Phys",
  ["journal of medical practice management"] = "J. Med. Pract. Manage.",
  ["journal of medical primatology"] = "J Med Primatol",
  ["journal of medical radiation sciences"] = "J Med Radiat Sci",
  ["journal of medical regulation"] = "J Med Regul",
  ["journal of medical research and innovation"] = "J Med Res Innov",
  ["journal of medical robotics research"] = "J Med Robot Res",
  ["journal of medical science"] = "J Med Sci",
  ["journal of medical sciences (faisalābād, pakistan)"] = "J Med Sci (Faisalabad, Pakistan)",
  ["journal of medical sciences (taipei, taiwan)"] = "J Med Sci",
  ["journal of medical screening"] = "J Med Screen",
  ["journal of medical signals and sensors"] = "J Med Signals Sens",
  ["journal of medical speech-language pathology"] = "J Med Speech Lang Pathol",
  ["journal of medical statistics and informatics"] = "J Med Stat Inform",
  ["journal of medical systems"] = "J Med Syst",
  ["journal of medical technology : official publication of american medical technologists [and] american society for medical technology"] = "J Med Technol",
  ["journal of medical toxicology"] = "J. Med. Toxicol.",
  ["journal of medical toxicology : official journal of the american college of medical toxicology"] = "J Med Toxicol",
  ["journal of medical ultrasonics"] = "J. Med. Ultrason.",
  ["journal of medical ultrasonics (2001)"] = "J Med Ultrason (2001)",
  ["journal of medical ultrasound"] = "J Med Ultrasound",
  ["journal of medical virology"] = "J Med Virol",
  ["journal of medicinal and pharmaceutical chemistry"] = "J Med Pharm Chem",
  ["journal of medicinal chemistry"] = "J Med Chem",
  ["journal of medicinal chemistry and drug design"] = "J Med Chem Drug Des",
  ["journal of medicinal food"] = "J Med Food",
  ["journal of medicinal plant research"] = "J Med Plant Res",
  ["journal of medicinal plants for economic development"] = "J. Med. Plants Econ. Dev.",
  ["journal of medicine"] = "J Med",
  ["journal of medicine and law"] = "J Med Law",
  ["journal of medicine and law / detroit college of law at michigan state university"] = "J Med Law",
  ["journal of medicine and life"] = "J Med Life",
  ["journal of medicine and philosophy"] = "J. Med. Philos.",
  ["journal of medicine and the person"] = "J Med Person",
  ["journal of medicine in the tropics"] = "J Med Trop",
  ["journal of medieval history"] = "J Mediev Hist",
  ["journal of medieval iberian studies"] = "J Mediev Iber Stud",
  ["journal of mediterranean anthropology and archaeology"] = "JMedAnthrA",
  ["journal of mediterranean archaeology"] = "JMA",
  ["journal of mega infrastructure & sustainable development"] = "J. Mega  Infrastruct. Sustainable Dev.",
  ["journal of melittology"] = "J Melittology",
  ["journal of membrane and separation technology"] = "J. Membr. Sep. Technol.",
  ["journal of membrane biology"] = "J. Membr. Biol.",
  ["journal of membrane computing"] = "J. Membr. Comput.",
  ["journal of membrane science"] = "J. Membr. Sci.",
  ["journal of membrane science & technology"] = "J Membr Sci Technol",
  ["journal of membrane science letters"] = "J. Membr. Sci. Lett.",
  ["journal of memory and language"] = "J Mem Lang",
  ["journal of men's health"] = "J Mens Health",
  ["journal of meningitis"] = "J Meningitis",
  ["journal of menopausal medicine"] = "J Menopausal Med",
  ["journal of mental deficiency research"] = "J Ment Defic Res",
  ["journal of mental deficiency research. monograph series"] = "J Ment Defic Res [Monogr Ser]",
  ["journal of mental health & clinical psychology"] = "J Ment Health Clin Psychol",
  ["journal of mental health (abingdon, england)"] = "J Ment Health",
  ["journal of mental health administration"] = "J Ment Health Adm",
  ["journal of mental health and aging"] = "J Ment Health Aging",
  ["journal of mental health counseling"] = "J Ment Health Couns",
  ["journal of mental health policy and economics"] = "J. Ment. Health Policy Econ.",
  ["journal of mental health research in intellectual disabilities"] = "J Ment Health Res Intellect Disabil",
  ["journal of metabolic syndrome"] = "J Metab Syndr",
  ["journal of metabolomics"] = "J Metabolomics",
  ["journal of metabolomics and systems biology"] = "J Metabolomics Syst Biol",
  ["journal of metabonomics & metabolites"] = "J Metabonomics Metab",
  ["journal of metallurg and materials science"] = "J. Metall. Mater. Sci.",
  ["journal of metallurgy and materials science"] = "J. Metall. Mater. Sci.",
  ["journal of metals, materials and minerals"] = "J. Met. Mater. Miner.",
  ["journal of metamorphic geology"] = "J. Metamorph. Geol.",
  ["journal of metastable and nanocrystalline materials"] = "J. Metastable Nanocryst. Mater.",
  ["journal of meteorological research"] = "J. Meteorolog. Res.",
  ["journal of meteorological research, japan"] = "J. Meteorol. Res. Japan",
  ["journal of meteorology"] = "J. Meteorol.",
  ["journal of methods and measurement in the social sciences"] = "J Methods Meas Soc Sci",
  ["journal of michigan dental association"] = "J. Mich. Dent. Assoc.",
  ["journal of micro - nano mechatronics"] = "J. Micro - Nano Mechatron.",
  ["journal of micro and nano-manufacturing"] = "J. Micro Nano-Manuf.",
  ["journal of micro- and nano-manufacturing"] = "J Micro Nanomanuf",
  ["journal of micro-bio robotics"] = "J Microbio Robot",
  ["journal of micro/nanolithography, mems and moems"] = "J. Micro/Nanolithogr. MEMS MOEMS",
  ["journal of micro/nanolithography, mems, and moems : jm3"] = "J Micro Nanolithogr MEMS MOEMS",
  ["journal of micro/nanolithography, microfabrican, and microsystems"] = "J. Micro/Nanolithog. Microfab. Microsys.",
  ["journal of micro/nanopatterning materials and metrology"] = "J. Micro/Nanopatterning Mater. Metrol.",
  ["journal of microbial & biochemical technology"] = "J Microb Biochem Technol",
  ["journal of microbiological methods"] = "J Microbiol Methods",
  ["journal of microbiology"] = "J. Microbiol.",
  ["journal of microbiology & biology education"] = "J Microbiol Biol Educ",
  ["journal of microbiology & experimentation"] = "J Microbiol Exp",
  ["journal of microbiology (seoul, korea)"] = "J Microbiol",
  ["journal of microbiology and antimicrobials"] = "J Microbiol Antimicrob",
  ["journal of microbiology and biotechnology"] = "J Microbiol Biotechnol",
  ["journal of microbiology and biotechnology research"] = "J Microbiol Biotechnol Res",
  ["journal of microbiology research"] = "J Microbiol Res (Rosemead, Calif)",
  ["journal of microbiology, biotechnology and food sciences"] = "J. Microbiol. Biotechnol. Food Sci.",
  ["journal of microbiology, epidemiology and immunobiology. zhurnal mikrobiologii, epidemiologii i immunobiologii"] = "J Microbiol Epidemiol Immunobiol",
  ["journal of microbiology, immunology and infection"] = "J. Microbiol. Immunol. Infect.",
  ["journal of microbiology, immunology, and infection"] = "J. Microbiol. Immunol. Infect.",
  ["journal of microbiology, immunology, and infection = wei mian yu gan ran za zhi"] = "J Microbiol Immunol Infect",
  ["journal of microcolumn separations"] = "J. Microcolumn Sep.",
  ["journal of microelectromechanical systems"] = "J. Microelectromech. Syst.",
  ["journal of microelectromechanical systems : a joint ieee and asme publication on microstructures, microactuators, microsensors, and microsystems"] = "J Microelectromech Syst",
  ["journal of microelectronic systems"] = "J. Microelectromech. Syst.",
  ["journal of microelectronics and electronic packaging"] = "J. Microelectron. Electron. Packag.",
  ["journal of microencapsulation"] = "J Microencapsul",
  ["journal of micrographics"] = "J. Microgr.",
  ["journal of microlithography, microfabrication, and microsystems"] = "J. Microlith. Microfab. Microsys.",
  ["journal of micromanufacturing"] = "J. Micromanuf.",
  ["journal of micromechanics and microengineering"] = "J. Micromech. Microeng.",
  ["journal of micromechanics and microengineering : structures, devices, and systems"] = "J Micromech Microeng",
  ["journal of micromechanics and molecular physics"] = "J Micromech Mol Phys",
  ["journal of micropalaeontology"] = "J Micropalaeontol",
  ["journal of micropaleontology"] = "J. Micropaleontol.",
  ["journal of microscopy"] = "J Microsc",
  ["journal of microscopy and ultrastructure"] = "J Microsc Ultrastruct",
  ["journal of microsurgery"] = "J Microsurg",
  ["journal of microwave power"] = "J. Microwave Power",
  ["journal of microwave power and electromagnetic energy"] = "J. Microwave Power Electromagn. Energy",
  ["journal of mid-life health"] = "J Midlife Health",
  ["journal of midwifery & women's health"] = "J Midwifery Womens Health",
  ["journal of midwifery & women’s health"] = "J Midwifery Womens Health",
  ["journal of midwifery and women’s health"] = "J. Midwifery Womens Health",
  ["journal of migration and health"] = "J Migr Health",
  ["journal of migration and refugee issues"] = "J Migr Refug Issues",
  ["journal of migration history"] = "J Migr Hist",
  ["journal of military and government counseling"] = "J Mil Gov Couns",
  ["journal of military and veterans' health"] = "J Mil Veterans Health",
  ["journal of military, veteran and family health"] = "J Mil Veteran Fam Health",
  ["journal of milk and food technology"] = "J. Milk Food Technol.",
  ["journal of mineralogical and petrological sciences"] = "J. Mineral. Petrol. Sci.",
  ["journal of mines, metals and fuels"] = "J. Mines Met. Fuels",
  ["journal of minimal access surgery"] = "J Minim Access Surg",
  ["journal of minimally invasive gynecology"] = "J Minim Invasive Gynecol",
  ["journal of mining and environment"] = "J. Min. Environ.",
  ["journal of mining and metallurgy, section a: mining"] = "J. Min. Metall. Sect. A.",
  ["journal of mining and metallurgy, section b: metallurgy"] = "J. Min. Metall. Sect. B.",
  ["journal of mining institute"] = "J. Min. Inst.",
  ["journal of mining science"] = "J. Min. Sci.",
  ["journal of mining science and technology"] = "J. Min. Sci. Technol.",
  ["journal of mithraic studies"] = "JMithrSt",
  ["journal of mitochondria, plastids and endosymbiosis"] = "J. Mitochondria, Plastids Endosymbiosis",
  ["journal of mixed methods research"] = "J Mix Methods Res",
  ["journal of mmij"] = "J. MMIJ",
  ["journal of mobile technology in medicine"] = "J Mob Technol Med",
  ["journal of modelling in management"] = "J. Modell. Manage.",
  ["journal of modern applied statistical methods : jmasm"] = "J Mod Appl Stat Methods",
  ["journal of modern dynamics"] = "J. Mod. Dyn.",
  ["journal of modern european history = zeitschrift für moderne europäische geschichte = revue d'histoire européenne contemporaine"] = "J Mod Eur Hist",
  ["journal of modern italian studies"] = "J Mod Ital Stud",
  ["journal of modern jewish studies"] = "J Mod Jew Stud",
  ["journal of modern mathematics and statistics"] = "J. Mod. Math. Stat.",
  ["journal of modern methods in numerical mathematics"] = "J. Mod. Methods Numer. Math.",
  ["journal of modern optics"] = "J. Mod. Opt.",
  ["journal of modern physiological research"] = "J Mod Physiol Res",
  ["journal of modern power systems and clean energy"] = "J. Mod. Power Syst. Clean Energy",
  ["journal of molecular and applied genetics"] = "J Mol Appl Genet",
  ["journal of molecular and cellular cardiology"] = "J Mol Cell Cardiol",
  ["journal of molecular and cellular immunology"] = "J. Mol. Cell. Immunol.",
  ["journal of molecular and cellular pathology"] = "J. Mol. Cell. Pathol.",
  ["journal of molecular and engineering materials"] = "J Mol Eng Mater",
  ["journal of molecular and genetic medicine"] = "J. Mol. Genet. Med.",
  ["journal of molecular and genetic medicine : an international journal of biomedical research"] = "J Mol Genet Med",
  ["journal of molecular biochemistry"] = "J Mol Biochem",
  ["journal of molecular biology"] = "J Mol Biol",
  ["journal of molecular biology & therapeutics"] = "J Mol Biol Ther",
  ["journal of molecular biology and methods"] = "J Mol Biol Methods",
  ["journal of molecular biology and molecular imaging"] = "J Mol Biol Mol Imaging",
  ["journal of molecular biology research"] = "J Mol Biol Res",
  ["journal of molecular biomarkers & diagnosis"] = "J Mol Biomark Diagn",
  ["journal of molecular catalysis"] = "J. Mol. Catal.",
  ["journal of molecular catalysis a: chemical"] = "J. Mol. Catal. A: Chem.",
  ["journal of molecular catalysis b: enzymatic"] = "J. Mol. Catal. B: Enzym.",
  ["journal of molecular catalysis. a, chemical"] = "J Mol Catal A Chem",
  ["journal of molecular catalysis. b, enzymatic"] = "J Mol Catal B Enzym",
  ["journal of molecular cell biology"] = "J Mol Cell Biol",
  ["journal of molecular cloning & genetic recombination"] = "J Mol Cloning Genet Recomb",
  ["journal of molecular diagnostics"] = "J. Mol. Diagn.",
  ["journal of molecular endocrinology"] = "J Mol Endocrinol",
  ["journal of molecular evolution"] = "J Mol Evol",
  ["journal of molecular genetics"] = "J Mol Genet",
  ["journal of molecular genetics (isleworth, london, england)"] = "J Mol Genet (Isleworth)",
  ["journal of molecular graphics"] = "J Mol Graph",
  ["journal of molecular graphics & modelling"] = "J Mol Graph Model",
  ["journal of molecular graphics and modelling"] = "J. Mol. Graphics Model.",
  ["journal of molecular histology"] = "J Mol Histol",
  ["journal of molecular imaging & dynamics"] = "J Mol Imaging Dyn",
  ["journal of molecular immunology"] = "J Mol Immunol",
  ["journal of molecular liquids"] = "J. Mol. Liq.",
  ["journal of molecular medicine"] = "J Mol Med",
  ["journal of molecular medicine (berlin, germany)"] = "J Mol Med (Berl)",
  ["journal of molecular medicine and clinical applications"] = "J Mol Med Clin Appl",
  ["journal of molecular microbiology and biotechnology"] = "J. Mol. Microbiol. Biotechnol.",
  ["journal of molecular modeling"] = "J Mol Model",
  ["journal of molecular neuroscience"] = "J. Mol. Neurosci.",
  ["journal of molecular neuroscience : mn"] = "J Mol Neurosci",
  ["journal of molecular pathology"] = "J. Mol. Pathol.",
  ["journal of molecular pharmaceutics & organic process research"] = "J Mol Pharm Org Process Res",
  ["journal of molecular psychiatry"] = "J Mol Psychiatry",
  ["journal of molecular recognition"] = "J. Mol. Recognit.",
  ["journal of molecular recognition : jmr"] = "J Mol Recognit",
  ["journal of molecular signaling"] = "J Mol Signal",
  ["journal of molecular spectroscopy"] = "J. Mol. Spectrosc.",
  ["journal of molecular structure"] = "J. Mol. Struct.",
  ["journal of molecular structure: theochem"] = "J. Mol. Struct. THEOCHEM",
  ["journal of molluscan studies"] = "J. Molluscan Stud.",
  ["journal of monetary economics"] = "J Monet Econ",
  ["journal of money, credit, and banking"] = "J Money Credit Bank",
  ["journal of moral education"] = "J Moral Educ",
  ["journal of moral philosophy"] = "J Moral Philos",
  ["journal of mormon history"] = "J Mormon Hist",
  ["journal of morphology"] = "J Morphol",
  ["journal of mosquito research"] = "J Mosq Res",
  ["journal of mother and child"] = "J Mother Child",
  ["journal of motor behavior"] = "J Mot Behav",
  ["journal of motor learning and development"] = "J Mot Learn Dev",
  ["journal of mountain agriculture on the balkans"] = "J Mt Agric Balk",
  ["journal of mountain science"] = "J Mt Sci",
  ["journal of movement disorders"] = "J Mov Disord",
  ["journal of mucosal immunology research"] = "J Mucosal Immunol Res",
  ["journal of multi-criteria decision analysis"] = "J. Multi-Criter. Decis. Anal.",
  ["journal of multicultural counseling and development"] = "J Multicult Couns Devel",
  ["journal of multicultural social work"] = "J Multicult Soc Work",
  ["journal of multidisciplinary evaluation"] = "J Multidiscip Eval",
  ["journal of multidisciplinary healthcare"] = "J Multidiscip Healthc",
  ["journal of multilingual and multicultural development"] = "J Multiling Multicult Dev",
  ["journal of multilingual communication disorders"] = "J Multiling Commun Disord",
  ["journal of multimedia"] = "J Multimed",
  ["journal of multimorbidity and comorbidity"] = "J. Multimorb. Comorb.",
  ["journal of multinational financial management"] = "J. Multinat. Finan. Manage.",
  ["journal of multiple sclerosis"] = "J Mult Scler (Foster City)",
  ["journal of multiple-valued logic and soft computing"] = "J. Mult.-Valued Logic Soft Comput.",
  ["journal of multiscale modeling"] = "J. Multiscale Model.",
  ["journal of multiscale modelling"] = "J. Multiscale Modell.",
  ["journal of multivariate analysis"] = "J. Multivariate Anal.",
  ["journal of muscle research and cell motility"] = "J Muscle Res Cell Motil",
  ["journal of musculoskeletal & neuronal interactions"] = "J Musculoskelet Neuronal Interact",
  ["journal of musculoskeletal disorders and treatment"] = "J Musculoskelet Disord Treat",
  ["journal of musculoskeletal pain"] = "J Musculoskelet Pain",
  ["journal of musculoskeletal research"] = "J Musculoskelet Res",
  ["journal of museum education : roundtable reports"] = "J Mus Educ",
  ["journal of music therapy"] = "J Music Ther",
  ["journal of muslim mental health"] = "J Muslim Ment Health",
  ["journal of muslim minority affairs"] = "J Muslim Minor Aff",
  ["journal of mycology"] = "J Mycol",
  ["journal of mycology and infection"] = "J Mycol Infect",
  ["journal of mycology and plant pathology"] = "J Mycol Plant Pathol",
  ["journal of mycopathological research"] = "J Mycopathol Res",
  ["journal of nal associates"] = "J Nal Assoc",
  ["journal of nanbiotechnology"] = "J. Nanobiotechnol.",
  ["journal of nanjing normal university"] = "J. Nanjing Norm. Univ. Nat. Sci. Ed.",
  ["journal of nanjing normal university. natural science edition. nanjing shida xuebao. ziran kexue ban"] = "J. Nanjing Norm. Univ. Nat. Sci. Ed.",
  ["journal of nanjing university. natural sciences"] = "J. Nanjing Univ. Nat. Sci.",
  ["journal of nano and electronic physics"] = "J. Nano Electron. Phys.",
  ["journal of nano education"] = "J. Nano Educ.",
  ["journal of nano education (print)"] = "J Nano Educ",
  ["journal of nano energy and power research"] = "J. Nano Energy Power Res.",
  ["journal of nano research"] = "J. Nano Res.",
  ["journal of nanobiotechnology"] = "J Nanobiotechnology",
  ["journal of nanoelectronics and optoelectronics"] = "J. Nanoelectron. Optoelectron.",
  ["journal of nanoengineering and nanomanufacturing"] = "J. Nanoeng. Nanomanuf.",
  ["journal of nanoeuroscience"] = "J. Nanoeurosci.",
  ["journal of nanomaterials"] = "J Nanomater",
  ["journal of nanomechanics & micromechanics"] = "J Nanomech Micromech",
  ["journal of nanomechanics and micromechanics"] = "J. Nanomech. Micromech.",
  ["journal of nanomedicine & biotherapeutic discovery"] = "J Nanomedine Biotherapeutic Discov",
  ["journal of nanomedicine & nanotechnology"] = "J Nanomed Nanotechnol",
  ["journal of nanomedicine research"] = "J Nanomed Res",
  ["journal of nanoneuroscience"] = "J Nanoneurosci",
  ["journal of nanoparticle research"] = "J. Nanopart. Res.",
  ["journal of nanoparticle research : an interdisciplinary forum for nanoscale science and technology"] = "J Nanopart Res",
  ["journal of nanoparticles"] = "J. Nanopart.",
  ["journal of nanopharmaceutics and drug delivery"] = "J Nanopharm Drug Deliv",
  ["journal of nanophotonics"] = "J. Nanophotonics",
  ["journal of nanoscience and nanomedicine"] = "J Nanosci Nanomed",
  ["journal of nanoscience and nanotechnology"] = "J Nanosci Nanotechnol",
  ["journal of nanoscience letters"] = "J Nanosci Lett",
  ["journal of nanostructure in chemistry"] = "J Nanostructure Chem",
  ["journal of nanostructures"] = "J. Nanostruct.",
  ["journal of nanotechnology"] = "J Nanotechnol",
  ["journal of nanotechnology and materials science"] = "J Nanotechnol Mater Sci",
  ["journal of nanotechnology in engineering and medicine"] = "J Nanotechnol Eng Med",
  ["journal of nanotoxicology and nanomedicine"] = "J Nanotoxicol Nanomed",
  ["journal of natal and zulu history"] = "J Natal Zulu Hist",
  ["journal of national black nurses association"] = "J. Natl. Black Nurses Assoc.",
  ["journal of national black nurses' association : jnbna"] = "J Natl Black Nurses Assoc",
  ["journal of national black nurses’ association : jnbna"] = "J Natl Black Nurses Assoc",
  ["journal of natural & physical sciences"] = "J. Natur. Phys. Sci.",
  ["journal of natural fibers"] = "J. Nat. Fibers",
  ["journal of natural gas chemistry"] = "J. Nat. Gas Chem.",
  ["journal of natural gas geoscience"] = "J. Nat. Gas Geosci.",
  ["journal of natural gas science and engineering"] = "J. Nat. Gas Sci. Eng.",
  ["journal of natural geometry"] = "J. Natur. Geom.",
  ["journal of natural history"] = "J. Nat. Hist.",
  ["journal of natural history museum"] = "J Nat Hist Mus",
  ["journal of natural medicines"] = "J Nat Med",
  ["journal of natural pesticide research"] = "J. Nat. Pestic. Res.",
  ["journal of natural products"] = "J Nat Prod",
  ["journal of natural resource policy research"] = "J. Nat. Resour. Policy Res.",
  ["journal of natural resources and life sciences education"] = "J. Nat. Resour. Life Sci. Educ.",
  ["journal of natural resources policy research"] = "J Nat Resour Policy Res",
  ["journal of natural science of hunan normal university. hunan shifan daxue ziran kexue xuebao"] = "J. Nat. Sci. Hunan Norm. Univ.",
  ["journal of natural science, biology, and medicine"] = "J Nat Sci Biol Med",
  ["journal of natural science. nanjing normal university"] = "J. Nat. Sci. Nanjing Norm. Univ.",
  ["journal of natural toxins"] = "J. Nat. Toxins",
  ["journal of nature and science"] = "J Nat Sci",
  ["journal of nature conservation"] = "J. Nat. Conserv.",
  ["journal of nature science and sustainable technology"] = "J. Nat. Sci. Sustainable Technol.",
  ["journal of navigation"] = "J. Navig.",
  ["journal of near eastern studies"] = "J Near East Stud",
  ["journal of near infrared spectroscopy"] = "J. Near Infrared Spectrosc.",
  ["journal of negative results : ecology & evolutionary biology"] = "J Negat Results",
  ["journal of negative results in biomedicine"] = "J Negat Results Biomed",
  ["journal of nematode morphology and systematics"] = "J Nematode Morphol System",
  ["journal of nematology"] = "J Nematol",
  ["journal of neonatal biology"] = "J Neonatal Biol",
  ["journal of neonatal nursing : jnn"] = "J Neonatal Nurs",
  ["journal of neonatal surgery"] = "J Neonatal Surg",
  ["journal of neonatal-perinatal medicine"] = "J Neonatal Perinatal Med",
  ["journal of nepal agricultural research council"] = "J Nepal Agric Res Counc",
  ["journal of nepal chemical society"] = "J. Nepal Chem. Soc.",
  ["journal of nepal health research council"] = "J Nepal Health Res Counc",
  ["journal of nephrology"] = "J Nephrol",
  ["journal of nephrology & therapeutics"] = "J Nephrol Ther",
  ["journal of nephrology and renal transplantation"] = "J Nephrol Ren Transplant",
  ["journal of nephrology nursing"] = "J Nephrol Nurs",
  ["journal of nephrology research"] = "J Nephrol Res",
  ["journal of nephropathology"] = "J Nephropathol",
  ["journal of nephropharmacology"] = "J Nephropharmacol",
  ["journal of nervous and mental disease"] = "J. Nerv. Ment. Dis.",
  ["journal of network and computer applications"] = "J. Network Comput. Appl.",
  ["journal of network and systems management"] = "J. Network Syst. Manage.",
  ["journal of neural engineering"] = "J Neural Eng",
  ["journal of neural transmission"] = "J Neural Transm",
  ["journal of neural transmission (vienna, austria : 1996)"] = "J Neural Transm (Vienna)",
  ["journal of neural transmission. general section"] = "J Neural Transm Gen Sect",
  ["journal of neural transmission. parkinson's disease and dementia section"] = "J Neural Transm Park Dis Dement Sect",
  ["journal of neural transmission. parkinsons disease and dementia section"] = "J. Neural Transm. Park. Dis. Dement. Sect.",
  ["journal of neural transmission. parkinson’s disease and dementia section"] = "J Neural Transm Park Dis Dement Sect",
  ["journal of neural transmission. supplementum"] = "J Neural Transm Suppl",
  ["journal of neural transmission: general section"] = "J. Neural Transm.: Gen. Sect.",
  ["journal of neural transplantation"] = "J Neural Transplant",
  ["journal of neural transplantation & plasticity"] = "J Neural Transplant Plast",
  ["journal of neural transplantation and plasticity"] = "J. Neural Transplant. Plast.",
  ["journal of neuro-aids"] = "J NeuroAIDS",
  ["journal of neuro-oncology"] = "J Neurooncol",
  ["journal of neuro-ophthalmology"] = "J. Neuroophthalmol.",
  ["journal of neuro-ophthalmology : the official journal of the north american neuro-ophthalmology society"] = "J Neuroophthalmol",
  ["journal of neuro-visceral relations"] = "J Neurovisc Relat",
  ["journal of neuroanaesthesiology and critical care"] = "J Neuroanaesth Crit Care",
  ["journal of neurobiology"] = "J Neurobiol",
  ["journal of neurobiology and physiology"] = "J Neurobiol Physiol",
  ["journal of neurochemistry"] = "J Neurochem",
  ["journal of neurocytology"] = "J Neurocytol",
  ["journal of neurodegeneration & regeneration"] = "J Neurodegener Regen",
  ["journal of neurodegenerative diseases"] = "J Neurodegener Dis",
  ["journal of neurodevelopmental disorders"] = "J Neurodev Disord",
  ["journal of neuroendocrinology"] = "J Neuroendocrinol",
  ["journal of neuroengineering and rehabilitation"] = "J Neuroeng Rehabil",
  ["journal of neurogastroenterology and motility"] = "J Neurogastroenterol Motil",
  ["journal of neurogenetics"] = "J Neurogenet",
  ["journal of neuroimaging"] = "J. Neuroimaging",
  ["journal of neuroimaging : official journal of the american society of neuroimaging"] = "J Neuroimaging",
  ["journal of neuroimaging in psychiatry & neurology"] = "J Neuroimaging Psychiatry Neurol",
  ["journal of neuroimmune pharmacology"] = "J. Neuroimmune Pharmacol.",
  ["journal of neuroimmune pharmacology : the official journal of the society on neuroimmune pharmacology"] = "J Neuroimmune Pharmacol",
  ["journal of neuroimmunology"] = "J Neuroimmunol",
  ["journal of neuroimmunology. supplement"] = "J Neuroimmunol Suppl",
  ["journal of neuroinfectious diseases"] = "J Neuroinfect Dis",
  ["journal of neuroinflammation"] = "J Neuroinflammation",
  ["journal of neuroinflammation and neurodegenerative diseases"] = "J Neuroinflamm Neurodegener Dis",
  ["journal of neurointerventional surgery"] = "J Neurointerv Surg",
  ["journal of neurolinguistics"] = "J Neurolinguistics",
  ["journal of neurologic physical therapy"] = "J. Neurol. Phys. Ther.",
  ["journal of neurologic physical therapy : jnpt"] = "J Neurol Phys Ther",
  ["journal of neurologic rehabilitation"] = "J Neurol Rehabil",
  ["journal of neurological disorders"] = "J Neurol Disord",
  ["journal of neurological disorders & stroke"] = "J Neurol Disord Stroke",
  ["journal of neurological research and therapy"] = "J Neurol Res Ther",
  ["journal of neurological sciences (turkish)"] = "J Neurol Sci Turk",
  ["journal of neurological sciences and psychiatry"] = "J Neurol Sci Psychiatry",
  ["journal of neurological surgery reports"] = "J Neurol Surg Rep",
  ["journal of neurological surgery. part a, central european neurosurgery"] = "J Neurol Surg A Cent Eur Neurosurg",
  ["journal of neurological surgery. part b, skull base"] = "J Neurol Surg B Skull Base",
  ["journal of neurology"] = "J Neurol",
  ["journal of neurology & neuromedicine"] = "J Neurol Neuromedicine",
  ["journal of neurology & neurophysiology"] = "J Neurol Neurophysiol",
  ["journal of neurology & translational neuroscience"] = "J Neurol Transl Neurosci",
  ["journal of neurology and experimental neuroscience"] = "J Neurol Exp Neurosci",
  ["journal of neurology and neurobiology"] = "J Neurol Neurobiol",
  ["journal of neurology and neurological disorders"] = "J Neurol Neurol Disord",
  ["journal of neurology and neurorehabilitation research"] = "J Neurol Neurorehabilit Res",
  ["journal of neurology and neuroscience"] = "J Neurol Neurosci",
  ["journal of neurology and neurosurgery"] = "J Neurol Neurosurg",
  ["journal of neurology and psychiatry"] = "J Neurol Psychiatry",
  ["journal of neurology and psychology"] = "J Neurol Psychol",
  ["journal of neurology research"] = "J Neurol Res",
  ["journal of neurology, neurological science and disorders"] = "J Neurol Neurol Sci Disord",
  ["journal of neurology, neurosurgery & psychiatry research"] = "J Neurol Neurosurg Psychiatry Res",
  ["journal of neurology, neurosurgery and psychiatry"] = "J. Neurol. Neurosurg. Psychiatry",
  ["journal of neurology, neurosurgery and spine"] = "J Neurol Neurosurg Spine",
  ["journal of neurology, neurosurgery, and psychiatry"] = "J Neurol Neurosurg Psychiatry",
  ["journal of neuromuscular diseases"] = "J Neuromuscul Dis",
  ["journal of neuroparasitology"] = "J Neuroparasitology",
  ["journal of neuropathic pain & symptom palliation"] = "J Neuropathic Pain Symptom Palliation",
  ["journal of neuropathology and clinical neurology"] = "J Neuropathol Clin Neurol",
  ["journal of neuropathology and experimental neurology"] = "J Neuropathol Exp Neurol",
  ["journal of neurophysiology"] = "J Neurophysiol",
  ["journal of neurophysiology and neurological disorders"] = "J Neurophysiol Neurol Disord",
  ["journal of neuropsychiatry"] = "J Neuropsychiatr",
  ["journal of neuropsychiatry (foster city, calif.)"] = "J Neuropsychiatry (Foster City)",
  ["journal of neuropsychiatry and clinical neurosciences"] = "J. Neuropsychiatry Clin. Neurosci.",
  ["journal of neuropsychology"] = "J Neuropsychol",
  ["journal of neuroradiology"] = "J. Neuroradiol.",
  ["journal of neuroradiology = journal de neuroradiologie"] = "J Neuroradiol",
  ["journal of neuroradiology. journal de neuroradiologie"] = "J Neuroradiol",
  ["journal of neuroscience"] = "J. Neurosci.",
  ["journal of neuroscience and neuroengineering"] = "J Neurosci Neuroeng",
  ["journal of neuroscience and neuropharmacology"] = "J Neurosci Neuropharmacol",
  ["journal of neuroscience and rehabilitation"] = "J Neurosci Rehabil",
  ["journal of neuroscience methods"] = "J Neurosci Methods",
  ["journal of neuroscience nursing"] = "J. Neurosci. Nurs.",
  ["journal of neuroscience research"] = "J Neurosci Res",
  ["journal of neuroscience, psychology, and economics"] = "J Neurosci Psychol Econ",
  ["journal of neurosciences in rural practice"] = "J Neurosci Rural Pract",
  ["journal of neurosurgery"] = "J Neurosurg",
  ["journal of neurosurgery. pediatrics"] = "J Neurosurg Pediatr",
  ["journal of neurosurgery. spine"] = "J Neurosurg Spine",
  ["journal of neurosurgical anesthesiology"] = "J Neurosurg Anesthesiol",
  ["journal of neurosurgical nursing"] = "J Neurosurg Nurs",
  ["journal of neurosurgical sciences"] = "J Neurosurg Sci",
  ["journal of neurotherapy"] = "J Neurother",
  ["journal of neurotrauma"] = "J Neurotrauma",
  ["journal of neurovirology"] = "J Neurovirol",
  ["journal of neutron research"] = "J. Neutron Res.",
  ["journal of new drugs"] = "J. New Drugs",
  ["journal of new england water environment association"] = "J. N. Engl. Water Environ. Assoc.",
  ["journal of new england water works association"] = "J. N. Engl. Water Works Assoc.",
  ["journal of new materials for electrochemical systems"] = "J. New Mater. Electrochem. Syst.",
  ["journal of new music research"] = "J New Music Res",
  ["journal of new zealand society of periodontology"] = "J. N. Z. Soc. Periodontol.",
  ["journal of nihon university school of dentistry"] = "J. Nihon Univ. Sch. Dent.",
  ["journal of ningxia university"] = "J. Ningxia Univ. Nat. Sci. Ed.",
  ["journal of nippon medical school"] = "J. Nippon Med. Sch.",
  ["journal of nippon medical school = nippon ika daigaku zasshi"] = "J Nippon Med Sch",
  ["journal of non-crystalline solids"] = "J. Non-Cryst. Solids",
  ["journal of non-crystalline solids x"] = "J. Non-Cryst. Solids X",
  ["journal of non-equilibrium thermodynamics"] = "J. Non-Equilib. Thermodyn.",
  ["journal of non-newtonian fluid mechanics"] = "J. Non-Newtonian Fluid Mech.",
  ["journal of noncommutative geometry"] = "J. Noncommut. Geom.",
  ["journal of nondestructive evaluation"] = "J. Nondestr. Eval.",
  ["journal of nondestructive evaluation, diagnostics & prognostics of engineering systems"] = "J. Nondestr. Eval. Diagn. Progn. Eng. Syst.",
  ["journal of nondestructive evaluation, diagnostics and prognostics of engineering systems"] = "J Nondestruct Eval Diagn Progn Eng Syst",
  ["journal of nonlinear analysis and optimization. theory and applications"] = "J. Nonlinear Anal. Optim.",
  ["journal of nonlinear and convex analysis"] = "J Nonlinear Convex Anal",
  ["journal of nonlinear and convex analysis. an international journal"] = "J. Nonlinear Convex Anal.",
  ["journal of nonlinear evolution equations and applications"] = "J. Nonlinear Evol. Equ. Appl.",
  ["journal of nonlinear mathematical physics"] = "J. Nonlinear Math. Phys.",
  ["journal of nonlinear optical physics and materials"] = "J. Nonlinear Opt. Phys. Mater.",
  ["journal of nonlinear science"] = "J. Nonlinear Sci.",
  ["journal of nonlinear sciences and applications. jnsa"] = "J. Nonlinear Sci. Appl.",
  ["journal of nonlinear systems and applications"] = "J. Nonlinear Syst. Appl.",
  ["journal of nonlinear, complex and data science"] = "J. Nonlinear Complex Data Sci.",
  ["journal of nonparametric statistics"] = "J Nonparametr Stat",
  ["journal of nonsmooth analysis and optimization"] = "J. Nonsmooth Anal. Optim.",
  ["journal of nonverbal behavior"] = "J Nonverbal Behav",
  ["journal of northeastern university"] = "J. Northeast. Univ. Nat. Sci.",
  ["journal of northeastern university. natural science. dongbei daxue xuebao. ziran kexue ban"] = "J. Northeast. Univ. Nat. Sci.",
  ["journal of northwest atlantic fishery science"] = "J. Northwest Atl. Fish. Sci.",
  ["journal of northwest university"] = "J. Northwest Univ.",
  ["journal of northwest university. natural science edition. xibei daxue xuebao. ziran kexue ban"] = "J. Northwest Univ. Nat. Sci.",
  ["journal of novel physiotherapies"] = "J Nov Physiother",
  ["journal of novel physiotherapy and physical rehabilitation"] = "J Nov Physiother Phys Rehabil",
  ["journal of nuclear and radiochemical sciences"] = "J. Nucl. Radiochem. Sci.",
  ["journal of nuclear biology and medicine"] = "J. Nucl. Biol. Med.",
  ["journal of nuclear biology and medicine (turin, italy : 1991)"] = "J Nucl Biol Med",
  ["journal of nuclear cardiology"] = "J. Nucl. Cardiol.",
  ["journal of nuclear cardiology : official publication of the american society of nuclear cardiology"] = "J Nucl Cardiol",
  ["journal of nuclear energy"] = "J. Nucl. Energy",
  ["journal of nuclear energy, part a: reactor science"] = "J. Nucl. Energy A",
  ["journal of nuclear energy, part b: reactor technology"] = "J. Nucl. Energy B",
  ["journal of nuclear energy, part c: plasma physics, accelerators, thermonuclear research"] = "J. Nucl. Energy C",
  ["journal of nuclear energy. part c, plasma physics, accelerators, thermonuclear research"] = "J. Nucl. Energy C",
  ["journal of nuclear engineering and radiation science"] = "J. Nucl. Eng. Radiat. Sci.",
  ["journal of nuclear materials"] = "J. Nucl. Mater.",
  ["journal of nuclear materials. journal des matériaux nucléaires"] = "J Nucl Mater",
  ["journal of nuclear medicine"] = "J. Nucl. Med.",
  ["journal of nuclear medicine & radiation therapy"] = "J Nucl Med Radiat Ther",
  ["journal of nuclear medicine : official publication, society of nuclear medicine"] = "J Nucl Med",
  ["journal of nuclear medicine and allied sciences"] = "J. Nucl. Med. Allied Sci.",
  ["journal of nuclear medicine technology"] = "J Nucl Med Technol",
  ["journal of nuclear science and technology"] = "J. Nucl. Sci. Technol.",
  ["journal of nuclear science and technology (abingdon, united kingdom)"] = "J. Nucl. Sci. Technol. (Abingdon, U. K.)",
  ["journal of nuclear science and technology (tokyo, japan)"] = "J. Nucl. Sci. Technol. (Tokyo, Jpn.)",
  ["journal of nucleic acids"] = "J Nucleic Acids",
  ["journal of nucleic acids investigation"] = "J Nucleic Acids Investig",
  ["journal of number theory"] = "J. Number Theory",
  ["journal of numerical analysis and approximation theory"] = "J. Numer. Anal. Approx. Theory",
  ["journal of numerical cognition"] = "J Numer Cogn",
  ["journal of numerical mathematics"] = "J. Numer. Math.",
  ["journal of numerical mathematics and stochastics"] = "J. Numer. Math. Stoch.",
  ["journal of nurse-midwifery"] = "J Nurse Midwifery",
  ["journal of nursing (luton, england)"] = "J Nurs (Luton)",
  ["journal of nursing administration"] = "J. Nurs. Adm.",
  ["journal of nursing and health"] = "J Nurs Health",
  ["journal of nursing and healthcare of chronic illness"] = "J Nurs Healthc Chronic Illn",
  ["journal of nursing care"] = "J. Nurs. Care",
  ["journal of nursing care quality"] = "J Nurs Care Qual",
  ["journal of nursing doctoral students scholarship"] = "J Nurs Dr Stud Scholarsh",
  ["journal of nursing education"] = "J. Nurs. Educ.",
  ["journal of nursing education and practice"] = "J Nurs Educ Pract",
  ["journal of nursing ethics"] = "J Nurs Ethics",
  ["journal of nursing history"] = "J. Nurs. Hist.",
  ["journal of nursing history : a publication of the nursing archives associates at boston university"] = "J Nurs Hist",
  ["journal of nursing law"] = "J Nurs Law",
  ["journal of nursing management"] = "J Nurs Manag",
  ["journal of nursing measurement"] = "J Nurs Meas",
  ["journal of nursing quality assurance"] = "J Nurs Qual Assur",
  ["journal of nursing regulation"] = "J Nurs Regul",
  ["journal of nursing research"] = "J. Nurs. Res.",
  ["journal of nursing scholarship"] = "J. Nurs. Scholarsh.",
  ["journal of nursing scholarship : an official publication of sigma theta tau international honor society of nursing"] = "J Nurs Scholarsh",
  ["journal of nursing scholarship : an official publication of sigma theta tau international honor society of nursing / sigma theta tau"] = "J Nurs Scholarsh",
  ["journal of nursing staff development"] = "J. Nurs. Staff Dev.",
  ["journal of nursing staff development : jnsd"] = "J Nurs Staff Dev",
  ["journal of nutrigenetics and nutrigenomics"] = "J Nutrigenet Nutrigenomics",
  ["journal of nutrition"] = "J. Nutr.",
  ["journal of nutrition & food sciences"] = "J Nutr Food Sci",
  ["journal of nutrition & health (northborough)"] = "J Nutr Health",
  ["journal of nutrition & intermediary metabolism"] = "J Nutr Intermed Metab",
  ["journal of nutrition and dietetics"] = "J Nutr Diet",
  ["journal of nutrition and metabolism"] = "J Nutr Metab",
  ["journal of nutrition education"] = "J Nutr Educ",
  ["journal of nutrition education and behavior"] = "J Nutr Educ Behav",
  ["journal of nutrition for the elderly"] = "J Nutr Elder",
  ["journal of nutrition in gerontology and geriatrics"] = "J Nutr Gerontol Geriatr",
  ["journal of nutrition, health & aging"] = "J. Nutr., Health Aging",
  ["journal of nutrition, health and aging"] = "J. Nutr. Health Aging",
  ["journal of nutritional & environmental medicine"] = "J. Nutr. Environ. Med.",
  ["journal of nutritional biochemistry"] = "J. Nutr. Biochem.",
  ["journal of nutritional biology"] = "J Nutr Biol",
  ["journal of nutritional disorders & therapy"] = "J Nutr Disord Ther",
  ["journal of nutritional ecology and food research"] = "J Nutr Ecol Food Res",
  ["journal of nutritional health & food science"] = "J Nutrit Health Food Sci",
  ["journal of nutritional immunology"] = "J. Nutr. Immunol.",
  ["journal of nutritional medicine and diet care"] = "J Nutr Med Diet Care",
  ["journal of nutritional oncology"] = "J. Nutr. Oncol.",
  ["journal of nutritional physiology"] = "J. Nutr. Physiol.",
  ["journal of nutritional science"] = "J Nutr Sci",
  ["journal of nutritional science and vitaminology"] = "J Nutr Sci Vitaminol (Tokyo)",
  ["journal of obesity"] = "J Obes",
  ["journal of obesity & metabolic syndrome"] = "J Obes Metab Syndr",
  ["journal of obesity & weight loss therapy"] = "J Obes Weight Loss Ther",
  ["journal of obesity and overweight"] = "J Obes Overweight",
  ["journal of obesity and weight-loss medication"] = "J Obes Weight Loss Medicat",
  ["journal of obsessive-compulsive and related disorders"] = "J Obsessive Compuls Relat Disord",
  ["journal of obstetric, gynecologic, and neonatal nursing"] = "J. Obstet. Gynecol. Neonatal Nurs.",
  ["journal of obstetric, gynecologic, and neonatal nursing : jognn"] = "J Obstet Gynecol Neonatal Nurs",
  ["journal of obstetric, gynecologic, and neonatal nursing : jognn / naacog"] = "J Obstet Gynecol Neonatal Nurs",
  ["journal of obstetrics & gynaecology of eastern and central africa"] = "J Obstet Gynaecol East Cent Africa",
  ["journal of obstetrics and gynaecology"] = "J Obstet Gynaecol (Lahore)",
  ["journal of obstetrics and gynaecology (tokyo, japan)"] = "J Obstet Gynaecol (Tokyo 1995)",
  ["journal of obstetrics and gynaecology : the journal of the institute of obstetrics and gynaecology"] = "J Obstet Gynaecol",
  ["journal of obstetrics and gynaecology canada"] = "J. Obstet. Gynaecol. Can.",
  ["journal of obstetrics and gynaecology canada : jogc = journal d'obstétrique et gynécologie du canada : jogc"] = "J Obstet Gynaecol Can",
  ["journal of obstetrics and gynaecology of india"] = "J Obstet Gynaecol India",
  ["journal of obstetrics and gynaecology of the british commonwealth"] = "J. Obstet. Gynaecol. Br. Commonw.",
  ["journal of obstetrics and gynaecology research"] = "J. Obstet. Gynaecol. Res.",
  ["journal of occupational accidents"] = "J. Occup. Accid.",
  ["journal of occupational and environmental hygiene"] = "J Occup Environ Hyg",
  ["journal of occupational and environmental medicine"] = "J Occup Environ Med",
  ["journal of occupational and environmental medicine / american college of occupational and environmental medicine"] = "J Occup Environ Med",
  ["journal of occupational and organizational psychology"] = "J Occup Organ Psychol",
  ["journal of occupational health"] = "J Occup Health",
  ["journal of occupational health psychology"] = "J Occup Health Psychol",
  ["journal of occupational medicine"] = "J. Occup. Med.",
  ["journal of occupational medicine and toxicology"] = "J. Occup. Med. Toxicol.",
  ["journal of occupational medicine and toxicology (london, england)"] = "J Occup Med Toxicol",
  ["journal of occupational medicine. : official publication of the industrial medical association"] = "J Occup Med",
  ["journal of occupational rehabilitation"] = "J Occup Rehabil",
  ["journal of occupational science"] = "J Occup Sci",
  ["journal of occupational therapy, schools & early intervention"] = "J Occup Ther Sch Early Interv",
  ["journal of ocean and climate"] = "J. Ocean Clim.",
  ["journal of ocean and coastal economics"] = "J Ocean Coast Econ",
  ["journal of ocean engineering and marine energy"] = "J Ocean Eng Mar Energy",
  ["journal of ocean engineering and science"] = "J. Ocean. Eng. Sci.",
  ["journal of ocean technology"] = "J. Ocean Technol.",
  ["journal of ocean university of china (english edition)"] = "J. Ocean Univ. China",
  ["journal of ocean university of china : jouc"] = "J Ocean Univ China",
  ["journal of oceanography"] = "J. Oceanogr.",
  ["journal of oceanology and limnology"] = "J Oceanol Limnol",
  ["journal of ocular biology"] = "J Ocul Biol",
  ["journal of ocular biology, diseases, and informatics"] = "J Ocul Biol Dis Infor",
  ["journal of ocular diseases and therapeutics"] = "J Ocul Dis Ther",
  ["journal of ocular pharmacology"] = "J Ocul Pharmacol",
  ["journal of ocular pharmacology and therapeutics"] = "J. Ocul. Pharmacol. Ther.",
  ["journal of ocular pharmacology and therapeutics : the official journal of the association for ocular pharmacology and therapeutics"] = "J Ocul Pharmacol Ther",
  ["journal of offender rehabilitation"] = "J Offender Rehabil",
  ["journal of official statistics"] = "J Off Stat",
  ["journal of offshore mechanics and arctic engineering"] = "J. Offshore Mech. Arct. Eng.",
  ["journal of oil palm research"] = "J Oil Palm Res",
  ["journal of oil, gas and petrochemical sciences"] = "J Oil Gas Petrochem Sci",
  ["journal of oilseeds research"] = "J Oilseeds Res",
  ["journal of oleo science"] = "J Oleo Sci",
  ["journal of oncobiomarkers"] = "J Oncobiomarkers",
  ["journal of oncology"] = "J Oncol",
  ["journal of oncology and cancer research"] = "J Oncol Cancer Res",
  ["journal of oncology management"] = "J. Oncol. Manag.",
  ["journal of oncology navigation & survivorship"] = "J Oncol Navig Surviv",
  ["journal of oncology pharmacy practice"] = "J. Oncol. Pharm. Pract.",
  ["journal of oncology pharmacy practice : official publication of the international society of oncology pharmacy practitioners"] = "J Oncol Pharm Pract",
  ["journal of oncology practice"] = "J Oncol Pract",
  ["journal of oncology research"] = "J Oncol Research",
  ["journal of oncology research and therapy"] = "J Oncol Res Ther",
  ["journal of oncology translational research"] = "J Oncol Transl Res",
  ["journal of oncopathology and clinical research"] = "J Oncopathol Clin Res",
  ["journal of online learning and teaching"] = "J Online Learn Teach",
  ["journal of open archaeology data"] = "J. Open Archaeol. Data",
  ["journal of open hardware"] = "J. Open Hardware",
  ["journal of open research software"] = "J Open Res Softw",
  ["journal of open source software"] = "J Open Source Softw",
  ["journal of operational meteorology"] = "J. Oper. Meteorol.",
  ["journal of operational oceanography"] = "J. Oper. Oceanogr.",
  ["journal of operations and strategic planning"] = "J. Oper. Strategic Plann.",
  ["journal of operations management"] = "J. Oper. Manage.",
  ["journal of operator theory"] = "J. Operator Theory",
  ["journal of ophthalmic & vision research"] = "J Ophthalmic Vis Res",
  ["journal of ophthalmic inflammation and infection"] = "J Ophthalmic Inflamm Infect",
  ["journal of ophthalmic nursing & technology"] = "J Ophthalmic Nurs Technol",
  ["journal of ophthalmic nursing and technology"] = "J. Ophthalmic Nurs. Technol.",
  ["journal of ophthalmology"] = "J Ophthalmol",
  ["journal of opioid management"] = "J Opioid Manag",
  ["journal of optical communication"] = "J. Opt. Commun.",
  ["journal of optical communications"] = "J. Opt. Commun.",
  ["journal of optical communications and networking"] = "J. Opt. Commun. Networking",
  ["journal of optical microsystems"] = "J. Opt. Microsyst.",
  ["journal of optics"] = "J. Opt.",
  ["journal of optics (2010)"] = "J Opt",
  ["journal of optics a: pure and applied optics"] = "J. Opt. A: Pure Appl. Opt.",
  ["journal of optics b: quantum and semiclassical optics"] = "J. Opt. B: Quantum Semiclassical Opt.",
  ["journal of optics. b, quantum and semiclassical optics : journal of the european optical society"] = "J Opt B Quantum Semiclassical Opt",
  ["journal of optimization"] = "J. Optim.",
  ["journal of optimization theory"] = "J. Optim. Theory",
  ["journal of optimization theory and applications"] = "J. Optim. Theory Appl.",
  ["journal of optimization, differential equations and their applications"] = "J. Optim. Differ. Equ. Appl.",
  ["journal of optoelectronics and advanced materials"] = "J. Optoelectron. Adv. Mater.",
  ["journal of optometry"] = "J Optom",
  ["journal of oral & facial pain and headache"] = "J Oral Facial Pain  Headache",
  ["journal of oral & maxillofacial research"] = "J Oral Maxillofac Res",
  ["journal of oral and maxillofacial pathology : jomfp"] = "J Oral Maxillofac Pathol",
  ["journal of oral and maxillofacial surgery"] = "J. Oral Maxillofac. Surg.",
  ["journal of oral and maxillofacial surgery : official journal of the american association of oral and maxillofacial surgeons"] = "J Oral Maxillofac Surg",
  ["journal of oral and maxillofacial surgery, medicine, and pathology"] = "J Oral Maxillofac Surg Med Pathol",
  ["journal of oral biology (northborough, mass.)"] = "J Oral Biol (Northborough)",
  ["journal of oral biology and craniofacial research"] = "J Oral Biol Craniofac Res",
  ["journal of oral biosciences"] = "J Oral Biosci",
  ["journal of oral diseases"] = "J Oral Dis",
  ["journal of oral hygiene & health"] = "J Oral Hyg Health",
  ["journal of oral implant and transplant surgery"] = "J. Oral Implant Transplant Surg.",
  ["journal of oral implantology"] = "J. Oral Implantol.",
  ["journal of oral medicine"] = "J Oral Med",
  ["journal of oral medicine and pain"] = "J Oral Med Pain",
  ["journal of oral microbiology"] = "J Oral Microbiol",
  ["journal of oral pathology"] = "J Oral Pathol",
  ["journal of oral pathology & medicine : official publication of the international association of oral pathologists and the american academy of oral pathology"] = "J Oral Pathol Med",
  ["journal of oral pathology and medicine"] = "J. Oral Pathol. Med.",
  ["journal of oral rehabilitation"] = "J Oral Rehabil",
  ["journal of oral science"] = "J Oral Sci",
  ["journal of oral surgery"] = "J Oral Surg (Chic)",
  ["journal of oral surgery (american dental association : 1965)"] = "J Oral Surg",
  ["journal of oral surgery, anesthesia, and hospital dental service"] = "J Oral Surg Anesth Hosp Dent Serv",
  ["journal of oral therapeutics and pharmacology"] = "J Oral Ther Pharmacol",
  ["journal of organ dysfunction"] = "J Organ Dysfunct",
  ["journal of organic and biomolecular simulations"] = "J Org Biomol Simul",
  ["journal of organic chemistry"] = "J. Org. Chem.",
  ["journal of organic semiconductors"] = "J. Org. Semicond.",
  ["journal of organizational and end user computing : an official publication of the information resources management association"] = "J Organ End User Comput",
  ["journal of organizational behavior"] = "J Organ Behav",
  ["journal of organizational behavior management"] = "J Organ Behav Manage",
  ["journal of organizational psychology"] = "J Organ Psychol",
  ["journal of organometallic chemistry"] = "J. Organomet. Chem.",
  ["journal of orissa mathematical society"] = "J. Orissa Math. Soc.",
  ["journal of ornithology"] = "J Ornithol",
  ["journal of orofacial orthopedics"] = "J. Orofac. Orthop.",
  ["journal of orofacial orthopedics = fortschritte der kieferorthopädie : organ/official journal deutsche gesellschaft für kieferorthopädie"] = "J Orofac Orthop",
  ["journal of orofacial pain"] = "J Orofac Pain",
  ["journal of orofacial sciences"] = "J Orofac Sci",
  ["journal of orthodontic science"] = "J Orthod Sci",
  ["journal of orthodontics"] = "J Orthod",
  ["journal of orthomolecular medicine : official journal of the academy of orthomolecular medicine"] = "J Orthomol Med",
  ["journal of orthopaedic & sports physical therapy"] = "J. Orthop. Sports Phys. Ther.",
  ["journal of orthopaedic and sports physical therapy"] = "J. Orthop. Sports Phys. Ther.",
  ["journal of orthopaedic case reports"] = "J Orthop Case Rep",
  ["journal of orthopaedic research"] = "J. Orthop. Res.",
  ["journal of orthopaedic research : official publication of the orthopaedic research society"] = "J Orthop Res",
  ["journal of orthopaedic science"] = "J. Orthop. Sci.",
  ["journal of orthopaedic science : official journal of the japanese orthopaedic association"] = "J Orthop Sci",
  ["journal of orthopaedic surgery (hong kong)"] = "J Orthop Surg (Hong Kong)",
  ["journal of orthopaedic surgery and research"] = "J Orthop Surg Res",
  ["journal of orthopaedic translation"] = "J Orthop Translat",
  ["journal of orthopaedic trauma"] = "J Orthop Trauma",
  ["journal of orthopaedics"] = "J Orthop",
  ["journal of orthopaedics and traumatology : official journal of the italian society of orthopaedics and traumatology"] = "J Orthop Traumatol",
  ["journal of orthopedic research and therapy"] = "J Orthop Res Ther",
  ["journal of orthopedics & rheumatology"] = "J Orthop Rheumatol",
  ["journal of orthoptera research"] = "J Orthoptera Res",
  ["journal of ortoptera research"] = "J. Ort. Res.",
  ["journal of osaka dental university"] = "J Osaka Dent Univ",
  ["journal of osteoporosis"] = "J Osteoporos",
  ["journal of osteoporosis and physical activity"] = "J Osteoporos Phys Act",
  ["journal of otolaryngology"] = "J. Otolaryngol.",
  ["journal of otolaryngology - head & neck surgery = le journal d'oto-rhino-laryngologie et de chirurgie cervico-faciale"] = "J Otolaryngol Head Neck Surg",
  ["journal of otolaryngology. supplement"] = "J. Otolaryngol. Suppl.",
  ["journal of otology"] = "J Otol",
  ["journal of otology & rhinology"] = "J Otol Rhinol",
  ["journal of outcome measurement"] = "J Outcome Meas",
  ["journal of outdoor and environmental education"] = "J. Outdoor Environ. Educ.",
  ["journal of ovarian research"] = "J Ovarian Res",
  ["journal of ovonic research"] = "J. Ovonic Res.",
  ["journal of packaging technology and reseach"] = "J. Packag. Technol. Res.",
  ["journal of packaging technology and research"] = "J Packag Technol Res",
  ["journal of paediatric dentistry"] = "J Paediatr Dent",
  ["journal of paediatrics and child health"] = "J Paediatr Child Health",
  ["journal of pain"] = "J. Pain",
  ["journal of pain & palliative care pharmacotherapy"] = "J Pain Palliat Care Pharmacother",
  ["journal of pain & relief"] = "J Pain Relief",
  ["journal of pain and palliative care pharmacotherapy"] = "J. Pain Palliat. Care Pharmacother.",
  ["journal of pain and symptom management"] = "J Pain Symptom Manage",
  ["journal of pain management"] = "J Pain Manag",
  ["journal of pain management & medicine"] = "J Pain Manag Med",
  ["journal of pain research"] = "J Pain Res",
  ["journal of pakistan psychiatric society"] = "J Pak Psychiatr Soc",
  ["journal of palaeogeography"] = "J. Palaeogeogr.",
  ["journal of palaeolimnology"] = "J. Paleolimnol.",
  ["journal of paleolimnology"] = "J. Paleolimnol.",
  ["journal of paleolithic archaeology"] = "J. Paleolithic Archaeol.",
  ["journal of paleontology"] = "J. Paleontol.",
  ["journal of palestine studies"] = "J Palest Stud",
  ["journal of palliative care"] = "J Palliat Care",
  ["journal of palliative care & medicine"] = "J Palliat Care Med",
  ["journal of palliative medicine"] = "J Palliat Med",
  ["journal of palynology"] = "J. Palynol.",
  ["journal of pancreatic cancer"] = "J Pancreat Cancer",
  ["journal of pancreatology"] = "J Pancreatol",
  ["journal of paper conservation"] = "J. Pap. Conserv.",
  ["journal of parallel and distributed computing"] = "J. Parallel Distrib. Comput.",
  ["journal of paramedical sciences"] = "J Paramed Sci",
  ["journal of parasitic diseases : official organ of the indian society for parasitology"] = "J Parasit Dis",
  ["journal of parasitology"] = "J. Parasitol.",
  ["journal of parasitology (faisalabad)"] = "J Parasitol (Faisalabad)",
  ["journal of parasitology research"] = "J Parasitol Res",
  ["journal of parenteral science and technology"] = "J. Parenter. Sci. Technol.",
  ["journal of parenteral science and technology : a publication of the parenteral drug association"] = "J Parenter Sci Technol",
  ["journal of park and recreation administration"] = "J Park Recreat Admi",
  ["journal of parkinson's disease"] = "J Parkinsons Dis",
  ["journal of parkinson's disease and alzheimer's disease"] = "J Parkinsons Dis Alzheimers Dis",
  ["journal of partial differential equations"] = "J. Partial Differ. Equ.",
  ["journal of participatory medicine"] = "J Particip Med",
  ["journal of participatory research methods"] = "J Particip Res Methods",
  ["journal of pastoral care"] = "J Pastoral Care",
  ["journal of pathogens"] = "J Pathog",
  ["journal of pathology"] = "J. Pathol.",
  ["journal of pathology and bacteriology"] = "J. Pathol. Bacteriol.",
  ["journal of pathology and translational medicine"] = "J Pathol Transl Med",
  ["journal of pathology informatics"] = "J Pathol Inform",
  ["journal of pathology: clinical research"] = "J. Pathol.: Clin. Res.",
  ["journal of patient account management"] = "J Patient Acc Manage",
  ["journal of patient experience"] = "J Patient Exp",
  ["journal of patient safety"] = "J Patient Saf",
  ["journal of patient safety and risk management"] = "J Patient Saf Risk Manag",
  ["journal of patient-centered research and reviews"] = "J Patient Cent Res Rev",
  ["journal of patient-reported outcomes"] = "J Patient Rep Outcomes",
  ["journal of peace research"] = "J Peace Res",
  ["journal of peace science"] = "J. Peace Sci.",
  ["journal of pediatric & perinatal nutrition"] = "J Pediatr Perinat Nutr",
  ["journal of pediatric & prenatal nutrition"] = "J. Pediatr. Prenatal Nutr.",
  ["journal of pediatric and adolescent gynecology"] = "J Pediatr Adolesc Gynecol",
  ["journal of pediatric and perinatal nutrition"] = "J. Pediatr. Perinat. Nutr.",
  ["journal of pediatric biochemistry"] = "J Pediatr Biochem",
  ["journal of pediatric endocrinology"] = "J. Pediatr. Endocrinol.",
  ["journal of pediatric endocrinology & metabolism : jpem"] = "J Pediatr Endocrinol Metab",
  ["journal of pediatric endocrinology (jersey city, n.j.)"] = "J Pediatr Endocrinol (Jersey City)",
  ["journal of pediatric endocrinology and metabolism"] = "J. Pediatr. Endocrinol. Metab.",
  ["journal of pediatric epilepsy"] = "J Pediatr Epilepsy",
  ["journal of pediatric gastroenterology and nutrition"] = "J Pediatr Gastroenterol Nutr",
  ["journal of pediatric genetics"] = "J Pediatr Genet",
  ["journal of pediatric health care"] = "J. Pediatr. Health Care",
  ["journal of pediatric health care : official publication of national association of pediatric nurse associates & practitioners"] = "J Pediatr Health Care",
  ["journal of pediatric hematology/oncology"] = "J Pediatr Hematol Oncol",
  ["journal of pediatric hematology/oncology : official journal of the american society of pediatric hematology/oncology"] = "J Pediatr Hematol Oncol",
  ["journal of pediatric infectious diseases"] = "J Pediatr Infect Dis",
  ["journal of pediatric intensive care"] = "J Pediatr Intensive Care",
  ["journal of pediatric neurology : jpn"] = "J Pediatr Neurol",
  ["journal of pediatric neuropsychology"] = "J Pediatr Neuropsychol",
  ["journal of pediatric neuroradiology"] = "J Pediatr Neuroradiol",
  ["journal of pediatric neurosciences"] = "J Pediatr Neurosci",
  ["journal of pediatric nursing"] = "J Pediatr Nurs",
  ["journal of pediatric oncology"] = "J Pediatr Oncol",
  ["journal of pediatric oncology nursing"] = "J. Pediatr. Oncol. Nurs.",
  ["journal of pediatric oncology nursing : official journal of the association of pediatric oncology nurses"] = "J Pediatr Oncol Nurs",
  ["journal of pediatric ophthalmology"] = "J Pediatr Ophthalmol",
  ["journal of pediatric ophthalmology and strabismus"] = "J Pediatr Ophthalmol Strabismus",
  ["journal of pediatric orthopaedics. part b"] = "J. Pediatr. Orthop. B",
  ["journal of pediatric orthopaedics. part b / european paediatric orthopaedic society, pediatric orthopaedic society of north america"] = "J Pediatr Orthop B",
  ["journal of pediatric orthopedics"] = "J Pediatr Orthop",
  ["journal of pediatric orthopedics. part b"] = "J Pediatr Orthop B",
  ["journal of pediatric psychology"] = "J Pediatr Psychol",
  ["journal of pediatric rehabilitation medicine"] = "J Pediatr Rehabil Med",
  ["journal of pediatric surgery"] = "J Pediatr Surg",
  ["journal of pediatric surgery case reports"] = "J Pediatr Surg Case Rep",
  ["journal of pediatric surgical nursing"] = "J Pediatr Surg Nurs",
  ["journal of pediatric urology"] = "J Pediatr Urol",
  ["journal of pediatrics"] = "J. Pediatr.",
  ["journal of pediatrics & child health care"] = "J Pediatr Child Health Care",
  ["journal of pediatrics & neonatal care"] = "J Pediatr Neonatal Care",
  ["journal of pediatrics and child nutrition"] = "J Pediatr Child Nutr",
  ["journal of pediatrics and mother care"] = "J Pediatr Mother Care",
  ["journal of pediatrics and pediatric medicine"] = "J Pediatr Pediatr Med",
  ["journal of pediatrics review"] = "J Pediatr Rev",
  ["journal of pediatrics, perinatology and child health"] = "J Pediatr Perinatol Child Health",
  ["journal of pedodontics"] = "J. Pedod.",
  ["journal of pelvic medicine & surgery"] = "J Pelvic Med Surg",
  ["journal of pension economics & finance"] = "J Pension Econ Financ",
  ["journal of peptide research"] = "J. Pept. Res.",
  ["journal of peptide science"] = "J. Pept. Sci.",
  ["journal of peptide science : an official publication of the european peptide society"] = "J Pept Sci",
  ["journal of performance health research"] = "J Perform Health Res",
  ["journal of performance of constructed facilities"] = "J. Perform. Constr. Facil",
  ["journal of perianesthesia nursing"] = "J. Perianesth. Nurs.",
  ["journal of perianesthesia nursing : official journal of the american society of perianesthesia nurses"] = "J Perianesth Nurs",
  ["journal of perianesthesia nursing : official journal of the american society of perianesthesia nurses / american society of perianesthesia nurses"] = "J Perianesth Nurs",
  ["journal of peridynamics and nonlocal modeling"] = "J. Peridyn. Nonlocal Model.",
  ["journal of perinatal and neonatal nursing"] = "J. Perinat. Neonatal Nurs.",
  ["journal of perinatal medicine"] = "J Perinat Med",
  ["journal of perinatology"] = "J. Perinatol.",
  ["journal of perinatology : official journal of the california perinatal association"] = "J Perinatol",
  ["journal of periodontal & implant science"] = "J Periodontal Implant Sci",
  ["journal of periodontal research"] = "J Periodontal Res",
  ["journal of periodontal research. supplement"] = "J Periodontal Res Suppl",
  ["journal of periodontology"] = "J Periodontol",
  ["journal of perioperative & critical intensive care nursing"] = "J Perioper Crit Intensiv Care Nurs",
  ["journal of perioperative practice"] = "J Perioper Pract",
  ["journal of personality"] = "J Pers",
  ["journal of personality and social psychology"] = "J Pers Soc Psychol",
  ["journal of personality assessment"] = "J Pers Assess",
  ["journal of personality disorders"] = "J Pers Disord",
  ["journal of personalized medicine"] = "J Pers Med",
  ["journal of personalized nanomedicine"] = "J Pers Nanomed",
  ["journal of personnel evaluation in education"] = "J Pers Eval Educ",
  ["journal of pest science"] = "J Pest Sci (2004)",
  ["journal of pesticide science"] = "J Pestic Sci",
  ["journal of petroleum & environmental biotechnology"] = "J Pet Environ Biotechnol",
  ["journal of petroleum exploration and production technology"] = "J Pet Explor Prod Technol",
  ["journal of petroleum geology"] = "J. Pet. Geol",
  ["journal of petroleum science & engineering"] = "J Pet Sci Eng",
  ["journal of petroleum science and technology"] = "J. Pet. Sci. Technol.",
  ["journal of petroleum technology"] = "J. Pet. Technol.",
  ["journal of petrology"] = "J. Petrol.",
  ["journal of pharmaceutical analysis"] = "J Pharm Anal",
  ["journal of pharmaceutical and biomedical analysis"] = "J Pharm Biomed Anal",
  ["journal of pharmaceutical and biomedical sciences"] = "J Pharm Biomed Sci",
  ["journal of pharmaceutical health care and sciences"] = "J Pharm Health Care Sci",
  ["journal of pharmaceutical health services research"] = "J. Pharm. Health Serv. Res.",
  ["journal of pharmaceutical health services research : an official journal of the royal pharmaceutical society of great britain"] = "J Pharm Health Serv Res",
  ["journal of pharmaceutical innovation"] = "J Pharm Innov",
  ["journal of pharmaceutical investigation"] = "J Pharm Investig",
  ["journal of pharmaceutical marketing & management"] = "J Pharm Mark Manage",
  ["journal of pharmaceutical marketing and management"] = "J. Pharm. Mark. Manage.",
  ["journal of pharmaceutical negative results"] = "J Pharm Negat Results",
  ["journal of pharmaceutical policy and practice"] = "J Pharm Policy Pract",
  ["journal of pharmaceutical research international"] = "J Pharm Res Int",
  ["journal of pharmaceutical science and technology"] = "J. Pharm. Sci. Technol.",
  ["journal of pharmaceutical science and technology : the official journal of pda"] = "J Pharm Sci Technol",
  ["journal of pharmaceutical sciences"] = "J Pharm Sci",
  ["journal of pharmaceutical sciences and experimental pharmacology"] = "J Pharm Sci Exp Pharmacol",
  ["journal of pharmaceutical sciences and pharmacology"] = "J Pharm Sci Pharmacol",
  ["journal of pharmaceutical statistics"] = "J. Pharm. Stat.",
  ["journal of pharmaceutical technology & drug research"] = "J Pharm Technol Drug Res",
  ["journal of pharmaceutics"] = "J Pharm (Cairo)",
  ["journal of pharmaceutics & drug delivery research"] = "J Pharm Drug Deliv Res",
  ["journal of pharmaceutics & pharmacology"] = "J Pharm Pharmacol (Los Angel)",
  ["journal of pharmaceutics and drug research"] = "J Pharm Drug Res",
  ["journal of pharmacobio-dynamics"] = "J Pharmacobiodyn",
  ["journal of pharmacogenomics & pharmacoproteomics"] = "J Pharmacogenomics Pharmacoproteomics",
  ["journal of pharmacognosy & natural products"] = "J Pharmacogn Nat Prod",
  ["journal of pharmacognosy and phytochemistry"] = "J Pharmacogn Phytochem",
  ["journal of pharmacokinetics and biopharmaceutics"] = "J Pharmacokinet Biopharm",
  ["journal of pharmacokinetics and pharmacodynamics"] = "J Pharmacokinet Pharmacodyn",
  ["journal of pharmacological & biomedical analysis"] = "J Pharmacol Biomed Anal",
  ["journal of pharmacological and toxicological methods"] = "J Pharmacol Toxicol Methods",
  ["journal of pharmacological methods"] = "J Pharmacol Methods",
  ["journal of pharmacological reports"] = "J Pharmacol Rep",
  ["journal of pharmacological sciences"] = "J Pharmacol Sci",
  ["journal of pharmacological sciences (tokyo, japan)"] = "J. Pharmacol. Sci. (Tokyo, Jpn.)",
  ["journal of pharmacology & clinical toxicology"] = "J Pharmacol Clin Toxicol",
  ["journal of pharmacology & pharmacotherapeutics"] = "J Pharmacol Pharmacother",
  ["journal of pharmacology and clinical trials"] = "J Pharmacol Clin Trials",
  ["journal of pharmacology and experimental therapeutics"] = "J. Pharmacol. Exp. Ther.",
  ["journal of pharmacology and pharmacotherapeutics"] = "J. Pharmacol. Pharmacother.",
  ["journal of pharmacopuncture"] = "J Pharmacopuncture",
  ["journal of pharmacotherapy"] = "J Pharmacother",
  ["journal of pharmacovigilance"] = "J Pharmacovigil",
  ["journal of pharmacy & bioallied sciences"] = "J Pharm Bioallied Sci",
  ["journal of pharmacy & pharmaceutical sciences : a publication of the canadian society for pharmaceutical sciences, societe canadienne des sciences pharmaceutiques"] = "J Pharm Pharm Sci",
  ["journal of pharmacy & pharmaceutical sciences : a publication of the canadian society for pharmaceutical sciences, société canadienne des sciences pharmaceutiques"] = "J Pharm Pharm Sci",
  ["journal of pharmacy & pharmacognosy research"] = "J Pharm Pharmacogn Res",
  ["journal of pharmacy and chemistry"] = "J. Pharm. Chem.",
  ["journal of pharmacy and pharmaceutical sciences"] = "J. Pharm. Pharm. Sci.",
  ["journal of pharmacy and pharmaceutics"] = "J Pharm Pharm",
  ["journal of pharmacy and pharmacology"] = "J. Pharm. Pharmacol.",
  ["journal of pharmacy practice"] = "J Pharm Pract",
  ["journal of pharmacy practice and research"] = "J. Pharm. Pract. Res.",
  ["journal of pharmacy research"] = "J. Pharm. Res.",
  ["journal of pharmacy technology"] = "J. Pharm. Technol.",
  ["journal of phase equilibria"] = "J. Phase Equilibria",
  ["journal of phase equilibria and diffusion"] = "J. Phase Equilib. Diffus.",
  ["journal of phase equilibrium"] = "J. Phase Equilib.",
  ["journal of phase equilibrium and diffusion"] = "J. Phase Equilib. Diffus.",
  ["journal of philippine development : a technical journal of the national economic and development authority"] = "J Philipp Dev",
  ["journal of philippine statistics"] = "J Philipp Stat",
  ["journal of philology"] = "JPhil",
  ["journal of philosophical logic"] = "J. Philos. Logic",
  ["journal of philosophy"] = "JPh",
  ["journal of philosophy, science and law"] = "J. Philos. Sci. Law",
  ["journal of phonetics"] = "J Phon",
  ["journal of photocatalysis"] = "J. Photocatal.",
  ["journal of photochemistry"] = "J. Photochem.",
  ["journal of photochemistry and photobiology"] = "J Photochem Photobiol",
  ["journal of photochemistry and photobiology a: chemistry"] = "J. Photochem. Photobiol., A",
  ["journal of photochemistry and photobiology b: biology"] = "J. Photochem. Photobiol., B",
  ["journal of photochemistry and photobiology c: photochemistry reviews"] = "J. Photochem. Photobiol., C",
  ["journal of photochemistry and photobiology, a: chemistry"] = "J. Photochem. Photobiol., A",
  ["journal of photochemistry and photobiology, b: biology"] = "J. Photochem. Photobiol., B",
  ["journal of photochemistry and photobiology, c: photochemistry reviews"] = "J. Photochem. Photobiol., C",
  ["journal of photochemistry and photobiology. a, chemistry"] = "J Photochem Photobiol A Chem",
  ["journal of photochemistry and photobiology. b, biology"] = "J Photochem Photobiol B",
  ["journal of photogrammetry and remote sensing"] = "J. Photogramm. Remote Sens.",
  ["journal of photonics for energy"] = "J. Photonics Energy",
  ["journal of photopolymer science and technology"] = "J. Photopolym. Sci. Technol.",
  ["journal of photopolymer science and technology = [fotoporimā konwakai shi]"] = "J Photopolym Sci Technol",
  ["journal of phycology"] = "J Phycol",
  ["journal of phylogenetics & evolutionary biology"] = "J Phylogenetics Evol Biol",
  ["journal of physical activity & health"] = "J Phys Act Health",
  ["journal of physical activity research"] = "J Phys Act Res",
  ["journal of physical activity, nutrition and rehabilitation"] = "J Phys Act Nutr Rehabil",
  ["journal of physical and chemical reference data"] = "J. Phys. Chem. Ref. Data",
  ["journal of physical and colloid chemistry"] = "J. Phys. Colloid Chem.",
  ["journal of physical chemistry"] = "J. Phys. Chem.",
  ["journal of physical chemistry & biophysics"] = "J Phys Chem Biophys",
  ["journal of physical chemistry a"] = "J. Phys. Chem. A",
  ["journal of physical chemistry b"] = "J. Phys. Chem. B",
  ["journal of physical chemistry c"] = "J. Phys. Chem. C",
  ["journal of physical chemistry letters"] = "J. Phys. Chem. Lett.",
  ["journal of physical education"] = "J Phys Educ",
  ["journal of physical education (dayton, ohio)"] = "J Phys Educ",
  ["journal of physical education and sport management"] = "J Physic Educ Sport Manag",
  ["journal of physical education, recreation & dance"] = "J Phys Educ Recreat Dance",
  ["journal of physical mathematics"] = "J. Phys. Math.",
  ["journal of physical medicine and rehabilitation (wilmington, del.)"] = "J Phys Med Rehabil",
  ["journal of physical oceanography"] = "J. Phys. Oceanogr.",
  ["journal of physical organic chemistry"] = "J. Phys. Org. Chem.",
  ["journal of physical science"] = "J. Phys. Sci.",
  ["journal of physical studies. zhurnal fizichnykh doslidzhen"] = "J. Phys. Stud.",
  ["journal of physical therapy and health promotion"] = "J Phys Ther Health Promot",
  ["journal of physical therapy science"] = "J Phys Ther Sci",
  ["journal of physics"] = "J Phys",
  ["journal of physics (moscow)"] = "J. Phys. (Moscow)",
  ["journal of physics a"] = "J. Phys. A",
  ["journal of physics a: general physics"] = "J. Phys. A: Gen. Phys.",
  ["journal of physics a: mathematical and general"] = "J. Phys. A: Math. Gen.",
  ["journal of physics a: mathematical and theoretical"] = "J. Phys. A: Math. Theor.",
  ["journal of physics a: mathematical, nuclear and general"] = "J. Phys. A: Math. Nucl. Gen.",
  ["journal of physics and chemistry of solids"] = "J. Phys. Chem. Solids",
  ["journal of physics b"] = "J. Phys. B",
  ["journal of physics b: atomic and molecular physics"] = "J. Phys. B: At. Mol. Phys.",
  ["journal of physics b: atomic, molecular and optical"] = "J. Phys. B",
  ["journal of physics b: atomic, molecular and optical physics"] = "J. Phys. B: At. Mol. Opt. Phys.",
  ["journal of physics c"] = "J. Phys. C",
  ["journal of physics c: solid state physics"] = "J. Phys. C: Solid State Phys.",
  ["journal of physics communications"] = "J. Phys. Commun.",
  ["journal of physics d"] = "J. Phys. D",
  ["journal of physics d: applied physics"] = "J. Phys. D: Appl. Phys.",
  ["journal of physics e"] = "J. Phys. E",
  ["journal of physics e: scientific instruments"] = "J. Phys. E: Sci. Instrum.",
  ["journal of physics f"] = "J. Phys. F",
  ["journal of physics f: metal physics"] = "J. Phys. F: Met. Phys.",
  ["journal of physics g"] = "J. Phys. G",
  ["journal of physics g: nuclear and particle physics"] = "J. Phys. G: Nucl. Part. Phys.",
  ["journal of physics g: nuclear physics"] = "J. Phys. G: Nucl. Phys.",
  ["journal of physics. a, mathematical and theoretical"] = "J Phys A Math Theor",
  ["journal of physics. a. mathematical and theoretical"] = "J. Phys. A",
  ["journal of physics. b, atomic, molecular, and optical physics : an institute of physics journal"] = "J Phys B At Mol Opt Phys",
  ["journal of physics. condensed matter : an institute of physics journal"] = "J Phys Condens Matter",
  ["journal of physics. conference series"] = "J. Phys. Conf. Ser.",
  ["journal of physics. e, scientific instruments"] = "J. Phys. [E]",
  ["journal of physics. g, nuclear and particle physics : an institute of physics journal"] = "J Phys G Nucl Part Phys",
  ["journal of physics: complexity"] = "J. Phys. Complexity",
  ["journal of physics: condensed matter"] = "J. Phys.: Condens. Matter",
  ["journal of physics: conference series"] = "J. Phys. Conf. Ser.",
  ["journal of physics: energy"] = "J. Phys. Energy",
  ["journal of physics: materials"] = "J. Phys. Mater.",
  ["journal of physics: photonics"] = "J. Phys. Photonics",
  ["journal of physiological anthropology"] = "J Physiol Anthropol",
  ["journal of physiological anthropology and applied human science"] = "J Physiol Anthropol Appl Human Sci",
  ["journal of physiological sciences"] = "J. Physiol. Sci.",
  ["journal of physiology"] = "J. Physiol.",
  ["journal of physiology (oxford, united kingdom)"] = "J. Physiol. (Oxford, U. K.)",
  ["journal of physiology and biochemistry"] = "J Physiol Biochem",
  ["journal of physiology and pharmacology"] = "J. Physiol. Pharmacol.",
  ["journal of physiology and pharmacology : an official journal of the polish physiological society"] = "J Physiol Pharmacol",
  ["journal of physiology, paris"] = "J Physiol Paris",
  ["journal of physiology-paris"] = "J. Physiol.-Paris",
  ["journal of physiotherapy"] = "J Physiother",
  ["journal of physiotherapy & physical rehabilitation"] = "J Physiother Phys Rehabil",
  ["journal of phytopathology"] = "J. Phytopathol.",
  ["journal of phytopathology (1986)"] = "J Phytopathol (1986)",
  ["journal of pierre fauchard academy"] = "J. Pierre Fauchard Acad.",
  ["journal of pierre fauchard academy (pierre fauchard academy. india section)"] = "J Pierre Fauchard Acad",
  ["journal of pigmentary disorders"] = "J Pigment Disord",
  ["journal of pineal research"] = "J Pineal Res",
  ["journal of pipeline science and engineering"] = "J. Pipeline Sci. Eng.",
  ["journal of pipeline systems engineering and practice"] = "J. Pipeline Syst. Eng. Pract.",
  ["journal of place management and development"] = "J. Place Manage. Dev.",
  ["journal of planar chromatography - modern tlc"] = "J. Planar. Chromatogr. - Mod. TLC",
  ["journal of planar chromatography–modern tlc"] = "J. Planar Chromatogr.–Mod. TLC",
  ["journal of plankton research"] = "J Plankton Res",
  ["journal of planning education and research"] = "J Plan Educ Res",
  ["journal of planning history"] = "J Plan Hist",
  ["journal of planning literature"] = "J Plan Lit",
  ["journal of plant biochemistry & physiology"] = "J Plant Biochem Physiol",
  ["journal of plant biochemistry and biotechnology"] = "J. Plant Biochem. Biotechnol.",
  ["journal of plant biology"] = "J. Plant Biol.",
  ["journal of plant biology = singmul hakhoe chi"] = "J Plant Biol",
  ["journal of plant biotechnology"] = "J Plant Biotechnol",
  ["journal of plant breeding and crop science"] = "J Plant Breed Crop Sci",
  ["journal of plant diseases and protection"] = "J. Plant Dis. Prot.",
  ["journal of plant diseases and protection : scientific journal of the german phytomedical society (dpg)"] = "J Plant Dis Prot (2006)",
  ["journal of plant ecology"] = "J. Plant Ecol.",
  ["journal of plant ecology : jpe"] = "J Plant Ecol",
  ["journal of plant growth regulation"] = "J. Plant Growth Regul.",
  ["journal of plant interactions"] = "J. Plant Interact.",
  ["journal of plant molecular biology and biotechnology"] = "J Plant Mol Biol Biotechnol",
  ["journal of plant nutrition"] = "J. Plant Nutr.",
  ["journal of plant nutrition and soil science"] = "J. Plant Nutr. Soil Sci.",
  ["journal of plant nutrition and soil science = zeitschrift für pflanzenernährung und bodenkunde"] = "J Plant Nutr Soil Sci (1999)",
  ["journal of plant pathology"] = "J. Plant Pathol.",
  ["journal of plant pathology & microbiology"] = "J Plant Pathol Microbiol",
  ["journal of plant pathology : an international journal of the italian phytopathological society"] = "J Plant Pathol",
  ["journal of plant physiology"] = "J Plant Physiol",
  ["journal of plant physiology & pathology"] = "J Plant Physiol Pathol",
  ["journal of plant protection research"] = "J. Plant Prot. Res.",
  ["journal of plant registrations"] = "J Plant Regist",
  ["journal of plant research"] = "J Plant Res",
  ["journal of plant studies"] = "J Plant Stud",
  ["journal of plasma physics"] = "J. Plasma Phys.",
  ["journal of plastic and reconstructive surgical nursing"] = "J. Plast. Reconstr. Surg. Nurs.",
  ["journal of plastic film & sheeting"] = "J. Plast. Film Sheeting",
  ["journal of plastic surgery and hand surgery"] = "J Plast Surg Hand Surg",
  ["journal of plastic, reconstructive & aesthetic surgery : jpras"] = "J Plast Reconstr Aesthet Surg",
  ["journal of plastic, reconstructive and aesthetic surgery"] = "J. Plast. Reconstr. Aesthet. Surg.",
  ["journal of poetry therapy"] = "J Poet Ther",
  ["journal of police and criminal psychology"] = "J Police Crim Psychol",
  ["journal of police crisis negotiations : an international journal"] = "J Police Crisis Negot",
  ["journal of policy analysis and management"] = "J. Pol. Anal. Manage.",
  ["journal of policy analysis and management : [the journal of the association for public policy analysis and management]"] = "J Policy Anal Manage",
  ["journal of policy and practice in intellectual disabilities"] = "J Policy Pract Intellect Disabil",
  ["journal of policy history : jph"] = "J Policy Hist",
  ["journal of policy modeling"] = "J Policy Model",
  ["journal of policy practice"] = "J Policy Pract",
  ["journal of policy reform"] = "J. Pol. Reform",
  ["journal of political & military sociology"] = "J. Polit. Mil. Soc.",
  ["journal of political & military sociology : jpms"] = "J Polit Mil Soc",
  ["journal of political ecology"] = "J Polit Ecol",
  ["journal of political economy"] = "J. Polit. Economy",
  ["journal of political science"] = "J Polit Sci",
  ["journal of politics and law"] = "J Politics Law",
  ["journal of pollination ecology"] = "J Pollinat Ecol",
  ["journal of polymer engineering"] = "J. Polym. Eng.",
  ["journal of polymer materials"] = "J. Polym. Mater.",
  ["journal of polymer research"] = "J. Polym. Res.",
  ["journal of polymer science"] = "J. Polym. Sci.",
  ["journal of polymer science (2020)"] = "J Polym Sci (2020)",
  ["journal of polymer science part a: polymer chemistry"] = "J. Polym. Sci., Part A: Polym. Chem.",
  ["journal of polymer science part b: polymer physics"] = "J. Polym. Sci., Part B: Polym. Phys.",
  ["journal of polymer science part c: polymer letters"] = "J. Polym. Sci., Part C: Polym. Lett.",
  ["journal of polymer science polymer symposium"] = "J. Polym. Sci. Polym. Symp.",
  ["journal of polymer science, part a: polymer chemistry"] = "J. Polym. Sci., Part A: Polym. Chem.",
  ["journal of polymer science, part b: polymer physics"] = "J. Polym. Sci., Part B: Polym. Phys.",
  ["journal of polymer science, polymer letters edition"] = "J. Polym. Sci. Polym. Lett. Ed.",
  ["journal of polymer science, polymer physics edition"] = "J. Polym. Sci. Polym. Phys. Ed.",
  ["journal of polymer science. part a, general papers"] = "J. Polym. Sci. A",
  ["journal of polymer science. part a, polymer chemistry"] = "J Polym Sci A Polym Chem",
  ["journal of polymer science. part a-1, polymer chemistry"] = "J Polym Sci A1",
  ["journal of polymer science. part b, polymer letters"] = "J. Polym. Sci. [B]",
  ["journal of polymer science. part b, polymer physics"] = "J Polym Sci B Polym Phys",
  ["journal of polymer science. part b: polymer letters"] = "J Polym Sci B",
  ["journal of polymer science: part a, general papers"] = "J Polym Sci A",
  ["journal of polymers"] = "J Polym",
  ["journal of polymers and the environment"] = "J. Polym. Environ.",
  ["journal of popular culture"] = "J Pop Cult",
  ["journal of population"] = "J Popul",
  ["journal of population ageing"] = "J Popul Ageing",
  ["journal of population economics"] = "J Popul Econ",
  ["journal of population research"] = "J Popul Res",
  ["journal of population research (canberra, a.c.t.)"] = "J Popul Res (Canberra)",
  ["journal of population therapeutics and clinical pharmacology = journal de la thérapeutique des populations et de la pharmacologie clinique"] = "J Popul Ther Clin Pharmacol",
  ["journal of porous materials"] = "J. Porous Mater.",
  ["journal of porous media"] = "J Porous Media",
  ["journal of porphyrins and phthalocyanines"] = "J. Porphyrins Phthalocyanines",
  ["journal of portfolio management"] = "J. Portfol. Manage.",
  ["journal of positive behavior interventions"] = "J Posit Behav Interv",
  ["journal of post anesthesia nursing"] = "J Post Anesth Nurs",
  ["journal of post keynesian economics"] = "J Post Keynes Econ",
  ["journal of postcolonial writing"] = "J Postcolon Writ",
  ["journal of postgraduate medicine"] = "J Postgrad Med",
  ["journal of posthuman studies : philosophy, technology, media"] = "J Posthum Stud",
  ["journal of postsecondary education and disability"] = "J Postsecond Educ Disabil",
  ["journal of poultry science"] = "J. Poult. Sci.",
  ["journal of poverty"] = "J Poverty",
  ["journal of poverty alleviation and international development"] = "J Poverty Alleviation Int Dev",
  ["journal of power electronics"] = "J. Power Electron.",
  ["journal of power sources"] = "J. Power Sources",
  ["journal of power sources advances"] = "J. Power Sources Adv.",
  ["journal of practical nursing"] = "J. Pract. Nurs.",
  ["journal of pragmatics"] = "J Pragmat",
  ["journal of pre-college engineering education research"] = "J Precoll Eng Educ Res",
  ["journal of precision medicine: health and disease"] = "J. Precis. Med.: Health Dis.",
  ["journal of precision respiratory medicine"] = "J Precis Respir Med",
  ["journal of pregnancy"] = "J Pregnancy",
  ["journal of pregnancy and child health"] = "J Pregnancy Child Health",
  ["journal of prehistoric religion"] = "JPR",
  ["journal of prenatal & perinatal psychology & health"] = "J Prenat Perinat Psychol Health",
  ["journal of prenatal medicine"] = "J Prenat Med",
  ["journal of presbyterian history"] = "J Presbyt Hist",
  ["journal of pressure vessel technology"] = "J. Pressure Vessel Technol.",
  ["journal of prevention"] = "J. Prev.",
  ["journal of prevention & intervention in the community"] = "J Prev Interv Community",
  ["journal of prevention and health promotion"] = "J Prev Health Promot",
  ["journal of preventive and social medicine : jopsom : a bi-annual journal of the national institute of preventive and social medicine"] = "J Prev Soc Med",
  ["journal of preventive cardiology"] = "J Preventive Cardiol",
  ["journal of preventive dentistry"] = "J. Prev. Dent.",
  ["journal of preventive medicine"] = "J Prev Med (Wilmington)",
  ["journal of preventive medicine and hygiene"] = "J Prev Med Hyg",
  ["journal of preventive medicine and public health"] = "J. Prev. Med. Pub. Health",
  ["journal of preventive medicine and public health = yebang ŭihakhoe chi"] = "J Prev Med Public Health",
  ["journal of preventive psychiatry"] = "J Prev Psychiatry",
  ["journal of primary care & community health"] = "J Prim Care Community Health",
  ["journal of primary health care"] = "J Prim Health Care",
  ["journal of primatology"] = "J Primatol",
  ["journal of prime research in mathematics"] = "J. Prime Res. Math.",
  ["journal of prison & jail health"] = "J Prison Jail Health",
  ["journal of prison health"] = "J Prison Health",
  ["journal of prisoners on prisons"] = "J Prison Prisons",
  ["journal of private enterprise"] = "J. Private Enterprise",
  ["journal of probability and statistics"] = "J. Probab. Stat.",
  ["journal of probiotics & health"] = "J Probiotics Health",
  ["journal of process control"] = "J. Process Control",
  ["journal of product innovation management"] = "J. Prod. Innovation Manage.",
  ["journal of production agriculture"] = "J. Prod. Agric.",
  ["journal of productivity analysis"] = "J. Prod. Anal.",
  ["journal of products liability"] = "J Prod Liability",
  ["journal of professional issues in engineering education and practice"] = "J. Prof. Issues Eng. Educ. Pract.",
  ["journal of professional nursing"] = "J. Prof. Nurs.",
  ["journal of professional nursing : official journal of the american association of colleges of nursing"] = "J Prof Nurs",
  ["journal of professional services marketing"] = "J Prof Serv Mark",
  ["journal of progressive human services"] = "J Progress Hum Serv",
  ["journal of progressive research in social sciences"] = "J Progress Res Soc Sci",
  ["journal of projective techniques"] = "J Proj Tech",
  ["journal of projective techniques & personality assessment"] = "J Proj Tech Pers Assess",
  ["journal of projective techniques and personality assessment"] = "J. Proj. Tech. Pers. Assess.",
  ["journal of propulsion and power"] = "J. Propul. Power",
  ["journal of prosthetic dentistry"] = "J. Prosthet. Dent.",
  ["journal of prosthetics and orthotics : jpo"] = "J Prosthet Orthot",
  ["journal of prosthodontic research"] = "J Prosthodont Res",
  ["journal of prosthodontics"] = "J. Prosthodont.",
  ["journal of prosthodontics : official journal of the american college of prosthodontists"] = "J Prosthodont",
  ["journal of protein chemistry"] = "J Protein Chem",
  ["journal of proteins and proteomics"] = "J Proteins Proteom",
  ["journal of proteolysis"] = "J Proteolysis",
  ["journal of proteome research"] = "J Proteome Res",
  ["journal of proteome science & computational biology"] = "J. Proteome Sci. Comput. Biol.",
  ["journal of proteome science and computational biology"] = "J Proteome Sci Comput Biol",
  ["journal of proteomics"] = "J Proteomics",
  ["journal of proteomics & bioinformatics"] = "J Proteomics Bioinform",
  ["journal of proteomics and genomics research"] = "J Proteom Genom Res",
  ["journal of proton therapy"] = "J Proton Ther",
  ["journal of protozoology"] = "J. Protozool.",
  ["journal of pseudo-differential operators and applications"] = "J. Pseudo-Differ. Oper. Appl.",
  ["journal of psoriasis and psoriatic arthritis"] = "J Psoriasis Psoriatic Arthritis",
  ["journal of psychedelic drugs"] = "J. Psychedelic Drugs",
  ["journal of psychedelic studies"] = "J Psychedelic Stud",
  ["journal of psychiatric and mental health nursing"] = "J Psychiatr Ment Health Nurs",
  ["journal of psychiatric education"] = "J Psychiatr Educ",
  ["journal of psychiatric nursing"] = "J Psychiatr Nurs",
  ["journal of psychiatric nursing and mental health services"] = "J Psychiatr Nurs Ment Health Serv",
  ["journal of psychiatric practice"] = "J Psychiatr Pract",
  ["journal of psychiatric research"] = "J Psychiatr Res",
  ["journal of psychiatric social work"] = "J Psychiatr Soc Work",
  ["journal of psychiatric treatment and evaluation"] = "J Psychiatr Treat Eval",
  ["journal of psychiatry"] = "J Psychiatry",
  ["journal of psychiatry & neuroscience : jpn"] = "J Psychiatry Neurosci",
  ["journal of psychiatry and behavioral science"] = "J Psychiatry Behav Sci",
  ["journal of psychiatry and brain science"] = "J Psychiatr Brain Sci",
  ["journal of psychiatry and cognitive behaviour"] = "J Psychiatry Cogn Behav",
  ["journal of psychiatry and neuroscience"] = "J. Psychiatry Neurosci.",
  ["journal of psychiatry and psychiatric disorders"] = "J Psychiatry Psychiatr Disord",
  ["journal of psychoactive drugs"] = "J Psychoactive Drugs",
  ["journal of psychoeducational assessment"] = "J Psychoeduc Assess",
  ["journal of psycholinguistic research"] = "J Psycholinguist Res",
  ["journal of psychological abnormalities in children"] = "J Psychol Abnorm Child",
  ["journal of psychological medicine and mental pathology (london, england : 1848)"] = "J Psychol Med Ment Pathol",
  ["journal of psychological medicine and mental pathology (london, england : 1875)"] = "J Psychol Med Ment Pathol (Lond)",
  ["journal of psychologists and counsellors in schools"] = "J Psychol Couns Sch",
  ["journal of psychology"] = "J. Psychol.",
  ["journal of psychology & behavior research"] = "J Psychol Behav Res",
  ["journal of psychology & human sexuality"] = "J Psychol Human Sex",
  ["journal of psychology & psychotherapy"] = "J Psychol Psychother",
  ["journal of psychology and ai"] = "J. Psychol. AI",
  ["journal of psychology and cognition"] = "J Psychol Cogn",
  ["journal of psychology and judaism"] = "J Psychol Judaism",
  ["journal of psychology and psychiatry"] = "J Psychol Psychiatr",
  ["journal of psychology and theology"] = "J Psychol Theol",
  ["journal of psychology in africa (south of the sahara, the caribbean, and afro-latin america)"] = "J Psychol Afr",
  ["journal of psychology, neuropsychiatric disorders and brain stimulation"] = "J Psychol Neuropsychiatr Disord Brain Stimul",
  ["journal of psychopathology and behavioral assessment"] = "J Psychopathol Behav Assess",
  ["journal of psychopharmacology"] = "J. Psychopharmacol.",
  ["journal of psychopharmacology (oxford, england)"] = "J Psychopharmacol",
  ["journal of psychophysiology"] = "J Psychophysiol",
  ["journal of psychosocial nursing and mental health services"] = "J Psychosoc Nurs Ment Health Serv",
  ["journal of psychosocial oncology"] = "J Psychosoc Oncol",
  ["journal of psychosocial oncology research and practice"] = "J Psychosoc Oncol Res Pract",
  ["journal of psychosocial rehabilitation and mental health"] = "J Psychosoc Rehabil Ment Health",
  ["journal of psychosomatic obstetrics and gynaecology"] = "J Psychosom Obstet Gynaecol",
  ["journal of psychosomatic research"] = "J Psychosom Res",
  ["journal of psychotherapy integration"] = "J Psychother Integr",
  ["journal of psychotherapy practice and research"] = "J. Psychother. Pract. Res.",
  ["journal of public administration research and theory : j-part"] = "J Public Adm Res Theory",
  ["journal of public affairs"] = "J Public Aff",
  ["journal of public and international affairs"] = "J. Public Int. Aff.",
  ["journal of public child welfare"] = "J Public Child Welf",
  ["journal of public economic theory"] = "J Public Econ Theory",
  ["journal of public economics"] = "J Public Econ",
  ["journal of public finance and public choice"] = "J. Public Finance Public Choice",
  ["journal of public health"] = "J Public Health (Bangkok)",
  ["journal of public health (oxford, england)"] = "J Public Health (Oxf)",
  ["journal of public health and emergency"] = "J Public Health Emerg",
  ["journal of public health and epidemiology"] = "J Public Health Epidemiol",
  ["journal of public health dentistry"] = "J Public Health Dent",
  ["journal of public health in africa"] = "J Public Health Afr",
  ["journal of public health issues and practices"] = "J Public Health Issues Pract",
  ["journal of public health management and practice"] = "J. Public Health Manag. Pract.",
  ["journal of public health management and practice : jphmp"] = "J Public Health Manag Pract",
  ["journal of public health medicine"] = "J Public Health Med",
  ["journal of public health policy"] = "J Public Health Policy",
  ["journal of public health research"] = "J Public Health Res",
  ["journal of public health, and sanitary review"] = "J Public Health Sanit Rev",
  ["journal of public mental health"] = "J Public Ment Health",
  ["journal of public policy"] = "J Public Policy",
  ["journal of public policy & marketing : jpp&m : an annual publication of the division of research, graduate school of business administration, the university of michigan"] = "J Public Policy Mark",
  ["journal of public transportation"] = "J Public Trans",
  ["journal of pulmonary & respiratory medicine"] = "J Pulm Respir Med",
  ["journal of pulp and paper science"] = "J. Pulp Pap. Sci.",
  ["journal of punjab academy of forensic medicine & toxicology"] = "J Punjab Acad Forensic Med Toxicol",
  ["journal of punjab academy of forensic medicine and toxicology"] = "J. Pubjab Acad. Forensic Med. Toxicol.",
  ["journal of pure & applied microbiology"] = "J Pure Appl Microbiol",
  ["journal of pure and applied algebra"] = "J. Pure Appl. Algebra",
  ["journal of pure and applied mathematics. advances and applications"] = "J. Pure Appl. Math. Adv. Appl.",
  ["journal of pure mathematics"] = "J. Pure Math.",
  ["journal of qingdao university"] = "J. Qingdao Univ. Nat. Sci. Ed.",
  ["journal of qualitative criminal justice & criminology : jqcjc"] = "J Qual Crim Justice Criminol",
  ["journal of quality assurance"] = "J. Qual. Assur.",
  ["journal of quality assurance : a publication of the national association of quality assurance professionals"] = "J Qual Assur",
  ["journal of quality in clinical practice"] = "J Qual Clin Pract",
  ["journal of quality in maintenance engineering"] = "J Qual Maint Eng",
  ["journal of quality technology"] = "J. Qual. Technol.",
  ["journal of quantitative analysis in sports"] = "J Quant Anal Sports",
  ["journal of quantitative criminology"] = "J Quant Criminol",
  ["journal of quantitative economics"] = "J. Quant. Econ.",
  ["journal of quantitative economics : journal of the indian econometric society"] = "J Quant Econ",
  ["journal of quantitative economics. jingji shuxue"] = "J. Quant. Econ.",
  ["journal of quantitative linguistics"] = "J Quant Linguist",
  ["journal of quantitative spectroscopy & radiative transfer"] = "J Quant Spectrosc Radiat Transf",
  ["journal of quantitative spectroscopy and radiative transfer"] = "J. Quant. Spectrosc. Radiat. Transfer",
  ["journal of quantitative trait loci : jqtl"] = "J Quant Trait Loci",
  ["journal of quantum computing"] = "J. Quantum Comput.",
  ["journal of quaternary science"] = "J Quat Sci",
  ["journal of race, ethnicity, and politics"] = "J Race Ethn Polit",
  ["journal of racial and ethnic health disparities"] = "J Racial Ethn Health Disparities",
  ["journal of radars"] = "J. Radars",
  ["journal of radiation oncology"] = "J Radiat Oncol",
  ["journal of radiation oncology and research"] = "J Radiat Oncol Res",
  ["journal of radiation research"] = "J Radiat Res",
  ["journal of radiation research and applied sciences"] = "J. Radiat. Res. Appl. Sci.",
  ["journal of radio electronics"] = "J. Radio Electron.",
  ["journal of radioanalytical and nuclear chemistry"] = "J. Radioanal. Nucl. Chem.",
  ["journal of radioanalytical and nuclear chemistry articles"] = "J. Radioanal. Nucl. Chem. Art.",
  ["journal of radioanalytical and nuclear chemistry letters"] = "J. Radioanal. Nucl. Chem. Lett.",
  ["journal of radiological protection"] = "J. Radiol. Prot.",
  ["journal of radiological protection : official journal of the society for radiological protection"] = "J Radiol Prot",
  ["journal of radiology and radiation therapy"] = "J Radiol Radiat Ther",
  ["journal of radiology case reports"] = "J Radiol Case Rep",
  ["journal of radiology nursing"] = "J Radiol Nurs",
  ["journal of radiosurgery and sbrt"] = "J Radiosurg SBRT",
  ["journal of radiotherapy in practice"] = "J. Radiother. Pract.",
  ["journal of rail and rapid transit"] = "J. Rail Rapid Transit",
  ["journal of rail transport planning and management"] = "J. Rail Transp. Plann. Manage.",
  ["journal of rajasthan academy of physical sciences"] = "J. Rajasthan Acad. Phys. Sci.",
  ["journal of raman specroscopy"] = "J. Raman Spectrosc.",
  ["journal of raman spectroscopy"] = "J. Raman Spectrosc.",
  ["journal of raman spectroscopy : jrs"] = "J Raman Spectrosc",
  ["journal of ramanujan society of mathematics and mathematical sciences"] = "J. Ramanujan Soc. Math. Math. Sci.",
  ["journal of rapid methods and automation in microbiology"] = "J. Rapid Methods Autom. Microbiol.",
  ["journal of raptor research"] = "J. Raptor Res.",
  ["journal of rare diseases research & treatment"] = "J Rare Dis Res Treat",
  ["journal of rare earths"] = "J. Rare Earths",
  ["journal of rational-emotive and cognitive-behavior therapy : ret"] = "J Ration Emot Cogn Behav Ther",
  ["journal of real estate finance and economics"] = "J. Real Estate Finance Econ.",
  ["journal of real estate literature"] = "J. Real Estate Lit.",
  ["journal of real estate portfolio management"] = "J. Real Estate Portfol. Manage.",
  ["journal of real estate practice and education"] = "J. Real Estate Practice Educ.",
  ["journal of real estate research"] = "J. Real Estate Res.",
  ["journal of real-time image processing"] = "J Real Time Image Process",
  ["journal of receptor and signal transduction research"] = "J Recept Signal Transduct Res",
  ["journal of receptor research"] = "J Recept Res",
  ["journal of receptor, ligand and channel research"] = "J Receptor Ligand Channel Res",
  ["journal of receptors and signal transduction"] = "J. Recept. Signal Transduction",
  ["journal of reconstructive microsurgery"] = "J Reconstr Microsurg",
  ["journal of reconstructive microsurgery open"] = "J Reconstr Microsurg Open",
  ["journal of recreational mathematics"] = "J. Recreational Math.",
  ["journal of reform judaism"] = "J Reform Jud",
  ["journal of refractive and corneal surgery"] = "J Refract Corneal Surg",
  ["journal of refractive surgery"] = "J. Refract. Surg.",
  ["journal of refractive surgery (thorofare, n.j. : 1995)"] = "J Refract Surg",
  ["journal of refugee studies"] = "J Refug Stud",
  ["journal of regenerative medicine"] = "J Regen Med",
  ["journal of regenerative medicine & tissue engineering"] = "J Regen Med Tissue Eng",
  ["journal of regional analysis and policy"] = "J. Reg. Anal. Pol.",
  ["journal of regional medical campuses"] = "J Reg Med Campuses",
  ["journal of regional policy"] = "J Reg Policy",
  ["journal of regional science"] = "J Reg Sci",
  ["journal of registry management"] = "J Registry Manag",
  ["journal of regulatory economics"] = "J Regul Econ",
  ["journal of rehabilitation"] = "J Rehabil",
  ["journal of rehabilitation and assistive technologies engineering"] = "J Rehabil Assist Technol Eng",
  ["journal of rehabilitation medicine"] = "J Rehabil Med",
  ["journal of rehabilitation medicine : official journal of the uems european board of physical and rehabilitation medicine"] = "J Rehabil Med",
  ["journal of rehabilitation medicine. clinical communications"] = "J Rehabil Med Clin Commun",
  ["journal of rehabilitation medicine. supplement"] = "J Rehabil Med Suppl",
  ["journal of rehabilitation r and d"] = "J. Rehabil. R D",
  ["journal of rehabilitation r&d"] = "J Rehabil R D",
  ["journal of rehabilitation r&d / veterans administration, department of medicine and surgery, rehabilitation r&d service"] = "J Rehabil R D",
  ["journal of rehabilitation research and development"] = "J Rehabil Res Dev",
  ["journal of rehabilitation research and development. clinical supplement"] = "J Rehabil Res Dev Clin Suppl",
  ["journal of rehabilitation research and development. clinical supplement / veterans administration"] = "J Rehabil Res Dev Clin Suppl",
  ["journal of reinforced plastics and composites"] = "J. Reinf. Plast. Compos.",
  ["journal of relationships research"] = "J Relatsh Res",
  ["journal of reliability science and engineering"] = "J. Reliab. Sci. Eng.",
  ["journal of reliable intelligent environments"] = "J Reliab Intell Environ",
  ["journal of religion"] = "JR",
  ["journal of religion & aging"] = "J Relig Aging",
  ["journal of religion & spirituality in social work"] = "J Relig Spiritual Soc Work",
  ["journal of religion and health"] = "J Relig Health",
  ["journal of religion in africa. religion en afrique"] = "J Relig Afr",
  ["journal of religion, spirituality & aging"] = "J Relig Spiritual Aging",
  ["journal of religious gerontology"] = "J Relig Gerontol",
  ["journal of religious history"] = "JRH",
  ["journal of remanufacturing"] = "J. Remanuf.",
  ["journal of remote sensing"] = "J. Remote Sens.",
  ["journal of renal cancer"] = "J Ren Cancer",
  ["journal of renal care"] = "J Ren Care",
  ["journal of renal injury prevention"] = "J Renal Inj Prev",
  ["journal of renal medicine"] = "J Ren Med",
  ["journal of renal nutrition"] = "J. Ren. Nutr.",
  ["journal of renal nutrition : the official journal of the council on renal nutrition of the national kidney foundation"] = "J Ren Nutr",
  ["journal of renewable & sustainable energy"] = "J. Renew. Sustain. Energy",
  ["journal of renewable and sustainable energy"] = "J. Renewable Sustainable Energy",
  ["journal of renewable energy"] = "J. Renewable Energy",
  ["journal of renewable energy and environment"] = "J. Renewable Energy Environ.",
  ["journal of renewable materials"] = "J Renew Mater",
  ["journal of reproduction & fertility. abstract series"] = "J Reprod Fertil Abstr Ser",
  ["journal of reproduction & infertility"] = "J Reprod Infertil",
  ["journal of reproduction and development"] = "J. Reprod. Dev.",
  ["journal of reproduction and fertility"] = "J Reprod Fertil",
  ["journal of reproduction and fertility. abstract series"] = "J. Reprod. Fertil. Abstr. Ser.",
  ["journal of reproduction and fertility. supplement"] = "J Reprod Fertil Suppl",
  ["journal of reproductive and infant psychology"] = "J Reprod Infant Psychol",
  ["journal of reproductive biology and health"] = "J Reprod Biol Health",
  ["journal of reproductive biotechnology & fertility"] = "J Reprod Biotechnol Fertil",
  ["journal of reproductive biotechnology and fertility"] = "J. Reprod. Biotechnol. Fertil.",
  ["journal of reproductive health and medicine"] = "J Reprod Health Med",
  ["journal of reproductive immunology"] = "J Reprod Immunol",
  ["journal of reproductive medicine"] = "J. Reprod. Med.",
  ["journal of research and practice for adult literacy, secondary, and basic education"] = "J Res Pract Adult Lit Second Basic Educ",
  ["journal of research in applied sciences"] = "J Res Appl Sci",
  ["journal of research in biological sciences"] = "J Res Biol Sci",
  ["journal of research in childhood education : jrce"] = "J Res Child Educ",
  ["journal of research in forestry, wildlife and environment"] = "J Res For Wildl Environ",
  ["journal of research in health sciences"] = "J Res Health Sci",
  ["journal of research in interprofessional practice and education"] = "J Res Interprof Pract Educ",
  ["journal of research in islamic economics"] = "J. Res. Islamic Econ.",
  ["journal of research in medical and dental science"] = "J Res Med Dent Sci",
  ["journal of research in medical sciences : the official journal of isfahan university of medical sciences"] = "J Res Med Sci",
  ["journal of research in nursing : jrn"] = "J Res Nurs",
  ["journal of research in personality"] = "J Res Pers",
  ["journal of research in pharmaceutical economics"] = "J. Res. Pharmaceut. Econ.",
  ["journal of research in pharmacy practice"] = "J Res Pharm Pract",
  ["journal of research in reading"] = "J Res Read",
  ["journal of research in science teaching"] = "J Res Sci Teach",
  ["journal of research in special educational needs : jorsen"] = "J Res Spec Educ Needs",
  ["journal of research of the national bureau of standards"] = "J. Res. Natl. Bur. Stand.",
  ["journal of research of the national bureau of standards (1977)"] = "J Res Natl Bur Stand (1977)",
  ["journal of research of the national bureau of standards, section a: physics and chemistry"] = "J. Res. Natl. Bur. Stand. Sec. A",
  ["journal of research of the national bureau of standards, section b: mathematical sciences"] = "J. Res. Natl. Bur. Stand. Sec. B",
  ["journal of research of the national bureau of standards, section c: engineering and instrumentation"] = "J. Res. Natl. Bur. Stand. Sec. C",
  ["journal of research of the national bureau of standards. section a, physics and chemistry"] = "J Res Natl Bur Stand A Phys Chem",
  ["journal of research of the national institute of standards and technology"] = "J Res Natl Inst Stand Technol",
  ["journal of research on adolescence : the official journal of the society for research on adolescence"] = "J Res Adolesc",
  ["journal of research on educational effectiveness"] = "J Res Educ Eff",
  ["journal of research on the lepidoptera"] = "J. Res. Lepid.",
  ["journal of research practice"] = "J Res Pract",
  ["journal of research updates in polymer science"] = "J. Res. Updates Polym. Sci.",
  ["journal of residuals science and technology"] = "J. Residuals Sci. Technol.",
  ["journal of resources and ecology"] = "J Resour Ecol",
  ["journal of resources, energy and development"] = "J. Resour. Energy Dev.",
  ["journal of responsible innovation"] = "J Responsible Innov",
  ["journal of responsible technology"] = "J. Responsible Technol.",
  ["journal of restorative medicine"] = "J Restor Med",
  ["journal of reward deficiency syndrome"] = "J Reward Defic Syndr",
  ["journal of reward deficiency syndrome and addiction science"] = "J Reward Defic Syndr Addict Sci",
  ["journal of rheology"] = "J. Rheol.",
  ["journal of rheology (melville, ny, united states)"] = "J. Rheol. (Melville, NY, U. S.)",
  ["journal of rheology (new york, ny, united states)"] = "J. Rheol. (N. Y., NY, U. S.)",
  ["journal of rheumatic diseases and treatment"] = "J Rheum Dis Treat",
  ["journal of rheumatology"] = "J. Rheumatol.",
  ["journal of rheumatology research"] = "J Rheumatol Res",
  ["journal of rheumatology. supplement"] = "J. Rheumatol. Suppl.",
  ["journal of rice research"] = "J Rice Res",
  ["journal of risk and insurance"] = "J. Risk Ins.",
  ["journal of risk and uncertainty"] = "J Risk Uncertain",
  ["journal of risk and uncertainty in engineering systems, part a: civil engineering"] = "J. Risk Uncertainty Eng. Syst. Part A: Civ. Eng.",
  ["journal of risk and uncertainty in engineering systems, part b: mechanical engineering"] = "J. Risk Uncertainty Eng. Syst. Part B: Mech. Eng.",
  ["journal of risk research"] = "J Risk Res",
  ["journal of rnai and gene silencing : an international journal of rna and gene targeting research"] = "J RNAi Gene Silencing",
  ["journal of road engineering"] = "J. Road Eng.",
  ["journal of road safety"] = "J. Road Saf.",
  ["journal of robotic surgery"] = "J Robot Surg",
  ["journal of robotic systems"] = "J Robot Syst",
  ["journal of robotics"] = "J. Rob.",
  ["journal of robotics, networking and artificial life"] = "J. Robot. Netw. Artif. Life",
  ["journal of rock mechanics and geotechnical engineering"] = "J. Rock Mech. Geotech. Eng.",
  ["journal of roman archaeology"] = "JRA",
  ["journal of roman military equipment studies"] = "JRMES",
  ["journal of roman pottery studies"] = "JRomPotSt",
  ["journal of roman studies"] = "JRS",
  ["journal of romanian literary studies [electronic resource]"] = "J Rom Lit Stud",
  ["journal of rubber research"] = "J. Rubber Res.",
  ["journal of rubber research (kuala lumpur, malaysia)"] = "J Rubber Res",
  ["journal of rural and community development"] = "J Rural Community Dev",
  ["journal of rural and community psychiatry"] = "J Rural Community Psychiatr",
  ["journal of rural community psychology"] = "J Rural Community Psychol",
  ["journal of rural community psychology (online)"] = "J Rural Community Psychol",
  ["journal of rural development"] = "J Rural Dev",
  ["journal of rural health"] = "J. Rural Health",
  ["journal of rural medicine : jrm"] = "J Rural Med",
  ["journal of rural social sciences"] = "J Rural Soc Sci",
  ["journal of rural studies"] = "J Rural Stud",
  ["journal of russian laser research"] = "J. Russ. Laser Res.",
  ["journal of russian phytopathological society"] = "J Russ Phytopathol Soc",
  ["journal of safety research"] = "J Safety Res",
  ["journal of safety science and management"] = "J. Saf. Sci. Manage.",
  ["journal of safety science and resilience"] = "J. Saf. Sci. Resilience",
  ["journal of safety, health and environmental research"] = "J Saf Health Environ Res",
  ["journal of sandwich structures and materials"] = "J. Sandwich Struct. Mater.",
  ["journal of saudi chemical society"] = "J. Saudi Chem. Soc.",
  ["journal of scandinavian studies in criminology and crime prevention"] = "J Scand Stud Criminol Crime Prev",
  ["journal of scheduling"] = "J. Sched.",
  ["journal of schizophrenia research"] = "J Schizophr Res",
  ["journal of school choice"] = "J Sch Choice",
  ["journal of school health"] = "J. Sch. Health",
  ["journal of school leadership"] = "J Sch Leadersh",
  ["journal of school nursing"] = "J. Sch. Nurs.",
  ["journal of school psychology"] = "J Sch Psychol",
  ["journal of school violence"] = "J Sch Violence",
  ["journal of science"] = "J Sci",
  ["journal of science & medicine"] = "J Sci Med",
  ["journal of science & technology in the tropics"] = "J Sci Technol Tropics",
  ["journal of science and medicine in sport"] = "J Sci Med Sport",
  ["journal of science and medicine in sport / sports medicine australia"] = "J Sci Med Sport",
  ["journal of science and technology policy management"] = "J. Sci. Technol. Policy Manage.",
  ["journal of science education and technology"] = "J. Sci. Educ. Technol.",
  ["journal of science in sport and exercise"] = "J. Sci. Sport Exercise",
  ["journal of science of the university of tehran"] = "J. Sci. Univ. Tehran Int. Ed.",
  ["journal of science teacher education"] = "J Sci Teacher Educ",
  ["journal of science, engineering and technology"] = "J Sci Eng Technol",
  ["journal of science, technology and environment"] = "J Sci Technol Environ",
  ["journal of science: advanced materials and devices"] = "J. Sci. Adv. Mater. Devices",
  ["journal of scientific & industrial research. c, biological sciences"] = "J Sci Ind Res (C)",
  ["journal of scientific & industrial research. c. biological sciences"] = "J Sci Ind Res (C)",
  ["journal of scientific and industrial research"] = "J Sci Ind Res (India)",
  ["journal of scientific and industrial research (new delhi, india : 1942)"] = "J Sci Ind Res (1942)",
  ["journal of scientific computing"] = "J. Sci. Comput.",
  ["journal of scientific conference proceedings"] = "J. Sci. Conf. Proc.",
  ["journal of scientific dentistry"] = "J Sci Dent",
  ["journal of scientific exploration : a publication of the society for scientific exploration"] = "J Sci Explor",
  ["journal of scientific instruments"] = "J. Sci. Instrum.",
  ["journal of scientific research and reports"] = "J Sci Res Rep",
  ["journal of scleroderma and related disorders"] = "J Scleroderma Relat Disord",
  ["journal of scottish historical studies"] = "J. Scott. Hist. Stud.",
  ["journal of sea research"] = "J. Sea Res.",
  ["journal of second language pronunciation"] = "J Second Lang Pronunciation",
  ["journal of second language writing"] = "J Second Lang Writ",
  ["journal of sedimentary petrology"] = "J. Sediment. Petrol.",
  ["journal of sedimentary research"] = "J. Sediment. Res.",
  ["journal of sedimentary research. section a, sedimentary petrology and processes : an international journal of sepm (society for sedimentary geology)"] = "J Sediment Res A Sediment Petrol Process",
  ["journal of seed science"] = "J. Seed Sci.",
  ["journal of seismic exploration"] = "J. Seismic Explor.",
  ["journal of seismology"] = "J. Seismolog.",
  ["journal of semantics"] = "J Semant",
  ["journal of semiconductor technology and science"] = "J. Semicond. Technol. Sci.",
  ["journal of semiconductors"] = "J. Semicond.",
  ["journal of semitic studies"] = "J Semit Stud",
  ["journal of sensor technology"] = "J Sens Technol",
  ["journal of sensors"] = "J Sens",
  ["journal of sensors and sensor systems"] = "J. Sens. Sens. Syst.",
  ["journal of sensory studies"] = "J Sens Stud",
  ["journal of separation science"] = "J Sep Sci",
  ["journal of service research"] = "J Serv Res",
  ["journal of sex & marital therapy"] = "J Sex Marital Ther",
  ["journal of sex and marital therapy"] = "J. Sex Marital Ther.",
  ["journal of sex education and therapy"] = "J Sex Educ Ther",
  ["journal of sex research"] = "J Sex Res",
  ["journal of sexual medicine"] = "J. Sex. Med.",
  ["journal of sexually transmitted diseases"] = "J Sex Transm Dis",
  ["journal of shandong university. natural science. shandong daxue xuebao. lixue ban"] = "J. Shandong Univ. Nat. Sci.",
  ["journal of shanghai jiaotong university"] = "J. Shanghai Jiaotong Univ. (Chin. Ed.)",
  ["journal of shanghai jiaotong university (science)"] = "J Shanghai Jiaotong Univ Sci",
  ["journal of shanghai jiaotong university. chinese edition. shanghai jiaotong daxue xuebao"] = "J. Shanghai Jiaotong Univ. (Chin. Ed.)",
  ["journal of shanghai university"] = "J Shanghai Univ",
  ["journal of shanghai university. natural science edition. shanghai daxue xuebao. ziran kexue ban"] = "J. Shanghai Univ. Nat. Sci.",
  ["journal of shellfish research"] = "J Shellfish Res",
  ["journal of ship production and design"] = "J. Ship. Prod. Des.",
  ["journal of ship research"] = "J. Ship Res.",
  ["journal of shoulder and elbow arthroplasty"] = "J Shoulder Elb Arthroplast",
  ["journal of shoulder and elbow surgery"] = "J Shoulder Elbow Surg",
  ["journal of sichuan union university"] = "J. Sichuan Union Univ. Eng. Sci. Ed.",
  ["journal of sichuan university. engineering science edition. sichuan daxue xuebao. gongcheng kexue ban"] = "J. Sichuan Univ. Eng. Sci. Ed.",
  ["journal of sichuan university. natural science edition. sichuan daxue xuebao. ziran kexue ban"] = "Sichuan Daxue Xuebao",
  ["journal of sickle cell disease and hemoglobinopathies"] = "J Sick Cell Dis Hemoglobinopathies",
  ["journal of signal processing systems"] = "J. Signal Process. Syst.",
  ["journal of signal transduction"] = "J Signal Transduct",
  ["journal of simulation"] = "J. Simul.",
  ["journal of simulation : jos"] = "J Simul",
  ["journal of singing : the official journal of the national association of teachers of singing"] = "J Sing",
  ["journal of singularities"] = "J. Singul.",
  ["journal of skin cancer"] = "J Skin Cancer",
  ["journal of skyscape archaeology"] = "J. Skyscape Archaeol.",
  ["journal of sleep disorders & therapy"] = "J Sleep Disord Ther",
  ["journal of sleep disorders and management"] = "J Sleep Disord Manag",
  ["journal of sleep disorders-- treatment & care"] = "J Sleep Disord Treat Care",
  ["journal of sleep medicine and disorders"] = "J Sleep Med Disord",
  ["journal of sleep research"] = "J Sleep Res",
  ["journal of small animal practice"] = "J. Small Anim. Pract.",
  ["journal of smart cities and society"] = "J. Smart Cities Soc.",
  ["journal of smoking cessation"] = "J Smok Cessat",
  ["journal of smooth muscle research"] = "J. Smooth Muscle Res.",
  ["journal of smooth muscle research (hirosaki-shi"] = "J Smooth Muscle Res",
  ["journal of smooth muscle research = nihon heikatsukin gakkai kikanshi"] = "J Smooth Muscle Res",
  ["journal of social and administrative pharmacy : jsap"] = "J Soc Adm Pharm",
  ["journal of social and biological structures"] = "J Soc Biol Struct",
  ["journal of social and clinical psychology"] = "J Soc Clin Psychol",
  ["journal of social and economic development"] = "J. Soc. Econ. Devel.",
  ["journal of social and economic studies"] = "J Soc Econ Stud",
  ["journal of social and personal relationships"] = "J Soc Pers Relat",
  ["journal of social archaeology"] = "J. Social Archaeol.",
  ["journal of social behavior and personality"] = "J Soc Behav Pers",
  ["journal of social casework"] = "J Soc Casework",
  ["journal of social computing"] = "J. Social Comput.",
  ["journal of social development in africa"] = "J Soc Dev Afr",
  ["journal of social distress and the homeless"] = "J Soc Distress Homeless",
  ["journal of social history"] = "J Soc Hist",
  ["journal of social hygiene"] = "J Soc Hyg",
  ["journal of social marketing"] = "J Soc Mark",
  ["journal of social philosophy"] = "J Soc Philos",
  ["journal of social policy"] = "J Soc Policy",
  ["journal of social psychology"] = "J. Soc. Psychol.",
  ["journal of social service research"] = "J Soc Serv Res",
  ["journal of social structure : joss"] = "J Soc Struct",
  ["journal of social work & human sexuality"] = "J Soc Work Hum Sex",
  ["journal of social work (london, england)"] = "J Soc Work (Lond)",
  ["journal of social work education"] = "J Soc Work Educ",
  ["journal of social work in disability & rehabilitation"] = "J Soc Work Disabil Rehabil",
  ["journal of social work in end-of-life & palliative care"] = "J Soc Work End Life Palliat Care",
  ["journal of social work in the global community"] = "J Soc Work Glob Community",
  ["journal of social work practice"] = "J Soc Work Pract",
  ["journal of social work practice in the addictions"] = "J Soc Work Pract Addict",
  ["journal of social work values and ethics"] = "J Soc Work Values Ethics",
  ["journal of social, behavioral and health sciences"] = "J Soc Behav Health Sci",
  ["journal of social, evolutionary & cultural psychology : jsec"] = "J Soc Evol Cult Psychol",
  ["journal of societal & social policy"] = "J Soc Soc Policy",
  ["journal of socio-economics"] = "J. Socio-Econ.",
  ["journal of sociology"] = "J. Sociol.",
  ["journal of sociology (melbourne, vic.)"] = "J Sociol (Melb)",
  ["journal of sociology and social welfare"] = "J Sociol Soc Welf",
  ["journal of software (malden, ma)"] = "J Softw (Malden)",
  ["journal of software engineering research and development"] = "J. Software Eng. Res. Dev.",
  ["journal of software for algebra and geometry"] = "J. Softw. Algebra Geom.",
  ["journal of software. ruanjian xuebao"] = "J. Softw.",
  ["journal of software: evolution and process"] = "J. Software: Evol. Process",
  ["journal of soil and water conservation"] = "J. Soil Water Conserv.",
  ["journal of soil science"] = "J. Soil Sci.",
  ["journal of soil science and plant nutrition"] = "J Soil Sci Plant Nutr",
  ["journal of soil science, the"] = "J. Soil Sci.",
  ["journal of soils and sediments"] = "J. Soils Sediments",
  ["journal of sol-gel science and technology"] = "J. Sol-Gel Sci. Technol.",
  ["journal of solar energy engineering"] = "J. Sol. Energy Eng.",
  ["journal of solid state chemistry"] = "J. Solid State Chem.",
  ["journal of solid state electrochemistry"] = "J. Solid State Electrochem.",
  ["journal of solid state electrochemistry : current research and development in science and technology"] = "J Solid State Electrochem",
  ["journal of solid tumors"] = "J Solid Tumors",
  ["journal of solution chemistry"] = "J. Solution Chem.",
  ["journal of sound and vibration"] = "J. Sound Vib.",
  ["journal of south american earth sciences"] = "J. South Amer. Earth Sci.",
  ["journal of south asian and middle eastern studies"] = "J South Asian Middle East Stud",
  ["journal of south china normal university"] = "J. South China Normal Univ. Natur. Sci. Ed.",
  ["journal of south china normal university. natural science edition. huanan shifan daxue xuebao. ziran kexue ban"] = "J. South China Normal Univ. Natur. Sci. Ed.",
  ["journal of southeast asian american education & advancement : jsaaea"] = "J Southeast Asian Am Educ Adv",
  ["journal of southeast asian earth sciences"] = "J. Southeast Asian Earth Sci.",
  ["journal of southeast asian studies"] = "J Southeast Asian Stud",
  ["journal of southeast university"] = "J. Southeast Univ.",
  ["journal of southeast university. english edition. dongnan daxue xuebao. yingwen ban"] = "J. Southeast Univ. (English Ed.)",
  ["journal of southeast university. natural science edition. dongnan daxue xuebao. ziran kexue ban"] = "J. Southeast Univ. Nat. Sci.",
  ["journal of southern african studies"] = "J South Afr Stud",
  ["journal of southern hemisphere earth systems science"] = "J. South. Hemisph. Earth Syst. Sci.",
  ["journal of southwest university (natural science edition)"] = "J of Southwest Univ",
  ["journal of soviet nationalities : a quarterly publication of the center on east-west trade, investment, and communications"] = "J Sov Natly",
  ["journal of space safety engineering"] = "J. Space Saf. Eng.",
  ["journal of space weather and space climate"] = "J. Space Weather Space Clim.",
  ["journal of spacecraft and rockets"] = "J Spacecr Rockets",
  ["journal of spatial econometrics"] = "J. Spatial Econom.",
  ["journal of spatial science"] = "J Spat Sci",
  ["journal of special education technology : a publication of utah state university, the association for special education technology, and the technology and media division of the council for exceptional children"] = "J Spec Educ Technol",
  ["journal of special operations medicine"] = "J. Special Oper. Med.",
  ["journal of special operations medicine : a peer reviewed journal for sof medical professionals"] = "J Spec Oper Med",
  ["journal of species research"] = "J Species Res",
  ["journal of spectral theory"] = "J. Spectral Theory",
  ["journal of spectroscopy"] = "J. Spectro.",
  ["journal of spectroscopy (hindawi)"] = "J Spectrosc (Hindawi)",
  ["journal of spectroscopy (new york, ny, united states)"] = "J. Spectrosc. (N. Y., NY, U. S.)",
  ["journal of spectroscopy and dynamics"] = "J Spectrosc Dyn",
  ["journal of speech and hearing disorders"] = "J. Speech Hear. Disord.",
  ["journal of speech and hearing disorders. monograph supplement"] = "J Speech Hear Disord Monogr Suppl",
  ["journal of speech and hearing research"] = "J Speech Hear Res",
  ["journal of speech pathology & therapy"] = "J Speech Pathol Ther",
  ["journal of speech, language, and hearing research"] = "J. Speech Lang. Hear. Res.",
  ["journal of speech, language, and hearing research : jslhr"] = "J Speech Lang Hear Res",
  ["journal of spinal cord medicine"] = "J. Spinal Cord Med.",
  ["journal of spinal disorders"] = "J Spinal Disord",
  ["journal of spinal disorders & techniques"] = "J Spinal Disord Tech",
  ["journal of spinal disorders and techniques"] = "J. Spinal Disord. Tech.",
  ["journal of spine"] = "J Spine",
  ["journal of spine & neurosurgery"] = "J Spine Neurosurg",
  ["journal of spine surgery (hong kong)"] = "J Spine Surg",
  ["journal of spirituality in mental health"] = "J Spiritual Ment Health",
  ["journal of sport & exercise psychology"] = "J Sport Exerc Psychol",
  ["journal of sport & social issues"] = "J. Sport Soc. Issues",
  ["journal of sport and health science"] = "J Sport Health Sci",
  ["journal of sport and social issues"] = "J Sport Soc Issues",
  ["journal of sport behavior"] = "J Sport Behav",
  ["journal of sport for development"] = "J Sport Dev",
  ["journal of sport history"] = "J Sport Hist",
  ["journal of sport psychology in action"] = "J Sport Psychol Action",
  ["journal of sport rehabilitation"] = "J Sport Rehabil",
  ["journal of sports analytics"] = "J. Sports Anal.",
  ["journal of sports economics"] = "J Sports Econom",
  ["journal of sports medicine"] = "J. Sports Med.",
  ["journal of sports medicine & doping studies"] = "J Sports Med Doping Stud",
  ["journal of sports medicine (hindawi publishing corporation)"] = "J Sports Med (Hindawi Publ Corp)",
  ["journal of sports medicine and physical fitness"] = "J. Sports Med. Phys. Fitness",
  ["journal of sports science & medicine"] = "J Sports Sci Med",
  ["journal of sports sciences"] = "J Sports Sci",
  ["journal of state government"] = "J. State Gov.",
  ["journal of statistical computation and simulation"] = "J. Stat. Comput. Simul.",
  ["journal of statistical distributions and applications"] = "J Stat Distrib Appl",
  ["journal of statistical mechanics (online)"] = "J Stat Mech",
  ["journal of statistical mechanics: theory and experiment"] = "J. Stat. Mech: Theory Exp.",
  ["journal of statistical physics"] = "J. Stat. Phys.",
  ["journal of statistical planning and inference"] = "J Stat Plan Inference",
  ["journal of statistical research"] = "J. Statist. Res.",
  ["journal of statistical science and application"] = "J Stat Sci Appl",
  ["journal of statistical software"] = "J Stat Softw",
  ["journal of statistical studies"] = "J. Stat. Stud.",
  ["journal of statistical theory and applications"] = "J. Stat. Theory Appl.",
  ["journal of statistical theory and applications : jsta"] = "J Stat Theory Appl",
  ["journal of statistical theory and practice"] = "J. Stat. Theory Pract.",
  ["journal of statistics & management systems"] = "J. Stat. Manag. Syst.",
  ["journal of statistics and data science education"] = "J. Stat. Data Sci. Educ.",
  ["journal of statistics and management systems"] = "J. Stat. Manage. Syst",
  ["journal of statistics applications & probability"] = "J Stat Appl Probab",
  ["journal of statistics education : an international journal on the teaching and learning of statistics"] = "J Stat Educ",
  ["journal of stem cell and regenerative biology"] = "J Stem Cell Regen Biol",
  ["journal of stem cell and transplantation biology"] = "J Stem Cell Transplant Biol",
  ["journal of stem cell research"] = "J Stem Cell Res (Overl Park)",
  ["journal of stem cell research & therapeutics"] = "J Stem Cell Res Ther (Edmond)",
  ["journal of stem cell research & therapy"] = "J Stem Cell Res Ther",
  ["journal of stem cell research and transplantation"] = "J Stem Cell Res Transplant",
  ["journal of stem cell therapy and transplantation"] = "J Stem Cell Ther Transplant",
  ["journal of stem cells"] = "J Stem Cells",
  ["journal of stem cells & regenerative medicine"] = "J Stem Cells Regen Med",
  ["journal of stem education : innovations and research"] = "J STEM  Educ",
  ["journal of stem outreach"] = "J STEM Outreach",
  ["journal of sterile services management"] = "J Sterile Serv Manage",
  ["journal of steroid biochemistry"] = "J. Steroid Biochem.",
  ["journal of steroid biochemistry and molecular biology"] = "J. Steroid Biochem. Mol. Biol.",
  ["journal of steroids & hormonal science"] = "J Steroids Horm Sci",
  ["journal of stochastic analysis. josa"] = "J. Stoch. Anal.",
  ["journal of stomatology, oral and maxillofacial surgery"] = "J Stomatol Oral Maxillofac Surg",
  ["journal of stone disease"] = "J. Stone Dis.",
  ["journal of stored products research"] = "J. Stored Prod. Res.",
  ["journal of strain analysis for engineering design"] = "J. Strain Anal. Eng. Des.",
  ["journal of strategic information systems"] = "J. Strategic Inf. Syst.",
  ["journal of strength & conditioning research"] = "J. Strength Cond. Res.",
  ["journal of strength and conditioning research"] = "J Strength Cond Res",
  ["journal of strength and conditioning research / national strength & conditioning association"] = "J Strength Cond Res",
  ["journal of stroke"] = "J Stroke",
  ["journal of stroke and cerebrovascular diseases : the official journal of national stroke association"] = "J Stroke Cerebrovasc Dis",
  ["journal of structural and functional genomics"] = "J Struct Funct Genomics",
  ["journal of structural biology"] = "J Struct Biol",
  ["journal of structural biology: x"] = "J Struct Biol X",
  ["journal of structural chemistry"] = "J. Struct. Chem.",
  ["journal of structural control"] = "J. Struct. Control",
  ["journal of structural design"] = "J. Struct. Des.",
  ["journal of structural engineering"] = "J. Struct. Eng.",
  ["journal of structural engineering (new york, n.y.)"] = "J Struct Eng (N Y N Y)",
  ["journal of structural fire engineering"] = "J. Struct. Fire Eng.",
  ["journal of structural geology"] = "J. Struct. Geol.",
  ["journal of structural heart disease"] = "J Struct Heart Dis",
  ["journal of structural integrity and maintenance"] = "J. Struct. Integrity Maint.",
  ["journal of student affairs research and practice"] = "J Stud Aff Res Pract",
  ["journal of student physical therapy research"] = "J Stud Phys Ther Res",
  ["journal of studies in international education"] = "J Stud Int Educ",
  ["journal of studies in social sciences"] = "J Stud Soc Sci",
  ["journal of studies on alcohol"] = "J Stud Alcohol",
  ["journal of studies on alcohol and drugs"] = "J Stud Alcohol Drugs",
  ["journal of studies on alcohol and drugs. supplement"] = "J Stud Alcohol Drugs Suppl",
  ["journal of studies on alcohol. supplement"] = "J Stud Alcohol Suppl",
  ["journal of submicroscopic cytology"] = "J Submicrosc Cytol",
  ["journal of submicroscopic cytology and pathology"] = "J Submicrosc Cytol Pathol",
  ["journal of substance abuse"] = "J Subst Abuse",
  ["journal of substance abuse and alcoholism"] = "J Subst Abus Alcohol",
  ["journal of substance abuse treatment"] = "J Subst Abuse Treat",
  ["journal of substance use"] = "J Subst Use",
  ["journal of sugar beet research"] = "J Sugar Beet Res",
  ["journal of sulfur chemistry"] = "J. Sulfur Chem.",
  ["journal of sulphur chemistry"] = "J Sulphur Chem",
  ["journal of supercomputing"] = "J. Supercomput.",
  ["journal of superconductivity"] = "J. Supercond.",
  ["journal of superconductivity and novel magnetism"] = "J. Supercond. Novel Magn.",
  ["journal of supercritical fluids"] = "J. Supercrit. Fluids",
  ["journal of superhard materials"] = "J. Superhard Mater.",
  ["journal of supportive oncology"] = "J. Support. Oncol.",
  ["journal of supramolecular chemistry"] = "J. Supramol. Chem.",
  ["journal of supramolecular structure"] = "J Supramol Struct",
  ["journal of supramolecular structure and cellular biochemistry"] = "J Supramol Struct Cell Biochem",
  ["journal of supramolecular structure and cellular biochemistry. supplement"] = "J Supramol Struct Cell Biochem Suppl",
  ["journal of supramolecular structure. supplement"] = "J Supramol Struct Suppl",
  ["journal of surface analysis (online)"] = "J Surf Anal",
  ["journal of surface engineered materials and advanced technology"] = "J Surf Eng Mater Adv Technol",
  ["journal of surface investigation: x-ray, synchrotron and neutron techniques"] = "J. Surf. Invest.",
  ["journal of surfactants and detergents"] = "J. Surfactants Deterg.",
  ["journal of surgery (lisle, il)"] = "J Surg (Lisle)",
  ["journal of surgery (new york, n.y.)"] = "J Surg (N Y N Y)",
  ["journal of surgery (northborough, mass.)"] = "J Surg (Northborough)",
  ["journal of surgery and emergency medicine"] = "J Surg Emerg Med",
  ["journal of surgery and research"] = "J Surg Res (Houst)",
  ["journal of surgery and science"] = "J Surg Sci",
  ["journal of surgical case reports"] = "J Surg Case Rep",
  ["journal of surgical education"] = "J Surg Educ",
  ["journal of surgical oncology"] = "J Surg Oncol",
  ["journal of surgical oncology (tallinn, estonia)"] = "J Surg Oncol (Tallinn)",
  ["journal of surgical oncology. supplement"] = "J Surg Oncol Suppl",
  ["journal of surgical orthopaedic advances"] = "J Surg Orthop Adv",
  ["journal of surgical radiology"] = "J Surg Radiol",
  ["journal of surgical research"] = "J. Surg. Res.",
  ["journal of surgical technique and case report"] = "J Surg Tech Case Rep",
  ["journal of surgical urology"] = "J Surg Urol",
  ["journal of survey in fisheries sciences"] = "J Surv Fish Sci",
  ["journal of survey statistics and methodology"] = "J Surv Stat Methodol",
  ["journal of surveying engineering"] = "J. Surv. Eng.",
  ["journal of sustainability research"] = "J. Sustainability Res.",
  ["journal of sustainability science and management"] = "J Sustain Sci Manag",
  ["journal of sustainable agriculture and environment"] = "J. Sustainable Agric. Environ.",
  ["journal of sustainable architecture and civil engineering"] = "J. Sustainable Archit. Civ. Eng.",
  ["journal of sustainable bioenergy systems"] = "J Sustain Bioenergy Syst",
  ["journal of sustainable cement-based materials"] = "J. Sustainable Cem.-Based Mater.",
  ["journal of sustainable development"] = "J Sustain Dev",
  ["journal of sustainable development of energy, water and environment systems"] = "J. Sustainable Dev. Energy Water Environ. Syst.",
  ["journal of sustainable energy engineering"] = "J. Sustainable Energy Eng.",
  ["journal of sustainable forestry"] = "J. Sustainable For.",
  ["journal of sustainable materials science and engineering"] = "J. Sustainable Mater. Sci. Eng.",
  ["journal of sustainable metallurgy"] = "J. Sustainable Metall.",
  ["journal of sustainable mining"] = "J. Sustainable Min.",
  ["journal of sustainable product design"] = "J. Sustainable Prod. Des.",
  ["journal of sustainable science and management"] = "J. Sustainable Sci. Manage.",
  ["journal of sustainable tourism"] = "J. Sustainable Tourism",
  ["journal of sustainable water in the built environment"] = "J Sustain Water Built Environ",
  ["journal of symbolic computation"] = "J. Symbolic Comput.",
  ["journal of symbolic logic"] = "J. Symb. Log.",
  ["journal of symplectic geometry"] = "J. Symplectic Geom.",
  ["journal of symptoms and signs"] = "J Symptoms Signs",
  ["journal of synchrotron radiat."] = "J. Synchrotron Radiat.",
  ["journal of synchrotron radiation"] = "J Synchrotron Radiat",
  ["journal of syndromes"] = "J Syndr",
  ["journal of synthetic lubrication"] = "J. Synth. Lubr.",
  ["journal of synthetic organic chemistry, japan"] = "J. Synth. Org. Chem Jpn.",
  ["journal of systematic palaeontology"] = "J. Syst. Paleontol.",
  ["journal of systematics and evolution"] = "J Syst Evol",
  ["journal of systematics evolution"] = "J. Syst. Evol.",
  ["journal of systemics, cybernetics and informatics"] = "J Syst Cybern Inf",
  ["journal of systems and integrative neuroscience"] = "J Syst Integr Neurosci",
  ["journal of systems and software"] = "J. Syst. Software",
  ["journal of systems architecture"] = "J. Syst. Archit.",
  ["journal of systems chemistry"] = "J. Syst. Chem.",
  ["journal of systems engineering"] = "J. Syst. Eng.",
  ["journal of systems engineering and electronics"] = "J. Syst. Eng. Electron.",
  ["journal of systems integration"] = "J. Syst. Integr.",
  ["journal of systems science & complexity"] = "J. Syst. Sci. Complex.",
  ["journal of systems science and complexity"] = "J Syst Sci Complex",
  ["journal of systems science and mathematical sciences"] = "J. Systems Sci. Math. Sci.",
  ["journal of systems science and mathematical sciences. xitong kexue yu shuxue"] = "J. Systems Sci. Math. Sci.",
  ["journal of systems science and systems engineering"] = "J. Syst. Sci. Syst. Eng.",
  ["journal of taibah university for science"] = "J. Taibah Univ. Sci.",
  ["journal of taibah university for science : jtusci"] = "J Taibah Univ Sci",
  ["journal of taibah university medical sciences"] = "J Taibah Univ Med Sci",
  ["journal of taxation"] = "J. Taxation",
  ["journal of teacher action research"] = "J Teach Action Res",
  ["journal of teacher education"] = "J Teach Educ",
  ["journal of teaching and research in chemistry"] = "J. Teach. Res. Chem.",
  ["journal of teaching in physical education : jtpe"] = "J Teach Phys Educ",
  ["journal of teaching in social work"] = "J Teach Soc Work",
  ["journal of technical physics"] = "J. Tech. Phys.",
  ["journal of technology and science education"] = "J Technol Sci Educ",
  ["journal of technology computer aided design tcad"] = "J. Technol. Comput. Aided Des. TCAD",
  ["journal of technology in behavioral science"] = "J Technol Behav Sci",
  ["journal of technology in human services"] = "J Technol Hum Serv",
  ["journal of technology transfer"] = "J. Technol. Transfer",
  ["journal of telemedicine and telecare"] = "J Telemed Telecare",
  ["journal of terramechanics"] = "J. Terramech.",
  ["journal of terrorism research"] = "J Terror Res",
  ["journal of testing and evaluation"] = "J. Test. Eval.",
  ["journal of textile design research and practice"] = "J. Text. Des. Res. Pract.",
  ["journal of textile science & engineering"] = "J Text Sci Eng",
  ["journal of textiles and fibrous materials"] = "J. Text. Fibrous Mater.",
  ["journal of texture studies"] = "J Texture Stud",
  ["journal of thai association for voluntary sterilization"] = "J Thai Assoc Volunt Steriliz",
  ["journal of thanatology"] = "J Thanatol",
  ["journal of the academy of consultation-liaison psychiatry"] = "J Acad Consult Liaison Psychiatry",
  ["journal of the academy of marketing science"] = "J Acad Mark Sci",
  ["journal of the academy of nutrition and dietetics"] = "J Acad Nutr Diet",
  ["journal of the academy of rehabilitative audiology"] = "J Acad Rehabil Audiol",
  ["journal of the acm"] = "J. ACM",
  ["journal of the acoustical society of america"] = "J. Acoust. Soc. Am.",
  ["journal of the adelaide botanic gardens"] = "J Adel Bot Gard",
  ["journal of the advanced practitioner in oncology"] = "J Adv Pract Oncol",
  ["journal of the aeronautical sciences"] = "J. Aeronaut. Sci.",
  ["journal of the afro-american historical and genealogical society. afro-american historical and genealogical society (washington, d.c.)"] = "J Afro Am Hist Geneal Soc",
  ["journal of the air & waste management association"] = "J Air Waste Manage Assoc",
  ["journal of the air & waste management association (1995)"] = "J Air Waste Manag Assoc",
  ["journal of the air and waste management association"] = "J. Air Waste Manage. Assoc.",
  ["journal of the air pollution control association"] = "J Air Pollut Control Assoc",
  ["journal of the air transport research society"] = "J. Air Transp. Res. Soc.",
  ["journal of the all india dental association"] = "J. All India Dent. Assoc.",
  ["journal of the all-india ophthalmological society"] = "J All India Ophthalmol Soc",
  ["journal of the american academy of applied nutrition. american academy of applied nutrition"] = "J Am Acad Appl Nutr",
  ["journal of the american academy of audiology"] = "J Am Acad Audiol",
  ["journal of the american academy of audiology (hamilton, ont."] = "J Am Acad Audiol",
  ["journal of the american academy of child and adolescent psychiatry"] = "J Am Acad Child Adolesc Psychiatry",
  ["journal of the american academy of child psychiatry"] = "J Am Acad Child Psychiatry",
  ["journal of the american academy of dermatology"] = "J Am Acad Dermatol",
  ["journal of the american academy of gold foil operators"] = "J. Am. Acad. Gold Foil Oper.",
  ["journal of the american academy of nurse practitioners"] = "J Am Acad Nurse Pract",
  ["journal of the american academy of orthopaedic surgeons"] = "J. Am. Acad. Orthop. Surg.",
  ["journal of the american academy of orthopaedic surgeons. global research & reviews"] = "J Am Acad Orthop Surg Glob Res Rev",
  ["journal of the american academy of physician assistants"] = "J Am Acad Physician Assist",
  ["journal of the american academy of psychiatry and the law"] = "J. Am. Acad. Psychiatry Law",
  ["journal of the american academy of psychoanalysis"] = "J. Am. Acad. Psychoanal.",
  ["journal of the american academy of psychoanalysis and dynamic psychiatry"] = "J. Am. Acad. Psychoanal. Dyn. Psychiatry",
  ["journal of the american academy of religion. american academy of religion"] = "J Am Acad Relig",
  ["journal of the american academy of special education professionals"] = "J Am Acad Spec Educ Prof",
  ["journal of the american aging association"] = "J Am Aging Assoc",
  ["journal of the american animal hospital association"] = "J Am Anim Hosp Assoc",
  ["journal of the american association for laboratory animal science"] = "J. Am. Assoc. Lab. Anim. Sci.",
  ["journal of the american association for laboratory animal science : jaalas"] = "J Am Assoc Lab Anim Sci",
  ["journal of the american association of gynecologic laparoscopists"] = "J. Am. Assoc. Gynecol. Laparosc.",
  ["journal of the american association of instructors and investigators of poultry husbandry"] = "J. Am. Assoc. Instr. Invest. Poult. Husb.",
  ["journal of the american association of nephrology nurses and technicians"] = "J. Am. Assoc. Nephrol. Nurses Tech.",
  ["journal of the american association of nurse anesthetists"] = "J. Am. Assoc. Nurse Anesth.",
  ["journal of the american association of nurse practitioners"] = "J Am Assoc Nurse Pract",
  ["journal of the american association of variable star observers"] = "JAAVSO",
  ["journal of the american audiology society"] = "J Am Audiol Soc",
  ["journal of the american auditory society"] = "J Am Aud Soc",
  ["journal of the american board of family medicine"] = "J. Am. Board Fam. Med.",
  ["journal of the american board of family medicine : jabfm"] = "J Am Board Fam Med",
  ["journal of the american board of family practice"] = "J. Am. Board Fam. Pract.",
  ["journal of the american ceramic society"] = "J. Am. Ceram. Soc.",
  ["journal of the american ceramic society. american ceramic society"] = "J Am Ceram Soc",
  ["journal of the american chemical society"] = "J Am Chem Soc",
  ["journal of the american college health association"] = "J Am Coll Health Assoc",
  ["journal of the american college of cardiology"] = "J Am Coll Cardiol",
  ["journal of the american college of clinical pharmacy"] = "J. Am. Coll. Clin. Pharm.",
  ["journal of the american college of clinical pharmacy : jaccp"] = "J Am Coll Clin Pharm",
  ["journal of the american college of dentists"] = "J. Am. Coll. Dent.",
  ["journal of the american college of emergency physicians open"] = "J Am Coll Emerg Physicians Open",
  ["journal of the american college of neuropsychiatrists"] = "J Am Coll Neuropsychiatr",
  ["journal of the american college of nutrition"] = "J Am Coll Nutr",
  ["journal of the american college of radiology"] = "J. Am. Coll. Radiol.",
  ["journal of the american college of radiology : jacr"] = "J Am Coll Radiol",
  ["journal of the american college of surgeons"] = "J Am Coll Surg",
  ["journal of the american college of toxicology"] = "J Am Coll Toxicol",
  ["journal of the american dental association"] = "J. Am. Dent. Assoc.",
  ["journal of the american dental association (1939)"] = "J Am Dent Assoc",
  ["journal of the american dental association (edizione italiana)"] = "J Am Dent Assoc (Ed Ital)",
  ["journal of the american dental hygienists association"] = "J. Am. Dent. Hyg. Assoc.",
  ["journal of the american dental society of anesthesiology"] = "J Am Dent Soc Anesthesiol",
  ["journal of the american dietetic association"] = "J Am Diet Assoc",
  ["journal of the american geriatrics society"] = "J Am Geriatr Soc",
  ["journal of the american heart association"] = "J Am Heart Assoc",
  ["journal of the american helicopter society. american helicopter society"] = "J Am Helicopter Soc",
  ["journal of the american historical society of germans from russia. american historical society of germans from russia"] = "J Am Hist Soc Ger Russ",
  ["journal of the american institute for conservation"] = "J. Am. Inst. Conserv.",
  ["journal of the american institute of electrical engineers"] = "J. Am. Inst. Electr. Eng.",
  ["journal of the american institute of homeopathy"] = "J Am Inst Homeopath",
  ["journal of the american institute of planners"] = "J Am Inst Plann",
  ["journal of the american leather chemists association"] = "J. Am. Leather Chem. Assoc.",
  ["journal of the american mathematical society"] = "J. Amer. Math. Soc.",
  ["journal of the american medical association"] = "J Am Med Assoc",
  ["journal of the american medical directors association"] = "J Am Med Dir Assoc",
  ["journal of the american medical informatics association"] = "J. Am. Med. Inf. Assoc.",
  ["journal of the american medical informatics association : jamia"] = "J Am Med Inform Assoc",
  ["journal of the american medical record association"] = "J Am Med Rec Assoc",
  ["journal of the american medical technologists"] = "J Am Med Technol",
  ["journal of the american medical women's association"] = "J Am Med Womens Assoc",
  ["journal of the american medical women's association (1972)"] = "J Am Med Womens Assoc (1972)",
  ["journal of the american medical womens association"] = "J. Am. Med. Wom. Assoc.",
  ["journal of the american medical women’s association"] = "J Am Med Womens Assoc",
  ["journal of the american medical women’s association (1972)"] = "J Am Med Womens Assoc",
  ["journal of the american mosquito control association"] = "J Am Mosq Control Assoc",
  ["journal of the american mosquito control association. supplement"] = "J Am Mosq Control Assoc Suppl",
  ["journal of the american nutrition association"] = "J. Am. Nutr. Assoc.",
  ["journal of the american oil chemists society"] = "J. Am. Oil Chem. Soc.",
  ["journal of the american oil chemists' society"] = "J Am Oil Chem Soc",
  ["journal of the american oil chemists’ society"] = "J. Am. Oil Chem. Soc.",
  ["journal of the american optometric association"] = "J Am Optom Assoc",
  ["journal of the american oriental society"] = "JAOS",
  ["journal of the american oriental society. american oriental society"] = "J Am Orient Soc",
  ["journal of the american osteopathic association"] = "J. Am. Osteopath. Assoc.",
  ["journal of the american osteopathic college of radiology"] = "J Am Osteopath Coll Radiol",
  ["journal of the american paraplegia society"] = "J. Am. Paraplegia Soc.",
  ["journal of the american pharmaceutical association"] = "J Am Pharm Assoc",
  ["journal of the american pharmaceutical association (washington, d.c. : 1996)"] = "J Am Pharm Assoc (Wash)",
  ["journal of the american pharmaceutical association (washington, d.c.)"] = "J. Am. Pharm. Assoc. (Wash.)",
  ["journal of the american pharmaceutical association (washington,d.c. : 1996)"] = "J Am Pharm Assoc (Wash)",
  ["journal of the american pharmaceutical association. american pharmaceutical association"] = "J Am Pharm Assoc Am Pharm Assoc (Baltim)",
  ["journal of the american pharmacists association"] = "J. Am. Pharm. Assoc.",
  ["journal of the american pharmacists association : japha"] = "J Am Pharm Assoc (2003)",
  ["journal of the american physical therapy association"] = "J Am Phys Ther Assoc",
  ["journal of the american planning association"] = "J. Am. Plann. Assoc.",
  ["journal of the american planning association. american planning association"] = "J Am Plann Assoc",
  ["journal of the american podiatric medical association"] = "J Am Podiatr Med Assoc",
  ["journal of the american podiatry association"] = "J Am Podiatry Assoc",
  ["journal of the american pomological society"] = "J. Am. Pomol. Soc.",
  ["journal of the american psychiatric nurses association"] = "J Am Psychiatr Nurses Assoc",
  ["journal of the american psychoanalytic association"] = "J Am Psychoanal Assoc",
  ["journal of the american real estate and urban economics association"] = "J. Amer. Real Estate Urban Econ. Assoc.",
  ["journal of the american research center in egypt"] = "JARCE",
  ["journal of the american scientific affiliation. american scientific affiliation"] = "J Am Sci Affil",
  ["journal of the american society for geriatric dentistry"] = "J. Am. Soc. Geriatr. Dent.",
  ["journal of the american society for horticultural science. american society for horticultural science"] = "J Am Soc Hortic Sci",
  ["journal of the american society for information science"] = "J. Am. Soc. Inf. Sci.",
  ["journal of the american society for information science and technology : jasist"] = "J Am Soc Inf Sci Technol",
  ["journal of the american society for information science. american society for information science"] = "J Am Soc Inf Sci",
  ["journal of the american society for mass spectrometry"] = "J Am Soc Mass Spectrom",
  ["journal of the american society for preventive dentistry"] = "J. Am. Soc. Prev. Dent.",
  ["journal of the american society for the study of orthodontics. american society for the study of orthodontics"] = "J Am Soc Study Orthod",
  ["journal of the american society of brewing chemists"] = "J. Am. Soc. Brew. Chem.",
  ["journal of the american society of brewing chemists. american society of brewing chemists"] = "J Am Soc Brew Chem",
  ["journal of the american society of cytopathology"] = "J Am Soc Cytopathol",
  ["journal of the american society of echocardiography"] = "J. Am. Soc. Echocardiogr.",
  ["journal of the american society of echocardiography : official publication of the american society of echocardiography"] = "J Am Soc Echocardiogr",
  ["journal of the american society of hypertension : jash"] = "J Am Soc Hypertens",
  ["journal of the american society of naval engineers"] = "J. Am. Soc. Naval Eng.",
  ["journal of the american society of nephrology"] = "J. Am. Soc. Nephrol.",
  ["journal of the american society of nephrology : jasn"] = "J Am Soc Nephrol",
  ["journal of the american society of psychosomatic dentistry and medicine"] = "J. Am. Soc. Psychosom. Dent. Med.",
  ["journal of the american statistical association"] = "J Am Stat Assoc",
  ["journal of the american venereal disease association"] = "J Am Vener Dis Assoc",
  ["journal of the american veterinary medical association"] = "J Am Vet Med Assoc",
  ["journal of the american water resources association"] = "J. Am. Water Resour. Assoc.",
  ["journal of the anatomical society of india"] = "J Anat Soc India",
  ["journal of the anus, rectum and colon"] = "J Anus Rectum Colon",
  ["journal of the argentine chemical society"] = "J. Argent. Chem. Soc.",
  ["journal of the arizona-nevada academy of science"] = "J. Ariz.-Nev. Acad. Sci.",
  ["journal of the arkansas academy of science"] = "J Ark Acad Sci",
  ["journal of the arkansas medical society"] = "J. Ark. Med. Soc.",
  ["journal of the arnold arboretum. arnold arboretum"] = "J Arnold Arbor",
  ["journal of the asahikawa national college of technology"] = "J. Asahikawa Nat. College Tech.",
  ["journal of the asean federation of endocrine societies"] = "J ASEAN Fed Endocr Soc",
  ["journal of the asia pacific economy"] = "J Asia Pac Econ",
  ["journal of the asiatic society"] = "J Asiat Soc",
  ["journal of the asiatic society of bangladesh. humanities"] = "J Asiat Soc Bangladesh Humanit",
  ["journal of the assam academy of mathematics (jaam)"] = "J. Assam Acad. Math.",
  ["journal of the assam science society"] = "J. Assam Sci. Soc.",
  ["journal of the association for academic minority physicians"] = "J. Assoc. Acad. Minor. Phys.",
  ["journal of the association for academic minority physicians : the official publication of the association for academic minority physicians"] = "J Assoc Acad Minor Phys",
  ["journal of the association for consumer research"] = "J Assoc Consum Res",
  ["journal of the association for crime scene reconstruction"] = "J. Assoc. Crime Scene Reconstr.",
  ["journal of the association for information science and technology"] = "J Assoc Inf Sci Technol",
  ["journal of the association for information systems"] = "J Assoc Inf Syst",
  ["journal of the association for mathematical research"] = "J. Assoc. Math. Res.",
  ["journal of the association for physical and mental rehabilitation"] = "J Assoc Phys Ment Rehabil",
  ["journal of the association for research in otolaryngology"] = "J. Assoc. Res. Otolaryngol.",
  ["journal of the association for research in otolaryngology : jaro"] = "J Assoc Res Otolaryngol",
  ["journal of the association for the care of children in hospitals"] = "J Assoc Care Child Hosp",
  ["journal of the association for the care of children's health"] = "J Assoc Care Child Health",
  ["journal of the association for the care of children’s health"] = "J. Assoc. Care Child. Health",
  ["journal of the association of asphalt paving technologists"] = "J. Assoc. Asphalt Paving Technol.",
  ["journal of the association of environmental and resource economists"] = "J Assoc Environ Resour Econ",
  ["journal of the association of food and drug officials"] = "J Assoc Food Drug Off",
  ["journal of the association of genetic technologists"] = "J Assoc Genet Technol",
  ["journal of the association of medical illustrators"] = "J Assoc Med Illus",
  ["journal of the association of medical microbiology and infectious disease canada = journal officiel de l’association pour la microbiologie médicale et l’infectiologie canada"] = "J Assoc Med Microbiol Infect Dis Can",
  ["journal of the association of nurses in aids care"] = "J. Assoc. Nurses AIDS Care",
  ["journal of the association of official agricultural chemists"] = "J. Assoc. Off. Agric. Chem.",
  ["journal of the association of pediatric oncology nurses"] = "J Assoc Pediatr Oncol Nurses",
  ["journal of the association of physicians of india"] = "J. Assoc. Physicians India",
  ["journal of the association of university technology managers"] = "J Assoc Univ Technol Manag",
  ["journal of the astronautical sciences"] = "J. Astronaut. Sci.",
  ["journal of the atmospheric sciences"] = "J. Atmos. Sci.",
  ["journal of the audio engineering society"] = "J. Audio Eng. Soc.",
  ["journal of the audio engineering society. audio engineering society"] = "J Audio Eng Soc",
  ["journal of the australasian ceramic society"] = "J. Australas. Ceram. Soc.",
  ["journal of the australasian college of road safety"] = "J Australas Coll Road Saf",
  ["journal of the australasian universities language and literature association"] = "AUMLA",
  ["journal of the australian ceramic society"] = "J. Aust. Ceram. Soc.",
  ["journal of the australian institute of surgical and dental technicians (melbourne)"] = "J Aust Inst Surg Dent Tech",
  ["journal of the australian mathematical society"] = "J. Aust. Math. Soc.",
  ["journal of the australian population association"] = "J Aust Popul Assoc",
  ["journal of the autonomic nervous system"] = "J Auton Nerv Syst",
  ["journal of the bahrain medical society = majallat jam'īyat al-atibbā' al-bahraynīyah"] = "J Bahrain Med Soc",
  ["journal of the baltimore college of dental surgery"] = "J Baltimore Coll Dent Surg",
  ["journal of the belgian society of radiology"] = "J Belg Soc Radiol",
  ["journal of the bergen county dental society"] = "J. Bergen Cty. Dent. Soc.",
  ["journal of the biological photographic association"] = "J Biol Photogr Assoc",
  ["journal of the black sea/mediterranean environment"] = "J. Black Sea/Mediterr. Environ.",
  ["journal of the board of greenkeeping research"] = "J. Board Greenkeep. Res.",
  ["journal of the botanical research institute of texas"] = "J Bot Res Inst Tex",
  ["journal of the botanical society of bengal"] = "J Bot Soc Bengal",
  ["journal of the brazialian computer society"] = "J. Braz. Comput. Sci.",
  ["journal of the brazilian chemical society"] = "J. Braz. Chem. Soc.",
  ["journal of the brazilian computer society"] = "J. Braz. Comput. Soc.",
  ["journal of the brazilian society of mechanical sciences and engineering"] = "J. Braz. Soc. Mech. Sci. Eng.",
  ["journal of the british academy"] = "J Br Acad",
  ["journal of the british archaeological association"] = "J. Br. Archaeol. Assoc.",
  ["journal of the british endodontic society"] = "J Br Endod Soc",
  ["journal of the british fertility society"] = "J Br Fer Soc",
  ["journal of the british institution of radio engineers"] = "J. Br. Inst. Radio Eng.",
  ["journal of the british interplanetary society"] = "J Br Interplanet Soc",
  ["journal of the british waterworks association"] = "J. Br. Waterworks Assoc.",
  ["journal of the california dental association"] = "J Calif Dent Assoc",
  ["journal of the california state dental association"] = "J Calif State Dent Assoc",
  ["journal of the canadian academy of child and adolescent psychiatry"] = "J. Can. Acad. Child Adolesc. Psychiatry",
  ["journal of the canadian academy of child and adolescent psychiatry = journal de l'académie canadienne de psychiatrie de l'enfant et de l'adolescent"] = "J Can Acad Child Adolesc Psychiatry",
  ["journal of the canadian association of gastroenterology"] = "J Can Assoc Gastroenterol",
  ["journal of the canadian association of radiologists"] = "J Can Assoc Radiol",
  ["journal of the canadian dental association"] = "J Can Dent Assoc (Tor)",
  ["journal of the canadian dietetic association"] = "J Can Diet Assoc",
  ["journal of the canadian health libraries association"] = "J. Can. Health Lib. Assoc.",
  ["journal of the cardiometabolic syndrome"] = "J Cardiometab Syndr",
  ["journal of the catholic nurses guild of england and wales"] = "J Cathol Nurses Guild Engl Wales",
  ["journal of the ceramic society of japan"] = "J. Ceram. Soc. Jpn.",
  ["journal of the charles h. tweed international foundation"] = "J Charles H. Tweed Int Found",
  ["journal of the chemical society"] = "J. Chem. Soc.",
  ["journal of the chemical society a: inorganic, physical, theoretical"] = "J. Chem. Soc. A",
  ["journal of the chemical society b: physical organic"] = "J. Chem. Soc. B",
  ["journal of the chemical society c: organic"] = "J. Chem. Soc. C",
  ["journal of the chemical society d: chemical communications"] = "J. Chem. Soc. D",
  ["journal of the chemical society of pakistan"] = "J. Chem. Soc. Pak.",
  ["journal of the chemical society of pakistan. chemical society of pakistan"] = "J Chem Soc Pak",
  ["journal of the chemical society, abstracts"] = "J. Chem. Soc. Abstr.",
  ["journal of the chemical society, chemical communications"] = "J. Chem. Soc., Chem. Commun.",
  ["journal of the chemical society, dalton transactions"] = "J. Chem. Soc., Dalton Trans.",
  ["journal of the chemical society, faraday transactions"] = "J. Chem. Soc., Faraday Trans.",
  ["journal of the chemical society, faraday transactions 1: physical chemistry in condensed phases"] = "J. Chem. Soc., Faraday Trans. 1 F",
  ["journal of the chemical society, faraday transactions 2: molecular and chemical physics"] = "J. Chem. Soc., Faraday Trans. 2",
  ["journal of the chemical society, perkin transactions 1"] = "J. Chem. Soc., Perkin Trans. 1",
  ["journal of the chemical society, perkin transactions 2"] = "J. Chem. Soc., Perkin Trans. 2",
  ["journal of the chemical society, transactions"] = "J. chem. Soc. Trans.",
  ["journal of the chemical society. chemical communications"] = "J Chem Soc Chem Commun",
  ["journal of the chemical society. perkin transactions 1"] = "J Chem Soc Perkin 1",
  ["journal of the chester archaeological society. chester archaeological society"] = "J Chester Archaeol Soc",
  ["journal of the chilean chemical society"] = "J. Chil. Chem. Soc.",
  ["journal of the china university of geosciences"] = "J. China Univ. Geosci.",
  ["journal of the chinese advanced materials society"] = "J. Chin. Adv. Mater. Soc.",
  ["journal of the chinese chemical society"] = "J. Chin. Chem. Soc.",
  ["journal of the chinese chemical society (taipei, taiwan)"] = "J. Chin. Chem. Soc. (Taipei, Taiwan)",
  ["journal of the chinese chemical society (weinheim, germany)"] = "J. Chin. Chem. Soc. (Weinheim, Ger.)",
  ["journal of the chinese chemical society. zhongguo hua xue hui (taipei, taiwan)"] = "J Chin Chem Soc",
  ["journal of the chinese institute of chemical engineers"] = "J. Chin. Inst. Chem. Eng,",
  ["journal of the chinese institute of engineers"] = "J. Chin. Inst. Eng.",
  ["journal of the chinese medical association"] = "J. Chin. Med. Assoc.",
  ["journal of the chinese medical association : jcma"] = "J Chin Med Assoc",
  ["journal of the chungcheong mathematical society"] = "J. Chungcheong Math. Soc.",
  ["journal of the cmsc"] = "J CMSC",
  ["journal of the college of general practitioners"] = "J. Coll. Gen. Pract.",
  ["journal of the college of general practitioners and research newsletter. college of general practitioners (great britain)"] = "J Coll Gen Pract Res Newsl",
  ["journal of the college of physicians and surgeons--pakistan : jcpsp"] = "J Coll Physicians Surg Pak",
  ["journal of the college of radiologists of australasia"] = "J Coll Radiol Australas",
  ["journal of the colorado dental association"] = "J Colo Dent Assoc",
  ["journal of the combustion society of japan"] = "J Combust Soc Jpn",
  ["journal of the community development society"] = "J Community Dev Soc",
  ["journal of the czech geological society"] = "J. Czech Geol. Soc.",
  ["journal of the deep foundations institute"] = "J. Deep Found. Inst.",
  ["journal of the dental association of south africa"] = "J. Dent. Assoc. S. Afr.",
  ["journal of the dental association of thailand"] = "J. Dent. Assoc. Thai.",
  ["journal of the dental auxiliaries"] = "J. Dent. Aux.",
  ["journal of the dental guidance council on the handicapped"] = "J Dent Guid Counc Handicap",
  ["journal of the dental school, national university of iran"] = "J Dent Sch Natl Univ Iran",
  ["journal of the dermatology nurses' association"] = "J Dermatol Nurses Assoc",
  ["journal of the dow university of health sciences"] = "J Dow Univ Health Sci",
  ["journal of the early republic"] = "J Early Repub",
  ["journal of the earth and space physics"] = "J. Earth Space Phys.",
  ["journal of the economic and social history of the orient"] = "JESHO",
  ["journal of the economic science association"] = "J Econ Sci Assoc",
  ["journal of the economics of ageing"] = "J Econ Ageing",
  ["journal of the economics of business"] = "J. Econ. of Bus.",
  ["journal of the egyptian mathematical society"] = "J. Egyptian Math. Soc.",
  ["journal of the egyptian medical association"] = "J. Egypt. Med. Assoc.",
  ["journal of the egyptian national cancer institute"] = "J Egypt Natl Canc Inst",
  ["journal of the egyptian public health association"] = "J. Egypt. Public Health Assoc.",
  ["journal of the egyptian society of obstetrics and gynecology"] = "J Egypt Soc Obstet Gynecol",
  ["journal of the egyptian society of parasitology"] = "J Egypt Soc Parasitol",
  ["journal of the electrochemical society"] = "J. Electrochem. Soc.",
  ["journal of the electrochemical society of india"] = "J. Electrochem. Soc. India",
  ["journal of the elisha mitchell scientific society"] = "J. Elisha Mitchell Sci. Soc.",
  ["journal of the endocrine society"] = "J Endocr Soc",
  ["journal of the energy institute"] = "J. Energy Inst.",
  ["journal of the entomological research society"] = "J Entomol Res Soc",
  ["journal of the entomological society of british columbia"] = "J. Entomol. Soc. B. C.",
  ["journal of the entomological society of british columbia. entomological society of british columbia"] = "J Entomol Soc B C",
  ["journal of the entomological society of ontario"] = "J Entomol Soc Ont",
  ["journal of the entomological society of southern africa"] = "J Entomol Soc South Afr",
  ["journal of the european academy of dermatology and venereology"] = "J. Eur. Acad. Dermatol. Venereol.",
  ["journal of the european academy of dermatology and venereology : jeadv"] = "J Eur Acad Dermatol Venereol",
  ["journal of the european association for health information and libraries"] = "J Eur Assoc Health Inf Libr",
  ["journal of the european ceramic society"] = "J. Eur. Ceram. Soc.",
  ["journal of the european economic association"] = "J Eur Econ Assoc",
  ["journal of the european mathematical society"] = "J. Eur. Math. Soc.",
  ["journal of the european mathematical society (jems)"] = "J. Eur. Math. Soc. (JEMS)",
  ["journal of the european mosquito control association"] = "J Eur Mosq Control Assoc",
  ["journal of the european optical society part a"] = "J. Eur. Opt. Soc. Part A",
  ["journal of the european optical society part b"] = "J. Eur. Opt. Soc. Part B",
  ["journal of the european optical society-rapid publications"] = "J. Eur. Opt. Soc.-Rapid Publ.",
  ["journal of the european optical society. rapid publications"] = "J Eur Opt Soc Rapid Publ",
  ["journal of the evangelical theological society. evangelical theological society"] = "J Evang Theol Soc",
  ["journal of the experimental analysis of behavior"] = "J Exp Anal Behav",
  ["journal of the faculty of agriculture, hokkaido university"] = "Hokkaido Daigaku Nogakubu kiyo|J. Fac. Agric. Hokkaido Univ.",
  ["journal of the faculty of agriculture, hokkaido university = hokkaido daigaku nōgakubu kiyo"] = "J Fac Agric Hokkaido Univ",
  ["journal of the faculty of agriculture, shinshu university"] = "J. Fac. Agric., Shinshu Univ.",
  ["journal of the faculty of engineering"] = "J. Fac. Eng. Chiba Univ.",
  ["journal of the faculty of liberal arts"] = "J. Fac. Liberal Arts Yamaguchi Univ. Natur. Sci.",
  ["journal of the faculty of medicine, baghdad"] = "J Fac Med Baghdad",
  ["journal of the faculty of science"] = "J. Fac. Sci. Shinshu Univ.",
  ["journal of the faculty of science and engineering"] = "J. Fac. Sci. Eng. Kinki Univ.",
  ["journal of the faculty of science. shinshu university"] = "J. Fac. Sci. Shinshu Univ.",
  ["journal of the fisheries research board of canada"] = "J. Fish. Res. Board Can.",
  ["journal of the florida medical association"] = "J. Fla. Med. Assoc.",
  ["journal of the florida state dental society"] = "J. Fla. State Dent. Soc.",
  ["journal of the forensic science society"] = "J. Forensic Sci. Soc.",
  ["journal of the formosan medical association"] = "J. Formos. Med. Assoc.",
  ["journal of the formosan medical association = taiwan yi zhi"] = "J Formos Med Assoc",
  ["journal of the franklin institute"] = "J. Franklin Inst.",
  ["journal of the franklin institute-engineering and applied mathematics"] = "J. Franklin Inst-. Eng. Appl. Math.",
  ["journal of the franklin institute. engineering and applied mathematics"] = "J. Franklin Inst.",
  ["journal of the gay and lesbian medical association"] = "J Gay Lesbian Med Assoc",
  ["journal of the geological society"] = "J Geol Soc London",
  ["journal of the geological society (london, united kingdom)"] = "J. Geol. Soc. (London, U.K.)",
  ["journal of the geological society of australia"] = "J. Geol. Soc. Aust.",
  ["journal of the geological society of china"] = "J. Geol. Soc. China",
  ["journal of the geological society of india"] = "J. Geol. Soc. India",
  ["journal of the geological society of jamaica"] = "J. Geol. Soc. Jam.",
  ["journal of the geological society of london"] = "J. Geol. Soc. London",
  ["journal of the georgia association of historians"] = "Proc Pap Ga Assoc Hist",
  ["journal of the georgia dental association"] = "J Ga Dent Assoc",
  ["journal of the georgia public health association"] = "J Ga Public Health Assoc",
  ["journal of the german statistical society"] = "J. German Statistical Society",
  ["journal of the germantown hospital"] = "J Germantown Hosp",
  ["journal of the ghana science association"] = "J. Ghana Sci. Assoc.",
  ["journal of the graduate school and faculty of engineering"] = "J. Grad. Sch. Fac. Eng. Univ. Tokyo Ser. B",
  ["journal of the greater houston dental society"] = "J. Gt. Houst. Dent. Soc.",
  ["journal of the hawaii dental association"] = "J Hawaii Dent Assoc",
  ["journal of the hawaii state dental association"] = "J Hawaii State Dent Assoc",
  ["journal of the hillside hospital"] = "J Hillside Hosp",
  ["journal of the history of biology"] = "J Hist Biol",
  ["journal of the history of dentistry"] = "J Hist Dent",
  ["journal of the history of economic thought"] = "J. Hist. Econ. Thought",
  ["journal of the history of economic thought : jhet"] = "J Hist Econ Thought",
  ["journal of the history of ideas"] = "J Hist Ideas",
  ["journal of the history of medicine and allied sciences"] = "J Hist Med Allied Sci",
  ["journal of the history of philosophy"] = "J Hist Philos",
  ["journal of the history of sexuality"] = "J Hist Sex",
  ["journal of the history of the behavioral sciences"] = "J Hist Behav Sci",
  ["journal of the history of the neurosciences"] = "J Hist Neurosci",
  ["journal of the hong kong branch of the royal asiatic society"] = "J Hong Kong Branch R Asiat Soc",
  ["journal of the houston district dental society"] = "J. Houston Dist. Dent. Soc.",
  ["journal of the human-environment system"] = "J Hum Environ Syst",
  ["journal of the icru"] = "J ICRU",
  ["journal of the idaho academy of science"] = "J Idaho Acad Sci",
  ["journal of the illinois state historical society. illinois state historical society"] = "J Ill State Hist Soc",
  ["journal of the illuminating engineering society"] = "J. Illum. Eng. Soc.",
  ["journal of the indian academy of dentistry"] = "J Indian Acad Dent",
  ["journal of the indian academy of forensic sciences"] = "J Indian Forensic Sci",
  ["journal of the indian academy of wood science"] = "J. Indian Acad. Wood Sci.",
  ["journal of the indian anthropological society"] = "J Indian Anthropol Soc",
  ["journal of the indian anthropological society / the indian anthropological society"] = "J Indian Anthropol Soc",
  ["journal of the indian chemical society"] = "J. Indian Chem. Soc.",
  ["journal of the indian dental association"] = "J Indian Dent Assoc",
  ["journal of the indian institute of science"] = "J. Indian Inst. Sci.",
  ["journal of the indian institute of science. a multidisciplinary reviews journal"] = "J. Indian Inst. Sci.",
  ["journal of the indian law institute"] = "J Indian Law Inst",
  ["journal of the indian leather technologists’ association"] = "J. Indian Leather Technol. Assoc.",
  ["journal of the indian mathematical society"] = "J. Indian Math. Soc.",
  ["journal of the indian medical association"] = "J Indian Med Assoc",
  ["journal of the indian medical profession"] = "J Indian Med Prof",
  ["journal of the indian ocean region"] = "J. Indian Ocean Reg.",
  ["journal of the indian orthodontic society"] = "J. Indian Orthod. Soc.",
  ["journal of the indian society for probability and statistics"] = "J. Indian Soc. Probab. Stat.",
  ["journal of the indian society of agricultural statistics"] = "J. Indian Soc. Agricultural Statist.",
  ["journal of the indian society of agricultural statistics. indian society of agricultural statistics"] = "J Indian Soc Agric Stat",
  ["journal of the indian society of pedodontics and preventive dentistry"] = "J Indian Soc Pedod Prev Dent",
  ["journal of the indian society of remote sensing"] = "J. Indian Soc. Remote Sens.",
  ["journal of the indian society of soil science"] = "J. Indian Soc. Soil Sci.",
  ["journal of the indian statistical association"] = "J. Indian Statist. Assoc.",
  ["journal of the indiana dental association"] = "J. Indiana Dent. Assoc.",
  ["journal of the indiana state dental association"] = "J. Indiana State Dent. Assoc.",
  ["journal of the indiana state medical association"] = "J. Indiana State Med. Assoc.",
  ["journal of the indianapolis district dental society"] = "J. Indianap. Dist. Dent. Soc.",
  ["journal of the indonesian mathematical society"] = "J. Indones. Math. Soc.",
  ["journal of the indonesian tropical animal agriculture"] = "J Indones Trop Anim Agric",
  ["journal of the institute for conservation"] = "J. Inst. Conserv.",
  ["journal of the institute of actuaries"] = "J Inst Actuar",
  ["journal of the institute of agriculture and animal science"] = "J Inst Agric Anim Sci",
  ["journal of the institute of brewing"] = "J. Inst. Brew.",
  ["journal of the institute of brewing. institute of brewing (great britain)"] = "J Inst Brew",
  ["journal of the institute of chartered accountants of sri lanka"] = "J Inst Chart Account Sri Lanka",
  ["journal of the institute of environment sciences"] = "J. Inst. Environ. Sci.",
  ["journal of the institute of hospital engineering"] = "J. Inst. Hosp. Eng.",
  ["journal of the institute of mathematics of jussieu"] = "J. Inst. Math. Jussieu",
  ["journal of the institute of mathematics of jussieu. jimj. journal de l’institut de mathematiques de jussieu"] = "J. Inst. Math. Jussieu",
  ["journal of the institute of medicine"] = "J Inst Med",
  ["journal of the institute of metals"] = "J. Inst. Met.",
  ["journal of the institute of petroleum"] = "J. Inst. Pet.",
  ["journal of the institution of chemists (india)"] = "J. Inst. Chem. (India)",
  ["journal of the institution of electrical engineers"] = "J. Inst. Electr. Eng.",
  ["journal of the institution of electronic and radio engineers"] = "J. Inst. Electron. Radio Eng.",
  ["journal of the institution of engineers (india) chemical engineering division"] = "J. Inst. Eng. (India) Chem. Eng. Div.",
  ["journal of the institution of engineers (india) metallurgical and materials engineering division"] = "J. Inst. Eng. (India) Metall. Mater. Eng. Div.",
  ["journal of the institution of engineers (india): series a"] = "J. Inst. Eng. India Ser. A",
  ["journal of the institution of engineers (india): series b"] = "J. Inst. Eng. India Ser. B",
  ["journal of the institution of engineers (india): series c"] = "J. Inst. Eng. India Ser. C",
  ["journal of the institution of engineers (india): series d"] = "J. Inst. Eng. India Ser. D",
  ["journal of the institution of engineers (india): series e"] = "J. Inst. Eng. India Ser. E",
  ["journal of the institution of locomotive engineers"] = "J. Inst. Locomot. Eng.",
  ["journal of the institution of production engineers"] = "J. Inst. Prod. Eng.",
  ["journal of the institution of water and environment management"] = "J. Inst. Water Environ. Manage.",
  ["journal of the instrument society of india"] = "J. Instrum. Soc. India",
  ["journal of the intensive care society"] = "J Intensive Care Soc",
  ["journal of the international academy of periodontology"] = "J Int Acad Periodontol",
  ["journal of the international aids society"] = "J Int AIDS Soc",
  ["journal of the international association of dentistry for children"] = "J Int Assoc Dent Child",
  ["journal of the international association of physicians in aids care"] = "J Int Assoc Physicians AIDS Care",
  ["journal of the international association of physicians in aids care (chicago, ill. : 2002)"] = "J Int Assoc Physicians AIDS Care (Chic)",
  ["journal of the international association of providers of aids care"] = "J Int Assoc Provid AIDS Care",
  ["journal of the international college of dentists, japan section"] = "J. Int. Coll. Dent. (Jpn.)",
  ["journal of the international college of dentists; japan section"] = "J Int Coll Dent (Jpn)",
  ["journal of the international college of surgeons"] = "J. Int. Coll. Surg.",
  ["journal of the international federation of clinical chemistry"] = "J Int Fed Clin Chem",
  ["journal of the international federation of clinical chemistry / ifcc"] = "J Int Fed Clin Chem",
  ["journal of the international federation of gynaecology and obstetrics"] = "J Int Fed Gynaecol Obstet",
  ["journal of the international neuropsychological society"] = "J. Int. Neuropsychol. Soc.",
  ["journal of the international neuropsychological society : jins"] = "J Int Neuropsychol Soc",
  ["journal of the international phonetic association"] = "J Int Phon Assoc",
  ["journal of the international society for respiratory protection"] = "J Int Soc Respir Prot",
  ["journal of the international society for telemedicine and ehealth"] = "J Int Soc Telemed eHealth",
  ["journal of the international society for the history of islamic medicine (ishim) : jishim"] = "J Int Soc History Islamic Med",
  ["journal of the international society of sports nutrition"] = "J Int Soc Sports Nutr",
  ["journal of the international wine & food society (london : 1989)"] = "J Int Wine Food Soc",
  ["journal of the iowa medical society"] = "J Iowa Med Soc",
  ["journal of the iran dental association"] = "J. Iran Dent. Assoc.",
  ["journal of the iranian chemical society"] = "J. Iran. Chem. Soc.",
  ["journal of the iranian mathematical society"] = "J. Iran. Math. Soc.",
  ["journal of the iranian statistical society. (jirss)"] = "J. Iran. Stat. Soc. (JIRSS)",
  ["journal of the irish colleges of physicians and surgeons"] = "J Ir Coll Physicians Surg",
  ["journal of the irish dental association"] = "J Ir Dent Assoc",
  ["journal of the irish medical association"] = "J Ir Med Assoc",
  ["journal of the japan institute of metals"] = "J. Jpn. Inst. Met.",
  ["journal of the japan petroleum institute"] = "J. Jpn. Pet. Inst.",
  ["journal of the japan society of powder and powder metallurgy"] = "J. Jpn. Soc. Powder Powder Metall.",
  ["journal of the japan statistical society (nihon tokei gakkai kaiho)"] = "J. Japan Statist. Soc.",
  ["journal of the japan statistical society (nihon tôkei gakkai kaihô)"] = "J. Japan Statist. Soc.",
  ["journal of the japan statistical society. japanese issue"] = "J. Jpn. Stat. Soc. Jpn. Issue",
  ["journal of the japanese and international economies"] = "J Jpn Int Econ",
  ["journal of the japanese obstetrical & gynecological society"] = "J Jpn Obstet Gynecol Soc",
  ["journal of the japanese obstetrical and gynecological society"] = "J. Jpn. Obstet. Gynecol. Soc.",
  ["journal of the japanese physical therapy association = rigaku ryōhō"] = "J Jpn Phys Ther Assoc",
  ["journal of the japanese society of computational statistics"] = "J. Japanese Soc. Comput. Statist.",
  ["journal of the kansas dental association"] = "J Kans Dent Assoc",
  ["journal of the kansas entomological society"] = "J Kans Entomol Soc",
  ["journal of the kansas medical society"] = "J. Kans. Med. Soc.",
  ["journal of the kansas state dental association"] = "J. Kans. State Dent. Assoc.",
  ["journal of the kentucky academy of science"] = "J Ky Acad Sci",
  ["journal of the kentucky dental association"] = "J Ky Dent Assoc",
  ["journal of the kentucky medical association"] = "J. Ky. Med. Assoc.",
  ["journal of the kenya chemical society"] = "J. Kenya Chem. Soc.",
  ["journal of the knowledge economy"] = "J. Knowl. Econ.",
  ["journal of the korea institute of applied superconductivity and cryogenics"] = "J. Korea Inst. Appl. Supercond. Cryog. Inst.",
  ["journal of the korea society of mathematical education"] = "J. Korea Soc. Math. Educ. Ser. B Pure Appl. Math.",
  ["journal of the korean association of oral and maxillofacial surgeons"] = "J Korean Assoc Oral Maxillofac Surg",
  ["journal of the korean astronomical society"] = "J. Korean Astron. Soc.",
  ["journal of the korean ceramic society"] = "J. Korean Ceram. Soc.",
  ["journal of the korean chemical society"] = "J. Korean Chem. Soc.",
  ["journal of the korean mathematical society"] = "J. Korean Math. Soc.",
  ["journal of the korean physical society"] = "J. Korean Phys. Soc.",
  ["journal of the korean research society for dental hypnosis"] = "J Korean Res Soc Dent Hypn",
  ["journal of the korean society for applied biological chemistry"] = "J Korean Soc Appl Biol Chem",
  ["journal of the korean society for industrial and applied mathematics"] = "J. Korean Soc. Ind. Appl. Math.",
  ["journal of the korean society of coloproctology"] = "J Korean Soc Coloproctol",
  ["journal of the korean society of hypertension"] = "J Korean Soc Hypertens",
  ["journal of the korean society of mathematical education. series b. the pure and applied mathematics"] = "J. Korean Soc. Math. Educ. Ser. B Pure Appl. Math.",
  ["journal of the korean society of mathematical education. series b. theoretical mathematics and pedagogical mathematics"] = "J. Korean Soc. Math. Educ. Ser. B. Theor. Math. Pedag. Math.",
  ["journal of the korean statistical society"] = "J. Korean Statist. Soc.",
  ["journal of the korean surgical society"] = "J Korean Surg Soc",
  ["journal of the lancaster county historical society. lancaster county historical society (pa.)"] = "J Lanc Cty Hist Soc",
  ["journal of the lepidopterists' society"] = "J Lepid Soc",
  ["journal of the lepidopterists’ society"] = "J. Lepid. Soc.",
  ["journal of the less-common metals"] = "J. Less-Common Met.",
  ["journal of the linnean society of london, botany"] = "J. Linn. Soc. London, Bot.",
  ["journal of the linnean society of london, zoology"] = "J. Linn. Soc. London, Zool.",
  ["journal of the london mathematical society"] = "J Lond Math Soc",
  ["journal of the london mathematical society. second series"] = "J. Lond. Math. Soc. (2)",
  ["journal of the louisiana dental association"] = "J La Dent Assoc",
  ["journal of the louisiana state medical society"] = "J. La. State Med. Soc.",
  ["journal of the macomb dental society"] = "J Macomb Dent Soc",
  ["journal of the magnetics society of japan"] = "J. Magn. Soc. Jpn.",
  ["journal of the maine medical association"] = "J. Maine Med. Assoc.",
  ["journal of the malaria institute of india"] = "J Malar Inst India",
  ["journal of the malaysian branch of the royal asiatic society"] = "J Malays Branch R Asiat Soc",
  ["journal of the mammalogical society of japan"] = "J Mammal Soc Jpn",
  ["journal of the marine biological association of the uk"] = "J. Mar. Biol. Assoc. U. K.",
  ["journal of the marine biological association of the united kingdom"] = "J. Mar. Biol. Assoc. U. K.",
  ["journal of the marine biological association of the united kingdom. marine biological association of the united kingdom"] = "J Mar Biol Assoc U K",
  ["journal of the marine biology association of the united kingdom"] = "J. mar. biol. Ass. U.K.",
  ["journal of the marine technology society"] = "J. Mar. Technol. Soc.",
  ["journal of the market research society. market research society"] = "J Mark Res Soc",
  ["journal of the maryland state dental association"] = "J Md State Dent Assoc",
  ["journal of the mass spectrometry society of japan"] = "J Mass Spectrom Soc Jpn",
  ["journal of the massachusetts dental society"] = "J Mass Dent Soc",
  ["journal of the mathematical society of japan"] = "J. Math. Soc. Japan",
  ["journal of the mechanical behavior of biomedical materials"] = "J Mech Behav Biomed Mater",
  ["journal of the mechanical behavior of materials"] = "J Mech Behav Mater",
  ["journal of the mechanics and physics of solids"] = "J. Mech. Phys. Solids",
  ["journal of the medical association for prevention of war"] = "J Med Assoc Prev War",
  ["journal of the medical association of georgia"] = "J Med Assoc Ga",
  ["journal of the medical association of thailand"] = "J. Med. Assoc. Thai.",
  ["journal of the medical association of thailand = chotmaihet thangphaet"] = "J Med Assoc Thai",
  ["journal of the medical association of the state of alabama"] = "J Med Assoc State Ala",
  ["journal of the medical library association"] = "J. Med. Libr. Assoc.",
  ["journal of the medical library association : jmla"] = "J Med Libr Assoc",
  ["journal of the medical society of new jersey"] = "J. Med. Soc. N. J.",
  ["journal of the mercer dental society"] = "J. Mercer Dent. Soc.",
  ["journal of the meteorological society of japan"] = "J. Meteorolog. Soc. Jpn.",
  ["journal of the mexican chemical society"] = "J. Mex. Chem. Soc.",
  ["journal of the michigan state dental association"] = "J. Mich. State Dent. Assoc.",
  ["journal of the middle east and africa"] = "J Middle East Afr",
  ["journal of the mine ventilation society of south africa. mine ventilation society of south africa"] = "J Mine Vent Soc S Afr",
  ["journal of the mining and materials processing institute of japan"] = "J. Min. Mater. Process.Inst. Jpn.",
  ["journal of the minnesota academy of science"] = "J Minn Acad Sci",
  ["journal of the mississippi academy of sciences. mississippi academy of sciences"] = "J Miss Acad Sci",
  ["journal of the mississippi state medical association"] = "J Miss State Med Assoc",
  ["journal of the missouri bar"] = "J Mo Bar",
  ["journal of the missouri dental association"] = "J Mo Dent Assoc",
  ["journal of the mongolian mathematical society"] = "J. Mong. Math. Soc.",
  ["journal of the mount sinai hospital, new york"] = "J Mt Sinai Hosp N Y",
  ["journal of the national analgesia society"] = "J. Natl. Analg. Soc.",
  ["journal of the national association of chiropodists"] = "J Natl Assoc Chirop",
  ["journal of the national association of seventh-day adventist dentists"] = "J Natl Assoc Seventh Day Advent Dent",
  ["journal of the national cancer institute"] = "J Natl Cancer Inst",
  ["journal of the national cancer institute (1988)"] = "J. Natl. Cancer Inst.",
  ["journal of the national cancer institute. monographs"] = "J Natl Cancer Inst Monogr",
  ["journal of the national cancer institute. monographs."] = "J. Natl. Cancer Inst. Monogr.",
  ["journal of the national comprehensive cancer network"] = "J. Natl. Compr. Canc. Netw.",
  ["journal of the national comprehensive cancer network : jnccn"] = "J Natl Compr Canc Netw",
  ["journal of the national medical association"] = "J Natl Med Assoc",
  ["journal of the national science foundation of sri lanka"] = "J Natl Sci Found",
  ["journal of the nebraska dental association"] = "J. Nebr. Dent. Assoc.",
  ["journal of the neurological sciences"] = "J Neurol Sci",
  ["journal of the neuromusculoskeletal system : jnms : a journal of the american chiropractic association, inc"] = "J Neuromusculoskelet Syst",
  ["journal of the new jersey dental association"] = "J N J Dent Assoc",
  ["journal of the new jersey dental hygienists association"] = "J N J Dent Hyg Assoc",
  ["journal of the new jersey state dental society"] = "J. N. J. State Dent. Soc.",
  ["journal of the new york state nurses association"] = "J. N. Y. State Nurses Assoc.",
  ["journal of the new york state school nurse teachers association"] = "J. N. Y. State Sch. Nurse Teach. Assoc.",
  ["journal of the new york state school nurse-teachers association"] = "J N Y State Sch Nurse Teach Assoc",
  ["journal of the new zealand society of periodontology"] = "J N Z Soc Periodontol",
  ["journal of the nigeria association of dental students"] = "J Niger Assoc Dent Stud",
  ["journal of the nigeria medical association"] = "J Nigeria Med Assoc",
  ["journal of the nigerian mathematical society"] = "J. Nigerian Math. Soc.",
  ["journal of the nigerian society of physical sciences"] = "J. Niger. Soc. Phys. Sci.",
  ["journal of the north american benthological society"] = "J North Am Benthol Soc",
  ["journal of the north atlantic"] = "J. North Atl.",
  ["journal of the north carolina academy of science"] = "J. N.C. Acad. Sci.",
  ["journal of the north carolina dental society"] = "J. N. C. Dent. Soc.",
  ["journal of the northwestern polytechnical university"] = "J. Northwest. Polytech. Univ.",
  ["journal of the ocean science foundation"] = "J Ocean Sci Found",
  ["journal of the oceanographical society of japan"] = "J. Oceanogr. Soc. Jpn.",
  ["journal of the ontario dental association"] = "J. Ont. Dent. Assoc.",
  ["journal of the operating room research institute"] = "J. Oper. Room Res. Inst.",
  ["journal of the operational research society"] = "J. Oper. Res. Soc.",
  ["journal of the operations research society of china"] = "J. Oper. Res. Soc. China",
  ["journal of the operations research society of japan"] = "J. Oper. Res. Soc. Japan",
  ["journal of the optical society of america"] = "J Opt Soc Am",
  ["journal of the optical society of america a"] = "J. Opt. Soc. Amer. A",
  ["journal of the optical society of america a, optics and image science"] = "J. Opt. Soc. Am. A",
  ["journal of the optical society of america a, optics, image science, and vision"] = "J. Opt. Soc. Am. A Opt. Image Sci. Vis.",
  ["journal of the optical society of america a. optics, image science, and vision"] = "J Opt Soc Am A Opt Image Sci Vis",
  ["journal of the optical society of america a: optics, image science & vision"] = "J. Opt. Soc. Am. A",
  ["journal of the optical society of america a: optics, image science, and vision"] = "J. Opt. Soc. Am. A:",
  ["journal of the optical society of america b"] = "J. Opt. Soc. Am. B",
  ["journal of the optical society of america b: optical physics"] = "J. Opt. Soc. Am. B: Opt. Phys.",
  ["journal of the optical society of america. a, optics and image science"] = "J Opt Soc Am A",
  ["journal of the optical society of america. a, optics, image science, and vision"] = "J Opt Soc Am A Opt Image Sci Vis",
  ["journal of the optical society of america. b, optical physics"] = "J Opt Soc Am B",
  ["journal of the optical society of korea"] = "J Opt Soc Korea",
  ["journal of the oregon dental association"] = "J. Oreg. Dent. Assoc.",
  ["journal of the orissa mathematical society"] = "J. Orissa Math. Soc.",
  ["journal of the osaka university dental school"] = "J. Osaka Univ. Dent. Sch.",
  ["journal of the oslo city hospitals"] = "J Oslo City Hosp",
  ["journal of the oto-laryngological society of australia"] = "J Otolaryngol Soc Aust",
  ["journal of the otto rank association. otto rank association"] = "J Otto Rank Assoc",
  ["journal of the pakistan medical association"] = "J Pak Med Assoc",
  ["journal of the parenteral drug association"] = "J Parenter Drug Assoc",
  ["journal of the patent and trademark office society"] = "J Pat Trademark Off Soc",
  ["journal of the patent office society. patent office society (u.s.)"] = "J Pat Off Soc",
  ["journal of the pediatric infectious diseases society"] = "J Pediatric Infect Dis Soc",
  ["journal of the pennsylvania academy of science"] = "J Pa Acad Sci",
  ["journal of the peripheral nervous system"] = "J. Peripher. Nerv. Syst.",
  ["journal of the peripheral nervous system : jpns"] = "J Peripher Nerv Syst",
  ["journal of the philadelphia county dental society"] = "J. Phila. Cty. Dent. Soc.",
  ["journal of the philippine dental association"] = "J. Philipp. Dent. Assoc.",
  ["journal of the philippine federation of private medical practitioners"] = "J Philipp Fed Priv Med Pract",
  ["journal of the philippine medical association"] = "J Philipp Med Assoc",
  ["journal of the philosophy of sport"] = "J Philos Sport",
  ["journal of the physical society of japan"] = "J. Phys. Soc. Jpn.",
  ["journal of the physicians association for aids care"] = "J Physicians Assoc AIDS Care",
  ["journal of the proceedings of the linnean society of london. botany"] = "J. Proc. Linn. Soc. London Bot.",
  ["journal of the proceedings of the linnean society of london. zoology"] = "J. Proc. Linn. Soc. London Zool.",
  ["journal of the ramanujan mathematical society"] = "J. Ramanujan Math. Soc.",
  ["journal of the renin-angiotensin-aldosterone system"] = "J. Renin-Angiotensin-Aldosterone Syst.",
  ["journal of the renin-angiotensin-aldosterone system : jraas"] = "J Renin Angiotensin Aldosterone Syst",
  ["journal of the reticuloendothelial society"] = "J Reticuloendothel Soc",
  ["journal of the rhode island state dental society"] = "J R I State Dent Soc",
  ["journal of the rocky mountain analgesia society"] = "J Rocky Mt Analg Soc",
  ["journal of the rocky mountain medieval and renaissance association. rocky mountain medieval and renaissance association"] = "J Rocky Mt Mediev Renaiss Assoc",
  ["journal of the roy aeronautical society"] = "J. R. Aeronaut. Soc.",
  ["journal of the royal aeronautical society"] = "J. R. Aeronaut. Soc.",
  ["journal of the royal anthropological institute"] = "J. R. Anthropol. Inst.",
  ["journal of the royal army medical corps"] = "J R Army Med Corps",
  ["journal of the royal asiatic society of great britain & ireland. royal asiatic society of great britain and ireland"] = "J R Asiat Soc GB Irel",
  ["journal of the royal asiatic society of great britain and ireland"] = "JAS",
  ["journal of the royal astronomical society of canada"] = "J. R. Astron. Soc. Can.",
  ["journal of the royal australian historical society"] = "J R Aust Hist Soc",
  ["journal of the royal college of general practitioners"] = "J. R. Coll. Gen. Pract.",
  ["journal of the royal college of general practitioners. occasional paper"] = "J. R. Coll. Gen. Pract. Occas. Pap.",
  ["journal of the royal college of physicians of london"] = "J R Coll Physicians Lond",
  ["journal of the royal college of surgeons of edinburgh"] = "J R Coll Surg Edinb",
  ["journal of the royal institute of chemistry"] = "J. R. Inst. Chem.",
  ["journal of the royal microscopical society"] = "J. R. Microsc. Soc.",
  ["journal of the royal naval medical service"] = "J R Nav Med Serv",
  ["journal of the royal nova scotia historical society"] = "J R N S Hist Soc",
  ["journal of the royal society of antiquaries of ireland. royal society of antiquaries of ireland"] = "J R Soc Antiq Irel",
  ["journal of the royal society of arts"] = "J R Soc Arts",
  ["journal of the royal society of health"] = "J R Soc Health",
  ["journal of the royal society of medicine"] = "J R Soc Med",
  ["journal of the royal society of new zealand"] = "J R Soc N Z",
  ["journal of the royal society of western australia"] = "J R Soc West Aust",
  ["journal of the royal society, interface"] = "J R Soc Interface",
  ["journal of the royal society, interface / the royal society"] = "J R Soc Interface",
  ["journal of the royal statistical society"] = "J. R. Stat. Soc.",
  ["journal of the royal statistical society series a (statistics in society)"] = "J. R. Stat. Soc. A",
  ["journal of the royal statistical society series b (statistical methodology)"] = "J. R. Stat. Soc. B",
  ["journal of the royal statistical society series c (applied statistics)"] = "J. R. Stat. Soc. C",
  ["journal of the royal statistical society series d (the statistician)"] = "J. R. Stat. Soc. D",
  ["journal of the royal statistical society, series a"] = "J. Roy. Statistical Society",
  ["journal of the royal statistical society. series a (general)"] = "J R Stat Soc Ser A",
  ["journal of the royal statistical society. series a, (statistics in society)"] = "J R Stat Soc Ser A Stat Soc",
  ["journal of the royal statistical society. series a. statistics in society"] = "J. Roy. Statist. Soc. Ser. A",
  ["journal of the royal statistical society. series b, statistical methodology"] = "J R Stat Soc Series B Stat Methodol",
  ["journal of the royal statistical society. series b. statistical methodology"] = "J. R. Stat. Soc. Ser. B. Stat. Methodol.",
  ["journal of the royal statistical society. series c, applied statistics"] = "J R Stat Soc Ser C Appl Stat",
  ["journal of the royal statistical society. series c. applied statistics"] = "J. R. Stat. Soc. Ser. C. Appl. Stat.",
  ["journal of the royal statistical society: series b (statistical methodology)"] = "J. R. Stat. Soc. B",
  ["journal of the russell society"] = "J. Russell Soc.",
  ["journal of the san antonio district dental society"] = "J. San Antonio Dent. Soc.",
  ["journal of the saudi heart association"] = "J Saudi Heart Assoc",
  ["journal of the saudi society of agricultural sciences"] = "J. Saudi Soc. Agric. Sci.",
  ["journal of the science of food and agriculture"] = "J Sci Food Agric",
  ["journal of the science society of thailand"] = "J Sci Soc Thailand",
  ["journal of the seattle district dental society"] = "J Seattle Dist Dent Soc",
  ["journal of the serbian chemical society"] = "J. Serb. Chem. Soc.",
  ["journal of the serbian society for computational mechanics"] = "J Serbian Soc Comput Mech",
  ["journal of the singapore paediatric society"] = "J. Singapore Paediatr. Soc.",
  ["journal of the social sciences"] = "J. Soc. Sci.",
  ["journal of the society for army historical research. society for army historical research (london, england)"] = "J Soc Army Hist Res",
  ["journal of the society for gynecologic investigation"] = "J Soc Gynecol Investig",
  ["journal of the society for health systems"] = "J Soc Health Syst",
  ["journal of the society for information display"] = "J Soc Inf Disp",
  ["journal of the society for integrative oncology"] = "J Soc Integr Oncol",
  ["journal of the society for radiological protection"] = "J. Soc. Radiol. Prot.",
  ["journal of the society for social work and research"] = "J Soc Social Work Res",
  ["journal of the society of architectural historians"] = "JSAH",
  ["journal of the society of archivists. society of archivists (great britain)"] = "J Soc Arch",
  ["journal of the society of chemical industry, london"] = "J. Soc. Chem. Ind. London",
  ["journal of the society of christian ethics"] = "J Soc Christ Ethics",
  ["journal of the society of cosmetic chemists"] = "J Soc Cosmet Chem",
  ["journal of the society of dyers and colourists"] = "J. Soc. Dyers Colour.",
  ["journal of the society of glass technology"] = "J. Soc. Glass Technol.",
  ["journal of the society of inorganic materials, japan"] = "J. Soc. Inorg. Mater., Jpn.",
  ["journal of the society of laparoendoscopic surgeons"] = "J. Soc. Laparoendosc. Surg.",
  ["journal of the society of leather technologists and chemists"] = "J. Soc. Leather Technol. Chem.",
  ["journal of the society of materials science, japan"] = "J. Soc. Mater. Sci., Jpn.",
  ["journal of the society of occupational medicine"] = "J. Soc. Occup. Med.",
  ["journal of the society of pediatric nurses"] = "J. Soc. Pediatr. Nurs.",
  ["journal of the society of pediatric nurses : jspn"] = "J Soc Pediatr Nurs",
  ["journal of the society of research administrators. society of research administrators"] = "J Soc Res Adm",
  ["journal of the society of telegraph engineers"] = "J. Soc. Telegraph Eng.",
  ["journal of the solar energy society of india"] = "J. Sol. Energy Soc. India",
  ["journal of the south african chemical institute. south african chemical institute"] = "J S Afr Chem Inst",
  ["journal of the south african institute of civil engineering"] = "J. South Afr. Inst. Civ. Eng.",
  ["journal of the south african institute of mining and metallurgy"] = "J. S. Afr. Inst. Min. Metall.",
  ["journal of the south african institution of civil engineering"] = "J. S. Afr. Inst. Civ. Eng.",
  ["journal of the south african logopedic society"] = "J S Afr Logop Soc",
  ["journal of the south african speech and hearing association"] = "J S Afr Speech Hear Assoc",
  ["journal of the south african veterinary association"] = "J S Afr Vet Assoc",
  ["journal of the south african veterinary medical association"] = "J S Afr Vet Med Assoc",
  ["journal of the south carolina academy of science"] = "J S C Acad Sci",
  ["journal of the south carolina medical association"] = "J S C Med Assoc",
  ["journal of the south carolina medical association (1975)"] = "J S C Med Assoc",
  ["journal of the southern african institute of mining and metallurgy"] = "J South Afr Inst Min Metall",
  ["journal of the southern california dental hygienists association"] = "J. South. Calif. State. Dent. Hyg. Assoc.",
  ["journal of the southern california dental hygienists' association"] = "J South Calif State Dent Hyg Assoc",
  ["journal of the southern california dental hygienists’ association"] = "J South Calif State Dent Hyg Assoc",
  ["journal of the southern orthopaedic association"] = "J South Orthop Assoc",
  ["journal of the southwest"] = "J Southwest",
  ["journal of the sri lanka branch of the royal asiatic society"] = "J Sri Lanka Branch R Asiat Soc",
  ["journal of the statistical and social inquiry society of ireland"] = "J. Statistical Soc. Inquiry Society Ireland",
  ["journal of the taiwan institute of chemical engineers"] = "J. Taiwan Inst. Chem. Eng.",
  ["journal of the technical university at plovdiv"] = "J. Tech. Univ. Plovdiv Fundam. Sci. Appl. Ser. A Pure Appl. Math.",
  ["journal of the tennessee dental association"] = "J. Tenn. Dent. Assoc.",
  ["journal of the tennessee medical association"] = "J Tenn Med Assoc",
  ["journal of the tensor society"] = "J. Tensor Soc.",
  ["journal of the texas dental hygienists association"] = "J. Texas Dent. Hyg. Assoc.",
  ["journal of the textile institute"] = "J. Text. Inst.",
  ["journal of the torrey botanical society"] = "J. Torrey Bot. Soc.",
  ["journal of the turkish chemical society section a chemistry"] = "J. Turk. Chem. Soc. Sect. A Chem.",
  ["journal of the turkish german gynecological association"] = "J Turk Ger Gynecol Assoc",
  ["journal of the urban and regional information systems association"] = "J Urban Reg Inf Syst Assoc",
  ["journal of the walters art gallery"] = "JWAG",
  ["journal of the warburg and courtauld institutes"] = "J Warburg Courtauld Inst",
  ["journal of the west"] = "J West",
  ["journal of the west african college of surgeons"] = "J West Afr Coll Surg",
  ["journal of the west australian nurses"] = "J West Aust Nurses",
  ["journal of the west virginia philosophical society"] = "J West Va Philo Soc",
  ["journal of the western society of periodontology / periodontal abstracts"] = "J. West. Soc. Periodontal Periodontal Abstr.",
  ["journal of the wisconsin dental association"] = "J. Wis. Dent. Assoc.",
  ["journal of the wisconsin state dental society"] = "J. Wis. State Dent. Soc.",
  ["journal of the world aquaculture society"] = "J World Aquac Soc",
  ["journal of the world federation of orthodontists"] = "J World Fed Orthod",
  ["journal of theological studies. oxford, clarendon press"] = "JThS",
  ["journal of theology for southern africa"] = "J Theol South Afr",
  ["journal of theoretical & computational chemistry"] = "J Theor Comput Chem",
  ["journal of theoretical and applied computer science"] = "J Theor Appl Comput Sci",
  ["journal of theoretical and applied information technology"] = "J Theor Appl Inf Technol",
  ["journal of theoretical and applied mechanics"] = "J. Theoret. Appl. Mech.",
  ["journal of theoretical and applied physics"] = "J. Theor. Appl. Phys.",
  ["journal of theoretical and computational acoustics"] = "J. Theor. Comput. Acoust.",
  ["journal of theoretical and computational chemistry"] = "J. Theor. Comput. Chem.",
  ["journal of theoretical archaeology"] = "JTheorA",
  ["journal of theoretical biology"] = "J Theor Biol",
  ["journal of theoretical politics"] = "J Theor Polit",
  ["journal of theoretical probability"] = "J. Theoret. Probab.",
  ["journal of theory and practice of dental public health"] = "J Theory Pract Dent Public Health",
  ["journal of theory construction & testing"] = "J Theory Constr Test",
  ["journal of therapeutic ultrasound"] = "J Ther Ultrasound",
  ["journal of thermal analysis"] = "J. Therm. Anal.",
  ["journal of thermal analysis and calorimetry"] = "J. Therm. Anal. Calorim.",
  ["journal of thermal biology"] = "J Therm Biol",
  ["journal of thermal envelope & building science"] = "J. Therm. Envelope Build. Sci.",
  ["journal of thermal envelope and building science"] = "J. Therm. Envelope Build. Sci.",
  ["journal of thermal insulation"] = "J. Therm. Insul.",
  ["journal of thermal insulation and building envelopes"] = "J. Therm. Insul. Build. Envelopes",
  ["journal of thermal science"] = "J. Therm. Sci.",
  ["journal of thermal science and engineering applications"] = "J Therm Sci Eng Appl",
  ["journal of thermal science and technology"] = "J. Therm. Sci. Technol.",
  ["journal of thermal spray technology"] = "J. Therm. Spray Technol.",
  ["journal of thermal stresses"] = "J. Therm. Stresses",
  ["journal of thermodynamics & catalysis"] = "J. Thermodyn. Catal.",
  ["journal of thermophysics and heat transfer"] = "J. Thermophys Heat Transfer",
  ["journal of thermoplastic composite materials"] = "J. Thermoplast. Compos. Mater.",
  ["journal of third world studies"] = "J Third World Stud",
  ["journal of thoracic and cardiovascular surgery"] = "J. Thorac. Cardiovasc. Surg.",
  ["journal of thoracic disease"] = "J Thorac Dis",
  ["journal of thoracic imaging"] = "J Thorac Imaging",
  ["journal of thoracic oncology : official publication of the international association for the study of lung cancer"] = "J Thorac Oncol",
  ["journal of thought"] = "J Thought",
  ["journal of threatened taxa"] = "J Threat Taxa",
  ["journal of thrombosis and haemostasis"] = "J. Thromb. Haemost.",
  ["journal of thrombosis and haemostasis : jth"] = "J Thromb Haemost",
  ["journal of thrombosis and thrombolysis"] = "J Thromb Thrombolysis",
  ["journal of thyroid disorders & therapy"] = "J Thyroid Disord Ther",
  ["journal of thyroid research"] = "J Thyroid Res",
  ["journal of tianjin normal university"] = "J. Tianjin Norm. Univ. Nat. Sci. Ed.",
  ["journal of tianjin university"] = "J. Tianjin Univ.",
  ["journal of time series analysis"] = "J. Time Ser. Anal.",
  ["journal of time series econometrics"] = "J. Time Ser. Econom.",
  ["journal of time use research"] = "J Time Use Res",
  ["journal of tissue culture methods : tissue culture association manual of cell, tissue, and organ culture procedures"] = "J Tissue Cult Methods",
  ["journal of tissue engineering"] = "J Tissue Eng",
  ["journal of tissue engineering and regenerative medicine"] = "J Tissue Eng Regen Med",
  ["journal of tissue science & engineering"] = "J Tissue Sci Eng",
  ["journal of tissue viability"] = "J Tissue Viability",
  ["journal of tongji medical university"] = "J. Tongji Med. Univ.",
  ["journal of tongji medical university = tong ji yi ke da xue xue bao"] = "J Tongji Med Univ",
  ["journal of topology"] = "J. Topol.",
  ["journal of topology\tand analysis"] = "J. Topol. Anal.",
  ["journal of topology and analysis"] = "J. Topol. Anal.",
  ["journal of toxicologic pathology"] = "J Toxicol Pathol",
  ["journal of toxicological education"] = "J Toxicol Educ",
  ["journal of toxicological sciences"] = "J. Toxicol. Sci.",
  ["journal of toxicology"] = "J Toxicol",
  ["journal of toxicology & environmental health part a: current issues"] = "J. Toxicol. Environ. Health Part A",
  ["journal of toxicology & environmental health part b: critical reviews"] = "J. Toxicol. Environ. Health Part B",
  ["journal of toxicology - clinical toxicology"] = "J. Toxicol. Clin. Toxicol.",
  ["journal of toxicology - cutaneous and ocular toxicology"] = "J. Toxicol. Cutaneous Ocul. Toxicol.",
  ["journal of toxicology - toxin reviews"] = "J. Toxicol. Toxin Rev.",
  ["journal of toxicology and environment health"] = "J. Toxicol. Environ. Health",
  ["journal of toxicology and environmental health"] = "J Toxicol Environ Health",
  ["journal of toxicology and environmental health part a: current issues"] = "J. Toxicol. Environ. Health Part A",
  ["journal of toxicology and environmental health part b: critical reviews"] = "J. Toxicol. Environ. Health Part B",
  ["journal of toxicology and environmental health sciences"] = "J Toxicol Environ Health Sci",
  ["journal of toxicology and environmental health, part a"] = "J. Toxicol. Environ. Health, Part A",
  ["journal of toxicology and environmental health, part a: current issues"] = "J. Toxicol. Environ. Health, Part A",
  ["journal of toxicology and environmental health, part b: critical reviews"] = "J. Toxicol. Environ. Health, Part B",
  ["journal of toxicology and environmental health. part a"] = "J Toxicol Environ Health A",
  ["journal of toxicology and environmental health. part b, critical reviews"] = "J Toxicol Environ Health B Crit Rev",
  ["journal of toxicology and environmental health. supplement"] = "J Toxicol Environ Health Suppl",
  ["journal of toxicology and pharmacology"] = "J Toxicol Pharmacol",
  ["journal of toxicology, clinical toxicology"] = "J. Toxicol. Clin. Toxicol.",
  ["journal of toxicology. clinical toxicology"] = "J Toxicol Clin Toxicol",
  ["journal of toxicology. cutaneous and ocular toxicology"] = "J Toxicol Cutaneous Ocul Toxicol",
  ["journal of toxicology. toxin reviews"] = "J Toxicol Toxin Rev",
  ["journal of trace and microprobe techniques"] = "J. Trace Microprobe Tech.",
  ["journal of trace elements and electrolytes in health and disease"] = "J Trace Elem Electrolytes Health Dis",
  ["journal of trace elements in experimental medicine"] = "J. Trace Elem. Exp. Med.",
  ["journal of trace elements in medicine and biology"] = "J. Trace Elem. Med. Biol.",
  ["journal of trace elements in medicine and biology : organ of the society for minerals and trace elements (gms)"] = "J Trace Elem Med Biol",
  ["journal of traditional and complementary medicine"] = "J Tradit Complement Med",
  ["journal of traditional chinese medicine"] = "J. Tradit. Chin. Med.",
  ["journal of traditional chinese medicine = chung i tsa chih ying wen pan"] = "J Tradit Chin Med",
  ["journal of traditional medicine & clinical naturopathy"] = "J Tradit Med Clin Naturop",
  ["journal of traffic and transportation engineering"] = "J. Traffic Transp. Eng.",
  ["journal of transcultural nursing"] = "J. Transcult. Nurs.",
  ["journal of transcultural nursing : official journal of the transcultural nursing society"] = "J Transcult Nurs",
  ["journal of transcultural nursing : official journal of the transcultural nursing society / transcultural nursing society"] = "J Transcult Nurs",
  ["journal of transfusion"] = "J Transfus",
  ["journal of translational autoimmunity"] = "J Transl Autoimmun",
  ["journal of translational genetics and genomics"] = "J Transl Genet Genom",
  ["journal of translational internal medicine"] = "J Transl Int Med",
  ["journal of translational medicine"] = "J Transl Med",
  ["journal of translational medicine & epidemiology"] = "J Transl Med Epidemiol",
  ["journal of translational neuroscience"] = "J Transl Neurosci (Beijing)",
  ["journal of translational neurosciences"] = "J Transl Neurosci",
  ["journal of translational science"] = "J Transl Sci",
  ["journal of transnational law & policy"] = "J Transnatl Law Policy",
  ["journal of transnational law & policy / florida state university"] = "J Transnatl Law Policy",
  ["journal of transnational management development"] = "J. Transnat. Manage. Devel.",
  ["journal of transplant coordination"] = "J. Transpl. Coord.",
  ["journal of transplant coordination : official publication of the north american transplant coordinators organization (natco)"] = "J Transpl Coord",
  ["journal of transplantation"] = "J Transplant",
  ["journal of transport & health"] = "J Transp Health",
  ["journal of transport and land use"] = "J Transp Land Use",
  ["journal of transport economics and policy"] = "J. Transp. Econ. Pol.",
  ["journal of transport geography"] = "J Transp Geogr",
  ["journal of transportation engineering"] = "J. Transp. Eng.",
  ["journal of transportation engineering part a. systems"] = "J. Transp. Eng. Part A. Syst.",
  ["journal of transportation engineering part b. pavements"] = "J. Transp. Eng. Part B. Pavements",
  ["journal of transportation engineering. part a, systems"] = "J Transp Eng A Syst",
  ["journal of transportation safety & security"] = "J. Transp. Saf. Secur.",
  ["journal of transportation security"] = "J. Transp. Secur.",
  ["journal of transportation technologies"] = "J Transp Technol",
  ["journal of trauma"] = "J. Trauma",
  ["journal of trauma & dissociation : the official journal of the international society for the study of dissociation (issd)"] = "J Trauma Dissociation",
  ["journal of trauma & treatment"] = "J Trauma Treat",
  ["journal of trauma management & outcomes"] = "J Trauma Manag Outcomes",
  ["journal of trauma nursing"] = "J. Trauma Nurs.",
  ["journal of trauma nursing : the official journal of the society of trauma nurses"] = "J Trauma Nurs",
  ["journal of traumatic stress"] = "J Trauma Stress",
  ["journal of traumatic stress disorders & treatment"] = "J Trauma Stress Disord Treat",
  ["journal of travel medicine"] = "J Travel Med",
  ["journal of travel medicine : official publication of the international society of travel medicine and the asia pacific travel health association"] = "J Travel Med",
  ["journal of travel research"] = "J Travel Res",
  ["journal of tribology"] = "J. Tribol.",
  ["journal of tropical agriculture"] = "J Trop Agric",
  ["journal of tropical agriculture and food science"] = "J Trop Agric Food Sci",
  ["journal of tropical biodiversity and biotechnology"] = "J Trop Biodivers Biotechnol",
  ["journal of tropical biology & conservation"] = "J Trop Biol Conserv",
  ["journal of tropical crop science"] = "J Trop Crop Sci",
  ["journal of tropical diseases"] = "J Trop Dis",
  ["journal of tropical diseases & public health"] = "J Trop Dis Public Health",
  ["journal of tropical ecology"] = "J Trop Ecol",
  ["journal of tropical forest science"] = "J. Trop. For. Sci.",
  ["journal of tropical life science"] = "J Trop Life Sci",
  ["journal of tropical medicine"] = "J Trop Med",
  ["journal of tropical medicine and hygiene"] = "J. Trop. Med. Hyg.",
  ["journal of tropical meteorology"] = "J. Trop. Meteorol.",
  ["journal of tropical microbiology and biotechnology"] = "J. Trop. Microbiol. Biotechnol.",
  ["journal of tropical pediatrics"] = "J Trop Pediatr",
  ["journal of tropical pediatrics (1967)"] = "J Trop Pediatr (1967)",
  ["journal of tropical pediatrics (london, england : 1955)"] = "J Trop Pediatr (Lond)",
  ["journal of tropical pediatrics and african child health"] = "J. Trop. Pediatr. Afr. Child Health",
  ["journal of tropical pediatrics and environmental child health"] = "J. Trop. Pediatr. Environ. Child Health",
  ["journal of tsinghua university"] = "J. Tsinghua Univ.",
  ["journal of tsinghua university. science and technology. qinghua daxue xuebao. ziran kexue ban"] = "J. Tsinghua Univ.",
  ["journal of tsuda college"] = "J. Tsuda College",
  ["journal of tuberculosis research"] = "J Tuberc Res",
  ["journal of turbomachinery"] = "J. Turbomach.",
  ["journal of turbulence"] = "J. Turbul.",
  ["journal of turfgrass management"] = "J. Turfgrass Manage.",
  ["journal of ultrafine grained and nanostructured materials"] = "J. Ultrafine Grained Nanostruct. Mater.",
  ["journal of ultrasonography"] = "J Ultrason",
  ["journal of ultrasound"] = "J Ultrasound",
  ["journal of ultrasound in medicine"] = "J. Ultrasound Med.",
  ["journal of ultrasound in medicine : official journal of the american institute of ultrasound in medicine"] = "J Ultrasound Med",
  ["journal of ultrastructure and molecular structure research"] = "J. Ultrastruct. Mol. Struct. Res.",
  ["journal of ultrastructure research"] = "J. Ultrastruct. Res.",
  ["journal of ultrastructure research. supplement"] = "J Ultrastruct Res Suppl",
  ["journal of umm al-qura university for applied sciences"] = "J. Umm Al-Qura Univ. Appl. Sci.",
  ["journal of umm al-qura university for engineering and architecture"] = "J. Umm Al-Qura Univ. Eng. Archit.",
  ["journal of uncertain systems"] = "J. Uncertain Syst.",
  ["journal of uncertainty analysis and applications"] = "J. Uncertainty Anal. Appl.",
  ["journal of unconventional oil and gas resources"] = "J. Unconv. Oil Gas Resour.",
  ["journal of undergraduate chemistry research"] = "J Undergrad Chem Res",
  ["journal of undergraduate neuroscience education"] = "J. Undergrad. Neurosci. Educ.",
  ["journal of undergraduate neuroscience education : june : a publication of fun, faculty for undergraduate neuroscience"] = "J Undergrad Neurosci Educ",
  ["journal of undergraduate reports in physics"] = "J. Undergrad. Rep. Phys.",
  ["journal of undergraduate research (gainesville, fla.)"] = "J Undergrad Res (Gainesv)",
  ["journal of universal computer science : j. ucs"] = "J Univers. Comput Sci",
  ["journal of university of babylon for engineering sciences"] = "J. Univ. Babylon Eng. Sci.",
  ["journal of university of babylon for pure and applied sciences"] = "J. Univ. Babylon Pure Appl. Sci.",
  ["journal of university of science and technology beijing, mineral, metallurgy, material"] = "J. Univ. Sci. Technol. Beijing, Miner. Metall. Mater.",
  ["journal of university of science and technology of china. zhongguo kexue jishu daxue xuebao"] = "J. Univ. Sci. Technol. China",
  ["journal of unmanned vehicle systems"] = "J Unmanned Veh Syst",
  ["journal of uoeh"] = "J UOEH",
  ["journal of uoeh, university of occupational and environmental health"] = "J. UOEH",
  ["journal of urban affairs"] = "J Urban Aff",
  ["journal of urban design"] = "J Urban Des (Abingdon)",
  ["journal of urban ecology"] = "J. Urban Ecol.",
  ["journal of urban economics"] = "J Urban Econ",
  ["journal of urban futures"] = "J. Urban Futures",
  ["journal of urban health"] = "J. Urban Health",
  ["journal of urban health : bulletin of the new york academy of medicine"] = "J Urban Health",
  ["journal of urban history"] = "J Urban Hist",
  ["journal of urban law"] = "J Urban Law",
  ["journal of urban law / university of detroit"] = "J Urban Law",
  ["journal of urban mobility"] = "J. Urban Mobility",
  ["journal of urban planning and development"] = "J. Urban Plann. Dev.",
  ["journal of urban technology"] = "J. Urban Technol.",
  ["journal of urbanism"] = "J Urban",
  ["journal of urological nursing"] = "J Urol Nurs",
  ["journal of urology"] = "J. Urol.",
  ["journal of urology and research"] = "J Urol Res",
  ["journal of us-china medical science"] = "J US China Med Sci",
  ["journal of usability studies"] = "J Usability Stud",
  ["journal of vaccines & immunization"] = "J Vaccines Immun",
  ["journal of vaccines & vaccination"] = "J Vaccines Vaccin",
  ["journal of vacuum science & technology a"] = "J. Vac. Sci. Technol. A",
  ["journal of vacuum science & technology a: vacuum, surfaces, and films"] = "J. Vac. Sci. Technol. A: Vac. Surf. Films",
  ["journal of vacuum science & technology b"] = "J. Vac. Sci. Technol. B",
  ["journal of vacuum science & technology b: microelectronics and nanometer structures"] = "J. Vac. Sci. Technol. B: Microelectron. Nanometer Struct.",
  ["journal of vacuum science & technology b: microelectronics processing and phenomena"] = "J. Vac. Sci. Technol. B: Microelectron. Process. Phenom.",
  ["journal of vacuum science & technology, a: vacuum, surfaces, and films"] = "J. Vac. Sci. Technol., A",
  ["journal of vacuum science & technology, b: microelectronics and nanometer structures–processing, measurement, and phenomena"] = "J. Vac. Sci. Technol., B: Microelectron. Nanometer Struct.–Process., Meas., Phenom.",
  ["journal of vacuum science & technology, b: microelectronics processing and phenomena"] = "J. Vac. Sci. Technol., B",
  ["journal of vacuum science & technology, b: nanotechnology & microelectronics: materials, processing, measurement, & phenomena"] = "J. Vac. Sci. Technol., B: Nanotechnol. Microelectron.: Mater., Process., Meas., Phenom.",
  ["journal of vacuum science & technology. a, vacuum, surfaces, and films : an official journal of the american vacuum society"] = "J Vac Sci Technol A",
  ["journal of vacuum science & technology. b, microelectronics and nanometer structures : processing, measurement, and phenomena : an official journal of the american vacuum society"] = "J Vac Sci Technol B Microelectron Nanometer Struct Process Meas Phenom",
  ["journal of vacuum science and technology"] = "J. Vac. Sci. Technol.",
  ["journal of vacuum science and technology a"] = "J. Vac. Sci. Technol., A",
  ["journal of vacuum science and technology b"] = "J. Vac. Sci. Technol., B",
  ["journal of vacuum science and technology b: microelectronics and nanometer structures"] = "J. Vac. Sci. Technol., B",
  ["journal of vacuum science and technology. b, nanotechnology & microelectronics : materials, processing, measurement, & phenomena : jvst b"] = "J Vac Sci Technol B Nanotechnol Microelectron",
  ["journal of vascular access"] = "J. Vasc. Access",
  ["journal of vascular and endovascular surgery"] = "J Vasc Endovasc Surg",
  ["journal of vascular and interventional neurology"] = "J Vasc Interv Neurol",
  ["journal of vascular and interventional radiology"] = "J. Vasc. Interv. Radiol.",
  ["journal of vascular and interventional radiology : jvir"] = "J Vasc Interv Radiol",
  ["journal of vascular diagnostics"] = "J Vasc Diagn",
  ["journal of vascular diagnostics and interventions"] = "J Vasc Diagn Interv",
  ["journal of vascular medicine & surgery"] = "J Vasc Med Surg",
  ["journal of vascular nursing"] = "J. Vasc. Nurs.",
  ["journal of vascular nursing : official publication of the society for peripheral vascular nursing"] = "J Vasc Nurs",
  ["journal of vascular research"] = "J Vasc Res",
  ["journal of vascular surgery"] = "J Vasc Surg",
  ["journal of vascular surgery : official publication, the society for vascular surgery [and] international society for cardiovascular surgery, north american chapter"] = "J Vasc Surg",
  ["journal of vascular surgery cases"] = "J Vasc Surg Cases",
  ["journal of vascular surgery cases and innovative techniques"] = "J Vasc Surg Cases Innov Tech",
  ["journal of vascular surgery. venous and lymphatic disorders"] = "J Vasc Surg Venous Lymphat Disord",
  ["journal of vector borne diseases"] = "J Vector Borne Dis",
  ["journal of vector ecology"] = "J. Vector Ecol.",
  ["journal of vector ecology : journal of the society for vector ecology"] = "J Vector Ecol",
  ["journal of vegetation science"] = "J. Veg. Sci.",
  ["journal of vegetation science : official organ of the international association for vegetation science"] = "J Veg Sci",
  ["journal of venom research"] = "J Venom Res",
  ["journal of venomous animals and toxins including tropical diseases"] = "J. Venomous Anim. Toxins Incl. Trop. Dis.",
  ["journal of venomous animals including tropical diseases"] = "J. Venomous Anim. Toxins Incl. Trop. Dis.",
  ["journal of verbal learning and verbal behavior"] = "J Verbal Learning Verbal Behav",
  ["journal of verification, validation, and uncertainty quantification"] = "J Verif Valid Uncertain Quantif",
  ["journal of vertebrate biology"] = "J Vertebr Biol",
  ["journal of vertebrate paleontology"] = "J Vertebr Paleontol",
  ["journal of vestibular research"] = "J. Vestibular Res.",
  ["journal of vestibular research : equilibrium & orientation"] = "J Vestib Res",
  ["journal of veterinary behavior : clinical applications and research : official journal of : australian veterinary behaviour interest group, international working dog breeding association"] = "J Vet Behav",
  ["journal of veterinary cardiology : the official journal of the european society of veterinary cardiology"] = "J Vet Cardiol",
  ["journal of veterinary dentistry"] = "J Vet Dent",
  ["journal of veterinary diagnostic investigation"] = "J. Vet. Diagn. Invest.",
  ["journal of veterinary diagnostic investigation : official publication of the american association of veterinary laboratory diagnosticians, inc"] = "J Vet Diagn Invest",
  ["journal of veterinary education research"] = "J. Vet. Educ. Res.",
  ["journal of veterinary emergency and critical care"] = "J. Vet. Emerg. Crit. Care",
  ["journal of veterinary emergency and critical care (san antonio, tex. : 2001)"] = "J Vet Emerg Crit Care (San Antonio)",
  ["journal of veterinary internal medicine"] = "J Vet Intern Med",
  ["journal of veterinary internal medicine / american college of veterinary internal medicine"] = "J Vet Intern Med",
  ["journal of veterinary medical education"] = "J Vet Med Educ",
  ["journal of veterinary medical science"] = "J. Vet. Med. Sci.",
  ["journal of veterinary medicine"] = "J Vet Med",
  ["journal of veterinary medicine and research"] = "J Vet Med Res",
  ["journal of veterinary medicine series a"] = "J. Vet. Med. Ser. A",
  ["journal of veterinary medicine series b"] = "J. Vet. Med. Ser. B",
  ["journal of veterinary medicine. a, physiology, pathology, clinical medicine"] = "J Vet Med A Physiol Pathol Clin Med",
  ["journal of veterinary medicine. b, infectious diseases and veterinary public health"] = "J Vet Med B Infect Dis Vet Public Health",
  ["journal of veterinary pharmacology and therapeutics"] = "J Vet Pharmacol Ther",
  ["journal of veterinary research"] = "J Vet Res",
  ["journal of veterinary science"] = "J Vet Sci",
  ["journal of veterinary science & medical diagnosis"] = "J Vet Sci Med Diagn",
  ["journal of veterinary science & technology"] = "J Vet Sci Technol",
  ["journal of veterinary science (suwon-si, korea)"] = "J Vet Sci",
  ["journal of vibration and acoustics"] = "J. Vib. Acoust.",
  ["journal of vibration and control"] = "J. Vib. Control",
  ["journal of vibration and control : jvc"] = "J Vib Control",
  ["journal of vibration engineering & technologies"] = "J. Vib. Eng. Technol.",
  ["journal of vibroengineering"] = "J. Vibroeng.",
  ["journal of victorian culture : jvc"] = "J Vic Cult",
  ["journal of vinyl and additive technology"] = "J. Vinyl Add. Tech.",
  ["journal of vinyl technology"] = "J. Vinyl Tech.",
  ["journal of viral hepatitis"] = "J Viral Hepat",
  ["journal of virological methods"] = "J Virol Methods",
  ["journal of virology"] = "J Virol",
  ["journal of virology & antiviral research"] = "J Virol Antivir Res",
  ["journal of virtual worlds research"] = "J Virtual Worlds Res",
  ["journal of virus eradication"] = "J Virus Erad",
  ["journal of viruses"] = "J Viruses",
  ["journal of visceral surgery"] = "J Visc Surg",
  ["journal of vision"] = "J Vis",
  ["journal of visual art practice"] = "J Vis Art Pract",
  ["journal of visual communication and image representation"] = "J. Visual Commun. Image Represent.",
  ["journal of visual communication in medicine"] = "J Vis Commun Med",
  ["journal of visual impairment & blindness"] = "J Vis Impair Blind",
  ["journal of visual languages and computing"] = "J Vis Lang Comput",
  ["journal of visualization"] = "J. Visualization",
  ["journal of visualized experiments"] = "J. Visualized Exp.",
  ["journal of visualized experiments : jove"] = "J Vis Exp",
  ["journal of visualized surgery"] = "J Vis Surg",
  ["journal of vitaminology"] = "J. Vitaminol. (Kyoto)",
  ["journal of viticulture and enology"] = "J. Vitic. Enol.",
  ["journal of vitreoretinal diseases"] = "J Vitreoretin Dis",
  ["journal of vlsi signal processing"] = "J. VLSI Sig. Proc.",
  ["journal of vlsi signal processing systems for signal, image, and video technology"] = "J VLSI Signal Process Syst Signal Image Video Technol",
  ["journal of vocational behavior"] = "J Vocat Behav",
  ["journal of vocational rehabilitation"] = "J Vocat Rehabil",
  ["journal of voice"] = "J. Voice",
  ["journal of voice : official journal of the voice foundation"] = "J Voice",
  ["journal of volcanology and geothermal research"] = "J. Volcanol. Geotherm. Res.",
  ["journal of volcanology and seismology"] = "J. Volcanol. Seismolog.",
  ["journal of voluntary action research"] = "J Volunt Action Res",
  ["journal of volunteer administration"] = "J. Volunt. Adm.",
  ["journal of war & culture studies"] = "J War Cult Stud",
  ["journal of water and climate"] = "J. Water Clim.",
  ["journal of water and climate change"] = "J. Water Clim. Change",
  ["journal of water and environment technology"] = "J Water Environ Technol",
  ["journal of water and environmental nanotechnology"] = "J. Water Environ. Nanotechnol.",
  ["journal of water and health"] = "J Water Health",
  ["journal of water and land development"] = "J. Water Land Dev.",
  ["journal of water chemistry and technology"] = "J. Water Chem. Technol.",
  ["journal of water management modeling"] = "J. Water Manage. Model.",
  ["journal of water process engineering"] = "J. Water Process Eng.",
  ["journal of water resource and protection"] = "J. Water Resour. Prot.",
  ["journal of water resources planning and management"] = "J. Water Resour. Plann. Manage.",
  ["journal of water reuse and desalination"] = "J. Water Reuse Desalin.",
  ["journal of water sanitation and hygiene for development"] = "J. Water Sanit. Hyg. Dev.",
  ["journal of water supply: research and technology - aqua"] = "J. Water Supply Res. Technol. AQUA",
  ["journal of water technology"] = "J. Water Technol.",
  ["journal of water, sanitation, and hygiene for development : a journal of the international water association"] = "J Water Sanit Hyg Dev",
  ["journal of waterway port, coastal, and ocean engineering"] = "J. Waterw. Port Coastal Ocean Eng.",
  ["journal of waterway, port, coastal, and ocean engineering"] = "J Waterw Port Coast Ocean Eng",
  ["journal of web semantics"] = "J. Web Semant.",
  ["journal of wetland archaeology"] = "J. Wetland Archaeol.",
  ["journal of wetlands ecology"] = "J. Wetlands Ecol.",
  ["journal of wide bandgap materials"] = "J. Wide Bandgap Mater.",
  ["journal of wilderness medicine"] = "J Wilderness Med",
  ["journal of wildlife and biodiversity"] = "J Wildl Biodivers",
  ["journal of wildlife diseases"] = "J Wildl Dis",
  ["journal of wildlife management"] = "J. Wildl. Manage.",
  ["journal of wind engineering and industrial aerodynamics"] = "J. Wind Eng. Ind. Aerodyn.",
  ["journal of wine economics"] = "J. Wine Econ.",
  ["journal of wine research"] = "J Wine Res",
  ["journal of wireless mobile networks, ubiquitous computing, and dependable applications"] = "J Wirel Mob Netw Ubiquitous Comput Dependable Appl",
  ["journal of women & aging"] = "J Women Aging",
  ["journal of women and aging"] = "J. Women Aging",
  ["journal of women and minorities in science and engineering"] = "J. Women Minorities Sci. Eng.",
  ["journal of women's health"] = "J Womens Health",
  ["journal of women's health & gender-based medicine"] = "J Womens Health Gend Based Med",
  ["journal of women's health (2002)"] = "J Womens Health (Larchmt)",
  ["journal of women's health and development"] = "J Womens Health Dev",
  ["journal of women's health and law"] = "J Womens Heal",
  ["journal of women's health care"] = "J Womens Health Care",
  ["journal of women's health physical therapy"] = "J Womens Health Phys Therap",
  ["journal of women's health, issues & care"] = "J Womens Health Issues Care",
  ["journal of women's history"] = "J Womens Hist",
  ["journal of women, politics & policy"] = "J Women Polit Policy",
  ["journal of womens health"] = "J Womens Health",
  ["journal of women’s health"] = "J. Womens Health",
  ["journal of women’s health & gender-based medicine"] = "J Womens Health Gend Based Med",
  ["journal of women’s health (2002)"] = "J Womens Health (Larchmt)",
  ["journal of women’s health / the official publication of the society for the advancement of women’s health research"] = "J Womens Health",
  ["journal of women’s health and gender-based medicine"] = "J. Womens Health Gend. Based Med.",
  ["journal of women’s health and law"] = "J Womens Heal",
  ["journal of women’s history"] = "J Womens Hist",
  ["journal of wood chemistry and technology"] = "J. Wood Chem. Technol.",
  ["journal of wood science"] = "J. Wood Sci.",
  ["journal of workplace behavioral health"] = "J Workplace Behav Health",
  ["journal of world business"] = "J. World Bus.",
  ["journal of world history : official journal of the world history association"] = "J World Hist",
  ["journal of world prehistory"] = "J World Prehist",
  ["journal of world trade"] = "J. World Trade",
  ["journal of world trade law"] = "J. World Trade Law",
  ["journal of world's poultry research"] = "J Worlds Poult Res",
  ["journal of wound care"] = "J Wound Care",
  ["journal of wound, ostomy and continence nursing"] = "J. Wound. Ostomy Continence Nurs.",
  ["journal of wound, ostomy, and continence nursing : official publication of the wound, ostomy and continence nurses society"] = "J Wound Ostomy Continence Nurs",
  ["journal of wound, ostomy, and continence nursing : official publication of the wound, ostomy and continence nurses society / wocn"] = "J Wound Ostomy Continence Nurs",
  ["journal of wrist surgery"] = "J Wrist Surg",
  ["journal of writing research"] = "J Writ Res",
  ["journal of wscg"] = "J WSCG (Plzen)",
  ["journal of wuhan university"] = "J. Wuhan Univ. Natur. Sci. Ed.",
  ["journal of wuhan university of technology materials science edition"] = "J. Wuhan Univ. Technol.\tMater. Sci. Ed.",
  ["journal of wuhan university of technology. materials science edition"] = "J Wuhan Univ Technol Mater Sci Ed",
  ["journal of wuhan university. natural science edition. wuhan daxue xuebao. lixue ban"] = "J. Wuhan Univ. Natur. Sci. Ed.",
  ["journal of x-ray science and technology"] = "J Xray Sci Technol",
  ["journal of xenobiotics"] = "J Xenobiot",
  ["journal of xiangya medicine"] = "J Xiangya Med",
  ["journal of yangzhou university"] = "J. Yangzhou Univ. Nat. Sci. Ed.",
  ["journal of yeast and fungal research"] = "J Yeast Fungal Res",
  ["journal of yiyang teachers’ college"] = "J. Yiyang Teachers College",
  ["journal of yoga & physical therapy"] = "J Yoga Phys Ther",
  ["journal of young investigators"] = "J Young Investig",
  ["journal of young pharmacists : jyp"] = "J Young Pharm",
  ["journal of youth and adolescence"] = "J Youth Adolesc",
  ["journal of youth development : bridging research and practice"] = "J Youth Dev",
  ["journal of youth studies"] = "J Youth Stud",
  ["journal of yunnan university. natural sciences. yunnan daxue xuebao. ziran kexue ban"] = "J. Yunnan Univ. Nat. Sci.",
  ["journal of zhejiang university-science a applied physics & engineering"] = "J. Zhejiang Univ.-Sci. A",
  ["journal of zhejiang university-science b biomedicine & biotechnology"] = "J. Zhejiang Univ.-Sci. B",
  ["journal of zhejiang university. science"] = "J Zhejiang Univ Sci",
  ["journal of zhejiang university. science edition. zhejiang daxue xuebao. lixue ban"] = "J. Zhejiang Univ. Sci. Ed.",
  ["journal of zhejiang university. science. b"] = "J Zhejiang Univ Sci B",
  ["journal of zhejiang university. science. b."] = "J Zhejiang Univ Sci B",
  ["journal of zhengzhou university"] = "J. Zhengzhou Univ. Nat. Sci. Ed.",
  ["journal of zhengzhou university. engineering science. zhengzhou daxue xuebao. gongxue ban"] = "J. Zhengzhou Univ. Eng. Sci.",
  ["journal of zhengzhou university. natural science edition. zhengzhou daxue xuebao. lixue ban"] = "J. Zhengzhou Univ. Nat. Sci. Ed.",
  ["journal of zoo and aquarium research"] = "J Zoo Aquar Res",
  ["journal of zoo and wildlife medicine"] = "J. Zoo Wildl. Med.",
  ["journal of zoo and wildlife medicine : official publication of the american association of zoo veterinarians"] = "J Zoo Wildl Med",
  ["journal of zoo biology"] = "J Zoo Biol",
  ["journal of zoological systematics and evolutionary research"] = "J. Zool. Syst. Evol. Res.",
  ["journal of zoological systematics and evolutionary research = zeitschrift für zoologische systematik und evolutionsforschung"] = "J Zool Syst Evol Res",
  ["journal of zoology"] = "J Zool",
  ["journal of zoology (london)"] = "J. Zool. (London)",
  ["journal of zoology (london, england : 1987)"] = "J Zool (1987)",
  ["journal officiel de la republique francaise, edition des lois et decrets"] = "J Off Repub Fr Ed Lois Decrets",
  ["journal on artificial intelligence"] = "J. Artif. Intell.",
  ["journal on big data"] = "J. Big Data",
  ["journal on data semantics"] = "J Data Semant",
  ["journal on developmental disabilities = le journal sur les handicaps du dévelopment [sic.]"] = "J Dev Disabl",
  ["journal on excellence in college teaching"] = "J Excell Coll Teach",
  ["journal on internet of things"] = "J. Internet Things",
  ["journal on migration and human security"] = "J Migr Hum Secur",
  ["journal on multimodal user interfaces"] = "J. Multimodal User Interface",
  ["journal on new biological reports"] = "J New Biol Rep",
  ["journal on numerical methods and computer applications"] = "J. Numer. Methods Comput. Appl.",
  ["journal on satisfiability, boolean modeling and computation"] = "J. Satisf. Boolean Model. Comput.",
  ["journal sogc : journal of the society of obstetricians and gynaecologists of canada"] = "J SOGC",
  ["journal watch. aids clinical care"] = "J Watch AIDS Clin Care",
  ["journal water pollution control federation"] = "J. Water Pollut. Control Fed.",
  ["journal, academy of general dentistry"] = "J. Acad. Gen. Dent.",
  ["journal, academy of hospital administration"] = "J. Acad. Hosp. Adm.",
  ["journal, alabama dental association"] = "J. Ala. Dent. Assoc.",
  ["journal, american academy of gnathologic orthopedics"] = "J. Am. Acad. Gnathol. Orthop.",
  ["journal, american association for medical transcription"] = "J. Am. Assoc. Med. Transcr.",
  ["journal, american health care association"] = "J. Am. Health Care Assoc.",
  ["journal, american intra-ocular implant society"] = "J. Am. Intraocul. Implant Soc.",
  ["journal, american medical record association"] = "J. Am. Med. Rec. Assoc.",
  ["journal, american society for the study of orthodontics"] = "J. Am. Soc. Study Orthod.",
  ["journal, anglo-continental dental society"] = "J. Anglocont. Dent. Soc.",
  ["journal, association for healthcare philanthropy"] = "J. Assoc. Healthc. Philanthr.",
  ["journal, association for hospital medical education"] = "J. Assoc. Hosp. Med. Educ.",
  ["journal, association of official analytical chemists"] = "J. Assoc. Off. Anal. Chem.",
  ["journal, australian college of midwives"] = "J. Aust. Coll. Midwives",
  ["journal, california dental association"] = "J. Calif. Dent. Assoc.",
  ["journal, canadian dental association"] = "J. Can. Dent. Assoc.",
  ["journal, connecticut state dental association"] = "J. Conn. State Dent. Assoc.",
  ["journal, contra costa dental society"] = "J. Contra Costa Dent. Soc.",
  ["journal, district of columbia dental society"] = "J. Dist. Columbia Dent. Soc.",
  ["journal, forensic science society"] = "J. Forensic Sci. Soc.",
  ["journal, genetic engineering & biotechnology"] = "J Genet Eng Biotechnol",
  ["journal, indiana dental association"] = "J. Indiana Dent. Assoc.",
  ["journal, institute of sterile services management"] = "J. Inst. Sterile Serv. Manage.",
  ["journal, maine dental association"] = "J. Maine Dent. Assoc.",
  ["journal, mississippi dental association"] = "J. Miss. Dent. Assoc.",
  ["journal, national association for hospital development"] = "J. Natl. Assoc. Hosp. Dev.",
  ["journal, national association of private psychiatric hospitals"] = "J. Natl. Assoc. Priv. Psychiatr. Hosp.",
  ["journal, new hampshire dental society"] = "J. N. H. Dent. Soc.",
  ["journal, oklahoma dental association"] = "J. Okla. Dent. Assoc.",
  ["journal, oklahoma state dental association"] = "J. Okla. State Dent. Assoc.",
  ["journal, oklahoma state medical association"] = "J. Okla. State Med. Assoc.",
  ["journal, physical therapy education"] = "J Phys Ther Educ",
  ["journal, seattle-king county dental society"] = "J. Seattle. King Cty. Dent. Soc.",
  ["journal, south carolina medical association"] = "J. S. C. Med. Assoc.",
  ["journal, southern california dental assistants association"] = "J. South. Calif. Dent. Assistants Assoc.",
  ["journal, southern california dental association"] = "J. South. Calif. Dent. Assoc.",
  ["journal, southern california state dental association"] = "J. South. Calif. State Dent. Assoc.",
  ["journal, tennessee state dental association"] = "J. Tenn. State Dent. Assoc.",
  ["journal, water pollution control federation"] = "J. Water Pollut. Control Fed.",
  ["journal. american association of medical record librarians"] = "J Am Assoc Med Rec Libr",
  ["journal. american public health association"] = "J Am Public Health Assoc",
  ["journal. american society of sugar cane technologists. florida division"] = "J Am Soc Sugar Cane Technol",
  ["journal. assam science society"] = "J Assam Science Society",
  ["journal. association of american medical colleges"] = "J Assoc Am Med Coll",
  ["journal. bengal tuberculosis association"] = "J Bengal Tuberc Assoc",
  ["journal. boston society of medical sciences"] = "J Boston Soc Med Sci",
  ["journal. bowman gray school of medicine"] = "J Bowman Gray Sch Med Wake For Coll",
  ["journal. children's asthma research institute and hospital (denver, colo.)"] = "J Child Asthma Res Inst Hosp Denver",
  ["journal. children’s asthma research institute and hospital, denver"] = "J Child Asthma Res Inst Hosp Denver",
  ["journal. colorado state dental association"] = "J Colo State Dent Assoc",
  ["journal. council for scientific and industrial research (australia)"] = "J Counc Sci Ind Res",
  ["journal. elisha mitchell scientific society (chapel hill, n.c.)"] = "J Elisha Mitchell Sci Soc Chapel Hill N C",
  ["journal. houston district dental society"] = "J Houston Dist Dent Soc",
  ["journal. hôtel-dieu de montréal"] = "J Hotel Dieu Montreal",
  ["journal. indianapolis district dental society"] = "J Indianap Dist Dent Soc",
  ["journal. institution of sanitary engineers (great britain)"] = "J Inst Sanit Eng Lond",
  ["journal. iowa state medical society"] = "J Iowa State Med Soc",
  ["journal. kansas city (mo.) southwest clinical society"] = "J Kans City Mo Southwest Clin Soc",
  ["journal. louisiana. state university and agricultural and mechanical college. school of medicine"] = "J La State Univ Agric Mech Coll",
  ["journal. massachusetts association of boards of health"] = "J Mass Assoc Boards Health",
  ["journal. medical association of eire"] = "J Med Assoc Eire",
  ["journal. mercer dental society, trenton"] = "J Mercer Dent Soc",
  ["journal. michigan state dental society"] = "J Mich State Dent Soc",
  ["journal. missouri state dental association"] = "J Mo State Dent Assoc",
  ["journal. national analgesia society"] = "J Natl Analg Soc",
  ["journal. national malaria society"] = "J Natl Malar Soc",
  ["journal. national malaria society (u.s.)"] = "J Natl Malar Soc",
  ["journal. national proctologic association"] = "J Natl Proctol Assoc",
  ["journal. obstetric and gynaecological society of northern india"] = "J Obstet Gynaecol Soc North India",
  ["journal. ohio state dental society"] = "J Ohio State Dent Soc",
  ["journal. omaha mid-west clinical society"] = "J Omaha Midwest Clin Soc",
  ["journal. oregon state dental association"] = "J Or State Dent Assoc",
  ["journal. palestine arab medical association"] = "J Palest Arab Med Assoc",
  ["journal. royal microscopical society (great britain)"] = "J R Microsc Soc",
  ["journal. royal sanitary institute (great britain)"] = "J R Sanit Inst",
  ["journal. san antonio district dental society"] = "J San Antonio Dent Soc",
  ["journal. second district dental society of the state of new york"] = "J Second Dist Dent Soc State N. Y.",
  ["journal. tennessee academy of science"] = "J Tenn Acad Sci",
  ["journal. washington academy of sciences, washington, d. c"] = "J Wash Acad Sci",
  ["journalen sykepleien"] = "J Sykepleien",
  ["journalism & communication monographs"] = "Journal Commun Monogr",
  ["journalism & mass communication quarterly"] = "Journal Mass Commun Q",
  ["journalism (london, england)"] = "Journalism (Lond)",
  ["journalism conference : [proceedings]. journalism conference"] = "Journal Conf Workshop ADA",
  ["journalism conference and workshop"] = "J Conf Workshop",
  ["journalism history"] = "Journal Hist",
  ["journalism monographs (austin, tex.)"] = "Journalism Monogr",
  ["journalism studies"] = "Journal Stud",
  ["journals of gerontology. series a, biological sciences and medical sciences"] = "J. Gerontol. A. Biol. Sci. Med. Sci.",
  ["journals of gerontology. series b, psychological sciences and social sciences"] = "J. Gerontol. B. Psychol. Sci. Soc. Sci.",
  ["journees annuelles de diabetologie de l hotel-dieu"] = "Journ Annu Diabetol Hotel Dieu",
  ["journees annuelles de diabetologie de l’hotel-dieu"] = "Journ. Annu. Diabetol. Hotel Dieu",
  ["journees equations aux derives partielles"] = "Journ. Equ. Deriv. Partielles",
  ["journées annuelles de diabétologie de l'hôtel-dieu"] = "Journ Annu Diabetol Hotel Dieu",
  ["jove chemistry"] = "JoVE Chem.",
  ["jōsai shika daigaku kiyō. the bulletin of the josai dental university"] = "Josai Shika Daigaku Kiyo",
  ["jōshi eiyō daigaku kiyō"] = "Joshi Eiyo Daigaku Kiyo",
  ["jp journal of biostatistics"] = "JP J Biostat",
  ["jpc journal of planar chromatography modern tlc"] = "JPC J. Planar Chromatogr. - Mod. TLC",
  ["jpen. journal of parenteral and enteral nutrition"] = "JPEN J Parenter Enteral Nutr",
  ["jpen: journal of parenteral and enteral nutrition"] = "JPEN J. Parenter. Enteral Nutr.",
  ["jpgn reports"] = "JPGN Rep",
  ["jphys energy"] = "JPhys Energy",
  ["jphys materials"] = "JPhys Mater.",
  ["jphys photonics"] = "JPhys Photonics",
  ["jpma. journal of the pakistan medical association"] = "JPMA J Pak Med Assoc",
  ["jpma. the journal of the pakistan medical association"] = "J Pak Med Assoc",
  ["jpma: journal of the pakistan medical association"] = "JPMA J. Pak. Med. Assoc.",
  ["jpo: journal of practical orthodontics"] = "JPO J. Pract. Orthod.",
  ["jpo: the journal of practical orthodontics"] = "JPO J Pract Orthod",
  ["jpras open"] = "JPRAS Open",
  ["jps conference proceedings"] = "JPS Conf Proc",
  ["jpss. journal of probability and statistical science"] = "JPSS J. Probab. Stat. Sci.",
  ["jrsm cardiovascular disease"] = "JRSM Cardiovasc Dis",
  ["jrsm open"] = "JRSM Open",
  ["jrsm short reports"] = "JRSM Short Rep",
  ["jsac grapevine. joint strategy and action committee (new york, n.y.)"] = "JSAC Grapevine",
  ["jses international"] = "JSES Int",
  ["jses open access"] = "JSES Open Access",
  ["jsiam letters"] = "JSIAM Lett.",
  ["jsls : journal of the society of laparoendoscopic surgeons"] = "JSLS",
  ["jsls : journal of the society of laparoendoscopic surgeons / society of laparoendoscopic surgeons"] = "JSLS",
  ["jsls: journal of the society of laparoendoscopic surgeons"] = "JSLS",
  ["jsm allergy and asthma"] = "JSM Allergy Asthma",
  ["jsm alzheimer's disease and related dementia"] = "JSM Alzheimers Dis Relat Dement",
  ["jsm atherosclerosis"] = "JSM Atheroscler",
  ["jsm biochemistry and molecular biology"] = "JSM Biochem Mol Biol",
  ["jsm bioinformatics, genomics and preteomics"] = "JSM Bioinform Genom Proteom",
  ["jsm biomarkers"] = "JSM Biomark",
  ["jsm biomedical imaging data papers"] = "JSM Biomed Imaging Data Pap",
  ["jsm biotechnology & biomedical engineering"] = "JSM Biotechnol Biomed Eng",
  ["jsm cell & developmental biology"] = "JSM Cell Dev Biol",
  ["jsm chemistry"] = "JSM Chem",
  ["jsm dental surgery"] = "JSM Dent Surg",
  ["jsm microbiology"] = "JSM Microbiol",
  ["jsm neurosurgery and spine"] = "JSM Neurosurg Spine",
  ["jsm tropical medicine and research"] = "JSM Trop Med Res",
  ["jsme international journal series a: mechanics and material engineering"] = "JSME Int J., Ser. A",
  ["jsme international journal series b: fluids and thermal engineering"] = "JSME Int J., Ser. B",
  ["jsme international journal series c: mechanical systems machine elements and manufacturing"] = "JSME Int J., Ser. C",
  ["jss research series in statistics"] = "JSS Res. Ser. Stat.",
  ["jtcvs techniques"] = "JTCVS Tech",
  ["jto clinical and research reports"] = "JTO Clin Res Rep",
  ["judaica librarianship"] = "Jud Librariansh",
  ["judgment and decision making"] = "Judgm Decis Mak",
  ["jugoslavenska ginekologija i opstetricija"] = "Jugosl Ginekol Opstet",
  ["jugoslavenska ginekologija i perinatologija"] = "Jugosl Ginekol Perinatol",
  ["jugoslovenski pregled; informativno dokumentarni prirucnik o jugoslaviji"] = "Jugosl Pregl",
  ["jugoslovenski pregled; informativno dokumentarni priručnik o jugoslaviji"] = "Jugosl Pregl",
  ["julkaisu. suomen naishammaslaakarit ryhma, helsingfors"] = "Julk Suom Naishammaslaak Ryhma",
  ["julkaisu. suomen naishammaslääkärit ryhmä, helsingfors"] = "Julk Suom Naishammaslaak Ryhma",
  ["jun shi yi xue ke xue yuan yuan kan"] = "Bulletin of the Academy of Military Medical Sciences|Chun Shih I Hsueh Ko Hsueh Yuan Yuan Kan",
  ["jun shi yi xue ke xue yuan yuan kan = bulletin of the academy of military medical sciences"] = "Chun Shih I Hsueh Ko Hsueh Yuan Yuan Kan",
  ["jun wu xi tong"] = "Mycosystema / Zhongguo ke xue yuan Wei sheng wu yan jiu suo. Zhongguo jun wu xue hui zhu ban|Jun Wu Xi Tong",
  ["jun wu xi tong = mycosystema"] = "Jun Wu Xi Tong",
  ["jun wu xue bao = mycosystema"] = "Jun Wu Xue Bao",
  ["jundishapur journal of microbiology"] = "Jundishapur J Microbiol",
  ["jundishapur journal of natural pharmaceutical products"] = "Jundishapur J Nat Pharm Prod",
  ["junior officer advisory group : a newsletter for junior officers by junior officers"] = "Jr Off Advis Group",
  ["juniper online journal material science"] = "Juniper Online J Mater Sci",
  ["juniper online journal of public health"] = "Juniper Online J Public Health",
  ["junwu yanjiu"] = "Junwu Yanjiu",
  ["juris doctor"] = "Juris Dr",
  ["jurist (washington, d.c.)"] = "Jurist",
  ["jurnal biologi"] = "J Biol (Denpasar)",
  ["jurnal biologi indonesia"] = "J Biologi Indones",
  ["jurnal bios logos"] = "J Bios Logos",
  ["jurnal biota"] = "J Biota",
  ["jurnal entomologi indonesia"] = "J Entomol Indones",
  ["jurnal fitopatologi = indonesian journal of phytopathology"] = "J Fitopatologi",
  ["jurnal iktiologi indonesia"] = "J Iktiologi Indones",
  ["jurnal kedokteran hewan"] = "J Kedokt Hewan (Banda Aceh)",
  ["jurnal kelautan"] = "J Kelaut",
  ["jurnal mikologi indonesia"] = "J Mikol Indones",
  ["jurnal natur indonesia : wacana sains indonesia"] = "J Nat Indones",
  ["jurnal penelitian tanaman industri = industrial crops research journal"] = "J Penelit Tanam Ind",
  ["jurnal pengolahan hasil perikanan indonesia"] = "J Pengolah Has Perikan Indones",
  ["jurnal teknologi"] = "J Teknol",
  ["justice quarterly : jq"] = "Justice Q",
  ["justice research and policy : journal of the justice research and statistics association : jrp"] = "Justice Res Policy",
  ["justus liebigs annalen der chemie"] = "Justus Liebigs Ann Chem",
  ["juvenile & family court journal"] = "Juv Fam Court J",
  ["jvs-vascular science"] = "JVS Vasc Sci",
  ["jyske samlinger"] = "Jyske Saml",
  ["jábega. revista de la disputación provincial de málaga"] = "Jabega",
  ["k [reports]. u.s. atomic energy commission"] = "K Rep",
  ["k reports"] = "K Rep.",
  ["k. k. birla academy monographs on the cultural, scientific and technological heritage of india"] = "Birla Acad. Monogr. Cult. Sci. Technol. Herit. India",
  ["kadmos. zeitschrift für vor- und frühgriechische epigraphik"] = "Kadmos",
  ["kafkas üniversitesi veteriner fakültesi dergisi"] = "Kafkas Univ Vet Fak Derg",
  ["kagaku (kyoto)"] = "Kagaku (Kyoto)",
  ["kagaku kogaku"] = "Kagaku Kogaku",
  ["kagaku kogaku ronbunshu"] = "Kagaku Kogaku Ronbunshu",
  ["kagaku to seibutsu"] = "Kagaku to Seibutsu",
  ["kagaku to seibutsu. [chemistry and biology]"] = "Kagaku To Seibutsu",
  ["kagakushi kenkyu. [journal of the history of science, japan"] = "Kagakushi Kenkyu",
  ["kagoshima daigaku shigakubu kiyō = annals of kagoshima university dental school"] = "Kagoshima Daigaku Shigakubu Kiyo",
  ["kaibogaku zasshi (journal of anatomy)"] = "Kaibogaku Zasshi",
  ["kaibogaku zasshi. journal of anatomy"] = "Kaibogaku Zasshi",
  ["kaiin dayori. [membership news]. nippon kontakuto renzu gakkai"] = "Kaiin Dayori Nihon Kontakuto Renzu Gakkai",
  ["kairos. journal of philosophy & science"] = "Kairos",
  ["kairos. zeitschrift für judaistik und religionswissenschaft"] = "Kairos",
  ["kajok kyehoek nonjip. journal of family planning studies"] = "Kajok Kyehoek Nonjip",
  ["kako gijutsu (osaka)"] = "Kako Gijutsu (Osaka)",
  ["kaku igaku (japanese journal of nuclear medicine)"] = "Kaku Igaku",
  ["kaku igaku [japanese journal of nuclear medicine]"] = "Kaku Igaku",
  ["kaku igaku. the japanese journal of nuclear medicine"] = "Kaku Igaku",
  ["kalapodi. ergebnisse der ausgrabungen im heiligtum der artemis und des apollon von hyampolis in der antiken phokis"] = "Kalapodi",
  ["kalathos. revista del seminario de arqueología y etnología turolense, universidad de teruel"] = "Kalathos",
  ["kalpa publications in computing"] = "Kalpa Publ. Comput.",
  ["kalós. arte in sicilia"] = "Kalos",
  ["kamper almanak"] = "Kampener Alm",
  ["kan, tan, sui"] = "Kan, Tan, Sui",
  ["kanagawa kenritsu hakubutsukan kenkyū hōkoku. shizen kagaku = bulletin of the kanagawa prefectural museum. natural science"] = "Kanagawa Kenritsu Hakubutsukan Kenkyu Hokoku Shizen Kagaku",
  ["kanagawa shigaku (journal of the kanagawa odontological society)"] = "Kanagawa Shigaku",
  ["kanagawa shigaku. the journal of the kanagawa odontological society"] = "Kanagawa Shigaku",
  ["kanagawa university"] = "Rep. Fac. Eng. Kanagawa Univ.",
  ["kanazawa ika daigaku zasshi"] = "Journal of Kanazawa Medical University|Kanazawa Ika Daigaku Zasshi",
  ["kanazawa ika daigaku zasshi = journal of kanazawa medical university"] = "Kanazawa Ika Daigaku Zasshi",
  ["kanazawa university"] = "Ann. Sci. Kanazawa Univ.",
  ["kango gijutsu (japanese journal of nursing art)"] = "Kango Gijutsu",
  ["kango kenkyu (japanese journal of nursing research)"] = "Kango Kenkyu",
  ["kango kenkyu. the japanese journal of nursing research"] = "Kango Kenkyu",
  ["kango kyoiku (japanese journal of nurses’ education)"] = "Kango Kyoiku",
  ["kango kyoshitsu (japanese journal of the nursing class)"] = "Kango Kyoshitsu",
  ["kango kyoshitsu. [nursing classroom]"] = "Kango Kyoshitsu",
  ["kango tenbo (japanese journal of nursing science)"] = "Kango Tenbo",
  ["kango tenbo. the japanese journal of nursing science"] = "Kango Tenbo",
  ["kango tenbō. the japanese journal of nursing science"] = "Kango Tenbo",
  ["kangogaku zasshi"] = "Kangogaku Zasshi",
  ["kangogaku zasshi (japanese journal of nursing)"] = "Kangogaku Zasshi",
  ["kanho hakhoe chi (journal of nurses’ academic society)"] = "Kanho Hakhoe Chi",
  ["kanho hakhoe chi [the journal of nurses academic society]"] = "Kanho Hakhoe Chi",
  ["kanhohak t'amgu"] = "Kanhohak Tamgu",
  ["kanhohak t’amgu"] = "Kanhohak Tamgu",
  ["kankyo kagaku sogo kenkyujo nenpo"] = "Kankyo Kagaku Sogo Kenkyujo Nenpo",
  ["kansai byōchūgai kenkyūkai hō"] = "Kansai Byochugai Kenkyukaiho",
  ["kansai daigaku keizai ronshu"] = "Kansai Daigaku Keizai Ronshu",
  ["kansai daigaku keizai ronshū"] = "Kansai Daigaku Keizai Ronshu",
  ["kansai university"] = "Tech. Rep. Kansai Univ.",
  ["kansantaloudellinen aikakauskirja"] = "Kansant. Aikak.",
  ["kansas history"] = "Kans Hist",
  ["kansas journal of medicine"] = "Kans J Med",
  ["kansas medicine"] = "Kans. Med.",
  ["kansas medicine : the journal of the kansas medical society"] = "Kans Med",
  ["kansas nurse"] = "Kans. Nurse",
  ["kansas statutes annotated : duly arranged, numbered, annotated, and indexed with histories, notes, and references as required by law / compiled and edited by franklin corrick, revisor of statutes. kansas"] = "Kans Statut Annot Kans",
  ["kansas statutes annotated : duly arranged, numbered, annotated, and indexed with histories, notes, and references as required by law. kansas"] = "Kans Statut Annot Kans",
  ["kansenshogaku zasshi (journal of the japanese association for infectious diseases)"] = "Kansenshogaku Zasshi",
  ["kansenshogaku zasshi. journal of the japanese association forinfectious diseases"] = "Kansenshogaku Zasshi",
  ["kansenshogaku zasshi. the journal of the japanese association for infectious diseases"] = "Kansenshogaku Zasshi",
  ["kansenshōgaku zasshi. the journal of the japanese association for infectious diseases"] = "Kansenshogaku Zasshi",
  ["kant-studien: philosophische zeitschrift"] = "KantStud",
  ["kantou tousan byougaichuu kenkyuukaihou"] = "Kantou Tousan Byougaichuu Kenkyuukaihou",
  ["kanzō = acta hepatologica japonica"] = "Kanzo",
  ["kao gu ren lei xue kan"] = "Kao Ku Jen Lei Hsueh Kan",
  ["kao-hsiung i hsueh ko hsueh tsa chih [kaohsiung journal of medical sciences]"] = "Kao Hsiung I Hsueh Ko Hsueh Tsa Chih",
  ["kaohsiung journal of medical sciences"] = "Kaohsiung J. Med. Sci.",
  ["karachi university"] = "Karachi Univ. J. Sci.",
  ["karbala international journal of modern science"] = "Karbala Int. J. Mod. Sci.",
  ["kardiochirurgia i torakochirurgia polska = polish journal of cardio-thoracic surgery"] = "Kardiochir Torakochirurgia Pol",
  ["kardiologia polska"] = "Kardiol Pol",
  ["karthago. revue d’archéologie mediterranéenne"] = "Karthago",
  ["kassenzahnarzt; colloquium med dent"] = "Kassenzahnarzt Colloq. Med. Dent.",
  ["kat'ollik taehak uihakpu nonmunjip"] = "Katollik Taehak Uihakpu Nonmunjip",
  ["kathmandu university medical journal"] = "Kathmandu Univ. Med. J.",
  ["kathmandu university medical journal (kumj)"] = "Kathmandu Univ Med J (KUMJ)",
  ["katilolehti (tidskrift for barnmorskor)"] = "Katilolehti",
  ["kat’ollik taehak uihakpu nonmunjip"] = "Katollik Taehak Uihakpu Nonmunjip",
  ["kauchuk i rezina"] = "Kauch. Rezina",
  ["kautschuk gummi kunststoffe"] = "Kautsch. Gummi Kunstst.",
  ["kavkazskij ėntomologičeskij bjulleten = caucasian entomological bulletin"] = "Kavk Entomol Bjul",
  ["kazanskii meditsinskii zhurnal"] = "Kazan Med Zh",
  ["kazanskiĭ meditsinskiĭ zhurnal"] = "Kazan Med Zh",
  ["kazanskiy gosudarstvennyy universitet"] = "Gravitatsiya i Teor. Otnositel\\cprime nosti",
  ["kazanskiy universitet"] = "Issled. Prikl. Mat.",
  ["kazı sonuçları toplantısı"] = "KST",
  ["kdd : proceedings. international conference on knowledge discovery & data mining"] = "KDD",
  ["ke ji dao bao = science & technology review"] = "Ke Ji Dao Bao",
  ["ke xue fa zhan"] = "Ke Xue Fa Zhan",
  ["keele cognition seminars"] = "Keele Cognition Sem.",
  ["keeping posted"] = "Keep Posted",
  ["keio economic studies"] = "Keio Econ Stud",
  ["keio journal of medicine"] = "Keio J. Med.",
  ["keio science and technology reports"] = "Keio Sci. Tech. Rep.",
  ["keio university"] = "Hiyoshi Rev. Natur. Sci.",
  ["keisei geka (japanese journal of plastic and reconstructive surgery)"] = "Keisei Geka",
  ["keisei geka. plastic & reconstructive surgery"] = "Keisei Geka",
  ["kek proceedings"] = "KEK Proc.",
  ["kekkaku (tuberculosis)"] = "Kekkaku",
  ["kekkaku : [tuberculosis]"] = "Kekkaku",
  ["kemia - kemi"] = "Kem.-Kemi",
  ["kemija u industriji"] = "Kem. Ind.",
  ["kemikaru enjiniyaringu"] = "Kemikaru Enjiniyaringu",
  ["kenchreai. eastern port of corinth. results of investigations by the university of chicago and indiana university for the american school of classical studies at athens"] = "Kenchreai",
  ["kendall’s library of statistics"] = "Kendall’s Lib. Statist.",
  ["kennedy institute of ethics journal"] = "Kennedy Inst Ethics J",
  ["kent archaeological review"] = "KentAR",
  ["kentucky bench & bar"] = "Ky Bench Bar",
  ["kentucky dental journal"] = "Ky Dent J",
  ["kentucky folklore record"] = "Ky Folkl Rec",
  ["kentucky hospitals magazine"] = "Ky Hosp Mag",
  ["kentucky journal of communication"] = "Ky J Commun",
  ["kentucky law journal (lexington, ky.)"] = "KY Law J",
  ["kentucky medical journal"] = "Ky Med J",
  ["kentucky nurse"] = "Ky Nurse",
  ["kentucky nurses association newsletter"] = "Ky. Nurses Assoc. Newsl.",
  ["kenya nursing journal"] = "Kenya Nurs J",
  ["kenya statistical digest"] = "Kenya Stat Dig",
  ["keos. results of excavations conducted by the university of cincinnati under the auspices of the american school of classical studies at athens"] = "Keos",
  ["kerameikos. ergebnisse der ausgrabungen"] = "Kerameikos",
  ["keramische zeitschrift"] = "Keram. Z.",
  ["kernos. revue internationale et pluridisciplinaire de religion grecque antique"] = "Kernos",
  ["kerteszet es szoleszet"] = "Kertesz Szolesz",
  ["kew bulletin"] = "Kew Bull.",
  ["kexue tongbao"] = "Kexue Tongbao (Chinese)",
  ["key engineering materials"] = "Key Eng. Mater.",
  ["key texts: classic studies in the history of ideas"] = "Key Texts Classic Stud. Hist. Ideas",
  ["keystone folklore quarterly"] = "Keystone Folkl Q",
  ["kgk, kautschuk gummi kunststoffe"] = "KGK, Kautsch. Gummi Kunstst.",
  ["khar\\cprime kovskiy universitet"] = "Vestnik Khar\\cprime kov. Univ.",
  ["khayyam journal of mathematics"] = "Khayyam J. Math.",
  ["khigiena i zdraveopazvane"] = "Hig Zdraveopaz",
  ["khimicheskaya fizika"] = "Khim. Fiz.",
  ["khimicheskaya promyshlennost (st. petersburg, russian federation)"] = "Khim. Prom-st. (St. Petersburg, Russ. Fed.)",
  ["khimicheskaya promyshlennost segodnya"] = "Khim. Prom-st. Segodnya",
  ["khimicheskie volokna"] = "Khim. Volokna",
  ["khimicheskoe i neftyanoe mashinostroenie"] = "Khim. Neft. Mashinostr.",
  ["khimiia i meditsina"] = "Khimiia Meditsina",
  ["khimiia prirodnykh soedinenii"] = "Him Prir Soedin",
  ["khimiia prirodnykh soedineniĭ"] = "Him Prir Soedin",
  ["khimiya drevesiny"] = "Khim. Drev.",
  ["khimiya i tekhnologiya topliv i masel"] = "Khim. Tekhnol. Topl. Masel",
  ["khimiya i tekhnologiya vody"] = "Khim. Tekhnol. Vody",
  ["khimiya tverdogo topliva (moscow, russian federation)"] = "Khim. Tverd. Topl. (Moscow, Russ. Fed.)",
  ["khimiya v interesakh ustoichivogo razvitiya"] = "Khim. Interesakh Ustoich. Razvit.",
  ["khimiya vysokikh energii"] = "Khim. Vys. Energ.",
  ["khimizatsiya sel’skogo khozyaistva"] = "Khim. Sel’sk. Khoz.",
  ["khoa học kỹ thuật thú y"] = "Khoa Hoc Ky Thuat Thu Y",
  ["khoa học và công nghệ"] = "Khoa Hoc Va Cong Nghe",
  ["ki kunstliche intelligenz"] = "KI Kunstliche Intell.",
  ["kias springer series in mathematics"] = "KIAS Springer Ser. Math.",
  ["kichin, kitosan kenkyu = chitin and chitosan research"] = "Kichin Kitosan Kenkyu",
  ["kidney & blood pressure research"] = "Kidney Blood Press Res",
  ["kidney and blood pressure research"] = "Kidney Blood Pressure Res.",
  ["kidney cancer (clifton, va.)"] = "Kidney Cancer",
  ["kidney cancer journal : official journal of the kidney cancer association"] = "Kidney Cancer J",
  ["kidney diseases (basel, switzerland)"] = "Kidney Dis (Basel)",
  ["kidney international"] = "Kidney Int",
  ["kidney international reports"] = "Kidney Int Rep",
  ["kidney international supplements"] = "Kidney Int Suppl (2011)",
  ["kidney international. supplement"] = "Kidney Int Suppl",
  ["kidney medicine"] = "Kidney Med",
  ["kidney research and clinical practice"] = "Kidney Res Clin Pract",
  ["kieler beitrage zur geschichte der medizin und pharmazie"] = "Kieler Beitr Gesch Med Pharm",
  ["kieler beiträge zur geschichte der medizin und pharmazie"] = "Kieler Beitr Gesch Med Pharm",
  ["kieler geographische schriften"] = "Kiel Geogr Schr",
  ["kieler studien"] = "Kiel. Stud.",
  ["kievskiy gosudarstvennyy universitet"] = "Issled. Operatsiy i ASU",
  ["kievskiy politekhnicheskiy institut"] = "Adapt. Sistemy Avtomat. Upravleniya",
  ["kievskiy universitet"] = "Vestnik Kiev. Univ. Model. Optim. Slozhn. Sist.",
  ["kikan riron-keizaigaku"] = "Kikan Riron Keizaigaku",
  ["killi-data series"] = "Killidata Ser",
  ["kinderarztliche praxis"] = "Kinderarztl. Prax.",
  ["kinderärztliche praxis"] = "Kinderarztl Prax",
  ["kinderkrankenschwester : organ der sektion kinderkrankenpflege"] = "Kinderkrankenschwester",
  ["kinderkrankenschwester : organ der sektion kinderkrankenpflege / deutsche gesellschaft fur sozialpadiatrie und deutsche gesellschaft fur kinderheilkunde"] = "Kinderkrankenschwester",
  ["kinematics and physics of celestial bodies"] = "Kinematics Phys. Celestial Bodies",
  ["kinesiology (zagreb, croatia)"] = "Kinesiology (Zagreb)",
  ["kinesiology review (champaign, ill.)"] = "Kinesiol Rev (Champaign)",
  ["kinetic and related models"] = "Kinet. Rel. Models",
  ["kinetics and catalysis"] = "Kinet. Catal.",
  ["kinetika i kataliz"] = "Kinet. Katal.",
  ["kinetoplastid biology and disease"] = "Kinetoplastid Biol Dis",
  ["king faisal specialist hospital medical journal"] = "King Faisal Spec. Hosp. Med. J.",
  ["king's college hospital gazette"] = "Kings Coll Hosp Gaz",
  ["king's college hospital gazette. clinical supplement"] = "Kings Coll Hosp Gaz Clin Suppl",
  ["king's law journal : klj"] = "Kings Law J",
  ["kinjin kenkyusho kenkyu hokoku"] = "Kinjin Kenkyusho Kenkyu Hokoku",
  ["kinki university"] = "J. Fac. Sci. Tech. Kinki Univ.",
  ["kinki university series on quantum computing"] = "Kinki Univ. Ser. Quantum Comput.",
  ["kino zairyo"] = "Kino Zairyo",
  ["kiph bulletin"] = "KIPH Bull",
  ["kirk-othmer encyclopedia of chemical technology"] = "Kirk-Othmer Encycl. Chem. Technol.",
  ["kiryat sefer"] = "Kiryat Sefer",
  ["kisaengch'unghak chapchi. the korean journal of parasitology"] = "Kisaengchunghak Chapchi",
  ["kisaengchunghak chapchi (korean journal of parasitology)"] = "Kisaengchunghak Chapchi",
  ["kisaengch’unghak chapchi. the korean journal of parasitology"] = "Kisaengchunghak Chapchi",
  ["kiseichugaku zasshi. japanese journal of parasitology"] = "Kiseichugaku Zasshi",
  ["kiserletes orvostudomany"] = "Kiserl Orvostud",
  ["kiso shinrigaku kenkyu"] = "Jpn J Psychon Sci",
  ["kiso shinrigaku kenkyū"] = "Jpn J Psychon Sci",
  ["kita kanto igaku. kitakanto medical journal"] = "Kita Kanto Igaku",
  ["kita kantō igaku. kitakanto medical journal"] = "Kita Kanto Igaku",
  ["kitasato archives of experimental medicine"] = "Kitasato Arch. Exp. Med.",
  ["kiyo. [reports]. kyoto daigaku. kekkaku kenkyusho"] = "Kyoto Daigaku Kekkaku Kenkyusho Kiyo",
  ["kiyo. [reports]. kyōto daigaku. kekkaku kenkyūjo"] = "Kyoto Daigaku Kekkaku Kenkyusho Kiyo",
  ["kísérletes orvostudomány"] = "Kiserl Orvostud",
  ["klagenfurter beitrage zur didaktik der mathematik"] = "Klagenf. Beitr. Didakt. Math.",
  ["klassische texte der wissenschaft"] = "Klass. Texte Wiss.",
  ["klearchos. bollettino dell’associazione amici del museo nazionale di reggio calabria"] = "Klearchos",
  ["kleine senckenbergreihe"] = "Kleine Senckenb.r.",
  ["kleos. estemporaneo di studi e testi sulla fortuna dell’antico"] = "Kleos",
  ["klimik dergisi"] = "KLIMIK Derg",
  ["klinicheskaia khirurgiia"] = "Klin Khir",
  ["klinicheskaia laboratornaia diagnostika"] = "Klin Lab Diagn",
  ["klinicheskaia meditsina"] = "Klin Med (Mosk)",
  ["klinichna khirurhiia"] = "Klin Khir",
  ["klinichna khirurhiia / ministerstvo okhorony zdorov’ia ukrainy, naukove tovarystvo khirurhiv ukrainy"] = "Klin Khir",
  ["klinicka mikrobiologie a infekcni lekarstvi"] = "Klin Mikrobiol Infekc Lek",
  ["klinická imunológia a alergológia"] = "Klin Imunol Alergol",
  ["klinická mikrobiologie a infekc̆ní lékar̆ství"] = "Klin Mikrobiol Infekc Lek",
  ["klinická onkologie : casopis ceské a slovenské onkologické spolecnosti"] = "Klin Onkol",
  ["klinik psikofarmakoloji bülteni = bulletin of clinical psychopharmacology"] = "Klinik Psikofarmakol Bülteni",
  ["klinik und praxis"] = "Klin Prax",
  ["klinika i lechenie zlokachestvennykh novoobrazovanii"] = "Klin. Lech. Zlokach. Novoobraz.",
  ["klinika i lechenie zlokachestvennykh novoobrazovaniĭ"] = "Klin Lech Zlokach Novoobraz",
  ["klinika oczna"] = "Klin Oczna",
  ["klinische anasthesiologie und intensivtherapie"] = "Klin Anasthesiol Intensivther",
  ["klinische anästhesiologie und intensivtherapie"] = "Klin Anasthesiol Intensivther",
  ["klinische medizin; osterreichische zeitschrift fur wissenschaftliche und praktische medizin"] = "Klin. Med. Osterr. Z. Wiss. Prakt. Med.",
  ["klinische medizin; österreichische zeitschrift für wissenschaftliche und praktische medizin"] = "Klin Med Osterr Z Wiss Prakt Med",
  ["klinische monatsblatter fur augenheilkunde"] = "Klin Monatsbl Augenheilkd",
  ["klinische monatsblatter fur augenheilkunde und fur augenarztliche fortbildung"] = "Klin Monatsblatter Augenheilkd Augenarztl Fortbild",
  ["klinische monatsblätter für augenheilkunde"] = "Klin Monbl Augenheilkd",
  ["klinische monatsblätter für augenheilkunde und für augenärztliche fortbildung"] = "Klin Monbl Augenheilkd Augenarztl Fortbild",
  ["klinische neuroradiologie"] = "Klin Neuroradiol",
  ["klinische padiatrie"] = "Klin Padiatr",
  ["klinische pädiatrie"] = "Klin Padiatr",
  ["klinische wochenschrift"] = "Klin Wochenschr",
  ["klinisk sygepleje"] = "Klin Sygepleje",
  ["klio. beiträge zur alten geschichte"] = "Klio",
  ["klitynna ta orhanna transplantolohii︠a︡"] = "Klitynna Orhanna Transplant",
  ["kluwer international handbooks of education"] = "Kluwer Int. Handb. Educ.",
  ["kluwer texts in the mathematical sciences"] = "Kluwer Texts Math. Sci.",
  ["km itl science and technology journal"] = "KM ITL Sci Technol J",
  ["kmutnb : ijast"] = "KMUTNB",
  ["kn - journal of cartography and geographic information"] = "KN J Cartogr Geogr Inf",
  ["kn journal of cartography and geographic information"] = "KN J. Cartogr. Geogr. Inf.",
  ["knee surgery & related research"] = "Knee Surg Relat Res",
  ["knee surgery, sports traumatology, arthroscopy"] = "Knee Surg. Sports Traumatol. Arthrosc.",
  ["knee surgery, sports traumatology, arthroscopy : official journal of the esska"] = "Knee Surg Sports Traumatol Arthrosc",
  ["knjiga. jugoslavenska akademija znanosti i umjetnosti. odjel za medicinske nauke"] = "Knjiga (Med)",
  ["knowledge and information systems"] = "Knowl Inf Syst",
  ["knowledge and management of aquatic ecosystems"] = "Knowl Manag Aquat Ecosyst",
  ["knowledge and process management"] = "Knowl. Process Manage.",
  ["knowledge engineering review"] = "Knowl. Eng. Rev.",
  ["knowledge organization for a sustainable world: challenges and perspectives for cultural, scientific, and technological sharing in a connected society : proceedings of the fourteenth international isko conference 27-29 september 2016 rio de janeiro, brazil. organized by the international society for knowledge organization (isko) isko-brazil são paulo state university"] = "Knowl Organ Sustain World Chall Perspect Cult Sci Technol Shar Connect Soc (2016)",
  ["knowledge-based systems"] = "Knowl Based Syst",
  ["ko hsueh tung pao"] = "Kexue Tongbao [English Language Edition]|Kexue Tongbao",
  ["ko hsueh tung pao = kexue tongbao [english language edition]"] = "Kexue Tongbao",
  ["kobe daigaku igakubu kiyo (medical journal of kobe university)"] = "Kobe Daigaku Igakubu Kiyo",
  ["kobe daigaku igakubu kiyo. medical journal of kobe university"] = "Kobe Daigaku Igakubu Kiyo",
  ["kobe economic and business review"] = "Kobe Econ. Bus. Rev.",
  ["kobe ika daigaku kiyo"] = "Kobe Ika Daigaku Kiyo",
  ["kobe journal of mathematics"] = "Kobe J. Math.",
  ["kobe journal of medical sciences"] = "Kobe J. Med. Sci.",
  ["kobe university economic review"] = "Kobe Univ. Econ. Rev.",
  ["kobunshi ronbunshu"] = "Kobunshi Ronbunshu",
  ["kocatepe veteriner dergisi"] = "Kocatepe Vet Derg",
  ["kochi university"] = "Mem. Fac. Sci. Kochi Univ. Ser. A Math.",
  ["kodai mathematical journal"] = "Kodai Math. J.",
  ["kodai. journal of ancient history"] = "Kodai",
  ["kogyo zairyo"] = "Kogyo Zairyo",
  ["koks i khimiya"] = "Koks Khim.",
  ["koku eisei gakkai zasshi"] = "Koku Eisei Gakkai Zasshi",
  ["kokubyo gakkai zasshi (journal of the stomatological society, japan)"] = "Kokubyo Gakkai Zasshi",
  ["kokubyo gakkai zasshi. the journal of the stomatological society, japan"] = "Kokubyo Gakkai Zasshi",
  ["kokumin-keizai zasshi"] = "Kokumin-Keizai Zasshi",
  ["kokuritsu iyakuhin shokuhin eisei kenkyujo hokoku (bulletin of national institute of health sciences)"] = "Kokuritsu Iyakuhin Shokuhin Eisei Kenkyusho Hokoku",
  ["kokuritsu iyakuhin shokuhin eisei kenkyusho hokoku (bulletin of national institute of health sciences)"] = "Kokuritsu Iyakuhin Shokuhin Eisei Kenkyusho Hokoku",
  ["kokuritsu iyakuhin shokuhin eisei kenkyūjo hōkoku = bulletin of national institute of health sciences"] = "Kokuritsu Iyakuhin Shokuhin Eisei Kenkyusho Hokoku",
  ["kokuritsu kagaku hakubutsukan senpō : memoirs of the national science museum. kokuritsu kagaku hakubutsukan (japan)"] = "Kokuritsu Kagaku Hakubutsukan Senpo",
  ["kokuritsu kogai kenkyusho tokubetsu kenkyu hokoku"] = "Kokuritsu Kogai Kenkyusho Kenkyu Hokoku",
  ["kokusai kyōryoku kenkyūshi"] = "Kokusai Kyoryoku Kenkyushi",
  ["kokusaigaku revyū = obirin review of international studies"] = "Kokusaigaku revyu",
  ["kokyu to junkan (respiration and circulation)"] = "Kokyu To Junkan",
  ["kokyu to junkan. respiration & circulation"] = "Kokyu To Junkan",
  ["koldewey-gesellschaft, vereinigung für baugeschichtliche forschung e. v. bericht über die tagung für ausgrabungswissenschaft und bauforschung"] = "KGB",
  ["koleopterologische rundschau"] = "Koleopterol Rundsch",
  ["kolleg philosophie"] = "Kolleg Philos.",
  ["kolloid zeitschrift"] = "Kolloid Z.",
  ["kolloid zeitschrift & zeitschrift fur polymere"] = "Kolloid Z. Z. Polym.",
  ["kolloid zeitschrift & zeitschrift für polymere"] = "Kolloid Z. Z. Polym.",
  ["kolloidchemische beihefte"] = "Kolloidchem. Beih.",
  ["kolloidn zhurnal"] = "Kolloidn Zh.",
  ["kolloidnyi zhurnal"] = "Kolloidn. Zh.",
  ["kolloquien zur allgemeinen und vergleichenden archäologie"] = "KollAVA",
  ["kolner medizinhistorische beitrage"] = "Koln Med Beitr",
  ["kolner zeitschrift fur soziologie und sozialpsychologie"] = "Kolner Z Soz Sozialpsychol (Aufl)",
  ["kompass nutrition & dietetics"] = "Kompass Nutr. Diet.",
  ["kompetenz-zentrum holz"] = "Kompet.-Zent. Holz",
  ["kompleksnoe ispol’zovanie mineral’nogo syr’ya"] = "Kompleksn. Ispol’z. Miner. Syr’ya",
  ["komunikaty mazursko-warminskie"] = "Komun Mazur Warm",
  ["komunikaty mazursko-warmińskie"] = "Komun Mazur Warm",
  ["kona : powder science and technology in japan"] = "Kona",
  ["kona powder and particle journal"] = "Kona Powder Part. J.",
  ["kongelige danske videnskabernes selskab, matematisk-fysiske meddelelser"] = "K. Dan. Vidensk. Selsk. Mater. Fys. Medd.",
  ["kongelige danske videnskabernes selskab: matematisk-fysiske meddelelser"] = "Kong. Dansk. Vidensk.",
  ["kongressband / deutsche gesellschaft fur chirurgie. deutsche gesellschaft fur chirurgie. kongress"] = "Kongressbd Dtsch Ges Chir Kongr",
  ["kongressband, deutsche gesellschaft fur chirurgie"] = "Kongressbd. Dtsch. Ges. Chir. Kongr.",
  ["kongressband. deutsche gesellschaft für chirurgie. kongress"] = "Kongressbd Dtsch Ges Chir Kongr",
  ["koninklijke nederlandse akademie van wetenschappen"] = "Indag. Math. (N.S.)",
  ["koninklijke nederlandse akademie van wetenschappen. indagationes mathematicae. new series"] = "Indag. Math. (N.S.)",
  ["konuralp journal of mathematics"] = "Konuralp J. Math.",
  ["koordinatsionnaya khimiya (russian journal of coordination chemistry)"] = "Koord. Khim.",
  ["kora bericht"] = "KORA Ber.",
  ["korea & world affairs"] = "Korea (South)",
  ["korea journal"] = "Korea J",
  ["korea journal of population and development"] = "Korea J Popul Dev",
  ["korea-australia rheology journal"] = "Korea-Aust. Rheol. J.",
  ["korean and korean-american studies bulletin"] = "Korean Korean Am Stud Bull",
  ["korean annals of mathematics"] = "Korean Ann. Math.",
  ["korean circulation journal"] = "Korean Circ J",
  ["korean diabetes journal"] = "Korean Diabetes J",
  ["korean economic review"] = "Korean Econ Rev",
  ["korean journal for food science of animal resources"] = "Korean J Food Sci Anim Resour",
  ["korean journal of anesthesiology"] = "Korean J Anesthesiol",
  ["korean journal of applied statistics"] = "Korean J. Appl. Statist.",
  ["korean journal of audiology"] = "Korean J Audiol",
  ["korean journal of biological sciences"] = "Korean J Biol Sci",
  ["korean journal of chemical engineering"] = "Korean J. Chem. Eng.",
  ["korean journal of critical care medicine"] = "Korean J Crit Care Med",
  ["korean journal of defense analysis"] = "Korean J. Def. Anal.",
  ["korean journal of entomology"] = "Korean J Entomology",
  ["korean journal of environmental agriculture"] = "Korean J. Environ. Agric.",
  ["korean journal of family medicine"] = "Korean J Fam Med",
  ["korean journal of gastroenterology"] = "Korean J. Gastroenterol.",
  ["korean journal of hepato-biliary-pancreatic surgery"] = "Korean J Hepatobiliary Pancreat Surg",
  ["korean journal of hepatology"] = "Korean J. Hepatol.",
  ["korean journal of internal medicine"] = "Korean J. Intern. Med.",
  ["korean journal of materials research"] = "Korean J. Mater. Res.",
  ["korean journal of medical education"] = "Korean J Med Educ",
  ["korean journal of metals and materials"] = "Korean J. Met. Mater.",
  ["korean journal of neurotrauma"] = "Korean J Neurotrauma",
  ["korean journal of ophthalmology"] = "Korean J. Ophthalmol.",
  ["korean journal of ophthalmology : kjo"] = "Korean J Ophthalmol",
  ["korean journal of orthodontics"] = "Korean J Orthod",
  ["korean journal of parasitology"] = "Korean J. Parasitol.",
  ["korean journal of pathology"] = "Korean J Pathol",
  ["korean journal of pediatrics"] = "Korean J Pediatr",
  ["korean journal of physiology & pharmacology"] = "Korean J. Physiol. Pharmacol.",
  ["korean journal of plant pathology"] = "Hangug Sigmul Byeongri Haghoeji",
  ["korean journal of radiology"] = "Korean J Radiol",
  ["korean journal of radiology : official journal of the korean radiological society"] = "Korean J Radiol",
  ["korean journal of soil science and fertilizer"] = "Korean J. Soil Sci. Fert.",
  ["korean journal of spine"] = "Korean J Spine",
  ["korean journal of systematic zoology"] = "Korean J Syst Zool",
  ["korean journal of urology"] = "Korean J Urol",
  ["korean journal of veterinary research"] = "Korean J Vet Res",
  ["korean mathematical society"] = "Commun. Korean Math. Soc.",
  ["korean mathematical society. communications"] = "Commun. Korean Math. Soc.",
  ["korot (jerusalem : 1952)"] = "Korot",
  ["koryŏ insam hakhoe chi"] = "KORYO INSAM HAKHOE CHI",
  ["koshu eiseiin kenkyu hokoku. bulletin of the institute of public health"] = "Koshu Eisei In Kenkyu Hokoku",
  ["kosmas : [journal of the czechoslovak society of arts and sciences (svu)]"] = "Kosmas",
  ["kosmicheskaia biologiia i aviakosmicheskaia meditsina"] = "Kosm Biol Aviakosm Med",
  ["kosmicheskaia biologiia i meditsina"] = "Kosm Biol Med",
  ["kosmos. seria a, biologia"] = "Kosm Ser A Biol",
  ["kosmos. seria a, biologia / polskie towarzystwo przyrodnikow im. kopernika"] = "Kosm Ser A Biol",
  ["kou qiang yi xue yan jiu = journal of oral science research"] = "Kou Qiang Yi Xue Yan Jui",
  ["kovove materialy"] = "Kovove Mater.",
  ["kovove materialy metallic materials"] = "Kovove Mater.",
  ["kozhevenno-obuvnaya promyshlennost"] = "Kozh.-Obuvn. Prom-st.",
  ["kōkūbyō gakkai zasshi. the journal of the stomatological society, japan"] = "Kokubyo Gakkai Zasshi",
  ["kōshū eiseiin kenkyū hōkoku. bulletin of the institute of public health"] = "Koshu Eisei In Kenkyu Hokoku",
  ["kölner zeitschrift für soziologie und sozialpsychologie"] = "Kolner Z Soz Sozpsychol",
  ["kragujevac journal of mathematics"] = "Kragujevac J. Math.",
  ["krankenhaus-hygiene + infektionsverhütung"] = "Krankenhaushyg Infektionsverhut",
  ["krankenpflege (frankfurt am main, germany)"] = "Krankenpflege (Frankf)",
  ["krankenpflege (soins infirmiers)"] = "Krankenpfl. Soins Infirm.",
  ["krankenpflege journal"] = "Krankenpfl J",
  ["krankenpflege. soins infirmiers"] = "Krankenpfl Soins Infirm",
  ["kratkie soobshcheniya po fizike"] = "Kratk. Soobshch. Fiz.",
  ["kratkie soobščenija instituta archeologii, kiev"] = "KSIAKiev",
  ["kratkie soobščenija o dokladach i polevych issledovanijach instituta archeologii"] = "KSIA",
  ["kratkije soobseniâ inst. arheol. akad. nauk sssr"] = "KSIA",
  ["kratylos. kritisches berichts- und rezensionsorgan für indogermanische und allgemeine sprachwissenschaft"] = "Kratylos",
  ["krebsforschung und krebsbekampfung"] = "Krebsforsch. Krebsbekampf.",
  ["krebsforschung und krebsbekämpfung"] = "Krebsforsch Krebsbekampf",
  ["kredit und kapital"] = "Kredit Kapital",
  ["kriminalistik und forensische wissenschaft"] = "Krim. Forensische Wiss.",
  ["krimskiĭ zhurnal ėksperimentalʹnoĭ i klinicheskoĭ medit︠s︡iny = krymsʹkyĭ z︠h︡urnal eksperymentalʹnoï ta klinichnoï medyt︠s︡yny = crimean journal of experimental and clinical medicine"] = "Krim Z Eksp Klin Med",
  ["kristall und technik"] = "Krist. Tech.",
  ["kristallografiya [soviet physics–crystallography]"] = "Kristallografiya [Sov. Phys. Crystallogr.]",
  ["kritiek (ravels, belgium)"] = "Kritiek",
  ["kritisches jahrbuch der philosophie"] = "Krit. Jahrb. Philos.",
  ["kroc foundation series"] = "Kroc Found Ser",
  ["kronika (ljubljana, slovenia)"] = "Kronika",
  ["kronos (bellville, south africa)"] = "Kronos",
  ["ksbb journal"] = "KSBB J",
  ["ksce journal of civil engineering"] = "KSCE J. Civ. Eng.",
  ["ksii transactions on internet and information systems"] = "KSII Trans. Internet Inf. Syst.",
  ["ktema. civilisations de l’orient, de la grèce et de rome antiques"] = "Ktema",
  ["kuang pu hseuh yu kuang pu fen hsi spectroscopy and spectral analysis (pei-ching"] = "Kuang Pu Hseuh Yu Kuang Pu Fen Hsi",
  ["kuba; revista de medicina tropical y parasitología"] = "Kuba",
  ["kula kula"] = "Kula Kula",
  ["kulak burun boğaz ihtisas dergisi : kbb = journal of ear, nose, and throat"] = "Kulak Burun Bogaz Ihtis Derg",
  ["kultur & technik : zeitschrift des deutschen museums munchen"] = "Kult Tech",
  ["kultur & technik : zeitschrift des deutschen museums münchen"] = "Kult Tech",
  ["kultura fizyczna"] = "Kult Fiz",
  ["kultura i spoleczenstwo"] = "Kult Spolecz",
  ["kultura i społeczeństwo"] = "Kult Spolecz",
  ["kulturgeschichte der zahnheilkunde in einzeldarstellungen"] = "Kult Zahnheilkd Einzeldarst",
  ["kul’tura iskusstvo anticnogo mira"] = "KIAM",
  ["kumamoto igakkai zasshi (journal of the kumamoto medical society)"] = "Kumamoto Igakkai Zasshi",
  ["kumamoto igakkai zasshi. the journal of the kumamoto medical society"] = "Kumamoto Igakkai Zasshi",
  ["kumamoto journal of mathematics"] = "Kumamoto J. Math.",
  ["kumamoto medical journal"] = "Kumamoto Med. J.",
  ["kumamoto university"] = "Mem. Fac. Ed. Kumamoto Univ. Natur. Sci.",
  ["kuml. årbog for jysk arkaeologisk selskab"] = "Kuml",
  ["kun chong fen lei xue bao"] = "Entomotaxonomia|Kun Chong Fen Lei Xue Bao",
  ["kun chong fen lei xue bao = entomotaxonomia"] = "Kun Chong Fen Lei Xue Bao",
  ["kun chong xue bao. acta entomologica sinica"] = "Kun Chong Xue Bao",
  ["kun chong zhi shi"] = "Kunchong zhishi / [Zhongguo kun chong xue hui]|Kun Chong Zhi Shi",
  ["kun chong zhi shi = kunchong zhishi"] = "Kun Chong Zhi Shi",
  ["kungliga krigsvetenskapsakademiens handlingar och tidskrift"] = "K Krigsvetenskapakad Handlingar Tidskr",
  ["kungliga vitterhets historie och antikvitets akademiens handlingar. antikvariska serien"] = "Handlingar",
  ["kunst & therapie"] = "Kunst Ther",
  ["kunst der welt in den berliner museen"] = "KuWeltBerlMus",
  ["kunst des orients"] = "Kunst Orient",
  ["kunstchronik. monatsschrift für kunstwissenschaft, museumswesen und denkmalpflege, mitteilungsblatt des verbandes deutscher kunsthistoriker"] = "Kunstchronik",
  ["kunstgeschichtliche anzeigen"] = "KuGeschAnz",
  ["kurinikaru sutadi (clinical study)"] = "Kurinikaru Sutadi",
  ["kurinikaru sutadī = clinical study"] = "Kurinikaru Sutadi",
  ["kurume medical journal"] = "Kurume Med. J.",
  ["kush. journal of the national corporation for antiquities and museums (ncam)"] = "Kush",
  ["kuwait journal of science"] = "Kuwait J. Sci.",
  ["kuwait journal of science & engineering"] = "Kuwait J. Sci. Eng.",
  ["künstliche intelligenz"] = "Kunstliche Intell (Oldenbourg)",
  ["kvantovaya elektronika (moscow)"] = "Kvantovaya Elektron. (Moscow)",
  ["kvantovaya elektronika (moscow) [soviet journal of quantum electronics]"] = "Kvant. Elektron. (Moscow) [Sov. J. Quantum Electron.]",
  ["kvartalsskrift. svenska nationalföreningen mot hjärt- och lungsjukdomar"] = "Kvartalsskr Sven Nationalforen Hjart Lungsjukd",
  ["kvartalsskrift. svenska nationalföreningen mot tuberkulos"] = "Kvartalsskrift",
  ["kwansei gakuin university"] = "Kwansei Gakuin Univ. Nat. Sci. Rev.",
  ["kwartalnik historii kultury materialnej"] = "Kwart Hist Kult Mater",
  ["kwartalnik historii nauki i techniki"] = "Kwart. Hist. Nauk. Tech.",
  ["kwartalnik historii nauki i techniki : kwartal'nyĭ zhurnal istorii nauki i tekhniki -"] = "Kwart Hist Nauki Tech",
  ["kwartalnik historii nauki i techniki : kwartal’nyi zhurnal istorii nauki i tekhniki -"] = "Kwart Hist Nauki Tech",
  ["kwartalnik historii ruchu zawodowego"] = "Kwart Hist Ruchu Zaw",
  ["kwartalnik historii żydów : jewish history quarterly"] = "Kwart Hist Zydow",
  ["kwartalnik historyczny"] = "Kwart Hist",
  ["ky [reports]. u.s. atomic energy commission"] = "KY Rep",
  ["ky reports"] = "KY Rep.",
  ["kybernetes. the international journal of cybernetics, systems and management sciences"] = "Kybernetes",
  ["kyklos (online)"] = "Kyklos (Oxford)",
  ["kyklos : jahrbuch des instituts fur geschichte der medizin an der universitat leipzig"] = "Kyklos",
  ["kyklos : jahrbuch des instituts für geschichte der medizin an der universität leipzig"] = "Kyklos",
  ["kyobu geka (japanese journal of thoracic surgery)"] = "Kyobu Geka",
  ["kyobu geka. japanese journal of thoracic surgery"] = "Kyobu Geka",
  ["kyobu geka. the japanese journal of thoracic surgery"] = "Kyobu Geka",
  ["kyoto daigaku kokukagaku kiyo (bulletin of stomatology, kyoto university)"] = "Kyoto Daigaku Kokukagaku Kiyo",
  ["kyoto daigaku kokukagaku kiyo. bulletin of stomatology, kyoto university"] = "Kyoto Daigaku Kokukagaku Kiyo",
  ["kyoto journal of mathematics"] = "Kyoto J. Math.",
  ["kyoto university"] = "Publ. Res. Inst. Math. Sci.",
  ["kyoto university economic review"] = "Kyoto Univ. Econ. Rev.",
  ["kyōrin igakkai zasshi. journal of the kyorin medical society"] = "Kyorin Igakkai Zasshi",
  ["kyŏngbuk uidae chapchi"] = "Gyenbug Uidai Jabji",
  ["kypriakai spoudai : deltion tes hetaireias kypriakon spoudon"] = "Kupr Spoud",
  ["kypriakai spoudai : deltion tēs hetaireias kypriakōn spoudōn"] = "Kupr Spoud",
  ["kyrkohistorisk årsskrift"] = "Kyrkohist Arsskr",
  ["kyungpook mathematical journal"] = "Kyungpook Math. J.",
  ["kyungpook national university"] = "Kyungpook Math. J.",
  ["kyushu journal of mathematics"] = "Kyushu J. Math.",
  ["kyushu journal of medical science"] = "Kyushu J Med Sci",
  ["kyushu sangyo university"] = "J. Fac. Int. Stud. Cult. Kyushu Sangyo Univ.",
  ["kǣn kasēt = khon kaen agriculture journal"] = "Kaen Kaset",
  ["kêmi. revue de philologie et d’archéologie égytiennes et coptes"] = "Kemi",
  ["kölner jahrbuch"] = "KoelnJb",
  ["kölner jahrbuch für vor- und frühgeschichte, hrsg. vom römisch- germanischen museum und der archäologischen gesellschaft köln"] = "KJ",
  ["kölner medizinhistorische beiträge"] = "Koln Med Beitr",
  ["kölner museums-bulletin. berichte und forschungen aus den museen der stadt köln"] = "KoelnMusB",
  ["kölner zeitschrift für soziologie und sozialpsychologie"] = "Koln. Z. Soziol. Sozialpsych.",
  ["kŏn’guk taehakkyo, haksul yŏn’guwon"] = "Hak-sul Chi",
  ["ḳatedrah be-toldot erets-yiśraʼel ṿe-yishuvah"] = "Katedrah Betoldot Erets Yisrael Veyishuvah",
  ["kώκαλοσ. studi pubblicati dall’istituto di storia antica dell’università di palermo"] = "Kokalos",
  ["l' alimentation et la vie"] = "Aliment Vie",
  ["l' année biologique"] = "Annee Biol",
  ["l' année sociologique"] = "Annee Sociol",
  ["l' année thérapeutique"] = "Annee Ther",
  ["l' arcispedale s. anna di ferrara"] = "Arcisp S Anna Ferrara",
  ["l' ateneo parmense. sezione i, acta bio-medica"] = "Ateneo Parmense 1",
  ["l' entomologiste"] = "Entomologiste (Paris)",
  ["l' ethnographie"] = "Ethnographie.",
  ["l' eurobiologiste"] = "Eurobiologiste",
  ["l' evolution médicale"] = "Evol Med",
  ["l' evolution psychiatrique"] = "Evol Psychiatr (Paris)",
  ["l' homéopathie française"] = "Homeopath Fr",
  ["l' homme; revue française d'anthropologie"] = "Homme",
  ["l' hôpital"] = "Hopital",
  ["l' infirmiere haïtienne"] = "Infirm Haiti",
  ["l' infirmière"] = "Infirmiere",
  ["l' infirmière canadienne"] = "Infirm Can",
  ["l' information dentaire"] = "Inf Dent",
  ["l' information médicale et paramédicale"] = "Inf Med Paramed (Montreal)",
  ["l' orthodontie française"] = "Orthod Fr",
  ["l' ospedale maggiore"] = "Osp Maggiore",
  ["l'action nationale"] = "Action Natl",
  ["l'aéronautique et l'astronautique : organe commun de l'association française des ingénieurs et techniciens de l'aéronautique de l'espace (a.f.i.t.a.e.) et de la société française d'astronautique (s.f.a.)"] = "Aeronaut Astronaut",
  ["l'afrique et l'asie modernes"] = "Afr Asie Mod",
  ["l'afrique littéraire et artistique"] = "Afr Litt Artist",
  ["l'année endocrinologique"] = "Annee Endocrinol",
  ["l'année philologique; bibliographie critique et analytique de l'antiquité gréco-latine"] = "Annee Philologique",
  ["l'année psychologique"] = "Annee Psychol",
  ["l'année sociale"] = "Annee Soc",
  ["l'année thérapeutique et clinique en ophtalmologie"] = "Annee Ther Clin Ophtalmol",
  ["l'antiquité classique"] = "Antiq Class",
  ["l'archiginnasio : bullettino della biblioteca comunale di bologna"] = "Archiginnasio",
  ["l'assistenza sociale"] = "Assist Soc",
  ["l'ateneo parmense. acta bio-medica : organo della società di medicina e scienze naturali di parma"] = "Ateneo Parmense Acta Biomed",
  ["l'ateneo parmense. acta naturalia : organo della società di medicina e scienze naturali di parma"] = "Ateneo Parmense Acta Nat",
  ["l'attualita medica"] = "Attual Med",
  ["l'attualità dietetica"] = "Attual Diet",
  ["l'autre europe"] = "Autre Eur",
  ["l'avenir médical"] = "Avenir Med",
  ["l'echo médical des cévennes"] = "Echo Med Cevennes",
  ["l'echo médical du nord"] = "Echo Med Nord",
  ["l'economia umana; rassegna medica internazionale"] = "Economia Umana Rassegna Medica Internazionale",
  ["l'economie de la réunion"] = "Econ Reun",
  ["l'egypte contemporaine"] = "Egypte Contemp",
  ["l'enfant en milieu tropical"] = "Enfant Milieu Trop",
  ["l'espace géographique"] = "Espace Geogr",
  ["l'économie familiale"] = "Econ Fam",
  ["l'hygiéne mentale"] = "Hyg Ment",
  ["l'igiene moderna"] = "Ig Mod",
  ["l'infirmière auxiliaire : revue de la corporation professionnelle des infirmières et infirmiers auxiliaires du québec"] = "Infirm Aux",
  ["l'infirmière du québec : revue officielle de l'ordre des infirmières et infirmiers du québec"] = "Infirm Que",
  ["l'infirmière francaise"] = "Infirm Fr",
  ["l'information d'histoire de l'art"] = "Inf Hist Art",
  ["l'information géographique"] = "Inf Geogr",
  ["l'information historique"] = "Inf Hist",
  ["l'information littéraire"] = "Inf Litt",
  ["l'information psychiatrique"] = "Inf Psychiatr",
  ["l'informatore agrario"] = "Inf Agrar",
  ["l'informatore medico"] = "Inf Med",
  ["l'oculiste français"] = "Ocul Fr",
  ["l'ospedale psichiatrico"] = "Osp Psichiatr",
  ["l'oto-rhino-laryngologie internationale"] = "Otorhinolaryngolog Int",
  ["l'oto-rino-laringologia italiana"] = "Otorinolaringol Ital",
  ["l'union médicale des praticiens français"] = "Union Med Prat Francais",
  ["l'unión médicale du canada"] = "Union Med Can",
  ["la [reports]. u.s. atomic energy commission"] = "LA Rep",
  ["la bibliofilia; rivista de storia del libro e delle arti grafiche di bibliografia ed erudizione"] = "Bibliofilia",
  ["la casana"] = "Casana",
  ["la cellule"] = "Cellule",
  ["la chirurgia degli organi di movimento"] = "Chir Organi Mov",
  ["la chronique du ceped"] = "Chron CEPED",
  ["la chronique du ceped / centre francais sur la population et le developpement"] = "Chron CEPED",
  ["la civilta cattolica"] = "Civ Cattol",
  ["la civiltà cattolica"] = "Civ Cattol",
  ["la clinica ginecologica"] = "Clin Ginecol",
  ["la clinica ortopedica"] = "Clin Ortop",
  ["la clinica ostetrica e ginecologica"] = "Clin Ostet Ginecol",
  ["la clinica otorinolaringoiatrica"] = "Clin Otorinolaringoiatr",
  ["la clinica pediatrica"] = "Clin Pediatr (Bologna)",
  ["la clinica terapeutica"] = "Clin Ter",
  ["la clinica termale"] = "Clin Term",
  ["la clinica veterinaria"] = "Clin Vet (Milano)",
  ["la critica sociologica"] = "Crit Sociol",
  ["la crónica médica"] = "Cron Med",
  ["la cultura scientifica"] = "Cult. Sci.",
  ["la diagnosi; rivista mensile di medicina pratica"] = "Diagnosi (Napoli)",
  ["la enfermera argentina"] = "Enferma Argent",
  ["la escuela de farmacia"] = "Esc Farm",
  ["la gaceta de la real sociedad matematica espanola"] = "Gac. R. Soc. Mat. Esp.",
  ["la gaceta de la real sociedad matemática española"] = "Gac. R. Soc. Mat. Esp.",
  ["la gazette des archives"] = "Gaz Arch",
  ["la ginecologia"] = "Ginecol",
  ["la juventud médica"] = "Juv Med",
  ["la linguistique"] = "Ling",
  ["la lutte contre le cancer"] = "Lutte Cancer",
  ["la matematica"] = "Matematica",
  ["la matematica per il 3+2"] = "La Mat. per il 3+2",
  ["la matematica. official journal of the association for women in mathematics"] = "Matematica",
  ["la medecine aeronautique"] = "Med Aeronaut",
  ["la medecine en france"] = "Med Fr (Bombay)",
  ["la medecine infantile"] = "Med Infant (Paris)",
  ["la medicina colonial"] = "Med Colon",
  ["la medicina del lavoro"] = "Med Lav",
  ["la medicina internazionale"] = "Med Int (Milano)",
  ["la medicina italiana"] = "Med Ital",
  ["la medicina tropical"] = "Med Trop (Madr)",
  ["la médecine aéronautique"] = "Med Aeronaut",
  ["la médecine en france"] = "Med Fr (Bombay)",
  ["la médecine infantile"] = "Med Infant (Paris)",
  ["la nouvelle clio"] = "NouvClio",
  ["la nouvelle presse medicale"] = "Nouv Presse Med",
  ["la nouvelle presse médicale"] = "Nouv Presse Med",
  ["la nouvelle revue des deux mondes"] = "Nouv Rev Deux Mondes",
  ["la nuova veterinaria"] = "Nuova Vet",
  ["la parola del passato"] = "Parola Passato",
  ["la parola del passato: rivista di studi antichi"] = "PP",
  ["la pathologie generale"] = "Pathol Gen",
  ["la pathologie générale"] = "Pathol Gen",
  ["la pediatria"] = "Pediatria (Napoli)",
  ["la pediatria del medico pratico"] = "Pediatr Med Prat",
  ["la pediatria medica e chirurgica : medical and surgical pediatrics"] = "Pediatr Med Chir",
  ["la pensee"] = "Pensee",
  ["la pensee et les hommes"] = "Pensee Hommes",
  ["la pensée"] = "Pensee",
  ["la pensée et les hommes"] = "Pensee Hommes",
  ["la pharmacie industrielle"] = "Pharm Ind",
  ["la pie"] = "Pie",
  ["la piê"] = "Pie",
  ["la prensa medica mexicana"] = "Prensa Med Mex",
  ["la prensa médica"] = "Prensa Med",
  ["la prensa médica mexicana"] = "Prensa Med Mex",
  ["la presse medicale"] = "Presse Med",
  ["la presse médicale"] = "Presse Med",
  ["la presse thermale et climatique"] = "Presse Therm Clim",
  ["la promotion dentaire"] = "Promot Dent",
  ["la prophylaxie antivenerienne"] = "Prophyl Antivenerienne",
  ["la prophylaxie antivénérienne"] = "Prophyl Antivenerienne",
  ["la prophylaxie sanitaire et morale"] = "Prophyl Sanit Morale",
  ["la provence medicale"] = "Prov Med",
  ["la provence médicale"] = "Prov Med",
  ["la provincia di lucca"] = "ProvLucca",
  ["la psychiatrie de l'enfant"] = "Psychiatr Enfant",
  ["la psychiatrie de l’enfant"] = "Psychiatr Enfant",
  ["la puériculture"] = "Puericulture",
  ["la quintessenza"] = "Quintessenza",
  ["la radiologia medica"] = "Radiol Med",
  ["la rassegna della letteratura italiana"] = "Rass Lett Ital",
  ["la rassegna di clinica, terapia e scienze affini"] = "Rass Clin Ter",
  ["la recherche"] = "Recherche",
  ["la reforma médica"] = "Reforma Med",
  ["la reports"] = "LA Rep.",
  ["la revista médica de yucatán"] = "Rev Med Yucatan",
  ["la revista médica del uruguay"] = "Rev Med Urug (Montev)",
  ["la revue de geographie de montreal"] = "Rev Geogr Montr",
  ["la revue de geriatrie"] = "Revue Geriatr",
  ["la revue de géographie de montréal"] = "Rev Geogr Montr",
  ["la revue de gériatrie"] = "Revue Geriatr",
  ["la revue de medecine"] = "Rev Med (Paris)",
  ["la revue de médecine"] = "Rev Med (Paris)",
  ["la revue de médecine interne"] = "Rev Med Interne",
  ["la revue de pediatrie"] = "Rev Pediatr",
  ["la revue de pédiatrie"] = "Rev Pediatr",
  ["la revue des arts. musées de france"] = "RArtMus",
  ["la revue des infirmieres et infirmiers auxiliaires du quebec"] = "Rev Infirm Infirm Aux Que",
  ["la revue des infirmières et infirmiers auxiliaires du québec"] = "Rev Infirm Infirm Aux Que",
  ["la revue du foie"] = "Rev Foie",
  ["la revue du louvre et des musees de france"] = "Rev Louvre Mus Fr",
  ["la revue du louvre et des musées de france"] = "Rev Louvre Mus Fr",
  ["la revue du praticien"] = "Rev Prat",
  ["la revue francaise de la prothese dentaire"] = "Rev Fr Prothese Dent",
  ["la revue francaise d’endocrinologie clinique, nutrition, et metabolisme"] = "Rev Fr Endocrinol Clin",
  ["la revue française d'endocrinologie clinique, nutrition, et métabolisme"] = "Rev Fr Endocrinol Clin",
  ["la revue française de la prothèse dentaire"] = "Rev Fr Prothese Dent",
  ["la revue lyonnaise de medecine"] = "Rev Lyon Med",
  ["la revue lyonnaise de médecine"] = "Rev Lyon Med",
  ["la revue medicale de picardie"] = "Rev Med Picardie",
  ["la revue médicale de picardie"] = "Rev Med Picardie",
  ["la revue odontologique"] = "Rev Odontol",
  ["la revue stomato-odontologique du nord de la france"] = "Rev Stomatoodontol Nord Fr",
  ["la ricerca in clinica e in laboratorio"] = "Ric Clin Lab",
  ["la ricerca scientifica"] = "Ric Sci",
  ["la ricerca scientifica. 2. ser., pt. 2: rendiconti. sezione b: biologica"] = "Ric Sci 2 Ser Pt 2 Rend B",
  ["la ricerca scientifica. 2. ser., pt.1: rivista"] = "Ric Sci 2 Ser Pt 1 Riv",
  ["la riforma medica"] = "Riforma Med",
  ["la rivista del nuovo cimento"] = "Riv. Nuovo Cimento (3)",
  ["la rivista del nuovo cimento della società italiana di fisica"] = "Riv. Nuovo Cimento Soc. Ital. Fis. (4)",
  ["la sante publique"] = "Sante Publique (Bucur)",
  ["la santé de l'homme"] = "Sante Homme",
  ["la santé publique"] = "Sante Publique (Bucur)",
  ["la science dans la république populaire roumaine"] = "Sci Repub Pop Roum",
  ["la science médicale pratique"] = "Sci Med Prat",
  ["la semaine des hopitaux : organe fonde par l’association d’enseignement medical des hopitaux de paris"] = "Sem Hop",
  ["la semaine des hopitaux de paris"] = "Sem Hop Paris",
  ["la semaine des hopitaux: therapeutique"] = "Sem Hop Ther Paris",
  ["la semaine des hôpitaux : organe fondé par l'association d'enseignement médical des hôpitaux de paris"] = "Sem Hop",
  ["la semaine des hôpitaux de paris"] = "Sem Hop Paris",
  ["la semaine des hôpitaux: thérapeutique"] = "Sem Hop Ther Paris",
  ["la semaine medicale [medecine sociale. edition: couverture jaune]"] = "Sem Medicale Med Soc",
  ["la semaine médicale [médecine sociale. edition: couverture jaune]"] = "Sem Medicale Med Soc",
  ["la semana medica"] = "Sem Med",
  ["la semana médica"] = "Sem Med",
  ["la serie t"] = "Ser. T",
  ["la société royale du canada"] = "C. R. Math. Rep. Acad. Sci. Canada",
  ["la technique de l'eau et de l'assainissement"] = "Tech Eau Assainissement",
  ["la technique de l’eau et de l’assainissement"] = "Tech Eau Assainissement",
  ["la technique sanitaire et municipale"] = "Tech Sanit Munic",
  ["la torretta. rivista quadrimestrale a cura della biblioteca comunale di blera"] = "Torretta",
  ["la tour de l’orle-d’or, semur-en-auxois"] = "TourOrleOr",
  ["la tourbe des philosophes : revue d'études alchimiques"] = "Tourbe Philos",
  ["la tourbe des philosophes : revue d’etudes alchimiques"] = "Tourbe Philos",
  ["la tribuna odontologica"] = "Trib Odontol (B Aires)",
  ["la tribuna odontológica"] = "Trib Odontol (B Aires)",
  ["la tunisie medicale"] = "Tunis Med",
  ["la tunisie médicale"] = "Tunis Med",
  ["la vie medicale au canada francais"] = "Vie Med Can Fr",
  ["la vie medicale. evolution medicale et therapeutique"] = "Vie Med Evolut Med Ther",
  ["la vie médicale au canada français"] = "Vie Med Can Fr",
  ["la vie médicale. evolution médicale et thérapeutique"] = "Vie Med Evolut Med Ther",
  ["la voix de la femme : magazine trimestriel d'information, d'éducation et de mobilisation de l'apdf"] = "Voix Femme",
  ["la voix de la femme : magazine trimestriel d’information, d’education et de mobilisation de l’apdf"] = "Voix Femme",
  ["la voix dentaire"] = "Voix Dent",
  ["lab animal"] = "Lab Anim (NY)",
  ["lab matters : analysis, answers, actions"] = "Lab Matters",
  ["lab on a chip"] = "Lab Chip",
  ["lab world"] = "Lab World",
  ["lab world magazine"] = "Lab World Mag",
  ["labciencia con noticias tecnicas del laboratorio"] = "LabCiencia Not. Tec. Lab.",
  ["labeo. rassegna di ritritto romano"] = "Labeo",
  ["labor (durham, n.c. : online)"] = "Labor (Durh)",
  ["labor history"] = "Labor Hist",
  ["labor law journal"] = "Labor Law J",
  ["labor studies journal"] = "Labor Stud J",
  ["laboratorio clinico"] = "Lab Clin",
  ["laboratornoe delo"] = "Lab Delo",
  ["laboratory and research methods in biology and medicine"] = "Lab Res Methods Biol Med",
  ["laboratory animal care"] = "Lab Anim Care",
  ["laboratory animal research"] = "Lab Anim Res",
  ["laboratory animal science"] = "Lab Anim Sci",
  ["laboratory animal science professional"] = "Lab Animal Sci Prof",
  ["laboratory animals"] = "Lab Anim",
  ["laboratory automation and information management"] = "Lab. Autom. Inf. Manage.",
  ["laboratory hematology : official publication of the international society for laboratory hematology"] = "Lab Hematol",
  ["laboratory investigation"] = "Lab. Invest.",
  ["laboratory investigation; a journal of technical methods and pathology"] = "Lab Invest",
  ["laboratory medicine"] = "Lab Med",
  ["laboratory phonology"] = "Lab Phonol",
  ["laboratory practice"] = "Lab Pract",
  ["laboratory robotics and automation"] = "Lab. Rob. Autom.",
  ["labour (committee on canadian labour history)"] = "Labour",
  ["labour (committee on canadian labour history)."] = "Labour",
  ["labour (rome, italy : online)"] = "Labour (Rome)",
  ["labour and society"] = "Labour Soc",
  ["labour capital and society. travail capital et societe"] = "Labour Cap Soc",
  ["labour capital and society. travail capital et société"] = "Labour Cap Soc",
  ["labour economics"] = "Labour Econ",
  ["labour history"] = "Labour Hist",
  ["labour history review"] = "Labour Hist Rev",
  ["lactarius : boletín de la asociación micológica"] = "Lactarius (Jaén)",
  ["lahey clinic foundation bulletin"] = "Lahey Clin Found Bull",
  ["lahore journal of economics"] = "Lahore J. Econ.",
  ["laietania. estudios d’arqueologia del maresme"] = "Laietania",
  ["lake and reservoir management"] = "Lake Reservoir Manage.",
  ["lake line"] = "Lake Line",
  ["lakes & reservoirs : research and management"] = "Lakes Reserv",
  ["lakes & reservoirs: research and management"] = "Lakes Reservoirs Res. Manage.",
  ["lakes & reservoirs: science, policy and management for sustainable use"] = "Lakes & Reservoirs: Science, Policy and Management for Sustainable Use",
  ["lakokrasochnye materialy i ikh primenenie"] = "Lakokras. Mater. Ikh Primen.",
  ["lampas. tijdschrift voor nederlandse classici"] = "Lampas",
  ["lancet (london, england)"] = "Lancet",
  ["lancet diabetes and endocrinology"] = "Lancet Diabetes Endocrinol.",
  ["lancet digital health"] = "Lancet Digit Health",
  ["lancet infectious diseases"] = "Lancet Infect. Dis.",
  ["lancet neurology"] = "Lancet Neurol.",
  ["lancet oncology"] = "Lancet Oncol.",
  ["lancet planetary health"] = "Lancet Planet. Heath",
  ["lancet regional health - southeast asia"] = "Lancet Reg. Health - Southeast Asia",
  ["lancia. revista de prehistoria, arqueología e historia antigua del noreste peninsular"] = "Lancia",
  ["land and water law review"] = "Land Water Law Rev",
  ["land degradation & development"] = "Land Degrad Dev",
  ["land degradation and development"] = "Land Degrad. Dev.",
  ["land economics"] = "Land Econ",
  ["land use policy"] = "Land use policy",
  ["landeshydrologie und -geologie, mitteilung"] = "Landeshydrol. -geol., Mitt.",
  ["landeskundliche vierteljahrsblätter. trier"] = "LandKunVierJBl",
  ["landscape and ecological engineering"] = "Landscape Ecol. Eng.",
  ["landscape and urban planning"] = "Landsc Urban Plan",
  ["landscape architecture frontiers"] = "Landsc Archit Front",
  ["landscape ecology"] = "Landsc Ecol",
  ["landscape history"] = "Landscape Hist.",
  ["landscape history : journal of the society for landscape studies"] = "Landsc Hist",
  ["landscape journal"] = "Landsc J",
  ["landscape research"] = "Landsc Res",
  ["landscape research record"] = "Landsc Res Rec",
  ["landscapes in logic"] = "Landsc. Log",
  ["landschaft aargau"] = "Landsch. Aargau",
  ["landschaftsentwicklung und umweltforschung"] = "Landsch.entwickl. Umweltforsch.",
  ["landschaftsgenese und landschaftsökologie"] = "Landsch.genes. Landsch.ökol.",
  ["landschaftspflege und vegetationskunde"] = "Landsch.pfl. Veg.kd.",
  ["landschaftsökologie und umweltforschung"] = "Landsch.ökol. Umweltforsch.",
  ["landwirtschaftliches jahrbuch der schweiz"] = "Landwirtsch. Jahrb. Schweiz",
  ["langenbeck's archives of surgery"] = "Langenbecks Arch Surg",
  ["langenbecks archiv fur chirurgie"] = "Langenbecks Arch. Chir.",
  ["langenbecks archiv fur chirurgie. supplement ii, verhandlungen der deutschen gesellschaft fur chirurgie"] = "Langenbecks Arch. Chir. Suppl. II Verh. Dtsch. Ges. Chir.",
  ["langenbecks archiv fur chirurgie. supplement ii, verhandlungen der deutschen gesellschaft fur chirurgie. deutsche gesellschaft fur chirurgie. kongress"] = "Langenbecks Arch Chir Suppl II Verh Dtsch Ges Chir",
  ["langenbecks archiv fur chirurgie. supplement, kongressband"] = "Langenbecks Arch. Chir. Suppl. Kongressbd.",
  ["langenbecks archiv fur chirurgie. supplement. kongressband"] = "Langenbecks Arch Chir Suppl Kongressbd",
  ["langenbecks archiv fur chirurgie. supplement. kongressband. deutsche gesellschaft fur chirurgie. kongress"] = "Langenbecks Arch Chir Suppl Kongressbd",
  ["langenbecks archiv für chirurgie"] = "Langenbecks Arch Chir",
  ["langenbecks archiv für chirurgie. supplement ii, verhandlungen der deutschen gesellschaft für chirurgie. deutsche gesellschaft für chirurgie. kongress"] = "Langenbecks Arch Chir Suppl II Verh Dtsch Ges Chir",
  ["langenbecks archiv für chirurgie. supplement. kongressband. deutsche gesellschaft für chirurgie. kongress"] = "Langenbecks Arch Chir Suppl Kongressbd",
  ["langenbecks archive of surgery"] = "Langenbecks Arch Surg",
  ["langenbeck’s archives of surgery"] = "Langenbecks Arch. Surg.",
  ["langenbeck’s archives of surgery / deutsche gesellschaft fur chirurgie"] = "Langenbecks Arch Surg",
  ["langmuir : the acs journal of surfaces and colloids"] = "Langmuir",
  ["language & communication"] = "Lang Commun",
  ["language acquisition"] = "Lang Acquis",
  ["language and cognition"] = "Lang Cogn",
  ["language and cognitive processes"] = "Lang Cogn Process",
  ["language and communication disorders"] = "Lang Commun Disord",
  ["language and linguistics"] = "Lang Linguist (Taipei)",
  ["language and linguistics compass"] = "Lang Linguist Compass",
  ["language and literature (harlow, england)"] = "Lang Lit (Harlow)",
  ["language and sociocultural theory"] = "Lang Sociocult Theory",
  ["language and speech"] = "Lang Speech",
  ["language assessment quarterly"] = "Lang Assess Q",
  ["language in society"] = "Lang. Soc.",
  ["language learning"] = "Lang Learn",
  ["language learning & technology : ll & t"] = "Lang Learn Technol",
  ["language learning and development : the official journal of the society for language development"] = "Lang Learn Dev",
  ["language policy"] = "Lang Policy",
  ["language research"] = "Lang Res",
  ["language resources and evaluation"] = "Lang Resour Eval",
  ["language sciences (oxford, england)"] = "Lang Sci",
  ["language variation and change"] = "Lang Var Change",
  ["language, cognition and neuroscience"] = "Lang Cogn Neurosci",
  ["language, education and society"] = "Lang. Ed. Soc.",
  ["language, speech, and communication"] = "Lang. Speech Commun.",
  ["language, speech, and hearing services in schools"] = "Lang Speech Hear Serv Sch",
  ["languages (basel, switzerland)"] = "Languages (Basel)",
  ["languedoc médical"] = "Languedoc Med",
  ["langues orientales anciennes. philologie et linguistique"] = "LangOrAnc",
  ["lankesteriana : la revista científica del jardín botánico lankester, universidad de costa rica"] = "Lankesteriana",
  ["laparoscopic surgery"] = "Laparosc Surg",
  ["laparoscopic, endoscopic, and robotic surgery"] = "Laparosc Endosc Robot Surg",
  ["larc medical"] = "LARC Med.",
  ["larc médical"] = "LARC Med",
  ["large animal review"] = "Large Anim. Rev.",
  ["large-scale annotation of biomedical data and expert label synthesis and hardware aware learning for medical imaging and computer assisted intervention : international workshops, labels 2019, hal-miccai 2019, and curious 2019, held in conjunction with miccai 2019, shenzhen, china, october 13 and 17, 2019, proceedings. labels (workshop) (4th : 2019 : shenzhen shi, china)"] = "Large Scale Annot Biomed Data Export Label Synth Hardw Aware Learn Med Imaging Comput Assist Interv (2019)",
  ["large-scale assessments in education"] = "Large Scale Assess Educ",
  ["laryngo- rhino- otologie"] = "Laryngorhinootologie",
  ["laryngologie, rhinologie, otologie"] = "Laryngol Rhinol Otol (Stuttg)",
  ["laryngologie, rhinologie, otologie und ihre grenzgebiete"] = "Laryngol Rhinol Otol (Stuttg)",
  ["laryngoscope investigative otolaryngology"] = "Laryngoscope Investig Otolaryngol",
  ["laser & photonics reviews"] = "Laser Photonics Rev.",
  ["laser and particle beams"] = "Laser Part. Beams",
  ["laser and photonics reviews"] = "Laser Photonics Rev.",
  ["laser chemistry"] = "Laser Chem.",
  ["laser focus world"] = "Laser Focus World",
  ["laser nursing"] = "Laser Nurs",
  ["laser physics"] = "Laser Phys.",
  ["laser physics letters"] = "Laser Phys. Lett.",
  ["laser technik journal"] = "Laser Tech. J.",
  ["laser therapy"] = "Laser Ther",
  ["lasers in dental science"] = "Lasers Dent Sci",
  ["lasers in engineering"] = "Lasers Eng.",
  ["lasers in manufacturing and materials processing"] = "Lasers Manuf. Mater. Process.",
  ["lasers in medical science"] = "Lasers Med Sci",
  ["lasers in surgery and medicine"] = "Lasers Surg Med",
  ["lasers in surgery and medicine. supplement"] = "Lasers Surg Med Suppl",
  ["lasers in the life sciences"] = "Lasers Life Sci",
  ["late imperial china = chʻing shih wen tʻi"] = "Late Imp China",
  ["latin america applied research"] = "Lat. Am. Appl. Res.",
  ["latin american economic review"] = "Lat Am Econ Rev",
  ["latin american journal of aquatic research"] = "Lat Am J Aquat Res",
  ["latin american journal of economics"] = "Lat Am J Econ",
  ["latin american journal of pharmacy"] = "Lat. Am. J. Pharm.",
  ["latin american journal of solids and structures"] = "Lat. Am. J. Solids Struct.",
  ["latin american mathematics series"] = "Lat. Amer. Math. Ser.",
  ["latin american mathematics series—ufscar subseries"] = "Lat. Amer. Math. Ser. UFSCar Subser.",
  ["latin american perspectives"] = "Lat Am Perspect",
  ["latin american policy"] = "Lat Am Policy",
  ["latin american policy journal"] = "Lat. Am. Policy J.",
  ["latin american politics and society"] = "Lat Am Polit Soc",
  ["latin american population history bulletin"] = "Lat Am Popul Hist Bull",
  ["latin american population history newsletter"] = "Lat Am Popul Hist News",
  ["latin american research review"] = "Lat Am Res Rev",
  ["latinitas. commentarii linguae latinae excolendae provehendae"] = "Latinitas",
  ["latino studies"] = "Lat Stud",
  ["latinskai︠a︡ amerika = américa latina"] = "Lat Am",
  ["latomus. revue d’études latines"] = "Latomus",
  ["latvian journal of physics and technical sciences"] = "Latv. J. Phys. Tech. Sci.",
  ["latvijas kimijas zurnals"] = "Latv. Kim. Z.",
  ["latvijas zinātn̦u akadēmijas vēstis. a dal̦a, sociālās un humanitārās zinātnes = proceedings of the latvian academy of sciences. section a"] = "Latv Zinat Akad Vestis A Humanit Zinat",
  ["latviy skaya akademiya nauk"] = "Latv. Mat. Ezhegodnik",
  ["laufener seminarbeiträge"] = "Laufener Semin.beitr.",
  ["laundry and dry cleaning journal of canada"] = "Laundry Dry Clean. J. Can.",
  ["laundry news"] = "Laund News",
  ["laval medical"] = "Laval Med.",
  ["laval médical"] = "Laval Med",
  ["laval théologique et philosophique"] = "LThPh",
  ["laverna. beiträge zur wirtschafts- und sozialgeschichte der alten welt"] = "Laverna",
  ["lavori dell'istituto di anatomia e istologia patologica, università degli studi di perugia"] = "Lav Ist Anat Istol Patol Univ Studi Perugia",
  ["lavori dell’istituto di anatomia e istologia patologica, universita degli studi di perugia"] = "Lav. Ist. Anat. Istol. Patol. Univ. Studi Perugia",
  ["lavoro umano"] = "Lav Um",
  ["law & inequality"] = "Law Inequal",
  ["law & justice"] = "Law Justice",
  ["law & liberty"] = "Law Lib",
  ["law & policy"] = "Law Policy",
  ["law & policy quarterly"] = "Law Policy Q",
  ["law & social inquiry : journal of the american bar foundation"] = "Law Soc Inq",
  ["law & society review"] = "Law Soc Rev",
  ["law and contemporary problems"] = "Law Contemp Probl",
  ["law and human behavior"] = "Law Hum Behav",
  ["law and mental health"] = "Law Ment Health",
  ["law and philosophy"] = "Law Philos",
  ["law and policy in international business"] = "Law Policy Int Bus",
  ["law and psychology review"] = "Law Psychol Rev",
  ["law and the human genome review"] = "Law Hum. Genome Rev.",
  ["law enforcement executive forum"] = "Law Enforc Exec Forum",
  ["law file"] = "Law File",
  ["law in context (bundoora, vic.)"] = "Law Context",
  ["law institute journal : the official organ of the law institute of victoria"] = "Law Inst J",
  ["law library journal"] = "Law Libr J",
  ["law review"] = "Law Rev",
  ["law, innovation and technology"] = "Law Innov Technol",
  ["law, medicine & health care : a publication of the american society of law & medicine"] = "Law Med Health Care",
  ["law, medicine and health care"] = "Law. Med. Health Care",
  ["law, probability & risk : a journal of reasoning under uncertainty"] = "Law Probab Risk",
  ["law, probability, and risk"] = "Law Probab. Risk",
  ["law-medicine news"] = "Law Med News",
  ["lawasia. law association for asia and the western pacific"] = "Lawasia",
  ["laws of puerto rico annotated. puerto rico"] = "Laws P R Annot P R",
  ["laws of the state of new mexico. new mexico"] = "Laws State N M N M",
  ["laws of the state of oregon. oregon"] = "Laws State Or Or",
  ["laws. illinois"] = "Laws Ill",
  ["lc gc : magazine of liquid and gas chromatography"] = "LC GC",
  ["lc gc europe"] = "LC GC Eur",
  ["lc gc north america"] = "LC GC N Am",
  ["lc-gc the magazine of separation science"] = "LC-GC",
  ["lcgc north america"] = "LCGC North Am.",
  ["lda journal"] = "LDA J",
  ["ldi issue brief"] = "LDI Issue Brief",
  ["le caducée"] = "Caducee",
  ["le centre-est medical"] = "Cent Est Med",
  ["le centre-est médical"] = "Cent Est Med",
  ["le chirurgien-dentiste de france"] = "Chir Dent Fr",
  ["le contrat social"] = "Contrat Soc",
  ["le corps medical"] = "Corps Med (Ettelbruck)",
  ["le corps médical"] = "Corps Med (Ettelbruck)",
  ["le courrier de l'unesco"] = "Courr Unesco",
  ["le courrier de l’unesco"] = "Courr Unesco",
  ["le courrier des pays de l'est"] = "Courr Pays Est",
  ["le courrier des pays de l’est / publie avec le concours du centre de documentation sur l’u.r.s.s. et les pays slaves de l’ecole pratique des hautes etudes"] = "Courr Pays Est",
  ["le debat"] = "Debat",
  ["le developpement voltaique"] = "Dev Volta",
  ["le débat"] = "Debat",
  ["le développement voltaïque"] = "Dev Volta",
  ["le diabete"] = "Diabete",
  ["le diabète"] = "Diabete",
  ["le genre humain"] = "GH",
  ["le infezioni in medicina"] = "Infez Med",
  ["le infezioni in medicina : rivista periodica di eziologia, epidemiologia, diagnostica, clinica e terapia delle patologie infettive"] = "Infez Med",
  ["le journal cannt = cannt journal : the journal of the canadian association of nephrology nurses and technicians"] = "J CANNT",
  ["le journal dentaire belge. belgisch blad voor tandheelkunde"] = "J Dent Belge",
  ["le journal dentaire du quebec"] = "J Dent Que",
  ["le journal dentaire du québec"] = "J Dent Que",
  ["le journal des médecines cunéiformes"] = "J Med Cuneif",
  ["le journal medical libanais. the lebanese medical journal"] = "J Med Liban",
  ["le journal médical libanais. the lebanese medical journal"] = "J Med Liban",
  ["le lait"] = "Lait",
  ["le livre et l'estampe"] = "Livre Estampe",
  ["le livre et l’estampe"] = "Livre Estampe",
  ["le mali medical"] = "Mali Med",
  ["le mali médical"] = "Mali Med",
  ["le marche. archeologia, storia, territorio"] = "Marche",
  ["le matematiche"] = "Matematiche (Catania)",
  ["le medecin de reserve"] = "Med Reserve (Paris)",
  ["le medecin d’usine; revue d’hygiene industrielle et des maladies professionnelles"] = "Med Usine Rev Hyg Ind Mal Prof",
  ["le medecin generaliste de france"] = "Med Gen Fr",
  ["le medecin libre"] = "Med Libre",
  ["le médecin d'usine; revue d'hygiène industrielle et des maladies professionnelles"] = "Med Usine Rev Hyg Ind Mal Prof",
  ["le médecin de réserve"] = "Med Reserve (Paris)",
  ["le médecin généraliste de france"] = "Med Gen Fr",
  ["le médecin libre"] = "Med Libre",
  ["le monde alpin et rhodanien"] = "Monde Alp Rhodan",
  ["le monde dentaire"] = "Monde Dent",
  ["le monde médical"] = "Monde Med",
  ["le moniteur des pharmacies et des laboratoires"] = "Monit Pharm Lab",
  ["le mouvement social"] = "Mouv Soc",
  ["le moyen âge: revue trimestrielle d’histoire et de philologie"] = "MA",
  ["le museon"] = "Museon",
  ["le muséon"] = "Museon",
  ["le muséon. revue d’études orientales"] = "Museon",
  ["le nourrisson"] = "Nourrisson",
  ["le pharmacien biologiste"] = "Pharm Biol",
  ["le poumon"] = "Poumon",
  ["le poumon et le coeur"] = "Poumon Coeur",
  ["le praticien en anesthésie réanimation"] = "Prat Anesth Reanim",
  ["le progres medical"] = "Progr Med (Paris)",
  ["le progres odonto-stomatologique"] = "Prog Odontostomatol",
  ["le progrés médical"] = "Prog Med (Paris)",
  ["le progrés odonto-stomatologique"] = "Prog Odontostomatol",
  ["le saguenay medical"] = "Saguenay Med",
  ["le saguenay médical"] = "Saguenay Med",
  ["le sang"] = "Sang",
  ["le savoir de mantice"] = "Savoir Mantice",
  ["le scalpel"] = "Scalpel (Brux)",
  ["le sud medical et chirurgical"] = "Sud Med Chir",
  ["le sud médical et chirurgical"] = "Sud Med Chir",
  ["le technicien belge en prothese dentaire"] = "Tech Belge Prothese Dent",
  ["le technicien belge en prothèse dentaire"] = "Tech Belge Prothese Dent",
  ["le temps de la réflexion"] = "TR",
  ["le travail humain"] = "Trav Hum",
  ["leadership (london)"] = "Leadership (Lond)",
  ["leadership and management in engineering"] = "Leadersh. Manage. Eng.",
  ["leadership and policy in schools"] = "Leadersh Policy Sch",
  ["leadership in health services"] = "Leadersh. Health Serv.",
  ["leadership in health services (bradford, england)"] = "Leadersh Health Serv (Bradf Engl)",
  ["leadership in health services = leadership dans les services de santé"] = "Leadersh Health Serv",
  ["leading edge"] = "Leading Edge",
  ["leading edge (tulsa, okla. : online)"] = "Lead Edge",
  ["leaflet agricultural research organization, division of forestry"] = "Leafl. Agric. Res. Organ., Div. For.",
  ["league exchange"] = "League Exch",
  ["league lines"] = "League Lines",
  ["lean construction journal"] = "Lean Construct. J.",
  ["learned publishing"] = "Learned Publ.",
  ["learned publishing : journal of the association of learned and professional society publishers"] = "Learn Publ",
  ["learning & behavior"] = "Learn Behav",
  ["learning & behavior : a psychonomic society publication"] = "Learn Behav",
  ["learning & memory (cold spring harbor, n.y.)"] = "Learn Mem",
  ["learning & perception"] = "Learn Percept",
  ["learning and behavior"] = "Learn. Behav.",
  ["learning and individual differences"] = "Learn Individ Differ",
  ["learning and instruction"] = "Learn Instr",
  ["learning and memory"] = "Learn. Mem.",
  ["learning and motivation"] = "Learn Motiv",
  ["learning curve"] = "Learn. Curve",
  ["learning disabilities (pittsburgh, pa.)"] = "Learn Disabil (Pittsbg)",
  ["learning disabilities (weston, mass.)"] = "Learn Disabil",
  ["learning disabilities research & practice : a publication of the division for learning disabilities, council for exceptional children"] = "Learn Disabil Res Pract",
  ["learning disability quarterly : journal of the division for children with learning disabilities"] = "Learn Disabil Q",
  ["learning environments research"] = "Learn Environ Res",
  ["learning health systems"] = "Learn Health Syst",
  ["learning in doing: social, cognitive, and computational perspectives"] = "Learn. Doing Soc. Cogn. Comput. Perspect.",
  ["learning landscapes"] = "Learn Landsc",
  ["learning materials in biosciences"] = "Learn. Mater. Biosci.",
  ["learning, culture and social interaction"] = "Learn Cult Soc Interact",
  ["learning, media and technology"] = "Learn Media Technol",
  ["lebens-kunst; cademario-nachrichten"] = "Lebens Kunst Cademario Nachr",
  ["lebensmittel-wissenschaft + [i.e. und] technologie. food science + technology. science + technologie alimentaire"] = "Lebensm Wiss Technol",
  ["lebensmittel-wissenschaft und -technologie"] = "Lebensm.-Wiss. Technol.",
  ["lebensmittel-wissenschaft und technologie"] = "Lebensm. Wiss. Technol.",
  ["lebensversicherungs medizin"] = "Lebensversicher Med",
  ["leber, magen, darm"] = "Leber Magen Darm",
  ["lecturas de economia"] = "Lecturas Econ.",
  ["lecturas matematicas"] = "Lect. Mat.",
  ["lecturas matemáticas"] = "Lect. Mat.",
  ["lecture and review series. naval medical research institute (u.s.)"] = "Lect Rev Ser",
  ["lecture notes from the first brazilian school on computer algebra"] = "Lecture Notes First Braz. School Comput. Algebra",
  ["lecture notes in applied and computational mechanics"] = "Lect. Notes Appl. Comput. Mech.",
  ["lecture notes in applied and numerical harmonic analysis"] = "Lect. Notes Appl. Numer. Harmon. Anal.",
  ["lecture notes in applied mathematics and mechanics"] = "Lect. Notes Appl. Math. Mech.",
  ["lecture notes in artif. intell."] = "Lecture Notes in Artif. Intell.",
  ["lecture notes in bioengineering"] = "Lect. Notes Bioeng.",
  ["lecture notes in bioinformatics"] = "Lect. Notes in Bioinform.",
  ["lecture notes in biomathematics"] = "Lect. Notes Biomath.",
  ["lecture notes in business information processing"] = "Lect. Notes Bus. Inf. Process.",
  ["lecture notes in chemistry"] = "Lecture Notes in Chem.",
  ["lecture notes in civil engineering"] = "Lect. Notes Civ. Eng.",
  ["lecture notes in computational science and engineering"] = "Lect. Notes Comput. Sci. Eng.",
  ["lecture notes in computational vision and biomechanics"] = "Lect. Notes Comput. Vis. Biomech.",
  ["lecture notes in computer science"] = "Lect. Notes Comput. Sci.",
  ["lecture notes in control and information sciences"] = "Lect. Notes Control Inf. Sci.",
  ["lecture notes in control and information sciences. proceedings"] = "Lect. Notes Control Inf. Sci. Proc.",
  ["lecture notes in earth sciences"] = "Lect. Notes Earth Sci.",
  ["lecture notes in earth system sciences"] = "Lect. Notes Earth Syst. Sci.",
  ["lecture notes in economics and mathematical systems"] = "Lecture Notes in Econom. and Math. Systems",
  ["lecture notes in electrical engineering"] = "Lect. Notes Electr. Eng.",
  ["lecture notes in energy"] = "Lect. Notes Energy",
  ["lecture notes in geosystems mathematics and computing"] = "Lect. Notes Geosyst. Math. Comput.",
  ["lecture notes in intelligent transportation and infrastructure"] = "Lect. Notes Intell. Transp. Infrastruct.",
  ["lecture notes in logic"] = "Lect. Notes Log.",
  ["lecture notes in logistics"] = "Lect. Notes Logist.",
  ["lecture notes in mathematical fluid mechanics"] = "Lect. Notes Math. Fluid Mech.",
  ["lecture notes in mathematics"] = "Lect. Notes Math.",
  ["lecture notes in mechanical engineering"] = "Lect. Notes Mech. Eng.",
  ["lecture notes in mechanics"] = "Lect. Notes Mech.",
  ["lecture notes in medical informatics"] = "Lect Notes Med Inform",
  ["lecture notes in mobility"] = "Lect. Notes Mobil.",
  ["lecture notes in morphogenesis"] = "Lect. Notes Morphog.",
  ["lecture notes in networks and systems"] = "Lect. Notes Netw. Syst.",
  ["lecture notes in nonlinear analysis"] = "Lect. Notes Nonlinear Anal.",
  ["lecture notes in numerical and applied analysis"] = "Lecture Notes Numer. Appl. Anal.",
  ["lecture notes in operations research"] = "Lect. Notes Oper. Res.",
  ["lecture notes in physics"] = "Lect. Notes Phys.",
  ["lecture notes in physics monographs"] = "Lect. Notes Phys. Monogr.",
  ["lecture notes in pure and applied mathematics"] = "Lecture Notes in Pure and Appl. Math.",
  ["lecture notes in social networks"] = "Lect. Notes Soc. Netw.",
  ["lecture notes in statistics"] = "Lect. Notes Stat.",
  ["lecture notes in statistics—proceedings"] = "Lect. Notes Stat. Proc.",
  ["lecture notes of seminario interdisciplinare di matematica"] = "Lect. Notes Semin. Interdiscip. Mat.",
  ["lecture notes of tbilisi international centre of mathematics and informatics"] = "Lect. Notes TICMI",
  ["lecture notes of the institute for computer sciences, social informatics and telecommunications engineering"] = "Lect. Notes Inst. Comput. Sci. Soc. Inform. Telecommun. Eng.",
  ["lecture notes of the unione matematica italiana"] = "Lect. Notes Unione Mat. Ital.",
  ["lecture notes on data engineering and communications technologies"] = "Lect. Notes Data Eng. Commun. Technol.",
  ["lecture notes on mathematical modelling in the life sciences"] = "Lect. Notes Math. Model. Life Sci.",
  ["lecture notes on numerical methods in engineering and sciences"] = "Lect. Notes Numer. Methods Eng. Sci.",
  ["lecture notes series"] = "Lecture Notes Ser.",
  ["lecture notes series on computing"] = "Lecture Notes Ser. Comput.",
  ["lecture notes series. institute for mathematical sciences. national university of singapore"] = "Lect. Notes Ser. Inst. Math. Sci. Natl. Univ. Singap.",
  ["lecture notes-monograph series"] = "Lect Notes Monogr Ser",
  ["lectures in applied mathematics"] = "Lectures in Appl. Math.",
  ["lectures in mathematics eth zurich"] = "Lectures Math. ETH Zurich",
  ["lectures in mathematics eth zürich"] = "Lectures Math. ETH Zürich",
  ["lectures in pure and applied mathematics"] = "Lect. Pure Appl. Math.",
  ["lectures on the scientific basis of medicine"] = "Lect Sci Basis Med",
  ["leeds dental journal"] = "Leeds Dent J",
  ["legal and criminological psychology"] = "Legal Criminol Psychol",
  ["legal aspects of medical practice"] = "Leg Aspects Med Pract",
  ["legal ethics (oxford, england)"] = "Leg Ethics",
  ["legal medical quarterly"] = "Leg Med Q",
  ["legal medicine"] = "Leg Med",
  ["legal medicine (tokyo, japan)"] = "Leg Med (Tokyo)",
  ["legal medicine annual"] = "Leg Med Annu",
  ["legal reference services quarterly"] = "Legal Ref Serv Q",
  ["legal studies (society of legal scholars)"] = "Leg Stud (Soc Leg Scholars)",
  ["legal studies (society of public teachers of law)"] = "Leg Stud",
  ["lege artis medicinae : uj magyar orvosi hirmondo"] = "Lege Artis Med",
  ["lege artis medicinae : új magyar orvosi hírmondó"] = "Lege Artis Med",
  ["legume genomics and genetics"] = "Legume Genomics Genet",
  ["legume research"] = "Legume Res.",
  ["legume science"] = "Legume Sci.",
  ["lehr- und handbucher zu geld, borse, bank und versicherung"] = "Lehr- Handb. Geld Borse Bank Versicher.",
  ["lehrbuch mathematik"] = "Lehrbuch Math.",
  ["lehrbücher und monographien zur didaktik der mathematik"] = "Lehrbücher Monogr. Didakt. Math.",
  ["leibniz from oxford"] = "Leibniz Oxford",
  ["leisure (waterloo, ont.)"] = "Leisure (Waterloo)",
  ["leisure science"] = "Leisure Sci.",
  ["leisure sciences"] = "Leis Sci",
  ["leitfaden der angewandten informatik"] = "Leitfaden Angew. Inform.",
  ["leitfäden der angewandten mathematik und mechanik"] = "Leitfäden Angew. Math. Mech.",
  ["leitfäden der informatik"] = "Leitfäden Inform.",
  ["leitfäden und monographien der informatik"] = "Leitfäden Monogr. Inform.",
  ["leitz, mitteilungen für wissenschaft und technik"] = "Leitz, Mitt. Wiss. Tech.",
  ["lekarska veda v zahranici"] = "Lek. Veda Zahr.",
  ["lekarske listy"] = "Lek List",
  ["lekarske prace"] = "Lek. Pr.",
  ["lekarsky obzor"] = "Lek Obz",
  ["lekarz wojskowy"] = "Lek Wojsk",
  ["lekárske práce"] = "Lek Pr",
  ["lekársky obzor"] = "Lek Obz",
  ["lemur news"] = "Lemur News",
  ["leningradskiy gosudarstvennyy ordena lenina universitet imeni a"] = "Leningrad. Gos. Univ. Uchen. Zap. Ser. Mat. Nauk",
  ["leningradskiy gosudarstvennyy universitet"] = "Vychisl. Tekhn. Voprosy Kibernet.",
  ["leningradskiy ordena lenina gosudarstvennyy universitet imeni a"] = "Metody Vychisl.",
  ["leningradskiy universitet"] = "Voprosy Teor. Sistem Avtomat. Upravleniya",
  ["lens and eye toxicity research"] = "Lens Eye Toxic Res",
  ["lens research"] = "Lens Res",
  ["leonardo (oxford, england)"] = "Leonardo (Oxf)",
  ["leonhardi euleri opera omnia, series quarta a: commercium epistolicum"] = "Leonhardi Euleri Opera Omnia Ser. 4A",
  ["lepidoptera novae"] = "Lepid Novae",
  ["leprosy in india"] = "Lepr India",
  ["leprosy review"] = "Lepr Rev",
  ["les annales archéologiques arabes syriennes"] = "AAS",
  ["les annales d'oto-laryngologie"] = "Ann Otolaryngol",
  ["les annales de l'i.f.o.r.d"] = "Ann IFORD",
  ["les annales de l’i.f.o.r.d"] = "Ann IFORD",
  ["les annales de médecine sociale"] = "Ann Med Soc",
  ["les annales des pays nivernais"] = "AnnNivern",
  ["les annales d’oto-laryngologie"] = "Ann Otolaryngol",
  ["les annales homeopathiques francaises"] = "Ann Homeopath Fr",
  ["les annales homéopathiques françaises"] = "Ann Homeopath Fr",
  ["les annales médico-chirurgicales de l'hôpital sainte-justine de montréal"] = "Ann Med Chir Hop Sainte Justine Montr",
  ["les bronches"] = "Bronches",
  ["les cahiers d'outre-mer"] = "Cah O M",
  ["les cahiers de fontenay"] = "Cah Fontenay",
  ["les cahiers de l'alliance israélite universelle"] = "Cah All Isr Univers",
  ["les cahiers de l'analyse des données"] = "Cah Anal Donnees",
  ["les cahiers de l'année gérontologique"] = "Cah Annee Gerontol",
  ["les cahiers de l'iroise"] = "Cah Iroise",
  ["les cahiers de l'orient"] = "Cah Orient",
  ["les cahiers de la recherche architecturale"] = "Cah Rech Archit",
  ["les cahiers de l’alliance israelite universelle"] = "Cah All Isr Univers",
  ["les cahiers de l’analyse des donnees"] = "Cah Anal Donnees",
  ["les cahiers de l’iroise"] = "Cah Iroise",
  ["les cahiers de l’orient"] = "Cah Orient",
  ["les cahiers de mariemont. bulletin du musée royal de mariemont"] = "CahMariemont",
  ["les cahiers de medecine"] = "Cah Med",
  ["les cahiers de medecine veterinaire"] = "Cah Med Vet",
  ["les cahiers de médecine"] = "Cah Med",
  ["les cahiers de médecine vétérinaire"] = "Cah Med Vet",
  ["les cahiers de prothese"] = "Cah Prothese",
  ["les cahiers de prothèse"] = "Cah Prothese",
  ["les cahiers de tunisie"] = "Cah Tunis",
  ["les cahiers des dix. les dix"] = "Cah Dix",
  ["les cahiers du bazadais"] = "Cah Bazadais",
  ["les cahiers du college de medecine des hopitaux de paris"] = "Cah Coll Med Hop Paris",
  ["les cahiers du collège de médecine des hôpitaux de paris"] = "Cah Coll Med Hop Paris",
  ["les cahiers du comité consultatif national d'éthique pour les sciences de la vie et de la santé"] = "Cah Com Consult Natl Ethique Sci Sante",
  ["les cahiers du nursing"] = "Cah Nurs",
  ["les cahiers d’outre-mer"] = "Cah O M",
  ["les cahiers haut-marnais"] = "Cah Haut Marnais",
  ["les cahiers lorrains"] = "Cah Lorrains",
  ["les cahiers rationalistes"] = "Cah Ration",
  ["les carnets de l'enfance = assignment children"] = "Carnets Enfance",
  ["les cours du cirm"] = "Cours CIRM",
  ["les dossiers du centre régional archéologique d’alet"] = "DossAlet",
  ["les dossiers d’archéologie"] = "DossAParis",
  ["les entretiens de bichat. medecine. entretiens de bichat"] = "Entret Bichat Med Entret Bichat",
  ["les entretiens de bichat. médecine. entretiens de bichat"] = "Entret Bichat Med Entret Bichat",
  ["les grandes voix de la recherche"] = "Gd. Voix Rech.",
  ["les grands classiques gauthier-villars"] = "Grands Class. Gauthier-Villars",
  ["les grands noms de la science et de la technique"] = "Grands Noms Sci. Tech.",
  ["les journées scientifiques du centre national de coordination des études et recherches sur la nutrition et l'alimentation"] = "Journ Sci Cent Nat Coord Etud Rech Nutr Aliment",
  ["les nouveaux rendez-vous d’archimede"] = "Nouv. Rendez-vous Archimede",
  ["les nouvelles. licensing executives society (u.s.a.)"] = "LES Nouv",
  ["les revues medicales normandes"] = "Rev Med Normandes",
  ["les revues médicales normandes"] = "Rev Med Normandes",
  ["les sciences hors d’occident au xxe siècle"] = "Sci. Occident XXe Siècle",
  ["les temps modernes"] = "Temps Mod",
  ["les études classiques"] = "LEC",
  ["les études classiques. revue trimestrielle de recherche et d’enseignement"] = "EtCl",
  ["letras de hoje"] = "Let Hoje",
  ["lettera matematica"] = "Lett. Mat.",
  ["lettera matematica pristem"] = "Lett Mat Pristem",
  ["lettere al nuovo cimento"] = "Lett. Nuovo Cimento",
  ["lettere italiane"] = "Lett Ital",
  ["letters in applied microbiology"] = "Lett Appl Microbiol",
  ["letters in biomathematics"] = "Lett. Biomath.",
  ["letters in drug design & discovery"] = "Lett. Drug Des. Discovery",
  ["letters in functional foods"] = "Lett. Funct. Foods",
  ["letters in health & biological sciences"] = "Lett Health Biol Sci",
  ["letters in heat and mass transfer"] = "Lett. Heat Mass Transfer",
  ["letters in high energy physics"] = "Lett. High Energy Phys.",
  ["letters in mathematical physics"] = "Lett. Math. Phys.",
  ["letters in organic chemistry"] = "Lett. Org. Chem.",
  ["letters in peptide science"] = "Lett. Pept. Sci.",
  ["letters in peptide science : lips"] = "Lett Pept Sci",
  ["letters in spatial and resource sciences"] = "Lett Spat Resour Sci",
  ["letters on materials"] = "Lett. Mater.",
  ["lettres orientales et classiques"] = "Lett. Orient. Class.",
  ["leukemia & lymphoma"] = "Leuk Lymphoma",
  ["leukemia : official journal of the leukemia society of america, leukemia research fund, u.k"] = "Leukemia",
  ["leukemia and lymphoma"] = "Leuk. Lymphoma",
  ["leukemia research"] = "Leuk Res",
  ["leukemia research and treatment"] = "Leuk Res Treatment",
  ["leukemia research reports"] = "Leuk Res Rep",
  ["leukemia supplements"] = "Leuk Suppl",
  ["leuven notes in mathematical and theoretical physics"] = "Leuven Notes Math. Theoret. Phys. Ser. B Theoret. Particle Phys.",
  ["leuvensche bijdragen"] = "Leuvense Bijdr",
  ["levant. journal of the british school of archaeology in jerusalem and the british institute at amman for archaeology and history"] = "Levant",
  ["leveltari kozlemenyek"] = "Leveltari Kozl",
  ["leveltari szemle"] = "Leveltari Sz",
  ["levéltári közlemények"] = "Leveltari Kozl",
  ["levéltári szemle"] = "Leveltari Sz",
  ["lex et scientia"] = "Lex Sci",
  ["lexikon der ägyptologie"] = "LAe",
  ["lexikon iconographicum mythologiae classicae"] = "LIMC",
  ["lexikon topographicum urbis romae"] = "LTUR",
  ["lexington books series in computer science"] = "Lexington Books Ser. Comput. Sci.",
  ["lexis district of columbia code. district of columbia"] = "Lexis DC Code DC",
  ["lexis. poetica, retorica e comunicazione nella tradizione classica"] = "Lexis",
  ["leyte-samar studies"] = "Leyte Samar Stud",
  ["lezioni lincee"] = "Lezioni Lincee",
  ["lékar̆ská vĕda v zahranic̆í"] = "Lek Veda Zahr",
  ["lékařské listy"] = "Lek List",
  ["lgbt health"] = "LGBT Health",
  ["lhb hydroscience journal"] = "LHB Hydrosci. J.",
  ["li shi dang an"] = "Li Shi Dang An",
  ["li shi yu yan yan jiu suo ji kan"] = "Li Shi Yu Yan Yan Jiu Suo Ji Kan",
  ["liang you shi pin ke ji"] = "Liang You Shi Pin Ke Ji",
  ["liberation (new york, n.y.)"] = "Liberation",
  ["liberian studies journal"] = "Liberian Stud J",
  ["libertas mathematica"] = "Lib Math",
  ["libertas mathematica. new series"] = "Lib. Math. (N.S.)",
  ["liberty, life, and family"] = "Lib Life Fam",
  ["libraries & culture"] = "Libr Cult",
  ["libraries & the cultural record"] = "Libr Cult Rec",
  ["library & information science research"] = "Libr Inf Sci Res",
  ["library association record. library association"] = "Libr Assoc Rec",
  ["library chronicle"] = "Libr Chron",
  ["library history"] = "Libr Hist",
  ["library journal"] = "Libr J",
  ["library of congress acquisitions. manuscript division. library of congress. manuscript division"] = "Lib Congr Acquis Manuscr Div",
  ["library of engineering mathematics"] = "Lib. Engrg. Math.",
  ["library of theoria"] = "Libr. Theor.",
  ["library resources & technical services"] = "Libr Resour Tech Serv",
  ["library resources and technical services"] = "Libr. Resources Tech. Serv.",
  ["library review. great britain. forestry commission"] = "Libr Rev (Lond)",
  ["library science with a slant to documentation"] = "Lib Sci Slant Doc",
  ["library technology reports"] = "Libr Technol Rep",
  ["library trends"] = "Libr Trends",
  ["libya antiqua"] = "LibyaAnt",
  ["libyan economic and business review"] = "Libyan Econ. Bus. Rev.",
  ["libyan studies"] = "LibStud",
  ["libyca. bulletin du service des antiquités. archéologie, épigraphie"] = "LibycaBServAnt",
  ["libyca. travaux du laboratoire d’anthropologie et d’archéologie préhistorique du musée du bardo"] = "LibycaTrav",
  ["lichenologist (london, england)"] = "Lichenologist (Lond)",
  ["lick observatory bulletin"] = "Lick Obs. Bull.",
  ["liebigs annalen - recueil"] = "Liebigs Ann. Recl.",
  ["liebigs annalen der chemie"] = "Liebigs Ann. Chem.",
  ["lietuvių tautos praeitis. lithuanian historical review"] = "Liet Tautos Praeit",
  ["lietuvos fizikos rinkinys"] = "Liet. Fiz. Rinkinys",
  ["lietuvos istorijos metrastis"] = "Liet Istor Metrast",
  ["lietuvos istorijos metraštis"] = "Liet Istor Metrast",
  ["lietuvos tsr aukstuju mokyklu mokslo darbai: biologija"] = "Liet Tsr Aukst Mokyklu Mokslo Darb",
  ["lietuvos tsr auks̆tu̧ju̧ mokyklu̧ mokslo darbai: biologija"] = "Liet Tsr Aukst Mokyklu Mokslo Darb",
  ["life (basel, switzerland)"] = "Life (Basel)",
  ["life (chicago, ill. : 1978)"] = "Life",
  ["life (jaipur, india)"] = "Life (Jaipur)",
  ["life and health"] = "Life Health",
  ["life and learning. university faculty for life. conference"] = "Life Learn Univ Fac Life Conf",
  ["life cycle reliability and safety engineering"] = "Life Cycle Reliab. Saf. Eng.",
  ["life in extreme environments"] = "Life Extreme Environ.",
  ["life science advances"] = "Life Sci Adv",
  ["life science advances. biochemistry : a journal of the council of scientific research integration"] = "Life Sci Adv Biochem",
  ["life science advances. experimental and clinical endocrinology : a journal of the council of scientific research integration"] = "Life Sci Adv Exp Clin Endocrinol",
  ["life science advances. genetics : a journal of the council of scientific research integration"] = "Life Sci Adv Genet",
  ["life science advances. plant physiology : a journal of the council of scientific research integration"] = "Life Sci Adv Plant Physiol",
  ["life science advances. steroid biochemistry : a journal of the council of scientific research integration"] = "Life Sci Adv Steroid Biochem",
  ["life science alliance"] = "Life Sci Alliance",
  ["life science journal"] = "Life Sci J",
  ["life sciences"] = "Life Sci",
  ["life sciences (1962)"] = "Life Sci (1962)",
  ["life sciences and space research"] = "Life Sci Space Res",
  ["life sciences in space research"] = "Life Sci Space Res (Amst)",
  ["life sciences, society and policy"] = "Life Sci Soc Policy",
  ["life sciences. part 1, physiology and pharmacology"] = "Life Sci. I",
  ["life sciences. part 2, biochemistry, general and molecular biology"] = "Life Sci. II",
  ["life sciences. pt. 1: physiology and pharmacology"] = "Life Sci I",
  ["life sciences. pt. 2: biochemistry, general and molecular biology"] = "Life Sci II",
  ["life span and disability"] = "Life Span Disabil",
  ["life support & biosphere science : international journal of earth space"] = "Life Support Biosph Sci",
  ["life support systems"] = "Life Support Syst.",
  ["life support systems : the journal of the european society for artificial organs"] = "Life Support Syst",
  ["life-threatening behavior"] = "Life Threat Behav",
  ["life: the excitement of biology"] = "Life Excit Biol",
  ["lifecycle journal"] = "Lifecycle J.",
  ["lifelong learning, the adult years"] = "Lifelong Learn Adult Years",
  ["lifelong learning, the adult years / adult education association of the united states of america"] = "Lifelong Learn Adult Years",
  ["lifelong learning; the adult years"] = "Lifelong Learn. Adult Years",
  ["lifestyle genomics"] = "Lifestyle Genom",
  ["lifetime data analysis"] = "Lifetime Data Anal",
  ["lifetime data analysis. an international journal devoted to statistical methods and applications for time-to-event data"] = "Lifetime Data Anal.",
  ["liga uruguaya contra la tuberculosis"] = "Liga Urug Contra Tuberc",
  ["light metals (warrendale, pennsylvania)"] = "Light Met. (Warrendale, Pa.)",
  ["light, science & applications"] = "Light Sci Appl",
  ["light: science & applications"] = "Light Sci. Appl.",
  ["lighting journal"] = "Light. J.",
  ["lighting research & technology (london, england : 2001)"] = "Light Res Technol",
  ["lighting research and technology"] = "Light. Res. Technol.",
  ["lightweight design"] = "Lightweight Des.",
  ["lightweight design worldwide"] = "Lightweight Des. Worldwide",
  ["lihua jianyan, huaxue fence"] = "Lihua Jianyan, Huaxue Fence",
  ["liiketaloudellinen aikakauskirja"] = "Liiketaloudellinen Aikak.",
  ["lijecnicki vjesnik"] = "Lijec Vjesn",
  ["lijec̆nic̆ki vjesnik"] = "Lijec Vjesn",
  ["likars'ka sprava"] = "Lik Sprava",
  ["likarska sprava"] = "Lik. Sprava",
  ["likars’ka sprava / ministerstvo okhorony zdorov’ia ukrainy"] = "Lik Sprava",
  ["lilith (fitzroy, vic.)"] = "Lilith",
  ["lille chirurgical"] = "Lille Chir",
  ["lille medical"] = "Lille Med.",
  ["lille medical : journal de la faculte de medecine et de pharmacie de l’universite de lille"] = "Lille Med",
  ["lille médical : journal de la faculté de médecine et de pharmacie de l'université de lille"] = "Lille Med",
  ["limesforschungen. studien zur organisation der römischen reichsgrenze an rhein und donau"] = "Limesforschungen",
  ["limnologica (online)"] = "Limnologica",
  ["limnologie aktuell"] = "Limnol. Akt.",
  ["limnology and freshwater biology"] = "Limnol Freshw Biol",
  ["limnology and oceanography"] = "Limnol Oceanogr",
  ["limnology and oceanography bulletin"] = "Limnol. Oceanogr. Bull.",
  ["limnology and oceanography e-lectures"] = "Limnol. Oceanogr. e-Lect.",
  ["limnology and oceanography letters"] = "Limnol Oceanogr Lett",
  ["limnology and oceanography, methods"] = "Limnol Oceanogr Methods",
  ["limnology and oceanography: fluids and environments"] = "Limnol. Oceanogr. Fluids Environ.",
  ["limnology and oceanography: methods"] = "Limnol. Oceanogr. Methods",
  ["lin chuang er bi yan hou ke za zhi (journal of clinical otorhinolaryngology)"] = "Lin Chuang Er Bi Yan Hou Ke Za Zhi",
  ["lin chuang er bi yan hou ke za zhi = journal of clinical otorhinolaryngology"] = "Lin Chuang Er Bi Yan Hou Ke Za Zhi",
  ["lin chuang er bi yan hou tou jing wai ke za zhi = journal of clinical otorhinolaryngology, head, and neck surgery"] = "Lin Chung Er Bi Yan Hou Tou Jing Wai Ke Za Zhi",
  ["lin chuang erh pi yen hou ko tsa chih journal of clinical otorhinolaryngology (wu-han shih"] = "Lin Chuang Erh Pi Yen Hou Ko Tsa Chih",
  ["lin chuang gan dan bing za zhi = chinese journal of clinical hepatology"] = "Lin Chuang Gan Dan Bing Za Zhi",
  ["lin chuang ping t"] = "Hsueh Tsa Chih",
  ["lin chuang yi xue za zhi = journal of clinical medicine"] = "Lin Chuang Yi Xue Za Zhi",
  ["lin ye ke xue (beijing, china : 1979)"] = "Lin Ye Ke Xue (1979)",
  ["lin ye ke xue yan jiu = forest research"] = "Lin Ye Ke Xue Yan Jiu",
  ["lin ye yu huan jing ke xue = forestry and environmental science"] = "Lin Ye Yu Huan Jing Ke Xue",
  ["linchan huaxue yu gongye"] = "Linchan Huaxue Yu Gongye",
  ["lincoln herald"] = "Linc Her",
  ["lincoln law review (san francisco, calif.)"] = "Linc Law Rev",
  ["lincoln review (washington, d.c.)"] = "Lincoln Rev",
  ["lincom textbooks in mathematics"] = "LINCOM Textb. Math.",
  ["lindleyana : the scientific journal of the american orchid society"] = "Lindleyana",
  ["lindos. fouilles et recherches"] = "Lindos",
  ["linea ecologica"] = "Linea Ecol.",
  ["linear algebra and its applications"] = "Linear Algebra Appl.",
  ["linear and multilinear algebra"] = "Linear Multilinear Algebra",
  ["linear and nonlinear analysis. an international journal"] = "Linear Nonlinear Anal.",
  ["linear operators and linear systems"] = "Linear Oper. Linear Syst.",
  ["linear topological spaces and complex analysis"] = "Linear Topol. Spaces Complex Anal.",
  ["linen supply news"] = "Linen Supply News",
  ["lingua e stile"] = "L&S",
  ["lingua nostra"] = "Lingua Nostra",
  ["lingua. international review of general linguistics. revue internationale de linguistique generale"] = "Lingua",
  ["lingua. international review of general linguistics. revue internationale de linguistique générale"] = "Lingua",
  ["linguistic approaches to bilingualism"] = "Linguist Approaches Biling",
  ["linguistic inquiry"] = "Ling. Inq.",
  ["linguistic inquiry (online)"] = "Linguist Inq",
  ["linguistic variation"] = "Linguist Var",
  ["linguistica biblica: interdisziplinäre zeitschrift für theologie und linguistik"] = "LingBibl",
  ["linguistica, epigrafia, filologia italica"] = "LingIt",
  ["linguistics and philosophy"] = "Linguist Philos",
  ["linguistics vanguard : multimodal online journal"] = "Linguist Vanguard",
  ["lingvisticae investigationes: supplementa"] = "Lingvist. Investig. Suppl.",
  ["links (new york, n.y.)"] = "Links",
  ["links : a newsletter on gender for oxfam gb staff and partners"] = "Links (Oxford)",
  ["linköping studies in science and technology"] = "Linköping Stud. Sci. Tech. Diss.",
  ["linnean society symposium series"] = "Linn. Soc. Symp. Ser.",
  ["linneana belgica"] = "Linneana Belg",
  ["linzer biologische beiträge"] = "Linz Biol Beitr",
  ["linzer biologische beiträge"] = "Linz. biol. Beitr.",
  ["lipics : leibniz international proceedings in informatics"] = "Lebniz Int Proc Inform",
  ["lipics. leibniz international proceedings in informatics"] = "LIPIcs. Leibniz Int. Proc. Inform.",
  ["lipid insights"] = "Lipid Insights",
  ["lipid technology"] = "Lipid Technol",
  ["lipids in health and disease"] = "Lipids Health Dis",
  ["lippincott health promotion letter"] = "Lippincott Health Promot. Lett.",
  ["lippincott's case management : managing the process of patient care"] = "Lippincotts Case Manag",
  ["lippincott's primary care practice"] = "Lippincotts Prim Care Pract",
  ["lippincott’s case management : managing the process of patient care"] = "Lippincotts Case Manag",
  ["lippincott’s primary care practice"] = "Lippincotts Prim. Care Pract.",
  ["lippische mitteilungen aus geschichte und landeskunde"] = "Lipp Mitt Gesch Landeskd",
  ["lipsey lectures"] = "Lipsey Lect.",
  ["liquid crystal today"] = "Liq. Cryst. Today",
  ["liquid crystals"] = "Liq. Cryst.",
  ["liquid crystals reviews"] = "Liq. Cryst. Rev.",
  ["liquid waste recovery"] = "Liq. Waste Recovery",
  ["lisboa médica"] = "Lisb Med",
  ["lishi yanjiu"] = "Lishi Yanjiu",
  ["liste d’abreviations de mots des titres de publications en serie. supplement"] = "Liste Abr. Mots Titres Publ. Ser. Suppl.",
  ["listener (london, england : 1967)"] = "Listener",
  ["listy filologické"] = "LF",
  ["liteinoe proizvodstvo"] = "Liteinoe Proizvod.",
  ["literacy & numeracy studies : an international journal in the education and training of adults"] = "Lit Numer Stud",
  ["literacy discussion"] = "Lit Discuss",
  ["literacy information and computer education journal"] = "Lit Inf Comput Educ J",
  ["literacy research : theory, method, and practice"] = "Lit Research",
  ["literacy research and instruction"] = "Lit Res Instr",
  ["literacy, teaching and learning"] = "Lit Teach Learn",
  ["literatura (san sebastian, spain)"] = "Literatura",
  ["literatura (san sebastián, spain)"] = "Literatura",
  ["literature & history"] = "Lit Hist",
  ["literature and medicine"] = "Lit Med",
  ["lithic technology"] = "Lithic Technol.",
  ["lithium (edinburgh, scotland)"] = "Lithium (Edinb.)",
  ["lithology and mineral resources"] = "Lithol. Min. Resour.",
  ["lithuanian academy of sciences"] = "Informatica (Vilnius)",
  ["lithuanian historical studies"] = "Lith Hist Stud",
  ["lithuanian journal of physics"] = "Lith. J. Phys.",
  ["lithuanian mathematical journal"] = "Lith. Math. J.",
  ["litterae numismaticae vindobonenses"] = "LNV",
  ["litteraria pragensia"] = "Litt Prag",
  ["litterature, medecine, societe"] = "Litt Med Soc",
  ["little histories"] = "Little Hist.",
  ["littérature, médecine, société"] = "Litt Med Soc",
  ["littérature, médicine, société"] = "LMS",
  ["liturgy and muse"] = "Liturg Condenda",
  ["liupanshui shi-zhuan xuebao"] = "Liupanshui Shi Zhuan Xuebao",
  ["liver cancer"] = "Liver Cancer",
  ["liver international"] = "Liver Int.",
  ["liver international : official journal of the international association for the study of the liver"] = "Liver Int",
  ["liver research"] = "Liver Res",
  ["liver transplantation"] = "Liver Transpl.",
  ["liver transplantation : official publication of the american association for the study of liver diseases and the international liver transplantation society"] = "Liver Transpl",
  ["liver transplantation and surgery"] = "Liver Transpl. Surg.",
  ["liver transplantation and surgery : official publication of the american association for the study of liver diseases and the international liver transplantation society"] = "Liver Transpl Surg",
  ["liverpool classical monthly"] = "LCM",
  ["liverpool classical monthly : lcm"] = "Liverp Classical Mon",
  ["livestock production science"] = "Livest Prod Sci",
  ["livestock science"] = "Livest Sci",
  ["living journal of computational molecular science"] = "Living J Comput Mol Sci",
  ["living reviews in computational astrophysics"] = "Living Rev Comput Astrophys",
  ["living reviews in landscape research"] = "Living Rev. Landscape Res.",
  ["living reviews in relativity"] = "Living Rev. Relativ.",
  ["living reviews in solar physics"] = "Living Rev. Sol. Phys.",
  ["living to 100 monograph"] = "Living 100 Monogr",
  ["lj special report"] = "LJ Spec Rep",
  ["llafur : journal of welsh labour history = cylchgrawn hanes llafur cymru"] = "Llafur",
  ["llano estacado heritage"] = "Llano Estac Heritage",
  ["lloyds bank review"] = "Lloyds Bank Rev.",
  ["llull : boletin de la sociedad espanola de historia de las ciencias"] = "Llull",
  ["llull : boletín de la sociedad española de historia de las ciencias"] = "Llull",
  ["llull. revista de la sociedad espanola de historia de las ciencias y de las tecnicas"] = "LLULL",
  ["lms journal of computation and mathematics"] = "LMS J. Comput. Math.",
  ["lmt : lab management today"] = "LMT",
  ["lmt; lab management today"] = "LMT",
  ["lo scalpello"] = "Scalpello",
  ["lo sperimentale"] = "Sperimentale",
  ["lobachevskii journal of mathematics"] = "Lobachevskii J. Math.",
  ["lobund reports"] = "Lobund reports",
  ["local and regional anesthesia"] = "Local Reg Anesth",
  ["local economy"] = "Local Econ",
  ["local environment"] = "Local Environ",
  ["local land and soil news"] = "Local land soil news",
  ["local population studies"] = "Local Popul Stud",
  ["locus (denton, tex.)"] = "Locus",
  ["locus (juiz de fora, brazil)"] = "Locus Juiz Fora Braz",
  ["loescher universita manueli"] = "Loescher Universita Manuali",
  ["logic and analysis"] = "Log. Anal.",
  ["logic and argumentation"] = "Log. Argum.",
  ["logic and cognitive systems"] = "Log. Cogn. Syst.",
  ["logic and computation in philosophy"] = "Logic Comput. Philos.",
  ["logic and logical philosophy"] = "Log. Log. Philos.",
  ["logic in asia: studia logica library"] = "Log. Asia Stud. Log. Libr.",
  ["logic journal of the igpl"] = "Log J IGPL",
  ["logic journal of the igpl. interest group in pure and applied logics"] = "Log. J. IGPL",
  ["logic phds"] = "Log. PhDs",
  ["logic, argumentation & reasoning"] = "Log. Argum. Reason",
  ["logic, epistemology, and the unity of science"] = "Log. Epistemol. Unity Sci.",
  ["logica nova"] = "Logica Nova",
  ["logica trianguli"] = "Log. Trianguli",
  ["logica universalis"] = "Log. Univers.",
  ["logical methods in computer science"] = "Log. Methods Comput. Sci.",
  ["logicheskie issledovaniya"] = "Log. Issled.",
  ["logicheskie issledovaniya. logical investigations"] = "Log. Issled.",
  ["logique et analyse"] = "Logique et Anal. (N.S.)",
  ["logique et analyse. nouvelle serie"] = "Log. Anal. (N.S.)",
  ["logistics and transportation review"] = "Logist. Transp. Rev.",
  ["logistics information management"] = "Logist. Inf. Manage.",
  ["logistics research"] = "Logist. Res.",
  ["logistics world"] = "Logist. World",
  ["logopedics, phoniatrics, vocology"] = "Logoped Phoniatr Vocol",
  ["logos (santa clara, calif.)"] = "Logos (Santa Clara)",
  ["logos. studien zur logik, sprachphilosophie und metaphysik"] = "Logos",
  ["loma linda university dentist magazine"] = "Loma Linda Univ Dent Mag",
  ["london and edinburgh monthly journal of medical science"] = "Lond Edinb Mon J Med Sci",
  ["london clinic medical journal"] = "Lond Clin Med J",
  ["london journal of medicine"] = "Lond J Med",
  ["london journal of primary care"] = "London J Prim Care (Abingdon)",
  ["london mathematical society lecture note series"] = "London Math. Soc. Lecture Note Ser.",
  ["london mathematical society monographs"] = "London Math. Soc. Monogr. (N.S.)",
  ["london mathematical society student texts"] = "London Math. Soc. Stud. Texts",
  ["london mathematical society. newsletter"] = "Lond. Math. Soc. Newsl.",
  ["long island forum"] = "Long Isl Forum",
  ["long range planning"] = "Long Range Plann",
  ["long term care"] = "Long Term Care (Don Mills)",
  ["long term care (don mills, ont.)"] = "Long Term Care (Don Mills)",
  ["long term care and health services administration quarterly"] = "Long Term Care Health Serv Adm Q",
  ["long term care quarterly"] = "Long Term Care Q.",
  ["long-term care quarterly"] = "Long Term Care Q",
  ["longevity & healthspan"] = "Longev Healthspan",
  ["longitudinal and life course studies"] = "Longit Life Course Stud",
  ["los alamos national laboratory, preprint archive, astrophysics"] = "Los Alamos Nat. Lab. Prepr. Arch. Astrophys.",
  ["los alamos national laboratory, preprint archive, condensed matter"] = "Los Alamos Nat. Lab. Prepr. Arch. Condens. Matter",
  ["los alamos national laboratory, preprint archive, general relativity and quantum cosmology"] = "Los Alamos Nat. Lab. Prepr. Arch. Gen. Relativ. Quantum Cosmol.",
  ["los alamos national laboratory, preprint archive, high energy physics-experiment"] = "Los Alamos Nat. Lab. Prepr. Arch. High Energy Phys. Exp.",
  ["los alamos national laboratory, preprint archive, high energy physics-lattice"] = "Los Alamos Nat. Lab. Prepr. Arch. High Energy Phys. Lattice",
  ["los alamos national laboratory, preprint archive, high energy physics-phenomenology"] = "Los Alamos Nat. Lab. Prepr. Arch. High Energy Phys. Phenomenol.",
  ["los alamos national laboratory, preprint archive, high energy physics-theory"] = "Los Alamos Nat. Lab. Prepr. Arch. High Energy Phys. Theory",
  ["los alamos national laboratory, preprint archive, nuclear experiment"] = "Los Alamos Nat. Lab. Prepr. Arch. High Energy Phys. Nucl. Exp.",
  ["los alamos national laboratory, preprint archive, nuclear theory"] = "Los Alamos Nat. Lab. Prepr. Arch. Nucl. Theory",
  ["los alamos national laboratory, preprint archive, physics"] = "Los Alamos Nat. Lab. Prepr. Arch. Phys.",
  ["los alamos national laboratory, preprint archive, quantitative biology"] = "Los Alamos Nat. Lab. Prepr. Arch. Quant. Biol.",
  ["los alamos national laboratory, preprint archive, quantum physics"] = "Los Alamos Nat. Lab. Prepr. Arch. Quantum Phys.",
  ["los alamos science"] = "Los Alamos Sci",
  ["los alamos science / los alamos scientific laboratory"] = "Los Alamos Sci",
  ["los angeles lawyer"] = "Los Angel Lawyer",
  ["loss, grief & care"] = "Loss Grief Care",
  ["lotta contro la tubercolosi"] = "Lotta Tuberc",
  ["lotta contro la tubercolosi e le malattie polmonari sociali"] = "Lotta Contro Tuberc Malat Polm Sociali",
  ["louisiana history"] = "La Hist",
  ["louisiana law review"] = "LA Law Rev",
  ["louisiana studies"] = "La Stud",
  ["louvain economic review"] = "Louvain Econ. Rev.",
  ["louvain medical"] = "Louv Med",
  ["louvain médical"] = "Louv Med",
  ["louvain studies"] = "Louv Stud",
  ["lovtidende. a, almindelige love, anordninger og bekendtgoerelser. denmark"] = "LOVTID K DAN A",
  ["low carbon economy"] = "Low Carbon Econ",
  ["low temperature physics"] = "Low Temp. Phys.",
  ["lower urinary tract symptoms"] = "Low Urin Tract Symptoms",
  ["loyola consumer law review"] = "Loyola Consum Law Rev",
  ["loyola law review"] = "Loyola Law Rev",
  ["loyola of los angeles international and comparative law journal"] = "Loyola Los Angel Int Comp Law J",
  ["loyola of los angeles law review"] = "Loyola Los Angel Law Rev",
  ["loyola university of chicago law journal. loyola university chicago. school of law"] = "Loyola Univ Chic Law J",
  ["loyola university of chicago law journal. loyola university of chicago. school of law"] = "Loyola Univ Chicago Law J",
  ["lso working papers in linguistics"] = "LSO Work Pap Linguist",
  ["ltc regulatory risk & liability advisor"] = "LTC Regul Risk Liabil Advis",
  ["ltcc advanced mathematics series"] = "LTCC Adv. Math. Ser.",
  ["lubricants (basel, switzerland)"] = "Lubricants",
  ["lubrication engineering"] = "Lubr. Eng.",
  ["lubrication science"] = "Lubr. Sci.",
  ["lucentum. anales de la universidad de alicante. prehistoria, arqueología e historia antigua"] = "Lucentum",
  ["luft- und raumfahrt"] = "Luft Raumfahrt",
  ["lumiere classique"] = "Lumiere Class.",
  ["luminescence : the journal of biological and chemical luminescence"] = "Luminescence",
  ["lund archaeological review"] = "LundAR",
  ["lundellia : journal of the plant resources center of the university of texas at austin"] = "Lundellia (Austin Tex)",
  ["lundiana : international journal of biodiversity"] = "Lundiana",
  ["lundqua report"] = "Lundqua Rep.",
  ["lung cancer"] = "Lung Cancer",
  ["lung cancer (amsterdam, netherlands)"] = "Lung Cancer",
  ["lung cancer (auckland, n.z.)"] = "Lung Cancer (Auckl)",
  ["lung cancer international"] = "Lung Cancer Int",
  ["lung cancer management"] = "Lung Cancer Manag",
  ["lung health professional magazine"] = "Lung Health Prof Mag",
  ["lung india : official organ of indian chest society"] = "Lung India",
  ["lungs and breathing"] = "Lungs Breath",
  ["luoyang university"] = "J. Luoyang Univ.",
  ["lupus (los angeles)"] = "Lupus (Los Angel)",
  ["lupus science & medicine"] = "Lupus Sci Med",
  ["luso-brazilian review"] = "Luso-Braz Rev",
  ["lusotopie : enjeux contemporains dans les espaces lusophones"] = "Lusotopie",
  ["lustrum. internationale forschungsberichte aus dem bereich des klassischen altertums"] = "Lustrum",
  ["lutheran quarterly (gettysburg, pa. : 1949)"] = "Lutheran Q",
  ["lutheran theological journal"] = "Lutheran Theol J",
  ["lutte contre le cancer"] = "Lutte Cancer",
  ["luzifer-amor : zeitschrift zur geschichte der psychoanalyse"] = "Luzif Amor",
  ["lwt food science and technology"] = "LWT Food Sci. Technol.",
  ["lwt–food science and technology"] = "LWT–Food Sci. Technol.",
  ["lychnos : lärdomshistoriska samfundets årsbok = annual of the swedish history of science society"] = "Lychnos Lardomshist Samf Arsb",
  ["lykia. anadolu-akdeniz kültürleri"] = "Lykia",
  ["lymphatic research and biology"] = "Lymphat Res Biol",
  ["lymphokine and cytokine research"] = "Lymphokine Cytokine Res",
  ["lymphokine research"] = "Lymphokine Res",
  ["lyon chirurgical"] = "Lyon Chir",
  ["lyon medical"] = "Lyon Med.",
  ["lyon mediterranee medical. medecine du sud-est"] = "Lyon Mediterr Med Med Sud Est",
  ["lyon médical"] = "Lyon Med",
  ["lyon méditerranée médical. médecine du sud-est"] = "Lyon Mediterr Med Med Sud Est",
  ["lyon pharmaceutique"] = "Lyon Pharm",
  ["löbf-mitteilungen, landesanstalt für ökologie, bodenordnung und forsten / landesamt für agrarordnung, nordrhein-westfalen"] = "LÖBF-Mitt., Landesanst. Ökol. Bodenordn. Forsten, Landesamt Agrarordn. Nordrh.-Westfal.",
  ["löbf-schriftenreihe, landesanstalt für ökologie, bodenordnung und forsten, nordrhein-westfalen"] = "LÖBF-Schr.reihe, Landesanst. Ökol. Bodenordn. Forsten Nordrh.-Westfal.",
  ["lölf-mitteilungen, landesanstalt für ökologie, landschaftsentwicklung und forstplanung, nordrhein-westfalen"] = "LÖLF-Mitt., Landesanst. Ökol. Landsch.entwickl.  Forstplan., Nordrh.-Westfal.",
  ["l’ alimentation et la vie"] = "Aliment Vie",
  ["l’ annee biologique"] = "Annee Biol",
  ["l’ annee psychologique"] = "Annee Psychol",
  ["l’ annee sociologique"] = "Annee Sociol",
  ["l’ annee therapeutique"] = "Annee Ther",
  ["l’ arcispedale s. anna di ferrara"] = "Arcisp S Anna Ferrara",
  ["l’ ateneo parmense. sezione i: acta bio-medica"] = "Ateneo Parmense [1]",
  ["l’ ethnographie"] = "Ethnographie.",
  ["l’ eurobiologiste"] = "Eurobiologiste",
  ["l’ evolution medicale"] = "Evol Med",
  ["l’ evolution psychiatrique"] = "Evol Psychiatr (Paris)",
  ["l’ homeopathie francaise"] = "Homeopath Fr",
  ["l’ homme; revue francaise d’anthropologie"] = "Homme",
  ["l’ hopital"] = "Hopital",
  ["l’ infirmiere"] = "Infirmiere",
  ["l’ infirmiere canadienne"] = "Infirm Can",
  ["l’ infirmiere haitienne"] = "Infirm Haiti",
  ["l’ information dentaire"] = "Inf Dent",
  ["l’ information medicale et paramedicale"] = "Inf Med Paramed (Montreal)",
  ["l’ orthodontie francaise"] = "Orthod Fr",
  ["l’ ospedale maggiore"] = "Osp Maggiore",
  ["l’academie en poche"] = "Acad. Poche",
  ["l’action nationale"] = "Action Natl",
  ["l’actualite chimique"] = "Actual Chim",
  ["l’actualite economique"] = "Actual Econ",
  ["l’actualité economique"] = "L’Actual. Econ.",
  ["l’aeronautique et l’astronautique : organe commun de l’association francaise des ingenieurs et techniciens de l’aeronautique de l’espace (a.f.i.t.a.e.) et de la societe francaise d’astronautique (s.f.a.)"] = "Aeronaut Astronaut",
  ["l’afrique et l’asie modernes"] = "Afr Asie Mod",
  ["l’afrique litteraire et artistique"] = "Afr Litt Artist",
  ["l’algebra e le sue applicazioni tra classico e moderno"] = "Algebra Appl. Class. Mod.",
  ["l’annee endocrinologique"] = "Annee Endocrinol",
  ["l’annee sociale / institut de sociologie, universite libre de bruxelles"] = "Annee Soc",
  ["l’annee therapeutique et clinique en ophtalmologie"] = "Annee Ther Clin Ophtalmol",
  ["l’année philologique"] = "APh",
  ["l’année épigraphique"] = "AnnEpigr",
  ["l’antiquite classique"] = "Antiq Class",
  ["l’antiquité classique"] = "AC",
  ["l’archeologo subacqueo. quadrimestrale di archeologia subacquea e navale"] = "ASub",
  ["l’archiginnasio : bullettino della biblioteca comunale di bologna"] = "Archiginnasio",
  ["l’ateneo parmense. acta bio-medica : organo della societa di medicina e scienze naturali di parma"] = "Ateneo Parmense Acta Biomed",
  ["l’ateneo parmense. acta naturalia : organo della societa di medicina e scienze naturali di parma"] = "Ateneo Parmense Acta Nat",
  ["l’attualita dietetica"] = "Attual Diet",
  ["l’avenir medical"] = "Avenir Med",
  ["l’echo medical des cevennes"] = "Echo Med Cevennes",
  ["l’echo medical du nord"] = "Echo Med Nord",
  ["l’economia umana; rassegna medica internazionale"] = "Economia Umana Rassegna Medica Internazionale",
  ["l’economie de la reunion"] = "Econ Reun",
  ["l’egypte contemporaine"] = "Egypte Contemp",
  ["l’enfant en milieu tropical"] = "Enfant Milieu Trop",
  ["l’enseignement mathematique"] = "Enseign. Math.",
  ["l’enseignement mathématique"] = "Enseign. Math. (2)",
  ["l’espace geographique"] = "Espace Geogr",
  ["l’evolution de l’humanite, synthese collective. premiere section, serie complementaire: la science dans l’antiquite"] = "L’Evolution de l’Humanite, Synthese Collective. Premiere Section, Serie Complementaire: La Science dans l’An",
  ["l’hygiene mentale"] = "Hyg Ment",
  ["l’igiene moderna"] = "Ig Mod",
  ["l’industria, nuova serie"] = "L’Industria",
  ["l’infirmiere auxiliaire : revue de la corporation professionnelle des infirmieres et infirmiers auxiliaires du quebec"] = "Infirm Aux",
  ["l’infirmiere du quebec : revue officielle de l’ordre des infirmieres et infirmiers du quebec"] = "Infirm Que",
  ["l’infirmiere francaise"] = "Infirm Fr",
  ["l’information d’histoire de l’art"] = "Inf Hist Art",
  ["l’information geographique"] = "Inf Geogr",
  ["l’information grammaticale"] = "IG",
  ["l’information historique"] = "IH",
  ["l’information litteraire"] = "Inf Litt",
  ["l’information littéraire"] = "IL",
  ["l’information psychiatrique"] = "Inf Psychiatr",
  ["l’informatore medico"] = "Inf Med",
  ["l’ospedale psichiatrico"] = "Osp Psichiatr",
  ["l’oto-rino-laringologia italiana"] = "Otorinolaringol Ital",
  ["l’union medicale du canada"] = "Union Med Can",
  ["l’urbe. rivista romana"] = "LUrbe",
  ["l’âne d’or"] = "Âne d’Or",
  ["l’écrit du temps"] = "EDT",
  ["m & b pharmaceutical bulletin"] = "M B Pharm Bull",
  ["m&t series"] = "M&T Ser.",
  ["m. crawford, roman republican coinage (london 1974)"] = "RRC",
  ["m. s. c. veterinarian"] = "M S C Vet",
  ["m. thompson – c. m. kraay – o. mørkholm, an inventory of greek coin hoards (new york 1973)"] = "IGCH",
  ["m.d. computing : computers in medical practice"] = "MD Comput",
  ["ma zui xue za zhi = anaesthesiologica sinica"] = "Ma Zui Xue Za Zhi",
  ["ma zui yu jian hu lun tan = forum of anesthesia and monitoring"] = "Ma Zui Yu Jian Hu Lun Tan",
  ["maa notes"] = "MAA Notes",
  ["maa problem books series"] = "MAA Probl. Books Ser.",
  ["maa spectrum"] = "MAA Spectrum",
  ["maa textbooks"] = "MAA Textb.",
  ["maandschrift - centraal bureau voor de statistiek. netherlands. centraal bureau voor de statistiek"] = "Maandschr Cent Bur Stat",
  ["maandschrift voor kindergeneeskunde"] = "Maandschr Kindergeneeskd",
  ["maandstatistiek van bevolking en volksgezondheid. netherlands. centraal bureau voor de statistiek"] = "Maandstat Bevolking Volksgezond",
  ["maandstatistiek van de bevolking (hague, netherlands : 1982)"] = "Maandstat Bevolking",
  ["maanedsskrift for praktisk laegegerning"] = "Manedsskr. Prakt. Laegegern.",
  ["maanedsskrift for praktisk laegegerning og social medicin"] = "Manedsskr Prakt Laegegern",
  ["maanedsskrift for praktisk lægegerning og social medicin"] = "Manedsskr Prakt Laegegern",
  ["maas journal of islamic science"] = "MAAS J Islam Sci",
  ["maastricht journal of european and comparative law"] = "Maastrich J Eur Comp Law",
  ["macedoniae acta archaeologica"] = "MacActaA",
  ["macedonian journal of chemistry and chemical engineering"] = "Maced. J. Chem. Chem. Eng.",
  ["macedonian veterinary review"] = "Maced Vet Rev",
  ["machine intelligence and pattern recognition"] = "Mach. Intelligence Pattern Recogn.",
  ["machine intelligence research"] = "Mach. Intell. Res.",
  ["machine learning"] = "Mach. Learn.",
  ["machine learning and interpretation in neuroimaging : 4th international workshop, mlini 2014, held at nips 2014, montreal qc, canada, december 13, 2014 : revised selected papers. mlini (workshop) (4th : 2014 : montréal, québec)"] = "Mach Learn Interpret Neuroimaging (2014)",
  ["machine learning and interpretation in neuroimaging : international workshop, mlini 2011, held at nips 2011, sierra nevada, spain, december 16-17, 2011 : revised selected and invited contributions. mlini (workshop) (2011 : sierra nevada, spain)"] = "Mach Learn Interpret Neuroimaging (2011)",
  ["machine learning and knowledge extraction"] = "Mach Learn Knowl Extr",
  ["machine learning and knowledge extraction : third ifip tc 5, tc 12, wg 8.4, wg 8.9, wg 12.9 international cross-domain conference, cd-make 2019, canterbury, uk, august 26-29, 2019 : proceedings. international cross-domain conference for machine learning and knowledge extraction (3rd : 2019 : canterbury, england)"] = "Mach Learn Knowl Extr (2019)",
  ["machine learning and medical engineering for cardiovascular health and intravascular imaging and computer assisted stenting : first international workshop, mlmech 2019, and 8th joint international workshop, cvii-stent 2019, held in conjunction with miccai 2019, shenzhen, china, october 13, 2019, proceedings. mlmech (workshop) (1st : 2019 : shenzhen shi, china)"] = "Mach Learn Med Eng Cardiovasc Health Intravasc Imaging Comput Assist Stenting (2019)",
  ["machine learning in medical imaging. mlmi (workshop)"] = "Mach Learn Med Imaging",
  ["machine learning with applications"] = "Mach. Learn. Appl.",
  ["machine learning: earth"] = "Mach. Learn.: Earth",
  ["machine learning: engineering"] = "Mach. Learn.: Eng.",
  ["machine learning: foundations, methodologies, and applications"] = "Mach. Learn. Found. Methodol. Appl.",
  ["machine learning: health"] = "Mach. Learn.: Health",
  ["machine learning: science and technology"] = "Mach. Learn.: Sci. Technol.",
  ["machine translation"] = "Mach. Transl.",
  ["machine translation : mt"] = "Mach Transl",
  ["machine vision and applications"] = "Mach Vis Appl",
  ["machining science and technology"] = "Mach. Sci. Technol.",
  ["macmillan new electronics series"] = "Macmillan New Electron. Ser.",
  ["macroeconomic dynamics"] = "Macroecon Dyn",
  ["macromolecular anticancer therapeutics"] = "Cancer. Drug. Discov. D.",
  ["macromolecular bioscience"] = "Macromol Biosci",
  ["macromolecular chemistry and physics"] = "Macromol. Chem. Phys.",
  ["macromolecular chemistry and physics supplement"] = "Macromol. Chem. Phys. Suppl.",
  ["macromolecular materials and engineering"] = "Macromol. Mater. Eng.",
  ["macromolecular rapid communications"] = "Macromol Rapid Commun",
  ["macromolecular reaction engineering"] = "Macromol. React. Eng.",
  ["macromolecular research"] = "Macromol. Res.",
  ["macromolecular symposia"] = "Macromol. Symp.",
  ["macromolecular symposium"] = "Macromol. Symp.",
  ["macromolecular theory and simulations"] = "Macromol. Theory Simul.",
  ["macromolecules (washington, dc, united states)"] = "Macromolecules (Washington, DC, U. S.)",
  ["madagascar conservation and development"] = "Madag Conserv Dev",
  ["madima'16 : proceedings of the 2nd international workshop on multimedia assisted dietary management : october 16, 2016, amsterdam, the netherlands. international workshop on multimedia assisted dietary management (2nd : 2016 : amsterdam, netherlands)"] = "MADiMa16 (2016)",
  ["madjalah kedokteran indonesia"] = "Madjalah Kedokt Indones",
  ["madjalah pdgi"] = "Madj Persat Dokt Gigi Indones",
  ["madjalah persatuan dokter gigi indonesia (journal of the indonesian dental association)"] = "Madj. Persat. Dokt. Gigi Indones.",
  ["madjalah persatuan dokter gigi indonesia. journal of the indonesian dental association"] = "Madj Persat Dokt Gigi Indones",
  ["madrider beiträge"] = "MB",
  ["madrider forschungen"] = "MF",
  ["madrider mitteilungen"] = "MM",
  ["madridge journal of cancer study & research"] = "Madridge J Cancer Study Res",
  ["madridge journal of dentistry and oral surgery"] = "Madridge J Dent Oral Surg",
  ["madridge journal of diabetes"] = "Madridge J Diabetes",
  ["madridge journal of pharmaceutical research"] = "Madridge J Pharm Res",
  ["madroño; a west american journal of botany"] = "Madrono",
  ["maecenas. studi sul mondo classico"] = "Maecenas",
  ["maejo international journal of science and technology"] = "Maejo Int. J. Sci. Technol.",
  ["magazin istoric"] = "Mag Istor",
  ["magazine antiques (new york, n.y. : 1971)"] = "Mag Antiq",
  ["magazine litteraire"] = "Mag Litt",
  ["magazine littéraire"] = "Mag Litt",
  ["magazine of concrete research"] = "Mag. Concr. Res.",
  ["magazine. cambridge university medical society"] = "Mag Camb Univ Med Soc",
  ["magazine. royal free hospital (london, england). school of medicine"] = "Mag R Free Hosp (Lond Engl)",
  ["maghreb mathematical review"] = "Maghreb Math. Rev.",
  ["maghreb médical = al-majallah al-ṭibbīyah lil-maghrib al-ʻarabī"] = "Maghreb Med",
  ["maghreb, machrek"] = "Maghreb Machrek",
  ["magic, ritual, and witchcraft"] = "Magic Ritual Witch",
  ["magma (new york, n.y.)"] = "MAGMA",
  ["magnesium and trace elements"] = "Magnes Trace Elem",
  ["magnesium research"] = "Magnes Res",
  ["magnesium research : official organ of the international society for the development of research on magnesium"] = "Magnes Res",
  ["magnetic resonance"] = "Magn. Reson.",
  ["magnetic resonance (göttingen, germany)"] = "Magn Reson (Gott)",
  ["magnetic resonance annual"] = "Magn Reson Annu",
  ["magnetic resonance imaging"] = "Magn Reson Imaging",
  ["magnetic resonance imaging clinics of north america"] = "Magn Reson Imaging Clin N Am",
  ["magnetic resonance in chemistry"] = "Magn. Reson. Chem.",
  ["magnetic resonance in chemistry : mrc"] = "Magn Reson Chem",
  ["magnetic resonance in medical sciences"] = "Magn. Reson. Med. Sci.",
  ["magnetic resonance in medical sciences : mrms : an official journal of japan society of magnetic resonance in medicine"] = "Magn Reson Med Sci",
  ["magnetic resonance in medicine"] = "Magn Reson Med",
  ["magnetic resonance in medicine : official journal of the society of magnetic resonance in medicine / society of magnetic resonance in medicine"] = "Magn Reson Med",
  ["magnetic resonance insights"] = "Magn Reson Insights",
  ["magnetic resonance letters"] = "Magn. Reson. Lett.",
  ["magnetic resonance materials in physics, biology and medicine"] = "Magn. Reson. Mater. Phys., Biol. Med.",
  ["magnetic resonance quarterly"] = "Magn Reson Q",
  ["magnetochemistry (basel, switzerland)"] = "Magnetochemistry",
  ["magyar állatorvosok lapja"] = "Magy Allatorvosok Lapja",
  ["magyar belorvosi archivum"] = "Magy Belorv Arch",
  ["magyar belorvosi archivum (1955)"] = "Magy Belorv Arch (1955)",
  ["magyar belorvosi archivum és ideggyógyászati szemle"] = "Magy Belorv Arch Ideggyogy Szle",
  ["magyar kemiai folyoirat"] = "Magy. Kem. Foly.",
  ["magyar kemiai folyoirat, kemiai kozlemenyek"] = "Magy. Kem. Foly., Kem. Kozl.",
  ["magyar könyvszemle"] = "Magyar Konyvsz",
  ["magyar noorvosok lapja"] = "Magy Noorv Lapja",
  ["magyar nöorvosok lapja"] = "Magy Noorv Lapja",
  ["magyar onkologia"] = "Magy Onkol",
  ["magyar radiologia"] = "Magy Radiol",
  ["magyar sebeszet"] = "Magy Seb",
  ["magyar sebészet"] = "Magy Seb",
  ["magyar traumatologia, orthopaedia es helyreallito sebeszet"] = "Magy. Traumatol. Orthop. Helyreallito Seb.",
  ["magyar traumatologia, ortopedia, kezsebeszet, plasztikai sebeszet"] = "Magy. Traumatol. Ortop. Kezseb. Plasztikai Seb.",
  ["magyar traumatológia, orthopaedia és helyreállító sebészet"] = "Magy Traumatol Orthop Helyreallito Seb",
  ["magyar traumatológia, ortopédia, kézsebészet, plasztikai sebészet"] = "Magy Traumatol Ortop Kezseb Plasztikai Seb",
  ["magyar tudomany : [a magyar tudomanyos akademia ertesitoje]"] = "Magy Tud",
  ["magyar tudomány : [a magyar tudományos akadémia értesítője]"] = "Magy Tud",
  ["maǧallaẗ al-baṣraẗ li-abḥāt̲ nahlaẗ al-tamr"] = "Magallat Albasrat Liabhat Nahlat Altamr",
  ["maǧallaẗ al-baṣraẗ li-l-abḥāt̲ al-bayṭariyyat̲"] = "Magallat Albasrat Lilabhat Albaytariyyat",
  ["maharashtra medical journal"] = "Maharashtra Med J",
  ["mahāpañjīkāra kī patrikā = registrar general's news letter. india. office of the registrar general"] = "Mahapanjikara Kipatrika",
  ["mahidol population gazette"] = "Mahidol Popul Q Gaz",
  ["mahidol population gazette / institute for population and social research, mahidol university"] = "Mahidol Popul Q Gaz",
  ["mai lei zuo wu xue bao"] = "Mai Lei Zuo Wu Xue Bao",
  ["maia. rivista di letterature classiche"] = "Maia",
  ["maia: rivista di letterature classiche"] = "Maia",
  ["maikotokishin : maikotokishin kenkyukai kaiho"] = "Proceedings of the Japanese Association of Mycotoxicology|Maikotokishin",
  ["maikotokishin : maikotokishin kenkyūkai kaihō = proceedings of the japanese association of mycotoxicology"] = "Maikotokishin",
  ["main group chemistry"] = "Main Group Chem.",
  ["main group metal chemistry"] = "Main Group Met. Chem.",
  ["mainake. estudios de arqueología malagueña"] = "Mainake",
  ["maine historical society quarterly"] = "Maine Hist Soc Q",
  ["maine law review"] = "Maine Law Rev",
  ["maine nurse"] = "Maine Nurse",
  ["maine revised statutes annotated, 1964 : being the tenth revision of the revised statutes of the state of maine, 1964 / prepared under the supervision of the committee on revision of statutes. maine"] = "Maine Revis Statut Annot 1964 Maine",
  ["maine revised statutes annotated, 1964 : being the tenth revision of the revised statutes of the state of maine, 1964. maine"] = "Maine Revis Statut Annot 1964 Maine",
  ["mainlines : the newsletter for the midwest alliance in nursing"] = "Mainlines",
  ["mains'l haul"] = "Mainsl Haul",
  ["mainzer archäologische zeitschrift"] = "MAZ",
  ["mainzer zeitschrift"] = "MainzZ",
  ["mainzer zeitschrift: mittelrheinisches jahrbuch für archäologie, kunst und geschichte"] = "MZ",
  ["maison poincare. regards mathematiques"] = "Maison Poincare. Regards Math.",
  ["maize genetics cooperation news letter"] = "Maize Genet Coop News Lett",
  ["majalah demografi indonesia"] = "Majalah Demografi Indones",
  ["majalah iptek institut teknologi sepuluh nopember 1945 surabaya"] = "Maj IPTEK Inst Teknol  Sepuluh Nop 1945 Surabaya",
  ["majalah kedokteran indonesia"] = "Majalah Kedokt Indones",
  ["majalah obstetri dan ginekologi indonesia"] = "Majalah Obstet Ginekol Indones",
  ["majallah-i dānish-i giyāh/pizishkī-i īrān"] = "Majallahi Danishi Giyah Pizishkii Iran",
  ["majallah-i dānishgāh-i ʻulūm-i pizishkī-i māzandarān"] = "Majallahi Danishgahi Ulumi Pizishkii Mazandaran",
  ["majallah-i dānishkadah-i dāmpizishkī"] = "Maj Danishkadahi Dampizishki",
  ["majallah-i hifāzat-i giyāhān"] = "Majallahi Hifazati Giyahan",
  ["majallah-i jāmiah-i dandānpizishkī-i īrān = the journal of the iranian dental association"] = "Majallah Jamia Dandan Pazshki",
  ["majallah-i mīkrub/shināsī-i pizishkī-i īrān"] = "Iran J Med Microbiol",
  ["majallah-i pizishkī-i hastahī̄ īrān"] = "Iran J Nucl Med",
  ["majallah-i ulūm va funūn-i daryāyī"] = "Majallahi Ulum Fununi Daryayi",
  ["majallah-i zīst/shināsī-i īrān"] = "Majallahi Zist Shinasii Iran",
  ["majallah-i ʻilmī pizhūhishī-i dānishgāh-i ʻulūm-i pizishkī va khadamāt-i bihdāshtī-i darmānī-i zanjān"] = "Majallahi Ilmi Pizhuhishii Danishgahi Ulumi Pizishki Khadamati Bihdashtii Darmanii Zanjan",
  ["majallah-i'ulūm-i pāyah-i pizishkī-i īrān = iranian journal of basic medical sciences"] = "Majallah Iulum I Payah I Pizishki I Iran",
  ["majallah-iʻ ilmī-pizhūhishī-i dānishgāh-iʻlūm-i pizishkī va khadamāt-i bihdāshtī darmānī-i arāk"] = "Majallahi Ilmipizhuhishii Danishgahilumi Pizishki Va Khadamati Bihdashti Darmanii Arak",
  ["majallah-i’ulum-i payah-i pizishki-i iran"] = "Iranian journal of basic medical sciences / Mashad University of Medical Sciences|Majallah iulum i Payah i Pizishki i Iran",
  ["majallat al-buhuth wa-al-dirasat al-’arabiyah"] = "Majallat Albuhuth Waaldirasat Alarabiyah",
  ["majallat al-buḥūth wa-al-dirāsāt al-ʻarabīyah"] = "Majallat Albuhuth Waaldirasat Alarabiyah",
  ["majallat al-diwan al-qawmi lil-usrah wa-al-’umran al-bashari"] = "Majallat Aldiwan Alqawmi Lilusrah Waal Umran lbashari",
  ["majallat al-sharq al-awsaṭ"] = "Majallat Alsharq Alawasat",
  ["majallat ma had al-makhtutat al-arabiyah"] = "Majallat Ma had al Makhtutat al Arabiy",
  ["majallat ma had al-makhtūtāt al-arabiyah"] = "Majallat Ma had al Makhtutat al Arabiy",
  ["majallat majmaʻ al-lughah al-ʻarabīyah bi-dimashq"] = "Rev Acad Arabe Damas",
  ["majallat majma’ al-lughah al-’arabiyah bi-dimashq"] = "Rev Acad Arabe Damas",
  ["majallat niqabat attiba’ al-asnan al-urduniyah"] = "Majallat Niqabat Attiba Alasnnan Alurduniyah",
  ["majallat niqābat attibā' al-asnān al-urdunīyah"] = "Majallat Niqabat Attiba Alasnnan Alurduniyah",
  ["majallat tibb al-asnan al-suriyah"] = "Majallat Tibb Alasnan Alsuriyah",
  ["majallat tibb al-asnān al-sūrīyah"] = "Majallat Tibb Alasnan Alsuriyah",
  ["majallat tibb al-famm al-suriyah"] = "Majallat Tibb Alfamm Alsuriyah",
  ["majallat tibb al-famm al-sūrīyah"] = "Majallat Tibb Alfamm Alsuriyah",
  ["majallāh-i dānishkadah-i pizishkī-i iṣfahān"] = "Majallahi Danishkadahi Pizishkii Isfahan",
  ["majāllat al-dīwān al-qawmī lil-usrah wa-al-ʻumrān al-basharī"] = "Majallat Aldiwan Alqawmi Lilusrah Waal Umran lbashari",
  ["major american universities ph"] = "Major Amer. Univ. Ph.D. Qualif. Questions Solut.",
  ["major problems in clinical pediatrics"] = "Major Probl Clin Pediatr",
  ["major problems in clinical surgery"] = "Major Probl Clin Surg",
  ["major problems in internal medicine"] = "Major Probl Intern Med",
  ["major problems in obstetrics and gynecology"] = "Major Probl Obstet Gynecol",
  ["major problems in pathology"] = "Major Probl Pathol",
  ["makara journal of science"] = "Makara J Sci",
  ["makedonska akademija na naukite i umetnostite"] = "Makedon. Akad. Nauk. Umet. Oddel. Mat.-Tehn. Nauk. Prilozi",
  ["makedonski medicinski pregled. revue medicale macedonienne"] = "Makedon Med Pregl",
  ["makedonski medicinski pregled. revue médicale macedonienne"] = "Makedon Med Pregl",
  ["makedonsko nasledstvo. spisanie za arheologija, istorija, istorija na umetnosta i etnologija"] = "MakedNasl",
  ["making the rounds in health, faith, & ethics"] = "Mak Rounds Health Faith Ethics",
  ["makinoa new series : bulletin of the makino botanical garden"] = "Makinoa New Ser",
  ["makromolekulare chemie"] = "Makromol. Chem.",
  ["makromolekulare chemie, macromolecular symposia"] = "Makromol. Chem. Macromol. Symp.",
  ["makromolekulare chemie, rapid communications"] = "Makromol. Chem. Rapid Commun.",
  ["makromolekulare chemie, supplement"] = "Makromol. Chem. Suppl.",
  ["makromolekulare chemie, theory and simulations"] = "Makromol. Chem. Theory Simul.",
  ["mala jazykova kniznice"] = "Mala Jazyk. Kn.",
  ["malacologica bohemoslovaca : journal about molluscs in central europe"] = "Malacol Bohemoslov",
  ["malagasy nature"] = "Malagasy Nat",
  ["malaria chemotherapy, control & elimination"] = "Malar Chemother Control Elimin",
  ["malaria control & elimination"] = "Malar Control Elimin",
  ["malaria journal"] = "Malar J",
  ["malaria research and treatment"] = "Malar Res Treat",
  ["malaria weekly"] = "Malar Wkly",
  ["malariaworld journal"] = "Malariaworld J",
  ["malattie cardiovascolari"] = "Mal Cardiovasc",
  ["malawi medical journal : the journal of medical association of malawi"] = "Malawi Med J",
  ["malaya journal of biosciences"] = "Malaya J Biosci",
  ["malaya journal of matematik"] = "Malaya J. Mat.",
  ["malayan economic review"] = "Malayan Econ. Rev.",
  ["malayan journal of tropical geography"] = "Malayan J. Trop. Geogr.",
  ["malayan nature journal, the"] = "Malayan Nat. J.",
  ["malaysian applied biology = biologi gunaan malaysia"] = "Malays Appl Biol",
  ["malaysian dental journal"] = "Malays Dent J",
  ["malaysian family physician : the official journal of the academy of family physicians of malaysia"] = "Malays Fam Physician",
  ["malaysian forester, the"] = "Malays. For.",
  ["malaysian journal of biochemistry and molecular biology"] = "Malays J Biochem Mol Biol",
  ["malaysian journal of economic studies : journal of the malaysian economic association and the faculty of economics and administration, university of malaya"] = "Malays J Econ Studies",
  ["malaysian journal of mathematical sciences"] = "Malays. J. Math. Sci.",
  ["malaysian journal of microbiology"] = "Malays J Microbiol",
  ["malaysian journal of nutrition"] = "Malays J Nutr",
  ["malaysian journal of pathology"] = "Malays. J. Pathol.",
  ["malaysian journal of reproductive health : a publication of the reproductive research centre of the national population and family development board, malaysia"] = "Malays J Reprod Health",
  ["malaysian journal of tropical geography"] = "Malays J Trop Geogr",
  ["malaysian journal of veterinary research"] = "Malays J Vet Res (Putrajaya)",
  ["malaysian mathematical society"] = "Bull. Malaysian Math. Soc. (2)",
  ["malaysian orthopaedic journal"] = "Malays Orthop J",
  ["male monografie pwn. technika"] = "Male Monogr. PWN Tech.",
  ["male nurses journal"] = "Male Nurses J.",
  ["malpractice digest"] = "Malpract Dig",
  ["mammal research"] = "Mamm Res",
  ["mammal review"] = "Mamm Rev",
  ["mammal study"] = "Mammal Study",
  ["mammalian biology"] = "Mamm. Biol.",
  ["mammalian biology - zeitschrift fur saugetierkunde"] = "Mamm. Biol. Z. Saugetierkd.",
  ["mammalian biology = zeitschrift für säugetierkunde"] = "Mamm Biol",
  ["mammalian genome"] = "Mamm. Genome",
  ["mammalian genome : official journal of the international mammalian genome society"] = "Mamm Genome",
  ["mammalian species"] = "Mamm. Species",
  ["man and medicine"] = "Man Med",
  ["man in india"] = "Man India",
  ["man-made textiles in india"] = "Man-Made Text. India",
  ["man; a monthly record of anthropological science"] = "Man (Lond)",
  ["managed care"] = "Manag. Care",
  ["managed care (langhorne, pa.)"] = "Manag Care",
  ["managed care interface"] = "Manag Care Interface",
  ["managed care quarterly"] = "Manag Care Q",
  ["managed care strategies"] = "Manag. Care Strateg.",
  ["managed care strategies (atlanta, ga.)"] = "Manag Care Strateg",
  ["managed medicare & medicaid news"] = "Manag Medicare Medicaid News",
  ["managed medicare and medicaid news"] = "Manag. Medicare Medicaid News",
  ["management & sustainability: an arab review"] = "Manage. Sustainability: Arab Rev.",
  ["management accounting"] = "Manage. Acc.",
  ["management and compliance series"] = "Manag Compliance Ser",
  ["management and compliance series / american society for hospital engineering of the american hospital association"] = "Manag Compliance Ser",
  ["management and industrial engineering"] = "Manag. Ind. Eng",
  ["management and information systems"] = "Management and Inform. Systems",
  ["management and organization review : the journal of the international association for chinese management research"] = "Manag Organ Rev",
  ["management communication quarterly"] = "Manag Commun Q",
  ["management focus"] = "Manage Focus",
  ["management for professionals"] = "Manag. Prof.",
  ["management international review : mir : journal of international business"] = "Manag Int Rev",
  ["management learning"] = "Manag Learn",
  ["management of biological invasions"] = "Manage. Biol. Invasions",
  ["management of biological invasions : international journal of applied research on biological invasions"] = "Manag Biol Invasion",
  ["management of environmental quality"] = "Manage. Environ. Qual.",
  ["management review"] = "Manage Rev",
  ["management science"] = "Manage Sci",
  ["management training review"] = "Manage. Train. Rev.",
  ["management world"] = "Manage World",
  ["manager (newton, mass.)"] = "Manag",
  ["managerial and decision economics"] = "Managerial Dec. Econ.",
  ["managerial and decision economics : mde"] = "MDE Manage Decis Econ",
  ["managing international development : mid"] = "Manag Int Dev",
  ["manchester medical gazette"] = "Manch. Med. Gaz.",
  ["manchester memoirs"] = "Manch Men",
  ["manchester memoirs / manchester literary and philosophical society"] = "Manch Men",
  ["manchester school"] = "Manchester Sch.",
  ["manchester school of economic and social studies"] = "Manchester Sch. Econ. Soc. Stud.",
  ["manchester university medical school gazette"] = "Manch Univ Med Sch Gaz",
  ["manglar (tumbes)"] = "Manglar (Tumbes)",
  ["mangroves and salt marshes"] = "Mangroves Salt Marshes",
  ["manitoba history"] = "Man Hist",
  ["manitoba law journal"] = "Manit Law J",
  ["manitoba medical review"] = "Manit Med Rev",
  ["manshurat majallat dirasat al-khalij wa-al-jazirah al-’arabiyah / jami’at al-kuwayt"] = "Manshur Atmajallat Dirasat Alkhalij Waal Jazirah Alarabiyah",
  ["manshūrāt majallat dirāsāt al-khalīj wa-al-jazīrah al-ʻarabīyah"] = "Manshur Atmajallat Dirasat Alkhalij Waal Jazirah Alarabiyah",
  ["mansoura journal of chemistry"] = "Mansoura J. Chem.",
  ["mantap : majalah ilmiah pkmi : journal of the indonesian association for secure contraception"] = "Mantap",
  ["manter (lincoln, neb.)"] = "Manter (Linc)",
  ["manual therapy"] = "Man Ther",
  ["manufacturing & service operations management : m & som"] = "Manuf Serv Oper Manag",
  ["manufacturing chemist"] = "Manuf. Chem.",
  ["manufacturing chemist and aerosol news"] = "Manuf Chem Aerosol News",
  ["manufacturing engineering and materials processing"] = "Manuf. Eng. Mater. Process.",
  ["manufacturing letters"] = "Manuf Lett",
  ["manufacturing review"] = "Manuf. Rev.",
  ["manuscripta mathematica"] = "Manuscripta Math.",
  ["manuscripts (new york, n.y.)"] = "Manuscripts (N Y)",
  ["maple technical newsletter, the"] = "Maple Tech. Newsl.",
  ["marburger beiträge zur antiken handels-, wirtschafts- und sozialgeschichte, vor jahrgang 2009 münstersche beiträge zur antiken handelsgeschichte"] = "MBAH",
  ["marburger geographische schriften"] = "Marbg Geogr Schr",
  ["marburger schriften zur medizingeschichte"] = "Marbg Schrift Medgesch",
  ["marburger winckelmann-programm"] = "MarbWPr",
  ["margin - the journal of applied economic research"] = "Margin J Appl Econ Res",
  ["marginalia dermatologica"] = "Marginalia Dermatol",
  ["marginalia otolaryngologica"] = "Marg Otolaryngol",
  ["margins (baltimore, md.)"] = "Margins (Baltim)",
  ["mari. annales de recherches interdisciplinaires"] = "Mari",
  ["marine & freshwater research"] = "Mar Freshw Res",
  ["marine and coastal fisheries"] = "Mar. Coastal Fish.",
  ["marine and coastal fisheries : dynamics, management , and ecosystem science"] = "Mar Coast Fish",
  ["marine and freshwater behaviour and physiology"] = "Mar Freshw Behav Physiol",
  ["marine and freshwater harmful algae : proceedings of the 16th international conference on harmful algae, 27th-31st october 2014, wellington, new zealand. international conference on harmful algae (16th : 2014 : wellington, n.z.), author"] = "Mar Freshw Harmful Algae (2014)",
  ["marine and freshwater research"] = "Mar. Freshwater Res.",
  ["marine and petroleum geology"] = "Mar. Pet. Geol.",
  ["marine behaviour and physiology"] = "Mar Behav Physiol",
  ["marine biodiversity"] = "Mar. Biodivers.",
  ["marine biodiversity : a journal of the senckenberg research institute"] = "Mar Biodivers",
  ["marine biodiversity records"] = "Mar Biodivers Rec",
  ["marine biology"] = "Mar. Biol.",
  ["marine biology research"] = "Mar. Biol. Res.",
  ["marine biology research (online)"] = "Mar Biol Res",
  ["marine biotechnology"] = "Mar. Biotechnol.",
  ["marine biotechnology (new york, n.y.)"] = "Mar Biotechnol (NY)",
  ["marine chemistry"] = "Mar. Chem.",
  ["marine drugs"] = "Mar Drugs",
  ["marine ecology"] = "Mar. Ecol.",
  ["marine ecology (berlin, germany)"] = "Mar Ecol (Berl)",
  ["marine ecology progress series"] = "Mar Ecol Prog Ser",
  ["marine economics and management"] = "Mar. Econ. Manage.",
  ["marine environment research"] = "Mar. Environ. Res.",
  ["marine environmental research"] = "Mar Environ Res",
  ["marine fisheries review"] = "Mar. Fish. Rev.",
  ["marine genomics"] = "Mar Genomics",
  ["marine geodesy"] = "Mar. Geod.",
  ["marine geology"] = "Mar. Geol.",
  ["marine geophysical research"] = "Mar. Geophys. Res.",
  ["marine geophysical researches"] = "Mar. Geophys. Res.",
  ["marine georesources and geotechnology"] = "Mar. Georesour. Geotechnol.",
  ["marine geotechnology"] = "Mar. Geotechnol.",
  ["marine life science & technology"] = "Mar Life Sci Technol",
  ["marine mammal science"] = "Mar Mamm Sci",
  ["marine microbial food webs"] = "Mar. Microb. Food Webs",
  ["marine micropaleontology"] = "Mar. Micropaleontol.",
  ["marine mining"] = "Mar. Min.",
  ["marine observer"] = "Mar. Obs.",
  ["marine ornithology"] = "Mar Ornithol",
  ["marine policy"] = "Mar Policy",
  ["marine pollution bulletin"] = "Mar Pollut Bull",
  ["marine resource economics"] = "Mar. Resour. Econ.",
  ["marine science and biology"] = "Mar Sci Biol",
  ["marine science and technology bulletin"] = "Mar Sci Technol Bull",
  ["marine structures"] = "Mar. struct.",
  ["marine systems & ocean technology"] = "Mar. Syst. Ocean Technol.",
  ["marine technology"] = "Mar. Technol.",
  ["marine technology society journal"] = "Mar. Technol. Soc. J.",
  ["mariners weather log"] = "Mar. Wea. Log",
  ["mario boella series on electromagnetism in information and communication"] = "Mario Boella Ser. Electromagn. Inf. Commun.",
  ["marisia. studii şi materiale. arheologie, istorie, etnografie"] = "Marisia",
  ["maritime affairs: journal of the national maritime foundation of india"] = "Marit. Aff.: J. Natl. Marit. Found. India",
  ["maritime occupational health"] = "Marit. Occup. Health",
  ["maritime policy & management"] = "Marit. Policy Manage.",
  ["maritime policy and management"] = "Maritime Pol. Manage.",
  ["maritime transport research"] = "Marit. Transport Res.",
  ["markers : the annual journal of the association for gravestone studies"] = "Markers",
  ["market forces"] = "Mark Forces",
  ["marketing and public policy conference proceedings. marketing and public policy conference"] = "Mark Public Policy Conf Proc",
  ["marketing health services"] = "Mark Health Serv",
  ["marketing letters"] = "Mark Lett",
  ["marketing science"] = "Marketing Sci.",
  ["markov processes and related fields"] = "Markov Process. Related Fields",
  ["marmara pharmaceutical journal"] = "Marmara Pharm J",
  ["maroc medical"] = "Maroc Med.",
  ["maroc médical"] = "Maroc Med",
  ["marquette business review"] = "Marquette Bus. Rev.",
  ["marquette intellectual property law review"] = "Marquette Intellect Prop Law Rev",
  ["marquette law review"] = "Marquette Law Rev",
  ["marquette medical review"] = "Marquette Med Rev",
  ["marriage & family review"] = "Marriage Fam Rev",
  ["marriage and family living"] = "Marriage Fam Living",
  ["marriage and family newsletter"] = "Marriage Fam Newsl",
  ["mars (los angeles, calif.)"] = "Mars (Los Angel)",
  ["marseille chirurgical"] = "Mars Chir",
  ["marseille medical"] = "Mars. Med.",
  ["marseille médical"] = "Mars Med",
  ["marshall journal of medicine"] = "Marshall J Med",
  ["marsyas. studies in the history of art"] = "Marsyas",
  ["martin gardner’s mathematical games"] = "Martin Gardner’s Math. Games",
  ["marxist perspectives"] = "Marx Perspect",
  ["maryland genealogical society bulletin"] = "Md Geneal Soc Bull",
  ["maryland health bulletin"] = "Md Health Bull",
  ["maryland historical magazine"] = "Md Hist Mag",
  ["maryland journal of contemporary legal issues"] = "Md J Contemp Leg Issues",
  ["maryland journal of international law and trade"] = "Md J Int Law Trade",
  ["maryland law forum"] = "Md Law Forum",
  ["maryland law review (baltimore, md. : 1936)"] = "MD Law Rev",
  ["maryland medical journal"] = "Md. Med. J.",
  ["maryland medical journal (baltimore, md. : 1985)"] = "Md Med J",
  ["maryland medicine"] = "Md. Med.",
  ["maryland medicine : mm : a publication of medchi, the maryland state medical society"] = "Md Med",
  ["maryland nurse"] = "Md. Nurse",
  ["maryland nursing news"] = "Md Nurs News",
  ["maryland state medical journal"] = "Md State Med J",
  ["masa review"] = "MASA Rev.",
  ["masa review : journal of the medical association of the state of alabama"] = "MASA Rev",
  ["masca journal. museum applied science center for archaeology, university of pennsylvania"] = "MascaJ",
  ["masca research papers in science and archaeology"] = "MascaP",
  ["mass communication & society"] = "Mass Commun Soc",
  ["mass spectrometry & purification techniques"] = "Mass Spectrom Purif Tech",
  ["mass spectrometry (tokyo, japan)"] = "Mass Spectrom (Tokyo)",
  ["mass spectrometry reviews"] = "Mass Spectrom Rev",
  ["massachusetts general laws annotated : under arrangement of the official general laws of massachusetts. massachusetts"] = "Mass Gen Laws Annot Mass",
  ["massachusetts journal of mental health"] = "Mass Journal Ment Health",
  ["massachusetts law review"] = "Mass Law Rev",
  ["massachusetts nurse"] = "Mass. Nurse",
  ["massachusetts physician"] = "Mass Physician",
  ["massachusetts reports : cases argued and determined in the supreme judicial court of massachusetts. massachusetts. supreme judicial court"] = "Mass Rep Mass Supreme Judic Court",
  ["mastering mathematical finance"] = "Mastering Math. Finance",
  ["masters abstracts international"] = "Masters Abstr Int",
  ["masters of modern physics"] = "Masters Modern Phys.",
  ["mastia. revista del museo arqueológico municipal de cartagena"] = "Mastia",
  ["mastozoología neotropical"] = "Mastozool Neotrop",
  ["masui (japanese journal of anesthesiology)"] = "Masui.",
  ["masui to sosei. hiroshima journal of anesthesia"] = "Masui To Sosei",
  ["masui. japanese journal of anesthesiology"] = "Masui",
  ["masui. the japanese journal of anesthesiology"] = "Masui",
  ["maszyny elektryczne - zeszyty problemowe"] = "Masz. Elektr. Zesz. Probl.",
  ["maszyny przepływowe"] = "Masz. Przepływ.",
  ["mat. serie a: conferencias, seminarios y trabajos de matematica"] = "MAT Ser. A Conf. Semin. Trab. Mat.",
  ["match (mülheim an der ruhr, germany)"] = "Match (Mulh)",
  ["match-communications in mathematical and in computer chemistry"] = "Match-Commun. Math. Comput. Chem.",
  ["match. communications in mathematical and in computer chemistry"] = "MATCH Commun. Math. Comput. Chem.",
  ["matej bel university"] = "Acta Univ. Mathaei Belii Nat. Sci. Ser. Ser. Math.",
  ["matekon: translations of russian & east european mathematical economics"] = "Matekon",
  ["matematica contemporanea"] = "Mat. Contemp.",
  ["matematica, cultura e societa. rivista dell’unione matematica italiana. serie i"] = "Mat. Cult. Soc. Riv. Unione Mat. Ital. (I)",
  ["matematiche complementari"] = "Mat. Complement.",
  ["matematicheskaya fizika i kompyuternoe modelirovanie. mathematical physics and computer simulation"] = "Mat. Fiz. Kompyut. Model.",
  ["matematicheskaya fizika, analiz, geometriya"] = "Mat. Fiz. Anal. Geom.",
  ["matematicheskaya teoriya igr i ee prilozheniya"] = "Mat. Teor. Igr Prilozh.",
  ["matematicheskie trudy"] = "Mat. Tr.",
  ["matematicheskie voprosy kriptografii. mathematical aspects of cryptography"] = "Mat. Vopr. Kriptografii",
  ["matematicheskie zametki"] = "Mat. Zametki",
  ["matematicheskii sbornik"] = "Mat. Sb.",
  ["matematichki bilten"] = "Mat. Bilten",
  ["matematichki bilten. bulletin mathematique de la societe des mathematiciens de la republique de macedoine"] = "Mat. Bilten",
  ["matematichki vesnik"] = "Mat. Vesnik",
  ["matematichni studii. pratsi lvivskogo matematichnogo tovaristva"] = "Mat. Stud.",
  ["matematicky seminar sntl"] = "Mat. Sem. SNTL",
  ["matematika i matematicheskoe modelirovanie"] = "Mat. Mat. Model.",
  ["matematika. malaysian journal of industrial and applied mathematics"] = "Matematika (Johor)",
  ["matematikai lapok"] = "Mat. Lapok (N.S.)",
  ["matematikai lapok. new series"] = "Mat. Lapok (N.S.)",
  ["matematikos ir informatikos institutas"] = "Liet. Mat. Rink.",
  ["matematisk-fysiske meddelelser"] = "Mat. Medd. Danske Vid. Selsk.",
  ["matemática contemporânea"] = "Mat. Contemp.",
  ["materia medica greca"] = "Mater Med Greca",
  ["materia medica nordmark"] = "Mater Med Nordmark",
  ["materia medica polona"] = "Mater. Med. Pol.",
  ["materia medica polona. polish journal of medicine and pharmacy"] = "Mater Med Pol",
  ["materia socio-medica"] = "Mater Sociomed",
  ["materia therapeutica : aus klinik und praxis für klinik und praxis"] = "Mater Ther",
  ["materia therapeutica dr. kutiak : aus klinik und praxis für klinik und praxis"] = "Mater Ther Dr Kutiak",
  ["material culture"] = "Mater Cult",
  ["material design & processing communications"] = "Mater. Des. Process. Commun.",
  ["material history bulletin. bulletin d'histoire de la culture matérielle"] = "Mater Hist Bull",
  ["material history bulletin. bulletin d’histoire de la culture materielle"] = "Mater Hist Bull",
  ["material religion : the journal of objects, art and belief"] = "Mater Relig",
  ["materiale şi cercetări arheologice"] = "MatCercA",
  ["materiales de construccion"] = "Mater. Constr.",
  ["materiali e contributi per la storia della narrativa greco-latina"] = "MCSN",
  ["materiali e discussioni per l’analisi dei testi classici"] = "MD",
  ["materiali in tehnologije"] = "Mater. Tehnol.",
  ["materialien zur allgemeinen und vergleichenden archäologie"] = "MAVA",
  ["materialien zur bevolkerungswissenschaft / bundesinstitut fur bevolkerungswissenschaft"] = "Mater Bevolkwiss",
  ["materialien zur bevölkerungswissenschaft"] = "Mater Bevolkwiss",
  ["materials & design"] = "Mater Des",
  ["materials (basel, switzerland)"] = "Materials (Basel)",
  ["materials advances"] = "Mater Adv",
  ["materials and corrosion"] = "Mater. Corros.",
  ["materials and design"] = "Mater. Des.",
  ["materials and emerging technologies for sustainability"] = "Mater. Emerging Technol. Sustainability",
  ["materials and manufacturing processes"] = "Mater. Manuf. Processes",
  ["materials and organisms"] = "Mater. Org.",
  ["materials and structures"] = "Mater. Struct.",
  ["materials at high temperatures"] = "Mater. High Temp.",
  ["materials australia"] = "Mater. Aust.",
  ["materials characterization"] = "Mater. Charact.",
  ["materials chemistry and physics"] = "Mater. Chem. Phys.",
  ["materials chemistry frontiers"] = "Mater Chem Front",
  ["materials circular economy"] = "Mater. Circ. Econ.",
  ["materials discovery"] = "Mater. Discover",
  ["materials engineering (modena, italy)"] = "Mater. Eng. (Modena, Italy)",
  ["materials evaluation"] = "Mater. Eval.",
  ["materials express"] = "Mater. Express",
  ["materials focus"] = "Mater. Focus",
  ["materials for quantum technology"] = "Mater. Quantum Technol.",
  ["materials for renewable and sustainable energy"] = "Mater. Renewable Sustainable Energy",
  ["materials forming, machining and tribology"] = "Mater. Form. Mach. Tribol.",
  ["materials forum"] = "Mater. Forum",
  ["materials horizons"] = "Mater. Horiz.",
  ["materials horizons: from nature to nanomaterials"] = "Mater. Horiz. From Nat. Nanomater.",
  ["materials informatics and data science"] = "Mater. Inf. Data Sci.",
  ["materials letters"] = "Mater. Lett.",
  ["materials letters: x"] = "Mater. Lett.: X",
  ["materials management in health care"] = "Mater Manag Health Care",
  ["materials open"] = "Mater. Open",
  ["materials performance"] = "Mater Perform",
  ["materials performance and characterization"] = "Mater. Perform. Charact.",
  ["materials physics and mechanics"] = "Mater. Phys. Mech.",
  ["materials protection"] = "Mater. Prot.",
  ["materials reports: energy"] = "Mater. Rep.: Energy",
  ["materials research"] = "Mater. Res.",
  ["materials research bulletin"] = "Mater. Res. Bull.",
  ["materials research express"] = "Mater. Res. Express",
  ["materials research innovations"] = "Mater. Res. Innovations",
  ["materials research letters"] = "Mater. Res. Lett.",
  ["materials research society symposia proceedings"] = "Mater. Res. Soc. Symp. Proc.",
  ["materials research society symposia proceedings. materials research society"] = "Mater Res Soc Symp Proc",
  ["materials research society symposium proceedings"] = "Mater. Res. Soc. Symp. Proc.",
  ["materials science"] = "Mater. Sci.",
  ["materials science & engineering b: solid-state materials for advanced technology"] = "Mater. Sci. Eng. B Solid-State Mater. Adv. Technol.",
  ["materials science & engineering, a: structural materials: properties, microstructure and processing"] = "Mater. Sci. Eng., A",
  ["materials science & engineering, b: advanced functional solid-state materials"] = "Mater. Sci. Eng., B",
  ["materials science & engineering, b: solid-state materials for advanced technology"] = "Mater. Sci. Eng., B",
  ["materials science & engineering, c: biomimetic and supramolecular systems"] = "Mater. Sci. Eng., C",
  ["materials science & engineering, c: materials for biological applications"] = "Mater. Sci. Eng., C",
  ["materials science & engineering, r: reports"] = "Mater. Sci. Eng., R",
  ["materials science & engineering. a, structural materials : properties, microstructure and processing"] = "Mater Sci Eng A Struct Mater",
  ["materials science & engineering. b, solid-state materials for advanced technology"] = "Mater Sci Eng B Solid State Mater Adv Technol",
  ["materials science & engineering. c, biomimetic and supramolecular systems"] = "Mater Sci Eng C Biomim Supramol Syst",
  ["materials science & engineering. c, biomimetic materials, sensors and systems"] = "Mater Sci Eng C Biomim Mater Sens Syst",
  ["materials science & engineering. c, materials for biological applications"] = "Mater Sci Eng C Mater Biol Appl",
  ["materials science & engineering. r, reports : a review journal"] = "Mater Sci Eng R Rep",
  ["materials science and engineering"] = "Mater. Sci. Eng.",
  ["materials science and engineering a"] = "Mater. Sci. Eng., A",
  ["materials science and engineering b"] = "Mater. Sci. Eng., B",
  ["materials science and engineering c"] = "Mater. Sci. Eng., C",
  ["materials science and engineering r: reports"] = "Mater. Sci. Eng. R Rep.",
  ["materials science and engineering reports"] = "Mater. Sci. Eng., R",
  ["materials science and engineering: c"] = "Mater. Sci. Eng. C",
  ["materials science and engineering: r: reports"] = "Mater. Sci. Eng.: R: Rep.",
  ["materials science and technology"] = "Mater. Sci. Technol.",
  ["materials science for energy technologies"] = "Mater Sci Energy Technol",
  ["materials science forum"] = "Mater. Sci. Forum",
  ["materials science in semiconductor processing"] = "Mater. Sci. Semicond. Process.",
  ["materials science reports"] = "Mater. Sci. Rep.",
  ["materials science research"] = "Materials Sci. Res.",
  ["materials science research international"] = "Mater. Sci. Res. Int.",
  ["materials science series"] = "Mater. Sci. Ser.",
  ["materials science-poland"] = "Mater. Sci.-Pol.",
  ["materials technology"] = "Mater. Technol.",
  ["materials technology (new york, n.y.)"] = "Mater Technol (N Y N Y)",
  ["materials testing"] = "Mater. Test.",
  ["materials theory"] = "Mater. Theory",
  ["materials today"] = "Mater. Today",
  ["materials today (kidlington, england)"] = "Mater Today (Kidlington)",
  ["materials today advances"] = "Mater. Today Adv.",
  ["materials today bio"] = "Mater. Today Bio",
  ["materials today chemistry"] = "Mater. Today Chem.",
  ["materials today communications"] = "Mater. Today Commun.",
  ["materials today electronics"] = "Mater. Today Electron.",
  ["materials today energy"] = "Mater. Today Energy",
  ["materials today nano"] = "Mater. Today Nano",
  ["materials today physics"] = "Mater. Today Phys.",
  ["materials today proceedings"] = "Mater. Today Proc.",
  ["materials today sustainability"] = "Mater. Today Sustainability",
  ["materials today. advances"] = "Mater Today Adv",
  ["materials today. bio"] = "Mater Today Bio",
  ["materials today. chemistry"] = "Mater Today Chem",
  ["materials today. communications"] = "Mater Today Commun",
  ["materials today. energy"] = "Mater Today Energy",
  ["materials today. nano"] = "Mater Today Nano",
  ["materials today. proceedings"] = "Mater Today Proc",
  ["materials today: proceedings"] = "Mater. Today:. Proc.",
  ["materials transactions"] = "Mater. Trans.",
  ["materials transactions, jim"] = "Mater. Trans., JIM",
  ["materials world"] = "Mater. World",
  ["materialwissenschaft und werkstofftechnik"] = "Materwiss Werksttech",
  ["materialy i issledovanija po archeologii sssr"] = "MatIsslA",
  ["materialy i prace antropologiczne"] = "Mater Pr Antropol",
  ["materialy po archeologii bssr"] = "MatABSSR",
  ["materialy po archeologii severnogo pričernomor’ja"] = "MatASevPri",
  ["materiaux & techniques"] = "Mater. Tech.",
  ["materiały badawcze. instytut meteorologii i gospodarki wodnej. seria, hydrologia i oceanologia"] = "Mater Bad Inst Meteor Gospod Wod Ser Hydrol Oceanol",
  ["materiały i prace antropologiczne"] = "Mater Pr Antropol",
  ["materiały starożytne"] = "MatStar",
  ["materiały starożytne i wczesnośredniowieczne"] = "MatStarWczes",
  ["materiały wczesnośredniowieczne"] = "MatWczes",
  ["maternal & child nutrition"] = "Matern Child Nutr",
  ["maternal and child health journal"] = "Matern Child Health J",
  ["maternal and child nutrition"] = "Matern. Child. Nutr.",
  ["maternal and pediatric nutrition"] = "Matern Pediatr Nutr",
  ["maternal health, neonatology and perinatology"] = "Matern Health Neonatol Perinatol",
  ["maternal-child nursing journal"] = "Matern Child Nurs J",
  ["maternal-fetal medicine (wolters kluwer health, inc.)"] = "Matern Fetal Med",
  ["maternidade e infancia"] = "Matern. Infanc. (Sao Paulo)",
  ["maternidade e infancia; arquivos medicos-sociais"] = "Matern Infanc (Sao Paulo)",
  ["maternidade e infância; arquivos médicos-sociais"] = "Matern Infanc (Sao Paulo)",
  ["maternita e infanzia"] = "Matern Infanz (1926)",
  ["maternite; revue pratique d’obstetrique et de puericulture"] = "Maternite",
  ["maternité; revue pratique d'obstétrique et de puériculture"] = "Maternite",
  ["math horizons"] = "Math Horiz.",
  ["math-for-industry (mi) lecture note series"] = "Math-for-Ind. (MI) Lect. Note Ser.",
  ["mathematic study resources"] = "Math. Study Resour.",
  ["mathematica applicanda. matematyka stosowana"] = "Math. Appl. (Warsaw)",
  ["mathematica applicata. yingyong shuxue"] = "Math. Appl. (Wuhan)",
  ["mathematica balkanica"] = "Math. Balkanica (N.S.)",
  ["mathematica balkanica. new series"] = "Math. Balkanica (N.S.)",
  ["mathematica graeca antiqua"] = "Math, Graeca Antiq.",
  ["mathematica japonica"] = "Math. Japon.",
  ["mathematica montisnigri"] = "Math. Montisnigri",
  ["mathematica montisnigri. matematika tsrne gore"] = "Math. Montisnigri",
  ["mathematica moravica"] = "Math. Morav.",
  ["mathematica numerica sinica"] = "Math. Numer. Sin.",
  ["mathematica numerica sinica. jisuan shuxue"] = "Math. Numer. Sin.",
  ["mathematica pannonica"] = "Math. Pannon.",
  ["mathematica pannonica. new series"] = "Math. Pannon. (N. S.)",
  ["mathematica scandinavica"] = "Math. Scand.",
  ["mathematica slovaca"] = "Math. Slovaca",
  ["mathematicae notae"] = "Math. Notae",
  ["mathematical & computational applications"] = "Math. Comput. Appl.",
  ["mathematical analysis and its applications"] = "Math. Anal. Appl.",
  ["mathematical and analytical techniques with applications to engineering"] = "Math. Anal. Tech. Appl. Eng.",
  ["mathematical and computational applications"] = "Math. Comput. Appl.",
  ["mathematical and computational biology and numerical analysis"] = "Math. Comput. Biol. Numer. Anal.",
  ["mathematical and computer modelling"] = "Math Comput Model",
  ["mathematical and computer modelling of dynamical systems"] = "Math. Comput. Model. Dyn. Syst.",
  ["mathematical and computer modelling of dynamical systems. methods, tools and applications in engineering and related sciences"] = "Math. Comput. Model. Dyn. Syst.",
  ["mathematical and information sciences report. series b"] = "Math. Inform. Sci. Rep. Ser. B",
  ["mathematical biosciences"] = "Math. Biosci.",
  ["mathematical biosciences and engineering"] = "Math. Biosci. Eng",
  ["mathematical biosciences and engineering : mbe"] = "Math Biosci Eng",
  ["mathematical biosciences and engineering. mbe"] = "Math. Biosci. Eng.",
  ["mathematical biosciences institute lecture series. stochastics in biological systems"] = "Math. Biosci. Inst. Lect. Ser. Stoch. Biol. Syst.",
  ["mathematical biosciences subseries"] = "Math. Biosci. Subser.",
  ["mathematical chemistry series"] = "Math. Chem. Ser.",
  ["mathematical cognition"] = "Math. Cognit.",
  ["mathematical communications"] = "Math. Commun.",
  ["mathematical concepts and methods in science and engineering"] = "Math. Concepts Methods Sci. Eng.",
  ["mathematical control and related fields"] = "Math. Control Relat. Fields",
  ["mathematical education"] = "Math. Ed.",
  ["mathematical engineering"] = "Math. Eng.",
  ["mathematical engineering in industry"] = "Math. Eng. Indust.",
  ["mathematical engineering, manufacturing, and management sciences"] = "Math. Eng. Manuf. Manag. Sci.",
  ["mathematical finance"] = "Math. Finance",
  ["mathematical finance. an international journal of mathematics, statistics and financial economics"] = "Math. Finance",
  ["mathematical forum"] = "Math. Forum",
  ["mathematical foundations for computer vision, machine learning, and robotics"] = "Math. Found. Comput. Vis. Mach. Learn. Robot.",
  ["mathematical foundations of computing"] = "Math. Found. Comput.",
  ["mathematical gazette"] = "Math. Gaz.",
  ["mathematical geology"] = "Math. Geol.",
  ["mathematical geosciences"] = "Math. Geosci.",
  ["mathematical inequalities & applications"] = "Math. Inequal. Appl.",
  ["mathematical intelligencer"] = "Math. Intell.",
  ["mathematical journal of ibaraki university"] = "Math. J. Ibaraki Univ.",
  ["mathematical journal of okayama university"] = "Math. J. Okayama Univ.",
  ["mathematical lectures from peking university"] = "Math. Lect. Peking Univ.",
  ["mathematical logic and foundations"] = "Math. Logic Found.",
  ["mathematical logic quarterly"] = "Math. Log. Q.",
  ["mathematical medicine and biology"] = "Math. Med. Biol.",
  ["mathematical medicine and biology : a journal of the ima"] = "Math Med Biol",
  ["mathematical medicine and biology. a journal of the ima"] = "Math. Med. Biol.",
  ["mathematical medley"] = "Math. Medley",
  ["mathematical methods and modeling"] = "Math. Methods Model.",
  ["mathematical methods in economics and finance"] = "Math. Methods Econ. Finance",
  ["mathematical methods in the applied sciences"] = "Math. Methods Appl. Sci.",
  ["mathematical methods of operations research"] = "Math. Methods Oper. Res.",
  ["mathematical methods of operations research (heidelberg, germany)"] = "Math Methods Oper Res (Heidelb)",
  ["mathematical methods of statistics"] = "Math. Methods Statist.",
  ["mathematical modeling"] = "Math. Model.",
  ["mathematical modeling and computation"] = "Math. Model. Comput.",
  ["mathematical modeling and computational experiment"] = "Math. Modeling Comput. Experiment",
  ["mathematical modelling"] = "Math. Modell.",
  ["mathematical modelling and analysis"] = "Math. Modell. Anal.",
  ["mathematical modelling and scientific computing"] = "Math. Modelling Sci. Comput.",
  ["mathematical modelling of natural phenomena"] = "Math Model Nat Phenom",
  ["mathematical modelling of systems"] = "Math. Model. Systems",
  ["mathematical modelling: theory and applications"] = "Math. Model. Theory Appl.",
  ["mathematical models & methods in applied sciences"] = "Math. Models Methods Appl. Sci.",
  ["mathematical models & methods in applied sciences : m3as"] = "Math Models Methods Appl Sci",
  ["mathematical models and computer simulations"] = "Math. Models Comput. Simul.",
  ["mathematical models and methods in applied sciences"] = "Math. Models Methods Appl. Sci.",
  ["mathematical neuroscience and applications"] = "Math. Neurosci. Appl.",
  ["mathematical notes"] = "Math. Notes",
  ["mathematical olympiad series"] = "Math. Olymp. Ser.",
  ["mathematical physics analysis and geometry"] = "Math. Phys. Anal. Geom.",
  ["mathematical physics and applied mathematics"] = "Math. Phys. Appl. Math.",
  ["mathematical physics electronic journal"] = "Math. Phys. Electron. J.",
  ["mathematical physics studies"] = "Math. Phys. Stud.",
  ["mathematical physics, analysis and geometry"] = "Math. Phys. Anal. Geom.",
  ["mathematical physics, analysis and geometry. an international journal devoted to the theory and applications of analysis and geometry to physics"] = "Math. Phys. Anal. Geom.",
  ["mathematical population studies"] = "Math Popul Stud",
  ["mathematical population studies. an international journal of mathematical demography"] = "Math. Popul. Stud.",
  ["mathematical problems in engineering"] = "Math. Probl. Eng.",
  ["mathematical problems of computer science"] = "Math. Probl. Comput. Sci.",
  ["mathematical proceedings of the cambridge philosophical society"] = "Math. Proc. Cambridge Philos. Soc.",
  ["mathematical proceedings of the royal irish academy"] = "Math. Proc. R. Ir. Acad.",
  ["mathematical programming"] = "Math. Program.",
  ["mathematical programming computation"] = "Math. Program. Comput.",
  ["mathematical reports (bucuresti)"] = "Math. Rep. (Bucur.)",
  ["mathematical research"] = "Math. Res.",
  ["mathematical research letters"] = "Math. Res. Lett.",
  ["mathematical reviews"] = "MR",
  ["mathematical sciences"] = "Math. Sci.",
  ["mathematical sciences & applications e-notes. an international electronic journal of mathematics"] = "Math. Sci. Appl. E-Notes",
  ["mathematical sciences professional directory"] = "Math. Sci. Prof. Dir.",
  ["mathematical sciences reference series"] = "Math. Sci. Ref. Ser.",
  ["mathematical sciences research hot-line"] = "Math. Sci. Res. Hot-Line",
  ["mathematical sciences research institute publications"] = "Math. Sci. Res. Inst. Publ.",
  ["mathematical social sciences"] = "Math. Social Sci.",
  ["mathematical society of japan. sugaku (mathematics)"] = "Sugaku",
  ["mathematical statistics and learning"] = "Math. Stat. Learn.",
  ["mathematical structures in computer science"] = "Math. Struct. Comput. Sci.",
  ["mathematical structures in computer science. a journal in the applications of categorical, algebraic and geometric methods in computer science"] = "Math. Structures Comput. Sci.",
  ["mathematical studies monograph series"] = "Math. Stud. Monogr. Ser.",
  ["mathematical surveys and monographs"] = "Math. Surveys Monogr.",
  ["mathematical theory and applications"] = "Math. Theory Appl. (Changsha)",
  ["mathematical thinking & learning"] = "Math Think Learn",
  ["mathematical thinking and learning"] = "Math. Thinking Learn.",
  ["mathematical topics"] = "Math. Top.",
  ["mathematical world"] = "Math. World",
  ["mathematics (basel, switzerland)"] = "Mathematics (Basel)",
  ["mathematics and biosciences in interaction"] = "Math. Biosci. Interact.",
  ["mathematics and computers in simulation"] = "Math Comput Simul",
  ["mathematics and financial economics"] = "Math. Financ. Econ.",
  ["mathematics and its applications"] = "Math. Appl.",
  ["mathematics and its applications (china series)"] = "Math. Appl. (China Ser.)",
  ["mathematics and its applications (soviet series)"] = "Math. Appl. (Soviet Ser.)",
  ["mathematics and its applications: modelling, engineering, and social sciences"] = "Math. Appl. Model. Engin. Soc. Sci.",
  ["mathematics and mechanics of complex systems"] = "Math. Mech. Complex Syst.",
  ["mathematics and mechanics of solids"] = "Math. Mech. Solids",
  ["mathematics and mechanics of solids : mms"] = "Math Mech Solids",
  ["mathematics and natural sciences"] = "Math. Nat. Sci.",
  ["mathematics and physics for science and technology"] = "Math. Phys. Sci. Technol.",
  ["mathematics and statistics series"] = "Math. Stat. Ser.",
  ["mathematics and the built environment"] = "Math. Built Environ.",
  ["mathematics and visualization"] = "Math Vis",
  ["mathematics education library"] = "Math. Ed. Lib.",
  ["mathematics for applications"] = "Math Appl",
  ["mathematics for industry (tokyo)"] = "Math. Ind. (Tokyo)",
  ["mathematics for teaching: a problem-based approach"] = "Math. Teach. Probl-Based Approach",
  ["mathematics in action"] = "MathS in Action",
  ["mathematics in computer science"] = "Math. Comput. Sci.",
  ["mathematics in engineering"] = "Math. Eng.",
  ["mathematics in industry"] = "Math. Ind.",
  ["mathematics in industry (philadelphia)"] = "Math. Ind. (Phila.)",
  ["mathematics in mind"] = "Math. Mind",
  ["mathematics in practice and theory"] = "Math. Practice Theory",
  ["mathematics in science and engineering"] = "Math. Sci. Eng.",
  ["mathematics journal of toyama university"] = "Math. J. Toyama Univ.",
  ["mathematics magazine"] = "Math. Mag.",
  ["mathematics newsletter"] = "Math. Newsl.",
  ["mathematics of climate and weather forecasting"] = "Math. Clim. Weather Forecasting",
  ["mathematics of computation"] = "Math. Comput.",
  ["mathematics of control signals and systems"] = "Math. Control Signals Syst.",
  ["mathematics of control, signals, and systems"] = "Math. Control Signals Systems",
  ["mathematics of data"] = "Math. Data",
  ["mathematics of operations research"] = "Math. Oper. Res.",
  ["mathematics of planet earth"] = "Math. Planet Earth",
  ["mathematics online first collections"] = "Math. Online First Collect.",
  ["mathematics open"] = "Math. Open",
  ["mathematics research developments"] = "Math. Res. Dev.",
  ["mathematics research developments series"] = "Math. Res. Dev. Ser.",
  ["mathematics research reports"] = "Math. Res. Rep.",
  ["mathematics studies (tartu)"] = "Math. Stud. (Tartu)",
  ["mathematics study resources"] = "Math. Study Resour.",
  ["mathematics teacher: learning and teaching prek-12"] = "Math. Teach.:  Learn. Teach. PreK-12",
  ["mathematics textbooks for science and engineering"] = "Math. Textb. Sci. Eng.",
  ["mathematics today"] = "Math. Today (Southend-on-Sea)",
  ["mathematics today (ahmedabad)"] = "Math. Today (Ahmedabad)",
  ["mathematics, computation and geometry of data"] = "Math. Comput. Geom. Data",
  ["mathematics, culture, and the arts"] = "Math. Cult. Arts",
  ["mathematics-in-industry case studies"] = "Math Ind Case Stud",
  ["mathematicsl methods in the applied sciences"] = "Math. Methods Appl. Sci.",
  ["mathematik fur anwendungen"] = "Math. Anwend.",
  ["mathematik fur ingenieure, naturwissenschaftler, okonomen und sonstige anwendungsorientierte berufe"] = "Math. Ingen. Naturwiss. Okonom. Sonstige Anwendungsorient. Berufe",
  ["mathematik fur studienanfanger"] = "Math. Studienanfanger",
  ["mathematik für ingenieure und naturwissenschaftler"] = "Math. Ingen. Naturwiss.",
  ["mathematik im kontext"] = "Math. Kontext",
  ["mathematik in beispielen"] = "Math. Beisp.",
  ["mathematik kompakt"] = "Math. Kompakt",
  ["mathematik und naturwissenschaften"] = "Math. Nat.wiss.",
  ["mathematik und wirtschaftswissenschaft"] = "Math. Wirtschaftswiss.",
  ["mathematika. a journal of pure and applied mathematics"] = "Mathematika",
  ["mathematiques & applications (berlin)"] = "Math. Appl. (Berlin)",
  ["mathematiques superieures et speciales"] = "Math. Super. Spec.",
  ["mathematisch centrum, afdeling mathematische besliskunde in samenwerking met het interuniversitair instituut bedrijfskunde delft/rotterdam"] = "Mathematisch Centrum, Afdeling Mathematische Besliskunde in samenwerking met het Interuniversitair Instituut Bedrijfskun",
  ["mathematisch-astronomische blätter—neue folge"] = "Math.-Astronom. Blätter—N.F.",
  ["mathematische annalen"] = "Math. Ann.",
  ["mathematische lehrbucher und monographien, herausgegeben von der deutschen akademie der wissenschaften zu berlin, forschungsinstitut fur mathematik. i. abteilung, mathematische lehrbucher"] = "Mathematische Lehrbucher und Monographien, herausgegeben von der Deutschen Akademie der Wissenschaften zu Berlin, Fors",
  ["mathematische leitfäden"] = "Math. Leitfäden",
  ["mathematische nachrichten"] = "Math. Nachr.",
  ["mathematische optimierung und wirtschaftsmathematik/mathematical optimization and economathematics"] = "Math. Optim. Wirtsch.math./Math. Optim. Economath.",
  ["mathematische schulerbucherei"] = "Math. Schulerbucherei",
  ["mathematische semesterberichte"] = "Math. Semesterber.",
  ["mathematische texte"] = "Math. Texte",
  ["mathematische und naturwissenschaftliche unterricht"] = "Math. Naturwiss. Unterr.",
  ["mathematische zeitschrift"] = "Math. Z.",
  ["mathews journal of case reports"] = "Mathews J Case Rep",
  ["mathews journal of ophthalmology"] = "Mathews J Ophthalmol",
  ["mathware & soft computing"] = "Mathware Soft Comput.",
  ["mathématiques appliquées"] = "Math. Appl.",
  ["mathématiques informatique et sciences humaines"] = "Math. Inform. Sci. Humaines",
  ["matimyas matematika"] = "Matimyas Mat.",
  ["matimyás matematika"] = "Matimyás Mat.",
  ["matrix (stuttgart, germany)"] = "Matrix",
  ["matrix (stuttgart, germany). supplement"] = "Matrix Suppl",
  ["matrix biology"] = "Matrix Biol.",
  ["matrix biology : journal of the international society for matrix biology"] = "Matrix Biol",
  ["matrix biology plus"] = "Matrix Biol Plus",
  ["matrix book series"] = "MATRIX Book Ser.",
  ["matrix. supplement"] = "Matrix. Suppl.",
  ["matter and radiation at extremes"] = "Matter Radiat. at Extremes",
  ["matters select"] = "Matters Sel",
  ["matériaux pour la carte géologique de la suisse, n.s."] = "Matér. Carte Géol. Suisse",
  ["matériaux pour le levé géobotanique de la suisse"] = "Matér. levé géobot. Suisse",
  ["maulana azad college of technology"] = "J. Maulana Azad College Tech.",
  ["max planck research library for the history and development of knowledge. textbooks"] = "Max Planck Res. Libr. Hist. Dev. Knowl., Textb.",
  ["maxillofacial plastic and reconstructive surgery"] = "Maxillofac Plast Reconstr Surg",
  ["maximum entropy and bayesian methods. international workshop on maximum entropy and bayesian methods"] = "Maximum Entropy Bayesian Methods",
  ["maxwell review"] = "Maxwell Rev",
  ["mayo clinic health letter"] = "Mayo Clin. Health Lett.",
  ["mayo clinic health letter (english ed.)"] = "Mayo Clin Health Lett",
  ["mayo clinic proceedings"] = "Mayo Clin Proc",
  ["mayo clinic proceedings. innovations, quality & outcomes"] = "Mayo Clin Proc Innov Qual Outcomes",
  ["mayo clinic proceedings. mayo clinic"] = "Mayo Clin Proc",
  ["mayo clinic women's healthsource"] = "Mayo Clin Womens Healthsource",
  ["mayo clinic women’s healthsource"] = "Mayo Clin Womens Healthsource",
  ["mbm metal bulletin monthly"] = "MBM Met. Bull. Mon.",
  ["mccall's (los angeles, calif. : 1921)"] = "McCalls",
  ["mccall’s (los angeles, calif. : 1921)"] = "McCalls",
  ["mcgeorge law review"] = "McGeorge Law Rev",
  ["mcgill dental review"] = "McGill Dent. Rev.",
  ["mcgill journal of education"] = "McGill J Educ",
  ["mcgill journal of law and health : mjlh = revue de droit et santé de mcgill : rdsm"] = "McGill J Law Health",
  ["mcgill journal of medicine : mjm : an international forum for the advancement of medical sciences by students"] = "Mcgill J Med",
  ["mcgill law journal. revue de droit de mcgill"] = "McGill Law J",
  ["mcgill medical journal"] = "McGill Med J",
  ["mcgill science undergraduate research journal : msurj"] = "McGill Sci Undergrad Res J",
  ["mcgill-queen's/hannah institute studies in the history of medicine, health, and society"] = "Mcgill Queens Hannah Inst Stud Hist Med Health Soc",
  ["mcgill-queen’s studies in the history of ideas"] = "McGill-Queen Stud. Hist. Ideas",
  ["mcgill-queen’s/hannah institute studies in the history of medicine, health, and society"] = "Mcgill Queens Hannah Inst Stud Hist Med Health Soc",
  ["mcgraw-hill series in industrial engineering and management science"] = "McGraw-Hill Ser. Indust. Engrg. Management Sci.",
  ["mcgraw-hill's medicine & health"] = "Mcgraw Hills Med Health",
  ["mcgraw-hill's washington report on medicine & health"] = "Mcgraw Hills Wash Rep Med Health",
  ["mcgraw-hill’s medicine & health"] = "Mcgraw Hills Med Health",
  ["mcgraw-hill’s medicine and health"] = "McGraw Hills Med. Health",
  ["mcgraw-hill’s washington report on medicine & health"] = "Mcgraw Hills Wash Rep Med Health",
  ["mcgraw-hill’s washington report on medicine and health"] = "McGraw Hills Wash. Rep. Med. Health",
  ["mcgregor clinic bulletin"] = "McGregor Clin Bull",
  ["mch news pac"] = "MCH News PAC",
  ["mckinney's consolidated laws of new york annotated : with annotations from state and federal courts and state agencies. new york (state)"] = "McKinneys Consol Laws N Y Annot N Y State",
  ["mckinney’s consolidated laws of new york annotated : with annotations from state and federal courts and state agencies. new york (state)"] = "McKinneys Consol Laws N Y Annot N Y State",
  ["mcn. the american journal of maternal child nursing"] = "MCN Am J Matern Child Nurs",
  ["mcn; american journal of maternal child nursing"] = "MCN. Am. J. Matern. Child Nurs.",
  ["mcv-q; medical college of virginia quarterly"] = "MCVQ",
  ["md (chicago, ill.)"] = "MD (Chic)",
  ["md advisor : a journal for new jersey medical community"] = "MD Advis",
  ["md anderson solid tumor oncology series"] = "Md. Anderson. Solid. Tu.",
  ["md computing"] = "MD Comput.",
  ["md medical newsmagazine"] = "MD",
  ["mdm policy & practice"] = "MDM Policy Pract",
  ["mead johnson symposium on perinatal and developmental medicine"] = "Mead Johnson Symp Perinat Dev Med",
  ["meander. miesięcznik poświęcony kulturze świata starożytnego"] = "Meander",
  ["measurement : interdisciplinary research and perspectives"] = "Measurement ( Mahwah N J)",
  ["measurement : journal of the international measurement confederation"] = "Measurement (Lond)",
  ["measurement and control"] = "Meas. Control",
  ["measurement and evaluation in counseling and development : the official publication of the association for measurement and evaluation in counseling and development"] = "Meas Eval Couns Dev",
  ["measurement in physical education and exercise science"] = "Meas Phys Educ Exerc Sci",
  ["measurement science & technology"] = "Meas Sci Technol",
  ["measurement science and technology"] = "Meas. Sci. Technol.",
  ["measurement techniques"] = "Meas. Tech.",
  ["measurement: food"] = "Meas.: Food",
  ["measurement: interdisciplinary research and perspectives"] = "Meas.: Interdiscip. Res. Perspect.",
  ["measurement: sensors"] = "Meas.: Sens.",
  ["meat science"] = "Meat Sci",
  ["mecanique & industries"] = "Mec. Ind.",
  ["mecanique and industries"] = "Mec. Ind.",
  ["meccanica. an international journal of theoretical and applied mechanics"] = "Meccanica",
  ["meccanica. international journal of theoretical and applied mechanics"] = "Meccanica",
  ["mechanical engineering"] = "Mech. Eng.",
  ["mechanical engineering (new york, n.y. : 1919)"] = "Mech Eng",
  ["mechanical engineering and solid mechanics series"] = "Mech. Eng. Solid Mech. Ser.",
  ["mechanical engineering research studies"] = "Mech. Engrg. Res. Stud.",
  ["mechanical engineering series"] = "Mech. Engrg. Ser.",
  ["mechanical sciences"] = "Mech. Sci.",
  ["mechanical systems and signal processing"] = "Mech. Syst. Sig. Process.",
  ["mechanics & chemistry of biosystems : mcb"] = "Mech Chem Biosyst",
  ["mechanics & industry"] = "Mech. Ind.",
  ["mechanics and mathematical methods. a series of handbooks. series ii: thermal stresses"] = "Mech. Math. Methods. Ser. II: Thermal Stresses",
  ["mechanics based design of structures and machines"] = "Mech. Based Des. Struct. Mach.",
  ["mechanics of advanced materials and modern processes"] = "Mech Adv Mater Mod Process",
  ["mechanics of advanced materials and structures"] = "Mech. Adv. Mater. Struct.",
  ["mechanics of cohesive-frictional materials"] = "Mech. Cohesive-frict. Mater.",
  ["mechanics of composite materials"] = "Mech. Compos. Mater.",
  ["mechanics of composite materials and structures"] = "Mech. Compos. Mater. Struct.",
  ["mechanics of elastic and inelastic solids"] = "Mech. Elastic Inelastic Solids",
  ["mechanics of fracture"] = "Mech. of Fracture",
  ["mechanics of materials"] = "Mech. Mater.",
  ["mechanics of materials : an international journal"] = "Mech Mater",
  ["mechanics of soft materials"] = "Mech Soft Mater",
  ["mechanics of solids"] = "Mech. Solids",
  ["mechanics of structures and machines"] = "Mech. Structures Mach.",
  ["mechanics of time dependent materials"] = "Mech. Time-Depend. Mater.",
  ["mechanics of time-dependent materials"] = "Mech Time Depend Mater",
  ["mechanics research communications"] = "Mech. Res. Commun.",
  ["mechanik, werkstoffe und konstruktion im bauwesen"] = "Mech. Werkst. Konstr. Bauwes.",
  ["mechanism and machine theory"] = "Mech. Mach. Theory",
  ["mechanisms and machine science"] = "Mech. Mach. Sci.",
  ["mechanisms of ageing and development"] = "Mech Ageing Dev",
  ["mechanisms of development"] = "Mech Dev",
  ["mechatronic systems and control"] = "Mechatron. Syst. Control",
  ["mechatronics : the science of intelligent machines"] = "Mechatronics (Oxf)",
  ["med (new york, n.y.)"] = "Med (N Y)",
  ["med one"] = "Med One",
  ["medcomm - biomaterials and applications"] = "MedComm - Biomater. Appl.",
  ["medcomm - future medicine"] = "MedComm - Future Med.",
  ["meddelande från familjehistoriska projektet, historiska institutionen, uppsala universitet"] = "Medd Fam Hist Proj Hist Inst Upps Univ",
  ["meddelanden fran flyg- och navalmedicinska namnden. statens namnd for flyg- och navalmedicinsk forsknings- och forsoksverksamhet"] = "Medd Flyg Navalmed Namnd Statens Namnd For Flyg Navalmed Forsk Forsoksvererksamhet",
  ["meddelanden från flyg- och navalmedicinska nämnden. statens nämnd för flyg- och navalmedicinsk forsknings- och försöksverksamhet"] = "Medd Flyg Navalmed Namnd Statens Namnd For Flyg Navalmed Forsk Forsoksvererksamhet",
  ["meddelanden från lunds universtitets historiska museum"] = "MeddelLund",
  ["meddelelser fra norsk farmaceutisk selskap"] = "Medd Nor Farm Selsk",
  ["meddelelser fra ny carlsberg glyptotek"] = "MeddelGlypt",
  ["meddelelser fra sundhedsstyrelsen. denmark. sundhedsstyrelsen. beredskabsafdelingen"] = "Medd Sundhedsstyr Beredskabsafdelingen",
  ["meddelelser fra thorvaldsens museum"] = "MeddelThor",
  ["meddelelser om groenland: geoscience"] = "Medd. Groenl.: Geosc.",
  ["meddelelser. denmark. sundhedsstyrelsen. beredskabsafdelingen"] = "Medd Sundhedsstyr Beredskabsafdelingen",
  ["medecine & chirurgie digestives"] = "Med Chir Dig",
  ["medecine aeronautique et spatiale"] = "Med Aeronaut Spat",
  ["medecine dans le monde"] = "Med Monde",
  ["medecine d’afrique noire"] = "Med Afr Noire",
  ["medecine d’egypte"] = "Med Egypte",
  ["medecine et armees"] = "Med Armees",
  ["medecine et audio vision"] = "Med. Audio Vis.",
  ["medecine et chirurgie digestives"] = "Med. Chir. Dig.",
  ["medecine et hygiene"] = "Med. Hyg. (Geneve)",
  ["medecine et laboratoire"] = "Med Lab",
  ["medecine et maladies infectieuses"] = "Med Mal Infect",
  ["medecine infantile"] = "Med. Infant. (Paris)",
  ["medecine interne"] = "Med. Interne",
  ["medecine legale et dommage corporel"] = "Med. Leg. Dommage Corpor.",
  ["medecine sciences : m/s"] = "Med Sci (Paris)",
  ["medecine tropicale"] = "Med. Trop. (Mars.)",
  ["medecine tropicale : revue du corps de sante colonial"] = "Med Trop (Mars)",
  ["mededelingen (rijksuniversiteit te gent. fakulteit van de landbouwkundige en toegepaste biologische wetenschappen)"] = "Meded Rijksuniv Gent Fak Landbouwkd Toegep Biol Wet",
  ["mededelingen van de faculteit landbouwwetenschappen, universiteit gent"] = "Meded. Fac. Landbouwwet., Rijksuniv. Gent",
  ["mededelingen van de koninklijke academie voor wetenschappen, letteren en schone kunsten van belgie. klasse der wetenschappen"] = "Meded K Acad Wet Lett Schone Kunsten Belg Kl Wet",
  ["mededelingen van de koninklijke academie voor wetenschappen, letteren en schone kunsten van belgië"] = "Med. Konink. Acad. Wetensch. België",
  ["mededelingen van de koninklijke academie voor wetensehappen, letteren & schone kunsten van belgië, kl. der letteren"] = "MAWBL",
  ["mededelingen van de landbouwhoogeschool te wageningen"] = "Meded. Landbouwhoogesch. Wagening.",
  ["mededelingen van het nederlands instituut te rome"] = "MededRom",
  ["mededelingen van het nederlandsch historisch instituut te rome"] = "MNIR",
  ["mededelingen van het wiskundig genootschap"] = "Meded. Wiskd. Genoot.",
  ["mededelingenblad. vereniging van vrienden allard pierson museum"] = "VerAmstMeded",
  ["mededportal : the journal of teaching and learning resources"] = "MedEdPORTAL",
  ["medelhavsmuseet. bulletin"] = "MedelhavsMusB",
  ["medeniyet medical journal"] = "Medeni Med J",
  ["medécine et armées"] = "Med Armees",
  ["medgenmed : medscape general medicine"] = "MedGenMed",
  ["medi-cal policy institute issue brief"] = "Medi-Cal Policy Inst Issue Brief",
  ["media & gender monitor"] = "Media Gend Monit",
  ["media & gender monitor / an occasional publication by the women’s programme at the world association for christian communication"] = "Media & Gend Monit",
  ["media and communication"] = "Media Commun",
  ["media asia"] = "Media Asia",
  ["media culture & society"] = "Media Cult. Soc.",
  ["media development"] = "Media Dev",
  ["media history"] = "Media Hist",
  ["media peternakan"] = "Media Peternakan Fakultas Peternakan Institut Pertanian Bogor",
  ["media psychology"] = "Media Psychol",
  ["media, culture, and society"] = "Media Cult Soc",
  ["mediaeval studies"] = "Mediaev Stud",
  ["mediaevalia philosophica polonorum"] = "Mediaev Philos Pol",
  ["mediastinum (hong kong, china)"] = "Mediastinum",
  ["mediators of inflammation"] = "Mediators Inflamm",
  ["medica britannica"] = "Med Br (Lond)",
  ["medical & biological engineering"] = "Med Biol Eng",
  ["medical & biological engineering & computing"] = "Med Biol Eng Comput",
  ["medical & biological illustration"] = "Med Biol Illus",
  ["medical & clinical research"] = "Med Clin Res",
  ["medical abstracts"] = "Med Abstr",
  ["medical acupuncture"] = "Med Acupunct",
  ["medical affairs"] = "Med Aff",
  ["medical and biological engineering"] = "Med. Biol. Eng.",
  ["medical and biological engineering and computing"] = "Med. Biol. Eng. Comput.",
  ["medical and biological illustration"] = "Med. Biol. Illus.",
  ["medical and pediatric oncology"] = "Med Pediatr Oncol",
  ["medical and pediatric oncology. supplement"] = "Med Pediatr Oncol Suppl",
  ["medical and veterinary entomology"] = "Med Vet Entomol",
  ["medical annals of the district of columbia"] = "Med. Ann. Dist. Columbia",
  ["medical anthropology"] = "Med Anthropol",
  ["medical anthropology newsletter"] = "Med Anthropol Newsl",
  ["medical anthropology quarterly"] = "Med Anthropol Q",
  ["medical archives (sarajevo, bosnia and herzegovina)"] = "Med Arch",
  ["medical art"] = "Med Art",
  ["medical arts and sciences"] = "Med Arts Sci",
  ["medical aspects of human sexuality"] = "Med Aspects Hum Sex",
  ["medical biochemistry"] = "Med Biochem",
  ["medical biology"] = "Med Biol",
  ["medical bookman and historian"] = "Med Bookm Hist",
  ["medical bulletin (ann arbor, mich.)"] = "Med Bull (Ann Arbor)",
  ["medical bulletin / u.s. army first army united states. army. army, 1st"] = "Med Bull U S Army Army 1st",
  ["medical bulletin of istanbul medical faculty"] = "Med Bull Istanbul Med Fac",
  ["medical bulletin of the u. s. army, europe. united states. army, europe. medical division"] = "Med Bull US Army Eur",
  ["medical bulletin of uganda"] = "Med Bull Uganda",
  ["medical bulletin, veterans administration"] = "Med. Bull. Vet. Adm.",
  ["medical bulletin. harrisburg polyclinic hospital"] = "Med Bull Harrisburg Polyclin Hosp",
  ["medical bulletin. st. louis university"] = "Med Bull St Louis Univ",
  ["medical bulletin. united states. army. army, 1st"] = "Med Bull U S Army Army 1st",
  ["medical bulletin. united states. army. european command. medical division"] = "Med Bull U S Army Eur Command Med Div",
  ["medical bulletin. united states. army. forces in the european theater. office of the theater chief surgeon"] = "Med Bull U S Army Force Europe Theater Off Theater Chief Surg",
  ["medical bulletin. united states. army. mediterranean theater of operations"] = "Med Bull U S Army Mediterr Theater Oper",
  ["medical bulletin. united states. veterans administration. department of medicine and surgery"] = "Med Bull Vet Adm",
  ["medical bulletin. united states. veterans administration. dept. of medicine and surgery"] = "Med Bull Vet Adm",
  ["medical cannabis and cannabinoids"] = "Med Cannabis Cannabinoids",
  ["medical care"] = "Med Care",
  ["medical care research and review"] = "Med. Care Res. Rev.",
  ["medical care research and review : mcrr"] = "Med Care Res Rev",
  ["medical care review"] = "Med Care Rev",
  ["medical case reports (wilmington, del.)"] = "Med Case Rep (Wilmington)",
  ["medical case reports and short reviews"] = "Med Case Rep Short Rev",
  ["medical claims management"] = "Med Claims Manag",
  ["medical clinics of north america"] = "Med. Clin. North Am.",
  ["medical communications"] = "Med Commun (N J)",
  ["medical computer vision : large data in medical imaging : third international miccai workshop, mcv 2013, nagoya, japan, september 26, 2013 : revised selected papers. mcv (workshop) (3rd : 2013 : nagoya-shi, japan)"] = "Med Comput Vis (2013)",
  ["medical computer vision and bayesian and graphical models for biomedical imaging : miccai 2016 international workshop, mcv and bambi, athens, greece, october 21, 2016 : revised selected papers"] = "Med Comput Vis Bayesian Graph Models Biomed Imaging (2016)",
  ["medical counterpoint"] = "Med Counterpoint",
  ["medical critic and psychological journal"] = "Med Crit Psychol J",
  ["medical decision making"] = "Med. Decis. Making",
  ["medical decision making : an international journal of the society for medical decision making"] = "Med Decis Making",
  ["medical design and material"] = "Med Des Mater",
  ["medical device technology"] = "Med Device Technol",
  ["medical devices & sensors"] = "Med Devices Sens",
  ["medical devices (auckland, n.z.)"] = "Med Devices (Auckl)",
  ["medical devices: evidence and research"] = "Med. Devices: Evidence Res.",
  ["medical digest"] = "Med Dig",
  ["medical dimensions"] = "Med Dimens",
  ["medical directory of chicago (including cook co.)"] = "Med Dir Chic",
  ["medical dosimetry"] = "Med. Dosim.",
  ["medical dosimetry : official journal of the american association of medical dosimetrists"] = "Med Dosim",
  ["medical economics"] = "Med Econ",
  ["medical education"] = "Med Educ",
  ["medical education (chicago, ill.)"] = "Med Educ (Chicago Ill)",
  ["medical education online"] = "Med Educ Online",
  ["medical electron microscopy : official journal of the clinical electron microscopy society of japan"] = "Med Electron Microsc",
  ["medical electronics"] = "Med Electron",
  ["medical electronics & biological engineering"] = "Med Electron Biol Eng",
  ["medical electronics and biological engineering"] = "Med. Electron. Biol. Eng.",
  ["medical encounter"] = "Med Encount",
  ["medical engineering & physics"] = "Med Eng Phys",
  ["medical engineering and physics"] = "Med. Eng. Phys.",
  ["medical entomology and zoology"] = "Med. Entomol. Zool.",
  ["medical epigenetics"] = "Med Epigenet",
  ["medical ethics (burlington, mass.)"] = "Med Ethics (Burlingt Mass)",
  ["medical ethics : journal of forum for medical ethics society"] = "Med Ethics",
  ["medical facts and observations"] = "Med Facts Obs",
  ["medical gas research"] = "Med Gas Res",
  ["medical group management"] = "Med Group Manage",
  ["medical group management journal"] = "Med Group Manage J",
  ["medical group management journal / mgma"] = "Med Group Manage J",
  ["medical gynaecology and sociology"] = "Med Gynaecol Sociol",
  ["medical gynaecology, andrology, and sociology"] = "Med Gynaecol Androl Sociol",
  ["medical heritage"] = "Med Herit",
  ["medical history"] = "Med Hist",
  ["medical history australia"] = "Med Hist Aust",
  ["medical history. supplement"] = "Med Hist Suppl",
  ["medical history: a quarterly journal devoted to the history of medicine and related sciences"] = "MedHist",
  ["medical humanities"] = "Med Humanit",
  ["medical humanities review"] = "Med Humanit Rev",
  ["medical hypotheses"] = "Med Hypotheses",
  ["medical hypotheses and research : mhr"] = "Med Hypotheses Res",
  ["medical hypothesis, discovery & innovation ophthalmology journal"] = "Med Hypothesis Discov Innov Ophthalmol",
  ["medical image analysis"] = "Med Image Anal",
  ["medical image understanding and analysis : 21st annual conference, miua 2017, edinburgh, uk, july 11-13, 2017, proceedings. medical image understanding and analysis (conference) (21st : 2017 : edinburgh, scotland)"] = "Med Image Underst Anal Conf (2017)",
  ["medical imaging and augmented reality : 7th international conference, miar 2016, bern, switzerland, august 24-26, 2016 : proceedings. miar (workshop) (7th : 2016 : bern, switzerland)"] = "Med Imaging Augment Real (2016)",
  ["medical immunology"] = "Med. Immunol.",
  ["medical immunology (london, england)"] = "Med Immunol",
  ["medical informatics"] = "Med. Inform. (Lond.)",
  ["medical informatics = médecine et informatique"] = "Med Inform (Lond)",
  ["medical informatics and the internet in medicine"] = "Med Inform Internet Med",
  ["medical innovation & business"] = "Med Innov Bus",
  ["medical insight"] = "Med Insight",
  ["medical instrumentation"] = "Med Instrum",
  ["medical instrumentation (luton, england)"] = "Med Instrum (Luton)",
  ["medical interface"] = "Med Interface",
  ["medical investigator (louisville, ky.)"] = "Med Investig Louisville",
  ["medical journal (fort sam houston, tex.)"] = "Med J (Ft Sam Houst Tex)",
  ["medical journal of australia"] = "Med. J. Aust.",
  ["medical journal of malaya"] = "Med. J. Malaya",
  ["medical journal of malaysia"] = "Med. J. Malaysia",
  ["medical journal of obstetrics and gynecology"] = "Med J Obstet Gynecol",
  ["medical journal of osaka university"] = "Med J Osaka Univ",
  ["medical journal of the islamic republic of iran"] = "Med J Islam Repub Iran",
  ["medical journal of zambia"] = "Med J Zambia",
  ["medical journal, armed forces india"] = "Med J Armed Forces India",
  ["medical laboratory sciences"] = "Med Lab Sci",
  ["medical laboratory technology"] = "Med Lab Technol",
  ["medical laser application"] = "Med. Laser Appl",
  ["medical laser application : international journal for laser treatment and research"] = "Med Laser Appl",
  ["medical law international"] = "Med Law Int",
  ["medical law review"] = "Med Law Rev",
  ["medical letter on drugs and therapeutics"] = "Med. Lett. Drugs Ther.",
  ["medical libraries"] = "Med Lib",
  ["medical library and historical journal"] = "Med Library Hist J",
  ["medical malpractice cost containment journal"] = "Med. Malpract. Cost Containment J.",
  ["medical management network"] = "Med Manag Netw",
  ["medical marketing & media"] = "Med Mark Media",
  ["medical marketing and media"] = "Med. Mark. Media",
  ["medical microbiology and immunology"] = "Med Microbiol Immunol",
  ["medical microbiology letters : an international journal for rapid communications on all aspects of medical and clinical microbiology"] = "Med Microbiol Lett",
  ["medical mission to austria, july 1-august 8, 1947. american unitarian service committee in cooperation with world health organization interim commission. [abridged report submitted by erwin kohn]. unitarian service committee"] = "Med Mission Austria",
  ["medical molecular morphology"] = "Med Mol Morphol",
  ["medical mycology"] = "Med Mycol",
  ["medical mycology : official publication of the international society for human and animal mycology"] = "Med Mycol",
  ["medical mycology case reports"] = "Med Mycol Case Rep",
  ["medical mycology journal"] = "Med Mycol J",
  ["medical mycology: open access"] = "Med Mycol Open Access",
  ["medical network strategy report"] = "Med Netw Strategy Rep",
  ["medical news letter. aviation supplement. united states. navy department department of aviation medicine"] = "Med News Lett Aviat Suppl",
  ["medical news schering"] = "Med News Schering",
  ["medical newsletter"] = "Med Newsl (Lond)",
  ["medical oncology"] = "Med. Oncol.",
  ["medical oncology (northwood, london, england)"] = "Med Oncol",
  ["medical oncology and tumor pharmacotherapy"] = "Med Oncol Tumor Pharmacother",
  ["medical opinion & review"] = "Med Opin Rev",
  ["medical opinion (new york)"] = "Med Opin (N Y)",
  ["medical photonics"] = "Med. Photonics",
  ["medical physics"] = "Med Phys",
  ["medical principles and practice : international journal of the kuwait university, health science centre"] = "Med Princ Pract",
  ["medical problems of performing artists"] = "Med Probl Perform Art",
  ["medical proceedings : a south african journal for the advancement of medical science = mediese bydraes : 'n suid-afrikaanse tydskrif vir die bevordering van die geneeskunde"] = "Med Proc (Johannesb)",
  ["medical products sales"] = "Med. Prod. Sales",
  ["medical products sales : mps : the official journal of the american surgical trade association"] = "Med Prod Sales",
  ["medical progress through technology"] = "Med Prog Technol",
  ["medical progress; a review of medical advances"] = "Med Prog",
  ["medical radiographer"] = "Med Radiogr",
  ["medical radiography and photography"] = "Med Radiogr Photogr",
  ["medical radiology"] = "Med Radiol",
  ["medical record (new york : 1934)"] = "Med Rec",
  ["medical record (reading, pa.)"] = "Med Rec (Reading)",
  ["medical record and annals"] = "Med Rec Ann",
  ["medical record and health care information journal"] = "Med Rec Health Care Inf J",
  ["medical record news"] = "Med Rec News",
  ["medical reference services quarterly"] = "Med Ref Serv Q",
  ["medical research and innovations"] = "Med Res Innov",
  ["medical research archives"] = "Med Res Arch",
  ["medical research council memorandum"] = "Memo Med Res Counc",
  ["medical research council special report series"] = "Med. Res. Counc. Spec. Rep. Ser. (Lond.)",
  ["medical research engineering"] = "Med Res Eng",
  ["medical research report. united states. navy. mine defense laboratory. medical research dept"] = "Med Res Rep",
  ["medical research reviews"] = "Med. Res. Rev.",
  ["medical science"] = "Med Sci",
  ["medical science educator"] = "Med Sci Educ",
  ["medical science monitor"] = "Med. Sci. Monit.",
  ["medical science monitor : international medical journal of experimental and clinical research"] = "Med Sci Monit",
  ["medical science monitor basic research"] = "Med Sci Monit Basic Res",
  ["medical science research"] = "Med Sci Res",
  ["medical sciences (basel, switzerland)"] = "Med Sci (Basel)",
  ["medical sciences (london, england)"] = "Prog Nucl Energy 7 Med Sci",
  ["medical section proceedings"] = "Med. Sect. Proc.",
  ["medical self-care"] = "Med Self Care",
  ["medical sensors & imaging"] = "Med. Sens. Imaging",
  ["medical service"] = "Med Serv",
  ["medical services journal, canada"] = "Med Serv J Can",
  ["medical society reporter"] = "Med Soc Report",
  ["medical staff counselor"] = "Med. Staff Couns.",
  ["medical student research journal"] = "Med Student Res J",
  ["medical teacher"] = "Med Teach",
  ["medical technicians bulletin"] = "Med Techn Bull",
  ["medical times"] = "Med Times",
  ["medical toxicology"] = "Med Toxicol",
  ["medical toxicology and adverse drug experience"] = "Med Toxicol Adverse Drug Exp",
  ["medical trial technique quarterly"] = "Med Trial Tech Q",
  ["medical tribune and medical news"] = "Med Trib Med News",
  ["medical ultrasonography"] = "Med Ultrason",
  ["medical ultrasound"] = "Med Ultrasound",
  ["medical ultrasound, and preterm, perinatal and paediatric image analysis"] = "Med Ultrasound Preterm Perinat Paediatr Image Anal (2020)",
  ["medical virology from pathogenesis to disease control"] = "Med. Virol. Pathog. Dis. Control",
  ["medical waste analyst"] = "Med Waste Anal",
  ["medical woman's journal"] = "Med Womans J",
  ["medical woman’s journal"] = "Med Womans J",
  ["medical world"] = "Med World",
  ["medical world news"] = "Med World News",
  ["medical writing"] = "Med. Writ.",
  ["medical/dental journal"] = "Med Dent J",
  ["medicamento, historia e sociedade"] = "Medicam Hist Soc",
  ["medicamento, história e sociedade"] = "Medicam Hist Soc",
  ["medicamentos de actualidad. drugs of today"] = "Med Actual",
  ["medicare & medicaid research review"] = "Medicare Medicaid Res Rev",
  ["medicare brief"] = "Medicare Brief",
  ["medicare brief / national academy of social insurance"] = "Medicare Brief",
  ["medicc review"] = "MEDICC Rev",
  ["medicina (florence, italy)"] = "Medicina (Firenze)",
  ["medicina (kaunas, lithuania)"] = "Medicina (Kaunas)",
  ["medicina (universidad de los andes)"] = "Med Bogota Colomb",
  ["medicina clinica"] = "Med Clin (Barc)",
  ["medicina clinica e sperimentale"] = "Med Clin Sper",
  ["medicina clínica"] = "Med Clin (Barc)",
  ["medicina clínica (english ed.)"] = "Med Clin (Engl Ed)",
  ["medicina contemporanea"] = "Med. Contemp.",
  ["medicina contemporanea (lisbon, portugal)"] = "Med Contemp",
  ["medicina contemporânea (lisbon, portugal)"] = "Med Contemp",
  ["medicina cutanea ibero-latino-americana"] = "Med. Cutan. Ibero. Lat. Am.",
  ["medicina cutánea ibero-latino-americana"] = "Med Cutan Ibero Lat Am",
  ["medicina de hoje"] = "Med Hoje",
  ["medicina del deporte y del trabajo"] = "Med Deporte Trab",
  ["medicina del lavoro"] = "Med. Lav.",
  ["medicina dello sport"] = "Med. Sport (Roma)",
  ["medicina dello sport; rivista di fisiopatologia dello sport"] = "Med Sport (Roma)",
  ["medicina e historia"] = "Med Hist (Barc)",
  ["medicina em revista"] = "Med Rev",
  ["medicina espanola"] = "Med. Esp.",
  ["medicina española"] = "Med Esp",
  ["medicina et pharmacologia experimentalis (international journal of experimental medicine)"] = "Med. Pharmacol. Exp. Int. J. Exp. Med.",
  ["medicina et pharmacologia experimentalis. international journal of experimental medicine"] = "Med Pharmacol Exp Int J Exp Med",
  ["medicina experimentalis (international journal of experimental medicine)"] = "Med. Exp. Int. J. Exp. Med.",
  ["medicina experimentalis : international journal of experimental medicine"] = "Med Exp Int J Exp Med",
  ["medicina experimentalis. international journal of experimental medicine"] = "Med Exp Int J Exp Med",
  ["medicina fisica y rehabilitacion"] = "Med. Fis. Rehabil.",
  ["medicina física y rehabilitación"] = "Med Fis Rehabil",
  ["medicina infantil"] = "Med Infant",
  ["medicina intensiva"] = "Med Intensiva",
  ["medicina intensiva / sociedad espanola de medicina intensiva y unidades coronarias"] = "Med Intensiva",
  ["medicina interna"] = "Med. Interna",
  ["medicina interna (bucharest, romania : 1991)"] = "Med Interna",
  ["medicina internacia revuo"] = "Med Int Rev",
  ["medicina latina; revista medica de colaboracion cientifica"] = "Med Lat",
  ["medicina latina; revista médica de colaboración científica"] = "Med Lat",
  ["medicina nei secoli"] = "Med Secoli",
  ["medicina nei secoli. arte e scienza"] = "MediSec",
  ["medicina oral"] = "Med. Oral",
  ["medicina oral : organo oficial de la sociedad espanola de medicina oral y de la academia iberoamericana de patologia y medicina bucal"] = "Med Oral",
  ["medicina oral : órgano oficial de la sociedad española de medicina oral y de la academia iberoamericana de patología y medicina bucal"] = "Med Oral",
  ["medicina oral, patologia oral y cirugia bucal"] = "Med Oral Patol Oral Cir Bucal",
  ["medicina oral, patología oral y cirugía bucal"] = "Med Oral Patol Oral Cir Bucal",
  ["medicina panamericana"] = "Med Panam",
  ["medicina practica"] = "Med Pract (Zaragoza)",
  ["medicina práctica"] = "Med Pract (Zaragoza)",
  ["medicina psicosomatica"] = "Med Psicosom",
  ["medicina thoracalis"] = "Med Thorac",
  ["medicina tradicional"] = "Med Tradic",
  ["medicina tropical"] = "Med. Trop. (Madr.)",
  ["medicina veterinaria brazil"] = "Med. Vet. Braz.",
  ["medicina y cirugia"] = "Med Cir (Bogota)",
  ["medicina y cirugia de guerra"] = "Med. Cir. Guerra",
  ["medicina y cirugía"] = "Med Cir (Bogota)",
  ["medicina y cirugía de guerra"] = "Med Cir Guerra",
  ["medicina y desarrollo"] = "Med Desarrollo",
  ["medicina y seguridad del trabajo"] = "Med Segur Trab (Madr)",
  ["medicina, cirurgia, farmacia"] = "Med Cir Farm",
  ["medicina, cirurgia, farmácia"] = "Med Cir Farm",
  ["medicinal & aromatic plants"] = "Med Aromat Plants (Los Angel)",
  ["medicinal and aromatic plant science and biotechnology"] = "Med Aromat Plant Sci Biotechnol",
  ["medicinal chemistry"] = "Med. Chem.",
  ["medicinal chemistry (shariqah, united arab emirates)"] = "Med Chem",
  ["medicinal chemistry (sharjah, united arab emirates)"] = "Med. Chem. (Sharjah, United Arab Emirates)",
  ["medicinal chemistry (shāriqah (united arab emirates))"] = "Med Chem",
  ["medicinal chemistry research"] = "Med. Chem. Res.",
  ["medicinal chemistry research : an international journal for rapid communications on design and mechanisms of action of biologically active agents"] = "Med Chem Res",
  ["medicinal plants"] = "Med. Plants",
  ["medicinal research reviews"] = "Med Res Rev",
  ["medicină internă"] = "Med Interna (Bucur)",
  ["medicină internă (bucharest, romania : 1991)"] = "Med Interna",
  ["medicine & health"] = "Med Health",
  ["medicine & health (1997)"] = "Med Health",
  ["medicine & science in sports & exercise"] = "Med. Sci. Sports Exercise",
  ["medicine (abingdon, england : uk ed.)"] = "Medicine (Abingdon)",
  ["medicine 2.0"] = "Med 2 0",
  ["medicine and global survival : m & gs"] = "Med Glob Surviv",
  ["medicine and health"] = "Med. Health",
  ["medicine and health, rhode island"] = "Med Health R I",
  ["medicine and law"] = "Med Law",
  ["medicine and pharmacy reports"] = "Med Pharm Rep",
  ["medicine and science in sports"] = "Med Sci Sports",
  ["medicine and science in sports and exercise"] = "Med Sci Sports Exerc",
  ["medicine and society"] = "Med Soc (Berkeley)",
  ["medicine and sport"] = "Med Sport",
  ["medicine and sport science"] = "Med Sport Sci",
  ["medicine and war"] = "Med War",
  ["medicine anthropology theory"] = "Med Anthropol Theory",
  ["medicine at emory"] = "Med Emory",
  ["medicine illustrated"] = "Med Illus",
  ["medicine in drug discovery"] = "Med Drug Discov",
  ["medicine in microecology"] = "Med. Microecol.",
  ["medicine in novel technology and devices"] = "Med Nov Technol Devices",
  ["medicine in omics"] = "Med. Omics",
  ["medicine international (uk ed.)"] = "Med Int",
  ["medicine of the year"] = "Med Year",
  ["medicine science"] = "Med Sci (Turkey)",
  ["medicine studies"] = "Med Stud",
  ["medicine today"] = "Med Today",
  ["medicine today & tomorrow"] = "Med Today Tomorrow",
  ["medicine's geographic heritage"] = "Med Geogr Herit",
  ["medicine, conflict and survival"] = "Med. Confl. Surviv.",
  ["medicine, conflict, and survival"] = "Med Confl Surviv",
  ["medicine, health care, and philosophy"] = "Med Health Care Philos",
  ["medicine, science and the law"] = "Med.Sci. Law",
  ["medicine, science, and the law"] = "Med Sci Law",
  ["medicines (basel, switzerland)"] = "Medicines (Basel)",
  ["medicines, ethics and practice : a guide for pharmacists. royal pharmaceutical society of great britain"] = "Med Ethics Pract Royal Pharm Soc G B",
  ["medicine’s geographic heritage"] = "Med Geogr Herit",
  ["medicinsk forum"] = "Med Forum",
  ["medicinska föreningens tidskrift"] = "Med Foren Tidskr",
  ["medicinski arhiv"] = "Med Arh",
  ["medicinski glasnik"] = "Med Glas",
  ["medicinski glasnik : official publication of the medical association of zenica-doboj canton, bosnia and herzegovina"] = "Med Glas (Zenica)",
  ["medicinski pregled"] = "Med Pregl",
  ["medicinski razgledi"] = "Med Razgl",
  ["medicinski vjesnik (osijek, croatia)"] = "Med Vjesn (Osijek)",
  ["medicínska etika a bioetika : časopis ústavu medicínskej etiky a bioetiky = medical ethics & bioethics : journal of the institute of medical ethics & bioethics"] = "Med Etika Bioet",
  ["medico boehringer: overseas ed"] = "Med Bohringer (Overseas)",
  ["medico-chirurgical journal : or, quarterly register of medical and surgical science"] = "Med Chir J",
  ["medico-chirurgical transactions"] = "Med Chir Trans",
  ["medico-legal bulletin"] = "Med Leg Bull",
  ["medico-legal journal"] = "Med. Leg. J.",
  ["medico-surgical suggestions"] = "Med Surg Suggest",
  ["medico; europa ausg"] = "Medico (Eur)",
  ["medico; international ed"] = "Medico (Int)",
  ["medicolegal and bioethics"] = "Medicoleg Bioeth",
  ["medicolegal news"] = "Medicoleg News",
  ["medieval archaeology"] = "Mediev Archaeol",
  ["medieval archaeology. journal of the society for medieval archaeology"] = "MedievA",
  ["medieval spanish medical text series"] = "Mediev Span Med Text Ser",
  ["medievalia et humanistica"] = "Mediev Humanist",
  ["medievalia et humanistica: studies in medieval and renaissance society"] = "M&H",
  ["medika; informativni casopis farmaceutiskih poduzeca nrh"] = "Medika (Zagreb)",
  ["medika; informativni c̆asopis farmaceutiskih poduzeća nrh"] = "Medika (Zagreb)",
  ["medinfo. medinfo"] = "Medinfo",
  ["medioevo greco"] = "MEG",
  ["medioevo latino"] = "Medioevo Lat",
  ["medioevo romanzo"] = "Medioev Romanzo",
  ["medisch contact"] = "Med Contact (Bussum)",
  ["medisch maandblad"] = "Med Maandbl",
  ["medisch-farmaceutische mededelingen"] = "Med Farm Meded",
  ["mediscope; journal of medicine & surgery"] = "Mediscope",
  ["mediterranean agricultural sciences"] = "Mediterr Agric Sci",
  ["mediterranean archaeology"] = "MedA",
  ["mediterranean archaeology: australian and new zealand journal for the archaeology of the mediterranean world"] = "MedArch",
  ["mediterranean botany"] = "Mediterr. Bot.",
  ["mediterranean conference on control & automation : [proceedings]. ieee mediterranean conference on control & automation"] = "Mediterr Conf Control Automation",
  ["mediterranean geoscience reviews"] = "Mediterr. Geosci. Rev.",
  ["mediterranean historical review"] = "Mediterr Hist Rev",
  ["mediterranean historical review, ed. by the aranne school of history, tel aviv univ"] = "MHR",
  ["mediterranean journal of hematology and infectious diseases"] = "Mediterr J Hematol Infect Dis",
  ["mediterranean journal of mathematics"] = "Mediterr. J. Math.",
  ["mediterranean journal of nutrition and metabolism"] = "Med J Nutrition Metab",
  ["mediterranean journal of rheumatology"] = "Mediterr J Rheumatol",
  ["mediterranean journal of rheumatology - ellīnikī reumatologia"] = "Mediterr J Rheumatol Ell Reumatol",
  ["mediterranean journal of social sciences"] = "Mediterr J Soc Sci",
  ["mediterranean marine science"] = "Mediterr Mar Sci",
  ["mediterranee medicale"] = "Mediterr Med",
  ["mediterraneo antico. economie, società, culture"] = "MedAnt",
  ["meditsina truda i promyshlennaia ekologiia"] = "Med Tr Prom Ekol",
  ["meditsinskaia parazitologiia i parazitarnye bolezni"] = "Med Parazitol (Mosk)",
  ["meditsinskaia promyshlennost sssr"] = "Med. Prom. SSSR",
  ["meditsinskaia promyshlennost' sssr"] = "Med Prom SSSR",
  ["meditsinskaia promyshlennost’ sssr"] = "Med Prom SSSR",
  ["meditsinskaia radiologiia"] = "Med Radiol (Mosk)",
  ["meditsinskaia sestra"] = "Med. Sestra",
  ["meditsinskaia tekhnika"] = "Med Tekh",
  ["meditsinski letopisi"] = "Med Letop",
  ["meditsinskii zhurnal uzbekistana"] = "Med Zh Uzb",
  ["meditsinskiĭ rabotnik"] = "Med Rab",
  ["medit︠s︡inskai︠a︡ sestra"] = "Med Sestra",
  ["medit︠s︡inskiĭ akademicheskiĭ zhurnal"] = "Med Akad Z",
  ["medit︠s︡inskiĭ zhurnal"] = "Med Zhurnal",
  ["medium aevum"] = "Medium Aevum",
  ["medizin in geschichte und kultur"] = "Med Gesch Kult",
  ["medizin und ernährung"] = "Med Ernahr",
  ["medizin und gesellschaft"] = "Med Ges",
  ["medizin und sport"] = "Med Sport (Berl)",
  ["medizin, gesellschaft, und geschichte : jahrbuch des instituts fur geschichte der medizin der robert bosch stiftung"] = "Med Ges Gesch",
  ["medizin, gesellschaft, und geschichte : jahrbuch des instituts für geschichte der medizin der robert bosch stiftung"] = "Med Ges Gesch",
  ["medizin, gesellschaft, und geschichte. beiheft : jahrbuch des instituts fur geschichte der medizin der robert bosch stiftung"] = "Med Ges Gesch Beih",
  ["medizin, gesellschaft, und geschichte. beiheft : jahrbuch des instituts für geschichte der medizin der robert bosch stiftung"] = "Med Ges Gesch Beih",
  ["medizin-historisches journal"] = "MHJ",
  ["medizinhistorische materialien"] = "Medhist Mater",
  ["medizinhistorisches journal"] = "Medizinhist J",
  ["medizinische dokumentation"] = "Med Dok",
  ["medizinische genetik"] = "Med. Gen.",
  ["medizinische genetik : mitteilungsblatt des berufsverbandes medizinische genetik e.v"] = "Med Genet",
  ["medizinische informatik und statistik"] = "Med. Inform. Statist.",
  ["medizinische klinik"] = "Med Klin",
  ["medizinische klinik (klinik-ausg.)"] = "Med Klin Klin",
  ["medizinische klinik (munich, germany : 1983)"] = "Med Klin (Munich)",
  ["medizinische klinik (munich, germany : 1983). supplement"] = "Med Klin Suppl",
  ["medizinische klinik (praxis-ausg.)"] = "Med Klin Prax",
  ["medizinische klinik, intensivmedizin und notfallmedizin"] = "Med Klin Intensivmed Notfmed",
  ["medizinische klinik. supplement"] = "Med. Klin. Suppl.",
  ["medizinische laboratorium"] = "Med. Lab. (Stuttg.)",
  ["medizinische monatsschrift"] = "Med Monatsschr",
  ["medizinische monatsschrift fur pharmazeuten"] = "Med. Monatsschr. Pharm.",
  ["medizinische monatsschrift für pharmazeuten"] = "Med Monatsschr Pharm",
  ["medizinische monatsschrift. beihefte"] = "Med Monatsschr Beih",
  ["medizinische nachrichten aus den vereinigten staaten"] = "Med Nachr Ver Staaten",
  ["medizinische novitaten"] = "Med Nov",
  ["medizinische novitäten"] = "Med Nov",
  ["medizinische technik"] = "Med Tech (Stuttg)",
  ["medizinische welt"] = "Med. Welt",
  ["medizinischer bild-dienst"] = "Med Bild Dienst",
  ["medizinrecht : medr"] = "Medizinrecht",
  ["medlab magazine"] = "MedLab Mag",
  ["medrxiv : the preprint server for health sciences"] = "medRxiv",
  ["medscape journal of medicine"] = "Medscape J Med",
  ["medscape women's health"] = "Medscape Womens Health",
  ["medscape women’s health"] = "Medscape Womens Health",
  ["medsurg nursing"] = "Medsurg Nurs.",
  ["medsurg nursing : official journal of the academy of medical-surgical nurses"] = "Medsurg Nurs",
  ["medtube science"] = "MEDtube Sci",
  ["medychne pravo"] = "Med Pravo",
  ["medychnyi zhurnal"] = "Med Zhurnal",
  ["medychnyĭ z︠h︡urnal"] = "Med Zhurnal",
  ["medycyna doswiadczalna"] = "Med. Dosw.",
  ["medycyna doswiadczalna i mikrobiologia"] = "Med Dosw Mikrobiol",
  ["medycyna doświadczalna"] = "Med Dosw",
  ["medycyna doświadczalna i mikrobiologia"] = "Med Dosw Mikrobiol",
  ["medycyna nowozytna : studia nad historia medycyny"] = "Med Nowozytna",
  ["medycyna nowozytna : studia nad historia medycyny / polska akademia nauk, instytut historii nauki"] = "Med Nowozytna",
  ["medycyna pracy"] = "Med Pr",
  ["medycyna weterynaryjna"] = "Med Weter",
  ["medycyna wieku rozwojowego"] = "Med Wieku Rozwoj",
  ["meikai daigaku shigaku zasshi"] = "Meikai Daigaku Shigaku Zasshi",
  ["meikai daigaku shigaku zasshi = the journal of meikai university school of dentistry"] = "Meikai Daigaku Shigaku Zasshi",
  ["mekhanika kompositnykh materialov (zinatne)"] = "Mekh. Kompoz. Mater. (Zinatne)",
  ["melanges de la casa de velazquez"] = "Melanges Casa Velazquez",
  ["melanges de l’ecole francaise de rome. moyen age, temps modernes"] = "Melanges Ec Fr Rome Moyen Age Temps Mod",
  ["melanoma management"] = "Melanoma Manag",
  ["melanoma research"] = "Melanoma Res",
  ["melbourne historical journal"] = "Melb Hist J",
  ["melbourne university law review"] = "Melb Univ Law Rev",
  ["meldinger fra norges landbrukshoegskole"] = "Meld. Nor. Landbrukshoegsk.",
  ["melita theologica : the review of the royal university students' theological association, malta"] = "Melita Theol",
  ["melita theologica : the review of the royal university students’ theological association, malta"] = "Melita Theol",
  ["mellen history of medicine"] = "Mellen Hist Med",
  ["melliand textilberichte"] = "Melliand Textilber.",
  ["melsheimer entomological series"] = "Meisheimer Entomol Ser",
  ["memai heiko igaku"] = "Memai Heiko Igaku",
  ["memai heikō igaku"] = "Memai Heiko Igaku",
  ["membrane & cell biology"] = "Membr Cell Biol",
  ["membrane and cell biology"] = "Membr. Cell Biol.",
  ["membrane and water treatment"] = "Membr. Water Treat.",
  ["membrane biochemistry"] = "Membr Biochem",
  ["membrane technology"] = "Membr. Technol.",
  ["membranes (basel, switzerland)"] = "Membranes (Basel, Switz.)",
  ["membranes and membrane technologies"] = "Membr. Membr. Technol.",
  ["membranes and membrane technology"] = "Membr. Membr. Technol.",
  ["memetic computing"] = "Memet Comput",
  ["memoir - american association of petroleum geologists"] = "Mem. Am. Assoc. Pet. Geol.",
  ["memoires de l academie royale de medecine de belgique"] = "Mem. Acad. R. Med. Belg.",
  ["memoires de la societe d’agriculture, commerce, sciences et arts du departement de la marne. societe d’agriculture, commerce, sciences et arts du departement de la marne, chalons-sur-marne"] = "Mem Soc Agric Commer Sci Arts Marne",
  ["memoires de la societe mathematique de france. nouvelle serie"] = "Mem. Soc. Math. Fr. (N.S.)",
  ["memoires de la societe royale des sciences de liege"] = "Mem. Soc. R. Sci. Liege",
  ["memoires de l’academie de chirurgie"] = "Mem. Acad. Chir. (Paris)",
  ["memoires de l’academie des sciences, inscriptions et belles-lettres de toulouse"] = "Mem Acad Sci Inscr B-Lett Toulouse",
  ["memoires de l’academie royale de medecine de belgique"] = "Mem Acad R Med Belg",
  ["memoires. academie de chirurgie (france)"] = "Mem Acad Chir (Paris)",
  ["memoirs and proceedings of the manchester literary & philosophical society. manchester literary and philosophical society"] = "Mem Proc Manch Lit Philos Soc",
  ["memoirs and proceedings, chemical society"] = "Mem. Proc. Chem. Soc.",
  ["memoirs of museum victoria"] = "Mem Mus Vic",
  ["memoirs of osaka institute of technology"] = "Mem. Osaka Inst. Tech.",
  ["memoirs of osaka kyoiku university"] = "Mem. Osaka Kyoiku Univ. III Natur. Sci. Appl. Sci.",
  ["memoirs of osaka kyoiku university. ser. iii. natural science and applied science"] = "Mem. Osaka Kyoiku Univ. Ser. III Nat. Sci. Appl. Sci.",
  ["memoirs of the american academy in rome"] = "MemAmAc",
  ["memoirs of the american mathematical society"] = "Mem. Amer. Math. Soc.",
  ["memoirs of the american philosophical society"] = "Mem. Amer. Philos. Soc.",
  ["memoirs of the college of education akita university"] = "Mem. College Ed. Akita Univ. Natur. Sci.",
  ["memoirs of the european mathematical society"] = "Mem. Eur. Math. Soc.",
  ["memoirs of the faculty of education"] = "Mem. Fac. Educ. Shiga Univ. Nat. Sci.",
  ["memoirs of the faculty of engineering and design kyoto institute of technology"] = "Mem. Fac. Eng. Design Kyoto Inst. Tech. Ser. Sci. Tech.",
  ["memoirs of the faculty of engineering hiroshima university"] = "Mem. Fac. Eng. Hiroshima Univ.",
  ["memoirs of the faculty of engineering miyazaki university"] = "Mem. Fac. Eng. Miyazaki Univ.",
  ["memoirs of the faculty of science and engineering shimane university"] = "Mem. Fac. Sci. Eng. Shimane Univ. Ser. B Math. Sci.",
  ["memoirs of the faculty of science, kochi university. series d, biology"] = "Mem Fac Sci Kochi Univ Ser D Biol",
  ["memoirs of the faculty of science, kyushu university, series c: chemistry"] = "Mem. Fac. Sci. Kyushu Univ., Ser. C",
  ["memoirs of the graduate school of science and engineering. shimane university. series b. mathematics"] = "Mem. Grad. Sch. Sci. Eng. Shimane Univ. Ser. B Math.",
  ["memoirs of the institute of science and engineering"] = "Mem. Inst. Sci. Eng. Ritsumeikan Univ.",
  ["memoirs of the institute of science and engineering. ritsumeikan university"] = "Mem. Inst. Sci. Engrg. Ritsumeikan Univ.",
  ["memoirs of the institute of scientific and industrial research, osaka university"] = "Mem. Insts. Sci. Ind. Res., Osaka Univ.",
  ["memoirs of the muroran institute of technology"] = "Mem. Muroran Inst. Technol.",
  ["memoirs of the national defense academy"] = "Mem. Nat. Defense Acad.",
  ["memoirs of the new york botanical garden"] = "Mem N Y Bot Gard",
  ["memoirs of the osaka institute of technology"] = "Mem. Osaka Inst. Tech. Ser. A",
  ["memoirs of the queensland museum"] = "Mem Queensl Mus",
  ["memoirs of the royal astronomical society"] = "Mem. R. Astron. Soc.",
  ["memoirs of the school of science and engineering"] = "Mem. School Sci. Eng. Waseda Univ.",
  ["memoirs on differential equations and mathematical physics"] = "Mem. Differ. Equ. Math. Phys.",
  ["memoranda societatis pro fauna et flora fennica"] = "Memo Soc Fauna Flora Fenn",
  ["memorandum report. naval medical research institute (u.s.)"] = "Memo Rep Nav Med Res Inst (US)",
  ["memorandum rm-. rand corporation"] = "Memo RM",
  ["memorandum, medical research council"] = "Memo. Med. Res. Counc.",
  ["memoria antiquitatis. acta musei petrodavensis. revista muzeului arheologic piatra neamţ"] = "MemAnt",
  ["memoria de el colegio nacional. colegio nacional (mexico)"] = "Mem Col Nac (Mex)",
  ["memoria storica. rivista del centro studi storici terni"] = "MemStor",
  ["memoria y sociedad : revista del departamento de historia y geografía"] = "Mem Soc",
  ["memoria. sociedad uruguaya de urología"] = "Mem Soc Urug Urol",
  ["memoria. universidad de barcelona, instituto de arqueología y prehistoria"] = "MemBarcelA",
  ["memorias da academia das ciencias de lisboa, classe de ciencias"] = "Mem Acad Cienc Lisb Cl Cienc",
  ["memorias de historia antigua"] = "MHA",
  ["memorias de historia antigua (universidad de oviedo)"] = "MemHistAnt",
  ["memorias de la real academia de ciencias exactas, físicas y naturales de madrid"] = "Mem. Real Acad. Cienc. Exact. Fís. Natur. Madrid",
  ["memorias de la real academia de ciencias y artes de barcelona"] = "Mem. Real Acad. Cienc. Artes Barcelona",
  ["memorias de la sociedad matematica mexicana"] = "Mem. Soc. Mat. Mex.",
  ["memorias do instituto butantan"] = "Mem. Inst. Butantan",
  ["memorias do instituto oswaldo cruz"] = "Mem Inst Oswaldo Cruz",
  ["memorias. congreso médico peninsular"] = "Memorias Congreso Med Penins",
  ["memorie del museo civico di storia naturale di verona. iia serie. monografie naturalistiche"] = "Mem Mus Civ Stor Nat Verona IIa Monogr Nat",
  ["memorie della classe di scienze morali e storiche dell’accademia dei lincei"] = "MAL",
  ["memorie dell’accademia delle scienze dell’istituto di bologna, cl. di sc. morali"] = "MAIB",
  ["memorie dell’accademia delle scienze di torino, classe di scienze morali, storiche e filologiche"] = "MAT",
  ["memorie dell’accademia di archeologia, lettere e belle arti di napoli"] = "MemNap",
  ["memorie dell’ist. lombardo, accademia di scienze e lettere, cl. di lett., sc. morali e storiche"] = "MIL",
  ["memorie storiche forogiuliesi"] = "MemStorFriuli",
  ["memories- materials, devices, circuits and systems"] = "Memories - Mater. Devices Circuits Syst.",
  ["memory & cognition"] = "Mem Cognit",
  ["memory (hove, england)"] = "Memory",
  ["memory and cognition"] = "Mem. Cognit.",
  ["memory studies"] = "Mem Stud",
  ["memórias da academia das ciências de lisboa, classe de ciências"] = "Mem Acad Cienc Lisb Cl Cienc",
  ["memórias do instituto butantan"] = "Mem Inst Butantan",
  ["memórias do instituto oswaldo cruz"] = "Mem Inst Oswaldo Cruz",
  ["memphis and mid-south medical journal"] = "Memphis Mid South Med J",
  ["memphis medical journal"] = "Memphis Med J",
  ["memphis state university law review"] = "Memphis State Univ Law Rev",
  ["memórias do instituto oswaldo cruz"] = "Mem. Inst. Oswaldo Cruz",
  ["men and masculinities"] = "Men Masc",
  ["men's reproductive health"] = "Men's Reprod Health",
  ["menara perkebunan"] = "Menara Perkeb",
  ["mendeleev communications"] = "Mendeleev Commun.",
  ["menemui matematik"] = "Menemui Mat.",
  ["menninger perspective"] = "Menninger Perspect",
  ["menopausal medicine : for clinicians who provide care for women"] = "Menopausal Med",
  ["menopause (new york, n.y.)"] = "Menopause",
  ["menopause international"] = "Menopause Int",
  ["menopause management"] = "Menopause Manag",
  ["mens agitat—colloq."] = "Mens Agit.—Colloquia",
  ["mens en onderneming"] = "Mens Ondernem",
  ["mens sana monographs"] = "Mens Sana Monogr",
  ["menschen – kulturen – traditionen. studien aus den forschungsclustern des deutschen archäologischen instituts"] = "MKT",
  ["mental and physical disability law reporter"] = "Ment Phys Disabil Law Rep",
  ["mental disability law reporter"] = "Ment Disabil Law Rep",
  ["mental disability law reporter / american bar association, commission on the mentally disabled"] = "Ment Disabil Law Rep",
  ["mental health"] = "Ment Health (Montr)",
  ["mental health & prevention"] = "Ment Health Prev",
  ["mental health (london)"] = "Ment Health (Lond)",
  ["mental health and addiction research"] = "Ment Health Addict Res",
  ["mental health and physical activity"] = "Ment Health Phys Act",
  ["mental health and society"] = "Ment Health Soc",
  ["mental health and substance use : dual diagnosis"] = "Ment Health Subst Use",
  ["mental health bulletin"] = "Ment Health Bull (Danville)",
  ["mental health care"] = "Ment Health Care",
  ["mental health in family medicine"] = "Ment Health Fam Med",
  ["mental health nursing"] = "Ment. Health Nurs.",
  ["mental health nursing : journal of the psychiatric nurses association"] = "Ment Health Nurs",
  ["mental health nursing : journal of the psychiatric nurses association / pna"] = "Ment Health Nurs",
  ["mental health review (brighton, england)"] = "Ment Health Rev (Brighton)",
  ["mental health services research"] = "Ment Health Serv Res",
  ["mental health special interest section quarterly"] = "Mental Health Spec Interest Sect Q",
  ["mental health statistical note"] = "Ment Health Stat Note",
  ["mental health today (brighton, england)"] = "Ment Health Today",
  ["mental health, religion & culture"] = "Ment Health Relig Cult",
  ["mental hospitals"] = "Ment Hosp",
  ["mental hygiene"] = "Ment Hyg (Lond)",
  ["mental hygiene survey"] = "Ment Hyg Surv",
  ["mental illness"] = "Ment Illn",
  ["mental retardation"] = "Ment Retard",
  ["mental retardation abstracts"] = "Ment Retard Abstr",
  ["mental retardation and developmental disabilities research reviews"] = "Ment Retard Dev Disabil Res Rev",
  ["mental retardation and the law"] = "Ment Retard Law",
  ["mental welfare"] = "Ment Welf",
  ["mentalities = mentalités"] = "Mentalities",
  ["mentoring & tutoring"] = "Mentor Tutoring",
  ["men’s reproductive health"] = "Men’s Reprod Health",
  ["mercer dental society newsletter"] = "Mercer Dent Soc Newsl",
  ["mercer law review"] = "Mercer Law Rev",
  ["meridian: crossing aesthetics"] = "Meridian Crossing Aesthet.",
  ["merkblatt für den forstpraktiker eidg. anstalt für das forstliche versuchswesen"] = "Merkbl. Forstprakt. Eidgenöss. Anst. forstl. Vers.wes.",
  ["merkblatt für die praxis"] = "Merkbl. Prax.",
  ["meroitica. schriften zur altsudanesischen geschichte und archäologie"] = "Meroitica",
  ["merrill-palmer quarterly"] = "Merrill Palmer Q",
  ["merrill-palmer quarterly (wayne state university. press)"] = "Merrill Palmer Q (Wayne State Univ Press)",
  ["merrill-palmer quarterly of behavior and development"] = "Merrill Palmer Q Behav Dev",
  ["mese sanitario"] = "Mese Sanit",
  ["mesoamerica (antigua, guatemala)"] = "Mesoamerica (Antigua Guatem)",
  ["mesoamerican herpetology"] = "Mesoam Herpetol",
  ["mesoamérica (antigua, guatemala)"] = "Mesoamerica (Antigua Guatem)",
  ["mesoporous biomaterials"] = "Mesoporous Biomater",
  ["mesopotamia. rivista di archeologia"] = "Mesopotamia",
  ["messana. rassegna di studi filologici, linguistici e storici"] = "Messana",
  ["messenger (los angeles, calif. : print)"] = "Messenger (Los Angel)",
  ["mesure et évaluation en éducation"] = "Mes Eval Educ",
  ["meta gene"] = "Meta Gene",
  ["metabarcoding and metagenomics"] = "Metabarcoding Metagenom",
  ["metabolic and pediatric ophthalmology"] = "Metab Pediatr Ophthalmol",
  ["metabolic bone disease & related research"] = "Metab Bone Dis Relat Res",
  ["metabolic bone disease and related research"] = "Metab. Bone Dis. Relat. Res.",
  ["metabolic brain disease"] = "Metab Brain Dis",
  ["metabolic engineering"] = "Metab Eng",
  ["metabolic engineering communications"] = "Metab Eng Commun",
  ["metabolic ophthalmology"] = "Metab Ophthalmol",
  ["metabolic ophthalmology, pediatric and systemic"] = "Metab Ophthalmol",
  ["metabolic syndrome and related disorders"] = "Metab Syndr Relat Disord",
  ["metabolic, pediatric and systemic ophthalmology"] = "Metab. Pediatr. Syst. Ophthalmol.",
  ["metabolic, pediatric, and systemic ophthalmology"] = "Metab Pediatr Syst Ophthalmol",
  ["metabolic, pediatric, and systemic ophthalmology (new york, n.y. : 1985)"] = "Metab Pediatr Syst Ophthalmol (1985)",
  ["metabolism and nutrition in oncology"] = "Metab. Nutr. Oncol.",
  ["metabolism clincal and experimental"] = "Metab. Clin. Exp.",
  ["metabolism clinical and experimental"] = "Metab. Clin. Exp.",
  ["metabolism open"] = "Metabol Open",
  ["metabolism, clinical and experimental"] = "Metab., Clin. Exp.",
  ["metabolism: clinical and experimental"] = "Metabolism",
  ["metabolomics : official journal of the metabolomic society"] = "Metabolomics",
  ["metabolomics : open access"] = "Metabolomics (Los Angel)",
  ["metacognition and learning"] = "Metacogn Learn",
  ["metagenomics (cairo, egypt)"] = "Metagenomics (Cairo)",
  ["metal finishing"] = "Met. Finish.",
  ["metal ions in biological systems"] = "Met Ions Biol Syst",
  ["metal ions in life sciences"] = "Met Ions Life Sci",
  ["metal powder report"] = "Met. Powder Rep.",
  ["metal science and heat treatment"] = "Met. Sci. Heat Treat.",
  ["metal-based drugs"] = "Met Based Drugs",
  ["metall (berlin)"] = "Metall (Berlin)",
  ["metallgesellschaft periodic review"] = "Metallges. Period. Rev.",
  ["metallofizika (kiev)"] = "Metallofizika (Akad. Nauk Ukr. SSR, Otd. Fiz. Astron.)",
  ["metallofizika i noveishie tekhnologii"] = "Metallofiz. Noveishie Tekhnol.",
  ["metallography, microstructure, and analysis"] = "Metall. Microstruct. Anal.",
  ["metallomics : integrated biometal science"] = "Metallomics",
  ["metalloorganicheskaya khimiya"] = "Metalloorg. Khim.",
  ["metalloproteinases in medicine"] = "Metalloproteinases Med",
  ["metallovedenie i termicheskaya obrabotka metallov"] = "Metalloved. Term. Obrab. Met.",
  ["metallurgia and metal forming"] = "Metall. Met. Forming",
  ["metallurgia italiana"] = "Metall. Ital.",
  ["metallurgical and materials engineering"] = "Metall. Mater. Eng.",
  ["metallurgical and materials transactions a: physical metallurgy and materials science"] = "Metall. Mater. Trans. A",
  ["metallurgical and materials transactions b-process metallurgy and materials processing science"] = "Metall. Mater. Trans. B Process Metall. Mater. Process. Sci.",
  ["metallurgical and materials transactions b: process metallurgy and materials processing science"] = "Metall. Mater. Trans. B",
  ["metallurgical and materials transactions e, materials for energy systems"] = "Metall. Mater. Trans. E",
  ["metallurgical and materials transactions. a. physical metallurgy and materials science"] = "Metall Mater Trans A Phys Metall Mater Sci",
  ["metallurgical research & technology"] = "Metall. Res. Technol.",
  ["metallurgical transactions a"] = "Metall. Trans. A",
  ["metallurgical transactions b"] = "Metall. Trans. B",
  ["metalogicon. international journal of pure and applied logic, linguistics and philosophy. new series"] = "Metalogicon (N.S.)",
  ["metals (basel, switzerland)"] = "Metals (Basel, Switz.)",
  ["metals and materials"] = "Met. Mater.",
  ["metals and materials international"] = "Met. Mater. Int.",
  ["metaphor and symbol"] = "Metaphor Symb",
  ["meteoritics & planetary science"] = "Meteorit. Planet. Sci.",
  ["meteorological applications"] = "Meteorol. Appl.",
  ["meteorological magazine"] = "Meteorol. Mag.",
  ["meteorological monographs"] = "Meteorol. Monogr.",
  ["meteorological technology international"] = "Meteorol Technol Int",
  ["meteorologische rundschau"] = "Meteor. Rundsch.",
  ["meteorologische zeitschrift"] = "Meteorol. Z.",
  ["meteorologiya i gidrologiya"] = "Meteor. Gidrol.",
  ["meteorology and atmospheric physics"] = "Meteorol. Atmos. Phys.",
  ["methoden der psychologie"] = "Methoden Psych.",
  ["methoden der regelungs- und automatisierungstechnik"] = "Methoden Regel. Autom.",
  ["methoden der regelungstechnik"] = "Method. Regelungstech.",
  ["methoden und verfahren der mathematischen physik"] = "Methoden Verfahren Math. Phys.",
  ["methoden, daten, analysen"] = "Methoden Daten Anal",
  ["methodist debakey cardiovascular journal"] = "Methodist Debakey Cardiovasc J",
  ["methodist history"] = "Methodist Hist",
  ["methodological innovations"] = "Method Innov",
  ["methodology : european journal of research methods for the behavioral & social sciences"] = "Methodology (Gott)",
  ["methodology and computing in applied probability"] = "Methodol. Comput. Appl. Probab.",
  ["methodology and science"] = "Methodol Sci",
  ["methodology in the social sciences"] = "Methodol. Soc. Sci.",
  ["methodology of music research"] = "Methodol. Music Res.",
  ["methodos series"] = "Methodos Ser.",
  ["methods (amsterdam, netherlands)"] = "Methods (Amsterdam, Neth.)",
  ["methods (oxford, united kingdom)"] = "Methods (Oxford, U. K.)",
  ["methods (san diego, ca, united states)"] = "Methods (San Diego, CA, U. S.)",
  ["methods (san diego, calif.)"] = "Methods",
  ["methods and achievements in experimental pathology"] = "Methods Achiev Exp Pathol",
  ["methods and applications in fluorescence"] = "Methods Appl Fluoresc",
  ["methods and applications of analysis"] = "Methods Appl. Anal.",
  ["methods and findings in experimental and clinical pharmacology"] = "Methods Find Exp Clin Pharmacol",
  ["methods and protocols"] = "Methods Protoc",
  ["methods in cell biology"] = "Methods Cell Biol",
  ["methods in cell science"] = "Methods Cell Sci.",
  ["methods in cell science : an official journal of the society for in vitro biology"] = "Methods Cell Sci",
  ["methods in ecology and evolution"] = "Methods Ecol Evol",
  ["methods in enzymology"] = "Methods Enzymol",
  ["methods in geochemistry and geophysics"] = "Methods Geochem. Geophys.",
  ["methods in medical research"] = "Methods Med Res",
  ["methods in molecular and cellular biology"] = "Methods Mol Cell Biol",
  ["methods in molecular biology"] = "Methods Mol. Biol.",
  ["methods in molecular biology (clifton, n.j.)"] = "Methods Mol Biol",
  ["methods in molecular medicine"] = "Methods Mol Med",
  ["methods in oceanography"] = "Methods Oceanogr.",
  ["methods in statistical ecology"] = "Methods Stat. Ecol.",
  ["methods of biochemical analysis"] = "Methods Biochem Anal",
  ["methods of experimental physics"] = "Methods Experiment. Phys.",
  ["methods of functional analysis and topology"] = "Methods Funct. Anal. Topology",
  ["methods of information in medicine"] = "Methods Inf Med",
  ["methods of information in medicine. supplement"] = "Methods Inf Med Suppl",
  ["methods report (rti press)"] = "Methods Rep RTI Press",
  ["metody diskretnogo analiza"] = "Metody Diskret. Anal.",
  ["metody optimalnykh reshenii"] = "Metody Optim. Reshenii",
  ["metody vychislenii"] = "Metody Vychisl.",
  ["metrika. international journal for theoretical and applied statistics"] = "Metrika",
  ["metropolitan museum journal"] = "MMJ",
  ["metropolitan museum studies"] = "MetrMusSt",
  ["metropolitan universities"] = "Metrop Univ",
  ["mexican journal of biotechnology"] = "Mex. J. Biotechnol.",
  ["mexican journal of scientific research"] = "Mex J Sci Res",
  ["mexican studies = estudios mexicanos"] = "Mex Stud",
  ["mezzogiorno d'europa"] = "Mezzog Eur",
  ["mezzogiorno d’europa"] = "Mezzog Eur",
  ["mètode science studies journal"] = "Metode Sci Stud J",
  ["médecine & chirurgie digestives"] = "Med Chir Dig",
  ["médecine aéronautique et spatiale"] = "Med Aeronaut Spat",
  ["médecine d'afrique noire"] = "Med Afr Noire",
  ["médecine d'egypte"] = "Med Egypte",
  ["médecine dans le monde"] = "Med Monde",
  ["médecine et audio vision"] = "Med Audio Vis",
  ["médecine et hygiène"] = "Med Hyg (Geneve)",
  ["médecine et laboratoire"] = "Med Lab",
  ["médecine et maladies infectieuses"] = "Med Mal Infect",
  ["médecine et santé tropicales"] = "Med Sante Trop",
  ["médecine interne"] = "Med Interne",
  ["médecine légale et dommage corporel"] = "Med Leg Dommage Corpor",
  ["médecine sciences : m/s"] = "Med Sci (Paris)",
  ["médecine thérapeutique médecine de la reproduction, gynécologie, endocrinologie"] = "Med Ther Med Reprod Gynecol Endocrinol",
  ["médecine tropicale : revue du corps de santé colonial"] = "Med Trop (Mars)",
  ["médica (matanzas, cuba)"] = "Medica",
  ["méditerranée médicale"] = "Mediterr Med",
  ["mélanges de l'ecole française de rome. moyen âge, temps modernes"] = "Melanges Ec Fr Rome Moyen Age Temps Mod",
  ["mélanges de la casa de velázquez"] = "Melanges Casa Velazquez",
  ["mémoires de biospéologie"] = "Mem Biospeol",
  ["mémoires de l'académie des sciences, inscriptions et belles-lettres de toulouse"] = "Mem Acad Sci Inscr B-Lett Toulouse",
  ["mémoires de l'académie royale de médecine de belgique"] = "Mem Acad R Med Belg",
  ["mémoires de la société d'agriculture, commerce, sciences et arts du département de la marne. société d'agriculture, commerce, sciences et arts du département de la marne, châlons-sur-marne"] = "Mem Soc Agric Commer Sci Arts Marne",
  ["mémoires du muséum national d'histoire naturelle (1993)"] = "Mem Mus Natl Hist Nat (1993)",
  ["mémoires. académie de chirurgie (france)"] = "Mem Acad Chir (Paris)",
  ["méxico farmacético"] = "Mex Farm",
  ["mĕsíc̆ník ciba"] = "Mesic Ciba",
  ["mgh news"] = "MGH News",
  ["mgh news / montreal general hospital"] = "MGH News",
  ["mgma connexion"] = "MGMA Connex",
  ["mgma connexion / medical group management association"] = "MGMA Connex",
  ["mh; mental hygiene"] = "MH",
  ["mhq : the quarterly journal of military history"] = "MHQ (NY NY)",
  ["mian hua xue bao = acta gossypii sinica"] = "Mian Hua Xue Bao",
  ["mian yi xue za zhi"] = "Journal of immunology|Mian Yi Xue Za Zhi",
  ["mian yi xue za zhi = journal of immunology"] = "Mian Yi Xue Za Zhi",
  ["michel serres and material futures"] = "Michel Serres Mater. Futur.",
  ["michie's annotated code of the public general laws of maryland. maryland"] = "Michies Annot Code Public Gen Laws Md Md",
  ["michie's west virginia code, annotated. west virginia"] = "Michies W Va Code Annot W Va",
  ["michie’s west virginia code, annotated / prepared by the editorial staff of the publisher. west virginia"] = "Michies W Va Code Annot W Va",
  ["michigan academician"] = "Mich Acad",
  ["michigan appeals reports : cases decided in the michigan court of appeals. michigan. court of appeals"] = "Mich Appeals Rep Mich Court Appeals",
  ["michigan compiled laws annotated. michigan"] = "Mich Compil Laws Annot Mich",
  ["michigan health & hospitals"] = "Mich Health Hosp",
  ["michigan health and hospitals"] = "Mich. Health Hosp.",
  ["michigan history"] = "Mich Hist",
  ["michigan hospitals"] = "Mich Hosp",
  ["michigan jewish history"] = "Mich Jew Hist",
  ["michigan journal of community service learning"] = "Mich J Community Serv Learn",
  ["michigan journal of gender & law"] = "Mich J Gend Law",
  ["michigan law review"] = "Mich Law Rev",
  ["michigan mathematical journal"] = "Mich. Math. J.",
  ["michigan medicine"] = "Mich Med",
  ["michigan nurse"] = "Mich. Nurse",
  ["michigan public health"] = "Mich Public Health",
  ["michigan quarterly review"] = "Mich Q Rev",
  ["michigan state international law review"] = "Mich State Int Law Rev",
  ["michigan statutes annotated. michigan"] = "Mich Statut Annot Mich",
  ["micologia aplicada international"] = "Micol Aplicada Int",
  ["micologia e vegetazione mediterranea"] = "Micol Veg Mediterr",
  ["micologia italiana"] = "Micol Ital",
  ["micologia toscana"] = "Micologia Toscana",
  ["micro & nano letters"] = "Micro Nano Lett",
  ["micro and nano engineering"] = "Micro Nano Eng.",
  ["micro and nano letters"] = "Micro Nano Lett.",
  ["micro and nano systems letters"] = "Micro Nano Syst. Lett.",
  ["micro and nanostructures"] = "Micro Nanostruct.",
  ["micro and nanosystems"] = "Micro Nanosyst.",
  ["microarrays (basel, switzerland)"] = "Microarrays (Basel)",
  ["microbe (washington, d.c.)"] = "Microbe Wash DC",
  ["microbes and environments"] = "Microbes Environ",
  ["microbes and infection"] = "Microbes Infect",
  ["microbes and infection / institut pasteur"] = "Microbes Infect",
  ["microbial & comparative genomics"] = "Microb Comp Genomics",
  ["microbial and comparative genomics"] = "Microb. Comp. Genomics",
  ["microbial bioactives"] = "Microb Bioact",
  ["microbial biotechnology"] = "Microb Biotechnol",
  ["microbial cell (graz, austria)"] = "Microb Cell",
  ["microbial cell factories"] = "Microb Cell Fact",
  ["microbial drug resistance"] = "Microb. Drug Resist.",
  ["microbial drug resistance (larchmont, n.y.)"] = "Microb Drug Resist",
  ["microbial ecology"] = "Microb Ecol",
  ["microbial ecology in health and disease"] = "Microb Ecol Health Dis",
  ["microbial genomics"] = "Microb Genom",
  ["microbial informatics and experimentation"] = "Microb Inform Exp",
  ["microbial pathogenesis"] = "Microb Pathog",
  ["microbial physiology"] = "Microb Physiol",
  ["microbial releases"] = "Microb. Releases",
  ["microbial releases : viruses, bacteria, fungi"] = "Microb Releases",
  ["microbial risk analysis"] = "Microb Risk Anal",
  ["microbiologia (madrid, spain)"] = "Microbiologia",
  ["microbiologia espanola"] = "Microbiol. Esp.",
  ["microbiologia, parazitologia, epidemiologia"] = "Microbiol Parazitol Epidemiol (Bucur)",
  ["microbiological research"] = "Microbiol Res",
  ["microbiological reviews"] = "Microbiol Rev",
  ["microbiological sciences"] = "Microbiol Sci",
  ["microbiología (madrid, spain)"] = "Microbiologia",
  ["microbiología española"] = "Microbiol Esp",
  ["microbiology (london, united kingdom)"] = "Microbiology (London, U. K.)",
  ["microbiology (moscow, russian federation)"] = "Microbiology (Moscow, Russ. Fed.)",
  ["microbiology (moscow, russian federation)(translation of mikrobiologiya)"] = "Microbiology (Moscow, Russ. Fed.)",
  ["microbiology (new york, ny, united states)"] = "Microbiology (N. Y., NY, U. S.)",
  ["microbiology (reading, england)"] = "Microbiology (Reading)",
  ["microbiology (reading, united kingdom)"] = "Microbiology (Reading, U. K.)",
  ["microbiology (road town, virgin islands (british))"] = "Microbiology (Road Town, V. I. (Br.))",
  ["microbiology and biotechnology letters"] = "Microbiol. Biotechnol. Lett.",
  ["microbiology and culture collections"] = "Microbiol Cult Collect",
  ["microbiology and immunology"] = "Microbiol Immunol",
  ["microbiology and molecular biology reviews"] = "Microbiol. Mol. Biol. Rev.",
  ["microbiology and molecular biology reviews : mmbr"] = "Microbiol Mol Biol Rev",
  ["microbiology australia"] = "Microbiol Aust",
  ["microbiology education"] = "Microbiol Educ",
  ["microbiology indonesia"] = "Microbiol Indones",
  ["microbiology insights"] = "Microbiol Insights",
  ["microbiology open"] = "Microbiol. Open",
  ["microbiology outlooks"] = "Microbiol. Outlooks",
  ["microbiology research"] = "Microbiol Res (Pavia)",
  ["microbiology research journal international"] = "Microbiol Res J Int",
  ["microbiology resource announcements"] = "Microbiol Resour Announc",
  ["microbiology spectrum"] = "Microbiol Spectr",
  ["microchemical journal"] = "Microchem. J.",
  ["microchemical journal : devoted to the application of microtechniques in all branches of science"] = "Microchem J",
  ["microchemical journal, devoted to the application of microtechniques in all branches of science"] = "Microchem J",
  ["microchimica acta"] = "Microchim. Acta",
  ["microcirculation (new york, n.y. : 1994)"] = "Microcirculation",
  ["microcirculation, endothelium, and lymphatics"] = "Microcirc Endothelium Lymphatics",
  ["microelectronic engineering"] = "Microelectron. Eng.",
  ["microelectronics and reliability"] = "Microelectron Reliab",
  ["microelectronics international"] = "Microelectron. Int.",
  ["microelectronics journal"] = "Microelectron. J.",
  ["microelectronics reliability"] = "Microelectron. Reliab.",
  ["microfluidics and nanofluidics"] = "Microfluid. Nanofluid.",
  ["microgram journal"] = "Microgram J.",
  ["microgravity quarterly : mgq"] = "Microgravity Q",
  ["microgravity science and technology"] = "Microgravity Sci Technol",
  ["micron (oxford, england : 1993)"] = "Micron",
  ["micron and microscopica acta"] = "Micron Microsc. Acta",
  ["microphysiological systems"] = "Microphysiol Syst",
  ["microplastics and nanoplastics"] = "Microplast. Nanoplast.",
  ["microporous and mesoporous materials"] = "Microporous Mesoporous Mater.",
  ["microporous and mesoporous materials : the official journal of the international zeolite association"] = "Microporous Mesoporous Mater",
  ["microporous materials"] = "Microporous Mater.",
  ["microprocessors and microsystems"] = "Microprocess. Microsyst.",
  ["micropublication biology"] = "MicroPubl Biol",
  ["microrna (shāriqah, united arab emirates)"] = "Microrna",
  ["microrna diagnostics and therapeutics"] = "Microrna Diagn Ther",
  ["microscale thermophysical engineering"] = "Microscale Thermophys. Eng.",
  ["microscope (carshalton beeches (surrey))"] = "Microscope",
  ["microscopia electronica y biologia celular"] = "Microsc. Electron. Biol. Celular",
  ["microscopia electronica y biologia celular : organo oficial de las sociedades latinoamericana de microscopia electronica e iberoamericana de biologia celular"] = "Microsc Electron Biol Celular",
  ["microscopic journal and structural record"] = "Microsc. J. Struct. Rec.",
  ["microscopica acta"] = "Microsc Acta",
  ["microscopica acta. supplement"] = "Microsc Acta Suppl",
  ["microscopía electrónica y biología celular : organo oficial de las sociedades latinoamericana de microscopía electrónica e iberoamericana de biología celular"] = "Microsc Electron Biol Celular",
  ["microscopy (oxford, england)"] = "Microscopy (Oxf)",
  ["microscopy : the journal of the quekett microscopical club"] = "Microscopy",
  ["microscopy and analysis"] = "Microsc. Anal.",
  ["microscopy and analysis (americas ed.)"] = "Microsc Anal (Am Ed)",
  ["microscopy and microanalysis"] = "Microsc. Microanal.",
  ["microscopy and microanalysis : the official journal of microscopy society of america, microbeam analysis society, microscopical society of canada"] = "Microsc Microanal",
  ["microscopy research and technique"] = "Microsc Res Tech",
  ["microscopy today"] = "Micros Today",
  ["microsystem technologies"] = "Microsyst. Technol.",
  ["microsystem technologies : sensors, actuators, systems integration"] = "Microsyst Technol",
  ["microsystems & nanoengineering"] = "Microsyst Nanoeng",
  ["microvascular research"] = "Microvasc Res",
  ["microwave and optical technology letters"] = "Microwave Opt. Technol. Lett.",
  ["microwave journal"] = "Microw J (Int Ed)",
  ["mid-america; an historical review"] = "Mid Am",
  ["mid-american review of sociology"] = "Mid Am Rev Sociol",
  ["mid-western educational researcher (akron, ohio : 1991)"] = "Midwest Educ Res (Akron)",
  ["midas journal"] = "Midas J",
  ["midcontinental journal of archaeology"] = "Midcont. J. Archaeol.",
  ["midcontinental journal of archaeology, mcja"] = "MidCont J Archaeol",
  ["middle east african journal of ophthalmology"] = "Middle East Afr J Ophthalmol",
  ["middle east business and economic review"] = "Middle East Bus. Econ. Rev.",
  ["middle east critique"] = "Middle East Crit",
  ["middle east dentistry & oral health"] = "Middle East Dent Oral Health",
  ["middle east dentistry and oral health"] = "Middle East Dent. Oral Health",
  ["middle east development journal"] = "Middle East Dev J",
  ["middle east economic papers"] = "Middle East Econ. Pap.",
  ["middle east fertility society journal"] = "Middle East Fertil Soc J",
  ["middle east health"] = "Middle East Health",
  ["middle east journal"] = "Middle East J.",
  ["middle east journal of anaesthesiology"] = "Middle East J Anaesthesiol",
  ["middle east journal of anesthesiology"] = "Middle East J Anesthesiol",
  ["middle east journal of cancer"] = "Middle East J Cancer",
  ["middle east journal of digestive diseases"] = "Middle East J Dig Dis",
  ["middle east journal of scientific research"] = "Middle East J Sci Res",
  ["middle east law and governance"] = "Middle East Law Gov",
  ["middle east report (new york, n.y.)"] = "Middle East Rep",
  ["middle east studies association bulletin"] = "Middle East Stud Assoc Bull",
  ["middle east technical university studies in development"] = "METU Stud. Devel.",
  ["middle eastern studies"] = "Middle East Stud",
  ["midland history"] = "Midl Hist",
  ["midstream; a monthly jewish review"] = "Midstream (N Y)",
  ["midwest alliance in nursing journal"] = "Midwest Alliance Nurs J",
  ["midwest alliance in nursing journal / main"] = "Midwest Alliance Nurs J",
  ["midwest medical ethics : a publication of the midwest bioethics center"] = "Midwest Med Ethics",
  ["midwest studies in philosophy"] = "Midwest Stud Philos",
  ["midwestern dentist"] = "Midwest Dent",
  ["midwestern druggist"] = "Midwest Drug",
  ["midwife and health visitor"] = "Midwife Health Visit",
  ["midwife, health visitor & community nurse"] = "Midwife Health Visit Community Nurse",
  ["midwife, health visitor and community nurse"] = "Midwife. Health Visit. Community Nurse",
  ["midwifery today and childbirth education"] = "Midwifery Today Childbirth Educ",
  ["midwifery today with international midwife"] = "Midwifery Today Int Midwife",
  ["midwives : official journal of the royal college of midwives"] = "Midwives (1995)",
  ["midwives chronicle"] = "Midwives Chron",
  ["mie medical journal"] = "Mie Med J",
  ["mieux comprende"] = "Mieux Compr.",
  ["migracijske teme"] = "Migr Teme",
  ["migration and development"] = "Migr Dev",
  ["migration letters : an international journal of migration studies"] = "Migrat Lett",
  ["migration news"] = "Migr News",
  ["migration studies"] = "Migr Stud",
  ["migration today (staten island, new york, n.y.)"] = "Migr Today",
  ["migration world magazine"] = "Migr World Mag",
  ["mikologiia i fitopatologiia"] = "Mikol Fitopatol",
  ["mikrobiologicheskii zhurnal"] = "Mikrobiol. Zh.",
  ["mikrobiologicheskiĭ zhurnal"] = "Mikrobiol Zh",
  ["mikrobiolohichnyi zhurnal"] = "Mikrobiol. Zh.",
  ["mikrobiolohichnyi zhurnal (kiev, ukraine : 1993)"] = "Mikrobiol Z",
  ["mikrobiolohichnyĭ zhurnal"] = "Mikrobiol Zh",
  ["mikrobiolohichnyĭ zhurnal (kiev, ukraine : 1993)"] = "Mikrobiol Z",
  ["mikrobiyoloji bulteni"] = "Mikrobiyol. Bul.",
  ["mikrobiyoloji bülteni"] = "Mikrobiyol Bul",
  ["mikrobnye biotekhnologii"] = "Mikrobn Biotekhnologii",
  ["mikrochimica acta"] = "Mikrochim Acta",
  ["mikrochimica et ichnoanalytica acta"] = "Mikrochim Ichnoanal Acta",
  ["milan journal of mathematics"] = "Milan J. Math.",
  ["milbank memorial fund quarterly"] = "Milbank Mem. Fund Q.",
  ["milbank memorial fund quarterly. health and society"] = "Milbank Mem. Fund Q. Health Soc.",
  ["milbank quarterly"] = "Milbank Q.",
  ["milchwissenschaft. milk science international"] = "Milchwissenschaft",
  ["milesische forschungen"] = "MilForsch",
  ["milet. ergebnisse der ausgrabungen und untersuchungen seit dem jahr 1899"] = "Milet",
  ["milieux (le creusot, france)"] = "Milieux",
  ["military affairs : journal of the american military institute"] = "Mil Aff",
  ["military behavioral health"] = "Mil Behav Health",
  ["military collector & historian"] = "Milit Collect Hist",
  ["military history of texas and the southwest"] = "Milit Hist Tex Southwest",
  ["military law review"] = "Mil Law Rev",
  ["military medical research"] = "Mil Med Res",
  ["military medicine"] = "Mil Med",
  ["military operations research"] = "Mil. Oper. Res.",
  ["military psychology : the official journal of the division of military psychology, american psychological association"] = "Mil Psychol",
  ["military review"] = "Mil Rev",
  ["military surgeon"] = "Mil Surg",
  ["militärgeschichtliche zeitschrift"] = "Militargesch Z",
  ["militärgeschichtliche mitteilungen"] = "MGM",
  ["milwaukee history"] = "Milw Hist",
  ["milwaukee medical times"] = "Milwaukee Med Times",
  ["mimeo series (university of north carolina (system). institute of statistics)"] = "Mimeo Ser",
  ["mims magazine weekly : the journal of quality patient care"] = "MIMS Mag",
  ["min guo dang an"] = "Min Guo Dang An",
  ["mind & brain : the journal of psychiatry"] = "Mind Brain",
  ["mind & language"] = "Mind Lang",
  ["mind and mental health magazine"] = "Mind Ment Health Mag",
  ["mind association occasional series"] = "Mind Assoc. Occas. Ser.",
  ["mind, brain and education : the official journal of the international mind, brain, and education society"] = "Mind Brain Educ",
  ["mind, culture, and activity"] = "Mind Cult Act",
  ["mind, meaning and metaphysics"] = "Mind Mean. Metaphys.",
  ["mind. a quarterly review of philosophy"] = "Mind",
  ["mind; a quarterly review of psychology and philosophy"] = "Mind",
  ["mindfulness & compassion"] = "Mindfulness Compassion",
  ["minds and machines"] = "Minds Mach (Dordr)",
  ["mine water and the environment"] = "Mine Water Environ.",
  ["mineral and electrolyte metabolism"] = "Miner Electrolyte Metab",
  ["mineral economics"] = "Miner. Econ.",
  ["mineral processing and extractive metallurgy"] = "Miner. Process. Extr. Metall.",
  ["mineral processing and extractive metallurgy review"] = "Miner. Process. Extr. Metall. Rev.",
  ["mineral resources engineering"] = "Min. Resour. Eng.",
  ["mineralia slovaca"] = "Miner. Slovaca",
  ["mineralium deposita"] = "Miner. Deposita",
  ["mineralogical magazine"] = "Mineral. Mag.",
  ["mineralogical record"] = "Mineral. Rec.",
  ["mineralogicheskii zhurnal"] = "Mineral. Zh.",
  ["mineralogy and petrology"] = "Mineral. Petrol.",
  ["minerals (basel, switzerland)"] = "Minerals (Basel)",
  ["minerals and metallurgical processing"] = "Miner. Metall. Process",
  ["minerals engineering"] = "Miner. Eng.",
  ["minerva (arlington, va.)"] = "Minerva (Arlingt Va)",
  ["minerva anestesiologica"] = "Minerva Anestesiol",
  ["minerva biotechnology and biomolecular research"] = "Minerva Biotechnol. Biomol. Res.",
  ["minerva biotecnologica"] = "Minerva Biotecnol",
  ["minerva cardioangiologica"] = "Minerva Cardioangiol",
  ["minerva cardioangiologica europea"] = "Minerva Cardioangiol Eur",
  ["minerva chirurgica"] = "Minerva Chir",
  ["minerva dermatologica"] = "Minerva Dermatol",
  ["minerva dietologica"] = "Minerva Dietol",
  ["minerva dietologica e gastroenterologica"] = "Minerva Dietol Gastroenterol",
  ["minerva ecologica idroclimatologica e fisiconucleare"] = "Minerva Ecol Idroclimatol Fisiconucl",
  ["minerva ecologica idroclimatologica fisicosanitaria"] = "Minerva Ecol Idroclimatol Fis Sanit",
  ["minerva ecologica, idroclimatologica e fisiconucleare"] = "Minerva Ecol. Idroclimatol. Fisiconucl.",
  ["minerva ecologica, idroclimatologica, fisicosanitaria"] = "Minerva Ecol. Idroclimatol. Fis. Sanit.",
  ["minerva endocrinologica"] = "Minerva Endocrinol",
  ["minerva farmaceutica"] = "Minerva Farm",
  ["minerva fisiconucleare"] = "Minerva Fisiconucl",
  ["minerva fisioterapica e radiobiologica"] = "Minerva Fisioter Radiobiol",
  ["minerva gastroenterologica"] = "Minerva Gastroenterol",
  ["minerva gastroenterologica e dietologica"] = "Minerva Gastroenterol Dietol",
  ["minerva ginecologica"] = "Minerva Ginecol",
  ["minerva medica"] = "Minerva Med",
  ["minerva medicolegale"] = "Minerva Medicoleg.",
  ["minerva medicolegale; archivio di antropologia criminale, psichiatria, e medicina legale"] = "Minerva Medicoleg",
  ["minerva nefrologica"] = "Minerva Nefrol",
  ["minerva neurochirurgica"] = "Minerva Neurochir",
  ["minerva nipiologica"] = "Minerva Nipiol",
  ["minerva nucleare"] = "Minerva Nucl",
  ["minerva oftalmologica"] = "Minerva Oftalmol",
  ["minerva ortognatodontica"] = "Minerva Ortognatod",
  ["minerva ortopedica"] = "Minerva Ortop",
  ["minerva ortopedica e traumatologica : organo ufficiale della società piemontese-ligure-lombarda di ortopedia e traumatologia"] = "Minerva Ortop Traumatol",
  ["minerva otorinolaringologica"] = "Minerva Otorinolaringol",
  ["minerva pediatrica"] = "Minerva Pediatr",
  ["minerva pneumologica"] = "Minerva Pneumol",
  ["minerva psichiatrica"] = "Minerva Psichiatr",
  ["minerva radiologica"] = "Minerva Radiol",
  ["minerva radiologica, fisioterapica e radiobiologica"] = "Minerva Radiol Fisioter Radiobiol",
  ["minerva stomatologica"] = "Minerva Stomatol",
  ["minerva urologica"] = "Minerva Urol",
  ["minerva urologica e nefrologica"] = "Minerva Urol. Nefrol.",
  ["minerva urologica e nefrologica = the italian journal of urology and nephrology"] = "Minerva Urol Nefrol",
  ["minerva. revista de filología clásica"] = "Minerva",
  ["minerva: revista de filología clásica"] = "Minerva(vall)",
  ["minerva: the international review of ancient art and archaeology"] = "Minerva(lond)",
  ["ming studies"] = "Ming Stud",
  ["mini - reviews in medicinal chemistry"] = "Mini Rev. Med. Chem.",
  ["mini reviews in medicinal chemistry"] = "Mini Rev Med Chem",
  ["mini-reviews in medicinal chemistry"] = "Mini-Rev. Med. Chem.",
  ["mini-reviews in organic chemistry"] = "Mini-Rev. Org. Chem.",
  ["minia journal of agricultural research and development"] = "Minia J Agric Res Dev",
  ["minima epigraphica et papyrologica. taccuini della cattedra e del laboratorio di epigrafia e papirologia giuridica dell’università degli studi di catanzaro »magna graecia«"] = "MinEpigrP",
  ["minimally invasive neurosurgery"] = "Minim. Invasive Neurosurg.",
  ["minimally invasive neurosurgery : min"] = "Minim Invasive Neurosurg",
  ["minimally invasive surgery"] = "Minim Invasive Surg",
  ["minimally invasive surgical nursing"] = "Minim Invasive Surg Nurs",
  ["minimally invasive therapy & allied technologies"] = "Minimally Invasive Ther. Allied Technol.",
  ["minimally invasive therapy & allied technologies : mitat : official journal of the society for minimally invasive therapy"] = "Minim Invasive Ther Allied Technol",
  ["minimax theory and its applications"] = "Minimax Theory Appl.",
  ["mining and scientific press"] = "Min. Sci. Press",
  ["mining engineering"] = "Min. Eng.",
  ["mining informational and analytical bulletin"] = "Min. Inf. Anal. Bull.",
  ["mining intelligence and knowledge exploration : third international conference, mike 2015, hyderabad, india, december 9-11, 2015. proceeedings. mike (conference) (3rd : 2015 : hyderabad, india)"] = "Min Intell Knowl Explor (2015)",
  ["mining journal"] = "Min. J.",
  ["mining magazine"] = "Min. Mag.",
  ["mining of mineral deposits"] = "Min. Miner. Deposits",
  ["mining science and technology"] = "Min. Sci. Technol.",
  ["mining technology"] = "Min. Technol.",
  ["mining, metallurgy & exploration"] = "Min Metall Explor",
  ["ministry of education, science and culture, japan. report of grant-in-aid for scientific research on priority areas"] = "Minist. Ed. Sci. Cult. Japan Rep. Grant Aid Sci. Res. Prior. Areas",
  ["minneapolis district dental journal"] = "Minneap. Dist. Dent. J.",
  ["minnesota history"] = "Minn Hist",
  ["minnesota journal of law, science & technology"] = "Minn J Law Sci Technol",
  ["minnesota law review"] = "Minn Law Rev",
  ["minnesota medicine"] = "Minn Med",
  ["minnesota nursing accent"] = "Minn Nurs Accent",
  ["minnesota pharmacist"] = "Minn Pharm",
  ["minnesota statutes annotated. minnesota"] = "Minn Statut Annot Minn",
  ["minnesota studies in the philosophy of science"] = "Minn. Stud. Philos. Sci.",
  ["minnesota symposia on child psychology (series)"] = "Minn Symp Child Psychol Ser",
  ["minnesota welfare"] = "Minn Welfare",
  ["minnetesol journal (2014)"] = "MinneTESOL J (2014)",
  ["minority nurse"] = "Minor Nurse",
  ["minority nurse newsletter"] = "Minor Nurse Newsl",
  ["minos. revista de filología egea"] = "Minos",
  ["minskiy radiotekhnicheskiy institut"] = "Avtomat. i Vychisl. Tekhn. (Minsk)",
  ["minufiya journal of agricultural research"] = "Minufiya J Agric Res",
  ["mir nauki, kulʻtury, obrazovaniâ"] = "Mir Nauki Kult Obraz",
  ["mira (san francisco, calif.)"] = "MIRA",
  ["mircen journal of applied microbiology and biotechnology"] = "MIRCEN J Appl Microbiol Biotechnol",
  ["mires and peat"] = "Mires Peat",
  ["mirovaia ekonomika i mezhdunarodnye otnosheniia / akademiia nauk sssr, institut mirovoi ekonomiki i mezhd’unarodnykh otnoshenii"] = "Mirovaia Ekon Mezdunar Otnosheiia",
  ["mirovai︠a︡ ėkonomika i mezhdunarodnye otnoshenii︠a︡"] = "Mirovaia Ekon Mezdunar Otnosheiia",
  ["mirror / georgetown university school of dentistry"] = "Georgetown Univ Sch Dent Mirror",
  ["mis quarterly : management information systems"] = "MIS Q",
  ["misaengmul hakhoe chi"] = "Misainmurhag Hoiji",
  ["misaengmul hakhoe chi (seoul, korea : 1997)"] = "Misaengmul Hakhoe Chi",
  ["misaengmul kwa sanop"] = "The microorganisms and industry|Misaengmul Gwa San-eob",
  ["misaengmul kwa sanŏp = the microorganisms and industry"] = "Misaengmul Gwa San Eob",
  ["miscelanea de estudios arabes y hebraicos"] = "Misc Estud Arabes Hebraicos",
  ["miscelanea matematica"] = "Miscelanea Mat.",
  ["miscelánea de estudios árabes y hebraicos"] = "Misc Estud Arabes Hebraicos",
  ["miscellanea di studi di letteratura christiana antica"] = "MiscCrAnt",
  ["miscellanea di studi storici"] = "MiscStStor",
  ["miscellanea greca e romana: studi pubblicati dall’ist. italiano per la storia antica"] = "MGR",
  ["miscellaneous papers (centre for entomological studies (ankara, turkey))"] = "Misc Pap",
  ["miscellaneous publications (university of michigan. museum of zoology)"] = "Misc Publ Mus Zool Univ Mich",
  ["miscelánea matemática"] = "Miscelánea Mat.",
  ["mises au point de l'accoucheur et du pédiatre"] = "Mises Point Accouch Pediatre",
  ["mises au point de l’accoucheur et du pediatre"] = "Mises Point Accouch Pediatre",
  ["miskolc mathematical notes. a publication of the university of miskolc"] = "Miskolc Math. Notes",
  ["missionalia hispanica"] = "Mission Hisp",
  ["missionalia hispánica"] = "Mission Hisp",
  ["mississippi college law review. mississippi college. school of law"] = "Miss Coll Law Rev",
  ["mississippi dental association journal"] = "Miss Dent Assoc J",
  ["mississippi law journal"] = "Miss Law J",
  ["mississippi rn"] = "Miss RN",
  ["mississippi valley journal of business and economics"] = "Miss. Val. J. Bus. Econ.",
  ["mississippi valley medical journal"] = "Miss Valley Med J",
  ["mississippi valley medical journal (quincy, ill)"] = "Miss Valley Med J",
  ["missouri dental journal"] = "Mo. Dent. J.",
  ["missouri dental journal (jefferson city, mo.)"] = "Mo Dent J",
  ["missouri historical review"] = "Mo Hist Rev",
  ["missouri journal of mathematical sciences"] = "Missouri J. Math. Sci.",
  ["missouri law review"] = "Miss Law Rev",
  ["missouri medicine"] = "Mo Med",
  ["missouri nurse"] = "Mo. Nurse",
  ["mit lincoln laboratory series"] = "MIT Linc. Lab. Ser.",
  ["mit press essential knowledge series"] = "MIT Press Essent. Knowl. Ser.",
  ["mit press series in artif. intell."] = "MIT Press Ser. Artif. Intell.",
  ["mit press series in logic programming"] = "MIT Press Ser. Logic Program.",
  ["mit press series on economic learning and social evolution"] = "MIT Press Ser. Econ. Learn. Soc. Evol.",
  ["mit sloan management review"] = "MIT Sloan Manag Rev",
  ["mit's technology review"] = "MITS Technol Rev",
  ["mitigation and adaptation strategies for global change"] = "Mitig Adapt Strateg Glob Chang",
  ["mitochondrial communications"] = "Mitochondrial Commun.",
  ["mitochondrial dna"] = "Mitochondrial DNA",
  ["mitochondrial dna part a dna mapping, sequencing, and analysis"] = "Mitochondrial DNA Part A",
  ["mitochondrial dna part b resources"] = "Mitochondrial DNA Part B",
  ["mitochondrial dna. part a, dna mapping, sequencing, and analysis"] = "Mitochondrial DNA A DNA Mapp Seq Anal",
  ["mitochondrial dna. part b, resources"] = "Mitochondrial DNA B Resour",
  ["mitteilungen - gesellschaft deutscher chemiker - fachgruppe geschichte der chemie"] = "Mitt. Gesell. Dtscher. Chem. Fachgruppe Gesch. Chem.",
  ["mitteilungen aus dem gebiete der lebensmittel-untersuchung un hygiene = travaux de chimie alimentaire et d'hygiène"] = "Mitt Geb Lebensmittelunters Hyg",
  ["mitteilungen aus dem gebiete der lebensmitteluntersuchung und hygiene"] = "Mitt. Geb. Lebensmittelunters. Hyg.",
  ["mitteilungen aus dem heinrich-schliemann-museum ankershagen"] = "MSchliemann",
  ["mitteilungen aus dem institut für allgemeine botanik hamburg. institut für allgemeine botanik hamburg"] = "Mitt Inst Allg Bot Hamb",
  ["mitteilungen aus dem mathematischen seminar giessen"] = "Mitt. Math. Sem. Giessen",
  ["mitteilungen aus dem museum für naturkunde in berlin. deutsche entomologische zeitschrift"] = "Mitt Mus Naturkunde Berl Dtsch Entomol Z",
  ["mitteilungen aus dem museum für naturkunde in berlin. fossil record"] = "Mitt Mus Nat Berl Foss Rec",
  ["mitteilungen aus dem museum für naturkunde in berlin. zoologische reihe"] = "Mitt Mus Naturkunde Berl Zoolog Reihe",
  ["mitteilungen aus der arbeitsmarkt- und berufsforschung"] = "Mitt Arbeitsmarkt Berufsforsch",
  ["mitteilungen aus der biologischen bundesanstalt für land- und forstwirtschaft, berlin-dahlem"] = "Mitt. Biol. Bundesanst. Land- Forstwirtsch., Berl.-Dahl.",
  ["mitteilungen aus der norddeutschen naturschutzakademie (nna)"] = "Mitt. Norddtsch. Nat.schutzakad.",
  ["mitteilungen aus der staatsforstverwaltung bayerns"] = "Mitt. Staatsforstverwalt. Bayerns",
  ["mitteilungen aus lebensmitteluntersuchung und hygiene = travaux de chimie alimentaire et d'hygiène"] = "Mitt Lebensmittelunters Hyg",
  ["mitteilungen der aargauischen naturforschenden gesellschaft"] = "Mitt. Aargau. nat.forsch. Ges.",
  ["mitteilungen der abteilung für botanik am landesmuseum joanneum in graz"] = "Mitt. Abt. Bot. Landesmus. Joanneum Graz",
  ["mitteilungen der abteilung für forstliche biometrie der universität freiburg"] = "Mitt. Abt. Forstl. Biom. Univ. Freibg.",
  ["mitteilungen der anthropologischen gesellschaft in wien"] = "MAnthrWien",
  ["mitteilungen der arbeitsgemeinschaft geobotanik in schleswig-holstein und hamburg"] = "Mitt. Arb.gem. Geobot. Schlesw.-Holst. Hambg.",
  ["mitteilungen der archäologischen gesellschaft graz"] = "MAGesGraz",
  ["mitteilungen der archäologischen gesellschaft steiermark"] = "MAGS",
  ["mitteilungen der bundesforschungsanstalt für forst- und holzwirtschaft"] = "Mitt. Bundesforsch.anst. Forst-  Holzwirtsch.",
  ["mitteilungen der deutschen bodenkundlichen gesellschaft"] = "Mitt. Dtsch. Bodenkdl. Ges.",
  ["mitteilungen der deutschen dendrologischen gesellschaft"] = "Mitt. Dtsch. dendrol. Ges.",
  ["mitteilungen der deutschen gesellschaft für allgemeine und angewandte entomologie"] = "Mitt Dtsch Ges Allg Angew Entomol",
  ["mitteilungen der deutschen gesellschaft für allgemeine und angewandte entomologie"] = "Mitt. Dtsch. Ges. allg. angew. Entomol.",
  ["mitteilungen der deutschen malakozoologischen gesellschaft"] = "Mitt Dtsch Malakozool Ges",
  ["mitteilungen der deutschen mathematiker-vereinigung"] = "Mitt. Dtsch. Math.-Ver.",
  ["mitteilungen der deutschen orient-gesellschaft zu berlin"] = "MDOG",
  ["mitteilungen der deutschen pharmazeutischen gesellschaft und der pharmazeutischen gesellschaft der ddr"] = "Mitt. Dtsch. Pharm. Ges. Pharm. Ges. DDR",
  ["mitteilungen der eidg. forschungsanstalt für wald, schnee und landschaft"] = "Mitt. Eidgenöss. Forsch.anst. Wald Schnee Landsch.",
  ["mitteilungen der entomologischen gesellschaft basel"] = "Mitt. Entomol. Ges. Basel",
  ["mitteilungen der floristisch-soziologischen arbeitsgemeinschaft"] = "Mitt. Florist.-soziol. Arb.gem.",
  ["mitteilungen der forstlichen bundes-versuchsanstalt wien"] = "Mitt. Forstl. Bundes-Vers.anst. Wien",
  ["mitteilungen der forstlichen versuchs- und forschungsanstalt baden-württemberg"] = "Mitt. Forstl. Vers.- Forsch.anst. Baden-Württ.",
  ["mitteilungen der geographisch-ethnographischen gesellschaft zürich"] = "Mitt. Geogr.-Ethnogr. Ges. Zür.",
  ["mitteilungen der gesellschaft fur salzburger landeskunde"] = "Mitt Ges Salzburger Landeskd",
  ["mitteilungen der gesellschaft für salzburger landeskunde"] = "Mitt Ges Salzburger Landeskd",
  ["mitteilungen der gesellschaft für angewandte mathematik und mechanik"] = "GAMM Mitt. Ges. Angew. Math. Mech.",
  ["mitteilungen der gesellschaft für bibliothekswesen und dokumentation des landbaues"] = "Mitt. Ges. Bibl.wes. Dok. Landbaues",
  ["mitteilungen der hessischen landesforstverwaltung"] = "Mitt. Hess. Landesforstverwalt.",
  ["mitteilungen der internationalen bodenkundlichen gesellschaft"] = "Mitt. Int. bodenkdl. Ges.",
  ["mitteilungen der mathematischen gesellschaft in hamburg"] = "Mitt. Math. Ges. Hamburg",
  ["mitteilungen der münchner entomologischen gesellschaft. münchner entomologische gesellschaft"] = "Mitt Munch Entomol Ges",
  ["mitteilungen der naturforschenden gesellschaft des kantons glarus"] = "Mitt. Nat.forsch. Ges. Kanton Glarus",
  ["mitteilungen der naturforschenden gesellschaft des kantons solothurn"] = "Mitt. Nat.forsch. Ges. Kanton Solothurn",
  ["mitteilungen der naturforschenden gesellschaft in bern"] = "Mitt. Nat.forsch. Ges. Bern",
  ["mitteilungen der naturforschenden gesellschaft in bern. naturforschende gesellschaft in bern"] = "Mitt Naturforsch Ges Bern",
  ["mitteilungen der naturforschenden gesellschaft luzern"] = "Mitt. Nat.forsch. Ges. Luzern",
  ["mitteilungen der naturforschenden gesellschaft schaffhausen"] = "Mitt. Nat.forsch. Ges. Schaffhausen",
  ["mitteilungen der naturwissenschaftlichen gesellschaft thun"] = "Mitt. Nat.wiss. Ges. Thun",
  ["mitteilungen der naturwissenschaftlichen gesellschaft winterthur"] = "Mitt. Nat.wiss. Ges. Winterthur",
  ["mitteilungen der niedersächsischen forstlichen versuchsanstalt"] = "Mitt. Niedersächs. forstl. Vers.anst.",
  ["mitteilungen der osterreichischen gesellschaft fur tropenmedizin und parasitologie"] = "Mitt Osterr Ges Tropenmed Parasitol",
  ["mitteilungen der osterreichischen sanitatsverwaltung"] = "Mitt Osterr Sanitatsverwalt",
  ["mitteilungen der österreichischen geographischen gesellschaft. österreichische geographische gesellschaft"] = "Mitt Osterr Geogr Ges",
  ["mitteilungen der österreichischen gesellschaft für tropenmedizin und parasitologie"] = "Mitt Osterr Ges Tropmed Parasitol",
  ["mitteilungen der österreichischen sanitätsverwaltung"] = "Mitt Osterr Sanitatsverwalt",
  ["mitteilungen der pollichia"] = "Mitt Pollichia",
  ["mitteilungen der prähistorischen kommission der österreichischen akademie der wissenschaften"] = "MPraehistKomWien",
  ["mitteilungen der schweizerischen entomologischen gesellschaft"] = "Mitt. Schweiz. Entomol. Ges.",
  ["mitteilungen der schweizerischen entomologischen gesellschaft = bulletin de la société entomologique suisse. schweizerische entomologische gesellschaft"] = "Mitt Schweiz Entomol Ges",
  ["mitteilungen der thurgauischen naturforschenden gesellschaft"] = "Mitt. Thurgau. Nat.forsch. Ges.",
  ["mitteilungen der versuchsanstalt für wasserbau, hydrologie und glaziologie an der eidg. technischen hochschule zürich"] = "Mitt. Vers.anst. Wasserbau Hydrol. Glaziol. Eidgenöss. Tech. Hochsch. Zür.",
  ["mitteilungen der österreichischen numismatischen gesellschaft"] = "MOeNumGes",
  ["mitteilungen des arbeitskreises wald und wasser"] = "Mitt. Arb.kr. Wald Wasser",
  ["mitteilungen des archäologischen instituts der ungarischen akademie der wissenschaften"] = "MAInstUngAk",
  ["mitteilungen des badischen landesvereins für naturkunde und naturschutz"] = "Mitt. Bad. Landesver. Nat.kd. Nat.schutz",
  ["mitteilungen des deutschen archaologischen instituts. deutsches archaologisches institut. abteilung kairo"] = "Mitt Dtsch Archaeol Inst Abt Kairo",
  ["mitteilungen des deutschen archäologischen instituts. deutsches archäologisches institut. abteilung kairo"] = "Mitt Dtsch Archaeol Inst Abt Kairo",
  ["mitteilungen des deutschen archäologen-verbandes e.v."] = "MDAVerb",
  ["mitteilungen des deutschen archäologischen instituts"] = "MdI",
  ["mitteilungen des deutschen archäologischen instituts (abt. istambul)"] = "MDAI(I)",
  ["mitteilungen des deutschen archäologischen instituts (abt. madrid)"] = "MDAI(M)",
  ["mitteilungen des deutschen archäologischen instituts, abteilung kairo"] = "MDAIK",
  ["mitteilungen des deutschen archäologischen instituts, athenische abteilung"] = "AM",
  ["mitteilungen des deutschen archäologischen instituts, römische abteilung"] = "RM",
  ["mitteilungen des deutschen archäologischen instituts (athen. abt.)"] = "MDAI(A)",
  ["mitteilungen des deutschen archäologischen instituts (röm. abt.)"] = "MDAI(R)",
  ["mitteilungen des eidg. institutes für schnee- und lawinenforschung"] = "Mitt. Eidgenöss. Inst. Schnee- Lawinenforsch.",
  ["mitteilungen des historischen vereins des kantons schwyz"] = "Mitt. Hist. Ver. Kanton Schwyz",
  ["mitteilungen des instituts für orientforschung. deutsche akademie der wissenschaften zu berlin. institut für orientforschung"] = "Mitt Inst Orientforsch",
  ["mitteilungen des instituts für orientforschung"] = "MIO",
  ["mitteilungen des instituts für österreichische geschichtsforschung"] = "MIöG",
  ["mitteilungen des kremser stadtarchivs. krems an der donau (austria). stadtarchiv"] = "Mitt Krems Stadtarch",
  ["mitteilungen des kunsthistorischen institutes in florenz"] = "MKuHistFlorenz",
  ["mitteilungen des kuratoriums für waldarbeit und forsttechnik"] = "Mitt. Kurat. Waldarb. Forsttech.",
  ["mitteilungen des vereins für forstliche standortskunde und forstpflanzenzüchtung"] = "Mitt. Ver. forstl. Standortskd. Forstpflanzenzücht.",
  ["mitteilungen eidg. anstalt für das forstliche versuchswesen"] = "Mitt. Eidgenöss. Anst. forstl. Vers.wes.",
  ["mitteilungen internationale vereinigung fur theoretische und angewandte limnologie"] = "Mitt. Int. Ver. Theor. Angew. Limnol.",
  ["mitteilungen klosterneuburg : rebe und wein, obstbau und früchteverwertung"] = "Mitt Klosterneuburg Rebe Wein Obstb Fruchteverwert",
  ["mitteilungen und forschungsbeitrage der cusanus-gesellschaft"] = "Mitt Forschungsbeitr Cusanus Ges",
  ["mitteilungen und forschungsbeiträge der cusanus-gesellschaft"] = "Mitt Forschungsbeitr Cusanus Ges",
  ["mitteilungen zur christlichen archäologie"] = "MiChA",
  ["mitteilungen zur frühchristlichen archäologie in österreich"] = "MFruehChrOe",
  ["mitteilungen zur kulturkunde"] = "MKul",
  ["mitteilungen zur spätantiken archäologie und byzantinischen kunstgeschichte"] = "MSpaetAByz",
  ["mitteilungen. deutsche pharmazeutische gesellschaft"] = "Mitt Dtsch Pharm Ges Pharm Ges DDR",
  ["mitteilungen. leichtweiss-institut für wasserbau der technischen universität braunschweig"] = "MInstWasser",
  ["mitteilungen. schweizerische unfallversicherungsanstalt. medizinische abteilung"] = "Mitt Med Abt Schweiz Unfallversicherunst",
  ["mitteilungsblatt - chemische gesellschaft der deutschen demokratischen republik"] = "Mitteilungsbl. - Chem. Ges. D. D. R.",
  ["mitteilungsblatt der schweizerischen gesellschaft für ur- und frühgeschichte"] = "MSchwUrFruehGesch",
  ["mitteilungsblatt des regionalen rechenzentrums erlangen"] = "Mitt. Region. Rechenzentrums Erlangen",
  ["mitteilungsblatt für vor- und frühgeschichte"] = "MBlVFruehGesch",
  ["mitteilungsblatt. osterreichische gesellschaft fur statistik und informatik"] = "Mittbl Osterr Ges Stat Inform",
  ["mitteilungsblatt. österreichische gesellschaft für statistik und informatik"] = "Mittbl Osterr Ges Stat Inform",
  ["mitteldeutsche familienkunde"] = "Mitteldtsch Famkd",
  ["mittellateinisches jahrbuch"] = "Mittellat Jahrb",
  ["mit’s technology review"] = "MITS Technol Rev",
  ["mix-hs'12 : proceedings of the 2nd international workshop on managing interoperability and complexity in health systems october 29, 2012, maui, hawaii, usa. international workshop on managing interoperability and complexity in health systems (2nd : 2012 : maui, hawaii)"] = "MIXHS 12 (2012)",
  ["miyagi national college of technology"] = "Res. Rep. Miyagi Nat. College Tech.",
  ["mizunami-shi kaseki hakubutsukan kenkyū hōkoku. bulletin of the mizunami fossil museum. mizunami-shi kaseki hakubutsukan"] = "Mizunamishi Kaseki Hakubutsukan Kenkyu Hokoku",
  ["mìkrobìologìâ ì bìotehnologìâ"] = "Mikrobiol Biotehnol",
  ["mla directory of periodicals"] = "MLA Dir Period",
  ["mlm [reports]. u.s. atomic energy commission"] = "MLM Rep",
  ["mlm reports"] = "MLM Rep.",
  ["mln bulletin"] = "MLN Bull",
  ["mlo: medical laboratory observer"] = "MLO Med Lab Obs",
  ["mlo; medical laboratory observer"] = "MLO. Med. Lab. Obs.",
  ["mlq. mathematical logic quarterly"] = "MLQ Math. Log. Q.",
  ["mmg. medizin, mensch, gesellschaft"] = "Mmg",
  ["mmhealth'16 : proceedings of the 2016 acm workshop on multimedia for personal health and health care : october 16, 2016, amsterdam, the netherlands. acm workshop on multimedia for personal health and health care (1st : 2016 : amsterdam, netherlands)"] = "MMHealth16 (2016)",
  ["mmhealth'17 : proceedings of the 2nd international workshop on multimedia for personal health and health care : october 23, 2017, mountain view, ca, usa. acm workshop on multimedia for personal health and health care (2nd : 2017 : mountain view, calif.)"] = "MMHealth17 (2017)",
  ["mmw fortschritte der medizin"] = "MMW Fortschr Med",
  ["mmw, munchener medizinische wochenschrift"] = "MMW Munch Med Wochenschr",
  ["mmw, münchener medizinische wochenschrift"] = "MMW Munch Med Wochenschr",
  ["mmw; munchener medizinische wochenschrift"] = "MMW Munch. Med. Wochenschr.",
  ["mmwr cdc surveillance summaries"] = "MMWR CDC Surveill. Summ.",
  ["mmwr supplements"] = "MMWR Suppl",
  ["mmwr surveillance summaries"] = "MMWR Surveill. Summ.",
  ["mmwr. cdc surveillance summaries : morbidity and mortality weekly report. cdc surveillance summaries"] = "MMWR CDC Surveill Summ",
  ["mmwr. cdc surveillance summaries : morbidity and mortality weekly report. cdc surveillance summaries / centers for disease control"] = "MMWR CDC Surveill Summ",
  ["mmwr. morbidity and mortality weekly report"] = "MMWR Morb Mortal Wkly Rep",
  ["mmwr. recommendations and reports : morbidity and mortality weekly report. recommendations and reports"] = "MMWR Recomm Rep",
  ["mmwr. recommendations and reports : morbidity and mortality weekly report. recommendations and reports / centers for disease control"] = "MMWR Recomm Rep",
  ["mmwr. surveillance summaries : morbidity and mortality weekly report. surveillance summaries / cdc"] = "MMWR Surveill Summ",
  ["mmwr; morbidity and mortality weekly report"] = "MMWR Morb. Mortal. Wkly. Rep.",
  ["mna accent"] = "MNA Accent",
  ["mna accent / minnesota nurses association"] = "MNA Accent",
  ["mnemosyne. a journal of classical studies"] = "Mnemosyne",
  ["mnemosyne. bibliotheca classica batava. supplementum"] = "Mnemosyne Bibl. Class. Batava Suppl.",
  ["mnemosyne: bibliotheca classica batava"] = "Mnemosyne",
  ["mni open research"] = "MNI Open Res",
  ["mobile dna"] = "Mob DNA",
  ["mobile genetic elements"] = "Mob Genet Elements",
  ["mobile information systems"] = "Mobile Inf. Syst.",
  ["mobile media & communication"] = "Mob Media Commun",
  ["mobile networks and applications"] = "Mobile Networks Appl.",
  ["moct-most: economic policy in transitional economies"] = "MOCT-MOST: Econ. Pol. Transitional Economies",
  ["model assisted statistics and applications"] = "Model Assisted Stat. Appl.",
  ["model assisted statistics and applications : an international journal"] = "Model Assist Stat Appl",
  ["modeling and optimization in science and technologies"] = "Model. Optim. Sci. Technol.",
  ["modeling and simulation in science, engineering and technology"] = "Model. Simul. Sci. Eng. Technol.",
  ["modeling and simulation in science, engineering, & technology"] = "Model. Simul. Sci. Eng. Technol.",
  ["modeling dynamic systems"] = "Model. Dyn. Syst.",
  ["modeling earth systems and environment"] = "Model Earth Syst Environ",
  ["modeling, identification and control"] = "MIC—Model. Identif. Control",
  ["modelirovanie i analiz informatsionnykh sistem. modeling and analysis of information systems"] = "Model. Anal. Inf. Sist.",
  ["modelling and simulation in engineering"] = "Modell. Simul. Eng.",
  ["modelling and simulation in materials science and engineering"] = "Model. Simul. Mater. Sci. Eng.",
  ["modelling measurement & control c"] = "Modell. Meas. Control C",
  ["modern & contemporary france"] = "Mod Contemp Fr",
  ["modern acoustics and signal processing"] = "Modern Acoust. Signal Process.",
  ["modern age (chicago, ill.)"] = "Mod Age",
  ["modern agriculture"] = "Mod. Agric.",
  ["modern applied science"] = "Mod Appl Sci",
  ["modern approaches in geophysics"] = "Modern Approaches Geophys.",
  ["modern asian studies"] = "Mod Asian Stud",
  ["modern birkhauser classics"] = "Mod. Birkhauser Class.",
  ["modern cell biology"] = "Mod Cell Biol",
  ["modern china"] = "Mod China",
  ["modern clinical medicine research"] = "Mod Clin Med Res",
  ["modern concepts of cardiovascular disease"] = "Mod Concepts Cardiovasc Dis",
  ["modern dental practice"] = "Mod Dent Pract",
  ["modern dentistry"] = "Mod Dent (N Y N Y 1931)",
  ["modern drug discovery"] = "Mod. Drug Discovery",
  ["modern electronic materials"] = "Mod. Electron. Mater.",
  ["modern fiction studies"] = "Mod Fiction Stud",
  ["modern french identities"] = "Mod. French Identities",
  ["modern genetics"] = "Mod Genet",
  ["modern greek studies yearbook"] = "Mod Greek Stud Yearb",
  ["modern healthcare"] = "Mod Healthc",
  ["modern healthcare (short term care)"] = "Mod. Healthc. (Short Term Care)",
  ["modern healthcare. [short-term care ed.]"] = "Mod Healthc (Short Term Care)",
  ["modern hospital"] = "Mod Hosp",
  ["modern italy : journal of the association for the study of modern italy"] = "Mod Italy",
  ["modern judaism"] = "Mod Jud",
  ["modern language notes"] = "Mod Lang Notes",
  ["modern logic"] = "Modern Logic",
  ["modern medicine"] = "Mod Med (Chic)",
  ["modern medicine of asia"] = "Mod Med Asia",
  ["modern medicine of canada"] = "Mod Med Can",
  ["modern midwife"] = "Mod Midwife",
  ["modern nursing home"] = "Mod Nurs Home",
  ["modern pathology"] = "Mod.  Pathol.",
  ["modern pathology : an official journal of the united states and canadian academy of pathology, inc"] = "Mod Pathol",
  ["modern pharmacy"] = "Mod Pharm",
  ["modern philology"] = "Mod Philol",
  ["modern physics letters a"] = "Mod. Phys. Lett. A",
  ["modern physics letters a. gravitation, cosmology, astrophysics, nuclear physics, particles and fields, accelerator physics, quantum information"] = "Modern Phys. Lett. A",
  ["modern physics letters a. particles and fields, gravitation, cosmology, astrophysics, nuclear physics, accelerator physics, quantum information"] = "Modern Phys. Lett. A",
  ["modern physics letters a. particles and fields, gravitation, cosmology, nuclear physics"] = "Modern Phys. Lett. A",
  ["modern physics letters b"] = "Mod. Phys. Lett. B",
  ["modern physics letters b. condensed matter physics, statistical physics, atomic, molecular and optical physics"] = "Modern Phys. Lett. B",
  ["modern physics letters. b, condensed matter physics, statistical physics, applied physics"] = "Mod Phys Lett B",
  ["modern probability and statistics"] = "Mod. Probab. Stat.",
  ["modern problems in ophthalmology"] = "Mod Probl Ophthalmol",
  ["modern problems in paediatrics"] = "Mod Probl Paediatr",
  ["modern problems of pharmacopsychiatry"] = "Mod Probl Pharmacopsychiatry",
  ["modern research in catalysis"] = "Mod. Res. Catal.",
  ["modern research in inflammation"] = "Mod Res Inflamm",
  ["modern rheumatology"] = "Mod Rheumatol",
  ["modern rheumatology / the japan rheumatism association"] = "Mod Rheumatol",
  ["modern rheumatology case reports"] = "Mod Rheumatol Case Rep",
  ["modern stochastics. theory and applications"] = "Mod. Stoch. Theory Appl.",
  ["modern theology"] = "Mod Theol",
  ["modern treatment"] = "Mod Treat",
  ["modern treatment yearbook"] = "Mod Treat Yearb",
  ["modern trends in financial engineering"] = "Mod. Trends Financ. Eng.",
  ["modern trends in human reproductive physiology"] = "Mod Trends Hum Reprod Physiol",
  ["modern trends in immunology"] = "Mod Trends Immunol",
  ["modern trends in medical virology"] = "Mod Trends Med Virol",
  ["modern trends in neurology"] = "Mod Trends Neurol",
  ["modern trends in orthopaedics"] = "Mod Trends Orthop",
  ["modern trends in pharmacopsychiatry"] = "Mod Trends Pharmacopsychiatry",
  ["modern trends in plastic surgery"] = "Mod Trends Plast Surg",
  ["modern trends in psychiatry"] = "Mod Trends Psychiatry",
  ["modern trends in radiotherapy"] = "Mod Trends Radiother",
  ["modern trends in rheumatology"] = "Mod Trends Rheumatol",
  ["modern trends in surgery"] = "Mod Trends Surg",
  ["modern veterinary practice"] = "Mod Vet Pract",
  ["modernist studies: literature & culture 1920-1940"] = "Mod Stud Lit Cult",
  ["modular mathematics series"] = "Modular Math. Ser.",
  ["modules in life sciences"] = "Modules Life Sci.",
  ["mohu xitong yu shuxue"] = "Mohu Xitong yu Shuxue",
  ["mois en afrique (paris, france : 1980)"] = "Mois Afr",
  ["moixeró : fongs i flora del pirineu"] = "Moixero",
  ["moj anatomy & physiology"] = "MOJ Anat Physiol",
  ["moj immunology"] = "MOJ Immunol",
  ["moj lymphology & phlebology"] = "MOJ Lymphology Phlebol",
  ["moj orthopedics & rheumatology"] = "MOJ Orthop Rheumatol",
  ["moj proteomics & bioinformatics"] = "MOJ Proteom Bioinform",
  ["moj solar and photoenergy systems"] = "MOJ Sol Photoenergy Syst",
  ["moj sports medicine"] = "MOJ Sports Med",
  ["moj toxicology"] = "MOJ Toxicol",
  ["mokuzai gakkai shi"] = "Mokuzai Gakkaishi",
  ["molecular & biochemical parasitology"] = "Mol. Biochem. Parasitol.",
  ["molecular & cellular biomechanics"] = "Mol. Cell. Biomech.",
  ["molecular & cellular biomechanics : mcb"] = "Mol Cell Biomech",
  ["molecular & cellular epilepsy"] = "Mol Cell Epilepsy",
  ["molecular & cellular oncology"] = "Mol Cell Oncol",
  ["molecular & cellular proteomics"] = "Mol. Cell. Proteomics",
  ["molecular & cellular proteomics : mcp"] = "Mol Cell Proteomics",
  ["molecular & cellular toxicology"] = "Mol. Cell. Toxicol.",
  ["molecular & general genetics : mgg"] = "Mol Gen Genet",
  ["molecular and biochemical parasitology"] = "Mol Biochem Parasitol",
  ["molecular and cell biology of human diseases series"] = "Mol Cell Biol Hum Dis Ser",
  ["molecular and cellular biochemistry"] = "Mol Cell Biochem",
  ["molecular and cellular biology"] = "Mol Cell Biol",
  ["molecular and cellular biomechanics"] = "Mol. Cell. Biomech.",
  ["molecular and cellular differentiation"] = "Mol Cell Differ",
  ["molecular and cellular endocrinology"] = "Mol Cell Endocrinol",
  ["molecular and cellular neuroscience"] = "Mol. Cell. Neurosci.",
  ["molecular and cellular neurosciences"] = "Mol Cell Neurosci",
  ["molecular and cellular pediatrics"] = "Mol Cell Pediatr",
  ["molecular and cellular pharmacology"] = "Mol. Cell. Pharm.",
  ["molecular and cellular probes"] = "Mol Cell Probes",
  ["molecular and cellular proteomics"] = "Mol. Cell. Proteomics",
  ["molecular and cellular therapies"] = "Mol. Cell. Ther.",
  ["molecular and chemical neuropathology"] = "Mol Chem Neuropathol",
  ["molecular and chemical neuropathology / sponsored by the international society for neurochemistry and the world federation of neurology and research groups on neurochemistry and cerebrospinal fluid"] = "Mol Chem Neuropathol",
  ["molecular and clinical oncology"] = "Mol Clin Oncol",
  ["molecular and general genetics"] = "Mol. Gen. Genet.",
  ["molecular and supramolecular photochemistry"] = "Mol. Supramol. Photochem.",
  ["molecular aspects of medicine"] = "Mol Aspects Med",
  ["molecular astrophysics"] = "Mol. Astrophys.",
  ["molecular autism"] = "Mol Autism",
  ["molecular based mathematical biology"] = "Mol Based Math Biol",
  ["molecular biology"] = "Mol Biol",
  ["molecular biology & medicine"] = "Mol Biol Med",
  ["molecular biology (los angeles, calif.)"] = "Mol Biol (Los Angel)",
  ["molecular biology (moscow, russian federation, english edition)"] = "Mol. Biol. (Moscow, Russ. Fed., Engl. Ed.)",
  ["molecular biology (moscow, russian federation, english edition) (translation of molekulyarnaya biologiya)"] = "Mol. Biol. (Moscow, Russ. Fed., Engl. Ed.)",
  ["molecular biology (new york, ny, united states, english edition)"] = "Mol. Biol. (N. Y., NY, U. S., Engl. Ed.)",
  ["molecular biology and evolution"] = "Mol Biol Evol",
  ["molecular biology and medicine"] = "Mol. Biol. Med.",
  ["molecular biology international"] = "Mol Biol Int",
  ["molecular biology of the cell"] = "Mol Biol Cell",
  ["molecular biology reports"] = "Mol Biol Rep",
  ["molecular biology research communications"] = "Mol Biol Res Commun",
  ["molecular biology today : the online journal of molecular biology"] = "Mol Biol Today",
  ["molecular biology, biochemistry and biophysics"] = "Mol. Biol. Biochem. Biophys.",
  ["molecular biology, biochemistry, and biophysics"] = "Mol Biol Biochem Biophys",
  ["molecular biomedicine"] = "Mol. Biomed.",
  ["molecular biosystems"] = "Mol. Biosyst.",
  ["molecular biotechnology"] = "Mol Biotechnol",
  ["molecular biotherapy"] = "Mol Biother",
  ["molecular brain"] = "Mol Brain",
  ["molecular brain research"] = "Mol. Brain. Res.",
  ["molecular breeding"] = "Mol. Breed.",
  ["molecular breeding : new strategies in plant improvement"] = "Mol Breed",
  ["molecular cancer"] = "Mol Cancer",
  ["molecular cancer biology"] = "Mol Cancer Biol",
  ["molecular cancer research"] = "Mol. Cancer Res.",
  ["molecular cancer research : mcr"] = "Mol Cancer Res",
  ["molecular cancer therapeutics"] = "Mol Cancer Ther",
  ["molecular carcinogenesis"] = "Mol Carcinog",
  ["molecular catalysis"] = "Mol. Catal.",
  ["molecular cell"] = "Mol Cell",
  ["molecular cell biology research communications"] = "Mol. Cell Biol. Res. Commun.",
  ["molecular cell biology research communications : mcbrc"] = "Mol Cell Biol Res Commun",
  ["molecular crystals & liquid crystals"] = "Mol. Cryst. Liq. Cryst.",
  ["molecular crystals and liquid crystals"] = "Mol. Cryst. Liq. Cryst.",
  ["molecular crystals and liquid crystals lett."] = "Mol. Cryst. Liq. Cryst. Lett.",
  ["molecular crystals and liquid crystals science and technology section a"] = "Mol. Cryst. Liq. Cryst. Sci. Technol., Sect. A",
  ["molecular crystals and liquid crystals science and technology section b"] = "Mol. Cryst. Liq. Cryst. Sci. Technol., Sect. B",
  ["molecular crystals and liquid crystals science and technology section c"] = "Mol. Cryst. Liq. Cryst. Sci. Technol., Sect. C",
  ["molecular cytogenetics"] = "Mol Cytogenet",
  ["molecular diagnosis"] = "Mol. Diagn.",
  ["molecular diagnosis & therapy"] = "Mol Diagn Ther",
  ["molecular diagnosis : a journal devoted to the understanding of human disease through the clinical application of molecular biology"] = "Mol Diagn",
  ["molecular diagnosis and therapy"] = "Mol. Diagn. Ther.",
  ["molecular diversity"] = "Mol Divers",
  ["molecular ecology"] = "Mol Ecol",
  ["molecular ecology notes"] = "Mol Ecol Notes",
  ["molecular ecology resources"] = "Mol Ecol Resour",
  ["molecular endocrinology"] = "Mol. Endocrinol.",
  ["molecular endocrinology (baltimore, md.)"] = "Mol Endocrinol",
  ["molecular entomology"] = "Mol Entomol",
  ["molecular frontiers journal"] = "Mol Front J",
  ["molecular genetic medicine"] = "Mol Genet Med",
  ["molecular genetics & genomic medicine"] = "Mol Genet Genomic Med",
  ["molecular genetics and genomics"] = "Mol. Genet. Genomics",
  ["molecular genetics and genomics : mgg"] = "Mol Genet Genomics",
  ["molecular genetics and metabolism"] = "Mol Genet Metab",
  ["molecular genetics and metabolism reports"] = "Mol Genet Metab Rep",
  ["molecular genetics, microbiology and virology"] = "Mol. Genet. Microbiol. Virol.",
  ["molecular genetics, microbiology and virology : molekulyarnaya genetika, mikrobiologiya i virusologiya"] = "Mol Gen Microbiol Virol",
  ["molecular horticulture"] = "Mol. Hortic.",
  ["molecular human reproduction"] = "Mol Hum Reprod",
  ["molecular imaging"] = "Mol Imaging",
  ["molecular imaging & biology"] = "Mol. Imag. Biol.",
  ["molecular imaging : official journal of the society for molecular imaging"] = "Mol Imaging",
  ["molecular imaging and biology"] = "Mol Imaging Biol",
  ["molecular imaging and biology : mib : the official publication of the academy of molecular imaging"] = "Mol Imaging Biol",
  ["molecular imaging and radionuclide therapy"] = "Mol Imaging Radionucl Ther",
  ["molecular imaging, reconstruction and analysis of moving body organs, and stroke imaging and treatment : fifth international workshop, cmmi 2017, second international workshop, rambo 2017, and first international workshop, switch 2017, held in conjunction with miccai 2017, québec city, qc, canada, september 14, 2017, proceedings. cmmi (workshop) (5th : 2017 : québec, québec)"] = "Mol Imaging Reconstr Anal Mov Body Organs Stroke Imaging Treat (2017)",
  ["molecular immunology"] = "Mol Immunol",
  ["molecular imprinting"] = "Mol. Imprint.",
  ["molecular informatics"] = "Mol Inform",
  ["molecular interventions"] = "Mol Interv",
  ["molecular life"] = "Mol Life",
  ["molecular marine biology and biotechnology"] = "Mol Mar Biol Biotechnol",
  ["molecular medicine"] = "Mol. Med.",
  ["molecular medicine & therapeutics"] = "Mol Med Ther",
  ["molecular medicine (cambridge, mass.)"] = "Mol Med",
  ["molecular medicine (manhasset, ny, united states)"] = "Mol. Med. (Manhasset, NY, U. S.)",
  ["molecular medicine reports"] = "Mol Med Rep",
  ["molecular medicine today"] = "Mol Med Today",
  ["molecular membrane biology"] = "Mol Membr Biol",
  ["molecular metabolism"] = "Mol Metab",
  ["molecular microbiology"] = "Mol Microbiol",
  ["molecular modeling and simulation. applications and perspectives"] = "Mol. Model. Simul. Appl. Perspect.",
  ["molecular neurobiology"] = "Mol Neurobiol",
  ["molecular neurodegeneration"] = "Mol Neurodegener",
  ["molecular neuropsychiatry"] = "Mol Neuropsychiatry",
  ["molecular nutrition & food research"] = "Mol Nutr Food Res",
  ["molecular nutrition and food research"] = "Mol. Nutr. Food Res.",
  ["molecular omics"] = "Mol Omics",
  ["molecular oncology"] = "Mol Oncol",
  ["molecular oral microbiology"] = "Mol Oral Microbiol",
  ["molecular pain"] = "Mol Pain",
  ["molecular pathology"] = "Mol. Pathol.",
  ["molecular pathology : mp"] = "Mol Pathol",
  ["molecular pharmaceutics"] = "Mol Pharm",
  ["molecular pharmacology"] = "Mol Pharmacol",
  ["molecular phylogenetics and evolution"] = "Mol Phylogenet Evol",
  ["molecular physics"] = "Mol. Phys.",
  ["molecular plant"] = "Mol Plant",
  ["molecular plant breeding"] = "Mol Plant Breed",
  ["molecular plant breeding = fen zi zhi wu yu zhong"] = "Mol Plant Breed",
  ["molecular plant pathology"] = "Mol Plant Pathol",
  ["molecular plant-microbe interactions"] = "Mol. Plant-Microbe Interact.",
  ["molecular plant-microbe interactions : mpmi"] = "Mol Plant Microbe Interact",
  ["molecular psychiatry"] = "Mol Psychiatry",
  ["molecular reproduction & development"] = "Mol. Reprod. Dev.",
  ["molecular reproduction and development"] = "Mol Reprod Dev",
  ["molecular screening news"] = "Mol. Screen. News",
  ["molecular simulation"] = "Mol. Simul.",
  ["molecular syndromology"] = "Mol Syndromol",
  ["molecular systems biology"] = "Mol Syst Biol",
  ["molecular systems design & engineering"] = "Mol Syst Des Eng",
  ["molecular systems design and engineering"] = "Mol. Syst. Des. Eng.",
  ["molecular therapy"] = "Mol. Ther.",
  ["molecular therapy - methods & clinical development"] = "Mol. Ther. Methods Clin. Dev.",
  ["molecular therapy - nucleic acids"] = "Mol. Ther. Nucleic Acids",
  ["molecular therapy - oncolytics"] = "Mol. Ther. Oncolytics",
  ["molecular therapy : the journal of the american society of gene therapy"] = "Mol Ther",
  ["molecular therapy oncolytics"] = "Mol Ther Oncolytics",
  ["molecular therapy — oncolytics"] = "Mol. Ther. Oncolytics",
  ["molecular therapy. methods & clinical development"] = "Mol Ther Methods Clin Dev",
  ["molecular therapy. nucleic acids"] = "Mol Ther Nucleic Acids",
  ["molecular therapy–methods & clinical development"] = "Mol. Ther.–Methods Clin. Dev.",
  ["molecular therapy–nucleic acids"] = "Mol. Ther.–Nucleic Acids",
  ["molecular toxicology"] = "Mol Toxicol",
  ["molecular urology"] = "Mol Urol",
  ["molecular vision"] = "Mol Vis",
  ["molecules (basel, switzerland)"] = "Molecules",
  ["molecules and cells"] = "Mol Cells",
  ["molecules online"] = "Mol. Online",
  ["molekuliarna meditsina"] = "Mol. Med. (Sofia)",
  ["molekuliarna meditsina = molecular medicine"] = "Mol Med (Sofia)",
  ["molekuliarnaia biologiia"] = "Mol Biol (Mosk)",
  ["molekuliarnaia genetika, mikrobiologia, i virusologa"] = "Mol Gen Mikrobiol Virusol",
  ["molekuliarnaia genetika, mikrobiologiia i virusologia"] = "Mol. Gen. Mikrobiol. Virusol.",
  ["molekuliarnaia genetika, mikrobiologiia i virusologiia"] = "Mol Gen Mikrobiol Virusol",
  ["molekulyarnaya biologiya (moscow)"] = "Mol. Biol. (Moscow)",
  ["molekulyarnaya genetika, mikrobiologiya i virusologiya"] = "Mol. Genet., Mikrobiol. Virusol.",
  ["molluscan research"] = "Molluscan Res",
  ["moment (new york, n.y.)"] = "Moment",
  ["momenti e problemi della storia del pensiero"] = "Momenti Probl. Storia Pensiero",
  ["momento económico"] = "Momento Econ.",
  ["mon-khmer studies"] = "MonKhmer Stud",
  ["monaldi archives for chest disease"] = "Monaldi Arch. Chest Dis.",
  ["monaldi archives for chest disease = archivio monaldi per le malattie del torace"] = "Monaldi Arch Chest Dis",
  ["monash bioethics review"] = "Monash Bioeth Rev",
  ["monash university law review. monash university. faculty of law"] = "Monash Univ Law Rev",
  ["monastic studies"] = "Monastic Stud",
  ["monatshefte fuer chemie"] = "Monatsh. Chem.",
  ["monatshefte fur chemie"] = "Monatsh. Chem.",
  ["monatshefte fur mathematik"] = "Monatsh. Math.",
  ["monatshefte fur tierheilkunde"] = "Monatsh Tierheilkd",
  ["monatshefte fur veterinarmedizin"] = "Monatsh. Veterinarmed.",
  ["monatshefte für chemie"] = "Monatsh Chem",
  ["monatshefte für mathematik"] = "Mon Hefte Math",
  ["monatshefte für tierheilkunde"] = "Monatsh Tierheilkd",
  ["monatshefte für veterinärmedizin"] = "Monatsh Veterinarmed",
  ["monatshefte für chemie"] = "Monatsh. Chem.",
  ["monatshefte für mathematik"] = "Monatsh. Math.",
  ["monatskurse fur die arztliche fortbildung. bildbeilage; farbige medizin"] = "Monatskurse Arztl Fortbild [Bildbeil]",
  ["monatskurse für die ärztliche fortbildung. bildbeilage; farbige medizin"] = "Monatskurse Arztl Fortbild Bildbeil",
  ["monatsschrift deutscher zahnarzte: der freie zahnarzt"] = "Monatsschr Dtsch Zahnarzte Freie Zahnarzt",
  ["monatsschrift deutscher zahnarzte; der freie zahnarzt"] = "Monatsschr. Dtsch. Zahnarzte Freie Zahnarzt",
  ["monatsschrift deutscher zahnärzte: der freie zahnarzt"] = "Monatsschr Dtsch Zahnarzte Freie Zahnarzt",
  ["monatsschrift fur brauwissenschaft"] = "Monatsschr. Brauwiss.",
  ["monatsschrift fur kinderheilkunde"] = "Monatsschr. Kinderheilkd.",
  ["monatsschrift fur ohrenheilkunde und laryngo-rhinologie"] = "Monatsschr. Ohrenheilkd. Laryngorhinol.",
  ["monatsschrift fur psychiatrie und neurologie"] = "Monatsschr Psychiatr Neurol",
  ["monatsschrift fur tuberkulosebekampfung"] = "Monatsschr Tuberkulosebekampf",
  ["monatsschrift fur unfallheilkunde"] = "Monatsschr. Unfallheilkd.",
  ["monatsschrift fur unfallheilkunde und versicherungsmedizin"] = "Monatsschr Unfallheilkd Versicherungsmed",
  ["monatsschrift fur unfallheilkunde, versicherungs-, versorgungs- und verkehrsmedizin"] = "Monatsschr. Unfallheilkd. Versicher. Versorg. Verkehrsmed.",
  ["monatsschrift für brauwissenschaft (internet)"] = "Monatsschrift Brauwiss",
  ["monatsschrift für geburtshülfe und gynäkologie"] = "Monatsschr Geburtshilfe Gynakol",
  ["monatsschrift für kinderheilkunde"] = "Monatsschr Kinderheilkd",
  ["monatsschrift für kriminologie und strafrechtsreform"] = "Monatsschr Kriminol",
  ["monatsschrift für ohrenheilkunde und laryngo-rhinologie"] = "Monatsschr Ohrenheilkd Laryngorhinol",
  ["monatsschrift für psychiatrie und neurologie"] = "Monatsschr Psychiatr Neurol",
  ["monatsschrift für tuberkulosebekämpfung"] = "Monatsschr Tuberkulosebekampf",
  ["monatsschrift für unfallheilkunde"] = "Monatsschr Unfallheilkd",
  ["monatsschrift für unfallheilkunde und versicherungsmedizin"] = "Monatsschr Unfallheilkd Versicherungsmed",
  ["monatsschrift für unfallheilkunde, versicherungs-, versorgungs- und verkehrsmedizin"] = "Monatsschr Unfallheilkd Versicher Versorg Verkehrsmed",
  ["monatsschrift für das württembergische forstwesen"] = "Mon.schr. württ. Forstwes.",
  ["monatsschrift kinderheilkunde"] = "Monatsschr. Kinderheilkd.",
  ["monatsschrift kinderheilkunde : organ der deutschen gesellschaft fur kinderheilkunde"] = "Monatsschr Kinderheilkd",
  ["monatsschrift kinderheilkunde : organ der deutschen gesellschaft für kinderheilkunde"] = "Monatsschr Kinderheilkd",
  ["mondai to kenkyu : chugoku jijo senmonshi"] = "Mondai To Kenkyu",
  ["mondai to kenkyū : chūgoku jijō senmonshi"] = "Mondai To Kenkyu",
  ["monday developments : a biweekly publication of interaction"] = "Monday Dev",
  ["monde dentaire"] = "Monde Dent.",
  ["monde medical"] = "Monde Med",
  ["monde médical"] = "Monde Med",
  ["mondes en developpement"] = "Mondes Dev",
  ["mondes et cultures : comptes rendus trimestriels des seances de l’academie des sciences d’outre-mer"] = "Mondes Cult",
  ["mondes et cultures : comptes rendus trimestriels des séances de l'académie des sciences d'outre-mer"] = "Mondes Cult",
  ["mondo aperto"] = "Mondo Aperto",
  ["mondo cinese"] = "Mondo Cin",
  ["mondo odontostomatologico"] = "Mondo Odontostomatol",
  ["mondo ortodontico"] = "Mondo Ortod",
  ["moneda y crédito"] = "Moneda Crédito",
  ["moneta e credito"] = "Moneta Credito",
  ["monetary and economic studies"] = "Monet. Econ. Stud.",
  ["mongolian journal of biological sciences"] = "Mong J Biol Sci",
  ["mongolian journal of demography : journal of population teaching and research center"] = "Mong J Demogr",
  ["mongolian mathematical journal"] = "Mong. Math. J.",
  ["monitor (associates of clinical pharmacology)"] = "Monitor (Assoc Clin Pharmacol)",
  ["monitor de la farmacía y de la terapéutica"] = "Monit Farm Ter",
  ["monitore ostetrico-ginecologico"] = "Monit Ostet Ginecolog",
  ["monitore ostetrico-ginecologico di endocrinologia e del metabolismo"] = "Monit Ostet Ginecol Endocrinol Metab",
  ["monoclonal antibodies in immunodiagnosis and immunotherapy"] = "Monoclon Antib Immunodiagn Immunother",
  ["monografia del bollettino. societa italiana di medicina e igiene tropicale. sezione eritrea"] = "Monogr Boll",
  ["monografias de la corporacion centro regional de la poblacion"] = "Monogr Corp Cent Reg Poblac",
  ["monografias de la real academia de ciencias exactas, fisicas, quimicas y naturales de zaragoza"] = "Monogr. Real Acad. Ci. Exact. Fis.-Quim. Nat. Zaragoza",
  ["monografias matematicas “garcia de galdeano”"] = "Monogr. Mat. Garcia Galdeano",
  ["monografie della scuola archeologica di atene e delle missioni italiane in oriente"] = "MSAtene",
  ["monografie di pagine di micologia"] = "Monogr Pagine Micol",
  ["monografie di quaderni internazionali di storia della medicina e della sanita"] = "Monogr Quad Int Stor Med Sanita",
  ["monografie di quaderni internazionali di storia della medicina e della sanità"] = "Monogr Quad Int Stor Med Sanita",
  ["monografie matematyczne"] = "Monogr. Mat.",
  ["monografii matematice"] = "Monogr. Mat.",
  ["monografii. novosibirskii gosudartstvennyi tekhnicheskii universitet"] = "Monografii NGTU",
  ["monografías de la corporación centro regional de la población"] = "Monogr Corp Cent Reg Poblac",
  ["monografías ciencia y tecnología"] = "Monogr. Cienc. Tecnol.",
  ["monografías de la academia de ciencias exactas, físicas, químicas y naturales de zaragoza"] = "Monogr. Acad. Ci. Exact. Fís.-Quím. Nat. Zaragoza",
  ["monograph of the american association of mental deficiency"] = "Monogr Am Assoc Ment Defic",
  ["monograph series of the parana’s mathematical society"] = "Monogr. Ser. Parana’s Math. Soc.",
  ["monograph series. world health organization"] = "Monogr Ser World Health Organ",
  ["monographien aus dem gesamtgebiete der neurologie und psychiatrie"] = "Monogr Gesamtgeb Neurol Psychiatr",
  ["monographien aus dem gesamtgebiete der psychiatrie"] = "Monogr Gesamtgeb Psychiatr Psychiatry Ser",
  ["monographien aus dem gesamtgebiete der psychiatrie. psychiatry series"] = "Monogr. Gesamtgeb. Psychiatr. Psychiatry Ser.",
  ["monographien zur modernen mathematik"] = "Monogr. Modernen Math.",
  ["monographies de linguistique mathematique"] = "Monogr. Linguist. Math.",
  ["monographs and research notes in mathematics"] = "Monogr. Res. Notes Math.",
  ["monographs and surveys in water resource engineering"] = "Monographs Surveys Water Res. Engrg.",
  ["monographs and textbooks in pure and applied mathematics"] = "Monogr. Textbooks Pure Appl. Math.",
  ["monographs in allergy"] = "Monogr Allergy",
  ["monographs in clinical cytology"] = "Monogr Clin Cytol",
  ["monographs in clinical neuroscience"] = "Monogr Clin Neurosci",
  ["monographs in computer science"] = "Monogr. Comput. Sci.",
  ["monographs in contemporary mathematics"] = "Monogr. Contemp. Math.",
  ["monographs in developmental biology"] = "Monogr Dev Biol",
  ["monographs in human genetics"] = "Monogr Hum Genet",
  ["monographs in inequalities"] = "Monogr. Inequal.",
  ["monographs in mathematical economics"] = "Monogr. Math. Econ.",
  ["monographs in mathematics"] = "Monogr. Math.",
  ["monographs in neural sciences"] = "Monogr Neural Sci",
  ["monographs in number theory"] = "Monogr. Number Theory",
  ["monographs in oral science"] = "Monogr Oral Sci",
  ["monographs in paediatrics"] = "Monogr Paediatr",
  ["monographs in pathology"] = "Monogr Pathol",
  ["monographs in population biology"] = "Monogr Popul Biol",
  ["monographs in soviet medical sciences"] = "Monogr Sov Med Sci",
  ["monographs in the surgical sciences"] = "Monogr Surg Sci",
  ["monographs in theoretical computer science"] = "Monogr. Theoret. Comput. Sci. EATCS Ser.",
  ["monographs in theoretical computer science. an eatcs series"] = "Monogr. Theoret. Comput. Sci. EATCS Ser.",
  ["monographs in virology"] = "Monogr Virol",
  ["monographs of the american association on mental deficiency"] = "Monogr Am Assoc Ment Defic",
  ["monographs of the american association on mental deficiency (1982)"] = "Monogr Am Assoc Ment Defic",
  ["monographs of the american association on mental retardation"] = "Monogr Am Assoc Ment Retard",
  ["monographs of the palaeontographical society"] = "Monogr. Palaeontogr. Soc.",
  ["monographs of the physiological society"] = "Monogr Physiol Soc",
  ["monographs of the society for research in child development"] = "Monogr Soc Res Child Dev",
  ["monographs of the western north american naturalist"] = "Monogr West N Am Nat",
  ["monographs on atherosclerosis"] = "Monogr Atheroscler",
  ["monographs on endocrinology"] = "Monogr Endocrinol",
  ["monographs on numerical analysis"] = "Monogr. Numer. Anal.",
  ["monographs on statistics and applied probability"] = "Monogr. Statist. Appl. Probab.",
  ["monographs on the physics and chemistry of materials"] = "Monogr. Phys. Chem. Mater.",
  ["monspeliensis hippocrates"] = "Monspel Hippocrates",
  ["montana : the magazine of western history"] = "Montana",
  ["montana code annotated. montana"] = "Mont Code Annot Mont",
  ["montana law review"] = "Mont Law Rev",
  ["montana wildlife (bozeman, mont.)"] = "Mont Wildl",
  ["monte carlo methods and applications"] = "Monte Carlo Methods Appl.",
  ["monte verità"] = "Monte Verità",
  ["monterrey médico"] = "Monterrey Med",
  ["montfort : vierteljahresschrift fur geschichte und gegenwartskunde vorarlbergs"] = "Montfort",
  ["montfort : vierteljahresschrift für geschichte und gegenwartskunde vorarlbergs"] = "Montfort",
  ["monthly bulletin of statistics (united nations. statistical office)"] = "Mon Bull Stat U N Stat Off",
  ["monthly bulletin of the bureau of health"] = "Mon Bull Bur Health",
  ["monthly bulletin of the ministry of health and the emergency public health laboratory service"] = "Mon Bull Minist Health Emerg Public Health Lab Serv",
  ["monthly bulletin of the ministry of health and the public health laboratory service"] = "Mon Bull Minist Health Public Health Lab Serv",
  ["monthly bulletin. great britain. emergency public health laboratory service"] = "Mon Bull G B Emerg Public Health Lab Serv",
  ["monthly bulletin. indiana state board of health"] = "Mon Bull Indiana State Board Health",
  ["monthly health bulletin. united states. army. army, 2nd"] = "Mon Heath Bull",
  ["monthly health report. united states. army. military district of washington"] = "Mon Health Rep U S Army",
  ["monthly journal of medical science"] = "Mon J Med Sci",
  ["monthly journal of medicine (edinburgh, scotland)"] = "Mon J Med (Edinb)",
  ["monthly labor review"] = "Mon Labor Rev",
  ["monthly labor review / u.s. department of labor, bureau of labor statistics"] = "Mon Labor Rev",
  ["monthly microscopical journal"] = "Mon. Microsc. J.",
  ["monthly notices of the royal astronomical society"] = "Mon Not R Astron Soc",
  ["monthly notices of the royal astronomical society letters"] = "Mon. Not. R. Astron. Soc. Lett.",
  ["monthly notices of the royal astronomical society. letters"] = "Mon Not R Astron Soc Lett",
  ["monthly review (new york, n.y. : 1949)"] = "Mon Rev",
  ["monthly review - new york state department of labor, division of industrial hygiene"] = "Mon Rev New York State Dep Labor Div Ind Hyg",
  ["monthly vital statistics report"] = "Mon Vital Stat Rep",
  ["monthly weather review"] = "Mon Weather Rev",
  ["monthy notices of the royal astronomical society"] = "Mon. Not. R. Astron. Soc.",
  ["monthy notices of the royal astronomical society letters"] = "Mon. Not. R. Astron. Soc. Lett.",
  ["monti e boschi"] = "Monti boschi",
  ["montpellier medical"] = "Montp. Med.",
  ["montpellier médical"] = "Montp Med",
  ["montreal medical"] = "Montreal Med",
  ["montréal médical"] = "Montreal Med",
  ["montroll memorial lecture series in mathematical physics"] = "Montroll Memorial Lecture Ser. Math. Phys.",
  ["monumenta artis romanae"] = "MARo",
  ["monumenta asiae minoris antiqua. publications of the american society for archaeological research in asia minor"] = "MAMA",
  ["monumenta nipponica"] = "Monum Nihon",
  ["monumenti antichi"] = "MonAnt",
  ["monumenti antichi, pubblicati dall’accademia dei lincei"] = "MonAL",
  ["monumenti della pittura antica scoperti in italia"] = "MonPitt",
  ["monumenti inediti pubblicati dall’istituto archeologico"] = "MonInst",
  ["monuments et mémoires publiés par l’académie des inscriptions et belles-lettres (fondation piot)"] = "MMAI",
  ["monuments et mémoires. fondation e. piot"] = "MonPiot",
  ["monuments historiques : mh"] = "Monum Hist",
  ["moral philosophy and politics"] = "Moral Philos Politics",
  ["morbidity and mortality"] = "Morb Mortal",
  ["morbidity and mortality weekly report. surveillance summaries (washington, d.c. : 2002)"] = "MMWR Surveill Summ",
  ["morbidity and mortality weekly report. surveillance summaries : mmwr / centers for disease control"] = "Morb Mortal Wkly Rep Surveill Summ",
  ["morfologiia (saint petersburg, russia)"] = "Morfologiia",
  ["morfologiia (sankt-peterburg"] = "Morfologiia",
  ["morfologii︠a︡ (saint petersburg, russia)"] = "Morfologiia",
  ["morgannwg : transactions of the glamorgan local history society"] = "Morgannwg",
  ["morningside lectures in mathematics"] = "Morningside Lect. Math.",
  ["moroccan journal of algebra and geometry with applications"] = "Moroc. J. Algebra Geom. Appl.",
  ["moroccan journal of chemistry"] = "Morrocan J. Chem.",
  ["moroccan journal of pure and applied analysis"] = "Morrocan J. Pure Appl. Anal.",
  ["morphologia medica"] = "Morphol Med",
  ["morphologiai es igazsagugyi orvosi szemle"] = "Morphol. Igazsagugyi Orv. Sz.",
  ["morphologiai és igazságügyi orvosi szemle"] = "Morphol Igazsagugyi Orv Sz",
  ["morphologie : bulletin de l'association des anatomistes"] = "Morphologie",
  ["morphologie : bulletin de l’association des anatomistes"] = "Morphologie",
  ["morphologie et embryologie"] = "Morphol Embryol (Bucur)",
  ["morphology (dordrecht, netherlands)"] = "Morphology (Dordr)",
  ["mortality (abingdon, england)"] = "Mortality (Abingdon)",
  ["mos-siam series on optimization"] = "MOS-SIAM Ser. Optim.",
  ["moscow journal of combinatorics and number theory"] = "Moscow J. Comb. Number Theory",
  ["moscow lectures"] = "Mosc. Lect.",
  ["moscow mathematical journal"] = "Moscow Math. J.",
  ["moscow physical society"] = "J. Moscow Phys. Soc.",
  ["moscow university biological sciences bulletin"] = "Moscow Univ Biol Sci Bull",
  ["moscow university chemistry bulletin"] = "Moscow Univ. Chem. Bull.",
  ["moscow university computational mathematics and cybernetics"] = "Moscow Univ. Comput. Math. Cybernet.",
  ["moscow university geology bulletin"] = "Moscow Univ. Geol. Bull.",
  ["moscow university mathematics bulletin"] = "Moscow Univ. Math. Bull.",
  ["moscow university mechanics bulletin"] = "Moscow Univ. Mech. Bull.",
  ["moscow university physics bulletin"] = "Moscow Univ. Phys. Bull.",
  ["moscow university soil science bulletin"] = "Moscow Univ. Soil Sci. Bull.",
  ["moskovskiy universitet"] = "Trudy Sem. Petrovsk.",
  ["mosquito news"] = "Mosq News",
  ["most newsletter"] = "MOST newsl",
  ["mother and child"] = "Mother Child",
  ["mother jones"] = "Mother Jones",
  ["mothers and children"] = "Mothers Child",
  ["motivation and emotion"] = "Motiv Emot",
  ["motivation science"] = "Motiv Sci",
  ["motivational interviewing : training, research, implementation, practice"] = "Motiv Interviewing",
  ["motor control"] = "Motor Control",
  ["motortechnische zeitschrift"] = "Mot. Z.",
  ["motriz : revista de educação física. unesp"] = "Motriz",
  ["mount sinai journal of medicine"] = "Mt. Sinai J. Med.",
  ["mountain geologist"] = "Mt. Geol.",
  ["mountain research and development"] = "Mt Res Dev",
  ["mouse genome"] = "Mouse Genome",
  ["mouth (american student dental association)"] = "Mouth",
  ["movement & sport science"] = "Mov. Sport Sci.",
  ["movement disorders"] = "Mov. Disord.",
  ["movement disorders : official journal of the movement disorder society"] = "Mov Disord",
  ["movement disorders clinical practice"] = "Mov Disord Clin Pract",
  ["movement ecology"] = "Mov Ecol",
  ["movimento operaio e socialista"] = "Mov Operaio Soc",
  ["moving boundaries vii: computational modelling of free and moving boundary problems"] = "Comp. Exptl. Methods.",
  ["moving worlds"] = "Mov Worlds",
  ["mozia. rapporto preliminare della missione congiunta con la soprintendenza alle antichità della sicilia occidentale"] = "Mozia",
  ["mp medycyna praktyczna. pediatria"] = "MP Med Prakt Pediatr",
  ["mps. medical products salesman"] = "MPS",
  ["mps; medical products salesman"] = "MPS",
  ["mpt metallurgical plant and technology"] = "MPT Metall. Plant Technol.",
  ["mpt metallurgical plant and technology international"] = "MPT Metall. Plant Technol. Int.",
  ["mrc news"] = "MRC News",
  ["mrs advances"] = "MRS Adv.",
  ["mrs bulletin"] = "MRS Bull.",
  ["mrs bulletin / materials research society"] = "MRS Bull",
  ["mrs communications"] = "MRS Commun.",
  ["mrs energy & sustainability"] = "MRS Energy & Sustainability",
  ["mrs internet journal of nitride semiconductor research"] = "MRS Internet J. Nitride Semicond. Res.",
  ["mrs online proceedings library"] = "MRS Online Proc. Lib.",
  ["mrs proceedings"] = "MRS Proc.",
  ["ms&a. modeling, simulation and applications"] = "MS&A. Model. Simul. Appl.",
  ["ms. (new york, n.y. : 1990)"] = "Ms",
  ["msda journal"] = "MSDA J.",
  ["msda journal : journal of the maryland state dental association"] = "MSDA J",
  ["msj memoirs"] = "MSJ Mem.",
  ["msn; monthly science news"] = "MSN Mon Sci News",
  ["msri mathematical circles library"] = "MSRI Math. Circ. Libr.",
  ["mucopolysaccharides biochimica et biophysica acta, specialized section on mucoproteins and mucopolysaccharides"] = "Biochim. Biophys. Acta, Spec. Sect. Mucoproteins",
  ["mucosal immunology"] = "Mucosal Immunol",
  ["mull und abfall"] = "Mull Abfall",
  ["multiagent and grid systems"] = "Multiagent Grid Syst.",
  ["multibody system dynamics"] = "Multibody Syst Dyn",
  ["multidimensional systems and signal processing"] = "Multidimension. Syst. Signal Process.",
  ["multidimensional systems and signal processing. an international journal"] = "Multidimens. Syst. Signal Process.",
  ["multidisciplinary and applied optics"] = "Multidiscip. Appl. Opt.",
  ["multidisciplinary journal of educational research"] = "Multidiscip J Educ Research",
  ["multidisciplinary respiratory medicine"] = "Multidiscip Respir Med",
  ["multidiscipline modeling in materials and structures"] = "Multidiscip. Model. Mater. Struct.",
  ["multifunctional materials"] = "Multifunct. Mater.",
  ["multimedia manual of cardiothoracic surgery : mmcts"] = "Multimed Man Cardiothorac Surg",
  ["multimedia systems"] = "Multimed Syst",
  ["multimedia tools and applications"] = "Multimed Tools Appl",
  ["multimodal brain image analysis : first international workshop, mbia 2011, held in conjunction with miccai 2011, toronto, canada, september 18, 2011 : proceedings"] = "Multimodal Brain Image Anal (2011)",
  ["multimodal brain image analysis : second international workshop, mbia 2012, held in conjunction with miccai 2012, nice, france, october 1-5, 2012 : proceedings. mbia (workshop) (2nd : 2012 : nice, france)"] = "Multimodal Brain Image Anal (2012)",
  ["multimodal brain image analysis : third international workshop, mbia 2013, held in conjunction with miccai 2013, nagoya, japan, september 22, 2013 : proceedings. mbia (workshop) (3rd : 2013 : nagoya-shi, japan)"] = "Multimodal Brain Image Anal (2013)",
  ["multimodal brain image analysis and mathematical foundations of computational anatomy : 4th international workshop, mbia 2019, and 7th international workshop, mfca 2019, held in conjunction with miccai 2019, shenzhen, china, october 17, 2019, proceedings. mbia (workshop) (4th : 2019 : shenzhen shi, china)"] = "Multimodal Brain Image Anal Math Found Comput Anat (2019)",
  ["multimodal learning for clinical decision support and clinical image-based procedures : 10th international workshop, ml-cds 2020, and 9th international workshop, clip 2020, held in conjunction with miccai 2020, lima, peru, october 4-8, 2020, proceedings. ml-cds (workshop) (10th : 2020 : online)"] = "Multimodal Learn Clin Decis Support Clin Image Based Proc (2020)",
  ["multinational finance journal"] = "Multinat. Finance J.",
  ["multiphase science and technology"] = "Multiphase Sci. Technol.",
  ["multiphysics modeling"] = "Multiphys. Model.",
  ["multiple criteria decision making"] = "Mult. Criteria Decis. Mak.",
  ["multiple sclerosis"] = "Mult. Scler.",
  ["multiple sclerosis (houndmills, basingstoke, england)"] = "Mult Scler",
  ["multiple sclerosis and demyelinating disorders"] = "Mult Scler Demyelinating Disord",
  ["multiple sclerosis and related disorders"] = "Mult Scler Relat Disord",
  ["multiple sclerosis international"] = "Mult Scler Int",
  ["multiple sclerosis journal - experimental, translational and clinical"] = "Mult Scler J Exp Transl Clin",
  ["multiscale and multidisciplinary modeling, experiments and design"] = "Multiscale Multidiscip. Model. Exp. Des.",
  ["multiscale modeling & simulation : a siam interdisciplinary journal"] = "Multiscale Model Simul",
  ["multiscale modeling & simulation. a siam interdisciplinary journal"] = "Multiscale Model. Simul.",
  ["multiscale modeling and simulation"] = "Multiscale Model. Simul.",
  ["multiscale science and engineering"] = "Multiscale Sci. Eng.",
  ["multisensory research"] = "Multisens Res",
  ["multivariate behavioral research"] = "Multivariate Behav Res",
  ["munca sanitara"] = "Munca Sanit.",
  ["munca sanitară"] = "Munca Sanit",
  ["munchener medizinische wochenschrift"] = "Munch. Med. Wochenschr.",
  ["munchener medizinische wochenschrift (1950)"] = "Munch Med Wochenschr",
  ["mund-, kiefer- und gesichtschirurgie"] = "Mund. Kiefer. Gesichtschir.",
  ["mund-, kiefer- und gesichtschirurgie : mkg"] = "Mund Kiefer Gesichtschir",
  ["mundo da saúde (1995)"] = "Mundo Saude (1995)",
  ["mundo medico"] = "Mundo Med",
  ["mundo médico"] = "Mundo Med",
  ["munger africana library notes"] = "Munger Afr Libr Notes",
  ["munibe. ciencias naturales"] = "Munibe",
  ["munster journal of mathematics"] = "Munster J. Math.",
  ["munster lectures in mathematics"] = "Munst. Lect. Math.",
  ["munstersche beitrage zur geschichte und theorie der medizin"] = "Munstersche Beitr Gesch Theor Med",
  ["muscle & nerve"] = "Muscle Nerve",
  ["muscle & nerve. supplement"] = "Muscle Nerve Suppl",
  ["muscle and nerve"] = "Muscle Nerve",
  ["muscle and nerve. supplement"] = "Muscle Nerve. Suppl.",
  ["muscle biology"] = "Muscle Biol",
  ["muscles, ligaments and tendons journal"] = "Muscles Ligaments Tendons J",
  ["musculoskeletal care"] = "Musculoskeletal Care",
  ["musculoskeletal regeneration"] = "Musculoskelet Regen",
  ["musculoskeletal science & practice"] = "Musculoskelet Sci Pract",
  ["musculoskeletal surgery"] = "Musculoskelet Surg",
  ["muse. annual of the museum of art and archaeology, university of missouri, columbia"] = "Muse",
  ["museen in köln. bulletin"] = "MusKoeln",
  ["musees de geneve"] = "Mus Geneve",
  ["musei e gallerie d’italia"] = "MusGalIt",
  ["musei ferraresi. bollettino annuale"] = "MusFerr",
  ["museo civico pontecorvo. quaderni"] = "QuadMusPontecorvo",
  ["museo de zaragoza, boletín"] = "BMZ",
  ["museo de zaragoza. boletín"] = "BMusZaragoza",
  ["museo in rivista. notiziario dei musei civici di pavia"] = "MusRiv",
  ["museum africum. west african journal of classical and related studies"] = "MusAfr",
  ["museum criticum"] = "MCr",
  ["museum haaretz, tel aviv. yearbook"] = "MusHaaretz",
  ["museum helveticum"] = "Mus Helv",
  ["museum helveticum: revue suisse pour l’étude de l’antiquité classique"] = "MH",
  ["museum history journal"] = "Mus Hist J",
  ["museum notes. american numismatic society"] = "ANSMusNotes",
  ["museum notes. the american numismatic society"] = "MusNotAmNumSoc",
  ["museum of northern arizona bulletin"] = "Mus North Ariz Bull",
  ["museum patavinum: rivista semestrale della facoltà di lettere e filosofia di padova"] = "MusPat",
  ["museum philologum londiniense"] = "MPhL",
  ["museum tusculanum"] = "MT",
  ["museum tusculanum. københavn"] = "MusTusc",
  ["museum. a quarterly review published by unesco"] = "MuseumUnesco",
  ["museums & social issues : a journal of reflective discourse"] = "Mus Soc Issues",
  ["musées de genève"] = "Mus Geneve",
  ["mushroom research"] = "Mushroom Res",
  ["music & letters"] = "Music Lett",
  ["music & science"] = "Music Sci (Lond)",
  ["music and medicine"] = "Music Med",
  ["music perception"] = "Music Percept",
  ["music therapy perspectives"] = "Music Ther Perspect",
  ["musicae scientiae : the journal of the european society for the cognitive sciences of music"] = "Music Sci",
  ["muslim world (hartford, conn.)"] = "Muslim World (Hartford)",
  ["musmat. brazilian journal of music and mathematics"] = "MusMat Braz. J. Music Math.",
  ["mutation research"] = "Mutat Res",
  ["mutation research - genetic toxicology and environmental mutagenesis"] = "Mutat. Res. Genet. Toxicol. Environ. Mutagen.",
  ["mutation research letters"] = "Mutat. Res. Lett.",
  ["mutation research, dna repair"] = "Mutat. Res. DNA Repair",
  ["mutation research, dna repair reports"] = "Mutat. Res. DNA Repair Rep.",
  ["mutation research, dnaging: genetic instability and aging"] = "Mutat. Res.  DNAging: Genet. Instab. Aging",
  ["mutation research, environmental mutagenesis and related subjects"] = "Mutat. Res. Environ. Mutagen. Relat. Subj.",
  ["mutation research, fundamental and molecular mechanisms of mutagenesis"] = "Mutat. Res. Fundam. Mol. Mech. Mutagen.",
  ["mutation research, genetic toxicology"] = "Mutat. Res. Genet. Toxicol.",
  ["mutation research, genetic toxicology and environmental mutagenesis"] = "Mutat. Res. Genet. Toxicol. Environ. Mutagen.",
  ["mutation research, genetic toxicology testing"] = "Mutat. Res. Genet. Toxicol. Test.",
  ["mutation research, mutation research genomics"] = "Mutat. Res.  Mutat. Res. Genomics",
  ["mutation research, reviews in genetic toxicology"] = "Mutat. Res. Rev. Genet. Toxicol.",
  ["mutation research, reviews in mutation research"] = "Mutat. Res. Rev. Mutat. Res.",
  ["mutation research. genetic toxicology and environmental mutagenesis"] = "Mutat Res Genet Toxicol Environ Mutagen",
  ["mutation research. reviews in mutation research"] = "Mutat Res Rev Mutat Res",
  ["muthanna journal of pure science"] = "Muthanna J Pure Sci",
  ["muzei i pametnici na kulturata"] = "MuzPamKul",
  ["múltunk : az mszmp központi bizottsága párttörténeti intézetének folyóirata"] = "Multunk",
  ["münchener medizinische wochenschrift (1950)"] = "Munch Med Wochenschr",
  ["mvp. modern veterinary practice"] = "MVP Mod Vet Pract",
  ["mycobacterial diseases : tuberculosis & leprosy"] = "Mycobact Dis",
  ["mycologia balcanica"] = "Mycol Balc",
  ["mycologia bavarica = bavarian journal of mycology"] = "Mycol Bavarica",
  ["mycologia helvetica"] = "Mycol. Helv.",
  ["mycologia iranica"] = "Mycol Iran",
  ["mycologia montenegrina"] = "Mycol Monten",
  ["mycological progress"] = "Mycol Prog",
  ["mycological research"] = "Mycol Res",
  ["mycopathologia et mycologia applicata"] = "Mycopathol Mycol Appl",
  ["mycosphere : journal of fungal biology"] = "Mycosphere",
  ["mycotoxin research"] = "Mycotoxin Res",
  ["mykosen. supplement"] = "Mykosen Suppl",
  ["mymensingh medical journal : mmj"] = "Mymensingh Med J",
  ["myrmecological news"] = "Myrmecol News",
  ["myrmecologische nachrichten"] = "Myrmecol Nachr",
  ["mysore journal of agricultural sciences"] = "Mysore J Agric Sci",
  ["mélanges de la casa de velázquez, madrid"] = "MCV",
  ["mélanges de la casa de velázquez. antiquité et moyen âge"] = "MelCasaVelazquez",
  ["mélanges de l’université saint-joseph"] = "MelBeyrouth",
  ["mélanges de l’école française de rome. antiquité"] = "MEFRA",
  ["mélanges de l’école française de rome. moyen âge et temps modernes"] = "MEFRM",
  ["mélanges de science religieuse"] = "MSR",
  ["mélanges d’archéologie et d’histoire de l’école française de rome, antiquité"] = "MEFRA",
  ["mémoires de la société mathématique de france"] = "Mém. Soc. Math. Fr. (N.S.)",
  ["mémoires de la société nationale des antiquaires de france"] = "MemAntFr",
  ["mémoires de la société vaudoise des sciences naturelles"] = "Mém. Soc. vaud. Sci. Nat.",
  ["mémoires de l’académie des inscriptions et belles lettres"] = "MAI",
  ["mémoires de l’académie des inscriptions et belles-lettres"] = "MemAcInscr",
  ["mémoires de l’institut national de france"] = "MemInstNatFr",
  ["mérida. ciudad y patrimonio"] = "Merida",
  ["mérida. excavaciones arqueológicas. memoria"] = "MeridaMem",
  ["méthodes et pratiques de l’ingénieur"] = "Méthodes Prat. Ingén.",
  ["méthodes mathématiques pour l’ingénieur"] = "Méthodes Math. Ingr.",
  ["métis. revue d’anthropologie du monde grec ancien"] = "Metis",
  ["münchener studien zur sprachwissenschaft"] = "MSS",
  ["münchner beiträge zur vor- und frühgeschichte"] = "MuenchBeitrVFG",
  ["münchner jahrbuch der bildenden kunst"] = "MueJb",
  ["münsterische beiträge zur antiken handelsgeschichte"] = "MBAH",
  ["münstersche beiträge zur geschichte und theorie der medizin"] = "Munstersche Beitr Gesch Theor Med",
  ["münzen und medaillen"] = "MuM",
  ["n & hc perspectives on community : official publication of the national league for nursing"] = "N HC Perspect Community",
  ["n and hc perspectives on community"] = "N HC Perspect. Community",
  ["n.m.o.s. bulletin"] = "NMOS Bull",
  ["naacog newsletter"] = "NAACOG Newsl",
  ["naacog technical bulletin"] = "NAACOG Tech Bull",
  ["naacog's clinical issues in perinatal and women's health nursing"] = "NAACOGS Clin Issu Perinat Womens Health Nurs",
  ["naacog’s clinical issues in perinatal and women’s health nursing"] = "NAACOGs Clin. Issu. Perinat. Womens Health Nurs.",
  ["naber report"] = "NABER Rep",
  ["nacdl journal"] = "NACDL J",
  ["nachrichten aus chemie technik und laboratorium"] = "Nachr. Chem. Tech. Lab.",
  ["nachrichten aus dem institut fur geschichte der naturwissenschaften hamburg"] = "Nachr Inst Gesch Naturwiss Hamburg",
  ["nachrichten aus dem institut für geschichte der naturwissenschaften hamburg"] = "Nachr Inst Gesch Naturwiss Hamburg",
  ["nachrichten aus der chemie"] = "Nachr. Chem.",
  ["nachrichten der akademie der wissenschaften in gottingen ii: mathematisch-physikalische klasse"] = "Nachr. Akad. Wiss. Gottingen II. Math. Phys. Kl.",
  ["nachrichten der akademie der wissenschaften in gottingen ii: mathematisch-physikalische klasse 1983"] = "Nachr. Akad. Wiss. Gottingen II: Math. Phys. Kl. 1983",
  ["nachrichten der akademie der wissenschaften in göttingen"] = "Nachr. Akad. Wiss. Göttingen Math.-Phys. Kl. II",
  ["nachrichten der akademie der wissenschaften in göttingen, philol.- hist. klasse"] = "NAWG",
  ["nachrichten der akademie der wissenschaften in göttingen. philologisch-historische klasse"] = "NAWG",
  ["nachrichten des entomologischen vereins apollo"] = "Nachr Entomol Ver Apollo",
  ["nachrichten von der akademie der wissenschaften in gottingen. philologisch-historische klasse"] = "Nachr Akad Wiss Gott Philol Hist Kl",
  ["nachrichten von der akademie der wissenschaften in göttingen. philologisch-historische klasse"] = "Nachr Akad Wiss Gott Philol Hist Kl",
  ["nachrichten von der gesellschaft der wissenschaften zu göttingen. philologisch-historische klasse"] = "NGWG",
  ["nachrichtenblatt arbeitskreis unterwasserarchäologie"] = "NachrArbUWA",
  ["nachrichtenblatt der bayerischen entomologen"] = "Nachr Bayer Entomol",
  ["nachrichtenblatt des deutschen pflanzenschutzdienstes"] = "Nachr Dtsch Pflanzenschutzd",
  ["nachrichtenblatt. deutsche gesellschaft fur geschichte der medizin, der naturwissenschaften und der technik"] = "Nachrichtenbl Dtsch Ges Gesch Med Naturwiss Tech",
  ["nachrichtenblatt. deutsche gesellschaft für geschichte der medizin, der naturwissenschaften und der technik"] = "Nachrichtenbl Dtsch Ges Gesch Med Naturwiss Tech",
  ["nachtrichtenblatt des deutschen pflanzenschutzdienstes"] = "Nachrichtenbl. Dtsch. Pflanzenschutzdienst",
  ["nacta journal"] = "NACTA J.",
  ["nacubo business officer : magazine of the national association of college and university business officers. national association of college and university business officers"] = "NACUBO Bus Off",
  ["nada (salisbury, zimbabwe)"] = "NADA",
  ["nadc-ac. united states. naval air development center, johnsville, pa. aeronautical computer laboratory"] = "NADC-AC",
  ["nadc-ma-. united states. naval air development center, johnsville, pa. aviation medical acceleration laboratory"] = "NADC MA United States Nav Air Dev Cen Johnsville Pa Aviat Med Accel Lab",
  ["nadc-ml [reports]. united states. naval air development center, johnsville, pa. aviation medical acceleration laboratory"] = "NADC-ML Rep",
  ["nadc-mr [reports]. united states. naval air development center, johnsville, pa. aerospace medical research department"] = "NADC-MR Rep",
  ["nadc-mr [reports]. united states. naval air development center, johnsville, pa. aerospace medical research dept"] = "NADC-MR Rep",
  ["nadl journal"] = "NADL J",
  ["nagasaki igakkai zasshi = nagasaki medical journal"] = "Nagasaki Igakkai Zasshi",
  ["nagoya gakuin daigaku ronshu. igaku, kenko kagaku, supotsu kagakuhen"] = "Nagoya Gakuin Daigaku Ronshu Igaku Kenko Kagaku Supotsu Kagakuhen",
  ["nagoya journal of medical science"] = "Nagoya J Med Sci",
  ["nagoya mathematical journal"] = "Nagoya Math. J.",
  ["nagoya medical journal"] = "Nagoya Med J",
  ["naham access management journal"] = "NAHAM Access Manag J",
  ["naham management journal"] = "NAHAM Manage. J.",
  ["naika (internal medicine)"] = "Naika",
  ["naika hokan (japanese archives of internal medicine)"] = "Naika Hokan",
  ["naika hokan. japanese archives of internal medicine"] = "Naika Hokan",
  ["naika. internal medicine"] = "Naika",
  ["nakadori byoin iho. nakadori hospital medical bulletin]"] = "Nakadori Byoin Iho",
  ["nan fang yi ke da xue xue bao (journal of southern medical university)"] = "Nan Fang Yi Ke Da Xue Xue Bao",
  ["nan fang yi ke da xue xue bao = journal of southern medical university"] = "Nan Fang Yi Ke Da Xue Xue Bao",
  ["nan kai da xue xue bao. zi ran ke xue ban"] = "Acta scientiarum naturalium Universitatis Nankaiensis|Nan Kai Da Xue Xue Bao Zi Ran Ke Xue Ban",
  ["nan kai da xue xue bao. zi ran ke xue ban = acta scientiarum naturalium universitatis nankaiensis"] = "Nan Kai Da Xue Xue Bao Zi Ran Ke Xue Ban",
  ["nan nü : men, women, and gender in early and imperial china"] = "Nan Nu",
  ["nan xing xue za zhi = chinese journal of andrology"] = "Nan Xing Xue Za Zhi",
  ["nan yang wen ti"] = "Nan Yang Wen Ti",
  ["nan yang wen ti yan qiu = southeast asian affairs"] = "Nanyang Wenti Yenchiu",
  ["nanfang nongye xuebao = journal of southern agriculture"] = "Nanfang Nongye Xuebao",
  ["nanfang shuichan kexue"] = "Nanfang Shuichan Kexue",
  ["nanhua daxue xuebao. ziran kexue ban = journal of nanhua university science and technology"] = "Nanhua Daxue Xuebao Ziran Kexue Ban",
  ["nanjing da xue xue bao. zhe xue, ren wen ke xue, she hui ke xue = journal of nanjing university"] = "Nanching Ta Hsueh Hsueh Pao",
  ["nanjing daxue xuebao"] = "Nanjing Daxue Xuebao Shuxue Bannian Kan",
  ["nanjing daxue xuebao. shuxue bannian kan. nanjing university. journal. mathematical biquarterly"] = "Nanjing Daxue Xuebao Shuxue Bannian Kan",
  ["nanjing nong ye da xue xue bao = journal of nanjing agricultural university"] = "Nanjing Nong Ye Da Xue Xue Bao",
  ["nanjing shi da xue bao. zi ran ke xue ban"] = "Nanjing shida xuebao|Nanjing Shi Da Xue Bao",
  ["nanjing shi da xue bao. zi ran ke xue ban = nanjing shida xuebao"] = "Nanjing Shi Da Xue Bao",
  ["nankai series in pure, applied mathematics and theoretical physics"] = "Nankai Ser. Pure Appl. Math. Theoret. Phys.",
  ["nankai tracts in mathematics"] = "Nankai Tracts Math.",
  ["nano biomedicine and engineering"] = "Nano Biomed Eng",
  ["nano communication networks"] = "Nano Commun. Networks",
  ["nano convergence"] = "Nano Converg",
  ["nano energy"] = "Nano Energy",
  ["nano express"] = "Nano Express",
  ["nano futures"] = "Nano Futures",
  ["nano letters"] = "Nano Lett.",
  ["nano life"] = "Nano Life",
  ["nano materials science"] = "Nano Mater. Sci.",
  ["nano micro letters"] = "Nano Micro Lett.",
  ["nano micro mechanics review"] = "Nano Micro Mech. Rev.",
  ["nano research"] = "Nano Res.",
  ["nano reviews"] = "Nano Rev",
  ["nano reviews & experiments"] = "Nano Rev Exp",
  ["nano science and nano technology"] = "Nano Sci. Nano Technol.",
  ["nano science and nano technology: an indian journal"] = "Nano Sci. Nano Technol.: Indian J.",
  ["nano select"] = "Nano Select",
  ["nano today"] = "Nano Today",
  ["nano-micro letters"] = "Nanomicro Lett",
  ["nano-structures & nano-objects"] = "Nano-Struct. Nano-Objects",
  ["nanobiotechnology : the journal at the intersection of nanotechnology, molecular biology, and biomedical sciences"] = "Nanobiotechnology",
  ["nanobiotechnology reports"] = "Nanobiotechnol. Rep.",
  ["nanomanufacturing and metrology"] = "Nanomanuf. Metrol.",
  ["nanomaterials (basel, switzerland)"] = "Nanomaterials (Basel)",
  ["nanomaterials and energy"] = "Nanomater. Energy",
  ["nanomaterials and nanotechnology"] = "Nanomater. Nanotechnol.",
  ["nanomaterials and the environment"] = "Nanomater Environ",
  ["nanomechanics science and technology: an international journal"] = "Nanomech. Sci. Technol.: Int. J.",
  ["nanomedicine & nanoscience research"] = "Nanomed Nanosci Res",
  ["nanomedicine (london, england)"] = "Nanomedicine (Lond)",
  ["nanomedicine (london, united kingdom)"] = "Nanomedicine (London, U. K.)",
  ["nanomedicine (new york, ny, united states)"] = "Nanomedicine (N. Y., NY, U. S.)",
  ["nanomedicine : design and applications of magnetic nanomaterials, nanosensors, and nanosystems"] = "Nanomedicine (Chichester)",
  ["nanomedicine : nanotechnology, biology, and medicine"] = "Nanomedicine",
  ["nanomedicine journal"] = "Nanomed J",
  ["nanomedicine research journal"] = "Nanomed. Res. J.",
  ["nanomedicine: nanotechnology, biology and medicine"] = "Nanomed. Nanotechnol. Biol. Med.",
  ["nanoscale advances"] = "Nanoscale Adv.",
  ["nanoscale and microscale thermophysical engineering"] = "Nanoscale Microscale Thermophys. Eng.",
  ["nanoscale horizons"] = "Nanoscale Horiz",
  ["nanoscale research letters"] = "Nanoscale Res Lett",
  ["nanoscale systems. mathematical modeling, theory and applications"] = "Nanoscale Syst. Math. Model. Theory Appl.",
  ["nanoscale systems: mathematical modeling, theory and applications"] = "Nanoscale Syst.: Math. Model. Theory Appl.",
  ["nanoscience and nanoengineering"] = "Nanosci Nanoeng",
  ["nanoscience and nanotechnolog letters"] = "Nanosci. Nanotechnol. Lett.",
  ["nanoscience and nanotechnology - asia"] = "Nanosci. Nanotechnol. - Asia",
  ["nanoscience and nanotechnology letters (print)"] = "Nanosci Nanotechnol Lett",
  ["nanoscience and technology"] = "NanoSci. Technol.",
  ["nanoscience and technology: an international journal"] = "Nanosci. Technol.: Int. J.",
  ["nanosistemi nanomateriali nanotehnologii"] = "Nanosist. Nanomater. Nanotehnol.",
  ["nanostructure science and technology"] = "Nanostruct. Sci. and Tech.",
  ["nanostructured materials"] = "Nanostruct. Mater.",
  ["nanotechnologies in construction"] = "Nanotechnol. Construct.",
  ["nanotechnologies in russia"] = "Nanotechnol. Russ.",
  ["nanotechnology and precision engineering"] = "Nanotechnol. Precis. Eng.",
  ["nanotechnology for environmental engineering"] = "Nanotechnol. Environ. Eng.",
  ["nanotechnology perceptions"] = "Nanotechnol Percept",
  ["nanotechnology reviews"] = "Nanotechnol Rev",
  ["nanotechnology, science and applications"] = "Nanotechnol Sci Appl",
  ["nanoworld journal"] = "NanoWorld J",
  ["nanr news"] = "NANR News",
  ["napa bulletin"] = "NAPA Bull",
  ["napht news"] = "NAPHT News",
  ["napoli nobilissima"] = "NapNobil",
  ["napt bulletin"] = "NAPT Bull",
  ["nar cancer"] = "NAR Cancer",
  ["nar genomics and bioinformatics"] = "NAR Genom Bioinform",
  ["naresa monograph"] = "NARESA Mongr",
  ["narodno zdravlje"] = "Nar Zdrav",
  ["narody azii i afriki"] = "Nar Azii Afr",
  ["narrative inquiry : ni"] = "Narrat Inq",
  ["narrative inquiry in bioethics"] = "Narrat Inq Bioeth",
  ["nasa contractor report. nasa cr. united states. national aeronautics and space administration"] = "NASA Contract Rep NASA CR",
  ["nasa monographs in systems and software engineering"] = "NASA Monogr. Syst. Softw. Eng.",
  ["nasa technical memorandum"] = "NASA Tech Memo",
  ["naselenie (sofia, bulgaria : 1983)"] = "Naselenie",
  ["nasn school nurse (print)"] = "NASN Sch Nurse",
  ["naspcp newsletter / national aids/std prevention and control program. national aids/std prevention and control program (philippines)"] = "NASPCP Newsl",
  ["naspcp newsletter. national aids/std prevention and control program (philippines)"] = "NASPCP Newsl",
  ["nassau medical news"] = "Nassau Med News",
  ["nassauische annalen"] = "NassAnn",
  ["nasu suteshon (journal of nursing administration for patients)"] = "Nasu Suteshon",
  ["nasza przeszlosc"] = "Nasza Przeslosc",
  ["nasza przeszłość"] = "Nasza Przeslosc",
  ["nation (new york, n.y. : 1865)"] = "Nation",
  ["nation's business"] = "Nations Bus",
  ["nation's health (washington, d.c.)"] = "Nations Health",
  ["national academy of science letters"] = "Nat. Acad. Sci. Lett.",
  ["national academy science letters"] = "Natl. Acad. Sci. Lett.",
  ["national academy science letters (india)"] = "Natl. Acad. Sci. Lett. (India)",
  ["national academy science letters. national academy of sciences, india"] = "Natl Acad Sci Lett",
  ["national aids bulletin"] = "Natl AIDS Bull",
  ["national aids bulletin / australian federation of aids organisations inc"] = "Natl AIDS Bull",
  ["national black law journal"] = "Natl Black Law J",
  ["national bureau of economic research bulletin on aging and health"] = "Natl Bur Econ Res Bull Aging Health",
  ["national bureau of standards (u.s.), circular"] = "Natl. Bur. Stand. Circ. (U.S.)",
  ["national bureau of standards (u.s.), miscellaneous publication"] = "Natl. Bur. Stand. Misc. Publ. (U.S.)",
  ["national bureau of standards (u.s.), special publication"] = "Natl. Bur. Stand. Spec. Publ. (U.S.)",
  ["national cancer institute carcinogenesis technical report series"] = "Natl Cancer Inst Carcinog Tech Rep Ser",
  ["national cancer institute monograph"] = "Natl Cancer Inst Monogr",
  ["national cancer institute monographs"] = "Natl. Cancer Inst. Monogr.",
  ["national cancer institute research report"] = "Natl Cancer Inst Res Rep",
  ["national catholic reporter"] = "Natl Cathol Report",
  ["national civic review"] = "Natl Civ Rev",
  ["national clearinghouse for poison control centers bulletin"] = "Natl. Clgh. Poison Control Cent. Bull.",
  ["national conference on dental public relations"] = "Natl Conf Dent Public Relat",
  ["national conference on women & hiv : innovation for care, policy and prevention. national conference on women & hiv: innovation for care, policy and prevention (1997 : los angeles, calif.)"] = "Natl Conf Women HIV Natl Conf Women HIV Innov Care Policy Prev 1997 Los Angel Calif",
  ["national dental association journal"] = "Natl Dent Assoc J",
  ["national dental health conference"] = "Natl. Dent. Health Conf.",
  ["national engineer"] = "Natl. Eng.",
  ["national family health survey bulletin"] = "Natl Fam Health Surv Bull",
  ["national federation of catholic physicians' guilds newsletter"] = "Natl Fed Cathol Physicians Guild Newsl",
  ["national federation of catholic physicians’ guilds newsletter"] = "Natl Fed Cathol Physicians Guild Newsl",
  ["national forum"] = "Natl Forum",
  ["national forum on hospital and health affairs"] = "Natl. Forum Hosp. Health Aff.",
  ["national geographic"] = "Natl Geogr Mag",
  ["national geographic magazine"] = "Nat. Geogr. Mag.",
  ["national geographic research"] = "Natl Geogr Res",
  ["national health statistics reports"] = "Natl Health Stat Report",
  ["national health; a monthly journal devoted to the interests of public health and hygiene, sanitation and sewage purification, town planning, housing, parks and gardens"] = "Natl Health",
  ["national hospital and health care"] = "Natl Hosp Health Care",
  ["national institute economic review"] = "Natl Inst Econ Rev",
  ["national institute of animal health quarterly"] = "Natl Inst Anim Health Q (Tokyo)",
  ["national institute on drug abuse research monograph series"] = "Natl Inst Drug Abuse Res Monogr Ser",
  ["national institutes of health consensus development conference consensus statement"] = "Natl Inst Health Consens Dev Conf Consens Statement",
  ["national institutes of health consensus development conference summaries"] = "Natl Inst Health Consens Dev Conf Summ",
  ["national institutes of health consensus development conference summary"] = "Natl Inst Health Consens Dev Conf Summ",
  ["national jewish law review"] = "Natl Jew Law Rev",
  ["national journal"] = "Natl J (Wash)",
  ["national journal of integrated research in medicine"] = "Natl J Integr Res Med",
  ["national journal of maxillofacial surgery"] = "Natl J Maxillofac Surg",
  ["national journal of physiology, pharmacy and pharmacology"] = "Natl J Physiol Pharm Pharmacol",
  ["national journal of sociology : njs"] = "Natl J Sociol",
  ["national medical care utilization and expenditure survey (series). series b, descriptive report"] = "Natl Med Care Util Expend Surv B",
  ["national medical care utilization and expenditure survey (series). series b, descriptive report."] = "Natl Med Care Util Expend Surv B",
  ["national medical care utilization and expenditure survey (series). series c, analytical report"] = "Natl Med Care Util Expend Surv C",
  ["national medical care utilization and expenditure survey (series). series c, analytical report."] = "Natl Med Care Util Expend Surv C",
  ["national medical care utilization and expenditure survey. series b, descriptive report"] = "Natl. Med. Care Util. Expend. Surv. B.",
  ["national medical care utilization and expenditure survey. series c, analytical report"] = "Natl. Med. Care Util. Expend. Surv. C.",
  ["national medical journal of india"] = "Natl. Med. J. India",
  ["national medical-legal journal"] = "Natl Med Leg J",
  ["national negro health news"] = "Natl Negro Health News",
  ["national network"] = "Natl. Netw.",
  ["national network (dallas, tex.)"] = "Natl Netw",
  ["national nosocomial infections study report"] = "Natl Nosocomial Infect Study Rep",
  ["national observer (washington, d.c.)"] = "Natl Obs",
  ["national population news bulletin : newsletter of the national population council"] = "Natl Popul News Bull",
  ["national report on subacute care"] = "Natl Rep Subacute Care",
  ["national review"] = "Natl Rev",
  ["national safety congress transactions"] = "Natl. Saf. Congr. Trans.",
  ["national safety news"] = "Natl Saf News",
  ["national science open"] = "Natl. Sci. Open",
  ["national science review"] = "Natl. Sci. Rev.",
  ["national tax journal"] = "Natl Tax J",
  ["national toxicology program genetically modified model report"] = "Natl Toxicol Program Genet Modif Model Rep",
  ["national toxicology program technical report series"] = "Natl Toxicol Program Tech Rep Ser",
  ["national underwriter"] = "Natl. Underwrit. [Life Health]",
  ["national university law review"] = "Natl Univ Law Rev",
  ["national vital statistics reports"] = "Natl. Vital Stat. Rep.",
  ["national vital statistics reports : from the centers for disease control and prevention, national center for health statistics, national vital statistics system"] = "Natl Vital Stat Rep",
  ["national weather digest"] = "Natl. Wea. Dig.",
  ["national westminster bank quarterly review"] = "Nat. Westminster Bank Quart. Rev.",
  ["national wildlife"] = "Natl Widl",
  ["national women's health report"] = "Natl Womens Health Rep",
  ["national women’s health report"] = "Natl Womens Health Rep",
  ["nationalism & ethnic politics"] = "Nationalism Ethn Polit",
  ["nationalities papers"] = "Natl Pap",
  ["nationalokonomisk tidsskrift : maanedskrift for samfundssporgsmaal, okonomi og handel"] = "Nationalokon Tidsskr",
  ["nationalpark-forschung in der schweiz"] = "Natl.park-Forsch. Schweiz",
  ["nationaløkonomisk tidsskrift"] = "Nationaløkon. Tidsskr.",
  ["nationaløkonomisk tidsskrift : maanedskrift for samfundsspørgsmaal, økonomi og handel"] = "Nationalokon Tidsskr",
  ["nations and nationalism"] = "Nations Natl",
  ["nation’s business"] = "Nations Bus.",
  ["nato advanced science institutes series a: life sciences"] = "NATO Adv. Sci. Inst. Ser. A Life Sci.",
  ["nato advanced science institutes series b: physics"] = "NATO Adv. Sci. Inst. Ser. B Phys.",
  ["nato advanced science institutes series c: mathematical and physical sciences"] = "NATO Adv. Sci. Inst. Ser. C Math. Phys. Sci.",
  ["nato advanced science institutes series e: applied sciences"] = "NATO Adv. Sci. Inst. Ser. E Appl. Sci.",
  ["nato advanced science institutes series f: computer and systems sciences"] = "NATO Adv. Sci. Inst. Ser. F Comput. Systems Sci.",
  ["nato advanced science institutes series, series i: global environmental change"] = "NATO ASI Ser. Ser. I Glob. Environ. Change",
  ["nato advanced study institutes series. series a, life sciences"] = "NATO Adv Stud Inst Ser A Life Sci",
  ["nato asi series, series a life sciences"] = "NATO ASI Ser., Ser. A",
  ["nato asi series, series a: life sciences"] = "NATO ASI Ser., Ser. A",
  ["nato asi series, series b physics"] = "NATO ASI Ser., Ser. B",
  ["nato asi series, series b: physics"] = "NATO ASI Ser., Ser. B",
  ["nato asi series, series c mathematical and physical sciences"] = "NATO ASI Ser., Ser. C",
  ["nato asi series, series c: mathematical and physical sciences"] = "NATO ASI Ser., Ser. C",
  ["nato asi series, series e applied physics"] = "NATO ASI Ser., Ser. E",
  ["nato asi series, series e: applied physics"] = "NATO ASI Ser., Ser. E",
  ["nato asi series, series g ecological sciences"] = "NATO ASI Ser., Ser. G",
  ["nato asi series, series g: ecological sciences"] = "NATO ASI Ser., Ser. G",
  ["nato asi series, series h cell biology"] = "NATO ASI Ser., Ser. H",
  ["nato asi series, series h: cell biology"] = "NATO ASI Ser., Ser. H",
  ["nato asi series, series i: global environmental change"] = "NATO ASI Ser., Ser. I",
  ["nato asi series. series a, life sciences"] = "NATO ASI Ser A Life Sci",
  ["nato asi series. series h, cell biology"] = "NATO ASI Ser H Cell Biol",
  ["nato challenges of modern society"] = "NATO Challenges Modern Soc.",
  ["nato conference series iv: marine sciences"] = "NATO Conf. Ser. IV: Marine Sci.",
  ["nato conference series vi: materials science"] = "NATO Conf. Ser. VI: Materials Sci.",
  ["nato science for peace and security series d: information and communication security"] = "NATO Sci. Peace Secur. Ser. D Inf. Commun. Secur.",
  ["natonal report on subacute care"] = "Natl. Rep. Subacute Care",
  ["natsional\\cprime naya akademiya nauk armenii"] = "Izv. Nats. Akad. Nauk Armenii Mat.",
  ["natsional\\cprime naya akademiya nauk ukrainy"] = "Gidromekh. (Kiev)",
  ["natsionalna akademiya nauk ukraini. institut matematiki. ukrainskii matematichnii zhurnal"] = "Ukrain. Mat. Zh.",
  ["natsionalnaya akademiya nauk armenii. izvestiya. matematika"] = "Izv. Nats. Akad. Nauk Armenii Mat.",
  ["natsionalnaya akademiya nauk ukrainy. institut kibernetiki im. v. m. glushkova. kibernetika i sistemnyi analiz"] = "Kibernet. Sistem. Anal.",
  ["natsionalnaya akademiya nauk ukrainy. institut prikladnoi matematiki i mekhaniki. mekhanika tverdogo tela"] = "Mekh. Tverd. Tela",
  ["natsionalnaya akademiya nauk ukrainy. problemy upravleniya i informatiki"] = "Problemy Upravlen. Inform.",
  ["natur und heimat (münster)"] = "Nat Heim (Munst)",
  ["natur und landschaft"] = "Nat. Landsch.",
  ["natur und mensch"] = "Nat. Mensch",
  ["natur und museum"] = "Nat Mus",
  ["natur und naturschutz in mecklenburg"] = "Nat. Nat.schutz Mecklenbg.",
  ["natur und philosophie"] = "Nat. Philos.",
  ["natur und recht"] = "Natur Recht",
  ["natur und technik"] = "Nat. Tech.",
  ["natur- und landschaftsschutzgebiete baden-württembergs"] = "Nat.- Landsch.schutzgeb. Baden-Württ.",
  ["natura croatica : periodicum musei historiae naturalis croatici"] = "casopis Hrvatskoga prirodoslovnog muzeja|Nat. Croat.",
  ["natura croatica : periodicum musei historiae naturalis croatici = časopis hrvatskoga prirodoslovnog muzeja"] = "Nat Croat",
  ["natural and engineering sciences"] = "Nat Eng Sci",
  ["natural areas journal"] = "Nat. Areas J.",
  ["natural communications with computers"] = "Natural Comm. Comput.",
  ["natural computing"] = "Nat Comput",
  ["natural computing series"] = "Nat. Comput. Ser.",
  ["natural computing. an international journal"] = "Nat. Comput.",
  ["natural gas & electricity"] = "Nat. Gas Electr.",
  ["natural gas industry b"] = "Nat. Gas Ind. B",
  ["natural hazards"] = "Nat. Hazard.",
  ["natural hazards (dordrecht, netherlands)"] = "Nat Hazards (Dordr)",
  ["natural hazards and earth system sciences"] = "Nat. Hazards Earth Syst. Sci.",
  ["natural hazards and earth system sciences discussions"] = "Nat. Hazards Earth Syst. Sci. Discuss.",
  ["natural hazards research"] = "Nat. Hazard. Res.",
  ["natural hazards review"] = "Nat. Hazard. Rev.",
  ["natural history"] = "Nat Hist",
  ["natural history collections and museomics"] = "Nat. Hist. Collect. Museomics",
  ["natural history research. special issue"] = "Nat Hist Res Spec Issue",
  ["natural history sciences : atti della società italiana di scienze naturali e del museo civico di storia naturale in milano"] = "Nat Hist Sci",
  ["natural immunity"] = "Nat Immun",
  ["natural immunity and cell growth regulation"] = "Nat Immun Cell Growth Regul",
  ["natural language & linguistic theory"] = "Nat Lang Linguist Theory",
  ["natural language engineering"] = "Nat Lang Eng",
  ["natural language processing journal"] = "Nat. Lang. Process. J.",
  ["natural law forum"] = "Nat Law Forum",
  ["natural product communications"] = "Nat Prod Commun",
  ["natural product letters"] = "Nat Prod Lett",
  ["natural product reports"] = "Nat Prod Rep",
  ["natural product research"] = "Nat Prod Res",
  ["natural product research part a: structure and synthesis"] = "Nat. Prod. Res., Part A",
  ["natural product research part b: bioactive natural products"] = "Nat. Prod. Res. Part B",
  ["natural product research, part a: structure and synthesis"] = "Nat. Prod. Res., Part A",
  ["natural product research, part b: bioactive natural products"] = "Nat. Prod. Res., Part B",
  ["natural product sciences"] = "Nat. Prod. Sci.",
  ["natural products and bioprospecting"] = "Nat Prod Bioprospect",
  ["natural products chemistry & research"] = "Nat Prod Chem Res",
  ["natural products communications"] = "Nat. Prod. Commun.",
  ["natural products journal"] = "Nat. Prod. J.",
  ["natural products letters"] = "Nat. Prod. Lett.",
  ["natural products reports"] = "Nat. Prod. Rep.",
  ["natural products research part a: structure and synthesis"] = "Nat. Prod. Res., Part A",
  ["natural products research part b: bioactive natural products"] = "Nat. Prod. Res. Part B",
  ["natural resource modeling"] = "Nat. Resour. Model.",
  ["natural resources & engineering"] = "Nat. Resour. Eng.",
  ["natural resources forum"] = "Nat Resour Forum",
  ["natural resources journal"] = "Nat Resour J",
  ["natural resources research"] = "Nat. Resour. Res.",
  ["natural rubber research"] = "Nat. Rubber Res.",
  ["natural science"] = "Nat Sci (Irvine)",
  ["natural science journal"] = "Nat. Sci. J.",
  ["natural science journal of xiangtan university"] = "Natur. Sci. J. Xiangtan Univ.",
  ["natural science report of the ochanomizu university"] = "Natur. Sci. Rep. Ochanomizu Univ.",
  ["natural sciences"] = "Nat. Sci.",
  ["natural sciences education"] = "Nat. Sci. Educ.",
  ["natural sciences journal of harbin normal university"] = "Natur. Sci. J. Harbin Normal Univ.",
  ["natural structural biology"] = "Nat. Struct. Biol.",
  ["natural toxins"] = "Nat Toxins",
  ["naturalia monspeliensia. serie botanique"] = "Nat Monspel Ser Bot",
  ["naturalia monspeliensia. série botanique"] = "Nat Monspel Ser Bot",
  ["naturalist, the"] = "Naturalist",
  ["naturaliste canadien (québec)"] = "Nat. Can. (Qué.)",
  ["naturalistes belges, les"] = "Nat. Belg.",
  ["nature (london)"] = "Nature (London)",
  ["nature (london, united kingdom)"] = "Nature (London, U. K.)",
  ["nature (paris, france)"] = "Nature (Paris, France)",
  ["nature africa"] = "Nat. Afr.",
  ["nature aging"] = "Nat Aging",
  ["nature and environment"] = "Nat. environ.",
  ["nature and human activities"] = "Nat Hum Activities",
  ["nature and life in southeast asia"] = "Nat. Life Southeast Asia",
  ["nature and resources"] = "Nat. Resources",
  ["nature and science"] = "Nat Sci (East Lansing)",
  ["nature and science of sleep"] = "Nat Sci Sleep",
  ["nature and system"] = "Nat Syst",
  ["nature astronomy"] = "Nat Astron",
  ["nature biomedical engineering"] = "Nat Biomed Eng",
  ["nature biotechnology"] = "Nat Biotechnol",
  ["nature cancer"] = "Nat Cancer",
  ["nature cardiovascular research"] = "Nat. Cardiovasc. Res.",
  ["nature catalysis"] = "Nat. Catal.",
  ["nature cell biology"] = "Nat Cell Biol",
  ["nature chemical biology"] = "Nat Chem Biol",
  ["nature chemical engineering"] = "Nat. Chem. Eng.",
  ["nature chemistry"] = "Nat Chem",
  ["nature cities"] = "Nat. Cities",
  ["nature climate change"] = "Nat Clim Chang",
  ["nature clinical practice. cardiovascular medicine"] = "Nat Clin Pract Cardiovasc Med",
  ["nature clinical practice. endocrinology & metabolism"] = "Nat Clin Pract Endocrinol Metab",
  ["nature clinical practice. endocrinology and metabolism"] = "Nat. Clin. Pract. Endocrinol. Metab.",
  ["nature clinical practice. gastroenterology & hepatology"] = "Nat Clin Pract Gastroenterol Hepatol",
  ["nature clinical practice. gastroenterology and hepatology"] = "Nat. Clin. Pract. Gastroenterol. Hepatol.",
  ["nature clinical practice. nephrology"] = "Nat Clin Pract Nephrol",
  ["nature clinical practice. neurology"] = "Nat Clin Pract Neurol",
  ["nature clinical practice. oncology"] = "Nat Clin Pract Oncol",
  ["nature clinical practice. rheumatology"] = "Nat Clin Pract Rheumatol",
  ["nature clinical practice. urology"] = "Nat Clin Pract Urol",
  ["nature communications"] = "Nat Commun",
  ["nature computational science"] = "Nat Comput Sci",
  ["nature conservation"] = "Nat Conserv",
  ["nature conservation bulgaria"] = "Nat. Conserv. Bulg.",
  ["nature conservation research = zapovednai︠a︡ nauka"] = "Nat Conserv Res",
  ["nature digest"] = "Nat. Dig.",
  ["nature ecology & evolution"] = "Nat Ecol Evol",
  ["nature electronics"] = "Nat Electron",
  ["nature energy"] = "Nat. Energy",
  ["nature et progrès"] = "Nat. Prog.",
  ["nature food"] = "Nat Food",
  ["nature genetics"] = "Nat Genet",
  ["nature geoscience"] = "Nat. Geosci.",
  ["nature health"] = "Nat. Health",
  ["nature human behaviour"] = "Nat Hum Behav",
  ["nature immunology"] = "Nat Immunol",
  ["nature india"] = "Nat. Ind.",
  ["nature information"] = "Nat. Inf.",
  ["nature italy"] = "Nat. Italy",
  ["nature journal (opole)"] = "Nat J (Opole)",
  ["nature machine intelligence"] = "Nat Mach Intell",
  ["nature materials"] = "Nat Mater",
  ["nature medicine"] = "Nat Med",
  ["nature medicine (new york, ny, united states)"] = "Nat. Med. (N. Y., NY, U. S.)",
  ["nature mental health"] = "Nat. Ment. Health",
  ["nature metabolism"] = "Nat Metab",
  ["nature methods"] = "Nat Methods",
  ["nature microbiology"] = "Nat Microbiol",
  ["nature nano"] = "Nat. Nano",
  ["nature nanotechnology"] = "Nat Nanotechnol",
  ["nature neuroscience"] = "Nat Neurosci",
  ["nature photonics"] = "Nat. Photonics",
  ["nature physics"] = "Nat. Phys.",
  ["nature plants"] = "Nat Plants",
  ["nature protocols"] = "Nat Protoc",
  ["nature reports. stem cells"] = "Nat Rep Stem Cells",
  ["nature reviews biodiversity"] = "Nat. Rev. Biodivers.",
  ["nature reviews bioengineering"] = "Nat. Rev. Bioeng.",
  ["nature reviews cancer"] = "Nat. Rev. Cancer",
  ["nature reviews cardiology"] = "Nat. Rev. Cardiol.",
  ["nature reviews chemistry"] = "Nat. Rev. Chem.",
  ["nature reviews clean technology"] = "Nat. Rev. Clean Technol.",
  ["nature reviews clinical oncology"] = "Nat. Rev. Clin. Oncol.",
  ["nature reviews disease primers"] = "Nat. Rev. Dis. Primers",
  ["nature reviews drug discovery"] = "Nat. Rev. Drug Discovery",
  ["nature reviews earth and environment"] = "Nat. Rev. Earth Environ.",
  ["nature reviews electrical engineering"] = "Nat. Rev. Electr. Eng.",
  ["nature reviews endrocrinology"] = "Nat. Rev. Endocrinol.",
  ["nature reviews gastroenterology & hepatology"] = "Nat. Rev. Gastroenterol. Hepatol.",
  ["nature reviews genetics"] = "Nat. Rev. Genet.",
  ["nature reviews immunology"] = "Nat. Rev. Immunol.",
  ["nature reviews materials"] = "Nat. Rev. Mater.",
  ["nature reviews methods primers"] = "Nat. Rev. Methods Primers",
  ["nature reviews microbiology"] = "Nat. Rev. Microbiol.",
  ["nature reviews molecular cell biology"] = "Nat. Rev. Mol. Cell Biol.",
  ["nature reviews nephrology"] = "Nat. Rev. Nephrol.",
  ["nature reviews neurology"] = "Nat. Rev. Neurol.",
  ["nature reviews neuroscience"] = "Nat. Rev. Neurosci.",
  ["nature reviews physics"] = "Nat. Rev. Phys.",
  ["nature reviews psychology"] = "Nat. Rev. Psychol.",
  ["nature reviews rheumatology"] = "Nat. Rev. Rheumatol.",
  ["nature reviews urology"] = "Nat. Rev. Urol.",
  ["nature reviews. cancer"] = "Nat Rev Cancer",
  ["nature reviews. cardiology"] = "Nat Rev Cardiol",
  ["nature reviews. chemistry"] = "Nat Rev Chem",
  ["nature reviews. clinical oncology"] = "Nat Rev Clin Oncol",
  ["nature reviews. disease primers"] = "Nat Rev Dis Primers",
  ["nature reviews. drug discovery"] = "Nat Rev Drug Discov",
  ["nature reviews. earth & environment"] = "Nat Rev Earth Environ",
  ["nature reviews. endocrinology"] = "Nat Rev Endocrinol",
  ["nature reviews. gastroenterology & hepatology"] = "Nat Rev Gastroenterol Hepatol",
  ["nature reviews. genetics"] = "Nat Rev Genet",
  ["nature reviews. immunology"] = "Nat Rev Immunol",
  ["nature reviews. materials"] = "Nat Rev Mater",
  ["nature reviews. microbiology"] = "Nat Rev Microbiol",
  ["nature reviews. molecular cell biology"] = "Nat Rev Mol Cell Biol",
  ["nature reviews. nephrology"] = "Nat Rev Nephrol",
  ["nature reviews. neurology"] = "Nat Rev Neurol",
  ["nature reviews. neuroscience"] = "Nat Rev Neurosci",
  ["nature reviews. rheumatology"] = "Nat Rev Rheumatol",
  ["nature reviews. urology"] = "Nat Rev Urol",
  ["nature sensors"] = "Nat. Sens.",
  ["nature structural & molecular biology"] = "Nat Struct Mol Biol",
  ["nature structural and molecular biology"] = "Nat. Struct. Mol. Biol.",
  ["nature structural biology"] = "Nat Struct Biol",
  ["nature sustainability"] = "Nat Sustain",
  ["nature synthesis"] = "Nat. Synth.",
  ["nature water"] = "Nat. Water",
  ["nature, environment and pollution technology"] = "Nat. Environ. Pollut. Technol.",
  ["nature, london"] = "Nature (London)",
  ["nature, paris"] = "Nature (Paris)",
  ["nature. new biology"] = "Nature. New Biol.",
  ["nature: new biology"] = "Nat New Biol",
  ["natures sciences societes"] = "Nat. Sci. Soc.",
  ["natureza on line"] = "Nat On Line",
  ["naturschutz heute"] = "Nat.schutz heute",
  ["naturschutz und landschaftspflege in hamburg"] = "Nat.schutz Landsch.pfl. Hambg.",
  ["naturschutz und landschaftspflege in niedersachsen"] = "Nat.schutz Landsch.pfl. Niedersachs.",
  ["naturschutz und landschaftsplanung"] = "Naturschutz Landschaftsplan.",
  ["naturwissenschaft im unterricht chemie"] = "Naturwiss. Unterr. Chem.",
  ["naturwissenschaft im unterricht physik"] = "Naturwiss. Unterr. Phys.",
  ["naturwissenschaften, die"] = "Naturwissenschaften",
  ["naturwissenschaftliche rundschau"] = "Naturwiss Rundsch",
  ["naturwissenschaftliche wochenschrift"] = "Naturwiss. Wochenschr.",
  ["nauchni trudove na nauchno-izsledovatelskiia stomatologichen institut"] = "Nauchni Tr. Nauchnoizsled. Stomatol. Inst. (Sofiia)",
  ["nauchni trudove na visshiia meditsinski institut, sofiia"] = "Nauchni Tr Vissh Med Inst Sofiia",
  ["nauchni trudove. sofia. nauchno-izsledovatelski stomatologichen institut"] = "Nauchni Tr Nauchnoizsled Stomatol Inst (Sofiia)",
  ["nauchnye doklady vysshei shkoly. biologicheskie nauki"] = "Nauchnye Doki Vyss Shkoly Biol Nauki",
  ["nauchnye doklady vyssheĭ shkoly. biologicheskie nauki"] = "Nauchnye Doki Vyss Shkoly Biol Nauki",
  ["nauchnyĭ obzor: voprosy organizatsii zdravookhraneniia i istorii meditsiny v sssr"] = "Nauchn Obz Vopr Org Zdravookhr Istor Med SSSR",
  ["nauka i tehnologii"] = "Nauka Tehnol",
  ["nauka i tekhnicheskii progress"] = "Nauka i Tekhn. Progress",
  ["naukovi dopovidi nat︠s︡ionalʹnoho universytetu bioresursiv i pryrodokorystuvanni︠a︡ ukrainy"] = "Naukovi Dopovidi Natsionalnoho Universytetu Bioresursiv Pryrodokorystuvannia Ukr",
  ["naukovij vìsnik nacìonalʹnogo medičnogo unìversitetu ìmenì o.o. bogomolʹcâ"] = "Naukovij Visn Nac Medicnogo Univ Im O O Bogomolca",
  ["naunyn-schmiedeberg's archives of pharmacology"] = "Naunyn Schmiedebergs Arch Pharmacol",
  ["naunyn-schmiedebergs archiv fur experimentelle pathologie und  pharmacologie"] = "Naunyn-Schmiedeberg’s Arch. Exp. Pathol. Pharmacol.",
  ["naunyn-schmiedebergs archiv fur experimentelle pathologie und pharmacologie"] = "Naunyn-Schmiedeberg’s Arch. Exp. Pathol. Pharmacol.",
  ["naunyn-schmiedebergs archiv fur experimentelle pathologie und pharmakologie"] = "Naunyn Schmiedebergs Arch Exp Pathol Pharmakol",
  ["naunyn-schmiedebergs archiv fur pharmakologie"] = "Naunyn Schmiedebergs Arch Pharmakol",
  ["naunyn-schmiedebergs archiv fur pharmakologie und experimentelle pathologie"] = "Naunyn Schmiedebergs Arch Pharmakol Exp Pathol",
  ["naunyn-schmiedebergs archiv für experimentelle pathologie und pharmakologie"] = "Naunyn Schmiedebergs Arch Exp Pathol Pharmakol",
  ["naunyn-schmiedebergs archiv für pharmakologie"] = "Naunyn Schmiedebergs Arch Pharmakol",
  ["naunyn-schmiedebergs archiv für pharmakologie und experimentelle pathologie"] = "Naunyn Schmiedebergs Arch Pharmakol Exp Pathol",
  ["naunyn-schmiedebergs archiv für experimentelle pathologie und pharmakologie"] = "Naunyn. Schmiedebergs Arch. Exp. Pathol. Pharmakol.",
  ["naunyn-schmiedebergs archiv für pharmakologie"] = "Naunyn. Schmiedebergs Arch. Pharmakol.",
  ["naunyn-schmiedebergs archives of pharmacology"] = "Naunyn Schmiedebergs Arch Pharmacol",
  ["naunyn-schmiedeberg’s archives of pharmacology"] = "Naunyn-Schmiedeberg’s Arch. Pharmacol.",
  ["nauplius : revista da sociedade brasileira de carcinologia"] = "Nauplius",
  ["naval engineers journal"] = "Nav. Eng. J.",
  ["naval research logistics"] = "Nav. Res. Logist.",
  ["naval research logistics quarterly"] = "Nav. Res. Logist. Q.",
  ["naval review (annapolis, md.)"] = "Nav Rev",
  ["navigation- journal of the institute of navigation"] = "Navig.- J. Inst. Navig.",
  ["navy medicine"] = "Navy Med",
  ["nápravná pedagogika"] = "Napravna Pedagog",
  ["nāmah-yi anjuman-i ḥasharahshināsān-i īrān"] = "Namahyi Anjumani Hasharahshinasani Iran",
  ["nāsu sutēshon : ns = nurses' station quarterly"] = "Nasu Suteshon",
  ["nber macroeconomics annual"] = "NBER Macroecon Annu",
  ["nbu journal of plant sciences"] = "NBU J Plant Sci",
  ["ncbhr communiqué = communiqué cnbrh"] = "NCBHR Commun",
  ["ncce news"] = "NCCE News",
  ["ncehr communiqué = communiqué cnerh"] = "NCEHR Commun",
  ["nchs data brief"] = "NCHS Data Brief",
  ["nci monographs"] = "NCI Monogr.",
  ["nci monographs : a publication of the national cancer institute"] = "NCI Monogr",
  ["ncp bulletin"] = "NCP Bull",
  ["ncp bulletin / nutrition center of the philippines"] = "NCP Bull",
  ["ncsdha dental hygienist"] = "NCSDHA Dent. Hyg.",
  ["ncsl international measure"] = "NCSL Int Meas",
  ["ncsl legisbrief"] = "NCSL Legisbrief",
  ["ncsli measure"] = "NCSLI Meas.",
  ["nda journal"] = "NDA J",
  ["ndss symposium"] = "NDDS Symp",
  ["ndt & e international"] = "NDT & E Int.",
  ["ndt & e international : independent nondestructive testing and evaluation"] = "NDT E Int",
  ["ndt and e international"] = "NDT and E Int.",
  ["ndt international"] = "NDT Int.",
  ["ndt plus"] = "NDT Plus",
  ["near eastern archaeology"] = "Near East. Archaeol.",
  ["near surface geophysics"] = "Near Surf. Geophys.",
  ["nebraska history"] = "Nebr Hist",
  ["nebraska journal of economics and business"] = "Nebr. J. Econ. Bus.",
  ["nebraska law review"] = "Neb Law Rev",
  ["nebraska medical journal"] = "Nebr. Med. J.",
  ["nebraska nurse"] = "Nebr Nurse",
  ["nebraska state medical journal"] = "Nebr. State Med. J.",
  ["nebraska symposium on motivation"] = "Nebr. Symp. Motiv.",
  ["nebraska symposium on motivation. nebraska symposium on motivation"] = "Nebr Symp Motiv",
  ["nec journal of advanced technology"] = "NEC J. Adv. Technol.",
  ["nec technical journal"] = "NEC Tech. J.",
  ["necas center series"] = "Necas Center Ser.",
  ["nederlands bosbouw-tijdschrift"] = "Ned. Bosb.-Tijdschr.",
  ["nederlands militair geneeskundig tijdschrift"] = "Ned Milit Geneeskd Tijdschr",
  ["nederlands tandartsenblad"] = "Ned Tandartsenbl",
  ["nederlands tijdschrift voor de psychologie en haar grensgebieden"] = "Ned Tijdschr Psychol",
  ["nederlands tijdschrift voor geneeskunde"] = "Ned Tijdschr Geneeskd",
  ["nederlands tijdschrift voor gerontologie"] = "Ned Tijdschr Gerontol",
  ["nederlands tijdschrift voor natuurkunde"] = "Ned. Tijdschr. Natuurkd.",
  ["nederlands tijdschrift voor tandheelkunde"] = "Ned Tijdschr Tandheelkd",
  ["nederlandsch archief voor kerkgeschiedenis"] = "Ned Arch Kerkgeschied",
  ["nederlandsch tijdschrift voor evidence based practice"] = "Ned. Tijdshr. Evidence Based Pract.",
  ["nederlandsch tijdschrift voor hygiene, microbiologie en serologie"] = "Ned. Tijdschr. Hyg., Microbiol. Serol.",
  ["nederlandsch tijdschrift voor verloskunde en gynaecologie"] = "Ned Tijdschr Verloskd Gynaecol",
  ["needham research institute studies"] = "Needham Res. Inst. Stud.",
  ["neerlands volksleven"] = "Neerl Volksleven",
  ["nefrologia : publicacion oficial de la sociedad espanola nefrologia"] = "Nefrologia",
  ["nefrología : publicación oficial de la sociedad española nefrologia"] = "Nefrologia",
  ["neftyanoe khozyaistvo (petroleum industry)"] = "Neft. Khoz.",
  ["negro history bulletin"] = "Negro Hist Bull",
  ["nehw health watch"] = "NEHW Health Watch",
  ["nei menggu nong ye da xue xue bao : zi ran ke xue ban"] = "Nei Menggu Nong Ye Da Xue Xue Bao",
  ["neijiang shifan xueyuan xuebao"] = "Neijiang Shifan Xueyuan Xuebao",
  ["neimenggu daxue xuebao"] = "Neimenggu Daxue Xuebao Ziran Kexue",
  ["neimenggu shida xuebao"] = "Neimenggu Shida Xuebao Ziran Kexue Hanwen Ban",
  ["neir̆ofiziologiia = neurophysiology"] = "Neirofiziologiia",
  ["nejm ai"] = "NEJM AI",
  ["nejm catalyst"] = "NEJM Catal",
  ["nejm catalyst innovations in care delivery"] = "NEJM Catal Innov Care Deliv",
  ["nejm evidence"] = "NEJM Evidence",
  ["nelineinaya dinamika"] = "Nelinein. Din.",
  ["neliniini kolivannya"] = "Neliniini Koliv.",
  ["nematologia brasileira"] = "Nematol Bras",
  ["nematologia mediterranea"] = "Nematol Mediterr",
  ["nematology : international journal of fundamental and applied nematological research"] = "Nematology",
  ["nenpo : annual reports. jinko mondai kenkyujo (japan)"] = "Jinko Mondai Kenkyusho Nenpo",
  ["nenpo : annual reports. jinkō mondai kenkyūjo (japan)"] = "Jinko Mondai Kenkyusho Nenpo",
  ["neobiota : advancing research on alien species and biological invasions"] = "NeoBiota",
  ["neonatal intensive care"] = "Neonatal Intensive Care",
  ["neonatal intensive care : the journal of perinatology-neonatology"] = "Neonatal Intensive Care",
  ["neonatal network"] = "Neonatal Netw.",
  ["neonatal network : nn"] = "Neonatal Netw",
  ["neonatal pharmacology quarterly"] = "Neonatal Pharmacol Q",
  ["neoplasia (new york, n.y.)"] = "Neoplasia",
  ["neoplastic disease at various sites"] = "Monogr Neoplast Dis Var Sites",
  ["neopsichiatria (u.s.l. n. 15)"] = "Neopsichiatria",
  ["neopsichiatria : rassegna di psichiatria, neurologia, endocrinologia"] = "Neopsichiatria (Pisa)",
  ["neorganicheskie materialy"] = "Neorg. Mater.",
  ["neotropical biodiversity"] = "Neotrop Biodivers",
  ["neotropical biology & conservation"] = "Neotrop. Biol. Conserv.",
  ["neotropical entomology"] = "Neotrop Entomol",
  ["neotropical helminthology"] = "Neotrop Helminthol",
  ["neotropical ichthyology"] = "Neotrop. Ichthyol.",
  ["neotropical ichthyology : official journal of the sociedade brasileira de ictiologia"] = "Neotrop Ichthyol",
  ["neotropical primates"] = "Neotrop. Primates",
  ["neotropical primates : a newsletter of the neotropical section of the iucn/ssc primate specialist group"] = "Neotrop Primates",
  ["nepal journal of epidemiology"] = "Nepal J Epidemiol",
  ["nepal journal of science and technology"] = "Nepal J Sci Technol",
  ["nepal medical college journal : nmcj"] = "Nepal Med Coll J",
  ["nepal population and development journal"] = "Nepal Popul Dev J",
  ["nepalese journal of agricultural economics"] = "Nepal J Agric Econ",
  ["nepalese journal of ophthalmology : a biannual peer-reviewed academic journal of the nepal ophthalmic society : nepjoph"] = "Nepal J Ophthalmol",
  ["nephro-urology monthly"] = "Nephrourol Mon",
  ["nephrologie & therapeutique"] = "Nephrol Ther",
  ["nephrologie et therapeutique"] = "Nephrol. Ther.",
  ["nephrology (carlton, vic.)"] = "Nephrology (Carlton)",
  ["nephrology news & issues"] = "Nephrol News Issues",
  ["nephrology news and issues"] = "Nephrol. News Issues",
  ["nephrology nurse"] = "Nephrol Nurse",
  ["nephrology nursing journal : journal of the american nephrology nurses' association"] = "Nephrol Nurs J",
  ["nephrology nursing journal : journal of the american nephrology nurses’ association"] = "Nephrol Nurs J",
  ["nephrology self-assessment program : nephsap"] = "Nephrol Self Assess Program",
  ["nephrology, dialysis, transplantation"] = "Nephrol. Dial. Transplant.",
  ["nephrology, dialysis, transplantation : official publication of the european dialysis and transplant association - european renal association"] = "Nephrol Dial Transplant",
  ["nephron extra"] = "Nephron Extra",
  ["nephron. clinical practice"] = "Nephron Clin Pract",
  ["nephron. experimental nephrology"] = "Nephron Exp Nephrol",
  ["nephron. physiology"] = "Nephron Physiol",
  ["nervnaia sistema"] = "Nerv Sist",
  ["nervnaia sistema / leningradskii ordena lenina gosudarstvennyi universitet imeni a.a. zhdanova, fiziologicheskii institut imeni akad. a.a. ukhtomskogo"] = "Nerv Sist",
  ["nestle nutrition workshop series. clinical & performance programme"] = "Nestle Nutr Workshop Ser Clin Perform Programme",
  ["nestle nutrition workshop series. clinical and performance programme"] = "Nestle Nutr. Workshop Ser. Clin. Perform. Programme",
  ["nestle nutrition workshop series. paediatric programme"] = "Nestle Nutr Workshop Ser Pediatr Program",
  ["nestlé nutrition institute workshop series"] = "Nestle Nutr Inst Workshop Ser",
  ["nestlé nutrition workshop series"] = "Nestle Nutr Workshop Ser",
  ["nestlé nutrition workshop series. clinical & performance programme"] = "Nestle Nutr Workshop Ser Clin Perform Programme",
  ["nestlé nutrition workshop series. paediatric programme"] = "Nestle Nutr Workshop Ser Pediatr Program",
  ["netherlands heart journal : monthly journal of the netherlands society of cardiology and the netherlands heart foundation"] = "Neth Heart J",
  ["netherlands journal of geosciences"] = "Neth. J. Geosci.",
  ["netherlands journal of medicine"] = "Neth. J. Med.",
  ["netherlands journal of psychology"] = "Neth J Psychol",
  ["netherlands journal of sea research"] = "Neth. J. Sea Res.",
  ["netherlands journal of surgery"] = "Neth. J. Surg.",
  ["netherlands milk and dairy journal"] = "Neth Milk Dairy J",
  ["network (bristol, england)"] = "Network",
  ["network (research triangle park, n.c.)"] = "Netw Res Triangle Park N C",
  ["network - international fertility research program"] = "Netw Int Fertil Res Program",
  ["network and systems medicine"] = "Netw Syst Med",
  ["network en francais"] = "Netw Fr",
  ["network en français"] = "Netw Fr",
  ["network modeling analysis in health informatics and bioinformatics"] = "Network Model. Anal. Health Inf. Bioinf.",
  ["network modeling and analysis in health informatics and bioinformatics"] = "Netw Model Anal Health Inform Bioinform",
  ["network neuroscience"] = "Network Neurosci.",
  ["network neuroscience (cambridge, mass.)"] = "Netw Neurosci",
  ["network science"] = "Network Sci.",
  ["network science (cambridge university press)"] = "Netw Sci (Camb Univ Press)",
  ["network: computation in neural systems"] = "Network: Comput. Neural Syst.",
  ["networking science"] = "Networking Sci.",
  ["networks & spatial economics"] = "Networks Spatial Econ.",
  ["networks (new york, ny)"] = "Networks (N Y)",
  ["networks and heterogeneous media"] = "Networks Heterogen. Media",
  ["networks and spatial economics"] = "Netw Spat Econ",
  ["networks and telecommunications series"] = "Netw. Telecommun. Ser.",
  ["networks. an international journal"] = "Networks",
  ["neue ephemeris für semitische epigraphik"] = "NEphemSemEpigr",
  ["neue heimat"] = "Neue Heimat",
  ["neue juristische wochenschrift"] = "Neue Jurist Wochenschr",
  ["neue landschaft"] = "Neue Landsch.",
  ["neue munchner beitrage zur geschichte der medizin und naturwissenschaften. medizinhistorische reihe"] = "Neue Munch. Beitr. Gesch. Med. Naturwiss. Medizinhist. Reihe",
  ["neue münchner beiträge zur geschichte der medizin und naturwissenschaften. medizinhistorische reihe"] = "Neue Munch Beitr Gesch Med Naturwiss Medizinhist Reihe",
  ["neue münchner beiträge zur geschichte der medizin und naturwissenschaften. naturwissenschaftshistorische reihe"] = "Neue Munch Beitr Gesch Med Naturwiss",
  ["neue osterreichische zeitschrift fur kinderheilkunde"] = "Neue Osterr Z Kinderheilkd",
  ["neue österreichische zeitschrift für kinderheilkunde"] = "Neue Osterr Z Kinderheilkd",
  ["neue politische literatur"] = "Neue Polit Lit",
  ["neue zürcher zeitung"] = "Neue Zür. Ztg.",
  ["neuere medizin- und wissenschaftsgeschichte"] = "Neuere Med Wiss Quellen Stud",
  ["neues jahrbuch fuer mineralogie, abhandlungen"] = "Neues Jahrb. Mineral., Abh.",
  ["neues jahrbuch fur geologie und palaontologie. abhandlungen"] = "Neues Jahrb Geol Palaontol Abh",
  ["neues jahrbuch fur mineralogie, abhandlungen"] = "Neues Jahrb. Mineral. Abh.",
  ["neues jahrbuch fur mineralogie, geologie und palaeontologie, abhandlungen abteiling a: mineralogie, petrographie"] = "Neues Jahrb. Mineral. Geol. Palaeontol. Abh. Abt. A",
  ["neues jahrbuch fur mineralogie, geologie und palaeontologie, abhandlungen abteiling b: geologie, palaeontologie"] = "Neues Jahrb. Mineral. Geol. Palaeontol. Abh. Abt. B",
  ["neues jahrbuch fur mineralogie, geologie und palaeontologie, referate"] = "Neues Jahrb. Mineral. Geol. Palaeontol. Ref.",
  ["neues jahrbuch fur mineralogie, monatshefte"] = "Neues Jahrb. Mineral. Monatsh..",
  ["neues jahrbuch für geologie und paläontologie. abhandlungen"] = "Neues Jahrb Geol Palaontol Abh",
  ["neues jahrbuch für mineralogie, abhandlungen"] = "Neues Jahrb. Mineral., Abh.",
  ["neues jahrbuch für mineralogie, monatshefte"] = "Neues Jahrb. Mineral., Monatsh.",
  ["neujahrsblatt der lesegesellschaft wädenswil"] = "Neujahrsbl. Leseges. Wädenswil",
  ["neujahrsblatt der naturforschenden gesellschaft in zürich"] = "Neujahrsbl. Nat.forsch. Ges. Zür.",
  ["neujahrsblatt der naturforschenden gesellschaft schaffhausen"] = "Neujahrsbl. Nat.forsch. Ges. Schaffhausen",
  ["neujahrsbote für das glarner hinterland"] = "Neujahrsbote Glarner Hinterland",
  ["neumologia y cirugia de torax"] = "Neumol Cir Torax",
  ["neumologia y cirugía de tórax"] = "Neumol Cir Torax",
  ["neuphilologische mitteilungen"] = "Neuphilol Mitt",
  ["neural computation"] = "Neural Comput",
  ["neural computing & applications"] = "Neural Comput Appl",
  ["neural computing and applications"] = "Neural Comput. Appl.",
  ["neural development"] = "Neural Dev",
  ["neural information processing series"] = "Neural Inf. Process. Ser.",
  ["neural monitoring"] = "Neurotraum.",
  ["neural network systems techniques and applications"] = "Neural Netw. Syst. Tech. Appl.",
  ["neural networks"] = "Neural Networks",
  ["neural networks : the official journal of the international neural network society"] = "Neural Netw",
  ["neural plasticity"] = "Neural Plast",
  ["neural processing letters"] = "Neural Process Lett",
  ["neural regeneration research"] = "Neural Regen Res",
  ["neural systems & circuits"] = "Neural Syst Circuits",
  ["neural systems and circuits"] = "Neural Syst. Circuits",
  ["neural, parallel & scientific computations"] = "Neural Parallel Sci. Comput.",
  ["neuro : open journal"] = "Neuro",
  ["neuro endocrinology letters"] = "Neuro Endocrinol Lett",
  ["neuro-degenerative diseases"] = "Neurodegener Dis",
  ["neuro-oncology advances"] = "Neurooncol Adv",
  ["neuro-oncology practice"] = "Neurooncol Pract",
  ["neuro-ophthalmology (aeolus press)"] = "Neuroophthalmology",
  ["neurobehavioral hiv medicine"] = "Neurobehav HIV Med",
  ["neurobehavioral toxicology"] = "Neurobehav Toxicol",
  ["neurobehavioral toxicology and teratology"] = "Neurobehav Toxicol Teratol",
  ["neurobiology (budapest, hungary)"] = "Neurobiology (Bp)",
  ["neurobiology of aging"] = "Neurobiol Aging",
  ["neurobiology of aging science"] = "Neurobiol. Aging Sci.",
  ["neurobiology of circadian rhythms"] = "Neurobiol. Circadian Rhythms",
  ["neurobiology of disease"] = "Neurobiol Dis",
  ["neurobiology of language"] = "Neurobiol. Lang.",
  ["neurobiology of language (cambridge, mass.)"] = "Neurobiol Lang (Camb)",
  ["neurobiology of learning and memory"] = "Neurobiol Learn Mem",
  ["neurobiology of lipids"] = "Neurobiol Lipids",
  ["neurobiology of pain"] = "Neurobiol. Pain",
  ["neurobiology of pain (cambridge, mass.)"] = "Neurobiol Pain",
  ["neurobiology of sleep and circadian rhythms"] = "Neurobiol Sleep Circadian Rhythms",
  ["neurobiology of stress"] = "Neurobiol Stress",
  ["neurocase : case studies in neuropsychology, neuropsychiatry, and behavioural neurology"] = "Neurocase",
  ["neurochemical journal"] = "Neurochem. J.",
  ["neurochemical pathology"] = "Neurochem Pathol",
  ["neurochemical research"] = "Neurochem Res",
  ["neurochemistry & neuropharmacology : open access"] = "Neurochem Neuropharmacol",
  ["neurochemistry international"] = "Neurochem Int",
  ["neurocirugia (asturias, spain)"] = "Neurocirugia (Astur)",
  ["neurocirugía (asturias, spain)"] = "Neurocirugia (Astur)",
  ["neurocritical care"] = "Neurocrit Care",
  ["neurodegeneration : a journal for neurodegenerative disorders, neuroprotection, and neuroregeneration"] = "Neurodegeneration",
  ["neurodegenerative disease management"] = "Neurodegener Dis Manag",
  ["neurodiagnostic journal"] = "Neurodiagn. J.",
  ["neuroembryology and aging"] = "Neuroembryology Aging",
  ["neurogastroenterology and motility"] = "Neurogastroenterol. Motil.",
  ["neurogastroenterology and motility : the official journal of the european gastrointestinal motility society"] = "Neurogastroenterol Motil",
  ["neurogenesis (austin, tex.)"] = "Neurogenesis (Austin)",
  ["neuroglia (basel, switzerland)"] = "Neuroglia",
  ["neurographics (2011)"] = "Neurographics (2011)",
  ["neuroimage. clinical"] = "Neuroimage Clin",
  ["neuroimaging clinics of north america"] = "Neuroimaging Clin N Am",
  ["neuroimmunology and neuroinflammation"] = "Neuroimmunol Neuroinflamm",
  ["neuroimmunology reports"] = "Neuroimmunol. Rep.",
  ["neurologia (barcelona, spain)"] = "Neurologia",
  ["neurologia croatica"] = "Neurol. Croat.",
  ["neurologia croatica : glasilo udruzenja neurologa jugoslavije = official journal of yugoslav neurological association"] = "Neurol Croat",
  ["neurologia i neurochirurgia polska"] = "Neurol Neurochir Pol",
  ["neurologia medico-chirurgica"] = "Neurol Med Chir (Tokyo)",
  ["neurologia, neurochirurgia i psychiatria polska"] = "Neurol Neurochir Psychiatr Pol",
  ["neurologia, neurocirugia, psiquiatria"] = "Neurol Neurocir Psiquiatr",
  ["neurologia, psihiatria, neurochirurgia"] = "Neurol Psihiatr Neurochir",
  ["neurologic clinics"] = "Neurol Clin",
  ["neurological disorders & epilepsy journal"] = "Neurol Disord Epilepsy J",
  ["neurological research"] = "Neurol Res",
  ["neurological research and practice"] = "Neurol Res Pract",
  ["neurological research and therapy"] = "Neurol Res Ther",
  ["neurological sciences"] = "Neurol. Sci.",
  ["neurological sciences : official journal of the italian neurological society and of the italian society of clinical neurophysiology"] = "Neurol Sci",
  ["neurological sciences and neurophysiology"] = "Neurol. Sci. Neurophysiol.",
  ["neurological sciences and neurosurgery"] = "Neurol Sci Neurosurg",
  ["neurologie a psychiatrie ceskoslovenska"] = "Neurol Psychiatr Ceskoslov",
  ["neurologie a psychiatrie c̆eskoslovenská"] = "Neurol Psychiatr Ceskoslov",
  ["neurologie et psychiatrie"] = "Neurol Psychiatr (Bucur)",
  ["neurologisch : fachmagazin für neurologie"] = "Neurologisch (Wien)",
  ["neurología (barcelona, spain)"] = "Neurologia",
  ["neurología, neurocirugía, psiquiatría"] = "Neurol Neurocir Psiquiatr",
  ["neurology & clinical neurophysiology : ncn"] = "Neurol Clin Neurophysiol",
  ["neurology (e-cronicon)"] = "Neurology (ECronicon)",
  ["neurology and clinical neurophysiology"] = "Neurol. Clin. Neurophysiol.",
  ["neurology and clinical neuroscience"] = "Neurol Clin Neurosci",
  ["neurology and neurobiology"] = "Neurol Neurobiol",
  ["neurology and neurobiology (tallinn, estonia)"] = "Neurol Neurobiol (Tallinn)",
  ["neurology and therapy"] = "Neurol Ther",
  ["neurology asia"] = "Neurol Asia",
  ["neurology india"] = "Neurol India",
  ["neurology international"] = "Neurol Int",
  ["neurology research international"] = "Neurol Res Int",
  ["neurology, neurophysiology, and neuroscience"] = "Neurol Neurophysiol Neurosci",
  ["neurology, psychiatry, and brain research"] = "Neurol Psychiatry Brain Res",
  ["neurology. clinical practice"] = "Neurol Clin Pract",
  ["neurology. genetics"] = "Neurol Genet",
  ["neurology® neuroimmunology & neuroinflammation"] = "Neurol Neuroimmunol Neuroinflamm",
  ["neuromodulation : journal of the international neuromodulation society"] = "Neuromodulation",
  ["neuromodulation: technology at the neural interface"] = "Neuromodulation: Technol. Neural Interface",
  ["neuromolecular medicine"] = "Neuromolecular Med",
  ["neuromorphic computing and engineering"] = "Neuromorph. Comput. Eng.",
  ["neuromuscular disorders"] = "Neuromuscul. Disord.",
  ["neuromuscular disorders : nmd"] = "Neuromuscul Disord",
  ["neuron glia biology"] = "Neuron Glia Biol",
  ["neuronal signaling"] = "Neuronal Signal",
  ["neurons, behavior, data analysis and theory"] = "Neuron Behav Data Anal Theory",
  ["neurons: methods and applications for the cell biologist"] = "Method. Cell Biol.",
  ["neuropathology : official journal of the japanese society of neuropathology"] = "Neuropathology",
  ["neuropathology and applied neurobiology"] = "Neuropathol Appl Neurobiol",
  ["neuropatologia polska"] = "Neuropatol Pol",
  ["neuropeptides (amsterdam, netherlands)"] = "Neuropeptides (Amsterdam, Neth.)",
  ["neuropeptides (oxford, united kingdom)"] = "Neuropeptides (Oxford, U. K.)",
  ["neurophysiologie clinique"] = "Neurophysiol. Clin.",
  ["neurophysiologie clinique = clinical neurophysiology"] = "Neurophysiol Clin",
  ["neuropsychiatric disease and treatment"] = "Neuropsychiatr Dis Treat",
  ["neuropsychiatric electrophysiology"] = "Neuropsychiatr Electrophysiol",
  ["neuropsychiatrie : klinik, diagnostik, therapie und rehabilitation : organ der gesellschaft osterreichischer nervenarzte und psychiater"] = "Neuropsychiatr",
  ["neuropsychiatrie : klinik, diagnostik, therapie und rehabilitation : organ der gesellschaft österreichischer nervenärzte und psychiater"] = "Neuropsychiatr",
  ["neuropsychiatrie de l'enfance et de l'adolescence"] = "Neuropsychiatr Enfance Adolesc",
  ["neuropsychiatrie de l’enfance et de l’adolescence"] = "Neuropsychiatr. Enfance Adolesc.",
  ["neuropsychiatry, neuropsychology, and behavioral neurology"] = "Neuropsychiatry Neuropsychol Behav Neurol",
  ["neuropsychological rehabilitation"] = "Neuropsychol Rehabil",
  ["neuropsychology review"] = "Neuropsychol Rev",
  ["neuropsychology, development, and cognition. section b, aging, neuropsychology and cognition"] = "Neuropsychol Dev Cogn B Aging Neuropsychol Cogn",
  ["neuropsychopharmacologia hungarica : a magyar pszichofarmakológiai egyesület lapja = official journal of the hungarian association of psychopharmacology"] = "Neuropsychopharmacol Hung",
  ["neuropsychopharmacology : official publication of the american college of neuropsychopharmacology"] = "Neuropsychopharmacology",
  ["neuropsychopharmacology reports"] = "Neuropsychopharmacol Rep",
  ["neuroquantology : an interdisciplinary journal of neuroscience and quantum physics"] = "Neuroquantology",
  ["neuroradiology journal"] = "Neuroradiol. J.",
  ["neurorehabilitation and neural repair"] = "Neurorehabil Neural Repair",
  ["neurorx : the journal of the american society for experimental neurotherapeutics"] = "NeuroRx",
  ["neuroscience & biobehavioral reviews"] = "Neurosci. Biobehav. Rev.",
  ["neuroscience (amsterdam, netherlands)"] = "Neuroscience (Amsterdam, Neth.)",
  ["neuroscience (oxford, united kingdom)"] = "Neuroscience (Oxford, U. K.)",
  ["neuroscience (san diego, ca, united states)"] = "Neuroscience (San Diego, CA, U. S.)",
  ["neuroscience and behavioral physiology"] = "Neurosci Behav Physiol",
  ["neuroscience and biobehavioral reviews"] = "Neurosci Biobehav Rev",
  ["neuroscience and medicine"] = "Neurosci Med",
  ["neuroscience and neuroeconomics"] = "Neurosci Neuroecon",
  ["neuroscience applied"] = "Neurosci. Appl.",
  ["neuroscience bulletin"] = "Neurosci Bull",
  ["neuroscience communications"] = "Neurosci Commun (Houst)",
  ["neuroscience informatics"] = "Neurosci. Inf.",
  ["neuroscience insights"] = "Neurosci Insights",
  ["neuroscience journal"] = "Neurosci J",
  ["neuroscience letters"] = "Neurosci Lett",
  ["neuroscience letters. supplement"] = "Neurosci Lett Suppl",
  ["neuroscience news"] = "Neurosci News",
  ["neuroscience of consciousness"] = "Neurosci Conscious",
  ["neuroscience research"] = "Neurosci Res",
  ["neuroscience research communications"] = "Neurosci Res Commun",
  ["neuroscience research. supplement"] = "Neurosci. Res. Suppl.",
  ["neuroscience research. supplement : the official journal of the japan neuroscience society"] = "Neurosci Res Suppl",
  ["neurosciences (riyadh, saudi arabia)"] = "Neurosciences (Riyadh)",
  ["neurosciences research"] = "Neurosci Res (N Y)",
  ["neurosciences research program bulletin"] = "Neurosci Res Program Bull",
  ["neurosurgery clinics of north america"] = "Neurosurg Clin N Am",
  ["neurosurgery quarterly"] = "Neurosurg Q",
  ["neurosurgical focus"] = "Neurosurg Focus",
  ["neurosurgical review"] = "Neurosurg Rev",
  ["neurotherapeutics : the journal of the american society for experimental neurotherapeutics"] = "Neurotherapeutics",
  ["neurotoxicity research"] = "Neurotox Res",
  ["neurotoxicology and teratology"] = "Neurotoxicol Teratol",
  ["neurotransmitter (houston, tex.)"] = "Neurotransmitter (Houst)",
  ["neurotrauma reports"] = "Neurotrauma Rep",
  ["neurourology and urodynamics"] = "Neurourol Urodyn",
  ["neurovascular imaging (london)"] = "Neurovasc Imaging",
  ["neusser jahrbuch"] = "Neuss Jahrb",
  ["neutron news"] = "Neutron News",
  ["nevada historical society quarterly (1961)"] = "Nev Hist Soc Q",
  ["nevada nurses association quarterly newsletter"] = "Nev Nurses Assoc Q Newsl",
  ["nevada revised statutes : containing all statute laws of nevada of a general, public, and permanent nature. nevada"] = "Nev Revis Statut Nev",
  ["nevada rnformation"] = "Nev Rnformation",
  ["nevrone; periodico di neurologia, psichiatria e scienze affini"] = "Nevrone Period Neurol Psichiatr Sci Affin",
  ["nevropatologiia i psikhiatriia"] = "Nevropatol Psikhiatriia",
  ["new african"] = "New Afr",
  ["new american review"] = "New Am Rev",
  ["new astronomy"] = "New Astron.",
  ["new astronomy review"] = "New Astron. Rev.",
  ["new astronomy reviews"] = "New Astron. Rev.",
  ["new atlantis (washington, d.c.)"] = "New Atlantis",
  ["new biologist"] = "New Biol.",
  ["new biotechnology"] = "N Biotechnol",
  ["new blackfriars"] = "New Blackfriars",
  ["new carbon materials"] = "New Carbon Mater.",
  ["new catholic world"] = "New Cathol World",
  ["new community"] = "New community",
  ["new comprehensive biochemistry"] = "New Compr. Biochem.",
  ["new critical idiom"] = "New. Crit. Idiom.",
  ["new dentist"] = "New Dent.",
  ["new developments in quantitative trading and investment"] = "New Dev. Quant. Trading Invest.",
  ["new diamond and frontier carbon technology"] = "New Diamond Front. Carbon Technol.",
  ["new directions for child and adolescent development"] = "New Dir Child Adolesc Dev",
  ["new directions for child development"] = "New Dir Child Dev",
  ["new directions for evaluation"] = "New Dir Eval",
  ["new directions for mental health services"] = "New Dir Ment Health Serv",
  ["new directions for student leadership"] = "New Dir Stud Leadersh",
  ["new directions for youth development"] = "New Dir Youth Dev",
  ["new directions in the philosophy of science"] = "New Dir. Philos. Sci.",
  ["new disease reports"] = "New Dis Rep",
  ["new economic windows"] = "New Econ. Windows",
  ["new economy"] = "New Economy",
  ["new egyptian journal of medicine"] = "New Egypt J Med",
  ["new egyptian journal of microbiology"] = "New Egypt. J. Microbiol.",
  ["new england and regional allergy proceedings"] = "N Engl Reg Allergy Proc",
  ["new england classical newsletter"] = "NECN",
  ["new england dental journal"] = "N Engl Dent J",
  ["new england economic review"] = "New England Econ. Rev.",
  ["new england journal of human services"] = "N Engl J Hum Serv",
  ["new england journal of medicine"] = "N. Engl. J. Med.",
  ["new england journal of medicine (overseas ed.)"] = "N Engl J Med Overseas Ed",
  ["new england journal of public policy : a journal of the john w. mccormack institute of public affairs, university of massachusetts at boston"] = "N Engl J Public Policy",
  ["new england journal on criminal and civil confinement"] = "N Engl J Crim Civ Confin",
  ["new england journal on prison law"] = "New Engl J Prison Law",
  ["new england law review"] = "New Engl Law Rev",
  ["new era nursing image international"] = "New Era Nurs. Image Int.",
  ["new era nursing image international : neni"] = "New Era Nurs Image Int",
  ["new forests"] = "New For (Dordr)",
  ["new frontiers in ophthalmology (london)"] = "New Front Ophthalmol",
  ["new frontiers in regional science: asian perspectives"] = "New Front. Reg. Sci. Asian Perspect.",
  ["new generation computing"] = "New Gener Comput",
  ["new genetics and society"] = "New Genet Soc",
  ["new german critique, ngc"] = "New Ger Crit",
  ["new guard"] = "New Guard",
  ["new hampshire health news"] = "N H Health News",
  ["new hampshire revised statutes annotated, 1955. new hampshire"] = "N H Revis Statut Annot N H",
  ["new health"] = "New Health",
  ["new hibernia review"] = "New Hibernia Rev",
  ["new hibernia review = iris éireannach nua"] = "New Hibernia Rev",
  ["new horizons"] = "New Horiz.",
  ["new horizons (baltimore, md.)"] = "New Horiz",
  ["new horizons in translational medicine"] = "New Horiz. Transl. Med.",
  ["new humanist (london, england)"] = "New Humanist",
  ["new icmi studies series"] = "New ICMI Stud. Ser.",
  ["new ideas in psychology"] = "New Ideas Psychol",
  ["new internationalist"] = "New Int",
  ["new istanbul contribution to clinical science"] = "New Istanbul Contrib Clin Sci",
  ["new jersey healthcare"] = "N J Healthc",
  ["new jersey historical commission newsletter. new jersey historical commission"] = "N J Hist Comm Newsl",
  ["new jersey history"] = "N J Hist",
  ["new jersey journal of pharmacy"] = "N J J Pharm",
  ["new jersey league for nursing news"] = "N J League Nurs News",
  ["new jersey medicine"] = "N. J. Med.",
  ["new jersey medicine : the journal of the medical society of new jersey"] = "N J Med",
  ["new jersey nurse"] = "N J Nurse",
  ["new jersey statutes annotated : permanent ed. laws of a general and permanent nature under arrangement of revised statutes, 1937, including laws of the 163rd legislature, 1939; with annotations from cases construing or applying the laws, tables, and index. new jersey"] = "N J Statut Annot N J",
  ["new journal (institute of health record & information management)"] = "New J (Inst Health Rec Inf Manag)",
  ["new journal of botany"] = "New J Botany",
  ["new journal of chemistry"] = "New J. Chem.",
  ["new journal of chemistry = nouveau journal de chimie"] = "New J Chem",
  ["new journal of physics"] = "New J. Phys.",
  ["new journal of science"] = "New J Sci",
  ["new leader (new york, n.y. : 1935)"] = "New Lead",
  ["new left review"] = "New Left Rev",
  ["new literary history"] = "New Lit Hist",
  ["new martin gardner mathematical library"] = "New Martin Gardner Math. Libr.",
  ["new materials, compounds and applications"] = "New Mater. Compd. Appl.",
  ["new mathematical library"] = "New Math. Library",
  ["new mathematical methods, systems and applications set"] = "New Math. Methods Syst. Appl. Set",
  ["new mathematical monographs"] = "New Math. Monogr.",
  ["new mathematics and natural computation"] = "New Math. Nat. Comput.",
  ["new media & society"] = "New Media Soc",
  ["new medieval literatures"] = "New Mediev Lit (Turnhout)",
  ["new mexico business"] = "New Mexico Bus.",
  ["new mexico dental journal"] = "N M Dent J",
  ["new mexico health officer"] = "N M Health Off",
  ["new mexico historical review"] = "N M Hist Rev",
  ["new mexico law review"] = "N M Law Rev",
  ["new mexico nurse"] = "N M Nurse",
  ["new mexico statutes, 1978 annotated. new mexico"] = "N M Statut 1978 Annot N M",
  ["new microbes and new infections"] = "New Microbes New Infect",
  ["new microbes new infections"] = "New Microbes New Infect.",
  ["new microbiologica"] = "New Microbiol.",
  ["new nano"] = "New Nano",
  ["new negatives in plant science"] = "New Negat. Plant Sci.",
  ["new perspectives on turkey"] = "New Perspect Turk",
  ["new perspectives quarterly : npq"] = "New Perspect Q",
  ["new philippines"] = "New Philipp",
  ["new physician"] = "New Physician",
  ["new phytologist"] = "New Phytol.",
  ["new political economy"] = "New Polit. Economy",
  ["new polymeric materials"] = "New Polym. Mater.",
  ["new republic (new york, n.y.)"] = "New Repub",
  ["new review of bioethics"] = "New Rev Bioeth",
  ["new scientist"] = "New Sci.",
  ["new scientist (1971)"] = "New Sci",
  ["new scientist and science journal"] = "New Sci Sci J",
  ["new series"] = "New Ser.",
  ["new society"] = "New Soc",
  ["new solutions : a journal of environmental and occupational health policy : ns"] = "New Solut",
  ["new south (atlanta, ga.)"] = "New South",
  ["new south wales public health bulletin"] = "N S W Public Health Bull",
  ["new space"] = "New Space",
  ["new statesman (london, england : 1957)"] = "New Statesman",
  ["new studies in american intellectual and cultural history"] = "New Stud. Am. Intellect. Cult. Hist.",
  ["new technology, work and employment"] = "New Technol Work Employ",
  ["new testament abstracts"] = "NTA",
  ["new testament studies: an international journal publ. quarterly under the auspices of studiorum novi testamenti societas"] = "NTS",
  ["new theatre quarterly : ntq"] = "NTQ New Theatre Q",
  ["new theology review"] = "New Theol Rev",
  ["new times"] = "New Times",
  ["new trends czech. econ."] = "New Trends Czech. Econ.",
  ["new trends in image analysis and processing -- iciap 2015 workshops : iciap 2015 international workshops, biofor, ctmr, rheuma, isca, madima, sbmi, and qoem, genoa, italy, september 7-8, 2015, proceedings"] = "New Trends Image Anal Process ICIAP 2015 Workshops (2015)",
  ["new trends in mathematical sciences"] = "New Trends Math. Sci.",
  ["new vision (kampala, uganda)"] = "New Vis",
  ["new voices in psychology"] = "New Voices Psychol",
  ["new west"] = "New West",
  ["new world (new orleans, la.)"] = "New World",
  ["new york"] = "New York",
  ["new york affairs"] = "N Y Aff",
  ["new york folklore"] = "N Y Folk",
  ["new york folklore quarterly"] = "NY Folkl Q",
  ["new york historical society quarterly. new-york historical society"] = "NY Hist Soc Q",
  ["new york history. new york state historical association"] = "N Y Hist",
  ["new york journal of dentistry"] = "N. Y. J. Dent.",
  ["new york journal of mathematics"] = "New York J. Math.",
  ["new york journal of mathematics. nyjm monographs"] = "NYJM Monogr.",
  ["new york law forum"] = "N Y Law Forum",
  ["new york law journal"] = "N Y Law J",
  ["new york law school human rights annual"] = "N Y Law School Hum Rights Annu",
  ["new york law school journal of human rights"] = "N Y Law Sch J Hum Rights",
  ["new york law school journal of international and comparative law"] = "N Y Law Sch J Int Comp Law",
  ["new york law school law review. new york law school"] = "NY Law Sch Law Rev",
  ["new york medical college news & notes"] = "N Y Med Coll News Notes",
  ["new york medicine"] = "NY Med",
  ["new york state bar journal"] = "N Y State Bar J",
  ["new york state dental journal"] = "N. Y. State Dent. J.",
  ["new york state journal of medicine"] = "N Y State J Med",
  ["new york state nurse"] = "N Y State Nurse",
  ["new york state pharmacist"] = "N Y State Pharm",
  ["new york state psychologist"] = "N Y State Psychol",
  ["new york supplement. second series"] = "N Y Suppl Second Ser",
  ["new york times"] = "N.Y. Times",
  ["new york times magazine"] = "N. Y. Times Mag.",
  ["new york university journal of dentistry"] = "N Y Univ J Dent",
  ["new york university journal of international law & politics. new york university. international law society"] = "N Y Univ J Int Law Polit",
  ["new york university journal of legislation and public policy"] = "N Y Univ J Legis Public Policy",
  ["new york university law review (1950)"] = "N Y Univ Law Rev",
  ["new york university medical quarterly"] = "N Y Univ Med Q",
  ["new yorker (new york, n.y. : 1925)"] = "New Yorker",
  ["new zealand bioethics journal"] = "N Z Bioeth J",
  ["new zealand cartography and geographic information systems : the journal of the new zealand cartographic society"] = "N Z Cartogr Geogr Inf Sys",
  ["new zealand dental journal"] = "N. Z. Dent. J.",
  ["new zealand economic papers"] = "New Zealand Econ. Pap.",
  ["new zealand engineering"] = "N.Z. Eng.",
  ["new zealand entomologist"] = "N.Z. Entomol.",
  ["new zealand environment"] = "N.Z. Environ.",
  ["new zealand fertilizer journal"] = "N.Z. Fert. J.",
  ["new zealand geographer"] = "N Z Geog",
  ["new zealand geological survey bulletin"] = "N.Z. Geol. Surv. Bull.",
  ["new zealand health & hospital"] = "N Z Health Hospital",
  ["new zealand health and hospital"] = "N. Z. Health Hospital",
  ["new zealand hospital"] = "N Z Hosp",
  ["new zealand journal of agricultural research"] = "N.Z. J. Agric. Res.",
  ["new zealand journal of agriculture"] = "N.Z. J. Agric.",
  ["new zealand journal of botany"] = "N Z J Bot",
  ["new zealand journal of crop and horticultural science"] = "N Z J Crop Hortic Sci",
  ["new zealand journal of dairy science and technology"] = "N.Z. J. Dairy Sci. Technol.",
  ["new zealand journal of ecology"] = "N Z J Ecol",
  ["new zealand journal of family planning"] = "N Z J Fam Plann",
  ["new zealand journal of forest science"] = "N.Z. J. For. Sci.",
  ["new zealand journal of forestry"] = "N.Z. J. For.",
  ["new zealand journal of forestry science"] = "N Z J For Sci",
  ["new zealand journal of geography"] = "N Z J Geogr",
  ["new zealand journal of geology and geophysics"] = "N.Z. J. Geol. Geophys.",
  ["new zealand journal of marine and freshwater research"] = "N.Z. J. Mar. Freshwater Res.",
  ["new zealand journal of mathematics"] = "New Zealand J. Math.",
  ["new zealand journal of medical laboratory science"] = "N. Z. J. Med. Lab. Sci.",
  ["new zealand journal of psychology"] = "NZ J Psychol",
  ["new zealand journal of science"] = "N.Z. J. Sci.",
  ["new zealand journal of science and technology section a"] = "N.Z. J. Sci. Technol. Sect: A",
  ["new zealand journal of science and technology section b"] = "N.Z. J. Sci. Technol. Sect: B",
  ["new zealand journal of technology"] = "N.Z. J. Technol.",
  ["new zealand journal of zoology"] = "N.Z. J. Zool.",
  ["new zealand libraries"] = "N Z Libr",
  ["new zealand mathematical society newsletter"] = "New Zealand Math. Soc. Newslett.",
  ["new zealand medical journal"] = "N. Z. Med. J.",
  ["new zealand nursing forum"] = "N Z Nurs Forum",
  ["new zealand nursing journal"] = "N. Z. Nurs. J.",
  ["new zealand plant protection. new zealand plant protection society"] = "N Z Plant Prot",
  ["new zealand population newsletter"] = "N Z Popul Newsl",
  ["new zealand population newsletter / new zealand demographic society (inc.)"] = "N Z Popul Newsl",
  ["new zealand population review"] = "N Z Popul Rev",
  ["new zealand population review / new zealand demographic society (inc.)"] = "N Z Popul Rev",
  ["new zealand school dental service gazette"] = "N Z Sch Dent Serv Gaz",
  ["new zealand slavonic journal"] = "N Z Slav J",
  ["new zealand veterinary journal"] = "N Z Vet J",
  ["newborn and infant nursing reviews : nainr"] = "Newborn Infant Nurs Rev",
  ["newcastle medical journal"] = "Newcastle Med J",
  ["newfoundland medical association journal"] = "Newfoundl Med Assoc J",
  ["newport history : bulletin of the newport historical society"] = "Newport Hist",
  ["news & views : rhode island nursing magazine"] = "News Views",
  ["news - women's international network. women's international network"] = "WIN News",
  ["news - women’s international network. women’s international network"] = "WIN News",
  ["news and views - bangladesh institute of research for promotion of essential & reproductive health and technologies"] = "News Views Bangladesh Inst Res Promot Essent Reprod Health Technol",
  ["news bulletin - indian dental association"] = "News Bull Indian Dent Assoc",
  ["news bulletin and calendar. worcester art museum"] = "NBWorcArtMus",
  ["news bulletin, indian dental association"] = "News Bull. Indian Dent. Assoc.",
  ["news in physiological sciences : an international journal of physiology produced jointly by the international union of physiological sciences and the american physiological society"] = "News Physiol Sci",
  ["news letter (aaf rheumatic fever control program (u.s.))"] = "News Lett AAF Rheum Fever Control Program",
  ["news letter (population centre bangalore)"] = "News Lett Popul Cent Bangalore",
  ["news letter, florence nightingale international nurses association"] = "News Lett. Florence Nightingale Int. Nurs. Assoc.",
  ["news letter, maine state nurses association"] = "News Lett. Maine State Nurses Assoc.",
  ["news letter. maine state nurses' association"] = "News Lett Maine State Nurses Assoc",
  ["news letter. maine state nurses’ association"] = "News Lett Maine State Nurses Assoc",
  ["news letter. national council on rehabilitation"] = "News Lett Natl Counc Rehabil",
  ["news letter. scandinavian society of forensic odontology"] = "Newsl Scand Soc Forensic Odontol",
  ["news letter. suffolk county medical society"] = "Newsl Suffolk Cty Med Soc",
  ["news letter; florence nightingale international nurses association"] = "News Lett Florence Nightingale Int Nurs Assoc",
  ["news notes / ohio dental journal"] = "News Notes Ohio Dent J",
  ["news notes, ohio dental journal"] = "News Notes Ohio Dent. J.",
  ["news of forest history"] = "News For. Hist.",
  ["news of norway"] = "News Nor",
  ["news views (michigan state nurses association. detroit district.)"] = "News Views",
  ["news-letter. american association of psychiatric social workers"] = "News Lett Am Assoc Psychiatr Soc Work",
  ["news. phi delta epsilon"] = "News Phi Delta Epsil",
  ["newsette - department of health national league of nurses, manila, philippines"] = "Newsette",
  ["newsletter (american academy of psychiatry and the law)"] = "Newsl Am Acad Psychiatry Law",
  ["newsletter (canadian society for the history of medicine)"] = "Newsl Can Soc Hist Med",
  ["newsletter (infant feeding action coalition)"] = "Newsl Infant Feed Action Coalit",
  ["newsletter (infant feeding action coalition)."] = "Newsl Infant Feed Action Coalit",
  ["newsletter (inter-african committee on traditional practices affecting the health of women and children)"] = "Newsl Inter Afr Comm Tradit Pract Affect Health Women Child",
  ["newsletter (international academy of periodontology)"] = "Newsl Int Acad Periodontol",
  ["newsletter (macro systems. institute for resource development. demographic and health surveys)"] = "Newsl Macro Syst Inst Resour Dev Demogr Health Surv",
  ["newsletter (medical sciences historical society (great britain))"] = "Newsl Med Sci Hist Soc",
  ["newsletter (society for ancient medicine & pharmacy)"] = "Newsl Soc Anc Med Pharm",
  ["newsletter (society for ancient medicine)"] = "Newsl Soc Anc Med",
  ["newsletter (women's global network on reproductive rights)"] = "Newsl Womens Glob Netw Reprod Rights",
  ["newsletter (women’s global network on reproductive rights)"] = "Newsl Womens Glob Netw Reprod Rights",
  ["newsletter - american academy of health administration. american academy of health administration"] = "Newsl Am Acad Health Adm",
  ["newsletter - american society of anesthesiologists"] = "Newsl Am Soc Anesthesiol",
  ["newsletter - family planning association of hong kong / family planning association of hong kong"] = "Newsl Fam Plan Assoc Hong Kong",
  ["newsletter - family planning association of hong kong. family planning association of hong kong"] = "Newsl Fam Plan Assoc Hong Kong",
  ["newsletter - family planning international assistance / planned parenthood federation of america. family planning international assistance"] = "Newsl Fam Plan Int Assist",
  ["newsletter - family planning international assistance. planned parenthood federation of america. family planning international assistance"] = "Newsl Fam Plan Int Assist",
  ["newsletter - group for the use of psychology in history. group for the use of psychology in history"] = "Newsl Group Use Psychol Hist",
  ["newsletter - springfield dental society"] = "Newsl Springfield Dent Soc",
  ["newsletter / american association of tissue banks"] = "Newsl Am Assoc Tissue Banks",
  ["newsletter / international union for the scientific study of population. international union for the scientific study of population"] = "Newsl Int Union Sci Study Popul",
  ["newsletter of biomedical safety and standards"] = "Newsl. Biomed. Saf. Stand.",
  ["newsletter of the american academy of implant dentistry"] = "Newsl. Am. Acad. Implant Dent.",
  ["newsletter of the netherlands institute at athens"] = "NewsletterAthen",
  ["newsletter of the society for ancient medicine and pharmacy"] = "SAMPhN",
  ["newsletter on feminism and philosophy"] = "Newsl Feminism Philos",
  ["newsletter on feminism and philosophy / american philosophical association"] = "Newsl Feminism Philos",
  ["newsletter on philosophy and medicine"] = "APA Newsl Philos Med",
  ["newsletter on philosophy and medicine / american philosophical association"] = "APA Newsl Philos Med",
  ["newsletter on science, technology & human values"] = "Newsl Sci Technol Human Values",
  ["newsletter, american academy of health administration"] = "Newsl. Am. Acad. Health Adm.",
  ["newsletter, history of anthropology"] = "Newsl Hist Anthropol",
  ["newsletter, international academy of periodontology"] = "Newsl. Int. Acad. Periodontol.",
  ["newsletter, international college of dentists, india section"] = "Newsl. Int. Coll. Dent. India Sect.",
  ["newsletter, springfield dental society"] = "Newsl. Springfield Dent. Soc.",
  ["newsletter, wisconsin league for nursing"] = "Newsl. Wis. League Nurs.",
  ["newsletter-social welfare history group. social welfare history group"] = "Newsl Soc Welf Hist Group",
  ["newsletter. american academy of implant dentistry"] = "Newsl Am Acad Implant Dent",
  ["newsletter. american academy of pediatrics"] = "Newsl Am Acad Pediatr",
  ["newsletter. american epilepsy league"] = "Newsl Am Epilepsy Leag",
  ["newsletter. british society for the history of science"] = "Newsl Br Soc Hist Sci",
  ["newsletter. department of pottery technology, university of leiden"] = "NewsletterPotTech",
  ["newsletter. institute of inter-american affairs (u.s.). health and sanitation division"] = "Newsl Inst Interam Aff",
  ["newsletter. international college of dentists. india section"] = "Newsl Int Coll Dent India Sect",
  ["newsletter. international union for the scientific study of population"] = "Newsl Int Union Sci Study Popul",
  ["newsletter. kentucky nurses association"] = "Newsl Ky Nurses Assoc",
  ["newsletter. nepal mathematical society"] = "Newsl. Nepal Math. Soc.",
  ["newsletter. oral history association"] = "Newsl Oral Hist Assoc",
  ["newsletter. wisconsin league for nursing"] = "Newsl Wis League Nurs",
  ["newsletters on stratigraphy"] = "Newsl Stratigr",
  ["newsline (people with aids coalition of new york)"] = "Newsline People AIDS Coalit N Y",
  ["newsmonthly. american dental society of anesthesiology"] = "Newsmonthly",
  ["newspaper research journal"] = "Newsp Res J",
  ["newswatch (ikeja, lagos state, nigeria)"] = "Newswatch",
  ["next generation, sequencing & applications"] = "Next Gener Seq Appl",
  ["nexus network journal"] = "Nexus Netw J",
  ["néphrologie & thérapeutique"] = "Nephrol Ther",
  ["nfais newsletter"] = "NFAIS Newsl",
  ["nfi bulletin"] = "NFI Bull",
  ["nfpa journal"] = "NFPA J.",
  ["nfpa journal : the official magazine of the national fire protection association"] = "NFPA J",
  ["nfs journal"] = "NFS J.",
  ["nghiên cứu lịch sử"] = "Nghien Cuu Lich Su",
  ["nhpf issue brief"] = "NHPF Issue Brief",
  ["nhpf issue brief / national health policy forum, george washington university"] = "NHPF Issue Brief",
  ["nhsa dialog"] = "NHSA Dialog",
  ["ni 2012 : 11th international congress on nursing informatics, june 23-27, 2012, montreal, canada. international congress in nursing informatics (11th : 2012 : montréal, québec)"] = "NI 2012 (2012)",
  ["niagara frontier"] = "Niagara Frontier",
  ["niaid aids agenda"] = "NIAID AIDS Agenda",
  ["niaid aids agenda / national institute of allergy and infectious diseases"] = "NIAID AIDS Agenda",
  ["nicaragua medica"] = "Nicar Med",
  ["nicaragua médica"] = "Nicar Med",
  ["nicaragua odontologica"] = "Nicar Odontol",
  ["nice historique"] = "Nice Hist",
  ["nice medical"] = "Nice Med",
  ["nice médical"] = "Nice Med",
  ["nichibunken newsletter = nihon bunka"] = "Nichibunken Newsl",
  ["nichidai igaku zasshi"] = "Nichidai Igaku Zasshi",
  ["nichidai koko kagaku (nihon university journal of oral science)"] = "Nichidai Koko Kagaku",
  ["nichidai kōkū kagaku = nihon university journal of oral science"] = "Nichidai Koko Kagaku",
  ["nicotine & tobacco research"] = "Nicotin Tob. Res.",
  ["nicotine & tobacco research : official journal of the society for research on nicotine and tobacco"] = "Nicotine Tob Res",
  ["nicotine and tobacco research"] = "Nicotine Tob. Res.",
  ["nida research monograph"] = "NIDA Res Monogr",
  ["niederdeutsche mitteilungen"] = "Niederdtsch Mitt",
  ["niederdeutsche mitteilungen / herausgegeben von der niederdeutschen arbeitsgemeinschaft zu lund"] = "Niederdtsch Mitt",
  ["niederrheinisches jahrbuch"] = "Niederhein Jahrb",
  ["niedersachsisches aerzteblatt"] = "Niedersachs Arztebl",
  ["niedersachsisches zahnarzteblatt"] = "Niedersachs. Zahnarztebl.",
  ["niedersächsisches aerzteblatt"] = "Niedersachs Arztebl",
  ["niedersächsisches zahnärzteblatt"] = "Niedersachs Zahnarztebl",
  ["niedersächsische akademie der geowissenschaften veröffentlichungen"] = "Niedersächs. Akad. Geowiss. Veröff.",
  ["nieuw archief voor wiskunde"] = "Nieuw Arch. Wisk. (4)",
  ["nieuw archief voor wiskunde. vijfde serie"] = "Nieuw Arch. Wiskd. (5)",
  ["nieuwe nederlandse bijdragen tot de geschiedenis der geneeskunde en der natuurwetenschappen"] = "Nieuwe Ned Bijdr Geschied Geneeskd Natuurwet",
  ["nieuwe west-indische gids"] = "Nieuwe West Indische Gids",
  ["nigeria and the classics"] = "Niger Classics",
  ["nigeria magazine"] = "Niger Mag",
  ["nigeria's population : quarterly journal of population activities in nigeria"] = "Niger Pop",
  ["nigerian behavioral sciences journal"] = "Niger Behav Sci J",
  ["nigerian dental journal"] = "Niger Dent J",
  ["nigerian food journal"] = "Niger. Food J.",
  ["nigerian journal of biotechnology"] = "Niger J Biotechnol",
  ["nigerian journal of clinical practice"] = "Niger J Clin Pract",
  ["nigerian journal of economic and social studies"] = "Nigerian J. Econ. Soc. Stud.",
  ["nigerian journal of medicine : journal of the national association of resident doctors of nigeria"] = "Niger J Med",
  ["nigerian journal of natural products and medicine"] = "Niger J Nat Prod Med",
  ["nigerian journal of nutrional sciences"] = "Niger. J. Nutr. Sci.",
  ["nigerian journal of paediatrics"] = "Niger J Paediatr",
  ["nigerian journal of parasitology"] = "Niger. J. Parasitol.",
  ["nigerian journal of pharmaceutical sciences"] = "Niger J Pharm Sci",
  ["nigerian journal of physiological sciences"] = "Niger. J. Physiol. Sci.",
  ["nigerian journal of physiological sciences : official publication of the physiological society of nigeria"] = "Niger J Physiol Sci",
  ["nigerian journal of surgery : official publication of the nigerian surgical research society"] = "Niger J Surg",
  ["nigerian journal of surgical sciences : official journal of the nigerian section of international college of surgeons"] = "Niger J Surg Sci",
  ["nigerian journal of technological development"] = "Niger. J. Technol. Dev.",
  ["nigerian medical journal"] = "Niger. Med. J.",
  ["nigerian medical journal : journal of the nigeria medical association"] = "Niger Med J",
  ["nigerian nurse"] = "Niger. Nurse",
  ["nigerian quarterly journal of hospital medicine"] = "Nig Q J Hosp Med",
  ["nigerian veterinary journal"] = "Niger Vet J",
  ["nigeria’s population : quarterly journal of population activities in nigeria"] = "Niger Pop",
  ["nih conference on retroviral integrase : molecular biology and pharmacology : a novel target for the treatment of aids, january 19-20, 1995. nih conference on retroviral integrase (1995 : bethesda, md.)"] = "NIH Conf Retroviral Integr NIH Conf Retroviral Integr 1995 Bethesda Md",
  ["nih consensus and state-of-the-science statements"] = "NIH Consens State Sci Statements",
  ["nih consensus development conference summary"] = "NIH Consens Dev Conf Summ",
  ["nih consensus statement"] = "NIH Consens Statement",
  ["nih guide for grants and contracts"] = "NIH Guide Grants Contracts",
  ["nih guide for grants and contracts (online)"] = "NIH Guide Grants Contracts",
  ["nih guide for grants and contracts / u.s. department of health, education, and welfare"] = "NIH Guide Grants Contracts",
  ["nihae bulletin"] = "NIHAE Bull",
  ["nihon ago kansetsu gakkai zasshi"] = "Nihon Ago Kansetsu Gakkai Zasshi",
  ["nihon arukoru yakubutsu igakkai zasshi"] = "Nihon Arukoru Yakubutsu Igakkai Zasshi",
  ["nihon arukoru yakubutsu igakkai zasshi (japanese journal of alcohol studies and drug dependence)"] = "Nihon Arukoru Yakubutsu Igakkai Zasshi",
  ["nihon arukōru yakubutsu igakkai zasshi = japanese journal of alcohol studies & drug dependence"] = "Nihon Arukoru Yakubutsu Igakkai Zasshi",
  ["nihon bentosu gakkai shi = benthos research"] = "Nihon Bentosu Gakkai Shi",
  ["nihon biseibutsu seitai gakkaihō = bulletin of japanese society of microbial ecology"] = "Nihon Biseibutsu Seitai Gakkaiho",
  ["nihon byori gakkai kaishi"] = "Nihon Byori Gakkai Kaishi",
  ["nihon chokucho komonbyo gakkai zasshi"] = "Nihon Chokucho Komonbyo",
  ["nihon daicho komonbyo gakkai zasshi"] = "Nihon Daicho Komonbyo Gakkai Zasshi",
  ["nihon dani gakkai shi = journal of the acarological society of japan"] = "Nihon Dani Gakkai Shi",
  ["nihon densenbyo gakkai zasshi"] = "Nihon Densenbyo Gakkai Zasshi",
  ["nihon dojō hiryōgaku zasshi : dojō, hiryō, shokubutsu eiyō"] = "Nihon Dojo Hiryogaku Zasshi",
  ["nihon eiseigaku zasshi. japanese journal of hygiene"] = "Nihon Eiseigaku Zasshi",
  ["nihon eiyō shokuryō gakkai shi = nippon eiyō shokuryō gakkaishi = journal of japanese society of nutrition and food science"] = "Nihon Eiyo Shokuryo Gakkai Shi",
  ["nihon funin gakkai zasshi"] = "Nihon Funin Gakkai Zasshi",
  ["nihon gan chiryo gakkai shi"] = "Nihon Gan Chiryo Gakkai Shi",
  ["nihon ganka kiyo"] = "Nihon Ganka Kiyo",
  ["nihon geka gakkai zasshi"] = "Nihon Geka Gakkai Zasshi",
  ["nihon geka hokan. archiv für japanische chirurgie"] = "Nihon Geka Hokan",
  ["nihon hansenbyo gakkai zasshi (japanese journal of leprosy)"] = "Nihon Hansenbyo Gakkai Zasshi",
  ["nihon hansenbyo gakkai zasshi. japanese journal of leprosy (higashimurayama- shi"] = "Nihon Hansenbyo Gakkai Zasshi",
  ["nihon hansenbyō gakkai zasshi = japanese journal of leprosy : official organ of the japanese leprosy association"] = "Nihon Hansenbyo Gakkai Zasshi",
  ["nihon heikatsukin gakkai zasshi"] = "Nihon Heikatsukin Gakkai Zasshi",
  ["nihon hifuka gakkai zasshi. the japanese journal of dermatology"] = "Nihon Hifuka Gakkai Zasshi",
  ["nihon hinyōkika gakkai zasshi. the japanese journal of urology"] = "Nihon Hinyokika Gakkai Zasshi",
  ["nihon hosenkin gakkai shi"] = "Actinomycetologica|Nippon Hosenkin Gakkaishi",
  ["nihon hotetsu shika gakkai zasshi"] = "Nihon Hotetsu Shika Gakkai Zasshi",
  ["nihon hōigaku zasshi = the japanese journal of legal medicine"] = "Nihon Hoigaku Zasshi",
  ["nihon hōsenkin gakkai shi = actinomycetologica"] = "Nippon Hosenkin Gakkaishi",
  ["nihon hōshasen gijutsu gakkai zasshi"] = "Nihon Hoshasen Gijutsu Gakkai Zasshi",
  ["nihon igaku hoshasen gakkai zasshi. nippon acta radiologica"] = "Nippon Igaku Hoshasen Gakkai Zasshi",
  ["nihon igaku hōshasen gakkai zasshi. nippon acta radiologica"] = "Nihon Igaku Hoshasen Gakkai Zasshi",
  ["nihon ika daigaku zasshi"] = "Nihon Ika Daigaku Zasshi",
  ["nihon ishigaku zasshi. [journal of japanese history of medicine]"] = "Nihon Ishigaku Zasshi",
  ["nihon ishikai zasshi. journal of the japan medical association"] = "Nihon Ishikai Zasshi",
  ["nihon ishinkin gakkai zasshi = japanese journal of medical mycology"] = "Nihon Ishinkin Gakkai Zasshi",
  ["nihon jibiinkoka gakkai kaiho"] = "Nihon Jibiinkoka Gakkai Kaiho",
  ["nihon jinzo gakkai shi"] = "Nihon Jinzo Gakkai Shi",
  ["nihon juigaku zasshi. the japanese journal of veterinary science"] = "Nihon Juigaku Zasshi",
  ["nihon kakin gakkaishi"] = "Japanese poultry science|Nippon Kakin Gakkaishi",
  ["nihon kakin gakkaishi = japanese poultry science"] = "Nippon Kakin Gakkaishi",
  ["nihon kango kagakkai shi (journal of japan academy of nursing science)"] = "Nihon Kango Kagakkaishi",
  ["nihon kango kagakkai shi = journal of japan academy of nursing science"] = "Nihon Kango Kagakkaishi",
  ["nihon kessen shiketsu gakkai shi = japanese journal of thrombosis and hemostasis"] = "Nihon Kessen Shiketsu Gakkai shi",
  ["nihon ketsueki gakkai zasshi : journal of japan haematological society"] = "Nihon Ketsueki Gakkai Zasshi",
  ["nihon kingakkai kaiho"] = "Nihon Kingakkai Kaiho",
  ["nihon koku geka gakkai zasshi"] = "Nihon Koku Geka Gakkai Zasshi",
  ["nihon kokyuki gakkai zasshi (journal of the japanese respiratory society)"] = "Nihon Kokyuki Gakkai Zasshi",
  ["nihon kokyūki gakkai zasshi = the journal of the japanese respiratory society"] = "Nihon Kokyuki Gakkai Zasshi",
  ["nihon kōsoshiki kenkyū gijutsu gakkai zasshi = journal of hard tissue biology. biology"] = "J Hard Tissue Biol",
  ["nihon kyobu shikkan gakkai zasshi"] = "Nihon Kyobu Shikkan Gakkai Zasshi",
  ["nihon kyōbu shikkan gakkai zasshi"] = "Nihon Kyobu Shikkan Gakkai Zasshi",
  ["nihon kyōsei shika gakkai zasshi = the journal of japan orthodontic society"] = "Nihon Kyosei Shika Gakkai Zasshi",
  ["nihon masu sukurīningu gakkai shi = journal of japanese society for mass-screening"] = "Nihon Masu Sukuriningu Gakkai Shi",
  ["nihon naibunpi gakkai zasshi"] = "Nihon Naibunpi Gakkai Zasshi",
  ["nihon naika gakkai zasshi. the journal of the japanese society of internal medicine"] = "Nihon Naika Gakkai Zasshi",
  ["nihon rai gakkai zasshi"] = "Nihon Rai Gakkai Zasshi",
  ["nihon rekishi = the nihon-rekishi"] = "Nihon Rekishi",
  ["nihon reoroji gakkaishi"] = "Nihon Reoroji Gakkaishi",
  ["nihon rinsho meneki gakkai kaishi"] = "Nihon Rinsho Meneki Gakkai Kaishi",
  ["nihon rinsho meneki gakkai kaishi (japanese journal of clinical immunology)"] = "Nihon Rinsho Meneki Gakkai Kaishi",
  ["nihon rinsho shishubyo danwakai kaishi"] = "Nihon Rinsho Shishubyo Danwakai Kaishi",
  ["nihon rinsho shishubyo danwakai kaishi (journal of the japanese society of periodontists)"] = "Nihon Rinsho Shishubyo Danwakai Kaishi",
  ["nihon rinsho. japanese journal of clinical medicine"] = "Nihon Rinsho",
  ["nihon rinshō men'eki gakkai kaishi = japanese journal of clinical immunology"] = "Nihon Rinsho Meneki Gakkai Kaishi",
  ["nihon rinshō shishūbyō danwakai kaishi"] = "Nihon Rinsho Shishubyo Danwakai Kaishi",
  ["nihon ronen igakkai zasshi. japanese journal of geriatrics"] = "Nihon Ronen Igakkai Zasshi",
  ["nihon saikingaku zasshi. japanese journal of bacteriology"] = "Nihon Saikingaku Zasshi",
  ["nihon sakumotsu gakkai kiji"] = "Jpn J Crop Sci",
  ["nihon sanka fujinka gakkai zasshi"] = "Nihon Sanka Fujinka Gakkai Zasshi",
  ["nihon sanshigaku zasshi"] = "The journal of sericultural science|Nihon Sanshigaku Zasshi",
  ["nihon sanshigaku zasshi = the journal of sericultural science"] = "Nihon Sanshigaku Zasshi",
  ["nihon seikeigeka gakkai zasshi"] = "Nihon Seikeigeka Gakkai Zasshi",
  ["nihon seirigaku zasshi. journal of the physiological society of japan"] = "Nihon Seirigaku Zasshi",
  ["nihon senchu gakkai shi"] = "Japanese journal of nematology|Nihon Senchu Gakkai Shi",
  ["nihon senchū gakkai shi = japanese journal of nematology"] = "Nihon Senchu Gakkai Shi",
  ["nihon shi kenkyu"] = "Nihonshi Kenkyu",
  ["nihon shi kenkyū"] = "Nihonshi Kenkyu",
  ["nihon shika ishikai zasshi"] = "Nihon Shika Ishikai Zasshi",
  ["nihon shika zairyo kikai gakkai zasshi. the journal of the japan research society of dental materials & appliances"] = "Nihon Shika Zairyo Kikai Gakkai Zasshi",
  ["nihon shinkei seishin yakurigaku zasshi (japanese journal of psychopharmacology)"] = "Nihon Shinkei Seishin Yakurigaku Zasshi",
  ["nihon shinkei seishin yakurigaku zasshi (kawasaki-shi"] = "Nihon Shinkei Seishin Yakurigaku Zasshi",
  ["nihon shinkei seishin yakurigaku zasshi = japanese journal of psychopharmacology"] = "Nihon Shinkei Seishin Yakurigaku Zasshi",
  ["nihon shishubyo gakkai kaishi"] = "Nihon Shishubyo Gakkai Kaishi",
  ["nihon shokaki geka gakkai zasshi"] = "Nihon Shokaki Geka Gakkai Zasshi",
  ["nihon shokakibyo gakkai zasshi = the japanese journal of gastro-enterology"] = "Nihon Shokakibyo Gakkai Zasshi",
  ["nihon shokubutsu byori gakkaiho"] = "Annals of the Phytopathological Society of Japan|Nippon Shokubutsu Byori Gakkaiho",
  ["nihon shokubutsu byōri gakkaiho = annals of the phytopathological society of japan"] = "Nippon Shokubutsu Byori Gakkaiho",
  ["nihon shonika gakkai zasshi. acta paediatrica japonica"] = "Nihon Shonika Gakkai Zasshi",
  ["nihon shoyu kenkyujo zasshi"] = "Journal of Japan Soy Sauce Research Institute|Nihon Shoyu Kenkyujo Zasshi",
  ["nihon shōni junkanki gakkai zasshi = pediatric cardiology and cardiac surgery"] = "Nihon Shoni Junkanki Gakkai Zasshi",
  ["nihon shōyu kenkyūjo zasshi = journal of japan soy sauce research institute"] = "Nihon Shoyu Kenkyujo Zasshi",
  ["nihon sutoma rihabiriteshon gakkaishi [journal of japanese society of stoma rehabilitation]"] = "Nihon Sutoma Rihabiriteshon Gakkaishi J Jpn Soc Stoma Rehabil",
  ["nihon yakurigaku zasshi. folia pharmacologica japonica"] = "Nihon Yakurigaku Zasshi",
  ["nihon yuketsu gakkai zasshi = journal of the japan society of blood transfusion"] = "Nihon Yuketsu Gakkai Zasshi",
  ["nihonkai mathematical journal"] = "Nihonkai Math. J.",
  ["niigata igakkai zasshi. niigata medical journal"] = "Niigata Igakkai Zasshi",
  ["nikephoros. zeitschrift für sport und kultur im altertum"] = "Nikephoros",
  ["nin. journal of gender studies in antiquity"] = "Nin",
  ["nina rapport"] = "NINA Rapp",
  ["ninchi shinkei kagaku"] = "Ninchi Shinkei Kagaku",
  ["nineteenth century (philadelphia, pa.)"] = "Ninet Century",
  ["nineteenth-century contexts"] = "Ninet Century Contexts",
  ["nineteenth-century fiction"] = "Nineteenth Century Fict",
  ["nineteenth-century french studies"] = "Ninet Century Fr Stud",
  ["nineteenth-century literature"] = "Ninet Century Lit",
  ["ninety nine"] = "Ninety Nine",
  ["ningxia yixueyuan xuebao. ningxia medical college"] = "Ningxia Yixueyuan Xuebao",
  ["niph annals"] = "NIPH Ann",
  ["nippon byori gakkai kaishi"] = "Nippon Byori Gakkai Kaishi",
  ["nippon byori gakkai kaishi (transactiones societatis pathologicae japonicae)"] = "Nippon Byori Gakkai Kaishi",
  ["nippon chokucho komonbyo gakkai zasshi"] = "Nippon Chokucho Komonbyo",
  ["nippon daicho komonbyo gakkai zasshi"] = "Nippon Daicho Komonbyo Gakkai Zasshi",
  ["nippon daicho komonbyo gakkai zasshi (journal of japan society of colo-proctology)"] = "Nippon Daicho Komonbyo Gakkai Zasshi",
  ["nippon densenbyo gakkai zasshi"] = "Nippon Densenbyo Gakkai Zasshi",
  ["nippon eiseigaku zasshi (japanese journal of hygiene)"] = "Nippon Eiseigaku Zasshi",
  ["nippon eiseigaku zasshi. japanese journal of hygiene"] = "Nippon Eiseigaku Zasshi",
  ["nippon funin gakkai zasshi"] = "Nippon Funin Gakkai Zasshi",
  ["nippon funin gakkai zasshi (japanese journal of fertility and sterility)"] = "Nippon Funin Gakkai Zasshi",
  ["nippon gan chiryo gakkai shi"] = "Nippon Gan Chiryo Gakkai Shi",
  ["nippon gan chiryo gakkai shi (journal of japan society for cancer therapy)"] = "Nippon Gan Chiryo Gakkai Shi",
  ["nippon ganka gakkai zasshi"] = "Nippon Ganka Gakkai Zasshi",
  ["nippon ganka gakkai zasshi (acta societatis ophthalmologicae japonicae)"] = "Nippon Ganka Gakkai Zasshi",
  ["nippon ganka gakkai zasshi. acta societatis ophthalmologicae japonicae"] = "Nippon Ganka Gakkai Zasshi",
  ["nippon ganka kiyo"] = "Nippon Ganka Kiyo",
  ["nippon ganka kiyo (folia ophthalmologica japonica. bulletin of japanese ophthalmology)"] = "Nippon Ganka Kiyo",
  ["nippon geka gakkai zasshi"] = "Nippon Geka Gakkai Zasshi",
  ["nippon geka gakkai zasshi (journal of japan surgical society)"] = "Nippon Geka Gakkai Zasshi",
  ["nippon geka gakkai zasshi. journal of japan surgical society"] = "Nippon Geka Gakkai Zasshi",
  ["nippon geka hokan (archiv fur japanische chirurgie)"] = "Nippon Geka Hokan.",
  ["nippon geka hokan. archiv fur japanische chirurgie"] = "Nippon Geka Hokan",
  ["nippon heikatsukin gakkai zasshi"] = "Nippon Heikatsukin Gakkai Zasshi",
  ["nippon hifuka gakkai zasshi (japanese journal of dermatology)"] = "Nippon Hifuka Gakkai Zasshi",
  ["nippon hifuka gakkai zasshi. the japanese journal of dermatology"] = "Nippon Hifuka Gakkai Zasshi",
  ["nippon hinyokika gakkai zasshi (japanese journal of urology)"] = "Nippon Hinyokika Gakkai Zasshi",
  ["nippon hinyokika gakkai zasshi. japanese journal of urology"] = "Nippon Hinyokika Gakkai Zasshi",
  ["nippon hinyokika gakkai zasshi. the japanese journal of urology"] = "Nippon Hinyokika Gakkai Zasshi",
  ["nippon hoigaku zasshi (japanese journal of legal medicine)"] = "Nippon Hoigaku Zasshi",
  ["nippon hoigaku zasshi. japanese journal of legal medicine"] = "Nippon Hoigaku Zasshi",
  ["nippon hoshasen gijutsu gakkai zasshi"] = "Nippon Hoshasen Gijutsu Gakkai Zasshi",
  ["nippon hotetsu shika gakkai zasshi"] = "Nippon Hotetsu Shika Gakkai Zasshi",
  ["nippon igaku hoshasen gakkai zasshi (nippon acta radiologica)"] = "Nippon Igaku Hoshasen Gakkai Zasshi",
  ["nippon igaku hoshasen gakkai zasshi. nippon acta radiologica"] = "Nippon Igaku Hoshasen Gakkai Zasshi",
  ["nippon ika daigaku zasshi"] = "Nippon Ika Daigaku Zasshi",
  ["nippon ika daigaku zasshi (journal of the nippon medical school)"] = "Nippon Ika Daigaku Zasshi",
  ["nippon ika daigaku zasshi. journal of the nippon medical school"] = "Nippon Ika Daigaku Zasshi",
  ["nippon ishikai zasshi (journal of the japan medical association)"] = "Nippon Ishikai Zasshi",
  ["nippon ishikai zasshi. journal of the japan medical association"] = "Nippon Ishikai Zasshi",
  ["nippon ishinkin gakkai zasshi (japanese journal of medical mycology)"] = "Nippon Ishinkin Gakkai Zasshi",
  ["nippon jibiinkoka gakkai kaiho"] = "Nippon Jibiinkoka Gakkai Kaiho",
  ["nippon jibiinkoka gakkai kaiho [journal of the oto-rhino-laryngological society of japan]"] = "Nippon Jibiinkoka Gakkai Kaiho",
  ["nippon jinzo gakkai shi"] = "Nippon Jinzo Gakkai Shi",
  ["nippon jinzo gakkai shi. japanese journal of nephrology"] = "Nippon Jinzo Gakkai Shi",
  ["nippon juigaku zasshi (japanese journal of veterinary science)"] = "Nippon Juigaku Zasshi",
  ["nippon juigaku zasshi. the japanese journal of veterinary science"] = "Nippon Juigaku Zasshi",
  ["nippon juishikai zasshi. journal of the japan veterinary medical association"] = "Nippon Juishikai Zasshi",
  ["nippon kagaku ryoho gakkai zasshi. chemotherapy"] = "Nippon Kagaku Ryoho Gakkai Zasshi",
  ["nippon ketsueki gakkai zasshi (acta haematologica japonica)"] = "Nippon Ketsueki Gakkai Zasshi",
  ["nippon ketsueki gakkai zasshi : journal of japan haematological society"] = "Nippon Ketsueki Gakkai Zasshi",
  ["nippon kikai gakkai ronbunshu, a-hen"] = "Nippon Kikai Gakkai Ronbunshu, A-hen",
  ["nippon kikai gakkai ronbunshu, b-hen"] = "Nippon Kikai Gakkai Ronbunshu, B-hen",
  ["nippon kinzoku gakkaishi"] = "Nippon Kinzoku Gakkaishi",
  ["nippon koku geka gakkai zasshi"] = "Nippon Koku Geka Gakkai Zasshi",
  ["nippon kontakuto renzu gakkai kaiin dayori (journal of japan contact lens society)"] = "Nippon Kontakuto Renzu Gakkai Kaiin Dayori.",
  ["nippon koshu eisei zasshi (japanese journal of public health)"] = "Nippon Koshu Eisei Zasshi",
  ["nippon koshu eisei zasshi [japanese journal of public health]"] = "Nippon Koshu Eisei Zasshi",
  ["nippon kyobu geka gakkai zasshi (journal of the japanese association for thoracic surgery)"] = "Nippon Kyobu Geka Gakkai Zasshi",
  ["nippon kyobu geka gakkai zasshi. journal of the japanese association for thoracic surgery"] = "Nippon Kyobu Geka Gakkai Zasshi",
  ["nippon kyobu shikkan gakkai zasshi (japanese journal of thoracic diseases)"] = "Nippon Kyobu Shikkan Gakkai Zasshi",
  ["nippon kyobu shikkan gakkai zasshi. japanese journal of thoracic diseases"] = "Nippon Kyobu Shikkan Gakkai Zasshi",
  ["nippon kyosei shika gakkai zasshi (journal of japan orthodontic society)"] = "Nippon Kyosei Shika Gakkai Zasshi",
  ["nippon naibunpi gakkai zasshi"] = "Nippon Naibunpi Gakkai Zasshi",
  ["nippon naibunpi gakkai zasshi (folia endocrinologica japonica)"] = "Nippon Naibunpi Gakkai Zasshi",
  ["nippon naika gakkai zasshi (journal of japanese society of internal medicine)"] = "Nippon Naika Gakkai Zasshi",
  ["nippon naika gakkai zasshi. journal of japanese society of internal medicine"] = "Nippon Naika Gakkai Zasshi",
  ["nippon nettai igakkai zasshi. japanese journal of tropical medicine and hygiene"] = "Nippon Nettai Igakkai Zasshi",
  ["nippon nôgeikagaku kaishi"] = "Nippon Nogeikagaku Kaishi",
  ["nippon oyo jiki gakkaishi"] = "Nippon Oyo Jiki Gakkaishi",
  ["nippon rai gakkai zasshi"] = "Nippon Rai Gakkai Zasshi",
  ["nippon rai gakkai zasshi (japanese journal of leprosy)"] = "Nippon Rai Gakkai Zasshi",
  ["nippon rinsho (japanese journal of clinical medicine)"] = "Nippon Rinsho",
  ["nippon rinsho. japanese journal of clinical medicine"] = "Nippon Rinsho",
  ["nippon ronen igakkai zasshi (japanese journal of geriatrics)"] = "Nippon Ronen Igakkai Zasshi",
  ["nippon ronen igakkai zasshi. japanese journal of geriatrics"] = "Nippon Ronen Igakkai Zasshi",
  ["nippon saikingaku zasshi (japanese journal of bacteriology)"] = "Nippon Saikingaku Zasshi",
  ["nippon saikingaku zasshi. japanese journal of bacteriology"] = "Nippon Saikingaku Zasshi",
  ["nippon sanka fujinka gakkai zasshi"] = "Nippon Sanka Fujinka Gakkai Zasshi",
  ["nippon sanka fujinka gakkai zasshi (acta obstetrica et gynaecologica japonica)"] = "Nippon Sanka Fujinka Gakkai Zasshi",
  ["nippon seikeigeka gakkai zasshi"] = "Nippon Seikeigeka Gakkai Zasshi",
  ["nippon seikeigeka gakkai zasshi (journal of the japanese orthopaedic association)"] = "Nippon Seikeigeka Gakkai Zasshi",
  ["nippon seirigaku zasshi (journal of the physiological society of japan)"] = "Nippon Seirigaku Zasshi",
  ["nippon seirigaku zasshi. journal of the physiological society of japan"] = "Nippon Seirigaku Zasshi",
  ["nippon seramikkusu kyokai gakujutsu ronbunshi"] = "Nippon Seramikkusu Kyokai Gakujutsu Ronbunshi",
  ["nippon shika hyoron (nippon dental review)"] = "Nippon Shika Hyoron.",
  ["nippon shika ishikai zasshi"] = "Nippon Shika Ishikai Zasshi",
  ["nippon shika ishikai zasshi (journal of the japan dental association)"] = "Nippon Shika Ishikai Zasshi",
  ["nippon shika zairyo kikai gakkai zasshi (journal of the japan research society of dental materials and appliances)"] = "Nippon Shika Zairyo Kikai Gakkai Zasshi",
  ["nippon shika zairyo kikai gakkai zasshi. the journal of the japan research society of dental materials & appliances"] = "Nippon Shika Zairyo Kikai Gakkai Zasshi",
  ["nippon shishubyo gakkai kaishi"] = "Nippon Shishubyo Gakkai Kaishi",
  ["nippon shishubyo gakkai kaishi (journal of the japanese association of periodontology)"] = "Nippon Shishubyo Gakkai Kaishi",
  ["nippon shokakibyo gakkai zasshi (japanese journal of gastroenterology)"] = "Nippon Shokakibyo Gakkai Zasshi",
  ["nippon shokakibyo gakkai zasshi the japanese journal of gastro-enterology"] = "Nippon Shokakibyo Gakkai Zasshi",
  ["nippon shokakibyo gakkai zasshi. japanese journal of gastroenterology"] = "Nippon Shokakibyo Gakkai Zasshi",
  ["nippon shokuhin kogyo gakkaishi"] = "Nippon Shokuhin Kogyo Gakkaishi",
  ["nippon shonika gakkai zasshi (acta paediatrica japonica)"] = "Nippon Shonika Gakkai Zasshi",
  ["nippon shonika gakkai zasshi. acta paediatrica japonica"] = "Nippon Shonika Gakkai Zasshi",
  ["nippon suisan gakkai shi. bulletin of the japanese society of scientific fisheries"] = "Nippon Suisan Gakkai Shi",
  ["nippon yakurigaku zasshi"] = "Nippon Yakurigaku Zasshi",
  ["nippon yakurigaku zasshi (folia pharmacologica japonica)"] = "Nippon Yakurigaku Zasshi",
  ["nippon yakurigaku zasshi. folia pharmacologica japonica"] = "Nippon Yakurigaku Zasshi",
  ["nir news"] = "NIR News",
  ["nishi nihon hinyokika = [west japan journal of urology]"] = "Nishi Nihon Hinyokika",
  ["nist special publication"] = "NIST Spec. Publ.",
  ["nitric oxide"] = "Nitric Oxide",
  ["nitric oxide : biology and chemistry"] = "Nitric Oxide",
  ["nitric oxide : biology and chemistry / official journal of the nitric oxide society"] = "Nitric Oxide",
  ["niwa information series"] = "NIWA Inf Ser",
  ["nizhegorodskiy gosudavstvennyy universitet"] = "Dinamika Sistem",
  ["njas - wageningen journal of life sciences"] = "NJAS Wageningen J. Life Sci.",
  ["njas: impact in agricultural and life sciences"] = "NJAS: Impact Agric. Life Sci.",
  ["njpc bulletin"] = "NJPC Bull",
  ["njsna newsletter"] = "NJSNA Newsl",
  ["nlm gateway"] = "NLM Gatew",
  ["nln convention papers"] = "NLN Conv Pap",
  ["nln publications"] = "NLN Publ",
  ["nln update"] = "NLN Update",
  ["nln update : connecting members of the national league for nursing"] = "NLN Update",
  ["nmc case report journal"] = "NMC Case Rep J",
  ["nmr in biomedicine"] = "NMR Biomed",
  ["no junkan taisha = cerebral blood flow and metabolism"] = "No Junkan Taisha",
  ["no shinkei geka (neurological surgery)"] = "No Shinkei Geka.",
  ["no shinkei geka. neurological surgery"] = "No Shinkei Geka",
  ["no to hattatsu"] = "No To Hattatsu",
  ["no to hattatsu = brain and development"] = "No To Hattatsu",
  ["no to hattatsu [brain and development]"] = "No To Hattatsu",
  ["no to hattatsu. brain and development"] = "No To Hattatsu",
  ["no to shinkei (brain and nerve)"] = "No To Shinkei",
  ["no to shinkei. brain and nerve"] = "No To Shinkei",
  ["noaa technical memorandum nmfs-swfsc"] = "NOAA Tech Memo NMFS SWFSC",
  ["noaa technical report nmfs"] = "NOAA Tech Rep NMFS",
  ["nobel laureate series"] = "Nobel Laureate Ser.",
  ["nobel symposium"] = "Nobel Symp",
  ["nodea. nonlinear differential equations and applications"] = "NoDEA Nonlinear Differential Equations Appl.",
  ["nodycon conference proceedings series"] = "NODYCON Conf. Proc. Ser.",
  ["noesis (pau, france)"] = "Noesis",
  ["noise & health"] = "Noise Health",
  ["noise & vibration worldwide"] = "Noise Vibr. Worldwide",
  ["noise control engineering journal"] = "Noise Control Eng J",
  ["nomadic peoples"] = "Nomad People",
  ["nomura research institute quarterly economic review"] = "Nomura Res. Inst. Quart. Econ. Rev.",
  ["non-coding rna"] = "Noncoding RNA",
  ["non-coding rna investigation"] = "Noncoding RNA Investig",
  ["non-coding rna research"] = "Noncoding RNA Res",
  ["non-destructive testing"] = "Non-Destr. Test.",
  ["non-pharmacological therapies in dementia"] = "Nonpharmacol Ther Dement",
  ["nonautonomous dynamical systems"] = "Nonauton. Dyn. Syst.",
  ["nonconvex optimization and its applications"] = "Nonconvex Optim. Appl.",
  ["nondestructive testing and evaluation"] = "Nondestr.Test. Eval.",
  ["nong ye huan jing ke xue xue bao = journal of agro-environment science"] = "Nong Ye Huan Jing Ke Xue Xue Bao",
  ["nong ye sheng wu ji shu xue bao"] = "Journal of agricultural biotechnology|Nong Ye Sheng Wu Ji Shu Xue Bao",
  ["nong ye sheng wu ji shu xue bao = journal of agricultural biotechnology"] = "Nong Ye Sheng Wu Ji Shu Xue Bao",
  ["nong'yag gwahag hoeji"] = "Nongyag Gwahag Hoeji",
  ["nongyaoxue xuebao"] = "Nongyaoxue Xuebao",
  ["nonlinear analysis"] = "Nonlinear Anal.",
  ["nonlinear analysis forum"] = "Nonlinear Anal. Forum",
  ["nonlinear analysis forum. an international journal"] = "Nonlinear Anal. Forum",
  ["nonlinear analysis, theory, methods & applications"] = "Nonlinear Anal Theory Methods Appl",
  ["nonlinear analysis. hybrid systems"] = "Nonlinear Anal. Hybrid Syst.",
  ["nonlinear analysis. modelling and control"] = "Nonlinear Anal. Model. Control",
  ["nonlinear analysis. real world applications"] = "Nonlinear Anal Real World Appl",
  ["nonlinear analysis. real world applications. an international multidisciplinary journal"] = "Nonlinear Anal. Real World Appl.",
  ["nonlinear analysis. theory, methods & applications. an international multidisciplinary journal"] = "Nonlinear Anal.",
  ["nonlinear analysis: hybrid systems"] = "Nonlinear Anal. Hybrid Syst",
  ["nonlinear analysis: modeling and control"] = "Nonlinear Anal. Model. Control",
  ["nonlinear analysis: modelling and control"] = "Nonlinear Anal.-Model. Control",
  ["nonlinear analysis: real world applications"] = "Nonlinear Anal. Real World Appl.",
  ["nonlinear analysis: theory methods & applications"] = "Nonlinear Anal. Theory Methods Appl.",
  ["nonlinear biomedical physics"] = "Nonlinear Biomed Phys",
  ["nonlinear differential equations and applications"] = "Nonlinear Differ. Equations Appl.",
  ["nonlinear dynamics"] = "Nonlinear Dyn.",
  ["nonlinear dynamics and systems theory. an international journal of research and surveys"] = "Nonlinear Dyn. Syst. Theory",
  ["nonlinear dynamics, psychology, and life sciences"] = "Nonlinear Dynamics Psychol Life Sci",
  ["nonlinear optics (mclc) section b"] = "Nonlinear Opt. (Mclc) Sect. B",
  ["nonlinear optics quantum optics"] = "Nonlinear Opt. Quantum Opt.",
  ["nonlinear oscillations"] = "Nonlinear Oscillations",
  ["nonlinear phenomena and complex systems"] = "Nonlinear Phenom. Complex Systems",
  ["nonlinear physical science"] = "Nonlinear Phys. Sci.",
  ["nonlinear processes in geophysics"] = "Nonlinear Processes Geophys.",
  ["nonlinear processes in geophysics discussions"] = "Nonlinear Processes Geophys. Discuss.",
  ["nonlinear science today"] = "Nonlinear Sci. Today",
  ["nonlinear studies"] = "Nonlinear Stud.",
  ["nonlinear studies. the international journal"] = "Nonlinear Stud.",
  ["nonlinear systems and complexity"] = "Nonlinear Syst. Complex.",
  ["nonlinear time series and chaos"] = "Nonlinear Time Ser. Chaos",
  ["nonlinear world"] = "Nonlinear World",
  ["nonlinearity in biology, toxicology, medicine"] = "Nonlinearity Biol Toxicol Med",
  ["nonprofit and voluntary sector quarterly"] = "Nonprofit Volunt Sect Q",
  ["nonprofit management & leadership"] = "Nonprofit Manag Leadersh",
  ["nonprofit management and leadership"] = "Nonprofit Manage. Leadership",
  ["norba. revista de arte, geografía e historia"] = "Norba",
  ["nordic journal of botany"] = "Nord J Bot",
  ["nordic journal of computing"] = "Nordic J. Comput.",
  ["nordic journal of freshwater research"] = "Nord J Freshw Res",
  ["nordic journal of human rights"] = "Nord J Hum Rights",
  ["nordic journal of migration research"] = "Nord J Migr Res",
  ["nordic journal of music therapy"] = "Nord J Music Ther",
  ["nordic journal of nursing research"] = "Nord J Nurs Res",
  ["nordic journal of philosophical logic"] = "Nordic J. Philos. Logic",
  ["nordic journal of political economy"] = "Nordic J. Polit. Economy",
  ["nordic journal of psychiatry"] = "Nord J Psychiatry",
  ["nordic journal of science and technology studies"] = "Nord J Sci Technol Stud",
  ["nordic journal of soviet and east european studies"] = "Nord J Sov East Eur Stud",
  ["nordic notes"] = "Nord Notes",
  ["nordic psychology"] = "Nord Psychol",
  ["nordic pulp & paper research journal"] = "Nord Pulp Paper Res J",
  ["nordic pulp and paper research journal"] = "Nordic Pulp Pap. Res. J.",
  ["nordic wittgenstein studies"] = "Nordic Wittgenstein Stud.",
  ["nordisk alkohol- & narkotikatidskrift : nat"] = "Nordisk Alkohol Nark",
  ["nordisk hygienisk tidskrift"] = "Nord Hyg Tidskr",
  ["nordisk hygienisk tidskrift. supplement"] = "Nord. Hyg. Tidskr. Suppl.",
  ["nordisk hygienisk tidskrift. supplementum"] = "Nord Hyg Tidskr Suppl",
  ["nordisk medicin"] = "Nord Med",
  ["nordisk medicinhistorisk arsbok"] = "Nord Medicinhist Arsb",
  ["nordisk medicinhistorisk årsbok"] = "Nord Medicinhist Arsb",
  ["nordisk medicinhistorisk årsbok. supplementum"] = "Nord Medicinhist Arsb Suppl",
  ["nordisk numismatisk årsskrift. utgiven av kungliga vitterhets historie och antikvitets akademien i samarbete med nordisk numismatisk union"] = "NordNumArs",
  ["nordisk psykiatrisk medlemsblad"] = "Nord Psyk Medl",
  ["nordisk psykiatrisk tidsskrift"] = "Nord. Psykiatr. Tidsskr.",
  ["nordisk psykiatrisk tidsskrift. nordic journal of psychiatry"] = "Nord Psykiatr Tidsskr",
  ["nordisk psykologi"] = "Nord Psykol",
  ["nordisk sexologi"] = "Nord Sexol",
  ["nordisk veterinaermedicin"] = "Nord Vet Med",
  ["nordisk øst-forum"] = "Nordisk Ost Forum",
  ["nordiskt folkmedicinmuseum meddelanden"] = "Nord Folkmedmus Medd",
  ["nordrhein-westfälische akademie der wissenschaften"] = "Vorträge Nordrhein-Westfälische Akad. Wiss.",
  ["norma, international journal for masculinity studies"] = "Norma Int J Masc Stud",
  ["normale und pathologische anatomie"] = "Norm Pathol Anat (Stuttg)",
  ["norman surgery series"] = "Norman Surg Ser",
  ["normat. nordisk matematisk tidskrift"] = "Normat",
  ["norsk antropologisk tidsskrift"] = "Norsk Antropol. Tidsskr.",
  ["norsk epidemiologi = norwegian journal of epidemiology"] = "Nor Epidemiol",
  ["norsk geografisk tidsskrift - norwegian journal of geography"] = "Norsk. Geogr. Tidsskr. - Norw. J. Geogr.",
  ["norsk geografisk tidsskrift. norwegian journal of geography"] = "Nor Geogr Tidsskr",
  ["norsk geologisk tidsskrift"] = "Nor. Geol. Tidsskr.",
  ["norsk theologisk tidsskrift"] = "Nor Teol Tidsskr",
  ["norske tannlaegeforenings tidende"] = "Nor. Tannlaegeforen. Tid.",
  ["north african and middle east epilepsy journal"] = "North Afr Middle East Epilepsy J",
  ["north american actuarial journal"] = "N. Am. Actuarial J.",
  ["north american actuarial journal : naaj"] = "N Am Actuar J",
  ["north american archaeologist"] = "N. Am. Archaeol.",
  ["north american dialogue : newsletter of the society for the anthropology of north america"] = "North Am Dialogue",
  ["north american fungi"] = "N Am Fungi",
  ["north american journal of aquaculture"] = "N Am J Aquac",
  ["north american journal of economics and finance"] = "N. Amer. J. Econ. Finance",
  ["north american journal of fisheries management"] = "N Am J Fish Manag",
  ["north american journal of medical sciences"] = "N Am J Med Sci",
  ["north american journal of medicine & science"] = "N Am J Med Sci (Boston)",
  ["north american journal of psychology"] = "N Am J Psychol",
  ["north american journal of sports physical therapy : najspt"] = "N Am J Sports Phys Ther",
  ["north carolina central law journal"] = "North Carol Centr Law J",
  ["north carolina dental gazette"] = "N. C. Dent. Gaz.",
  ["north carolina dental gazette : a publication of north carolina dental society"] = "N C Dent Gaz",
  ["north carolina dental journal"] = "N C Dent J",
  ["north carolina folklore"] = "N C Folklor",
  ["north carolina journal of law & technology"] = "N C J Law Technol",
  ["north carolina journal of mental health"] = "N C J Ment Health",
  ["north carolina law review"] = "North Carol Law Rev",
  ["north carolina medical journal"] = "N C Med J",
  ["north carolina medical journal (relston-salem nc"] = "N C Med J",
  ["north carolina state data center newsletter"] = "NC State Data Cent Newsl",
  ["north dakota history"] = "N D Hist",
  ["north dakota law review"] = "N D Law Rev",
  ["north eastern reporter. second series"] = "North East Rep Second Ser",
  ["north western reporter. second series"] = "North West Rep Second Ser",
  ["north-eastern naturalist"] = "North East Nat",
  ["north-holland mathematical library"] = "North-Holland Math. Library",
  ["north-holland mathematics studies"] = "North-Holland Math. Stud.",
  ["north-holland series in applied mathematics and mechanics"] = "North-Holland Ser. Appl. Math. Mech.",
  ["north-western european journal of mathematics"] = "North-West. Eur. J. Math.",
  ["north-western journal of zoology"] = "North West J Zool",
  ["northeast african studies"] = "Northeast Afr Stud",
  ["northeast folklore"] = "Northeast Folk",
  ["northeast historical archaeology"] = "Northeast Hist Archaeology",
  ["northeast journal of contemporary research"] = "Northeast J Contemp Res",
  ["northeastern mathematical journal"] = "Northeast. Math. J.",
  ["northeastern naturalist"] = "Northeast Nat (Steuben)",
  ["northern clinics of istanbul"] = "North Clin Istanb",
  ["northern history"] = "North Hist",
  ["northern illinois university law review"] = "North Ill Univ Law Rev",
  ["northern journal of applied forestry"] = "North. J. Appl. For.",
  ["northern kentucky law review"] = "North KY Law Rev",
  ["northern kentucky state law forum"] = "North Ky State Law Forum",
  ["northern new york medical journal"] = "North N Y Med J",
  ["northern scotland"] = "North Scotl",
  ["northern territory naturalist"] = "North Territ Nat",
  ["northern virginia heritage"] = "North Va Herit",
  ["northwest dentistry"] = "Northwest Dent",
  ["northwest medicine"] = "Northwest Med",
  ["northwest ohio quarterly"] = "Northwest Ohio Q",
  ["northwest public health"] = "Northwest Public Health",
  ["northwest report : a newsletter of the northwest area foundation"] = "Northwest Rep",
  ["northwest science"] = "Northwest Sci.",
  ["northwest science : official publication of the northwest scientific association"] = "Northwest Sci",
  ["northwestern dental research"] = "Northwest Dent Res",
  ["northwestern journal of law and social policy"] = "Northwest J Law Soc Policy",
  ["northwestern naturalist"] = "Northwest. Nat.",
  ["northwestern naturalist : a journal of vertebrate biology"] = "Northwest Nat (Olymp. Wash.)",
  ["northwestern university bulletin, dental research and graduate study"] = "Northwest. Univ. Bull. Dent. Res. Grad. Study",
  ["northwestern university law review"] = "Northwest Univ Law Rev",
  ["norwegian archaeological review"] = "Norw. Archaeol. Rev.",
  ["norwegian journal of entomology"] = "Norw. J. Entomol.",
  ["norwegian journal of entomology (oslo, norway : 1999)"] = "Nor J Entomol (1999)",
  ["norwegian journal of geology"] = "Norw. J. Geol.",
  ["nos oiseaux"] = "Nos Oiseaux",
  ["noshuyo byori (brain tumor pathology)"] = "Noshuyo Byori",
  ["nosokomeiaka chronika"] = "Nosokom Chron",
  ["nostro paese, il"] = "Nostro paese",
  ["not man apart"] = "Not Man Apart",
  ["nota lepidopterologica"] = "Nota Lepidopterol",
  ["notas de algebra y analisis"] = "Notas Algebra Anal.",
  ["notas de la sociedad de matemática de chile"] = "Notas Soc. Mat. Chile",
  ["notas de lógica matemática"] = "Notas Lógica Mat.",
  ["notas de matemática discreta"] = "Notas Mat. Discreta",
  ["notas de poblacion"] = "Notas Poblacion",
  ["notas de población"] = "Notas Poblacion",
  ["notas de salud pública"] = "Notas Salud Publica",
  ["notas do instituto de matematica e estatistica da universidade de sao paulo, serie matematica"] = "Notas Inst. Mat. Estatist. Univ. Sao Paulo Ser. Mat.",
  ["notas e comunicacoes de matematica"] = "Notas Comun. Mat.",
  ["notas em matematica aplicada"] = "Notas Mat. Apl.",
  ["notas quirúrgicas del sanatorio deschamps"] = "Notas Quir Sanat Deschamps",
  ["notatki plockie"] = "Notatki Plockie",
  ["notatki płockie"] = "Notatki Plockie",
  ["note di matematica"] = "Note Mat.",
  ["note e riviste di psichiatria"] = "Note Riv Psichiatr",
  ["note technique groupement technique forestier"] = "Note tech. Group. tech. for.",
  ["notes and queries"] = "Notes Queries",
  ["notes and records"] = "Notes Rec.",
  ["notes and records of the royal society of london"] = "Notes Rec R Soc Lond",
  ["notes and records. the royal society journal of the history of science"] = "Notes and Rec.",
  ["notes from the underground (new york, n.y.)"] = "Notes Undergr",
  ["notes on intuitionistic fuzzy sets"] = "Notes IFS",
  ["notes on number theory and discrete mathematics"] = "Notes Number Theory Discret Math",
  ["notes on numerical fluid mechanics"] = "Notes Numer. Fluid Mech.",
  ["notes on numerical fluid mechanics and multidisciplinary design"] = "Notes Numer. Fluid Mech. Multidiscip. Des.",
  ["notes techniques du centre d’écologie forestière et rurale"] = "Notes tech. Cent. écol. for. rural",
  ["notes. czechoslovak medical association in great britain"] = "Notes Czechoslov Med Assoc G B",
  ["notfall & rettungsmedizin"] = "Notf Rett Med",
  ["notice pour le praticien"] = "Not. prat.",
  ["notices from the isms"] = "Not. ISMS",
  ["notices of the american mathematical society"] = "Notices Amer. Math. Soc.",
  ["notices of the american mathematical society. american mathematical society"] = "Not Am Math Soc",
  ["notices of the south african mathematical society. mededelings van die suid-afrikaanse wiskundevereniging"] = "Not. S. Afr. Math. Soc.",
  ["noticiario arqueológico hispánico. arqueología"] = "NotAHisp",
  ["noticiario arqueológico hispánico. prehistoria"] = "NotAHispPrehistoria",
  ["noticias clínicas"] = "Not Clin",
  ["noticiero médico español"] = "Not Med Esp",
  ["notiziario dell'amministrazione sanitaria. italy. alto commissariato per l'igiene e la sanità pubblica"] = "Not Ammin Sanit (1948)",
  ["notiziario farmaceutico"] = "Not Farm",
  ["notiziario. museo civico, associazione archeologica, allumiere"] = "NotAllumiere",
  ["notiziario. soprintendenza archeologica della lombardia"] = "NotALomb",
  ["notizie archeologiche bergomensi. periodico di archeologia del civico museo archeologico di bergamo"] = "NotABerg",
  ["notizie dal chiostro del monastero maggiore. rassegna di studi del civico museo archeologico e del civico gabinetto numismatico di milano"] = "NotMilano",
  ["notizie degli scavi di antichità"] = "NSA",
  ["notizie di politeia"] = "Not Polit",
  ["notizie per la pratica"] = "Not. prat.",
  ["notre dame journal of formal logic"] = "Notre Dame J. Formal Logic",
  ["notre dame journal of law, ethics & public policy"] = "Notre Dame J Law Ethics Public Policy",
  ["notre dame mathematical lectures"] = "Notre Dame Math. Lect.",
  ["notulae algarum"] = "Not Algarum",
  ["notulae botanicae horti agrobotanici cluj-napoca"] = "Not Bot Horti Agrobot Cluj Napoca",
  ["notulae naturae of the academy of natural sciences of philadelphia"] = "Not Nat Acad Nat Sci Philadelphia",
  ["notulae odonatologicae"] = "Notulae odonatologicae",
  ["notulae scientia biologicae"] = "Not Sci Biol",
  ["nous (detroit, mich.)"] = "Nous",
  ["nouson kougaku kenkyuujo gihou = technical report of the national institute for rural engineering"] = "Nouson Kougaku Kenkyuujo Gihou",
  ["nouveau journal de chimie"] = "Nouv J Chim",
  ["nouveau praticien veterinaire (le) canine & feline"] = "Nouv. Prat. Vet. canine feline",
  ["nouveau praticien veterinaire (le) elevages & sante"] = "Nouv. Prat. Vet. elevages sante",
  ["nouveau praticien veterinaire (le) equine"] = "Nouv. Prat. Vet. equine",
  ["nouvelle bibliotheque mathematique"] = "Nouv. Bibl. Math.",
  ["nouvelle bibliothèque mathématique"] = "Nouv. Bibl. Math.",
  ["nouvelle ecole"] = "Nouv Ec",
  ["nouvelle école"] = "Nouv Ec",
  ["nouvelle presse medicale"] = "Nouv. Presse Med.",
  ["nouvelle revue du xvie siecle"] = "Nouv Rev 16e Siecle",
  ["nouvelle revue du xvie siècle"] = "Nouv Rev 16e Siecle",
  ["nouvelle revue d’optique"] = "Nouv. Rev. Opt.",
  ["nouvelle revue d’optique appliqu?e"] = "Nouv. Rev. Opt. Appl.",
  ["nouvelle revue d’optique appliqute"] = "Nouv. Rev. Opt. Appl.",
  ["nouvelle revue francaise d’hematologie"] = "Nouv. Rev. Fr. Hematol.",
  ["nouvelle revue francaise d’hematologie (blood cells)"] = "Nouv. Rev. Fr. Hematol. Blood Cells",
  ["nouvelle revue francaise d’hematologie; blood cells"] = "Nouv Rev Fr Hematol Blood Cells",
  ["nouvelle revue française d'hématologie"] = "Nouv Rev Fr Hematol",
  ["nouvelle revue française d'hématologie; blood cells"] = "Nouv Rev Fr Hematol Blood Cells",
  ["nouvelle revue théologique"] = "NRTh",
  ["nouvelles - comite international catholique des infirmieres et assistantes medico-sociales. news - international committee of catholic nurses and medico-social workers"] = "Nouv Com Int Cathol Infirm Assist Med Soc",
  ["nouvelles - comité international catholique des infirmières et assistantes médico-sociales. news - international committee of catholic nurses and medico-social workers"] = "Nouv Com Int Cathol Infirm Assist Med Soc",
  ["nouvelles de la republique des lettres (naples, italy)"] = "Nouv Repub Lett",
  ["nouvelles de la république des lettres (naples, italy)"] = "Nouv Repub Lett",
  ["nouvelles der landeshydrologie und -geologie"] = "Nouv. Landeshydrol. -geol.",
  ["nouvelles visions des sciences"] = "Nouv. Vis. Sci.",
  ["noûs (detroit, mich.)"] = "Nous",
  ["nova acta científica compostelana. bioloxía : nacc. bioloxía"] = "Nova Acta Cient Compostel Biol",
  ["nova acta leopoldina"] = "Nova Acta Leopold.",
  ["nova acta leopoldina : abhandlungen der kaiserlich leopoldinisch-carolinisch deutschen akademie der naturforscher"] = "Nova Acta Leopold",
  ["nova acta stomatologica"] = "Nova Acta Stomatol",
  ["nova biotechnologica et chimica"] = "Nova Biotechnol. Chim.",
  ["nova economia (belo horizonte, brazil)"] = "Nova Econ",
  ["nova hedwigia"] = "Nova Hedwigia",
  ["nova hedwigia. beiheft"] = "Nova Hedwig Beih",
  ["nova journal of mathematics, game theory and algebra"] = "Nova J. Math. Game Theory Algebra",
  ["nova journal of theoretical physics"] = "Nova J. Theor. Phys.",
  ["nova law journal"] = "Nova Law J",
  ["nova law review"] = "Nova Law Rev",
  ["nova scotia historical review"] = "N S Hist Rev",
  ["nova scotia medical bulletin"] = "N. S. Med. Bull.",
  ["novaensia. badania ekspedycji archeologicznej uniwersytetu warszawskiego w novae"] = "Novaensia",
  ["novaia i noveishaia istoriia"] = "Nov Novejsaja Istor",
  ["novaia khirurgiia"] = "Nov Khir",
  ["novai︠a︡ i noveĭshai︠a︡ istorii︠a︡"] = "Nov Novejsaja Istor",
  ["novapex : trimestriel de la société belge de malacologie = novapex : quarterly of the belgian malacological society"] = "Novapex",
  ["novartis foundation symposium"] = "Novartis Found Symp",
  ["novedades medicas"] = "Noved Med",
  ["novel approaches in cancer study"] = "Nov Approaches Cancer Study",
  ["novi sad journal of mathematics"] = "Novi Sad J. Math.",
  ["novice iz varstva gozdov"] = "Novice Varst Gozdov",
  ["novitates botanicae universitatis carolinae"] = "Novit Bot Univ Carol",
  ["novitates caribaea"] = "Novit Caribaea",
  ["novon : a journal for botanical nomenclature"] = "Novon (St Louis)",
  ["novon: a journal for botanical nomenclature"] = "Novon: J. Bot. Nomencl.",
  ["novosti meditsinskogo priborostroeniia"] = "Nov Med Priborostr",
  ["novosti meditsinskoi tekhniki"] = "Nov. Med. Tekh.",
  ["novosti meditsinskoĭ tekhniki"] = "Nov Med Tekh",
  ["novosti sistematiki nizshikh rasteniĭ"] = "Nov Sist Nizs Rast",
  ["novum testamentum: an international quarterly for new testament and related studies"] = "NT",
  ["novyĭ khirurgicheskiĭ arkhiv"] = "Novyi Khirurgicheskii Arkhiv",
  ["nowe drogi"] = "Nowe Drogi",
  ["nowiny lekarskie : organ wydziału lekarskiego towarzystwa przyjaciół nauk posnańskiego"] = "Now Lekarskie Organ Wydz Lekarskiego Tow Przyj Nauk Posnanskiego",
  ["nông nghiệp & nông thôn"] = "Nong Nghiep Nong Thon",
  ["nō to shinkei = brain and nerve"] = "No To Shinkei",
  ["nōshuyō byōri = brain tumor pathology"] = "Noshuyo Byori",
  ["nöro psikiyatri arşivi"] = "Noro Psikiyatr Ars",
  ["nőgyógyászati és szülészeti továbbképző szemle"] = "Nogyogyaszati Szuleszeti Tovabbkepzo Szemle",
  ["np news"] = "NP News",
  ["np news (bellevue, wash.)"] = "NP News",
  ["npg asia materials"] = "NPG Asia Mater.",
  ["npg forum series"] = "NPG Forum Ser",
  ["npj 2d materials and applications"] = "NPJ 2D Mater Appl",
  ["npj aging and mechanisms of disease"] = "NPJ Aging Mech Dis",
  ["npj biodiversity"] = "npj Biodivers.",
  ["npj biofilms and microbiomes"] = "NPJ Biofilms Microbiomes",
  ["npj breast cancer"] = "NPJ Breast Cancer",
  ["npj clean water"] = "NPJ Clean Water",
  ["npj climate and atmospheric science"] = "NPJ Clim Atmos Sci",
  ["npj computational materials"] = "npj Comput. Mater.",
  ["npj digital medicine"] = "NPJ Digit Med",
  ["npj flexible electronics"] = "npj Flexible Electron.",
  ["npj genomic medicine"] = "NPJ Genom Med",
  ["npj materials degradation"] = "Npj Mater Degrad",
  ["npj microgravity"] = "NPJ Microgravity",
  ["npj molecular phenomics"] = "npj Mol. Phenomics",
  ["npj ocean sustainability"] = "npj Ocean Sustainability",
  ["npj parkinson's disease"] = "NPJ Parkinsons Dis",
  ["npj parkinson’s disease"] = "npj Parkinson’s Dis.",
  ["npj pollution control"] = "npj Pollut. Control",
  ["npj precision oncology"] = "NPJ Precis Oncol",
  ["npj primary care respiratory medicine"] = "NPJ Prim Care Respir Med",
  ["npj quantum information"] = "npj Quantum Inf.",
  ["npj quantum materials"] = "NPJ Quantum Mater",
  ["npj regenerative medicine"] = "NPJ Regen Med",
  ["npj robotics"] = "npj Rob.",
  ["npj schizophrenia"] = "NPJ Schizophr",
  ["npj science of food"] = "NPJ Sci Food",
  ["npj science of learning"] = "NPJ Sci Learn",
  ["npj systems biology and applications"] = "NPJ Syst Biol Appl",
  ["npj vaccines"] = "NPJ Vaccines",
  ["npn medecine"] = "NPN Med",
  ["npn médecine"] = "NPN Med",
  ["nriag journal of astronomy and geophysics"] = "NRIAG J. Astron. Geophys.",
  ["nrl review (1988)"] = "NRL Rev",
  ["nsf-cbms regional conference series in probability and statistics"] = "NSF-CBMS Regional Conf. Ser. Probab. Statist.",
  ["nt learning curve"] = "NT Learn Curve",
  ["ntm journal of the history of science, technology and medicine"] = "NTM J. Hist. Sci. Technol. Med.",
  ["ntm zeitschrift fur geschichte der wissenschaften, technik und medizin"] = "NTM Z. Gesch. Wissen. Tech. Med.",
  ["ntp cerhr mon"] = "NTP CERHR MON",
  ["ntp monograph"] = "NTP Monogr",
  ["nubia christiana"] = "NubChr",
  ["nubian letters"] = "NubLet",
  ["nuclear analysis"] = "Nucl Anal.",
  ["nuclear and chemical waste management"] = "Nucl. Chem. Waste Manage.",
  ["nuclear and particle physics proceedings"] = "Nucl. Part. Phys. Proc.",
  ["nuclear data section a"] = "Nucl. Data Sect. A",
  ["nuclear data section b"] = "Nucl. Data Sect. B",
  ["nuclear data sheets"] = "Nucl. Data Sheets",
  ["nuclear data, section a"] = "Nucl. Data, Sec. A",
  ["nuclear data, section b"] = "Nucl. Data, Sec. B",
  ["nuclear energy and technology"] = "Nucl. Energy Technol.",
  ["nuclear engineering and design"] = "Nucl. Eng. Des.",
  ["nuclear engineering and design/fusion : an international journal devoted to the thermal, mechanical, materials, structural, and design problems of fusion energy"] = "Nucl Eng Des/Fusion",
  ["nuclear engineering and technology"] = "Nucl. Eng. Technol.",
  ["nuclear fusion"] = "Nucl. Fusion",
  ["nuclear geophysics"] = "Nucl. Geophys.",
  ["nuclear instruments"] = "Nucl. Instrum.",
  ["nuclear instruments & methods"] = "Nucl. Instrum. Methods",
  ["nuclear instruments & methods in physics research, section a: accelerators, spectrometers, detectors, and associated equipment"] = "Nucl. Instrum. Methods Phys. Res., Sect. A",
  ["nuclear instruments & methods in physics research, section b: beam interactions with materials and atoms"] = "Nucl. Instrum. Methods Phys. Res., Sect. B",
  ["nuclear instruments & methods in physics research. section a, accelerators, spectrometers, detectors and associated equipment"] = "Nucl Instrum Methods Phys Res A",
  ["nuclear instruments & methods in physics research. section b, beam interactions with materials and atoms"] = "Nucl Instrum Methods Phys Res B",
  ["nuclear instruments and methods"] = "Nucl. Instrum. Methods",
  ["nuclear instruments and methods in physics research"] = "Nucl. Instrum. Methods Phys. Res.",
  ["nuclear instruments and methods in physics research section a: accelerators, spectrometers, detectors and associated equipment"] = "Nucl. Instrum. Methods Phys. Res., Sect. A",
  ["nuclear instruments and methods in physics research section b"] = "Nucl. Instrum. Methods Phys. Res. Sect. B",
  ["nuclear instruments and methods in physics research section b: beam interactions with materials and atoms"] = "Nucl. Instrum. Methods Phys. Res., Sect. B",
  ["nuclear materials and energy"] = "Nucl. Mater. Energy",
  ["nuclear medicine and biology"] = "Nucl Med Biol",
  ["nuclear medicine and biomedical imaging"] = "Nucl Med Biomed Imaging",
  ["nuclear medicine and molecular imaging"] = "Nucl Med Mol Imaging",
  ["nuclear medicine communications"] = "Nucl Med Commun",
  ["nuclear medicine review. central & eastern europe"] = "Nucl Med Rev Cent East Eur",
  ["nuclear medicine review. central & eastern europe : journal of bulgarian, czech, macedonian, polish, romanian, russian, slovak, yugoslav societies of nuclear medicine and ukrainian society of radiology"] = "Nucl Med Rev Cent East Eur",
  ["nuclear physics"] = "Nucl. Phys.",
  ["nuclear physics a"] = "Nucl. Phys. A",
  ["nuclear physics and atomic energy"] = "Nucl. Phys. At. Energy",
  ["nuclear physics b"] = "Nucl. Phys. B",
  ["nuclear physics b - proceedings supplements"] = "Nucl. Phys. B Proc. Suppl.",
  ["nuclear physics b, proceedings supplements"] = "Nucl. Phys. B, Proc. Suppl.",
  ["nuclear physics news"] = "Nucl. Phys. News",
  ["nuclear physics. a"] = "Nucl Phys A",
  ["nuclear physics. b"] = "Nucl Phys B",
  ["nuclear physics. b, proceedings, supplements"] = "Nucl Phys B Proc Suppl",
  ["nuclear physics. b. theoretical, phenomenological, and experimental high energy physics. quantum field theory and statistical systems"] = "Nuclear Phys. B",
  ["nuclear receptor"] = "Nucl Recept",
  ["nuclear receptor research"] = "Nucl Receptor Res",
  ["nuclear receptor signaling"] = "Nucl Recept Signal",
  ["nuclear science abstracts"] = "Nucl Sci Abstr",
  ["nuclear science and engineering"] = "Nucl. Sci. Eng.",
  ["nuclear science and engineering : the journal of the american nuclear society"] = "Nucl Sci Eng",
  ["nuclear science and techniques"] = "Nucl. Sci. Tech.",
  ["nuclear structural engineering"] = "Nucl. Struct. Eng.",
  ["nuclear technology"] = "Nucl Technol",
  ["nuclear track detection"] = "Nucl. Track Detect.",
  ["nuclear tracks"] = "Nucl. Tracks",
  ["nuclear tracks and radiation measurement"] = "Nucl. Tracks Radiat. Meas.",
  ["nuclear tracks and radiation measurements"] = "Nucl Tracks Radiat Meas",
  ["nuclear tracks and radiation measurements (oxford, england : 1993)"] = "Nucl Tracks Radiat Meas",
  ["nucleic acid therapeutics"] = "Nucleic Acid Ther",
  ["nucleic acids and molecular biology"] = "Nucleic Acids Mol Biol",
  ["nucleic acids research"] = "Nucleic Acids Res",
  ["nucleic acids research supplement"] = "Nucleic Acids Res. Suppl.",
  ["nucleic acids research. special publication"] = "Nucleic Acids Res Spec Publ",
  ["nucleic acids research. supplement (2001)"] = "Nucleic Acids Res Suppl",
  ["nucleic acids symposium series"] = "Nucleic Acids Symp Ser",
  ["nucleic acids symposium series (2004)"] = "Nucleic Acids Symp Ser (Oxf)",
  ["nucleosides & nucleotides"] = "Nucleosides Nucleotides",
  ["nucleosides and nucleotides"] = "Nucleosides Nucleotides",
  ["nucleosides, nucleotides & nucleic acids"] = "Nucleosides Nucleotides Nucleic Acids",
  ["nucleosides, nucleotides and nucleic acids"] = "Nucleosides Nucleotides Nucleic Acids",
  ["nucleosides, nucleotides, and nucleic acids"] = "Nucleosides Nucleotides Nucleic Acids",
  ["nucleus (austin, tex.)"] = "Nucleus",
  ["nucleus (cambridge, mass.)"] = "Nucleus (Camb)",
  ["nuestra historia"] = "Nuestra Hist",
  ["nueva enfermeria"] = "Nueva Enferm",
  ["nuevas ideas; revista de medicina"] = "Nuevas Ideas Rev Med",
  ["nuevos libros"] = "Nuevos Libros",
  ["nuevos libros / acuerdo de cartagena, junta, biblioteca"] = "Nuevos Libros",
  ["nufusbilim dergisi"] = "Nufusbil Derg",
  ["nuklearmedizin. nuclear medicine"] = "Nuklearmedizin",
  ["numantia. investigaciones arqueológicas en castilla y león"] = "Numantia",
  ["numen. international review for the history of religions"] = "Numen",
  ["numerical algebra, control and optimization"] = "Numer. Algebra, Control Optim.",
  ["numerical algorithms"] = "Numer. Algorithms",
  ["numerical analysis and applications"] = "Numer. Anal. Appl.",
  ["numerical functional analysis and optimization"] = "Numer. Funct. Anal. Optim.",
  ["numerical functional analysis and optimization. an international journal"] = "Numer. Funct. Anal. Optim.",
  ["numerical heat transfer part a"] = "Numer. Heat Transfer, Part A",
  ["numerical heat transfer part a: applications"] = "Numer. Heat Transfer, Part A",
  ["numerical heat transfer part b"] = "Numer. Heat Transfer, Part B",
  ["numerical heat transfer part b: fundamentals"] = "Numer. Heat Transfer, Part B",
  ["numerical heat transfer. part a, applications"] = "Numeri Heat Transf A Appl",
  ["numerical insights"] = "Numer. Insights",
  ["numerical linear algebra with applications"] = "Numer. Linear Algebra Appl.",
  ["numerical mathematics"] = "Numer. Math. J. Chinese Univ.",
  ["numerical mathematics (hong kong, china)"] = "Numer Math",
  ["numerical mathematics and scientific computation"] = "Numer. Math. Sci. Comput.",
  ["numerical mathematics theory methods and applications"] = "Numer. Math. Theory Methods Appl.",
  ["numerical mathematics. a journal of chinese universities. gaodeng xuexiao jisuan shuxue xuebao"] = "Numer. Math. J. Chinese Univ.",
  ["numerical mathematics. theory, methods and applications"] = "Numer. Math. Theory Methods Appl.",
  ["numerical methods for partial differential equations"] = "Numer Methods Partial Differ Equ",
  ["numerical methods for partial differential equations. an international journal"] = "Numer. Methods Partial Differential Equations",
  ["numerical methods in engineering series"] = "Numer. Methods Eng. Ser.",
  ["numerische mathematik"] = "Numer. Math.",
  ["numisma. revista de la sociedad iberoamericana de estudios numismáticos"] = "Numisma",
  ["numismatic chronicle"] = "NC",
  ["numismatic circular"] = "NCirc",
  ["numismatic literature"] = "NL",
  ["numismatica e antichità classiche. quaderni ticinesi"] = "NumAntCl",
  ["numismatica e antichità classiche: quaderni ticinesi"] = "NAC",
  ["numismatica. periodico di cultura e di informazione numismatica"] = "NumismaticaRom",
  ["numismatika i èpigraphika"] = "NE",
  ["numismatische zeitschrift"] = "NZ",
  ["numismatisches nachrichtenblatt: organ des verbandes der dt. münzvereine"] = "NNB",
  ["numizmatika i epigrafika"] = "NumEpigr",
  ["numizmatičar. casopis za anticki i stari jugoslavenski novac"] = "Numizmaticar",
  ["nummus. sociedade portuguesa de numismatica"] = "Nummus",
  ["nuncius / istituto e museo di storia della scienza"] = "Nuncius",
  ["nuntius radiologicus"] = "Nunt Radiol",
  ["nuova antologia"] = "Nuova Antologia",
  ["nuova rivista di neurologia"] = "Nuova Riv Neurol",
  ["nuova rivista storica"] = "Nuova Riv Stor",
  ["nuovi annali d'igiene e microbiologia"] = "Nuovi Ann Ig Microbiol",
  ["nuovi annali della facoltà di magistero dell’università di messina"] = "NAFM",
  ["nuovi annali d’igiene e microbiologia"] = "Nuovi Ann. Ig. Microbiol.",
  ["nuovi studi fanesi"] = "NStFan",
  ["nuovo bullettino archeologico sardo"] = "BASard",
  ["nuovo cimento"] = "Nuovo Cimento",
  ["nuovo cimento a"] = "Nuovo Cimento A",
  ["nuovo cimento b"] = "Nuovo Cimento B",
  ["nuovo cimento c"] = "Nuovo Cimento C",
  ["nuovo cimento d"] = "Nuovo Cimento D",
  ["nuovo cimento della societa italiana di fisica a"] = "Nuovo Cimento Soc. Ital. Fis., A",
  ["nuovo cimento della societa italiana di fisica b"] = "Nuovo Cimento Soc. Ital. Fis., B",
  ["nuovo cimento della societa italiana di fisica c"] = "Nuovo Cimento Soc. Ital. Fis., C",
  ["nuovo cimento della societa italiana di fisica d"] = "Nuovo Cimento Soc. Ital. Fis., D",
  ["nuovo didaskaleion"] = "NuovDidask",
  ["nuovo giornale botanico italiano"] = "Nuovo g. bot. ital.",
  ["nur al-mantiq wal-ma\\rasp rifah"] = "Nur al-Mantiq wal-Ma\\rasp rifah",
  ["nurse anesthesia"] = "Nurse Anesth",
  ["nurse author & editor"] = "Nurse Author Ed",
  ["nurse author and editor"] = "Nurse Author Ed.",
  ["nurse education in practice"] = "Nurse Educ Pract",
  ["nurse education today"] = "Nurse Educ Today",
  ["nurse educator"] = "Nurse Educ",
  ["nurse educators opportunities and innovations"] = "Nurse Educ Oppor Innov",
  ["nurse leader"] = "Nurse Lead",
  ["nurse managers bookshelf"] = "Nurse Managers Bookshelf",
  ["nurse managers' bookshelf"] = "Nurse Managers Bookshelf",
  ["nurse managers’ bookshelf"] = "Nurse Managers Bookshelf",
  ["nurse practitioner"] = "Nurse Pract.",
  ["nurse practitioner forum"] = "Nurse Pract Forum",
  ["nurse researcher"] = "Nurse Res",
  ["nurses lamp"] = "Nurses Lamp",
  ["nursing & health care : official publication of the national league for nursing"] = "Nurs Health Care",
  ["nursing & health sciences"] = "Nurs Health Sci",
  ["nursing (auckland, n.z.)"] = "Nursing (Auckl)",
  ["nursing administration quarterly"] = "Nurs Adm Q",
  ["nursing and health (alhambra, calif.)"] = "Nurs Health (Alhambra)",
  ["nursing and health care"] = "Nurs. Health Care",
  ["nursing and health care (winfield, ill.)"] = "Nurs Health Care (Winfield)",
  ["nursing and health care perspectives"] = "Nurs Health Care Perspect",
  ["nursing and midwifery studies"] = "Nurs Midwifery Stud",
  ["nursing and palliative care"] = "Nurs Palliat Care",
  ["nursing bc"] = "Nurs BC",
  ["nursing bc / registered nurses association of british columbia"] = "Nurs BC",
  ["nursing care"] = "Nurs Care",
  ["nursing careers"] = "Nurs Careers",
  ["nursing case management"] = "Nurs. Case Manag.",
  ["nursing case management : managing the process of patient care"] = "Nurs Case Manag",
  ["nursing children and young people"] = "Nurs Child Young People",
  ["nursing clinics of north america"] = "Nurs. Clin. North Am.",
  ["nursing diagnosis"] = "Nurs. Diagn.",
  ["nursing diagnosis : nd : the official journal of the north american nursing diagnosis association"] = "Nurs Diagn",
  ["nursing digest"] = "Nurs Dig",
  ["nursing dynamics"] = "Nurs Dyn",
  ["nursing economic$"] = "Nurs Econ",
  ["nursing economics"] = "Nurs. Econ.",
  ["nursing education monographs"] = "Nurs Educ Monogr",
  ["nursing education perspectives"] = "Nurs Educ Perspect",
  ["nursing educators microworld"] = "Nurs Educ Microworld",
  ["nursing ethics"] = "Nurs Ethics",
  ["nursing focus"] = "Nurs Focus",
  ["nursing for women's health"] = "Nurs Womens Health",
  ["nursing for women’s health"] = "Nurs Womens Health",
  ["nursing forum"] = "Nurs Forum",
  ["nursing history review"] = "Nurs. Hist. Rev.",
  ["nursing history review : official journal of the american association for the history of nursing"] = "Nurs Hist Rev",
  ["nursing homes"] = "Nurs Homes",
  ["nursing homes and senior citizen care"] = "Nurs Homes Sr Citiz Care",
  ["nursing in critical care"] = "Nurs Crit Care",
  ["nursing inquiry"] = "Nurs Inq",
  ["nursing journal"] = "Nurs J (Manila)",
  ["nursing journal of india"] = "Nurs. J. India",
  ["nursing journal of singapore"] = "Nurs. J. Singapore",
  ["nursing law & ethics"] = "Nurs Law Ethics",
  ["nursing law & ethics,"] = "Nurs Law Ethics",
  ["nursing law and ethics"] = "Nurs. Law Ethics",
  ["nursing law's regan report"] = "Nurs Law Regan Rep",
  ["nursing law’s regan report"] = "Nurs Law Regan Rep",
  ["nursing leadership"] = "Nurs Leadersh",
  ["nursing leadership (toronto, ont.)"] = "Nurs Leadersh (Tor Ont)",
  ["nursing leadership forum"] = "Nurs Leadersh Forum",
  ["nursing life"] = "Nurs. Life",
  ["nursing management"] = "Nurs Manage",
  ["nursing management (harrow, london, england : 1994)"] = "Nurs Manag (Harrow)",
  ["nursing mirror"] = "Nurs Mirror",
  ["nursing mirror and midwives journal"] = "Nurs Mirror Midwives J",
  ["nursing montreal"] = "Nurs. Montreal",
  ["nursing montréal"] = "Nurs Montreal",
  ["nursing new zealand"] = "Nurs N Z",
  ["nursing new zealand (wellington, n.z. : 1995)"] = "Nurs N Z",
  ["nursing news"] = "Nurs News",
  ["nursing older people"] = "Nurs Older People",
  ["nursing open"] = "Nurs Open",
  ["nursing outlook"] = "Nurs Outlook",
  ["nursing papers"] = "Nurs. Pap.",
  ["nursing papers. perspectives en nursing"] = "Nurs Pap",
  ["nursing philosophy : an international journal for healthcare professionals"] = "Nurs Philos",
  ["nursing practice"] = "Nurs. Pract.",
  ["nursing practice (edinburgh, scotland)"] = "Nurs Pract",
  ["nursing praxis in new zealand"] = "Nurs. Prax. N. Z.",
  ["nursing praxis in new zealand inc"] = "Nurs Prax N Z",
  ["nursing quality connection"] = "Nurs Qual Connect",
  ["nursing quebec"] = "Nurs. Que.",
  ["nursing québec"] = "Nurs Que",
  ["nursing reports"] = "Nurs Rep",
  ["nursing research"] = "Nurs Res",
  ["nursing research and practice"] = "Nurs Res Pract",
  ["nursing research conference"] = "Nurs Res Conf",
  ["nursing research report"] = "Nurs Res Rep",
  ["nursing review (north sydney, n.s.w.)"] = "Nurs Rev",
  ["nursing rsa"] = "Nurs. RSA",
  ["nursing rsa = verpleging rsa"] = "Nurs RSA",
  ["nursing science"] = "Nurs Sci",
  ["nursing science quarterly"] = "Nurs Sci Q",
  ["nursing spectrum (d.c./baltimore metro ed.)"] = "Nurs Spectr (Wash D C)",
  ["nursing spectrum (d.c./baltimore metro edition)"] = "Nurs. Spectr. (Wash. D. C.)",
  ["nursing spectrum (florida ed.)"] = "Nurs Spectr (Fla Ed)",
  ["nursing spectrum (florida edition)"] = "Nurs. Spectr. (Fla. Ed.)",
  ["nursing spectrum (greater chicago/ne illinois and nw indiana edition)"] = "Nurs. Spectr. (Gt. Chic. NE Ill. NW Indiana Ed.)",
  ["nursing spectrum (greater philadelphia/tri-state ed)"] = "Nurs. Spectr. (Gt. Phila. Tri State Ed.)",
  ["nursing spectrum (greater philadelphia/tri-state ed.)"] = "Nurs Spectr (Gt Phila Tri State Ed)",
  ["nursing spectrum (illinois ed.)"] = "Nurs Spectr (Gt Chic Ne Ill Nw Indiana Ed)",
  ["nursing spectrum (new england ed.)"] = "Nurs Spectr (N Engl Ed)",
  ["nursing spectrum (new england edition)"] = "Nurs. Spectr. (N. Engl. Ed.)",
  ["nursing staff development insider"] = "Nurs Staff Dev Insid",
  ["nursing standard"] = "Nurs. Stand.",
  ["nursing standard (royal college of nursing (great britain) : 1987)"] = "Nurs Stand",
  ["nursing standard (royal college of nursing (great britain) : 1987). special supplement"] = "Nurs Stand Spec Suppl",
  ["nursing standard (royal college of nursing (great britain) : 1987). special supplement."] = "Nurs Stand Spec Suppl",
  ["nursing standard : official newspaper of the royal college of nursing"] = "Nurs Stand",
  ["nursing standard. special supplement"] = "Nurs. Stand. Spec. Suppl.",
  ["nursing success today"] = "Nurs Success Today",
  ["nursing the elderly"] = "Nurs. Elder.",
  ["nursing the elderly : in hospital, homes and the community"] = "Nurs Elder",
  ["nursing times"] = "Nurs Times",
  ["nursing times. nursing homes"] = "Nurs Times Nurs Homes",
  ["nursing world"] = "Nurs World",
  ["nutrafoods : international journal of science and marketing for nutraceutical actives, raw materials, finish products"] = "Nutrafoods",
  ["nutricion hospitalaria"] = "Nutr. Hosp.",
  ["nutricion hospitalaria : organo oficial de la sociedad espanola de nutricion parenteral y enteral"] = "Nutr Hosp",
  ["nutrición hospitalaria"] = "Nutr Hosp",
  ["nutrient cycling in agroecosystems"] = "Nutr. Cycling Agroecosyst.",
  ["nutritio et dieta (european review of nutrition and dietetics)"] = "Nutr. Dieta Eur. Rev. Nutr. Diet.",
  ["nutritio et dieta; european review of nutrition and dietetics"] = "Nutr Dieta Eur Rev Nutr Diet",
  ["nutrition & diabetes"] = "Nutr Diabetes",
  ["nutrition & dietetics"] = "Nutr. Diet.",
  ["nutrition & dietetics: the journal of the dietitians association of australia"] = "Nutr Diet",
  ["nutrition & food science"] = "Nutr. Food Sci.",
  ["nutrition & metabolism"] = "Nutr Metab (Lond)",
  ["nutrition (burbank, los angeles county, calif.)"] = "Nutrition",
  ["nutrition abstracts and reviews"] = "Nutr Abstr Rev",
  ["nutrition abstracts and reviews. series a: human and experimental"] = "Nutr Abstr Rev Ser Hum Exp",
  ["nutrition and aging"] = "Nutr. Aging",
  ["nutrition and aging (amsterdam, netherlands)"] = "Nutr Aging (Amst)",
  ["nutrition and cancer"] = "Nutr Cancer",
  ["nutrition and dietary supplements"] = "Nutr Diet Suppl",
  ["nutrition and food science"] = "Nutr Food Sci",
  ["nutrition and health"] = "Nutr Health",
  ["nutrition and health (berkhamsted, hertfordshire)"] = "Nutr Health",
  ["nutrition and healthy aging"] = "Nutr Healthy Aging",
  ["nutrition and metabolic insights"] = "Nutr Metab Insights",
  ["nutrition and metabolism"] = "Nutr Metab",
  ["nutrition bulletin"] = "Nutr Bull",
  ["nutrition for healthy living"] = "Nutr Healthy Living",
  ["nutrition for life"] = "Nutr Life",
  ["nutrition history notes"] = "Nutr Hist Notes",
  ["nutrition history notes / vanderbilt medical center library"] = "Nutr Hist Notes",
  ["nutrition in clinical care"] = "Nutr. Clin. Care",
  ["nutrition in clinical care : an official publication of tufts university"] = "Nutr Clin Care",
  ["nutrition in clinical practice"] = "Nutr. Clin. Pract.",
  ["nutrition in clinical practice : official publication of the american society for parenteral and enteral nutrition"] = "Nutr Clin Pract",
  ["nutrition journal"] = "Nutr J",
  ["nutrition news"] = "Nutr News (Rosemt)",
  ["nutrition reports international"] = "Nutr Rep Int",
  ["nutrition research"] = "Nutr. Res.",
  ["nutrition research (new york, n.y.)"] = "Nutr Res",
  ["nutrition research (new york, ny, united states)"] = "Nutr. Res. (N. Y., NY, U. S.)",
  ["nutrition research and practice"] = "Nutr Res Pract",
  ["nutrition research reviews"] = "Nutr Res Rev",
  ["nutrition review. mother and child"] = "Nutr Rev Mother Child",
  ["nutrition reviews"] = "Nutr Rev",
  ["nutrition today"] = "Nutr Today",
  ["nutrition update"] = "Nutr Update",
  ["nutrition, metabolism, and cardiovascular diseases : nmcd"] = "Nutr Metab Cardiovasc Dis",
  ["nutritional neuroscience"] = "Nutr Neurosci",
  ["nutzholzverbrauch in der schweiz"] = "Nutzholzverbrauch Schweiz",
  ["nuytsia : bulletin of the western australian herbarium"] = "Nuytsia",
  ["nüfusbilim dergisi"] = "Nufusbil Derg",
  ["nwb-studienbucher wirtschaftswissenschaften"] = "NWB-Studienbuch. Wirtschaftswissenschaft.",
  ["nwsa journal : a publication of the national women's studies association"] = "NWSA J",
  ["nyo [reports]. u.s. atomic energy commission"] = "NYO Rep",
  ["nyo reports"] = "NYO Rep.",
  ["nyssnta journal"] = "NYSSNTA J",
  ["nürnberger blätter zur archäologie. publikationsreihe des bildungszentrums der stadt nürnberg, fachbereich archäologie"] = "NueBlA",
  ["o arqueólogo português"] = "APort",
  ["o biologico"] = "Biol",
  ["o lar do médico"] = "Lar Do Med",
  ["o medico; semanario de assuntos medicos e paramedicos"] = "Medico (Porto)",
  ["o medico; semanario de assuntos médicos e paramédicos"] = "Medico (Porto)",
  ["o público e o privado : caderno dos núcleos e grupos de pesquisa vinculados ao mestrado acadêmico em políticas públicas e sociedade da universidade estadual do ceará"] = "Publico Priv",
  ["oa alcohol"] = "OA Alcohol",
  ["oa autism"] = "OA Autism",
  ["oa family medicine"] = "OA Fam Med",
  ["oa genetics"] = "OA Genet",
  ["oa journal of pregnancy and child care"] = "OA J Pregnancy Child Care",
  ["oa microbiology"] = "OA Microbiol",
  ["oa minimally invasive surgery"] = "OA Minim Invasive Surg",
  ["oa molecular oncology"] = "OA Mol Oncol",
  ["oa women's health"] = "OA Womens Health",
  ["oasics: openaccess series in informatics"] = "OASIcs OpenAccess Ser. Inform.",
  ["ob. gyn. news"] = "Ob Gyn News",
  ["ob/gyn digest"] = "Ob Gyn Dig",
  ["ob/gyn world"] = "Ob/Gyn World",
  ["obere extremität"] = "Obere Extrem",
  ["oberwolfach reports"] = "Oberwolfach Rep.",
  ["oberwolfach seminars"] = "Oberwolfach Semin.",
  ["obesity & control therapies : open access"] = "Obes Control Ther",
  ["obesity (silver spring, md.)"] = "Obesity (Silver Spring)",
  ["obesity and endocrinology"] = "Obes. Endocrinol.",
  ["obesity and metabolism"] = "Obe Metab",
  ["obesity and weight management"] = "Obes Weight  Manag",
  ["obesity facts"] = "Obes Facts",
  ["obesity management"] = "Obes Manag",
  ["obesity medicine"] = "Obes Med",
  ["obesity research"] = "Obes Res",
  ["obesity research & clinical practice"] = "Obes Res Clin Pract",
  ["obesity reviews"] = "Obes. Rev.",
  ["obesity reviews : an official journal of the international association for the study of obesity"] = "Obes Rev",
  ["obesity science & practice"] = "Obes Sci Pract",
  ["obesity surgery"] = "Obes Surg",
  ["obesity surgery : the official journal of the american society for bariatric surgery and of the obesity surgery society of australia and new zealand"] = "Obes Surg",
  ["obesity, open access"] = "Obes Open Access",
  ["objets et mondes : la revue du musee de l’homme, museum national d’histoire naturelle"] = "Objets Mondes",
  ["objets et mondes : la revue du musée de l'homme, muséum national d'histoire naturelle"] = "Objets Mondes",
  ["obm genetics"] = "OBM Genet",
  ["obm geriatrics"] = "OBM Geriat",
  ["obm integrative and complimentary medicine"] = "OBM Integr Compliment Med",
  ["obm neurobiology"] = "OBM Neurobiol",
  ["obm transplantation"] = "OBM Transplant",
  ["obradoiro de historia moderna"] = "Obradoiro Hist Mod",
  ["observational studies"] = "Obs Stud",
  ["observations et diagnostics economiques. revue de l’ofce"] = "Obs Diagn Econ",
  ["observations et diagnostics économiques. revue de l'ofce"] = "Obs Diagn Econ",
  ["observer (london, england)"] = "Observer (Lond)",
  ["obstetric medicine"] = "Obstet Med",
  ["obstetrica si ginecologia"] = "Obstet. Ginecol. (Bucur.)",
  ["obstetrical & gynecological survey"] = "Obstet Gynecol Surv",
  ["obstetrical and gynecological survey"] = "Obstet. Gynecol. Surv.",
  ["obstetrică şi ginecologia"] = "Obstet Ginecol (Bucur)",
  ["obstetricia pratica"] = "Obstet Prat",
  ["obstetricia prática"] = "Obstet Prat",
  ["obstetricia y ginecologia latino-americanas"] = "Obstet. Ginecol. Lat. Am.",
  ["obstetricia y ginecología latino-americanas"] = "Obstet Ginecol Lat Am",
  ["obstetrics & gynecology international journal"] = "Obstet Gynecol Int J",
  ["obstetrics & gynecology science"] = "Obstet Gynecol Sci",
  ["obstetrics and gynaecology cases - reviews"] = "Obstet Gynaecol Cases Rev",
  ["obstetrics and gynecology"] = "Obstet Gynecol",
  ["obstetrics and gynecology annual"] = "Obstet Gynecol Annu",
  ["obstetrics and gynecology clinics of north america"] = "Obstet Gynecol Clin North Am",
  ["obstetrics and gynecology international"] = "Obstet Gynecol Int",
  ["obstetrics, gynaecology and reproductive medicine"] = "Obstet Gynaecol Reprod Med",
  ["occasional newsletter (lindsay club)"] = "Occas Newsl Lindsay Club",
  ["occasional newsletter, lindsay club"] = "Occas. Newsl. Lindsay Club",
  ["occasional paper (royal college of general practitioners)"] = "Occas Pap R Coll Gen Pract",
  ["occasional paper, royal college of general practitioners"] = "Occas. Pap. R. Coll. Gen. Pract.",
  ["occasional papers (society for history in the federal government (u.s.))"] = "Occas Pap Soc Hist Fed Gov",
  ["occasional papers (texas tech university. museum)"] = "Occas Pap Tex Tech Univ Mus",
  ["occasional papers of the institute of mennonite studies"] = "Occ. P. Inst. Men. Stud.",
  ["occasional papers of the museum of zoology, university of michigan"] = "Occas Pap Mus Zool Univ Mich",
  ["occasional papers on medical history australia"] = "Occas Pap Med Hist Aust",
  ["occasional publications in classical studies"] = "OccasPublClSt",
  ["occident and orient"] = "OccOr",
  ["occupational and environmental medicine"] = "Occup Environ Med",
  ["occupational ergonomics"] = "Occup. Ergon.",
  ["occupational ergonomics : the journal of the international society for occupational ergonomics and safety"] = "Occup Ergon",
  ["occupational health"] = "Occup Health (Auckl)",
  ["occupational health & safety (waco, tex.)"] = "Occup Health Saf",
  ["occupational health (washington, d.c.)"] = "Occup Health (Wash D C)",
  ["occupational health and safety"] = "Occup. Health Saf.",
  ["occupational health nurse"] = "Occup. Health Nurse (Auckl.)",
  ["occupational health nursing"] = "Occup Health Nurs",
  ["occupational health review"] = "Occup Health Rev",
  ["occupational health science"] = "Occup Health Sci",
  ["occupational health; a journal for occupational health nurses"] = "Occup Health (Lond)",
  ["occupational medicine"] = "Occup Med (Chic Ill)",
  ["occupational medicine & health affairs"] = "Occup Med Health Aff",
  ["occupational medicine (oxford, england)"] = "Occup Med (Lond)",
  ["occupational medicine (philadelphia, pa.)"] = "Occup Med",
  ["occupational outlook quarterly"] = "OCCUP Outlook Q",
  ["occupational outlook quarterly / united states department of labor, bureau of labor statistics in cooperation with veterans administration"] = "OCCUP Outlook Q",
  ["occupational psychology"] = "Occup Psychol",
  ["occupational therapy and rehabilitation"] = "Occup Ther Rehabil",
  ["occupational therapy in health care"] = "Occup Ther Health Care",
  ["occupational therapy in mental health"] = "Occup Ther Ment Health",
  ["occupational therapy international"] = "Occup Ther Int",
  ["occupations; the vocational guidance journal"] = "Occupations",
  ["ocds bulletin"] = "OCDS Bull",
  ["ocean & coastal management"] = "Ocean Coast Manag",
  ["ocean and coastal management"] = "Ocean Coast. Manage.",
  ["ocean and coastal research"] = "Ocean Coastal Res",
  ["ocean and polar research"] = "Ocean Polar Res.",
  ["ocean and shoreline management"] = "Ocean Shoreline Manage.",
  ["ocean development & international law"] = "Ocean Dev. Int. Law",
  ["ocean dynamics"] = "Ocean Dyn.",
  ["ocean engineering"] = "Ocean Eng.",
  ["ocean engineering & oceanography"] = "Ocean Eng. Oceanogr.",
  ["ocean manage."] = "Ocean Manage.",
  ["ocean modelling"] = "Ocean Model (Oxf)",
  ["ocean physics and engineering"] = "Ocean Phys. Eng.",
  ["ocean science"] = "Ocean Sci.",
  ["ocean science and technology"] = "Ocean Sci. Technol.",
  ["ocean science discussions"] = "Ocean Sci. Discuss.",
  ["ocean science journal"] = "Ocean Sci. J.",
  ["ocean science journal : osj"] = "Ocean Sci J",
  ["ocean university of qingdao"] = "J. Ocean Univ. Qingdao",
  ["ocean-land-atmosphere research"] = "Ocean-Land-Atmos. Res.",
  ["oceania; a journal devoted to the study of the native peoples of australia, new guinea, and the islands of the pacific"] = "Oceania",
  ["oceanographic and marine biology"] = "Oceanogr. Mar. Biol.",
  ["oceanographic magazine"] = "Oceanogr. Mag.",
  ["oceanography & fisheries open access journal"] = "Oceanogr Fish Open Access J",
  ["oceanography (washington, d.c.)"] = "Oceanography (Wash D C)",
  ["oceanography and meteorology"] = "Oceanogr. Meteorol.",
  ["oceanologica acta"] = "Oceanolog. Acta",
  ["oceanological and hydrobiological studies"] = "Oceanol Hydrobiol Stud",
  ["ochsner clinic reports"] = "Ochsner Clin Rep",
  ["ocp: on continuing practice"] = "Ocp",
  ["octogon mathematical magazine"] = "Octogon Math. Mag.",
  ["ocular immunology and inflammation"] = "Ocul Immunol Inflamm",
  ["ocular oncology and pathology"] = "Ocul Oncol Pathol",
  ["ocular surface"] = "Ocul. Surf.",
  ["odonto-stomatologie tropicale"] = "Odontostomatol. Trop.",
  ["odonto-stomatologie tropicale = tropical dental journal"] = "Odontostomatol Trop",
  ["odontoiatria e protesi dentaria"] = "Odontoiatr Protesi Dent",
  ["odontoiatria pratica"] = "Odontoiatr Prat",
  ["odontoiatria; revista ibero-americana de medicina de la boca"] = "Odontoiatr. Rev. Iberoam. Med. Boca",
  ["odontoiatría; revista ibero-americana de medicina de la boca"] = "Odontoiatr Rev Iberoam Med Boca",
  ["odontologia (paris, france)"] = "Odontologia",
  ["odontologia atual"] = "Odontol Atual",
  ["odontologia bonaerense"] = "Ondontol. Bonaer.",
  ["odontologia bonaerense. / fopba, federacion odontologica de la pcia. de bs. as"] = "Ondontol Bonaer",
  ["odontologia capixaba"] = "Odontol Capixaba",
  ["odontologia chilena"] = "Odontol. Chil.",
  ["odontologia clinica"] = "Odontol Clin",
  ["odontologia de america"] = "Odontol Am",
  ["odontologia de américa"] = "Odontol Am",
  ["odontologia de postgrado"] = "Odontol. Postgrado",
  ["odontologia dinamica"] = "Odontol. Din.",
  ["odontologia dinámica"] = "Odontol Din",
  ["odontologia jalisciense"] = "Odontol Jalisc",
  ["odontologia peruana"] = "Odontol Peru",
  ["odontologia practica"] = "Odontol Pract",
  ["odontologia uruguaya"] = "Odontol Urug",
  ["odontological bulletin"] = "Odontol Bull",
  ["odontologie conservatrice"] = "Odontol Conserv",
  ["odontologisk revy"] = "Odontol Revy",
  ["odontologisk revy. supplement"] = "Odontol Revy Suppl",
  ["odontologisk tidskrift"] = "Odontol Tidskr",
  ["odontologiska foreningens tidskrift"] = "Odontol. Foren. Tidskr.",
  ["odontologiska föreningens tidskrift"] = "Odontol Foren Tidskr",
  ["odontologiska samfundet i finland arsbok"] = "Odontol. Samf. Finl. Arsb.",
  ["odontología bonaerense"] = "Ondontol Bonaer",
  ["odontología chilena"] = "Odontol Chil",
  ["odontología de postgrado"] = "Odontol Postgrado",
  ["odontología infantil"] = "Odontol Infant",
  ["odontology / the society of the nippon dental university"] = "Odontology",
  ["odontostomatologia e implantoprotesi"] = "Odontostomatol Implantoprotesi",
  ["odontostomatologike proodos"] = "Odontostomatol Proodos",
  ["oecd economic studies"] = "OECD Econ. Stud.",
  ["oeconomica polona"] = "Oecon Pol",
  ["oekologia plantarum"] = "Oekol. Plant.",
  ["oekologie der vögel"] = "Oekol. Vögel",
  ["oeno one"] = "OENO one",
  ["oesterreichische apotheker zeitung"] = "Osterr Apoth Ztg",
  ["oesterreichische dentisten zeitschrift"] = "Osterr Dent Z",
  ["oesterreichische wasser- und abfallwirtschaft"] = "Oesterr. Wasser Abfallwirtsch.",
  ["oeuvres completes de d’alembert. serie iii. opuscules et memoires mathematiques, 1757–1783"] = "OEuvres Complet. d’Alembert Ser. III. Opuscules et mem. math. 1757–1783",
  ["off our backs"] = "Off Our Backs",
  ["offa. berichte und mitteilungen zur urgeschiche, frühgeschichte und mittelalterarchäologie"] = "Offa",
  ["offender programs report"] = "Offender Programs Rep",
  ["offentliche gesundheitsdienst"] = "Offentl. Gesundheitsdienst",
  ["offentliche gesundheitswesen"] = "Offentl. Gesundheitswes.",
  ["officers, committees, and roster of fellows. medical society of the state of north carolina"] = "Off Comm Roster Fellows",
  ["official bulletin. american hospital association"] = "Off Bull Am Hosp Assoc",
  ["official journal of the canadian association of critical care nurses"] = "Off J Can Assoc Crit Care Nurs",
  ["official journal of the canadian association of critical care nurses / caccn"] = "Off J Can Assoc Crit Care Nurs",
  ["official publication, dental hygienists association of the state of new york"] = "Off. Publ. Dent. Hyg. Assoc. State N. Y.",
  ["official publication. dental hygienists' association of the state of new york"] = "Off Publ Dent Hyg Assoc State N Y",
  ["official publication. dental hygienists’ association of the state of new york"] = "Off Publ Dent Hyg Assoc State N Y",
  ["official records. world health organization"] = "Off Rec World Health Organ",
  ["oftalmologia (bucharest, romania : 1990)"] = "Oftalmologia",
  ["oftalmologicheskii zhurnal"] = "Oftalmol. Zh.",
  ["oftalmologicheskiĭ zhurnal"] = "Oftalmol Zh",
  ["ogam. bulletin des amis de la tradition celtique"] = "Ogam",
  ["oh. osteopathic hospitals"] = "OH",
  ["oh; osteopathic hospitals"] = "OH",
  ["ohio dental journal"] = "Ohio Dent. J.",
  ["ohio history"] = "Ohio Hist",
  ["ohio history / ohio historical society"] = "Ohio Hist",
  ["ohio journal of public health"] = "Ohio J Public Health",
  ["ohio medicine"] = "Ohio Med.",
  ["ohio medicine : journal of the ohio state medical association"] = "Ohio Med",
  ["ohio northern university law review"] = "Ohio North Univ Law Rev",
  ["ohio nurses review"] = "Ohio Nurses Rev",
  ["ohio public health"] = "Ohio Public Health",
  ["ohio state law journal"] = "Ohio State Law J",
  ["ohio state medical journal"] = "Ohio State Med. J.",
  ["ohio state university bulletin of business research"] = "Ohio State U. Bull. Bus. Res.",
  ["ohio state university mathematical research institute publications"] = "Ohio State Univ. Math. Res. Inst. Publ.",
  ["ohms digest"] = "OHMS Dig",
  ["oikos (copenhagen, denmark)"] = "Oikos",
  ["oikumene. studia ad historiam antiquam classicam et orientalem pertinentia"] = "Oikumene",
  ["oil & gas science and technology"] = "Oil Gas Sci. Technol.",
  ["oil and chemical pollution"] = "Oil Chem. Pollut.",
  ["oil and energy trends"] = "Oil Energy Trends",
  ["oil and energy trends: annual statistical review"] = "Oil Energy Trends: Annu. Stat. Rev.",
  ["oil and gas international"] = "Oil Gas Int.",
  ["oil and gas journal"] = "Oil Gas J.",
  ["oil and petrochemical pollution"] = "Oil Petrochem. Pollut.",
  ["oil crop science"] = "Oil Crop Sci.",
  ["oil, gas (hamburg, germany)"] = "Oil, Gas (Hamburg, Ger.)",
  ["oil, paint, and drug reporter"] = "Oil Paint Drug Report",
  ["oilseeds and fats, crops and lipids"] = "Oilseeds fats, Crops Lipids",
  ["oita university"] = "Rep. Fac. Eng. Oita Univ.",
  ["okajimas folia anatomica japonica"] = "Okajimas Folia Anat Jpn",
  ["okayama igakkai zasshi"] = "Okayama Igakkai zasshi",
  ["oklahoma city university law review. oklahoma city university"] = "Oklahoma City Univ Law Rev",
  ["oklahoma law review"] = "Oklahoma Law Rev",
  ["oklahoma nurse"] = "Okla. Nurse",
  ["oklahoma statutes annotated. oklahoma"] = "Okla Statut Annot Okla",
  ["okonomi og politik"] = "Okon. Polit.",
  ["okonomische analyse des rechts"] = "Okon. Anal. Rechts",
  ["olba. mersin üniversitesi kilikia arkeolojisini araştırma merkezi yayınları"] = "Olba",
  ["old english newsletter"] = "Old Engl Newsl",
  ["oldenburgische familienkunde"] = "Oldenburg Familienkd",
  ["oleagineux, corps gras, lipides"] = "Ol. Corps Gras, Lipides",
  ["olympia. die ergebnisse der von dem deutschen reich veranstalteten ausgrabung (berlin 1890–­1897)"] = "Olympia",
  ["olympische forschungen"] = "OF",
  ["omaly sy anio. hier et aujourd'hui"] = "Omaly Anio",
  ["omaly sy anio. hier et aujourd’hui"] = "Omaly Anio",
  ["oman journal of ophthalmology"] = "Oman J Ophthalmol",
  ["oman medical journal"] = "Oman Med J",
  ["omics : a journal of integrative biology"] = "OMICS",
  ["omics journal of radiology"] = "OMICS J Radiol",
  ["omics: a journal of integrative biology"] = "OMICS: J. Integr. Biol.",
  ["omni (new york, n.y.)"] = "Omni",
  ["omnia medica"] = "Omnia Med",
  ["omnia medica et therapeutica"] = "Omnia Med Ther",
  ["omnia medica. supplemento"] = "Omnia Med Suppl",
  ["omnia therapeutica"] = "Omnia Ther",
  ["omnia therapeutica. supplemento"] = "Omnia Ther Suppl",
  ["omphalina : newsletter of foray newfoundland and labrador"] = "Omphalina",
  ["oms, organic mass spectrometry"] = "Org Mass Spectrom",
  ["on the nutritive value of the major nutrients of irradiated foods and appraisal of the toxicity of irradiated foods; progress report. university of illinois (urbana-champaign campus). division of animal nutrition"] = "Nutr Value Major Nutr Irradiat Food Apprais Toxic Irradiat Food",
  ["on the state of the public health. great britain. ministry of health"] = "On State public health (1921)",
  ["on thinking"] = "On Think.",
  ["on-line journal of nursing informatics"] = "Online J Nurs Inform",
  ["ona journal"] = "ONA J",
  ["onco therapeutics"] = "Onco Ther.",
  ["oncodevelopmental biology and medicine"] = "Oncodev. Biol. Med.",
  ["oncodevelopmental biology and medicine : the journal of the international society for oncodevelopmental biology and medicine"] = "Oncodev Biol Med",
  ["oncogene research"] = "Oncogene Res",
  ["oncologica / vereniging van oncologie verpleegkundigen"] = "Oncologica",
  ["oncologie (paris, france)"] = "Oncologie (Paris)",
  ["oncology & hematology review"] = "Oncol Hematol Rev",
  ["oncology (williston park, n.y.)"] = "Oncology (Williston Park)",
  ["oncology and therapy"] = "Oncol Ther",
  ["oncology discovery"] = "Oncol Discov",
  ["oncology fellows"] = "Oncol Fellows",
  ["oncology issues"] = "Oncol Issues",
  ["oncology letters"] = "Oncol Lett",
  ["oncology nurse advisor"] = "Oncol Nurse Advis",
  ["oncology nurse edition"] = "Oncol Nurse Ed",
  ["oncology nursing forum"] = "Oncol Nurs Forum",
  ["oncology nursing news"] = "Oncol Nurs News",
  ["oncology nutrition connection : a publication of the on dpg, oncology nutrition dietetic practice group"] = "Oncol Nutr Connect",
  ["oncology reports"] = "Oncol Rep",
  ["oncology research"] = "Oncol Res",
  ["oncology research and treatment"] = "Oncol Res Treat",
  ["oncology reviews"] = "Oncol Rev",
  ["oncology signaling"] = "Oncol Signal",
  ["oncolytic virotherapy"] = "Oncolytic Virother",
  ["oncotargets and therapy"] = "Onco Targets Ther",
  ["onderstepoort journal of veterinary research"] = "Onderstepoort J. Vet. Res.",
  ["one earth"] = "One Earth",
  ["one earth (cambridge, mass.)"] = "One Earth",
  ["one ecosystem"] = "One Ecosyst.",
  ["one health (amsterdam, netherlands)"] = "One Health",
  ["one health outlook"] = "One Health Outlook",
  ["one on one"] = "One One",
  ["online international interdisciplinary research journal : an international multidisciplinary journal"] = "Online Int Interdiscip Res J",
  ["online journal of analytic combinatorics"] = "Online J. Anal. Comb.",
  ["online journal of bioinformatics : ojb"] = "Online J Bioinform",
  ["online journal of biological sciences"] = "Online J Biol Sci",
  ["online journal of communication and media technologies"] = "Online J Commun Media Technol",
  ["online journal of complementary & alternative medicine"] = "Online J Complement Altern Med",
  ["online journal of cultural competence in nursing and healthcare"] = "Online J Cult Competence Nurs Healthc",
  ["online journal of current clinical trials"] = "Online J. Curr. Clin. Trials",
  ["online journal of health and allied sciences : ojhas"] = "Online J Health Allied Sci",
  ["online journal of health ethics"] = "Online J Health Ethics",
  ["online journal of issues in nursing"] = "Online J Issues Nurs",
  ["online journal of public health informatics"] = "Online J Public Health Inform",
  ["online journal of rural and urban research"] = "Online J Rural Urban Res",
  ["online journal of rural nursing and health care : the official journal of the rural nurse organization"] = "Online J Rural Nurs Health Care",
  ["online journal of rural research and policy"] = "Online J Rural Res Policy",
  ["online journal of veterinary research : ojvr"] = "Online J Vet Res",
  ["online learning : the official journal of the online learning consortium"] = "Online Learn",
  ["online social networks and media"] = "Online Soc Netw Media",
  ["online, adl-nachrichten"] = "Online, ADL-Nachr.",
  ["ons amsterdam"] = "Ons Amst",
  ["ons connect"] = "ONS Connect",
  ["ons heem"] = "Ons Heem",
  ["ons news"] = "ONS News",
  ["ons news / oncology nursing society"] = "ONS News",
  ["ontario dentist"] = "Ont Dent",
  ["ontario health technology assessment series"] = "Ont Health Technol Assess Ser",
  ["ontario history"] = "Ont Hist",
  ["ontario medical review"] = "Ont Med Rev",
  ["ontos mathematical logic"] = "Ontos Math. Log.",
  ["ooq; occupational outlook quarterly"] = "OOQ Occup. Outlook Q.",
  ["oostvlaamsche zanten"] = "Oostvlaam Zanten",
  ["opec energy review"] = "OPEC Energy Rev.",
  ["opec review"] = "OPEC Rev.",
  ["open access animal physiology"] = "Open Access Anim Physiol",
  ["open access bioinformatics"] = "Open Access Bioinformatics",
  ["open access emergency medicine : oaem"] = "Open Access Emerg Med",
  ["open access insect physiology"] = "Open access insect physiol",
  ["open access journal of behavioural science & psychology"] = "Open Access J Behav Sci Psychol",
  ["open access journal of biomedical engineering and biosciences"] = "Open Access J Biomed Eng Biosci",
  ["open access journal of biomedical engineering and its applications"] = "Open Access J Biomed Eng Appl",
  ["open access journal of clinical trials"] = "Open Access J Clin Trials",
  ["open access journal of contraception"] = "Open Access J Contracept",
  ["open access journal of gerontology & geriatric medicine"] = "Open Access J Gerontol Geriatr Med",
  ["open access journal of neurology & neurosurgery"] = "Open Access J Neurol Neurosurg",
  ["open access journal of ophthalmology"] = "Open Access J Ophthalmol",
  ["open access journal of sports medicine"] = "Open Access J Sports Med",
  ["open access journal of toxicology"] = "Open Access J Toxicol",
  ["open access journal of translational medicine & research"] = "Open Access J Transl Med Res",
  ["open access journal of urology"] = "Open Access J Urol",
  ["open access macedonian journal of medical sciences"] = "Open Access Maced J Med Sci",
  ["open access medical statistics"] = "Open Access Med Stat",
  ["open access rheumatology : research and reviews"] = "Open Access Rheumatol",
  ["open agriculture"] = "Open Agric",
  ["open applied mathematics journal"] = "Open Appl. Math. J.",
  ["open astronomy"] = "Open Astron.",
  ["open atmospheric science journal"] = "Open Atmos. Sci. J.",
  ["open biochemistry journal"] = "Open Biochem. J.",
  ["open biochemistry letters"] = "Open Biochem. Lett.",
  ["open bioinformatics journal"] = "Open Bioinform. J.",
  ["open biology"] = "Open Biol",
  ["open biomaterials journal"] = "Open Biomater. J.",
  ["open biomedical engineering journal"] = "Open Biomed. Eng. J.",
  ["open biomedical engineering letters"] = "Open Biomed. Eng.Lett.",
  ["open biotechnology journal"] = "Open Biotechnol. J.",
  ["open ceramics"] = "Open Ceram.",
  ["open chemical engineering journal"] = "Open Chem. Eng. J.",
  ["open chemistry"] = "Open Chem",
  ["open communications in nonlinear mathematical physics"] = "Open Commun. Nonlinear Math. Phys.",
  ["open cybernetics and systemics journal"] = "Open Cybern. Syst. J.",
  ["open drug delivery journal"] = "Open Drug Delivery J.",
  ["open drug delivery reviews"] = "Open Drug Delivery Rev.",
  ["open economies review"] = "Open Economies Rev.",
  ["open engineering"] = "Open Eng.",
  ["open food science journal"] = "Open Food Sci. J.",
  ["open forum infectious diseases"] = "Open Forum Infect Dis",
  ["open geosciences"] = "Open Geosci.",
  ["open geospatial data, software and standards"] = "Open Geospatial Data Software Stand.",
  ["open glycoscience"] = "Open Glycosci",
  ["open health data"] = "Open Health Data",
  ["open heart"] = "Open Heart",
  ["open inorganic chemistry journal"] = "Open Inorg. Chem. J.",
  ["open inorganic chemistry reviews"] = "Open Inorg. Chem. Rev.",
  ["open journal for students and new scientists research ojssr"] = "Open J. Stud. New Sci. Res.",
  ["open journal of anesthesiology"] = "Open J Anesthesiol",
  ["open journal of animal sciences"] = "Open J Anim Sci",
  ["open journal of apoptosis"] = "Open J Apoptosis",
  ["open journal of biophysics"] = "Open J Biophys",
  ["open journal of bioresources"] = "Open J Bioresour",
  ["open journal of cardiovascular surgery"] = "Open J Cardiovasc Surg",
  ["open journal of clinical & medical case reports"] = "Open J Clin Med Case Rep",
  ["open journal of clinical diagnostics"] = "Open J Clin Diagn",
  ["open journal of discrete applied mathematics. odam"] = "Open J. Discrete Appl. Math.",
  ["open journal of ecology"] = "Open J Ecol",
  ["open journal of endocrine and metabolic diseases"] = "Open J Endocr Metab Dis",
  ["open journal of epidemiology"] = "Open J Epidemiol",
  ["open journal of forestry"] = "Open J For",
  ["open journal of gastroenterology"] = "Open J Gastroenterol",
  ["open journal of genetics"] = "Open J Genet",
  ["open journal of hematology"] = "Open J Hematol",
  ["open journal of immunology"] = "Open J Immunol",
  ["open journal of inorganic chemistry"] = "Open J Inorg Chem",
  ["open journal of inorganic non-metallic materials"] = "Open J. Inorg. Non-Met. Mater.",
  ["open journal of mathematical optimization (ojmo)"] = "Open J. Math. Optim.",
  ["open journal of medical microbiology"] = "Open J Med Microbiol",
  ["open journal of medical psychology"] = "Open J Med Psychol",
  ["open journal of medicinal chemistry"] = "Open J Med Chem",
  ["open journal of metal"] = "Open J Met",
  ["open journal of modern linguistics"] = "Open J Mod Linguist",
  ["open journal of molecular and integrative physiology"] = "Open J Mol Integr Physiol",
  ["open journal of nephrology"] = "Open J Nephrol",
  ["open journal of neuroscience"] = "Open J Neurosci",
  ["open journal of nursing"] = "Open J Nurs",
  ["open journal of obstetrics and gynecology"] = "Open J Obstet Gynecol",
  ["open journal of ophthalmology"] = "Open J Ophthalmol",
  ["open journal of organic polymer materials"] = "Open J. Org. Polym. Mater.",
  ["open journal of orthopedics"] = "Open J Orthop",
  ["open journal of pediatrics"] = "Open J Pediatr",
  ["open journal of philosophy"] = "Open J Philos",
  ["open journal of physical chemistry"] = "Open J Phys Chem",
  ["open journal of political science"] = "Open J Polit Sci",
  ["open journal of polymer chemistry"] = "Open J. Polym. Chem.",
  ["open journal of preventive medicine"] = "Open J Prev Med",
  ["open journal of psychiatry"] = "Open J Psychiatr",
  ["open journal of psychiatry & allied sciences"] = "Open J Psychiatry Allied Sci",
  ["open journal of radiology"] = "Open J Radiol",
  ["open journal of rheumatology and autoimmune diseases"] = "Open J Rheumatol Autoimmune Dis",
  ["open journal of social sciences"] = "Open J Soc Sci",
  ["open journal of soil science"] = "Open J. Soil Sci.",
  ["open journal of statistics"] = "Open J Stat",
  ["open journal of stomatology"] = "Open J Stomatol",
  ["open journal of synthesis theory and applications"] = "Open J. Synth. Theory Appl.",
  ["open journal of veterinary medicine"] = "Open J Vet Med",
  ["open life sciences"] = "Open Life Sci",
  ["open linguistics"] = "Open Linguist",
  ["open longevity science"] = "Open Longev Sci",
  ["open magnetic resonance journal"] = "Open Magn. Reson. J.",
  ["open mathematics"] = "Open Math.",
  ["open mathematics journal"] = "Open Math. J.",
  ["open medicinal chemistry journal"] = "Open Med. Chem. J.",
  ["open medicine (warsaw, poland)"] = "Open Med (Wars)",
  ["open medicine : a peer-reviewed, independent, open-access journal"] = "Open Med",
  ["open medicine journal"] = "Open Med J",
  ["open memo. national health and medical research council (australia). fatigue laboratory"] = "Open Memo Natl Health Med Res Counc Aust",
  ["open mind"] = "Open Mind",
  ["open mind : discoveries in cognitive science"] = "Open Mind (Camb)",
  ["open nanoscience journal"] = "Open Nanosci. J.",
  ["open neuroendocrinology journal (online)"] = "Open Neuroendocrinol J",
  ["open numerical methods journal"] = "Open Numer. Methods J.",
  ["open nutraceuticals journal"] = "Open Nutraceuticals J.",
  ["open organic chemistry journal"] = "Open Org. Chem. J.",
  ["open organic chemistry reviews"] = "Open Inorg. Chem. Rev.",
  ["open physical chemistry journal"] = "Open Phys. Chem. J.",
  ["open physical chemistry reviews"] = "Open Phys. Chem. Rev.",
  ["open physics"] = "Open Phys.",
  ["open proceedings in life sciences"] = "Open Proc. Life Sci.",
  ["open psychology"] = "Open Psychol",
  ["open quaternary"] = "Open Quat",
  ["open research europe"] = "Open Res. Europe",
  ["open spectroscopy journal"] = "Open Spectrosc. J.",
  ["open surgical oncology journal (online)"] = "Open Surg Oncol J",
  ["open systems & information dynamics"] = "Open Syst. Inf. Dyn.",
  ["open systems and information dynamics"] = "Open Syst. Inf. Dyn.",
  ["open veterinary journal"] = "Open Vet J",
  ["open water"] = "Open Water",
  ["opera botanica"] = "Opera Bot.",
  ["opera medica et physiologica"] = "Opera Med Physiol",
  ["opera pharmaceutica rariora"] = "Opera Pharm Rariora",
  ["operating and programming systems series"] = "Oper. Program. Systems Ser.",
  ["operational research an international journal"] = "Oper. Res. Int. J.",
  ["operational research quarterly"] = "Oper Res Q",
  ["operations management research"] = "Oper. Manage. Res.",
  ["operations research"] = "Oper. Res.",
  ["operations research for health care"] = "Oper Res Health Care",
  ["operations research forum"] = "Oper. Res. Forum",
  ["operations research letters"] = "Oper. Res. Lett.",
  ["operations research perspectives"] = "Oper. Res. Perspect.",
  ["operations research proceedings"] = "Oper. Res. Proc.",
  ["operations research series"] = "Oper. Res. Ser.",
  ["operations research spektrum"] = "OR Spektrum",
  ["operations research transactions. yunchouxue xuebao"] = "Oper. Res. Trans.",
  ["operations research/computer science interfaces series"] = "Oper. Res./Comput. Sci. Interfaces Ser.",
  ["operative dentistry"] = "Oper Dent",
  ["operative dentistry. supplement"] = "Oper Dent Suppl",
  ["operative neurosurgery (hagerstown, md.)"] = "Oper Neurosurg (Hagerstown)",
  ["operative orthopadie und traumatologie"] = "Oper Orthop Traumatol",
  ["operative orthopädie und traumatologie"] = "Oper Orthop Traumatol",
  ["operative techniques in orthopaedics"] = "Oper Tech Orthop",
  ["operative techniques in otolaryngology--head and neck surgery"] = "Oper Tech Otolayngol Head Neck Surg",
  ["operative techniques in sports medicine"] = "Oper Tech Sports Med",
  ["operative techniques in thoracic and cardiovascular surgery : a comparative atlas : an official publication of the american association for thoracic surgery"] = "Oper Tech Thorac Cardiovasc Surg",
  ["operator theory: advances and applications"] = "Oper. Theory Adv. Appl.",
  ["operators and matrices"] = "Oper. Matrices",
  ["ophtalmologie : organe de la societe francaise d’ophtalmologie"] = "Ophtalmologie",
  ["ophtalmologie : organe de la société française d'ophtalmologie"] = "Ophtalmologie",
  ["ophthalmic & physiological optics : the journal of the british college of ophthalmic opticians (optometrists)"] = "Ophthalmic Physiol Opt",
  ["ophthalmic and physiological optics"] = "Ophthalmic Physiol. Opt.",
  ["ophthalmic epidemiology"] = "Ophthalmic Epidemiol",
  ["ophthalmic genetics"] = "Ophthalmic Genet",
  ["ophthalmic literature"] = "Ophthalmic Lit",
  ["ophthalmic medical image analysis : 6th international workshop, omia 2019, held in conjunction with miccai 2019, shenzhen, china, october 17, proceedings. omia (workshop) (6th : 2019 : shenzhen shi, china)"] = "Ophthalmic Med Image Anal (2019)",
  ["ophthalmic paediatrics and genetics"] = "Ophthalmic Paediatr Genet",
  ["ophthalmic plastic and reconstructive surgery"] = "Ophthalmic Plast Reconstr Surg",
  ["ophthalmic research"] = "Ophthalmic Res",
  ["ophthalmic seminars"] = "Ophthalmic Semin",
  ["ophthalmic surgery"] = "Ophthalmic Surg",
  ["ophthalmic surgery and lasers"] = "Ophthalmic Surg Lasers",
  ["ophthalmic surgery, lasers & imaging : the official journal of the international society for imaging in the eye"] = "Ophthalmic Surg Lasers Imaging",
  ["ophthalmic surgery, lasers & imaging retina"] = "Ophthalmic Surg Lasers Imaging Retina",
  ["ophthalmic surgery, lasers and imaging"] = "Ophthalmic Surg. Lasers Imaging",
  ["ophthalmologia ibero americana"] = "Ophthalmol Ibero Am",
  ["ophthalmologica. journal international d'ophtalmologie. international journal of ophthalmology. zeitschrift für augenheilkunde"] = "Ophthalmologica",
  ["ophthalmologica. journal international d’ophtalmologie. international journal of ophthalmology. zeitschrift fur augenheilkunde"] = "Ophthalmologica",
  ["ophthalmology and eye diseases"] = "Ophthalmol Eye Dis",
  ["ophthalmology and therapy"] = "Ophthalmol Ther",
  ["ophthalmology clinics of north america"] = "Ophthalmol Clin North Am",
  ["ophthalmology international"] = "Ophthalmol Int",
  ["ophthalmology research and reports"] = "Ophthalmol Res Rep",
  ["ophthalmology. glaucoma"] = "Ophthalmol Glaucoma",
  ["ophthalmology. retina"] = "Ophthalmol Retina",
  ["opinions of the attorney general--state of maryland. maryland. attorney general's office"] = "Opin Atty Gen State Md Md Atty Gen Off",
  ["opsearch. quarterly journal of the operational research society of india"] = "Opsearch",
  ["optica acta"] = "Opt Acta (Lond)",
  ["optica quantum"] = "Optica Quantum",
  ["optical and quantum electronics"] = "Opt. Quantum Electron.",
  ["optical developments"] = "Opt Dev",
  ["optical engineering"] = "Opt. Eng.",
  ["optical engineering (redondo beach, calif.)"] = "Opt Eng",
  ["optical fiber technology"] = "Opt. Fiber Technol.",
  ["optical journal and review of optometry"] = "Opt J Rev Optom",
  ["optical materials"] = "Opt. Mater.",
  ["optical materials express"] = "Opt. Mater. Express",
  ["optical materials: x"] = "Opt. Mater.: X",
  ["optical memory and neural networks"] = "Opt. Mem. Neural Networks",
  ["optical nanoscopy"] = "Opt. Nanosc.",
  ["optical review"] = "Opt. Rev.",
  ["optical switching and networking"] = "Opt. Switching Networking",
  ["opticheskii zhurnal"] = "Opt. Zh.",
  ["optics & laser technology"] = "Opt. Laser Technol.",
  ["optics & photonics news"] = "Opt. Photonics News",
  ["optics and laser technology"] = "Opt Laser Technol",
  ["optics and lasers in engineering"] = "Opt. Lasers Eng.",
  ["optics and photonics letters"] = "Opt. Photonics Lett.",
  ["optics and photonics news"] = "Opt Photonics News",
  ["optics and spectroscopy"] = "Opt. Spectrosc.",
  ["optics and spectroscopy (ussr) [translation of optika i spektroskopiya]"] = "Opt. Spectrosc.",
  ["optics communication"] = "Opt. Commun.",
  ["optics communications"] = "Opt. Commun.",
  ["optics continuum"] = "Opt. Continuum",
  ["optics express"] = "Opt Express",
  ["optics letters"] = "Opt Lett",
  ["optics news"] = "Opt. News",
  ["optik & photonik"] = "Opt. Photonik",
  ["optik (stuttgart)"] = "Optik (Stuttgart)",
  ["optik - international journal for light and electron optics"] = "Opt. - Int. J. Light Electron Opt.",
  ["optika i spektroskopiya"] = "Opt. Spektrosk.",
  ["optika i spektroskopiya [optics and spectroscopy (ussr)]"] = "Opt. Spektrosk. [Opt. Spectrosc. (USSR)]",
  ["optiko-mekhanicheskaya promyshlennost"] = "Opt.-Mekh. Prom-st.",
  ["optiko–mekhanicheskaya promyshlennost [soviet journal of optical technology]"] = "Opt.–Mekh. Prom. [Sov. J. Opt. Technol.]",
  ["optimal control applications & methods"] = "Optimal Control Appl. Methods",
  ["optimal control applications and methods"] = "Optim. Control. Appl. Methods",
  ["optimization and engineering"] = "Optim. Eng.",
  ["optimization and engineering. international multidisciplinary journal to promote optimization theory & applications in engineering sciences"] = "Optim. Eng.",
  ["optimization letters"] = "Optim. Lett.",
  ["optimization methods & software"] = "Optim Methods Softw",
  ["optimization methods and software"] = "Optim. Methods Software",
  ["optimization. a journal of mathematical programming and operations research"] = "Optimization",
  ["options (van nuys, calif.)"] = "Options (Van Nuys)",
  ["options for policy and practice"] = "Options Policy Pract",
  ["opto-electronics review"] = "Opto-Electron. Rev.",
  ["optoelectronic materials and devices"] = "Optoelectron Mater Devices",
  ["optoelectronics and advanced materials, rapid communications"] = "Optoelectron. Adv. Mater. Rapid Commun.",
  ["optoelectronics and communications conference"] = "Optoelectron Commun Conf",
  ["optoelectronics letters"] = "Optoelectron. Lett.",
  ["optoelectronics, instrumentation and data processing"] = "Optoelectron. Instrum. Data Process.",
  ["optomechatronic technologies (isot), 2010 international symposium on : 25-27 oct. 2010 : [toronto, on]. international symposium on optomechatronic technologies (2010 : toronto, ont.)"] = "Optomechatron Technol ISOT 2010 Int Symp (2010)",
  ["optometric glaucoma society e-journal"] = "Optom Glaucoma Soc E J",
  ["optometry & visual performance"] = "Optom Vis Perform",
  ["optometry (st. louis, mo.)"] = "Optometry",
  ["optometry and vision science"] = "Optom. Vis. Sci.",
  ["optometry and vision science : official publication of the american academy of optometry"] = "Optom Vis Sci",
  ["optometry clinics"] = "Optom. Clin.",
  ["optometry clinics : the official publication of the prentice society"] = "Optom Clin",
  ["optometry, open access"] = "Optom Open Access",
  ["opus. rivista internazionale per la storia economica e sociale dell’antichità"] = "Opus",
  ["opuscula atheniensia"] = "OpAth",
  ["opuscula atheniensia: acta inst. athen. regni sueciae"] = "OAth",
  ["opuscula instituti romani finlandiae"] = "OpFin",
  ["opuscula mathematica"] = "Opuscula Math.",
  ["opuscula medica"] = "Opusc Med",
  ["opuscula philolichenum"] = "Opusc Philolichenum",
  ["opuscula pompeiana"] = "OpPomp",
  ["opuscula romana"] = "OpRom",
  ["opuscula romana: acta inst. rom. regni sueciae"] = "ORom",
  ["opuscula zoologica instituti zoosystematici et oecologici universitatis budapestinensis"] = "Opusc Zool Inst Zoosyst Oecol Univ Bp",
  ["or 2.0 context-aware operating theaters, computer assisted robotic endoscopy, clinical image-based procedures, and skin image analysis : first international workshop, or 2.0 2018, 5th international workshop, care 2018, 7th international workshop, clip 2018, third international workshop, isic 2018, held in conjunction with miccai 2018, granada, spain, september 16-20, 2018 : proceedings. or 2.0 (workshop) (1st : 2018 : granada, spain)"] = "OR 2.0 Context Aware Oper Theaters Comput Assist Robot Endosc Clin Image Based Proced Skin Image Anal (2018)",
  ["or manager"] = "OR Manager",
  ["or spectrum"] = "OR Spectrum",
  ["or spectrum : quantitative approaches in management"] = "OR Spectr",
  ["or spectrum. quantitative approaches in management"] = "OR Spectrum",
  ["or spektrum"] = "OR Spektrum",
  ["or tech"] = "OR Tech",
  ["oral & implantology"] = "Oral Implantol (Rome)",
  ["oral and maxillofacial surgery"] = "Oral Maxillofac Surg",
  ["oral and maxillofacial surgery cases"] = "Oral Maxillofac Surg Cases",
  ["oral and maxillofacial surgery clinics of north america"] = "Oral Maxillofac Surg Clin North Am",
  ["oral biosciences & medicine : obm"] = "Oral Biosci Med",
  ["oral diseases"] = "Oral Dis",
  ["oral health"] = "Oral Health",
  ["oral health & preventive dentistry"] = "Oral Health Prev Dent",
  ["oral health and dental management"] = "Oral Health Dent Manag",
  ["oral health and dental studies"] = "Oral Health Dent Stud",
  ["oral health and preventive dentistry"] = "Oral Health Prev. Dent.",
  ["oral health case reports"] = "Oral Health Case Rep",
  ["oral history"] = "Oral Hist (Colch)",
  ["oral hygiene"] = "Oral Hyg",
  ["oral implantology"] = "Oral Implantol",
  ["oral microbiology and immunology"] = "Oral Microbiol Immunol",
  ["oral oncology"] = "Oral Oncol",
  ["oral radiology"] = "Oral Radiol",
  ["oral science international"] = "Oral Sci Int",
  ["oral sciences reviews"] = "Oral Sci Rev",
  ["oral surgery"] = "Oral Surg",
  ["oral surgery, oral diagnosis"] = "Oral Surg. Oral Diagn.",
  ["oral surgery, oral diagnosis : osd"] = "Oral Surg Oral Diagn",
  ["oral surgery, oral medicine, and oral pathology"] = "Oral Surg Oral Med Oral Pathol",
  ["oral surgery, oral medicine, oral pathology"] = "Oral Surg. Oral Med. Oral Pathol.",
  ["oral surgery, oral medicine, oral pathology and oral radiology"] = "Oral Surg Oral Med Oral Pathol Oral Radiol",
  ["oral surgery, oral medicine, oral pathology, oral radiology and endodontics"] = "Oral Surg. Oral Med. Oral Pathol. Oral Radiol. Endod.",
  ["oral surgery, oral medicine, oral pathology, oral radiology andendodontics oral surg oral med oral pathol oral radio"] = "Endod",
  ["oral surgery, oral medicine, oral pathology, oral radiology, and endodontics"] = "Oral Surg Oral Med Oral Pathol Oral Radiol Endod",
  ["oral-prophylaxe / herausgeber, verein fur zahnhygiene e.v"] = "Oralprophylaxe",
  ["orale implantologie"] = "Orale Implantol",
  ["orange county dental society bulletin"] = "Orange Cty Dent Soc Bull",
  ["orbis economicus"] = "Orbis Econ.",
  ["orbis terrarum. internationale zeitschrift für historische geographie der alten welt"] = "OrbTerr",
  ["orbit (amsterdam, netherlands)"] = "Orbit",
  ["orbita mathematicae"] = "Orbita Math.",
  ["orchids (west palm beach, fla.)"] = "Orchids",
  ["order (dordrecht, netherlands)"] = "Order (Dordr)",
  ["order. a journal on the theory of ordered sets and its applications"] = "Order",
  ["ordona. rapports et études"] = "Ordona",
  ["ore and energy resource geology"] = "Ore Energy Resour. Geol.",
  ["ore geology reviews"] = "Ore Geol. Rev.",
  ["oregon business review"] = "Oregon Bus. Rev.",
  ["oregon health bulletin"] = "Or Health Bull",
  ["oregon historical quarterly. oregon historical society"] = "Oreg Hist Q",
  ["oregon law review"] = "Oregon Law Rev",
  ["oregon nurse"] = "Oreg. Nurse",
  ["oregon state dental journal"] = "Oreg State Dent J",
  ["organic & biomolecular chemistry"] = "Org Biomol Chem",
  ["organic agriculture"] = "Org. Agric.",
  ["organic and biomolecular chemistry"] = "Org. Biomol. Chem.",
  ["organic and medicinal chemistry letters"] = "Org Med Chem Lett",
  ["organic chemistry frontiers"] = "Org. Chem. Front.",
  ["organic chemistry frontiers : an international journal of organic chemistry"] = "Org Chem Front",
  ["organic chemistry international"] = "Org Chem Int",
  ["organic electronics"] = "Org. Electron.",
  ["organic geochemistry"] = "Org. Geochem.",
  ["organic letters"] = "Org Lett",
  ["organic magnetic resonance"] = "Org. Magn. Reson.",
  ["organic mass spectrometry"] = "Org. Mass Spectrom.",
  ["organic preparations and procedures international"] = "Org. Prep. Proced. Int.",
  ["organic process research & development"] = "Org Process Res Dev",
  ["organic process research and development"] = "Org. Process Res. Dev.",
  ["organic reaction mechanics"] = "Org. React. Mech.",
  ["organic reaction mechanisms"] = "Org. React. Mech.",
  ["organic reactions"] = "Org. React.",
  ["organic syntheses"] = "Org. Synth.",
  ["organic syntheses; an annual publication of satisfactory methods for the preparation of organic chemicals"] = "Organic Synth",
  ["organisator, der"] = "Organisator",
  ["organisms diversity & evolution"] = "Org. Divers. Evol.",
  ["organisms, diversity & evolution"] = "Org Divers Evol",
  ["organization & environment"] = "Organ Environ",
  ["organization (london, england)"] = "Organization (Lond)",
  ["organization and environment"] = "Organ. Environ.",
  ["organizational behavior and human decision processes"] = "Organ Behav Hum Decis Process",
  ["organizational behavior and human performance"] = "Organ Behav Hum Perform",
  ["organizational cybersecurity journal: practice, process and people"] = "Organ. Cybersecur. J.: Pract. Process People",
  ["organizational dynamics"] = "Organ Dyn",
  ["organizational ethics : healthcare, business, and policy : oe"] = "Organ Ethic",
  ["organizational research methods"] = "Organ Res Methods",
  ["organizatsiya vychislitelnykh struktur i protsessov"] = "Organiz. Vychisl. Struktur. Process.",
  ["organohalogen compounds"] = "Organohalogen Compd.",
  ["organometallic chemistry"] = "Organomet. Chem.",
  ["organon f"] = "Organon F (Bratisl)",
  ["orgyn : organon's magazine on women & health"] = "ORGYN",
  ["orgyn : organon’s magazine on women & health"] = "ORGYN",
  ["oriens antiquus. rivista del centro per le antichità e la storia dell’arte del vicino oriente"] = "OrAnt",
  ["oriens antiquus: rivista del centro per le antichità e la storia dell’arte del vicino oriente"] = "OA",
  ["oriens christianus"] = "OrChr",
  ["oriens christianus: hefte für die kunde des christlichen orients"] = "OC",
  ["oriens extremus"] = "Oriens Extremus",
  ["orient. the reports of the society for near eastern studies in japan"] = "Orient",
  ["orientación médica"] = "Orientac Med",
  ["oriental insects"] = "Orient Insects",
  ["oriental institute publications"] = "OIP",
  ["oriental journal of chemistry"] = "Orient. J. Chem.",
  ["oriental pharmacy and experimental medicine"] = "Orient Pharm Exp Med",
  ["orientalia (pontificio istituto biblico)"] = "Or",
  ["orientalia christiana periodica"] = "OCP",
  ["orientalia lovaniensia periodica"] = "OLP",
  ["orientalia suecana. an international journal of indological, iranian, semitic, turkic studies"] = "OrSu",
  ["orientalistische literaturzeitung"] = "OLZ",
  ["oriente cristiano"] = "Oriente Crist",
  ["oriente moderno"] = "Oriente Mod",
  ["origini. preistoria e protostoria delle civiltà antiche"] = "Origini",
  ["origins of life"] = "Orig Life",
  ["origins of life and evolution of the biosphere"] = "Origins Life Evol. Biosphere",
  ["origins of life and evolution of the biosphere : the journal of the international society for the study of the origin of life"] = "Orig Life Evol Biosph",
  ["orins [reports]. u.s. atomic energy commission"] = "ORINS Rep US At Energy Comm",
  ["orizzonti della ortopedia odierna e della riabilitazione"] = "Orizz Ortop Odie Riabil",
  ["orizzonti. rassegna di archeologia"] = "Orizzonti",
  ["orl-head and neck nursing"] = "ORL. Head Neck Nurs.",
  ["orl-head and neck nursing : official journal of the society of otorhinolaryngology and head-neck nurses"] = "ORL Head Neck Nurs",
  ["orl; journal for oto-rhino-laryngology and its related specialties"] = "ORL J Otorhinolaryngol Relat Spec",
  ["orl; journal of oto-rhino-laryngology and its related specialties"] = "ORL. J. Otorhinolaryngol. Relat. Spec.",
  ["ornac journal"] = "ORNAC J",
  ["ornis fennica"] = "Ornis Fenn",
  ["ornis norvegica"] = "Ornis Norv",
  ["ornis svecica"] = "Ornis Svec",
  ["ornithological applications"] = "Ornithol. Appl.",
  ["ornithological monographs"] = "Ornithol Monogr",
  ["ornithological science"] = "Ornithol Sci",
  ["ornithologische abhandlungen"] = "Ornithol. Abh.",
  ["ornithologische beobachter, der"] = "Ornithol. Beob.",
  ["ornithologische jahreshefte für baden-württemberg"] = "Ornithol. Jahresh. Baden-Württ.",
  ["ornithology research"] = "Ornith. Res.",
  ["ornitología neotropical"] = "Ornitol Neotrop",
  ["ornl-nsic [reports]. u.s. atomic energy commission"] = "ORNL-NSIC Rep",
  ["ornl-tm [reports]. u.s. atomic energy commission"] = "ORNL-TM Rep",
  ["ornl. u.s. atomic energy commission"] = "ORNL",
  ["oro [reports]. u.s. atomic energy commission"] = "ORO Rep",
  ["oro reports"] = "ORO Rep.",
  ["orovosok lapja (magyar orvosok szabad szakszervezete)"] = "Orvosok Lapja Nepeu",
  ["orphanet journal of rare diseases"] = "Orphanet J Rare Dis",
  ["orpheus. journal of indo-european, palaeo-balkan and thracian studies"] = "OrpheusThracSt",
  ["orpheus. rivista di umanità classica e cristiana"] = "Orpheus",
  ["orpheus: rivista di umanità classica e cristiana"] = "Orpheus",
  ["orthodontic review"] = "Orthod Rev",
  ["orthodontic waves (english ed.)"] = "Orthod Waves",
  ["orthodontics & craniofacial research"] = "Orthod Craniofac Res",
  ["orthodontics : the art and practice of dentofacial enhancement"] = "Orthodontics (Chic.)",
  ["orthodontics and craniofacial research"] = "Orthod. Craniofac. Res.",
  ["orthodontie francaise"] = "Orthod. Fr.",
  ["orthodontike epitheorese"] = "Orthod. Epitheorese",
  ["orthodontike epitheorese : epiotemoniko periodiko tes orthodontikes etaireias tes ellados / oee"] = "Orthod Epitheorese",
  ["orthodontikē epitheōrēsē : epiotēmoniko periodiko tēs orthodontikēs etaireias tēs ellados"] = "Orthod Epitheor",
  ["orthopaedic audio-synopsis continuing medical education"] = "Orthop Audiosynop Contin Med Educ",
  ["orthopaedic journal of sports medicine"] = "Orthop J Sports Med",
  ["orthopaedic knowledge online"] = "Orthop Knowl Online",
  ["orthopaedic nursing"] = "Orthop. Nurs.",
  ["orthopaedic nursing / national association of orthopaedic nurses"] = "Orthop Nurs",
  ["orthopaedic physical therapy clinics of north america"] = "Orthop Phys Ther Clin N Am",
  ["orthopaedic review"] = "Orthop Rev",
  ["orthopaedic surgery"] = "Orthop Surg",
  ["orthopaedics & traumatology, surgery & research : otsr"] = "Orthop Traumatol Surg Res",
  ["orthopaedics and trauma"] = "Orthop Trauma",
  ["orthopädie-technik (english ed.)"] = "Orthop Tech (Engl Ed)",
  ["orthopedic & muscular system : current research"] = "Orthop Muscular Syst",
  ["orthopedic clinics of north america"] = "Orthop. Clin. North Am.",
  ["orthopedic nursing"] = "Orthop Nurs",
  ["orthopedic nursing / national association of orthopedic nurses"] = "Orthop Nurs",
  ["orthopedic research and reviews"] = "Orthop Res Rev",
  ["orthopedic reviews"] = "Orthop Rev (Pavia)",
  ["orthopedics and sports medicine : open access journal"] = "Orthop Sports Med",
  ["ortopedia e traumatologia dell'apparato motore"] = "Ortop Traumatol Appar Mot",
  ["ortopedia maxilar"] = "Ortop Maxilar",
  ["ortopedia, traumatologia, rehabilitacja"] = "Ortop Traumatol Rehabil",
  ["ortopediia travmatologiia i protezirovanie"] = "Ortop Travmatol Protez",
  ["oruen : the cns journal"] = "Oruen",
  ["orvosi hetilap"] = "Orv Hetil",
  ["orvostorteneti kozlemenyek"] = "Orvostort Kozl",
  ["orvostörténeti közlemények"] = "Orvostort Kozl",
  ["oryx : the journal of the fauna preservation society"] = "Oryx",
  ["os urbanitas"] = "Urbanitas",
  ["osa continuum"] = "OSA Contin",
  ["osa technical digest (online)"] = "OSA Tech. Dig. Online",
  ["osaka city medical journal"] = "Osaka City Med J",
  ["osaka daigaku shigaku zasshi (journal of the osaka university dental society)"] = "Osaka Daigaku Shigaku Zasshi",
  ["osaka economic papers"] = "Osaka Econ. Pap.",
  ["osaka journal of mathematics"] = "Osaka J. Math.",
  ["osaka university law review"] = "Osaka Univ Law Rev",
  ["osgoode hall law journal"] = "Osgoode Hall Law J",
  ["osler library newsletter"] = "Osler Libr Newsl",
  ["osler library studies in the history of medicine"] = "Osler Libr Stud Hist Med",
  ["osnabrucker schriften zur mathematik. reihe d: mathematisch-didaktische manuskripte"] = "Osnabruck. Schrift. Math. Reihe D Math.-Didakt. Manuskr.",
  ["osnabrucker schriften zur mathematik. reihe i: informatik"] = "Osnabruck. Schrift. Math. Reihe I Inform.",
  ["osnabrucker schriften zur mathematik. reihe p: preprints"] = "Osnabruck. Schrift. Math. Reihe P Preprints",
  ["osnabrucker schriften zur mathematik. reihe u: materialien zum mathematikunterricht"] = "Osnabruck. Schrift. Math. Reihe U Mater. Mathematikunterr.",
  ["osnabrücker studien zur mathematik"] = "Osnabrück. Stud. Math.",
  ["osong public health and research perspectives"] = "Osong Public Health Res Perspect",
  ["ospedale maggiore"] = "Osp. Maggiore",
  ["ospedale psichiatrico"] = "Osp. Psichiatr.",
  ["ospedali d'italia - chirurgia"] = "Osp Ital Chir",
  ["ospedali d’italia - chirurgia"] = "Osp Ital Chir",
  ["ospedali d’italia chirurgia"] = "Osp. Ital. Chir.",
  ["osservatore romano (english ed.)"] = "Oss. rom.",
  ["ostbairische grenzmarken. passauer jahrbuch für geschichte, kunst und volkskunde"] = "OstbGrenzm",
  ["osteoarthritis and cartilage"] = "Osteoarthritis Cartilage",
  ["osteoarthritis and cartilage / oars, osteoarthritis research society"] = "Osteoarthritis Cartilage",
  ["osteoarthritis and cartilage open"] = "Osteoarthr Cartil Open",
  ["osteopathic hospital"] = "Osteopath. Hosp.",
  ["osteopathic hospital leadership"] = "Osteopath Hosp Leadersh",
  ["osteopathic medicine and primary care"] = "Osteopath Med Prim Care",
  ["osteopathic medicine and surgery"] = "Osteopath Med Surg",
  ["osteoporosis and sarcopenia"] = "Osteoporos Sarcopenia",
  ["osteoporosis international"] = "Osteoporos. Int.",
  ["osteoporosis international : a journal established as result of cooperation between the european foundation for osteoporosis and the national osteoporosis foundation of the usa"] = "Osteoporos Int",
  ["osterreichische akademie der wissenschaften, philosophisch-historische klasse, sitzungsberichte, 252. band, 5. abhandlung; veroffentlichungen der kommission fur geschichte der erziehung und des unterrichts"] = "Osterreichische Akademie der Wissenschaften, Philosophisch-Historische Klasse, Sitzungsberichte, 252. Band, 5. Abhandl",
  ["osterreichische akademie der wissenschaften.kommission fur byzantinistik.institut fur byzantinistik der universitat wien"] = "Osterreichische Akademie der Wissenschaften.Kommission fur Byzantinistik.Institut fur By",
  ["osterreichische arztezeitung"] = "Osterr Arzteztg",
  ["osterreichische botanische zeitschrift"] = "Osterr. Bot. Z.",
  ["osterreichische chemie zeitschrift"] = "Osterr. Chem. Z.",
  ["osterreichische dentisten zeitschrift"] = "Osterr. Dent. Z.",
  ["osterreichische hebammenzeitung"] = "Osterr. Hebammenztg.",
  ["osterreichische kneipp-magazin"] = "Osterr. Kneipp. Mag.",
  ["osterreichische krankenpflegezeitschrift"] = "Osterr. Krankenpflegez.",
  ["osterreichische pflegezeitschrift : organ des osterreichischen gesundheits- und krankenpflegeverbands"] = "Osterreichische Pflegezeitschrift",
  ["osterreichische schwesternzeitung"] = "Osterr. Schwesternztg.",
  ["osterreichische zahnarzte-zeitung"] = "Osterr. Zahnarzteztg.",
  ["osterreichische zahnarzts-zeitung; mitteilungen der bundesfachgruppe fur zahnheilkunde der osterreichischen arztekammer"] = "Osterr Zahnarzteztg",
  ["osterreichische zahnprothetik"] = "Osterr. Zahnprothet.",
  ["osterreichische zahntechniker handwerk"] = "Osterr. Zahntech. Handwerk",
  ["osterreichische zeitschrift fur erforschung und bekampfung der krebskrankheit"] = "Osterr. Z. Erforsch. Bekampf. Krebskr.",
  ["osterreichische zeitschrift fur onkologie. austrian journal of oncology"] = "Osterr Z Onkol",
  ["osterreichische zeitschrift fur statistik und informatik"] = "Osterr Z Stat Inform",
  ["osterreichische zeitschrift fur stomatologie"] = "Osterr. Z. Stomatol.",
  ["ostkirchliche studien"] = "OS",
  ["ostomy/wound management"] = "Ostomy Wound Manage",
  ["ostraka. rivista di antichità"] = "Ostraka",
  ["ot practice"] = "OT Pract",
  ["ota international : the open access journal of orthopaedic trauma"] = "OTA Int",
  ["otago law review"] = "Otago Law Rev",
  ["otechestvennai︠a︡ istorii︠a︡"] = "Otechestvennaia Istor",
  ["otechestvennye arkhivy"] = "Otecestv Arh",
  ["otjr : occupation, participation and health"] = "OTJR (Thorofare N J)",
  ["oto open"] = "OTO Open",
  ["oto-rhino-laryngologia danubiana"] = "Otorhinolaryngol Danub",
  ["oto-rino-laringologia italiana"] = "Otorinolaringol. Ital.",
  ["oto-rino-laringologia şi oftalmologia"] = "Otorinolaringol Oftalmol",
  ["otolaryngologia polska"] = "Otolaryngol. Pol.",
  ["otolaryngologia polska = the polish otolaryngology"] = "Otolaryngol Pol",
  ["otolaryngologia polska. the polish otolaryngology"] = "Otolaryngol Pol",
  ["otolaryngologic clinics of north america"] = "Otolaryngol Clin North Am",
  ["otolaryngology (sunnyvale, calif.)"] = "Otolaryngol (Sunnyvale)",
  ["otolaryngology - head and neck surgery"] = "Otolaryngol. Head Neck Surg.",
  ["otolaryngology and head and neck surgery"] = "Otolaryngol Head Neck Surg (1979)",
  ["otolaryngology--head and neck surgery : official journal of american academy of otolaryngology-head and neck surgery"] = "Otolaryngol Head Neck Surg",
  ["otology & neurotology : official publication of the american otological society, american neurotology society [and] european academy of otology and neurotology"] = "Otol Neurotol",
  ["otology and neurotology"] = "Otol. Neurotol.",
  ["otorhinolaryngology-head and neck surgery"] = "Otorhinolaryngol Head Neck Surg",
  ["ottawa law review"] = "Ottawa Law Rev",
  ["ou daigaku shigakushi"] = "Ou Daigaku Shigakushi",
  ["ou mei yan jiu = euramerica"] = "Ou Mei Yan Jiu",
  ["oudheidkundige mededeelingen van het rijksmuseum van oudheden te leiden = nuntii ex museo antiquario leidensi"] = "Oudheidkd Meded Rijksmus Oudheiden Leiden",
  ["oudheidkundige mededelingen uit het rijksmuseum van oudheden te leiden"] = "OudhMeded",
  ["oudheidkundige mededelingen uit het rijksmuseum van oudheiden te leiden"] = "OMRL",
  ["our planet : the magazine of the united nations environment programme"] = "Our Planet",
  ["outbreak, surveillance and investigation reports"] = "Outbreak Surveill Investig Rep",
  ["outcomes management"] = "Outcomes Manag",
  ["outcomes management for nursing practice"] = "Outcomes Manag Nurs Pract",
  ["outlook and bulletin, southern dental society of new jersey"] = "Outlook Bull. South. Dent. Soc. N. J.",
  ["outlook and bulletin. southern dental society of new jersey"] = "Outlook Bull South Dent Soc N J",
  ["outlook for the blind and the teachers forum"] = "Outlook Blind Teach Forum",
  ["outlook magazine"] = "Outlook Mag",
  ["outlook on agriculture"] = "Outlook Agric",
  ["outstanding contributions to logic"] = "Outst. Contrib. Log.",
  ["ouverture philosophique"] = "Ouvert. Philos.",
  ["overijssel (enschede, netherlands)"] = "Overijssel",
  ["oxford applied mathematics and computing science series"] = "Oxford Appl. Math. Comput. Sci. Ser.",
  ["oxford bulletin of economics and statistics"] = "Oxf Bull Econ Stat",
  ["oxford classic texts in the physical sciences"] = "Oxf. Class. Texts Phys. Sci.",
  ["oxford development studies"] = "Oxford Devel. Stud.",
  ["oxford economic papers"] = "Oxf Econ Pap",
  ["oxford engineering science series"] = "Oxford Eng. Sci. Ser.",
  ["oxford german studies"] = "Oxf Ger Stud",
  ["oxford graduate texts"] = "Oxf. Grad. Texts",
  ["oxford graduate texts in mathematics"] = "Oxf. Grad. Texts Math.",
  ["oxford journal of archaeology"] = "Oxford J. Archaeol.",
  ["oxford journal of legal studies"] = "Oxf J Leg Stud",
  ["oxford lecture series in mathematics and its applications"] = "Oxford Lecture Ser. Math. Appl.",
  ["oxford logic guides"] = "Oxford Logic Guides",
  ["oxford master series in physics"] = "Oxf. Master Ser. Phys.",
  ["oxford master series in statistical computational, and theoretical physics"] = "Oxf. Master Ser. Stat. Comput. Theor. Phys.",
  ["oxford mathematical monographs"] = "Oxford Math. Monogr.",
  ["oxford medical case reports"] = "Oxf Med Case Reports",
  ["oxford medical school gazette"] = "Oxf Med Sch Gaz",
  ["oxford open climate change"] = "Oxford Open Clim. Change",
  ["oxford open digital health"] = "Oxford Open Digital Health",
  ["oxford open energy"] = "Oxford Open Energy",
  ["oxford open immunology"] = "Oxf Open Immunol",
  ["oxford open infrastructure and health"] = "Oxford Open Infrastruct. Health",
  ["oxford open materials science"] = "Oxford Open Mater. Sci.",
  ["oxford open neuroscience"] = "Oxford Open Neurosci.",
  ["oxford philosophical concepts"] = "Oxford Philos. Concepts",
  ["oxford philosophical monographs"] = "Oxf. Philos. Monogr.",
  ["oxford readings in philosophy"] = "Oxf. Read. Philos.",
  ["oxford review of economic policy"] = "Oxf Rev Econ Policy",
  ["oxford review of education"] = "Oxf Rev Educ",
  ["oxford reviews of reproductive biology"] = "Oxf Rev Reprod Biol",
  ["oxford science publications"] = "Oxford Sci. Publ.",
  ["oxford series in social cognition and social neuroscience"] = "Oxf Ser Soc Cogn Soc Neurosci",
  ["oxford series on materials modelling"] = "Oxf. Ser. Mater. Model.",
  ["oxford series on realism in quantum physics"] = "Oxford Ser. Realism Quantum Phys.",
  ["oxford slavonic papers"] = "Oxf Slav Pap",
  ["oxford statistical science series"] = "Oxford Statist. Sci. Ser.",
  ["oxford studies in ancient philosophy"] = "Oxf Stud Anc Philos",
  ["oxford studies in philosophy of science"] = "Oxf. Stud. Philos. Sci.",
  ["oxford studies in probability"] = "Oxford Stud. Probab.",
  ["oxford studies in theoretical linguistics"] = "Oxf. Stud. Theor. Linguist.",
  ["oxford surveys on eukaryotic genes"] = "Oxf Surv Eukaryot Genes",
  ["oxidants and antioxidants in medical science"] = "Oxid Antioxid Med Sci",
  ["oxidation communications"] = "Oxid. Commun.",
  ["oxidation of metals"] = "Oxid. Met.",
  ["oxidative medicine and cellular longevity"] = "Oxid Med Cell Longev",
  ["oyo butsuri"] = "Oyo Butsuri",
  ["oyo yakuri"] = "Oyo Yakuri",
  ["ozone science and engineering"] = "Ozone Sci. Eng.",
  ["ozone: science & engineering"] = "Ozone Sci Eng",
  ["ozone: science and engineering"] = "Ozone: Sci. Eng.",
  ["ōu daigaku shigakushi"] = "Ou Daigaku Shigakushi",
  ["österreich in geschichte und literatur mit geographie"] = "Osterr Gesch Lit Geogr",
  ["österreichische ärztezeitung"] = "Osterr Arzteztg",
  ["österreichische dental-revue"] = "Osterr Dent Rev",
  ["österreichische hebammenzeitung"] = "Osterr Hebammenztg",
  ["österreichische krankenpflegezeitschrift"] = "Osterr Krankenpflegez",
  ["österreichische militärische zeitschrift"] = "Osterr Mil Z",
  ["österreichische osthefte"] = "Osterr Osth",
  ["österreichische pflegezeitschrift : organ des österreichischen gesundheits- und krankenpflegeverbands"] = "Osterr Pflegezeitschrift",
  ["österreichische schwesternzeitung"] = "Osterr Schwesternztg",
  ["österreichische zahnärzts-zeitung; mitteilungen der bundesfachgruppe für zahnheilkunde der österreichischen ärztekammer"] = "Osterr Zahnarzteztg",
  ["österreichische zahnprothetik"] = "Osterr Zahnprothet",
  ["österreichische zeitschrift für erforschung und bekämpfung der krebskrankheit"] = "Osterr Z Erforsch Bekampf Krebskr",
  ["österreichische zeitschrift für geschichtswissenschaften : özg"] = "Osterr Z Geschwiss",
  ["österreichische zeitschrift für kinderheilkunde und kinderfürsorge"] = "Osterr Z Kinderheilkd Kinderfuersorge",
  ["österreichische zeitschrift für onkologie. austrian journal of oncology"] = "Osterr Z Onkol",
  ["österreichische zeitschrift für pilzkunde"] = "Osterr Z Pilzkd",
  ["österreichische zeitschrift für statistik und informatik"] = "Osterr Z Stat Inform",
  ["österreichische zeitschrift für stomatologie"] = "Osterr Z Stomatol",
  ["özs, österreichische zeitschrift für soziologie"] = "OZS Osterr Z Soziol",
  ["o’zbekiston kimyo jurnali"] = "O’zb. Kim. J.",
  ["p & s quarterly"] = "P S Q",
  ["p & t : a peer-reviewed journal for formulary management"] = "P T",
  ["p-adic numbers, ultrametric analysis, and applications"] = "p-Adic Numbers Ultrametric Anal. Appl.",
  ["pa journal"] = "PA J.",
  ["paa. pattern analysis and applications"] = "PAA Pattern Anal. Appl.",
  ["pace law review"] = "Pace Law Rev",
  ["pachart history of astronomy series"] = "Pachart Hist. Astronom. Ser.",
  ["pacific affairs"] = "Pac Aff",
  ["pacific aids alert bulletin"] = "Pac AIDS Alert Bull",
  ["pacific aids alert bulletin / south pacific commission"] = "Pac AIDS Alert Bull",
  ["pacific and asian journal of energy"] = "Pacific Asian J. Energy",
  ["pacific asia inquiry : multidisciplinary perspectives"] = "Pac Asia Inq",
  ["pacific conservation biology"] = "Pac. Conserv. Biol.",
  ["pacific conservation biology : a journal devoted to conservation and land management in the pacific region"] = "Pac Conserv Biol",
  ["pacific economic bulletin"] = "Pacific Econ. Bull.",
  ["pacific economic review"] = "Pacific Econ. Rev.",
  ["pacific health dialog"] = "Pac Health Dialog",
  ["pacific health dialog : a publication of the pacific basin officers training program and the fiji school of medicine"] = "Pac Health Dialog",
  ["pacific historical review"] = "Pac Hist Rev",
  ["pacific journal of applied mathematics"] = "Pac. J. Appl. Math.",
  ["pacific journal of mathematics"] = "Pacific J. Math.",
  ["pacific journal of mathematics for industry"] = "Pac. J. Math. Ind.",
  ["pacific journal of medical sciences"] = "Pac J Med Sci",
  ["pacific journal of optimization. an international journal"] = "Pac. J. Optim.",
  ["pacific law journal (sacramento, calif.)"] = "Pac Law J",
  ["pacific medicine and surgery"] = "Pac Med Surg",
  ["pacific northwest fungi"] = "Pac Northwest Fungi",
  ["pacific northwest quarterly"] = "Pac Northwest Q",
  ["pacific philosophical quarterly"] = "Pac Philos Q",
  ["pacific reporter. second series"] = "Pac Rep Second Ser",
  ["pacific rim international journal of nursing research"] = "Pac Rim Int J Nurs Res Thail",
  ["pacific science"] = "Pac Sci",
  ["pacific science review"] = "Pac. Sci. Rev.",
  ["pacific science review a: natural science and engineering"] = "Pac. Sci. Rev. A: Nat. Sci. Eng",
  ["pacific sociological review"] = "Pac Sociol Rev",
  ["pacific studies"] = "Pac Stud",
  ["pacific symposium on biocomputing"] = "Pac. Symp. Biocomput.",
  ["pacific symposium on biocomputing. pacific symposium on biocomputing"] = "Pac Symp Biocomput",
  ["pacific viewpoint"] = "Pac Viewp",
  ["pacific-basin finance journal"] = "Pacific-Basin Finance J.",
  ["pacifica : journal of the melbourne college of divinity"] = "Pacifica",
  ["pacing and clinical electrophysiology"] = "Pacing Clin. Electrophysiol.",
  ["pacing and clinical electrophysiology : pace"] = "Pacing Clin Electrophysiol",
  ["packaging technology and science"] = "Packag. Technol. Sci.",
  ["packaging, transport, storage & security of radioactive material"] = "Packag., Transp., Storage Secur. Radioact. Mater.",
  ["pact. revue du groupe européen d’études pour les techniques physiques, chimiques et mathématiques appliquées à l’archéologie"] = "Pact",
  ["pada (han'guk haeyang hakhoe)"] = "Pada",
  ["paddy and water environment"] = "Paddy Water Environ.",
  ["padiatrie und grenzgebiete"] = "Padiatr. Grenzgeb.",
  ["padiatrie und padologie"] = "Padiatr. Padol.",
  ["padiatrie und padologie. supplementum"] = "Padiatr. Padol. Suppl.",
  ["padusa. bolletino del centro polesano di studi storici, archeologici ed etnografici, rovigo"] = "Padusa",
  ["paedagogica historica"] = "Paedagog Hist",
  ["paediatria japonica"] = "Paediatr Jpn",
  ["paediatria universitatis tokyo"] = "Paediatr Univ Tokyo",
  ["paediatric anaesthesia"] = "Paediatr Anaesth",
  ["paediatric and perinatal drug therapy"] = "Paediatr Perinat Drug Ther",
  ["paediatric and perinatal epidemiology"] = "Paediatr Perinat Epidemiol",
  ["paediatric drugs"] = "Paediatr Drugs",
  ["paediatric nursing"] = "Paediatr Nurs",
  ["paediatric respiratory reviews"] = "Paediatr Respir Rev",
  ["paediatrica indonesiana"] = "Paediatr Indones",
  ["paediatrics & child health"] = "Paediatr Child Health",
  ["paediatrics and child health"] = "Paediatr Child Health (Oxford)",
  ["paediatrics and child health canada"] = "Paediatr. Child Health Can.",
  ["paediatrics and international child health"] = "Paediatr Int Child Health",
  ["pag bulletin"] = "Pag Bull",
  ["pageoph topical volumes"] = "Pageoph Top. Vol.",
  ["pages documentaires"] = "Pages Doc",
  ["pagine di archeologia. studi e materiali"] = "PagA",
  ["pagine di storia della medicina"] = "Pagine Stor Med",
  ["pagine di storia della medicina. collana"] = "Pagine Stor Med Collana",
  ["pahlavi medical journal"] = "Pahlavi Med J",
  ["paho today : the newsletter of the pan american health organization"] = "PAHO Today",
  ["paideuma. mitteilungen zur kulturkunde"] = "Paideuma",
  ["pain and headache"] = "Pain Headache",
  ["pain and therapy"] = "Pain Ther",
  ["pain clinical updates"] = "Pain Clin Updates",
  ["pain management"] = "Pain Manag",
  ["pain management nursing"] = "Pain Manag. Nurs.",
  ["pain management nursing : official journal of the american society of pain management nurses"] = "Pain Manag Nurs",
  ["pain medicine"] = "Pain Med.",
  ["pain medicine (malden, mass.)"] = "Pain Med",
  ["pain physician"] = "Pain Physician",
  ["pain practice"] = "Pain Pract.",
  ["pain practice : the official journal of world institute of pain"] = "Pain Pract",
  ["pain reports"] = "Pain Rep",
  ["pain research & management"] = "Pain Res Manag",
  ["pain research and management"] = "Pain Res. Manage.",
  ["pain research and treatment"] = "Pain Res Treat",
  ["pain. supplement"] = "Pain Suppl",
  ["pakistan & gulf economist"] = "Pak Gulf Econ",
  ["pakistan dental review"] = "Pak. Dent. Rev.",
  ["pakistan development review"] = "Pak Dev Rev",
  ["pakistan economic and social review"] = "Pak Econ Soc Rev",
  ["pakistan journal of agricultural sciences"] = "Pak J Agric Sci",
  ["pakistan journal of analytical & environmental chemistry"] = "pak. J. Anal. Environ. Chem.",
  ["pakistan journal of analytical chemistry"] = "Pak. J. Anal. Chem.",
  ["pakistan journal of applied economics"] = "Pakistan J. Appl. Econ.",
  ["pakistan journal of biochemistry and molecular biology"] = "Pak. J. Biochem. Mol. Biol.",
  ["pakistan journal of biological sciences"] = "Pak. J. Biol. Sci.",
  ["pakistan journal of biological sciences : pjbs"] = "Pak J Biol Sci",
  ["pakistan journal of biological sciences: pjbs"] = "Pak J Biol Sci",
  ["pakistan journal of botany"] = "Pak. J. Bot.",
  ["pakistan journal of family planning"] = "Pak J Fam Plann",
  ["pakistan journal of health"] = "Pak J Health",
  ["pakistan journal of life and social sciences"] = "Pak J Life Soc Sci",
  ["pakistan journal of medical research"] = "Pak J Med Res",
  ["pakistan journal of medical sciences"] = "Pak J Med Sci",
  ["pakistan journal of nematology : an official publication of pakistan society of nematologists"] = "Pak J Nematol",
  ["pakistan journal of nutrition : pjn"] = "Pak J Nutr",
  ["pakistan journal of pharmaceutical sciences"] = "Pak J Pharm Sci",
  ["pakistan journal of pharmacology"] = "Pak. J. Pharmacol.",
  ["pakistan journal of psychology"] = "Pak J Psychol",
  ["pakistan journal of science"] = "Pak J Sci",
  ["pakistan journal of scientific and industrial research"] = "Pak J Sci Ind Res",
  ["pakistan journal of scientific research"] = "Pak. J. Sci. Res.",
  ["pakistan journal of statistics"] = "Pakistan J. Statist.",
  ["pakistan journal of statistics and operation research"] = "Pak. J. Stat. Oper. Res.",
  ["pakistan journal of surgery, gynaecology and obstetrics"] = "Pak. J. Surg. Gynaecol. Obstet.",
  ["pakistan journal of zoology"] = "Pak J Zool",
  ["pakistan nursing and health review"] = "Pak. Nurs. Health Rev.",
  ["pakistan pediatric journal"] = "Pak Pediatr J",
  ["pakistan population review"] = "Pak Popul Rev",
  ["pakistan veterinary journal"] = "Pak Vet J",
  ["paladyn : journal of behavioral robotics"] = "Paladyn",
  ["palaeobiodiversity and palaeoenvironments"] = "Palaeobiodivers. Palaeoenviron.",
  ["palaeogeography, palaeoclimatology, palaeoecology"] = "Palaeogeogr Palaeoclimatol Palaeoecol",
  ["palaeohistoria. acta et communicationes instituti bio-archaeologici universitatis groninganae"] = "Palaeohistoria",
  ["palaeontographica abteilung a palaozoologie stratigraphie"] = "Palaeontogr. Abt. A Palazool. Stratigr.",
  ["palaeontographica abteilung b palaeophytologie"] = "Palaeontogr. Abt. B Palaeophytol.",
  ["palaeontologia electronica"] = "Palaeontol. Electronica",
  ["palaeontologia electronica (online)"] = "Palaeontol Electronica",
  ["palaeontologische zeitschrift"] = "Palaentolog. Z.",
  ["palarch’s journal of vertebrate palaeontology"] = "PalArch’s J.  Vertebr. Palaeontol.",
  ["paläontologische zeitschrift"] = "Palaontol Z",
  ["paleoceanography and paleoclimatology"] = "Paleoceanogr Paleoclimatol",
  ["paleoclimate research"] = "Paleoclim. Res.",
  ["paleohispánica. revista sobre lenguas y culturas de la hispania antigua"] = "Paleohispanica",
  ["paleontologia mexicana"] = "Paleontol. Mex.",
  ["paleontological bulletin"] = "Paleontol. Bull.",
  ["paleontological contributions"] = "Paleontol. Contrib.",
  ["paleontological journal"] = "Paleontolog. J.",
  ["paleontological research"] = "Paleontol. Res.",
  ["paleontological society papers"] = "Paleontol. Soc. Pap.",
  ["paleopathology association monograph"] = "Paleopathol Assoc Monogr",
  ["paleopathology newsletter"] = "Paleopathol Newsl",
  ["palestine exploration quarterly"] = "Palest Explor Q",
  ["palestine journal of mathematics"] = "Palest. J. Math.",
  ["palgrave communications"] = "Palgrave Commun",
  ["palgrave philosophy today"] = "Palgrave Philos. Today",
  ["palgrave pivots in sports economics"] = "Palgrave Pivots Sports Econ.",
  ["palgrave studies in educational philosophy and theory"] = "Palgrave Stud. Educ. Philos. Theory",
  ["palgrave studies in literature, science and medicine"] = "Palgrave Stud. Lit. Sci. Med.",
  ["palgrave studies in sustainability, environment and macroeconomics"] = "Palgrave Stud. Sustain. Environ. Macroecon.",
  ["palgrave studies in the history of science and technology"] = "Palgrave Stud. Hist. Sci. Technol.",
  ["palgrave texts in econometrics"] = "Palgrave Texts Econom.",
  ["palladio. rivista di storia dell’architettura"] = "Palladio",
  ["pallas. revue d’études antiques"] = "Pallas",
  ["palliative & supportive care"] = "Palliat Support Care",
  ["palliative and supportive care"] = "Palliat. Support. Care",
  ["palliative care"] = "Palliat Care",
  ["palliative care and social practice"] = "Palliat Care Soc Pract",
  ["palliative medicine"] = "Palliat Med",
  ["palliative medicine & care : open access"] = "Palliat Med Care",
  ["palliative medicine reports"] = "Palliat Med Rep",
  ["palmet. sadberk hanım müzesi yıllığı"] = "Palmet",
  ["památky archeologické"] = "PamA",
  ["pamm. proceedings in applied mathematics and mechanics"] = "PAMM. Proc. Appl. Math. Mech.",
  ["pamâtniki kul’tury (monuments of culture: new discoveries)"] = "PK",
  ["pan (nigeria)"] = "PAN",
  ["pan american health"] = "Pan Am Health",
  ["pan american medical woman's journal"] = "Pan Am Med Womans J",
  ["pan american medical woman’s journal"] = "Pan Am Med Womans J",
  ["pan-african journal"] = "Pan Afr J",
  ["pan-american journal of aquatic sciences"] = "Panam J Aquat Sci",
  ["pan-pacific entomologist"] = "Pan-Pac. Entomol.",
  ["pan. studi dell’istituto di filologia latina"] = "Pan",
  ["panamerican mathematical journal"] = "PanAmer. Math. J.",
  ["pancreas (fairfax, va.)"] = "Pancreas (Fairfax)",
  ["pancreatic disorders & therapy"] = "Pancreat Disord Ther",
  ["panhandle-plains historical review"] = "Panhand-Plains Hist Rev",
  ["panminerva medica"] = "Panminerva Med",
  ["panoramas et syntheses"] = "Panor. Syntheses",
  ["panoramas et synthèses"] = "Panor. Synthèses",
  ["panstwo i prawo : organ zrzeszenia prawnikow demokratow w polsce"] = "Panstwo Prawo",
  ["państwo i prawo : organ zrzeszenia prawników demokratów w polsce"] = "Panstwo Prawo",
  ["papeles bilbilitanos"] = "PapBilb",
  ["papeles de economía española"] = "Pap. Econ. Española",
  ["papeles de poblacion / centro de investigacion y estudios avanzados de la poblacion, universidad autonoma del estado de mexico"] = "Papeles Poblac",
  ["papeles de población"] = "Papeles Poblac",
  ["paper - geological survey of canada"] = "Pap. - Geol. Surv. Can.",
  ["paper series (united hospital fund of new york)"] = "Pap Ser United Hosp Fund N Y",
  ["paper series, united hospital fund of new york"] = "Pap. Ser. United Hosp. Fund N. Y.",
  ["paperi ja puu"] = "Pap. Puu",
  ["papers / national conference for professional nurses and physicians"] = "Pap Natl Conf Prof Nurses Physicians",
  ["papers / regional science association. regional science association. meeting"] = "Pap Reg Sci Assoc",
  ["papers and proceedings. royal society of tasmania"] = "Pap Proc",
  ["papers in applied geography"] = "Pap. Appl. Geogr.",
  ["papers in meteorology and geophysics"] = "Pap. Meteorol. Geophys.",
  ["papers in palaeontology"] = "Pap Palaeontol",
  ["papers in physical oceanography and meteorology"] = "Pap. Phys. Oceanogr. Meteorol.",
  ["papers in regional science"] = "Pap. Reg. Sci.",
  ["papers in regional science : the journal of the regional science association international"] = "Pap Reg Sci",
  ["papers in slovene studies"] = "Pap Slov Stud",
  ["papers of the bibliographical society of canada. cahiers de la societe bibliographique du canada. bibliographical society of canada"] = "Pap Bibliogr Soc Can",
  ["papers of the bibliographical society of canada. cahiers de la société bibliographique du canada. bibliographical society of canada"] = "Pap Bibliogr Soc Can",
  ["papers of the british school at rome"] = "Pap Br Sch Rome",
  ["papers of the laboratory of tree ring research"] = "Pap. Lab. Tree Ring Res.",
  ["papers of the michigan academy of science, arts and letters"] = "Pap. Mich. Acad. Sci. Arts Lett.",
  ["papers on far eastern history"] = "Pap Far East Hist",
  ["papers on french seventeenth century literature"] = "Pap Fr Seventeenth Century Lit",
  ["papers on non-market decision making"] = "Pap. Non-Market Dec. Making",
  ["papers presented to the workshop on the evolution of the martian atmosphere : kona, hawaii, june 29-july 1, 1992"] = "Pap Present Workshop Evol Martian Atmos",
  ["papers, national conference for professional nurses and physicians"] = "Pap. Natl. Conf. Prof. Nurses Physicians",
  ["papers. bibliographical society of america"] = "Pap Bibliogr Soc Am",
  ["papers. kroeber anthropological society"] = "Pap Kroeber Anthropol Soc",
  ["papers. national conference for professional nurses and physicians"] = "Pap Natl Conf Prof Nurses Physicians",
  ["papers. regional science association. meeting"] = "Pap Reg Sci Assoc",
  ["papéis avulsos de zoologia"] = "Pap Avulsos Zool",
  ["papillomavirus research (amsterdam, netherlands)"] = "Papillomavirus Res",
  ["papir a celuloza"] = "Pap. Celul.",
  ["papua and new guinea medical journal"] = "P N G Med J",
  ["papua new guinea medical journal"] = "P. N. G. Med. J.",
  ["papyri. bollettino del museo del papiro"] = "Papyri",
  ["papyrologica lupiensia"] = "PLup",
  ["paracelsus; archiv der praktischen medizin"] = "Paracelsus Arch Prakt Med",
  ["parade (new york, n.y.)"] = "Parade",
  ["paragraph (cambridge, england)"] = "Paragraph (Mod Crit Theory Group)",
  ["parallel algorithms and applications"] = "Parallel Algorithms Appl.",
  ["parallel and distributed computing series"] = "Parallel Distrib. Comput. Ser.",
  ["parallel computing"] = "Parallel Comput.",
  ["parallel computing. systems & applications"] = "Parallel Comput.",
  ["parallel processing letters"] = "Parallel Process. Lett.",
  ["parallel robots: theory and applications"] = "Parallel Robots: Theory Appl.",
  ["paramedics international"] = "Paramed Int",
  ["parameters : journal of the us army war college"] = "Parameters",
  ["parasite (paris, france)"] = "Parasite",
  ["parasite epidemiology and control"] = "Parasite Epidemiol Control",
  ["parasite immunology"] = "Parasite Immunol",
  ["parasites & vectors"] = "Parasit Vectors",
  ["parasites and vectors"] = "Parasites Vectors",
  ["parasitology international"] = "Parasitol Int",
  ["parasitology open"] = "Parasitol Open",
  ["parasitology research"] = "Parasitol Res",
  ["parasitology today"] = "Parasitol. Today",
  ["parasitology today (personal ed.)"] = "Parasitol Today",
  ["parasitología al día"] = "Parasitol. día",
  ["parasitología latinoamericana"] = "Parasitol. latinoam.",
  ["parenting, science and practice"] = "Parent Sci Pract",
  ["parents magazine (new york, n.y. : 1985)"] = "Parents Mag (NY)",
  ["parents' magazine (bergenfield, n.j.)"] = "Parents Mag",
  ["parents’ magazine (bergenfield, n.j.)"] = "Parents Mag",
  ["paris medical"] = "Paris Med",
  ["paris médical"] = "Paris Med",
  ["parisankhyan samikkha"] = "Parisankhyan Samikkha",
  ["parivar ayojan"] = "Parivar Ayojan",
  ["park science"] = "Park Sci",
  ["parkinson's disease"] = "Parkinsons Dis",
  ["parkinsonism & related disorders"] = "Parkinsonism Relat Disord",
  ["parliamentary affairs"] = "Parliam Aff",
  ["parma medica"] = "Parma Med",
  ["parnassiana archives"] = "Parnass Arch",
  ["parodontologia e stomatologia (nuova)"] = "Parodontol. Stomatol. (Nuova)",
  ["parodontologia e stomatologia (nuova) : organo ufficiale dell'arpa italiana, della società italiana jonoforesi stomatologica e della accademia ligustica di stomatologia"] = "Parodontol Stomatol (Nuova)",
  ["parodontologia e stomatologia (nuova) : organo ufficiale dell’arpa italiana, della societa italiana jonoforesi stomatologica e della accademia ligustica di stomatologia"] = "Parodontol Stomatol (Nuova)",
  ["parodontologie (berlin, germany)"] = "Parodontol",
  ["parodontologie (zurich, switzerland : 1954)"] = "Parodontologie",
  ["parodontologie and academy review"] = "Parodontol Acad Rev",
  ["paroi arterielle"] = "Paroi Arterielle",
  ["paroi artérielle"] = "Paroi Arterielle",
  ["part a news : independent news to legally maximize medicare part a dollars"] = "Part A News",
  ["part b news"] = "Part B News",
  ["partenope. rivista trimestrale di cultura napoletana"] = "Partenope",
  ["parthica. incontri di culture nel mondo antico"] = "Parthica",
  ["partial differential equations and applications"] = "Partial Differ. Equations Appl.",
  ["partial differential equations and measure theory"] = "Partial Differ. Equ. Meas. Theory",
  ["partial differential equations in applied mathematics"] = "Partial Differ. Equations Appl. Math.",
  ["participant journal"] = "Part J",
  ["particle & particle systems characterization : measurement and description of particle properties and behavior in powders and other disperse systems"] = "Part Part Syst Charact",
  ["particle and fibre toxicology"] = "Part Fibre Toxicol",
  ["particle and particle systems characterization"] = "Part. Part. Syst. Char.",
  ["particle technology series"] = "Part. Technol. Ser.",
  ["particles and fields subseries"] = "Particles and Fields Subser.",
  ["particulate science and technology"] = "Part. Sci. Technol.",
  ["partisan review (new york, n.y. : 1936)"] = "Partis Rev",
  ["partner abuse"] = "Partner Abuse",
  ["parttorteneti kozlemenyek. magyar szocialista munkaspart. kozponti bizottsag. parttorteneti intezet"] = "Parttort Kozl",
  ["party politics"] = "Party Politics",
  ["pas reporter"] = "PAS Rep",
  ["passato e presente (florence, italy)"] = "Passato Presente",
  ["passauer schriften zur psychologiegeschichte"] = "Passau Schr Psychologiegesch",
  ["passauer schriften zur psychologiegeschichte / herausgegeben vom institut fur geschichte der neueren psychologie der universitat passau"] = "Passau Schr Psychologiegesch",
  ["past & present"] = "Past Present",
  ["past and present: a journal of historical studies"] = "P&P",
  ["past global changes magazine"] = "Past Global Changes Mag.",
  ["past imperfect (edmonton, alta.)"] = "Past Imperfect",
  ["pastoral care in education"] = "Pastor Care Educ",
  ["pastoral psychology"] = "Pastoral Psychol",
  ["pastoralism : research, policy and practice"] = "Pastoralism",
  ["patch-based techniques in medical imaging : 4th international workshop, patch-mi 2018, held in conjunction with miccai 2018, granada, spain, september 20, 2018 : proceedings. patch-mi (workshop) (4th : 2018 : granada, spain)"] = "Patch Based Tech Med Imaging (2018)",
  ["patch-based techniques in medical imaging : first international workshop, patch-mi 2015, held in conjunction with miccai 2015, munich, germany, october 9, 2015, revised selected papers. patch-mi (workshop) (1st : 2015 : munich, germany)"] = "Patch Based Tech Med Imaging (2015)",
  ["patch-based techniques in medical imaging : second international workshop, patch-mi 2016, held in conjunction with miccai 2016, athens, greece, october 17, 2016 : proceedings. patch-mi (workshop) (2nd : 2016 : athens, greece)"] = "Patch Based Tech Med Imaging (2016)",
  ["patch-based techniques in medical imaging : third international workshop, patch-mi 2017, held in conjunction with miccai 2017, quebec city, qc, canada, september 14, 2017, proceedings. patch-mi (workshop) (3rd : 2017 : québec, québec)"] = "Patch Based Tech Med Imaging (2017)",
  ["pathobiology : journal of immunopathology, molecular and cellular biology"] = "Pathobiology",
  ["pathobiology annual"] = "Pathobiol Annu",
  ["pathobiology of aging & age related diseases"] = "Pathobiol Aging Age Relat Dis",
  ["pathobiology of aging & age-related diseases"] = "Pathobiol. Aging Age-Relat. Dis.",
  ["pathogenesis : the journal of mechanisms in disease processes"] = "Pathogenesis (Amst)",
  ["pathogens & immunity"] = "Pathog Immun",
  ["pathogens (basel, switzerland)"] = "Pathogens",
  ["pathogens and disease"] = "Pathog Dis",
  ["pathogens and global health"] = "Pathog Glob Health",
  ["pathologia et microbiologia"] = "Pathol Microbiol (Basel)",
  ["pathologia europaea"] = "Pathol Eur",
  ["pathologia veterinaria"] = "Pathol Vet",
  ["pathologie biologie"] = "Pathol Biol (Paris)",
  ["pathologie et biologie"] = "Pathol Biol",
  ["pathology & oncology research"] = "Pathol. Oncol. Res.",
  ["pathology (philadelphia, pa.)"] = "Pathology (Phila)",
  ["pathology and immunopathology research"] = "Pathol Immunopathol Res",
  ["pathology and laboratory medicine international"] = "Pathol Lab Med Int",
  ["pathology and oncology research"] = "Pathol. Oncol. Res.",
  ["pathology annual"] = "Pathol Annu",
  ["pathology case reviews"] = "Pathol Case Rev",
  ["pathology discovery"] = "Pathol Discov",
  ["pathology international"] = "Pathol Int",
  ["pathology oncology research"] = "Pathol. Oncol. Res.",
  ["pathology oncology research : por"] = "Pathol Oncol Res",
  ["pathology research international"] = "Patholog Res Int",
  ["pathology, research and practice"] = "Pathol Res Pract",
  ["pathophysiology : the official journal of the international society for pathophysiology"] = "Pathophysiology",
  ["pathophysiology : the official journal of the international society for pathophysiology / isp"] = "Pathophysiology",
  ["pathophysiology of haemostasis and thrombosis"] = "Pathophysiol Haemost Thromb",
  ["pathpapers : a series of occasional papers on innovative projects supported by the pathfinder fund"] = "Pathpapers",
  ["pathways (chestnut hill, mass.)"] = "Pathways",
  ["pathways in mathematics"] = "Pathw. Math.",
  ["patient accounts"] = "Patient Acc",
  ["patient care"] = "Patient Care",
  ["patient care management"] = "Patient Care Manag",
  ["patient counselling and health education"] = "Patient Couns Health Educ",
  ["patient education and counseling"] = "Patient Educ Couns",
  ["patient education newsletter"] = "Patient Educ Newsl",
  ["patient experience journal"] = "Patient Exp J",
  ["patient intelligence"] = "Patient Intell",
  ["patient preference and adherence"] = "Patient Prefer Adherence",
  ["patient related outcome measures"] = "Patient Relat Outcome Meas",
  ["patient safety in surgery"] = "Patient Saf Surg",
  ["patient-focused care"] = "Patient. Focus. Care",
  ["patient-focused care : the health care executive's guide to organizational restructuring"] = "Patient Focus Care",
  ["patient-focused care : the health care executive’s guide to organizational restructuring"] = "Patient Focus Care",
  ["patient-focused care and satisfaction"] = "Patient Focus Care Satisf",
  ["patient-focused care and satisfaction / american health consultants"] = "Patient Focus Care Satisf",
  ["patma-banasirakan handes"] = "Patma Banasirakan Handes",
  ["patma-banasirakan handes: revue historico-philologique"] = "PBH",
  ["patologia e clinica ostetrica e ginecologica"] = "Patol Clin Ostet Ginecol",
  ["patologia polska"] = "Patol Pol",
  ["patologicheskaia fiziologiia i eksperimentalnaia terapiia"] = "Patol. Fiziol. Eksp. Ter.",
  ["patologicheskaia fiziologiia i eksperimental’naia terapiia"] = "Patol Fiziol Eksp Ter",
  ["patologicheskaia fiziologiia i èksperimental'naia terapiia"] = "Patol Fiziol Eksp Ter",
  ["pattern analysis and applications"] = "Pattern Anal. Appl.",
  ["pattern analysis and applications : paa"] = "Pattern Anal Appl",
  ["pattern recognition"] = "Pattern Recognit",
  ["pattern recognition and image analysis"] = "Pattern Recognit Image Anal.",
  ["pattern recognition in physics"] = "Pattern Recognit. Phys.",
  ["pattern recognition letters"] = "Pattern Recognit. Lett.",
  ["patterns (new york, n.y.)"] = "Patterns (N Y)",
  ["patterns of prejudice"] = "Patterns Prejudice",
  ["paulys realencyclopädie der classischen altertumswissenschaft"] = "RE",
  ["pavlov journal of higher nervous activity"] = "Pavlov J High Nerv Act",
  ["pavlovian journal of biological science"] = "Pavlov. J. Biol. Sci.",
  ["pawlow-zeitschrift fur hohere nerventatigkeit. deutsche ausg"] = "Pawlow Z Hohere Nerventatigkeit",
  ["pawlow-zeitschrift für höhere nerventätigkeit. deutsche ausg"] = "Pawlow Z Hohere Nerventatigkeit",
  ["párttörténeti közlemények. magyar szocialista munkáspárt. központi bizottság. párttörténeti intézet"] = "Parttort Kozl",
  ["pāsonariti kenkyū"] = "Pasonariti Kenkyu",
  ["pädiatrie und grenzgebiete"] = "Padiatr Grenzgeb",
  ["pädiatrie und pädologie"] = "Padiatr Padol",
  ["pädiatrie und pädologie. supplementum"] = "Padiatr Padol Suppl",
  ["pci journal"] = "PCI J.",
  ["pcr methods and applications"] = "PCR Methods Appl",
  ["pcr reporter"] = "PCR Rep.",
  ["pda journal of pharmaceutical science and technology"] = "PDA J Pharm Sci Technol",
  ["pda journal of pharmaceutical science and technology / pda"] = "PDA J Pharm Sci Technol",
  ["pdm: physicians' drug manual"] = "PDM",
  ["pdm: physicians’ drug manual"] = "PDM",
  ["pdm; physicians drug manual"] = "PDM.",
  ["pdpta '19 : proceedings of the 2019 international conference on parallel & distributed processing techniquess & applications. international conference on parallel and distributed processing techniques and applications (2019 : las vegas, nev.)"] = "PDPTA 19 (2019)",
  ["peace and conflict : journal of peace psychology : the journal of the division of peace psychology of the american psychological association"] = "Peace Confl",
  ["peace research society international papers"] = "Peace Res. Society Int. Pap.",
  ["peace science society (international) papers"] = "Peace Sci. Society",
  ["peasant studies"] = "Peasant Stud",
  ["pedagogy in health promotion"] = "Pedagogy Health Promot",
  ["pediatria de las americas"] = "Pediatr Am",
  ["pediatria de las américas"] = "Pediatr Am",
  ["pediatria internazionale"] = "Pediatr Int (Roma)",
  ["pediatria medica e chirurgica"] = "Pediatr. Med. Chir.",
  ["pediatria panamericana"] = "Pediatr Panam",
  ["pediatria polska"] = "Pediatr Pol",
  ["pediatria pratica"] = "Pediatr. Prat.",
  ["pediatria prática"] = "Pediatr Prat",
  ["pediatric aids and hiv infection"] = "Pediatr AIDS HIV Infect",
  ["pediatric alert"] = "Pediatr Alert",
  ["pediatric allergy and immunology"] = "Pediatr. Allergy Immunol.",
  ["pediatric allergy and immunology : official publication of the european society of pediatric allergy and immunology"] = "Pediatr Allergy Immunol",
  ["pediatric allergy, immunology, and pulmonology"] = "Pediatr Allergy Immunol Pulmonol",
  ["pediatric and adolescent gynecology : official journal of the international federation of infantile and juvenile gynecology"] = "Pediatr Adolesc Gynecol",
  ["pediatric and developmental pathology"] = "Pediatr. Dev. Pathol.",
  ["pediatric and developmental pathology : the official journal of the society for pediatric pathology and the paediatric pathology society"] = "Pediatr Dev Pathol",
  ["pediatric annals"] = "Pediatr Ann",
  ["pediatric asthma, allergy & immunology"] = "Pediatr Asthma Allergy Immunol",
  ["pediatric blood & cancer"] = "Pediatr Blood Cancer",
  ["pediatric blood and cancer"] = "Pediatr. Blood Cancer",
  ["pediatric cardiology"] = "Pediatr Cardiol",
  ["pediatric care (wilmington, del.)"] = "Pediatr Care (Wilmington)",
  ["pediatric case reviews"] = "Pediatr. Case Rev.",
  ["pediatric case reviews (print)"] = "Pediatr Case Rev",
  ["pediatric clinics of india"] = "Pediatr Clin India",
  ["pediatric clinics of north america"] = "Pediatr Clin North Am",
  ["pediatric critical care medicine : a journal of the society of critical care medicine and the world federation of pediatric intensive and critical care societies"] = "Pediatr Crit Care Med",
  ["pediatric dental journal : international journal of japanese society of pediatric dentistry"] = "Pediatr Dent J",
  ["pediatric dentistry"] = "Pediatr Dent",
  ["pediatric dermatology"] = "Pediatr Dermatol",
  ["pediatric diabetes"] = "Pediatr Diabetes",
  ["pediatric dimensions"] = "Pediatr Dimens",
  ["pediatric drugs"] = "Pediatr. Drugs",
  ["pediatric emergency care"] = "Pediatr Emerg Care",
  ["pediatric emergency medicine practice"] = "Pediatr Emerg Med Pract",
  ["pediatric endocrinology reviews"] = "Pediatr. Endocrinol. Rev.",
  ["pediatric endocrinology reviews : per"] = "Pediatr Endocrinol Rev",
  ["pediatric endocrinology, diabetes, and metabolism"] = "Pediatr Endocrinol Diabetes Metab",
  ["pediatric exercise science"] = "Pediatr Exerc Sci",
  ["pediatric gastroenterology, hepatology & nutrition"] = "Pediatr Gastroenterol Hepatol Nutr",
  ["pediatric health"] = "Ped Health",
  ["pediatric health, medicine and therapeutics"] = "Pediatric Health Med Ther",
  ["pediatric hematology and oncology"] = "Pediatr Hematol Oncol",
  ["pediatric infectious disease"] = "Pediatr Infect Dis",
  ["pediatric infectious disease journal"] = "Pediatr. Infect. Dis. J.",
  ["pediatric investigation"] = "Pediatr Investig",
  ["pediatric medicine (hong kong, china)"] = "Pediatr Med",
  ["pediatric nephrology"] = "Pediatr. Nephrol.",
  ["pediatric nephrology (berlin, germany)"] = "Pediatr Nephrol",
  ["pediatric neurology"] = "Pediatr Neurol",
  ["pediatric neurology briefs"] = "Pediatr Neurol Briefs",
  ["pediatric neuroscience"] = "Pediatr Neurosci",
  ["pediatric neurosurgery"] = "Pediatr Neurosurg",
  ["pediatric news"] = "Pediatr News",
  ["pediatric nursing"] = "Pediatr Nurs",
  ["pediatric obesity"] = "Pediatr Obes",
  ["pediatric oncall"] = "Pediatr Oncall",
  ["pediatric pathology"] = "Pediatr Pathol",
  ["pediatric pathology & laboratory medicine : journal of the society for pediatric pathology, affiliated with the international paediatric pathology association"] = "Pediatr Pathol Lab Med",
  ["pediatric pathology & molecular medicine"] = "Pediatr Pathol Mol Med",
  ["pediatric pathology / affiliated with the international paediatric pathology association"] = "Pediatr Pathol",
  ["pediatric pathology and laboratory medicine"] = "Pediatr. Pathol. Lab. Med.",
  ["pediatric pathology and molecular medicine"] = "Pediatr. Pathol. Mol. Med.",
  ["pediatric pharmacology"] = "Pediatr. Pharmacol. (New York)",
  ["pediatric pharmacology (new york, n.y.)"] = "Pediatr Pharmacol (New York)",
  ["pediatric physical therapy : the official publication of the section on pediatrics of the american physical therapy association"] = "Pediatr Phys Ther",
  ["pediatric pulmonology"] = "Pediatr Pulmonol",
  ["pediatric pulmonology. supplement"] = "Pediatr Pulmonol Suppl",
  ["pediatric quality & safety"] = "Pediatr Qual Saf",
  ["pediatric radiology"] = "Pediatr Radiol",
  ["pediatric rehabilitation"] = "Pediatr Rehabil",
  ["pediatric reports"] = "Pediatr Rep",
  ["pediatric research"] = "Pediatr Res",
  ["pediatric rheumatology online journal"] = "Pediatr Rheumatol Online J",
  ["pediatric surgery international"] = "Pediatr Surg Int",
  ["pediatric transplantation"] = "Pediatr Transplant",
  ["pediatricke listy"] = "Pediatr Listy",
  ["pediatrické listy"] = "Pediatr Listy",
  ["pediatrics & therapeutics : current research"] = "Pediatr Ther",
  ["pediatrics and neonatal nursing : open journal"] = "Pediatr Neonatal Nurs",
  ["pediatrics and neonatology"] = "Pediatr Neonatol",
  ["pediatrics in review"] = "Pediatr Rev",
  ["pediatrics in review / american academy of pediatrics"] = "Pediatr Rev",
  ["pediatrics international"] = "Pediatr. Int.",
  ["pediatrics international : official journal of the japan pediatric society"] = "Pediatr Int",
  ["pediatrie (bucharest, romania)"] = "Pediatrie (Bucur)",
  ["pediatriia akusherstvo i ginekologiia"] = "Pediatr Akus Ginekol",
  ["pediatriia, akusherstvo i ginekologiia"] = "Pediatr. Akush. Ginekol.",
  ["pediatrii︠a︡ akusherstvo i ginekologii︠a︡"] = "Pediatr Akus Ginekol",
  ["pediatrik cerrahi dergisi"] = "Pediatr Cerrahi Derg",
  ["pediatría panamericana"] = "Pediatr Panam",
  ["pedodontie francaise"] = "Pedod Fr",
  ["peer review : emerging trends and key debates in undergraduate education"] = "Peer Rev",
  ["peer-to-peer networking and applications"] = "Peer Peer Netw Appl",
  ["peerj analytical chemistry"] = "PeerJ Anal. Chem.",
  ["peerj computer science"] = "PeerJ Comput. Sci.",
  ["peerj inorganic chemistry"] = "PeerJ Inorg. Chem.",
  ["peerj life and environment"] = "PeerJ Life Environ.",
  ["peerj materials science"] = "PeerJ Mater. Sci.",
  ["peerj open advances in ecology"] = "PeerJ Open Adv. Ecol.",
  ["peerj open advances in marine biology"] = "PeerJ Open Adv. Mar. Biol.",
  ["peerj open advances in plant science"] = "PeerJ Open Adv. Plant Sci.",
  ["peerj open advances in zoology"] = "PeerJ Open Adv. Zool.",
  ["peerj organic chemistry"] = "PeerJ Org. Chem.",
  ["peerj physical chemistry"] = "PeerJ Phys Chem",
  ["peerj preprints"] = "PeerJ Prepr",
  ["peerj. computer science"] = "PeerJ Comput Sci",
  ["pegasus. berliner beiträge zum nachleben der antike"] = "Pegasus",
  ["peking mathematical journal"] = "Peking Math. J.",
  ["peking university series in mathematics"] = "Peking Univ. Ser. Math.",
  ["pelican news"] = "Pelican News",
  ["pembroke persian papers"] = "Pemb. Pers. Pap.",
  ["pendik veteriner mikrobiyoloji dergisi"] = "Pendik Vet Mikrobiyol Derg",
  ["penelope (paris, france)"] = "Penelope",
  ["penn bioethics journal"] = "Penn Bioeth J",
  ["penn dental journal"] = "Penn Dent. J. (Phila.)",
  ["penn gse perspectives on urban education"] = "Penn GSE Perspect Urban Educ",
  ["penn medicine"] = "Penn Med",
  ["penn medicine / university of pennsylvania medical center"] = "Penn Med",
  ["pennsylvania archaeologist"] = "Pa Archaeol",
  ["pennsylvania bar association quarterly. pennsylvania bar association"] = "PA Bar Assoc Q",
  ["pennsylvania dental journal"] = "Pa Dent J (Harrisb)",
  ["pennsylvania district and county reports"] = "Pa Dist Cty Rep",
  ["pennsylvania folklife"] = "Pa Folklife",
  ["pennsylvania health & you"] = "Pa Health You",
  ["pennsylvania health and you"] = "Pa. Health You",
  ["pennsylvania heritage (1974)"] = "Pa Herit",
  ["pennsylvania history"] = "Pa Hist",
  ["pennsylvania medical journal"] = "Pa. Med. J.",
  ["pennsylvania medical journal (1897)"] = "Pa Med J",
  ["pennsylvania medical journal (1928)"] = "Pa Med J",
  ["pennsylvania medicine"] = "Pa Med",
  ["pennsylvania mennonite heritage"] = "Pa Mennon Herit",
  ["pennsylvania nurse"] = "Pa. Nurse",
  ["pennsylvania's health : official journal of the pennsylvania state dept. of health"] = "Pa Health",
  ["pensamiento educativo : revista de investigación educacional latinoamericana"] = "Pensam Educ",
  ["pensamiento iberoamericano: revista de economía política"] = "Pensa. Iberamer.: Revista Econ. Política",
  ["pensiero medico"] = "Pensiero Med (Milano)",
  ["people & the planet"] = "People Planet",
  ["people & the planet / ippf, unfpa, iucn"] = "People Planet",
  ["people and nature"] = "People Nat.",
  ["people and nature (hoboken, n.j.)"] = "People Nat (Hoboken)",
  ["people and place"] = "People Place",
  ["people count"] = "People Count",
  ["people weekly"] = "People (Chicago)",
  ["pepperdine law review"] = "Pepperdine Law Rev",
  ["peptide and protein reviews"] = "Peptide Protein Rev",
  ["peptide research"] = "Pept Res",
  ["peptide science"] = "Pept. Sci.",
  ["peptide science (hoboken, n.j.)"] = "Pept Sci (Hoboken)",
  ["peptides (amsterdam, netherlands)"] = "Peptides (Amsterdam, Neth.)",
  ["peptides (new york, ny, united states)"] = "Peptides (N. Y., NY, U. S.)",
  ["perception & psychophysics"] = "Percept Psychophys",
  ["perception and psychophysics"] = "Percept. Psychophys.",
  ["perceptual and motor skills"] = "Percept Mot Skills",
  ["perfiles latinoamericanos : revista de la sede academica de mexico de la facultad latinoamericana de ciencias sociales"] = "Perf Latinoam",
  ["perfiles latinoamericanos : revista de la sede académica de méxico de la facultad latinoamericana de ciencias sociales"] = "Perf Latinoam",
  ["performance enhancement & health"] = "Perform Enhanc Health",
  ["performance evaluation"] = "Perform. Eval.",
  ["performance improvement"] = "Perform. Improv.",
  ["performance improvement advisor"] = "Perform Improv Advis",
  ["performance research"] = "Perform Res",
  ["pergamenische forschungen"] = "PF",
  ["pergamon studies in neuroscience"] = "Pergamon Stud. Neurosci.",
  ["perinatal care"] = "Perinat Care",
  ["perinatologia y reproduccion humana / inper"] = "Perinatol Reprod Hum",
  ["perinatología y reproducción humana"] = "Perinatol Reprod Hum",
  ["periodica mathematica hungarica"] = "Period Math Hung",
  ["periodica mathematica hungarica. journal of the janos bolyai mathematical society"] = "Period. Math. Hungar.",
  ["periodica polytechnica"] = "Period. Polytech. Mech. Eng.",
  ["periodica polytechnica, chemical engineering"] = "Period. Polytech., Chem. Eng.",
  ["periodica polytechnica, civil engineering"] = "Period. Polytech., Civ. Eng.",
  ["periodica polytechnica, engineering"] = "Period. Polytech., Eng.",
  ["periodica polytechnica, mechanical engineering"] = "Period. Polytech., Mech. Eng.",
  ["periodical (council on america's military past)"] = "Period Counc Am Mil Past",
  ["periodical (council on america’s military past)"] = "Period Counc Am Mil Past",
  ["periodicum biologorum"] = "Period Biol",
  ["periodontal abstracts"] = "Periodontal Abstr",
  ["periodontal abstracts / journal of the western society of periodontology"] = "Periodontal Abstr.",
  ["periodontal case reports"] = "Periodontal Case Rep.",
  ["periodontal case reports : a publication of the northeastern society of periodontists"] = "Periodontal Case Rep",
  ["periodontal clinical investigations"] = "Periodontal Clin. Investig.",
  ["periodontal clinical investigations : official publication of the northeastern society of periodontists"] = "Periodontal Clin Investig",
  ["periodontology 2000"] = "Periodontol 2000",
  ["perioperative care and operating room management"] = "Perioper Care Oper Room Manag",
  ["perioperative medicine (london, england)"] = "Perioper Med (Lond)",
  ["perioperative nursing quarterly"] = "Perioper Nurs Q",
  ["periplus : jahrbuch für aussereuropäische geschichte"] = "Periplus",
  ["peristil. zbornik radova za povijest umjetnosti"] = "Peristil",
  ["peritoneal dialysis international"] = "Perit. Dial. Int.",
  ["peritoneal dialysis international (downsview, ontario"] = "Perit Dial Int",
  ["peritoneal dialysis international : journal of the international society for peritoneal dialysis"] = "Perit Dial Int",
  ["perkins journal"] = "Perkins J",
  ["permafrost and periglacial processes"] = "Permafr Periglac Process",
  ["permanente foundation medical bulletin"] = "Perm Found Med Bull",
  ["permian historical annual"] = "Permian Hist Annu",
  ["persica. jaarboek van het genootschap nederland-iran, stichting voor culturele betrekkingen"] = "Persica",
  ["persona y bioética"] = "Pers Bioet",
  ["persona y derecho"] = "Pers Derecho",
  ["personal and ubiquitous computing"] = "Pers Ubiquitous Comput",
  ["personal relationships"] = "Pers Relatsh",
  ["personality & social psychology bulletin"] = "Pers Soc Psychol Bull",
  ["personality and individual differences"] = "Pers Individ Dif",
  ["personality and mental health"] = "Personal Ment Health",
  ["personality and social psychology bulletin"] = "Pers Soc Psychol Bull",
  ["personality and social psychology review : an official journal of the society for personality and social psychology, inc"] = "Pers Soc Psychol Rev",
  ["personality disorders"] = "Personal Disord",
  ["personality neuroscience"] = "Personal Neurosci",
  ["personalized medicine"] = "Per Med",
  ["personalized medicine in oncology"] = "Pers Med Oncol",
  ["personalized medicine in psychiatry"] = "Pers Med Psychiatry",
  ["personnel administrator"] = "Pers. Adm.",
  ["personnel journal"] = "Pers. J.",
  ["personnel psychology"] = "Pers Psychol",
  ["perspectivas bioéticas en las américas"] = "Perspect Bioet Am",
  ["perspectivas en nutrición humana : órgano de divulgación academica de la escuela de nutrición y dietética de la universidad de antioquia"] = "Perspect Nutr Hum",
  ["perspectivas internacionales en planificacion familiar"] = "Perspect Int Planif Fam",
  ["perspectivas internacionales en planificación familiar"] = "Perspect Int Planif Fam",
  ["perspective infirmiere : revue officielle de l’ordre des infirmieres et infirmiers du quebec"] = "Perspect Infirm",
  ["perspective infirmière : revue officielle de l'ordre des infirmières et infirmiers du québec"] = "Perspect Infirm",
  ["perspectives (american probation and parole association)"] = "Perspect (Am Probat Parole Assoc)",
  ["perspectives (gerontological nursing association (canada))"] = "Perspectives",
  ["perspectives (washington, d.c. : 1984)"] = "Perspectives (Wash)",
  ["perspectives in american history"] = "Perspect Am Hist",
  ["perspectives in biology and medicine"] = "Perspect Biol Med",
  ["perspectives in biosecurity"] = "Perspect Biosecur",
  ["perspectives in clinical research"] = "Perspect Clin Res",
  ["perspectives in drug discovery and design"] = "Perspect. Drug Discovery Des.",
  ["perspectives in drug discovery and design : pd3"] = "Perspect Drug Discov Des",
  ["perspectives in ecology and conservation"] = "Perspect Ecol Conserv",
  ["perspectives in formal induction, revision and evolution"] = "Perspect. Form. Induction Revis. Evol.",
  ["perspectives in health : the magazine of the pan american health organization"] = "Perspect Health",
  ["perspectives in health information management"] = "Perspect Health Inf Manag",
  ["perspectives in healthcare risk management"] = "Perspect Healthc Risk Manage",
  ["perspectives in healthcare risk management / american society for healthcare risk management of the american hospital association"] = "Perspect Healthc Risk Manage",
  ["perspectives in logic"] = "Perspect. Log.",
  ["perspectives in mathematical logic"] = "Perspect. Math. Logic",
  ["perspectives in mathematics"] = "Perspect. Math.",
  ["perspectives in medical virology"] = "Perspect Med Virol",
  ["perspectives in medicinal chemistry"] = "Perspect Medicin Chem",
  ["perspectives in mexican american studies"] = "Perspect Mex Am Stud",
  ["perspectives in nephrology and hypertension"] = "Perspect Nephrol Hypertens",
  ["perspectives in neural computing"] = "Perspect. Neural Comput.",
  ["perspectives in pediatric pathology"] = "Perspect Pediatr Pathol",
  ["perspectives in plant ecology, evolution and systematics"] = "Perspect. Plant Ecol. Evol. Syst.",
  ["perspectives in psychiatric care"] = "Perspect Psychiatr Care",
  ["perspectives in public health"] = "Perspect Public Health",
  ["perspectives in respiratory nursing"] = "Perspect. Respir. Nurs.",
  ["perspectives in respiratory nursing : a publication of the respiratory nursing society"] = "Perspect Respir Nurs",
  ["perspectives in science"] = "Perspect. Sci.",
  ["perspectives in social work"] = "Perspect Soc Work",
  ["perspectives in vascular surgery and endovascular therapy"] = "Perspect Vasc Surg Endovasc Ther",
  ["perspectives internationales du planning familial"] = "Perspect Internation Plan Fam",
  ["perspectives of earth and space scientists"] = "Perspect. Earth Space Sci.",
  ["perspectives of the asha special interest groups"] = "Perspect ASHA Spec Interest Groups",
  ["perspectives on accreditation"] = "Perspect Accredit",
  ["perspectives on addictions nursing"] = "Perspect. Addict. Nurs.",
  ["perspectives on addictions nursing : a publication of the national nurses society on addictions"] = "Perspect Addict Nurs",
  ["perspectives on augmentative and alternative communication"] = "Perspect Augment altern commun",
  ["perspectives on behavior science"] = "Perspect Behav Sci",
  ["perspectives on developmental neurobiology"] = "Perspect Dev Neurobiol",
  ["perspectives on fluency and fluency disorders"] = "Perspect Fluen Fluen Disord",
  ["perspectives on gerontology"] = "Perspect Gerontol",
  ["perspectives on hearing and hearing disorders in childhood"] = "Perspect Hear Hear Disord Child",
  ["perspectives on hearing and hearing disorders. research and research diagnostics"] = "Perspect Hear Hear Disord Res Res Diagn",
  ["perspectives on labour and income"] = "Perspect. Lab. Income",
  ["perspectives on language and literacy"] = "Perspect Lang Lit",
  ["perspectives on language learning and education"] = "Perspect Lang Learn Educ",
  ["perspectives on medicaid and medicare management"] = "Perspect Medicaid Medicare Manage",
  ["perspectives on medicaid management"] = "Perspect Medicaid Manage",
  ["perspectives on medical education"] = "Perspect Med Educ",
  ["perspectives on neurophysiology and neurogenic speech and language disorders"] = "Perspect Neurophysiol Neurogenic Speech Lang Disord",
  ["perspectives on psychological science : a journal of the association for psychological science"] = "Perspect Psychol Sci",
  ["perspectives on public management and governance"] = "Perspect Public Manag Gov",
  ["perspectives on science"] = "Perspect. Sci.",
  ["perspectives on science : historical, philosophical, social"] = "Perspect Sci",
  ["perspectives on science and christian faith : journal of the american scientific affiliation"] = "Perspect Sci Christ Faith",
  ["perspectives on science. historical, philosophical, social"] = "Perspect. Sci.",
  ["perspectives on sexual and reproductive health"] = "Perspect Sex Reprod Health",
  ["perspectives on social work : the journal of the doctoral students of the university of houston graduate school of social work"] = "Perspect Soc Work (Houst)",
  ["perspectives on speech science and orofacial disorders"] = "Perspect Speech Sci Orofac Disord",
  ["perspectives on the professions"] = "Perspectives Prof",
  ["perspectives on undergraduate research and mentoring : purm"] = "Perspect Undergrad Res Mentor",
  ["perspektiven der mathematikdidaktik"] = "Perspekt. Math.didakt.",
  ["perspektiven der philosophie: neues jahrbuch"] = "PPH",
  ["perspektivy razvitiya vychislitelnoi tekhniki"] = "Perspekt. Razvit. Vychisl. Tekh.",
  ["pertanika journal of science & technology"] = "Pertanika J Sci Technol",
  ["pertanika journal of tropical agricultural science"] = "Pertanika J. Trop. Agric. Sci.",
  ["pervasive and mobile computing"] = "Pervasive Mob. Comput.",
  ["pervasive displays 2017 : the 6th acm international symposium on pervasive displays : proceedings : lugano, switzerland, june 7-9, 2017. international symposium on pervasive displays (6th : 2017 : lugano, switzerland)"] = "Pervasive Disp 2017 (2017)",
  ["pesquisa agropecuaria brasileira"] = "Pesqui. Agropecu. Bras.",
  ["pesquisa agropecuária brasileira"] = "Pesqui Agropecu Bras",
  ["pesquisa agropecuária tropical"] = "Pesqui Agropecu Trop",
  ["pesquisa agropecuária brasileira"] = "Pesqui. Agropecu. Bras.",
  ["pesquisa brasileira em odontopediatria e clínica integrada"] = "Pesqui Bras Odontopediatria Clin Integr",
  ["pesquisa e planejamento economico"] = "Pesqui Planej Econ",
  ["pesquisa e planejamento econômico"] = "Pesqui Planej Econ",
  ["pesquisa e planejamento econômico"] = "Pesquisa Planejamento Econ.",
  ["pesquisa florestal brasileira"] = "Pesqui Florest Bras",
  ["pesquisa odontologica brasileira"] = "Pesqui. Odontol. Bras.",
  ["pesquisa odontológica brasileira = brazilian oral research"] = "Pesqui Odontol Bras",
  ["pesquisa veterinária brasileira : revista do colégio brasileiro de patologia animal = brazilian journal of veterinary research"] = "Pesqui Vet Bras",
  ["pest control"] = "Pest Control",
  ["pest management in horticultural ecosystems"] = "Pest. Manage. Hortic. Ecosyst.",
  ["pest management science"] = "Pest Manag Sci",
  ["pesticide biochemistry and physiology"] = "Pestic. Biochem. Physiol.",
  ["pesticide outlook"] = "Pestic. Outlook",
  ["pesticide science"] = "Pestic Sci",
  ["pesticides monitoring journal"] = "Pestic Monit J",
  ["pesticides, people and nature"] = "Pestic. People Nat.",
  ["pesticidi i fitomedicina = pesticides and phytomedicine"] = "Pestic Fitomed",
  ["pests and their control"] = "Pest Control",
  ["pet clinics"] = "PET Clin",
  ["petermanns geographische mitteilungen"] = "Petermanns Geogr Mitt",
  ["petria (roma)"] = "Petria (Roma)",
  ["petroleum and coal"] = "Pet. Coal",
  ["petroleum chemistry"] = "Pet. Chem.",
  ["petroleum engineering"] = "Pet. Eng.",
  ["petroleum exploration and development"] = "Pet. Explor. Dev.",
  ["petroleum geoscience"] = "Pet. Geosci.",
  ["petroleum research"] = "Pet. Res.",
  ["petroleum science"] = "Pet. Sci.",
  ["petroleum science and technology"] = "Pet. Sci. Technol.",
  ["petronian society newsletter"] = "PSN",
  ["peuce. studii şi comunicări de istorie veche, arheologie şi numismatică"] = "Peuce",
  ["peuples mediterraneens. mediterranean peoples"] = "Mediterr Peoples",
  ["peuples méditerranéens. mediterranean peoples"] = "Mediterr Peoples",
  ["pénélope (paris, france)"] = "Penelope",
  ["pfca review"] = "PFCA Rev",
  ["pfg-journal of photogrammetry remote sensing and geoinformation science"] = "PFG- J. Photogramm. Remote Sens. Geoinf. Sci.",
  ["pflanzenschutz-nachrichten bayer"] = "Pflanzenschutz-Nachr. Bayer",
  ["pflege aktuell"] = "Pflege Aktuell",
  ["pflege aktuell / dbfk, deutscher berufsverband fur pflegeberufe"] = "Pflege Aktuell",
  ["pflege zeitschrift"] = "Pflege Z",
  ["pfluegers archiv"] = "Pfluegers Arch.",
  ["pflugers archiv (european journal of physiology)"] = "Pflugers Arch.",
  ["pflugers archiv - european journal of physiology"] = "Pflugers Arch. - Eur. J. Physiol.",
  ["pflugers archiv : european journal of physiology"] = "Pflugers Arch",
  ["pflugers archiv fur die gesamte physiologie des menschen und der tiere"] = "Pflugers Arch. Gesamte Physiol. Menschen Tiere",
  ["pflugers archiv. european journal of physiology"] = "Pflugers Arch",
  ["pflügers archiv : european journal of physiology"] = "Pflugers Arch",
  ["pflügers archiv für die gesamte physiologie des menschen und der tiere"] = "Pflugers Arch Gesamte Physiol Menschen Tiere",
  ["pga proceedings of the geologists’ association"] = "PGA Proc. Geol. Assoc.",
  ["phage (new rochelle, n.y.)"] = "Phage (New Rochelle)",
  ["phage: therapy, applications, and research"] = "PHAGE: Ther. Appl. Res.",
  ["pharma chemica"] = "Pharma Chem.",
  ["pharma focus asia"] = "Pharma Focus Asia",
  ["pharmacal advance"] = "Pharmacal Adv",
  ["pharmaceutica acta helvetiae"] = "Pharm Acta Helv",
  ["pharmaceutica analytica acta"] = "Pharm Anal Acta",
  ["pharmaceutical & diagnostic innovation"] = "Pharm Diagn Innov",
  ["pharmaceutical and pharmacological letters"] = "Pharm. Pharmacol. Lett.",
  ["pharmaceutical biology"] = "Pharm Biol",
  ["pharmaceutical bioprocessing"] = "Pharm Bioprocess",
  ["pharmaceutical biotechnology"] = "Pharm Biotechnol",
  ["pharmaceutical bulletin"] = "Pharm Bull",
  ["pharmaceutical chemistry journal"] = "Pharm. Chem. J.",
  ["pharmaceutical crops"] = "Pharm Crop",
  ["pharmaceutical development and regulation"] = "Pharm Dev Regul",
  ["pharmaceutical development and technology"] = "Pharm Dev Technol",
  ["pharmaceutical frontiers"] = "Pharm Front",
  ["pharmaceutical historian"] = "Pharm Hist (Lond)",
  ["pharmaceutical medicine"] = "Pharmaceut Med",
  ["pharmaceutical methods"] = "Pharm. Methods",
  ["pharmaceutical nanotechnology"] = "Pharm Nanotechnol",
  ["pharmaceutical outsourcing"] = "Pharm Outsourcing",
  ["pharmaceutical patent analyst"] = "Pharm Pat Anal",
  ["pharmaceutical regulatory affairs : open access"] = "Pharm Regul Aff",
  ["pharmaceutical research"] = "Pharm Res",
  ["pharmaceutical reviews"] = "Pharm Rev",
  ["pharmaceutical science & technology today"] = "Pharm. Sci. Technol. Today",
  ["pharmaceutical statistics"] = "Pharm Stat",
  ["pharmaceutical technology"] = "Pharm. Technol.",
  ["pharmaceutical technology in hospital pharmacy"] = "Pharm Technol Hosp Pharm",
  ["pharmaceuticals (basel, switzerland)"] = "Pharmaceuticals (Basel)",
  ["pharmaceutisch weekblad"] = "Pharm Weekbl",
  ["pharmaceutisch weekblad. scientific edition"] = "Pharm Weekbl Sci",
  ["pharmacia sinica"] = "Pharm. Sin.",
  ["pharmacien de france"] = "Pharm Fr",
  ["pharmacoeconomics & outcomes news"] = "PharmacoEcon Outcomes News",
  ["pharmacoeconomics - open"] = "Pharmacoecon Open",
  ["pharmacoepidemiology and drug safety"] = "Pharmacoepidemiol Drug Saf",
  ["pharmacogenetics and genomics"] = "Pharmacogenet Genomics",
  ["pharmacogenomics and personalized medicine"] = "Pharmgenomics Pers Med",
  ["pharmacogenomics journal"] = "Pharmacogenomics J.",
  ["pharmacognosy magazine"] = "Pharmacogn Mag",
  ["pharmacognosy research"] = "Pharmacognosy Res",
  ["pharmacognosy reviews"] = "Pharmacogn. Rev.",
  ["pharmacological reports"] = "Pharmacol. Rep.",
  ["pharmacological reports : pr"] = "Pharmacol Rep",
  ["pharmacological research"] = "Pharmacol Res",
  ["pharmacological research - modern chinese medicine"] = "Pharmacol. Res. Mod. Chin. Med.",
  ["pharmacological research : the official journal of the italian pharmacological society"] = "Pharmacol Res",
  ["pharmacological research communications"] = "Pharmacol. Res. Commun.",
  ["pharmacological reviews"] = "Pharmacol Rev",
  ["pharmacology & pharmacy"] = "Pharmacol Pharm",
  ["pharmacology & therapeutics"] = "Pharmacol Ther",
  ["pharmacology & therapeutics part a: chemotherapy, toxicology and metabolic inhibitors"] = "Pharmacol. Ther. Part A",
  ["pharmacology & therapeutics part b: general and systemic pharmacology"] = "Pharmacol. Ther. Part B",
  ["pharmacology & therapeutics part c: clinical pharmacology and therapeutics"] = "Pharmacol. Ther. Part C",
  ["pharmacology & therapeutics, part a: chemotherapy, toxicology and metabolic inhibitors"] = "Pharmacol. Ther., Part A",
  ["pharmacology & therapeutics, part b: general & systematic pharmacology"] = "Pharmacol. Ther., Part B",
  ["pharmacology & therapeutics, part c: clinical pharmacology and therapeutics"] = "Pharmacol. Ther., Part C",
  ["pharmacology & therapeutics. part b: general & systematic pharmacology"] = "Pharmacol Ther B",
  ["pharmacology & toxicology"] = "Pharmacol Toxicol",
  ["pharmacology & toxicology (copenhagen)"] = "Pharmacol. Toxicol. (Copenhagen)",
  ["pharmacology and therapeutics"] = "Pharmacol. Ther.",
  ["pharmacology and therapeutics in dentistry"] = "Pharmacol Ther Dent",
  ["pharmacology and therapeutics. part b, general and systematic pharmacology"] = "Pharmacol. Ther. [B]",
  ["pharmacology and toxicology"] = "Pharmacol. Toxicol.",
  ["pharmacology biochemistry and behavior"] = "Pharmacol. Biochem. Behav.",
  ["pharmacology biochemistry and behavior."] = "Pharmacol. Biochem. Behav",
  ["pharmacology communications"] = "Pharmacol Commun",
  ["pharmacology for physicians"] = "Pharmacol Physicians",
  ["pharmacology research & perspectives"] = "Pharmacol Res Perspect",
  ["pharmacology, biochemistry and behavior"] = "Pharmacol.  Biochem. Behav.",
  ["pharmacology, biochemistry, and behavior"] = "Pharmacol Biochem Behav",
  ["pharmacology, drug development & therapeutics"] = "Pharmacol Drug Dev Ther",
  ["pharmacy & pharmacology international journal"] = "Pharm Pharmacol Int J",
  ["pharmacy (basel, switzerland)"] = "Pharmacy (Basel)",
  ["pharmacy and pharmacology communications"] = "Pharm. Pharmacol. Commun.",
  ["pharmacy history australia : the newsletter of the australian academy of the history of pharmacy"] = "Pharm Hist Aust",
  ["pharmacy in history"] = "Pharm Hist",
  ["pharmacy international"] = "Pharm Int",
  ["pharmacy management combined with the american journal of pharmacy"] = "Pharm. Manage. Comb. Am. J. Pharm.",
  ["pharmacy management combined with the american journal of pharmacy : pm"] = "Pharm Manage Comb Am J Pharm",
  ["pharmacy practice"] = "Pharm Pract (Granada)",
  ["pharmacy practice management quarterly"] = "Pharm Pract Manag Q",
  ["pharmacy times"] = "Pharm Times",
  ["pharmacy today"] = "Pharm. Today",
  ["pharmacy update"] = "Pharm Update",
  ["pharmacy world & science : pws"] = "Pharm World Sci",
  ["pharmacy world and science"] = "Pharm. World Sci.",
  ["pharmakopsychiatrie neuro-psychopharmakologie"] = "Pharmakopsychiatr. Neuropsychopharmakol.",
  ["pharmakopsychiatrie, neuro-psychopharmakologie"] = "Pharmakopsychiatr Neuropsychopharmakol",
  ["pharmazeutische praxis"] = "Pharm Prax",
  ["pharmazeutische rundschau"] = "Pharm Rundsch",
  ["pharmazeutische zeitung"] = "Pharm Ztg (1856)",
  ["pharmazeutische zentralhalle fur deutschland"] = "Pharm. Zentralhalle Dtschl.",
  ["pharmazeutische zentralhalle für deutschland"] = "Pharm Zentralhalle Dtschl",
  ["pharmazie in unserer zeit"] = "Pharm Unserer Zeit",
  ["pharmaziehistorische bibliographie : phb"] = "Pharmaziehist Bibliogr",
  ["pharmaziehistorische forschungen"] = "Pharmaziehist Forsch",
  ["pharmeuropa bio"] = "Pharmeuropa Bio",
  ["pharmeuropa bio & scientific notes"] = "Pharmeur Bio Sci Notes",
  ["pharmeuropa bio / the biological standardisation programme, edqm"] = "Pharmeuropa Bio",
  ["pharmeuropa scientific notes"] = "Pharmeur Sci Notes",
  ["pharmeuropa. special issue biologicals"] = "Pharmeuropa Spec Issue Biol",
  ["pharos of alpha omega alpha honor medical society"] = "Pharos",
  ["pharos. journal of the netherlands institute at athens"] = "Pharos",
  ["phase transition and critical phenomena"] = "Phase Transit. Crit. Phenom.",
  ["phase transitions"] = "Phase Transitions",
  ["phc4 fyi"] = "PHC4 FYI",
  ["phc4 fyi / pennsylvania health care cost containment council"] = "PHC4 FYI",
  ["phenomenology and the cognitive sciences"] = "Phenomenol Cogn Sci",
  ["phi delta kappan"] = "Phi Delta Kappan",
  ["phi kappa phi journal"] = "Phi Kappa Phi J",
  ["philadelphia health bulletin"] = "Phila Health Bull",
  ["philadelphia inquirer (philadelphia, pa. : 1969)"] = "Phila Inq",
  ["philadelphia medicine"] = "Phila Med",
  ["philadelphia social innovations journal"] = "Phila Soc Innov J",
  ["philippine agricultural science"] = "Philipp. Agric. Sci.",
  ["philippine development"] = "Philipp Dev",
  ["philippine economic journal"] = "Philippine Econ. J.",
  ["philippine entomologist"] = "Philipp Entomol",
  ["philippine geographical journal"] = "Philipp Geogr J",
  ["philippine journal of business and economics"] = "Philippine J. Bus. Econ.",
  ["philippine journal of business and finance"] = "Philippine J. Bus. Finance",
  ["philippine journal of cancer"] = "Philipp J Cancer",
  ["philippine journal of cardiology"] = "Philipp J Cardiol",
  ["philippine journal of crop science"] = "Philipp. J. Crop Sci.",
  ["philippine journal of internal medicine"] = "Philipp J Intern Med",
  ["philippine journal of mental health"] = "Philipp J Ment Health",
  ["philippine journal of nursing"] = "Philipp. J. Nurs.",
  ["philippine journal of nutrition"] = "Philipp J Nutr",
  ["philippine journal of obstetrics & gynecology : official publication, philippine obstetrical and gynecological society"] = "Philipp J Obstet Gynecol",
  ["philippine journal of ophthalmology"] = "Philipp J Ophthalmol",
  ["philippine journal of ophthalmology and otolaryngology"] = "Philipp J Ophthalmol Otolaryngol",
  ["philippine journal of psychology"] = "Philipp J Psychol",
  ["philippine journal of public administration"] = "Philipp J Public Adm",
  ["philippine journal of science"] = "Philipp J Sci",
  ["philippine journal of surgery"] = "Philipp J Surg",
  ["philippine journal of surgery and surgical specialties"] = "Philipp J Surg Surg Spec",
  ["philippine journal of surgical specialties"] = "Philipp J Surg Spec",
  ["philippine journal of systematic biology"] = "Philipp J Syst Biol",
  ["philippine journal of veterinary and animal sciences"] = "Philipp J Vet Anim Sci",
  ["philippine law journal"] = "Philipp Law J",
  ["philippine population journal"] = "Philipp Popul J",
  ["philippine population newsletter"] = "Philipp Popul Newsl",
  ["philippine quarterly of culture and society"] = "Philipp Q Cult Soc",
  ["philippine review of economics"] = "Philippine Rev. Econ.",
  ["philippine review of economics and business"] = "Philippine Rev. Econ. Bus.",
  ["philippine science letters"] = "Philipp Sci Lett",
  ["philippine social sciences and humanities review"] = "Philipp Soc Sci Humanit Rev",
  ["philippine sociological review"] = "Philipp Sociol Rev",
  ["philippine studies"] = "Philipp Stud",
  ["philips journal of research"] = "Philips J. Res.",
  ["philips research reports"] = "Philips Res. Rep.",
  ["phillip journal"] = "Phillip J",
  ["phillip journal fur restaurative zahnmedizin"] = "Phillip J. Restaur. Zahnmed.",
  ["phillip journal für restaurative zahnmedizin"] = "Phillip J Restaur Zahnmed",
  ["philologia classica: recueil interuniversitaire périodique"] = "PhilClass",
  ["philological quarterly"] = "PhQ",
  ["philologus. zeitschrift für das klassische altertum"] = "Philologus",
  ["philologus: zeitschrift für antike literatur und ihre rezeption"] = "Philologus",
  ["philosophes medievaux"] = "Philos. Mediev.",
  ["philosophia (ramat-gan, israel)"] = "Philosophia (Ramat Gan)",
  ["philosophia antiqua"] = "Philos. Antiq.",
  ["philosophia mathematica"] = "Philos. Math.",
  ["philosophia mathematica. series iii"] = "Philos. Math. (3)",
  ["philosophia naturalis"] = "Philos Nat",
  ["philosophia scientiae. travaux d’histoire des sciences et de philosophie"] = "Philos. Sci. (Paris)",
  ["philosophic exchange"] = "Philos Exch",
  ["philosophic research and analysis"] = "Philos Res Analysis",
  ["philosophical explorations : an international journal for the philosophy of mind and action"] = "Philos Explor",
  ["philosophical mag. abingdon england"] = "Philos. Mag. Abingdon Engl.",
  ["philosophical magazine"] = "Philos. Mag.",
  ["philosophical magazine (abingdon, england)"] = "Philos Mag (Abingdon)",
  ["philosophical magazine a"] = "Philos. Mag. A",
  ["philosophical magazine a: physics of condensed matter: defects and mechanical properties"] = "Philos. Mag. A",
  ["philosophical magazine b"] = "Philos. Mag. B",
  ["philosophical magazine b: physics of condensed matter: statistical mechanics, electronic, optical and magnetic properties"] = "Philos. Mag. B",
  ["philosophical magazine b: physics of condensed matter: structural, electronic, optical and magnetic properties"] = "Philos. Mag. B",
  ["philosophical magazine letters"] = "Philos. Mag. Lett.",
  ["philosophical perspectives"] = "Philos. Perspect.",
  ["philosophical perspectives. a supplement to nous"] = "Philos. Perspect.",
  ["philosophical psychology"] = "Philos Psychol",
  ["philosophical review"] = "PhR",
  ["philosophical studies"] = "Philos Stud",
  ["philosophical studies series"] = "Philos. Stud. Ser.",
  ["philosophical studies. an international journal for philosophy in the analytic tradition"] = "Philos. Stud.",
  ["philosophical transactions of the royal society"] = "Philos. Trans. R. Soc.",
  ["philosophical transactions of the royal society a"] = "Philos. Trans. R. Soc. A",
  ["philosophical transactions of the royal society a. mathematical, physical and engineering sciences"] = "Philos. Trans. Roy. Soc. A",
  ["philosophical transactions of the royal society a: mathematical, physical and engineering sciences"] = "Philos. Trans. R. Soc. A: Math. Phys. Eng. Sci.",
  ["philosophical transactions of the royal society b"] = "Phil. Trans. R. Soc. B",
  ["philosophical transactions of the royal society of london"] = "Philos. Trans. Roy. Soc. London",
  ["philosophical transactions of the royal society of london, series a"] = "Philos. Trans. R. Soc. London, Ser. A",
  ["philosophical transactions of the royal society of london, series a: mathematical and physical sciences"] = "Philos. Trans. R. Soc. London, Ser. A",
  ["philosophical transactions of the royal society of london, series a: mathematical, physical and engineering sciences"] = "Philos. Trans. R. Soc. London, Ser. A",
  ["philosophical transactions of the royal society of london, series b"] = "Philos. Trans. R. Soc. London, Ser. B",
  ["philosophical transactions of the royal society of london, series b: biological sciences"] = "Philos. Trans. R. Soc. London, Ser. B",
  ["philosophical transactions of the royal society of london. a"] = "Philos Trans R Soc Lond A",
  ["philosophical transactions of the royal society of london. series a. mathematical, physical and engineering sciences"] = "Philos. Trans. R. Soc. Lond. Ser. A Math. Phys. Eng. Sci.",
  ["philosophical transactions of the royal society of london. series a: mathematical and physical sciences"] = "Philos Trans R Soc Lond A",
  ["philosophical transactions of the royal society of london. series b, biological sciences"] = "Philos Trans R Soc Lond B Biol Sci",
  ["philosophical transactions of the royal society of london. series b: biological sciences"] = "Philos Trans R Soc Lond B Biol Sci",
  ["philosophical transactions. physical sciences and engineering"] = "Philos Trans Phys Sci Eng",
  ["philosophical transactions. series a, mathematical, physical, and engineering sciences"] = "Philos Trans A Math Phys Eng Sci",
  ["philosophische abhandlungen"] = "Philos. Abhandlungen",
  ["philosophische analyse/philosophical analysis"] = "Philos. Anal.",
  ["philosophische bibliothek"] = "Philos. Bibliothek",
  ["philosophische rundschau: eine zeitschrift für philosophische kritik"] = "PhRdschau",
  ["philosophische texte und studien"] = "Philos. Texte Stud.",
  ["philosophischer literaturanzeiger"] = "PhLA",
  ["philosophisches jahrbuch"] = "PhJ",
  ["philosophy & public affairs"] = "Philos Public Aff",
  ["philosophy & public policy quarterly"] = "Philos Public Policy Q",
  ["philosophy & public policy quarterly / the institute for philosophy and public policy, school of public affairs, university of maryland"] = "Philos Public Policy Q",
  ["philosophy & social criticism"] = "Philos Soc Crit",
  ["philosophy & technology"] = "Philos Technol",
  ["philosophy (london, england)"] = "Philosophy",
  ["philosophy and cultural studies revisited/historisch-genetische studien zur philosophie und kulturgeschichte"] = "Philos. Cult. Stud. Revisit./Hist-genet. Stud. Philos. Kult.gesch.",
  ["philosophy and literature"] = "Ph&Lit",
  ["philosophy and phenomenological research"] = "Philos Phenomenol Res",
  ["philosophy and rhetoric"] = "Ph&Rh",
  ["philosophy and technology"] = "Philos. Technol.",
  ["philosophy compass"] = "Philos Compass",
  ["philosophy east & west"] = "Philos East West",
  ["philosophy in context"] = "Philos Context",
  ["philosophy of engineering and technology"] = "Philos. Eng. Technol.",
  ["philosophy of management"] = "Philos Manag",
  ["philosophy of science"] = "Philos Sci",
  ["philosophy of the social sciences"] = "Philos Soc Sci",
  ["philosophy research archives (bowling green, ohio : 1982)"] = "Philos Res Arch",
  ["philosophy, ethics, and humanities in medicine"] = "Philos. Ethics Humanit. Med.",
  ["philosophy, ethics, and humanities in medicine : pehm"] = "Philos Ethics Humanit Med",
  ["philosophy, psychiatry, & psychology : ppp"] = "Philos Psychiatr Psychol",
  ["philosophy, theology"] = "Philos Theol",
  ["phispc monograph series on history of philosophy, science and culture in india"] = "PHISPC Monogr. Ser. Hist. Philos. Sci. Cult. India",
  ["phlebitis bulletin"] = "Phlebitis Bull",
  ["phoebe (oneonta, n.y.)"] = "Phoebe (Oneonta N Y)",
  ["phoenix. bulletin uitgegeven door het vooraziatisch-egyptisch genootschap »ex oriente lux«"] = "PhoenixExOrLux",
  ["phoenix. the journal of the classical association of canada"] = "Phoenix",
  ["phoenix: journal of the classical association of canada"] = "Phoenix",
  ["phoibos. bulletin du cercle de philologie classique et orientale de l’université libre de bruxelles"] = "Phoibos",
  ["phosphorus sulfur and related elements"] = "Phosphorus Sulfur Rel. Elem.",
  ["phosphorus sulfur silicon and the related elements"] = "Phosphorus, Sulfur Silicon Relat. Elem.",
  ["phosphorus, sulfur and silicon and the related elements"] = "Phosphorus, Sulfur Silicon Relat. Elem.",
  ["phosphorus, sulfur, and silicon and the related elements"] = "Phosphorus Sulfur Silicon Relat Elem",
  ["photobiomodulation, photomedicine, and laser surgery"] = "Photobiomodul Photomed Laser Surg",
  ["photochemical & photobiological sciences"] = "Photochem. Photobiol. Sci.",
  ["photochemical & photobiological sciences : official journal of the european photochemistry association and the european society for photobiology"] = "Photochem Photobiol Sci",
  ["photochemical and photobiological sciences"] = "Photochem. Photobiol. Sci.",
  ["photochemistry & photobiological sciences"] = "Photochem. Photobiol. Sci.",
  ["photochemistry and photobiology"] = "Photochem Photobiol",
  ["photochemistry-a specialist periodical report"] = "Photochem-Spec. Perio.",
  ["photodermatology, photoimmunology & photomedicine"] = "Photodermatol Photoimmunol Photomed",
  ["photodermatology, photoimmunology and photomedicine"] = "Photodermatol. Photoimmunol. Photomed.",
  ["photodiagnosis and photodynamic therapy"] = "Photodiagnosis Photodyn Ther",
  ["photogrammetric engineering and remote sensing"] = "Photogramm. Eng. Remote Sens.",
  ["photogrammetric record"] = "Photogramm. Rec.",
  ["photogrammetric record, the"] = "Photogramm. Rec.",
  ["photographic science and engineering"] = "Photogr. Sci. Eng.",
  ["photomedicine and laser surgery"] = "Photomed Laser Surg",
  ["photonic network communications"] = "Photonic Network Commun.",
  ["photonic sensors"] = "Photonic Sensors",
  ["photonics & lasers in medicine"] = "Photonics Lasers Med",
  ["photonics and nanostructures - fundamentals and applications"] = "Photonics Nanostruct. Fundam. Appl.",
  ["photonics and nanostructures : fundamentals and applications"] = "Photonics Nanostruct",
  ["photonics insights"] = "Photonics Insights",
  ["photonics letters of poland"] = "Photonics Lett Pol",
  ["photonics research"] = "Photonics Res.",
  ["photonics spectra"] = "Photonics Spectra",
  ["photosynthesis research"] = "Photosynth Res",
  ["photovoltaics bulletin"] = "Photovoltaics Bull.",
  ["phronesis (barcelona, spain)"] = "Phronesis",
  ["phronesis. a journal for ancient philosophy"] = "Phronesis",
  ["phycological research"] = "Phycological Res",
  ["phykos : journal of the phycological society (india)"] = "Phykos",
  ["phyllomedusa : journal of neotropical herpetology"] = "Phyllomedusa",
  ["phylon (1960)"] = "Phylon",
  ["physica (utrecht)"] = "Physica (Utrecht)",
  ["physica a"] = "Physica A",
  ["physica a. statistical mechanics and its applications"] = "Phys. A",
  ["physica a: statistical mechanics and its applications"] = "Physica A",
  ["physica a: statistical mechanics and its applications (amsterdam, netherlands)"] = "Physica A (Amsterdam, Neth.)",
  ["physica b"] = "Phys. B",
  ["physica b: condensed matter"] = "Physica B",
  ["physica b: condensed matter (amsterdam, netherlands)"] = "Phys. B (Amsterdam, Neth.)",
  ["physica c"] = "Phys. C",
  ["physica c: superconductivity"] = "Physica C",
  ["physica c: superconductivity and its applications"] = "Phys. C: Supercond. Its Appl.",
  ["physica c: superconductivity and its applications (amsterdam, netherlands)"] = "Phys. C (Amsterdam, Neth.)",
  ["physica d"] = "Phys. D",
  ["physica d. nonlinear phenomena"] = "Physica D",
  ["physica d: nonlinear phenomena"] = "Physica D",
  ["physica d: nonlinear phenomena (amsterdam, netherlands)"] = "Physica D (Amsterdam, Neth.)",
  ["physica e"] = "Phys. E",
  ["physica e-low-dimensional systems & nanostructures"] = "Physica E",
  ["physica e: low-dimensional systems and nanostructures"] = "Physica E",
  ["physica e: low-dimensional systems and nanostructures (amsterdam, netherlands)"] = "Physica E (Amsterdam, Neth.)",
  ["physica medica"] = "Physica Med.",
  ["physica medica : pm : an international journal devoted to the applications of physics to medicine and biology : official journal of the italian association of biomedical physics (aifb)"] = "Phys Med",
  ["physica scripta"] = "Phys. Scr.",
  ["physica scripta t"] = "Phys. Scr. T",
  ["physica scripta, t"] = "Phys. Scr., T",
  ["physica scripta, t."] = "Phys. Scr. T.",
  ["physica status solidi"] = "Phys. Status Solidi",
  ["physica status solidi (a)"] = "Phys. Status Solidi A",
  ["physica status solidi (a): applied research"] = "Physica Status Solidi A Appl Res",
  ["physica status solidi (b)"] = "Phys. Status Solidi B",
  ["physica status solidi (c)"] = "Phys. Status Solidi C",
  ["physica status solidi (rrl) – rapid research letters"] = "Phys. Status Solidi RRL",
  ["physica status solidi a"] = "Phys. Status Solidi A",
  ["physica status solidi a\tapplications and material science"] = "Phys. Status Solidi A",
  ["physica status solidi a applications and material science"] = "Phys. Status Solidi A",
  ["physica status solidi a: applications and materials science"] = "Phys. Status Solidi A",
  ["physica status solidi a: applied research"] = "Phys. Status Solidi A: Appl. Res.",
  ["physica status solidi b"] = "Phys. Status Solidi B",
  ["physica status solidi b\tbasic solid state physics"] = "Phys. Status Solidi B",
  ["physica status solidi b basic solid state physics"] = "Phys. Status Solidi B",
  ["physica status solidi b: basic research"] = "Phys. Status Solidi B: Basic Res.",
  ["physica status solidi b: basic solid state physics"] = "Phys. Status Solidi B: Basic Solid State Phys.",
  ["physica status solidi c"] = "Phys. Status Solidi C",
  ["physica status solidi c\tcurrent topics in solid state physics"] = "Phys. Status Solidi C",
  ["physica status solidi c current topics in solid state physics"] = "Phys. Status Solidi C",
  ["physica status solidi c: conferences and critical reviews"] = "Phys. Status Solidi C",
  ["physica status solidi c: current topics in solid state physics"] = "Phys. Status Solidi C: Curr. Top. Solid State Phys.",
  ["physica status solidi rapid research letters"] = "Phys. Status Solidi RRL",
  ["physica status solidi rrl: rapid research letters"] = "Phys. Status Solidi RRL",
  ["physica status solidi. b, basic research"] = "Phys Status Solidi B Basic Res",
  ["physica status solidi. b, basic solid state physics : pss"] = "Phys Status Solidi B Basic Solid State Phys",
  ["physica status solidi. rapid research letters : pss-rrl"] = "Phys Status Solidi Rapid Res Lett",
  ["physica. b, condensed matter"] = "Physica B Condens Matter",
  ["physica. c, superconductivity"] = "Physica C Supercond",
  ["physica. e, low-dimensional systems & nanostructures"] = "Physica E Low Dimens Syst Nanostruct",
  ["physical & occupational therapy in geriatrics"] = "Phys Occup Ther Geriatr",
  ["physical & occupational therapy in pediatrics"] = "Phys Occup Ther Pediatr",
  ["physical activity and health"] = "Phys. Act. Health",
  ["physical activity and nutrition"] = "Phys Act Nutr",
  ["physical and engineering sciences in medicine"] = "Phys Eng Sci Med",
  ["physical biology"] = "Phys Biol",
  ["physical chemistry chemical physics"] = "Phys. Chem. Chem. Phys.",
  ["physical chemistry chemical physics : pccp"] = "Phys Chem Chem Phys",
  ["physical chemistry research"] = "Phys. Chem. Res.",
  ["physical communication"] = "Phys. Commun.",
  ["physical disabilities : education and related services"] = "Phys Disabil",
  ["physical education and sport pedagogy"] = "Phys Educ Sport Pedagogy",
  ["physical geography"] = "Phys Geogr",
  ["physical medicine and rehabilitation clinics of north america"] = "Phys Med Rehabil Clin N Am",
  ["physical medicine and rehabilitation international"] = "Phys Med Rehabil Int",
  ["physical medicine and rehabilitation research"] = "Phys Med Rehabil Res",
  ["physical mesomechanics"] = "Phys. Mesomech.",
  ["physical oceanography"] = "Phys. Oceanogr.",
  ["physical properties of the low-dimensional a3b6 and a3b3c62 compounds"] = "Condens. Matter. Res. T.",
  ["physical review"] = "Phys. Rev.",
  ["physical review a"] = "Phys. Rev. A",
  ["physical review a: atomic, molecular, and optical physics"] = "Phys. Rev. A: At. Mol. Opt. Phys.",
  ["physical review a: general physics"] = "Phys Rev A Gen Phys",
  ["physical review accelerators and beams"] = "Phys. Rev. Accel. Beams",
  ["physical review applied"] = "Phys. Rev. Applied",
  ["physical review b"] = "Phys. Rev. B",
  ["physical review b: condensed matter"] = "Phys. Rev. B: Condens. Matter",
  ["physical review b: condensed matter and materials physics"] = "Phys. Rev. B: Condens. Matter Mater. Phys.",
  ["physical review c"] = "Phys. Rev. C",
  ["physical review c: nuclear physics"] = "Phys. Rev. C: Nucl. Phys.",
  ["physical review d"] = "Phys. Rev. D",
  ["physical review d: particles and fields"] = "Phys. Rev. D: Part. Fields",
  ["physical review d: particles, fields, gravitation, and cosmology"] = "Phys. Rev. D: Part., Fields, Gravitation, Cosmol.",
  ["physical review e"] = "Phys. Rev. E",
  ["physical review e: statistical physics, plasmas, fluids, and related interdisciplinary topics"] = "Phys. Rev. E: Stat. Phys. Plasmas Fluids Relat. Interdisciplin. Top.",
  ["physical review e: statistical, nonlinear, and soft matter physics"] = "Phys. Rev. E: Stat. Nonlinear Soft Matter Phys.",
  ["physical review fluids"] = "Phys. Rev. Fluids",
  ["physical review letters"] = "Phys. Rev. Lett.",
  ["physical review materials"] = "Phys. Rev. Materials",
  ["physical review physics education research"] = "Phys. Rev. Phys. Educ. Res.",
  ["physical review research"] = "Phys Rev Res",
  ["physical review special topics - accelerators and beams"] = "Phys. Rev. Spec. Top. Accel Beams",
  ["physical review special topics - physics education research"] = "Phys. Rev. Spec. Top. Phys. Educ. Res.",
  ["physical review special topics – accelerators and beams"] = "Phys. Rev. Spec. Top. Accel. Beams",
  ["physical review x"] = "Phys. Rev. X",
  ["physical review. a"] = "Phys Rev A  (Coll Park)",
  ["physical review. a, atomic, molecular, and optical physics"] = "Phys Rev A",
  ["physical review. a, general physics"] = "Phys Rev A Gen Phys",
  ["physical review. b"] = "Phys Rev B",
  ["physical review. b, condensed matter"] = "Phys Rev B Condens Matter",
  ["physical review. b, condensed matter and materials physics"] = "Phys Rev B Condens Matter Mater Phys",
  ["physical review. c"] = "Phys Rev C",
  ["physical review. c, nuclear physics"] = "Phys Rev C Nucl Phys",
  ["physical review. d, particles and fields"] = "Phys Rev D Part Fields",
  ["physical review. d. (2016)"] = "Phys Rev D",
  ["physical review. e"] = "Phys Rev E",
  ["physical review. e, statistical physics, plasmas, fluids, and related interdisciplinary topics"] = "Phys Rev E Stat Phys Plasmas Fluids Relat Interdiscip Topics",
  ["physical review. e, statistical, nonlinear, and soft matter physics"] = "Phys Rev E Stat Nonlin Soft Matter Phys",
  ["physical review. physics education research"] = "Phys Rev Phys Educ Res",
  ["physical review. x"] = "Phys Rev X",
  ["physical separaion in science and engineering"] = "Phys. Sep. Sci. Eng.",
  ["physical therapy"] = "Phys Ther",
  ["physical therapy and rehabilitation"] = "Phys Ther Rehabil",
  ["physical therapy in sport : official journal of the association of chartered physiotherapists in sports medicine"] = "Phys Ther Sport",
  ["physical therapy research"] = "Phys Ther Res",
  ["physical therapy reviews : ptr"] = "Phys Ther Rev",
  ["physician assistant"] = "Physician Assist",
  ["physician assistant (american academy of physician assistants)"] = "Physician Assist",
  ["physician assistant clinics"] = "Physician Assist Clin",
  ["physician assistant. health practitioner"] = "Physician Assist Health Pract",
  ["physician east"] = "Phys East",
  ["physician executive"] = "Physician Exec",
  ["physician leadership journal"] = "Physician Leadersh J",
  ["physician performance & payment report"] = "Physician Perform Paym Rep",
  ["physician relations update"] = "Physician Relat Update",
  ["physician relations update / american health consultants"] = "Physician Relat Update",
  ["physician's bulletin"] = "Physician's Bull",
  ["physician's management"] = "Physicians Manage",
  ["physician's world"] = "Physicians World",
  ["physicians management"] = "Physicians Manage.",
  ["physician’s management"] = "Physicians Manage",
  ["physician’s world"] = "Physicians World",
  ["physicochemical problems of mineral processing"] = "Physicochem. Probl. Miner. Process.",
  ["physics (new york)"] = "Physics (N.Y.)",
  ["physics and chemistry of glasses"] = "Phys. Chem. Glasses",
  ["physics and chemistry of glasses: european journal of glass science and technology, part b"] = "Phys. Chem. Glasses: Eur. J. Glass Sci. Technol., Part B",
  ["physics and chemistry of liquids"] = "Phys. Chem. Liq.",
  ["physics and chemistry of minerals"] = "Phys. Chem. Miner.",
  ["physics and chemistry of solids"] = "Phys. Chem. Solids",
  ["physics and chemistry of the earth"] = "Phys. Chem. Earth.",
  ["physics and chemistry of the earth (2002)"] = "Phys Chem Earth (2002)",
  ["physics and chemistry of the earth a/b/c"] = "Phys. Chem. Earth. A/B/C/",
  ["physics and chemistry of the earth b"] = "Phys. Chem Earth B",
  ["physics and chemistry of the earth, part a: solid earth and geodesy"] = "Phys. Chem. Earth Part A.",
  ["physics and chemistry of the earth, part b: hydrology, oceans and atmosphere"] = "Phys. Chem. Earth Part B",
  ["physics and chemistry of the earth, part c: solar, terrestrial & planetary science"] = "Phys. Chem. Earth Part C",
  ["physics and imaging in radiation oncology"] = "Phys Imaging Radiat Oncol",
  ["physics education"] = "Phys. Educ.",
  ["physics education research"] = "Phys. Educ. Res.",
  ["physics educator"] = "Phys. Educ.",
  ["physics essays"] = "Phys. Essays",
  ["physics in medicine"] = "Phys. Med.",
  ["physics in medicine & biology"] = "Phys. Med. Biol.",
  ["physics in medicine and biology"] = "Phys Med Biol",
  ["physics in perspective"] = "Phys Perspect",
  ["physics in technology"] = "Phys. Technol.",
  ["physics letters"] = "Phys. Lett.",
  ["physics letters a"] = "Phys. Lett. A",
  ["physics letters b"] = "Phys. Lett. B",
  ["physics letters. [part b]"] = "Phys Lett B",
  ["physics letters. a"] = "Phys Lett A",
  ["physics letters. b. particle physics, nuclear physics and cosmology"] = "Phys. Lett. B",
  ["physics notes"] = "Phys. Notes",
  ["physics of atomic nuclei"] = "Phys. At. Nucl.",
  ["physics of atoms and molecules"] = "Phys. Atoms and Molecules",
  ["physics of earth and space environments"] = "Phys. Earth Space Environ.",
  ["physics of fluids"] = "Phys. Fluids",
  ["physics of fluids (woodbury, n.y. : 1994)"] = "Phys Fluids (1994)",
  ["physics of fluids a"] = "Phys. Fluids A",
  ["physics of fluids a: fluid dynamics"] = "Phys. Fluids A",
  ["physics of fluids b"] = "Phys. Fluids B",
  ["physics of fluids b: plasma physics"] = "Phys. Fluids B",
  ["physics of life reviews"] = "Phys Life Rev",
  ["physics of metals"] = "Phys. Met.",
  ["physics of metals and metallography"] = "Phys. Met. Metall.",
  ["physics of metals and metallography (ussr) [translation of fizika metallovi metallovedenie]"] = "Phys. Met. Metallogr. (USSR)",
  ["physics of particles and nuclei"] = "Phys. Part. Nucl.",
  ["physics of particles and nuclei letters"] = "Phys. Part. Nucl. Lett.",
  ["physics of plasmas"] = "Phys. Plasmas",
  ["physics of the dark universe"] = "Phys. Dark Universe",
  ["physics of the earth and planetary interiors"] = "Phys. Earth Planet. Inter.",
  ["physics of the solid state"] = "Phys. Solid State",
  ["physics of wave phenomenon"] = "Phys. Wave Phenom.",
  ["physics open"] = "Phys. Open",
  ["physics physique fizika"] = "Phys. Phys. Fiz.",
  ["physics procedia"] = "Physics Procedia",
  ["physics reports"] = "Phys. Rep.",
  ["physics reports of kumamoto university"] = "Phys. Rep. Kumamoto Univ.",
  ["physics reports. a review section of physics letters"] = "Phys. Rep.",
  ["physics teacher"] = "Phys. Teach.",
  ["physics today"] = "Phys. Today",
  ["physics world"] = "Phys. World",
  ["physics world discovery"] = "Phys. World Discov.",
  ["physik der kondensierten materie"] = "Phys. Kondens. Mater.",
  ["physik in unserer zeit"] = "Phys. Unserer Zeit",
  ["physik journal"] = "Phys. J.",
  ["physikalische berichte"] = "Phys. Ber.",
  ["physikalische blätter"] = "Phys. Bl.",
  ["physikalische monographien"] = "Phys. Monogr.",
  ["physikalische zeitschrift"] = "Phys. Z.",
  ["physikalische zeitschrift der sowjetunion"] = "Phys. Z. Sowjetunion",
  ["physikalishce zeitschrift"] = "Phys. Z.",
  ["physiologia bohemoslovaca"] = "Physiol Bohemoslov",
  ["physiologia bohemoslovenica"] = "Physiol Bohemoslov",
  ["physiologia comparata et oecologia; an international journal of comparative physiology and ecology"] = "Physiol Comp Ocol Int J Comp Physiol Ecol",
  ["physiologia plantarum"] = "Physiol Plant",
  ["physiological and biochemical zoology"] = "Physiol. Biochem. Zool.",
  ["physiological and biochemical zoology : pbz"] = "Physiol Biochem Zool",
  ["physiological and molecular plant pathology"] = "Physiol. Mol. Plant Pathol.",
  ["physiological chemistry and physics"] = "Physiol Chem Phys",
  ["physiological chemistry and physics and medical nmr"] = "Physiol Chem Phys Med NMR",
  ["physiological entomology"] = "Physiol Entomol",
  ["physiological genomics"] = "Physiol Genomics",
  ["physiological measurement"] = "Physiol Meas",
  ["physiological plant pathology"] = "Physiol. Plant Pathol.",
  ["physiological reports"] = "Physiol Rep",
  ["physiological research"] = "Physiol Res",
  ["physiological research / academia scientiarum bohemoslovaca"] = "Physiol Res",
  ["physiological reviews"] = "Physiol Rev",
  ["physiological reviews. supplement"] = "Physiol Rev Suppl",
  ["physiological zoology"] = "Physiol Zool",
  ["physiologie (bucarest)"] = "Physiologie",
  ["physiologie vegetale"] = "Physiol Veg",
  ["physiologie végétale"] = "Physiol Veg",
  ["physiology & behavior"] = "Physiol Behav",
  ["physiology & pharmacology for physicians"] = "Physiol Pharmacol Physicians",
  ["physiology (bethesda, md.)"] = "Physiology (Bethesda)",
  ["physiology and behavior"] = "Physiol. Behav.",
  ["physiology and molecular biology of plants"] = "Physiol. Mol. Biol. Plants",
  ["physiology and molecular biology of plants : an international journal of functional plant biology"] = "Physiol Mol Biol Plants",
  ["physiology and pharmacology"] = "Physiol. Pharmacol.",
  ["physiology and pharmacology for physicians"] = "Physiol. Pharmacol. Physicians",
  ["physiology for physicians"] = "Physiol Physicians",
  ["physiology international"] = "Physiol Int",
  ["physiology journal"] = "Physiol J",
  ["physiology teacher"] = "Physiol. Teach.",
  ["physiotherapy canada"] = "Physiother. Can.",
  ["physiotherapy canada. physiotherapie canada"] = "Physiother Can",
  ["physiotherapy canada. physiothérapie canada"] = "Physiother Can",
  ["physiotherapy practice and research"] = "Physiother Pract Res",
  ["physiotherapy research international"] = "Physiother. Res. Int.",
  ["physiotherapy research international : the journal for researchers and clinicians in physical therapy"] = "Physiother Res Int",
  ["physiotherapy theory and practice"] = "Physiother Theory Pract",
  ["physis (rio de janeiro, brazil)"] = "Physis",
  ["physis; rivista internazionale di storia della scienza"] = "Physis Riv Int Stor Sci",
  ["physische geographie"] = "Phys. Geogr.",
  ["physis—rivista internazionale di storia della scienza"] = "Physis Riv. Internaz. Storia Sci. (N.S.)",
  ["physis—rivista internazionale di storia della scienza. nuova serie"] = "Physis Riv. Internaz. Storia Sci. (N.S.)",
  ["phytobiomes journal"] = "Phytobiomes J",
  ["phytochemical analysis"] = "Phytochem. Anal.",
  ["phytochemical analysis : pca"] = "Phytochem Anal",
  ["phytochemistry (elsevier)"] = "Phytochemistry (Elsevier)",
  ["phytochemistry letters"] = "Phytochem. Lett.",
  ["phytochemistry reviews"] = "Phytochem. Rev.",
  ["phytochemistry reviews : proceedings of the phytochemical society of europe"] = "Phytochem Rev",
  ["phytologia balcanica"] = "Phytol Balc (Sofia)",
  ["phytomedicine : international journal of phytotherapy and phytopharmacology"] = "Phytomedicine",
  ["phytomedicine plus"] = "Phytomed. Plus",
  ["phyton-international journal of experimental botany"] = "Phyton-Int. J. Exp. Bot.",
  ["phyton; annales rei botanicae"] = "Phyton",
  ["phytoparasitica; israel journal of plant protection sciences"] = "Phytoparasitica",
  ["phytopathogenic mollicutes"] = "Phytopathogenic Mollicutes",
  ["phytopathologia mediterranea"] = "Phytopathol Mediterr",
  ["phytopathologia polonica"] = "Phytopathol Pol",
  ["phytopathologische zeitschrift"] = "Phytopathol. Z.",
  ["phytopathologische zeitschrift. journal of phytopathology"] = "J Phytopathol",
  ["phytopathology research"] = "Phytopathol. Res.",
  ["phytotherapy research"] = "Phytother. Res.",
  ["phytotherapy research : ptr"] = "Phytother Res",
  ["phytothérapie (paris, france)"] = "Phytotherapie (Paris)",
  ["pi mu epsilon journal"] = "Pi Mu Epsilon J.",
  ["pi perspective"] = "PI Perspect",
  ["piact papers"] = "PIACT Pap",
  ["piact papers / program for the introduction and adaptation of contraceptive technology"] = "PIACT Pap",
  ["piact product news. program for the introduction and adaptation of contraceptive technology (u.s.)"] = "PIACT Prod News",
  ["piact product newsletter. program for the introduction and adaptation of contraceptive technology (u.s.)"] = "PIACT Prod Newsl",
  ["picus. studi e ricerche sulle marche nell’antichità"] = "Picus",
  ["pidsa abstracts"] = "PIDSA Abstr",
  ["pielegniarka i polozna"] = "Pieleg. Polozna",
  ["pielȩgniarka i połozna"] = "Pieleg Polozna",
  ["pigment and resin technology"] = "Pigm. Resin Technol.",
  ["pigment cell & melanoma research"] = "Pigment Cell Melanoma Res",
  ["pigment cell research"] = "Pigment Cell Res",
  ["pigment cell research / sponsored by the european society for pigment cell research and the international pigment cell society"] = "Pigment Cell Res",
  ["pilot and feasibility studies"] = "Pilot Feasibility Stud",
  ["pinheiros terapeutico"] = "Pinheiros Ter",
  ["pinheiros terapêutico"] = "Pinheiros Ter",
  ["pioneer america society transactions : p.a.s.t. pioneer america society"] = "Pioneer Am Soc Trans",
  ["pioneers in arts, humanities, science, engineering, practice"] = "Pioneers Arts Humanit. Sci. Eng. Pract.",
  ["pis\\cprime ma v zhurnal èksperimental\\cprime noy i teoreticheskoy fiziki"] = "Pis\\cprime ma Zh. Èksper. Teoret. Fiz.",
  ["pishchevaya promyshlennost (moscow)"] = "Pishch. Prom-st. (Moscow)",
  ["pisum genetics"] = "Pisum Genet",
  ["pis’ma v astronomicheskii zhurnal"] = "Pis’ma Astron. Zh.",
  ["pis’ma v astronomicheskii zhurnal [soviet astronomy letters]"] = "Pis’ma Astron. Zh. [Sov. Astron. Lett.]",
  ["pis’ma v zhurnal eksperimentalnoi i teoreticheskoi fiziki [jetp letters]"] = "Pis’ma Zh. Eksp. Teor. Fiz. [JETP Lett.]",
  ["pis’ma v zhurnal eksperimental’noi i teoreticheskoi fiziki"] = "Pis’ma Zh. Eksp. Teor. Fiz.",
  ["pis’ma v zhurnal tekhnicheskoi fiziki"] = "Pis’ma Zh. Tekh. Fiz.",
  ["pis’ma v zhurnal tekhnicheskoi fiziki [soviet technical physics letters]"] = "Pis’ma Zh. Tekh. Fiz. [Sov. Tech. Phys. Lett.]",
  ["pitannâ klasicnoï filologìï (questions de philologie classique)"] = "PKFil",
  ["pitman monographs and surveys in pure and applied mathematics"] = "Pitman Monogr. Surveys Pure Appl. Math.",
  ["pitman research notes in mathematics series"] = "Pitman Res. Notes Math. Ser.",
  ["pittsburgh medical bulletin"] = "Pittsbg Med Bull",
  ["pizhuhish va sazandidi"] = "Research & reconstruction|Pizhuhish Va Sazandidi",
  ["pizhuhish va sazandidi = research & reconstruction"] = "Pizhuhish Va Sazandidi",
  ["pizhūhish/hā-yi kārburdī dar giyāhpizishkī"] = "Pizhuhishhayi Karburdi Giyahpizishki",
  ["pje. peabody journal of education"] = "Peabody J Educ",
  ["placenta. supplement"] = "Placenta Suppl",
  ["plain dealer (cleveland, ohio : 1961)"] = "Plain Dealer",
  ["plains anthropologist"] = "Plains Anthropol",
  ["plan og arbeid"] = "Plan Arb",
  ["planejamento agora"] = "Planej Agora",
  ["planetary and space science"] = "Planet. Space Sci.",
  ["planetary science"] = "Planet Sci",
  ["planetary science journal"] = "Planet. Sci. J.",
  ["planfed news : newsletter of the planned parenthood federation of nigeria"] = "PLANFED news",
  ["plankton & benthos research"] = "Plankton Benthos Res",
  ["plankton and benthos research"] = "Plankton Benthos Res.",
  ["planned parenthood"] = "Planned Parent (India)",
  ["planned parenthood challenges"] = "Plan Parent Chall",
  ["planned parenthood challenges / international planned parenthood federation"] = "Plan Parent Chall",
  ["planned parenthood in europe = planning familial en europe"] = "Plan Parent Eur",
  ["planned parenthood in europe regional information bulletin = planning familial en europe bulletin d'information régional = familienplanung in europe regionale informationen"] = "IPPF Eur Reg Inf",
  ["planned parenthood review"] = "Plan Parent Rev",
  ["planned parenthood-world population washington memo"] = "Plan Parent World Popul Wash Memo",
  ["planning history"] = "Plann Hist",
  ["planning perspectives : pp"] = "Plan Perspect",
  ["plant & cell physiology"] = "Plant Cell Physiol",
  ["plant and cell physiology"] = "Plant Cell Physiol.",
  ["plant and fungal systematics"] = "Plant Fungal Syst",
  ["plant and soil"] = "Plant Soil",
  ["plant archives"] = "Plant Arch",
  ["plant biology"] = "Plant Biol",
  ["plant biology (stuttgart, germany)"] = "Plant Biol (Stuttg)",
  ["plant biosystems"] = "Plant Biosyst",
  ["plant biosystems - an international journal dealing with all aspects of plant biology"] = "Plant Biosyst. - Int. J. Dealing Aspects Plant Biosyst.",
  ["plant biotechnology"] = "Plant Biotechnol.",
  ["plant biotechnology (sheffield, england)"] = "Plant Biotechnol",
  ["plant biotechnology (tokyo, japan)"] = "Plant Biotechnol (Tokyo)",
  ["plant biotechnology for sustainable production of energy and co-products"] = "Biotech. Agr. Forest.",
  ["plant biotechnology journal"] = "Plant Biotechnol J",
  ["plant biotechnology reports"] = "Plant Biotechnol Rep",
  ["plant breeding"] = "Plant Breed.",
  ["plant breeding = zeitschrift für pflanzenzüchtung"] = "Plant Breed",
  ["plant breeding and biotechnology"] = "Plant Breed Biotechnol",
  ["plant breeding reviews"] = "Plant Breed Rev",
  ["plant cell"] = "Plant Cell",
  ["plant cell and environment"] = "Plant Cell Environ.",
  ["plant cell biotechnology and molecular biology"] = "Plant Cell Biotechnol Mol Biol",
  ["plant cell reports"] = "Plant Cell Rep",
  ["plant cell, tissue and organ culture"] = "Plant Cell Tissue Organ Cult",
  ["plant communications"] = "Plant Commun",
  ["plant direct"] = "Plant Direct",
  ["plant disease"] = "Plant Dis",
  ["plant disease management reports : pdmr"] = "Plant Dis Manag Rep",
  ["plant diversity"] = "Plant Divers",
  ["plant diversity and evolution"] = "Plant Divers Evol",
  ["plant diversity and resources"] = "Plant Diversity Resour",
  ["plant ecology"] = "Plant Ecol",
  ["plant ecology & diversity"] = "Plant Ecol Divers",
  ["plant ecology and diversity"] = "Plant Ecolog. Divers.",
  ["plant ecology and evolution"] = "Plant Ecol Evol",
  ["plant foods for human nutrition"] = "Plant Foods Hum. Nutr.",
  ["plant foods for human nutrition (dordrecht, netherlands)"] = "Plant Foods Hum Nutr",
  ["plant gene"] = "Plant Gene",
  ["plant gene & trait"] = "Plant Gene Trait",
  ["plant genetic resources"] = "Plant Genet. Resour.",
  ["plant genetic resources : characterization and utilization"] = "Plant Genet Resour",
  ["plant genetic resources newsletter (rome, italy : 1979)"] = "Plant Genet Resour Newsl",
  ["plant genome"] = "Plant Genome",
  ["plant growth regulation"] = "Plant Growth Regul",
  ["plant health cases"] = "Plant Health Cases",
  ["plant health progress"] = "Plant Health Prog",
  ["plant journal"] = "Plant J.",
  ["plant methods"] = "Plant Methods",
  ["plant molecular biology"] = "Plant Mol Biol",
  ["plant molecular biology reporter"] = "Plant Mol Biol Report",
  ["plant molecular biology reporter / ispmb"] = "Plant Mol Biol Rep",
  ["plant omics"] = "Plant Omics",
  ["plant pathology"] = "Plant Pathol",
  ["plant pathology & quarantine journal of fungal biology"] = "Plant Pathol Quar J Fungal Biolog",
  ["plant pathology journal"] = "Plant Pathol J (Faisalabad)",
  ["plant phenome journal"] = "Plant Phenome J.",
  ["plant phenomics"] = "Plant Phenomics",
  ["plant phenomics (washington, d.c.)"] = "Plant Phenomics",
  ["plant physiology"] = "Plant Physiol",
  ["plant physiology and biochemistry"] = "Plant Physiol. Biochem.",
  ["plant physiology and biochemistry (amsterdam, netherlands)"] = "Plant Physiol. Biochem. (Amsterdam, Neth.)",
  ["plant physiology and biochemistry (issy les moulineaux, france)"] = "Plant Physiol. Biochem. (Issy les Moulineaux, Fr.)",
  ["plant physiology and biochemistry (issy-les-moulineaux, france)"] = "Plant Physiol. Biochem. (Issy-les-Moulineaux, Fr.)",
  ["plant physiology and biochemistry : ppb"] = "Plant Physiol Biochem",
  ["plant physiology and biochemistry : ppb / societe francaise de physiologie vegetale"] = "Plant Physiol Biochem",
  ["plant physiology reports"] = "Plant Physiol. Rep.",
  ["plant physiology. supplement"] = "Plant Physiol Suppl (Bethesda)",
  ["plant production science"] = "Plant Prod Sci",
  ["plant protection quarterly"] = "Plant Prot Q",
  ["plant protection science"] = "Plant Prot. Sci.",
  ["plant protection science = ochrana rostlin"] = "Plant Prot Sci",
  ["plant reproduction"] = "Plant Reprod",
  ["plant science"] = "Plant Sci.",
  ["plant science (amsterdam, netherlands)"] = "Plant Sci. (Amsterdam, Neth.)",
  ["plant science (limerick, ireland)"] = "Plant Sci. (Limerick, Irel.)",
  ["plant science (shannon, ireland)"] = "Plant Sci. (Shannon, Irel.)",
  ["plant science : an international journal of experimental plant biology"] = "Plant Sci",
  ["plant science journal"] = "Plant Sci J",
  ["plant science letters"] = "Plant Sci Lett",
  ["plant signaling & behavior"] = "Plant Signal Behav",
  ["plant sociology"] = "Plant Sociol.",
  ["plant species biology"] = "Plant Species Biol",
  ["plant stress"] = "Plant Stress",
  ["plant systematics and evolution"] = "Plant Syst. Evol.",
  ["plant systematics and evolution = entwicklungsgeschichte und systematik der pflanzen"] = "Plant Syst Evol",
  ["plant systematics and evolution. supplementum"] = "Entwicklungsgeschichte und Systematik der Pflanzen. Supplementum|Plant Sys. Evol., Suppl.",
  ["plant systematics and evolution. supplementum = entwicklungsgeschichte und systematik der pflanzen. supplementum"] = "Plant Sys Evol Suppl",
  ["plant tissue culture & biotechnology"] = "Plant Tissue Cult Biotechnol",
  ["plant viruses"] = "Plant Viruses",
  ["plant, cell & environment"] = "Plant Cell Environ",
  ["plant, cell and environment"] = "Plant Cell Environ.",
  ["plant, soil and environment"] = "Plant Soil Environ",
  ["plant-environment interactions"] = "Plant-Environ. Interact.",
  ["plant/operations progress"] = "Plant/Oper. Prog.",
  ["planta medica"] = "Planta Med",
  ["plantation health"] = "Plant Health",
  ["plants (basel, switzerland)"] = "Plants (Basel)",
  ["plants, people, planet,"] = "Plants People Planet",
  ["plantsman (london, england : 2002)"] = "Plantsman",
  ["plasma chemistry and plasma processing"] = "Plasma Chem. Plasma Process.",
  ["plasma devices and operations"] = "Plasma Devices Oper.",
  ["plasma medicine"] = "Plasma Med",
  ["plasma physics"] = "Plasma Phys.",
  ["plasma physics and controlled fusion"] = "Plasma Phys. Controlled Fusion",
  ["plasma physics and controlled nuclear fusion research"] = "Plasma Phys. Controlled Nucl. Fusion Res.",
  ["plasma physics reports"] = "Plasma Phys. Rep.",
  ["plasma processes and polymers"] = "Plasma Processes Polym.",
  ["plasma processes and polymers (print)"] = "Plasma Process Polym",
  ["plasma research express"] = "Plasma Res. Express",
  ["plasma science and technology"] = "Plasma Sci. Technol",
  ["plasma sources science & technology"] = "Plasma Sources Sci Technol",
  ["plasma sources science and technology"] = "Plasma Sources Sci. Technol.",
  ["plasmas and polymers"] = "Plasmas Polym.",
  ["plasmonics (norwell, mass.)"] = "Plasmonics",
  ["plastic and aesthetic research"] = "Plast Aesthet Res",
  ["plastic and reconstructive surgery"] = "Plast Reconstr Surg",
  ["plastic and reconstructive surgery (1946)"] = "Plast Reconstr Surg (1946)",
  ["plastic and reconstructive surgery and the transplantation bulletin"] = "Plast Reconstr Surg Transplant Bull",
  ["plastic and reconstructive surgery. global open"] = "Plast Reconstr Surg Glob Open",
  ["plastic surgery (oakville, ont.)"] = "Plast Surg (Oakv)",
  ["plastic surgery international"] = "Plast Surg Int",
  ["plastic surgical nursing"] = "Plast. Surg. Nurs.",
  ["plastic surgical nursing : official journal of the american society of plastic and reconstructive surgical nurses"] = "Plast Surg Nurs",
  ["plasticheskie massy"] = "Plast. Massy",
  ["plastics engineering"] = "Plast. Eng.",
  ["plastics rubber and composites processing and applications"] = "Plast. Rubber Compos. Process. Appl.",
  ["plastics, rubber and composites"] = "Plast., Rubber Compos.",
  ["platax = hai yang sheng wu xue kan : an official publication of the national museum of marine biology & aquarium"] = "Platax",
  ["plating and surface finishing"] = "Plat. Surf. Finish.",
  ["platinum metals review"] = "Platinum Met. Rev.",
  ["pleura (thousand oaks, ventura county, calif.)"] = "Pleura (Thousand Oaks)",
  ["pleura and peritoneum"] = "Pleura Peritoneum",
  ["pliska studia mathematica"] = "Pliska Stud. Math.",
  ["pliska. pliiska"] = "Pliska",
  ["plos biology"] = "PLoS Biol",
  ["plos climate"] = "PLOS Clim.",
  ["plos clinical trials"] = "PLoS Clin Trials",
  ["plos complex systems"] = "PLOS Complex Syst.",
  ["plos computational biology"] = "PLoS Comput Biol",
  ["plos currents"] = "PLoS Curr",
  ["plos currents: disasters"] = "PLOS Curr.: Disasters",
  ["plos currents: evidence on genomic tests"] = "PLOS Curr.:  Evidence Genomic Tests",
  ["plos currents: huntington disease"] = "PLOS Curr.: Huntington Dis.",
  ["plos currents: muscular dystrophy"] = "PLOS Curr.: Muscular Dystrophy",
  ["plos currents: outbreaks"] = "PLOS Curr.: Outbreaks",
  ["plos currents: tree of life"] = "PLOS Curr.: Tree Life",
  ["plos digital health"] = "PLOS Digital Health",
  ["plos genetics"] = "PLoS Genet",
  ["plos global public health"] = "PLOS Global Public Health",
  ["plos medicine"] = "PLoS Med",
  ["plos neglected tropical diseases"] = "PLoS Negl Trop Dis",
  ["plos one"] = "PLoS One",
  ["plos pathogens"] = "PLoS Pathog",
  ["plos sustainability and transformation"] = "PLOS Sustainability Transform.",
  ["plos water"] = "PLOS Water",
  ["plovdivski universitet “paisii khilendarski”. nauchni trudove. matematika"] = "Plovdiv. Univ. Paisii Khilendarski Nauchn. Trud. Mat.",
  ["plucne bolesti"] = "Plucne Bolesti",
  ["plucne bolesti i tuberkuloza"] = "Plucne Bolesti Tuberk.",
  ["plućne bolesti : casopis udruzenja pneumoftiziologa jugoslavije = the journal of yugoslav association of phthisiology and pneumology"] = "Plucne Bolesti",
  ["plućne bolesti i tuberkuloza"] = "Plucne Bolesti Tuberk",
  ["plugger (waterloo, ia.)"] = "Plugger",
  ["plugger; iowa dental assistants journal"] = "Plugger.",
  ["plural societies"] = "Plural Soc",
  ["plus de lumière"] = "Plus Lumière",
  ["plzensky lekarsky sbornik"] = "Plzen Lek Sb",
  ["plzen̆ský lékar̆ský sborník"] = "Plzen Lek Sb",
  ["pm & r : the journal of injury, function, and rehabilitation"] = "PM R",
  ["pmc biophysics"] = "PMC Biophys",
  ["pmc physics a"] = "PMC Phys. A",
  ["pmc physics b"] = "PMC Phys. B",
  ["pmse preprints"] = "PMSE Prepr.",
  ["pmse preprints. american chemical society. division of polymeric materials: science and engineering. meeting"] = "PMSE Preprints",
  ["pneumo news"] = "Pneumo News",
  ["pneumoftiziologia : revista societatii romane de pneumoftiziologie / [societatea romana de pneumoftiziologie]"] = "Pneumoftiziologia",
  ["pneumoftiziologia : revista societătii române de pneumoftiziologie"] = "Pneumoftiziologia",
  ["pneumologia (bucharest, romania)"] = "Pneumologia",
  ["pneumologie (stuttgart, germany)"] = "Pneumologie",
  ["pneumonia (nathan qld.)"] = "Pneumonia (Nathan)",
  ["pneumonologia danubiana"] = "Pneumonol Danub",
  ["pneumonologia i alergologia polska"] = "Pneumonol Alergol Pol",
  ["pneumonologia i alergologia polska : organ polskiego towarzystwa ftyzjopneumonologicznego, polskiego towarzystwa alergologicznego, i instytutu gruzlicy i chorob pluc"] = "Pneumonol Alergol Pol",
  ["pneumonologia polska"] = "Pneumonol Pol",
  ["pneumonologie. pneumonology"] = "Pneumonologie",
  ["pneumonologike kai phymatiologike epitheoresis"] = "Pneumonol Phymatiologike Epitheor",
  ["pnlde subseries in control"] = "PNLDE Subser. Control",
  ["poblacion (consejo nacional de poblacion (peru) : 1995)"] = "Poblacion",
  ["poblacion y desarrollo (1991)"] = "Poblac Desarro",
  ["población (consejo nacional de población (peru) : 1995)"] = "Poblacion",
  ["población y desarrollo (1991)"] = "Poblac Desarro",
  ["población y salud en mesoamérica"] = "Poblac Salud Mesoam",
  ["podiplomski seminar iz matematike"] = "Podiplomski Sem. Mat.",
  ["podstawowe algorytmy numeryczne"] = "Podstaw. Algorytmy Numer.",
  ["podstawowe nauki techniczne"] = "Podstaw. Nauki Tech. Monogr.",
  ["poetics (hague, netherlands)"] = "Poetics (Amst)",
  ["poggendorff’s annalen der physik"] = "Poggendorff’s Ann. Phys.",
  ["pogon sahoe yon’gu"] = "Pogon Sahoe Yongu",
  ["pogŏn sahoe nonjip = journal of population, health, and social welfare"] = "Bogeon sahoe nonjib",
  ["pogŏn sahoe yŏn'gu"] = "Pogon Sahoe Yongu",
  ["poiesis & praxis : international journal of ethics of science and technology assessment"] = "Poiesis Prax",
  ["poincare journal of analysis and applications"] = "Poincare J. Anal. Appl.",
  ["point of care"] = "Point Care",
  ["poj nursing practice & research"] = "POJ Nurs Pract Res",
  ["polar biology"] = "Polar Biol",
  ["polar bioscience"] = "Polar Biosci",
  ["polar geography"] = "Polar Geogr.",
  ["polar geography (palm beach, fla.)"] = "Polar Geogr (Palm Beach)",
  ["polar journal"] = "Polar J.",
  ["polar record"] = "Polar Rec.",
  ["polar research"] = "Polar Res",
  ["polar science"] = "Polar Sci.",
  ["polarnet technical report"] = "Polarnet Tech Rep",
  ["polevye archeologičeskie issledovanija"] = "PAI",
  ["police practice & research : an international journal"] = "Police Pract Res",
  ["police quarterly"] = "Police Q",
  ["policing & society"] = "Policing Soc",
  ["policing (bradford, england)"] = "Policing",
  ["policlinico infantile"] = "Policlin Infant",
  ["policlinico. sezione chirurgica"] = "Policlinico. [Chir.]",
  ["policlinico. sezione medica"] = "Policlinico. [Med.]",
  ["policlinico. sezione pratica"] = "Policlinico. [Prat.]",
  ["policlinico; sezione practica"] = "Policlinico [Prat]",
  ["policy & internet"] = "Policy Internet",
  ["policy & society"] = "Policy Soc",
  ["policy analysis"] = "Policy Anal",
  ["policy analysis brief. h series"] = "Policy Anal Brief H Ser",
  ["policy analysis brief. h series / project hope, center for health affairs"] = "Policy Anal Brief H Ser",
  ["policy analysis brief. w series"] = "Policy Anal Brief W Ser",
  ["policy analysis brief. w series / project hope, walsh center for rural health analysis"] = "Policy Anal Brief W Ser",
  ["policy and politics"] = "Policy Polit",
  ["policy brief (alexandria, va.)"] = "Policy Brief Health Care Technol Inst",
  ["policy brief (center for home care policy and research (u.s.))"] = "Policy Brief (Cent Home Care Policy Res)",
  ["policy brief (commonwealth fund)"] = "Policy Brief Commonw Fund",
  ["policy brief (george washington university. center for health services research and policy)"] = "Policy Brief George Wash Univ Cent Health Serv Res Policy",
  ["policy brief (ucla center for health policy research)"] = "Policy Brief UCLA Cent Health Policy Res",
  ["policy insights from the behavioral and brain sciences"] = "Policy Insights Behav Brain Sci",
  ["policy matters"] = "Policy Matters Inst Res Public Policy",
  ["policy review"] = "Policy Rev",
  ["policy sciences"] = "Policy Sci",
  ["policy statement / royal college of general practitioners. royal college of general practitioners"] = "Policy Statement R Coll Gen Pract",
  ["policy statement, royal college of general practitioners"] = "Policy Statement R. Coll. Gen. Pract.",
  ["policy statement. royal college of general practitioners"] = "Policy Statement R Coll Gen Pract",
  ["policy studies"] = "Pol. Stud.",
  ["policy studies (policy studies institute)"] = "Policy Stud",
  ["policy studies journal: the journal of the policy studies organization"] = "Policy Stud J",
  ["policy studies review"] = "Policy Stud Rev",
  ["policy, politics & nursing practice"] = "Policy Polit Nurs Pract",
  ["polimery (warsaw)"] = "Polimery (Warsaw)",
  ["polimery w medycynie"] = "Polim Med",
  ["polin (series)"] = "Polin",
  ["polio news"] = "Polio news",
  ["polis (bologna, italy)"] = "Polis",
  ["polis. revista de ideas y formas políticas de la antigüedad clásica"] = "Polis",
  ["polish academy of sciences"] = "Acta Phys. Polon. A",
  ["polish academy of sciences. institute of fundamental technological research. archives of mechanics (archiwum mechaniki stosowanej)"] = "Arch. Mech. (Arch. Mech. Stos.)",
  ["polish american studies"] = "Pol Am Stud",
  ["polish archaeology in the mediterranean"] = "PolAMed",
  ["polish archives of internal medicine"] = "Pol Arch Intern Med",
  ["polish botanical journal"] = "Pol Bot J",
  ["polish contemporary philosophy and philosophical humanities"] = "Pol. Contemp. Philos. Philos. Humanit.",
  ["polish ecological studies"] = "Pol Ecol Stud",
  ["polish hyperbaric research"] = "Pol Hyperb Res",
  ["polish journal of chemical technology"] = "Pol. J. Chem. Technol.",
  ["polish journal of chemistry"] = "Pol J Chem",
  ["polish journal of ecology"] = "Pol J Ecol",
  ["polish journal of entomology"] = "Pol. J. Entomol.",
  ["polish journal of environmental studies"] = "Pol J Environ Stud",
  ["polish journal of food and nutrition sciences"] = "Pol J Food Nutr Sci",
  ["polish journal of medical physics and engineering"] = "Pol. J. Med. Phys. Eng.",
  ["polish journal of microbiology"] = "Pol J Microbiol",
  ["polish journal of natural science"] = "Pol. J. Nat. Sci.",
  ["polish journal of occupational medicine"] = "Pol J Occup Med",
  ["polish journal of occupational medicine and environmental health"] = "Pol J Occup Med Environ Health",
  ["polish journal of pathology"] = "Pol. J. Pathol.",
  ["polish journal of pathology : official journal of the polish society of pathologists"] = "Pol J Pathol",
  ["polish journal of pharmacology"] = "Pol J Pharmacol",
  ["polish journal of pharmacology and pharmacy"] = "Pol J Pharmacol Pharm",
  ["polish journal of radiology"] = "Pol J Radiol",
  ["polish journal of soil science"] = "Pol. J. Soil Sci.",
  ["polish journal of veterinary sciences"] = "Pol J Vet Sci",
  ["polish journal of zoology"] = "Pol. J. Zool.",
  ["polish maritime research"] = "Pol. Marit. Res.",
  ["polish medical history and science bulletin"] = "Pol Med Hist Sci Bull",
  ["polish medical journal"] = "Pol Med J",
  ["polish medical science and history bulletin"] = "Pol Med Sci Hist Bull",
  ["polish medical sciences and history bulletin"] = "Pol Med Sci Hist Bull",
  ["polish orthopedics and traumatology"] = "Pol Orthop Traumatol",
  ["polish perspectives"] = "Pol Perspect",
  ["polish polar research"] = "Pol Polar Res",
  ["polish population review"] = "Pol Popul Rev",
  ["polish population review / polish demographic society [and] central statistical office"] = "Pol Popul Rev",
  ["polish western affairs"] = "Pol West Aff",
  ["politechnika poznańska"] = "Rozprawy Politech. Poznań",
  ["politechnika wrocławska"] = "Prace Nauk. Inst. Cybernet. Tech. Politech. Wrocław. Ser. Monograf.",
  ["politica economica"] = "Politica Econ.",
  ["politica internazionale"] = "Polit Int",
  ["politica y sociedad (madrid, spain)"] = "Polit Soc (Madrid)",
  ["political affairs"] = "Polit Aff",
  ["political analysis : an annual publication of the methodology section of the american political science association"] = "Polit Anal",
  ["political and legal anthropology review : polar"] = "Polit Leg Anthropol Rev",
  ["political behavior"] = "Polit Behav",
  ["political communication"] = "Polit Commun",
  ["political geography"] = "Polit Geogr",
  ["political psychology"] = "Polit Psychol",
  ["political research quarterly"] = "Polit Res Q",
  ["political science (wellington, n.z.)"] = "Polit Sci",
  ["political science quarterly"] = "Polit Sci Q",
  ["political science research and methods"] = "Political Sci Res Methods",
  ["political studies"] = "Polit Stud (Oxf)",
  ["political theory"] = "Polit Theory",
  ["politicka ekonomie"] = "Polit Ekon",
  ["politická ekonomie"] = "Polit Ekon",
  ["politická ekonomie"] = "Politická Ekon.",
  ["politico (pavia, italy)"] = "Politico (Pavia)",
  ["politics & policy"] = "Polit. Policy",
  ["politics & policy (statesboro, ga.)"] = "Polit Policy",
  ["politics & society"] = "Polit Soc",
  ["politics (kensington, n.s.w.)"] = "Politics (Syd)",
  ["politics and religion"] = "Politics Relig",
  ["politics and the life sciences"] = "Polit. Life Sci.",
  ["politics and the life sciences : the journal of the association for politics and the life sciences"] = "Politics Life Sci",
  ["politics philosophy & economics"] = "Polit Philos Econ",
  ["politics today"] = "Polit Today",
  ["politics, groups & identities"] = "Polit Groups Identities",
  ["politique africaine (paris, france : 1981)"] = "Polit Afr",
  ["politique etrangere"] = "Polit Etrang",
  ["politique étrangère"] = "Polit Etrang",
  ["politique internationale"] = "Polit Intern",
  ["politiques de population"] = "Politiq Popul",
  ["politische studien"] = "Polit Stud",
  ["politische vierteljahresschrift"] = "Polit Vierteljahresschr",
  ["politische ökologie"] = "Polit. Ökol.",
  ["polito springer series"] = "PoliTO Springer Ser.",
  ["política y sociedad (madrid, spain)"] = "Polit Soc (Madrid)",
  ["poljoprivredna znanstvena smotra = agriculturae conspectus scientificus : acs"] = "Poljopr Znan Smotra",
  ["pollen et spores"] = "Pollen Spores",
  ["pollution atmospherique"] = "Pollut. Atmos.",
  ["pollution research"] = "Pollut Res",
  ["polski merkuriusz lekarski"] = "Pol. Merkuriusz Lek.",
  ["polski merkuriusz lekarski : organ polskiego towarzystwa lekarskiego"] = "Pol Merkur Lekarski",
  ["polski przeglad chirurgiczny"] = "Pol Przegl Chir",
  ["polski przeglad radiologii"] = "Pol Przegl Radiol",
  ["polski przeglad radiologii / polskie lekarskie towarzystwo radiologiczne"] = "Pol Przegl Radiol",
  ["polski przeglad radiologii i medycyny nuklearnej"] = "Pol. Przegl. Radiol. Med. Nukl.",
  ["polski przegla̜d radiologii i medycyny nuklearnej"] = "Pol Przegl Radiol Med Nukl",
  ["polski przegla̧d radiologiczny"] = "Pol Prz Radiol",
  ["polski tygodnik lekarski"] = "Pol Tyg Lek (Wars)",
  ["polski tygodnik lekarski (warsaw, poland : 1960)"] = "Pol Tyg Lek",
  ["polskie archiwum medycyny wewnetrznej"] = "Pol Arch Med Wewn",
  ["polskie archiwum medycyny wewnętrznej"] = "Pol Arch Med Wewn",
  ["polskie archiwum weterynaryjne"] = "Pol Arch Weter",
  ["polskie towarzystwo cybernetyczne"] = "Post\\polhk epy Cybernet.",
  ["polskiego towarzystwa matematycznego"] = "Comment. Math. Prace Mat.",
  ["polycyclic aromatic compounds"] = "Polycyclic Aromat. Compd.",
  ["polymer (korea)"] = "Polymer (Korea)",
  ["polymer - plastics technology and engineering"] = "Polym. Plast. Technol. Eng.",
  ["polymer bulletin"] = "Polym. Bull.",
  ["polymer bulletin (berlin)"] = "Polym. Bull. (Berlin)",
  ["polymer bulletin (berlin, germany)"] = "Polym Bull (Berl)",
  ["polymer bulletin (heidelberg, germany)"] = "Polym. Bull. (Heidelberg, Ger.)",
  ["polymer chemistry"] = "Polym. Chem.",
  ["polymer composites"] = "Polym. Compos.",
  ["polymer crystallization"] = "Polym. Cryst.",
  ["polymer degradation and stability"] = "Polym. Degrad. Stab.",
  ["polymer engineering & science"] = "Polym. Eng. Sci.",
  ["polymer engineering and science"] = "Polym. Eng. Sci.",
  ["polymer gels and networks"] = "Polym. Gels Networks",
  ["polymer international"] = "Polym. Int.",
  ["polymer journal"] = "Polym. J.",
  ["polymer journal (tokyo, japan)"] = "Polym. J. (Tokyo, Jpn.)",
  ["polymer networks and blends"] = "Polym. Networks Blends",
  ["polymer news"] = "Polym. News",
  ["polymer photochemistry"] = "Polym. Photochem.",
  ["polymer preprints (american chemical society, division of polymer chemistry)"] = "Polym. Prepr. (Am. Chem. Soc., Div. Polym. Chem.)",
  ["polymer reaction engineering"] = "Polym. React. Eng.",
  ["polymer recycling"] = "Polym. Recycl.",
  ["polymer reviews"] = "Polym. Rev.",
  ["polymer reviews (philadelphia, pa, united states)"] = "Polym. Rev. (Philadelphia, PA, U. S.)",
  ["polymer reviews (philadelphia, pa.)"] = "Polym Rev (Phila Pa)",
  ["polymer science and technology"] = "Polymer Sci. Tech.",
  ["polymer science series a - polymer physics"] = "Polym. Sci. Ser. A Polym. Phys.",
  ["polymer science series b - polymer chemistry"] = "Polym. Sci. Ser. B Polym. Chem.",
  ["polymer science series c - selected topics"] = "Polym. Sci. Ser. C Sel. Top.",
  ["polymer science series d - glues and sealing materials"] = "Polym. Sci. Ser. D Glues Sealing Mater.",
  ["polymer science, series a"] = "Polym. Sci., Ser. A",
  ["polymer science, series b"] = "Polym. Sci., Ser. B",
  ["polymer science, series c"] = "Polym. Sci., Ser. C",
  ["polymer science, series d"] = "Polym. Sci., Ser. D",
  ["polymer science. series a, chemistry, physics"] = "Polym Sci Ser A Chem Phys",
  ["polymer testing"] = "Polym. Test.",
  ["polymer-plastics technology and engineering"] = "Polym Plast Technol Eng",
  ["polymer-plastics technology and materials"] = "Polym.-Plast. Technol. Mater.",
  ["polymeric materials science and engineering"] = "Polym. Mater. Sci. Eng.",
  ["polymeric preprints (american chemical society, division of polymer chemistry)"] = "Polym. Prepr. (Am. Chem. Soc., Div. Polym. Chem.)",
  ["polymers & polymer composites"] = "Polym. Polym. Compos.",
  ["polymers (basel, switzerland)"] = "Polymers (Basel, Switz.)",
  ["polymers and polymer composites"] = "Polym. Polym. Compos.",
  ["polymers for advanced technologies"] = "Polym. Adv. Technol.",
  ["polymers from renewable resources"] = "Polym. Renewable Resour.",
  ["polytechnica posnaniensis"] = "Fasc. Math.",
  ["polytechnica posnaniensis. institutum mathematicum. fasciculi mathematici"] = "Fasc. Math.",
  ["pomeranian journal of life sciences"] = "Pomeranian J Life Sci",
  ["pompei: pitture e mosaici. enciclopedia dell’arte antica classica e orientale"] = "PPM",
  ["pompeii, herculaneum, stabiae. bolletino. associazione internazionale amici di pompei"] = "PompHercStab",
  ["pontica. studii şi materiale de istorie, arheologie şi muzeografie, constanţa"] = "Pontica",
  ["pontificae academiae scientiarum scripta varia"] = "Pontif. Acad. Sci. Scr. Varia",
  ["pop sahel : bulletin d'information sur la population et le développement"] = "Pop Sahel",
  ["pop sahel : bulletin d’information sur la population et le developpement"] = "Pop Sahel",
  ["popcen news letter. population centre (lucknow, india)"] = "POPCEN Newsl",
  ["popin bulletin"] = "POPIN Bull",
  ["popin bulletin / international population information network"] = "POPIN Bull",
  ["populacao & desenvolvimento"] = "Popul Desenvolv",
  ["populacni zpravy"] = "Popul Zpr",
  ["populační zprávy"] = "Popul Zpr",
  ["população & desenvolvimento"] = "Popul Desenvolv",
  ["popular government"] = "Pop Gov",
  ["population (new york, n.y.)"] = "Population (NY)",
  ["population and community biology"] = "Population Community Biol.",
  ["population and development review"] = "Popul Dev Rev",
  ["population and economics"] = "Popul Econ",
  ["population and environment"] = "Popul Environ",
  ["population and environmental psychology newsletter"] = "Popul Environ Psychol Newsl",
  ["population briefs : reports on population council research"] = "Popul Briefs",
  ["population briefs : reports on population council research,"] = "Popul Briefs",
  ["population bulletin"] = "Popul Bull",
  ["population bulletin of ecwa"] = "Popul Bull ECWA",
  ["population bulletin of escwa"] = "Popul Bull ESCWA",
  ["population bulletin of the united nations"] = "Popul Bull UN",
  ["population bulletin of the united nations economic commission for western asia"] = "Popul Bull U N Econ Comm West Asia",
  ["population chronicle"] = "Popul Chron",
  ["population communication: technical documentation"] = "Popul Commun Tech Doc",
  ["population concern news"] = "Popul Concern News",
  ["population data information service"] = "Popul Data Inf Serv",
  ["population dynamics quarterly"] = "Popul Dyn Q",
  ["population ecology"] = "Popul Ecol",
  ["population education in asia and oceania newsletter"] = "Popul Educ Asia Ocean Newsl",
  ["population education in asia and the pacific newsletter"] = "Popul Educ Asia Pac Newsl Forum",
  ["population education interchange"] = "Popul Educ Interchange",
  ["population et avenir"] = "Popul Avenir",
  ["population et famille"] = "Popul Famille",
  ["population et societes; bulletin mensuel d’informations demographiques, economiques, sociales"] = "Popul Soc (Paris)",
  ["population et sociétés; bulletin mensuel d'informations démographiques, économiques, sociales"] = "Popul Soc (Paris)",
  ["population forum : monthly newsletter of the commission on population"] = "Popul Forum",
  ["population geography : a journal of the association of population geographers of india"] = "Popul Geogr",
  ["population headliners"] = "Popul Headl",
  ["population health management"] = "Popul Health Manag",
  ["population health metrics"] = "Popul Health Metr",
  ["population index"] = "Popul Index",
  ["population manager : icomp review"] = "Popul Manag",
  ["population manager : icomp review / icomp,"] = "Popul Manag",
  ["population medicine"] = "Popul Med",
  ["population newsletter"] = "Popul Newsl",
  ["population newsletter / issued by the population division of the department of economic and social affairs, united nations"] = "Popul Newsl",
  ["population notes (university of wisconsin--madison. applied population laboratory)"] = "Popul Notes",
  ["population policy compendium"] = "Popul Policy Compend",
  ["population report"] = "Popul Rep",
  ["population report (honolulu, hawaii)"] = "Popul Rep",
  ["population reports (washington, d.c.). series c, female sterilization"] = "Popul Rep C",
  ["population reports (washington, d.c.). series c,, female sterilization"] = "Popul Rep C",
  ["population reports (washington, d.c.). series d, male sterilization"] = "Popul Rep D",
  ["population reports (washington, d.c.). series d, sterilization"] = "Popul Rep D",
  ["population reports (washington, d.c.). series d, sterilization."] = "Popul Rep D",
  ["population reports. series a, oral contraceptives"] = "Popul Rep A",
  ["population reports. series a: oral contraceptives"] = "Popul Rep A",
  ["population reports. series a; oral contraceptives"] = "Popul Rep A",
  ["population reports. series b, intrauterine devices"] = "Popul Rep B",
  ["population reports. series b: intrauterine devices"] = "Popul Rep B",
  ["population reports. series b; intrauterine devices"] = "Popul Rep B",
  ["population reports. series c, female sterilization"] = "Popul. Rep. C",
  ["population reports. series c, sterilization female"] = "Popul. Rep. C",
  ["population reports. series c: sterilization [female]"] = "Popul Rep C",
  ["population reports. series d, male sterilization"] = "Popul. Rep. D",
  ["population reports. series d, sterilization male"] = "Popul. Rep. D",
  ["population reports. series e, law and policy"] = "Popul Rep E",
  ["population reports. series e; law and policy"] = "Popul Rep E",
  ["population reports. series f, pregnancy termination"] = "Popul Rep F",
  ["population reports. series f: pregnancy termination"] = "Popul Rep F",
  ["population reports. series g, prostaglandins"] = "Popul Rep G",
  ["population reports. series h, barrier methods"] = "Popul Rep H",
  ["population reports. series h: barrier methods"] = "Popul Rep H",
  ["population reports. series h; barrier methods"] = "Popul Rep H",
  ["population reports. series i, periodic abstinence"] = "Popul Rep I",
  ["population reports. series i: periodic abstinence"] = "Popul Rep I",
  ["population reports. series j, family planning programs"] = "Popul Rep J",
  ["population reports. series j: family planning programs"] = "Popul Rep J",
  ["population reports. series k, injectables and implants"] = "Popul Rep K",
  ["population reports. series k: injectables and implants"] = "Popul Rep K",
  ["population reports. series l, issues in world health"] = "Popul Rep L",
  ["population reports. series l: issues in world health"] = "Popul Rep L",
  ["population reports. series m, special topic monographs"] = "Popul Rep M",
  ["population reports. series m, special topics"] = "Popul Rep M",
  ["population reports. series m. special topics"] = "Popul Rep M",
  ["population reports. special topic monographs"] = "Popul Rep Spec Top Monogr",
  ["population research (peking, china)"] = "Popul Res",
  ["population research abstract : a contribution to research on india's population"] = "Popul Res Abstr",
  ["population research abstract : a contribution to research on india’s population"] = "Popul Res Abstr",
  ["population research and policy review"] = "Popul Res Policy Rev",
  ["population research bulletin. pakistan institute of development economics"] = "Popul Res Bull",
  ["population research leads"] = "Popul Res Leads",
  ["population researches and studies"] = "Popul Res Stud",
  ["population review"] = "Popul Rev",
  ["population sciences (cairo, egypt)"] = "Popul Sci",
  ["population studies"] = "Popul Stud (Camb)",
  ["population studies of japan"] = "Popul. Stud. Japan",
  ["population studies translation series"] = "Popul Stud Transl Ser",
  ["population talk point"] = "Popul Talk Point",
  ["population today"] = "Popul Today",
  ["population trends"] = "Popul Trends",
  ["population trends and public policy"] = "Popul Trends Public Policy",
  ["population, space and place"] = "Popul Space Place",
  ["population. english selection"] = "Popul",
  ["porcine health management"] = "Porcine Health Manag",
  ["porcine research"] = "Porcine Res",
  ["poromechanics v : proceedings of the fifth biot conference on poromechanics, july 10-12, 2013, vienna, austria. biot conference on poromechanics (5th : 2013 : vienna, austria)"] = "Poromechanics V (2013)",
  ["poroshkovaya metallurgiya (kiev)"] = "Poroshk. Metall. (Kiev)",
  ["porto biomedical journal"] = "Porto Biomed J",
  ["portugal medico"] = "Port Med",
  ["portugal médico"] = "Port Med",
  ["portugaliae acta biologica. série b, sistemática, ecologia, biogeografia e paleontologia"] = "Port Acta Biol Ser B Sist Ecol Biogeogr Paleontol",
  ["portugaliae electrochimica acta"] = "Port. Electrochim. Acta",
  ["portugaliae mathematica"] = "Port. Math.",
  ["portugaliae mathematica. a journal of the portuguese mathematical society"] = "Port. Math.",
  ["portuguese studies review"] = "Portuguese Stud Rev",
  ["portugália. revista do instituto de arqueologia da faculdade de letras da universidade do porto"] = "Portugalia",
  ["pos, proceedings of science"] = "Pos Proc Sci",
  ["posebna izdanja"] = "Posebna Izdan.",
  ["posebna izdanja. odjeljenje prirodnih i matematiickih nauka"] = "Poseb. izd. od. prir. mat. nauka",
  ["positions (durham, n.c.)"] = "Positions (Durh N C)",
  ["positive directions news : a support and information network of people with hiv/aids, their families, friends and providers"] = "Posit Dir News",
  ["positive health news"] = "Posit Health News",
  ["positive living (los angeles, calif.)"] = "Posit Living",
  ["positive outlook : the newsletter of the aids alternative health project"] = "Posit Outlook",
  ["positively aware : the monthly journal of the test positive aware network"] = "Posit Aware",
  ["positivity. an international mathematics journal devoted to theory and applications of positivity"] = "Positivity",
  ["post reproductive health"] = "Post Reprod Health",
  ["post-communist economies"] = "Postcommunist Econ",
  ["post-graduate course in ophthalmology; proceedings"] = "Postgraduate Course Ophthalmol",
  ["post-medieval archaeology"] = "Postmediev Archaeol",
  ["post-soviet affairs"] = "Post Sov Aff",
  ["post-soviet geography"] = "Post Sov Geogr",
  ["post-soviet geography and economics"] = "Post Sov Geogr Econ",
  ["postal history international"] = "Postal Hist Int",
  ["postal history journal"] = "Post Hist",
  ["postdigital science and education"] = "Postdigital Sci. Educ.",
  ["postdiplomski seminar iz fizike"] = "Postdiplom. Sem. Fiz.",
  ["postdoc journal : a journal of postdoctoral research and postdoctoral affairs"] = "Postdoc J",
  ["postepy biochemii"] = "Postepy Biochem",
  ["postepy chirurgii"] = "Postep Chir",
  ["postepy higieny i medycyny doswiadczalnej"] = "Postepy Hig Med Dosw",
  ["postepy higieny i medycyny doswiadczalnej (online)"] = "Postepy Hig Med Dosw (Online)",
  ["postepy neurologii, neurochirurgii, i psychiatrii"] = "Postep Neurol Neurochir Psychiatr",
  ["postepy reumatologii"] = "Postep Reumatol",
  ["postepy wiedzy medycznej"] = "Postep Wiedzy Med",
  ["postȩpy chirurgii"] = "Postep Chir",
  ["postȩpy dermatologii i alergologii"] = "Postepy Dermatol Alergol",
  ["postȩpy higieny i medycyny doświadczalnej"] = "Postepy Hig Med Dosw",
  ["postȩpy higieny i medycyny doświadczalnej (online)"] = "Postepy Hig Med Dosw (Online)",
  ["postȩpy neurologii, neurochirurgii, i psychiatrii"] = "Postep Neurol Neurochir Psychiatr",
  ["postȩpy reumatologii"] = "Postep Reumatol",
  ["postȩpy wiedzy medycznej"] = "Postep Wiedzy Med",
  ["postępy w kardiologii interwencyjnej = advances in interventional cardiology"] = "Postepy Kardiol Interwencyjnej",
  ["postgraduate dentistry"] = "Postgrad Dent",
  ["postgraduate dentistry / dental learning systems"] = "Postgrad Dent",
  ["postgraduate doctor. africa"] = "Postgrad Doct Afr",
  ["postgraduate medical journal"] = "Postgrad Med J",
  ["postgraduate medicine"] = "Postgrad Med",
  ["postgraduate seminar. american urological association. north central section"] = "Postgrad Semin Am Urol Assoc North Cent",
  ["postharvest biology and technology"] = "Postharvest Biol Technol",
  ["postmedieval : a journal of medieval cultural studies"] = "Postmedieval",
  ["potato research"] = "Potato Res.",
  ["potential analysis"] = "Potential Anal.",
  ["potential analysis. an international journal devoted to the interactions between potential theory, probability theory, geometry and functional analysis"] = "Potential Anal.",
  ["potravinarstvo slovak journal of food sciences"] = "Potravinarstvo Slovak J. Food Sci.",
  ["potsdamer forschungen. naturwissenschaftliche reihe b."] = "Potsdamer Forsch. Naturwiss. Reihe B",
  ["poultry science"] = "Poult Sci",
  ["poumon et le coeur"] = "Poumon Coeur",
  ["pour la science"] = "Pour Sci",
  ["poverty & public policy"] = "Poverty Public Policy",
  ["povijesni prilozi"] = "Povij Pril",
  ["povolzhskiĭ ėkologicheskiĭ zhurnal"] = "Povolz Ekol Z",
  ["powder diffraction"] = "Powder Diffr.",
  ["powder metallurgy"] = "Powder Metall.",
  ["powder metallurgy and metal ceramics"] = "Powder Metall. Met. Ceram.",
  ["powder metallurgy technology"] = "Powder Metall. Technol.",
  ["powder technology"] = "Powder Technol.",
  ["power electronic devices and components"] = "Power Electron. Devices Compon.",
  ["power electronics and power systems"] = "Power Electron. Power Syst.",
  ["power engineer"] = "Power Eng.",
  ["power systems"] = "Power Syst.",
  ["poznań studies in the philosophy of the sciences and the humanities"] = "Poznań Stud. Philos. Sci. Humanities",
  ["poznań university of technology"] = "Found. Comput. Decision Sci.",
  ["poznańskie towarzystwo przyjaciół nauk"] = "Studia Automat. Inform.",
  ["pp news"] = "PP News",
  ["ppar research"] = "PPAR Res",
  ["ppo-special report"] = "PPO Spec Rep",
  ["prace archeologiczne"] = "PraceA",
  ["prace i materialy muzeum archeologicznego i etnograficznego w łodzi"] = "PraceMatLodz",
  ["prace i studia geograficzne"] = "Prac Studi Geograf",
  ["prace i studia geograficzne / uniwersytet warszawski, wydzial geografii i studiow regionalnych"] = "Prac Studi Geograf",
  ["prace ippt/iftr reports"] = "Prace IPPT/IFTR Rep.",
  ["prace komisji medycyny doswiadczalnej / poznanskie towarzystwo przyjaciol nauk, wydzial lekarski"] = "Pr Kom Med Dosw",
  ["prace komisji medycyny doświadczalnej"] = "Pr Kom Med Dosw",
  ["prace naukowe instytutu cybernetyki technicznej politechniki wrocławskiej"] = "Prace Nauk. Inst. Cybernet. Tech. Politech. Wrocław.",
  ["prace naukowe instytutu informatyki stosowanej politechniki wroclawskiej"] = "Prace Nauk. Inst. Inform. Stosow. Politech. Wroc.",
  ["prace naukowe instytutu technologii nieorganicznej i nawozow mineralnych politechniki wroclawskiej"] = "Pr. Nauk Inst. Technol. Nieorg. Nawozow Miner. Politech. Wroclaw.",
  ["prace naukowe uniwersytetu śląskiego w katowicach"] = "Pr. Nauk. Uniw. Śl. Katow.",
  ["prace z dejin prirodnich ved"] = "Pr Dejin Prir Ved",
  ["prace, lodzkie towarzystwo naukowe, wydzial iv, nauk lekarskich"] = "Pr. Lodz. Tow. Nauk. [IV]",
  ["prace. lodzkie towarzystwo naukowe. wydzial iv, nauk lekarskich"] = "Pr Lodz Tow Nauk [IV]",
  ["prace. lódzkie towarzystwo naukowe. wydział iv, nauk lekarskich"] = "Pr Lodz Tow Nauk IV",
  ["pracovni lekarstvi"] = "Prac. Lek.",
  ["pracovní lékar̆stvi"] = "Prac Lek",
  ["practica farmaceutica / ministerul sanatatii, directia farmaceutica si a aparaturii medicale"] = "Pract Farm",
  ["practica farmaceutică"] = "Pract Farm",
  ["practica médica"] = "Supl Pract Med",
  ["practica odontologica"] = "Pract. Odontol.",
  ["practica oto-rhino-laryngologica"] = "Pract Otorhinolaryngol (Basel)",
  ["practical dental monographs"] = "Pract Dent Monogr",
  ["practical dermatology"] = "Pract Dermatol",
  ["practical diabetes international : the journal for diabetes care teams worldwide"] = "Pract Diabetes Int",
  ["practical failure analysis"] = "Pract. Fail. Anal.",
  ["practical gastroenterology"] = "Pract Gastroenterol",
  ["practical laboratory medicine"] = "Pract Lab Med",
  ["practical neurology"] = "Pract Neurol",
  ["practical neurology (fort washington, pa.)"] = "Pract Neurol (Fort Wash Pa)",
  ["practical pain management"] = "Pract Pain Manag",
  ["practical periodical on structural design and construction"] = "Pract. Period. Struct. Des. Constr.",
  ["practical periodontics and aesthetic dentistry"] = "Pract. Periodontics Aesthet. Dent.",
  ["practical periodontics and aesthetic dentistry : ppad"] = "Pract Periodontics Aesthet Dent",
  ["practical procedures & aesthetic dentistry : ppad"] = "Pract Proced Aesthet Dent",
  ["practical proteomics"] = "Pract. Proteomics",
  ["practical psychology for physicians"] = "Pract Psychol Physicians",
  ["practical radiation oncology"] = "Pract Radiat Oncol",
  ["practical spectroscopy"] = "Pract. Spectrosc.",
  ["practice (birmingham, england)"] = "Practice (Birm)",
  ["practice and experience in advanced research computing 2018 : seamless creativity : july 22-26 2017, pittsburgh, pennsylvania. practice and experience in advanced research computing (conference) (2018 : pittsburgh, pa.)"] = "Pract Exp Adv Res Comput 2018 (2018)",
  ["practice digest"] = "Pract Dig",
  ["practice innovations (washington, d.c.)"] = "Pract Innov (Wash D C)",
  ["practice periodical of hazardous, toxic, waste management"] = "Pract. Period. Hazard. Toxic Radioact. Waste Manage.",
  ["practice periodical on structural design and construction"] = "Pract. Period. Struct. Des. Constr.",
  ["practice resource"] = "Pract Resour",
  ["practicing anthropology"] = "Pract Anthropol",
  ["practicing midwife"] = "Pract. Midwife",
  ["practitioner series"] = "Pract. Ser.",
  ["prader-willi perspectives : a journal of prader-willi syndrome information for parents & professionals"] = "Prader Willi Perspect",
  ["praehistorische zeitschrift"] = "Praehist. Z.",
  ["pragmatic and observational research"] = "Pragmat Obs Res",
  ["pragmatic case studies in psychotherapy : pcsp"] = "Pragmat Case Stud Psychother",
  ["prague economic papers"] = "Prague Econ. Pap.",
  ["prague medical report"] = "Prague Med Rep",
  ["prairie rose"] = "Prairie Rose",
  ["prakticke zubni lekarstvi"] = "Prakt. Zubn. Lek.",
  ["praktické zubní lékar̆ství"] = "Prakt Zubn Lek",
  ["prakticky lekar"] = "Prakt Lek",
  ["praktický lékar̆"] = "Prakt Lek",
  ["praktika tes akademias athenon"] = "PAA",
  ["praktika tes en athenais archaiologikes etaireias"] = "PAAH",
  ["praktische anaesthesie, wiederbelebung und intensivtherapie"] = "Prakt. Anaesth.",
  ["praktische anasthesie, wiederbelebung und intensivtherapie"] = "Prakt Anaesth",
  ["praktische anästhesie, wiederbelebung und intensivtherapie"] = "Prakt Anaesth",
  ["praktische forstwirt für die schweiz, der"] = "Prakt. Forstwirt Schweiz",
  ["praktische kieferorthopadie"] = "Prakt. Kieferorthop.",
  ["praktische kieferorthopädie"] = "Prakt Kieferorthop",
  ["praktische metallographie"] = "Prakt. Metallogr.",
  ["pramana – journal of physics"] = "Pramana - J. Phys.",
  ["pras open"] = "PRAS Open",
  ["pratiques et organisation des soins"] = "Prat Organ Soins",
  ["pratsi institutu matematiki natsionalnoi akademii nauk ukraini. matematika ta ii zastosuvannya"] = "Pr. Inst. Mat. Nats. Akad. Nauk Ukr. Mat. Zastos.",
  ["praxis (oneonta, n.y.)"] = "Praxis (Oneonta N Y)",
  ["praxis der kinderpsychologie und kinderpsychiatrie"] = "Prax Kinderpsychol Kinderpsychiatr",
  ["praxis der kinderpsychologie und kinderpsychiatrie. beiheft"] = "Prax Kinderpsychol Kinderpsychiatr Beih",
  ["praxis der naturwissenschaften, biologie in der schule"] = "Prax. Naturwiss. Biol. Sch.",
  ["praxis der naturwissenschaften, chemie"] = "Prax. Naturwiss. Chem.",
  ["praxis der naturwissenschaften, chemie in der schule"] = "Prax. Naturwiss. Chem. Sch.",
  ["praxis der naturwissenschaften, physik"] = "Prax. Naturwiss. Phys.",
  ["praxis der pneumologie"] = "Prax Pneumol",
  ["praxis der pneumologie vereinigt mit der tuberkulosearzt"] = "Prax Pneumol",
  ["praxis der psychotherapie"] = "Prax Psychother",
  ["praxis der psychotherapie und psychosomatik"] = "Prax Psychother Psychosom",
  ["praxis und klinik der pneumologie"] = "Prax Klin Pneumol",
  ["práce z dějin přírodních věd"] = "Pr Dejin Prir Ved",
  ["práctica odontológica"] = "Pract Odontol",
  ["prb report"] = "PRB Rep",
  ["pre- and peri-natal psychology journal : official journal of the pre and peri-natal psychology association of north america"] = "Pre- Peri-Nat Psychol J",
  ["pre-medical journal of columbia university"] = "Premed J Columbia Univ",
  ["precambrian research"] = "Precambrian Res.",
  ["precis analytique des travaux. academie des sciences, belles-lettres et arts de rouen"] = "Precis Anal Trav",
  ["precision agriculture"] = "Precis Agric",
  ["precision cancer medicine"] = "Precis Cancer Med",
  ["precision chemistry"] = "Precis. Chem.",
  ["precision clinical medicine"] = "Precis Clin Med",
  ["precision engineering"] = "Precis. Eng.",
  ["precision medicine"] = "Precis Med",
  ["precision medicine sciences"] = "Precis. Med. Sci.",
  ["precision metal"] = "Precis. Met.",
  ["precision nanomedicine"] = "Precis Nanomed",
  ["precision nutrition"] = "Precis. Nutr.",
  ["precision radiation oncology"] = "Precis Radiat Oncol",
  ["predictive intelligence in medicine. prime (workshop)"] = "Predict Intell Med",
  ["pregnancy hypertension"] = "Pregnancy Hypertens",
  ["pregon (madrid, spain)"] = "Pregon",
  ["pregón (madrid, spain)"] = "Pregon",
  ["prehospital and disaster medicine"] = "Prehosp Disaster Med",
  ["prehospital and disaster medicine : the official journal of the national association of ems physicians and the world association for emergency and disaster medicine in association with the acute care foundation"] = "Prehosp Disaster Med",
  ["prehospital emergency care"] = "Prehosp. Emerg. Care",
  ["prehospital emergency care : official journal of the national association of ems physicians and the national association of state ems directors"] = "Prehosp Emerg Care",
  ["preistoria alpina. museo tridentino di scienze naturali"] = "PreistAlp",
  ["premier cycle des etudes medicales"] = "Prem. Cycle Etud. Med.",
  ["prenatal and neonatal medicine : the international journal of basic and clinical research and practice"] = "Prenat Neonatal Med",
  ["prenatal diagnosis"] = "Prenat Diagn",
  ["prensa medica argentina"] = "Prensa Med. Argent.",
  ["prensa medica mexicana"] = "Prensa Med. Mex.",
  ["prensa médica argentina"] = "Prensa Med Argent",
  ["prensa pediatrica; revista americana de puericultura y pediatria"] = "Prensa Pediatr Rev Am Puericu Pediatr",
  ["prensa pediátrica; revista americana de puericultura y pediatría"] = "Prensa Pediatr Rev Am Puericu Pediatr",
  ["prentice hall information and system sciences series"] = "Prentice Hall Inform. System Sci. Ser.",
  ["prentice hall international series in computer science"] = "Prentice Hall Int. Ser. Comput. Sci.",
  ["prentice hall international series in computer systems science and engineering"] = "Prentice Hall Internat. Ser. Comput. Systems Sci. Engrg.",
  ["prentice hall international series in industrial and systems engineering"] = "Prentice Hall Int. Ser. Indust. Systems Eng.",
  ["prentice hall series in advanced communications technologies"] = "Prentice Hall Ser. Adv. Commun. Technol.",
  ["prentice hall series in computer science"] = "Prentice Hall Ser. Comput. Sci.",
  ["prentice hall series in statistics"] = "Prentice Hall Ser. Statist.",
  ["prentice hall series in technical mathematics"] = "Prentice Hall Ser. Tech. Math.",
  ["prentice hall series on environmental and intelligent manufacturing systems"] = "Prentice Hall Ser. Environ. Intell. Manuf. Syst.",
  ["prentice hall software series"] = "Prentice Hall Software Ser.",
  ["prentice practica"] = "Prentice Pract.",
  ["preparative biochemistry"] = "Prep Biochem",
  ["preparative biochemistry & biotechnology"] = "Prep Biochem Biotechnol",
  ["preparative biochemistry and biotechnology"] = "Prep. Biochem. Biotechnol.",
  ["prepodavanie istorii v shkole"] = "Prepod Ist Sk",
  ["preprint papers - american chemical society, division of fuel chemistry"] = "Prepr. Pap. - Am. Chem. Soc., Div. Fuel Chem.",
  ["preprints - american chemical society, division of energy & fuels"] = "Prepr. - Am. Chem. Soc., Div. Energy Fuels",
  ["preprints - american chemical society, division of petroleum chemistry"] = "Prepr. - Am. Chem. Soc., Div. Pet. Chem.",
  ["preprints of extended abstracts presented at the acs national meeting, american chemical society, division of environmental chemistry"] = "Prepr. Ext. Abstr. ACS Natl. Meet., Am. Chem. Soc., Div. Environ. Chem.",
  ["preprints of papers - american chemical society, division of fuel chemistry"] = "Prepr. Pap. - Am. Chem. Soc., Div. Fuel Chem.",
  ["preprints of symposia - american chemical society, division of fuel chemistry"] = "Prepr. Symp. - Am. Chem. Soc., Div. Fuel Chem.",
  ["presbyterian-st. luke's hospital medical bulletin"] = "Presbyt St Lukes Hosp Med Bull",
  ["presbyterian-st. luke’s hospital medical bulletin"] = "Presbyt St Lukes Hosp Med Bull",
  ["prescribers' journal"] = "Prescr J",
  ["prescribers’ journal"] = "Prescr J",
  ["prescrire international"] = "Prescrire Int",
  ["presence (cambridge, mass.)"] = "Presence (Camb)",
  ["presence africaine (paris, france : 1967)"] = "Presence Afr",
  ["presence normande"] = "Presence Normande",
  ["present tense (new york, n.y.)"] = "Present Tense",
  ["president's council on physical fitness and sports research digest"] = "Pres Counc Phys Fit Sports Res Dig",
  ["presidential studies quarterly"] = "Pres Stud Q",
  ["presse medicale"] = "Presse Med.",
  ["presse medicale (paris, france : 1983)"] = "Presse Med",
  ["presse médicale (paris, france : 1983)"] = "Presse Med",
  ["presse thermale et climatique"] = "Presse Therm. Clim.",
  ["prevencion social"] = "Prev Soc",
  ["preventing chronic disease"] = "Prev Chronic Dis",
  ["preventing school failure : alternative education for children and youth"] = "Prev Sch Fail",
  ["prevention and control : the official journal of the world heart federation"] = "Prev Control",
  ["prevention in counseling psychology : theory, research, practice and training"] = "Prev Couns Psychol",
  ["prevention in human services"] = "Prev Hum Serv",
  ["prevention science"] = "Prev. Sci.",
  ["prevention science : the official journal of the society for prevention research"] = "Prev Sci",
  ["prevention strategist"] = "Prev Strateg",
  ["preventive cardiology"] = "Prev Cardiol",
  ["preventive medicine"] = "Prev Med",
  ["preventive medicine and community health"] = "Prev Med Community Health",
  ["preventive medicine reports"] = "Prev Med Rep",
  ["preventive nutrition and food science"] = "Prev Nutr Food Sci",
  ["preventive veterinary medicine"] = "Prev Vet Med",
  ["prevenzione & assistenza dentale"] = "Prev Assist Dent",
  ["prevenzione e assistenza dentale"] = "Prev. Assist. Dent.",
  ["prevenzione oggi. ispesl"] = "Prev Oggi",
  ["prevenzione stomatologica"] = "Prev Stomatol",
  ["previdenza sociale"] = "Previdenza Soc",
  ["prevision social"] = "Previs Soc",
  ["précis analytique des travaux. académie des sciences, belles-lettres et arts de rouen"] = "Precis Anal Trav",
  ["présence africaine (paris, france : 1967)"] = "Presence Afr",
  ["présence normande"] = "Presence Normande",
  ["pribory i tekhnika eksperimenta [instruments and experimental techniques (ussr)]"] = "Prib. Tekh. E ksp. [Instrum. Exp. Tech. (USSR)]",
  ["pride institute journal of long term home health care"] = "Pride Inst J Long Term Home Health Care",
  ["prikladnaia biokhimiia i mikrobiologiia"] = "Prikl Biokhim Mikrobiol",
  ["prikladnaya biokhimiya i mikrobiologiya"] = "Prikl. Biokhim. Mikrobiol.",
  ["prikladnaya diskretnaya matematika"] = "Prikl. Diskretn. Mat.",
  ["prilozi (makedonska akademija na naukite i umetnostite. oddelenie za medicinski nauki)"] = "Pril (Makedon Akad Nauk Umet Odd Med Nauki)",
  ["prilozi instituta za arheologiju u zagrebu"] = "PriloziZagreb",
  ["prilozi. oddelenie za prirodno-matematichki i biotekhnichki nauki. contributions. section of natural, mathematical and biotechnical sciences"] = "Pril. Odd. Prir.-Mat. Biotekhnichki Nauki",
  ["primary cardiology"] = "Primary Cardiol",
  ["primary care"] = "Prim Care",
  ["primary care & community psychiatry"] = "Prim Care Community Psychiatr",
  ["primary care companion to the journal of clinical psychiatry"] = "Prim Care Companion J Clin Psychiatry",
  ["primary care diabetes"] = "Prim Care Diabetes",
  ["primary care respiratory journal : journal of the general practice airways group"] = "Prim Care Respir J",
  ["primary care update for ob/gyns"] = "Prim Care Update Ob Gyns",
  ["primary care; clinics in office practice"] = "Prim Care",
  ["primary dental care"] = "Prim. Dent. Care",
  ["primary dental care : journal of the faculty of general dental practitioners (uk)"] = "Prim Dent Care",
  ["primary dental journal"] = "Prim Dent J",
  ["primary health care : open access"] = "Prim Health Care",
  ["primary health care research & development"] = "Prim Health Care Res Dev",
  ["primary prevention insights"] = "Prim Prev Insights",
  ["primary psychiatry"] = "Prim psychiatry",
  ["primary sensory neuron : the international interdisciplinary journal reporting basic and clinical research on sensory receptors and primary afferent neurons"] = "Prim Sens Neuron",
  ["primate biology"] = "Primate Biol",
  ["primate conservation"] = "Primate Conserv.",
  ["primate conservation : the newsletter and journal of the iucn/ssc primate specialist group"] = "Primate Conserv",
  ["primate report"] = "Primate Rep",
  ["primates in medicine"] = "Primates Med",
  ["primates; journal of primatology"] = "Primates",
  ["primatologie : revue publiée sous l'égide de la société francophone de primatologie"] = "Primatologie",
  ["primena radioaktivnih izotopa i jonizujućih zrac̆enja u medicini; bilten"] = "Primena Radioakt Izot Joniz Zrac Med",
  ["primer (leawood, kan.)"] = "PRiMER",
  ["primers in complex systems"] = "Primers Complex Syst.",
  ["primus : problems, resources, and issues in mathematics undergraduate studies"] = "PRIMUS (Terre Ht)",
  ["princess takamatsu symposia"] = "Princess Takamatsu Symp",
  ["princeton foundations of contemporary philosophy"] = "Princet. Found. Contemp. Philos.",
  ["princeton history"] = "Princet Hist",
  ["princeton journal of bioethics"] = "Princet J Bioeth",
  ["princeton landmarks in mathematics"] = "Princeton Landmarks Math.",
  ["princeton landmarks in physics"] = "Princeton Landmarks Phys.",
  ["princeton legacy library"] = "Princet. Leg. Libr.",
  ["princeton lifesaver study guide"] = "Princet. Lifesav. Study Guide",
  ["princeton mathematical series"] = "Princeton Math. Ser.",
  ["princeton paperbacks"] = "Princeton Paperbacks",
  ["princeton science library"] = "Princeton Sci. Lib.",
  ["princeton series in applied mathematics"] = "Princeton Ser. Appl. Math.",
  ["princeton series in astrophysics"] = "Princeton Ser. Astrophys.",
  ["princeton series in computer science"] = "Princeton Ser. Comput. Sci.",
  ["princeton series in finance"] = "Princet. Ser. Finance",
  ["princeton series in physics"] = "Princeton Ser. Phys.",
  ["princeton series in theoretical and computational biology"] = "Princeton Ser. Theor. Comput. Biol.",
  ["principe de viana / diputacion foral, institucion principe de viana, consejo de cultura de navarra"] = "Princ Viana",
  ["principe di viana"] = "PrincViana",
  ["principles and practice of clinical research (2015)"] = "Princ Pract Clin Res",
  ["print quarterly"] = "Print Q",
  ["priroda (moscow)"] = "Priroda (Moscow)",
  ["priroda, clovek in zdravje"] = "Prir Clovek Zdravje",
  ["priroda, c̆lovek in zdravje"] = "Prir Clovek Zdravje",
  ["prison service journal"] = "Prison Serv J",
  ["prispevki za novejšo zgodovino"] = "Prispev Novejso Zgod",
  ["privacy in statistical databases. psd (conference : 2004-)"] = "Priv Stat Databases",
  ["private practice"] = "Priv Pact",
  ["príncipe de viana"] = "Princ Viana",
  ["prn forum"] = "PRN Forum",
  ["pro austria romana"] = "AustrRom",
  ["pro familia informationen"] = "Pro Fam Inf",
  ["pro infirmis"] = "Pro Infirm",
  ["pro mathematica"] = "Pro Math.",
  ["pro medico; ärztliches referatenblatt"] = "Pro Med Arztl Ref",
  ["pro re nata"] = "Pro Re Nata",
  ["pro re nata prn : the official publication of the utah nurses' association"] = "Pro Re Nata",
  ["pro re nata prn : the official publication of the utah nurses’ association"] = "Pro Re Nata",
  ["pro-fono : revista de atualizacao cientifica"] = "Pro Fono",
  ["probabilistic engineering mechanics"] = "Probab. Eng. Mech.",
  ["probability and its applications"] = "Probab. Appl.",
  ["probability and its applications (new york)"] = "Probab. Appl. (N. Y.)",
  ["probability and mathematical physics"] = "Probab. Math. Phys.",
  ["probability and mathematical statistics"] = "Probab. Math. Statist.",
  ["probability and stochastics series"] = "Probab. Stochastics Ser.",
  ["probability in the engineering and informational sciences"] = "Probab. Eng. Inf. Sci.",
  ["probability surveys"] = "Probab. Surv.",
  ["probability theory and related fields"] = "Probab. Theory Related Fields",
  ["probability theory and stochastic modelling"] = "Probab. Theory Stoch. Model.",
  ["probability, uncertainty and quantitative risk"] = "Probab Uncertain Quant Risk",
  ["probate and property (chicago, ill. : 1987)"] = "Probate Prop",
  ["probate law journal"] = "Probate Law J",
  ["probate law journal / national college of probate judges and boston university school of law"] = "Probate Law J",
  ["probe (ottawa, ont.)"] = "Probe",
  ["probiotics and antimicrobial proteins"] = "Probiotics Antimicrob Proteins",
  ["problem books in mathematics"] = "Probl. Books in Math.",
  ["problem solving in mathematics and beyond"] = "Probl. Solving Math. Beyond",
  ["problemas del desarrollo"] = "Probl Desarro",
  ["probleme de cardiologie"] = "Probl Cardiol",
  ["probleme de matematica si fizica: serie pentru liceu"] = "Probl. Mat. Fiz. Ser. Pentru Liceu",
  ["probleme de patologie infecţioasă [din literatura sovietică de specialitate]"] = "Probl Patol Infect",
  ["probleme de reumatologie"] = "Probl Reumatol",
  ["probleme de terapeutica"] = "Probl Ter",
  ["probleme de terapeutică"] = "Probl Ter",
  ["probleme de tuberculoza"] = "Probl Tuberculoza",
  ["probleme de tuberculoză"] = "Probl Tuberculoza",
  ["probleme globale ale omenirii"] = "Probl. Globale Omen.",
  ["probleme. symposien aktueller therapeutischer"] = "Probl Symp Aktuell Ther",
  ["problemes actuels d oto-rhino-laryngologie"] = "Probl. Actuels Otorhinolaryngol.",
  ["problemes actuels d’endocrinologie et de nutrition"] = "Probl. Actuels Endocrinol. Nutr.",
  ["problemes actuels d’oto-rhino-laryngologie"] = "Probl Actuels Otorhinolaryngol",
  ["problemes d’amerique latine"] = "Probl Am Lat",
  ["problemes economiques"] = "Probl Econ",
  ["problemi na geografiiata"] = "Probl Geogr",
  ["problemi na geografii︠a︡ta"] = "Probl Geogr",
  ["problemi na izkustvoto. trimesecno spisanie za estetika, teorija, istorija i kritika na izkustvoto"] = "ProblIsk",
  ["problemi na khigienata"] = "Probl Khig",
  ["problemi na onkologiiata"] = "Probl Onkol",
  ["problemi na tekhnicheskata kibernetika i robotikata"] = "Problemi Tekhn. Kibernet. Robot.",
  ["problems in veterinary medicine"] = "Probl Vet Med",
  ["problems of economic transition"] = "Probl Econ Transit",
  ["problems of economics"] = "Probl Econ",
  ["problems of hematology and blood transfusion. problemy gematologii i perelivaniia krovi"] = "Probl Hematol Blood Transfus",
  ["problems of information transmission"] = "Probl. Inf. Transm.",
  ["problems of oncology. voprosy onkologii"] = "Probl Oncol",
  ["problems of virology"] = "Probl Virol",
  ["problemy analiza. issues of analysis"] = "Probl. Anal. Issues Anal.",
  ["problemy biokhimii v michurinskoĭ biologii"] = "Probl Biokhimii Michurinskoi Biol",
  ["problemy endokrinologii"] = "Probl. Endokrinol. (Mosk.)",
  ["problemy endokrinologii i gormonoterapii"] = "Probl. Endokrinol. Gormonoter.",
  ["problemy ėndokrinologii"] = "Probl Endokrinol (Mosk)",
  ["problemy ėndokrinologii i gormonoterapii"] = "Probl Endokrinol Gormonoter",
  ["problemy fiziki, matematiki i tekhniki"] = "Probl. Fiz. Mat. Tekh.",
  ["problemy fiziologicheskoi akustiki"] = "Probl Fiziol Akust",
  ["problemy fiziologicheskoĭ optiki"] = "Probl Fizol Opt",
  ["problemy gematologii i perelivaniia krovi"] = "Probl Gematol Pereliv Krovi",
  ["problemy istoricheskoi demografii sssr"] = "Probl Istor Demogr SSSR",
  ["problemy istoricheskoĭ demografii sssr"] = "Probl Istor Demogr SSSR",
  ["problemy kosmicheskoi biologii"] = "Probl Kosm Biol",
  ["problemy kosmicheskoĭ biologii"] = "Probl Kosm Biol",
  ["problemy lesovedeniia i lesovodstva : sbornik nauchnykh trudov il nan belarusi"] = "Probl Lesoved I Lesovod",
  ["problemy medicinskoj mikologii"] = "Probl Med Mikol",
  ["problemy medycyny wieku rozwojowego"] = "Probl Med Wieku Rozwoj",
  ["problemy osobo opasnykh infektsiĭ"] = "Probl Osobo Opas Infekc",
  ["problemy prochnosti"] = "Probl. Prochn.",
  ["problemy programmirovaniya. problems in programming"] = "Probl. Program.",
  ["problemy radiat︠s︡iĭnoï medyt︠s︡yny ta radiobiolohiï"] = "Probl Radiac Med Radiobiol",
  ["problemy rodziny"] = "Probl Rodziny",
  ["problemy severa"] = "Probl Sev",
  ["problemy sotsialnoi gigieny i istoriia meditsiny"] = "Probl. Sotsialnoi Gig. Istor. Med.",
  ["problemy sot︠s︡ialʹnoĭ gigieny, zdravookhranenii︠a︡ i istorii medit︠s︡iny"] = "Probl Sotsialnoi Gig Zdravookhranenniiai Istor Med",
  ["problemy sot︠s︡ialʹnoĭ gigieny i istorii︠a︡ medit︠s︡iny"] = "Probl Sotsialnoi Gig Istor Med",
  ["problemy sovremennoĭ neĭrokhirurgii"] = "Probl Sovrem Neirokhirurgii",
  ["problemy teorii gravitatsii i èlementarnykh chastits"] = "Problemy Teor. Gravitatsii i Element. Chastits",
  ["problemy tuberkuleza"] = "Probl Tuberk",
  ["problemy tuberkuleza i boleznei legkikh"] = "Probl Tuberk Bolezn Legk",
  ["problemy tuberkuleza i bolezneĭ legkikh"] = "Probl Tuberk Bolezn Legk",
  ["problemy yadernoy fiziki i kosmicheskikh luchey"] = "Problemy Yadern. Fiz. i Kosm. Luchey",
  ["problèmes actuels d'endocrinologie et de nutrition"] = "Probl Actuels Endocrinol Nutr",
  ["problèmes actuels d'oto-rhino-laryngologie"] = "Probl Actuels Otorhinolaryngol",
  ["problèmes d'amérique latine"] = "Probl Am Lat",
  ["problèmes économiques"] = "Probl Econ",
  ["problèmes de mathématiques"] = "Probl. Math. Écrit CAPES 1991–1996",
  ["procare : das forbildungsmagazin fur pflegeberufe"] = "Procare",
  ["procedia chemistry"] = "Procedia Chem",
  ["procedia cirp"] = "Procedia CIRP",
  ["procedia computer science"] = "Procedia Comput. Sci.",
  ["procedia earth and planetary science"] = "Procedia Earth Planet. Sci.",
  ["procedia engineering"] = "Procedia Eng",
  ["procedia environmental sciences"] = "Procedia Environ. Sci.",
  ["procedia food science"] = "Procedia Food Sci.",
  ["procedia in vaccinology"] = "Procedia Vaccinol",
  ["procedia iutam"] = "Procedia IUTAM",
  ["procedia manufacturing"] = "Procedia Manuf",
  ["procedia materials science"] = "Procedia Mater. Sci.",
  ["procedia structural integrity"] = "Procedia Struct. Integrity",
  ["procedia technology"] = "Procedia Technol.",
  ["procedia, social and behavioral sciences"] = "Procedia Soc Behav Sci",
  ["proceedings (baylor university. medical center)"] = "Proc (Bayl Univ Med Cent)",
  ["proceedings (mdpi)"] = "Proceedings (MDPI)",
  ["proceedings - electrochemical society"] = "Proc. - Electrochem. Soc.",
  ["proceedings - electronic components & technology conference"] = "Proc. - Electron. Compon. Conf.",
  ["proceedings - indian academy of sciences, chemical sciences"] = "Proc. - Indian Acad. Sci., Chem. Sci.",
  ["proceedings - international conference on software engineering. international conference on software engineering"] = "Proc Int Conf Softw Eng",
  ["proceedings - mathematical sciences"] = "Proc. Math. Sci.",
  ["proceedings - royal society of edinburgh. section b: natural environment"] = "Proc R Soc Edinb Nat Environ",
  ["proceedings - society of photo-optical instrumentation engineers"] = "Proc Soc Photo Opt Instrum Eng",
  ["proceedings - united states naval institute"] = "Proc. U.S. Nav. Inst.",
  ["proceedings - united states naval institute. united states naval institute"] = "Proc U S Nav Inst",
  ["proceedings / annual meeting of the american association for cancer research [and] annual meeting of the american society of clinical oncology. american association for cancer research. meeting"] = "Proc Annu Meet Am Assoc Cancer Res",
  ["proceedings / ieee computational systems bioinformatics conference, csb. ieee computational systems bioinformatics conference"] = "Proc IEEE Comput Syst Bioinform Conf",
  ["proceedings / ieee computer society bioinformatics conference. ieee computer society bioinformatics conference"] = "Proc IEEE Comput Soc Bioinform Conf",
  ["proceedings / state secretaries management conference, american dental association"] = "Proc State Secr Manage Conf Am Dent Assoc",
  ["proceedings : a conference of the american medical informatics association. amia fall symposium"] = "Proc AMIA Annu Fall Symp",
  ["proceedings : annual international computer software and applications conference. compsac"] = "Proc COMPSAC",
  ["proceedings : automotiveui 2014 : 6th international conference on automotive user interfaces and interactive vehicular applications : september 17-19, 2014, seattle, washington, usa. international conference on automotive user interfaces and interactive vehicular applications (6th : 2014 : seattle, wash.)"] = "Proc Int Conf Automot User Interfaces Interact Veh Appl (2014)",
  ["proceedings : icslp. international conference on spoken language processing"] = "Proc Int Conf Spok Lang Process",
  ["proceedings [of the] annual meeting. american student health association. pacific coast section"] = "Proc Annu Meet Am Stud Health Assoc Pac Coast  Sect",
  ["proceedings [of the] annual meeting. central society for clinical research (u.s.)"] = "Proc Annu Meet Cent Soc Clin Res U S",
  ["proceedings a"] = "Proc. A.",
  ["proceedings and papers of the annual conference of the california mosquito control association"] = "Proc. Pap. Annu. Conf. Calif. Mosq. Control Assoc.",
  ["proceedings and papers of the annual conference of the california mosquito control association, inc"] = "Proc Pap Annu Conf Calif Mosq Control Assoc",
  ["proceedings and transactions of the nova scotian institute of science"] = "Proc. Trans. N. S. Inst. Sci.",
  ["proceedings ieee international symposium on field-programmable custom computing machines : fccm 2011 : 1-3 may 2011, salt lake city, utah, usa. ieee symposium on fpgas for custom computing machines (19th : 2011 : salt lake city, utah)"] = "Proc IEEE Int Symp Field Program Cust Comput Mach",
  ["proceedings in applied mathematics and mechanics"] = "Proc. Appl. Math. Mech.",
  ["proceedings in pure and applied mathematics"] = "Proc. Pure Appl. Math.",
  ["proceedings of annual meeting. industrial hygiene foundation of america"] = "Proc Annu Meet",
  ["proceedings of coling. international conference on computational linguistics"] = "Proc Int Conf Comput Ling",
  ["proceedings of forum acusticum"] = "Proc. Forum Acust",
  ["proceedings of ieee sensors. ieee international conference on sensors"] = "Proc IEEE Sens",
  ["proceedings of ieee southeastcon. ieee southeastcon"] = "Proc IEEE Southeastcon",
  ["proceedings of ilia vekua institute of applied mathematics"] = "Proc. I. Vekua Inst. Appl. Math.",
  ["proceedings of institute of mathematics and mechanics. national academy of sciences of azerbaijan"] = "Proc. Inst. Math. Mech. Natl. Acad. Sci. Azerb.",
  ["proceedings of lunar and planetary science"] = "Proc Lunar Planet Sci",
  ["proceedings of machine learning research"] = "Proc Mach Learn Res",
  ["proceedings of machine learning research (pmlr)"] = "Proc. Mach. Learn. Res. (PMLR)",
  ["proceedings of measuring behavior 2016 : 10th international conference on methods and techniques in behavioral research. international conference on methods and techniques in behavioral research (10th : 2016 : dublin, ireland)"] = "Proc Meas Behav 2016 (2016)",
  ["proceedings of meetings on acoustics. acoustical society of america"] = "Proc Meet Acoust",
  ["proceedings of rijeka school of economics: journal of economics and business"] = "Proceed. Rijeka Sch. Econ.: J. Econ. Bus.",
  ["proceedings of site. society for information technology and teacher education. international conference"] = "Proc SITE",
  ["proceedings of spie"] = "Proc. SPIE",
  ["proceedings of spie - the international society for optical engineering"] = "Proc. SPIE-Int. Soc. Opt. Eng.",
  ["proceedings of spie the international society for optical engineering"] = "Proc. SPIE Int. Soc. Opt. Eng.",
  ["proceedings of spie--the international society for optical engineering"] = "Proc SPIE Int Soc Opt Eng",
  ["proceedings of staff meeting. tulsa, okla. hillcrest memorial hospital"] = "Proc Staff Meet Tulsa Okla Hillcrest Meml Hosp",
  ["proceedings of stem cell research and oncogenesis"] = "Proc Stem Cell Res Oncog",
  ["proceedings of symposia in applied mathematics"] = "Proc. Sympos. Appl. Math.",
  ["proceedings of symposia in pure mathematics"] = "Proc. Sympos. Pure Math.",
  ["proceedings of the -- usenix symposium on operating systems design and implementation (osdi). usenix symposium on operating systems design and implementation"] = "Proc USENIX Symp Oper Syst Des Implement (OSDI)",
  ["proceedings of the 17th international web for all conference"] = "Proc 17th Int Web All Conf (2020)",
  ["proceedings of the 18th acm international conference on interaction design and children"] = "Proc 18th ACM Int Conf Interact Des Child (2019)",
  ["proceedings of the 1st acm sigspatial workshop on geospatial humanities. acm sigspatial workshop on geospatial humanities (1st : 2017 : redondo beach, calif.)"] = "Proc 1st ACM SIGSPATIAL Workshop Geospat Humanit (2017)",
  ["proceedings of the 1st workshop on scholarly web mining. workshop on scholarly web mining (1st : 2017 : cambridge, england)"] = "Proc 1st Workshop Sch Web Min (2017)",
  ["proceedings of the 2012 beliv workshop : beyond time and errors - novel evaluation methods for visualization : seattle wa, usa - october 14-15, 2012. beliv (conference) (2012 : seattle, wash.)"] = "Proc 2012 BELIV Workshop (2012)",
  ["proceedings of the 2014 6th international advanced research workshop on in silico oncology and cancer investigation : the chic project workshop (iarwisoci) : athens, greece, 3-4 november 2014. international advanced research workshop on in silico oncology and cancer investigation (6th : 2014 : athens, greece)"] = "Proc 2014 6th Int Adv Res Workshop In Silico Oncol Cancer Investig (2014)",
  ["proceedings of the 2014 acm southeast regional conference"] = "Proc 2014 ACM Southeast Reg Conf",
  ["proceedings of the 2015 workshop on visual analytics in healthcare"] = "Proc 2015 Workshop Vis Anal Healthc (2015)",
  ["proceedings of the 2nd workshop on human-in-the-loop data analytics. workshop on human-in-the-loop data analytics (2nd : 2017 : chicago, ill.)"] = "Proc 2nd Workshop Hum Loop Data Anal (2017)",
  ["proceedings of the 5th international workshop on exploratory search in databases and the web. international workshop on exploratory search in databases and the web (5th : 2018 : houston, tex.)"] = "Proc 5th Int Workshop Explor Search Databases Web (2018)",
  ["proceedings of the academy of natural sciences of philadelphia"] = "Proc. Acad. Nat. Sci. Philadelphia",
  ["proceedings of the academy of political science"] = "Proc Acad Polit Sci",
  ["proceedings of the acm conference on health, inference, and learning"] = "Proc ACM Conf Health Inference Learn (2020)",
  ["proceedings of the acm on computer graphics and interactive techniques"] = "Proc. ACM Comput. Graphics Comput. Syst.",
  ["proceedings of the acm on human-computer interaction"] = "Proc ACM Hum Comput Interact",
  ["proceedings of the acm on interactive, mobile, wearable and ubiquitous technologies"] = "Proc ACM Interact Mob Wearable Ubiquitous Technol",
  ["proceedings of the acm on management of data"] = "Proc. ACM Manage. Data",
  ["proceedings of the acm on measurement and analysis of computing systems"] = "Proc. ACM Meas. Anal. Comput. Syst.",
  ["proceedings of the acm on networking"] = "Proc. ACM Networking",
  ["proceedings of the acm on programming languages"] = "Proc. ACM Program. Lang.",
  ["proceedings of the acm on software engineering"] = "Proc. ACM Software Eng.",
  ["proceedings of the acm symposium on document engineering. acm symposium on document engineering"] = "Proc ACM Symp Doc Eng",
  ["proceedings of the acm symposium on user interface software and technology. acm symposium on user interface software and technology"] = "Proc ACM Symp User Interface Softw Tech",
  ["proceedings of the acm symposium on virtual reality software and technology. acm symposium on virtual reality software and technology"] = "Proc ACM Symp Virtual Real Softw Technol",
  ["proceedings of the aesf annual technical conference"] = "Proc. - AESF Annu. Tech. Conf.",
  ["proceedings of the american antiquarian society. american antiquarian society"] = "Proc Am Antiq Soc",
  ["proceedings of the american association for cancer research"] = "Proc Am Assoc Cancer Res",
  ["proceedings of the american catholic philosophical association"] = "Proc Am Cathol Philos Assoc",
  ["proceedings of the american catholic philosophical association / office of the national secretary of the association, catholic university of america"] = "Proc Am Cathol Philos Assoc",
  ["proceedings of the american mathematical society"] = "Proc. Amer. Math. Soc.",
  ["proceedings of the american mathematical society. american mathematical society"] = "Proc Am Math Soc",
  ["proceedings of the american mathematical society. series b"] = "Proc. Amer. Math. Soc. Ser. B",
  ["proceedings of the american philosophical society"] = "Proc Am Philos Soc",
  ["proceedings of the american society for information science & technology"] = "Proc. Am. Soc. Inf. Sci. Technol.",
  ["proceedings of the american society of agronomy"] = "Proc. Am. Soc. Agron.",
  ["proceedings of the american society of brewing chemists"] = "Proc. Am. Soc. Brew. Chem.",
  ["proceedings of the american thoracic society"] = "Proc Am Thorac Soc",
  ["proceedings of the analytical division of the chemical society"] = "Proc. Anal. Div. Chem. Soc.",
  ["proceedings of the annual clinical spinal cord injury conference"] = "Proc. Annu. Clin. Spinal Cord Inj. Conf.",
  ["proceedings of the annual conference of the prognostics and health management society. prognostics and health management society. conference"] = "Proc Annu Conf Progn Health Manag Soc",
  ["proceedings of the annual conference on research in medical education"] = "Proc. Annu. Conf. Res. Med. Educ.",
  ["proceedings of the annual meeting of the american psychopathological association"] = "Proc Annu Meet Am Psychopathol Assoc",
  ["proceedings of the annual meeting. american student health association"] = "Proc Annu Meet Am Stud Health Assoc",
  ["proceedings of the annual meeting. hawaiian academy of science, honolulu"] = "Hawaii Acad Sci Honol",
  ["proceedings of the annual session - gulf and caribbean fisheries institute. gulf and caribbean fisheries institute"] = "Proc Gulf Caribb Fish Inst Univ Miami",
  ["proceedings of the annual symposium of the eugenics society"] = "Proc Annu Symp Eugen Soc",
  ["proceedings of the aristotelian society"] = "PAS",
  ["proceedings of the aristotelian society. supplementary volume."] = "PAS(suppl)",
  ["proceedings of the asme conference on smart materials, adaptive structures, and intelligent systems. asme conference on smart materials, adaptive structures, and intelligent systems"] = "Proc ASME Conf Smart Mater Adapt Struct Intell Syst",
  ["proceedings of the asme dynamic systems and control conference. asme dynamic systems and control conference"] = "Proc ASME Dyn Syst Control Conf",
  ["proceedings of the asme international conference on manufacturing science and engineering. asme international conference on manufacturing science and engineering"] = "Proc ASME Int Conf Manuf Sci Eng",
  ["proceedings of the asme micro/nanoscale heat and mass transfer international conference -- 2012 : presented at the asme 2012 3rd micro/nanoscale heat and mass transfer international conference, march 3-6, 2012, atlanta, georgia. micro/nanoscale heat and mass transfer international conference (2012 : atlanta, georgia)"] = "Proc ASME Micro Nanoscale Heat Mass Transf Int Conf (2012)",
  ["proceedings of the asme verification and validation symposium -2019- : vvs2019 : presented at asme 2019 verification and validation symposium, may 15-17, 2019, las vegas, nevada, usa. asme verification and validation symposium (2019 : las vegas, nev.)"] = "Proc ASME Verification Valid Symp 2019 (2019)",
  ["proceedings of the association for information science and technology. association for information science and technology"] = "Proc Assoc Inf Sci Technol",
  ["proceedings of the association of american physicians"] = "Proc Assoc Am Physicians",
  ["proceedings of the australian association of neurologists"] = "Proc Aust Assoc Neurol",
  ["proceedings of the australian physiological and pharmacological society"] = "Proc Aust Physiol Pharmacol Soc",
  ["proceedings of the biological society of washington"] = "Proc. Biol. Soc. Wash.",
  ["proceedings of the botanical society of the british isles. botanical society of the british isles"] = "Proc Bot Soc Br Isles",
  ["proceedings of the british academy"] = "Proc Br Acad",
  ["proceedings of the british institution of radio engineers"] = "Proc. Br. Inst. Radio Eng.",
  ["proceedings of the british paedodontic society"] = "Proc Br Paedod Soc",
  ["proceedings of the british society of dental and maxillofacial radiology"] = "Proc Br Soc Dent Maxillofac Radiol",
  ["proceedings of the cambridge philological society"] = "PCPhS",
  ["proceedings of the cambridge philosophical society"] = "Proc. Cambridge Philos. Soc.",
  ["proceedings of the canadian society for history and philosophy of mathematics/la societe canadienne d’histoire et de philosophie des mathematiques"] = "Proceed. Canad. Soc. Hist. Philos. Math.",
  ["proceedings of the cardiff medical society"] = "Proc. Cardiff Med. Soc.",
  ["proceedings of the centre for mathematics and its applications"] = "Proc. Cent. Math. Appl.",
  ["proceedings of the centre for mathematics and its applications, australian national university"] = "Proc. Centre Math. Appl. Austral. Nat. Univ.",
  ["proceedings of the chemical society"] = "Proc. Chem. Soc.",
  ["proceedings of the chinese academy of medical sciences and the peking union medical college"] = "Proc. Chin. Acad. Med. Sci. Peking Union Med. Coll.",
  ["proceedings of the chinese academy of medical sciences and the peking union medical college = chung-kuo i hsüeh k'o hsüeh yüan, chung-kuo hsieh ho i k'o ta hsüeh hsüeh pao"] = "Proc Chin Acad Med Sci Peking Union Med Coll",
  ["proceedings of the classical association"] = "PCA",
  ["proceedings of the clinical dialysis and transplant forum"] = "Proc Clin Dial Transplant Forum",
  ["proceedings of the combustion institute"] = "Proc. Combust. Inst.",
  ["proceedings of the combustion institute. international symposium on combustion"] = "Proc Combust Inst",
  ["proceedings of the conference of army physicians, central mediterranean forces : held at the institute superiore di sanità : rome, 29th january to 3rd february 1945. conference of army physicians, central mediterranean forces (1945 : istituto superiore di sanità)"] = "Proc Conf Army Physicians Cent Mediterr Forces",
  ["proceedings of the conference on empirical methods in natural language processing. conference on empirical methods in natural language processing"] = "Proc Conf Empir Methods Nat Lang Process",
  ["proceedings of the conference on hazardous waste research. conference on hazardous waste research (kansas state university)"] = "Proc Conf Hazard Waste Res",
  ["proceedings of the conference. association for computational linguistics. meeting"] = "Proc Conf Assoc Comput Linguist Meet",
  ["proceedings of the conference. association for computational linguistics. north american chapter. meeting"] = "Proc Conf",
  ["proceedings of the congress of c.m.f. army surgeons. congress of c.m.f. army surgeons (1945 : rome, italy)"] = "Proc Congress C M F Army Surg",
  ["proceedings of the council - massachusetts medical society"] = "Proc Counc Mass Med Soc",
  ["proceedings of the danish institute at athens"] = "ProcDanInstAth",
  ["proceedings of the dorset natural history and archaeological society. dorset natural history and archaeological society"] = "Proc Dorset Nat Hist Archeol Soc",
  ["proceedings of the edinburgh mathematical society"] = "Proc. Edinburgh Math. Soc.",
  ["proceedings of the edinburgh mathematical society section a: mathematics"] = "Proc. Edinburgh Math. Soc. Sect. A: Math.",
  ["proceedings of the edinburgh mathematical society. series ii"] = "Proc. Edinb. Math. Soc. (2)",
  ["proceedings of the eighth wireless of the students, by the students, and for the students workshop. workshop on wireless of the students, by the students, for the students (8th : 2016 : new york, n.y.)"] = "Proc Eighth Wirel Stud Stud Stud Workshop (2016)",
  ["proceedings of the entomological society of british columbia"] = "Proc. Entomol. Soc. B. C.",
  ["proceedings of the entomological society of london"] = "Proc. Entomol. Soc. London",
  ["proceedings of the entomological society of washington"] = "Proc Entomol Soc Wash",
  ["proceedings of the estonian academy of sciences"] = "Proc. Est. Acad. Sci.",
  ["proceedings of the estonian academy of sciences, chemistry"] = "Proc. Est. Acad. Sci., Chem.",
  ["proceedings of the estonian academy of sciences. biology, ecology"] = "Eesti Teaduste Akadeemia toimetised. Bioloogia, okoloogia|Proc. Estonian Acad. Sci. Biol. Ecol.",
  ["proceedings of the estonian academy of sciences. biology, ecology = eesti teaduste akadeemia toimetised. bioloogia, ökoloogia"] = "Proc Estonian Acad Sci Biol Ecol",
  ["proceedings of the european conference on antennas and propagation. european conference on antennas and propagation"] = "Proc Eur Conf Antennas Propag",
  ["proceedings of the european dialysis and transplant association"] = "Proc. Eur. Dial. Transplant Assoc.",
  ["proceedings of the european dialysis and transplant association - european renal association"] = "Proc. Eur. Dial. Transplant Assoc. Eur. Ren. Assoc.",
  ["proceedings of the european dialysis and transplant association - european renal association. european dialysis and transplant association - european renal association. congress"] = "Proc Eur Dial Transplant Assoc Eur Ren Assoc",
  ["proceedings of the european dialysis and transplant association. european dialysis and transplant association"] = "Proc Eur Dial Transplant Assoc",
  ["proceedings of the european prosthodontic association"] = "Proc. Eur. Prosthodontic Assoc.",
  ["proceedings of the exploredb'17. international workshop on exploratory search in databases and the web (4th : 2017 : chicago, ill.)"] = "Proc ExploreDB17 (2017)",
  ["proceedings of the finnish dental society"] = "Proc. Finn. Dent. Soc.",
  ["proceedings of the finnish dental society. suomen hammaslaakariseuran toimituksia"] = "Proc Finn Dent Soc",
  ["proceedings of the finnish dental society. suomen hammaslääkäriseuran toimituksia"] = "Proc Finn Dent Soc",
  ["proceedings of the first acm sigcas conference on computing and sustainable societies (compass) 2018 : tech museum, san jose, ca, usa, facebook, menlo park, ca, usa, june 20-22, 2018. acm sigcas conference on computing and sustainable societies (1st : 2018 : san jose, calif. ; menlo park, calif.)"] = "Proc First ACM SIGCAS Conf Comput Sustain Soc Compass 2018 (2018)",
  ["proceedings of the first joint nasa cardiopulmonary workshop. joint nasa cardiopulmonary workshop (1st : 1990 : houston, tex.)"] = "Proc First Jt NASA Cardiopulm Workshop",
  ["proceedings of the foundation for orthodontic research"] = "Proc Found Orthod Res",
  ["proceedings of the fourth workshop on data analytics at scale (danac 2015) : may 31st, 2015, melbourne, australia. workshop on data analytics in the cloud (4th : 2015 : melbourne, vic.)"] = "Proc Fourth Workshop Data Anal Scale Danac 2015 (2015)",
  ["proceedings of the frontiers in the convergence of bioscience and information technologies : jeju island, korea, october 11-13, 2007. frontiers in the convergence of bioscience and information technologies (2007 : cheju-do, korea)"] = "Proc Front Converg Biosci Inf Technol (2007)",
  ["proceedings of the genetic and evolutionary computation conference. genetic and evolutionary computation conference"] = "Proc Genet Evol Comput Conf",
  ["proceedings of the geologists' association. geologists' association"] = "Proc Geol Assoc",
  ["proceedings of the geologists’ association"] = "Proc. Geol. Assoc.",
  ["proceedings of the geologists’ association. geologists’ association"] = "Proc Geol Assoc",
  ["proceedings of the glasgow mathematical association"] = "Proc. Glasgow Math. Assoc.",
  ["proceedings of the hawaiian entomological society. hawaiian entomological society"] = "Proc Hawaii Entomol Soc",
  ["proceedings of the health policy forum"] = "Proc. Health Policy Forum",
  ["proceedings of the health policy forum. health policy forum"] = "Proc Health Policy Forum",
  ["proceedings of the helminthological society of washington"] = "Proc Helminthol Soc Wash",
  ["proceedings of the huguenot society of london. huguenot society of london"] = "Proc Huguenot Soc Lond",
  ["proceedings of the human factors and ergonomics society annual meeting"] = "Proc. Hum. Factors Ergon. Soc. Annu. Meet.",
  ["proceedings of the ieee"] = "Proc. IEEE",
  ["proceedings of the ieee international conference on cloud engineering. ieee international conference on cloud engineering"] = "Proc IEEE Int Conf Cloud Eng",
  ["proceedings of the ieee. institute of electrical and electronics engineers"] = "Proc IEEE Inst Electr Electron Eng",
  ["proceedings of the ifac world congress. international federation of automatic control. world congress"] = "Proc IFAC World Congress",
  ["proceedings of the indian academy of sciences mathematical sciences"] = "Proc. Indian Acad. Sci. Math. Sci.",
  ["proceedings of the indian national science academy"] = "Proc. Indian Nat. Sci. Acad.",
  ["proceedings of the indian national science academy. part b, biological sciences"] = "Proc Indian Natl Sci Acad B Biol Sci",
  ["proceedings of the indiana academy of science"] = "Proc Indiana Acad Sci",
  ["proceedings of the institute of acoustics. institute of acoustics (great britain)"] = "Proc Inst Acoust",
  ["proceedings of the institute of mathematics of the national academy of sciences of ukraine. mathematics and its applications"] = "Proc. Inst. Math. Natl. Acad. Sci. Ukr. Math. Appl.",
  ["proceedings of the institute of medicine of chicago"] = "Proc. Inst. Med. Chic.",
  ["proceedings of the institute of statistical mathematics"] = "Proc. Inst. Statist. Math.",
  ["proceedings of the institution of civil engineers"] = "Proc. Inst. Civ. Eng.",
  ["proceedings of the institution of civil engineers. urban design and planning"] = "Proc Inst Civ Eng Urban Des Plan",
  ["proceedings of the institution of civil engineers: bridge engineering"] = "Proc. Inst. Civ. Eng. Bridge Eng.",
  ["proceedings of the institution of civil engineers: civil engineering"] = "Proc. Inst. Civ. Eng. Civ. Eng.",
  ["proceedings of the institution of civil engineers: construction materials"] = "Proc. Inst. Civ. Eng. Constr. Mater",
  ["proceedings of the institution of civil engineers: energy"] = "Proc. Inst. Civ. Eng. Energy",
  ["proceedings of the institution of civil engineers: engineering and computational mechanics"] = "Proc. Inst. Civ. Eng. Eng. Comput. Mech.",
  ["proceedings of the institution of civil engineers: engineering history and heritage"] = "Proc. Inst. Civ. Eng. Eng. Hist. Heritage",
  ["proceedings of the institution of civil engineers: engineering sustainability"] = "Proc. Inst. Civ. Eng. Eng. Sustainability",
  ["proceedings of the institution of civil engineers: forensic engineering"] = "Proc. Inst. Civ. Eng. Forensic Eng.",
  ["proceedings of the institution of civil engineers: geotechnical engineering"] = "Proc. Inst. Civ. Eng. Geotech. Eng.",
  ["proceedings of the institution of civil engineers: ground improvement"] = "Proc. Inst. Civ. Eng. Ground Improv.",
  ["proceedings of the institution of civil engineers: management, procurement and law"] = "Proc. Inst. Civ. Eng. Manage. Procure. Law",
  ["proceedings of the institution of civil engineers: maritime engineering"] = "Proc. Inst. Civ. Eng. Marit. Eng.",
  ["proceedings of the institution of civil engineers: municipal engineer"] = "Proc. Inst. Civ. Eng. Munic. Eng.",
  ["proceedings of the institution of civil engineers: smart infrastructure and construction"] = "Proc. Inst. Civ. Eng. Smart Infrastruct. Constr.",
  ["proceedings of the institution of civil engineers: structures and buildings"] = "Proc. Inst. Civ. Eng. Struct. Build.",
  ["proceedings of the institution of civil engineers: transport"] = "Proc. Inst. Civ. Eng. Transp.",
  ["proceedings of the institution of civil engineers: urban planning and design"] = "Proc. Inst. Civ. Eng. Urban Plann. Des.",
  ["proceedings of the institution of civil engineers: waste and resource management"] = "Proc. Inst. Civ. Eng. Waste Resour. Manag.",
  ["proceedings of the institution of civil engineers: waste resource management"] = "Proc. Inst. Civ. Eng. Waste Resour. Manag.",
  ["proceedings of the institution of civil engineers: water and maritime engineering"] = "Proc. Inst. Civ. Eng. Water Marit. Eng.",
  ["proceedings of the institution of civil engineers: water management"] = "Proc. Inst. Civ. Eng. Water Manage.",
  ["proceedings of the institution of civil engineers: water, maritime and energy"] = "Proc. Inst. Civ. Eng. Water Marit. Energy",
  ["proceedings of the institution of electrical engineers"] = "Proc. Inst. Electr. Eng.",
  ["proceedings of the institution of electronic and radio engineers"] = "Proc. Inst. Electron. Radio Eng.",
  ["proceedings of the institution of mechanical engineers, imeche conference"] = "Proc. Inst. Mech. Eng., IMechE Conf.",
  ["proceedings of the institution of mechanical engineers, part a: journal of power and energy"] = "Proc. Inst. Mech. Eng. Part A J. Power Energy",
  ["proceedings of the institution of mechanical engineers, part b: journal of engineering manufacture"] = "Proc. Inst. Mech. Eng. Part B J. Eng. Manuf.",
  ["proceedings of the institution of mechanical engineers, part c: journal of mechanical engineering science"] = "Proc. Inst. Mech. Eng. Part C J. Mech. Eng. Sci.",
  ["proceedings of the institution of mechanical engineers, part d: journal of automobile engineering"] = "Proc. Inst. Mech. Eng. Part D J. Automob. Eng.",
  ["proceedings of the institution of mechanical engineers, part e: journal of process mechanical engineering"] = "Proc. Inst. Mech. Eng. Part E J. Process Mech. Eng.",
  ["proceedings of the institution of mechanical engineers, part f: journal of rail and rapid transit"] = "Proc. Inst. Mech. Eng. Part F J. Rail Rapid Transit",
  ["proceedings of the institution of mechanical engineers, part g: journal of aerospace engineering"] = "Proc. Inst. Mech. Eng. Part G J. Aerosp. Eng.",
  ["proceedings of the institution of mechanical engineers, part h: journal of engineering in medicine"] = "Proc. Inst. Mech. Eng. Part H J. Eng. Med.",
  ["proceedings of the institution of mechanical engineers, part i: journal of systems and control engineering"] = "Proc. Inst. Mech. Eng., Part I: J. Syst. Control Eng.",
  ["proceedings of the institution of mechanical engineers, part j: journal of engineering tribology"] = "Proc. Inst. Mech. Eng. Part J J. Eng. Tribol.",
  ["proceedings of the institution of mechanical engineers, part k: journal of multi-body dynamics"] = "Proc. Inst. Mech. Eng. Part K J. Multi-body Dyn.",
  ["proceedings of the institution of mechanical engineers, part l: journal of materials: design and applications"] = "Proc. Inst. Mech. Eng. Part L J. Mater. Des. Appl.",
  ["proceedings of the institution of mechanical engineers, part m: journal of engineering for the maritime environment"] = "Proc. Inst. Mech. Eng. Part M J. Eng. Marit. Environ.",
  ["proceedings of the institution of mechanical engineers, part n: journal of nanomaterials, nanoengineering and nanosystems"] = "Proc. Inst. Mech. Eng., Part N: J. Nanomater. Nanoeng. Nanosyst.",
  ["proceedings of the institution of mechanical engineers, part o: journal of risk and reliability"] = "Proc. Inst. Mech. Eng., Part O: J. Risk Reliab.",
  ["proceedings of the institution of mechanical engineers. part b, journal of engineering manufacture"] = "Proc Inst Mech Eng B J Eng Manuf",
  ["proceedings of the institution of mechanical engineers. part c. journal of mechanical engineering science"] = "Proc Inst Mech Eng C J Mech Eng Sci",
  ["proceedings of the institution of mechanical engineers. part f, journal of rail and rapid transit"] = "Proc Inst Mech Eng F J Rail Rapid Transit",
  ["proceedings of the institution of mechanical engineers. part g, journal of aerospace engineering"] = "Proc Inst Mech Eng G J Aerosp Eng",
  ["proceedings of the institution of mechanical engineers. part h, journal of engineering in medicine"] = "Proc Inst Mech Eng H",
  ["proceedings of the institution of mechanical engineers. part i: journal of systems and control engineering"] = "Proc. Inst. Mech. Eng. Part I J. Syst. Control Eng.",
  ["proceedings of the institution of mechanical engineers. part o, journal of risk and reliability"] = "Proc Inst Mech Eng O J Risk Reliab",
  ["proceedings of the institution of mechanical engineers. part p, journal of sports engineering and technology"] = "Proc Inst Mech Eng P J Sport Eng Technol",
  ["proceedings of the institution of mechanical engineers. part p: journal of sports engineering and technology"] = "Proc. Inst. Mech. Eng., Part P: J. Sports Eng. Technol.",
  ["proceedings of the integrated ocean drilling program. integrated ocean drilling program"] = "Proc Integr Ocean Drill Program",
  ["proceedings of the international academy of oral pathology"] = "Proc Int Acad Oral Pathol",
  ["proceedings of the international association of hydrological sciences"] = "Proc. Int. Assoc. Hydrol. Sci.",
  ["proceedings of the international astronomical congress"] = "Proc. Int. Astron. Congr.",
  ["proceedings of the international astronomical union"] = "Proc. Int. Astron. Union",
  ["proceedings of the international astronomical union. international astronomical union"] = "Proc Int Astron Union",
  ["proceedings of the international centre for heat and mass transfer"] = "Proc. Internat. Centre Heat Mass Transfer",
  ["proceedings of the international conference on health care systems engineering. international conference on health care systems engineering (2013 : milan, italy)"] = "Proc Int Conf Health Care Syst Eng  (2013)",
  ["proceedings of the international conference on lasers"] = "Proc. Int. Conf. Lasers",
  ["proceedings of the international congress of ecology"] = "Proc. Int. Congr. Ecol.",
  ["proceedings of the international geometry center"] = "Proc. Int. Geom. Cent.",
  ["proceedings of the international power sources symposium"] = "Proc. Int. Power Sources Symp.",
  ["proceedings of the international pyrotechnics seminar"] = "Proc. Int. Pyrotech. Semin.",
  ["proceedings of the international symposium of human factors and ergonomics in healthcare. international symposium of human factors and ergonomics in healthcare"] = "Proc Int Symp Hum Factors Ergon Healthc",
  ["proceedings of the international symposium on boreal forest"] = "Proc. Int. Symp. Boreal For.",
  ["proceedings of the international symposium on human factors and ergonomics in health care"] = "Proc. Int. Symp. Hum. Factors Ergon. Health Care",
  ["proceedings of the intersociety energy conversion engineering conference"] = "Proc. Intersoc. Energy Convers. Eng. Conf.",
  ["proceedings of the ire"] = "Proc. IRE",
  ["proceedings of the jangjeon mathematical society. memoirs of the jangjeon mathematical society"] = "Proc. Jangjeon Math. Soc.",
  ["proceedings of the japan academy"] = "Proc Jpn Acad",
  ["proceedings of the japan academy, series a, mathematical sciences"] = "Proc. Jpn. Acad. Ser. A, Math. Sci.",
  ["proceedings of the japan academy. series b, physical and biological sciences"] = "Proc Jpn Acad Ser B Phys Biol Sci",
  ["proceedings of the koninklijke nederlandse akademie van wetenschappen"] = "Proc. K. Ned. Akad. Wet.",
  ["proceedings of the koninklijke nederlandse akademie van wetenschappen. biological, chemical, geological, physical, and medical sciences"] = "Proc Koninklijke Nederlandse Akademie Wetenschappen",
  ["proceedings of the koninklijke nederlandse akademie van wetenschappen. series c. biological and medical sciences"] = "Proc K Ned Akad Wet C",
  ["proceedings of the koninklijke nederlandse akademie van wetenschappen. series c: biological and medical sciences"] = "Proc. K. Ned. Akad. Wet. C",
  ["proceedings of the latvian academy of sciences, section b: natural, exact and applied sciences"] = "Proc. Latv. Acad. Sci. Sect. B",
  ["proceedings of the latvian academy of sciences. section b, natural, exact, and applied sciences = latvijas zinātnu akadēmijas vēstis. b dala, dabaszinātnes"] = "Proc Latv Acad Sci B Nat Exact Appl Sci",
  ["proceedings of the linnean society of london"] = "Proc. Linn. Soc. London",
  ["proceedings of the linnean society of new south wales"] = "Proc. Linn. Soc. N. S. W.",
  ["proceedings of the linnean society of new south wales. linnean society of new south wales"] = "Proc Linn Soc N S W",
  ["proceedings of the london mathematical society"] = "Proc. London Math. Soc.",
  ["proceedings of the london mathematical society. third series"] = "Proc. Lond. Math. Soc. (3)",
  ["proceedings of the massachusetts historical society. massachusetts historical society"] = "Mass Hist Soc Proc",
  ["proceedings of the mathematical and physical society of egypt"] = "Proc. Math. Phys. Soc. Egypt",
  ["proceedings of the mathematical society"] = "Proc. Math. Soc.",
  ["proceedings of the mine medical officers association"] = "Proc. Mine Med. Off. Assoc.",
  ["proceedings of the mine medical officers association of sa"] = "Proc. Mine Med. Off. Assoc. SA",
  ["proceedings of the mine medical officers' association"] = "Proc Mine Med Off Assoc",
  ["proceedings of the mine medical officers' association of s. a"] = "Proc Mine Med Off Assoc SA",
  ["proceedings of the mine medical officers’ association"] = "Proc Mine Med Off Assoc",
  ["proceedings of the mine medical officers’ association of s. a"] = "Proc Mine Med Off Assoc SA",
  ["proceedings of the national academy of sciences"] = "Proc. Natl. Acad. Sci.",
  ["proceedings of the national academy of sciences of the united states of america"] = "Proc Natl Acad Sci U S A",
  ["proceedings of the national academy of sciences of the united states of america, early edition"] = "Proc. Natl. Acad. Sci. U.S.A., Early Ed.",
  ["proceedings of the national academy of sciences, india"] = "Proc. Nat. Acad. Sci. India",
  ["proceedings of the national academy of sciences, india section a: physical sciences"] = "Proc. Natl. Acad. Sci., India, Sect. A Phys. Sci.",
  ["proceedings of the national academy of sciences, india section b: biological sciences"] = "Proc. Natl. Acad. Sci., India, Sect. B Biol. Sci.",
  ["proceedings of the national academy of sciences, india, section a: physical sciences"] = "Proc. Natl. Acad. Sci., India, Sect. A",
  ["proceedings of the national academy of sciences, india. section a. physical sciences"] = "Proc. Nat. Acad. Sci. India Sect. A",
  ["proceedings of the national academy of sciences, india. section b"] = "Proc Natl Acad Sci India Sect B Biol Sci",
  ["proceedings of the national cancer conference"] = "Proc. Natl. Cancer Conf.",
  ["proceedings of the national institute of ecology of the republic of korea"] = "Proc Natl Inst Ecol Repub Korea",
  ["proceedings of the national science council, republic of china. part b, life sciences"] = "Proc Natl Sci Counc Repub China B",
  ["proceedings of the nec research symposia"] = "Proc. NEC Res. Sympos.",
  ["proceedings of the neuropsychiatric conference of the sixth service command : john b. murphy memorial auditorium, american college of surgeons, 50 east erie street, chicago, illinois, 16-17 november 1945. neuropsychiatric conference of the sixth service command (1945 : chicago, ill.)"] = "Proc Neuropsychiatr Conf Sixth Serv Command",
  ["proceedings of the new york state association of public health laboratories"] = "Proc N Y State Assoc Public Health Lab",
  ["proceedings of the nipr symposium on antarctic meteorites. nipr symposium on antarctic meteorites"] = "Proc NIPR Symp Antarct Meteorites",
  ["proceedings of the nova scotian institute of science"] = "Proc N S Inst Sci",
  ["proceedings of the nutrition society"] = "Proc. Nutr. Soc.",
  ["proceedings of the nutrition society of india"] = "Proc Nutr Soc India",
  ["proceedings of the ocean drilling program, part b: scientific results"] = "Proc. Ocean Drill. Program Part B Sci. Results.",
  ["proceedings of the ocean drilling program. scientific results. ocean drilling program"] = "Proc Ocean Drill Prog Sci Results",
  ["proceedings of the oklahoma academy of science"] = "Proc Okla Acad Sci",
  ["proceedings of the p. n. lebedev physics institute"] = "Proc. Lebedev Phys. Inst.",
  ["proceedings of the pakistan academy of sciences"] = "Proc. Pakistan Acad. Sci.",
  ["proceedings of the pakistan academy of sciences. a. physical and computational sciences"] = "Proc. Pak. Acad. Sci. A",
  ["proceedings of the physical society"] = "Proc. Phys. Soc.",
  ["proceedings of the physical society b"] = "Proc. Phys. Soc. B",
  ["proceedings of the physical society of london"] = "Proc. Phys. Soc. London",
  ["proceedings of the physical society, london"] = "Proc. Phys. Soc. London",
  ["proceedings of the physical society, london, section a"] = "Proc. Phys. Soc. London, Sec. A",
  ["proceedings of the physical society, london, section b"] = "Proc. Phys. Soc. London, Sec. B",
  ["proceedings of the physical society, section a"] = "Proc. Phys. Soc., Sect. A",
  ["proceedings of the physical society, section b"] = "Proc. Phys. Soc., Sect. B",
  ["proceedings of the physical society. section a"] = "Proc. Phys. Soc. London, Sect. A",
  ["proceedings of the physical society. section b"] = "Proc. Phys. Soc. London, Sect. B",
  ["proceedings of the physiological society"] = "Proc. Physiol. Soc.",
  ["proceedings of the prehistoric society"] = "Proc. Prehist. Soc",
  ["proceedings of the professional staff, brooke general hospital. brooke general hospital"] = "Proc Prof Staff",
  ["proceedings of the romanian academy series a mathematics physics technical sciences information science"] = "Proc. Rom. Acad. Sci. Ser. A Math. Phys. Tech. Sci. Inf. Sci.",
  ["proceedings of the romanian academy. series a. mathematics, physics, technical sciences, information science"] = "Proc. Rom. Acad. Ser. A Math. Phys. Tech. Sci. Inf. Sci.",
  ["proceedings of the royal college of physicians of edinburgh"] = "Proc R Coll Physicians Edinb",
  ["proceedings of the royal entomological society of london"] = "Proc. R. Entomol. Soc. London",
  ["proceedings of the royal entomological society of london series b taxonomy"] = "Proc. R. Entomol. Soc. London Ser. B Taxon.",
  ["proceedings of the royal institution of great britain"] = "Proc R Inst G B",
  ["proceedings of the royal irish academy"] = "Proc. Roy. Irish Acad.",
  ["proceedings of the royal irish academy. section b: biological, geological and chemical science"] = "Proc. R. Ir. Acad. [B]",
  ["proceedings of the royal irish academy. section b: biological, geological, and chemical science"] = "Proc R Ir Acad B",
  ["proceedings of the royal irish academy. section c. archaeology, celtic studies, history, linguistics and literature. royal irish academy"] = "Proc R Ir Acad C Archaeol Celt Stud Hist Linguist Lit",
  ["proceedings of the royal society"] = "Proc. Roy. Soc. London",
  ["proceedings of the royal society a"] = "Proc. R. Soc. A",
  ["proceedings of the royal society b"] = "Proc. R. Soc. B",
  ["proceedings of the royal society of edinburgh"] = "Proc. Roy. Soc. Edinburgh",
  ["proceedings of the royal society of edinburgh. section a. mathematics"] = "Proc. Roy. Soc. Edinburgh Sect. A",
  ["proceedings of the royal society of edinburgh. section b: biology"] = "Proc R Soc Edinb Biol",
  ["proceedings of the royal society of london"] = "Proc. R. Soc. London",
  ["proceedings of the royal society of london a"] = "Proc. R. Soc. Lond. A",
  ["proceedings of the royal society of london series a"] = "Proc. R. Soc. London, Ser. A",
  ["proceedings of the royal society of london series b"] = "Proc. R. Soc. London, Ser. B",
  ["proceedings of the royal society of london, series a: mathematical and physical sciences"] = "Proc. R. Soc. London, Ser. A: Math. Phys. Sci.",
  ["proceedings of the royal society of london, series b"] = "Proc. R. Soc. London, Ser. B",
  ["proceedings of the royal society of london. series a"] = "Proc. R. Soc. London, Ser. A.",
  ["proceedings of the royal society of london. series a - mathematical and physical sciences"] = "Proc. R. Soc. London A - Math Phys. Sci.",
  ["proceedings of the royal society of london. series a, containing papers of a mathematical and physical character"] = "Proc. R. Soc. London A.",
  ["proceedings of the royal society of london. series a, mathematical and physical sciences"] = "Proc R Soc Lond A Math Phys Sci",
  ["proceedings of the royal society of london. series b"] = "Proc. R. Soc. London B",
  ["proceedings of the royal society of london. series b, biological sciences"] = "Proc R Soc Lond B Biol Sci",
  ["proceedings of the royal society of london. series b, containing papers of a biological character. royal society (great britain)"] = "Proc R Soc Lond B Biol Sci",
  ["proceedings of the royal society of london. series b: biological sciences"] = "Proc. R. Soc. Lond. B Biol. Sci.",
  ["proceedings of the royal society of medicine"] = "Proc R Soc Med",
  ["proceedings of the royal society of victoria"] = "Proc. R. Soc. Victoria",
  ["proceedings of the royal society series b biological sciences"] = "Proc. R. Soc. Ser. B Biol. Sci.",
  ["proceedings of the rudolf virchow medical society in the city of new york"] = "Proc Rudolf Virchow Med Soc City N Y",
  ["proceedings of the school of science of tokai university"] = "Proc. Sch. Sci. Tokai Univ.",
  ["proceedings of the second workshop on data management for end-to-end machine learning. workshop on data management for end-to-end machine learning (2nd : 2018 : houston, tex.)"] = "Proc Second Workshop Data Manag End End Mach Learn (2018)",
  ["proceedings of the sigchi conference on human factors in computing systems. chi conference"] = "Proc SIGCHI Conf Hum Factor Comput Syst",
  ["proceedings of the singapore national academy of science"] = "Proc. Singapore Natl. Acad. Sci.",
  ["proceedings of the social statistics section. american statistical association. social statistics section"] = "Proc Soc Stat Sect",
  ["proceedings of the society for analytical chemistry"] = "Proc. Soc. Anal. Chem.",
  ["proceedings of the society for applied bacteriology"] = "Proc. Soc. Appl. Bacteriol.",
  ["proceedings of the society for experimental biology and medicine"] = "Proc. Soc. Exp. Biol. Med.",
  ["proceedings of the society for experimental biology and medicine. society for experimental biology and medicine (new york, n.y.)"] = "Proc Soc Exp Biol Med",
  ["proceedings of the society of agricultural bacteriologists"] = "Proc. Soc. Agric. Bacteriol.",
  ["proceedings of the soil science society of america"] = "Proc. Soil Sci. Soc. Am.",
  ["proceedings of the south dakota academy of science"] = "Proc S Dak Acad Sci",
  ["proceedings of the spi annual technical/marketing conference"] = "Proc. SPI Annu. Tech./Mark. Conf.",
  ["proceedings of the staff meetings. honolulu. clinic"] = "Proc Staff Meet Honol Clin",
  ["proceedings of the staff meetings. mayo clinic"] = "Proc Staff Meet Mayo Clin",
  ["proceedings of the staff meetings. pethah-tiqva, israel. beilinson hospital"] = "Proc Staff Meet Pethah Tiqva Isr Beilinson Hosp",
  ["proceedings of the steklov institute of mathematics"] = "Proc. Steklov Inst. Math.",
  ["proceedings of the stekov institute of mathematics"] = "Proc. Stekov Inst. Math.",
  ["proceedings of the university of otago medical school"] = "Proc Univ Otago Med Sch",
  ["proceedings of the veterans administration spinal cord injury conference"] = "Proc. Veterans Adm. Spinal Cord Inj. Conf.",
  ["proceedings of the virchow-pirquet medical society"] = "Proc Virchow Pirquet Med Soc",
  ["proceedings of the virgil society"] = "PVS",
  ["proceedings of the vldb endowment"] = "Proc. VLDB Endow.",
  ["proceedings of the vldb endowment. international conference on very large data bases"] = "Proceedings VLDB Endowment",
  ["proceedings of the weekly seminar in neurology"] = "Proc Wkly Semin Neurol",
  ["proceedings of the wesley historical society. wesley historical society"] = "Proc Wesley Hist Soc",
  ["proceedings of the west virginia academy of science"] = "Proc W Va Acad Sci",
  ["proceedings of the western pharmacology society"] = "Proc West Pharmacol Soc",
  ["proceedings of the workshop on vitamin d"] = "Proc. Workshop Vitam. D",
  ["proceedings of the yorkshire geological society"] = "Proc. Yorkshire Geol. Soc.",
  ["proceedings of the zoological society"] = "Proc. Zool. Soc.",
  ["proceedings of wild trout. wild trout symposium (yellowstone national park)"] = "Proc Wild Trout Wild Trout Symp",
  ["proceedings one hta : yesterday, today & tomorrow, htai 2004, may 30-june 2, 2004, krakow, poland. health technology assessment international. meeting (1st : 2004 : krakow, poland)"] = "Proc One HTA Health Technol Assess Int Meet 1st 2004 Krakow Pol",
  ["proceedings rmrs"] = "Proc RMRS",
  ["proceedings soil science society of america"] = "Proc. Soil Sci. Soc. Am.",
  ["proceedings, 2015 resilience week (rsw) : hilton philadelphia at penn's landing, philadelphia, pa, 18-20 august, 2015. resilience week (2015 : philadelphia, pa.)"] = "Proc 2015 Resil Week RSW (2015)",
  ["proceedings, amia annual fall symposium"] = "Proc. AMIA Annu. Fall Symp.",
  ["proceedings, amia annual symposium"] = "Proc. AMIA Symp.",
  ["proceedings, amia symposium"] = "Proc. AMIA Symp.",
  ["proceedings, annual management conference - american dental association"] = "Proc Annu Manage Conf Am Dent Assoc",
  ["proceedings, annual management conference, american dental association"] = "Proc. Annu. Manage. Conf. Am. Dent. Assoc.",
  ["proceedings, annual meeting - new jersey mosquito extermination association"] = "Proc Annu Meet N J  Mosqu Exterm Assoc",
  ["proceedings, annual meeting of the medical section of the american council of life insurance"] = "Proc. Annu. Meet. Med. Sect. Am. Counc. Life Insur.",
  ["proceedings, annual meeting of the medical section of the american life convention"] = "Proc Annu Meet Med Sect Am Life Conv",
  ["proceedings, annual meeting of the medical section of the american life insurance association"] = "Proc. Annu. Meet. Med. Sect. Am. Life Insur. Assoc.",
  ["proceedings, annual meeting of the united states animal health association"] = "Proc Annu Meet U S Anim Health Assoc",
  ["proceedings, annual meeting of the united states livestock sanitary association"] = "Proc Annu Meet U S Livest Sanit Assoc",
  ["proceedings, annual symposium on computer applications in medical care"] = "Proc. Annu. Symp. Comput. Appl. Med. Care",
  ["proceedings, foundation for orthodontic research"] = "Proc. Found. Orthod. Res.",
  ["proceedings, governance of technology, information and policies : addressing the challenges of worldwide interconnectivity. workshop on governance of technology, information and policies (2010 : austin, tex.)"] = "Proc Gov Technol Inf Policies (2010)",
  ["proceedings, ieee micro electro mechanical systems. ieee workshop on micro electro mechanical systems"] = "Proc IEEE Micro Electro Mech Syst",
  ["proceedings, international conference on intelligent systems for molecular biology"] = "Proc. Int. Conf. Intell. Syst. Mol. Biol.",
  ["proceedings, national conference on methadone treatment"] = "Proc. Natl. Conf. Methadone Treat.",
  ["proceedings, nursing theory conference"] = "Proc. Nurs. Theory Conf.",
  ["proceedings, royal society of edinburgh. section b: biology"] = "Proc. R. Soc. Edinb. [Biol.]",
  ["proceedings, royal society of edinburgh. section b: natural environment"] = "Proc. R. Soc. Edinb. [Nat. Environ.]",
  ["proceedings, state secretaries management conference, american dental association"] = "Proc. State Secr. Manage. Conf. Am. Dent. Assoc.",
  ["proceedings, the annual meeting of the medical section of the american council of life insurance"] = "Proc Annu Meet Med Sect Am Counc Life Insur",
  ["proceedings, the annual meeting of the medical section of the american life insurance association"] = "Proc Annu Meet Med Sect Am Life Insur Assoc",
  ["proceedings. aaai artificial intelligence and interactive digital entertainment conference"] = "Proc AAAI Artif Intell Interact Digit Enterain Conf",
  ["proceedings. academy of natural sciences of philadelphia"] = "Proc Acad Nat Sci Phila",
  ["proceedings. acm siggraph symposium on interactive 3d graphics and games"] = "Proc ACM SIGGRAPH Symp Interact 3D Graph Games",
  ["proceedings. acm-sigmod international conference on management of data"] = "Proc ACM SIGMOD Int Conf Manag Data",
  ["proceedings. american association for cancer research. annual meeting"] = "Proc Annu Meet Am Assoc Cancer Res",
  ["proceedings. american federation for clinical research"] = "Proc Am Fed Clin Res",
  ["proceedings. american statistical association. annual meeting"] = "Proc Am Stat Assoc",
  ["proceedings. amia symposium"] = "Proc AMIA Symp",
  ["proceedings. animal care panel (u.s.)"] = "Proc Anim Care Panel",
  ["proceedings. biological sciences"] = "Proc Biol Sci",
  ["proceedings. biological sciences / the royal society"] = "Proc Biol Sci",
  ["proceedings. biological society of washington"] = "Proc Biol Soc Wash",
  ["proceedings. california academy of sciences"] = "Proc Calif Acad Sci",
  ["proceedings. canadian cancer conference"] = "Proc Can Cancer Conf",
  ["proceedings. canadian conference on computer and robot vision"] = "Proc Can Conf Comput Robot Vis",
  ["proceedings. chemical society (great britain)"] = "Proc Chem Soc",
  ["proceedings. clinical spinal cord injury conference"] = "Proc Annu Clin Spinal Cord Inj Conf",
  ["proceedings. colloquium in biological sciences"] = "Proc Annu Biol Colloq",
  ["proceedings. consortium on revolutionary europe, 1750-1850"] = "Proc Consort Revolut Eur",
  ["proceedings. data compression conference"] = "Proc Data Compress Conf",
  ["proceedings. design automation conference"] = "Proc Des Autom Conf",
  ["proceedings. euromicro international conference on parallel, distributed, and network-based processing"] = "Proc Euromicro Int Conf Parallel Distrib Netw Based Process",
  ["proceedings. eye tracking research & applications symposium"] = "Proc Eye Track Res Appl Symp",
  ["proceedings. florida conference on recent advances in robotics"] = "Proc Fla Conf Recent Adv Robot",
  ["proceedings. ibm medical symposium"] = "Proc IBM Med Symp",
  ["proceedings. ieee computational systems bioinformatics conference"] = "Proc IEEE Comput Syst Bioinform Conf",
  ["proceedings. ieee computer society bioinformatics conference"] = "Proc IEEE Comput Soc Bioinform Conf",
  ["proceedings. ieee computer society conference on computer vision and pattern recognition"] = "Proc IEEE Comput Soc Conf Comput Vis Pattern Recognit",
  ["proceedings. ieee conference on intelligent transportation systems"] = "Proc IEEE Conf Intell Transp Syst",
  ["proceedings. ieee conference on virtual reality and 3d user interfaces"] = "Proc IEEE Conf Virtual Real 3D User Interfaces",
  ["proceedings. ieee global humanitarian technology conference"] = "Proc IEEE Glob Humanit Technol Conf",
  ["proceedings. ieee international automated software engineering conference"] = "Proc IEEE Int Autom Softw Eng Conf",
  ["proceedings. ieee international conference on bioinformatics and biomedicine"] = "Proceedings (IEEE Int Conf Bioinformatics Biomed)",
  ["proceedings. ieee international conference on cluster computing"] = "Proc IEEE Int Conf Clust Comput",
  ["proceedings. ieee international conference on computer vision"] = "Proc IEEE Int Conf Comput Vis",
  ["proceedings. ieee international conference on data mining"] = "Proc IEEE Int Conf Data Min",
  ["proceedings. ieee international conference on healthcare informatics, imaging and systems biology"] = "Proc IEEE Int Conf Healthc Inform Imaging Syst Biol",
  ["proceedings. ieee international conference on micro electro mechanical systems"] = "Proc IEEE Int Conf Micro Electro Mech Syst",
  ["proceedings. ieee international conference on multimedia and expo"] = "Proc (IEEE Int Conf Multimed Expo)",
  ["proceedings. ieee international conference on robotics and automation"] = "Proc IEEE Int Conf Robot Autom",
  ["proceedings. ieee international conference on semantic computing"] = "Proc IEEE Int Conf  Semant Comput",
  ["proceedings. ieee international congress on big data"] = "Proc IEEE Int Congr Big Data",
  ["proceedings. ieee international requirements engineering conference"] = "Proc IEEE Int Requir Eng Conf",
  ["proceedings. ieee international symposium on bioinformatics and bioengineering"] = "Proc IEEE Int Symp Bioinformatics Bioeng",
  ["proceedings. ieee international symposium on biomedical imaging"] = "Proc IEEE Int Symp Biomed Imaging",
  ["proceedings. ieee international symposium on computer-based medical systems"] = "Proc IEEE Int Symp Comput Based Med Syst",
  ["proceedings. ieee international symposium on high-assurance systems engineering"] = "Proc IEEE Int Symp High Assur Syst Eng",
  ["proceedings. ieee international symposium on information theory"] = "Proc IEEE Int Symp Info Theory",
  ["proceedings. ieee international symposium on policies for distributed systems and networks"] = "Proc IEEE Int Symp Policies Distrib Syst Netw",
  ["proceedings. ieee radio and wireless symposium"] = "Proc IEEE Radio Wirel Symp",
  ["proceedings. ieee southwest symposium on image analysis and interpretation"] = "Proc IEEE Southwest Symp Image Anal Interpret",
  ["proceedings. ieee symposium on 3d user interfaces"] = "Proc IEEE Symp 3D User Interfaces",
  ["proceedings. ieee symposium on biological data visualization"] = "Proc IEEE Symp Biol Data Vis",
  ["proceedings. ieee symposium on computer-based medical systems"] = "Proc IEEE Symp Comput Based Med Syst",
  ["proceedings. ieee symposium on computers and communications"] = "Proc IEEE Symp Comput Commun",
  ["proceedings. ieee symposium on large-scale data analysis and visualization"] = "Proc IEEE Symp Large Scale Data Anal Vis",
  ["proceedings. ieee symposium on security and privacy"] = "Proc IEEE Symp Secur Priv",
  ["proceedings. ieee symposium on security and privacy. workshops"] = "Proc IEEE Symp Secur Priv Workshops",
  ["proceedings. ieee symposium on visual analytics science and technology"] = "Proc IEEE Symp Vis Anal Sci Technol",
  ["proceedings. ieee ultrasonics symposium"] = "Proc IEEE Ultrason Symp",
  ["proceedings. ieee virtual reality conference"] = "Proc IEEE Virtual Real Conf",
  ["proceedings. ieee western new york image and signal processing workshop"] = "Proc IEEE West N Y Image Signal Process Workshop",
  ["proceedings. ieee workshop on applications of computer vision"] = "Proc IEEE Workshop Appl Comput Vis",
  ["proceedings. ieee workshop on automatic speech recognition and understanding"] = "Proc IEEE Workshop Autom Speech Recognit Underst",
  ["proceedings. ieee workshop on mobile computing systems and applications"] = "Proc IEEE Workshop Mob Comput Syst Appl",
  ["proceedings. ieee/wic/acm international conference on web intelligence and intelligent agent technology"] = "Proc IEEE WIC ACM Int Conf Web Intell Intell Agent Technol",
  ["proceedings. ifsa world congress"] = "Proc IFSA World Congr",
  ["proceedings. information theory workshop"] = "Proc Inf Theory Workshop",
  ["proceedings. international conference on 3d vision"] = "Proc Int Conf 3D Vis",
  ["proceedings. international conference on availability, reliability and security"] = "Proc Int Conf Availab Reliab Secur",
  ["proceedings. international conference on data engineering"] = "Proc Int Conf Data Eng",
  ["proceedings. international conference on distributed computing systems"] = "Proc Int Conf Distrib Comput Syst",
  ["proceedings. international conference on image processing"] = "Proc Int Conf Image Proc",
  ["proceedings. international conference on intelligent systems for molecular biology"] = "Proc Int Conf Intell Syst Mol Biol",
  ["proceedings. international conference on machine learning and cybernetics"] = "Proc Int Conf Mach Learn Cybern",
  ["proceedings. international conference on quality software"] = "Proc Int Conf Qual Softw",
  ["proceedings. international congress on yak"] = "Proc Int Congr Yak",
  ["proceedings. international coral reef symposium"] = "Proc Int Coral Reef Symp",
  ["proceedings. international fall workshop vision, modeling, and visualization"] = "Proc Int Fall Workshop Vis Model Vis",
  ["proceedings. international ircobi conference on the biomechanics of impacts"] = "Proc Int IRCOBI Conf Biomech Impacts",
  ["proceedings. international meshing roundtable"] = "Proc Int Meshing Roundtable",
  ["proceedings. international symposium on computer architecture"] = "Proc Int Symp Comput Archit",
  ["proceedings. international symposium on symbolic and numeric algorithms for scientific computing"] = "Proc Int Symp Symb Numer Algorithms Sci Comput",
  ["proceedings. international symposium on wearable computers"] = "Proc Int Symp Wearable Comput",
  ["proceedings. international workshop on social informatics"] = "Proc Int Workshop Soc Inform",
  ["proceedings. ipdps (conference)"] = "Proc IPDPS (Conf)",
  ["proceedings. mathematical, physical, and engineering sciences"] = "Proc Math Phys Eng Sci",
  ["proceedings. national cancer conference"] = "Proc Natl Cancer Conf",
  ["proceedings. national conference on methadone treatment"] = "Proc Natl Conf Methadone Treat",
  ["proceedings. nursing theory conference (university of kansas)"] = "Proc Nurs Theory Conf",
  ["proceedings. ohio collaborative conference on bioinformatics"] = "Proc Ohio Collab Conf Bioinform",
  ["proceedings. pacific conference on computer graphics and applications"] = "Proc Pac Conf Comput Graph Appl",
  ["proceedings. royal australasian college of physicians"] = "Proc R Australas Coll Physicians",
  ["proceedings. society for the study of fertility"] = "Proc Soc Study Fertil",
  ["proceedings. state secretaries management conference"] = "Proc State Secr Manage Conf",
  ["proceedings. symposium on computer applications in medical care"] = "Proc Annu Symp Comput Appl Med Care",
  ["proceedings. symposium on computer architecture and high performance computing"] = "Proc Symp Comput Archit High Perform Comput",
  ["proceedings. symposium on haptic interfaces for virtual environment and teleoperator systems"] = "Proc Symp Haptic Interface Virtual Env Teleoperator Syst",
  ["proceedings. transvaal mine medical officers' association"] = "Proc Transvaal Mine Med Off Assoc",
  ["proceedings. usnc-ursi radio science meeting"] = "Proc (USNC URSI Radio Sci Meet)",
  ["proceedings. veterans administration spinal cord injury conference"] = "Proc Veterans Adm Spinal Cord Inj Conf",
  ["proceedings. workshop on mathematical methods in biomedical image analysis"] = "Proc Workshop Math Methods Biomed Image Analysis",
  ["proceedings: plant science"] = "Proc.: Plant Sci.",
  ["procesos históricos"] = "Process Hist",
  ["process biochemistry"] = "Process Biochem.",
  ["process biochemistry (amsterdam, netherlands)"] = "Process Biochem. (Amsterdam, Neth.)",
  ["process biochemistry (barking, london, england)"] = "Process Biochem",
  ["process biochemistry (oxford, united kingdom)"] = "Process Biochem. (Oxford, U. K.)",
  ["process control and quality"] = "Process Control Qual.",
  ["process integration and optimization for sustainability"] = "Process Integr. Optim. Sustainability",
  ["process safety and environment protection"] = "Process Saf. Environ. Prot.",
  ["process safety and environmental protection"] = "Process Saf. Environ. Prot.",
  ["process safety and environmental protection : transactions of the institution of chemical engineers, part b"] = "Process Saf Environ Prot",
  ["process safety progress"] = "Process Saf. Prog.",
  ["processes (basel, switzerland)"] = "Processes (Basel)",
  ["processing and application of ceramics"] = "Process. Appl. Ceram.",
  ["procès-verbaux, exposés et discussions [du] congrès de la société internationale de chirurgie. international society of surgery"] = "P V Expo Discuss Congr Soc Int Chir",
  ["proclins cardiology"] = "Proclins Cardiol",
  ["production & manufacturing research"] = "Prod Manuf Res",
  ["production and operations management"] = "Prod Oper Manag",
  ["production engineering"] = "Prod. Eng.",
  ["produits & problemes pharmaceutiques"] = "Prod Probl Pharm",
  ["produits & problêmes pharmaceutiques"] = "Prod Probl Pharm",
  ["produits et problemes pharmaceutiques"] = "Prod. Probl. Pharm.",
  ["produits pharmaceutiques"] = "Prod Pharm",
  ["produktion von leiterplatten und systemen"] = "Prod. Leiterplatten Sys.",
  ["profamilia : planificacion, poblacion y desarollo"] = "Profamilia",
  ["profamilia : planificación, población y desarollo"] = "Profamilia",
  ["professional and industrial computing series"] = "Prof. Indust. Comput. Ser.",
  ["professional animal scientist"] = "Prof. Anim. Sci.",
  ["professional care of mother and child"] = "Prof Care Mother Child",
  ["professional case management"] = "Prof Case Manag",
  ["professional development (philadelphia, pa.)"] = "Prof Dev (Phila)",
  ["professional development series (chicago, ill.)"] = "Prof Dev Ser (Chic Ill)",
  ["professional ethics (gainesville, fla.)"] = "Prof Ethics",
  ["professional ethics report : newsletter of the american association for the advancement of science, committee on scientific freedom & responsibility, professional society ethics group"] = "Prof Ethics Rep",
  ["professional flashes"] = "Prof Flashes",
  ["professional geographer"] = "Prof. Geogr.",
  ["professional nurse"] = "Prof. Nurse",
  ["professional nurse (london, england)"] = "Prof Nurse",
  ["professional nursing home"] = "Prof Nurs Home",
  ["professional psychology"] = "Prof Psychol",
  ["professional psychology, research and practice"] = "Prof Psychol Res Pr",
  ["professional safety"] = "Prof Saf",
  ["professional sanitation management"] = "Prof Sanit Manage",
  ["professioni infermieristiche"] = "Prof Inferm",
  ["profilakticheskai︠a︡ i klinicheskai︠a︡ medit︠s︡ina"] = "Profil Klin Med",
  ["profile of medical practice"] = "Profile Med Pract",
  ["profile of medical practice / center for health services research and development, american medical association"] = "Profile Med Pract",
  ["profiles in healthcare communications"] = "Profiles Healthc Commun",
  ["profiles in healthcare marketing"] = "Profiles Healthc Mark",
  ["profiles in hospital marketing"] = "Profiles Hosp Mark",
  ["profiles of drug substances, excipients, and related methodology"] = "Profiles Drug Subst Excip Relat Methodol",
  ["progess in agricultural engineering sciences"] = "Prog. Agric. Eng. Sci.",
  ["program & abstracts : first hiv drp symposium understanding antiviral drug resistance, december 3-6, 2000, westfields internaitonal conference center, chantilly, virginia. hiv drp symposium understanding antiviral drug resistance (1st : 2000 : chantilly, virginia)"] = "Program Abstr HIV DRP Symp Underst Antivir Drug Resist 1st 2000 Chantilly Va",
  ["program and abstracts of the 3rd conference on retroviruses and opportunistic infections : sheraton washington hotel, january 28-february 2 [i.e. 1], 1996, washington, dc. conference on retroviruses and opportunistic infections (3rd : 1996 : washington, d.c.)"] = "Program Abstr 3rd Conf Retrovir Oppor Infect Conf Retrovir Oppor Infect 3rd 1996 Wash D C",
  ["program and abstracts of the 4th conference on retroviruses and opportunistic infections : sheraton washington hotel, january 22-26, 1997, washington, dc. conference on retroviruses and opportunistic infections (4th : 1997 : washington, d.c.)"] = "Program Abstr 4th Conf Retrovir Oppor Infect Conf Retrovir Oppor Infect 4th 1997 Wash DC",
  ["program and abstracts of the 5th conference on retroviruses and opportunistic infections : sheraton chicago hotel and towers, february 1-5 1998, chicago, il. conference on retroviruses and opportunistic infections (5th : 1998 : chicago, ill.)"] = "Program Abstr 5th Conf Retrovir Oppor Infect Conf Retrovir Oppor Infect 5th 1998 Chic Ill",
  ["program and abstracts of the 6th conference on retroviruses and opportunistic infections : sheraton chicago hotel & towers, january 31 - february 4, 1999. conference on retroviruses and opportunistic infections (6th : 1999 : chicago, ill.)"] = "Program Abstr 6th Conf Retrovir Oppor Infect Conf Retrovir Oppor Infect 6th 1999 Chic Ill",
  ["program and abstracts of the first national conference on human retroviruses and related infections : sheraton washington hotel, 12-16 december 1993, washington, dc. national conference on human retroviruses and related infections (1st : 1993 : washington, d.c.)"] = "Program Abstr First Natl Conf Hum Retrovir Relat Infect Natl Conf Hum Retrovir Relat Infect 1st 1993 Wash DC",
  ["program and abstracts of the hiv infection in women conference : setting a new agenda : february 22-24, 1995, sheraton washington hotel, washington, d.c. hiv infection in women conference (1995 : washington, d.c.)"] = "Program Abstr HIV Infect Women Conf HIV Infect Women Conf 1995 Wash DC",
  ["program and abstracts of the second national conference on human retroviruses and related infections : sheraton washington hotel, 29 january-2 february 1995, washington, dc. national conference on human retroviruses and related infections (2nd : 1995 : washington, d.c.)"] = "Program Abstr Second Natl Conf Hum Retrovir Relat Infect Natl Conf Hum Retrovir Relat Infect 2nd 1995 Wash DC",
  ["program notes (association of university programs.)"] = "Program Notes Assoc Univ Programs Health Adm",
  ["program notes, association of university programs in health administration"] = "Program Notes Assoc. Univ. Programs Health Adm.",
  ["program quarterly - the asia foundation. the asia foundation"] = "Program Q Asia Found",
  ["programming and computer software"] = "Program. Comput. Softw.",
  ["progres en urologie"] = "Prog. Urol.",
  ["progres en urologie : journal de l’association francaise d’urologie et de la societe francaise d’urologie"] = "Prog Urol",
  ["progres medical"] = "Progr. Med. (Paris.)",
  ["progres odonto-stomatologique"] = "Prog. Odontostomatol.",
  ["progresos de patología y clínica"] = "Prog Patol Clin",
  ["progresos de pediatria y puericultura"] = "Progr Pediatr Puericult",
  ["progresos de pediatría y puericultura"] = "Prog Pediatr Pueric",
  ["progresos de terapeutica clinica"] = "Progr. Ter. Clin.",
  ["progresos de terapéutica clínica"] = "Prog Ter Clin",
  ["progress in additive manufacturing"] = "Prog. Addit. Manuf.",
  ["progress in advanced computing and intelligent engineering : proceedings of icacie 2016. international conference on advanced computing and intelligent engineering (2016 : bhubaneswar, india)"] = "Prog Adv Comput Intell Eng",
  ["progress in aerospace sciences"] = "Prog. Aerosp. Sci.",
  ["progress in aids pathology"] = "Prog AIDS Pathol",
  ["progress in allergy"] = "Prog Allergy",
  ["progress in artificial intelligence"] = "Prog. Artif. Intell.",
  ["progress in atomic medicine"] = "Prog At Med",
  ["progress in behavior modification"] = "Prog Behav Modif",
  ["progress in biochemical pharmacology"] = "Prog Biochem Pharmacol",
  ["progress in biocybernetics"] = "Prog Biocybern",
  ["progress in biological sciences"] = "Prog Biol Sci",
  ["progress in biomaterials"] = "Prog Biomater",
  ["progress in biomedical engineering"] = "Prog. Biomed. Eng.",
  ["progress in biomedical engineering (bristol, england)"] = "Prog Biomed Eng (Bristol)",
  ["progress in biometeorology, division a: progress in human biometeorology"] = "Prog. Biometeorol.",
  ["progress in biometeorology. division a: progress in human biometeorology"] = "Prog Biometeorol",
  ["progress in biophysics and biophysical chemistry"] = "Prog Biophys Biophys Chem",
  ["progress in biophysics and molecular biology"] = "Prog Biophys Mol Biol",
  ["progress in biotechnology"] = "Prog. Biotechnol.",
  ["progress in botany. fortschritte der botanik"] = "Prog Bot",
  ["progress in brain research"] = "Prog Brain Res",
  ["progress in cardiovascular diseases"] = "Prog Cardiovasc Dis",
  ["progress in cardiovascular nursing"] = "Prog Cardiovasc Nurs",
  ["progress in cell cycle research"] = "Prog Cell Cycle Res",
  ["progress in chemical toxicology"] = "Prog Chem Toxicol",
  ["progress in clinical and biological research"] = "Prog Clin Biol Res",
  ["progress in clinical cancer"] = "Prog Clin Cancer",
  ["progress in clinical immunology"] = "Prog Clin Immunol",
  ["progress in clinical parasitology"] = "Prog Clin Parasitol",
  ["progress in clinical pathology"] = "Prog Clin Pathol",
  ["progress in colloid & polymer science"] = "Prog Colloid Polym Sci",
  ["progress in colloid and polymer science"] = "Prog. Colloid Polym. Sci.",
  ["progress in community health partnerships : research, education, and action"] = "Prog Community Health Partnersh",
  ["progress in computational fluid dynamics"] = "Prog. Comput. Fluid Dyn.",
  ["progress in computational fluid dynamics. an international journal"] = "Prog. Comput. Fluid Dyn.",
  ["progress in computer science and applied logic"] = "Progr. Comput. Sci. Appl. Logic",
  ["progress in creep and fracture"] = "Progr. Creep Fracture",
  ["progress in crystal growth and characterization"] = "Prog. Cryst. Growth Charact.",
  ["progress in crystal growth and characterization of materials"] = "Prog. Cryst. Growth Charact. Mater.",
  ["progress in disaster science"] = "Prog. Disaster Sci.",
  ["progress in drug research"] = "Prog. Drug Res.",
  ["progress in drug research. fortschritte der arzneimittelforschung. progres des recherches pharmaceutiques"] = "Prog Drug Res",
  ["progress in drug research. fortschritte der arzneimittelforschung. progrès des recherches pharmaceutiques"] = "Prog Drug Res",
  ["progress in earth and planetary science"] = "Prog. Earth Planet. Sci.",
  ["progress in electromagnetics research b. pier b"] = "Prog Electromagn Res B Pier B",
  ["progress in electromagnetics research c. pier c"] = "Prog Electromagn Res C Pier C",
  ["progress in electromagnetics research m. pier m"] = "Prog Electromagn Res M Pier M",
  ["progress in electromagnetics research symposium : [proceedings]. progress in electromagnetics research symposium"] = "Prog Electromagn Res Symp",
  ["progress in energy"] = "Prog. Energy",
  ["progress in energy and combustion science"] = "Prog. Energy Combust. Sci.",
  ["progress in environmental geography"] = "Prog. Environ. Geogr.",
  ["progress in experimental personality & psychopathology research"] = "Prog Exp Pers Psychopathol Res",
  ["progress in experimental personality and psychopathology research"] = "Prog. Exp. Pers. Psychopathol. Res.",
  ["progress in experimental personality research"] = "Prog Exp Pers Res",
  ["progress in experimental tumor research"] = "Prog Exp Tumor Res",
  ["progress in experimental tumor research. fortschritte der experimentellen tumorforschung. progres de la recherche experimentale des tumeurs"] = "Prog Exp Tumor Res",
  ["progress in food & nutrition science"] = "Prog Food Nutr Sci",
  ["progress in food and nutrition science"] = "Prog. Food Nutr. Sci.",
  ["progress in growth factor research"] = "Prog Growth Factor Res",
  ["progress in gynecology"] = "Prog Gynecol",
  ["progress in hematology"] = "Prog Hematol",
  ["progress in hemostasis and thrombosis"] = "Prog Hemost Thromb",
  ["progress in heterocyclic chemistry"] = "Prog. Heterocycl. Chem.",
  ["progress in histochemistry and cytochemistry"] = "Prog. Histochem. Cytochem.",
  ["progress in human geography"] = "Prog Hum Geogr",
  ["progress in human reproduction research"] = "Prog Hum Reprod Res",
  ["progress in immunobiological standardization"] = "Prog Immunobiol Stand",
  ["progress in industrial ecology"] = "Prog. Ind. Ecol.",
  ["progress in industrial microbiology"] = "Prog. Ind. Microbiol.",
  ["progress in inorganic biochemistry and biophysics"] = "Prog Inorg Biochem Biophys",
  ["progress in inorganic chemistry"] = "Prog. Inorg. Chem.",
  ["progress in is"] = "Prog. IS",
  ["progress in lipid research"] = "Prog Lipid Res",
  ["progress in liver diseases"] = "Prog Liver Dis",
  ["progress in materials science"] = "Prog. Mater Sci.",
  ["progress in mathematical physics"] = "Prog. Math. Phys.",
  ["progress in mathematics"] = "Progr. Math.",
  ["progress in medical genetics"] = "Prog Med Genet",
  ["progress in medical virology"] = "Prog. Med. Virol.",
  ["progress in medical virology. fortschritte der medizinischen virusforschung. progres en virologie medicale"] = "Prog Med Virol",
  ["progress in medical virology. fortschritte der medizinischen virusforschung. progrès en virologie médicale"] = "Prog Med Virol",
  ["progress in medicinal chemistry"] = "Prog Med Chem",
  ["progress in medicine"] = "Prog Med",
  ["progress in metal physics"] = "Prog. Met. Phys.",
  ["progress in molecular and subcellular biology"] = "Prog Mol Subcell Biol",
  ["progress in molecular biology and translational science"] = "Prog Mol Biol Transl Sci",
  ["progress in natural science"] = "Prog. Nat. Sci.",
  ["progress in natural science : communication of state key laboratories of china"] = "Prog Nat Sci",
  ["progress in natural science: materials international"] = "Prog. Nat. Sci. Mater. Int.",
  ["progress in neuro-psychopharmacology"] = "Prog Neuropsychopharmacol",
  ["progress in neuro-psychopharmacology & biological psychiatry"] = "Prog Neuropsychopharmacol Biol Psychiatry",
  ["progress in neuro-psychopharmacology and biological psychiatry"] = "Prog. Neuro-Psychopharmacol. Biol. Psychiatry",
  ["progress in neurobiology"] = "Prog Neurobiol",
  ["progress in neurological surgery"] = "Prog Neurol Surg",
  ["progress in neurology and psychiatry"] = "Prog Neurol Psychiatry",
  ["progress in nonlinear differential equations and their applications"] = "Progr. Nonlinear Differential Equations Appl.",
  ["progress in nuclear energy"] = "Prog. Nucl. Energy",
  ["progress in nuclear magnetic resonance spectroscopy"] = "Prog Nucl Magn Reson Spectrosc",
  ["progress in nuclear medicine"] = "Prog Nucl Med",
  ["progress in nucleic acid research and molecular biology"] = "Prog Nucleic Acid Res Mol Biol",
  ["progress in oceanography"] = "Prog. Oceanogr.",
  ["progress in optical science and photonics"] = "Prog. Opt. Sci. Photonics",
  ["progress in organic coatings"] = "Prog. Org. Coat.",
  ["progress in orthodontics"] = "Prog Orthod",
  ["progress in pain research and management"] = "Prog Pain Res Manag",
  ["progress in palliative care"] = "Prog Palliat Care",
  ["progress in paper recycling"] = "Prog. Pap. Recycl.",
  ["progress in particle and nuclear physics"] = "Prog. Part. Nucl. Phys.",
  ["progress in pediatric cardiology"] = "Prog Pediatr Cardiol",
  ["progress in pediatric surgery"] = "Prog Pediatr Surg",
  ["progress in pharmacology and clinical pharmacology"] = "Prog Pharmacol Clin Pharmacol",
  ["progress in photovoltaics"] = "Prog Photovolt",
  ["progress in photovoltaics: research and applications"] = "Prog. Photovoltaics Res. Appl.",
  ["progress in physical geography"] = "Prog. Phys. Geogr.",
  ["progress in physical geography: earth and environment"] = "Prog. Phys. Geogr.: Earth Environ.",
  ["progress in physical therapy"] = "Prog Phys Ther",
  ["progress in physics"] = "Progr. Phys.",
  ["progress in planning"] = "Prog Plann",
  ["progress in polymer science"] = "Prog. Polym. Sci.",
  ["progress in preventive medicine (new york, n.y.)"] = "Prog Prev Med (N Y)",
  ["progress in probability"] = "Progr. Probab.",
  ["progress in psychobiology and physiological psychology"] = "Prog Psychobiol Physiol Psychol",
  ["progress in psychotherapy"] = "Prog Psychother",
  ["progress in pure and applied discrete mathematics"] = "Progr. Pure Appl. Discrete Math.",
  ["progress in quantum electronics"] = "Prog. Quantum Electron.",
  ["progress in radiation therapy"] = "Prog Radiat Ther",
  ["progress in reaction kinetics"] = "Prog. React. Kinet.",
  ["progress in reaction kinetics and mechanism"] = "Prog. React. Kinet. Mech.",
  ["progress in rehabilitation medicine"] = "Prog Rehabil Med",
  ["progress in respiration research"] = "Prog Respir Res",
  ["progress in respiratory research"] = "Prog. Respir. Res.",
  ["progress in retinal and eye research"] = "Prog Retin Eye Res",
  ["progress in rubber, plastics and recycling technology"] = "Prog. Rubber Plast. Recycl. Technol.",
  ["progress in solid state chemistry"] = "Prog. Solid State Chem.",
  ["progress in stem cell"] = "Prog Stem Cell",
  ["progress in stereochemistry"] = "Prog Stereochem",
  ["progress in structural engineering and materials"] = "Prog. Struct. Mater. Eng.",
  ["progress in surface science"] = "Prog. Surf. Sci.",
  ["progress in surgery"] = "Prog Surg",
  ["progress in systems and control theory"] = "Progr. Systems Control Theory",
  ["progress in the chemistry of fats and other lipids"] = "Prog Chem Fats Other Lipids",
  ["progress in the chemistry of organic natural products"] = "Prog Chem Org Nat Prod",
  ["progress in theoretical computer science"] = "Progr. Theoret. Comput. Sci.",
  ["progress in transplantation (aliso viejo, calif.)"] = "Prog Transplant",
  ["progress in tumor research"] = "Prog Tumor Res",
  ["progress in veterinary microbiology and immunology"] = "Prog Vet Microbiol Immunol",
  ["progress of mathematics"] = "Progr. Math. (Varanasi)",
  ["progress of theoretical and experimental physics"] = "Prog. Theor. Exp. Phys.",
  ["progress of theoretical physics"] = "Prog. Theor. Phys.",
  ["progress of theoretical physics supplement"] = "Prog. Theor. Phys. Suppl.",
  ["progress of theoretical physics supplements"] = "Prog. Theor. Phys. Suppl.",
  ["progress on chemistry and application of chitin and its derivatives"] = "Prog Chem Appl Chitin Deriv",
  ["progress pediatric study. american academy of pediatrics. committee for the study of child health services"] = "Prog Pediat Study",
  ["progress report of the center for population research, national institute of child health and human development. center for population research (national institute of child health and human development (u.s.))"] = "Prog Rep Cent Popul Res Natl Inst Child Health Hum Dev",
  ["progress reports on health & development in southern africa"] = "Prog Rep Health Dev South Afr",
  ["progressi di terapia"] = "Progress Ter",
  ["progressive architecture"] = "Prog Arch",
  ["progressive nurse"] = "Progress. Nurse",
  ["progresso medico"] = "Progr. Med. (Napoli)",
  ["progresso odontoiatrico"] = "Prog Odontoiatr",
  ["progressos da medicina"] = "Prog Med",
  ["progressus rei botanicae"] = "Progress. Rei Bot.",
  ["progrès en urologie : journal de l'association française d'urologie et de la société française d'urologie"] = "Prog Urol",
  ["project inform perspective"] = "Proj Inf Perspect",
  ["project mathematics!"] = "Proj. MATH!",
  ["project report. usaf school of aviation medicine"] = "Proj Rep USAF Sch Aviat Med",
  ["prologue (washington, d.c.)"] = "Prologue J Natl Arch",
  ["promed komplementär"] = "Promed Komplement",
  ["promega notes magazine"] = "Promega Notes",
  ["prometeo (milan, italy : 1983)"] = "Prometeo (Milan)",
  ["prometheus. rivista quadrimestrale di studi classici"] = "Prometheus",
  ["promoting health"] = "Promot Health",
  ["promotion & education"] = "Promot Educ",
  ["promotion dentaire"] = "Promot. Dent.",
  ["promotion et education"] = "Promot. Educ.",
  ["propellants, explosives, pyrotechnics"] = "Propellants Explos. Pyrotech.",
  ["prophylaxie sanitaire et morale"] = "Prophyl. Sanit. Morale",
  ["propulsion and power research"] = "Propul. Power Res.",
  ["propyläen kunstgeschichte"] = "PKG",
  ["prose studies"] = "Prose Stud",
  ["prosopographia imperii romani"] = "LPIR",
  ["prospettiva. rivista di storia dell’arte antica e moderna"] = "Prospettiva",
  ["prospezioni archeologiche. quaderni"] = "ProspAQuad",
  ["prospezioni. bollettino di informazioni scientifiche"] = "Prospezioni",
  ["prostaglandins & other lipid mediators"] = "Prostaglandins Other Lipid Mediat",
  ["prostaglandins and medicine"] = "Prostaglandins Med",
  ["prostaglandins and other lipid mediators"] = "Prostaglandins Other Lipid Mediators",
  ["prostaglandins leukotrienes and essential fatty acids"] = "Prostaglandins Leukot. Essent. Fatty Acids",
  ["prostaglandins, leukotrienes and essential fatty acids"] = "Prostaglandins, Leukotrienes Essent. Fatty Acids",
  ["prostaglandins, leukotrienes and medicine"] = "Prostaglandins. Leukot. Med.",
  ["prostaglandins, leukotrienes, and essential fatty acids"] = "Prostaglandins Leukot Essent Fatty Acids",
  ["prostaglandins, leukotrienes, and medicine"] = "Prostaglandins Leukot Med",
  ["prostate cancer"] = "Prostate Cancer",
  ["prostate cancer and prostatic diseases"] = "Prostate Cancer Prostatic Dis",
  ["prostate cancer imaging : image analysis and image-guided interventions : international workshop held in conjunction with miccai 2011, toronto, canada, september 22, 2011 : proceedings"] = "Prostate Cancer Imaging (2011)",
  ["prostate international"] = "Prostate Int",
  ["prostate. supplement"] = "Prostate. Suppl.",
  ["prosthetics and orthotics international"] = "Prosthet Orthot Int",
  ["protección social"] = "Prot Soc",
  ["protecting children"] = "Prot Child",
  ["protecting human subjects"] = "Prot Hum Subj",
  ["protection and control of modern power systems"] = "Prot. Control Mod. Power Syst.",
  ["protection de la nature"] = "Prot. nat.",
  ["protection of metals"] = "Prot. Met.",
  ["protection of metals and physial chemistry of surfaces"] = "Prot. Met. Phys. Chem. Surf.",
  ["protection of metals and physical chemistry of surfaces"] = "Prot. Met. Phys. Chem",
  ["protein & cell"] = "Protein Cell",
  ["protein & peptide letters"] = "Protein Pept. Lett.",
  ["protein alterations in erythrocytes at low temperatures. bjorksten research foundation"] = "Protein Alter Erythroc Low Temp",
  ["protein and peptide letters"] = "Protein Pept Lett",
  ["protein engineering"] = "Protein Eng",
  ["protein engineering, design & selection"] = "Protein Eng. Des. Sel.",
  ["protein engineering, design & selection : peds"] = "Protein Eng Des Sel",
  ["protein engineering, design and selection"] = "Protein Eng. Des. Sel.",
  ["protein expression and purification"] = "Protein Expr Purif",
  ["protein journal"] = "Protein J.",
  ["protein profile"] = "Protein Profile",
  ["protein reviews on the web"] = "Protein Rev. Web",
  ["protein science"] = "Protein Sci.",
  ["protein science : a publication of the protein society"] = "Protein Sci",
  ["protein sequences & data analysis"] = "Protein Seq Data Anal",
  ["protein sequences and data analysis"] = "Protein Seq. Data Anal.",
  ["proteins: structure, function, and bioinformatics"] = "Proteins Struct. Funct. Bioinf.",
  ["proteins: structure, function, and genetics"] = "Proteins: Struct., Funct., Genet.",
  ["proteoglycan research"] = "Proteoglycan Res.",
  ["proteome science"] = "Proteome Sci",
  ["proteomics & bioinformatics : current research"] = "Proteom Bioinform",
  ["proteomics - clinical applications"] = "Proteomics Clin. Appl.",
  ["proteomics insights"] = "Proteomics Insights",
  ["proteomics. clinical applications"] = "Proteomics Clin Appl",
  ["proteomics: clinical applications"] = "Proteomics: Clin. Appl.",
  ["proteómica : seprot journal"] = "Proteomica",
  ["protesis clínica; periódico técnico informativo para la profesión dental"] = "Protes Clin",
  ["protesista dental"] = "Protes. Dent.",
  ["protetyka stomatologiczna"] = "Protet Stomatol",
  ["proteus (shippensburg, pa.)"] = "Proteus",
  ["protides of the biological fluids; proceedings of the colloquium"] = "Protides Biol Fluid Proc Colloq",
  ["protist genomics"] = "Protist Genomics",
  ["protocol exchange"] = "Protoc Exch",
  ["provence historique"] = "ProvHist",
  ["provence historique; revue trimestrielle"] = "Provence Hist",
  ["provider (washington, d.c.)"] = "Provider",
  ["provider studies research note"] = "Provid Stud Res Note Agency Health Care Policy Res",
  ["provincial china"] = "Prov China",
  ["provincial medical & surgical journal"] = "Prov Med Surg J",
  ["provincial medical journal and retrospect of the medical sciences"] = "Prov Med J Retrosp Med Sci",
  ["proyecciones. journal of mathematics"] = "Proyecciones",
  ["pró-fono : revista de atualização científica"] = "Pro Fono",
  ["prx energy"] = "PRX Energy",
  ["prx life"] = "PRX Life",
  ["prx quantum"] = "PRX Quantum",
  ["przeglad dermatologiczny"] = "Przegl. Dermatol.",
  ["przeglad epidemiologiczny"] = "Przegl Epidemiol",
  ["przeglad geograficzny. polskii geograficheskii obzor. polish geographical review. revue polonaise de geographie"] = "Prz Geogr",
  ["przeglad historyczny"] = "PHist",
  ["przeglad lekarski"] = "Przegl Lek",
  ["przeglad orientalistyczny"] = "Prz Orient",
  ["przeglad polonijny"] = "Przegl Pol",
  ["przeglad zachodni"] = "Prz Zach",
  ["przegla̜d gastroenterologiczny"] = "Prz Gastroenterol",
  ["przegla̜d menopauzalny = menopause review"] = "Prz Menopauzalny",
  ["przegla̧d dermatologiczny"] = "Przegl Dermatol",
  ["przegla̧d epidemiologiczny"] = "Przegl Epidemiol",
  ["przegla̧d lekarski"] = "Przegl Lek",
  ["przegląd geograficzny. polśkiĭ geograficheskiĭ obzor. polish geographical review. revue polonaise de géographie"] = "Prz Geogr",
  ["przegląd historyczno-oświatowy"] = "Prz Hist Osw",
  ["przegląd historyczny"] = "Przegl Hist",
  ["przegląd orientalistyczny"] = "Prz Orient",
  ["przegląd polonijny"] = "Przegl Pol",
  ["przegląd zachodni"] = "Prz Zach",
  ["przemysl chemiczny"] = "Przem. Chem.",
  ["przeszlosc demograficzna polski"] = "Przeszlosc Demograf Pol",
  ["przeszłość demograficzna polski"] = "Przeszlosc Demograf Pol",
  ["prähistorische archäologie in südosteuropa"] = "PAS",
  ["prähistorische bronzefunde"] = "PBF",
  ["prähistorische zeitschrift"] = "PZ",
  ["prépublications de l’institut de recherche mathématique avancée"] = "Prépubl. Inst. Rech. Math. Av.",
  ["ps, political science & politics"] = "PS Polit Sci Polit",
  ["psd communication newsletter"] = "PSD Commun Newsl",
  ["pseudo-differential operators. theory and applications"] = "Pseudo Diff. Oper.",
  ["psi chi journal of undergraduate research"] = "Psi Chi J Undergrad Res",
  ["psi newsletter"] = "PSI Newsl",
  ["psichiatria generale e dell' età evolutiva"] = "Psichiatr Gen Evol",
  ["psicoanalisi applicata alla medicina, pedagogia, sociologia, letteratura ed arte"] = "Psicoanal Appl Med Pedagog Sociol Lett Arte",
  ["psicologia em estudo"] = "Psicol Estud",
  ["psicologia sociale"] = "Psicol Soc (Bologna)",
  ["psicologia, reflexão e crítica : revista semestral do departamento de psicologia da ufrgs"] = "Psicol Reflex Crit",
  ["psicología conductual"] = "Psicol Conductual",
  ["psicología educativa"] = "Psicol Educ (Madr)",
  ["psicológica : revista de metodología y psicología experimental"] = "Psicologica (Valencia)",
  ["psikhologicheskii zhurnal"] = "Psikholog Zh",
  ["psikhologicheskiĭ zhurnal"] = "Psikholog Zh",
  ["psikhologyah : ketav ʻet madaʻi yiśreʼeli le-ʻiyun ule-meḥḳar"] = "Psikhologyah",
  ["psiquiatria y ciencias afines actas luso esp neurol psiquiatr cien"] = "Afines",
  ["psoriasis (auckland, n.z.)"] = "Psoriasis (Auckl)",
  ["psoriasis forum"] = "Psoriasis Forum",
  ["psr quarterly (baltimore, md.)"] = "PSR Q",
  ["psych journal"] = "Psych J",
  ["psyche: a journal of entomology"] = "Psyche: J. Entomol.",
  ["psyche; a journal of entomology"] = "Psyche (Camb Mass)",
  ["psyche; revue internationale des sciences de l’homme et de psychanalyse"] = "Psyche Rev Int Sci Homme Psychanal",
  ["psychedelic medicine"] = "Psychedelic Med.",
  ["psyché; revue internationale des sciences de l'homme et de psychanalyse"] = "Psyche Rev Int Sci Homme Psychanal",
  ["psychiatria clinica"] = "Psychiatr Clin (Basel)",
  ["psychiatria danubina"] = "Psychiatr Danub",
  ["psychiatria et neurologia"] = "Psychiatr Neurol (Basel)",
  ["psychiatria hungarica : a magyar pszichiatriai tarsasag tudomanyos folyoirata"] = "Psychiatr Hung",
  ["psychiatria hungarica : a magyar pszichiátriai társaság tudományos folyóirata"] = "Psychiatr Hung",
  ["psychiatria polska"] = "Psychiatr Pol",
  ["psychiatria, neurologia, neurochirurgia"] = "Psychiatr Neurol Neurochir",
  ["psychiatric annals"] = "Psychiatr Ann",
  ["psychiatric bulletin (2014)"] = "Psychiatr Bull (2014)",
  ["psychiatric bulletin (london, england)"] = "Psychiatr Bull R Coll Psychiatr",
  ["psychiatric clinics of north america"] = "Psychiatr. Clin. North Am.",
  ["psychiatric communications"] = "Psychiatr Commun",
  ["psychiatric developments"] = "Psychiatr Dev",
  ["psychiatric forum"] = "Psychiatr Forum",
  ["psychiatric genetics"] = "Psychiatr Genet",
  ["psychiatric hospital"] = "Psychiatr. Hosp.",
  ["psychiatric journal of the university of ottawa"] = "Psychiatr. J. Univ. Ott.",
  ["psychiatric journal of the university of ottawa : revue de psychiatrie de l'université d'ottawa"] = "Psychiatr J Univ Ott",
  ["psychiatric journal of the university of ottawa : revue de psychiatrie de l’universite d’ottawa"] = "Psychiatr J Univ Ott",
  ["psychiatric medicine"] = "Psychiatr Med",
  ["psychiatric medicine update: massachusetts general hospital reviews for physicians"] = "Psychiatr Med Update Mass Gen Hosp Rev Physicians",
  ["psychiatric news"] = "Psychiatr News",
  ["psychiatric opinion"] = "Psychiatr Opin",
  ["psychiatric quarterly"] = "Psychiatr. Q.",
  ["psychiatric quarterly. supplement"] = "Psychiatr. Q. Suppl.",
  ["psychiatric rehabilitation journal"] = "Psychiatr Rehabil J",
  ["psychiatric research reports"] = "Psychiatr Res Rep Am Psychiatr Assoc",
  ["psychiatric services"] = "Psychiatr. Serv.",
  ["psychiatric services (washington, d.c.)"] = "Psychiatr Serv",
  ["psychiatrie (stuttgart, germany)"] = "Psychiatrie (Stuttg)",
  ["psychiatrie de l’enfant"] = "Psychiatr. Enfant",
  ["psychiatrie, neurologie und medizinische psychologie"] = "Psychiatr. Neurol. Med. Psychol. (Leipz.)",
  ["psychiatrie, neurologie und medizinische psychologie. beihefte"] = "Psychiatr Neurol Med Psychol Beih",
  ["psychiatrie, neurologie, und medizinische psychologie"] = "Psychiatr Neurol Med Psychol (Leipz)",
  ["psychiatrikē = psychiatriki"] = "Psychiatriki",
  ["psychiatrische praxis"] = "Psychiatr Prax",
  ["psychiatry (abingdon, england)"] = "Psychiatry (Abingdon)",
  ["psychiatry (edgmont (pa. : township))"] = "Psychiatry (Edgmont)",
  ["psychiatry and clinical neurosciences"] = "Psychiatry Clin Neurosci",
  ["psychiatry and clinical psychopharmacology"] = "Psychiatry Clin. Psychopharmacol.",
  ["psychiatry digest"] = "Psychiatry Dig",
  ["psychiatry in medicine"] = "Psychiatry Med",
  ["psychiatry investigation"] = "Psychiatry Investig",
  ["psychiatry journal"] = "Psychiatry J",
  ["psychiatry research"] = "Psychiatry Res",
  ["psychiatry research. neuroimaging"] = "Psychiatry Res Neuroimaging",
  ["psychiatry, psychology, and law : an interdisciplinary journal of the australian and new zealand association of psychiatry, psychology and law"] = "Psychiatr Psychol Law",
  ["psycho-social medicine"] = "Psychosoc Med",
  ["psychoanalysis and history"] = "Psychoanal Hist",
  ["psychoanalysis and the psychoanalytic review"] = "Psychoanal Psychoanal Rev",
  ["psychoanalysis and the social sciences"] = "Psychoanal Soc Sci",
  ["psychoanalytic dialogues"] = "Psychoanal Dialogues",
  ["psychoanalytic inquiry"] = "Psychoanal Inq",
  ["psychoanalytic psychology : the official journal of the division of psychoanalysis, american psychological association, division 39"] = "Psychoanal Psychol",
  ["psychoanalytic psychotherapy"] = "Psychoanal Psychother",
  ["psychoanalytic quarterly"] = "Psychoanal. Q.",
  ["psychoanalytic review"] = "Psychoanal Rev",
  ["psychoanalytic study of the child"] = "Psychoanal. Study Child",
  ["psychobiology (austin, tex.)"] = "Psychobiology (Austin, Tex)",
  ["psychodynamic counselling"] = "Psychodyn Couns",
  ["psychodynamic practice"] = "Psychodyn Pract",
  ["psychodynamic psychiatry"] = "Psychodyn Psychiatry",
  ["psychogeriatrics : the official journal of the japanese psychogeriatric society"] = "Psychogeriatrics",
  ["psychologia : an international journal of psychology in the orient"] = "Psychologia",
  ["psychologica belgica"] = "Psychol Belg",
  ["psychological abstracts"] = "Psychol Abstr",
  ["psychological assessment"] = "Psychol Assess",
  ["psychological bulletin"] = "Psychol Bull",
  ["psychological injury and law"] = "Psychol Inj Law",
  ["psychological inquiry"] = "Psychol Inq",
  ["psychological issues"] = "Psychol Issues",
  ["psychological medicine"] = "Psychol Med",
  ["psychological medicine. monograph supplement"] = "Psychol Med Monogr Suppl",
  ["psychological methods"] = "Psychol Methods",
  ["psychological monographs"] = "Psychol Monogr",
  ["psychological reports"] = "Psychol Rep",
  ["psychological research"] = "Psychol Res",
  ["psychological review"] = "Psychol Rev",
  ["psychological science"] = "Psychol Sci",
  ["psychological science : a journal of the american psychological society / aps"] = "Psychol Sci",
  ["psychological science agenda"] = "Psychol Sci Agenda",
  ["psychological science in the public interest : a journal of the american psychological society"] = "Psychol Sci Public Interest",
  ["psychological services"] = "Psychol Serv",
  ["psychological studies"] = "Psychol Stud (Mysore)",
  ["psychological studies in population/family planning"] = "Psychol Stud Popul Fam Plan",
  ["psychological test and assessment modeling"] = "Psychol Test Assess Model",
  ["psychological trauma : theory, research, practice and policy"] = "Psychol Trauma",
  ["psychologie & gezondheid"] = "Psychol Gezondh",
  ["psychologie & neuropsychiatrie du vieillissement"] = "Psychol Neuropsychiatr Vieil",
  ["psychologie et neuropsychiatrie du vieillissement"] = "Psychol. Neuropsychiatr. Vieil.",
  ["psychologie medicale"] = "Psychol Med (Paris)",
  ["psychologie médicale"] = "Psychol Med (Paris)",
  ["psychologische forschung"] = "Psychol Forsch",
  ["psychologische praxis"] = "Psychol. Prax.",
  ["psychologische praxis; schriftenreihe fur erziehung und jugendpflege"] = "Psychol Prax",
  ["psychologische praxis; schriftenreihe für erziehung und jugendpflege"] = "Psychol Prax",
  ["psychologische rundschau; ueberblick uber die fortschritte der psychologie in deutschland, oesterreich, und der schweiz"] = "Psychol Rundsch",
  ["psychologische rundschau; ueberblick über die fortschritte der psychologie in deutschland, oesterreich, und der schweiz"] = "Psychol Rundsch",
  ["psychology & health"] = "Psychol Health",
  ["psychology & marketing"] = "Psychol Mark",
  ["psychology & neuroscience"] = "Psychol Neurosci",
  ["psychology (irvine, calif.)"] = "Psychology (Irvine)",
  ["psychology (savannah, ga.)"] = "Psychology",
  ["psychology and aging"] = "Psychol Aging",
  ["psychology and behavioral sciences (new york, n.y. 2012)"] = "Psychol Behav Sci",
  ["psychology and cognitive sciences : open journal"] = "Psychol Cogn Sci",
  ["psychology and developing societies"] = "Psychol Dev Soc J",
  ["psychology and education"] = "Psychol Educ",
  ["psychology and mental health care : open access"] = "Psychol Ment Health Care",
  ["psychology and psychotherapy"] = "Psychol Psychother",
  ["psychology and sexuality"] = "Psychol Sex",
  ["psychology in the schools"] = "Psychol Sch",
  ["psychology of addictive behaviors : journal of the society of psychologists in addictive behaviors"] = "Psychol Addict Behav",
  ["psychology of aesthetics, creativity, and the arts"] = "Psychol Aesthet Creat Arts",
  ["psychology of consciousness (washington, d.c.)"] = "Psychol Conscious (Wash D C)",
  ["psychology of men & masculinity"] = "Psychol Men Masc",
  ["psychology of music"] = "Psychol Music",
  ["psychology of popular media culture"] = "Psychol Pop Media Cult",
  ["psychology of religion and spirituality"] = "Psycholog Relig Spiritual",
  ["psychology of sexual orientation and gender diversity"] = "Psychol Sex Orientat Gend Divers",
  ["psychology of sexualities review"] = "Psychol Sex Rev",
  ["psychology of sport and exercise"] = "Psychol Sport Exerc",
  ["psychology of violence"] = "Psychol Violence",
  ["psychology of well-being"] = "Psychol Well Being",
  ["psychology of women quarterly"] = "Psychol Women Q",
  ["psychology research (libertyville, ill.)"] = "Psychol Res (Libertyville)",
  ["psychology research and behavior management"] = "Psychol Res Behav Manag",
  ["psychology science quarterly"] = "Psychol Sci Q",
  ["psychology today"] = "Psychol Today",
  ["psychology, crime & law : pc & l"] = "Psychol Crime Law",
  ["psychology, health & medicine"] = "Psychol Health Med",
  ["psychology, public policy, and law : an official law review of the university of arizona college of law and the university of miami school of law"] = "Psychol Public Policy Law",
  ["psychonomic bulletin & review"] = "Psychon Bull Rev",
  ["psychonomic science"] = "Psychon Sci",
  ["psychopathology review"] = "Psychopathol Rev",
  ["psychopharm review : timely reports in psychopharmacology and device-based therapies"] = "Psychopharm Rev",
  ["psychopharmacology (berlin, germany)"] = "Psychopharmacology (Berlin, Ger.)",
  ["psychopharmacology (heidelberg, germany)"] = "Psychopharmacology (Heidelberg, Ger.)",
  ["psychopharmacology bulletin"] = "Psychopharmacol Bull",
  ["psychopharmacology communications"] = "Psychopharmacol Commun",
  ["psychopharmacology series"] = "Psychopharmacol Ser",
  ["psychopharmacology service center bulletin"] = "Psychopharmacol Serv Cent Bull",
  ["psychopharmacology. supplementum"] = "Psychopharmacology Suppl",
  ["psychosocial rehabilitation journal"] = "Psychosoc Rehabil J",
  ["psychosomatic medicine"] = "Psychosom Med",
  ["psychosomatik und konsiliarpsychiatrie"] = "Psychosom Konsiliarpsychiatr",
  ["psychotherapie und medizinische psychologie"] = "Psychother Med Psychol (Stuttg)",
  ["psychotherapie, medizinische psychologie"] = "Psychother Med Psychol (Stuttg)",
  ["psychotherapie, psychosomatik, medizinische psychologie"] = "Psychother Psychosom Med Psychol",
  ["psychotherapy (chicago, ill.)"] = "Psychotherapy (Chic)",
  ["psychotherapy and psychosomatics"] = "Psychother Psychosom",
  ["psychotherapy research : journal of the society for psychotherapy research"] = "Psychother Res",
  ["psychothérapies (geneva, switzerland)"] = "Psychotherapies (Geneva)",
  ["psychotropes (paris, france)"] = "Psychotr (Paris)",
  ["pszichologiai tanulmanyok"] = "Pszichol Tanulman",
  ["pszichológiai tanulmányok"] = "Pszichol Tanulman",
  ["ptep. progress of theoretical and experimental physics"] = "PTEP. Prog. Theor. Exp. Phys.",
  ["ptolemaeus arabus et latinus. studies"] = "Ptolemaeus Arab. Lat., Stud.",
  ["pubblicazioni del centro di sperimentazione agricola e forestale"] = "Pubbl. Cent. sper. agric. for.",
  ["pubblicazioni della stazione zoologica di napoli. section ii: history and philosophy of the life sciences"] = "Pubbl. Stn. Zool. Napoli II",
  ["pubblicazioni dell’università di ferrara"] = "Pubbl. Univ. Ferrara",
  ["public administration"] = "Public Adm",
  ["public administration and development : a journal of the royal institute of public administration"] = "Public Adm Dev",
  ["public administration review"] = "Public Adm Rev",
  ["public affairs quarterly"] = "Public Aff Q",
  ["public affairs report"] = "Public Aff. Rep.",
  ["public affairs report : bulletin of the bureau of public administration"] = "Public Aff Rep",
  ["public budgeting and finance"] = "Public Budg. Finance",
  ["public choice"] = "Public Choice",
  ["public citizen (washington, d.c.)"] = "Public Citiz",
  ["public finance"] = "Public Finance",
  ["public finance = finances publiques"] = "Public Financ",
  ["public finance quarterly"] = "Public Finan Q",
  ["public finance review"] = "Public Finance Rev.",
  ["public finance review : pfr"] = "Public Finance Rev",
  ["public health"] = "Public Health",
  ["public health (fairfax, va.)"] = "Public Health (Fairfax)",
  ["public health action"] = "Public Health Action",
  ["public health bulletin"] = "Public Health Bulletin (Wash D C)",
  ["public health ethics"] = "Public Health Ethics",
  ["public health frontier"] = "Public Health Front",
  ["public health genomics"] = "Public Health Genomics",
  ["public health in practice (oxford, england)"] = "Public Health Pract (Oxf)",
  ["public health monograph"] = "Public Health Monogr",
  ["public health news"] = "Public Health News",
  ["public health nursing"] = "Public Health Nurs",
  ["public health nursing (boston, mass.)"] = "Public Health Nurs",
  ["public health nutrition"] = "Public Health Nutr",
  ["public health panorama : journal of the who regional office for europe = panorama obshchestvennogo zdravookhranenii︠a︡"] = "Public Health Panor",
  ["public health papers"] = "Public Health Pap",
  ["public health papers and reports"] = "Public Health Pap Rep",
  ["public health reports"] = "Public Health Rep",
  ["public health reports (washington, d.c. : 1896)"] = "Public Health Rep",
  ["public health reports (washington, d.c. : 1974)"] = "Public Health Rep",
  ["public health research & practice"] = "Public Health Res Pract",
  ["public health research (southampton, england)"] = "Public Health Res (Southampt)",
  ["public health reviews"] = "Public Health Rev",
  ["public health technical monograph. united states. public health service"] = "Public health tech monogr US",
  ["public history news"] = "Public Hist News",
  ["public interest"] = "Public Interest",
  ["public law"] = "Public Law",
  ["public law forum"] = "Public Law Forum",
  ["public management sources"] = "Public Management Sources",
  ["public opinion quarterly"] = "Public Opin Q",
  ["public personnel management"] = "Public Pers Manage",
  ["public policy"] = "Public Policy",
  ["public policy and administration"] = "Public Policy Adm",
  ["public relations journal"] = "Public Relat. J.",
  ["public relations review"] = "Public Relat Rev",
  ["public sector contracting report"] = "Public Sect. Contract. Rep.",
  ["public sector contracting report : the monthly guide to medicare and medicaid managed care"] = "Public Sect Contract Rep",
  ["public sector, health care risk management"] = "Public Sector Health Care Risk Manage",
  ["public service review. uk science & technology"] = "Public Serv Rev UK Sci Technol",
  ["public transport"] = "Public Transp.",
  ["public understanding of science (bristol, england)"] = "Public Underst Sci",
  ["public welfare"] = "Public Welf",
  ["public welfare in indiana"] = "Public Welf Indiana",
  ["public works management & policy"] = "Public Works Manage. Policy",
  ["publicacion electronica de la asociacion paleontologica argentina"] = "Publ. Electron. Paleontol. Asoc.",
  ["publicaciones científicas alter"] = "Publ Cient Alter",
  ["publicaciones de la institución »tello téllez de meneses«"] = "PublInstTTMeneses",
  ["publicaciones del departamento de matemáticas, universidad de extremadura"] = "Publ. Dep. Mat. Univ. Extremadura",
  ["publicaciones del museo de la farmacia"] = "Publ Mus Farm",
  ["publicaciones del observatorio de santiago, ii. consejo superior de investigaciones cientificas, instituto nacional de geofisica"] = "Publicaciones del Observatorio de Santiago, II. Consejo Superior de Investigaciones Cientificas, Instituto Nacional de G",
  ["publicaciones matematicas del uruguay"] = "Publ. Mat. Urug.",
  ["publicaciones matemáticas del uruguay"] = "Publ. Mat. Urug.",
  ["publicaciones médicas"] = "Publ Med",
  ["publicaciones. buenos aires (argentina). centro de investigaciones tisiológicas"] = "Publ B Aires Cent Investig",
  ["publicaciones. buenos aires. centro de investigaciones tisiologicas"] = "Publ B Aires Cent Investig",
  ["publicaciones. universidad de chile. laboratorio de medicina experimental"] = "Publ Univ Chile",
  ["publicacions de la seccio de mathematiques"] = "Publ. Sec. Mat.",
  ["publicacions matematiques"] = "Publ. Mat.",
  ["publicacions matemàtiques"] = "Publ. Mat.",
  ["publicacions matemã tiques"] = "Publ. Mat.",
  ["publicación científica - oficina sanitaria panamericana"] = "Publ Cient Oficina Sanit Panam",
  ["publicacoes do departamento de matematica"] = "Publ. Dep. Mat.",
  ["publicacoes matematicas do impa"] = "Publ. Mat. IMPA",
  ["publicações farmacêuticas"] = "Publ Farm (Sao Paulo)",
  ["publicações médicas"] = "Publicacoes Med",
  ["publication - american institute of the history of pharmacy"] = "Publ Am Inst Hist Pharm",
  ["publication - group for the advancement of psychiatry"] = "Publ Group Adv Psychiatry",
  ["publication de l’institut de recherche mathématique avancée"] = "Publ. Inst. Rech. Math. Av.",
  ["publication of the society for literature and science: psls"] = "Publ Soc Lit Sci",
  ["publication, group for the advancement of psychiatry"] = "Publ. Group Adv. Psychiatry",
  ["publication. cayenne, french guiana. institut pasteur de la guyane francaise et de l’inini"] = "Publ Inst Pasteur Guyane Fr Inini",
  ["publication. cayenne, french guiana. institut pasteur de la guyane française et de l'inini"] = "Publ Inst Pasteur Guyane Fr Inini",
  ["publicationes mathematicae debrecen"] = "Publ. Math. Debrecen",
  ["publications de la cour européenne des droits de l'homme. série a, arrêts et décisions = publications of the european court of human rights. series a, judgments and decisions. european court of human rights"] = "Publ Cour Eur Droits Homme Sr A Arrets Decis",
  ["publications de la societe historique et archeologique dans le limbourg. limburgs geschied- en oudheidkundig genootschap"] = "Publ Soc Hist Archeol Limbg",
  ["publications de la société historique et archéologique dans le limbourg. limburgs geschied- en oudheidkundig genootschap"] = "Publ Soc Hist Archeol Limbg",
  ["publications de l’institut de mathematique de l’universite de strasbourg, xiv. actualites scientifiques et industrielles"] = "Publications de l’Institut de Mathematique de l’Universite de Strasbourg, XIV. Actualites Scientifiques et Industr",
  ["publications de l’institut de mathematique de l’universite de strasbourg, xv. actualites scientifiques et industrielles"] = "Publications de l’Institut de Mathematique de l’Universite de Strasbourg, XV. Actualites Scientifiques et Industri",
  ["publications de l’institut de recherche mathématique de rennes"] = "Publ. Inst. Rech. Math. Rennes",
  ["publications des archives henri-poincare"] = "Publ. Arch. Henri-Poincare",
  ["publications du centre de recherches en mathématiques pures, série ii"] = "Publ. Centre Rech. Math. Pures Sér. II",
  ["publications du centre de recherches en mathématiques pures, séries i"] = "Publ. Cent. Rech. Math. Pures Sér. I",
  ["publications in child development. university of california, berkeley"] = "Publ Child Dev Univ Calif",
  ["publications in indiana medical history"] = "Publ Ind Med Hist",
  ["publications in psychology"] = "Publ Psychol",
  ["publications in psychology. university of california (1868-1952)"] = "Publ Psychol",
  ["publications in public health"] = "Publ Public Health Univ Calif",
  ["publications in public health. university of california (1868-1952)"] = "Publ Public Health Univ Calif",
  ["publications mathematiques de besancon. algebre et theorie des nombres"] = "Publ. Math. Besancon Algebre Theorie Nr.",
  ["publications mathematiques de l’ihes"] = "Publ. Math. IHES",
  ["publications mathematiques. institut de hautes etudes scientifiques"] = "Publ. Math. Inst. Hautes Etudes Sci.",
  ["publications mathématiques de la faculté des sciences de besançon"] = "Publ. Math. Fac. Sci. Besançon",
  ["publications mathématiques de l’université paris 7—denis diderot"] = "Publ. Math. Univ. Paris 7—Denis Diderot",
  ["publications mathématiques de l’université pierre et marie curie"] = "Publ. Math. Univ. Pierre et Marie Curie",
  ["publications mathématiques d’orsay"] = "Publ. Math. Orsay",
  ["publications of the astronomical society of australia"] = "Publ. Astron. Soc. Aust.",
  ["publications of the astronomical society of australia : pasa"] = "Publ Astron Soc Aust",
  ["publications of the astronomical society of japan"] = "Publ. Astron. Soc. Jpn.",
  ["publications of the astronomical society of japan. nihon tenmon gakkai"] = "Publ Astron Soc Jpn Nihon Tenmon Gakkai",
  ["publications of the astronomical society of the pacific"] = "Publ. Astron. Soc. Pac.",
  ["publications of the astronomical society of the pacific. astronomical society of the pacific"] = "Publ Astron Soc Pac",
  ["publications of the australasian institute of mining and metallurgy"] = "Publ. Australas. Inst. Min. Metall.",
  ["publications of the austrian ludwig wittgenstein society. new series"] = "Publ. Austrian Ludwig Wittgenstein Soc. New Ser.",
  ["publications of the faculty of electrical engineering"] = "Publ. Fac. Electr. Eng. Ser. Automat. Control",
  ["publications of the german historical institute"] = "Publ. Ger. Hist. Inst.",
  ["publications of the hochiminh city mathematical society"] = "Publ. HoChiMinh City Math. Soc.",
  ["publications of the international research center, salzburg, austria"] = "Publ. Int. Res. Cent. Salzbg. Austria",
  ["publications of the netherlands interdisciplinary demographic institute (n.i.d.i) and the population and family study centre (c.b.g.s.)"] = "Publ Neth Interuniv Demogr Inst N.I.D.I. Popul Fam Study Cent C.B.G.S.",
  ["publications of the newton institute"] = "Publ. Newton Inst.",
  ["publications of the research institute for mathematical sciences"] = "Publ. Res. Inst. Math. Sci.",
  ["publications of the university of eastern finland. reports and studies in forestry and natural sciences"] = "Publ. Univ. East. Finl. Rep. Stud. For. Nat. Sci.",
  ["publications of the university of joensuu, series bi"] = "Publ. Univ. Joensuu, Ser. BI",
  ["publikationen des henri-poincaré-archivs"] = "Publ. Henri-Poincaré-Arch.",
  ["publishing research quarterly"] = "Publ. Res. Q.",
  ["puerto rico health sciences journal"] = "P R Health Sci J",
  ["puerto rico y su enfermera"] = "P R Enferm",
  ["pulaski county historical review"] = "Pulaski Cty Hist Rev",
  ["pulmonary and critical care medicine"] = "Pulm Crit Care Med",
  ["pulmonary circulation"] = "Pulm Circ",
  ["pulmonary medicine"] = "Pulm Med",
  ["pulmonary pharmacology"] = "Pulm Pharmacol",
  ["pulmonary pharmacology & therapeutics"] = "Pulm Pharmacol Ther",
  ["pulmonary pharmacology and therapeutics"] = "Pulm. Pharmacol. Ther.",
  ["pulmonary research and respiratory medicine : open journal"] = "Pulm Res Respir Med",
  ["pulmonary therapy"] = "Pulm Ther",
  ["pulpudeva. semaines philippopolitaines de l’histoire et de la culture thrace"] = "Pulpudeva",
  ["pulse (basel, switzerland)"] = "Pulse (Basel)",
  ["pulse of the montana state nurses association"] = "Pulse",
  ["punishment & society"] = "Punishm Soc",
  ["punjab journal of politics"] = "Punjab J Polit",
  ["punjab university economist"] = "Punjab Univ. Econ.",
  ["purchasing administration"] = "Purch Adm",
  ["purdon's pennsylvania statutes, annotated. pennsylvania"] = "Purdons Pa Statut Annot Pa",
  ["purdon’s pennsylvania statutes, annotated. pennsylvania"] = "Purdons Pa Statut Annot Pa",
  ["pure and applied analysis"] = "Pure Appl. Anal.",
  ["pure and applied chemistry"] = "Pure Appl. Chem.",
  ["pure and applied chemistry. chimie pure et appliquee"] = "Pure Appl Chem",
  ["pure and applied chemistry. chimie pure et appliquée"] = "Pure Appl Chem",
  ["pure and applied functional analysis"] = "Pure Appl. Funct. Anal.",
  ["pure and applied geophysics"] = "Pure Appl. Geophys.",
  ["pure and applied mathematical quarterly"] = "Pure Appl. Math. Q.",
  ["pure and applied mathematics"] = "Pure Appl. Math.",
  ["pure and applied mathematics (amsterdam)"] = "Pure Appl. Math. (Amst.)",
  ["pure and applied mathematics (boca raton)"] = "Pure Appl. Math. (Boca Raton)",
  ["pure and applied mathematics (hoboken)"] = "Pure Appl. Math. (Hoboken)",
  ["pure and applied mathematics quarterly"] = "Pure Appl. Math. Q.",
  ["pure and applied mathematika sciences"] = "Pure Appl. Math. Sci.",
  ["pure and applied optics"] = "Pure Appl. Opt.",
  ["pure and applied physics"] = "Pure Appl. Phys.",
  ["pure and applied undergraduate texts"] = "Pure Appl. Undergrad. Texts",
  ["pure culture study of bacteria"] = "Pure Cult Study Bact",
  ["pure mathematics and applications"] = "Pure Math. Appl.",
  ["pure mathematics and applications. pu.m.a."] = "Pure Math. Appl. (PU.M.A.)",
  ["pure mathematics manuscript"] = "Pure Math. Manuscript",
  ["purinergic signalling"] = "Purinergic Signal",
  ["pusat media pelatihan aids untuk wartawan newsletter"] = "Pus Media Pelatih AIDS Untuk War Newsl",
  ["puteoli. studi di storia antica"] = "Puteoli",
  ["pyrenae. crónica arqueológica"] = "Pyrenae",
  ["pyrethrum post"] = "Pyrethrum Post",
  ["pz wissenschaft : pharmazeutische zeitung, wissenschaftsausgabe"] = "PZ Wiss",
  ["q.a. brief"] = "QA Brief",
  ["qa review"] = "QA Rev.",
  ["qa review : quality assurance news and views"] = "QA Rev",
  ["qatar medical journal"] = "Qatar Med J",
  ["qatar university science journal"] = "Qatar Univ. Sci. J.",
  ["qing shi yan jiu = qingshi yanjiu"] = "Qing Shi Yan Jiu",
  ["qjm : monthly journal of the association of physicians"] = "QJM",
  ["qmw maths notes"] = "QMW Maths Notes",
  ["qp–pq: quantum probability and white noise analysis"] = "QP–PQ: Quantum Probab. White Noise Anal.",
  ["qrb discovery"] = "QRB Discov",
  ["qrb. quality review bulletin"] = "QRB Qual Rev Bull",
  ["qrb; quality review bulletin"] = "QRB Qual. Rev. Bull.",
  ["qrc advisor"] = "QRC Advis",
  ["qsar & combinatorial science"] = "QSAR Comb. Sci.",
  ["qscience proceedings"] = "Qsci Proc",
  ["quaderni aquileiesi"] = "QuadAquil",
  ["quaderni catanesi di cultura classica e medievale"] = "QCCCM",
  ["quaderni catanesi di studi classici e medievali"] = "QC",
  ["quaderni del dipartimento delle arti, università della calabria"] = "QuadACal",
  ["quaderni del gruppo archeologico ostigliese"] = "QuadAOst",
  ["quaderni del museo archeologico f. ribezzo di brindisi"] = "QMAB",
  ["quaderni del museo archeologico regionale antonino salinas"] = "QuadMusSalinas",
  ["quaderni del raggruppamento tosco-umbro-emiliano di storia della medicina"] = "Quad Raggruppamento Toscoumbroemiliano Stor Med",
  ["quaderni della nutrizione"] = "Quad Nutr",
  ["quaderni della soprintendenza archeologica del piemonte"] = "QuadAPiem",
  ["quaderni dell’ accademia pontaniana"] = "Quad. Accad. Pontaniana",
  ["quaderni dell’ingegnere chimico italiano"] = "Quad. Ing. Chim. Ital.",
  ["quaderni dell’ist. di filosofia, univ. degli studi di perugia, fac. di magistero"] = "QIFP",
  ["quaderni dell’istituto di archeologia e storia antica dell’univ. g. d’annunzio, chieti"] = "QIASA",
  ["quaderni dell’istituto di archeologia e storia antica, università di chieti"] = "QuadChieti",
  ["quaderni dell’istituto di filologia greca"] = "QuadIstFilGr",
  ["quaderni dell’istituto di filologia greca di cagliari"] = "QIFG",
  ["quaderni dell’istituto di lingua e letteratura latina, università di roma"] = "QuadIstLat",
  ["quaderni dell’istituto di lingua e letteratura latina. università di roma, fac. di magistero"] = "QILL",
  ["quaderni di antibiotica"] = "Quad Antibiot",
  ["quaderni di archeologia del mantovano"] = "QuadAMant",
  ["quaderni di archeologia del veneto"] = "QuadAVen",
  ["quaderni di archeologia della libia"] = "QAL",
  ["quaderni di archeologia, università di messina"] = "QuadAMess",
  ["quaderni di azione sociale"] = "Quad Azione Soc",
  ["quaderni di clinica ostetrica e ginecologica"] = "Quad Clin Ostet Ginecol",
  ["quaderni di criminologia clinica"] = "Quad Criminol Clin",
  ["quaderni di cultura e di tradizione classica"] = "QCTC",
  ["quaderni di economia e finanza"] = "Quaderni Econ. Finanza",
  ["quaderni di filologia classica dell’università di trieste, ist. di filol. class."] = "QFC",
  ["quaderni di gerico"] = "QuadGerico",
  ["quaderni di matematica"] = "Quad. Mat.",
  ["quaderni di merceologia"] = "Quad Merceol",
  ["quaderni di protostoria. università di perugia"] = "QuadProtost",
  ["quaderni di radiologia"] = "Quad Radiol",
  ["quaderni di ricerca in didattica"] = "Quad. Ric. Didatt.",
  ["quaderni di statistica"] = "Quad Stat",
  ["quaderni di storia"] = "QS",
  ["quaderni di storia della medicina e della scienza"] = "Quad Stor Med Sci",
  ["quaderni d’archeologia reggiana"] = "QuadAReggio",
  ["quaderni erbesi. civico museo archeologico di erba"] = "QuadErb",
  ["quaderni friulani di archeologia"] = "QuadFriulA",
  ["quaderni internazionali di storia della medicina e della sanita"] = "Quad Int Stor Med Sanita",
  ["quaderni internazionali di storia della medicina e della sanità"] = "Quad Int Stor Med Sanita",
  ["quaderni linguistici e filologici: ricerche svolte presso l’univ. degli studi di macerata"] = "QLF",
  ["quaderni medici"] = "Quadern  Med",
  ["quaderni per la storia dell'università di padova"] = "Quad Stor Univ Padova",
  ["quaderni per la storia dell’universita di padova"] = "Quad Stor Univ Padova",
  ["quaderni sclavo di diagnostica clinica e di laboratorio"] = "Quad Sclavo Diagn",
  ["quaderni storici"] = "Quad Stor",
  ["quaderni storici delle marche"] = "Quaderni Storici delle Marche",
  ["quaderni urbinati di cultura classica"] = "QUCC",
  ["quaderni. centro studi lunensi"] = "QuadStLun",
  ["quaderni. istituto di lingue e letterature classiche, facoltà di magistero, università degli studi, lecce"] = "QuadLecce",
  ["quaderni. scuola normale superiore di pisa (nuova serie)"] = "Quad. Sc. Norm. Super. di Pisa (N. S.)",
  ["quaderni. scuola normale superiore di pisa. selections"] = "Quad. Sc. Norm. Super. Pisa Sel.",
  ["quaderni. soprintendenza archeologica per la provincie di cagliari e oristano"] = "QuadACagl",
  ["quaderns de prehistòria i arqueologia de castelló"] = "QuadCast",
  ["quaderns d’arqueologia i història de la ciutat (barcelona)"] = "QuadABarcel",
  ["quadrature. magazine de mathematiques pures et epicees"] = "Quadrature",
  ["quaestiones mathematicae"] = "Quaestiones Math.",
  ["quaestiones mathematicae : journal of the south african mathematical society"] = "Quaest Math",
  ["quaestiones mathematicae. journal of the south african mathematical society"] = "Quaest. Math.",
  ["quaker history"] = "Quaker Hist",
  ["qualitative and quantitative analysis of scientific and scholarly communication"] = "Qual. Quant. Anal. Sci. Sch. Commun.",
  ["qualitative health research"] = "Qual Health Res",
  ["qualitative inquiry : qi"] = "Qual Inq",
  ["qualitative psychology (washington, d.c.)"] = "Qual Psychol",
  ["qualitative report (online)"] = "Qual Rep",
  ["qualitative research"] = "Qual. Res.",
  ["qualitative research : qr"] = "Qual Res",
  ["qualitative research in medicine & healthcare"] = "Qual Res Med Healthc",
  ["qualitative research in psychology"] = "Qual Res Psychol",
  ["qualitative research in sport and exercise"] = "Qual Res Sport Exerc",
  ["qualitative research in sport, exercise and health"] = "Qual Res Sport Exerc Health",
  ["qualitative research journal"] = "Qual. Res. J.",
  ["qualitative social work : qsw : research and practice"] = "Qual Soc Work",
  ["qualitative sociology"] = "Qual Sociol",
  ["qualitative theory of dynamical systems"] = "Qual. Theory Dyn. Syst.",
  ["quality & quantity"] = "Qual Quant",
  ["quality & safety in health care"] = "Qual Saf Health Care",
  ["quality (new york, n.y.)"] = "Quality",
  ["quality and quantity"] = "Qual. Quant.",
  ["quality and reliability"] = "Qual. Reliab.",
  ["quality and reliability engineering international"] = "Qual. Reliab. Eng. Int.",
  ["quality and user experience"] = "Qual User Exp",
  ["quality assurance"] = "Qual. Assur.",
  ["quality assurance & safety of crops and food"] = "Qual. Assur. Saf. Crops Food",
  ["quality assurance (san diego, calif.)"] = "Qual Assur",
  ["quality assurance and utilization review"] = "Qual. Assur. Util. Rev.",
  ["quality assurance and utilization review : official journal of the american college of utilization review physicians"] = "Qual Assur Util Rev",
  ["quality assurance in health care"] = "Qual. Assur. Health Care",
  ["quality assurance in health care : the official journal of the international society for quality assurance in health care"] = "Qual Assur Health Care",
  ["quality assurance in health care : the official journal of the international society for quality assurance in health care / isqa"] = "Qual Assur Health Care",
  ["quality connection"] = "Qual. Connect.",
  ["quality connection (brookline, mass.)"] = "Qual Connect",
  ["quality engineering"] = "Qual. Eng.",
  ["quality in ageing : policy, practice and research"] = "Qual Ageing",
  ["quality in ageing and older adults"] = "Qual Ageing Older Adults",
  ["quality in health care"] = "Qual. Health Care",
  ["quality in health care : qhc"] = "Qual Health Care",
  ["quality in primary care"] = "Qual Prim Care",
  ["quality letter for healthcare leaders"] = "Qual. Lett. Healthc. Lead.",
  ["quality management in health care"] = "Qual Manag Health Care",
  ["quality of life and cardiovascular care"] = "Qual Life Cardiovasc Care",
  ["quality of life research"] = "Qual. Life Res.",
  ["quality of life research : an international journal of quality of life aspects of treatment, care and rehabilitation"] = "Qual Life Res",
  ["quality progress"] = "Qual Prog",
  ["quality technology & quantitative management"] = "Qual Technol Quant Manag",
  ["quantitative and network biology"] = "Quant. Netw. Biol.",
  ["quantitative applications in the social sciences"] = "Quantitative Appl. in the Social Sciences",
  ["quantitative bio-science"] = "Quant Biosci",
  ["quantitative biology"] = "Quant. Biol.",
  ["quantitative biology (beijing, china)"] = "Quant Biol",
  ["quantitative economics"] = "Quant Econom",
  ["quantitative finance"] = "Quant. Finance",
  ["quantitative geology and geostatistics"] = "Quant. Geol. Geostat.",
  ["quantitative imaging in medicine and surgery"] = "Quant Imaging Med Surg",
  ["quantitative infrared themography journal"] = "Quant. InfraRed Thermogr. J.",
  ["quantitative infrared thermography journal"] = "Quant Infrared Thermogr J",
  ["quantitative methoden der unternehmungsplanung"] = "Quantitative Method. Unternehmungsplanung",
  ["quantitative methods in the humanities and social sciences"] = "Quant. Methods Humanit. Soc. Sci.",
  ["quantitative microbiology"] = "Quant. Microbiol.",
  ["quantitative perspectives on behavioral economics and finance"] = "Quant. Perspect. Behav. Econ. Finance",
  ["quantitative plant biology"] = "Quant. Plant Biol.",
  ["quantitative science studies"] = "Quant. Sci. Stud.",
  ["quantitative structure-activity relationships"] = "Quant. Struct.-Act. Relat.",
  ["quantitative studies in social relations"] = "Quantitative Stud. Social Relations",
  ["quantum and semiclassical optics"] = "Quantum Semiclassical Opt.",
  ["quantum beam science"] = "Quantum Beam Sci.",
  ["quantum bio-informatics v : proceedings of the quantum bio-informatics 2011, tokyo university of science, japan, 7-12 march 2011. quantum bio-informatics (conference) (5th : 2011 : tokyo, japan)"] = "Quantum Bioinform V (2011)",
  ["quantum bioinformatics iv : from quantum information to bio-informatics : tokyo university of science, japan, 10-13 march 2010"] = "Quantum Bioinform IV (2010)",
  ["quantum electronics"] = "Quantum Electron.",
  ["quantum electronics—principles and applications"] = "Quantum Electron.—Principles Appl.",
  ["quantum engineering"] = "Quantum Eng.",
  ["quantum frontiers"] = "Quantum Front.",
  ["quantum information & computation"] = "Quantum Inf Comput",
  ["quantum information and computation"] = "Quantum Inf. Comput.",
  ["quantum information processing"] = "Quantum Inf. Process.",
  ["quantum machine intelligence"] = "Quantum Mach Intell",
  ["quantum optics"] = "Quantum Opt.",
  ["quantum reports"] = "Quantum Rep.",
  ["quantum science and technology"] = "Quantum Sci. Technol.",
  ["quantum studies. mathematics and foundations"] = "Quantum Stud. Math. Found.",
  ["quantum studies: mathematics and foundations"] = "Quantum Stud.: Math. Found.",
  ["quantum topology"] = "Quantum Topol.",
  ["quarterly (plant growth regulator society of america)"] = "PGRSA Q",
  ["quarterly - northeastern nevada historical society. northeastern nevada historical society"] = "Northeast Nev Hist Soc Q",
  ["quarterly bulletin (commonwealth medical association)"] = "Q Bull Commonw Med Assoc",
  ["quarterly bulletin of chinese bibliography"] = "Q Bull Chin Bibliogr",
  ["quarterly bulletin of sea view hospital. new york. sea view hospital, staten island. clinical society"] = "Q Bull Sea View Hosp",
  ["quarterly bulletin of the international association of agricultural information specialists"] = "Q. Bull. Int. Assoc. Agric. Inf. Spec.",
  ["quarterly bulletin of the south african library = kwartaalblad van die suid-afrikaanse biblioteek"] = "Q Bull S Afr Libr",
  ["quarterly bulletin. louisiana. board of health"] = "Q Bull La Board Health",
  ["quarterly bulletin. new york (n.y.). department of health"] = "Q Bull",
  ["quarterly bulletin. northwestern university (evanston, ill.). medical school"] = "Q Bull Northwest Univ Med Sch",
  ["quarterly demographic bulletin"] = "Q Demogr Bull",
  ["quarterly dental review"] = "Q Dent Rev",
  ["quarterly economic commentary"] = "Q Econ Comment",
  ["quarterly journal of austrian economics"] = "Quart. J. Austrian Econ.",
  ["quarterly journal of business and economics"] = "Quart. J. Bus. Econ.",
  ["quarterly journal of crude drug research = vierteljährliche zietschrift für drogen-forschung = revue trimestrielle des recherches sur les matières premières"] = "Q J Crude Drug Res",
  ["quarterly journal of economics"] = "Quart. J. Econ.",
  ["quarterly journal of engineering geology"] = "Q. J. Eng. Geol.",
  ["quarterly journal of engineering geology and hydrogeology"] = "Q. J. Eng. Geol. Hydrogeol.",
  ["quarterly journal of experimental physiology"] = "Q. J. Exp. Physiol.",
  ["quarterly journal of experimental physiology (cambridge, england)"] = "Q J Exp Physiol",
  ["quarterly journal of experimental physiology and cognate medical sciences"] = "Q J Exp Physiol Cogn Med Sci",
  ["quarterly journal of experimental psychology"] = "Q. J. Exp. Psychol.",
  ["quarterly journal of experimental psychology (2006)"] = "Q J Exp Psychol (Hove)",
  ["quarterly journal of experimental psychology. a, human experimental psychology"] = "Q. J. Exp. Psychol. A",
  ["quarterly journal of experimental psychology. b, comparative and physiological psychology"] = "Q. J. Exp. Psychol. B",
  ["quarterly journal of indian economic studies in social science"] = "Quart. J. Indian Econ. Stud. Soc. Sci.",
  ["quarterly journal of international agriculture"] = "Q. J. Int. Agric.",
  ["quarterly journal of mathematics"] = "Q. J. Math.",
  ["quarterly journal of mechanics and applied mathematics"] = "Q. J. Mech. Appl. Math.",
  ["quarterly journal of medicine"] = "Q. J. Med.",
  ["quarterly journal of mental health"] = "Q J Ment Health",
  ["quarterly journal of microscopical science"] = "Q. J. Microsc. Sci.",
  ["quarterly journal of nuclear medicine"] = "Q. J. Nucl. Med.",
  ["quarterly journal of nuclear medicine and molecular imaging"] = "Q. J. Nucl. Med. Mol. Imaging",
  ["quarterly journal of pharmacy and pharmacology"] = "Q J Pharm Pharmacol",
  ["quarterly journal of political science"] = "Quart J Polit Sci",
  ["quarterly journal of science, literature, and the arts"] = "Q. J. Sci. Lit. Arts",
  ["quarterly journal of speech"] = "QJS",
  ["quarterly journal of studies on alcohol"] = "Q. J. Stud. Alcohol",
  ["quarterly journal of the chemical society"] = "Q. J. Chem. Soc.",
  ["quarterly journal of the geological society of london"] = "Q. J. Geol. Soc. London",
  ["quarterly journal of the royal astronomical society"] = "Q. J. R. Astron. Soc.",
  ["quarterly journal of the royal meteorological society"] = "Q. J. R. Meteorolog. Soc.",
  ["quarterly journal of the royal meteorological society. royal meteorological society (great britain)"] = "Q J R Meteorol Soc",
  ["quarterly journal. library of congress"] = "Q J Libr Cong",
  ["quarterly medical review"] = "Q Med Rev",
  ["quarterly of applied mathematics"] = "Q. Appl. Math.",
  ["quarterly of magnetic resonance in biology and medicine"] = "Q. Magn. Reson. Biol. Med.",
  ["quarterly of the national dental association"] = "Q. Natl. Dent. Assoc.",
  ["quarterly population bulletin. new zealand. department of statistics"] = "Q Popul Bull",
  ["quarterly population bulletin. new zealand. dept. of statistics"] = "Q Popul Bull",
  ["quarterly progress report. united states. air force. radiation laboratory, university of chicago"] = "Q Prog Rep United States Air Force Radiat Lab Univ Chic",
  ["quarterly research report. usaf school of aviation medicine"] = "Q Res Rep",
  ["quarterly review"] = "Q. Rev. DC. Nurses. Assoc.",
  ["quarterly review (london, england)"] = "Q Rev",
  ["quarterly review of agricultural economics"] = "Quart. Rev. Agr. Econ.",
  ["quarterly review of allergy and applied immunology"] = "Q Rev Allergy Appl Immunol",
  ["quarterly review of biology"] = "Q. Rev. Biol.",
  ["quarterly review of distance education"] = "Q Rev Distance Educ",
  ["quarterly review of economics and business"] = "Quart. Rev. Econ. Bus.",
  ["quarterly review of economics and finance"] = "Quart. Rev. Econ. Finance",
  ["quarterly review of medicine"] = "Q Rev Med",
  ["quarterly review of obstetrics and gynecology"] = "Q Rev Obstet Gynecol",
  ["quarterly review of otorhinolaryngology and broncho-esophagology"] = "Q Rev Otorhinolaryngol Bronchoesophagol",
  ["quarterly review of pediatrics"] = "Q Rev Pediatr",
  ["quarterly review of psychiatry and neurology"] = "Q Rev Psychiatry Neurol",
  ["quarterly review of surgery"] = "Q Rev Surg",
  ["quarterly review of surgery, obstetrics and gynecology"] = "Q Rev Surg Obstet Gynecol",
  ["quarterly review of the rural economy"] = "Quart. Rev. Rural Economy",
  ["quarterly review. british red cross society"] = "Q Rev Br Red Cross Soc",
  ["quarterly review. united states. veterans administration. coordination service. research division"] = "Q Rev U S Veterans Adm",
  ["quarterly reviews of biology"] = "Q. Rev. Biol.",
  ["quarterly reviews of biophysics"] = "Q Rev Biophys",
  ["quarterly reviews on drug metabolism and drug interactions"] = "Q Rev Drug Metab Drug Interact",
  ["quarterly reviews, chemical society"] = "Q. Rev. Chem. Soc.",
  ["quarterly. phi beta pi medical fraternity"] = "Quarterly Phi Beta Pi Med Fratern",
  ["quarternary international"] = "Quart. Int.",
  ["quasigroups and related systems"] = "Quasigroups Related Systems",
  ["quaternaria. storia naturale e culturale del quaternario"] = "Quaternaria",
  ["quaternary geochronology"] = "Quat. Geochronol.",
  ["quaternary international"] = "Quat. Int.",
  ["quaternary international : the journal of the international union for quaternary research"] = "Quat Int",
  ["quaternary research"] = "Quat. Res.",
  ["quaternary science advances"] = "Quat. Sci. Adv.",
  ["quaternary science reviews"] = "Quat. Sci. Rev.",
  ["que sais-je?"] = "Que Sais-Je?",
  ["queen's gazette"] = "Queens Gazette",
  ["queen's law journal"] = "Queens Law J",
  ["queen's medical magazine"] = "Queens Med Mag (1972)",
  ["queen's nursing journal"] = "Queens Nurs J",
  ["queen's quarterly"] = "Queen's Q",
  ["queens nursing journal"] = "Queens Nurs. J.",
  ["queensland nurse"] = "Qld. Nurse",
  ["queensland nurses journal"] = "Qld Nurses J",
  ["queen’s gazette"] = "Queens Gazette",
  ["queen’s law journal"] = "Queens Law J",
  ["queen’s nursing journal"] = "Queens Nurs J",
  ["queen’s papers in pure and applied mathematics"] = "Queen’s Papers in Pure and Appl. Math.",
  ["queen’s quarterly"] = "Queen’s Q",
  ["quellen und beitrage zur geschichte der universitat jena"] = "Quellen Beitr. Gesch. Univ. Jena",
  ["quellen und forschungen aus italienischen archiven und bibliotheken"] = "Quellen Forsch Ital Arch Bibl",
  ["quellen und studien zur geschichte der pharmazie"] = "Quellen Stud Gesch Pharm",
  ["quellen zum leben und wirken adam ries’ und seiner sohne"] = "Quellen Leben Wirk. Adam Ries Sohne",
  ["quercy recherche"] = "Quercy Rech",
  ["quest (grand rapids, mich.)"] = "Quest",
  ["questions [d'internat]"] = "Quest D Int",
  ["questions and answers in general topology"] = "Questions Answers Gen. Topology",
  ["queueing systems"] = "Queueing Syst",
  ["queueing systems. theory and applications"] = "Queueing Syst.",
  ["qufu shifan daxue xuebao"] = "Qufu Shifan Daxue Xuebao Ziran Kexue Ban",
  ["quimica analitica"] = "Quim. Anal. (Barcelona)",
  ["quimica no brasil"] = "Quim. Bras.",
  ["quimica nova"] = "Quim. Nova",
  ["quinnipiac health law journal"] = "Quinnipiac Health Law J",
  ["quintessence international"] = "Quintessence Int (Berl)",
  ["quintessence international (berlin, germany : 1985)"] = "Quintessence Int",
  ["quintessence international, dental digest"] = "Quintessence Int Dent Dig",
  ["quintessence international. edition francaise"] = "Quintessence Int Fr",
  ["quintessence of dental technology"] = "Quintessence Dent Technol",
  ["quintessencia de protese de laboratorio"] = "Quintessencia Protese Lab.",
  ["quintessenz der zahntechnik"] = "Quintessenz Zahntech.",
  ["quintessenz journal"] = "Quintessenz J",
  ["quintessenz-impulse : magazin fur kommunikation & praxismanagement"] = "Quintessenz Impulse",
  ["quintessenz-impulse : magazin für kommunikation & praxismanagement"] = "Quintessenz Impulse",
  ["quintessência de prótese de laboratório"] = "Quintessencia Protese Lab",
  ["quinto centenario"] = "Quinto Cent",
  ["química nova"] = "Quim Nova",
  ["r & s report. hawaii. department of health. research and statistics office"] = "R S Rep",
  ["r & s report. hawaii. dept. of health. research and statistics office"] = "R S Rep",
  ["r journal"] = "R J.",
  ["r. borger, handbuch der keilschriftliteratur (berlin 1967–1975)"] = "HKL",
  ["r.i. medical journal"] = "R I Med J",
  ["rabies bulletin europe"] = "Rabies Bull Eur",
  ["race & class"] = "Race Cl",
  ["race & society : official journal of the association of black sociologists"] = "Race Soc",
  ["race and justice"] = "Race Justice",
  ["race and social problems"] = "Race Soc Probl",
  ["race, ethnicity and education"] = "Race Ethn Educ",
  ["race, gender & class (towson, md.)"] = "Race Gend Cl",
  ["race, gender, and science"] = "Race Gend. Sci.",
  ["rad conference proceedings"] = "RAD Conf. Proc.",
  ["rad hrvatske akademije znanosti i umjetnosti"] = "Rad Hrvatske Akad. Znan. Umjet.",
  ["rad hrvatske akademije znanosti i umjetnosti. matematicke znanosti"] = "Rad Hrvat. Akad. Znan. Umjet. Mat. Znan.",
  ["rad jugoslavenske akademije znanosti i umjetnosti. razred za medicinske znanosti"] = "Rad Jugosl Akad Znan Umjet Razred Med Znan",
  ["radc-tdr-. rome air development center"] = "RADC TDR",
  ["radiation and environmental biophysics"] = "Radiat Environ Biophys",
  ["radiation biology"] = "Radiat. Biol.",
  ["radiation data and reports"] = "Radiat Data Rep",
  ["radiation detection technology and methods"] = "Radiat. Detect. Technol. Methods",
  ["radiation effects"] = "Radiat Eff",
  ["radiation effects and defects in solids"] = "Radiat. Eff. Defects Solids",
  ["radiation measurements"] = "Radiat. Meas.",
  ["radiation medicine"] = "Radiat Med",
  ["radiation medicine and protection"] = "Radiat Med Prot",
  ["radiation oncology"] = "Radiat. Oncol.",
  ["radiation oncology (london, england)"] = "Radiat Oncol",
  ["radiation oncology investigations"] = "Radiat Oncol Investig",
  ["radiation oncology journal"] = "Radiat Oncol J",
  ["radiation physics and chemistry"] = "Radiat. Phys. Chem.",
  ["radiation physics and chemistry (oxford, england : 1993)"] = "Radiat Phys Chem Oxf Engl 1993",
  ["radiation protection dosimetry"] = "Radiat Prot Dosimetry",
  ["radiation research"] = "Radiat Res",
  ["radiation research. supplement"] = "Radiat Res Suppl",
  ["radiatsionnaia biologiia, radioecologiia"] = "Radiats Biol Radioecol",
  ["radiatsionnaia biologiia, radioecologiia / rossiiskaia akademiia nauk"] = "Radiats Biol Radioecol",
  ["radiatsionnaya bezopasnost i zashchita aes"] = "Radiats. Bezop. Zashch. AES",
  ["radical america"] = "Radic Am",
  ["radical history review"] = "Radic Hist Rev",
  ["radio and electronic engineer"] = "Radio Electron. Eng.",
  ["radio engineering and electronic physics (ussr) [translation of radiotekhnika i elektronika]"] = "Radio Eng. Electron. Phys. (USSR)",
  ["radio engineering and electronics (ussr) [translation of radiotekhnika i elektronika]"] = "Radio Eng. Electron. (USSR)",
  ["radio physics and radio astronomy"] = "Radio Phys. Radio Astron.",
  ["radio science"] = "Radio Sci.",
  ["radioactive and natural strontium in human bone; u. k. results"] = "Radioactiv Nat Strontium Hum Bone",
  ["radiobiologia, radioterapia e fisica medica"] = "Radiobiol. Radioter. Fis. Med.",
  ["radiobiologia, radioterapia, e fisica medica"] = "Radiobiol Radioter Fis Med",
  ["radiobiologia, radiotherapia"] = "Radiobiol Radiother (Berl)",
  ["radiocarbon. an international journal of cosmogenic isotope research"] = "Radiocarbon",
  ["radiochemistry (moscow, russian federation)"] = "Radiochemistry (Moscow, Russ. Fed.)",
  ["radiochemistry (moscow, russian federation)(translation of radiokhimiya)"] = "Radiochemistry (Moscow, Russ. Fed.)",
  ["radiochemistry (new york, ny, united states)"] = "Radiochemistry (N. Y., NY, U. S.)",
  ["radiochimica acta"] = "Radiochim. Acta",
  ["radioelectronics and communications systems"] = "Radioelectron. Commun. Syst.",
  ["radiographics : a review publication of the radiological society of north america, inc"] = "Radiographics",
  ["radiography (london, england : 1995)"] = "Radiography (Lond)",
  ["radiography and clinical photography"] = "Radiogr Clin Photogr",
  ["radiography today"] = "Radiogr Today",
  ["radiologia austriaca"] = "Radiol Austriaca",
  ["radiologia brasileira"] = "Radiol Bras",
  ["radiologia clinica"] = "Radiol Clin",
  ["radiologia clinica et biologica"] = "Radiol Clin Biol",
  ["radiologia diagnostica"] = "Radiol Diagn (Berl)",
  ["radiologia interamericana"] = "Radiol Interam",
  ["radiologia medica"] = "Radiol. Med. (Torino)",
  ["radiologia pratica"] = "Radiol Prat",
  ["radiologia sperimentale"] = "Radiol Sper",
  ["radiologia; rassegna internazionale trimestrale di radiobiologia, radioterapia, radiodiagnostica, terapia fisica e fisica applicata alla medicina"] = "Radiologia (Roma)",
  ["radiologic clinics of north america"] = "Radiol Clin North Am",
  ["radiologic technology"] = "Radiol Technol",
  ["radiological health data and reports"] = "Radiol Health Data Rep",
  ["radiological physics and technology"] = "Radiol Phys Technol",
  ["radiología interamericana"] = "Radiol Interam",
  ["radiology - open journal"] = "Radiol Open J",
  ["radiology and oncology"] = "Radiol Oncol",
  ["radiology case reports"] = "Radiol Case Rep",
  ["radiology management"] = "Radiol Manage",
  ["radiology of infectious diseases"] = "Radiol. Infect. Dis.",
  ["radiology of infectious diseases (beijing, china)"] = "Radiol Infect Dis",
  ["radiology research and practice"] = "Radiol Res Pract",
  ["radiology. artificial intelligence"] = "Radiol Artif Intell",
  ["radiology. cardiothoracic imaging"] = "Radiol Cardiothorac Imaging",
  ["radiology. imaging cancer"] = "Radiol Imaging Cancer",
  ["radiology: artifical intelligence"] = "Radiol. Artif. Intell.",
  ["radiology: cardiothoracic imaging"] = "Radiol. Cardiothorac. Imaging",
  ["radiology: imaging cancer"] = "Radiol. Imaging Cancer",
  ["radiophysics and quantum electronics"] = "Radiophys. Quantum Electron.",
  ["radiotekhnika i elektronika (moscow)"] = "Radiotekh. Elektron. (Moscow)",
  ["radiotekhnika i elektronika [radio engineering and electronic physics (ussr)]"] = "Radiotekh. Elektron. [Radio Eng. Electron. Phys. (USSR)]",
  ["radiotekhnika i elektronika [radio engineering and electronics (ussr)]"] = "Radiotekh. Elektron. [Radio Eng. Electron. (USSR)]",
  ["radioterapia, radiobiologia e fisica medica"] = "Radioter Radiobiol Fis Medica",
  ["radiotherapy and oncology"] = "Radiother. Oncol.",
  ["radiotherapy and oncology : journal of the european society for therapeutic radiology and oncology"] = "Radiother Oncol",
  ["radon series on computational and applied mathematics"] = "Radon Ser. Comput. Appl. Math.",
  ["radovi matematički"] = "Rad. Mat.",
  ["radovi medicinskog faculteta u zagrebu"] = "Rad Med Fak Zagrebu",
  ["radovi medicinskog fakulteta u zagrebu"] = "Rad. Med. Fak. Zagrebu",
  ["radovi zavoda za povijesne znanosti, hrvatska akademija znanosti i umjetnosti u zadru"] = "RadAkZadar",
  ["radovi. razdio povijesnih znanosti"] = "RadSplit",
  ["raffles bulletin of zoology"] = "Raffles Bull. Zool.",
  ["raggi. zeitschrift für kunstgeschichte und archäologie"] = "Raggi",
  ["railway engineering science"] = "Railway Eng. Sci.",
  ["railway sciences"] = "Railway Sci.",
  ["rairo - operations research"] = "RAIRO Oper. Res.",
  ["rairo - theoretical informatics and applications"] = "RAIRO Theor. Inf. Appl.",
  ["rairo informatique théorique et applications"] = "RAIRO Inform. Théor. Appl.",
  ["rairo modélisation mathématique et analyse numérique"] = "RAIRO Modél. Math. Anal. Numér.",
  ["rairo operations research"] = "RAIRO Oper. Res.",
  ["rairo recherche opérationnelle"] = "RAIRO Rech. Opér.",
  ["rairo theoretical informatics and applications (rairo: ita)"] = "RAIRO Theor. Inform. Appl. (RAIRO:ITA)",
  ["rairo theoretical informatics and applications. informatique theorique et applications"] = "RAIRO Theor. Inform. Appl.",
  ["rajasthan medical journal"] = "Rajasthan Med J",
  ["ramanujan journal. an international journal devoted to the areas of mathematics influenced by ramanujan"] = "Ramanujan J.",
  ["ramanujan mathematical society collected works series"] = "Ramanujan Math. Soc. Collect. Works Ser.",
  ["ramanujan mathematical society lecture notes series"] = "Ramanujan Math. Soc. Lect. Notes Ser.",
  ["rambam maimonides medical journal"] = "Rambam Maimonides Med J",
  ["rand health quarterly"] = "Rand Health Q",
  ["rand journal of economics"] = "RAND J. Econ.",
  ["random & computational dynamics"] = "Random Comput. Dynam.",
  ["random matrices. theory and applications"] = "Random Matrices Theory Appl.",
  ["random matrices: theory and application"] = "Random Matrices: Theory Appl.",
  ["random operators and stochastic equations"] = "Random Oper. Stoch. Equ.",
  ["random structures & algorithms"] = "Random Structures Algorithms",
  ["random structures and algorithms"] = "Random Struct. Algorithms",
  ["ranf review"] = "RANF Rev",
  ["range management and agroforestry"] = "Range Manage. Agrofor.",
  ["rangeland ecology & management"] = "Rangel Ecol Manag",
  ["rangeland journal"] = "Rangeland J.",
  ["ranliao huaxue xuebao"] = "Ranliao Huaxue Xuebao",
  ["rap bulletin of biological assessment"] = "RAP Bull. Biol. Assess.",
  ["rapid communications in mass spectrometry"] = "Rapid Commun. Mass Spectrom.",
  ["rapid communications in mass spectrometry : rcm"] = "Rapid Commun Mass Spectrom",
  ["rapid prototyping journal"] = "Rapid Prototyp J",
  ["raporty wykopaliskowe"] = "RapWyk",
  ["rapport sur le fonctionnement technique. cayenne, french guiana. institut pasteur de la guyane française et de l'inini"] = "Rapp Fonct Tec Inst Pasteur Guyane Fr",
  ["rapporter fra statistisk sentralbyra"] = "Rapp Stat Sentralbyra",
  ["rapports. congrès national de la tuberculose"] = "Rapp Congr Natl Tuberc",
  ["rapra review reports"] = "Rapra Rev. Rep.",
  ["rare cancers and therapy"] = "Rare Cancers Ther",
  ["rare diseases (austin, tex.)"] = "Rare Dis",
  ["rare metals"] = "Rare Met.",
  ["rare metals (beijing, china)"] = "Rare Met. (Beijing, China)",
  ["rare metals materials and engineering"] = "Rare Met. Mater. Eng.",
  ["rare tumors"] = "Rare Tumors",
  ["ras techniques and instruments"] = "RAS Tech. Instrum.",
  ["rasayan journal of chemistry"] = "Rasayan J. Chem.",
  ["rasprave i grada za povijest znanosti / hrvatska akademija znanosti i umjetnosti"] = "Raspr Grada Povij Znan",
  ["rasprave i građa za povijest znanosti"] = "Raspr Grada Povij Znan",
  ["rassegna clinico-scientifica"] = "Rass Clin Sci",
  ["rassegna degli archivi di stato"] = "Rass Arch Stato",
  ["rassegna del lazio"] = "RassLazio",
  ["rassegna di archeologia. associazione archeologica piombinese"] = "RassAPiomb",
  ["rassegna di clinica terapia e scienze affini"] = "Rass. Clin. Ter.",
  ["rassegna di dermatologia e di sifilografia"] = "Rass Dermatol Sifilogr",
  ["rassegna di dermatologia e sifilografia"] = "Rass. Dermatol. Sifilogr.",
  ["rassegna di fisiopatologia clinica e terapeutica"] = "Rass Fisiopatol Clin Ter",
  ["rassegna di medicina industriale e di igiene del lavoro"] = "Rass Med Ind Ig Lav",
  ["rassegna di medicina sperimentale"] = "Rass Med Sper",
  ["rassegna di neurologia vegetativa"] = "Rass Neurol Veg",
  ["rassegna di neuropsichiatria e scienze affini"] = "Rass Neuropsichiatr",
  ["rassegna di studi psichiatrici"] = "Rass Studi Psichiatr",
  ["rassegna economica"] = "Rassegna Econ.",
  ["rassegna giuliana di medicina"] = "Rass Giuliana Med",
  ["rassegna internazionale di clinica e terapia"] = "Rass Int Clin Ter",
  ["rassegna internazionale di stomatologia pratica"] = "Rass Int Stomatol Prat",
  ["rassegna italiana d'ottalmologia"] = "Rass Ital Ottalmol",
  ["rassegna italiana di chirurgia e medicina"] = "Rass Ital Chir Med",
  ["rassegna italiana di gastro-enterologia"] = "Rass Ital Gastroenterol",
  ["rassegna italiana di sociologia"] = "Rass Ital Sociol",
  ["rassegna medica"] = "Rass Medica",
  ["rassegna medica sarda"] = "Rass Med Sarda",
  ["rassegna odontotecnica"] = "Rass Odontotec",
  ["rassegna penitenziaria e criminologica"] = "Rass Penititenziaria Crim",
  ["rassegna storica del risorgimento : organo della societa nazionale per la storia del risorgimento italiano"] = "Rass Stor Risorgim",
  ["rassegna storica del risorgimento : organo della società nazionale per la storia del risorgimento italiano"] = "Rass Stor Risorgim",
  ["rassegna storica salernitana"] = "RSS",
  ["rassegna trimestrale di odontoiatria"] = "Rass Trimest Odontoiatr",
  ["rassegna volterrana"] = "RassVolt",
  ["rasy i narody"] = "Rasy Nar Sovrem Etn Ras Probl",
  ["rasy i narody / akademiia nauk sssr, institut etnografii im. n.n. miklukho-maklaia"] = "Rasy Nar Sovrem Etn Ras Probl",
  ["ratiariensia. studi e materiali mesici e danubiani"] = "Ratiariensia",
  ["ratio (new series)"] = "Ratio (N.S.)",
  ["ratio juris"] = "Ratio Juris",
  ["rational drug therapy"] = "Ration Drug Ther",
  ["rationality and society"] = "Ration. Soc.",
  ["raumforschung und raumordnung"] = "Raumforsch Raumordn",
  ["raumplanung, informationshefte"] = "Raumplan., Inf.h.",
  ["ravenna studi e ricerche"] = "RavStRic",
  ["ravintsara : collection medecine, sante publique"] = "Ravintsara",
  ["ravintsara : collection médecine, santé publique"] = "Ravintsara",
  ["raydan. journal of the ancient yemeni antiquities and epigraphy"] = "Raydan",
  ["rber, review of business and economic research"] = "RBER Rev Bus Econ Res",
  ["rbm : a journal of rare books, manuscripts, and cultural heritage"] = "RBM",
  ["rca review"] = "RCA Rev.",
  ["rcm midwives : the official journal of the royal college of midwives"] = "RCM Midwives",
  ["rcm midwives journal : official journal of the royal college of midwives"] = "RCM Midwives J",
  ["rcn nursing standard"] = "RCN Nurs Stand",
  ["re dai hai yang xue bao = journal of tropical oceanography"] = "Re Dai Hai Yang Xue Bao",
  ["re dai nong ye ke xue"] = "Re Dai Nong Ye Ke Xue",
  ["re dai ya re dai zhi wu xue bao = journal of tropical and subtropical botany"] = "J Trop Subtrop Botany",
  ["re dai yi xue za zhi"] = "Journal of tropical medicine|Re Dai Yi Xue Za Zhi",
  ["re dai yi xue za zhi = journal of tropical medicine"] = "Re Dai Yi Xue Za Zhi",
  ["re dai zuo wu xue bao = chinese journal of tropical crops"] = "Re Dai Zuo Wu Xue Bao",
  ["re/views in environmental science and bio/technology"] = "Rev Environ Sci Biotechnol",
  ["re:gen open"] = "Re:GEN Open",
  ["reach - reviews in human space exploration"] = "REACH Rev. Hum. Space Explor.",
  ["reaching out"] = "Reach Out",
  ["reaction chemistry & engineering"] = "React. Chem. Eng.",
  ["reaction kinetics and catalysis letters"] = "React. Kinet. Catal. Lett.",
  ["reaction kinetics mechanisms and catalysis"] = "React. Kinet. Mech. Catal.",
  ["reaction kinetics, mechanisms and catalysis"] = "React. Kinet., Mech. Catal.",
  ["reactions weekly"] = "React. Wkly",
  ["reactive & functional polymers"] = "React Funct Polym",
  ["reactive and functional polymers"] = "React. Funct. Polym.",
  ["reactive oxygen species (apex, n.c.)"] = "React Oxyg Species (Apex)",
  ["reactive polymers"] = "React. Polym.",
  ["reactivity of solids"] = "React. Solids",
  ["readers digest"] = "Read. Dig.",
  ["reading & writing quarterly : overcoming learning difficulties"] = "Read Writ Q",
  ["reading and writing"] = "Read Writ",
  ["reading psychology"] = "Read Psychol",
  ["reading research and instruction : the journal of the college reading association"] = "Read Res Instr",
  ["reading research quarterly"] = "Read Res Q",
  ["readings in mathematics"] = "Read. Math.",
  ["readings in philosophy"] = "Read. Philos.",
  ["real academia de ciencias exactas, físicas y naturales de madrid"] = "Rev. Real Acad. Cienc. Exact. Fís. Natur. Madrid",
  ["real academia de farmacia de barcelona. sesión inaugural del curso"] = "Real Acad Farm Barcelona Sesion Inaug Curso",
  ["real analysis exchange"] = "Real Anal. Exchange",
  ["real estate economics"] = "Real Estate Econ.",
  ["real property, probate, and trust journal"] = "Real Property Probate Trust J",
  ["real-time systems"] = "Real-Time Syst.",
  ["reales sitios"] = "R Sitios",
  ["realidad, datos y espacio : revista internacional de estadística y geografía"] = "Real Datos Espacio",
  ["realites cliniques"] = "Real. Clin.",
  ["realites cliniques : revue europeenne d’odontologie"] = "Real Clin",
  ["reallexikon der assyriologie und vorderasiatischen archäologie"] = "RlA",
  ["reallexikon für antike und christentum"] = "RLAC",
  ["reallexikon zur byzantinischen kunst"] = "LRBK",
  ["reanimation et organes artificiels"] = "Reanim. Organes Artif.",
  ["reanimation et organes artificiels. wiederbelebung und kunstliche organe. reanimation and artificial organs"] = "Reanim Organes Artif",
  ["reason papers"] = "Reason Pap",
  ["recenseamento geral da populaçao. instituto nacional de estatística (portugal)"] = "Recens Geral Popul",
  ["recent advances & research updates"] = "Recent Adv Res Updat",
  ["recent advances in biological psychiatry"] = "Recent Adv Biol Psychiatry",
  ["recent advances in cardiovascular drug discovery"] = "Recent Adv Cardiovasc Drug Discov",
  ["recent advances in clinical nuclear medicine"] = "Recent Adv Clin Nucl Med",
  ["recent advances in computational chemistry"] = "Recent Adv. Comput. Chem.",
  ["recent advances in dna & gene sequences"] = "Recent Adv. DNA Gene Sequences",
  ["recent advances in drug delivery and formulation"] = "Recent Adv Drug Deliv Formul",
  ["recent advances in electrical and electronic engineering"] = "Recent Adv. Electr. Electron. Eng.",
  ["recent advances in nursing"] = "Recent Adv Nurs",
  ["recent advances in phytochemistry"] = "Recent Adv Phytochem",
  ["recent advances in studies on cardiac structure and metabolism"] = "Recent Adv Stud Cardiac Struct Metab",
  ["recent developments in alcoholism"] = "Recent Dev. Alcohol.",
  ["recent developments in alcoholism : an official publication of the american medical society on alcoholism, the research society on alcoholism, and the national council on alcoholism"] = "Recent Dev Alcohol",
  ["recent innovations in chemical engineering"] = "Recent Innovations Chem. Eng.",
  ["recent pat. biotechnol."] = "Recent Pat. Biotechnol.",
  ["recent patents on anti-cancer drug discovery"] = "Recent Pat Anticancer Drug Discov",
  ["recent patents on anti-infective drug discovery"] = "Recent Pat Antiinfect Drug Discov",
  ["recent patents on biomarkers"] = "Recent Pat Biomark",
  ["recent patents on biomedical engineering"] = "Recent Pat Biomed Eng",
  ["recent patents on biotechnology"] = "Recent Pat Biotechnol",
  ["recent patents on cardiovascular drug discovery"] = "Recent Pat Cardiovasc Drug Discov",
  ["recent patents on cns drug discovery"] = "Recent Pat. CNS Drug Discovery",
  ["recent patents on dna & gene sequences"] = "Recent Pat DNA Gene Seq",
  ["recent patents on drug delivery & formulation"] = "Recent Pat Drug Deliv Formul",
  ["recent patents on endocrine, metabolic & immune drug discovery"] = "Recent Pat Endocr Metab Immune Drug Discov",
  ["recent patents on engineering"] = "Recent Pat. Eng.",
  ["recent patents on food, nutrition & agriculture"] = "Recent Pat Food Nutr Agric",
  ["recent patents on inflammation & allergy drug discovery"] = "Recent Pat Inflamm Allergy Drug Discov",
  ["recent patents on mechanical engineering"] = "Recent Pat. Mech. Eng.",
  ["recent patents on medical imaging"] = "Recent Pat Med Imaging",
  ["recent patents on nanomedicine"] = "Recent Pat Nanomed",
  ["recent patents on nanotechnology"] = "Recent Pat Nanotechnol",
  ["recent patents on regenerative medicine"] = "Recent Pat Regen Med",
  ["recent patents on space technology"] = "Recent Pat Space Technol",
  ["recent progress in hormone research"] = "Recent Prog Horm Res",
  ["recent progress in materials"] = "Recent Prog Mater",
  ["recent progress in psychiatry"] = "Recent Progr Psychiatr",
  ["recent research developments in immunology"] = "Rec Res Dev Immunol",
  ["recent research developments in neuroscience"] = "Recent Res Dev Neurosci",
  ["recent research developments in organic chemistry"] = "Recent Res Dev Org Chem",
  ["recent research developments in virology"] = "Recent Res Dev Virol",
  ["recent results in cancer research"] = "Recent Results Cancer Res.",
  ["recent results in cancer research. fortschritte der krebsforschung. progres dans les recherches sur le cancer"] = "Recent Results Cancer Res",
  ["recent results in cancer research. fortschritte der krebsforschung. progrès dans les recherches sur le cancer"] = "Recent Results Cancer Res",
  ["recenti progressi in medicina"] = "Recenti Prog Med",
  ["recently published articles"] = "Recent Publ Articles",
  ["receptors & channels"] = "Recept Channels",
  ["receptors & clinical investigation"] = "Receptors Clin Investig",
  ["receptors & signal transduction"] = "Recept Signal Transduct",
  ["receptors and channels"] = "Receptors Channels",
  ["receptors and signal transduction"] = "Recept. Signal Transduct.",
  ["recerques del museu d’alcoi"] = "RecMusAlcoi",
  ["recherche en soins infirmiers"] = "Rech Soins Infirm",
  ["recherche sociale"] = "Rech Soc",
  ["recherche, la"] = "Recherche",
  ["recherches & éducations"] = "Rech Educ",
  ["recherches albanologiques"] = "RechAlb",
  ["recherches amerindiennes au quebec"] = "Rech Amerindien Que",
  ["recherches amérindiennes au québec"] = "Rech Amerindien Que",
  ["recherches archéologiques. institut d’archéologie de l’université de cracovie"] = "RechACrac",
  ["recherches augustiniennes (supplément à reaug)"] = "RecAug",
  ["recherches de science religieuse"] = "RecSR",
  ["recherches de théologie ancienne et médiévale"] = "RecTh",
  ["recherches economiques de louvain"] = "Rech. Econ. Louvain",
  ["recherches en mathematiques appliquees"] = "Rech. Math. Appl.",
  ["recherches en mathématiques appliquées"] = "Rech. Math. Appl.",
  ["recherches économiques de louvain"] = "Rech Econ Louvain",
  ["recherches feministes"] = "Rech fem",
  ["recherches féministes"] = "Rech fem",
  ["recherches geographiques a strasbourg"] = "Rech Geogr Strasbg",
  ["recherches géographiques à strasbourg"] = "Rech Geogr Strasbg",
  ["recherches sociographiques"] = "Rech Sociogr",
  ["recherches sociologiques et anthropologiques : rs & a"] = "Rech Sociol Anthropol",
  ["recherches sur la philosophie et le langage"] = "Rech. Philos. Lang.",
  ["rechtshistoriches journal"] = "RJ",
  ["rechtsmedizin (berlin, germany)"] = "Rechtsmedizin (Berl)",
  ["recombinant dna technical bulletin"] = "Recomb DNA Tech Bull",
  ["reconstruction surgery and traumatology"] = "Reconstr Surg Traumatol",
  ["record of chemical progress"] = "Rec Chem Prog",
  ["records of buckinghamshire, or, papers and notes on the history, antiquities, and architecture of the county, together with the proceedings of the architectural and archaeological society for the county of buckingham"] = "Rec Bucks",
  ["records of natural products"] = "Rec. Nat. Prod.",
  ["records of the australian museum"] = "Rec Aust Mus",
  ["records of the canterbury museum"] = "Rec Canterb Mus",
  ["records of the columbia historical society, washington, d.c"] = "Rec Columb Hist Soc",
  ["records of the western australian museum. western australian museum"] = "Rec West Aust Mus",
  ["records of the zoological survey of india"] = "Rec Zool Surv India",
  ["recovery journal"] = "Recovery J",
  ["recreation, parks, and tourism in public health"] = "Recreat Park Tour Public Health",
  ["recreational mathematics magazine"] = "Recreat. Math. Mag.",
  ["recruit & retain : a monthly publication for professionals concerned with nurse recruitment and retention"] = "Recruit Retain",
  ["recruit and retain"] = "Recruit Retain",
  ["recruitment & retention report"] = "Recruit Retent Rep",
  ["recruitment and retention report"] = "Recruit. Retent. Rep.",
  ["recruitment, retention & restructuring report"] = "Recruit Retent Restruct Rep",
  ["recruitment, retention and restructuring report"] = "Recruit. Retent. Restruct. Rep.",
  ["recueil de medecine veterinaire"] = "Rec Med Vet Ec Alfort",
  ["recueil de médecine vétérinaire"] = "Rec Med Vet Ec Alfort",
  ["recueil de travaux relatifs à philologie et l’archéologie égyptiennes et assyriennes"] = "RecTrav",
  ["recueil des notices et memoires de la société archéologique du département de constantine"] = "RecConstantine",
  ["recueil des travaux chimiques des pays-bas"] = "Recl. Trav. Chim. Pays-Bas",
  ["recueil des travaux chimiques des pays-bas (leiden, netherlands : 1920)"] = "Recl Trav Chim Pays Bas",
  ["recueil des travaux chimiques des pays-bas journal of the royal netherlands chemical society"] = "Recl. Trav. Chim. Pays-Bas",
  ["recusant history"] = "Recusant Hist",
  ["red river valley historical review"] = "Red River Val Hist Rev",
  ["redbook : the magazine for young adults"] = "Redbook",
  ["redia (firenze)"] = "Redia (Firenze)",
  ["redox biology"] = "Redox Biol",
  ["redox experimental biology"] = "Redox Exp. Biol.",
  ["redox report"] = "Redox Rep.",
  ["redox report : communications in free radical research"] = "Redox Rep",
  ["refereed e-journal"] = "Refereed ejournal",
  ["reference guide for the health care technology industry"] = "Ref Guide Health Care Technol Ind",
  ["reference sources for the social sciences and humanities"] = "Ref. Sources Social Sci. Humanities",
  ["reference sources in science and technology"] = "Ref. Source Sci. Technol.",
  ["references en gynecologie obstetrique"] = "Ref Gynecol Obstet",
  ["references sciences"] = "Ref. Sci.",
  ["reflections (long beach, calif.)"] = "Reflections (Long Beach)",
  ["reflections / sigma theta tau"] = "Reflections",
  ["reflections on nursing leadership"] = "Reflect Nurs Leadersh",
  ["reflections on nursing leadership / sigma theta tau international, honor society of nursing"] = "Reflect Nurs Leadersh",
  ["reflets de la physique"] = "Reflets Phys.",
  ["refractive & corneal surgery"] = "Refract Corneal Surg",
  ["refractive and corneal surgery"] = "Refract. Corneal Surg.",
  ["refractories and industrial ceramics"] = "Refract. Ind. Ceram",
  ["refresher courses in anesthesiology"] = "Refresh Courses Anesthesiol",
  ["refrigeration science and technology"] = "Refrig. Sci. Technol.",
  ["refuat ha-shinayim"] = "Refu. Hashinayim",
  ["refugee reports"] = "Refug Rep",
  ["refugees (office of the united nations high commissioner for refugees : 1984)"] = "Refugees",
  ["refuʼah ṿeṭerinarit : rivʻon histadrut ha-rofʼim ha-ṿeṭerinariyim be-erets-yiśraʼel"] = "Refu Vet",
  ["refuʼat ha-peh ṿeha-shinayim (1993)"] = "Refuat Hapeh Vehashinayim (1993)",
  ["refuʼat ha-peh ṿeha-shinayim (tel aviv, israel : 1969)"] = "Refuat Hapeh Vehashinayim",
  ["refuʼat ha-shinayim (tel aviv, israel : 1944)"] = "Refuat Hashinayim",
  ["refuʼat ha-shinayim (tel aviv, israel : 1983)"] = "Refuat Hashinayim",
  ["refu’ah veterinarit : riv’on histadrut ha-rof’im ha-veterinariyim be-erets-yisra’el"] = "Refu Vet",
  ["refu’at ha-peh veha-shinayim (1993)"] = "Refuat Hapeh Vehashinayim",
  ["refu’at ha-peh veha-shinayim (tel aviv, israel : 1969)"] = "Refuat Hapeh Vehashinayim",
  ["refu’at ha-shinayim (tel aviv, israel : 1944)"] = "Refuat Hashinayim",
  ["refu’at ha-shinayim (tel aviv, israel : 1983)"] = "Refuat Hashinayim",
  ["regan report on nursing law"] = "Regan Rep. Nurs. Law",
  ["regards sur l'actualité"] = "Regards Actual",
  ["regards sur l’actualite"] = "Regards Actual",
  ["regeneration (oxford, england)"] = "Regeneration (Oxf)",
  ["regenerative biomaterials"] = "Regen Biomater",
  ["regenerative engineering and translational medicine"] = "Regener. Eng. Transl. Med.",
  ["regenerative medicine"] = "Regen Med",
  ["regenerative medicine frontiers"] = "Regen Med Front",
  ["regenerative medicine research"] = "Regen Med Res",
  ["regenerative therapy"] = "Regen Ther",
  ["regensburger jahrbuch fur arztliche fortbildung"] = "Regensb Jahrb Arztl Fortbild",
  ["regensburger jahrbuch für ärztliche fortbildung"] = "Regensb Jahrb Arztl Fortbild",
  ["regensburger mathematische schriften"] = "Regensburger Math. Schriften",
  ["regio basiliensis"] = "Reg. Basil.",
  ["regional anaesthesie"] = "Reg. Anaesth.",
  ["regional anesthesia"] = "Reg Anesth",
  ["regional anesthesia and pain medicine"] = "Reg Anesth Pain Med",
  ["regional development dialogue"] = "Reg Dev Dialogue",
  ["regional environmental change"] = "Reg Environ Change",
  ["regional immunology"] = "Reg Immunol",
  ["regional science and urban economics"] = "Reg Sci Urban Econ",
  ["regional science perspectives"] = "Reg Sci Perspect",
  ["regional studies"] = "Reg Stud",
  ["regional studies in marine science"] = "Reg Stud Mar Sci",
  ["regional studies, regional science"] = "Reg Stud Reg Sci",
  ["regional sustainability"] = "Reg. Sustainability",
  ["regions & cohesion : the journal of the consortium for comparative research on regional integration and social cohesion"] = "Reg Cohes",
  ["registered nurse"] = "Regist. Nurse",
  ["registered nurse (toronto, ont.)"] = "Regist Nurse",
  ["regnum vegetabile"] = "Regnum Veg",
  ["regulae benedicti studia: annuarium internationale"] = "RBS",
  ["regular and chaotic dynamics"] = "Regul. Chaotic Dyn.",
  ["regular and chaotic dynamics. international scientific journal"] = "Regul. Chaotic Dyn.",
  ["regulated rivers: research & management"] = "Regul. Rivers: Res. Manage.",
  ["regulation & governance"] = "Regul Gov",
  ["regulatory analyst. medical waste"] = "Regul Anal Med Waste",
  ["regulatory mechanisms in biosystems"] = "Regul Mech Biosyst",
  ["regulatory peptides"] = "Regul. Pept.",
  ["regulatory peptides. supplement"] = "Regul Pept Suppl",
  ["regulatory toxicology and pharmacology"] = "Regul. Toxicol. Pharmacol.",
  ["regulatory toxicology and pharmacology : rtp"] = "Regul Toxicol Pharmacol",
  ["regulyarnaya & khaoticheskaya dinamika"] = "Regul. Khaoticheskaya Din.",
  ["rehab management"] = "Rehab Manag",
  ["rehabilitace a fyzikalni lekarstvi"] = "Rehabil Fyz Lek",
  ["rehabilitace a fyzikální́́ lékařství"] = "Rehabil Fyz Lek",
  ["rehabilitation counseling bulletin"] = "Rehabil Couns Bull",
  ["rehabilitation counselors and educators journal"] = "Rehabil Couns Educ J",
  ["rehabilitation literature"] = "Rehabil Lit",
  ["rehabilitation nursing"] = "Rehabil. Nurs.",
  ["rehabilitation nursing : the official journal of the association of rehabilitation nurses"] = "Rehabil Nurs",
  ["rehabilitation oncology (american physical therapy association. oncology section)"] = "Rehabil Oncol",
  ["rehabilitation psychology"] = "Rehabil Psychol",
  ["rehabilitation record"] = "Rehabil Rec",
  ["rehabilitation research and practice"] = "Rehabil Res Pract",
  ["rehabilitation: sozialmedizin, physikalische medizin, praventivmedizin; internationale zeitschrift mit zentralblatt"] = "Rehabilitation (Bonn)",
  ["rehabilitation: sozialmedizin, physikalische medizin, präventivmedizin; internationale zeitschrift mit zentralblatt"] = "Rehabilitation (Bonn)",
  ["rehabilitation; internationale zeitschrift für physikalische medizin und rehabilitation mit zentralblatt"] = "Rehabilitation (Uelzen)",
  ["rehabilitation; sozialmedizin, physikalischemedizin, praventivmedizin"] = "Rehabilitation (Bonn)",
  ["rei cretariae romanae fautores. communicationes"] = "ReiCretCommunic",
  ["rei cretariae romanae fautorum acta"] = "ReiCretActa",
  ["reihe wissenschaft"] = "Reihe Wiss.",
  ["reihe, medizinische forschung"] = "Reihe Med Forsch",
  ["reihe: planung, information und unternehmungsfuhrung"] = "Reihe: Planung, Inform. Unternehmungsfuhr.",
  ["rein et foie"] = "Rein Foie",
  ["rein et foie, maladies de la nutrition; actualites"] = "Rein Foie",
  ["rein et foie, maladies de la nutrition; actualités"] = "Rein Foie",
  ["reinforced plastics"] = "Reinf. Plast.",
  ["reinforced plastics and composites world"] = "Reinf. Plast. Compos. World",
  ["rejuvenation research"] = "Rejuvenation Res",
  ["rekishigaku kenkyū"] = "Rekishigaku Kenkyu",
  ["relaciones (colegio de michoacán)"] = "Relaciones (Zamora)",
  ["relaciones internacionales"] = "Relac Int",
  ["relevé épidémiologique hebdomadaire"] = "Wkly Epidemiol Rec",
  ["reliability computing"] = "Reliab. Comput.",
  ["reliability engineerign"] = "Reliab. Eng.",
  ["reliability engineering"] = "Reliab. Eng.",
  ["reliability engineering & system safety"] = "Reliab. Eng. Syst. Saf.",
  ["reliable computing"] = "Reliab. Comput.",
  ["religion & gender"] = "Relig Gend",
  ["religion in life (new york, n.y.)"] = "Relig Life",
  ["religion, brain & behavior"] = "Religion Brain Behav",
  ["religion, state & society : the keston journal"] = "Relig State Soc",
  ["religious education (chicago, ill.)"] = "Relig Educ",
  ["religious humanism"] = "Relig Humanism",
  ["religious studies"] = "RelStud",
  ["religious studies review"] = "Relig Stud Rev",
  ["remaking economics: eminent post-war economists"] = "Remaking Econ. Eminent Post-War Econ.",
  ["remedial and special education : rase"] = "Remedial Spec Educ",
  ["remediation (new york, n.y.)"] = "Remediation (N Y)",
  ["remediation journal"] = "Rem. J.",
  ["remedy open access"] = "Rem Open Access",
  ["remote sensing"] = "Remote Sens (Basel)",
  ["remote sensing applications : society and environment"] = "Remote Sens Appl",
  ["remote sensing applications: society and environment"] = "Remote Sens. Appl.: Soc. Environ.",
  ["remote sensing in earth systems sciences"] = "Remote Sens Earth Syst Sci",
  ["remote sensing in ecology and conservation"] = "Remote Sens Ecol Conserv",
  ["remote sensing letters"] = "Remote Sens. Lett.",
  ["remote sensing letters (print)"] = "Remote Sens Lett",
  ["remote sensing of environment"] = "Remote Sens. Environ.",
  ["remote sensing of the environment"] = "Remote Sens. Environ.",
  ["remote sensing series"] = "Remote Sens. Ser.",
  ["ren kou dong tai"] = "Ren Kou Dong Tai",
  ["ren kou xue kan (changchun shi, china)"] = "Ren Kou Xue Kan",
  ["ren kou xue kan / kuo li tai-wan ta hsueh"] = "Ren Kou Xue Kan",
  ["ren kou xue kan / kuo li tai-wan ta hsüeh"] = "Ren Kou Xue Kan",
  ["ren kou yan jiu = renkou yanjiu"] = "Renkou Yanjiu",
  ["ren kou yu jing ji = renkou yu jingji"] = "Renkou Yu Jingji",
  ["ren wen ji she hui ke xue ji kan"] = "Ren Wen Ji She Hui Ke Xue Ji Kan",
  ["renaissance and modern studies"] = "Renaiss Mod Stud",
  ["renaissance and reformation"] = "Renaiss Reform",
  ["renaissance drama (evanston, ill. : 1964)"] = "Renaiss Drama",
  ["renaissance quarterly"] = "Renaiss Q",
  ["renaissance studies : journal of the society for renaissance studies"] = "Renaiss Stud",
  ["renal failure"] = "Ren Fail",
  ["renal nutrition forum"] = "Renal Nutr Forum",
  ["renal physiology"] = "Ren Physiol",
  ["renal physiology and biochemistry"] = "Ren Physiol Biochem",
  ["renal replacement therapy"] = "Ren Replace Ther",
  ["rendiconti - istituto superiore di sanita"] = "Rend Ist Sup Sanit",
  ["rendiconti - istituto superiore di sanità"] = "Rend Ist Sup Sanit",
  ["rendiconti / accademia nazionale del xl"] = "Rend Accad Naz XL",
  ["rendiconti del circolo matematico di palermo"] = "Rend. Circ. Mat. Palermo",
  ["rendiconti del circolo matematico di palermo. second series"] = "Rend. Circ. Mat. Palermo (2)",
  ["rendiconti del seminario della facoltà di scienze dell’università di cagliari"] = "Rend. Sem. Fac. Sci. Univ. Cagliari",
  ["rendiconti del seminario matematico della universita di padova"] = "Rend. Semin. Mat. Univ. Padova",
  ["rendiconti del seminario matematico della università di padova"] = "Rend. Sem. Mat. Univ. Padova",
  ["rendiconti del seminario matematico di messina"] = "Rend. Sem. Mat. Messina Ser. II",
  ["rendiconti del seminario matematico e fisico di milano"] = "Rend. Sem. Mat. Fis. Milano",
  ["rendiconti del seminario matematico. universita e politecnico torino"] = "Rend. Semin. Mat. Univ. Politec. Torino",
  ["rendiconti della accademia di archeologia, lettere e belle arti, napoli"] = "RendNap",
  ["rendiconti della accademia nazionale delle scienze detta dei xl"] = "Rend. Accad. Naz. Sci. XL Mem. Sci. Fis. Natur. (5)",
  ["rendiconti della pontificia accademia di archeologia"] = "RPAA",
  ["rendiconti dell’accademia delle scienze dell’istituto di bologna, classe di scienze morali"] = "RAIB",
  ["rendiconti dell’accademia di archeologia, lettere e belle arti di napoli"] = "RAAN",
  ["rendiconti dell’accademia nazionale dei lincei, classe di scienze morali, storiche e filologiche"] = "RendLinc",
  ["rendiconti dell’istituto di matematica dell’universita di trieste. an international journal of mathematics"] = "Rend. Istit. Mat. Univ. Trieste",
  ["rendiconti dell’istituto di matematica dell’università di trieste"] = "Rend. Istit. Mat. Univ. Trieste",
  ["rendiconti dell’istituto lombardo, classe di lettere, scienze morali e storiche"] = "RIL",
  ["rendiconti di matematica e delle sue applicazioni"] = "Rend. Mat. Appl. (7)",
  ["rendiconti di matematica e delle sue applicazioni. serie vii"] = "Rend. Mat. Appl. (7)",
  ["rendiconti lincei-matematica e applicazioni"] = "Rend. Lincei-Mat. Appl.",
  ["rendiconti lincei. scienze fisiche e naturali"] = "Rend Lincei Sci Fis Nat",
  ["rendiconti online societa geologica italiana"] = "Rend. Online Soc. Geol. Ital.",
  ["rendiconti. atti della pontificia accademia romana di archeologia"] = "RendPontAc",
  ["rendiconti. b, scienze biologiche e mediche"] = "Rend Ist Lombardo Accad Sci Lett Sez",
  ["rendiconti. classe di lettere e scienze morali e storiche, istituto lombardo, accademia di scienze e lettere"] = "RendIstLomb",
  ["renewable & sustainable energy reviews"] = "Renewable Sustainable Energy Rev.",
  ["renewable agriculture and food systems"] = "Renewable Agric. Food Syst.",
  ["renewable and sustainable energy reviews"] = "Renewable Sustainable Energy Rev.",
  ["renewable and sustainable energy transition"] = "Renewable Sustainable Energy Transition",
  ["renewable energy"] = "Renewable Energy",
  ["renewable energy and environmental sustainability"] = "Renewable Energy Environ. Sustainability",
  ["renewable energy and power quality journal"] = "Renewable Energy Power Qual. J.",
  ["renewable energy focus"] = "Renewable Energy Focus",
  ["renewable resources journal"] = "Renewable Resour. J.",
  ["renewables: wind, water, and solar"] = "Renew Wind Water Sol",
  ["rengong jingti xuebao"] = "Rengong Jingti Xuebao",
  ["report & transactions of the devonshire association for the advancement of science, literature and art. devonshire association for the advancement of science, literature and art"] = "Rep Trans Devonshire Assoc",
  ["report (group for the advancement of psychiatry : 1984)"] = "Rep Group Adv Psychiatry",
  ["report (group for the advancement of psychiatry)"] = "Rep Group Adv Psychiatry",
  ["report (u.s. naval medical research laboratory)"] = "Rep U S Nav Med Res Lab",
  ["report (u.s. naval submarine medical center)"] = "Rep US Nav Submar Med Cent",
  ["report - commission on accreditation of rehabilitation facilities"] = "Rep Comm Accredit Rehabil Facil",
  ["report - institute on human values in medicine"] = "Rep Inst Hum Values",
  ["report - navy medical neuropsychiatric research unit"] = "Rep Nav Med Neuropsychiatr Res Unit",
  ["report - u. s. naval medical research laboratory"] = "Rep US Nav Med Res Lab",
  ["report / group for the advancement of psychiatry"] = "Rep Group Adv Psychiatry",
  ["report / victor-bostrom fund for the international planned parenthood federation. victor-bostrom fund"] = "Victor Bostrom Fund Rep",
  ["report and list of members. worshipful society of apothecaries of london. faculty of the history and philosophy of medicine and pharmacy"] = "Rep List Memb Fac Hist Phil Med Pharm Worshipf Soc Apoth London",
  ["report and transactions - the devonshire association for the advancement of science, literature and art. devonshire association for the advancement of science, literature and art"] = "Devon Assoc Adv Sci",
  ["report from the institute for philosophy & public policy"] = "Rep Inst Philos Public Policy",
  ["report mech 1983"] = "Rep. MECH 1983",
  ["report of health care"] = "Rep Health Care",
  ["report of proceedings / the scottish society of the history of medicine. scottish society of the history of medicine"] = "Rep Proc Scott Soc Hist Med",
  ["report of proceedings. scottish society of the history of medicine"] = "Rep Proc Scott Soc Hist Med",
  ["report of the congress, european orthodontic society"] = "Rep. Congr. Eur. Orthod. Soc.",
  ["report of the congress. european orthodontic society"] = "Rep Congr Eur Orthod Soc",
  ["report of the department of antiquities, cyprus"] = "RDAC",
  ["report of the keio institute of cultural and linguistic studies"] = "RKI",
  ["report of the new york state veterinary college at cornell university. new york state veterinary college"] = "Rep N Y State Vet Coll Cornell Univ",
  ["report of the tomato genetics cooperative. tomato genetics cooperative"] = "Rep Tomato Genet Coop Tomato Genet Coop",
  ["report on carcinogens : carcinogen profiles"] = "Rep Carcinog",
  ["report on carcinogens : carcinogen profiles / u.s. dept. of health and human services, public health service, national toxicology program"] = "Rep Carcinog",
  ["report on carcinogens background document for [substance name]"] = "Rep Carcinog Backgr Doc",
  ["report on emotional & behavioral disorders in youth"] = "Rep Emot Behav Disord Youth",
  ["report on medical guidelines & outcomes research"] = "Rep Med Guidel Outcomes Res",
  ["report on social norms"] = "Rep Soc Norms",
  ["report on the working of the museum department, malta, department of information"] = "RepMalta",
  ["report series in theoretical physics"] = "Rep. Ser. Theoret. Phys.",
  ["report to congress (united states. physician payment review commission : series)"] = "Rep Congr (Ser)",
  ["report to congress. united states. physician payment review commission"] = "Rep Congr Physician Paym Rev Comm",
  ["report, commission on accreditation of rehabilitation facilities"] = "Rep. Comm. Accredit. Rehabil. Facil.",
  ["report, group for the advancement of psychiatry"] = "Rep. Group Adv. Psychiatry",
  ["report, navy medical neuropsychiatric research unit"] = "Rep. Nav. Med. Neuropsychiatr. Res. Unit",
  ["report, u.s. army medical research and nutrition laboratory, denver"] = "Rep. US. Army Med. Res. Nutr. Lab. Denver",
  ["report, u.s. army medical research laboratory"] = "Rep. US. Army Med. Res. Lab.",
  ["report, u.s. navy experimental diving unit"] = "Rep. US. Navy Exp. Diving Unit",
  ["report, us naval submarine medical center"] = "Rep. US Nav. Submar. Med. Cent.",
  ["report. army medical research laboratory (u.s.)"] = "Rep US Army Med Res Lab",
  ["report. associated medical services, inc"] = "Rep Assoc Med Serv Inc Hannah Inst Hist Med",
  ["report. canada. defence research chemical laboratories"] = "Rep Can Def Res Chem Lab",
  ["report. canada. defence research kingston laboratory"] = "Rep Can Def Res Kings Lab",
  ["report. cucurbit genetics cooperative"] = "Rep Cucurbit Genet Coop",
  ["report. great britain. department of health for scotland"] = "Rep Dep Health Scotl",
  ["report. joint fao/who expert committee on nutrition"] = "Report",
  ["report. u.s. army medical research and nutrition laboratory"] = "Rep US Army Med Res Nutr Lab Denver",
  ["report. u.s. army medical research laboratory"] = "Rep US Army Med Res Lab",
  ["report. united states. army medical research and nutrition laboratory, denver"] = "Rep US Army Med Res Nutr Lab Denver",
  ["report. united states. navy experimental diving unit, washington, d. c"] = "Rep US Navy Exp Diving Unit",
  ["report. victor-bostrom fund"] = "Victor Bostrom Fund Rep",
  ["reporter (wilmington, del.)"] = "Rep Del Nurses Assoc",
  ["reporter on human reproduction and the law"] = "Report Hum Reprod Law",
  ["reporter, delaware nurses association"] = "Rep. Del. Nurses Assoc.",
  ["reports (international development research centre (canada))"] = "Rep Int Dev Res Cent Can",
  ["reports - american universities field staff. american universities field staff"] = "Rep Am Univ Field Staff",
  ["reports in advances of physical sciences"] = "Rep. Adv. Phys. Sci.",
  ["reports in medical imaging"] = "Rep. Med. Imaging",
  ["reports of biochemistry & molecular biology"] = "Rep Biochem Mol Biol",
  ["reports of biochemistry and molecular biology"] = "Rep. Biochem. Mol. Biol.",
  ["reports of cases argued and determined in the superior court, appellate division, chancery division, law division, and in the county courts of the state of new jersey. new jersey. superior court"] = "Rep Cases Argued Determ Super Court Appell Div Chancery Div Law Div Cty Courts State N J N J Super Court",
  ["reports of cases argued and determined in the supreme court of new jersey. new jersey. supreme court"] = "Rep Cases Argued Determ Supreme Court N J N J Supreme Court",
  ["reports of cases decided in the appellate division of the supreme court, state of new york. new york (state). supreme court. appellate division"] = "Rep Cases Decided Appell Div Supreme Court State NY NY State Supreme Court Appell Div",
  ["reports of cases determined in the courts of appeal of the state of california"] = "Rep Cases Determ Court Appeal State Calif",
  ["reports of enlarged sessions of seminar of i. vekua institute of applied mathematics"] = "Rep. Enlarged Sess. Semin. I. Vekua Appl. Math.",
  ["reports of faculty of science, shizuoka university"] = "Rep Fac Sci Shizuoka Univ",
  ["reports of practical oncology and radiotherapy : journal of greatpoland cancer center in poznań and polish society of radiation oncology"] = "Rep Pract Oncol Radiother",
  ["reports of selected cases decided in courts of the state of new york other than the court of appeals and the appellate division of the supreme court : miscellaneous reports"] = "Rep Sel Cases Decided Courts State N Y Court Appeals Appell Div Supreme Court",
  ["reports of statistical application research"] = "Rep. Statist. Appl. Res. Un. Japan. Sci. Engrs.",
  ["reports of the faculty of science"] = "Rep. Fac. Sci. Kagoshima Univ.",
  ["reports of the faculty of science and engineering"] = "Rep. Fac. Sci. Eng. Saga Univ. Math.",
  ["reports of the faculty of science and technology"] = "Rep. Fac. Sci. Technol. Meijo Univ.",
  ["reports of the institute of mathematics"] = "Rep. Inst. Math.",
  ["reports on health and social subjects"] = "Rep Health Soc Subj (Lond)",
  ["reports on mathematical logic"] = "Rep. Math. Logic",
  ["reports on mathematical physics"] = "Rep. Math. Phys.",
  ["reports on population/family planning"] = "Rep Popul Fam Plann",
  ["reports on progress in physics"] = "Rep. Prog. Phys.",
  ["reports on progress in physics. physical society (great britain)"] = "Rep Prog Phys",
  ["reports on public health and medical subjects"] = "Rep Public Health Med Subj (Lond)",
  ["reports on rheumatic diseases"] = "Rep Rheum Dis",
  ["reports scm"] = "Rep. SCM",
  ["reppal. revue des études phénicienne-puniques et des antiquités libyques"] = "Reppal",
  ["representation (mcdougall trust)"] = "Representation (McDougall Trust)",
  ["representation and mind"] = "Represent. Mind",
  ["representation theory"] = "Represent. Theory",
  ["representation theory. an electronic journal of the american mathematical society"] = "Represent. Theory",
  ["representations (berkeley, calif.)"] = "Representations (Berkeley)",
  ["representative research in social psychology"] = "Represent Res Soc Psychol",
  ["reprints in theory and applications of categories"] = "Repr. Theory Appl. Categ.",
  ["reproduction (bristol, united kingdom)"] = "Reproduction (Bristol, U. K.)",
  ["reproduction (cambridge, england)"] = "Reproduction",
  ["reproduction (cambridge, england) supplement"] = "Reprod Suppl",
  ["reproduction (cambridge, england). abstract series"] = "Reprod Abstr Ser",
  ["reproduction and contraception"] = "Reprod Contracept",
  ["reproduction fertility and development"] = "Reprod. Fertil. Dev.",
  ["reproduction in domestic animals"] = "Reprod. Domest. Anim.",
  ["reproduction in domestic animals = zuchthygiene"] = "Reprod Domest Anim",
  ["reproduction, fertility, and development"] = "Reprod Fertil Dev",
  ["reproduction, nutrition, development"] = "Reprod Nutr Dev",
  ["reproduction, nutrition, developpement"] = "Reprod. Nutr. Dev.",
  ["reproduction, nutrition, développement"] = "Reprod Nutr Dev",
  ["reproduction. abstract series"] = "Reprod. Abstr. Ser.",
  ["reproduction. supplement"] = "Reprod. Suppl.",
  ["reproductive and genetic engineering"] = "Reprod Genet Eng",
  ["reproductive biology"] = "Reprod Biol",
  ["reproductive biology and endocrinology"] = "Reprod. Biol. Endocrinol.",
  ["reproductive biology and endocrinology : rb&e"] = "Reprod Biol Endocrinol",
  ["reproductive biomedicine & society online"] = "Reprod Biomed Soc Online",
  ["reproductive biomedicine online"] = "Reprod Biomed Online",
  ["reproductive freedom news"] = "Reprod Freedom News",
  ["reproductive freedom news / from the center for reproductive law & policy"] = "Reprod Freedom News",
  ["reproductive health"] = "Reprod Health",
  ["reproductive health matters"] = "Reprod Health Matters",
  ["reproductive medicine (basel, switzerland)"] = "Reprod Med (Basel)",
  ["reproductive medicine and biology"] = "Reprod Med Biol",
  ["reproductive sciences"] = "Reprod. Sci.",
  ["reproductive sciences (thousand oaks, calif.)"] = "Reprod Sci",
  ["reproductive system & sexual disorders : current research"] = "Reprod Syst Sex Disord",
  ["reproductive toxicology"] = "Reprod. Toxicol.",
  ["reproductive toxicology (elmsford, n.y.)"] = "Reprod Toxicol",
  ["reprowatch (youth edition)"] = "ReproWatch (Youth Ed)",
  ["reptile rap"] = "Reptile Rap",
  ["reptiles & amphibians : conservation and natural history"] = "Reptiles Amphib",
  ["repura (leprosy)"] = "Repura",
  ["repura. leprosy"] = "Repura",
  ["república de venezuela"] = "Repúb. Venezuela Bol. Acad. Cienc. Fís. Mat. Natur.",
  ["requirements engineering"] = "Requir Eng",
  ["res ipsa loquitur (washington, d.c.)"] = "Res Ipsa Loquitur",
  ["res publica (liverpool, england)"] = "Res Publica",
  ["res publica litterarum"] = "Res Publica Litt",
  ["res publica litterarum: studies in the classical tradition"] = "RPL",
  ["research & knowledge"] = "Res Knowl",
  ["research & reviews in polymer"] = "Res. Rev. Polym.",
  ["research & reviews. drug delivery"] = "Res Rev Drug Deliv",
  ["research & reviews. journal of engineering and technology"] = "Res Rev J Eng Technol",
  ["research & reviews. journal of hospital and clinical pharmacy"] = "Res Rev J Hosp Clin Pharm",
  ["research & reviews. journal of microbiology and biotechnology"] = "Res Rev J Microbiol Biotechnol",
  ["research & reviews. journal of nursing and health sciences"] = "Res Rev J Nurs Health Sci",
  ["research & reviews. journal of pharmacy and pharmaceutical sciences"] = "Res Rev J Pharm Pharm Sci",
  ["research & reviews. journal of statistics and mathematical sciences"] = "Res Rev J Stat Math Sci",
  ["research (lambertville, n. j.)"] = "Research (Lambertville)",
  ["research (washington, d.c.)"] = "Research (Wash D C)",
  ["research activities"] = "Res. Act. Fac. Sci. Eng. Tokyo Denki Univ.",
  ["research agenda brief"] = "Res Agenda Brief",
  ["research and advances in psychiatry"] = "Res Adv Psychiatry",
  ["research and clinical studies in headache"] = "Res Clin Stud Headache",
  ["research and development technical report. united states. naval radiological defense laboratory, san francisco"] = "Res Dev Tech Rep",
  ["research and exposition in mathematics"] = "Res. Exp. Math.",
  ["research and practice for persons with severe disabilities : the journal of tash"] = "Res Pract Persons Severe Disabl",
  ["research and practice in alzheimer's disease"] = "Res Pract Alzheimers Dis",
  ["research and practice in technology enhanced learning"] = "Res Pract Technol Enhanc Learn",
  ["research and practice in the schools : the official journal of the texas association of school psychologists"] = "Res Prac Sch",
  ["research and practice in thrombosis and haemostasis"] = "Res Pract Thromb Haemost",
  ["research and reports in biochemistry"] = "Res Rep Biochem",
  ["research and reports in biology"] = "Res Rep Biol",
  ["research and reports in endocrine disorders"] = "Res Rep Endocr Disord",
  ["research and reports in forensic medical science"] = "Res. Rep. Forensic Med. Sci.",
  ["research and reports in medicinal chemistry"] = "Res Rep Med Chem",
  ["research and reports in neonatology"] = "Res Rep Neonatol",
  ["research and reports in nuclear medicine"] = "Res Rep Nucl Med",
  ["research and reports in tropical medicine"] = "Res Rep Trop Med",
  ["research and reports in urology"] = "Res Rep Urol",
  ["research and reviews in biosciences"] = "Res Rev Biosci",
  ["research and statistics note - health care financing administration, office of policy, planning, and research"] = "Res Stat Note Health Care Financ Adm Off Policy Plan Res",
  ["research and statistics note - social security administration, office of research and statistics"] = "Res Stat Note",
  ["research and statistics note, health care financing administration, office of policy, planning, and research"] = "Res. Stat. Note Health Care Financ. Adm. Off. Policy. Plan. Res.",
  ["research and statistics note, social security administration, office of research and statistics"] = "Res. Stat. Note",
  ["research and theory for nursing practice"] = "Res Theory Nurs Pract",
  ["research brief"] = "Res Brief",
  ["research bulletin (sūn chīwawitthayā thāng thalē phūket)"] = "Res Bull (Sun Chiwawitthaya Thang Thale Phuket)",
  ["research bulletin of the panjab university"] = "Res. Bull. Panjab Univ. Sci.",
  ["research bulletin. united states. army air forces. training command. psychological section"] = "Res Bull U. S.",
  ["research bulletins of the college experiment forests, hokkaido university"] = "Res. Bull. Coll. Exp. For., Hokkaido Univ.",
  ["research communications - institute for fermentation, osaka"] = "Res Commun Inst Ferment",
  ["research communications in biochemistry and cell & molecular biology"] = "Res Commun Biochem Cell Mol Biol",
  ["research communications in chemical pathology and pharmacology"] = "Res Commun Chem Pathol Pharmacol",
  ["research communications in molecular pathology and pharmacology"] = "Res Commun Mol Pathol Pharmacol",
  ["research communications in pharmacology and toxicology"] = "Res. Commun. Pharmacol. Toxicol.",
  ["research communications in psychology, psychiatry and behavior"] = "Res Commun Psychol Psychiatr Behav",
  ["research communications in substance abuse"] = "Res Commun Subst Abuse",
  ["research communications in substances of abuse"] = "Res. Commun. Subst. Abuse",
  ["research communications, institute for fermentation"] = "Res. Commun. Inst. Ferment.",
  ["research disclosure"] = "Res. Discl.",
  ["research ethics"] = "Res Ethics",
  ["research evaluation"] = "Res Eval",
  ["research frontiers in fertility regulation : rffr"] = "Res Front Fertil Regul",
  ["research frontiers in fertility regulation : rffr / parfr"] = "Res Front Fertil Regul",
  ["research ideas & outcomes"] = "Res. Ideas Outcomes",
  ["research ideas and outcomes"] = "Res Ideas Outcomes",
  ["research in agriculture and agronomy"] = "Res Agric Agron",
  ["research in astronomy and astrophysics"] = "Res. Astron. Astrophys.",
  ["research in autism spectrum disorders"] = "Res Autism Spectr Disord",
  ["research in biology"] = "Res Biol",
  ["research in biotechnology"] = "Res Biotechnol",
  ["research in cardiovascular medicine"] = "Res Cardiovasc Med",
  ["research in cold and arid regions"] = "Res. Cold Arid Reg.",
  ["research in community and mental health"] = "Res Community Ment Health",
  ["research in comparative and international education"] = "Res Comp Int Educ",
  ["research in developmental disabilities"] = "Res Dev Disabil",
  ["research in diagnostic and interventional imaging"] = "Res. Diagn. Interventional Imaging",
  ["research in drama education"] = "Res Drama Educ",
  ["research in economic anthropology"] = "Res Econ Anthropol",
  ["research in economic history"] = "Res Econ Hist",
  ["research in economics"] = "Res. Econ.",
  ["research in economics = ricerche economiche"] = "Res Econ",
  ["research in engineering design"] = "Res. Eng. Des.",
  ["research in experimental medicine"] = "Res. Exp. Med.",
  ["research in experimental medicine. zeitschrift fur die gesamte experimentelle medizin einschliesslich experimenteller chirurgie"] = "Res Exp Med (Berl)",
  ["research in experimental medicine. zeitschrift für die gesamte experimentelle medizin einschliesslich experimenteller chirurgie"] = "Res Exp Med (Berl)",
  ["research in gerontological nursing"] = "Res Gerontol Nurs",
  ["research in health economics"] = "Res Health Econ",
  ["research in health science"] = "Res Health Sci",
  ["research in higher education"] = "Res High Educ",
  ["research in higher education journal"] = "Res High Educ J",
  ["research in human capital and development"] = "Res Hum Cap Dev",
  ["research in human development"] = "Res Hum Dev",
  ["research in immunology"] = "Res Immunol",
  ["research in international business and finance"] = "Res Int Bus Finance",
  ["research in law and economics"] = "Res. Law Econ.",
  ["research in life sciences"] = "Res. Life Sci.",
  ["research in mathematics"] = "Res. Math.",
  ["research in mathematics education"] = "Res. Math. Educ.",
  ["research in microbiology"] = "Res Microbiol",
  ["research in nondestructive evaluation"] = "Res. Nondestr. Eval.",
  ["research in number theory"] = "Res. Number Theory",
  ["research in nursing & health"] = "Res Nurs Health",
  ["research in nursing and health"] = "Res. Nurs. Health",
  ["research in organizational behavior"] = "Res Organ Behav",
  ["research in pharmaceutical sciences"] = "Res Pharm Sci",
  ["research in plant biology"] = "Res Plant Biol",
  ["research in population economics"] = "Res Popul Econ",
  ["research in prostaglandins"] = "Res Prostaglandins",
  ["research in psychotherapy (milano)"] = "Res Psychother",
  ["research in reproduction"] = "Res Reprod",
  ["research in rural sociology and development"] = "Res Rural Sociol Dev",
  ["research in science education"] = "Res Sci Educ",
  ["research in social & administrative pharmacy : rsap"] = "Res Social Adm Pharm",
  ["research in social stratification and mobility"] = "Res Soc Stratif Mobil",
  ["research in sociology of education"] = "Res Sociol Educ",
  ["research in sociology of education and socialization"] = "Res Sociol Educ Social",
  ["research in sports medicine"] = "Res. Sports Med.",
  ["research in sports medicine (print)"] = "Res Sports Med",
  ["research in statistics"] = "Res. Stat.",
  ["research in the mathematical sciences"] = "Res. Math. Sci.",
  ["research in the schools : a nationally refereed journal sponsored by the mid-south educational research association and the university of alabama"] = "Res Sch",
  ["research in the service of medicine"] = "Res Serv Med",
  ["research in the social scientific study of religion"] = "Res Soc Sci Study Relig",
  ["research in the sociology of health care"] = "Res Sociol Health Care",
  ["research in the sociology of work"] = "Res Sociol Work",
  ["research in urban economics"] = "Res Urban Econ",
  ["research in veterinary science"] = "Res Vet Sci",
  ["research in virology"] = "Res Virol",
  ["research initiative, treatment action : rita"] = "Res Initiat Treat Action",
  ["research integrity and peer review"] = "Res Integr Peer Rev",
  ["research involvement and engagement"] = "Res Involv Engagem",
  ["research journal of agriculture and biological sciences"] = "Res J Agric Biol Sci",
  ["research journal of applied sciences"] = "Res J Appl Sci",
  ["research journal of biotechnology"] = "Res J Biotechnol",
  ["research journal of chemical and environmental sciences"] = "Res J Chem Environ Sci",
  ["research journal of chemistry and environment"] = "Res J Chem Environ",
  ["research journal of environmental sciences"] = "Res. J. Environ. Sci.",
  ["research journal of health sciences"] = "Res J Health Sci",
  ["research journal of life sciences, bioinformatics, pharmaceutical and chemical sciences"] = "Res J Life Sci Bioinform Pharm Chem Sci",
  ["research journal of medical sciences"] = "Res J Med Sci",
  ["research journal of microbiology"] = "Res J Microbiol",
  ["research journal of pharmaceutical, biological and chemical sciences"] = "Res J Pharm Biol Chem Sci",
  ["research journal of pharmacy and technology"] = "Res J Pharm Technol",
  ["research journal of recent sciences"] = "Res J Recent Sci",
  ["research journal of textile and apparel"] = "Res. J. Text. Apparel",
  ["research journal of veterinary sciences"] = "Res J Vet Sci",
  ["research journal of women's health"] = "Res J Womens Health",
  ["research letters in biochemistry"] = "Res Lett Biochem",
  ["research letters in signal processing"] = "Res Lett Signal Process",
  ["research management"] = "Res Manage",
  ["research management review"] = "Res Manag Rev",
  ["research methods in medicine & health sciences"] = "Res. Methods Med. Health Sci.",
  ["research newsletter. college of general practitioners (great britain)"] = "Res Newsl",
  ["research notes in mathematics"] = "Res. Notes Math.",
  ["research notes of the aas"] = "Res. Notes AAS",
  ["research on aging"] = "Res Aging",
  ["research on biomedical engineering"] = "Res. Biomed. Eng.",
  ["research on chemical intermediates"] = "Res. Chem. Intermed.",
  ["research on child and adolescent psychopathology"] = "Res Child Adolesc Psychopathol",
  ["research on economic inequality"] = "Res Econ Inequal",
  ["research on engineering structures and materials"] = "Res. Eng. Struct. Mater.",
  ["research on intelligent manufacturing (rims)"] = "Res. Intell. Manuf. (RIMS)",
  ["research on language and social interaction"] = "Res Lang Soc Interact",
  ["research on social work practice"] = "Res Soc Work Pract",
  ["research on steroids"] = "Res Steroids (Amst)",
  ["research opinions in animal & veterinary sciences"] = "Res Opin Anim Vet Sci",
  ["research paper usda forest service"] = "Res. Pap. USDA For. Serv.",
  ["research papers (university of arkansas for medical sciences. history of medicine associates)"] = "Res Pap Hist Med Assoc",
  ["research papers in education"] = "Res Pap Educ",
  ["research perspectives ghent analysis and pde center"] = "Res. Perspect. Ghent Anal. PDE Cent.",
  ["research perspectives on migration : a joint project of the international migration policy program of the carnegie endowment for international peace and the urban institute"] = "Res Perspect Migr",
  ["research policy"] = "Res Policy",
  ["research program. united states. army. medical department"] = "Res Program",
  ["research progress in organic, biological and medicinal chemistry"] = "Res Prog Org Biol Med Chem",
  ["research publications - association for research in nervous and mental disease"] = "Res Publ Assoc Res Nerv Ment Dis",
  ["research publications of the wellcome unit for the history of medicine"] = "Res Publ Wellcome Unit Hist Med",
  ["research publications, association for research in nervous and mental disease"] = "Res. Publ. Assoc. Res. Nerv. Ment. Dis.",
  ["research quarterly"] = "Res Q",
  ["research quarterly for exercise and sport"] = "Res Q Exerc Sport",
  ["research report (health effects institute)"] = "Res Rep Health Eff Inst",
  ["research report / health effects institute"] = "Res Rep Health Eff Inst",
  ["research report, health effects institute"] = "Res. Rep. Health Eff. Inst.",
  ["research report, united states naval school of aviation medicine"] = "Res. Rep. U. S. Nav. Sch. Aviat. Med.",
  ["research report. brooke army medical center. surgical research unit"] = "Res Rep Brooke Army Med Cent",
  ["research report. naval medical research institute (u.s.)"] = "Res Rep Nav Med Res Inst (US)",
  ["research report. naval school of aviation medicine (u.s.)"] = "Res Rep U S Nav Sch Aviat Med",
  ["research reports"] = "Res. Rep.",
  ["research reports (montgomery)"] = "Res Rep (Montgomery, Ala.)",
  ["research reports in clinical cardiology"] = "Res. Rep. Clin. Cardiol.",
  ["research reports of the anan college of technology"] = "Res. Rep. Anan College Tech.",
  ["research reports on information sciences"] = "Res. Rep. Inf. Sci.",
  ["research reports, united states naval medical field research laboratory"] = "Res. Rep. U. S. Nav. Med. Field Res. Lab.",
  ["research resources reporter"] = "Res Resour Rep",
  ["research results in pharmacology"] = "Res. Results Pharmacol.",
  ["research reviews. united states. office of naval research"] = "Res Rev",
  ["research square"] = "Res Sq",
  ["research studies (pullman, wash.)"] = "Res Stud",
  ["research summaries"] = "Res Summ (Nav Med Res Inst)",
  ["research synthesis methods"] = "Res Synth Methods",
  ["research today"] = "Res Today",
  ["research trends in physics"] = "Res. Trends Phys.",
  ["research; a journal of science and its applications"] = "Research",
  ["researches on population ecology"] = "Res Popul Ecol (Kyoto)",
  ["reseaux - ciephum"] = "Reseaux Ciephum",
  ["resena historica (mexico city, mexico)"] = "Resena Hist",
  ["resenha clinico-cientifica"] = "Resen. Clin. Cient.",
  ["resenha clínico-científica"] = "Resen Clin Cient",
  ["resenha médica"] = "Resen Med",
  ["resenhas do instituto de matemática e estatística da universidade de são paulo"] = "Resenhas",
  ["reseña histórica (mexico city, mexico)"] = "Resena Hist",
  ["resident and staff physician"] = "Resid Staff Physician",
  ["residential treatment for children & youth"] = "Resid Treat Child Youth",
  ["residue reviews"] = "Residue Rev",
  ["resilient cities and structures"] = "Resilient Cities Struct.",
  ["resource and energy economics"] = "Resour Energy Econ",
  ["resource conservation"] = "Resour. Conserv.",
  ["resource engineering and technology for sustainable world"] = "Resour. Eng. Technol. Sustainable World",
  ["resource geology"] = "Resour. Geol.",
  ["resource recovery and conservation"] = "Resour. Recovery Conserv.",
  ["resource-efficient technologies"] = "Resour.-Effic. Technol.",
  ["resources and energy"] = "Resources Energy",
  ["resources chemicals and materials"] = "Resour. Chem. Mater.",
  ["resources for biomedical research and education"] = "Resour Biomed Res Educ",
  ["resources for feminist research : rfr = documentation sur la recherche féministe : drf"] = "Resour Fem Res",
  ["resources policy"] = "Resour. Policy",
  ["resources, conservation & recycling advances"] = "Resour. Conserv. Recyl. Adv.",
  ["resources, conservation and recycling"] = "Resour. Conserv. Recycl.",
  ["resources, conservation and recycling: x"] = "Resour. Conserv. Recycl.: X",
  ["resources, conservation, and recycling"] = "Resour Conserv Recycl",
  ["resources, environment and sustainability"] = "Resour. Environ. Sustainability",
  ["respiration physiology"] = "Respir Physiol",
  ["respiration; international review of thoracic diseases"] = "Respiration",
  ["respiratory care"] = "Respir Care",
  ["respiratory care clinics of north america"] = "Respir Care Clin N Am",
  ["respiratory investigation"] = "Respir Investig",
  ["respiratory management"] = "Respir Manage",
  ["respiratory medicine"] = "Respir Med",
  ["respiratory medicine and research"] = "Respir Med Res",
  ["respiratory medicine case reports"] = "Respir Med Case Rep",
  ["respiratory medicine cme"] = "Respir Med CME",
  ["respiratory medicine: x"] = "Respir Med X",
  ["respiratory physiology & neurobiology"] = "Respir Physiol Neurobiol",
  ["respiratory physiology and neurobiology"] = "Respir. Physiol. Neurobiol.",
  ["respiratory research"] = "Respir Res",
  ["respiratory therapy"] = "Respir Ther",
  ["respirology (carlton, vic.)"] = "Respirology",
  ["respirology case reports"] = "Respirol Case Rep",
  ["responsa meridiana"] = "Responsa Meridiana",
  ["restaurants & institutions"] = "Restaurants Inst",
  ["restaurants and institutions"] = "Restaurants Inst.",
  ["restoration ecology"] = "Restor. Ecol.",
  ["restoration quarterly"] = "Restor Q",
  ["restorative dentistry"] = "Restorative Dent",
  ["restorative dentistry & endodontics"] = "Restor Dent Endod",
  ["restorative neurology and neuroscience"] = "Restor Neurol Neurosci",
  ["results and problems in cell differentiation"] = "Results Probl Cell Differ",
  ["results in applied mathematics"] = "Results Appl. Math.",
  ["results in chemistry"] = "Results Chem",
  ["results in control and optimization"] = "Results Control Optim.",
  ["results in engineering"] = "Results Eng.",
  ["results in geochemistry"] = "Results in Geochem.",
  ["results in geophysical sciences"] = "Results Geophys. Sci.",
  ["results in immunology"] = "Results Immunol",
  ["results in materials"] = "Results Mater.",
  ["results in mathematics"] = "Results Math.",
  ["results in optics"] = "Results Opt.",
  ["results in pharma sciences"] = "Results Pharma Sci",
  ["results in physics"] = "Results Phys.",
  ["results in surfaces and interfaces"] = "Results Surf. Interfaces",
  ["resuscitation plus"] = "Resusc Plus",
  ["resúmenes de trabajos médicos venezolanos"] = "Resum Trab Med Venez",
  ["rethinking ecology"] = "Rethinking Ecol.",
  ["rethinking history"] = "Rethink Hist",
  ["retina (philadelphia, pa.)"] = "Retina",
  ["retinal cases & brief reports"] = "Retin Cases Brief Rep",
  ["retinal physician"] = "Retin Physician",
  ["retrovirology : research and treatment"] = "Retrovirology (Auckl)",
  ["reumatologia polska"] = "Reumatol Pol",
  ["reumatología clinica"] = "Reumatol Clin",
  ["revealing antiquity"] = "Reveal. Antiq.",
  ["revenue-cycle strategist"] = "Revenue-cycle Strateg",
  ["review (federation of american health systems)"] = "Rev Fed Am Health Syst",
  ["review (international commission of jurists (1952- )"] = "Rev Int Comm Jurists",
  ["review (patient focused care association)"] = "Rev Patient Focus Care Assoc",
  ["review - federation of american hospitals"] = "Rev Fed Am Hosp",
  ["review for religious"] = "Rev Relig",
  ["review journal of autism and developmental disorders"] = "Rev J Autism Dev Disord",
  ["review journal of chemistry"] = "Rev J Chem",
  ["review of accounting studies"] = "Rev. Acc. Stud.",
  ["review of african political economy"] = "Rev Afr Polit Econ",
  ["review of agricultural economics"] = "Rev. Agr. Econ.",
  ["review of agricultural, food and environmental studies"] = "Rev. Agric., Food Environ. Stud.",
  ["review of allergy"] = "Rev Allergy",
  ["review of allergy and applied immunology"] = "Rev Allergy Appl Immunol",
  ["review of asset pricing studies"] = "Rev Asset Pricing Stud",
  ["review of austrian economics"] = "Rev. Austrian Econ.",
  ["review of automotive engineering"] = "Rev. Automot. Eng.",
  ["review of black political economy"] = "Rev. Black Polit. Economy",
  ["review of business and economic research"] = "Rev. Bus. Econ. Res.",
  ["review of czechoslovak medicine"] = "Rev Czech Med",
  ["review of derivatives research"] = "Rev. Derivatives Res.",
  ["review of development economics"] = "Rev Dev Econ",
  ["review of disability studies"] = "Rev Disabil Stud",
  ["review of eastern medical sciences"] = "Rev East Med Sci",
  ["review of economic conditions in italy"] = "Rev Econ Cond Italy",
  ["review of economic design"] = "Rev. Econ. Des.",
  ["review of economic dynamics"] = "Rev Econ Dyn",
  ["review of economic studies"] = "Rev. Econ. Stud.",
  ["review of economics and statistics"] = "Rev. Econ. Statist.",
  ["review of economics of the household"] = "Rev Econ Househ",
  ["review of educational research"] = "Rev Educ Res",
  ["review of environmental economics and policy"] = "Rev Environ Econ Policy",
  ["review of european studies"] = "Rev Eur Stud",
  ["review of european, comparative & international environmental law"] = "Rev Eur Comp Int Environ Law",
  ["review of finance"] = "Rev Financ",
  ["review of financial economics"] = "Rev. Finan. Econ.",
  ["review of financial studies"] = "Rev. Finan. Stud.",
  ["review of general psychology : journal of division 1, of the american psychological association"] = "Rev Gen Psychol",
  ["review of historical demography"] = "Rev Hist Demogr",
  ["review of human factors and ergonomics"] = "Rev Hum Factors Ergon",
  ["review of income and wealth"] = "Rev. Income Wealth",
  ["review of industrial organization"] = "Rev. Ind. Organ.",
  ["review of international economics"] = "Rev Int Econ",
  ["review of international organizations"] = "Rev. Int. Organ.",
  ["review of international political economy"] = "Rev Int Polit Econ",
  ["review of international studies"] = "Rev Int Stud",
  ["review of law and social change. new york university"] = "Rev Law Soc Change",
  ["review of marketing and agricultural economics"] = "Rev. Marketing Agr. Econ.",
  ["review of metaphysics"] = "RMeta",
  ["review of modern logic"] = "Rev. Mod. Logic",
  ["review of palaeobotany and palynology"] = "Rev Palaeobot Palynol",
  ["review of philosophy and psychology"] = "Rev Philos Psychol",
  ["review of physical chemistry of japan"] = "Rev. Phys. Chem. Jpn.",
  ["review of physics in technology"] = "Rev. Phys. Technol.",
  ["review of plant pathology"] = "Rev. Plant Pathol.",
  ["review of political economy"] = "Rev. Polit. Economy",
  ["review of progress in coloration and related topics"] = "Rev. Prog. Color. Relat. Top.",
  ["review of public data use"] = "Rev Public Data Use",
  ["review of public personnel administration"] = "Rev Public Pers Adm",
  ["review of quantitative finance and accounting"] = "Rev. Quant. Finance Acc.",
  ["review of radical political economics"] = "Rev. Radical Polit. Econ.",
  ["review of regional research"] = "Rev. Reg. Res.",
  ["review of regional studies"] = "Rev. Reg. Stud.",
  ["review of religious research"] = "Rev Relig Res",
  ["review of scientific instruments"] = "Rev. Sci. Instrum.",
  ["review of social economy"] = "Rev Soc Econ",
  ["review of social policy"] = "Rev Soc Policy",
  ["review of socialist law"] = "Rev Social Law",
  ["review of socionetwork strategies"] = "Rev. Socionetwork Strategies",
  ["review of surgery"] = "Rev Surg",
  ["review of symbolic logic"] = "Rev. Symb. Log.",
  ["review of the activities metallgesellschaft ag"] = "Rev. Act. Metallges. AG",
  ["review of urban and regional development studies"] = "Rev. Urban Reg. Devel. Stud.",
  ["review of urban and regional development studies : rurds : journal of the applied regional conference"] = "RURDS Rev Urb Reg Dev Stud",
  ["review of world economics"] = "Rev. World Econ.",
  ["review, federation of american health systems"] = "Rev. Fed. Am. Health Syst.",
  ["review, federation of american hospitals"] = "Rev. Fed. Am. Hosp.",
  ["review, patient focused care association"] = "Rev. Patient Focus. Care Assoc.",
  ["reviews and advances in chemistry"] = "Rev. Adv. Chem.",
  ["reviews and research in medical microbiology"] = "Rev. Res. Med. Microbiol.",
  ["reviews in american history"] = "Rev Am Hist",
  ["reviews in analgesia"] = "Rev Analg",
  ["reviews in analytical chemistry"] = "Rev. Anal. Chem",
  ["reviews in aquaculture"] = "Rev Aquac",
  ["reviews in cardiovascular medicine"] = "Rev Cardiovasc Med",
  ["reviews in chemical engineering"] = "Rev. Chem. Eng.",
  ["reviews in clinical & basic pharmacology"] = "Rev Clin Basic Pharm",
  ["reviews in clinical and basic pharmacology"] = "Rev. Clin. Basic Pharm.",
  ["reviews in clinical and experimental hematology"] = "Rev Clin Exp Hematol",
  ["reviews in clinical gerontology"] = "Rev Clin Gerontol",
  ["reviews in computational chemistry"] = "Rev. Comput. Chem.",
  ["reviews in conservation"] = "Rev. Conserv.",
  ["reviews in endocrine & metabolic disorders"] = "Rev Endocr Metab Disord",
  ["reviews in endocrine and metabolic disorders"] = "Rev. Endocr. Metab. Disord.",
  ["reviews in environmental science and bio/technology"] = "Rev. Environ. Sci. Bio/Technol.",
  ["reviews in environmental science and biotechnology"] = "Rev. Environ. Sci. Biotechnol.",
  ["reviews in environmental toxicology"] = "Rev. Environ. Toxicol.",
  ["reviews in fish biology and fisheries"] = "Rev Fish Biol Fish",
  ["reviews in fisheries science & aquaculture"] = "Rev. Fish. Sci. Aquacult.",
  ["reviews in fluorescence"] = "Rev Fluoresc",
  ["reviews in gastroenterological disorders"] = "Rev Gastroenterol Disord",
  ["reviews in health care"] = "Rev Health Care",
  ["reviews in immunogenetics"] = "Rev Immunogenet",
  ["reviews in infection"] = "Rev Infect",
  ["reviews in inorganic chemistry"] = "Rev. Inorg. Chem.",
  ["reviews in mathematical physics"] = "Rev. Math. Phys.",
  ["reviews in mathematical physics. a journal for both review and original research papers in the field of mathematical physics"] = "Rev. Math. Phys.",
  ["reviews in mathematics and mathematical physics"] = "Rev. Math. Math. Phys.",
  ["reviews in medical microbiology : a journal of the pathological society of great britain and ireland"] = "Rev Med Microbiol",
  ["reviews in medical virology"] = "Rev Med Virol",
  ["reviews in mineralogy and geochemistry"] = "Rev. Mineral. Geochem.",
  ["reviews in molecular biotechnology"] = "Rev. Mol. Biotechnol.",
  ["reviews in neurological diseases"] = "Rev Neurol Dis",
  ["reviews in obstetrics & gynecology"] = "Rev Obstet Gynecol",
  ["reviews in pain"] = "Rev Pain",
  ["reviews in physics"] = "Rev. Phys.",
  ["reviews in pure & applied pharmacological sciences"] = "Rev Pure Appl Pharmacol Sci",
  ["reviews in pure and applied pharmacological sciences"] = "Rev. Pure Appl. Pharmacol. Sci.",
  ["reviews in the neurosciences"] = "Rev Neurosci",
  ["reviews in toxicology"] = "Rev. Toxicol.",
  ["reviews in urology"] = "Rev Urol",
  ["reviews of accelerator science and technology"] = "Rev. Accel. Sci. Technol.",
  ["reviews of environment contamination and toxicology"] = "Rev. Environ. Contam. Toxicol.",
  ["reviews of environmental contamination and toxicology"] = "Rev Environ Contam Toxicol",
  ["reviews of geophysics"] = "Rev. Geophys.",
  ["reviews of geophysics (washington, d.c. : 1985)"] = "Rev Geophys",
  ["reviews of geophysics and space physics"] = "Rev. Geophys. Space Phys.",
  ["reviews of human factors and ergonomics"] = "Rev. Hum. Factors Ergon.",
  ["reviews of infectious diseases"] = "Rev Infect Dis",
  ["reviews of modern physics"] = "Rev. Mod. Phys.",
  ["reviews of modern plasma physics"] = "Rev Mod Plasma Phys",
  ["reviews of nonlinear dynamics and complexity"] = "Rev. Nonlinear Dyn. Complex.",
  ["reviews of oculomotor research"] = "Rev Oculomot Res",
  ["reviews of physiology biochemistry and pharmacology"] = "Rev. Physiol., Biochem. Pharmacol.",
  ["reviews of physiology, biochemistry and pharmacology"] = "Rev Physiol Biochem Pharmacol",
  ["reviews of reproduction"] = "Rev Reprod",
  ["reviews on advanced materials science"] = "Rev. Adv. Mater. Sci.",
  ["reviews on drug metabolism and drug interactions"] = "Rev Drug Metab Drug Interact",
  ["reviews on environmental health"] = "Rev Environ Health",
  ["reviews on recent clinical trials"] = "Rev Recent Clin Trials",
  ["revija za sociologiju"] = "Rev Sociol",
  ["revised statutes : under arrangement of the official louisiana revised statutes of 1950. louisiana"] = "Revis Statut La",
  ["revisiones sobre biologia celular"] = "Revis. Biol. Celular",
  ["revisiones sobre biologia celular : rbc"] = "Revis Biol Celular",
  ["revisiones sobre biología celular : rbc"] = "Revis Biol Celular",
  ["revista (cámara de comercio de bogotá)"] = "Rev Camara Coner Bogota",
  ["revista (ordem dos medicos (portugal))"] = "Rev Ordem Med",
  ["revista (ordem dos médicos (portugal))"] = "Rev Ordem Med",
  ["revista abp-apal"] = "Rev ABPAPAL",
  ["revista abp-apal / associacao brasileira de psiquiatria and asociacion psiquiatrica de america latina"] = "Rev ABPAPAL",
  ["revista adicción y ciencia"] = "Rev Adiccion Cienc",
  ["revista adm"] = "Rev. ADM",
  ["revista adm (asociacion dental mexicana : 1988)"] = "Rev ADM",
  ["revista adm (asociación dental mexicana : 1988)"] = "Rev ADM",
  ["revista adm : organo oficial de la asociacion dental mexicana"] = "Rev ADM",
  ["revista adm : organo oficial de la asociación dental mexicana"] = "Rev ADM",
  ["revista agrária acadêmica"] = "Rev Agrar Acad",
  ["revista alergia"] = "Rev. Alerg.",
  ["revista alergia : organo oficial de la sociedad mexicana de alergia e inmunlogia"] = "Rev Alerg",
  ["revista alergia : organo oficial de la sociedad mexicana de alergia e inmunlogía"] = "Rev Alerg",
  ["revista alergia mexico"] = "Rev. Alerg. Mex.",
  ["revista alergia mexico (tecamachalco, puebla, mexico : 1993)"] = "Rev Alerg Mex",
  ["revista alergia méxico"] = "Rev Alerg Mex",
  ["revista alimentar"] = "Rev Aliment",
  ["revista andaluza de medicina del deporte"] = "Rev Andal Med Deport",
  ["revista argentina de alergia"] = "Rev Argent Alerg",
  ["revista argentina de anestesia y analgesia"] = "Rev Argent Anest Analg",
  ["revista argentina de antropologia biologica"] = "Rev. Argent. Antropol. Biol.",
  ["revista argentina de antropología biológica"] = "Rev Argent Antropol Biol",
  ["revista argentina de bronco-esofagología"] = "Rev Argent Broncoesofagol",
  ["revista argentina de cardiologia"] = "Rev Argent Cardiol",
  ["revista argentina de cardiología"] = "Rev Argent Cardiol",
  ["revista argentina de ciencias del comportamiento"] = "Rev Argent Cienc Comport",
  ["revista argentina de cirugia"] = "Rev. Argent. Cir.",
  ["revista argentina de cirugía"] = "Rev Argent Cir",
  ["revista argentina de dermatosifilología"] = "Rev Argent Dermatosifilol",
  ["revista argentina de endocrinologia y metabolismo"] = "Rev. Argent. Endocrinol. Metab.",
  ["revista argentina de endocrinología y metabolismo"] = "Rev Argent Endocrinol Metab",
  ["revista argentina de implantologia estomatologica"] = "Rev Argent Implantol Estomatol",
  ["revista argentina de microbiologia"] = "Rev. Argent. Microbiol.",
  ["revista argentina de microbiología"] = "Rev Argent Microbiol",
  ["revista argentina de neurologia y psiquiatria"] = "Rev Argent Neurol Psiquiatr",
  ["revista argentina de neurología y psiquiatría"] = "Rev Argent Neurol Psiquiatr",
  ["revista argentina de neurología, psiquiatría y neurocirugía"] = "Rev Argent Neurol Psiquiatr Neurocir",
  ["revista argentina de puericultura e neonatología"] = "Rev Argent Pueric Neonatol",
  ["revista argentina de reumatologia"] = "Rev Argent Reumatol",
  ["revista argentina de reumatología"] = "Rev Argent Reumatol",
  ["revista argentina de salud pública"] = "Rev Argent Salud Publica",
  ["revista argentina de tuberculosis y enfermedades pulmonares"] = "Rev Argent Tuberc Enferm Pulm",
  ["revista argentina de urologia"] = "Rev Argent Urol",
  ["revista argentina de urologia y nefrologia"] = "Rev. Argent. Urol. Nefrol.",
  ["revista argentina de urología"] = "Rev Argent Urol",
  ["revista argentina de urología y nefrología"] = "Rev Argent Urol Nefrol",
  ["revista argentina de zoonosis y enfermedades infecciosas emergentes"] = "Rev Argent Zoonosis Enferm Infecc Emerg",
  ["revista arhivelor"] = "Rev Arh",
  ["revista arvore"] = "Rev. Arvore",
  ["revista atlántica-\\-mediterránea de prehistoria y arqueología social"] = "RAtlMed",
  ["revista biodiversidad neotropical"] = "Rev Biodivers Neotrop",
  ["revista boliviana de quimica"] = "Rev. Boliv. Quim.",
  ["revista brasileira de agroecologia"] = "Rev Bras Agroecol",
  ["revista brasileira de análise do comportamento"] = "Rev Bras Anal Comport",
  ["revista brasileira de anestesiologia"] = "Rev Bras Anestesiol",
  ["revista brasileira de aplicacoes de vacuo"] = "Rev. Bras. Apl. Vacuo",
  ["revista brasileira de biociências = brazilian journal of biosciences"] = "Rev Bras Biocienc",
  ["revista brasileira de biologia"] = "Rev Bras Biol",
  ["revista brasileira de botânica : brazilian journal of botany"] = "Rev Bras Bot",
  ["revista brasileira de ciência avi̧cola"] = "Rev Bras Cienc Avic",
  ["revista brasileira de ciência do solo"] = "Rev Bras Cienc Solo",
  ["revista brasileira de cirurgia"] = "Rev Bras Cir",
  ["revista brasileira de cirurgia cardiovascular : orgao oficial da sociedade brasileira de cirurgia cardiovascular"] = "Rev Bras Cir Cardiovasc",
  ["revista brasileira de cirurgia cardiovascular : órgão oficial da sociedade brasileira de cirurgia cardiovascular"] = "Rev Bras Cir Cardiovasc",
  ["revista brasileira de ciência do solo"] = "Rev. Bras. Cienc. Solo",
  ["revista brasileira de crescimento e desenvolvimento humano"] = "Rev Bras Crescimento Desenvolv Hum",
  ["revista brasileira de economia"] = "Revista Brasileira Econ.",
  ["revista brasileira de educação médica"] = "Rev Bras Educ Med",
  ["revista brasileira de enfermagem"] = "Rev Bras Enferm",
  ["revista brasileira de engenharia agricola e ambiental"] = "Rev. Bras. Eng. Agric. Ambiental",
  ["revista brasileira de ensino de quimica"] = "Rev. Bras. Ensino Quim.",
  ["revista brasileira de entomologia"] = "Rev Bras Entomol",
  ["revista brasileira de epidemiologia = brazilian journal of epidemiology"] = "Rev Bras Epidemiol",
  ["revista brasileira de estatistica"] = "Rev Bras Estat",
  ["revista brasileira de estatística"] = "Rev Bras Estat",
  ["revista brasileira de estudos de populacao"] = "Rev Bras Estud Popul",
  ["revista brasileira de estudos de população"] = "Rev Bras Estud Popul",
  ["revista brasileira de farmacognosia"] = "Rev. Bras. Farmacogn.",
  ["revista brasileira de farmacognosia : orgão oficial da sociedade brasileira de farmacognosia"] = "Rev Bras Farmacogn",
  ["revista brasileira de farmácia"] = "Rev Bras Farm",
  ["revista brasileira de fisica"] = "Rev. Bras. Fis.",
  ["revista brasileira de fisiologia vegetal"] = "Braz J Plant Physiol",
  ["revista brasileira de fisioterapia (são carlos (são paulo, brazil))"] = "Rev Bras Fisioter",
  ["revista brasileira de fruticultura"] = "Rev Bras Frutic",
  ["revista brasileira de gastroenterologia"] = "Rev Bras Gastroenterol",
  ["revista brasileira de genética"] = "Rev Bras Genet",
  ["revista brasileira de geofisica"] = "Rev. Bras. Geofis.",
  ["revista brasileira de geografia"] = "Rev Bras Geogr",
  ["revista brasileira de ginecologia e obstetrícia : revista da federação brasileira das sociedades de ginecologia e obstetrícia"] = "Rev Bras Ginecol Obstet",
  ["revista brasileira de hematologia e hemoterapia"] = "Rev Bras Hematol Hemoter",
  ["revista brasileira de historia da matematica. an international journal on the history of mathematics"] = "Rev. Bras. Hist. Mat.",
  ["revista brasileira de história"] = "Rev Bras Hist",
  ["revista brasileira de horticultura ornamental"] = "Rev Bras Hortic Ornam",
  ["revista brasileira de leprologia"] = "Rev Bras Leprol",
  ["revista brasileira de malariologia e doencas tropicais"] = "Rev. Bras. Malariol. Doencas Trop.",
  ["revista brasileira de malariologia e doencas tropicais. publicacoes avulsas"] = "Rev Bras Malariol Doencas Trop",
  ["revista brasileira de malariologia e doenças tropicais. publicações avulsas"] = "Rev Bras Malariol Doencas Trop",
  ["revista brasileira de medicina"] = "Rev Bras Med",
  ["revista brasileira de medicina do trabalho : publicação oficial da associação nacional de medicina do trabalho-anamt"] = "Rev Bras Med Trab",
  ["revista brasileira de medicina veterinaria"] = "Rev Bras Med Vet",
  ["revista brasileira de milho e sorgo"] = "Rev Bras Milho Sorgo",
  ["revista brasileira de oceanografia"] = "Rev. Bras. Oceangr.",
  ["revista brasileira de odontologia"] = "Rev Bras Odontol",
  ["revista brasileira de odontologia militar"] = "Rev. Bras. Odontol. Mil.",
  ["revista brasileira de odontologia militar : orgao oficial da academia brasileira de odontologia militar"] = "Rev Bras Odontol Mil",
  ["revista brasileira de odontologia militar : órgão oficial da academia brasileira de odontologia militar"] = "Rev Bras Odontol Mil",
  ["revista brasileira de oftalmologia"] = "Rev Bras Oftalmol",
  ["revista brasileira de ornitologia"] = "Rev Bras Ornitol",
  ["revista brasileira de ortopedia"] = "Rev Bras Ortop",
  ["revista brasileira de oto-rino-laringologia"] = "Rev Bras Otorrinolaringol",
  ["revista brasileira de otorrinolaringologia"] = "Rev Bras Otorinolaringol",
  ["revista brasileira de otorrinolaringologia (english ed.)"] = "Rev Bras Otorrinolaringol (Engl Ed)",
  ["revista brasileira de paleontologia"] = "Rev. Bras. Paleontol.",
  ["revista brasileira de parasitologia veterinária = brazilian journal of veterinary parasitology : órgão oficial do colégio brasileiro de parasitologia veterinária"] = "Rev Bras Parasitol Vet",
  ["revista brasileira de pesquisas medicas e biologicas"] = "Rev. Bras. Pesqui. Med. Biol.",
  ["revista brasileira de pesquisas médicas e biológicas"] = "Rev Bras Pesqui Med Biol",
  ["revista brasileira de psiquiatria (sao paulo, brazil : 1999)"] = "Rev Bras Psiquiatr",
  ["revista brasileira de psiquiatria (são paulo, brazil : 1999)"] = "Braz J Psychiatry",
  ["revista brasileira de reumatologia"] = "Rev Bras Reumatol",
  ["revista brasileira de sementes = brazilian seed journal"] = "Rev Bras Sementes",
  ["revista brasileira de terapia intensiva"] = "Rev Bras Ter Intensiva",
  ["revista brasileira de tuberculose e doencas toracicas"] = "Rev Bras Tuberc Doencas Torac",
  ["revista brasileira de tuberculose e doenças torácicas"] = "Rev Bras Tuberc Doencas Torac",
  ["revista brasileira de zoologia"] = "Rev Bras Zool",
  ["revista catalana de micologia"] = "Rev Catalana Micol",
  ["revista catarinense de odontologia"] = "Rev Catarinense Odontol",
  ["revista centro america odontologica"] = "Rev. Cent. Am. Odontol.",
  ["revista centro américa odontológica"] = "Rev Cent Am Odontol",
  ["revista centroamericana de economia"] = "Rev Centroam Econ",
  ["revista centroamericana de economía"] = "Rev Centroam Econ",
  ["revista cepae"] = "Rev CEPAE",
  ["revista chapingo. serie: horticultura"] = "Rev Chapingo Ser Hortic",
  ["revista chilena de entomología : publicación de la faculdad de filosofía y educación, universidad de chile, y de la sociedad chilena de entomología"] = "Rev Chil Entomol",
  ["revista chilena de higiene y medicina preventiva"] = "Rev Chil Hig Med Prev",
  ["revista chilena de historia natural"] = "Rev. Chil. Hist. Nat.",
  ["revista chilena de historia natural (valparaíso, chile : 1983)"] = "Rev Chil Hist Nat",
  ["revista chilena de historia y geografia"] = "Rev Chil Hist Geogr",
  ["revista chilena de historia y geografía"] = "Rev Chil Hist Geogr",
  ["revista chilena de infectologia : organo oficial de la sociedad chilena de infectologia"] = "Rev Chilena Infectol",
  ["revista chilena de infectología : órgano oficial de la sociedad chilena de infectología"] = "Rev Chilena Infectol",
  ["revista chilena de neuro-psiquiatría"] = "Rev Chil Neuropsiquiatr",
  ["revista chilena de nutricion : organo oficial de la sociedad chilena de nutricion, bromatologia y toxicologia"] = "Rev Chil Nutr",
  ["revista chilena de nutrición : órgano oficial de la sociedad chilena de nutrición, bromatología y toxicología"] = "Rev Chil Nutr",
  ["revista chilena de obstetricia y ginecologia"] = "Rev. Chil. Obstet. Ginecol.",
  ["revista chilena de obstetricia y ginecología"] = "Rev Chil Obstet Ginecol",
  ["revista chilena de odontoestomatologia"] = "Rev. Chil. Odontoestomatol.",
  ["revista chilena de odontoestomatología"] = "Rev Chil Odontoestomatol",
  ["revista chilena de pediatria"] = "Rev. Chil. Pediatr.",
  ["revista chilena de pediatría"] = "Rev Chil Pediatr",
  ["revista chilena de urología"] = "Rev Chil Urol",
  ["revista ciencias matemáticas"] = "Cienc. Mat. (Havana)",
  ["revista científica (universidad del zulia. facultad de ciencias veterinarias. división de investigación)"] = "Rev Cient (Maracaibo)",
  ["revista científica rural"] = "Rev Cient Rural",
  ["revista clinica de sao paulo"] = "Rev Clin Sao Paulo",
  ["revista clinica de são paulo"] = "Rev Clin Sao Paulo",
  ["revista clinica espanola"] = "Rev Clin Esp",
  ["revista clinica, instituto maternal, lisbon"] = "Rev. Clin. Inst. Matern. Lisb.",
  ["revista clinica. lisbon. instituto maternal"] = "Rev Clin Inst Matern Lisb",
  ["revista clínica espanõla"] = "Rev Clin Esp (Barc)",
  ["revista clínica española"] = "Rev Clin Esp",
  ["revista clínica. lisbon. instituto maternal"] = "Rev Clin Inst Matern Lisb",
  ["revista colombiana de biotecnología"] = "Rev Colomb Biotecnol",
  ["revista colombiana de entomologia"] = "Rev. Colomb. Entomol.",
  ["revista colombiana de entomología"] = "Rev Colomb Entomol",
  ["revista colombiana de estadistica"] = "Rev. Colombiana Estadist.",
  ["revista colombiana de estadística"] = "Rev Colomb Estad",
  ["revista colombiana de gastroenterología"] = "Rev Colomb Gastroenterol",
  ["revista colombiana de matematicas"] = "Rev. Colombiana Mat.",
  ["revista colombiana de matemáticas"] = "Rev. Colombiana Mat.",
  ["revista colombiana de obstetricia y ginecologia"] = "Rev. Colomb. Obstet. Ginecol.",
  ["revista colombiana de obstetricia y ginecología"] = "Rev Colomb Obstet Ginecol",
  ["revista colombiana de pediatria y puericultura"] = "Rev. Colomb. Pediatr. Pueric.",
  ["revista colombiana de pediatría y puericultura"] = "Rev Colomb Pediatr Pueric",
  ["revista colombiana de psiquiatría"] = "Rev Colomb Psiquiatr",
  ["revista colombiana de quimica"] = "Rev. Colomb. Quim.",
  ["revista colombiana estadística"] = "Rev. Colombiana Estadíst.",
  ["revista complutense de historia de américa"] = "Rev Complut Hist Am",
  ["revista costarricense de ciencias medicas"] = "Rev. Costarric. Cienc. Med.",
  ["revista costarricense de ciencias médicas"] = "Rev Costarric Cienc Med",
  ["revista crítica de ciencias sociais"] = "Rev Crit Cienc Sociais",
  ["revista cubana de administracion de salud"] = "Rev Cuhana Adm Salud",
  ["revista cubana de administración de salud"] = "Rev Cuhana Adm Salud",
  ["revista cubana de cardiologia"] = "Rev Cubana Cardiol",
  ["revista cubana de cardiología"] = "Rev Cubana Cardiol",
  ["revista cubana de enfermeria"] = "Rev. Cubana Enferm.",
  ["revista cubana de enfermería"] = "Rev Cubana Enferm",
  ["revista cubana de estomatologia"] = "Rev. Cubana Estomatol.",
  ["revista cubana de estomatología"] = "Rev Cubana Estomatol",
  ["revista cubana de higiene y epidemiologia"] = "Rev Cubana Hig Epidemiol",
  ["revista cubana de higiene y epidemiología"] = "Rev Cubana Hig Epidemiol",
  ["revista cubana de laboratorio clínico"] = "Rev Cubana Lab Clin",
  ["revista cubana de medicina"] = "Rev Cubana Med",
  ["revista cubana de medicina soviética"] = "Rev Cuba Med Sov",
  ["revista cubana de medicina tropical"] = "Rev Cubana Med Trop",
  ["revista cubana de pediatria"] = "Rev Cubana Pediatr",
  ["revista cubana de pediatría"] = "Rev Cubana Pediatr",
  ["revista cubana de salud pública"] = "Rev Cub Salud Publica",
  ["revista cubana de tuberculosis"] = "Rev Cuba Tuberc",
  ["revista da associacao medica brasileira"] = "Rev Assoc Med Bras",
  ["revista da associacao medica brasileira (1992)"] = "Rev Assoc Med Bras",
  ["revista da associacao medica de minas gerais"] = "Rev Assoc Med Minas Gerais",
  ["revista da associacao medica do rio grande do sul"] = "Rev Assoc Med Rio Grande Do Sul",
  ["revista da associacao paulista de cirurgioes dentistas"] = "Rev. Assoc. Paul. Cir. Dent.",
  ["revista da associacao paulista de cirurgioes dentistas regional de aracatuba"] = "Rev Assoc Paul Cir Dent Reg Aracatuba",
  ["revista da associação médica brasileira"] = "Rev Assoc Med Bras",
  ["revista da associação médica brasileira (1992)"] = "Rev Assoc Med Bras (1992)",
  ["revista da associação médica de minas gerais"] = "Rev Assoc Med Minas Gerais",
  ["revista da associação médica do rio grande do sul"] = "Rev Assoc Med Rio Grande Do Sul",
  ["revista da associação paulista de cirurgiões dentistas"] = "Rev Assoc Paul Cir Dent",
  ["revista da associação paulista de cirurgiões dentistas regional de araçatuba"] = "Rev Assoc Paul Cir Dent Reg Aracatuba",
  ["revista da escola de enfermagem da u s p"] = "Rev Esc Enferm USP",
  ["revista da escola de enfermagem da usp"] = "Rev. Esc. Enferm. USP",
  ["revista da faculdade de farmacia e odontologia de araraquara"] = "Rev. Fac. Farm. Odontol. Araraquara",
  ["revista da faculdade de farmacia e odontologia de ribeirao preto"] = "Rev. Fac. Farm. Odontol. Ribeiro. Preto",
  ["revista da faculdade de farmácia e odontologia de araraquara"] = "Rev Fac Farm Odontol Araraquara",
  ["revista da faculdade de farmácia e odontologia de ribeirão prêto"] = "Rev Fac Farm Odontol Ribeiro Preto",
  ["revista da faculdade de medicina veterinária"] = "Rev Fac Med Vet Univ Sao Paulo",
  ["revista da faculdade de odontologia da fzl"] = "Rev Faculdade Odontol FZL",
  ["revista da faculdade de odontologia da universidade de sao paulo"] = "Rev. Fac. Odontol. Sao Paulo",
  ["revista da faculdade de odontologia da universidade de são paulo"] = "Rev Fac Odontol Sao Paulo",
  ["revista da faculdade de odontologia da universidade federal da bahia"] = "Rev Fac Odontol Univ Fed Bahia",
  ["revista da faculdade de odontologia de aracatuba"] = "Rev. Fac. Odontol. Aracatuba",
  ["revista da faculdade de odontologia de araçatuba"] = "Rev Fac Odontol Aracatuba",
  ["revista da faculdade de odontologia de lins"] = "Rev Faculdade Odontol Lins",
  ["revista da faculdade de odontologia de pernambuco"] = "Rev Fac Odontol Pernambuco",
  ["revista da faculdade de odontologia de ribeirao preto"] = "Rev. Fac. Odontol. Ribeiro Preto",
  ["revista da faculdade de odontologia de ribeirão prêto"] = "Rev Fac Odontol Ribeiro Preto",
  ["revista da faculdade de odontologia de sao jose dos campos"] = "Rev. Fac. Odontol. Sao Jose dos Campos",
  ["revista da faculdade de odontologia de são josé dos campos"] = "Rev Fac Odontol Sao Jose Dos Campos",
  ["revista da faculdade de odontologia, porto alegre"] = "Rev. Fac. Odontol. P. Alegre",
  ["revista da faculdade de odontologia, pôrto alegre"] = "Rev Fac Odontol P Alegre",
  ["revista da facultade de letras. serie de historia. universidade do porto"] = "RPorto",
  ["revista da flora medicinal"] = "Rev Flora Med",
  ["revista da sociedade brasileira de historia da ciencia"] = "Rev Soc Bras Hist Cienc",
  ["revista da sociedade brasileira de historia da ciência"] = "Rev Soc Bras Hist Cienc",
  ["revista da sociedade brasileira de medicina tropical"] = "Rev Soc Bras Med Trop",
  ["revista de actualidad estomatologica espanola"] = "Rev. Actual. Estomatol. Esp.",
  ["revista de actualidad estomatologica espanola / ilustre consejo general de colegios de odontologos y estomatologos de espana"] = "Rev Actual Estomatol Esp",
  ["revista de actualidad estomatológica española"] = "Rev Actual Estomatol Esp",
  ["revista de actualidad odontoestomatologica espanola"] = "Rev. Actual. Odontoestomatol. Esp.",
  ["revista de actualidad odontoestomatologica espanola : boletin de informacion, del ilustre consejo general de colegios de odontologos y estomatologos de espana"] = "Rev Actual Odontoestomatol Esp",
  ["revista de actualidad odontoestomatológica española : boletín de información, del ilustre consejo general de colegios de odontólogos y estomatólogos de españa"] = "Rev Actual Odontoestomatol Esp",
  ["revista de administracao municipal"] = "Rev Adm Munic",
  ["revista de administração municipal"] = "Rev Adm Munic",
  ["revista de administraçao d emprêsas"] = "Revista Admin. Emprêsas",
  ["revista de analisis economico"] = "Revista Anal. Econ.",
  ["revista de antropología social"] = "Rev Antopol Soc",
  ["revista de archivos, bibliotecas y museos"] = "RArchBiblMus",
  ["revista de archivos, bibliotecas y museos (madrid, spain : 1897)"] = "Rev Arch Bibl Mus",
  ["revista de arqueología"] = "RAMadrid",
  ["revista de bioética y derecho"] = "Rev Bioet Derecho",
  ["revista de biologia oral"] = "Rev Biol Oral",
  ["revista de biologia tropical"] = "Rev Biol Trop",
  ["revista de biología marina y oceanografía"] = "Rev Biol Mar Oceanogr",
  ["revista de biología tropical"] = "Rev Biol Trop",
  ["revista de calidad asistencial : organo de la sociedad española de calidad asistencial"] = "Rev Calid Asist",
  ["revista de chimie"] = "Rev. Chim.",
  ["revista de chimie (bucharest, romania)"] = "Rev. Chim. (Bucharest, Rom.)",
  ["revista de chirurgie, oncologie, radiologie, o. r. l., oftalmologie, stomatologie. oftalmologie"] = "Rev Chir Oncol Radiol O R L Oftalmol Stomatol Ser Oftalmol",
  ["revista de chirurgie, oncologie, radiologie, o. r. l., oftalmologie, stomatologie. oncologia"] = "Rev Chir Oncol Radiol ORL Oftalmol Stomatol Oncol",
  ["revista de chirurgie, oncologie, radiologie, o. r. l., oftalmologie, stomatologie. seria: oftalmologie"] = "Rev Chir Oncol Radiol O R L Oftalmol Stomatol Ser Oftalmol",
  ["revista de chirurgie, oncologie, radiologie, o. r. l., oftalmologie, stomatologie. seria: stomatologie"] = "Rev Chir Oncol Radiol O R L Oftalmol Stomatol Ser Stomatol",
  ["revista de chirurgie, oncologie, radiologie, o.r.l., oftalmologie, stomatologie. chirurgie"] = "Rev Chir Oncol Radiol O R L Oftalmol Stomatol Chir",
  ["revista de chirurgie, oncologie, radiologie, o.r.l., oftalmologie, stomatologie. oto-rino-laringologia"] = "Rev Chir Oncol Radiol O R L Oftalmol Stomatol Otorinolaringol",
  ["revista de chirurgie, oncologie, radiologie, o.r.l., oftalmologie, stomatologie. seria oftalmologie"] = "Rev. Chir. Oncol. Radiol. O. R. L. Oftalmol. Stomatol. Ser. Oftalmol.",
  ["revista de chirurgie, oncologie, radiologie, o.r.l., oftalmologie, stomatologie. seria stomatologie"] = "Rev. Chir. Oncol. Radiol. O. R. L. Oftalmol. Stomatol. Ser. Stomatol.",
  ["revista de ciencias biomedicas"] = "Rev Cinc Biomed",
  ["revista de ciencias sociales"] = "Rev Cienc Soc",
  ["revista de ciências biomédicas"] = "Rev Cinc Biomed",
  ["revista de cirugia"] = "Rev Cir (Mex)",
  ["revista de cirugia del uruguay"] = "Rev. Cir. Urug.",
  ["revista de cirugía"] = "Rev Cir (Mex)",
  ["revista de cirugía del uruguay"] = "Rev Cir Urug",
  ["revista de cirurgia de são paulo"] = "Rev Cir Sao Paulo",
  ["revista de criminología y policía científica"] = "Rev Criminol Polic Cient",
  ["revista de demografía histórica"] = "Rev Demogr Hist",
  ["revista de derecho y genoma humano = law and the human genome review"] = "Rev Derecho Genoma Hum",
  ["revista de dialectologia y tradiciones populares / consejo superior de investigaciones cientificas, instituto antonio de nebrija, seccion de tradiciones populares"] = "Rev Dialoctol Tradiciones Pop (Madrid)",
  ["revista de dialectología y tradiciones populares"] = "Rev Dialoctol Tradiciones Pop (Madrid)",
  ["revista de divulgación técnica agrícola y agroindustrial"] = "Rev Divulg Tec Agric Agroind",
  ["revista de econometria"] = "Revista Econometria",
  ["revista de economia aplicada"] = "Revista Econ. Aplicada",
  ["revista de economia política"] = "Revista Econ. Política",
  ["revista de economía"] = "Revista Econ.",
  ["revista de economía institucional"] = "Revista Econ. Inst.",
  ["revista de economía y estadística, n.s."] = "Revista Econ. y Estadíst.",
  ["revista de enfermagem"] = "Rev Enferm (Lisboa)",
  ["revista de enfermeria"] = "Rev. Enferm.",
  ["revista de enfermeria (barcelona, spain)"] = "Rev Enferm",
  ["revista de enfermería (barcelona, spain)"] = "Rev Enferm",
  ["revista de estadística"] = "Revista Estadíst.",
  ["revista de estudios clásicos"] = "REC",
  ["revista de estudios de la vida local"] = "Rev Estud Vida Local",
  ["revista de estudios extremenos"] = "Rev Estud Extremenos",
  ["revista de estudios extremeños"] = "Rev Estud Extremenos",
  ["revista de estudios ibéricos"] = "REstIber",
  ["revista de estudios políticos"] = "Revista Estud. Polít.",
  ["revista de farmacia e bioquimica da universidade de sao paulo"] = "Rev. Farm. Bioquim. Univ. Sao Paulo",
  ["revista de farmacia e odontologia"] = "Rev Farm Odontol",
  ["revista de farmacología y medicina experimental"] = "Rev Farmacol Med Exp",
  ["revista de farmácia e bioquímica da universidade de são paulo"] = "Rev Farm Bioquim Univ Sao Paulo",
  ["revista de filosofía"] = "RF(Argentina)",
  ["revista de fomento social"] = "Rev Fom Soc",
  ["revista de gastroenterologia de mexico"] = "Rev Gastroenterol Mex",
  ["revista de gastroenterologia del peru"] = "Rev. Gastroenterol. Peru",
  ["revista de gastroenterologia del peru : organo oficial de la sociedad de gastroenterologia del peru"] = "Rev Gastroenterol Peru",
  ["revista de gastroenterología de méxico"] = "Rev Gastroenterol Mex",
  ["revista de gastroenterología del perú : órgano oficial de la sociedad de gastroenterología del perú"] = "Rev Gastroenterol Peru",
  ["revista de geofisica"] = "Rev. Geofis.",
  ["revista de ginecologia e d'obstetrícia"] = "Rev Ginecol Obstet (Sao Paulo)",
  ["revista de ginecologia e d’obstetricia"] = "Rev. Ginecol. Obstet. (Sao Paulo)",
  ["revista de ginecologia y obstetricia"] = "Rev Gynecol Obstet",
  ["revista de guimarães"] = "RGuimar",
  ["revista de historia"] = "Rev Hist (Arg)",
  ["revista de historia americana y argentina"] = "Rev Hist Am Argent",
  ["revista de historia canaria"] = "Rev Hist Canaria",
  ["revista de historia de la psicologia"] = "Rev Hist Psicol",
  ["revista de historia de la psicología"] = "Rev Hist Psicol",
  ["revista de historia económica"] = "Revista Hist. Econ.",
  ["revista de historia industrial"] = "Revista Hist. Ind.",
  ["revista de historia jeronimo zurita"] = "Rev Hist Jeronimo Zurita",
  ["revista de historia jerónimo zurita"] = "Rev Hist Jeronimo Zurita",
  ["revista de historia militar"] = "Rev Hist Mil",
  ["revista de historia naval"] = "Rev Hist Naval",
  ["revista de historia naval / instituto de historia y cultura naval, armada espanola"] = "Rev Hist Naval",
  ["revista de história"] = "Rev Hist (Brazil)",
  ["revista de igiena, bacteriologie, virusologie, parazitologie, epidemiologie, pneumoftiziologie. bacteriologia, virusologia, parazitologia, epidemiologia"] = "Rev. Ig. Bacteriol. Virusol. Parazitol. Epidemiol. Pneumoftiziol. Bacteriol. Virusol. Parazitol. Epi",
  ["revista de igiena, bacteriologie, virusologie, parazitologie, epidemiologie, pneumoftiziologie. pneumoftiziologia"] = "Rev. Ig. Bacteriol. Virusol. Parazitol. Epidemiol. Pneumoftiziol. Pneumoftiziol.",
  ["revista de igiena, medicina muncii, medicina sociala, bacteriologie, virusologie, parazitologie, epidemiologie, pneumoftiziologie. pneumoftiziologie / uniunea societatilor de stiinte medicale din romania"] = "Rev Ig Med Muncii Med Soc Bacteriol Virusol Parazitol Epidemiol Pneumoftiziol Pneumoftiziol",
  ["revista de igienă, bacteriologie, virusologie, parazitologie, epidemiologie, pneumoftiziologie. bacteriologia, virusologia, parazitologia, epidemiologia"] = "Rev Ig Bacteriol Virusol Parazitol Epidemiol Pneumoftiziol Bacteriol Virusol Parazitol Epidemiol",
  ["revista de igienă, bacteriologie, virusologie, parazitologie, epidemiologie, pneumoftiziologie. pneumoftiziologia"] = "Rev Ig Bacteriol Virusol Parazitol Epidemiol Pneumoftiziol Pneumoftiziol",
  ["revista de igienă, medicina muncii, medicină socială, bacteriologie, virusologie, parazitologie, epidemiologie, pneumoftiziologie. pneumoftiziologie"] = "Rev Ig Med Muncii Med Soc Bacteriol Virusol Parazitol Epidemiol Pneumoftiziol Pneumoftiziol",
  ["revista de indias"] = "Rev Indias",
  ["revista de informacao legislativa"] = "Rev Inf Legis",
  ["revista de informacion medico-terapeutica"] = "Rev Inform Med Ter",
  ["revista de información médico-terapéutica"] = "Rev Inform Med Ter",
  ["revista de informação legislativa"] = "Rev Inf Legis",
  ["revista de instituciones europeas"] = "Rev Inst Eur",
  ["revista de investigacion"] = "Rev. Invest. (Guadalajara)",
  ["revista de investigacion clinica"] = "Rev. Invest. Clin.",
  ["revista de investigacion clinica; organo del hospital de enfermedades de la nutricion"] = "Rev Invest Clin",
  ["revista de investigacion en salud publica"] = "Rev. Invest. Salud Publica",
  ["revista de investigaciones marinas"] = "Rev Investig Mar",
  ["revista de investigaciones veterinarias del perú (online)"] = "Rev Investig Vet Peru",
  ["revista de investigación"] = "Rev Invest (Guadalajara)",
  ["revista de investigación clínica; organo del hospital de enfermedades de la nutrición"] = "Rev Invest Clin",
  ["revista de investigación en salud pública"] = "Rev Invest Salud Publica",
  ["revista de istorie"] = "Rev Istor",
  ["revista de la academia canaria de ciencias"] = "Rev. Acad. Canaria Cienc.",
  ["revista de la academia colombiana de ciencias exactas, fisicas y naturales"] = "Rev Acad Colomb Cienc Exactas Fis Nat",
  ["revista de la academia de ciencias exactas, físicas químicas y naturales de zaragoza"] = "Rev. Acad. Cienc. Zaragoza (2)",
  ["revista de la administración nacional del agua. argentina. administración nacional del agua"] = "Rev Adm Nac Agua",
  ["revista de la agrupacion odontologica de la capital federal"] = "Rev. Agrup. Odontol. Cap. Fed.",
  ["revista de la agrupación odontológica de la capital federal"] = "Rev Agrup Odontol Cap Fed",
  ["revista de la asociacion argentina de microbiologia"] = "Rev. Asoc. Argent. Microbiol.",
  ["revista de la asociacion de profesionales del hospital nacional de odontologia"] = "Rev Asoc Prof Hosp Nac Odontol",
  ["revista de la asociacion geologica argentina"] = "Rev. Asoc. Geol. Argent.",
  ["revista de la asociacion medica argentina"] = "Rev. Asoc. Med. Argent.",
  ["revista de la asociacion mexicana de enfermeras"] = "Rev. Asoc. Mex. Enferm.",
  ["revista de la asociacion odontologica argentina"] = "Rev. Asoc. Odontol. Argent.",
  ["revista de la asociacion odontologica de costa rica"] = "Rev. Asoc. Odontol. Costa Rica",
  ["revista de la asociación argentina de microbiología"] = "Rev Asoc Argent Microbiol",
  ["revista de la asociación bioquímica argentina. asociación bioquímica argentina"] = "Rev Asoc Bioquim Argent",
  ["revista de la asociación mexicana de enfermeras"] = "Rev Asoc Mex Enferm",
  ["revista de la asociación médica argentina"] = "Rev Asoc Med Argent",
  ["revista de la asociación odontológica argentina"] = "Rev Asoc Odontol Argent",
  ["revista de la asociación odontológica de costa rica"] = "Rev Asoc Odontol Costa Rica",
  ["revista de la biblioteca nacional jose marti. biblioteca nacional jose marti"] = "Rev Bibl Nac Jose Marti",
  ["revista de la biblioteca nacional josé martí. biblioteca nacional josé martí"] = "Rev Bibl Nac Jose Marti",
  ["revista de la construccion"] = "Rev. Constr.",
  ["revista de la escuela de odontologia, universidad nacional de tucuman, facultad de medicina"] = "Rev Esc Odontol Tucuman",
  ["revista de la facultad de ciencas económicas y de administración de  montevideo"] = "Revista Fac. Cienc. Econ. Admin. Montevideo",
  ["revista de la facultad de ciencias agrarias"] = "Rev Fac Cienc Agrar",
  ["revista de la facultad de ciencias medicas"] = "Rev Fac Cienc Med",
  ["revista de la facultad de ciencias medicas (cordoba, argentina)"] = "Rev Fac Cien Med Univ Nac Cordoba",
  ["revista de la facultad de ciencias medicas / universidad nacional de cordoba"] = "Rev Fac Cien Med Univ Nac Cordoba",
  ["revista de la facultad de ciencias medicas de cordoba"] = "Rev. Fac. Cienc. Med. Cordoba",
  ["revista de la facultad de ciencias medicas de la universidad nacional de cordoba"] = "Rev. Fac. Cien. Med. Univ. Nac. Cordoba",
  ["revista de la facultad de ciencias medicas, universidad nacional de cordoba"] = "Rev. Fac. Cien. Med. Univ. Nac. Cordoba",
  ["revista de la facultad de ciencias médicas"] = "Rev Fac Cienc Med",
  ["revista de la facultad de ciencias médicas (córdoba, argentina)"] = "Rev Fac Cien Med Univ Nac Cordoba",
  ["revista de la facultad de ciencias médicas de córdoba"] = "Rev Fac Cienc Med Cordoba",
  ["revista de la facultad de ciencias médicas de la universidad nac. de córdoba"] = "Rev Fac Cien Med Univ Nac Cordoba",
  ["revista de la facultad de ciencias médicas, universidad nacional de córdoba"] = "Rev Fac Cienc Med Univ Nac Cordoba",
  ["revista de la facultad de ciencias químicas (química y farmacia). la plata. universidad nacional. facultad de química y farmacia"] = "Rev Fac Cienc Quim Quim Farm",
  ["revista de la facultad de farmacia, universidad central de venezuela"] = "Rev Fac Farm Univ Cent Venez",
  ["revista de la facultad de medicina veterinaria de la universidad nacional mayor de san marcos"] = "Rev Fac Med Vet (Lima)",
  ["revista de la facultad de medicina, universidad nacional autonoma de mexico"] = "Rev Fac Med Univ Nac Auton Mex",
  ["revista de la facultad de medicina, universidad nacional autónoma de méxico"] = "Rev Fac Med Univ Nac Auton Mex",
  ["revista de la facultad de medicina, universidad nacional de colombia"] = "Rev Fac Med Univ Nac Colomb",
  ["revista de la facultad de odontologia"] = "Rev. Fac. Odontol. Univ. Nac. (Cordoba)",
  ["revista de la facultad de odontologia (san miguel de tucuman, argentina)"] = "Rev Fac Odontol Tucuman",
  ["revista de la facultad de odontologia (santiago, chile)"] = "Rev Fac Odontol Univ Chile",
  ["revista de la facultad de odontologia de la universidad nacional de colombia"] = "Rev. Fac. Odontol. Univ. Nac. Colomb.",
  ["revista de la facultad de odontologia universidad de antioquia"] = "Rev Fac Odontol Univ Antioq",
  ["revista de la facultad de odontologia, universidad de antioquia"] = "Rev. Fac. Odontol. Univ. Antioq.",
  ["revista de la facultad de odontologia, universidad nacional de tucuman"] = "Rev. Fac. Odontol. Tucuman",
  ["revista de la facultad de odontología"] = "Rev Fac Odontol Univ Nac (Cordoba)",
  ["revista de la facultad de odontología (san miguel de tucumán, argentina)"] = "Rev Fac Odontol Tucuman",
  ["revista de la facultad de odontología (santiago, chile)"] = "Rev Fac Odontol Univ Chile",
  ["revista de la facultad de odontología de la universidad nacional de colombia"] = "Rev Fac Odontol Univ Nac Colomb",
  ["revista de la facultad de odontología universidad de antioquia"] = "Rev Fac Odontol Univ Antioq",
  ["revista de la federacion odontologica colombiana"] = "Rev. Fed. Odontol. Colomb.",
  ["revista de la federacion odontologica ecuatoriana"] = "Rev. Fed. Odontol. Ecuat.",
  ["revista de la federación odontológica colombiana"] = "Rev Fed Odontol Colomb",
  ["revista de la federación odontológica ecuatoriana"] = "Rev Fed Odontol Ecuat",
  ["revista de la oim sobre migraciones en américa latina"] = "Rev OIM Sobre Migr Am Lat",
  ["revista de la real academia de ciencias exactas, fisicas y naturales (espana)"] = "Rev. R. Acad. Cienc. Exactas Fis. Nat. (Esp.)",
  ["revista de la real academia de ciencias exactas, fisicas y naturales. serie a. matematicas"] = "Rev. R. Acad. Cienc. Exactas, Fis. Nat. Ser. A Mat.",
  ["revista de la real academia de ciencias exactas, fisicas y naturales. serie a. matematicas. racsam"] = "Rev. R. Acad. Cienc. Exactas Fis. Nat. Ser. A Mat. RACSAM",
  ["revista de la real academia de ciencias exactas, fisicas, quimicas y naturales de zaragoza. serie 2"] = "Rev. R. Acad. Cienc. Exactas Fis. Quim. Nat. Zaragoza (2)",
  ["revista de la real academia de ciencias exactas, físicas y naturales. serie a, matemáticas"] = "Rev R Acad Cienc Exactas Fis Nat A Mat",
  ["revista de la real academia de ciencias exactas, físicas y naturales (españa)"] = "Rev. R. Acad. Cienc. Exactas Fís. Nat. (Esp.)",
  ["revista de la sanidad de policia"] = "Rev Sanid Polic",
  ["revista de la sanidad de policía"] = "Rev Sanid Polic",
  ["revista de la sanidad militar"] = "Rev Sanid Mil",
  ["revista de la sanidad militar argentina"] = "Rev Sanid Milit Argent",
  ["revista de la sanidad militar del peru"] = "Rev Sanid Mil Peru",
  ["revista de la sanidad militar del perú"] = "Rev Sanid Mil Peru",
  ["revista de la sociedad argentina de biologia"] = "Rev. Soc. Argent. Biol.",
  ["revista de la sociedad argentina de biología"] = "Rev Soc Argent Biol",
  ["revista de la sociedad bolivariana de venezuela. sociedad bolivariana de venezuela"] = "Rev Soc Bolivar Venez",
  ["revista de la sociedad colombiana de endocrinologia"] = "Rev Soc Colomb Endocrinol",
  ["revista de la sociedad colombiana de endocrinología"] = "Rev Soc Colomb Endocrinol",
  ["revista de la sociedad colombiana de ortodoncia"] = "Rev. Soc. Colomb. Ortod.",
  ["revista de la sociedad colombiana de pediatria y puericultura"] = "Rev Soc Colomb Pediatr Pueric",
  ["revista de la sociedad colombiana de pediatría y puericultura"] = "Rev Soc Colomb Pediatr Pueric",
  ["revista de la sociedad de ginecologia y obstetricia"] = "Rev Soc Ginecol Obstet",
  ["revista de la sociedad de pediatría del litoral"] = "Rev Soc Pediatr Litoral",
  ["revista de la sociedad entomologica argentina"] = "Rev. Soc. Entomol. Argent.",
  ["revista de la sociedad entomológica argentina"] = "Rev Soc Entomol Argent",
  ["revista de la sociedad mexicana de historia natural"] = "Rev Soc Mex Hist Nat",
  ["revista de la sociedad médica de trujillo"] = "Rev Soc Med Trujillo",
  ["revista de la sociedad médico-quirúrgica del zulia"] = "Rev Soc Med  Quir Zulia",
  ["revista de la sociedad odontologica de la plata"] = "Rev. Soc. Odontol. La. Plata",
  ["revista de la sociedad odontológica de la plata"] = "Rev Soc Odontol La Plata",
  ["revista de la sociedad peruana de endocrinologia"] = "Rev. Soc. Peru. Endocrinol.",
  ["revista de la sociedad peruana de endocrinología"] = "Rev Soc Peru Endocrinol",
  ["revista de la sociedad quimica de mexico"] = "Rev. Soc. Quim. Mex.",
  ["revista de la sociedad quimica del peru"] = "Rev. Soc. Quim. Peru",
  ["revista de la sociedad venezolana de historia de la medicina"] = "Rev Soc Venez Hist Med",
  ["revista de la union matematica argentina"] = "Rev. Union Mat. Argent.",
  ["revista de la universidad nacional de cordoba. universidad nacional de cordoba"] = "Rev Univ Nac Cordoba",
  ["revista de la universidad nacional de cordoba. universidad nacional de córdoba"] = "Rev Univ Nac Cordoba",
  ["revista de la unión matemática argentina"] = "Rev. Un. Mat. Argentina",
  ["revista de letras"] = "RL",
  ["revista de marina (valparaiso, chile)"] = "Rev Mar",
  ["revista de marina (valparaíso, chile)"] = "Rev Mar",
  ["revista de matematica. teoria y aplicaciones"] = "Rev. Mat. Teor. Apl.",
  ["revista de matemática de la universidad del zulia"] = "Divulg. Mat.",
  ["revista de matemática e estatística"] = "Rev. Mat. Estatíst.",
  ["revista de matemáticas aplicadas"] = "Rev. Mat. Apl.",
  ["revista de medicina"] = "Rev Med (Rio J)",
  ["revista de medicina de la universidad de navarra"] = "Rev Med Univ Navarra",
  ["revista de medicina do rio grande do sul"] = "Rev Med Rio Gd Sul",
  ["revista de medicina e cirurgia de são paulo"] = "Rev Med Cir Sao Paulo",
  ["revista de medicina interna, neurologe, psihiatrie, neurochirurgie, dermato-venerologie. medicina interna"] = "Rev Med Interna Neurol Psihiatr Neurochir Dermatovenerol Med Interna",
  ["revista de medicina interna, neurologie, psihiatrie, neurochirurgie, dermato-venerologie. medicina interna"] = "Rev. Med. Interna Neurol. Psihiatr. Neurochir. Dermatovenerol. Med. Intern.",
  ["revista de medicina interna, neurologie, psihiatrie, neurochirurgie, dermato-venerologie. neurologie, psihiatrie, neurochirurgie"] = "Rev. Med. Interna Neurol. Psihiatr. Neurochir. Dermatovenerol. Neurol. Psihiatr. Neurochir.",
  ["revista de medicina legal"] = "Rev Med Leg",
  ["revista de medicina legal de colombia"] = "Rev Med Leg Colomb",
  ["revista de medicina militar"] = "Rev Med Mil",
  ["revista de medicina tropical y parasitología, bacteriología, clínica y laboratorio"] = "Rev Med Trop Parasitol Bacteriol Clin Lab",
  ["revista de medicina veterinaria"] = "Rev Med Vet (Bogota)",
  ["revista de medicina y alimentación"] = "Rev Med Aliment",
  ["revista de medicina y ciencias afines"] = "Rev Med Cienc Afines",
  ["revista de medicina y cirugia de la habana"] = "Rev Med Cir La Habana",
  ["revista de medicina y cirugía"] = "Rev Med Cir",
  ["revista de medicina, cancerología"] = "Rev Med Cancerol",
  ["revista de medicină internă, neurologe, psihiatrie, neurochirurgie, dermato-venerologie. medicină internă"] = "Rev Med Interna Neurol Psihiatr Neurochir Dermatovenerol Med Interna",
  ["revista de medicină internă, neurologie, psihiatrie, neurochirurgie, dermato-venerologie. neurologie, psihiatrie, neurochirurgie"] = "Rev Med Interna Neurol Psihiatr Neurochir Dermatovenerol Neurol Psihiatr Neurochir",
  ["revista de metalurgia (madrid, spain)"] = "Rev. Metal.",
  ["revista de neuro-psiquiatria"] = "Rev. Neuropsiquiatr.",
  ["revista de neuro-psiquiatría"] = "Rev Neuropsiquiatr",
  ["revista de neurologia"] = "Rev Neurol",
  ["revista de obras sanitarias de la nación. obras sanitarias de la nación (argentina)"] = "Rev Obras Sanit Nac B Aires",
  ["revista de obstetricia y ginecologia de venezuela"] = "Rev. Obstet. Ginecol. Venez.",
  ["revista de obstetricia y ginecología de venezuela"] = "Rev Obstet Ginecol Venez",
  ["revista de occidente (madrid, spain : 1923)"] = "Rev Occidente",
  ["revista de odontologia da unesp"] = "Rev Odontol UNESP",
  ["revista de odontologia da unesp / universidade estadual paulista (unesp)"] = "Rev Odontol UNESP",
  ["revista de odontologia da unicid"] = "Rev Odontol UNICID",
  ["revista de odontologia da universidade de sao paulo"] = "Rev. Odontol. Univ. Sao Paulo",
  ["revista de odontologia da universidade de sao paulo / usp"] = "Rev Odontol Univ Sao Paulo",
  ["revista de odontologia da universidade de são paulo"] = "Rev Odontol Univ Sao Paulo",
  ["revista de odontologia da universidade federal de santa catarina"] = "Rev Odontol St Catarina",
  ["revista de ortopedia y traumatologia. edicion latino americana"] = "Rev Ortop Traumatol Ed Lat Am",
  ["revista de ortopedia y traumatología. edición latino americana"] = "Rev Ortop Traumatol Ed Lat Am",
  ["revista de otorrinolaringologia"] = "Rev Otorrinolaringol",
  ["revista de otorrinolaringología"] = "Rev Otorrinolaringol",
  ["revista de pediatrie, obstetrica si ginecologie. obstetrica si ginecologie"] = "Rev. Pediatr. Obstet. Ginecol. Obstet. Ginecol.",
  ["revista de pediatrie, obstetrica si ginecologie. pediatria"] = "Rev. Pediatr. Obstet. Ginecol. Pediatr.",
  ["revista de pediatrie, obstetrică si ginecologie. obstetrică si ginecologie"] = "Rev Pediatr Obstet Ginecol Obstet Ginecol",
  ["revista de pediatrie, obstetrică şi ginecologie. pediatria"] = "Rev Pediatr Obstet Ginecol Pediatr",
  ["revista de planeacion y desarrollo"] = "Rev Planeac Desarro",
  ["revista de planeación y desarrollo"] = "Rev Planeac Desarro",
  ["revista de plasticos modernos"] = "Rev. Plast. Mod.",
  ["revista de protección vegetal"] = "Rev Prot Veg",
  ["revista de psicoanalisis"] = "Rev Psicoanal",
  ["revista de psicoanálisis"] = "Rev Psicoanal",
  ["revista de psicologia social"] = "Rev Psicol Soc",
  ["revista de psicología general y aplicada"] = "Rev Psicol Gen Apl",
  ["revista de psiquiatria clínica"] = "Rev Psiquiatr Clín",
  ["revista de psiquiatria y psicologia medica de europa y america latinas"] = "Rev Psiquiatr Psicol Med Eur Am Lat",
  ["revista de psiquiatría y psicología médica de europa y américa latinas"] = "Rev Psiquiatr Psicol Med Eur Am Lat",
  ["revista de psiquiatrí́a y salud mental"] = "Rev Psiquiatr Salud Ment",
  ["revista de publicaciones navales; suplemento de sanidad naval"] = "Rev Publ Nav",
  ["revista de quimica e farmacia"] = "Rev Quim Farm",
  ["revista de quimica textil"] = "Rev. Quim. Text.",
  ["revista de química e farmácia"] = "Rev Quim Farm",
  ["revista de química y farmacia"] = "Rev Quim Farm",
  ["revista de radiología y fisioterapía"] = "Rev Radiol Fisioter",
  ["revista de salud ambiental"] = "Rev Salud Ambient",
  ["revista de salud fronteriza = journal of border health"] = "Rev Salud Front",
  ["revista de salud publica (bogota, colombia)"] = "Rev Salud Publica (Bogota)",
  ["revista de salud pública (bogotá, colombia)"] = "Rev Salud Publica (Bogota)",
  ["revista de sanidad"] = "Rev Sanid",
  ["revista de sanidad de aeronáutica"] = "Rev Sanid Aeronaut",
  ["revista de sanidad e higiene publica"] = "Rev. Sanid. Hig. Publica (Madr.)",
  ["revista de sanidad e higiene pública"] = "Rev Sanid Hig Publica (Madr)",
  ["revista de sanidad militar"] = "Rev Sanid Milit",
  ["revista de sanidad veterinaria"] = "Rev Sanid Vet",
  ["revista de sanidad y asistencia social"] = "Rev Sanid Asist Soc",
  ["revista de saude publica"] = "Rev Saude Publica",
  ["revista de saúde pública"] = "Rev Saude Publica",
  ["revista de saúde pública"] = "Rev. Saúde Pública",
  ["revista de sifilografía, leprología y dermatología"] = "Rev Sifilogr Leprolog Dermatol",
  ["revista de statistica"] = "Rev Stat",
  ["revista de trabajo social"] = "Revista Trab Soc (Santiago)",
  ["revista de tuberculosis del uruguay"] = "Rev Tuberc Urug",
  ["revista de urologia"] = "Rev Urol(mexico)",
  ["revista de urología"] = "Rev Urol (Mex)",
  ["revista de veterinaria militar"] = "Rev Vet Mil",
  ["revista de viernes medico"] = "Rev Viernes Med",
  ["revista de viernes médico"] = "Rev Viernes Med",
  ["revista del ateneo argentino de odontologia"] = "Rev. Ateneo Argent. Odontol.",
  ["revista del ateneo argentino de odontologia : r.a.a.o"] = "Rev Ateneo Argent Odontol",
  ["revista del ateneo argentino de odontología : r.a.a.o"] = "Rev Ateneo Argent Odontol",
  ["revista del ateneo de la catedra de tecnica de operatoria dental"] = "Rev Ateneo Catedra Tec Oper Dent",
  ["revista del centro policlinico valencia"] = "Rev Cent Policlin Valencia",
  ["revista del cim sobre migraciones en américa latina"] = "Rev CIM Migr Am Lat",
  ["revista del circulo argentino de odontologia"] = "Rev. Circ. Argent. Odontol.",
  ["revista del circulo odontologico de cordoba"] = "Rev. Circ. Odontol. Cordoba",
  ["revista del circulo odontologico de rosario"] = "Rev. Circul. Ondontol. Ros.",
  ["revista del circulo odontologico del sur"] = "Rev Circ Odontol Sur",
  ["revista del círculo odontológico de córdoba"] = "Rev Circ Odontol Cordoba",
  ["revista del círculo odontólogico del oeste. círculo odontológico del oeste, buenos aires"] = "Rev Circ Odontol Oeste",
  ["revista del colegio de medicos y cirujanos de guatemala"] = "Rev Col Med Cir Guatem",
  ["revista del colegio de médicos y cirujanos de guatemala"] = "Rev Col Med Cir Guatem",
  ["revista del colegio estomatologico de guatemala"] = "Rev. Col. Estomatol. Guatem.",
  ["revista del colegio estomatológico de guatemala"] = "Rev Col Estomatol Guatem",
  ["revista del colegio medico de guatemala"] = "Rev. Col. Med. Guatem.",
  ["revista del colegio médico"] = "Rev Col Med",
  ["revista del colegio médico de guatemala"] = "Rev Col Med Guatem",
  ["revista del colegio médico del estado falcón"] = "Rev Col Med Estado Falcon",
  ["revista del colegio nacional de enfermeras"] = "Rev Col Nac Enferm",
  ["revista del desarrollo internacional. revue du développement international. international development review"] = "Rev Desarro Int",
  ["revista del hospital del nino"] = "Rev Hosp Nino (Lima)",
  ["revista del hospital del niño"] = "Rev Hosp Nino (Lima)",
  ["revista del instituto de salubridad y enfermedades tropicales"] = "Rev Inst Salubr Enferm Trop",
  ["revista del instituto mexicano de ingenieros quimicos"] = "Rev. Inst. Mex. Ing. Chim.",
  ["revista del instituto mexicano del petroleo"] = "Rev. Inst. Mex. Pet.",
  ["revista del instituto nacional de cancerologia"] = "Rev. Inst. Nac. Cancerol. (Mex.)",
  ["revista del instituto nacional de cancerología"] = "Rev Inst Nac Cancerol (Mex)",
  ["revista del museo argentino de ciencias naturales nueva serie"] = "Rev. Mus. Argent. Cienc. Nat. Nueva Ser.",
  ["revista del museo de la facultad de odontologia de buenos aires"] = "Rev. Museo Fac. Odontol. B. Aires",
  ["revista del museo de la facultad de odontología de buenos aires"] = "Rev Museo Fac Odontol B Aires",
  ["revista del museo nacional. museo nacional (peru)"] = "Rev Mus Nac",
  ["revista dental"] = "Rev Dent (St Domingo)",
  ["revista dental de chile"] = "Rev Dent Chile",
  ["revista dimension"] = "Rev Dimens",
  ["revista do centro de ciencias biomedicas da universidade federal de uberlandia"] = "Rev. Centro Cienc. Biomed. Univ. Fed. Uberlandia",
  ["revista do centro de ciências biomédicas da universidade federal de uberlândia"] = "Rev Centro Cienc Biomed Univ Fed Uberlandia",
  ["revista do colégio brasileiro de cirurgiões"] = "Rev Col Bras Cir",
  ["revista do hospital das clinicas"] = "Rev Hosp Clin Fac Med Sao Paulo",
  ["revista do hospital das clinicas, faculdade de medicina da universidade de sao paulo"] = "Rev. Hosp. Clin. Fac. Med. Sao Paulo",
  ["revista do hospital das clinicas; faculdade de medicina da universidade de sao paulo"] = "Rev Hosp Clin Fac Med Sao Paulo",
  ["revista do hospital das clínicas"] = "Rev Hosp Clin Fac Med Sao Paulo",
  ["revista do instituto adolfo lutz"] = "Rev Inst Adolfo Lutz",
  ["revista do instituto de antibioticos, universidade federal de pernambuco"] = "Rev. Inst. Antibiot. (Recife)",
  ["revista do instituto de antibióticos, universidade federal de pernambuco"] = "Rev Inst Antibiot (Recife)",
  ["revista do instituto de medicina tropical de sao paulo"] = "Rev Inst Med Trop Sao Paulo",
  ["revista do instituto de medicina tropical de são paulo"] = "Rev Inst Med Trop Sao Paulo",
  ["revista do instituto de medicina tropical de são paulo"] = "Rev. Inst. Med. trop. S. Paulo",
  ["revista do instituto historico e geographico brazileiro (1906)"] = "Rev Inst Hist Geogr Bras",
  ["revista do servico nacional de tuberculose"] = "Rev Serv Nac Tubers",
  ["revista do serviço nacional de tuberculose"] = "Rev Serv Nac Tubers",
  ["revista d’arqueologia de ponent"] = "RAPon",
  ["revista economica"] = "Rev Econ",
  ["revista economica do nordeste"] = "Rev Econ Nordeste",
  ["revista economică"] = "Rev Econ",
  ["revista econômica do nordeste"] = "Rev Econ Nordeste",
  ["revista ecuatoriana de higiene y medicina tropical"] = "Rev Ecuat Hig Med Trop",
  ["revista ecuatoriana de medicina y ciencias biologicas"] = "Rev Ecuat Med Cienc Biol",
  ["revista ecuatoriana de medicina y ciencias biológicas"] = "Rev Ecuat Med Cienc Biol",
  ["revista ecuatoriana de pediatria"] = "Rev Ecuat Pediatr",
  ["revista ecuatoriana de pediatria y puericultura"] = "Rev Ecuat Pediatr Pueric",
  ["revista ecuatoriana de pediatría"] = "Rev Ecuat Pediatr",
  ["revista ecuatoriana de pediatría y puericultura"] = "Rev Ecuat Pediatr Pueric",
  ["revista electrónica de investigación psicoeducativa : reipe = electronic journal of research in educational psychology"] = "Rev Electron Investig Psicoeduc Psigopedag",
  ["revista electrônica de comunicação, informação & inovação em saúde : reciis"] = "Rev Electron Comun Inf Inov Saude",
  ["revista enfermagem em novas dimensoes"] = "Rev. Enferm. Nov. Dimens.",
  ["revista enfermagem em novas dimensões"] = "Rev Enferm Nov Dimens",
  ["revista espanola de anestesiologia"] = "Rev. Esp. Anestesiol.",
  ["revista espanola de anestesiologia y reanimacion"] = "Rev Esp Anestesiol Reanim",
  ["revista espanola de antropologia americana / departamento de antropologia y etnologia de america"] = "Rev Esp Antropol Am",
  ["revista espanola de cardiologia"] = "Rev Esp Cardiol",
  ["revista espanola de endodoncia"] = "Rev. Esp. Endodoncia",
  ["revista espanola de endodoncia / aede"] = "Rev Esp Endodoncia",
  ["revista espanola de enfermedades digestivas"] = "Rev. Esp. Enferm. Dig.",
  ["revista espanola de enfermedades digestivas : organo oficial de la sociedad espanola de patologia digestiva"] = "Rev Esp Enferm Dig",
  ["revista espanola de estomatologia"] = "Rev. Esp. Estomatol.",
  ["revista espanola de fisiologia"] = "Rev Esp Fisiol",
  ["revista espanola de geriatria y gerontologia"] = "Rev Esp Geriatr Gerontol",
  ["revista espanola de las enfermedades del aparato digestivo"] = "Rev. Esp. Enferm. Apar. Dig.",
  ["revista espanola de las enfermedades del aparato digestivo y de la nutricion"] = "Rev Esp Enferm Apar Dig Nutr",
  ["revista espanola de medicina nuclear"] = "Rev. Esp. Med. Nucl.",
  ["revista espanola de obstetricia y ginecologia"] = "Rev. Esp. Obstet. Ginecol.",
  ["revista espanola de oncologia"] = "Rev. Esp. Oncol.",
  ["revista espanola de oto-neuro-oftalmologia y neurocirugia"] = "Rev. Esp. Otoneurooftalmol. Neurocir.",
  ["revista espanola de paleontologia"] = "Rev Esp Paleontol",
  ["revista espanola de paradoncia"] = "Rev Esp Parad",
  ["revista espanola de pediatria"] = "Rev. Esp. Pediatr.",
  ["revista espanola de quimioterapia"] = "Rev. Esp. Quimioter.",
  ["revista espanola de quimioterapia : publicacion oficial de la sociedad espanola de quimioterapia"] = "Rev Esp Quimioter",
  ["revista espanola de reumatismo y enfermedades osteo-articulares"] = "Rev. Esp. Reum. Enferm. Osteoartic.",
  ["revista espanola de reumatismo y enfermedades osteoarticulares"] = "Rev Esp Reum Enferm Osteoartic",
  ["revista espanola de reumatologia : organo oficial de la sociedad espanola de reumatologia"] = "Rev Esp Reumatol",
  ["revista espanola de salud publica"] = "Rev Esp Salud Publica",
  ["revista espanola de tuberculosis"] = "Rev. Esp. Tuberc.",
  ["revista espanola la de documentacion cientifica"] = "Rev Esp Doc Cient",
  ["revista española de anestesiología"] = "Rev Esp Anestesiol",
  ["revista española de anestesiología y reanimación"] = "Rev Esp Anestesiol Reanim",
  ["revista española de antropología americana"] = "Rev Esp Antropol Am",
  ["revista española de cardiología"] = "Rev Esp Cardiol",
  ["revista española de cardiología (english ed.)"] = "Rev Esp Cardiol (Engl Ed)",
  ["revista española de cirugía"] = "Rev Esp Cir (Madr 1944)",
  ["revista española de cirugía ortopédica y traumatología"] = "Rev Esp Cir Ortop Traumatol",
  ["revista española de cirugía, traumatología y ortopedia"] = "Rev Eesp Cir Traumatol Ortop",
  ["revista española de endodoncia"] = "Rev Esp Endodoncia",
  ["revista española de enfermedades digestivas : organo oficial de la sociedad española de patología digestiva"] = "Rev Esp Enferm Dig",
  ["revista española de esclerosis múltiple"] = "Rev Esp Escler Mult",
  ["revista española de estomatología"] = "Rev Esp Estomatol",
  ["revista española de fisiología"] = "Rev Esp Fisiol",
  ["revista española de geriatría y gerontología"] = "Rev Esp Geriatr Gerontol",
  ["revista española de herpetología"] = "Rev Esp Herpetol",
  ["revista española de las enfermedades del aparato digestivo"] = "Rev Esp Enferm Apar Dig",
  ["revista española de las enfermedades del aparato digestivo y de la nutrición"] = "Rev Esp Enferm Apar Dig Nutr",
  ["revista española de lingüística aplicada"] = "Rev Esp Linguist Apl",
  ["revista española de medicina nuclear"] = "Rev Esp Med Nucl",
  ["revista española de medicina nuclear e imagen molecular"] = "Rev Esp Med Nucl Imagen Mol",
  ["revista española de medicina y cirugía de guerra"] = "Rev Esp Med Cir Guerra",
  ["revista española de obstetricia y ginecología"] = "Rev Esp Obstet Ginecol",
  ["revista española de oncología"] = "Rev Esp Oncol",
  ["revista española de oto-neuro-oftalmología y neurocirugía"] = "Rev Esp Otoneurooftalmol Neurocir",
  ["revista española de paleontología"] = "Rev Esp Paleontol",
  ["revista española de patología : publicación oficial de la sociedad española de anatomía patológica y de la sociedad española de citología"] = "Rev Esp Patol",
  ["revista española de pediatría"] = "Rev Esp Pediatr",
  ["revista española de quimioterapia : publicación oficial de la sociedad española de quimioterapia"] = "Rev Esp Quimioter",
  ["revista española de reumatismo y enfermedades osteoarticulares"] = "Rev Esp Reum Enferm Osteoartic",
  ["revista española de reumatología : órgano oficial de la sociedad española de reumatología"] = "Rev Esp Reumatol",
  ["revista española de salud pública"] = "Rev Esp Salud Publica",
  ["revista española de sanidad penitenciaria"] = "Rev Esp Sanid Penit",
  ["revista española de tuberculosis"] = "Rev Esp Tuberc",
  ["revista española la de documentación científica"] = "Rev Esp Doc Cient",
  ["revista española de economía"] = "Revista Española Econ.",
  ["revista española de linguística"] = "REspLing",
  ["revista estadistica (havana, cuba)"] = "Rev Estad",
  ["revista estadística (havana, cuba)"] = "Rev Estad",
  ["revista estomatológica de cuba"] = "Rev Estomatol Cuba",
  ["revista europea de estudios latinoamericanos y del caribe = european review of latin american and caribbean studies"] = "Rev Eur Estud Latinoam Caribe",
  ["revista europea de odonto-estomatologia"] = "Rev. Eur. Odontoestomatol.",
  ["revista europea de odonto-estomatología"] = "Rev Eur Odontoestomatol",
  ["revista f o a"] = "Rev FOA",
  ["revista facultad nacional de agronomia medellin"] = "Rev Fac Nac Agron Medellin",
  ["revista farmacéutica de cuba"] = "Rev Farm Cuba",
  ["revista farmacéutica peruana"] = "Rev Farm Peru",
  ["revista fitotecnia mexicana"] = "Rev. Fitotec. Mex.",
  ["revista foa"] = "Rev. FOA",
  ["revista gaucha de enfermagem"] = "Rev. Gaucha Enferm.",
  ["revista gaucha de enfermagem / eenfufrgs"] = "Rev Gaucha Enferm",
  ["revista gaucha de odontologia"] = "Rev. Gaucha Odontol.",
  ["revista gaúcha de enfermagem"] = "Rev Gaucha Enferm",
  ["revista gaúcha de odontologia"] = "Rev Gaucha Odontol",
  ["revista general de marina"] = "Rev Gen Mar",
  ["revista geofisica"] = "Rev. Geofis.",
  ["revista geografica del instituto panamericano de geografia e historia"] = "Rev Geogr Inst Panam Geogr Hist",
  ["revista geográfica del instituto panamericano de geografía e historia"] = "Rev Geogr Inst Panam Geogr Hist",
  ["revista geologica de chile"] = "Rev. Geol. Chile",
  ["revista goiana de medicina"] = "Rev Goiana Med",
  ["revista griot"] = "Rev Griot",
  ["revista guatemalteca de estomatologia"] = "Rev. Guatem. Estomatol.",
  ["revista guatemalteca de estomatología"] = "Rev Guatem Estomatol",
  ["revista iberica de endocrinologia"] = "Rev. Iber. Endocrinol.",
  ["revista iberica de parasitologia"] = "Rev Iber Parasitol",
  ["revista ibero-americana de ortodoncia"] = "Rev. Iberoam. Ortod.",
  ["revista ibero-americana de ortodoncia : publicacion oficial de la asociacion ibero-americana ortodoncistas"] = "Rev Iberoam Ortod",
  ["revista ibero-americana de ortodoncia : publicación oficial de la asociación ibero-americana ortodoncistas"] = "Rev Iberoam Ortod",
  ["revista iberoamericana de bioética"] = "Rev Iberoam Bioet",
  ["revista iberoamericana de estudios de desarrollo = iberoamerican journal of development studies"] = "Rev Iberoam Estud Desarro",
  ["revista iberoamericana de micologia : organo de la asociacion espanola de especialistas en micologia"] = "Rev Iberoam Micol",
  ["revista iberoamericana de micología"] = "Rev Iberoam Micol",
  ["revista ibérica de aracnología"] = "Rev Iber Aracnol",
  ["revista ibérica de endocrinología"] = "Rev Iber Endocrinol",
  ["revista ibérica de parasitología"] = "Rev Iber Parasitol",
  ["revista ibys"] = "Rev Ibys",
  ["revista ieee america latina"] = "Rev. IEEE Am. Lat.",
  ["revista ingenierias univsidad de medellin"] = "Rev. Ing. Univ. Medellin",
  ["revista ingeniería biomédica"] = "Rev Ing Biomed",
  ["revista integracion. temas de matematicas"] = "Rev. Integr. Temas Mat.",
  ["revista integración"] = "Rev. Integr. Temas Mat.",
  ["revista interamericana de bibliografía = review of inter-american bibliography"] = "Rev Interam Bibliogr",
  ["revista interamericana de planificacion"] = "Rev Interam Planif",
  ["revista interamericana de planificación"] = "Rev Interam Planif",
  ["revista interamericana de psicología = interamerican journal of psychology"] = "Interam J Psychol",
  ["revista interamericana de radiologia"] = "Rev. Interam. Radiol.",
  ["revista interamericana de radiología"] = "Rev Interam Radiol",
  ["revista internacional de andrología"] = "Rev Int Androl",
  ["revista internacional de ciencias del deporte"] = "Rev Int Cienc Deporte",
  ["revista internacional de metodos numericos para calculo y diseno en ingenieria"] = "Rev. Int. Metodos Numer. Calc. Diseno Ing.",
  ["revista internacional de métodos numéricos para cálculo y diseño en ingeniería"] = "Rev. Internac. Métod. Numér. Cálc. Diseñ. Ingr.",
  ["revista internacional de psicología y terapia psicológica = international journal of psychology and psychological therapy"] = "Rev Int Psicol Ter Psicol",
  ["revista internacional de sociología"] = "Rev Int Sociol",
  ["revista istorică"] = "Rev Istor",
  ["revista juridica de la universidad de puerto rico"] = "Rev Jurid Univ P R",
  ["revista jurídica de la universidad de puerto rico"] = "Rev Jurid Univ P R",
  ["revista kairós : gerontologia"] = "Rev Kairos",
  ["revista kuba de medicina tropical y parasitologia"] = "Rev Kuba Med Trop Parasitol",
  ["revista kuba de medicina tropical y parasitología"] = "Rev Kuba Med Trop Parasitol",
  ["revista lasallista de investigacion"] = "Rev Lasallista Investig",
  ["revista latina de sociología"] = "Rev Lat Sociol",
  ["revista latino-americana de enfermagem"] = "Rev Lat Am Enfermagem",
  ["revista latinoamericana de anatomia patologica"] = "Rev Latinoam Anat Patol",
  ["revista latinoamericana de anatomía patológica"] = "Rev Latinoam Anat Patol",
  ["revista latinoamericana de bioética"] = "Rev Latinoam Bioet",
  ["revista latinoamericana de cirugia plastica"] = "Rev. Latinoam. Cir. Plast.",
  ["revista latinoamericana de cirugía plástica"] = "Rev Latinoam Cir Plast",
  ["revista latinoamericana de etnomatematica. perspectivas socioculturales de la educacion matematica"] = "Rev. Latinoam. Etnomat.",
  ["revista latinoamericana de filosofía"] = "RLAF",
  ["revista latinoamericana de ingenieria quimica y quimica aplicada"] = "Rev. Latinoam. Ing. Quim. Quim. Apl.",
  ["revista latinoamericana de microbiologia"] = "Rev Latinoam Microbiol",
  ["revista latinoamericana de microbiologia y parasitologia"] = "Rev. Latinoam. Microbiol. Parasitol. (Mex.)",
  ["revista latinoamericana de microbiología"] = "Rev Latinoam Microbiol (1958)",
  ["revista latinoamericana de microbiología y parasitología"] = "Rev Latinoam Microbiol Parasitol (Mex)",
  ["revista latinoamericana de patologia"] = "Rev. Latinoam. Patol.",
  ["revista latinoamericana de patología"] = "Rev Latinoam Patol",
  ["revista latinoamericana de perinatologia : organo oficial de la federacion latinoamericana de asociaciones de medicina [perinatal] (f.l.a.m.p.)"] = "Rev Latinoam Perinatol",
  ["revista latinoamericana de perinatología : órgano oficial de la federación latinoamericana de asociaciones de medicina [perinatal] (f.l.a.m.p.)"] = "Rev Latinoam Perinatol",
  ["revista latinoamericana de psicologia"] = "Rev Latinoam Psicol",
  ["revista latinoamericana de psicología"] = "Rev Latinoam Psicol",
  ["revista latinoamericana de quimica"] = "Rev. Latinoam. Quim.",
  ["revista latinoamericana de química"] = "Rev Latinoam Quim",
  ["revista loteria"] = "Loteria (Panama)",
  ["revista lotería"] = "Loteria (Panama)",
  ["revista matematica complutense"] = "Rev. Mat. Complutense",
  ["revista matematica de la universidad del zulia. divulgaciones matematicas"] = "Divulg. Mat.",
  ["revista matematica iberoamericana"] = "Rev. Mat. Iberoam.",
  ["revista matemática complutense"] = "Rev Mat Complut",
  ["revista matemática complutense"] = "Rev. Mat. Complut.",
  ["revista matemática de la universidad complutense de madrid"] = "Rev. Mat. Univ. Complut. Madrid",
  ["revista matemática iberoamericana"] = "Rev. Mat. Iberoamericana",
  ["revista medica (instituto mexicano del seguro social)"] = "Rev Med (Mex)",
  ["revista medica cubana"] = "Rev Med Cubana",
  ["revista medica da aeronautica"] = "Rev Med Aeron",
  ["revista medica da aeronautica do brasil"] = "Rev Med Aeron Braz",
  ["revista medica de chile"] = "Rev Med Chil",
  ["revista medica de cordoba"] = "Rev Med Cordoba",
  ["revista medica de costa rica"] = "Rev. Med. Costa Rica",
  ["revista medica de mocambique"] = "Rev Med Moamb",
  ["revista medica de panama"] = "Rev Med Panama",
  ["revista medica de valparaiso"] = "Rev Med Valparaiso",
  ["revista medica del hospital central del empleado"] = "Rev Med Hosp Cent Empl",
  ["revista medica del hospital espanol"] = "Rev Medica Hosp Esp",
  ["revista medica del hospital general"] = "Rev. Med. Hosp. Gen. (Mex.)",
  ["revista medica del hospital obrero"] = "Rev Med Hosp Obrero",
  ["revista medica del instituto mexicano del seguro social"] = "Rev Med Inst Mex Seguro Soc",
  ["revista medica del ipss / instituto peruano de seguridad social"] = "Rev Med IPSS",
  ["revista medica do sul de minas"] = "Rev Medica Minas",
  ["revista medica hondurena"] = "Rev. Med. Hondur.",
  ["revista medica peruana"] = "Rev Med Peru",
  ["revista medico-chirurgicala a societatii de medici si naturalisti din iasi"] = "Rev Med Chir Soc Med Nat Iasi",
  ["revista medico-chirurgicală̆ a societă̆ţ̜ii de medici ş̧i naturaliş̧ti din iaş̧i"] = "Rev Med Chir Soc Med Nat Iasi",
  ["revista meteorologica"] = "Rev. Meteor.",
  ["revista mexicana de análisis de la conducta = mexican journal of behavior analysis"] = "Rev Mex Anal Conducta",
  ["revista mexicana de astronomia y astrofisica"] = "Rev. Mex. Astron. Astrofis.",
  ["revista mexicana de astronomica y astrofisica"] = "Rev. Mex. Astron. Astrofis.",
  ["revista mexicana de astronomía y astrofísica"] = "Rev Mex Astron Astrofis",
  ["revista mexicana de biodiversidad"] = "Rev Mex Biodivers",
  ["revista mexicana de ciencias agrícolas"] = "Rev Mex De Cienc Agric",
  ["revista mexicana de ciencias farmaceuticas"] = "Rev. Mex. Cienc. Farm.",
  ["revista mexicana de ciencias farmacéutica"] = "Rev Mex Cienc Farm",
  ["revista mexicana de ciencias forestales"] = "Rev Mex Cienc For",
  ["revista mexicana de ciencias médicas y biológicas"] = "Rev Mex Cienc Med Bio",
  ["revista mexicana de ciencias pecuarias"] = "Rev Mex Cienc Pecu",
  ["revista mexicana de ciencias politicas y sociales"] = "Rev Mex Cienc Polit Soc",
  ["revista mexicana de ciencias políticas y sociales"] = "Rev Mex Cienc Polit Soc",
  ["revista mexicana de cirugia, ginecologia y cancer"] = "Rev. Mex. Cir. Ginecol. Cancer",
  ["revista mexicana de cirugía, ginecología y cáncer"] = "Rev Mex Cir Ginecol Cancer",
  ["revista mexicana de fisica"] = "Rev. Mex. Fis.",
  ["revista mexicana de fisica e"] = "Rev. Mex. Fis. E",
  ["revista mexicana de fisica e. publicacion de ensenanza, historia y filosofia de la sociedad mexicana de fisica"] = "Rev. Mex. Fis. E",
  ["revista mexicana de fitopatologia : la revista oficial de la sociedad mexicana de fitopatologia"] = "Rev Mex Fitopatol",
  ["revista mexicana de física"] = "Rev. Mexicana Fís.",
  ["revista mexicana de ingenieria biomedica"] = "Rev. Mex. Ing. Biomed.",
  ["revista mexicana de ingenieria quimica"] = "Rev. Mex. Ing. Chim.",
  ["revista mexicana de ingeniería química"] = "Rev Mex Ing Quim",
  ["revista mexicana de micología : órgano oficial de la sociedad mexicana de micología"] = "Rev Mex Micol",
  ["revista mexicana de sociologia"] = "Rev Mex Sociol",
  ["revista mexicana de sociología"] = "Rev Mex Sociol",
  ["revista mexicana de trastornos alimentarios = mexican journal of eating disorders"] = "Rev Mex Trastor Aliment",
  ["revista mexicana de tuberculosis y enfermedades del aparato respiratorio"] = "Rev Mex Tuberc Enferm Apar Respir",
  ["revista mexicana de urologia"] = "Rev. Mex. Urol.",
  ["revista mexicana de urología"] = "Rev Mex Urol",
  ["revista médica"] = "Rev Med (Puebla)",
  ["revista médica (instituto mexicano del seguro social)"] = "Rev Med (Mex)",
  ["revista médica brasileira"] = "Rev Med Bras",
  ["revista médica cubana"] = "Rev Med Cubana",
  ["revista médica da aeronáutica"] = "Rev Med Aeron",
  ["revista médica da aeronáutica do brasil"] = "Rev Med Aeron Braz",
  ["revista médica da bahia"] = "Rev Med Bahia",
  ["revista médica de chile"] = "Rev Med Chil",
  ["revista médica de chile. suplemento"] = "Rev Med Chile Supl",
  ["revista médica de costa rica"] = "Rev Med Costa Rica",
  ["revista médica de córdoba"] = "Rev Med Cordoba",
  ["revista médica de moçambique"] = "Rev Med Moamb",
  ["revista médica de panamá"] = "Rev Med Panama",
  ["revista médica de pernambuco"] = "Rev Med Pernamb",
  ["revista médica de rosario"] = "Rev Med Rosario",
  ["revista médica de valparaíso"] = "Rev Med Valparaiso",
  ["revista médica del hospital británico"] = "Rev Med Hosp Britanico",
  ["revista médica del hospital central del empleado"] = "Rev Med Hosp Cent Empl",
  ["revista médica del hospital español"] = "Rev Medica Hosp Esp",
  ["revista médica del hospital general"] = "Rev Med Hosp Gen (Mex)",
  ["revista médica del hospital obrero"] = "Rev Med Hosp Obrero",
  ["revista médica del instituto mexicano del seguro social"] = "Rev Med Inst Mex Seguro Soc",
  ["revista médica del ipss"] = "Rev Med IPSS",
  ["revista médica del paraguay"] = "Rev Med Parag",
  ["revista médica do sul de minas"] = "Rev Medica Minas",
  ["revista médica dominicana"] = "Rev Med Dominic",
  ["revista médica hondureña"] = "Rev Med Hondur",
  ["revista médica latino-americana"] = "Rev Med Latinoam",
  ["revista médica municipal"] = "Rev Med Munic",
  ["revista médica pasteur"] = "Rev Med Pasteur",
  ["revista médica peruana"] = "Rev Med Peru",
  ["revista médica veracruzana"] = "Rev Med Veracruzana",
  ["revista médica xalapeña"] = "Rev Med Xalapena",
  ["revista médica. ateneo ramón y cajal, mexico"] = "Rev Med",
  ["revista médico-quirúrgica de oriente"] = "Rev Med Quir Oriente",
  ["revista médico-quirúrgica de patología femenina"] = "Rev Med Quir Patol Fem",
  ["revista militar de remonta e veterinaria"] = "Rev Mil Remonta Vet",
  ["revista mvz córdoba"] = "Rev MVZ Cordoba",
  ["revista neurologica de buenos aires"] = "Rev Neurol B Aires",
  ["revista observaciones filosóficas"] = "Rev Obs Filos",
  ["revista odonto-estomatologica"] = "Rev. Odontoestomatol.",
  ["revista odonto-estomatológica"] = "Rev Odontoestomatol",
  ["revista odontologica"] = "Rev Odontol (La Paz)",
  ["revista odontologica de concepcion"] = "Rev. Odontol. Concepcion",
  ["revista odontologica de costa rica"] = "Rev. Odontol. Costa Rica",
  ["revista odontologica de puerto rico"] = "Rev. Odontol. P. R.",
  ["revista odontologica del circulo de odontologos del paraguay"] = "Rev. Odontol. Circ. Odontol. Parag.",
  ["revista odontologica do parana"] = "Rev Odontol Parana",
  ["revista odontologica ecuatoriana"] = "Rev. Odontol. Ecuat.",
  ["revista odontologica, facultad de odontologia, universidad nacional de cordoba"] = "Rev. Odontol. (Cordoba)",
  ["revista odontologica; facultad de odontologia, universidad nacional de cordoba"] = "Rev Odontol (Cordoba)",
  ["revista odontológica"] = "Rev Odontol Circ Odontol Parag",
  ["revista odontológica de concepción"] = "Rev Odontol Concepcion",
  ["revista odontológica de costa rica"] = "Rev Odontol Costa Rica",
  ["revista odontológica de puerto rico"] = "Rev Odontol P R",
  ["revista odontológica do paraná"] = "Rev Odontol Parana",
  ["revista odontológica ecuatoriana"] = "Rev Odontol Ecuat",
  ["revista organon"] = "Rev Organon",
  ["revista oto-neuro-oftalmologica y cirugia neurologica sudamericana"] = "Rev Oto Neuro Oftalmol Cir Neurol Sud Am",
  ["revista oto-neuro-oftalmológica y cirugía neurologica sudamericana"] = "Rev Oto Neuro Oftalmol Cir Neurol Sud Am",
  ["revista pan-amazônica de saúde"] = "Rev Panamazonica Saude",
  ["revista panamericana de medicina y cirugía del tórax"] = "Rev Panam Med Cir Torax",
  ["revista panamericana de salud publica"] = "Rev. Panam. Salud Publica",
  ["revista panamericana de salud pública = pan american journal of public health"] = "Rev Panam Salud Publica",
  ["revista paulista de endodontia"] = "Rev Paul Endodontia",
  ["revista paulista de enfermagem"] = "Rev Paul Enferm",
  ["revista paulista de hospitais"] = "Rev Paul Hosp",
  ["revista paulista de medicina"] = "Rev Paul Med",
  ["revista paulista de odontologia"] = "Rev Paul Odontol",
  ["revista paulista de pediatria : orgão oficial da sociedade de pediatria de são paulo"] = "Rev Paul Pediatr",
  ["revista paulista de tisiologia e do tórax"] = "Rev Paul Tisiol Torax",
  ["revista pernambucana de odontologia"] = "Rev Pernambucana Odontol",
  ["revista peruana de biología"] = "Rev Peru Biol",
  ["revista peruana de medicina de experimental y salud pública"] = "Rev. Peru. Med. Exp. Salud Publica",
  ["revista peruana de medicina experimental y salud pública"] = "Rev Peru Med Exp Salud Publica",
  ["revista peruana de pediatria"] = "Rev Peru Pediatr",
  ["revista peruana de pediatría"] = "Rev Peru Pediatr",
  ["revista peruana de poblacion"] = "Rev Peru Poblac",
  ["revista peruana de población"] = "Rev Peru Poblac",
  ["revista peruana de quimica e ingenieria quimica"] = "Rev. Peru. Quim. Ing. Quim.",
  ["revista portuguesa de arqueologia"] = "RPortA",
  ["revista portuguesa de cardiologia"] = "Rev. Port. Cardiol.",
  ["revista portuguesa de cardiologia : orgão oficial da sociedade portuguesa de cardiologia = portuguese journal of cardiology : an official journal of the portuguese society of cardiology"] = "Rev Port Cardiol",
  ["revista portuguesa de cirurgia cardio-toracica e vascular : orgao oficial da sociedade portuguesa de cirurgia cardio-toracica e vascular"] = "Rev Port Cir Cardiotorac Vasc",
  ["revista portuguesa de cirurgia cardio-torácica e vascular : órgão oficial da sociedade portuguesa de cirurgia cardio-torácica e vascular"] = "Rev Port Cir Cardiotorac Vasc",
  ["revista portuguesa de estomatologia e cirurgia maxilo-facial"] = "Rev Port Estomatol Cir Maxilofac",
  ["revista portuguesa de estomatologia e cirurgia maxilofacial"] = "Rev. Port. Estomatol. Cir. Maxilofac.",
  ["revista portuguesa de farmacia"] = "Rev Port Farm",
  ["revista portuguesa de farmácia"] = "Rev Port Farm",
  ["revista portuguesa de historia"] = "Rev Port Hist",
  ["revista portuguesa de história"] = "Rev Port Hist",
  ["revista portuguesa de medicina militar"] = "Rev Port Med Mil",
  ["revista portuguesa de pediatria"] = "Rev Port Pediatr",
  ["revista portuguesa de pediatria e puericultura"] = "Rev Port Pediatr Pueric",
  ["revista portuguesa de pneumologia"] = "Rev Port Pneumol",
  ["revista portuguesa de quimica"] = "Rev. Port. Quim.",
  ["revista puertorriqueña de medicina y salud pública"] = "Rev Puertoriq Med Salud Publica",
  ["revista puertorriqueña de psicología"] = "Rev Puertorriquena Psicol",
  ["revista quirurgica espanola"] = "Rev. Quir. Esp.",
  ["revista quirurgica espanola : rqe"] = "Rev Quir Esp",
  ["revista quirúrgica española : rqe"] = "Rev Quir Esp",
  ["revista regional de aracatuba associacao paulista de cirurgioes dentistas"] = "Rev. Reg. Aracatuba Assoc. Paul. Cir. Dent.",
  ["revista regional de aracatuba, associacao paulista de cirurgioes dentistas"] = "Rev Reg Aracatuba Assoc Paul Cir Dent",
  ["revista regional de araçatuba, associação paulista de cirurgiões dentistas"] = "Rev Reg Aracatuba Assoc Paul Cir Dent",
  ["revista romana de bioetica"] = "Rev Rom Bioet",
  ["revista romana de materiale"] = "Rev. Rom. Mater.",
  ["revista romana de statistica : organ al comisiei nationale pentru statistica"] = "Rev Rom Stat",
  ["revista română de medicină de laborator"] = "Rev Rom Med Lab",
  ["revista română de statistică : organ al comisiei naționale pentru statistică"] = "Rev Rom Stat",
  ["revista românǎ de bioeticǎ"] = "Rev Rom Bioet",
  ["revista stiintelor medicale: medicina interna"] = "Rev Stiint Med",
  ["revista sudamericana de morfología"] = "Rev Sudam Morfol",
  ["revista ştiinţelor medicale: medicina internă"] = "Rev Stiint Med",
  ["revista universitaria"] = "Rev Univ",
  ["revista uruguaya de psicoanalisis"] = "Rev Urug Psicoanal",
  ["revista uruguaya de psicoanálisis"] = "Rev Urug Psicoanal",
  ["revista venezolana de filosofía"] = "RVF",
  ["revista venezolana de sanidad y asistencia social"] = "Rev Venez Sanid Asist Soc",
  ["revista venezolana de urologia"] = "Rev. Venez. Urol.",
  ["revista venezolana de urología"] = "Rev Venez Urol",
  ["revista virtual de estudos da linguagem"] = "Rev Virtual Estud Ling",
  ["revista zootécnica"] = "Rev Zootec (B. Aires)",
  ["revista. asociación colombiana de médicos veterinarios"] = "Rev Asoc Columb Med Vet",
  ["revista. asociación médica latinoamericana"] = "Rev Asoc Med Latinoam",
  ["revista. asociación médica mexicana"] = "Rev Asoc Medica Mex",
  ["revista. caracas (venezuela). policlínica caracas"] = "Rev Policlin Caracas",
  ["revista. centro de estudos demograficos (portugal)"] = "Rev Cent Estud Demograficos",
  ["revista. centro de estudos demográficos (portugal)"] = "Rev Cent Estud Demograficos",
  ["revista. chile. servicio nacional de salud"] = "Rev Serv Nac Salud",
  ["revista. circulo argentino de odontologia"] = "Rev Circ Argent Odontol",
  ["revista. circulo odontologico de rosario"] = "Rev Circul Ondontol Ros",
  ["revista. círculo argentino de odontología"] = "Rev Circ Argent Odontol",
  ["revista. círculo odontológico de rosario"] = "Rev Circul Ondontol Ros",
  ["revista. facultad de medicina de tucumán"] = "Rev Fac Med Tucuman",
  ["revista. red cross. colombia. cruz roja nacional"] = "Rev Red Cross",
  ["revista. sociedad colombiana de ortodoncia"] = "Rev Soc Colomb Ortod",
  ["revista. sociedad peruana de prótesis dental y máxilo-facial, lima"] = "Rev Soc Peru Protes Dent Maxilofac Lima",
  ["revista. tucuman, argentine republic. universidad. facultad de medicina"] = "Rev Fac Med Tucuman",
  ["revista. união odontologica brasileira"] = "Rev Uniao Odontol Bras",
  ["revista. universidad nacional mayor de san marcos. facultad de farmacia y bioquimica"] = "Rev Univ San Marcos Fac Farm Bioquim",
  ["revista/review interamericana"] = "Rev Rev Interam",
  ["revmatologiia (moscow, russia)"] = "Revmatologiia (Mosk)",
  ["revolution (oakland, calif.)"] = "Revolution",
  ["revolution (staten island, n.y.)"] = "Revolution",
  ["revstat statistical journal"] = "Revstat Stat J",
  ["revue africaine de developpement"] = "Revue Africaine Devel.",
  ["revue annuelle - societe odonto-stomatologique du nord-est"] = "Rev Annu Soc Odontostomatol Nordest",
  ["revue annuelle - société odonto-stomatologique du nord-est"] = "Rev Annu Soc Odontostomatol Nordest",
  ["revue annuelle, societe odonto-stomatologique du nord-est"] = "Rev. Annu. Soc. Odontostomatol. Nordest.",
  ["revue archeologique du centre de la france"] = "Rev Archeol Centre France",
  ["revue archéologique de l’est et du centre-est"] = "RAE",
  ["revue archéologique du centre de la france"] = "Rev Archeol Centre France",
  ["revue archéologique"] = "RA",
  ["revue archéologique de l’est et du centre-est"] = "RAE",
  ["revue archéologique de narbonnaise"] = "RAN",
  ["revue archéologique de picardie"] = "RAPic",
  ["revue archéologique du centre de la france"] = "RACFr",
  ["revue archéologique du centre [de la france]"] = "RACF",
  ["revue belge d'archéologie et d'histoire de l'art"] = "Rev Belge Archeol Hist Art",
  ["revue belge d'histoire contemporaine. belgisch tijdschrift voor nieuwste geschiedenis"] = "Belg Tijdschr Nieuwste Geschied",
  ["revue belge d'histoire militaire = belgisch tijdschrift voor militaire geschiedenis"] = "Rev Belge Hist Milit",
  ["revue belge d'homoeopathie"] = "Rev Belg Homoeopath",
  ["revue belge de geographie"] = "Rev Belge Geogr",
  ["revue belge de géographie"] = "Rev Belge Geogr",
  ["revue belge de medecine dentaire"] = "Rev. Belge Med. Dent.",
  ["revue belge de medecine dentaire. belgisch tijdschrift voor tandheelkunde"] = "Rev Belge Med Dent",
  ["revue belge de médecine dentaire"] = "Rev Belge Med Dent (1984)",
  ["revue belge de médecine dentaire. belgisch tijdschrift voor tandheelkunde"] = "Rev Belge Med Dent",
  ["revue belge de numismatique et de sigillographie"] = "RBN",
  ["revue belge de pathologie et de medecine experimentale"] = "Rev. Belg. Pathol. Med. Exp.",
  ["revue belge de pathologie et de médecine expérimentale"] = "Rev Belg Pathol Med Exp",
  ["revue belge de philologie et d'histoire. belgisch tijdschrift voor philologie en geschiedenis"] = "Rev Belge Philol Hist",
  ["revue belge de philologie et d’histoire"] = "RBPh",
  ["revue belge de philologie et d’histoire. belgisch tijdschrift voor philologie en geschiedenis"] = "Rev Belge Philol Hist",
  ["revue belge de science dentaire. belgisch tijdschrift voor tandheelkunde"] = "Rev Belge Sci Dent",
  ["revue belge de stomatologie. belgisch tijdschrift voor stomatologie"] = "Rev Belge Stomatol",
  ["revue belge des sciences médicales"] = "Rev Belge Sci Med",
  ["revue belge d’archeologie et d’histoire de l’art"] = "Rev Belge Archeol Hist Art",
  ["revue belge d’archéologie et d’histoire de l’art"] = "RBA",
  ["revue belge d’histoire contemporaine. belgisch tijdschrift voor nieuwste geschiedenis"] = "Belg Tijdschr Nieuwste Geschied",
  ["revue belge d’homoeopathie"] = "Rev Belg Homoeopath",
  ["revue biblique"] = "RBi",
  ["revue bryologique et lichénologique"] = "Rev. bryol. lichénol.",
  ["revue bénédictine"] = "RBen",
  ["revue canadienne d'études du développement = canadian journal of development studies"] = "Rev Can Etudes Dev",
  ["revue canadienne de biologie"] = "Rev Can Biol",
  ["revue canadienne de biologie / editee par l’universite de montreal"] = "Rev Can Biol",
  ["revue canadienne de biologie experimentale"] = "Rev. Can. Biol. Exp.",
  ["revue canadienne de biologie expérimentale"] = "Rev Can Biol Exp",
  ["revue canadienne des sciences de l’administration"] = "Revue Can. Sci. Admin.",
  ["revue coloniale de médecine & chirurgie"] = "Rev Coloniale Med Chir",
  ["revue d epidemiologie et de sante publique"] = "Rev Epidemiol Sante Publique",
  ["revue d'allemagne"] = "Rev Allem",
  ["revue d'alsace"] = "Rev Alsace",
  ["revue d'archéologie moderne et d'archéologie générale : ramage"] = "Rev Archeol Mod Archeol Gen",
  ["revue d'assyriologie et d'archéologie orientale"] = "Rev Assyriol Archeol Orient",
  ["revue d'auvergne"] = "Rev Auvergne",
  ["revue d'ethnologie du québec"] = "Rev Ethnol Que",
  ["revue d'écologie"] = "Rev Ecol",
  ["revue d'économie du développement"] = "Rev Econ Dev",
  ["revue d'économie financière"] = "Rev Econ Financ",
  ["revue d'économie politique"] = "Rev Econ Polit",
  ["revue d'économie régionale et urbaine (adicueer (association))"] = "Rev Econ Reg Urbaine",
  ["revue d'égyptologie"] = "Rev Egyptol",
  ["revue d'électroencéphalographie et de neurophysiologie clinique"] = "Rev Electroencephalogr Neurophysiol Clin",
  ["revue d'élevage et de médecine vétérinaire des pays tropicaux"] = "Rev Elev Med Vet Pays Trop",
  ["revue d'épidémiologie et de santé publique"] = "Rev Epidemiol Sante Publique",
  ["revue d'épidémiologie, médecine sociale et santé publique"] = "Rev Epidemiol Med Soc Sante Publique",
  ["revue d'études comparatives est-ouest"] = "Rev Etud Comp Est Ouest",
  ["revue d'études palestiniennes : revue trimestrielle publiée par l'institut des études palestiniennes"] = "Rev Etud Palest",
  ["revue d'hématologie"] = "Rev Hematol",
  ["revue d'histoire de l'amérique française"] = "Rev Hist Am Fr",
  ["revue d'histoire de l'art dentaire"] = "Rev Hist Art Dent",
  ["revue d'histoire de l'église de france"] = "Rev Hist Eglise Fr",
  ["revue d'histoire de la deuxième guerre mondiale"] = "Rev Hist Deuxieme Guerre Mond",
  ["revue d'histoire de la médecine hébraique"] = "Rev Hist Med Heb",
  ["revue d'histoire de la pharmacie"] = "Rev Hist Pharm (Paris)",
  ["revue d'histoire des sciences"] = "Rev Hist Sci Paris",
  ["revue d'histoire des sciences et de leurs applications"] = "Rev Hist Sci (Paris)",
  ["revue d'histoire des textes"] = "Rev Hist Textes",
  ["revue d'histoire du théâtre"] = "Rev Hist Theatre",
  ["revue d'histoire ecclésiastique"] = "Rev Hist Eccles",
  ["revue d'histoire et de philosophie religieuses"] = "Rev Hist Philos Relig",
  ["revue d'histoire économique et sociale"] = "Rev Hist Econ Soc",
  ["revue d'histoire littéraire de la france"] = "Rev Hist Litt Fr",
  ["revue d'histoire maghrebine"] = "Rev Hist Maghreb",
  ["revue d'histoire moderne et contemporaine"] = "Rev Hist Mod Contemp",
  ["revue d'hygiène et de médecine sociale"] = "Rev Hyg Med Soc",
  ["revue d'immunologie"] = "Rev Immunol (Paris)",
  ["revue d'immunologie et de thérapie antimicrobienne"] = "Rev Immunol Ther Antimicrob",
  ["revue d'odonto-stomatologie"] = "Rev Odontostomatol (Paris)",
  ["revue d'odonto-stomatologie du midi de la france"] = "Rev Odontostomatol Midi Fr",
  ["revue d'orthopédie dento-faciale"] = "Rev Orthop Dento Faciale",
  ["revue d'orthopédie et de chirurgie de l'appareil moteur"] = "Rev Orthop Chir Appar Mot",
  ["revue d'oto-neuro-ophtalmologie"] = "Rev Otoneuroophtalmol",
  ["revue de chirurgie"] = "Rev Chir",
  ["revue de chirurgie orthopedique et reparatrice de l appareil moteur"] = "Rev Chir Orthop Reparatrice Appar Mot",
  ["revue de chirurgie orthopedique et reparatrice de l’appareil moteur"] = "Rev. Chir. Orthop. Reparatrice Appar. Mot.",
  ["revue de chirurgie orthopédique et réparatrice de l'appareil moteur"] = "Rev Chir Orthop Reparatrice Appar Mot",
  ["revue de comminges (pyrenees centrales)"] = "Rev Comminges",
  ["revue de comminges (pyrénées centrales)"] = "Rev Comminges",
  ["revue de coree"] = "Rev Coree",
  ["revue de corée"] = "Rev Coree",
  ["revue de cytologie clinique"] = "Rev Cytol Clin",
  ["revue de diététique"] = "Rev Diet",
  ["revue de droit"] = "Rev Droit",
  ["revue de geographie alpine / publiee par l’institute de geographie alpine (universite de grenoble)"] = "Rev Geogr Alp",
  ["revue de geographie de lyon"] = "Rev Geogr Lyon",
  ["revue de gerontologic d’expression francaise"] = "Rev Gerontol Expr Fr",
  ["revue de gerontologie d’expression francaise"] = "Rev. Gerontol. Expr. Fr.",
  ["revue de géographie alpine"] = "Rev Geogr Alp",
  ["revue de géographie de lyon"] = "Rev Geogr Lyon",
  ["revue de gérontologic d'expression française"] = "Rev Gerontol Expr Fr",
  ["revue de kinésithérapie"] = "Rev Kinesither",
  ["revue de l'académie polonaise des sciences. polska akademia nauk"] = "Rev Acad Pol Sci",
  ["revue de l'athérosclérose"] = "Rev Atheroscler",
  ["revue de l'athérosclérose et des artériopathies périphériques"] = "Rev Atheroscler (Paris)",
  ["revue de l'avranchin et du pays de granville"] = "Rev Avranchin Pays Granville",
  ["revue de l'enseignement philosophique"] = "Rev Enseign Philos",
  ["revue de l'enseignement supérieur"] = "Rev Enseign Super",
  ["revue de l'est"] = "Rev Est",
  ["revue de l'histoire des religions"] = "Rev Hist Relig",
  ["revue de l'hygiène professionnelle"] = "Rev Hyg Prof",
  ["revue de l'infirmière"] = "Rev Infirm",
  ["revue de l'infirmière et de l'assistante sociale"] = "Rev Infirm Assist Soc",
  ["revue de l'infirmière. informations"] = "Rev Infirm Inf",
  ["revue de l'institut d'hygiène des mines"] = "Rev Inst Hyg Mines (Hasselt)",
  ["revue de l'institut de sociologie"] = "Rev Inst Sociol",
  ["revue de l'institut napoléon (paris, france : 1954)"] = "Rev Inst Napoleon",
  ["revue de l'occident musulman et de la méditerranée"] = "Rev Occident Musulman Mediterr",
  ["revue de l'ouïe"] = "Rev Ouie",
  ["revue de l'université d'ottawa. university of ottawa"] = "Rev Univ Ottawa",
  ["revue de l'université de bruxelles"] = "Rev Univ Brux",
  ["revue de la bibliotheque nationale. bibliotheque nationale (france)"] = "Rev Bibl Natl",
  ["revue de la bibliothèque nationale. bibliothèque nationale (france)"] = "Rev Bibl Natl",
  ["revue de la haute-auvergne"] = "Rev Haute Auvergne",
  ["revue de la saintonge et de l'aunis"] = "Rev Saintonge Aunis",
  ["revue de la saintonge et de l’aunis"] = "RSaintonge",
  ["revue de la societe de biometrie humaine"] = "Rev Soc Biom Hum",
  ["revue de la societe francaise d’histoire de l’art dentaire"] = "Rev Soc Fr Hist Art Dent",
  ["revue de la societe francaise d’histoire des hopitaux"] = "Rev Soc Fr Hist Hop",
  ["revue de la société de biométrie humaine"] = "Rev Soc Biom Hum",
  ["revue de la société française d'histoire des hôpitaux"] = "Rev Soc Fr Hist Hop",
  ["revue de la société française d'histoire de l'art dentaire"] = "Rev Soc Fr Hist Art Dent",
  ["revue de la tuberculose"] = "Rev Tuberc",
  ["revue de laryngologie - otologie - rhinologie"] = "Rev Laryngol Otol Rhinol (Bord)",
  ["revue de laryngologie otologie rhinologie"] = "Rev. Laryngol. Otol. Rhinol. (Bord.)",
  ["revue de louisiane. louisiana review"] = "Rev Louisiane",
  ["revue de l’academie polonaise des sciences. polska akademia nauk"] = "Rev Acad Pol Sci",
  ["revue de l’atherosclerose"] = "Rev Atheroscler",
  ["revue de l’atherosclerose et des arteriopathies peripheriques"] = "Rev. Atheroscler. (Paris)",
  ["revue de l’avranchin et du pays de granville"] = "Rev Avranchin Pays Granville",
  ["revue de l’electricite et de l’electronique"] = "Rev. Electr. Electron.",
  ["revue de l’enseignement philosophique"] = "REPh",
  ["revue de l’enseignement superieur"] = "Rev Enseign Super",
  ["revue de l’est"] = "Rev Est",
  ["revue de l’histoire des religions"] = "RHR",
  ["revue de l’hygiene professionnelle"] = "Rev Hyg Prof",
  ["revue de l’infirmiere"] = "Rev. Infirm.",
  ["revue de l’infirmiere et de l’assistante sociale"] = "Rev. Infirm. Assist. Soc.",
  ["revue de l’infirmiere. informations"] = "Rev. Infirm. [Inf.]",
  ["revue de l’institut d hygiene des mines"] = "Rev. Inst. Hyg. Mines (Hasselt)",
  ["revue de l’institut de sociologie"] = "Rev Inst Sociol",
  ["revue de l’institut d’hygiene des mines"] = "Rev Inst Hyg Mines (Hasselt)",
  ["revue de l’institut napoleon (paris, france : 1954)"] = "Rev Inst Napoleon",
  ["revue de l’occident musulman et de la mediterranee"] = "Rev Occident Musulman Mediterr",
  ["revue de l’ofce"] = "Revue L’OFCE",
  ["revue de l’universite de bruxelles"] = "Rev Univ Brux",
  ["revue de l’universite d’ottawa. university of ottawa"] = "Rev Univ Ottawa",
  ["revue de l’université de bruxelles"] = "RUB",
  ["revue de medecine aeronautique"] = "Rev. Med. Aeronaut.",
  ["revue de medecine de toulouse"] = "Rev Med Toulouse",
  ["revue de medecine de tours"] = "Rev Med Tours",
  ["revue de medecine et d’hygiene d’outre-mer"] = "Rev Med Hyg Outre Mer",
  ["revue de medecine interne"] = "Rev. Med. Interne",
  ["revue de medecine psychosomatique (1985)"] = "Rev Med Psychosom",
  ["revue de medecine psychosomatique et de psychologie medicale"] = "Rev. Med. Psychosom. Psychol. Med.",
  ["revue de medecine veterinaire"] = "Rev Med Vet (Toulouse)",
  ["revue de metallurgie / cahiers d’informations techniques"] = "Rev. Metall. / Cah. Inf. Tech.",
  ["revue de metaphysique et de morale (paris, france : 1945)"] = "Rev Metaphys Morale",
  ["revue de médecine aéronautique"] = "Rev Med Aeronaut",
  ["revue de médecine de toulouse"] = "Rev Med Toulouse",
  ["revue de médecine de tours"] = "Rev Med Tours",
  ["revue de médecine et d'hygiène d'outre-mer"] = "Rev Med Hyg Outre Mer",
  ["revue de médecine navale (métropole et outre-mer); travaux scientifiques des médecins et pharmaciens-chimistes de la marine"] = "Rev Med Nav Metrop O-M",
  ["revue de médecine psychosomatique"] = "Rev Med Psychosom",
  ["revue de médecine psychosomatique et de psychologie médicale"] = "Rev Med Psychosom Psychol Med",
  ["revue de médecine vétérinaire"] = "Rev Med Vet (Toulouse)",
  ["revue de métaphysique et de morale (paris, france : 1945)"] = "Rev Metaphys Morale",
  ["revue de micropaleontologie"] = "Rev. Micropaleontol.",
  ["revue de micropaléontologie (online)"] = "Rev Micropaleontol",
  ["revue de micropaléontologie"] = "Rev. Micropaleontol./font>",
  ["revue de métaphysique et de morale"] = "RMM",
  ["revue de neuropsychiatrie infantile et d'hygiène mentale de l'enfance"] = "Rev Neuropsychiatr Infant",
  ["revue de neuropsychiatrie infantile et d’hygiene mentale de l’enfance"] = "Rev. Neuropsychiatr. Infant.",
  ["revue de paleobiologie"] = "Rev. Paleobiol.",
  ["revue de pathologie comparee"] = "Rev. Pathol. Comp.",
  ["revue de pathologie comparée"] = "Rev Pathol Comp",
  ["revue de pathologie generale et de physiologie clinique"] = "Rev Pathol Gen Physiol Clin",
  ["revue de pathologie générale et de physiologie clinique"] = "Rev Pathol Gen Physiol Clin",
  ["revue de pau et du bearn"] = "Rev Pau Bearn",
  ["revue de pau et du béarn"] = "Rev Pau Bearn",
  ["revue de philologie, de litterature et d’histoire anciennes"] = "Rev Philol Litt Hist Anc",
  ["revue de philologie, de littérature et d'histoire anciennes"] = "Rev Philol Litt Hist Anc",
  ["revue de philologie, de littérature et d’histoire anciennes"] = "RPhil",
  ["revue de philologie, de littérature et d’histoire anciennes"] = "RPh",
  ["revue de philosophie ancienne"] = "RPhA",
  ["revue de phytothérapie"] = "Rev Phytother (Paris)",
  ["revue de pneumologie clinique"] = "Rev Pneumol Clin",
  ["revue de psychoéducation"] = "Rev Psychoeduc",
  ["revue de psychologie appliquée"] = "Rev Psychol Appl",
  ["revue de recherches du service canadien des forêts"] = "Rev. rech. Serv. can. for.",
  ["revue de science financière"] = "Revue Sci. Financ.",
  ["revue de statistique appliquee"] = "Rev Stat Appl",
  ["revue de statistique appliquée"] = "Rev Stat Appl",
  ["revue de stomatologie"] = "Revue Stomatol",
  ["revue de stomatologie et de chirurgie maxillo-faciale"] = "Rev Stomatol Chir Maxillofac",
  ["revue de stomatologie, de chirurgie maxillo-faciale et de chirurgie orale"] = "Rev Stomatol Chir Maxillofac Chir Orale",
  ["revue de synthese / centre international de synthese"] = "Rev Synth",
  ["revue de synthese. 6e serie"] = "Rev. Synth. (6)",
  ["revue de synthese. 7e serie"] = "Rev. Synth. (7)",
  ["revue de synthèse"] = "Rev Synth",
  ["revue de synthèse"] = "RS",
  ["revue de théologie et de philosophie"] = "RThPh",
  ["revue de tuberculose et de pneumologie"] = "Rev Tuberc Pneumol (Paris)",
  ["revue dentaire libanaise (lebanese dental magazine)"] = "Rev. Dent. Liban.",
  ["revue dentaire libanaise. lebanese dental magazine"] = "Rev Dent Liban",
  ["revue des archéologues et historiens d’art de louvain"] = "RAArtLouv",
  ["revue des composites et des materiaux avances"] = "Rev. Compos. Mater. Av.",
  ["revue des corps de sante des armees, terre, mer, air"] = "Rev. Corps Sante Armees. Terre. Mer. Air",
  ["revue des corps de sante des armees: terre, mer, air, et du corps veterinaire"] = "Rev Corps Sante Armees Terre Mer Air",
  ["revue des corps de santé des armées terre, mer, air"] = "Rev Corps Sante Armees Terre Mer Air",
  ["revue des deux mondes (paris, france : 1948)"] = "Rev Deux Mondes",
  ["revue des eaux et forêts"] = "Rev. eaux for.",
  ["revue des etudes byzantines"] = "Rev Etud Byz",
  ["revue des etudes grecques"] = "Rev Etud Grec",
  ["revue des etudes italiennes"] = "Rev Etud Ital",
  ["revue des etudes juives"] = "Rev Etud Juives",
  ["revue des etudes latines"] = "Rev Etud Lat",
  ["revue des etudes slaves"] = "Rev Etud Slaves",
  ["revue des etudes sud-est europeennes"] = "Rev Etud Sud Est Eur",
  ["revue des études byzantines"] = "Rev Etud Byz",
  ["revue des études grecques"] = "Rev Etud Grec",
  ["revue des études italiennes"] = "Rev Etud Ital",
  ["revue des études juives"] = "Rev Etud Juives",
  ["revue des études latines"] = "Rev Etud Lat",
  ["revue des études slaves"] = "Rev Etud Slaves",
  ["revue des études sud-est européennes"] = "Rev Etud Sud Est Eur",
  ["revue des infirmieres et infirmiers auxiliaires du quebec"] = "Rev. Infirm. Infirm. Aux. Que.",
  ["revue des langues romanes"] = "Rev Lang Rom",
  ["revue des maladies respiratoires"] = "Rev Mal Respir",
  ["revue des maladies respiratoires actualités"] = "Rev Malad Respir Actual",
  ["revue des questions scientifiques"] = "Rev Quest Sci",
  ["revue des sciences de l'éducation"] = "Rev Sci Edu",
  ["revue des sciences de l’eau"] = "Rev. Sci. Eau",
  ["revue des sciences de l’education"] = "Rev Sci Edu",
  ["revue des sciences fondamentales et appliquées"] = "Rev Sci Fondam Appl",
  ["revue des sciences humaines"] = "Rev Sci Hum",
  ["revue des sciences medicales"] = "Rev Sci Med",
  ["revue des sciences médicales"] = "Rev Sci Med",
  ["revue des sciences morales & politiques : travaux de l'académie des sciences morales et politiques"] = "Rev Sci Morales Polit",
  ["revue des sciences morales & politiques : travaux de l’academie des sciences morales et politiques"] = "Rev Sci Morales Polit",
  ["revue des sciences philosophiques et theologiques"] = "Rev Sci Philos Theol",
  ["revue des sciences philosophiques et théologiques"] = "Rev Sci Philos Theol",
  ["revue des sciences philosophiques et théologiques"] = "RSPh",
  ["revue des sciences religieuses"] = "RSR",
  ["revue des sciences sociales de la france de l'est"] = "Rev Sci Soc Fr Est",
  ["revue des sciences sociales de la france de l’est"] = "Rev Sci Soc Fr Est",
  ["revue des societes savantes de haute-normandie"] = "Rev Soc Savantes Haute Normandie",
  ["revue des sociétés savantes de haute-normandie"] = "Rev Soc Savantes Haute Normandie",
  ["revue des travaux de l'académie des sciences morales et politiques et comptes rendus de ses séances"] = "Rev Trav Acad Sci Morales Polit C R Seances",
  ["revue des travaux de l’academie des sciences morales et politiques et comptes rendus de ses seances"] = "Rev Trav Acad Sci Morales Polit C R Seances",
  ["revue des études anciennes"] = "REA",
  ["revue des études arméniennes"] = "REArm",
  ["revue des études augustiniennes"] = "REAug",
  ["revue des études byzantines"] = "REByz",
  ["revue des études grecques"] = "REG",
  ["revue des études géorgiennes et caucasiennes"] = "RGeorgCauc",
  ["revue des études juives"] = "REJ",
  ["revue des études latines"] = "REL",
  ["revue des études sud-est européennes"] = "RESE",
  ["revue doptique, théorique et instrumentale"] = "Rev. Opt. Theor. Instrum.",
  ["revue dromoise"] = "Rev Dromoise",
  ["revue du bois et de ses applications"] = "Rev. bois appl.",
  ["revue du centre d'étude des pays de l'est et du centre national pour l'étude des états de l'est. brussels (belgium). université libre. centre d'étude des pays de l'est"] = "Rev Cent Etud Pays Est",
  ["revue du centre d’etude des pays de l’est et du centre national pour l’etude des etats de l’est. brussels (belgium). universite libre. centre d’etude des pays de l’est"] = "Rev Cent Etud Pays Est",
  ["revue du corps de sante militaire"] = "Rev Corps Sante Mil",
  ["revue du corps de santé militaire"] = "Rev Corps Sante Mil",
  ["revue du louvre. la revue des musées de france"] = "RLouvre",
  ["revue du monde musulman et de la méditerranée"] = "Rev Monde Musulman Mediterr",
  ["revue du nord"] = "Rev Nord",
  ["revue du paludisme et de medecine tropicale"] = "Rev Palud Med Trop",
  ["revue du paludisme et de médecine tropicale"] = "Rev Palud Med Trop",
  ["revue du praticien"] = "Rev. Prat.",
  ["revue du rhumatisme"] = "Rev Rhum",
  ["revue du rhumatisme (ed. francaise : 1993)"] = "Rev Rhum Ed Fr",
  ["revue du rhumatisme (ed. française : 1993)"] = "Rev Rhum Ed Fr",
  ["revue du rhumatisme (english ed.)"] = "Rev Rhum Engl Ed",
  ["revue du rhumatisme et des maladies osteo-articulaires"] = "Rev. Rhum. Mal. Osteoartic.",
  ["revue du rhumatisme et des maladies ostéo-articulaires"] = "Rev Rhum Mal Osteoartic",
  ["revue du rhumatisme. edition francaise"] = "Rev. Rhum. Ed. Fr.",
  ["revue du rhumatisme. english edition"] = "Rev. Rhum. Engl. Ed.",
  ["revue du rouergue"] = "Rev Rouergue",
  ["revue du sesda"] = "Rev SESDA",
  ["revue du tarn"] = "Rev Tarn",
  ["revue du vivarais"] = "Rev Vivarais",
  ["revue d’allemagne"] = "Rev Allem",
  ["revue d’alsace"] = "Rev Alsace",
  ["revue d’analyse economique"] = "Rev. Anal. Econ.",
  ["revue d’analyse numérique et de théorie de l’approximation"] = "Rev. Anal. Numér. Théor. Approx.",
  ["revue d’archeologie moderne et d’archeologie generale : ramage"] = "Rev Archeol Mod Archeol Gen",
  ["revue d’archéométrie"] = "RArcheom",
  ["revue d’archéometrie"] = "Rev. Archéom.",
  ["revue d’assyriologie et d’archeologie orientale"] = "Rev Assyriol Archeol Orient",
  ["revue d’assyriologie et d’archéologie orientale"] = "RAssyr",
  ["revue d’auvergne"] = "Rev Auvergne",
  ["revue d’economie du developpement"] = "Rev Econ Dev",
  ["revue d’economie du développement"] = "Revue Econ. Dével.",
  ["revue d’economie financière"] = "Revue Econ. Financ.",
  ["revue d’economie industrielle"] = "Revue Econ. Ind.",
  ["revue d’economie politique"] = "Revue Econ. Politique",
  ["revue d’economie regionale et urbaine"] = "Revue Econ. Reg. Urbaine",
  ["revue d’egyptologie"] = "Rev Egyptol",
  ["revue d’electroencephalographie et de neurophysiologie clinique"] = "Rev. Electroencephalogr. Neurophysiol. Clin.",
  ["revue d’elevage et de medecine veterinaire des pays tropicaux"] = "Rev. Elev. Med. Vet. Pays Trop.",
  ["revue d’epidemiologie et de sante publique"] = "Rev. Epidemiol. Sante Publique",
  ["revue d’epidemiologie, medecine sociale et sante publique"] = "Rev. Epidemiol. Med. Soc. Sante Publique",
  ["revue d’ethnologie du quebec"] = "Rev Ethnol Que",
  ["revue d’etudes comparatives est-ouest"] = "Revue Etud. Compar. Est-Ouest",
  ["revue d’égyptologie (kairo)"] = "RdE",
  ["revue d’hematologie"] = "Rev Hematol",
  ["revue d’histoire de la deuxieme guerre mondiale"] = "Rev Hist Deuxieme Guerre Mond",
  ["revue d’histoire de la medecine hebraique"] = "Rev Hist Med Heb",
  ["revue d’histoire de la pharmacie"] = "RHPHARM",
  ["revue d’histoire de l’amerique francaise"] = "Rev Hist Am Fr",
  ["revue d’histoire de l’art dentaire"] = "Rev. Hist. Art Dent.",
  ["revue d’histoire de l’eglise de france"] = "Rev Hist Eglise Fr",
  ["revue d’histoire des mathematiques. journal for the history of mathematics"] = "Rev. Histoire Math.",
  ["revue d’histoire des mathématiques"] = "Rev. Histoire Math.",
  ["revue d’histoire des sciences"] = "Rev. Hist. Sci.",
  ["revue d’histoire des sciences et de leurs applications"] = "RHS",
  ["revue d’histoire des textes"] = "RHT",
  ["revue d’histoire du droit (tijdschrift voor rechtsgeschiedenis)"] = "RHD",
  ["revue d’histoire ecclesiastique"] = "Rev Hist Eccles",
  ["revue d’histoire ecclésiastique"] = "RHE",
  ["revue d’histoire economique et sociale"] = "Rev Hist Econ Soc",
  ["revue d’histoire et de philosophie religieuses"] = "RHPhR",
  ["revue d’histoire litteraire de la france"] = "Rev Hist Litt Fr",
  ["revue d’histoire maghrebine"] = "Rev Hist Maghreb",
  ["revue d’histoire moderne et contemporaine"] = "Rev Hist Mod Contemp",
  ["revue d’hygiene et de medecine sociale"] = "Rev. Hyg. Med. Soc.",
  ["revue d’immunologie"] = "Rev. Immunol. (Paris)",
  ["revue d’immunologie et de therapie antimicrobienne"] = "Rev. Immunol. Ther. Antimicrob.",
  ["revue d’odonto-stomatologie"] = "Rev. Odontostomatol. (Paris)",
  ["revue d’odonto-stomatologie du midi de la france"] = "Rev. Odontostomatol. Midi Fr.",
  ["revue d’orthopedie dento-faciale"] = "Rev. Orthop. Dento. Faciale",
  ["revue d’orthopedie et de chirurgie de l’appareil moteur"] = "Rev Orthop Chir Appar Mot",
  ["revue d’oto-neuro-ophtalmologie"] = "Rev. Otoneuroophtalmol.",
  ["revue d’égyptologie"] = "REgypt",
  ["revue d’études ligures"] = "RELig",
  ["revue economique"] = "Revue Econ.",
  ["revue economique de madagascar"] = "Revue Econ. Madagascar",
  ["revue economique du sud-ouest"] = "Rev Econ Sud Ouest",
  ["revue economique et sociale"] = "Rev Econ Soc",
  ["revue europeenne de dermatologie et de mst"] = "Rev Eur Dermatol MST",
  ["revue europeenne des migrations internationales"] = "Rev Eur Migr Int",
  ["revue europeenne d’etudes cliniques et biologiques"] = "Rev. Eur. Etud. Clin. Biol.",
  ["revue europeenne d’etudes cliniques et biologiques. european journal of clinical and biological research"] = "Rev Eur Etud Clin Biol",
  ["revue européenne d'études cliniques et biologiques. european journal of clinical and biological research"] = "Rev Eur Etud Clin Biol",
  ["revue européenne de dermatologie et de mst"] = "Rev Eur Dermatol MST",
  ["revue européenne des migrations internationales"] = "Rev Eur Migr Int",
  ["revue européenne des éléments finis"] = "Rev. Européenne Élém. Finis",
  ["revue économique du sud-ouest"] = "Rev Econ Sud Ouest",
  ["revue économique et sociale"] = "Rev Econ Soc",
  ["revue finance contrôle stratégie"] = "Revue Finance Contrôle Stratégie",
  ["revue forestière française"] = "Rev. for. fr.",
  ["revue francaise de geotechnique"] = "Rev. Fr. Geotech.",
  ["revue francaise de gerontologie"] = "Rev Fr Gerontol",
  ["revue francaise de gynecologie et d’obstetrique"] = "Rev. Fr. Gynecol. Obstet.",
  ["revue francaise de la prothese dentaire"] = "Rev. Fr. Prothese Dent.",
  ["revue francaise de psychanalyse"] = "Rev. Fr. Psychanal.",
  ["revue francaise de science politique"] = "Rev Fr Sci Polit",
  ["revue francaise de sociologie"] = "Rev. Fr. Sociol.",
  ["revue francaise de transfusion"] = "Rev. Fr. Transfus.",
  ["revue francaise de transfusion et d’hemobiologie"] = "Rev. Fr. Transfus. Hemobiol.",
  ["revue francaise de transfusion et d’hemobiologie : bulletin de la societe nationale de transfusion sanguine"] = "Rev Fr Transfus Hemobiol",
  ["revue francaise de transfusion et immuno-hematologie"] = "Rev. Fr. Transfus. Immunohematol.",
  ["revue francaise des affaires sociales"] = "Rev Fr Aff Soc",
  ["revue francaise des maladies respiratoires"] = "Rev. Fr. Mal. Respir.",
  ["revue francaise des prothesistes dentaires"] = "Rev. Fr. Prothes. Dent.",
  ["revue francaise d’allergologie"] = "Rev. Fr. Allergol.",
  ["revue francaise d’endocrinologie clinique, nutrition et metabolisme"] = "Rev. Fr. Endocrinol. Clin.",
  ["revue francaise d’endodontie"] = "Rev. Fr. Endod.",
  ["revue francaise d’endodontie : publication officielle de la societe francaise d’endodontie"] = "Rev Fr Endod",
  ["revue francaise d’etudes americaines"] = "Rev Fr Etud Am",
  ["revue francaise d’etudes cliniques et biologiques"] = "Rev. Fr. Etud. Clin. Biol.",
  ["revue francaise d’histoire du livre"] = "Rev Fr Hist Livre",
  ["revue francaise d’histoire d’outre-mer / societe francaise d’histoire d’outre-mer"] = "Rev Fr Hist Outre Mer",
  ["revue francaise d’odonto-stomatologie"] = "Rev. Fr. Odontostomatol.",
  ["revue francaise d’oenologie"] = "Rev. Fr. Oenol.",
  ["revue francophone des laboratoires : rfl"] = "Rev Francoph Lab",
  ["revue française d'allergie"] = "Rev Fr Allerg",
  ["revue française d'allergologie"] = "Rev Fr Allergol",
  ["revue française d'allergologie (2009)"] = "Rev Fr Allergol (2009)",
  ["revue française d'allergologie et d'immunologie clinique"] = "Rev Fr Allergol Immunol Clin",
  ["revue française d'endodontie : publication officielle de la société française d'endodontie"] = "Rev Fr Endod",
  ["revue française d'études américaines"] = "Rev Fr Etud Am",
  ["revue française d'études cliniques et biologiques"] = "Rev Fr Etud Clin Biol",
  ["revue française d'histoire d'outre-mer"] = "Rev Fr Hist Outre Mer",
  ["revue française d'histoire du livre"] = "Rev Fr Hist Livre",
  ["revue française d'odonto-stomatologie"] = "Rev Fr Odontostomatol",
  ["revue française de gérontologie"] = "Rev Fr Gerontol",
  ["revue française de gynécologie et d'obstétrique"] = "Rev Fr Gynecol Obstet",
  ["revue française de psychanalyse"] = "Rev Fr Psychanal",
  ["revue française de science politique"] = "Rev Fr Sci Polit",
  ["revue française de sérologie et de chimiothérapie"] = "Rev Francaise Serol Chimiother",
  ["revue française de sociologie"] = "Rev Fr Sociol",
  ["revue française de transfusion"] = "Rev Fr Transfus",
  ["revue française de transfusion et d'hémobiologie : bulletin de la société nationale de transfusion sanguine"] = "Rev Fr Transfus Hemobiol",
  ["revue française de transfusion et immuno-hématologie"] = "Rev Fr Transfus Immunohematol",
  ["revue française des affaires sociales"] = "Rev Fr Aff Soc",
  ["revue française des maladies respiratoires"] = "Rev Fr Mal Respir",
  ["revue française des prothésistes dentaires"] = "Rev Fr Prothes Dent",
  ["revue française de gestion"] = "Revue Française Gestion",
  ["revue française d’economie"] = "Revue Française Econ.",
  ["revue generale de droit"] = "Rev Gen Droit",
  ["revue generale de thermique"] = "Rev. Gen. Therm.",
  ["revue generale des caoutchoucs & plastiques"] = "Rev. Gen. Caoutch. Plast.",
  ["revue generale des sciences pures et appliquees et bulletin de l’association francaise pour l’avancement des sciences"] = "Rev Gen Sci Pures Appl Bull Assoc Fr Av Sci",
  ["revue generale nucleaire"] = "Rev. Gen. Nucl.",
  ["revue geographique de l’est"] = "Rev Geogr Est",
  ["revue geographique des pyrenees et du sud-ouest"] = "Rev Geogr Pyren Sud Ouest",
  ["revue générale de droit"] = "Rev Gen Droit",
  ["revue générale de médecine & de chirurgie de l'union française"] = "Rev Gen Med Chir Union Fr",
  ["revue générale des sciences pures et appliquées et bulletin de l'association française pour l'avancement des sciences"] = "Rev Gen Sci Pures Appl Bull Assoc Fr Av Sci",
  ["revue géographique de l'est"] = "Rev Geogr Est",
  ["revue géographique des pyrénées et du sud-ouest"] = "Rev Geogr Pyren Sud Ouest",
  ["revue générale de botanique"] = "Rev. gen. Bot.",
  ["revue historique"] = "RH",
  ["revue historique (paris, france)"] = "Rev Hist (Paris)",
  ["revue historique de bordeaux et du departement de la gironde"] = "Rev Hist Bordeaux Dep Gironde",
  ["revue historique de bordeaux et du département de la gironde"] = "Rev Hist Bordeaux Dep Gironde",
  ["revue historique de droit français et étranger"] = "RD",
  ["revue historique des armees"] = "Rev Hist Armees",
  ["revue historique des armées"] = "Rev Hist Armees",
  ["revue historique des armées"] = "RHistArmees",
  ["revue historique vaudoise"] = "RHV",
  ["revue hittite et asianique"] = "RHA",
  ["revue informatique et statistique dans les sciences humaines"] = "RISSH",
  ["revue international des services de sante des armees de terre, de mer et de l’air"] = "Rev Int Serv Sante Armees",
  ["revue international des services de santé des armées de terre, de mer et de l'air"] = "Rev Int Serv Sante Armees",
  ["revue internationale d'hépatologie"] = "Rev Int Hepatol",
  ["revue internationale d'histoire de la psychanalyse"] = "Rev Int Hist Psychanal",
  ["revue internationale d'histoire de la psychiatrie"] = "Rev Int Hist Psychiatr",
  ["revue internationale d'histoire militaire"] = "Rev Int Hist Mil",
  ["revue internationale de criminologie et de police technique"] = "Rev Int Criminol Police Tech",
  ["revue internationale de la croix-rouge"] = "Rev Int Croix Rouge",
  ["revue internationale de pediatrie"] = "Rev Int Pediatr",
  ["revue internationale de pédiatrie"] = "Rev Int Pediatr",
  ["revue internationale de philosophie"] = "Rev Int Philos",
  ["revue internationale des droits de l’antiquité"] = "RDroitsAnt",
  ["revue internationale des droits de l’antiquité"] = "RIDA",
  ["revue internationale des services de sante des armees de terre, de mer et de l air"] = "Rev. Int. Serv. Sante Armees",
  ["revue internationale des services de santé des forces armées : organe du comité international de médecine et de pharmacie militaires = international review of the armed forces medical services : organ of the international committee of military medicine and pharmacy"] = "Rev Int Serv Sante Forces Armees",
  ["revue internationale du froid"] = "Int J Refrig",
  ["revue internationale du trachome"] = "Rev. Int. Trach.",
  ["revue internationale du trachome et de pathologie oculaire tropicale et subtropicale"] = "Rev. Int. Trach. Pathol. Ocul. Trop. Subtrop.",
  ["revue internationale du trachome et de pathologie oculaire tropicale et subtropicale : organe de la ligue contre le trachome avec la collaboration de l'international organization against trachoma et des organisations nationales et internationales de santé publique"] = "Rev Int Trach Pathol Ocul Trop Subtrop",
  ["revue internationale du trachome et de pathologie oculaire tropicale et subtropicale et de sante publique"] = "Rev. Int. Trach. Pathol. Ocul. Trop. Subtrop. Sante Publique",
  ["revue internationale du trachome et de pathologie oculaire tropicale et subtropicale et de santé publique : organe de la ligue contre le trachome avec la collaboration de l'international organization against trachoma et des organisations nationales et internationales de santé publique"] = "Rev Int Trach Pathol Ocul Trop Subtrop Sante Publique",
  ["revue internationale du trachome et de pathologie oculaire tropicale etsubtropicale et de sante publique rev int trach pathol ocul trop subtro"] = "Sante Publique",
  ["revue internationale du trachome. international review of trachoma"] = "Rev Int Trach",
  ["revue internationale d’hepatologie"] = "Rev. Int. Hepatol.",
  ["revue internationale d’histoire de la psychanalyse"] = "Rev Int Hist Psychanal",
  ["revue internationale d’histoire de la psychiatrie"] = "Rev Int Hist Psychiatr",
  ["revue internationale d’histoire militaire"] = "Rev Int Hist Mil",
  ["revue internationale d’onomastique"] = "RIO",
  ["revue internationale geomatique"] = "Revue Int. Geomatique",
  ["revue lyonnaise de medecine"] = "Rev. Lyon. Med.",
  ["revue malienne d'infectiologie et de microbiologie"] = "Rev Mali Infect Microbiol",
  ["revue medicale de bruxelles"] = "Rev Med Brux",
  ["revue medicale de la suisse romande"] = "Rev Med Suisse Romande",
  ["revue medicale de liege"] = "Rev Med Liege",
  ["revue medicale de louvain"] = "Rev Med Louvain",
  ["revue medicale de nancy"] = "Rev Med Nancy",
  ["revue medicale des alpes francaises"] = "Rev Med Alpes Fr",
  ["revue medicale du moyen-orient"] = "Rev. Med. Moyen. Orient",
  ["revue medicale internationale de photo, cinema, television (international journal of medical photography, cinematography and television)"] = "Rev. Med. Int. Photo Cinema Telev.",
  ["revue medicale internationale de photo, cinema, television. international journal of medical photography, cinematography & television"] = "Rev Med Int Photo Cinema Telev",
  ["revue medicale miniere"] = "Rev Med Miniere",
  ["revue medicale suisse"] = "Rev. Med. Suisse",
  ["revue medico-chirurgicale des maladies du foie"] = "Rev. Med. Chir. Mal. Foie",
  ["revue mediterraneenne des sciences medicales"] = "Rev Mediterr Sci Med",
  ["revue mensuelle des musées et collections de la ville genève"] = "Rev. mens. Mus. Collect. ville Genève",
  ["revue médicale de bruxelles"] = "Rev Med Brux",
  ["revue médicale de france"] = "Rev Med Fr",
  ["revue médicale de la suisse romande"] = "Rev Med Suisse Romande",
  ["revue médicale de liège"] = "Rev Med Liege",
  ["revue médicale de louvain"] = "Rev Med Louvain",
  ["revue médicale de nancy"] = "Rev Med Nancy",
  ["revue médicale des alpes françaises"] = "Rev Med Alpes Fr",
  ["revue médicale du moyen-orient"] = "Rev Med Moyen Orient",
  ["revue médicale internationale de photo, cinéma, télévision. international journal of medical photography, cinematography & television"] = "Rev Med Int Photo Cinema Telev",
  ["revue médicale minière"] = "Rev Med Miniere",
  ["revue médicale suisse"] = "Rev Med Suisse",
  ["revue médico-chirurgicale des maladies du foie"] = "Rev Med Chir Mal Foie",
  ["revue méditerranéenne des sciences médicales"] = "Rev Mediterr Sci Med",
  ["revue neurologique"] = "Rev Neurol (Paris)",
  ["revue nouvelle (brussels, belgium)"] = "Rev Nouv",
  ["revue numismatique"] = "Rev Numis",
  ["revue odonto stomatologique du nord-est"] = "Rev Odontostomatol Nordest",
  ["revue odonto-implantologique"] = "Rev Odontoimplantol",
  ["revue odonto-stomatologique du nord-est"] = "Rev. Odontostomatol. Nordest.",
  ["revue philosophique de la france et de l'étranger"] = "Rev Philos France Let",
  ["revue philosophique de la france et de l’étranger"] = "RPhilos",
  ["revue philosophique de louvain"] = "Rev Philos Louv",
  ["revue roumaine d'endocrinologie"] = "Rev Roum Endocrinol",
  ["revue roumaine d'histoire"] = "Rev Roum Hist",
  ["revue roumaine d'inframicrobiologie"] = "Rev Roum Inframicrobiol",
  ["revue roumaine de chimie"] = "Rev. Roum. Chim.",
  ["revue roumaine de geophysique"] = "Rev. Roum. Geophys.",
  ["revue roumaine de mathematiques pures et appliquees. romanian journal of pure and applied mathematics"] = "Rev. Roumaine Math. Pures Appl.",
  ["revue roumaine de mathématiques pures et appliquées"] = "Rev. Roumaine Math. Pures Appl.",
  ["revue roumaine de medecine"] = "Rev. Roum. Med.",
  ["revue roumaine de medecine interne"] = "Rev. Roum. Med. Intern.",
  ["revue roumaine de médecine (1974)"] = "Rev Roum Med",
  ["revue roumaine de médecine interne (1964)"] = "Rev Roum Med Intern",
  ["revue roumaine de morphologie et d'embryologie"] = "Rev Roum Morphol Embryol",
  ["revue roumaine de morphologie et de physiologie"] = "Rev Roum Morphol Physiol",
  ["revue roumaine de morphologie et d’embryologie"] = "Rev. Roum. Morphol. Embryol.",
  ["revue roumaine de neurologie"] = "Rev Roum Neurol",
  ["revue roumaine de neurologie et de psychiatrie"] = "Rev Roum Neurol Psychiatr",
  ["revue roumaine de physiologie"] = "Rev Roum Physiol",
  ["revue roumaine de physiologie (bucharest, romania : 1990)"] = "Rev Roum Physiol  (1990)",
  ["revue roumaine de virologie"] = "Rev Roum Virol",
  ["revue roumaine de virologie (bucharest, romania : 1990)"] = "Rev Roum Virol",
  ["revue roumaine des sciences sociales; série sciences economiques"] = "Revue Roumaine Sci. Soc. Série Sci. Econ.",
  ["revue roumaine des sciences techniques"] = "Rev. Roumaine Sci. Tech. Sér. Méc. Appl.",
  ["revue roumaine du chimie"] = "Rev Roum Chim",
  ["revue roumaine d’endocrinologie"] = "Rev. Roum. Endocrinol.",
  ["revue roumaine d’histoire"] = "Rev Roum Hist",
  ["revue roumaine d’inframicrobiologie"] = "Rev. Roum. Inframicrobiol.",
  ["revue science et technique. série sciences humaines"] = "Rev Sci Tech Ser Sci Hum",
  ["revue scientifique et technique"] = "Rev Sci Tech",
  ["revue scientifique et technique (international office of epizootics)"] = "Rev Sci Tech",
  ["revue scientifique et technique, office international des epizooties"] = "Rev. Sci. Tech.",
  ["revue stomato-odontologique du nord de la france"] = "Rev. Stomatoodontol. Nord Fr.",
  ["revue suisse de viticulture arboriculture horticulture"] = "Rev. suisse vitic. arboric. hortic.",
  ["revue suisse de viticulture et aboriculture"] = "Rev. Suisse Vitic. Aboric.",
  ["revue suisse de zoologie"] = "Rev. Suisse Zool.",
  ["revue suisse de zoologie; annales de la societe zoologique suisse et du museum d’histoire naturelle de geneve"] = "Rev Suisse Zool",
  ["revue suisse de zoologie; annales de la société zoologique suisse et du muséum d'histoire naturelle de genève"] = "Rev Suisse Zool",
  ["revue suisse d’agriculture"] = "Rev. Suisse Agric.",
  ["revue suisse d’art et d’archéologie"] = "RSAA",
  ["revue suisse d’hydrologie"] = "Rev. suisse Hydrol.",
  ["revue théologique de louvain"] = "RThL",
  ["revue tiers monde"] = "Revue Tiers Monde",
  ["revue tiers-monde"] = "Tiers Monde",
  ["revue tunisienne de geographie"] = "Rev Tunis Geogr",
  ["revue tunisienne de géographie"] = "Rev Tunis Geogr",
  ["revue tunisienne de sciences sociales"] = "Rev Tunis Sci Soc",
  ["revue tunisienne des etudes de population"] = "Rev Tunis Etud Popul",
  ["revue tunisienne des études de population"] = "Rev Tunis Etud Popul",
  ["revue vétérinaire militaire"] = "Rev Vet Mil (Paris)",
  ["revue égyptologique (paris)"] = "RevEg",
  ["revue. palais de la decouverte (paris, france)"] = "Rev Palais Decouv",
  ["revue. palais de la découverte (paris, france)"] = "Rev Palais Decouv",
  ["rezultaty issledovanii po mezhdunarodnym proektam"] = "Rezultaty Issled. Mezhdunar. Geofiz. Proekt.",
  ["réalités cliniques : revue européenne d'odontologie"] = "Real Clin",
  ["réalités. edition médecine"] = "Realites",
  ["réanimation : journal de la société de réanimation de langue française"] = "Reanimation",
  ["réanimation et organes artificiels. wiederbelebung und künstliche organe. reanimation and artificial organs"] = "Reanim Organes Artif",
  ["rééducation orthophonique"] = "Reeduc Orthoph",
  ["références en gynécologie obstétrique"] = "Ref Gynecol Obstet",
  ["répertoire d’épigraphie sémitique (paris 1900–1950)"] = "RES",
  ["réseaux - ciephum"] = "Reseaux Ciephum",
  ["rf illustrated"] = "RF Illus",
  ["rguhs journal of pharmaceutical sciences"] = "RGUHS J Pharm Sci",
  ["rheedea : official journal of indian association for angiosperm taxonomy"] = "Rheedea",
  ["rheinische lebensbilder"] = "Rheinische Lebensbilder",
  ["rheinische vierteljahrsblatter"] = "Rhein Vierteljahrsbl",
  ["rheinische vierteljahrsblätter"] = "Rhein Vierteljahrsbl",
  ["rheinisches arzteblatt"] = "Rheinisches Arztebl",
  ["rheinisches ärzteblatt"] = "Rheinisches Arztebl",
  ["rheinisches museum fur philologie"] = "Rhein Mus Philol",
  ["rheinisches museum für philologie"] = "Rhein Mus Philol",
  ["rheinisches museum für philologie"] = "RhM",
  ["rheologica acta"] = "Rheol. Acta",
  ["rhetoric society quarterly"] = "Rhetor Soc Q",
  ["rhetoric, professional communication, and globalization"] = "Rhetor Prof Commun Glob",
  ["rheuma plus"] = "Rheuma Plus",
  ["rheumatic diseases clinics of north america"] = "Rheum Dis Clin North Am",
  ["rheumatologia, balneologia, allergologia"] = "Rheumatol Balneol Allergol",
  ["rheumatology (oxford, england)"] = "Rheumatology (Oxford)",
  ["rheumatology (sunnyvale, calif.)"] = "Rheumatology (Sunnyvale)",
  ["rheumatology advances in practice"] = "Rheumatol Adv Pract",
  ["rheumatology and physical medicine"] = "Rheumatol Phys Med",
  ["rheumatology and rehabilitation"] = "Rheumatol Rehabil",
  ["rheumatology and therapy"] = "Rheumatol Ther",
  ["rheumatology international"] = "Rheumatol Int",
  ["rhinology online"] = "Rhinol Online",
  ["rhinology. supplement"] = "Rhinol Suppl",
  ["rhode island dental journal"] = "R I Dent J",
  ["rhode island history"] = "R I Hist",
  ["rhode island medical journal"] = "R I Med J",
  ["rhode island medical journal (2013)"] = "R I Med J (2013)",
  ["rhode island medicine"] = "R I Med",
  ["rhodesia nurse"] = "Rhod. Nurse",
  ["rhodesian nurse"] = "Rhod. Nurse",
  ["riai  revista iberoamericana de automatica e informatica industrial"] = "Rev. Iberoam. Autom. Inf. Ind.",
  ["riai revista iberoamericana de automatica e informatica industrial"] = "Rev. Iberoam. Autom. Inf. Ind.",
  ["ric reviews"] = "RIC Rev.",
  ["rice (new york, n.y.)"] = "Rice (N Y)",
  ["rice biotechnology quarterly"] = "Rice Biotechnol Q",
  ["rice genetics newsletter"] = "Rice Genet Newsl",
  ["rice genomics and genetics"] = "Rice Genom Genet",
  ["rice science"] = "Rice Sci.",
  ["ricerca in clinica e in laboratorio"] = "Ric. Clin. Lab.",
  ["ricerca scientifica"] = "Ric. Sci.",
  ["ricerca scientifica. 2 ser. parte 1, rivista"] = "Ric. Sci. 2 Ser. Pt. 1 Riv.",
  ["ricerca scientifica. 2 ser. parte 2, rendiconti. sezione b, biologica"] = "Ric. Sci. 2 Ser. Pt 2 Rend. [B]",
  ["ricerche di egittologia e di antichità copte"] = "RicEgAntCopt",
  ["ricerche di matematica"] = "Ric. Mat.",
  ["ricerche di matematica. a journal of pure and applied mathematics"] = "Ric. Mat.",
  ["ricerche e studi. museo francesco ribezzo, brindisi"] = "RicStBrindisi",
  ["ricerche economiche"] = "Ricerche Econ.",
  ["ricerche teologiche"] = "RicTeol",
  ["richmond county history"] = "Richmond Cty Hist",
  ["ricognizioni archeologiche"] = "RicognA",
  ["riforma medica"] = "Riforma Med.",
  ["riggisberger berichte"] = "RigBer",
  ["rim monographs in mathematics"] = "RIM Monogr. Math.",
  ["rinascimento (florence, italy)"] = "Rinascimento",
  ["rinascimento: rivista dell’istituto nazionale di studi sul rinascimento"] = "Rinascimento",
  ["ringing & migration"] = "Ringing Migr.",
  ["rinsho biseibutshu jinsoku shindan kenkyukai shi (journal of the association for rapid method and automation in microbiology)"] = "Rinsho Biseibutshu Jinsoku Shindan Kenkyukai Shi",
  ["rinsho byori (japanese journal of clinical pathology)"] = "Rinsho Byori",
  ["rinsho byori. japanese journal of clinical pathology"] = "Rinsho Byori",
  ["rinsho byori. the japanese journal of clinical pathology"] = "Rinsho Byori",
  ["rinsho ganka. japanese journal of clinical ophthalmology"] = "Rinsho Ganka",
  ["rinsho geka. journal of clinical surgery"] = "Rinsho Geka",
  ["rinsho hoshasen (japanese journal of clinical radiology)"] = "Rinsho Hoshasen",
  ["rinsho hoshasen. clinical radiography"] = "Rinsho Hoshasen",
  ["rinsho ketsueki (japanese journal of clinical hematology)"] = "Rinsho Ketsueki",
  ["rinsho ketsueki. japanese journal of clinical hematology"] = "Rinsho Ketsueki",
  ["rinsho kyobu geka (japanese annals of thoracic surgery)"] = "Rinsho Kyobu Geka",
  ["rinsho naika shonika. internal medicine and pediatrics"] = "Rinsho Naika Shonika Intern Med",
  ["rinsho shika (clinical dentistry)"] = "Rinsho Shika",
  ["rinsho shinkeigaku (clinical neurology)"] = "Rinsho Shinkeigaku",
  ["rinsho shinkeigaku. clinical neurology"] = "Rinsho Shinkeigaku",
  ["rinsho shishubyo danwakai kaishi"] = "Rinsho Shishubyo Danwakai Kaishi",
  ["rinsho to uirusu. clinical virology"] = "Rinsho To Uirusu",
  ["rinsho yakuri"] = "Rinsho Yakuri",
  ["rinshō biseibutsu jinsoku shindan kenkyūkai shi = jarmam : journal of the association for rapid method and automation in microbiology"] = "Rinsho Biseibutshu Jinsoku Shindan Kenkyukai Shi",
  ["rinshō kyōbu geka = japanese annals of thoracic surgery"] = "Rinsho Kyobu Geka",
  ["rinshō shinkeigaku = clinical neurology"] = "Rinsho Shinkeigaku",
  ["rinshō shishūbyō danwakai kaishi"] = "Rinsho Shishubyo Danwakai Kaishi",
  ["riogrande odontológico"] = "Riogrande Odontol",
  ["ripon forum"] = "Ripon Forum",
  ["risec: international review of economics and business"] = "RISEC: Int. Rev. Econ. Bus.",
  ["risk (concord, nh)"] = "Risk",
  ["risk analysis"] = "Risk Anal.",
  ["risk analysis : an official publication of the society for risk analysis"] = "Risk Anal",
  ["risk decision and policy"] = "Risk Dec. Pol.",
  ["risk engineering"] = "Risk Eng.",
  ["risk management"] = "Risk Manage.",
  ["risk management (new york, n.y.)"] = "Risk Manage",
  ["risk management and healthcare policy"] = "Risk Manag Healthc Policy",
  ["risk management in canadian health care"] = "Risk Manag Can Health Care",
  ["risk, hazards & crisis in public policy"] = "Risk Hazards Crisis Public Policy",
  ["risk, issues in health & safety"] = "Risk Issues Health Saf",
  ["riso report"] = "RISO Rep.",
  ["risorgimento (milan, italy)"] = "Risorgimento",
  ["risø report"] = "RISO Rep",
  ["rittenhouse : journal of the american scientific instrument enterprise"] = "Rittenhouse",
  ["river publishers series in mathematical and engineering sciences"] = "River Publ. Ser. Math. Eng. Sci.",
  ["river publishers series in mathematical, statistical and computational modelling for engineering"] = "River Publ. Ser. Math. Stat. Comput. Model. Eng.",
  ["river publishers series in software engineering"] = "River Publ. Ser. Softw. Eng.",
  ["river research and applications"] = "River Res. Appl.",
  ["river systems"] = "River Syst.",
  ["rivista archeologica dell’antica provincia e diocesi di como"] = "RAComo",
  ["rivista biblica"] = "RivBibl",
  ["rivista critica di clinica medica"] = "Riv Crit Clin Med",
  ["rivista critica di storia della filosofia"] = "Riv Crit Stor Filos",
  ["rivista d'ostetricia e ginecologia pratica"] = "Riv Ostet Ginecol Prat",
  ["rivista degli infortuni e delle malattie professionali"] = "Riv Infort Mal Prof",
  ["rivista degli studi orientali"] = "Riv Studi Orient",
  ["rivista dei combustibili e dell’industia chimica"] = "Riv. Combust. Ind. Chim.",
  ["rivista del nuovo cimento"] = "Riv. Nuovo Cimento",
  ["rivista dell'infermiere"] = "Riv Inferm",
  ["rivista dell'istituto sieroterapico italiano"] = "Riv Ist Sieroter Ital",
  ["rivista dell'istituto vaccinogeno e dei consorzi provinciali antitubercolari"] = "Riv Ist Vaccinogeno Consorzi Prov Antituberc",
  ["rivista della tubercolosi e delle malattie dell'apparato respiratorio"] = "Riv Tuberc Mal Appar Respir",
  ["rivista della tubercolosi e delle malattie dell’apparato respiratorio"] = "Riv Tuberc Mal Appar Respir",
  ["rivista dell’infermiere"] = "Riv. Inferm.",
  ["rivista dell’istituto nazionale di archeologia e storia dell’arte"] = "RIA",
  ["rivista dell’istituto nazionale d’archeologia e storia dell’arte"] = "RIA",
  ["rivista dell’istituto sieroterapico italiano"] = "Riv. Ist. Sieroter. Ital.",
  ["rivista dell’istituto vaccinogeno e dei consorzi provinciali antitubercolari"] = "Riv Ist Vaccinogeno Consorzi Prov Antituberc",
  ["rivista di anatomia patologica e di oncologia"] = "Riv Anat Patol Oncol",
  ["rivista di archeologia"] = "RdA",
  ["rivista di archeologia cristiana"] = "RAC",
  ["rivista di biologia"] = "Riv Biol",
  ["rivista di biologia coloniale"] = "Riv Biol Coloniale",
  ["rivista di chirurgia e medicina"] = "Riv Chir Med",
  ["rivista di chirurgia pediatrica"] = "Riv Chir Pediatr",
  ["rivista di clinica pediatrica"] = "Riv Clin Pediatr",
  ["rivista di cultura classica e medioevale"] = "RCCM",
  ["rivista di difesa sociale"] = "Riv Dif Soc",
  ["rivista di emoterapia ed immunoematologia"] = "Riv Emoter Immunoematol",
  ["rivista di estetica"] = "Riv Estet",
  ["rivista di filologia e d'istruzione classica"] = "Riv Filol Istruz Classica",
  ["rivista di filologia e di istruzione classica"] = "RFIC",
  ["rivista di filologia e d’istruzione classica"] = "Riv Filol Istruz Classica",
  ["rivista di filosofia"] = "Riv Filos",
  ["rivista di filosofia neo-scolastica"] = "Riv Filos Neo Scolast",
  ["rivista di filosofia neoscolastica"] = "RFN",
  ["rivista di gastro-enterologia"] = "Riv Gastroenterol",
  ["rivista di istochimica, normale e patologica"] = "Riv Istochim Norm Patol",
  ["rivista di linguistica"] = "Riv Linguist",
  ["rivista di malariologia"] = "Riv Malariol",
  ["rivista di matematica della universita di parma. new series. a journal of pure and applied mathematics"] = "Riv. Math. Univ. Parma (N.S.)",
  ["rivista di matematica della università di parma"] = "Riv. Mat. Univ. Parma (5)",
  ["rivista di matematica per le scienze economiche e sociali"] = "Riv. Mat. Sci. Econom. Social.",
  ["rivista di matematica pura ed applicata"] = "Riv. Mat. Pura Appl.",
  ["rivista di medicina aeronautica"] = "Riv Med Aeronaut",
  ["rivista di medicina aeronautica e spaziale"] = "Riv Med Aeronaut Spaz",
  ["rivista di medicina legale e legislazione sanitaria"] = "Riv Med Leg Legis Sanit",
  ["rivista di meteorologia aeronautica"] = "Riv Meteorol Aeronaut",
  ["rivista di micologia : bollettino dell'associazione micologica bresadola"] = "Riv Micol",
  ["rivista di neurobiologia"] = "Riv. Neurobiol.",
  ["rivista di neurobiologia : organo ufficiale della societa dei neurologi, neuroradiologi e neurochirurghi ospedalieri"] = "Riv Neurobiol",
  ["rivista di neurobiologia : organo ufficiale della società dei neurologi, neuroradiologi e neurochirurghi ospedalieri"] = "Riv Neurobiol",
  ["rivista di neurologia"] = "Riv Neurol",
  ["rivista di odontoiatria degli amici di brugg"] = "Riv Odontoiatr Amici Brugg",
  ["rivista di odontojatria e ortognatodonzia"] = "Riv Odontojatria Ortognatod",
  ["rivista di odontostomatologia e implantoprotesi"] = "Riv Odontostomatol Implantoprotesi",
  ["rivista di ostetricia e ginecologia"] = "Riv Ostet Ginecol",
  ["rivista di ostetricia ginecologia pratica e medicina perinatale"] = "Riv Ostet Ginecol Prat Med Perinat",
  ["rivista di parassitologia"] = "Riv Parassitol",
  ["rivista di patologia clinica e sperimentale"] = "Riv Patol Clin Sper",
  ["rivista di patologia e clinica"] = "Riv Patol Clin",
  ["rivista di patologia e clinica della tubercolosi"] = "Riv Patol Clin Tuberc",
  ["rivista di patologia nervosa e mentale"] = "Riv Patol Nerv Ment",
  ["rivista di politica economica"] = "Rivista Politica Econ.",
  ["rivista di psichiatria"] = "Riv Psichiatr",
  ["rivista di psicoanalisi"] = "Riv Psicoanal",
  ["rivista di psicolinguistica applicata"] = "Riv Psicolinguist Appl",
  ["rivista di psicologia"] = "Riv Psicol (1912)",
  ["rivista di psicologia (1945)"] = "Riv Psicol (1945)",
  ["rivista di psicologia (1955)"] = "Riv Psicol (1955)",
  ["rivista di psicologia della scrittura"] = "Riv Psicol Scr",
  ["rivista di psicologia sociale e archivio italiano di psicologia generale e del lavoro"] = "Riv Psicol Soc Arch Ital Psicol Gen",
  ["rivista di radiologia"] = "Riv Radiol",
  ["rivista di scienze preistoriche"] = "RScPreist",
  ["rivista di statistica applicata"] = "Rivista Statist. Appl.",
  ["rivista di storia della chiesa in italia"] = "Riv Stor Chiesa Ital",
  ["rivista di storia della filosofia (milan, italy : 1984)"] = "Riv Stor Filos",
  ["rivista di storia della medicina"] = "Riv Stor Med",
  ["rivista di storia della scienza"] = "Riv Stor Sci",
  ["rivista di storia delle scienze mediche e naturali"] = "Riv Stor Sci Mediche Nat",
  ["rivista di storia e letteratura religiosa"] = "RSLR",
  ["rivista di storia economica, n.s."] = "Rivista Storia Econ.",
  ["rivista di studi bizantini e neoellenici"] = "RSBN",
  ["rivista di studi bizantini e slavi"] = "RSBS",
  ["rivista di studi classici"] = "Riv Studi Class",
  ["rivista di studi fenici"] = "RStFen",
  ["rivista di studi fenici, pubbl. dal centro di studio per la civiltà fenicia e punica"] = "RStudFen",
  ["rivista di studi liguri"] = "RStLig",
  ["rivista di studi marchigiani"] = "RStMarch",
  ["rivista di studi pompeiani"] = "RSP",
  ["rivista di studi punici"] = "RStPun",
  ["rivista di topografia antica"] = "RTopAnt",
  ["rivista d’ostetricia e ginecologia pratica"] = "Riv. Ostet. Ginecol. Prat.",
  ["rivista europea per le scienze mediche e farmacologiche"] = "Riv. Eur. Sci. Med. Farmacol.",
  ["rivista europea per le scienze mediche e farmacologiche = european review for medical and pharmacological sciences = revue européenne pour les sciences médicales et pharmacologiques"] = "Riv Eur Sci Med Farmacol",
  ["rivista geografica italiana"] = "Riv Geogr Ital",
  ["rivista ingauna e intemelia"] = "RIngIntem",
  ["rivista internazionale di filosofia del diritto"] = "RIFD",
  ["rivista internazionale di filosofia e psicologia"] = "Riv Internazionale Filos Psicol",
  ["rivista internazionale di psicologia e ipnosi"] = "Riv Int Psicol Ipn",
  ["rivista internazionale di scienze economiche e commerciali"] = "Rivista Int. Sci. Econ. Com.",
  ["rivista internazionale di scienze sociali"] = "Riv Int Sci Sociali",
  ["rivista italiana d'igiene"] = "Riv Ital Ig",
  ["rivista italiana degli economisti"] = "Rivista Italiana Econ.",
  ["rivista italiana degli odontotecnici"] = "Riv Ital Odontotec",
  ["rivista italiana del tracoma e di patologia oculare esotica"] = "Riv Ital Trac Patol Ocul Esotica",
  ["rivista italiana della medicina di laboratorio"] = "Riv. Ital. Med. Lab.",
  ["rivista italiana di economia, demografia e statistica"] = "Riv Ital Econ Demogr Stat",
  ["rivista italiana di ginecologia"] = "Riv Ital Ginecol",
  ["rivista italiana di medicina legale : dottrina, casistica, ricerca sperimentale, giurisprudenza e legislazione"] = "Riv Ital Med Leg",
  ["rivista italiana di numismatica e scienze affini"] = "RIN",
  ["rivista italiana di odontoiatria infantile"] = "Riv. Ital. Odontoiatr. Infant.",
  ["rivista italiana di odontoiatria infantile : organo ufficiale della societa italiana di odontoiatria infantile"] = "Riv Ital Odontoiatr Infant",
  ["rivista italiana di odontoiatria infantile : organo ufficiale della società italiana di odontoiatria infantile"] = "Riv Ital Odontoiatr Infant",
  ["rivista italiana di paleontologia e stratigrafia"] = "Riv. Ital. Paleontol. Stratigr.",
  ["rivista italiana di pediatria = the italian journal of pediatrics"] = "Riv Ital Pediatr",
  ["rivista italiana di stomatologia"] = "Riv Ital Stomatol",
  ["rivista italiana d’igiene"] = "Riv. Ital. Ig.",
  ["rivista marittima"] = "Riv Maritt",
  ["rivista micologica romana"] = "Riv Micol Romana",
  ["rivista militare"] = "Riv Mil",
  ["rivista oto-neuro-oftalmologica"] = "Riv Otoneurooftalmol",
  ["rivista pediatrica siciliana"] = "Riv Pediatr Sicil",
  ["rivista rosminiana di filosofia e di cultura"] = "Riv Rosmin Filos Cult",
  ["rivista siciliana della tubercolosi"] = "Riv Sicil Tuberc",
  ["rivista siciliana della tubercolosi e delle malattie dell'apparato respiratorio"] = "Riv Sicil Tuberc Mal Appar Respir",
  ["rivista sperimentale di freniatria e medicina legale delle alienazioni mentali"] = "Riv Sper Freniatr Med Leg Alien Ment",
  ["rivista storica calabrese"] = "Riv Stor Calabr",
  ["rivista storica dell’antichità"] = "RStorAnt",
  ["rivista storica dell’antichità"] = "RSA",
  ["rivista storica italiana"] = "Riv Stor Ital",
  ["rizhskiy politekhnicheskiy institut"] = "Voprosy Dinamiki i Prochnosti",
  ["rmd open"] = "RMD Open",
  ["rmle online : research in middle level education"] = "RMLE Online",
  ["rms. research in mathematics & statistics"] = "RMS. Res. Math. Stat.",
  ["rn (for managers)"] = "RN (For Managers)",
  ["rn idaho"] = "RN Ida",
  ["rna & disease (houston, tex.)"] = "RNA Dis",
  ["rna (new york, n.y.)"] = "RNA",
  ["rna biology"] = "RNA Biol",
  ["rna technologies"] = "RNA Technol.",
  ["rnabc news"] = "RNABC News",
  ["rnabc news, registered nurses association of british columbia"] = "RNABC News.",
  ["rnao news"] = "RNAO News",
  ["road materials and pavement design"] = "Road Mater. Pavement Des.",
  ["robomech journal"] = "ROBOMECH J.",
  ["robotic intelligence and automation"] = "Rob. Intell. Autom.",
  ["robotic surgery (auckland)"] = "Robot Surg",
  ["robotic surgery: research and reviews"] = "Rob. Surg.: Res. Rev.",
  ["robotics and autonomous systems"] = "Rob Auton Syst",
  ["robotics and biomimetics"] = "Robotics Biomim",
  ["robotics and computer-integrated manufacturing"] = "Rob. Comput. Integr. Manuf.",
  ["robotics and mechatronics : proceedings of the 4th iftomm international symposium on robotics and mechatronics. iftomm international symposium on robotics and mechatronics (4th : 2015 : poitiers, france)"] = "Robot Mechatron (2015)",
  ["robotics reports"] = "Rob. Rep.",
  ["robotics science and systems : online proceedings"] = "Robot Sci Syst",
  ["robototekhnika i gibkie proizvodstvennye sistemy"] = "Robototekhn. Gibk. Proizvod. Sist.",
  ["roche review"] = "Roche Rev",
  ["rochester history"] = "Rochester Hist",
  ["rock mechanics and rock engineering"] = "Rock Mech. Rock Eng.",
  ["rock mechanics bulletin"] = "Rock Mech. Bull.",
  ["rocks and minerals"] = "Rocks Miner.",
  ["rocky mountain forest and range experiment station, forest series"] = "Rocky Mt. For. Range  Exp. Stn., For. Ser.",
  ["rocky mountain geology"] = "Rocky Mt Geol.",
  ["rocky mountain journal of mathematics"] = "Rocky Mt J Math",
  ["rocky mountain medical journal"] = "Rocky Mt Med J",
  ["rocznik muzeum narodowego w warszawie"] = "RoczMuzWarsz",
  ["rocznik pomorskiej akademii medycznej im. gen. karola swierczewskiego w szczecinie"] = "Rocz. Pomor. Akad. Med.",
  ["rocznik pomorskiej akademii medycznej im. gen. karola świerczewskiego w szczecinie"] = "Rocz Pomor Akad Med",
  ["roczniki akademii medycznej im. juliana marchlewskiego w bialymstoku"] = "Rocz. Akad. Med. Im. Juliana Marchlewskiego Bialymst.",
  ["roczniki akademii medycznej im. juliana marchlewskiego w bialymstoku. suplement"] = "Rocz Akad Med Im Juliana Marchlewskiego Bialymst Suppl",
  ["roczniki akademii medycznej im. juliana marchlewskiego w bialymstoku. supplement"] = "Rocz. Akad. Med. Im. Juliana Marchlewskiego Bialymst. Suppl.",
  ["roczniki akademii medycznej im. juliana marchlewskiego w białymstoku"] = "Rocz Akad Med Im Juliana Marchlewskiego Bialymst",
  ["roczniki akademii medycznej im. juliana marchlewskiego w białymstoku. suplement"] = "Rocz Akad Med Im Juliana Marchlewskiego Bialymst Suppl",
  ["roczniki akademii medycznej w bialymstoku"] = "Rocz. Akad. Med. Bialymst.",
  ["roczniki akademii medycznej w bialymstoku (1995)"] = "Rocz Akad Med Bialymst",
  ["roczniki akademii medycznej w bialymstoku. supplement"] = "Rocz. Akad. Med. Bialymst. Supl.",
  ["roczniki akademii medycznej w białymstoku (1995)"] = "Rocz Akad Med Bialymst",
  ["roczniki akademii medycznej w białymstoku = annales academiae medicae bialostocensis"] = "Rocz Akad Med Bialymst",
  ["roczniki akademii medycznej w białymstoku. suplement = annales academiae medicae bialostocensis. supplementum"] = "Rocz Akad Med Bialymst Supl",
  ["roczniki filozoficzne. annales de philosophie"] = "Rocz Filoz",
  ["roczniki humanistyczne"] = "Rocz Humanist",
  ["roczniki humanistyczne. 3, filologia klasyczna"] = "RHum",
  ["roczniki panstwowego zakladu higieny"] = "Rocz Panstw Zakl Hig",
  ["roczniki państwowego zakładu higieny"] = "Rocz Panstw Zakl Hig",
  ["roczniki polskiego towarzystwa matematycznego"] = "Mat. Stos.",
  ["rodo kagaku. the journal of science of labour"] = "Rodo Kagaku",
  ["roentgen-europ; radiodiagnostic clinique europeen"] = "Rontgeneur Radiodiagn Clin Eur",
  ["roentgen-europ; radiodiagnostic clinique européen"] = "Rontgeneur Radiodiagn Clin Eur",
  ["roeper review"] = "Roeper Rev",
  ["rofo : fortschritte auf dem gebiete der rontgenstrahlen und der nuklearmedizin"] = "Rofo",
  ["rofo fortschritte auf dem gebiete der rontgenstrahlen und der nuklearmedizin"] = "RoFo Fortschr. Geb. Rontgenstr. Nuklearmed.",
  ["rofo. fortschritte auf dem gebiete der rontgenstrahlen und der neuenbildgebenden verfahren rofo fortschr geb rontgenstr neuen bildge"] = "Verfahr",
  ["rofo. fortschritte auf dem gebiete der rontgenstrahlen und der nuklearmedizin"] = "ROFO. Fortschr. Geb. Rontgenstr. Nuklearmed.",
  ["rofo; fortschritte auf dem gebiete der rontgenstrahlen und der neuen bildgebenden verfahren"] = "Rofo. Fortschr. Geb. Rontgenstr. Neuen Bildgeb. Verfahr.",
  ["rogerian nursing science news"] = "Rogerian Nurs. Sci. News",
  ["rogerian nursing science news : newsletter of the society of rogerian scholars"] = "Rogerian Nurs Sci News",
  ["rolf nevanlinna institute research reports a"] = "Rolf Nevanlinna Inst. Res. Rep. A",
  ["rolling stone (san francisco, calif.)"] = "Roll Stone",
  ["romai journal"] = "ROMAI J.",
  ["roman provincial coinage"] = "RPC",
  ["romana gens. bollettino dell’associazione archeologica romana"] = "RomGens",
  ["romance linguistics 2008 : interactions in romance : selected papers from the 38th linguistic symposium on romance languages (lsrl), urbana-champaign, april 2008. linguistic symposium on romance languages (38th : 2008 : urbana-champaign, ill.)"] = "Roman Linguist 2008 (2008)",
  ["romance notes"] = "Roman Notes",
  ["romanian agricultural research"] = "Rom Agric Res",
  ["romanian biotechnological letters"] = "Rom Biotechnol Lett",
  ["romanian biotechnology letters"] = "Rom. Biotechnol. Lett.",
  ["romanian journal of anaesthesia and intensive care"] = "Rom J Anaesth Intensive Care",
  ["romanian journal of biochemistry"] = "Rom. J. Biochem.",
  ["romanian journal of biophysics"] = "Rom J Biophys",
  ["romanian journal of chemistry"] = "Rom. J. Chem.",
  ["romanian journal of diabetes nutrition and metabolic diseases"] = "Rom. J. Diabetes Nutr. Metab. Dis.",
  ["romanian journal of diabetes, nutrition, & metabolic diseases"] = "Rom J Diabetes Nutr Metab Dis",
  ["romanian journal of endocrinology"] = "Rom J Endocrinol",
  ["romanian journal of endocrinology / sponsore [sic] by the academy of medical sciences"] = "Rom J Endocrinol",
  ["romanian journal of gastroenterology"] = "Rom J Gastroenterol",
  ["romanian journal of geography"] = "Rom. J. Geogr.",
  ["romanian journal of gerontology and geriatrics"] = "Rom J Gerontol Geriatr",
  ["romanian journal of gerontology and geriatrics / the national institute of gerontology and geriatrics, bucharest"] = "Rom J Gerontol Geriatr",
  ["romanian journal of information science and technology"] = "Rom. J. Inf. Sci. Technol.",
  ["romanian journal of internal medicine"] = "Rom. J. Intern. Med.",
  ["romanian journal of internal medicine = revue roumaine de médecine interne"] = "Rom J Intern Med",
  ["romanian journal of legal medicine"] = "Rom. J. Legal Med.",
  ["romanian journal of materials"] = "Rom. J. Mater.",
  ["romanian journal of mathematics and computer science"] = "Rom. J. Math. Comput. Sci.",
  ["romanian journal of morphology and embryology"] = "Rom. J. Morphol. Embryol.",
  ["romanian journal of morphology and embryology = revue roumaine de morphologie et embryologie"] = "Rom J Morphol Embryol",
  ["romanian journal of neurology and psychiatry"] = "Rom. J. Neurol. Psychiatry",
  ["romanian journal of neurology and psychiatry = revue roumaine de neurologie et psychiatrie"] = "Rom J Neurol Psychiatry",
  ["romanian journal of ophthalmology"] = "Rom J Ophthalmol",
  ["romanian journal of parasitic diseases"] = "Romanian J Parasit Dis",
  ["romanian journal of physics"] = "Rom. J. Phys.",
  ["romanian journal of physiology"] = "Rom. J. Physiol.",
  ["romanian journal of physiology : physiological sciences"] = "Rom J Physiol",
  ["romanian journal of physiology : physiological sciences / [academia de stiinte medicale]"] = "Rom J Physiol",
  ["romanian journal of technical sciences. applied mechanics"] = "Rom. J. Tech. Sci. Appl. Mech.",
  ["romanian journal of virology"] = "Rom J Virol",
  ["romanian medical review"] = "Rom Med Rev",
  ["romanian reports in physics"] = "Rom Rep Phys",
  ["romanische forschungen: vierteljahresschrift für romanische sprachen und literaturen"] = "RomForsch",
  ["romanobarbarica. contributi allo studio dei rapporti culturali tra il mondo latino e mondo barbarico"] = "Romanobarbarica",
  ["romanobarbarica: contributi allo studio dei rapporti culturali tra mondo latino e mondo barbarico"] = "RomBarb",
  ["romische historische mitteilungen"] = "Rom Hist Mitt",
  ["ronald e. mcnair postbaccalaureate achievement program : [journal]"] = "Ronald E McNair Postbac Achiev Program",
  ["ronenbyo. [geriatrics]"] = "Ronenbyo",
  ["rontgen- und laboratoriumspraxis"] = "Rontgen Laborator",
  ["rontgen-blatter; zeitschrift fur rontgen-technik und medizinisch-wissenschaftliche photographie"] = "Rontgenblatter",
  ["rontgendiagnostik; ergebnisse"] = "Rontgendiagnostik Ergeb",
  ["rontgenpraxis; zeitschrift fur radiologische technik"] = "Rontgenpraxis",
  ["rooilijn (amsterdam)"] = "Rooilijn",
  ["rorschach research exchange and journal of projective techniques"] = "Rorschach Res Exch J Proj Tech",
  ["rose-hulman undergraduate math journal"] = "Rose-Hulman Undergrad. Math J.",
  ["rose-hulman undergraduate mathematics journal"] = "Rose-Hulman Undergrad. Math. J.",
  ["rossiiskaya akademiya nauk. algebra i analiz"] = "Algebra i Analiz",
  ["rossiiskaya akademiya nauk. avtomatika i telemekhanika"] = "Avtomat. i Telemekh.",
  ["rossiiskaya akademiya nauk. doklady akademii nauk"] = "Dokl. Akad. Nauk",
  ["rossiiskaya akademiya nauk. funktsionalnyi analiz i ego prilozheniya"] = "Funktsional. Anal. i Prilozhen.",
  ["rossiiskaya akademiya nauk. izvestiya. mekhanika zhidkosti i gaza"] = "Izv. Ross. Akad. Nauk Mekh. Zhidk. Gaza",
  ["rossiiskaya akademiya nauk. izvestiya. seriya matematicheskaya"] = "Izv. Ross. Akad. Nauk Ser. Mat.",
  ["rossiiskaya akademiya nauk. kristallografiya"] = "Kristallografiya",
  ["rossiiskaya akademiya nauk. matematicheskie zametki"] = "Mat. Zametki",
  ["rossiiskaya akademiya nauk. matematicheskii sbornik"] = "Mat. Sb.",
  ["rossiiskaya akademiya nauk. matematicheskoe modelirovanie"] = "Mat. Model.",
  ["rossiiskaya akademiya nauk. moskovskoe matematicheskoe obshchestvo. uspekhi matematicheskikh nauk"] = "Uspekhi Mat. Nauk",
  ["rossiiskaya akademiya nauk. prikladnaya matematika i mekhanika"] = "Prikl. Mat. Mekh.",
  ["rossiiskaya akademiya nauk. problemy peredachi informatsii"] = "Problemy Peredachi Informatsii",
  ["rossiiskaya akademiya nauk. programmirovanie"] = "Programmirovanie",
  ["rossiiskaya akademiya nauk. sankt-peterburgskoe otdelenie. matematicheskii institut im. v. a. steklova. zapiski nauchnykh seminarov (pomi)"] = "Zap. Nauchn. Sem. S.-Peterburg. Otdel. Mat. Inst. Steklov. (POMI)",
  ["rossiiskaya akademiya nauk. sibirskoe otdelenie. institut gidrodinamiki. dinamika sploshnoi sredy"] = "Dinamika Sploshn. Sredy",
  ["rossiiskaya akademiya nauk. sibirskoe otdelenie. institut matematiki im. s. l. soboleva. sibirskii matematicheskii zhurnal"] = "Sibirsk. Mat. Zh.",
  ["rossiiskaya akademiya nauk. teoreticheskaya i matematicheskaya fizika"] = "Teoret. Mat. Fiz.",
  ["rossiiskaya akademiya nauk. teoriya veroyatnostei i ee primeneniya"] = "Teor. Veroyatn. Primen.",
  ["rossiiskii fiziologicheskii zhurnal imeni i. m. sechenova"] = "Ross. Fiziol. Zh. Im. I. M. Sechenova",
  ["rossiiskii fiziologicheskii zhurnal imeni i.m. sechenova"] = "Ross. Fiziol. Zh. Im. I M Sechenova",
  ["rossiiskii fiziologicheskii zhurnal imeni i.m. sechenova / rossiiskaia akademiia nauk"] = "Ross Fiziol Zh Im I M Sechenova",
  ["rossiiskii gastroenterologicheskii zhurnal : ezhekvartalnyi nauchno-prakticheskii zhurnal"] = "Ross Gastroenterol Zh",
  ["rossiiskii khimicheskii zhurnal"] = "Ross. Khim. Zh.",
  ["rossiiskii meditsinskii zhurnal"] = "Ross. Med. Zh.",
  ["rossiiskii meditsinskii zhurnal : organ ministerstva zdravookhraneniia rsfsr"] = "Ross Med Zh",
  ["rossiĭskii fiziologicheskiĭ zhurnal imeni i.m. sechenova"] = "Ross Fiziol Zh Im I M Sechenova",
  ["rossiĭskiĭ gastroėnterologicheskiĭ zhurnal : ezhekvartalńyĭ nauchno-prakticheskiĭ zhurnal"] = "Ross Gastroenterol Zh",
  ["rossiĭskiĭ immunologicheskiĭ zhurnal"] = "Ross Immunol Zhurnal",
  ["rossiĭskiĭ medit︠s︡inskiĭ zhurnal : organ ministerstva zdravookhranenii︠a︡ rsfsr"] = "Ross Med Zh",
  ["rossijskaja archeologija"] = "RossA",
  ["rossiyskaya akademiya nauk"] = "Sibirsk. Mat. Zh.",
  ["rossíiskaya akademiya nauk"] = "Algebra i Analiz",
  ["rostocker mathematisches kolloquium"] = "Rostock. Math. Kolloq.",
  ["roumanian archives of microbiology and immunology"] = "Roum Arch Microbiol Immunol",
  ["routledge contemporary readings in philosophy"] = "Routledge Contemp. Read. Philos.",
  ["routledge european sociological association studies in european societies"] = "Routl. Eur. Sociol. Ass.",
  ["routledge inem advances in economic methodology"] = "Routledge INEM Adv. Econ. Methodol.",
  ["routledge philosophers"] = "Routl Philos",
  ["routledge research in early modern history"] = "Routledge Res. Early Mod. Hist.",
  ["routledge studies in ecological economics"] = "Routl. Stud. Ecol. Econ.",
  ["routledge studies in the philosophy of mathematics and physics"] = "Routledge Stud. Philos. Math. Phys.",
  ["routledge studies in the philosophy of religion"] = "Rout. Stud. Philos. Rel.",
  ["roux's archives of developmental biology : the official organ of the edbo"] = "Rouxs Arch Dev Biol",
  ["royal canadian dental corps quarterly"] = "R. Can. Dent. Corps Q.",
  ["royal institute of chemistry, lecture series"] = "R. Inst. Chem. Lect. Ser.",
  ["royal institute of chemistry, lectures, monographs, and reports"] = "R. Inst. Chem. Lect. Monogr. Rep.",
  ["royal institute of chemistry, reviews"] = "R. Inst. Chem. Rev.",
  ["royal institute of philosophy supplement"] = "Roy. Inst. Philos. Suppl.",
  ["royal institute of public health and hygiene journal"] = "R. Inst. Public Health Hyg. J.",
  ["royal melbourne hospital clinical reports"] = "R Melb Hosp Clin Rep",
  ["royal naval medical bulletin"] = "R Nav Med Bull",
  ["royal society of health journal"] = "R Soc Health J",
  ["royal society open science"] = "R Soc Open Sci",
  ["royal statistical society lecture note series"] = "Roy. Statist. Soc. Lecture Note Ser.",
  ["rozhledy v chirurgii"] = "Rozhl. Chir.",
  ["rozhledy v chirurgii : mesicnik ceskoslovenske chirurgicke spolecnosti"] = "Rozhl Chir",
  ["rozhledy v chirurgii : měsíčník československé chirurgické společnosti"] = "Rozhl Chir",
  ["rozhledy v tuberkulose"] = "Rozhl Tuberk",
  ["rozpravy české akademie věd a umění. třída ii, mathematicko-přírodovědecká"] = "Rozpr Ces Akad Ved Umeni Trida II Mat Prirodoved",
  ["rozprawy monografie"] = "Rozpr. Monogr.",
  ["rozprawy wydzialu nauk medycznych"] = "Rozpr. Wydz. Nauk Med.",
  ["rozprawy wydziału nauk medycznych"] = "Rozpr Wydz Nauk Med",
  ["rozprawy. polska akademia umiejętności. wydział lekarski"] = "Rozprawy",
  ["röfo : fortschritte auf dem gebiete der röntgenstrahlen und der nuklearmedizin"] = "Rofo",
  ["römische historische mitteilungen"] = "Rom Hist Mitt",
  ["röntgen- und laboratoriumspraxis"] = "Rontgen Laborator",
  ["röntgen-blätter; zeitschrift für röntgen-technik und medizinisch-wissenschaftliche photographie"] = "Rontgenblatter",
  ["röntgendiagnostik; ergebnisse"] = "Rontgendiagnostik Ergeb",
  ["röntgenphotographie; medizinsche photographie und medizinische laboratoriumspraxis"] = "Rontgenphotogr Med Photogr Med Lab Prax",
  ["röntgenpraxis; diagnostik, röntgen-, radium-, lichttherapie"] = "Rontgenprax Diagn Rontgen Radium Lichtther",
  ["röntgenpraxis; zeitschrift für radiologische technik"] = "Rontgenpraxis",
  ["rps pharmacy and pharmacology reports"] = "RPS Pharm. Pharmacol. Rep.",
  ["rsc advances"] = "RSC Adv.",
  ["rsc applied interfaces"] = "RSC Appl. Interfaces",
  ["rsc applied polymers"] = "RSC Appl. Polym.",
  ["rsc biomolecular sciences"] = "RSC Biomol Sci",
  ["rsc chemical biology"] = "RSC Chem Biol",
  ["rsc medicinal chemistry"] = "RSC Med Chem",
  ["rsc sustainability"] = "RSC Sustainability",
  ["rsme springer series"] = "RSME Springer Ser.",
  ["rsr. reference services review"] = "Ref Serv Rev",
  ["rsyn chemical sciences"] = "RSYN Chem. Sci.",
  ["rsyn proceedings"] = "RSYN Proc.",
  ["rsyn research journal"] = "RSYN Res. J.",
  ["rt-pcr protocols, second edition"] = "Methods. Mol. Biol.",
  ["rubber chemistry and technology"] = "Rubber Chem. Technol.",
  ["ruch prawniczy, ekonomiczny i socjologiczny (1962)"] = "Ruch Praw Ekon Socjol",
  ["rudiae. ricerche sul mondo classico"] = "Rudiae",
  ["rudiae: ricerche sul mondo classico"] = "Rudiae",
  ["rudolstadter heimathefte"] = "Rudolstaedter Heimath",
  ["rudolstädter heimathefte"] = "Rudolstaedter Heimath",
  ["rumah tangga dan kesehatan"] = "Rumah Tangga Kesehatan",
  ["rumanian medical review"] = "Rum Med Rev",
  ["ruptures (université de montréal. groupe de recherche interdisciplinaire en santé)"] = "Ruptures (Univ Montr Groupe Rech Interdiscip Sante)",
  ["rural africana"] = "Rural Afr",
  ["rural and remote health"] = "Rural Remote Health",
  ["rural conditions and trends"] = "Rural Cond Trends",
  ["rural demography"] = "Rural Demogr",
  ["rural history : economy, society, culture"] = "Rural Hist",
  ["rural landscapes: society, environment, history"] = "Rural Landscapes: Soc. Environ. Hist.",
  ["rural mental health"] = "Rural Ment Health",
  ["rural policy brief"] = "Rural Policy Brief",
  ["rural policy brief / rupri rural health panel"] = "Rural Policy Brief",
  ["rural reconstruction forum"] = "Rural Reconstr Forum",
  ["rural sociology"] = "Rural Sociol",
  ["rural sustainability research"] = "Rural Sustainability Res.",
  ["russ coile's health trends"] = "Russ Coiles Health Trends",
  ["russ coile’s health trends"] = "Russ Coiles Health Trends",
  ["russian academy of sciences"] = "Russian Acad. Sci. Izv. Math.",
  ["russian aeronautics"] = "Russ. Aeronaut.",
  ["russian agricultural sciences"] = "Russ Agric Sci",
  ["russian and east european finance and trade"] = "Russian E. Europ. Finance Trade",
  ["russian chemical bulletin"] = "Russ. Chem. Bull.",
  ["russian chemical bulletin = izvestii︠a︡ akademii nauk. serii︠a︡ khimicheskai︠a︡"] = "Russ Chem Bull",
  ["russian chemical reviews"] = "Russ. Chem. Rev.",
  ["russian economic trends"] = "Russian Econ. Trends",
  ["russian electrical engineering"] = "Russ. Electr. Eng.",
  ["russian engineering research"] = "Russ. Eng. Res.",
  ["russian entomological journal"] = "Russkii entomologicheskii zhurnal|Russ. Entomol. J.",
  ["russian entomological journal = russkiĭ ėntomologicheskiĭ zhurnal"] = "Russ Entomol J",
  ["russian geology and geophysics"] = "Russ. Geol. Geophys.",
  ["russian history. histoire russe"] = "Russ Hist (Pittsburgh)",
  ["russian journal of agricultural and socio-economic sciences"] = "Russ J Agric Socioecon Sci",
  ["russian journal of applied chemistry"] = "Russ. J. Appl. Chem.",
  ["russian journal of biological invasions"] = "Russ J Biol Invasions",
  ["russian journal of bioorganic chemistry"] = "Russ. J. Bioorg. Chem.",
  ["russian journal of coordination chemistry"] = "Russ. J. Coord. Chem.",
  ["russian journal of developmental biology"] = "Russ J Dev Biol",
  ["russian journal of earth sciences"] = "Russ. J. Earth Sci.",
  ["russian journal of ecology"] = "Russ J Ecol",
  ["russian journal of ecosystem ecology"] = "Russ J Ecosyst Ecol",
  ["russian journal of electrochemistry"] = "Russ. J. Electrochem.",
  ["russian journal of general chemistry"] = "Russ. J. Gen. Chem.",
  ["russian journal of genetics"] = "Russ. J. Genet.",
  ["russian journal of genetics. applied research"] = "Russ J Genet Appl Res",
  ["russian journal of genetics: applied research"] = "Russ. J. Genet.: Appl. Res.",
  ["russian journal of herpetology"] = "Russ J Herpetol",
  ["russian journal of immunology : rji : official journal of russian society of immunology"] = "Russ J Immunol",
  ["russian journal of inorganic chemistry"] = "Russ. J. Inorg. Chem.",
  ["russian journal of marine biology"] = "Russ. J. Mar. Biol.",
  ["russian journal of mathematical physics"] = "Russ. J. Math. Phys.",
  ["russian journal of nematology"] = "Russ J Nematol",
  ["russian journal of non-ferrous metals"] = "Russ. J. Non-Ferrous Met.",
  ["russian journal of nondestructive testing"] = "Russ. J. Nondestr. Test.",
  ["russian journal of nonlinear dynamics"] = "Russ. J. Nonlinear Dyn.",
  ["russian journal of numerical analysis and mathematical modelling"] = "Russian J. Numer. Anal. Math. Modelling",
  ["russian journal of organic chemistry"] = "Russ. J. Org. Chem.",
  ["russian journal of pacific geology"] = "Russ. J. Pac. Geol.",
  ["russian journal of physical chemistry"] = "Russ. J. Phys. Chem.",
  ["russian journal of physical chemistry [translation of zhurnal fizicheskoi khimii]"] = "Russ. J. Phys. Chem.",
  ["russian journal of physical chemistry a"] = "Russ. J. Phys. Chem. A",
  ["russian journal of physical chemistry a, focus on chemistry"] = "Russ. J. Phys. Chem. A",
  ["russian journal of physical chemistry b"] = "Russ. J. Phys. Chem. B",
  ["russian journal of physical chemistry b, focus on physics"] = "Russ. J. Phys. Chem. B",
  ["russian journal of plant physiology"] = "Russ. J. Plant Physiol.",
  ["russian journal of plant physiology: a comprehensive russian journal on modern phytophysiology"] = "Russ J Plant Physiol",
  ["russian journal of theriology"] = "Russ J Theriol",
  ["russian mathematical surveys"] = "Russ. Math. Surv.",
  ["russian mathematics"] = "Russ. Math.",
  ["russian mathematics (izvestiya vuz"] = "Russian Math. (Iz. VUZ)",
  ["russian mathematics (izvestiya vuz. matematika)"] = "Russian Math. (Iz. VUZ)",
  ["russian metallurgy"] = "Russ. Metall.",
  ["russian meteorology and hydrology"] = "Russ. Meteorol. Hydrol.",
  ["russian microelectronics"] = "Russ. Microlectron.",
  ["russian physics journal"] = "Russ. Phys. J.",
  ["russian review of biology"] = "Russ Rev Biol",
  ["russian social science review : a journal of translations"] = "Russ Soc Sci Rev",
  ["russian studies in history"] = "Russ Stud Hist",
  ["russion journal of coordination chemistry"] = "Russ. J. Coord. Chem.",
  ["russkiĭ zhurnal spid, rak i obshchestvennoe zdorovʹe = russian journal aids, cancer, and public health"] = "Rus Zhurnal SPID Rak Obshchestvennoe Zdor",
  ["russkij ornithologičeskij žurnal"] = "Rus Ornithol Z",
  ["rutgers camden law journal"] = "Rutgers Camden Law J",
  ["rutgers computer & technology law journal"] = "Rutgers Comput Technol Law J",
  ["rutgers journal--computers, technology, and the law"] = "Rutgers J Comput Technol Law",
  ["rutgers law journal"] = "Rutgers Law J",
  ["rutgers law review"] = "Rutgers Law Rev",
  ["rutgers series on self and social identity"] = "Rutg. Self Soc. Id.",
  ["ruthenica : rossiĭskiĭ malakologicheskiĭ zhurnal = russian malacological journal"] = "Ruthenica",
  ["rutland record : journal of the rutland record society"] = "Rutland Rec",
  ["ryan advisory for health services governing boards"] = "Ryan Advis. Health Serv. Gov. Boards",
  ["ryoikibetsu shokogun shirizu"] = "Ryoikibetsu Shokogun Shirizu",
  ["ryoikibetsu shokogun shirizu (series on symptomatic syndrome by area)"] = "Ryoikibetsu Shokogun Shirizu",
  ["ryōikibetsu shōkōgun shirīzu"] = "Ryoikibetsu Shokogun Shirizu",
  ["ryukyu mathematical journal"] = "Ryukyu Math. J.",
  ["ryumachi. [rheumatism]"] = "Ryumachi",
  ["ryūmachi. [rheumatism]"] = "Ryumachi",
  ["région et développement"] = "Région Dével.",
  ["römisch-germanische forschungen"] = "RGF",
  ["römische quartalschrift für christliche altertumskunde und für kirchengeschichte"] = "RQA",
  ["römische quartalschrift für christliche altertumskunde und kirchengeschichte"] = "RoemQSchr",
  ["römisches österreich. jahresschrift der österreichischen gesellschaft für archäologie"] = "RoemOe",
  ["römisches österreich: jahresschrift der österreichischen gesellschaft für archäologie"] = "Rö",
  ["s & ta & nu. rivista di scienza e tecnologia degli alimenti e di nutrizione umana"] = "S TA NU",
  ["s e ta e nu; rivista di scienza e tecnologia degli alimenti e di nutrizione umana"] = "S TA NU",
  ["s. e. m"] = "S E M",
  ["s.a. nursing journal = s.a. verplegingstydskrif"] = "S A Nurs J",
  ["sa journal of radiology"] = "SA J Radiol",
  ["sa nursing journal"] = "SA Nurs. J.",
  ["sa nursing journal. sa verplegingstydskrif"] = "SA Nurs J",
  ["sa. sociological analysis"] = "Sociol Anal",
  ["saad digest"] = "SAAD Dig",
  ["saalburg-jahrbuch. bericht des saalburg-museums"] = "SaalbJb",
  ["saarbrücker beiträge zur altertumskunde"] = "SaarBeitr",
  ["saarbrücker studien und materialien zur altertumskunde"] = "SaarStMat",
  ["saarlandisches aerzteblatt"] = "Saarl Arztebl",
  ["saarländisches aerzteblatt"] = "Saarl Arztebl",
  ["saas bulletin, biochemistry and biotechnology"] = "SAAS Bull Biochem Biotechnol",
  ["saas-fee advanced course"] = "Saas-Fee Adv. Course",
  ["saber (cumaná, venezuela)"] = "Saber (Cumana)",
  ["sabrao journal of breeding and genetics"] = "SABRAO J Breed Genet",
  ["sac explorer"] = "Sac Explor",
  ["sacema quarterly"] = "SACEMA Q",
  ["sacer. bollettino della associazione storica saccarese"] = "Sacer",
  ["sacitra ayurveda"] = "Sacitra Ayurveda",
  ["sacitra āyurveda"] = "Sacitra Ayurveda",
  ["sacris erudiri: jaarboek voor godsdienstwetenschappen"] = "SEJG",
  ["sadhana. academy proceedings in engineering sciences"] = "Sadhana",
  ["sadj : journal of the south african dental association = tydskrif van die suid-afrikaanse tandheelkundige vereniging"] = "SADJ",
  ["sae international journal of advances and current practices in mobility"] = "SAE Int J Adv Curr Pract Mobil",
  ["sae international journal of aerospace"] = "SAE Int. J. Aerosp.",
  ["sae international journal of commercial vehicles"] = "SAE Int. J. Commer. Veh.",
  ["sae international journal of connected and automated vehicles"] = "SAE Int. J. Connected Autom. Veh.",
  ["sae international journal of electrified vehicles"] = "SAE Int. J. Electrified Veh.",
  ["sae international journal of engines"] = "SAE Int. J. Engines",
  ["sae international journal of fuels and lubricants"] = "SAE Int. J. Fuels Lubr.",
  ["sae international journal of manufacturing & materials"] = "SAE Int. J. Manuf. Mater.",
  ["sae international journal of materials and manufacturing"] = "SAE Int. J. Mater. Manuf.",
  ["sae international journal of passenger cars - electronic and electrical systems"] = "SAE Int. J. Passenger Cars Electron. Electr. Syst.",
  ["sae international journal of passenger cars - mechanical systems"] = "SAE Int. J. Passenger Cars Mech. Syst.",
  ["sae international journal of passenger cars. electronic and electrical systems"] = "SAE Int J Passeng Cars Electron Electr Syst",
  ["sae international journal of sustainable transportation, energy, environment, & policy"] = "SAE Int. J. Sustainable Transp. Energy Environ. Policy",
  ["sae international journal of transportation cybersecurity and privacy"] = "SAE Int. J. Transp. Cybersecur. Privacy",
  ["sae international journal of transportation safety"] = "SAE Int J Transp Saf",
  ["sae international journal of vehicle dynamics, stability, and nvh"] = "SAE Int J Veh Dyn Stab NVH",
  ["sae journal"] = "SAE J.",
  ["sae quarterly transactions"] = "SAE Q. Trans.",
  ["sae transactions"] = "SAE Trans.",
  ["saeculum. jahrbuch für universalgeschichte"] = "Saeculum",
  ["saengmyeong gwahag hoeji"] = "Saengmyeong Gwahag Hoeji",
  ["saengyak hakhoe chi"] = "Saengyak Hakhoe Chi",
  ["safaids news : southern africa aids information dissemination service bulletin"] = "SAfAIDS News",
  ["safe journal"] = "SAFE J",
  ["safe motherhood"] = "Safe Mother",
  ["safety (basel, switzerland)"] = "Safety (Basel)",
  ["safety and health at work"] = "Saf Health Work",
  ["safety and reliability"] = "Saf. Reliab.",
  ["safety in extreme environments"] = "Saf. Extreme Environ.",
  ["safety science"] = "Saf. Sci.",
  ["sagamore army materials research conference proceedings"] = "Sagamore Army Materials Res. Conf. Proc.",
  ["sage (atlanta, ga.)"] = "Sage",
  ["sage benchmarks in social research methods"] = "Sage Benchmarks Soc. Res. Methods",
  ["sage focus editions"] = "Sage Focus Ed.",
  ["sage open"] = "Sage Open",
  ["sage open engineering"] = "Sage Open Eng.",
  ["sage open medical case reports"] = "SAGE Open Med Case Rep",
  ["sage open medicine"] = "SAGE Open Med",
  ["sage open nursing"] = "SAGE Open Nurs",
  ["saggi fenici"] = "SaggiFen",
  ["saggi scientifici"] = "Saggi Sci.",
  ["sago palm"] = "Sago Palm",
  ["saguntum. papeles del laboratorio de arqueología de valencia"] = "Saguntum",
  ["sağlık bilimleri dergisi"] = "Saglik Bilim Derg",
  ["sah bulletin"] = "SAH Bull.",
  ["sahara (segrate, italy)"] = "Sahara (Segrate)",
  ["sahara j : journal of social aspects of hiv/aids research alliance"] = "SAHARA J",
  ["sahara j : journal of social aspects of hiv/aids research alliance / sahara , human sciences research council"] = "SAHARA J",
  ["saibo kogaku"] = "Saibo Kogaku",
  ["saiee africa research journal"] = "SAIEE Afr. Res. J.",
  ["sains indonesia"] = "Sains Indones",
  ["sains malaysiana"] = "Sains Malays",
  ["saint anthony messenger; a national catholic family magazine"] = "St Anthony Messenger",
  ["saint louis university journal of health law & policy"] = "St Louis U J Health Law Policy",
  ["saint louis university law journal"] = "St Louis Univ Law J",
  ["saint louis university public law review"] = "St Louis Univ Public Law Rev",
  ["saint-luc medical. sint-lucas tijdschrift"] = "St Luc Med",
  ["saint-luc médical. sint-lucas tijdschrift"] = "St Luc Med",
  ["sairaanhoidon vuosikirja"] = "Sairaanh Vuosik",
  ["sairaanhoitaja (helsinki, finland : 1991)"] = "Sairaanhoitaja",
  ["sairaanhoitaja. sjukskoterskan"] = "Sairaanhoitaja",
  ["sairaanhoitaja. sjuksköterskan"] = "Sairaanhoitaja",
  ["sais review (paul h. nitze school of advanced international studies)"] = "SAIS Rev",
  ["saishin igaku"] = "Saishin Igaku",
  ["saishin igaku (recent medicine)"] = "Saishin Igaku",
  ["saishin igaku. modern medicine"] = "Saishin Igaku",
  ["saisons d'alsace"] = "Saisons Alsace",
  ["saisons d’alsace"] = "Saisons Alsace",
  ["saitabi. noticiario de historia, arte y arqueología de levante"] = "Saitabi",
  ["saitama mathematical journal"] = "Saitama Math. J.",
  ["sajch : the south african journal of child health"] = "SAJCH",
  ["salamandra (frankfurt)"] = "Salamandra (Frankf)",
  ["salduie. estudios de prehistoria y arqueología"] = "Salduie",
  ["saline systems"] = "Saline Syst",
  ["salisbury medical bulletin"] = "Salisbury Med Bull",
  ["sallim kwahak yŏn'gu"] = "Sallim Kwahak Yongu",
  ["salmagundi (saratoga springs, n.y.)"] = "Salmagundi",
  ["salubridad y asistencia"] = "Salubr Asist",
  ["salud bucal"] = "Salud Bucal",
  ["salud bucal / confederacion odontologica de la republica argentina"] = "Salud Bucal",
  ["salud colectiva"] = "Salud Colect",
  ["salud mental (mexico city, mexico)"] = "Salud Ment (Mex)",
  ["salud publica de mexico"] = "Salud Publica Mex",
  ["salud pública de méxico"] = "Salud Publica Mex",
  ["salud pública de méxico"] = "Salud pública Méx.",
  ["salud reproductiva y sociedad : organo informativo del programa salud reproductiva y sociedad de el colegio de mexico"] = "Salud Reprod Soc",
  ["salud reproductiva y sociedad : órgano informativo del programa salud reproductiva y sociedad de el colegio de méxico"] = "Salud Reprod Soc",
  ["salud uis : revista de la universidad industrial de santander, facultad de salud"] = "Salud UIS",
  ["salud y drogas"] = "Salud Drogas",
  ["salud y sociedad"] = "Salud Soc",
  ["salud y vida"] = "Salud Vida",
  ["salus / department of national health and population development"] = "Salus",
  ["salzburger beitrage zur paracelsusforschung"] = "Salzburger Beitr Paracelsusforsch",
  ["salzburger beiträge zur paracelsusforschung"] = "Salzburger Beitr Paracelsusforsch",
  ["same-day surgery"] = "Same Day Surg",
  ["sammlung tusculum"] = "Samml. Tusculum",
  ["sammlung zwangloser abhandlungen aus dem gebiete der psychiatrie und neurologie"] = "Samml Zwangl Abh Geb Psychiatr Neurol",
  ["samoa medical journal"] = "Samoa Med J",
  ["samothrace. excavations conducted by the institute of fine arts of new york university"] = "Samothrace",
  ["sampe journal"] = "SAMPE J.",
  ["sampe journal. society for the advancement of material and process engineering"] = "Sampe J",
  ["sample registration bulletin"] = "Sample Regist Bull",
  ["sampling theory in signal and image processing. an international journal"] = "Sampl. Theory Signal Image Process.",
  ["sampling theory, signal processing, and data analysis"] = "Sampling Theory, Signal Process., Data Anal.",
  ["san fernando valley dental society bulletin"] = "San Fernando Val Dent Soc Bull",
  ["san francisco medicine"] = "San Franc Med",
  ["san francisco state university series in philosophy"] = "San Franc. State Univ. Ser. Philos.",
  ["san gabriel valley dental society bulletin"] = "San Gabriel Val Dent Soc Bull",
  ["san jose studies"] = "San Jose Stud",
  ["san josé studies"] = "San Jose Stud",
  ["sandalion. quaderni di cultura classica, cristiana e medievale"] = "Sandalion",
  ["sanfujinka no jissai (practice of gynecology and obstetrics)"] = "Sanfujinka No Jissai",
  ["sanfujinka no jissai. practice of gynecology and obstetrics"] = "Sanfujinka No Jissai",
  ["sangyo eiseigaku zasshi"] = "Sangyo Eiseigaku Zasshi",
  ["sangyo eiseigaku zasshi (journal of occupational health)"] = "Sangyo Eiseigaku Zasshi",
  ["sangyo igaku (japanese journal of industrial health)"] = "Sangyo Igaku",
  ["sangyo igaku. japanese journal of industrial health"] = "Sangyo Igaku",
  ["sangyo ika daigaku zasshi"] = "Sangyo Ika Daigaku Zasshi",
  ["sangyō eiseigaku zasshi = journal of occupational health"] = "Sangyo Eiseigaku Zasshi",
  ["sangyō igaku. japanese journal of industrial health"] = "Sangyo Igaku",
  ["sanidad aeronáutica : organo informativo técnico"] = "Sanid Aeronaut",
  ["sanidad y beneficencia municipal; revista médico social"] = "Sanid Benefic Munic",
  ["sanita pubblica"] = "Sanita Pubblica",
  ["sanità pubblica"] = "Sanita Pubblica",
  ["sankhya a. the indian journal of statistics"] = "Sankhya A",
  ["sankhya b. applied and interdisciplinary statistics"] = "Sankhya B",
  ["sankhya b. the indian journal of statistics"] = "Sankhya B",
  ["sankhya. series b. [methodological.]"] = "Sankhya Ser B",
  ["sankhyā. series a. (2008)"] = "Sankhya Ser A",
  ["sankhyā. series b (2008)"] = "Sankhya B (2008)",
  ["sankhyā. series b. [methodological.]"] = "Sankhya Ser B",
  ["sankyo kenkyusho nenpo"] = "Sankyo Kenkyusho Nenpo",
  ["sanop misaengmul hakhoe chi"] = "Sanop Misaengmul Hakhoe Chi",
  ["sanop misaengmul hakhoechi"] = "Sanop Misaengmul Hakhoechi",
  ["sanŏp misaengmul hakhoe chi"] = "Sanop Misaengmul Hakhoe Chi",
  ["santa clara computer and high-technology law journal"] = "Santa Clara Comput High Technol Law J",
  ["santa clara law review"] = "Santa Clara Law Rev",
  ["santa clara lawyer (1961)"] = "Santa Clara Lawyer",
  ["santa fe institute studies in the sciences of complexity"] = "Santa Fe Inst. Stud. Sci. Complexity Lecture Notes",
  ["sante (montrouge, france)"] = "Sante",
  ["sante mentale au quebec"] = "Sante Ment. Que.",
  ["sante publique"] = "Sante Publique (Bucur.)",
  ["sante publique (vandoeuvre-les-nancy"] = "Sante Publique",
  ["sante publique (vandoeuvre-les-nancy, france)"] = "Sante Publique",
  ["sante quebec"] = "Sante Que.",
  ["sante quebec : revue de la corporation professionnelle des infirmieres et infirmiers auxiliaires du quebec"] = "Sante Que",
  ["sante salud"] = "Sante Salud",
  ["santé (montrouge, france)"] = "Sante",
  ["santé mentale au québec"] = "Sante Ment Que",
  ["santé publique"] = "Sante Publique (Paris)",
  ["santé publique (vandoeuvre-lès-nancy, france)"] = "Sante Publique",
  ["santé québec : revue de la corporation professionnelle des infirmières et infirmiers auxiliaires du québec"] = "Sante Que",
  ["santé salud"] = "Sante Salud",
  ["santiago / [universidad de oriente]"] = "Santiago",
  ["santo tomas journal of medicine"] = "St Tomas J Med",
  ["santo tomas nursing journal"] = "St Tomas Nurs J",
  ["sao paulo journal of mathematical sciences"] = "Sao Palo J. Math. Sci.",
  ["sapporo igaku zasshi (sapporo medical journal)"] = "Sapporo Igaku Zasshi",
  ["sapporo igaku zasshi. the sapporo medical journal"] = "Sapporo Igaku Zasshi",
  ["sar and qsar in environmental research"] = "SAR QSAR Environ Res",
  ["sarajevo journal of mathematics"] = "Sarajevo J. Math.",
  ["sarcoidosis, vasculitis, and diffuse lung diseases"] = "Sarcoidosis Vasc. Diffuse Lung Dis.",
  ["sarcoidosis, vasculitis, and diffuse lung diseases : official journal of wasog"] = "Sarcoidosis Vasc Diffuse Lung Dis",
  ["sarcoidosis, vasculitis, and diffuse lung diseases : official journal of wasog / world association of sarcoidosis and other granulomatous disorders"] = "Sarcoidosis Vasc Diffuse Lung Dis",
  ["sardis. publications of the american society for the excavation of sardis"] = "Sardis",
  ["sargetia. acta musei regionalis devensis"] = "Sargetia",
  ["sas global forum"] = "SAS Glob Forum",
  ["sas journal"] = "SAS J",
  ["saskatchewan history"] = "Sask Hist",
  ["saskatchewan law review"] = "Sask Law Rev",
  ["satellite navigation"] = "Satell. Navig.",
  ["saturday review"] = "Saturday Rev",
  ["saturday review of the sciences"] = "Saturday Rev Sci",
  ["saturday review/world"] = "Saturday Rev World",
  ["saudi journal for health sciences"] = "Saudi J Health Sci",
  ["saudi journal of anaesthesia"] = "Saudi J Anaesth",
  ["saudi journal of biological sciences"] = "Saudi J Biol Sci",
  ["saudi journal of gastroenterology : official journal of the saudi gastroenterology association"] = "Saudi J Gastroenterol",
  ["saudi journal of kidney diseases and transplantation : an official publication of the saudi center for organ transplantation, saudi arabia"] = "Saudi J Kidney Dis Transpl",
  ["saudi journal of medicine & medical sciences"] = "Saudi J Med Med Sci",
  ["saudi journal of ophthalmology : official journal of the saudi ophthalmological society"] = "Saudi J Ophthalmol",
  ["saudi medical journal"] = "Saudi Med J",
  ["saudi pharmaceutical journal : spj : the official publication of the saudi pharmaceutical society"] = "Saudi Pharm J",
  ["sautuola. revista del instituto de prehistoria y arqueologia sautuola"] = "Sautuola",
  ["sauvegarde de l'enfance"] = "Sauvegarde enfance (Paris, 1946)",
  ["savaria. bulletin der museen des komitats vas"] = "Savaria",
  ["savaşta erbaşlar"] = "Savasta Erbaslar",
  ["savoirs actuels"] = "Savoirs Actuels",
  ["savoirs scientifiques & pratiques d’enseignement"] = "Savoirs Sci. Prat. Enseign.",
  ["são paulo medical journal = revista paulista de medicina"] = "Sao Paulo Med J",
  ["são paulo médico"] = "Sao Paulo Med",
  ["sbc : newsletter of the society for bioethics consultation"] = "SBC",
  ["sbornik archivnich praci"] = "Sb Arch Pr",
  ["sbornik ceske geograficke spolecnosti"] = "Sbs Ces Geogr Spol",
  ["sbornik lekarsky"] = "Sb Lek",
  ["sbornik mathematics"] = "Sbornik Math.",
  ["sbornik praci filosoficke fakulty brnenske university. c, rada historicka"] = "Sb Pr Filos Fak Brnenske Univ Rada Hist",
  ["sbornik pro pathofysiologii traveni a vyzivy; gastroenterologia bohema"] = "Sb Pathofysiol Traveni Vyz Gastroenterol Bohema",
  ["sbornik trudov. azerbaidzhanskii gosudarstvennyi meditsinskii institut im. n. narimanova"] = "Sb Tr Azerbaidzhanskii Gos Meditsinskii Inst N Narimanova",
  ["sbornik trudov. azerbaĭdzhanskiĭ gosudarstvennyĭ medit︠s︡inskiĭ institut"] = "Sb Tr Azerbaidzhanskii Gos Meditsinskii Inst N Narimanova",
  ["sbornik vedeckych praci lekarske fakulty karlovy university v hradci kralove"] = "Sb Ved Pr Lek Fak Karlovy Univerzity Hradci Kralove",
  ["sbornik vedeckych praci lekarske fakulty karlovy univerzity v hradci kralove"] = "Sb. Ved. Pr. Lek. Fak. Karlovy Univerzity Hradci Kralove",
  ["sbornik vedeckych praci lekarske fakulty karlovy univerzity v hradci kralove. supplementum"] = "Sb. Ved. Pr. Lek. Fak. Karlovy Univerzity Hradci Kralove. Suppl.",
  ["sbornik vedeckych praci, vysoka skola chemickotechnologicka pardubice"] = "Sb. Ved. Pr., Vys. Sk. Chemickotechnol. Pardubice",
  ["sbornik. mathematics"] = "Sb. Math.",
  ["sbornik: mathematics"] = "Sb. Math.",
  ["sborník archivních prací"] = "Sb Arch Pr",
  ["sborník české geografické společnosti"] = "Sbs Ces Geogr Spol",
  ["sborník lékar̆ský"] = "Sb Lek",
  ["sborník národního musea v praze. a, historický = acta musei nationalis prague. a, historia"] = "Sb Nar Muz Praze Rada A Hist",
  ["sborník národního musea v praze. b, přírodovědný = acta musei nationalis pragae. b, hist. naturalis"] = "Sb Nar Muz Praze Rada B",
  ["sborník národního muzea v praze. řada c, literární historie = acta musei nationalis pragae. series c, historia litterarum"] = "Sb Nar Muz Pr Rada C Lit Hist",
  ["sborník prací filosofické fakulty brněnské university. c, řada historická"] = "Sb Pr Filos Fak Brnenske Univ Rada Hist",
  ["sborník prací filozofické fakulty brněnské univerzity. m, rada archeologická"] = "SborBrno",
  ["sborník pro pathofysiologii trávení a výz̆ivy; gastroenterologia bohema"] = "Sb Pathofysiol Traveni Vyz Gastroenterol Bohema",
  ["sborník vědeckých prací lékařské fakulty karlovy university v hradci králové"] = "Sb Ved Pr Lek Fak Karlovy Univerzity Hradci Kralove",
  ["sborník vědeckých prací lékařské fakulty karlovy univerzity v hradci králové. supplementum"] = "Sb Ved Pr Lek Fak Karlovy Univerzity Hradci Kralove Suppl",
  ["sborník prací filosofické fak. brnenské univ. rada archeol. klas."] = "SPFB",
  ["sborník prací filosofické fakulty brnenské university"] = "SPFB(klas)",
  ["sc [reports]. u.s. atomic energy commission"] = "SC Rep",
  ["sc reports"] = "SC Rep.",
  ["sc trodent"] = "SC Trodent",
  ["scada journal"] = "SCADA J",
  ["scalable computing and communications"] = "Scalable Comput. Commun.",
  ["scalpel & tongs : american journal of medical philately"] = "Scalpel Tongs",
  ["scalpel and quill"] = "Scapel Quill",
  ["scandinavian actuarial journal"] = "Scand Actuar J",
  ["scandinavian audiology"] = "Scand Audiol",
  ["scandinavian audiology. supplementum"] = "Scand Audiol Suppl",
  ["scandinavian cardiovascular journal"] = "Scand. Cardiovasc. J.",
  ["scandinavian cardiovascular journal : scj"] = "Scand Cardiovasc J",
  ["scandinavian cardiovascular journal. supplement"] = "Scand Cardiovasc J Suppl",
  ["scandinavian economic history review"] = "Scand. Econ. Hist. Rev.",
  ["scandinavian journal of caring sciences"] = "Scand J Caring Sci",
  ["scandinavian journal of child and adolescent psychiatry and psychology"] = "Scand J Child Adolesc Psychiatr Psychol",
  ["scandinavian journal of clinical and laboratory investigation"] = "Scand J Clin Lab Invest",
  ["scandinavian journal of clinical and laboratory investigation. supplement"] = "Scand. J. Clin. Lab. Invest. Suppl.",
  ["scandinavian journal of clinical and laboratory investigation. supplementum"] = "Scand J Clin Lab Invest Suppl",
  ["scandinavian journal of dental research"] = "Scand J Dent Res",
  ["scandinavian journal of design history"] = "Scan J Des Hist",
  ["scandinavian journal of development alternatives"] = "Scand J Dev Altern",
  ["scandinavian journal of disability research : sjdr"] = "Scand J Disabil Res",
  ["scandinavian journal of economics"] = "Scand. J. Econ.",
  ["scandinavian journal of food & nutrition"] = "Scand J Food Nutr",
  ["scandinavian journal of forest research"] = "Scand. J. For. Res.",
  ["scandinavian journal of gastroenterology"] = "Scand J Gastroenterol",
  ["scandinavian journal of gastroenterology. supplement"] = "Scand J Gastroenterol Suppl",
  ["scandinavian journal of haematology"] = "Scand J Haematol",
  ["scandinavian journal of haematology. supplementum"] = "Scand J Haematol Suppl",
  ["scandinavian journal of history"] = "Scand J Hist",
  ["scandinavian journal of immunology"] = "Scand J Immunol",
  ["scandinavian journal of immunology. supplement"] = "Scand J Immunol Suppl",
  ["scandinavian journal of infectious diseases"] = "Scand J Infect Dis",
  ["scandinavian journal of infectious diseases. supplementum"] = "Scand J Infect Dis Suppl",
  ["scandinavian journal of laboratory animal science"] = "Scand. J. Lab. Anim. Sci.",
  ["scandinavian journal of laboratory animal science = scand las nyt : official quarterly journal of the scandinavian federation for laboratory animal science"] = "Scand J Lab Anim Sci",
  ["scandinavian journal of medicine & science in sports"] = "Scand J Med Sci Sports",
  ["scandinavian journal of medicine and science in sports"] = "Scand. J. Med. Sci. Sports",
  ["scandinavian journal of metallurgy"] = "Scand. J. Metall.",
  ["scandinavian journal of occupational therapy"] = "Scand J Occup Ther",
  ["scandinavian journal of pain"] = "Scand J Pain",
  ["scandinavian journal of plastic and reconstructive surgery"] = "Scand J Plast Reconstr Surg",
  ["scandinavian journal of plastic and reconstructive surgery and hand surgery"] = "Scand J Plast Reconstr Surg Hand Surg",
  ["scandinavian journal of plastic and reconstructive surgery and hand surgery / nordisk plastikkirurgisk forening [and] nordisk klubb for handkirurgi"] = "Scand J Plast Reconstr Surg Hand Surg",
  ["scandinavian journal of plastic and reconstructive surgery and hand surgery. supplementum"] = "Scand J Plast Reconstr Surg Hand Surg Suppl",
  ["scandinavian journal of plastic and reconstructive surgery. supplementum"] = "Scand J Plast Reconstr Surg Suppl",
  ["scandinavian journal of primary health care"] = "Scand J Prim Health Care",
  ["scandinavian journal of primary health care. supplement"] = "Scand J Prim Health Care Suppl",
  ["scandinavian journal of psychology"] = "Scand J Psychol",
  ["scandinavian journal of public health"] = "Scand J Public Health",
  ["scandinavian journal of public health. supplement"] = "Scand J Public Health Suppl",
  ["scandinavian journal of rehabilitation medicine"] = "Scand J Rehabil Med",
  ["scandinavian journal of rehabilitation medicine. supplement"] = "Scand J Rehabil Med Suppl",
  ["scandinavian journal of respiratory diseases"] = "Scand J Respir Dis",
  ["scandinavian journal of respiratory diseases. supplementum"] = "Scand J Respir Dis Suppl",
  ["scandinavian journal of rheumatology"] = "Scand J Rheumatol",
  ["scandinavian journal of rheumatology. supplement"] = "Scand J Rheumatol Suppl",
  ["scandinavian journal of social medicine"] = "Scand J Soc Med",
  ["scandinavian journal of social medicine. supplementum"] = "Scand J Soc Med Suppl",
  ["scandinavian journal of social welfare"] = "Scand J Soc Welf",
  ["scandinavian journal of statistics"] = "Scand. J. Stat.",
  ["scandinavian journal of statistics, theory and applications"] = "Scand Stat Theory Appl",
  ["scandinavian journal of statistics. theory and applications"] = "Scand. J. Stat.",
  ["scandinavian journal of surgery"] = "Scand. J. Surg.",
  ["scandinavian journal of surgery : sjs : official organ for the finnish surgical society and the scandinavian surgical society"] = "Scand J Surg",
  ["scandinavian journal of the old testament : sjot"] = "Scand J Old Testam",
  ["scandinavian journal of theology"] = "STh",
  ["scandinavian journal of thoracic and cardiovascular surgery"] = "Scand J Thorac Cardiovasc Surg",
  ["scandinavian journal of thoracic and cardiovascular surgery. supplementum"] = "Scand J Thorac Cardiovasc Surg Suppl",
  ["scandinavian journal of trauma, resuscitation and emergency medicine"] = "Scand J Trauma Resusc Emerg Med",
  ["scandinavian journal of urology"] = "Scand J Urol",
  ["scandinavian journal of urology and nephrology"] = "Scand J Urol Nephrol",
  ["scandinavian journal of urology and nephrology. supplementum"] = "Scand J Urol Nephrol Suppl",
  ["scandinavian journal of work, environment & health"] = "Scand J Work Environ Health",
  ["scandinavian journal of work, environment and health"] = "Scand. J. Work. Environ. Health",
  ["scandinavian political studies"] = "Scan Polit Stud",
  ["scandinavian population studies"] = "Scand Popul Stud",
  ["scandinavian society of forensic odontology newsletter"] = "Scand. Soc. Forensic Odontol. Newsl.",
  ["scandinavian studies : publication of the society for the advancement of scandinavian study"] = "Scand Stud",
  ["scandinavian studies in law"] = "Scand Stud Law",
  ["scanning electron microscopy"] = "Scan Electron Microsc",
  ["scanning microscopy"] = "Scanning Microsc",
  ["scanning microscopy (elk grove village, il"] = "Scanning Microsc",
  ["scanning microscopy for nanotechnology"] = "Scanning Microsc. Nanotechnol.",
  ["scanning microscopy. supplement"] = "Scanning Microsc Suppl",
  ["scars, burns & healing"] = "Scars Burn Heal",
  ["scha journal"] = "SCHA J",
  ["schaffhauser beitrage zur vaterlandischen geschichte"] = "Schaffhauser Beitr Vaterl Gesch",
  ["schaffhauser beiträge zur vaterländischen geschichte"] = "Schaffhauser Beitr Vaterl Gesch",
  ["schaum’s outline series in computers"] = "Schaum’s Outline Ser. Comput.",
  ["schede medievali: rassegna a cura dell’officina di studi medievali"] = "SMed",
  ["schiff und zeit"] = "Schiff Zeit",
  ["schild von steier. beiträge zur steirischen vor- und frühgeschichte und münzkunde"] = "SchildStei",
  ["schizophrenia bulletin"] = "Schizophr Bull",
  ["schizophrenia bulletin open"] = "Schizophr Bull Open",
  ["schizophrenia research"] = "Schizophr Res",
  ["schizophrenia research and treatment"] = "Schizophr Res Treatment",
  ["schizophrenia research. cognition"] = "Schizophr Res Cogn",
  ["schmalenbachs zeitschrift fur betriebswirtschaftliche forschung"] = "Schmalenbachs Z. betriebswirtsch. Forsch.",
  ["schmerz (berlin, germany)"] = "Schmerz",
  ["schmollers jahrbuch für wirtschafts- und sozialwissenschaften"] = "Schmollers Jahr.",
  ["schmollers jahrbuch: zeitschrift für wirtschafts- und sozialwissenschaften"] = "Schmollers Jahr.: Z. Wirtsch. Sozialwissen.",
  ["scholarly inquiry for nursing practice"] = "Sch Inq Nurs Pract",
  ["scholarpedia journal"] = "Scholarpedia J",
  ["scholars academic journal of biosciences"] = "Scholars Acad J Biosci",
  ["scholarship and practice of undergraduate research"] = "Scholarsh Pract Undergrad Res",
  ["scholarship of teaching and learning in psychology"] = "Scholarsh Teach Learn Psychol",
  ["scholastic update (teachers' ed.)"] = "Scholast Update",
  ["scholastic update (teachers’ ed.)"] = "Scholast Update",
  ["scholia. natal studies in classical antiquity"] = "Scholia",
  ["school community journal"] = "Sch Comm J",
  ["school dental service gazette, new zealand"] = "Sch Dent Serv Gaz N Z",
  ["school dental services gazette"] = "Sch. Dent. Serv. Gaz. N. Z.",
  ["school effectiveness and school improvement : an international journal of research, policy and practice"] = "Sch Eff Sch Improv",
  ["school health review"] = "Sch Health Rev",
  ["school mental health"] = "School Ment Health",
  ["school nurse news"] = "School Nurse News",
  ["school of medicine. university of southern california. school of medicine"] = "Univ South Calif Bull Sch Med",
  ["school psychology (washington, d.c.)"] = "Sch Psychol",
  ["school psychology forum, research in practice"] = "Sch Psychol Forum Res Pract",
  ["school psychology international"] = "Sch Psychol Int",
  ["school psychology quarterly : the official journal of the division of school psychology, american psychological association"] = "Sch Psychol Q",
  ["school psychology review"] = "School Psych Rev",
  ["school science and mathematics"] = "Sch Sci Math",
  ["school science review"] = "Sch. Sci. Rev.",
  ["school social work journal"] = "Sch Soc Work J",
  ["schriften aus der forstlichen fakultät der universität göttingen und der niedersächsischen forstlichen versuchsanstalt"] = "Schr. Forstl. Fak. Univ. Gött. Niedersächs.  forstl. Vers.anst.",
  ["schriften der carl friedrich von siemens stiftung"] = "Schr. Carl Friedrich von Siemens Stift.",
  ["schriften der koenigsberger gelehrten gesellschaft, naturwissenschaftliche klasse"] = "Schr. Koenigsb. Gelehrten Ges., Naturwiss. Kl.",
  ["schriften der konigsberger gelehrten gesellschaft"] = "Schr. Konigsb. Gelehrten Ges.",
  ["schriften des adam-ries-bundes annaberg-buchholz"] = "Schrift. Adam-Ries-Bundes Annaberg-Buchholz",
  ["schriften des museums der universitat tubingen (mut)"] = "Schrift. Mus. Univ. Tubingen (MUT)",
  ["schriften des vereines zur verbreitung naturwissenschaftlicher kenntnisse in wien"] = "Schr. Naturwissen. Verbr. Naturwissen. Kennt. Wien",
  ["schriften des zentrums für archäologie und kulturgeschichte des schwarzmeerraumes"] = "ZAKSSchriften",
  ["schriften zu genetischen ressourcen"] = "Schr Genet Ressour",
  ["schriften zur geschichte und kultur des alten orients"] = "Schr Gesch Kult Alten Orients",
  ["schriften zur malakozoologie aus dem haus der natur, cismar"] = "Schr Malakozool Haus Nat Cismar",
  ["schriften zur wirtschaftsinformatik"] = "Schr. Wirtsch.inform.",
  ["schriften zur wirtschaftspolitik. neue folge"] = "Schrift. Wirtschaftspolit. N.F.",
  ["schriften zur wirtschaftswissenschaftlichen forschung"] = "Schriften Wirtschaftswiss. Forsch.",
  ["schriftenreihe aus dem gebiete des offentlichen gesundheitswesens"] = "Schriftenr. Geb. Off. Gesundheitswes.",
  ["schriftenreihe aus dem gebiete des öffentlichen gesundheitswesens"] = "Schriftenr Geb Off Gesundheitswes",
  ["schriftenreihe bayerisches landesamt für umweltschutz"] = "Schr.reihe Bayer. Landesamt Umweltschutz",
  ["schriftenreihe der badischen forstlichen versuchsanstalt freiburg"] = "Schr.reihe Bad. Forstl. Vers.anst., Freibg.",
  ["schriftenreihe der forstwissenschaftlichen fakultät der albert-ludwigs-universität freiburg i. br."] = "Schr.reihe Forstwiss. Fak. Albert-Ludwigs-Univ.  Freibg. i.B.",
  ["schriftenreihe der institut fur mathematik bei der deutschen akademie der wissenschaften zu berlin, reihe a. reine mathematik"] = "Schriftenreihe der Institut fur Mathematik bei der Deutschen Akademie der Wissenschaften zu Berlin, Reihe A. Reine Mat",
  ["schriftenreihe der institut fur mathematik bei der deutschen akademie der wissenschaften zu berlin, reihe b. angewandte mathematik und mechanik"] = "Schriftenreihe der Institut fur Mathematik bei der Deutschen Akademie der Wissenschaften zu Berlin, Reihe B. Angewandt",
  ["schriftenreihe der institute fur systemdynamik (isd) und optische systeme (ios)"] = "Schriftenr. Inst. Syst.dyn. (ISD) Opt. Syst. (IOS)",
  ["schriftenreihe der landesanstalt für ökologie, landschaftsentwicklung und forstplanung nordrhein-westfalen"] = "Schr.reihe Landesanst. Ökol. Landsch.entwickl. Forstplan. Nordrh.-Westfal.",
  ["schriftenreihe der landesforstverwaltung baden-württemberg"] = "Schr.reihe Landesforstverwalt. Baden-Württ.",
  ["schriftenreihe der landesstelle für naturschutz und landschaftspflege in nordrhein-westfalen"] = "Schr.reihe Landesstelle Nat.schutz Landsch.pfl.  Nordrh.-Westfal.",
  ["schriftenreihe der pädagogischen hochschule heidelberg"] = "Schriftenreihe Pädagog. Hochsch. Heidelberg",
  ["schriftenreihe des bayerischen staatsministeriums für ernährung, landwirtschaft und forsten"] = "Schr.reihe Bayer. Staatsminist. Ernähr.  Landwirtsch. Forsten",
  ["schriftenreihe des deutschen rates für landespflege"] = "Schr.reihe Dtsch. Rat Landespfl.",
  ["schriftenreihe des institutes für landespflege der universität freiburg"] = "Schr.reihe Inst. Landespfl. Univ. Freibg.",
  ["schriftenreihe des instituts fur empirische wirtschaftsforschung der universitat zurich"] = "Schriftenreihe Inst. Empirische Wirtschaftsforsch. Univ. Zurich",
  ["schriftenreihe des mathematischen instituts der universität münster, 3"] = "Schriftenreihe Math. Inst. Univ. Münster 3. Ser.",
  ["schriftenreihe des vereins fur wasser-, boden- und lufthygiene"] = "Schriftenr Ver Wasser Boden Lufthyg",
  ["schriftenreihe des vereins fur wasser-, boden-, und lufthygiene"] = "Schriftenr. Ver. Wasser. Boden. Lufthyg.",
  ["schriftenreihe des vereins für wasser-, boden- und lufthygiene"] = "Schriftenr Ver Wasser Boden Lufthyg",
  ["schriftenreihe erfahrungen des deutschen sanitätsdienstes im zweiten weltkrieg"] = "Schriftenr Erfahr Dtsch Sanit Zweiten Weltkrieg",
  ["schriftenreihe fur geologische wissenschaften"] = "Schriftenr. Geol. Wissen.",
  ["schriftenreihe für geschichte der naturwissenschaften, technik und medizin"] = "NTM",
  ["schriftenreihe für landschaftspflege und naturschutz"] = "Schr.reihe Landsch.pfl. Nat.schutz",
  ["schriftenreihe für vegetationskunde"] = "Schr.reihe Veg.kd.",
  ["schriftenreihe lebensraum vorarlberg"] = "Schr.reihe Lebensraum Vorarlberg",
  ["schriftenreihe neurologie"] = "Schriftenr Neurol",
  ["schriftenreihe neurologie (neurology series)"] = "Schriftenr. Neurol.",
  ["schriftenreihe neurologie. neurology series"] = "Schriftenr Neurol",
  ["schriftenreihe umwelt"] = "Schr.reihe Umw.",
  ["schriftenreihe zentralblatt fur arbeitsmedizin arbeitsschutz und prophylaxe"] = "Schriftenr. Zentralbl. Arbeitsmed. Arbeitsschutz Prophyl.",
  ["schriftenreihe zentralblatt fur arbeitsmedizin, arbeitsschutz und prophylaxe"] = "Schriftenr Zentralbl Arbeitsmed Arbeitsschutz Prophyl",
  ["schriftenreihe zentralblatt fur arbeitsmedizin, arbeitsschutz, prophylaxe, und ergonomie"] = "Schriftenr. Zentralbl. Arbeitsmed. Arbeitsschutz. Prophyl. Ergonomie",
  ["schriftenreihe zentralblatt für arbeitsmedizin, arbeitsschutz und prophylaxe"] = "Schriftenr Zentralbl Arbeitsmed Arbeitsschutz Prophyl",
  ["schriftenreihe zentralblatt für arbeitsmedizin, arbeitsschutz, prophylaxe, und ergonomie"] = "Schriftenr Zentralbl Arbeitsmed Arbeitsschutz Prophyl Ergonomie",
  ["schriftenreihe zur geschichte der versammlungen deutscher naturforscher und aerzte"] = "Schriftenr Gesch Versamml Dtsch Naturforsch Arzte",
  ["schriftenreihe “die versicherung”"] = "Schriftenreihe Versicher.",
  ["schriftenreihe “wissenschaft und frieden”"] = "Schriftenreihe Wiss. Frieden",
  ["schriftenreihe. altnurnberger landschaft e.v"] = "Altnurnb Landsch",
  ["schriftenreihe. altnürnberger landschaft e.v"] = "Altnurnb Landsch",
  ["schriftenreihe. gesellschaft der freunde der medizinischen hochschule hannover"] = "Schriftenr Ges Freunde Med Hochsch Hannover",
  ["schriftenreihe. medizinische hochschule hannover. gesellschaft der freunde"] = "Schriftenr Ges Freunde Med Hochsch Hannover",
  ["schs studies"] = "SCHS Stud",
  ["schumannia : ein sonderheft der deutschen kakteen-gesellschaft e.v"] = "Schumannia",
  ["schweiss- und prueftechnik"] = "Schweiss Prueftech.",
  ["schweizer archiv der tierheilkunde"] = "Schweiz. Arch. Tierheilkd.",
  ["schweizer archiv fur neurologie und psychiatrie"] = "Schweiz. Arch. Neurol. Psychiatr.",
  ["schweizer archiv fur neurologie und psychiatrie (zurich, switzerland : 1985)"] = "Schweiz Arch Neurol Psychiatr",
  ["schweizer archiv fur neurologie und psychiatrie. archives suisses de neurologie et de psychiatrie. archivio svizzero di neurologia e psichiatria"] = "Schweiz Arch Neurol Psychiatr",
  ["schweizer archiv fur neurologie, neurochirurgie und psychiatrie"] = "Schweiz. Arch. Neurol. Neurochir. Psychiatr.",
  ["schweizer archiv fur tierheilkunde"] = "Schweiz Arch Tierheilkd",
  ["schweizer archiv für neurologie und psychiatrie (zurich, switzerland : 1985)"] = "Schweiz Arch Neurol Psychiatr (1985)",
  ["schweizer archiv für neurologie und psychiatrie. archives suisses de neurologie et de psychiatrie. archivio svizzero di neurologia e psichiatria"] = "Schweiz Arch Neurol Psychiatr",
  ["schweizer archiv für neurologie, neurochirurgie und psychiatrie = archives suisses de neurologie, neurochirurgie et de psychiatrie"] = "Schweiz Arch Neurol Neurochir Psychiatr",
  ["schweizer archiv für tierheilkunde"] = "Schweiz Arch Tierheilkd",
  ["schweizer chemiker-zeitung und technik-industrie"] = "Schweizer Chem Ztg Tech Ind",
  ["schweizer förster, der"] = "Schweiz. Förster",
  ["schweizer holz-börse"] = "Schweiz. Holz-Börse",
  ["schweizer holzwirtschaft"] = "Schweiz. Holzwirtsch.",
  ["schweizer holzzeitung"] = "Schweiz. Holzztg.",
  ["schweizer ingenieur und architekt"] = "Schweiz. Ing. Archit.",
  ["schweizer landtechnik"] = "Schweiz. Landtech.",
  ["schweizer monatshefte"] = "Schweiz Monatsh",
  ["schweizer monatsschrift fur zahnmedizin"] = "Schweiz. Monatsschr. Zahnmed.",
  ["schweizer monatsschrift für zahnmedizin = revue mensuelle suisse d'odonto-stomatologie = rivista mensile svizzera di odontologia e stomatologia"] = "Schweiz Monatsschr Zahnmed",
  ["schweizer münzblätter"] = "SchwMueBl",
  ["schweizer naturschutz"] = "Schweiz. Nat.schutz",
  ["schweizer volkskunde"] = "Schweiz Volkskd",
  ["schweizer wald"] = "Schweiz. Wald",
  ["schweizerische aktuarvereinigung"] = "Schweiz. Aktuarver. Mitt.",
  ["schweizerische apotheker zeitung"] = "Schweiz Apoth Ztg",
  ["schweizerische apotheker-zeitung"] = "Schweiz. Apoth. Ztg.",
  ["schweizerische arztezeitung. bulletin des medecins suisses. bollettino dei medici svizzeri"] = "Schweiz Arzteztg",
  ["schweizerische ärztezeitung für standesfragen. bulletin professionnel des médecins suisses. bollettino dei medici svizzeri per interessi professionali"] = "Schweiz Arzteztg Standesfr",
  ["schweizerische ärztezeitung. bulletin des médecins suisses. bollettino dei medici svizzeri"] = "Schweiz Arzteztg",
  ["schweizerische beiträge zur dendrologie"] = "Schweiz. Beitr. Dendrol.",
  ["schweizerische blätter für krankenpflege. revue suisse des infirmières"] = "Schweiz Bl Krankenpfl",
  ["schweizerische drogisten zeitung"] = "Schweiz Drog Ztg (1901)",
  ["schweizerische gesellschaft für phytomedizin, info"] = "Schweiz.  Ges. Phytomed., Info",
  ["schweizerische interessengemeinschaft industrieholz, anleitung"] = "Schweiz. Interessengem. Ind.holz, Anleit.",
  ["schweizerische krankenkassen-zeitung"] = "Schweiz Krankenkassen Ztg",
  ["schweizerische laboratoriums zeitschrift"] = "Schweiz. Lab. Z.",
  ["schweizerische landwirtschaftliche forschung"] = "Schweiz. landwirtsch. Forsch.",
  ["schweizerische medizinische wochenschrift"] = "Schweiz Med Wochenschr",
  ["schweizerische medizinische wochenschrift. journal suisse de medecine"] = "Schweiz Med Wochenschr",
  ["schweizerische medizinische wochenschrift. supplementum"] = "Schweiz Med Wochenschr Suppl",
  ["schweizerische mineralogische und petrographische mitteilungen"] = "Schweiz. Mineral. Petrogr. Mitt.",
  ["schweizerische monatsschrift fur zahnmedizin"] = "Schweiz. Monatsschr. Zahnmed.",
  ["schweizerische monatsschrift für zahnheilkunde = revue mensuelle suisse d'odonto-stomatologie"] = "SSO Schweiz Monatsschr Zahnheilkd",
  ["schweizerische monatsschrift für zahnmedizin = revue mensuelle suisse d'odonto-stomatologie = rivista mensile svizzera di odontologia e stomatologia"] = "Schweiz Monatsschr Zahnmed (1984)",
  ["schweizerische nationalbank quartalsheft"] = "Schweiz. Nationalbank Quart.",
  ["schweizerische numismatische rundschau"] = "SNR",
  ["schweizerische rundschau fur medizin praxis"] = "Schweiz. Rundsch. Med. Prax.",
  ["schweizerische rundschau für medizin praxis = revue suisse de médecine praxis"] = "Schweiz Rundsch Med Prax",
  ["schweizerische schreinerzeitung"] = "Schweiz. Schreinerztg.",
  ["schweizerische wochenschrift fur chemie und pharmacie"] = "Schweiz. Wochenschr. Chem. Pharm.",
  ["schweizerische zeitschrift fur geschichte. revue suisse d’histoire. rivista storica svizzera"] = "Schweiz Z Gesch",
  ["schweizerische zeitschrift fur gynakologie und geburtshilfe. revue suisse de gynecologie et d’obstetrique"] = "Schweiz Z Gynakol Geburtshilfe",
  ["schweizerische zeitschrift fur homoopathie. journal suisse d’homoeopathie"] = "Schweiz Z Homoopath",
  ["schweizerische zeitschrift fur medizin und traumatologie"] = "Schweiz. Z. Med. Traumatol.",
  ["schweizerische zeitschrift fur pathologie und bakteriologie. revue suisse de pathologie et de bacteriologie"] = "Schweiz Z Pathol Bakteriol",
  ["schweizerische zeitschrift fur psychologie und ihre andwendungen. revue suisse de psychologie, pure et appliquee"] = "Schweiz Z Psychol Anwend",
  ["schweizerische zeitschrift fur psychologie und ihre anwendungen"] = "Schweiz. Z. Psychol. Anwend.",
  ["schweizerische zeitschrift fur soziologie. revue suisse de sociologie"] = "Schweiz Z Soziol",
  ["schweizerische zeitschrift fur sportmedizin"] = "Schweiz. Z. Sportmed.",
  ["schweizerische zeitschrift fur tuberkulose und pneumonologie. revue suisse de la tuberculose et de pneumonologie. rivista svizzera della tubercolosi e della pneumonologia"] = "Schweiz Z Tuberc Pneumonol",
  ["schweizerische zeitschrift fur tuberkulose. revue suisse de la tuberculose. rivista svizzera della tubercolosi"] = "Schweiz Z Tuberk",
  ["schweizerische zeitschrift für geschichte. revue suisse d'histoire. rivista storica svizzera"] = "Schweiz Z Gesch",
  ["schweizerische zeitschrift für gynäkologie und geburtshilfe. revue suisse de gynécologie et d'obstétrique"] = "Schweiz Z Gynakol Geburtshilfe",
  ["schweizerische zeitschrift für homöopathie. journal suisse d'homoeopathie"] = "Schweiz Z Homoopath",
  ["schweizerische zeitschrift für medizin und traumatologie = revue suisse pour médecine et traumatologie"] = "Schweiz Z Med Traumatol",
  ["schweizerische zeitschrift für pathologie und bakteriologie. revue suisse de pathologie et de bactériologie"] = "Schweiz Z Pathol Bakteriol",
  ["schweizerische zeitschrift für pilzkunde"] = "Schweiz Z Pilzkd",
  ["schweizerische zeitschrift für psychologie und ihre andwendungen. revue suisse de psychologie, pure et appliquée"] = "Schweiz Z Psychol Anwend",
  ["schweizerische zeitschrift für soziologie. revue suisse de sociologie"] = "Schweiz Z Soziol",
  ["schweizerische zeitschrift für sportmedizin"] = "Schweiz Z Sportmed",
  ["schweizerische zeitschrift für tuberkulose und pneumonologie. revue suisse de la tuberculose et de pneumonologie. rivista svizzera della tubercolosi e della pneumonologia"] = "Schweiz Z Tuberc Pneumonol",
  ["schweizerische zeitschrift für tuberkulose. revue suisse de la tuberculose. rivista svizzera della tubercolosi"] = "Schweiz Z Tuberk",
  ["schweizerische zeitschrift für volkswirtschaft und statistik"] = "Schweiz Z Volkswirtsch Stat",
  ["schweizerische zeitschrift für zahnheilkunde. journal suisse d'odontologie. odontologia svizzera. swiss journal of dental surgery"] = "Schweiz Z Zahnheilkd",
  ["schweizerische zeitschrift für forstwesen"] = "Schweiz. Z. Forstwes.",
  ["schweizerische zeitschrift für geschichte"] = "SZG",
  ["schweizerische zeitschrift für hydrologie"] = "Schweiz. Z. Hydrol.",
  ["schweizerische zeitschrift für pilzkunde"] = "Schweiz. Z. Pilzkd.",
  ["schweizerische zeitschrift für volkswirtschaft und statistik"] = "Schweiz. Z. Volkswirtsch. Statist.",
  ["schweizerisches medizinisches jahrbuch"] = "Schweiz Med Jahrb",
  ["schwestern revue"] = "Schwest Rev",
  ["sci nursing"] = "SCI Nurs.",
  ["sci nursing : a publication of the american association of spinal cord injury nurses"] = "SCI Nurs",
  ["sci sustainability"] = "SCI Sustainability",
  ["sci. adv. mater. science of advanced materials"] = "Sci. Adv. Mater. Science of Advanced Materials",
  ["sciamvs. sources and commentaries in exact sciences"] = "SCIAMVS",
  ["science & christian belief"] = "Sci Christ Belief",
  ["science & diplomacy"] = "Sci Dipl",
  ["science & education"] = "Sci. Educ.",
  ["science & government report"] = "Sci Gov Rep",
  ["science & justice"] = "Sci. Justice",
  ["science & justice : journal of the forensic science society"] = "Sci Justice",
  ["science & medicine"] = "Sci Med (Phila)",
  ["science & practice perspectives"] = "Sci Pract Perspect",
  ["science & practice perspectives / a publication of the national institute on drug abuse, national institutes of health"] = "Sci Pract Perspect",
  ["science & public policy"] = "Sci Public Policy",
  ["science & recherche odontostomatologiques"] = "Sci Rech Odontostomatol",
  ["science & sports"] = "Sci Sports",
  ["science & technology education library"] = "Sci. Technol. Educ. Libr.",
  ["science & technology in china"] = "Sci Technol China",
  ["science & technology libraries"] = "Sci Technol Libr (New York, NY)",
  ["science & technology review (livermore, calif.)"] = "Sci Technol Rev",
  ["science (new york, n.y.)"] = "Science",
  ["science (washington, dc, united states)"] = "Science (Washington, DC, U. S.)",
  ["science activities"] = "Sci Act",
  ["science advances"] = "Sci Adv",
  ["science and culture"] = "Sci Cult",
  ["science and engineering ethics"] = "Sci Eng Ethics",
  ["science and engineering of composite materials"] = "Sci. Eng. Compos. Mater.",
  ["science and fiction"] = "Sci. Fict.",
  ["science and its conceptual foundations"] = "Sci. Concept. Found.",
  ["science and justice"] = "Sci. Justice",
  ["science and medicine in football"] = "Sci. Med. Football",
  ["science and philosophy in the twentieth century"] = "Sci. Philos. Twent. Century Basic Works Log. Empir.",
  ["science and society"] = "Sci Soc",
  ["science and technology for energy transition"] = "Sci. Technol. Energy Transition",
  ["science and technology for the built environment"] = "Sci Technol Built Environ",
  ["science and technology of advanced materials"] = "Sci Technol Adv Mater",
  ["science and technology of advanced materials: methods"] = "Sci. Technol. Adv. Mater.: Methods",
  ["science and technology of archaeological research"] = "Sci Technol Archaeol Res",
  ["science and technology of energetic materials"] = "Sci. Technol. Energetic Mater.",
  ["science and technology of materials"] = "Sci. Technol. Mater.",
  ["science and technology of nuclear installations"] = "Sci. Technol. Nucl. Install.",
  ["science and technology of welding and joining"] = "Sci. Technol. Weld. Joining",
  ["science and technology of welding and joining /font>"] = "Sci. Technol. Weld. Joining",
  ["science as culture"] = "Sci Cult (Lond)",
  ["science bulletin"] = "Sci Bull (Beijing)",
  ["science bulletin (kuo chia kʻo hsüeh wei yüan hui)"] = "Sci Bull Sci Found Philipp",
  ["science bulletin of josai university"] = "Sci. Bull. Josai Univ.",
  ["science bulletin of the faculty of education"] = "Sci. Bull. Fac. Ed. Nagasaki Univ.",
  ["science china chemistry"] = "Sci. China Chem.",
  ["science china earth sciences"] = "Sci. China Earth Sci.",
  ["science china information sciences"] = "Sci. China Inf. Sci.",
  ["science china life sciences"] = "Sci. China Life Sci.",
  ["science china materials"] = "Sci. China Mater.",
  ["science china mathematics"] = "Sci. China Math.",
  ["science china physics, mechanics & astronomy"] = "Sci. China Phys., Mech. Astron.",
  ["science china technological sciences"] = "Sci. China Technol. Sci.",
  ["science china. chemistry"] = "Sci China Chem",
  ["science china. earth sciences"] = "Sci China Earth Sci",
  ["science china. information sciences"] = "Sci. China Inf. Sci.",
  ["science china. life sciences"] = "Sci China Life Sci",
  ["science china. mathematics"] = "Sci China Math",
  ["science china. physics, mechanics & astronomy"] = "Sci China Phys Mech Astron",
  ["science china. technological sciences"] = "Sci China Technol Sci",
  ["science china: chemistry"] = "Sci. China: Chem.",
  ["science communication"] = "Sci Commun",
  ["science digest"] = "Sci Dig",
  ["science digest (new york, n.y.)"] = "Sci Dig (NY)",
  ["science diliman"] = "Sci Diliman",
  ["science dimension"] = "Sci. Dimens.",
  ["science du sol"] = "Sci. Sol",
  ["science education"] = "Sci Educ",
  ["science education & civic engagement : an international journal"] = "Sci Educ Civ Engagem",
  ["science educator : the national science education leadership association journal"] = "Sci Educ (Arlingt)",
  ["science et recherche odontostomatologiques"] = "Sci. Rech. Odontostomatol.",
  ["science for conservation (wellington)"] = "Sci Conserv",
  ["science for the people"] = "Sci People",
  ["science forum"] = "Sci Forum",
  ["science foundation in china"] = "Sci. Found. China",
  ["science illustrated"] = "Sci Illus",
  ["science immunology"] = "Sci Immunol",
  ["science in china (scientia sinica)"] = "Sci. China Ser. A",
  ["science in china series c: life sciences"] = "Sci. China, Ser. C: Life Sci.",
  ["science in china series d: earth sciences"] = "Sci. China, Ser. D: Earth Sci.",
  ["science in china series e: technological sciences"] = "Sci. China Ser. E: Technol. Sci.",
  ["science in china series f: information sciences"] = "Sci. China, Ser. F: Info. Sci.",
  ["science in china series g: physics, mechanics & astronomy"] = "Sci. China, Ser. G: Phys., Mech., Astron.",
  ["science in china, series a: mathematics"] = "Sci. China, Ser. A Math.",
  ["science in china, series b: chemistry"] = "Sci. China, Ser. B Chem.",
  ["science in china, series b: chemistry, life sciences, & earth sciences"] = "Sci. China, Ser. B: Chem.",
  ["science in china, series c: life sciences"] = "Sci. China, Ser. C Life Sci.",
  ["science in china, series d: earth sciences"] = "Sci. China, Ser. D Earth Sci.",
  ["science in china, series e: engineering & materials science"] = "Sci. China, Ser. E Eng. Mater. Sci.",
  ["science in china, series e: technological sciences"] = "Sci. China, Ser. E: Tech. Sci.",
  ["science in china, series g: physics mechanics and astronomy"] = "Sci. China, Ser. G",
  ["science in china, series g: physics, mechanics and astronomy"] = "Sci. China G: Phys. Mech. Astron.",
  ["science in china. series a, mathematics"] = "Sci China Ser A Math",
  ["science in china. series a, mathematics, physics, astronomy"] = "Sci China Ser A Math Phys Astron",
  ["science in china. series b, chemistry"] = "Sci China B Chem",
  ["science in china. series b, chemistry, life sciences & earth sciences"] = "Sci China B",
  ["science in china. series b, chemistry, life sciences and earth sciences"] = "Sci. China B",
  ["science in china. series c, life sciences"] = "Sci China C Life Sci",
  ["science in china. series c, life sciences / chinese academy of sciences"] = "Sci China C Life Sci",
  ["science in china. series d, earth sciences"] = "Sci China Ser D Earth Sci",
  ["science in context"] = "Sci Context",
  ["science in new guinea"] = "Sci New Guinea",
  ["science in progress"] = "Sci Prog (New Haven)",
  ["science international (lahore)"] = "Sci Int (Lahore)",
  ["science letters"] = "Sci. Lett.",
  ["science letters journal"] = "Sci Lett J",
  ["science museum group journal"] = "Sci Mus Group J",
  ["science networks"] = "Sci. Networks Hist. Stud.",
  ["science networks. historical studies"] = "Sci. Networks Hist. Stud.",
  ["science news"] = "Sci News",
  ["science news letter"] = "Sci News Lett",
  ["science of advanced materials"] = "Sci Adv Mater",
  ["science of aging knowledge environment : sage ke"] = "Sci Aging Knowledge Environ",
  ["science of biology journal"] = "Sci Bio J",
  ["science of computer programming"] = "Sci Comput Program",
  ["science of nature"] = "Sci. Nat.",
  ["science of remote sensing"] = "Sci. Remote Sens.",
  ["science of sintering"] = "Sci. Sintering",
  ["science of synthesis, knowledge updates"] = "Sci. Synth., Knowl. Updates",
  ["science of the total environment"] = "Sci. Total Environ.",
  ["science postprint"] = "Sci Postprint",
  ["science progress"] = "Sci Prog",
  ["science progress in the twentieth century"] = "Sci Prog Twent Century",
  ["science reports of kagoshima university"] = "Sci. Rep. Kagoshima Univ.",
  ["science reports of the faculty of education"] = "Sci. Rep. Fac. Ed. Gifu Univ. Natur. Sci.",
  ["science reports of the hirosaki university"] = "Sci. Rep. Hirosaki Univ.",
  ["science reports of the research institute, tohoku university. ser. c, medicine"] = "Sci. Rep. Res. Inst. Tohoku Univ. [Med.]",
  ["science reports of the yokohama national university"] = "Sci. Rep. Yokohama Nat. Univ. Sect. I Math. Phys. Chem.",
  ["science robotics"] = "Sci Robot",
  ["science scope (washington, d.c.)"] = "Sci Scope (Wash D C)",
  ["science signaling"] = "Sci Signal",
  ["science studies"] = "Sci Stud (St Bonaventure)",
  ["science studies (helsinki, finland)"] = "Sci Stud",
  ["science studies/wetenschapsstudies"] = "Sci. Stud./Wet.",
  ["science talks"] = "Sci. Talks",
  ["science teacher (normal, ill.)"] = "Sci Teach",
  ["science translational medicine"] = "Sci Transl Med",
  ["science vision"] = "Sci Vis",
  ["science's stke : signal transduction knowledge environment"] = "Sci STKE",
  ["science, medicine and man"] = "Sci Med Man",
  ["science, medicine, and man"] = "Sci. Med. Man",
  ["science, medicine, and technology in east asia"] = "Sci Med Technol East Asia",
  ["science, technology & human values"] = "Sci Technol Human Values",
  ["science, technology and arts research"] = "Sci Technol Arts Res",
  ["science, technology and society"] = "Sci., Technol. Soc.",
  ["science, technology, and management series"] = "Sci. Technol. Manag. Ser.",
  ["science, technology, and medicine in ancient cultures"] = "Sci. Technol. Med. Anc. Cult.",
  ["scienceasia : journal of the science society of thailand"] = "Sci Asia",
  ["scienceopen research"] = "ScienceOpen Res.",
  ["sciences de la musique. serie etudes"] = "Sci. Musique Ser. Etudes",
  ["sciences des aliments"] = "Sci Aliments",
  ["sciences des structures et de la matiere"] = "Sci. Struct. Matiere",
  ["sciences economiques"] = "Sci. Econom.",
  ["sciences et avenir"] = "Sci Avenir",
  ["sciences et savoirs"] = "Sci. Savoirs",
  ["sciences et techniques en perspective"] = "Sci Tech Perspect",
  ["sciences et techniques en perspective. iie serie"] = "Sci. Tech. Perspect. (2)",
  ["sciences geologiques: memoire"] = "Sci Geol Mem",
  ["sciences géologiques: mémoire"] = "Sci Geol Mem",
  ["sciences in cold and arid regions"] = "Sci Cold Arid Reg",
  ["sciences sociales et sante"] = "Sci Soc Sante",
  ["sciences sociales et santé"] = "Sci Soc Sante",
  ["sciences: concepts et problemes"] = "Sci. Concepts Probl.",
  ["science’s stke"] = "Sci. STKE",
  ["science’s stke : signal transduction knowledge environment"] = "Sci STKE",
  ["scientia (bristol, england)"] = "Scientia (Bristol)",
  ["scientia agraria"] = "Sci Agrar",
  ["scientia agricola"] = "Sci Agric",
  ["scientia canadensis"] = "Sci Can",
  ["scientia forestalis"] = "Sci For",
  ["scientia fungorum"] = "Sci Fungorum",
  ["scientia genetica; periodico di genetica per i paesi latini"] = "Sci Genet",
  ["scientia graeco-arabica"] = "Sci. Graeco-Arabica",
  ["scientia horticulturae"] = "Sci Hortic",
  ["scientia iranica"] = "Sci. Iran.",
  ["scientia iranica. transactions d, computer science & engineering, electrical engineering"] = "Sci Iran D Comput Sci Eng Electr Eng",
  ["scientia marina"] = "Sci Mar",
  ["scientia medica"] = "Sci Med (Porto Alegre)",
  ["scientia medica italica. english ed"] = "Sci Med Ital",
  ["scientia pharmaceutica"] = "Sci. Pharm.",
  ["scientia sinica"] = "Sci Sin",
  ["scientia sinica chimica"] = "Sci. Sin. Chim.",
  ["scientia sinica. series b, chemical, biological, agricultural, medical & earth sciences"] = "Sci Sin B",
  ["scientia sinica. series b, chemical, biological, agricultural, medical & earth sciences / chung-kuo k’o hsueh yuan, chu pan"] = "Sci Sin [B]",
  ["scientia sinica. series b, chemical, biological, agricultural, medical and earth sciences"] = "Sci. Sin. [B]",
  ["scientia. serie physico-mathematique"] = "Scientia. Ser. Phys.-Math.",
  ["scientia. series a. mathematical sciences. new series"] = "Sci. Ser. A Math. Sci. (N.S.)",
  ["scientia; rivista di scienza"] = "Scientia",
  ["scientiae mathematicae japonicae"] = "Sci. Math. Jpn.",
  ["scientific african"] = "Sci Afr",
  ["scientific american"] = "Sci Am",
  ["scientific american library paperback"] = "Sci. Amer. Lib. Paperback",
  ["scientific american mind"] = "Sci Am Mind",
  ["scientific american neurology"] = "Sci Am Neurol",
  ["scientific and educational bulletin"] = "Sci Educ Bull",
  ["scientific and educational journal"] = "Sci Educ J",
  ["scientific and engineering computation"] = "Sci. Engrg. Comput.",
  ["scientific annals of computer science"] = "Sci. Ann. Comput. Sci.",
  ["scientific annals of the danube delta institute"] = "Sci. Ann. Danube Delta Inst.",
  ["scientific basis of medicine annual reviews"] = "Sci. Basis Med. Annu. Rev.",
  ["scientific computation"] = "Sci. Comput.",
  ["scientific data"] = "Sci Data",
  ["scientific drilling"] = "Sci. Drill.",
  ["scientific issues. jan dlugosz university in czestochowa. mathematics"] = "Sci. Issues Jan Dlugosz Univ. Czest. Math.",
  ["scientific journal of medical research"] = "Sci J Med Res",
  ["scientific mining journal"] = "Sci. Min. J.",
  ["scientific modeling and simulation"] = "Sci. Model. Simul.",
  ["scientific modeling and simulation smns"] = "Sci. Model. Simul. SMNS",
  ["scientific online letters on the atmosphere"] = "SOLA",
  ["scientific pages of pulmonology"] = "Sci Pages Pulmonol",
  ["scientific papers (university of kansas. natural history museum)"] = "Sci Pap Univ Kansas Nat Hist Mus",
  ["scientific phone apps and mobile devices"] = "Sci. Phone Apps Mobile Devices",
  ["scientific proceedings of the cardiff medical society"] = "Sci Proc Cardiff Med Soc",
  ["scientific programming"] = "Sci Program",
  ["scientific psychology"] = "Sci. Psychol.",
  ["scientific reports"] = "Sci Rep",
  ["scientific reports of the istituto superiore di sanita"] = "Sci Rep Ist Super Sanita",
  ["scientific reports of the istituto superiore di sanità"] = "Sci Rep Ist Super Sanita",
  ["scientific review"] = "Guigoz Sci Rev",
  ["scientific studies and research. series mathematics and informatics"] = "Sci. Stud. Res. Ser. Math. Inform.",
  ["scientific studies of reading : the official journal of the society for the scientific study of reading"] = "Sci Stud Read",
  ["scientific study of literature"] = "Sci Study Lit",
  ["scientific temperance journal"] = "Sci Temper J",
  ["scientific times journal of diabetes"] = "Sci Times J Diabetes",
  ["scientific transactions in environment and technovation"] = "Sci Trans Environ Technovation",
  ["scientific world journal"] = "Sci. World J.",
  ["scientific writings from the ancient and medieval world"] = "Sci. Writ. Anc. Mediev. World",
  ["scientist (philadelphia, pa.)"] = "Scientist",
  ["scientist and citizen"] = "Sci Citizen",
  ["scientists of wales"] = "Sci. Wales",
  ["scienze dell’antichità. storia, archeologia, antropologia"] = "ScAnt",
  ["scienze matematiche"] = "Sci. Mat.",
  ["scienze, le"] = "Scienze",
  ["scifed journal of cardiology"] = "Scifed J Cardiol",
  ["scifed virology research journal"] = "SciFed Virol Res J",
  ["scimedicine journal"] = "SciMed J",
  ["scipost physics"] = "Scipost Phys.",
  ["scipost physics proceedings"] = "Scipost Phys. Proc.",
  ["scm noticies"] = "SCM Not.",
  ["scn news"] = "SCN News",
  ["scn news / united nations, administrative committee on coordination, subcommittee on nutrition"] = "SCN News",
  ["scna newsletter"] = "SCNA Newsl",
  ["scoliosis and spinal disorders"] = "Scoliosis Spinal Disord",
  ["scottish affairs"] = "Scott Aff",
  ["scottish archaeological journal"] = "Scott. Archaeol. J.",
  ["scottish economic & social history"] = "Scott Econ Soc Hist",
  ["scottish educational review"] = "Scott Edu Rev",
  ["scottish forestry"] = "Scott. For.",
  ["scottish geographical journal"] = "Scott. Geogr. J.",
  ["scottish geographical magazine"] = "Scott Geogr Mag",
  ["scottish graduate series"] = "Scott. Grad. Ser.",
  ["scottish journal of geology"] = "Scott. J. Geol.",
  ["scottish journal of political economy"] = "Scott J Polit Econ",
  ["scottish medical journal"] = "Scott Med J",
  ["screening : journal of the international society of neonatal screening"] = "Screening",
  ["scripta classica israelica. yearbook of the israel society for the promotion of classical studies"] = "ScrClIsr",
  ["scripta classica israelica: yearbook of the israel soc. for the promotion of classical studies"] = "SCI",
  ["scripta ethnologica (centro argentino de etnologia americana)"] = "Scr Ethnol",
  ["scripta ethnologica (centro argentino de etnología americana)"] = "Scr Ethnol",
  ["scripta geobotanica"] = "Scr. Geobot.",
  ["scripta geologica"] = "Scr Geol",
  ["scripta hierosolymitana. publications of the hebrew university, jerusalem"] = "ScrHieros",
  ["scripta instituti donneriani aboensis"] = "Scr Instituti Donneriani Abo",
  ["scripta materialia"] = "Scr. Mater.",
  ["scripta medica"] = "Scr Med (Brno)",
  ["scripta mediterranea. bulletin of the society for mediterranean studies, toronto"] = "ScrMed",
  ["scripta metallugica et materialia"] = "Scr. Metall. Mater.",
  ["scripta metallurgica"] = "Scr. Metall.",
  ["scripta metallurgica et materialia"] = "Scr. Metall. Mater.",
  ["scripta minora regiae societatis humaniorum litterarum lundensis = studier"] = "Scr Minora",
  ["scriptorium: revue internationale des études relatives aux manuscrits"] = "Scriptorium",
  ["scrittura e civiltà"] = "S&C",
  ["sda dentist"] = "SDA Dent",
  ["sdc magazine"] = "Sdc Mag",
  ["sdrp journal of biomedical engineering"] = "SDRP J Biomed Eng",
  ["sdrp journal of plant science"] = "SDRP J Plant Sci",
  ["se pu = chinese journal of chromatography"] = "Se Pu",
  ["se pu chinese journal of chromatography (ta-lien shih"] = "Se Pu",
  ["sea view hospital bulletin"] = "Sea View Hosp Bull",
  ["seabird (thetford, england : 2008)"] = "Seabird",
  ["seagull theatre quarterly"] = "Seagull Theatre Q",
  ["seagull theatre quarterly : stq"] = "Seagull Theatre Q",
  ["sealing technology"] = "Sealing Technol.",
  ["seaprap research report : a report of research undertaken with the assistance of an award from the southeast asia population research awards program (seaprap), institute of southeast asian studies"] = "Seaprap Res Rep",
  ["seara medica"] = "Seara Med",
  ["seara médica"] = "Seara Med",
  ["searcher (medford, n.j.)"] = "Search Medford N J",
  ["seattle university law review"] = "Seattle Univ Law Rev",
  ["seattle university law review / seattle university"] = "Seattle Univ Law Rev",
  ["seb experimental biology series"] = "SEB Exp Biol Ser",
  ["secolas annals : journal of the southeastern council on latin american studies. southeastern council on latin american studies"] = "SECOLAS Ann",
  ["second language research"] = "Second Lang Res",
  ["second messengers and phosphoproteins"] = "Second Messengers Phosphoproteins",
  ["second opinion"] = "Second Opin.",
  ["second opinion (chicago, ill. : 1999)"] = "Second Opin (Chic)",
  ["second opinion (park ridge, ill.)"] = "Second Opin",
  ["second opinions of health care issues"] = "Second Opin. Health Care Issues",
  ["second opinions on health care issues"] = "Second Opin Health Care Issues",
  ["secuencia (mexico city, mexico)"] = "Secuencia",
  ["security and communication networks"] = "Secur. Commun. Netw.",
  ["security and safety"] = "Secur. Saf.",
  ["security dialogue"] = "Secur Dialogue",
  ["security informatics"] = "Secur Inform",
  ["sedimentary geology"] = "Sediment. Geol.",
  ["seed science research"] = "Seed Sci. Res.",
  ["seed treatment: progress and prospects"] = "Br. Crop. Pr.",
  ["seeds (new york, n.y.)"] = "Seeds",
  ["seeing and perceiving"] = "Seeing Perceiving",
  ["sefer asya"] = "Sefer Asya",
  ["seg discovery"] = "SEG Discovery",
  ["segno e testo"] = "S&T",
  ["sei roka kango daigaku kiyo"] = "Seiroka Kango Daigaku Kiyo",
  ["sei roka kango daigaku kiyō"] = "Seiroka Kango Daigaku Kiyo",
  ["seibutsu kankyo chosetsu. [environment control in biology"] = "Seibutsu Kankyo Chosetsu",
  ["seibutsu kogaku kaishi"] = "Seibutsu Kogaku Kaishi",
  ["seibutsu seisangaku kenkyū : hiroshima daigaku seibutsu seisan gakubu kiyō"] = "Seibutsu Seisangaku Kenkyu",
  ["seibutsu shiryo bunseki"] = "Journal of analytical bio-science|Seibutsu Shiryo Bunseki",
  ["seibutsu shiryō bunseki = journal of analytical bio-science"] = "Seibutsu Shiryo Bunseki",
  ["seikagaku (journal of japanese biochemical society)"] = "Seikagaku",
  ["seikagaku. journal of japanese biochemical society"] = "Seikagaku",
  ["seikagaku. the journal of japanese biochemical society"] = "Seikagaku",
  ["seikatsu eisei"] = "Seikatsu Eisei",
  ["seikatsu eisei. [life sanitation"] = "Seikatsu Eisei",
  ["seikei geka (orthopedic surgery)"] = "Seikei Geka",
  ["seikeigeka. orthopedic surgery"] = "Seikei Geka",
  ["seimei kagaku kenkyujo kiyō"] = "Seimei Kagaku Kenkyujo Kiyo",
  ["seiri seitai. [physiology and ecology]"] = "Seiri Seital",
  ["seiroka kango daigaku kiyo"] = "Seiroka Kango Daigaku Kiyo",
  ["seishin igaku kenkyusho gyosekishu (bulletin of the seishin-igaku institute)"] = "Seishin Igaku Kenkyusho Gyosekishu",
  ["seishin igaku kenkyūjo gyōsekishū = bulletin of the seishin-igaku institute"] = "Seishin Igaku Kenkyusho Gyosekishu",
  ["seishin igaku. clinical psychiatry"] = "Seishin Igaku",
  ["seishin shinkeigaku zasshi (psychiatria et neurologia japonica)"] = "Seishin Shinkeigaku Zasshi",
  ["seishin shinkeigaku zasshi = psychiatria et neurologia japonica"] = "Seishin Shinkeigaku Zasshi",
  ["seishin shinkeigaku zasshi. psychiatria et neurologia japonica"] = "Seishin Shinkeigaku Zasshi",
  ["seismic instruments"] = "Seismic Instrum.",
  ["seismological research letters"] = "Seismol. Res. Lett.",
  ["seitai no kagaku"] = "Seitai no Kagaku",
  ["seizure : the journal of the british epilepsy association"] = "Seizure",
  ["sej surgery and pain"] = "SEJ Surg Pain",
  ["selecciones matematicas"] = "Sel. Mat.",
  ["selecoes odontologicas"] = "Sel. Odontol. (Sao Paulo)",
  ["selecta mathematica"] = "Selecta Math. (N.S.)",
  ["selecta mathematica new series"] = "Sel. Math. New Ser.",
  ["selecta mathematica. new series"] = "Selecta Math. (N.S.)",
  ["selected annual reviews of the analytical sciences"] = "Sel. Annu. Rev. Anal. Sci.",
  ["selected chapters of number theory: special numbers"] = "Sel. Chapters Number Theory: Spec. Numbers",
  ["selected lectures in mathematics"] = "Sel. Lectures Math.",
  ["selected papers. consortium on revolutionary europe, 1750-1850"] = "Sel Pap Consort Revolut Eur 1759 1850",
  ["selected works in probability and statistics"] = "Sel. Works Probab. Stat.",
  ["selected works of benoit b"] = "Sel. Works B. B. Mandelbrot",
  ["selective cancer therapeutics"] = "Sel Cancer Ther",
  ["seleções odontológicas"] = "Sel Odontol (Sao Paulo)",
  ["self and identity : the journal of the international society for self and identity"] = "Self Identity",
  ["self-care, dependent-care & nursing : the official journal of the international orem society"] = "Self Care Depend Care Nurs",
  ["self-surrender (prapatti) to god in srivaisnavism: tamil cats and sanskrit monkeys"] = "Routl. Hindu. Stud. Ser.",
  ["self-teaching guides on mathematics"] = "Self-Teach. Guides Math.",
  ["selʹskokhozi︠a︡ĭstvennai︠a︡ biologii︠a︡"] = "Selskokhoziaĭstvennaia Biol",
  ["sema journal"] = "SeMA J.",
  ["sema journal. boletin de la sociedad espannola de matematica aplicada"] = "SeMA J.",
  ["sema journal. boletin de la sociedad espanola de matematica aplicada"] = "SeMA J.",
  ["sema simai springer series"] = "SEMA SIMAI Springer Ser.",
  ["semaine des hopitaux"] = "Sem. Hop.",
  ["semaine des hopitaux therapeutique"] = "Sem. Hop. Ther.",
  ["semaine des hopitaux. therapeutique"] = "Sem Hop Ther",
  ["semaine des hopitaux: informations"] = "Sem Hop Inf",
  ["semaine des hôpitaux. therapeutique"] = "Sem Hop Ther",
  ["semaine des hôpitaux: informations"] = "Sem Hop Inf",
  ["semaine medicale professionelle et medico-sociale"] = "Sem Med Prof Med Soc",
  ["semaine médicale professionelle et médico-sociale"] = "Sem Med Prof Med Soc",
  ["semaine therapeutique"] = "Sem. Ther.",
  ["semaine thérapeutique"] = "Sem Ther",
  ["semana epidemiologica"] = "Sem Epidemiol",
  ["semana medica de mexico"] = "Sem Med Mex",
  ["semana médica de méxico"] = "Sem Med Mex",
  ["semana médica española"] = "Sem Med Esp",
  ["semantic web"] = "Semant Web",
  ["semantics and pragmatics"] = "Semant Pragmat",
  ["semiconductor physics, quantum electronics and optoelectronics"] = "Semicond. Phys. Quantum Electron. Optoelectron.",
  ["semiconductor science and technology"] = "Semicond. Sci. Technol.",
  ["semiconductor today"] = "Semicond. Today",
  ["semigroup forum"] = "Semigroup Forum",
  ["semina. ciências agrárias"] = "Semin Cienc Agrar",
  ["semina: ciencias agrarias"] = "Semina Cienc. Agrar.",
  ["seminaire de mathematique (nouvelle serie)"] = "Sem. Math. (N. S.)",
  ["seminaire de probabilites"] = "Semin. Probab.",
  ["seminaire lotharingien de combinatoire"] = "Sem. Lothar. Combin.",
  ["seminaire: equations aux derivees partielles"] = "Semin. Equ. Deriv. Partielles",
  ["seminaires et congres"] = "Semin. Congr.",
  ["seminar international"] = "Semin Int",
  ["seminar of ilia vekua institute of applied mathematics. reports"] = "Semin. I. Vekua Inst. Appl. Math. Rep.",
  ["seminar on mathematical sciences"] = "Sem. Math. Sci.",
  ["seminar report. merck sharp & dohme"] = "Semin Rep Merck Sharp Dohme",
  ["seminar series in mathematics"] = "Sem. Ser. Math. Algebra",
  ["seminari romani di cultura greca"] = "SemRom",
  ["seminario de ensenanza de la odontopediatria"] = "Semin Ensen Odontopediatr",
  ["seminars for nurse managers"] = "Semin Nurse Manag",
  ["seminars in adolescent medicine"] = "Semin Adolesc Med",
  ["seminars in anesthesia"] = "Semin Anesth",
  ["seminars in arthritis and rheumatism"] = "Semin Arthritis Rheum",
  ["seminars in arthroplasty"] = "Semin Arthroplasty",
  ["seminars in breast disease"] = "Semin Breast Dis",
  ["seminars in cancer biology"] = "Semin Cancer Biol",
  ["seminars in cardiothoracic and vascular anesthesia"] = "Semin Cardiothorac Vasc Anesth",
  ["seminars in cell & developmental biology"] = "Semin Cell Dev Biol",
  ["seminars in cell and developmental biology"] = "Semin. Cell Dev. Biol.",
  ["seminars in cell biology"] = "Semin Cell Biol",
  ["seminars in clinical neuropsychiatry"] = "Semin Clin Neuropsychiatry",
  ["seminars in colon & rectal surgery"] = "Semin Colon Rectal Surg",
  ["seminars in cutaneous medicine and surgery"] = "Semin Cutan Med Surg",
  ["seminars in dental hygiene"] = "Semin Dent Hyg",
  ["seminars in dermatology"] = "Semin Dermatol",
  ["seminars in developmental biology"] = "Semin Dev Biol",
  ["seminars in diagnostic pathology"] = "Semin Diagn Pathol",
  ["seminars in dialysis"] = "Semin Dial",
  ["seminars in drug treatment"] = "Semin Drug Treat",
  ["seminars in family medicine"] = "Semin Fam Med",
  ["seminars in fetal & neonatal medicine"] = "Semin Fetal Neonatal Med",
  ["seminars in fetal and neonatal medicine"] = "Semin. Fetal. Neonatal Med.",
  ["seminars in gastrointestinal disease"] = "Semin Gastrointest Dis",
  ["seminars in hearing"] = "Semin Hear",
  ["seminars in hematology"] = "Semin Hematol",
  ["seminars in immunology"] = "Semin Immunol",
  ["seminars in immunopathology"] = "Semin Immunopathol",
  ["seminars in interventional cardiology"] = "Semin. Interv. Cardiol.",
  ["seminars in interventional cardiology : siic"] = "Semin Interv Cardiol",
  ["seminars in interventional radiology"] = "Semin Intervent Radiol",
  ["seminars in laparoscopic surgery"] = "Semin Laparosc Surg",
  ["seminars in liver disease"] = "Semin Liver Dis",
  ["seminars in medical practice"] = "Semin Med Pract",
  ["seminars in musculoskeletal radiology"] = "Semin Musculoskelet Radiol",
  ["seminars in neonatology"] = "Semin. Neonatol.",
  ["seminars in neonatology : sn"] = "Semin Neonatol",
  ["seminars in nephrology"] = "Semin Nephrol",
  ["seminars in neurology"] = "Semin Neurol",
  ["seminars in nuclear medicine"] = "Semin Nucl Med",
  ["seminars in oncology"] = "Semin Oncol",
  ["seminars in oncology nursing"] = "Semin Oncol Nurs",
  ["seminars in ophthalmology"] = "Semin Ophthalmol",
  ["seminars in orthodontics"] = "Semin Orthod",
  ["seminars in pediatric infectious diseases"] = "Semin Pediatr Infect Dis",
  ["seminars in pediatric neurology"] = "Semin Pediatr Neurol",
  ["seminars in pediatric surgery"] = "Semin Pediatr Surg",
  ["seminars in perinatology"] = "Semin Perinatol",
  ["seminars in perioperative nursing"] = "Semin Perioper Nurs",
  ["seminars in plastic surgery"] = "Semin Plast Surg",
  ["seminars in psychiatry"] = "Semin Psychiatry",
  ["seminars in radiation oncology"] = "Semin Radiat Oncol",
  ["seminars in reproductive endocrinology"] = "Semin Reprod Endocrinol",
  ["seminars in reproductive medicine"] = "Semin Reprod Med",
  ["seminars in respiratory and critical care medicine"] = "Semin Respir Crit Care Med",
  ["seminars in respiratory infections"] = "Semin Respir Infect",
  ["seminars in roentgenology"] = "Semin Roentgenol",
  ["seminars in speech and language"] = "Semin Speech Lang",
  ["seminars in spine surgery"] = "Semin Spine Surg",
  ["seminars in surgical oncology"] = "Semin Surg Oncol",
  ["seminars in surgical oncology (new york, ny"] = "Semin Surg Oncol",
  ["seminars in thoracic and cardiovascular surgery"] = "Semin Thorac Cardiovasc Surg",
  ["seminars in thoracic and cardiovascular surgery. pediatric cardiac surgery annual"] = "Semin Thorac Cardiovasc Surg Pediatr Card Surg Annu",
  ["seminars in thrombosis and hemostasis"] = "Semin Thromb Hemost",
  ["seminars in ultrasound, ct and mr"] = "Semin. Ultrasound. CT MR",
  ["seminars in ultrasound, ct and mri"] = "Semin. Ultrasound CT MRI",
  ["seminars in ultrasound, ct, and mr"] = "Semin Ultrasound CT MR",
  ["seminars in urologic oncology"] = "Semin Urol Oncol",
  ["seminars in urology"] = "Semin Urol",
  ["seminars in vascular medicine"] = "Semin Vasc Med",
  ["seminars in vascular surgery"] = "Semin Vasc Surg",
  ["seminars in veterinary medicine and surgery (small animal)"] = "Semin Vet Med Surg Small Anim",
  ["seminars in virology"] = "Semin Virol",
  ["semitica. cahiers publiés par l’institut d’études sémitiques du college de france. avec le concours du centre national de la recherche scientifique"] = "Semitica",
  ["semstat elements"] = "SemStat Elem.",
  ["senckenbergiana biologica"] = "Senckenb Biol",
  ["senckenbergiana biologie"] = "Senckenbergiana Biol.",
  ["senckenbergiana lethaea"] = "Senckenbergiana Lethaea",
  ["senior nurse"] = "Sr Nurse",
  ["senior scholastic"] = "Sr Scholast",
  ["senior scholastic (teacher[s] ed.)"] = "Sr Scholast (Teach Ed)",
  ["seniors housing & care journal"] = "Seniors Hous Care J",
  ["sensecam 2013 : proceedings of the 4th sensecam conference : sensecam and pervasive imaging 2013 : san diego, usa, november 18-19, 2013. sensecam (conference) (4th : 2013 : san diego, calif.)"] = "SenseCam 2013",
  ["sensing and bio-sensing research"] = "Sens. Bio-Sens. Res.",
  ["sensing and imaging"] = "Sens Imaging",
  ["sensing and instrumentation for food quality and safety"] = "Sens. Instrum. Food Qual. Saf.",
  ["sensor letters"] = "Sens Lett",
  ["sensor review"] = "Sens. Rev.",
  ["sensors (basel, switzerland)"] = "Sensors (Basel)",
  ["sensors (switzerland)"] = "Sens. Switz.",
  ["sensors and actuators (warrendale, pa.)"] = "Sens Actuators (Warrendale Pa)",
  ["sensors and actuators a: physical"] = "Sens. Actuators, A",
  ["sensors and actuators b: chemical"] = "Sens. Actuators, B",
  ["sensors and actuators reports"] = "Sens. Actuators Rep.",
  ["sensors and actuators, a: physical"] = "Sens. Actuators, A",
  ["sensors and actuators, b: chemical"] = "Sens. Actuators, B",
  ["sensors and actuators, b: chemical sensors and materials"] = "Sens. Actuators, B",
  ["sensors and actuators. a, physical"] = "Sens Actuators A Phys",
  ["sensors and actuators. b, chemical"] = "Sens Actuators B Chem",
  ["sensors and materials"] = "Sens. Mater.",
  ["sensors international"] = "Sens. Int.",
  ["sensors update"] = "Sensors Update",
  ["sensory neuron"] = "Sens Neur",
  ["sensory processes"] = "Sens Processes",
  ["sensory systems"] = "Sens Syst",
  ["sentinel event alert"] = "Sentinel Event Alert",
  ["sentinel event alert / joint commission on accreditation of healthcare organizations"] = "Sentinel Event Alert",
  ["sen’i gakkaishi"] = "Sen’i Gakkaishi",
  ["seohae-an ji-yeog baljeon yeon-gu = the west coast area development institute"] = "Seohaean Jiyeog Baljeon Yeongu",
  ["seoul journal of economics"] = "Seoul J. Econ.",
  ["separation and purification methods"] = "Sep. Purif. Methods",
  ["separation and purification reviews"] = "Sep. Purif. Rev.",
  ["separation and purification technology"] = "Sep. Purif. Technol.",
  ["separation science and technology"] = "Sep. Sci. Technol.",
  ["separation science and technology (philadelphia, pa, united states)"] = "Sep. Sci. Technol. (Philadelphia, PA, U. S.)",
  ["separation science plus"] = "Sep Sci Plus",
  ["separations technology"] = "Sep. Technol.",
  ["sequential analysis"] = "Sequential Anal.",
  ["sequential analysis. design methods & applications"] = "Sequential Anal.",
  ["ser (revista médico-social)"] = "Ser Rev Med Soc",
  ["serbian astronomical journal"] = "Serb. Astron. J.",
  ["serdica journal of computing"] = "Serdica J. Comput.",
  ["serdica. mathematical journal. serdika. matematichesko spisanie"] = "Serdica Math. J.",
  ["seria filologia rosyjska"] = "Ser Filol Ros",
  ["seria filologia rosyjska / uniwersytet im. adama michiewicza w poznaniu, wydzial filologiczny"] = "Ser Filol Ros",
  ["seria fizyka (uniwersytet im. adama mickiewicza w poznaniu)"] = "Ser. Fiz. (Uniw. im. Adama Mickiewicza Poznaniu)",
  ["seria: konferencje"] = "Ser. Konf. (Wroc.)",
  ["serie cientifica"] = "Ser. Cient.",
  ["serie di matematica e fisica"] = "Ser. Mat. Fis.",
  ["serie sciences de l’homme"] = "Cah Am Lat Ser Sci Homme",
  ["serie scientifica"] = "Ser. Sci.",
  ["serie systemes automatises"] = "Ser. Syst. Autom.",
  ["serie: ecole nationale de la statistique et de l’administration economique et du centre d’etudes des programmes economiques"] = "Serie: Ecole Nationale de la Statistique et de l’Administration economique et du Centre d’Etudes des Programmes",
  ["series : journal of the spanish economic association"] = "SERIEs (Berl)",
  ["series haematologica"] = "Ser. Haematol.",
  ["series haematologica (1968)"] = "Ser Haematol",
  ["series in algebra"] = "Ser. Algebra",
  ["series in algebraic and differential geometry"] = "Ser. Algebr. Differ. Geom.",
  ["series in analysis"] = "Ser. Anal.",
  ["series in applied and computational mathematics"] = "Ser. Appl. Comput. Math.",
  ["series in approximations and decompositions"] = "Ser. Approx. Decompos.",
  ["series in bioengineering"] = "Ser. BioEng.",
  ["series in computational and physical processes in mechanics and thermal sciences"] = "Ser. Comput. Phys. Process. Mech. Thermal Sci.",
  ["series in computer vision"] = "Ser. Comput. Vis.",
  ["series in contemporary applied mathematics cam"] = "Ser. Contemp. Appl. Math. CAM",
  ["series in contemporary mathematics"] = "Ser. Contemp. Math.",
  ["series in display science and technology"] = "Ser. Disp. Sci. Tech.",
  ["series in electrical and computer engineering"] = "Ser. Electr. Comput. Eng.",
  ["series in intelligent control and intelligent automation"] = "Ser. Intell. Control Intell. Autom.",
  ["series in international business and economics"] = "Ser. Internat. Bus. Econom.",
  ["series in machine perception and artificial intelligence"] = "Ser. Mach. Percept. Artif. Intell.",
  ["series in mathematical analysis and applications"] = "Ser. Math. Anal. Appl.",
  ["series in mathematical biology and medicine"] = "Ser. Math. Biol. Med.",
  ["series in modern condensed matter physics"] = "Ser. Modern Condensed Matter Phys.",
  ["series in number theory"] = "Ser. Number Theory",
  ["series in optics and optoelectronics"] = "Ser. Opt. Optoelectron.",
  ["series in pure mathematics"] = "Ser. Pure Math.",
  ["series in quantitative finance"] = "Ser. Quant. Financ.",
  ["series in real analysis"] = "Ser. Real Anal.",
  ["series in soft condensed matter"] = "Ser. Soft Condens. Matter",
  ["series on advances in mathematics for applied sciences"] = "Ser. Adv. Math. Appl. Sci.",
  ["series on advances in statistical mechanics"] = "Ser. Adv. Statist. Mech.",
  ["series on analysis, applications and computation"] = "Ser. Anal. Appl. Comput.",
  ["series on applied mathematics"] = "Ser. Appl. Math.",
  ["series on biomechanics"] = "Ser Biomech",
  ["series on coding theory and cryptology"] = "Ser. Coding Theory Cryptol.",
  ["series on complexity science"] = "Ser. Complex. Sci.",
  ["series on complexity, nonlinearity and chaos"] = "Ser. Complex. Nonlinearity Chaos",
  ["series on computers and operations research"] = "Ser. Comput. Oper. Res.",
  ["series on concrete and applicable mathematics"] = "Ser. Concr. Appl. Math.",
  ["series on knots and everything"] = "Ser. Knots Everything",
  ["series on mathematics education"] = "Ser. Math. Educ.",
  ["series on multivariate analysis"] = "Ser. Multivariate Anal.",
  ["series on new frontiers in advanced mathematics of the istituto per la ricerca di base"] = "Ser. New Front. Adv. Math. Ist. Ric. Base",
  ["series on number theory and its applications"] = "Ser. Number Theory Appl.",
  ["series on nursing administration"] = "Ser Nurs Adm",
  ["series on optimization"] = "Ser. Optim.",
  ["series on optimization and its applications"] = "Ser. Optim. Appl.",
  ["series on soviet and east european mathematics"] = "Ser. Soviet East European Math.",
  ["series on stability, vibration and control of structures"] = "Ser. Stab. Vib. Control Struct.",
  ["series on stability, vibration and control of systems"] = "Ser. Stab. Vib. Control Syst. Ser. B",
  ["series on stability, vibration and control of systems. series a: textbooks, monographs and treatises"] = "Ser. Stab. Vib. Control Syst. Ser. A",
  ["series on the foundations of natural science and technology"] = "Ser. Found. Nat. Sci. Tech.",
  ["series on the pure and applied mathematics"] = "Ser. Pure Appl. Math.",
  ["series on theoretical computer science"] = "Ser. Theor. Comput. Sci.",
  ["series on university mathematics"] = "Ser. Univ. Math.",
  ["series on work and organization"] = "Ser. Work Organ.",
  ["series paedopsychiatrica"] = "Ser Paedopsychiatr",
  ["serket = sarkat"] = "Serket",
  ["serono symposia publications from raven press"] = "Serono Symp Publ Raven Press",
  ["service de la recherche terres et forêts, québec"] = "Serv. rech. terres for., Qué.",
  ["service oriented computing and applications"] = "Serv. Oriented Comput. Appl.",
  ["service science"] = "Serv Sci",
  ["service social"] = "Serv Soc Que",
  ["servir (lisbon, portugal)"] = "Servir",
  ["servizio migranti"] = "Serv Migr",
  ["session laws of hawaii passed by the state legislature. hawaii"] = "Sess Laws Hawaii Passed State Legis Hawaii",
  ["session laws of the state of minnesota. minnesota"] = "Sess Laws State Minn Minn",
  ["session laws. guam"] = "Sess Laws Guam",
  ["set-valued analysis"] = "Set-Valued Anal.",
  ["set-valued and variational analysis"] = "Set-Valued Var. Anal.",
  ["set-valued and variational analysis. theory and applications"] = "Set-Valued Var. Anal.",
  ["seton hall law review"] = "Seton Hall Law Rev",
  ["seton hall legislative journal"] = "Seton Hall Legis J",
  ["settimana medica"] = "Settim Med",
  ["setúbal arqueológica"] = "SetubalA",
  ["seven days"] = "Seven Days",
  ["seventeenth-century french studies"] = "Seventeenth Century Fr Stud",
  ["seventh conference on health survey research methods. conference on health survey research methods (7th : 1999 : williamsburg, va.)"] = "Seventh Conf Health Survey Res Methods (1999)",
  ["sewage and industrial wastes"] = "Sewage Ind Waste",
  ["sewage works journal"] = "Sewage Work J",
  ["sex education"] = "Sex Educ",
  ["sex roles"] = "Sex Roles",
  ["sex weekly plus"] = "Sex Wkly Plus",
  ["sexual & reproductive healthcare : official journal of the swedish association of midwives"] = "Sex Reprod Healthc",
  ["sexual abuse"] = "Sex. Abuse",
  ["sexual abuse : a journal of research and treatment"] = "Sex Abuse",
  ["sexual addiction & compulsivity"] = "Sex Addict Compulsivity",
  ["sexual and marital therapy : journal of the association of sexual and marital therapists"] = "Sex Marital Ther",
  ["sexual and relationship therapy : journal of the british association for sexual and relationship therapy"] = "Sex Relation Ther",
  ["sexual and reproductive health matters"] = "Sex Reprod Health Matters",
  ["sexual development : genetics, molecular biology, evolution, endocrinology, embryology, and pathology of sex determination and differentiation"] = "Sex Dev",
  ["sexual health"] = "Sex Health",
  ["sexual health exchange"] = "Sex Health Exch",
  ["sexual medicine"] = "Sex Med",
  ["sexual medicine reviews"] = "Sex Med Rev",
  ["sexual medicine today"] = "Sex Med Today",
  ["sexual plant reproduction"] = "Sex Plant Reprod",
  ["sexualidad, salud y sociedad : revista latinoamericana"] = "Sex Salud Soc (Rio J)",
  ["sexualidade e planeamento familiar"] = "Sex Planeam Fam",
  ["sexuality & culture"] = "Sex Cult",
  ["sexuality and disability"] = "Sex Disabil",
  ["sexuality research & social policy : journal of nsrc : sr & sp"] = "Sex Res Social Policy",
  ["sexuality, gender & policy"] = "Sex Gend Policy",
  ["sexuality, reproduction & menopause"] = "Sex Reprod Menopause",
  ["sexually transmitted diseases"] = "Sex Transm Dis",
  ["sexually transmitted infections"] = "Sex Transm Infect",
  ["sexualpadagogik und familienplanung"] = "Sex Padag Fam Plan",
  ["sécheresse. agence universitaire de la francophonie"] = "Secheresse",
  ["série sciences de l'homme"] = "Cah Am Lat Ser Sci Homme",
  ["sf journal of biotechnology and biomedical engineering"] = "SF J Biotechnol Biomed Eng",
  ["sf journal of medical oncology and cancer"] = "SF J Med Oncol Cancer",
  ["sh'ma : a journal of jewish responsibility"] = "Shma",
  ["shaanxi keji daxue xuebao"] = "Shaanxi Keji Daxue Xuebao",
  ["shaanxi normal university"] = "J. Shaanxi Normal Univ. Nat. Sci. Ed.",
  ["shaanxi normal university. journal. natural science edition. shaanxi shifan daxue xuebao. ziran kexue ban"] = "J. Shaanxi Normal Univ. Nat. Sci. Ed.",
  ["shakai keizai shigaku"] = "Shakai Keizai Shigaku",
  ["shakai keizai shigaku / shakai keizai shigakkai"] = "Shakai Keizai Shigaku",
  ["shakespeare studies"] = "Shakespeare Stud",
  ["shan di nong ye sheng wu xue bao = journal of mountain agriculture and biology"] = "Shan Di Nong Ye Sheng Wu Xue Bao",
  ["shandong da xue xue bao. li xue ban"] = "Journal of Shandong University. Natural science|Shandong Da Xue Xue Bao",
  ["shandong da xue xue bao. li xue ban /natural science"] = "Shandong Da Xue Xue Bao Li Xue Ban",
  ["shandong da xue xue bao. yi xue ban"] = "Shandong Da Xue Xue Bao",
  ["shandong daxue xuebao"] = "Shandong Daxue Xuebao Ziran Kexue Ban",
  ["shandong kuangye xueyuan xuebao"] = "Shandong Kuangye Xueyuan Xuebao",
  ["shandong li gong da xue xue bao. she hui ke xue ban"] = "Shandong Li Gong Da Xue Xue Bao She Hui Ke Xue Ban",
  ["shandong nongye kexue"] = "Shandong Nongye Kexue",
  ["shang hai yi xue = injury medicine"] = "Inj Med",
  ["shanghai archives of psychiatry"] = "Shanghai Arch Psychiatry",
  ["shanghai chest"] = "Shanghai Chest",
  ["shanghai hai yang da xue xue bao"] = "Shanghai Hai Yang Da Xue Xue Bao",
  ["shanghai huanjing kexue"] = "Shanghai Huanjing Kexue",
  ["shanghai institute of railway technology"] = "J. Shanghai Inst. Railway Tech.",
  ["shanghai jiao tong da xue xue bao. yi xue ban"] = "Shanghai Jiao Tong Da Xue Xue Bao Yi Xue Ban",
  ["shanghai jiaotong daxue xuebao. nongye kexueban"] = "Shanghai Jiaotong Daxue Xuebao. Nongye Kexueban",
  ["shanghai kou qiang yi xue = shanghai journal of stomatology"] = "Shanghai Kou Qiang Yi Xue",
  ["shanghai mian yi xue za zhi"] = "Shanghai Mien I Hsueh Tsa Chih",
  ["shanghai nong xue yuan xue bao = journal of shanghai agricultural college = shanghai nongxueyuan xuebao"] = "Shanghai Nong Xue Yuan Xue Bao",
  ["shanghai nong ye xue bao = acta agriculturae shanghai"] = "Shanghai Nong Ye Xue Bao",
  ["shanghai shifan daxue xuebao"] = "Shanghai Shifan Daxue Xuebao Ziran Kexue Ban",
  ["shanghai yixue jianyan zazhi"] = "Shanghai journal of medical laboratory sciences|Shanghai Yixue Jianyan Zazhi",
  ["shanghai yixue jianyan zazhi = shanghai journal of medical laboratory sciences"] = "Shanghai Yixue Jianyan Zazhi",
  ["shanxi da xue xue bao. zi ran ke xue ban = shanxi university journal"] = "Shanxi Da Xue Xue Bao Zi Ran Ke Xue Ban",
  ["shanxi nong ye da xue xue bao. zi ran ke xue ban"] = "Shanxi Nong Ye Da Xue Xue Bao Zi Ran Ke Xue Ban",
  ["shanxi shi da xue bao. zi ran ke xue ban = journal of shaanxi normal university"] = "Shanxi Shi Da Xue Bao Zi Ran Ke Xue Ban",
  ["shape in medical imaging : international workshop, shapemi 2018, held in conjunction with miccai 2018, granada, spain, september 20, 2018 : proceedings. shapemi (workshop) (2018 : granada, spain)"] = "Shape Med Imaging (2018)",
  ["shape in medical imaging : international workshop, shapemi 2020, held in conjunction with miccai 2020, lima, peru, october 4, 2020, proceedings"] = "Shape Med Imaging (2020)",
  ["shape memory and superelasticity"] = "Shape Mem. Superelasticity",
  ["shb'12 : proceedings of the 2012 acm international workshop on smart health and wellbeing : october 29, 2012, maui, hawaii, usa. international workshop on smart health and wellbeing (2012 : maui, hawaii)"] = "SHB12 (2012)",
  ["she ji: the journal of design, economics, and innovation"] = "She Ji: J. Des. Econ. Innovation",
  ["she zhi"] = "Journal of snake|She Zhi",
  ["she zhi = journal of snake"] = "She Zhi",
  ["shendetesia popullore"] = "Shendet Pop",
  ["sheng li hsueh pao [acta physiologica sinica]"] = "Sheng Li Hsueh Pao",
  ["sheng li ke xue jin zhan (progress in physiology)"] = "Sheng Li Ke Xue Jin Zhan",
  ["sheng li ke xue jin zhan [progress in physiology]"] = "Sheng Li Ke Xue Jin Zhan",
  ["sheng li ko hsueh chin chan [progress in physiological sciences]"] = "Sheng Li Ko Hsueh Chin Chan",
  ["sheng li xue bao (acta physiologica sinica)"] = "Sheng Li Xue Bao",
  ["sheng li xue bao : [acta physiologica sinica]"] = "Sheng Li Xue Bao",
  ["sheng ming ke xue yan jiu"] = "Sheng Ming Ke Xue Yan Jiu",
  ["sheng ming ke xue yan jiu / hunan shi fan da xue zhu ban"] = "Life science research / Hunan Normal University|Sheng Ming Ke Xue Yan Jiu",
  ["sheng tai huan jing xue bao = ecology and environmental sciences"] = "Sheng Tai Huan Jing Xue Bao",
  ["sheng tai xue bao = acta ecologica sinica"] = "Sheng Tai Xue Bao",
  ["sheng wu duo yang xing"] = "Chinese biodiversity|Sheng Wu Duo Yang Xing",
  ["sheng wu duo yang xing = chinese biodiversity"] = "Sheng Wu Duo Yang Xing",
  ["sheng wu gong cheng xue bao = chinese journal of biotechnology"] = "Sheng Wu Gong Cheng Xue Bao",
  ["sheng wu hua xue yu sheng wu wu li jin zhan"] = "Sheng Wu Hua Xue Yu Sheng Wu Wu Li Jin Zhan",
  ["sheng wu hua xue yu sheng wu wu li xue bao acta biochimica et biophysica sinica"] = "Sheng Wu Hua Xue Yu Sheng Wu Wu Li Xue Bao (Shanghai)",
  ["sheng wu hua xue za zhi"] = "Chinese biochemical journal / Zhongguo sheng wu hua xue hui zhu ban|Sheng Wu Hua Hsueh Tsa Chih",
  ["sheng wu hua xue za zhi = chinese biochemical journal"] = "Sheng Wu Hua Hsueh Tsa Chih",
  ["sheng wu i hsueh kung cheng hsueh tsa chih journal of biomedical engineering (cheng-tu"] = "Sheng Wu I Hsueh Kung Cheng Hsueh Tsa Chih",
  ["sheng wu ji shu tong bao = biotechnology bulletin"] = "Sheng Wu Ji Shu Tong Bao",
  ["sheng wu jia gong guo cheng"] = "Sheng Wu Jia Gong Guo Cheng",
  ["sheng wu wu li hsueh bao"] = "Sheng Wu Wu Li Hsueh Bao",
  ["sheng wu xue za zhi = journal of biology"] = "Sheng Wu Xue Za Zhi",
  ["sheng wu yi xue gong cheng xue za zhi (journal of biomedical engineering)"] = "Sheng Wu Yi Xue Gong Cheng Xue Za Zhi",
  ["sheng wu yi xue gong cheng xue za zhi = journal of biomedical engineering = shengwu yixue gongchengxue zazhi"] = "Sheng Wu Yi Xue Gong Cheng Xue Za Zhi",
  ["sheng zhi yu bi yun = reproduction and contraception"] = "Shengzhi Yu Biyun",
  ["shengtai huanjing"] = "Shengtai Huanjing",
  ["shengwu huaxue yu shengwu wuli jinzhan"] = "Shengwu Huaxue Yu Shengwu Wuli Jinzhan",
  ["shengwu huaxue yu shengwu wuli xuebao"] = "Shengwu Huaxue Yu Shengwu Wuli Xuebao",
  ["shengwu huaxue zazhi"] = "Shengwu Huaxue Zazhi",
  ["shengwu jishu"] = "Biotechnology|Shengwu Jishu",
  ["shengwu jishu = biotechnology"] = "Shengwu Jishu",
  ["shengwu jishu tongxun"] = "Shengwu Jishu Tongxun",
  ["shengwu xinxixue"] = "Shengwu Xinxixue",
  ["shengwu ziyuan"] = "Shengwu Ziyuan",
  ["shenyang nongye daxue xuebao"] = "Shenyang Nongye Daxue Xuebao",
  ["shenyang yao ke da xue xue bao"] = "Journal of Shenyang Pharmaceutical University|Shenyang Yao Ke Da Xue Xue Bao",
  ["shenyang yao ke da xue xue bao = journal of shenyang pharmaceutical university"] = "Shenyang Yao Ke Da Xue Xue Bao",
  ["shenyang yaoxueyuan xuebao"] = "Shenyang Yaoxueyuan Xuebao",
  ["shëndetësia popullore"] = "Shendet Pop",
  ["shi da sheng wu xue bao"] = "Shi Da Sheng Wu Xue Bao",
  ["shi da xue bao"] = "Shi Da Xue Bao",
  ["shi jie hua ren xiao hua za zhi"] = "World Chinese journal of digestology|Shi Jie Hua Ren Xiao Hua Za Zhi",
  ["shi jie hua ren xiao hua za zhi = world chinese journal of digestology"] = "Shi Jie Hua Ren Xiao Hua Za Zhi",
  ["shi jie li shi = shi jie li shi"] = "Shi Jie Li Shi",
  ["shi pin yu fa xiao gong ye = shipin yu faxiao gongye"] = "Shi Pin Yu Fa Xiao Gong Ye",
  ["shi xue yue kan"] = "Shi Xue Yue Kan",
  ["shi yan li xue = journal of experimental mechanics"] = "Shi Yan Li Xue",
  ["shi yan sheng wu xue bao"] = "Shi Yan Sheng Wu Xue Bao",
  ["shi yan sheng wu xue bao (bulletin of experimental biology)"] = "Shi Yan Sheng Wu Xue Bao",
  ["shi yan yu jian yan yi xue"] = "Shi Yan Yu Jian Yan Yi Xue",
  ["shi yong jun : shuang yue kan = edible fungi"] = "Shi Yong Jun",
  ["shi yong jun xue bao (online)"] = "Shi Yong Jun Xue Bao",
  ["shigaku (odontology)"] = "Shigaku",
  ["shigaku = odontology; journal of nihon dental college"] = "Shigaku",
  ["shigaku zasshi"] = "Shigaku Zasshi",
  ["shigaku. odontology; journal of nippon dental college"] = "Shigaku",
  ["shih yen sheng wu hsueh pao journal of experimental biology"] = "Shih Yen Sheng Wu Hsueh Pao",
  ["shihezi daxue xuebao. ziran kexue ban = journal of shihezi university. natural science"] = "Shihezi Daxue Xuebao",
  ["shika gakuho. dental science reports"] = "Shikwa Gakuho",
  ["shika igaku (journal of osaka odontological society)"] = "Shika Igaku",
  ["shika kiso igakkai zasshi (japanese journal of oral biology)"] = "Shika Kiso Igakkai Zasshi",
  ["shika kiso igakkai zasshi = japanese journal of oral biology"] = "Shika Kiso Igakkai Zasshi",
  ["shika rikogaku zasshi (journal of the japan society for dental apparatus and materials)"] = "Shika Rikogaku Zasshi",
  ["shika rikogaku zasshi. journal of the japan society for dental apparatus and materials"] = "Shika Rikogaku Zasshi",
  ["shika rikōgaku zasshi. journal of the japan society for dental apparatus and materials"] = "Shika Rikogaku Zasshi",
  ["shika zairyo, kikai (journal of the japanese society for dental materials and devices)"] = "Shika Zairyo Kikai",
  ["shika zairyō, kikai = journal of the japanese society for dental materials and devices"] = "Shika Zairyo Kikai",
  ["shikai tenbo (dental outlook)"] = "Shikai Tenbo",
  ["shikai tenbo = dental outlook"] = "Shikai Tenbo",
  ["shikwa gakuho (dental science reports)"] = "Shikwa Gakuho",
  ["shilap : revista de lepidopterología"] = "SHILAP Soc Hispano Luso Am Lepid",
  ["shimane journal of medical science"] = "Shimane J Med Sci",
  ["shinkei ganka = neuro-ophthalmology japan"] = "Shinkei Ganka",
  ["shinkei kenkyu no shimpo (advances in neurological sciences)"] = "Shinkei Kenkyu No Shimpo",
  ["shinkei kenkyu no shimpo. advances in neurological sciences"] = "Shinkei Kenkyu No Shimpo",
  ["shinkei kenkyū no shimpo. advances in neurological sciences"] = "Shinkei Kenkyu No Shimpo",
  ["shinrigaku kenkyu (japanese journal of psychology)"] = "Shinrigaku Kenkyu",
  ["shinrigaku kenkyu : the japanese journal of psychology"] = "Shinrigaku Kenkyu",
  ["shinrigaku kenkyu. japanese journal of psychology"] = "Shinrigaku Kenkyu",
  ["ship technology research"] = "Ship Technol. Res.",
  ["shipin anquan zhiliang jiance xuebao"] = "Shipin Anquan ZhiliangJjiance Xuebao",
  ["shipin kexue"] = "Shipin Kexue (Beijing)",
  ["shipin kexue (beijing)"] = "Shipin Kexue (Beijing)",
  ["shipin yu shengwu jishu xuebao"] = "Shipin Yu Shengwu Jishu Xuebao",
  ["ships and offshore structures"] = "Ships Offshore Struct.",
  ["shiraz e-medical journal"] = "Shiraz E Med J",
  ["shirin (kyoto, japan)"] = "Shirin (Kyoto 1916)",
  ["shishitsu seikagaku kenkyu. [studies on biochemistry of lipids]"] = "Shishitsu Seikagaku Kenkyu",
  ["shiyan wulixeu congshu"] = "Shiyan Wulixue Congshu",
  ["shiyo (journal of the tokyo dental association)"] = "Shiyo",
  ["shiyo. the journal of the tokyo dental association"] = "Shiyo",
  ["shiyou huagong"] = "Shiyou Huagong",
  ["shiyou xuebao, shiyou jiagong"] = "Shiyou Xuebao, Shiyou Jiagong",
  ["shizen to bunka (fukuoka, japan)"] = "Nat Cult",
  ["shizhen guo yi guo yao = lishizhen medicine and materia medica research"] = "Shizhen Guo Yi Guo Yao",
  ["shizuoka daigaku nōgakubu kenkyū hōkoku = bulletin of the faculty of agriculture, shizuoka university"] = "Shizuoka Daigaku Nogakubu Kenkyu Hokoku",
  ["shock (augusta, ga.)"] = "Shock",
  ["shock and vibration"] = "Shock Vib.",
  ["shock wave and high pressure phenomena"] = "Shock Wave High Press. Phenom.",
  ["shock waves"] = "Shock Waves",
  ["shofar (west lafayette, ind.)"] = "Shofar (West Lafayette Indiana)",
  ["shokubutsu kenkyu zasshi. the journal of japanese botany"] = "Shokubutsu Kenkyu Zasshi",
  ["shokuhin eiseigaku zasshi. journal of the food hygienic society of japan"] = "Shokuhin Eiseigaku Zasshi",
  ["shokuseishi kenkyu = japanese journal of historical botany"] = "Shokuseishi Kenkyu",
  ["shoni shikagaku zasshi (japanese journal of pedodontics)"] = "Shoni Shikagaku Zasshi",
  ["shoni shikagaku zasshi. the japanese journal of pedodontics"] = "Shoni Shikagaku Zasshi",
  ["shonika kiyo. annales paediatrici japonici"] = "Shonika Kiyo",
  ["shonika. pediatrics of japan"] = "Shonika",
  ["short textbooks in logic"] = "Short Textb. Log.",
  ["shou lei xue bao"] = "Acta theriologica sinica|Shou Lei Xue Bao",
  ["shoulder & elbow"] = "Shoulder Elbow",
  ["showa igakkai zasshi (journal of the showa medical association)"] = "Showa Igakkai Zasshi",
  ["showa shigakkai zasshi (journal of showa university dental society)"] = "Showa Shigakkai Zasshi",
  ["shòu lèi xuè bào = acta theriologica sinica"] = "Shou Lei Xue Bao",
  ["shōni shikagaku zasshi. the japanese journal of pedodontics"] = "Shoni Shikagaku Zasshi",
  ["shōwa igakkai zasshi = the journal of the showa medical association"] = "Showa Igakkai Zasshi",
  ["shōwa shigakkai zasshi = the journal of showa university dental society"] = "Showa Shigakkai Zasshi",
  ["shui chan yan jiu = journal of taiwan fisheries research"] = "Shui Chan Yan Jiu",
  ["shui chu li ji shu"] = "Shui Chu Li Ji Shu",
  ["shui li yu ye"] = "Shui Li Yu Ye",
  ["shui sheng sheng wu hsueh bao = acta hydrobiologica sinica"] = "Shui Sheng Sheng Wu Hsueh Bao",
  ["shui sheng tai xue za zhi = journal of hydroecology"] = "Shui Sheng Tai Xue Za Zhi",
  ["shuichan xuebao"] = "Shuichan Xuebao",
  ["shuichuli jishu"] = "Shuichuli Jishu",
  ["shujutsu (operation)"] = "Shujutsu",
  ["shujutsu. operation"] = "Shujutsu",
  ["sh’ma : a journal of jewish responsibility"] = "Shma",
  ["si yu yan"] = "Si Yu Yan",
  ["siam journal of scientific computing"] = "SIAM J. Sci. Comput.",
  ["siam journal on applied algebra and geometry"] = "SIAM J. Appl. Algebra Geom.",
  ["siam journal on applied dynamical systems"] = "SIAM J. Appl. Dyn. Syst.",
  ["siam journal on applied mathematics"] = "SIAM J. Appl. Math.",
  ["siam journal on computing"] = "SIAM J. Comput.",
  ["siam journal on control and optimization"] = "SIAM J. Control Optim.",
  ["siam journal on discrete mathematics"] = "SIAM J. Discrete Math.",
  ["siam journal on financial mathematics"] = "SIAM J. Financ. Math.",
  ["siam journal on imaging sciences"] = "SIAM J. Imag. Sci.",
  ["siam journal on mathematical analysis"] = "SIAM J. Math. Anal.",
  ["siam journal on mathematics of data science"] = "SIAM J. Math. Data Sci.",
  ["siam journal on matrix analysis and applications"] = "SIAM J. Matrix Anal. Appl.",
  ["siam journal on matrix analysis and applications : a publication of the society for industrial and applied mathematics"] = "SIAM J Matrix Anal Appl",
  ["siam journal on numerical analysis"] = "SIAM J. Numer. Anal.",
  ["siam journal on optimization"] = "SIAM J. Optim.",
  ["siam journal on optimization : a publication of the society for industrial and applied mathematics"] = "SIAM J Optim",
  ["siam journal on scientific and statistical computing"] = "SIAM J. Sci. Stat. Comput",
  ["siam journal on scientific computing"] = "SIAM J. Sci. Comput.",
  ["siam journal on scientific computing : a publication of the society for industrial and applied mathematics"] = "SIAM J Sci Comput",
  ["siam monographs on mathematical modeling and computation"] = "SIAM Monogr. Math. Model. Comput.",
  ["siam review"] = "SIAM Rev.",
  ["siam review. society for industrial and applied mathematics"] = "SIAM Rev Soc Ind Appl Math",
  ["siam studies in applied mathematics"] = "SIAM Stud. Appl. Math.",
  ["siam undergraduate research online"] = "SIAM Undergrad Res Online",
  ["siam/asa journal on uncertainty quantification"] = "SIAM/ASA J. Uncertainty Quantif.",
  ["siauliai mathematical seminar"] = "Siauliai Math. Semin.",
  ["siberian advances in mathematics"] = "Siberian Adv. Math.",
  ["siberian journal of computer mathematics"] = "Siberian J. Comput. Math.",
  ["siberian journal of differential equations"] = "Siberian J. Differential Equations",
  ["siberian mathematical journal"] = "Sib. Math. J.",
  ["siberian school of algebra and logic"] = "Siberian School Algebra Logic",
  ["sibirskie elektronnye matematicheskie izvestiya. siberian electronic mathematical reports"] = "Sib. Elektron. Mat. Izv.",
  ["sibirskii ekologicheskii zhurnal"] = "Sibirskii Ekolog Z",
  ["sibirskii zhurnal chistoi i prikladnoi matematiki"] = "Sib. Zh. Chist. Prikl. Mat.",
  ["sibirskii zhurnal industrialnoi matematiki"] = "Sib. Zh. Ind. Mat.",
  ["sibirskii zhurnal vychislitelnoi matematiki. siberian journal of numerical mathematics. rossiiskaya akademiya nauk. sibirskoe otdelenie"] = "Sib. Zh. Vychisl. Mat.",
  ["sibirskiĭ vestnik selʹskokhozi︠a︡ĭstvennoĭ nauki"] = "Sibirskii Vestnik Selskokhoziaistvennoi Nauki",
  ["sibrium. collana di studi e documentazioni"] = "Sibrium",
  ["sice journal of control, measurement, and system integration"] = "SICE J. Control Meas. Syst. Integr.",
  ["sichuan da xue xue bao = sichuandaxue xuebao"] = "Sichuan Da Xue Xue Bao",
  ["sichuan da xue xue bao. yi xue ban (journal of sichuan university. medical science edition)"] = "Sichuan Da Xue Xue Bao Yi Xue Ban",
  ["sichuan da xue xue bao. yi xue ban = journal of sichuan university. medical science edition"] = "Sichuan Da Xue Xue Bao Yi Xue Ban",
  ["sichuan da xue xue bao. zi ran ke xue ban"] = "Sichuan daxue xuebao|Sichuan Da Xue Xue Bao Zi Ran Ke Xue Ban",
  ["sichuan da xue xue bao. zi ran ke xue ban = sichuan daxue xuebao"] = "Sichuan Da Xue Xue Bao Zi Ran Ke Xue Ban",
  ["sichuan daxue xuebao"] = "Sichuan Daxue Xuebao",
  ["sichuan dong wu = sichuan dongwu = sichuan journal of zoology"] = "Sichuan Dong Wu",
  ["sichuan lin ye ke ji = journal of sichuan forestry science and technology"] = "Sichuan Lin Ye Ke Ji",
  ["sichuan shifan daxue xuebao"] = "Sichuan Shifan Daxue Xuebao Ziran Kexue Ban",
  ["sichuan shifan daxue xuebao (ziran kexue ban)"] = "Sichuan Shifan Daxue Xuebao",
  ["sichuan yi xue yuan xue bao (acta academiae medicinae sichuan)"] = "Sichuan Yi Xue Yuan Xue Bao",
  ["sichuan yi xue yuan xue bao = acta academiae medicinae sichuan"] = "Sichuan Yi Xue Yuan Xue Bao",
  ["sicilia archeologica"] = "SicA",
  ["sicilia medica"] = "Sicilia Med",
  ["sicilia sanitaria"] = "Sicil Sanit",
  ["sics software-intensive cyber-physical systems"] = "SICS Software-Intensive Cyber-Phys. Syst.",
  ["siculorum gymnasium"] = "SicGymn",
  ["siculorum gymnasium: rassegna semestrale della facoltà di lettere e filosofia dell’università di catania"] = "SicGymn",
  ["sid symposium digest of technical papers"] = "SID Symp. Dig. Tech. Pap.",
  ["sidahora : un proyecto del departamento de publicaciones del pwa coalition, ny"] = "Sidahora",
  ["sidahora : un proyecto del departamento de publicaciónes del pwa coalition, ny"] = "Sidahora",
  ["sieccan journal"] = "Sieccan J",
  ["siecus report"] = "SIECUS Rep",
  ["siemens review"] = "Siemens Rev.",
  ["siggraph asia 2017 symposium on visualization. siggraph asia symposium on visualization (2017 : bangkok, thailand)"] = "SIGGRAPH Asia 2017 Symp Vis (2017)",
  ["sight and life magazine"] = "Sight Life Mag",
  ["sight-saving review"] = "Sight. Sav. Rev.",
  ["sightsaving review"] = "Sightsav. Rev.",
  ["sigkdd explorations : newsletter of the special interest group (sig) on knowledge discovery & data mining"] = "SIGKDD Explor",
  ["sigma journal of engineering and natural sciences"] = "Sigma J. Eng. Nat. Sci.",
  ["sigma series in applied mathematics"] = "Sigma Ser. Appl. Math.",
  ["sigma series in pure mathematics"] = "Sigma Ser. Pure Math.",
  ["sigma. symmetry, integrability and geometry. methods and applications"] = "SIGMA Symmetry Integrability Geom. Methods Appl.",
  ["sigmod record"] = "SIGMOD Rec",
  ["sigmul bunryu hag-hoeji"] = "Korean journal of plant taxonomy|Sigmul Bunryu Hag-hoeji",
  ["sigmul bunryu hag-hoeji = korean journal of plant taxonomy"] = "Sigmul Bunryu Hag Hoeji",
  ["sigmul saengmyeong gong haghoeji"] = "Korean journal of plant biotechnology|Sigmul Saengmyeong Gong Haghoeji",
  ["sigmul saengmyeong gong haghoeji = korean journal of plant biotechnology"] = "Sigmul Saengmyeong Gong Haghoeji",
  ["sigmulbyeong yeon'gu"] = "Sigmulbyeong Yeongu",
  ["sigmund freud house bulletin"] = "Sigmund Freud House Bull",
  ["sign language and linguistics"] = "Sign Lang Linguist",
  ["sign language studies"] = "Sign Lang Stud",
  ["signal and image processing : an international journal"] = "Signal Image Process",
  ["signal processing"] = "Signal Process.",
  ["signal processing and digital filtering"] = "Signal Process. Digit. Filtering",
  ["signal processing. image communication"] = "Signal Process Image Commun",
  ["signal processing: image communication"] = "Signal Process. Image Commun.",
  ["signal transduction"] = "Signal Transduct",
  ["signal transduction and targeted therapy"] = "Signal Transduct Target Ther",
  ["signal transduction insights"] = "Sign Transduct Insights",
  ["signal, image and video processing"] = "Signal Image Video Process",
  ["signals and communication technology"] = "Signals Commun. Technol.",
  ["signature (ramsey, n.j.)"] = "Signature",
  ["significance (oxford, england)"] = "Signif (Oxf)",
  ["signs and society (chicago, ill.)"] = "Signs Soc (Chic)",
  ["sileno. rivista di studi classici e cristiani"] = "Sileno",
  ["silica and associated respirable mineral particles"] = "Silica Assoc Respirable Miner Part (2012)",
  ["silicates industriels"] = "Silic. Indus.",
  ["silicon chemistry"] = "Silicon Chem.",
  ["silk road art and archaeology. journal of the institute of silk road studies, kamakura"] = "SilkRoadArtA",
  ["sillūl va bāft"] = "Sillul Baft",
  ["silva balcanica"] = "Silva Balcan.",
  ["silva fennica"] = "Silva Fenn.",
  ["silva fennica (helsinki, finland : 1967)"] = "Silva Fenn (Hels)",
  ["silva lusitana"] = "Silva Lusit",
  ["silvae genetica"] = "Silvae Genet",
  ["silvaecultura tropica et subtropica"] = "Silvaecult. Trop. Subtrop.",
  ["simblos. scritti di storia antica"] = "Simblos",
  ["simon wiesenthal center annual"] = "Simon Wiesenthal Cent Annu",
  ["simons symposia"] = "Simons Symp.",
  ["simula springerbriefs on computing"] = "Simula SpringerBriefs Comput.",
  ["simulation & games"] = "Simul Games",
  ["simulation & gaming"] = "Simul Gaming",
  ["simulation foundations, methods and applications"] = "Simul. Found. Methods Appl.",
  ["simulation in healthcare : journal of the society for simulation in healthcare"] = "Simul Healthc",
  ["simulation modelling practice and theory"] = "Simul Model Pract Theory",
  ["simulation practice and theory"] = "Simul. Pract. Theory",
  ["sinai hospital journal"] = "Sinai Hosp J (Balt)",
  ["sinet: ethiopian journal of science"] = "SINET",
  ["singapore dental journal"] = "Singapore Dent J",
  ["singapore economic review"] = "Singapore Econ. Rev.",
  ["singapore journal of obstetrics & gynaecology"] = "Singapore J Obstet Gynecol",
  ["singapore journal of tropical geography"] = "Singap J Trop Geogr",
  ["singapore libraries"] = "Singap Libr",
  ["singapore medical journal"] = "Singapore Med J",
  ["singapore statistical bulletin. singapore. national statistical commission"] = "Singap Stat Bull",
  ["single cell biology"] = "Single Cell Biol",
  ["singmul hakhoe chi. the korean journal of botany"] = "Singmul Hakhoe Chi",
  ["sinh học"] = "Sinh Hoc",
  ["sino-western cultural relations journal = zhong xi wen hua jiao liu shi za zhi"] = "Sino West Cult Relat J",
  ["sintesi medica"] = "Sintesi Medica",
  ["siriraj hospital gazette"] = "Siriraj Hosp Gaz",
  ["siriraj medical journal"] = "Siriraj Med J",
  ["sissa springer series"] = "SISSA Springer Ser.",
  ["sistema nervoso"] = "Sist Nerv",
  ["sistemas e tecnologias de informação : atas de 12a conferência ibérica de sistemas e tecnologias de informação (cisti'2017) : 21 a 24 de junho de 2017, lisboa, portugal = information systems and technologies : proceedings of the 12th iberian conference on information systems and technologies (cisti'2017) : 21 to 24 of june 2017, lisbon, portugal. iberian conference on information systems and technologies (12th : 2017 : lisbon, portugal)"] = "Sist Tecnol Inf (2017)",
  ["sistemi urbani"] = "Sist Urbani",
  ["sitzungsberichte bayerische akademie der wissenschaften, mathematische - naturwissenschaftliche klasse"] = "Sitzungsber. Bayer. Akad. Wiss. Math. Naturwiss. Kl.",
  ["sitzungsberichte der akademie der wissenschaften der ddr, mathematik, naturwissenschaften, technik"] = "Sitzungsber. Akad. Wiss. DDR, Math., Naturwiss., Tech.",
  ["sitzungsberichte der akademie der wissenschaften der ddr, mathematik—naturwissenschaften—technik, jahrgang 1976"] = "Sitzungsber. Akad. Wiss. DDR Math. Naturwiss. Tech. Jahrg. 1976",
  ["sitzungsberichte der akademie der wissenschaften in berlin, gesellschaftswissenschaften (g)"] = "SAWDDR",
  ["sitzungsberichte der bayerischen akademie der wissenschaften, philos.- hist. klasse"] = "SBAW",
  ["sitzungsberichte der deutschen akademie der wissenschaften zu berlin. klasse für sprache, literatur und kunst"] = "SBBerlin",
  ["sitzungsberichte der heidelberger akademie der wissenschaften, mathematisch-naturwissenschaftliche klasse"] = "Sitzungsber Heidelb Akad Wiss Math Naturwiss Kl",
  ["sitzungsberichte der heidelberger akademie der wissenschaften, mathematische - naturwissenschaftliche klasse"] = "Sitzungsber. Heidelb. Akad. Wiss. Math. Naturwiss. Kl.",
  ["sitzungsberichte der heidelberger akademie der wissenschaften, philos.- hist. klasse"] = "SHAW",
  ["sitzungsberichte der koniglichen preussischen akademie der wissenschaften"] = "Sitzungsber. K. Preuss. Akad. Wiss.",
  ["sitzungsberichte der mathematisch-naturwissenschaftlichen klasse"] = "Sitzungsber. Math.-Naturwiss. Kl.",
  ["sitzungsberichte der sachsischen akademie der wissenschaften zu leipzig. mathematisch-naturwissenschaftliche klasse. band 116"] = "Sitzungsber. Sachsischen Akad. Wissenschaft. Leipzig. Math. Nat. Sci. Sect. Band 116",
  ["sitzungsberichte der sächsischen akademie der wissenschaften zu leipzig"] = "SBLeipzig",
  ["sitzungsberichte der wissenschaftlichen gesellschaft an der johann wolfgang goethe-universität frankfurt am main"] = "Sitzungsber. Wiss. Ges. Johann Wolfgang Goethe-Univ. Frankfurt am Main",
  ["sitzungsberichte der österreichischen akademie der wissenschaft in wien, philos.-hist. klasse"] = "SAWW",
  ["sitzungsberichte. österreichische akademie der wissenschaften"] = "SBWien",
  ["sixth north american paleontological convention : abstracts and program : field museum of natural history, june 9-june 12, 1996. north american paleontological convention (6th : 1996 : smithsonian institution, washington, d.c.)"] = "Sixth N Am Paleontol Conv N Am Paleontol Conv 6th 1996 Smithson Inst Washington DC",
  ["sixth-generation computer technology series"] = "Sixth-Gener. Comput. Tech. Ser.",
  ["síntesis de trabajos científicos"] = "Sint Trab Cient",
  ["sjukskoterskan / svensk sjukskoterskeforening"] = "Sjukskoterskan",
  ["skandinavisches archiv fur physiologie"] = "Skand. Arch. Physiol.",
  ["skandinavisches archiv für physiologie"] = "Skand Arch Physiol",
  ["skandinavisk veterinär-tidsskrift för bakteriologi, patologi, samt kjöttoch mjölkhygienen"] = "Skandinavisk Vet Tidsskr Bakteriol Patol Samt Kjottoch Mjokhygienen",
  ["skandinaviska banken quarterly review"] = "Skand. Banken Quart. Rev.",
  ["skeletal muscle"] = "Skelet Muscle",
  ["skeletal radiology"] = "Skeletal Radiol",
  ["skin appendage disorders"] = "Skin Appendage Disord",
  ["skin pharmacology"] = "Skin Pharmacol.",
  ["skin pharmacology : the official journal of the skin pharmacology society"] = "Skin Pharmacol",
  ["skin pharmacology and applied skin physiology"] = "Skin Pharmacol Appl Skin Physiol",
  ["skin pharmacology and physiology"] = "Skin Pharmacol Physiol",
  ["skin research and technology : official journal of international society for bioengineering and the skin (isbs) [and] international society for digital imaging of skin (isdis) [and] international society for skin imaging (issi)"] = "Skin Res Technol",
  ["skin therapy letter"] = "Skin Therapy Lett",
  ["skrifter utgivna av svenska institutet i rom. opuscula archaeologica"] = "OpArch",
  ["skripten zur mathematischen statistik"] = "Skript. Math. Statist.",
  ["skrypty uniwersytetu śląskiego"] = "Skrypty Uniw. Śląskiego",
  ["skuast journal of research"] = "SKUAST J Res",
  ["skull base reports"] = "Skull Base Rep",
  ["skull base surgery"] = "Skull Base Surg",
  ["sky and telescope"] = "Sky Telescope",
  ["skyllis. zeitschrift für unterwasserarchäologie"] = "Skyllis",
  ["slas discovery"] = "SLAS Discovery",
  ["slas discovery : advancing life sciences r & d"] = "SLAS Discov",
  ["slas discovery: advancing the science of drug discovery"] = "SLAS Discovery",
  ["slas technology"] = "SLAS Technol",
  ["slas technology: translating life sciences innovation"] = "SLAS Technol.",
  ["slaski kwartalnik historyczny sobotka / wroclawskie towarzystwo milosnikow historii"] = "Slaski Kwat Hist Sabotka",
  ["slavery & abolition"] = "Slavery Abol",
  ["slavic review"] = "Slavic Rev",
  ["slavonic and east european review"] = "Slav East Eur Rev",
  ["sleep & breathing = schlaf & atmung"] = "Sleep Breath",
  ["sleep advances"] = "SLEEP Adv.",
  ["sleep advances : a journal of the sleep research society"] = "Sleep Adv",
  ["sleep and biological rhythms"] = "Sleep Biol Rhythms",
  ["sleep and breathing"] = "Sleep Breathing",
  ["sleep and hypnosis : an international journal of sleep, dream, and hypnosis"] = "Sleep Hypn",
  ["sleep and vigilance"] = "Sleep Vigil",
  ["sleep disorders"] = "Sleep Disord",
  ["sleep health"] = "Sleep Health",
  ["sleep medicine"] = "Sleep Med",
  ["sleep medicine and disorders : international journal"] = "Sleep Med Disord",
  ["sleep medicine clinics"] = "Sleep Med Clin",
  ["sleep medicine research"] = "Sleep Med Res",
  ["sleep medicine reviews"] = "Sleep Med Rev",
  ["sleep medicine: x"] = "Sleep Med X",
  ["sleep research online"] = "Sleep Res. Online",
  ["sleep research online : sro"] = "Sleep Res Online",
  ["sleep science"] = "Sleep Sci.",
  ["sleep science (são paulo, brazil)"] = "Sleep Sci",
  ["sleep science and practice"] = "Sleep Sci Pract",
  ["slezsky sbornik"] = "Slezsky Sb",
  ["slezský sborník"] = "Slezsky Sb",
  ["sloan management review"] = "Sloan Manage Rev",
  ["slovansky prehled"] = "Slov Prehl",
  ["slovanský přehled"] = "Slov Prehl",
  ["slovenska archivistika"] = "Slov Arch",
  ["slovenská archivistika"] = "Slov Arch",
  ["slovenski veterinarski zbornik = slovenian veterinary research"] = "Slov Vet Zb",
  ["slovensky lekar"] = "Slov Lek",
  ["slovensky lekár"] = "Slov Lek",
  ["slovenská archeológia"] = "SlovA",
  ["slovenská numizmatika"] = "SlovNum",
  ["slownik biograficzny polskich nauk medycznych xx wieku / instytut historii nauki, oswiaty i techniki pan, pracownia historii nauk medycznych"] = "Slown Biogr Pol Med XX Wieku",
  ["sm journal of clinical and medical imaging"] = "SM J Clin Med Imaging",
  ["sm journal of community medicine"] = "SM J Community Med",
  ["sm journal of reproductive health & infertility"] = "SM J Reprod Health Infertil",
  ["sm radiology journal"] = "SM Radiol J",
  ["sm tropical medicine journal"] = "SM Trop Med J",
  ["smai journal of computational mathematics"] = "SMAI J. Comput. Math.",
  ["small (weinheim an der bergstrasse, germany)"] = "Small",
  ["small and advanced healthcare materials"] = "Small Adv. Healthcare Mater.",
  ["small business economics"] = "Small Bus. Econ.",
  ["small carnivore conservation : the newsletter and journal of the iucn/ssc mustelid, viverrid & procynid specialist group"] = "Small Carniv Conserv",
  ["small group research"] = "Small Group Res",
  ["small gtpases"] = "Small GTPases",
  ["small methods"] = "Small Methods",
  ["small molecules"] = "Small Mol.",
  ["small ruminant research"] = "Small Ruminant Res.",
  ["small ruminant research : the journal of the international goat association"] = "Small Rumin Res",
  ["small science"] = "Small Sci.",
  ["small structures"] = "Small Struct",
  ["small-scale forest economics, management and policy"] = "Small-scale For. Econ. Manage. Policy",
  ["small-scale forestry"] = "Small-Scale For.",
  ["smarandache notions journal"] = "Smarandache Notions J.",
  ["smart agricultural technology"] = "Smart Agric. Technol.",
  ["smart and resilient transportation"] = "Smart Resilient Transp.",
  ["smart and sustainable built environment"] = "Smart Sustainable Built Environ.",
  ["smart and sustainable manufacturing systems"] = "Smart Sustain Manuf Syst",
  ["smart energy"] = "Smart Energy",
  ["smart health"] = "Smart Health",
  ["smart health (amsterdam, netherlands)"] = "Smart Health (Amst)",
  ["smart homecare technology and telehealth"] = "Smart Homecare Technol Telehealth",
  ["smart homes and beyond"] = "Assist. Technol. Res. S.",
  ["smart infrastructure and construction"] = "Smart Infrastruct. Constr.",
  ["smart innovation, systems and technologies"] = "Smart Innov. Syst. Technol.",
  ["smart learning environments"] = "Smart Learn. Environ.",
  ["smart manufacturing"] = "Smart Manuf.",
  ["smart materials & structures"] = "Smart Mater Struct",
  ["smart materials and structures"] = "Smart Mater. Struct.",
  ["smart materials bulletin"] = "Smart Mater. Bull.",
  ["smart materials in manufacturing"] = "Smart Mater. Manuf.",
  ["smart materials in medicine"] = "Smart Mater Med",
  ["smart medicine"] = "Smart Med.",
  ["smart science"] = "Smart Sci.",
  ["smart structures and materials. nondestructive evaluation for health monitoring and diagnostics"] = "Smart Struct Mater Nondestruct Eval Health Monit Diagn",
  ["smart structures and systems"] = "Smart Struct Syst",
  ["smart water"] = "Smart Water",
  ["smf journee annuelle"] = "SMF Journ. Annu.",
  ["smf journée annuelle"] = "SMF Journ. Annu.",
  ["smith college studies in social work"] = "Smith Coll Stud Soc Work",
  ["smith-hurd illinois annotated statutes. illinois"] = "Smith-Hurd Ill Annot Statut Ill",
  ["smithiana. bulletin"] = "Smithiana Bull",
  ["smithsonian annals of flight"] = "Smithson Ann Flight",
  ["smithsonian contributions to the marine sciences"] = "Smithson Contrib Mar Sci",
  ["smithsonian contributions to zoology"] = "Smithson Contrib Zool",
  ["smithsonian studies in history and technology. smithsonian institution"] = "Smithson Stud Hist Technol",
  ["smpte motion imaging journal"] = "SMPTE Motion Imaging J",
  ["smst : proceedings of the international conference on shape memory and superelastic technologies. international conference on shape memory and superelastic technologies"] = "SMST",
  ["smu law review"] = "SMU Law Rev.",
  ["smu law review : a publication of southern methodist university school of law"] = "SMU Law Rev",
  ["sn applied sciences"] = "SN Appl Sci",
  ["sn comprehensive clinical medicine"] = "SN Compr Clin Med",
  ["sn computer science"] = "SN Comput Sci",
  ["sn operations research forum"] = "SN Oper. Res. Forum",
  ["sn partial differential equations and applications"] = "SN Partial Differ Equ Appl",
  ["snakeroot extract"] = "Snakeroot Extr",
  ["snudebiller : studies on taxonomy, biology, and ecology of curculionoidea"] = "Snudebiller (Monchengladb)",
  ["soa--chʻŏngsonyŏn chŏngsin ŭihak = journal of child & adolescent psychiatry"] = "Soa Chongsonyon Chongsin Uihak",
  ["soap and sanitary chemicals"] = "Soap Sanit Chem",
  ["soap and sanitary chemicals. blue book and catalog ed"] = "Soap Sanit Chem Blue Book Cat Ed",
  ["soap, cosmetics, chemical specialties"] = "Soap, Cosmet., Chem. Spec.",
  ["soap, perfumery & cosmetics"] = "Soap Perfum Cosmet (Lond)",
  ["soccer and society"] = "Soccer Soc",
  ["social & cultural geography"] = "Soc Cult Geogr",
  ["social & legal studies"] = "Soc Leg Stud",
  ["social action"] = "Soc Action",
  ["social analysis (adelaide, s.a.)"] = "Soc Anal",
  ["social and economic administration"] = "Soc Econ Adm (Lond)",
  ["social and economic studies"] = "Soc Econ Stud",
  ["social and labour bulletin"] = "Soc Labour Bull",
  ["social and personality psychology compass"] = "Soc Personal Psychol Compass",
  ["social and rehabilitation record"] = "Soc. Rehabil. Rec.",
  ["social anthropology : the journal of the european association of social anthropologists = anthropologie sociale"] = "Soc Anthropol",
  ["social behavior and personality"] = "Soc Behav Pers",
  ["social behaviour"] = "Soc Behav",
  ["social biology"] = "Soc Biol",
  ["social biology and human affairs"] = "Soc Biol Hum Aff",
  ["social casework"] = "Soc Casework",
  ["social change"] = "Soc Change",
  ["social choice and welfare"] = "Soc Choice Welfare",
  ["social cognition"] = "Soc Cogn",
  ["social cognition: the basis of human interaction"] = "Front. Soc. Psychol.",
  ["social cognitive and affective neuroscience"] = "Soc Cogn Affect Neurosci",
  ["social compass"] = "Soc Compass",
  ["social computing, behavioral-cultural modeling, and prediction : 8th international conference, sbp 2015, washington, dc, usa, march 31-april 3, 2015. proceedings. international conference on social computing, behavioral modeling and prediction (8th : 2015 : washington, d.c.)"] = "Soc Comput Behav Cult Model Predict (2015)",
  ["social currents"] = "Soc Curr",
  ["social development (oxford, england)"] = "Soc Dev",
  ["social development issues"] = "Soc Dev Issues",
  ["social dynamics"] = "Soc Dyn",
  ["social epistemology"] = "Soc Epistemol",
  ["social forces"] = "Soc. Forces",
  ["social forces; a scientific medium of social study and interpretation"] = "Soc Forces",
  ["social history"] = "Soc Hist",
  ["social history of medicine : the journal of the society for the social history of medicine"] = "Soc Hist Med",
  ["social history of medicine : the journal of the society for the social history of medicine / sshm"] = "Soc Hist Med",
  ["social inclusion"] = "Soc Incl",
  ["social indicators research"] = "Soc Indic Res",
  ["social influence"] = "Soc Influ",
  ["social informatics : 8th international conference, socinfo 2016, bellevue, wa, usa, november 11-14, 2016, proceedings. part ii. socinfo (conference) (8th : 2016 : bellevue, wash.)"] = "Soc Inform (2016)",
  ["social issues and policy review"] = "Soc Issues Policy Rev",
  ["social justice (san francisco, calif.)"] = "Soc Justice",
  ["social justice research"] = "Soc Justice Res",
  ["social justice review"] = "Soc Justice Rev",
  ["social marketing forum"] = "Soc Mark Forum",
  ["social marketing quarterly"] = "Soc Mar Q",
  ["social marketing update"] = "Soc Mark Update",
  ["social media + society"] = "Soc Media Soc",
  ["social medicine (social medicine publication group)"] = "Soc Med (Soc Med Publ Group)",
  ["social movement studies"] = "Soc Mov Stud",
  ["social network analysis and mining"] = "Soc Netw Anal Min",
  ["social networking"] = "Soc Netw",
  ["social networks"] = "Soc Networks",
  ["social neuroscience"] = "Soc Neurosci",
  ["social philosophy & policy"] = "Soc Philos Policy",
  ["social policy"] = "Soc Policy",
  ["social policy & administration"] = "Soc Policy Adm",
  ["social policy and society : a journal of the social policy association"] = "Soc Policy Soc",
  ["social policy report"] = "Soc Policy Rep",
  ["social policy report / society for research in child development"] = "Soc Policy Rep",
  ["social politics"] = "Soc Polit",
  ["social praxis"] = "Soc Prax",
  ["social problems"] = "Soc Probl",
  ["social psychiatry"] = "Soc. Psychiatry",
  ["social psychiatry and psychiatric epidemiology"] = "Soc Psychiatry Psychiatr Epidemiol",
  ["social psychiatry. sozialpsychiatrie. psychiatrie sociale"] = "Soc Psychiatry",
  ["social psychological and personality science"] = "Soc Psychol Personal Sci",
  ["social psychology"] = "Soc Psychol",
  ["social psychology of education : an international journal"] = "Soc Psychol Educ",
  ["social psychology quarterly"] = "Soc Psychol Q",
  ["social research"] = "Soc Res (New York)",
  ["social responsibility, journalism, law, medicine"] = "Soc Responsib Journal Law Med",
  ["social science"] = "Soc Sci",
  ["social science & medicine"] = "Soc Sci Med",
  ["social science & medicine (1982)"] = "Soc Sci Med",
  ["social science & medicine. medical anthropology"] = "Soc Sci Med Med Anthropol",
  ["social science & medicine. medical economics"] = "Soc Sci Med Med Econ",
  ["social science & medicine. medical geography"] = "Soc Sci Med Med Geogr",
  ["social science & medicine. medical psychology & medical sociology"] = "Soc Sci Med Med Psychol Med Sociol",
  ["social science & medicine. part a, medical sociology"] = "Soc Sci Med A",
  ["social science & medicine. part b, medical anthropology"] = "Soc Sci Med B",
  ["social science & medicine. part c, medical economics"] = "Soc Sci Med C",
  ["social science & medicine. part d, medical geography"] = "Soc Sci Med D",
  ["social science & medicine. part e, medical psychology"] = "Soc Sci Med E",
  ["social science & medicine. part f, medical & social ethics"] = "Soc Sci Med F",
  ["social science and dentistry"] = "Soc Sci Dent",
  ["social science and medicine"] = "Soc. Sci. Med.",
  ["social science and medicine. medical anthropology"] = "Soc. Sci. Med. [Med. Anthropol.]",
  ["social science and medicine. medical economics"] = "Soc. Sci. Med. [Med. Econ.]",
  ["social science and medicine. medical geography"] = "Soc. Sci. Med. [Med. Geogr.]",
  ["social science and medicine. medical psychology and medical sociology"] = "Soc. Sci. Med. [Med. Psychol. Med. Sociol.]",
  ["social science and medicine. part a, medical sociology"] = "Soc. Sci. Med. [A]",
  ["social science and medicine. part b, medical anthropology"] = "Soc. Sci. Med. [B]",
  ["social science and medicine. part c, medical economics"] = "Soc. Sci. Med. [C]",
  ["social science and medicine. part d, medical geography"] = "Soc. Sci. Med. [D]",
  ["social science and medicine. part e, medical psychology"] = "Soc. Sci. Med. [E]",
  ["social science and medicine. part f, medical and social ethics"] = "Soc. Sci. Med. [F]",
  ["social science computer review"] = "Soc Sci Comput Rev",
  ["social science diliman"] = "Soc Sci Diliman",
  ["social science history"] = "Soc Sci Hist",
  ["social science information studies : ssis"] = "Soc Sci Inf Stud",
  ["social science japan journal"] = "Soc Sci Jpn J",
  ["social science news letter"] = "Soc Sci New Lett",
  ["social science quarterly"] = "Soc Sci Q",
  ["social science research"] = "Soc Sci Res",
  ["social sciences"] = "Soc Sci",
  ["social sciences (basel, switzerland)"] = "Soc Sci (Basel)",
  ["social sciences in china"] = "Soc Sci China",
  ["social sciences information. information sur les sciences sociales"] = "Soc Sci Inf (Paris)",
  ["social scientist"] = "Soc Sci (New Delhi)",
  ["social security bulletin"] = "Soc Secur Bull",
  ["social security bulletin. annual statistical supplement"] = "Soc Secur Bull Annu Stat Suppl",
  ["social security bulletin. annual statistical supplement / u.s. dept. of health, education, and welfare, social security administration"] = "Soc Secur Bull Annu Stat Suppl",
  ["social security bulletin. annual statistical supplement."] = "Soc. Secur. Bull. Annu. Stat. Suppl.",
  ["social service review"] = "Soc. Serv. Rev.",
  ["social studies (maynooth, ireland)"] = "Soc Stud (Maynooth)",
  ["social studies (philadelphia, pa. : 1953)"] = "Soc Stud",
  ["social studies of science"] = "Soc Stud Sci",
  ["social theory & health : sth"] = "Soc Theory Health",
  ["social theory and practice"] = "Soc Theory Pract",
  ["social thought"] = "Soc Thought",
  ["social welfare"] = "Soc Welfare (India)",
  ["social work"] = "Soc Work",
  ["social work & social sciences review"] = "Soc Work Soc Sci Rev",
  ["social work & society"] = "Soc Work Soc",
  ["social work education and development newsletter"] = "Soc Work Educ Dev Newsl",
  ["social work forum (new york, n.y.)"] = "Soc Work Forum (N Y N Y)",
  ["social work in education"] = "Soc Work Educ",
  ["social work in health care"] = "Soc Work Health Care",
  ["social work in mental health"] = "Soc Work Ment Health",
  ["social work in public health"] = "Soc Work Public Health",
  ["social work research"] = "Soc Work Res",
  ["social work research & abstracts"] = "Soc Work Res Abstr",
  ["social work research and abstracts"] = "Soc. Work Res. Abstr.",
  ["social work with groups"] = "Soc Work Groups",
  ["social, cultural, and behavioral modeling : 9th international conference, sbp-brims 2016, washington, dc, usa, june 28-july 1, 2016, proceedings. international conference on social computing, behavioral modeling and prediction (9th : 2016 : washington, d.c.)"] = "Soc Cult Behav Model (2016)",
  ["social-medicinsk tidskrift"] = "Soc Med Tidskr",
  ["sociale wetenschappen"] = "Soc Wetenschap",
  ["socialist review"] = "Social Rev",
  ["socialist revolution"] = "Social Revolut",
  ["socialmedicinsk tidskrift. skriftserie"] = "Socialmed Tidskr Skriftser",
  ["sociedad matemática mexicana"] = "Bol. Soc. Mat. Mexicana (3)",
  ["sociedade brasiliera de matemática applicada e computacional"] = "Mat. Apl. Comput.",
  ["sociedade paranaense de matemática"] = "Bol. Soc. Parana. Mat. (2)",
  ["societa e storia"] = "Soc Stor",
  ["societa nazionale di scienze, lettere e arti in napoli. rendiconto dell’accademia delle scienze fisiche e matematiche. serie iv"] = "Rend. Accad. Sci. Fis. Mat. Napoli (4)",
  ["societate şi politică"] = "Soc Politica",
  ["società e storia"] = "Soc Stor",
  ["societes d’afrique & sida : newsletter of the societes d’afrique & sida network"] = "Soc Afr SIDA",
  ["societies (basel, switzerland)"] = "Societies (Basel)",
  ["society & animals"] = "Soc. Anim.",
  ["society & animals : social scientific studies of the human experience of other animals"] = "Soc Anim",
  ["society & natural resources"] = "Soc Nat Resour",
  ["society and culture"] = "Soc Cult",
  ["society and mental health"] = "Soc Ment Health",
  ["society and natural resources"] = "Soc. Nat. Resour.",
  ["society for ancient medicine review"] = "Soc Anc Med Rev",
  ["society for applied bacteriology symposium series"] = "Soc Appl Bacteriol Symp Ser",
  ["society for nursing history gazette"] = "Soc. Nurs. Hist. Gaz.",
  ["society of general physiologists series"] = "Soc Gen Physiol Ser",
  ["society of photographic instrumentation engineers journal"] = "Soc. Photogr. Instrum. Eng. J.",
  ["society of reproduction and fertility supplement"] = "Soc Reprod Fertil Suppl",
  ["società italiana di ecologia, atti"] = "Soc. Ital. Ecol., Atti",
  ["società italiana di fisica"] = "Nuovo Cimento A (11)",
  ["società nazionale di scienze, lettere e arti in napoli"] = "Rend. Accad. Sci. Fis. Mat. Napoli (4)",
  ["sociétés d'afrique & sida : newsletter of the sociétés d'afrique & sida network"] = "Soc Afr SIDA",
  ["socio-ecological practice research"] = "Socioeco Prac Res",
  ["socio-economic planning sciences"] = "Socioecon Plann Sci",
  ["socio-economic review"] = "Socioecon Rev",
  ["socioaffective neuroscience & psychology"] = "Socioaffect Neurosci Psychol",
  ["socioeconomic issues of health"] = "Socioecon Issues Health",
  ["sociologia (lisbon, portugal)"] = "Sociologia",
  ["sociologia del diritto"] = "Sociol Diritto",
  ["sociologia del lavoro"] = "Sociol Lav",
  ["sociologia neerlandica"] = "Sociol Neerl",
  ["sociologia ruralis"] = "Sociol Ruralis",
  ["sociologia urbana e rurale"] = "Sociol Urbana Rurale (Testo Stamp)",
  ["sociological bulletin"] = "Sociol Bull",
  ["sociological focus"] = "Sociol Focus",
  ["sociological forum"] = "Sociol. Forum",
  ["sociological forum (randolph, n.j.)"] = "Sociol Forum (Randolph N J)",
  ["sociological inquiry"] = "Sociol Inq",
  ["sociological methodology"] = "Sociol Methodol",
  ["sociological methods & research"] = "Sociol Methods Res",
  ["sociological perspectives"] = "Sociol. Perspect.",
  ["sociological perspectives : sp : official publication of the pacific sociological association"] = "Sociol Perspect",
  ["sociological quarterly"] = "Sociol. Q.",
  ["sociological research"] = "Sociol Res",
  ["sociological research online"] = "Sociol Res Online",
  ["sociological review"] = "Sociol. Rev.",
  ["sociological review mongraph"] = "Sociol Rev [Monogr]",
  ["sociological review monograph"] = "Sociol Rev Monogr",
  ["sociological science"] = "Sociol Sci",
  ["sociological spectrum"] = "Sociol. Spectr.",
  ["sociological spectrum : the official journal of the mid-south sociological association"] = "Sociol Spectr",
  ["sociological studies of children and youth"] = "Sociol Stud Child Youth",
  ["sociological symposium"] = "Sociol Symp",
  ["sociological theory"] = "Sociol Theory",
  ["sociological theory and methods"] = "Sociol. Theor. Methods",
  ["sociologicky casopis"] = "Sociol Cas",
  ["sociologicky casopis-czech sociological review"] = "Sociol. Cas.",
  ["sociologický časopis"] = "Sociol Cas",
  ["sociologie du travail"] = "Sociol Trav",
  ["sociologie et societes"] = "Sociol Soc",
  ["sociologie et sociétés"] = "Sociol Soc",
  ["sociologie romaneasca"] = "Sociol Rom",
  ["sociologie românească"] = "Sociol Rom",
  ["sociologija sela"] = "Sociol Sela",
  ["sociologische gids"] = "Sociol Gids",
  ["sociologisk forskning"] = "Sociol. Forsk.",
  ["sociologus; zeitschrift fur empirische soziologie, sozialpsychologische und ethnologische forschung. a journal for empirical sociology, social psychology, and ethnic research"] = "Sociologus",
  ["sociologus; zeitschrift für empirische soziologie, sozialpsychologische und ethnologische forschung. a journal for empirical sociology, social psychology, and ethnic research"] = "Sociologus",
  ["sociology and anthropology (alhambra, calif.)"] = "Sociol Anthropol (Alhambra)",
  ["sociology and social research"] = "Sociol Soc Res",
  ["sociology compass"] = "Sociol Compass",
  ["sociology of development (oakland, calif.)"] = "Sociol Dev (Oakl)",
  ["sociology of education"] = "Sociol Educ",
  ["sociology of health & illness"] = "Sociol Health Illn",
  ["sociology of health and illness"] = "Sociol. Health Illn.",
  ["sociology of race and ethnicity (thousand oaks, calif.)"] = "Sociol Race Ethn (Thousand Oaks)",
  ["sociology of religion"] = "Sociol Relig",
  ["sociology of sport journal"] = "Sociol Sport J",
  ["sociology of the sciences. yearbook"] = "Sociol. Sci. Yearbook",
  ["sociology of work and occupations"] = "Sociol Work Occup",
  ["sociology-the journal of the british sociological association"] = "Sociol.-J. Brit. Sociol. Assoc.",
  ["socius : sociological research for a dynamic world"] = "Socius",
  ["société vaudoise des sciences naturelles"] = "Bull. Soc. Vaud. Sci. Nat.",
  ["socra source"] = "SOCRA Source",
  ["sodininkyste ir darzininkyste"] = "Sodinink Darzinink",
  ["soft and biological matter"] = "Soft Biol. Matter",
  ["soft computing"] = "Soft comput",
  ["soft computing letters"] = "Soft Comput. Lett.",
  ["soft materials"] = "Soft Mater.",
  ["soft matter"] = "Soft Matter",
  ["soft robotics"] = "Soft Robot",
  ["software and systems modeling"] = "Softw Syst Model",
  ["software engineering in healthcare systems (sehs), ieee/acm international workshop on"] = "Softw Eng Healthc Syst SEHS IEEE ACM Int Workshop",
  ["software impacts"] = "Software Impacts",
  ["software in healthcare"] = "Softw Healthc",
  ["software law journal"] = "Softw Law J",
  ["software quality journal"] = "Software Qual. J.",
  ["software testing, verification and reliability"] = "Softw. Test. Verif. Reliab.",
  ["software, environments, and tools"] = "Software Environ. Tools",
  ["software: practice & experience"] = "Softw Pract Exp",
  ["software: practice and experience"] = "Softw.: Pract. Exper.",
  ["sogo igaku. medicine"] = "Sogo Igaku",
  ["sogo kango (comprehensive nursing quarterly)"] = "Sogo Kango",
  ["sogo kango. comprehensive nursing, quarterly"] = "Sogo Kango",
  ["soil & sediment contamination"] = "Soil Sediment Contam",
  ["soil & tillage research"] = "Soil Tillage Res",
  ["soil and sediment contamination"] = "Soil Sediment Contam.",
  ["soil and tillage research"] = "Soil Tillage Res.",
  ["soil and water research"] = "Soil Water Res.",
  ["soil biology & biochemistry"] = "Soil Biol Biochem",
  ["soil biology and biochemistry"] = "Soil Biol. Biochem.",
  ["soil discussions"] = "Soil Discuss.",
  ["soil dynamics and earthquake engineering"] = "Soil Dyn. Earthquake Eng.",
  ["soil ecology letters"] = "Soil Ecol. Lett.",
  ["soil horizons"] = "Soil Horiz.",
  ["soil mechanics and foundation engineering"] = "Soil Mech. Found. Eng.",
  ["soil organisms"] = "Soil Org",
  ["soil research"] = "Soil Res.",
  ["soil research (collingwood, vic.)"] = "Soil Res",
  ["soil science"] = "Soil Sci",
  ["soil science & plant nutrition"] = "Soil Sci. Plant Nutr.",
  ["soil science and plant nutrition"] = "Soil Sci Plant Nutr",
  ["soil science society of america journal"] = "Soil Sci. Soc. Am. J.",
  ["soil science society of america journal. soil science society of america"] = "Soil Sci Soc Am J",
  ["soil security"] = "Soil Secur.",
  ["soil systems"] = "Soil Syst",
  ["soil technology"] = "Soil Technol.",
  ["soil use and management"] = "Soil Use Manag",
  ["soils and foundations"] = "Soils Found.",
  ["soins. cardiologie"] = "Soins Cardiol",
  ["soins. chirurgie"] = "Soins Chir",
  ["soins. chirurgie (paris, france : 1982)"] = "Soins Chir",
  ["soins. chirurgie generale et specialisee"] = "Soins. Chir. Gen. Spec.",
  ["soins. chirurgie generale et specialisée"] = "Soins Chir Gen Spec",
  ["soins. formation, pedagogie, encadrement"] = "Soins. Form. Pedagog. Encadr.",
  ["soins. formation, pedagogie, encadrement : avec la participation du ceeiec"] = "Soins Form Pedagog Encadr",
  ["soins. formation, pédagogie, encadrement : avec la participation du ceeiec"] = "Soins Form Pedagog Encadr",
  ["soins. gerontologie"] = "Soins. Gerontol.",
  ["soins. gérontologie"] = "Soins Gerontol",
  ["soins. gynecologie, obstetrique, puericulture"] = "Soins. Gynecol. Obstet. Pueric.",
  ["soins. gynecologie, obstetrique, puericulture, pediatrie"] = "Soins. Gynecol. Obstet. Pueric. Pediatr.",
  ["soins. gynécologie, obstétrique, puériculture"] = "Soins Gynecol Obstet Pueric",
  ["soins. gynécologie, obstétrique, puériculture, pédiatrie"] = "Soins Gynecol Obstet Pueric Pediatr",
  ["soins. pathologie tropicale"] = "Soins Pathol Trop",
  ["soins. pediatrie, puericulture"] = "Soins. Pediatr. Pueric.",
  ["soins. pédiatrie, puériculture"] = "Soins Pediatr Pueric",
  ["soins. psychiatrie"] = "Soins Psychiatr",
  ["soins; la revue de reference infirmiere"] = "Soins",
  ["soins; la revue de référence infirmière"] = "Soins",
  ["soj anesthesiology & pain management"] = "SOJ Anesthesiol Pain Manag",
  ["soj biochemistry"] = "SOJ Biochem",
  ["soj immunology"] = "SOJ Immunol",
  ["soj microbiology & infectious diseases"] = "SOJ Microbiol Infect Dis",
  ["soj neurology"] = "SOJ Neurol",
  ["soj nursing & health care"] = "SOJ Nurs Health Care",
  ["soj pharmacy & pharmaceutical sciences"] = "SOJ Pharm Pharm Sci",
  ["soj psychology"] = "SOJ Psychol",
  ["soj surgery"] = "SOJ Surg",
  ["soj urology and nephrology open access"] = "SOJ Urol Nephrol Open Access",
  ["sojourn (singapore)"] = "Sojourn",
  ["sokoto journal of medical laboratory science"] = "Sokoto J Med Lab Sci",
  ["solaiat (sociedad odontologica latino-americana de implantes aloplasticos y transplantes)"] = "SOLAIAT",
  ["solaiat; sociedad odontologica latino-americana de implantes aloplasticos y transplantes"] = "SOLAIAT",
  ["solar & wind technology"] = "Solar Wind Technol.",
  ["solar cells"] = "Solar Cells",
  ["solar compass"] = "Sol. Compass",
  ["solar energy"] = "Sol. Energy",
  ["solar energy (phoenix, ariz.)"] = "Sol Energy",
  ["solar energy advances"] = "Sol. Energy Adv.",
  ["solar energy materials"] = "Sol. Energy Mater.",
  ["solar energy materials & solar cells"] = "Sol. Energy Mater. Sol. Cells",
  ["solar energy materials amd solar cells"] = "Sol. Energy Mater. Sol. Cells.",
  ["solar energy materials and solar cells"] = "Sol. Energy Mater. Sol. Cells",
  ["solar energy materials and solar cells : an international journal devoted to photovoltaic, photothermal, and photochemical solar energy conversion"] = "Sol Energy Mater Sol Cells",
  ["solar physics"] = "Sol. Phys.",
  ["solar rrl"] = "Sol. RRL",
  ["solar system research"] = "Sol. Syst. Res.",
  ["soldering & surface mount technology"] = "Soldering Surf. Mount Technol.",
  ["solicitors' journal (london, england : 1928)"] = "Solicit J",
  ["solicitors’ journal (london, england : 1928)"] = "Solicit J",
  ["solid earth"] = "Solid Earth",
  ["solid earth discussions"] = "Solid Earth Discuss.",
  ["solid earth science"] = "Solid Earth Sci.",
  ["solid earth sciences"] = "Solid Earth Sci.",
  ["solid freeform fabrication symposium proceedings. solid freeform fabrication symposium"] = "Solid Freeform Fabr Symp Proc",
  ["solid fuel chemistry"] = "Solid Fuel Chem.",
  ["solid mechanics and its applications"] = "Solid Mech. Appl.",
  ["solid state communications"] = "Solid State Commun.",
  ["solid state electronics"] = "Solid-State Electron.",
  ["solid state electronics letters"] = "Solid State Electron. Lett.",
  ["solid state ionics"] = "Solid State Ionics",
  ["solid state nuclear magnetic resonance"] = "Solid State Nucl Magn Reson",
  ["solid state sciences"] = "Solid State Sci.",
  ["solid-state electronics"] = "Solid-State Electron.",
  ["solid–state electronics"] = "Solid–State Electron.",
  ["solubility data series"] = "Solubility Data Ser.",
  ["solutions (burlington, vt.)"] = "Solutions (Burlingt)",
  ["solvent extraction and ion exchange"] = "Solvent Extr. Ion Exch.",
  ["solvent extraction research and development, japan"] = "Solvent Extr. Res. Dev., Jpn.",
  ["somatic cell and molecular genetics"] = "Somat Cell Mol Genet",
  ["somatic cell genetics"] = "Somatic Cell Genet",
  ["somatosensory & motor research"] = "Somatosens Mot Res",
  ["somatosensory and motor research"] = "Somatosens. Mot. Res.",
  ["somatosensory research"] = "Somatosens Res",
  ["somnologie : schlafforschung und schlafmedizin = somnology : sleep research and sleep medicine"] = "Somnologie (Berl)",
  ["sonderband der zeitschrift strahlentherapie und onkologie"] = "Sonderb. Z. Strahlenther. Onkol.",
  ["sonderbande zur strahlentherapie und onkologie"] = "Sonderb. Strahlenther. Onkol.",
  ["sonderbände zur strahlentherapie und onkologie"] = "Sonderb Strahlenther Onkol",
  ["sonderdruck aus schriften der sudetendeutschen akademie der wissenschaften und küntse"] = "Sonderdr. Schr. Sudet.dtsch. Akad. Wiss. Künste Forsch.beitr. Nat.wiss. Kl.",
  ["sonderschriften des deutschen archäologischen instituts, abteilung kairo"] = "SDAIK",
  ["soobseniâ gosudarstvennogo èrmitaza (reports of the hermitage museum)"] = "SGE",
  ["soobshcheniia akademii nauk gruzinskoi ssr (1963)"] = "Soobshch Akad Nauk Gruz Ssr",
  ["soobshchenii︠a︡ akademii nauk gruzinskoĭ ssr (1963)"] = "Soobshch Akad Nauk Gruz Ssr",
  ["soobshcheniya akademii nauk gruzii"] = "Soobshch. Akad. Nauk Gruz. SSR",
  ["soobščenija gosudarstvennogo ėrmitaža"] = "SoobErmit",
  ["soobščenija gosudarstvennogo muzeja izobrazitel’nych isskustv imeni a. s. puškina"] = "SoobMuzMoskva",
  ["soochow journal of mathematics"] = "Soochow J. Math.",
  ["sophia studies in cross-cultural philosophy of traditions and cultures"] = "Sophia Stud. Cross-Cult. Philos. Tradit. Cult.",
  ["sort (barcelona)"] = "Sort (Barc)",
  ["sort. statistics and operations research transactions"] = "SORT",
  ["sosiaalilaaketieteellinen aikakauslehti"] = "Sos Laaketiet Aikak",
  ["sosiaalilääketieteellinen aikakauslehti"] = "Sos Laaketiet Aikak",
  ["sotahistoriallinen aikakauskirja"] = "Sotahist Aikak",
  ["sotilaslaaketieteellinen aikakauslehti"] = "Sotilaslaak. Aikak.",
  ["sotilaslääketieteellinen aikakauslehti"] = "Sotilaslaak Aikak",
  ["sotsial'naia gigiena, organizatsiia zdravookhraneniia i istoriia meditsiny"] = "Sots Gig Organ Zdravookhr Istor Med",
  ["sotsiologicheskie issledovaniia"] = "Sotsiol Issled",
  ["sotsiologicheskie issledovaniya"] = "Sotsiologicheskie Issled.",
  ["sot︠s︡iologicheskie issledovanii︠a︡"] = "Sotsiol Issled",
  ["soudni lekarstvi"] = "Soud. Lek.",
  ["soudni lekarstvi / casopis sekce soudniho lekarstvi cs. lekarske spolecnosti j. ev. purkyne"] = "Soud Lek",
  ["soudní lékarství"] = "Soud Lek",
  ["souls : a critical journal of black politics, culture, and society"] = "Souls",
  ["sound & vibration"] = "Sound Vib.",
  ["source (new york, n.y. : 1981)"] = "Source Notes Hist Art",
  ["source (united nations development programme)"] = "Source",
  ["source code for biology and medicine"] = "Source Code Biol Med",
  ["source journal of genomics"] = "Source J Genom",
  ["source journal of obstetrics and gynaecology"] = "Source J Obstet Gynaecol",
  ["sources and studies in the history of mathematics and physical sciences"] = "Sources Stud. Hist. Math. Phys. Sci.",
  ["sources in the history of mathematics and physical sciences"] = "Sources Hist. Math. Phys. Sci.",
  ["sources: travaux historiques"] = "STHist",
  ["south african archaeological bulletin"] = "S. Afr. Archaeol. Bull.",
  ["south african cancer bulletin = suid-afrikaanse kankerbulletin"] = "S Afr Cancer Bull",
  ["south african computer journal"] = "S. Afr. Comput. J.",
  ["south african dental journal. suid afrikaanse tandarts tydskrif"] = "S Afr Dent J",
  ["south african family practice : official journal of the south african academy of family practice/primary care"] = "S Afr Fam Pract (2004)",
  ["south african geographical journal"] = "S. Afr. Geogr. J.",
  ["south african health review"] = "S Afr Health Rev",
  ["south african humanities"] = "S. Afr. Humanit.",
  ["south african journal for research in sport physical education and recreation"] = "S. Afr. J. Res. Sport Phys. Educ. Recreat.",
  ["south african journal of agricultural extension"] = "S. Afr. J. Agric. Ext.",
  ["south african journal of animal science"] = "S Afr J Anim Sci",
  ["south african journal of bioethics and law"] = "S Afr J Bioeth Law",
  ["south african journal of botany"] = "S. Afr. J. Bot.",
  ["south african journal of botany : official journal of the south african association of botanists"] = "Suid-Afrikaanse tydskrif vir plantkunde : amptelike tydskrif van die Suid-Afrikaanse Genootskap van Plantkundiges|S. Afr. J. Bot.",
  ["south african journal of botany : official journal of the south african association of botanists = suid-afrikaanse tydskrif vir plantkunde : amptelike tydskrif van die suid-afrikaanse genootskap van plantkundiges"] = "S Afr J Bot",
  ["south african journal of chemical engineering"] = "S. Afr. J. Chem. Eng.",
  ["south african journal of chemistry"] = "S. Afr. J. Chem.",
  ["south african journal of chemistry. suid-afrikaanse tydskrif vir chemie"] = "S Afr J Chem",
  ["south african journal of clinical nutrition"] = "S. Afr. J. Clin. Nutr.",
  ["south african journal of clinical science"] = "S. Afr. J. Clin. Sci.",
  ["south african journal of clinical science. suid-afrikaanse tydskrif vir kliniese wetenskap"] = "S Afr J Clin Sci",
  ["south african journal of communication disorders"] = "S. Afr. J. Commun. Disord.",
  ["south african journal of communication disorders. die suid-afrikaanse tydskrif vir kommunikasieafwykings"] = "S Afr J Commun Disord",
  ["south african journal of economic and management sciences"] = "South African J. Econ. Manage. Sci.",
  ["south african journal of economic and management sciences, n.s."] = "South African J. Econ. Manage. Sci.",
  ["south african journal of economics"] = "South African J. Econ.",
  ["south african journal of education"] = "S Afr J Educ",
  ["south african journal of enology and viticulture"] = "S. Afr. J. Enol. Vitic.",
  ["south african journal of geology"] = "S. Afr. J. Geol.",
  ["south african journal of geology : being the transactions of the geological society of south africa = suid-afrikaanse tydskrif vir geologie : syndie die verhandelinge van die geologiese vereniging van suid-afrika"] = "South Afr J Geol",
  ["south african journal of geomatics"] = "S. Afr. J. Geomatics",
  ["south african journal of industrial engineering"] = "S. Afr. J. Ind. Eng.",
  ["south african journal of industrial psychology"] = "S. Afr. J. Ind. Psychol.",
  ["south african journal of information management"] = "S Afr J Inf Manag",
  ["south african journal of laboratory and clinical medicine"] = "S. Afr. J. Lab. Clin. Med.",
  ["south african journal of laboratory and clinical medicine. suid-afrikaanse tydskrif vir laboratorium- en kliniekwerk"] = "S Afr J Lab Clin Med",
  ["south african journal of marine science"] = "S. Afr. J. Mar. Sci.",
  ["south african journal of medical sciences"] = "S. Afr. J. Med. Sci.",
  ["south african journal of obstetrics and gynaecology"] = "S Afr J Obstet Gynaecol",
  ["south african journal of obstetrics and gynaecology (1999)"] = "S Afr J Obstet Gynaecol (1999)",
  ["south african journal of plant and soil"] = "S. Afr. J. Plant Soil",
  ["south african journal of plant and soil = suid-afrikaanse tydskrif vir plant en grond"] = "S Afr J Plant Soil",
  ["south african journal of psychology = suid-afrikaanse tydskrif vir sielkunde"] = "S Afr J Psychol",
  ["south african journal of radiology"] = "S. Afr. J. Radiol.",
  ["south african journal of science"] = "S Afr J Sci",
  ["south african journal of sports medicine"] = "S. Afr. J. Sports Med.",
  ["south african journal of surgery"] = "S. Afr. J. Surg.",
  ["south african journal of surgery. suid-afrikaanse tydskrif vir chirurgie"] = "S Afr J Surg",
  ["south african journal of wildlife research"] = "S. Afr. J. Wildl. Res.",
  ["south african journal on human rights"] = "S Afr J Hum Rights",
  ["south african labour bulletin"] = "S Afr Labour Bull",
  ["south african law journal"] = "S Afr Law J",
  ["south african law reports. die suid-afrikaanse hofverslae. beslissings van: die hooggeregshof van suid-afrika, die suidwes-afrika afdeling, die hoërhof van rhodesië. decisions of: the supreme court of south africa, the south-west africa division, the high court of rhodesia"] = "S Afr Law Rep",
  ["south african medical journal"] = "S. Afr. Med. J.",
  ["south african medical journal = suid-afrikaanse tydskrif vir geneeskunde"] = "S Afr Med J",
  ["south african nursing journal. suid-afrikaanse verplegingstydskrif"] = "S Afr Nursing J",
  ["south african pharmaceutical journal. suid-afrikaanse tydskrif vir apteekwese"] = "S Afr Pharm J",
  ["south african respiratory journal"] = "S Afr Respir J",
  ["south african sociological review"] = "S Afr Sociol Rev",
  ["south african statistical journal"] = "S. Afr. Stat. J.",
  ["south african statistical journal = suid-afrikaanse statistiese tydskrif"] = "S Afr Stat J",
  ["south african statistical journal. suid-afrikaanse statistiese tydskrif"] = "South African Statist. J.",
  ["south american indian studies"] = "South Am Indian Stud",
  ["south american journal of herpetology"] = "South Am J Herpetol",
  ["south american journal of logic"] = "South Amer. J. Log.",
  ["south asia (nedlands, w.a.)"] = "South Asia",
  ["south asia bulletin"] = "South Asia Bull",
  ["south asia research"] = "South Asia Res",
  ["south asian history and culture"] = "South Asian Hist Cult",
  ["south asian journal of cancer"] = "South Asian J Cancer",
  ["south asian review"] = "South Asian Rev",
  ["south carolina dental journal"] = "S C Dent J",
  ["south carolina historical magazine"] = "S C Hist Mag",
  ["south carolina history illustrated"] = "S C Hist Illus",
  ["south carolina law review"] = "S C Law Rev",
  ["south carolina nurse"] = "S. C. Nurse",
  ["south carolina nurse (columbia, s.c. : 1994)"] = "S C Nurse (1994)",
  ["south carolina nursing"] = "S C Nurs",
  ["south dakota farm & home research"] = "S D Farm Home Res",
  ["south dakota historical collections"] = "S D Hist Collect",
  ["south dakota journal of medicine"] = "S D J Med",
  ["south dakota law review"] = "S D Law Rev",
  ["south dakota medicine"] = "S. D. Med.",
  ["south dakota medicine : the journal of the south dakota state medical association"] = "S D Med",
  ["south dakota nurse"] = "S. D. Nurse",
  ["south east asia research"] = "South East Asia Res",
  ["south east asian journal of mathematics and mathematical sciences"] = "South East Asian J. Math. Math. Sci.",
  ["south eastern european journal of public health"] = "South East Eur J Public Health",
  ["south eastern reporter. second series"] = "South East Report Second Ser",
  ["south european society & politics"] = "South Eur Soc Polit",
  ["south pacific bulletin"] = "South Pac Bull",
  ["south pacific journal of natural and applied sciences"] = "South Pac. J. Nat. Appl. Sci.",
  ["south texas law journal"] = "South Tex Law J",
  ["south texas law review"] = "South Tex Law Rev",
  ["southampton medical journal"] = "Southampt Med J",
  ["southeast asian bulletin of mathematics"] = "Southeast Asian Bull. Math.",
  ["southeast asian journal of sciences"] = "Southeast Asian J Sci",
  ["southeast asian journal of social science"] = "Southeast Asian J Soc Sci",
  ["southeast asian journal of surgery"] = "Southeast Asian J Surg",
  ["southeast asian journal of tropical medicine and public health"] = "Southeast Asian J. Trop. Med. Public Health",
  ["southeast europe series"] = "Southeast Eur Ser",
  ["southeastern archaeology"] = "Southeast. Archaeol.",
  ["southeastern europe. l'europe du sud-est"] = "Southeast Eur",
  ["southeastern europe. l’europe du sud-est"] = "Southeast Eur",
  ["southeastern geographer"] = "Southeast Geogr",
  ["southeastern geology"] = "Southeast. Geol.",
  ["southeastern naturalist"] = "Southeast. Nat.",
  ["southeastern naturalist (steuben, me.)"] = "Southeast Nat",
  ["southern africa political & economic monthly"] = "South Afr Polit Econ Mon",
  ["southern african institute of mining and metallurgy"] = "SAIMM",
  ["southern african journal of demography = suidelike afrikaanse tydskrif vir demografie"] = "South Afr J Demogr",
  ["southern african journal of hiv medicine"] = "South Afr J HIV Med",
  ["southern african journal of infectious diseases"] = "S Afr J Infect Dis",
  ["southern california interdisciplinary law journal"] = "South Calif Interdiscip Law J",
  ["southern california law review"] = "South Calif Law Rev",
  ["southern california quarterly"] = "South Calif Q",
  ["southern california review of law and women's studies"] = "South Calif Rev Law Womens Stud",
  ["southern california review of law and women’s studies"] = "South Calif Rev Law Womens Stud",
  ["southern clinics of istanbul eurasia"] = "South Clin Istanb Eurasia",
  ["southern communication journal"] = "South Commun J",
  ["southern economic journal"] = "South Econ J",
  ["southern exposure"] = "South Expos",
  ["southern folklore quarterly"] = "South Folk Q",
  ["southern forests"] = "South For",
  ["southern forests: a journal of forest science"] = "South. For.: J. For. Sci.",
  ["southern history"] = "South Hist",
  ["southern hospitals"] = "South Hosp",
  ["southern illinois university law journal. southern illinois university at carbondale. school of law"] = "South Ill Univ Law J",
  ["southern indian studies"] = "South Indian Stud",
  ["southern journal of agricultural economics"] = "Southern J. Agr. Econ.",
  ["southern journal of applied forestry"] = "South. J. Appl. For.",
  ["southern journal of optometry"] = "South J Optom",
  ["southern med review"] = "South Med Rev",
  ["southern medical journal"] = "South Med J",
  ["southern medicine"] = "South Med",
  ["southern medicine and surgery"] = "South Med Surg",
  ["southern online journal of nursing research"] = "South Online J Nurs Res",
  ["southern quarterly"] = "Southern Quart.",
  ["southern reporter. second series. cases argued and determined in the courts of alabama, florida, louisiana, mississippi"] = "South Report Second Ser Cases Argued Determ Courts Ala Fla La Miss",
  ["southern studies"] = "South Stud",
  ["southern university law review. southern university and a & m college. school of law"] = "South Univ Law Rev",
  ["southwest journal of pulmonary & critical care"] = "Southwest J Pulm Crit Care",
  ["southwest journal of pure and applied mathematics"] = "Southwest J. Pure Appl. Math.",
  ["southwest review"] = "Southwest Rev",
  ["southwestern entomologist"] = "Southwest. Entomolog.",
  ["southwestern journal of anthropology"] = "Southwest J Anthropol",
  ["southwestern law journal"] = "Southwest Law J",
  ["southwestern medicine"] = "Southwest Med",
  ["southwestern naturalist"] = "Southwest. Nat.",
  ["southwestern university law review"] = "Southwest Univ Law Rev",
  ["sovetskaia arkhitektura (kommunisticheskaia akademiia (soviet union). sektsiia sotsrasselenia i zhilishchno-bytogo stroitel’stvo)"] = "Sov Arkh",
  ["sovetskaia etnografiia / akademiia nauk sssr i narodnyi komissariat prosveshcheniia rsfsr"] = "Sov Etnogr",
  ["sovetskaia meditsina"] = "Sov Med",
  ["sovetskai︠a︡ arkhitektura (kommunisticheskai︠a︡ akademii︠a︡ (soviet union). sekt︠s︡ii︠a︡ sot︠s︡rasseleni︠a︡ i zhilishchno-bytogo stroitelʹstvo)"] = "Sov Arkh",
  ["sovetskai︠a︡ ėtnografii︠a︡"] = "Sov Etnogr",
  ["sovetskaja archeologija"] = "SovA",
  ["sovetskaja arheologia"] = "SA",
  ["sovetskii vrachebnyi sbornik"] = "Sov Varchebnyii Sb",
  ["sovetskiĭ vrachebnyĭ sbornik"] = "Sov Varchebnyii Sb",
  ["sovetskoe gosudarstvo i pravo"] = "Sov Gos Pravo",
  ["sovetskoe slavi︠a︡novedenie (moscow, russia : 1965)"] = "Sov Slavanovednie",
  ["sovetskoe zdravookhranenie"] = "Sov Zdravookhr",
  ["sovetskoe zdravookhranenie / ministerstvo zdravookhraneniia sssr"] = "Sov Zdravookhr",
  ["sovetskoe zdravookhranenie kirgizii"] = "Sov Zdravookhr Kirg",
  ["soviet and eastern european foreign trade"] = "Soviet E. Europ. Foreign Trade",
  ["soviet anthropology and archeology"] = "Sov Anthropol Archeol",
  ["soviet astronomy [translation of astronomicheskii zhurnal]"] = "Sov. Astron.",
  ["soviet astronomy letters [translation of pisma v astronomicheskii zhurnal]"] = "Soc. Astron. Lett.",
  ["soviet economy"] = "Soviet Economy",
  ["soviet economy (silver spring, md.)"] = "Sov Econ",
  ["soviet education"] = "Sov Educ",
  ["soviet genetics"] = "Sov Genet",
  ["soviet geography"] = "Sov Geogr",
  ["soviet jewish affairs"] = "Sov Jew Aff",
  ["soviet jewish affairs / ija, institute of jewish affairs, in association with the world jewish congress"] = "Sov Jew Aff",
  ["soviet journal of atomic energy [translation of atomnaya energiya]"] = "Sov. J. At. Energy",
  ["soviet journal of bioorganic chemistry"] = "Sov J Bioorg Chem",
  ["soviet journal of developmental biology"] = "Sov. J. Dev. Biol.",
  ["soviet journal of ecology"] = "Sov. J. Ecol.",
  ["soviet journal of low temperature physics [translation of fizika nizkikh temperatur]"] = "Sov. J. Low Temp. Phys.",
  ["soviet journal of nuclear physics [translation of yadernaya fizika]"] = "Sov. J. Nucl. Phys.",
  ["soviet journal of optical technology [translation of optiko–mekhanicheskaya promyshlennost]"] = "Sov. J. Opt. Technol.",
  ["soviet journal of particles and nuclei [translation of fizika elementarnykh chastitsi atomnogo yadra]"] = "Sov. J. Part. Nucl.",
  ["soviet journal of plasma physics [translation of fizika plazmy]"] = "Sov. J. Plasma Phys.",
  ["soviet journal of quantum electronics"] = "Sov. J. Quantum Electron.",
  ["soviet journal of quantum electronics [translation of kvantovaya elektronika (moscow)]"] = "Sov. J. Quantum Electron.",
  ["soviet law and government"] = "Sov Law Gov",
  ["soviet physics journal [translation of izvestiya vysshikh uchebnykh zavedenii, fizika]"] = "Sov. Phys. J.",
  ["soviet physics – solid state"] = "Sov. Phys. - Solid State",
  ["soviet physics, crystallography"] = "Sov Phys Crystallogr",
  ["soviet physics–acoustics [translation of akusticheskii zhurnal]"] = "Sov. Phys. Acoust.",
  ["soviet physics–crystallography [translation of kristallografiya]"] = "Sov. Phys. Crystallogr.",
  ["soviet physics–doklady [translation of doklady akademii nauk sssr]"] = "Sov. Phys. Dokl.",
  ["soviet physics–jetp [translation of zhurnal eksperimentalnoi i teoreticheskoi fiziki]"] = "Sov. Phys. JETP",
  ["soviet physics–semiconductors [translation of fizika i tekhnika poluprovodnikov]"] = "Sov. Phys. Semicond.",
  ["soviet physics–solid state [translation of fizika tverdogo tela (leningrad)]"] = "Sov. Phys. Solid State",
  ["soviet physics–technical physics [translation of zhurnal tekhnicheskoi fiziki]"] = "Sov. Phys. Tech. Phys.",
  ["soviet physics–uspekhi [translation of uspekhi fizicheskikh nauk]"] = "Sov. Phys. Usp.",
  ["soviet radiophysics [translation of izvestiya vysshikh uchebnykh zavedenii, radiofizika]"] = "Sov. Radiophys.",
  ["soviet scientific reviews/section a: physics reviews"] = "Soviet Sci. Rev. Sect. A Phys. Rev.",
  ["soviet sociology"] = "Sov Soc",
  ["soviet studies"] = "Sov Stud",
  ["soviet studies in history"] = "Sov Stud Hist",
  ["soviet studies in philosophy"] = "Sov Stud Philos",
  ["soviet technical physics letters [translation of pisma v zhurnal tekhnicheskoi fiziki]"] = "Sov. Tech. Phys. Lett.",
  ["soviet union. union sovietique"] = "Sov Union",
  ["soviet union. union soviétique"] = "Sov Union",
  ["sovremennai︠a︡ gerpetologii︠a︡"] = "Sovrem Gerpetol",
  ["sovremennaya matematika. fundamentalnye napravleniya"] = "Sovrem. Mat. Fundam. Napravl.",
  ["sovremennye fiziko-tekhnicheskie problemy"] = "Sovrem. Fiz.-Tekh. Probl.",
  ["sovremennye problemy matematiki"] = "Sovrem. Probl. Mat.",
  ["sovremennye problemy onkologii"] = "Sovrem Probl Onkol",
  ["sovremennye problemy tuberkuleza"] = "Sovrem Probl Tuberk",
  ["sovremennye tekhnologii v medit︠s︡ine"] = "Sovrem Tekhnologii Med",
  ["sowjetwissenschaft; naturwissenschaftliche beitrage"] = "Sowjetwiss Naturwiss Beitr",
  ["sowjetwissenschaft; naturwissenschaftliche beiträge"] = "Sowjetwiss Naturwiss Beitr",
  ["soybean genetics newsletter"] = "Soyb Genet Newsl",
  ["sozial- und praventivmedizin"] = "Soz Praventivmed",
  ["sozial- und präventivmedizin"] = "Soz Praventivmed",
  ["soziale welt : zeitschrift fuer sozialwissenschaftliche forschung und praxis"] = "Soz Welt",
  ["soziale welt-zeitschrift für sozialwissenschaftliche forschung und praxis"] = "Sozial Welt-Zeit. Sozialwiss. Forsch. Praxis",
  ["sozialgeschichte der medizin"] = "Sozialgesch Med",
  ["sozialistische forstwirtschaft, die"] = "Sozial. Forstwirtsch.",
  ["sozialmedizinische und padagogische jugendkunde"] = "Sozialmed. Padagog. Jugendkd.",
  ["sozialmedizinische und pädagogische jugendkunde"] = "Sozialmed Padagog Jugendkd",
  ["sozialwissenschaftliche informationen"] = "Soz Wiss Inf",
  ["space & polity"] = "Space Polity",
  ["space (beaconsfield, buckinghamshire)"] = "Space",
  ["space and culture : the journal"] = "Space Cult",
  ["space communications"] = "Space Commun",
  ["space conferences and exposition : aiaa space 2016"] = "SPACE Conf Expo (2016)",
  ["space life sciences"] = "Space Life Sci",
  ["space policy"] = "Space Policy",
  ["space research today"] = "Space Res. Today",
  ["space science and technologies"] = "Space Sci. Technol.",
  ["space science review"] = "Space Sci. Rev.",
  ["space science reviews"] = "Space Sci. Rev.",
  ["space technology (oxford, england)"] = "Space Technol",
  ["space technology library"] = "Space Technol. Libr.",
  ["space technology proceedings"] = "Space Technol. Proc.",
  ["space weather"] = "Space Weather",
  ["space weather : the international journal of research & applications"] = "Space Weather",
  ["space weather quarterly"] = "Space Weather Q.",
  ["space world"] = "Space World",
  ["space: science & technology"] = "Space Sci. Technol.",
  ["spacecraft and satellites"] = "Spacecraft Satell.",
  ["spaces & flows : an international journal of urban and extraurban studies"] = "Spaces Flows",
  ["spagna contemporanea"] = "Spagna Contemp",
  ["spal. revista de prehistoria y arqueología de la universidad de sevilla"] = "Spal",
  ["spanish economic review"] = "Spanish Econ. Rev.",
  ["spanish journal of agricultural research"] = "Span. J. Agric. Res.",
  ["spanish journal of agricultural research = revista de investigación agraria"] = "Span J Agric Res",
  ["spanish journal of palaeontology"] = "Span. J. Palaeontol.",
  ["spanish journal of psychology"] = "Span. J. Psychol.",
  ["spanish journal of soil science"] = "Span. J. Soil Sci.",
  ["spartan medical research journal"] = "Spartan Med Res J",
  ["spas book series"] = "SPAS Book Ser.",
  ["spastics' quarterly"] = "Spastic Quart",
  ["spatial and spatio-temporal epidemiology"] = "Spat Spatiotemporal Epidemiol",
  ["spatial cognition & computation"] = "Spatial Cognit. Comput.",
  ["spatial cognition and computation"] = "Spat Cogn Comput",
  ["spatial demography"] = "Spat Demogr",
  ["spatial economic analysis"] = "Spat Econ Anal",
  ["spatial information research"] = "Spatial Inf. Res.",
  ["spatial research and planning"] = "Spatial Res. Plann.",
  ["spatial statistics"] = "Spat Stat",
  ["spatial vision"] = "Spat Vis",
  ["spatio-temporal image analysis for longitudinal and time-series image data : second international workshop, stia 2012, held in conjunction with miccai 2012, nice, france, october 1, 2012, proceedings. stia (conference) (2nd : 2012 : nice, france)"] = "Spatiotemporal Image Anal Longitud Time Ser Image Data (2012)",
  ["spatio-temporal image analysis for longitudinal and time-series image data : third international workshop, stia 2014, held in conjunction with miccai 2014, boston, ma, usa, september 18, 2014 : revised selected papers. stia (conference) (3rd : 2014 : boston, mass.)"] = "Spatiotemporal Image Anal Longitud Time Ser Image Data (2014)",
  ["spe drilling & completion"] = "SPE Drill. Complet",
  ["spe formation evaluation"] = "SPE Form. Eval.",
  ["spe journal"] = "SPE J.",
  ["spe polymers"] = "SPE Polym.",
  ["spe production & facilities"] = "SPE Prod. Facil.",
  ["spe production & operations"] = "SPE Prod. Oper.",
  ["spe production engineering"] = "SPE Prod. Eng.",
  ["spe reservoir engineering"] = "SPE Reservoir Eng",
  ["spe reservoir evaluation & engineering"] = "SPE Reservoir Eval. Eng",
  ["special bulletin of the japanese society of coleopterology"] = "Spec Bull Jpn Soc Coleopterol",
  ["special care in dentistry"] = "Spec. Care Dentist.",
  ["special care in dentistry : official publication of the american association of hospital dentists, the academy of dentistry for the handicapped, and the american society for geriatric dentistry"] = "Spec Care Dentist",
  ["special collections"] = "Spec Collect",
  ["special discussions of the faraday society"] = "Spec. Discuss. Faraday Soc.",
  ["special education"] = "Spec Educ",
  ["special education: forward trends"] = "Spec Educ Forward Trends",
  ["special issue of imvi"] = "Spec. Issue IMVI",
  ["special libraries"] = "Spec Libr",
  ["special matrices"] = "Spec. Matrices",
  ["special operations journal"] = "Special Oper. J.",
  ["special paper of the geological survey of finland"] = "Spec. Pap. Geol. Surv. Finl.",
  ["special publication (american philosophical society)"] = "Spec Publ Am Philos Soc",
  ["special publication (european aquaculture society)"] = "Spec Publ (Eur Aquac Soc)",
  ["special publication - royal society of chemistry"] = "Spec. Publ. - R. Soc. Chem.",
  ["special publications (texas tech university. museum)"] = "Spec Publ Tex Tech Univ Mus",
  ["special report series (medical research council (great britain))"] = "Spec Rep Ser Med Res Counc (G B)",
  ["special report series - indian council of medical research"] = "Spec Rep Ser Indian Counc Med Res",
  ["special report series, indian council of medical research"] = "Spec. Rep. Ser. Indian Counc. Med. Res.",
  ["special topics & reviews in porous media: an international journal"] = "Spec. Top. Rev. Porous Media: Int. J.",
  ["special topics in endocrinology and metabolism"] = "Spec Top Endocrinol Metab",
  ["specialty law digest. health care (monthly)"] = "Spec Law Dig Health Care (Mon)",
  ["specialty law digest. health care law"] = "Spec Law Dig Health Care Law",
  ["species diversity : an international journal for taxonomy, systematics, speciation, biogeography, and life history research of animals"] = "Species Divers",
  ["specifying engineer"] = "Specif Eng",
  ["specimina nova dissertationum ex instituto historico universitatis quinqueecclesiensis de iano pannonio nominatae"] = "SpNov",
  ["spectator (london, england : 1828)"] = "Spectator",
  ["spectra analyse"] = "Spectra Anal.",
  ["spectra biologie"] = "Spectra Biol.",
  ["spectral and shape analysis in medical imaging : first international workshop, sesami 2016, held in conjunction with miccai 2016, athens, greece, october 21, 2016 : revised selected papers. sesami (workshop) (1st : 2016 : athens, greece)"] = "Spectr Shape Anal Med Imaging (2016)",
  ["spectrochimica acta"] = "Spectrochim. Acta",
  ["spectrochimica acta part a: molecular and biomolecular spectroscopy"] = "Spectrochim. Acta, Part A",
  ["spectrochimica acta part b: atomic spectroscopy"] = "Spectrochim. Acta, Part B",
  ["spectrochimica acta, part a: molecular and biomolecular spectroscopy"] = "Spectrochim. Acta, Part A",
  ["spectrochimica acta, part a: molecular spectroscopy"] = "Spectrochim. Acta, Part A",
  ["spectrochimica acta, part b: atomic spectroscopy"] = "Spectrochim. Acta, Part B",
  ["spectrochimica acta. part a, molecular and biomolecular spectroscopy"] = "Spectrochim Acta A Mol Biomol Spectrosc",
  ["spectrochimica acta. part a: molecular spectroscopy"] = "Spectrochim Acta A",
  ["spectrochimica acta. part b, atomic spectroscopy"] = "Spectrochim Acta Part B At Spectrosc",
  ["spectroscopic properties of inorganic and organometallic compounds"] = "Spectrosc. Prop. Inorg. Organomet. Compd.",
  ["spectroscopy (amsterdam, netherlands)"] = "Spectroscopy (Amsterdam, Neth.)",
  ["spectroscopy (springfield, or.)"] = "Spectroscopy (Springf)",
  ["spectroscopy asia"] = "Spectrosc. Asia",
  ["spectroscopy europe"] = "Spectrosc. Eur.",
  ["spectroscopy letters"] = "Spectrosc. Lett.",
  ["spectroscopy letters; an international journal for rapid communication"] = "Spectrosc Lett",
  ["spectrum (lexington, ky.)"] = "Spectrum",
  ["spectrum slovakia"] = "Spectr. Slovak.",
  ["speculations in science and technology"] = "Speculations Sci Technol",
  ["speech communication"] = "Speech Commun",
  ["speech pathology and therapy"] = "Speech Pathol Ther",
  ["speech prosody (urbana, ill.)"] = "Speech Prosody",
  ["speech, language and hearing (london, england)"] = "Speech Lang Hear",
  ["spektrum der augenheilkunde : zeitschrift der österreichischen ophthalmologischen gesellschaft, öog"] = "Spektrum Augenheilkd",
  ["spektrum der wissenschaft"] = "Spektrum Wiss.",
  ["spektrum lehrbuch"] = "Spektrum Lehrbuch",
  ["spekulation und erfahrung"] = "Spekulation Erfahrung Abt. II Unters.",
  ["speleobiology notes"] = "Speleobiology notes",
  ["spg biomed"] = "SPG Biomed",
  ["spie milestone series"] = "SPIE Milest. Ser.",
  ["spie newsroom"] = "SPIE Newsroom",
  ["spie reviews"] = "SPIE Rev.",
  ["spiegel historiael"] = "Spieg Hist",
  ["spill science and technology bulletin"] = "Spill Sci. Technol. Bull.",
  ["spinal cord"] = "Spinal Cord",
  ["spinal cord : the official journal of the international medical society of paraplegia"] = "Spinal Cord",
  ["spinal cord series and cases"] = "Spinal Cord Ser Cases",
  ["spine (philadelphia, pa. : 1986)"] = "Spine (Phila Pa 1986)",
  ["spine deformity"] = "Spine Deform",
  ["spine journal"] = "Spine J.",
  ["spine surgery and related research"] = "Spine Surg Relat Res",
  ["spinner (new bedford, mass.)"] = "Spinner",
  ["spirituality in clinical practice (washington, d.c.)"] = "Spiritual Clin Pract (Wash D C )",
  ["spisanie na bulgarskata akademiia na naukite. bulgarska akademiia na naukite"] = "Spis Blg Akad Nauk",
  ["spisanie na bŭlgarskata akademii︠a︡ na naukite. bŭlgarska akademii︠a︡ na naukite"] = "Spis Blg Akad Nauk",
  ["spisy pedagogicke fakulty v ostrave"] = "Spisy Ped. Fak. Ostrave",
  ["splice : the splice of life"] = "Splice Life",
  ["spoletium. rivista di arte, storia, cultura"] = "Spoletium",
  ["spora : a journal of biomathematics"] = "Spora",
  ["spore (english edition)"] = "Spore",
  ["sport exercise and performance psychology"] = "Sport Exercise Perform. Psychol.",
  ["sport history review"] = "Sport Hist Rev",
  ["sport in society"] = "Sport Soc",
  ["sport sciences for health"] = "Sport Sci Health",
  ["sport, education and society"] = "Sport Educ Soc",
  ["sport, ethics and philosophy"] = "Sport Ethics Philos",
  ["sport, exercise, and performance psychology"] = "Sport Exerc Perform Psychol",
  ["sports (basel, switzerland)"] = "Sports (Basel)",
  ["sports biomechanics"] = "Sports Biomech",
  ["sports biomechanics / international society of biomechanics in sports"] = "Sports Biomech",
  ["sports engineering"] = "Sports Eng.",
  ["sports health"] = "Sports Health",
  ["sports medicine"] = "Sports Med.",
  ["sports medicine (auckland, n.z.)"] = "Sports Med",
  ["sports medicine - open"] = "Sports Med Open",
  ["sports medicine and arthroscopy review"] = "Sports Med Arthrosc Rev",
  ["sports medicine and health science"] = "Sports Med. Health Sci.",
  ["sports medicine international open"] = "Sports Med Int Open",
  ["sports medicine, arthroscopy, rehabilitation, therapy & technology : smartt"] = "Sports Med Arthrosc Rehabil Ther Technol",
  ["sportverletzung sportschaden"] = "Sportverletz. Sportschaden",
  ["sportverletzung sportschaden : organ der gesellschaft fur orthopadisch-traumatologische sportmedizin"] = "Sportverletz Sportschaden",
  ["sportverletzung sportschaden : organ der gesellschaft für orthopädisch-traumatologische sportmedizin"] = "Sportverletz Sportschaden",
  ["sprachkunst. beiträge zur literaturwissenschaft"] = "Sprachkunst",
  ["sprawozdania z czynnosci i posiedzen - lodzkie towarzystwo naukowe"] = "Spraw Czynnosci Posiedz Lodz Tow Nauk",
  ["sprawozdania z czynności i posiedzeń - lódzkie towarzystwo naukowe"] = "Spraw Czynnosci Posiedz Lodz Tow Nauk",
  ["springer actuarial"] = "Springer Actuar.",
  ["springer actuarial textbooks"] = "Springer Actuar. Textb.",
  ["springer aerospace technology"] = "Springer Aerosp. Technol.",
  ["springer biographies"] = "Springer Biogr.",
  ["springer briefs in business"] = "Springer Briefs Bus.",
  ["springer briefs in environmental science"] = "Springer Briefs Environ. Sci.",
  ["springer briefs in operations research"] = "Springer Briefs Oper. Res.",
  ["springer briefs in water science and technology"] = "Springer Briefs Water Sci. Technol.",
  ["springer climate"] = "Springer Clim.",
  ["springer collected works in mathematics"] = "Springer Collect. Works Math.",
  ["springer finance"] = "Springer Finance",
  ["springer finance lecture notes"] = "Springer Finance Lect. Notes",
  ["springer finance textbooks"] = "Springer Finance Textb.",
  ["springer geology"] = "Springer Geol.",
  ["springer geophysics"] = "Springer Geophys.",
  ["springer graduate texts in philosophy"] = "Springer Grad. Texts Philos.",
  ["springer handbook of auditory research"] = "Springer Handb Auditory Res",
  ["springer handbooks of computational statistics"] = "Springer Handb. Comput. Stat.",
  ["springer indam series"] = "Springer INdAM Ser.",
  ["springer international handbooks of education"] = "Springer Int. Handb. Educ.",
  ["springer lehrbuch mathematik"] = "Springer Lehrbuch Math.",
  ["springer monographs in mathematics"] = "Springer Monogr. Math.",
  ["springer optimization and its applications"] = "Springer Optim. Appl.",
  ["springer proceedings in advanced robotics"] = "Springer Proc. Adv. Robot.",
  ["springer proceedings in business and economics"] = "Springer Proc. Bus. Econ.",
  ["springer proceedings in complexity"] = "Springer Proc. Complexity",
  ["springer proceedings in earth and environmental sciences"] = "Springer Proc. Earth Environ. Sci.",
  ["springer proceedings in mathematics"] = "Springer Proc. Math.",
  ["springer proceedings in mathematics & statistics"] = "Springer Proc. Math. Stat.",
  ["springer proceedings in physics"] = "Springer Proc. Phys.",
  ["springer science reviews"] = "Springer Sci Rev",
  ["springer seminars in immunopathology"] = "Springer Semin Immunopathol",
  ["springer series in advanced manufacturing"] = "Springer Ser. Adv. Manuf.",
  ["springer series in astrophysics and cosmology"] = "Springer Ser. Astrophys. Cosmol.",
  ["springer series in astrostatistics"] = "Springer Ser. Astrostatistics",
  ["springer series in bio-/neuroinformatics"] = "Springer Ser. Bio-/Neuroinform.",
  ["springer series in biophysics"] = "Springer Ser Biophys",
  ["springer series in chemical physics"] = "Springer Ser. Chem. Phys.",
  ["springer series in computational mathematics"] = "Springer Ser. Comput. Math.",
  ["springer series in discrete mathematics and theoretical computer science"] = "Springer Ser. Discrete Math. Theor. Comput. Sci.",
  ["springer series in game theory"] = "Springer Ser. Game Theory",
  ["springer series in information sciences"] = "Springer Ser. Inform. Sci.",
  ["springer series in light scattering"] = "Springer Ser. Light Scatt.",
  ["springer series in materials science"] = "Springer Ser. Mater. Sci.",
  ["springer series in measurement science and technology"] = "Springer Ser. Meas. Sci. Technol.",
  ["springer series in nonlinear dynamics"] = "Springer Ser. Nonlinear Dynam.",
  ["springer series in operations research"] = "Springer Ser. Oper. Res.",
  ["springer series in operations research and financial engineering"] = "Springer Ser. Oper. Res. Financ. Eng.",
  ["springer series in opical sciences"] = "Springer Ser. Opt. Sci.",
  ["springer series in optical sciences"] = "Springer Ser. Opt. Sci.",
  ["springer series in pharmaceutical statistics"] = "Springer Ser. Pharm. Stat.",
  ["springer series in plasma science and technology"] = "Springer Ser. Plasma Sci. Technol.",
  ["springer series in reliability engineering"] = "Springer Ser. Reliab. Eng.",
  ["springer series in solid and structural mechanics"] = "Springer Ser. Solid Struct. Mech.",
  ["springer series in solid-state sciences"] = "Springer Ser. Solid-State Sci.",
  ["springer series in statistics"] = "Springer Ser. Statist.",
  ["springer series in statistics. perspectives in statistics"] = "Springer Ser. Statist. Perspect. Statist.",
  ["springer series in statistics: perspectives in statistics"] = "Springer Ser. Statist. Perspect. Statist.",
  ["springer series in statistics: statistics in the health sciences"] = "Springer Ser. Statist. Statist. Health Sci.",
  ["springer series in supply chain management"] = "Springer Ser. Supply Chain Manag.",
  ["springer series in surface sciences"] = "Springer Ser. Surface Sci.",
  ["springer series in synergetics"] = "Springer Ser. Synergetics",
  ["springer series in the data sciences"] = "Springer Ser. Data Sci.",
  ["springer series on atomic, optical, and plasma physics"] = "Springer Ser. At. Opt. Plasma Phys.",
  ["springer series on chemical sensors and biosensors"] = "Springer Ser. Chem. Sens. Biosens.",
  ["springer series on cultural computing"] = "Springer Ser. Cult. Comput.",
  ["springer series on health care and society"] = "Springer Ser Health Care Soc",
  ["springer series on naval architecture, marine engineering, shipbuilding and shipping"] = "Springer Ser. Nav. Archit. Mar. Eng. Shipbuild. Shipp.",
  ["springer series on wave phenomena"] = "Springer Ser. Wave Phenomena",
  ["springer studies in the history of economic thought"] = "Springer Stud. Hist. Econ. Thought",
  ["springer studium mathematik—master"] = "Springer Stud. Math. Master",
  ["springer textbooks in earth sciences, geography and environment"] = "Springer Textb. Earth Sci. Geogr. Environ.",
  ["springer texts in business and economics"] = "Springer Texts Bus. Econ.",
  ["springer texts in social sciences"] = "Springer Texts Soc. Sci.",
  ["springer texts in statistics"] = "Springer Texts Statist.",
  ["springer topics in signal processing"] = "Springer Top. Signal Process.",
  ["springer tracts in advanced robotics"] = "Springer Tracts Adv. Robot.",
  ["springer tracts in civil engineering"] = "Springer Tracts Civ. Eng.",
  ["springer tracts in mechanical engineering"] = "Springer Tracts Mech. Eng.",
  ["springer tracts in modern physics"] = "Springer Tracts Modern Phys.",
  ["springer tracts in natural philosophy"] = "Springer Tracts Nat. Philos.",
  ["springer tracts in nature-inspired computing"] = "Springer Tracts Nat.-Inspir. Comput.",
  ["springer tracts on transportation and traffic"] = "Springer Tracts Transp. Traffic",
  ["springer transactions in civil and environmental engineering"] = "Springer Trans. Civil Environ. Eng.",
  ["springer uncertainty research"] = "Springer Uncertain. Res.",
  ["springer undergraduate mathematics series"] = "Springer Undergrad. Math. Ser.",
  ["springer undergraduate textbooks in statistics"] = "Springer Undergrad. Textb. Stat.",
  ["springer undergraduate texts in mathematics and technology"] = "Springer Undergrad. Texts Math. Technol.",
  ["springer undergraduate texts in philosophy"] = "Springer Undergrad. Texts Philos.",
  ["springer-lehrbuch masterclass"] = "Springer-Lehrb. Mastercl.",
  ["springer-praxis books in astronomy and planetary sciences"] = "Springer-Praxis Books Astron. Planet. Sci.",
  ["springer-praxis books in environmental sciences"] = "Springer-Praxis Books Environ. Sci.",
  ["springer-praxis books. popular science"] = "Springer-Praxis Books. Pop. Sci.",
  ["springerbriefs in advanced information and knowledge processing"] = "SpringerBriefs Adv. Inf. Knowl. Process.",
  ["springerbriefs in applied sciences and technology"] = "SpringerBriefs Appl. Sci. Technol.",
  ["springerbriefs in applied statistics and econometrics"] = "SpringerBriefs Appl. Stat. Econom.",
  ["springerbriefs in astronomy"] = "SpringerBriefs Astron.",
  ["springerbriefs in bioengineering"] = "SpringerBriefs Bioeng.",
  ["springerbriefs in cognitive mathematics"] = "SpringerBriefs Cogn. Math.",
  ["springerbriefs in complexity"] = "SpringerBriefs Complex.",
  ["springerbriefs in computational intelligence"] = "SpringerBriefs Comput. Intell.",
  ["springerbriefs in computational mechanics"] = "SpringerBriefs Comput. Mech.",
  ["springerbriefs in computer science"] = "SpringerBriefs Comput. Sci.",
  ["springerbriefs in continuum mechanics"] = "SpringerBriefs Contin. Mech.",
  ["springerbriefs in control, automation and robotics"] = "SpringerBriefs Control Autom. Robot.",
  ["springerbriefs in earth sciences"] = "SpringerBriefs Earth Sci.",
  ["springerbriefs in earth system sciences"] = "SpringerBriefs Earth Syst. Sci.",
  ["springerbriefs in ecology"] = "SpringerBriefs Ecol.",
  ["springerbriefs in economics"] = "SpringerBriefs Econ.",
  ["springerbriefs in electrical and computer engineering"] = "SpringerBriefs Electr. Comput. Eng.",
  ["springerbriefs in energy"] = "SpringerBriefs Energy",
  ["springerbriefs in finance"] = "SpringerBriefs Finance",
  ["springerbriefs in geography"] = "SpringerBriefs Geogr.",
  ["springerbriefs in history of science and technology"] = "SpringerBriefs Hist. Sci. Technol.",
  ["springerbriefs in information security and cryptography"] = "SpringerBriefs Inf. Secur. Cryptogr.",
  ["springerbriefs in intelligent systems"] = "SpringerBriefs Intell. Syst.",
  ["springerbriefs in mathematical methods"] = "SpringerBriefs Math. Methods",
  ["springerbriefs in mathematical physics"] = "SpringerBriefs Math. Phys.",
  ["springerbriefs in mathematics"] = "SpringerBriefs Math.",
  ["springerbriefs in mathematics of planet earth"] = "SpringerBriefs Math. Planet Earth",
  ["springerbriefs in molecular science"] = "SpringerBriefs Mol. Sci.",
  ["springerbriefs in nonlinear circuits"] = "SpringerBriefs Nonlinear Circuits",
  ["springerbriefs in optimization"] = "SpringerBriefs Optim.",
  ["springerbriefs in philosophy"] = "SpringerBriefs Philos.",
  ["springerbriefs in physics"] = "SpringerBriefs Phys.",
  ["springerbriefs in population studies"] = "SpringerBriefs Popul. Stud.",
  ["springerbriefs in probability and mathematical statistics"] = "SpringerBriefs Probab. Math. Stat.",
  ["springerbriefs in quantitative finance"] = "SpringerBriefs Quant. Finance",
  ["springerbriefs in signal processing"] = "SpringerBriefs Signal Process.",
  ["springerbriefs in statistics"] = "SpringerBriefs Stat.",
  ["springerbriefs in statistics—abe"] = "SpringerBriefs Stat. ABE",
  ["springerbriefs in structural mechanics"] = "SpringerBriefs Struct. Mech.",
  ["springerbriefs in the mathematics of materials"] = "SpringerBriefs Math. Mater.",
  ["springerbriefs in thermal engineering and applied science"] = "SpringerBriefs Therm. Eng. Appl. Sci.",
  ["springerbriefs on pdes and data science"] = "SpringerBriefs PDEs Data Sci.",
  ["springers kurzlehrbucher der wirtschaftswissenschaften"] = "Springers Kurzlehrb. Wirtsch.",
  ["spums journal"] = "SPUMS J",
  ["spums journal / south pacific underwater medicine society"] = "SPUMS J",
  ["spvn : journal of the society for peripheral vascular nursing"] = "SPVN",
  ["squibb memoranda"] = "Squibb Memo",
  ["sra journal"] = "SRA J",
  ["sri lanka journal of population studies"] = "Sir Lanka J Popul Stud",
  ["sri lanka journal of social sciences"] = "Sri Lanka J Soc Sci",
  ["sri lanka population digest"] = "Sri Lanka Popul Dig",
  ["sri lankan journal of applied statistics"] = "Sri Lankan J Appl Stat",
  ["srl neurology & neurosurgery"] = "SRL Neurol Neurosurg",
  ["srpski arhiv za celokupno lekarstvo"] = "Srp Arh Celok Lek",
  ["srs bulletin"] = "SRS Bull",
  ["srx pharmacology"] = "SRX Pharmacol",
  ["ssm - population health"] = "SSM Popul Health",
  ["sso; schweizerische monatsschrift fur zahnheilkunde"] = "SSO Schweiz. Monatsschr. Zahnheilkd.",
  ["st luke’s hospital gazette"] = "St Lukes Hosp. Gaz. (Guardamangia)",
  ["st. barnabas hospital medical bulletin"] = "St Barnabas Hosp Med Bull",
  ["st. bartholomew's hospital journal"] = "St Bartholomews Hosp J",
  ["st. bartholomew’s hospital journal"] = "St Bartholomews Hosp J",
  ["st. john's law review"] = "St Johns Law Rev",
  ["st. john’s law review"] = "St Johns Law Rev",
  ["st. luke's hospital staff clinics bulletin"] = "St. Lukes Hosp Staff Clin Bull",
  ["st. luke's journal of theology"] = "St Lukes J Theol",
  ["st. luke’s journal of theology"] = "St Lukes J Theol",
  ["st. mary's hospital gazette"] = "St Marys Hosp Gaz",
  ["st. mary's law journal"] = "St Marys Law J",
  ["st. mary’s hospital gazette"] = "St Marys Hosp Gaz",
  ["st. mary’s law journal"] = "St Marys Law J",
  ["st. petersburg mathematical journal"] = "St. Petersburg Math. J.",
  ["st. petersburg polytechnical university journal: physics and mathematics"] = "St. Petersburg Polytech. Univ. J.: Phys. Math.",
  ["st. thomas law review"] = "St Thomas Law Rev",
  ["st. thomas's hospital gazette"] = "St Thomas Hosp Gaz",
  ["st. vincent's hospital medical bulletin"] = "ST Vincents Hosp Med Bull",
  ["st. vincent’s hospital medical bulletin"] = "ST Vincents Hosp Med Bull",
  ["st. vladimir's theological quarterly"] = "St Vladmirs Theol Q",
  ["st. vladimir’s theological quarterly"] = "St Vladmirs Theol Q",
  ["stability and control: theory, methods and applications"] = "Stability Control Theory Methods Appl.",
  ["stability, oscillations and optimization of systems"] = "Stab. Oscil. Optim. Syst.",
  ["stadion (cologne, germany)"] = "Stadion (Koln)",
  ["stadion. internationale zeitschrift für geschichte des sports"] = "Stadion",
  ["staff bulletin. easton, pa. hospital"] = "Staff Bull Easton Hosp (Easton, Pa.)",
  ["staff papers - international monetary fund. international monetary fund"] = "Staff Pap Int Monet Fund",
  ["staff proceedings. hillcrest medical center (tulsa, okla.)"] = "Staff Proc",
  ["stain technology"] = "Stain Technol",
  ["stal, sciences et techniques de l’animal de laboratoire"] = "STAL, Sci. Tech. Anim. Lab.",
  ["stal. sciences et techniques de l'animal de laboratoire"] = "Stal",
  ["stal. sciences et techniques de l’animal de laboratoire"] = "Stal",
  ["stammering research : an on-line journal published by the british stammering association"] = "Stammering Res",
  ["standardization news : sn"] = "Stand News",
  ["standards in genomic sciences"] = "Stand Genomic Sci",
  ["stanford environmental law journal"] = "Stanf Environ Law J",
  ["stanford french review"] = "Stanford Fr Rev",
  ["stanford journal of international law"] = "Stanford J Int Law",
  ["stanford journal of law, science & policy"] = "Stanf J Law Sci Policy",
  ["stanford law & policy review"] = "Stanford Law Pol Rev",
  ["stanford law review"] = "Stanford Law Rev",
  ["stanford lectures"] = "Stanford Lectures",
  ["stanford medical bulletin"] = "Stanford Med Bull",
  ["stanford technology law review : stlr : an online high-technology law journal from stanford law school"] = "Stanf Technol Law Rev",
  ["stanford text technologies"] = "Stanf. Text Technol.",
  ["stapp car crash journal"] = "Stapp Car Crash J",
  ["star protocols"] = "STAR Protoc",
  ["star: science & technology of archaeological research"] = "STAR: Sci. Technol. Archaeol. Res.",
  ["starch - starke"] = "Starch - Starke",
  ["starinararheoloki institut beograd"] = "Starinar",
  ["stat (international statistical institute)"] = "Stat (Int Stat Inst)",
  ["stat; bulletin of the wisconsin nurses association"] = "Stat.",
  ["state coverage initiatives issue brief : a national initiative of the robert wood johnson foundation"] = "State Coverage Initiat Issue Brief",
  ["state government"] = "State Gov.",
  ["state government (denver, colo.)"] = "State Gov",
  ["state health legislation report"] = "State Health Legis Rep",
  ["state legislatures"] = "State Legis",
  ["state nursing legislation quarterly"] = "State Nurs Legis Q",
  ["state of health care in america"] = "State Health Care Am.",
  ["state of the art reports"] = "State Art Rep.",
  ["state politics & policy quarterly"] = "State Polit Policy Q",
  ["state reproductive health monitor : legislative proposals and actions"] = "State Reprod Health Monit",
  ["state reproductive health monitor : legislative proposals and actions / prepared by the alan guttmacher institute"] = "State Reprod Health Monit",
  ["state-of-the-arts series"] = "State. Arts. Ser.",
  ["states of health"] = "States Health",
  ["static & dynamic game theory: foundations & applications"] = "Static Dyn. Game Theory Found. Appl.",
  ["statistica neerlandica"] = "Stat Neerl",
  ["statistica neerlandica. journal of the netherlands society for statistics and operations research"] = "Stat. Neerl.",
  ["statistica sinica"] = "Statist. Sinica",
  ["statistical analysis and data mining"] = "Stat. Anal. Data Min.",
  ["statistical applications in genetics and molecular biology"] = "Stat Appl Genet Mol Biol",
  ["statistical atlases and computational models of the heart. stacom (workshop)"] = "Stat Atlases Comput Models Heart",
  ["statistical bulletin (metropolitan life foundation)"] = "Stat Bull Metrop Life Found",
  ["statistical bulletin (metropolitan life insurance company : 1984)"] = "Stat Bull Metrop Insur Co",
  ["statistical bulletin (metropolitan life insurance company)"] = "Stat Bull Metropol Life Insur Co",
  ["statistical bulletin / metropolitan insurance companies"] = "Stat Bull Metrop Insur Co",
  ["statistical bulletin of czechoslovakia. czechoslovak republic. státni urad statistický"] = "Stat Bull Czechoslov",
  ["statistical bulletin, metropolitan insurance companies"] = "Stat. Bull. Metrop. Insur. Co.",
  ["statistical bulletin, metropolitan life foundation"] = "Stat. Bull. Metrop. Life Found.",
  ["statistical bulletin, metropolitan life insurance company"] = "Stat. Bull. Metropol. Life Insur. Co.",
  ["statistical communications in infectious diseases"] = "Stat Commun Infect Dis",
  ["statistical ecology series"] = "Statist. Ecology Ser.",
  ["statistical inference for stochastic processes"] = "Stat. Inference Stochastic Processes",
  ["statistical inference for stochastic processes. an international journal devoted to time series analysis and the statistics of continuous time processes and dynamical systems"] = "Stat. Inference Stoch. Process.",
  ["statistical information bulletin for africa = bulletin d'information statistique pour l'afrique"] = "Stat Inf Bull Afr",
  ["statistical journal"] = "Statistical J.",
  ["statistical journal of the iaos"] = "Stat J IAOS",
  ["statistical journal of the united nations economic commission for europe"] = "Stat J UN Econ Comm Eur",
  ["statistical methodology"] = "Stat. Methodol.",
  ["statistical methods & applications"] = "Stat Methods Appt",
  ["statistical methods & applications. journal of the italian statistical society"] = "Stat. Methods Appl.",
  ["statistical methods in medical research"] = "Stat Methods Med Res",
  ["statistical modeling and decision science"] = "Statist. Model. Decis. Sci.",
  ["statistical modelling"] = "Stat Modelling",
  ["statistical modelling. an international journal"] = "Stat. Model.",
  ["statistical news. poland. glowny urzad statystyczny"] = "Stat News Pol",
  ["statistical news. poland. główny urzad statystyczny"] = "Stat News Pol",
  ["statistical notes for health planners"] = "Stat Notes Health Plann",
  ["statistical notes of japan"] = "Stat Notes Jpn",
  ["statistical papers"] = "Statist. Papers",
  ["statistical papers (berlin, germany)"] = "Stat Pap (Berl)",
  ["statistical reporter / office of federal statistical policy and standards"] = "Stat Report",
  ["statistical science"] = "Stat. Sci.",
  ["statistical science : a review journal of the institute of mathematical statistics"] = "Stat Sci",
  ["statistical science and interdisciplinary research"] = "Stat. Sci. Interdiscip. Res.",
  ["statistical science. a review journal of the institute of mathematical statistics"] = "Statist. Sci.",
  ["statistical theory and method abstracts"] = "STMA",
  ["statistical theory and related fields"] = "Stat. Theory Relat. Fields",
  ["statisticheskie metody v ekonomicheskikh issledovaniyakh"] = "Statist. Metody Ekonom. Issled.",
  ["statistický zpravodaj. czechoslovak republic. státni urad statistický"] = "Stat Zprav",
  ["statistics & decisions"] = "Statist. Decisions",
  ["statistics & probability letters"] = "Statist. Probab. Lett.",
  ["statistics & risk modeling with applications in finance and insurance"] = "Stat. Risk Model.",
  ["statistics and applications"] = "Stat Appl",
  ["statistics and computing"] = "Stat. Comput.",
  ["statistics and its interface"] = "Stat. Interface",
  ["statistics and public policy (philadelphia, pa.)"] = "Stat Public Policy (Phila)",
  ["statistics for biology and health"] = "Stat. Biol. Health",
  ["statistics for engineering and physical science"] = "Stat. Eng. Phys. Sci.",
  ["statistics for industry and technology"] = "Stat. Ind. Technol.",
  ["statistics for industry, technology, and engineering"] = "Stat. Ind. Technol. Eng.",
  ["statistics for social and behavioral sciences"] = "Stat. Soc. Behav. Sci.",
  ["statistics in biopharmaceutical research"] = "Stat Biopharm Res",
  ["statistics in biosciences"] = "Stat Biosci",
  ["statistics in medicine"] = "Stat Med",
  ["statistics in practice"] = "Statist. Practice",
  ["statistics in the health sciences"] = "Statist. Health Sci.",
  ["statistics in the social and behavioral sciences series"] = "Stat. Soc. Behav. Sci. Ser.",
  ["statistics of navy medicine"] = "Stat Navy Med",
  ["statistics surveys"] = "Stat. Surv.",
  ["statistics, optimization & information computing"] = "Stat. Optim. Inf. Comput.",
  ["statistics, politics, and policy"] = "Stat Politics Policy",
  ["statistics. a journal of theoretical and applied statistics"] = "Statistics",
  ["statistics: textbooks and monographs"] = "Stat. Textb. Monogr.",
  ["statistique et decisions economiques"] = "Statist. Decisions Econom.",
  ["statistique mathématique et probabilité"] = "Stat. Math. Probab.",
  ["statistiques & etudes: midi-pyrenees"] = "Stat Etud Midi Pyren",
  ["statistiques & études: midi-pyrénees"] = "Stat Etud Midi Pyren",
  ["statistiques demographiques. institut national de statistique (belgium)"] = "State Demogr",
  ["statistiques démographiques. institut national de statistique (belgium)"] = "State Demogr",
  ["statistische nachrichten. österreichisches statistisches zentralamt"] = "Stat Nachr Osterr Stat Zent Amt",
  ["statistisches jahrbuch der schweiz"] = "Stat. Jahrb. Schweiz",
  ["statistisk tidskrift. statistical review. sweden. statistiska centralbyran"] = "Stat Tidskr",
  ["statistisk tidskrift. statistical review. sweden. statistiska centralbyrån"] = "Stat Tidskr",
  ["statistiska meddelanden. serie be"] = "Statist Medd Ser Be Befolk Val",
  ["statistiske efterretninger. befolkning og valg"] = "Statist Efterret",
  ["statistiske efterretninger. befolkning og valg / danmarks statistik"] = "Statist Efterret",
  ["statisztikai szemle : a magyar kozponti statisztikai hivatal folyoirata"] = "Stat Szle",
  ["statisztikai szemle : a magyar központi statisztikai hivatal folyóirata"] = "Stat Szle",
  ["statsokonomisk tidsskrift"] = "Statsokon. Tidsskr.",
  ["statutes of california and digests of measures / compiled by bion m. gregory, legislative counsel. california"] = "Statut Calif Dig Meas Calif",
  ["statutes of california and digests of measures. california"] = "Statut Calif Dig Meas Calif",
  ["staub - reinhaltung der luft"] = "Staub - Reinhalt. Luft",
  ["std bulletin"] = "STD Bull",
  ["std news : a quarterly newsletter of the american social health association"] = "STD News",
  ["steam-h: science, technology, engineering, agriculture, mathematics & health"] = "STEAM-H: Sci. Technol. Eng. Agric. Math. Health",
  ["steel and composite structures"] = "Steel Compos. Struct.",
  ["steel construction - design and research"] = "Steel Constr. Des. Res.",
  ["steel in translation"] = "Steel Transl.",
  ["steel research"] = "Steel Res.",
  ["steel research international"] = "Steel Res. Int.",
  ["steklo i keramika"] = "Steklo Keram.",
  ["stem cell and translational investigation"] = "Stem Cell Transl Investig",
  ["stem cell epigenetics"] = "Stem Cell Epigenet",
  ["stem cell investigation"] = "Stem Cell Investig",
  ["stem cell reports"] = "Stem Cell Reports",
  ["stem cell research"] = "Stem Cell Res",
  ["stem cell research & therapeutics"] = "Stem Cell Res Ther (Walnut)",
  ["stem cell research & therapy"] = "Stem Cell Res Ther",
  ["stem cell review and reports"] = "Stem Cell Rev. Rep.",
  ["stem cell reviews"] = "Stem Cell Rev",
  ["stem cell reviews and reports"] = "Stem Cell Rev Rep",
  ["stem cell studies"] = "Stem Cell Stud",
  ["stem cell: advanced research and therapy"] = "Stem Cell Adv Res Ther",
  ["stem cells"] = "Stem Cells",
  ["stem cells (dayton, ohio)"] = "Stem Cells",
  ["stem cells and cloning : advances and applications"] = "Stem Cells Cloning",
  ["stem cells and cloning: advances and applications"] = "Stem Cells Cloning: Adv. Appl.",
  ["stem cells and development"] = "Stem Cells Dev",
  ["stem cells international"] = "Stem Cells Int",
  ["stem cells translational medicine"] = "Stem Cells Transl Med",
  ["stem education"] = "STEM Educ.",
  ["stem fellowship journal"] = "STEM Fellowship J.",
  ["step perspective"] = "STEP Perspect",
  ["stephan mueller special publication series"] = "Stephan Mueller Spec. Publ. Ser.",
  ["stereotactic and functional neurosurgery"] = "Stereotact Funct Neurosurg",
  ["sterile world"] = "Sterile World",
  ["sterne und weltraum"] = "Sterne Weltraum",
  ["steroids and lipids research"] = "Steroids Lipids Res",
  ["stetson law review"] = "Stetson Law Rev",
  ["stewart william's glamorgan historian"] = "Glamorgan Hist",
  ["stewart william’s glamorgan historian"] = "Glamorgan Hist",
  ["stiftung reusstal, jahresbericht"] = "Stift. Reusstal, Jahresber.",
  ["stiftung zur förderung der hamburgischen kunstsammlungen. erwerbungen"] = "StiftHambKuSamml",
  ["stigma and health"] = "Stigma Health",
  ["stigma research and action"] = "Stigma Res Action",
  ["stimmen der zeit"] = "Stimmen Zeit",
  ["stn's journal of trauma nursing : the official journal of the society of trauma nurses"] = "STNS J Trauma Nurs",
  ["stns journal of trauma nursing"] = "STNS J. Trauma Nurs.",
  ["stn’s journal of trauma nursing : the official journal of the society of trauma nurses"] = "STNS J Trauma Nurs",
  ["stochastic analysis and applications"] = "Stochastic Anal. Appl.",
  ["stochastic environmental research and risk assessment"] = "Stochastic Environ. Res. Risk Assess.",
  ["stochastic environmental research and risk assessment : research journal"] = "Stoch Environ Res Risk Assess",
  ["stochastic modeling"] = "Stochastic Model.",
  ["stochastic modeling series"] = "Stochastic Model. Ser.",
  ["stochastic modelling and applied probability"] = "Stoch. Model. Appl. Probab.",
  ["stochastic models"] = "Stoch Model",
  ["stochastic models in survival analysis and reliability set"] = "Stoch. Models Surviv. Anal. Reliab. Set",
  ["stochastic partial differential equations : analysis and computations"] = "Stoch Partial Differ Equ",
  ["stochastic partial differential equations. analysis and computations"] = "Stoch. Partial Differ. Equ. Anal. Comput.",
  ["stochastic processes and their applications"] = "Stochastic Process. Appl.",
  ["stochastic systems"] = "Stochastic Syst.",
  ["stochastics (abingdon, england : 2005)"] = "Stochastics (Abingdon)",
  ["stochastics and dynamics"] = "Stochastics Dyn.",
  ["stochastics and partial differential equations-analysis and computations"] = "Stochastics Partial Differ. Equations-Anal. Comput.",
  ["stochastics and partial differential equations. analysis and computations"] = "Stoch. Partial Differ. Equ. Anal. Comput.",
  ["stochastics and quality control"] = "Stoch. Qual. Control",
  ["stochastics and stochastics reports"] = "Stochastics Stochastics Rep.",
  ["stochastics monographs"] = "Stochastics Monogr.",
  ["stochastics partial differential equations. analysis and computations"] = "Stoch. Partial Differ. Equ. Anal. Comput.",
  ["stochastics. an international journal of probability and stochastic processes"] = "Stochastics",
  ["stoma (lisbon, portugal)"] = "Stoma (Lisb)",
  ["stoma (thessalonike, greece)"] = "Stoma (Thessaloniki)",
  ["stoma (thessalonikē, greece)"] = "Stoma (Thessaloniki)",
  ["stomatologia mediterranea"] = "Stomatol. Mediterr.",
  ["stomatologia mediterranea : sm"] = "Stomatol Mediterr",
  ["stomatologicke zpravy"] = "Stomatol. Zpr.",
  ["stomatologicke zpravy / spofa"] = "Stomatol Zpr",
  ["stomatologické zprávy"] = "Stomatol Zpr",
  ["stomatologicky vestnik"] = "Stomatol Vestn",
  ["stomatologický vĕstník"] = "Stomatol Vestn",
  ["stomatologie (bucharest, romania)"] = "Stomatologie",
  ["stomatologie der ddr"] = "Stomatol DDR",
  ["stomatologiia. stomatology"] = "Stomatologiia (Sofiia)",
  ["stomatologika chronika"] = "Stomatol. Chron. (Athenai)",
  ["stomatologika chronika. acta stomatologica hellenica"] = "Stomatol Chron (Athenai)",
  ["stomatoloski glasnik srbije"] = "Stomatol. Glas. Srb.",
  ["stomatoloski vjesnik (stomatological review)"] = "Stomatol. Vjesn.",
  ["stomatoloski vjesnik. stomatological review"] = "Stomatol Vjesn",
  ["stomatolos̆ki glasnik srbije"] = "Stomatol Glas Srb",
  ["stomatolos̆ki vjesnik. stomatological review"] = "Stomatol Vjesn",
  ["storia contemporanea"] = "Stor Contemp",
  ["storia della storiografia"] = "SStor",
  ["storia e critica della psicologia"] = "Stor Crit Psicol",
  ["storia e medicina popolare"] = "Stor Med Pop",
  ["storia in lombardia"] = "Stor Lomb",
  ["storia urbana"] = "Stor Urbana",
  ["strahlenschutz in forschung und praxis"] = "Strahlenschutz Forsch Prax",
  ["strahlentherapie und onkologie"] = "Strahlenther. Onkol.",
  ["strahlentherapie. sonderbande"] = "Strahlentherapie [Sonderb.]",
  ["strahlentherapie. sonderbände"] = "Strahlentherapie Sonderb",
  ["strasbourg medical"] = "Strasb. Med.",
  ["strasbourg médical"] = "Strasb Med",
  ["strategic organization"] = "Strateg Organ",
  ["strategic planning for energy and the environment"] = "Strategic Plann. Energy Environ.",
  ["strategies (la jolla, san diego, calif.)"] = "Strategies",
  ["strategies for healthcare excellence"] = "Strateg. Healthc. Excell.",
  ["strategies for healthcare excellence : organizational productivity, quality and effectiveness"] = "Strateg Healthc Excell",
  ["strategies in molecular biology"] = "Strateg Mol Biol",
  ["strategies in trauma and limb reconstruction"] = "Strategies Trauma Limb Reconstr",
  ["strategy & tactics (cambria, calif.)"] = "Strategy Tactics",
  ["stratigraphy and geological correlation"] = "Stratigr. Geol. Correl.",
  ["stratigraphy and geological correlation = stratigrafiya, geologicheskaya korrelyatsiya"] = "Stratigr Geol Correl",
  ["strength & conditioning journal"] = "Strength Cond. J.",
  ["strength and conditioning journal"] = "Strength Cond J",
  ["strength of materials"] = "Strength Mater.",
  ["strength, fracture and complexity"] = "Strength Fract Complex",
  ["strenna dei romanisti"] = "StrennaRom",
  ["strenna storica bolognese"] = "Strenna Stor Bolognese",
  ["stress (amsterdam, netherlands)"] = "Stress",
  ["stress and health : journal of the international society for the investigation of stress"] = "Stress Health",
  ["stress biology"] = "Stress Biol.",
  ["stress medicine"] = "Stress Med",
  ["stroke and vascular neurology"] = "Stroke Vasc Neurol",
  ["stroke research and treatment"] = "Stroke Res Treat",
  ["stroke: vascular and interventional neurology"] = "Stroke: Vasc. Interventional Neurol.",
  ["stroke; a journal of cerebral circulation"] = "Stroke",
  ["struct.–process., meas., phenom. journal of vacuum science & technology, b: microelectronics and nanometer structures–processing, measurement, and phenomena"] = "J. Vac. Sci. Technol., B: Microelectron. Nanometer",
  ["structural and multidisciplinary optimization"] = "Struct. Multidiscip.  Optim.",
  ["structural and multidisciplinary optimization : journal of the international society for structural and multidisciplinary optimization"] = "Struct Multidiscipl Optim",
  ["structural change and economic dynamics"] = "Struct. Change Econ. Dynam.",
  ["structural chemistry"] = "Struct. Chem.",
  ["structural concrete"] = "Struct. Concr.",
  ["structural control & health monitoring"] = "Struct Control Health Monit",
  ["structural control and health monitoring"] = "Struct. Control Health Monit.",
  ["structural design of tall and special buildings"] = "Struct. Des. Tall Special Build.",
  ["structural design of tall buildings"] = "Struct. Des. Tall Build.",
  ["structural durability & health monitoring"] = "Struct. Durability Health Monit.",
  ["structural dynamics"] = "Struct. Dyn.",
  ["structural dynamics (melville, n.y.)"] = "Struct Dyn",
  ["structural engineering and mechanics"] = "Struct. Eng. Mech.",
  ["structural engineering international"] = "Struct. Eng. Int.",
  ["structural equation modeling : a multidisciplinary journal"] = "Struct Equ Modeling",
  ["structural equation modeling. a multidisciplinary journal"] = "Struct. Equ. Model.",
  ["structural equation modelling a multidisciplinary journal"] = "Struct. Equation Modell. Multidiscip. J.",
  ["structural health monitoring"] = "Struct Health Monit",
  ["structural heart : the journal of the heart team"] = "Struct Heart",
  ["structural integrity"] = "Struct. Integr.",
  ["structural safety"] = "Struct. Saf.",
  ["structural survey"] = "Struct. Surv.",
  ["structural topology"] = "Structural Topology",
  ["structure (cambridge, ma, united states)"] = "Structure (Cambridge, MA, U. S.)",
  ["structure (london, england : 1993)"] = "Structure",
  ["structure (oxford, united kingdom)"] = "Structure (Oxford, U. K.)",
  ["structure and bonding"] = "Struct. Bond.",
  ["structure and infrastructure engineering"] = "Struct. Infrastruct. Eng.",
  ["structure with folding and design"] = "Structure Fold. Des.",
  ["stud. med"] = "Studenterraad Med",
  ["student lawyer (chicago, ill. : 1972)"] = "Stud Lawyer",
  ["student mathematical library"] = "Stud. Math. Libr.",
  ["student medicine"] = "Stud Med",
  ["student und praktikant"] = "Stud. Prakt.",
  ["students’ quarterly journal"] = "Stud. Q. J.",
  ["studi bitontini"] = "StBitont",
  ["studi classici e orientali"] = "Stud Cl Orient",
  ["studi dell’istituto gramsci siciliano"] = "Studi Ist. Gramsci Sicil.",
  ["studi di antichità. università di lecce"] = "StAnt",
  ["studi di demografia"] = "Stud Demogr",
  ["studi di egittologia e di antichità puniche"] = "StEgAntPun",
  ["studi di filologia italiana : bulletino dell' accademia della crusca"] = "Studi Filol Ital",
  ["studi di filologia italiana : bulletino dell’ accademia della crusca"] = "Studi Filol Ital",
  ["studi di medicina e chirurgia dello sport"] = "Studi Med Chir Sport",
  ["studi di sociologia"] = "Studi Sociol",
  ["studi e documenti di archeologia"] = "StDocA",
  ["studi e materiali di storia delle religioni"] = "SMSR",
  ["studi e materiali. soprintendenza ai beni archeologici per la toscana"] = "SteMat",
  ["studi e ricerche dell’ist. di civilità classica, cristiana, medievale"] = "SRIC",
  ["studi e saggi linguistici"] = "SSL",
  ["studi eblaiti"] = "StEbla",
  ["studi economici"] = "Stud. Econ.",
  ["studi emigrazione"] = "Stud Emigrazione",
  ["studi emigrazione : international journal of migration studies"] = "Studi Emigr",
  ["studi emigrazione. etudes migrations"] = "Studi Emigr",
  ["studi epigrafici e linguistici sul vicino oriente antico"] = "StEpigrLing",
  ["studi etruschi"] = "StEtr",
  ["studi etruschi. roma, giorgio bretschneider"] = "SE",
  ["studi francesi"] = "Studi Fr",
  ["studi galileiani"] = "Studi Galileiani",
  ["studi genuensi"] = "SG",
  ["studi italiani di filologia classica"] = "Studi Ital Filol Cl",
  ["studi italiani di linguistica teorica e applicata"] = "SILTA",
  ["studi latini e italiani"] = "StLatIt",
  ["studi magrebini"] = "StudMagr",
  ["studi medievali"] = "StudMed",
  ["studi medievali (turin, italy : 1928)"] = "Studi Medievali",
  ["studi micenei ed egeo-anatolici"] = "SMEA",
  ["studi piemontesi"] = "Studi Piemont",
  ["studi romagnoli"] = "StudRomagn",
  ["studi romani"] = "StRom",
  ["studi romani: rivista bimestrale dell’istituto di studi romani"] = "StudRom",
  ["studi salentini"] = "StudSal",
  ["studi sardi"] = "SS",
  ["studi sassaresi"] = "Studi Sassar",
  ["studi storici"] = "Stud Stor",
  ["studi storici: rivista trimestrale dell’ist. gramsci"] = "StudStor",
  ["studi storico-religiosi"] = "StStorRel",
  ["studi sull’oriente cristiano"] = "StOrCr",
  ["studi tardoantichi"] = "StTardoant",
  ["studi trentini di scienze naturali. acta geologica"] = "Studi Trentini Sci Nat Acta Geol",
  ["studi trentini di scienze storiche. sezione seconda"] = "StTrentStor",
  ["studi urbinati di storia, filosofia e letteratura"] = "StudUrb(B)",
  ["studi urbinati. b, scienze umane. 3, linguistica, letteratura, arte"] = "StUrbin",
  ["studi veneziani"] = "Studi Veneziani",
  ["studi. centro interuniversitario per la storia delle universita italiane"] = "Studi Cent. Interuniv. Stor. Univ. Ital.",
  ["studia (lisbon, portugal)"] = "Studia",
  ["studia aegyptiaca. budapest"] = "StAeg",
  ["studia albanica"] = "StAlb",
  ["studia archaeologica"] = "StA",
  ["studia archeologiczne. uniwersytet warszawski, instytut archeologii"] = "StAWarsz",
  ["studia asiatica (bucharest, romania)"] = "Stud Asiat",
  ["studia botanica hungarica"] = "Stud Bot Hung",
  ["studia cartesiana"] = "Stud Cartesiana",
  ["studia comeniana et historica"] = "Stud Comeniana Hist",
  ["studia demograficzne"] = "Studia Demogr",
  ["studia dipterologica"] = "Stud Dipterol",
  ["studia et documenta historiae et iuris"] = "SDHI",
  ["studia filozoficzne"] = "Stud Filoz",
  ["studia forestalia suecica"] = "Stud. for. suec.",
  ["studia geographica"] = "Stud Geogr",
  ["studia geophysica et geodaetica"] = "Stud. Geophys. Geod.",
  ["studia hibernica"] = "Stud Hibernica",
  ["studia historiae oeconomicae"] = "Stud Hist Oecon",
  ["studia historica gandensia"] = "Stud Hist Gandensia",
  ["studia historica historia antigua"] = "SHHA",
  ["studia historica. ha. contemporánea"] = "Stud Hist Ha Contemp",
  ["studia historica. ha. moderna"] = "Stud Hist Ha Mod",
  ["studia historica. historia antigua"] = "StHist",
  ["studia i materialy z dziejow nauki polskiej"] = "Stud Mater Dziejow Nauk Pol",
  ["studia i materialy z dziejow nauki polskiej. seria b: historia nauk biologicznych i medycznych"] = "Stud Mater Dziej Nauk Pol Ser B",
  ["studia i materiały z dziejów nauki polskiej"] = "Stud Mater Dziejow Nauk Pol",
  ["studia i materiały z dziejów nauki polskiej. seria b: historia nauk biologicznych i medycznych"] = "Stud Mater Dziej Nauk Pol Ser B",
  ["studia islamica"] = "Stud Islam",
  ["studia leibnitiana"] = "Stud Leibnitiana",
  ["studia linguistica"] = "Stud Linguist",
  ["studia linguistica: revue de linguistique générale et comparée"] = "SL",
  ["studia logica"] = "Stud Log",
  ["studia logica. an international journal for symbolic logic"] = "Studia Logica",
  ["studia mathematica"] = "Studia Math.",
  ["studia monastica"] = "StudMon",
  ["studia neophilologica"] = "Stud Neophilol",
  ["studia odonatologica hungarica"] = "Stud Odonatol Hung",
  ["studia oliveriana"] = "SOliv",
  ["studia orientalia, helsinki"] = "StOr",
  ["studia papyrologica"] = "Stud Papyrol",
  ["studia patavina: rivista di scienze religiose"] = "StudPat",
  ["studia philologica salmanticensia. univ. de salamanca, acta salm. filos. y letras"] = "SPhS",
  ["studia philosophica"] = "Stud. Philos.",
  ["studia philosophica et historica"] = "Stud. Philos. Hist.",
  ["studia philosophica: annuaire de la société suisse de philosophie"] = "StudPhil",
  ["studia picena"] = "StudPic",
  ["studia psychologica"] = "Stud Psychol (Bratisl)",
  ["studia quaternaria"] = "Stud. Quat.",
  ["studia rosenthaliana"] = "Stud Rosenthaliana",
  ["studia scientiarum mathematicarum hungarica"] = "Studia Sci. Math. Hungar.",
  ["studia scientiarum mathematicarum hungarica. a quarterly of the hungarian academy of sciences"] = "Studia Sci. Math. Hungar.",
  ["studia scientiarum mathematicarum hungarica. combinatorics, geometry and topology (cogeto)"] = "Studia Sci. Math. Hungar.",
  ["studia scientiarum mathematicarum hungaricae"] = "Stud. Sci. Math. Hung.",
  ["studia troica"] = "StTroica",
  ["studia universitatis babes-bolyai, chemia"] = "Stud. Univ. Babes-Bolyai, Chem.",
  ["studia universitatis babeș-bolyai. biologia"] = "Stud Univ Babes Bolyai Biol",
  ["studia z zakresu inzynierii"] = "Stud. Zakresu Inz.",
  ["studia zamorensia"] = "SZ",
  ["studia. universitatis babes-bolyai mathematica"] = "Stud. Univ. Babes-Bolyai Math.",
  ["studia. universitatis babes-bolyai. informatica"] = "Stud. Univ. Babes-Bolyai Inform.",
  ["studien und mitteilungen zur geschichte des benediktiner-ordens und seiner zweige"] = "Stud Mitt Gesch Benediktinerorden",
  ["studien zu den bogazköy-texten"] = "StBoT",
  ["studien zum buch- und bibliothekswesen"] = "SBB",
  ["studien zur altägyptischen kultur"] = "SAK",
  ["studien zur angewandten wirtschaftsforschung und statistik aus dem institut fur statistik und okonometrie der universitat hamburg"] = "Stud. Angew. Wirtschaftsforsch. Statist. Inst. Statist. Okonom. Univ. Hamburg",
  ["studien zur archäologie und geschichte altägyptens"] = "SAGA",
  ["studien zur geschichte der universität wien"] = "Stud Gesch Univ Wien",
  ["studien zur geschichte des krankenhauswesens"] = "Stud Gesch Krankenhauswesens",
  ["studien zur medizingeschichte des neunzehnten jahrhunderts"] = "Stud Medizingesch Neunzehnten Jahrhunderts",
  ["studien zur wissenschafts-, sozial- und bildungsgeschichte der mathematik"] = "Stud. Wiss. Soz. Bildungsgesch. Math.",
  ["studien zur österreichischen philosophie"] = "Stud. Österr. Philos.",
  ["studienbucher der technischen wissenschaften"] = "Studienbucher Tech. Wiss.",
  ["studienbucher mathematik"] = "Studienbucher Math.",
  ["studienbucher wirtschaftsmathematik"] = "Stud. Wirtsch.math.",
  ["studies - hastings center"] = "Stud Hastings Cent",
  ["studies and texts"] = "Stud. Texts",
  ["studies from the rockefeller institute for medical research. reprints. rockefeller institute for medical research"] = "Stud Rockefeller Inst Med Res Repr",
  ["studies from the tokugawa institute. tokugawa seibutsugaku kenkyusho, tokyo"] = "Stud Tokugawa Inst",
  ["studies from the tokugawa institute. tokugawa seibutsugaku kenkyūsho, tokyo"] = "Stud Tokugawa Inst",
  ["studies in advanced mathematics"] = "Stud. Adv. Math.",
  ["studies in ancient medicine"] = "Stud Anc Med",
  ["studies in applied mathematics"] = "Stud. Appl. Math.",
  ["studies in applied mathematics (cambridge, mass.)"] = "Stud Appl Math",
  ["studies in applied mechanics"] = "Stud. Appl. Mech.",
  ["studies in applied philosophy, epistemology and rational ethics"] = "Stud. Appl. Philos. Epistemol. Ration. Ethics",
  ["studies in behavioural adaptation"] = "Stud. Behav. Adapt.",
  ["studies in bibliography"] = "Stud Bibliogr",
  ["studies in big data"] = "Stud. Big Data",
  ["studies in burke and his time"] = "Stud Burke Time",
  ["studies in choice and welfare"] = "Stud. Choice Welf.",
  ["studies in christian ethics"] = "Stud Christ Ethics",
  ["studies in church history (chicago, ill.)"] = "Stud Chur Hist",
  ["studies in classification, data analysis, and knowledge organization"] = "Stud. Classification Data Anal. Knowledge Organ.",
  ["studies in comparative international development"] = "Stud Comp Int Dev",
  ["studies in computational intelligence"] = "Stud. Comput. Intell.",
  ["studies in computational mathematics"] = "Stud. Comput. Math.",
  ["studies in computer science and artif. intell."] = "Stud. Comput. Sci. Artif. Intell.",
  ["studies in conservation"] = "Stud. Conserv.",
  ["studies in conservation = études de conservation"] = "Stud Conserv",
  ["studies in dynamical economic science"] = "Stud. Dynam. Econom. Sci.",
  ["studies in economic design"] = "Stud. Econ. Des.",
  ["studies in economic theory"] = "Stud. Econom. Theory",
  ["studies in economics and finance"] = "Stud. Econ. Finance",
  ["studies in eighteenth-century culture"] = "Stud Eighteenth Cent Cult",
  ["studies in empirical economics"] = "Stud. Empir. Econom.",
  ["studies in english literature"] = "Stud Engl Lit",
  ["studies in ethics, law, and technology"] = "Stud Ethics Law Technol",
  ["studies in ethnicity and nationalism"] = "Stud Ethn Natl",
  ["studies in european thought"] = "Stud Eur Thought",
  ["studies in family planning"] = "Stud Fam Plann",
  ["studies in fungi"] = "Stud Fungi",
  ["studies in fuzziness"] = "Stud. Fuzziness",
  ["studies in fuzziness and soft computing"] = "Stud. Fuzziness Soft Comput.",
  ["studies in gender and sexuality"] = "Stud Gend Sex",
  ["studies in general philosophy of science"] = "Stud. Gen. Philos. Sci.",
  ["studies in health technology and informatics"] = "Stud Health Technol Inform",
  ["studies in high energy physics cosmology and gravitation"] = "Stud. High Energy Phys. Cosmol. Gravit.",
  ["studies in history"] = "Stud Hist (Sahibabad)",
  ["studies in history and philosophy of biological and biomedical sciences"] = "Stud Hist Philos Biol Biomed Sci",
  ["studies in history and philosophy of science"] = "Stud Hist Philos Sci",
  ["studies in history and philosophy of science part a"] = "Stud. Hist. Philos. Sci. Part A",
  ["studies in history and philosophy of science part b: studies in history and philosophy of modern physics"] = "Stud. Hist. Philos. Sci. Part B: Stud. Hist. Philos. Mod. Phys.",
  ["studies in history and philosophy of science part c: studies in history and philosophy of biological and biomedical sciences"] = "Stud. Hist. Philos. Sci. Part C: Stud. Hist. Philos. Biol. Biomed. Sci.",
  ["studies in history and philosophy of science. part b. studies in history and philosophy of modern physics"] = "Stud. Hist. Philos. Sci. B Stud. Hist. Philos. Modern Phys.",
  ["studies in history of biology"] = "Stud Hist Biol",
  ["studies in history of medicine"] = "Stud Hist Med",
  ["studies in history of medicine and science"] = "Stud Hist Med Sci",
  ["studies in human ecology"] = "Stud Hum Ecol",
  ["studies in infrastructure and control"] = "Stud. Infrastruct. Control",
  ["studies in intercultural philosophy"] = "Stud. Intercult. Phil.",
  ["studies in knowledge"] = "Stud. Knowledge",
  ["studies in labour history"] = "Stud Labour Hist",
  ["studies in language"] = "Stud Lang",
  ["studies in language companion series (slcs)"] = "Stud. Language Companion Ser. (SLCS)",
  ["studies in language: international journal sponsored by the foundations of language"] = "StudLang",
  ["studies in latin american popular culture"] = "Stud Lat Am Pop Cult",
  ["studies in linguistics and philosophy"] = "Stud. Linguist. Philos.",
  ["studies in logic (london)"] = "Stud. Log. (Lond.)",
  ["studies in logic and computation"] = "Stud. Logic Comput.",
  ["studies in logic and practical reasoning"] = "Stud. Log. Pract. Reason.",
  ["studies in logic and the foundations of mathematics"] = "Stud. Logic Found. Math.",
  ["studies in logic, language and information"] = "Stud. Logic Lang. Inform.",
  ["studies in mathematical and managerial economics"] = "Stud. Math. Managerial Econom.",
  ["studies in mathematical physics"] = "Stud. Math. Phys.",
  ["studies in mathematics and its applications"] = "Stud. Math. Appl.",
  ["studies in mathematics education series"] = "Stud. Math. Ed. Ser.",
  ["studies in mathematics in the arts and humanities"] = "Stud. Math. Arts Humanit.",
  ["studies in mechanobiology, tissue engineering and biomaterials"] = "Stud. Mechanobiol. Tissue Eng. Biomater.",
  ["studies in media and communication"] = "Stud Media Commun",
  ["studies in medieval culture"] = "Stud Mediev Cult",
  ["studies in mediterranean archaeology"] = "SIMA",
  ["studies in mental inefficiency"] = "Stud Ment Ineffic",
  ["studies in morphology"] = "Stud. Morphol.",
  ["studies in mycology"] = "Stud Mycol",
  ["studies in natural language and linguistic theory"] = "Stud. Nat. Lang. Linguist Theory",
  ["studies in natural products chemistry"] = "Stud. Nat. Prod. Chem.",
  ["studies in nepali history and society"] = "Stud Nepali Hist Soc",
  ["studies in neuro-anatomy"] = "Stud Neuroanat",
  ["studies in nonlinear dynamics & econometrics"] = "Stud. Nonlinear Dyn. Econom.",
  ["studies in nonlinear dynamics and econometrics"] = "Stud. Nonlinear Dyn. Econom.",
  ["studies in organic chemistry (amsterdam)"] = "Stud. Org. Chem. (Amsterdam)",
  ["studies in philology"] = "Stud Philol",
  ["studies in philosophy and the history of philosophy"] = "Stud. Philos. Hist. Philos.",
  ["studies in philosophy of medicine"] = "Stud Philos Med",
  ["studies in platonism, neoplatonism, and the platonic tradition"] = "Stud. Platonism Neoplatonism Platonic Tradit.",
  ["studies in political economy : a socialist review"] = "Stud Polit Econ",
  ["studies in profertility series"] = "Stud Profertility Ser",
  ["studies in public choice"] = "Stud. Public Choice",
  ["studies in religion (sciences religieuses)"] = "SR",
  ["studies in religion. sciences religieuses"] = "Stud Relig",
  ["studies in romanticism"] = "Stud Romanticism",
  ["studies in science education"] = "Stud Sci Educ",
  ["studies in scottish literature"] = "Stud Scott Lit",
  ["studies in second language acquisition"] = "Stud Second Lang Acquis",
  ["studies in social science research"] = "Stud Soc Sci Res",
  ["studies in soviet thought"] = "Stud Sov Thought",
  ["studies in surface science and catalysis"] = "Stud Surf Sci Catal",
  ["studies in symbolic interaction"] = "Stud Symb Interact",
  ["studies in systems, decision and control"] = "Stud. Syst. Decis. Control",
  ["studies in the american renaissance"] = "Stud Am Renaiss",
  ["studies in the decorative arts"] = "Stud Decor Arts",
  ["studies in the history and philosophy of mathematics"] = "Stud. Hist. Philos. Math.",
  ["studies in the history of gardens & designed landscapes"] = "Stud. Hist. Gardens Des. Landscapes",
  ["studies in the history of mathematical enquiry"] = "Stud. Hist. Math. Enq.",
  ["studies in the renaissance"] = "Stud Renaissance",
  ["studies in the social sciences (west georgia college)"] = "Stud Soc Sci West Ga Coll",
  ["studies in theoretical and applied statistics. selected papers of the statistical societies"] = "Stud. Theor. Appl. Stat. Sel. Papers Stat. Soc.",
  ["studies in theoretical philosophy"] = "Stud. Theor. Philos.",
  ["studies in theoretical psycholinguistics"] = "Stud. Theor. Psycholinguist.",
  ["studies in third world societies"] = "Stud Third World Soc",
  ["studies in universal logic"] = "Stud. Univers. Log.",
  ["studies of nonlinear phenomena in life science"] = "Stud. Nonlinear Phenom. Life Sci.",
  ["studies of the university of transport and communications in žilina"] = "Stud. Univ. Transp. Commun. Žilina Math.-Phys. Ser.",
  ["studies of the university of zilina. mathematical series"] = "Stud. Univ. Zilina Math. Ser.",
  ["studies of university in žilina"] = "Stud. Univ. Žilina Math. Phys. Ser.",
  ["studies on medical and population subjects"] = "Stud Med Popul Subj",
  ["studies on neotropical fauna and environment"] = "Stud Neotrop Fauna Environ",
  ["studies on russian economic development"] = "Stud Russ Econ Dev",
  ["studies on the semantic web"] = "Stud. Semant. Web",
  ["studies on tropical andean ecosystems"] = "Stud. Trop. Andean Ecosyst.",
  ["studies on voltaire and the eighteenth century"] = "Stud Voltaire 18th Century",
  ["studies on voltaire and the eighteenth century / edited by theodore besterman"] = "Stud Voltaire 18th Century",
  ["studies. institutum divi thomae"] = "Stud Inst Divi Thomae",
  ["studii clasice"] = "Stud Cl",
  ["studii clasice. societatea de studii clasice din republica socialistă romănia"] = "StCl",
  ["studii si cercetari de endocrinologie"] = "Stud. Cercet. Endocrinol.",
  ["studii si cercetari de fiziologie"] = "Stud. Cercet. Fiziol.",
  ["studii si cercetari de inframicrobiologie"] = "Stud. Cercet. Inframicrobiol.",
  ["studii si cercetari de medicina interna"] = "Stud. Cercet. Med. Interna",
  ["studii si cercetari de neurologie"] = "Stud. Cercet. Neurol.",
  ["studii si cercetari de virusologie"] = "Stud. Cercet. Virusol.",
  ["studii si cercetari economice"] = "Stud. Cercet. Econ.",
  ["studii si cercetări de neurologie"] = "Stud Cercet Neurol",
  ["studii și materiale de istorie medie. institutul de istorie (academia republicii populare romîne)"] = "Stud Mater Ist Medie",
  ["studii şi cercetari de endocrinologie"] = "Stud Cercet Endocrinol",
  ["studii şi cercetări de fiziologie"] = "Stud Cercet Fiziol",
  ["studii şi cercetări de inframicrobiologie"] = "Stud Cercet Inframicrobiol",
  ["studii şi cercetări de istorie veche şi arheologie"] = "StCercIstorV",
  ["studii şi cercetări de medicină internă"] = "Stud Cercet Med Interna",
  ["studii şi cercetări de numismatică"] = "StCercNum",
  ["studii şi cercetări de virusologie"] = "Stud Cercet Virusol",
  ["studii şi cercetări de mecanică aplicată"] = "Stud. Cerc. Mec. Apl.",
  ["studii şi cercetări matematice"] = "Stud. Cerc. Mat.",
  ["studime historike"] = "Studime Hist",
  ["studium (rotterdam, netherlands)"] = "Studium (Rotterdam)",
  ["studium biblicum franciscanum. liber annuus"] = "StBiFranc",
  ["studium generale"] = "Stud. Gen. (Berl.)",
  ["studium generale; zeitschrift fur die einheit der wissenschaften im zusammenhang ihrer begriffsbildungen und forschungsmethoden"] = "Stud Gen (Berl)",
  ["studium generale; zeitschrift für die einheit der wissenschaften im zusammenhang ihrer begriffsbildungen und forschungsmethoden"] = "Stud Gen (Berl)",
  ["studium ovetense: revista del centro sup. de estudios eclesiásticos la asunción"] = "StudOv",
  ["study encounter"] = "Study Encounter",
  ["study sessions - canadian catholic historical association. canadian catholic historical association"] = "Study Sess Can Cathol Hist Assoc",
  ["stuttgarter beitraege zur naturkunde a"] = "Stuttgart. Beitr. Naturkd. A",
  ["stuttgarter beiträge zur naturkunde"] = "Stuttg. Beitr. Naturkd.",
  ["stylus: cuadernos de filología"] = "Stylus",
  ["sub-cellular biochemistry"] = "Subcell Biochem",
  ["subacute care"] = "Subacute Care",
  ["subj. biochimica et biophysica acta, specialized section on nucleic acids and related subjects"] = "Biochim. Biophys. Acta, Spec. Sect. Nucleic Acids Relat.",
  ["subsea control and data acquisition for oil and gas production systems"] = "Adv. Underwat. Technol.",
  ["subseries in control"] = "Subser. Control",
  ["subsidia medica"] = "Subsid Med",
  ["substance abuse"] = "Subst Abus",
  ["substance abuse : official publication of the association for medical education and research in substance abuse"] = "Subst Abus",
  ["substance abuse : research and treatment"] = "Subst Abuse",
  ["substance abuse and rehabilitation"] = "Subst Abuse Rehabil",
  ["substance abuse treatment, prevention, and policy"] = "Subst Abuse Treat Prev Policy",
  ["substance and alcohol actions/misuse"] = "Subst Alcohol Actions Misuse",
  ["substance use & misuse"] = "Subst Use Misuse",
  ["substance use and misuse"] = "Subst. Use Misuse",
  ["subterranean biology"] = "Subterr Biol",
  ["sud-est médical"] = "Sudest Med",
  ["sudan & nubia : the sudan archaeological research society bulletin"] = "Sudan Nubia",
  ["sudan journal of population studies"] = "Sudan J Popul Stud",
  ["sudan medical journal"] = "Sudan Med J",
  ["sudan notes and records"] = "Sudan Notes Rec",
  ["sudanese journal of paediatrics"] = "Sudan J Paediatr",
  ["suddeutsche optikerzeitung"] = "Suddtsch Optikerztg",
  ["sudebno-meditsinskaia ekspertiza"] = "Sud Med Ekspert",
  ["sudhoffs archiv"] = "Sudhoffs Arch",
  ["sudhoffs archiv fur geschichte der medizin und der naturwissenschaften"] = "Sudhoffs Arch Gesch Med Naturwiss",
  ["sudhoffs archiv für geschichte der medizin und der naturwissenschaften"] = "Sudhoffs Arch Gesch Med Naturwiss",
  ["sudhoffs archiv für geschichte der medizin"] = "Sudhoffs Arch. Gesch. Med.",
  ["sudhoffs archiv für geschichte der medizin und der naturwissenschaften"] = "Sudhoffs Arch. Gesch. Med. Naturwiss.",
  ["sudhoffs archiv. vierteljahresschrift fur geschichte der medizin und der naturwissenschaften, der pharmazie und der mathematik"] = "Sudhoffs Arch., Vierteljahresschr. fur Gesch. Med. Naturwiss. Pharm. Math.",
  ["sudhoffs archiv. zeitschrift für wissenschaftsgeschichte"] = "Sudhoffs Arch., Z. Wissenschaftsgesch.",
  ["sudhoffs archiv: zeitschrift für wissenschaftsgeschichte"] = "ZWG",
  ["sudhoffs archiv; vierteljahrsschrift fur geschichte der medizin und der naturwissenschaften, der pharmazie und der mathematik"] = "Sudhoffs Arch. Vierteljahrsschr. Gesch. Med. Naturwiss. Pharm. Math.",
  ["sudhoffs archiv; zeitschrift fur wissenschaftsgeschichte"] = "Sudhoffs Arch. Z. Wissenschaftsgesch.",
  ["sudhoffs archiv; zeitschrift fur wissenschaftsgeschichte. beihefte"] = "Sudhoffs Arch Z Wissenschaftsgesch Beih",
  ["sudhoffs archiv; zeitschrift für wissenschaftsgeschichte. beihefte"] = "Sudhoffs Arch Z Wissenschaftsgesch Beih",
  ["sudhoffs klassiker der medizin (leipzig, germany : 1981)"] = "Sudhoffs Klass Med",
  ["sudost europa : [monatsschrift der abteilung gegenwartsforschung des sudost-instituts]"] = "Sudosteuropa",
  ["suffield technical note. canada. suffield experimental station, ralston, alta"] = "Suffield Tech Note Can Suffield Exp Stn Ralston Alta",
  ["suffield technical paper. canada. suffield experimental station, ralston, alta"] = "Suffield Tech Pap Can Suffield Exp Stn Ralston Alta",
  ["suffolk transnational law review"] = "Suffolk Transnatl Law R",
  ["suffolk university law review"] = "Suffolk Univ Law Rev",
  ["sugaku expositions"] = "Sugaku Expositions",
  ["sugar tech"] = "Sugar Tech",
  ["sugar tech : an international journal of sugar crops & related industries"] = "Sugar Tech",
  ["suhayl. international journal for the history of the exact and natural sciences in islamic civilisation"] = "Suhayl",
  ["suicide & life-threatening behavior"] = "Suicide Life Threat Behav",
  ["suicide and life-threatening behavior"] = "Suicide Life. Threat. Behav.",
  ["suicidology online"] = "Suicidol Online",
  ["suid-afrikaanse historiese joernaal"] = "S Afr Hist J",
  ["suid-afrikaanse tydskrif vir ekonomiese en bestuurswetenskappe"] = "Suid-Afrikaanse Tydskr. Ekon. Bestuurswetenskappe",
  ["suid-afrikaanse tydskrif vir strafreg en kriminologie = south african journal of criminal law and criminology"] = "S Afr Tydskr Strafr Kriminol",
  ["suisan ikushu"] = "Suisan Ikushu",
  ["suisan sōgō kenkyū sentā kenkyū hōkoku = bulletin of fisheries research agency"] = "Suisan Sogo Kenkyu Senta Kenkyu Hokoku",
  ["suisan zōshoku"] = "Suisan Zoshoku",
  ["sujil bojeon"] = "Sujil Bojeon",
  ["sulfur letters"] = "Sulfur Lett.",
  ["sultan qaboos university medical journal"] = "Sultan Qaboos Univ Med J",
  ["sumer. a journal of archaeology (and history) in iraq"] = "Sumer",
  ["sumerianz journal of medical and healthcare"] = "Sumer J Med Healthc",
  ["summa phytopathológica"] = "Summa Phytopathol",
  ["summer computer simulation conference : (scsc 2014) : 2014 summer simulation multi-conference : monterey, california, usa, 6-10 july 2014. summer computer simulation conference (2014 : monterey, calif.)"] = "Summer Comput Simul Conf (2014)",
  ["summer computer simulation conference : (scsc 2015) : 2015 summer simulation multi-conference (summersim'15) : chicago, illinois, usa, 26-29 july 2015. summer computer simulation conference (2015 : chicago, illinois)"] = "Summer Comput Simul Conf (2015)",
  ["summit on translational bioinformatics"] = "Summit Transl Bioinform",
  ["sums readings"] = "SUMS Read.",
  ["sun (baltimore, md. : 1837)"] = "Sun",
  ["sunday times (london, england : 1931)"] = "Sunday Times",
  ["suny series in ancient greek philosophy"] = "SUNY Ser. Anc. Greek Philos.",
  ["suny series in logic and language"] = "SUNY Ser. Logic Lang.",
  ["suny series in science, technology, and society"] = "SUNY Ser. Sci. Tech. Soc.",
  ["suny series, reform in mathematics education"] = "SUNY Ser. Reform Math. Educ.",
  ["suomalainen tiedeakatemia toimituksia. sar. a.4: biologica"] = "Ann Acad Sci Fenn Biol",
  ["suomalaisen tiedeakatemian toimituksia. sar. a.5, medica-anthropologica"] = "Suom Tiedeakat Toim A5 Medica Anthropol",
  ["suomen apteekkarilehti. finlands apotekartidning"] = "Suom Apteenkkaril",
  ["suomen hammaslaakarilehti"] = "Suom. Hammaslaakarilehti",
  ["suomen hammaslaakariseuran toimituksia"] = "Suom. Hammaslaak. Toim.",
  ["suomen hammaslääkärilehti = finlands tandläkartidning"] = "Suom Hammaslaakarilehti",
  ["suomen hammaslääkäriseuran toimituksia = finska tandläkarsällskapets förhandlingar"] = "Suom Hammaslaak Toim",
  ["suomen laakarilehti"] = "Suom. Laakaril.",
  ["suomen laakarilehti. finlands lakartidning"] = "Suom Laakaril",
  ["suomen lääkärilehti. finlands läkartidning"] = "Suom Laakaril",
  ["suomen naishammaslaakarit ryhma julkaisu"] = "Suom. Naishammaslaak. Julk.",
  ["suomen ortopedia ja traumatologia = ortopedi och traumatologi i finland = finnish journal of orthopaedics and traumatology"] = "Suom Ortoped Traumatol",
  ["supercomputing frontiers and innovations"] = "Supercomput Front Innov",
  ["superconductor science & technology"] = "Supercond Sci Technol",
  ["superconductor science and technology"] = "Supercond. Sci. Technol.",
  ["superlattices and microstructures"] = "Superlattices Microstruct.",
  ["supervisor nurse"] = "Superv Nurse",
  ["supervisory management"] = "Superv Manage",
  ["suplemento de la revista mexicana de fã­sica"] = "Supl. Rev. Mex. Fis",
  ["supplement to international journal of gynecology and obstetrics"] = "Suppl Int J Gynecol Obstet",
  ["supplement to the european journal neuroscience"] = "Suppl. Eur. J. Neurosci.",
  ["supplementi di tumori"] = "Suppl. Tumori",
  ["supplements to clinical neurophysiology"] = "Suppl Clin Neurophysiol",
  ["supplements to the bulletin of the history of medicine"] = "Bull Hist Med Suppl",
  ["supplementum ad thrombosis and haemostasis"] = "Suppl. Thromb. Haemost.",
  ["supplementum epigraphicum graecum"] = "SEG",
  ["supplementum orientale. corpus medicorum graecorum"] = "Corpus Med Graecorum Suppl Orient",
  ["supportive cancer therapy"] = "Support Cancer Ther",
  ["supportive care in cancer"] = "Support. Care Cancer",
  ["supportive care in cancer : official journal of the multinational association of supportive care in cancer"] = "Support Care Cancer",
  ["supramolecular chemistry"] = "Supramol. Chem.",
  ["supramolecular materials"] = "Supramol. Mater.",
  ["supramolecular photosensitive and electroactive materials"] = "Supramol. Photosensit. Electroact. Mater.",
  ["supramolecular science"] = "Supramol. Sci.",
  ["supreme court economic review"] = "Supreme Court Econ. Rev.",
  ["surface & coatings technology"] = "Surf Coat Technol",
  ["surface and coatings technology"] = "Surf. Coat. Technol.",
  ["surface and interface analysis"] = "Surf. Interface Anal.",
  ["surface and interface analysis : sia"] = "Surf Interface Anal",
  ["surface coatings international"] = "Surf. Coat. Int.",
  ["surface coatings international part a: coatings journal"] = "Surf. Coat. Int.: Part A",
  ["surface coatings international part b: coatings transactions"] = "Surf. Coat. Int.: Part B",
  ["surface engineering"] = "Surf. Eng.",
  ["surface engineering and applied electrochemistry"] = "Surf. Eng. Appl. Electrochem.",
  ["surface innovations"] = "Surf Innov",
  ["surface physics"] = "Ccast. Wl. Sw.",
  ["surface review and letters"] = "Surf. Rev. Lett.",
  ["surface science"] = "Surf. Sci.",
  ["surface science letters"] = "Surf. Sci. Lett.",
  ["surface science reports"] = "Surf. Sci. Rep.",
  ["surface science spectra"] = "Surf. Sci. Spectra",
  ["surface technology"] = "Surf. Technol.",
  ["surface topography : metrology and properties"] = "Surf Topogr",
  ["surface topography: metrology and properties"] = "Surf. Topogr.: Metrol. Prop.",
  ["surfaces and interfaces"] = "Surf. Interfaces",
  ["surgeon's circular letter. united states. far east command. medical section"] = "Surg Circ Lett U S Army Far East Command Med Sect",
  ["surgery (oxford, oxfordshire)"] = "Surgery (Oxf)",
  ["surgery and medical technologies"] = "Surg. Med. Technol.",
  ["surgery annual"] = "Surg Annu",
  ["surgery for obesity and related diseases"] = "Surg. Obes. Relat. Dis.",
  ["surgery for obesity and related diseases : official journal of the american society for bariatric surgery"] = "Surg Obes Relat Dis",
  ["surgery journal (new york, n.y.)"] = "Surg J (N Y)",
  ["surgery open science"] = "Surg Open Sci",
  ["surgery research and practice"] = "Surg Res Pract",
  ["surgery today"] = "Surg Today",
  ["surgery, gynecology & obstetrics"] = "Surg Gynecol Obstet",
  ["surgery, gynecology and obstetrics"] = "Surg. Gynecol. Obstet.",
  ["surgical and experimental pathology"] = "Surg. Exp. Pathol.",
  ["surgical and radiologic anatomy"] = "Surg. Radiol. Anat.",
  ["surgical and radiologic anatomy : sra"] = "Surg Radiol Anat",
  ["surgical business"] = "Surg Bus",
  ["surgical case reports"] = "Surg Case Rep",
  ["surgical case reports (tallinn, estonia)"] = "Surg Case Rep (Tallinn)",
  ["surgical clinics of north america"] = "Surg. Clin. North Am.",
  ["surgical endoscopy"] = "Surg Endosc",
  ["surgical equipment"] = "Surg Equip",
  ["surgical forum"] = "Surg Forum",
  ["surgical gastroenterology"] = "Surg Gastroenterol",
  ["surgical infections"] = "Surg Infect (Larchmt)",
  ["surgical innovation"] = "Surg Innov",
  ["surgical laparoscopy & endoscopy"] = "Surg Laparosc Endosc",
  ["surgical laparoscopy and endoscopy"] = "Surg. Laparosc. Endosc.",
  ["surgical laparoscopy, endoscopy & percutaneous techniques"] = "Surg Laparosc Endosc Percutan Tech",
  ["surgical laparoscopy, endoscopy and percutaneous techniques"] = "Surg. Laparosc. Endosc. Percutan. Tech.",
  ["surgical neurology"] = "Surg Neurol",
  ["surgical neurology international"] = "Surg Neurol Int",
  ["surgical oncology"] = "Surg Oncol",
  ["surgical oncology clinics of north america"] = "Surg Oncol Clin N Am",
  ["surgical pathology clinics"] = "Surg Pathol Clin",
  ["surgical practice"] = "Surg Pract",
  ["surgical research communications"] = "Surg Res Commun",
  ["surgical rounds"] = "Surg Rounds",
  ["surgical science"] = "Surg Sci",
  ["surgical staff seminars. united states. veterans administration. hospital, minneapolis"] = "Surg Staff  Semin U S Veterans Adm Hosp Minneap",
  ["surgical technologist"] = "Surg. Technol.",
  ["surgical technology international"] = "Surg Technol Int",
  ["surgo; glasgow university medical journal"] = "Surgo Glasg Univ Med J",
  ["surveillance & society"] = "Surveill Soc",
  ["surveillance efficace"] = "Surveil Effic",
  ["survey (london, england : 1961)"] = "Survey (Lond)",
  ["survey and synthesis of pathology research"] = "Surv Synth Pathol Res",
  ["survey graphic"] = "Surv Graph",
  ["survey methodology"] = "Surv Methodol",
  ["survey methods - insights from the field"] = "Surv Methods Insights Field",
  ["survey of biological progress"] = "Surv Biol Prog",
  ["survey of current business"] = "Surv. Curr. Bus.",
  ["survey of immunologic research"] = "Surv Immunol Res",
  ["survey of ophthalmology"] = "Surv Ophthalmol",
  ["survey practice"] = "Surv Pract",
  ["survey research methods"] = "Surv Res Methods",
  ["survey review"] = "Surv. Rev.",
  ["surveys and tutorials in the applied mathematical sciences"] = "Surv. Tutor. Appl. Math. Sci.",
  ["surveys in applied mathematics"] = "Surveys Appl. Math.",
  ["surveys in approximation theory"] = "Surv. Approx. Theory",
  ["surveys in differential geometry"] = "Surv. Differ. Geom.",
  ["surveys in geophysics"] = "Surv. Geophys.",
  ["surveys in high energy physics"] = "Surv. High Energy Phys.",
  ["surveys in mathematics and its applications"] = "Surv. Math. Appl.",
  ["surveys in operations research and management science"] = "Surv. Oper. Res. Manag. Sci.",
  ["surveys of modern mathematics"] = "Surv. Mod. Math.",
  ["surveys on mathematics for industry"] = "Surveys Math. Indust.",
  ["survival news (atlanta, ga.)"] = "Surviv News (Atlanta Ga)",
  ["susan haeyang gyoyuk yeon-gu"] = "Susan Haeyang Gyoyuk Yeongu",
  ["suspop news : the sustainable population society newsletter"] = "SusPop News",
  ["sustainability & circularity now"] = "Sustainability Circularity NOW",
  ["sustainability (new rochelle, n.y.)"] = "Sustainability (New Rochelle)",
  ["sustainability analytics and modeling"] = "Sustainability Anal. Model.",
  ["sustainability and climate change"] = "Sustainability Clim. Change",
  ["sustainability management forum"] = "Sustainability Manage. Forum",
  ["sustainability of water quality and ecology"] = "Sustainability Water Qual. Ecol.",
  ["sustainability science"] = "Sustainability Sci.",
  ["sustainability water quality and ecology"] = "Sustainability Water Qual. Ecol.",
  ["sustainability: science, practice and policy"] = "Sustainability: Sci., Pract. Policy",
  ["sustainable and resilient infrastructure"] = "Sustain Resilient Infrastruct",
  ["sustainable buildings"] = "Sustainable Build.",
  ["sustainable chemical processes"] = "Sustainable Chem. Processes",
  ["sustainable chemistry and pharmacy"] = "Sustain Chem Pharm",
  ["sustainable chemistry for climate action"] = "Sustainable Chem. Clim. Action",
  ["sustainable cities and society"] = "Sustain Cities Soc",
  ["sustainable computing informatics & systems"] = "Sustainable Comput. Inf. Syst.",
  ["sustainable development"] = "Sustainable Dev.",
  ["sustainable earth"] = "Sustainable Earth",
  ["sustainable energy & fuels"] = "Sustain Energy Fuels",
  ["sustainable energy research"] = "Sustainable Energy Res.",
  ["sustainable energy technologies and assessments"] = "Sustainable Energy Technol. Assess.",
  ["sustainable energy, grids and networks"] = "Sustainable Energy Grids Networks",
  ["sustainable environment"] = "Sustainable Environ.",
  ["sustainable environment research"] = "Sustainable Environ. Res.",
  ["sustainable food proteins"] = "Sustainable Food Proteins",
  ["sustainable food technology"] = "Sustainable Food Technol.",
  ["sustainable futures"] = "Sustainable Futures",
  ["sustainable horizons"] = "Sustainable Horiz.",
  ["sustainable manufacturing and service economics"] = "Sustainable Manuf. Serv. Econ.",
  ["sustainable materials and technologies"] = "Sustainable Mater.Technol.",
  ["sustainable operations and computers"] = "Sustainable Oper. Comput.",
  ["sustainable production and consumption"] = "Sustain Prod Consum",
  ["sustainable technology and entrepreneurship"] = "Sustainable Technol. Entrepreneurship",
  ["sustainable water resources management"] = "Sustain Water Resour Manag",
  ["sut journal of mathematics"] = "SUT J. Math.",
  ["suŭi kawahak yon'gu nonmunjip"] = "Suui Kawahak Yongu Nonmunjip",
  ["suvremenna meditsina"] = "Suvr. Med. (Sofiia)",
  ["sŭvremenna medit︠s︡ina"] = "Suvr Med (Sofiia)",
  ["süddeutsche apotheker-zeitung"] = "Suddeutsch Apoth Ztg",
  ["süddeutsche optikerzeitung"] = "Suddtsch Optikerztg",
  ["südost europa : [monatsschrift der abteilung gegenwartsforschung des südost-instituts]"] = "Sudosteuropa",
  ["südwestdeutsches ärzteblatt"] = "Sudwestdtsch Arztebl",
  ["svarochnoe proizvodstvo"] = "Svar. Proizvod.",
  ["sveikatos apsauga"] = "Sveikatos Apsauga",
  ["svensk botanisk tidskrift"] = "Sven. Bot. Tidskr.",
  ["svensk farmaceutisk tidskrift"] = "Sven Farm Tidskr",
  ["svensk geografisk arsbok"] = "Sven Geogr Arsb",
  ["svensk geografisk årsbok"] = "Sven Geogr Arsb",
  ["svensk juristtidning"] = "Sven Juristtidn",
  ["svensk medicinhistorisk tidskrift"] = "Sven Med Tidskr",
  ["svensk mykologisk tidskrift"] = "Sven Mykol Tidskr",
  ["svensk papperstidning"] = "Sven. Papperstidn.",
  ["svensk sjukkasse-tidning"] = "Svensk Sjukkasse Tidn",
  ["svensk tandlakare tidskrift. swedish dental journal"] = "Sven Tandlak Tidskr",
  ["svensk tandlakaretidskrift"] = "Sven. Tandlak. Tidskr.",
  ["svensk tandläkare tidskrift. swedish dental journal"] = "Sven Tandlak Tidskr",
  ["svensk tidskrift (uppsala, sweden)"] = "Sven Tidskr",
  ["svenska lakartidningen"] = "Sven Lakartidn",
  ["svenska läkartidningen"] = "Sven Lakartidn",
  ["svenska linnésällskapets årsskrift = yearbook of the swedish linnaeus society"] = "Sven Linnesallskap Arsskr",
  ["sveriges tandlakarforbund tidning"] = "Sver. Tandlakarforb. Tidn.",
  ["sverkhprovodimost: fizika, khimiya, tekhnika"] = "Sverkhprovodimost: Fiz., Khim., Tekh.",
  ["svobodnai︠a︡ myslʹ (moscow, russia : 1991)"] = "Svob Misl",
  ["svu international journal of basic sciences"] = "SVU Int J Basic Sci",
  ["swarm and evolutionary computation"] = "Swarm Evol. Comput.",
  ["swarm intelligence"] = "Swarm Intell.",
  ["swasth hind"] = "Swasth Hind",
  ["swedish dental journal"] = "Swed Dent J",
  ["swedish dental journal. supplement"] = "Swed Dent J Suppl",
  ["swedish economic policy review"] = "Swedish Econ. Pol. Rev.",
  ["swedish journal of economics"] = "Swedish J. Econ.",
  ["swiss biotech"] = "Swiss Biotech",
  ["swiss dent"] = "Swiss Dent",
  ["swiss dental journal"] = "Swiss Dent J",
  ["swiss journal of economics and statistics"] = "Swiss J Econ Stat",
  ["swiss journal of geosciences"] = "Swiss J. Geosci.",
  ["swiss journal of palaeontology"] = "Swiss J Palaeontol",
  ["swiss journal of psychology : official publication of the swiss psychological society schweizerische zeitschrift für psychologie = revue suisse de psychologie"] = "Swiss J Psychol",
  ["swiss medical weekly"] = "Swiss Med Wkly",
  ["swiss medical weekly : official journal of the swiss society of infectious diseases, the swiss society of internal medicine, the swiss society of pneumology"] = "Swiss Med Wkly",
  ["swiss political science review"] = "Swiss Polit. Sci. Rev.",
  ["swiss surgery"] = "Swiss Surg.",
  ["swiss surgery = schweizer chirurgie = chirurgie suisse = chirurgia svizzera"] = "Swiss Surg",
  ["swiss surgery. supplement"] = "Swiss Surg Suppl",
  ["sxi—springer per l’innovazione/sxi—springer for innovation"] = "Sxi Springer Innov.",
  ["sydney university medical journal"] = "Sydn Univ Med J",
  ["sydsvenska medicinhistoriska sallskapets arsskrift"] = "Sydsven Medicinhist Sallsk Arsskr",
  ["sydsvenska medicinhistoriska sallskapets arsskrift. supplementum"] = "Sydsven Medicinhist Sallsk Arsskr Suppl",
  ["sydsvenska medicinhistoriska sällskapets årsskrift"] = "Sydsven Medicinhist Sallsk Arsskr",
  ["sydsvenska medicinhistoriska sällskapets årsskrift. supplementum"] = "Sydsven Medicinhist Sallsk Arsskr Suppl",
  ["sykepleien fag"] = "Sykepl. Fag",
  ["sykepleien. fag"] = "Sykepl Fag",
  ["syllecta classica"] = "SyllClass",
  ["sylloge nummorum graecorum"] = "SNG",
  ["sylva mala. bollettino del centro di studi archeologici di boscoreale, boscotrecase e trecase"] = "SylvaMala",
  ["symbiosis (philadelphia, pa.)"] = "Symbiosis",
  ["symbolae osloenses"] = "Symb Oslo",
  ["symbolae osloenses / auspiciis societatis graeco-latinae"] = "Symb Oslo",
  ["symbolae osloenses, auspiciis societatis graeco-latine"] = "SO",
  ["symbolae philologorum posnanensium"] = "SPhP",
  ["symbolic and numeric computation series"] = "Symbol. Numer. Comput. Ser.",
  ["symbolic computation. computer graphics"] = "Symbol. Comput. Comput. Graph.",
  ["symbolic interaction"] = "Symb Interact",
  ["symmetry: culture and science"] = "Symmetry Cult. Sci.",
  ["symposia gaussiana"] = "Sympos. Gaussiana",
  ["symposia mathematica"] = "Sympos. Math.",
  ["symposia of the society for experimental biology"] = "Symp Soc Exp Biol",
  ["symposia of the swedish nutrition foundation"] = "Symp Swed Nutr Found",
  ["symposium (international) on combustion"] = "Symp. (Int.) Combust.",
  ["symposium (international) on combustion, [proceedings]"] = "Symp. (Int.) Combust., [Proc.]",
  ["symposium - international astronomical union"] = "Symp. - Int. Astron. Union",
  ["symposium international astronomical union"] = "Symp. Int. Astron. Union",
  ["symposium of the faraday society"] = "Symp. Faraday Soc.",
  ["symposium of the pharmacology, therapeutics and toxicology group, international association for dental research"] = "Symp Pharmacol Ther Toxicol Group",
  ["symposium of the society for developmental biology"] = "Symp. Soc. Dev. Biol.",
  ["symposium on fundamental cancer research"] = "Symp Fundam Cancer Res",
  ["symposium on geometry processing : [proceedings]. symposium on geometry processing"] = "Symp Geom Process",
  ["symposium on oral sensation and perception"] = "Symp Oral Sens Percept",
  ["symposium on theory of modeling & simulation : devs integrative m&s symposium (tms-devs). symposium on theory of modeling and simulation"] = "Symp Theory Model Simul",
  ["symposium on vlsi circuits : [proceedings]. symposium on vlsi circuits"] = "Symp VLSI Circuits",
  ["symposium series (society for applied microbiology)"] = "Symp Ser Soc Appl Microbiol",
  ["symposium series, society for applied microbiology"] = "Symp. Ser. Soc. Appl. Microbiol.",
  ["synapse (hoboken, nj, united states)"] = "Synapse (Hoboken, NJ, U. S.)",
  ["synapse (new york, n.y.)"] = "Synapse",
  ["synapse (new york, ny, united states)"] = "Synapse (N. Y., NY, U. S.)",
  ["synchrotron radiation news"] = "Synchrotron Radiat News",
  ["synergie—syntropie—nichtlineare systeme"] = "Synerg. Syntropie Nichtlineare Syst.",
  ["synergist (akron, ohio)"] = "Synergist (Akron)",
  ["synlett : accounts and rapid communications in synthetic organic chemistry"] = "Synlett",
  ["syntax (oxford, england)"] = "Syntax",
  ["synthese library"] = "Synthese Lib.",
  ["synthese library. studies in epistemology, logic, methodology, and philosophy of science"] = "Synth. Libr.",
  ["synthese. an international journal for epistemology, methodology and philosophy of science"] = "Synthese",
  ["synthesis and reactivity in inorganic and metal-organic chemistry"] = "Synth. React. Inorg. Met.-Org. Chem.",
  ["synthesis and reactivity in inorganic, metal-organic, and nano-metal chemistry"] = "Synth. React. Inorg., Met.-Org., Nano-Met. Chem.",
  ["synthesis lectures on algorithms and software in engineering"] = "Synth. Lect. Algorithms Softw. Eng.",
  ["synthesis lectures on artificial intelligence and machine learning"] = "Synth. Lect. Artif. Intell. Mach. Learn.",
  ["synthesis lectures on computer science"] = "Synth. Lect. Comput. Sci.",
  ["synthesis lectures on computer vision"] = "Synth. Lect. Comput. Vis.",
  ["synthesis lectures on data mining and knowledge discovery"] = "Synth. Lect. Data Min. Knowl. Discov.",
  ["synthesis lectures on data, semantics, and knowledge"] = "Synth. Lect. Data Semant. Knowl.",
  ["synthesis lectures on digital circuits and systems"] = "Synth. Lect. Digit. Circuits Syst.",
  ["synthesis lectures on engineering, science, and technology"] = "Synth. Lect. Eng. Sci. Technol.",
  ["synthesis lectures on games and computational intelligence"] = "Synth. Lect. Games Comput. Intell.",
  ["synthesis lectures on human language technologies"] = "Synth. Lect. Hum. Lang. Technol.",
  ["synthesis lectures on intelligent technologies"] = "Synth. Lect. Intell. Technol.",
  ["synthesis lectures on learning, networks, and algorithms"] = "Synth. Lect. Learn. Netw. Algorithms",
  ["synthesis lectures on mathematics and statistics"] = "Synth. Lect. Math. Stat.",
  ["synthesis lectures on mechanical engineering"] = "Synth. Lect. Mech. Eng.",
  ["synthesis lectures on operations research and applications"] = "Synth. Lect. Oper. Res. Appl.",
  ["synthesis lectures on theoretical computer science"] = "Synth. Lect. Theor. Comput. Sci.",
  ["synthesis lectures on visual computing"] = "Synth. Lect. Vis. Comput.",
  ["synthesis lectures on wave phenomena in the physical sciences"] = "Synth. Lect. Wave Phenom. Phys. Sci.",
  ["synthesis philosophica"] = "Synth Philos",
  ["synthetic and systems biotechnology"] = "Synth Syst Biotechnol",
  ["synthetic biology"] = "Synth. Biol.",
  ["synthetic biology (oxford, england)"] = "Synth Biol (Oxf)",
  ["synthetic communications"] = "Synth. Commun.",
  ["synthetic fibers"] = "Synth. Fibres",
  ["synthetic lubrication"] = "Synth. Lubr.",
  ["synthetic metals"] = "Synth. Met.",
  ["syracuse law & civic engagement"] = "Syracuse Law Civ Engagem",
  ["syracuse law review"] = "Syracuse Law Rev",
  ["syria. revue d’art oriental et d’archéologie"] = "Syria",
  ["syro-mesopotamian studies"] = "SyrMesopSt",
  ["system & control letters"] = "Syst. Control Lett.",
  ["system development foundation benchmark series"] = "System Develop. Found. Benchmark Ser.",
  ["system dynamics review"] = "Syst Dyn Rev",
  ["systema ascomycetum"] = "Systema Ascomycetum",
  ["systematic and applied acarology"] = "Syst Appl Acarol",
  ["systematic and applied microbiology"] = "Syst Appl Microbiol",
  ["systematic biology"] = "Syst Biol",
  ["systematic botany"] = "Syst Bot",
  ["systematic botany monographs : monographic series of the american society of plant taxonomists"] = "Syst Bot Monogr",
  ["systematic entomology"] = "Syst Entomol",
  ["systematic parasitology"] = "Syst Parasitol",
  ["systematic reviews"] = "Syst Rev",
  ["systematic reviews in pharmacy"] = "Syst. Rev. Pharm.",
  ["systematic reviews, evidence for action : abstracts for workshops and scienitific [sic] sessions. international cochrane colloquium (6th : 1998 : baltimore, md.)"] = "Syst Rev Evid Action Int Cochrane Colloq 6th 1998 Baltim Md",
  ["systematic zoology"] = "Syst Zool",
  ["systematics and biodiversity"] = "Syst. Biodivers.",
  ["systematics and geography of plants"] = "Syst Geogr Plants",
  ["systematics association special volume"] = "Syst Assoc Spec Vol",
  ["systemic practice and action research"] = "Syst Pract Action Res",
  ["systems & control letters"] = "Syst Control Lett",
  ["systems & control: foundations & applications"] = "Systems Control Found. Appl.",
  ["systems and soft computing"] = "Syst. Soft Comput.",
  ["systems and synthetic biology"] = "Syst Synth Biol",
  ["systems biology"] = "Syst Biol (Stevenage)",
  ["systems biology in reproductive medicine"] = "Syst Biol Reprod Med",
  ["systems biomedicine (austin, tex.)"] = "Syst Biomed (Austin)",
  ["systems engineering"] = "Syst. Eng.",
  ["systems engineering - theory & practice"] = "Syst. Eng. Theory Pract.",
  ["systems engineering procedia"] = "Syst. Eng. Procedia",
  ["systems evaluation, prediction, and decision-making series"] = "Syst. Eval. Predict. Decis.-Mak. Ser.",
  ["systems medicine (new rochelle, n.y.)"] = "Syst Med (New Rochelle)",
  ["systems microbiology and biomanufacturing"] = "Syst. Microbiol. Biomanuf.",
  ["systems pharmacology"] = "Syst Pharmacol",
  ["systems research : the official journal of the international federation for systems research"] = "Syst Res",
  ["systems research and behavioral science"] = "Syst Res Behav Sci",
  ["systems research forum"] = "Syst. Res. Forum",
  ["systems science and control engineering"] = "Syst. Sci. Control Eng.",
  ["systems science and mathematical sciences"] = "Systems Sci. Math. Sci.",
  ["systems, control and information"] = "Systems Control Inform.",
  ["szczecinskie towarzystwo naukowe, wydzial nauk lekarskich"] = "Szczecin Tow Nauk Wydz Nauk Lek",
  ["szczecińskie towarzystwo naukowe, wydział nauk lekarskich"] = "Szczecin Tow Nauk Wydz Nauk Lek",
  ["szpitalnictwo polskie"] = "Szpit Pol",
  ["säugetierkundliche mitteilungen"] = "Säugetierkd. Mitt.",
  ["séminaire de mathématique de luxembourg"] = "Sém. Math. Luxembourg",
  ["séminaire de théorie spectrale et géométrie"] = "Sémin. Théor. Spectr. Géom.",
  ["séminaire d’analyse"] = "Sémin. Anal. Univ. Blaise Pascal (Clermont II)",
  ["séminaire d’analyse moderne"] = "Sém. Anal. Moderne",
  ["séminaire lotharingien de combinatoire"] = "Sém. Lothar. Combin.",
  ["séminaire: équations aux dérivées partielles"] = "Sémin. Équ. Dériv. Partielles",
  ["séminaires et congrès"] = "Sémin. Congr.",
  ["słownik biograficzny polskich nauk medycznych xx wieku"] = "Slown Biogr Pol Med XX Wieku",
  ["słupskie prace matematyczno-przyrodnicze"] = "Słup. Prace Mat. Przyr. Mat. Fiz.",
  ["ślas̨ki kwartalnik historyczny sobótka"] = "Slaski Kwat Hist Sabotka",
  ["šumarski list"] = "Sumar List",
  ["ṣāmid al-iqtiṣādī = samed"] = "Samid Aliqtisadi",
  ["şişli etfal hastanesi tıp bülteni"] = "Sisli Etfal Hastan Tıp Bul",
  ["t. brookhaven national laboratory"] = "Tech Rep Brookhaven Natl Lab",
  ["t. h. carpenter (hrsg.), beazley addenda \\textsuperscript{2}(oxford 1989)"] = "BeazleyAddenda2",
  ["t.-i.-t. journal of life sciences"] = "TIT J Life Sci",
  ["t[echnical report]. brookhaven national laboratory"] = "Tech Rep Brookhaven Natl Lab",
  ["t\\lasp bilisis a"] = "T\\lasp bilis. A. Razmadzis Saxel. Mat\\lasp. Inst. Shromebi",
  ["tablet (brooklyn, new york, n.y.)"] = "Tablet",
  ["tabriz university series"] = "Tabriz Univ. Ser.",
  ["tabula imperii byzantini"] = "TIB",
  ["tabulae biologicae"] = "Tabulae Biol",
  ["taehan ch'ikkwa uisa hyŏphoe chi"] = "Taehan Chikkwa Uisa Hyophoe Chi",
  ["taehan chikkwa uisa hyophoe chi (journal of the korean dental association)"] = "Taehan Chikkwa Uisa Hyophoe Chi",
  ["taehan ch’ikkwa uisa hyophoe chi"] = "Taehan Chikkwa Uisa Hyophoe Chi",
  ["taehan imsang misaengmul hakhoe chi"] = "Korean journal of clinical microbiology|Taehan Imsang Misaengmul Hakhoe Chi",
  ["taehan imsang misaengmul hakhoe chi = korean journal of clinical microbiology"] = "Taehan Imsang Misaengmul Hakhoe Chi",
  ["taehan imsang pyongni hakhoe chi"] = "Korean journal of clinical pathology|Korean J. Clin. Pathol.",
  ["taehan imsang pyŏngni hakhoe chi = korean journal of clinical pathology"] = "Korean J Clin Pathol",
  ["taehan kan hakhoe chi = the korean journal of hepatology"] = "Taehan Kan Hakhoe Chi",
  ["taehan kanho (korean nurse)"] = "Taehan Kanho",
  ["taehan kanho hakhoe chi"] = "Taehan Kanho Hakhoe Chi",
  ["taehan kanho. the korean nurse"] = "Taehan Kanho",
  ["taehan kumsok hakhoechi"] = "Taehan Kumsok Hakhoechi",
  ["taehan kumsok, chaeryo hakhoechi"] = "Taehan Kumsok, Chaeryo Hakhoechi",
  ["taehan misaengmul hakhoe chi"] = "The journal of the Korean Society for Microbiology|Taehan Misaengmul Hakhoe Chi",
  ["taehan misaengmul hakhoe chi = the journal of the korean society for microbiology"] = "Taehan Misaengmul Hakhoe Chi",
  ["taehan naekwa hakhoe chapchi = the korean journal of internal medicine"] = "Taehan Naekwa Hakhoe Chapchi",
  ["taehan pairŏsŭ hakhoe chi"] = "Taehan Pairosu Hakhoe Chi",
  ["taehan pinyogikwa hakhoe chapchi = the korean journal of urology"] = "Taehan Pinyogikwa Hakhoe Chapchi",
  ["taehan sanbuinkwa hakhoe chapchi = korean journal of obstetrics and gynecology"] = "Taehan Sanbuinkwa Hakhoe Chapchi",
  ["taehan suŭi hakhoe chi. taehan suŭi hakhoe"] = "Taehan Suui Hakhoe Chi Taehan Suui Hakhoe",
  ["taehan uihak hyophoe chi. the journal of the korean medical association"] = "Taehan Uihak Hyophoe Chi",
  ["taehan ŭihak hyŏphoe chi. the journal of the korean medical association"] = "Taehan Uihak Hyophoe Chi",
  ["taehan ŭijinkyun hakhoe chi = korean journal of medical mycology"] = "Taehan Uijinkyun Hakhoe Chi",
  ["tag. theoretical and applied genetics. theoretische und angewandte genetik"] = "Theor Appl Genet",
  ["tages-anzeiger magazin"] = "Tages-Anz. Mag.",
  ["tagungsbericht. österreichische ärztetagung"] = "Tag Ber Osterr Arztetag",
  ["tahqiqat-e eqtesadi (quarterly journal of economic research)"] = "Tahq. Eq.",
  ["taḥqīqāt-i zhinitīk va iṣlāḥ-i giyāhān-i martaʻī-i va jangalī-i īrān"] = "Tahqiqati Zhinitik Islahi Giyahani Martaii Jangalii Iran",
  ["tai da li shi xue bao"] = "Tai Da Li Shi Xue Bao",
  ["tai da wen shi zhe xue bao"] = "Tai Da Wen Shi Zhe Xue Bao",
  ["tairyoku, eiyō men'ekigaku zasshi = journal of physical fitness, nutrition and immunology"] = "Tairyoku Eiyo Menekigaku Zasshi",
  ["taishō 2-nenmatsu jinkō seitai chōsa no kekka ni yoru teikoku jinkō gaisetsu / naikaku tōkeikyoku hensan = exposé général des résultats du recensement de la population de l'empire du japon à la fin de 1913 : cabinet impérial, bureau de la statistique générale"] = "Skin Res Technol",
  ["taiwan economic review"] = "Taiwan Econ. Rev.",
  ["taiwan jing ji yu ce"] = "Taiwan Jing Ji Yu Ce",
  ["taiwan journal of ophthalmology"] = "Taiwan J Ophthalmol",
  ["taiwan lin ye ke xue"] = "Taiwan Lin Ye Ke Xue",
  ["taiwan nong ye hua xue yu shi pin ke xue"] = "Taiwanese journal of agricultural chemistry and food science|Taiwan Nong Ye Hua Xue Yu Shi Pin Ke Xue",
  ["taiwan nong ye hua xue yu shi pin ke xue = taiwanese journal of agricultural chemistry and food science"] = "Taiwan Nong Ye Hua Xue Yu Shi Pin Ke Xue",
  ["taiwan nong ye yan jiu"] = "Taiwan Nong Ye Yan Jiu",
  ["taiwan sheng wu duo yang xing yan jiu"] = "Taiwan Sheng Wu Duo Yang Xing Yan Jiu",
  ["taiwan veterinary journal"] = "Taiwan Vet. J.",
  ["taiwan xu mu shou yi xue hui hui bao. tʻai-wan sheng hsü mu shou i hsüeh hui"] = "Taiwan Xu Mu Shou Yi Xue Hui Hui Bao",
  ["taiwan yi xue hui za zhi (journal of the formosan medical association)"] = "Taiwan Yi Xue Hui Za Zhi",
  ["taiwan yi xue hui za zhi. journal of the formosan medical association"] = "Taiwan Yi Xue Hui Za Zhi",
  ["taiwan yi xue ren wen xue kan"] = "Taiwan Yi Xue  Ren Wen Xue Kan",
  ["taiwanese journal of mathematics"] = "Taiwanese J. Math.",
  ["taiwanese journal of obstetrics & gynecology"] = "Taiwan J Obstet Gynecol",
  ["taiwanese journal of orthodontics"] = "Taiwan J Orthod",
  ["taiyuan li gong da xue xue bao = journal of taiyuan university of technology"] = "Taiyuan Li Gong Da Xue Xue Bao",
  ["takamine kenkyūjo nenpō the annual report of takamine laboratory. sankyō kabushiki kaisha. takamine kenkyūjo"] = "Takamine Kenkyusho Nenpo",
  ["talanta open"] = "Talanta Open",
  ["talimu nongken daxue xuebao"] = "Talimu Nongken Daxue Xuebao",
  ["tamkang journal of mathematics"] = "Tamkang J. Math.",
  ["tampa bay history"] = "Tampa Bay Hist",
  ["tamsui oxford journal of information and mathematical sciences"] = "Tamsui Oxf. J. Inf. Math. Sci.",
  ["tamsui oxford journal of management sciences"] = "Tamsui Oxford J. Management Sci.",
  ["tamsui oxford journal of mathematical sciences"] = "Tamsui Oxf. J. Math. Sci.",
  ["tandlaegernes nye tidsskrift"] = "Tandlaegernes Tidsskr",
  ["taniguchi symposia on brain sciences"] = "Tanaguchi Symp Brain Sci",
  ["tanisal ve girisimsel radyoloji : tibbi goruntuleme ve girisimsel radyoloji dernegi yayin organi"] = "Tani Girisim Radyol",
  ["tannlaeknabladid : blad tannlaeknafélags íslands = icelandic dental journal"] = "Tannlaeknabladid",
  ["tanpakushitsu kakusan koso"] = "Tanpakushitsu Kakusan Koso",
  ["tanpakushitsu kakusan koso (protein, nucleic acid, enzyme)"] = "Tanpakushitsu Kakusan Koso.",
  ["tanpakushitsu kakusan koso. protein, nucleic acid, enzyme"] = "Tanpakushitsu Kakusan Koso",
  ["tanzania health research bulletin"] = "Tanzan Health Res Bull",
  ["tanzania journal of health research"] = "Tanzan J Health Res",
  ["tanzania notes and records"] = "Tanzan Notes Rec",
  ["tanzanian economic review"] = "Tanzan Econ Rev",
  ["tanısal ve girişimsel radyoloji : tıbbi görüntüleme ve girişimsel radyoloji derneği yayın organı"] = "Tani Girisim Radyol",
  ["tappi journal"] = "Tappi J.",
  ["taprobanica : the journal of south asian biodiversity, with special attention to sri lanka ands western ghats"] = "Taprobanica",
  ["tar heel nurse"] = "Tar Heel Nurse",
  ["taras. rivista di archeologia"] = "Taras",
  ["target proteins database"] = "Target Proteins Database",
  ["targeted diagnosis and therapy"] = "Targeted Diagn Ther",
  ["targeted oncology"] = "Target Oncol",
  ["tartu ülikooli toimetised"] = "Tartu Ül. Toimetised",
  ["tata institute of fundamental research lectures on mathematics and physics"] = "Tata Inst. Fund. Res. Lectures on Math. and Phys.",
  ["tata institute of fundamental research studies in mathematics"] = "Tata Inst. Fundam. Res. Stud. Math.",
  ["tatra mountains mathematical publications"] = "Tatra Mt. Math. Publ.",
  ["tập chí khoa học kỹ thuật nông lâm nghiệp"] = "Tap Chi Khoa Hoc Ky Thuat Nong Lam Nghiep",
  ["tāksūnūmī va bīyūsīstimātīk"] = "Taksunumi Biyusistimatik",
  ["tạp chí công nghê sinh học"] = "Tap Chi Cong Nghe Sinh Hoc",
  ["tạp chí hóa học"] = "Tap Chi Hoa Hoc",
  ["tạp chí khoa học & giáo dục"] = "Tap Chi Khoa Hoc Giao Duc",
  ["tạp chí khoa học (hue)"] = "Tap Chi Khoa Hoc (Hue)",
  ["tạp chí khoa học = journal of science"] = "Tap Chi Khoa Hoc",
  ["tạp chí khoa học tự nhiên"] = "Tap Chi Khoa Hoc Tu Nhien",
  ["tạp chí khoa học và công nghệ"] = "Tap Chi Khoa Hoc Va Cong Nghe",
  ["tạp chí khoa học và công nghệ nông nghiệp việt nam"] = "Tap Chi Khoa Hoc Va Cong Nghe Nong Nghiep Viet Nam",
  ["tạp chí khoa học văn lang"] = "Tap Chi Khoa Hoc Van Lang",
  ["tạp chí khoa học, khoa học xã hội"] = "Tap Chi Khoa Hoc Khoa Hoc Xa Hoi",
  ["tạp chí nghiên cứu y học"] = "Tap Chi Nghien Y Hoc",
  ["tạp chí sinh học"] = "Tap Chi Sinh Hoc",
  ["tạp chí y học dự phòng = journal of preventive medicine"] = "Tap Chi Y Hoc Du Phong",
  ["tạp chí y tế công cộng"] = "Tạp Chi Y Te Cong Cong",
  ["tb & hiv"] = "TB HIV",
  ["tbilisi mathematical journal"] = "Tbilisi Math. J.",
  ["tbilisskiy gosudarstvennyy universitet"] = "Tbiliss. Gos. Univ. Inst. Prikl. Mat. Trudy",
  ["tbilisskiy universitet"] = "Trudy Tbiliss. Univ. Mat. Mekh. Astronom.",
  ["tdr news"] = "TDR News",
  ["teach-in : the journal for junior hospital doctors and senior medical students"] = "Teach In",
  ["teacher education and special education"] = "Teach Educ Spec Educ",
  ["teachers college record"] = "Teach Coll Rec",
  ["teachers college record (1970)"] = "Teach Coll Rec (1970)",
  ["teaching & learning"] = "Teach Learn",
  ["teaching and learning in medicine"] = "Teach Learn Med",
  ["teaching and learning in nursing : official journal of the national organization for assciate degree nursing"] = "Teach Learn Nurs",
  ["teaching and teacher education"] = "Teach Teach Educ",
  ["teaching business and economics"] = "Teaching Bus. Econ.",
  ["teaching children mathematics"] = "Teach Child Math",
  ["teaching exceptional children"] = "Teach Except Child",
  ["teaching mathematics and its applications"] = "Teach. Math. Appl.",
  ["teaching notes on population"] = "Teach Notes Popul",
  ["teaching of psychology (columbia, mo.)"] = "Teach Psychol",
  ["teaching philosophy"] = "Teach Philos",
  ["teaching political science"] = "Teach Polit Sci",
  ["teaching sociology"] = "Teach Sociol",
  ["teaching statistics"] = "Teach Stat",
  ["tec bulletin"] = "TEC Bull (Online)",
  ["tec bulletin (online)"] = "TEC Bull (Online)",
  ["technical bulletin of the registry of medical technologists"] = "Tech. Bull. Regist. Med. Technol.",
  ["technical bulletin of the registry of medical technologists. american society of clinical pathologists. registry of medical technologists"] = "Tech Bull Regist Med Technol",
  ["technical bulletin. united states. veterans administration"] = "Tech Bull United States Veterans Admin",
  ["technical communication"] = "Tech Commun",
  ["technical communication quarterly"] = "Tech. Commun. Q",
  ["technical digest. international electron devices meeting"] = "Tech Dig Int Electron Devices Meet",
  ["technical documentary report, arl-tdr. united states. aeromedical research laboratory, holloman air force base, n. m"] = "Tech Doc Rep ARL TDR",
  ["technical documentary report, rtd tdr. united states. air force. systems command. research and technology division"] = "Tech  Doc Rep RTD TDR",
  ["technical documentary report. sam-tdr. usaf school of aerospace medicine"] = "Tech Doc Rep SAMTDR USAF Sch Aerosp Med",
  ["technical documentary report. united states. air force. systems command. electronic systems division"] = "Tech Doc Rep U S Air Force Syst Command Electron Syst Div",
  ["technical documentary report; aal-tdr. arctic aeromedical laboratory (u.s.)"] = "Tech Doc Rep Arct Aeromed Lab US",
  ["technical documents in hydrology"] = "Tech. Doc. Hy.",
  ["technical innovations & patient support in radiation oncology"] = "Tech Innov Patient Support Radiat Oncol",
  ["technical manual. u.s. army biological laboratories"] = "Tech Man US Army Biol Lab",
  ["technical manuscript. u.s. army biological laboratories"] = "Tech Manuscr US Army Biol Lab",
  ["technical note, afmdc tn. air force missile development center (u.s.)"] = "Tech Note AFMDC TN",
  ["technical note. united states. national aeronautics and space administration"] = "Tech Note U S Natl Aeronaut Space Adm",
  ["technical note; tn. arctic aeromedical laboratory (u.s.)"] = "Tech Note Arct Aeromed Lab (US)",
  ["technical physics"] = "Tech. Phys.",
  ["technical physics letters"] = "Tech. Phys. Lett.",
  ["technical quarterly & the mbaa communicator"] = "Tech. Q. MBAA Commun.",
  ["technical quarterly - master brewers association of the americas"] = "Tech. Q. Master Brew. Assoc. Am.",
  ["technical report"] = "Tech. Rep.",
  ["technical report forest engineering research institute of canada"] = "Tech. Rep. For. Eng. Res. Inst. Can.",
  ["technical report, cp. quartermaster research & development center (u.s.). chemicals and plastics division"] = "Tech Rep CP",
  ["technical report. crdlr. u.s. army chemical research and development laboratories"] = "Tech Rep CRDLR US Army Chem Res Dev Lab",
  ["technical report. crdlr. us army edgewood arsenal chemical research and development laboratories"] = "Tech Rep CRDLR US Army Chem Res Dev Lab",
  ["technical report. cwlr. u.s. army chemical warfare laboratories"] = "Tech Rep CWLR US Army Chem Warf Lab",
  ["technical report. new york naval shipyard. material laboratory"] = "Tech Rep NY Nav Shipyard Mater Lab",
  ["technical report.; tr. arctic aeromedical laboratory (u.s.)"] = "Tech Rep Arct Aeromed Lab US",
  ["technical report: navtradevcen. naval training device center"] = "Tech Rep NAVTRADEVCEN",
  ["technical reports in hydrology and water resources"] = "Tech. Rep. Hydrol. Water Resour.",
  ["technical reports of the australian museum. australian museum"] = "Tech Rep Aust Mus",
  ["technical sciences: advances in electronics"] = "Tech. Sci. Adv. Electron.",
  ["technical tips online"] = "Tech Tips Online",
  ["technical university of wrocław"] = "Systems Sci.",
  ["technicien belge en prothese dentaire"] = "Tech. Belge Prothese Dent.",
  ["technik geschichte"] = "Technikgeschichte",
  ["technikgeschichte in einzeldarstellungen"] = "Technikgesch Einzeldarst",
  ["techniki tworczego myslenia"] = "Tech. Tworczego Myslenia",
  ["techniques and innovations in gastrointestinal endoscopy"] = "Tech Innov Gastrointest Endosc",
  ["techniques de l’ingenieur, constantes physico-chimiques"] = "Tech. Ing. Constantes Phys. Chim.",
  ["techniques de l’ingenieur, environnement"] = "Tech. Ing. Environ.",
  ["techniques de l’ingenieur, genie des procede"] = "Tech. Ing. Genie Procedes",
  ["techniques de l’ingenieur, genie nuclear"] = "Tech. Ing. Genie Nucl.",
  ["techniques de l’ingenieur, materiaux fonctionnels"] = "Tech. Ing. Mater. Fond.",
  ["techniques de l’ingenieur, materiaux metalliques"] = "Tech. Ing. Mater. Met.",
  ["techniques de l’ingenieur, plastiques et composites"] = "Tech. Ing. Plast. Compos.",
  ["techniques de l’ingenieur, sciences fondamentales: mathematiques pour l’ingenieur"] = "Tech. Ing. Sci. Fondam. Math. Ing.",
  ["techniques de l’ingenieur, sciences fondamentales: physique, chemie"] = "Tech. Ing. Sci. Fondam. Phys. Chim.",
  ["techniques hospitalieres, medico-sociales et sanitaires"] = "Tech Hosp Med Soc Sanit",
  ["techniques hospitalières, médico-sociales et sanitaires"] = "Tech Hosp Med Soc Sanit",
  ["techniques in coloproctology"] = "Tech Coloproctol",
  ["techniques in foot & ankle surgery"] = "Tech Foot Ankle Surg",
  ["techniques in gastrointestinal endoscopy"] = "Tech Gastrointest Endosc",
  ["techniques in hand & upper extremity surgery"] = "Tech Hand Up Extrem Surg",
  ["techniques in orthopaedics (rockville, md.)"] = "Tech Orthop",
  ["techniques in regional anesthesia & pain management"] = "Tech Reg Anesth Pain Manag",
  ["techniques in shoulder & elbow surgery"] = "Tech Shoulder Elb Surg",
  ["techniques in urology"] = "Tech Urol",
  ["techniques in vascular and interventional radiology"] = "Tech Vasc Interv Radiol",
  ["techniques orthopediques"] = "Tech Orthop",
  ["technische mitteilungen"] = "Tech. Mitt.",
  ["technische mitteilungen aeg-telefunken"] = "Tech. Mitt. AEG-Telefunken",
  ["technische rundschau"] = "Tech. Rundsch.",
  ["technische uberwachung"] = "Tech. Uberwach.",
  ["technisches messen"] = "Tech. Mess.",
  ["technoetic arts : a journal of speculative research"] = "Technoetic Arts",
  ["technological forecasting and social change"] = "Technol Forecast Soc Change",
  ["technological sustainability"] = "Technol. Sustainability",
  ["technology (elmsford, n.y.)"] = "Technology",
  ["technology analysis & strategic management"] = "Technol Anal Strateg Manag",
  ["technology analysis and strategic management"] = "Tech. Anal. Strategic Manage.",
  ["technology and aging"] = "Assist. Techn. Res. Ser.",
  ["technology and culture"] = "Technol Cult",
  ["technology and disability"] = "Technol Disabil",
  ["technology and economics of smart grids and sustainable energy"] = "Technol. Econ. Smart Grids Sustainable Energy",
  ["technology and health care"] = "Technol. Health Care",
  ["technology and health care : official journal of the european society for engineering and medicine"] = "Technol Health Care",
  ["technology and innovation"] = "Technol Innov",
  ["technology and investment"] = "Technol Invest",
  ["technology evaluation center assessment program. executive summary"] = "Technol Eval Cent Assess Program Exec Summ",
  ["technology in cancer research & treatment"] = "Technol Cancer Res Treat",
  ["technology in cancer research and treatment"] = "Technol. Cancer Res. Treat.",
  ["technology in education series"] = "Tech. Ed. Ser.",
  ["technology in society"] = "Technol Soc",
  ["technology reports of the osaka university"] = "Tech. Rep. Osaka Univ.",
  ["technology review"] = "Technol Rev",
  ["technology science"] = "Technol Sci",
  ["technology, knowledge and learning"] = "Technol., Knowl. Learn.",
  ["technology, mind, and behavior"] = "Technol. Mind Behav.",
  ["technology, risk, and society"] = "Technol. Risk Soc.",
  ["technometrics : a journal of statistics for the physical, chemical, and engineering sciences"] = "Technometrics",
  ["technometrics. a journal of statistics for the physical, chemical and engineering sciences"] = "Technometrics",
  ["techtrends : for leaders in education & training"] = "TechTrends",
  ["tecnologia em metalurgia, materiais e mineracao"] = "Tecnol. Metal., Mater. Min.",
  ["tecnologia em metalurgia, materiais e mineração"] = "Tecnol Metal Mater Min",
  ["tecnologica (chicago, ill.)"] = "Tecnologica",
  ["tecnologica. map supplement"] = "Tecnologica. MAP Suppl.",
  ["tecnologica. map supplement. blue cross and blue shield association. medical advisory panel"] = "Tecnologica MAP Suppl",
  ["tegen de tuberculose"] = "Tegen Tuberc",
  ["teheraner forschungen"] = "TeherF",
  ["teion kōgaku = cryogenic engineering : [official journal of the cryogenic association of japan]"] = "Teion Kogaku",
  ["teiresias. a review and continuing bibliography of boiotian studies"] = "Teiresias",
  ["tekhnicheskiya universitet v plovdiv"] = "Izv. Tekhn. Univ. Plovdiv. Fund. Nauk. Prilozhen.",
  ["tel aviv. journal of the institute of archaeology of tel aviv university"] = "TelAvivJA",
  ["tel aviver jahrbuch fur deutsche geschichte / herausgegeben vom institut fur deutsche geschichte"] = "Tel Aviver Jahrb Dtsch Gesch",
  ["tel aviver jahrbuch für deutsche geschichte"] = "Tel Aviver Jahrb Dtsch Gesch",
  ["telecommunication systems"] = "Telecommun Syst",
  ["telecommunications and radio engineering"] = "Telecommun. Radio Eng.",
  ["telecommunications policy"] = "Telecomm Policy",
  ["telemedicine and e-health"] = "Telemed. e-Health",
  ["telemedicine and telehealth networks"] = "Telemed. Telehealth Netw.",
  ["telemedicine and telehealth networks : newsmagazine of distance healthcare"] = "Telemed Telehealth Netw",
  ["telemedicine and virtual reality"] = "Telemed Virtual Real",
  ["telemedicine journal"] = "Telemed. J.",
  ["telemedicine journal : the official journal of the american telemedicine association"] = "Telemed J",
  ["telemedicine journal and e-health"] = "Telemed. J. E Health",
  ["telemedicine journal and e-health : the official journal of the american telemedicine association"] = "Telemed J E Health",
  ["telemedicine reports"] = "Telemed Rep",
  ["telemedicine today"] = "Telemed Today",
  ["telinde's operative gynecology updates"] = "TeLindes Oper Gynecol Updates",
  ["telinde’s operative gynecology updates"] = "TeLindes Oper Gynecol Updates",
  ["tellus a: dynamic meteorology and oceanography"] = "Tellus A: Dyn. Meteorol. Oceanogr.",
  ["tellus b: chemical and physical meteorology"] = "Tellus B: Chem. Phys. Meteorol.",
  ["tellus series a"] = "Tellus Ser. A",
  ["tellus series b"] = "Tellus Ser. B",
  ["tellus. series b, chemical and physical meteorology"] = "Tellus B Chem Phys Meteorol",
  ["tema. tendencias em matematica aplicada e computacional"] = "TEMA Tend. Mat. Apl. Comput.",
  ["temas de la agenda pública"] = "Temas Agenda Publica",
  ["temas de poblacion"] = "Temas Poblac",
  ["temas de población"] = "Temas Poblac",
  ["temas odontologicos"] = "Temas Odontol.",
  ["temas odontológicos"] = "Temas Odontol",
  ["temida : časopis o viktimizaciji, ljudskim pravima i rodu"] = "Temida (Beogr)",
  ["temperature (austin, tex.)"] = "Temperature (Austin)",
  ["temple dental review"] = "Temple Dent Rev",
  ["temple international and comparative law journal"] = "Temple Int Comp Law J",
  ["temple law quarterly (philadelphia, pa. : 1946)"] = "Temple Law Q",
  ["temple law review"] = "Temple Law Rev",
  ["tempo (rio de janeiro, brazil)"] = "Tempo (Rio J)",
  ["tennessee code annotated / prepared under the supervision of the tennessee code commission. tennessee"] = "Tenn Code Annot Tenn",
  ["tennessee code annotated. tennessee"] = "Tenn Code Annot Tenn",
  ["tennessee folklore society bulletin"] = "Tenn Folk Soc Bull",
  ["tennessee historical quarterly"] = "Tenn Hist Q",
  ["tennessee law review"] = "Tenn Law Rev",
  ["tennessee medicine"] = "Tenn. Med.",
  ["tennessee medicine : journal of the tennessee medical association"] = "Tenn Med",
  ["tennessee nurse"] = "Tenn Nurse",
  ["tennessee nurse / tennessee nurses association"] = "Tenn Nurse",
  ["tennessee public welfare record"] = "Tenn Publ Welfare Rec",
  ["tennessee studies in literature"] = "Tenn Stud Lit",
  ["tenside, surfactants, detergents"] = "Tenside, Surfactants, Deterg.",
  ["teoreticheskai︠a︡ i prikladnai︠a︡ ėkologii︠a︡"] = "Teor Prikl Ekol",
  ["teoreticheskaya elektrotekhnika"] = "Teoret. Elektrotekhn.",
  ["teoreticheskaya i eksperimental’naya khimiya"] = "Teor. Eksp. Khim.",
  ["teoreticheskaya i matematicheskaya fizika"] = "Teoret. Mat. Fiz.",
  ["teoreticheskie osnovy khimicheskoi tekhnologii"] = "Teor. Osn. Khim. Tekhnol.",
  ["teorie & modelli"] = "Teor Model",
  ["teorie a praxe telesne vychovy a sportu"] = "Teor Prax Teles Vychovy",
  ["teorie a praxe tĕlesné výchovy a sportu"] = "Teor Prax Teles Vychovy",
  ["teoriia i praktika fizicheskoi kul’tury"] = "Teor Prak Fiz Kult",
  ["teoriia i praktika fizicheskoĭ kul'tury"] = "Teor Prak Fiz Kult",
  ["teoriya funktsiy, funktsional\\cprime nyy analiz i ikh prilozheniya"] = "Teor. Funktsiy Funktsional. Anal. i Prilozhen.",
  ["teoriya imovirnostei ta matematichna statistika. kiivskii universitet imeni tarasa shevchenka"] = "Teor. Imovir. Mat. Stat.",
  ["teoriya veroyatnostei i ee primeneniya"] = "Teor. Veroyatn. Primen.",
  ["teploenergetika (moscow)"] = "Teploenergetika (Moscow)",
  ["teplofizika vysokikh temperatur"] = "Teplofiz. Vys. Temp.",
  ["teplofizika vysokikh temperatur [high temperature (ussr)]"] = "Teplofiz. Vys. Temp. [High Temp. (USSR)]",
  ["terahertz science and technology"] = "Terahertz Sci. Technol.",
  ["terapevticheskii arkhiv"] = "Ter Arkh",
  ["terapevticheskiĭ arkhiv"] = "Ter Arkh",
  ["teratogenesis, carcinogenesis, and mutagenesis"] = "Teratog Carcinog Mutagen",
  ["terra (helsinki, finland)"] = "Terra (Hels)",
  ["terra ameriga"] = "Terra Ameriga",
  ["terra dei volsci. annali del museo archeologico di frosinone"] = "TerraVolsci",
  ["terra latinoamericana : organo científico de la sociedad mexicana de la ciencia del suelo, a.c"] = "Terra Latinoam",
  ["terra nova"] = "Terra Nova",
  ["terrae incognitae"] = "Terr Incogn",
  ["terramycine informations"] = "Terramycine Inf",
  ["terre et la vie, la"] = "Terre Vie",
  ["terrestrial arthropod reviews"] = "Terr Arthropod Rev",
  ["terrestrial magnetism"] = "Terr. Magn.",
  ["terrestrial magnetism and atmospheric electricity"] = "Terr. Magn. Atmos. Electr.",
  ["terrestrial, atmospheric and oceanic sciences"] = "Terr., Oceanic Atmos. Sci.",
  ["territory, politics, governance"] = "Territ Politic Gov",
  ["teruel. instituto de estudios turolenses"] = "Teruel",
  ["tesi di perfezionamento"] = "Tesi Perfez.",
  ["tesi. scuola normale superiore di pisa (nuova series)"] = "Tesi Sc. Norm. Super. Pisa (N. S.)",
  ["test (madrid, spain)"] = "Test (Madr)",
  ["test. an official journal of the spanish society of statistics and operations research"] = "TEST",
  ["tetrahedron chem"] = "Tetrahedron Chem",
  ["tetrahedron green chem"] = "Tetrahedron Green Chem",
  ["tetrahedron letters"] = "Tetrahedron Lett.",
  ["tetrahedron, asymmetry"] = "Tetrahedron Asymmetry",
  ["tetrahedron: asymmetry"] = "Tetrahedron: Asymmetry",
  ["tetsu to hagane"] = "Tetsu to Hagane",
  ["teubner skripten zur mathematischen stochastik"] = "Teubner Skr. Math. Stochastik",
  ["teubner skripten zur numerik"] = "Teubner Skr. Numer.",
  ["teubner studienbucher physik/chemie"] = "Teubner Studienbuch. Phys./Chem.",
  ["teubner studienbucher wirtschaftswissenschaften"] = "Teubner Stud.buch. Wirtsch.wiss.",
  ["teubner studienbücher mathematik"] = "Teubner Studienbüch. Math.",
  ["teubner studienskripten: mathematik, ingenieurwissenschaften"] = "Teubner Studienskr. Math. Ingen.",
  ["teubner-archiv zur mathematik"] = "Teubner-Arch. Math.",
  ["teubner-texte zur informatik"] = "Teubner-Texte Inform.",
  ["teubner-texte zur mathematik"] = "Teubner-Texte Math.",
  ["texas bar journal"] = "Texas Bar J",
  ["texas cancer bulletin"] = "Tex Cancer Bull",
  ["texas dental assistants association bulletin"] = "Tex Dent Assist Assoc Bull",
  ["texas dental journal"] = "Tex Dent J",
  ["texas education review (austin, tex.)"] = "Tex Educ Rev (Austin)",
  ["texas heart institute journal"] = "Tex Heart Inst J",
  ["texas heart institute journal / from the texas heart institute of st. luke’s episcopal hospital, texas children’s hospital"] = "Tex Heart Inst J",
  ["texas hospitals"] = "Tex Hosp",
  ["texas international law journal"] = "Tex Int Law J",
  ["texas journal of health-system pharmacy"] = "Tex J Health Syst Pharm",
  ["texas journal of pharmacy"] = "Tex J Pharm",
  ["texas journal of women and the law"] = "Tex J Women Law",
  ["texas law review"] = "Tex Law Rev",
  ["texas medical journal (austin, tex.)"] = "Tex Med J (Austin)",
  ["texas medicine"] = "Tex Med",
  ["texas monthly (austin, tex.)"] = "Tex Mon",
  ["texas nursing"] = "Tex Nurs",
  ["texas public health journal"] = "Tex Public Health J",
  ["texas reports on biology and medicine"] = "Tex Rep Biol Med",
  ["texas review of law & politics"] = "Tex Rev Law Polit",
  ["texas state journal of medicine"] = "Tex State J Med",
  ["texas studies in literature and language"] = "Texas Stud Lit Lang",
  ["texas tech law review"] = "Tex Tech Law Rev",
  ["text, speech and dialogue. tsd"] = "Text Speech Dialog",
  ["textbook series in physical sciences"] = "Textb. Ser. Phys. Sci.",
  ["textbooks in electrical and electronic engineering"] = "Textb. Electr. Electron. Eng.",
  ["textbooks in mathematical sciences"] = "Textbooks Math. Sci.",
  ["textbooks in mathematics"] = "Textb. Math.",
  ["texte zur didaktik der mathematik"] = "Texte Didakt. Math.",
  ["texte zur geschichte der präventivmedizin : tgp"] = "Texte Gesch Praventivmed",
  ["textes et etudes du moyen age"] = "Textes Etudes Moyen Age",
  ["textile chemist and colorist"] = "Text. Chem. Color.",
  ["textile chemist and colorist & american dyestuff reporter"] = "Text. Chem. Color. Am. Dyest Rep.",
  ["textile history"] = "Text Hist",
  ["textile progress"] = "Text. Prog.",
  ["textile rental"] = "Text Rent",
  ["textile research journal"] = "Text. Res. J.",
  ["textile research journal : publication of textile research institute, inc. and the textile foundation"] = "Text Res J",
  ["textile science and technology"] = "Text. Sci. Tech.",
  ["textile world"] = "Text. World",
  ["textile: the journal of cloth and culture"] = "Text.: J. Cloth Cult.",
  ["textiles anciens. bulletin du centre international d’étude des textiles anciens"] = "TextilAnc",
  ["textiles and clothing sustainability"] = "Text. Clothing Sustainability",
  ["textos de matematica"] = "Textos Mat.",
  ["textos de matematica/mathematics texts"] = "Textos Mat./Math. Texts",
  ["textos de matemática"] = "Textos Mat. Sér. B",
  ["textos del bronce"] = "Textos Bronce",
  ["textos nepo"] = "Textos NEPO",
  ["textos y manuales de ensenanza"] = "Textos Man. Ensen.",
  ["texts and monographs in physics"] = "Texts Monogr. Phys.",
  ["texts and monographs in symbolic computation"] = "Texts Monogr. Symbol. Comput.",
  ["texts and monographs in theoretical physics"] = "Texts Monogr. Theor. Phys.",
  ["texts and readings in mathematics"] = "Texts Read. Math.",
  ["texts and readings in physical sciences"] = "Texts Read. Phys. Sci.",
  ["texts for quantitative critical thinking"] = "Texts Quant. Crit. Think.",
  ["texts in algorithmics"] = "Texts Algorithmics",
  ["texts in applied mathematics"] = "Texts Appl. Math.",
  ["texts in computational science and engineering"] = "Texts Comput. Sci. Eng.",
  ["texts in computer science"] = "Texts Comput. Sci.",
  ["texts in computing"] = "Texts Comput.",
  ["texts in logic and reasoning"] = "Texts Log. Reason.",
  ["texts in mathematics"] = "Texts Math.",
  ["texts in philosophy"] = "Texts Philos.",
  ["texts in statistical science series"] = "Texts Statist. Sci. Ser.",
  ["texts in theoretical computer science"] = "Texts Theoret. Comput. Sci. EATCS Ser.",
  ["texts in theoretical computer science. an eatcs series"] = "Texts Theoret. Comput. Sci. Eatcs Ser.",
  ["textual practice"] = "Textual Pract",
  ["tezisy dokladov. mezhrespublikanskaia konferentsiia po voprosam istorii estestvoznaniia i tekhniki v pribaltike"] = "Tezisy Dokl Mezhrespub Konf",
  ["tèyou shengwù yánjiu"] = "Teyou Shengwu Yanjiu",
  ["th open : companion journal to thrombosis and haemostasis"] = "TH Open",
  ["thai forest bulletin"] = "Thai For Bull",
  ["thai journal of agricultural science"] = "Thai J Agric Sci",
  ["thai journal of genetics"] = "Thai J Genet",
  ["thai journal of mathematics"] = "Thai J. Math.",
  ["thai journal of nursing"] = "Thai J Nurs",
  ["thai journal of nursing research"] = "Thai J Nurs Res",
  ["thai journal of pharmaceutical sciences"] = "Thai J. Pharm. Sci.",
  ["thai journal of science and technology"] = "Thai J Sci Technol",
  ["thalamus & related systems"] = "Thalamus Relat Syst",
  ["thalassas : revista de ciencias del mar"] = "Thalassas",
  ["thalassas: international journal of marine sciences"] = "Thalassas Int. J. Mar. Sci.",
  ["thalassia salentina"] = "Thalassia Salentina",
  ["thammasat international journal of science and technology : tijsat"] = "Thammasat Int J Sci Tech",
  ["thang long journal of science. mathematics and mathematical sciences"] = "Thang Long J. Sci. Math. Math. Sci.",
  ["the $r$ series"] = "$R$ Ser.",
  ["the 16th web for all conference : san francisco, 2019. web for all conference (16th : 2019 : san francisco, calif.)"] = "16th Web All Conf (2019)",
  ["the 2010 acm international conference on bioinformatics and computational biology : acm-bcb 2010 : niagara falls, new york, u.s.a., august 2-4, 2010. acm international conference on bioinformatics and computational biology (1st : 2010 : niagara falls, n.y.)"] = "ACM Int Conf Bioinform Comput Biol (2010)",
  ["the a-v"] = "AV (Phila)",
  ["the a.s.t.a. journal"] = "ASTA J",
  ["the aaps journal"] = "AAPS J",
  ["the abnf journal : official journal of the association of black nursing faculty in higher education, inc"] = "ABNF J",
  ["the academic nurse : the journal of the columbia university school of nursing"] = "Acad Nurse",
  ["the academy bookman"] = "Acad Bookman",
  ["the academy of management annals"] = "Acad Manag Ann",
  ["the academy of management perspectives"] = "Acad Manag Perspect",
  ["the academy today : advancing orthotic and prosthetic care through knowledge"] = "Acad Today",
  ["the accounting historians journal"] = "Account Hist J",
  ["the aces bulletin. association for comparative economic studies (u.s.)"] = "ACES Bull",
  ["the adhd report"] = "ADHD Rep",
  ["the admitting management journal"] = "Admit Manage J",
  ["the afcs-nature molecule pages"] = "AFCS Nat Mol Pages",
  ["the afl-cio american federationist. afl-cio"] = "AFL CIO Am Fed",
  ["the african archaeological review"] = "Afr Archaeol Rev",
  ["the african journal of medical sciences"] = "Afr J Med Sci",
  ["the african journal of mycology and biotechnology"] = "Afr J Mycol Biotechnol",
  ["the african journal of psychiatry"] = "Afr J Psychiatry",
  ["the african journal of psychiatry. le journal africain de psychiatrie"] = "Afr J Psychiatry",
  ["the african journal of sexually transmitted diseases"] = "Afr J Sex Transmi Dis",
  ["the aging male : the official journal of the international society for the study of the aging male"] = "Aging Male",
  ["the agricultural history review"] = "Agric Hist Rev",
  ["the ahfad journal"] = "Ahfad J",
  ["the aids reader"] = "AIDS Read",
  ["the air force law review"] = "Air Force Law Rev",
  ["the air surgeon's bulletin"] = "Air Surg Bull",
  ["the alabama historical quarterly"] = "Ala Hist Q",
  ["the alabama journal of medical sciences"] = "Ala J Med Sci",
  ["the alabama nurse"] = "Ala Nurse",
  ["the alabama review"] = "Ala Rev",
  ["the alan guttmacher institute annual report. alan guttmacher institute"] = "Alan Guttmacher Inst Annu Rep",
  ["the alaska nurse"] = "Alaska Nurse",
  ["the albuquerque tribune"] = "Albuq Tribune",
  ["the alexandria medical journal"] = "Alexandria Med J",
  ["the aligarh bulletin of mathematics"] = "Aligarh Bull. Math.",
  ["the aligarh journal of statistics"] = "Aligarh J. Statist.",
  ["the alkaloids. chemistry and biology"] = "Alkaloids Chem Biol",
  ["the all england law reports"] = "All Engl Law Rep",
  ["the all results journals. biol"] = "All Results J Biol",
  ["the alma mater"] = "Alma Mater (Baltimore)",
  ["the alpha omegan"] = "Alpha Omegan",
  ["the alumnae magazine"] = "Alumnae Mag",
  ["the alumni magazine"] = "Alumni Mag",
  ["the american archivist"] = "Am Arch",
  ["the american art journal"] = "Am Art J",
  ["the american asian review"] = "Am Asian Rev",
  ["the american baby"] = "Am Baby",
  ["the american behavioral scientist"] = "Am Behav Sci",
  ["the american benedictine review"] = "Am Benedictine Rev",
  ["the american biology teacher"] = "Am Biol Teach",
  ["the american bookman"] = "Am Bookm",
  ["the american college of cardiology extended learning : accel"] = "Am Coll Cardiol Ext Learn",
  ["the american college of cardiology extended learning. supplement tape"] = "Am Coll Cardiol Extended Learn Suppl Tape",
  ["the american criminal law review"] = "Am Crim Law Rev",
  ["the american ecclesiastical review"] = "Am Eccles Rev",
  ["the american economic journal. macroeconomics"] = "Am Econ J Macroecon",
  ["the american economic review"] = "Am Econ Rev",
  ["the american economist"] = "Am Econ",
  ["the american heart hospital journal"] = "Am Heart Hosp J",
  ["the american historical review"] = "Am Hist Rev",
  ["the american homeopath"] = "Am Homeopath",
  ["the american imago; a psychoanalytic journal for the arts and sciences"] = "Am Imago",
  ["the american institute of chemical engineers"] = "AIChE J.",
  ["the american jewish year book"] = "Am Jew",
  ["the american journal for nurse practitioners"] = "Am J Nurse Pract",
  ["the american journal of anatomy"] = "Am J Anat",
  ["the american journal of anesthesiology"] = "Am J Anesthesiol",
  ["the american journal of bioethics : ajob"] = "Am J Bioeth",
  ["the american journal of cancer"] = "Am J Cancer",
  ["the american journal of cardiology"] = "Am J Cardiol",
  ["the american journal of cardiovascular pathology"] = "Am J Cardiovasc Pathol",
  ["the american journal of case reports"] = "Am J Case Rep",
  ["the american journal of chinese medicine"] = "Am J Chin Med",
  ["the american journal of clinical hypnosis"] = "Am J Clin Hypn",
  ["the american journal of clinical nutrition"] = "Am J Clin Nutr",
  ["the american journal of comparative law"] = "Am J Comp Law",
  ["the american journal of dental science"] = "Am J Dent Sci",
  ["the american journal of dermatopathology"] = "Am J Dermatopathol",
  ["the american journal of digestive diseases"] = "Am J Dig Dis",
  ["the american journal of distance education"] = "Am J Distance Educ",
  ["the american journal of drug and alcohol abuse"] = "Am J Drug Alcohol Abuse",
  ["the american journal of eeg technology"] = "Am J EEG Technol",
  ["the american journal of emergency medicine"] = "Am J Emerg Med",
  ["the american journal of evaluation"] = "Am J Eval",
  ["the american journal of family therapy"] = "Am J Fam Ther",
  ["the american journal of forensic medicine and pathology"] = "Am J Forensic Med Pathol",
  ["the american journal of forensic medicine and pathology : official publication of the national association of medical examiners"] = "Am J Forensic Med Pathol",
  ["the american journal of forensic psychiatry"] = "Am J Forensic Psychiatry",
  ["the american journal of forensic psychology"] = "Am J Forensic Psychol",
  ["the american journal of gastroenterology"] = "Am J Gastroenterol",
  ["the american journal of geriatric cardiology"] = "Am J Geriatr Cardiol",
  ["the american journal of geriatric pharmacotherapy"] = "Am J Geriatr Pharmacother",
  ["the american journal of geriatric psychiatry : official journal of the american association for geriatric psychiatry"] = "Am J Geriatr Psychiatry",
  ["the american journal of gynecologic health"] = "Am J Gynecol Health",
  ["the american journal of hospice & palliative care"] = "Am J Hosp Palliat Care",
  ["the american journal of hospice care"] = "Am J Hosp Care",
  ["the american journal of international law"] = "Am J Int Law",
  ["the american journal of islamic social sciences"] = "Am J Islam Soc Sci",
  ["the american journal of jurisprudence"] = "Am J Jurisprud",
  ["the american journal of knee surgery"] = "Am J Knee Surg",
  ["the american journal of legal history"] = "Am J Leg Hist",
  ["the american journal of managed care"] = "Am J Manag Care",
  ["the american journal of medical electronics"] = "Am J Med Electron",
  ["the american journal of medical technology"] = "Am J Med Technol",
  ["the american journal of medicine"] = "Am J Med",
  ["the american journal of nursing"] = "Am J Nurs",
  ["the american journal of obstetrics and diseases of women and children"] = "Am J Obstet Dis Women Child",
  ["the american journal of occupational therapy : official publication of the american occupational therapy association"] = "Am J Occup Ther",
  ["the american journal of occupational therapy. : official publication of the american occupational therapy association"] = "Am J Occup Ther",
  ["the american journal of orthopedic surgery"] = "Am J Orthop Surg",
  ["the american journal of orthopedics"] = "Am J Orthop",
  ["the american journal of orthopsychiatry"] = "Am J Orthopsychiatry",
  ["the american journal of otology"] = "Am J Otol",
  ["the american journal of pathology"] = "Am J Pathol",
  ["the american journal of pediatric hematology/oncology"] = "Am J Pediatr Hematol Oncol",
  ["the american journal of pharmacy"] = "Am J Pharm",
  ["the american journal of pharmacy benefits"] = "Am J Pharm Benefits",
  ["the american journal of physiology"] = "Am J Physiol",
  ["the american journal of psychiatry"] = "Am J Psychiatry",
  ["the american journal of psychiatry residents' journal"] = "Am J Psychiatry Resid J",
  ["the american journal of psychology"] = "Am J Psychol",
  ["the american journal of roentgenology and radium therapy"] = "Am J Roentgenol Radium Ther",
  ["the american journal of roentgenology, radium therapy, and nuclear medicine"] = "Am J Roentgenol Radium Ther Nucl Med",
  ["the american journal of social psychiatry"] = "Am J Soc Psychiatry",
  ["the american journal of sociology"] = "Am J Sociol",
  ["the american journal of sports medicine"] = "Am J Sports Med",
  ["the american journal of surgical pathology"] = "Am J Surg Pathol",
  ["the american journal of the medical sciences"] = "Am J Med Sci",
  ["the american journal of trial advocacy"] = "Am J Trial Advocacy",
  ["the american journal of tropical medicine and hygiene"] = "Am J Trop Med Hyg",
  ["the american journal on addictions"] = "Am J Addict",
  ["the american journal on addictions / american academy of psychiatrists in alcoholism and addictions"] = "Am J Addict",
  ["the american mathematical monthly"] = "Amer. Math. Monthly",
  ["the american mathematical monthly : the official journal of the mathematical association of america"] = "Am Math Mon",
  ["the american midland naturalist"] = "Am Midl Nat",
  ["the american mineralogist"] = "Am Mineral",
  ["the american naturalist"] = "Am Nat",
  ["the american neptune"] = "Am Neptune",
  ["the american nurse"] = "Am Nurse",
  ["the american orthoptic journal"] = "Am Orthopt J",
  ["the american philatelist"] = "Am Philat",
  ["the american political science review"] = "Am Polit Sci Rev",
  ["the american prospect"] = "Am Prospect",
  ["the american psychologist"] = "Am Psychol",
  ["the american review of canadian studies"] = "Am Rev Can Stud",
  ["the american review of respiratory disease"] = "Am Rev Respir Dis",
  ["the american scholar"] = "Am Scholar",
  ["the american school board journal"] = "Am Sch Board J",
  ["the american society legion of honor magazine"] = "Am Soc Legion Honor Mag",
  ["the american sociologist"] = "Am Sociol",
  ["the american spectator"] = "Am Spect",
  ["the american statistician"] = "Am Stat",
  ["the american surgeon"] = "Am Surg",
  ["the american university journal of gender, social policy & the law"] = "Am Univ J Gend Soc Policy Law",
  ["the american university journal of international law and policy"] = "Am Univ J Int Law Policy",
  ["the american university journal of international law and policy / washington college of law"] = "Am Univ J Int Law Policy",
  ["the american university law review"] = "Am Univ Law Rev",
  ["the amicus journal"] = "Amic J",
  ["the analysis of verbal behavior"] = "Anal Verbal Behav",
  ["the analyst"] = "Analyst",
  ["the anatomical record"] = "Anat Rec",
  ["the anatomical record. part a, discoveries in molecular, cellular, and evolutionary biology"] = "Anat Rec A Discov Mol Cell Evol Biol",
  ["the anatomical record. supplement : an offical publication of the american association of anatomists"] = "Anat Rec Suppl",
  ["the ancient history bulletin"] = "AHB",
  ["the ancient india"] = "AncInd",
  ["the ancient world"] = "AncW",
  ["the angle orthodontist"] = "Angle Orthod",
  ["the anglo-american law review"] = "Anglo Am Law Rev",
  ["the animals' agenda"] = "Anim Agenda",
  ["the animals' defender"] = "Anim Def",
  ["the animals’ agenda"] = "Anim Agenda",
  ["the annals of applied biology"] = "Ann Appl Biol",
  ["the annals of applied probability"] = "Ann. Appl. Probab.",
  ["the annals of applied probability : an official journal of the institute of mathematical statistics"] = "Ann Appl Probab",
  ["the annals of applied statistics"] = "Ann Appl Stat",
  ["the annals of long-term care : the official journal of the american medical directors association"] = "Ann Longterm Care",
  ["the annals of occupational hygiene"] = "Ann Occup Hyg",
  ["the annals of otology, rhinology & laryngology. supplement"] = "Ann Otol Rhinol Laryngol Suppl",
  ["the annals of otology, rhinology, and laryngology"] = "Ann Otol Rhinol Laryngol",
  ["the annals of pharmacotherapy"] = "Ann Pharmacother",
  ["the annals of physiological anthropology = seiri jinruigaku kenkyūkai kaishi"] = "Ann Physiol Anthropol",
  ["the annals of probability"] = "Ann. Probab.",
  ["the annals of regional science"] = "Ann Reg Sci",
  ["the annals of statistics"] = "Ann. Statist.",
  ["the annals of the american academy of political and social science"] = "Ann Am Acad Pol Soc Sci",
  ["the annals of the chinese historical society of the pacific northwest"] = "Ann Chin Hist Soc Pac Northwest",
  ["the annals of the chinese historical society of the pacific northwest / mei-kuo hsi pei hua jen li shih hsueh hui"] = "Ann Chin Hist Soc Pac Northwest",
  ["the annals of the new york academy of dentistry"] = "Annals N Y Acad Dent",
  ["the annals of thoracic surgery"] = "Ann Thorac Surg",
  ["the annotated code of the public general laws of maryland. health-general. maryland"] = "Annot Code Public Gen Laws Md Md",
  ["the annual of leeds university oriental society"] = "AnnLeedsUnOrSoc",
  ["the annual of the american schools of oriental research"] = "AASO",
  ["the annual of the british school at athens"] = "BSA",
  ["the annual of the society of christian ethics. society of christian ethics (u.s.)"] = "Annu Soc Christ Ethics",
  ["the annual report of the fogg art museum"] = "AnnRepFoggArtMus",
  ["the annual report of the library company of philadelphia. library company of philadelphia"] = "Annu Rep Libr Co Phila",
  ["the antioch review"] = "Antioch Rev",
  ["the antiquaries journal"] = "AntJ",
  ["the antiquaries journal: being the journal of the society of antiquaries of london"] = "AntJ",
  ["the antiseptic"] = "Antiseptic",
  ["the anziam journal"] = "ANZIAM J.",
  ["the anziam journal. the australian & new zealand industrial and applied mathematics journal"] = "ANZIAM J.",
  ["the apdt chronicle of the dog"] = "APDT Chron Dog",
  ["the apothecary"] = "Apoth (Beirut)",
  ["the application of clinical genetics"] = "Appl Clin Genet",
  ["the appraisal journal"] = "Appraisal J",
  ["the arab economist"] = "Arab Econ",
  ["the arab gulf journal"] = "Arab Gulf J",
  ["the arab journal of laboratory medicine"] = "Arab J Lab Med",
  ["the arab journal of the social sciences"] = "Arab J Soc Sci",
  ["the arabian journal for science and engineering"] = "Arab. J. Sci. Eng. Sect. A Sci.",
  ["the arabidopsis book"] = "Arabidopsis Book",
  ["the archaeological advertiser"] = "AAdv",
  ["the archaeological journal"] = "AJ",
  ["the archaeological news letter"] = "ANL",
  ["the archimedes palimpsest publications"] = "Archimedes Palimpsest Publ.",
  ["the archives of bone and joint surgery"] = "Arch Bone Jt Surg",
  ["the archivist"] = "Archivist",
  ["the arizona dental journal"] = "Ariz Dent J",
  ["the arizona nurse"] = "Ariz Nurse",
  ["the ark-light newsletter"] = "Ark Light Newsl",
  ["the arkansas dental journal"] = "Ark Dent J",
  ["the arkansas historical quarterly"] = "Ark Hist Q",
  ["the armenian review"] = "Armen Rev",
  ["the army lawyer"] = "Army Lawyer",
  ["the army quarterly and defence journal"] = "Army Q Def J",
  ["the art bulletin"] = "Art Bull",
  ["the art bulletin: a quarterly published by the college art association of america"] = "ABull",
  ["the art of discrete and applied mathematics"] = "Art Discrete Appl. Math.",
  ["the art of medication"] = "Art Med",
  ["the artech house microwave library"] = "Artech House Microw. Lib.",
  ["the artech house signal processing library"] = "Artech House Signal Process. Libr.",
  ["the articulator"] = "Articulator (Syd)",
  ["the arts in psychotherapy"] = "Arts Psychother",
  ["the asbury theological journal"] = "Asbury Theol J",
  ["the asia foundation news"] = "Aisa Found News",
  ["the asia magazine"] = "Asia Mag",
  ["the asian and australasian journal of plant science and biotechnology"] = "Asian Australas J Plant Sci Biotechnol",
  ["the asian economic review"] = "Asian Econ Rev",
  ["the asian journal of economics : a journal of the society for asian development"] = "Asian J Econ",
  ["the asian journal of mathematics"] = "Asian J. Math.",
  ["the assyrian dictionary of the oriental institute of the university of chicago"] = "CAD",
  ["the astronomical journal"] = "Astron J",
  ["the astrophysical journal"] = "Astrophys J",
  ["the astrophysical journal letters"] = "Astrophys. J. Lett.",
  ["the astrophysical journal supplement series"] = "Astrophys. J. Suppl. Ser.",
  ["the astrophysical journal. letters"] = "Astrophys J Lett",
  ["the astrophysical journal. supplement series"] = "Astrophys J Suppl Ser",
  ["the athenian agora"] = "Agora",
  ["the atlanta historical journal"] = "Atlanta Hist J.",
  ["the auk"] = "Auk",
  ["the auk: ornithological advances"] = "The Auk: Ornithol. Adv.",
  ["the australasian catholic record"] = "Australas Cathol Rec",
  ["the australasian journal of combinatorics"] = "Australas. J. Combin.",
  ["the australasian journal of dermatology"] = "Australas J Dermatol",
  ["the australasian journal of logic"] = "Australas. J. Log.",
  ["the australasian journal of optometry"] = "Australas J Optom",
  ["the australasian journal of pharmacy"] = "Australas J Pharm",
  ["the australasian medical journal"] = "Australas Med J",
  ["the australasian nurses journal"] = "Australas Nurses J",
  ["the australasian nurses' journal"] = "Australas Nurses J",
  ["the australian & new zealand grapegrower & winemaker"] = "Aust N Z Grapegrow Winemak",
  ["the australian & new zealand journal of criminology"] = "Aust N Z J Criminol",
  ["the australian & new zealand journal of obstetrics & gynaecology"] = "Aust N Z J Obstet Gynaecol",
  ["the australian & new zealand journal of obstetrics & gynaecology. supplement"] = "Aust N Z J Obstet Gynaecol Suppl",
  ["the australian and new zealand general practitioner"] = "Aust N Z Gen Pract",
  ["the australian and new zealand journal of family therapy"] = "Aust N Z J Fam Ther",
  ["the australian and new zealand journal of mental health nursing"] = "Aust N Z J Ment Health Nurs",
  ["the australian and new zealand journal of psychiatry"] = "Aust N Z J Psychiatry",
  ["the australian and new zealand journal of sociology"] = "Aust N Z J Sociol",
  ["the australian and new zealand journal of surgery"] = "Aust N Z J Surg",
  ["the australian dental mirror, incorporating the queensland dental magazine"] = "Aust Dent mirror",
  ["the australian economic review"] = "Aust Econ Rev",
  ["the australian economic review / institute of applied economic research"] = "Aust Econ Rev",
  ["the australian entomologist"] = "Aust Entomol (Brisb)",
  ["the australian geographer"] = "Aust Geogr",
  ["the australian hospital"] = "Aust Hosp",
  ["the australian journal of advanced nursing : a quarterly publication of the royal australian nursing federation"] = "Aust J Adv Nurs",
  ["the australian journal of agricultural and resource economics"] = "Aust J Agric Resour Econ",
  ["the australian journal of anthropology"] = "Aust J Anthropol",
  ["the australian journal of chinese affairs = ao chung"] = "Aust J Chin Aff",
  ["the australian journal of clinical hypnotherapy"] = "Aust J Clin Hypnother",
  ["the australian journal of clinical hypnotherapy / australian society of clinical hypnotherapists"] = "Aust J Clin Hypnother",
  ["the australian journal of dentistry"] = "Aust J Dent",
  ["the australian journal of experimental biology and medical science"] = "Aust J Exp Biol Med Sci",
  ["the australian journal of forensic sciences"] = "Aust J Forensic Sci",
  ["the australian journal of holistic nursing"] = "Aust J Holist Nurs",
  ["the australian journal of mathematical analysis and applications"] = "Aust. J. Math. Anal. Appl.",
  ["the australian journal of mental retardation"] = "Aust J Ment Retard",
  ["the australian journal of optometry"] = "Aust J Optom",
  ["the australian journal of physiotherapy"] = "Aust J Physiother",
  ["the australian journal of politics and history"] = "Aust J Polit Hist",
  ["the australian journal of rural health"] = "Aust J Rural Health",
  ["the australian journal of science"] = "Aust J Sci",
  ["the australian journal of social issues"] = "Aust J Soc Issues",
  ["the australian journal of statistics"] = "Aust J Stat",
  ["the australian law journal"] = "Aust Law J",
  ["the australian library journal"] = "Aust Libr J",
  ["the australian mathematical society"] = "Austral. Math. Soc. Gaz.",
  ["the australian mathematical society. gazette"] = "Austral. Math. Soc. Gaz.",
  ["the australian nurses' journal. royal australian nursing federation"] = "Aust Nurses J",
  ["the australian nurses’ journal"] = "Aust Nurses J",
  ["the australian outlook"] = "Aust Outlook",
  ["the australian quarterly"] = "Aust Q",
  ["the australian zoologist"] = "Aust Zool",
  ["the b. e. journal of theoretical economics"] = "B. E. J. Theor. Econ.",
  ["the b.e. journal of economic analysis & policy"] = "B E J Econom Anal Policy",
  ["the bangkok medical journal"] = "Bangk Med J",
  ["the bangkok post"] = "Bangkok Post",
  ["the bangladesh journal of scientific research"] = "Bangladesh J Sci Res",
  ["the baptist quarterly"] = "Baptist Q",
  ["the baxter health policy review"] = "Baxter Health Policy Rev",
  ["the baylor dental journal"] = "Baylor Dent J",
  ["the beaver"] = "Beaver",
  ["the behavior analyst"] = "Behav Anal",
  ["the behavior analyst / maba"] = "Behav Anal",
  ["the behavior analyst today"] = "Behav Anal Today",
  ["the behavior therapist"] = "Behav Ther (N Y N Y)",
  ["the behavioral and brain sciences"] = "Behav Brain Sci",
  ["the bekhterev review of psychiatry and medical psychology = obozrenie psikhiatrii i medit︠s︡inskoĭ psikhologii imeni v.m. bekhtereva"] = "Bekhterev Rev Psychiatry Med Psychol",
  ["the bengal public health journal"] = "Bengal Public Health J",
  ["the best evidence for health care : the role of the cochrane collaboration. cochrane colloquium (7th : 1999 : università s. tommaso d'aquino)"] = "Best Evid Health Care Cochrane Colloq 7th 1999 Univ S Tommaso Daquino",
  ["the biblical archaeologist"] = "Biblic Archaeol",
  ["the biblical archaeologist. the american school of oriental research new haven"] = "BibAr",
  ["the bibliotheck"] = "Bibliotheck",
  ["the big questions"] = "Big Quest.",
  ["the biochemical journal"] = "Biochem J",
  ["the biochemist"] = "Biochem (Lond)",
  ["the bioethics bulletin"] = "Bioeth Bull",
  ["the biological bulletin"] = "Biol Bull",
  ["the biological review of the city college"] = "Biol Rev City Coll",
  ["the bioscan (ranchi)"] = "Bioscan (Ranchi)",
  ["the birth gazette"] = "Birth Gaz",
  ["the black law journal"] = "Black Law J",
  ["the black scholar"] = "Black Scholar",
  ["the blue sheet"] = "Blue Sheet",
  ["the bodleian library record"] = "Bodleian Libr Rec",
  ["the body forum"] = "Body Forum",
  ["the body positive"] = "Body Posit",
  ["the bombay hospital journal"] = "Bombay Hosp J",
  ["the bone & joint journal"] = "Bone Joint J",
  ["the book & paper group annual"] = "Book Pap Group Annu",
  ["the book collector"] = "Book Collect",
  ["the bookman"] = "Bookman",
  ["the boston globe"] = "Boston Globe",
  ["the boston medical and surgical journal"] = "Boston Med Surg J",
  ["the boston university public interest law journal"] = "Boston Univ Public Interest Law J",
  ["the botanical review; interpreting botanical progress"] = "Bot Rev",
  ["the botulinum journal"] = "Botulinum J",
  ["the bovine practitioner"] = "Bov Pract (Stillwater)",
  ["the brazilian journal of infectious diseases : an official publication of the brazilian society of infectious diseases"] = "Braz J Infect Dis",
  ["the breast journal"] = "Breast J",
  ["the bridge"] = "Bridge (Kans City)",
  ["the bristol digest"] = "Bristol Dig",
  ["the british academy classical and medieval logic texts"] = "British Acad. Classical Medieval Logic Texts",
  ["the british and foreign medical review"] = "Br Foreign Med Rev",
  ["the british and foreign medico-chirurgical review"] = "Br Foreign Med Chir Rev",
  ["the british and irish orthoptic journal"] = "Br Ir Orthopt J",
  ["the british chemical digest"] = "Br Chem Dig",
  ["the british chiropody journal"] = "Br Chirop J",
  ["the british dental surgery assistant"] = "Br Dent Surg Assist",
  ["the british food journal and hygienic review"] = "Br Food J Hyg Rev",
  ["the british homoeopathic journal"] = "Br Homeopath J",
  ["the british journal for eighteenth-century studies"] = "Br J 18th Cent Stud",
  ["the british journal for the philosophy of science"] = "Br J Philos Sci",
  ["the british journal of addiction to alcohol and other drugs"] = "Br J Addict Alcohol Other Drugs",
  ["the british journal of cancer. supplement"] = "Br J Cancer Suppl",
  ["the british journal of cardiology"] = "Br J Cardiol",
  ["the british journal of clinical practice"] = "Br J Clin Pract",
  ["the british journal of clinical psychology"] = "Br J Clin Psychol",
  ["the british journal of clinical psychology / the british psychological society"] = "Br J Clin Psychol",
  ["the british journal of criminology"] = "Br J Criminol",
  ["the british journal of dermatology"] = "Br J Dermatol",
  ["the british journal of dermatology and syphilis"] = "Br J Dermatol Syph",
  ["the british journal of developmental psychology"] = "Br J Dev Psychol",
  ["the british journal of diabetes & vascular disease"] = "Br J Diabetes Vasc Dis",
  ["the british journal of disorders of communication"] = "Br J Disord Commun",
  ["the british journal of educational psychology"] = "Br J Educ Psychol",
  ["the british journal of family planning"] = "Br J Fam Plann",
  ["the british journal of general practice : the journal of the royal college of general practitioners"] = "Br J Gen Pract",
  ["the british journal of mathematical and statistical psychology"] = "Br J Math Stat Psychol",
  ["the british journal of medical psychology"] = "Br J Med Psychol",
  ["the british journal of nursing"] = "Br J Nurs",
  ["the british journal of nutrition"] = "Br J Nutr",
  ["the british journal of occupational therapy"] = "Br J Occup Ther",
  ["the british journal of ophthalmology"] = "Br J Ophthalmol",
  ["the british journal of oral & maxillofacial surgery"] = "Br J Oral Maxillofac Surg",
  ["the british journal of oral surgery"] = "Br J Oral Surg",
  ["the british journal of physical medicine : including its application to industry"] = "Br J Phys Med",
  ["the british journal of physical medicine, including its application to industry"] = "Br J Phys Med",
  ["the british journal of physiological optics"] = "Br J Physiol Opt",
  ["the british journal of psychiatry : the journal of mental science"] = "Br J Psychiatry",
  ["the british journal of psychiatry. supplement"] = "Br J Psychiatry Suppl",
  ["the british journal of psychology. general section"] = "Br J Psychol Gen Sect",
  ["the british journal of radiology"] = "Br J Radiol",
  ["the british journal of social and clinical psychology"] = "Br J Soc Clin Psychol",
  ["the british journal of social psychology"] = "Br J Soc Psychol",
  ["the british journal of social psychology / the british psychological society"] = "Br J Soc Psychol",
  ["the british journal of sociology"] = "Br J Sociol",
  ["the british journal of sports history"] = "Br J Sports Hist",
  ["the british journal of surgery"] = "Br J Surg",
  ["the british journal of theatre nursing : natnews : the official journal of the national association of theatre nurses"] = "Br J Theatre Nurs",
  ["the british journal of tuberculosis and diseases of the chest"] = "Br J Tuberc Dis Chest",
  ["the british journal of venereal diseases"] = "Br J Vener Dis",
  ["the british journal of visual impairment"] = "Br J Vis Impair",
  ["the british library journal. british library"] = "Br Libr J",
  ["the british museum quarterly"] = "BMQ",
  ["the british museum quarterly. news supplement"] = "BMQNSuppl",
  ["the british museum yearbook"] = "BrMusYearbook",
  ["the british orthoptic journal"] = "Br Orthopt J",
  ["the british studies monitor"] = "Br Stud Monit",
  ["the british veterinary journal"] = "Br Vet J",
  ["the broadway"] = "Broadway",
  ["the brookings review"] = "Brookings Rev",
  ["the brooklyn hospital journal"] = "Brooklyn Hosp J",
  ["the bryologist"] = "Bryologist",
  ["the bulletin - american association of dental editors"] = "Bull Am Assoc Dent Ed",
  ["the bulletin - essex county dental society"] = "Bull Essex Cty Dent Soc",
  ["the bulletin - missouri historical society. missouri historical society"] = "Bull Mo Hist Soc",
  ["the bulletin [of] the university of miami school of medicine and jackson memorial hospital. university of miami. school of medicine"] = "Bull Univ Miami Sch Med Jackson Meml Hosp",
  ["the bulletin of akron city hospital"] = "Bull Akron City Hosp",
  ["the bulletin of louis a. weiss memorial hospital"] = "Bull Louis A Weiss Mem Hosp",
  ["the bulletin of mathematical biophysics"] = "Bull Math Biophys",
  ["the bulletin of symbolic logic"] = "Bull. Symb. Log.",
  ["the bulletin of the academy of general dentistry"] = "Bull Acad Gen Dent",
  ["the bulletin of the american academy of psychiatry and the law"] = "Bull Am Acad Psychiatry Law",
  ["the bulletin of the american college of physicians"] = "Bull Am Coll Physicians",
  ["the bulletin of the american society of papyrologists"] = "BAmSocP",
  ["the bulletin of the american society of papyrologists. american society of papyrologists"] = "Bull Am Soc Papyrol",
  ["the bulletin of the atomic scientists"] = "Bull At Sci",
  ["the bulletin of the bergen county dental society"] = "Bull Bergen Cty Dent Soc",
  ["the bulletin of the cincinnati dental society"] = "Bull Cincinnati Dent Soc",
  ["the bulletin of the cleveland dental society"] = "Bull Cleve Dent Soc",
  ["the bulletin of the cleveland medical library. cleveland medical library association"] = "Bull Clevel Med Libr",
  ["the bulletin of the cleveland museum of art"] = "BClevMus",
  ["the bulletin of the fifth district dental society"] = "Bull Fifth Dist Dent Soc (Fresno)",
  ["the bulletin of the high institute of public health"] = "Bull High Inst Public Health",
  ["the bulletin of the illinois dental hygienist's association. illinois dental hygienists' association"] = "Bull Ill Dent Hyg Assoc",
  ["the bulletin of the illinois dental hygienist’s association. illinois dental hygienists’ association"] = "Bull Ill Dent Hyg Assoc",
  ["the bulletin of the kanagawa dental college : bkdc"] = "Bull Kanagawa Dent Coll",
  ["the bulletin of the kanagawa dental college : bkdc / kds"] = "Bull Kanagawa Dent Coll",
  ["the bulletin of the london mathematical society"] = "Bull Lond Math Soc",
  ["the bulletin of the michigan dental hygienists' association"] = "Bull Mich Dent Hyg Assoc",
  ["the bulletin of the michigan dental hygienists’ association"] = "Bull Mich Dent Hyg Assoc",
  ["the bulletin of the michigan state dental hygienists' association"] = "Bull Mich State Dent Hyg Assoc",
  ["the bulletin of the orange county medical association"] = "Bull Orange Cty Med Assoc",
  ["the bulletin of the passaic county medical society"] = "Bull Passaic Cty Med Soc",
  ["the bulletin of the philadelphia county dental society"] = "Bull Phila Cty Dent Soc",
  ["the bulletin of the society of university cartographers. society of university cartographers"] = "Bull Soc Univ Cartogr",
  ["the bulletin of the sri lanka college of microbiologists"] = "Bull Sri Lanka Coll Microbiol",
  ["the bulletin of the toledo dental society"] = "Bull Toledo Dent Soc",
  ["the bulletin of the tri-county dental society"] = "Bull Tri Cty Dent Soc",
  ["the bulletin of the tulane medical faculty"] = "Bull Tulane Univ Med Fac",
  ["the bulletin of the virginia state dental association"] = "Bull Va State Dent Assoc",
  ["the bulletin of the yamaguchi medical school"] = "Bull Yamaguchi Med Sch",
  ["the bulletin of tokyo dental college"] = "Bull Tokyo Dent Coll",
  ["the bulletin of tokyo medical and dental university"] = "Bull Tokyo Med Dent Univ",
  ["the bulletin of zoological nomenclature"] = "Bull Zool Nomencl",
  ["the bur"] = "Bur",
  ["the burlington magazine"] = "Burlingt Mag",
  ["the calcutta historical journal"] = "Calcutta Hist J",
  ["the california clinician"] = "Calif Clin",
  ["the california journal of emergency medicine"] = "Cal J Emerg Med",
  ["the california journal of health-system pharmacy : cjhp"] = "Calif J Health Syst Pharm",
  ["the california school psychologist : casp"] = "Calif School Psychol",
  ["the california school psychologist : casp / california association of school psychologists"] = "Calif School Psychol",
  ["the californians"] = "Californians",
  ["the caliper"] = "Caliper",
  ["the cambridge ancient history"] = "CAH",
  ["the cambridge law journal"] = "Camb Law J",
  ["the canadian"] = "Canadian",
  ["the canadian applied mathematics quarterly"] = "Canad. Appl. Math. Quart.",
  ["the canadian bar review"] = "Can Bar Rev",
  ["the canadian child and adolescent psychiatry review = la revue canadienne de psychiatrie de l'enfant et de l'adolescent"] = "Can Child Adolesc Psychiatr Rev",
  ["the canadian dental hygienist"] = "Can Dent Hyg",
  ["the canadian entomologist"] = "Can Entomol",
  ["the canadian field-naturalist"] = "Can Field Nat",
  ["the canadian forum"] = "Can Forum",
  ["the canadian geographer. geographe canadien"] = "Can Geogr",
  ["the canadian geographer. géographe canadien"] = "Can Geogr",
  ["the canadian historical review"] = "Can Hist Rev",
  ["the canadian journal of addiction"] = "Can J Addict",
  ["the canadian journal of cardiology"] = "Can J Cardiol",
  ["the canadian journal of chemical engineering"] = "Can J Chem Eng",
  ["the canadian journal of clinical pharmacology = journal canadien de pharmacologie clinique"] = "Can J Clin Pharmacol",
  ["the canadian journal of community dentistry = la revue canadienne de la dentisterie communautaire"] = "Can J Community Dent",
  ["the canadian journal of critical care nursing"] = "Can J Crit Care Nurs",
  ["the canadian journal of economics. revue canadienne d'économique"] = "Can J Econ",
  ["the canadian journal of economics. revue canadienne d’economique"] = "Can J Econ",
  ["the canadian journal of hospital pharmacy"] = "Can J Hosp Pharm",
  ["the canadian journal of human sexuality"] = "Can J Hum Sex",
  ["the canadian journal of infection control : the official journal of the community & hospital infection control association-canada = revue canadienne de prévention des infections"] = "Can J Infect Control",
  ["the canadian journal of infectious diseases & medical microbiology = journal canadien des maladies infectieuses et de la microbiologie médicale"] = "Can J Infect Dis Med Microbiol",
  ["the canadian journal of infectious diseases = journal canadien des maladies infectieuses"] = "Can J Infect Dis",
  ["the canadian journal of law and jurisprudence"] = "Can J Law Jurisprud",
  ["the canadian journal of medical radiation technology"] = "Can J Med Radiat Technol",
  ["the canadian journal of medical radiation technology / camrt"] = "Can J Med Radiat Technol",
  ["the canadian journal of native studies"] = "Can J Native Stud",
  ["the canadian journal of neurological sciences. le journal canadien des sciences neurologiques"] = "Can J Neurol Sci",
  ["the canadian journal of nursing research = revue canadienne de recherche en sciences infirmières"] = "Can J Nurs Res",
  ["the canadian journal of ob/gyn & women's health care"] = "Can J Ob Gyn Womens Health Care",
  ["the canadian journal of ob/gyn & women’s health care"] = "Can J Ob Gyn Womens Health Care",
  ["the canadian journal of oncology"] = "Can J Oncol",
  ["the canadian journal of plastic surgery = journal canadien de chirurgie plastique"] = "Can J Plast Surg",
  ["the canadian journal of program evaluation = la revue canadienne d'évaluation de programme"] = "Can J Program Eval",
  ["the canadian journal of psychiatric nursing"] = "Can J Psychiatr Nurs",
  ["the canadian journal of radiography, radiotherapy, nuclear medicine"] = "Can J Radiogr Radiother Nucl Med",
  ["the canadian journal of regional science = la revue canadienne des sciences régionales"] = "Can J Reg Sci",
  ["the canadian journal of statistics"] = "Canad. J. Statist.",
  ["the canadian journal of statistics = revue canadienne de statistique"] = "Can J Stat",
  ["the canadian journal of statistics. la revue canadienne de statistique"] = "Canad. J. Statist.",
  ["the canadian journal of urology"] = "Can J Urol",
  ["the canadian journal of women's health care for physicians addressing women's health issues"] = "Can J Womens Health Care Phys Addressing Womens Health Issues",
  ["the canadian journal of women’s health care for physicians addressing women’s health issues"] = "Can J Womens Health Care Phys Addressing Womens Health Issues",
  ["the canadian modern language review = la revue canadienne des langues vivantes"] = "Can Mod Lang Rev",
  ["the canadian nurse"] = "Can Nurse",
  ["the canadian nurse and hospital review"] = "Can Nurse Hosp Rev",
  ["the canadian review of sociology and anthropology. la revue canadienne de sociologie et d'anthropologie"] = "Can Rev Sociol Anthropol",
  ["the canadian review of sociology and anthropology. la revue canadienne de sociologie et d’anthropologie"] = "Can Rev Sociol Anthropol",
  ["the canadian veterinary journal = la revue vétérinaire canadienne"] = "Can Vet J",
  ["the canadian veterinary journal. la revue veterinaire canadienne"] = "Can Vet J",
  ["the canadian yearbook of international law. annuaire canadien de droit international"] = "Can Yearb Int Law",
  ["the cancer journal from scientific american"] = "Cancer J Sci Am",
  ["the capital chemist"] = "Cap Chem",
  ["the career development quarterly"] = "Career Dev Q",
  ["the carl g. hempel lecture series"] = "Carl G. Hempel Lect. Ser.",
  ["the carmarthen antiquary"] = "Carmarthen Antiq",
  ["the carmarthensire hisorian"] = "Carmarthensh Hist",
  ["the carolina journal of pharmacy"] = "Carol J Pharm",
  ["the cartographic journal"] = "Cartogr J",
  ["the case manager"] = "Case Manager",
  ["the catholic historical review"] = "Cathol Hist Rev",
  ["the catholic lawyer"] = "Cathol Lawyer",
  ["the catholic medical quarterly"] = "Cathol Med Q",
  ["the catholic mind"] = "Cathol Mind",
  ["the catholic nurse"] = "Cathol Nurse",
  ["the cato journal"] = "Cato J",
  ["the ccl news"] = "CCL News",
  ["the centaur of alpha kappa kappa"] = "Centaur Alpha Kappa Kappa",
  ["the centennial review"] = "Centen Rev",
  ["the center magazine"] = "Cent Mag",
  ["the central african journal of medicine"] = "Cent Afr J Med",
  ["the cerebral palsy journal"] = "Cereb Palsy J",
  ["the cervix and the lower female genital tract"] = "Cervix Low Female Genital Tract",
  ["the ceylon medical journal"] = "Ceylon Med J",
  ["the chaucer review"] = "Chaucer Rev",
  ["the chemist and druggist"] = "Chem Drug",
  ["the chemist and druggist export review"] = "Chem Drug Export Rev",
  ["the chesopiean"] = "Chesopiean",
  ["the chesterton review"] = "Chesterton Rev",
  ["the chicago medical recorder"] = "Chic Med Rec",
  ["the chicago medical school quarterly"] = "Chic Med Sch Q",
  ["the child"] = "Child",
  ["the china business review"] = "China Bus Rev",
  ["the china geographer"] = "Chnia Geogr",
  ["the china quarterly"] = "China Q",
  ["the chinese journal of dental research : the official journal of the scientific section of the chinese stomatological association (csa)"] = "Chin J Dent Res",
  ["the chinese journal of physiology"] = "Chin J Physiol",
  ["the chinese review of tropical medicine"] = "Chin Rev Trop Med",
  ["the chronicle"] = "Chronicle",
  ["the chronicle of higher education"] = "Chron High Educ",
  ["the chronicle of mentoring & coaching"] = "Chron Mentor Coach",
  ["the cincinnati art museum bulletin"] = "CincArtB",
  ["the cincinnati dental society bulletin"] = "Cincinnati Dent Soc Bull",
  ["the cincinnati historical society bulletin"] = "Cinci Hist Soc Bull",
  ["the civil liberties review"] = "Civ Liberties Rev",
  ["the clao journal : official publication of the contact lens association of ophthalmologists, inc"] = "CLAO J",
  ["the classical bulletin"] = "CB",
  ["the classical journal"] = "Class J",
  ["the classical outlook"] = "Class Outlook",
  ["the classical outlook: journal of the american classical league"] = "CO",
  ["the classical quarterly"] = "ClQ",
  ["the classical review"] = "Classical Rev",
  ["the classical world"] = "CW",
  ["the cleft palate journal"] = "Cleft Palate J",
  ["the cleft palate-craniofacial journal : official publication of the american cleft palate-craniofacial association"] = "Cleft Palate Craniofac J",
  ["the clergy review"] = "Clergy Rev",
  ["the clinical biochemist. reviews"] = "Clin Biochem Rev",
  ["the clinical biochemist. reviews / australian association of clinical biochemists"] = "Clin Biochem Rev",
  ["the clinical bulletin of the school of medicine of western reserve university and its associated hospitals. western reserve university. school of medicine"] = "Clin Bull Sch Med West Reserve Univ Assoc Hosp",
  ["the clinical bulletin. university hospitals of cleveland (ohio)"] = "Clin Bull Univ Hosp Clevel (Ohio)",
  ["the clinical investigator"] = "Clin Investig",
  ["the clinical journal"] = "Clin J",
  ["the clinical journal of pain"] = "Clin J Pain",
  ["the clinical neuropsychologist"] = "Clin Neuropsychol",
  ["the clinical psychologist"] = "Clin Psychol",
  ["the clinical respiratory journal"] = "Clin Respir J",
  ["the clinical supervisor"] = "Clin Superv",
  ["the clinical teacher"] = "Clin Teach",
  ["the clinician"] = "Clinician (Goa)",
  ["the cobasheca (community based health care) newsletter"] = "COBASHECA",
  ["the cochrane database of systematic reviews"] = "Cochrane Database Syst Rev",
  ["the cochrane library"] = "Cochrane Libr",
  ["the coevolution quarterly"] = "Coevol Q",
  ["the colby library quarterly"] = "Colby Libr Q",
  ["the coleopterists' bulletin"] = "Coleopt Bull",
  ["the college mathematics journal"] = "College Math. J.",
  ["the colorado magazine"] = "Colo Mag",
  ["the columbia journal of world business"] = "Columbia J World Bus",
  ["the columbia optometrist"] = "Columbia Optom",
  ["the columbia science and technology law review"] = "Columbia Sci Technol Law Rev",
  ["the comparative and international law journal of southern africa = tydskrif vir regsvergelyking en internasionale reg van suidelik afrika = jornal de direito comparativo e internacional para os paises do sul da africa = journal de droit comparé et international des pays de l'afrique australe = zeitschrift für rechtsvergleichung und internationales recht des südlichen afrika"] = "Comp Int Law J South Afr",
  ["the compendium of continuing education in dentistry"] = "Compend Contin Educ Dent",
  ["the compendium on continuing education in general dentistry"] = "Compend Contin Educ Gen Dent",
  ["the computer journal"] = "Comput J",
  ["the computer science library"] = "Comput. Sci. Lib.",
  ["the condor"] = "Condor",
  ["the condor: ornithological applications"] = "The Condor: Ornithol. Appl.",
  ["the conference board record"] = "Conf Board Rec",
  ["the connecticut historical society bulletin"] = "Conn Hist Soc Bull",
  ["the connecticut probate law journal"] = "Conn Probate Law J",
  ["the consultant pharmacist : the journal of the american society of consultant pharmacists"] = "Consult Pharm",
  ["the contemporary pacific"] = "Contemp Pac",
  ["the control of nature"] = "Control Nat.",
  ["the cornell hotel and restaurant administration quarterly"] = "Cornell Hotel Restaur Adm Q",
  ["the cornell journal of social relations"] = "Cornell J Soc Relat",
  ["the cornell veterinarian"] = "Cornell Vet",
  ["the counseling psychologist"] = "Couns Psychol",
  ["the courier"] = "Courier",
  ["the crippled child"] = "Crippl Child",
  ["the crispr journal"] = "CRISPR J",
  ["the critical quarterly"] = "Crit Q",
  ["the crop journal"] = "Crop J",
  ["the crown agents review"] = "Crown Agents Rev",
  ["the cryosphere"] = "Cryosphere",
  ["the cte journal"] = "CTE J",
  ["the current digest of the post-soviet press"] = "Curr Dig Post Sov Press",
  ["the current digest of the soviet press"] = "Curr Dig Sov Press",
  ["the curriculum journal"] = "Curric J",
  ["the cyber-physical systems series"] = "Cyber-Phys. Syst. Ser.",
  ["the dacca university studies: part a. university of dacca"] = "Dacca Univ Stud A",
  ["the daily washington law reporter"] = "Dly Wash Law Report",
  ["the dalhousie review"] = "Dalhousie Rev",
  ["the dallas medical journal"] = "Dallas Med J",
  ["the davis nursing survey"] = "Davis Nurs Surv",
  ["the deccan geographer"] = "Deccan Geogr",
  ["the dental assistant"] = "Dent Assist",
  ["the dental assistant journal : journal of the american dental assistants association"] = "Dent Assist J",
  ["the dental gazette"] = "Dent Gaz",
  ["the dental journal of australia"] = "Dent J Aust",
  ["the dental journal of malaysia & singapore"] = "Dent J Malaysia Singapore",
  ["the dental magazine and oral topics"] = "Dent Mag Oral Top",
  ["the dental outlook"] = "Dent Outlook",
  ["the dental practitioner and dental record"] = "Dent Pract Dent Rec",
  ["the dental record"] = "Dent Rec (London)",
  ["the dental technician"] = "Dent Tech",
  ["the denver westerners roundup"] = "Denver West Roundup",
  ["the department of state bulletin"] = "Dep State Bull",
  ["the depositional record : a journal of biological, physical and geochemical sedimentary processes"] = "Depos Rec",
  ["the developing economies"] = "Dev Econ",
  ["the devon historian"] = "Devon Hist",
  ["the dhaka university journal of pharmaceutical sciences"] = "Dhaka Univ J Pharm Sci",
  ["the dhaka university studies. part b"] = "Dhaka Univ Stud B Biol Stud",
  ["the diabetes educator"] = "Diabetes Educ",
  ["the dialog"] = "Dialog Fairleigh Dickinson Univ Sch Dent",
  ["the dialog / fairleigh dickinson university school of dentistry"] = "Dialog Fairleigh Dickinson Univ Sch Dent",
  ["the dickensian"] = "Dickensian",
  ["the digest"] = "Digest (Wash D C)",
  ["the diplomate"] = "Dipl",
  ["the doctor; a medical penny magazine"] = "Dr Med Penny Mag",
  ["the dolciani mathematical expositions"] = "Dolciani Math. Exp.",
  ["the drug monitor"] = "Drug Monit",
  ["the durham thomas harriot seminar. occasional paper"] = "Durham Thomas Harriot Sem.",
  ["the durham university journal. university of durham"] = "Durham Univ J",
  ["the duxbury advanced series in statistics and decision sciences"] = "Duxbury Adv. Ser. Statist. Decis. Sci.",
  ["the e-journal of neonatology research"] = "EJ Neonatol Res",
  ["the earth times"] = "Earth Times (N Y)",
  ["the east african geographical review"] = "East Afr Geogr Rev",
  ["the east african health research journal"] = "East Afr Health Res J",
  ["the east tennessee historical society's publications. east tennessee historical society"] = "East Tenn Soc Publ",
  ["the east tennessee historical society’s publications. east tennessee historical society"] = "East Tenn Soc Publ",
  ["the eastern anthropologist"] = "East Anthropol",
  ["the ecologist"] = "Ecologist",
  ["the econometrics journal"] = "Econom J",
  ["the economic and social review"] = "Econ Soc Rev (Irel)",
  ["the economic bulletin of ghana"] = "Econ Bull Ghana",
  ["the economic history review"] = "Econ Hist Rev",
  ["the economic journal of nepal"] = "Econ J Nepal",
  ["the economic record"] = "Econ Rec",
  ["the economic weekly"] = "Econ Wkly",
  ["the economics of information, communication and entertainment"] = "Econ. Inf. Commun. Entertain.",
  ["the ecumenical review"] = "Ecum Rev",
  ["the edinburgh monthly journal of medical science"] = "Edinb Mon J Med Sci",
  ["the educational forum"] = "Educ Forum",
  ["the educational record"] = "Educ Rec",
  ["the egyptian heart journal : (ehj) : official bulletin of the egyptian society of cardiology"] = "Egypt Heart J",
  ["the egyptian journal of chest diseases and tuberculosis"] = "Egypt J Chest Dis Tuberc",
  ["the egyptian journal of haematology : the official journal of the egyptian society of haematology"] = "Egypt J Haematol",
  ["the egyptian journal of hospital medicine"] = "Egypt J Hosp Med",
  ["the egyptian journal of immunology"] = "Egypt J Immunol",
  ["the egyptian journal of immunology / egyptian association of immunologists"] = "Egypt J Immunol",
  ["the egyptian journal of internal medicine"] = "Egypt J Intern Med",
  ["the egyptian journal of neurology, psychiatry and neurosurgery"] = "Egypt J Neurol Psychiatr Neurosurg",
  ["the egyptian journal of psychiatry : official journal of the egyptian psychiatric association"] = "Egypt J Psychiatry",
  ["the egyptian population and family planning review"] = "Egypt Popul Fam Plann Rev",
  ["the egyptian statistical journal"] = "Egyptian Statist. J.",
  ["the einstein journal of biology and medicine : ejbm"] = "Einstein J Biol Med",
  ["the einstein quarterly journal of biology and medicine"] = "Einstein Q J Biol Med",
  ["the elder law journal"] = "Elder Law J",
  ["the electrical engineering handbook series"] = "Electr. Eng. Handbook Ser.",
  ["the electrochemical society interface"] = "Electrochem Soc Interface",
  ["the electronic journal of australian and new zealand history"] = "Electron J Aust N Z Hist",
  ["the electronic journal of human sexuality"] = "Electronic J Hum Sex",
  ["the electronic journal of science education"] = "Electron J Sci Educ",
  ["the elementary school journal"] = "Elem Sch J",
  ["the embalmers' monthly and national funeral director"] = "Embalmers Mon Natl Funer Dir",
  ["the embo journal"] = "EMBO J",
  ["the emporia state research studies"] = "Emporia State Res Stud",
  ["the emporia state research studies / kansas state teachers college"] = "Emporia State Res Stud",
  ["the emt journal"] = "EMT J",
  ["the emu"] = "Emu",
  ["the endocrinologist"] = "Endocrinologist",
  ["the endodontic report"] = "Endod Rep",
  ["the engineering economist"] = "Eng Econ",
  ["the engineers journal"] = "Eng J",
  ["the english historical review"] = "Engl Hist Rev",
  ["the environmentalist"] = "Environmentalist",
  ["the enzymes"] = "Enzymes",
  ["the epma journal"] = "EPMA J",
  ["the essence of computing series"] = "Essence Comput. Ser.",
  ["the eta sigma gamman"] = "Eta Sigma Gamman",
  ["the ethiopian herald"] = "Ethiop Herald",
  ["the ethiopian journal of health development = yaʼityoṗyā ṭénā lemāt maṣḥét"] = "Ethiop J Health Dev",
  ["the eugenical news"] = "Eugen News",
  ["the eugenics review"] = "Eugen Rev",
  ["the eurasian journal of medicine"] = "Eurasian J Med",
  ["the eurobiotech journal"] = "Eurobiotech J",
  ["the european accounting review"] = "Eur Account Rev",
  ["the european journal of cardiovascular medicine"] = "Eur J Cardiovasc Med",
  ["the european journal of cognitive psychology"] = "Eur J Cogn Psychol",
  ["the european journal of contraception & reproductive health care : the official journal of the european society of contraception"] = "Eur J Contracept Reprod Health Care",
  ["the european journal of development research"] = "Eur J Dev Res",
  ["the european journal of developmental psychology"] = "Eur J Dev Psychol",
  ["the european journal of esthetic dentistry : official journal of the european academy of esthetic dentistry"] = "Eur J Esthet Dent",
  ["the european journal of general practice"] = "Eur J Gen Pract",
  ["the european journal of genetic toxicology"] = "Eur J Genet Mol Toxicol",
  ["the european journal of health economics : hepac : health economics in prevention and care"] = "Eur J Health Econ",
  ["the european journal of international affairs"] = "Eur J Int Aff",
  ["the european journal of medicine"] = "Eur J Med",
  ["the european journal of neuroscience"] = "Eur J Neurosci",
  ["the european journal of prosthodontics and restorative dentistry"] = "Eur J Prosthodont Restor Dent",
  ["the european journal of psychiatry"] = "Eur J Psychiatry",
  ["the european journal of surgery = acta chirurgica"] = "Eur J Surg",
  ["the european journal of surgery. supplement. : = acta chirurgica. supplement"] = "Eur J Surg Suppl",
  ["the european journal of the history of economic thought"] = "Eur J Hist Econ Thought",
  ["the european journal on artificial intelligence"] = "Eur. J. Artif. Intell.",
  ["the european legacy, toward new paradigms : journal of the international society for the study of european ideas"] = "Eur Leg Towar New Paradig",
  ["the european physical journal"] = "Eur. Phys. J. D",
  ["the european physical journal a"] = "Eur. Phys. J. A",
  ["the european physical journal a: hadrons and nuclei"] = "Eur. Phys. J. A",
  ["the european physical journal ap"] = "Eur. Phys. J. AP",
  ["the european physical journal ap: applied physics"] = "Eur. Phys. J. AP",
  ["the european physical journal b"] = "Eur. Phys. J. B",
  ["the european physical journal b. condensed matter and complex systems"] = "Eur. Phys. J. B",
  ["the european physical journal b: condensed matter and complex systems"] = "Eur. Phys. J. B",
  ["the european physical journal c"] = "Eur. Phys. J. C",
  ["the european physical journal c: particles and fields"] = "Eur. Phys. J. C",
  ["the european physical journal d"] = "Eur. Phys. J. D",
  ["the european physical journal d: atomic, molecular, optical and plasma physics"] = "Eur. Phys. J. D",
  ["the european physical journal e"] = "Eur. Phys. J. E",
  ["the european physical journal e: soft matter and biological physics"] = "Eur. Phys. J. E",
  ["the european physical journal h"] = "Eur. Phys. J. H",
  ["the european physical journal plus"] = "Eur. Phys. J. Plus",
  ["the european physical journal special topics"] = "Eur. Phys. J. Special Topics",
  ["the european physical journal. a, hadrons and nuclei"] = "Eur Phys J A Hadron Nucl",
  ["the european physical journal. b"] = "Eur Phys J B",
  ["the european physical journal. c, particles and fields"] = "Eur Phys J C Part Fields",
  ["the european physical journal. d, atomic, molecular, and optical physics"] = "Eur Phys J D At Mol Opt Phys",
  ["the european physical journal. e, soft matter"] = "Eur Phys J E Soft Matter",
  ["the european physical journal. special topics"] = "Eur Phys J Spec Top",
  ["the european respiratory journal"] = "Eur Respir J",
  ["the european respiratory journal : official journal of the european society for clinical respiratory physiology"] = "Eur Respir J",
  ["the european respiratory journal. supplement"] = "Eur Respir J Suppl",
  ["the european sports history review"] = "Eur Sport Hist Rev",
  ["the european studies journal"] = "Eur Stud J",
  ["the european zoological journal"] = "Eur Zool J",
  ["the euthanasia review"] = "Euthan Rev",
  ["the excavations at dura-europos conducted by yale university and the french academy of inscriptions and letters"] = "Dura-Europos",
  ["the explorer : the journal of the school of dentistry of the university of missouri at kansas city. university of missouri at kansas city. school of dentistry"] = "Explorer (Kansas City)",
  ["the expository times"] = "Expo Times",
  ["the extractive industries and society"] = "Extr Ind Soc",
  ["the family coordinator"] = "Fam Coord",
  ["the family law reporter"] = "Fam Law Rep (Wash D C)",
  ["the family life coordinator"] = "Fam Life Coord",
  ["the family planner"] = "Fam Plann (Palo Alto)",
  ["the family planning manager"] = "Fam Plan Manag",
  ["the family practice research journal"] = "Fam Pract Res J",
  ["the family psychologist : bulletin of the division of family psychology (43)"] = "Fam Psychol",
  ["the farrier's magazine : or, the archives of veterinary science"] = "Farriers Mag",
  ["the faseb journal : official publication of the federation of american societies for experimental biology"] = "FASEB J",
  ["the febs journal"] = "FEBS J",
  ["the federal court reports. australia. federal court"] = "Fed Court Rep",
  ["the federal law reports : being reports of cases decided by the federal courts (other than the high court), state courts exercising federal jurisdiction, and courts of territories"] = "Fed Law Rep",
  ["the federal reporter"] = "Fed Report",
  ["the female patient"] = "Female Patient (Parsippany)",
  ["the fern gazette"] = "Fern Gaz",
  ["the festivus : a publication of the san diego shell club"] = "Festivus",
  ["the fibonacci quarterly"] = "Fibonacci Quart.",
  ["the fibonacci quarterly. the official journal of the fibonacci association"] = "Fibonacci Quart.",
  ["the fight against disease"] = "Fight Against",
  ["the filson club history quarterly"] = "Filson Club Hist Q",
  ["the first annual conference on vaccine research : basic science, product development, clinical & field studies : final program and abstract book : may 30-june 1, 1998, marriott at metro center, washington, d.c. conference on vaccine research (1st : 1998 : washington, d.c.)"] = "First Annu Conf Vaccine Res Conf Vaccine Res 1st 1998 Wash D C",
  ["the first people"] = "First People",
  ["the fletcher forum"] = "Fletcher Forum",
  ["the fletcher forum of world affairs"] = "Fletcher Forum World Aff",
  ["the florida bar journal"] = "FLA Bar J",
  ["the florida entomologist"] = "Fla Entomol",
  ["the florida geographer"] = "Fla Geogr",
  ["the florida historical quarterly"] = "Fla Hist Q",
  ["the florida nurse"] = "Fla Nurse",
  ["the fookien times yearbook"] = "Fookien Times Yearb",
  ["the fordham urban law journal"] = "Fordham Urban Law J",
  ["the fortnightly review of the chicago dental society"] = "Fortn Rev Chic Dent Soc",
  ["the frank j. fabozzi series"] = "Frank J. Fabozzi Ser.",
  ["the frater of psi omega"] = "Frater Psi Omega",
  ["the freeman"] = "Freeman",
  ["the frontiers collection"] = "Front. Coll.",
  ["the functional orthodontist"] = "Funct Orthod",
  ["the furrow"] = "Furrow",
  ["the future of children"] = "Future Child",
  ["the future of children / center for the future of children, the david and lucile packard foundation"] = "Future Child",
  ["the futurist"] = "Futurist",
  ["the gallup opinion index"] = "Gallup Opin Index",
  ["the gardens' bulletin, singapore"] = "Gard Bull (Singapore)",
  ["the gastroenterologist"] = "Gastroenterologist",
  ["the gastroenterology report"] = "Gastroenterol Rep",
  ["the gazette of the egyptian paediatric association"] = "Gaz Egypt Paediatr Assoc",
  ["the gelfand mathematical seminars"] = "Gelfand Math. Sem.",
  ["the gem state rn news letter"] = "Gem State RN News Lett",
  ["the gender cg newsletter"] = "Gend CG Newsl",
  ["the general psychologist"] = "Gen Psychol",
  ["the genetic resource"] = "Genetic Resour",
  ["the geneva papers on risk and insurance. issues and practice"] = "Geneva Pap Risk Insur Issues Pract",
  ["the geneva risk and insurance review"] = "Geneva Risk Insur Rev",
  ["the geographical journal"] = "Geogr J",
  ["the geographical magazine"] = "Geogr Mag",
  ["the geographical observer"] = "Geogr Obs",
  ["the george washington journal of international law and economics"] = "George Washington J Int Law Econ",
  ["the george washington law review"] = "George Washington Law Rev",
  ["the george wright forum"] = "George Wright Forum",
  ["the georgetown journal of gender and the law"] = "Georget J Gend Law",
  ["the georgetown journal of legal ethics"] = "Georget J Leg Ethics",
  ["the georgetown law journal"] = "Georgetown Law J",
  ["the georgetown university right-to-life journal"] = "Georgetown Univ Right to Life J",
  ["the georgia historical quarterly"] = "Ga Hist Q",
  ["the georgia review"] = "Ga Rev",
  ["the gerontologist"] = "Gerontologist",
  ["the ghanaian nurse"] = "Ghana Nurse",
  ["the gifted child quarterly"] = "Gift Child Q",
  ["the glaxo volume; an occasional contribution to the science and art of medicine"] = "Glaxo Vol Occas Contrib Sci Art Med",
  ["the grace hospital bulletin"] = "Grace Hosp Bul",
  ["the graduate journal of mathematics"] = "Grad. J. Math.",
  ["the great basin naturalist"] = "Great Basin Nat",
  ["the great circle : journal of the australian association for maritime history"] = "Great Circle",
  ["the great concern"] = "Gt Concern",
  ["the great ideas today"] = "Great Ideas Today",
  ["the great lakes entomologist"] = "Great Lakes Entomol",
  ["the greater milwaukee dental bulletin"] = "Greater Milw Dent Bull",
  ["the greater st. louis dental society bulletin"] = "Greater St Louis Dent Soc Bull",
  ["the group health journal"] = "Group Health J",
  ["the guam medical association journal"] = "Guam Med Assoc J",
  ["the guildhall miscellany"] = "Guildhall Misc",
  ["the gulf journal of oncology"] = "Gulf J Oncolog",
  ["the gunma journal of medical sciences"] = "Gunma J Med Sci",
  ["the guthrie bulletin of the donald guthrie medical center"] = "Guthrie Bull",
  ["the guthrie journal of the donald guthrie foundation for medical research"] = "Guthrie J Donald Guthrie Found Med Res",
  ["the guttmacher report on public policy"] = "Guttmacher Rep Public Policy",
  ["the hahnemannian"] = "Hahnemannian",
  ["the hamlyn symposium on medical robotics : proceedings. hamlyn symposium on medical robotics (9th : 2016 : london, england)"] = "Hamyln Symp Med Robot (2016)",
  ["the hand"] = "Hand",
  ["the harlem hospital bulletin"] = "Harlem Hosp Bull (N Y)",
  ["the harvard blackletter journal"] = "Harv Blacklett J",
  ["the harvard environmental law review : helr"] = "HELR Harvard Environ Law Rev",
  ["the harvard journal of minority public health"] = "Harv J Minor Public Health",
  ["the harvard mental health letter"] = "Harv Ment Health Lett",
  ["the harvard mental health letter / from harvard medical school"] = "Harv Ment Health Lett",
  ["the harvard theological review"] = "Harv Theol Rev",
  ["the hastings center report"] = "Hastings Cent Rep",
  ["the hastings law journal"] = "Hastings Law J",
  ["the hawaii nurse : the official monthly newsletter of the hawaii nurses' association"] = "Hawaii Nurse",
  ["the hawaii nurse : the official monthly newsletter of the hawaii nurses’ association"] = "Hawaii Nurse",
  ["the hawaiian journal of history"] = "Hawaii J Hist",
  ["the health bulletin"] = "Health Bull (Raleigh)",
  ["the health care manager"] = "Health Care Manag (Frederick)",
  ["the health care supervisor"] = "Health Care Superv",
  ["the health center journal"] = "Health Cent J",
  ["the health of the people"] = "Health People",
  ["the health press"] = "Health Press",
  ["the health service journal"] = "Health Serv J",
  ["the healthcare forum journal"] = "Healthc Forum J",
  ["the healthcare strategist"] = "Healthc Strateg",
  ["the hearing journal"] = "Hear J",
  ["the heart bulletin"] = "Heart Bull",
  ["the heart surgery forum"] = "Heart Surg Forum",
  ["the hedgehog review"] = "Hedgehog Rev",
  ["the hematology journal : the official journal of the european haematology association"] = "Hematol J",
  ["the hematology journal : the official journal of the european haematology association / eha"] = "Hematol J",
  ["the hemingway review"] = "Hemingway Rev",
  ["the henry e. sigerist supplements to the bulletin of the history of medicine"] = "Henry E Sigerist Suppl Bull Hist Med",
  ["the herbal review"] = "Herbal Rev",
  ["the herbarist"] = "Herbarist",
  ["the herpetological bulletin"] = "Herpetol Bull",
  ["the herpetological journal"] = "Herpetol J",
  ["the hillside journal of clinical psychiatry"] = "Hillside J Clin Psychiatry",
  ["the hip"] = "Hip",
  ["the hispanic american historical review"] = "Hisp Am Hist Rev",
  ["the histochemical journal"] = "Histochem J",
  ["the historian; a journal of history"] = "Historian",
  ["the history of anaesthesia society proceedings"] = "Hist Anaesth Soc Proc",
  ["the history of the family : an international quarterly"] = "Hist Fam",
  ["the history teacher"] = "Hist Teacher",
  ["the holocene"] = "Holocene",
  ["the homiletic and pastoral review"] = "Homilet Pastor Rev",
  ["the homoeopathic physician"] = "Homoeopath Physician",
  ["the homoeopathic recorder"] = "Homoeopath Rec",
  ["the hong kong journal of paediatrics : the journal of hong kong paediatric society = hsiang-kang erh k'o i hsueh hui hui k'an"] = "Hong Kong J Paediatr",
  ["the hong kong journal of social work = xiang gang she hui gong zuo xue bao"] = "Hong Kong J Soc Work",
  ["the hopkins hiv report : a bimonthly newsletter for healthcare providers"] = "Hopkins HIV Rep",
  ["the hopkins hiv report : a bimonthly newsletter for healthcare providers / johns hopkins university aids service"] = "Hopkins HIV Rep",
  ["the horticulture journal"] = "Hort J",
  ["the hospice journal"] = "Hosp J",
  ["the hospital"] = "Hospital (Lond 1886)",
  ["the hospital and health review"] = "Hosp Health Rev",
  ["the hospital and health services review"] = "Hosp Health Serv Rev",
  ["the hospital magazine"] = "Hosp Mag",
  ["the hospital manager"] = "Hosp Manager",
  ["the hospital manager / american hospital association"] = "Hosp Manager",
  ["the hospital medical staff"] = "Hosp Med Staff",
  ["the hospitalist"] = "Hospitalist",
  ["the houston review: history and culture of the gulf coast"] = "Houston Rev",
  ["the howard journal of communications"] = "Howard J Commun",
  ["the hugo journal"] = "Hugo J",
  ["the human life review"] = "Hum Life Rev",
  ["the humanist"] = "Humanist",
  ["the huntington library quarterly"] = "Huntingt Libr Q",
  ["the ibis"] = "Ibis (Lond 1859)",
  ["the ictp series in theoretical physics"] = "ICTP Ser. Theoret. Phys.",
  ["the idrc reports"] = "IDRC Rep",
  ["the ieee intelligent informatics bulletin"] = "IEEE Intell Inform Bull",
  ["the ieee mtt-s 2019 international microwave biomedical conference (imbioc 2019) : proceedings : may 06-08, 2019, international conference hotel of nanjing, nanjing, china. ieee mtt-s international microwave bio conference (2019 : nanjing shi, china)"] = "IEEE MTT-S 2019 Int Microw Biomed Conf IMBioC 2019 (2019)",
  ["the ifcolog journal of logics and their applications"] = "IfCoLog J. Log. Appl.",
  ["the ihs primary care provider"] = "IHS Prim Care Provid",
  ["the illinois cancer bulletin"] = "Ill Cancer Bull",
  ["the illinois medical journal"] = "Ill Med J",
  ["the illustrated london news"] = "ILN",
  ["the ima volumes in mathematics and its applications"] = "IMA Vol. Math. Appl.",
  ["the implant society : [periodical]"] = "Implant Soc",
  ["the indexer"] = "Indexer",
  ["the india journal of medical science"] = "India J Med Sci",
  ["the indian dental journal"] = "Indian Dent J",
  ["the indian economic and social history review"] = "Indian Econ Soc Hist Rev",
  ["the indian economic journal : the quarterly journal of the indian economic association"] = "Indian Econ J",
  ["the indian historian"] = "Indian Hist",
  ["the indian historical review"] = "Indian Hist Rev",
  ["the indian journal of animal sciences"] = "Indian J Anim Sci",
  ["the indian journal of chest diseases"] = "Indian J Chest Dis",
  ["the indian journal of chest diseases & allied sciences"] = "Indian J Chest Dis Allied Sci",
  ["the indian journal of child health"] = "Indian J Child Health",
  ["the indian journal of entomology"] = "Indian J Entomol",
  ["the indian journal of genetics & plant breeding : official publication of the indian society of genetics & plant breeding"] = "Indian J Genet Plant Breed",
  ["the indian journal of horticulture"] = "Indian J Hortic",
  ["the indian journal of hospital pharmacy"] = "Indian J Hosp Pharm",
  ["the indian journal of labour economics : the quarterly journal of the indian society of labour economics"] = "Indian J Labour Econ",
  ["the indian journal of medical research"] = "Indian J Med Res",
  ["the indian journal of medicine & surgery"] = "Indian J Med Surg",
  ["the indian journal of nutrition and dietetics"] = "Indian J Nutr Diet",
  ["the indian journal of pharmacy"] = "Indian J Pharm",
  ["the indian journal of public administration : quarterly journal of the indian institute of public administration"] = "Indian J Public Adm",
  ["the indian journal of radiology & imaging"] = "Indian J Radiol Imaging",
  ["the indian journal of social work"] = "Indian J Soc Work",
  ["the indian journal of surgery"] = "Indian J Surg",
  ["the indian journal of tuberculosis"] = "Indian J Tuberc",
  ["the indian medical gazette"] = "Ind Med Gaz",
  ["the indian physician"] = "Indian Physician",
  ["the indian political science review"] = "Indian Pol Sci Rev",
  ["the indian practitioner"] = "Indian Pract",
  ["the indian veterinary journal"] = "Indian Vet J",
  ["the indiana nurse"] = "Indiana Nurse",
  ["the indiana pharmacist"] = "Indiana Pharm",
  ["the indonesian journal of geography"] = "Indones J Geogr",
  ["the indonesian quarterly"] = "Indones Q",
  ["the industrial robot"] = "Ind Rob",
  ["the information retrieval series"] = "Inf. Retr. Ser.",
  ["the innovation journal : the public sector innovation journal"] = "Innov J",
  ["the insight journal"] = "Insight J",
  ["the institute of mathematics & its applications monograph series"] = "IMA Monogr. Ser.",
  ["the institute of mathematics and its applications conference series"] = "Inst. Math. Appl. Conf. Ser. New Ser.",
  ["the institute of nautical archaeology quarterly"] = "InstNautAQ",
  ["the interdisciplinary journal of problem-based learning"] = "Interdiscip J Probl Based Learn",
  ["the international and comparative law quarterly"] = "Int Comp Law Q",
  ["the international arabic journal of antimicrobial agents"] = "Int Arab J Antimicrob Agents",
  ["the international cryogenics monograph series"] = "Internat. Cryogenics Monogr. Ser.",
  ["the international electronic journal of health education"] = "Int Electron J Health Educ",
  ["the international history review"] = "Int Hist Rev",
  ["the international journal for the psychology of religion"] = "Int J Psychol Relig",
  ["the international journal of adult orthodontics and orthognathic surgery"] = "Int J Adult Orthodon Orthognath Surg",
  ["the international journal of african historical studies"] = "Int J Afr Hist Stud",
  ["the international journal of angiology : official publication of the international college of angiology, inc"] = "Int J Angiol",
  ["the international journal of applied philosophy"] = "Int J Appl Philos",
  ["the international journal of applied radiation and isotopes"] = "Int J Appl Radiat Isot",
  ["the international journal of artificial organs"] = "Int J Artif Organs",
  ["the international journal of aviation psychology"] = "Int J Aviat Psychol",
  ["the international journal of behavioral nutrition and physical activity"] = "Int J Behav Nutr Phys Act",
  ["the international journal of bilingualism : cross-disciplinary, cross-linguistic studies of language behavior"] = "Int J Billing",
  ["the international journal of biochemistry"] = "Int J Biochem",
  ["the international journal of biochemistry & cell biology"] = "Int J Biochem Cell Biol",
  ["the international journal of biological markers"] = "Int J Biol Markers",
  ["the international journal of biostatistics"] = "Int J Biostat",
  ["the international journal of cardiovascular imaging"] = "Int J Cardiovasc Imaging",
  ["the international journal of childbirth education : the official publication of the international childbirth education association"] = "Int J Childbirth Educ",
  ["the international journal of clinical and experimental hypnosis"] = "Int J Clin Exp Hypn",
  ["the international journal of cognitive technology : the official journal of the practical memory institute"] = "Int J Cogn Technol",
  ["the international journal of design education"] = "Int J Des Educ",
  ["the international journal of developmental biology"] = "Int J Dev Biol",
  ["the international journal of diversity in organisations, communities and nations"] = "Int J Divers Organ Communities Nations",
  ["the international journal of eating disorders"] = "Int J Eat Disord",
  ["the international journal of educational and psychological assessment"] = "Int J Educ Psychol Assess",
  ["the international journal of emotional education"] = "Int J Emot Educ",
  ["the international journal of engineering and science"] = "Int J Eng Sci (Ghaziabad)",
  ["the international journal of engineering education"] = "Int J Eng Educ",
  ["the international journal of environmental studies"] = "Int J Environ Stud",
  ["the international journal of esthetic dentistry"] = "Int J Esthet Dent",
  ["the international journal of forensic dentistry"] = "Int J Forensic Dent",
  ["the international journal of health planning and management"] = "Int J Health Plann Manage",
  ["the international journal of health, wellness & society"] = "Int J Health Wellness Soc",
  ["the international journal of high performance computing applications"] = "Int J High Perform Comput Appl",
  ["the international journal of intelligent control and systems"] = "Int. J. Intell. Control Syst.",
  ["the international journal of life cycle assessment"] = "Int J Life Cycle Assess",
  ["the international journal of lower extremity wounds"] = "Int J Low Extrem Wounds",
  ["the international journal of medical robotics + computer assisted surgery : mrcas"] = "Int J Med Robot",
  ["the international journal of medicine and law : ijml"] = "Int J Med Law",
  ["the international journal of mental health promotion"] = "Int J Ment Health Promot",
  ["the international journal of multiphysics"] = "Int J Multiphys",
  ["the international journal of neuropsychopharmacology"] = "Int J Neuropsychopharmacol",
  ["the international journal of neuropsychopharmacology / official scientific journal of the collegium internationale neuropsychopharmacologicum (cinp)"] = "Int J Neuropsychopharmacol",
  ["the international journal of neuroscience"] = "Int J Neurosci",
  ["the international journal of occupational and environmental medicine"] = "Int J Occup Environ Med",
  ["the international journal of occupational health & safety"] = "Int J Occup Health Saf",
  ["the international journal of oral & maxillofacial implants"] = "Int J Oral Maxillofac Implants",
  ["the international journal of oral implantology : implantologist"] = "Int J Oral Implantol",
  ["the international journal of oral myology"] = "Int J Oral Myol",
  ["the international journal of orofacial myology : official publication of the international association of orofacial myology"] = "Int J Orofacial Myology",
  ["the international journal of pediatric nephrology"] = "Int J Pediatr Nephrol",
  ["the international journal of periodontics & restorative dentistry"] = "Int J Periodontics Restorative Dent",
  ["the international journal of pharmacy practice"] = "Int J Pharm Pract",
  ["the international journal of press/politics"] = "Int J Press Polit",
  ["the international journal of prosthodontics"] = "Int J Prosthodont",
  ["the international journal of psychiatric nursing research"] = "Int J Psychiatr Nurs Res",
  ["the international journal of psycho-analysis"] = "Int J Psychoanal",
  ["the international journal of religion and spirituality in society"] = "Int J Religion Spiritual Soc",
  ["the international journal of research on service-learning and community engagement"] = "Int J Res Serv Learn Community Engagem",
  ["the international journal of risk & safety in medicine"] = "Int J Risk Saf Med",
  ["the international journal of robotics research"] = "Int J Rob Res",
  ["the international journal of science in society"] = "Int J Sci Soc",
  ["the international journal of social education : official journal of the indiana council for the social studies"] = "Int J Soc Educ",
  ["the international journal of social psychiatry"] = "Int J Soc Psychiatry",
  ["the international journal of sociology and social policy"] = "Int J Sociol Soc Policy",
  ["the international journal of soil, sediment and water : documenting the cutting edge of environmental stewardship"] = "Int J Soil Sediment Water",
  ["the international journal of structural changes in solids : mechanics and applications"] = "Int J Struct Changes Sol",
  ["the international journal of supercomputer applications"] = "Int J Supercomput Appl",
  ["the international journal of targeted therapies in cancer"] = "Int J Target Ther Cancer",
  ["the international journal of the addictions"] = "Int J Addict",
  ["the international journal of the history of sport"] = "Int J Hist Sport",
  ["the international journal of transgenderism"] = "Int J Transgend",
  ["the international journal of transitional justice"] = "Int J  Transit Justice",
  ["the international journal of tuberculosis and lung disease : the official journal of the international union against tuberculosis and lung disease"] = "Int J Tuberc Lung Dis",
  ["the international journal of virtual reality : a multimedia publication for professionals"] = "Int J Virtual Real",
  ["the international journal on drug policy"] = "Int J Drug Policy",
  ["the international journal on school disaffection"] = "Int J School Disaffect",
  ["the international journal, advanced manufacturing technology"] = "Int J Adv Manuf Technol",
  ["the international law news"] = "Int Law News",
  ["the international library of critical writings in economics"] = "Internat. Lib. Crit. Writ. Econ.",
  ["the international migration review"] = "Int Migr Rev",
  ["the international monitor. rheumatology"] = "Int Monit Rheumatol",
  ["the international review of psycho-analysis"] = "Int Rev Psychoanal",
  ["the international spectator : a quarterly journal of the istituto affari internazionali"] = "Int Spect (Rome)",
  ["the international tinnitus journal"] = "Int Tinnitus J",
  ["the international undergraduate journal for service-learning, leadership, and social change"] = "Int Undergrad J Serv Learn Leadersh Soc Change",
  ["the international zoo yearbook"] = "Int Zoo Yearb",
  ["the interne"] = "Interne",
  ["the internet and higher education"] = "Internet High Educ",
  ["the internet journal of allied health sciences and practice"] = "Internet J Allied Health Sci Pract",
  ["the internet journal of hematology"] = "Internet J Hematol",
  ["the internet journal of mental health"] = "Internet J Ment Health",
  ["the internet journal of microbiology"] = "Internet J Microbiol",
  ["the internet journal of nephrology"] = "Internet J Nephrol",
  ["the internet journal of nutrition and wellness"] = "Internet J Nutr Wellness",
  ["the internet journal of pediatrics and neonatology"] = "Internet J Pediatr Neonatol",
  ["the internist"] = "Internist",
  ["the interpreter and translator trainer"] = "Interpret Transl Train",
  ["the interservice/industry training, simulation, and education conference : i/itsec. interservice/industry training, simulation, and education conference"] = "Interserv Ind Train Simul Educ Conf",
  ["the interstudy quality edge : measurement and management of clinical outcomes"] = "Interstudy Qual Edge",
  ["the investor-owned hospital review"] = "Investor Owned Hosp Rev",
  ["the iowa dental bulletin"] = "Iowa Dent Bull",
  ["the iowa dental journal"] = "Iowa Dent J",
  ["the iowa dental journal / iowa state dental society"] = "Iowa Dent J",
  ["the iowa orthopaedic journal"] = "Iowa Orthop J",
  ["the iranian journal of botany"] = "Iran J Bot",
  ["the irish banking review"] = "Ir Bank Rev",
  ["the irish journal of psychology"] = "Ir J Psychol",
  ["the irish naturalists' journal"] = "Ir Nat J",
  ["the irish nursing news"] = "Ir Nurs News",
  ["the irish sword : the journal of the military history society of ireland"] = "Ir Sword",
  ["the irish theological quarterly"] = "Irish Theol Q",
  ["the irish times"] = "Irish Times",
  ["the isaac newton institute series of lectures"] = "Isaac Newton Inst. Ser. Lectures",
  ["the islamic quarterly"] = "Islam Q",
  ["the island magazine"] = "Isl Mag",
  ["the isme journal"] = "ISME J",
  ["the israel annals of psychiatry and related disciplines"] = "Isr Ann Psychiatr Relat Discip",
  ["the israel journal of psychiatry and related sciences"] = "Isr J Psychiatry Relat Sci",
  ["the israel medical association journal : imaj"] = "Isr Med Assoc J",
  ["the israel museum journal"] = "IsrMusJ",
  ["the israel museum news"] = "IsrMusN",
  ["the israeli journal of aquaculture = bamidgeh"] = "Isr J Aquac",
  ["the italian journal of biochemistry"] = "Ital J Biochem",
  ["the italian journal of gastroenterology"] = "Ital J Gastroenterol",
  ["the italian journal of surgical sciences"] = "Ital J Surg Sci",
  ["the italian journal of surgical sciences / sponsored by societa italiana di chirurgia"] = "Ital J Surg Sci",
  ["the italian journal of zoology"] = "Ital J Zool (Modena)",
  ["the j. paul getty museum journal"] = "GMusJ",
  ["the jackson clinic bulletin"] = "Jackson Clin Bull",
  ["the jamaican nurse"] = "Jamaican Nurse",
  ["the japanese dental science review"] = "Jpn Dent Sci Rev",
  ["the japanese economic review. the journal of the japanese economic association"] = "Jpn. Econ. Rev.",
  ["the japanese journal of antibiotics"] = "Jpn J Antibiot",
  ["the japanese journal of behaviormetrics"] = "Jpn. J. Behaviormetrics",
  ["the japanese journal of dermatology. ser. b"] = "Jpn J Dermatol B",
  ["the japanese journal of experimental medicine"] = "Jpn J Exp Med",
  ["the japanese journal of human genetics"] = "Jpn J Hum Genet",
  ["the japanese journal of obstetrics and gynecology"] = "Jpn J Obstet Gynecol",
  ["the japanese journal of physiology"] = "Jpn J Physiol",
  ["the japanese journal of population"] = "Jpn J Popul",
  ["the japanese journal of psychiatry and neurology"] = "Jpn J Psychiatry Neurol",
  ["the japanese journal of surgery"] = "Jpn J Surg",
  ["the japanese journal of thoracic and cardiovascular surgery : official publication of the japanese association for thoracic surgery = nihon kyōbu geka gakkai zasshi"] = "Jpn J Thorac Cardiovasc Surg",
  ["the japanese journal of tuberculosis"] = "Jpn J Tuberc",
  ["the japanese journal of tuberculosis and chest diseases"] = "Jpn J Tuberc Chest Dis",
  ["the japanese journal of veterinary research"] = "Jpn J Vet Res",
  ["the japanese psychological research"] = "Jpn Psychol Res",
  ["the jefferson-hillman hospital bulletin"] = "Jefferson-Hillman Hosp Bull",
  ["the jerusalem quarterly"] = "Jerusalem Q",
  ["the jewish digest"] = "Jew Dig",
  ["the jewish memorial hospital bulletin"] = "Jew Mem Hosp Bull",
  ["the john marshall journal of computer & information law"] = "John Marshall J Comput Inf Law",
  ["the john marshall journal of practice and procedure"] = "John Marshall J Pract Proced",
  ["the john marshall law review"] = "John Marshall Law Rev",
  ["the johns hopkins alumni magazine"] = "Johns Hopkins Alumni Mag",
  ["the johns hopkins medical journal"] = "Johns Hopkins Med J",
  ["the johns hopkins medical letter health after 50"] = "Johns Hopkins Med Lett Health After 50",
  ["the johns hopkins nurses alumnae magazine"] = "Johns Hopkins Nurses Alumnae Mag",
  ["the johns hopkins nurses' alumni magazine"] = "Johns Hopkins Nurses Alumni Mag",
  ["the johns hopkins nurses’ alumni magazine"] = "Johns Hopkins Nurses Alumni Mag",
  ["the joint commission journal on quality improvement"] = "Jt Comm J Qual Improv",
  ["the jordan dental journal"] = "Jordan Dent J",
  ["the journal"] = "Journal",
  ["the journal for hospital admitting management : official publication of the national association of hospital admitting managers"] = "J Hosp Admit Manage",
  ["the journal for hospital admitting management : official publication of the national association of hospital admitting managers / naham"] = "J Hosp Admit Manage",
  ["the journal for nurse practitioners : jnp"] = "J Nurse Pract",
  ["the journal of abdominal surgery"] = "J Abdom Surg",
  ["the journal of adhesion"] = "J Adhes",
  ["the journal of adhesive dentistry"] = "J Adhes Dent",
  ["the journal of adolescent health : official publication of the society for adolescent medicine"] = "J Adolesc Health",
  ["the journal of advanced prosthodontics"] = "J Adv Prosthodont",
  ["the journal of aging research & clinical practice"] = "J Aging Res Clin Pract",
  ["the journal of agricultural science"] = "J Agric Sci",
  ["the journal of air medical transport"] = "J Air Med Transp",
  ["the journal of allergy"] = "J Allergy",
  ["the journal of allergy and clinical immunology"] = "J Allergy Clin Immunol",
  ["the journal of allergy and clinical immunology. in practice"] = "J Allergy Clin Immunol Pract",
  ["the journal of ambulatory care management"] = "J Ambul Care Manage",
  ["the journal of american folk-lore"] = "J Am Folk",
  ["the journal of american health care"] = "J Am Health Care",
  ["the journal of american health policy"] = "J Am Health Policy",
  ["the journal of american science"] = "J Am Sci",
  ["the journal of analysis"] = "J. Anal.",
  ["the journal of analytical psychology"] = "J Anal Psychol",
  ["the journal of animal ecology"] = "J Anim Ecol",
  ["the journal of antibiotics"] = "J Antibiot (Tokyo)",
  ["the journal of antibiotics. ser. b"] = "J Antibiot B",
  ["the journal of antimicrobial chemotherapy"] = "J Antimicrob Chemother",
  ["the journal of applied analysis and computation"] = "J. Appl. Anal. Comput.",
  ["the journal of applied bacteriology"] = "J Appl Bacteriol",
  ["the journal of applied behavioral science"] = "J Appl Behav Sci",
  ["the journal of applied ecology"] = "J Appl Ecol",
  ["the journal of applied laboratory medicine"] = "J Appl Lab Med",
  ["the journal of applied management and entrepreneurship"] = "J Appl Manag Entrep",
  ["the journal of applied nutrition"] = "J Appl Nutr",
  ["the journal of applied poultry research"] = "J Appl Poult Res",
  ["the journal of applied psychology"] = "J Appl Psychol",
  ["the journal of applied research"] = "J Appl Res",
  ["the journal of applied research on children : informing policy for children at risk"] = "J Appl Res Child",
  ["the journal of applied science"] = "J Appl Sci (Thailand)",
  ["the journal of applied social sciences"] = "J Appl Soc Sci",
  ["the journal of arachnology"] = "J Arachnol",
  ["the journal of arizona history"] = "J Ariz Hist",
  ["the journal of arthroplasty"] = "J Arthroplasty",
  ["the journal of artificial intelligence research"] = "J Artif Intell Res",
  ["the journal of asian studies"] = "J Asian Stud",
  ["the journal of asthma : official journal of the association for the care of asthma"] = "J Asthma",
  ["the journal of asthma research"] = "J Asthma Res",
  ["the journal of at-risk issues : a joint publication of the national dropout prevention center and network"] = "J Risk Issues",
  ["the journal of audiovisual media in medicine"] = "J Audiov Media Med",
  ["the journal of auditory research"] = "J Aud Res",
  ["the journal of automatic chemistry"] = "J Automat Chem",
  ["the journal of aviation medicine"] = "J Aviat Med",
  ["the journal of behavioral economics"] = "J Behav Econ",
  ["the journal of behavioral finance"] = "J Behav Financ",
  ["the journal of behavioral health services & research"] = "J Behav Health Serv Res",
  ["the journal of biocommunication"] = "J Biocommun",
  ["the journal of biolaw & business"] = "J Biolaw Bus",
  ["the journal of biological chemistry"] = "J Biol Chem",
  ["the journal of biological sciences"] = "J Biol Sci",
  ["the journal of biophysical and biochemical cytology"] = "J Biophys Biochem Cytol",
  ["the journal of bioscience and medicine"] = "J Biosci Med",
  ["the journal of black psychology"] = "J Black Psychol",
  ["the journal of bone and joint surgery"] = "J Bone Joint Surg",
  ["the journal of bone and joint surgery. american volume"] = "J Bone Joint Surg Am",
  ["the journal of bone and joint surgery. british volume"] = "J Bone Joint Surg Br",
  ["the journal of breast health"] = "J Breast Health",
  ["the journal of british studies"] = "J Br Stud",
  ["the journal of burma studies"] = "J Burma Stud",
  ["the journal of burn care & rehabilitation"] = "J Burn Care Rehabil",
  ["the journal of business and social studies"] = "J Bus Soc Stud",
  ["the journal of business of the university of chicago"] = "J Bus Univ Chic",
  ["the journal of business strategy"] = "J Bus Strategy",
  ["the journal of cancer research"] = "J Cancer Res",
  ["the journal of cardiothoracic trauma"] = "J Cardiothorac Trauma",
  ["the journal of cardiovascular management : the official journal of the american college of cardiovascular administrators"] = "J Cardiovasc Manag",
  ["the journal of cardiovascular medicine"] = "J Cardiovasc Med",
  ["the journal of cardiovascular nursing"] = "J Cardiovasc Nurs",
  ["the journal of cardiovascular surgery"] = "J Cardiovasc Surg (Torino)",
  ["the journal of caribbean history"] = "J Caribb Hist",
  ["the journal of caribbean ornithology"] = "J Caribb Ornithol",
  ["the journal of cell biology"] = "J Cell Biol",
  ["the journal of cetacean research and management"] = "J Cetacean Res Manag",
  ["the journal of chemical physics"] = "J Chem Phys",
  ["the journal of chemical thermodynamics"] = "J Chem Thermodyn",
  ["the journal of child psychiatry"] = "J Child Psychiatry",
  ["the journal of chinese medicine"] = "J Chin Med",
  ["the journal of chiropractic education"] = "J Chiropr Educ",
  ["the journal of civil war medicine : official quarterly publication of the society of civil war surgeons, inc"] = "J Civ War Med",
  ["the journal of clinical and aesthetic dermatology"] = "J Clin Aesthet Dermatol",
  ["the journal of clinical dentistry"] = "J Clin Dent",
  ["the journal of clinical dysmorphology"] = "J Clin Dysmorphol",
  ["the journal of clinical endocrinology"] = "J Clin Endocrinol",
  ["the journal of clinical endocrinology and metabolism"] = "J Clin Endocrinol Metab",
  ["the journal of clinical ethics"] = "J Clin Ethics",
  ["the journal of clinical investigation"] = "J Clin Invest",
  ["the journal of clinical nutrition"] = "J Clin Nutr",
  ["the journal of clinical pediatric dentistry"] = "J Clin Pediatr Dent",
  ["the journal of clinical pharmacology and new drugs"] = "J Clin Pharmacol New Drugs",
  ["the journal of clinical pharmacology and the journal of new drugs"] = "J Clin Pharmacol J New Drugs",
  ["the journal of clinical psychiatry"] = "J Clin Psychiatry",
  ["the journal of college and university law"] = "J Coll Univ Law",
  ["the journal of commonwealth & comparative politics"] = "J Commonw Comp Polit",
  ["the journal of commonwealth literature"] = "J Commonw Lit",
  ["the journal of communicable diseases"] = "J Commun Dis",
  ["the journal of communication"] = "J Commun",
  ["the journal of communication inquiry"] = "J Commun Inq",
  ["the journal of community and supportive oncology"] = "J Community Support Oncol",
  ["the journal of comparative neurology"] = "J Comp Neurol",
  ["the journal of comparative pathology and therapeutics"] = "J Comp Pathol Ther",
  ["the journal of compliance in health care : jchc"] = "J Compliance Health Care",
  ["the journal of computed tomography"] = "J Comput Tomogr",
  ["the journal of conflict resolution"] = "J Conflict Resolut",
  ["the journal of consumer affairs"] = "J Consum Aff",
  ["the journal of consumer marketing"] = "J Consum Mark",
  ["the journal of consumer research"] = "J Consum Res",
  ["the journal of contemporary china = dang dai zhongguo"] = "J Contemp China",
  ["the journal of contemporary dental practice"] = "J Contemp Dent Pract",
  ["the journal of contemporary health law and policy"] = "J Contemp Health Law Policy",
  ["the journal of continuing education in the health professions"] = "J Contin Educ Health Prof",
  ["the journal of contraception"] = "J Contracept",
  ["the journal of corporation law"] = "J Corp Law",
  ["the journal of cranio-mandibular practice"] = "J Craniomandibular Pract",
  ["the journal of cranio-maxillofacial trauma"] = "J Craniomaxillofac Trauma",
  ["the journal of craniofacial surgery"] = "J Craniofac Surg",
  ["the journal of creative behavior"] = "J Creat Behav",
  ["the journal of criminal law & criminology"] = "J Crim Law Criminol",
  ["the journal of criminal law and criminology, including the american journal of police science"] = "J Crim Law Criminol (1931)",
  ["the journal of criminal law, criminology, and police science"] = "J Crim Law Criminol Police Sci",
  ["the journal of critical illness"] = "J Crit Illn",
  ["the journal of current adolescent medicine"] = "J Curr Adolesc Med",
  ["the journal of dairy research"] = "J Dairy Res",
  ["the journal of dental medicine"] = "J Dent Med",
  ["the journal of dental symposia"] = "J Dent Symp",
  ["the journal of dermatologic surgery"] = "J Dermatol Surg",
  ["the journal of dermatologic surgery and oncology"] = "J Dermatol Surg Oncol",
  ["the journal of dermatological treatment"] = "J Dermatolog Treat",
  ["the journal of dermatology"] = "J Dermatol",
  ["the journal of development and administrative studies"] = "J Dev Admin Stud",
  ["the journal of development communication"] = "J Dev Comm",
  ["the journal of development studies"] = "J Dev Stud",
  ["the journal of developmental and learning disorders"] = "J Dev Learn Disord",
  ["the journal of developmental processes"] = "J Dev Process",
  ["the journal of diabetic complications"] = "J Diabet Complications",
  ["the journal of documentation; devoted to the recording, organization and dissemination of specialized knowledge"] = "J Doc",
  ["the journal of early adolescence"] = "J Early Adolesc",
  ["the journal of east asian affairs"] = "J East Asian Aff",
  ["the journal of ecclesiastical history"] = "J Ecclesiast Hist",
  ["the journal of ecology"] = "J Ecol",
  ["the journal of economic education"] = "J Econ Educ",
  ["the journal of economic history"] = "J Econ Hist",
  ["the journal of economic perspectives : a journal of the american economic association"] = "J Econ Perspect",
  ["the journal of ect"] = "J ECT",
  ["the journal of education in perioperative medicine : jepm"] = "J Educ Perioper Med",
  ["the journal of educational research"] = "J Educ Res",
  ["the journal of egyptian archaeology"] = "J Egypt Archaeol",
  ["the journal of electronic publishing : jep"] = "J Electron Publ",
  ["the journal of emergency medicine"] = "J Emerg Med",
  ["the journal of endocrine genetics : jeg"] = "J Endocr Genet",
  ["the journal of endocrinology"] = "J Endocrinol",
  ["the journal of environment & development"] = "J Environ Dev",
  ["the journal of environmental education"] = "J Environ Educ",
  ["the journal of equity in health"] = "J Equity Health",
  ["the journal of erie studies"] = "J Erie Stud",
  ["the journal of ethics"] = "J Ethics",
  ["the journal of ethnic studies"] = "J Ethn Stud",
  ["the journal of eukaryotic microbiology"] = "J Eukaryot Microbiol",
  ["the journal of european economic history"] = "J Eur Econ Hist",
  ["the journal of evidence-based dental practice"] = "J Evid Based Dent Pract",
  ["the journal of experimental biology"] = "J Exp Biol",
  ["the journal of experimental education"] = "J. Exp. Educ.",
  ["the journal of experimental medicine"] = "J Exp Med",
  ["the journal of experimental secondary science"] = "J Exp Second Sci",
  ["the journal of experimental zoology"] = "J Exp Zool",
  ["the journal of experimental zoology. supplement : published under auspices of the american society of zoologists and the division of comparative physiology and biochemistry"] = "J Exp Zool Suppl",
  ["the journal of experimental zoology. supplement : published under auspices of the american society of zoologists and the division of comparative physiology and biochemistry / the wistar institute of anatomy and biology"] = "J Exp Zool Suppl",
  ["the journal of extra-corporeal technology"] = "J Extra Corpor Technol",
  ["the journal of faculty development"] = "J Fac Dev",
  ["the journal of family and culture"] = "J Fam Cult",
  ["the journal of family health care"] = "J Fam Health Care",
  ["the journal of family planning and reproductive health care"] = "J Fam Plann Reprod Health Care",
  ["the journal of family planning and reproductive health care / faculty of family planning & reproductive health care, royal college of obstetricians & gynaecologists"] = "J Fam Plann Reprod Health Care",
  ["the journal of family practice"] = "J Fam Pract",
  ["the journal of family welfare"] = "J Fam Welf",
  ["the journal of finance"] = "J Finance",
  ["the journal of foot and ankle surgery : official publication of the american college of foot and ankle surgeons"] = "J Foot Ankle Surg",
  ["the journal of foot surgery"] = "J Foot Surg",
  ["the journal of forensic odonto-stomatology"] = "J Forensic Odontostomatol",
  ["the journal of forensic psychiatry & psychology"] = "J Forens Psychiatry Psychol",
  ["the journal of fourier analysis and applications"] = "J. Fourier Anal. Appl.",
  ["the journal of frailty & aging"] = "J Frailty Aging",
  ["the journal of functional and logic programming"] = "J. Funct. Logic Programming",
  ["the journal of gender-specific medicine : jgsm : the official journal of the partnership for women's health at columbia"] = "J Gend Specif Med",
  ["the journal of gender-specific medicine : jgsm : the official journal of the partnership for women’s health at columbia"] = "J Gend Specif Med",
  ["the journal of gene medicine"] = "J Gene Med",
  ["the journal of general and applied microbiology"] = "J Gen Appl Microbiol",
  ["the journal of general education"] = "J Gen Educ",
  ["the journal of general physiology"] = "J Gen Physiol",
  ["the journal of general psychology"] = "J Gen Psychol",
  ["the journal of general virology"] = "J Gen Virol",
  ["the journal of genetic genealogy"] = "J Genet Geneal",
  ["the journal of genetic psychology"] = "J Genet Psychol",
  ["the journal of geography"] = "J Geog",
  ["the journal of geology"] = "J Geol",
  ["the journal of geometric analysis"] = "J. Geom. Anal.",
  ["the journal of glaciology"] = "J Glaciol",
  ["the journal of global drug policy and practice"] = "J Glob Drug Policy Pract",
  ["the journal of global radiology"] = "J Glob Radiol",
  ["the journal of gnathology"] = "J Gnathol",
  ["the journal of hand surgery"] = "J Hand Surg Am",
  ["the journal of hand surgery asian-pacific volume"] = "J Hand Surg Asian Pac Vol",
  ["the journal of hand surgery, european volume"] = "J Hand Surg Eur Vol",
  ["the journal of head trauma rehabilitation"] = "J Head Trauma Rehabil",
  ["the journal of headache and pain"] = "J Headache Pain",
  ["the journal of headache and pain : official journal of the italian society for the study of headaches"] = "J Headache Pain",
  ["the journal of health administration education"] = "J Health Adm Educ",
  ["the journal of health and physical education"] = "J Health Phys Educ",
  ["the journal of health, environment & education"] = "J Health Environ Educ",
  ["the journal of heart and lung transplantation : the official publication of the international society for heart transplantation"] = "J Heart Lung Transplant",
  ["the journal of heart failure : the official journal of the international society of heart failure"] = "J Heart Fail",
  ["the journal of heart transplantation"] = "J Heart Transplant",
  ["the journal of heart valve disease"] = "J Heart Valve Dis",
  ["the journal of hellenic studies"] = "J Hell Stud",
  ["the journal of hellenic studies / the society for the promotion of hellenic studies"] = "J Hell Stud",
  ["the journal of heredity"] = "J Hered",
  ["the journal of high energy physics"] = "J. High Energy Phys.",
  ["the journal of higher education"] = "J Higher Educ",
  ["the journal of histochemistry and cytochemistry : official journal of the histochemistry society"] = "J Histochem Cytochem",
  ["the journal of horticultural science"] = "J Hortic Sci",
  ["the journal of horticultural science & biotechnology"] = "J Hortic Sci Biotechnol",
  ["the journal of hospital infection"] = "J Hosp Infect",
  ["the journal of human resources"] = "J Hum Resour",
  ["the journal of humanistic education and development"] = "J Humanist Educ Dev",
  ["the journal of hygiene"] = "J Hyg (Lond)",
  ["the journal of ima"] = "J IMA",
  ["the journal of imaging science and technology"] = "J Imaging Sci Technol",
  ["the journal of imperial and commonwealth history"] = "J Imp Commonw Hist",
  ["the journal of indian orthodontic society"] = "J Indian Orthod Soc",
  ["the journal of industrial economics"] = "J Ind Econ",
  ["the journal of industrial hygiene and toxicology"] = "J Ind Hyg Toxicol",
  ["the journal of industrial relations"] = "J Ind Relat",
  ["the journal of infection"] = "J Infect",
  ["the journal of infectious diseases"] = "J Infect Dis",
  ["the journal of information and image management : jiim"] = "J Inf Image Manage",
  ["the journal of infusional chemotherapy"] = "J Infus Chemother",
  ["the journal of innovations in cardiac rhythm management"] = "J Innov Card Rhythm Manag",
  ["the journal of interactive technology and pedagogy"] = "J Interact Technol Pedagog",
  ["the journal of interdisciplinary history"] = "J Interdiscip Hist",
  ["the journal of international advanced otology"] = "J Int Adv Otol",
  ["the journal of international law and economics"] = "J Int Law Econ",
  ["the journal of international medical research"] = "J Int Med Res",
  ["the journal of international trade & economic development"] = "J Int Trade Econ Dev",
  ["the journal of invasive cardiology"] = "J Invasive Cardiol",
  ["the journal of invasive fungal infections"] = "J Invasive Fungal Infect",
  ["the journal of investigative dermatology"] = "J Invest Dermatol",
  ["the journal of investigative dermatology. symposium proceedings"] = "J Investig Dermatol Symp Proc",
  ["the journal of investigative dermatology. symposium proceedings / the society for investigative dermatology, inc. [and] european society for dermatological research"] = "J Investig Dermatol Symp Proc",
  ["the journal of israeli history"] = "J Isr Hist",
  ["the journal of japanese studies"] = "J Jpn Stud",
  ["the journal of juristic papyrology"] = "JJurP",
  ["the journal of knee surgery"] = "J Knee Surg",
  ["the journal of kurdish studies"] = "J Kurd Stud",
  ["the journal of laboratory and clinical medicine"] = "J Lab Clin Med",
  ["the journal of laryngology and otology"] = "J Laryngol Otol",
  ["the journal of laryngology and otology. supplement"] = "J Laryngol Otol Suppl",
  ["the journal of latin american and caribbean anthropology"] = "J Lat Am Caribb Anthropol",
  ["the journal of law & economics"] = "J Law Econ",
  ["the journal of law & politics"] = "J Law Polit",
  ["the journal of law and religion"] = "J Law Relig",
  ["the journal of law and technology"] = "J Law Technol",
  ["the journal of law, medicine & ethics : a journal of the american society of law, medicine & ethics"] = "J Law Med Ethics",
  ["the journal of legal history"] = "J Legal Hist",
  ["the journal of legal medicine"] = "J Leg Med",
  ["the journal of legal studies"] = "J Legal Stud",
  ["the journal of libertarian studies"] = "J Libert Stud",
  ["the journal of lithotripsy & stone disease"] = "J Lithotr Stone Dis",
  ["the journal of logic and algebraic programming"] = "J Log Algebr Program",
  ["the journal of long island history"] = "J Long Isl Hist",
  ["the journal of long term care administration"] = "J Long Term Care Adm",
  ["the journal of long term home health care : the pride institute journal"] = "J Long Term Home Health Care",
  ["the journal of macrotrends in health and medicine"] = "J Macrotrends Health Med",
  ["the journal of manual & manipulative therapy"] = "J Man Manip Ther",
  ["the journal of materials education"] = "J Mater Educ",
  ["the journal of maternal-fetal & neonatal medicine : the official journal of the european association of perinatal medicine, the federation of asia and oceania perinatal societies, the international society of perinatal obstetricians"] = "J Matern Fetal Neonatal Med",
  ["the journal of maternal-fetal medicine"] = "J Matern Fetal Med",
  ["the journal of mathematical sociology"] = "J Math Sociol",
  ["the journal of media literacy education"] = "J Media Lit Educ",
  ["the journal of medical chemical defense"] = "J Med Chem Biol Radiol Def",
  ["the journal of medical humanities"] = "J Med Humanit",
  ["the journal of medical humanities and bioethics"] = "J Med Humanit Bioeth",
  ["the journal of medical investigation : jmi"] = "J Med Invest",
  ["the journal of medical laboratory technology"] = "J Med Lab Technol",
  ["the journal of medical practice management : mpm"] = "J Med Pract Manage",
  ["the journal of medical research"] = "J Med Res",
  ["the journal of medical, chemical, biological, and radiological defense"] = "J Med Chem Biol Radiol Def",
  ["the journal of medicine and philosophy"] = "J Med Philos",
  ["the journal of medieval and renaissance studies"] = "J Mediev Renaiss Stud",
  ["the journal of membrane biology"] = "J Membr Biol",
  ["the journal of men's health & gender : the official journal of the international society for men's health & gender"] = "J Mens Health Gend",
  ["the journal of men's studies"] = "J Mens Stud",
  ["the journal of mental health policy and economics"] = "J Ment Health Policy Econ",
  ["the journal of mental science"] = "J Ment Sci",
  ["the journal of mental subnormality"] = "J Ment Subnorm",
  ["the journal of men’s health & gender : the official journal of the international society for men’s health & gender"] = "J Mens Health Gend",
  ["the journal of mexican american history"] = "J Mex Am Hist",
  ["the journal of microbiology, biotechnology and food sciences"] = "J Microbiol Biotechnol Food Sci",
  ["the journal of microcolumn separations : jms"] = "J Microcolumn Sep",
  ["the journal of micrographics"] = "J Microgr",
  ["the journal of microwave power"] = "J Microw Power",
  ["the journal of microwave power and electromagnetic energy : a publication of the international microwave power institute"] = "J Microw Power Electromagn Energy",
  ["the journal of military history"] = "J Mil Hist",
  ["the journal of military medicine in the pacific"] = "J Mil Med Pac",
  ["the journal of mississippi history"] = "J Miss Hist",
  ["the journal of modern african studies"] = "J Mod Afr Stud",
  ["the journal of modern history"] = "J Mod Hist",
  ["the journal of molecular and cellular immunology : jmci"] = "J Mol Cell Immunol",
  ["the journal of molecular diagnostics : jmd"] = "J Mol Diagn",
  ["the journal of molluscan studies"] = "J Molluscan Stud",
  ["the journal of multicultural nursing & health : official journal of the center for the study of multiculturalism and health care, inc"] = "J Multicult Nurs Health",
  ["the journal of musculoskeletal medicine"] = "J Musculoskelet Med",
  ["the journal of natural sciences and mathematics"] = "J. Natur. Sci. Math.",
  ["the journal of negro education"] = "J Negro Educ",
  ["the journal of negro history"] = "J Negro Hist",
  ["the journal of nephrology social work"] = "J Nephrol Soc Work",
  ["the journal of nervous and mental disease"] = "J Nerv Ment Dis",
  ["the journal of neurology and psychopathology"] = "J Neurol Psychopathol",
  ["the journal of neuropsychiatry and clinical neurosciences"] = "J Neuropsychiatry Clin Neurosci",
  ["the journal of neuroscience : the official journal of the society for neuroscience"] = "J Neurosci",
  ["the journal of neuroscience nursing : journal of the american association of neuroscience nurses"] = "J Neurosci Nurs",
  ["the journal of new drugs"] = "J New Drugs",
  ["the journal of nih research : life sciences research and news about the national institutes of health and the alcohol, drug abuse, and mental health administration"] = "J NIH Res",
  ["the journal of nihon university school of dentistry"] = "J Nihon Univ Sch Dent",
  ["the journal of northeast agricultural university"] = "J Northeast Agric Univ",
  ["the journal of nuclear biology and medicine"] = "J Nucl Biol Med",
  ["the journal of nuclear medicine and allied sciences"] = "J Nucl Med Allied Sci",
  ["the journal of nursing administration"] = "J Nurs Adm",
  ["the journal of nursing care"] = "J Nurs Care",
  ["the journal of nursing education"] = "J Nurs Educ",
  ["the journal of nursing home research sciences"] = "J Nurs Home Res Sci",
  ["the journal of nursing research : jnr"] = "J Nurs Res",
  ["the journal of nutrition"] = "J Nutr",
  ["the journal of nutrition, health & aging"] = "J Nutr Health Aging",
  ["the journal of nutritional biochemistry"] = "J Nutr Biochem",
  ["the journal of obstetrics and gynaecology of the british commonwealth"] = "J Obstet Gynaecol Br Commonw",
  ["the journal of obstetrics and gynaecology of the british empire"] = "J Obstet Gynaecol Br Emp",
  ["the journal of obstetrics and gynaecology research"] = "J Obstet Gynaecol Res",
  ["the journal of oncology management : the official journal of the american college of oncology administrators"] = "J Oncol Manag",
  ["the journal of oncopathology"] = "J Oncopathol",
  ["the journal of open source education"] = "J Open Source Educ",
  ["the journal of ophthalmic photography"] = "J Ophthalmic Photogr",
  ["the journal of oral implant and transplant surgery"] = "J Oral Implant Transplant Surg",
  ["the journal of oral implantology"] = "J Oral Implantol",
  ["the journal of organic chemistry"] = "J Org Chem",
  ["the journal of orthopaedic and sports physical therapy"] = "J Orthop Sports Phys Ther",
  ["the journal of osaka university dental school"] = "J Osaka Univ Dent Sch",
  ["the journal of osteopathy"] = "J Osteopath (Kirksvill)",
  ["the journal of otolaryngology"] = "J Otolaryngol",
  ["the journal of otolaryngology. supplement"] = "J Otolaryngol Suppl",
  ["the journal of pacific history"] = "J Pac Hist",
  ["the journal of pain"] = "J Pain",
  ["the journal of pain : official journal of the american pain society"] = "J Pain",
  ["the journal of pan african studies"] = "J Pan Afr Stud",
  ["the journal of parapsychology"] = "J Parapsychol",
  ["the journal of parasitology"] = "J Parasitol",
  ["the journal of parenteral therapy"] = "J Parenter Ther",
  ["the journal of pastoral care & counseling : jpcc"] = "J Pastoral Care Counsel",
  ["the journal of pathology"] = "J Pathol",
  ["the journal of pathology and bacteriology"] = "J Pathol Bacteriol",
  ["the journal of pathology. clinical research"] = "J Pathol Clin Res",
  ["the journal of peasant studies"] = "J Peasant Stud",
  ["the journal of pediatric endocrinology"] = "J Pediatr Endocrinol",
  ["the journal of pediatric pharmacology and therapeutics : jppt : the official journal of ppag"] = "J Pediatr Pharmacol Ther",
  ["the journal of pediatrics"] = "J Pediatr",
  ["the journal of pediatrics: x"] = "J Pediatr X",
  ["the journal of pedodontics"] = "J Pedod",
  ["the journal of peptide research : official journal of the american peptide society"] = "J Pept Res",
  ["the journal of perinatal & neonatal nursing"] = "J Perinat Neonatal Nurs",
  ["the journal of perinatal education"] = "J Perinat Educ",
  ["the journal of perinatal education : an aspo/lamaze publication"] = "J Perinat Educ",
  ["the journal of periodontology"] = "J Periodontol",
  ["the journal of pharmacology and experimental therapeutics"] = "J Pharmacol Exp Ther",
  ["the journal of pharmacy & law"] = "J Pharm Law",
  ["the journal of pharmacy & law / ohio northern university, the pharmacy-law institute"] = "J Pharm Law",
  ["the journal of pharmacy and pharmacology"] = "J Pharm Pharmacol",
  ["the journal of pharmacy technology : jpt : official publication of the association of pharmacy technicians"] = "J Pharm Technol",
  ["the journal of philosophy"] = "J Philos",
  ["the journal of philosophy, science & law"] = "J Philos Sci Law",
  ["the journal of physical and colloid chemistry"] = "J Phys Colloid Chem",
  ["the journal of physical chemistry"] = "J Phys Chem",
  ["the journal of physical chemistry a"] = "J. Phys. Chem. A",
  ["the journal of physical chemistry letters"] = "J Phys Chem Lett",
  ["the journal of physical chemistry. a"] = "J Phys Chem A",
  ["the journal of physical chemistry. b"] = "J Phys Chem B",
  ["the journal of physical chemistry. c, nanomaterials and interfaces"] = "J Phys Chem C Nanomater Interfaces",
  ["the journal of physical fitness and sports medicine"] = "J Phys Fit Sports Med",
  ["the journal of physician assistant education : the official journal of the physician assistant education association"] = "J Physician Assist Educ",
  ["the journal of physics and chemistry of solids"] = "J Phys Chem Solids",
  ["the journal of physiological sciences : jps"] = "J Physiol Sci",
  ["the journal of physiology"] = "J Physiol",
  ["the journal of plant hydraulics"] = "J Plant Hydraul",
  ["the journal of plastic and reconstructive surgical nursing : official organ of the american society of plastic and reconstructive surgical nurses"] = "J Plast Reconstr Surg Nurs",
  ["the journal of political economy"] = "J Polit Econ",
  ["the journal of political philosophy"] = "J Polit Philos",
  ["the journal of politics"] = "J Polit",
  ["the journal of popular film and television : jpf & t"] = "J Pop Film Telev",
  ["the journal of positive psychology"] = "J Posit Psychol",
  ["the journal of poultry science"] = "J Poult Sci",
  ["the journal of poverty and social justice : research, policy, practice"] = "J Poverty Soc Justice",
  ["the journal of practical nursing"] = "J Pract Nurs",
  ["the journal of precision medicine"] = "J Precis Med",
  ["the journal of prevention of alzheimer's disease"] = "J Prev Alzheimers Dis",
  ["the journal of preventive dentistry"] = "J Prev Dent",
  ["the journal of primary prevention"] = "J Prim Prev",
  ["the journal of privacy and confidentiality"] = "J Priv Confid",
  ["the journal of problem solving"] = "J Probl Solving",
  ["the journal of product innovation management"] = "J Prod Innov Manage",
  ["the journal of prosthetic dentistry"] = "J Prosthet Dent",
  ["the journal of protozoology"] = "J Protozool",
  ["the journal of protozoology research"] = "J Protozool Res",
  ["the journal of psychiatry & law"] = "J Psychiatry Law",
  ["the journal of psychohistory"] = "J Psychohist",
  ["the journal of psychology"] = "J Psychol",
  ["the journal of psychotherapy practice and research"] = "J Psychother Pract Res",
  ["the journal of raptor research"] = "J Raptor Res",
  ["the journal of rare disorders"] = "J Rare Disord",
  ["the journal of religion"] = "J Relig",
  ["the journal of religious ethics"] = "J Relig Ethics",
  ["the journal of religious history"] = "J Relig Hist",
  ["the journal of reproduction and development"] = "J Reprod Dev",
  ["the journal of reproductive medicine"] = "J Reprod Med",
  ["the journal of research administration"] = "J Res Adm",
  ["the journal of research and education in indian medicine"] = "J Res Educ Indian Med",
  ["the journal of research in crime and delinquency"] = "J Res Crime Delinq",
  ["the journal of research in indian medicine"] = "J Res Indian Med",
  ["the journal of research on the lepidoptera"] = "J Res Lepid",
  ["the journal of respiratory diseases"] = "J Respir Dis",
  ["the journal of retirement"] = "J Retire",
  ["the journal of rheumatology"] = "J Rheumatol",
  ["the journal of rheumatology. supplement"] = "J Rheumatol Suppl",
  ["the journal of risk and insurance"] = "J Risk Insur",
  ["the journal of rural health : official journal of the american rural health association and the national rural health care association"] = "J Rural Health",
  ["the journal of san diego history"] = "J San Diego Hist",
  ["the journal of school health"] = "J Sch Health",
  ["the journal of school nursing : the official publication of the national association of school nurses"] = "J Sch Nurs",
  ["the journal of science policy & governance"] = "J Sci Policy Gov",
  ["the journal of sex education"] = "J Sex Educ",
  ["the journal of sexual aggression"] = "J Sex Aggress",
  ["the journal of sexual medicine"] = "J Sex Med",
  ["the journal of small animal practice"] = "J Small Anim Pract",
  ["the journal of social and political affairs"] = "J Soc Polit Aff",
  ["the journal of social and political studies"] = "J Soc Polit Stud",
  ["the journal of social issues"] = "J Soc Issues",
  ["the journal of social psychology"] = "J Soc Psychol",
  ["the journal of social studies"] = "J Soc Stud",
  ["the journal of social welfare & family law"] = "J Soc Welf Fam Law",
  ["the journal of social welfare law"] = "J Soc Welfare Law",
  ["the journal of social, political, and economic studies"] = "J Soc Polit Econ Stud",
  ["the journal of socio-economics"] = "J Socio Econ",
  ["the journal of sociological studies"] = "J Sociol Stud",
  ["the journal of southern history"] = "J South Hist",
  ["the journal of space safety engineering"] = "J Space Saf Eng",
  ["the journal of special education"] = "J Spec Educ",
  ["the journal of speech and hearing disorders"] = "J Speech Hear Disord",
  ["the journal of speech and language pathology, applied behavior analysis"] = "J Speech Lang Pathol Appl Behav Anal",
  ["the journal of speech disorders"] = "J Speech Disord",
  ["the journal of spinal cord medicine"] = "J Spinal Cord Med",
  ["the journal of sports medicine"] = "J Sports Med",
  ["the journal of sports medicine and physical fitness"] = "J Sports Med Phys Fitness",
  ["the journal of state government"] = "J State Gov",
  ["the journal of state government / national conference of state legislatures [and] the council of state governments"] = "J State Gov",
  ["the journal of steroid biochemistry and molecular biology"] = "J Steroid Biochem Mol Biol",
  ["the journal of stone disease"] = "J Stone Dis",
  ["the journal of strain analysis for engineering design"] = "J Strain Anal Eng Des",
  ["the journal of stuttering therapy, advocacy, and research"] = "J Stutt Ther Advocacy Res",
  ["the journal of supercomputing"] = "J Supercomput",
  ["the journal of supercritical fluids"] = "J Supercrit Fluids",
  ["the journal of supportive oncology"] = "J Support Oncol",
  ["the journal of surgical research"] = "J Surg Res",
  ["the journal of swimming research"] = "J Swim Res",
  ["the journal of symbolic logic"] = "J. Symb. Log.",
  ["the journal of symplectic geometry"] = "J. Symplectic Geom.",
  ["the journal of systems and software"] = "J Syst Softw",
  ["the journal of targeted therapies in cancer"] = "J Target Ther Cancer",
  ["the journal of technology transfer"] = "J Technol Transf",
  ["the journal of tehran heart center"] = "J Tehran Heart Cent",
  ["the journal of the acoustical society of america"] = "J Acoust Soc Am",
  ["the journal of the alabama academy of science"] = "J Ala Acad Sci",
  ["the journal of the albert einstein medical center, philadelphia"] = "J Albert Einstein Med Cent (Phila)",
  ["the journal of the all india dental association"] = "J All India Dent Assoc",
  ["the journal of the american academy of gold foil operators"] = "J Am Acad Gold Foil Oper",
  ["the journal of the american academy of orthopaedic surgeons"] = "J Am Acad Orthop Surg",
  ["the journal of the american academy of psychiatry and the law"] = "J Am Acad Psychiatry Law",
  ["the journal of the american academy of psychoanalysis"] = "J Am Acad Psychoanal",
  ["the journal of the american academy of psychoanalysis and dynamic psychiatry"] = "J Am Acad Psychoanal Dyn Psychiatry",
  ["the journal of the american association of gynecologic laparoscopists"] = "J Am Assoc Gynecol Laparosc",
  ["the journal of the american association of nephrology nurses & technicians"] = "J Am Assoc Nephrol Nurses Tech",
  ["the journal of the american association of nurse anesthetists"] = "J Am Assoc Nurse Anesth",
  ["the journal of the american board of family practice"] = "J Am Board Fam Pract",
  ["the journal of the american board of family practice / american board of family practice"] = "J Am Board Fam Pract",
  ["the journal of the american college of certified wound specialists"] = "J Am Col Certif Wound Spec",
  ["the journal of the american college of clinical wound specialists"] = "J Am Coll Clin Wound Spec",
  ["the journal of the american college of dentists"] = "J Am Coll Dent",
  ["the journal of the american dental hygienists' association"] = "J Am Dent Hyg Assoc",
  ["the journal of the american dental hygienists’ association"] = "J Am Dent Hyg Assoc",
  ["the journal of the american killifish association"] = "J Am Killifish Assoc",
  ["the journal of the american osteopathic association"] = "J Am Osteopath Assoc",
  ["the journal of the american paraplegia society"] = "J Am Paraplegia Soc",
  ["the journal of the american society for geriatric dentistry"] = "J Am Soc Geriatr Dent",
  ["the journal of the american society for preventive dentistry"] = "J Am Soc Prev Dent",
  ["the journal of the american society for psychical research"] = "J Am Soc Psych Res",
  ["the journal of the american society of psychosomatic dentistry and medicine"] = "J Am Soc Psychosom Dent Med",
  ["the journal of the anthropological survey of india"] = "J Anthropol Surv India",
  ["the journal of the arkansas medical society"] = "J Ark Med Soc",
  ["the journal of the asian federation of obstetrics and gynaecology"] = "J Asian Fed Obstet Gynaecol",
  ["the journal of the association for persons with severe handicaps : official publication of the association for persons with severe handicaps"] = "J Assoc Pers Sev Handicaps",
  ["the journal of the association of medical women in india"] = "J Assoc Med Women India",
  ["the journal of the association of nurses in aids care : janac"] = "J Assoc Nurses AIDS Care",
  ["the journal of the association of physicians of india"] = "J Assoc Physicians India",
  ["the journal of the astronautical sciences"] = "J Astronaut Sci",
  ["the journal of the bergen county dental society. bergen county dental society"] = "J Bergen Cty Dent Soc",
  ["the journal of the bombay natural history society"] = "J Bombay Nat Hist Soc",
  ["the journal of the british archaeological association"] = "JBAA",
  ["the journal of the british menopause society"] = "J Br Menopause Soc",
  ["the journal of the california alliance for the mentally ill"] = "J Calif Alliance Ment Ill",
  ["the journal of the canadian chiropractic association"] = "J Can Chiropr Assoc",
  ["the journal of the canadian church historical society"] = "J Can Church Hist Soc",
  ["the journal of the canadian health libraries association"] = "J Can Health Libr Assoc",
  ["the journal of the canadian medical services"] = "J Can Med Serv",
  ["the journal of the christian medical association of india"] = "J Christ Med Assoc India",
  ["the journal of the college of general practitioners"] = "J Coll Gen Pract",
  ["the journal of the dental association of south africa = die tydskrif van die tandheelkundige vereniging van suid-afrika"] = "J Dent Assoc S Afr",
  ["the journal of the dental association of thailand"] = "J Dent Assoc Thai",
  ["the journal of the dental auxiliaries, malaya, malaysia"] = "J Dent Aux",
  ["the journal of the dental society of the state of new york"] = "J Dent Soc State  N Y",
  ["the journal of the egyptian medical association"] = "J Egypt Med Assoc",
  ["the journal of the egyptian public health association"] = "J Egypt Public Health Assoc",
  ["the journal of the faculty of radiologists. faculty of radiologists (great britain)"] = "J Fac Radiol",
  ["the journal of the florida medical association"] = "J Fla Med Assoc",
  ["the journal of the florida medical association. florida medical association"] = "J Fla Med Assoc",
  ["the journal of the florida state dental society"] = "J Fla State Dent Soc",
  ["the journal of the friends' historical society"] = "J Friends Hist Soc",
  ["the journal of the friends’ historical society"] = "J Friends Hist Soc",
  ["the journal of the greater houston dental society"] = "J Gt Houst Dent Soc",
  ["the journal of the hattori botanical laboratory"] = "Hattori Shokubutsu Kenkyujo hokoku|J. Hattori Bot. Lab.",
  ["the journal of the hattori botanical laboratory = hattori shokubutsu kenkyūjo hōkoku"] = "J Hattori Bot Lab",
  ["the journal of the history of childhood and youth"] = "J Hist Child Youth",
  ["the journal of the indian mathematical society"] = "J. Indian Math. Soc. (N.S.)",
  ["the journal of the indian mathematical society. new series"] = "J. Indian Math. Soc. (N.S.)",
  ["the journal of the indiana dental association"] = "J Indiana Dent Assoc",
  ["the journal of the indiana state dental association"] = "J Indiana State Dent Assoc",
  ["the journal of the indiana state medical association"] = "J Indiana State Med Assoc",
  ["the journal of the institute of hospital engineering"] = "J Inst Hosp Eng",
  ["the journal of the interdenominational theological center. interdenominational theological center (atlanta, ga.)"] = "J Interdenominatl Theol Cent",
  ["the journal of the international association of medical science educators : jiamse"] = "J Int Assoc Med Sci Educ",
  ["the journal of the international college of surgeons"] = "J Int Coll Surg",
  ["the journal of the international society of physical and rehabilitation medicine"] = "J Int Soc Phys Rehabil Med",
  ["the journal of the kansas bar association. kansas bar association"] = "J Kans Bar Assoc",
  ["the journal of the kansas medical society"] = "J Kans Med Soc",
  ["the journal of the kansas state dental association"] = "J Kans State Dent Assoc",
  ["the journal of the kentucky medical association"] = "J Ky Med Assoc",
  ["the journal of the kentucky state medical association"] = "J Ky State Med Assoc",
  ["the journal of the korean physical society"] = "J Korean Phys Soc",
  ["the journal of the korean society for microbiology"] = "J Korean Soc Microbiol",
  ["the journal of the kuwait medical association"] = "J Kuwait Med Assoc",
  ["the journal of the london mathematical society"] = "J. London Math. Soc. (2)",
  ["the journal of the louisiana state medical society : official organ of the louisiana state medical society"] = "J La State Med Soc",
  ["the journal of the maine medical association"] = "J Maine Med Assoc",
  ["the journal of the medical society of new jersey"] = "J Med Soc N J",
  ["the journal of the michigan dental association"] = "J Mich Dent Assoc",
  ["the journal of the michigan state dental association"] = "J Mich State Dent Assoc",
  ["the journal of the nebraska dental association"] = "J Nebr Dent Assoc",
  ["the journal of the new jersey state dental society"] = "J N J State Dent Soc",
  ["the journal of the new york state nurses' association"] = "J N Y State Nurses Assoc",
  ["the journal of the new york state nurses’ association"] = "J N Y State Nurses Assoc",
  ["the journal of the new york state school nurse teachers association"] = "J N Y State Sch Nurse Teach Assoc",
  ["the journal of the north carolina dental society"] = "J N C Dent Soc",
  ["the journal of the oklahoma state medical association"] = "J Okla State Med Assoc",
  ["the journal of the ontario dental association"] = "J Ont Dent Assoc",
  ["the journal of the operating room research institute : jorri"] = "J Oper Room Res Inst",
  ["the journal of the operational research society"] = "J Oper Res Soc",
  ["the journal of the oregon dental association"] = "J Oreg Dent Assoc",
  ["the journal of the pharmacy society of wisconsin : jpsw"] = "J Pharm Soc Wis",
  ["the journal of the philadelphia county dental society"] = "J Phila Cty Dent Soc",
  ["the journal of the philippine dental association"] = "J Philipp Dent Assoc",
  ["the journal of the polynesian society. polynesian society (n.z.)"] = "J Polyn Soc",
  ["the journal of the robert h. lurie cancer center of northwestern university"] = "J Robert H Lurie Cancer Cent Northwest Univ",
  ["the journal of the royal anthropological institute"] = "J R Anthropol Inst",
  ["the journal of the royal army veterinary corps"] = "J R Army Vet Corps",
  ["the journal of the royal astronomical society of canada. royal astronomical society of canada"] = "J R Astron Soc Can",
  ["the journal of the royal college of general practitioners"] = "J R Coll Gen Pract",
  ["the journal of the royal college of general practitioners. occasional paper"] = "J R Coll Gen Pract Occas Pap",
  ["the journal of the royal college of physicians of edinburgh"] = "J R Coll Physicians Edinb",
  ["the journal of the royal faculty of medicine of iraq"] = "J R Fac Med Iraq",
  ["the journal of the royal institute of public health and hygiene"] = "J R Inst Public Health",
  ["the journal of the royal society for the promotion of health"] = "J R Soc Promot Health",
  ["the journal of the rutgers university library"] = "J Rutgers Univ Libr",
  ["the journal of the siam society"] = "J Siam Soc",
  ["the journal of the singapore paediatric society"] = "J Singapore Paediatr Soc",
  ["the journal of the society of occupational medicine"] = "J Soc Occup Med",
  ["the journal of the tennessee dental association"] = "J Tenn Dent Assoc",
  ["the journal of the tennessee state medical association. tennessee state medical association"] = "J Tn State Med Assoc",
  ["the journal of the texas dental hygienists' association. texas dental hygienists' association"] = "J Texas Dent Hyg Assoc",
  ["the journal of the texas dental hygienists’ association. texas dental hygienists’ association"] = "J Texas Dent Hyg Assoc",
  ["the journal of the torrey botanical society"] = "J Torrey Bot Soc",
  ["the journal of the walters art gallery"] = "JWaltersArtGal",
  ["the journal of the west virginia historical association. west virginia historical association"] = "J W Va Hist Assoc",
  ["the journal of the western society of periodontology/periodontal abstracts"] = "J West Soc Periodontol Periodontal Abstr",
  ["the journal of the wisconsin dental association"] = "J Wis Dent Assoc",
  ["the journal of the wisconsin state dental society"] = "J Wis State Dent Soc",
  ["the journal of thoracic and cardiovascular surgery"] = "J Thorac Cardiovasc Surg",
  ["the journal of thoracic surgery"] = "J Thorac Surg",
  ["the journal of toxicological sciences"] = "J Toxicol Sci",
  ["the journal of trauma"] = "J Trauma",
  ["the journal of trauma and acute care surgery"] = "J Trauma Acute Care Surg",
  ["the journal of tropical geography"] = "J Trop Geogr",
  ["the journal of tropical medicine"] = "J Trop Med",
  ["the journal of tropical medicine and hygiene"] = "J Trop Med Hyg",
  ["the journal of tropical pediatrics and african child health"] = "J Trop Pediatr Afr Child Health",
  ["the journal of tropical pediatrics and environmental child health"] = "J Trop Pediatr Environ Child Health",
  ["the journal of urban psychiatry"] = "J Urban Psychiatry",
  ["the journal of urology"] = "J Urol",
  ["the journal of value inquiry"] = "J Value Inq",
  ["the journal of vascular access"] = "J Vasc Access",
  ["the journal of venereal disease information"] = "J Vener Dis Inf",
  ["the journal of venomous animals and toxins including tropical diseases"] = "J Venom Anim Toxins Incl Trop Dis",
  ["the journal of veterinary medical science"] = "J Vet Med Sci",
  ["the journal of veterinary medical science / the japanese society of veterinary science"] = "J Vet Med Sci",
  ["the journal of viral diseases"] = "J Viral Dis",
  ["the journal of vitaminology"] = "J Vitaminol (Kyoto)",
  ["the journal of volunteer administration"] = "J Volunt Adm",
  ["the journal of west african languages"] = "J West Afr Lang",
  ["the journal of wildlife management"] = "J Wildl Manage",
  ["the journal of workplace learning"] = "J Workplace Learn",
  ["the journal of world intellectual property"] = "J World Intellect Prop",
  ["the journal of writing analytics"] = "J Writ Anal",
  ["the journal-lancet"] = "J Lancet",
  ["the journalism quarterly"] = "Journal Q",
  ["the journals of gerontology. series a, biological sciences and medical sciences"] = "J Gerontol A Biol Sci Med Sci",
  ["the journals of gerontology. series b, psychological sciences and social sciences"] = "J Gerontol B Psychol Sci Soc Sci",
  ["the judges' journal"] = "Judges J",
  ["the judges’ journal"] = "Judges J",
  ["the juridical review"] = "Jurid Rev",
  ["the kansas geographer"] = "Kans Geogr",
  ["the kansas historical quarterly"] = "Kans Hist Q",
  ["the kansas nurse"] = "Kans Nurse",
  ["the kaohsiung journal of medical sciences"] = "Kaohsiung J Med Sci",
  ["the keio journal of medicine"] = "Keio J Med",
  ["the key reporter"] = "Key Report",
  ["the kidney"] = "Kidney",
  ["the king faisal specialist hospital medical journal"] = "King Faisal Spec Hosp Med J",
  ["the kitasato archives of experimental medicine"] = "Kitasato Arch Exp Med",
  ["the kitasato medical journal"] = "Kitasato Med J",
  ["the kiva"] = "Kiva",
  ["the kluwer international series in engineering and computer science"] = "Kluwer Int. Ser. Eng. Comput. Sci.",
  ["the kluwer international series in engineering and computer science. knowledge representation, learning and expert systems"] = "Kluwer Internat. Ser. Engrg. Comput. Sci. Knowledge Represent., Learn. Expert Systems",
  ["the kluwer international series in software engineering"] = "Kluwer Int. Ser. Softw. Eng.",
  ["the kluwer international series on discrete event dynamic systems"] = "Kluwer Int. Ser. Discrete Event Dyn. Syst.",
  ["the knee"] = "Knee",
  ["the knowledge engineering review"] = "Knowl Eng Rev",
  ["the kobe journal of medical sciences"] = "Kobe J Med Sci",
  ["the korean journal of biochemistry : official journal of the korean biochemical society"] = "Korean J Biochem",
  ["the korean journal of chemical engineering"] = "Korean J Chem Eng",
  ["the korean journal of computational & applied mathematics"] = "Korean J. Comput. Appl. Math.",
  ["the korean journal of gastroenterology = taehan sohwagi hakhoe chi"] = "Korean J Gastroenterol",
  ["the korean journal of genetics"] = "Korean J Genet",
  ["the korean journal of hematology"] = "Korean J Hematol",
  ["the korean journal of hepatology"] = "Korean J Hepatol",
  ["the korean journal of internal medicine"] = "Korean J Intern Med",
  ["the korean journal of laboratory medicine"] = "Korean J Lab Med",
  ["the korean journal of malacology"] = "Korean J Malacol",
  ["the korean journal of mathematics"] = "Korean J. Math.",
  ["the korean journal of pain"] = "Korean J Pain",
  ["the korean journal of parasitology"] = "Korean J Parasitol",
  ["the korean journal of physiology & pharmacology : official journal of the korean physiological society and the korean society of pharmacology"] = "Korean J Physiol Pharmacol",
  ["the korean journal of sericultural science"] = "Korean J Sericult Sci",
  ["the korean journal of thoracic and cardiovascular surgery"] = "Korean J Thorac Cardiovasc Surg",
  ["the kumamoto medical journal"] = "Kumamoto Med J",
  ["the kurume medical journal"] = "Kurume Med J",
  ["the kuwait medical journal : kmj : the official journal of the kuwait medical association"] = "Kuwait Med J",
  ["the labor lawyer"] = "Labor Lawyer",
  ["the laboratory"] = "Laboratory",
  ["the laboratory digest"] = "Lab Dig",
  ["the lactation review"] = "Lactation Rev",
  ["the ladies' home journal"] = "Ladies Home J",
  ["the ladies’ home journal"] = "Ladies Home J",
  ["the lahey clinic bulletin"] = "Lahey Clin Bull",
  ["the lamp"] = "Lamp",
  ["the lancet infectious diseases"] = "Lancet Infect Dis",
  ["the lancet oncology"] = "Lancet Oncol",
  ["the lancet regional health. western pacific"] = "Lancet Reg Health West Pac",
  ["the lancet. child & adolescent health"] = "Lancet Child Adolesc Health",
  ["the lancet. diabetes & endocrinology"] = "Lancet Diabetes Endocrinol",
  ["the lancet. digital health"] = "Lancet Digit Health",
  ["the lancet. gastroenterology & hepatology"] = "Lancet Gastroenterol Hepatol",
  ["the lancet. global health"] = "Lancet Glob Health",
  ["the lancet. haematology"] = "Lancet Haematol",
  ["the lancet. healthy longevity"] = "Lancet Healthy Longev",
  ["the lancet. hiv"] = "Lancet HIV",
  ["the lancet. infectious diseases"] = "Lancet Infect Dis",
  ["the lancet. microbe"] = "Lancet Microbe",
  ["the lancet. neurology"] = "Lancet Neurol",
  ["the lancet. oncology"] = "Lancet Oncol",
  ["the lancet. planetary health"] = "Lancet Planet Health",
  ["the lancet. psychiatry"] = "Lancet Psychiatry",
  ["the lancet. public health"] = "Lancet Public Health",
  ["the lancet. respiratory medicine"] = "Lancet Respir Med",
  ["the lancet. rheumatology"] = "Lancet Rheumatol",
  ["the laryngoscope"] = "Laryngoscope",
  ["the latin american journal of aquatic mammals : (lajam)"] = "Lat Am J Aquat Mamm",
  ["the law quarterly review"] = "Law Q Rev",
  ["the law reports. house of lords, and judicial committee of the privy council, and peerage cases"] = "Law Rep House Lords",
  ["the law teacher"] = "Law Teach",
  ["the leadership quarterly"] = "Leadersh Q",
  ["the lebanese pharmaceutical journal. la revue pharmaceutique libanaise"] = "Leban Pharm J",
  ["the leech"] = "Leech (Wales)",
  ["the library"] = "Library (Lond)",
  ["the library chronicle of the university of texas at austin"] = "Libr Chron Univ Tex Austin",
  ["the library quarterly"] = "Libr Q",
  ["the libyan journal of medicine"] = "Libyan J Med",
  ["the licentiate; a monthly journal of medicine and surgery"] = "Licentiate",
  ["the linacre quarterly"] = "Linacre Q",
  ["the linnean : newsletter and proceedings of the linnean society of london. linnean society of london"] = "Linnean",
  ["the lippincott health promotion letter"] = "Lippincott Health Promot Lett",
  ["the liverpool law review"] = "Liverp Law Rev",
  ["the liverpool medico-chirurgical journal"] = "Liverp Med Chir J",
  ["the local historian"] = "Local Hist",
  ["the log"] = "Log",
  ["the log of mystic seaport"] = "Log Mystic Seapt",
  ["the london archaeologist"] = "Lond Archaeol",
  ["the london hospital gazette"] = "Lond Hosp Gaz",
  ["the london journal"] = "Lond J",
  ["the london medical and physical journal"] = "Lond Med Phys J",
  ["the london medical journal"] = "Lond Med J",
  ["the london, edinburgh and dublin philosophical magazine and journal of science"] = "Lond. Edinb. Dublin Philos. Mag. J. Sci.",
  ["the long island historical journal"] = "Long Island Hist J",
  ["the los angeles times"] = "Los Angeles Times",
  ["the lovelace clinic review"] = "Lovelace Clin Rev",
  ["the magazine of albemarle county history"] = "Mag Albemarle Cty Hist",
  ["the maghreb review. majallat al-maghrib"] = "Maghreb Rev",
  ["the malawian geographer"] = "Malawian Geogr",
  ["the malayan economic review : the journal of the economic society of singapore, the department of economics and statistics and the economic research centre of the university of singapore"] = "Malay Econ Rev",
  ["the malaysian journal of medical sciences : mjms"] = "Malays J Med Sci",
  ["the malaysian journal of pathology"] = "Malays J Pathol",
  ["the male nurses' journal"] = "Male Nurses J",
  ["the male nurses’ journal"] = "Male Nurses J",
  ["the malloch room newsletter"] = "Malloch Room Newsl",
  ["the manchester medical gazette"] = "Manch Med Gaz",
  ["the manchester physics series"] = "Manchester Phys. Ser.",
  ["the manchester school of economic and social studies"] = "Manchester Sch Econ Soc Stud",
  ["the mankind quarterly"] = "Mankind Q",
  ["the marine corps gazette"] = "Mar Corps Gaz",
  ["the mariner's mirror"] = "Mar Mirror",
  ["the mariner’s mirror"] = "Mar Mirror",
  ["the maryland bar journal"] = "Md Bar J",
  ["the maryland family doctor"] = "Md Fam Dr",
  ["the maryland historian"] = "Md Hist",
  ["the maryland nurse"] = "Md Nurse",
  ["the massachusetts health journal"] = "Mass Health J",
  ["the massachusetts nurse"] = "Mass Nurse",
  ["the massachusetts review"] = "Mass Rev",
  ["the masterkey for indian lore and history"] = "Masterkey Ind Lore Hist",
  ["the mathematica journal"] = "Math J",
  ["the mathematical gazette"] = "Math. Gaz.",
  ["the mathematical intelligencer"] = "Math. Intelligencer",
  ["the mathematical scientist"] = "Math. Sci.",
  ["the mathematics education"] = "Math. Ed. (Siwan)",
  ["the mathematics student"] = "Math. Student",
  ["the mayan"] = "Mayan",
  ["the mcgill dental review"] = "Mcgill Dent Rev",
  ["the mcgill news"] = "McGill News",
  ["the medical and dental bulletin"] = "Med Dent Bull",
  ["the medical and physical journal"] = "Med Phys J",
  ["the medical annals of the district of columbia"] = "Med Ann Dist Columbia",
  ["the medical annual"] = "Med Annu",
  ["the medical bulletin"] = "Med Bull (N Y)",
  ["the medical bulletin of the u. s. army far east"] = "Med Bull US",
  ["the medical clinics of north america"] = "Med Clin North Am",
  ["the medical comment"] = "Med Comment",
  ["the medical journal of australia"] = "Med J Aust",
  ["the medical journal of cairo university"] = "Med J Cairo Univ",
  ["the medical journal of malaya"] = "Med J Malaya",
  ["the medical journal of malaysia"] = "Med J Malaysia",
  ["the medical journal of the egyptian armed forces"] = "Med J Egypt Armed Forces",
  ["the medical journal of the south-west"] = "Med J Southwest",
  ["the medical letter on drugs and therapeutics"] = "Med Lett Drugs Ther",
  ["the medical malpractice cost containment journal"] = "Med Malpract Cost Containment J",
  ["the medical officer"] = "Med Off",
  ["the medical press"] = "Med Press",
  ["the medical press of egypt"] = "Med Press Egypt",
  ["the medical quarterly review"] = "Med Q Rev",
  ["the medical staff counselor"] = "Med Staff Couns",
  ["the medical way"] = "Med Way",
  ["the medical world"] = "Med World (New York)",
  ["the medical-moral newsletter"] = "Med Moral Newsl",
  ["the medico-chirurgical journal and review"] = "Med Chir J Rev",
  ["the medico-chirurgical review"] = "Med Chir Rev",
  ["the medico-chirurgical review and journal of medical science. analytical series"] = "Med Chir Rev J Med Sci Anal Ser",
  ["the medico-legal and criminological review"] = "Med Leg Criminol Rev",
  ["the medico-legal journal"] = "Med Leg J",
  ["the medicus"] = "Medicus",
  ["the mediæval journal"] = "Mediaev J",
  ["the meharri-dent"] = "Meharri Dent",
  ["the mendel newsletter; archival resources for the history of genetics & allied sciences"] = "Mendel Newsl",
  ["the mennonite quarterly review"] = "Mennon Q Rev",
  ["the mental health clinician"] = "Ment Health Clin",
  ["the mental lexicon"] = "Ment Lex",
  ["the merck report"] = "Merck Rep",
  ["the metropolitan museum of art bulletin"] = "BMetrMus",
  ["the michigan historical review"] = "Mich Hist Rev",
  ["the michigan mathematical journal"] = "Michigan Math. J.",
  ["the michigan nurse"] = "Mich Nurse",
  ["the middle east journal"] = "Middle East J",
  ["the middle east medical journal"] = "Middle East Med J",
  ["the middlesex hospital journal"] = "Middx Hosp J",
  ["the middlesex veterinarian"] = "Middx Vet",
  ["the midwest quarterly"] = "Midwest Q (Pittsb)",
  ["the milbank memorial fund quarterly"] = "Milbank Mem Fund Q",
  ["the milbank memorial fund quarterly. health and society"] = "Milbank Mem Fund Q Health Soc",
  ["the milbank quarterly"] = "Milbank Q",
  ["the military engineer"] = "Mil Eng",
  ["the minneapolis district dental journal"] = "Minneap Dist Dent J",
  ["the minor planet bulletin"] = "Minor Planet Bull",
  ["the mississippi doctor"] = "Miss Doct",
  ["the mississippi quarterly"] = "Miss Q",
  ["the missouri nurse"] = "Mo Nurse",
  ["the mit press series in structural mechanics"] = "MIT Press Ser. Structural Mech.",
  ["the moccasin"] = "Moccasin",
  ["the modern churchman"] = "Mod Churchm",
  ["the modern law review"] = "Mod Law Rev",
  ["the modern schoolman"] = "Mod Schoolman",
  ["the modern schoolman: a quarterly journal of philosophy"] = "ModSch",
  ["the monist"] = "Monist",
  ["the month"] = "Month",
  ["the morgan kaufmann series in computer graphics and geometric modeling"] = "Morgan Kaufmann Ser. Comput. Graph. Geom. Model.",
  ["the morgan kaufmann series in data management systems"] = "Morgan Kaufmann Ser. Data Management Systems",
  ["the mother"] = "Mother",
  ["the mount sinai journal of medicine, new york"] = "Mt Sinai J Med",
  ["the mouth mirror"] = "Mouth Mirror",
  ["the muslim world"] = "Muslim World",
  ["the mycologist"] = "Mycologist",
  ["the naham management journal"] = "NAHAM Manage J",
  ["the national catholic bioethics quarterly"] = "Natl Cathol Bioeth Q",
  ["the national dean's list"] = "Deans List",
  ["the national dean’s list"] = "Deans List",
  ["the national eclectic medical quarterly"] = "Natl Elect Med Q",
  ["the national engineer"] = "Natl Eng",
  ["the national jewish monthly"] = "Natl Jew Mon",
  ["the national law journal"] = "Natl Law J",
  ["the national medical journal of india"] = "Natl Med J India",
  ["the national underwriter. life & health insurance edition"] = "Natl Underwrit Life Health",
  ["the national union catalog of manuscript collections"] = "Natl Union Cat Manuscr Collect",
  ["the nation’s health"] = "Nations Health",
  ["the natural history bulletin of the siam society"] = "Nat Hist Bull Siam Soc",
  ["the natural products journal"] = "Nat Prod J",
  ["the nautilus"] = "Nautilus (Philadelphia)",
  ["the ncsdha dental hygienist"] = "NCSDHA Dent Hyg",
  ["the nebraska medical journal"] = "Nebr Med J",
  ["the nebraska state medical journal"] = "Nebr State Med J",
  ["the negro educational review"] = "Negro Educ Rev",
  ["the nepali mathematical sciences report"] = "Nepali Math. Sci. Rep.",
  ["the nervous child"] = "Nerv Child",
  ["the netherlands journal of medicine"] = "Neth J Med",
  ["the netherlands journal of sociology : sociologia neerlandica"] = "Neth J Sociol",
  ["the netherlands journal of surgery"] = "Neth J Surg",
  ["the netherlands' journal of social sciences : a publication of the netherlands' sociological and anthropological society"] = "Neth J Soc Sci",
  ["the neurips '18 competition : from machine learning to intelligent conversations"] = "NeurIPS 18 Compet (2018)",
  ["the neurodiagnostic journal"] = "Neurodiagn J",
  ["the neurohospitalist"] = "Neurohospitalist",
  ["the neurologist"] = "Neurologist",
  ["the neurologist (hyderabad, india)"] = "Neurologist (Hyderabad)",
  ["the neuroradiology journal"] = "Neuroradiol J",
  ["the neuroscientist : a review journal bringing neurobiology, neurology and psychiatry"] = "Neuroscientist",
  ["the new bioethics : a multidisciplinary journal of biotechnology and the body"] = "New Bioeth",
  ["the new biologist"] = "New Biol",
  ["the new dentist : the official journal of the american student dental association"] = "New Dent",
  ["the new england historical and genealogical register"] = "N Engl Hist Geneal Regist",
  ["the new england historical and genealogical register / new-england historic genealogical society"] = "N Engl Hist Geneal Regist",
  ["the new england journal of history"] = "New Engl J Hist",
  ["the new england journal of medicine"] = "N Engl J Med",
  ["the new england quarterly"] = "New Engl Q",
  ["the new hungarian quarterly"] = "New Hung Q",
  ["the new iraqi journal of medicine"] = "New Iraqi J Med",
  ["the new law journal"] = "New Law J",
  ["the new microbiologica"] = "New Microbiol",
  ["the new microbiologica : official journal of the italian society for medical, odontoiatric, and clinical microbiology (simmoc)"] = "New Microbiol",
  ["the new orleans medical and surgical journal"] = "New Orleans Med Surg J",
  ["the new philosophy"] = "New Philos",
  ["the new physician"] = "New Physician",
  ["the new phytologist"] = "New Phytol",
  ["the new review of hypermedia and multimedia"] = "New Rev Hypermedia Multimed",
  ["the new scholasticism"] = "New Scholasticism",
  ["the new school psychology bulletin"] = "New School Psychol Bull",
  ["the new universities quarterly"] = "New Univ Q",
  ["the new world of irish nursing : official journal of irish nurses organisation and national council of nurses"] = "New World Ir Nurs",
  ["the new york dental journal"] = "NY Dent J",
  ["the new york journal of dentistry"] = "N Y J Dent",
  ["the new york medical quarterly"] = "N Y Med Q",
  ["the new york physician"] = "N Y Physician",
  ["the new york review of books"] = "New York Rev Books",
  ["the new york state dental journal"] = "N Y State Dent J",
  ["the new york times"] = "NY Times (Print)",
  ["the new york times magazine"] = "N Y Times Mag",
  ["the new york times on the web"] = "N Y Times Web",
  ["the new zealand dental journal"] = "N Z Dent J",
  ["the new zealand entomologist"] = "N Z Entomol",
  ["the new zealand family physician"] = "N Z Fam Physician",
  ["the new zealand journal of history"] = "N Z J Hist",
  ["the new zealand journal of science and technology. b. general section"] = "N Z J Sci Technol Sect B",
  ["the new zealand law journal : nzlj"] = "N Z Law J",
  ["the new zealand law reports"] = "N Z Law Rep",
  ["the new zealand medical journal"] = "N Z Med J",
  ["the new zealand nursing journal. kai tiaki"] = "N Z Nurs J",
  ["the new zealand psychologist"] = "NZ Psychol",
  ["the new-england galaxy"] = "N Engl Galaxy",
  ["the newsletter of biomedical safety & standards"] = "Newsl Biomed Saf Stand",
  ["the nigerian journal of economic and social studies"] = "Niger J Econ Soc Stud",
  ["the nigerian medical practitioner"] = "Niger Med Pract",
  ["the nigerian nurse"] = "Niger Nurse",
  ["the nigerian postgraduate medical journal"] = "Niger Postgrad Med J",
  ["the nih catalyst : a publication for nih intramural scientists"] = "NIH Catal",
  ["the nihon university journal of medicine"] = "Nihon Univ J Med",
  ["the norden news"] = "Norden News",
  ["the north american review"] = "North Am Rev",
  ["the north american veterinarian"] = "North Am Vet",
  ["the north carolina historical review"] = "N C Hist Rev",
  ["the north wing; the magazine of the sheffield medical and dental schools"] = "North Wing",
  ["the northern ireland law reports"] = "North Irel Law Rep",
  ["the northern ireland legal quarterly"] = "North Irel Leg Q",
  ["the northern mariner : journal of the canadian nautical research society = le marin du nord : revue de société canadienne pour la recherche nautique"] = "North Mar",
  ["the notre dame law review"] = "Notre Dame Law Rev",
  ["the notre dame lawyer"] = "Notre Dame Lawyer",
  ["the nova scotia historical quarterly"] = "NS Hist Q",
  ["the nova scotia medical bulletin"] = "N S Med Bull",
  ["the nucleus : an international journal of cytology and allied topics"] = "Nucleus (Calcutta)",
  ["the nucleus; an international journal of cytology and allied topics"] = "Nucleus",
  ["the numismatic chronicle. the journal of the royal society"] = "NumChron",
  ["the nurse practitioner"] = "Nurse Pract",
  ["the nursing clinics of north america"] = "Nurs Clin North Am",
  ["the nursing journal of india"] = "Nurs J India",
  ["the nursing journal of singapore"] = "Nurs J Singapore",
  ["the oalma journal"] = "Oalma J",
  ["the obstetrician & gynaecologist : the journal for continuing professional development from the royal college of obstetricians & gynaecologists"] = "Obstet Gynaecol",
  ["the occupational health nurse"] = "Occup Health Nurse (Auckl)",
  ["the ochsner journal"] = "Ochsner J",
  ["the ocular surface"] = "Ocul Surf",
  ["the oecd observer. organisation for economic co-operation and development"] = "OECD Obs",
  ["the office"] = "Office",
  ["the ohio dental journal"] = "Ohio Dent J",
  ["the ohio journal of science"] = "Ohio J Sci",
  ["the ohio state medical journal"] = "Ohio State Med J",
  ["the okinawan journal of american studies"] = "Okinawan J Am Stud",
  ["the oklahoma nurse"] = "Okla Nurse",
  ["the old northwest"] = "Old Northwest",
  ["the oncologist"] = "Oncologist",
  ["the oncology nurse"] = "Oncol Nurse",
  ["the oncology pharmacist"] = "Oncol Pharm",
  ["the onderstepoort journal of veterinary research"] = "Onderstepoort J Vet Res",
  ["the onderstepoort journal of veterinary science and animal industry"] = "Onderstepoort J Vet Sci Anim Ind",
  ["the online journal of current clinical trials"] = "Online J Curr Clin Trials",
  ["the online journal of distance education and e-learning : tojdel"] = "Online J Distance Educ Elearn",
  ["the online journal of knowledge synthesis for nursing"] = "Online J Knowl Synth Nurs",
  ["the ontario psychologist"] = "Ont Psychol",
  ["the open addiction journal"] = "Open Addict J",
  ["the open agriculture journal"] = "Open Agric J",
  ["the open aids journal"] = "Open AIDS J",
  ["the open allergy journal"] = "Open Allergy J",
  ["the open analytical chemistry journal"] = "Open Anal Chem J",
  ["the open anthropology journal"] = "Open Anthropol J",
  ["the open autoimmunity journal"] = "Open Autoimmun J",
  ["the open biochemistry journal"] = "Open Biochem J",
  ["the open bioinformatics journal"] = "Open Bioinforma J",
  ["the open biology journal"] = "Open Biol J",
  ["the open biomarkers journal"] = "Open Biomark J",
  ["the open biomedical engineering journal"] = "Open Biomed Eng J",
  ["the open biotechnology journal"] = "Open Biotechnol J",
  ["the open bone journal"] = "Open Bone J",
  ["the open book series"] = "Open Book Ser.",
  ["the open cancer immunology journal"] = "Open Cancer Immunol J",
  ["the open cancer journal"] = "Open Cancer J",
  ["the open cardiovascular medicine journal"] = "Open Cardiovasc Med J",
  ["the open cell development & biology journal"] = "Open Cell Dev Biol J",
  ["the open cell signaling journal"] = "Open Cell Signal J",
  ["the open circulation & vascular journal"] = "Open Circ Vasc J",
  ["the open clinical cancer journal"] = "Open Clin Cancer J",
  ["the open complementary medicine journal"] = "Open Complement Med J",
  ["the open conference proceedings journal"] = "Open Conf Proc J",
  ["the open critical care medicine journal"] = "Open Crit Care Med J",
  ["the open dentistry journal"] = "Open Dent J",
  ["the open dermatology journal"] = "Open Dermatol J",
  ["the open diabetes journal"] = "Open Diabetes J",
  ["the open drug discovery journal"] = "Open Drug Discov J",
  ["the open drug metabolism journal"] = "Open Drug Metab J",
  ["the open endocrinology journal"] = "Open Endocrinol J",
  ["the open entomology journal"] = "Open Entomol J",
  ["the open enzyme inhibition journal"] = "Open Enzym Inhib J",
  ["the open epidemiology journal"] = "Open Epidemiol J",
  ["the open family studies journal"] = "Open Fam Stud J",
  ["the open forensic science journal"] = "Open Forensic Sci J",
  ["the open gene therapy journal"] = "Open Gene Ther J",
  ["the open genomics journal"] = "Open Genomics J",
  ["the open geriatric medicine journal"] = "Open Geriatr Med J",
  ["the open health services and policy journal"] = "Open Health Serv Policy J",
  ["the open hypertension journal"] = "Open Hypertens J",
  ["the open immunology journal"] = "Open Immunol J",
  ["the open infectious diseases journal"] = "Open Infect Dis J",
  ["the open inflammation journal"] = "Open Inflamm J",
  ["the open information systems journal"] = "Open Inf Syst J",
  ["the open journal of occupational therapy"] = "Open J Occup Ther",
  ["the open magnetic resonance reviews"] = "Open Magn Reson Rev",
  ["the open marine biology journal"] = "Open Mar Biol J",
  ["the open medical imaging journal"] = "Open Med Imaging J",
  ["the open medical informatics journal"] = "Open Med Inform J",
  ["the open medicinal chemistry journal"] = "Open Med Chem J",
  ["the open microbiology journal"] = "Open Microbiol J",
  ["the open mycology journal"] = "Open Mycol J",
  ["the open nanomedicine journal"] = "Open Nanomed J",
  ["the open neuroimaging journal"] = "Open Neuroimag J",
  ["the open neurology journal"] = "Open Neurol J",
  ["the open neuropsychopharmacology journal"] = "Open Neuropsychopharmacol J",
  ["the open neuroscience journal"] = "Open Neurosci J",
  ["the open nuclear medicine journal"] = "Open Nucl Med J",
  ["the open nursing journal"] = "Open Nurs J",
  ["the open nutraceuticals journal"] = "Open Nutraceuticals J",
  ["the open nutrition journal"] = "Open Nutr J",
  ["the open obesity journal"] = "Open Obes J",
  ["the open ophthalmology journal"] = "Open Ophthalmol J",
  ["the open ornithology journal"] = "Open Ornithol J",
  ["the open orthopaedics journal"] = "Open Orthop J",
  ["the open pacing, electrophysiology & therapy journal"] = "Open Pacing Electrophysiol Ther J",
  ["the open pain journal"] = "Open Pain J",
  ["the open parasitology journal"] = "Open Parasitol J",
  ["the open pathology journal"] = "Open Pathol J",
  ["the open pediatric medicine journal"] = "Open Pediatr Med Journal",
  ["the open pharmacology journal"] = "Open Pharmacol J",
  ["the open prostate cancer journal"] = "Open Prost Cancer J",
  ["the open proteomics journal"] = "Open Proteomics J",
  ["the open psychiatry journal"] = "Open Psychiatr J",
  ["the open psychology journal"] = "Open Psychol J",
  ["the open public health journal"] = "Open Public Health J",
  ["the open rehabilitation journal"] = "Open Rehabil J",
  ["the open reproductive science journal"] = "Open Reprod Sci J",
  ["the open respiratory medicine journal"] = "Open Respir Med J",
  ["the open rheumatology journal"] = "Open Rheumatol J",
  ["the open sleep journal"] = "Open Sleep J",
  ["the open spectroscopy journal"] = "Open Spectrosc J",
  ["the open sports sciences journal"] = "Open Sports Sci J",
  ["the open stem cell journal"] = "Open Stem Cell J",
  ["the open systems biology journal"] = "Open Syst Biol J",
  ["the open tissue engineering and regenerative medicine journal"] = "Open Tissue Eng Regen Med J",
  ["the open toxicology journal"] = "Open Toxicol J",
  ["the open toxinology journal"] = "Open Toxinology J",
  ["the open translational medicine journal"] = "Open Transl Med J",
  ["the open transplantation journal"] = "Open Transplant J",
  ["the open tropical medicine journal"] = "Open Trop Med J",
  ["the open urology & nephrology journal"] = "Open Urol Nephrol J",
  ["the open vaccine journal"] = "Open Vaccine J",
  ["the open veterinary science journal"] = "Open Vet Sci J",
  ["the open virology journal"] = "Open Virol J",
  ["the open women's health journal"] = "Open Womens Health J",
  ["the open zoology journal"] = "Open Zool J",
  ["the opera quarterly"] = "Opera Q",
  ["the optician"] = "Optician",
  ["the optometric weekly"] = "Optom Wkly",
  ["the oral history review"] = "Oral Hist Rev",
  ["the oregon nurse"] = "Oreg Nurse",
  ["the organ"] = "Organ",
  ["the orthodontist"] = "Orthodontist",
  ["the orthopaedic journal at harvard medical school"] = "Orthop J Harv Med Sch",
  ["the orthopedic clinics of north america"] = "Orthop Clin North Am",
  ["the osteopathic hospital"] = "Osteopath Hosp",
  ["the osteopathic magazine"] = "Osteopath Mag",
  ["the osteopathic profession"] = "Osteopath Prof",
  ["the ostrich"] = "Ostrich",
  ["the overseas post-graduate medical journal"] = "Overseas Postgrad Med J",
  ["the p. a. journal"] = "PA J",
  ["the pacific historian"] = "Pac Hist",
  ["the pacific northwesterner"] = "Pac Northwest",
  ["the pakistan dental review"] = "Pak Dent Rev",
  ["the pakistan journal of surgery, gynaecology, and obstetrics"] = "Pak J Surg Gynaecol Obstet",
  ["the pakistan nursing and health review"] = "Pak Nurs Health Rev",
  ["the palawan scientist"] = "Palawan Sci",
  ["the palimpsest"] = "Palimpsest (Iowa City)",
  ["the pamphlets of lewis carroll"] = "Pam. Lewis Carroll",
  ["the pan african medical journal"] = "Pan Afr Med J",
  ["the pan-pacific entomologist"] = "Pan-Pac Entomol",
  ["the patient"] = "Patient",
  ["the patna journal of medicine"] = "Patna J Med",
  ["the pavlovian journal of biological science"] = "Pavlov J Biol Sci",
  ["the pcr reporter"] = "PCR Rep",
  ["the pedagogical seminary and journal of genetic psychology"] = "Pedagog Semin J Genet Psychol",
  ["the pediatric infectious disease journal"] = "Pediatr Infect Dis J",
  ["the penn dental journal"] = "Penn Dent J (Phila)",
  ["the pennsylvania geographer"] = "Pa Geogr",
  ["the pennsylvania hospital bulletin"] = "Pa Hosp Bull",
  ["the pennsylvania magazine of history and biography"] = "Pa Mag Hist Biogr",
  ["the pennsylvania nurse"] = "Pa Nurse",
  ["the pennsylvania pharmacist"] = "Pa Pharm",
  ["the permanente journal"] = "Perm J",
  ["the personalist"] = "Personalist",
  ["the personnel administrator"] = "Pers Adm",
  ["the personnel and guidance journal"] = "Pers Guid J",
  ["the personnel journal"] = "Pers J",
  ["the pezcoller foundation journal : news from the pezcoller foundation world"] = "Pezcoller Found J",
  ["the pharma innovation"] = "Pharma Innov",
  ["the pharmaceutical journal"] = "Pharm J",
  ["the pharmaceutical journal of new zealand"] = "Pharm J N Z",
  ["the pharmacogenomics journal"] = "Pharmacogenomics J",
  ["the pharmacologist"] = "Pharmacologist",
  ["the pharos of alpha omega alpha-honor medical society. alpha omega alpha"] = "Pharos Alpha Omega Alpha Honor Med Soc",
  ["the phi lambda kappa quarterly"] = "Phi Lambda Kappa Q",
  ["the philippine agricultural scientist"] = "Philipp Agric Sci",
  ["the philippine economic journal"] = "Philipp Econ J",
  ["the philippine journal of crop science"] = "Philipp J Crop Sci",
  ["the philippine journal of nursing"] = "Philipp J Nurs",
  ["the philippine journal of pediatrics"] = "Philipp J Pediatr",
  ["the philippine medical world"] = "Philipp Med World",
  ["the philippine review of economics"] = "Philipp Rev Econ",
  ["the philippine review of economics & business"] = "Philipp Rev Econ Bus",
  ["the philosophical forum"] = "Philos Forum",
  ["the philosophical forum: a philosophical quarterly publ. with the cooper. of the dept. of philosophy of boston univ."] = "PF",
  ["the philosophical journal"] = "Philos J",
  ["the philosophical quarterly"] = "Philos Q",
  ["the philosophical review"] = "Philos Rev",
  ["the philosophy forum"] = "Philos Forum",
  ["the philosophy of science in a european perspective"] = "Philos. Sci. Eur. Perspect.",
  ["the physical review"] = "Phys Rev",
  ["the physical therapy review"] = "Phys Ther Rev",
  ["the physician and sportsmedicine"] = "Phys Sportsmed",
  ["the physics of fluids"] = "Phys Fluids",
  ["the physics teacher"] = "Phys Teach",
  ["the physiologist"] = "Physiologist",
  ["the physiology teacher"] = "Physiol Teach",
  ["the physiotherapy review"] = "Physiotherap Rev",
  ["the pilot-evanston hospital"] = "Pilot Evanst Hosp",
  ["the planetary report"] = "Planet Rep",
  ["the planetary science journal"] = "Planet Sci J",
  ["the plant cell"] = "Plant Cell",
  ["the plant genome"] = "Plant Genome",
  ["the plant journal : for cell and molecular biology"] = "Plant J",
  ["the plant pathology journal"] = "Plant Pathol J",
  ["the polar journal"] = "Polar J",
  ["the polar record"] = "Polar Rec (Gr Brit)",
  ["the police chief"] = "Police Chief",
  ["the police journal"] = "Police J",
  ["the polish review"] = "Pol Rev",
  ["the political quarterly"] = "Polit Q",
  ["the pope speaks"] = "Pope Speaks",
  ["the practical lawyer"] = "Pract Lawyer",
  ["the practising midwife"] = "Pract Midwife",
  ["the practitioner"] = "Practitioner",
  ["the prairie rose"] = "Prairie Rose",
  ["the prescriber"] = "Prescriber",
  ["the prevention researcher"] = "Prev Res",
  ["the priest"] = "Priest",
  ["the primary care companion for cns disorders"] = "Prim Care Companion CNS Disord",
  ["the princeton university library chronicle"] = "Princeton Univ Libr Chron",
  ["the prison journal"] = "Prison J",
  ["the probe"] = "Probe (Lond)",
  ["the proceedings of the african classical associations"] = "ProcAfrClAss",
  ["the proceedings of the cardiff medical society"] = "Proc Cardiff Med Soc",
  ["the proceedings of the institute of medicine of chicago"] = "Proc Inst Med Chic",
  ["the proceedings of the institute of statistical mathematics"] = "Proc. Inst. Statist. Math.",
  ["the proceedings of the nutrition society"] = "Proc Nutr Soc",
  ["the proceedings of the pharmaceutical society of egypt"] = "Proc Pharm Soc Egypt",
  ["the proceedings of the south carolina historical association. south carolina historical association"] = "Proc S C Hist Assoc",
  ["the professional animal scientist"] = "Prof Anim Sci",
  ["the professional geographer : the journal of the association of american geographers"] = "Prof Geogr",
  ["the progressive"] = "Progressive",
  ["the progressive nurse"] = "Progress Nurse",
  ["the prostate"] = "Prostate",
  ["the prostate. supplement"] = "Prostate Suppl",
  ["the protein journal"] = "Protein J",
  ["the psychiatric clinics of north america"] = "Psychiatr Clin North Am",
  ["the psychiatric hospital"] = "Psychiatr Hosp",
  ["the psychiatric quarterly"] = "Psychiatr Q",
  ["the psychiatric quarterly. supplement"] = "Psychiatr Q Suppl",
  ["the psychiatric times"] = "Psychiatr Times",
  ["the psychiatrist"] = "Psychiatrist",
  ["the psychoanalytic quarterly"] = "Psychoanal Q",
  ["the psychoanalytic study of the child"] = "Psychoanal Study Child",
  ["the psychohistory review"] = "Psychohist Rev",
  ["the psychological clinic"] = "Psychol Clin",
  ["the psychological record"] = "Psychol Rec",
  ["the psychologist"] = "Psychologist",
  ["the psychologist manager journal"] = "Psychol Manag J",
  ["the psychology of learning and motivation"] = "Psychol Learn Motiv",
  ["the public health laboratory"] = "Public Health Lab",
  ["the public historian"] = "Public Hist",
  ["the public interest"] = "Public Interest",
  ["the public policy and aging report"] = "Public Policy Aging Rep",
  ["the public relations journal"] = "Public Relat J",
  ["the publications of the thoresby society"] = "Publ Thoresby Soc",
  ["the puerto rico journal of public health and tropical medicine"] = "PR J Public Health Trop Med",
  ["the pulse of the montana state nurses' association"] = "Pulse",
  ["the pulse of the montana state nurses’ association"] = "Pulse",
  ["the pump journal of undergraduate research"] = "PUMP J. Undergrad. Res.",
  ["the punjab medical journal"] = "Punjab Med J (Banga)",
  ["the punjab university"] = "Punjab Univ. J. Math. (Lahore)",
  ["the punjab university. journal of mathematics"] = "Punjab Univ. J. Math. (Lahore)",
  ["the purdue pharmacist"] = "Purdue Pharm",
  ["the quality letter for healthcare leaders"] = "Qual Lett Healthc Lead",
  ["the quantitative methods for psychology"] = "Quant Method Psychol",
  ["the quarterly bulletin of the indiana university medical center"] = "Q Bull Indiana Univ Med Cent",
  ["the quarterly journal of administration"] = "Q J Adm",
  ["the quarterly journal of administration / institute of administration, university of ife"] = "Q J Adm",
  ["the quarterly journal of economics"] = "Q J Econ",
  ["the quarterly journal of experimental psychology"] = "Q J Exp Psychol",
  ["the quarterly journal of experimental psychology. a, human experimental psychology"] = "Q J Exp Psychol A",
  ["the quarterly journal of experimental psychology. b, comparative and physiological psychology"] = "Q J Exp Psychol B",
  ["the quarterly journal of finance"] = "Q J Finance",
  ["the quarterly journal of mathematics"] = "Q. J. Math.",
  ["the quarterly journal of mechanics and applied mathematics"] = "Quart. J. Mech. Appl. Math.",
  ["the quarterly journal of medicine"] = "Q J Med",
  ["the quarterly journal of microscopical science"] = "Q J Microsc Sci",
  ["the quarterly journal of nuclear medicine : official publication of the italian association of nuclear medicine (aimn) [and] the international association of radiopharmacology (iar)"] = "Q J Nucl Med",
  ["the quarterly journal of nuclear medicine and molecular imaging : official publication of the italian association of nuclear medicine (aimn) [and] the international association of radiopharmacology (iar), [and] section of the society of radiopharmaceutical chemistry and biology"] = "Q J Nucl Med Mol Imaging",
  ["the quarterly journal of social affairs"] = "Q J Soc Aff",
  ["the quarterly journal of speech"] = "Q J Speech",
  ["the quarterly of the department of antiquities in palestine"] = "QDAP",
  ["the quarterly of the national dental association, inc"] = "Q Natl Dent Assoc",
  ["the quarterly review"] = "Q Rev DC Nurses Assoc",
  ["the quarterly review of biology"] = "Q Rev Biol",
  ["the quarterly review of economics and business"] = "Q Rev Econ Bus",
  ["the quarterly review of economics and finance : journal of the midwest economics association"] = "Q Rev Econ Finance",
  ["the queensland dental journal"] = "Qld Dent J",
  ["the queensland health education council annual report to the honourable the minister for health. queensland health education council"] = "Annu Rep Qld Health Educ Counc Honour Minist Health Home Aff",
  ["the queensland nurse"] = "Qld Nurse",
  ["the r journal"] = "R J",
  ["the raffles bulletin of zoology"] = "Raffles Bull Zool",
  ["the ramanujan journal"] = "Ramanujan J",
  ["the rand journal of economics"] = "Rand J Econ",
  ["the reader's digest"] = "Read Dig",
  ["the reader’s digest"] = "Read Dig",
  ["the reading teacher"] = "Read Teach",
  ["the record of the association of the bar of the city of new york"] = "Rec Assoc Bar City New York",
  ["the recorder of the columbia medical society of richland county, s. c"] = "Rec Columbia Med Soc Richland Cty SC",
  ["the red cross courier"] = "Red Cross Cour",
  ["the reese record"] = "Reese Rec",
  ["the reference librarian"] = "Ref Libr",
  ["the reformed journal"] = "Reform J",
  ["the regan report on nursing law"] = "Regan Rep Nurs Law",
  ["the register of the kentucky historical society"] = "Regist Ky Hist Soc",
  ["the regulatory affairs journal"] = "Regul Aff J",
  ["the research bulletin of the faculty of education"] = "Res. Bull. Fac. Ed. Oita Univ.",
  ["the research file"] = "Res file",
  ["the responsive community : rights and responsibilities"] = "Responsive Community",
  ["the review of black political economy"] = "Rev Black Polit Econ",
  ["the review of communication"] = "Rev Comm",
  ["the review of diabetic studies : rds"] = "Rev Diabet Stud",
  ["the review of economic studies"] = "Rev Econ Stud",
  ["the review of economics and statistics"] = "Rev Econ Stat",
  ["the review of faith & international affairs"] = "Rev Faith Int Aff",
  ["the review of financial studies"] = "Rev Financ Stud",
  ["the review of gastroenterology"] = "Rev Gastroenterol",
  ["the review of higher education"] = "Rev High Ed",
  ["the review of income and wealth"] = "Rev Income Wealth",
  ["the review of metaphysics"] = "Rev Metaphys",
  ["the review of policy research"] = "Rev Policy Res",
  ["the review of politics"] = "Rev Polit",
  ["the review of radical political economics"] = "Rev Radic Polit Econ",
  ["the review of regional studies"] = "Rev Reg Stud",
  ["the review of scientific instruments"] = "Rev Sci Instrum",
  ["the review of symbolic logic"] = "Rev. Symb. Log.",
  ["the review of the polish academy of sciences. polska akademia nauk"] = "Rev Pol Acad Sci",
  ["the rheumatic diseases"] = "Rheum Dis",
  ["the rhodesia nurse"] = "Rhod Nurse",
  ["the rhodesian nurse"] = "Rhod Nurse",
  ["the rocky mountain journal of mathematics"] = "Rocky Mountain J. Math.",
  ["the rocky mountain social science journal"] = "Rocky Mt Soc Sci J",
  ["the rostrum"] = "Rostrum",
  ["the round table"] = "Round Table",
  ["the royal canadian dental corps quarterly"] = "R Can Dent Corps Q",
  ["the royal dental hospital magazine"] = "R Dent Hosp Mag",
  ["the royal institute of public health and hygiene journal"] = "R Inst Public Health Hyg J",
  ["the royal society of london"] = "R. Soc. Lond. Philos. Trans. Ser. A Math. Phys. Eng. Sci.",
  ["the rsa series in transdisciplinary rhetoric"] = "RSA Ser. Transdiscipl. Rhetor.",
  ["the rt-11 technical user’s series"] = "RT-11 Tech. User’s Ser.",
  ["the russell sage foundation journal of the social sciences : rsf"] = "RSF",
  ["the russian review"] = "Russ Rev",
  ["the rutgers journal of bioethics"] = "Rutgers J Bioeth",
  ["the rutgers lectures in philosophy"] = "Rutgers Lect. Philos.",
  ["the ryan advisory for health services governing boards"] = "Ryan Advis Health Serv Gov Boards",
  ["the s.a. journal of continuing medical education = die s.a. tydskrif van voortgesette geneeskundige onderrig"] = "SA J Contin Med Educ",
  ["the saint george's hospital gazette"] = "St Georg Hosp Gaz",
  ["the san diego law review"] = "San Diego Law Rev",
  ["the sanitarian"] = "Sanitarian",
  ["the sanitary review and journal of public health"] = "Sanit Rev J Public Health",
  ["the saudi dental journal"] = "Saudi Dent J",
  ["the saudi journal for dental research"] = "Saudi J Dent Res",
  ["the scalpel"] = "Scalpel",
  ["the scalpel of alpha epsilon delta"] = "Scalpel Alpha Epsil Delta",
  ["the scandinavian economic history review"] = "Scand Econ Hist Rev",
  ["the scandinavian economic history review / [the scandanavian society for economic and social history and historical geography]"] = "Scand Econ Hist Rev",
  ["the scandinavian journal of economics"] = "Scand J Econ",
  ["the school counselor"] = "Sch Couns",
  ["the school psychology digest"] = "Sch Psychol Dig",
  ["the school science review"] = "Sch Sci Rev",
  ["the science of the total environment"] = "Sci Total Environ",
  ["the science reports of kanazawa university"] = "Sci. Rep. Kanazawa Univ.",
  ["the science reports of the kanazawa university"] = "Sci. Rep. Kanazawa Univ.",
  ["the science reports of the research institutes, tohoku university. ser. c, medicine. tohoku daigaku"] = "Sci Rep Res Inst Tohoku Univ [Med]",
  ["the science reports of the research institutes, tohoku university. ser. c, medicine. tōhoku daigaku"] = "Sci Rep Res Inst Tohoku Univ Med",
  ["the science reports of the tohoku university. seventh series: (geography) tohoku daigaku"] = "Sci Rep Tohoku Univ Ser 7",
  ["the science reports of the tôhoku university. seventh series: (geography). tōhoku daigaku"] = "Sci Rep Tohoku Univ Ser 7",
  ["the sciences"] = "Sciences (New York)",
  ["the scientific basis of medicine annual reviews"] = "Sci Basis Med Annu Rev",
  ["the scientific monthly"] = "Sci Mon",
  ["the scientific proceedings of the royal dublin society. royal dublin society"] = "Sci Proc R Dublin Soc",
  ["the scottish historical review"] = "Scott Hist Rev",
  ["the scottish journal of criminal justice studies : the journal of the scottish association for the study of delinquency"] = "Scott J Crim Justice Stud",
  ["the scottish journal of sociology"] = "Scott J Sociol",
  ["the second century: a journal of early christian studies"] = "SCent",
  ["the senior care pharmacist"] = "Sr Care Pharm",
  ["the seoul journal of medicine"] = "Seoul J Med",
  ["the serials librarian"] = "Ser Libr",
  ["the seventeenth century"] = "Seventeenth Century",
  ["the sight-saving review"] = "Sight Sav Rev",
  ["the sightsaving review : official publication of the national society for the prevention of blindness"] = "Sightsav Rev",
  ["the sind medical journal"] = "Sind Med J",
  ["the singapore family physician"] = "Singapore Fam Physician",
  ["the singapore public health bulletin"] = "Singapore Public Health Bull",
  ["the sixteenth century journal"] = "Sixt Century J",
  ["the social and rehabilitation record"] = "Soc Rehabil Rec",
  ["the social history of alcohol and drugs"] = "Soc Hist Alcohol Drugs",
  ["the social history of alcohol review"] = "Soc Hist Alcohol Rev",
  ["the social science journal"] = "Soc Sci J",
  ["the social service review"] = "Soc Serv Rev",
  ["the social work practitioner-researcher"] = "Soc Work Pract Res",
  ["the social worker"] = "Soc Work",
  ["the society for nursing history gazette"] = "Soc Nurs Hist Gaz",
  ["the society for the social history of medicine bulletin"] = "Soc Soc Hist Med Bull (Lond)",
  ["the society of malawi journal. society of malawi (historical and scientific)"] = "Soc Malawi J",
  ["the society of malaŵi journal. society of malawi (historical and scientific)"] = "Soc Malawi J",
  ["the sociological quarterly"] = "Sociol Q",
  ["the sociological review"] = "Sociol Rev",
  ["the south african geographical journal, being a record of the proceedings of the south african geographical society"] = "S Afr Geogr J",
  ["the south african journal of african affairs"] = "S Afr J Afr Aff",
  ["the south african journal of clinical nutrition : sajcn : the official journal of the south african society of parenteral and enteral nutrition and the association for dietetics in south africa"] = "South Afr J Clin Nutr",
  ["the south african journal of communication disorders = die suid-afrikaanse tydskrif vir kommunikasieafwykings"] = "S Afr J Commun Disord",
  ["the south african journal of communication disorders. die suid-afrikaanse tydskrif vir kommunikasieafwykings"] = "S Afr J Commun Disord",
  ["the south african journal of economic history : [journal of the economic history society of southern africa]"] = "S Afr J Econ Hist",
  ["the south african journal of economics. suid-afrikaanse tydskrif vir ekonomie"] = "S Afr J Econ",
  ["the south african journal of medical sciences"] = "S Afr J Med Sci",
  ["the south african journal of physiotherapy"] = "S Afr J Physiother",
  ["the south african journal of psychiatry : sajp : the journal of the society of psychiatrists of south africa"] = "S Afr J Psychiatr",
  ["the south atlantic quarterly"] = "South Atl Q",
  ["the south carolina nurse"] = "S C Nurse",
  ["the south carolina nurse / south carolina nurses’ association"] = "S C Nurse",
  ["the south dakota journal of medicine and pharmacy"] = "SD J Med Pharm",
  ["the south dakota nurse"] = "S D Nurse",
  ["the southeast asian journal of tropical medicine and public health"] = "Southeast Asian J Trop Med Public Health",
  ["the southern african journal of epidemiology & infection : official journal of the sexually transmitted diseases, infectious diseases and epidemiological societies of southern africa"] = "South Afr J Epidemiol Infect",
  ["the southern journal of educational research"] = "South J Educ Res",
  ["the southern journal of philosophy"] = "South J Philos",
  ["the southern medical journal"] = "South Med J",
  ["the southern surgeon"] = "South Surg",
  ["the southwestern entomologist"] = "Southwest Entomol",
  ["the southwestern historical quarterly"] = "Southwest Hist Q",
  ["the southwestern journal of philosophy"] = "Southwest J Philos",
  ["the southwestern naturalist"] = "Southwest Nat",
  ["the southwestern reporter"] = "Southwest Report",
  ["the soviet journal of developmental biology"] = "Sov J Dev Biol",
  ["the soviet journal of ecology"] = "Sov J Ecol",
  ["the soviet review"] = "Sov Rev",
  ["the spanish journal of psychology"] = "Span J Psychol",
  ["the spastic review"] = "Spastic Rev",
  ["the spine journal : official journal of the north american spine society"] = "Spine J",
  ["the sport journal"] = "Sport J",
  ["the sport psychologist"] = "Sport Psychol",
  ["the sports historian"] = "Sports Hist",
  ["the springer series on demographic methods and population analysis"] = "Springer Ser. Demogr. Methods Popul. Anal.",
  ["the springfielder"] = "Springfielder",
  ["the st. luke's hospital gazette"] = "St Lukes Hosp Gaz (Guardamangia)",
  ["the st. luke’s hospital gazette"] = "St Lukes Hosp Gaz (Guardamangia)",
  ["the star"] = "Star",
  ["the stata journal"] = "Stata J",
  ["the state of black america. national urban league"] = "State Black Am",
  ["the state of health care in america"] = "State Health Care Am",
  ["the statistician : journal of the institute of statisticians"] = "Statistician",
  ["the studia philonica annual: studies in hellenistic judaism"] = "StudPhilon",
  ["the subnuclear series"] = "Subnucl. Ser.",
  ["the sudan journal of veterinary research"] = "Sudan J Vet Res",
  ["the summary"] = "Summ Shute Inst",
  ["the sunday times magazine"] = "Sunday Times Magazine",
  ["the supreme court review"] = "Supreme Court Rev",
  ["the surgeon : journal of the royal colleges of surgeons of edinburgh and ireland"] = "Surgeon",
  ["the surgeon’s circular letter. united states. army. far east command. medical section"] = "Surg Circ Lett U S Army Far East Command Med Sect",
  ["the surgical clinics of north america"] = "Surg Clin North Am",
  ["the surgical technologist"] = "Surg Technol",
  ["the swedish cyprus expedition"] = "SCE",
  ["the swedish pioneer historical quarterly"] = "Swed Pioneer Hist Q",
  ["the swedish-american historical quarterly"] = "Swed Am Hist Q",
  ["the swedish-american historical quarterly / swedish pioneer historical society"] = "Swed Am Hist Q",
  ["the sydney law review"] = "Syd Law Rev",
  ["the symposium / the society for developmental biology. society for developmental biology. symposium"] = "Symp Soc Dev Biol",
  ["the synthesis project. research synthesis report"] = "Synth Proj Res Synth Rep",
  ["the systems thinker"] = "Syst Thinker",
  ["the taylor & francis systems and control book series"] = "Taylor & Francis Syst. Control Book Ser.",
  ["the tensor society"] = "Tensor (N.S.)",
  ["the tensor society. tensor. new series"] = "Tensor (N.S.)",
  ["the tenth international symposium on infections in the immunocompromised host, 21-24 june 1998, davos, switzerland. international symposium on infections in the immunocompromised host (10th : 1998 : davos, switzerland)"] = "Tenth Int Symp Infect Immunocompromised Host 21 24 June 1998 Davos Switz Int Symp Infect Immunocompromised Host 10th 1998 Davos Switz",
  ["the texas journal of science"] = "Tex J Sci",
  ["the third ieee international conference on big data security on cloud, ieee bigdatasecurity 2017 : the third ieee international conference on high performance and smart computing, ieee hpsc 2017 ; the second ieee international conference on intelligent data and security, ieee ids 2017 : proceedings : 26-28 may 2017, beijing, china. ieee international conference on big data security on cloud (3rd : 2017 : beijing, china)"] = "Third IEEE Int Conf Big Data Secur Cloud IEEE BigDataSecur 2017 (2017)",
  ["the thomist"] = "Thomist",
  ["the thoracic and cardiovascular surgeon"] = "Thorac Cardiovasc Surg",
  ["the thoracic and cardiovascular surgeon reports"] = "Thorac Cardiovasc Surg Rep",
  ["the times literary supplement"] = "TLS",
  ["the tohoku journal of experimental medicine"] = "Tohoku J Exp Med",
  ["the tohoku mathematical journal. second series"] = "Tohoku Math. J. (2)",
  ["the tokai journal of experimental and clinical medicine"] = "Tokai J Exp Clin Med",
  ["the tokushima journal of experimental medicine"] = "Tokushima J Exp Med",
  ["the torah u-madda journal"] = "Torah U Madda J",
  ["the town planning review"] = "Town Plan Rev",
  ["the toxicologist : an official publication of the society of toxicology"] = "Toxicologist",
  ["the trained nurse and hospital review"] = "Trained Nurse Hosp Rev",
  ["the training school bulletin"] = "Train Sch Bull (Vinel)",
  ["the transactions of the association of industrial medical officers"] = "Trans Assoc Ind Med Off",
  ["the transactions of the hunterian society"] = "Trans Hunter Soc",
  ["the transactions of the society of occupational medicine"] = "Trans Soc Occup Med",
  ["the travelers standard"] = "Travel Stand",
  ["the turkish journal of gastroenterology : the official journal of turkish society of gastroenterology"] = "Turk J Gastroenterol",
  ["the turkish journal of pediatrics"] = "Turk J Pediatr",
  ["the turnbull library record"] = "Turnbull Libr Rec",
  ["the tōhoku mathematical journal"] = "Tōhoku Math. J. (2)",
  ["the ukrainian quarterly"] = "Ukr Q",
  ["the ulster medical journal"] = "Ulster Med J",
  ["the ultrasound journal"] = "Ultrasound J",
  ["the umap journal"] = "UMAP J.",
  ["the undergraduate journal of philosophy"] = "Undergrad J Philos",
  ["the unesco courier"] = "UNESCO Cour",
  ["the university college hospital magazine"] = "Univ Coll Hosp Mag",
  ["the university of chicago law review. university of chicago. law school"] = "Univ Chic Law Rev",
  ["the university of chicago legal forum"] = "Univ Chic Leg Forum",
  ["the university of illinois law forum"] = "Univ Ill Law Forum",
  ["the university of louisville journal of respiratory infections"] = "Univ Louisville J Respir Infect",
  ["the university of manitoba medical journal"] = "Univ Manit Med J",
  ["the university of maryland law forum"] = "Univ Md Law Forum",
  ["the university of memphis law review"] = "Univ Memphis Law Rev",
  ["the university of new south wales law journal"] = "Univ N S W Law J",
  ["the university of queensland law journal"] = "Univ Qld Law J",
  ["the university of tokyo"] = "J. Math. Sci. Univ. Tokyo",
  ["the university of tokyo. journal of mathematical sciences"] = "J. Math. Sci. Univ. Tokyo",
  ["the university of toronto law journal"] = "Univ Tor Law J",
  ["the university series in higher mathematics"] = "Univ. Ser. Higher Math.",
  ["the university series in mathematics"] = "Univ. Ser. Math.",
  ["the urban & social change review"] = "Urban Soc Change Review",
  ["the urban book series"] = "Urban Book Ser.",
  ["the urban lawyer"] = "Urban Lawyer",
  ["the urban league review. national urban league. research dept"] = "Urban Leag Rev",
  ["the urban review"] = "Urban Rev",
  ["the urologic and cutaneous review"] = "Urol Cutaneous Rev",
  ["the urologic clinics of north america"] = "Urol Clin North Am",
  ["the utah women's health review"] = "Utah Womens Health Rev",
  ["the utne reader"] = "Utne Read",
  ["the uttar pradesh state dental journal : an official publication of the state u.p. dental branch of the indian dental association"] = "Uttar Pradesh State Dent J",
  ["the vad journal : the journal of mechanical assisted circulation and heart failure"] = "VAD J",
  ["the veliger"] = "Veliger",
  ["the veterinarian"] = "Veterinarian",
  ["the veterinary bulletin"] = "Vet Bull",
  ["the veterinary clinics of north america"] = "Vet Clin North Am",
  ["the veterinary clinics of north america. equine practice"] = "Vet Clin North Am Equine Pract",
  ["the veterinary clinics of north america. exotic animal practice"] = "Vet Clin North Am Exot Anim Pract",
  ["the veterinary clinics of north america. food animal practice"] = "Vet Clin North Am Food Anim Pract",
  ["the veterinary clinics of north america. large animal practice"] = "Vet Clin North Am Large Anim Pract",
  ["the veterinary clinics of north america. small animal practice"] = "Vet Clin North Am Small Anim Pract",
  ["the veterinary quarterly"] = "Vet Q",
  ["the veterinary record"] = "Vet Rec",
  ["the victorian historical journal"] = "Vic Hist J",
  ["the victorian historical magazine"] = "Vic Hist Mag",
  ["the victorian newsletter"] = "Vic Newsl",
  ["the vikram mathematical journal"] = "Vikram Math. J.",
  ["the virginia explorer"] = "Va Explor",
  ["the virginia magazine of history and biography"] = "Va Mag Hist Biogr",
  ["the virginia quarterly review"] = "Va Q Rev",
  ["the virtual laboratory"] = "Virtual Lab.",
  ["the virtual mentor : vm"] = "Virtual Mentor",
  ["the visual computer"] = "Vis Comput",
  ["the vldb journal : very large data bases : a publication of the vldb endowment"] = "VLDB J",
  ["the volta review"] = "Volta Rev",
  ["the volunteer leader"] = "Volunt Leader",
  ["the walters art gallery bulletin"] = "BWaltersArtGal",
  ["the washington monthly"] = "Wash Mon",
  ["the washington nurse"] = "Wash Nurse",
  ["the washington post"] = "Washington Post",
  ["the washington quarterly"] = "Wash Q",
  ["the washington university medical alumni quarterly"] = "Wash Univ Med Alumni Q",
  ["the washingtonian"] = "Washingtonian",
  ["the weather vane"] = "Weather Vane",
  ["the weekly law reports"] = "Wkly Law Rep",
  ["the welfare bulletin"] = "Welf Bull",
  ["the welsh history review. cylchgrawn hanes cymru"] = "Welsh Hist Rev",
  ["the west african medical journal"] = "West Afr Med J",
  ["the west african medical journal and nigerian medical & dental practitioner"] = "West Afr Med J Niger Med Dent Pract",
  ["the west african medical journal and nigerian practitioner"] = "West Afr Med J Niger Pract",
  ["the west indian medical journal"] = "West Indian Med J",
  ["the west tennessee historical society papers. west tennessee historical society"] = "West Tenn Hist Soc Pap",
  ["the west texas historical association year book. west texas historical association"] = "West Tex Hist Asso Year B",
  ["the west virginia medical journal"] = "W V Med J",
  ["the western australian naturalist"] = "West Aust Nat",
  ["the western canadian journal of anthropology"] = "West Can J Anthropol",
  ["the western druggist"] = "West Drug (Vanc)",
  ["the western historical quarterly"] = "West Hist Q",
  ["the western humanities review"] = "West Hum Rev",
  ["the western journal of black studies"] = "West J Black Stud",
  ["the western journal of emergency medicine"] = "West J Emerg Med",
  ["the western journal of medicine"] = "West J Med",
  ["the western pennsylvania historical magazine"] = "West Pa Hist Mag",
  ["the western political quarterly"] = "West Polit Q",
  ["the william and mary bill of rights journal : a student publication of the marshall-wythe school of law"] = "William Mary Bill Rights J",
  ["the william and mary quarterly"] = "William Mary Q",
  ["the wilson bulletin"] = "Wilson Bull",
  ["the wilson journal of ornithology"] = "Wilson J Ornithol",
  ["the wilson quarterly"] = "Wilson Q",
  ["the wisconsin bar bulletin"] = "Wis Bar Bull",
  ["the wisconsin sociologist"] = "Wisc Sociol",
  ["the wistar institute symposium monograph"] = "Wistar Inst Symp Monogr",
  ["the woman physician"] = "Woman Physician",
  ["the women's annual"] = "Womens Annu",
  ["the women's health data book : a profile of women's health in the united states"] = "Womens Health Data Book",
  ["the women's watch"] = "Womens Watch",
  ["the women’s annual"] = "Womens Annu",
  ["the women’s health data book : a profile of women’s health in the united states / [jacob’s institute of women’s health]"] = "Womens Health Data Book",
  ["the women’s watch / international women’s right action watch"] = "Womens Watch",
  ["the world allergy organization journal"] = "World Allergy Organ J",
  ["the world bank economic review"] = "World Bank Econ Rev",
  ["the world bank research observer"] = "World Bank Res Obs",
  ["the world economy"] = "World Econ",
  ["the world journal of biological psychiatry : the official journal of the world federation of societies of biological psychiatry"] = "World J Biol Psychiatry",
  ["the world journal of men's health"] = "World J Mens Health",
  ["the world of learning"] = "World Learn (Lond)",
  ["the world today"] = "World Today",
  ["the world wide web journal of biology"] = "World Wide Web J Biol",
  ["the worm breeder's gazette"] = "Worm Breed Gaz",
  ["the x-ray technician"] = "Xray Tech",
  ["the xi psi phi quarterly"] = "Xi Psi Phi Q",
  ["the yale journal of biology and medicine"] = "Yale J Biol Med",
  ["the yale journal of criticism"] = "YJC",
  ["the yale law journal"] = "Yale Law J",
  ["the yale leibniz"] = "Yale Leibniz",
  ["the yale university library gazette. yale university. library"] = "Yale Univ Libr Gaz",
  ["the year in immunology"] = "Year Immunol",
  ["the yearbook of the research centre for german and austrian exile studies"] = "Yearb Res Cent Ger Austrian Exile Stud",
  ["the zambia nurse"] = "Zambia Nurse",
  ["the zambia nurse journal"] = "Zambia Nurse J",
  ["the zimbabwe journal of economics"] = "Zimb J Econ",
  ["the zimbabwe nurse"] = "Zimbabwe Nurse",
  ["theatre survey"] = "Theatre Surv",
  ["themes in economics"] = "Themes Econ.",
  ["themes in modern econometrics"] = "Themes Mod. Econom.",
  ["theochem journal of molecular structure"] = "THEOCHEM",
  ["theologiai szemle"] = "Theol Sz",
  ["theological studies"] = "Theol Stud",
  ["theologie und philosophie"] = "Th&Ph",
  ["theologische bibliothek topelmann"] = "Theol. Bibl. Topelmann",
  ["theologische literaturzeitung: monatsschrift für das gesamte gebiet der theologie und religionswissenschaft"] = "ThLZ",
  ["theologische quartalschrift"] = "ThQ",
  ["theologische revue"] = "ThRev",
  ["theologische rundschau"] = "ThRdschau",
  ["theologische zeitschrift"] = "ThZ",
  ["theology digest"] = "Theol Dig",
  ["theology today (princeton, n.j.)"] = "Theol Today",
  ["theoretica chimica acta"] = "Theor. Chim. Acta",
  ["theoretical and applied climatology"] = "Theor. Appl. Climatol.",
  ["theoretical and applied fracture mechanics"] = "Theor. Appl. Fract. Mech.",
  ["theoretical and applied genetics"] = "Theor. Appl. Genet.",
  ["theoretical and applied mechanics"] = "Theoret. Appl. Mech.",
  ["theoretical and applied mechanics letters"] = "Theor. Appl. Mech. Lett.",
  ["theoretical and computational fluid dynamics"] = "Theor. Comput. Fluid Dyn.",
  ["theoretical and experimental chemistry"] = "Theor. Exp. Chem.",
  ["theoretical and experimental plant physiology"] = "Theor Exp Plant Physiol",
  ["theoretical and mathematical physics"] = "Theor. Math. Phys.",
  ["theoretical biology"] = "Theor. Biol.",
  ["theoretical biology & medical modelling"] = "Theor Biol Med Model",
  ["theoretical biology and medical modelling"] = "Theor. Biol. Med. Modell.",
  ["theoretical biology forum"] = "Theor Biol Forum",
  ["theoretical chemistry accounts"] = "Theor. Chem. Acc.",
  ["theoretical computer science"] = "Theoret. Comput. Sci.",
  ["theoretical criminology"] = "Theor Criminol",
  ["theoretical ecology"] = "Theor Ecol",
  ["theoretical economics"] = "Theor. Econ.",
  ["theoretical foundations of chemical engineering"] = "Theor. Found. Chem. Eng.",
  ["theoretical issues in ergonomics science"] = "Theor Issues Ergon Sci",
  ["theoretical linguistics"] = "Theor Linguist",
  ["theoretical medicine"] = "Theor Med",
  ["theoretical medicine and bioethics"] = "Theor Med Bioeth",
  ["theoretical population biology"] = "Theor Popul Biol",
  ["theoretical surgery"] = "Theor Surg",
  ["theoria et historia scientiarum"] = "Theoria Hist. Sci.",
  ["theoria. a swedish journal of philosophy"] = "Theoria",
  ["theoria. revista de teoria, historia y fundamentos de la ciencia. segunda epoca. an international journal for theory, history and foundations of science. second series"] = "Theoria (San Sebastian) (2)",
  ["theoria: a swedish journal of philosophy and psychology"] = "Theoria",
  ["theorie und praxis der korperkultur"] = "Theor Prax Korperkult",
  ["theorie und praxis der körperkultur"] = "Theor Prax Korperkult",
  ["theory & psychology"] = "Theory Psychol",
  ["theory and applications of categories"] = "Theory Appl. Categ.",
  ["theory and applications of computability"] = "Theory Appl. Comput.",
  ["theory and applications of graphs"] = "Theory Appl. Graphs",
  ["theory and applications of mathematics & computer science"] = "Theory Appl. Math. Comput. Sci.",
  ["theory and applications of natural language processing"] = "Theory Appl. Nat. Lang. Process.",
  ["theory and applications of transport in porous media"] = "Theory Appl. Transp. Porous Media",
  ["theory and decision"] = "Theory Decis",
  ["theory and decision library"] = "Theory Decis. Lib. Ser. B Math. Statist. Methods",
  ["theory and decision library c. game theory, social choice, decision theory, and optimization"] = "Theory Decis. Libr. C, Game Theory Soc. Choice Decis. Theory Optim.",
  ["theory and decision library. series c: game theory, mathematical programming and operations research"] = "Theory Decis. Libr. Ser. C Game Theory Math. Program. Oper. Res.",
  ["theory and decision. an international journal for multidisciplinary advances in decision science"] = "Theory and Decision",
  ["theory and engineering applications of computational methods"] = "Theory Eng. Appl. Comput. Methods",
  ["theory and practice of logic programming"] = "Theory Pract. Log. Program.",
  ["theory and research in education : tre"] = "Theory Res Educ",
  ["theory and research in social education"] = "Theory Res Soc Educ",
  ["theory and society"] = "Theory Soc",
  ["theory in biosciences"] = "Theory Biosci.",
  ["theory in biosciences = theorie in den biowissenschaften"] = "Theory Biosci",
  ["theory into practice"] = "Theory Pract",
  ["theory of computing systems"] = "Theory Comput. Syst.",
  ["theory of computing. an open access journal"] = "Theory Comput.",
  ["theory of probability and its applications"] = "Theory Probab. Appl.",
  ["theory of probability and mathematical statistics"] = "Theory Probab. Math. Statist.",
  ["theory of stochastic processes"] = "Theory Stoch. Process.",
  ["theory, culture & society"] = "Theory Cult Soc",
  ["therapeutic advances in cardiovascular disease"] = "Ther Adv Cardiovasc Dis",
  ["therapeutic advances in chronic disease"] = "Ther Adv Chronic Dis",
  ["therapeutic advances in drug safety"] = "Ther Adv Drug Saf",
  ["therapeutic advances in endocrinology and metabolism"] = "Ther Adv Endocrinol Metab",
  ["therapeutic advances in gastroenterology"] = "Therap Adv Gastroenterol",
  ["therapeutic advances in gastrointestinal endoscopy"] = "Ther Adv Gastrointest Endosc",
  ["therapeutic advances in hematology"] = "Ther Adv Hematol",
  ["therapeutic advances in infectious disease"] = "Ther Adv Infect Dis",
  ["therapeutic advances in medical oncology"] = "Ther Adv Med Oncol",
  ["therapeutic advances in musculoskeletal disease"] = "Ther Adv Musculoskelet Dis",
  ["therapeutic advances in neurological disorders"] = "Ther Adv Neurol Disord",
  ["therapeutic advances in ophthalmology"] = "Ther Adv Ophthalmol",
  ["therapeutic advances in psychopharmacology"] = "Ther Adv Psychopharmacol",
  ["therapeutic advances in reproductive health"] = "Ther Adv Reprod Health",
  ["therapeutic advances in respiratory disease"] = "Ther Adv Respir Dis",
  ["therapeutic advances in urology"] = "Ther Adv Urol",
  ["therapeutic advances in vaccines"] = "Ther Adv Vaccines",
  ["therapeutic advances in vaccines and immunotherapy"] = "Ther Adv Vaccines Immunother",
  ["therapeutic apheresis"] = "Ther. Apher.",
  ["therapeutic apheresis : official journal of the international society for apheresis and the japanese society for apheresis"] = "Ther Apher",
  ["therapeutic apheresis and dialysis"] = "Ther. Apher. Dial.",
  ["therapeutic apheresis and dialysis : official peer-reviewed journal of the international society for apheresis, the japanese society for apheresis, the japanese society for dialysis therapy"] = "Ther Apher Dial",
  ["therapeutic communities"] = "Ther Communities",
  ["therapeutic delivery"] = "Ther Deliv",
  ["therapeutic drug monitoring"] = "Ther Drug Monit",
  ["therapeutic hypothermia and temperature management"] = "Ther Hypothermia Temp Manag",
  ["therapeutic immunology"] = "Ther Immunol",
  ["therapeutic innovation & regulatory science"] = "Ther Innov Regul Sci",
  ["therapeutic notes"] = "Ther Notes",
  ["therapeutic patient education"] = "Ther Patient Educ",
  ["therapeutic problems of today"] = "Ther Probl Today",
  ["therapeutic radiology and oncology"] = "Ther Radiol Oncol",
  ["therapeutic recreation journal"] = "Ther Recreation J",
  ["therapeutic targets for neurological diseases"] = "Ther Targets Neurol Dis",
  ["therapeutica nova"] = "Ther Nova",
  ["therapeutics and clinical risk management"] = "Ther Clin Risk Manag",
  ["therapeutikon (karlsruhe, germany)"] = "Therapeutikon",
  ["therapeutique (la semaine des hopitaux)"] = "Therapeutique",
  ["therapeutique (la semaine des hôpitaux)"] = "Therapeutique",
  ["therapeutique / entretiens de bichat pitie-salpetriere. entretiens de bichat pitie-salpetriere"] = "Entret Bichat Pitie Salpetriere Ther",
  ["therapeutische berichte"] = "Ther Ber",
  ["therapeutische umschau"] = "Ther. Umsch.",
  ["therapeutische umschau und medizinische bibliographie. revue therapeutique et bibliographie medicale"] = "Ther Umsch Med Bibliogr",
  ["therapeutische umschau und medizinische bibliographie. revue thérapeutique et bibliographie médicale"] = "Ther Umsch Med Bibliogr",
  ["therapeutische umschau. revue therapeutique"] = "Ther Umsch",
  ["therapeutische umschau. revue thérapeutique"] = "Ther Umsch",
  ["therapia hungarica"] = "Ther Hung",
  ["therapia hungarica (english edition)"] = "Ther Hung",
  ["therapie der gegenwart"] = "Ther Ggw",
  ["therapy (london, england : 2004)"] = "Therapy",
  ["thermal engineering"] = "Therm. Eng.",
  ["thermal science and engineering progress"] = "Therm. Sci. Eng. Prog.",
  ["thermochimica acta"] = "Thermochim. Acta",
  ["thermophysics and aeromechanics"] = "Thermophys. Aeromech.",
  ["thesaurus cultus et rituum antiquorum"] = "ThesCRA",
  ["theta series in advanced mathematics"] = "Theta Ser. Adv. Math.",
  ["thérapeutique. entretiens de bichat pitié-salpêtrière"] = "Entret Bichat Pitie Salpetriere Ther",
  ["thin solid films"] = "Thin Solid Films",
  ["thin walled structures"] = "Thin Walled Struct.",
  ["thin-walled structures"] = "Thin-Walled Struct.",
  ["think (london, england)"] = "Think (Lond)",
  ["thinking & reasoning"] = "Think Reason",
  ["thinking skills and creativity"] = "Think Skills Creat",
  ["third republic = troisième république"] = "Third Repub",
  ["third world planning review"] = "Third World Plann Rev",
  ["third world quarterly"] = "Third World Q",
  ["thisweek (lagos, nigeria)"] = "ThisWeek",
  ["thomas jefferson law review"] = "Thomas Jefferson Law Rev",
  ["thoracic and cardiovascular surgeon"] = "Thorac. Cardiovasc. Surg.",
  ["thoracic cancer"] = "Thorac Cancer",
  ["thoracic surgery clinics"] = "Thorac Surg Clin",
  ["thoracic surgical science"] = "Thorac Surg Sci",
  ["thoraxchirurgie und vaskulare chirurgie"] = "Thoraxchir Vask Chir",
  ["thoraxchirurgie und vaskuläre chirurgie"] = "Thoraxchir Vask Chir",
  ["thoraxchirurgie vaskulare chirurgie"] = "Thoraxchir. Vask. Chir.",
  ["thoraxchirurgie, vaskulare chirurgie"] = "Thoraxchir Vask Chir",
  ["thoraxchirurgie, vaskuläre chirurgie"] = "Thoraxchir Vask Chir",
  ["thought (hoboken, n.j.)"] = "Thought (Hoboken)",
  ["thrombosis and haemostasis"] = "Thromb Haemost",
  ["thrombosis et diathesis haemorrhagica"] = "Thromb Diath Haemorrh",
  ["thrombosis et diathesis haemorrhagica. supplementum"] = "Thromb Diath Haemorrh Suppl",
  ["thrombosis journal"] = "Thromb J",
  ["thrombosis research"] = "Thromb Res",
  ["thrombosis research. supplement"] = "Thromb Res Suppl",
  ["thurgauer jahrbuch"] = "Thurg. Jahrb.",
  ["thurgood marshall law review"] = "Thurgood Marshall Law Rev",
  ["thymos (harriman, tenn.)"] = "Thymos (Harriman)",
  ["thyroid : official journal of the american thyroid association"] = "Thyroid",
  ["thyroid research"] = "Thyroid Res",
  ["thyroidology / a.p.r.i.m"] = "Thyroidology",
  ["tianjin da xue xue bao"] = "Journal of Tianjin University|Tianjin Daxue Xuebao",
  ["tianjin da xue xue bao = journal of tianjin university"] = "Tianjin Daxue Xuebao",
  ["tianjin university"] = "Trans. Tianjin Univ.",
  ["tianjin yi ke da xue xue bao"] = "Journal of Tianjin Medical University|Tianjin Yi Ke Da Xue Xue Bao",
  ["tianjin yi ke da xue xue bao = journal of tianjin medical university"] = "Tianjin Yi Ke Da Xue Xue Bao",
  ["tianjin yi yao = tianjin medical journal"] = "Tianjin Yi Yao",
  ["tianran chanwu yanjiu yu kaifa"] = "Tianran Chanwu Yanjiu Yu Kaifa",
  ["tibet news review"] = "Tibet News Rev",
  ["tibetan buddhist literature and praxis: studies in its formative period, 900-1400"] = "Brills. Tibet. Stu. Lib.",
  ["tibiscus. istorie, arheologie. muzeul banatului timişoara"] = "Tibiscus",
  ["ticks and tick-borne diseases"] = "Ticks Tick Borne Dis",
  ["tid [reports]. u.s. atomic energy commission"] = "TID Rep",
  ["tid reports"] = "TID Rep.",
  ["tidens tann"] = "Tid Tann",
  ["tidning. sveriges tandlakarforbund"] = "Sver Tandlakarforb Tidn",
  ["tidning. sveriges tandläkarförbund"] = "Sver Tandlakarforb Tidn",
  ["tidskrift for sjukvardspedagoger"] = "Tidskr. Sjukvardspedagog.",
  ["tidskrift for sveriges sjukskoterskor"] = "Tidskr. Sver. Sjukskot.",
  ["tidskrift för sjukvårdspedagoger"] = "Tidskr Sjukvardspedagog",
  ["tidskrift för sveriges sjuksköterskor"] = "Tidskr Sver Sjukskot",
  ["tidskrift i militar halsovard"] = "Tidskr Mil Halsov",
  ["tidskrift i militär hälsovård"] = "Tidskr Mil Halsov",
  ["tidsskrift for den norske laegeforening"] = "Tidsskr. Nor. Laegeforen.",
  ["tidsskrift for den norske lægeforening : tidsskrift for praktisk medicin, ny række"] = "Tidsskr Nor Laegeforen",
  ["tidsskrift for kjemi, bergvesen, og metallurgi"] = "Tidsskr. Kjemi Bergv. Metall",
  ["tidsskrift for praktiserende tandlaeger"] = "Tidsskr Prakt Tandlaeg",
  ["tidsskrift for rettsvidenskap"] = "Tidsskr Rettsvitenskap",
  ["tidsskrift for samfunnsforskning"] = "Tidsskr Samfunnsforsk",
  ["tidsskrift for skogbruk"] = "Tidsskr. Skogbruk",
  ["tidsskrift for sygeplejersker"] = "Tidsskr Sygepl",
  ["tidsskrift for tandlaeger"] = "Tidsskr. Tandlaeger",
  ["tidsskrift for tandlaeger (copenhagen, denmark : 1981)"] = "Tidsskr Tandlaeger",
  ["tidsskriftet sykepleien"] = "Tidsskr Sykepl",
  ["tierarztliche praxis"] = "Tierarztl. Prax.",
  ["tierarztliche praxis ausgabe g grosstiere nutztiere"] = "Tierarztl. Prax. Ausg. G Grosstiere Nutztiere",
  ["tierarztliche praxis ausgabe k kleintiere heimtiere"] = "Tierarztl. Prax. Ausg. K Kleintiere Heimtiere",
  ["tierarztliche praxis. ausgabe g, grosstiere nutztiere"] = "Tierarztl Prax Ausg G Grosstiere Nutztiere",
  ["tierarztliche praxis. ausgabe g, grosstiere/nutztiere"] = "Tierarztl. Prax. Ausg. G Grosstiere Nutztiere",
  ["tierarztliche praxis. ausgabe k, kleintiere/heimtiere"] = "Tierarztl Prax Ausg K Klientiere Heimtiere",
  ["tierarztliche praxis. supplement"] = "Tierarztl Prax Suppl",
  ["tierarztliche umschau"] = "Tierarztl. Umsch.",
  ["tierärztliche praxis"] = "Tierarztl Prax",
  ["tierärztliche praxis. ausgabe g, grosstiere/nutztiere"] = "Tierarztl Prax Ausg G Grosstiere Nutztiere",
  ["tierärztliche praxis. ausgabe k, kleintiere/heimtiere"] = "Tierarztl Prax Ausg K Kleintiere Heimtiere",
  ["tierärztliche praxis. supplement"] = "Tierarztl Prax Suppl",
  ["tierärztliche umschau"] = "Tierarztl Umsch",
  ["tierras de león"] = "TiLeon",
  ["tijdschrift voor bedrijfs- en verzekeringsgeneeskunde"] = "Tijdschr Bedr Verzekeringsgeneeskd",
  ["tijdschrift voor bejaarden- kraam- en ziekenverzorging"] = "Tijdschr. Bejaarden. Kraam. Ziekenverzorging",
  ["tijdschrift voor bejaarden-, kraam- en ziekenversorging : bkz"] = "Tijdschr Bejaarden Kraam Ziekenverzorging",
  ["tijdschrift voor de geschiedenis der geneeskunde, natuurwetenschappen, wiskunde en techniek"] = "Tijdschr Geschied Geneeskd Natuurwet Wiskd Tech",
  ["tijdschrift voor diergeneeskunde"] = "Tijdschr Diergeneeskd",
  ["tijdschrift voor economie"] = "Tijdschr. Econ.",
  ["tijdschrift voor economie en management"] = "Tijdschr. Econ. Manage.",
  ["tijdschrift voor economische en sociale geografie = journal of economic and social geography = revue de géographie économique et humaine = zeitschrift für ökonomische und soziale geographie = revista de geografía económica y social"] = "Tijdschr Econ Soc Geogr",
  ["tijdschrift voor entomologie"] = "Tijdschr Entomol",
  ["tijdschrift voor filosofie"] = "TF",
  ["tijdschrift voor gastro-enterologie"] = "Tijdschr Gastroenterol",
  ["tijdschrift voor geneeskunde"] = "Tijdschr Geneeskd",
  ["tijdschrift voor gerontologie en geriatrie"] = "Tijdschr Gerontol Geriatr",
  ["tijdschrift voor geschiedenis"] = "TG",
  ["tijdschrift voor geschiedenis (1920)"] = "Tijdschr Geschied",
  ["tijdschrift voor geschiedenis (1920)."] = "Tijdschr Geschied",
  ["tijdschrift voor gezondheidszorg & ethiek : tge"] = "Tijdschr Gezondh Ethiek",
  ["tijdschrift voor kindergeneeskunde"] = "Tijdschr Kindergeneeskd",
  ["tijdschrift voor mediterrane archeologie"] = "TMA",
  ["tijdschrift voor nu[c]leaire geneeskunde"] = "Tijdschr Nucl Geneeskd",
  ["tijdschrift voor philosophie"] = "Tijdschr Filos",
  ["tijdschrift voor psychiatrie"] = "Tijdschr Psychiatr",
  ["tijdschrift voor sociale geneeskunde"] = "Tijdschr Soc Geneeskd",
  ["tijdschrift voor sociale geschiedenis"] = "Tijdschr Soc Geschied",
  ["tijdschrift voor sociale gezondheidszorg : tsg : 14-daags blad van de algemene nederlandse vereniging voor sociale gezondheidszorg"] = "Tijdschr Soc Gezondheidsz",
  ["tijdschrift voor sociale wetenschappen"] = "Tijdschr Soc Wet",
  ["tijdschrift voor sociologie"] = "Tijdschr Sociol",
  ["tijdschrift voor tandheelkunde"] = "Tijdschr Tandheelkd",
  ["tijdschrift voor verzorgenden"] = "Tijdschr Verzorg",
  ["tijdschrift voor ziekenverpleging"] = "Tijdschr Ziekenverpl",
  ["tijdschrift vor klei, glas en keramiek"] = "Tijd. Klei Glas Keram.",
  ["tijeretazos sobre malaria"] = "Tijeretazos Malar",
  ["tile and till"] = "Tile Till",
  ["timarit hjukrunarfelags islands"] = "Timarit Hjukrunarfel. Isl.",
  ["timber bulletin for europe"] = "Timber Bull. Eur.",
  ["time & society"] = "Time Soc",
  ["timeline (columbus, ohio)"] = "Timeline",
  ["timely topics in medicine. cardiovascular diseases"] = "Timely Top Med Cardiovasc Dis",
  ["times (london, england : 1788)"] = "Times (Lond)",
  ["times / tennessee hospital association"] = "Times",
  ["timing & time perception (leiden, netherlands)"] = "Timing Time Percept",
  ["timisoara medicala"] = "Timisoara Med",
  ["timişoara medicală"] = "Timisoara Med",
  ["tip fakultesi mecmuasi"] = "Tip Fak Mecm",
  ["tip fakultesi mecmuasi, istanbul universitesi"] = "Tip Fak. Mecm.",
  ["tip fakültesi mecmuasi"] = "Tip Fak Mecm",
  ["tire science & technology"] = "Tire Sci Technol",
  ["tire science and technology"] = "Tire Sci. Technol.",
  ["tiroler heimat"] = "Tirol Heim",
  ["tiroler heimatblatter : zeitschrift fur geschichte, natur und volkskunde / herausgegeben vom verein fur heimatschutz und heimatpflege in tirol"] = "Tirol Heimatbl",
  ["tiroler heimatblätter : zeitschrift für geschichte, natur und volkskunde"] = "Tirol Heimatbl",
  ["tiryns. forschungen und berichte"] = "Tiryns",
  ["tissue & cell"] = "Tissue Cell",
  ["tissue and cell"] = "Tissue Cell",
  ["tissue antigens"] = "Tissue Antigens",
  ["tissue barriers"] = "Tissue Barriers",
  ["tissue engineering"] = "Tissue Eng.",
  ["tissue engineering and regenerative medicine"] = "Tissue Eng Regen Med",
  ["tissue engineering part a"] = "Tissue Eng. Part A",
  ["tissue engineering part b"] = "Tissue Eng. Part B",
  ["tissue engineering part c"] = "Tissue Eng. Part C",
  ["tissue engineering. part a"] = "Tissue Eng Part A",
  ["tissue engineering. part b, reviews"] = "Tissue Eng Part B Rev",
  ["tissue engineering. part c, methods"] = "Tissue Eng Part C Methods",
  ["tit journal of life sciences"] = "TIT J. Life Sci.",
  ["tímarit hjúkrunarfélags islands"] = "Timarit Hjukrunarfel Isl",
  ["tm - technisches messen"] = "Tm - Tech. Mess.",
  ["tmj update"] = "TMJ Update",
  ["to hellēniko periodiko gia stomatikē & gnathoprosōpikē cheirourgikē"] = "Hell Period Stomat Gnathopathoprosopike Cheir",
  ["tobacco control"] = "Tob Control",
  ["tobacco induced diseases"] = "Tob Induc Dis",
  ["tobacco prevention & cessation"] = "Tob Prev Cessat",
  ["tobacco regulatory science"] = "Tob Regul Sci",
  ["tobacco use insights"] = "Tob Use Insights",
  ["today's catholic teacher"] = "Todays Cathol Teach",
  ["today's christian doctor : the journal of the christian medical & dental society"] = "Todays Christ Dr",
  ["today's fda : official monthly journal of the florida dental association"] = "Todays FDA",
  ["today's health"] = "Todays Health",
  ["today's nursing home"] = "Todays Nurs Home",
  ["today's or nurse"] = "Todays OR Nurse",
  ["today's surgical nurse"] = "Todays Surg Nurse",
  ["todays fda"] = "Todays FDA",
  ["todays nursing home"] = "Todays Nurs. Home",
  ["today’s catholic teacher"] = "Todays Cathol Teach",
  ["today’s christian doctor : the journal of the christian medical & dental society"] = "Todays Christ Dr",
  ["today’s engineer"] = "Today’s Engineer",
  ["today’s fda : official monthly journal of the florida dental association"] = "Todays FDA",
  ["today’s health"] = "Todays Health",
  ["today’s nursing home"] = "Todays Nurs Home",
  ["today’s or nurse"] = "Todays OR Nurse",
  ["today’s surgical nurse"] = "Todays Surg. Nurse",
  ["todo social"] = "Todo Soc",
  ["toho igakkai zasshi. journal of the medical society of toho university"] = "Toho Igakkai Zasshi",
  ["tohoku igaku zasshi"] = "Tohoku Igaku Zasshi",
  ["tohoku journal of experimental medicine"] = "Tohoku J. Exp. Med.",
  ["tohoku mathematical journal"] = "Tohoku Math. J.",
  ["tohoku mathematical publications"] = "Tohoku Math. Publ.",
  ["tohoku shika daigaku gakkai shi"] = "Tohoku Shika Daigaku Gakkai Shi",
  ["tokai journal of experimental and clinical medicine"] = "Tokai J. Exp. Clin. Med.",
  ["toko-ginecologia practica"] = "Tokoginecol. Pract.",
  ["toko-ginecología práctica"] = "Tokoginecol Pract",
  ["tokushima journal of experimental medicine"] = "Tokushima J. Exp. Med.",
  ["tokyo igaku zasshi. the tokyo journal of medical sciences"] = "Tokyo Igaku Zasshi",
  ["tokyo igaku. the tokyo journal of medical sciences"] = "Tokyo Igaku",
  ["tokyo ika daigaku zasshi (journal of tokyo medical college)"] = "Tokyo Ika Daigaku Zasshi",
  ["tokyo ika shika daigaku iyo kizai kenkyusho hokoku (reports of the institute for medical and dental engineering, tokyo medical and dental university)"] = "Tokyo Ika Shika Daigaku Iyo Kizai Kenkyusho Hokoku",
  ["tokyo journal of mathematics"] = "Tokyo J. Math.",
  ["tomography (ann arbor, mich.)"] = "Tomography",
  ["tomography of materials and structures"] = "Tomogr. Mater. Struct",
  ["tomorrow's technology transfer"] = "Tomorrows Technol Transf",
  ["tonan ajia kenkyu"] = "Tonan Ajia Kenkyu",
  ["tong ji yi ke da xue xue bao = acta universitatis medicinae tongji"] = "Tong Ji Yi Ke Da Xue Xue Bao",
  ["tonga munhwa"] = "Tonga Munhwa",
  ["tongji daxue xuebao"] = "Tongji Daxue Xuebao Ziran Kexue Ban",
  ["tongmul hakhoe chi"] = "Tongmul Hakhoe Chi",
  ["tonkie khimicheskie tekhnologii"] = "Tonkie Khim. Tekhnol.",
  ["tools and algorithms for the construction and analysis of systems : 26th international conference, tacas 2020, held as part of the european joint conferences on theory and practice of software, etaps 2020, dublin, ireland, april 25-30, 2020, proceedings. part i. tacas (conference) (26th : 2020 : dublin, ireland)"] = "Tools Algorithms Constr Anal Syst I (2020)",
  ["top. an official journal of the spanish society of statistics and operations research"] = "TOP",
  ["top. transactions in operations research"] = "TOP",
  ["topical problems of psychotherapy"] = "Top Probl Psychother",
  ["topics in advanced geoinformatics"] = "Top. Adv. Geoinformatics",
  ["topics in anti-cancer research"] = "Top Anticancer Res",
  ["topics in antiviral medicine"] = "Top Antivir Med",
  ["topics in applied physics"] = "Top. Appl. Phys.",
  ["topics in biodiversity and conservation"] = "Top. Biodivers. Conserv.",
  ["topics in biological and inorganic chemistry"] = "Top. Biol. Inorg. Chem.",
  ["topics in catalysis"] = "Top. Catal.",
  ["topics in clinical nursing"] = "Top Clin Nurs",
  ["topics in clinical nutrition"] = "Top Clin Nutr",
  ["topics in cognitive science"] = "Top Cogn Sci",
  ["topics in companion animal medicine"] = "Top Companion Anim Med",
  ["topics in computer mathematics"] = "Topics in Comput. Math.",
  ["topics in current chemistry"] = "Top. Curr. Chem.",
  ["topics in current chemistry (cham)"] = "Top Curr Chem (Cham)",
  ["topics in current genetics"] = "Top Curr Genet",
  ["topics in discrete mathematics"] = "Topics Discrete Math.",
  ["topics in early childhood special education"] = "Topics Early Child Spec Educ",
  ["topics in economic analysis & policy"] = "Top Econ Anal Policy",
  ["topics in emergency medicine"] = "Top Emerg Med",
  ["topics in engineering"] = "Topics Eng.",
  ["topics in fluorescence spectroscopy"] = "Top. Fluoresc. Spectrosc.",
  ["topics in geriatric rehabilitation"] = "Top Geriatr Rehabil",
  ["topics in health care financing"] = "Top Health Care Financ",
  ["topics in health information management"] = "Top Health Inf Manage",
  ["topics in health record management"] = "Top Health Rec Manage",
  ["topics in heterocyclic chemistry"] = "Top Heterocycl Chem",
  ["topics in hiv medicine : a publication of the international aids society, usa"] = "Top HIV Med",
  ["topics in hospital law"] = "Top Hosp Law",
  ["topics in hospital pharmacy management"] = "Top Hosp Pharm Manage",
  ["topics in hospital pharmacy management / aspen systems corporation"] = "Top Hosp Pharm Manage",
  ["topics in inorganic chemistry"] = "Top. Inorg. Chem.",
  ["topics in intelligent engineering and informatics"] = "Top. Intell. Eng. Inform.",
  ["topics in language disorders"] = "Top Lang Disord",
  ["topics in magnetic resonance imaging"] = "Top. Magn. Reson. Imaging",
  ["topics in magnetic resonance imaging : tmri"] = "Top Magn Reson Imaging",
  ["topics in organometallic chemistry"] = "Top. Organomet. Chem.",
  ["topics in photographic preservation"] = "Topics Photogr Preserv",
  ["topics in physical chemistry"] = "Top. Phys. Chem.",
  ["topics in safety, risk, reliability and quality"] = "Top. Saf. Risk Reliab. Qual.",
  ["topics in spinal cord injury rehabilitation"] = "Top Spinal Cord Inj Rehabil",
  ["topics in stereochemistry"] = "Top Stereochem",
  ["topics in stroke rehabilitation"] = "Top Stroke Rehabil",
  ["toplum ve bilim"] = "Toplum Bilim",
  ["topoi : an international review of philosophy"] = "Topoi (Dordr)",
  ["topoi. an international review of philosophy"] = "Topoi",
  ["topoi: an international review of philosophy"] = "Topoi(Dordrecht)",
  ["topoi: orient-occident"] = "Topoi(Lyon)",
  ["topological algebra and its applications"] = "Topol. Algebra Appl.",
  ["topological methods in nonlinear analysis"] = "Topol. Methods Nonlinear Anal.",
  ["topology and its applications"] = "Topology Appl.",
  ["topology proceedings"] = "Topology Proc.",
  ["toronto journal of theology"] = "Tor J Theol",
  ["toronto star"] = "Tor Star",
  ["torreon medico"] = "Torreon Med",
  ["torreon médico"] = "Torreon Med",
  ["tort & insurance law journal"] = "Tort Insur Law J",
  ["tort and insurance law journal"] = "Tort Insur. Law J.",
  ["tort trial & insurance practice law journal"] = "Tort Trial Insur Pract Law J",
  ["tortenelmi szemle"] = "Tortenelmi Sz",
  ["torteneti statisztikai tanulmanyok / kozponti statisztikai hivatal konyvtara, magyar orszagos leveltar"] = "Tort Stat Tanulm",
  ["torture : quarterly journal on rehabilitation of torture victims and prevention of torture"] = "Torture",
  ["total patient care in hiv & hcv"] = "Total Patient Care HIV HCV",
  ["tottel's journal of professional negligence"] = "Tottels J Prof Neglig",
  ["tottori rinshō kagaku kenkyūkai shi"] = "Tottori Rinsho Kagaku Kenkyukai Shi",
  ["touchstone (nashville, tenn.)"] = "Touchstone (Nashv)",
  ["toulouse medical"] = "Toulouse Med",
  ["toulouse médical"] = "Toulouse Med",
  ["tourism and hospitality management"] = "Tourism Hospitality Manage.",
  ["tourism economics"] = "Tourism Econ.",
  ["tourism management"] = "Tour Manag",
  ["tourism management perspectives"] = "Tour Manag Perspect",
  ["tourist studies"] = "Tour Stud",
  ["touro law review"] = "Touro Law Rev",
  ["town and country planning : the quarterly review of the town and country planning association"] = "Town Ctry Plann",
  ["town hall journal"] = "Town Hall J",
  ["toxic substance mechanisms"] = "Toxic Subst Mech",
  ["toxicity report series"] = "Toxic Rep Ser",
  ["toxicologic pathology"] = "Toxicol Pathol",
  ["toxicological & environmental chemistry"] = "Toxicol. Environ. Chem.",
  ["toxicological and environmental chemistry"] = "Toxicol. Environ. Chem.",
  ["toxicological european research"] = "Toxicol. Eur. Res.",
  ["toxicological european research. recherche europeenne en toxicologie"] = "Toxicol Eur Res",
  ["toxicological european research. recherche européenne en toxicologie"] = "Toxicol Eur Res",
  ["toxicological research"] = "Toxicol Res",
  ["toxicological reviews"] = "Toxicol Rev",
  ["toxicological sciences"] = "Toxicol. Sci.",
  ["toxicological sciences : an official journal of the society of toxicology"] = "Toxicol Sci",
  ["toxicologie analytique et clinique"] = "Toxicol. Anal. Clin.",
  ["toxicology and applied pharmacology"] = "Toxicol Appl Pharmacol",
  ["toxicology and environmental health sciences"] = "Toxicol Environ Health Sci",
  ["toxicology and forensic medicine : open journal"] = "Toxicol Forensic Med",
  ["toxicology and industrial health"] = "Toxicol Ind Health",
  ["toxicology communications"] = "Toxicol Commun",
  ["toxicology in vitro"] = "Toxicol. in Vitro",
  ["toxicology in vitro : an international journal published in association with bibra"] = "Toxicol In Vitro",
  ["toxicology international"] = "Toxicol Int",
  ["toxicology letters"] = "Toxicol Lett",
  ["toxicology mechanisms and methods"] = "Toxicol Mech Methods",
  ["toxicology reports"] = "Toxicol Rep",
  ["toxicology research"] = "Toxicol Res (Camb)",
  ["toxicology research (cambridge, united kingdom)"] = "Toxicol. Res. (Cambridge, U. K.)",
  ["toxicology research and application"] = "Toxicol. Res. Appl.",
  ["toxicology: open access"] = "Toxicol Open Access",
  ["toxicon : official journal of the international society on toxinology"] = "Toxicon",
  ["toxicon: x"] = "Toxicon X",
  ["toxin reviews"] = "Toxin Rev",
  ["toyama mathematical journal"] = "Toyama Math. J.",
  ["toyoshi kenkyu / [toyoshi kenkyukai]"] = "Toyoshi Kenkyu",
  ["tōhoku shika daigaku gakkai shi"] = "Tohoku Shika Daigaku Gakkai Shi",
  ["tōhō igakkai zasshi = journal of the medical society of toho university"] = "Tōhō Igakkai Zasshi",
  ["tōkyō nōgyō daigaku nōgaku shūhō"] = "Tokyo Nogyo Daigaku Nogaku Shuho",
  ["tōnan ajia kenkyū"] = "Tonan Ajia Kenkyu",
  ["tōyōshi kenkyū"] = "Toyoshi Kenkyu",
  ["történelmi szemle"] = "Tortenelmi Sz",
  ["történeti statisztikai tanulmányok"] = "Tort Stat Tanulm",
  ["tpm. testing, psychometrics, methodology in applied psychology"] = "TPM Test Psychom Methodol Appl Psychol",
  ["trabajos de arqueología de navarra"] = "TrabNavarra",
  ["trabajos de la cátedra de historia de la medicina"] = "Trab Catedra Hist Med",
  ["trabajos de prehistoria"] = "TrabPrehist",
  ["trabajos del instituto cajal"] = "Trab Inst Cajal",
  ["trabajos del instituto cajal / consejo superior de investigaciones cientificas"] = "Trab Inst Cajal",
  ["trabajos del instituto cajal de investigaciones biologicas"] = "Trab. Inst. Cajal Invest. Biol.",
  ["trabajos del instituto cajal de investigaciones biológicas"] = "Trab Inst Cajal Invest Biol",
  ["trabajos y memorias. sociedad nacional de cirugía, havana"] = "Trab Memorias",
  ["trabalhos da associação dos arqueólogos portugueses"] = "TrabAssArqPort",
  ["trabalhos de antropologia e etnologia"] = "TrabAntrEtn",
  ["trabalhos de arqueologia"] = "TrabArq",
  ["trabalhos de investigação"] = "Trab. Invest.",
  ["trac trends in analytical chemistry"] = "TrAC, Trends Anal. Chem.",
  ["trac, trends in analytical chemistry"] = "TrAC, Trends Anal. Chem.",
  ["trace elements and electrolytes"] = "Trace Elem. Electrolytes",
  ["trace elements in medicine"] = "Trace Elem Med",
  ["tracking report"] = "Track Rep",
  ["tracking report / center for studying health system change"] = "Track Rep",
  ["tractrix : yearbook for the history of science, medicine, technology and mathematics"] = "Tractrix",
  ["tradematic. metodologie matematiche e tecnologie informatiche per la finanza"] = "Tradematic",
  ["tradimus : the magazine for rcn nursing students"] = "Tradimus",
  ["traditio. studies in ancient and medieval history, thought and religion"] = "Traditio",
  ["traditio: studies in ancient and medieval history, thought, and religion"] = "Traditio",
  ["tradition (rabbinical council of america)"] = "Tradition",
  ["tradition (rabbinical council of america)."] = "Tradition",
  ["traditional medicine and modern medicine"] = "Tradit. Med. Mod. Med.",
  ["traffic (copenhagen, denmark)"] = "Traffic",
  ["traffic (oxford, united kingdom)"] = "Traffic (Oxford, U. K.)",
  ["traffic injury prevention"] = "Traffic Inj Prev",
  ["traffic related air pollution and internal combustion engines"] = "Air Water Soil Sci T",
  ["training (new york, n.y.)"] = "Training",
  ["training and development journal"] = "Train Dev J",
  ["training and education in professional psychology"] = "Train Educ Prof Psychol",
  ["training school bulletin"] = "Train. Sch. Bull. (Vinel)",
  ["traité des nouvelles technologies"] = "Traité Nouvelles Tech.",
  ["trajectos ciencia"] = "Traject. Cienc.",
  ["trakya university journal of natural sciences"] = "Trak Univ J Nat Sci",
  ["transaction of the international conference on endodontics. international conference on endodontics"] = "Trans Int Conf Endod",
  ["transactions & studies of the college of physicians of philadelphia"] = "Trans Stud Coll Physicians Phila",
  ["transactions (english ceramic circle)"] = "Trans Engl Ceram Circle",
  ["transactions (institute of british geographers : 1965)"] = "Trans Inst Br Geogr",
  ["transactions - american academy of ophthalmology and otolaryngology. american academy of ophthalmology and otolaryngology"] = "Trans Am Acad Ophthalmol Otolaryngol",
  ["transactions - american society for artificial internal organs"] = "Trans Am Soc Artif Intern Organs",
  ["transactions - american society of ophthalmologic and otolaryngologic allergy"] = "Trans Am Soc Ophthalmol Otolaryngol Allergy",
  ["transactions - bristol and gloucestershire archaeological society. bristol and gloucestershire archaeological society"] = "Trans Bristol Glos Archaeol Soc",
  ["transactions - indiana academy of ophthalmology and otolaryngology"] = "Trans Indiana Acad Ophthalmol Otolaryngol",
  ["transactions - pennsylvania academy of ophthalmology and otolaryngology"] = "Trans Pa Acad Ophthalmol Otolaryngol",
  ["transactions - the leicestershire archaeological and historical society. leicestershire archaeological and historical society"] = "Trans Leicester Archaeol Hist Soc",
  ["transactions and proceedings of the american philological association"] = "TAPhA",
  ["transactions and proceedings of the american philological association. american philological association"] = "Trans Proc Am Philol Assoc",
  ["transactions and report / liverpool medical institution. liverpool medical institution"] = "Trans Rep Liverp Med Inst",
  ["transactions and report. liverpool medical institution"] = "Trans Rep Liverp Med Inst",
  ["transactions and studies of the college of physicians of philadelphia"] = "Trans. Stud. Coll. Physicians Phila.",
  ["transactions bulletin. industrial hygiene foundation of america"] = "Trans Bull Ind Hyg Found Am",
  ["transactions in earth, environment, and sustainability"] = "Trans. Earth, Environ., Sustainability",
  ["transactions in gis"] = "Trans. GIS.",
  ["transactions in gis : tg"] = "Trans GIS",
  ["transactions in planning and urban research"] = "Trans. Plann. Urban Res.",
  ["transactions in urban data, science, and technology"] = "Trans. Urban Data, Sci., Technol.",
  ["transactions of a. razmadze mathematical institute"] = "Trans. A. Razmadze Math. Inst.",
  ["transactions of azerbaijan national academy of sciences. series of physical-technical and mathematical sciences. physics and astronomy"] = "Trans. Azerb. Natl. Acad. Sci. Ser. Phys.-Tech. Math. Sci. Phys. Astron.",
  ["transactions of beijing institute of technology. beijing ligong daxue xuebao"] = "Trans. Beijing Inst. Tech.",
  ["transactions of famena"] = "Trans. FAMENA",
  ["transactions of indian national academy of engineering"] = "Trans. Indian Natl. Acad. Eng.",
  ["transactions of mathematics and its applications"] = "Trans. Math. Appl.",
  ["transactions of mathematics and its applications. a journal of the ima"] = "Trans. Math. Appl.",
  ["transactions of national academy of sciences of azerbaijan. series of physical-technical and mathematical sciences"] = "Trans. Natl. Acad. Sci. Azerb. Ser. Phys.-Tech. Math. Sci.",
  ["transactions of society for mining, metallurgy, and exploration, inc"] = "Trans Soc Min Metall Explor Inc",
  ["transactions of society of mining engineers"] = "Trans. Soc. Min. Eng.",
  ["transactions of the american academy of insurance medicine"] = "Trans. Am. Acad. Insur. Med.",
  ["transactions of the american academy of insurance medicine : annual meeting. american academy of insurance medicine. meeting"] = "Trans Am Acad Insur Med",
  ["transactions of the american association of genito-urinary surgeons"] = "Trans Am Assoc Genitourin Surg",
  ["transactions of the american climatological and clinical association. american climatological and clinical association"] = "Trans Am Climatol Clin Assoc",
  ["transactions of the american clinical and climatological association"] = "Trans Am Clin Climatol Assoc",
  ["transactions of the american crystallographic association"] = "Trans. Am. Crystallogr. Assoc.",
  ["transactions of the american entomological society"] = "Trans. Am. Entomol. Soc.",
  ["transactions of the american fisheries society"] = "Trans Am Fish Soc",
  ["transactions of the american geophysical union"] = "Trans. Am. Geophys. Union",
  ["transactions of the american gynecological society"] = "Trans Am Gynecol Soc",
  ["transactions of the american institute of electrical engineers"] = "Trans. Am. Inst. Electr. Eng.",
  ["transactions of the american institute of electrical engineers part 1: communication and electronics"] = "Trans. Am. Inst. Electr. Eng. Part 1",
  ["transactions of the american institute of electrical engineers part 2: applications and industry"] = "Trans. Am. Inst. Electr. Eng. Part 2",
  ["transactions of the american institute of electrical engineers part 3: power apparatus and systems"] = "Trans. Am. Inst. Electr. Eng. Part 3",
  ["transactions of the american institute of metals"] = "Trans. Am. Inst. Met.",
  ["transactions of the american institute of mining, metallurgical and petroleum engineers"] = "Trans. Am. Inst. Min. Metall. Pet. Eng.",
  ["transactions of the american institute of mining, metallurgical and petroleum engineers, society of mining engineers of aime"] = "Trans. Am. Inst. Min. Metall. Pet. Eng. Soc. Min. Eng. AIME",
  ["transactions of the american laryngological, rhinological and otological society"] = "Trans. Am. Laryngol. Rhinol. Otol. Soc.",
  ["transactions of the american laryngological, rhinological and otological society, inc"] = "Trans Am Laryngol Rhinol Otol Soc",
  ["transactions of the american mathematical society"] = "Trans. Amer. Math. Soc.",
  ["transactions of the american mathematical society. series b"] = "Trans. Amer. Math. Soc. Ser. B",
  ["transactions of the american microscopical society"] = "Trans Am Microsc Soc",
  ["transactions of the american neurological association"] = "Trans Am Neurol Assoc",
  ["transactions of the american nuclear society"] = "Trans Am Nucl Soc",
  ["transactions of the american ophthalmological society"] = "Trans Am Ophthalmol Soc",
  ["transactions of the american otological society"] = "Trans Am Otol Soc",
  ["transactions of the american philosophical society"] = "Trans. Amer. Philos. Soc.",
  ["transactions of the american society for metals"] = "Trans. Am. Soc. Met.",
  ["transactions of the american society for neurochemistry"] = "Trans Am Soc Neurochem",
  ["transactions of the american society of mechanical engineers"] = "Trans. Am. Soc. Mech. Eng.",
  ["transactions of the annual meeting of the american broncho-esophagological association"] = "Trans. Annu. Meet. Am. Bronchoesophagol. Assoc.",
  ["transactions of the annual meeting of the american laryngological association"] = "Trans. Am. Laryngol. Assoc.",
  ["transactions of the annual meeting of the orthopaedic research society. orthopaedic research society"] = "Trans Annu Meet Orthop Res Soc",
  ["transactions of the annual meeting. national tuberculosis association"] = "Trans Annu Meet Natl Tuberc Assoc",
  ["transactions of the asabe"] = "Trans. ASABE",
  ["transactions of the asae. american society of agricultural engineers"] = "Trans ASAE",
  ["transactions of the association for computational linguistics"] = "Trans Assoc Comput Linguist",
  ["transactions of the association of american physicians"] = "Trans Assoc Am Physicians",
  ["transactions of the association of industrial medical officers"] = "Trans. Assoc. Ind. Med. Off.",
  ["transactions of the association of life insurance medical directors of america"] = "Trans Assoc Life Insur Med Dir Am",
  ["transactions of the australian college of ophthalmologists"] = "Trans Aust Coll Ophthalmol",
  ["transactions of the british ceramic society"] = "Trans. Br. Ceram. Soc.",
  ["transactions of the british mycological society"] = "Trans. Br. Mycol. Soc.",
  ["transactions of the british society for the history of pharmacy"] = "Trans Br Soc Hist Pharm",
  ["transactions of the british society for the study of orthodontics"] = "Trans Br Soc Study Orthod",
  ["transactions of the cambridge bibliographical society"] = "Trans Camb Bibliogr Soc",
  ["transactions of the canadian ophthalmological society"] = "Trans Can Opthalmolog Soc",
  ["transactions of the canadian society for mechanical engineering"] = "Trans. Can. Soc. Mech. Eng.",
  ["transactions of the egyptian society of chemical engineers"] = "Trans. Egypt. Soc. Chem. Eng.",
  ["transactions of the entomological society of america"] = "Trans. Entomol. Soc. Am.",
  ["transactions of the european orthodontic society"] = "Trans. Eur. Orthod. Soc.",
  ["transactions of the faraday society"] = "Trans. Faraday Soc.",
  ["transactions of the illinois state academy of science. illinois state academy of science"] = "Trans Ill State Acad Sci",
  ["transactions of the imf"] = "Trans. IMF",
  ["transactions of the indian ceramic society"] = "Trans. Indian Ceram. Soc.",
  ["transactions of the indian institute of metals"] = "Trans. Indian Inst. Met.",
  ["transactions of the institute of brewing"] = "Trans. Inst. Brew.",
  ["transactions of the institute of british geographers"] = "Trans. Inst. Br. Geogr.",
  ["transactions of the institute of indian geographers"] = "Trans Inst Indian Geogr",
  ["transactions of the institute of measurement and control"] = "Trans. Inst. Meas. Control",
  ["transactions of the institute of metal finishing"] = "Trans. Inst. Met. Finish.",
  ["transactions of the institution of civil engineers"] = "Trans. Inst. Civ. Eng.",
  ["transactions of the institution of metal finishing"] = "Trans. Inst. Met. Finish.",
  ["transactions of the institution of mining and metallurgy section a"] = "Trans. Inst. Min. Metall., Sect. A",
  ["transactions of the institution of mining and metallurgy section b"] = "Trans. Inst. Min. Metall., Sect. B",
  ["transactions of the institution of mining and metallurgy section c"] = "Trans. Inst. Min. Metall., Sect. C",
  ["transactions of the institution of mining and metallurgy section c: mineral processing and extractive metallurgy"] = "Trans. Inst. Min. Metall., Sect. C",
  ["transactions of the international astronomical union"] = "Trans. Int. Astron. Union",
  ["transactions of the international conference on endodontics"] = "Trans. Int. Conf. Endod.",
  ["transactions of the international conference on oral surgery"] = "Trans Int Conf Oral Surg",
  ["transactions of the international society for music information retrieva"] = "Trans. Int. Soc. Music Inf. Retr.",
  ["transactions of the japan academy"] = "Trans. Japan Acad.",
  ["transactions of the japan institute of electronics packaging"] = "Trans. Jpn. Inst. Electron. Packag.",
  ["transactions of the japan society for aeronautical and space sciences"] = "Trans Jpn Soc Aeronaut Space Sci",
  ["transactions of the japanese society for artificial intelligence = jinkō chinō gakkai ronbunshi"] = "Trans Jpn Soc Artif Intell",
  ["transactions of the kansas academy of science"] = "Trans. Kansas Acad. Sci.",
  ["transactions of the kansas academy of science. kansas academy of science"] = "Trans Kans Acad Sci",
  ["transactions of the korean institute of electrical engineers"] = "Trans. Korean Inst. Electr. Eng.",
  ["transactions of the korean society of mechanical engineers, a"] = "Trans. Korean Soc. Mech. Eng. A",
  ["transactions of the korean society of mechanical engineers, b"] = "Trans. Korean Soc. Mech. Eng. B",
  ["transactions of the laboratory club"] = "Trans. Lab. Club",
  ["transactions of the lancashire and cheshire antiquarian society. lancashire and cheshire antiquarian society"] = "Trans Lancaster Cheshire Antiq Soc",
  ["transactions of the lepidopterological society of japan"] = "Trans Lepidopterol Soc Jpn",
  ["transactions of the linnean society of london"] = "Trans. Linn. Soc. London",
  ["transactions of the london and middlesex archaeological society"] = "TransactLond",
  ["transactions of the london mathematical society"] = "Trans. London Math. Soc.",
  ["transactions of the materials research society of japan"] = "Trans. Mater. Res. Soc. Jpn",
  ["transactions of the medical society of london"] = "Trans Med Soc Lond",
  ["transactions of the metallurgical society of aime"] = "Trans. Metall. Soc. AIME",
  ["transactions of the microscopical society of london"] = "Trans. Microsc. Soc. London",
  ["transactions of the missouri academy of science. missouri academy of science"] = "Trans Mo Acad Sci",
  ["transactions of the moscow mathematical society"] = "Trans. Moscow Math. Soc.",
  ["transactions of the nebraska academy of sciences and affiliated societies"] = "Trans Nebr Acad Sci Affil Soc",
  ["transactions of the new england obstetrical and gynecological society"] = "Trans N Engl Obstet Gynecol Soc",
  ["transactions of the new orleans academy of ophthalmology"] = "Trans New Orleans Acad Ophthalmol",
  ["transactions of the new york academy of sciences"] = "Trans N Y Acad Sci",
  ["transactions of the new york academy of sciences, series ii"] = "Trans. New York Acad. Sci. Ser. II",
  ["transactions of the nonferrous metals society of china"] = "Trans. Nonferrous Met. Soc. China",
  ["transactions of the ophthalmological societies of the united kingdom"] = "Trans Ophthalmol Soc U K",
  ["transactions of the ophthalmological society of australia"] = "Trans Ophthalmol Soc Aust",
  ["transactions of the ophthalmological society of new zealand"] = "Trans Ophthalmol Soc N Z",
  ["transactions of the optical society"] = "Trans. Opt. Soc.",
  ["transactions of the pacific coast obstetrical and gynecological society"] = "Trans Pac Coast Obstet Gynecol Soc",
  ["transactions of the pacific coast oto-ophthalmological society annual meeting"] = "Trans Pac Coast Otoophthalmol Soc Annu Meet",
  ["transactions of the philological society"] = "TPhS",
  ["transactions of the philological society. philological society (great britain)"] = "Trans Philol Soc",
  ["transactions of the royal entomological society of london"] = "Trans. R. Entomol. Soc. London",
  ["transactions of the royal historical society"] = "Trans R Hist Soc",
  ["transactions of the royal schools of dentistry, stockholm and umea"] = "Trans. R. Sch. Dent. Stockh. Umea",
  ["transactions of the royal schools of dentistry, stockholm and umea. stockholm. tandlakarhogskolan"] = "Trans R Sch Dent Stockh Umea",
  ["transactions of the royal schools of dentistry, stockholm and umeå. stockholm. tandläkarhögskolan"] = "Trans R Sch Dent Stockh Umea",
  ["transactions of the royal society of canada"] = "Trans. R. Soc. Can.",
  ["transactions of the royal society of canada = mémoires de la société royale du canada"] = "Trans R Soc Can",
  ["transactions of the royal society of canada. section 5, biological sciences. royal society of canada"] = "Trans R Soc Can 5 Biol Sci",
  ["transactions of the royal society of canada. sections 1, 2, and 3 = mémoires de la société royale du canada. sections 1, 2, et 3"] = "Trans R Soc Can (1963)",
  ["transactions of the royal society of edinburgh earth science"] = "Trans. R. Soc. Edinburgh: Earth Sci.",
  ["transactions of the royal society of edinburgh, earth sciences"] = "Trans. R. Soc. Edinb., Earth Sci.",
  ["transactions of the royal society of edinburgh: earth sciences"] = "Trans R Soc Edinb Earth Sci",
  ["transactions of the royal society of south africa"] = "Trans. R. Soc. S. Afr.",
  ["transactions of the royal society of south australia"] = "Trans. R. Soc. S. Aust.",
  ["transactions of the royal society of south australia, incorporated. royal society of south australia"] = "Trans R Soc S Aust",
  ["transactions of the royal society of tropical medicine & hygiene"] = "Trans. R. Soc. Trop. Med. Hyg.",
  ["transactions of the royal society of tropical medicine and hygiene"] = "Trans R Soc Trop Med Hyg",
  ["transactions of the saest"] = "Trans. SAEST",
  ["transactions of the society for advancement of electrochemical science and technology"] = "Trans. Soc. Adv. Electrochem. Sci. Technol.",
  ["transactions of the society for mining, metallurgy, and exploration"] = "Trans. Soc. Min. Metall. Explor.",
  ["transactions of the society of mining engineers of aime"] = "Trans. Soc. Min. Eng. AIME",
  ["transactions of the society of occupational medicine"] = "Trans. Soc. Occup. Med.",
  ["transactions of the society of rheology"] = "Trans. Soc. Rheol.",
  ["transactions of the south african institute of electrical engineering"] = "Trans. South Afr. Inst. Electr. Eng.",
  ["transactions of the southern surgical association. southern surgical association (u.s.)"] = "Trans South Surg Assoc",
  ["transactions of the st johns hospital dermatological society"] = "Trans. St Johns Hosp. Dermatol. Soc.",
  ["transactions of the st. john's hospital dermatological society"] = "Trans St Johns Hosp Dermatol Soc",
  ["transactions of the st. john’s hospital dermatological society"] = "Trans St Johns Hosp Dermatol Soc",
  ["transactions of the zoological society of london"] = "Trans. Zool. Soc. London",
  ["transactions of tianjin university"] = "Trans. Tianjin Univ.",
  ["transactions on combinatorics"] = "Trans. Comb.",
  ["transactions on data privacy"] = "Trans Data Priv",
  ["transactions on electrical and electronic materials"] = "Trans. Electr. Electron. Mater.",
  ["transactions on emerging telecommunications technologies"] = "Trans. Emerging Telecommun. Technol.",
  ["transactions, american academy of ophthalmology and otolaryngology"] = "Trans. Am. Acad. Ophthalmol. Otolaryngol.",
  ["transactions, american otological society"] = "Trans. Am. Otol. Soc.",
  ["transactions, american society for artificial internal organs"] = "Trans. Am. Soc. Artif. Intern. Organs",
  ["transactions, indiana academy of ophthalmology and otolaryngology"] = "Trans. Indiana Acad. Ophthalmol. Otolaryngol.",
  ["transactions, new orleans academy of ophthalmology"] = "Trans. New Orleans Acad. Ophthalmol.",
  ["transactions, pennsylvania academy of ophthalmology and otolaryngology"] = "Trans. Pa. Acad. Ophthalmol. Otolaryngol.",
  ["transactions. american college of cardiology"] = "Trans Am Coll Cardiol",
  ["transactions. american proctologic society"] = "Trans Am Proctol Soc",
  ["transactions. american urological association. south central section"] = "Trans South Cent Sect Am Urol Assoc",
  ["transactions. american urological association. southeastern section"] = "Trans Southeast Sect Am Urol Assoc",
  ["transactions. american urological association. western section"] = "Trans West Sect Am Urol Assoc",
  ["transactions. australasian medical congress (british medical association)"] = "Trans Aust Med Congr",
  ["transactions. conference on liver injury"] = "Trans Conf Liver Inj",
  ["transactions. conference on metabolic aspects of convalescence"] = "Trans Conf Metab Asp Conval",
  ["transactions. edinburgh obstetrical society"] = "Trans Edinb Obstet Soc",
  ["transactions. epidemiological society of london"] = "Trans Epidemiol Soc Lond",
  ["transactions. european orthodontic society"] = "Trans Eur Orthod Soc",
  ["transactions. medico-chirurgical society of edinburgh"] = "Trans Med Chir Soc Edinb",
  ["transactions. national safety council. national safety congress"] = "Trans Natl Saf Congr",
  ["transactions. ophthalmological society of the united kingdom"] = "Trans Ophthalmol Soc U K",
  ["transactions. section c, mineral processing & extractive metallurgy"] = "Miner Process Extr Metall",
  ["transactions. section on ophthalmology. american academy of ophthalmology and otolaryngology"] = "Trans Sect Ophthalmol Am Acad Ophthalmol Otolaryngol",
  ["transactions. section on otolaryngology. american academy of ophthalmology and otolaryngology"] = "Trans Sect Otolaryngol Am Acad Ophthalmol Otolaryngol",
  ["transactions. western surgical association"] = "Trans West Surg Assoc",
  ["transboundary and emerging diseases"] = "Transbound Emerg Dis",
  ["transcriptomics: open access"] = "Transcr Open Access",
  ["transcultural psychiatry"] = "Transcult Psychiatry",
  ["transcultural research—heidelberg studies on asia and europe in a global context"] = "Transcult. Res. Heidelb. Stud. Asia Europe Glob. Context",
  ["transdisciplinary journal of engineering & science"] = "Transdiscipl J Eng Sci",
  ["transfer (brussels, belgium)"] = "Transfer (Bruss)",
  ["transformation groups"] = "Transform. Groups",
  ["transformat︠s︡ii︠a︡ ėkosistem"] = "Transform Ekosistem",
  ["transforming digital worlds : 13th international conference, iconference 2018, sheffield, uk, march 25-28, 2018, proceedings. international conference on transforming digital worlds (13th : 2018 : sheffield, england)"] = "Transform Digit Worlds (2018)",
  ["transfusion alternatives in transfusion medicine : tatm"] = "Transfus Altern Transfus Med",
  ["transfusion and apheresis science"] = "Transfus. Apheresis Sci.",
  ["transfusion and apheresis science : official journal of the world apheresis association : official journal of the european society for haemapheresis"] = "Transfus Apher Sci",
  ["transfusion clinique et biologique"] = "Transfus. Clin. Biol.",
  ["transfusion clinique et biologique : journal de la societe francaise de transfusion sanguine"] = "Transfus Clin Biol",
  ["transfusion clinique et biologique : journal de la société française de transfusion sanguine"] = "Transfus Clin Biol",
  ["transfusion medicine"] = "Transfus. Med.",
  ["transfusion medicine (oxford, england)"] = "Transfus Med",
  ["transfusion medicine and hemotherapy : offizielles organ der deutschen gesellschaft fur transfusionsmedizin und immunhamatologie"] = "Transfus Med Hemother",
  ["transfusion medicine and hemotherapy : offizielles organ der deutschen gesellschaft fur̈ transfusionsmedizin und immunham̈atologie"] = "Transfus Med Hemother",
  ["transfusion medicine reviews"] = "Transfus Med Rev",
  ["transfusion science"] = "Transfus Sci",
  ["transgender health"] = "Transgend Health",
  ["transgene : biological analysis through dna transfer"] = "Transgene",
  ["transgenic research"] = "Transgenic Res",
  ["transgenics : biological analysis through dna transfer"] = "Transgenics",
  ["transilvania university of brașov"] = "Bull. Transilv. Univ. Brașov Ser. A (N.S.)",
  ["transition metal chemistry"] = "Transition Met. Chem.",
  ["transition metal chemistry (dordrecht, netherlands)"] = "Transition Met. Chem. (Dordrecht, Neth.)",
  ["transition metal chemistry (london)"] = "Transition Met. Chem. (London)",
  ["transitional waters bulletin"] = "Transit Water Bull",
  ["translation (austin, tex.)"] = "Translation (Austin)",
  ["translation (santa barbara, calif.)"] = "Translation (St Barbara, Calif)",
  ["translational and clinical pharmacology"] = "Transl Clin Pharmacol",
  ["translational andrology and urology"] = "Transl Androl Urol",
  ["translational animal science"] = "Transl Anim Sci",
  ["translational behavioral medicine"] = "Transl Behav Med",
  ["translational biomedicine"] = "Transl Biomed",
  ["translational biophotonics"] = "Transl Biophotonics",
  ["translational brain rhythmicity"] = "Transl Brain Rhythm",
  ["translational cancer research"] = "Transl Cancer Res",
  ["translational developmental psychiatry"] = "Transl Dev Psychiatry",
  ["translational endocrinology & metabolism"] = "Transl Endocrinol Metab",
  ["translational gastroenterology and hepatology"] = "Transl Gastroenterol Hepatol",
  ["translational gastrointestinal cancer"] = "Transl Gastrointest Cancer",
  ["translational issues in psychological science"] = "Transl Issues Psychol Sci",
  ["translational journal of the american college of sports medicine"] = "Transl J Am Coll Sports Med",
  ["translational lung cancer research"] = "Transl Lung Cancer Res",
  ["translational materials research"] = "Transl. Mater. Res.",
  ["translational medicine (sunnyvale, calif.)"] = "Transl Med (Sunnyvale)",
  ["translational medicine @ unisa"] = "Transl Med UniSa",
  ["translational medicine communications"] = "Transl Med Commun",
  ["translational medicine of aging"] = "Transl Med Aging",
  ["translational metabolic syndrome research"] = "Transl. Metab. Syndr. Res.",
  ["translational neurodegeneration"] = "Transl Neurodegener",
  ["translational neuroscience"] = "Transl Neurosci",
  ["translational oncogenomics"] = "Transl Oncogenomics",
  ["translational oncology"] = "Transl Oncol",
  ["translational pediatrics"] = "Transl Pediatr",
  ["translational perioperative and pain medicine"] = "Transl Perioper Pain Med",
  ["translational proteomics"] = "Transl. Proteomics",
  ["translational psychiatry"] = "Transl Psychiatry",
  ["translational research"] = "Transl. Res.",
  ["translational research : the journal of laboratory and clinical medicine"] = "Transl Res",
  ["translational respiratory medicine"] = "Transl Respir Med",
  ["translational science"] = "Transl Sci",
  ["translational science of rare diseases"] = "Transl Sci Rare Dis",
  ["translational sports medicine"] = "Transl Sports Med",
  ["translational stroke research"] = "Transl Stroke Res",
  ["translational systems sciences"] = "Transl. Syst. Sci.",
  ["translational vision science & technology"] = "Transl Vis Sci Technol",
  ["translations of mathematical monographs"] = "Transl. Math. Monogr.",
  ["transnational corporations"] = "Transnat. Corporations",
  ["transpersonal psychology review"] = "Transpers Psychol Rev",
  ["transplant immunology"] = "Transpl Immunol",
  ["transplant infectious disease"] = "Transpl. Infect. Dis.",
  ["transplant infectious disease : an official journal of the transplantation society"] = "Transpl Infect Dis",
  ["transplant international"] = "Transplant Int.",
  ["transplant international : official journal of the european society for organ transplantation"] = "Transpl Int",
  ["transplant research and risk management"] = "Transplant Res. Risk Manage.",
  ["transplantation and cellular therapy"] = "Transplant Cell Ther",
  ["transplantation bulletin"] = "Transplant Bull",
  ["transplantation direct"] = "Transplant Direct",
  ["transplantation proceedings"] = "Transplant Proc",
  ["transplantation research"] = "Transplant Res",
  ["transplantation reviews"] = "Transplant Rev",
  ["transplantation reviews (orlando, fla.)"] = "Transplant Rev (Orlando)",
  ["transplantation science"] = "Transplant Sci",
  ["transport history"] = "Trans Hist",
  ["transport in porous media"] = "Transp. Porous Media",
  ["transport policy"] = "Transp Policy (Oxf)",
  ["transport reviews"] = "Transp Rev",
  ["transport theory and statistical physics"] = "Transp Theory Stat Phys",
  ["transportation cybersecurity and privacy"] = "Transp. Cybersecur. Privacy",
  ["transportation engineering"] = "Transp. Eng.",
  ["transportation geotechnics"] = "Transp. Geotech.",
  ["transportation in developing economies"] = "Transp. Dev. Econ.",
  ["transportation infrastructure geotechnology"] = "Transp. Infrastruct. Geotechnol.",
  ["transportation journal"] = "Transp. J.",
  ["transportation letters"] = "Transp. Lett.",
  ["transportation planning and technology"] = "Transp. Plann. Technol.",
  ["transportation research"] = "Transp. Res.",
  ["transportation research interdisciplinary perspectives"] = "Transp Res Interdiscip Perspect",
  ["transportation research part a policy and practice"] = "Transp. Res. Part A Policy Pract.",
  ["transportation research part b methodological"] = "Transp. Res. Part B Methodol.",
  ["transportation research part c emerging technologies"] = "Transp. Res. Part C Emerging Technol.",
  ["transportation research part d transport and environment"] = "Transp. Res. Part D Transp. Environ.",
  ["transportation research part e logistics and transportation review"] = "Transp. Res. Part E Logist. Transp. Rev.",
  ["transportation research part f traffic psychology and behaviour"] = "Transp. Res. Part F Psychol. Behav.",
  ["transportation research procedia"] = "Transp. Res. Procedia",
  ["transportation research record"] = "Transp. Res. Rec.",
  ["transportation research. part a, policy and practice"] = "Transp Res Part A Policy Pract",
  ["transportation research. part c, emerging technologies"] = "Transp Res Part C Emerg Technol",
  ["transportation research. part d, transport and environment"] = "Transp Res D Transp Environ",
  ["transportation research. part e, logistics and transportation review"] = "Transp Res E Logist Transp Rev",
  ["transportation research. part f, traffic psychology and behaviour"] = "Transp Res Part F Traffic Psychol Behav",
  ["transportation research: part a: policy and practice"] = "Transp. Res.: Part A: Pol. Practice",
  ["transportation research: part b: methodological"] = "Transp. Res.: Part B: Methodological",
  ["transportation research: part d: transport and environment"] = "Transp. Res.: Part D: Transport Environ.",
  ["transportation research: part e: logistics and transporation review"] = "Transp. Res.: Part E: Logist. Transp. Rev.",
  ["transportation safety and environment"] = "Transp. Saf.  Environ.",
  ["transportmetrica a: transport science"] = "Transportmetrica A: Transport Sci.",
  ["transportmetrica b: transport dynamics"] = "Transportmetrica B: Transport Dyn.",
  ["trastornos adictivos"] = "Trastor Adict",
  ["trauma & acute care"] = "Trauma Acute Care",
  ["trauma (london, england)"] = "Trauma",
  ["trauma and emergency care"] = "Trauma Emerg Care",
  ["trauma case reports"] = "Trauma Case Rep",
  ["trauma monthly"] = "Trauma Mon",
  ["trauma surgery & acute care open"] = "Trauma Surg Acute Care Open",
  ["trauma und berufskrankheit"] = "Trauma Berufskrankh",
  ["trauma, violence & abuse"] = "Trauma Violence Abuse",
  ["travail et emploi"] = "Travail Emploi",
  ["travaux de la societe de pharmacie de montpellier"] = "Trav Soc Pharm Montp",
  ["travaux de la société de pharmacie de montpellier"] = "Trav Soc Pharm Montp",
  ["travaux de logique"] = "Travaux Log.",
  ["travaux de l’institut d’art préhistorique, université de toulouse – le mirail"] = "TravToulouse",
  ["travaux des laboratoires de matiere medicale et de pharmacie galenique de la faculte de pharmacie de paris"] = "Trav Lab Matiere Med Pharm Galenique Fac Pharm Paris",
  ["travaux des laboratoires de matière médicale et de pharmacie galénique de la faculté de pharmacie de paris"] = "Trav Lab Matiere Med Pharm Galenique Fac Pharm Paris",
  ["travaux du centre de recherches sémiologiques"] = "Travaux Centre Rech. Sémiol.",
  ["travaux du cercle linguistique d’aix-en-provence"] = "TCLA",
  ["travaux en cours"] = "Travaux en Cours",
  ["travaux et mémoires. centre de recherche d’histoire et civilisation byzantine, paris"] = "TravMem",
  ["travaux et mémoires du centre de recherches d’hist. et civil. byzantines"] = "T&MBYZ",
  ["travaux mathematiques"] = "Trav. Math.",
  ["travaux neuchâtelois de linguistique"] = "TraNeL",
  ["travaux scientifiques du parc national de la vanoise"] = "Trav. Sci. Parc Natl. Vanoise",
  ["travaux station de recherches des eaux et forêts"] = "Trav. Stn. rech. eaux for.",
  ["travel behaviour & society"] = "Travel Behav Soc",
  ["travel medicine and infectious disease"] = "Travel Med Infect Dis",
  ["trazos; entretenimientos de clínica y terapéutica"] = "Trazos (Bilbao)",
  ["treatment guidelines from the medical letter"] = "Treat Guidel Med Lett",
  ["treatment review"] = "Treat Rev",
  ["treatment services bulletin. canada. department of veterans' affairs"] = "Treat Serv Bull",
  ["treatment services bulletin. canada. dept. of veterans’ affairs"] = "Treat Serv Bull",
  ["treatment strategies. diabetes"] = "Treat Strategies Diabetes",
  ["treatment strategies. hematology"] = "Treat Strategies Hematol",
  ["treatment strategies. oncology"] = "Treat Strateg Oncol",
  ["treatments in endocrinology"] = "Treat Endocrinol",
  ["treatments in respiratory medicine"] = "Treat Respir Med",
  ["tree genetics & genomes"] = "Tree Genet Genomes",
  ["tree physiology"] = "Tree Physiol",
  ["tree-ring bulletin"] = "Tree-ring Bull.",
  ["tree-ring research"] = "Tree Ring Res",
  ["trees (berlin, germany : west)"] = "Trees (Berl West)",
  ["trees, forests and people"] = "Trees, Forests People",
  ["trees-structure and function"] = "Trees-Struct. Funct.",
  ["tremor and other hyperkinetic movements (new york, n.y.)"] = "Tremor Other Hyperkinet Mov (N Y)",
  ["trends & techniques in the contemporary dental laboratory"] = "Trends Tech Contemp Dent Lab",
  ["trends and techniques in the contemporary dental laboratory"] = "Trends Tech. Contemp. Dent. Lab.",
  ["trends in abstract and applied analysis"] = "Trends Abstr. Appl. Anal.",
  ["trends in amplification"] = "Trends Amplif",
  ["trends in analytical chemistry"] = "Trends Anal. Chem.",
  ["trends in analytical chemistry : trac"] = "Trends Analyt Chem",
  ["trends in applied sciences research"] = "Trends Appl Sci Res",
  ["trends in applied spectroscopy"] = "Trends Appl Spectrosc",
  ["trends in biochemical sciences"] = "Trends Biochem Sci",
  ["trends in biochemical sciences (cambridge"] = "Trends Biochem Sci",
  ["trends in biomaterials & artificial organs"] = "Trends Biomater Artif Organs",
  ["trends in biomaterials and artificial organs"] = "Trends Biomater. Artif. Organs",
  ["trends in biosciences"] = "Trends Biosci.",
  ["trends in biotechnology"] = "Trends Biotechnol",
  ["trends in biotechnology research (solapur)"] = "Trends Biotechnol Res",
  ["trends in cancer"] = "Trends Cancer",
  ["trends in cancer research"] = "Trends Cancer Res",
  ["trends in carbohydrate research"] = "Trends Carbohydr Res",
  ["trends in cardiovascular medicine"] = "Trends Cardiovasc Med",
  ["trends in cell & molecular biology"] = "Trends Cell Mol Biol",
  ["trends in cell biology"] = "Trends Cell Biol",
  ["trends in chemistry"] = "Trends Chem",
  ["trends in classics"] = "TC",
  ["trends in classics—supplementary volumes"] = "Trends Class. Suppl. Vol.",
  ["trends in cognitive sciences"] = "Trends Cogn Sci",
  ["trends in computational and applied mathematics"] = "Trends Comput. Appl. Math.",
  ["trends in developmental biology"] = "Trends Dev Biol",
  ["trends in diabetes and metabolism"] = "Trends Diabetes Metab",
  ["trends in ecology & evolution"] = "Trends Ecol Evol",
  ["trends in ecology & evolution (personal edition)"] = "Trends Ecol Evol",
  ["trends in ecology and evolution"] = "Trends Ecol. Evol.",
  ["trends in endocrinology and metabolism"] = "Trends Endocrinol. Metab.",
  ["trends in endocrinology and metabolism: tem"] = "Trends Endocrinol Metab",
  ["trends in entomology"] = "Trends Entomol",
  ["trends in environmental analytical chemistry"] = "Trends Environ. Anal. Chem.",
  ["trends in evolutionary biology"] = "Trends Evol Biol",
  ["trends in food science & technology"] = "Trends Food Sci Technol",
  ["trends in food science and technology"] = "Trends Food Sci. Technol.",
  ["trends in genetics"] = "Trends Genet.",
  ["trends in genetics : tig"] = "Trends Genet",
  ["trends in glycoscience and glycotechnology : tigg"] = "Trends Glycosci Glycotechnol",
  ["trends in health care, law & ethics"] = "Trends Health Care Law Ethics",
  ["trends in health care, law and ethics"] = "Trends Health Care Law Ethics",
  ["trends in hearing"] = "Trends Hear",
  ["trends in history"] = "Trends Hist",
  ["trends in immunology"] = "Trends Immunol",
  ["trends in logic—studia logica library"] = "Trends Log. Stud. Log. Libr.",
  ["trends in mathematics"] = "Trends Math.",
  ["trends in mathematics. research perspectives"] = "Trends Math. Res. Perspect.",
  ["trends in mathematics. research perspectives crm barcelona"] = "Trends Math. Res. Perspect. CRM Barc.",
  ["trends in microbiology"] = "Trends Microbiol",
  ["trends in molecular medicine"] = "Trends Mol Med",
  ["trends in neuroscience and education"] = "Trends Neurosci Educ",
  ["trends in neurosciences"] = "Trends Neurosci",
  ["trends in organic chemistry"] = "Trends Org Chem",
  ["trends in organized crime"] = "Trends Organ Crime",
  ["trends in parasitology"] = "Trends Parasitol",
  ["trends in pharmacological sciences"] = "Trends Pharmacol Sci",
  ["trends in plant science"] = "Trends Plant Sci",
  ["trends in polymer science"] = "Trends Polym Sci",
  ["trends in psychiatry and psychotherapy"] = "Trends Psychiatry Psychother",
  ["trends in reproductive biology"] = "Tren Reprod Bio",
  ["trends in research"] = "Trends Res",
  ["trends in sport sciences"] = "Trends Sport Sci",
  ["trends in the history of science"] = "Trends Hist. Sci.",
  ["trenie i iznos"] = "Trenie Iznos",
  ["trial (boston, mass.)"] = "Trial",
  ["trial lawyers quarterly"] = "Trial Lawyers Q",
  ["trials in vaccinology"] = "Trials Vaccinol",
  ["triangle; the sandoz journal of medical science"] = "Triangle",
  ["tribologie und schmierungstechnik"] = "Tribol. Schmierungstech.",
  ["tribology - materials, surfaces and & interfaces"] = "Tribol. Mater. Surf. Interfaces",
  ["tribology - materials, surfaces and interfaces"] = "Tribol. Mater. Surf. Interfaces",
  ["tribology and interface engineering series"] = "Tribol. Interface Eng. Ser.",
  ["tribology international"] = "Tribol. Int.",
  ["tribology letters"] = "Tribol. Lett.",
  ["tribology transactions"] = "Tribol. Trans.",
  ["tribulus : bulletin of the emirates natural history group"] = "Tribulus (Abu Dhabi)",
  ["tribuna d’arqueologia"] = "TribArq",
  ["tribuna farmacêutica"] = "Trib Farm",
  ["tribuna médico-social"] = "Trib Med Soc",
  ["tribuna odontologica"] = "Trib. Odontol. (B. Aires)",
  ["tribuna odontologica do sindicato dos odontologistas do estado da guanabara"] = "Trib Odontol (Guanabara)",
  ["tribune (international women's tribune centre : english ed.)"] = "Trib Int Womens Trib Cent",
  ["tribune (international women’s tribune centre : english ed.)"] = "Trib Int Womens Trib Cent",
  ["trierer winckelmannsprogramme"] = "TrWPr",
  ["trierer zeitschrift für geschichte und kunst des trierer landes und seiner nachbargebiete"] = "TrZ",
  ["trierer zeitschrift, für geschichte und kunst des trierer landes und seiner nachbargebiete"] = "TZ",
  ["trinity journal, publ. by trinity evangelical divinity school"] = "TJ",
  ["triple helix"] = "Triple Helix",
  ["triple helix (heidelberg)"] = "Triple Helix (Heidelb)",
  ["triumph (washington, d.c.)"] = "Triumph",
  ["tropenmedizin und parasitologie"] = "Tropenmed Parasitol",
  ["tropical agriculturist"] = "Trop Agric",
  ["tropical and geographical medicine"] = "Trop Geogr Med",
  ["tropical and subtropical agroecosystems"] = "Trop. Subtrop. Agroecosyt.",
  ["tropical animal health and production"] = "Trop Anim Health Prod",
  ["tropical biomedicine"] = "Trop Biomed",
  ["tropical bryology"] = "Trop Bryol",
  ["tropical conservation science"] = "Trop Conserv Sci",
  ["tropical cyclone research and review"] = "Trop. Cyclone Res. Rev.",
  ["tropical diseases bulletin"] = "Trop Dis Bull",
  ["tropical diseases, travel medicine and vaccines"] = "Trop Dis Travel Med Vaccines",
  ["tropical doctor"] = "Trop Doct",
  ["tropical ecology"] = "Trop Ecol",
  ["tropical gastroenterology"] = "Trop. Gastroenterol.",
  ["tropical gastroenterology : official journal of the digestive diseases foundation"] = "Trop Gastroenterol",
  ["tropical grasslands"] = "Trop. Grasslands",
  ["tropical health"] = "Trop Heal",
  ["tropical journal of natural product research"] = "Trop. J. Nat. Prod. Res.",
  ["tropical journal of obstetrics and gynaecology"] = "Trop J Obstet Gynaecol",
  ["tropical journal of pharmaceutical research"] = "Trop. J. Pharm. Res.",
  ["tropical journal of pharmaceutical research : tjpr"] = "Trop J Pharm Res",
  ["tropical lepidoptera research"] = "Trop Lepid Res",
  ["tropical life sciences research"] = "Trop Life Sci Res",
  ["tropical medicine & international health : tm & ih"] = "Trop Med Int Health",
  ["tropical medicine & surgery"] = "Trop Med Surg",
  ["tropical medicine and health"] = "Trop Med Health",
  ["tropical medicine and infectious disease"] = "Trop Med Infect Dis",
  ["tropical medicine and international health"] = "Trop. Med. Int. Health",
  ["tropical medicine and parasitology"] = "Trop. Med. Parasitol.",
  ["tropical medicine and parasitology : official organ of deutsche tropenmedizinische gesellschaft and of deutsche gesellschaft fur technische zusammenarbeit (gtz)"] = "Trop Med Parasitol",
  ["tropical medicine and parasitology : official organ of deutsche tropenmedizinische gesellschaft and of deutsche gesellschaft für technische zusammenarbeit (gtz)"] = "Trop Med Parasitol",
  ["tropical medicine news"] = "Trop Med News",
  ["tropical natural history"] = "Trop Nat Hist",
  ["tropical parasitology"] = "Trop Parasitol",
  ["tropical plant biology"] = "Trop Plant Biol",
  ["tropical plant pathology"] = "Trop Plant Pathol",
  ["tropical zoology"] = "Trop. Zool.",
  ["trudy (institut biologii vnutrennikh vod im. i.d. papanina)"] = "Trudy (Inst Biol Vnutr Vod Im ID Papanina)",
  ["trudy - institut fiziologii imeni i. p. pavlova"] = "Tr Inst Fiziol Im I P Pavlova",
  ["trudy - institut normal'noĭ i patologicheskoĭ fiziologii"] = "Tr Inst Norm Patol Fiziol",
  ["trudy - institut normal’noi i patologicheskoi fiziologii"] = "Tr Inst Norm Patol Fiziol",
  ["trudy belorusskogo gosudarstvennogo universiteta. serii︠a︡ fiziologicheskie, biohimicheskie i molekuli︠a︡rnye osnovy funkt︠s︡ionirovanii︠a︡ sistem"] = "Tr Beloruss Gos Univ Ser Fiziol Biohim Mol Osn Funkc Sist",
  ["trudy bgtu"] = "Trudy BGTU",
  ["trudy fiziologicheskikh laboratorii akademika i.p. pavlova"] = "Tr Fiziol Lab Akad I P Pavlova",
  ["trudy fiziologicheskikh laboratoriĭ akademika i.p. pavlova"] = "Tr Fiziol Lab Akad I P Pavlova",
  ["trudy gelʹmintologicheskoĭ laboratorii"] = "Tr Gelmintol Lab",
  ["trudy geofizicheskogo instituta, akademiya nauk sssr"] = "Tr. Geofiz. Inst., Akad. Nauk SSSR",
  ["trudy glavnoi geofizicheskoi observatorii"] = "Tr. Gl. Geofiz. Obs.",
  ["trudy gosudarstbennogo èrmitaza (travaux du musée d’état de l’ermitage)"] = "TÈ",
  ["trudy gosudarstvennogo ėrmitaža"] = "TrudyErmit",
  ["trudy instituta biologii"] = "Tr Inst Biol",
  ["trudy instituta fiziologii akademiia nauk gruzinskoi ssr"] = "Tr. Inst. Fiz. Akad. Nauk Gruz. SSR",
  ["trudy instituta fiziologii, akademiia nauk gruzinskoi ssr"] = "Tr Inst Fiz Akad Nauk Gruz Ssr",
  ["trudy instituta fiziologii, akademiia nauk gruzinskoĭ ssr"] = "Tr Inst Fiz Akad Nauk Gruz Ssr",
  ["trudy instituta imeni pastera"] = "Tr Inst Im Pastera",
  ["trudy instituta matematiki"] = "Tr. Inst. Mat.",
  ["trudy instituta matematiki i mekhaniki"] = "Tr. Inst. Mat. Mekh.",
  ["trudy instituta mikrobiologii"] = "Trudy Inst Mikrobiol",
  ["trudy instituta morfologii zhivotnykh im. a. n. severtsova"] = "Tr Inst Morfol Zhivotn An Severtsova",
  ["trudy instituta prikladnoi matematiki i mekhaniki"] = "Tr. Inst. Prikl. Mat. Mekh.",
  ["trudy leningradskogo instituta epidemiologii i mikrobiologii im. pastera"] = "Tr. Leningr. Inst. Epidemiol. Mikrobiol.",
  ["trudy leningradskogo instituta ėpidemiologii i mikrobiologii im. pastera"] = "Tr Leningr Inst Epidemiol Mikrobiol",
  ["trudy leningradskogo instituta ėpidemiologii, mikrobiologii, i gigieny im. pastera"] = "Trudy Leningr Inst Epidemiol Mikrobiol Gig Im Pastera",
  ["trudy leningradskogo nauchno-issledovatelskogo instituta epidemiologii i mikrobiologii im. pastera"] = "Tr. Leningr. Nauchnoissled. Inst. Epidemiol. Mikrobiol.",
  ["trudy leningradskogo nauchno-issledovatelʹskogo instituta ėpidemiologii i mikrobiologii imeni pastera"] = "Tr Leningr Nauchnoissled Inst Epidemiol Mikrobiol",
  ["trudy leningradskogo nauchno-issledovatel’skogo instituta epidemiologii i mikrobiologii imeni pastera"] = "Tr Leningr Nauchnoissled Inst Epidemiol Mikrobiol",
  ["trudy leningradskogo sanitarno-gigienicheskogo meditsinskogo instituta"] = "Tr Leningr Sanitarnogig Med Inst",
  ["trudy matematicheskogo instituta imeni v"] = "Tr. Mat. Inst. Steklova",
  ["trudy matematicheskogo instituta imeni v. a. steklova"] = "Tr. Mat. Inst. Steklova",
  ["trudy matematicheskogo instituta imeni v. a. steklova. rossiiskaya akademiya nauk"] = "Tr. Mat. Inst. Steklova",
  ["trudy moskovskogo matematicheskogo obshchestva"] = "Tr. Mosk. Mat. Obs.",
  ["trudy moskovskogo nauchno-issledovatel'skogo instituta epidemiologii i mikrobiologii"] = "Tr Mosk Nauchnoissled Inst Epidmiol Mikrobiol",
  ["trudy moskovskogo nauchno-issledovatel’skogo instituta epidemiologii i mikrobiologii"] = "Tr Mosk Nauchnoissled Inst Epidmiol Mikrobiol",
  ["trudy petrozavodskogo gosudarstvennogo universiteta"] = "Tr. Petrozavodsk. Gos. Univ. Ser. Mat.",
  ["trudy russkogo èntomologičeskogo obŝestva"] = "Tr Russ Entomol Obs (2000)",
  ["trudy sankt-peterburgskogo matematicheskogo obshchestva"] = "Trudy S.-Peterburg. Mat. Obshch.",
  ["trudy seminara po kraevym zadacham"] = "Trudy Sem. Kraev. Zadacham",
  ["trudy seminara po vektornomu i tenzornomu analizu s ikh prilozheniyami k geometrii, mekhanike i fizike"] = "Trudy Sem. Vektor. Tenzor. Anal.",
  ["trudy seminara s. l. soboleva"] = "Trudy Sem. S. L. Soboleva",
  ["trudy zoologicheskogo instituta"] = "Tr Zool Inst",
  ["trudy, akademiia nauk sssr, institut genetiki"] = "Tr. Akad. Nauk SSSR. Inst. Genet.",
  ["trudy. akademiia nauk sssr. institut genetiki"] = "Tr Akad Nauk SSSR Inst Genet",
  ["trudy. institut genetiki (akademii︠a︡ nauk sssr)"] = "Tr Akad Nauk SSSR Inst Genet",
  ["trudy. institut grudnoi khirurgii (akademiia meditsinskikh nauk sssr)"] = "Tr Akad Med Nauk Sssr Mosc Inst Grudn Khir",
  ["trudy. institut grudnoĭ khirurgii (akademii︠a︡ medit︠s︡inskikh nauk sssr)"] = "Tr Akad Med Nauk Sssr Mosc Inst Grudn Khir",
  ["trudy. institut tuberkuleza (akademii︠a︡ medit︠s︡inskikh nauk sssr)"] = "Tr Inst Tuberk",
  ["trudy. mikrobiologijas instituts (latvijas psr zinatnu akademija)"] = "Tr Latv Padomju Soc Repub Zinat Akad Mikrobiol Inst",
  ["trudy. mikrobiologijas institūts (latvijas psr zinātnu akadēmija)"] = "Tr Latv Padomju Soc Repub Zinat Akad Mikrobiol Inst",
  ["trudy. moscow (russia). vsesoiuznyi nauchno-issledovatel’skii vitaminnyi institut"] = "Trudy Mosc Russ Vsesoiuznyi Nauchno Issled Vitam Inst",
  ["trudy. moscow (russia). vsesoiuznyĭ nauchno-issledovatel'skiĭ vitaminnyĭ institut"] = "Trudy Mosc Russ Vsesoiuznyi Nauchno Issled Vitam Inst",
  ["trudy. tsentral'nyĭ nauchno-issledovatel'skiĭ institut protezirovaniia i protezostroeniia (moscow, r.s.f.s.r.)"] = "Trudy Tsentralnyi Naucnoissled Inst Protez Protezostr (Mosc)",
  ["trudy. vsesoiuznoe fiziologicheskoe obshchestvo imeni i. p. pavlova. orenburgskoe otdelenie"] = "Trudy Vsesoiuznoe Fiziol Obshchestvo Im I P Pavlov Orenb Otd",
  ["trustee : the journal for hospital governing boards"] = "Trustee",
  ["trusts and estates"] = "Trusts Estates",
  ["trustworthy eternal systems via evolving software, data and knowledge : second international workshop, eternals 2012, montpellier, france, august 28, 2012, revised selected papers. eternals (workshop) (2nd : 2012 : montpellier, france)"] = "Trust Eternal Syst Via Evol Softw Data Knowl (2012)",
  ["trésors monétaires"] = "TMon",
  ["tschermaks mineralogische und petrographische mitteilungen"] = "Tschermaks Mineral. Petrogr. Mitt.",
  ["tsetse and trypanosomiasis information"] = "Tsetse Trypanos Inf",
  ["tsetse and trypanosomiasis information quarterly"] = "Tsetse Trypanos Inf Q",
  ["tsg : tijdschrift voor gezondheidswetenschappen"] = "TSG",
  ["tsinghua science and technology"] = "Tsinghua Sci. Technol.",
  ["tsitologiia i genetika"] = "Tsitol Genet",
  ["tsuchi to biseibutsu = soil microorganisms"] = "Tsuchi To Biseibutsu",
  ["tsukuba daigaku jinbun chirigaku kenkyu"] = "Tsukuba Daigaku Jinbun Chirigaku Kenkyu",
  ["tsukuba daigaku jinbun chirigaku kenkyū"] = "Tsukuba Daigaku Jinbun Chirigaku Kenkyu",
  ["tsukuba journal of mathematics"] = "Tsukuba J. Math.",
  ["tsurumi shigaku (tsurumi university dental journal)"] = "Tsurumi Shigaku",
  ["tsurumi shigaku. tsurumi university dental journal"] = "Tsurumi Shigaku",
  ["tsvetnaya metally"] = "Tsvetn. Met. (Moscow)",
  ["tsvetnye metally (moscow)"] = "Tsvetn. Met. (Moscow)",
  ["tu rang xue bao = acta pedologica sinica"] = "Tu Rang Xue Bao",
  ["tubercle and lung disease"] = "Tuber. Lung Dis.",
  ["tubercle and lung disease : the official journal of the international union against tuberculosis and lung disease"] = "Tuber Lung Dis",
  ["tuberculology and thoracic diseases"] = "Tuberculol Thorac Dis",
  ["tuberculosis (edinburgh, scotland)"] = "Tuberculosis (Edinb)",
  ["tuberculosis and airborne disease weekly"] = "Tuberc Airborne Dis Wkly",
  ["tuberculosis and respiratory diseases"] = "Tuberc Respir Dis (Seoul)",
  ["tuberculosis research and treatment"] = "Tuberc Res Treat",
  ["tuberkulëz i bolezni lëgkikh"] = "Tuberk Biolezni Legkih",
  ["tuberkulose und ihre grenzgebiete in einzeldarstellungen"] = "Tuberk. Grenzgeb. Einzeldarst.",
  ["tuberkuloz ve toraks"] = "Tuberk Toraks",
  ["tuberkulozis es tudobetegsegek"] = "Tuberk Tudobetegsegek",
  ["tuberkulozis kerdesei"] = "Tuberk Kerdesei",
  ["tuberkulózis és tüdöbetegségek"] = "Tuberk Tudobetegsegek",
  ["tuberkulózis kérdései"] = "Tuberk Kerdesei",
  ["tufts dental outlook"] = "Tufts Dent Outlook",
  ["tufts folia medica"] = "Tufts Folia Med",
  ["tufts health science review"] = "Tufts Health Sci Rev",
  ["tufts medical journal"] = "Tufts Med J",
  ["tugium : jahrbuch des staatsarchivs des kantons zug, des amtes für denkmalpflege und archäologie, des kantonalen museums für urgeschichte zug und der burg zug"] = "Tugium",
  ["tuhinga : records of the museum of new zealand te papa tongarewa"] = "Tuhinga",
  ["tulane law review"] = "Tulane Law Rev",
  ["tulsa law journal"] = "Tulsa Law J",
  ["tumor & microenvironment"] = "Tumor Microenviron",
  ["tumor biology"] = "Tumor Biol.",
  ["tumor microenvironment and therapy"] = "Tumor Microenviron Ther",
  ["tumor research"] = "Tumor Res",
  ["tumour biology"] = "Tumour Biol.",
  ["tumour biology : the journal of the international society for oncodevelopmental biology and medicine"] = "Tumour Biol",
  ["tumour virus research"] = "Tumour Virus Res",
  ["tung wu hsüeh tsa chih : dongwuxue zazhi"] = "Dongwuxue Zazhi",
  ["tunisian journal of mathematics"] = "Tunis. J. Math.",
  ["tunisian journal of plant protection"] = "Tunis J Plant Prot",
  ["tunisie medicale"] = "Tunis. Med.",
  ["tunnelling and underground space technology"] = "Tunnelling Underground Space Technol.",
  ["turk botanik dergisi"] = "Turkish journal of botany|Turk. J. Bot.",
  ["turk hemsireler dergisi"] = "Turk Hemsire Derg",
  ["turk hemsireler dergisi (turkish journal of nursing)"] = "Turk Hemsire. Derg.",
  ["turk hijiyen ve deneysel biyoloji dergisi (turkish bulletin of hygiene and experimental biology)"] = "Turk Hij. Deney. Biyol. Derg.",
  ["turk hijiyen ve deneysel biyoloji dergisi. turkish bulletin of hygiene and experimental biology"] = "Turk Hij Deney Biyol Derg",
  ["turk hijiyen ve tecrubi biyoloji dergisi"] = "Turk Hij. Tecr. Biyol. Derg.",
  ["turk hijiyen ve tecrubi biyoloji dergisi (turkish bulletin of hygiene and experimental biology)"] = "Turk Hij. Tecr. Biyol. Derg.",
  ["turk hijiyen ve tecrubi biyoloji dergisi. turkish bulletin of hygiene and experimental biology"] = "Turk Hij Tecr Biyol Derg",
  ["turk ortodonti dergisi"] = "Turk Ortodonti Derg.",
  ["turk tip cemiyeti mecmuasi"] = "Turk Tip Cemiy. Mecm.",
  ["turk tip dernegi dergisi"] = "Turk Tip Dernegi Derg.",
  ["turkic world mathematical society journal of applied and engineering mathematics"] = "Turk. World Math. Soc. J. App. Eng. Math.",
  ["turkisch journal of physics"] = "Turk. J. Phys.",
  ["turkish archives of otorhinolaryngology"] = "Turk Arch Otorhinolaryngol",
  ["turkish computational and theoretical chemistry"] = "Turk. Comput. Theor. Chem.",
  ["turkish journal of agriculture & forestry = türk tarım ve ormancılık dergisi"] = "Turk J Agric For",
  ["turkish journal of agriculture and forestry"] = "Turk. J. Agric. For.",
  ["turkish journal of anaesthesiology and reanimation"] = "Turk J Anaesthesiol Reanim",
  ["turkish journal of biochemistry"] = "Turk. J. Biochem.",
  ["turkish journal of biology"] = "Turk. J. Biol.",
  ["turkish journal of biology = türk biyoloji dergisi"] = "Turk J Biol",
  ["turkish journal of botany"] = "Turk J Botany",
  ["turkish journal of chemistry"] = "Turk. J. Chem.",
  ["turkish journal of civil engineering"] = "Turk. J. Civ. Eng.",
  ["turkish journal of computer and mathematics education"] = "Turk. J. Comput. Math. Educ.",
  ["turkish journal of earth sciences"] = "Turk. J. Earth Sci.",
  ["turkish journal of electrical engineering and computer sciences"] = "Turk. J. Electr. Eng. Comput. Sci.",
  ["turkish journal of emergency medicine"] = "Turk J Emerg Med",
  ["turkish journal of engineering and environmental science"] = "Turk. J. Eng. Environ. Sci.",
  ["turkish journal of field crops"] = "Turk. J. Field Crops",
  ["turkish journal of fisheries and aquatic sciences"] = "Turk J Fish Aquat Sci",
  ["turkish journal of haematology : official journal of turkish society of haematology"] = "Turk J Haematol",
  ["turkish journal of hematology"] = "Turk. J. Hematol.",
  ["turkish journal of immunology"] = "Turk. J. Immunol.",
  ["turkish journal of mathematics"] = "Turkish J. Math.",
  ["turkish journal of medical sciences"] = "Turk J Med Sci",
  ["turkish journal of obstetrics and gynecology"] = "Turk J Obstet Gynecol",
  ["turkish journal of ophthalmology"] = "Turk J Ophthalmol",
  ["turkish journal of orthodontics"] = "Turk J Orthod",
  ["turkish journal of pathology"] = "Turk. J. Pathol.",
  ["turkish journal of pediatrics"] = "Turk. J. Pediatr.",
  ["turkish journal of pharmaceutical sciences"] = "Turk J Pharm Sci",
  ["turkish journal of physical medicine and rehabilitation"] = "Turk J Phys Med Rehabil",
  ["turkish journal of physics"] = "Turk. J. Phys.",
  ["turkish journal of physiotherapy and rehabilitation"] = "Turk. J. Physiother. Rehabil.",
  ["turkish journal of surgery"] = "Turk J Surg",
  ["turkish journal of urology"] = "Turk J Urol",
  ["turkish journal of veterinary and animal sciences"] = "Turk J Vet Anim Sci",
  ["turkish journal of zoology"] = "Turk. J. Zool.",
  ["turkish neurosurgery"] = "Turk Neurosurg",
  ["turkish thoracic journal"] = "Turk Thorac J",
  ["turkishi journal of endocrinology and metabolism"] = "Turk. J. Endocrinol. Metab.",
  ["turtle island journal of indigenous health"] = "Turt Isl J Indig Health",
  ["turtox news"] = "Turtox News",
  ["turun historiallinen arkisto"] = "Turun Hist Ark (Finl)",
  ["tutkimuksia—research reports"] = "Tutkimuksia—Res. Rep.",
  ["tutmondaj sciencoj kaj teknikoj"] = "Tutmondaj Sciencoj Tek",
  ["tutorials in quantitative methods for psychology"] = "Tutor Quant Methods Psychol",
  ["tutorials, schools, and workshops in the mathematical sciences"] = "Tutor. Sch. Workshops Math. Sci.",
  ["tüberküloz ve toraks"] = "Tuberk Toraks",
  ["türk biyokimya dergisi = turkish journal of biochemistry"] = "Turk Biyokim Derg",
  ["türk geriatri dergisi"] = "Turk Geriatri Derg",
  ["türk göğüs kalp damar cerrahisi dergisi"] = "Turk Gogus Kalp Damar Cerrahisi Derg",
  ["türk hemşireler dergisi"] = "Turk Hemsire Derg",
  ["türk hijiyen ve deneysel biyoloji dergisi. turkish bulletin of hygiene and experimental biology"] = "Turk Hij Deney Biyol Derg",
  ["türk hijiyen ve tecrübi biyoloji dergisi"] = "Turk Hij Tecr Biyol Derg",
  ["türk hijiyen ve tecrübî biyoloji dergisi. turkish bulletin of hygiene and experimental biology"] = "Turk Hij Tecr Biyol Derg",
  ["türk kardiyoloji derneği arşivi : türk kardiyoloji derneğinin yayın organıdır"] = "Turk Kardiyol Dern Ars",
  ["türk mikrobiyoloji cemiyeti dergisi"] = "Turk Mikrobiyol Cemiy Derg",
  ["türk nöroloji dergisi = turkish journal of neurology"] = "Turk Noroloji Dergisi",
  ["türk oftalmolojı dergısı"] = "Turk Oftalmol Derg",
  ["türk ortodonti dergisi : ortodonti derneğ'nin resmi yayin organidir = turkish journal of orthodontics"] = "Turk Ortodonti Derg",
  ["türk patoloji dergisi"] = "Turk Patoloji Derg",
  ["türk pediatri arşivi"] = "Turk Pediatri Ars",
  ["türk psikiyatri dergisi = turkish journal of psychiatry"] = "Turk Psikiyatri Derg",
  ["türk silahlı kuvvetleri koruyucu hekimlik bülteni"] = "Turk Silahlı Kuvvetleri Koruyucu Hekim Bul",
  ["türk tarım gıda bilim ve teknoloji dergisi"] = "Turk Tarim Gida Bilim Teknol Derg",
  ["türk tip cemīyetī mecmuasi"] = "Turk Tip Cemiy Mecm",
  ["türk tip derneğī dergīsī"] = "Turk Tip Dernegi Derg",
  ["türk zooloji dergisi = turkish journal of zoology"] = "Turk Zool Derg",
  ["türkiye entomoloji dergisi"] = "Turk Entomol Derg",
  ["türkiye klinikleri. immünoloji alerji"] = "Turk Klin Immunol Alerji",
  ["türkiye klinikleri. tıp etiği-hukuku-tarihi = türkiye klinikleri journal of medical ethics, law, and history"] = "Turk Klin Tip Etigi Hukuku Tarihi",
  ["türkiye parazitolojii dergisi"] = "Turkiye Parazitol Derg",
  ["türkiye tarımsal araştırmalar dergisi"] = "Turk Tarim Arast Derg",
  ["tvz : het vakblad voor de verpleging"] = "TVZ",
  ["twentieth-century china = er shi shi ji zhongguo"] = "Twent Century China",
  ["twin research"] = "Twin Res.",
  ["twin research : the official journal of the international society for twin studies"] = "Twin Res",
  ["twin research and human genetics"] = "Twin Res. Hum. Genet.",
  ["twin research and human genetics : the official journal of the international society for twin studies"] = "Twin Res Hum Genet",
  ["twms journal of applied and engineering mathematics"] = "TWMS J. App. Eng. Math.",
  ["twms journal of pure and applied mathematics"] = "TWMS J. Pure Appl. Math.",
  ["tyche. beiträge zur alten geschichte, papyrologie und epigraphik"] = "Tyche",
  ["tydskrif vir hedendaagse romeins-holandse reg"] = "Tydskrif Hedendaagse Romeins Holandse Reg",
  ["tydskrif vir rasse-aangeleenthede. journal of racial affairs"] = "TydsKr Rasse Aangel",
  ["tydskrif vir studies in ekonomie en ekonometrie = journal for studies in economics and econometrics"] = "Tydskr Stud Ekon Ekon",
  ["tzu chi medical journal"] = "Tzu Chi Med J",
  ["tätigkeitsbericht der naturforschenden gesellschaft baselland"] = "Tätigk.ber. Nat.forsch. Ges. Baselland",
  ["tübinger atlas des vorderen orients"] = "TAVO",
  ["tübinger geographische studien"] = "Tüb. Geogr. Stud.",
  ["türk arkeoloji dergisi"] = "TAD",
  ["türk tarih kurumu yayınları"] = "TTKY",
  ["türkiye bilimler akademisi arkeoloji dergisi"] = "TueBA-Ar",
  ["tıp tarihi araştırmaları = history of medicine studies"] = "Tip Tarihi Arastirmalari",
  ["tự nhiên và công nghệ : khoa học"] = "Tu Nhien Va Cong Nghe",
  ["tʻung pao. tʻoung pao"] = "Toung Pao",
  ["t’ung pao. t’oung pao"] = "Toung Pao",
  ["t︠s︡itologii︠a︡ i genetika"] = "Tsitol Genet",
  ["u m c"] = "U M C Cordoba",
  ["u. s. medicine"] = "US Med",
  ["u. s. navy medicine"] = "US Navy Med",
  ["u. s. pharmacist"] = "US Pharm",
  ["u. thieme – f. becker (hrsg.), allgemeines lexikon der bildenden künstler"] = "Thieme-Becker",
  ["u.c. davis law review. university of california, davis. school of law"] = "UC Davis Law Rev",
  ["u.p.b. scientific bulletin series a applied mathematics and physics"] = "U.P.B. Sci. Bull. Ser. A",
  ["u.p.b. scientific bulletin series b chemistry and materials science"] = "U.P.B. Sci. Bull. Ser. B",
  ["u.p.b. scientific bulletin series c electrical engineering and computer science"] = "U.P.B. Sci. Bull. Ser. C",
  ["u.p.b. scientific bulletin series d mechanical engineering"] = "U.P.B. Sci. Bull. Ser. D",
  ["u.s. army medical department journal"] = "US Army Med Dep J",
  ["u.s. catholic"] = "US Cathol",
  ["u.s. census bureau center for economic studies research paper series"] = "US Census Bur Cent Econ Stud Res Pap Ser",
  ["u.s. chinese journal of lymphology and oncology"] = "U S Chin J Lymphology Oncol",
  ["u.s. geological survery water resources  investigations report"] = "U.S. Geol. Surv. Water Resour. Invest. Rep.",
  ["u.s. geological survey bulletin"] = "U.S. Geol. Surv. Bull.",
  ["u.s. geological survey professional paper"] = "U.S. Geol. Surv. Prof. Pap.",
  ["u.s. geological survey scientific investigations report"] = "U.S. Geol. Surv. Sci. Invest. Rep.",
  ["u.s. healthcare"] = "US Healthc",
  ["u.s. long-term review"] = "US Long Term Rev",
  ["u.s. medicine"] = "US. Med.",
  ["u.s. navy medicine"] = "US Navy Med.",
  ["u.s. news & world report"] = "US News World Rep",
  ["u.s.-japan women's journal. english supplement = nichi-bei josei jānaru. english supplement"] = "US Jpn Womens J Engl Suppl",
  ["ubersichten zur tierernahrung"] = "Ubersich. Tierernahr.",
  ["ubiquitous computing, electronics & mobile communication conference (uemcon), ieee annual"] = "Ubiquitous Comput Electron Mob Commun Conf (UEMCON) IEEE Annu",
  ["ucd [reports]. u.s. atomic energy commission"] = "UCD Rep",
  ["uchenye zapiski fizicheskogo fakulʹteta mgu"] = "Uchenye Zap Fiz Fak MGU",
  ["uchenye zapiski instituta farmakologii i khimioterapii amn sssr"] = "Uch Zap Inst Farmakol Khimioter AMN SSSR",
  ["uchenye zapiski kazanskogo universiteta. seriya fiziko-matematicheskie nauki"] = "Uch. Zap. Kazan. Univ. Ser. Fiz.-Mat. Nauki",
  ["uchenye zapiski tavricheskogo nat︠s︡ionalʹnogo universiteta im. v.i. vernadskogo"] = "Uc Zap Tavriceskogo Nac Univ Im V I Vemadskogo",
  ["uchenye zapiski tsentralnogo aero-gidrodinamicheskogo instituta (tsagi)"] = "Uchen. Zap. TsAGI",
  ["uchenye zapiski. artsakhskii gosudarstvennyi universitet"] = "Uch. Zap. Artsakh. Gos. Univ.",
  ["uchenye zapiski. moscow (russia). moskovskiĭ gorodskoĭ pedagogicheskiĭ institut"] = "Uchenye Zap (Mosc)",
  ["uchenye zapiski. seriia biologicheskaia. leningrad. universitet"] = "Uchenye Zap Seriia Biolog Leningr Univ",
  ["uchenye zapiski. vtoroi moskovskii ordena lenina gosudarstvennyi meditsinskii institut im n.i. pirogova"] = "Uchenye Zap Mosk Meditsinskii Inst",
  ["uchenye zapiski. vtoroĭ moskovskiĭ ordena lenina gosudarstvennyĭ medit︠s︡inskiĭ institut im n.i. pirogova"] = "Uchenye Zap Mosk Meditsinskii Inst",
  ["uchu koku kankyo igaku / nihon uchu koku kankyo igakkai"] = "Uchu Koku Kankyo Igaku",
  ["uchū kōkū kankyō igaku"] = "Uchu Koku Kankyo Igaku",
  ["uchū seibutsu kagaku"] = "Biol Sci Space",
  ["ucl open"] = "UCL Open",
  ["ucl open environment"] = "UCL Open Environ",
  ["ucla [reports]. u.s. atomic energy commission"] = "UCLA Rep",
  ["ucla forum in medical sciences"] = "UCLA Forum Med Sci",
  ["ucla law review. university of california, los angeles. school of law"] = "UCLA Law Rev",
  ["ucla pacific basin law journal"] = "UCLA Pac Basin Law J",
  ["ucla reports"] = "UCLA Rep.",
  ["ucla symposia on molecular and cellular biology"] = "UCLA Symp Mol Cell Biol",
  ["ucla symposia on molecular and cellular biology, new series"] = "UCLA Symp. Mol. Cell. Bio., New Ser.",
  ["ucla women's law journal"] = "UCLA Womens Law J",
  ["ucla women’s law journal"] = "UCLA Womens Law J",
  ["ucrl [reports]. u.s. atomic energy commission"] = "UCRL US At Energy Comm",
  ["uddannelse (copenhagen, denmark)"] = "Uddannelse",
  ["ufa mathematical journal"] = "Ufa Math. J.",
  ["ufsi reports"] = "UFSI Rep",
  ["uganda medical journal"] = "Uganda Med J",
  ["ugarit-forschungen. internationales jahrbuch für die altertumskunde syrien-palästinas"] = "UF",
  ["ugeskrift for laeger"] = "Ugeskr Laeger",
  ["ui journal"] = "UI J",
  ["ui sahak"] = "Uisahak",
  ["uirusu. journal of virology"] = "Uirusu",
  ["ujcd. union des jeunes chirurgiens-dentistes"] = "UJCD Union Jeunes Chir Dent",
  ["ujcd; union des jeunes chirurgiens-dentistes"] = "UJCD. Union Jeunes Chir. Dent.",
  ["uk energy research centre"] = "UK Energy Res. Cent.",
  ["ukrainian antarctic journal"] = "Ukr. Antarct. J.",
  ["ukrainian biochemical journal"] = "Ukr. Biochem. J.",
  ["ukrainian botanical journal"] = "Ukr. Botan. J.",
  ["ukrainian economic review"] = "Ukrainian Econ. Rev.",
  ["ukrainian food journal"] = "Ukr. Food J.",
  ["ukrainian journal of ecology"] = "Ukr J Ecol",
  ["ukrainian journal of food science"] = "Ukr. J. Food Sci.",
  ["ukrainian journal of forest and wood science"] = "Ukr. J. For. Wood Sci.",
  ["ukrainian journal of physical optics"] = "Ukr. J. Phys. Opt.",
  ["ukrainian journal of physics"] = "Ukr. J. Phys.",
  ["ukrainian journal of radiology and oncology"] = "Ukr. J. Radiol. Oncol.",
  ["ukrainian mathematical journal"] = "Ukrainian Math. J.",
  ["ukrainian mathematics journal"] = "Ukr. Math. J.",
  ["ukrainian physics journal [translation of ukrainskii fizicheskii zhurnal (russian edition)]"] = "Ukr. Phys. J.",
  ["ukrainian review (london, england)"] = "Ukr Rev (Lond)",
  ["ukrainica bioorganica acta"] = "Ukr. Bioorg. Acta",
  ["ukrainskii biokhimicheskii zhurnal"] = "Ukr Biokhim Zh",
  ["ukrainskii fizicheskii zhurnal"] = "Ukr. Fiz. Zh.",
  ["ukrainskii fizicheskii zhurnal (russian edition)"] = "Ukr. Fiz. Zh. (Russ. Ed.)",
  ["ukrainskii khimicheskii zhurnal"] = "Ukr. Khim. Zh.",
  ["ukrainskii khimicheskii zhurnal (russian edition)"] = "Ukr. Khim. Zh. (Russ. Ed.)",
  ["ukrainskii matematichnii visnik"] = "Ukr. Mat. Visn.",
  ["ukrainskiĭ biokhimicheskiĭ zhurnal (1978)"] = "Ukr Biokhim Zh (1978)",
  ["ukrainskiy geometricheskiy sbornik"] = "Ukrain. Geom. Sb.",
  ["ukrainskyi biokhimichnyi zhurnal"] = "Ukr. Biokhim. Zh.",
  ["ukrains’kii biokhimichnii zhurnal"] = "Ukr. Biokhim. Zh.",
  ["ukrains’kii fizichnii zhurnal"] = "Ukr. Fiz. Zh.",
  ["ukrains’kyi biokhimichnyi zhurnal"] = "Ukr Biokhim Zh",
  ["ukrains’kyi botanichnyi zhurnal. ukrainian botanical review. ukrains’ke botanichne tovarystvo"] = "Ukr Bot Z Ukr Bot Tovarystvo",
  ["ukrains’kyi istorychnyi zhurnal"] = "Ukr Istor Z",
  ["ukraïns'kyĭ biokhimichnyĭ zhurnal"] = "Ukr Biokhim Zh",
  ["ukraïns'kyǐ botanichnyǐ z︠h︡urnal (kiev, u.r.s.r. : 1956)"] = "Ukr Bot Z",
  ["ukraïnsʹkyĭ biokhimichnyĭ z︠h︡urnal (1999 )"] = "Ukr Biokhim Zh (1999)",
  ["ukraïnsʹkyĭ botanichnyĭ z︠h︡urnal. ukrainian botanical review. ukraïnsʹke botanichne tovarystvo"] = "Ukr Bot Z Ukr Bot Tovarystvo",
  ["ukraïnsʹkyĭ istorychnyĭ zhurnal"] = "Ukr Istor Z",
  ["ulam quarterly"] = "Ulam Quart.",
  ["ulster folklife"] = "Ulster Folklife",
  ["ulster medical journal"] = "Ulster Med. J.",
  ["ultra scientist of physical sciences"] = "Ultra Sci. Phys. Sci.",
  ["ultrafast science"] = "Ultrafast Sci.",
  ["ultraschall in der medizin"] = "Ultraschall Med.",
  ["ultraschall in der medizin (stuttgart, germany : 1980)"] = "Ultraschall Med",
  ["ultrasonic imaging"] = "Ultrason Imaging",
  ["ultrasonics sonochemistry"] = "Ultrason Sonochem",
  ["ultrasonography (seoul, korea)"] = "Ultrasonography",
  ["ultrasound (leeds, england)"] = "Ultrasound",
  ["ultrasound clinics"] = "Ultrasound Clin",
  ["ultrasound in medicine & biology"] = "Ultrasound Med Biol",
  ["ultrasound in medicine and biology"] = "Ultrasound Med. Biol.",
  ["ultrasound in obstetrics & gynecology : the official journal of the international society of ultrasound in obstetrics and gynecology"] = "Ultrasound Obstet Gynecol",
  ["ultrasound in obstetrics and gynecology"] = "Ultrasound Obstet. Gynecol.",
  ["ultrasound international open"] = "Ultrasound Int Open",
  ["ultrasound journal"] = "Ultrasound J.",
  ["ultrasound quarterly"] = "Ultrasound Q",
  ["ultrastructural pathology"] = "Ultrastruct Pathol",
  ["ulusal cerrahi dergisi"] = "Ulus Cerrahi Derg",
  ["ulusal travma dergisi = turkish journal of trauma & emergency surgery : tjtes"] = "Ulus Travma Derg",
  ["ulusal travma ve acil cerrahi dergisi (turkish journal of trauma and emergency surgery)"] = "Ulus. Travma Acil Cerrahi Derg.",
  ["ulusal travma ve acil cerrahi dergisi = turkish journal of trauma & emergency surgery : tjtes"] = "Ulus Travma Acil Cerrahi Derg",
  ["umkc law review"] = "UMKC Law Rev",
  ["umschau in wissenschaft und technik"] = "Umsch Wiss Tech",
  ["umschau, die"] = "Umschau",
  ["umwelt und gesundheit online"] = "Umw Gesundh Online",
  ["umweltmedizine in forschung und praxis"] = "Umweltmed. Forsch. Prax.",
  ["umweltrecht in der praxis"] = "Umweltrecht Prax.",
  ["umweltschutz in der schweiz"] = "Umweltschutz Schweiz",
  ["umweltwissenschaften und schadstoff-forschung"] = "Umweltwiss. Schadst. Forsch.",
  ["un chronicle"] = "UN Chron",
  ["un monthly chronicle"] = "UN Mon Chron",
  ["una communique"] = "UNA Commun.",
  ["una communiqué"] = "UNA Commun",
  ["una nursing journal"] = "UNA Nurs J",
  ["uncertainty and operations research"] = "Uncertain. Oper. Res.",
  ["uncertainty for safe utilization of machine learning in medical imaging and clinical image-based procedures : first international workshop, unsure 2019, and 8th international workshop, clip 2019, held in conjunction with miccai 2019, shenzhen, china, october 17, 2019, proceedings. unsure (workshop) (1st : 2019 : shenzhen shi, china)"] = "Uncertain Safe Util Machine Learn Med Imaging Clin Image Based Proced (2019)",
  ["uncertainty for safe utilization of machine learning in medical imaging, and graphs in biomedical image analysis : second international workshop, unsure 2020, and third international workshop, grail 2020, held in conjunction with miccai 2020, lima, peru, october 8, 2020, proceedings"] = "Uncertain Safe Util Mach Learn Med Imaging Graph Biomed Image Anal (2020)",
  ["uncertainty, computational techniques, and decision intelligence"] = "Uncertain. Comput. Tech. Decis. Intell.",
  ["unconventional resources"] = "Unconv. Resour.",
  ["unctad review"] = "UNCTAD Rev.",
  ["undergraduate lecture notes in physics"] = "Undergrad. Lect. Notes Phys.",
  ["undergraduate texts in computer science"] = "Undergrad. Texts Comput. Sci.",
  ["undergraduate texts in mathematics"] = "Undergrad. Texts Math.",
  ["undergraduate texts in mathematics. readings in mathematics"] = "Undergrad. Texts Math. Read. Math.",
  ["undergraduate texts in physics"] = "Undergrad. Texts Phys.",
  ["undergraduate topics in computer science"] = "Undergrad. Top. Comput. Sci.",
  ["underground space"] = "Underground Space",
  ["undersea & hyperbaric medicine : journal of the undersea and hyperbaric medical society, inc"] = "Undersea Hyperb Med",
  ["undersea and hyperbaric medicine"] = "Undersea Hyperb. Med.",
  ["undersea biomedical research"] = "Undersea Biomed Res",
  ["understanding and interpreting machine learning in medical image computing applications : first international workshops, mlcn 2018, dlf 2018, and imimic 2018, held in conjunction with miccai 2018, granada, spain, september 16-20, 2018 : proceedings. mlcn (workshop) (1st : 2018 : granada, spain)"] = "Underst Interpret Mach Learn Med Image Comput Appl (2018)",
  ["understanding chemical reactivity"] = "Underst. Chem. React.",
  ["understanding complex systems"] = "Underst Complex Syst",
  ["understanding interventions that broaden participation in research careers"] = "Unders Interv Broaden Particip Res Careers",
  ["understanding the child; a magazine for teachers"] = "Underst Child",
  ["undp news : networking publication of undp staff worldwide. united nations development programme"] = "UNDP News",
  ["unesco bulletin for libraries"] = "Unesco Bull Libr",
  ["unesco courier (paris, france : 1989)"] = "Unesco Cour",
  ["unesco sources"] = "Unesco Sources",
  ["unicef news"] = "UNICEF News",
  ["unidad sanitaria"] = "Unidad Sanit",
  ["uniform distribution theory"] = "Unif. Distrib. Theory",
  ["union medicale du canada"] = "Union Med. Can.",
  ["union of burma journal of life sciences"] = "Union Burma J Life Sci",
  ["union seminary quarterly review"] = "USQR Union Semin Q Rev",
  ["unione matematica italiana"] = "Boll. Un. Mat. Ital. A (7)",
  ["unipa springer series"] = "UNIPA Springer Ser.",
  ["unitas (helsinki, finland : english edition)"] = "Unitas",
  ["united european gastroenterology journal"] = "United European Gastroenterol J",
  ["united evangelical action"] = "United Evangelical Action",
  ["united nations monthly chronicle"] = "UN Mon. Chron.",
  ["united states armed forces medical journal"] = "U S Armed Forces Med J",
  ["united states code annotated. united states"] = "U S Code Annot U S",
  ["united states department of agriculture bulletin"] = "U.S. Dep. Agric. Bull.",
  ["united states department of agriculture miscellaneous publication"] = "U.S. Dep. Agric. Misc. Publ.",
  ["united states naval medical bulletin"] = "U S Nav Med Bull",
  ["united states navy medical news letter"] = "US Navy Med News Lett",
  ["united states statutes at large"] = "US Statut Large",
  ["united states supreme court reports. united states. supreme court"] = "US Supreme Court Rep",
  ["united synagogue review"] = "United Synag Rev",
  ["unitext for physics"] = "Unitext Phys.",
  ["universal access in human-computer interaction : 5th international conference, uahci 2009, held as part of hci international 2009, san diego, ca, usa, july 19-24, 2009 : proceedings. international conference on universal access in human-computer interaction (5th : 2009 : san diego, ca)"] = "Univers Access Hum Comput Interact (2009)",
  ["universal access in human-computer interaction : designing novel interactions : 11th international conference, uahci 2017, held as part of hci international 2017, vancouver, bc, canada, july 9-14, 2017, proceedings. part ii. international conference on universal access in human-computer interaction (11th : 2017 : vancouver, b.c.)"] = "Univers Access Hum Comput Interact (2017)",
  ["universal access in the information society"] = "Univers Access Inf Soc",
  ["universal journal of agricultural research"] = "Univers J Agric Res",
  ["universal journal of electrical and electronic engineering"] = "Univers. J. Electr. Electron. Eng.",
  ["universal journal of microbiology research"] = "Univers J Microbiol Res",
  ["universal journal of public health"] = "Univers J Public Health",
  ["universale economica"] = "Universale Econom.",
  ["universidad del zulia"] = "Ciencia",
  ["universidad nacional de tucumán"] = "Univ. Nac. Tucumán Rev. Ser. A",
  ["universidad y salud"] = "Univ Salud",
  ["universitas (university of ghana)"] = "Universitas (Ghana)",
  ["universitas odontológica : revista cientifica de la facultad de odontológica"] = "Univ Odontol",
  ["universitas scientiarum"] = "Univ Sci (Bogota)",
  ["universitat de barcelona"] = "Collect. Math.",
  ["universitatis babeș-bolyai"] = "Studia Univ. Babeș-Bolyai Math.",
  ["universitatis debreceniensis"] = "Publ. Math. Debrecen",
  ["universitatis iagellonicae"] = "Univ. Iagel. Acta Math.",
  ["universitatis iagellonicae. acta mathematica"] = "Univ. Iagel. Acta Math.",
  ["universitatis masarykianae brunensis"] = "Arch. Math. (Brno)",
  ["universitatis masarykianae brunensis. facultas scientiarum naturalium. archivum mathematicum"] = "Arch. Math. (Brno)",
  ["universitats-lehr- und studienbucher. naturwissenschaften"] = "Univ. Lehr. Stud. Naturwissenschaft.",
  ["universite de grenoble. annales de l’institut fourier"] = "Ann. Inst. Fourier (Grenoble)",
  ["universitet po arkhitektura, stroitelstvo i geodeziya sofiya"] = "God. Univ. Arkhit. Stroit. Geod. Sofiya Svit\\cdprime k II Mat. Mekh.",
  ["university chemistry education"] = "Univ. Chem. Educ.",
  ["university hospital bulletin"] = "Univ Hosp Bull",
  ["university lecture series"] = "Univ. Lecture Ser.",
  ["university of alexandria"] = "Bull. Fac. Sci. Alexandria Univ.",
  ["university of ankara"] = "Comm. Fac. Sci. Univ. Ankara Ser. A1 Math. Statist.",
  ["university of arizona bulletin"] = "Univ. Ariz. Bull.",
  ["university of arkansas at little rock law journal"] = "Univ Ark Little Rock Law J",
  ["university of baltimore law review. university of baltimore. school of law"] = "Univ Baltimore Law Rev",
  ["university of british columbia law review. university of british columbia"] = "Univ B C Law Rev",
  ["university of california publications in pharmacology"] = "Univ Calif Publ Pharmacol",
  ["university of california publications in physiology"] = "Univ Calif Publ Physiol",
  ["university of california publications in zoology"] = "Univ Calif Publ Zool",
  ["university of california, davis law review"] = "Univ Calif Davis Law Rev",
  ["university of cincinnati law review. university of cincinnati. college of law"] = "Univ Cincinnati Law Rev",
  ["university of colorado law review. university of colorado (boulder campus). school of law"] = "Univ Colo Law Rev",
  ["university of dayton law review. university of dayton. law school"] = "Univ Dayton Law Rev",
  ["university of detroit journal of urban law"] = "Univ Detroit J Urban Law",
  ["university of detroit law review"] = "Univ Detroit Law Rev",
  ["university of detroit mercy law review"] = "Univ Detroit Mercy Law Rev",
  ["university of durham medical gazette"] = "Univ Durh Med Gaz",
  ["university of edinburgh journal"] = "Univ Edinb J",
  ["university of florida journal of law and public policy : jl & pp"] = "Univ Fla J Law Public Policy",
  ["university of florida law review"] = "Univ Fla Law Rev",
  ["university of ghana law journal"] = "Univ Ghana Law J",
  ["university of illinois law review"] = "Univ Ill Law Rev",
  ["university of joensuu, department of mathematics report series"] = "Univ. Joensuu Dept. Math. Rep. Ser.",
  ["university of joensuu, department of physics and mathematics. report series"] = "Univ. Joensuu Dep. Phys. Math. Rep. Ser.",
  ["university of jordan publications"] = "Univ. Jordan Publ.",
  ["university of kansas law review. university of kansas. school of law"] = "Univ Kans Law Rev",
  ["university of leeds medical magazine"] = "Univer Leeds Med Mag",
  ["university of lodz. department of logic. bulletin of the section of logic"] = "Bull. Sect. Logic Univ. Lodz",
  ["university of louisville journal of family law"] = "Univ Louisv J Fam Law",
  ["university of louisville journal of family law / university of louisville school of law"] = "Univ Louisv J Fam Law",
  ["university of miami entertainment & sports law review"] = "Univ Miami Entertain Sports Law Rev",
  ["university of miami law review"] = "Univ Miami Law Rev",
  ["university of michigan journal of law reform. university of michigan. law school"] = "Univ Mich J Law Reform",
  ["university of michigan medical center journal"] = "Univ Mich Med Cent J",
  ["university of minnesota medical bulletin"] = "Univ Minn Med Bull",
  ["university of missouri business and government review"] = "Univ. Missouri Bus. Govt. Rev.",
  ["university of new mexico orthopaedics research journal"] = "Univ N M Orthop Res J",
  ["university of newcastle upon tyne medical gazette"] = "Univ Newcastle Tyne Med Gaz",
  ["university of niš"] = "Facta Univ. Ser. Mech. Automat. Control Robot.",
  ["university of north carolina news letter"] = "Univ N C News Lett",
  ["university of oulu report series in physical sciences"] = "Univ. Oulu Rep. Ser. Phys. Sci.",
  ["university of pennsylvania journal of constitutional law"] = "Univ Pa J Const Law",
  ["university of pennsylvania law review"] = "Univ PA Law Rev",
  ["university of pennsylvania medical bulletin"] = "Univ Pa Med Bull",
  ["university of pittsburgh law review. university of pittsburgh. school of law"] = "Univ Pittsbg Law Rev",
  ["university of puget sound law review. university of puget sound. school of law"] = "Univ Puget Sound Law Rev",
  ["university of richmond law review. university of richmond"] = "Univ Richmond Law Rev",
  ["university of san fernando valley law review"] = "Univ San Fernando Valley Law Review",
  ["university of san francisco law review. university of san francisco. school of law"] = "Univ San Francisco Law Rev",
  ["university of tasmania law review"] = "Univ Tasman Law Rev",
  ["university of tehran science and humanities series"] = "Univ. Tehran Sci. Humanit. Ser.",
  ["university of the ryukyus"] = "Bull. College Sci. Univ. Ryukyus",
  ["university of toledo law review. university of toledo. college of law"] = "Univ Toledo Law Rev",
  ["university of toronto dental journal"] = "Univ Tor Dent J",
  ["university of toronto faculty of law review"] = "Univ Tor Fac Law Rev",
  ["university of toronto medical journal"] = "Univ Toronto Med J",
  ["university of toronto studies. biological series"] = "Univ Tor Stud Biol Ser",
  ["university of toronto undergraduate dental journal"] = "Univ Toronto Undergrad Dent J",
  ["university of washington business review"] = "Univ. Wash. Bus. Rev.",
  ["university of west los angeles law review"] = "Univ West Los Angel Law Rev",
  ["university of western ontario law review"] = "Univ West Ont Law Rev",
  ["university of western ontario medical journal"] = "Univ West Ont Med J",
  ["university of western ontario series in philosophy of science"] = "Univ. Western Ontario Ser. Philos. Sci.",
  ["university of wyoming publications (laramie, wyo. : 1935)"] = "Univ Wyo Publ",
  ["university of łódź"] = "Bull. Sect. Logic Univ. Łódź",
  ["university politehnica of bucharest scientific bulletin\tseries a applied mathematics and physics"] = "Univ. Politeh. Bucharest Sci. Bull. Ser. A",
  ["university politehnica of bucharest scientific bulletin\tseries b chemistry and materials science"] = "Univ. Politeh. Bucharest Sci. Bull. Ser. B",
  ["university politehnica of bucharest scientific bulletin\tseries c electrical engineering and computer science"] = "Univ. Politeh. Bucharest Sci. Bull. Ser. C",
  ["university politehnica of bucharest scientific bulletin\tseries d mechanical engineering"] = "Univ. Politeh. Bucharest Sci. Bull. Ser. D",
  ["university texts in the mathematical sciences"] = "Univ. Texts Math. Sci.",
  ["università e politecnico di torino"] = "Rend. Sem. Mat. Univ. Politec. Torino",
  ["universitätsforschungen zur prähistorischen archäologie"] = "UPA",
  ["université de grenoble"] = "Ann. Inst. Fourier (Grenoble)",
  ["universităţii din timișoara"] = "An. Univ. Timișoara Ser. Mat.-Inform.",
  ["universităţii tehnice din timișoara"] = "Bul. Ştiinţ. Univ. Tehn. Timișoara Mat. Fiz.",
  ["univerzitet u beogradu"] = "Univ. Beograd. Publ. Elektrotehn. Fak. Ser. Mat.",
  ["univerzitet u kragujevcu"] = "Zb. Rad. (Kragujevac)",
  ["univerzitet u nisu. prirodno-matematicki fakultet. filomat"] = "Filomat",
  ["univerzitet u nišu"] = "Filomat",
  ["uniwersytet im"] = "Funct. Approx. Comment. Math.",
  ["uniwersytet im. adama mickiewicza w poznaniu. wydzial matematyki i informatyki. functiones et approximatio commentarii mathematici"] = "Funct. Approx. Comment. Math.",
  ["unmanned systems"] = "Unmanned Syst.",
  ["unos update"] = "UNOS Update",
  ["unscn nutrition"] = "UNSCN Nutr",
  ["unser wald"] = "Unser Wald",
  ["unsere heimat"] = "Unsere Heim",
  ["unsolved problems in intuitive mathematics"] = "Unsolved Probl. in Intuitive Math.",
  ["unsupervised and semi-supervised learning"] = "Unsupervised and Semi-Supervised Learn.",
  ["untertoggenburger neujahrsblätter"] = "Untertoggenburg. Neujahrsbl.",
  ["untimely meditations"] = "Untimely Medit.",
  ["update (council for accreditation in occupational hearing conservation)"] = "Update Counc Accredit Occup Hearing Conserv",
  ["update (loma linda university. ethics center)"] = "Update",
  ["update (music educators national conference (u.s.))"] = "Update Univ S C Dep Music",
  ["update (national minority aids council)"] = "Update Natl Minor AIDS Counc",
  ["update in pediatric dentistry"] = "Update Pediatr Dent",
  ["update on cancer therapeutics"] = "Update Cancer Ther",
  ["update on ethics"] = "Update Ethics",
  ["updates in surgery"] = "Updates Surg",
  ["upravlenie bolshimi sistemami"] = "Upr. Bolsh. Sist.",
  ["upsala journal of medical sciences"] = "Ups J Med Sci",
  ["upsala journal of medical sciences. supplement"] = "Ups J Med Sci Suppl",
  ["upsala lakareforenings forhandlingar"] = "Upsala Lakareforen Forh",
  ["upsala läkareförenings förhandlingar"] = "Upsala Lakareforen Forh",
  ["upstream oil and gas technology"] = "Upstream Oil Gas Technol.",
  ["upv journal of natural sciences"] = "UPV J Nat Sci",
  ["ur [reports]. u.s. atomic energy commission"] = "UR Rep",
  ["ur reports"] = "UR Rep.",
  ["ur-schweiz. la suisse primitive"] = "UrSchw",
  ["ural mathematical journal"] = "Ural Math. J.",
  ["ural\\cprime skiy gosudarstvennyy universitet im"] = "Ural. Gos. Univ. Mat. Zap.",
  ["urban affairs quarterly"] = "Urban Aff Q",
  ["urban affairs review (thousand oaks, calif.)"] = "Urban Aff Rev Thousand Oaks Calif",
  ["urban agriculture & regional food systems"] = "Urban Agric. Reg. Food Syst.",
  ["urban anthropology"] = "Urban anthropol",
  ["urban anthropology and studies of cultural systems and world economic development"] = "Urban Anthropol Stud Cult Syst World Econ Dev",
  ["urban climate"] = "Urban Clim.",
  ["urban ecology"] = "Urban Ecol.",
  ["urban ecosystems"] = "Urban Ecosyst",
  ["urban education"] = "Urban Educ (Beverly Hills Calif)",
  ["urban forestry & urban greening"] = "Urban For Urban Green",
  ["urban geography"] = "Urban Geogr",
  ["urban health"] = "Urban Health",
  ["urban history"] = "Urban History",
  ["urban history review. revue d'histoire urbaine"] = "Urban Hist Rev",
  ["urban history review. revue d’histoire urbaine"] = "Urban Hist Rev",
  ["urban history yearbook"] = "Urban Hist Yearb",
  ["urban informatics"] = "Urban Inf.",
  ["urban planning"] = "Urban Plan",
  ["urban rail transit"] = "Urban Rail Transit",
  ["urban research & practice"] = "Urban Res Pract",
  ["urban science (basel, switzerland)"] = "Urban Sci",
  ["urban social work"] = "Urban Soc Work",
  ["urban studies"] = "Urban Stud.",
  ["urban studies (edinburgh, scotland)"] = "Urban Stud",
  ["urban transformations"] = "Urban Transform",
  ["urban water journal"] = "Urban Water J",
  ["urban, planning and transport research"] = "Urban Plan Transp Res",
  ["urbana (caracas, venezuela)"] = "Urbana",
  ["urbanisation and health newsletter"] = "Urban Health Newsl",
  ["uremia investigation"] = "Uremia Invest",
  ["urine (amsterdam, netherlands)"] = "Urine (Amst)",
  ["urologe. ausgabe a"] = "Urologe. A",
  ["urologia internationalis"] = "Urol Int",
  ["urologia polska"] = "Urol Pol",
  ["urologic clinics of north america"] = "Urol. Clin. North Am.",
  ["urologic nursing"] = "Urol Nurs",
  ["urologic nursing : official journal of the american urological association allied"] = "Urol Nurs",
  ["urologic oncology"] = "Urol Oncol",
  ["urologic radiology"] = "Urol Radiol",
  ["urological research"] = "Urol Res",
  ["urological science"] = "Urol Sci",
  ["urological survey"] = "Urol Surv",
  ["urologicke listy"] = "Urol Listy",
  ["urologiia (moscow, russia : 1923)"] = "Urol Mosc",
  ["urologiia (moscow, russia : 1999)"] = "Urologiia",
  ["urologiia i nefrologiia"] = "Urol Nefrol (Mosk)",
  ["urologii︠a︡ (moscow, russia : 1923)"] = "Urol Mosc",
  ["urologii︠a︡ (moscow, russia : 1999)"] = "Urologiia",
  ["urologii︠a︡ i nefrologii︠a︡"] = "Urol Nefrol (Mosk)",
  ["urologists in cancer care"] = "Urol Cancer Care",
  ["urology annals"] = "Urol Ann",
  ["urology case reports"] = "Urol Case Rep",
  ["urology journal"] = "Urol J",
  ["urology practice"] = "Urol Pract",
  ["urology times"] = "Urol Times",
  ["urology video journal"] = "Urol Video J",
  ["ursi radio science bulletin"] = "URSI Radio Sci. Bull.",
  ["ursus (international association for bear research and management)"] = "Ursus",
  ["us clivar reports"] = "US CLIVAR Rep",
  ["us endocrinology"] = "US Endocrinol",
  ["us gastroenterology & hepatology review"] = "US Gastroenterol Hepatol Rev",
  ["us neurology"] = "US Neurol",
  ["us news and world report"] = "US News World Rep.",
  ["us obstetrics & gynaecology"] = "US Obstet Gynecol",
  ["us oncology"] = "US Oncol",
  ["us oncology & hematology"] = "US Oncol Hematol",
  ["us ophthalmic review"] = "US Ophthalmic Rev",
  ["us psychiatry"] = "US Psyc",
  ["us respiratory & pulmonary diseases"] = "US Respir Pulm Dis",
  ["usa today"] = "USA Today",
  ["usa weekend"] = "USA Weekend",
  ["usaf medical service digest"] = "USAF Med Serv Dig",
  ["usaf medical service digest / united states air force"] = "USAF Med Serv Dig",
  ["usaid developments"] = "USAID Dev",
  ["usaid highlights"] = "USAID Highlights",
  ["user modeling and user-adapted interaction"] = "User Model User-adapt Interact",
  ["uspekhi biologicheskoi khimii"] = "Usp Biol Khim",
  ["uspekhi biologicheskoĭ khimii"] = "Usp Biol Khim",
  ["uspekhi fizicheskikh nauk [soviet physics–uspekhi]"] = "Usp. Fiz. Nauk [Sov. Phys. Usp.]",
  ["uspekhi fiziologicheskikh nauk"] = "Usp Fiziol Nauk",
  ["uspekhi khimii"] = "Usp Khim",
  ["uspekhi matematicheskikh nauk"] = "Uspekhi Mat. Nauk",
  ["uspekhi sovremennoi biologii"] = "Usp. Sovrem. Biol.",
  ["uspekhi sovremennoĭ biologii"] = "Usp Sovrem Biol",
  ["utah code annotated 1953. utah"] = "Utah Code Annot 1953 Utah",
  ["utah historical quarterly"] = "Utah Hist Q",
  ["utah law review"] = "Utah Law Rev",
  ["utah nurse"] = "Utah Nurse",
  ["utah science"] = "Utah Sci",
  ["utilitas mathematica"] = "Util. Math.",
  ["utilities policy"] = "Util Policy",
  ["utokyo engineering course/basic mathematics"] = "UTokyo Eng. Course/Basic Math.",
  ["utopian studies"] = "Utop Stud",
  ["utsunomiya university"] = "Bull. Fac. Ed. Utsunomiya Univ. Sect. 2",
  ["uttar pradesh state dental journal"] = "Uttar Pradesh State Dent. J.",
  ["uwfl [reports]. u.s. atomic energy commission"] = "UWFL Rep",
  ["uwfl reports"] = "UWFL Rep.",
  ["uzbek mathematical journal"] = "Uzbek Math. J.",
  ["uzbekskii khimicheskii zhurnal"] = "Uzb. Khim. Zh.",
  ["ŭngyong tʻonggye yŏn'gu = the korean journal of applied statistics"] = "Ungyong Tonggye Yongu",
  ["vaccination research : open journal"] = "Vaccin Res",
  ["vaccine & immunization news : the newsletter of the global programme for vaccines and immunization"] = "Vaccine Immun News",
  ["vaccine (auckland, n.z.)"] = "Vaccine (Auckl)",
  ["vaccine reports"] = "Vaccine Rep",
  ["vaccine weekly"] = "Vaccine Wkly",
  ["vaccine: x"] = "Vaccine X",
  ["vadose zone journal"] = "Vadose Zone J.",
  ["vadose zone journal : vzj"] = "Vadose Zone J",
  ["vakblad voor biologen"] = "Vakbl Biol",
  ["vakuum in forschung und praxis"] = "Vak. Forsch. Prax.",
  ["valparaiso university law review. valparaiso university. school of law"] = "Valparaiso Univ Law Rev",
  ["value in health"] = "Value Health",
  ["value in health : the journal of the international society for pharmacoeconomics and outcomes research"] = "Value Health",
  ["value in health regional issues"] = "Value Health Reg Issues",
  ["van nostrand reinhold mathematics studies"] = "Van Nostrand Reinhold Math. Stud.",
  ["vancouver studies in cognitive science"] = "Vancouver Stud. Cogn. Sci.",
  ["vancouver sun (1986)"] = "Vanc Sun",
  ["vancouver sun (1986)."] = "Vanc Sun",
  ["vanderbilt journal of entertainment and technology law"] = "Vanderbilt J Entertain Technol Law",
  ["vanderbilt law review"] = "Vanderbilt Law Rev",
  ["varahmihir journal of mathematical sciences"] = "Varahmihir J. Math. Sci.",
  ["vard i norden"] = "Vard Nord. Utveckl. Forsk.",
  ["variorum collected studies series"] = "Variorum Collect. Stud. Ser.",
  ["various publications series"] = "Various Publ. Ser.",
  ["varstvo spomenikov"] = "VarSpom",
  ["vasa. supplementum"] = "Vasa Suppl",
  ["vasa. zeitschrift fur gefasskrankheiten. journal for vascular diseases"] = "Vasa",
  ["vasa. zeitschrift für gefässkrankheiten"] = "Vasa",
  ["vascular and endovascular surgery"] = "Vasc Endovascular Surg",
  ["vascular biology"] = "Vasc. Biol.",
  ["vascular biology (bristol, england)"] = "Vasc Biol",
  ["vascular cell"] = "Vasc Cell",
  ["vascular disease prevention"] = "Vasc Dis Prev",
  ["vascular diseases"] = "Vasc Dis",
  ["vascular health and risk management"] = "Vasc Health Risk Manag",
  ["vascular investigation and therapy"] = "Vasc Investig Ther",
  ["vascular medicine"] = "Vasc. Med.",
  ["vascular medicine (london, england)"] = "Vasc Med",
  ["vascular pharmacology"] = "Vascul Pharmacol",
  ["vascular specialist international"] = "Vasc Specialist Int",
  ["vascular surgery"] = "Vasc Surg",
  ["vavilovskiĭ zhurnal genetiki i selekt︠s︡ii"] = "Vavilovskii Zhurnal Genet Selektsii",
  ["vård i norden"] = "Vard Nord Utveckl Forsk",
  ["vb news"] = "VB News",
  ["vecteur environnement"] = "Vecteur Environ.",
  ["vector borne and zoonotic diseases"] = "Vector Borne Zoonotic Dis.",
  ["vector borne and zoonotic diseases (larchmont, n.y.)"] = "Vector Borne Zoonotic Dis",
  ["vector optimization"] = "Vector Optim.",
  ["vector-borne and zoonotic diseases"] = "Vector-Borne Zoonotic Dis.",
  ["vegetable crops research bulletin"] = "Veg. Crops  Res. Bull.",
  ["vegetarische bode"] = "Veg Bode",
  ["vegetation classification and survey"] = "Veg. Classif. Surv.",
  ["vegetation history and archaeobotany"] = "Veg Hist Archaeobot",
  ["vegetos (bareilly, india)"] = "Vegetos",
  ["vehicle system dynamics"] = "Veh. Syst. Dyn.",
  ["vehicular communications"] = "Veh. Commun.",
  ["vekove. dvumesečno spisanie. bălgarsko istoričesko družestvo"] = "Vekove",
  ["veleia. revista de prehistoria, historia antigua, arqueología y filología clásicas"] = "Veleia",
  ["veleia: revista de prehistoria, historia antigua, arqueologá y filología clásicas"] = "Veleia",
  ["venereology : official publication of the national venereology council of australia"] = "Venereology",
  ["venezia arti. bolletino del dipartimento di storia e critica delle arti dell’università di venezia"] = "VenArt",
  ["venezuela odontologica"] = "Venez. Odontol.",
  ["venezuela odontológica"] = "Venez Odontol",
  ["venoms and toxins"] = "Venoms Toxins",
  ["venture capital"] = "Venture Capital",
  ["venus (fukuyama-shi, japan)"] = "Venus",
  ["verbanus. rassegna per la cultura, l’arte, la storia del lago"] = "Verbanus",
  ["verbum (nancy, france)"] = "Verbum (Nancy)",
  ["verbum: revue de linguistique publ. par l’univ. de nancy ii"] = "Verbum",
  ["verhandelingen - koninklijke academie voor geneeskunde van belgie"] = "Verh K Acad Geneeskd Belg",
  ["verhandelingen - koninklijke academie voor geneeskunde van belgië"] = "Verh K Acad Geneeskd Belg",
  ["verhandelingen - koninklijke vlaamse academie voor geneeskunde van belgie"] = "Verh K Vlaam Acad Geneeskd Belg",
  ["verhandelingen - koninklijke vlaamse academie voor geneeskunde van belgië"] = "Verh K Vlaam Acad Geneeskd Belg",
  ["verhandelingen, koninklijke academie voor geneeskunde van belgie"] = "Verh. K. Acad. Geneeskd. Belg.",
  ["verhandelingen, koninklijke vlaamse academie voor geneeskunde van belgie"] = "Verh. K. Vlaam. Acad. Geneeskd. Belg.",
  ["verhandlungen - internationale vereinigung fur theoretische und angewandte limnologie"] = "Verh. Int. Ver. Theor. Angew. Limnol.",
  ["verhandlungen - internationale vereinigung für theoretische und angewandte limnologie"] = "Verh. - Int. Ver. Theor. Angew. Limnol.",
  ["verhandlungen der anatomischen gesellschaft"] = "Verh Anat Ges",
  ["verhandlungen der deutschen gesellschaft fur herz- und kreislaufforschung"] = "Verh. Dtsch. Ges. Herz. Kreislaufforsch.",
  ["verhandlungen der deutschen gesellschaft fur innere medizin"] = "Verh. Dtsch. Ges. Inn. Med.",
  ["verhandlungen der deutschen gesellschaft fur kreislaufforschung"] = "Verh. Dtsch. Ges. Kreislaufforsch.",
  ["verhandlungen der deutschen gesellschaft fur pathologie"] = "Verh Dtsch Ges Pathol",
  ["verhandlungen der deutschen gesellschaft fur rheumatologie"] = "Verh. Dtsch. Ges. Rheumatol.",
  ["verhandlungen der deutschen gesellschaft für herz- und kreislaufforschung"] = "Verh Dtsch Ges Herz Kreislaufforsch",
  ["verhandlungen der deutschen gesellschaft für innere medizin"] = "Verh Dtsch Ges Inn Med",
  ["verhandlungen der deutschen gesellschaft für kreislaufforschung"] = "Verh Dtsch Ges Kreislaufforsch",
  ["verhandlungen der deutschen gesellschaft für pathologie"] = "Verh Dtsch Ges Pathol",
  ["verhandlungen der deutschen gesellschaft für rheumatologie"] = "Verh Dtsch Ges Rheumatol",
  ["verhandlungen der deutschen physikalischen gesellschaft"] = "Verh. Dtsch. Phys. Ges.",
  ["verhandlungen der gesellschaft für ökologie"] = "Verh. Ges. Ökol.",
  ["verhandlungen der internationalen vereinigung für theoretische und angewandte limnologie. international association of theoretical and applied limnology"] = "Verh Int Ver Theor Angew Limnol",
  ["verhandlungen der internationalen vereinigung für theoretische und angewandte limnologie"] = "Verh. Int. Ver. Theoret. Angew. Limnol.",
  ["verhandlungen der naturforschenden gesellschaft in basel"] = "Verh. Nat.forsch. Ges. Basel",
  ["verhandlungen der schweizerischen naturforschenden gesellschaft"] = "Verh. Schweiz. nat.forsch. Ges.",
  ["verhandlungen der zoologisch-botanischen gesellschaft in österreich"] = "Verh Zool Bot Ges Osterr",
  ["verhandlungen der zoologisch-botanischen gesellschaft in wien. zoologisch-botanische gesellschaft in wien"] = "Verh Zool Bot Ges Wien",
  ["verhandlungen des naturwissenschaftlichen vereins in hamburg (1980)"] = "Verh Naturwiss Ver Hambg (1980)",
  ["veridian e-journal science and technology silpakorn university"] = "Veridian EJ Sci Technol Silpakorn Univ",
  ["verkundigung und forschung: wissenschaft vom neuen testament. beitr. zu evangelische theologie"] = "V&F",
  ["vermessung, photogrammetrie, kulturtechnik"] = "Vermess. Photogramm. Kult.tech.",
  ["vermont history"] = "Vt Hist",
  ["vermont history news"] = "Vt Hist News",
  ["vermont law review"] = "Vt Law Rev",
  ["vermont registered nurse"] = "Vt Regist Nurse",
  ["vermont statutes annotated. vermont"] = "Vt Statut Annot Vt",
  ["vernacular architecture"] = "Vernac. Archit.",
  ["vernon's annotated revised civil statutes of the state of texas. texas"] = "Vernons Annot Revis Civ Statut State Tex Tex",
  ["vernon’s annotated revised civil statutes of the state of texas. texas"] = "Vernons Annot Revis Civ Statut State Tex Tex",
  ["veroffentlichungen aus der morphologischen pathologie"] = "Veroff. Morphol. Pathol.",
  ["veroffentlichungen aus der pathologie"] = "Veroff Pathol",
  ["veroffentlichungen der internationalen gesellschaft fur geschichte der pharmazie e. v"] = "Veroff Int Ges Gesch Pharm",
  ["veroffentlichungen der schweizerischen gesellschaft fur geschichte der medizin und der naturwissenschaften"] = "Veroff Schweiz Ges Gesch Med Naturwiss",
  ["veroffentlichungen der schweizerischen gesellschaft fur geschichte der pharmazie / herausgegeben vom vorstand der schweizerischen gesellschaft fur geschichte der pharmazie (sggp)"] = "Veroff Schweiz Ges Gesch Pharm",
  ["veroffentlichungen des deutschen historischen instituts london"] = "Veroff. Dtsch. Hist. Inst. London",
  ["veroffentlichungen des forschungsinstituts des deutschen museums fur die geschichte der naturwissenschaften und der technik, reihe a: kleine mitteilungen"] = "Veroffentlichungen des Forschungsinstituts des Deutschen Museums fur die Geschichte der Naturwissenschaften und der",
  ["veroffentlichungen des forschungsinstituts des deutschen museums fur die geschichte der naturwissenschaften und der technik, reihe c: quellentexte und ubersetzungen"] = "Veroffentlichungen des Forschungsinstituts des Deutschen Museums fur die Geschichte der Naturwissenschaften und der",
  ["veroffentlichungen des forschungsinstituts des deutschen museums fur die geschichte der naturwissenschaften und der technik. reihe a: kleine mitteilungen"] = "Veroffentlichungen des Forschungsinstituts des Deutschen Museums fur die Geschichte der Naturwissenschaften und der",
  ["veroffentlichungen des forschungsinstituts des deutschen museums fur die geschichte der naturwissenschaften und der technik. reihe c: quellentexte und ubersetzungen"] = "Veroffentlichungen des Forschungsinstituts des Deutschen Museums fur die Geschichte der Naturwissenschaften und der",
  ["veroffentlichungen des instituts fur bibliothekswissenschaft und bibliothekarausbildung der freien universitat berlin"] = "Veroffentl. Inst. Bibliothekswissen. Bibliothekarausbild. Freien Univ. Berlin",
  ["veroffentlichungen zur geschichte der naturwissenschaften, mathematik und medizin"] = "Veroff. Gesch. Naturwiss. Math. Med.",
  ["veröffentlichungen aus der morphologischen pathologie"] = "Veroff Morphol Pathol",
  ["veröffentlichungen aus der pathologie"] = "Veroff Pathol",
  ["veröffentlichungen der internationalen gesellschaft für geschichte der pharmazie e. v"] = "Veroff Int Ges Gesch Pharm",
  ["veröffentlichungen der schweizerischen gesellschaft für geschichte der medizin und der naturwissenschaften"] = "Veroff Schweiz Ges Gesch Med Naturwiss",
  ["veröffentlichungen des naturkundemuseums erfurt"] = "Veroff Natkdmus (Erfurt)",
  ["verpleegkundigen en gemeenschapszorg : tijdschrift van de het nationaal verbond der katholieke vlaamse verplegenden"] = "Verpleegkd Gem",
  ["versicherungsmedizin / herausgegeben von verband der lebensversicherungs-unternehmen e.v. und verband der privaten krankenversicherung e.v"] = "Versicherungsmedizin",
  ["verslagen en mededelingen - koninklijke academie voor nederlandse taal- en letterkunde. koninklijke academie voor nederlandse taal- en letterkunde"] = "Versl Meded K Viaam Acad Taal Lett",
  ["verslagen en mededelingen betreffende de volksgezondheid"] = "Verslag Meded Betreff Volksgezond",
  ["verslagen en mededelingen van de koninklijke vlaamse academie voor taal- en letterkunde. koninklijke vlaamse academie voor taal- en letterkunde"] = "Versl Meded K Vlaam Acad Taal Lett",
  ["verständliche forschung"] = "Verständl. Forsch.",
  ["versuchsanstalt für wasserbau, hydrologie und glaziologie"] = "Vers.anst. Wasserbau Hydrol. Glaziol.",
  ["vertebrate anatomy morphology palaeontology"] = "Vertebr. Anat. Morphol. Paleontol.",
  ["vertebrate zoology"] = "Vertebr Zool",
  ["vertex (buenos aires, argentina)"] = "Vertex",
  ["very short introductions"] = "Very Short Introd.",
  ["veröffentlichungen der gesellschaft pro vindonissa"] = "VGesVind",
  ["veröffentlichungen der landesstelle für naturschutz und landschaftspflege in baden-württemberg"] = "Veröff. Landesstelle Nat.schutz Landsch.pfl. Baden-Württ.",
  ["veröffentlichungen der schweizerischen gesellschaft für geschichte der pharmazie"] = "Veroff Schweiz Ges Gesch Pharm",
  ["veröffentlichungen des geobotanischen institutes der eidg. technischen hochschule, stiftung rübel in zürich"] = "Veröff. Geobot. Inst. Eidgenöss. Tech. Hochsch., Stift. Rübel  Zür.",
  ["veröffentlichungen des leibniz-archivs"] = "Veröffentl. Leibniz-Archivs",
  ["veröffentlichungen natur-museum luzern"] = "Veröff. Nat.-Mus. Luzern",
  ["vesalius : acta internationales historiae medicinae"] = "Vesalius",
  ["vessel plus"] = "Vessel Plus",
  ["vesta company series in pure and applied mathematics"] = "Vesta Co. Ser. Pure Appl. Math.",
  ["vestis. izvestiia. latvijas psr zinatnu akademija"] = "Vestis Lat Psr Zinat Akad",
  ["vestis. izvestiia. latvijas psr zinātnu akadēmija"] = "Vestis Lat Psr Zinat Akad",
  ["vestnik - leningradskogo universiteta. geologiia, geografiia. leningradskii gosudarstvennyi universitet"] = "Vestn Leningr Univ Geol Geogr",
  ["vestnik - leningradskogo universiteta. geologii︠a︡, geografii︠a︡. leningradskiĭ gosudarstvennyĭ universitet"] = "Vestn Leningr Univ Geol Geogr",
  ["vestnik akademii meditsinskikh nauk sssr"] = "Vestn Akad Med Nauk SSSR",
  ["vestnik akademii nauk sssr"] = "Vestn Akad Nauk SSSR",
  ["vestnik belorusskogo gosudarstvennogo universiteta"] = "Vestn. Beloruss. Gos. Univ. Ser. 1 Fiz. Mat. Inform.",
  ["vestnik belorusskogo gosudarstvennogo universiteta imeni v.i. lenina. seriíà 2, khimiíà, biologiíà, geografiíà"] = "Vestn Beloruss Gos Univ Im VI Lenina Ser 2 Him Biol Geol Geogr",
  ["vestnik bgu. seriâ 2, himiâ, biologiâ, geografiâ"] = "Vestn BGU Ser 2 Him Biol Geogr",
  ["vestnik dermatologii i venerologii"] = "Vestn Dermatol Venerol",
  ["vestnik drevnei istorii. journal of ancient history"] = "Vestn Drevnej Istor (Moskva)",
  ["vestnik drevneĭ istorii. journal of ancient history"] = "Vestn Drevnej Istor (Moskva)",
  ["vestnik drevnej istorii"] = "VDI",
  ["vestnik drevnej istorii (revue d’histoire ancienne)"] = "VDI",
  ["vestnik irkutskogo gosudarstvennogo tehničeskogo universiteta"] = "Vestn Irkutsk Gos Teh Univ",
  ["vestnik kamčatskogo gosudarstvennogo tehničeskogo universiteta"] = "Vestn Kamcatsk Gos Teh Univ",
  ["vestnik khirurgii imeni i. i. grekova"] = "Vestn Khir Im I I Grek",
  ["vestnik kraunts. fiziko-matematicheskie nauki"] = "Vestn. KRAUNTS. Fiz.-Mat. Nauki",
  ["vestnik l\\cprime vovskogo politekhnicheskogo instituta"] = "Vestnik L\\cprime vov. Politekhn. Inst.",
  ["vestnik leningradskogo universiteta, ser. 2, istor."] = "VLUIST",
  ["vestnik leningradskogo universiteta, ser. 6, filos."] = "VLUFILOS",
  ["vestnik leningradskogo universiteta. biologiia"] = "Vestn Leningr Univ Biol",
  ["vestnik leningradskogo universiteta. seriia biologii"] = "Vestn Leningr Univ Ser Biol",
  ["vestnik ministerstva nauki–akademiya nauk respubliki kazakhstan"] = "Vestn. Minister. Nauki Akad. Nauk Resp. Kaz.",
  ["vestnik mordovskogo universiteta"] = "Vestn. Mord. Univ.",
  ["vestnik moskovskogo universiteta"] = "Vestnik Moskov. Univ. Ser. I Mat. Mekh.",
  ["vestnik moskovskogo universiteta (filol. sekcija)"] = "VMUFILOL",
  ["vestnik moskovskogo universiteta (filos. sekcija)"] = "VMUFILOS",
  ["vestnik moskovskogo universiteta (ist. sekcija)"] = "VMUIST",
  ["vestnik moskovskogo universiteta, seriya 2: khimiya"] = "Vestn. Mosk. Univ., Ser. 2: Khim.",
  ["vestnik moskovskogo universiteta, seriya 3: fizika, astronomiya"] = "Vestn. Mosk. Univ., Ser. 3: Fiz., Astron.",
  ["vestnik moskovskogo universiteta. seriia v, geografiia"] = "Vestn Mosk Univ Ser V",
  ["vestnik moskovskogo universiteta. seriia viii: istoriia. moskovskii gosudarstvennyi universitet im. m.v. lomonosova"] = "Vestn Mosk Univ Ser 8 Istor",
  ["vestnik moskovskogo universiteta. serii︠a︡ v, geografii︠a︡"] = "Vestn Mosk Univ Ser V",
  ["vestnik moskovskogo universiteta. serii︠a︡ viii: istorii︠a︡. moskovskiĭ gosudarstvennyĭ universitet im. m.v. lomonosova"] = "Vestn Mosk Univ Ser 8 Istor",
  ["vestnik moskovskogo universiteta. seriya i. matematika, mekhanika"] = "Vestnik Moskov. Univ. Ser. I Mat. Mekh.",
  ["vestnik moskovskogo universiteta. seriya iii. fizika, astronomiya"] = "Vestnik Moskov. Univ. Ser. III Fiz. Astronom.",
  ["vestnik moskovskogo universiteta. seriya xv. vychislitelnaya matematika i kibernetika"] = "Vestnik Moskov. Univ. Ser. XV Vychisl. Mat. Kibernet.",
  ["vestnik oftalmologii"] = "Vestn Oftalmol",
  ["vestnik ohotovedeniâ"] = "Vestn Ohotoved",
  ["vestnik otorinolaringologii"] = "Vestn Otorinolaringol",
  ["vestnik permskogo nat︠s︡ionalʹnogo issledovatelʹskogo politekhnicheskogo universiteta. prikladnai︠a︡ ėkologii︠a︡, urbanistika"] = "Vestn Permsk Natsionalnogo Issledovatelskogo Politekh Univ Prikl Ekol Urban",
  ["vestnik permskogo universiteta : seriâ biologiâ"] = "Vestn Permsk Univ Ser Biol",
  ["vestnik rentgenologii i radiologii"] = "Vestn Rentgenol Radiol",
  ["vestnik rossiiskoi akademii meditsinskikh nauk"] = "Vestn. Ross. Akad. Med. Nauk",
  ["vestnik rossiiskoi akademii meditsinskikh nauk / rossiiskaia akademiia meditsinskikh nauk"] = "Vestn Ross Akad Med Nauk",
  ["vestnik rossiĭskogo universiteta druzhby narodov. serii︠a︡ agronomii︠a︡ i zhivotnovodstvo"] = "Vestn Ross Univ Druhzby Narodov Ser Agron Zhivotnovod",
  ["vestnik rossiĭskoĭ akademii meditsinskikh nauk"] = "Vestn Ross Akad Med Nauk",
  ["vestnik rossiĭskoĭ akademii sel'skokhozi︠a︡ĭstvennykh nauk : dvukhmesi︠a︡chnyĭ nauchno-teoreticheskiĭ zhurnal"] = "Vestnik Rossiiskoi Akademii Selskokhoziaistvennykh Nauk",
  ["vestnik samarskogo universiteta. estestvennonauchnaya seriya"] = "Vestn. Samar. Univ. Estestvennonauchn. Ser.",
  ["vestnik sankt-peterburgskogo universiteta"] = "Vestnik S.-Peterburg. Univ. Fiz. Khim.",
  ["vestnik sankt-peterburgskogo universiteta, seriya 4: fizika, khimiya"] = "Vestn. S.-Peterb. Univ., Ser. 4: Fiz., Khim.",
  ["vestnik sankt-peterburgskogo universiteta. matematika. mekhanika. astronomiya"] = "Vestn. St.-Peterbg. Univ. Mat. Mekh. Astron.",
  ["vestnik sankt-peterburgskogo universiteta. prikladnaya matematika. informatika. protsessy upravleniya"] = "Vestn. St.-Peterbg. Univ. Prikl. Mat. Inform. Protsessy Upr.",
  ["vestnik sankt-peterburgskogo universiteta. serii︠a︡ 2, istorii︠a︡, i︠a︡zykoznanie, literaturovedenie"] = "Vestn St Peterbg Univ Ser Istor Azykozn Literaturoved",
  ["vestnik saratovskogo gosudarstvennogo universiteta"] = "Vestnik Saratovskogo Gosudarstvennogo Universiteta",
  ["vestnik saratovskoj gosudarstvennoj akademii prava"] = "Vestn Sarat Gos Akad Prava",
  ["vestnik severnogo meždunarodnogo universiteta"] = "Vestn Sev Mezdunar Univ",
  ["vestnik severo-vostochnogo nauchnogo t︠s︡entra dalʹnevostochnogo otdelenii︠a︡ ran"] = "Vestn Severo Vost Nauchnogo Tsentra Dalnevost Otd RAN",
  ["vestnik slovenskega kemijskega drustva"] = "Vestn. Slov. Kem. Drus.",
  ["vestnik st"] = "Vestnik St. Petersburg Univ. Math.",
  ["vestnik st. petersburg university, mathematics"] = "Vestn. St.Petersburg Univ., Math.",
  ["vestnik st. petersburg university. mathematics"] = "Vestnik St. Petersburg Univ. Math.",
  ["vestnik statistiki (moscow, russia : 1949)"] = "Vestn Statistiki",
  ["vestnik tomskogo gosudarstvennogo universiteta"] = "Vestn Tomsk Gos Univ",
  ["vestnik tomskogo gosudarstvennogo universiteta. biologiâ"] = "Vestn Tomsk Gos Univ Biol",
  ["vestnik tomskogo gosudarstvennogo universiteta. matematika i mechanika"] = "Vestn. Tomsk. Gos. Univ. Mat. Mech.",
  ["vestnik udmurtskogo universiteta. matematika. mekhanika. kompyuternye nauki"] = "Vestn. Udmurt. Univ. Mat. Mekh. Kompyut. Nauki",
  ["vestnik venerologii i dermatologii"] = "Vestn Venerol Dermatol",
  ["vestnik vogis"] = "Vestn VOGiS",
  ["vestnik voronezhskogo gosudarstvennogo universiteta. seriya fizika, matematika"] = "Vestn. Voronezh. Gos. Univ. Ser. Fiz., Mat.",
  ["vestnik zashchity rasteniĭ = plant protection news"] = "Vestn Zas Rast",
  ["vestnik zoologii"] = "Vestn Zool",
  ["vestsi akademii navuk belarusi, seryya biyalagichnykh navuk"] = "Vestsi Akad. Navuk BSSR, Ser. Biyal. Navuk",
  ["vestsi akademii navuk belarusi, seryya fizika-energetychnykh navuk"] = "Vestsi Akad. Navuk BSSR, Ser. Fiz.-Energ. Navuk",
  ["vestsi akademii navuk belarusi, seryya khimichnykh navuk"] = "Vestsi Akad. Navuk BSSR, Ser. Khim. Navuk",
  ["vestsi natsyi︠a︡nalʹnaĭ akadėmii navuk belarusi. seryi︠a︡ bii︠a︡lahichnykh navuk"] = "Vestsi Natsyianalnai Akademii Navuk Belarusi",
  ["vestsi natsyyanalnai akademii navuk belarusi. seryya fizika-matematychnykh navuk. izvestiya natsionalnoi akademii nauk belarusi. seriya fiziko-matematicheskikh nauk. proceedings of the national academy of sciences of belarus. physics and mathematics series"] = "Vestsi Nats. Akad. Navuk Belarusi Ser. Fiz.-Mat. Navuk",
  ["vestsi natsyyanal’nai akademii navuk belarusi, seryya khimichnykh navuk"] = "Vestsi Nats. Akad. Navuk Belarusi, Ser. Khim. Navuk",
  ["vetera christianorum"] = "VetChr",
  ["veterinair historisch genootschap cahier"] = "Vet Historisch Genoot Cah",
  ["veterinar-medizinische nachrichten"] = "Vet Med Nachr",
  ["veterinaria italiana"] = "Vet Ital",
  ["veterinariâ (abaj)"] = "Veterinaria Abaj",
  ["veterinarii︠a︡ (ukraine. ministerstvo silʹsʹkoho hospodarstva)"] = "Vet (Kiev)",
  ["veterinarii︠a︡ segodni︠a︡ = veterinary science today"] = "Vet Segodnia",
  ["veterinarni medicina"] = "Vet. Med.",
  ["veterinarno-meditsinski nauki"] = "Vet Med Nauki",
  ["veterinarnomeditsinski nauki"] = "Vet. Med. Nauki",
  ["veterinarski arhiv"] = "Vet Arh",
  ["veterinarski glasnik"] = "Vet Glas",
  ["veterinary anaesthesia and analgesia"] = "Vet Anaesth Analg",
  ["veterinary and animal science"] = "Vet Anim Sci",
  ["veterinary and comparative oncology"] = "Vet Comp Oncol",
  ["veterinary and comparative orthopaedics and traumatology"] = "Vet. Comp. Orthop. Traumatol.",
  ["veterinary and comparative orthopaedics and traumatology : v.c.o.t"] = "Vet Comp Orthop Traumatol",
  ["veterinary and human toxicology"] = "Vet Hum Toxicol",
  ["veterinary clinical pathology"] = "Vet Clin Pathol",
  ["veterinary clinical pathology / american society for veterinary clinical pathology"] = "Vet Clin Pathol",
  ["veterinary clinics of north america"] = "Vet. Clin. N. Am.",
  ["veterinary clinics of north america. equine practice"] = "Vet. Clin. North Am. Equine Pract.",
  ["veterinary clinics of north america. exotic animal practice"] = "Veterinary Clin. North Am. Exot. Anim. Pract.",
  ["veterinary clinics of north america. food animal practice"] = "Vet. Clin. North Am. Food Anim. Pract.",
  ["veterinary clinics of north america. large animal practice"] = "Vet. Clin. North Am. Large Anim. Pract.",
  ["veterinary clinics of north america. small animal practice"] = "Vet. Clin. North Am. Small Anim. Pract.",
  ["veterinary clinics of north america: equine practice"] = "Vet. Clin. N. Am.: Equine Pract.",
  ["veterinary clinics of north america: exotic animal practice"] = "Vet. Clin. N. Am.: Exot. Anim. Pract.",
  ["veterinary clinics of north america: food animal practice"] = "Vet. Clin. N. Am.: Food Anim. Pract.",
  ["veterinary clinics of north america: large animal practice"] = "Vet. Clin. N. Am.: Large Anim. Pract.",
  ["veterinary clinics of north america: small animal practice"] = "Vet. Clin. N. Am.: Small Anim. Pract.",
  ["veterinary dermatology"] = "Vet Dermatol",
  ["veterinary extension quarterly"] = "Vet Ext Q",
  ["veterinary heritage : bulletin of the american veterinary history society"] = "Vet Herit",
  ["veterinary history"] = "Vet Hist",
  ["veterinary immunology and immunopathology"] = "Vet Immunol Immunopathol",
  ["veterinary journal"] = "Vet. J.",
  ["veterinary journal (london, england : 1997)"] = "Vet J",
  ["veterinary medical journal (jīzah, egypt : 1995)"] = "Vet Med J",
  ["veterinary medical review"] = "Vet Med Rev",
  ["veterinary medicine"] = "Vet Med",
  ["veterinary medicine (auckland, n.z.)"] = "Vet Med (Auckl)",
  ["veterinary medicine and science"] = "Vet Med Sci",
  ["veterinary medicine international"] = "Vet Med Int",
  ["veterinary medicine, small animal clinician"] = "Vet. Med. Small Anim. Clin.",
  ["veterinary medicine, small animal clinician : vm, sac"] = "Vet Med Small Anim Clin",
  ["veterinary medicine: research and reports"] = "Vet. Med.: Res. Rep.",
  ["veterinary microbiology"] = "Vet Microbiol",
  ["veterinary ophthalmology"] = "Vet Ophthalmol",
  ["veterinary parasitology"] = "Vet Parasitol",
  ["veterinary parasitology, regional studies and reports"] = "Vet Parasitol Reg Stud Reports",
  ["veterinary parasitology: x"] = "Vet Parasitol X",
  ["veterinary pathology"] = "Vet Pathol",
  ["veterinary pathology. supplement"] = "Vet Pathol Suppl",
  ["veterinary quarterly"] = "Vet. Q",
  ["veterinary radiology & ultrasound"] = "Vet. Radiol. Ultrasound",
  ["veterinary radiology & ultrasound : the official journal of the american college of veterinary radiology and the international veterinary radiology association"] = "Vet Radiol Ultrasound",
  ["veterinary radiology and ultrasound"] = "Vet. Radiol. Ultrasound",
  ["veterinary record"] = "Vet. Rec.",
  ["veterinary record case reports"] = "Vet Rec Case Rep",
  ["veterinary record open"] = "Vet Rec Open",
  ["veterinary research"] = "Vet Res",
  ["veterinary research communications"] = "Vet Res Commun",
  ["veterinary research forum : an international quarterly journal"] = "Vet Res Forum",
  ["veterinary reviews and annotations"] = "Vet Rev Annot",
  ["veterinary science communications"] = "Vet Sci Commun",
  ["veterinary science in china"] = "Vet Sci China",
  ["veterinary sciences"] = "Vet Sci",
  ["veterinary surgery"] = "Vet. Surg.",
  ["veterinary surgery : vs"] = "Vet Surg",
  ["veterinary surgery : vs : the official journal of the american college of veterinary surgeons"] = "Vet Surg",
  ["veterinary therapeutics"] = "Vet. Ther.",
  ["veterinary therapeutics : research in applied veterinary medicine"] = "Vet Ther",
  ["veterinary world"] = "Vet World",
  ["veterinární medicína"] = "Vet Med (Praha)",
  ["veterinär-medizinische nachrichten"] = "Vet Med Nachr",
  ["veteriner fakültesi dergisi"] = "Vet Fak Derg",
  ["veterynarna medyt︠s︡yna"] = "Veterynama Medytsyna",
  ["vetus testamentum: quarterly publ. by the international organization of old testament scholars"] = "VT",
  ["vĕstník c̆eskoslovenských lékar̆ů"] = "Vestn Ceskoslov Lekaru",
  ["vgb kraftwerkstechnik"] = "VGB Kraftwerkstech.",
  ["viata medicala"] = "Viata Med",
  ["viata medicala; revista de informare profesionala si stiintifica a cadrelor medii sanitare"] = "Viata Med. Rev. Inf. Prof. Stiint. Cadrelor Medii Sanit.",
  ["viaţa medicală"] = "Viata Med",
  ["viaţa medicală; revistă de informare profesională şi ştiinţifică a cadrelor medii sanitare"] = "Viata Med Rev Inf Prof Stiint Cadrelor Medii Sanit",
  ["vibrant : virtual brazilian anthropology"] = "Vibrant (Brasilia)",
  ["vibrational spectroscopy"] = "Vib. Spectrosc.",
  ["vich-infekt︠s︡ii︠a︡ i immunosupresii"] = "Vichinfektsiia Immunosuppr",
  ["vichiana. rassegna di studi filologici e storici"] = "Vichiana",
  ["vichiana: rassegna di studi filologici e storici"] = "Vichiana",
  ["vicino oriente"] = "VicOr",
  ["victims & offenders"] = "Vict Offender",
  ["victoria university of wellington law review / law faculty"] = "Law Rev",
  ["victorian literature and culture"] = "Vic Lit Cult",
  ["victorian periodicals newsletter"] = "Vic Period Newsl",
  ["victorian periodicals review"] = "Vic Period Rev",
  ["victorian studies"] = "Vic Stud",
  ["victorian veterinary proceedings"] = "Vic Vet Proc",
  ["victorian veterinary proceedings / australian veterinary association (victorian division)"] = "Vic Vet Proc",
  ["vida médica"] = "Vida Med",
  ["vida nueva"] = "Vida Nueva",
  ["vida odontologica"] = "Vida Odontol",
  ["vida y ética"] = "Vida Etica",
  ["video journal of prosthetic urology"] = "Video J Prosthet Urol",
  ["video-assisted thoracic surgery"] = "Video Assist Thorac Surg",
  ["videogie : an official video journal of the american society for gastrointestinal endoscopy"] = "VideoGIE",
  ["videourology (new rochelle, n.y.)"] = "Videourology (New Rochelle)",
  ["vie et langage"] = "Vie Lang",
  ["vie et milieu (paris, france : 1980)"] = "Vie Milieu",
  ["vie et sante"] = "Vie Sante",
  ["vie et santé"] = "Vie Sante",
  ["vie medicale (paris, france : 1920)"] = "Vie Med",
  ["vie medicale au canada francais"] = "Vie Med. Can. Fr.",
  ["vie médicale (paris, france : 1920)"] = "Vie Med",
  ["vienna circle institute library"] = "Vienna Circ. Inst. Libr.",
  ["vienna circle institute yearbook"] = "Vienna Circ. Inst. Yearb.",
  ["vienna yearbook of population research"] = "Vienna Yearb Popul Res",
  ["vierteljahresberichte / forschungsinstitut der friedrich-ebert-stiftung"] = "Vierteljahresberichte",
  ["vierteljahresschrift fur dermatologie und syphilis"] = "Vierteljahresschr. Dermatol. Syph.",
  ["vierteljahrschrift fur sozial- und wirtschaftsgeschichte"] = "Vierteljahrschr Soz Wirtschaftsgesch",
  ["vierteljahrschrift für sozial- und wirtschaftsgeschichte"] = "Vierteljahrschr Soz Wirtschaftsgesch",
  ["vierteljahrshefte fur zeitgeschichte"] = "Vierteljahrsh Zeitgesch",
  ["vierteljahrshefte für zeitgeschichte"] = "Vierteljahrsh Zeitgesch",
  ["vierteljahrsheftze zur wirtschaftsforschung"] = "Vierteljahrs. Wirtschaftsforschung",
  ["vierteljahrsschrift der naturforschenden gesellschaft in zürich"] = "Vierteljahrsschr. Nat.forsch. Ges. Zür.",
  ["vierteljahrsschrift fur schweizerische sanitatsoffiziere"] = "Vierteljahrsschr. Schweiz. Sanitatsoff.",
  ["vierteljahrsschrift fur schweizerische sanitatsoffiziere. journal trimestriel des officiers suisses du service de sante. rivista trimestrale degli ufficiali sanitari svizzeri"] = "Vierteljahrsschr Schweiz Sanitatsoff",
  ["vierteljahrsschrift für schweizerische sanitätsoffiziere. journal trimestriel des officiers suisses du service de santé. rivista trimestrale degli ufficiali sanitari svizzeri"] = "Vierteljahrsschr Schweiz Sanitatsoff",
  ["vietnam journal of chemistry"] = "Vietnam J. Chem.",
  ["vietnam journal of computer science"] = "Vietnam J. Comput. Sci.",
  ["vietnam journal of mathematics"] = "Vietnam J. Math.",
  ["vietnam journal of science and technology"] = "Vietnam J Sci Technol",
  ["vietnamese studies (hanoi, vietnam : 1966)"] = "Vietnam Stud",
  ["vieweg mathematik"] = "Vieweg Math.",
  ["vieweg programmbibliothek. taschenrechner"] = "Vieweg Programmbibl. Taschenrechner",
  ["vieweg studium: aufbaukurs mathematik"] = "Vieweg Stud. Aufbaukurs Math.",
  ["vieweg studium: grundkurs mathematik"] = "Vieweg Stud. Grundkurs Math.",
  ["vieweg textbook mathematics"] = "Vieweg Textbook Math.",
  ["vieweg tracts in pure and applied physics"] = "Vieweg Tracts Pure Appl. Phys.",
  ["viewpoints on digestive diseases"] = "Viewp Dig Dis",
  ["vigiliae christianae"] = "VigChr",
  ["vigiliae christianae: a review of early christian life and language"] = "VChr",
  ["vigo international journal of applied linguistics"] = "Vigo Int J Appl Linguist",
  ["viitorul social"] = "Viitorul Soc",
  ["vijnana parishad anusandhan patrika"] = "Vijnana Parishad Anusandhan Patrika",
  ["vijnana parishad anusandhan patrika. the research journal of the hindi science academy"] = "Vijnana Parishad Anusandhan Patrika",
  ["vijnana parishad of india"] = "Bull. Vijnana Parishad India",
  ["villaclara médica"] = "Villaclara Med",
  ["village voice (greenwich village, new york, n.y.)"] = "Village Voice",
  ["villanova law review"] = "Villanova Law Rev",
  ["viminacium. zbornik radova narodnog muzeja"] = "Viminacium",
  ["vina quarterly"] = "VINA Q",
  ["vingtieme siecle (paris, france : 1984)"] = "Vingtieme Siecle",
  ["vingtième siècle (paris, france : 1984)"] = "Vingtieme Siecle",
  ["violence against women"] = "Violence Against Women",
  ["violence and gender"] = "Violence Gend",
  ["violence and victims"] = "Violence Vict",
  ["violence update"] = "Violence Update",
  ["vipimage 2017 : proceedings of the vi eccomas thematic conference on computational vision and medical image processing porto, portugal, october 18-20, 2017. vipimage (conference) (2017 : porto, portugal)"] = "VipIMAGE 2017 (2017)",
  ["vipimage 2019 : proceedings of the vii eccomas thematic conference on computational vision and medical image processing, october 16-18, 2019, porto, portugal. vipimage (conference) (2019 : porto, portugal)"] = "VipIMAGE 2019 (2019)",
  ["viral hepatitis in practice"] = "Viral Hepat Pract",
  ["viral immunology"] = "Viral Immunol",
  ["virchows archiv"] = "Virchows Arch.",
  ["virchows archiv : an international journal of pathology"] = "Virchows Arch",
  ["virchows archiv fur pathologische anatomie und physiologie und fur klinische medizin"] = "Virchows Arch. Pathol. Anat. Physiol. Klin. Med.",
  ["virchows archiv für pathologische anatomie und physiologie und für klinische medizin"] = "Virchows Arch Pathol Anat Physiol Klin Med",
  ["virchows archiv. a, pathological anatomy and histology"] = "Virchows Arch A Pathol Anat Histol",
  ["virchows archiv. a, pathological anatomy and histopathology"] = "Virchows Arch A Pathol Anat Histopathol",
  ["virchows archiv. a, pathology, pathologische anatomie"] = "Virchows Arch. A Pathol. Pathol. Anat.",
  ["virchows archiv. a, pathology. pathologische anatomie"] = "Virchows Arch A Pathol Pathol Anat",
  ["virchows archiv. a. pathological anatomy and histology"] = "Virchows Arch. A Pathol. Anat. Histol.",
  ["virchows archiv. a: pathology. pathologische anatomie"] = "Virchows Arch A Pathol Pathol Anat",
  ["virchows archiv. b, cell pathology"] = "Virchows Arch B Cell Pathol",
  ["virchows archiv. b, cell pathology including molecular pathology"] = "Virchows Arch B Cell Pathol Incl Mol Pathol",
  ["virchows archiv. b. cell pathology"] = "Virchows Arch. B Cell Pathol.",
  ["virchows archiv. b: cell pathology"] = "Virchows Arch B Cell Pathol",
  ["virginia bar news"] = "Va Bar News",
  ["virginia cavalcade"] = "Va Cavalcade",
  ["virginia dental journal"] = "Va Dent J",
  ["virginia health bulletin"] = "Va Health Bull",
  ["virginia journal of international law"] = "VA J Int Law",
  ["virginia journal of science"] = "Va J Sci",
  ["virginia journal of social policy & the law"] = "Va J Soc Policy Law",
  ["virginia law review"] = "Va Law Rev",
  ["virginia medical"] = "Va Med",
  ["virginia medical monthly"] = "Va Med Mon (1918)",
  ["virginia medical quarterly"] = "Va. Med. Q.",
  ["virginia medical quarterly : vmq"] = "Va Med Q",
  ["virginia nurse"] = "Va Nurse",
  ["virginia nurse quarterly"] = "Va Nurse Q",
  ["virginia review of sociology"] = "Va Rev Sociol",
  ["virologica sinica"] = "Virol Sin",
  ["virologie (montrouge, france)"] = "Virologie (Montrouge)",
  ["virology & mycology : infectious diseases"] = "Virol Mycol",
  ["virology (hyderabad)"] = "Virology (Hyderabad)",
  ["virology : research and treatment"] = "Virology (Auckl)",
  ["virology journal"] = "Virol J",
  ["virology monographs"] = "Virol. Monogr.",
  ["virology monographs. die virusforschung in einzeldarstellungen"] = "Virol Monogr",
  ["virology reports"] = "Virol. Rep.",
  ["virtual and physical prototyping"] = "Virtual Phys Prototyp",
  ["virtual journal of atomic quantum fluids"] = "Virtual J. At. Quantum Fluids",
  ["virtual reality"] = "Virtual Real",
  ["virtual series on symplectic geometry"] = "Virtual Ser. Symplectic Geom.",
  ["virtual, augmented and mixed reality : industrial and everyday life applications : 12th international conference, vamr 2020, held as part of the 22nd hci international conference, hcii 2020, copenhagen, denmark, july 19-24, 2020, proceedings. part ii. vamr (conference) (12th : 2020 : online)"] = "Virtual Augment Mixed Real (2020)",
  ["virus evolution"] = "Virus Evol",
  ["virus genes"] = "Virus Genes",
  ["virus information exchange newsletter for south-east asia and the western pacific"] = "Virus Inf Exch Newsl South East Asia West Pac",
  ["virus research"] = "Virus Res",
  ["virus research. supplement"] = "Virus Res Suppl",
  ["virus reviews & research : journal of the brazilian society for virology"] = "Virus Rev Res",
  ["visceral medicine"] = "Visc Med",
  ["vishwa international journal of graph theory"] = "Vishwa Int. J. Graph Theory",
  ["visible language"] = "Visible Lang",
  ["visible religion"] = "VisRel",
  ["visible religion: annual for religious iconography"] = "VRel",
  ["vision (basel, switzerland)"] = "Vision (Basel)",
  ["vision development and rehabilitation"] = "Vis Dev Rehabil",
  ["vision research"] = "Vision Res",
  ["vision systems design"] = "Vision Syst. Des.",
  ["visnyk kharkivsʹkoho nat︠s︡ionalʹnoho universytetu imeni v.n. karazina. serii︠a︡ biolohii︠a︡"] = "Visnik Harkivskogo Nacionalnogo Universitetu Imeni V N Karazina Seria Biologia",
  ["visnyk kyïvsʹkoho nat︠s︡ionalʹnoho universytetu imeni tarasa shevchenka. biolohii︠a︡"] = "Visnyk Kyivskoho Natsionalnoho Universytetu Imeni Tarasa Shevchenka Biolohiia",
  ["visnyk l'vivs'koho universytetu. seriia biolohichna"] = "Visnyk Lvivskoho Universytetu Seriia Biolohichna",
  ["visnyk lʹvivsʹkoho universytetu. seriia biolohichna"] = "Visnyk Lvivskoho Universytetu Seriia Biolohichna",
  ["visnyk ukraïnsʹkoho tovarystva henetykiv i selekt︠s︡ioneriv"] = "Vis Ukr Tov Genet Sel",
  ["vistas in astronomy"] = "Vistas in Astron.",
  ["visual anthropology"] = "Vis Anthropol",
  ["visual anthropology review : journal of the society for visual anthropology"] = "Vis Anthropol Rev",
  ["visual cognition"] = "Vis cogn",
  ["visual computer"] = "Visual Comput.",
  ["visual computing for industry, biomedicine, and art"] = "Vis Comput Ind Biomed Art",
  ["visual culture in britain"] = "Vis Cult Br",
  ["visual geosciences"] = "Visual Geosci.",
  ["visual impairment research"] = "Vis Impair Res",
  ["visual journal of emergency medicine"] = "Vis J Emerg Med",
  ["visual neuroscience"] = "Vis Neurosci",
  ["visual studies"] = "Vis Stud",
  ["visualization and data analysis"] = "Vis Data Anal",
  ["visualization in engineering"] = "Visualization Eng.",
  ["visualization of mechanical processes: an international online journal"] = "Visualization Mech. Processes: Int. Online J.",
  ["visualization, image processing and computation in biomedicine"] = "Visualization Image Proc. Comput. Biomed.",
  ["visualized cancer medicine"] = "Visualized Cancer Med.",
  ["vita humana. internationale zeitschrift fur lebensaltersforschung. international journal of human development. journal international de developpement humain"] = "Vita Hum Int Z Lebensalterforsch",
  ["vita humana. internationale zeitschrift für lebensaltersforschung. international journal of human development. journal international de développement humain"] = "Vita Hum Int Z Lebensalterforsch",
  ["vita latina"] = "VL",
  ["vita malacologica"] = "Vita Malacol",
  ["vita mathematica"] = "Vita Math.",
  ["vitae scholasticae"] = "Vitae Scholasticae",
  ["vital & health statistics. series 14, data from the national health survey"] = "Vital Health Stat 14",
  ["vital & health statistics. series 14, data from the national inventory of family planning services"] = "Vital Health Stat 14",
  ["vital & health statistics. series 3, analytical and epidemiological studies"] = "Vital Health Stat 3",
  ["vital & health statistics. series 3, analytical and epidemiological studies / [u.s. dept. of health and human services, public health service, national center for health statistics]"] = "Vital Health Stat 3",
  ["vital and health statistics. ser. 1, programs and collection procedures"] = "Vital Health Stat 1",
  ["vital and health statistics. ser. 4, documents and committee reports"] = "Vital Health Stat 4",
  ["vital and health statistics. ser. 4: documents and committee reports"] = "Vital Health Stat 4",
  ["vital and health statistics. series 10, data from the national health survey"] = "Vital Health Stat 10",
  ["vital and health statistics. series 10: data from the national health survey"] = "Vital Health Stat. 10.",
  ["vital and health statistics. series 11, data from the national health survey"] = "Vital Health Stat 11",
  ["vital and health statistics. series 11: data from the national health survey"] = "Vital Health Stat. 11.",
  ["vital and health statistics. series 13, data from the national health survey"] = "Vital Health Stat 13",
  ["vital and health statistics. series 13: data from the national health survey"] = "Vital Health Stat. 13.",
  ["vital and health statistics. series 14, data on national health resources"] = "Vital Health Stat 14",
  ["vital and health statistics. series 14: data from the national health survey"] = "Vital Health Stat. 14.",
  ["vital and health statistics. series 14: data from the national inventory of family planning services"] = "Vital Health Stat. 14.",
  ["vital and health statistics. series 14: data on national health resources"] = "Vital Health Stat. 14.",
  ["vital and health statistics. series 16, compilations of advance data from vital and health statistics"] = "Vital Health Stat Ser 16",
  ["vital and health statistics. series 1: programs and collection procedures"] = "Vital Health Stat. 1.",
  ["vital and health statistics. series 2, data evaluation and methods research"] = "Vital Health Stat 2",
  ["vital and health statistics. series 20, data from the national vital statistics system"] = "Vital Health Stat 20",
  ["vital and health statistics. series 20, data on mortality"] = "Vital Health Stat 20 Data Mortal",
  ["vital and health statistics. series 20: data from the national vital statistics system"] = "Vital Health Stat. 20. Data Natl. Vital Stat. Syst.",
  ["vital and health statistics. series 20: data from the national vitalstatistics system vital health stat 20 data natl vital sta"] = "Syst",
  ["vital and health statistics. series 21, data from the national vital statistics system"] = "Vital Health Stat 21",
  ["vital and health statistics. series 21, data on natality, marriage, and divorce"] = "Vital Health Stat 21",
  ["vital and health statistics. series 21: data from the national vital statistics system"] = "Vital Health Stat. 21.",
  ["vital and health statistics. series 21: data on natality, marriage, and divorce"] = "Vital Health Stat. 21.",
  ["vital and health statistics. series 23, data from the national survey of family growth"] = "Vital Health Stat 23",
  ["vital and health statistics. series 23: data from the national survey of family growth"] = "Vital Health Stat. 23.",
  ["vital and health statistics. series 24, compilations of data on natality, mortality, marriage, divorce, and induced terminations of pregnancy"] = "Vital Health Stat Ser 24",
  ["vital and health statistics. series 2: data evaluation and methods research"] = "Vital Health Stat. 2.",
  ["vital and health statistics. series 3, analytical studies"] = "Vital Health Stat 3",
  ["vital and health statistics. series 3: analytical and epidemiological studies"] = "Vital Health Stat. 3.",
  ["vital and health statistics. series 3: analytical studies"] = "Vital Health Stat. 3.",
  ["vital and health statistics. series 4: documents and committee reports"] = "Vital Health Stat. 4.",
  ["vital and health statistics. series 5, comparative international vital and health statistics reports"] = "Vital Health Stat 5",
  ["vital and health statistics. series 5: comparative international vital and health statistics reports"] = "Vital Health Stat. 5.",
  ["vital signs"] = "Vital Signs",
  ["vital speeches of the day"] = "Vital Speeches Day",
  ["vital statistics"] = "Vital Stat (Ott Annu Ed)",
  ["vital statistics of the united states"] = "Vital Stat U S (Hyattsville)",
  ["vital statistics, special reports"] = "Vital Stat Spec Rep",
  ["vitamin d digest"] = "Vitam D Dig",
  ["vitamine und hormone"] = "Vitam Horm Leipzig",
  ["vitamins and hormones"] = "Vitam Horm",
  ["vitamins and hormones (new york)"] = "Vitam. Horm. (N.Y.)",
  ["viticulture data journal"] = "Vitic. Data J.",
  ["viva origino. [origin of life]"] = "Viva Orig",
  ["vivarium scyllacense. bollettino dell’istituto di studi su cassiodoro e sul medioevo in calabria"] = "VivScyl",
  ["vivarium: a journal for mediaeval philosophy and the intellectual life of the middle ages"] = "Vivarium",
  ["vivechan international journal of research"] = "Vivechan Int J Res",
  ["vivre autrement"] = "Vivre Autrement",
  ["vivre en val-d'oise"] = "Vivre Val D",
  ["vizantijskij vremennik"] = "VizVrem",
  ["vjesnik arheološkog muzeja u zagrebu"] = "VjesAMuzZagreb",
  ["vjesnik za arheologiju i historiju dalmatinsku. bulletin d’archéologie et d’histoire dalmates"] = "VjesDal",
  ["vlaams diergeneeskundig tijdschrift"] = "Vlaams Diergeneeskd Tijdschr",
  ["vladikavkazskii matematicheskii zhurnal"] = "Vladikavkaz. Mat. Zh.",
  ["vldb journal"] = "VLDB J.",
  ["vlsi design"] = "VLSI Des.",
  ["vnitrni lekarstvi"] = "Vnitr Lek",
  ["vnitr̆ní lékar̆ství"] = "Vnitr Lek",
  ["vnr new mathematics library"] = "VNR New Math. Library",
  ["voces desde el trabajo social"] = "Voces Desde Trab Soc",
  ["voenno meditsinsko delo"] = "Voen Med Delo",
  ["voenno-istoricheskii zhurnal"] = "Voen Istor Zh",
  ["voenno-meditsinskii zhurnal"] = "Voen Med Zh",
  ["voenno-meditsinskiĭ sbornik"] = "Voen Meditsinskii Sb",
  ["voenno-medit︠s︡inskiĭ zhurnal"] = "Voen Med Zh",
  ["vogelwarte, die"] = "Vogelwarte",
  ["vogelwelt, die"] = "Vogelwelt",
  ["voice & speech review"] = "Voice Speech Rev",
  ["voix dentaire"] = "Voix Dent.",
  ["vojenske zdravotnicke listy"] = "Vojen. Zdrav. Listy",
  ["vojenské zdravotnické listy"] = "Vojen Zdrav Listy",
  ["vojenskozdravotnicka knihovna"] = "Vojenskozdrav. Knih.",
  ["vojenskozdravotnická knihovna"] = "Vojenskozdrav Knih",
  ["vojno-istoriski glasnik"] = "Vojnois Glas",
  ["vojnosanitetski pregled"] = "Vojnosanit Pregl",
  ["vojnosanitetski pregled. military-medical and pharmaceutical review"] = "Vojnosanit Pregl",
  ["volksgesundheit (bern, switzerland)"] = "Volksgesundheit Bad Hombg",
  ["volkskunde (amsterdam, netherlands)"] = "Volkskunde",
  ["volta voices"] = "Volta Voices",
  ["volucella : (malsch)"] = "Volucella",
  ["volume graphics. international symposium on volume graphics"] = "Vol Graph",
  ["voluntary action leadership"] = "Volunt Action Leadersh",
  ["voluntas : international journal of voluntary and nonprofit organizations"] = "Voluntas",
  ["volunteer administration"] = "Volunt Adm",
  ["volunteer leader"] = "Volunt. Leader",
  ["volúmenes de homenaje"] = "Vol. Homenaje",
  ["vom zahlstein zum computer"] = "Vom Zahlstein zum Comput.",
  ["voprosy biokhimii mozga"] = "Vopr Biokhim Mozga",
  ["voprosy biokhimii. [problems of biochemistry]"] = "Vopr Biokhimii",
  ["voprosy ekonomiki / akademiia nauk sssr, institut ekonomiki"] = "Vopr Ekon",
  ["voprosy eksperimentalnoi i klinicheskoi urologii"] = "Vopr Eksper Klinicheskoi Urol",
  ["voprosy elektropatologii, elektrotravmatizma i elektrobezopasnosti"] = "Vopr Elektropatol Elektrotravmatizma Elektrobezop",
  ["voprosy ėkonomiki"] = "Vopr Ekon",
  ["voprosy e̊lektropatologii, e̊lektrotravmatizma i e̊lektrobezopasnosti"] = "Vopr Elektropatol Elektrotravmatizma Elektrobezop",
  ["voprosy filosofii"] = "Vopr Filos",
  ["voprosy filosofii (problèmes de philosophie)"] = "VopFilos",
  ["voprosy filosofii / akademiia nauk sssr, institut filosofii"] = "Vopr Filos",
  ["voprosy fiziologii"] = "Vopr Fiziol",
  ["voprosy geografii"] = "Vopr Geogr",
  ["voprosy istorii (moscow, russia : 1945)"] = "Vopr Istor",
  ["voprosy istorii (problèmes d’histoire)"] = "VopIst",
  ["voprosy istorii estestvoznaniia i tekhniki (institut istorii estestvoznaniia i tekhniki (akademiia nauk sssr))"] = "Vopr Istor Estestvozn Tekh",
  ["voprosy istorii estestvoznanii︠a︡ i tekhniki (institut istorii estestvoznanii︠a︡ i tekhniki (akademii︠a︡ nauk sssr))"] = "Vopr Istor Estestvozn Tekh",
  ["voprosy istorii kpss (moscow, russia)"] = "Vopr Istor KPSS",
  ["voprosy klassiceskoj filologii"] = "VKF",
  ["voprosy klinicheskoĭ i eksperimental'noĭ oftalmologiĭ"] = "Vopr Klin Eksp Oftalmol",
  ["voprosy klinicheskoĭ i e̊ksperimental'noĭ urologii; sbornik trudov"] = "Vopr Klin Eksp Urol",
  ["voprosy kliniki i lecheniia zlokachestvennykh novoobrazovanii"] = "Vopr Klin Lecheniia Zlokachestvennykh Novoobraz",
  ["voprosy kliniki i lecheniia zlokachestvennykh novoobrazovaniĭ"] = "Vopr Klin Lecheniia Zlokachestvennykh Novoobraz",
  ["voprosy kurortologii, fizioterapii i lechebnoi fizicheskoi kultury"] = "Vopr. Kurortol. Fizioter. Lech. Fiz. Kult.",
  ["voprosy kurortologii, fizioterapii, i lechebnoi fizicheskoi kultury"] = "Vopr Kurortol Fizioter Lech Fiz Kult",
  ["voprosy kurortologii, fizioterapii, i lechebnoĭ fizicheskoĭ kultury"] = "Vopr Kurortol Fizioter Lech Fiz Kult",
  ["voprosy meditsinskof virusologii"] = "Vopr Med Virusol",
  ["voprosy meditsinskoi khimii"] = "Vopr Med Khim",
  ["voprosy meditsinskoĭ khimii"] = "Vopr Med Khimii",
  ["voprosy medit︠s︡inskoĭ khimii"] = "Vopr Med Khim",
  ["voprosy neirokhirurgii"] = "Vopr. Neirokhir.",
  ["voprosy neĭrokhirurgii"] = "Vopr Neirokhir",
  ["voprosy okhrany materinstva i detstva"] = "Vopr Okhr Materin Det",
  ["voprosy onkologii"] = "Vopr Onkol",
  ["voprosy optimalnogo planirovaniya i upravleniya sotsialisticheskoi ekonomikoi"] = "Voprosy Optim. Plan. Upravl. Sots. Ekonom.",
  ["voprosy patologii serdechno-sosudistoi sistemy"] = "Vopr Patol Serdechno Sosud Sist",
  ["voprosy patologii serdechno-sosudistoĭ sistemy"] = "Vopr Patol Serdechno Sosud Sist",
  ["voprosy pediatrii"] = "Vopr Pediatrii",
  ["voprosy pitaniia"] = "Vopr Pitan",
  ["voprosy psikhiatrii i nevropatologii"] = "Vopr. Psikhiatr. Nevropatol.",
  ["voprosy psikhiatrii i nevropatologii; sbornik trudov"] = "Vopr Psikhiatr Nevropatol",
  ["voprosy psikhīatrīi i nevrologīi"] = "Vopr Psikhiatrii Nevrol",
  ["voprosy psikhologii"] = "Vopr Psikhol",
  ["voprosy revmatizma"] = "Vopr Revm",
  ["voprosy sotsial'noĭ gigieny, organizatsii zdravookhraneniia i istorii meditsiny"] = "Vopr Sots Gig Organ Zdravookhr Istor Med",
  ["voprosy statistiki (moscow, russia : 1994)"] = "Vopros Stat",
  ["voprosy virusologii"] = "Vopr Virusol",
  ["voprosy vychislitelnoi i prikladnoi matematiki. hisoblash va amaliy matematika masalalari"] = "Vopr. Vychisl. Prikl. Mat.",
  ["voprosy âzykoznaniâ (problèmes de linguistique)"] = "Vopåz",
  ["vorarlberger naturschau"] = "Vorarlb Naturschau",
  ["vorlesungen aus dem fachbereich mathematik der universität gh essen"] = "Vorlesungen Fachbereich Math. Univ. Essen",
  ["vorlesungen aus dem mathematischen institut giessen"] = "Vorlesungen Math. Inst. Giessen",
  ["vorläufiger bericht über die von dem deutschen archäologischen institut und der deutschen orient-gesellschaft aus den mitteln der deutschen forschungsgemeinschaft unternommenen ausgrabungen in uruk-warka"] = "UVB",
  ["vostok (moscow, russia : 1991)"] = "Vostok (Mosk)",
  ["vox latina: commentarii periodici"] = "VoxLat",
  ["vox patrum"] = "VoxP",
  ["vox romanica: annales helvetici explorandis linguis romanicis destinati"] = "VR",
  ["vox sanguinis"] = "Vox Sang",
  ["voyages en mathematiques"] = "Voyag. Math.",
  ["vrachebnoe delo"] = "Vrach Delo",
  ["vtt symposium"] = "VTT Symp.",
  ["vulnerable children and youth studies"] = "Vulnerable Child Youth Stud",
  ["vutreshni bolesti"] = "Vutr. Boles.",
  ["vŭtreshni bolesti"] = "Vutr Boles",
  ["vychislitel\\cprime naya i prikladnaya matematika"] = "Vychisl. Prikl. Mat. (Kiev)",
  ["vychislitel\\cprime naya tekhnika"] = "Vychisl. Tekhn. Sistemy Upravlenie",
  ["vychislitel\\cprime naya tekhnika i voprosy kibernetiki"] = "Vychisl. Tekhn. Vopr. Kibern.",
  ["vychislitelnaya tekhnika sotsialisticheskikh stran"] = "Vychisl. Tekhn. Sots. Stran",
  ["vychislitelnye tekhnologii. computational technologies"] = "Vychisl. Tekhnol.",
  ["vychislitelnyi tsentr moskovskogo gosudarstvennogo universiteta. seriya: statistika i stokhasticheskie sistemi,"] = "Vychislitelnyi Tsentr Moskovskogo Gosudarstvennogo Universiteta. Seriya: Statistika i Stokhasticheskie",
  ["vychislitelnyi tsentr moskovskogo gosudarstvennogo universiteta. seriya: statistika i stokhasticheskie sistemy"] = "Vychislitelnyi Tsentr Moskovskogo Gosudarstvennogo Universiteta. Seriya: Statistika i Stokhasticheskie",
  ["vychislitelnyi tsentr moskovskogo gosudarstvennogo universiteta. seriya: statistika i stokhasticheskie sistemy,"] = "Vychislitelnyi Tsentr Moskovskogo Gosudarstvennogo Universiteta. Seriya: Statistika i Stokhasticheskie",
  ["vysokochistye veshchestva"] = "Vysokochist. Veshchestva",
  ["vysokomolekulyarnye soedineniya, seriya a"] = "Vysokomol. Soedin., Ser. A",
  ["vysokomolekulyarnye soedineniya, seriya a i seriya b"] = "Vysokomol. Soedin., Ser. A Ser. B",
  ["vysokomolekulyarnye soedineniya, seriya b: kratkie soobshcheniya"] = "Vysokomol. Soedin., Ser. B",
  ["vysoké učení technické v brně"] = "Sb. Věd. Odb. Prací",
  ["vyziva lidu"] = "Vyziva Lidu",
  ["výz̆iva lidu"] = "Vyziva Lidu",
  ["vögel der heimat"] = "Vögel Heim.",
  ["w. dittenberger, orientis graeci inscriptiones selectae (leipzig 1903–­1905)"] = "OGIS",
  ["w. dittenberger, sylloge inscriptionum graecarum (leipzig 1915–­1924)"] = "SIG",
  ["w. h. roscher, ausführliches lexikon der griechischen und römischen mythologie"] = "RoscherML",
  ["w. helbig, führer durch die öffentlichen sammlungen klassischer altertümer in rom"] = "Helbig",
  ["w. von soden, akkadisches handwörterbuch (wiesbaden 1965–1981)"] = "AHw",
  ["wad-al-hayara. revista de estudios de la institución provincial de cultura »marqués de santillana« de guadalajara"] = "Wad-al-Hayara",
  ["wadc technical note. united states. air force. wright air development center, dayton, ohio"] = "WADC Tech Note United States Air Force Wright Air Dev Cen Day Ohio",
  ["wadc technical note. wright air development center"] = "WADC Tech Note United States Air Force Wright Air Dev Cen Day Ohio",
  ["wadc technical report. united states. air force. wright air development center, dayton, ohio"] = "WADC Tech Rep United States Air Force Wright Air Dev Cent Day Ohio",
  ["wadc technical report. wright air development center"] = "WADC Tech Rep United States Air Force Wright Air Dev Cent Day Ohio",
  ["waimm journal of earth resources engineering"] = "WAIMM J. Earth Resour. Eng.",
  ["wakayama medical reports"] = "Wakayama Med Rep",
  ["wakayama university"] = "Bull. Fac. Ed. Wakayama Univ. Natur. Sci.",
  ["wakayama university. faculty of education. bulletin. natural science"] = "Bull. Fac. Ed. Wakayama Univ. Natur. Sci.",
  ["wake forest law review"] = "Wake Forest Law Rev",
  ["waking and sleeping"] = "Waking Sleeping",
  ["walailak journal of science and technology"] = "Walailak J Sci Technol",
  ["wald und holz"] = "Wald Holz",
  ["wald, der"] = "Wald",
  ["wald- und holzwirtschaft, die"] = "Wald- Holzwirtsch.",
  ["waldarbeit, die"] = "Waldarbeit",
  ["walkerana : transactions of the poets society"] = "Walkerana (Ann Arbor Mich)",
  ["wall street journal (eastern ed.)"] = "Wall St J (East Ed)",
  ["wall street journal (midwest ed.)"] = "Wall St J Midwest Ed",
  ["wall street journal (southwest ed.)"] = "Wall Str J",
  ["wall street journal. midwest edition"] = "Wall St. J. [Midwest. Ed.]",
  ["walter eucken institut. wirtschaftswissenschaftliche und wirtschaftsrechtliche untersuchungen"] = "Walter Eucken Inst. Wirtschaftswissensch. Wirtschaftsrechtl. Unters.",
  ["war & society"] = "War Soc",
  ["war department technical bulletin. tb med"] = "War Dep Tech Bull TB Med",
  ["war in history"] = "War Hist",
  ["war medicine"] = "War Med (Chic 1941)",
  ["war on hunger; a report from the agency for international development"] = "War Hung",
  ["warasan phesatchasat"] = "Mahidol University journal of pharmaceutical sciences / Mahawitthayalai Mahidon|Warasan Phesatchasat",
  ["warasan phrueksasat thai"] = "Warasan Phrueksasat Thai",
  ["warasan phuetsat songkhlanakharin"] = "Warasan Phuetsat Songkhlanakharin",
  ["warasan songkhla nakharin / mahawitthayalai songkhla nakharin, witthayakhet hat yai"] = "Songklanakarin journal / Prince of Songkla University, Hat Yai Campus|Warasan Songkhla Nakharin",
  ["warasan technology suranaree"] = "Warasan Technol Suranaree",
  ["warasan wichai lae songsoem wichakan kaset"] = "Warasan Wichai Lae Songsoem Wichakan Kaset",
  ["warasan wichai mo kho. (chobap bandit sueksa)"] = "Warasan Wichai Mo Kho (Chobap Bandit Sueksa)",
  ["warasan wichakan witthayasat lae the noloyi mahawithayalai ratchaphat nakhon sawan"] = "Warasan Wichakan Witthayasat Lae The Noloyi Mahawithayalai Ratchaphat Nakhon Sawan",
  ["warasan witthayasart burapha"] = "Warasan Witthayasart Burapha",
  ["warasan witthayasat mahawitthayalai naresuan"] = "Warasan Witthayasat Mahawitthayalai Naresuan",
  ["warasān wičhai mō̧.khō̧. = kku research journal"] = "Warasan Wichai Mo Kho",
  ["warburg institute studies and texts"] = "Warburg Inst. Stud. Texts",
  ["warreport : bulletin of the institute for war & peace reporting"] = "War Rep",
  ["warsan wichai witthayasat / khana witthayasat chulalongkon mahawitthayalai"] = "The Journal of scientific research / Faculty of Science, Chulalongkorn University|J. Sci. Res. Chulalongkorn Univ.",
  ["warta demografi"] = "Warta Demografi",
  ["warwickshire history"] = "Warwickshire Hist",
  ["wash [reports]. u.s. atomic energy commission"] = "WASH Rep",
  ["wash reports"] = "WASH Rep.",
  ["washburn law journal"] = "Washburn Law J",
  ["washington and lee law review"] = "Wash Lee Law Rev",
  ["washington drug & device letter"] = "Wash Drug Device Lett",
  ["washington journal of environmental law & policy"] = "Wash J Environ Law Policy",
  ["washington law review (seattle, wash. : 1962)"] = "Wash Law Rev",
  ["washington memo"] = "Wash Memo Alan Guttmacher Inst",
  ["washington nurse"] = "Wash. Nurse",
  ["washington post (washington, d.c. : 1974)"] = "Washington Post",
  ["washington report on medicine & health"] = "Wash Rep Med Health",
  ["washington report on medicine and health"] = "Wash. Rep. Med. Health",
  ["washington state dental journal"] = "Wash State Dent J",
  ["washington state journal of nursing"] = "Wash State J Nurs",
  ["washington state journal of public health practice"] = "Wash State J Public Health Pract",
  ["washington times (washington, d.c. : 1982)"] = "Wash Times",
  ["washington university dental journal"] = "Wash Univ Dent J",
  ["washington university journal of american indian & alaska native health"] = "Wash Univ J Am Indian Alsk Native Health",
  ["washington university journal of urban and contemporary law"] = "Wash Univ J Urban Contemp Law",
  ["washington university law quarterly. washington university (saint louis, mo.). school of law"] = "Wash Univ Law Q",
  ["wasser und boden"] = "Wasser Boden",
  ["wasser, energie, luft"] = "Wasser energ. luft",
  ["wasser- und energiewirtschaft"] = "Wasser- Energ.wirtsch.",
  ["wasserwirtschaft wassertechnik"] = "Wasserwirtsch. Wassertech.",
  ["waste and biomass valorization"] = "Waste Biomass Valorization",
  ["waste disposal & sustainable energy"] = "Waste Dispos Sustain Energy",
  ["waste management"] = "Waste Manage. (Oxford)",
  ["waste management & research : the journal of the international solid wastes and public cleansing association, iswa"] = "Waste Manag Res",
  ["waste management (new york, n.y.)"] = "Waste Manag",
  ["waste management (oxford)"] = "Waste Manage. (Oxford)",
  ["waste management (tucson, arizona)"] = "Waste Manage. (Tucson, Ariz.)",
  ["waste management and research"] = "Waste Manage. Res.",
  ["waste treatment and recovery"] = "Waste Treat. Recovery",
  ["water & sewage works"] = "Water Sew Works",
  ["water alternatives"] = "Water Altern.",
  ["water and energy international"] = "Water Energy Int.",
  ["water and environment journal"] = "Water Environ. J.",
  ["water and environment journal : the journal"] = "Water Environ J",
  ["water and environmental management"] = "Water Environ. Manage.",
  ["water asset management international"] = "Water Asset Manage. Int.",
  ["water biology and security"] = "Water Biol. Secur.",
  ["water conservation and management"] = "Water Conserv. Manage.",
  ["water conservation science and engineering"] = "Water Conserv. Sci. Eng.",
  ["water cycle"] = "Water Cycle",
  ["water economics and policy"] = "Water Econ. Policy",
  ["water environment journal"] = "Water Environ. J.",
  ["water environment research"] = "Water Environ. Res.",
  ["water environment research : a research publication of the water environment federation"] = "Water Environ Res",
  ["water history"] = "Water Hist",
  ["water intelligence online"] = "Water Intell. Online",
  ["water international"] = "Water Int.",
  ["water policy"] = "Water Policy",
  ["water policy : official journal of the world water council"] = "Water Policy",
  ["water practice & technology"] = "Water Pract Technol",
  ["water practice technology"] = "Water Pract. Technol.",
  ["water quality and ecosystems modeling"] = "Water Qual. Ecosyst. Model.",
  ["water quality research journal"] = "Water Qual. Res. J.",
  ["water quality research journal of canada"] = "Water Qual. Res. J. Can.",
  ["water quality, exposure and health"] = "Water Qual. Exposure Health",
  ["water quality, exposure, and health"] = "Water Qual Expo Health",
  ["water research"] = "Water Res",
  ["water research x"] = "Water Res X",
  ["water resources"] = "Water Resour.",
  ["water resources and economics"] = "Water Resour. Econ.",
  ["water resources and industry"] = "Water Resour. Ind.",
  ["water resources and rural development"] = "Water Resour. Rural Dev.",
  ["water resources investigations report (united states geological survey)"] = "Water Resour. Invest. Rep. U.S. Geol. Surv.",
  ["water resources management"] = "Water Resour. Manage.",
  ["water resources protection"] = "Water Resour. Prot.",
  ["water resources research"] = "Water Resour. Res.",
  ["water reuse"] = "Water Reuse",
  ["water sa"] = "Water SA",
  ["water science"] = "Water Sci.",
  ["water science & technology, water supply"] = "Water Sci Technol Water Supply",
  ["water science and engineering"] = "Water Sci. Eng.",
  ["water science and technology"] = "Water Sci. Technol.",
  ["water science and technology : a journal of the international association on water pollution research"] = "Water Sci Technol",
  ["water science and technology library"] = "Water Sci. Technol. Lib.",
  ["water science and technology water supply"] = "Water Sci. Technol. Water Supply",
  ["water security"] = "Water Secur",
  ["water sewage and effluent"] = "Water Sewage Effluent",
  ["water supply"] = "Water Supply",
  ["water utilities management international"] = "Water Util. Manage. Int.",
  ["water waves"] = "Water Waves",
  ["water waves. an interdisciplinary journal"] = "Water Waves",
  ["water works and sewerage"] = "Water Works Sewerage",
  ["water, air and soil pollution"] = "Water Air Soil Pollut.",
  ["water, air, & soil pollution"] = "Water Air Soil Pollut.",
  ["water, air, & soil pollution: focus"] = "Water Air Soil Pollut. Focus",
  ["water, air, and soil pollution"] = "Water Air Soil Pollut",
  ["water, bodem, lucht"] = "Water Bodem Lucht",
  ["water-energy nexus"] = "Water-Energy Nexus",
  ["watermark (archivists and librarians in the history of the health sciences)"] = "Watermark (Arch Libr Hist Health Sci)",
  ["watermark (association of librarians in the history of the health sciences)"] = "Watermark",
  ["watershed ecology and the environment"] = "Watershed Ecol. Environ.",
  ["wave motion"] = "Wave Motion",
  ["wave motion (north-holland publishing company)"] = "Wave Motion",
  ["wave motion. an international journal reporting research on wave phenomena"] = "Wave Motion",
  ["wavelet analysis and its applications"] = "Wavelet Anal. Appl.",
  ["waves in random and complex media"] = "Waves Random Complex Medium",
  ["waves in random and complex media. propagation, scattering and imaging"] = "Waves Random Complex Media",
  ["waves in random media"] = "Waves Random Media",
  ["waves, wavelets and fractals"] = "Waves Wavelets Fractals",
  ["wayne law review"] = "Wayne Law Rev",
  ["wārasān kānpramong = thai fisheries gazette"] = "Warasan Kanpramong",
  ["wārasān khana witthayāsāt, mahāwitthayālai chīang mai = journal of the science faculty of chiang mai university"] = "Warasan Khana Witthayasat Maha Witthayalai Chiang Mai",
  ["wārasān phēsatchasāt = mahidol university journal of pharmaceutical sciences"] = "Warasan Phesatchasat",
  ["wārasān prachākō̜n læ sangkhom = journal of population and social studies"] = "Warasan Prachakon Lae Sangkhom",
  ["wārasān rōk phư̄t"] = "Warasan Rok Phut",
  ["wārasān songkhlā nakharin"] = "Warasan Songkhla Nakharin",
  ["wārasān wichākān kasēt : thai agricultural research journal"] = "Warasan Wichakan Kaset",
  ["wārasān wičhai witthayāsāt kānphǣt = thai journal of health research"] = "Warasan Wichai Witthayasat Kanphaet",
  ["wārsān wičhai witthayāsāt"] = "J Sci Res Chulalongkorn Univ",
  ["wda journal"] = "WDA J",
  ["we international"] = "WE Int",
  ["we the blind"] = "We Blind",
  ["wear : an international journal on the science and technology of friction lubrication and wear"] = "Wear",
  ["wearable technologies"] = "Wearable Technol.",
  ["weather and climate dynamics"] = "Weather Clim. Dyn.",
  ["weather and climate extremes"] = "Weather Clim. Extremes",
  ["weather and forecasting"] = "Weather Forecast",
  ["weather vane"] = "Weather Vane",
  ["weather, climate, and society"] = "Weather Clim. Soc.",
  ["weather, climate, and society (print)"] = "Weather Clim Soc",
  ["web ecology"] = "Web Ecol.",
  ["web semantics (online)"] = "Web Semant",
  ["weed & turfgrass science"] = "Weed Turfgrass Sci",
  ["weed biology and management"] = "Weed Biol Manag",
  ["weed research"] = "Weed Res",
  ["weed science"] = "Weed Sci",
  ["weed technology"] = "Weed Technol.",
  ["weed technology : a journal of the weed science society of america"] = "Weed Technol",
  ["weeds world : the international electronic arabidopsis newsletter"] = "Weeds World",
  ["weekend argus"] = "Weekend Argus",
  ["weekend magazine"] = "Weekend Mag",
  ["weekly abstract of sanitary reports. united states. marine hospital service"] = "Wkly Abstr Sanit Rep",
  ["weekly bulletin; epidemiological information received. interim commission of the world health organization. epidemiological intelligence station, singapore"] = "Wkly Bull Epidemiol Inf Receiv",
  ["weekly epidemiological record"] = "Wkly. Epidemiol. Rec.",
  ["weekly mail (johannesburg, south africa)"] = "Weekly Mail",
  ["weekly record of convention diseases at ports and other localities at home and abroad. great britain. ministry of health"] = "Wkly Rec Conv Dis Ports Other Localities Home Abroad",
  ["weekly report. illinois. division of communicable diseases"] = "Wkly Rep Ill",
  ["weevil news"] = "Weevil News",
  ["wehrmedizinische mitteilungen"] = "Wehrmed Mitt",
  ["wehrmedizinische monatsschrift"] = "Wehrmed Monatsschr",
  ["wei sheng wu hsueh pao [acta microbiologica sinica]"] = "Wei Sheng Wu Hsueh Pao",
  ["wei sheng wu xue bao (acta microbiologica sinica)"] = "Wei Sheng Wu Xue Bao",
  ["wei sheng wu xue bao = acta microbiologica sinica"] = "Wei Sheng Wu Xue Bao",
  ["wei sheng wu xue mian yi xue jin zhan"] = "Wei Sheng Wu Xue Mian Yi Xue Jin Zhan",
  ["wei sheng wu xue tong bao"] = "Wei Sheng Wu Xue Tong Bao",
  ["wei sheng wu xue za zhi"] = "Wei Sheng Wu Xue Za Zhi",
  ["wei sheng wu yu gan ran"] = "Wei Sheng Wu Yu Gan Ran",
  ["wei sheng yan jiu (journal of hygiene research)"] = "Wei Sheng Yan Jiu",
  ["wei sheng yan jiu = journal of hygiene research"] = "Wei Sheng Yan Jiu",
  ["wei sheng yen chiu journal of hygiene research (pei-ching shih"] = "Wei Sheng Yen Chiu",
  ["wei ti gu sheng wu xue bao"] = "Acta micropalaeontologica Sinica|Wei Ti Gu Sheng Wu Xue Bao",
  ["wei ti gu sheng wu xue bao = acta micropalaeontologica sinica"] = "Wei Ti Gu Sheng Wu Xue Bao",
  ["weichangbingxue he ganbingxue zazhi"] = "Chinese journal of gastroenterology and hepatology|Weichangbingxue He Ganbingxue Zazhi",
  ["weichangbingxue he ganbingxue zazhi = chinese journal of gastroenterology and hepatology"] = "Weichangbingxue He Ganbingxue Zazhi",
  ["wein wissenschaft"] = "Wein Wiss.",
  ["weinwirtschaft technik"] = "Weinwirtsch. Tech.",
  ["welding in the world"] = "Weld. World",
  ["welding international"] = "Weld. Int.",
  ["welding journal"] = "Weld J",
  ["weleda korrespondenzblatter fur aerzte"] = "Weleda Korrespondenzbl Aerzte",
  ["weleda korrespondenzblätter für aerzte"] = "Weleda Korrespondenzbl Aerzte",
  ["welfare in review"] = "Welf Rev",
  ["wellbeing, space and society"] = "Wellbeing Space Soc.",
  ["wellcome open research"] = "Wellcome Open Res",
  ["weltwirtschaftliches archiv"] = "Weltwirtsch Arch",
  ["wen shi zhe xue bao"] = "Bull. Coll. Lib. Arts",
  ["wen shi zhe xue bao / guo li taiwan da xue"] = "Bull Coll Lib Arts",
  ["wen wu (1959)"] = "Wen Wu",
  ["wereld en zending"] = "Wereld Zending",
  ["werkstatt bionik und evolutionstechnik"] = "Werkstatt Bion. Evol.",
  ["werkstatt geschichte"] = "Werkstatt Gesch",
  ["werkstoffe und korrosion"] = "Werkst. Korros.",
  ["werkstoffe und korrosion. materials and corrosion"] = "Werkst Korros",
  ["wesleyan theological journal"] = "Wesleyan Theol J",
  ["west africa"] = "West Afr",
  ["west african journal of applied ecology"] = "West Afr. J. Appl. Ecol.",
  ["west african journal of medicine"] = "West Afr J Med",
  ["west african journal of pharmacology and drug research"] = "West Afr. J. Pharmacol. Drug Res.",
  ["west african journal of radiology"] = "West Afr J Radiol",
  ["west african journal of ultrasound"] = "West Afr J Ultrasound",
  ["west african medical journal"] = "West Afr. Med. J.",
  ["west african medical journal and nigerian medical and dental practitioner"] = "West Afr. Med. J. Niger. Med. Dent. Pract.",
  ["west african medical journal and nigerian practitioner"] = "West Afr. Med. J. Niger. Pract.",
  ["west african pharmacist"] = "West Afr Pharm",
  ["west european politics"] = "West Eur Polit",
  ["west indian medical journal"] = "West Indian Med. J.",
  ["west london medical journal"] = "West Lond Med J",
  ["west of england medical journal"] = "West Engl Med J",
  ["west virginia dental journal"] = "W V Dent J",
  ["west virginia history"] = "W Va Hist",
  ["west virginia law review"] = "West VA Law Rev",
  ["west virginia medical journal"] = "W. V. Med. J.",
  ["west's annotated california codes. california"] = "Wests Annot Calif Codes Calif",
  ["west's annotated indiana code : under arrangement of the official indiana code. indiana"] = "Wests Annot Indiana Code",
  ["west's atlantic reporter"] = "West's Atl Report",
  ["west's california reporter"] = "Wests Calif Report",
  ["west's federal reporter : cases argued and determined in the united states courts of appeals and temporary emergency court of appeals"] = "Wests Fed Rep",
  ["west's federal rules decisions : opinions, decisions and rulings involving the federal rules of civil procedure and federal rules of criminal procedure"] = "Wests Fed Rules Decis",
  ["west's federal supplement"] = "Wests Fed Suppl",
  ["west's federal supplement (permanent edition)"] = "Wests Fed Suppl (Perm Ed)",
  ["west's florida statutes annotated : under arrangement of the official florida statutes. florida"] = "Wests Fla Statut Annot Fla",
  ["west's new york supplement"] = "Wests N Y Suppl",
  ["west's north eastern reporter"] = "Wests North East Rep",
  ["west's north western reporter"] = "Wests North West Rep",
  ["west's pacific reporter"] = "Wests Pac Report",
  ["west's revised code of washington annotated. washington (state)"] = "Wests Revis Code Wash Annot Wash State",
  ["west's south eastern reporter"] = "Wests South East Report",
  ["west's south western reporter"] = "Wests South West Report",
  ["west's southern reporter"] = "Wests South Report",
  ["west's supreme court reporter"] = "Wests Supreme Court Report",
  ["west's wisconsin statutes annotated. wisconsin"] = "Wests Wis Statut Annot Wis",
  ["westchester medical bulletin"] = "Westchest Med Bull",
  ["western birds"] = "West Birds",
  ["western criminology review"] = "West Crim Rev",
  ["western dental society bulletin"] = "West. Dent. Soc. Bull.",
  ["western economic journal"] = "Western Econ. J.",
  ["western folklore"] = "West Folk",
  ["western illinois regional studies"] = "West Ill Reg Stud",
  ["western indian ocean journal of marine science"] = "West. Indian Ocean J. Mar. Sci.",
  ["western journal of agricultural economics"] = "Western J. Agr. Econ.",
  ["western journal of applied forestry"] = "West. J. Appl. For.",
  ["western journal of communication"] = "West J Commun",
  ["western journal of medicine"] = "West. J. Med.",
  ["western journal of medicine and surgery"] = "West J Med Surg",
  ["western journal of nursing research"] = "West J Nurs Res",
  ["western journal of speech communication : wjsc"] = "West J Speech Commun",
  ["western journal of surgery, obstetrics, and gynecology"] = "West J Surg Obstet Gynecol",
  ["western medicine; the medical journal of the west"] = "West Med Med J West",
  ["western new england law review"] = "West New Engl Law Rev",
  ["western north american naturalist"] = "West N Am Nat",
  ["western ontario law review"] = "West Ont Law Rev",
  ["western ontario series in philosophy of science"] = "West. Ont. Ser. Philos. Sci.",
  ["western pacific surveillance and response journal : wpsar"] = "Western Pac Surveill Response J",
  ["western philosophy series"] = "West. Philos. Ser.",
  ["western state university law review"] = "West State Univ Law Rev",
  ["western states jewish historical quarterly"] = "West States Jew Hist Q",
  ["western states jewish history"] = "West States Jew Hist",
  ["western wildlife (albany, calif.)"] = "West Wildl",
  ["westminster institute review"] = "Westminst Inst Rev",
  ["west’s annotated california codes. california"] = "Wests Annot Calif Codes Calif",
  ["west’s annotated indiana code : under arrangement of the official indiana code. indiana"] = "Wests Annot Indiana Code",
  ["west’s atlantic reporter"] = "West’s Atl Report",
  ["west’s california reporter"] = "Wests Calif Report",
  ["west’s federal reporter : cases argued and determined in the united states courts of appeals and temporary emergency court of appeals"] = "Wests Fed Rep",
  ["west’s federal supplement"] = "Wests Fed Suppl",
  ["west’s florida statutes annotated : under arrangement of the official florida statutes. florida"] = "Wests Fla Statut Annot Fla",
  ["west’s new york supplement"] = "Wests N Y Suppl",
  ["west’s north eastern reporter"] = "Wests North East Rep",
  ["west’s north western reporter"] = "Wests North West Rep",
  ["west’s pacific reporter"] = "Wests Pac Report",
  ["west’s revised code of washington annotated. washington (state)"] = "Wests Revis Code Wash Annot Wash State",
  ["west’s south eastern reporter"] = "Wests South East Report",
  ["west’s south western reporter"] = "Wests South West Report",
  ["west’s southern reporter"] = "Wests South Report",
  ["west’s supreme court reporter"] = "Wests Supreme Court Report",
  ["west’s wisconsin statutes annotated / under arrangement of the official wisconsin statutes. wisconsin"] = "Wests Wis Statut Annot Wis",
  ["weta : news bulletin of the entomological society of new zealand inc"] = "Weta",
  ["wetenschappelijke tijdingen"] = "Wet Tijd",
  ["wetland science"] = "Wetland Sci.",
  ["wetland science and practice"] = "Wetland Sci. Pract.",
  ["wetlands (wilmington, n.c.)"] = "Wetlands (Wilmington)",
  ["wetlands ecology and management"] = "Wetlands Ecol. Manage.",
  ["wetter und leben"] = "Wetter Leben",
  ["wētchasān sattawaphaet = the thai journal of veterinary medicine"] = "Wetchasan Sattawaphaet",
  ["what's new"] = "Whats New",
  ["what's new in industrial hygiene"] = "Whats New Ind Hyg",
  ["what's new in plant physiology"] = "Whats New Plant Physiol",
  ["what’s new in forest research"] = "What’s New For. Res.",
  ["whittier law review"] = "Whittier Law Rev",
  ["who chronicle"] = "WHO Chron",
  ["who features"] = "WHO Feature",
  ["who offset publication"] = "WHO Offset Publ",
  ["who regional publications european series"] = "WHO Reg. Publ. Eur. Ser.",
  ["who regional publications. european series"] = "WHO Reg Publ Eur Ser",
  ["who south-east asia journal of public health"] = "WHO South East Asia J Public Health",
  ["who's who in american art; a biographical directory of contemporary artists, editors, critics, executives, etc"] = "Whos Who Am Art",
  ["whole earth review"] = "Whole Earth Rev",
  ["why the sciences of the ancient world matter"] = "Why Sci. Anc. World Matter",
  ["wiadomosci lekarskie"] = "Wiad. Lek.",
  ["wiadomosci lekarskie (warsaw, poland : 1948)"] = "Wiad Lek Wars Pol 1948",
  ["wiadomosci lekarskie (warsaw, poland : 1960)"] = "Wiad Lek",
  ["wiadomosci matematyczne"] = "Wiad. Mat.",
  ["wiadomosci parazytologiczne"] = "Wiad Parazytol",
  ["wiadomosci statystyczne (warsaw, poland : 1956)"] = "Wiad Stat (Warsaw)",
  ["wiadomości archeologiczne. bulletin archéologique polonais"] = "WiadA",
  ["wiadomości lekarskie (warsaw, poland : 1948)"] = "Wiad Lek Wars Pol 1948",
  ["wiadomości lekarskie (warsaw, poland : 1960)"] = "Wiad Lek",
  ["wiadomości parazytologiczne"] = "Wiad Parazytol",
  ["wiadomości statystyczne (warsaw, poland : 1956)"] = "Wiad Stat (Warsaw)",
  ["widening participation and lifelong learning : the journal of the institute for access studies and the european access network"] = "Widening Partic Lifelong Learn",
  ["wideochirurgia i inne techniki małoinwazyjne = videosurgery and other miniinvasive techniques"] = "Wideochir Inne Tech Maloinwazyjne",
  ["wiederherstellungschirurgie und traumatologie"] = "Wiederherstellungschir. Traumatol.",
  ["wiederherstellungschirurgie und traumatologie. reconstruction surgery and traumatology"] = "Wiederherstellungschir Traumatol",
  ["wiener arbeiten zur philosophie"] = "Wien. Arb. Philos. Reihe A Univ.stud.",
  ["wiener archiv fur psychologie, psychiatrie und neurologie"] = "Wien Arch Psychol Psychiatr Neurologie",
  ["wiener archiv für psychologie, psychiatrie und neurologie"] = "Wien Arch Psychol Psychiatr Neurologie",
  ["wiener beitrage zur chirurgie"] = "Wien Beitr Chir",
  ["wiener beitrage zur dermatologie"] = "Wien Beitr Dermatol",
  ["wiener beitrage zur geburtshilfe und gynakologie"] = "Wien Beitr Geburtshilfe Gynakol",
  ["wiener beitrage zur geschichte der neuzeit"] = "Wien Beitr Gesch Neuzeit",
  ["wiener beitrage zur hals-, nasen- und ohrenheilkunde"] = "Wien Beitr Hals Nasen Ohrenheilkd",
  ["wiener beitrage zur hygiene"] = "Wien Beitr Hyg",
  ["wiener beitrage zur kinderheilkunde"] = "Wien Beitr Kinderheilkd",
  ["wiener beitrage zur zahnheilkunde"] = "Wien Beitr Zahnheilkd",
  ["wiener beiträge zur chirurgie"] = "Wien Beitr Chir",
  ["wiener beiträge zur dermatologie"] = "Wien Beitr Dermatol",
  ["wiener beiträge zur geburtshilfe und gynäkologie"] = "Wien Beitr Geburtshilfe Gynakol",
  ["wiener beiträge zur geschichte der neuzeit"] = "Wien Beitr Gesch Neuzeit",
  ["wiener beiträge zur hals-, nasen- und ohrenheilkunde"] = "Wien Beitr Hals Nasen Ohrenheilkd",
  ["wiener beiträge zur hygiene"] = "Wien Beitr Hyg",
  ["wiener beiträge zur inneren medizin"] = "Wien Beitr Inneren Med",
  ["wiener beiträge zur kinderheilkunde"] = "Wien Beitr Kinderheilkd",
  ["wiener beiträge zur zahnheilkunde"] = "Wien Beitr Zahnheilkd",
  ["wiener beiträge zur geschichte der medizin (universitaẗ wien. institut für geschichte der medizin)"] = "Wien Beitr Gesch Med",
  ["wiener geschichtsblatter / herausgegeben vom verein fur geschichte der stadt wien"] = "Wien Geschichtsbl",
  ["wiener geschichtsblätter"] = "Wien Geschichtsbl",
  ["wiener humanistische blätter"] = "WHB",
  ["wiener klinische wochenschrift"] = "Wien Klin Wochenschr",
  ["wiener klinische wochenschrift. education"] = "Wien Klin Wochenschr Educ",
  ["wiener klinische wochenschrift. supplementum"] = "Wien Klin Wochenschr Suppl",
  ["wiener klinisches magazin : beilage zur wiener klinischen wochenschrift"] = "Wien Klin Mag",
  ["wiener medizinische wochenschrift"] = "Wien. Med. Wochenschr.",
  ["wiener medizinische wochenschrift (1946)"] = "Wien Med Wochenschr",
  ["wiener medizinische wochenschrift. beihefte"] = "Wien Med Wochenschr Beih",
  ["wiener medizinische wochenschrift. supplement"] = "Wien Med Wochenschr Suppl",
  ["wiener medizinische wochenschrift.supplement"] = "Wien Med Wochenschr Suppl",
  ["wiener reihe"] = "Wien. Reihe",
  ["wiener studien"] = "WSt",
  ["wiener studien zur medizin, geschichte und philosophie"] = "Wien Stud Med Gesch Philos",
  ["wiener studien: zeitschrift für klassische philologie und patristik"] = "WS",
  ["wiener tierarztliche monatsschrift"] = "Wien. Tierarztl. Monatsschr.",
  ["wiener tierärztliche monatsschrift"] = "Wien Tierarztl Monatsschr",
  ["wiener zeitschrift fur die kunde des morgenlandes"] = "Wien Z Kunde Morgenl",
  ["wiener zeitschrift fur innere medizin und ihre grenzgebiete"] = "Wien. Z. Inn. Med.",
  ["wiener zeitschrift fur nervenheilkunde und deren grenzgebiete"] = "Wien. Z. Nervenheilkd. Grenzgeb.",
  ["wiener zeitschrift für die kunde des morgenlandes"] = "Wien Z Kunde Morgenl",
  ["wiener zeitschrift für innere medizin und ihre grenzgebiete"] = "Wien Z Inn Med",
  ["wiener zeitschrift für nervenheilkunde und deren grenzgebiete"] = "Wien Z Nervenheilkd Grenzgeb",
  ["wiener zeitschrift für praktische psychologie"] = "Wien Z Prakt Psychol",
  ["wiener zeitschrift für die kunde des morgenlandes"] = "WZKM",
  ["wiener zeitschrift zur geschichte der neuzeit"] = "Wien Z Gesch Neuzeit",
  ["wies wspolczesna; pismo ruchu ludowego"] = "Wies Wspolcz",
  ["wieś współczesna; pismo ruchu ludowego"] = "Wies Wspolcz",
  ["wijsgerig perspectief op maatschappij en wetenschap"] = "Wijsgerig Perspect Maatsch Wet",
  ["wildbach- und lawinenverbau"] = "Wildbach- Lawinenverbau",
  ["wilderness & environmental medicine"] = "Wilderness Environ Med",
  ["wilderness and environmental medicine"] = "Wilderness Environ. Med.",
  ["wildfowl. wildfowl trust"] = "Wildfowl (Slimbr)",
  ["wildlife biology"] = "Wildlife Biol",
  ["wildlife biology in practice (online)"] = "Wildl Biol Pract",
  ["wildlife disease"] = "Wildl Dis",
  ["wildlife letters"] = "Wildl. Lett.",
  ["wildlife monographs"] = "Wildl. Monogr.",
  ["wildlife research"] = "Wildl. Res.",
  ["wildlife research (east melbourne, melbourne, vic.)"] = "Wildl Res",
  ["wildlife society bulletin"] = "Wildl Soc Bull",
  ["wiley classics library"] = "Wiley Classics Lib.",
  ["wiley finance editions"] = "Wiley Finance Ed.",
  ["wiley finance series"] = "Wiley Finance Ser.",
  ["wiley handbook in financial engineering and econometrics"] = "Wiley Handb. Finance Eng. Econom.",
  ["wiley interdisciplinary reviews"] = "WIRES",
  ["wiley interdisciplinary reviews. climate change"] = "Wiley Interdiscip Rev Clim Change",
  ["wiley interdisciplinary reviews. cognitive science"] = "Wiley Interdiscip Rev Cogn Sci",
  ["wiley interdisciplinary reviews. computational molecular science"] = "Wiley Interdiscip Rev Comput Mol Sci",
  ["wiley interdisciplinary reviews. computational statistics"] = "Wiley Interdiscip Rev Comput Stat",
  ["wiley interdisciplinary reviews. computational statistics (wires)"] = "Wiley Interdiscip. Rev. Comput. Stat.",
  ["wiley interdisciplinary reviews. data mining and knowledge discovery"] = "Wiley Interdiscip Rev Data Min Knowl Discov",
  ["wiley interdisciplinary reviews. developmental biology"] = "Wiley Interdiscip Rev Dev Biol",
  ["wiley interdisciplinary reviews. energy and environment"] = "Wiley Interdiscip Rev Energy Environ",
  ["wiley interdisciplinary reviews. membrane transport and signaling"] = "Wiley Interdiscip Rev Membr Transp Signal",
  ["wiley interdisciplinary reviews. nanomedicine and nanobiotechnology"] = "Wiley Interdiscip Rev Nanomed Nanobiotechnol",
  ["wiley interdisciplinary reviews. rna"] = "Wiley Interdiscip Rev RNA",
  ["wiley interdisciplinary reviews. systems biology and medicine"] = "Wiley Interdiscip Rev Syst Biol Med",
  ["wiley interdisciplinary reviews: climate change"] = "Wiley Interdiscip. Rev. Clim. Change",
  ["wiley interdisciplinary reviews: cognitive science"] = "Wiley Interdiscip. Rev. Cognit. Sci.",
  ["wiley interdisciplinary reviews: computational molecular science"] = "Wiley Interdiscip. Rev.: Comput. Mol. Sci.",
  ["wiley interdisciplinary reviews: computational statistics"] = "Wiley Interdiscip. Rev. Comput. Stat.",
  ["wiley interdisciplinary reviews: data mining and knowledge discovery"] = "Wiley Interdiscip. Rev.: Data Min. Knowl. Discovery",
  ["wiley interdisciplinary reviews: developmental biology"] = "Wiley Interdiscip. Rev.: Dev. Biol.",
  ["wiley interdisciplinary reviews: energy and environment"] = "Wiley Interdiscip. Rev.: Energy Environ.",
  ["wiley interdisciplinary reviews: forensic science"] = "Wiley Interdiscip. Rev.: Forensic Sci.",
  ["wiley interdisciplinary reviews: membrane transport and signaling"] = "Wiley Interdiscip. Rev.: Membr. Transp. Signaling",
  ["wiley interdisciplinary reviews: nanomedicine and nanobiotechnology"] = "Wiley Interdiscip. Rev. Nanomed. Nanobiotechnol.",
  ["wiley interdisciplinary reviews: rna"] = "Wiley Interdiscip. Rev.: RNA",
  ["wiley interdisciplinary reviews: systems biology and medicine"] = "Wiley Interdiscip. Rev. Syst. Biol. Med.",
  ["wiley interdisciplinary reviews: water"] = "Wiley Interdiscip. Rev.: Water",
  ["wiley professional paperback series"] = "Wiley Prof. Paperb. Ser.",
  ["wiley series in beam physics and accelerator technology"] = "Wiley Ser. Beam Phys. Accel. Tech.",
  ["wiley series in computational and quantitative social science"] = "Wiley Ser. Comput. Quant. Soc. Sci.",
  ["wiley series in computational statistics"] = "Wiley Ser. Comput. Stat.",
  ["wiley series in discrete mathematics and optimization"] = "Wiley Ser. Discrete Math. Optim.",
  ["wiley series in mathematical methods in practice"] = "Wiley Ser. Math. Methods Pract.",
  ["wiley series in methods and applications of statistics"] = "Wiley Ser. Methods Appl. Statist.",
  ["wiley series in modeling and simulation"] = "Wiley Ser. Model. Simul.",
  ["wiley series in nonlinear science"] = "Wiley Ser. Nonlinear Sci.",
  ["wiley series in operations research and management science"] = "Wiley Ser. Oper. Res. Manag. Sci.",
  ["wiley series in probability and mathematical statistics"] = "Wiley Ser. Probab. Math. Statist.",
  ["wiley series in probability and statistics"] = "Wiley Ser. Probab. Stat.",
  ["wiley series in probability and statistics: applied probability and statistics"] = "Wiley Ser. Probab. Statist. Appl. Probab. Statist.",
  ["wiley series in probability and statistics: probability and statistics"] = "Wiley Ser. Probab. Statist. Probab. Statist.",
  ["wiley series in probability and statistics: texts and references section"] = "Wiley Ser. Probab. Stat. Texts Ref. Sect.",
  ["wiley series in survey methodology"] = "Wiley Ser. Surv. Methodol.",
  ["wiley series in systems engineering"] = "Wiley Ser. Syst. Eng.",
  ["wiley series on adaptive and cognitive dynamic systems"] = "Wiley Ser. Adapt. Cogn. Dyn. Syst.",
  ["wiley series on bioinformatics: computational techniques and engineering"] = "Wiley Ser. Bioinform. Comput. Tech. Eng.",
  ["wiley series on parallel and distributed computing"] = "Wiley Ser. Parallel Distrib. Comput.",
  ["wiley-asme press series"] = "Wiley-ASME Press Ser.",
  ["wiley-interscience series in discrete mathematics and optimization"] = "Wiley-Intersci. Ser. Discrete Math. Optim.",
  ["wiley-interscience series in systems and optimization"] = "Wiley-Intersci. Ser. Systems Optim.",
  ["wiley-teubner computing"] = "Wiley-Teubner Comput.",
  ["wiley-teubner series advances in numerical mathematics"] = "Wiley-Teubner Ser. Adv. Numer. Math.",
  ["wilhelm roux' archiv für entwicklungsmechanik der organismen"] = "Wilhelm Roux Arch Entwickl Mech Org",
  ["wilhelm roux's archives of developmental biology"] = "Wilehm Roux Arch Dev Biol",
  ["wilhelm roux’ archiv fuer entwicklungsmechanik der organismen"] = "Wilhelm Roux’ Arch. Entwicklungsmech. Org.",
  ["willamette law journal"] = "Willamette Law J",
  ["willamette law review"] = "Willamette Law Rev",
  ["william & mary journal of women and the law"] = "William Mary J Women Law",
  ["william & mary policy review"] = "William Mary Policy Rev",
  ["william and mary law review"] = "William Mary Law Rev",
  ["william mitchell law review"] = "William Mitchell Law Rev",
  ["wilson and wilson's comprehensive analytical chemistry"] = "Wilson Wilsons Compr Anal Chem",
  ["wilson journal of ornithology"] = "Wilson J. Ornith.",
  ["wilson library bulletin"] = "Wilson Libr Bull",
  ["winckelmannsprogramm der archäologischen gesellschaft zu berlin"] = "BWPr",
  ["wind and structures"] = "Wind Struct.",
  ["wind energy"] = "Wind Energy",
  ["wind energy (chichester, england)"] = "Wind Energy (Chichester)",
  ["wind energy science"] = "Wind Energy Sci.",
  ["wind engineering"] = "Wind Eng.",
  ["windows in time"] = "Windows Time",
  ["wine economics and policy"] = "Wine Econ. Policy",
  ["winnipeg clinic quarterly"] = "Winnipeg Clin Q",
  ["winter braids lecture notes"] = "Winter Braids Lect. Notes",
  ["winterberichte des eidg. institutes für schnee- und lawinenforschung"] = "Winterber. Eidgenöss. Inst. Schnee- Lawinenforsch.",
  ["winterthur portfolio"] = "Winterthur Portf",
  ["wireless communications & mobile computing"] = "Wirel Commun Mob Comput",
  ["wireless communications and mobile computing"] = "Wireless Commun. Mobile Comput.",
  ["wireless networks"] = "Wireless Netw.",
  ["wireless personal communications"] = "Wirel Pers Commun",
  ["wireless power transfer"] = "Wireless Power Transfer",
  ["wires climate change"] = "WIREs Clim. Change",
  ["wires cognitive science"] = "WIREs Cognit. Sci.",
  ["wires computational molecular science"] = "WIREs Comput. Mol. Sci.",
  ["wires computational statistics"] = "WIREs Comput. Stat.",
  ["wires data mining and knowledge discovery"] = "WIREs Data Min. Knowl. Discovery",
  ["wires developmental biology"] = "WIREs Dev. Biol.",
  ["wires energy and environment"] = "WIREs Energy Environ.",
  ["wires forensic science"] = "WIREs Forensic Sci.",
  ["wires mechanisms of disease"] = "WIREs Mech. Dis.",
  ["wires membrane transport and signaling"] = "WIREs Membr. Transp. Signaling",
  ["wires nanomedicine and nanobiotechnology"] = "WIREs Nanomed. Nanobiotechnol.",
  ["wires rna"] = "WIREs RNA",
  ["wires systems biology and medicine"] = "WIREs Syst. Biol. Med.",
  ["wires water"] = "WIREs Water",
  ["wires. water"] = "WIREs Water",
  ["wiring installations and supplies"] = "Wiring Install. Supplies",
  ["wirtschaft und recht"] = "Wirtsch. Recht",
  ["wirtschaft und statistik"] = "Wirtsch Stat",
  ["wirtschaftsdienst (hamburg, germany : 1949)"] = "Wirtschaftsdienst",
  ["wirtschaftspolitische blätter"] = "Wirtschaftspolitische Blätter",
  ["wirtschaftswissenschaftliches studium"] = "Wirtschaftswissenschaftliches Stud",
  ["wirtschaftswissenschaftliches studium—taschenbucher"] = "WiSt—Taschenbucher",
  ["wisconsin academy review"] = "Wis Acad Rev",
  ["wisconsin dental association journal"] = "Wis Dent Assoc J",
  ["wisconsin dental association journal (milwaukee, wis. : 1992)"] = "Wis Dent Assoc J",
  ["wisconsin dental association journal / wda"] = "Wis Dent Assoc J",
  ["wisconsin law review"] = "Wis L Rev",
  ["wisconsin magazine of history"] = "Wis Mag Hist",
  ["wisconsin medical journal"] = "Wis Med J",
  ["wisconsin publications in the history of science and medicine"] = "Wis Publ Hist Sci Med",
  ["wisconsin session laws. wisconsin"] = "Wis Sess Laws Wis",
  ["wisconsin welfare"] = "Wis Welf",
  ["wisconsin women's law journal"] = "Wis Womens Law J",
  ["wisconsin women’s law journal"] = "Wis Womens Law J",
  ["wiss. meeresunters."] = "Wissenschaftliche Meeresuntersuchungen",
  ["wissen kompakt : fortbildung für zahnärzte"] = "Wissen Kompakt (Heidelb)",
  ["wissenschaft im 20"] = "Wiss. 20 Jhd. Transdiszip. Reflex.",
  ["wissenschaft und glaube"] = "Wiss Glaube",
  ["wissenschaftliche alpenvereinshefte"] = "Wiss. Alp.ver.heft",
  ["wissenschaftliche beitrage der martin-luther-universitat halle-wittenberg"] = "Wiss Beitr Martin Luther Univ Halle Wittenberg",
  ["wissenschaftliche beitrage zur geschichte der seelenheilkunde"] = "Wiss Beitr Gesch Seelenheilkd",
  ["wissenschaftliche beiträge der martin-luther-universität halle-wittenberg"] = "Wiss Beitr Martin Luther Univ Halle Wittenberg",
  ["wissenschaftliche beiträge zur geschichte der seelenheilkunde"] = "Wiss Beitr Gesch Seelenheilkd",
  ["wissenschaftliche bucherei"] = "Wiss. Bucherei",
  ["wissenschaftliche mitteilungen des bosnischen landesmuseums, a. archäologie"] = "WissMBosn",
  ["wissenschaftliche mitteilungen. österreichischer apothekerverein, vienna. pharmazeutisches forschungsinstitut"] = "Wiss Mitt",
  ["wissenschaftliche schriftenreihe mathematik"] = "Wiss. Schriftenr. Math.",
  ["wissenschaftliche veroffentlichungen. deutsche gesellschaft fur ernahrung"] = "Wiss Veroff Dtsch Ges Ernahr",
  ["wissenschaftliche veröffentlichungen. deutsche gesellschaft für ernährung"] = "Wiss Veroff Dtsch Ges Ernahr",
  ["wissenschaftliche veröffentlichungen der deutschen orient-gesellschaft"] = "WVDOG",
  ["wissenschaftliche zeitschrift der ernst-moritz-arndt-univ., greifswald, gesellsch.- & sprachwiss. reihe."] = "WZGREIFS",
  ["wissenschaftliche zeitschrift der ernst-moritz-arndt-universitat, greifswald. gesellschaftswissenschaftliche reihe"] = "Wiss Z Ernst Moritz Arndt Univ [Ges Sprachwiss]",
  ["wissenschaftliche zeitschrift der ernst-moritz-arndt-universität, greifswald. gesellschaftswissenschaftliche reihe"] = "Wiss Z Ernst Moritz Arndt Univ Ges Sprachwiss",
  ["wissenschaftliche zeitschrift der friedrich-schiller-univ. jena, gesellsch.- & sprachwiss. reihe."] = "WZJENA",
  ["wissenschaftliche zeitschrift der friedrich-schiller-universitat jena. gesellschafts- und sprachwissenschaftliche reihe"] = "Wiss Z Friedrich Schiller Univ Jena Ges Sprachwiss",
  ["wissenschaftliche zeitschrift der friedrich-schiller-universität jena. gesellschafts- und sprachwissenschaftliche reihe"] = "Wiss Z Friedrich Schiller Univ Jena Ges Sprachwiss",
  ["wissenschaftliche zeitschrift der friedrich-schiller-universität jena"] = "WissZJena",
  ["wissenschaftliche zeitschrift der humboldt-univ. berlin, gesellsch.- & sprachwiss. reihe"] = "WZBERLIN",
  ["wissenschaftliche zeitschrift der humboldt-universitat zu berlin. gesellschafts- und sprachwissenschaftliche reihe"] = "Wiss Z Humboldt Univ Berl (Ges Sprachwiss)",
  ["wissenschaftliche zeitschrift der humboldt-universitat zu berlin. mathematisch-naturwissenschaftliche reihe"] = "Wiss. Z. Humboldt Univ. Berl. [Math. Naturwiss.]",
  ["wissenschaftliche zeitschrift der humboldt-universität zu berlin. gesellschafts- und sprachwissenschaftliche reihe"] = "Wiss Z Humboldt Univ Berl (Ges Sprachwiss)",
  ["wissenschaftliche zeitschrift der humboldt-universität zu berlin. gesellschaftswissenschaftliche reihe"] = "Wiss Z Humboldt Univ Berl Ges Sprachwiss Reihe",
  ["wissenschaftliche zeitschrift der humboldt-universität zu berlin. mathematisch-naturwissenschaftliche reihe"] = "Wiss Z Humboldt Univ Berl Math Naturwiss",
  ["wissenschaftliche zeitschrift der humboldt-universität zu berlin. gesellschafts- und sprachwissenschaftliche reihe"] = "WissZBerl",
  ["wissenschaftliche zeitschrift der karl-marx-univ. leipzig, gesellsch.- & sprachwlss. reihe."] = "WZLEIPZIG",
  ["wissenschaftliche zeitschrift der karl-marx-universitat leipzig. mathematisch-naturwissenschaftliche reihe"] = "Wiss Z Karl Marx Univ Math Naturwiss",
  ["wissenschaftliche zeitschrift der karl-marx-universität leipzig. mathematisch-naturwissenschaftliche reihe"] = "Wiss Z Karl Marx Univ Math Naturwiss",
  ["wissenschaftliche zeitschrift der martin-luther-univ. halle-wittenberg"] = "WZHALLE",
  ["wissenschaftliche zeitschrift der technischen universität dresden"] = "Wiss. Z. Tech. Univ. Dresden",
  ["wissenschaftliche zeitschrift der technischen universität dresden, separatreihe 5"] = "Wiss. Z. Tech. Univ. Dresd., Sep.r. 5",
  ["wissenschaftliche zeitschrift der universität rostock"] = "WissZRostock",
  ["wissenschaftliche zeitschrift der wilhelm-pieck-univ. rostock gesellsch.- & sprachwiss. reihe."] = "WZROSTOCK",
  ["wissenschaftliche zeitschrift der wilhelm-pieck-universitat rostock. gesellschafts- und sprachwissenschaftliche reihe"] = "Wiss Z Wilhelm Pieck Univ Rostock [Ges Sprachwiss]",
  ["wissenschaftliche zeitschrift der wilhelm-pieck-universität rostock. gesellschafts- und sprachwissenschaftliche reihe"] = "Wiss Z Wilhelm Pieck Univ Rostock Ges Sprachwiss",
  ["wissenschaftliche zeitschrift. gesellschafts- und sprachwissenschaftliche reihe (karl-marx-universitat leipzig)"] = "Wiss Z Karl Marx Univ Leipzig [Ges Sprachwiss]",
  ["wissenschaftliche zeitschrift. gesellschafts- und sprachwissenschaftliche reihe (karl-marx-universität leipzig)"] = "Wiss Z Karl Marx Univ Leipzig Ges Sprachwiss",
  ["wissenschaftliche zeitschrift. martin-luther-universität halle-wittenberg"] = "WissZHalle",
  ["wissenschaftliche zeitschrift. mathematisch-naturwissenschaftliche reihe. ernst-moritz-arndt-universitat greifswald"] = "Wiss Z Ernst Moritz Arndt Univ [Math]",
  ["wissenschaftliche zeitschrift. mathematisch-naturwissenschaftliche reihe. ernst-moritz-arndt-universität greifswald"] = "Wiss Z Ernst Moritz Arndt Univ Math",
  ["wissenschaftlicher dienst sudosteuropa"] = "Wiss Dienst Sudosteur",
  ["wissenschaftlicher dienst südosteuropa"] = "Wiss Dienst Sudosteur",
  ["wissenschaftliches jahrbuch der tiroler landesmuseen"] = "Wiss Jahrb Tirol Landesmuseen",
  ["wissenshaftliche zeitschrift der martin luther universitat halle-wittenberg"] = "Wiss Z. Martin-Luther Univ. Halle-Wittenberg",
  ["wistar institute symposium monograph"] = "Wistar Inst. Symp. Monogr.",
  ["wit transactions on ecology and the environment"] = "WIT Trans. Ecol. Environ.",
  ["wit transactions on engineering sciences"] = "WIT Trans. Eng. Sci.",
  ["wit transactions on modelling and simulation"] = "WIT Trans Modelling Simul",
  ["wit transactions on the built environment"] = "WIT Trans. Built Environ.",
  ["wits journal of clinical medicine"] = "Wits J Clin Med",
  ["witthayasan kasetsat"] = "The Kasetsart journal|Witthayasan Kasetsat",
  ["witthayasat mo kho"] = "Witthayasat Mo Kho",
  ["witthayāsān kasētsāt = the kasetsart journal"] = "Witthayasan Kasetsat",
  ["witthayāsān kasētsāt. sākhā sangkhommasāt = the kasetsart journal. social sciences"] = "Kasetsart J Soc Sci",
  ["wmf press bulletins"] = "WMF Press Bull",
  ["wmj : official publication of the state medical society of wisconsin"] = "WMJ",
  ["wochenbericht. deutsches institut für wirtschaftsforschung"] = "Wochenber Dtsch Inst Wirtschaftsforsch",
  ["wochenblatt fur papierfabrikation"] = "Wochenbl. Papierfabr.",
  ["wochenblatt für papierfabrikation"] = "Wochenbl. Papierfabr.",
  ["wochenschrift fur brauerei"] = "Wochenschr. Brau.",
  ["wolfenbutteler renaissance mitteilungen"] = "Wolfenbutteler Renaiss Mitt",
  ["wolfenbütteler renaissance mitteilungen"] = "Wolfenbutteler Renaiss Mitt",
  ["woman physician"] = "Woman Physician",
  ["women & criminal justice"] = "Women Crim Justice",
  ["women & health"] = "Women Health",
  ["women & politics"] = "Women Polit",
  ["women & therapy"] = "Women Ther",
  ["women 2000"] = "Women 2000",
  ["women alive (los angeles, calif.)"] = "Women Alive",
  ["women and aids"] = "Women AIDS",
  ["women and birth : journal of the australian college of midwives"] = "Women Birth",
  ["women and environments"] = "Women Environ",
  ["women and health"] = "Women Health",
  ["women envision"] = "Women Envis",
  ["women health care and issues"] = "Women Health Care Issues",
  ["women in action (rome, italy)"] = "Women Action",
  ["women in antiquity"] = "Women Antiq.",
  ["women in medicine; official quarterly of american medical women's association, inc"] = "Women Med",
  ["women in sport & physical activity journal"] = "Women Sport Phys Act J",
  ["women in the history of philosophy and sciences"] = "Women Hist. Philos. Sci.",
  ["women lawyers' journal"] = "Women Lawyers J",
  ["women lawyers’ journal"] = "Women Lawyers J",
  ["women of vietnam"] = "Women Vietnam",
  ["women's health & urban life : an international and interdisciplinary journal"] = "Womens Health Urban Life",
  ["women's health (hillsdale, n.j.)"] = "Womens Health",
  ["women's health (london, england)"] = "Womens Health (Lond)",
  ["women's health bulletin"] = "Womens Health Bull",
  ["women's health care : a practical journal for nurse practitioners"] = "Womens Health Care",
  ["women's health issues : official publication of the jacobs institute of women's health"] = "Womens Health Issues",
  ["women's health journal"] = "Womens Health J",
  ["women's health reports (new rochelle, n.y.)"] = "Womens Health Rep (New Rochelle)",
  ["women's health update"] = "Womens Health Update",
  ["women's health weekly"] = "Womens Health Wkly",
  ["women's healthcare (doylestown, pa.)"] = "Womens Healthc (Doylestown)",
  ["women's history"] = "Womens Hist",
  ["women's history review"] = "Womens Hist Rev",
  ["women's midlife health"] = "Womens Midlife Health",
  ["women's reproductive health (philadelphia, pa.)"] = "Womens Reprod Health (Phila)",
  ["women's rights law reporter"] = "Women's Rights Law Report",
  ["women's studies"] = "Womens Stud",
  ["women's studies forum"] = "Womens Stud Forum",
  ["women's studies international forum"] = "Womens Stud Int Forum",
  ["women's studies international quarterly"] = "Womens Stud Int Q",
  ["women's world (kampala, uganda)"] = "Womens World",
  ["women's writing : the elizabethan to victorian period"] = "Womens Writ",
  ["womens health"] = "Womens Health",
  ["womens health issues"] = "Womens Health Issues",
  ["womens health newsletter"] = "Womens Health Newsl",
  ["women’s health"] = "Women’s Health",
  ["women’s health (hillsdale, n.j.)"] = "Womens Health",
  ["women’s health data book"] = "Womens Health Data Book",
  ["women’s health issues"] = "Womens Health Issues",
  ["women’s health issues : official publication of the jacobs institute of women’s health"] = "Womens Health Issues",
  ["women’s health journal / isis international, latin american and caribbean women’s health network"] = "Womens Health J",
  ["women’s health update"] = "Womens Health Update",
  ["women’s health weekly"] = "Womens Health Wkly",
  ["women’s rights law reporter"] = "Women’s Rights Law Report",
  ["women’s studies"] = "Womens Stud",
  ["women’s studies forum"] = "Womens Stud Forum",
  ["women’s studies international forum"] = "Womens Stud Int Forum",
  ["women’s studies international quarterly"] = "Womens Stud Int Q",
  ["women’s world (kampala, uganda)"] = "Womens World",
  ["wood and fiber science"] = "Wood Fiber Sci.",
  ["wood design focus"] = "Wood Des Focus",
  ["wood material science & engineering"] = "Wood Mater. Sci. Eng.",
  ["wood research"] = "Wood Res.",
  ["wood science and technology"] = "Wood Sci. Technol.",
  ["woodhead publishing in mechanical engineering"] = "Woodhead Publ. Mech. Eng.",
  ["woodhead publishing series in biomaterials"] = "Woodhead Publ Ser Biomater",
  ["woodhead publishing series in biomedicine"] = "Woodhead Publ. Ser. Biomed.",
  ["woodhead publishing series in composites science and engineering"] = "Woodhead Publ. Ser. Compos. Sci. Eng.",
  ["worcester medical news"] = "Worcester Med News",
  ["word & image"] = "Word Image (Lond)",
  ["word & world"] = "Word World",
  ["word (new york, n.y. : 1945)"] = "Word (N Y : 1945)",
  ["work (reading, mass.)"] = "Work",
  ["work and health"] = "Work Health",
  ["work and occupations"] = "Work Occup",
  ["work and stress"] = "Work Stress",
  ["work employment and society"] = "Work Employ. Soc.",
  ["work, aging and retirement"] = "Work Aging Retire",
  ["work, employment & society : a journal of the british sociological association"] = "Work Employ Soc",
  ["work, environment, health"] = "Work Environ Health",
  ["work-life policies. national symposium on family issues (15th : 2007 : pennsylvania state university)"] = "Work Life Policies",
  ["working paper (center for community development investments)"] = "Work Paper (Cent Community Dev Invest)",
  ["working paper (université catholique de louvain (1970- ). département de démographie)"] = "Work Pap",
  ["working paper series (national bureau of economic research)"] = "NBER Work Pap Ser",
  ["working papers for a new society"] = "Work Pap New Soc",
  ["working papers magazine"] = "Work Pap Mag",
  ["working with older people (brighton, england)"] = "Work Older People",
  ["working woman"] = "Work Woman",
  ["workplace health & safety"] = "Workplace Health Saf",
  ["workshop on countering space adaptation with exercise : current issues"] = "Workshop Counter Space Adapt Exerc",
  ["workshop on speech and language processing for assistive technologies"] = "Workshop Speech Lang Process Assist Technol",
  ["workshops & poster abstracts : [proceedings]. ieee computational systems bioinformatics conference"] = "Workshops Poster Abstr",
  ["workshops in computing"] = "Workshops Comput.",
  ["world (oakland, calif. : 1993)"] = "WORLD",
  ["world academy of science, engineering and technology"] = "World Acad Sci Eng Technol",
  ["world academy of sciences journal"] = "World Acad Sci J",
  ["world aids day features"] = "World AIDS Day Features",
  ["world aids day newsletter"] = "World AIDS Day Newsl",
  ["world allergy organization journal"] = "World Allergy Organ. J.",
  ["world applied sciences journal"] = "World Appl Sci J",
  ["world archaeology"] = "World Archaeol",
  ["world bank economic review"] = "World Bank Econ. Rev.",
  ["world bank research news"] = "World Bank Res News",
  ["world bank research observer"] = "World Bank Res. Observer",
  ["world citizen news : newsletter of the world government of world citizens"] = "World Citz News",
  ["world conservation"] = "World Conserv",
  ["world cultural psychiatry research review : official journal of world association of cultural psychiatry"] = "World Cult Psychiatry Res Rev",
  ["world development"] = "World Dev",
  ["world development forum"] = "World Dev Forum",
  ["world development perspectives"] = "World Dev Perspect",
  ["world economy"] = "World Economy",
  ["world education reports"] = "World Educ Rep",
  ["world electric vehicle journal"] = "World Electr. Veh. J.",
  ["world environment report"] = "World Environ Rep",
  ["world futures"] = "World Futures",
  ["world futures review"] = "World Futures Rev",
  ["world haptics 2007 : second joint eurohaptics conference and symposium on haptic interfaces for virtual environment and teleoperator systems : 22-24 march, 2007, tsukuba, japan. world haptics conference (2nd : 2007 : tsukuba-shi, japan)"] = "World Haptics 2007 (2007)",
  ["world health"] = "World Health",
  ["world health & population"] = "World Health Popul",
  ["world health design"] = "World Health Des",
  ["world health forum"] = "World Health Forum",
  ["world health organization aids technical bulletin"] = "WHO AIDS Tech Bull",
  ["world health organization monograph series"] = "World Health Organ. Monogr. Ser.",
  ["world health organization technical report series"] = "World Health Organ Tech Rep Ser",
  ["world health statistics quarterly"] = "World Health Stat. Q.",
  ["world health statistics quarterly. rapport trimestriel de statistiques sanitaires mondiales"] = "World Health Stat Q",
  ["world health statistics report"] = "World Health Stat. Rep.",
  ["world health statistics report. rapport de statistiques sanitaires mondiales"] = "World Health Stat Rep",
  ["world heart journal"] = "World Heart J",
  ["world hospitals"] = "World Hosp",
  ["world hospitals and health services"] = "World Hosp. Health Serv.",
  ["world hospitals and health services : the official journal of the international hospital federation"] = "World Hosp Health Serv",
  ["world journal for pediatric & congenital heart surgery"] = "World J Pediatr Congenit Heart Surg",
  ["world journal of acupuncture-moxibustion"] = "World J Acupunct Moxibustion",
  ["world journal of agricultural research"] = "World J Agric Res",
  ["world journal of aids"] = "World J AIDS",
  ["world journal of biological chemistry"] = "World J Biol Chem",
  ["world journal of biological psychiatry"] = "World J. Biol. Psychiatry",
  ["world journal of cardiology"] = "World J Cardiol",
  ["world journal of cardiovascular diseases"] = "World J Cardiovasc Dis",
  ["world journal of cardiovascular surgery"] = "World J Cardiovasc Surg",
  ["world journal of clinical cases"] = "World J Clin Cases",
  ["world journal of clinical infectious diseases"] = "World J Clin Infect Dis",
  ["world journal of clinical oncology"] = "World J Clin Oncol",
  ["world journal of clinical pediatrics"] = "World J Clin Pediatr",
  ["world journal of clinical urology"] = "World J Clin Urol",
  ["world journal of colorectal surgery"] = "World J Colorectal Surg",
  ["world journal of condensed matter physics"] = "World J. Condens. Matter Phys.",
  ["world journal of critical care medicine"] = "World J Crit Care Med",
  ["world journal of diabetes"] = "World J Diabetes",
  ["world journal of educational research (los angeles, calif.)"] = "World J Educ Res",
  ["world journal of emergency medicine"] = "World J Emerg Med",
  ["world journal of emergency surgery : wjes"] = "World J Emerg Surg",
  ["world journal of engineering"] = "World J. Eng.",
  ["world journal of experimental medicine"] = "World J Exp Med",
  ["world journal of gastroenterology"] = "World J Gastroenterol",
  ["world journal of gastroenterology : wjg"] = "World J Gastroenterol",
  ["world journal of gastrointestinal endoscopy"] = "World J Gastrointest Endosc",
  ["world journal of gastrointestinal oncology"] = "World J Gastrointest Oncol",
  ["world journal of gastrointestinal pathophysiology"] = "World J Gastrointest Pathophysiol",
  ["world journal of gastrointestinal pharmacology and therapeutics"] = "World J Gastrointest Pharmacol Ther",
  ["world journal of gastrointestinal surgery"] = "World J Gastrointest Surg",
  ["world journal of gynecology & womens health"] = "World J Gynecol Womens Health",
  ["world journal of hepatology"] = "World J Hepatol",
  ["world journal of hypertension"] = "World J Hypertens",
  ["world journal of immunology"] = "World J Immunol",
  ["world journal of medical genetics"] = "World J Med Genet",
  ["world journal of meta-analysis"] = "World J Metaanal",
  ["world journal of methodology"] = "World J Methodol",
  ["world journal of microbiology & biotechnology"] = "World J Microbiol Biotechnol",
  ["world journal of microbiology and biotechnology"] = "World J. Microbiol. Biotechnol.",
  ["world journal of nephrology"] = "World J Nephrol",
  ["world journal of nephrology and urology"] = "World J Nephrol Urol",
  ["world journal of neurology"] = "World J Neurol",
  ["world journal of neuroscience"] = "World J Neurosci",
  ["world journal of nuclear medicine"] = "World J Nucl Med",
  ["world journal of obstetrics and gynecology"] = "World J Obstet Gynecol",
  ["world journal of oncology"] = "World J Oncol",
  ["world journal of ophthalmology"] = "World J Ophthalmol",
  ["world journal of orthodontics"] = "World J Orthod",
  ["world journal of orthopedics"] = "World J Orthop",
  ["world journal of otorhinolaryngology"] = "World J Otorhinolaryngol",
  ["world journal of otorhinolaryngology - head and neck surgery"] = "World J Otorhinolaryngol Head Neck Surg",
  ["world journal of pediatrics : wjp"] = "World J Pediatr",
  ["world journal of pharmaceutical research"] = "World J Pharm Res",
  ["world journal of pharmacology"] = "World J Pharmacol",
  ["world journal of pharmacy and pharmaceutical sciences"] = "World J Pharm Pharm Sci",
  ["world journal of plastic surgery"] = "World J Plast Surg",
  ["world journal of psychiatry"] = "World J Psychiatry",
  ["world journal of psychiatry and mental health research"] = "World J Psychiatry Ment Health Res",
  ["world journal of radiology"] = "World J Radiol",
  ["world journal of research and review"] = "World J Res Rev",
  ["world journal of respirology"] = "World J Respirol",
  ["world journal of rheumatology"] = "World J Rheumatol",
  ["world journal of science, technology and sustainable development"] = "World J. Sci. Technol. Sustainable Dev.",
  ["world journal of stem cells"] = "World J Stem Cells",
  ["world journal of surgery"] = "World J Surg",
  ["world journal of surgical oncology"] = "World J Surg Oncol",
  ["world journal of surgical procedures"] = "World J Surg Proced",
  ["world journal of traditional chinese medicine"] = "World J Tradit Chin Med",
  ["world journal of translational medicine"] = "World J Transl Med",
  ["world journal of transplantation"] = "World J Transplant",
  ["world journal of urology"] = "World J Urol",
  ["world journal of vaccines"] = "World J Vaccines",
  ["world journal of virology"] = "World J Virol",
  ["world leisure journal"] = "World Leis J",
  ["world medical & health policy"] = "World Med Health Policy",
  ["world medical journal"] = "World Med J",
  ["world medicine"] = "World Med",
  ["world meteorological organization bulletin"] = "WMO Bull.",
  ["world mycotoxin journal"] = "World Mycotoxin J",
  ["world neighbors in action"] = "World Neighb Action",
  ["world neurology"] = "World Neurol",
  ["world neurosurgery"] = "World Neurosurg",
  ["world neurosurgery: x"] = "World Neurosurg X",
  ["world news on maxillofacial radiology"] = "World News Maxillofac Radiol",
  ["world of irish nursing"] = "World Ir Nurs",
  ["world of irish nursing (dublin, ireland : 1995)"] = "World Ir Nurs",
  ["world of mining: surface and underground"] = "World Min. Surf. Underground",
  ["world of wireless mobile and multimedia networks. ieee international symposium on a world of wireless mobile and multimedia networks"] = "World Wirel Mob Multimed Netw",
  ["world of work : the magazine of the ilo"] = "World Work",
  ["world policy journal"] = "World Policy J",
  ["world politics"] = "World Polit",
  ["world psychiatry : official journal of the world psychiatric association (wpa)"] = "World Psychiatry",
  ["world rabbit science"] = "World Rabbit Sci.",
  ["world research journal of agricultural biotechnology"] = "World Res J Agric Biotechnol",
  ["world resource review"] = "World Resour. Rev.",
  ["world review (brisbane, qld.)"] = "World Review",
  ["world review of nutrition and dietetics"] = "World Rev Nutr Diet",
  ["world review of science technology and sustainable development"] = "World Rev. Sci. Technol. Sustainable Dev.",
  ["world scientific annual review of artificial intelligence"] = "World Sci. Annu. Rev. Artif. Intell.",
  ["world scientific annual review of biomechanics"] = "World Sci. Annu. Rev. Biomech.",
  ["world scientific annual review of cancer immunology"] = "World Sci. Annu. Rev. Cancer Immunol.",
  ["world scientific annual review of data science"] = "World Sci. Annu. Rev. Data Sci.",
  ["world scientific annual review of fintech"] = "World Sci. Annu. Rev. FinTech",
  ["world scientific annual review of functional materials"] = "World Sci. Annu. Rev. Funct. Mate.",
  ["world scientific annual review of particle physics"] = "World Sci. Annu. Rev. Part. Phys.",
  ["world scientific annual review of vaccine design"] = "World Sci. Annu. Rev. Vaccine Des.",
  ["world scientific handbook in financial economics series"] = "World Sci. Handb. Financ. Econ. Ser.",
  ["world scientific lecture notes in economics and policy"] = "World Sci. Lect. Notes Econ. Policy",
  ["world scientific lecture notes in finance"] = "World Sci. Lect. Notes Finance",
  ["world scientific lecture notes in physics"] = "World Sci. Lecture Notes Phys.",
  ["world scientific news"] = "World Sci News",
  ["world scientific series in 20th century mathematics"] = "World Sci. Ser. 20th Century Math.",
  ["world scientific series in 20th century physics"] = "World Sci. Ser. 20th Century Phys.",
  ["world scientific series in 21st century mathematics"] = "World Sci. Ser. 21st Century Math.",
  ["world scientific series in applicable analysis"] = "World Sci. Ser. Appl. Anal.",
  ["world scientific series in economic theory"] = "World Sci. Ser. Econ. Theory",
  ["world scientific series in finance"] = "World Sci. Ser. Finance",
  ["world scientific series in information studies"] = "World Sci. Ser. Inf. Stud.",
  ["world scientific series in robotics and intelligent systems"] = "World Sci. Ser. Robot. Intell. Systems",
  ["world scientific series on nonlinear science"] = "World Sci. Ser. Nonlinear Sci. Ser. A Monogr. Treatises",
  ["world scientific series on nonlinear science. series a: monographs and treatises"] = "World Sci. Ser. Nonlinear Sci. Ser. A Monogr. Treatises",
  ["world scientific series on nonlinear science. series b: special theme issues and proceedings"] = "World Sci. Ser. Nonlinear Sci. Ser. B Spec. Theme Issues Proc.",
  ["world scientific series on probability theory and its applications"] = "World Sci. Ser. Probab. Theory Appl.",
  ["world scientific series on quantum algorithms, information, and learning"] = "World. Sci. Ser. Quantum Algorithms Inf. Learn.",
  ["world scientific—now publishers series in business"] = "World Sci. Now Publ. Ser. Bus.",
  ["world smoking & health"] = "World Smoking Health",
  ["world tobacco"] = "World Tob",
  ["world watch"] = "World Watch",
  ["world water policy"] = "World Water Policy",
  ["world wide web"] = "World Wide Web",
  ["world wood"] = "World Wood",
  ["world's poultry science journal"] = "Worlds Poult Sci J",
  ["world-wide abstracts of general medicine"] = "World Wide Abstr Gen Med",
  ["worlds poultry science journal"] = "Worlds Poult. Sci. J.",
  ["worldviews on evidence-based nursing"] = "Worldviews Evid Based Nurs",
  ["worldviews on evidence-based nursing / sigma theta tau international, honor society of nursing"] = "Worldviews Evid Based Nurs",
  ["worldwide waste: journal of interdisciplinary studies"] = "Worldwide Waste: J. Interdiscip. Stud.",
  ["world’s poultry science journal"] = "World’s Poult. Sci. J.",
  ["wormbook : the online review of c. elegans biology"] = "WormBook",
  ["wort und dienst: jahrbuch der kirchlichen hochschule bethel (bei bielefeld)"] = "W&D",
  ["wosp 2017 : proceedings of the 6th international workshop on mining scientific publications. international workshop on mining scientific publications (6th : 2017 : toronto, ont.)"] = "WOSP 2017 (2017)",
  ["wound management & prevention"] = "Wound Manag Prev",
  ["wound repair and regeneration"] = "Wound Repair Regener.",
  ["wound repair and regeneration : official publication of the wound healing society [and] the european tissue repair society"] = "Wound Repair Regen",
  ["wounds : a compendium of clinical research and practice"] = "Wounds",
  ["wri issues and ideas"] = "WRI Issues Ideas",
  ["writing systems research"] = "Writ Syst Res",
  ["writings on american history: a subject bibliography of articles"] = "Writ Am Hist",
  ["written communication"] = "Writ Commun",
  ["written language and literacy"] = "Writ Lang Lit",
  ["wseas transactions on mathematics"] = "WSEAS Trans Math",
  ["wu li hua xue xue bao = acta physico-chimica sinica"] = "Wu Li Hua Xue Xue Bao",
  ["wuhan da xue xue bao. li xue ban"] = "Journal of Wuhan University. Natural science edition|Wuhan Da Xue Xue Bao Li Xue Ban",
  ["wuhan da xue xue bao. li xue ban = journal of wuhan university. natural science edition"] = "Wuhan Da Xue Xue Bao Li Xue Ban",
  ["wuhan da xue xue bao. zi ran ke xue ban = journal of wuhan university. natural sciences edition"] = "Wuhan Da Xue Xue Bao Zi Ran Ke Xue Ban",
  ["wuhan university"] = "Wuhan Univ. J. Nat. Sci.",
  ["wuhan university journal of natural sciences"] = "Wuhan Univ. J. Nat. Sci.",
  ["wuhan university. journal of natural sciences"] = "Wuhan Univ. J. Nat. Sci.",
  ["wuhan zhi wu xue yan jiu"] = "Wuhan botanical research|Wuhan Zhi Wu Xue Yan Jiu",
  ["wuhan zhi wu xue yan jiu = wuhan botanical research"] = "Wuhan Zhi Wu Xue Yan Jiu",
  ["wuji huaxue xuebao"] = "Wuji Huaxue Xuebao",
  ["wuli huaxue xuebao"] = "Wuli Huaxue Xuebao",
  ["wuli xuebao"] = "Wuli Xuebao",
  ["wuli xuebao (acta physica sinica) [chinese journal of physics]"] = "Wuli Xuebao (Acta Phys. Sin.) [Chin. J. Phys.]",
  ["wurzburger medizinhistorische forschungen"] = "Wurzbg Medizinhist Forsch",
  ["wurzburger medizinhistorische mitteilungen / im auftrage der wurzburger medizinhistorischen gesellschaft und in verbindung mit dem institut fur geschichte der medizin der universitat wurzburg"] = "Wurzbg Medizinhist Mitt",
  ["wyoming nurses newsletter"] = "Wyo. Nurses Newsl.",
  ["wyoming nurses' newsletter"] = "Wyo Nurses Newsl",
  ["wyoming nurses’ newsletter"] = "Wyo Nurses Newsl",
  ["wyoming statutes, annotated. wyoming"] = "Wyo Statut Annot Wyo",
  ["wythe county historical review"] = "Wythe Cty Hist Rev",
  ["wythe county historical review / wythe county historical society"] = "Wythe Cty Hist Rev",
  ["wyższa szkoła pedagogiczna im. komisji edukacji narodowej w krakowie. rocznik naukowo-dydaktyczny. prace matematyczne"] = "Rocznik Nauk.-Dydakt. Prace Mat.",
  ["würzburger jahrbücher für die altertumswissenschaft"] = "WJA",
  ["würzburger medizinhistorische forschungen"] = "Wurzbg Medizinhist Forsch",
  ["würzburger medizinhistorische mitteilungen"] = "Wurzbg Medizinhist Mitt",
  ["x-ray spectrometry"] = "X-Ray Spectrom.",
  ["x-ray spectrometry : xrs"] = "Xray Spectrom",
  ["x-ray structure analysis online"] = "X-Ray Struct. Anal. Online",
  ["xenia antiqua"] = "XeniaAnt",
  ["xenia. konstanzer althistorische vorträge und forschungen"] = "XeniaKonst",
  ["xenia. semestrale di antichità"] = "Xenia",
  ["xenobiotica; the fate of foreign compounds in biological systems"] = "Xenobiotica",
  ["xi bao sheng wu xue za zhi"] = "Xi Bao Sheng Wu Xue Za Zhi",
  ["xi bao yu fen zi mian yi xue za zhi = chinese journal of cellular and molecular immunology"] = "Xi Bao Yu Fen Zi Mian Yi Xue Za Zhi",
  ["xi bei nong lin ke ji da xue xue bao. zi ran ke xue ban = journal of northwest sci-tech university of agriculture and forestry. natural science edition"] = "Xi Bei Nong Lin Ke Ji Da Xue Xue Bao Zi Ran Ke Xue Ban.",
  ["xi bei nong ye xue bao = acta agriculturae boreali-occidentalis sinica"] = "Xi Bei Nong Ye Xue Bao",
  ["xi bei ren kou = xibeirenkou"] = "Xi Bei Ren Kou",
  ["xi bei zhi wu xue bao"] = "Acta botanica Boreali-Occidentalia Sinica|Xibei Zhiwu Xuebao",
  ["xi bei zhi wu xue bao = acta botanica boreali-occidentalia sinica"] = "Xibei Zhiwu Xuebao",
  ["xi nan nong ye da xue xue bao = journal of southwest agricultural university"] = "Xi Nan Nong Ye Da Xue Xue Bao",
  ["xi nan nong ye xue bao"] = "Southwest China journal of agricultural sciences|Xi Nan Nong Ye Xue Bao",
  ["xi nan nong ye xue bao = southwest china journal of agricultural sciences"] = "Xi Nan Nong Ye Xue Bao",
  ["xi psi phi quarterly"] = "Xi Psi Phi Q.",
  ["xi'nan minzu daxue xuebao. ziran kexue ban"] = "Xinan Minzu Daxue Xuebao Ziran Kexue Ban",
  ["xiamen da xue xue bao. zi ran ke xue ban"] = "Journal of Xiamen University. Natural science|Xiamen Da Xue Xue Bao Zi Ran Ke Xue Ban",
  ["xiamen da xue xue bao. zi ran ke xue ban = journal of xiamen university. natural science"] = "Xiamen Da Xue Xue Bao Zi Ran Ke Xue Ban",
  ["xiamen daxue xuebao"] = "Xiamen Daxue Xuebao Ziran Kexue Ban",
  ["xiamen daxue xuebao. ziran kexue ban. journal of xiamen university. natural science"] = "Xiamen Daxue Xuebao Ziran Kexue Ban",
  ["xian dai hua gong = modern chemical industry"] = "Xian Dai Hua Gong",
  ["xian dai sheng wu yi xue jin zhan"] = "Xian Dai Sheng Wu Yi Xue Jin Zhan",
  ["xian dai shi yong yi xue"] = "Xian Dai Shi Yong Yi Xue",
  ["xian dai yu fang yi xue = modern preventive medicine"] = "Xian Dai Yu Fang Yi Xue",
  ["xiandai shipin keji = modern food science and technology"] = "Xiandai Shipin Keji",
  ["xianggang hu li za zhi (hong kong nursing journal)"] = "Xianggang Hu Li Za Zhi",
  ["xianggang hu li za zhi. the hong kong nursing journal"] = "Xianggang Hu Li Za Zhi",
  ["xianggang she hui ke xue xue bao = hong kong journal of social sciences"] = "Xianggang She Hui Ke Xue Xue Bao",
  ["xianggang zhong wen da xue xue bao. chinese university of hong kong"] = "J Chin Univ Hong Kong",
  ["xibu linye kexue = journal of west china forestry science"] = "Xibu Linye Kexue",
  ["xin li ke xue jin zhan"] = "Xin Li Ke Xue Jin Zhan",
  ["xin li xue bao. acta psychologica sinica"] = "Xin Li Xue Bao",
  ["xinan shifan daxue xuebao"] = "Xinan Shifan Daxue Xuebao",
  ["xinjiang nong ye da xue xue bao = journal of xinjiang agricultural university"] = "Xinjiang Nong Ye Da Xue Xue Bao",
  ["xinjiang nong ye ke xue"] = "Xinjiang Nong Ye Ke Xue",
  ["xinjiang university"] = "J. Xinjiang Univ. Natur. Sci.",
  ["xi’an jiaotong daxue xuebao"] = "Xi’an Jiaotong Daxue Xuebao",
  ["xi’an jiaotong daxue xuebao. journal of xi’an jiaotong university"] = "Xi’an Jiaotong Daxue Xuebao",
  ["xrds: crossroads, the acm magazine for students"] = "XRDS: Crossroads, ACM Mag. Stud.",
  ["xu mu shou yi xue bao"] = "Acta veterinaria et zootechnica Sinica / [Zhongguo xu mu shou yi xue hui]|Xu Mu Shou Yi Xue Bao",
  ["xu mu shou yi xue bao = acta veterinaria et zootechnica sinica"] = "Xu Mu Shou Yi Xue Bao",
  ["xumu yu shouyi"] = "Xumu Yu Shouyi",
  ["xviie siècle"] = "XVIIe Siecle",
  ["xviii international aids conference, 2010 : abstracts. international conference on aids (18th : 2010 : vienna, austria)"] = "AIDS 2010",
  ["y (dayton, ohio)"] = "Y",
  ["y [report]. u.s. atomic energy commission"] = "Y Rep",
  ["y hoc viêt nam"] = "Y Hoc Viet Nam",
  ["y học thành phó̂ hò̂ chí minh"] = "Y Hoc Thanh Pho Ho Chi Minh",
  ["y reports"] = "Y Rep.",
  ["yad vashem studies"] = "Yad Vashem Stud Eur Jew Catastrophe Resist",
  ["yadernaya fizika"] = "Yad. Fiz.",
  ["yadernaya fizika [soviet journal of nuclear physics]"] = "Yad. Fiz. [Sov. J. Nucl. Phys.]",
  ["yakhak hoechi"] = "Yakhak Hoechi",
  ["yakhakhoe chi. journal of the pharmaceutical society of korea"] = "Yakhakhoe Chi",
  ["yakhteh = the cell"] = "Yakhteh",
  ["yakubutsu, seishin, kodo (japanese journal of psychopharmacology)"] = "Yakubutsu Seishin Kodo",
  ["yakubutsu, seishin, kōdō = japanese journal of psychopharmacology"] = "Yakubutsu Seishin Kodo",
  ["yakugaku zasshi"] = "Yakugaku Zasshi",
  ["yakugaku zasshi (journal of the pharmaceutical society of japan)"] = "Yakugaku Zasshi",
  ["yakugaku zasshi : journal of the pharmaceutical society of japan"] = "Yakugaku Zasshi",
  ["yakugaku zasshi. journal of the pharmaceutical society of japan"] = "Yakugaku Zasshi",
  ["yakuri to chiryo"] = "Yakuri to Chiryo",
  ["yakushigaku zasshi"] = "Yakushigaku Zasshi",
  ["yakushigaku zasshi. the journal of japanese history of pharmacy"] = "Yakushigaku Zasshi",
  ["yale classical studies"] = "YCIS",
  ["yale economic essays"] = "Yale Econ. Essays",
  ["yale journal of biology and medicine"] = "Yale J. Biol. Med.",
  ["yale journal of health policy, law, and ethics"] = "Yale J Health Policy Law Ethics",
  ["yale journal of law & the humanities"] = "Yale J Law Humanit",
  ["yale journal of law and feminism"] = "Yale J Law Fem",
  ["yale journal on regulation"] = "Yale J Regul",
  ["yale law & policy review"] = "Yale Law Policy Rev",
  ["yale law journal"] = "Yale Law J.",
  ["yale mathematical monographs"] = "Yale Math. Monogr.",
  ["yale scientific magazine"] = "Yale Sci Mag",
  ["yale university art gallery bulletin"] = "YaleUnivB",
  ["yale university library gazette"] = "YLG",
  ["yalkut moreshet"] = "Yalkut Moreshet",
  ["yalḳuṭ moreshet"] = "Yalkut Moreshet",
  ["yamashina chōrui kenkyūjo kenkyū hōkōku"] = "Yamashina Chorui Kenkyujo Kenkyu Hokoku",
  ["yamashina chōruigaku zasshi"] = "Yamashina Choruigaku Zasshi",
  ["yan jiu bao gao (taiwan)"] = "Yan Jiu Bao Gao (Taiwan)",
  ["yan ke xue bao (2016)"] = "Yan Ke Xue Bao (Hong Kong)",
  ["yan ke xue bao (eye science)"] = "Yan Ke Xue Bao",
  ["yan ke xue bao = eye science"] = "Yan Ke Xue Bao",
  ["yancao keji"] = "Yancao Keji",
  ["yang zhou da xue xue bao. nong ye yu sheng ming ke xue ban"] = "Yang Zhou Da Xue Xue Bao Nong Ye Yu Sheng Ming Ke Xue Ban",
  ["yao hsueh hsueh pao [acta pharmaceutica sinica]"] = "Yao Hsueh Hsueh Pao",
  ["yao wu fen xi za zhi = yaowu fenxi zazhi"] = "Yao Wu Fen Xi Za Zhi",
  ["yao wu sheng wu ji shu"] = "Yao Wu Sheng Wu Ji Shu",
  ["yao wu shi pin fen xi = journal of food and drug analysis"] = "Yao Wu Shi Pin Fen Xi",
  ["yao xue xue bao (acta pharmaceutica sinica)"] = "Yao Xue Xue Bao",
  ["yao xue xue bao = acta pharmaceutica sinica"] = "Yao Xue Xue Bao",
  ["yaowu fenxi zazhi"] = "Yaowu Fenxi Zazhi",
  ["yaoxue xuebao"] = "Yaoxue Xuebao",
  ["yapi kredi economic review"] = "Yapi Kredi Econ. Rev.",
  ["year book"] = "Year B Leo Baeck Inst",
  ["year book - american philosophical society"] = "Yearb Am Philos Soc",
  ["year book - carnegie institution of washington"] = "Year B Carnegie Inst Wash",
  ["year book - royal society of edinburgh"] = "Year B R Soc Edinb",
  ["year book. academy of applied osteopathy"] = "Yearb Acad Appl Osteopathy",
  ["year in immunology"] = "Year Immunol.",
  ["yearbook (conference of latin americanist geographers)"] = "Yearb Conf Lat Am Geogr",
  ["yearbook of german-american studies"] = "Yearb Ger Am Stud",
  ["yearbook of medical informatics"] = "Yearb Med Inform",
  ["yearbook of physical anthropology"] = "Yearb Phys Anthropol",
  ["yearbook of population research in finland = väestöntutkimuksen vuosikirja"] = "Yearb Popul Res Finl",
  ["yearbook of the estonian naturalist society"] = "Yearb. Eston. Nat. Soc.",
  ["yeast (chichester, england)"] = "Yeast",
  ["yejin fenxi"] = "Yejin Fenxi",
  ["yen ko hsueh pao [eye science] (chung-kuo kuang-chou"] = "Yen Ko Hsueh Pao",
  ["yeni tıp tarihi araştırmaları = the new history of medicine studies"] = "Yeni Tip Tarihi Arastirmalari",
  ["yenisei series in pure and applied mathematics"] = "Yenisei Ser. Pure Appl. Math.",
  ["yeungnam university journal of medicine"] = "Yeungnam Univ J Med",
  ["yi chuan = hereditas"] = "Yi Chuan",
  ["yi chuan xue bao (acta genetica sinica)"] = "Yi Chuan Xue Bao",
  ["yi chuan xue bao = acta genetica sinica"] = "Yi Chuan Xue Bao",
  ["yi yong sheng wu li xue"] = "Yi Yong Sheng Wu Li Xue",
  ["ying yong kun chong xue bao = chinese journal of applied entomology"] = "Ying Yong Kun Chong Xue Bao",
  ["ying yong sheng tai xue bao = the journal of applied ecology"] = "Ying Yong Sheng Tai Xue Bao",
  ["ying yong yu huan jing sheng wu xue bao"] = "Chinese journal of applied and environmental biology / Zhongguo ke xue yuan Chengdu sheng wu yan jiu suo zhu ban|Ying Yong Yu Huan Jing Sheng Wu Xue Bao",
  ["ying yong yu huan jing sheng wu xue bao = chinese journal of applied and environmental biology"] = "Ying Yong Yu Huan Jing Sheng Wu Xue Bao",
  ["yingxiang kexue yu guang huaxue"] = "Yingxiang Kexue Yu Guang Huaxue",
  ["yingyong huaxue"] = "Yingyong Huaxue",
  ["yngre laeger"] = "Yngre Laeger",
  ["yokohama mathematical journal"] = "Yokohama Math. J.",
  ["yokohama medical bulletin"] = "Yokohama Med Bull",
  ["yokufuen chosa kenkyu kiyo. acta gerontologica japonica"] = "Yokufuen Chosa Kenkyu Kiyo",
  ["yokufūen chōsa kenkyū kiyō. acta gerontologica japonica"] = "Yokufuen Chosa Kenkyu Kiyo",
  ["yonago acta medica"] = "Yonago Acta Med",
  ["yonago igaku zasshi. the journal of the yonago medical association"] = "Yonago Igaku Zasshi",
  ["yonsei medical journal"] = "Yonsei Med J",
  ["york state tradition"] = "York State Trad",
  ["yorkshire archaeological journal"] = "Yorkshire Archaeol. J.",
  ["yorkshire bull. econ. soc. res."] = "Yorkshire Bull. Econ. Soc. Res.",
  ["yosetsu gakkai ronbunshu"] = "Yosetsu Gakkai Ronbunshu",
  ["youji huaxue"] = "Youji Huaxue",
  ["young (stockholm, sweden)"] = "Young",
  ["young children"] = "Young Child",
  ["young minds"] = "Young Minds",
  ["your child patient"] = "Your Child Patient",
  ["your health"] = "Your Health (Salt Lake City)",
  ["your oklahoma dental association journal"] = "Your Okla Dent Assoc J",
  ["youth & society"] = "Youth Soc",
  ["youth violence and juvenile justice"] = "Youth Violence Juv Justice",
  ["yu mi ke xue"] = "Yu Mi Ke Xue",
  ["yuan yi xue bao"] = "Yuan Yi Xue Bao",
  ["yugoslav journal of operations research"] = "Yugosl. J. Oper. Res.",
  ["yugoslav journal of operations research. an international journal dealing with theoretical and computational aspects of operations research, systems science, and management science"] = "Yugosl. J. Oper. Res.",
  ["yugoslav survey"] = "Yugosl Surv",
  ["yuki gosei kagaku kyokaishi"] = "Yuki Gosei Kagaku Kyokaishi",
  ["yunnan nong ye da xue xue bao"] = "Journal of Yunnan Agricultural University|Yunnan Nong Ye Da Xue Xue Bao",
  ["yunnan nong ye da xue xue bao = journal of yunnan agricultural university"] = "Yunnan Nong Ye Da Xue Xue Bao",
  ["yunnan zhi wu yan jiu = yun nan zhi wu yan jiu"] = "Yunnan Zhi Wu Yan Jiu",
  ["z magazine (boston, mass.)"] = "Z Mag",
  ["z pola walki"] = "Z Pola Walki",
  ["zagazig veterinary journal"] = "Zagazig Vet J",
  ["zagreb international review of economics and business"] = "Zagreb Int. Rev. Econ. Bus.",
  ["zahn-, mund- und kieferheilkunde in vorträgen"] = "Zahnheilkd Mundheilkd Kieferheilkd Vortr",
  ["zahn-, mund-, und kieferheilkunde mit zentralblatt"] = "Zahn Mund Kieferheilkd Zentralbl",
  ["zahnarzteblatt baden-wurttemberg"] = "Zahnarztebl. Baden. Wurttemb.",
  ["zahnarztliche mitteilungen"] = "Zahnarztl. Mitt.",
  ["zahnarztliche nachrichten sudwurttemburg-hohenzollern"] = "Zahnarztl Nachr Sudwurttemb",
  ["zahnarztliche praxis"] = "Zahnarztl. Prax.",
  ["zahnarztliche praxisfuhrung"] = "Zahnarztl Praxisfuhr",
  ["zahnarztliche rundschau"] = "Zahnarztl. Rundsch.",
  ["zahnarztliche welt"] = "Zahnarztl Welt",
  ["zahnarztliche welt, zahnarztliche reform, zwr"] = "Zahnarztl Welt Zahnarztl Reform Zwr",
  ["zahnarztliche welt, zahnarztliche rundschau, zwr, vereinigt mit zahnarztliche reform"] = "Zahnarztl. Welt. Zahnarztl. Rundsch. ZWR. Zahnarztl. Reform",
  ["zahnarztlicher gesundheitsdienst"] = "Zahnarztl. Gesundheitsdienst",
  ["zahnarztlicher gesundheitsdienst : offizielles organ des bundesverbandes der zahnarzte des offentlichen gesundheitsdienstes e.v"] = "Zahnarztl Gesundheitsdienst",
  ["zahnãrztliche welt, zahnãrztliche rundschau, zwr, vereinigt mit zahnãrztliche reform"] = "Zahnarztl Welt Zahnarztl Rundsch ZWR Zahnarztl Reform",
  ["zahnärzteblatt baden-württemberg"] = "Zahnarztebl Baden Wurttemb",
  ["zahnärztliche mitteilungen"] = "Zahnarztl Mitt",
  ["zahnärztliche praxis"] = "Zahnarztl Prax",
  ["zahnärztliche rundschau"] = "Zahnarztl Rundsch",
  ["zahnärztliche welt"] = "Zahnarztl Welt",
  ["zahnärztliche welt, zahnärztliche reform, zwr"] = "Zahnarztl Welt Zahnarztl Reform Zwr",
  ["zahnärztlicher gesundheitsdienst : offizielles organ des bundesverbandes der zahnärzte des öffentlichen gesundheitsdienstes e.v"] = "Zahnarztl Gesundheitsdienst",
  ["zahntechnik (zurich, switzerland)"] = "Zahntechnik (Zur)",
  ["zahntechnik; zeitschrift fur theorie und praxis der wissenschaftlichen zahntechnik"] = "Zahntechnik (Berl)",
  ["zahntechnik; zeitschrift für theorie und praxis der wissenschaftlichen zahntechnik"] = "Zahntechnik (Berl)",
  ["zahradnictví (prague, czech republic : 1992)"] = "Zahradnictvi (Praha)",
  ["zambia nurse"] = "Zambia Nurse",
  ["zambia nurse (kitwe, zambia : 1978)"] = "Zambia Nurse",
  ["zambia nurse journal"] = "Zambia Nurse J.",
  ["zamm. zeitschrift fur angewandte mathematik und mechanik. journal of applied mathematics and mechanics"] = "ZAMM Z. Angew. Math. Mech.",
  ["zanco journal of medical sciences"] = "Zanco J Med Sci",
  ["zanco journal of pure and applied sciences"] = "Zanco J Pure Appl Sci",
  ["zapiski vsesoyuznogo mineralogicheskogo obshchestva"] = "Zap. Vses. Mineral. O-va.",
  ["zaragoza (zaragoza (spain))"] = "Zaragoza",
  ["zaragoza (zaragoza (spain))."] = "Zaragoza",
  ["zashchita i karantin rasteniĭ"] = "Zas Karantin Rast",
  ["zashchita metallov"] = "Zashch. Met.",
  ["zasshi. tokyo ika daigaku"] = "Zasshi Tokyo Ika Daigaku",
  ["zaštita bilja"] = "Zast Bilja",
  ["zbirka izbranih poglavij iz fizike"] = "Zbirka Izbran. Poglav. Fiz.",
  ["zbirka izbranih poglavij iz matematike"] = "Zbirka Izbran. Poglav. Mat.",
  ["zbirnik prats institutu matematiki nan ukraini"] = "Zb. Pr. Inst. Mat. NAN Ukr.",
  ["zbornik biotehniške fakultete univerze v ljubljani. kmetijstvo (1990)"] = "Zb Bioteh Fak Univ Ljubl Kmet Supl",
  ["zbornik instituta za historijske nauke u zadru"] = "ZborZadar",
  ["zbornik matice srpske za istoriju"] = "Zb Matice Srp Istor",
  ["zbornik matice srpske za prirodne nauke"] = "Zb Matice Srp Prir Nauk",
  ["zbornik narodnog muzeja beograd"] = "ZborMuzBeograd",
  ["zbornik odsjeka za povijesne znanosti zavoda za povijesne i društvene znanosti hrvatske akademije znanosti i umjetnosti"] = "Zb Odsjeka Povij Znan Zovoda Povij Drus Znan Hrvat Akad Znan Umjet",
  ["zbornik prac. univerzita pavla jozefa safarika v kosiciach. lekarska fakulta"] = "Zb Lek Fak Kosice",
  ["zbornik radova"] = "Zb. Rad. (Beogr.)",
  ["zbornik radova ekonomskog fakulteta rijeka"] = "Zbornik Radova Ekon. Fakul. Rijeka",
  ["zbornik radova ekonomskog fakulteta u. rijeci"] = "Zbornik Radova Ekon. Fakul. U. Rijeci",
  ["zbornik radova nauc̆nog sastanka"] = "Zb Rad Naunog Sastanka",
  ["zbornik radova vizantološkog instituta. recueil des travaux de l’institut d’études byzantines, beograd"] = "ZborRadBeograd",
  ["zbornik radova. (beograd)"] = "Zb. Rad. (Beogr.)",
  ["zbornik vojnomedicinske akademije"] = "Zb. Vojnomed. Akad.",
  ["zbornik za istoriju"] = "Zb Istor",
  ["zbornik za istoriju bosne i hercegovine = recueil de l'histoire de bosnie et herzegovine"] = "Zb Istor Bosne Herceg",
  ["zbornik za zgodovino naravoslovja in tehnike / slovenska matica v ljubljani"] = "Zb Zgodovino Naravoslovja Teh",
  ["zbornik za zgodovino naravoslovja in tehnike = history reviews, science and technology"] = "Zb Zgodovino Naravoslovja Teh",
  ["zbornik. vojnomedicinska akademija (yugoslavia)"] = "Zb Vojnomed Akad",
  ["zborník prác. univerzita pavla jozefa safárika v kosiciach. lekárska fakulta"] = "Zb Lek Fak Kosice",
  ["zdm : the international journal on mathematics education"] = "ZDM",
  ["zdm mathematics education"] = "ZDM Math. Educ.",
  ["zdravno delo"] = "Zdr Delo",
  ["zdravookhranenie (minsk, belarus)"] = "Zdravookhranenie (Minsk)",
  ["zdravookhranenie belarusi (belorusskii tsentr nauchnoi meditsinskoi informatsii)"] = "Zdravookhranenie Belarusi",
  ["zdravookhranenie belarusi (belorusskiĭ tsentr nauchnoĭ meditsinskoĭ informatsii)"] = "Zdravookhranenie Belarusi",
  ["zdravookhranenie belorussii"] = "Zdravookhr Beloruss",
  ["zdravookhranenie kazakhstana"] = "Zdravookhranenie Kazakhstana",
  ["zdravookhranenie kirgizii"] = "Zdravookhr Kirg",
  ["zdravookhranenie rossiiskoi federatsii"] = "Zdravookhr. Ross. Fed.",
  ["zdravookhranenie rossiiskoi federatsii / ministerstvo zdravookhraneniia rsfsr"] = "Zdravookhr Ross Fed",
  ["zdravookhranenie rossiĭskoĭ federatsii"] = "Zdravookhr Ross Fed",
  ["zdravookhranenie turkmenistana"] = "Zdravookhr Turkmenistana",
  ["zdravotnicka pracovnice"] = "Zdrav. Prac.",
  ["zdravotnicka revue; vestnik ministerstva zdravotnictvi"] = "Zdrav Rev Vestn Minist Zdrav",
  ["zdravotnická pracovnice"] = "Zdrav Prac",
  ["zdravotnická revue; vĕstník ministerstva zdravotnictví"] = "Zdrav Rev Vestn Minist Zdrav",
  ["zdravotnicke aktuality"] = "Zdrav. Aktual.",
  ["zdravotnicke aktuality,"] = "Zdrav Aktual",
  ["zdravotnické aktuality"] = "Zdrav Aktual",
  ["zdravstveni radnik"] = "Zdrav Radn",
  ["zdravstveni vestnik"] = "Zdrav Vestn",
  ["zdravstveno varstvo"] = "Zdr Varst",
  ["zdrowie publiczne"] = "Zdrow Publiczne",
  ["zdrowie publiczne / polskie towarzystwo higjeniczne"] = "Zdrow Publiczne",
  ["zeiss – informationen"] = "Zeiss – Inf.",
  ["zeitschrift der arbeitsgemeinschaft österreichischer entomologen"] = "Z. Arb.gem. österr. Entomol.",
  ["zeitschrift der deutschen geologischen gesellschaft"] = "Z. Dtsch. Geol. Ges.",
  ["zeitschrift der deutschen gesellschaft fur geowissenschaften"] = "Z. Dtsch. Ges. Geowiss",
  ["zeitschrift der deutschen morgenlandischen gesellschaft"] = "Z Dtsch Morgenl Ges",
  ["zeitschrift der deutschen morgenländischen gesellschaft"] = "Z Dtsch Morgenl Ges",
  ["zeitschrift der deutschen morgenländischen gesellschaft"] = "ZDMG",
  ["zeitschrift der gesellschaft fur schleswig-holsteinische geschichte. gesellschaft fur schleswig-holsteinische geschichte"] = "Z Ges Schleswig Holstein Gesch",
  ["zeitschrift der gesellschaft für schleswig-holsteinische geschichte. gesellschaft für schleswig-holsteinische geschichte"] = "Z Ges Schleswig Holstein Gesch",
  ["zeitschrift der ostpreussischen meschkinnes forschung"] = "Z. Ostpreuss. Meschkinnes Forsch.",
  ["zeitschrift der savigny-stiftung fur rechtsgeschichte. germanistische abteilung"] = "Z Savigny Stift Rechtsgesch Ger Abt",
  ["zeitschrift der savigny-stiftung fur rechtsgeschichte. kanonistische abteilung"] = "Z Savigny Stift Rechtsgesch Kanon Abt",
  ["zeitschrift der savigny-stiftung für rechtsgeschichte. germanistische abteilung"] = "Z Savigny Stift Rechtsgesch Ger Abt",
  ["zeitschrift der savigny-stiftung für rechtsgeschichte. kanonistische abteilung"] = "Z Savigny Stift Rechtsgesch Kanon Abt",
  ["zeitschrift der savigny-stiftung für rechtsgeschichte (romanistische abteilung)"] = "ZRG",
  ["zeitschrift der savigny-stiftung für rechtsgeschichte. romanistische abteilung"] = "ZSav",
  ["zeitschrift der unfug gesellschaft"] = "Z. Unfug Ges.",
  ["zeitschrift des aachener geschichtsvereins. aachener geschichtsverein"] = "Z Aachener Geschichtsver",
  ["zeitschrift des deutschen palästina-vereins"] = "ZPalV",
  ["zeitschrift des historischen vereines fur steiermark"] = "Z Hist Ver Steiermark",
  ["zeitschrift des historischen vereines für steiermark"] = "Z Hist Ver Steiermark",
  ["zeitschrift des historischen vereins fur schwaben. historischer verein fur schwaben"] = "Z Hist Ver Schwaben Neuburg",
  ["zeitschrift des historischen vereins für schwaben. historischer verein für schwaben"] = "Z Hist Ver Schwaben Neuburg",
  ["zeitschrift des koìˆniglichen preussischen statistischen bureaus"] = "Z. K. Preuss. Stat. Bur.",
  ["zeitschrift des vereines der deutschen zucker industrie"] = "Z. Ver. Dtsch. Zucker Ind.",
  ["zeitschrift des vereins deutscher ingenieure"] = "Z. Ver. Dtsch. Ing.",
  ["zeitschrift des vereins fur lubeckische geschichte und altertumskunde. verein fur lubeckische geschichte und altertumskunde"] = "Z Ver Lubeck Gesch Altertumskd",
  ["zeitschrift des vereins für lübeckische geschichte und altertumskunde. verein für lübeckische geschichte und altertumskunde"] = "Z Ver Lubeck Gesch Altertumskd",
  ["zeitschrift fuer angewandte mathematik und physik"] = "Z Angew Math Phys",
  ["zeitschrift fuer anorganische und allgemeine chemie"] = "Z. Anorg. Allg. Chem.",
  ["zeitschrift fuer kristallographie"] = "Z. Kristallogr.",
  ["zeitschrift fuer kristallographie - crystalline materials"] = "Z. Kristallogr. - Cryst. Mater.",
  ["zeitschrift fuer kristallographie - new crystal structures"] = "Z. Kristallogr. - New Cryst. Struct.",
  ["zeitschrift fuer lebensmittel-untersuchung und -forschung a: european food research and technology"] = "Z. Lebensm.-Unters. Forsch. A",
  ["zeitschrift fuer metallkunde"] = "Z. Metallkd.",
  ["zeitschrift fuer naturforschung, a: physical sciences"] = "Z. Naturforsch., A: Phys. Sci.",
  ["zeitschrift fuer naturforschung, b: a journal of chemical sciences"] = "Z. Naturforsch., B: J. Chem. Sci.",
  ["zeitschrift fuer naturforschung, b: chemical sciences"] = "Z. Naturforsch., B: Chem. Sci.",
  ["zeitschrift fuer naturforschung, c: journal of biosciences"] = "Z. Naturforsch., C: J. Biosci.",
  ["zeitschrift fuer physikalische chemie (muenchen, germany)"] = "Z. Phys. Chem. (Muenchen, Ger.)",
  ["zeitschrift fur agyptische sprache und altertumskunde"] = "Z Aegypt Sprach Altertumskd",
  ["zeitschrift fur allgemeine mikrobiologie"] = "Z. allg. Mikrobiol.",
  ["zeitschrift fur allgemeine wissenschaftstheorie. journal for general philosophy of science"] = "Z Allg Wissenschaftstheor",
  ["zeitschrift fur allgemeinmedizin"] = "Z. Allgemeinmed.",
  ["zeitschrift fur alternsforschung"] = "Z. Alternsforsch.",
  ["zeitschrift fur analysis und ihre anwendungen"] = "Z. Anal. Anwend.",
  ["zeitschrift fur analysis und ihre anwendungen. journal of analysis and its applications"] = "Z. Anal. Anwend.",
  ["zeitschrift fur analytische chemie"] = "Z. Anal. Chem.",
  ["zeitschrift fur anatomie und entwicklungsgeschichte"] = "Z. Anat. Entwicklungsgesch.",
  ["zeitschrift fur angewandte bader- und klimaheilkunde"] = "Z Angew Bader Klimaheilkd",
  ["zeitschrift fur angewandte entomologie"] = "Z. Angew. Entomol.",
  ["zeitschrift fur angewandte geologie"] = "Z. Angew. Geol.",
  ["zeitschrift fur angewandte ichthyologie"] = "Journal of applied ichthyology|J. Appl. Ichthyol.",
  ["zeitschrift fur angewandte mathematik und mechanik"] = "ZAMM",
  ["zeitschrift fur angewandte mathematik und physik"] = "ZAMP",
  ["zeitschrift fur angewandte mathematik und physik. zamp. journal of applied mathematics and physics. journal de mathematiques et de physique appliquees"] = "Z. Angew. Math. Phys.",
  ["zeitschrift fur angewandte photographie in wissenschaft und technik"] = "Z. Angew. Photogr. Wiss. Tech.",
  ["zeitschrift fur angewandte physik und chemie"] = "Z. Elektrochem. Angew. Phys. Chem.",
  ["zeitschrift fur anorganische chemie"] = "Z. Anorg. Chem.",
  ["zeitschrift fur anorganische und allgemeine chemie"] = "ZAAC",
  ["zeitschrift fur arbeitswissenschaft"] = "Z Arbeitswiss",
  ["zeitschrift fur arznei und gewurzpflanzen"] = "Z. Arznei Gewurzpflanzen",
  ["zeitschrift fur arztliche fortbildung"] = "Z. Arztl. Fortbild.",
  ["zeitschrift fur arztliche fortbildung (berlin)"] = "Z Arztl Fortbild (Berl)",
  ["zeitschrift fur arztliche fortbildung und qualitatssicherung"] = "Z. Arztl. Fortbild. Qualitatssich.",
  ["zeitschrift fur arztliche fortbildung und qualitatssicherung : in zusammenarbeit mit der kaiserin-friedrich-stiftung fur das arztliche fortbildungswesen"] = "Z Arztl Fortbild Qualitatssich",
  ["zeitschrift fur arztliche fortbildung. beiheft"] = "Z. Arztl. Fortbild. Beih. (Jena)",
  ["zeitschrift fur bayerische landesgeschichte"] = "Z Bayer Landesgesch",
  ["zeitschrift fur bevolkerungswissenschaft"] = "Z Bevolkerungswiss",
  ["zeitschrift fur bibliothekswesen und bibliographie"] = "Z Bibliothekswes Bibliogr",
  ["zeitschrift fur bienenforschung"] = "Z. Bienenforsch.",
  ["zeitschrift fur biologie"] = "Z. Biol.",
  ["zeitschrift fur chemie"] = "Z. Chem.",
  ["zeitschrift fur das gesamte schiess und sprengstoffwesen"] = "Z. Gesamte Schiess Sprengstoffwes.",
  ["zeitschrift fur den erdkundeunterricht"] = "Z Erdkundeunterr",
  ["zeitschrift fur deutsches alterthum und deutsche litteratur"] = "Z Dtsch Altert Dtsch Lit",
  ["zeitschrift fur didaktik der naturwissenschaften"] = "Z. Didakt. Naturwiss.",
  ["zeitschrift fur die gesamte experimentelle medizin"] = "Z Gesamte Exp Med",
  ["zeitschrift fur die gesamte experimentelle medizin einschliesslich experimentelle chirurgie"] = "Z. Gesamte Exp. Med.",
  ["zeitschrift fur die gesamte hygiene und ihre grenzgebiete"] = "Z. Gesamte Hyg.",
  ["zeitschrift fur die gesamte innere medizin und ihre grenzgebiete"] = "Z. Gesamte Inn. Med.",
  ["zeitschrift fur elektrochemie"] = "Z. Elektrochem.",
  ["zeitschrift fur elektrochemie und angewandte physikalische chemie"] = "Z. Elektrochem. Angew. Phys. Chem.",
  ["zeitschrift fur elektrotechnik und elektrochemie"] = "Z. Elektrotech. Elektrochem.",
  ["zeitschrift fur entwicklungspsychologie und padagogische psychologie"] = "Z Entwicklungspsychol Padagog Psychol",
  ["zeitschrift fur erkrankungen der atmungsorgane"] = "Z. Erkr. Atmungsorgane",
  ["zeitschrift fur erkrankungen der atmungsorgane mit folia bronchologica"] = "Z Erkr Atmungsorgane Folia Bronchol",
  ["zeitschrift fur erkrankungen der atmungsorgane, mit folia bronchologica"] = "Z. Erkr. Atmungsorgane. Folia Bronchol.",
  ["zeitschrift fur ernahrungswissenschaft"] = "Z Ernahrungswiss",
  ["zeitschrift fur ernahrungswissenschaft (journal of nutritional sciences) supplementa"] = "Z. Ernahrungswiss. Suppl.",
  ["zeitschrift fur ernahrungswissenschaft. journal of nutritional sciences. supplementa"] = "Z Ernahrungswiss Suppl",
  ["zeitschrift fur erzbergbau und metallhuttenwesen"] = "Z. Erzbergbau MetallhuttenWes.",
  ["zeitschrift fur ethnologie"] = "Z Ethnol",
  ["zeitschrift fur experimentelle chirurgie"] = "Z. Exp. Chir.",
  ["zeitschrift fur experimentelle chirurgie, transplantation, und kunstliche organe"] = "Z. Exp. Chir. Transplant. Kunstliche Organe",
  ["zeitschrift fur experimentelle chirurgie, transplantation, und kunstliche organe : organ der sektion experimentelle chirurgie der gesellschaft fur chirurgie der ddr"] = "Z Exp Chir Transplant Kunstliche Organe",
  ["zeitschrift fur experimentelle pathologie und therapie"] = "Z. Exp. Pathol. Ther.",
  ["zeitschrift fur experimentelle psychologie"] = "Z. Exp. Psychol.",
  ["zeitschrift fur experimentelle psychologie : organ der deutschen gesellschaft fur psychologie"] = "Z Exp Psychol",
  ["zeitschrift fur experimentelle und angewandte psychologie"] = "Z. Exp. Angew. Psychol.",
  ["zeitschrift fur familienforschung"] = "Z Fam Forsch",
  ["zeitschrift fur feldherpetologie"] = "Z Feldherpetol",
  ["zeitschrift fur flugtechnik und motorluftschiffahrt"] = "Z. Flugtech. Motorluftschiffahrt",
  ["zeitschrift fur flugwissenschaften und weltraumforschung"] = "Z. Flugwiss. Weltraumforsch.",
  ["zeitschrift fur gastroenterologie"] = "Z Gastroenterol",
  ["zeitschrift fur gastroenterologie. verhandlungsband"] = "Z. Gastroenterol. Verh.",
  ["zeitschrift fur geburtshilfe und gynakologie"] = "Z Geburtshilfe Gynakol",
  ["zeitschrift fur geburtshilfe und neonatologie"] = "Z Geburtshilfe Neonatol",
  ["zeitschrift fur geburtshilfe und perinatologie"] = "Z. Geburtshilfe Perinatol.",
  ["zeitschrift fur geologische wissenschaften"] = "Z. Geol. Wiss.",
  ["zeitschrift fur geomorphologie"] = "Z. Geomorphol.",
  ["zeitschrift fur geomorphologie supplementband"] = "Z. Geomorphol. Suppl.",
  ["zeitschrift fur geophysik"] = "Z. Geophys.",
  ["zeitschrift fur gerontologie"] = "Z. Gerontol.",
  ["zeitschrift fur gerontologie und geriatrie"] = "Z. Gerontol. Geriatr.",
  ["zeitschrift fur gerontologie und geriatrie : organ der deutschen gesellschaft fur gerontologie und geriatrie"] = "Z Gerontol Geriatr",
  ["zeitschrift fur geschichte der arabisch-islamischen wissenschaften"] = "Z. Gesch. Arab.-Islam. Wiss.",
  ["zeitschrift fur geschichtswissenschaft"] = "Z Geschichtswiss",
  ["zeitschrift fur geschiebeforschung"] = "Z. Geschiebeforsch.",
  ["zeitschrift fur gletscherkunde und glazialgeologie"] = "Z. Gletscherkd. Glazialgeol.",
  ["zeitschrift fur haut- und geschlechtskrankheiten"] = "Z. Haut. Geschlechtskr.",
  ["zeitschrift fur hautkrankheiten"] = "Z. Hautkr.",
  ["zeitschrift fur heereskunde"] = "Z Heereskd",
  ["zeitschrift fur historische forschung"] = "Z Hist Forsch",
  ["zeitschrift fur hochfrequenztechnik"] = "Z. Hochfrequenztech.",
  ["zeitschrift fur hygiene"] = "Z. Hyg.",
  ["zeitschrift fur hygiene und infektionskrankheiten"] = "Z. Hyg. Infektionskrankh.",
  ["zeitschrift fur hygiene und infektionskrankheiten; medizinische mikrobiologie, immunologie und virologie"] = "Z Hyg Infektionskr",
  ["zeitschrift fur immunitats- und allergieforschung"] = "Z Immunitats Allergieforsch",
  ["zeitschrift fur immunitatsforschung (immunobiology)"] = "Z. Immunitatsforsch. Immunobiol.",
  ["zeitschrift fur immunitatsforschung und experimentalle therapie"] = "Z. Immunitatsforsch. Exp. Ther.",
  ["zeitschrift fur immunitatsforschung und experimentelle therapie"] = "Z Immun exp ther",
  ["zeitschrift fur immunitatsforschung, allergie und klinische immunologie"] = "Z. Immunitatsforsch. Allerg. Klin. Immunol.",
  ["zeitschrift fur immunitatsforschung, experimentelle und klinische immunologie"] = "Z. Immunitatsforsch. Exp. Klin. Immunol.",
  ["zeitschrift fur immunitatsforschung. immunobiology"] = "Z Immunitatsforsch Immunobiol",
  ["zeitschrift fur induktive abstammungs- und vererbungslehre"] = "Z Indukt Abstamm Vererbungsl",
  ["zeitschrift fur instrumentenkunde"] = "Z. Instrumentenkd.",
  ["zeitschrift fur jagdwissenschaft"] = "Z. Jagdwiss.",
  ["zeitschrift fur kardiologie"] = "Z Kardiol",
  ["zeitschrift fur kardiologie. supplementum"] = "Z. Kardiol. Suppl.",
  ["zeitschrift fur kinder- und jugendpsychiatrie"] = "Z. Kinder. Jugenpsychiatr.",
  ["zeitschrift fur kinder- und jugendpsychiatrie und psychotherapie"] = "Z. Kinder. Jugendpsychiatr. Psychother.",
  ["zeitschrift fur kinderchirurgie"] = "Z. Kinderchir.",
  ["zeitschrift fur kinderchirurgie und grenzgebiete"] = "Z. Kinderchir. Grenzgeb.",
  ["zeitschrift fur kinderheilkunde"] = "Z. Kinderheilkd.",
  ["zeitschrift fur kirchengeschichte"] = "Z Kirchengesch",
  ["zeitschrift fur klinische chemie und klinische biochemie"] = "Z. Klin. Chem. Klin. Biochem.",
  ["zeitschrift fur klinische medicin"] = "Z. Klin. Med.",
  ["zeitschrift fur klinische medizin"] = "Z. Klin. Med.",
  ["zeitschrift fur klinische psychologie und psychotherapie"] = "Z. Klin. Psychol. Psychother.",
  ["zeitschrift fur klinische psychologie, psychiatrie und psychotherapie"] = "Z. Klin. Psychol. Psychiatr. Psychother.",
  ["zeitschrift fur klinische psychologie, psychiatrie und psychotherapie / im auftrag der gorres-gesellschaft"] = "Z Klin Psychol Psychiatr Psychother",
  ["zeitschrift fur klinische psychologie, psychopathologie und psychotherapie"] = "Z. Klin. Psychol. Psychopathol. Psychother.",
  ["zeitschrift fur klinische psychologie, psychopathologie und psychotherapie / im auftrag der gorres-gesellschaft"] = "Z Klin Psychol Psychopathol Psychother",
  ["zeitschrift fur krankenpflege"] = "Z. Krankenpfl.",
  ["zeitschrift fur krankenpflege. revue suisse des infirmieres"] = "Z Krankenpfl",
  ["zeitschrift fur krebsforschung"] = "Z. Krebsforsch.",
  ["zeitschrift fur krebsforschung und klinische onkologie (cancer research and clinical oncology)"] = "Z. Krebsforsch. Klin. Onkol. Cancer Res. Clin. Oncol.",
  ["zeitschrift fur krebsforschung und klinische onkologie. cancer research and clinical oncology"] = "Z Krebsforsch Klin Onkol Cancer Res Clin Oncol",
  ["zeitschrift fur kreislaufforschung"] = "Z. Kreislaufforsch.",
  ["zeitschrift fur kristallograhie"] = "Z. Kristallogr.",
  ["zeitschrift fur kristallograhie - new crystral stuctures"] = "Z. Kristallogr. - New Cryst. Struct.",
  ["zeitschrift fur kristallographie"] = "Z. Kristallogr.",
  ["zeitschrift fur kristallographie -  crystalline materials"] = "Z. Kristallogr. - Cryst. Mater.",
  ["zeitschrift fur kristallographie - new crystral stuctures"] = "Z. Kristallogr. - New Cryst. Struct.",
  ["zeitschrift fur kristallographie, kristallgeometrie, kristallphysik, kristallchemie"] = "Z. Kristallogr. Kristallgeom. Kristallphys. Kristallchem.",
  ["zeitschrift fur kristallographie, mineralogie und petrographie"] = "Z. Kristallogr. Mineral. Petrogr.",
  ["zeitschrift fur kunstgeschichte"] = "Z Kunstgesch",
  ["zeitschrift fur laryngologie, rhinologie, otologie und ihre grenzgebiete"] = "Z. Laryngol. Rhinol. Otol.",
  ["zeitschrift fur lebensmittel-untersuchung und -forschung"] = "Z. Lebensm. Unters. Forsch.",
  ["zeitschrift fur lymphologie (journal of lymphology)"] = "Z. Lymphol.",
  ["zeitschrift fur lymphologie. journal of lymphology"] = "Z Lymphol",
  ["zeitschrift fur medizinische isotopenforschung und deren grenzgebiete"] = "Z. Med. Isotopenforsch. Deren Grenzgeb.",
  ["zeitschrift fur medizinische laboratoriumsdiagnostik"] = "Z. Med. Laboratoriumsdiagn.",
  ["zeitschrift fur medizinische labortechnik"] = "Z. Med. Labortech.",
  ["zeitschrift fur medizinische mikrobiologie und immunologie"] = "Z. Med. Mikrobiol. Immunol.",
  ["zeitschrift fur medizinische physik"] = "Z. Med. Phys.",
  ["zeitschrift fur medizinische psychologie : zmp"] = "Z Med Psychol",
  ["zeitschrift fur menschliche vererbungs- und konstitutionslehre"] = "Z Mensch Vererb Konstitutionsl",
  ["zeitschrift fur metailkunde"] = "Z. MetaIlkd.",
  ["zeitschrift fur mikroskopisch-anatomische forschung"] = "Z. Mikrosk. Anat. Forsch.",
  ["zeitschrift fur morphologie und anthropologie"] = "Z Morphol Anthropol",
  ["zeitschrift fur naturforschung a: journal of physical sciences"] = "Z. Naturforsch., A: Phys. Sci.",
  ["zeitschrift fur naturforschung b: journal of chemical sciences"] = "Z. Naturforsch., B: Chem. Sci.",
  ["zeitschrift fur naturforschung c: journal of biosciences"] = "Z. Naturforsch., C: Biosci.",
  ["zeitschrift fur naturforschung. c, journal of biosciences"] = "Z Naturforsch [C]",
  ["zeitschrift fur naturforschung. section c, biosciences"] = "Z. Naturforsch. [C]",
  ["zeitschrift fur naturforschung. section c, journal of biosciences"] = "Z. Naturforsch. [C]",
  ["zeitschrift fur naturforschung. section c. journal of biosciences"] = "Z Naturforsch [C]",
  ["zeitschrift fur naturforschung. section c: biosciences"] = "Z Naturforsch [C]",
  ["zeitschrift fur naturforschung. teil b, anorganische chemie, organische chemie, biochemie, biophysik, biologie"] = "Z. Naturforsch. [B]",
  ["zeitschrift fur naturforschung. teil b, chemie, biochemie, biophysik, biologie und verwandte gebiete"] = "Z. Naturforsch. B",
  ["zeitschrift fur naturforschung. teil b. anorganische chemie, organische chemie, biochemie, biophysik, biologie"] = "Z Naturforsch [B]",
  ["zeitschrift fur naturforschung. teil b: chemie, biochemie, biophysik, biologie"] = "Z Naturforsch B",
  ["zeitschrift fur naturforschung. teil c: biochemie, biophysik, biologie, virologie"] = "Z Naturforsch [C]",
  ["zeitschrift fur naturwissenschaftlich-medizinische grundlagenforschung"] = "Z. Naturwiss. Med. Grundlagenforsch.",
  ["zeitschrift fur naturwissenschaftlichmedizinische grundlagenforschung"] = "Z Naturwiss Med Grundlagenforsch",
  ["zeitschrift fur neurologie"] = "Z. Neurol.",
  ["zeitschrift fur orthopadie und ihre grenzgebiete"] = "Z Orthop Ihre Grenzgeb",
  ["zeitschrift fur orthopadie und unfallchirurgie"] = "Z Orthop Unfall",
  ["zeitschrift fur ostforschung"] = "Z Ostforsch",
  ["zeitschrift fur ostpreussische meschkinnes forschung"] = "Z. Ostpreuss. Meschkinnes Forsch.",
  ["zeitschrift fur papyrologie und epigraphik"] = "Z Papyrologie Epigraphik",
  ["zeitschrift fur parasitenkunde"] = "Z. Parasitenkd.",
  ["zeitschrift fur parasitenkunde (berlin, germany)"] = "Z Parasitenkd",
  ["zeitschrift fur pflanzenernahrung dungung bodenkunde"] = "Z. Pflanzenernahr. Dung. Bodenkd.",
  ["zeitschrift fur pflanzenernahrung und bodenkunde"] = "Z. Pflanzenernahr. Bodenkd.",
  ["zeitschrift fur pflanzenkrankheiten und pflanzenschutz"] = "Z. Pflanzenkrankh. Pflanzenschutz",
  ["zeitschrift fur pflanzenphysiologie"] = "Z. Pflanzenphysiol.",
  ["zeitschrift fur philosophische forschung"] = "Z Philos Forsch",
  ["zeitschrift fur physik"] = "Z. Angew. Phys.",
  ["zeitschrift fur physik a: atomes and nuclei"] = "Z. Phys. A: At. Nucl.",
  ["zeitschrift fur physik a: hadrons and nuclei"] = "Z. Phys. A: Hadrons Nucl.",
  ["zeitschrift fur physik b: condensed matter"] = "Z. Phys. B: Condens. Matter",
  ["zeitschrift fur physik b: condensed matter and quanta"] = "Z. Phys. B: Condens. Matter Quanta",
  ["zeitschrift fur physik c: particle fields"] = "Z. Phys. C: Part. Fields",
  ["zeitschrift fur physik c: particles and fields"] = "Z. Phys. C: Part. Fields",
  ["zeitschrift fur physik d: atoms, molecules and clusters"] = "Z. Phys. D: At. Mol. Clusters",
  ["zeitschrift fur physik. a, atomic nuclei"] = "Z Phys A At Nucl",
  ["zeitschrift fur physikalische chemie"] = "Z. Phys. Chem.",
  ["zeitschrift fur physikalische chemie, abteilung a"] = "Z. Phys. Chem. Abt. A:",
  ["zeitschrift fur physikalische chemie, abteilung b"] = "Z. Phys. Chem. Abt. B:",
  ["zeitschrift fur physikalische chemie, stoeechiometrie und verwandschaftslehre"] = "Z. Phys. Chem. Stoechiom. Verwandtschafts.",
  ["zeitschrift fur physiologische chemie"] = "Z. Physiol. Chem.",
  ["zeitschrift fur physiotherapie"] = "Z. Physiother.",
  ["zeitschrift fur phytotherapie"] = "Z. Phytother.",
  ["zeitschrift fur phytotherapie : offizielles organ der ges. f. phytotherapie e.v"] = "Z Phytother",
  ["zeitschrift fur plastische chirurgie"] = "Z. Plast. Chir.",
  ["zeitschrift fur politik"] = "Z Polit",
  ["zeitschrift fur praktische anasthesie und wiederbelebung"] = "Z. Prakt. Anasth.",
  ["zeitschrift fur praktische anasthesie, wiederbelebung und intensivtherapie"] = "Z. Prakt. Anasth. Wiederbeleb. Intensivther.",
  ["zeitschrift fur praventivmedizin"] = "Z. Praventivmed.",
  ["zeitschrift fur praventivmedizin. revue de medecine preventive"] = "Z Praventivmed",
  ["zeitschrift fur preussische geschichte und landeskunde"] = "Z. preuss. Gesch. Landeskd.",
  ["zeitschrift fur psycho-somatische medizin"] = "Z Psychosom Med",
  ["zeitschrift fur psychologie mit zeitschrift fur angewandte psychologie"] = "Z. Psychol. Z. Angew. Psychol.",
  ["zeitschrift fur psychosomatische medizin und psychoanalyse"] = "Z. Psychosom. Med. Psychoanal.",
  ["zeitschrift fur psychosomatische medizin und psychotherapie"] = "Z. Psychosom. Med. Psychother.",
  ["zeitschrift fur psychotherapie und medizinische psychologie"] = "Z. Psychother. Med. Psychol.",
  ["zeitschrift fur rechtsmedizin"] = "Z. Rechtsmed.",
  ["zeitschrift fur rechtsmedizin (journal of legal medicine)"] = "Z. Rechtsmed.",
  ["zeitschrift fur rechtsmedizin. journal of legal medicine"] = "Z Rechtsmed",
  ["zeitschrift fur religions- und geistesgeschichte"] = "Z Relig Geistesgesch",
  ["zeitschrift fur rheumaforschung"] = "Z. Rheumaforsch.",
  ["zeitschrift fur rheumatologie"] = "Z Rheumatol",
  ["zeitschrift fur romanische philologie"] = "Z Rom Philol",
  ["zeitschrift fur saugetierkunde"] = "Z Saugetierkd",
  ["zeitschrift fur sozialreform"] = "Z Sozialreform",
  ["zeitschrift fur soziologie"] = "Z Soziol",
  ["zeitschrift fur stomatologie"] = "Z. Stomatol.",
  ["zeitschrift fur stomatologie (1984)"] = "Z Stomatol",
  ["zeitschrift fur technische physik"] = "Z. Tech. Phys.",
  ["zeitschrift fur theoretische und angewandte genetik"] = "Z. Theor. Angew. Genet.",
  ["zeitschrift fur tierernahrung und futtermittelkunde"] = "Z. Tierernahr. Futtermittelkd.",
  ["zeitschrift fur tierphysiologie tierernahrung und futtermittelkunde"] = "Z. Tierphysiol. Tierernahr. Futtermittelkd.",
  ["zeitschrift fur tierphysiologie, tierernahrung und futtermittelkunde"] = "Z. Tierphysiol. Tierernahr. Futtermittelkd.",
  ["zeitschrift fur tierpsychologie"] = "Z. Tierpsychol.",
  ["zeitschrift fur tierzuchtung und zuchtungsbiologie"] = "Z. Tierz. Zuchtungsbiol.",
  ["zeitschrift fur tierzuchtung und zuchtungsbiologie : organ der reichsarbeitsgemeinschaft tierzucht im forschungsdienst"] = "Z Tierzuecht Zuechtungsbiol",
  ["zeitschrift fur tropenmedizin und parasitologie"] = "Z. Tropenmed. Parasitol.",
  ["zeitschrift fur tuberkulose"] = "Z Tuberk",
  ["zeitschrift fur tuberkulose und erkrankungen der thoraxorgane"] = "Z. Tuberk. Erkr. Thoraxorg.",
  ["zeitschrift fur unfallchirurgie und versicherungsmedizin"] = "Z. Unfallchir. Versicherungsmed.",
  ["zeitschrift fur unfallchirurgie, versicherungsmedizin und berufskrankheiten"] = "Z. Unfallchir. Versicherungsmed. Berufskr.",
  ["zeitschrift fur unfallmedizin und berufskrankheiten"] = "Z. Unfallmed. Berufskr.",
  ["zeitschrift fur unfallmedizin und berufskrankheiten. revue de medecine des accidents et des maladies professionelles"] = "Z Unfallmed Berufskr",
  ["zeitschrift fur urologie"] = "Z Urol",
  ["zeitschrift fur urologie und nephrologie"] = "Z. Urol. Nephrol.",
  ["zeitschrift fur vererbungslehre"] = "Z. Vererbungsl.",
  ["zeitschrift fur vergleichende physiologie"] = "Z. Vgl. Physiol.",
  ["zeitschrift fur vermessungswesen"] = "Z. Vermessungswes.",
  ["zeitschrift fur versuchstierkunde"] = "Z. Versuchstierkd.",
  ["zeitschrift fur vitamin- hormon- und fermentforschung"] = "Z. Vitam. Horm. Fermentforsch.",
  ["zeitschrift fur vitamin-, hormon- und fermentforschung"] = "Z Vitam Horm Fermentforsch",
  ["zeitschrift fur wasser- und abwasserforschung"] = "Z. Wasser Abwasser-Forsch.",
  ["zeitschrift fur wirtschafts- und sozialwissenschaften"] = "Z Wirtsch Sozialwiss",
  ["zeitschrift fur wirtschaftsgeographie"] = "Z Wirtschgeogr",
  ["zeitschrift fur wissenschaftliche mikroskopie und fur mikroskopische technik"] = "Z. Wiss. Mikrosk. Mikrosk. Tech.",
  ["zeitschrift fur wissenschaftliche mikroskopie und mikroskopische technik"] = "Z Wiss Mikrosk",
  ["zeitschrift fur wissenschaftliche photographie photophysik und photochemie"] = "Z. Wiss. Photogr. Photophys. Photochem.",
  ["zeitschrift fur wissenschartliche zoologie"] = "Z Wiss Zool",
  ["zeitschrift fur wurttembergische landesgeschichte"] = "Z Wurttemb Landesgesch",
  ["zeitschrift fur zellforschung und mikroskopische anatomie"] = "Z. Zellforsch. Mikrosk. Anat.",
  ["zeitschrift fur zellforschung und mikroskopische anatomie (vienna, austria : 1948)"] = "Z Zellforsch Mikrosk Anat",
  ["zeitschrift fur zellforschung und mikroskopische anatomie. abteilung histochemie"] = "Z Zellforch Microsk Anat Histochem",
  ["zeitschrift fur zuchtung reihe b"] = "Z. Zucht. Reihe B",
  ["zeitschrift fũr erkrankungen der atmungsorgane mit folia bronchologica"] = "Z Erkr Atmungsorgane Folia Bronchol",
  ["zeitschrift für agrargeschichte und agrarsoziologie"] = "Z Agrarges Agrarsoziol",
  ["zeitschrift für allgemeine mikrobiologie"] = "Z Allg Mikrobiol",
  ["zeitschrift für allgemeine wissenschaftstheorie. journal for general philosophy of science"] = "Z Allg Wissenschaftstheor",
  ["zeitschrift für allgemeinmedizin"] = "Z Allgemeinmed",
  ["zeitschrift für alternsforschung"] = "Z Alternsforsch",
  ["zeitschrift für anatomie und entwicklungsgeschichte"] = "Z Anat Entwicklungsgesch",
  ["zeitschrift für angewandte bäder- und klimaheilkunde"] = "Z Angew Bader Klimaheilkd",
  ["zeitschrift für angewandte ichthyologie = journal of applied ichthyology"] = "J Appl Ichthyol",
  ["zeitschrift für angewandte mathematik und mechanik"] = "Z Angew Math Mech",
  ["zeitschrift für anorganische und allgemeine chemie"] = "Z Anorg Allg Chem",
  ["zeitschrift für arbeitswissenschaft"] = "Z Arbeitswiss",
  ["zeitschrift für ägyptische sprache und altertumskunde"] = "Z Aegypt Sprach Altertumskd",
  ["zeitschrift für ärztliche fortbildung"] = "Z Arztl Fortbild (Jena)",
  ["zeitschrift für ärztliche fortbildung (berlin)"] = "Z Arztl Fortbild (Berl)",
  ["zeitschrift für ärztliche fortbildung und qualitätssicherung"] = "Z Arztl Fortbild Qualitatssich",
  ["zeitschrift für ärztliche fortbildung. beiheft"] = "Z Arztl Fortbild Beih (Jena)",
  ["zeitschrift für bayerische landesgeschichte"] = "Z Bayer Landesgesch",
  ["zeitschrift für bevölkerungswissenschaft"] = "Z Bevolkerungswiss",
  ["zeitschrift für bibliothekswesen und bibliographie"] = "Z Bibliothekswes Bibliogr",
  ["zeitschrift für biologie"] = "Z Biol",
  ["zeitschrift für den erdkundeunterricht"] = "Z Erdkundeunterr",
  ["zeitschrift für deutsches alterthum und deutsche litteratur"] = "Z Dtsch Altert Dtsch Lit",
  ["zeitschrift für die gesamte experimentelle medizin"] = "Z Gesamte Exp Med",
  ["zeitschrift für die gesamte experimentelle medizin einschliesslich experimentelle chirurgie"] = "Z Gesamte Exp Med",
  ["zeitschrift für die gesamte hygiene und ihre grenzgebiete"] = "Z Gesamte Hyg",
  ["zeitschrift für die gesamte innere medizin und ihre grenzgebiete"] = "Z Gesamte Inn Med",
  ["zeitschrift für die gesamte staatswissenschaft = journal of institutional and theoretical economics"] = "Z Gesamte Staatswiss",
  ["zeitschrift für entwicklungspsychologie und pädagogische psychologie"] = "Z Entwicklungspsychol Padagog Psychol",
  ["zeitschrift für erkrankungen der atmungsorgane"] = "Z Erkr Atmungsorgane",
  ["zeitschrift für ernährungswissenschaft"] = "Z Ernahrungswiss",
  ["zeitschrift für ernährungswissenschaft. journal of nutritional sciences. supplementa"] = "Z Ernahrungswiss Suppl",
  ["zeitschrift für erziehungswissenschaft : zfe"] = "Z Erziehwiss",
  ["zeitschrift für ethnologie"] = "Z Ethnol",
  ["zeitschrift für evidenz, fortbildung und qualität im gesundheitswesen"] = "Z Evid Fortbild Qual Gesundhwes",
  ["zeitschrift für experimentelle chirurgie"] = "Z Exp Chir",
  ["zeitschrift für experimentelle chirurgie, transplantation, und künstliche organe : organ der sektion experimentelle chirurgie der gesellschaft für chirurgie der ddr"] = "Z Exp Chir Transplant Kunstliche Organe",
  ["zeitschrift für experimentelle psychologie : organ der deutschen gesellschaft für psychologie"] = "Z Exp Psychol",
  ["zeitschrift für experimentelle und angewandte psychologie"] = "Z Exp Angew Psychol",
  ["zeitschrift für flugwissenschaften und weltraumforschung"] = "Z Flugwiss Weltraumforsch",
  ["zeitschrift für gastroenterologie"] = "Z Gastroenterol",
  ["zeitschrift für gastroenterologie. verhandlungsband"] = "Z Gastroenterol Verh",
  ["zeitschrift für geburtshilfe und gynäkologie"] = "Z Geburtshilfe Gynakol",
  ["zeitschrift für geburtshilfe und neonatologie"] = "Z Geburtshilfe Neonatol",
  ["zeitschrift für geburtshilfe und perinatologie"] = "Z Geburtshilfe Perinatol",
  ["zeitschrift für gerontologie"] = "Z Gerontol",
  ["zeitschrift für gerontologie und geriatrie"] = "Z Gerontol Geriatr",
  ["zeitschrift für geschichte der arabisch-islamischen wissenschaften"] = "Z Gesch Arab Islam Wiss",
  ["zeitschrift für geschichtswissenschaft"] = "Z Geschichtswiss",
  ["zeitschrift für gesundheitswissenschaften = journal of public health"] = "Z Gesundh Wiss",
  ["zeitschrift für haut- und geschlechtskrankheiten"] = "Z Haut Geschlechtskr",
  ["zeitschrift für hautkrankheiten"] = "Z Hautkr",
  ["zeitschrift für heereskunde"] = "Z Heereskd",
  ["zeitschrift für herz-, thorax- und gefässchirurgie"] = "Z Herz Thorax Gefasschir",
  ["zeitschrift für historische forschung"] = "Z Hist Forsch",
  ["zeitschrift für hygiene und infektionskrankheiten; medizinische mikrobiologie, immunologie und virologie"] = "Z Hyg Infektionskr",
  ["zeitschrift für immunitats- und allergieforschung"] = "Z Immunitats Allergieforsch",
  ["zeitschrift für immunitätsforschung und experimentelle therapie"] = "Z Immun exp ther",
  ["zeitschrift für immunitätsforschung, allergie und klinische immunologie"] = "Z Immunitatsforsch Allerg Klin Immunol",
  ["zeitschrift für immunitätsforschung, experimentelle und klinische immunologie"] = "Z Immunitatsforsch Exp Klin Immunol",
  ["zeitschrift für immunitätsforschung. immunobiology"] = "Z Immunitatsforsch Immunobiol",
  ["zeitschrift für immunitätsforschung. immunobiology. supplemente"] = "Z Immunitatsforsch Immunobiol Suppl",
  ["zeitschrift für induktive abstammungs- und vererbungslehre"] = "Z Indukt Abstamm Vererbungsl",
  ["zeitschrift für jagdwissenschaft"] = "Z Jagdwiss",
  ["zeitschrift für kardiologie"] = "Z Kardiol",
  ["zeitschrift für kardiologie. supplementum"] = "Z Kardiol Suppl",
  ["zeitschrift für kinder- und jugendpsychiatrie"] = "Z Kinder Jugendpsychiatr",
  ["zeitschrift für kinder- und jugendpsychiatrie und psychotherapie"] = "Z Kinder Jugendpsychiatr Psychother",
  ["zeitschrift für kinderchirurgie : organ der deutschen, der schweizerischen und der osterreichischen gesellschaft für kinderchirurgie = surgery in infancy and childhood"] = "Z Kinderchir",
  ["zeitschrift für kinderchirurgie und grenzgebiete"] = "Z Kinderchir Grenzgeb",
  ["zeitschrift für kinderheilkunde"] = "Z Kinderheilkd",
  ["zeitschrift für kinderpsychiatrie. revue de psychiatrie infantile"] = "Z Kinderpsychiatr",
  ["zeitschrift für kirchengeschichte"] = "Z Kirchengesch",
  ["zeitschrift für klinische chemie und klinische biochemie"] = "Z Klin Chem Klin Biochem",
  ["zeitschrift für klinische medizin"] = "Z Klin Med",
  ["zeitschrift für klinische medizin (berlin, germany : 1985)"] = "Z Klin Med",
  ["zeitschrift für klinische psychologie und psychotherapie"] = "Z Klin Psychol Psychother",
  ["zeitschrift für klinische psychologie und psychotherapie (göttingen, germany)"] = "Z Klin Psychol Psychother (Gott)",
  ["zeitschrift für klinische psychologie, psychiatrie und psychotherapie"] = "Z Klin Psychol Psychiatr Psychother",
  ["zeitschrift für klinische psychologie, psychopathologie und psychotherapie"] = "Z Klin Psychol Psychopathol Psychother",
  ["zeitschrift für krankenpflege. revue suisse des infirmières"] = "Z Krankenpfl",
  ["zeitschrift für krebsforschung"] = "Z Krebsforsch",
  ["zeitschrift für krebsforschung und klinische onkologie. cancer research and clinical oncology"] = "Z Krebsforsch Klin Onkol Cancer Res Clin Oncol",
  ["zeitschrift für kreislaufforschung"] = "Z Kreislaufforsch",
  ["zeitschrift für kristallographie. crystalline materials"] = "Z Kristallogr Cryst Mater",
  ["zeitschrift für kunstgeschichte"] = "Z Kunstgesch",
  ["zeitschrift für laryngologie, rhinologie, otologie und ihre grenzgebiete"] = "Z Laryngol Rhinol Otol",
  ["zeitschrift für lebensmittel-untersuchung und -forschung"] = "Z Lebensm Unters Forsch",
  ["zeitschrift für lebensmittel-untersuchung und -forschung. a, food research and technology"] = "Z Lebensm Unters Forsch A Food Res Technol",
  ["zeitschrift für lymphologie. journal of lymphology"] = "Z Lymphol",
  ["zeitschrift für medizinische ethik : wissenschaft, kultur, religion"] = "Z Med Ethik",
  ["zeitschrift für medizinische isotopenforschung und deren grenzgebiete"] = "Z Med Isotopenforsch Deren Grenzgeb",
  ["zeitschrift für medizinische laboratoriumsdiagnostik"] = "Z Med Lab Diagn",
  ["zeitschrift für medizinische labortechnik"] = "Z Med Labortech",
  ["zeitschrift für medizinische mikrobiologie und immunologie"] = "Z Med Mikrobiol Immunol",
  ["zeitschrift für medizinische physik"] = "Z Med Phys",
  ["zeitschrift für medizinische psychologie : zmp"] = "Z Med Psychol",
  ["zeitschrift für menschliche vererbungs- und konstitutionslehre"] = "Z Mensch Vererb Konstitutionsl",
  ["zeitschrift für mikroskopisch-anatomische forschung"] = "Z Mikrosk Anat Forsch",
  ["zeitschrift für morphologie und anthropologie"] = "Z Morphol Anthropol",
  ["zeitschrift für morphologie und oekologie der tiere"] = "Z Morphol Okol Tiere",
  ["zeitschrift für mykologie"] = "Z Mykol",
  ["zeitschrift für naturforschung. a, journal of physical sciences"] = "Z Nat Forsch A J Phys Sci",
  ["zeitschrift für naturforschung. b, journal of chemical sciences"] = "Z Naturforsch B J Chem Sci",
  ["zeitschrift für naturforschung. c, journal of biosciences"] = "Z Naturforsch C J Biosci",
  ["zeitschrift für naturforschung. section c, biosciences"] = "Z Naturforsch C Biosci",
  ["zeitschrift für naturforschung. teil b, chemie, biochemie, biophysik, biologie und verwandte gebiete"] = "Z Naturforsch B",
  ["zeitschrift für naturforschung. teil b. anorganische chemie, organische chemie, biochemie, biophysik, biologie"] = "Z Naturforsch B",
  ["zeitschrift für naturforschung. teil c: biochemie, biophysik, biologie, virologie"] = "Z Naturforsch C",
  ["zeitschrift für naturwissenschaftlichmedizinische grundlagenforschung"] = "Z Naturwiss Med Grundlagenforsch",
  ["zeitschrift für neurologie"] = "Z Neurol",
  ["zeitschrift für orthopädie und ihre grenzgebiete"] = "Z Orthop Ihre Grenzgeb",
  ["zeitschrift für orthopädie und unfallchirurgie"] = "Z Orthop Unfall",
  ["zeitschrift für ostforschung"] = "Z Ostforsch",
  ["zeitschrift für papyrologie und epigraphik"] = "Z Papyrologie Epigraphik",
  ["zeitschrift für parasitenkunde (berlin, germany)"] = "Z Parasitenkd",
  ["zeitschrift für pflanzenkrankheiten und pflanzenschutz"] = "Z Pflanzenkr Pflanzenschutz",
  ["zeitschrift für pflanzenphysiologie. international journal of plant physiology"] = "Z Pflanzenphysiol",
  ["zeitschrift für philosophische forschung"] = "Z Philos Forsch",
  ["zeitschrift für physik. a, atomic nuclei"] = "Z Phys A At Nucl",
  ["zeitschrift für physikalische chemie (frankfurt am main, germany)"] = "Z Phys Chem (N F)",
  ["zeitschrift für physikalische therapie, bäder- und klimaheilkunde"] = "Z Phys Ther Bader Klimanheikd",
  ["zeitschrift für physiotherapie"] = "Z Physiother",
  ["zeitschrift für phytotherapie : offizielles organ der ges. f. phytotherapie e.v"] = "Z Phytother",
  ["zeitschrift für plastische chirurgie"] = "Z Plast Chir",
  ["zeitschrift für politik"] = "Z Polit",
  ["zeitschrift für praktische anästhesie und wiederbelebung"] = "Z Prakt Anasth",
  ["zeitschrift für praktische anästhesie, wiederbelebung und intensivtherapie"] = "Z Prakt Anasth Wiederbeleb Intensivther",
  ["zeitschrift für präventivmedizin. revue de médecine préventive"] = "Z Praventivmed",
  ["zeitschrift für psycho-somatische medizin"] = "Z Psychosom Med",
  ["zeitschrift für psychologie"] = "Z Psychol",
  ["zeitschrift für psychologie mit zeitschrift für angewandte psychologie"] = "Z Psychol Z Angew Psychol",
  ["zeitschrift für psychosomatische medizin und psychoanalyse"] = "Z Psychosom Med Psychoanal",
  ["zeitschrift für psychosomatische medizin und psychotherapie"] = "Z Psychosom Med Psychother",
  ["zeitschrift für psychotherapie und medizinische psychologie"] = "Z Psychother Med Psychol",
  ["zeitschrift für rechtsmedizin. journal of legal medicine"] = "Z Rechtsmed",
  ["zeitschrift für religions- und geistesgeschichte"] = "Z Relig Geistesgesch",
  ["zeitschrift für rheumaforschung"] = "Z Rheumaforsch",
  ["zeitschrift für rheumatologie"] = "Z Rheumatol",
  ["zeitschrift für romanische philologie"] = "Z Rom Philol",
  ["zeitschrift für säugetierkunde"] = "Z Saugetierkd",
  ["zeitschrift für sexualforschung (stuttgart, germany)"] = "Z Sex Forsch",
  ["zeitschrift für sozialreform"] = "Z Sozialreform",
  ["zeitschrift für soziologie"] = "Z Soziol",
  ["zeitschrift für sportpsychologie"] = "Z Sportpsychol",
  ["zeitschrift für stomatologie"] = "Z Stomatol (1921)",
  ["zeitschrift für stomatologie (1984)"] = "Z Stomatol",
  ["zeitschrift für tierphysiologie, tierernährung und futtermittelkunde"] = "Z Tierphysiol Tierernahr Futtermittelkd",
  ["zeitschrift für tierpsychologie"] = "Z Tierpsychol",
  ["zeitschrift für tierzüchtung und züchtungsbiologie : organ der reichsarbeitsgemeinschaft tierzucht im forschungsdienst"] = "Z Tierzuecht Zuechtungsbiol",
  ["zeitschrift für tropenmedizin und parasitologie"] = "Z Tropenmed Parasitol",
  ["zeitschrift für tuberkulose"] = "Z Tuberk",
  ["zeitschrift für tuberkulose und erkrankungen der thoraxorgane"] = "Z Tuberk Erkr Thoraxorg",
  ["zeitschrift für unfallchirurgie und versicherungsmedizin : offizielles organ der schweizerischen gesellschaft für unfallmedizin und berufskrankheiten = revue de traumatologie et d'assicurologie : organe officiel de la société suisse de médecine des accidents et des maladies professionnelles"] = "Z Unfallchir Versicherungsmed",
  ["zeitschrift für unfallchirurgie, versicherungsmedizin und berufskrankheiten : offizielles organ der schweizerischen gesellschaft für unfallmedizin und berufskrankheiten = revue de traumatologie, d'assicurologie et des maladies professionnelles : organe officiel de la société suisse de médecine des accidents et des maladies professionnelles"] = "Z Unfallchir Versicherungsmed Berufskr",
  ["zeitschrift für unfallmedizin und berufskrankheiten. revue de médecine des accidents et des maladies professionelles"] = "Z Unfallmed Berufskr",
  ["zeitschrift für urologie"] = "Z Urol",
  ["zeitschrift für urologie und nephrologie"] = "Z Urol Nephrol",
  ["zeitschrift für vererbungslehre"] = "Z Vererbungsl",
  ["zeitschrift für vergleichende physiologie"] = "Z Vgl Physiol",
  ["zeitschrift für versuchstierkunde"] = "Z Versuchstierkd",
  ["zeitschrift für vitamin-, hormon- und fermentforschung"] = "Z Vitam Horm Fermentforsch",
  ["zeitschrift für vitaminforschung"] = "Z Vitam Forsch",
  ["zeitschrift für wirtschafts- und sozialwissenschaften"] = "Z Wirtsch Sozialwiss",
  ["zeitschrift für wirtschaftsgeographie"] = "Z Wirtschgeogr",
  ["zeitschrift für wissenschaftliche mikroskopie und mikroskopische technik"] = "Z Wiss Mikrosk",
  ["zeitschrift für wissenschartliche zoologie"] = "Z Wiss Zool",
  ["zeitschrift für württembergische landesgeschichte"] = "Z Wurttemb Landesgesch",
  ["zeitschrift für zellforschung und mikroskopische anatomie (vienna, austria : 1948)"] = "Z Zellforsch Mikrosk Anat",
  ["zeitschrift für zellforschung und mikroskopische anatomie. abteilung histochemie"] = "Z Zellforch Microsk Anat Histochem",
  ["zeitschrift für agrargeschichte und agrarsoziologie"] = "Z. Agrargesch. Agrarsoziol.",
  ["zeitschrift für allgemaine und textile marktwirts schaft"] = "Z. Allg. Textile Marktwirtsch.",
  ["zeitschrift für analysis und ihre anwendungen"] = "Z. Anal. Anwendungen",
  ["zeitschrift für analytische chemie"] = "Z. Anal. Chem.",
  ["zeitschrift für angewandte entomologie"] = "Z. Angew. Entomol.",
  ["zeitschrift für angewandte mathematik und mechanik"] = "Z. Angew. Math. Mech.",
  ["zeitschrift für angewandte mathematik und physik"] = "Z. Angew. Math. Phys.",
  ["zeitschrift für angewandte physik"] = "Z. Angew. Phys.",
  ["zeitschrift für angewandte physik und chemie"] = "Z. Elektrochem. Angew. Phys. Chem.",
  ["zeitschrift für anorganische chemie"] = "Z. Anorg. Chem.",
  ["zeitschrift für anorganische und allgemeine chemie"] = "Z. Anorg. Allg. Chem.",
  ["zeitschrift für antikes christentum"] = "ZAC",
  ["zeitschrift für archäologie"] = "ZfA",
  ["zeitschrift für archäologie aussereuropäischer kulturen"] = "ZAAK",
  ["zeitschrift für assyriologie und vorderasiatische archäologie"] = "ZA",
  ["zeitschrift für astrophysik"] = "Z. Astrophys.",
  ["zeitschrift für betriebswirtschaft"] = "Z. Betriebswirtschaft",
  ["zeitschrift für das forst- und jagdwesen"] = "Z. Forst- Jagdwes.",
  ["zeitschrift für die alttestamentliche wissenschaft"] = "ZATW",
  ["zeitschrift für die gesamte staatswissenschaft (jite)"] = "Z. ges. Staatswissen. (JITE)",
  ["zeitschrift für die neutestamentliche wissenschaft und die kunde der älteren kirche"] = "ZNTW",
  ["zeitschrift für elektrochemie"] = "Z. Elektrochem.",
  ["zeitschrift für ethnologie der deutschen gesellschaft für völkerkunde und der berliner gesellschaft für anthropologie, ethnologie und urgeschichte"] = "ZEthn",
  ["zeitschrift für forstgenetik und forstpflanzenzüchtung"] = "Z. Forstgenet. Forstpflanzenzücht.",
  ["zeitschrift für geologische wissenschaften"] = "Z. Geol. Wiss.",
  ["zeitschrift für geomorphologie"] = "Z. Geomorphol.",
  ["zeitschrift für geschichte der arabisch-islamischen wissenschaften"] = "Z. Gesch. Arab.-Islam. Wiss.",
  ["zeitschrift für gletscherkunde und glazialgeologie"] = "Z. Gletsch.kd. Glazialgeol.",
  ["zeitschrift für jagdwissenschaft"] = "Z. Jagdwiss.",
  ["zeitschrift für katholische theologie"] = "ZKTh",
  ["zeitschrift für kirchengeschichte"] = "ZKG",
  ["zeitschrift für kristallographie"] = "Z. Krist.",
  ["zeitschrift für kristallographie - new crystal structures"] = "Z. Kristallogr. - New Cryst. Struct.",
  ["zeitschrift für kristallographie, kristallgeometrie, kristallphysik, kristallchemie"] = "Z. Kristallogr. Kristallgeom. Kristallphys. Kristallchem.",
  ["zeitschrift für kulturtechnik und flurbereinigung"] = "Z. Kult.tech. Flurbereinig.",
  ["zeitschrift für kunstgeschichte"] = "ZKuGesch",
  ["zeitschrift für lebensmittel-untersuchung und -forschung"] = "Z. Lebensm.-Unters. Forsch.",
  ["zeitschrift für metallkunde"] = "Z. Metallkd.",
  ["zeitschrift für meteorologie"] = "Z. Meteor.",
  ["zeitschrift für mykologie"] = "Z. Mykol.",
  ["zeitschrift für nationalökonomie"] = "Z. Nationalökon.",
  ["zeitschrift für naturforschung"] = "Z. Naturforsch.",
  ["zeitschrift für naturforschung, a: physical sciences"] = "Z. Naturforsch., A: Phys. Sci.",
  ["zeitschrift für naturforschung, b: chemical sciences"] = "Z. Naturforsch., B: Chem. Sci.",
  ["zeitschrift für naturforschung, c: biosciences"] = "Z. Naturforsch., C: Biosci.",
  ["zeitschrift für naturforschung, c: journal of biosciences"] = "Z. Naturforsch., C: J. Biosci.",
  ["zeitschrift für naturforschung, teil a: physik, physikalische chemie, kosmophysik"] = "Z. Naturforsch. Teil A",
  ["zeitschrift für numismatik"] = "ZfNum",
  ["zeitschrift für papyrologie und epigraphik"] = "ZPE",
  ["zeitschrift für pflanzenernährung und bodenkunde"] = "Z. Pflanzenernähr. Bodenkd.",
  ["zeitschrift für pflanzenkrankheiten und pflanzenschutz"] = "Z. Pflanzenkrankh. Pflanzenschutz",
  ["zeitschrift für philosophische forschung"] = "ZPhF",
  ["zeitschrift für photogrammetrie und fernerkundung"] = "Z. Photogramm. Fernerkund.",
  ["zeitschrift für physik"] = "Z. Phys.",
  ["zeitschrift für physik a"] = "Z. Phys. A",
  ["zeitschrift für physik a hadrons and nuclei"] = "Z. Phys. A Hadrons Nucl.",
  ["zeitschrift für physik a: atoms and nuclei"] = "Z. Phys. A: At. Nucl.",
  ["zeitschrift für physik a: hadrons and nuclei"] = "Z. Phys. A: Hadrons Nucl.",
  ["zeitschrift für physik b"] = "Z. Phys. B",
  ["zeitschrift für physik b condensed matter"] = "Z. Phys. B Condens. Matter",
  ["zeitschrift für physik b condensed matter and quanta"] = "Z. Phys. B Condens. Matter Quanta",
  ["zeitschrift für physik b: condensed matter"] = "Z. Phys. B: Condens. Matter",
  ["zeitschrift für physik b: condensed matter and quanta"] = "Z. Phys. B: Condens. Matter Quanta",
  ["zeitschrift für physik c"] = "Z. Phys. C",
  ["zeitschrift für physik c: particle fields"] = "Z. Phys. C: Part. Fields",
  ["zeitschrift für physik c: particles and fields"] = "Z. Phys. C: Part. Fields",
  ["zeitschrift für physik d"] = "Z. Phys. D",
  ["zeitschrift für physik d: atoms, molecules and clusters"] = "Z. Phys. D: At. Mol. Clust.",
  ["zeitschrift für physikalische chemie (frankfurt am main)"] = "Z. Phys. Chem. (Frankurt am Main)",
  ["zeitschrift für physikalische chemie (leipzig)"] = "Z. Phys. Chem. (Leipzig)",
  ["zeitschrift für physikalische chemie (muenchen, germany)"] = "Z. Phys. Chem. (Muenchen, Ger.)",
  ["zeitschrift für physikalische chemie, abteilung a"] = "Z. Phys. Chem. Abt. A:",
  ["zeitschrift für physikalische chemie, abteilung a: chemische thermodynamik, kinetik, elektrochemie, eigenschaftslehre"] = "Z. Phys. Chem. Abt. A",
  ["zeitschrift für physikalische chemie, abteilung b"] = "Z. Phys. Chem. Abt. B:",
  ["zeitschrift für physikalische chemie, abteilung b: chemie der elementarprozesse, aufbau der materie"] = "Z. Phys. Chem. Abt. B",
  ["zeitschrift für physikalische chemie, stoeechiometrie und verwandschaftslehre"] = "Z. Phys. Chem. Stoechiom. Verwandtschafts.",
  ["zeitschrift für physikalisch–chemische materialforschung"] = "Z. Phys. Chem. Materialforsch.",
  ["zeitschrift für physiologische chemie"] = "Z. Physiol. Chem.",
  ["zeitschrift für pilzkunde"] = "Z. Pilzkd.",
  ["zeitschrift für religions- und geistesgeschichte"] = "ZRGG",
  ["zeitschrift für romanische philologie"] = "ZRPh",
  ["zeitschrift für schweizerische archäologie und kunstgeschichte"] = "ZSchwA",
  ["zeitschrift für semiotik"] = "ZS",
  ["zeitschrift für soziologie"] = "Z. Soziol.",
  ["zeitschrift für säugetierkunde"] = "Z. Säugetierkd.",
  ["zeitschrift für tierpsychologie"] = "Z. Tierpsychol.",
  ["zeitschrift für vergleichende sprachforschung"] = "ZVS",
  ["zeitschrift für wasser- und abwasserforschung"] = "Z. Wasser Wasser-Forsch.",
  ["zeitschrift für weltforstwirtschaft"] = "Z. Weltforstwirtsch.",
  ["zeitschrift für wirtschafts- und sozialwissenschaften"] = "Z. Wirtsch. Sozialwissen.",
  ["zeitschrift für wirtschaftspolitik"] = "Z. Wirtschaftspolitik",
  ["zeitschrift für wissenschaftliche zoologie"] = "Z. wiss. Zool.",
  ["zeitschrift für zellforschung"] = "Z. Zellforsch.",
  ["zeitschrift für ägyptische sprache und altertumskunde"] = "ZAS",
  ["zeitschrift für ökologie und naturschutz"] = "Z. Ökol. Nat.schutz",
  ["zellstoff und papier"] = "Zellst. Papier",
  ["zement kalk gips"] = "Zem. Kalk Gips",
  ["zentral-zeitung fur optik und mechanik"] = "Zent.-Ztg. Opt. Mech.",
  ["zentralasiatische studien"] = "Zentralasiat Stud",
  ["zentralblatt fur allgemeine pathologie und pathologische anatomie"] = "Zentralbl. Allg. Pathol.",
  ["zentralblatt fur arbeitsmedizin und arbeitsschutz"] = "Zentralbl. Arbeitsmed.",
  ["zentralblatt fur arbeitsmedizin und arbeitsschutz. beihefte"] = "Zent Arb Arb Beih",
  ["zentralblatt fur arbeitsmedizin, arbeitsschutz und ergonomie"] = "Zentralbl. Arbeitsmed. Arbeitsschutz Ergon.",
  ["zentralblatt fur arbeitsmedizin, arbeitsschutz und prophylaxe"] = "Zentralbl. Arbeitsmed. Arbeitsschutz Prophyl.",
  ["zentralblatt fur arbeitsmedizin, arbeitsschutz, prophylaxe und ergonomie"] = "Zentralbl. Arbeitsmed. Arbeitsschutz. Prophyl. Ergonomie",
  ["zentralblatt fur bakteriologie"] = "Zentralbl Bakteriol",
  ["zentralblatt fur bakteriologie : international journal of medical microbiology"] = "Zentralbl Bakteriol",
  ["zentralblatt fur bakteriologie, mikrobiologie und hygiene. 1. abt. originale b, hygiene"] = "Zentralbl. Bakteriol. Mikrobiol. Hyg. [B]",
  ["zentralblatt fur bakteriologie, mikrobiologie und hygiene. serie b, umwelthygiene, krankenhaushygiene, arbeitshygiene, praventive medizin"] = "Zentralbl. Bakteriol. Mikrobiol. Hyg. [B]",
  ["zentralblatt fur bakteriologie, mikrobiologie, und hygiene. series a, medical microbiology, infectious diseases, virology, parasitology"] = "Zentralbl. Bakteriol. Mikrobiol. Hyg. [A]",
  ["zentralblatt fur bakteriologie, parasitenkunde, infektionskrankheiten und hygiene. 1. abt. medizinisch-hygienische bakteriologie, virusforschung und parasitologie. originale"] = "Zentralbl Bakteriol [Orig]",
  ["zentralblatt fur bakteriologie, parasitenkunde, infektionskrankheiten und hygiene. 1. abt.,"] = "Zentralbl. Bakteriol. [Orig.]",
  ["zentralblatt fur bakteriologie, parasitenkunde, infektionskrankheiten und hygiene. erste"] = "Zentralbl. Bakteriol. [Orig. B]",
  ["zentralblatt fur bakteriologie, parasitenkunde, infektionskrankheiten und hygiene. erste abteilung originale. reihe a: medizinische mikrobiologie und parasitologie"] = "Zentralbl Bakteriol [Orig A]",
  ["zentralblatt fur bakteriologie, parasitenkunde, infektionskrankheiten und hygiene. erste abteilung originale. reihe b: hygiene, betriebshygiene, praventive medizin"] = "Zentralbl Bakteriol [B]",
  ["zentralblatt fur bakteriologie, parasitenkunde, infektionskrankheiten und hygiene. erste abteilung originale. reihe b: hygiene, praventive medizin"] = "Zentralbl Bakteriol [Orig B]",
  ["zentralblatt fur bakteriologie, parasitenkunde, infektionskrankheiten und hygiene. erste abteilung, originale. reihe b, hygiene, betriebshygiene, praventive medizin"] = "Zentralbl. Bakteriol. [B]",
  ["zentralblatt fur bakteriologie, parasitenkunde, infektionskrankheiten und hygiene. zweite"] = "Zentralbl. Bakteriol. Parasitenkd. Infektionskr. Hyg.",
  ["zentralblatt fur bakteriologie, parasitenkunde, infektionskrankheiten und hygiene. zweite naturwissenschaftliche abt.: allgemeine, landwirtschaftliche und technische mikrobiologie"] = "Zentralbl Bakteriol Parasitenkd Infektionskr Hyg",
  ["zentralblatt fur bakteriologie, parasitenkunde, infektionskrankheiten und hygiene. zweite naturwissenschaftliche abteilung: mikrobiologie der landwirtschaft der technologie und des umweltschutzes"] = "Zentralbl Bakteriol Naturwiss",
  ["zentralblatt fur bakteriologie. 1. abt. originale b, hygiene, krankenhaushygiene, betriebshygiene, praventive medizin"] = "Zentralbl Bakteriol [B]",
  ["zentralblatt fur bakteriologie. 1. abt. originale. a: medizinische mikrobiologie, infektionskrankheiten und parasitologie"] = "Zentralbl Bakteriol A",
  ["zentralblatt fur bakteriologie. 1. abt., originale. a, medizinische mikrobiologie, infektionskrankheiten und parasitiologie"] = "Zentralbl. Bakteriol. A.",
  ["zentralblatt fur bakteriologie. 1. abt., originale. b, hygiene, krankenhaushygiene, betriebshygiene, praventive medizin"] = "Zentralbl. Bakteriol. [B]",
  ["zentralblatt fur bauverwaltung"] = "Zentralbl. Bauverwaltung",
  ["zentralblatt fur biochemie und biophysik"] = "Zentralbl. Biochem. Biophys.",
  ["zentralblatt fur biologische aerosolforschung"] = "Zentralbl. Biol. Aerosolforsch.",
  ["zentralblatt fur chirurgie"] = "Zentralbl Chir",
  ["zentralblatt fur geologie und palaeontologie teil 1: allgemeine, angewandte, regionale, und historische geologie"] = "Zentralbl. Geol. Palaeontol. Teil 1",
  ["zentralblatt fur geologie und palaeontologie teil 2: palaeontologie"] = "Zentralbl. Geol. Palaeontol. Teil 2",
  ["zentralblatt fur gynakologie"] = "Zentralbl Gynakol",
  ["zentralblatt fur haut- und geschlechtskrankheiten, sowie deren grenzgebiete"] = "Zentralbl Haut Geschlechtskr Grenzgeb",
  ["zentralblatt fur hygiene und umweltmedizin"] = "Zentralbl. Hyg. Umweltmed.",
  ["zentralblatt fur mikrobiologie"] = "Zentralbl. Mikrobiol.",
  ["zentralblatt fur mineralogie teil 1: kristallographie und mineralogie"] = "Zentralbl. Mineral. Teil 1",
  ["zentralblatt fur mineralogie teil 2: petrographie, technische mineralogie, geochemie und lagerstattenkunde"] = "Zentralbl. Mineral. Teil 2",
  ["zentralblatt fur mineralogie, geologie und palaeontologie"] = "Zentralbl. Mineral. Geol. Palaeontol.",
  ["zentralblatt fur neurochirurgie"] = "Zentralbl Neurochir",
  ["zentralblatt fur pathologie"] = "Zentralbl. Pathol.",
  ["zentralblatt fur pharmazie"] = "Zentralbl. Pharm.",
  ["zentralblatt fur pharmazie, pharmakotherapie und laboratoriumsdiagnostik"] = "Zentralbl. Pharm. Pharmakother. Laboratoriumsdiagn.",
  ["zentralblatt fur phlebologie"] = "Zentralbl. Phlebol.",
  ["zentralblatt fur verkehrs-medizin, verkehrs-psychologie, luft- und raumfahrt-medizin"] = "Zentralbl Verkehrsmed Verkehrspsychol Luft Raumfahrtmed",
  ["zentralblatt fur veterinarmedizin"] = "Zentralbl. Veterinarmed.",
  ["zentralblatt fur veterinarmedizin. reihe a"] = "Zentralbl Veterinarmed A",
  ["zentralblatt fur veterinarmedizin. reihe b"] = "Zentralbl. Veterinarmed. [B]",
  ["zentralblatt fur veterinarmedizin. reihe b. journal of veterinary medicine. series b"] = "Zentralbl Veterinarmed B",
  ["zentralblatt fur veterinarmedizin. reihe c, anatomia, histologia, embryologia"] = "Zentralbl. Veterinarmed. [C]",
  ["zentralblatt fur veterinarmedizin. reihe c: anatomie, histologie, embryologie"] = "Zentralbl Veterinarmed [C]",
  ["zentralblatt für allgemeine pathologie u. pathologische anatomie"] = "Zentralbl Allg Pathol",
  ["zentralblatt für arbeitsmedizin und arbeitsschutz"] = "Zentralbl Arbeitsmed",
  ["zentralblatt für arbeitsmedizin und arbeitsschutz. beihefte"] = "Zent Arb Arb Beih",
  ["zentralblatt für arbeitsmedizin, arbeitsschutz und ergonomie"] = "Zentralbl Arbeitsmed Arbeitsschutz Ergon",
  ["zentralblatt für arbeitsmedizin, arbeitsschutz und prophylaxe"] = "Zentralbl Arbeitsmed Arbeitsschutz Prophyl",
  ["zentralblatt für arbeitsmedizin, arbeitsschutz, prophylaxe und ergonomie"] = "Zentralbl Arbeitsmed Arbeitsschutz Prophyl Ergonomie",
  ["zentralblatt für arbeitswissenschaft"] = "Zent Arbeitswissenschaft",
  ["zentralblatt für bakteriologie : international journal of medical microbiology"] = "Zentralbl Bakteriol",
  ["zentralblatt für bakteriologie, mikrobiologie und hygiene. 1. abt. originale a, medizinische mikrobiologie, infektionskrankheiten und parasitologie = international journal of microbiology and hygiene. a, medical microbiology, infectiousdiseases, parasitology"] = "Zentralbl Bakteriol Mikrobiol Hyg A",
  ["zentralblatt für bakteriologie, mikrobiologie und hygiene. 1. abt. originale b, hygiene"] = "Zentralbl Bakteriol Mikrobiol Hyg B",
  ["zentralblatt für bakteriologie, mikrobiologie und hygiene. serie b, umwelthygiene, krankenhaushygiene, arbeitshygiene, präventive medizin"] = "Zentralbl Bakteriol Mikrobiol Hyg B",
  ["zentralblatt für bakteriologie, mikrobiologie, und hygiene. series a, medical microbiology, infectious diseases, virology, parasitology"] = "Zentralbl Bakteriol Mikrobiol Hyg A",
  ["zentralblatt für bakteriologie, parasitenkunde, infektionskrankheiten und hygiene. 1. abt. medizinisch-hygienische bakteriologie, virusforschung und parasitologie. originale"] = "Zentralbl Bakteriol Orig",
  ["zentralblatt für bakteriologie, parasitenkunde, infektionskrankheiten und hygiene. 1. abt. medizinisch-hygienische bakteriologie, virusforschung und parasitologie. referate"] = "Zentralblatt Bakteriol Parasitenkd Infekt Hyg",
  ["zentralblatt für bakteriologie, parasitenkunde, infektionskrankheiten und hygiene. erste abteilung originale. reihe a: medizinische mikrobiologie und parasitologie"] = "Zentralbl Bakteriol Orig A",
  ["zentralblatt für bakteriologie, parasitenkunde, infektionskrankheiten und hygiene. erste abteilung originale. reihe b: hygiene, betriebshygiene, präventive medizin"] = "Zentralbl Bakteriol B",
  ["zentralblatt für bakteriologie, parasitenkunde, infektionskrankheiten und hygiene. erste abteilung originale. reihe b: hygiene, präventive medizin"] = "Zentralbl Bakteriol Orig B",
  ["zentralblatt für bakteriologie, parasitenkunde, infektionskrankheiten und hygiene. zweite naturwissenschaftliche abt.: allgemeine, landwirtschaftliche und technische mikrobiologie"] = "Zentralbl Bakteriol Parasitenkd Infektionskr Hyg",
  ["zentralblatt für bakteriologie, parasitenkunde, infektionskrankheiten und hygiene. zweite naturwissenschaftliche abteilung: mikrobiologie der landwirtschaft der technologie und des umweltschutzes"] = "Zentralbl Bakteriol Naturwiss",
  ["zentralblatt für bakteriologie. 1. abt. originale b, hygiene, krankenhaushygiene, betriebshygiene, präventive medizin"] = "Zentralbl Bakteriol B",
  ["zentralblatt für bakteriologie. 1. abt. originale. a: medizinische mikrobiologie, infektionskrankheiten und parasitologie"] = "Zentralbl Bakteriol A",
  ["zentralblatt für bibliothekswesen"] = "Zent Bibliothekswesen",
  ["zentralblatt für biologische aerosolforschung"] = "Zentralbl Biol Aerosolforsch",
  ["zentralblatt für chirurgie"] = "Zentralbl Chir",
  ["zentralblatt für die gesamte rechtsmedizin und ihre grenzgebiete"] = "Zentralbl Gesamte Rechtsmed Grenzgeb",
  ["zentralblatt für gynäkologie"] = "Zentralbl Gynakol",
  ["zentralblatt für haut- und geschlechtskrankheiten, sowie deren grenzgebiete"] = "Zentralbl Haut Geschlechtskr Grenzgeb",
  ["zentralblatt für hygiene und umweltmedizin = international journal of hygiene and environmental medicine"] = "Zentralbl Hyg Umweltmed",
  ["zentralblatt für mikrobiologie"] = "Zentralbl Mikrobiol",
  ["zentralblatt für neurochirurgie"] = "Zentralbl Neurochir",
  ["zentralblatt für pathologie"] = "Zentralbl Pathol",
  ["zentralblatt für phlebologie"] = "Zentralbl Phlebol",
  ["zentralblatt für verkehrs-medizin, verkehrs-psychologie, luft- und raumfahrt-medizin"] = "Zentralbl Verkehrsmed Verkehrspsychol Luft Raumfahrtmed",
  ["zentralblatt für veterinärmedizin. reihe a"] = "Zentralbl Veterinarmed A",
  ["zentralblatt für veterinärmedizin. reihe b. journal of veterinary medicine. series b"] = "Zentralbl Veterinarmed B",
  ["zentralblatt für veterinärmedizin. reihe c: anatomie, histologie, embryologie"] = "Zentralbl Veterinarmed C",
  ["zentralblatt für bakteriologie, supplement"] = "Zentralbl. Bakteriol., Suppl.",
  ["zentralblatt für bibliothekswesen"] = "ZBB",
  ["zentralblatt für mathematik und ihre grenzgebiete"] = "Zbl",
  ["zentralblatt haut- und geschlechtskrankheiten"] = "Zentralbl Haut Und Geschlechtskr",
  ["zentralsterilisation (wiesbaden)"] = "Zentralsterilisation (Wiesb)",
  ["zephyrus. revista de prehistoria y arqueología"] = "Zephyrus",
  ["zero population growth national reporter"] = "Zero Popul Growth Natl Rep",
  ["zero to three"] = "Zero Three",
  ["zeszyty historyczne (paris, france)"] = "Zest Hist",
  ["zeszyty naukowe geometria"] = "Zeszyty Nauk. Geom.",
  ["zeszyty naukowe politechniki rzeszowskiej"] = "Zeszyty Nauk. Politech. Rzeszowskiej Mat.",
  ["zeszyty naukowe uniwersytetu jagiellonskiego. prace geograficzne"] = "Zesz Nauk Uniw Jagiell Pr Geogr",
  ["zeszyty naukowe wyższa szkoła pedagogiczna im. powstańców śl\\polhk askich w opolu. matematyka"] = "Zeszyty Nauk. Wyż. Szkoły Ped. Powstańców Śl. Opolu Mat.",
  ["zfa - zeitschrift für allgemeinmedizin"] = "ZFA - Z. Allgemeinmed.",
  ["zfa. zeitschrift fur allgemeinmedizin"] = "ZFA (Stuttgart)",
  ["zfa. zeitschrift fur alternsforschung"] = "ZFA",
  ["zfa. zeitschrift für allgemeinmedizin"] = "ZFA (Stuttgart)",
  ["zfa; zeitschrift fur allgemeinmedizin"] = "ZFA (Stuttgart)",
  ["zfa; zeitschrift fur alternsforschung"] = "ZFA",
  ["zfsl, zeitschrift fur franzosische sprache und literatur"] = "Z Fr Spr Lit",
  ["zfsl, zeitschrift für französische sprache und literatur"] = "Z Fr Spr Lit",
  ["zfw—zeitschrift für flugwissenschaften und weltraumforschung"] = "Z. Flugwiss. Weltraumforsch.",
  ["zhangzhou shiyuan xuebao (ziran kexue ban)"] = "Zhangzhou Shiyuan Xuebao (Ziran Kexue Ban)",
  ["zhejiang da xue xue bao. li xue ban = journal of zhejiang university. sciences edition"] = "Zhejiang Da Xue Xue Bao Li Xue Ban",
  ["zhejiang da xue xue bao. yi xue ban = journal of zhejiang university. medical sciences"] = "Zhejiang Da Xue Xue Bao Yi Xue Ban",
  ["zhejiang daxue xue bao. nong ye yu sheng ming ke xue ban = journal of zhejiang university. agriculture and life sciences"] = "Zhejiang Daxue Xue Bao Nong Ye Yu Sheng Ming Ke Xue Ban",
  ["zhejiang daxue xuebao"] = "Zhejiang Daxue Xuebao Ziran Kexue Ban",
  ["zhejiang haiyang xueyuan xuebao ziran kexue ban"] = "Zhejiang Haiyang Xueyuan Xuebao Ziran Kexue Ban",
  ["zhejiang nong ye da xue xue bao"] = "Acta Agriculturae Universitatis Chekianensis|Zhejiang Nong Ye Da Xue Xue Bao",
  ["zhejiang nong ye da xue xue bao = acta agriculturae universitatis chekianensis"] = "Zhejiang Nong Ye Da Xue Xue Bao",
  ["zhejiang nong ye xue bao = acta agriculturae zhejiangensis"] = "Zhejiang Nong Ye Xue Bao",
  ["zhejiang yi ke da xue xue bao = journal of zhejiang medical university"] = "Zhe Jiang Yi Ke Da Xue Xue Bao",
  ["zhejiang yu fang yi xue"] = "Zhejiang Yu Fang Yi Xue",
  ["zhen ci yan jiu = acupuncture research"] = "Zhen Ci Yan Jiu",
  ["zhengzhou da xue xue bao. yi xue ban = journal of zhengzhou university. medical sciences = zhengzhou daxue xuebao. yixueban"] = "Zhengzhou Da Xue Xue Bao Yi Xue Ban",
  ["zhi wu bao hu = plant protection"] = "Zhi Wu Bao Hu",
  ["zhi wu bao hu xue bao"] = "Acta phytophylacica sinica|Zhi Wu Bao Hu Xue Bao",
  ["zhi wu bao hu xue bao = acta phytophylacica sinica"] = "Zhi Wu Bao Hu Xue Bao",
  ["zhi wu bao hu xue hui hui kan plant protection bulletin"] = "Zhi Wu Bao Hu Xue Hui Hui Kan",
  ["zhi wu bing li xue bao"] = "Acta phytopathologica sinica|Zhi Wu Bing Li Xue Bao",
  ["zhi wu bing li xue bao = acta phytopathologica sinica"] = "Zhi Wu Bing Li Xue Bao",
  ["zhi wu bing li xue hui kan"] = "Plant pathology bulletin|Zhi Wu Bing Li Xue Hui Kan",
  ["zhi wu bing li xue hui kan = plant pathology bulletin"] = "Zhi Wu Bing Li Xue Hui Kan",
  ["zhi wu fen lei xue bao = acta phytotaxonomica sinica"] = "Zhi Wu Fen Lei Xue Bao",
  ["zhi wu jian yi"] = "Zhi Wu Jian Yi",
  ["zhi wu sheng li xue bao (2011)"] = "Zhi Wu Sheng Li Xue Bao (2011)",
  ["zhi wu sheng li xue bao = acta phytophysiologica sinica"] = "Zhi Wu Sheng Li Xue Bao",
  ["zhi wu sheng li xue tong xun"] = "Zhi Wu Sheng Li Xue Tong Xun",
  ["zhi wu sheng li yu fen zi sheng wu xue xue bao (journal of plant physiology and molecular biology)"] = "Zhi Wu Sheng Li Yu Fen Zi Sheng Wu Xue Xue Bao",
  ["zhi wu sheng li yu fen zi sheng wu xue xue bao = journal of plant physiology and molecular biology"] = "Zhi Wu Sheng Li Yu Fen Zi Sheng Wu Xue Xue Bao",
  ["zhi wu yan jiu (2001)"] = "Zhi Wu Yan Jiu",
  ["zhi wu yi xue qi kan"] = "Zhi Wu Yi Xue Qi Kan",
  ["zhinitīk dar hizārah-i sivvum"] = "Zhinitik Hizarahi Sivvum",
  ["zhinitīk-i nuvīn"] = "Zhinitik I Nuvin",
  ["zhipu xuebao"] = "Zhipu Xuebao",
  ["zhiwu xuebao = chinese bulletin of botany"] = "Zhiwu Xuebao",
  ["zhiwu yichuan ziyuan xuebao"] = "Zhiwu Yichuan Ziyuan Xuebao",
  ["zhiwu ziyuan yu huanjing xuebao"] = "Zhiwu Ziyuan Yu Huanjing Xuebao",
  ["zhiwuxue tongbao"] = "Zhiwuxue Tongbao",
  ["zhong cao yao"] = "Chinese traditional and herbal drugs|Zhong Cao Yao",
  ["zhong cao yao = chinese traditional and herbal drugs"] = "Zhong Cao Yao",
  ["zhong gong yan jiu = studies on chinese communism"] = "Zhong Gong Yan Jiu",
  ["zhong guo ke ji lun wen zai xian"] = "Zhong Guo Ke Ji Lun Wen Zai Xian",
  ["zhong liu fang zhi za zhi = china journal of cancer prevention and treatment"] = "Zhong Liu Fang Zhi Za Zhi",
  ["zhong liu za zhi"] = "Zhong Liu Za Zhi",
  ["zhong nan da xue xue bao. yi xue ban (journal of central south university. medical sciences)"] = "Zhong Nan Da Xue Xue Bao Yi Xue Ban",
  ["zhong nan da xue xue bao. yi xue ban = journal of central south university. medical sciences"] = "Zhong Nan Da Xue Xue Bao Yi Xue Ban",
  ["zhong xi yi jie he xue bao (journal of chinese integrative medicine)"] = "Zhong Xi Yi Jie He Xue Bao",
  ["zhong xi yi jie he xue bao = journal of chinese integrative medicine"] = "Zhong Xi Yi Jie He Xue Bao",
  ["zhong xi yi jie he za zhi (chinese journal of modern developments in traditional medicine)"] = "Zhong Xi Yi Jie He Za Zhi",
  ["zhong xi yi jie he za zhi = chinese journal of modern developments in traditional medicine"] = "Zhong Xi Yi Jie He Za Zhi",
  ["zhong yang yan jiu yuan jin dai shi yan jiu suo ji kan. zhong yang yan jiu yuan. jin dai shi yan jiu suo"] = "Bull Inst Mod Hist Acad Sin",
  ["zhong yao cai = zhongyaocai = journal of chinese medicinal materials"] = "Zhong Yao Cai",
  ["zhong yao tong bao (beijing, china : 1981)"] = "Zhong Yao Tong Bao",
  ["zhong yao tong bao (bulletin of chinese materia medica)"] = "Zhong Yao Tong Bao",
  ["zhongguo baozhuang ke-ji bolan"] = "Zhongguo Baozhuang Keji Bolan",
  ["zhongguo bing du bing za zhi"] = "Zhongguo Bing Du Bing Za Zhi",
  ["zhongguo bing du xue"] = "Virologica Sinica|Zhongguo Bing Du Xue",
  ["zhongguo bing du xue = virologica sinica"] = "Zhongguo Bing Du Xue",
  ["zhongguo bing yuan sheng wu xue za zhi"] = "Journal of pathogen biology|Zhongguo Bing Yuan Sheng Wu Xue Za Zhi",
  ["zhongguo bing yuan sheng wu xue za zhi = journal of pathogen biology"] = "Zhongguo Bing Yuan Sheng Wu Xue Za Zhi",
  ["zhongguo cao shi dong wu"] = "China herbivores|Zhongguo Cao Shi Dong Wu",
  ["zhongguo cao shi dong wu = china herbivores"] = "Zhongguo Cao Shi Dong Wu",
  ["zhongguo caodi xuebao = chinese journal of grassland"] = "Zhongguo Caodi Xuebao",
  ["zhongguo chao sheng yi xue za zhi = zhongguo chaosheng yixue zazhi = ultrasonic medical journal of china"] = "Zhongguo Chao Sheng Yixue Za Zhi",
  ["zhongguo dang dai er ke za zhi = chinese journal of contemporary pediatrics"] = "Zhongguo Dang Dai Er Ke Za Zhi",
  ["zhongguo di fang bing xue za zhi = zhongguo difangbingxue zazhi = chinese journal of endemiology"] = "Zhong Guo Di Fang Bing Xue Za Zhi",
  ["zhongguo dong wu chuan ran bing xue bao"] = "Chin J Anim Infect Dis",
  ["zhongguo dongwu jianyi"] = "Chinese journal of animal quaratine|Zhongguo Dongwu Jianyi",
  ["zhongguo dongwu jianyi = chinese journal of animal quaratine"] = "Zhongguo Dongwu Jianyi",
  ["zhongguo fei ai za zhi = chinese journal of lung cancer"] = "Zhongguo Fei Ai Za Zhi",
  ["zhongguo gong gong wei sheng"] = "China public health|Zhongguo Gong Gong Wei Sheng",
  ["zhongguo gong gong wei sheng = china public health"] = "Zhongguo Gong Gong Wei Sheng",
  ["zhongguo gu shang = china journal of orthopaedics and traumatology"] = "Zhongguo Gu Shang",
  ["zhongguo hai yang da xue xue bao. zi ran ke xue ban = zhongguo haiyang daxue xuebao. ziran kexue ban"] = "Zhongguo Haiyang Daxue Xuebao Ziran Kexue Ban",
  ["zhongguo hai yang yao wu = chinese journal of marine drugs"] = "Zhongguo Hai Yang Yao Wu",
  ["zhongguo hsu mu za zhi"] = "Zhongguo Hsu Mu Za Zhi",
  ["zhongguo ji hua mian yi"] = "Chinese journal of vaccines and immunization|Zhongguo Ji Hua Mian Yi",
  ["zhongguo ji sheng chong bing fang zhi za zhi = chinese journal of parasitic disease control"] = "Zhongguo Ji Sheng Chong Bing Fang Zhi Za Zhi",
  ["zhongguo ji sheng chong xue yu ji sheng chong bing za zhi (chinese journal of parasitology and parasitic diseases)"] = "Zhongguo Ji Sheng Chong Xue Yu Ji Sheng Chong Bing Za Zhi",
  ["zhongguo ji sheng chong xue yu ji sheng chong bing za zhi = chinese journal of parasitology & parasitic diseases"] = "Zhongguo Ji Sheng Chong Xue Yu Ji Sheng Chong Bing Za Zhi",
  ["zhongguo jiaqin"] = "Zhongguo Jiaqin",
  ["zhongguo jiguang"] = "Zhongguo Jiguang",
  ["zhongguo kang sheng su za zhi"] = "Chinese journal of antibiotics|Zhongguo Kang Sheng Su Za Zhi",
  ["zhongguo kang sheng su za zhi = chinese journal of antibiotics"] = "Zhongguo Kang Sheng Su Za Zhi",
  ["zhongguo kangshengsu zazhi"] = "Zhongguo Kangshengsu Zazhi",
  ["zhongguo ke ji shi liao = china historical materials of science and technology"] = "Zhongguo Ke Ji Shi Liao",
  ["zhongguo ke xue : sheng ming ke xue = scientia sinica vitae"] = "Zhongguo Ke Xue",
  ["zhongguo lin chuang yao li xue za zhi"] = "The Chinese journal of clinical pharmacology / [bian ji, Zhongguo lin chuang yao li xue za zhi bian ji bu]|Zhongguo Lin Chuang Yao Li Xue Za Zhi",
  ["zhongguo lin chuang yao li xue za zhi = the chinese journal of clinical pharmacology"] = "Zhongguo Lin Chuang Yao Li Xue Za Zhi",
  ["zhongguo mei jie sheng wu xue ji kong zhi za zhi : zhongguo yu fang yi xue hui xi lie za zhi = chinese journal of vector biology and control : serial journal of china preventive medical association = zhongguo meijieshengwuxue ji kongzhi zazhi"] = "Zhongguo Mei Jie Sheng Wu Xue Ji Kong Zhi Za Zhi",
  ["zhongguo mian yi xue za zhi"] = "Chinese journal of immunology|Zhongguo Mian Yi Xue Za Zhi",
  ["zhongguo mian yi xue za zhi = chinese journal of immunology"] = "Zhongguo Mian Yi Xue Za Zhi",
  ["zhongguo niang zao = zhongguo niangzao"] = "Zhongguo Niangzao",
  ["zhongguo nong shi"] = "Zhonguo Nongshi",
  ["zhongguo nong xue tong bao"] = 'Chinese agricultural science bulletin / "Zhongguo nong xue tong bao" bian ji bu|Zhongguo Nong Xue Tong Bao',
  ["zhongguo nong xue tong bao = chinese agricultural science bulletin"] = "Zhongguo Nong Xue Tong Bao",
  ["zhongguo nong ye da xue xue bao"] = "Journal of China Agricultural University|Zhongguo Nong Ye Da Xue Xue Bao",
  ["zhongguo nong ye da xue xue bao = journal of china agricultural university"] = "Zhongguo Nong Ye Da Xue Xue Bao",
  ["zhongguo nong ye ke xue"] = "Zhongguo nongye kexue|Zhongguo Nong Ye Ke Xue",
  ["zhongguo nong ye ke xue = zhongguo nongye kexue"] = "Zhongguo Nong Ye Ke Xue",
  ["zhongguo re dai yi xue"] = "China Trop Med",
  ["zhongguo ren kou ke xue = population science of china"] = "Zhongguo Ren Kou Ke Xue",
  ["zhongguo ren shou gong huan bing za zhi"] = "Chinese journal of zoonoses|Zhongguo Ren Shou Gong Huan Bing Za Zhi",
  ["zhongguo ren shou gong huan bing za zhi = chinese journal of zoonoses"] = "Zhongguo Ren Shou Gong Huan Bing Za Zhi",
  ["zhongguo rupin gongye = china dairy industry"] = "Zhonggue Rupin Gongye",
  ["zhongguo she hui xue kan"] = "Zhongguo She Hui Xue Kan",
  ["zhongguo shen jing ke xue za zhi"] = "Chinese journal of neuroscience|Chin. J. Neurosci.",
  ["zhongguo shen jing ke xue za zhi = chinese journal of neuroscience"] = "Chin J Neurosci",
  ["zhongguo sheng hua yao wu za zhi"] = "Zhongguo Sheng Hua Yao Wu Za Zhi",
  ["zhongguo sheng tai nong ye xue bao"] = "Zhongguo Sheng Tai Nong Ye Xue Bao",
  ["zhongguo sheng wu fang zhi"] = "Zhongguo Sheng Wu Fang Zhi",
  ["zhongguo sheng wu gong cheng za zhi"] = "Journal of Chinese biotechnology|Zhongguo Sheng Wu Gong Cheng Za Zhi",
  ["zhongguo sheng wu gong cheng za zhi = journal of chinese biotechnology"] = "Zhongguo Sheng Wu Gong Cheng Za Zhi",
  ["zhongguo sheng wu hua xue yu fen zi sheng wu xue bao"] = "Chinese journal of biochemistry and molecular biology|Zhongguo Sheng Wu Hua Xue Yu Fen Zi Sheng Wu Xue Bao",
  ["zhongguo sheng wu hua xue yu fen zi sheng wu xue bao = chinese journal of biochemistry and molecular biology"] = "Zhongguo Sheng Wu Hua Xue Yu Fen Zi Sheng Wu Xue Bao",
  ["zhongguo sheng wu zhi pin xue za zhi : zhonghua yu fang yi xue hui xi lie za zhi = zhongguo shengwu zhipin xue zazhi = chinese journal of biologicals : serial journal of china preventive medical association"] = "Zhongguo Sheng Wu Zhi Pin Xue Za Zhi",
  ["zhongguo shengwu huaxue yu fenzi shengwu xuebao"] = "Zhongguo Shengwu Huaxue Yu Fenzi Shengwu Xuebao",
  ["zhongguo shi yan dong wu xue bao = acta laboratorium animalis scientia sinica"] = "Zhongguo Shi Yan Dong Wu Xue Bao",
  ["zhongguo shi yan fang ji xue za zhi = chinese journal of experimental traditional medical formulae"] = "Zhongguo Shi Yan Fang Ji Xue Za Zhi",
  ["zhongguo shi yan jiu = zhongguoshi yanjiu"] = "Zhongguo Shi Yan Jiu",
  ["zhongguo shi yan xue ye xue za zhi"] = "Zhongguo Shi Yan Xue Ye Xue Za Zhi",
  ["zhongguo shipin xuebao = journal of chinese institute of food science and technology"] = "Zhongguo Shipin Xuebao",
  ["zhongguo shiyan zhenduanxue"] = "Zhongguo Shiyan Zhenduanxue",
  ["zhongguo shou yao za zhi"] = "Zhongguo Shou Yao Zazhi",
  ["zhongguo shou yi ke-ji = chinese journal of veterinary science and technology"] = "Zhongguo Shouyi Ke Ji",
  ["zhongguo shou yi xue bao"] = "Chinese journal of veterinary science|Chin. J. Vet. Sci.",
  ["zhongguo shou yi xue bao = chinese journal of veterinary science"] = "Chin J Vet Sci",
  ["zhongguo shou yi za zhi"] = "Chinese journal of veterinary medicine|Zhongguo Shou Yi Za Zhi",
  ["zhongguo shou yi za zhi = chinese journal of veterinary medicine"] = "Zhongguo Shou Yi Za Zhi",
  ["zhongguo shouyi ke-ji"] = "Chinese journal of veterinary science and technology|Zhongguo Shouyi Ke-ji",
  ["zhongguo shu cai"] = "Chinese vegetables|Zhongguo Shu Cai",
  ["zhongguo shu cai = chinese vegetables"] = "Zhongguo Shu Cai",
  ["zhongguo shui chan ke xue"] = "Journal of fishery sciences of China / Zhongguo shui chan ke xue yan jiu yuan bian ji chu ban|Zhongguo Shui Chan Ke Xue",
  ["zhongguo shui chan ke xue = journal of fishery sciences of china"] = "Zhongguo Shui Chan Ke Xue",
  ["zhongguo shuidao kexue"] = "Zhongguo Shuidao Kexue",
  ["zhongguo tu rang yu fei liao = soil and fertilizer sciences in china"] = "Zhongguo Tu Rang Yu Fei Liao",
  ["zhongguo wei sheng jian yan"] = "Zhongguo Wei Sheng Jian Yan",
  ["zhongguo wei sheng tai xue za zhi"] = "Zhongguo Weishengtaixue Zazhi",
  ["zhongguo wei zhong bing ji jiu yi xue = chinese critical care medicine = zhongguo weizhongbing jijiuyixue"] = "Zhongguo Wei Zhong Bing Ji Jiu Yi Xue",
  ["zhongguo wuji fenxi huaxue"] = "Zhongguo Wuji Fenxi Huaxue",
  ["zhongguo xian dai shen jing ji bing za zhi"] = "Zhongguo Xian Dai Shen Jing Ji Bing Za Zhi",
  ["zhongguo xitu xuebao"] = "Zhongguo Xitu Xuebao",
  ["zhongguo xiu fu chong jian wai ke za zhi (chinese journal of reparative and reconstructive surgery)"] = "Zhongguo Xiu Fu Chong Jian Wai Ke Za Zhi",
  ["zhongguo xiu fu chong jian wai ke za zhi = zhongguo xiufu chongjian waike zazhi = chinese journal of reparative and reconstructive surgery"] = "Zhongguo Xiu Fu Chong Jian Wai Ke Za Zhi",
  ["zhongguo xu mu shou yi = zhongguo xumu shouyi = china animal husbandry & veterinary medicine"] = "Zhongguo Xu Mu Shou Yi",
  ["zhongguo xu mu xue hui hui zhi"] = "Journal of the Chinese Society of Animal Science|J. Chin. Soc. Anim. Sci.",
  ["zhongguo xu mu xue hui hui zhi = journal of the chinese society of animal science"] = "J Chin Soc Anim Sci",
  ["zhongguo xu qin chuan ran bing"] = "Zhongguo Xu Qin Chuan Ran Bing",
  ["zhongguo xue xi chong bing fang zhi za zhi = chinese journal of schistosomiasis control"] = "Zhongguo Xue Xi Chong Bing Fang Zhi Za Zhi",
  ["zhongguo yan cao xue bao = acta tabacaria sinica = zhongguo yancao xuebao"] = "Zhongguo Yan Cao Xue Bao",
  ["zhongguo yao li xue bao (acta pharmacologica sinica)"] = "Zhongguo Yao Li Xue Bao",
  ["zhongguo yao li xue bao = acta pharmacologica sinica"] = "Zhongguo Yao Li Xue Bao",
  ["zhongguo yao shi"] = "Zhongguo Yao Shi",
  ["zhongguo yao xue za zhi (zhongguo yao xue hui : 1989)"] = "Zhongguo Yao Xue Za Zhi",
  ["zhongguo yaoli xuebao"] = "Zhongguo Yaoli Xuebao",
  ["zhongguo yi liao qi xie za zhi (chinese journal of medical instrumentation)"] = "Zhongguo Yi Liao Qi Xie Za Zhi",
  ["zhongguo yi liao qi xie za zhi = chinese journal of medical instrumentation"] = "Zhongguo Yi Liao Qi Xie Za Zhi",
  ["zhongguo yi miao he mian yi"] = "Zhongguo Yi Miao He Mian Yi",
  ["zhongguo yi shi za zhi"] = "Zhongguo Yi Shi Za Zhi",
  ["zhongguo yi xue ke xue yuan xue bao (acta academiae medicinae sinicae)"] = "Zhongguo Yi Xue Ke Xue Yuan Xue Bao",
  ["zhongguo yi xue ke xue yuan xue bao. acta academiae medicinae sinicae"] = "Zhongguo Yi Xue Ke Xue Yuan Xue Bao",
  ["zhongguo ying yong sheng li xue za zhi (chinese journal of applied physiology)"] = "Zhongguo Ying Yong Sheng Li Xue Za Zhi",
  ["zhongguo ying yong sheng li xue za zhi = zhongguo yingyong shenglixue zazhi = chinese journal of applied physiology"] = "Zhongguo Ying Yong Sheng Li Xue Za Zhi",
  ["zhongguo yiyao gongye zazhi"] = "Zhongguo Yiyao Gongye Zazhi",
  ["zhongguo yiyao shengwu jishu"] = "Zhongguo Yiyao Shengwu Jishu",
  ["zhongguo you liao zuo wu xue bao"] = "Chinese journal of oil crop sciences / Zhongguo nong ye ke xue yuan you liao zuo wu yan jiu suo zhu ban|Zhongguo You Liao Zuo Wu Xue Bao",
  ["zhongguo you liao zuo wu xue bao = chinese journal of oil crop sciences"] = "Zhongguo You Liao Zuo Wu Xue Bao",
  ["zhongguo yuanyi wenzhai"] = "Zhongguo Yuanyi Wenzhai",
  ["zhongguo zhen jiu = chinese acupuncture & moxibustion"] = "Zhongguo Zhen Jiu",
  ["zhongguo zhen jun xue za zhi"] = "Zhongguo Zhen Jun Xue Za Zhi",
  ["zhongguo zhi bao dao kan = china plant protection"] = "Zhongguo Zhi Bao Dao Kan",
  ["zhongguo zhong xi yi jie he za zhi (chinese journal of integrated traditional and western medicine)"] = "Zhongguo Zhong Xi Yi Jie He Za Zhi",
  ["zhongguo zhong xi yi jie he za zhi zhongguo zhongxiyi jiehe zazhi = chinese journal of integrated traditional and western medicine"] = "Zhongguo Zhong Xi Yi Jie He Za Zhi",
  ["zhongguo zhong yao za zhi (china journal of chinese materia medica)"] = "Zhongguo Zhong Yao Za Zhi",
  ["zhongguo zhong yao za zhi = zhongguo zhongyao zazhi = china journal of chinese materia medica"] = "Zhongguo Zhong Yao Za Zhi",
  ["zhonghua bing li xue za zhi (chinese journal of pathology)"] = "Zhonghua Bing Li Xue Za Zhi",
  ["zhonghua bing li xue za zhi = chinese journal of pathology"] = "Zhonghua Bing Li Xue Za Zhi",
  ["zhonghua bing li xue za zhi chinese journal of pathology"] = "Zhonghua Bing Li Xue Za Zhi",
  ["zhonghua chuan ran bing za zhi = zhonghua chuanranbing zazhi = chinese journal of infectious diseases"] = "Zhonghua Chuan Ran Bing Za Zhi",
  ["zhonghua er bi yan hou ke za zhi"] = "Zhonghua Er Bi Yan Hou Ke Za Zhi",
  ["zhonghua er bi yan hou ke za zhi (chinese journal of otorhinolaryngology)"] = "Zhonghua Er Bi Yan Hou Ke Za Zhi",
  ["zhonghua er bi yan hou tou jing wai ke za zhi (chinese journal of otorhinolaryngology head and neck surgery)"] = "Zhonghua Er Bi Yan Hou Tou Jing Wai Ke Za Zhi",
  ["zhonghua er bi yan hou tou jing wai ke za zhi = chinese journal of otorhinolaryngology head and neck surgery"] = "Zhonghua Er Bi Yan Hou Tou Jing Wai Ke Za Zhi",
  ["zhonghua er ke za zhi = chinese journal of pediatrics"] = "Zhonghua Er Ke Za Zhi",
  ["zhonghua er ke za zhi. chinese journal of pediatrics"] = "Zhonghua Er Ke Za Zhi",
  ["zhonghua fang she xue za zhi (chinese journal of radiology)"] = "Zhonghua Fang She Xue Za Zhi",
  ["zhonghua fang she xue za zhi chinese journal of radiology"] = "Zhonghua Fang She Xue Za Zhi",
  ["zhonghua fang she yi xue yu fang hu za zhi = chinese journal of radiological medicine and protection"] = "Zhonghua Fang She Yi Xue Yu Fang Hu Za Zhi",
  ["zhonghua fangsh yixue yu fanghu zazhi"] = "Zhonghua Fangsh Yixue Yu Fanghu Zazhi",
  ["zhonghua fu chan ke za zhi"] = "Zhonghua Fu Chan Ke Za Zhi",
  ["zhonghua fu chan ke za zhi (chinese journal of obstetrics and gynecology)"] = "Zhonghua Fu Chan Ke Za Zhi",
  ["zhonghua gan zang bing za zhi (chinese journal of hepatology)"] = "Zhonghua Gan Zang Bing Za Zhi",
  ["zhonghua gan zang bing za zhi = zhonghua ganzangbing zazhi = chinese journal of hepatology"] = "Zhonghua Gan Zang Bing Za Zhi",
  ["zhonghua hu li za zhi (chinese journal of nursing)"] = "Zhonghua Hu Li Za Zhi",
  ["zhonghua hu li za zhi = chinese journal of nursing"] = "Zhonghua Hu Li Za Zhi",
  ["zhonghua jian yan yi xue za zhi = chinese journal of laboratory medicine"] = "Zhonghua Jian Yan Yi Xue Za Zhi",
  ["zhonghua jie he he hu xi xi ji bing za zhi (chinese journal of tuberculosis and respiratory diseases)"] = "Zhonghua Jie He He Hu Xi Xi Ji Bing Za Zhi",
  ["zhonghua jie he he hu xi xi ji bing za zhi = chinese journal of tuberculosis and respiratory diseases"] = "Zhonghua Jie He He Hu Xi Xi Ji Bing Za Zhi",
  ["zhonghua jie he he hu xi za zhi (chinese journal of tuberculosis and respiratory diseases)"] = "Zhonghua Jie He He Hu Xi Za Zhi",
  ["zhonghua jie he he hu xi za zhi = zhonghua jiehe he huxi zazhi = chinese journal of tuberculosis and respiratory diseases"] = "Zhonghua Jie He He Hu Xi Za Zhi",
  ["zhonghua kou qiang ke za zhi (chinese journal of stomatology)"] = "Zhonghua Kou Qiang Ke Za Zhi",
  ["zhonghua kou qiang ke za zhi [chinese journal of stomatology]"] = "Zhonghua Kou Qiang Ke Za Zhi",
  ["zhonghua kou qiang yi xue za zhi (chinese journal of stomatology)"] = "Zhonghua Kou Qiang Yi Xue Za Zhi",
  ["zhonghua kou qiang yi xue za zhi = zhonghua kouqiang yixue zazhi = chinese journal of stomatology"] = "Zhonghua Kou Qiang Yi Xue Za Zhi",
  ["zhonghua lao dong wei sheng zhi ye bing za zhi = zhonghua laodong weisheng zhiyebing zazhi = chinese journal of industrial hygiene and occupational diseases"] = "Zhonghua Lao Dong Wei Sheng Zhi Ye Bing Za Zhi",
  ["zhonghua lin xue ji kan = quarterly journal of chinese forestry"] = "Zhonghua Linxue Jikan",
  ["zhonghua liu xing bing xue za zhi"] = "Zhonghua Liu Xing Bing Xue Za Zhi",
  ["zhonghua liu xing bing xue za zhi = zhonghua liuxingbingxue zazhi"] = "Zhonghua Liu Xing Bing Xue Za Zhi",
  ["zhonghua min guo taiwan ren kou tong ji ji kan. china (republic : 1949- ). nei zheng bu"] = "Zhonghua Min Guo Taiwan Jen Kou Tong Ji Ji Kan",
  ["zhonghua minguo fu chan ke yi xue hui za zhi. the journal of obstetrics and gynecology of the republic of china"] = "Chung Hua Min Kuo Fu Chan Ko I Hsueh Hui",
  ["zhonghua minguo shou yi xue hui za zhi journal of the chinese society of veterinary science"] = "Zhonghua Min Guo Shou Yi Xue Hui Za Zhi",
  ["zhonghua minguo wei sheng wu ji mian yi xue za zhi (chinese journal of microbiology and immunology)"] = "Zhonghua Min Guo Wei Sheng Wu Ji Mian Yi Xue Za Zhi",
  ["zhonghua minguo wei sheng wu ji mian yi xue za zhi = chinese journal of microbiology and immunology"] = "Zhonghua Min Guo Wei Sheng Wu Ji Mian Yi Xue Za Zhi",
  ["zhonghua minguo wei sheng wu xue za zhi (chinese journal of microbiology)"] = "Zhonghua Min Guo Wei Sheng Wu Xue Za Zhi",
  ["zhonghua minguo wei sheng wu xue za zhi = chinese journal of microbiology"] = "Zhonghua Min Guo Wei Sheng Wu Xue Za Zhi",
  ["zhonghua minguo xiao er ke yi xue hui za zhi"] = "Zhonghua Min Guo Xiao Er Ke Yi Xue Hui Za Zhi",
  ["zhonghua minguo xiao er ke yi xue hui za zhi [journal]. zhonghua minguo xiao er ke yi xue hui"] = "Zhonghua Min Guo Xiao Er Ke Yi Xue Hui Za Zhi",
  ["zhonghua minguo xin zang xue hui za zhi = acta cardiologica sinica"] = "Zhonghua Minguo Xin Zang Xue Hui Za Zhi",
  ["zhonghua nan ke xue = national journal of andrology"] = "Zhonghua Nan Ke Xue",
  ["zhonghua nei ke za zhi"] = "Zhonghua Nei Ke Za Zhi",
  ["zhonghua nei ke za zhi (chinese journal of internal medicine)"] = "Zhonghua Nei Ke Za Zhi",
  ["zhonghua nei ke za zhi [chinese journal of internal medicine]"] = "Zhonghua Nei Ke Za Zhi",
  ["zhonghua shao shang za zhi"] = "Zhonghua Shao Shang Za Zhi",
  ["zhonghua shao shang za zhi = zhonghua shaoshang zazhi = chinese journal of burns"] = "Zhonghua Shao Shang Za Zhi",
  ["zhonghua shen jing jing shen ke za zhi (chinese journal of neurology and psychiatry)"] = "Zhonghua Shen Jing Jing Shen Ke Za Zhi",
  ["zhonghua shen jing jing shen ke za zhi = chinese journal of neurology and psychiatry"] = "Zhonghua Shen Jing Jing Shen Ke Za Zhi",
  ["zhonghua shen jing ke za zhi = chinese journal of neurology"] = "Zhonghua Shen Jing Ge Za Zhi",
  ["zhonghua sheng wu yi xue gong cheng za zhi"] = "Zhonghua Sheng Wu Yi Xue Gong Cheng Za Zhi",
  ["zhonghua shi yan he lin chuang bing du xue za zhi = zhonghua shiyan he linchuang bingduxue zazhi = chinese journal of experimental and clinical virology"] = "Zhonghua Shi Yan He Lin Chuang Bing Du Xue Za Zhi",
  ["zhonghua shi yan wai ke za zhi = zhonghua shiyan waike zazhi = chinese journal of experimental surgery"] = "Zhonghua Shi Yan Wai Ke Za Zhi",
  ["zhonghua wai ke za zhi (chinese journal of surgery)"] = "Zhonghua Wai Ke Za Zhi",
  ["zhonghua wai ke za zhi [chinese journal of surgery]"] = "Zhonghua Wai Ke Za Zhi",
  ["zhonghua wei chang wai ke za zhi = chinese journal of gastrointestinal surgery"] = "Zhonghua Wei Chang Wai Ke Za Zhi",
  ["zhonghua wei sheng wu xue he mian yi xue za zhi = zhonghua weishengwuxue he mianyixue zazhi"] = "Zhonghua Wei Sheng Wu Xue He Mian Yi Xue Za Zhi",
  ["zhonghua wei zhong bing ji jiu yi xue"] = "Zhonghua Wei Zhong Bing Ji Jiu Yi Xue",
  ["zhonghua xiao hua za zhi = chinese journal of digestion"] = "Chung Hua Hsiao Hua Tsa Chih",
  ["zhonghua xin xue guan bing za zhi"] = "Zhonghua Xin Xue Guan Bing Za Zhi",
  ["zhonghua xin xue guan bing za zhi (chinese journal of cardiovascular diseases)"] = "Zhonghua Xin Xue Guan Bing Za Zhi",
  ["zhonghua xin xue guan bing za zhi [chinese journal of cardiovascular diseases]"] = "Zhonghua Xin Xue Guan Bing Za Zhi",
  ["zhonghua xue ye xue za zhi = zhonghua xueyexue zazhi"] = "Zhonghua Xue Ye Xue Za Zhi",
  ["zhonghua ya yi xue hui za zhi"] = "Zhonghua Ya Yi Xue Hui Za Zhi",
  ["zhonghua yi shi za zhi (beijing, china : 1980)"] = "Zhonghua Yi Shi Za Zhi",
  ["zhonghua yi xue yi chuan xue za zhi (chinese journal of medical genetics)"] = "Zhonghua Yi Xue Yi Chuan Xue Za Zhi",
  ["zhonghua yi xue yi chuan xue za zhi = zhonghua yixue yichuanxue zazhi = chinese journal of medical genetics"] = "Zhonghua Yi Xue Yi Chuan Xue Za Zhi",
  ["zhonghua yi xue za zhi"] = "Zhonghua Yi Xue Za Zhi",
  ["zhonghua yi xue za zhi (chinese medical journal, taipei)"] = "Zhonghua Yi Xue Za Zhi (Taipei)",
  ["zhonghua yi xue za zhi = chinese medical journal; free china ed"] = "Zhonghua Yi Xue Za Zhi (Taipei)",
  ["zhonghua yi yuan gan ran xue za zhi"] = "Chinese journal of nosocomiology|Zhonghua Yi Yuan Gan Ran Xue Za Zhi",
  ["zhonghua yi yuan gan ran xue za zhi = chinese journal of nosocomiology"] = "Zhonghua Yi Yuan Gan Ran Xue Za Zhi",
  ["zhonghua yu fang yi xue za zhi (chinese journal of preventive medicine)"] = "Zhonghua Yu Fang Yi Xue Za Zhi",
  ["zhonghua yu fang yi xue za zhi [chinese journal of preventive medicine]"] = "Zhonghua Yu Fang Yi Xue Za Zhi",
  ["zhonghua zheng xing shao shang wai ke za zhi (chinese journal of plastic surgery and burns)"] = "Zhonghua Zheng Xing Shao Shang Wai Ke Za Zhi",
  ["zhonghua zheng xing shao shang wai ke za zhi = zhonghua zheng xing shao shang waikf [i.e. waike] zazhi = chinese journal of plastic surgery and burns"] = "Zhonghua Zheng Xing Shao Shang Wai Ke Za Zhi",
  ["zhonghua zheng xing wai ke za zhi"] = "Zhonghua Zheng Xing Wai Ke Za Zhi",
  ["zhonghua zheng xing wai ke za zhi = zhonghua zhengxing waike zazhi = chinese journal of plastic surgery"] = "Zhonghua Zheng Xing Wai Ke Za Zhi",
  ["zhonghua zhong liu za zhi (chinese journal of oncology)"] = "Zhonghua Zhong Liu Za Zhi",
  ["zhonghua zhong liu za zhi [chinese journal of oncology]"] = "Zhonghua Zhong Liu Za Zhi",
  ["zhongnan minzu daxue xuebao. ziran kexue ban = journal of south-central university for nationalities. natural science edition"] = "Zhongnan Minzu Daxue Xuebao Ziran Kexue Ban",
  ["zhongshan da xue xue bao. zi ran ke xue ban"] = "Acta scientiarum naturalium Universitatis Sunyatseni|Zhongshan Da Xue Xue Bao Zi Ran Ke Xue Ban",
  ["zhongshan da xue xue bao. zi ran ke xue ban = acta scientiarum naturalium universitatis sunyatseni"] = "Zhongshan Da Xue Xue Bao Zi Ran Ke Xue Ban",
  ["zhongshan yi ke da xue xue bao = zhongshan yike daxue xuebao = academic journal of sun yat-sen university of medical sciences"] = "Zhongshan Yi Ke Da Xue Xue Bao",
  ["zhurnal belorusskogo gosudarstvennogo universiteta. matematika. informatika"] = "Zh. Beloruss. Gos. Univ. Mat. Inform.",
  ["zhurnal eksperimental'noĭ i klinicheskoĭ meditsiny"] = "Zh Eksp Klin Med",
  ["zhurnal eksperimentalnoi i klinicheskoi meditsiny"] = "Zh. Eksp. Klin. Med.",
  ["zhurnal eksperimentalnoi i teoreticheskoi fiziki [soviet physics–jetp]"] = "Zh. Eksp. Teor. Fiz. [Sov. Phys. JETP]",
  ["zhurnal eksperimental’noi i klinicheskoi meditsiny"] = "Zh Eksp Klin Med",
  ["zhurnal eksperimental’noi i teoreticheskoi fiziki"] = "Zh. Eksp. Teor. Fiz.",
  ["zhurnal evoliutsionnoi biokhimii i fiziologii"] = "Zh Evol Biokhim Fiziol",
  ["zhurnal evoliutsionnoĭ biokhimii i fiziologii"] = "Zh Evol Biokhim Fiziol",
  ["zhurnal evolyutsionnoi biokhimii i fiziologii"] = "Zh. Evol. Biokhim. Fiziol.",
  ["zhurnal fizicheskoi khimii"] = "Zh. Fiz. Khim.",
  ["zhurnal fizicheskoi khimii [russian journal of physical chemistry]"] = "Zh. Fiz. Khim. [Russ. J. Phys. Chem.]",
  ["zhurnal infektologii"] = "Z Infektol",
  ["zhurnal matematicheskoi fiziki, analiza, geometrii. journal of mathematical physics, analysis, geometry"] = "Zh. Mat. Fiz. Anal. Geom.",
  ["zhurnal mikrobiologii, epidemiologii i immunobiologii"] = "Zh. Mikrobiol. Epidemiol. Immunobiol.",
  ["zhurnal mikrobiologii, epidemiologii, i immunobiologii"] = "Zh Mikrobiol Epidemiol Immunobiol",
  ["zhurnal mikrobiologii, ėpidemiologii i immunobiologii"] = "Zh Mikrobiol Epidemiol Immunobiol",
  ["zhurnal neorganicheskoi khimii"] = "Zh. Neorg. Khim.",
  ["zhurnal nevrologii i psikhiatrii imeni s.s. korsakova"] = "Zh Nevrol Psikhiatr Im S S Korsakova",
  ["zhurnal nevrologii i psikhiatrii imeni s.s. korsakova / ministerstvo zdravookhraneniia i meditsinskoi promyshlennosti rossiiskoi federatsii, vserossiiskoe obshchestvo nevrologov [i] vserossiiskoe obshchestvo psikhiatrov"] = "Zh Nevrol Psikhiatr Im S S Korsakova",
  ["zhurnal nevropatologii i psikhiatrii imeni s. s. korsakova"] = "Zh. Nevropatol. Psikhiatr. Im. S. S. Korsakova",
  ["zhurnal nevropatologii i psikhiatrii imeni s.s. korsakova (moscow, russia : 1952)"] = "Zh Nevropatol Psikhiatr Im S S Korsakova",
  ["zhurnal obshchei biologii"] = "Zh Obshch Biol",
  ["zhurnal obshchei khimii"] = "Zh. Obshch. Khim.",
  ["zhurnal obshcheĭ biologii"] = "Zh Obshch Biol",
  ["zhurnal organicheskoi khimii"] = "Zh. Org. Khim.",
  ["zhurnal organichnoi ta farmatsevtichnoi khimii"] = "Zh. Org. Farm. Khim.",
  ["zhurnal prikladnoi khimii"] = "Zh. Prikl. Khim.",
  ["zhurnal prikladnoi khimii (s. -peterburg)"] = "Zh. Prikl. Khim. (Leningrad)",
  ["zhurnal prikladnoi spektroskopii"] = "Zh. Prikl. Spektrosk.",
  ["zhurnal prikladnoi spektroskopii [journal of applied spectroscopy (ussr)]"] = "Zh. Prikl. Spektrosk. [J. Appl. Spectrosc. (USSR)]",
  ["zhurnal prikladnoĭ khimii (leningrad, r.s.f.s.r.)"] = "Zhurnal Prikl Him",
  ["zhurnal sibirskogo federalnogo universiteta. matematika i fizika"] = "Zh. Sib. Fed. Univ. Mat. Fiz.",
  ["zhurnal tekhnicheskoi fiziki [soviet physics–technical physics]"] = "Zh. Tekh. Fiz. [Sov. Phys. Tech. Phys.]",
  ["zhurnal ushnykh, nosovykh i gorlovykh boleznei"] = "Zh. Ushn. Nos. Gorl. Bolezn.",
  ["zhurnal ushnykh, nosovykh i gorlovykh bolezneĭ = the journal of otology, rhinology, and laryngologie [sic]"] = "Zh Ushn Nos Gorl Bolezn",
  ["zhurnal voprosy neirokhirurgii imeni n. n. burdenko"] = "Zh Vopr Neirokhir Im N N Burdenko",
  ["zhurnal voprosy neĭrokhirurgii imeni n. n. burdenko"] = "Zh Vopr Neirokhir Im N N Burdenko",
  ["zhurnal vsesoyuznogo khimicheskogo obshchestva im. d. i. mendeleeva"] = "Zh. Vses. Khim. O-va. im. D. I. Mendeleeva",
  ["zhurnal vychislitel\\cprime noy matematiki i matematicheskoy fiziki"] = "Zh. Vychisl. Mat. Mat. Fiz.",
  ["zhurnal vychislitelnoi matematiki i matematicheskoi fiziki. rossiiskaya akademiya nauk"] = "Zh. Vychisl. Mat. Mat. Fiz.",
  ["zhurnal vysshei nervnoi deiatelnosti imeni i p pavlova"] = "Zh Vyssh Nerv Deiat Im I P Pavlova",
  ["zhurnal vysshei nervnoi deiatelnosti imeni i. p. pavlova"] = "Zh. Vyssh. Nerv. Deiat. Im. I. P. Pavlova",
  ["zhurnal vyssheĭ nervnoĭ deiatelnosti imeni i p pavlova"] = "Zh Vyssh Nerv Deiat Im I P Pavlova",
  ["zhurnal éksperimentalnoi i teoreticheskoi fiziki"] = "Zh. Eksp. Teor. Fiz",
  ["zhuxing xuebao"] = "Zhuxing Xuebao",
  ["zi dong hua xue bao = acta automatica sinica"] = "Zidonghua Xuebao",
  ["zi ran ke xue (irvine, calif.)"] = "Zi Ran Ke Xue",
  ["zi ran ke xue jin zhan"] = "Zi Ran Ke Xue Jin Zhan",
  ["zi ran ke xue shi yan = studies in the history of natural sciences"] = "Zi Ran Ke Xue Shi Yan",
  ["zi ran za zhi. ziran zazhi"] = "Zi Ran Za Zhi",
  ["zi you zhongguo zhi gong ye"] = "Ind Free China",
  ["zimbabwe nurse"] = "Zimbabwe Nurse",
  ["zimbabwe rhodesia nurse"] = "Zimbabwe Rhod Nurse",
  ["ziva antika (antiquité vivante)"] = "ZAnt",
  ["ziva antika. antiquite vivante"] = "Ziva Antika",
  ["zīst/fanāvarī-i giyāhān-i zirāī"] = "Zist Fanavarii Giyahani Zirai",
  ["zīstʹshināsī-i mīkrūʹurgānīsmʹhā : biological journal of microorganism"] = "Zistshinasii Mikruurganismha",
  ["zkg international"] = "ZKG Int.",
  ["znamenatel'nye daty istorii zdravookhraneniia i meditsiny azerbaĭdzhana"] = "Znam Daty Istor Zdravookhr Med Azerb",
  ["zobozdravstveni vestnik"] = "Zobozdrav Vestn",
  ["zoo biology"] = "Zoo Biol",
  ["zoologia (curitiba, brazil)"] = "Zoologia (Curitiba)",
  ["zoologia caboverdiana"] = "Zool Caboverdiana",
  ["zoologiâ bespozvonoc̆nyh"] = "Zool Bespozvon",
  ["zoologica poloniae"] = "Zool Pol",
  ["zoologica scripta"] = "Zool Scr",
  ["zoologica; scientific contributions of the new york zoological society"] = "Zool Sci Contrib N Y Zool Soc",
  ["zoological journal of the linnean society"] = "Zool J Linn Soc",
  ["zoological letters"] = "Zoological Lett",
  ["zoological research"] = "Zool Res",
  ["zoological science"] = "Zoolog Sci",
  ["zoological studies"] = "Zool Stud",
  ["zoological systematics"] = "Zool Syst",
  ["zoologicheskii zhurnal"] = "Zool. Zh.",
  ["zoologicheskiĭ zhurnal"] = "Zool Zhurnal",
  ["zoologische abhandlungen"] = "Zool Abh",
  ["zoologische garten"] = "Zool. Garten",
  ["zoologische jahrbucher. abteilung fur anatomie und ontogenie der tiere"] = "Zool Jahrb Abt Anat Ontogenie Tiere",
  ["zoologische jahrbücher. abteilung für anatomie und ontogenie der tiere"] = "Zool Jahrb Abt Anat Ontogenie Tiere",
  ["zoologische mededeelingen"] = "Zool Meded",
  ["zoologischer anzeiger"] = "Zool Anz",
  ["zoologischer garten n.f."] = "Zool. Gart. N.F.",
  ["zoologisches jahrbuch, abteilung für systematik"] = "Zool. Jahrb. Abt. Syst.",
  ["zoology (jena, germany)"] = "Zoology (Jena)",
  ["zoology and ecology"] = "Zool. Ecol.",
  ["zoology and ecology (print)"] = "Zool Ecol",
  ["zoology in the middle east"] = "Zool Middle East",
  ["zoonoses and public health"] = "Zoonoses Public Health",
  ["zoonoses research"] = "Zoonoses Res",
  ["zoos' print journal"] = "Zoos Print J",
  ["zoosystematica rossica"] = "Zoosyst Ross",
  ["zoosystematics and evolution"] = "Zoosyst. Evol.",
  ["zoosystematics and evolution : mitteilungen aus dem museum für naturkunde in berlin"] = "Zoosyst Evol",
  ["zootecnia tropical"] = "Zootec Trop",
  ["zorg + welzijn"] = "Zorg Welz",
  ["zorgvisie : vakopinieblad voor de zorgsector"] = "Zorgvisie",
  ["zpg reporter"] = "ZPG Report",
  ["zpravy geografickeho ustavu csav"] = "Zpr Geogr Ust CSAV",
  ["zprávy geografického ústavu čsav"] = "Zpr Geogr Ust CSAV",
  ["zprávy jednoty klasickaych filologu"] = "ZJKF",
  ["zu guo"] = "Zu Guo",
  ["zubolekarski pregled"] = "Zubolekarski Pregl",
  ["zuger nachrichten"] = "Zuger Nachr.",
  ["zuger neujahrsblatt"] = "Zuger Neujahrsbl.",
  ["zuo wu xue bao"] = "Zuo Wu Xue Bao",
  ["zur modernen deutung der aristotelischen logik"] = "Mod. Deutung Aristot. Log.",
  ["zurcher medizingeschichtliche abhandlungen"] = "Zur Medizingesch Abh",
  ["zurich lectures in advanced mathematics"] = "Zur. Lect. Adv. Math.",
  ["zurich lectures on multiphase flow"] = "Zur. Lect. Multiph. Flow",
  ["zürcher medizingeschichtliche abhandlungen"] = "Zur Medizingesch Abh",
  ["zvestí c̆erveného kríz̆a"] = "Zvesti Cerv Kriza",
  ["zwingliana : mitteilungen zur geschichte zwinglis der reformation"] = "Zwingliana",
  ["zygote (cambridge, england)"] = "Zygote",
  ["züchter, der"] = "Züchter",
  ["zürcher geographische schriften"] = "Zür. Geogr. Schr.",
  ["zürcher wald"] = "Zür.ld",
  ["živa antika. antiquité vivante"] = "Ziva Antika",
  ["ägypten und levante. egypt and the levant. internationale zeitschrift für ägyptische archäologie und deren nachbargebiete"] = "ÄgLev",
  ["échos du monde classique (classical views)"] = "EMC",
  ["école antique de nîmes. bulletin annuel"] = "EcAntNimes",
  ["école pratique des hautes études"] = "Éc. Prat. Ht. études IV$\\sp e$ Sect. Sci. Hist. Philol. V Ht. études Médiév. Mod.",
  ["études celtiques"] = "EC",
  ["études classiques. faculté des lettres et sciences humaines d’aix"] = "EtClAix",
  ["études crétoises"] = "EtCret",
  ["études de lettres: bulletin de la faculté des lettres de l’univ. de lausanne et de la soc. des études de lettres"] = "EL",
  ["études de papyrologie"] = "EtP",
  ["études d’archéologie classique"] = "EtACl",
  ["études et travaux. studia i prace. travaux du centre d’archéologie méditerranéenne de l’académie des sciences polonaise"] = "EtTrav",
  ["études indo-européennes"] = "EIE",
  ["études philosophiques"] = "EPh",
  ["études préliminaires aux religions orientales dans l’empire romain"] = "EPRO",
  ["études sur pézenas et l’hérault"] = "EtPezenas",
  ["ökologie und landbau"] = "Ökol. Landbau",
  ["österreichische akademie der wissenschaften"] = "Anz. Österreich. Akad. Wiss. Math.-Natur. Kl.",
  ["österreichische akademie der wissenschaften mathematisch-naturwissenschaftliche klasse"] = "Österreich. Akad. Wiss. Math.-Natur. Kl. Sitzungsber. II",
  ["österreichische akademie der wissenschaften, philosophisch-historische klasse. denkschriften"] = "DenkschrWien",
  ["österreichische akademie der wissenschaften. almanach"] = "AlmanachWien",
  ["österreichische botanische zeitschrift"] = "Österr. bot. Z.",
  ["österreichische forstzeitung"] = "Österr. Forstztg.",
  ["österreichische wasser- und abfallwirtschaft"] = "Österr. Wasser- Abfallwirtsch.",
  ["überblicke mathematik"] = "Überbl. Math.",
  ["ŭi sahak"] = "Uisahak",
  ["ʻādīyāt ḥalab"] = "Adiyat Halab",
  ["ʻālam al-fikr"] = "Alam Al Firkr",
  ["ʻīrāq (1977)"] = "Iraq",
  ["ʻulūm va tiknuluzhī-i khalīj-i fars"] = "Ulum Va Tiknuluzhii Khaliji Fars",
  ["άνθρωποσ. όργανο τησ ανθρωπολογικήσ εταιρείασ ελλάδοσ"] = "Anthropos",
  ["ανθρωπολογικά και αρχαιολογικά χρονικά"] = "AnthrAChron",
  ["αρχαιολογικά ανάλεκτα εξ αθηνών"] = "AAA",
  ["αρχαιολογική eφημερίσ"] = "AEphem",
  ["αρχαιολογικόν δελτίον (μελέτεσ)"] = "ADelt A",
  ["αρχαιολογικόν δελτίον (χρονικά)"] = "Adelt B",
  ["αρχείον ευβοϊκών μελετών"] = "ArchEubMel",
  ["αρχείον των βυζαντινών μνημείων τησ ελλάδοσ"] = "ArchByzMnem",
  ["βυζαντινά. επιστημονικόν όργανον κέντρου βυζαντινών ερευνών φιλοσοφικήσ σχολήσ αριστοτελείου πανεπιστημίου θεσσαλονίκησ"] = "Byzantina",
  ["δελτίον τησ χριστιανικήσ αρχαιολογικήσ εταιρείασ"] = "DeltChrA",
  ["εγνατία. επιστημονική επετηρίδα τησ φιλοσοφικήσ σχολήσ, αριστοτέλειο πανεπιστήμιο θεσσαλονίκησ, τμήμα ιστορίασ και αρχαιολογίασ"] = "Egnatia",
  ["εικασμόσ. quaderni bolognesi di filologia classica"] = "Eikasmos",
  ["ελληνικά. φιλολογικόν, ιστορικόν και λαογραφικόν περιοδικόν σύγγραμμα"] = "Ellenika",
  ["ελληνικά: φιλολογικό, ιστορικό και λαογραφικό περιοδικό σύγγραμμα"] = "Hellenica",
  ["επετηρίσ τησ εταιρείασ βοιωτικών μελετών"] = "EpetBoiotMel",
  ["επετηρίσ τησ εταιρείασ βυζαντινών σπουδών"] = "EpetByzSpud",
  ["επετηρίσ τησ εταιρείασ κυκλαδικών μελετών"] = "EpetKyklMel",
  ["επιστημονική επετηρίδα τησ πολυτεχνικήσ σχολήσ, αριστοτέλειο πανεπιστήμιο θεσσαλονίκησ, τμήμα αρχιτεκτόνων"] = "EpistEpetPolytThess",
  ["επιστημονική επετηρίδα τησ φιλοσοφικήσ σχολήσ του πανεπιστημίου θεσσαλονίκησ"] = "EpistEpetThess",
  ["επιστημονική επετηρίσ τησ φιλοσοφικήσ σχολήσ του πανεπιστημίου αθηνών"] = "EpistEpetAth",
  ["ευλιμένη (μεσογειακή αρχαιολογική εταιρεία)"] = "Eulimene",
  ["η θεσσαλονίκη"] = "Thessalonike",
  ["ηόροσ. ἔνα ἀρχαιογνωστικò περιοδικό"] = "Horos",
  ["θρακικά χρονικά"] = "ThrakChron",
  ["θρακική επετηρίδα"] = "ThrakEp",
  ["κρητικά χρονικά"] = "KretChron",
  ["κυπριακαì σπουδαί"] = "KypSpud",
  ["κυπριακή αρχαιολογία. archaeologia cypria"] = "KypA",
  ["μακεδονικά. σύγγραμμα περιοδικόν τησ εταιρείασ μακεδονικών σπουδών"] = "Makedonika",
  ["μουσείο μπενάκη"] = "MusBenaki",
  ["νέα ῥώμη. rivista di ricerche bizantinistiche"] = "Nea Rhome",
  ["νομισματικά χρονικά. περιοδική έκδωσισ τησ ελληνικήσ νομισματικήσ εταιρείασ"] = "NomChron",
  ["πλάτον. δελτίον τησ εταιρείασ ελλήνων φιλολόγων"] = "Platon",
  ["πολέμων. αρχαιολογικόν περιοδικόν"] = "Polemon",
  ["πρακτικά τησ ακαδημίασ αθηνών"] = "PraktAkAth",
  ["πρακτικά τησ εν αθήναισ αρχαιολογικήσ εταιρείασ"] = "Prakt",
  ["ρήμα. mitteilungen zur indogermanischen, vornehmlich indo-iranischen wortkunde sowie zur holothetischen sprachtheorie"] = "Rema",
  ["τάλαντα. proceedings of the dutch archaeological and historical society"] = "Talanta",
  ["το αρχαιολογικό έργο στη μακεδονία και θράκη"] = "AErgoMak",
  ["το έργον τησ αρχαιολογικήσ εταιρείασ"] = "Ergon",
  ["τόποι. orient – occident"] = "Topoi",
  ["‘atiqot. journal of the israel department of antiquities"] = "Atiqot",
  ["‘atiqot. journal of the israel department of antiquities. hebrew series"] = "AtiqotHeb",
  ["’adiyat halab"] = "Adiyat Halab",
  ["’alon mahleqat ha-’atiqot šel medinat jisra’el"] = "AlonJisrael",
  ["’iraq (1977)"] = "Iraq",
  ["“politehnica” university of bucharest. scientific bulletin. series a. applied mathematics and physics"] = "Politehn. Univ. Bucharest Sci. Bull. Ser. A Appl. Math. Phys.",
}

unabbrevs = {
  ['"g3-genes genomes genet"'] = "G3-Genes Genomes Genetics",
  ["$k$-monogr math"] = "$K$-Monographs in Mathematics",
  ["$r$ ser"] = "The $R$ Series",
  ["10th ieee int conf big knowl (2019)"] = "10th IEEE International Conference on Big Knowledge : proceedings : 10-11 November 2019, Beijing, China. IEEE International Conference on Big Knowledge (10th : 2019 : Beijing, China)",
  ["1199 news"] = "1199 news. National Union of Hospital and Health Care Employees. District 1199",
  ["16th web all conf (2019)"] = "The 16th Web for All Conference : San Francisco, 2019. Web for All Conference (16th : 2019 : San Francisco, Calif.)",
  ["19 (lond)"] = "19 : interdisciplinary studies in the long nineteenth century",
  ["1999 z sozial 20 21 jhd"] = "1999, Zeitschrift für Sozialgeschichte des 20. und 21. Jahrhunderts",
  ["19th century music"] = "19th century music",
  ["20 century br hist"] = "20 century British history",
  ["2013 ieee 26th int conf micro electro mech syst mems 2013 (2013)"] = "2013 IEEE 26th International Conference on Micro Electro Mechanical Systems (MEMS 2013) : Taipei, Taiwan, 20-24 January 2013. IEEE International Conference on Micro Electro Mechanical Systems (26th : 2013 : Taipei, Taiwan)",
  ["2015 ieee signal process signal process educ workshop sp spe (2015)"] = "2015 IEEE Signal Processing and Signal Processing Education Workshop (SP/SPE)",
  ["2017 des med devices conf (2017)"] = "2017 Design of Medical Devices Conference : April 10-13, 2017, Minneapolis, Minnesota, USA. Design of Medical Devices Conference (2017 : Minneapolis, Minn.)",
  ["2017 first ieee mtts int microw bio conf (imbioc) (2017)"] = "2017 First IEEE MTT-S International Microwave Bio Conference (IMBioC) : 15-17 May, 2017, Gothenburg, Sweden. IEEE MTT-S International Microwave Bio Conference (1st : 2017 : Göteborg, Sweden)",
  ["2017 ieee sens appl symp (sas) (2017)"] = "2017 IEEE Sensors Applications Symposium (SAS). IEEE Staff",
  ["2017 ieee workshop vis anal healthc vahc (2017)"] = "2017 IEEE Workshop on Visual Analytics in Healthcare (VAHC) : 1 October 2017. Workshop on Visual Analytics in Healthcare (8th : 2017 : Phoenix, Ariz.)",
  ["2018 des med devices conf (2018)"] = "2018 Design of Medical Devices Conference : April 9-12, 2018, Minneapolis, Minnesota, USA. Design of Medical Devices Conference (2018 : Minneapolis, Minn.)",
  ["2018 ieee conf cogn comput asp situat manag cogsima (2018)"] = "2018 IEEE Conference on Cognitive and Computational Aspects of Situation Management (CogSIMA) : proceedings : Boston, MA, USA, 11-14 June 2018. IEEE Conference on Cognitive and Computational Aspects of Situation Management (8th : 2018 : Boston, Mass.)",
  ["2018 ieee int conf healthc inform workshop (2018)"] = "2018 IEEE International Conference on Healthcare Informatics Workshops : proceedings : 4-7 June 2018, New York, NY. International Conference on Healthcare Informatics (6th : 2018 : New York, N.Y.). Workshops",
  ["2018 int symp big data vis immers analyt (bdva) (2018)"] = "2018 International Symposium on Big Data Visual and Immersive Analytics (BDVA) : Konstanz, Germany, October 17 -19, 2018. IEEE International Symposium on Big Data Visual and Immersive Analytics (4th : 2018 : Konstanz, Germany)",
  ["2019 des med devices conf (2019)"] = "2019 Design of Medical Devices Conference : April 15-18, 2019, Minneapolis, Minnesota, USA. Design of Medical Devices Conference (2019 : Minneapolis, Minn.)",
  ["2019 ieee smartworld ubiquitous intell comput adv trust comput scalable comput commun cloud big data comput internet people smart city innov (2019)"] = "2019 IEEE SmartWorld, Ubiquitous Intelligence & Computing, Advanced & Trusted Computing, Scalable Computing & Communications, Cloud & Big Data Computing, Internet of People and Smart City Innovation : SmartWorld/UIC/ATC/SCALCOM/IOP/SCI 2019",
  ["21st century sci technol"] = "21st century science & technology",
  ["24th int conf 3d web technol (2019)"] = "24th International Conference on 3D Web Technology",
  ["2d mater"] = "2D Materials",
  ["2nd int conf bioinform comput biol (2010)"] = "2nd International Conference on Bioinformatics and Computational Biology 2010, (BICoB-2010), Honolulu, Hawaii, USA, 24-26 March 2010. International Conference on Bioinformatics and Computational Biology (2nd : 2010 : Honolulu, Hawaii)",
  ["3 biotech"] = "3 Biotech",
  ["30$\\\\sp {\\\\rm o}$ coloq bras mat"] = "30$\\\\sp {\\\\rm o}$ Coloquio Brasileiro de Matematica",
  ["31$\\\\sp {\\\\rm o}$ coloq bras mat"] = "31$\\\\sp {\\\\rm o}$ Coloquio Brasileiro de Matematica",
  ["32 $\\\\sp {\\\\rm o}$ coloq bras mat"] = "32 $\\\\sp {\\\\rm o}$ Coloquio Brasileiro de Matematica",
  ["33 $\\\\sp {\\\\rm o}$ coloq bras mat"] = "33 $\\\\sp {\\\\rm o}$ Coloquio Brasileiro de Matematica",
  ["34 $\\\\sp {\\\\rm o}$ coloq bras mat"] = "34 $\\\\sp {\\\\rm o}$ Coloquio Brasileiro de Matematica",
  ["3d print addit manuf"] = "3D Printing and Additive Manufacturing",
  ["3d print med"] = "3D printing in medicine",
  ["48th aiaa fluid dyn conf 2018 (2018)"] = "48th AIAA Fluid Dynamics Conference 2018 : held at the AIAA Aviation Forum 2018 : Atlanta, Georgia, USA, 25-29 June 2018. AIAA Fluid Dynamics Conference (48th : 2018 : Atlanta, Ga.)",
  ["4or - q j oper res"] = "4OR - A Quarterly Journal of Operations Research",
  ["50th anniv cracow univ technol anniv issue"] = "50th Anniversary Cracow University of Technology",
  ["7th conf retrovir oppor infect jan 30 feb 2 2000 conf retrovir oppor infect 7th 2000 san franc calif"] = "7th Conference on Retroviruses and Opportunistic Infections, January 30-February 2, 2000. Conference on Retroviruses and Opportunistic Infections (7th : 2000 : San Francisco, Calif.)",
  ["8th annu symp nonhum primate models aids symp nonhum primate models aids 8th 1990 new orleans la"] = "8th Annual Symposium on Nonhuman Primate Models for AIDS : November 28-30, 1990, the Westin Canal Place Hotel, New Orleans, Louisiana. Symposium on Nonhuman Primate Models for AIDS (8th : 1990 : New Orleans, La.)",
  ["9th conf retrovir oppor infect feb 24 28 2002 wash state conv trade cent seattle wash conf retrovir oppor infect 9th 2002 seattle wash"] = "9th Conference on Retroviruses and Opportunistic Infections, February 24-28, 2002, Washington State Convention and Trade Center, Seattle, WA. Conference on Retroviruses and Opportunistic Infections (9th : 2002 : Seattle, Wash.)",
  ["a & m"] = "A&M : advertising and marketing",
  ["a a case rep"] = "A & A case reports",
  ["a a g bijdr"] = "A.A.G. bijdragen",
  ["a a pract"] = "A&A practice",
  ["a m"] = "A&M : advertising and marketing",
  ["a m a arch ind hyg occup med"] = "A. M. A. archives of industrial hygiene and occupational medicine",
  ["a u u st ge"] = "Acta Universitatis Upsaliensis : Studia Germanistica Upsaliensia",
  ["aace clin case rep"] = "AACE clinical case reports",
  ["aachen kunstbl"] = "Aachener Kunstblätter (DuMont Buchverlag)",
  ["aacn adv crit care"] = "AACN advanced critical care",
  ["aacn clin issues"] = "AACN clinical issues",
  ["aacn clin issues crit care nurs"] = "AACN clinical issues in critical care nursing",
  ["aacr meet abstr online"] = "AACR meeting abstracts online. American Association for Cancer Research. Annual Meeting",
  ["aade ed j"] = "AADE editors' journal",
  ["aade pract"] = "AADE in practice",
  ["aahs j"] = "AAHS journal",
  ["aana j"] = "AANA journal",
  ["aannt j"] = "AANNT journal",
  ["aaohn j"] = "AAOHN journal : official journal of the American Association of Occupational Health Nurses",
  ["aap news"] = "AAP news",
  ["aapg bull"] = "AAPG Bulletin",
  ["aapg mem"] = "AAPG Memoir",
  ["aapg memoir"] = "AAPG Memoirs",
  ["aapi nexus"] = "AAPI nexus : Asian Americans and Pacific Islanders, policy practice and community",
  ["aapl newsl"] = "AAPL newsletter",
  ["aapm rep"] = "AAPM report",
  ["aappo j"] = "AAPPO journal : the journal of the American Association of Preferred Provider Organizations",
  ["aapps bull"] = "AAPPS Bulletin",
  ["aaps j"] = "The AAPS journal",
  ["aaps open"] = "AAPS open",
  ["aaps pharmsci"] = "AAPS pharmSci",
  ["aaps pharmscitech"] = "AAPS PharmSciTech",
  ["aarc times"] = "AARCTimes",
  ["aarn news lett"] = "AARN news letter",
  ["aas open res"] = "AAS open research",
  ["aastaraam eesti mat selts"] = "Aastaraamat. Eesti Matemaatika Selts",
  ["aastaraam eesti tead seltsi rootsis"] = "Aastaraamat - Eesti Teadusliku Seltsi Rootsis. Annales - Societatis Litterarum Estonicae in Svecia. Eesti Teaduslik Selts Rootsis",
  ["aatcc j res"] = "AATCC Journal of Research",
  ["aatcc rev"] = "AATCC review",
  ["aaup bull"] = "AAUP bulletin : quarterly publication of the American Association of University Professors",
  ["ab bookm wkly"] = "AB bookman's weekly : for the specialist book world",
  ["aba health esource"] = "ABA health eSource",
  ["aba j"] = "ABA journal",
  ["abb rev"] = "ABB Review",
  ["abc decor"] = "ABC décor",
  ["abdom imaging"] = "Abdominal imaging",
  ["abdom imaging (2013)"] = "Abdominal imaging : computation and clinical applications : 5th International Workshop, held in conjunction with MICCAI 2013, Nagoya, Japan, September 22, 2013 : proceedings. Abdominal Imaging (Workshop) (5th : 2013 : Nagoya-shi, Japan)",
  ["abdom imaging (2014)"] = "Abdominal imaging : computational and clinical applications : 6th International Workshop, ABDI 2014, held in conjunction with MICCAI 2014, Cambridge, MA, USA, September 14, 2014. ABDI (Workshop) (6th : 2014 : Cambridge, Mass.)",
  ["abdom radiol (ny)"] = "Abdominal radiology (New York)",
  ["abdom surg"] = "Abdominal surgery",
  ["abel symp"] = "Abel Symposia",
  ["aberdeen univ rev"] = "Aberdeen University review",
  ["abh akad wiss ddr"] = "Abhandlungen der Akademie der Wissenschaften der DDR",
  ["abh akad wiss ddr, abt math, naturwiss, tech"] = "Abhandlungen der Akademie der Wissenschaften der DDR, Abteilung Mathematik, Naturwissenschaften, Technik",
  ["abh akad wiss gottingen neue folge"] = "Abhandlungen der Akademie der Wissenschaften zu Gottingen. Neue Folge",
  ["abh akad wiss hamburg"] = "Abhandlungen der Akademie der Wissenschaften in Hamburg",
  ["abh ber dtsch mus"] = "Abhandlungen und Berichte (Deutsches Museum (Germany))",
  ["abh ber staatl mus voelkerkd dres"] = "Abhandlungen und Berichte des Staatlichen Museums für Völkerkunde, Dresden. Staatliches Museum für Völkerkunde Dresden",
  ["abh braunschweig wiss ges"] = "Abhandlungen der Braunschweigischen Wissenschaftlichen Gesellschaft",
  ["abh geistes sozwiss kl  akad wiss lit"] = "Abhandlungen der Geistes- und Sozialwissenschaftlichen Klasse",
  ["abh gesch med naturwiss"] = "Abhandlungen zur Geschichte der Medizin und der Naturwissenschaften",
  ["abh math sem univ hamburg"] = "Abhandlungen aus dem Mathematischen Seminar der Universität Hamburg",
  ["abh math semin univ hambg"] = "Abhandlungen aus dem Mathematischen Seminar der Universitat Hamburg",
  ["abh math semin univ hamburg"] = "Abhandlungen aus dem Mathematischen Seminar der Universität Hamburg",
  ["abh math-naturwiss kl akad wiss lit mainz"] = "Abhandlungen der Mathematisch-Naturwissenschaftlichen Klasse",
  ["abh natforsch ges zür"] = "Abhandlungen der Naturforschenden Gesellschaft in Zürich",
  ["abh natwiss ver bremen"] = "Abhandlungen des Naturwissenschaftlichen Vereins zu Bremen",
  ["abh philos gesch"] = "Abhandlungen zur Philosophie und ihrer Geschichte",
  ["abh sachs akad wiss leipzig math-natwiss kl"] = "Abhandlungen der Sachsischen Akademie der Wissenschaften zu Leipzig. Mathematisch-Naturwissenschaftliche Klasse",
  ["abh senckb naturforsch ges"] = "Abhandlungen der Senckenbergischen Naturforschenden Gesellschaft",
  ["abh westfal mus naturkunde"] = "Abhandlungen aus dem Westfälischen Museum für Naturkunde",
  ["abh westfäl mus natkd"] = "Abhandlungen aus dem Westfälischen Museum für Naturkunde",
  ["abhandlung zur erlangung der wurde eines doktors der naturwissenschaften der eidgenossischen technischen hochschule"] = "Abhandlung zur Erlangung der Wurde eines Doktors der Naturwissenschaften der Eidgenossischen Technischen Hochschule Zurich",
  ["abhandlung zur erlangung der wurde eines doktors der technischen wissenschaften der eidgenossischen technischen hoch"] = "Abhandlung zur Erlangung der Wurde eines Doktors der technischen Wissenschaften der Eidgenossischen Technischen Hochschule Zurich",
  ["abi tech"] = "ABI Technik",
  ["abnf j"] = "The ABNF journal : official journal of the Association of Black Nursing Faculty in Higher Education, Inc",
  ["aboric j"] = "Arboricultural Journal",
  ["aborig health bull"] = "Aboriginal health bulletin",
  ["aborig hist"] = "Aboriginal history",
  ["aborig isl health work j"] = "Aboriginal And Islander Health Worker Journal",
  ["abort res notes"] = "Abortion research notes",
  ["abortion rev"] = "Abortion Review",
  ["abr nahain"] = "Abr-Nahrain : an annual under the auspices of the Department of Semitic Studies, University of Melbourne",
  ["abstr 10th conf retrovir oppor infect feb 10 14 2003 hynes conv cent boston mass usa conf retrovir oppor infect 10th 2003 boston mass"] = "Abstracts 10th Conference on Retroviruses and Opportunistic Infections, February 10-14, 2003, Hynes Convention Center, Boston, Massachusetts, USA. Conference on Retroviruses and Opportunistic Infections (10th : 2003: Boston, Mass.)",
  ["abstr annu meet am soc microbiol"] = "Abstracts of the Annual meeting of the American Society for Microbiology",
  ["abstr appl anal"] = "Abstract and Applied Analysis",
  ["abstr book 2003 natl hiv prev conf july 27 30 2003 hyatt regency atlanta hotel atlanta ga natl hiv prev conf 2003 atlanta ga"] = "Abstract book, 2003 National HIV Prevention Conference, July 27-30, 2003, Hyatt Regency Atlanta Hotel, Atlanta, Georgia. National HIV Prevention Conference (2003 : Atlanta, Ga.)",
  ["abstr book 2005 natl hiv prev conf june 12 15 2005 hyatt regency atlanta hotel atlanta ga natl hiv prev conf 2005 atlanta ga"] = "Abstract book, 2005 National HIV Prevention Conference, June 12-15, 2005, Hyatt Regency Atlanta Hotel, Atlanta, Georgia. National HIV Prevention Conference (2005 : Atlanta, Ga.)",
  ["abstr bot"] = "Abstracta Botanica",
  ["abstr int congr trop med malar"] = "Abstracts. International Congress on Tropical Medicine and Malaria (4th : 1948 : Washington, D. C.)",
  ["abstr pap am chem soc"] = "Abstracts of papers - American Chemical Society",
  ["abstr search tools 1999 natl hiv prev conf natl hiv prev conf 1999 atlanta ga"] = "Abstract search tools (for the 1999 National HIV Prevention Conference). National HIV Prevention Conference (1999 : Atlanta, Ga.)",
  ["abstr soc neurosci"] = "Abstracts - Society for Neuroscience. Society for Neuroscience",
  ["abstr translat sci library"] = "Abstracts and translations from the science library. Institute of Living",
  ["abstracts amer math soc"] = "Abstracts of Papers Presented to the American Mathematical Society",
  ["aca trans"] = "ACA Transactions",
  ["acad bookman"] = "The Academy bookman",
  ["acad bras filos"] = "Academia Brasileira de Filosofia",
  ["acad emerg med"] = "Academic emergency medicine : official journal of the Society for Academic Emergency Medicine",
  ["acad forensic pathol"] = "Academic forensic pathology",
  ["acad j agric res"] = "Academia journal of agricultural research",
  ["acad j pediatr neonatol"] = "Academic journal of pediatric and neonatology",
  ["acad j xian jiaotong univ"] = "Academic journal of Xia̕n Jiaotong University : AJXJTU = Xi'an jiao tong da xue xue bao (Yi wen ban)",
  ["acad manag ann"] = "The Academy of Management annals",
  ["acad manag perspect"] = "The Academy of Management perspectives",
  ["acad manage j"] = "Academy of Management journal. Academy of Management",
  ["acad manage rev"] = "Academy of management review. Academy of Management",
  ["acad med"] = "Academic medicine : journal of the Association of American Medical Colleges",
  ["acad news"] = "Academy news (Milwaukee, Wis.)",
  ["acad nurse"] = "The Academic nurse : the journal of the Columbia University School of Nursing",
  ["acad pathol"] = "Academic pathology",
  ["acad pediatr"] = "Academic pediatrics",
  ["acad peru cir"] = "Academia Peruana de Cirugia",
  ["acad poche"] = "L’Academie en Poche",
  ["acad psychiatry"] = "Academic psychiatry : the journal of the American Association of Directors of Psychiatric Residency Training and the Association for Academic Psychiatry",
  ["acad radiol"] = "Academic radiology",
  ["acad rev calif acad periodontol"] = "Academy review of the California Academy of Periodontology, United States Section, ARPA Internationale",
  ["acad roy belg bull cl sci (6)"] = "Académie Royale de Belgique",
  ["acad roy belg cl sci mém collect 8$\\sp \\rm o$ (3)"] = "Académie Royale de Belgique",
  ["acad serbe sci arts glas"] = "Académie Serbe des Sciences et des Arts",
  ["acad today"] = "The Academy today : advancing orthotic and prosthetic care through knowledge",
  ["academia econ pap"] = "Academia Economic Papers",
  ["academic press ser cognition and perception"] = "Academic Press Series in Cognition and Perception",
  ["academic psychol bull"] = "Academic psychology bulletin",
  ["acc chem res"] = "Accounts of chemical research",
  ["acc curr j rev"] = "ACC current journal review",
  ["acc mater res"] = "Accounts of materials research",
  ["acc rev"] = "Accounting Review",
  ["accad bibl ital"] = "Accademie e biblioteche d'Italia",
  ["accad medica"] = "Accademia medica",
  ["accad naz lincei rome"] = "Accademia nazionale dei Lincei. Fondazione Leone Caetani",
  ["accao med"] = "Acção médica",
  ["accel phys"] = "Accelerator Physics",
  ["access microbiol"] = "Access microbiology",
  ["accid anal prev"] = "Accident; analysis and prevention",
  ["accid emerg nurs"] = "Accident and emergency nursing",
  ["account hist j"] = "The Accounting historians journal",
  ["account res"] = "Accountability in research",
  ["accredit qual assur"] = "Accreditation and Quality Assurance",
  ["accès sci"] = "Accès Sciences",
  ["ace q"] = "ACE quarterly",
  ["aces bull"] = "The ACES bulletin. Association for Comparative Economic Studies (U.S.)",
  ["acetic acid bact"] = "Acetic acid bacteria",
  ["acg case rep j"] = "ACG case reports journal",
  ["ach - models chem"] = "ACH - Models in Chemistry",
  ["achiev life sci"] = "Achievements in the Life Sciences",
  ["achiev solut mech eng ii (2019)"] = "Achievements and solutions in mechanical engineering II : selected, peer reviewed papers from the 5th International Conference of Mechanical Engineering (ICOME) 2019, October 24-25, 2019, Craiova, Romania. ICOME (Conference) (5th : 2019 : Craiova, Romania)",
  ["aci 2017 improv relat (2017)"] = "ACI2017, improving relations : Fourth International Conference on Animal-Computer Interaction : proceedings : 21-23 November 2017, Milton Keynes, United Kingdom. International Conference on Animal-Computer Interaction (4th : 2017 : Milton Keynes, England)",
  ["aci mater j"] = "ACI Materials Journal",
  ["aci open"] = "ACI open",
  ["aci struct j"] = "ACI structural journal",
  ["acm ai lett"] = "ACM AI Letters",
  ["acm commun comput algebra"] = "ACM Communications in Computer Algebra",
  ["acm comput surv"] = "ACM computing surveys",
  ["acm conf bioinform comput biol biomed inform (2013)"] = "2013 ACM Conference on Bioinformatics, Computational Biology and Biomedical Informatics : ACM - BCB 2013 : Washington, D.C., U.S.A., September 22 - 25, 2013. ACM Conference on Bioinformatics, Computational Biology and Biomedical Informatics (4th : 2013 : Washington, D.C.)",
  ["acm digital threats: res pract"] = "ACM Digital Threats: Research and Practice",
  ["acm disting theses"] = "ACM Distinguished Theses",
  ["acm inroads"] = "ACM Inroads",
  ["acm int conf bioinform comput biol (2010)"] = "The 2010 ACM International Conference on Bioinformatics and Computational Biology : ACM-BCB 2010 : Niagara Falls, New York, U.S.A., August 2-4, 2010. ACM International Conference on Bioinformatics and Computational Biology (1st : 2010 : Niagara Falls, N.Y.)",
  ["acm j auton transp syst"] = "ACM Journal on Autonomous Transportation Systems",
  ["acm j comput cult heritage"] = "ACM Journal on Computing and Cultural Heritage",
  ["acm j comput sustainable soc"] = "ACM Journal on Computing and Sustainable Societies",
  ["acm j data inf qual"] = "ACM Journal of Data and Information Quality",
  ["acm j emerg technol comput syst"] = "ACM journal on emerging technologies in computing systems",
  ["acm j emerging technol comput"] = "ACM Journal on Emerging Technologies in Computing",
  ["acm j emerging technol comput syst"] = "ACM Journal on Emerging Technologies in Computing Systems",
  ["acm j exp algorithmics"] = "ACM Journal of Experimental Algorithmics",
  ["acm j responsible comput"] = "ACM Journal on Responsible Computing",
  ["acm queue"] = "ACM queue : tomorrow's computing today",
  ["acm sigapp appl comput rev"] = "ACM SIGAPP applied computing review : a publication of the Special Interest Group on Applied Computing",
  ["acm trans access comput"] = "ACM transactions on accessible computing",
  ["acm trans accessible comput"] = "ACM Transactions on Accessible Computing",
  ["acm trans ai sci"] = "ACM Transactions on AI for Science",
  ["acm trans algorithms"] = "ACM Transactions on Algorithms",
  ["acm trans appl percept"] = "ACM transactions on applied perception",
  ["acm trans archit code optim"] = "ACM Transactions on Architecture and Code Optimization",
  ["acm trans asian lang inf process"] = "ACM Transactions on Asian Language Information Processing",
  ["acm trans asian low-resour lang inf process"] = "ACM Transactions on Asian and Low-Resource Language Information Processing",
  ["acm trans auton adapt syst"] = "ACM Transactions on Autonomous and Adaptive Systems",
  ["acm trans comput educ"] = "ACM Transactions on Computing Education",
  ["acm trans comput healthc"] = "ACM transactions on computing for healthcare",
  ["acm trans comput healthcare"] = "ACM Transactions on Computing for Healthcare",
  ["acm trans comput hum interact"] = "ACM transactions on computer-human interaction : a publication of the Association for Computing Machinery",
  ["acm trans comput log"] = "ACM Transactions on Computational Logic",
  ["acm trans comput syst"] = "ACM Transactions on Computer Systems",
  ["acm trans comput theory"] = "ACM Transactions on Computation Theory",
  ["acm trans comput-hum interact"] = "ACM Transactions on Computer-Human Interaction",
  ["acm trans cyber-phys syst"] = "ACM Transactions on Cyber-Physical Systems",
  ["acm trans database syst"] = "ACM Transactions on Database Systems",
  ["acm trans des autom electron syst"] = "ACM Transactions on Design Automation of Electronic Systems",
  ["acm trans econ comput"] = "ACM Transactions on Economics and Computation",
  ["acm trans embed comput syst"] = "ACM transactions on embedded computing systems : TECS",
  ["acm trans embedded comput syst"] = "ACM Transactions on Embedded Computing Systems",
  ["acm trans evol learn"] = "ACM Transactions on Evolutionary Learning",
  ["acm trans evol learn optim"] = "ACM Transactions on Evolutionary Learning and Optimization",
  ["acm trans graph"] = "ACM transactions on graphics",
  ["acm trans graphics"] = "ACM Transactions on Graphics",
  ["acm trans hum robot interact"] = "ACM transactions on human-robot interaction",
  ["acm trans hum-rob interact"] = "ACM Transactions on Human-Robot Interaction",
  ["acm trans inf syst"] = "ACM transactions on information systems",
  ["acm trans inf syst secur"] = "ACM Transactions on Information and System Security",
  ["acm trans intell syst technol"] = "ACM Transactions on Intelligent Systems and Technology",
  ["acm trans interact intell syst"] = "ACM Transactions on Interactive Intelligent Systems",
  ["acm trans internet technol"] = "ACM transactions on Internet technology",
  ["acm trans internet things"] = "ACM Transactions on Internet of Things",
  ["acm trans knowl discov data"] = "ACM transactions on knowledge discovery from data",
  ["acm trans knowl discovery data"] = "ACM Transactions on Knowledge Discovery from Data",
  ["acm trans manag inf syst"] = "ACM transactions on management information systems",
  ["acm trans manage inf syst"] = "ACM Transactions on Management Information Systems",
  ["acm trans math softw"] = "ACM transactions on mathematical software. Association for Computing Machinery",
  ["acm trans math software"] = "ACM Transactions on Mathematical Software",
  ["acm trans model comput simul"] = "ACM Transactions on Modeling and Computer Simulation",
  ["acm trans model perform eval comput syst"] = "ACM Transactions on Modeling and Performance Evaluation of Computing Systems",
  ["acm trans multimedia comput commun appl"] = "ACM Transactions on Multimedia Computing Communications and Applications",
  ["acm trans parallel comput"] = "ACM Transactions on Parallel Computing",
  ["acm trans privacy secur"] = "ACM Transactions on Privacy and Security",
  ["acm trans probab mach learn"] = "ACM Transactions on Probabilistic Machine Learning",
  ["acm trans program lang syst"] = "ACM Transactions on Programming Languages and Systems",
  ["acm trans quantum comput"] = "ACM Transactions on Quantum Computing",
  ["acm trans recommender syst"] = "ACM Transactions on Recommender Systems",
  ["acm trans reconfigurable technol syst"] = "ACM Transactions on Reconfigurable Technology and Systems",
  ["acm trans sens netw"] = "ACM transactions on sensor networks",
  ["acm trans social comput"] = "ACM Transactions on Social Computing",
  ["acm trans software eng methodol"] = "ACM Transactions on Software Engineering and Methodology",
  ["acm trans spatial algorithms syst"] = "ACM Transactions on Spatial Algorithms and Systems",
  ["acm trans speech lang process"] = "ACM Transactions on Speech and Language Processing",
  ["acm trans storage"] = "ACM Transactions on Storage",
  ["acm trans web"] = "ACM Transactions on the Web",
  ["acm transact des autom electron syst"] = "ACM transactions on design automation of electronic systems",
  ["acm/ims j data sci"] = "ACM/IMS Journal of Data Science",
  ["acm/ims trans data sci"] = "ACM/IMS Transactions on Data Science",
  ["acog clin rev"] = "ACOG clinical review",
  ["acog comm opin"] = "ACOG committee opinion",
  ["acog educ bull"] = "ACOG educational bulletin",
  ["acog nurse bull"] = "ACOG nurses bulletin",
  ["acog tech bull"] = "ACOG technical bulletin",
  ["acoust aust"] = "Acoustics Australia",
  ["acoust imaging"] = "Acoustical Imaging",
  ["acoust phys"] = "Acoustical Physics",
  ["acoust res lett online"] = "Acoustics research letters online : ARLO",
  ["acoust sci technol"] = "Acoustical Science and Technology",
  ["acoust today"] = "Acoustics Today",
  ["acoustics (basel)"] = "Acoustics (Basel, Switzerland)",
  ["acp j club"] = "ACP journal club",
  ["acquir immunodefic syndr (aids) wkly surveill rep"] = "Acquired immunodeficiency syndrome (AIDS) weekly surveillance report",
  ["acquis med recent"] = "Acquisitions médicales récentes",
  ["acr open rheumatol"] = "ACR open rheumatology",
  ["acrh rep"] = "ACRH [reports]. U.S. Atomic Energy Commission",
  ["across board (n y)"] = "Across the Board",
  ["across board (ny)"] = "Across the board",
  ["acs agric sci technol"] = "ACS Agricultural Science & Technology",
  ["acs appl bio mater"] = "ACS applied bio materials",
  ["acs appl electron mater"] = "ACS applied electronic materials",
  ["acs appl energy mater"] = "ACS Applied Energy Materials",
  ["acs appl eng mater"] = "ACS Applied Engineering Materials",
  ["acs appl mater interfaces"] = "ACS Applied Materials & Interfaces",
  ["acs appl nano mater"] = "ACS Applied Nano Materials",
  ["acs appl opt mater"] = "ACS Applied Optical Materials",
  ["acs appl polym mater"] = "ACS applied polymer materials",
  ["acs bio & med chem au"] = "ACS Bio & Med Chem Au",
  ["acs biomater sci eng"] = "ACS biomaterials science & engineering",
  ["acs case rev surg"] = "ACS case reviews in surgery",
  ["acs catal"] = "ACS Catalysis",
  ["acs cent sci"] = "ACS central science",
  ["acs chem biol"] = "ACS chemical biology",
  ["acs chem health saf"] = "ACS Chemical Health and Safety",
  ["acs chem neurosci"] = "ACS chemical neuroscience",
  ["acs comb sci"] = "ACS combinatorial science",
  ["acs earth space chem"] = "ACS earth & space chemistry",
  ["acs electrochem"] = "ACS Electrochemistry",
  ["acs energy lett"] = "ACS Energy Letters",
  ["acs eng au"] = "ACS Engineering Au",
  ["acs environ au"] = "ACS Environmental Au",
  ["acs es&t air"] = "ACS ES&T Air",
  ["acs es&t eng"] = "ACS ES&T Engineering",
  ["acs es&t water"] = "ACS ES&T Water",
  ["acs food sci technol"] = "ACS Food Science & Technology",
  ["acs infect dis"] = "ACS infectious diseases",
  ["acs macro lett"] = "ACS Macro Letters",
  ["acs mater au"] = "ACS Materials Au",
  ["acs mater lett"] = "ACS materials letters",
  ["acs meas sci au"] = "ACS Measurement Science Au",
  ["acs med chem lett"] = "ACS medicinal chemistry letters",
  ["acs nano"] = "ACS nano",
  ["acs nanosci au"] = "ACS Nanoscience Au",
  ["acs omega"] = "ACS omega",
  ["acs org inorg au"] = "ACS Organic & Inorganic Au",
  ["acs pharmacol transl sci"] = "ACS pharmacology & translational science",
  ["acs photonics"] = "ACS Photonics",
  ["acs phys chem au"] = "ACS Physical Chemistry Au",
  ["acs polym au"] = "ACS Polymers Au",
  ["acs sens"] = "ACS sensors",
  ["acs sustain chem eng"] = "ACS sustainable chemistry & engineering",
  ["acs sustainable chem eng"] = "ACS Sustainable Chemistry & Engineering",
  ["acs sustainable resour manage"] = "ACS Sustainable Resource Management",
  ["acs symp ser"] = "ACS Symposium Series",
  ["acs symp ser am chem soc"] = "ACS symposium series. American Chemical Society",
  ["acs synth biol"] = "ACS synthetic biology",
  ["acsms health fit j"] = "ACSM's health & fitness journal",
  ["act acad ab"] = "Acta Academiae Aboensis : Ser B Math Et Phys",
  ["act adapt aging"] = "Activities, adaptation & aging",
  ["act dir q alzheimers other dement patients"] = "Activities directors' quarterly for Alzheimer's & other dementia patients",
  ["act hepato"] = "Acta hepatologica",
  ["act nerv super (2007)"] = "Activitas nervosa superior (2007)",
  ["act nerv super (praha)"] = "Activitas nervosa superior",
  ["act passive electron compon"] = "Active and Passive Electronic Components",
  ["acta acad abo ser b"] = "Acta Academiae Aboensis",
  ["acta acad med wuhan"] = "Acta Academiae Medicinae Wuhan = Wu-han i hsüeh yüan hsüeh pao",
  ["acta acad sci pol"] = "Acta Academiae Scientiarum Polonae",
  ["acta acad sci taurinensis"] = "Acta Academiae Scientiarum Taurinensis",
  ["acta academiae medicinae jiangxi|jiangxi yixueyuan xuebao"] = "Jiangxi yixueyuan xuebao",
  ["acta acust"] = "Acta Acustica",
  ["acta acust united acust"] = "Acta acustica united with acustica : the journal of the European Acoustics Association (EEIG)",
  ["acta adriat"] = "Acta Adriatica",
  ["acta agral fenn"] = "Acta Agralia Fennica",
  ["acta agrar silv, ser lesn (krak)"] = "Acta Agraria et Silvestria, Seria Lesna, Krakau",
  ["acta agric scand a anim sci"] = "Acta agriculturæ Scandinavica. Section A, Animal science",
  ["acta agric scand b soil plant sci"] = "Acta agriculturæ Scandinavica. Section B, Soil and plant science",
  ["acta agric scand sect a"] = "Acta Agriculturae Scandinavica Section A: Animal Science",
  ["acta agric scand sect b"] = "Acta Agriculturae Scandinavica Section B: Soil and Plant Science",
  ["acta agric slov"] = "Acta agriculturae slovenica",
  ["acta agriculturae universitatis chekianensis|zhejiang nong ye da xue xue bao"] = "Zhejiang nong ye da xue xue bao",
  ["acta agriculturae universitatis jilinensis|jilin nong ye da xue xue bao"] = "Jilin nong ye da xue xue bao",
  ["acta agrobot"] = "Acta agrobotanica",
  ["acta agron"] = "Acta agronomica",
  ["acta agron hung"] = "Acta Agronomica Hungarica",
  ["acta agron sin"] = "Acta Agronomica Sinica",
  ["acta aliment"] = "Acta alimentaria",
  ["acta allergol"] = "Acta allergologica",
  ["acta allergol suppl (copenh)"] = "Acta allergologica. Supplementum",
  ["acta amazon"] = "Acta amazonica",
  ["acta amazonica"] = "Acta Amazonica",
  ["acta anaesthesiol"] = "Acta anaesthesiologica",
  ["acta anaesthesiol belg"] = "Acta anaesthesiologica Belgica",
  ["acta anaesthesiol ital"] = "Acta anaesthesiologica italica",
  ["acta anaesthesiol scand"] = "Acta anaesthesiologica Scandinavica",
  ["acta anaesthesiol scand suppl"] = "Acta anaesthesiologica Scandinavica. Supplementum",
  ["acta anaesthesiol sin"] = "Acta anaesthesiologica Sinica",
  ["acta anaesthesiol taiwan"] = "Acta anaesthesiologica Taiwanica : official journal of the Taiwan Society of Anesthesiologists",
  ["acta anal"] = "Acta analytica : philosophy and psychology",
  ["acta anal funct appl"] = "Acta Analysis Functionalis Applicata. AAFA. Yingyong Fanhanfenxi Xuebao",
  ["acta anat (basel)"] = "Acta anatomica",
  ["acta anat suppl (basel)"] = "Acta anatomica. Supplementum",
  ["acta anthropogenet"] = "Acta anthropogenetica",
  ["acta antiq hung"] = "Acta antiqua Academiae Scientiarum Hungaricae",
  ["acta appl math"] = "Acta Applicandae Mathematicae",
  ["acta arachnol"] = "Acta arachnologica",
  ["acta archaeol"] = "Acta Archaeologica",
  ["acta argent fisiol fisiopatol"] = "Acta argentina de fisiología y fisiopatología",
  ["acta arith"] = "Acta Arithmetica",
  ["acta astron"] = "Acta Astronomica",
  ["acta astronaut"] = "Acta astronautica",
  ["acta astronom sinica"] = "Acta Astronomica Sinica",
  ["acta astrophys sinica"] = "Acta Astrophysica Sinica",
  ["acta autom sin"] = "Acta Automatica Sinica",
  ["acta automat sinica"] = "Acta Automatica Sinica",
  ["acta balt"] = "Acta Baltica",
  ["acta belg arte med pharm mil"] = "Acta Belgica de arte medicinali et pharmaceutica militari",
  ["acta belg hist med"] = "Acta Belgica historiae medicinae : official journal of the Belgian Association for the History of Medicine",
  ["acta belg med phys"] = "Acta Belgica. Medica physica : organe officiel de la Société royale belge de médecine physique et de réhabilitation",
  ["acta bern"] = "Acta Bernensia",
  ["acta biochim biophys acad sci hung"] = "Acta biochimica et biophysica; Academiae Scientiarum Hungaricae",
  ["acta biochim biophys hung"] = "Acta biochimica et biophysica Hungarica",
  ["acta biochim biophys sin"] = "Acta Biochimica et Biophysica Sinica",
  ["acta biochim biophys sin (shanghai)"] = "Acta biochimica et biophysica Sinica",
  ["acta biochim pol"] = "Acta Biochimica Polonica",
  ["acta bioeng biomech"] = "Acta of bioengineering and biomechanics",
  ["acta bioeth"] = "Acta bioethica",
  ["acta biol acad sci hung"] = "Acta biologica Academiae Scientiarum Hungaricae",
  ["acta biol crac ser bot"] = "Acta biologica Cracoviensia. Series: Botanica",
  ["acta biol exp (warsz)"] = "Acta biologiae experimentalis",
  ["acta biol hung"] = "Acta Biologica Hungarica",
  ["acta biol marisiensis"] = "Acta biologica marisiensis",
  ["acta biol med (gdansk)"] = "Acta biologica et medica",
  ["acta biol med ger"] = "Acta Biologica et Medica Germanica",
  ["acta biol mont"] = "Acta Biologica Montana",
  ["acta biol plant agriensis"] = "Acta biologica plantarum Agriensis : a journal of plant biology",
  ["acta biol sib"] = "Acta biologica Sibirica",
  ["acta biol slov"] = "Acta biologica Slovenica : ABS",
  ["acta biol turc"] = "Acta biologica turcica",
  ["acta biolo colomb"] = "Acta biológica colombiana",
  ["acta biomater"] = "Acta biomaterialia",
  ["acta biomater odontol scand"] = "Acta biomaterialia odontologica Scandinavica",
  ["acta biomed"] = "Acta bio-medica : Atenei Parmensis",
  ["acta biomed ateneo parmense"] = "Acta bio-medica de L'Ateneo parmense : organo della Società di medicina e scienze naturali di Parma",
  ["acta biomed sci"] = "Acta biomedica scientifica",
  ["acta bioquim clin latinoam"] = "Acta Bioquimica Clinica Latinoamericana",
  ["acta biotechnol"] = "Acta biotechnologica",
  ["acta biotheor"] = "Acta biotheoretica",
  ["acta borealia"] = "Acta Borealia",
  ["acta bot brasilica"] = "Acta botanica Brasílica",
  ["acta bot croat"] = "Acta botanica Croatica",
  ["acta bot fenn"] = "Acta Botanica Fennica",
  ["acta bot gall"] = "Acta botanica Gallica : bulletin de la Société botanique de France",
  ["acta bot gallica"] = "Acta botanica Gallica : bulletin de la Societe botanique de France",
  ["acta bot hung"] = "Acta botanica Hungarica",
  ["acta bot mex"] = "Acta botánica mexicana",
  ["acta bot neerl"] = "Acta Botanica Neerlandica",
  ["acta bot sin"] = "Acta botanica Sinica",
  ["acta bot venez"] = "Acta botánica venezuélica",
  ["acta botan croat"] = "Acta Botanica Croatica",
  ["acta botanica boreali-occidentalia sinica|xibei zhiwu xuebao"] = "Xi bei zhi wu xue bao",
  ["acta brevia neerl physiol pharmacol microbiol e a"] = "Acta brevia Neerlandica de physiologia, pharmacologia, microbiologia e.a",
  ["acta cancerol (lima)"] = "Acta cancerológica",
  ["acta cardiol"] = "Acta cardiologica",
  ["acta cardiol sin"] = "Acta Cardiologica Sinica",
  ["acta cardiol suppl"] = "Acta cardiologica. Supplementum",
  ["acta chem scand"] = "Acta Chemica Scandinavica",
  ["acta chem scand a"] = "Acta chemica Scandinavica. Series A: Physical and inorganic chemistry",
  ["acta chem scand b"] = "Acta chemica Scandinavica. Series B: Organic chemistry and biochemistry",
  ["acta chem scand ser a"] = "Acta Chemica Scandinavica, Series A: Physical and Inorganic Chemistry",
  ["acta chem scand ser b"] = "Acta Chemica Scandinavica, Series B: Organic Chemistry and Biochemistry",
  ["acta chim hung"] = "Acta Chimica Hungarica",
  ["acta chim sin"] = "Acta Chimica Sinica",
  ["acta chim sinica"] = "Acta Chimica Sinica",
  ["acta chim slov"] = "Acta chimica Slovenica",
  ["acta chim slovaca"] = "Acta Chimica Slovaca",
  ["acta chimi sin"] = "Acta Chimica Sinica",
  ["acta chir acad sci hung"] = "Acta chirurgica Academiae Scientiarum Hungaricae",
  ["acta chir austriaca"] = "Acta chirurgica Austriaca",
  ["acta chir belg"] = "Acta chirurgica Belgica",
  ["acta chir hung"] = "Acta chirurgica Hungarica",
  ["acta chir ital"] = "Acta chirurgica Italica",
  ["acta chir iugosl"] = "Acta chirurgica Iugoslavica",
  ["acta chir orthop traumatol cech"] = "Acta chirurgiae orthopaedicae et traumatologiae Cechoslovaca",
  ["acta chir patav"] = "Acta chirurgica Patavina",
  ["acta chir plast"] = "Acta chirurgiae plasticae",
  ["acta chir scand"] = "Acta chirurgica Scandinavica",
  ["acta chir scand suppl"] = "Acta chirurgica Scandinavica. Supplementum",
  ["acta chiropt"] = "Acta chiropterologica",
  ["acta chiropterol"] = "Acta Chiropterologica",
  ["acta chromatogr"] = "Acta Chromatographica",
  ["acta cienc indica math"] = "Acta Ciencia Indica",
  ["acta cienc indica, chem"] = "Acta Ciencia Indica Chemistry",
  ["acta cienc indica, phys"] = "Acta Ciencia Indica Phyics",
  ["acta cient venez"] = "Acta Cientifica Venezolana",
  ["acta cient venezolana"] = "Acta Cientifica Venezolana. Asociacion Venezolana para el Avance de la Ciencia",
  ["acta cir bras"] = "Acta cirúrgica brasileira",
  ["acta cl"] = "Acta classica",
  ["acta clin belg"] = "Acta clinica Belgica",
  ["acta clin belg suppl"] = "Acta clinica Belgica. Supplementum",
  ["acta clin croat"] = "Acta clinica Croatica",
  ["acta clin odontol"] = "Acta Clinica Odontologica",
  ["acta comment univ tartu math"] = "Acta et Commentationes Universitatis Tartuensis de Mathematica",
  ["acta cryst sect a"] = "Acta Crystallographica",
  ["acta crystallogr"] = "Acta crystallographica",
  ["acta crystallogr a"] = "Acta crystallographica. Section A, Crystal physics, diffraction, theoretical and general crystallography",
  ["acta crystallogr a found adv"] = "Acta crystallographica. Section A, Foundations and advances",
  ["acta crystallogr b"] = "Acta crystallographica. Section B: Structural crystallography and crystal chemistry",
  ["acta crystallogr b struct sci cryst eng mater"] = "Acta crystallographica Section B, Structural science, crystal engineering and materials",
  ["acta crystallogr c"] = "Acta crystallographica. Section C, Crystal structure communications",
  ["acta crystallogr c struct chem"] = "Acta crystallographica. Section C, Structural chemistry",
  ["acta crystallogr d biol crystallogr"] = "Acta crystallographica. Section D, Biological crystallography",
  ["acta crystallogr d struct biol"] = "Acta crystallographica. Section D, Structural biology",
  ["acta crystallogr e crystallogr commun"] = "Acta crystallographica. Section E, Crystallographic communications",
  ["acta crystallogr f struct biol commun"] = "Acta crystallographica. Section F, Structural biology communications",
  ["acta crystallogr sec a"] = "Acta Crystallographica, Section A: Crystal Physics, Diffraction,Theoretical and General Crystallography",
  ["acta crystallogr sec b"] = "Acta Crystallographica, Section B: Structural Crystallography and Crystal Chemistry",
  ["acta crystallogr sect a"] = "Acta Crystallographica Section A. Foundations and Advances",
  ["acta crystallogr sect e struct rep online"] = "Acta crystallographica. Section E, Structure reports online",
  ["acta crystallogr sect f struct biol cryst commun"] = "Acta crystallographica. Section F, Structural biology and crystallization communications",
  ["acta crystallogr, sect a: found adv"] = "Acta Crystallographica, Section A: Foundations and Advances",
  ["acta crystallogr, sect a: found crystallogr"] = "Acta Crystallographica Section A: Foundations",
  ["acta crystallogr, sect b: struct sci"] = "Acta Crystallographica Section B: Structural Science",
  ["acta crystallogr, sect b: struct sci, cryst eng mater"] = "Acta Crystallographica, Section B: Structural Science, Crystal Engineering and Materials",
  ["acta crystallogr, sect c: cryst struct commun"] = "Acta Crystallographica Section C: Crystal Structure Communications",
  ["acta crystallogr, sect c: struct chem"] = "Acta Crystallographica, Section C: Structural Chemistry",
  ["acta crystallogr, sect d: biol crystallogr"] = "Acta Crystallographica Section D: Biological Crystallography",
  ["acta crystallogr, sect e: crystallogr commun"] = "Acta Crystallographica, Section E: Crystallographic Communications",
  ["acta crystallogr, sect e: struct rep online"] = "Acta Crystallographica Section E Structure Reports Online",
  ["acta crystallogr, sect f: struct biol commun"] = "Acta Crystallographica, Section F: Structural Biology Communications",
  ["acta crystallogr, sect f: struct biol cryst commun"] = "Acta Crystallographica Section F: Structural Biology and Crystallization Communications",
  ["acta crystallograph sect f struct biol cryst commun"] = "Acta Crystallographica. Section F, Structural Biology and Crystallization Communications",
  ["acta cybernet"] = "Acta Cybernetica",
  ["acta cytol"] = "Acta cytologica",
  ["acta demogr"] = "Acta demographica",
  ["acta derm venereol"] = "Acta dermato-venereologica",
  ["acta derm venereol suppl (stockh)"] = "Acta dermato-venereologica. Supplementum",
  ["acta dermatol kyoto engl ed"] = "Acta dermatologica-Kyoto. English edition",
  ["acta dermatovenerol alp pannonica adriat"] = "Acta dermatovenerologica Alpina, Pannonica, et Adriatica",
  ["acta dermatovenerol alp panonica adriat"] = "Acta dermatovenerologica Alpina, Panonica, et Adriatica",
  ["acta dermatovenerol croat"] = "Acta dermatovenerologica Croatica : ADC",
  ["acta diabetol"] = "Acta diabetologica",
  ["acta diabetol lat"] = "Acta diabetologica latina",
  ["acta ecol sin"] = "Acta Ecologica Sinica",
  ["acta embryol exp (palermo)"] = "Acta embryologiae experimentalis",
  ["acta embryol morphol exp"] = "Acta embryologiae et morphologiae experimentalis",
  ["acta endocrinol (buchar)"] = "Acta endocrinologica (Bucharest, Romania : 2005)",
  ["acta endocrinol (copenh)"] = "Acta endocrinologica",
  ["acta endocrinol iber"] = "Acta endocrinologica Iberica",
  ["acta endocrinol panam"] = "Acta Endocrinologica Panamericana",
  ["acta endocrinol suppl (copenh)"] = "Acta endocrinologica. Supplementum",
  ["acta endoscopica"] = "Acta endoscopica",
  ["acta entomol mus natl pragae"] = "Acta Entomologica Musei Nationalis Pragae",
  ["acta ethnol linguist"] = "Acta ethnologica et linguistica",
  ["acta ethol"] = "Acta ethologica",
  ["acta eur fertil"] = "Acta Europaea fertilitatis",
  ["acta fac pharm univ comen"] = "Acta Facultatis Pharmaceuticae Universitatis Comenianae",
  ["acta farm bonaerense"] = "Acta Farmaceutica Bonaerense",
  ["acta for fenn"] = "Acta Forestalia Fennica",
  ["acta gastroenterol belg"] = "Acta gastro-enterologica Belgica",
  ["acta gastroenterol latinoam"] = "Acta gastroenterologica Latinoamericana",
  ["acta genet med gemellol (roma)"] = "Acta geneticae medicae et gemellologiae",
  ["acta genet sin"] = "Acta Genetica Sinica",
  ["acta genet stat med"] = "Acta genetica et statistica medica",
  ["acta genetica sinica|chin j genet"] = "Chinese journal of genetics",
  ["acta geochim"] = "Acta Geochimica",
  ["acta geochimica"] = "Acta geochimica",
  ["acta geod geophys"] = "Acta Geodaetica et Geophysica",
  ["acta geod geophys hung"] = "Acta Geodaetica et Geophysica Hungarica",
  ["acta geodyn geomater"] = "Acta Geodynamics et Geomaterialia",
  ["acta geogr"] = "Acta geographica",
  ["acta geogr croat"] = "Acta geographica Croatica : glasilo Geografskog odjela PMF-a Sveučilišta u Zagrebu",
  ["acta geogr geol meteorol debr geol gemorfol termeszfoldr sor"] = "Acta geographica ac geologica et meteorologica Debrecina. Geológia, geomorfológia, természetföldrajz sorozat",
  ["acta geogr sin"] = "Di li xue bao",
  ["acta geol hung"] = "Acta Geologica Hungarica",
  ["acta geol pol"] = "Acta Geologica Polonica",
  ["acta geol sin"] = "Acta Geologica Sinica",
  ["acta geol slovaca"] = "Acta geologica Slovaca : AGEOS",
  ["acta geophys"] = "Acta Geophysica",
  ["acta geotech"] = "Acta Geotechnica",
  ["acta gerontol (milano)"] = "Acta gerontologica",
  ["acta ginecol (madr)"] = "Acta ginecológica",
  ["acta gynaecol obstet hisp lusit"] = "Acta gynaecologica et obstetrica Hispano-Lusitana",
  ["acta haematol"] = "Acta haematologica",
  ["acta haematol pol"] = "Acta haematologica Polonica",
  ["acta hepatogastroenterol (stuttg)"] = "Acta hepato-gastroenterologica",
  ["acta hepatosplenol"] = "Acta hepato-splenologica",
  ["acta herpetol"] = "Acta herpetologica",
  ["acta hist leopold"] = "Acta Historica Leopoldina",
  ["acta hist leopoldina"] = "Acta historica Leopoldina",
  ["acta hist med pharm vet (beograd)"] = "Acta historica medicinae, pharmaciae, veterinae",
  ["acta hist med vallisoletana monogr"] = "Acta histórico-médica vallisoletana. Monografías",
  ["acta hist rerum nat tech"] = "Acta historiae rerum naturalium necnon technicarum. Special issue",
  ["acta hist sci nat med"] = "Acta historica scientiarum naturalium et medicinalium",
  ["acta histochem"] = "Acta histochemica",
  ["acta histochem cytochem"] = "Acta histochemica et cytochemica",
  ["acta histochem suppl"] = "Acta histochemica. Supplementband",
  ["acta hortic"] = "Acta horticulturae",
  ["acta hortic regiotect"] = "Acta horticulturae et regiotectuare",
  ["acta hosp"] = "Acta hospitalia",
  ["acta hydroch hydrob"] = "Acta Hydrochimica et Hydrobiologica",
  ["acta hydrophys"] = "Acta Hydrophysica",
  ["acta iber radiol cancerol"] = "Acta ibérica radiológica-cancerológica",
  ["acta ichthyol piscat"] = "Acta ichthyologica et piscatoria",
  ["acta imeko (2012)"] = "Acta IMEKO (2012)",
  ["acta inform"] = "Acta Informatica",
  ["acta inform med"] = "Acta informatica medica : AIM : journal of the Society for Medical Informatics of Bosnia & Herzegovina : časopis Društva za medicinsku informatiku BiH",
  ["acta inst for zvolen"] = "Acta Instituti Forestalis Zvolenensis",
  ["acta investig psicol"] = "Acta de investigación psicológica",
  ["acta isot (padova)"] = "Acta isotopica",
  ["acta jurid (cape town)"] = "Acta juridica (Cape Town, South Africa)",
  ["acta leiden"] = "Acta Leidensia",
  ["acta leprol"] = "Acta leprologica",
  ["acta limnol bras"] = "Acta Limnologica Brasiliensia",
  ["acta linguist hung"] = "Acta linguistica Hungarica",
  ["acta manila ser a"] = "Acta Manilana",
  ["acta mater"] = "Acta Materialia",
  ["acta math"] = "Acta Mathematica",
  ["acta math acad paedagog nyhazi"] = "Acta Mathematica. Academiae Paedagogicae Nyiregyhaziensis",
  ["acta math acad paedagog nyhazi (ns)"] = "Acta Mathematica. Academiae Paedagogicae Nyiregyhaziensis. New Series",
  ["acta math acad sci hung"] = "Acta Mathematica Academiae Scientiarum Hungaricae",
  ["acta math appl sin"] = "Acta mathematicae applicatae Sinica (English series)",
  ["acta math appl sin engl ser"] = "Acta Mathematicae Applicatae Sinica. English Series",
  ["acta math appl sinica"] = "Acta Mathematicae Applicatae Sinica",
  ["acta math appl sinica (english ser)"] = "Acta Mathematicae Applicatae Sinica",
  ["acta math hung"] = "Acta Mathematica Hungarica",
  ["acta math hungar"] = "Acta Mathematica Hungarica",
  ["acta math inform univ ostraviensis"] = "Acta Mathematica et Informatica Universitatis Ostraviensis",
  ["acta math sci"] = "Acta mathematica scientia = Shu xue wu li xue bao",
  ["acta math sci (chinese)"] = "Acta Mathematica Scientia",
  ["acta math sci (english ed)"] = "Acta Mathematica Scientia",
  ["acta math sci ser a (chin ed)"] = "Acta Mathematica Scientia. Series A. Shuxue Wuli Xuebao. Chinese Edition",
  ["acta math sci ser a (chinese ed)"] = "Acta Mathematica Scientia. Series A. Shuxue Wuli Xuebao. Chinese Edition",
  ["acta math sci ser b (engl ed)"] = "Acta Mathematica Scientia. Series B. English Edition",
  ["acta math sin"] = "Acta Mathematica Sinica",
  ["acta math sin (engl ser)"] = "Acta Mathematica Sinica (English Series)",
  ["acta math sin engl ser"] = "Acta mathematica Sinica, English series",
  ["acta math sinica"] = "Acta Mathematica Sinica",
  ["acta math sinica (chin ser)"] = "Acta Mathematica Sinica. Chinese Series",
  ["acta math sinica (chinese ser)"] = "Acta Mathematica Sinica. Chinese Series",
  ["acta math sinica (ns)"] = "Acta Mathematica Sinica",
  ["acta math univ comenian (ns)"] = "Acta Mathematica Universitatis Comenianae. New Series",
  ["acta math vietnam"] = "Acta Mathematica Vietnamica",
  ["acta mech"] = "Acta Mechanica",
  ["acta mech sin"] = "Acta Mechanica Sinica",
  ["acta mech sinica (beijing)"] = "Acta Mechanica Sinica",
  ["acta mech sinica (english ed)"] = "Acta Mechanica Sinica",
  ["acta mech solida sin"] = "Acta Mechanica Solida Sinica",
  ["acta med"] = "Acta medica",
  ["acta med acad"] = "Acta medica academica",
  ["acta med acad sci hung"] = "Acta medica Academiae Scientiarum Hungaricae",
  ["acta med austriaca"] = "Acta medica Austriaca",
  ["acta med austriaca suppl"] = "Acta medica Austriaca. Supplement",
  ["acta med auxol (milano)"] = "Acta medica auxologica",
  ["acta med biol (niigata)"] = "Acta medica et biologica",
  ["acta med cathol"] = "Acta medica Catholica",
  ["acta med colomb"] = "Acta médica colombiana : AMC : organo de la Asociación Columbiana de Medicina Interna",
  ["acta med costarric"] = "Acta médica costarricense",
  ["acta med croatica"] = "Acta Medica Croatica",
  ["acta med dom"] = "Acta médica dominicana : revista científica para médicos",
  ["acta med hidalguense"] = "Acta médica hidalguense",
  ["acta med hisp"] = "Acta médica hispánica",
  ["acta med hist adriat"] = "Acta medico-historica adriatica : AMHA",
  ["acta med hist patav"] = "Acta medicae historiae Patavina",
  ["acta med hung"] = "Acta medica Hungarica",
  ["acta med indones"] = "Acta medica Indonesiana",
  ["acta med iran"] = "Acta medica Iranica",
  ["acta med ital mal infett parassit"] = "Acta medica Italica di malattie infettive e parassitarie",
  ["acta med ital med trop subtrop gastroenterol"] = "Acta medica Italica di medicina tropicale e subtropicale e di gastroenterologia",
  ["acta med iugosl"] = "Acta medica Iugoslavica",
  ["acta med kinki univ"] = "Acta medica Kinki University",
  ["acta med leg"] = "Acta medicinae legalis",
  ["acta med leg soc (liege)"] = "Acta medicinae legalis et socialis",
  ["acta med litu"] = "Acta medica Lituanica",
  ["acta med marisiensis"] = "Acta medica marisiensis",
  ["acta med nagasaki"] = "Acta medica Nagasakiensia",
  ["acta med okayama"] = "Acta medica Okayama",
  ["acta med orient"] = "Acta medica Orientalia",
  ["acta med patav"] = "Acta medica Patavina",
  ["acta med patavina suppl"] = "Acta medica Patavina. Supplemento",
  ["acta med philipp"] = "Acta medica Philippina",
  ["acta med pol"] = "Acta medica Polona",
  ["acta med port"] = "Acta médica portuguesa",
  ["acta med rom"] = "Acta medica romana; annali della Facoltà di medicina e chirurgia, Università cattolica del Sacro Cuore",
  ["acta med scand"] = "Acta medica Scandinavica",
  ["acta med scand suppl"] = "Acta medica Scandinavica. Supplementum",
  ["acta med tenerife"] = "Acta médica de Tenerife",
  ["acta med turc"] = "Acta medica Turcica",
  ["acta med valle"] = "Acta médica del valle",
  ["acta med vet (napoli)"] = "Acta medica veterinaria",
  ["acta med vietnam"] = "Acta medica Vietnamica",
  ["acta medica (hradec kralove)"] = "Acta medica (Hradec Králové)",
  ["acta medica (hradec kralove) suppl"] = "Acta Medica (Hradec Kralove). Supplementum",
  ["acta medica cordoba"] = "Acta medica",
  ["acta medica transilvanica"] = "Acta medica transilvanica",
  ["acta metall"] = "Acta Metallurgica",
  ["acta metall mater"] = "Acta Metallurgica et Materialia",
  ["acta metall sin"] = "Acta Metallurgica Sinica",
  ["acta metall sin engl lett"] = "Acta Metallurgica Sinica- English Letters",
  ["acta metall sinica"] = "Acta Metallurgica Sinica",
  ["acta metall slovaca"] = "Acta Metallurgica Slovaca",
  ["acta microbiol acad sci hung"] = "Acta microbiologica Academiae Scientiarum Hungaricae",
  ["acta microbiol bulg"] = "Acta Microbiologica Bulgarica",
  ["acta microbiol hell"] = "Acta Microbiologica Hellenica",
  ["acta microbiol hung"] = "Acta microbiologica Hungarica",
  ["acta microbiol immunol hung"] = "Acta microbiologica et immunologica Hungarica",
  ["acta microbiol pol"] = "Acta microbiologica Polonica",
  ["acta microbiol pol (1952)"] = "Acta microbiologica Polonica (1952)",
  ["acta microbiol pol a"] = "Acta microbiologica Polonica. Series A: Microbiologia generalis",
  ["acta microbiol pol b"] = "Acta microbiologica Polonica. Series B: Microbiologia applicata",
  ["acta microbiol sin"] = "Acta microbiologica Sinica",
  ["acta microbiol virol immunol (sofiia)"] = "Acta microbiologica, virologica et immunologica",
  ["acta micropalaeontologica sinica|wei ti gu sheng wu xue bao"] = "Wei ti gu sheng wu xue bao",
  ["acta montan slovaca"] = "Acta Montanistica Slovaca",
  ["acta morphol acad sci hung"] = "Acta morphologica Academiae Scientiarum Hungaricae",
  ["acta morphol acad sci hung suppl"] = "Acta morphologica Academiae Scientiarum Hungaricae. Supplementum",
  ["acta morphol hung"] = "Acta morphologica Hungarica",
  ["acta morphol neerl scand"] = "Acta morphologica Neerlando-Scandinavica",
  ["acta morphol neerl-scand"] = "Acta Morphologica Neerlando-Scandinavica",
  ["acta mus napocensis"] = "Acta Musei Napocensis",
  ["acta musei sil sci natur"] = "Acta Musei Silesiae. Scientiae naturales",
  ["acta mycol"] = "Acta mycologica",
  ["acta myol"] = "Acta myologica : myopathies and cardiomyopathies : official journal of the Mediterranean Society of Myology",
  ["acta méd costarric"] = "Acta Médica Costarricense",
  ["acta naturae"] = "Acta naturae",
  ["acta neerl morphol norm pathol"] = "Acta Neerlandica morphologiae normalis et pathologicae",
  ["acta neurobiol exp"] = "Acta Neurobiologiae Experimentalis",
  ["acta neurobiol exp (wars)"] = "Acta neurobiologiae experimentalis",
  ["acta neurobiol exp (warsz)"] = "Acta Neurobiologiae Experimentalis",
  ["acta neurochir (wien)"] = "Acta neurochirurgica",
  ["acta neurochir suppl"] = "Acta neurochirurgica. Supplement",
  ["acta neurochir suppl (wien)"] = "Acta neurochirurgica. Supplementum",
  ["acta neurol (napoli)"] = "Acta neurologica",
  ["acta neurol [quad] (napoli)"] = "Acta Neurologica. Quaderni",
  ["acta neurol belg"] = "Acta neurologica Belgica",
  ["acta neurol latinoam"] = "Acta neurológica latinoamericana",
  ["acta neurol psychiatr belg"] = "Acta neurologica et psychiatrica Belgica",
  ["acta neurol quad (napoli)"] = "Acta neurologica. Quaderni",
  ["acta neurol scand"] = "Acta neurologica Scandinavica",
  ["acta neurol scand suppl"] = "Acta neurologica Scandinavica. Supplementum",
  ["acta neurol taiwan"] = "Acta neurologica Taiwanica",
  ["acta neuropathol"] = "Acta neuropathologica",
  ["acta neuropathol (berl)"] = "Acta Neuropathologica",
  ["acta neuropathol commun"] = "Acta neuropathologica communications",
  ["acta neuropathol suppl"] = "Acta neuropathologica. Supplementum",
  ["acta neuropathol suppl (berl)"] = "Acta Neuropathologica. Supplement",
  ["acta neuropsychiatr"] = "Acta neuropsychiatrica",
  ["acta neuropsychol"] = "Acta neuropsychologica : the official journal of the Polish Neuropsychological Society",
  ["acta neuroveg (wien)"] = "Acta neurovegetativa",
  ["acta neuroveg suppl"] = "Acta neurovegetativa. Supplementum",
  ["acta numer"] = "Acta Numerica",
  ["acta numis"] = "Acta numismática",
  ["acta obstet ginecol hisp lusit"] = "Acta obstétrica y ginecológica hispano-lusitana",
  ["acta obstet ginecol hisp lusit suppl"] = "Acta obstetrica y ginecologica hispano-lusitana. Suplemento",
  ["acta obstet gynaecol jpn"] = "Acta obstetrica et gynaecologica Japonica",
  ["acta obstet gynecol scand"] = "Acta obstetricia et gynecologica Scandinavica",
  ["acta obstet gynecol scand suppl"] = "Acta obstetricia et gynecologica Scandinavica. Supplement",
  ["acta oceanolog sin"] = "Acta Oceanologica Sinica",
  ["acta odontol latinoam"] = "Acta odontológica latinoamericana : AOL",
  ["acta odontol pediatr"] = "Acta de odontología pediátrica",
  ["acta odontol scand"] = "Acta odontologica Scandinavica",
  ["acta odontol scand suppl"] = "Acta odontologica Scandinavica. Supplementum",
  ["acta odontol venez"] = "Acta odontológica venezolana",
  ["acta oecol"] = "Acta Oecologica",
  ["acta oecol (montrouge)"] = "Acta oecologica (Montrouge, France)",
  ["acta oecon"] = "Acta oeconomica : periodical of the Hungarian Academy of Sciences",
  ["acta oncol"] = "Acta oncologica (Stockholm, Sweden)",
  ["acta oncol (madr)"] = "Acta oncológica",
  ["acta ophthalmol"] = "Acta ophthalmologica",
  ["acta ophthalmol (copenh)"] = "Acta Ophthalmologica",
  ["acta ophthalmol scand"] = "Acta ophthalmologica Scandinavica",
  ["acta ophthalmol scand suppl"] = "Acta ophthalmologica Scandinavica. Supplement",
  ["acta ophthalmol suppl"] = "Acta ophthalmologica. Supplementum",
  ["acta ophthalmol suppl (oxf )"] = "Acta ophthalmologica. Supplement",
  ["acta ornithol"] = "Acta ornithologica",
  ["acta orthop"] = "Acta orthopaedica",
  ["acta orthop belg"] = "Acta orthopaedica Belgica",
  ["acta orthop scand"] = "Acta orthopaedica Scandinavica",
  ["acta orthop scand suppl"] = "Acta orthopaedica Scandinavica. Supplementum",
  ["acta orthop suppl"] = "Acta orthopaedica. Supplementum",
  ["acta orthop traumatol turc"] = "Acta orthopaedica et traumatologica turcica",
  ["acta ortop bras"] = "Acta ortopedica brasileira",
  ["acta ortop mex"] = "Acta ortopédica mexicana",
  ["acta otolaryngol"] = "Acta oto-laryngologica",
  ["acta otolaryngol (stockh)"] = "Acta Oto-Laryngologica",
  ["acta otolaryngol case rep"] = "Acta oto-laryngologica case reports",
  ["acta otolaryngol orient"] = "Acta oto-laryngologica orientalia",
  ["acta otolaryngol suppl"] = "Acta oto-laryngologica. Supplementum",
  ["acta otolaryngol suppl (stockh)"] = "Acta Oto-Laryngologica. Supplement",
  ["acta otorhinolaryngol belg"] = "Acta oto-rhino-laryngologica Belgica",
  ["acta otorhinolaryngol ital"] = "Acta otorhinolaryngologica Italica : organo ufficiale della Società italiana di otorinolaringologia e chirurgia cervico-facciale",
  ["acta otorinolaryngol iber am"] = "Acta oto-rino-laringológica ibero-americana",
  ["acta otorrinolaringol esp"] = "Acta otorrinolaringológica española",
  ["acta paediatr"] = "Acta paediatrica (Oslo, Norway : 1992)",
  ["acta paediatr acad sci hung"] = "Acta paediatrica Academiae Scientiarum Hungaricae",
  ["acta paediatr belg"] = "Acta paediatrica Belgica",
  ["acta paediatr hung"] = "Acta paediatrica Hungarica",
  ["acta paediatr jpn"] = "Acta Paediatrica Japonica",
  ["acta paediatr lat"] = "Acta paediatrica Latina",
  ["acta paediatr scand"] = "Acta paediatrica Scandinavica",
  ["acta paediatr scand suppl"] = "Acta paediatrica Scandinavica. Supplement",
  ["acta paediatr suppl"] = "Acta paediatrica (Oslo, Norway : 1992). Supplement",
  ["acta paediatr taiwan"] = "Acta paediatrica Taiwanica = Taiwan er ke yi xue hui za zhi",
  ["acta paedolog"] = "Acta paedologica",
  ["acta paedopsychiatr"] = "Acta paedopsychiatrica",
  ["acta palaeobot"] = "Acta Palaeobotanica",
  ["acta palaeontol pol"] = "Acta palaeontologica Polonica",
  ["acta palaeontol rom"] = "Acta Palaeontologica Romaniae",
  ["acta palaeontologica sinica|gu sheng wu xue bao"] = "Gu sheng wu xue bao",
  ["acta parasitol"] = "Acta parasitologica",
  ["acta pathol jpn"] = "Acta pathologica japonica",
  ["acta pathol microbiol immunol scand [a]"] = "Acta Pathologica, Microbiologica, et Immunologica Scandinavica. Section A, Pathology",
  ["acta pathol microbiol immunol scand [b]"] = "Acta Pathologica, Microbiologica, et Immunologica Scandinavica. Section B, Microbiology",
  ["acta pathol microbiol immunol scand [c]"] = "Acta Pathologica, Microbiologica, et Immunologica Scandinavica. Section C, Immunology",
  ["acta pathol microbiol immunol scand a"] = "Acta pathologica, microbiologica, et immunologica Scandinavica. Section A, Pathology",
  ["acta pathol microbiol immunol scand b"] = "Acta pathologica, microbiologica, et immunologica Scandinavica. Section B, Microbiology",
  ["acta pathol microbiol immunol scand c"] = "Acta pathologica, microbiologica, et immunologica Scandinavica. Section C, Immunology",
  ["acta pathol microbiol immunol scand suppl"] = "Acta pathologica, microbiologica, et immunologica Scandinavica. Supplement",
  ["acta pathol microbiol scand"] = "Acta pathologica et microbiologica Scandinavica",
  ["acta pathol microbiol scand [a]"] = "Acta Pathologica et Microbiologica Scandinavica. Section A, Pathology",
  ["acta pathol microbiol scand [b]"] = "Acta Pathologica et Microbiologica Scandinavica. Section B, Microbiology",
  ["acta pathol microbiol scand [b] microbiol immunol"] = "Acta Pathologica et Microbiologica Scandinavica. Section B, Microbiology and Immunology",
  ["acta pathol microbiol scand [c]"] = "Acta Pathologica et Microbiologica Scandinavica. Section C, Immunology",
  ["acta pathol microbiol scand a"] = "Acta pathologica et microbiologica Scandinavica. Section A, Pathology",
  ["acta pathol microbiol scand b"] = "Acta pathologica et microbiologica Scandinavica. Section B, Microbiology",
  ["acta pathol microbiol scand b microbiol immunol"] = "Acta pathologica et microbiologica Scandinavica. Section B: Microbiology and immunology",
  ["acta pathol microbiol scand c"] = "Acta pathologica et microbiologica Scandinavica. Section C, Immunology",
  ["acta pathol microbiol scand suppl"] = "Acta pathologica et microbiologica Scandinavica. Supplement",
  ["acta pediatr esp"] = "Acta pediátrica española",
  ["acta pharm"] = "Acta pharmaceutica (Zagreb, Croatia)",
  ["acta pharm (zagreb, croatia)"] = "Acta Pharmaceutica (Zagreb, Croatia)",
  ["acta pharm hung"] = "Acta pharmaceutica Hungarica",
  ["acta pharm int"] = "Acta pharmaceutica internationalia",
  ["acta pharm jugosl"] = "Acta pharmaceutica Jugoslavica",
  ["acta pharm nord"] = "Acta pharmaceutica Nordica",
  ["acta pharm sin b"] = "Acta pharmaceutica Sinica. B",
  ["acta pharm suec"] = "Acta pharmaceutica Suecica",
  ["acta pharmacol sin"] = "Acta pharmacologica Sinica",
  ["acta pharmacol toxicol (copenh)"] = "Acta pharmacologica et toxicologica",
  ["acta phys"] = "Acta Physica",
  ["acta phys acad sci hung"] = "Acta physica Academiae Scientiarum Hungaricae",
  ["acta phys austriaca"] = "Acta Physica Austriaca",
  ["acta phys chim debrecina"] = "Acta Physica et Chimica Debrecina",
  ["acta phys chim sin"] = "Acta Physico-Chimica Sinica",
  ["acta phys debrecina"] = "Acta Physica Debrecina",
  ["acta phys hung"] = "Acta Physica Hungarica",
  ["acta phys hung a"] = "Acta Physica Hungarica Series A, Heavy Ion Physics",
  ["acta phys hung b"] = "Acta Physica Hungarica B Quantum Electronics",
  ["acta phys pol"] = "Acta Physica Polonica",
  ["acta phys pol a"] = "Acta Physica Polonica A",
  ["acta phys pol b"] = "Acta Physica Polonica B",
  ["acta phys pol b proc suppl"] = "Acta Physica Polonica Proceedings Supplement",
  ["acta phys pol, a"] = "Acta Physica Polonica, A",
  ["acta phys pol, b"] = "Acta Physica Polonica, B",
  ["acta phys pol-ser gen phys"] = "Acta Physica Polonica-Series A General Physics",
  ["acta phys polon a"] = "Polish Academy of Sciences",
  ["acta phys polon b"] = "Jagellonian University. Institute of Physics. Acta Physica Polonica B",
  ["acta phys sinica"] = "Acta Physica Sinica",
  ["acta physiol"] = "Acta Physiologica",
  ["acta physiol (oxf)"] = "Acta physiologica (Oxford, England)",
  ["acta physiol acad sci hung"] = "Acta physiologica Academiae Scientiarum Hungaricae",
  ["acta physiol hung"] = "Acta Physiologica Hungarica",
  ["acta physiol lat am"] = "Acta physiologica latino americana",
  ["acta physiol pharmacol bulg"] = "Acta physiologica et pharmacologica Bulgarica",
  ["acta physiol pharmacol latinoam"] = "Acta physiologica et pharmacologica latinoamericana : organo de la Asociación Latinoamericana de Ciencias Fisiológicas y de la Asociación Latinoamericana de Farmacología",
  ["acta physiol pharmacol neerl"] = "Acta physiologica et pharmacologica Neerlandica",
  ["acta physiol pharmacol ther latinoam"] = "Acta physiologica, pharmacologica et therapeutica latinoamericana : órgano de la Asociación Latinoamericana de Ciencias Fisiológicas y [de] la Asociación Latinoamericana de Farmacología",
  ["acta physiol plant"] = "Acta physiologiae plantarum",
  ["acta physiol pol"] = "Acta physiologica Polonica",
  ["acta physiol scand"] = "Acta Physiologica Scandinavica",
  ["acta physiol scand suppl"] = "Acta physiologica Scandinavica. Supplementum",
  ["acta physiother rheumatol belg"] = "Acta physiotherapica et rheumatologica Belgica",
  ["acta phytogeogr suec"] = "Acta Phytogeographica Suecica",
  ["acta phytomed"] = "Acta Phytomedica",
  ["acta phytopathol entomol hung"] = "Acta phytopathologica et entomologica Hungarica",
  ["acta phytopathologica sinica|zhi wu bing li xue bao"] = "Zhi wu bing li xue bao",
  ["acta phytophylacica sinica|zhi wu bao hu xue bao"] = "Zhi wu bao hu xue bao",
  ["acta phytotaxon geobotan"] = "Acta Phytotaxonomica et Geobotanica",
  ["acta pol hist"] = "Acta Poloniae historica",
  ["acta pol pharm"] = "Acta poloniae pharmaceutica",
  ["acta polym"] = "Acta Polymerica",
  ["acta polytech ctu proc"] = "Acta Polytechnica CTU proceedings",
  ["acta polytech hung"] = "Acta Polytechnica Hungarica",
  ["acta polytech scand math manage comput eng ser"] = "Acta Polytechnica Scandinavica",
  ["acta polytech scand, chem technol ser"] = "Acta Polytechnica Scandinavica - Chemical Technology Series",
  ["acta protozool"] = "Acta protozoologica",
  ["acta psiquiatr psicol am lat"] = "Acta psiquiátrica y psicológica de América latina",
  ["acta psychiatr belg"] = "Acta psychiatrica Belgica",
  ["acta psychiatr neurol"] = "Acta psychiatrica et neurologica",
  ["acta psychiatr neurol scand"] = "Acta psychiatrica et neurologica Scandinavica",
  ["acta psychiatr neurol scand suppl"] = "Acta psychiatrica et neurologica Scandinavica. Supplementum",
  ["acta psychiatr neurol suppl"] = "Acta psychiatrica et neurologica. Supplementum",
  ["acta psychiatr scand"] = "Acta psychiatrica Scandinavica",
  ["acta psychiatr scand suppl"] = "Acta psychiatrica Scandinavica. Supplementum",
  ["acta psychol (amst)"] = "Acta psychologica",
  ["acta psychother psychosom"] = "Acta psychotherapeutica et psychosomatica",
  ["acta psychother psychosom orthopaedagog"] = "Acta psychotherapeutica, psychosomatica et orthopaedagogica",
  ["acta radiol"] = "Acta radiologica (Stockholm, Sweden : 1987)",
  ["acta radiol cancerol"] = "Acta radiologica et cancerologica. Bohemoslovencia",
  ["acta radiol diagn (stockh)"] = "Acta radiologica: diagnosis",
  ["acta radiol oncol"] = "Acta radiologica. Oncology",
  ["acta radiol oncol radiat phys biol"] = "Acta radiologica: oncology, radiation, physics, biology",
  ["acta radiol oncol radiat ther phys biol"] = "Acta radiologica. Oncology, radiation therapy, physics and biology",
  ["acta radiol open"] = "Acta radiologica open",
  ["acta radiol short rep"] = "Acta radiologica short reports",
  ["acta radiol suppl"] = "Acta radiologica. Supplementum",
  ["acta radiol suppl (stockholm)"] = "Acta radiologica. Supplement",
  ["acta radiol ther phys biol"] = "Acta radiologica: therapy, physics, biology",
  ["acta reprod turc"] = "Acta reproductiva Turcica",
  ["acta reumatol port"] = "Acta reumatológica portuguesa",
  ["acta rheumatol scand"] = "Acta rheumatologica Scandinavica",
  ["acta rheumatol scand suppl"] = "Acta rheumatologica Scandinavica. Supplementum",
  ["acta rhumatol"] = "Acta rhumatologica",
  ["acta rhumatol belg"] = "Acta rhumatologica Belgica",
  ["acta salmant hist med"] = "Acta Salmanticensia. Historia de la medicina",
  ["acta salmanticensia ser filos letra"] = "Acta Salmanticensia. Filosofía y letras",
  ["acta salmanticensia: coleccion de memorias y trabajos cientificos editados por la universidad de salamanca, serie"] = "Acta Salmanticensia: Coleccion de Memorias y Trabajos Cientificos Editados por la Universidad de Salamanca, Serie de Ciencias",
  ["acta sch med univ kioto"] = "Acta Scholae Medicinalis Universitatis in Kioto",
  ["acta sci"] = "Acta scientiarum. Animal sciences",
  ["acta sci agron"] = "Acta scientiarum. Agronomy",
  ["acta sci biol sci"] = "Acta scientiarum. Biological sciences",
  ["acta sci cancer biol"] = "Acta scientific cancer biology",
  ["acta sci math"] = "Acta Scientiarum Mathematicarum",
  ["acta sci math (szeged)"] = "Acta Universitatis Szegediensis. Acta Scientiarum Mathematicarum",
  ["acta sci nat acad sci bohemoslov"] = "Acta scientiarum naturalium Academiae scientiarum Bohemoslovaca",
  ["acta sci natur univ jilin"] = "Acta Scientiarum Naturalium Universitatis Jilinensis",
  ["acta sci natur univ norm hunan"] = "Acta Scientiarum Naturalium Universitatis Normalis Hunanensis",
  ["acta sci natur univ sunyatseni"] = "Acta Scientiarum Naturalium Universitatis Sunyatseni. Zhongshan Daxue Xuebao. Ziran Kexue Ban",
  ["acta sci neurol"] = "Acta scientific neurology",
  ["acta sci pol technol aliment"] = "Acta scientiarum polonorum. Technologia alimentaria",
  ["acta sci sin"] = "Acta scientia Sinica = Chung-kuo k'o hsüeh",
  ["acta sci vet"] = "Acta scientiae veterinariae",
  ["acta sci-agron"] = "Acta Scientiarum-Agronomy",
  ["acta scientiarum naturalium universitatis nankaiensis|nan kai da xue xue bao zi ran ke xue ban"] = "Nan kai da xue xue bao. Zi ran ke xue ban",
  ["acta scientiarum naturalium universitatis sunyatseni|zhongshan da xue xue bao zi ran ke xue ban"] = "Zhongshan da xue xue bao. Zi ran ke xue ban",
  ["acta seismol sin"] = "Acta Seismologica Sinica",
  ["acta soc bot pol pol tow bot"] = "Acta Societatis Botanicorum Poloniae. Polskie Towarzystwo Botaniczne",
  ["acta soc fauna flora fenn"] = "Acta Societatis pro Fauna et Flora Fennica",
  ["acta soc med ups"] = "Acta Societatis Medicorum Upsaliensis",
  ["acta soc med ups suppl"] = "Acta Societatis Medicorum Upsaliensis. Supplementum",
  ["acta soc zool bohem"] = "Acta Societatis Zoologicae Bohemoslovacae",
  ["acta sociol"] = "Acta sociologica (Copenhagen, Denmark)",
  ["acta sociomed scand"] = "Acta socio-medica Scandinavica",
  ["acta sociomed scand suppl"] = "Acta socio-medica Scandinavica. Supplement",
  ["acta stereol"] = "Acta stereologica",
  ["acta stomatol belg"] = "Acta stomatologica Belgica",
  ["acta stomatol croat"] = "Acta stomatologica Croatica",
  ["acta stomatol int"] = "Acta stomatologica internationalia : revue scientifique et bulletin officiel de l'Association stomatologique internationale",
  ["acta suom laak duodecim a"] = "Acta Societatis medicorum Fennicae Duodecim. Series A. Suomalainen lääkäriseura Duodecim",
  ["acta suom laak duodecim b"] = "Acta Societatis medicorum Fennicae Duodecim. Series B. Suomalainen lääkäriseura Duodecim",
  ["acta tech"] = "Acta Technica",
  ["acta tech csav"] = "Acta Technica CSAV",
  ["acta ther"] = "Acta therapeutica",
  ["acta theriol"] = "Acta Theriologica",
  ["acta theriol (warsz)"] = "Acta theriologica",
  ["acta theriologica sinica|shou lei xue bao"] = "Shou lei xue bao",
  ["acta trop"] = "Acta tropica",
  ["acta trop suppl"] = "Acta tropica. Supplementum",
  ["acta tuberc belg"] = "Acta tuberculosea Belgica",
  ["acta tuberc jpn"] = "Acta tuberculosea Japonica",
  ["acta tuberc pneumol belg"] = "Acta tuberculosea et pneumologica Belgica",
  ["acta tuberc pneumol scand"] = "Acta tuberculosea et pneumologica Scandinavica",
  ["acta tuberc pneumol scand suppl"] = "Acta tuberculosea et pneumologica Scandinavica. Supplementum",
  ["acta tuberc scand"] = "Acta tuberculosea Scandinavica",
  ["acta tuberc scand suppl"] = "Acta tuberculosea Scandinavica. Supplementum",
  ["acta unio int contra cancrum"] = "Acta - Unio Internationalis Contra Cancrum",
  ["acta univ"] = "Acta universitaria",
  ["acta univ agric brno sb spisy fak vet"] = "Acta Universitatis Agriculturae, Brno. Sborník. Rada B: Spisy Fakulty veterinární",
  ["acta univ agric fac agron"] = "Acta Universitatis Agriculturae. Facultas agronomica. Vysoká škola zemědělská v Brně. Fakulta agronomická",
  ["acta univ apulensis math inform"] = "Acta Universitatis Apulensis. Mathematics. Informatics",
  ["acta univ carol [med] (praha)"] = "Acta Universitatis Carolinae. Medica",
  ["acta univ carol geogr"] = "Acta Universitatis Carolinae. Geographica. Univerzita Karlova",
  ["acta univ carol med (praha)"] = "Acta Universitatis Carolinae. Medica",
  ["acta univ carol med monogr"] = "Acta Universitatis Carolinae. Medica. Monographia",
  ["acta univ carol philos hist"] = "Acta Universitatis Carolinae. Philosophica et historica",
  ["acta univ carol univerzita karlov"] = "Acta Universitatis Carolinae. Historia Universitatis Carolinae Pragensis. Univerzita Karlova",
  ["acta univ carolin math phys"] = "Acta Universitatis Carolinae",
  ["acta univ lodz folia math"] = "Acta Universitatis Lodziensis",
  ["acta univ lodz folia philos"] = "Acta Universitatis Lodziensis",
  ["acta univ lund"] = "Acta Universitatis Lundensis. Sectio I, Theologica, juridica, humaniora",
  ["acta univ m belii ser math"] = "Acta Universitatis Matthiae Belii. Series Mathematics",
  ["acta univ mathaei belii nat sci ser ser math"] = "Matej Bel University",
  ["acta univ oulu ser a sci rerum natur"] = "Acta Universitatis Ouluensis",
  ["acta univ palack olomuc fac med"] = "Acta Universitatis Palackianae Olomucensis Facultatis Medicae",
  ["acta univ palack olomuc fac rerum natur math"] = "Acta Universitatis Palackianae Olomucensis. Facultas Rerum Naturalium. Mathematica",
  ["acta univ palacki olomuc fac med"] = "Acta Universitatis Palackianae Olomucensis Facultatis Medicae",
  ["acta univ sapientiae matem"] = "Acta universitatis sapientiae. Mathematica",
  ["acta univ sapientiae math"] = "Acta Universitatis Sapientiae. Mathematica. An International Scientific Journal of Sapientia University",
  ["acta univ szeged attila jozsef nominatae"] = "Acta Universitatis Szegediensis de Attila József Nominatae. Acta juridica et politica",
  ["acta univ tamper ser a"] = "Acta Universitatis Tamperensis",
  ["acta univ tampere"] = "Acta Universitatis Tamperensis",
  ["acta univ ups"] = "Acta Universitatis Upsaliensis. Studia sociologica Upsaliensia",
  ["acta univ ups stud philos ups"] = "Acta Universitatis Upsaliensis",
  ["acta univ ups stud stat ups"] = "Acta Universitatis Upsaliensis",
  ["acta univ ups symb bot ups"] = "Acta Universitatis Upsaliensis. Symbolae botanicae Upsalienses : Arbeten från Botaniska institutionerna i Uppsala",
  ["acta univ upsaliensis skr uppsala univ c organ hist"] = "Acta Universitatis Upsaliensis",
  ["acta univ wratislav"] = "Acta Universitatis Wratislaviensis",
  ["acta urol"] = "Acta urologica",
  ["acta urol belg"] = "Acta urologica Belgica",
  ["acta vet (beogr)"] = "Acta Veterinaria",
  ["acta vet acad sci hung"] = "Acta veterinaria Academiae Scientiarum Hungaricae",
  ["acta vet brno"] = "Acta veterinaria",
  ["acta vet hung"] = "Acta veterinaria Hungarica",
  ["acta vet indones"] = "Acta veterinaria Indonesiana",
  ["acta vet scand"] = "Acta veterinaria Scandinavica",
  ["acta vet scand suppl"] = "Acta veterinaria Scandinavica. Supplementum",
  ["acta veterinaria et zootechnica sinica / [zhongguo xu mu shou yi xue hui]|xu mu shou yi xue bao"] = "Xu mu shou yi xue bao",
  ["acta virol"] = "Acta virologica",
  ["acta vitaminol"] = "Acta vitaminologica",
  ["acta vitaminol enzymol"] = "Acta vitaminologica et enzymologica",
  ["acta zoobot austria"] = "Acta ZooBot Austria",
  ["acta zool"] = "Acta zoologica (Stockholm, Sweden)",
  ["acta zool academ sci hung"] = "Acta zoologica Academiae Scientiarum Hungaricae (Budapest, Hungary : 1994)",
  ["acta zool bulg"] = "Acta zoologica Bulgarica",
  ["acta zool litu"] = "Acta zoologica Lituanica",
  ["acta zool mex"] = "Acta zoologica mexicana",
  ["acta zool pathol antverp"] = "Acta zoologica et pathologica Antverpiensia",
  ["acta zool taiwanica"] = "Acta zoologica Taiwanica (1992)",
  ["acta zootaxonomica sinica|dong wu fen lei xue bao"] = "Dong wu fen lei xue bao",
  ["actaaarthist sa"] = "Acta ad archaeologiam et artium historiam pertinentia. Series altera in 8",
  ["actas ciba (rio j)"] = "Actas Ciba",
  ["actas congr “dr antonio a r monteiro”"] = "Actas del Congreso “Dr. Antonio A. R. Monteiro”",
  ["actas dermatol dermatopatol"] = "Actas de dermatología & dermatopatología",
  ["actas dermosifiliogr"] = "Actas dermo-sifiliográficas",
  ["actas esp psiquiatr"] = "Actas españolas de psiquiatría",
  ["actas luso esp neurol psiquiatr"] = "Actas luso-españolas de neurología y psiquiatría",
  ["actas luso esp neurol psiquiatr cienc afines"] = "Actas luso-españolas de neurología, psiquiatría y ciencias afines",
  ["actas reun cient cuerfo fac inst policlin"] = "Actas de las reuniones científicas del cuerpo facultativo del Instituto Policlínico",
  ["actas urol esp"] = "Actas urologicas españolas",
  ["actes rech sci soc"] = "Actes de la recherche en sciences sociales",
  ["actes rencontres cirm"] = "Actes des Rencontres du CIRM",
  ["actes soc helv sci nat"] = "Actes de la Société helvétique des sciences naturelles",
  ["actes soc jura émul"] = "Actes de la Société jurassienne d’émulation",
  ["actex acad ser"] = "ACTEX Academic Series",
  ["actex prof ser"] = "ACTEX Professional Series",
  ["actinomycetologica|nippon hosenkin gakkaishi"] = "Nihon Hosenkin Gakkai shi",
  ["action child"] = "Action For Children",
  ["action contre sida"] = "Action contre le SIDA",
  ["action learn"] = "Action learning",
  ["action natl"] = "Action (National Fund for Research into Crippling Diseases)",
  ["action res (lond)"] = "Action research (London)",
  ["actionaid disabil news"] = "ActionAid disability news : the newsletter of Disability Division, ActionAid, India",
  ["actual anatpathol"] = "Actualites anatomo-pathologiques",
  ["actual biol"] = "Actualites biologiques",
  ["actual cardiol angeiol int (paris)"] = "Actualites Cardiologiques et Angeiologiques Internationales",
  ["actual chim"] = "Actualite Chimique",
  ["actual econ"] = "L’Actualite economique",
  ["actual endocrinol (paris)"] = "Actualites Endocrinologiques",
  ["actual hepatogastroenterol (paris)"] = "Actualites Hepato-Gastro-Enterologiques",
  ["actual jurid aranzadi"] = "Actualidad jurídica Aranzadi",
  ["actual med peru"] = "Actualidad médica peruana",
  ["actual medica"] = "Actualidad médica",
  ["actual neurophysiol (paris)"] = "Actualites Neurophysiologiques",
  ["actual odonto estomatol esp"] = "Actualidad odontoestomatologica española",
  ["actual odontostomatol (paris)"] = "Actualites Odonto-Stomatologiques",
  ["actual pediatr (granada)"] = "Actualidad pediátrica; revista de bibliografía internacional",
  ["actual pharm"] = "Actualites pharmaceutiques",
  ["actual pharmacol (paris)"] = "Actualites Pharmacologiques",
  ["actual physiol pathol (paris)"] = "Actualites de Physiologie Pathologique",
  ["actual psychiatr"] = "Actualites psychiatriques",
  ["actualites sci agronom inra"] = "Actualites Scientifiques et Agronomiques de l’INRA",
  ["actualités math"] = "Actualités Mathématiques",
  ["actualités sci indust"] = "Actualités Scientifiques et Industrielles",
  ["actuel urol"] = "Actuelle Urologie",
  ["acupunct electrother res"] = "Acupuncture and Electro-Therapeutics Research",
  ["acupunct med"] = "Acupuncture in medicine : journal of the British Medical Acupuncture Society",
  ["acute card care"] = "Acute Cardiac Care",
  ["acute care"] = "Acute Care",
  ["acute crit care"] = "Acute and critical care",
  ["acute med"] = "Acute medicine",
  ["acute med surg"] = "Acute medicine & surgery",
  ["ad astra"] = "Ad astra (Washington, D.C.)",
  ["ad hoc networks"] = "Ad Hoc Networks",
  ["ad nurse"] = "AD Nurse",
  ["ada news"] = "ADA News",
  ["adapt behav"] = "Adaptive Behavior",
  ["adapt comput mach learn"] = "Adaptive Computation and Machine Learning",
  ["adapt hum behav physiol"] = "Adaptive Human Behavior and Physiology",
  ["adapt learn optim"] = "Adaptation, Learning, and Optimization",
  ["adapt learn syst signal process commun control"] = "Adaptive and Learning Systems for Signal Processing, Communications, and Control",
  ["adapt phys activ q"] = "Adapted physical activity quarterly : APAQ",
  ["adapt sistemy avtomat upravleniya"] = "Kievskiy Politekhnicheskiy Institut",
  ["adcaij: adv distrib comput and artif intell j"] = "ADCAIJ: Advances in Distributed Computing and Artificial Intelligence Journal",
  ["addict behav"] = "Addictive behaviors",
  ["addict behav rep"] = "Addictive behaviors reports",
  ["addict biol"] = "Addiction biology",
  ["addict dis"] = "Addictive diseases",
  ["addict disord their treat"] = "Addictive disorders & their treatment",
  ["addict health"] = "Addiction & health",
  ["addict neurosci"] = "Addiction Neuroscience",
  ["addict res theory"] = "Addiction Research & Theory",
  ["addict sci clin pract"] = "Addiction science & clinical practice",
  ["addison-wesley ser comput sci inform process"] = "Addison-Wesley Series in Computer Science and Information Processing",
  ["addison-wesley ser math"] = "Addison-Wesley Series in Mathematics",
  ["addison-wesley ser statist"] = "Addison-Wesley Series in Statistics",
  ["addit manuf"] = "Additive Manufacturing",
  ["addit manuf lett"] = "Additive Manufacturing Letters",
  ["addit polym"] = "Additives for Polymers",
  ["additamenta folia med neerl"] = "Folia medica Neerlandica. Additamenta",
  ["adelt a"] = "Αρχαιολογικόν Δελτίον (Μελέτες)",
  ["adelt b"] = "Αρχαιολογικόν Δελτίον (Χρονικά)",
  ["adhaes kleben dicten"] = "Adhaesion–Kleben & Dichten",
  ["adhas kleben dicten"] = "Adhasion–Kleben & Dichten",
  ["adhd rep"] = "The ADHD report",
  ["adhes age"] = "Adhesives Age",
  ["adiyat halab"] = "ʻĀdīyāt Ḥalab",
  ["adjac gov"] = "Adjacent government",
  ["adler mus bull"] = "Adler Museum bulletin",
  ["adm change"] = "Administrative change",
  ["adm dev"] = "Administration for development",
  ["adm law j"] = "Administrative law journal",
  ["adm law rev"] = "Administrative law review",
  ["adm manage"] = "Administrative management",
  ["adm ment health"] = "Administration in mental health",
  ["adm policy ment health"] = "Administration and policy in mental health",
  ["adm radiol"] = "Administrative radiology : AR",
  ["adm radiol j"] = "Administrative radiology journal : AR",
  ["adm sci"] = "Administrative sciences",
  ["adm sci q"] = "Administrative science quarterly",
  ["adm soc"] = "Administration & society",
  ["adm soc work"] = "Administration in social work",
  ["admet dmpk"] = "ADMET & DMPK",
  ["admit manage j"] = "The Admitting management journal",
  ["adolesc educ newsl"] = "Adolescence education newsletter",
  ["adolesc fam health"] = "Adolescent & family health : A & FH",
  ["adolesc health med ther"] = "Adolescent health, medicine and therapeutics",
  ["adolesc med"] = "Adolescent medicine (Philadelphia, Pa.)",
  ["adolesc med clin"] = "Adolescent medicine clinics",
  ["adolesc med state art rev"] = "Adolescent medicine: state of the art reviews",
  ["adolesc pediatr gynecol"] = "Adolescent and pediatric gynecology",
  ["adolesc psychiatry"] = "Adolescent psychiatry",
  ["adolesc psychiatry (hilversum)"] = "Adolescent psychiatry (Hilversum, Netherlands)",
  ["adolesc res rev"] = "Adolescent research review",
  ["adolesc saude"] = "Adolescência & saúde",
  ["adopt foster"] = "Adoption & fostering",
  ["adopt q"] = "Adoption quarterly",
  ["adsorp sci technol"] = "Adsorption science & technology : interface science for advanced materials & technologies",
  ["adsorpt sci technol"] = "Adsorption Science and Technology",
  ["adsorption (boston)"] = "Adsorption : journal of the International Adsorption Society",
  ["adult basic educ lit j"] = "Adult basic education and literacy journal",
  ["adult educ"] = "Adult education",
  ["adult educ q (am assoc adult contin educ)"] = "Adult education quarterly (American Association for Adult and Continuing Education)",
  ["adult foster care j"] = "Adult foster care journal : AFCJ",
  ["adult resid care j"] = "Adult residential care journal",
  ["adv abort care"] = "Advances in abortion care",
  ["adv acoust vib"] = "Advances in acoustics and vibration",
  ["adv acoust vibr"] = "Advances in Acoustics and Vibrations",
  ["adv adapt data anal"] = "Advances in adaptive data analysis",
  ["adv adm lab"] = "Advance for administrators of the laboratory",
  ["adv adolesc mental health"] = "Advances in adolescent mental health",
  ["adv aerodyn"] = "Advances in Aerodynamics",
  ["adv aging res"] = "Advances in aging research",
  ["adv agric"] = "Advances in Agriculture",
  ["adv agrochem"] = "Advanced Agrochem",
  ["adv agron"] = "Advances in Agronomy",
  ["adv air pollut ser"] = "Advances in Air Pollution Series",
  ["adv aircr spacecr sci"] = "Advances in aircraft & spacecraft science",
  ["adv alcohol subst abuse"] = "Advances in alcohol & substance abuse",
  ["adv alzheimer dis"] = "Advances in Alzheimer's disease",
  ["adv anal geom"] = "Advances in Analysis and Geometry",
  ["adv anat embryol cell biol"] = "Advances in anatomy, embryology, and cell biology",
  ["adv anat pathol"] = "Advances in anatomic pathology",
  ["adv anesth"] = "Advances in anesthesia",
  ["adv anim biosci"] = "Advances in Animal Biosciences",
  ["adv anim vet sci"] = "Advances in animal and veterinary sciences",
  ["adv antivir drug des"] = "Advances in antiviral drug design",
  ["adv appl biodivers sci"] = "Advances in Applied Biodiversity Science",
  ["adv appl bioinf chem"] = "Advances and Applications in Bioinformatics and Chemistry",
  ["adv appl bioinform chem"] = "Advances and applications in bioinformatics and chemistry : AABC",
  ["adv appl ceram"] = "Advances in Applied Ceramics",
  ["adv appl clifford algebr"] = "Advances in applied Clifford algebras",
  ["adv appl clifford algebras"] = "Advances in Applied Clifford Algebras",
  ["adv appl energy"] = "Advances in Applied Energy",
  ["adv appl math"] = "Advances in Applied Mathematics",
  ["adv appl math mech"] = "Advances in Applied Mathematics and Mechanics",
  ["adv appl math sci"] = "Advances and Applications in Mathematical Sciences",
  ["adv appl mech"] = "Advances in Applied Mechanics",
  ["adv appl microbiol"] = "Advances in applied microbiology",
  ["adv appl pattern recogn"] = "Advanced Applications in Pattern Recognition",
  ["adv appl probab"] = "Advances in Applied Probability",
  ["adv appl sociol"] = "Advances in applied sociology",
  ["adv appl stat"] = "Advances and applications in statistics",
  ["adv appl stat sci"] = "Advances and applications in statistical sciences",
  ["adv appl statist"] = "Advances in Applied Statistics",
  ["adv archaeol pract"] = "Advances in Archaeological Practice",
  ["adv archaeomater"] = "Advances in Archaeomaterials",
  ["adv artif intell"] = "Advances in artificial intelligence. Canadian Society for Computational Studies of Intelligence. Conference",
  ["adv artif intell mach learn"] = "Advances in Artificial Intelligence and Machine Learning",
  ["adv astrobiol biogeophys"] = "Advances In Astrobiology And Biogeophysics",
  ["adv astron"] = "Advances in Astronomy",
  ["adv astronaut sci technol"] = "Advances in Astronautics Science and Technology",
  ["adv at mol opt phy"] = "Advances in Atomic Molecular and Optical Physics",
  ["adv at mol opt phys"] = "Advances in Atomic Molecular and Optical Physics",
  ["adv at mol phys"] = "Advances in Atomic and Molecular Physics",
  ["adv atmos sci"] = "Advances in Atmospheric Sciences",
  ["adv behav biol"] = "Advances in behavioral biology",
  ["adv bio tech"] = "Advanced bio tech : India's biotechnology magazine",
  ["adv biochem"] = "Advances in biochemistry",
  ["adv biochem eng biotechnol"] = "Advances in biochemical engineering/biotechnology",
  ["adv biochem eng/biotechnol"] = "Advances in Biochemical Engineering / Biotechnology",
  ["adv biochem psychopharmacol"] = "Advances in biochemical psychopharmacology",
  ["adv bioeth"] = "Advances in bioethics",
  ["adv bioinf"] = "Advances in Bioinformatics",
  ["adv bioinformatics"] = "Advances in bioinformatics",
  ["adv biol"] = "Advanced biology",
  ["adv biol chem"] = "Advances in biological chemistry",
  ["adv biol med phys"] = "Advances in biological and medical physics",
  ["adv biol regul"] = "Advances in biological regulation",
  ["adv biol res (rennes)"] = "Advances in biological research",
  ["adv biol sci res"] = "Advances in biological sciences research",
  ["adv biol skin"] = "Advances in biology of skin",
  ["adv biomark sci technol"] = "Advances in biomarker sciences and technology",
  ["adv biomarker sci technol"] = "Advances in Biomarker Sciences and Technology",
  ["adv biomed eng"] = "Advanced Biomedical Engineering",
  ["adv biomed eng med phys"] = "Advances in biomedical engineering and medical physics",
  ["adv biomed eng res"] = "Advances in biomedical engineering research",
  ["adv biomed res"] = "Advanced biomedical research",
  ["adv bionics biokinet biocybern"] = "Advanced Bionics, Biokinetics and Biocybernetics",
  ["adv biophys"] = "Advances in Biophysics",
  ["adv biores"] = "Advances in bioresearch",
  ["adv biosci"] = "Advances in the biosciences",
  ["adv biosci bioeng (n y)"] = "Advances in bioscience and bioengineering (New York, NY)",
  ["adv biosci biotechnol"] = "Advances in bioscience and biotechnology (Print)",
  ["adv biosci clin med"] = "Advances in bioscience and clinical medicine",
  ["adv biosens bioelectron"] = "Advances in biosensors and bioelectronics",
  ["adv biosyst"] = "Advanced biosystems",
  ["adv biotechnol microbiol"] = "Advances in biotechnology & microbiology (Newbury, Calif.)",
  ["adv biotechnol processes"] = "Advances in biotechnological processes",
  ["adv book class"] = "Advanced Book Classics",
  ["adv bot res"] = "Advances in botanical research",
  ["adv bound elem ser"] = "Boundary Elements XXIV: Incorporating Meshless Solutions",
  ["adv breast cancer res"] = "Advances in breast cancer research",
  ["adv bridge eng"] = "Advances in Bridge Engineering",
  ["adv build energy res"] = "Advances in Building Energy Research",
  ["adv cad vlsi"] = "Advances in CAD for VLSI",
  ["adv calc var"] = "Advances in Calculus of Variations",
  ["adv cancer prev"] = "Advances in cancer prevention",
  ["adv cancer res"] = "Advances in cancer research",
  ["adv cancer res treat"] = "Advances in cancer: research & treatment",
  ["adv carbohydr chem"] = "Advances in carbohydrate chemistry",
  ["adv carbohydr chem biochem"] = "Advances in carbohydrate chemistry and biochemistry",
  ["adv card surg"] = "Advances in cardiac surgery",
  ["adv cardiol"] = "Advances in cardiology",
  ["adv cardiopulm dis"] = "Advances in cardiopulmonary diseases",
  ["adv catal"] = "Advances in Catalysis",
  ["adv cell biol"] = "Advances in cell biology",
  ["adv cell gene ther"] = "Advances in cell and gene therapy",
  ["adv cell mol biol plants"] = "Advances in cellular and molecular biology of plants",
  ["adv cem based mater"] = "Advanced Cement Based Materials",
  ["adv cem res"] = "Advances in Cement Research",
  ["adv chem phys"] = "Advances in Chemical Physics",
  ["adv chem res"] = "Advances in chemistry research",
  ["adv chem ser"] = "Advances in chemistry series",
  ["adv chemother"] = "Advances in chemotherapy",
  ["adv child dev behav"] = "Advances in child development and behavior",
  ["adv chromatogr"] = "Advances in Chromatography",
  ["adv chromatogr (boca raton, fl, u s)"] = "Advances in Chromatography (Boca Raton, FL, United States)",
  ["adv chromatogr (n y, ny, u s)"] = "Advances in Chromatography (New York, NY, United States)",
  ["adv chronic kidney dis"] = "Advances in chronic kidney disease",
  ["adv civ eng"] = "Advances in Civil Engineering",
  ["adv civ eng mater"] = "Advances in civil engineering materials",
  ["adv clim change res"] = "Advances in Climate Change Research",
  ["adv clin care"] = "Advancing clinical care : official journal of NOAADN",
  ["adv clin chem"] = "Advances in clinical chemistry",
  ["adv clin endocrinol metab"] = "Advances in clinical endocrinology and metabolism",
  ["adv clin exp med"] = "Advances in clinical and experimental medicine : official organ Wroclaw Medical University",
  ["adv clin neurosci rehabil"] = "Advances in clinical neuroscience & rehabilitation : ACNR",
  ["adv clin path"] = "Advances in clinical pathology : the official journal of Adriatic Society of Pathology",
  ["adv clin pharmacol"] = "Advances in clinical pharmacology",
  ["adv clin radiol"] = "Advances in Clinical Radiology",
  ["adv clin rehabil"] = "Advances in clinical rehabilitation",
  ["adv cogn neurodyn"] = "Advances in Cognitive Neurodynamics",
  ["adv cogn psychol"] = "Advances in cognitive psychology",
  ["adv colloid interface sci"] = "Advances in colloid and interface science",
  ["adv comb"] = "Advances in Combinatorics",
  ["adv comp physiol biochem"] = "Advances in comparative physiology and biochemistry",
  ["adv complex syst"] = "Advances in Complex Systems",
  ["adv compos bull"] = "Advanced Composites Bulletin",
  ["adv compos hybrid mater"] = "Advanced composites and hybrid materials",
  ["adv compos lett"] = "Advanced Composites Letters",
  ["adv compos mater"] = "Advanced Composite Materials",
  ["adv comput electr eng (acee) book ser"] = "Advances in Computer and Electrical Engineering (ACEE) Book Series",
  ["adv comput fluid dyn"] = "Advances in Computational Fluid Dynamics",
  ["adv comput intell"] = "Advances in Computational Intelligence",
  ["adv comput intell robot acir book ser"] = "Advances in Computational Intelligence and Robotics (ACIR) Book Series",
  ["adv comput manag sci"] = "Advances in Computational Management Science",
  ["adv comput math"] = "Advances in Computational Mathematics",
  ["adv comput sci"] = "Advances in Computing Science",
  ["adv comput sci eng"] = "Advances in Computational Science and Engineering",
  ["adv comput sci technol"] = "Advances in Computer Science and Technology",
  ["adv comput vis pattern recognit"] = "Advances in Computer Vision and Pattern Recognition",
  ["adv concr constr"] = "Advances in Concrete Construction",
  ["adv condens matter phys"] = "Advances in Condensed Matter Physics",
  ["adv consum res"] = "Advances in consumer research. Association for Consumer Research (U.S.)",
  ["adv contin discrete models"] = "Advances in Continuous and Discrete Models",
  ["adv contin mech"] = "Advances in Continuum Mechanics",
  ["adv contracept"] = "Advances in contraception : the official journal of the Society for the Advancement of Contraception",
  ["adv contracept deliv syst"] = "Advances in contraceptive delivery systems : CDS",
  ["adv control appl"] = "Advanced Control for Applications",
  ["adv course nucl eng"] = "An Advanced Course in Nuclear Engineering",
  ["adv courses math crm barcelona"] = "Advanced Courses in Mathematics. CRM Barcelona",
  ["adv cryog eng"] = "Advances in cryogenic engineering",
  ["adv cyclic nucleotide protein phosphorylation res"] = "Advances in cyclic nucleotide and protein phosphorylation research",
  ["adv cyclic nucleotide res"] = "Advances in cyclic nucleotide research",
  ["adv cytopharmacol"] = "Advances in cytopharmacology",
  ["adv damage mech"] = "Advances in Damage Mechanics",
  ["adv data"] = "Advance data",
  ["adv data anal classif"] = "Advances in data analysis and classification",
  ["adv data min"] = "Advances in data mining. Industrial Conference on Data Mining",
  ["adv data sci adapt anal"] = "Advances in Data Science and Adaptive Analysis",
  ["adv database technol"] = "Advances in database technology : proceedings. International Conference on Extending Database Technology",
  ["adv decis sci"] = "Advances in Decision Sciences",
  ["adv delays dyn"] = "Advances in Delays and Dynamics",
  ["adv dent res"] = "Advances in dental research",
  ["adv dermatol"] = "Advances in dermatology",
  ["adv des control"] = "Advances in Design and Control",
  ["adv dev biol"] = "Advances in Developmental Biology",
  ["adv dev hum resour"] = "Advances in developing human resources",
  ["adv device mater"] = "Advanced Device Materials",
  ["adv devices instrum"] = "Advanced Devices & Instrumentation",
  ["adv differ equ"] = "Advances in difference equations",
  ["adv differ equations"] = "Advances in Difference Equations",
  ["adv difference equ"] = "Advances in Difference Equations",
  ["adv differential equations"] = "Advances in Differential Equations",
  ["adv dis control prev"] = "Advances in disease control and prevention",
  ["adv dis surveill"] = "Advances in disease surveillance",
  ["adv drug alcohol res"] = "Advances in Drug and Alcohol Research",
  ["adv drug deliv rev"] = "Advanced drug delivery reviews",
  ["adv drug delivery rev"] = "Advanced Drug Delivery Reviews",
  ["adv drug res"] = "Advances in drug research",
  ["adv dual diagn"] = "Advances in dual diagnosis",
  ["adv dyn syst appl"] = "Advances in Dynamical Systems and Applications",
  ["adv eat disord"] = "Advances in eating disorders (Abingdon, England )",
  ["adv ecol res"] = "Advances in ecological research",
  ["adv ecol rs"] = "Advances in ecological research",
  ["adv econ bus"] = "Advances in economics and business",
  ["adv econom"] = "Advances in Econometrics",
  ["adv electrochem sci eng"] = "Advances in Electrochemical Science and Engineering",
  ["adv electron mater"] = "Advanced Electronic Materials",
  ["adv emerg nurs j"] = "Advanced emergency nursing journal",
  ["adv endocrinol metab"] = "Advances in endocrinology and metabolism",
  ["adv energy convers"] = "Advanced Energy Conversion",
  ["adv energy mater"] = "Advanced Energy Materials",
  ["adv energy sustainability res"] = "Advanced Energy and Sustainability Research",
  ["adv eng educ"] = "Advances in Engineering Education",
  ["adv eng inf"] = "Advanced Engineering Informatics",
  ["adv eng mater"] = "Advanced Engineering Materials",
  ["adv eng softw"] = "Advances in engineering software (Barking, London, England : 1992)",
  ["adv eng software"] = "Advances in Engineering Software",
  ["adv entomol"] = "Advances in Entomology (Irvine, Calif.)",
  ["adv environ biol"] = "Advances in environmental biology",
  ["adv environ eng green technol (aeegt) book ser"] = "Advances in Environmental Engineering and Green Technologies (AEEGT) Book Series",
  ["adv environ res"] = "Advances in Environmental Research",
  ["adv enzyme regul"] = "Advances in Enzyme Regulation",
  ["adv enzyme res"] = "Advances in enzyme research",
  ["adv enzymol relat areas mol biol"] = "Advances in Enzymology and Related Areas of Molecular Biology",
  ["adv enzymol relat subj biochem"] = "Advances in enzymology and related subjects of biochemistry",
  ["adv epidemiol"] = "Advances in epidemiology",
  ["adv ethol"] = "Advances in Ethology",
  ["adv evol biol"] = "Advances in evolutionary biology",
  ["adv evol equ"] = "Advances in Evolution Equations",
  ["adv exp biol"] = "Advances in experimental biology",
  ["adv exp med biol"] = "Advances in experimental medicine and biology",
  ["adv exp philos"] = "Advances in Experimental Philosophy",
  ["adv exp soc psychol"] = "Advances in experimental social psychology",
  ["adv fertil control"] = "Advances in fertility control",
  ["adv fiber mater"] = "Advanced Fiber Materials",
  ["adv filtr sep technol"] = "Advances in Filtration and Separation Technology",
  ["adv finance account econ (afae) book ser"] = "Advances in Finance, Accounting, and Economics (AFAE) Book Series",
  ["adv fluid mech"] = "Advances in Fluid Mechanics",
  ["adv fluorine res"] = "Advances in fluorine research and dental caries prevention",
  ["adv food nutr res"] = "Advances in food and nutrition research",
  ["adv food res"] = "Advances in food research",
  ["adv food res suppl"] = "Advances in food research. Supplement",
  ["adv forensic haemogenet"] = "Advances in Forensic Haemogenetics",
  ["adv free radical biol med"] = "Advances in Free Radical Biology and Medicine",
  ["adv funct mater"] = "Advanced Functional Materials",
  ["adv fuzzy syst"] = "Advances in Fuzzy Systems",
  ["adv fuzzy systems appl theory"] = "Advances in Fuzzy Systems—Applications and Theory",
  ["adv genet"] = "Advances in genetics",
  ["adv genet (hoboken)"] = "Advanced genetics (Hoboken, N.J.)",
  ["adv genet eng"] = "Advancements in genetic engineering",
  ["adv genet res"] = "Advances in genetics research",
  ["adv genomics genet"] = "Advances in genomics and genetics",
  ["adv geoecol"] = "Advances in Geoecology",
  ["adv geom"] = "Advances in Geometry",
  ["adv geophys environ mech math"] = "Advances in Geophysical and Environmental Mechanics and Mathematics",
  ["adv geosci"] = "Advances in Geosciences",
  ["adv geriatr"] = "Advances in geriatrics",
  ["adv geriatr med res"] = "Advances in geriatric medicine and research",
  ["adv gerontol"] = "Advances in gerontology = Uspekhi gerontologii",
  ["adv gerontol res"] = "Advances in gerontological research",
  ["adv group decis negot"] = "Advances in Group Decision and Negotiation",
  ["adv group process"] = "Advances in group processes",
  ["adv group theory appl"] = "Advances in Group Theory and Applications",
  ["adv gut microbiome res"] = "Advanced Gut & Microbiome Research",
  ["adv health care manag"] = "Advances in health care management",
  ["adv health care technol"] = "Advanced health care technologies",
  ["adv health econ health serv res"] = "Advances in health economics and health services research",
  ["adv health econ health serv res suppl"] = "Advances in health economics and health services research. Supplement",
  ["adv health sci educ theory pract"] = "Advances in health sciences education : theory and practice",
  ["adv healthc mater"] = "Advanced healthcare materials",
  ["adv healthcare mater"] = "Advanced Healthcare Materials",
  ["adv heat transf"] = "Advances in heat transfer",
  ["adv hematol"] = "Advances in hematology",
  ["adv hepatol"] = "Advances in hepatology",
  ["adv heterocycl chem"] = "Advances in Heterocyclic Chemistry",
  ["adv high energy phys"] = "Advances in High Energy Physics",
  ["adv hortic sci"] = "Advances in horticultural science",
  ["adv hum comput interact"] = "Advances in Human-Computer Interaction",
  ["adv hum genet"] = "Advances in human genetics",
  ["adv hydrogen energy"] = "Advances in Hydrogen Energy",
  ["adv image comm"] = "Advances in Image Communication",
  ["adv imaging"] = "Advanced imaging (Woodbury, N.Y.)",
  ["adv immun cancer ther"] = "Advances in immunity and cancer therapy",
  ["adv immunol"] = "Advances in immunology",
  ["adv in appl math"] = "Advances in Applied Mathematics",
  ["adv in appl probab"] = "Advances in Applied Probability",
  ["adv in insect phys"] = "Advances in insect physiology",
  ["adv ind control"] = "Advances in Industrial Control",
  ["adv ind eng polym res"] = "Advanced Industrial and Engineering Polymer Research",
  ["adv ind manuf eng"] = "Advances in Industrial and Manufacturing Engineering",
  ["adv inf knowl process"] = "Advanced Information and Knowledge Processing",
  ["adv inf secur"] = "Advances in Information Security",
  ["adv infect dis"] = "Advances in infectious diseases",
  ["adv inorg biochem"] = "Advances in inorganic biochemistry",
  ["adv inorg chem"] = "Advances in Inorganic Chemistry",
  ["adv integr med"] = "Advances in integrative medicine",
  ["adv intell data anal"] = "Advances in intelligent data analysis. International Symposium on Intelligent Data Analysis",
  ["adv intell discovery"] = "Advanced Intelligent Discovery",
  ["adv intell syst"] = "Advanced intelligent systems (Weinheim an der Bergstrasse, Germany)",
  ["adv intell syst comput"] = "Advances in Intelligent Systems and Computing",
  ["adv intern med"] = "Advances in internal medicine",
  ["adv j emerg med"] = "Advanced journal of emergency medicine",
  ["adv j food sci technol"] = "Advance journal of food science and technology : AJFST",
  ["adv j mol imaging"] = "Advances in molecular imaging",
  ["adv j urol nephrol"] = "Advancements in journal of urology and nephrology",
  ["adv jpn bus econ"] = "Advances in Japanese Business and Economics",
  ["adv knowl acquis transf manag (akatm)"] = "Advances in Knowledge Acquisition, Transfer and Management (AKATM)",
  ["adv knowl discov data min (2017)"] = "Advances in Knowledge Discovery and Data Mining : 21st Pacific-Asia Conference, PAKDD 2017, Jeju, South Korea, May 23-26, 2017, Proceedings. Part I. Pacific-Asia Conference on Knowledge Discovery and Data Mining (21st : 2017 : Cheju Island, Korea)",
  ["adv lect math (alm)"] = "Advanced Lectures in Mathematics (ALM)",
  ["adv lectures math"] = "Advanced Lectures in Mathematics",
  ["adv legume syst"] = "Advances in legume systematics",
  ["adv life course res"] = "Advances in life course research",
  ["adv life sci"] = "Advancements in life sciences",
  ["adv limnol"] = "Advances in limnology",
  ["adv lipid res"] = "Advances in Lipid Research",
  ["adv lung cancer (irvine)"] = "Advances in lung cancer (Irvine)",
  ["adv magn opt reson"] = "Advances in Magnetic and Optical Resonance",
  ["adv magn reson"] = "Advances in Magnetic Resonance",
  ["adv manuf"] = "Advances in manufacturing",
  ["adv manuf polym compos sci"] = "Advanced Manufacturing: Polymer & Composites Science",
  ["adv mar biol"] = "Advances in marine biology",
  ["adv mass spectrom"] = "Advances in mass spectrometry",
  ["adv mat res"] = "Advanced materials research",
  ["adv mat technol"] = "Advances Materials Technologies",
  ["adv mater"] = "Advanced Materials",
  ["adv mater (weinheim, ger)"] = "Advanced Materials (Weinheim, Germany)",
  ["adv mater deerfield"] = "Advanced materials (Deerfield Beach, Fla.)",
  ["adv mater forum"] = "Advanced Materials Forum",
  ["adv mater interfaces"] = "Advanced Materials Interfaces",
  ["adv mater lett"] = "Advanced materials letters",
  ["adv mater opt electron"] = "Advanced Materials for Optics and Electronics",
  ["adv mater process technol"] = "Advances in Materials and Processing Technologies",
  ["adv mater processes"] = "Advanced Materials and Processes",
  ["adv mater res"] = "Advances in Materials Research",
  ["adv mater sci appl"] = "Advances in materials science and applications",
  ["adv mater sci eng"] = "Advances in Materials Science and Engineering",
  ["adv mater technol"] = "Advanced Materials Technologies",
  ["adv math"] = "Advances in Mathematics",
  ["adv math (china)"] = "Advances in Mathematics (China)",
  ["adv math (n y)"] = "Advances in mathematics",
  ["adv math commun"] = "Advances in Mathematics of Communications",
  ["adv math econ"] = "Advances in Mathematical Economics",
  ["adv math educ"] = "Advances in Mathematics Education",
  ["adv math fluid mech"] = "Advances in Mathematical Fluid Mechanics",
  ["adv math phys"] = "Advances in Mathematical Physics",
  ["adv math sci"] = "Advances in the Mathematical Sciences",
  ["adv math sci appl"] = "Advances in Mathematical Sciences and Applications",
  ["adv mech"] = "Advances in Mechanics",
  ["adv mech eng"] = "Advances in Mechanical Engineering",
  ["adv mech math"] = "Advances in Mechanics and Mathematics",
  ["adv mechatron mech eng (amme) book ser"] = "Advances in Mechatronics and Mechanical Engineering (AMME) Book Series",
  ["adv med"] = "Advances in medicine",
  ["adv med biol"] = "Advances in medicine and biology",
  ["adv med dent sci"] = "Advances in Medical and Dental Sciences",
  ["adv med educ pract"] = "Advances in medical education and practice",
  ["adv med ethics"] = "Advances in medical ethics",
  ["adv med sci"] = "Advances in medical sciences",
  ["adv med sociol"] = "Advances in medical sociology",
  ["adv membr"] = "Advanced Membranes",
  ["adv ment health intellect disabil"] = "Advances in mental health and intellectual disabilities",
  ["adv metab disord"] = "Advances in metabolic disorders",
  ["adv metaheuristics"] = "Advances in Metaheuristics",
  ["adv meteorol"] = "Advances in Meteorology",
  ["adv methods pract psychol sci"] = "Advances in methods and practices in psychological science",
  ["adv microb ecol"] = "Advances in microbial ecology",
  ["adv microb physiol"] = "Advances in microbial physiology",
  ["adv microbiol"] = "Advances in microbiology",
  ["adv mil technol"] = "Advances In Military Technology",
  ["adv mind body med"] = "Advances in mind-body medicine",
  ["adv mod oncol res"] = "Advances in modern oncology research",
  ["adv model anal a"] = "Advances in Modelling and Analysis A",
  ["adv model anal b"] = "Advances in Modelling and Analysis B",
  ["adv model simul eng sci"] = "Advanced Modeling and Simulation in Engineering Sciences",
  ["adv modell anal a"] = "Advances in Modelling and Analysis A",
  ["adv modell anal b"] = "Advances in Modelling and Analysis B",
  ["adv mol pathol"] = "Advances in Molecular Pathology",
  ["adv mol relax interact processes"] = "Advances in Molecular Relaxation and Interaction Processes",
  ["adv molten salt chem"] = "Advances in Molten Salt Chemistry",
  ["adv morphog"] = "Advances in morphogenesis",
  ["adv motiv achiev"] = "Advances in motivation and achievement : a research annual",
  ["adv motiv sci"] = "Advances in motivation science",
  ["adv multimed inf process - pcm 2013 (2013)"] = "Advances in multimedia information processing - PCM 2013 : 14th Pacific-Rim Conference on Multimedia, Nanjing, China, December 13-16, 2013 : proceedings. IEEE Pacific Rim Conference on Multimedia (14th : 2013 : Nanjing, China)",
  ["adv multimedia"] = "Advances in Multimedia",
  ["adv myocardiol"] = "Advances in myocardiology",
  ["adv myochem"] = "Advances in myochemistry",
  ["adv nano res"] = "Advances in nano research",
  ["adv nanobiomed res"] = "Advanced nanobiomed research",
  ["adv nanomed nanotechnol res"] = "Advances in nanomedicine and nanotechnology research",
  ["adv nanopart"] = "Advances in nanoparticles",
  ["adv nat appl sci"] = "Advances in Natural and Applied Science",
  ["adv nat sci: nanosci nanotechnol"] = "Advances in Natural Sciences: Nanoscience and Nanotechnology",
  ["adv neonatal care"] = "Advances in neonatal care : official journal of the National Association of Neonatal Nurses",
  ["adv nephrol necker hosp"] = "Advances in nephrology from the Necker Hospital",
  ["adv neural inf process syst"] = "Advances in neural information processing systems",
  ["adv neurobiol"] = "Advances in neurobiology",
  ["adv neurodev disord"] = "Advances in neurodevelopmental disorders",
  ["adv neuroimmune biol"] = "Advances in neuroimmune biology",
  ["adv neuroimmunol"] = "Advances in neuroimmunology",
  ["adv neurol"] = "Advances in neurology",
  ["adv neurosci (hindawi)"] = "Advances in neuroscience (Hindawi)",
  ["adv neurotoxicol"] = "Advances in neurotoxicology",
  ["adv nonlinear anal"] = "Advances in Nonlinear Analysis",
  ["adv nonlinear stud"] = "Advanced Nonlinear Studies",
  ["adv nonlinear var inequal"] = "Advances in Nonlinear Variational Inequalities. An International Journal",
  ["adv nps pas"] = "Advance for NPs & PAs",
  ["adv nucl phys"] = "Advances in Nuclear Physics",
  ["adv numer anal"] = "Advances in Numerical Analysis",
  ["adv numer math"] = "Advances in Numerical Mathematics",
  ["adv nurs"] = "Advances in nursing",
  ["adv nurs dr educ res"] = "Advances in nursing doctoral education & research",
  ["adv nurse pract"] = "Advance for nurse practitioners",
  ["adv nutr"] = "Advances in nutrition (Bethesda, Md.)",
  ["adv nutr res"] = "Advances in nutritional research",
  ["adv obes weight manag control"] = "Advances in obesity, weight management & control",
  ["adv occup ther pract"] = "Advance for occupational therapy practitioners",
  ["adv oceanogr limnol"] = "Advances in Oceanography and Limnology",
  ["adv oper orthop"] = "Advances In Operative Orthopaedics",
  ["adv oper res"] = "Advances in Operations Research",
  ["adv oper theory"] = "Advances in Operator Theory",
  ["adv ophthalmic plast reconstr surg"] = "Advances in ophthalmic plastic and reconstructive surgery",
  ["adv ophthalmol"] = "Advances in ophthalmology = Fortschritte der Augenheilkunde = Progrès en ophtalmologie",
  ["adv ophthalmol optom"] = "Advances in ophthalmology and optometry",
  ["adv ophthalmol vis syst"] = "Advances in ophthalmology & visual system",
  ["adv opt mater"] = "Advanced Optical Materials",
  ["adv opt photonics"] = "Advances in Optics and Photonics",
  ["adv opt technol"] = "Advanced Optical Technologies",
  ["adv optoelectron"] = "Advances in OptoElectronics",
  ["adv oral biol"] = "Advances in oral biology",
  ["adv org synth"] = "Advances in Organic Synthesis",
  ["adv organ biol"] = "Advances in organ biology",
  ["adv organomet chem"] = "Advances in Organometallic Chemistry",
  ["adv orthop"] = "Advances in orthopedics",
  ["adv otolaryngol"] = "Advances in otolaryngology",
  ["adv otorhinolaryngol"] = "Advances in oto-rhino-laryngology",
  ["adv pain res ther"] = "Advances in pain research and therapy",
  ["adv parallel comput"] = "Advances in Parallel Computing",
  ["adv parasitol"] = "Advances in parasitology",
  ["adv partial differ equ (basel)"] = "Advances in Partial Differential Equations (Basel)",
  ["adv partial differential equations"] = "Advances in Partial Differential Equations",
  ["adv pathobiol"] = "Advances in pathobiology",
  ["adv pediatr"] = "Advances in pediatrics",
  ["adv pediatr infect dis"] = "Advances in pediatric infectious diseases",
  ["adv pediatr res"] = "Advances in pediatric research",
  ["adv perform mater"] = "Advanced Performance Materials",
  ["adv perit dial"] = "Advances in peritoneal dialysis. Conference on Peritoneal Dialysis",
  ["adv pest control res"] = "Advances in pest control research",
  ["adv pharm bull"] = "Advanced pharmaceutical bulletin",
  ["adv pharm sci"] = "Advances in pharmaceutical sciences",
  ["adv pharmacoepidemiol drug saf"] = "Advances in pharmacoepidemiology & drug safety",
  ["adv pharmacol"] = "Advances in pharmacology (San Diego, Calif.)",
  ["adv pharmacol chemother"] = "Advances in pharmacology and chemotherapy",
  ["adv pharmacol clin trials"] = "Advances in pharmacology and clinical trials",
  ["adv pharmacol pharm sci"] = "Advances in pharmacological and pharmaceutical sciences",
  ["adv pharmacol sci"] = "Advances in pharmacological sciences",
  ["adv photochem"] = "Advances in Photochemistry",
  ["adv photonics"] = "Advanced Photonics",
  ["adv photonics nexus"] = "Advanced Photonics Nexus",
  ["adv photonics res"] = "Advanced photonics research",
  ["adv phys"] = "Advances in Physics",
  ["adv phys org chem"] = "Advances in Physical Organic Chemistry",
  ["adv phys res"] = "Advanced Physics Research",
  ["adv phys ther rehab med"] = "Advance for physical therapy and rehab medicine",
  ["adv phys x"] = "Advances in physics: X",
  ["adv phys: x"] = "Advances in Physics: X",
  ["adv physiol educ"] = "Advances in physiology education",
  ["adv physiother"] = "Advances in physiotherapy",
  ["adv pineal res"] = "Advances in pineal research",
  ["adv plan parent"] = "Advances in planned parenthood",
  ["adv planar lipid bilayers liposomes"] = "Advances in planar lipid bilayers and liposomes",
  ["adv plant pathol"] = "Advances in Plant Pathology",
  ["adv polar sci"] = "Advances in polar science",
  ["adv polym sci"] = "Advances in Polymer Science",
  ["adv polym tech"] = "Advances in Polymer Technology",
  ["adv polym technol"] = "Advances in Polymer Technology",
  ["adv popul"] = "Advances in population : psychosocial perspectives",
  ["adv powder mater"] = "Advanced Powder Materials",
  ["adv powder metall part mater"] = "Advances in Powder Metallurgy and Particulate Materials",
  ["adv powder technol"] = "Advanced Powder Technology",
  ["adv power electron"] = "Advances in Power Electronics",
  ["adv pract nurs q"] = "Advanced practice nursing quarterly",
  ["adv pract nurse"] = "Advanced practice nurse : APN",
  ["adv prev med"] = "Advances in preventive medicine",
  ["adv probab theory"] = "Advances in Probability Theory",
  ["adv prod eng manage"] = "Advances in Production Engineering & Management",
  ["adv prostaglandin thromboxane leukot res"] = "Advances in prostaglandin, thromboxane, and leukotriene research",
  ["adv prostaglandin thromboxane res"] = "Advances in prostaglandin and thromboxane research",
  ["adv protein chem"] = "Advances in Protein Chemistry",
  ["adv protein chem struct biol"] = "Advances in protein chemistry and structural biology",
  ["adv protein phosphatases"] = "Advances in protein phosphatases",
  ["adv psych"] = "Advances in Psychology",
  ["adv psychiatr treat"] = "Advances in psychiatric treatment : the Royal College of Psychiatrists' journal of continuing professional development",
  ["adv psychobiol"] = "Advances in psychobiology",
  ["adv psychol res"] = "Advances in psychology research",
  ["adv psychol study"] = "Advances in psychology study",
  ["adv psychosom med"] = "Advances in psychosomatic medicine",
  ["adv public health"] = "Advances in public health",
  ["adv public inter acc"] = "Extending Schumacher's Concept of Total Accounting and Accountability Into The 21st Century",
  ["adv pulm hypertens"] = "Advances in pulmonary hypertension",
  ["adv pure appl math"] = "Advances in Pure and Applied Mathematics",
  ["adv quant anal finance account"] = "Advances in Quantitative Analysis of Finance and Accounting",
  ["adv quant tech soc sci"] = "Advanced Quantitative Techniques in the Social Sciences",
  ["adv quantum chem"] = "Advances in Quantum Chemistry",
  ["adv quantum technol"] = "Advanced Quantum Technologies",
  ["adv radiat biol"] = "Advances in radiation biology",
  ["adv radiat oncol"] = "Advances in radiation oncology",
  ["adv radio sci"] = "Advances in Radio Science",
  ["adv radiol"] = "Advances in radiology",
  ["adv rare dis"] = "Advances in rare diseases",
  ["adv redox res"] = "Advances in Redox Research",
  ["adv regener biol"] = "Advances in Regenerative Biology",
  ["adv ren replace ther"] = "Advances in renal replacement therapy",
  ["adv reprod physiol"] = "Advances in reproductive physiology",
  ["adv res"] = "Advances in research",
  ["adv res comput softw sci"] = "Advanced Research in Computing and Software Science",
  ["adv res endocrinol metab"] = "Advance research on endocrinology & metabolism",
  ["adv res gastroenterol hepatol"] = "Advanced research in gastroenterology & hepatology",
  ["adv respir med"] = "Advances in respiratory medicine",
  ["adv rheumatol"] = "Advances in rheumatology (London, England)",
  ["adv rob"] = "Advanced Robotics",
  ["adv rob res"] = "Advanced Robotics Research",
  ["adv robot"] = "Advanced robotics : the international journal of the Robotics Society of Japan",
  ["adv sample prep"] = "Advances in Sample Preparation",
  ["adv sch ment health promot"] = "Advances in school mental health promotion",
  ["adv sci"] = "Advancement of science",
  ["adv sci (weinh)"] = "Advanced science (Weinheim, Baden-Württemberg, Germany)",
  ["adv sci eng med"] = "Advanced science, engineering and medicine",
  ["adv sci lett"] = "Advanced Science Letters",
  ["adv sci res"] = "Advances in Science and Research",
  ["adv sci tech ussr chem ser"] = "Advances in Science and Technology in the USSR: Chemistry Series",
  ["adv sci technol"] = "Advances in science and technology (Owerri, Nigeria)",
  ["adv second messenger phosphoprotein res"] = "Advances in Second Messenger and Phosphoprotein Research",
  ["adv sens energy mater"] = "Advanced Sensor and Energy Materials",
  ["adv sens res"] = "Advanced Sensor Research",
  ["adv ser appl phys"] = "Advanced Series in Applied Physics",
  ["adv ser dir high energy phys"] = "Advanced Series on Directions in High Energy Physics",
  ["adv ser dynam systems"] = "Advanced Series in Dynamical Systems",
  ["adv ser electr comput eng"] = "Advanced Series in Electrical and Computer Engineering",
  ["adv ser math phys"] = "Advanced Series in Mathematical Physics",
  ["adv ser math psychol"] = "Advanced Series on Mathematical Psychology",
  ["adv ser math sci eng"] = "Advanced Series in Mathematical Science and Engineering",
  ["adv ser neurosci"] = "Advanced Series in Neuroscience",
  ["adv ser nonlinear dynam"] = "Advanced Series in Nonlinear Dynamics",
  ["adv ser stat sci appl probab"] = "Advanced Series on Statistical Science & Applied Probability",
  ["adv ser theoret phys sci"] = "Advanced Series on Theoretical Physical Science",
  ["adv sex horm res"] = "Advances in sex hormone research",
  ["adv shock res"] = "Advances in shock research",
  ["adv simul"] = "Advances In Simulation",
  ["adv simul (lond)"] = "Advances in simulation (London, England)",
  ["adv skin wound care"] = "Advances in skin & wound care",
  ["adv small anim care"] = "Advances in Small Animal Care",
  ["adv small anim med surg"] = "Advances in small animal medicine and surgery",
  ["adv soc sci cult"] = "Advances in social science and culture",
  ["adv soc sci res j"] = "Advances in social sciences research journal",
  ["adv soc work"] = "Advances in social work",
  ["adv sociodent res"] = "Advances in socio-dental research",
  ["adv space biol med"] = "Advances in space biology and medicine",
  ["adv space res"] = "Advances in Space Research",
  ["adv spat sci"] = "Advances in Spatial Science",
  ["adv spectrosc"] = "Advances in Spectroscopy",
  ["adv stat"] = "Advances in statistics",
  ["adv stat anal"] = "Advances in statistical analysis : AStA : a journal of the German Statistical Society",
  ["adv stat climatol meteorol oceanogr"] = "Advances in Statistical Climatology, Meteorology and Oceanography",
  ["adv steel constr"] = "Advanced Steel Construction",
  ["adv stem cells"] = "Advances in stem cells",
  ["adv steroid biochem pharmacol"] = "Advances in steroid biochemistry and pharmacology",
  ["adv struct chem imaging"] = "Advanced structural and chemical imaging",
  ["adv struct eng"] = "Advances in Structural Engineering",
  ["adv struct mater"] = "Advanced Structured Materials",
  ["adv stud biol"] = "Advanced studies in biology",
  ["adv stud complex syst"] = "Advanced Studies in Complex Systems",
  ["adv stud euro-tbil math j"] = "Advanced Studies. Euro-Tbilisi Mathematical Journal",
  ["adv stud pure math"] = "Advanced Studies in Pure Mathematics",
  ["adv stud theor phys"] = "Advanced Studies in Theoretical Physics",
  ["adv stud theoret appl econometrics"] = "Advanced Studies in Theoretical and Applied Econometrics",
  ["adv stud: euro-tbilisi math j"] = "Advanced Studies: Euro-Tbilisi Mathematical Journal",
  ["adv study behav"] = "Advances in the study of behavior",
  ["adv surg"] = "Advances in surgery",
  ["adv sustain syst"] = "Advanced sustainable systems",
  ["adv sustainable syst"] = "Advanced Sustainable Systems",
  ["adv synchrotron radiat"] = "Advances in Synchrotron Radiation",
  ["adv synth catal"] = "Advanced Synthesis and Catalysis",
  ["adv syst sci appl"] = "Advances in systems science and applications",
  ["adv tech biol med"] = "Advanced techniques in biology & medicine",
  ["adv tech stand neurosurg"] = "Advances and technical standards in neurosurgery",
  ["adv textb control signal process"] = "Advanced Textbooks in Control and Signal Processing",
  ["adv textb math"] = "Advanced Textbooks in Mathematics",
  ["adv textb phys"] = "Advanced Textbooks in Physics",
  ["adv texts econometrics"] = "Advanced Texts in Econometrics",
  ["adv theor math phys"] = "Advances in Theoretical and Mathematical Physics",
  ["adv theor simul"] = "Advanced Theory and Simulations",
  ["adv theory simul"] = "Advanced theory and simulations",
  ["adv ther"] = "Advances in therapy",
  ["adv ther (weinh)"] = "Advanced therapeutics",
  ["adv top math"] = "Advanced Topics in Mathematics",
  ["adv topics mech engrg ser"] = "Advanced Topics in Mechanical Engineering Series",
  ["adv tracer methodol"] = "Advances in tracer methodology",
  ["adv transp stud"] = "Advances in Transportation Studies",
  ["adv trends soft comput (2013)"] = "Advance trends in soft computing : proceedings of WCSC 2013, December 16-18, San Antonio, Texas, USA. WCSC (Conference : Soft computing) (2013 : San Antonio, Tex.)",
  ["adv tribol"] = "Advances in Tribology",
  ["adv tuberc res"] = "Advances in tuberculosis research. Fortschritte der Tuberkuloseforschung. Progrès de l'exploration de la tuberculose",
  ["adv tumor virol"] = "Advances in tumor virology",
  ["adv underwat technol"] = "Subsea Control and Data Acquisition for Oil and Gas Production Systems",
  ["adv urol"] = "Advances in urology",
  ["adv vasc med"] = "Advances in vascular medicine",
  ["adv vet med"] = "Advances in veterinary medicine",
  ["adv vet sci"] = "Advances in veterinary science",
  ["adv vet sci comp med"] = "Advances in veterinary science and comparative medicine",
  ["adv virol"] = "Advances in virology",
  ["adv virus res"] = "Advances in virus research",
  ["adv water res"] = "Advances in Water Resources",
  ["adv water resour"] = "Advances in water resources",
  ["adv wound care"] = "Advances in wound care : the journal for prevention and healing",
  ["adv wound care (new rochelle)"] = "Advances in wound care",
  ["adv x-ray anal"] = "Advances in X-Ray Analysis",
  ["advanc cons interest"] = "Advancing the Consumer Interest",
  ["advent herit"] = "Adventist heritage",
  ["advers resil sci"] = "Adversity and resilience science",
  ["adverse drug react acute poisoning rev"] = "Adverse drug reactions and acute poisoning reviews",
  ["adverse drug react bull"] = "Adverse drug reaction bulletin",
  ["adverse drug react toxicol rev"] = "Adverse Drug Reactions and Toxicological Reviews",
  ["advert age"] = "Advertising age",
  ["advocacy now"] = "Advocacy now",
  ["aea pap proc"] = "AEA papers and proceedings. American Economic Association",
  ["aecu rep"] = "AECU [reports]. U.S. Atomic Energy Commission",
  ["aehe iv"] = "Annuaire de l’école pratique des Hautes études, IV’ sect., Sciences hist. & philol",
  ["aehe v"] = "Annuaire de l’école pratique des Hautes études, Ve sect., Sciences religieuses",
  ["aem educ train"] = "AEM education and training",
  ["aeolian res"] = "Aeolian Research",
  ["aequ math"] = "Aequationes mathematicae",
  ["aequationes math"] = "Aequationes Mathematicae",
  ["aer j"] = "AER journal : research and practice in visual impairment and blindness",
  ["aera open"] = "AERA open",
  ["aere perennius"] = "Aere perennius",
  ["aerobiologia (bologna)"] = "Aerobiologia",
  ["aeromed acta"] = "Aeromedica acta",
  ["aeromed rev"] = "Aeromedical reviews",
  ["aeronaut astronaut"] = "L'Aéronautique et l'astronautique : organe commun de l'Association française des ingénieurs et techniciens de l'aéronautique de l'espace (A.F.I.T.A.E.) et de la Société française d'astronautique (S.F.A.)",
  ["aeronaut j"] = "Aeronautical Journal",
  ["aeronaut q"] = "Aeronautical Quarterly",
  ["aerosol air qual res"] = "Aerosol and air quality research",
  ["aerosol res"] = "Aerosol Research",
  ["aerosol sci eng"] = "Aerosol Science and Engineering",
  ["aerosol sci technol"] = "Aerosol Science and Technology",
  ["aerosp am"] = "Aerospace America",
  ["aerosp eng"] = "Aerospace engineering (Society of Automotive Engineers)",
  ["aerosp hist"] = "Aerospace historian",
  ["aerosp med"] = "Aerospace medicine",
  ["aerosp med hum perform"] = "Aerospace medicine and human performance",
  ["aerosp sci technol"] = "Aerospace Science and Technology",
  ["aerosp ser"] = "Aerospace Series",
  ["aerosp syst"] = "Aerospace Systems",
  ["aerospace (basel)"] = "Aerospace (Basel, Switzerland)",
  ["aerotec missili spazio"] = "Aerotecnica Missili & Spazio",
  ["aerztlicher feldpostbrief"] = "Aerztlicher Feldpostbrief",
  ["aesthet surg j"] = "Aesthetic surgery journal",
  ["aesthet surg j open forum"] = "Aesthetic surgery journal. Open forum",
  ["aesthetic plast surg"] = "Aesthetic plastic surgery",
  ["aesthetic surg j"] = "Aesthetic Surgery Journal",
  ["aeu int j electron commun"] = "AEU International Journal of Electronics and Communications",
  ["afcs nat mol pages"] = "The AFCS-nature molecule pages",
  ["afd inform"] = "Afdeling Informatica",
  ["afd numer wisk"] = "Afdeling Numerieke Wiskunde",
  ["aff soc int"] = "Affari sociali internazionali",
  ["affect sci"] = "Affective science",
  ["afghan med j"] = "Afghan medical journal",
  ["afips conf proc"] = "AFIPS Conference Proceedings",
  ["afl cio am fed"] = "The AFL-CIO American federationist. AFL-CIO",
  ["afmdc tr"] = "AFMDC TR. Air Force Missile Development Center (U.S.)",
  ["afosr tn united states air force off sci res"] = "AFOSR TN. United States. Air Force. Office of Scientific Research",
  ["afosr tr united states air force off sci res"] = "AFOSR TR. United States. Air Force. Office of Scientific Research",
  ["afr aff (lond)"] = "African affairs",
  ["afr alternat"] = "African alternatives",
  ["afr anal"] = "Africa analysis",
  ["afr anthropol"] = "African anthropology : journal of the Pan African Anthropological Association = Anthropologie africaine : revue de l'Association panafricaine de l'anthropologie",
  ["afr archaeol rev"] = "The African archaeological review",
  ["afr asie mod"] = "L'Afrique et l'Asie modernes",
  ["afr brain j"] = "African Brain Journal",
  ["afr bull"] = "Africana bulletin (Boston, Mass.)",
  ["afr concord"] = "African concord",
  ["afr contemp"] = "Afrique contemporaine",
  ["afr crop sci j"] = "African crop science journal",
  ["afr demogr"] = "African demography",
  ["afr dent j"] = "African dental journal : official publication of the Federation of African Dental Associations = Journal dentaire africain",
  ["afr dev"] = "Africa development. Afrique et développement",
  ["afr dev rev"] = "African development review = Revue africaine de développement",
  ["afr diaspora j math"] = "African Diaspora Journal of Mathematics",
  ["afr diaspora j math new ser"] = "African Diaspora Journal of Mathematics. New Series",
  ["afr doc"] = "Afrique documents",
  ["afr econ hist"] = "African economic history",
  ["afr entomol"] = "African entomology : journal of the Entomological Society of Southern Africa",
  ["afr environ"] = "African environment",
  ["afr farmer"] = "African farmer : the key to Africa's future",
  ["afr fr chir"] = "Afrique française chirurgicale",
  ["afr francaise chir"] = "Afrique francaise chirurgicale",
  ["afr geogr rev"] = "African Geographical Review",
  ["afr health"] = "Africa health",
  ["afr health sci"] = "African health sciences",
  ["afr hist stud"] = "African historical studies",
  ["afr insight"] = "Africa insight",
  ["afr invertebr"] = "African invertebrates",
  ["afr j agric res"] = "African journal of agricultural research",
  ["afr j aids res"] = "African journal of AIDS research : AJAR",
  ["afr j appl res"] = "African Journal of Applied Research",
  ["afr j appl stat"] = "African Journal of Applied Statistics",
  ["afr j aquat sci"] = "African Journal of Aquatic Science",
  ["afr j biochem res"] = "African Journal of Biochemistry Research",
  ["afr j biol sci"] = "African Journal of Biological Sciences",
  ["afr j biomed res"] = "African journal of biomedical research : AJBR",
  ["afr j biotechnol"] = "African journal of biotechnology",
  ["afr j cell pathol"] = "African journal of cellular pathology",
  ["afr j chem educ"] = "African Journal of Chemical Education",
  ["afr j clin exp microbiol"] = "African Journal of Clinical and Experimental Microbiology",
  ["afr j comput ict"] = "African journal of computing and ICT",
  ["afr j disabil"] = "African journal of disability",
  ["afr j drug alcohol stud"] = "African journal of drug and alcohol studies",
  ["afr j ecol"] = "African journal of ecology",
  ["afr j econ sustainable dev"] = "African Journal of Economic and Sustainable Development",
  ["afr j emerg med"] = "African journal of emergency medicine : Revue africaine de la médecine d'urgence",
  ["afr j eng technol"] = "African Journal of Engineering & Technology",
  ["afr j environ health sci"] = "African journal of environmental health sciences",
  ["afr j environ sci tech"] = "African journal of environmental science and technology",
  ["afr j environ sci technol"] = "African Journal of Environmental Science and Technology",
  ["afr j fertil sexual reprod heal"] = "African journal of fertility, sexuality, and reproductive health",
  ["afr j food agric nutr dev"] = "African Journal of Food, Agriculture, Nutrition and Development",
  ["afr j food sci"] = "African Journal of Food Science",
  ["afr j health prof educ"] = "African journal of health professions education",
  ["afr j health sci"] = "African journal of health sciences",
  ["afr j herpetol"] = "African journal of herpetology : journal of the Herpetological Association of Africa",
  ["afr j infect dis"] = "African journal of infectious diseases",
  ["afr j lab med"] = "African journal of laboratory medicine",
  ["afr j mar sci"] = "African Journal of Marine Science",
  ["afr j med health sci"] = "African journal of medical and health sciences",
  ["afr j med med sci"] = "African journal of medicine and medical sciences",
  ["afr j med pract"] = "African journal of medical practice",
  ["afr j med sci"] = "The African journal of medical sciences",
  ["afr j microbiol res"] = "African journal of microbiology research",
  ["afr j midwifery womens health"] = "African journal of midwifery and women's health",
  ["afr j mycol biotechnol"] = "The African journal of mycology and biotechnology",
  ["afr j nephrol"] = "African journal of nephrology : official publication of the African Association of Nephrology. African Association of Nephrology",
  ["afr j nurs midwifery"] = "Africa journal of nursing and midwifery",
  ["afr j paediatr surg"] = "African journal of paediatric surgery : AJPS",
  ["afr j pharm pharmacol"] = "African journal of pharmacy and pharmacology",
  ["afr j phys health educ recreat dance"] = "African journal for physical health education, recreation, and dance",
  ["afr j plant sci"] = "African Journal of Plant Science",
  ["afr j prim health care fam med"] = "African journal of primary health care & family medicine",
  ["afr j psychiatry"] = "The African journal of psychiatry",
  ["afr j psychiatry (johannesbg)"] = "African journal of psychiatry",
  ["afr j psychol study soc issues"] = "African journal for the psychological study of social issues : journal of the African Society for the Psychological Study of Social Issues",
  ["afr j range forage sci"] = "African Journal of Range and Forage Science",
  ["afr j reprod health"] = "African journal of reproductive health",
  ["afr j res math sci technol educ"] = "African Journal of Research in Mathematics, Science and Technology Education",
  ["afr j root tuber crops"] = "African journal of root and tuber crops",
  ["afr j sci innovation dev"] = "African Journal of Science, Technology, Innovation and Development",
  ["afr j sex transmi dis"] = "The African journal of sexually transmitted diseases",
  ["afr j sociol"] = "African journal of sociology",
  ["afr j tradit complement altern med"] = "African journal of traditional, complementary, and alternative medicines : AJTCAM",
  ["afr j trauma stress"] = "African journal of traumatic stress",
  ["afr j urol"] = "African journal of urology : the official journal of the Pan African Urological Surgeons' Association (PAUSA)",
  ["afr j wildl res"] = "African journal of wildlife research",
  ["afr link"] = "Africa link : a publication of the Africa Region",
  ["afr litt artist"] = "L'Afrique littéraire et artistique",
  ["afr marburgen"] = "Africana Marburgensia",
  ["afr mat"] = "Afrika Matematika",
  ["afr math ann afma"] = "Africa Mathematics Annals. AFMA. A Pan African Journal of Mathematics",
  ["afr med"] = "Afrique médicale",
  ["afr media rev"] = "Africa media review",
  ["afr notes"] = "African notes : bulletin of the Institute of African Studies, University of Ibadan",
  ["afr notes news"] = "Africana notes and news. Africana aantekeninge en nuus",
  ["afr plant prot"] = "African plant protection",
  ["afr popul dev bull"] = "Africa's population and development bulletin",
  ["afr popul newsl"] = "African population newsletter",
  ["afr primates"] = "African primates",
  ["afr q"] = "Africa quarterly",
  ["afr recovery"] = "Africa recovery",
  ["afr rep"] = "Africa report",
  ["afr res bull"] = "Africana research bulletin",
  ["afr res doc"] = "African research & documentation",
  ["afr rev"] = "African review (Dar es Salaam, Tanzania)",
  ["afr soc res"] = "African social research",
  ["afr south"] = "Africa South of the Sahara",
  ["afr stat"] = "Afrika Statistika",
  ["afr stat j"] = "African statistical journal = Journal statistique africain",
  ["afr stud"] = "African studies",
  ["afr stud rev"] = "African studies review",
  ["afr study monogr"] = "African study monographs",
  ["afr today"] = "Africa today",
  ["afr uebersee"] = "Afrika und Übersee",
  ["afr urban notes"] = "African urban notes",
  ["afr urban notes b"] = "African urban notes. Ser. B",
  ["afr urban q"] = "African urban quarterly",
  ["afr urban stud"] = "African urban studies (Michigan State University. African Studies Center)",
  ["afr woman"] = "Africa woman",
  ["afr women health"] = "Africa women & health : a Safe Motherhood magazine",
  ["afr zool"] = "African zoology",
  ["africa (lond)"] = "Africa : journal of the International Institute of African Languages and Cultures",
  ["africa (roma)"] = "Africa : notiziario dell’Associazione fra le imprese italiane in Africa",
  ["african devel rev"] = "African Development Review",
  ["african econ hist"] = "African Economic History",
  ["african finance j"] = "African Finance Journal",
  ["afrika mat (3)"] = "Afrika Matematika",
  ["afrika spec"] = "Afrika Spectrum",
  ["afrique 2000"] = "Afrique 2000 (Geneva, Switzerland)",
  ["afro am n y life hist"] = "Afro-Americans in New York life and history",
  ["afro tech pap"] = "AFRO technical papers",
  ["against grain (charlest)"] = "Against the grain (Charleston, S.C.)",
  ["agal res"] = "Agal Research",
  ["age (dordr)"] = "Age (Dordrecht, Netherlands)",
  ["age (omaha)"] = "Age",
  ["age ageing"] = "Age and ageing",
  ["age dordr neth"] = "Age (Dordrecht, Netherlands)",
  ["aged care serv rev"] = "Aged care & services review",
  ["ageing int"] = "Ageing international",
  ["ageing res"] = "Ageing research",
  ["ageing res rev"] = "Ageing research reviews",
  ["ageing soc"] = "Ageing and society",
  ["agenda salud"] = "Agenda salud",
  ["agent dir simul symp"] = "Agent-Directed Simulation Symposium (ADS). Agent-directed Simulation Symposium",
  ["agents actions"] = "Agents and actions",
  ["agents actions suppl"] = "Agents and actions. Supplements",
  ["aggiorn clinico ter"] = "Aggiornamenti clinicoterapeutici",
  ["aggiorn fisiol"] = "Aggiornamenti di fisiologia",
  ["aggiorn ostet ginecol"] = "Aggiornamenti in ostetricia e ginecologia",
  ["aggiorn pediatr"] = "Aggiornamento pediatrico",
  ["aggiorn sulle mal infez"] = "Aggiornamenti sulle malattie da infezione",
  ["aggiorn ter oftalmol"] = "Aggiornamenti di terapia oftalmologica",
  ["aggress behav"] = "Aggressive behavior",
  ["aggress violent behav"] = "Aggression and violent behavior",
  ["agid geo in"] = "Agid Report Series : Geosciences in International Development",
  ["aging (albany ny)"] = "Aging",
  ["aging (milano)"] = "Aging (Milan, Italy)",
  ["aging brain"] = "Aging brain",
  ["aging cancer"] = "Aging and cancer",
  ["aging cell"] = "Aging cell",
  ["aging clin exp res"] = "Aging clinical and experimental research",
  ["aging dis"] = "Aging and disease",
  ["aging health"] = "Aging health",
  ["aging health res"] = "Aging and health research",
  ["aging leis living"] = "Aging & leisure living",
  ["aging male"] = "The aging male : the official journal of the International Society for the Study of the Aging Male",
  ["aging med (milton)"] = "Aging medicine (Milton (N.S.W))",
  ["aging ment health"] = "Aging & mental health",
  ["aging neurodegener"] = "Aging and neurodegeneration",
  ["aging pathobiol ther"] = "Aging pathobiology and therapeutics",
  ["aging soc"] = "Aging and society : an interdisciplinary journal",
  ["aging today"] = "Aging today : the bimonthly newspaper of the American Society on Aging",
  ["aging trends"] = "Aging trends (Hyattsville, Md.)",
  ["aging work"] = "Aging and work",
  ["agnes karll schwest krankenpfleger"] = "Die Agnes Karll-Schwester, der Krankenpfleger",
  ["agora heterodoxias"] = "Agora de heterodoxias",
  ["agora math"] = "Agora Mathematica",
  ["agr econ"] = "Agricultural Economics",
  ["agr econ res"] = "Agricultural Economics Research",
  ["agr finance rev"] = "Agricultural Finance Review",
  ["agr human values"] = "Agriculture and Human Values",
  ["agr resource econ rev"] = "Agricultural and Resource Economics Review",
  ["agr resources quart"] = "Agriculture and Resources Quarterly",
  ["agrarforsch schweiz"] = "Agrarforschung Schweiz",
  ["agratort szle"] = "Agrártörténeti szemle = Historia rerum rusticarum",
  ["agri gene"] = "Agri gene",
  ["agribusiness (n y n y)"] = "Agribusiness (New York, N.Y.)",
  ["agric agric sci procedia"] = "Agriculture and Agricultural Science Procedia",
  ["agric biol"] = "Agricultural biology",
  ["agric biol chem"] = "Agricultural and biological chemistry",
  ["agric econ"] = "Agricultural economics (Amsterdam, Netherlands)",
  ["agric econ res"] = "Agricultural economics research",
  ["agric ecosyst environ"] = "Agriculture Ecosystems and Environment",
  ["agric eng"] = "Agricultural Engineering",
  ["agric environ"] = "Agriculture and Environment",
  ["agric environ lett"] = "Agricultural & Environmental Letters",
  ["agric food econ"] = "Agricultural and Food Economics",
  ["agric food sci"] = "Agricultural and Food Science",
  ["agric food secur"] = "Agriculture & food security",
  ["agric for bull"] = "Agriculture and Forestry Bulletin",
  ["agric for entomol"] = "Agricultural and Forest Entomology",
  ["agric for meteorol"] = "Agricultural and Forest Meteorology",
  ["agric handb"] = "Agriculture Handbook",
  ["agric hist"] = "Agricultural history",
  ["agric hist rev"] = "The Agricultural history review",
  ["agric hum values"] = "Agriculture and Human Values",
  ["agric human values"] = "Agriculture and human values",
  ["agric inf bull usda for serv"] = "Agricultural Information Bulletin USDA Forest Service",
  ["agric inf dev bull"] = "Agricultural information development bulletin",
  ["agric ital"] = "Agricoltura Italiana",
  ["agric meteorol"] = "Agricultural Meteorology",
  ["agric nat resour"] = "Agriculture and Natural Resources",
  ["agric res"] = "Agricultural Research",
  ["agric res technol"] = "Agricultural research & technology : open access journal",
  ["agric resour econ rev"] = "Agricultural and resource economics review",
  ["agric sci china"] = "Agricultural sciences in China",
  ["agric sci dev"] = "Agriculture science developments",
  ["agric sci j"] = "Agricultural science journal",
  ["agric syst"] = "Agricultural systems",
  ["agric technol (thail)"] = "Agricultural technology",
  ["agric ticin"] = "Agricoltore ticinese",
  ["agric wastes"] = "Agricultural Wastes",
  ["agric water manag"] = "Agricultural water management",
  ["agric water manage"] = "Agricultural Water Management",
  ["agriculture (basel)"] = "Agriculture (Basel, Switzerland)",
  ["agro food ind hi tech"] = "Agro food industry hi-tech",
  ["agroecol sustainable food syst"] = "Agroecology and Sustainable Food Systems",
  ["agrofor syst"] = "Agroforestry Systems",
  ["agrolife sci j"] = "AgroLife Scientific Journal",
  ["agron ambient"] = "Agronomía & ambiente",
  ["agron colomb"] = "Agronomía Colombiana",
  ["agron j"] = "Agronomy journal",
  ["agron mesoam"] = "Agronomía mesoamericana : órgano divulgativo del PCCMCA, Programa Cooperativo Centroamericano de Mejoramiento de Cultivos y Animales",
  ["agron sustain dev"] = "Agronomy for sustainable development",
  ["agron sustainable dev"] = "Agronomy for Sustainable Development",
  ["agronomia mesoamericana"] = "Agronomia Mesoamericana",
  ["agronomy (basel)"] = "Agronomy (Basel, Switzerland)",
  ["agrosyst geosci environ"] = "Agrosystems, Geosciences & Environment",
  ["agrég math"] = "Agrégation de Mathématiques",
  ["agso j aust geol geophys"] = "AGSO Journal of Australian Geology and Geophysics",
  ["agu adv"] = "AGU Advances",
  ["aha hosp technol ser"] = "AHA hospital technology series",
  ["ahfad j"] = "The Ahfad journal",
  ["ahip cover"] = "AHIP Coverage",
  ["ahme j"] = "AHME journal",
  ["ahot beyisrael"] = "ha-Ahot be-Yisrael",
  ["ahp j"] = "AHP journal",
  ["ahrq webm&m"] = "AHRQ WebM&M : morbidity & mortality rounds on the Web",
  ["ai bian ji bian tu bian"] = "Ai bian, ji bian, tu bian",
  ["ai civ eng"] = "AI in Civil Engineering",
  ["ai commun"] = "AI Communications",
  ["ai edam"] = "AI EDAM",
  ["ai ethics"] = "AI and Ethics",
  ["ai mag"] = "AI magazine",
  ["ai matters"] = "AI matters",
  ["ai open"] = "AI Open",
  ["ai soc"] = "AI & society",
  ["ai switz"] = "AI Switzerland",
  ["ai zheng"] = "Ai zheng = Aizheng = Chinese journal of cancer",
  ["aia j"] = "AIA journal. American Institute of Architects",
  ["aiaa educ ser"] = "AIAA Education Series",
  ["aiaa j"] = "AIAA journal. American Institute of Aeronautics and Astronautics",
  ["aiaa j air transp"] = "AIAA Journal of Air Transportation",
  ["aiaa stud j"] = "AIAA student journal. American Institute of Aeronautics and Astronautics",
  ["aicc econ rev"] = "AICC economic review. Indian National Congress. All India Congress Committee",
  ["aiccm bull"] = "AICCM Bulletin",
  ["aiche j"] = "AIChE journal. American Institute of Chemical Engineers",
  ["aiche symp ser"] = "AIChE Symposium Series",
  ["aichi gakuin daigaku shigakkai shi"] = "Aichi Gakuin Daigaku Shigakkai shi",
  ["aichi gakuin dent sci"] = "Aichi-Gakuin dental science",
  ["aid eval news"] = "A.I.D. evaluation news",
  ["aidos news"] = "AIDoS news : bimonthly bulletin of the Italian Association for Women in Development. Associazione italiana donne per lo sviluppo",
  ["aids 2006"] = "AIDS 2006 : time to deliver : abstract CD-ROM. International Conference on AIDS (16th : 2006 : Toronto, Canada)",
  ["aids 2008"] = "AIDS 2008 : abstract CD-ROM, XVII International AIDS Conference, Mexico City, Mexico, 3-8 August 2008. International Conference on AIDS (17th : 2008 : Mexico City)",
  ["aids 2010"] = "XVIII International AIDS Conference, 2010 : abstracts. International Conference on AIDS (18th : 2010 : Vienna, Austria)",
  ["aids action"] = "AIDS action",
  ["aids action policy brief"] = "AIDS Action policy brief",
  ["aids action update"] = "AIDS action update",
  ["aids alert"] = "AIDS alert",
  ["aids anal afr"] = "AIDS analysis Africa",
  ["aids anal asia"] = "AIDS analysis Asia",
  ["aids asia"] = "Aids Asia : voice of the Asian solidarity against AIDS",
  ["aids behav"] = "AIDS and behavior",
  ["aids care"] = "AIDS care",
  ["aids clin care"] = "AIDS clinical care",
  ["aids clin rev"] = "AIDS clinical review",
  ["aids educ prev"] = "AIDS education and prevention : official publication of the International Society for AIDS Education",
  ["aids forsch"] = "AIDS-Forschung : AIFO = Acquired immune deficiency syndrome research",
  ["aids health promot exch"] = "AIDS health promotion exchange",
  ["aids illus"] = "AIDS illustrated",
  ["aids inf exch"] = "AIDS information exchange",
  ["aids inst newsl"] = "AIDS Institute newsletter. New York (State). AIDS Institute",
  ["aids lit law rev"] = "AIDS literature & law review",
  ["aids newsl"] = "AIDS newsletter (London, England)",
  ["aids patient care"] = "AIDS patient care",
  ["aids patient care stds"] = "AIDS patient care and STDs",
  ["aids policy law"] = "AIDS policy & law",
  ["aids public policy j"] = "AIDS & public policy journal",
  ["aids read"] = "The AIDS reader",
  ["aids res"] = "AIDS research",
  ["aids res hum retroviruses"] = "AIDS research and human retroviruses",
  ["aids res ther"] = "AIDS research and therapy",
  ["aids res treat"] = "AIDS research and treatment",
  ["aids rev"] = "AIDS reviews",
  ["aids soc"] = "AIDS & society",
  ["aids std health promot exch"] = "AIDS/STD health promotion exchange",
  ["aids treat news"] = "AIDS treatment news",
  ["aids updates"] = "AIDS updates",
  ["aids vaccine 2001"] = "AIDS vaccine 2001 : September 5-8, 2001, Philadelphia Marriott, Philadelphia, PA, USA",
  ["aids watch"] = "AIDS watch",
  ["aids wkly"] = "AIDS weekly",
  ["aids wkly plus"] = "AIDS weekly plus",
  ["aidsed newsl"] = "AIDSED newsletter : newsletter of the WHO/Unesco AIDS Education and Health Promotion Materials Exchange Centre for Asia and the Pacific",
  ["aiha j"] = "AIHA Journal",
  ["aiha j (fairfax, va)"] = "AIHA journal : a journal for the science of occupational and environmental health and safety",
  ["aims agric food"] = "AIMS Agriculture and Food",
  ["aims allergy immunol"] = "AIMS allergy and immunology",
  ["aims bioeng"] = "AIMS bioengineering",
  ["aims biophys"] = "AIMS biophysics",
  ["aims electron electr eng"] = "AIMS Electronics and Electrical Engineering",
  ["aims energy"] = "AIMS Energy",
  ["aims environ sci"] = "AIMS environmental science",
  ["aims genet"] = "AIMS genetics",
  ["aims geosci"] = "AIMS geosciences",
  ["aims libr ser"] = "African Institute of Mathematics (AIMS) Library Series",
  ["aims mater sci"] = "AIMS materials science",
  ["aims math"] = "AIMS Mathematics",
  ["aims med sci"] = "AIMS medical science",
  ["aims microbiol"] = "AIMS microbiology",
  ["aims mol sci"] = "AIMS molecular science",
  ["aims neurosci"] = "AIMS neuroscience",
  ["aims public health"] = "AIMS public health",
  ["aims rep"] = "AIMS report",
  ["aims ser appl math"] = "AIMS Series on Applied Mathematics",
  ["aims ser differ equ dyn syst"] = "AIMS Series on Differential Equations & Dynamical Systems",
  ["aims ser random comput dyn"] = "AIMS Series on Random and Computational Dynamics",
  ["ain shams eng j"] = "Ain Shams Engineering Journal",
  ["ain shams med j"] = "Ain Shams medical journal",
  ["aip adv"] = "AIP Advances",
  ["aip conf proc"] = "AIP Conference Proceedings",
  ["aip ser comput appl math phys"] = "AIP Series in Computational and Applied Mathematical Physics",
  ["aip ser modern acoust signal process"] = "AIP Series in Modern Acoustics and Signal Processing",
  ["aipla q j"] = "AIPLA quarterly journal",
  ["air force law rev"] = "The Air Force law review",
  ["air line pilot"] = "Air line pilot",
  ["air med j"] = "Air medical journal",
  ["air med news lett"] = "Air medical news-letter. Great Britain. Admiralty. Medical Department",
  ["air qual atmos health"] = "Air Quality, Atmosphere and Health",
  ["air sea rescue bull"] = "Air sea rescue bulletin",
  ["air soil water res"] = "Air, Soil and Water Research",
  ["air space"] = "Air & space Smithsonian",
  ["air surg bull"] = "The air surgeon's bulletin",
  ["air univ rev"] = "Air University quarterly review",
  ["air univ rev (us ed)"] = "Air University review (United States edition)",
  ["air waste"] = "Air & waste : journal of the Air & Waste Management Association",
  ["air water pollut"] = "Air and Water Pollution",
  ["air water soil sci t"] = "Traffic Related Air Pollution and Internal Combustion Engines",
  ["aircr des"] = "Aircraft Design",
  ["aircr eng"] = "Aircraft Engineering",
  ["aircr eng aerosp technol"] = "Aircraft Engineering and Aerospace Technology",
  ["airo (lisboa)"] = "Airo",
  ["airo springer ser"] = "AIRO Springer Series",
  ["aisa found news"] = "The Asia Foundation news",
  ["ajalooline ajak"] = "Ajalooline ajakiri (Tartu, Estonia : 1998)",
  ["ajnr am j neuroradiol"] = "AJNR. American journal of neuroradiology",
  ["ajob empir bioeth"] = "AJOB empirical bioethics",
  ["ajob neurosci"] = "AJOB neuroscience",
  ["ajob prim res"] = "AJOB primary research",
  ["ajp rep"] = "AJP reports",
  ["ajr am j roentgenol"] = "AJR. American journal of roentgenology",
  ["ajs rev"] = "AJS review. Association for Jewish Studies",
  ["ajsp rev rep"] = "AJSP: reviews & reports",
  ["ak peters/crc recreat math ser"] = "AK Peters/CRC Recreational Mathematics Series",
  ["akad dzive"] = "Akadēmiskā Dzīve",
  ["akad iatr"] = "Akademaike iatrike",
  ["akad nauk armenii dokl"] = "Akademiya Nauk Armenii",
  ["akad nauk respub kazakhstan trudy astrofiz inst"] = "Akademiya Nauk Respubliki Kazakhstan",
  ["akad nauka umjet bosne hercegov rad odjelj prirod mat nauka"] = "Akademija Nauka i Umjetnosti Bosne i Hercegovine",
  ["akad nauka umjet bosne hercegov rad odjelj tehn nauka"] = "Akademija Nauka i Umjetnosti Bosne i Hercegovine",
  ["akce int j graphs comb"] = "AKCE International Journal of Graphs and Combinatorics",
  ["akdeniz univ ziraat fak derg"] = "Akdeniz Üniversitesi Ziraat Fakültesi dergisi = Journal of Faculty of Agriculture Akdeniz University",
  ["akp class"] = "AKP Classics",
  ["akron bus econ rev"] = "Akron business and economic review",
  ["akron law rev"] = "Akron law review",
  ["aktual vopr vet biol"] = "Aktualʹnye voprosy veterinarnoj biologii",
  ["aktuel endokrinol stoffwechsel"] = "Aktuelle Endokrinologie und Stoffwechsel",
  ["aktuel ernahrungsmed"] = "Aktuelle Ernährungsmedizin",
  ["aktuel probl phoniatr logop"] = "Aktuelle Probleme der Phoniatrie und Logopädie. Current problems in phoniatrics and logopedics. Problèmes actuels de phoniatrie et de logopédie",
  ["aktuelle derm"] = "Aktuelle Dermatologie",
  ["aktuelle gerontol"] = "Aktuelle Gerontologie",
  ["aktuelle otorhinolaryngol"] = "Aktuelle Oto-Rhino-Laryngologie",
  ["aktuelle probl  dermatol"] = "Aktuelle Probleme der Dermatologie. Current problems in dermatology. Problèmes actuels de dermatologie",
  ["aktuelle probl chir"] = "Aktuelle Probleme in der Chirurgie",
  ["aktuelle probl chir orthop"] = "Aktuelle Probleme in Chirurgie und Orthopadie",
  ["aktuelle radiol"] = "Aktuelle Radiologie",
  ["aktuelle traumatol"] = "Aktuelle Traumatologie",
  ["aktuelle urol"] = "Aktuelle Urologie",
  ["akush ginekol (mosk)"] = "Akusherstvo i ginekologii︠a︡",
  ["akush ginekol (sofiia)"] = "Akusherstvo i Ginekologiia",
  ["akust zh [sov phys acoust]"] = "Akusticheskii Zhurnal [Soviet PhysicsAcoustics]",
  ["akwesasne notes"] = "Akwesasne notes",
  ["al azhar med j"] = "Al Azhar medical journal = Majallat al-Ṭibb al-Azhar",
  ["al darat"] = "Al Darat",
  ["al majallah al tibbiyah al iraqiyah"] = "al-Majallah al-tibbiyah al-'Iraqiyah",
  ["al mihan al tibbiyah"] = "al-Mihan al-tibbiyah",
  ["al qantara"] = "al-Qanṭara",
  ["al raida"] = "al-Raida / Institute for Women's Studies in the Arab World, Beirut University College",
  ["al serv bosco paesaggio"] = "Al servizio del bosco e del paesaggio",
  ["al usrah wa al umran al bashari"] = "al-Usrah wa-al-ʻumrān al-basharī",
  ["ala dent rev"] = "Alabama dental review",
  ["ala herit"] = "Alabama heritage",
  ["ala hist q"] = "The Alabama historical quarterly",
  ["ala j math"] = "Alabama Journal of Mathematics",
  ["ala j med sci"] = "The Alabama journal of medical sciences",
  ["ala law rev"] = "Alabama law review",
  ["ala med"] = "Alabama medicine : journal of the Medical Association of the State of Alabama",
  ["ala nurse"] = "The Alabama nurse",
  ["ala rev"] = "The Alabama review",
  ["ala soc welf"] = "Alabama social welfare",
  ["alam al firkr"] = "ʻĀlam al-fikr",
  ["alan guttmacher inst annu rep"] = "The Alan Guttmacher Institute annual report. Alan Guttmacher Institute",
  ["alaska j"] = "Alaska journal (Juneau, Alaska : 1971)",
  ["alaska med"] = "Alaska medicine",
  ["alaska nurse"] = "The Alaska nurse",
  ["alba regia"] = "Alba Regia: annales Musei Stephani Regis",
  ["albanian j math"] = "Albanian Journal of Mathematics",
  ["albany law j sci technol"] = "Albany law journal of science & technology",
  ["albany law rev"] = "Albany law review",
  ["alber-reihe philos"] = "Alber-Reihe: Philosophie",
  ["alberta hist"] = "Alberta history",
  ["alberta med bull"] = "Alberta medical bulletin",
  ["albertus magnus blaett"] = "Albertus Magnus Blätter",
  ["albion math appl ser"] = "Albion Mathematics & Applications Series",
  ["albrecht von graefes arch klin exp ophthalmol"] = "Albrecht von Graefes Archiv für klinische und experimentelle Ophthalmologie. Albrecht von Graefe's archive for clinical and experimental ophthalmology",
  ["albrecht von graefes arch ophthalmol"] = "Albrecht von Graefe's Archiv für Ophthalmologie",
  ["albuq tribune"] = "The Albuquerque tribune",
  ["alche j"] = "AIChE Journal",
  ["alche symp ser"] = "AIChE Symposium Series",
  ["alcheringa: australas j palaeontol"] = "Alcheringa: An Australasian Journal of Palaeontology",
  ["alcohol (ny)"] = "Alcohol (New York)",
  ["alcohol alcohol"] = "Alcohol and alcoholism (Oxford, Oxfordshire)",
  ["alcohol alcohol suppl"] = "Alcohol and alcoholism (Oxford, Oxfordshire). Supplement",
  ["alcohol clin exp res"] = "Alcoholism, clinical and experimental research",
  ["alcohol drug addict"] = "Alcoholism and Drug Addiction",
  ["alcohol drug res"] = "Alcohol and drug research",
  ["alcohol drugs driving"] = "Alcohol, drugs, and driving : abstracts and reviews",
  ["alcohol health res world"] = "Alcohol health and research world",
  ["alcohol hyg"] = "Alcohol hygiene",
  ["alcohol res"] = "Alcohol research : current reviews",
  ["alcohol res health"] = "Alcohol research & health : the journal of the National Institute on Alcohol Abuse and Alcoholism",
  ["alcohol treat q"] = "Alcoholism treatment quarterly",
  ["alcohol: clin exp res"] = "Alcoholism: Clinical and Experimental Research",
  ["aldrichimica acta"] = "Aldrichimica acta",
  ["alea lat am j probab math stat"] = "ALEA. Latin American Journal of Probability and Mathematical Statistics",
  ["alembic club repr"] = "Alembic Club reprints",
  ["aleph (jerus)"] = "Aleph (Jerusalem)",
  ["alex dent j"] = "Alexandria dental journal : ADJ",
  ["alex j pharm sci"] = "Alexandria journal of pharmaceutical sciences",
  ["alex j vet sci"] = "Alexandria journal of veterinary sciences",
  ["alexander blain hosp bull"] = "Alexander Blain Hospital bulletin",
  ["alexandria eng j"] = "Alexandria Engineering Journal",
  ["alexandria j agric res"] = "Alexandria Journal of Agricultural Research",
  ["alexandria j med"] = "Alexandria Journal of Medicine",
  ["alexandria j vet sci"] = "Alexandria Journal of Veterinary Sciences",
  ["alexandria med j"] = "The Alexandria medical journal",
  ["algal res"] = "Algal research",
  ["algebr appl"] = "Algebra and Applications",
  ["algebr comb"] = "Algebraic Combinatorics",
  ["algebr geom"] = "Algebraic Geometry",
  ["algebr geom topol"] = "Algebraic & Geometric Topology",
  ["algebr represent theory"] = "Algebras and Representation Theory",
  ["algebr stat"] = "Algebraic Statistics",
  ["algebra and logic"] = "Algebra and Logic",
  ["algebra appl class mod"] = "L’Algebra e le sue Applicazioni tra Classico e Moderno",
  ["algebra ber"] = "Algebra Berichte",
  ["algebra colloq"] = "Algebra Colloquium",
  ["algebra discrete math"] = "Algebra and Discrete Mathematics",
  ["algebra i analiz"] = "Rossiiskaya Akademiya Nauk. Algebra i Analiz",
  ["algebra log"] = "Algebra and Logic",
  ["algebra logic"] = "Algebra and Logic",
  ["algebra logic appl"] = "Algebra, Logic and Applications",
  ["algebra logika"] = "Algebra i Logika. Institut Diskretnoi Matematiki i Informatiki",
  ["algebra number theory"] = "Algebra & Number Theory",
  ["algebra universalis"] = "Algebra universalis",
  ["algebraic geom"] = "Algebraic Geometry",
  ["algebraic geom phys"] = "Algebraic Geometry and Physics",
  ["algebraic geom topol"] = "Algebraic & Geometric Topology",
  ["algebras groups geom"] = "Algebras, Groups and Geometries",
  ["algebras represent theory"] = "Algebras and representation theory",
  ["algebras representation theory"] = "Algebras and Representation Theory",
  ["alger medicale"] = "Algérie médicale",
  ["algol stud (2007)"] = "Algological studies (Stuttgart, Germany : 2007)",
  ["algologiia (kiev)"] = "Algologiia (Kiev, Ukraine)",
  ["algorithmic finance"] = "Algorithmic Finance",
  ["algorithms combin"] = "Algorithms and Combinatorics",
  ["algorithms comput math"] = "Algorithms and Computation in Mathematics",
  ["algorithms intell syst"] = "Algorithms for Intelligent Systems",
  ["algorithms mol biol"] = "Algorithms for molecular biology : AMB",
  ["aligarh bull math"] = "The Aligarh Bulletin of Mathematics",
  ["aligarh j statist"] = "The Aligarh Journal of Statistics",
  ["aliment pharmacol ther"] = "Alimentary pharmacology & therapeutics",
  ["aliment vie"] = "L' Alimentation et la vie",
  ["alkalmaz mat lapok"] = "Alkalmazott Matematikai Lapok",
  ["alkaloids chem biol"] = "The Alkaloids. Chemistry and biology",
  ["alkohol narkom"] = "Alkoholizm i narkomania",
  ["all earth"] = "All Earth",
  ["all engl law rep"] = "The all England law reports",
  ["all life"] = "All Life",
  ["all life methods"] = "All Life Methods",
  ["all results j biol"] = "The all results journals. Biol",
  ["all stat arch"] = "Allgemeines statistisches Archiv",
  ["allelopathy j"] = "Allelopathy journal : official publication of International Allelopathy Foundation",
  ["allem aujourdhui"] = "Allemagne d'aujourd'hui : revue française d'information sur les deux l'Allemagnes",
  ["allerg asthma (leipz)"] = "Allergie und Asthma",
  ["allerg asthmaforsch"] = "Allergie- und Asthmaforschung",
  ["allerg immunol (leipz)"] = "Allergie und Immunologie",
  ["allerg immunol (paris)"] = "Allergie et immunologie",
  ["allerg pointe claire"] = "Allergie (Pointe-Claire, Québec)",
  ["allergo j"] = "Allergo Journal : interdisziplinäre Zeitschrift für Allergologie und Umweltmedizin : Organ der Deutschen Gesellschaft für Allergie- und Immunitätsforschung",
  ["allergo j int"] = "Allergo journal international",
  ["allergol immunopathol (madr)"] = "Allergologia et immunopathologia",
  ["allergol int"] = "Allergology international : official journal of the Japanese Society of Allergology",
  ["allergol select"] = "Allergologie select",
  ["allergy asthma clin immunol"] = "Allergy, asthma, and clinical immunology : official journal of the Canadian Society of Allergy and Clinical Immunology",
  ["allergy asthma immunol res"] = "Allergy, asthma & immunology research",
  ["allergy asthma proc"] = "Allergy and asthma proceedings",
  ["allergy clin immunol int"] = "Allergy & clinical immunology international : official organ of the International Association of Allergology and Clinical Immunology",
  ["allergy proc"] = "Allergy proceedings : the official journal of regional and state allergy societies",
  ["allergy rhinol (providence)"] = "Allergy & rhinology (Providence, R.I.)",
  ["allg forst jagdztg"] = "Allgemeine Forst und Jagdzeitung",
  ["allg forst z waldwirtsch umweltvorsorge"] = "Allgemeine Forst Zeitschrift für Waldwirtschaft und Umweltvorsorge [AFZ-Der Wald]",
  ["allg forst- jagdztg"] = "Allgemeine Forst- und Jagdzeitung",
  ["allg homoopath ztg"] = "Allgemeine homöopathische Zeitung für wissenschaftliche und praktische Homöopathie",
  ["allgemeines statistisches arch"] = "Allgemeines Statistisches Archiv",
  ["allied health behav sci"] = "Allied health and behavioral sciences",
  ["alma mater (baltimore)"] = "The Alma mater",
  ["alma mater philipp"] = "Alma Mater Philippina",
  ["almagallat alarabiyyat lilbiat algaffat"] = "al-Majallah al-ʻArabīyah lil-bīʼāt al-jāffah",
  ["almagallat aliraqiyyat libuhut alsuq wahimayat almustahlik"] = "Al-Mağallaẗ al-'irāqiyyaẗ li-buḥūt̲ al-sūq wa-ḥimāyaẗ al-mustahlik",
  ["almagallat altibbiyyat albaytariyyat aliraqiyyat"] = "Al-maǧallaẗ al-ṭibbiyyaẗ al-bayṭariyyaẗ al-’irāqiyyaẗ = The Iraqi journal of veterinary medicine",
  ["almustaqbal alarabi"] = "al-Mustaqbal al-ʻArabī",
  ["alp bot"] = "Alpine botany",
  ["alp entomol"] = "Alpine Entomology",
  ["alp mediterr quat"] = "Alpine And Mediterranean Quaternary",
  ["alpha omega fr"] = "Alpha Omega France",
  ["alpha omegan"] = "The Alpha omegan",
  ["alsk health"] = "Alaska's health",
  ["alsk health profile"] = "Alaska health profile",
  ["alsk hist"] = "Alaska history (Anchorage, Alaska)",
  ["alsk statut 1962 alsk"] = "Alaska statutes, 1962 : containing the general and permanent laws of the State of Alaska. Alaska",
  ["alt thuring"] = "Alt-Thüringen",
  ["alta law rev"] = "Alberta law review",
  ["alta rn"] = "Alberta RN",
  ["alterites (univ montr dep anthropol)"] = "Altérités (Université de Montréal. Département d'anthropologie)",
  ["altern complement ther"] = "Alternative & complementary therapies : a new bimonthly publication for health care practitioners",
  ["altern entwickl aging dev"] = "Altern und Entwicklung. Aging and development",
  ["altern integr med"] = "Alternative & integrative medicine",
  ["altern lab anim"] = "Alternatives to laboratory animals : ATLA",
  ["altern med rev"] = "Alternative medicine review : a journal of clinical therapeutic",
  ["altern ther health med"] = "Alternative therapies in health and medicine",
  ["alternation (durb)"] = "Alternation",
  ["alternative (nga pae maramatanga (organ))"] = "AlterNative (Auckland, N.Z. : 2005)",
  ["alternative lifestyles"] = "Alternative lifestyles",
  ["alternatives (boulder)"] = "Alternatives (Boulder, Colo.)",
  ["altnurnb landsch"] = "Schriftenreihe. Altnürnberger Landschaft e.V",
  ["altsprachi unterr"] = "Der Altsprachliche Unterricht",
  ["alum trans"] = "Aluminum Transactions",
  ["aluminium (duesseldorf)"] = "Aluminium (Duesseldorf)",
  ["alumnae mag"] = "The Alumnae magazine",
  ["alumnae mag columbia univ presbyt hosp sch nurs alumnae assoc"] = "Alumnae magazine (Columbia University-Presbyterian Hospital School of Nursing Alumnae Association)",
  ["alumnae mag johns hopkins hosp sch nurs alumnae assoc"] = "Alumnae Magazine, Johns Hopkins Hospital School of Nursing Alumnae Association",
  ["alumni bull long island coll med brooklyn"] = "Alumni bulletin. Long Island College of Medicine. Alumni Association",
  ["alumni bull sch dent indiana univ"] = "Alumni bulletin - School of Dentistry, Indiana University",
  ["alumni bull univ mich sch dent"] = "Alumni bulletin. University of Michigan. School of Dentistry",
  ["alumni bull west reserv univ sch med"] = "Alumni bulletin. Western Reserve University. School of Medicine",
  ["alumni mag"] = "The Alumni magazine",
  ["alumni mag columbia univ presbyt hosp sch nurs alumni assoc"] = "Alumni magazine (Columbia University-Presbyterian Hospital School of Nursing Alumni Association)",
  ["alzheimer dis assoc disord"] = "Alzheimer disease and associated disorders",
  ["alzheimers care q"] = "Alzheimer's care quarterly",
  ["alzheimers care today"] = "Alzheimer's care today",
  ["alzheimers dement"] = "Alzheimer's & dementia : the journal of the Alzheimer's Association",
  ["alzheimers dement (amst)"] = "Alzheimer's & dementia (Amsterdam, Netherlands)",
  ["alzheimers dement (n y)"] = "Alzheimer's & dementia (New York, N. Y.)",
  ["alzheimers dement cogn neurol"] = "Alzheimer's, dementia & cognitive neurology",
  ["alzheimers dis dement"] = "Alzheimers disease & dementia",
  ["alzheimers dis res j"] = "Alzheimer's disease research journal",
  ["alzheimers rep"] = "Alzheimer's reports",
  ["alzheimers res ther"] = "Alzheimer's research & therapy",
  ["alzheimers res ther open access"] = "Alzheimer's research & therapy open access",
  ["am ann deaf"] = "American annals of the deaf",
  ["am anthropol"] = "American anthropologist",
  ["am antiq"] = "American antiquity",
  ["am arch"] = "The American archivist",
  ["am arch rehabil ther"] = "American archives of rehabilitation therapy",
  ["am art j"] = "The American art journal",
  ["am asian rev"] = "The American Asian review",
  ["am assoc ind nurses j"] = "American Association of Industrial Nurses journal",
  ["am assoc pet geol bull"] = "American Association of Petroleum Geologists Bulletin",
  ["am baby"] = "The American baby",
  ["am baptist q"] = "American Baptist quarterly",
  ["am bar assoc j"] = "American Bar Association journal. American Bar Association",
  ["am bar found res j"] = "American Bar Foundation research journal. American Bar Foundation",
  ["am bee j"] = "American Bee Journal",
  ["am behav sci"] = "The American behavioral scientist",
  ["am benedictine rev"] = "The American Benedictine review",
  ["am biol teach"] = "The American biology teacher",
  ["am biotechnol lab"] = "American biotechnology laboratory",
  ["am book prices"] = "American book prices current",
  ["am bookm"] = "The American bookman",
  ["am ceram soc bull"] = "American Ceramic Society Bulletin",
  ["am chem j"] = "American Chemical Journal",
  ["am chem sci j"] = "American Chemical Science Journal",
  ["am christmas tree j"] = "American Christmas Tree Journal",
  ["am clin"] = "America clínica",
  ["am clin lab"] = "American clinical laboratory",
  ["am coll cardiol ext learn"] = "The American College of Cardiology Extended Learning : ACCEL",
  ["am coll cardiol extended learn suppl tape"] = "The American College of Cardiology Extended Learning. Supplement tape",
  ["am coll physicians obs"] = "American College of Physicians observer",
  ["am concr inst, sp"] = "American Concrete Institute, SP",
  ["am correct ther j"] = "American corrective therapy journal",
  ["am crim law rev"] = "The American criminal law review",
  ["am demogr"] = "American demographics",
  ["am drug"] = "American druggist",
  ["am eccles rev"] = "The American ecclesiastical review",
  ["am econ"] = "The American economist",
  ["am econ j appl econ"] = "American economic journal. Applied economics",
  ["am econ j econ policy"] = "American economic journal. Economic policy",
  ["am econ j macroecon"] = "The American economic journal. Macroeconomics",
  ["am econ j microecon"] = "American economic journal. Microeconomics",
  ["am econ rev"] = "The American economic review",
  ["am econ rev insights"] = "American economic review. Insights",
  ["am educ"] = "American education",
  ["am educ res j"] = "American educational research journal",
  ["am enterp"] = "American enterprise (Washington, D.C.)",
  ["am entomol"] = "American Entomologist",
  ["am ethnol"] = "American ethnologist",
  ["am eurasian j agric environ sci"] = "American-Eurasian journal of agricultural & environmental sciences",
  ["am fam physician"] = "American family physician",
  ["am fam physician gp"] = "American family physician/GP",
  ["am fern j"] = "American fern journal",
  ["am health"] = "American health",
  ["am health drug benefits"] = "American health & drug benefits",
  ["am heart hosp j"] = "The American heart hospital journal",
  ["am heart j"] = "American heart journal",
  ["am herit"] = "American heritage",
  ["am herit invent technol"] = "American heritage of invention & technology",
  ["am hist assoc inst serv program"] = "American Historical Association, Institutional Services Program",
  ["am hist illus"] = "American history illustrated",
  ["am hist life"] = "America, history and life",
  ["am hist rev"] = "The American historical review",
  ["am homeopath"] = "The American homeopath",
  ["am horseman"] = "American horseman",
  ["am imago"] = "The American imago; a psychoanalytic journal for the arts and sciences",
  ["am ind hyg assoc j"] = "American Industrial Hygiene Association Journal",
  ["am ind hyg assoc q"] = "American Industrial Hygiene Association quarterly",
  ["am indian alsk native ment health res"] = "American Indian and Alaska native mental health research (Online)",
  ["am indian alsk native ment health res monogr ser"] = "American Indian and Alaska native mental health research (Monographic series)",
  ["am indian art mag"] = "American Indian art magazine",
  ["am indian cult res j"] = "American Indian culture and research journal",
  ["am indian q"] = "American Indian quarterly",
  ["am indig"] = "America indígena",
  ["am int j contemp res"] = "American international journal of contemporary research",
  ["am int j res form appl nat sci"] = "American international journal of research in formal, applied and natural sciences",
  ["am j  bioinform comput biol"] = "American journal of bioinformatics and computational biology",
  ["am j accountable care"] = "American journal of accountable care",
  ["am j acupunct"] = "American journal of acupuncture",
  ["am j addict"] = "The American journal on addictions",
  ["am j adv drug deliv"] = "American journal of advanced drug delivery",
  ["am j aging sci res"] = "American journal of aging science and research",
  ["am j agric biol sci"] = "American journal of agricultural and biological sciences",
  ["am j agric econ"] = "American journal of agricultural economics",
  ["am j altern agric"] = "American Journal of Alternative Agriculture",
  ["am j alzheimers dis (columbia)"] = "American journal of Alzheimer's disease (Columbia, Mo.)",
  ["am j alzheimers dis other demen"] = "American journal of Alzheimer's disease and other dementias",
  ["am j analyt chem"] = "American journal of analytical chemistry",
  ["am j anat"] = "The American journal of anatomy",
  ["am j anesthesiol"] = "The American journal of anesthesiology",
  ["am j anim vet sci"] = "American journal of animal and veterinary sciences",
  ["am j appl math stat"] = "American journal of applied mathematics and statistics",
  ["am j appl sci"] = "American journal of applied sciences",
  ["am j archaeol"] = "American Journal Of Archaeology",
  ["am j art ther"] = "American journal of art therapy",
  ["am j audiol"] = "American journal of audiology",
  ["am j biochem biotechnol"] = "American Journal of Biochemistry and Biotechnology",
  ["am j biochem mol biol"] = "American journal of biochemistry and molecular biology",
  ["am j bioeth"] = "The American journal of bioethics : AJOB",
  ["am j bioethics"] = "American Journal of Bioethics",
  ["am j biol anthropol"] = "American Journal of Biological Anthropology",
  ["am j biomed eng"] = "American journal of biomedical engineering",
  ["am j biomed sci"] = "American journal of biomedical sciences",
  ["am j biomed sci res"] = "American journal of biomedical science & research",
  ["am j biostat"] = "American journal of biostatistics",
  ["am j blood res"] = "American journal of blood research",
  ["am j bot"] = "American journal of botany",
  ["am j cancer"] = "The American journal of cancer",
  ["am j cancer prev"] = "American journal of cancer prevention",
  ["am j cancer res"] = "American journal of cancer research",
  ["am j cancer sci"] = "American journal of cancer science",
  ["am j cancer ther pharmacol"] = "American journal of cancer therapy and pharmacology",
  ["am j card imaging"] = "American journal of cardiac imaging",
  ["am j cardiol"] = "The American journal of cardiology",
  ["am j cardiovasc dis"] = "American journal of cardiovascular disease",
  ["am j cardiovasc drugs"] = "American journal of cardiovascular drugs : drugs, devices, and other interventions",
  ["am j cardiovasc pathol"] = "The American journal of cardiovascular pathology",
  ["am j case rep"] = "The American journal of case reports",
  ["am j chin med"] = "The American journal of Chinese medicine",
  ["am j chin med (gard city n y)"] = "The American journal of Chinese medicine",
  ["am j chin stud"] = "American journal of Chinese studies",
  ["am j clim change"] = "American journal of climate change",
  ["am j clin cancer res"] = "American journal of clinical cancer research",
  ["am j clin dermatol"] = "American journal of clinical dermatology",
  ["am j clin exp immunol"] = "American journal of clinical and experimental immunology",
  ["am j clin exp obstet gynecol"] = "American journal of clinical and experimental obstetrics and gynecology",
  ["am j clin exp urol"] = "American journal of clinical and experimental urology",
  ["am j clin hypn"] = "The American journal of clinical hypnosis",
  ["am j clin nutr"] = "American Journal of Clinical Nutrition",
  ["am j clin oncol"] = "American journal of clinical oncology",
  ["am j clin pathol"] = "American journal of clinical pathology",
  ["am j community psychol"] = "American journal of community psychology",
  ["am j comp law"] = "The American journal of comparative law",
  ["am j contact dermat"] = "American journal of contact dermatitis : official journal of the American Contact Dermatitis Society",
  ["am j correct"] = "American journal of correction",
  ["am j crim justice"] = "American journal of criminal justice : AJCJ",
  ["am j crim law"] = "American journal of criminal law",
  ["am j crit care"] = "American journal of critical care : an official publication, American Association of Critical-Care Nurses",
  ["am j cult sociol"] = "American journal of cultural sociology",
  ["am j dance ther"] = "American journal of dance therapy",
  ["am j dent"] = "American journal of dentistry",
  ["am j dent sci"] = "The American journal of dental science",
  ["am j dermatopathol"] = "The American Journal of dermatopathology",
  ["am j dig dis"] = "The American journal of digestive diseases",
  ["am j dig dis (madison)"] = "American journal of digestive disease",
  ["am j dis child"] = "American journal of diseases of children (1960)",
  ["am j disaster med"] = "American journal of disaster medicine",
  ["am j distance educ"] = "The American journal of distance education",
  ["am j drug alcohol abuse"] = "The American journal of drug and alcohol abuse",
  ["am j drug delivery"] = "American Journal of Drug Delivery",
  ["am j econ sociol"] = "American journal of economics and sociology",
  ["am j educ (chic ill)"] = "American journal of education (Chicago, Ill.)",
  ["am j educ res"] = "American journal of educational research",
  ["am j eeg technol"] = "The American journal of EEG technology",
  ["am j electroneurodiagnostic technol"] = "American journal of electroneurodiagnostic technology",
  ["am j emerg med"] = "The American journal of emergency medicine",
  ["am j eng educ"] = "American journal of engineering education",
  ["am j eng res"] = "American Journal of Engineering Research",
  ["am j enol vitic"] = "American journal of enology and viticulture",
  ["am j entrep"] = "American journal of entrepreneurship",
  ["am j environ engineer"] = "American journal of environmental engineering",
  ["am j environ sci"] = "American journal of environmental sciences",
  ["am j epidemiol"] = "American journal of epidemiology",
  ["am j ethics med"] = "American journal of ethics & medicine : a national medical student journal published by the University of Pennsylvania School of Medicine",
  ["am j eval"] = "The American journal of evaluation",
  ["am j fam law"] = "American journal of family law",
  ["am j fam ther"] = "The American journal of family therapy",
  ["am j food technol"] = "American Journal of Food Technology",
  ["am j forensic med pathol"] = "The American journal of forensic medicine and pathology",
  ["am j forensic psychiatry"] = "The American journal of forensic psychiatry",
  ["am j forensic psychol"] = "The American journal of forensic psychology",
  ["am j gastroenterol"] = "The American journal of gastroenterology",
  ["am j gastroenterol suppl"] = "American journal of gastroenterology supplements (Print)",
  ["am j geriatr cardiol"] = "The American journal of geriatric cardiology",
  ["am j geriatr pharmacother"] = "The American journal of geriatric pharmacotherapy",
  ["am j geriatr psychiatry"] = "The American journal of geriatric psychiatry : official journal of the American Association for Geriatric Psychiatry",
  ["am j gynecol health"] = "The American journal of gynecologic health",
  ["am j health behav"] = "American journal of health behavior",
  ["am j health econ"] = "American journal of health economics",
  ["am j health educ"] = "American journal of health education",
  ["am j health plann"] = "American journal of health planning",
  ["am j health promot"] = "American journal of health promotion : AJHP",
  ["am j health sci"] = "American journal of health sciences",
  ["am j health stud"] = "American journal of health studies",
  ["am j health syst pharm"] = "American journal of health-system pharmacy : AJHP : official journal of the American Society of Health-System Pharmacists",
  ["am j health-syst pharm"] = "American Journal of Health System Pharmacy",
  ["am j hematol"] = "American journal of hematology",
  ["am j hematol oncol"] = "American journal of hematology/oncology",
  ["am j hosp care"] = "The American journal of hospice care",
  ["am j hosp med"] = "American journal of hospital medicine",
  ["am j hosp palliat care"] = "The American journal of hospice & palliative care",
  ["am j hosp pharm"] = "American journal of hospital pharmacy",
  ["am j hum biol"] = "American journal of human biology : the official journal of the Human Biology Council",
  ["am j hum genet"] = "American journal of human genetics",
  ["am j hyg"] = "American journal of hygiene",
  ["am j hypertens"] = "American journal of hypertension",
  ["am j immunol"] = "American journal of immunology",
  ["am j ind med"] = "American journal of industrial medicine",
  ["am j inf manag"] = "American journal of information management",
  ["am j infect control"] = "American journal of infection control",
  ["am j infect dis"] = "American journal of infectious diseases",
  ["am j int law"] = "The American journal of international law",
  ["am j intellect dev disabil"] = "American journal on intellectual and developmental disabilities",
  ["am j islam soc sci"] = "The American journal of Islamic social sciences",
  ["am j jurisprud"] = "The American journal of jurisprudence",
  ["am j kidney dis"] = "American journal of kidney diseases : the official journal of the National Kidney Foundation",
  ["am j knee surg"] = "The American journal of knee surgery",
  ["am j lab med"] = "American journal of laboratory medicine",
  ["am j law med"] = "American journal of law & medicine",
  ["am j leg hist"] = "The American journal of legal history",
  ["am j leuk res"] = "American journal of leukemia research",
  ["am j lifestyle med"] = "American journal of lifestyle medicine",
  ["am j manag care"] = "The American journal of managed care",
  ["am j math"] = "American Journal of Mathematics",
  ["am j math manage sci"] = "American Journal of Mathematical and Management Sciences",
  ["am j med"] = "The American journal of medicine",
  ["am j med case rep"] = "American journal of medical case reports",
  ["am j med electron"] = "The American journal of medical electronics",
  ["am j med genet"] = "American Journal of Medical Genetics",
  ["am j med genet a"] = "American journal of medical genetics. Part A",
  ["am j med genet b neuropsychiatr genet"] = "American journal of medical genetics. Part B, Neuropsychiatric genetics : the official publication of the International Society of Psychiatric Genetics",
  ["am j med genet c semin med genet"] = "American journal of medical genetics. Part C, Seminars in medical genetics",
  ["am j med genet part a"] = "American Journal of Medical Genetics Part A",
  ["am j med genet part b: neuropsychiatr genet"] = "American Journal of Medical Genetics Part B: Neuropsychiatric Genetics",
  ["am j med genet part c: semin med genet"] = "American Journal of Medical Genetics Part C: Seminars in Medical Genetics",
  ["am j med genet suppl"] = "American journal of medical genetics. Supplement",
  ["am j med qual"] = "American journal of medical quality : the official journal of the American College of Medical Quality",
  ["am j med res (n y)"] = "American journal of medical research (New York, N.Y.)",
  ["am j med sci"] = "The American journal of the medical sciences",
  ["am j med technol"] = "The American journal of medical technology",
  ["am j media psychol"] = "American journal of media psychology",
  ["am j mens health"] = "American journal of men's health",
  ["am j ment defic"] = "American journal of mental deficiency",
  ["am j ment retard"] = "American journal of mental retardation : AJMR",
  ["am j microbiol biotechnol"] = "American journal of microbiology and biotechnology",
  ["am j microbiol res"] = "American journal of microbiological research",
  ["am j min metall"] = "American Journal of Mining and Metallurgy",
  ["am j mod chromatogr"] = "American journal of modern chromatography",
  ["am j mol biol"] = "American journal of molecular biology",
  ["am j nephrol"] = "American journal of nephrology",
  ["am j neurodegener dis"] = "American journal of neurodegenerative disease",
  ["am j neuroprot neuroregen"] = "American journal of neuroprotection and neuroregeneration",
  ["am j neuroprot neuroregener"] = "American Journal of Neuroprotection and Neuroregeneration",
  ["am j neurosci"] = "American journal of neuroscience",
  ["am j nucl med mol imaging"] = "American journal of nuclear medicine and molecular imaging",
  ["am j nurs"] = "The American journal of nursing",
  ["am j nurse pract"] = "The American journal for nurse practitioners",
  ["am j obstet dis women child"] = "The American journal of obstetrics and diseases of women and children",
  ["am j obstet gynecol"] = "American journal of obstetrics and gynecology",
  ["am j obstet gynecol mfm"] = "American journal of obstetrics & gynecology MFM",
  ["am j occup ther"] = "The American journal of occupational therapy : official publication of the American Occupational Therapy Association",
  ["am j ophthalmol"] = "American journal of ophthalmology",
  ["am j ophthalmol case rep"] = "American journal of ophthalmology case reports",
  ["am j optom arch am acad optom"] = "American journal of optometry and archives of American Academy of Optometry",
  ["am j optom physiol opt"] = "American journal of optometry and physiological optics",
  ["am j orthod"] = "American journal of orthodontics",
  ["am j orthod dentofacial orthop"] = "American journal of orthodontics and dentofacial orthopedics : official publication of the American Association of Orthodontists, its constituent societies, and the American Board of Orthodontics",
  ["am j orthod oral surg"] = "American journal of orthodontics and oral surgery",
  ["am j orthop"] = "The American journal of orthopedics",
  ["am j orthop (belle mead nj)"] = "American journal of orthopedics (Belle Mead, N.J.)",
  ["am j orthop surg"] = "The American journal of orthopedic surgery",
  ["am j orthopsychiatry"] = "The American journal of orthopsychiatry",
  ["am j otol"] = "The American journal of otology",
  ["am j otolaryngol"] = "American journal of otolaryngology",
  ["am j pathol"] = "The American journal of pathology",
  ["am j pediatr hematol oncol"] = "The American journal of pediatric hematology/oncology",
  ["am j perinatol"] = "American journal of perinatology",
  ["am j pharm"] = "The American journal of pharmacy",
  ["am j pharm benefits"] = "The American journal of pharmacy benefits",
  ["am j pharm educ"] = "American journal of pharmaceutical education",
  ["am j pharm sci support public health"] = "American journal of pharmacy and the sciences supporting public health",
  ["am j pharmacogenomics"] = "American journal of pharmacogenomics : genomics-related research in drug development and clinical practice",
  ["am j pharmacol sci"] = "American journal of pharmacological sciences",
  ["am j pharmacol toxicol"] = "American Journal of Pharmacology and Toxicology",
  ["am j philol"] = "American journal of philology",
  ["am j phys"] = "American journal of physics",
  ["am j phys anthropol"] = "American journal of physical anthropology",
  ["am j phys med"] = "American journal of physical medicine",
  ["am j phys med rehabil"] = "American journal of physical medicine & rehabilitation",
  ["am j physiol"] = "American Journal of Physiology",
  ["am j physiol cell physiol"] = "American journal of physiology. Cell physiology",
  ["am j physiol endocrinol metab"] = "American journal of physiology. Endocrinology and metabolism",
  ["am j physiol gastrointest liver physiol"] = "American journal of physiology. Gastrointestinal and liver physiology",
  ["am j physiol heart  circ physiol"] = "American Journal of Physiology Heart and Circulatory Physiology",
  ["am j physiol heart circ physiol"] = "American journal of physiology. Heart and circulatory physiology",
  ["am j physiol imaging"] = "American journal of physiologic imaging",
  ["am j physiol lung cell mol physiol"] = "American journal of physiology. Lung cellular and molecular physiology",
  ["am j physiol regul integr comp physiol"] = "American journal of physiology. Regulatory, integrative and comparative physiology",
  ["am j physiol regul intgr comp physiol"] = "American Journal of Physiology Regulatory, Integrative and Comparative Physiology",
  ["am j physiol renal physiol"] = "American journal of physiology. Renal physiology",
  ["am j phytomed clin ther"] = "American Journal of Phytomedicine and Clinical Therapeutics",
  ["am j plant sci"] = "American Journal of Plant Science",
  ["am j play"] = "American journal of play",
  ["am j pol sci"] = "American journal of political science",
  ["am j potato res"] = "American Journal of Potato Research",
  ["am j pract nurs"] = "American journal of practical nursing",
  ["am j prev cardiol"] = "American journal of preventive cardiology",
  ["am j prev med"] = "American journal of preventive medicine",
  ["am j primatol"] = "American journal of primatology",
  ["am j proctol"] = "American journal of proctology",
  ["am j proctol gastroenterol colon rectal surg"] = "American journal of proctology, gastroenterology & colon & rectal surgery",
  ["am j psychiatr rehabil"] = "American journal of psychiatric rehabilitation",
  ["am j psychiatry"] = "The American journal of psychiatry",
  ["am j psychiatry resid j"] = "The American journal of psychiatry residents' journal",
  ["am j psychoanal"] = "American journal of psychoanalysis",
  ["am j psychol"] = "The American journal of psychology",
  ["am j psychother"] = "American journal of psychotherapy",
  ["am j public health"] = "American journal of public health",
  ["am j public health (n y)"] = "American journal of public health (New York, N.Y. : 1912)",
  ["am j public health nations health"] = "American journal of public health and the nation's health",
  ["am j public health res"] = "American journal of public health research",
  ["am j public hygiene"] = "American journal of public hygiene",
  ["am j recreat ther"] = "American journal of recreation therapy",
  ["am j reprod immunol"] = "American journal of reproductive immunology (New York, N.Y. : 1989)",
  ["am j reprod immunol microbiol"] = "American journal of reproductive immunology and microbiology : AJRIM",
  ["am j res commun"] = "American journal of research communication",
  ["am j respir cell mol biol"] = "American Journal of Respiratory Cell and Molecular Biology",
  ["am j respir crit care med"] = "American journal of respiratory and critical care medicine",
  ["am j respir med"] = "American journal of respiratory medicine : drugs, devices, and other interventions",
  ["am j rhinol"] = "American journal of rhinology",
  ["am j rhinol allergy"] = "American journal of rhinology & allergy",
  ["am j robot surg"] = "American journal of robotic surgery",
  ["am j roentgenol"] = "American Journal of Roentgenology",
  ["am j roentgenol radium ther"] = "The American journal of roentgenology and radium therapy",
  ["am j roentgenol radium ther nucl med"] = "The American journal of roentgenology, radium therapy, and nuclear medicine",
  ["am j safe patient handl mov"] = "American journal of safe patient handling & movement",
  ["am j sci"] = "American Journal of Science",
  ["am j sex educ"] = "American journal of sexuality education",
  ["am j soc psychiatry"] = "The American journal of social psychiatry",
  ["am j sociol"] = "The American journal of sociology",
  ["am j speech lang pathol"] = "American journal of speech-language pathology",
  ["am j sports med"] = "The American journal of sports medicine",
  ["am j stem cells"] = "American journal of stem cells",
  ["am j surg"] = "American journal of surgery",
  ["am j surg pathol"] = "The American journal of surgical pathology",
  ["am j syph gonorrhea vener dis"] = "American journal of syphilis, gonorrhea, and venereal diseases",
  ["am j ther"] = "American journal of therapeutics",
  ["am j transl res"] = "American journal of translational research",
  ["am j transplant"] = "American journal of transplantation : official journal of the American Society of Transplantation and the American Society of Transplant Surgeons",
  ["am j trial advocacy"] = "The American journal of trial advocacy",
  ["am j trop med hyg"] = "The American journal of tropical medicine and hygiene",
  ["am j undergrad res"] = "American journal of undergraduate research",
  ["am j urol res"] = "American journal of urology research",
  ["am j vet med"] = "American journal of veterinary medicine",
  ["am j vet res"] = "American journal of veterinary research",
  ["am j virol"] = "American journal of virology",
  ["am jew"] = "The American Jewish year book",
  ["am jew arch"] = "American Jewish archives",
  ["am jew hist"] = "American Jewish history",
  ["am jew hist q"] = "American Jewish historical quarterly",
  ["am lab"] = "American Laboratory",
  ["am lat"] = "América latina",
  ["am lat hist econ"] = "América Latina en la historia económica : boletín de fuentes",
  ["am laund dig"] = "American laundry digest",
  ["am law rep alr 3rd cases annot"] = "American law reports. ALR 3d. Cases and annotations",
  ["am libr"] = "American libraries",
  ["am lit"] = "American literature: a journal of literary history, criticism and bibliography",
  ["am lit hist"] = "American literary history",
  ["am lung assoc bull"] = "American Lung Association bulletin",
  ["am malacol bull"] = "American malacological bulletin",
  ["am math mon"] = "The American mathematical monthly : the official journal of the Mathematical Association of America",
  ["am med j"] = "American medical journal",
  ["am med news"] = "American medical news",
  ["am med stud res j"] = "American medical student research journal",
  ["am midl nat"] = "The American midland naturalist",
  ["am mineral"] = "American Mineralogist",
  ["am mus novit"] = "American Museum novitates",
  ["am nat"] = "The American naturalist",
  ["am neptune"] = "The American neptune",
  ["am nurse"] = "The American nurse",
  ["am nurse today"] = "American nurse today",
  ["am opt vis"] = "American optical vision",
  ["am orthopt j"] = "The American orthoptic journal",
  ["am ose rev"] = "American OSE review",
  ["am pharm"] = "American pharmacy",
  ["am pharm rev"] = "American pharmaceutical review",
  ["am philat"] = "The American philatelist",
  ["am philos q"] = "American philosophical quarterly",
  ["am polit q"] = "American politics quarterly",
  ["am polit sci rev"] = "The American political science review",
  ["am potato j"] = "American potato journal",
  ["am pract dig treat"] = "American practitioner and digest of treatment",
  ["am presbyt"] = "American Presbyterian",
  ["am prof pharm"] = "American professional pharmacist",
  ["am prospect"] = "The American prospect",
  ["am psychol"] = "The American psychologist",
  ["am q"] = "American quarterly",
  ["am rehabil"] = "American rehabilitation",
  ["am rep"] = "AM [reports]. United States. Office of Aviation Medicine",
  ["am rev can stud"] = "The American review of Canadian studies",
  ["am rev public adm"] = "American review of public administration",
  ["am rev respir dis"] = "The American review of respiratory disease",
  ["am rev sov med"] = "American review of Soviet medicine",
  ["am rev tuberc"] = "American review of tuberculosis",
  ["am sch board j"] = "The American school board journal",
  ["am scholar"] = "The American scholar",
  ["am sci"] = "American scientist",
  ["am soc clin oncol educ book"] = "American Society of Clinical Oncology educational book. American Society of Clinical Oncology. Annual Meeting",
  ["am soc legion honor mag"] = "The American Society Legion of Honor magazine",
  ["am sociol"] = "The American sociologist",
  ["am sociol rev"] = "American sociological review",
  ["am spect"] = "The American spectator",
  ["am speech"] = "American speech",
  ["am stat"] = "The American statistician",
  ["am stud (lawrence)"] = "American studies (Lawrence, Kan.)",
  ["am stud int"] = "American studies international",
  ["am surg"] = "The American surgeon",
  ["am univ field staff rep afr"] = "American Universities Field Staff reports: Africa",
  ["am univ field staff rep asia"] = "American Universities Field Staff reports: Asia",
  ["am univ field staff rep north am"] = "American Universities Field Staff reports: North America",
  ["am univ field staff rep south am"] = "American Universities Field Staff reports: South America",
  ["am univ j gend soc policy law"] = "The American University journal of gender, social policy & the law",
  ["am univ j int law policy"] = "The American University journal of international law and policy",
  ["am univ law rev"] = "The American University law review",
  ["am univ stud ser ix hist"] = "American University Studies",
  ["am zool"] = "American Zoologist",
  ["ama am j dis child"] = "A.M.A. American journal of diseases of children",
  ["ama arch derm"] = "A.M.A. archives of dermatology",
  ["ama arch derm syphilol"] = "A.M.A. archives of dermatology and syphilology",
  ["ama arch gen psychiatry"] = "A.M.A. archives of general psychiatry",
  ["ama arch ind health"] = "A.M.A. archives of industrial health",
  ["ama arch ind hyg occup med"] = "A.M.A. archives of industrial hygiene and occupational medicine",
  ["ama arch intern med"] = "A.M.A. archives of internal medicine",
  ["ama arch neurol"] = "A.M.A. archives of neurology",
  ["ama arch neurol psychiatry"] = "A.M.A. archives of neurology and psychiatry",
  ["ama arch ophthalmol"] = "A.M.A. archives of ophthalmology",
  ["ama arch otolaryngol"] = "A.M.A. archives of otolaryngology",
  ["ama arch pathol"] = "A.M.A. archives of pathology",
  ["ama arch surg"] = "A.M.A. archives of surgery",
  ["ama j dis child"] = "A.M.A. journal of diseases of children",
  ["ama j ethics"] = "AMA journal of ethics",
  ["amatus lusit"] = "Amatus Lusitanus; revista de medicina e cirurgia",
  ["amb express"] = "AMB Express",
  ["amb rev assoc med bras"] = "AMB : revista da Associação Médica Brasileira",
  ["ambito encuentros"] = "Ámbito de encuentros",
  ["ambul care"] = "Ambulatory care",
  ["ambul outreach"] = "Ambulatory outreach",
  ["ambul pediatr"] = "Ambulatory pediatrics : the official journal of the Ambulatory Pediatric Association",
  ["ambul surg"] = "Ambulatory surgery",
  ["amd tr rep"] = "AMD-TR [reports]. United States. Aerospace Medical Division",
  ["amdi boll"] = "AMDI bollettino",
  ["ame case rep"] = "AME case reports",
  ["ame med j"] = "AME medical journal",
  ["amer econ rev"] = "American Economic Review",
  ["amer economist"] = "American Economist",
  ["amer enterprise"] = "American Enterprise",
  ["amer historical rev"] = "American Historical Review",
  ["amer j agr econ"] = "American Journal of Agricultural Economics",
  ["amer j econ sociology"] = "American Journal of Economics and Sociology",
  ["amer j math"] = "American Journal of Mathematics",
  ["amer j math management sci"] = "American Journal of Mathematical and Management Sciences",
  ["amer j phys"] = "American Journal of Physics",
  ["amer law econ rev"] = "American Law and Economics Review",
  ["amer math monthly"] = "American Mathematical Monthly",
  ["amer math soc colloq publ"] = "American Mathematical Society Colloquium Publications",
  ["amer math soc transl ser 2"] = "American Mathematical Society Translations, Series 2",
  ["amer philos quart monogr ser"] = "American Philosophical Quarterly Monograph Series",
  ["amer polit sci rev"] = "American Political Science Review",
  ["amer prospect"] = "American Prospect",
  ["amer real estate urban econ assoc j"] = "American Real Estate and Urban Economics Association Journal",
  ["amer sci"] = "American Scientist",
  ["amer statist"] = "The American Statistician",
  ["amer statistician"] = "American Statistician",
  ["amer univ stud ser v philos"] = "American University Studies",
  ["amerasia j"] = "Amerasia journal",
  ["america (ny)"] = "America",
  ["americas (acad am francisc hist)"] = "Americas (Academy of American Franciscan History)",
  ["americas (engl ed)"] = "Américas",
  ["amfar rep"] = "AmFAR report",
  ["amhc forum"] = "AMHC forum",
  ["amia annu symp proc"] = "AMIA Annual Symposium Proceedings",
  ["amia jt summits transl sci proc"] = "AMIA Joint Summits on Translational Science proceedings. AMIA Joint Summits on Translational Science",
  ["amic j"] = "The Amicus journal",
  ["amici molluscarum"] = "Amici Molluscarum",
  ["amino acids"] = "Amino acids",
  ["ammonia plant saf relat facil"] = "Ammonia Plant Safety and Related Facilities",
  ["ampersand (oxford)"] = "Ampersand (Oxford, UK)",
  ["amphib reptil"] = "Amphibia-reptilia : publication of the Societas Europaea Herpetologica",
  ["amphib reptile conserv"] = "Amphibian & reptile conservation",
  ["amrc open res"] = "AMRC open research",
  ["amrl tr"] = "AMRL-TR. Aerospace Medical Research Laboratories (U.S.)",
  ["ams progr math lecture ser"] = "AMS Progress in Mathematics Lecture Series",
  ["ams short course lecture notes"] = "AMS Short Course Lecture Notes",
  ["ams-maa joint lecture ser"] = "AMS-MAA Joint Lecture Series",
  ["ams/ip stud adv math"] = "AMS/IP Studies in Advanced Mathematics",
  ["ams/maa probl books ser"] = "AMS/MAA Problem Books Series",
  ["ams/maa spectr"] = "AMS/MAA Spectrum",
  ["ams/maa textb"] = "AMS/MAA Textbooks",
  ["amst stud theory hist linguist sci ser iii stud hist lang sci"] = "Amsterdam Studies in the Theory and History of Linguistic Science. Series III: Studies in the History of the Language Sciences",
  ["amst stud theory hist linguist sci ser iv curr issues linguist theory"] = "Amsterdam Studies in the Theory and History of Linguistic Science",
  ["amyotroph lateral scler"] = "Amyotrophic lateral sclerosis : official publication of the World Federation of Neurology Research Group on Motor Neuron Diseases",
  ["amyotroph lateral scler frontotemporal degener"] = "Amyotrophic lateral sclerosis & frontotemporal degeneration",
  ["amyotroph lateral scler other motor neuron disord"] = "Amyotrophic lateral sclerosis and other motor neuron disorders : official publication of the World Federation of Neurology, Research Group on Motor Neuron Diseases",
  ["aménage territ"] = "Aménagement du territoire",
  ["an  azevedos"] = "Anais Azevedos",
  ["an acad bras cienc"] = "Anais da Academia Brasileira de Ciências",
  ["an acad brasil cienc"] = "Anais da Academia Brasileira de Ciencias. Annals of the Brazilian Academy of Sciences",
  ["an acad brasil ciênc"] = "Anais da Academia Brasileira de Ciências",
  ["an acad cienc medicas fisicas nat habana"] = "Anales. Real Academia de Ciencias Médicas, Físicas y Naturales de La Habana",
  ["an acad geogr hist guatem"] = "Anales de la Academia de Geografía e Historia de Guatemala. Academia de Geografía e Historia de Guatemala",
  ["an academ med medellin"] = "Anales. Academia de Medicina, Medellín",
  ["an antropol"] = "Anales de antropología",
  ["an argent oftalmol"] = "Anales argentinos de oftalmología",
  ["an ark nurse"] = "AN. Arkansas nurse",
  ["an asoc fis argent"] = "Anales de la Asociacion Fisica Argentina",
  ["an asoc quim argent"] = "Anales de la Asociacion Quimica Argentina",
  ["an asoc quim farm uruguay"] = "Anales. Asociación de Química y Farmacia del Uruguay",
  ["an ateneo clin quir montev"] = "Anales. Ateneo de Clínica Quirúrgica, Montevideo",
  ["an biol"] = "Anales de biología",
  ["an bras derm sifilogr"] = "Anais brasileiros de dermatologia e sifilografia",
  ["an bras dermatol"] = "Anais brasileiros de dermatologia",
  ["an bras ginecol"] = "Anais brasileiros de ginecologia",
  ["an bromatol"] = "Anales de Bromatologia",
  ["an casa salud valdecilla"] = "Anales de la Casa de Salud Valdecilla",
  ["an catedra patol clin tuberc"] = "Anales de la Cátedra de Patología y Clínica de la Tuberculosis",
  ["an chil hist med"] = "Anales chilenos de historia de la medicina",
  ["an cir (rosario)"] = "Anales de cirugía",
  ["an enferm"] = "Anais de enfermagem",
  ["an esc nac cienc biol"] = "Anales. Escuela Nacional de Ciencias Biológicas (Mexico)",
  ["an esc nacl saude publica med trop (lisb)"] = "Anais da Escola Nacional de Saude Publica e de Medicina Tropical",
  ["an esp odontoestomatol"] = "Anales españoles de odontoestomatología",
  ["an esp pediatr"] = "Anales Espanoles de Pediatria",
  ["an fac farm porto"] = "Anais da Faculdade de Farmácia do Porto",
  ["an fac med lima"] = "Anales. Universidad Nacional Mayor de San Marcos. Facultad de Medicina",
  ["an fac med montev"] = "Anales de la Facultad de Medicina de Montevideo",
  ["an fac med porto alegre"] = "Anais da Faculdade de Medicina de Pôrto Alegre",
  ["an fac med recife"] = "Anais da Faculdade de Medicina do Recife",
  ["an fac med univ repub montev urug"] = "Anales de la Facultad de Medicina, Universidad de la Republica, Montevideo, Uruguay",
  ["an fac med univ sao paulo"] = "Anais. Universidade de São Paulo. Faculdade de Medicina",
  ["an fac odontol"] = "Anales de la Facultad de Odontologia",
  ["an fac odontol univ fed pernambuco"] = "Anais da Faculdade de Odontologia da Universidade Federal de Pernambuco",
  ["an fac quim farm (santiago)"] = "Anales de la Facultad de Química y Farmacia",
  ["an farm hosp (madr)"] = "Anales de farmacia hospitalaria",
  ["an fís monogr"] = "Anales de Física",
  ["an gazi husrev begove bibl"] = "Anali Gazi Husrev-begove biblioteke. Gazi Husrevbegova biblioteka u Sarajevu",
  ["an hist"] = "Anais de história : publicação do Departamento de História da Faculdade de Filosofia, Ciências e Letras de Assis",
  ["an hist antigua mediev"] = "Anales de historia antigua y medieval",
  ["an hist contemp"] = "Anales de historia contemporánea",
  ["an hosp alvear b aires"] = "Anales. Hospital Torcuato de Alvear (Buenos Aires, Argentina). Ateneo de Patologia y Clínica Médica",
  ["an hosp st cruz san pablo"] = "Anales del Hospital de la Santa Cruz y San Pablo",
  ["an hosp varsovia"] = "Anales. Hospital Varsovia",
  ["an hui ji shu shi fan xue yuan xue bao"] = "An hui ji shu shi fan xue yuan xue bao",
  ["an ilustre col of med"] = "Anales (Ilustre Colegio Oficial de Médicos (Lerida, Spain))",
  ["an inst barraquer"] = "Anales del Instituto Barraquer",
  ["an inst bio"] = "Anales del Instituto de Biología",
  ["an inst estud madr"] = "Anales del Instituto de Estudios Madrileños. Instituto de Estudios Madrileños (Consejo Superior de Investigaciones Científicas)",
  ["an inst hig med trop (lisb)"] = "Anais do Instituto de Higiene e Medicina Tropical",
  ["an inst invest odontol (maracaibo)"] = "Anales - Instituto de Investigaciones Odontológicas, Universidad del Zulia",
  ["an inst invest vet"] = "Anales del Instituto de Investigaciones Veterinarias",
  ["an inst mat univ nac autónoma méxico"] = "Anales del Instituto de Matemáticas",
  ["an inst med trop (lisb)"] = "Anais do Instituto de Medicina Tropical",
  ["an inst nac antropol hist"] = "Anales del Instituto Nacional de Antropologia e Historia",
  ["an jard bot madr (1979)"] = "Anales del Jardín Botánico de Madrid (1979)",
  ["an lima"] = "Anales. Universidad Nacional Mayor de San Marcos. Facultad de Farmacia y Bioquimica",
  ["an matern sao paulo"] = "Anais. São Paulo (Brazil). Maternidade",
  ["an med (barc)"] = "Anales de medicina",
  ["an med (lima)"] = "Anales de medicina (Lima, Peru)",
  ["an med ateneo ramon cajal mex"] = "Anales de medicina. Ateneo Ramón y Cajal, Mexico",
  ["an med cir"] = "Anales de medicina: Cirugía",
  ["an med concepc"] = "Anales médicos de Concepción",
  ["an med espec"] = "Anales de medicina: Especialidades",
  ["an med interna"] = "Anales de medicina interna (Madrid, Spain : 1984)",
  ["an med med"] = "Anales de medicina: Medicina",
  ["an med publica"] = "Anales de medicina pública",
  ["an microbiol (rio j)"] = "Anais de microbiologia",
  ["an nestle (b aires)"] = "Anales Nestlé",
  ["an nestle (rio de janeiro)"] = "Anais Nestle",
  ["an nestle (rio j)"] = "Anais Nestlé",
  ["an nuevo leon mex"] = "Anales. Nuevo León, Mexico. Universidad, Monterey. Instituto de Investigaciones Científicas",
  ["an otorrinolaringol ibero am"] = "Anales Otorrinolaringologicos Iberoamericanos",
  ["an paul med cir"] = "Anais paulistas de medicina e cirurgia",
  ["an pediatr"] = "Anales de Pediatria",
  ["an pediatr (barc)"] = "Anales de pediatría (Barcelona, Spain : 2003)",
  ["an pediatr (engl ed)"] = "Anales de pediatría",
  ["an programa acad med (lima)"] = "Anales del Programa Académico de Medicina",
  ["an quim"] = "Anales de Quimica",
  ["an r acad med zaragoza"] = "Anales de la Real Academia de Medicina de Zaragoza. Sesión inaugural",
  ["an r acad nac med (madr)"] = "Anales de la Real Academia Nacional de Medicina",
  ["an real acad farm"] = "Anales de la Real Academia de Farmacia",
  ["an rom sov ser med gen"] = "Analele Româno-Sovietice. Seria medicină generală",
  ["an sist sanit navar"] = "Anales del sistema sanitario de Navarra",
  ["an soc geogr hist"] = "Anales de la Sociedad de Geografía e Historia. Sociedad de Geografía e Historia de Guatemala",
  ["an soc med quir guayas"] = "Anales. Sociedad Médico-Quirúrgica del Guayas",
  ["an soc mex hist cienc tec"] = "Anales de la Sociedad Mexicana de Historia de la Ciencia y de la Tecnología",
  ["an soc mex oftalmol otorinolaringol"] = "Anales. Sociedad Mexicana de Oftalmología y Oto-rino-laringología",
  ["an soc pueric b aires"] = "Anales de la Sociedad de Puericultura de Buenos Aires",
  ["an stiint univ al i cuza iasi mat (ns)"] = "Analele Stiintifice ale Universitatii “Al. I. Cuza” din Iasi. Serie Noua. Matematica. Annals of the Alexandru Cuza University. Mathematics",
  ["an stiint univ “ovidius” constanta ser mat"] = "Analele Stiintifice ale Universitatii “Ovidius” Constanta. Seria Matematica. Mathematical Journal of the Ovidius University of Constantza",
  ["an univ bucur inform"] = "Analele Universitatii Bucuresti. Informatica",
  ["an univ bucur istor"] = "Analele Universității București: Istorie. Universitatea din București",
  ["an univ bucurești mat"] = "Analele Universităţii București",
  ["an univ craiova ser mat inform"] = "Analele Universitatii din Craiova. Seria Matematica-Informatica. Annals of the University of Craiova. Mathematics and Computer Science Series",
  ["an univ dunarea de jos galati fasc ii mat fiz mec teor"] = "Analele Universitatii “Dunarea de Jos” din Galati. Fascicula II. Matematica, Fizica, Mecanica Teoretica. Annals of the “Dunarea de Jos” University of Galati. Fascicle II. Mathematics, Physics, Theoretical Mechanics",
  ["an univ galaţi fasc ii mat fiz mec teoret"] = "Analele Universităţii din Galaţi",
  ["an univ nac asuncion"] = "Anales. Universidad Nacional de Asunción. Facultad de Ciencias Médicas",
  ["an univ oradea fasc mat"] = "Analele Universitatii din Oradea. Fascicola Matematica",
  ["an univ spiru haret ser mat-inform"] = "Analele Universitatii Spiru Haret. Seria Matematica-Informatica. Annals of Spiru Haret University. Mathematics-Informatics Series",
  ["an univ st domingo"] = "Anales. Universidad de Santo Domingo",
  ["an univ timișoara ser mat-inform"] = "Universităţii din Timișoara",
  ["an univ timișoara ser ştiinţ fiz"] = "Analele Universităţii din Timișoara",
  ["an univ târgoviște fasc i"] = "Analele Universităţii din Târgoviște",
  ["an univ vest timis ser mat-inform"] = "Analele. Universitatii de Vest din Timisoara. Seria Matematica-Informatica. Annals of West University of Timisoara. Mathematics and Computer Science",
  ["ana clin conf"] = "ANA clinical conferences",
  ["ana clin sess"] = "ANA clinical sessions",
  ["ana publ"] = "American Nurses Association Publications",
  ["anadolu kardiyol derg"] = "Anadolu kardiyoloji dergisi : AKD = the Anatolian journal of cardiology",
  ["anadolu klinigi"] = "Anadolu kliniği",
  ["anadolu psikiyatri derg"] = "Anadolu psikiyatri dergisi",
  ["anaesth anaesth"] = "Anaesthesia and anaesthetics",
  ["anaesth cases"] = "Anaesthesia cases",
  ["anaesth crit care pain med"] = "Anaesthesia, critical care & pain medicine",
  ["anaesth intensive care"] = "Anaesthesia and intensive care",
  ["anaesth rep"] = "Anaesthesia reports",
  ["anaesth resusc intensive ther"] = "Anaesthesia, resuscitation, and intensive therapy",
  ["anaesth surg open access j"] = "Anaesthesia & surgery open access journal",
  ["anaesthesiol intensive ther"] = "Anaesthesiology intensive therapy",
  ["anaesthesiol reanim"] = "Anaesthesiologie und Reanimation",
  ["anais hist alem mar"] = "Anais de história de além-mar",
  ["anal appl"] = "Analysis and Applications",
  ["anal appl (singap)"] = "Analysis and Applications",
  ["anal bioanal chem"] = "Analytical and bioanalytical chemistry",
  ["anal bioanalchem"] = "Analytical and Bioanalytical Chemistry",
  ["anal biochem"] = "Analytical Biochemistry",
  ["anal cell pathol"] = "Analytical cellular pathology : the journal of the European Society for Analytical Cellular Pathology",
  ["anal cell pathol (amst)"] = "Analytical cellular pathology (Amsterdam)",
  ["anal chem"] = "Analytical Chemistry",
  ["anal chem (washington, dc, u s)"] = "Analytical Chemistry (Washington, DC, United States)",
  ["anal chem insights"] = "Analytical chemistry insights",
  ["anal chem lett"] = "Analytical Chemistry Letters",
  ["anal chem res"] = "Analytical Chemistry Research",
  ["anal chem symp ser"] = "Analytical Chemistry Symposia Series",
  ["anal chim acta"] = "Analytica chimica acta",
  ["anal chim acta x"] = "Analytica chimica acta: X",
  ["anal chim acta: x"] = "Analytica Chimica Acta: X",
  ["anal commun"] = "Analytical Communications",
  ["anal gambl behav"] = "Analysis of gambling behavior",
  ["anal geom metr spaces"] = "Analysis and Geometry in Metric Spaces",
  ["anal geom number theory"] = "Analysis, Geometry and Number Theory. An International Journal",
  ["anal lett"] = "Analytical Letters",
  ["anal math"] = "Analysis Mathematica",
  ["anal math phys"] = "Analysis and Mathematical Physics",
  ["anal methods"] = "Analytical methods : advancing methods and applications",
  ["anal methods accid res"] = "Analytic methods in accident research",
  ["anal methods soc res"] = "Analytical Methods for Social Research",
  ["anal methods spec funct"] = "Analytical Methods and Special Functions",
  ["anal modern apl"] = "Analiză Modernă şi Aplicaţii",
  ["anal pde"] = "Analysis & PDE",
  ["anal pdes set"] = "Analysis for PDEs Set",
  ["anal philos"] = "Analytic Philosophy",
  ["anal polit"] = "Análisis político (Bogotá, Colombia)",
  ["anal proc"] = "Analytical Proceedings",
  ["anal proc incl anal commun"] = "Analytical Proceedings including Analytical Communications",
  ["anal prévision"] = "Analyse Prévision",
  ["anal psychol"] = "Analytische Psychologie",
  ["anal quant cytol"] = "Analytical and quantitative cytology",
  ["anal quant cytol histol"] = "Analytical and quantitative cytology and histology",
  ["anal quant cytopathol histpathol"] = "Analytical and quantitative cytopathology and histopathology",
  ["anal sci"] = "Analytical sciences : the international journal of the Japan Society for Analytical Chemistry",
  ["anal sci adv"] = "Analytical Science Advances",
  ["anal sci: x-ray struct anal online"] = "Analytical Sciences: X-Ray Structure Analysis Online",
  ["anal sens"] = "Analysis & Sensing",
  ["anal soc"] = "Analise Social",
  ["anal soc biol bogota"] = "Anales. Sociedad de Biologia de Bogotá",
  ["anal soc issues public policy"] = "Analyses of social issues and public policy : ASAP",
  ["anal spectrosc lib"] = "Analytical Spectroscopy Library",
  ["anal theory appl"] = "Analysis in Theory and Applications",
  ["anal verbal behav"] = "The Analysis of verbal behavior",
  ["analecta calas"] = "Analecta Calasanctiana",
  ["analecta med"] = "Analecta médica",
  ["analecta med hist"] = "Analecta medico-historica",
  ["analecta sacra tarracon"] = "Analecta sacra Tarraconensia; revista de ciencias histórico-eclesiásticas",
  ["analecta vet"] = "Analecta veterinaria",
  ["analectas farm gerundenses"] = "Analectas farmacéutico-gerundenses",
  ["analg resusc"] = "Analgesia & resuscitation : current research",
  ["analgesia (elmsford n y)"] = "Analgesia (Elmsford, N.Y.)",
  ["analog circuits signal process"] = "Analog Circuits and Signal Processing",
  ["analog dialogue"] = "Analog Dialogue",
  ["analog integr circuits signal process"] = "Analog integrated circuits and signal processing",
  ["analog sci fict sci fact"] = "Analog science fiction/science fact",
  ["analysis (berlin)"] = "Analysis. International Mathematical Journal of Analysis and its Applications",
  ["analysis (munich)"] = "Analysis",
  ["analysis (oxford)"] = "Analysis",
  ["analyst (cambridge, u k)"] = "Analyst (Cambridge, United Kingdom)",
  ["analyst (lond)"] = "The Analyst",
  ["anasth intensivther notf med"] = "Anasthesie, Intensivtherapie, Notfallmedizin",
  ["anasth intensivther notfallmed"] = "Anästhesie, Intensivtherapie, Notfallmedizin",
  ["anasthesiol intensivmed notfallmed schmerzther"] = "Anästhesiologie, Intensivmedizin, Notfallmedizin, Schmerztherapie : AINS",
  ["anasthesiol intensivmed prax"] = "Anästhesiologische und intensivmedizinische Praxis",
  ["anat anz"] = "Anatomischer Anzeiger",
  ["anat cell biol"] = "Anatomy & cell biology",
  ["anat chir"] = "Anatomia e chirurgia",
  ["anat clin"] = "Anatomia clinica",
  ["anat embryol (berl)"] = "Anatomy and embryology",
  ["anat histol embryol"] = "Anatomia, histologia, embryologia",
  ["anat pathol"] = "Anatomic pathology (Chicago, Ill. : annual)",
  ["anat physiol"] = "Anatomy & physiology : current research",
  ["anat rec"] = "Anatomical Record",
  ["anat rec (hoboken)"] = "Anatomical record (Hoboken, N.J. : 2007)",
  ["anat rec a discov mol cell evol biol"] = "The anatomical record. Part A, Discoveries in molecular, cellular, and evolutionary biology",
  ["anat rec b new anat"] = "Anatomical record. Part B, New anatomist",
  ["anat rec part a"] = "Anatomical Record Part A: Discoveries in Molecular, Cellular, and Evolutionary Biology",
  ["anat rec part b"] = "Anatomical Record Part B: The Anatomist",
  ["anat rec suppl"] = "The Anatomical record. Supplement : an offical publication of the American Association of Anatomists",
  ["anat res int"] = "Anatomy research international",
  ["anat sci educ"] = "Anatomical sciences education",
  ["anat sci int"] = "Anatomical science international",
  ["anatol j cardiol"] = "Anatolian journal of cardiology",
  ["anatol sport res"] = "Anatolia sport research",
  ["anatom histol embryol"] = "Anatomia, Histologia, Embryologia",
  ["anbar j vet sci"] = "Al-Anbar journal of veterinary sciences",
  ["anc biomol"] = "Ancient biomolecules",
  ["anc comment aristotle"] = "Ancient Commentators on Aristotle",
  ["anc mediterr mediev texts contexts"] = "Ancient Mediterranean and Medieval Texts and Contexts",
  ["anc philos"] = "Ancient philosophy (Garland Publishing)",
  ["anc sci"] = "Ancient science",
  ["anc sci life"] = "Ancient science of life",
  ["ancient biomol"] = "Ancient Biomolecules",
  ["andean geol"] = "Andean Geology",
  ["andover newton q"] = "Andover Newton quarterly",
  ["androg clin res ther"] = "Androgens: clinical research and therapeutics",
  ["andrology (los angel)"] = "Andrology : open access",
  ["anest rianim"] = "Anestesia e rianimazione",
  ["anesteziol reanimatol"] = "Anesteziologiia i reanimatologiia",
  ["anestezjol intens ter"] = "Anestezjologia intensywna terapia",
  ["anesth anal"] = "Anesthésie et analgésie",
  ["anesth analg"] = "Anesthesia and analgesia",
  ["anesth analg (paris)"] = "Anesthésie, analgésie, réanimation",
  ["anesth essays res"] = "Anesthesia, essays and researches",
  ["anesth pain control dent"] = "Anesthesia & pain control in dentistry",
  ["anesth pain med"] = "Anesthesiology and pain medicine",
  ["anesth pain med (seoul)"] = "Anesthesia and pain medicine",
  ["anesth prog"] = "Anesthesia progress",
  ["anesthesiol case rep"] = "Anesthesiology case reports",
  ["anesthesiol clin"] = "Anesthesiology clinics",
  ["anesthesiol clin north am"] = "Anesthesiology clinics of North America",
  ["anesthesiol clin north america"] = "Anesthesiology clinics of North America",
  ["anesthesiol res pract"] = "Anesthesiology research and practice",
  ["anesthesiol rev"] = "Anesthesiology review",
  ["angeiol ann soc fr angeiol histopathol"] = "Angéiologie et annales de la Société française d'angéiologie et d'histopathologie",
  ["angew carabidol"] = "Angewandte Carabidologie",
  ["angew chem"] = "Angewandte Chemie",
  ["angew chem - int ed"] = "Angewandte Chemie International Edition",
  ["angew chem int ed"] = "Angewandte Chemie International Edition",
  ["angew chem int ed engl"] = "Angewandte Chemie (International ed. in English)",
  ["angew chem weinheim bergstr ger"] = "Angewandte Chemie (Weinheim an der Bergstrasse, Germany)",
  ["angew chem, int ed"] = "Angewandte Chemie, International Edition",
  ["angew chem, int ed engl"] = "Angewandte Chemie, International Edition in English",
  ["angew entomol"] = "Angewandte Entomologie",
  ["angew landschökol"] = "Angewandte Landschaftsökologie",
  ["angew makromol chem"] = "Angewandte Makromolekulare Chemie",
  ["angew math"] = "Angewandte Mathematik",
  ["angew parasitol"] = "Angewandte Parasitologie",
  ["angew pflanzensoziol"] = "Angewandte Pflanzensoziologie",
  ["angew statist ökonom"] = "Angewandte Statistik und Ökonometrie",
  ["angewandte botanik / vereinigung fur angewandte botanik [und] deutsche gesellschaft fur qualitatsforschung|j appl bot food qual"] = "Journal of applied botany and food quality",
  ["angiol open access"] = "Angiology: open access",
  ["angiol sosud khir"] = "Angiologii︠a︡ i sosudistai︠a︡ khirurgii︠a︡ = Angiology and vascular surgery",
  ["angl episcop hist"] = "Anglican and Episcopal history",
  ["angle orthod"] = "The Angle orthodontist",
  ["anglican theol rev"] = "Anglican theological review",
  ["anglo am law rev"] = "The Anglo-American law review",
  ["anglo ger med rev"] = "Anglo-German medical review. Deutsch-englische medizinische Rundschau",
  ["anglo sax engl"] = "Anglo-Saxon England",
  ["anhui nong ye da xue xue bao"] = "Anhui nong ye da xue xue bao",
  ["anhui nongxueyuan xuebao"] = "Anhui nongxueyuan xuebao = Journal of Anhui Agricultural College",
  ["anhui nongye kexue"] = "Anhui nongye kexue",
  ["anhui shida xuebao ziran kexue ban"] = "Anhui Shida Xuebao",
  ["anim agenda"] = "The animals' agenda",
  ["anim behav"] = "Animal behaviour",
  ["anim behav cogn"] = "Animal behavior and cognition",
  ["anim behav monogr"] = "Animal Behaviour Monographs",
  ["anim biodivers conserv"] = "Animal biodiversity and conservation",
  ["anim biol anim husb"] = "Animal biology & animal husbandry",
  ["anim biol leiden neth"] = "Animal biology (Leiden, Netherlands)",
  ["anim biosci"] = "Animal bioscience",
  ["anim biotechnol"] = "Animal biotechnology",
  ["anim biotechnol bull"] = "Animal biotechnology bulletin",
  ["anim biotelem"] = "Animal Biotelemetry",
  ["anim blood groups biochem genet"] = "Animal blood groups and biochemical genetics",
  ["anim blood groups biochem genet suppl"] = "Animal Blood Groups and Biochemical Genetics. Supplement",
  ["anim cells syst"] = "Animal Cells and Systems",
  ["anim cells syst (seoul)"] = "Animal cells and systems",
  ["anim cogn"] = "Animal cognition",
  ["anim cognit"] = "Animal Cognition",
  ["anim conserv"] = "Animal conservation",
  ["anim def"] = "The Animals' defender",
  ["anim dis"] = "Animal Diseases",
  ["anim feed sci technol"] = "Animal Feed Science and Technology",
  ["anim front"] = "Animal frontiers : the review magazine of animal agriculture",
  ["anim gene"] = "Animal Gene",
  ["anim genet"] = "Animal genetics",
  ["anim genet resour"] = "Animal genetic resources = Ressources génétiques animales = Recursos genéticos animales",
  ["anim health res rev"] = "Animal health research reviews",
  ["anim learn behav"] = "Animal learning & behavior",
  ["anim microbiome"] = "Animal microbiome",
  ["anim models exp med"] = "Animal Models and Experimental Medicine",
  ["anim nutr"] = "Animal nutrition (Zhongguo xu mu shou yi xue hui)",
  ["anim nutr feed technol"] = "Animal Nutrition and Feed Technology",
  ["anim open space"] = "Animal - Open Space",
  ["anim prod"] = "Animal production",
  ["anim prod sci"] = "Animal production science",
  ["anim reprod"] = "Animal reproduction",
  ["anim reprod sci"] = "Animal reproduction science",
  ["anim sci"] = "Animal Science",
  ["anim sci cases"] = "Animal Science Cases",
  ["anim sci j"] = "Animal science journal = Nihon chikusan Gakkaihō",
  ["anim sci pap rep"] = "Animal science papers and reports",
  ["anim sci technol"] = "Animal Science and Technology",
  ["anim syst evol diversity"] = "Animal systematics, evolution and diversity",
  ["anim taxon ecol"] = "Animal Taxonomy and Ecology",
  ["anim technol"] = "Animal technology : journal of the Institute of Animal Technicians",
  ["anim welf"] = "Animal welfare (South Mimms, England)",
  ["anim welf etol tartastechnol"] = "Animal welfare, etológia és tartástechnológia",
  ["anim welfare"] = "Animal Welfare",
  ["anim welfare inf cent bull"] = "Animal Welfare Information Center Bulletin",
  ["animal model exp med"] = "Animal models and experimental medicine",
  ["animals (basel)"] = "Animals : an open access journal from MDPI",
  ["anjisuan he shengwu ziyuan"] = "Anjisuan he shengwu ziyuan = Amino acids & biotic resources",
  ["ankara univ hekim fak derg"] = "Ankara Üniversitesi Diş Hekimliği Fakültesi dergisi = The Journal of the Dental Faculty of Ankara University",
  ["anl rep"] = "ANL",
  ["ann 3d print med"] = "Annals of 3D Printed Medicine",
  ["ann acad med singap"] = "Annals of the Academy of Medicine, Singapore",
  ["ann acad med singapore"] = "Annals of the Academy of Medicine, Singapore",
  ["ann acad med stetin"] = "Annales Academiae Medicae Stetinensis",
  ["ann acad med stetin supl"] = "Annales Academiae Medicae Stetinensis. Suplement",
  ["ann acad rom sci"] = "Annals of the Academy of Romanian Scientists",
  ["ann acad rom sci ser math appl"] = "Annals of the Academy of Romanian Scientists. Series on Mathematics and its Applications",
  ["ann acad sci fenn [biol]"] = "Annales Academiae Scientiarum Fennicae. Series A IV, Biologica",
  ["ann acad sci fenn a"] = "Annales Academiae Scientiarum Fennicae. Ser. A.5, Medica",
  ["ann acad sci fenn a 3 geol geogr"] = "Annales Academiae Scientiarum Fennicae. Series A. 3, Geologica--geographica",
  ["ann acad sci fenn biol"] = "Suomalainen Tiedeakatemia toimituksia. Sar. A.4: Biologica",
  ["ann acad sci fenn math"] = "Annales Academiae Scientiarum Fennicae. Mathematica",
  ["ann acad sci fenn math diss"] = "Academiæ Scientiarum Fennicae. Annales. Mathematica. Dissertationes",
  ["ann accad ital sci for"] = "Annali Accademia Italiana di Scienze Forestali",
  ["ann adv automot med"] = "Annals of advances in automotive medicine. Association for the Advancement of Automotive Medicine. Annual Scientific Conference",
  ["ann afr med"] = "Annals of African medicine",
  ["ann agrar sci"] = "Annals of agrarian science",
  ["ann agric crop sci"] = "Annals of agricultural & crop sciences",
  ["ann agric environ med"] = "Annals of agricultural and environmental medicine : AAEM",
  ["ann agric sci"] = "Annals of agricultural science",
  ["ann allergy"] = "Annals of allergy",
  ["ann allergy asthma immunol"] = "Annals of allergy, asthma & immunology : official publication of the American College of Allergy, Asthma, & Immunology",
  ["ann am acad pol soc sci"] = "The Annals of the American Academy of Political and Social Science",
  ["ann am assoc geogr"] = "Annals of the American Association of Geographers",
  ["ann am thorac soc"] = "Annals of the American Thoracic Society",
  ["ann amer acad polit soc sci"] = "Annals of the American Academy of Political and Social Science",
  ["ann anat"] = "Annals of anatomy = Anatomischer Anzeiger : official organ of the Anatomische Gesellschaft",
  ["ann anat pathol (paris)"] = "Annales d'anatomie pathologique",
  ["ann anesthesiol fr"] = "Annales de l'anesthésiologie française",
  ["ann anim sci"] = "Annals of Animal Science",
  ["ann anthropol pract"] = "Annals of anthropological practice",
  ["ann appl biol"] = "Annals of Applied Biology",
  ["ann appl biosci"] = "Annals of applied bio-sciences",
  ["ann appl math"] = "Annals of Applied Mathematics. Yingyong Shuxue Niankan",
  ["ann appl probab"] = "The Annals of Applied Probability",
  ["ann appl sport sci"] = "Annals of applied sport science",
  ["ann appl stat"] = "The annals of applied statistics",
  ["ann archeol syrie"] = "Annales archéologiques arabes syriennes",
  ["ann arid zone"] = "Annals of arid zone",
  ["ann assoc am geogr"] = "Annals of the Association of American Geographers. Association of American Geographers",
  ["ann aust coll dent surg"] = "Annals of the Australian College of Dental Surgeons",
  ["ann behav med"] = "Annals of behavioral medicine : a publication of the Society of Behavioral Medicine",
  ["ann behav neurosci"] = "Annals of behavioral neuroscience",
  ["ann behav sci med educ"] = "Annals of behavioral science and medical education : journal of the Association for the Behavioral Sciences and Medical Education",
  ["ann belg med mil"] = "Annales belges de médecine militaire. Belgisch tijdschrift voor militaire geneeskunde",
  ["ann biochem exp med"] = "Annals of biochemistry and experimental medicine",
  ["ann biochim clin que"] = "Annales de Biochimie Clinque du Quebec",
  ["ann bioeth"] = "Annals of bioethics",
  ["ann biol"] = "Annals of biology",
  ["ann biol anim biochim biophys"] = "Annales de biologie animale, biochimie, biophysique",
  ["ann biol clin"] = "Annales de Biologie Clinique",
  ["ann biol clin (paris)"] = "Annales de biologie clinique",
  ["ann biol clin que"] = "Annales de Biologie Clinique du Quebec",
  ["ann biol res"] = "Annals of biological research",
  ["ann biom biostat"] = "Annals of biometrics & biostatistics",
  ["ann biomed eng"] = "Annals of biomedical engineering",
  ["ann blood"] = "Annals of blood",
  ["ann bogor"] = "Annales Bogorienses",
  ["ann bot"] = "Annals of botany",
  ["ann bot (lond)"] = "Annals of botany",
  ["ann bot fenn"] = "Annales botanici Fennici",
  ["ann botan fenn"] = "Annales Botanici Fennici",
  ["ann bourgogne"] = "Annales de Bourgogne",
  ["ann breast cancer ther"] = "Annals of breast cancer and therapy",
  ["ann bretagne payes ouest"] = "Annales de Bretagne et des pays de l'Ouest",
  ["ann burns fire disasters"] = "Annals of burns and fire disasters",
  ["ann can soc hist philos math/soc can hist philos math"] = "Annals of the Canadian Society for History and Philosophy of Mathematics/Societe Canadienne d’Histoire et de Philosophie des Mathematiques",
  ["ann cancer epidemiol"] = "Annals of cancer epidemiology",
  ["ann cancer res"] = "Annals of cancer research",
  ["ann card anaesth"] = "Annals of cardiac anaesthesia",
  ["ann cardiol angeiol (paris)"] = "Annales de cardiologie et d'angéiologie",
  ["ann cardiothorac surg"] = "Annals of cardiothoracic surgery",
  ["ann carnegie mus"] = "Annals of the Carnegie Museum",
  ["ann cerc roy hist archeol ath"] = "Annales du Cercle royal d'histoire et d'archéologie d'Ath et de la région et musées athois",
  ["ann cercle archeol canton soignies"] = "Annales du Cercle archéologique du canton de Soignies",
  ["ann cercle archeol enghien"] = "Annales du Cercle archéologique d'Enghien",
  ["ann chim"] = "Annali di chimica",
  ["ann chim (cachan, fr)"] = "Annales de Chimie (Cachan, France)",
  ["ann chim (paris)"] = "Annales de chimie",
  ["ann chim (paris, fr)"] = "Annales de Chimie (Paris, France)",
  ["ann chim (rome)"] = "Annali di Chimica",
  ["ann chim (rome, italy)"] = "Annali di Chimica (Rome, Italy)",
  ["ann chim - sci mater"] = "Annales de Chimie - Science des Materiaux",
  ["ann chim anal"] = "Annals Chimie Analytique",
  ["ann chim anal rev chim anal reunies"] = "Annales de Chimie Analytique et Revue de Chimie Analytique Reunies",
  ["ann chim phys"] = "Annales de Chimie et de Physique",
  ["ann chim sci mat"] = "Annales de Chimie Science des Materiaux",
  ["ann chimie sci materiaux"] = "Annales de chimie--science des matériaux",
  ["ann chin hist soc pac northwest"] = "The Annals of the Chinese Historical Society of the Pacific Northwest",
  ["ann chir"] = "Annales de chirurgie",
  ["ann chir gynaecol"] = "Annales chirurgiae et gynaecologiae",
  ["ann chir gynaecol fenn"] = "Annales chirurgiae et gynaecologiae Fenniae",
  ["ann chir gynaecol fenn suppl"] = "Annales chirurgiae et gynaecologiae Fenniae. Supplementum",
  ["ann chir gynaecol suppl"] = "Annales chirurgiae et gynaecologiae. Supplementum",
  ["ann chir infant"] = "Annales de chirurgie infantile",
  ["ann chir main"] = "Annales de chirurgie de la main : organe officiel des sociétés de chirurgie de la main",
  ["ann chir main memb super"] = "Annales de chirurgie de la main et du membre supérieur : organe officiel des sociétés de chirurgie de la main = Annals of hand and upper limb surgery",
  ["ann chir plast"] = "Annales de chirurgie plastique",
  ["ann chir plast esthet"] = "Annales de chirurgie plastique et esthétique",
  ["ann chir thorac cardiovasc"] = "Annales de chirurgie thoracique et cardio-vasculaire",
  ["ann cisalp hist soc"] = "Annales cisalpines d'histoire sociale. Série I",
  ["ann clin biochem"] = "Annals of clinical biochemistry",
  ["ann clin case rep"] = "Annals of clinical case reports",
  ["ann clin cytol pathol"] = "Annals of clinical cytology and pathology",
  ["ann clin exp hypertens"] = "Annals of clinical and experimental hypertension",
  ["ann clin lab res"] = "Annals of clinical and laboratory research",
  ["ann clin lab sci"] = "Annals of clinical and laboratory science",
  ["ann clin microbiol antimicrob"] = "Annals of clinical microbiology and antimicrobials",
  ["ann clin pathol"] = "Annals of clinical pathology",
  ["ann clin psychiatry"] = "Annals of clinical psychiatry : official journal of the American Academy of Clinical Psychiatrists",
  ["ann clin res"] = "Annals of clinical research",
  ["ann clin transl neurol"] = "Annals of clinical and translational neurology",
  ["ann coloproctol"] = "Annals of coloproctology",
  ["ann comb"] = "Annals of Combinatorics",
  ["ann compos"] = "Annales des Composites",
  ["ann data sci"] = "Annals of Data Science",
  ["ann demogr hist (paris)"] = "Annales de démographie historique",
  ["ann dent"] = "Annals of dentistry",
  ["ann dent oral disord"] = "Annals of dentistry and oral disorders",
  ["ann depress anxiety"] = "Annals of depression and anxiety",
  ["ann dermatol"] = "Annals of dermatology",
  ["ann dermatol syphiligr (paris)"] = "Annales de dermatologie et de syphiligraphie",
  ["ann dermatol venereol"] = "Annales de dermatologie et de vénéréologie",
  ["ann diagn pathol"] = "Annals of diagnostic pathology",
  ["ann differential equations"] = "Annals of Differential Equations",
  ["ann discrete math"] = "Annals of Discrete Mathematics",
  ["ann droit int med"] = "Annales de droit international médical",
  ["ann dyslexia"] = "Annals of dyslexia",
  ["ann earth"] = "Annals of earth : a publication of Ocean Arks International and the Lindisfarne Association",
  ["ann econ finance"] = "Annals of Economics and Finance",
  ["ann econ soc civilis"] = "Annales (Paris, France : 1946)",
  ["ann econ soc meas"] = "Annals of Economic and Social Measurement",
  ["ann econ stat"] = "Annals of economics and statistics",
  ["ann econ stat (1986)"] = "Annales d'économie et de statistique",
  ["ann econ statist"] = "Annales d’Economie et de Statistique",
  ["ann emerg dispatch response"] = "Annals of emergency dispatch & response",
  ["ann emerg med"] = "Annals of emergency medicine",
  ["ann endocrinol (paris)"] = "Annales d'endocrinologie",
  ["ann entomol fenn"] = "Annales Entomologici Fennici",
  ["ann entomol soc am"] = "Annals of the Entomological Society of America",
  ["ann epidemiol"] = "Annals of epidemiology",
  ["ann epidemiol public health"] = "Annals of epidemiology and public health",
  ["ann est"] = "Annales de l'Est",
  ["ann etud int"] = "Annales d'études internationales. Annals of international studies",
  ["ann eugen"] = "Annals of eugenics",
  ["ann eye sci"] = "Annals of eye science",
  ["ann fac econ com"] = "Annali della facoltà di Economia e Commercio, Università Palermo",
  ["ann fac econ commer"] = "Annali della Facoltà di economia e commercio (Università di Messina)",
  ["ann fac lett sci hum"] = "Annales de la Faculté des lettres et sciences humaines (Dakar, Senegal)",
  ["ann fac lett sci hum nice"] = "Annales. Faculté des lettres et sciences humaines de Nice",
  ["ann fac med chir univ studi perugia"] = "Annali della Facoltà di medicina e chirurgia della Università degli studi di Perugia",
  ["ann fac sci toulouse math (6)"] = "Annales de la Faculte des Sciences de Toulouse. Mathematiques. Serie 6",
  ["ann fac sci univ kinshasa"] = "Annales de la Faculté des Sciences",
  ["ann falsif fraud"] = "Annales des falsifications et des fraudes",
  ["ann fam med"] = "Annals of family medicine",
  ["ann fenn math"] = "Annales Fennici Mathematici",
  ["ann finance"] = "Annals of Finance",
  ["ann fluid dyn"] = "Annals of Fluid Dynamics",
  ["ann fond louis de broglie"] = "Fondation Louis de Broglie. Annales",
  ["ann fond luigi einaudi"] = "Annali della Fondazione Luigi Einaudi. Fondazione Luigi Einaudi",
  ["ann for res"] = "Annals of forest research",
  ["ann for sci"] = "Annals of forest science",
  ["ann forensic res anal"] = "Annals of forensic research and analysis",
  ["ann fr anesth reanim"] = "Annales françaises d'anesthèsie et de rèanimation",
  ["ann funct anal"] = "Annals of Functional Analysis",
  ["ann fuzzy math inform"] = "Annals of Fuzzy Mathematics and Informatics",
  ["ann gastroenterol"] = "Annals of gastroenterology",
  ["ann gastroenterol dig disord"] = "Annals of gastroenterology and digestive disorders",
  ["ann gastroenterol hepatol (paris)"] = "Annales de gastroentérologie et d'hépatologie",
  ["ann gastroenterol surg"] = "Annals of gastroenterological surgery",
  ["ann gastroentol hepatol"] = "Annals of gastroenterology & hepatology",
  ["ann gen hosp psychiatry"] = "Annals of general hospital psychiatry",
  ["ann gen psychiatry"] = "Annals of general psychiatry",
  ["ann genet"] = "Annales de Genetique",
  ["ann genet sel anim"] = "Annales de génétique et de sélection animale",
  ["ann geofis"] = "Annali di Geofisica",
  ["ann geophys"] = "Annales Geophysicae",
  ["ann georgr"] = "Annales de géographie",
  ["ann geriatr med res"] = "Annals of geriatric medicine and research",
  ["ann gerontol geriatr res"] = "Annals of gerontology and geriatric research",
  ["ann gis"] = "Annals of GIS",
  ["ann glaciol"] = "Annals of Glaciology",
  ["ann glob anal geom (dordr)"] = "Annals of global analysis and geometry",
  ["ann glob health"] = "Annals of global health",
  ["ann global anal geom"] = "Annals of Global Analysis and Geometry",
  ["ann h lebesgue"] = "Annales Henri Lebesgue",
  ["ann health law"] = "Annals of health law",
  ["ann health law life sci"] = "Annals of health law and life sciences",
  ["ann hematol"] = "Annals of hematology",
  ["ann hematol oncol"] = "Annals of hematology & oncology",
  ["ann henri poincare"] = "Annales Henri Poincaré",
  ["ann henri poincaré"] = "Annales Henri Poincaré",
  ["ann hepatobiliary pancreat surg"] = "Annals of hepato-biliary-pancreatic surgery",
  ["ann hepatol"] = "Annals of hepatology",
  ["ann hist comput"] = "Annals of the history of computing",
  ["ann hist mantois"] = "Annales historiques du Mantois",
  ["ann hist revolut fr"] = "Annales historiques de la Révolution française",
  ["ann hist-nat mus natl hung"] = "Annales Historico-Naturales Musei Nationalis Hungarici",
  ["ann histochim"] = "Annales d'histochimie",
  ["ann homeopath fr"] = "Les Annales homéopathiques françaises",
  ["ann hum biol"] = "Annals of human biology",
  ["ann hum genet"] = "Annals of human genetics",
  ["ann hydrogr"] = "Annales Hydrographiques",
  ["ann hydrogr marit meteorol"] = "Annalen der Hydrographie und maritimen Meteorologie",
  ["ann hyg"] = "Annals of hygiene",
  ["ann hyg med colon"] = "Annales d'hygiène et de médecine coloniales",
  ["ann hyg publique ind soc (1923)"] = "Annales d'hygiène publique, industrielle et sociale",
  ["ann i h p probab stat"] = "Annales de l'I.H.P. Probabilités et statistiques",
  ["ann ib postgrad med"] = "Annals of Ibadan postgraduate medicine",
  ["ann icrp"] = "Annals of the ICRP",
  ["ann iford"] = "Les Annales de l'I.F.O.R.D",
  ["ann ig"] = "Annali di igiene : medicina preventiva e di comunità",
  ["ann ig (roma)"] = "Annali d'igiene",
  ["ann immunol"] = "Annals of immunology",
  ["ann immunol (paris)"] = "Annales d'immunologie",
  ["ann immunol hung"] = "Annales immunologiae Hungaricae",
  ["ann indian acad neurol"] = "Annals of Indian Academy of Neurology",
  ["ann infect"] = "Annals of infection",
  ["ann insee"] = "Annales de l’INSEE",
  ["ann inst archeol luxemb"] = "Annales de l'Institut archéologique du Luxembourg. Institut archéologique du Luxembourg (Arlon, Belgium)",
  ["ann inst fourier"] = "Annales de le Institue Fourier",
  ["ann inst fourier (grenoble)"] = "Universite de Grenoble. Annales de l’Institut Fourier",
  ["ann inst h poincare anal non lineaire"] = "Annales de l’Institut Henri Poincare. Analyse Non Lineaire",
  ["ann inst h poincare c anal non lineaire"] = "Annales de l’Institut Henri Poincare C. Analyse Non Lineaire",
  ["ann inst h poincaré anal non linéaire"] = "Annales de l’Institut Henri Poincaré",
  ["ann inst h poincaré phys théor"] = "Annales de l’Institut Henri Poincaré",
  ["ann inst h poincaré probab statist"] = "Annales de l’Institut Henri Poincaré",
  ["ann inst henri poincare"] = "Annales de l’Institute Henri Poincare",
  ["ann inst henri poincare c"] = "Annales de l’Institute Henri Poincare C",
  ["ann inst henri poincare d"] = "Annales de l’Institute Henri Poincare D",
  ["ann inst henri poincare probab stat"] = "Annales de l’Institut Henri Poincare Probabilites et Statistiques",
  ["ann inst henri poincare, probab stat"] = "Annales de l’Institute Henri Poincare, Probabilites et Statistiques",
  ["ann inst henri poincaré"] = "Annales de l’Institut Henri Poincaré",
  ["ann inst henri poincaré, a"] = "Annales de l’Institut Henri Poincaré, Section A: Physique Theorique",
  ["ann inst henri poincaré, b"] = "Annales de l’Institut Henri Poincaré, Section B: Calculdes Probabilités et Statistique",
  ["ann inst hydrol climatol"] = "Annales de l'Institut d'hydrologie et de climatologie",
  ["ann inst pasteur (paris)"] = "Annales de l'Institut Pasteur",
  ["ann inst pasteur actual"] = "Annales de l'Institut Pasteur. Actualités",
  ["ann inst pasteur imm"] = "Annales de'l Institut Pasteur-immunology",
  ["ann inst pasteur immunol"] = "Annales de l'Institut Pasteur. Immunology",
  ["ann inst pasteur immunol (1985)"] = "Annales de l'Institut Pasteur. Immunologie",
  ["ann inst pasteur lille"] = "Annales de l'Institut Pasteur de Lille",
  ["ann inst pasteur mic"] = "Annales de'l Institut Pasteur-microbiologie",
  ["ann inst pasteur microbiol"] = "Annales de l'Institut Pasteur. Microbiology",
  ["ann inst pasteur microbiol (1985)"] = "Annales de l'Institut Pasteur. Microbiologie",
  ["ann inst pasteur vir"] = "Annales De L Institut Pasteur-virology",
  ["ann inst pasteur virol"] = "Annales de l'Institut Pasteur. Virology",
  ["ann inst stat math"] = "Annals of the Institute of Statistical Mathematics",
  ["ann inst statist math"] = "Annals of the Institute of Statistical Mathematics",
  ["ann int commun assoc"] = "Annals of the International Communication Association",
  ["ann int soc dynam games"] = "Annals of the International Society of Dynamic Games",
  ["ann integr mol med"] = "Annals of integrative molecular medicine",
  ["ann intensive care"] = "Annals of intensive care",
  ["ann intern med"] = "Annals of internal medicine",
  ["ann internat soc dynam games"] = "Annals of the International Society of Dynamic Games",
  ["ann iowa"] = "Annals of Iowa",
  ["ann isnardi auxol norm patol"] = "Annali Isnardi di auxologia normale e patologica",
  ["ann israel phys soc"] = "Annals of the Israel Physical Society",
  ["ann ist carlo forlanini"] = "Annali dell’Istituto Carlo Forlanini",
  ["ann ist giangiacomo feltrinelli"] = "Annali - Istituto Giangiacomo Feltrinelli",
  ["ann ist mus stor sci fir"] = "Annali dell'Istituto e museo di storia della scienza di Firenze",
  ["ann ist orient napoli"] = "Annali",
  ["ann ist sper assest for alpic"] = "Annali dell’Istituto Sperimentale per l’Assestamento Forestale e per l’Alpicoltura",
  ["ann ist sper selvic"] = "Annali dell’Istituto Sperimentale per la Selvicoltura",
  ["ann ist stor italo ger"] = "Annali dell'Istituto storico italo germanico",
  ["ann ist stor italo ger trento"] = "Annali dell'Istituto storico italo-germanico in Trento. Contributi = Jahrbuch des Italienisch-Deutschen Historischen Instituts in Trient. Beiträge",
  ["ann ist super sanita"] = "Annali dell'Istituto superiore di sanità",
  ["ann isup"] = "Annales de l’I.S.U.P.",
  ["ann ital chir"] = "Annali italiani di chirurgia",
  ["ann ital dermatol sifilogr"] = "Annali italiani di dermatologia e sifilografia",
  ["ann ital med int"] = "Annali Italiani di Medicina Interna",
  ["ann ital pediatr"] = "Annali italiani di pediatria",
  ["ann japan assoc philos sci"] = "Annals of the Japan Association for Philosophy of Science",
  ["ann jt"] = "Annals of joint",
  ["ann k-theory"] = "Annals of K-Theory",
  ["ann king edw med univ"] = "Annals of King Edward Medical University",
  ["ann lab med"] = "Annals of laboratory medicine",
  ["ann laparosc endosc surg"] = "Annals of laparoscopic and endoscopic surgery",
  ["ann laringol otol rinol faringol"] = "Annali di laringologia, otologia, rinologia, faringologia",
  ["ann lgbtq public popul health"] = "Annals of LGBTQ public and population health",
  ["ann limnol"] = "Annales de limnologie",
  ["ann longterm care"] = "The annals of long-term care : the official journal of the American Medical Directors Association",
  ["ann lung cancer"] = "Annals of lung cancer",
  ["ann lymphoma"] = "Annals of lymphoma",
  ["ann mat pura appl"] = "Annali di matematica pura ed applicata",
  ["ann mat pura appl (4)"] = "Annali di Matematica Pura ed Applicata. Series IV",
  ["ann math"] = "Annals of Mathematics",
  ["ann math artif intell"] = "Annals of Mathematics and Artificial Intelligence",
  ["ann math blaise pascal"] = "Annales Mathematiques Blaise Pascal",
  ["ann math inform"] = "Annales Mathematicae et Informaticae",
  ["ann math log"] = "Annals of Mathematical Logic",
  ["ann math que"] = "Annales Mathematiques du Quebec",
  ["ann math sci appl"] = "Annals of Mathematical Sciences and Applications",
  ["ann math sil"] = "Annales Mathematicae Silesianae",
  ["ann math stat"] = "Annals of Mathematical Statistics",
  ["ann maxillofac surg"] = "Annals of maxillofacial surgery",
  ["ann med"] = "Annals of medicine",
  ["ann med (milano)"] = "Annali medici",
  ["ann med chir hop sainte justine montr"] = "Les Annales médico-chirurgicales de l'Hôpital Sainte-Justine de Montréal",
  ["ann med exp biol fenn"] = "Annales medicinae experimentalis et biologiae Fenniae",
  ["ann med health sci res"] = "Annals of medical and health sciences research",
  ["ann med hist"] = "Annals of medical history",
  ["ann med intern fenn"] = "Annales medicinae internae Fenniae",
  ["ann med intern fenn suppl"] = "Annales medicinae internae Fenniae. Supplementum",
  ["ann med interne (paris)"] = "Annales de Medecine Interne",
  ["ann med leg criminol police sci toxicol"] = "Annales de médecine légale, criminologie, police scientifique et toxicologie",
  ["ann med nancy"] = "Annales médicales de Nancy",
  ["ann med nav (roma)"] = "Annali di medicina navale",
  ["ann med nav trop"] = "Annali di medicina navale e tropicale",
  ["ann med phys (lille)"] = "Annales de médecine physique",
  ["ann med psychol (paris)"] = "Annales médico-psychologiques",
  ["ann med res"] = "Annals of medical research",
  ["ann med sect pol acad sci"] = "Annals of the Medical Section of the Polish Academy of Sciences",
  ["ann med soc"] = "Les Annales de médecine sociale",
  ["ann med sondalo"] = "Annali medici di Sondalo",
  ["ann med surg (lond)"] = "Annals of medicine and surgery (2012)",
  ["ann med univ bialyst pol"] = "Annals of the Medical University, Białystok, Poland = Roczniki Akademii Medycznej w Białymstoku",
  ["ann med vet"] = "Annales de médecine vétérinaire",
  ["ann meteor"] = "Annalen der Meteorologie",
  ["ann microbiol"] = "Annals of microbiology",
  ["ann microbiol (paris)"] = "Annales de microbiologie",
  ["ann microbiol enzimol"] = "Annali di microbiologia ed enzimologia",
  ["ann midi"] = "Annales du Midi",
  ["ann mo bot gard"] = "Annals of the Missouri Botanical Garden. Missouri Botanical Garden",
  ["ann musculoskelet med"] = "Annals of musculoskeletal medicine",
  ["ann n y acad sci"] = "Annals of the New York Academy of Sciences",
  ["ann nat hist mus wien ser b bot zool"] = "Annalen des Naturhistorischen Museums in Wien. Serie B, Für Botanik und Zoologie",
  ["ann natl acad med sci"] = "Annals of the National Academy of Medical Sciences (India)",
  ["ann natl assoc geogr"] = "Annals of the National Association of Geographers, India",
  ["ann naturhist mus wien ser a mineral petrogr geol palaontol anthropol prahist"] = "Annalen des Naturhistorischen Museums in Wien. Serie A, Für Mineralogie und Petrographie, Geologie und Paläontologie, Anthropologie und Prähistorie",
  ["ann nestle [eng]"] = "Annales Nestle [English ed.]",
  ["ann nestle [fr]"] = "Annales Nestle",
  ["ann nestle [ger]"] = "Annales Nestle. [Deutsche Aufl.]",
  ["ann nestle eng"] = "Annales Nestlé [English ed.]",
  ["ann nestle fr"] = "Annales Nestlé. [Ed. française]",
  ["ann nestle ger"] = "Annales Nestlé. [Deutsche Aufl.]",
  ["ann neurol"] = "Annals of neurology",
  ["ann neuropsichiatr psicoanal"] = "Annali di neuropsichiatria e psicoanalisi",
  ["ann neurosci"] = "Annals of neurosciences",
  ["ann neurosci psychol"] = "Annals of neuroscience and psychology",
  ["ann new york acad sci"] = "Annals of the New York Academy of Sciences",
  ["ann noninvasive electrocardiol"] = "Annals of noninvasive electrocardiology : the official journal of the International Society for Holter and Noninvasive Electrocardiology, Inc",
  ["ann normandie"] = "Annales de Normandie",
  ["ann nucl cardiol"] = "Annals of nuclear cardiology",
  ["ann nucl energy"] = "Annals of Nuclear Energy",
  ["ann nucl eng"] = "Annals of Nuclear Engineering",
  ["ann nucl med"] = "Annals of nuclear medicine",
  ["ann nucl sci eng"] = "Annals of Nuclear Science and Engineering",
  ["ann numer math"] = "Annals of Numerical Mathematics",
  ["ann nutr aliment"] = "Annales de la nutrition et de l'alimentation",
  ["ann nutr metab"] = "Annals of nutrition & metabolism",
  ["ann ny acad sci"] = "Annals of the New York Academy of Sciences",
  ["ann occup environ med"] = "Annals of occupational and environmental medicine",
  ["ann occup hyg"] = "Annals of Occupational Hygiene",
  ["ann ocul (paris)"] = "Annales d'oculistique",
  ["ann odontostomatol (lyon)"] = "Annales odonto-stomatologiques",
  ["ann of math (2)"] = "Annals of Mathematics. Second Series",
  ["ann of math stud"] = "Annals of Mathematics Studies",
  ["ann of sci"] = "Annals of Science",
  ["ann oncol"] = "Annals of oncology : official journal of the European Society for Medical Oncology",
  ["ann oper res"] = "Annals of Operations Research",
  ["ann ophthalmol"] = "Annals of ophthalmology",
  ["ann ophthalmol (skokie)"] = "Annals of ophthalmology (Skokie, Ill.)",
  ["ann orgone inst"] = "Annals. Orgone Institute",
  ["ann osp maria vittoria torino"] = "Annali dell'Ospedale Maria Vittoria di Torino",
  ["ann ostet ginecol"] = "Annali di ostetricia e ginecologia",
  ["ann ostet ginecol med perinat"] = "Annali di ostetricia, ginecologia, medicina perinatale",
  ["ann otol rhinol laryngol"] = "The Annals of otology, rhinology, and laryngology",
  ["ann otol rhinol laryngol suppl"] = "The Annals of otology, rhinology & laryngology. Supplement",
  ["ann otolaryngol"] = "Les Annales d'oto-laryngologie",
  ["ann otolaryngol chir cervicofac"] = "Annales d'oto-laryngologie et de chirurgie cervico faciale : bulletin de la Société d'oto-laryngologie des hôpitaux de Paris",
  ["ann ottalmol clin ocul"] = "Annali di ottalmologia e clinica oculistica",
  ["ann paediatr"] = "Annales paediatrici. International review of pediatrics",
  ["ann paediatr fenn"] = "Annales paediatriae Fenniae",
  ["ann paediatr int rev pediatr"] = "Annales Paediatrici (International Review of Pediatrics)",
  ["ann paediatr rheumatol"] = "Annals of paediatric rheumatology",
  ["ann paleontol"] = "Annales de Paleontologie",
  ["ann palliat med"] = "Annals of palliative medicine",
  ["ann pancreat cancer"] = "Annals of pancreatic cancer",
  ["ann parasitol"] = "Annals of parasitology",
  ["ann parasitol hum comp"] = "Annales de parasitologie humaine et comparée",
  ["ann pathol"] = "Annales de pathologie",
  ["ann pde"] = "Annals of PDE",
  ["ann pediatr (paris)"] = "Annales de pédiatrie",
  ["ann pediatr cardiol"] = "Annals of pediatric cardiology",
  ["ann pediatr child health"] = "Annals of pediatrics & child health",
  ["ann pediatr endocrinol metab"] = "Annals of pediatric endocrinology & metabolism",
  ["ann pediatr res"] = "Annals of pediatric research",
  ["ann periodontol"] = "Annals of periodontology",
  ["ann pharm"] = "Annalen der Pharmacie",
  ["ann pharm fr"] = "Annales Pharmaceutiques Francaises",
  ["ann pharmacol pharm"] = "Annals of pharmacology and pharmaceutics",
  ["ann pharmacother"] = "The Annals of pharmacotherapy",
  ["ann phys"] = "Annalen der Physik",
  ["ann phys (amsterdam, neth)"] = "Annals of Physics (Amsterdam, Netherlands)",
  ["ann phys (berl)"] = "Annalen der Physik",
  ["ann phys (berlin)"] = "Annalen der Physik",
  ["ann phys (leipzig)"] = "Annalen der Physik (Leipzig)",
  ["ann phys (n y)"] = "Annals of physics",
  ["ann phys (ny)"] = "Annals of Physics (New York)",
  ["ann phys (oxford, u k)"] = "Annals of Physics (Oxford, United Kingdom)",
  ["ann phys (paris)"] = "Annales de physique",
  ["ann phys (san diego, ca, u s)"] = "Annals of Physics (San Diego, CA, United States)",
  ["ann phys chem"] = "Annual Review of Physical Chemistry",
  ["ann phys med"] = "Annals of physical medicine",
  ["ann phys rehabil med"] = "Annals of physical and rehabilitation medicine",
  ["ann physics"] = "Annals of Physics",
  ["ann physik (8)"] = "Annalen der Physik (8)",
  ["ann physiol anthropol"] = "The Annals of physiological anthropology = Seiri Jinruigaku Kenkyūkai kaishi",
  ["ann phytomed"] = "Annals of phytomedicine",
  ["ann plant prot sci"] = "Annals of Plant Protection Sciences",
  ["ann plast surg"] = "Annals of plastic surgery",
  ["ann polon math"] = "Annales Polonici Mathematici",
  ["ann ponts chaussees"] = "Annales des Ponts et Chausses",
  ["ann probab"] = "The Annals of Probability",
  ["ann psychiatry ment health"] = "Annals of psychiatry and mental health",
  ["ann public coop econ"] = "Annals of Public and Cooperative Economics",
  ["ann public health res"] = "Annals of public health and research",
  ["ann pulm crit care med"] = "Annals of pulmonary and critical care medicine",
  ["ann pure appl log"] = "Annals of pure and applied logic",
  ["ann pure appl logic"] = "Annals of Pure and Applied Logic",
  ["ann r australas coll dent surg"] = "Annals of the Royal Australasian College of Dental Surgeons",
  ["ann r coll physicians surg can"] = "Annals (Royal College of Physicians and Surgeons of Canada)",
  ["ann r coll surg engl"] = "Annals of the Royal College of Surgeons of England",
  ["ann radiat ther oncol"] = "Annals of radiation therapy and oncology",
  ["ann radiol (paris)"] = "Annales de radiologie",
  ["ann radiol diagn (bologna)"] = "Annali di radiologia diagnostica",
  ["ann readapt med phys"] = "Annales de réadaptation et de médecine physique : revue scientifique de la Société française de rééducation fonctionnelle de réadaptation et de médecine physique",
  ["ann rech vet"] = "Annales de recherches vétérinaires. Annals of veterinary research",
  ["ann reg sci"] = "The Annals of regional science",
  ["ann rehabil med"] = "Annals of rehabilitation medicine",
  ["ann represent theory"] = "Annals of Representation Theory",
  ["ann res hosp"] = "Annals of research hospitals",
  ["ann rev automat programming"] = "Annual Review in Automatic Programming",
  ["ann rev cell dev biol"] = "Annual Review of Cell and Developmental Biology",
  ["ann rev mar sci"] = "Annual review of marine science",
  ["ann rev nuclear particle sci"] = "Annual Review of Nuclear and Particle Science",
  ["ann rev res"] = "Annals of reviews and research",
  ["ann rheum dis"] = "Annals of the rheumatic diseases",
  ["ann rom soc cell biol"] = "Annals of the Romanian society for cell biology",
  ["ann sanita pubblica"] = "Annali della sanità pubblica",
  ["ann saudi med"] = "Annals of Saudi medicine",
  ["ann sc norm super pisa"] = "Annali della Scuola normale superiore di Pisa. Classe di lettere e filosofia. Scuola normale superiore (Italy)",
  ["ann sc norm super pisa cl sci"] = "Annali della Scuola Normale Superiore di Pisa Classe di Scienze",
  ["ann sc norm super pisa cl sci (5)"] = "Annali della Scuola Normale Superiore di Pisa. Classe di Scienze. Serie V",
  ["ann schweiz meteorol anst"] = "Annalen der Schweizerischen meteorologischen Anstalt",
  ["ann sci"] = "Annals of science",
  ["ann sci ec norm super"] = "Annales Scientifiques de le Ecole Normale Superieure",
  ["ann sci ec norm super (4)"] = "Annales Scientifiques de l’Ecole Normale Superieure. Quatrieme Serie",
  ["ann sci econ appl"] = "Annales de Sciences Economiques Appliqués",
  ["ann sci for"] = "Annales des sciences forestières",
  ["ann sci kanazawa univ"] = "Kanazawa University",
  ["ann sci math québec"] = "Annales des Sciences Mathématiques du Québec",
  ["ann sci nat zool biol anim"] = "Annales des Sciences Naturelles - Zoologie et Biologie Animale",
  ["ann sci technol policy"] = "Annals of Science and Technology Policy",
  ["ann sci univ besancon med"] = "Annales scientifiques de l'Université de Besançon. Médecine",
  ["ann sci école norm sup (4)"] = "Annales Scientifiques de l’École Normale Supérieure",
  ["ann sclavo"] = "Annali Sclavo; rivista di microbiologia e di immunologia",
  ["ann sclavo collana monogr"] = "Annali Sclavo. Collana monografica",
  ["ann scuola norm sup pisa cl sci (4)"] = "Annali della Scuola Normale Superiore di Pisa",
  ["ann serv antiq egypte"] = "Annales du Service des antiquités de l'Egypte",
  ["ann soc angeiol histopathologie"] = "Annales. Société française d'angéiologie et d'histopathologie",
  ["ann soc belg hist hop"] = "Annales de la Société belge d'histoire des hôpitaux",
  ["ann soc belg med trop"] = "Annales de la Société belge de médecine tropicale",
  ["ann soc belg med trop (1920)"] = "Annales de la Société belge de médecine tropicale (1920)",
  ["ann soc belges med trop parasitol mycol"] = "Annales des Sociétés belges de médecine tropicale, de parasitologie, et de mycologie",
  ["ann soc emul bruges"] = "Annales de la Société d'émulation de Bruges",
  ["ann soc entomol fr"] = "Annales de la Société entomologique de France. Société entomologique de France",
  ["ann soc entomol france"] = "Annales de la Societe Entomologique de France",
  ["ann soc r archeol brux"] = "Annales de la Société royale d'archéologie de Bruxelles. Société royale d'archéologie de Bruxelles (Belgium)",
  ["ann soc r sci med nat brux"] = "Annales de la Société royale des sciences médicales et naturelles de Bruxelles",
  ["ann soc sci counc niger"] = "Annals of the Social Science Council of Nigeria",
  ["ann software eng"] = "Annals of Software Engineering",
  ["ann solid struct mech"] = "Annals of Solid and Structural Mechanics",
  ["ann sper agrar"] = "Annali della Sperimentazione Agraria",
  ["ann sports med res"] = "Annals of sports medicine and research",
  ["ann stat"] = "Annals of statistics",
  ["ann statist"] = "The Annals of Statistics",
  ["ann stem cell res"] = "Annals of stem cell research",
  ["ann stem cell res ther"] = "Annals of stem cell research & therapy",
  ["ann stomatol (roma)"] = "Annali di stomatologia",
  ["ann surg"] = "Annals of surgery",
  ["ann surg innov res"] = "Annals of surgical innovation and research",
  ["ann surg oncol"] = "Annals of surgical oncology",
  ["ann surg open"] = "Annals of surgery open : perspectives of surgical history, education, and clinical approaches",
  ["ann surg treat res"] = "Annals of surgical treatment and research",
  ["ann telecommun"] = "Annals of Telecommunications",
  ["ann thorac cardiovasc surg"] = "Annals of thoracic and cardiovascular surgery : official journal of the Association of Thoracic and Cardiovascular Surgeons of Asia",
  ["ann thorac med"] = "Annals of thoracic medicine",
  ["ann thorac surg"] = "The Annals of thoracic surgery",
  ["ann thyroid"] = "Annals of thyroid",
  ["ann thyroid res"] = "Annals of thyroid research",
  ["ann tiberiu popoviciu semin funct equ approx convexity"] = "Annals of the Tiberiu Popoviciu Seminar of Functional Equations, Approximation and Convexity",
  ["ann tour res"] = "Annals of tourism research",
  ["ann touris res"] = "Annals Of Tourism Research",
  ["ann toxicol anal"] = "Annales de Toxicologie Analytique",
  ["ann transl med"] = "Annals of translational medicine",
  ["ann transl med epidemiol"] = "Annals of translational medicine & epidemiology",
  ["ann transplant"] = "Annals of transplantation",
  ["ann trauma acute care"] = "Annals of trauma and acute care",
  ["ann trop med parasitol"] = "Annals of tropical medicine and parasitology",
  ["ann trop med public health"] = "Annals of tropical medicine and public health",
  ["ann trop paediatr"] = "Annals of tropical paediatrics",
  ["ann tuberc"] = "Annals of tuberculosis",
  ["ann univ abidjan ser g"] = "Annales de l'Université d'Abidjan. Série G: Géographie. Université d'Abidjan",
  ["ann univ ankara"] = "Ankara Üniversitesi Yilliği. Annales de l'Université d'Ankara. Ankara Üniversitesi",
  ["ann univ buc chim"] = "Analele Universitatii Bucurersti. Chimie",
  ["ann univ buchar math ser"] = "Annals of the University of Bucharest. Mathematical Series",
  ["ann univ ferrara sez vii (ns)"] = "Annali dell’Università di Ferrara",
  ["ann univ ferrara sez vii sci mat"] = "Annali dell’Universita di Ferrara. Sezione VII. Scienze Matematiche",
  ["ann univ mariae curie sklodowska [med]"] = "Annales Universitatis Mariae Curie-Sklodowska. Sectio D, Medicina",
  ["ann univ mariae curie sklodowska med"] = "Annales Universitatis Mariae Curie-Skłodowska. Sectio D: Medicina",
  ["ann univ mariae curie sklodowska sect ee zootech"] = "Annales Universitatis Mariae Curie-Skłodowska. Section EE, Zootechnica",
  ["ann univ mariae curie-sklodowska sect a"] = "Annales. Universitatis Mariae Curie-Sklodowska. Sectio A. Mathematica",
  ["ann univ mariae curie-sklodowska sect ai-inform"] = "Annales Universitatis Mariae Curie-Sklodowska. Sectio AI. Informatica",
  ["ann univ mariae curie-skłodowska sect a"] = "Annales Universitatis Mariae Curie-Skłodowska",
  ["ann univ paedagog crac stud math"] = "Annales Universitatis Paedagogicae Cracoviensis. Studia Mathematica",
  ["ann univ sarav [med]"] = "Annales Universitatis Saraviensis. Medizin",
  ["ann univ sarav med"] = "Annales Universitatis Saraviensis. Medizin",
  ["ann univ sarav ser math"] = "Annales Universitatis Saraviensis",
  ["ann univ sci budap rolando eotvos nom sect geogr"] = "Annales Universitatis Scientiarum Budapestinensis de Rolando Eötvös Nominatae. Sectio geographica. Eötvös Loránd Tudományegyetem",
  ["ann univ sci budap rolando eotvos nominatae [biol]"] = "Annales Universitatis Scientiarum Budapestinensis de Rolando Eotvos Nominatae. Sectio Biologica",
  ["ann univ sci budap rolando eotvos nominatae biol"] = "Annales Universitatis Scientiarum Budapestinensis de Rolando Eötvös Nominatae. Sectio Biologica",
  ["ann univ sci budapest eotvos sect math"] = "Annales Universitatis Scientiarum Budapestinensis de Rolando Eotvos Nominatae. Sectio Mathematica",
  ["ann univ sci budapest eötvös sect math"] = "Annales Universitatis Scientiarum Budapestinensis de Rolando Eötvös Nominatae",
  ["ann univ sci budapest sect comput"] = "Annales Universitatis Scientiarum Budapestinensis de Rolando Eotvos Nominatae. Sectio Computatorica",
  ["ann univ turku ser a i"] = "Annales Universitatis Turkuensis",
  ["ann urol (paris)"] = "Annales d'urologie",
  ["ann vasc dis"] = "Annals of vascular diseases",
  ["ann vasc med res"] = "Annals of vascular medicine and research",
  ["ann vasc surg"] = "Annals of vascular surgery",
  ["ann virol res"] = "Annals of virology and research",
  ["ann west med surg"] = "Annals of western medicine and surgery",
  ["ann work expo health"] = "Annals of work exposures and health",
  ["ann work exposures health"] = "Annals of Work Exposures and Health",
  ["ann wyo"] = "Annals of Wyoming (Cheyenne, Wyo. : 1925)",
  ["ann zool"] = "Annales Zoologici",
  ["ann zool fenn"] = "Annales Zoologici Fennici",
  ["ann zool fennici"] = "Annales zoologici Fennici",
  ["ann éc natl eaux for"] = "Annales de l’École nationale des eaux et forêts et de la station de recherches et expériences forestières",
  ["ann économ statist"] = "Annales d’Économie et de Statistique",
  ["anna j"] = "ANNA journal",
  ["annal droit"] = "Annales de droit",
  ["annal zool"] = "Annales zoologici",
  ["annales (esc)"] = "Annales: économies, sociétés, civilisations",
  ["annales (hss)"] = "Annales: histoire, sciences sociales",
  ["annals n y acad dent"] = "The annals of the New York Academy of Dentistry",
  ["annals of the phytopathological society of japan|nippon shokubutsu byori gakkaiho"] = "Nihon Shokubutsu Byori Gakkaiho",
  ["annee biol"] = "L' Année biologique",
  ["annee endocrinol"] = "L'Année endocrinologique",
  ["annee philologique"] = "L'Année philologique; bibliographie critique et analytique de l'antiquité gréco-latine",
  ["annee psychol"] = "L'année psychologique",
  ["annee soc"] = "L'Année sociale",
  ["annee sociol"] = "L' Année sociologique",
  ["annee ther"] = "L' Année thérapeutique",
  ["annee ther clin ophtalmol"] = "L'Année thérapeutique et clinique en ophtalmologie",
  ["anneli lax new math libr"] = "Anneli Lax New Mathematical Library",
  ["annot code public gen laws md md"] = "The annotated code of the public general laws of Maryland. Health-general. Maryland",
  ["annot laws mass"] = "Annotated laws of Massachusetts : containing all the laws of Massachusetts of a general and permanent nature. Massachusetts",
  ["annot laws mass mass"] = "Annotated laws of Massachusetts : containing all the laws of Massachusetts of a general and permanent nature / completely annotated by the editorial staff of the publisher. Massachusetts",
  ["annot zool jpn"] = "Annotationes zoologicae Japonenses",
  ["annu afr nord"] = "Annuaire de l'Afrique du Nord",
  ["annu bibliogr hist print bool libr"] = "ABHB. Annual bibliography of the history of the printed book and libraries",
  ["annu book astm stand sect 11 water environ technol"] = "Annual book of ASTM standards. Section 11, Water and environmental technology. ASTM Committee E-35 on Pesticides",
  ["annu conf australas soc hiv med"] = "Annual Conference / Australasian Society For Hiv Medicine",
  ["annu conf expo (am soc eng educ)"] = "Annual Conference & Exposition : final program and proceedings. American Society for Engineering Education",
  ["annu conf res med educ"] = "Annual Conference on Research in Medical Education. Conference on Research in Medical Education",
  ["annu int conf ieee eng med biol soc"] = "Annual International Conference of the IEEE Engineering in Medicine and Biology Society. IEEE Engineering in Medicine and Biology Society. Annual International Conference",
  ["annu int conf syiah kuala univ"] = "Annual International Conference Syiah Kuala University (AIC-UNSYIAH)",
  ["annu mediaev"] = "Annuale mediaevale",
  ["annu meet am inst oral biol"] = "Annual meeting - American Institute of Oral Biology",
  ["annu meet amer climatol assoc"] = "Annual Meeting of the American Climatological Association. American Climatological Association. Annual Meeting",
  ["annu meet fla state hort soc"] = "Annual meeting of the Florida State Horticultural Society. Florida State Horticultural Society. Meeting",
  ["annu ornl biomed sci eng cent conf"] = "Annual ORNL Biomedical Science and Engineering Center Conference. ORNL Biomedical Science and Engineering Center Conference",
  ["annu pays ocean indien"] = "Annuaire des pays de l'océan indien",
  ["annu proc assoc adv automot med"] = "Annual proceedings. Association for the Advancement of Automotive Medicine",
  ["annu prog rep univ calif lab nucl radiat biol"] = "Annual progress report - University of California, Laboratory of Nuclear Medicine and Radiation Biology",
  ["annu prog reprod med"] = "Annual progress in reproductive medicine",
  ["annu rep anal at spectrosc"] = "Annual Reports on Analytical Atomic Spectroscopy",
  ["annu rep bean improv coop"] = "Annual report of the Bean Improvement Cooperative. Bean Improvement Cooperative",
  ["annu rep board regents smithson inst"] = "Annual report of the Board of Regents of the Smithsonian Institution. Smithsonian Institution. Board of Regents",
  ["annu rep british guiana"] = "Annual report. British Guiana. Mosquito Control Service",
  ["annu rep comput chem"] = "Annual reports in computational chemistry",
  ["annu rep div biol med res argonne natl lab"] = "Annual report - Division of Biological and Medical Research, Argonne National Laboratory. Argonne National Laboratory. Division of Biological and Medical Research",
  ["annu rep gorgas meml lab rep audit"] = "Annual report of the Gorgas Memorial Laboratory and report of audit. Gorgas Memorial Laboratory",
  ["annu rep la dep public welf"] = "Annual report. Louisiana. Department of Public Welfare",
  ["annu rep libr co phila"] = "The Annual report of the Library Company of Philadelphia. Library Company of Philadelphia",
  ["annu rep med chem"] = "Annual reports in medicinal chemistry",
  ["annu rep n y state  dept health division lab res"] = "Annual report. New York (State). Department of Health. Division of Laboratories and Research",
  ["annu rep nmr spectrosc"] = "Annual reports on NMR spectroscopy",
  ["annu rep prog chem sect a: inorg chem"] = "Annual Reports on the Progress of Chemistry, Section A: Inorganic Chemistry",
  ["annu rep prog chem sect b: org chem"] = "Annual Reports on the Progress of Chemistry, Section B: Organic Chemistry",
  ["annu rep prog chem sect c: phys chem"] = "Annual Reports on the Progress of Chemistry, Section C: Physical Chemistry",
  ["annu rep qld health educ counc honour minist health home aff"] = "The Queensland Health Education Council annual report to the Honourable the Minister for Health. Queensland Health Education Council",
  ["annu rep r i"] = "Annual report. Rhode Island. Department of Public Welfare",
  ["annu rep r soc chem sect a: inorg chem"] = "Annual Reports - Royal Society of Chemistry, Section Section A: Inorganic Chemistry",
  ["annu rep r soc chem sect b org chem"] = "Annual Reports - Royal Society of Chemistry, Section Section B: Organic Chemistry",
  ["annu rep r soc chem sect c phys chem"] = "Annual Reports - Royal Society of Chemistry, Section Section C: Physical Chemistry",
  ["annu rep r soc promot health"] = "Annual report of the Council. Royal Society of Health (Great Britain)",
  ["annu rep res inst environ med nagoya univ"] = "Annual report of the Research Institute of Environmental Medicine, Nagoya University",
  ["annu rep united states natl mus"] = "Annual report. United States National Museum",
  ["annu res br"] = "Annual research briefs. Center for Turbulence Research (U.S.)",
  ["annu res rev biol"] = "Annual research & review in biology",
  ["annu rev anal chem"] = "Annual Review of Analytical Chemistry",
  ["annu rev anal chem (palo alto calif)"] = "Annual review of analytical chemistry (Palo Alto, Calif.)",
  ["annu rev anim biosci"] = "Annual review of animal biosciences",
  ["annu rev anthropol"] = "Annual review of anthropology",
  ["annu rev appl linguist"] = "Annual review of applied linguistics",
  ["annu rev astron astrophys"] = "Annual review of astronomy and astrophysics",
  ["annu rev at mol phys"] = "Annual Review of Atomic and Molecular Physics",
  ["annu rev biochem"] = "Annual Review of Biochemistry",
  ["annu rev biomed data sci"] = "Annual review of biomedical data science",
  ["annu rev biomed eng"] = "Annual review of biomedical engineering",
  ["annu rev biophys"] = "Annual review of biophysics",
  ["annu rev biophys bioeng"] = "Annual review of biophysics and bioengineering",
  ["annu rev biophys biomol struct"] = "Annual Review of Biophysics and Biomolecular Structure",
  ["annu rev biophys biophys chem"] = "Annual review of biophysics and biophysical chemistry",
  ["annu rev cancer biol"] = "Annual review of cancer biology",
  ["annu rev cell biol"] = "Annual review of cell biology",
  ["annu rev cell dev biol"] = "Annual review of cell and developmental biology",
  ["annu rev chem biomol eng"] = "Annual review of chemical and biomolecular engineering",
  ["annu rev clin psychol"] = "Annual review of clinical psychology",
  ["annu rev comput sci"] = "Annual Review of Computer Science",
  ["annu rev condens matter phys"] = "Annual Review of Condensed Matter Physics",
  ["annu rev control"] = "Annual Review in Control",
  ["annu rev control rob auton syst"] = "Annual Review of Control, Robotics, and Autonomous Systems",
  ["annu rev control robot auton syst"] = "Annual review of control, robotics, and autonomous systems",
  ["annu rev criminol"] = "Annual review of criminology",
  ["annu rev dev psychol"] = "Annual review of developmental psychology",
  ["annu rev earth planet sci"] = "Annual Review of Earth and Planetary Sciences",
  ["annu rev earth planet sci lett"] = "Annual Review of Earth and Planetary Science Letters",
  ["annu rev ecol evol syst"] = "Annual Review of Ecology, Evolution, and Systematics",
  ["annu rev ecol syst"] = "Annual review of ecology and systematics",
  ["annu rev econom"] = "Annual review of economics",
  ["annu rev energy env"] = "Annual Review of Energy and the Environment",
  ["annu rev entomol"] = "Annual review of entomology",
  ["annu rev environ resour"] = "Annual review of environment and resources",
  ["annu rev fish dis"] = "Annual review of fish diseases",
  ["annu rev fluid mech"] = "Annual Review of Fluid Mechanics",
  ["annu rev food sci technol"] = "Annual review of food science and technology",
  ["annu rev genet"] = "Annual review of genetics",
  ["annu rev genomics hum genet"] = "Annual review of genomics and human genetics",
  ["annu rev gerontol geriatr"] = "Annual review of gerontology & geriatrics",
  ["annu rev heat transfer"] = "Annual Review of Heat Transfer",
  ["annu rev immunol"] = "Annual review of immunology",
  ["annu rev law soc sci"] = "Annual review of law and social science",
  ["annu rev linguist"] = "Annual review of linguistics",
  ["annu rev mar sci"] = "Annual Review of Marine Science",
  ["annu rev mater res"] = "Annual Review of Materials Research",
  ["annu rev mater sci"] = "Annual Review of Materials Science",
  ["annu rev med"] = "Annual review of medicine",
  ["annu rev microbiol"] = "Annual review of microbiology",
  ["annu rev mod quant financ"] = "Annual Reviews in Modern Quantitative Finance. Including Current Aspects of Fintech, Risk and Investments",
  ["annu rev neurosci"] = "Annual review of neuroscience",
  ["annu rev nucl part sci"] = "Annual Review of Nuclear and Particle Science",
  ["annu rev nucl sci"] = "Annual review of nuclear science",
  ["annu rev nurs res"] = "Annual review of nursing research",
  ["annu rev nutr"] = "Annual review of nutrition",
  ["annu rev pathol"] = "Annual review of pathology",
  ["annu rev pathol: mech dis"] = "Annual Review of Pathology: Mechanisms of Disease",
  ["annu rev pathol: pathol mech dis"] = "Annual Review of Pathology: Mechanisms of Disease",
  ["annu rev pharmacol"] = "Annual review of pharmacology",
  ["annu rev pharmacol toxicol"] = "Annual Review of Pharmacology and Toxicology",
  ["annu rev phys chem"] = "Annual Review of Physical Chemistry",
  ["annu rev physiol"] = "Annual review of physiology",
  ["annu rev phytopathol"] = "Annual review of phytopathology",
  ["annu rev plant biol"] = "Annual review of plant biology",
  ["annu rev plant physiol"] = "Annual review of plant physiology",
  ["annu rev plant physiol plant mol biol"] = "Annual review of plant physiology and plant molecular biology",
  ["annu rev polit sci (palo alto)"] = "Annual review of political science (Palo Alto, Calif.)",
  ["annu rev popul law"] = "Annual review of population law",
  ["annu rev psychol"] = "Annual review of psychology",
  ["annu rev public health"] = "Annual review of public health",
  ["annu rev rehabil"] = "Annual review of rehabilitation",
  ["annu rev resour economics"] = "Annual review of resource economics",
  ["annu rev sex res"] = "Annual review of sex research",
  ["annu rev sociol"] = "Annual review of sociology",
  ["annu rev stat appl"] = "Annual review of statistics and its application",
  ["annu rev virol"] = "Annual review of virology",
  ["annu rev vis sci"] = "Annual review of vision science",
  ["annu rev vision sci"] = "Annual Review of Vision Science",
  ["annu soc christ ethics"] = "The annual of the Society of Christian Ethics. Society of Christian Ethics (U.S.)",
  ["annu stat assist previd soc"] = "Annuario statistico dell'assistenza e della previdenza sociale. Istituto centrale di statistica (Italy)",
  ["annu stat belg congo belge"] = "Annuaire statistique de la Belgique et du Congo belge",
  ["annu surv am law"] = "Annual survey of American law",
  ["annu symp nurs fac pract"] = "Annual Symposium on Nursing Faculty Practice",
  ["annual rep fac ed iwate univ"] = "Iwate University",
  ["annual rev comput phys"] = "Annual Reviews of Computational Physics",
  ["anot pediatr"] = "Anotaciones pediátricas",
  ["anphi pap"] = "ANPHI papers",
  ["ans adv nurs sci"] = "ANS. Advances in nursing science",
  ["anshu keji"] = "Anshu keji",
  ["antarct j us"] = "Antarctic journal of the United States",
  ["antarct rec"] = "Antarctic Record",
  ["antarct res ser"] = "Antarctic research series",
  ["antarct sci"] = "Antarctic Science",
  ["antenor (paris)"] = "Antenor (Paris)",
  ["anthem stud wittgenstein"] = "Anthem Studies in Wittgenstein",
  ["anthropocene coasts"] = "Anthropocene Coasts",
  ["anthropocene rev"] = "Anthropocene Review",
  ["anthropocene sci"] = "Anthropocene Science",
  ["anthropol action"] = "Anthropology in action : newsletter of the British Association for Social Anthropology in Policy and Practice (BASAPP)",
  ["anthropol anz"] = "Anthropologischer Anzeiger; Bericht über die biologisch-anthropologische Literatur",
  ["anthropol archeol eurasia"] = "Anthropology & archeology of Eurasia",
  ["anthropol conscious"] = "Anthropology of consciousness",
  ["anthropol educ q"] = "Anthropology & education quarterly",
  ["anthropol forum"] = "Anthropological forum",
  ["anthropol kozl"] = "Anthropologiai közlemények",
  ["anthropol med"] = "Anthropology & medicine",
  ["anthropol now"] = "Anthropology now",
  ["anthropol pap am mus nat hist"] = "Anthropological papers of the American Museum of Natural History",
  ["anthropol philos"] = "Anthropology and Philosophy",
  ["anthropol q"] = "Anthropological quarterly",
  ["anthropol sci"] = "Anthropological science : journal of the Anthropological Society of Nippon = Jinruigaku zasshi",
  ["anthropol soc"] = "Anthropologie et sociétés",
  ["anthropol today"] = "Anthropology today",
  ["anthropol ucla"] = "Anthropology U.C.L.A",
  ["anthropol work rev"] = "Anthropology of work review : AWR",
  ["anti-cancer agents med chem"] = "Anti-Cancer Agents in Medicinal Chemistry",
  ["anti-cancer drug des"] = "Anti-Cancer Drug Design",
  ["anti-cancer drugs"] = "Anti-Cancer Drugs",
  ["anti-corros methods mater"] = "Anti-Corrosion Methods and Materials",
  ["anti-infect agents"] = "Anti-infective Agents",
  ["anti-infect agents med chem"] = "Anti-infective Agents in Medicinal Chemistry",
  ["anti-inflammatory anti-allergy agents med chem"] = "Anti-Inflammatory & Anti-Allergy Agents in Medicinal Chemistry",
  ["antib ther"] = "Antibody therapeutics",
  ["antibiot annu"] = "Antibiotics annual",
  ["antibiot chemother"] = "Antibiotics and Chemotherapy",
  ["antibiot chemother (1971)"] = "Antibiotics and chemotherapy",
  ["antibiot chemother (northfield)"] = "Antibiotics & chemotherapy (Northfield, Ill.)",
  ["antibiot khimioter"] = "Antibiotiki i khimioterapii︠a︡ = Antibiotics and chemoterapy [sic]",
  ["antibiot med biotekhnol"] = "Antibiotiki i meditsinskaia biotekhnologiia = Antibiotics and medical biotechnology",
  ["antibiotic med clin ther"] = "Antibiotic medicine & clinical therapy",
  ["antibiotic med clin ther (new york)"] = "Antibiotic medicine & clinical therapy (New York, NY)",
  ["antibiotica [quad]"] = "Antibiotica. Quaderni",
  ["antibiotics (basel)"] = "Antibiotics (Basel, Switzerland)",
  ["antibiotics (basel, switz)"] = "Antibiotics (Basel, Switzerland)",
  ["antibiotiques (paris)"] = "Antibiotiques (Paris, France : 1999)",
  ["antibodies (basel)"] = "Antibodies (Basel, Switzerland)",
  ["antibody ther"] = "Antibody Therapeutics",
  ["anticancer agents med chem"] = "Anti-cancer agents in medicinal chemistry",
  ["anticancer drug des"] = "Anti-cancer drug design",
  ["anticancer drugs"] = "Anti-cancer drugs",
  ["anticancer res"] = "Anticancer Research",
  ["anticip sci"] = "Anticipation Science",
  ["antiinfect agents"] = "Anti-infective agents",
  ["antiinfect agents med chem"] = "Anti-infective agents in medicinal chemistry",
  ["antiinflamm antiallergy agents med chem"] = "Anti-inflammatory & anti-allergy agents in medicinal chemistry",
  ["antike welt"] = "Antike Welt",
  ["antimicrob agents chemother"] = "Antimicrobial Agents and Chemotherapy",
  ["antimicrob comb devices (2019)"] = "Antimicrobial combination devices",
  ["antimicrob resist infect control"] = "Antimicrobial resistance and infection control",
  ["antimicrob stewardship healthcare epidemiol"] = "Antimicrobial Stewardship and Healthcare Epidemiology",
  ["antimicrobial agents chemother (bethesda)"] = "Antimicrobial agents and chemotherapy",
  ["antioch rev"] = "The Antioch review",
  ["antioquia med"] = "Antioquia médica",
  ["antioxid redox signal"] = "Antioxidants & redox signaling",
  ["antioxid redox signaling"] = "Antioxidants & Redox Signaling",
  ["antioxidants (basel)"] = "Antioxidants (Basel, Switzerland)",
  ["antiq class"] = "L'Antiquité classique",
  ["antiq math"] = "Antiquitates Mathematicae. Roczniki Polskiego Towarzystwa Matematycznego. Seria VI",
  ["antisense nucleic acid drug dev"] = "Antisense and Nucleic Acid Drug Development",
  ["antisense res dev"] = "Antisense research and development",
  ["antitrust bull"] = "Antitrust Bulletin",
  ["antitrust law econ"] = "Antitrust Law and Economics Review",
  ["antivir chem chemother"] = "Antiviral chemistry & chemotherapy",
  ["antivir ther"] = "Antiviral therapy",
  ["antiviral chem chemother"] = "Antiviral Chemistry and Chemotherapy",
  ["antiviral res"] = "Antiviral Research",
  ["antol med ital arch"] = "Antologia medica italiana. Archivio",
  ["antonie van leeuwenhoek"] = "Antonie van Leeuwenhoek",
  ["antropologia (rome)"] = "Antropologia (Rome, Italy)",
  ["anu bras odontol"] = "Anuario Brasileiro De Odontologia",
  ["anu estad peru"] = "Anuario estadístico del Perú. Peru. Dirección nacional de estadística",
  ["anu estud am"] = "Anuario de estudios americanos",
  ["anu estud atl"] = "Anuario de estudios Atlánticos",
  ["anu estud centroam"] = "Anuario de estudios centroamericanos",
  ["anu estud mediev"] = "Anuario de estudios medievales",
  ["anu filol univ barc"] = "Anuari de Filologia (Universitat de Barcelona)",
  ["anu hist j"] = "ANU historical journal",
  ["anu iehs"] = "Anuario IEHS",
  ["anu indig"] = "Anuario indigenista",
  ["anu investig adicciones"] = "Anuario de investigación en adicciones",
  ["anwendorientier math"] = "Anwendungsorientierte Mathematik",
  ["anwendorientier stat"] = "Anwendungsorientierte Statistik",
  ["anxiety stress coping"] = "Anxiety, stress, and coping",
  ["anz altertwiss"] = "Anzeiger für die Altertumswissenschaft",
  ["anz j surg"] = "ANZ journal of surgery",
  ["anz osterr akad wiss philos-hist kl"] = "Anzeiger. Osterreichische Akademie der Wissenschaften. Philosophisch-Historiche Klasse",
  ["anz philol hist kl osterr akad wiss"] = "Anzeiger (Österreichische Akademie der Wissenschaften. Philosophisch-Historische Klasse)",
  ["anz schädlkd pflanzenschutz umweltschutz"] = "Anzeiger für Schädlingskunde, Pflanzenschutz, Umweltschutz",
  ["anz österreich akad wiss math-natur kl"] = "Österreichische Akademie der Wissenschaften",
  ["anzhes j"] = "ANZHES journal. Australian and New Zealand History of Education Society",
  ["anziam j"] = "The ANZIAM Journal",
  ["aob plants"] = "AoB PLANTS",
  ["aoha prog"] = "AOHA Progress",
  ["aoha today"] = "AOHA today",
  ["aorn j"] = "AORN journal",
  ["aorta (stamford)"] = "Aorta (Stamford, Conn.)",
  ["apa newsl"] = "APA newsletters",
  ["apa newsl philos med"] = "Newsletter on philosophy and medicine",
  ["apacph clear house bull"] = "APACPH-Clearing House bulletin",
  ["apar respir tuberc"] = "Aparato respiratorio y tuberculosis",
  ["apcbee procedia"] = "APCBEE Procedia",
  ["apdi newsl"] = "APDI newsletter. United Nations Asian and Pacific Development Institute",
  ["apdt chron dog"] = "The APDT chronicle of the dog",
  ["aperito j cell mol biol"] = "Aperito journal of cellular and molecular biology",
  ["apl bioeng"] = "APL bioengineering",
  ["apl comput phys"] = "APL Computational Physics",
  ["apl electron devices"] = "APL Electronic Devices",
  ["apl energy"] = "APL Energy",
  ["apl mach learn"] = "APL Machine Learning",
  ["apl mater"] = "APL Materials",
  ["apl photonics"] = "APL Photonics",
  ["apl quantum"] = "APL Quantum",
  ["apl: org electron photonics"] = "APL: Organic Electronics and Photonics",
  ["apmis suppl"] = "APMIS. Supplementum",
  ["apollonia (sydney)"] = "Apollonia",
  ["aportaciones mat"] = "Aportaciones Matematicas",
  ["aportaciones mat comun"] = "Aportaciones Matematicas: Comunicaciones",
  ["aportaciones mat investig"] = "Aportaciones Matematicas: Investigacion",
  ["aportaciones mat notas investigación"] = "Aportaciones Matemáticas: Notas de Investigación",
  ["aportaciones mat textos"] = "Aportaciones Matematicas: Textos",
  ["apoth (beirut)"] = "The Apothecary",
  ["apoth kunst"] = "Apotheker und Kunst",
  ["appalach j"] = "Appalachian journal",
  ["appea j"] = "APPEA Journal",
  ["appita j"] = "Appita Journal",
  ["appl acoust"] = "Applied Accoustics",
  ["appl adhes sci"] = "Applied Adhesion Science",
  ["appl ai lett"] = "Applied AI Letters",
  ["appl algebra eng comm comput"] = "Applicable Algebra in Engineering, Communication and Computing",
  ["appl algebra eng commun comput"] = "Applicable Algebra in Engineering, Communication and Computing",
  ["appl algebra engrg comm comput"] = "Applicable Algebra in Engineering, Communication and Computing",
  ["appl anal"] = "Applicable analysis",
  ["appl anal discrete math"] = "Applicable Analysis and Discrete Mathematics",
  ["appl anal optim"] = "Applied Analysis and Optimization",
  ["appl anim behav sci"] = "Applied animal behaviour science",
  ["appl anim ethol"] = "Applied Animal Ethology",
  ["appl anim sci"] = "Applied Animal Science",
  ["appl anthropol"] = "Applied anthropology",
  ["appl appl math"] = "Applications and Applied Mathematics",
  ["appl artif intell"] = "Applied Artifical Intelligence",
  ["appl behav sci rev"] = "Applied behavioral science review",
  ["appl biochem biotechnol"] = "Applied Biochemistry and Biotechnology",
  ["appl biochem microbiol"] = "Applied Biochemistry and Microbiology",
  ["appl bioinf"] = "Applied Bioinformatics",
  ["appl bioinformatics"] = "Applied bioinformatics",
  ["appl biol chem"] = "Applied biological chemistry",
  ["appl biol res"] = "Applied biological research",
  ["appl bionics biomech"] = "Applied bionics and biomechanics",
  ["appl biosaf"] = "Applied biosafety : journal of the American Biological Safety Association",
  ["appl biotechnol food sci policy"] = "Applied biotechnology, food science and policy",
  ["appl cardiol"] = "Applied cardiology",
  ["appl cardiopulm pathophysiol"] = "Applied cardiopulmonary pathophysiology : ACP",
  ["appl catal"] = "Applied catalysis",
  ["appl catal a gen"] = "Applied catalysis. A, General",
  ["appl catal b"] = "Applied catalysis. B, Environmental",
  ["appl catal, a"] = "Applied Catalysis A: General",
  ["appl catal, b"] = "Applied Catalysis B: Environmental",
  ["appl categ structures"] = "Applied Categorical Structures. A Journal Devoted to Applications of Categorical Methods in Algebra, Analysis, Computer Science, Logic, Order and Topology",
  ["appl categor struct"] = "Applied Categorical Structures",
  ["appl clay sci"] = "Applied Clay Science",
  ["appl clin genet"] = "The application of clinical genetics",
  ["appl clin inf"] = "Applied Clinical Informatics",
  ["appl clin inform"] = "Applied clinical informatics",
  ["appl clin trials"] = "Applied clinical trials",
  ["appl cogn psychol"] = "Applied cognitive psychology",
  ["appl commun theory"] = "Applications of Communications Theory",
  ["appl compos mater"] = "Applied Composite Materials",
  ["appl comput electromagn soc j"] = "Applied Computational Electromagnetics Society journal",
  ["appl comput geosci"] = "Applied Computing and Geosciences",
  ["appl comput harmon anal"] = "Applied and Computational Harmonic Analysis",
  ["appl comput inf"] = "Applied Computing and Informatics",
  ["appl comput intell soft comput"] = "Applied Computational Intelligence and Soft Computing",
  ["appl comput math"] = "Applied and Computational Mathematics. An International Journal",
  ["appl comput sci"] = "Applied Computer Science. Berichte zur Praktischen Informatik",
  ["appl cytogenet"] = "Applied cytogenetics : journal of the Association of Cytogenetic Technologists",
  ["appl demor"] = "Applied demography",
  ["appl dev sci"] = "Applied developmental science",
  ["appl discrete math theoret comput sci"] = "Applied Discrete Mathematics and Theoretical Computer Science",
  ["appl earth sci"] = "Applied Earth Science",
  ["appl ecol abstr"] = "Applied Ecology Abstracts",
  ["appl ecol environ res"] = "Applied ecology and environmental research",
  ["appl econ"] = "Applied economics",
  ["appl econ discuss pap ser"] = "Applied economics discussion paper series",
  ["appl econ lett"] = "Applied economics letters",
  ["appl econ letters"] = "Applied Economics Letters",
  ["appl econ perspect policy"] = "Applied economic perspectives and policy",
  ["appl energy"] = "Applied Energy",
  ["appl energy combust sci"] = "Applications in Energy and Combustion Science",
  ["appl eng agric"] = "Applied engineering in agriculture",
  ["appl eng math ser"] = "Applied and Engineering Mathematics Series",
  ["appl eng sci"] = "Applications in Engineering Science",
  ["appl entomol phytopathol"] = "Applied entomology and phytopathology",
  ["appl entomol zool"] = "Applied entomology and zoology",
  ["appl entomol zool (jpn)"] = "Applied entomology and zoology",
  ["appl environ biotechnol"] = "Applied environmental biotechnology",
  ["appl environ educ commun"] = "Applied environmental education and communication (Print)",
  ["appl environ microbiol"] = "Applied and Environmental Microbiology",
  ["appl environ soil sci"] = "Applied and environmental soil science",
  ["appl ergon"] = "Applied Ergonomics",
  ["appl evol comput (2014)"] = "Applications of Evolutionary Computation : 17th European Conference, EvoApplications 2014, Granada, Spain, April 23-25, 2014 : revised selected papers. EvoApplications (Conference) (17th : 2014 : Granada, Spain)",
  ["appl finan econ"] = "Applied Financial Economics",
  ["appl finance account"] = "Applied finance and accounting",
  ["appl food biotechol"] = "Applied Food Biotechnology",
  ["appl food res"] = "Applied Food Research",
  ["appl gen topol"] = "Applied General Topology",
  ["appl genomics proteomics"] = "Applied genomics and proteomics",
  ["appl geochem"] = "Applied Geochemistry",
  ["appl geogr"] = "Applied Geography",
  ["appl geogr dev"] = "Applied geography and development",
  ["appl geogr devel"] = "Applied Geography and Development",
  ["appl geomatics"] = "Applied Geomatics",
  ["appl geophys"] = "Applied Geophysics",
  ["appl health econ health policy"] = "Applied health economics and health policy",
  ["appl hum factors ergon conf"] = "Applied Human Factors and Ergonomics Conference",
  ["appl human sci"] = "Applied human science : journal of physiological anthropology",
  ["appl immunohistochem mol morphol"] = "Applied Immunohistochemistry and Molecular Morphology",
  ["appl in vitro toxicol"] = "Applied in vitro toxicology",
  ["appl inf"] = "Applied Informatics",
  ["appl inform (berl)"] = "Applied informatics",
  ["appl intell"] = "Applied Intelligence",
  ["appl linguist"] = "Applied linguistics",
  ["appl log ser"] = "Applied Logic Series",
  ["appl magn reson"] = "Applied magnetic resonance",
  ["appl manag sci"] = "Applications of Management Science",
  ["appl mater today"] = "Applied Materials Today",
  ["appl math"] = "Applications of Mathematics",
  ["appl math (irvine)"] = "Applied mathematics",
  ["appl math (prague)"] = "Applications of mathematics (Prague, Czechoslovakia)",
  ["appl math (warsaw)"] = "Applicationes Mathematicae",
  ["appl math comput"] = "Applied Mathematics and Computation",
  ["appl math comput sci"] = "Applied Mathematics and Computer Science",
  ["appl math e-notes"] = "Applied Mathematics E-Notes",
  ["appl math eng sci texts"] = "Applied Mathematics and Engineering Science Texts",
  ["appl math finance"] = "Applied Mathematical Finance",
  ["appl math inf sci"] = "Applied Mathematics & Information Sciences",
  ["appl math inform mech"] = "Applied Mathematics, Informatics and Mechanics",
  ["appl math j chin univ"] = "Applied Mathematics A Journal of Chinese Universities",
  ["appl math j chinese univ ser a"] = "Applied Mathematics. A Journal of Chinese Universities. Series A. Gaoxiao Yingyong Shuxue Xuebao",
  ["appl math j chinese univ ser b"] = "Applied Mathematics. A Journal of Chinese Universities. Ser. B",
  ["appl math lett"] = "Applied Mathematics Letters",
  ["appl math math comput"] = "Applied Mathematics and Mathematical Computation",
  ["appl math mech"] = "Applied Mathematics and Mechanics",
  ["appl math mech (english ed)"] = "Applied Mathematics and Mechanics. English Edition",
  ["appl math mod challenges"] = "Applied Mathematics for Modern Challenges",
  ["appl math model"] = "Applied mathematical modelling",
  ["appl math modell"] = "Applied Mathematical Modelling",
  ["appl math nonlinear sci"] = "Applied Mathematics and Nonlinear Sciences",
  ["appl math optim"] = "Applied Mathematics and Optimization",
  ["appl math res express"] = "Applied Mathematics Research eXpress",
  ["appl math sci"] = "Applied Mathematical Sciences",
  ["appl math sci eng"] = "Applied Mathematics in Science and Engineering",
  ["appl mech rev"] = "Applied Mechanics Reviews",
  ["appl mech,"] = "Applied Mechanics",
  ["appl med inform"] = "Applied medical informatics",
  ["appl microbiol"] = "Applied microbiology",
  ["appl microbiol (los angel)"] = "Applied microbiology (Los Angeles, CA)",
  ["appl microbiol biotechnol"] = "Applied Microbiology and Biotechnology",
  ["appl micrograv technol"] = "Applied microgravity technology",
  ["appl microsc"] = "Applied microscopy",
  ["appl modern tech business"] = "Applications of Modern Technology in Business",
  ["appl nanosci"] = "Applied Nanoscience",
  ["appl netw sci"] = "Applied network science",
  ["appl network sci"] = "Applied Network Science",
  ["appl neurophysiol"] = "Applied neurophysiology",
  ["appl neuropsychol"] = "Applied neuropsychology",
  ["appl neuropsychol adult"] = "Applied neuropsychology. Adult",
  ["appl neuropsychol child"] = "Applied neuropsychology. Child",
  ["appl nmr spectrosc"] = "Applications of NMR spectroscopy",
  ["appl numer anal comput math"] = "Applied Numerical Analysis & Computational Mathematics",
  ["appl numer harmon anal"] = "Applied and Numerical Harmonic Analysis",
  ["appl numer math"] = "Applied numerical mathematics : transactions of IMACS",
  ["appl nurs res"] = "Applied nursing research : ANR",
  ["appl occup environ hyg"] = "Applied occupational and environmental hygiene",
  ["appl ocean res"] = "Applied Ocean Research",
  ["appl ontol"] = "Applied ontology",
  ["appl opt"] = "Applied Optics",
  ["appl optim"] = "Applied Optimization",
  ["appl organomet chem"] = "Applied Organometallic Chemistry",
  ["appl parasitol"] = "Applied parasitology",
  ["appl pathol"] = "Applied pathology",
  ["appl petrochem res"] = "Applied Petrochemical Research",
  ["appl philos"] = "Applied philosophy (Fort Pierce, Fla.)",
  ["appl phycol"] = "Applied Phycology",
  ["appl phys a"] = "Applied Physics A",
  ["appl phys a mater sci process"] = "Applied physics. A, Materials science & processing",
  ["appl phys a: mater sci process"] = "Applied Physics A: Materials Science & Processing",
  ["appl phys b"] = "Applied Physics B",
  ["appl phys b: lasers opt"] = "Applied Physics B: Lasers and Optics",
  ["appl phys express"] = "Applied Physics Express",
  ["appl phys lett"] = "Applied Physics Letters",
  ["appl phys mater sci process"] = "Applied Physics A: Materials Science & Processing",
  ["appl phys res"] = "Applied Physics Research",
  ["appl phys rev"] = "Applied Physics Reviews",
  ["appl physiol nutr metab"] = "Applied Physiology, Nutrition and Metabolism",
  ["appl plant sci"] = "Applications in plant sciences",
  ["appl polym compos"] = "Applied Polymer Composites",
  ["appl polym symp"] = "Applied polymer symposia",
  ["appl prev psychol"] = "Applied & preventive psychology : journal of the American Association of Applied and Preventive Psychology",
  ["appl probab index"] = "Applied Probability",
  ["appl psychol"] = "Applied psychology = Psychologie appliquée",
  ["appl psychol crim justice"] = "Applied psychology in criminal justice",
  ["appl psychol health well being"] = "Applied psychology. Health and well-being",
  ["appl psychol meas"] = "Applied psychological measurement",
  ["appl psycholinguist"] = "Applied psycholinguistics",
  ["appl psychophysiol biofeedback"] = "Applied psychophysiology and biofeedback",
  ["appl quant finance"] = "Applied Quantitative Finance",
  ["appl radiat isot"] = "Applied Radiation and Isotopes",
  ["appl radiat oncol"] = "Applied radiation oncology",
  ["appl radiol"] = "Applied radiology",
  ["appl res"] = "Applied Research",
  ["appl res ment retard"] = "Applied research in mental retardation",
  ["appl res qual life"] = "Applied research in quality of life",
  ["appl rheol"] = "Applied rheology (Lappersdorf, Germany : Online)",
  ["appl sci"] = "APPS. Applied Sciences",
  ["appl sci (basel)"] = "Applied sciences (Basel, Switzerland)",
  ["appl sci eng"] = "Applied Sciences and Engineering",
  ["appl sci monogr"] = "Applied Sciences. Monographs",
  ["appl sci res"] = "Applied Scientific Research",
  ["appl sci switz"] = "Applied Sciences (Switzerland)",
  ["appl sci-basel"] = "Applied Sciences-Basel",
  ["appl soc res methods ser"] = "Applied Social Research Methods Series",
  ["appl soft comput"] = "Applied Soft Computing",
  ["appl soil ecol"] = "Applied soil ecology : a section of Agriculture, Ecosystems & Environment",
  ["appl sol energy"] = "Applied Solar Energy",
  ["appl solar energy"] = "Applied Solar Energy",
  ["appl spat anal policy"] = "Applied spatial analysis and policy",
  ["appl spatial anal policy"] = "Applied Spatial Analysis and Policy",
  ["appl spectrosc"] = "Applied Spectroscopy",
  ["appl spectrosc pract"] = "Applied Spectroscopy Practica",
  ["appl spectrosc rev"] = "Applied Spectroscopy Reviews",
  ["appl stat"] = "Applied statistics",
  ["appl stoch methods ser"] = "Applied Stochastic Methods Series",
  ["appl stoch models bus ind"] = "Applied stochastic models in business and industry",
  ["appl stochastic models bus ind"] = "Applied Stochastic Models in Business and Industry",
  ["appl stochastic models data anal"] = "Applied Stochastic Models and Data Analysis",
  ["appl supercond"] = "Applied Superconductivity",
  ["appl surf sci"] = "Applied Surface Science",
  ["appl surf sci adv"] = "Applied Surface Science Advances",
  ["appl theor electrophor"] = "Applied and theoretical electrophoresis : the official journal of the International Electrophoresis Society",
  ["appl ther"] = "Applied therapeutics",
  ["appl therm eng"] = "Applied Thermal Engineering",
  ["appl transl genom"] = "Applied & translational genomics",
  ["appl transl genomics"] = "Applied & Translational Genomics",
  ["appl transl geonomics"] = "Applied and Translational Geonomics",
  ["appl turfgrass sci"] = "Applied Turfgrass Science",
  ["appl veg sci"] = "Applied Vegetation Science",
  ["appl water sci"] = "Applied Water Science",
  ["appraisal j"] = "The Appraisal journal",
  ["approaches semiotics"] = "Approaches to Semiotics",
  ["approx optim"] = "Approximation & Optimization",
  ["approx theory appl (ns)"] = "Approximation Theory and its Applications",
  ["approximation theory appl"] = "Approximation Theory and its Applications",
  ["appunti sc norm super pisa (n s)"] = "Appunti. Scuola Normale Superiore di Pisa (Nuova Serie)",
  ["apres demain"] = "Après-demain",
  ["aps obs"] = "APS observer",
  ["apsipa trans signal inf process"] = "APSIPA Transactions on Signal and Information Processing",
  ["apsp j case rep"] = "APSP journal of case reports",
  ["aptamers (oxf)"] = "Aptamers (Oxford, England)",
  ["aptechn delo"] = "Aptechnoe delo",
  ["apunt psicol"] = "Apuntes de psicología",
  ["apunts sports med"] = "Apunts Sports Medicine",
  ["aqeic bol tec"] = "AQEIC Boletin Tecnico",
  ["aqua planta"] = "Aqua planta",
  ["aqua water infrastruct ecosyst soc"] = "AQUA - Water Infrastructure, Ecosystems and Society",
  ["aquac aquar conserv legis"] = "Aquaculture, aquarium, conservation & legislation : international journal of the Bioflux Society",
  ["aquac assoc can spec"] = "Aquaculture Association of Canada Special Publication",
  ["aquac eng"] = "Aquacultural engineering",
  ["aquac environ interact"] = "Aquaculture Environment Interactions",
  ["aquac fish"] = "Aquaculture and fisheries",
  ["aquac int"] = "Aquaculture international : journal of the European Aquaculture Society",
  ["aquac nutr"] = "Aquaculture nutrition",
  ["aquac rep"] = "Aquaculture reports",
  ["aquac res"] = "Aquaculture research",
  ["aquacult econ manage"] = "Aquaculture Economics & Management",
  ["aquacult eng"] = "Aquacultural Engineering",
  ["aquacult fish"] = "Aquaculture and Fisheries",
  ["aquacult fish fish"] = "Aquaculture, Fish and Fisheries",
  ["aquacult int"] = "Aquaculture International",
  ["aquacult nutr"] = "Aquaculture Nutrition",
  ["aquacult rep"] = "Aquaculture Reports",
  ["aquacult res"] = "Aquaculture Research",
  ["aquaculture econ manage"] = "Aquaculture Economics and Management",
  ["aquarium sci conserv"] = "Aquarium Sciences and Conservation",
  ["aquat biol"] = "Aquatic biology",
  ["aquat biosyst"] = "Aquatic biosystems",
  ["aquat bot"] = "Aquatic botany",
  ["aquat conserv"] = "Aquatic conservation : marine and freshwater ecosystems",
  ["aquat conserv mar freshwater ecosyst"] = "Aquatic Conservation: Marine and Freshwater Ecosystems",
  ["aquat ecol"] = "Aquatic ecology",
  ["aquat ecosyst health manag"] = "Aquatic ecosystem health & management",
  ["aquat ecosyst health manage"] = "Aquatic Ecosystem Health & Management",
  ["aquat geochem"] = "Aquatic Geochemistry",
  ["aquat insects"] = "Aquatic insects",
  ["aquat invaders"] = "Aquatic invaders",
  ["aquat invasions"] = "Aquatic invasions",
  ["aquat living resour"] = "Aquatic living resources",
  ["aquat mamm"] = "Aquatic mammals",
  ["aquat microb ecol"] = "Aquatic Microbial Ecology",
  ["aquat procedia"] = "Aquatic procedia",
  ["aquat sci"] = "Aquatic Sciences",
  ["aquat toxicol"] = "Aquatic Toxicology",
  ["aquilo ser bot"] = "Aquilo. Ser. botanica",
  ["arab aff"] = "Arab affairs (London, England)",
  ["arab econ"] = "The Arab economist",
  ["arab gulf j"] = "The Arab Gulf journal",
  ["arab gulf j sci res"] = "Arab Gulf Journal of Scientific Research",
  ["arab j basic appl sci"] = "Arab journal of basic and applied sciences",
  ["arab j biotechnol"] = "Arab journal of biotechnology",
  ["arab j gastroenterol"] = "Arab journal of gastroenterology : the official publication of the Pan-Arab Association of Gastroenterology",
  ["arab j lab med"] = "The Arab journal of laboratory medicine",
  ["arab j math (springer)"] = "Arabian Journal of Mathematics",
  ["arab j math sci"] = "Arab Journal of Mathematical Sciences",
  ["arab j nephrol transplant"] = "Arab journal of nephrology and transplantation",
  ["arab j sci eng"] = "Arabian journal for science and engineering",
  ["arab j sci eng sect a sci"] = "The Arabian Journal for Science and Engineering",
  ["arab j sci eng sect b eng"] = "The Arabian Journal for Science and Engineering",
  ["arab j sci eng sect c theme issues"] = "The Arabian Journal for Science and Engineering",
  ["arab j soc sci"] = "The Arab journal of the social sciences",
  ["arab j urol"] = "Arab journal of urology",
  ["arab sci philos"] = "Arabic sciences and philosophy : a historical journal",
  ["arab stud q"] = "Arab studies quarterly",
  ["arabian archaeol epigr"] = "Arabian Archaeology and Epigraphy",
  ["arabian j chem"] = "Arabian Journal of Chemistry",
  ["arabian j geosci"] = "Arabian Journal of Geosciences",
  ["arabian j math"] = "Arabian Journal of Mathematics",
  ["arabian j sci eng"] = "Arabian Journal for Science and Engineering",
  ["arabic sci philos"] = "Arabic Sciences and Philosophy. A Historical Journal",
  ["arabidopsis book"] = "The arabidopsis book",
  ["arachnol mitt"] = "Arachnologische mitteilungen",
  ["arachnol mitt: arachnol lett"] = "Arachnologische Mitteilungen: Arachnology Letters",
  ["arag natl newsl"] = "A.R.A.G. national newsletter",
  ["aramco world mag"] = "Aramco world magazine",
  ["arb anat inst kais-jpn univ sendai"] = "Arbeiten aus dem Anatomischen Institut der Kaiserlich-Japanischen Universitaet zu Sendai",
  ["arb dlg"] = "Arbeiten der DLG",
  ["arb gesch med giessen"] = "Arbeiten zur Geschichte der Medizin in Giessen",
  ["arb kais biol anst land forstwirtsch"] = "Arbeiten der Kaiserlichen Biologischen Anstalt fur Land- und Forstwirtschaft",
  ["arb paul ehrlich inst bundesamt sera impfstoffe frankf a m"] = "Arbeiten aus dem Paul-Ehrlich-Institut (Bundesamt für Sera und Impfstoffe) zu Frankfurt a.M",
  ["arb paul ehrlich inst bundesinstitut impfstoffe biomed arzneim langen hess"] = "Arbeiten aus dem Paul-Ehrlich-Institut (Bundesinstitut für Impfstoffe und biomedizinische Arzneimittel) Langen/Hessen",
  ["arb paul ehrlich inst georg speyer haus ferdinand blum inst frankf a m"] = "Arbeiten aus dem Paul-Ehrlich-Institut, dem Georg-Speyer-Haus und dem Ferdinand-Blum-Institut zu Frankfurt a.M",
  ["arb soz"] = "Arbeit und Soziales",
  ["arbeitsber inst math masch datenverarb inform"] = "Arbeitsberichte des Instituts für Mathematische Maschinen und Datenverarbeitung (Informatik)",
  ["arbeitsber inst math masch datenverarb, band 15"] = "Arbeitsberichte des Instituts fur Mathematische Maschinen und Datenverarbeitung, Band 15",
  ["arbeitsmed sozialmed praventivmed"] = "Arbeitsmedizin, Sozialmedizin, Präventivmedizin",
  ["arbeitspapiere sonderforschungsbereichs 340 sprachtheor grundlagen computerlinguist ber"] = "Arbeitspapiere des Sonderforschungsbereichs 340, Sprachtheoretische Grundlagen für die Computerlinguistik",
  ["arbetarrorel arsb"] = "Arbetarrörelsens årsbok",
  ["arbor scientiarum: beitrage wissenschaftsgeschichte, reihe a: abhandlungen"] = "Arbor Scientiarum: Beitrage zur Wissenschaftsgeschichte, Reihe A: Abhandlungen",
  ["arboric j"] = "Arboricultural journal",
  ["arboric urban for"] = "Arboriculture & urban forestry",
  ["arc cercle"] = "Arc de cercle : an international journal of the history of the mind-sciences",
  ["arc j addict"] = "ARC journal of addiction",
  ["arc j diabetes endocrinol"] = "ARC journal of diabetes and endocrinology",
  ["arc j public health community med"] = "ARC journal of public health and community medicine",
  ["arch acad emerg med"] = "Archives of academic emergency medicine",
  ["arch acker pflanzenbau bodenkd"] = "Archiv für Acker- und Pflanzenbau und Bodenkunde",
  ["arch agron soil sci"] = "Archives of Agronomy and Soil Science",
  ["arch aids res"] = "Archives of AIDS research",
  ["arch am med"] = "Archivos americanos de medicina",
  ["arch anat cytol pathol"] = "Archives d'anatomie et de cytologie pathologiques",
  ["arch anat histol embryol"] = "Archives d'anatomie, d'histologie et d'embryologie normales et expérimentales",
  ["arch anat microsc morphol exp"] = "Archives d'anatomie microscopique et de morphologie expérimentale",
  ["arch anat pathol (paris)"] = "Archives d'anatomie pathologique",
  ["arch androl"] = "Archives of andrology",
  ["arch anesthesiol"] = "Archives of anesthesiology",
  ["arch anim breed"] = "Archives animal breeding",
  ["arch anim nutr"] = "Archives of animal nutrition",
  ["arch antropol crim psichiatr med leg"] = "Archivio di antropologia criminale, psichiatria e medicina legale",
  ["arch appl mech"] = "Archive of applied mechanics = Ingenieur-Archiv",
  ["arch appl sci res"] = "Archives of applied science research",
  ["arch argent dermatol"] = "Archivos argentinos de dermatología",
  ["arch argent enferm apar dig"] = "Archivos argentinos de enfermedades del aparato digestivo",
  ["arch argent kinesiol kinesiter kinefilax"] = "Archivos argentinos de kinesiología; kinesiterapia y kinefilaxia",
  ["arch argent pediatr"] = "Archivos argentinos de pediatría",
  ["arch argent repub secr salud publica"] = "Archivos. Argentina. Secretaría de Salud Pública",
  ["arch argent tisiol"] = "Archivos argentinos de tisiología",
  ["arch argent tisiol neumonol"] = "Archivos argentinos de tisiología y neumonología",
  ["arch asoc evit ceguera mex"] = "Archivos. Asociación para Evitar la Ceguera en México",
  ["arch atti"] = "Archivio ed atti. Società italiana di chirurgia",
  ["arch autoimmune dis"] = "Archives of autoimmune diseases",
  ["arch automot eng"] = "Archives of Automotive Engineering",
  ["arch basic appl med"] = "Archives of basic and applied medicine",
  ["arch belg"] = "Archives belges = Belgisch archief",
  ["arch belg dermatol"] = "Archives belges de dermatologie",
  ["arch belg dermatol syphiligr"] = "Archives belges de dermatologie et de syphiligraphie",
  ["arch belg med soc"] = "Archives belges de médecine sociale, hygiène, médecine du travail et médecine légale. Belgisch archief van sociale geneeskunde, hygiëne, arbeidsgeneeskunde en gerechtelijke geneeskunde",
  ["arch biochem"] = "Archives of biochemistry",
  ["arch biochem biophys"] = "Archives of Biochemistry and Biophysics",
  ["arch biol (liege)"] = "Archives de biologie",
  ["arch biol andina"] = "Archivos de biología andina",
  ["arch biol med exp"] = "Archivos de biología y medicina experimentales",
  ["arch biol med exp (santiago)"] = "Archivos de Biologia y Medicina Experimentales",
  ["arch biol sci"] = "Archives of biological sciences",
  ["arch biomed eng biotechnol"] = "Archives in biomedical engineering & biotechnology",
  ["arch bioquim quim farm tucuman"] = "Archivos de bioquímica, química y farmacia, Tucumán",
  ["arch bone jt surg"] = "The archives of bone and joint surgery",
  ["arch bras med"] = "Archivos brasileiros de medicina",
  ["arch bronconeumol"] = "Archivos de bronconeumología",
  ["arch cancer biol ther"] = "Archives of cancer biology and therapy",
  ["arch cancer res"] = "Archives in cancer research",
  ["arch cardiol mex"] = "Archivos de cardiología de México",
  ["arch cardiovasc dis"] = "Archives of cardiovascular diseases",
  ["arch cas"] = "Archivní časopis",
  ["arch child health"] = "Archives of child health",
  ["arch chir neerl"] = "Archivum chirurgicum Neerlandicum",
  ["arch chir torac cardiovasc"] = "Archivio di chirurgia toracica e cardiovascolare",
  ["arch civ mech eng"] = "Archives of Civil and Mechanical Engineering",
  ["arch clin biomed res"] = "Archives of clinical and biomedical research",
  ["arch clin case rep"] = "Archives of clinical case reports",
  ["arch clin infect dis"] = "Archives of clinical infectious diseases",
  ["arch clin med case rep"] = "Archives of clinical and medical case reports",
  ["arch clin microbiol"] = "Archives of clinical microbiology",
  ["arch clin neuropsychol"] = "Archives of clinical neuropsychology : the official journal of the National Academy of Neuropsychologists",
  ["arch clin toxicol (middlet)"] = "Archives of clinical toxicology",
  ["arch col med el salv"] = "Archivos del Colegio Medico de El Salvador",
  ["arch combust"] = "Archivum Combustionis",
  ["arch comput methods eng"] = "Archives of Computational Methods in Engineering",
  ["arch control sci"] = "Archives of Control Sciences",
  ["arch corrisp sci italiani"] = "Archivio della Corrispondenza degli Scienziati Italiani",
  ["arch craniofac surg"] = "Archives of craniofacial surgery",
  ["arch cuba"] = "Archivos. Instituto Nacional de Hidrología y Climatología Médicas (Cuba)",
  ["arch cuba cancerol"] = "Archivos cubanos de cancerología",
  ["arch de vecchi anat patol"] = "Archivio de Vecchi per l’Anatomia Patologica e la Medicina Clinica",
  ["arch depress anxiety"] = "Archives of depression and anxiety",
  ["arch derm syphilol"] = "Archives of dermatology and syphilology",
  ["arch dermatol"] = "Archives of dermatology",
  ["arch dermatol forsch"] = "Archiv für dermatologische Forschung",
  ["arch dermatol res"] = "Archives of dermatological research",
  ["arch dermatol syph"] = "Archiv für Dermatologie und Syphilis",
  ["arch dis child"] = "Archives of disease in childhood",
  ["arch dis child educ pract ed"] = "Archives of disease in childhood. Education and practice edition",
  ["arch dis child fetal neonatal ed"] = "Archives of disease in childhood. Fetal and neonatal edition",
  ["arch domin pediatr"] = "Archivos dominicanos de pediatría",
  ["arch drug inf"] = "Archives of drug information",
  ["arch eisenhuttenwes"] = "Archiv fur das Eisenhuttenwesen",
  ["arch electr eng"] = "Archives of Electrical Engineering",
  ["arch elektrotech"] = "Archiv für Elektrotechnik",
  ["arch emerg med"] = "Archives of emergency medicine",
  ["arch emerg med crit care"] = "Archives of emergency medicine and critical care",
  ["arch endocrinol metab"] = "Archives of endocrinology and metabolism",
  ["arch enferm coraz vasos"] = "Archivos de enfermedades del corazón y vasos",
  ["arch entwicklmech org"] = "Archiv für Entwicklungsmechanik der Organismen",
  ["arch environ contam toxicol"] = "Archives of Environmental Contamination and Toxicology",
  ["arch environ health"] = "Archives of environmental health",
  ["arch environ occup health"] = "Archives of environmental & occupational health",
  ["arch epidemiol"] = "Archives of epidemiology",
  ["arch esp arte"] = "Archivo español de arte",
  ["arch esp morfol"] = "Archivo español de morfología",
  ["arch esp urol"] = "Archivos españoles de urología",
  ["arch eur sociol"] = "Archives européennes de sociologie. European journal of sociology. Europäisches Archiv für Soziologie",
  ["arch exp med biol"] = "Archives of Experimental Medicine and Biology",
  ["arch exp pathol pharmakol"] = "Archiv für Experimentalle Pathologie und Pharmakologie",
  ["arch exp veterinarmed"] = "Archiv für experimentelle Veterinärmedizin",
  ["arch facial plast surg"] = "Archives of facial plastic surgery",
  ["arch fam med"] = "Archives of family medicine",
  ["arch farmacol sper sci affin"] = "Archivio di farmacologia sperimentale e scienze affini",
  ["arch farmacol toxicol"] = "Archivos de farmacología y toxicología",
  ["arch fisiol"] = "Archivio di fisiologia",
  ["arch forstwes"] = "Archiv für Forstwesen",
  ["arch foundry eng"] = "Archives of Foundry Engineering",
  ["arch fr mal app dig"] = "Archives françaises des maladies de l'appareil digestif",
  ["arch fr pediatr"] = "Archives françaises de pédiatrie",
  ["arch franciscanum hist"] = "Archivum franciscanum historicum : periodica publicatio trimestris cura pp. Collegii D. Bonaventurae",
  ["arch fund roux ocefa"] = "Archivos de la Fundacion Roux-Ocefa",
  ["arch gen intern med"] = "Archives of general internal medicine",
  ["arch gen psychiatry"] = "Archives of General Psychiatry",
  ["arch genet (zur)"] = "Archiv für Genetik",
  ["arch gerontol geriatr"] = "Archives of gerontology and geriatrics",
  ["arch gerontol geriatr suppl"] = "Archives of gerontology and geriatrics. Supplement",
  ["arch gesamte psychol"] = "Archiv für die gesamte Psychologie",
  ["arch gesamte virusforsch"] = "Archiv für die gesamte Virusforschung",
  ["arch gesch buchwes"] = "Archiv für Geschichte des Buchwesens",
  ["arch gesch naturwiss"] = "Archiv der Geschichte der Naturwissenschaften",
  ["arch gesch oberfranken"] = "Archiv für die Geschichte von Oberfranken",
  ["arch geschwulstforsch"] = "Archiv für Geschwulstforschung",
  ["arch gewerbepathol gewerbehyg"] = "Archiv für Gewerbepathologie und Gewerbehygiene",
  ["arch ginecol obstet"] = "Archivos de ginecología y obstetricia",
  ["arch gynakol"] = "Archiv für Gynäkologie",
  ["arch gynecol"] = "Archives of gynecology",
  ["arch gynecol obstet"] = "Archives of gynecology and obstetrics",
  ["arch hematol blood dis"] = "Archives of hematology and blood diseases",
  ["arch hisp"] = "Archivo hispalense",
  ["arch hist doctrin litt moyen age"] = "Archives d'histoire doctrinale et littéraire du moyen âge",
  ["arch hist exact sci"] = "Archive for History of Exact Sciences",
  ["arch hist filoz med"] = "Archiwum historii i filozofii medycyny",
  ["arch hist med (warsz)"] = "Archiwum historii medycyny",
  ["arch hist med port"] = "Archivos de historia da medicina portugueza",
  ["arch hist pontif"] = "Archivum historiae pontificiae",
  ["arch histol cytol"] = "Archives of histology and cytology",
  ["arch histol jpn"] = "Archivum histologicum Japonicum = Nihon soshikigaku kiroku",
  ["arch histol norm patol"] = "Archivos de histología normal y patológica",
  ["arch hosp (paris)"] = "Archives hospitalières",
  ["arch hosp univ"] = "Archivos. Hospital Universitario General Calixto García",
  ["arch hydrobiol"] = "Archiv für Hydrobiologie",
  ["arch hydrobiol suppl"] = "Archiv für Hydrobiologie, Supplement",
  ["arch hydrobiol suppl algol stud"] = "Algological studies",
  ["arch hyg bakteriol"] = "Archiv für Hygiene und Bakteriologie",
  ["arch iatr epistem"] = "Archeion iatrikōn epistēmōn",
  ["arch iatr hetaireon"] = "Archeia iatrikōn hetaireōn = Archives of the Hellenic medical societies",
  ["arch ibero am"] = "Archivo ibero-americano",
  ["arch immunol ther exp"] = "Archivum Immunolgiae et Therapiae Experimentalis",
  ["arch immunol ther exp (warsz)"] = "Archivum immunologiae et therapiae experimentalis",
  ["arch ind hyg occup med"] = "Archives of industrial hygiene and occupational medicine",
  ["arch inn med"] = "Archiv für innere Medizin",
  ["arch insect biochem physiol"] = "Archives of Insect Biochemistry and Physiology",
  ["arch inst biol andina"] = "Archivos del Instituto de Biologia Andina",
  ["arch inst cardiol mex"] = "Archivos del Instituto de Cardiologia de Mexico",
  ["arch inst farmacol exp (madr)"] = "Archivos del Instituto de Farmacología Experimental (Medicina)",
  ["arch inst hessarek"] = "Archives de l'Institut d'Hessarek",
  ["arch inst pasteur alger"] = "Archives de l'Institut Pasteur d'Algérie. Institut Pasteur d'Algérie",
  ["arch inst pasteur hell"] = "Archives de l'Institut Pasteur hellénique",
  ["arch inst pasteur madagascar"] = "Archives de l’Institut Pasteur de Madagascar",
  ["arch inst pasteur maroc"] = "Archives. Institut Pasteur du Maroc",
  ["arch inst pasteur martinique"] = "Archives de l'Institut Pasteur de la Martinique",
  ["arch inst pasteur tunis"] = "Archives de l’Institut Pasteur de Tunis",
  ["arch int claude bernard"] = "Archives internationales Claude Bernard",
  ["arch int ethnogr"] = "Internationales Archiv für Ethnographie",
  ["arch int hist idees"] = "Archives internationales d'histoire des idées = International archives of the history of ideas",
  ["arch int hist sci"] = "Archives Internationales d’Histoire des Sciences",
  ["arch int hist sci (paris)"] = "Archives internationales d'histoire des sciences",
  ["arch int neurol"] = "Archives internationales de neurologie, des maladies héréditaires, de médecine mentale et psychosomatique",
  ["arch int pharmacodyn ther"] = "Archives internationales de pharmacodynamie et de thérapie",
  ["arch int physiol"] = "Archives internationales de physiologie",
  ["arch int physiol biochim"] = "Archives internationales de physiologie et de biochimie",
  ["arch int physiol biochim biophys"] = "Archives internationales de physiologie, de biochimie et de biophysique",
  ["arch interam rheumatol"] = "Archives of interamerican rheumatology : A.I.R",
  ["arch intern med"] = "Archives of internal medicine",
  ["arch intern med (chic)"] = "Archives of internal medicine (Chicago, Ill. : 1908)",
  ["arch internazionale studi neurol"] = "Archivio internazionale di studi neurologici",
  ["arch invest med (mex)"] = "Archivos de investigación médica",
  ["arch iran med"] = "Archives of Iranian medicine",
  ["arch ital anat embriol"] = "Archivio italiano di anatomia e di embriologia. Italian journal of anatomy and embryology",
  ["arch ital anat istol patol"] = "Archivio italiano di anatomia e istologia patologica",
  ["arch ital biol"] = "Archives italiennes de biologie",
  ["arch ital chir"] = "Archivio italiano di chirurgia",
  ["arch ital dermatol sifilogr venereol"] = "Archivio italiano di dermatologia, sifilografia, e venereologia",
  ["arch ital dermatol venereol sessuol"] = "Archivio italiano di dermatologia, venereologia, e sessuologia",
  ["arch ital laringol"] = "Archivii italiani di laringologia",
  ["arch ital mal appar dig"] = "Archivio italiano delle malattie dell'apparato digerente",
  ["arch ital otol rinol laringol"] = "Archivio italiano di otologia, rinologia e laringologia",
  ["arch ital otol rinol laringol patol cervicofacc"] = "Archivio italiano di otologia, rinologia, laringologia, e patologia cervico-facciale",
  ["arch ital otol rinol laringol patol cervicofacc suppl"] = "Archivio italiano di otologia, rinologia, laringologia e patologia cervico-facciale. Supplemento",
  ["arch ital otol rinol laringol suppl"] = "Archivio italiano di otologia, rinologia e laringologia. Supplemento",
  ["arch ital patol clin tumori"] = "Archivio italiano di patologia e clinica dei tumori",
  ["arch ital pediatr pueric"] = "Archivio italiano di pediatria e puericoltura",
  ["arch ital sci farmacol"] = "Archivio italiano di scienze farmacologiche",
  ["arch ital sci med trop parassitol"] = "Archivio italiano di scienze mediche tropicali e di parassitologia",
  ["arch ital urol"] = "Archivio italiano di urologia",
  ["arch ital urol androl"] = "Archivio italiano di urologia, andrologia : organo ufficiale [di] Società italiana di ecografia urologica e nefrologica",
  ["arch ital urol nefrol"] = "Archivio italiano di urologia e nefrologia",
  ["arch ital urol nefrol androl"] = "Archivio italiano di urologia, nefrologia, andrologia : organo ufficiale dell'Associazione per la ricerca in urologia = Urological, nephrological, and andrological sciences",
  ["arch julius klaus stift vererbungsforsch sozialanthropol rassenhyg"] = "Archiv der Julius Klaus-Stiftung fur Vererbungsforschung, Sozialanthropologie und Rassenhygiene",
  ["arch kinderheilkd"] = "Archiv für Kinderheilkunde",
  ["arch kinderheilkd suppl"] = "Archiv für Kinderheilkunde. Beihefte",
  ["arch klin exp dermatol"] = "Archiv für klinische und experimentelle Dermatologie",
  ["arch klin exp ohren nasen kehlkopfheilkd"] = "Archiv für klinische und experimentelle Ohren- Nasen- und Kehlkopfheilkunde",
  ["arch klin med"] = "Archiv für klinische Medizin",
  ["arch kreislaufforsch"] = "Archiv für Kreislaufforschung",
  ["arch krim"] = "Archiv fur Kriminologie",
  ["arch kriminol"] = "Archiv für Kriminologie",
  ["arch kulturgesch"] = "Archiv für Kultur-geschichte",
  ["arch lagerstattenforsch"] = "Archiv für Lagerstattenforschung",
  ["arch latinoam nutr"] = "Archivos Latinoamericanos de Nutricion",
  ["arch lebensmittelhyg"] = "Archiv für lebensmittelhygiene",
  ["arch mal appar dig mal nutr"] = "Archives des maladies de l'appareil digestif et des maladies de la nutrition",
  ["arch mal coeur vaiss"] = "Archives des maladies du coeur et des vaisseaux",
  ["arch mal coeur vaiss pratique"] = "Archives des maladies du coeur et des vaisseaux. Pratique",
  ["arch mal prof"] = "Archives des maladies professionnelles de médecine du travail et de sécurité sociale",
  ["arch mal prof environ"] = "Archives des Maladies Professionnelles et de l’Environment",
  ["arch mater sci eng"] = "Archives Of Materials Science And Engineering",
  ["arch math"] = "Archiv der Mathematik. Archives of mathematics. Archives mathématiques",
  ["arch math (basel)"] = "Archiv der Mathematik",
  ["arch math (brno)"] = "Universitatis Masarykianae Brunensis. Facultas Scientiarum Naturalium. Archivum Mathematicum",
  ["arch math log"] = "Archive for mathematical logic",
  ["arch math logic"] = "Archive for Mathematical Logic",
  ["arch mech"] = "Archives of Mechanics",
  ["arch mech (arch mech stos)"] = "Polish Academy of Sciences. Institute of Fundamental Technological Research. Archives of Mechanics (Archiwum Mechaniki Stosowanej)",
  ["arch med"] = "Archivalia médica",
  ["arch med (oviedo)"] = "Archives of medicine",
  ["arch med case rep"] = "Archives of medical case reports",
  ["arch med cuba"] = "Archivos médicos de Cuba",
  ["arch med deporte"] = "Archivos de medicina del deporte : publicación de la Federación Española de Medicina del Deporte",
  ["arch med exp"] = "Archivos de medicina experimental; trabajos del Instituto Nacional de Ciencias Médicas",
  ["arch med gen trop"] = "Archives de médecine générale et tropicale",
  ["arch med infant"] = "Archivos de medicina infantil",
  ["arch med intern"] = "Archivos de medicina interna (Montevideo, Uruguay)",
  ["arch med interna"] = "Archivio di medicina interna",
  ["arch med mutual"] = "Archivio di medicina mutualistica",
  ["arch med ouest"] = "Archives médicales de l'Ouest: Angers/Nantes/Rennes",
  ["arch med panamenos"] = "Archivos médicos panameños",
  ["arch med res"] = "Archives of medical research",
  ["arch med sadowej kryminol"] = "Archiwum medycyny sa̧dowej i kryminologii",
  ["arch med sci"] = "Archives of medical science : AMS",
  ["arch med sci atheroscler dis"] = "Archives of medical sciences. Atherosclerotic diseases",
  ["arch med soc"] = "Archives de médecine sociale",
  ["arch med vet"] = "Archivos de medicina veterinaria",
  ["arch medicos mex"] = "Archivos médicos mexicanos",
  ["arch mediterr med"] = "Archives méditerranéennes de médecine",
  ["arch metall mater"] = "Archives of Metallurgy and Materials",
  ["arch meteor geophys bioklimatol"] = "Archiv für Meteorologie, Geophysik und Bioklimatologie",
  ["arch meteorol geophys bioklimatol [b]"] = "Archiv fur Meteorologie, Geophysik und Bioklimatologie. Serie B: Klimatologie, Umweltmeteorologie, Strahlungsforschung",
  ["arch meteorol geophys bioklimatol a"] = "Archiv für Meteorologie, Geophysik, und Bioklimatologie Serie A Meteorologie und Geophysik",
  ["arch meteorol geophys bioklimatol b"] = "Archiv für Meteorologie, Geophysik und Bioklimatologie. Serie B: Klimatologie, Umweltmeteorologie, Strahlungsforschung",
  ["arch meteorol geophys bioklimatol, ser b"] = "Archiv für Meteorologie, Geophysik und Bioklimatologie, Serie B",
  ["arch mex venereol dermatol"] = "Archivos mexicanos de venereología y dermatología",
  ["arch microbiol"] = "Archives of Microbiology",
  ["arch middx hosp"] = "Archives. Middlesex Hospital",
  ["arch mikrobiol"] = "Archiv für Mikrobiologie",
  ["arch mikrosk anat"] = "Archiv fur mikroskopische Anatomie",
  ["arch mikrosk anat enwicklmech"] = "Archiv für mikroskopische Anatomie und Entwicklungsmechanik",
  ["arch min sci"] = "Archives of Mining Sciences",
  ["arch mold"] = "Archiva Moldaviæ",
  ["arch molluskenkd"] = "Archiv fur Molluskenkunde",
  ["arch molluskenkd 2004"] = "Archiv für Molluskenkunde",
  ["arch monaldi"] = "Archivio Monaldi per la tisiologia e le malattie dell'apparato respiratorio",
  ["arch monaldi mal torace"] = "Archivio Monaldi per le malattie del torace",
  ["arch nat conserv landscape res"] = "Archives of Nature Conservation and Landscape Research",
  ["arch nat hist"] = "Archives of natural history",
  ["arch natschutz landschforsch"] = "Archiv für Naturschutz und Landschaftsforschung",
  ["arch neerl physiol homme anim"] = "Archives néerlandaises de physiologie de l'homme et des animaux",
  ["arch neerl sci exactes nat"] = "Archives Neerlandaises des Sciences Exactes et Naturelles",
  ["arch neerl sci exactes nat ser ii"] = "Archives Neerlandaises des Sciences Exactes et Naturelles. Serie II",
  ["arch neurobiol (madr)"] = "Archivos de neurobiologiá",
  ["arch neurocir"] = "Archivos de neurocirugía",
  ["arch neurol"] = "Archives of Neurology",
  ["arch neurol neuro disord"] = "Archives of neurology and neuro disorders",
  ["arch neurol neurosci"] = "Archives in neurology & neuroscience",
  ["arch neurol psiquiatr mex"] = "Archivos de neurología y psiquiatría de México : organo de la Sociedad mexicana de Neurología y Psiquiatría",
  ["arch neurol psychiatry"] = "Archives of neurology and psychiatry",
  ["arch neurosci"] = "Archives of neuroscience",
  ["arch obstet gynaecol"] = "Archives of obstetrics and gynaecology",
  ["arch odonto estomatol"] = "Archivos de odonto estomatologia",
  ["arch oftalmol b aires"] = "Archivos de oftalmología de Buenos Aires",
  ["arch ohren nasen kehlkopfheilkd"] = "Archiv für Ohren-, Nasen- und Kehlkopfheilkunde",
  ["arch oncol"] = "Archive of oncology",
  ["arch ophtalmol (paris)"] = "Archives d'ophtalmologie",
  ["arch ophtalmol rev gen ophtalmol"] = "Archives d'ophtalmologie et revue générale d'ophtalmologie",
  ["arch ophthal"] = "Archives of ophthalmology (Chicago, Ill. : 1929)",
  ["arch ophthalmol"] = "Archives of Ophthalmology",
  ["arch oral biol"] = "Archives of oral biology",
  ["arch orient"] = "Archiv orientální",
  ["arch orientforsch"] = "Archiv für Orientforschung",
  ["arch orthop rheumatol"] = "Archives of orthopedics and rheumatology",
  ["arch orthop trauma surg"] = "Archives of orthopaedic and trauma surgery",
  ["arch orthop unfallchir"] = "Archiv für orthopädische und Unfall-Chirurgie",
  ["arch ortop"] = "Archivio di ortopedia",
  ["arch osp mare"] = "Archivio - Ospedale al mare",
  ["arch osteoporos"] = "Archives of osteoporosis",
  ["arch ostet ginecol"] = "Archivio di ostetricia e ginecologia",
  ["arch otolaryngol"] = "Archives of otolaryngology (Chicago, Ill. : 1960)",
  ["arch otolaryngol head neck surg"] = "Archives of Otolaryngology – Head and Neck Surgery",
  ["arch otorhinolaryngol"] = "Archives of oto-rhino-laryngology",
  ["arch otorhinolaryngol suppl"] = "Archives of oto-rhino-laryngology. Supplement = Archiv für Ohren-, Nasen- und Kehlkopfheilkunde. Supplement",
  ["arch ottamol"] = "Archivio di ottalmologia",
  ["arch ottoman"] = "Archivum Ottomanicum",
  ["arch p amer ant asso"] = "Archeological Papers of The American Anthropological Association",
  ["arch palliat care med"] = "Archives of palliative care and medicine",
  ["arch panama city"] = "Archivos. Hospital Santo Tomás (Panama, Panama)",
  ["arch papyrusforsch"] = "Archiv für Papyrusforschung und verwandte Gebiete",
  ["arch papyrusforsch verw geb"] = "Archiv fur Papyrusforschung und Verwandte Gebiete",
  ["arch pathol"] = "Archives of Pathology",
  ["arch pathol (chic)"] = "Archives of pathology",
  ["arch pathol lab med"] = "Archives of pathology & laboratory medicine",
  ["arch patol clin med"] = "Archivio di patologia e clinica medica",
  ["arch pediatr"] = "Archives of pediatrics",
  ["arch pediatr (barc)"] = "Archivos de pediatría",
  ["arch pediatr adolesc med"] = "Archives of pediatrics & adolescent medicine",
  ["arch pediatr infect dis"] = "Archives of pediatric infectious diseases",
  ["arch pediatr urug"] = "Archivos de pediatría del Uruguay",
  ["arch peru patol clin"] = "Archivos peruanos de patología y clínica",
  ["arch pharm"] = "Archiv der Pharmazie",
  ["arch pharm (weinheim)"] = "Archiv der Pharmazie",
  ["arch pharm (weinheim, ger)"] = "Archiv der Pharmazie (Weinheim, Germany)",
  ["arch pharm ber dtsch pharm ges"] = "Archiv der Pharmazie und Berichte der Deutschen Pharmazeutischen Gesellschaft",
  ["arch pharm chem (kbh)"] = "Archiv for pharmaci og chemi",
  ["arch pharm pract"] = "Archives of pharmacy practice",
  ["arch pharm res"] = "Archives of pharmacal research",
  ["arch pharmacal res"] = "Archives of Pharmacal Research",
  ["arch pharmacol ther"] = "Archives of pharmacology and therapeutics",
  ["arch philos"] = "Archives de philosophie",
  ["arch phys med"] = "Archives of physical medicine",
  ["arch phys med rehabil"] = "Archives of physical medicine and rehabilitation",
  ["arch phys ther"] = "Archives of physical therapy",
  ["arch phys ther (leipz)"] = "Archiv für physikalische Therapie",
  ["arch physiol biochem"] = "Archives of Physiology and Biochemistry",
  ["arch physiother"] = "Archives of physiotherapy",
  ["arch phytopathol pflanzenschutz"] = "Archiv für Phytopathologie und Pflanzenschutz",
  ["arch phytopathol plant prot"] = "Archives of Phytopathology and Plant Protection",
  ["arch plast surg"] = "Archives of plastic surgery",
  ["arch prev riesgos labor"] = "Archivos de prevención de riesgos laborales",
  ["arch proteom bioinform"] = "Archives of proteomics and bioinformatics",
  ["arch protistenkd"] = "Archiv fur Protistenkunde",
  ["arch protistenkunde"] = "Archiv für Protistenkunde",
  ["arch psicol neurol psichiatr"] = "Archivio di psicologia, neurologia e psichiatria",
  ["arch psychiatr nervenkr"] = "Archiv für Psychiatrie und Nervenkrankheiten",
  ["arch psychiatr nervenkr z gesamte neurol psychiatr"] = "Archiv für Psychiatrie und Nervenkrankheiten, vereinigt mit Zeitschrift für die gesamte Neurologie und Psychiatrie",
  ["arch psychiatr nurs"] = "Archives of psychiatric nursing",
  ["arch psychol (chic)"] = "Archives of psychology (Chicago, Ill.)",
  ["arch psychol (frankf)"] = "Archiv für Psychologie",
  ["arch psychol (geneve)"] = "Archives de psychologie",
  ["arch psychol relig"] = "Archive for the psychology of religion = Archiv für Religionspsychologie",
  ["arch public health"] = "Archives of public health = Archives belges de santé publique",
  ["arch putti chir organi mov"] = "Archivio Putti di Chirurgia degli Organi di Movimento",
  ["arch radiol"] = "Archivio di radiologia",
  ["arch ration mech anal"] = "Archive for Rational Mechanics and Analysis",
  ["arch rational mech anal"] = "Archive for Rational Mechanics and Analysis",
  ["arch razi inst"] = "Archives of Razi Institute",
  ["arch rec (abingdon)"] = "Archives and records (Abingdon, England)",
  ["arch reformation hist"] = "Archiv für Reformationsgeschichte",
  ["arch rehabil res clin transl"] = "Archives of rehabilitation research and clinical translation",
  ["arch report"] = "Archive reporter. Data Archive on Adolescent Pregnancy and Pregnancy Prevention",
  ["arch rheumatol"] = "Archives of rheumatology",
  ["arch rhumatol"] = "Archives de rhumatologie",
  ["arch roum pathol exp microbiol"] = "Archives roumaines de pathologie expérimentales et de microbiologie",
  ["arch sci"] = "Archives des sciences",
  ["arch sci biol (bologna)"] = "Archivio di scienze biologiche",
  ["arch sci compte rendu seances soc"] = "Archives des sciences et compte rendu des séances de la Société",
  ["arch sci med (torino)"] = "Archivio per le scienze mediche",
  ["arch sci physiol (paris)"] = "Archives des sciences physiologiques",
  ["arch sci psychol"] = "Archives of scientific psychology",
  ["arch sci soc relig"] = "Archives de sciences sociales des religions",
  ["arch sex behav"] = "Archives of sexual behavior",
  ["arch sicil med chir 4 acta med mediterr"] = "Archivio siciliano di medicina e chirurgia. 4, Acta medica mediterranea",
  ["arch soc agent anat norm patol"] = "Archivos. Sociedad Argentina de Anatomía Normal y Patológica",
  ["arch soc biol montev"] = "Archivos. Sociedad de Biología de Montevideo",
  ["arch soc cir chile"] = "Archivos de la Sociedad de Cirujanos de Chile",
  ["arch soc cir hosp"] = "Archivos. Sociedad de Cirujanos de Hospital, Santiago de Chile",
  ["arch soc esp oftalmol"] = "Archivos de la Sociedad Española de Oftalmología",
  ["arch soc estud clin habana"] = "Archivos. Sociedad de Estudios Clínicos de la Habana",
  ["arch soc rom stor patria"] = "Archivio della Società romana di storia patria",
  ["arch soc zool bot fenn vanamo"] = "Archivum Societatis Zoologicae Botanicae Fennicae Vanamo",
  ["arch sozialgesch"] = "Archiv für Sozialgeschichte",
  ["arch std hiv res"] = "Archives of STD/HIV research",
  ["arch stem cell res"] = "Archives of stem cell research",
  ["arch stem cell ther"] = "Archives of stem cell and therapy",
  ["arch stomatol"] = "Archives de stomatologie",
  ["arch stomatol (napoli)"] = "Archivio stomatologico",
  ["arch stor ital"] = "Archivio storico italiano",
  ["arch stor lomb"] = "Archivio storico lombardo",
  ["arch stor prov napol"] = "Archivio storico per le province napoletane",
  ["arch stor prov parm"] = "Archivio storico per le province Parmensi",
  ["arch stor sicil"] = "Archivio storico siciliano (Palermo, Italy : 1946)",
  ["arch suicide res"] = "Archives of suicide research : official journal of the International Academy for Suicide Research",
  ["arch suisses anthropol gen"] = "Archives suisses d'anthropologie générale",
  ["arch surg"] = "Archives of surgery (Chicago, Ill. : 1960)",
  ["arch tierernahr"] = "Archiv für Tierernährung",
  ["arch tierz"] = "Archiv für Tierzucht",
  ["arch tisiol"] = "Archivio di tisiologia",
  ["arch tisiol mal appar respir"] = "Archivio di tisiologia e delle malattie dell'apparato respiratorio",
  ["arch toxicol"] = "Archives of Toxicology",
  ["arch toxicol suppl"] = "Archives of Toxicology. Supplement",
  ["arch toxikol"] = "Archiv für Toxikologie",
  ["arch trauma res"] = "Archives of trauma research",
  ["arch urkundenforsch"] = "Archiv fur Urkundenforschung",
  ["arch urol res"] = "Archive of urological research",
  ["arch urug med cir espec"] = "Archivos uruguayos de medicina, cirugía y especialidades",
  ["arch veneto"] = "Archivio veneto (1927)",
  ["arch venez med trop parasitol med"] = "Archivos venezolanos de medicina tropical y parasitología médica",
  ["arch venez pueric pediatr"] = "Archivos venezolanos de puericultura y pediatría",
  ["arch vet ital"] = "Archivio veterinario italiano",
  ["arch vet pol"] = "Archivum veterinarium Polonicum",
  ["arch virol"] = "Archives of Virology",
  ["arch virol suppl"] = "Archives of virology. Supplementum",
  ["arch wiss"] = "Archiv und Wissenschaft; Schriftenreihe der Archivalischen Zeitschrift",
  ["arch womens ment health"] = "Archives of women's mental health",
  ["arch z"] = "Archivalische Zeitschrift",
  ["archaeol aeliana"] = "Archaeologia aeliana, or, Miscellaneous tracts relating to antiquity",
  ["archaeol anthropol sci"] = "Archaeological and anthropological sciences",
  ["archaeol anz"] = "Archäologischer Anzeiger",
  ["archaeol balt"] = "Archaeologica Baltica",
  ["archaeol cantiana"] = "Archaeologia cantiana",
  ["archaeol dialogues"] = "Archaeological Dialogues",
  ["archaeol j"] = "Archaeological Journal",
  ["archaeol nachrichtenbl"] = "Archaeologisches Nachrichtenblatt",
  ["archaeol oceania"] = "Archaeology in Oceania",
  ["archaeol pap am anthropol assoc"] = "Archaeological Papers of the American Anthropological Association",
  ["archaeol prospect"] = "Archaeological prospection",
  ["archaeol rep"] = "Archaeological Reports",
  ["archaeol res asia"] = "Archaeological Research in Asia",
  ["archeol pap am anthr"] = "Housework: Craft Production and Domestic Economy in Ancient Mesoamerica",
  ["archeosci-rev archeom"] = "ArcheoSciences-Revue d Archeometrie",
  ["archief meded k zeeuwsch genoot wet"] = "Archief. Mededelingen van het Koninklijk zeeuwsch genootschap der wetenschappen. Zeeuwsch genootschap der wetenschappen",
  ["archimedes palimpsest publ"] = "The Archimedes Palimpsest Publications",
  ["archit aujourdhui"] = "AA : l'architecture d'aujourd'hui",
  ["archit cult"] = "Architecture and Culture",
  ["archit eng des manage"] = "Architectural Engineering and Design Management",
  ["archit intell"] = "Architectural Intelligence",
  ["archit rec"] = "Architectural record",
  ["archit sci rev"] = "Architectural science review",
  ["archit struct construct"] = "Architecture, Structures and Construction",
  ["archit theory rev"] = "Architectural Theory Review",
  ["archiv hibernicum"] = "Archivium hibernicum : or, Irish historical records",
  ["archiv math log"] = "Archive for Mathematical Logic",
  ["archives econ hist"] = "Archives of Economic History",
  ["archivio chir torace"] = "Archivio di chirurgia del torace",
  ["archivos soc oftalmol hisp am"] = "Archivos de la Sociedad Oftalmológica Hispano-Americana",
  ["archäol schweiz"] = "Archäologie der Schweiz",
  ["archéol neuchâtel"] = "Archéologie neuchâteloise",
  ["arcisp s anna ferrara"] = "L' Arcispedale S. Anna di Ferrara",
  ["arct alp res"] = "Arctic and Alpine Research",
  ["arct antarct alp res"] = "Arctic, antarctic, and alpine research",
  ["arct anthropol"] = "Arctic Anthropology",
  ["arct environ res"] = "Arctic environmental research",
  ["arct sci"] = "Arctic Science",
  ["arct, antarc, alp res"] = "Arctic, Antarctic, and Alpine Research",
  ["arctic anthropol"] = "Arctic anthropology",
  ["arctic med res"] = "Arctic medical research",
  ["area (oxf)"] = "Area (Oxford, England)",
  ["arealstat schweiz"] = "Arealstatistik der Schweiz",
  ["argent repub laws statut"] = "Argentine Republic. Laws, statutes, etc",
  ["argum libr"] = "Argumentation Library",
  ["arh farm (belgr)"] = "Arhiv za farmaciju",
  ["arh hig rada"] = "Arhiv za higijenu rada",
  ["arh hig rada toksikol"] = "Arhiv za higijenu rada i toksikologiju",
  ["arh zast"] = "Arhiv za zas̆titu majke i djeteta. Archives of mother and child welfare. Archives pour la protection de la mère et de l'enfant",
  ["arid ecosyst"] = "Arid Ecosystems",
  ["arid land res manage"] = "Arid Land Research and Management",
  ["ariel (jerus)"] = "Ariel. Ariʼel",
  ["arima rev afr rech inform math appl"] = "ARIMA. Revue Africaine de la Recherche en Informatique et Mathematiques Appliquees",
  ["aristote trad etudes"] = "Aristote. Traductions et Etudes",
  ["ariz dent j"] = "The Arizona dental journal",
  ["ariz geriatr soc j"] = "Arizona Geriatrics Society journal",
  ["ariz j int comp law"] = "Arizona journal of international and comparative law",
  ["ariz law rev"] = "Arizona law review",
  ["ariz med"] = "Arizona medicine",
  ["ariz nurse"] = "The Arizona nurse",
  ["ariz revis statut"] = "Arizona revised statutes. Arizona",
  ["ariz revis statut annot ariz"] = "Arizona revised statutes, annotated : prepared under legislative authority, Laws 1956, chapter 129. Arizona",
  ["ariz state law j"] = "Arizona State law journal",
  ["ariz west"] = "Arizona and the West",
  ["arizona j environ law policy"] = "Arizona journal of environmental law & policy",
  ["ark archeol"] = "Arkansas archeology",
  ["ark dent"] = "Arkansas dentistry",
  ["ark dent j"] = "The Arkansas dental journal",
  ["ark fys"] = "Arkiv foer Physik",
  ["ark geofys"] = "Arkiv foer Geohysik",
  ["ark hist q"] = "The Arkansas historical quarterly",
  ["ark kemi mineral geol"] = "Arkiv foer Kemi, Mineralogi och Geologi",
  ["ark law rev"] = "Arkansas law review",
  ["ark light newsl"] = "The Ark-light newsletter",
  ["ark mat"] = "Arkiv for Matematik",
  ["ark mat astron fys"] = "Arkiv foer Matematik, Astronomi, och Fysik",
  ["arkh anat gistol embriol"] = "Arkhiv anatomii, gistologii i émbriologii",
  ["arkh patol"] = "Arkhiv patologii",
  ["arkivoc (gainesville, fl, u s)"] = "ARKIVOC (Gainesville, FL, United States)",
  ["armed forces med j india"] = "Armed Forces medical journal, India",
  ["armed forces soc"] = "Armed forces and society",
  ["armen j math"] = "Armenian Journal of Mathematics",
  ["armen rev"] = "The Armenian review",
  ["army lawyer"] = "The Army lawyer",
  ["army materials tech conf ser"] = "Army Materials Technology Conference Series",
  ["army med corps j india"] = "Army Medical Corps journal, India",
  ["army med dep bull"] = "Army Medical Department bulletin",
  ["army q def j"] = "The Army quarterly and defence journal",
  ["arn j"] = "ARN journal : the official journal of the Association of Rehabilitation Nurses",
  ["arne ryde meml lect ser"] = "Arne Ryde Memorial Lectures Series",
  ["arnold appl stat"] = "Arnold Applications of Statistics",
  ["arnold math j"] = "Arnold Mathematical Journal",
  ["arq bras cardiol"] = "Arquivos brasileiros de cardiologia",
  ["arq bras cir dig"] = "Arquivos brasileiros de cirurgia digestiva : ABCD = Brazilian archives of digestive surgery",
  ["arq bras endocrinol metabol"] = "Arquivos brasileiros de endocrinologia e metabologia",
  ["arq bras med"] = "Arquivos brasileiros de medicina",
  ["arq bras med nav"] = "Arquivos brasileiros de medicina naval",
  ["arq bras med vet zootec"] = "Arquivo brasileiro de medicina veterinária e zootecnia",
  ["arq bras oftalmol"] = "Arquivos brasileiros de oftalmologia",
  ["arq bras psicol"] = "Arquivos brasileiros de psicologia",
  ["arq cent cult port"] = "Arquivos do Centro Cultural Português. Centre culturel portugais",
  ["arq cent estud curso odontol"] = "Arquivos do Centro de Estudos do curso de odontologia",
  ["arq cent estud curso odontol univ fed minas gerais"] = "Arquivos do Centro de Estudos, Curso de Odontologia Universidade Federal de Minas Gerais",
  ["arq cent estud fac odontol ufmg (belo horiz)"] = "Arquivos do Centro de Estudos da Faculdade de Odontologia da U. F. M. G",
  ["arq cir clin exp"] = "Arquivos de cirurgia clinica e experimental",
  ["arq clin (rio de j)"] = "Arquivos de clínica",
  ["arq fac hig saude publica univ sao paulo"] = "Arquivos da Faculdade de Higiene e Saude Publica da Universidade de Sao Paulo",
  ["arq gastroenterol"] = "Arquivos de gastroenterologia",
  ["arq hig saude publica"] = "Arquivos de higiene e saúde pública",
  ["arq inst biol (sao paulo)"] = "Arquivos do Instituto Biologico",
  ["arq inst bras invest tubers"] = "Arquivos. Instituto Brasileiro para Investigação da Tuberculose",
  ["arq inst farm"] = "Arquivos. Coimbra. Universidade. Instituto de Farmacologia e Terapêutica Experimental",
  ["arq med"] = "Arquivos de medicina : revista de ciência e arte médicas",
  ["arq med cirur pernamb"] = "Arquivos de medicina e cirurgia de Pernambuco",
  ["arq min leprol"] = "Arquivos mineiros de leprologia",
  ["arq mus nac"] = "Arquivos do Museu Nacional. Museu Nacional (Brazil)",
  ["arq neuropsiquiatr"] = "Arquivos de neuro-psiquiatria",
  ["arq oncol"] = "Arquivos de oncología",
  ["arq patol"] = "Arquivo de patologia",
  ["arq polic civ sao paul"] = "Arquivos da polícia civil de São Paulo",
  ["arq port bioquim"] = "Arquivos portugueses de bioquímica",
  ["arq port oftalmol"] = "Arquivos portugueses de oftalmologia",
  ["arq rio gd sul braz state"] = "Arquivos. Rio Grande do Sul (Brazil : State). Departamento Estadual de Saúde",
  ["arq zool"] = "Arquivos de zoologia",
  ["arqueol mex"] = "Arqueologia Mexicana",
  ["arquivo inst"] = "Arquivo do Instituto Gulbenkian de Ciência",
  ["arrhythm electrophysiol rev"] = "Arrhythmia & electrophysiology review",
  ["arrhythmia electrophysiol rev"] = "Arrhythmia & Electrophysiology Review",
  ["arrows change"] = "Arrows for change",
  ["ars combin"] = "Ars Combinatoria",
  ["ars curandi odontol"] = "Ars curandi em odontologia",
  ["ars inven anal"] = "Ars Inveniendi Analytica",
  ["ars math contemp"] = "Ars Mathematica Contemporanea",
  ["ars med"] = "Ars medici",
  ["ars med rev int med prat"] = "Ars medici; revue internationale pour le médecin praticien",
  ["arsb goteb tandlak sallsk"] = "Årsbok. Göteborgs tandläkare-sällskap",
  ["arsb odontol samf finl"] = "Årsbok. Odontologiska samfundet i Finland",
  ["arsberet kbh univ med hist inst mus"] = "Aarsberetning - Københavns universitets medicinsk-historiske institut og museum",
  ["arsberet kobenhavns univ med hist mus"] = "Aarsberetning - Københavns universitets medicinsk-historiske museum",
  ["arsskr sydven medicinhist sallsskapet"] = "Aarsskrift - Sydsvenska medicinhistoriska sällskapet",
  ["art am"] = "Art in America",
  ["art bull"] = "The Art bulletin",
  ["art discrete appl math"] = "The Art of Discrete and Applied Mathematics",
  ["art hist"] = "Art history",
  ["art med"] = "The Art of medication",
  ["art ther (alex)"] = "Art therapy : journal of the American Art Therapy Association",
  ["artech house antenna lib"] = "Artech House Antenna Library",
  ["artech house microw lib"] = "The Artech House Microwave Library",
  ["artech house signal process libr"] = "The Artech House Signal Processing Library",
  ["artech house telecommun libr"] = "Artech House Telecommunications Library",
  ["arteres veines"] = "Artères et veines",
  ["arterioscler thromb"] = "Arteriosclerosis and thrombosis : a journal of vascular biology",
  ["arterioscler thromb vasc biol"] = "Arteriosclerosis, thrombosis, and vascular biology",
  ["arterioscler, thromb, vasc biol"] = "Arteriosclerosis, Thrombosis, and Vascular Biology",
  ["artery res"] = "Artery research",
  ["artes mex"] = "Artes de México",
  ["artha vijnana"] = "Artha vijñāna : journal of the Gokhale Institute of Politics and Economics, Poona (India) = Artha vijñāna : Gokhale Artha Śāstra Samsthā Dvārā Prakāśita Traimēsika Pātrikā",
  ["arthritis care res"] = "Arthritis care and research : the official journal of the Arthritis Health Professions Association",
  ["arthritis care res (hoboken)"] = "Arthritis care & research",
  ["arthritis res"] = "Arthritis research",
  ["arthritis res ther"] = "Arthritis research & therapy",
  ["arthritis rheum"] = "Arthritis and rheumatism",
  ["arthritis rheum (munch)"] = "Arthritis + Rheuma",
  ["arthritis rheumatol"] = "Arthritis & rheumatology (Hoboken, N.J.)",
  ["arthroplast today"] = "Arthroplasty today",
  ["arthropod manage tests"] = "Arthropod Management Tests",
  ["arthropod plant interact"] = "Arthropod-plant interactions",
  ["arthropod struct dev"] = "Arthropod structure & development",
  ["arthropod syst phylog"] = "Arthropod Systematics & Phylogeny",
  ["arthropod syst phylogeny"] = "Arthropod systematics & phylogeny",
  ["arthropod-plant interact"] = "Arthropod-Plant Interactions",
  ["arthropoda sel"] = "Arthropoda selecta",
  ["arthrosc sports med rehabil"] = "Arthroscopy, sports medicine, and rehabilitation",
  ["arthrosc tech"] = "Arthroscopy techniques",
  ["articulator (columb)"] = "Articulator",
  ["articulator (syd)"] = "The Articulator",
  ["artif cells blood substit immobil biotechnol"] = "Artificial cells, blood substitutes, and immobilization biotechnology",
  ["artif cells nanomed biotechnol"] = "Artificial cells, nanomedicine, and biotechnology",
  ["artif cells, blood substitues, immobilization biotechnol"] = "Artificial Cells Blood Substitutes and Immobilization Biotechnology",
  ["artif dna pna xna"] = "Artificial DNA, PNA & XNA",
  ["artif intell"] = "Artificial intelligence",
  ["artif intell agric"] = "Artificial Intelligence in Agriculture",
  ["artif intell appl"] = "Artificial intelligence and applications (Commerce, Calif.)",
  ["artif intell eng des anal manuf"] = "Artificial intelligence for engineering design, analysis and manufacturing : AI EDAM",
  ["artif intell found theory algorithms"] = "Artificial Intelligence: Foundations, Theory, and Algorithms",
  ["artif intell geosci"] = "Artificial Intelligence in Geosciences",
  ["artif intell health (2018)"] = "Artificial intelligence in health : first International Workshop, AIH 2018, Stockholm, Sweden, July 13-14, 2018, Revised selected papers. AIH (Workshop) (1st : 2018 : Stockholm, Sweden)",
  ["artif intell law"] = "Artificial Intelligence and Law",
  ["artif intell law (dordr)"] = "Artificial intelligence and law",
  ["artif intell life sci"] = "Artificial Intelligence in the Life Sciences",
  ["artif intell med"] = "Artificial intelligence in medicine",
  ["artif intell med (2015)"] = "Artificial intelligence in medicine : 15th Conference on Artificial Intelligence in Medicine, AIME 2015, Pavia, Italy, June 17-20, 2015 : proceedings. Conference on Artificial Intelligence in Medicine (2005-) (15th : 2015 : Pavia, Italy)",
  ["artif intell med (2017)"] = "Artificial intelligence in medicine : 16th Conference on Artificial Intelligence in Medicine, AIME 2017, Vienna, Austria, June 21-24, 2017 : proceedings. Conference on Artificial Intelligence in Medicine (2005-) (16th : 2017 : Vienna, Austria)",
  ["artif intell med conf artif intell med (2005-)"] = "Artificial intelligence in medicine. Conference on Artificial Intelligence in Medicine (2005-)",
  ["artif intell res"] = "Artificial intelligence research",
  ["artif intell rev"] = "Artificial intelligence review",
  ["artif intell soft comp ind transform"] = "Artificial Intelligence and Soft Computing for Industrial Transformation",
  ["artif intell surg"] = "Artificial Intelligence Surgery",
  ["artif life"] = "Artificial life",
  ["artif life rob"] = "Artificial Life and Robotics",
  ["artif life robot"] = "Artificial life and robotics",
  ["artif limbs"] = "Artificial limbs",
  ["artif organs"] = "Artificial organs",
  ["artif satell"] = "Artifical Satellites",
  ["artif: dna: pna xna"] = "Artificial DNA: PNA & XNA",
  ["arts health"] = "Arts & health",
  ["arts psychother"] = "The Arts in psychotherapy",
  ["arukoru kenkyuto yakubutsu ison"] = "Arukōru kenkyū to yakubutsu izon = Japanese journal of alcohol studies & drug dependence",
  ["arx misc zool"] = "Arxius de miscel·lània zoològica",
  ["arxivorg, e-print arch astrophys"] = "arXiv.org, e-Print Archive Astrophysics",
  ["arxivorg, e-print arch condens matter"] = "arXiv.org, e-Print Archive Condensed Matter",
  ["arxivorg, e-print arch gen relativ quantum cosmol"] = "arXiv.org, e-Print Archive General Relativity and Quantum Cosmology",
  ["arxivorg, e-print arch high energy phys exp"] = "arXiv.org, e-Print Archive High Energy Physics - Experimental",
  ["arxivorg, e-print arch high energy phys lattice"] = "arXiv.org, e-Print Archive High Energy Physics - Lattice",
  ["arxivorg, e-print arch high energy phys phenomenol"] = "arXiv.org, e-Print Archive High Energy Physics - Phenomenology",
  ["arxivorg, e-print arch high energy phystheor"] = "arXiv.org, e-Print Archive High Energy Physics - Theory",
  ["arxivorg, e-print arch nucl exp"] = "arXiv.org, e-Print Archive Nuclear Experiment",
  ["arxivorg, e-print arch nucl theory"] = "arXiv.org, e-Print Archive Nuclear Theory",
  ["arxivorg, e-print arch phys"] = "arXiv.org, e-Print Archive Physics",
  ["arxivorg, e-print arch quant biol"] = "arXiv.org, e-Print Archive Quantitative Biology",
  ["arxivorg, e-print arch quantum phys"] = "arXiv.org, e-Print Archive Quantum Physics",
  ["arya atheroscler"] = "ARYA atherosclerosis",
  ["arzneim forsch"] = "Arzneimittel Forschung",
  ["arzt christ"] = "Arzt und Christ",
  ["arztebl rheinl pfalz"] = "Aerzteblatt Rheinland-Pfalz",
  ["arztl dienst"] = "Der Ärztliche Dienst",
  ["arztl forsch"] = "Arztliche Forschung",
  ["arztl fortbild"] = "Die Ärztliche Fortbildung",
  ["arztl jugendkd"] = "Arztliche Jugendkunde",
  ["arztl kosmetol"] = "Ärztliche Kosmetologie = Medical cosmetology",
  ["arztl mitt"] = "Ärztliche Mitteilungen : nebst Anzeiger und wissenschaftlicher Beilage",
  ["arztl monatshefte berufl fortbild"] = "Ärztliche Monatshefte für berufliche Fortbildung. Cahiers mensuels de médecine",
  ["arztl prax"] = "Aerztliche Praxis",
  ["arztl sammelbl"] = "Ärztliche Sammelblätter",
  ["arztl wochensch"] = "Ärztliche Wochenschrift",
  ["asa decen conf ser"] = "Future of Anthropological Knowledge",
  ["asa newsl"] = "ASA newsletter",
  ["asa-crc ser stat reason sci soc"] = "ASA-CRC Series on Statistical Reasoning in Science and Society",
  ["asa-siam ser stat appl probab"] = "ASA-SIAM Series on Statistics and Applied Probability",
  ["asaio j"] = "ASAIO journal (American Society for Artificial Internal Organs : 1992)",
  ["asaio trans"] = "ASAIO transactions",
  ["asbury theol j"] = "The Asbury theological journal",
  ["asce asme j risk uncertain eng syst a civ eng"] = "ASCE-ASME journal of risk and uncertainty in engineering systems. Part A, Civil Engineering",
  ["asce open"] = "ASCE OPEN",
  ["asce-asme j risk uncertainty eng syst part a: civ eng"] = "ASCE-ASME Journal of Risk and Uncertainty in Engineering Systems, Part A: Civil Engineering",
  ["asce-asme j risk uncertainty eng syst part b: mech eng"] = "ASCE-ASME Journal of Risk and Uncertainty in Engineering Systems, Part B: Mechanical Engineering",
  ["aschkenas (wien)"] = "Aschkenas",
  ["asd tech rep"] = "ASD technical report. United States. Air Force. Systems Command. Aeronautical Systems Division",
  ["asda news"] = "ASDA news (American Student Dental Association : 1981)",
  ["asdc j dent child"] = "ASDC journal of dentistry for children",
  ["asean econ bull"] = "ASEAN economic bulletin",
  ["asean heart j"] = "ASEAN heart journal : Official journal of the ASEAN Federation of Cardiology",
  ["asean j open distance learn"] = "ASEAN journal of Open and Distance Learning",
  ["asgsb bull"] = "ASGSB bulletin : publication of the American Society for Gravitational and Space Biology",
  ["asha lead"] = "ASHA leader",
  ["asha monogr"] = "ASHA monographs",
  ["asha suppl"] = "ASHA. Supplement",
  ["ashrae j"] = "ASHRAE journal",
  ["ashrae trans"] = "ASHRAE transactions",
  ["ashrae winter conf pap"] = "ASHRAE winter conference papers. American Society of Heating, Refrigeration and Air-Conditioning Engineers",
  ["ashuu koukogaku"] = "Ashuu koukogaku",
  ["asia eur j"] = "Asia Europe journal",
  ["asia life sci"] = "Asia life sciences",
  ["asia mag"] = "The Asia magazine",
  ["asia major"] = "Asia major",
  ["asia ocean j nucl med biol"] = "Asia Oceania journal of nuclear medicine & biology",
  ["asia oceania j obstet gynaecol"] = "Asia-Oceania journal of obstetrics and gynaecology",
  ["asia pac allergy"] = "Asia Pacific allergy",
  ["asia pac biotech news"] = "Asia-Pacific biotech news",
  ["asia pac econ lit"] = "Asian-Pacific economic literature",
  ["asia pac fam med"] = "Asia Pacific family medicine",
  ["asia pac j atmos sci"] = "Asia-Pacific journal of atmospheric sciences",
  ["asia pac j blood types genes"] = "Asia-Pacific journal of blood types and genes",
  ["asia pac j cardiol"] = "Asia-Pacific journal of cardiology",
  ["asia pac j clin nutr"] = "Asia Pacific journal of clinical nutrition",
  ["asia pac j clin oncol"] = "Asia-Pacific journal of clinical oncology",
  ["asia pac j couns psychother"] = "Asia pacific journal of counselling and psychotherapy",
  ["asia pac j health law ethics"] = "Asia Pacific journal of health law & ethics",
  ["asia pac j med toxicol"] = "Asia Pacific Journal of Medical Toxicology",
  ["asia pac j mol biol biotechnol"] = "Asia-Pacific journal of molecular biology and biotechnology",
  ["asia pac j oncol nurs"] = "Asia-Pacific journal of oncology nursing",
  ["asia pac j ophthalmol (phila)"] = "Asia-Pacific journal of ophthalmology (Philadelphia, Pa.)",
  ["asia pac j public health"] = "Asia-Pacific journal of public health",
  ["asia pac j rural dev"] = "Asia-Pacific journal of rural development",
  ["asia pac j soc work"] = "Asia Pacific journal of social work",
  ["asia pac j soc work dev"] = "Asia Pacific journal of social work and development",
  ["asia pac j speech lang hear"] = "Asia Pacific journal of speech, language, and hearing",
  ["asia pac j sports med arthrosc rehabil technol"] = "Asia-Pacific journal of sports medicine, arthroscopy, rehabilitation and technology",
  ["asia pac math newsl"] = "Asia Pacific Mathematics Newsletters",
  ["asia pac obs"] = "Asia Pacific observer : a quarterly newsletter from the East-West Center",
  ["asia pac phys newsl"] = "Asia Pacific Physics Newsletter",
  ["asia pac policy stud"] = "Asia & the Pacific policy studies",
  ["asia pac pop policy"] = "Asia-Pacific population & policy",
  ["asia pac popin bull"] = "Asia-Pacific POPIN bulletin",
  ["asia pac popul j"] = "Asia-Pacific population journal",
  ["asia pac popul res abstr"] = "Asia-Pacific population research abstracts",
  ["asia pac psychiatry"] = "Asia-Pacific psychiatry : official journal of the Pacific Rim College of Psychiatrists",
  ["asia pac viewp"] = "Asia Pacific viewpoint",
  ["asia pacific bus rev"] = "Asia Pacific Business Review",
  ["asia pacific j econ bus"] = "Asia Pacific Journal of Economics and Business",
  ["asia q"] = "Asia quarterly",
  ["asia taepyongyang sangdam yongu"] = "Asia T'aep'yŏngyang sangdam yŏn'gu",
  ["asia-pac biotech news"] = "Asia-Pacific Biotech News",
  ["asia-pac j atmos sci"] = "Asia-Pacific Journal of Atmospheric Sciences",
  ["asia-pac j chem eng"] = "Asia-Pacific Journal of Chemical Engineering",
  ["asia-pac j oper res"] = "Asia-Pacific Journal of Operational Research",
  ["asia-pac popin newsl"] = "Asia-Pacific POPIN newsletter",
  ["asia-pacific devel j"] = "Asia-Pacific Development Journal",
  ["asia-pacific econ rev"] = "Asia-Pacific Economic Review",
  ["asia-pacific finan markets"] = "Asia-Pacific Financial Markets",
  ["asia-pacific j oper res"] = "Asia-Pacific Journal of Operational Research",
  ["asian aff (lond)"] = "Asian affairs",
  ["asian afr stud"] = "Asian and African studies",
  ["asian agrihist"] = "Asian agri-history",
  ["asian am j psychol"] = "Asian American journal of psychology",
  ["asian am pac isl j health"] = "Asian American and Pacific Islander journal of health",
  ["asian anthropol"] = "Asian anthropology",
  ["asian australas j plant sci biotechnol"] = "The Asian and Australasian journal of plant science and biotechnology",
  ["asian bioeth rev"] = "Asian bioethics review",
  ["asian biomed"] = "Asian Biomedicine",
  ["asian biomed (res rev news)"] = "Asian biomedicine : research, reviews and news",
  ["asian cardiovasc thorac ann"] = "Asian cardiovascular & thoracic annals",
  ["asian control conf"] = "Asian Control Conference. Asian Control Conference",
  ["asian dev rev"] = "Asian development review",
  ["asian devel rev"] = "Asian Development Review",
  ["asian econ"] = "Asian economies",
  ["asian econ j"] = "Asian Economic Journal",
  ["asian econ rev"] = "The Asian economic review",
  ["asian eur j math"] = "Asian European Journal Of Mathematics",
  ["asian fish sci"] = "Asian fisheries science",
  ["asian folkl stud"] = "Asian folklore studies",
  ["asian forum newsl"] = "Asian Forum newsletter : official organ of the Asian Forum of Parliamentarians on Population and Development",
  ["asian geogr"] = "Asian geographer",
  ["asian herpetol res"] = "Asian Herpetological Research",
  ["asian j aesthet dent"] = "Asian journal of aesthetic dentistry",
  ["asian j agr econ"] = "Asian Journal of Agricultural Economics",
  ["asian j agric food sci"] = "Asian journal of agriculture and food science",
  ["asian j androl"] = "Asian journal of andrology",
  ["asian j anesthesiol"] = "Asian journal of anesthesiology",
  ["asian j anim sci"] = "Asian journal of animal sciences",
  ["asian j anim vet adv"] = "Asian journal of animal and veterinary advances",
  ["asian j atmos environ"] = "Asian Journal of Atmospheric Environment",
  ["asian j biochem"] = "Asian Journal of Biochemistry",
  ["asian j biol life sci"] = "Asian journal of biological and life sciences",
  ["asian j biotechnol"] = "Asian Journal of Biotechnology",
  ["asian j biotechnol bioresour technol"] = "Asian journal of biotechnology and bioresource technology",
  ["asian j chem"] = "Asian Journal of Chemistry",
  ["asian j civ eng"] = "Asian Journal of Civil Engineering",
  ["asian j commun"] = "Asian journal of communication",
  ["asian j conserv biol"] = "Asian journal of conservation biology",
  ["asian j control"] = "Asian Journal of Control",
  ["asian j criminol"] = "Asian journal of criminology",
  ["asian j econ"] = "The Asian journal of economics : a journal of the Society for Asian Development",
  ["asian j econ soc stud"] = "Asian journal of economics and social studies",
  ["asian j endosc surg"] = "Asian journal of endoscopic surgery",
  ["asian j exp biol sci"] = "Asian Journal of Experimental Biological Sciences",
  ["asian j exp sci"] = "Asian Journal of Experimental Sciences",
  ["asian j gambl issues public health"] = "Asian journal of gambling issues and public health",
  ["asian j infect dis"] = "Asian journal of infectious diseases",
  ["asian j mater sci"] = "Asian Journal of Materials Science",
  ["asian j math"] = "Asian Journal of Mathematics",
  ["asian j math educ"] = "Asian Journal for Mathematics Education",
  ["asian j med sci"] = "Asian journal of medical sciences",
  ["asian j microbiol biotechnol environ exp sci"] = "Asian Journal of Microbiology, Biotechnology & Environmental Sciences",
  ["asian j microbiol biotechnol environ sci"] = "Asian journal of microbiology, biotechnology & environmental sciences : AJMBES",
  ["asian j mycol"] = "Asian journal of mycology",
  ["asian j neurosurg"] = "Asian journal of neurosurgery",
  ["asian j occup ther"] = "Asian journal of occupational therapy",
  ["asian j org chem"] = "Asian Journal of Organic Chemistry",
  ["asian j pharm"] = "Asian journal of pharmaceutics",
  ["asian j pharm clin res"] = "Asian journal of pharmaceutical and clinical research : AJPCR",
  ["asian j pharm pharmacol"] = "Asian journal of pharmacy and pharmacology",
  ["asian j pharm sci"] = "Asian journal of pharmaceutical sciences",
  ["asian j phys"] = "Asian Journal of Physics",
  ["asian j plant pathol"] = "Asian journal of plant pathology",
  ["asian j plant sci"] = "Asian journal of plant science",
  ["asian j psychiatr"] = "Asian journal of psychiatry",
  ["asian j res anim vet sci"] = "Asian journal of research in animal and veterinary sciences",
  ["asian j res crop sci"] = "Asian journal of research in crop science",
  ["asian j res soc sci humanit"] = "Asian journal of research in social sciences and humanities",
  ["asian j sci res"] = "Asian journal of scientific research",
  ["asian j soc psychol"] = "Asian journal of social psychology",
  ["asian j soc sci"] = "Asian journal of social science",
  ["asian j spectro"] = "Asian Journal of Spectroscopy",
  ["asian j sport exercise psychol"] = "Asian Journal of Sport and Exercise Psychology",
  ["asian j sports med"] = "Asian journal of sports medicine",
  ["asian j surg"] = "Asian journal of surgery",
  ["asian j technol innovation"] = "Asian Journal of Technology Innovation",
  ["asian j tradit med"] = "Asian journal of traditional medicines",
  ["asian j transfus sci"] = "Asian journal of transfusion science",
  ["asian j urol"] = "Asian journal of urology",
  ["asian j water environ pollut"] = "Asian Journal of Water, Environment and Pollution",
  ["asian j women stud"] = "Asian journal of women's studies",
  ["asian med (leiden)"] = "Asian medicine (Leiden, Netherlands)",
  ["asian med j"] = "Asian journal of medicine",
  ["asian migr"] = "Asian migrant",
  ["asian nurs res (korean soc nurs sci)"] = "Asian nursing research",
  ["asian pac cens forum"] = "Asian and Pacific census forum",
  ["asian pac census newsl"] = "Asian and Pacific census newsletter",
  ["asian pac isl nurs j"] = "Asian/Pacific Island nursing journal",
  ["asian pac j allergy immunol"] = "Asian Pacific journal of allergy and immunology",
  ["asian pac j cancer prev"] = "Asian Pacific journal of cancer prevention : APJCP",
  ["asian pac j trop biomed"] = "Asian Pacific journal of tropical biomedicine",
  ["asian pac j trop dis"] = "Asian Pacific journal of tropical disease",
  ["asian pac j trop med"] = "Asian Pacific journal of tropical medicine",
  ["asian pac migr j"] = "Asian and Pacific migration journal : APMJ",
  ["asian pac popul forum"] = "Asian and Pacific population forum",
  ["asian pac popul programme news"] = "Asian-Pacific population programme news",
  ["asian perspect"] = "Asian perspective",
  ["asian popul programme news"] = "Asian population programme news",
  ["asian popul stud"] = "Asian population studies",
  ["asian popul stud ser"] = "Asian population studies series",
  ["asian primates j"] = "Asian primates journal : a journal of the Southeast Asia, South Asia and China of the IUCN/SSC Primate Specialist Group",
  ["asian profile"] = "Asian profile",
  ["asian soc sci"] = "Asian social science",
  ["asian spine j"] = "Asian spine journal",
  ["asian stud"] = "Asian studies (Quezon City, Philippines)",
  ["asian stud rev"] = "Asian studies review",
  ["asian surv"] = "Asian survey",
  ["asian text j"] = "Asian Textile Journal",
  ["asian thought soc"] = "Asian thought & society",
  ["asian transport stud"] = "Asian Transport Studies",
  ["asian-australas j anim sci"] = "Asian-Australasian journal of animal sciences",
  ["asian-eur j math"] = "Asian-European Journal of Mathematics",
  ["asian-pacific econ lit"] = "Asian-Pacific Economic Literature",
  ["asiat forsch"] = "Asiatische Forschungen",
  ["asiat herpetol res"] = "Asiatic herpetological research",
  ["asiat j biotechnol resour"] = "Asiatic journal of biotechnology resources",
  ["asien afr lateinam"] = "Asien, Afrika, Lateinamerika (Berlin, Germany : 1973)",
  ["ask sihhiye derg"] = "Askerî sihhiye dergisi",
  ["ask vet mecm"] = "Askerı̂ veteriner mecmuasi",
  ["aslib proc"] = "Aslib proceedings",
  ["asm anim microbiol"] = "ASM Animal Microbiology",
  ["asm case rep"] = "ASM Case Reports",
  ["asm food microbiol"] = "ASM Food Microbiology",
  ["asm news"] = "ASM news",
  ["asme appl mech rev"] = "ASME Applied Mechanics Reviews",
  ["asme j appl mech"] = "ASME Journal of Applied Mechanics",
  ["asme j auton veh syst"] = "ASME Journal of Autonomous Vehicles and Systems",
  ["asme j biomech eng"] = "ASME Journal of Biomechanical Engineering",
  ["asme j comput inf sci eng"] = "ASME Journal of Computing and Information Science in Engineering",
  ["asme j comput nonlinear dyn"] = "ASME Journal of Computational and Nonlinear Dynamics",
  ["asme j convergence biomed eng biotechnol"] = "ASME Journal of Convergence in Biomedical Engineering & Biotechnology",
  ["asme j dyn syst meas control"] = "ASME Journal of Dynamic Systems Measurement & Control",
  ["asme j electrochem energy convers storage"] = "ASME Journal of Electrochemical Energy Conversion and Storage",
  ["asme j electron packag"] = "ASME Journal of Electronic Packaging",
  ["asme j energy res technol"] = "ASME Journal of Energy Resources Technology",
  ["asme j eng gas turbines power"] = "ASME Journal of Engineering for Gas Turbines and Power",
  ["asme j eng mater technol"] = "ASME Journal of Engineering Materials and Technology",
  ["asme j eng sci med diagn ther"] = "ASME Journal of Engineering and Science in Medical Diagnostics and Therapy",
  ["asme j eng sustainable build cities"] = "ASME Journal of Engineering for Sustainable Buildings and Cities",
  ["asme j fluids eng"] = "ASME Journal of Fluids Engineering",
  ["asme j fuel cell sci technol"] = "ASME Journal of Fuel Cell Science and Technology",
  ["asme j heat transfer"] = "ASME Journal of Heat Transfer",
  ["asme j manuf sci eng"] = "ASME Journal of Manufacturing Science and Engineering",
  ["asme j mech des"] = "ASME Journal of Mechanical Design",
  ["asme j mech rob"] = "ASME Journal of Mechanisms and Robotics",
  ["asme j med devices"] = "ASME Journal of Medical Devices",
  ["asme j med diagn"] = "ASME Journal of Medical Diagnostics",
  ["asme j micro nano-manuf"] = "ASME Journal of Micro and Nano-Manufacturing",
  ["asme j nanotechnol eng med"] = "ASME Journal of Nanotechnology in Engineering and Medicine",
  ["asme j nondestr eval diagn progn eng syst"] = "ASME Journal of Nondestructive Evaluation, Diagnostics & Prognostics of Engineering Systems",
  ["asme j nucl eng radiat sci"] = "ASME Journal of Nuclear Engineering and Radiation Science",
  ["asme j offshore mech arct eng"] = "ASME Journal of Offshore Mechanics and Arctic Engineering",
  ["asme j pressure vessel technol"] = "ASME Journal of Pressure Vessel Technology",
  ["asme j sol energy eng"] = "ASME Journal of Solar Energy Engineering",
  ["asme j therm sci eng appl"] = "ASME Journal of Thermal Science and Engineering Applications",
  ["asme j tribol"] = "ASME Journal of Tribology",
  ["asme j turbomach"] = "ASME Journal of Turbomachinery",
  ["asme j verif validation uncertainty quantif"] = "ASME Journal of Verification, Validation, and Uncertainty Quantification",
  ["asme j vib acoust"] = "ASME Journal of Vibration and Acoustics",
  ["asme lett dyn syst control"] = "ASME Letters in Dynamic Systems and Control",
  ["asme mech eng"] = "ASME Mechanical Engineering",
  ["asn neuro"] = "ASN neuro",
  ["asna rep"] = "ASNA reporter",
  ["asp appl biol"] = "Aspects of applied biology",
  ["asp komplexer systeme"] = "Aspekte Komplexer Systeme",
  ["asp stat serv cant stat geneve scs"] = "Aspects statistiques",
  ["aspects homogeneous catal"] = "Aspects of Homogeneous Catalysis",
  ["aspects math"] = "Aspects of Mathematics",
  ["aspen emphysema conf"] = "Aspen Emphysema Conference",
  ["aspens advis nurse exec"] = "Aspen's advisor for nurse executives",
  ["assay drug dev technol"] = "Assay and drug development technologies",
  ["assem autom"] = "Assembly Automation",
  ["assertive nurse"] = "Assertive Nurse",
  ["assess dev matters"] = "Assessment & development matters",
  ["assess educ"] = "Assessment in education : principles, policy & practice",
  ["assess eff interv"] = "Assessment for effective intervention : official journal of the Council for Educational Diagnostic Services",
  ["assess eval high educ"] = "Assessment and evaluation in higher education",
  ["asset anal"] = "Asset Analytics",
  ["assia jew med ethics"] = "Assia--Jewish medical ethics",
  ["assignment child"] = "Assignment children",
  ["assist inferm ric"] = "Assistenza infermieristica e ricerca : AIR",
  ["assist reprod rev"] = "Assisted reproduction reviews",
  ["assist soc"] = "L'Assistenza sociale",
  ["assist techn res ser"] = "Technology and Aging",
  ["assist technol"] = "Assistive technology : the official journal of RESNA",
  ["assist technol outcomes benefits"] = "Assistive technology outcomes and benefits",
  ["assist technol res s"] = "Smart Homes and Beyond",
  ["assist technol res ser"] = "Assistive technology research series",
  ["assistive technol"] = "Assistive Technology",
  ["assiut vet med j"] = "Assiut veterinary medical journal",
  ["assoc american publ electron manuscript ser"] = "Association of American Publishers Electronic Manuscript Series",
  ["assoc can map lib arch bull"] = "Association Of Canadian Map Libraries And Archives Bulletin",
  ["assoc manage"] = "Association management",
  ["assoc math res monogr"] = "Association for Mathematical Research Monographs",
  ["assoc med j"] = "Association medical journal",
  ["assoc soc manager"] = "Association & society manager",
  ["assoc women math ser"] = "Association for Women in Mathematics Series",
  ["asta adv stat anal"] = "AStA Advances in Statistical Analysis",
  ["asta j"] = "The A.S.T.A. journal",
  ["asthet med (berl)"] = "Asthetische Medizin",
  ["asthma gen pract"] = "Asthma in general practice : journal of the GPs in Asthma Group",
  ["asthma res pract"] = "Asthma research and practice",
  ["astin bull"] = "Astin Bulletin",
  ["astm spec tech publ"] = "ASTM Special Technical Publication",
  ["astm stand news"] = "ASTM Standardization News",
  ["astra proc"] = "ASTRA Proceedings",
  ["astron astrophys"] = "Astronomy and Astrophysics",
  ["astron astrophys rev"] = "Astronomy and Astrophysics Review",
  ["astron astrophys suppl ser"] = "Astronomy & astrophysics. Supplement series",
  ["astron astrophys trans"] = "Astronomical & Astrophysical Transactions",
  ["astron astrophys, suppl ser"] = "Astronomy & Astrophysics, Supplement Series",
  ["astron comput"] = "Astronomy and Computing",
  ["astron educ rev"] = "Astronomy Education Review",
  ["astron geophys"] = "Astronomy & Geophysics",
  ["astron j"] = "Astronomical Journal",
  ["astron lett"] = "Astronomy Letters",
  ["astron nachr"] = "Astronomische Nachrichten",
  ["astron nachrichten"] = "Astronomische Nachrichten",
  ["astron q"] = "Astronomy Quarterly",
  ["astron rep"] = "Astronomy Reports",
  ["astron rev"] = "Astronomical Review",
  ["astron zh [sov astron]"] = "Astronomicheskii Zhurnal [Soviet Astronomy]",
  ["astronom astrophys lib"] = "Astronomy and Astrophysics Library",
  ["astronom nachr"] = "Astronomische Nachrichten",
  ["astropart phys"] = "Astroparticle Physics",
  ["astrophy space sci"] = "Astrophysics and Space Science",
  ["astrophys bull"] = "Astrophysical Bulletin",
  ["astrophys j"] = "The Astrophysical journal",
  ["astrophys j lett"] = "The astrophysical journal. Letters",
  ["astrophys j suppl ser"] = "The Astrophysical journal. Supplement series",
  ["astrophys lett"] = "Astrophysical Letters",
  ["astrophys lett commun"] = "Astrophysical letters & communications",
  ["astrophys space sci"] = "Astrophysics and Space Science",
  ["astrophys space sci lib"] = "Astrophysics and Space Science Library",
  ["astrophys space sci libr"] = "Astrophysics and Space Science Library",
  ["astrophys space sci trans"] = "Astrophysics and Space Science Transactions",
  ["aswan heart cent sci pract ser"] = "Aswan Heart Centre science & practice series",
  ["asymptot anal"] = "Asymptotic Analysis",
  ["at data nucl data tables"] = "Atomic Data and Nuclear Data Tables",
  ["at energ [sov j at energy]"] = "Atomnaya Energiya [Soviet Journal of Atomic Energy]",
  ["at energy"] = "Atomic Energy",
  ["at energy (n y, ny, u s)"] = "Atomic Energy (New York, NY, United States)",
  ["at energy biophys biol med"] = "Atomic energy in biophysics, biology and medicine",
  ["at energy law j"] = "Atomic energy law journal",
  ["at energy rev"] = "Atomic energy review",
  ["at spectrosc"] = "Atomic Spectroscopy",
  ["atalanta (munch)"] = "Atalanta (Munich, Germany)",
  ["atcc connect"] = "ATCC connection : the newsletter of the American Type Culture Collection",
  ["aten primaria"] = "Atencion primaria",
  ["atene roma"] = "Atene e Roma",
  ["ateneo parmense 1"] = "L' Ateneo parmense. Sezione I, Acta bio-medica",
  ["ateneo parmense [1]"] = "Ateneo Parmense. Sezione I, Acta Bio-Medica",
  ["ateneo parmense acta biomed"] = "L'Ateneo parmense. Acta bio-medica : organo della Società di medicina e scienze naturali di Parma",
  ["ateneo parmense acta nat"] = "L'Ateneo parmense. Acta naturalia : organo della Società di medicina e scienze naturali di Parma",
  ["ateneo veneto"] = "Ateneo Veneto : revista di scienze, lettere ed arti",
  ["athena sci optim comput ser"] = "Athena Scientific Optimization and Computation Series",
  ["athenaeum econom"] = "Athenaeum Economics",
  ["athenaum monogr philos"] = "Athenaum Monografien. Philosophie",
  ["athens j sci"] = "Athens journal of sciences",
  ["atheroscler suppl"] = "Atherosclerosis. Supplements",
  ["atherosclerosis (amsterdam, neth)"] = "Atherosclerosis (Amsterdam, Netherlands)",
  ["atherosclerosis (shannon, irel)"] = "Atherosclerosis (Shannon, Ireland)",
  ["athl ther today"] = "Athletic therapy today : the journal for sports health care professionals",
  ["athl train sports health care"] = "Athletic training & sports health care",
  ["atl econ j"] = "Atlantic economic journal : AEJ",
  ["atl geol"] = "Atlantic Geology",
  ["atl j commun"] = "Atlantic journal of communication",
  ["atl mon"] = "Atlantic monthly (Boston, Mass. : 1993)",
  ["atl report"] = "Atlantic reporter",
  ["atl stud (abingdon)"] = "Atlantic studies (Abingdon, England)",
  ["atla abstr"] = "ATLA abstracts",
  ["atla, altern lab anim"] = "ATLA, Alternatives to Laboratory Animals",
  ["atlanta hist"] = "Atlanta history",
  ["atlanta hist j"] = "The Atlanta historical journal",
  ["atlanta med"] = "Atlanta medicine : bulletin of the Medical Association of Atlanta",
  ["atlanta wkly"] = "Atlanta weekly",
  ["atlantic econ j"] = "Atlantic Economic Journal",
  ["atlantis (montr)"] = "Atlantis (Montréal, Quebec)",
  ["atlantis (wolfv)"] = "Atlantis (Wolfville, N.S.)",
  ["atlantis briefs differ equ"] = "Atlantis Briefs in Differential Equations",
  ["atlantis comput intell syst"] = "Atlantis Computational Intelligence Systems",
  ["atlantis stud astropart phys cosmol"] = "Atlantis Studies in Astroparticle Physics and Cosmology",
  ["atlantis stud comput"] = "Atlantis Studies in Computing",
  ["atlantis stud comput finance financ eng"] = "Atlantis Studies in Computational Finance and Financial Engineering",
  ["atlantis stud differ equ"] = "Atlantis Studies in Differential Equations",
  ["atlantis stud dyn syst"] = "Atlantis Studies in Dynamical Systems",
  ["atlantis stud math"] = "Atlantis Studies in Mathematics",
  ["atlantis stud math eng sci"] = "Atlantis Studies in Mathematics for Engineering and Science",
  ["atlantis stud math phys theory appl"] = "Atlantis Studies in Mathematical Physics: Theory and Applications",
  ["atlantis stud probab stat"] = "Atlantis Studies in Probability and Statistics",
  ["atlantis stud sci comput electromagn"] = "Atlantis Studies in Scientific Computing in Electromagnetics",
  ["atlantis stud var geom"] = "Atlantis Studies in Variational Geometry",
  ["atlantis trans geom"] = "Atlantis Transactions in Geometry",
  ["atlas genet cytogenet oncol haematol"] = "Atlas of genetics and cytogenetics in oncology and haematology",
  ["atlas j biol"] = "Atlas journal of biology",
  ["atlas j plant biol"] = "Atlas journal of plant biology",
  ["atlas oral maxillofac surg clin north am"] = "Atlas of the oral and maxillofacial surgery clinics of North America",
  ["atlas radiol clin presse med"] = "Atlas de radiologie clinique de la Presse médicale",
  ["atlas world press rev"] = "Atlas world press review",
  ["atmos chem phys"] = "Atmospheric Chemistry and Physics",
  ["atmos chem phys discuss"] = "Atmospheric Chemistry and Physics Discussions",
  ["atmos environ"] = "Atmospheric Environment",
  ["atmos environ (1994)"] = "Atmospheric environment (Oxford, England : 1994)",
  ["atmos environ part a"] = "Atmospheric Environment Part A General Topics",
  ["atmos environ part b"] = "Atmospheric Environment Part B Urban Atmosphere",
  ["atmos environ x"] = "Atmospheric Environment: X",
  ["atmos environ, part a"] = "Atmospheric Environment, Part A: General Topics",
  ["atmos environ: x"] = "Atmospheric Environment: X",
  ["atmos meas tech"] = "Atmospheric Measurement Techniques",
  ["atmos meas tech discuss"] = "Atmospheric Measurement Techniques Discussions",
  ["atmos ocean"] = "Atmosphere-Ocean",
  ["atmos oceanic opt"] = "Atmospheric and Oceanic Optics",
  ["atmos oceanic sci lett"] = "Atmospheric and Oceanic Science Letters",
  ["atmos oceanogr sci libr"] = "Atmospheric and Oceanographic Sciences Library",
  ["atmos pollut res"] = "Atmospheric pollution research",
  ["atmos res"] = "Atmospheric Research",
  ["atmos res lett"] = "Atmospheric Research Letters",
  ["atmos sci lett"] = "Atmospheric Science Letters",
  ["atmosphere (basel)"] = "Atmosphere",
  ["atoll res bull"] = "Atoll research bulletin",
  ["atomization sprays"] = "Atomization and Sprays",
  ["ats sch"] = "ATS scholar",
  ["attach hum dev"] = "Attachment & human development",
  ["attachment (lond)"] = "Attachment (London, England)",
  ["atten defic hyperact disord"] = "Attention deficit and hyperactivity disorders",
  ["atten percept psychophys"] = "Attention, perception & psychophysics",
  ["atti accad fisiocrit siena"] = "Atti della R. Accademia dei fisiocritici in Siena",
  ["atti accad fisiocrit siena [med fis]"] = "Atti della Accademia dei Fisiocritici in Siena. Sezione Medico-Fisica",
  ["atti accad fisiocrit siena med fis"] = "Atti della Accademia dei fisiocritici in Siena. Sezione medico-fisica",
  ["atti accad ligure sci lett (5)"] = "Accademia Ligure di Scienze e Lettere",
  ["atti accad med lomb"] = "Atti della Accademia medica lombarda",
  ["atti accad naz lincei"] = "Atti della Accademia nazionale dei Lincei",
  ["atti accad naz lincei cl sci fis mat nat rend"] = "Atti della Accademia Nazionale dei Lincei, Classe di Scienze Fisiche, Matematiche e Naturali, Rendiconti",
  ["atti accad naz lincei cl sci fis mat natur rend lincei (9) mat appl"] = "Atti della Accademia Nazionale dei Lincei",
  ["atti accad naz lincei cl sci fis mat natur rend lincei (9) suppl"] = "Atti della Accademia Nazionale dei Lincei",
  ["atti accad naz lincei rend lincei mat appl"] = "Atti della Accademia Nazionale dei Lincei. Rendiconti Lincei. Matematica e Applicazioni",
  ["atti accad peloritana pericolanti cl sci fis mat natur"] = "Atti della Accademia Peloritana dei Pericolanti. Classe di Scienze, Fisiche, Matematiche e Naturali. AAPP. Physical, Mathematical, and Natural Sciences",
  ["atti accad pontaniana (ns)"] = "Atti della Accademia Pontaniana",
  ["atti accad sci ist bologna classe sci fis rend"] = "Atti della Accademia delle scienze dell'Istituto di Bologna, Classe di scienze fisiche. Memorie",
  ["atti accad sci istit bologna cl sci fis rend (14)"] = "Atti della Accademia delle Scienze dell’Istituto di Bologna",
  ["atti accad sci torino cl sci fis mat natur"] = "Atti della Accademia delle Scienze di Torino",
  ["atti clin otorinolaringoitr univ palermo"] = "Atti. Università di Palermo. Clinica oto-rino-laringoiatrica",
  ["atti congr naz soc ital med leg"] = "Atti del congresso nazionale. Società italiana di medicina legale e delle assicurazioni",
  ["atti congr naz stor med"] = "Atti. Congresso nazionale di storia della medicina",
  ["atti convegni aicm"] = "Atti dei Convegni dell’A",
  ["atti mem accad ital stor farm"] = "Atti e memorie della Accademia italiana di storia della farmacia",
  ["atti mem accad naz sci lett arti modena"] = "Atti e memorie - Accademia nazionale di scienze, lettere e arti, Modena. Accademia di scienze, lettere e arti (Modena, Italy)",
  ["atti mem accad stor arte sanit"] = "Atti e memorie della Accademia di storia dell'arte sanitaria",
  ["atti mem accad toscana sci lett colombaria"] = "Atti e memorie dell'Accademia toscana di scienze e lettere La Colombaria",
  ["atti mem deput stor patria antiche prov modenesi"] = "Atti e memorie - Deputazione di storia patria per le antiche provincie modenesi. Deputazione di storia patria per le antiche provincie modenesi",
  ["atti mem mem sci giuridiche lett (8)"] = "Atti e Memorie. Memorie Scientifiche, Giuridiche, Letterarie. Serie VIII",
  ["atti mem r accad petrarca lett arti sci"] = "Atti e memorie della Reale Accademia Petrarca di lettere, arti e scienze",
  ["atti sem mat fis univ modena"] = "Atti del Seminario Matematico e Fisico dell’Università di Modena",
  ["atti soc ital cardiol"] = "Atti della Società italiana di cardiologia",
  ["attiv didat"] = "Attività didattico-scientifica nell'anno accademico. Università di Roma. Istituto di storia della medicina",
  ["attual dent"] = "Attualità dentale",
  ["attual diet"] = "L'Attualità dietetica",
  ["attual ematol"] = "Attualita in ematologia",
  ["attual med"] = "L'Attualita medica",
  ["attual ostet ginecol"] = "Attualità di ostetricia e ginecologia",
  ["atw int z kernerg"] = "ATW Internationale Zeitschrift fur Kernenergie",
  ["aua news"] = "AUA news",
  ["aua update ser"] = "AUA update series",
  ["auaa j"] = "AUAA journal : official journal of the American Urological Association Allied",
  ["auckl univ law rev"] = "Auckland University law review",
  ["auckl waikate hist j"] = "Auckland-Waikato historical journal",
  ["audiol med"] = "Audiological medicine",
  ["audiol neurootol"] = "Audiology & neuro-otology",
  ["audiol neurotol extra"] = "Audiology and neurotology. Extra",
  ["audiol res"] = "Audiology research",
  ["audiol today"] = "Audiology today : bulletin of the American Academy of Audiology",
  ["audiology and neurotology"] = "Audiol. Neurotol.",
  ["audit neurosci"] = "Auditory neuroscience",
  ["audit percept cogn"] = "Auditory perception & cognition",
  ["audit unit news"] = "Audit Unit News",
  ["aufbaukurs math"] = "Aufbaukurs Mathematik",
  ["aufklarung (hambg)"] = "Aufklärung",
  ["augment altern commun"] = "Augmentative and alternative communication (Baltimore, Md. : 1985)",
  ["augment environ comput assist interv (2014)"] = "Augmented environments for computer-assisted interventions : 9th International Workshop, AE-CAI 2014, held in conjunction with MICCAI 2014, Boston, MA, USA, September 14, 2014 : proceedings. AE-CAI (Workshop) (9th : 2014 : Boston, Mass.)",
  ["augment hum res"] = "Augmented Human Research",
  ["augsbg math-naturwiss schrift"] = "Augsburger Mathematisch-Naturwissenschaftliche Schriften",
  ["augsbg ökol schr"] = "Augsburger Ökologische Schriften",
  ["auris nasus larynx"] = "Auris, nasus, larynx",
  ["aurora dover mod math orig"] = "Aurora Dover Modern Math Originals",
  ["ausgrab funde"] = "Ausgrabungen und Funde",
  ["aust agder arv"] = "Aust-Agder arv",
  ["aust archaeol"] = "Australian Archaeology",
  ["aust bull labour"] = "Australian bulletin of labour",
  ["aust clin rev"] = "Australian clinical review",
  ["aust coll midwives inc j"] = "Australian College of Midwives Incorporated journal",
  ["aust community psychol"] = "Australian community psychologist (Online)",
  ["aust crit care"] = "Australian critical care : official journal of the Confederation of Australian Critical Care Nurses",
  ["aust cult hist"] = "Australian cultural history",
  ["aust dent j"] = "Australian dental journal",
  ["aust dent mirror"] = "The Australian dental mirror, incorporating the Queensland dental magazine",
  ["aust dent pract"] = "Australian dental practice",
  ["aust econ hist rev"] = "Australian economic history review",
  ["aust econ pap"] = "Australian economic papers",
  ["aust econ rev"] = "The Australian economic review",
  ["aust educ res"] = "Australian educational researcher",
  ["aust endod j"] = "Australian endodontic journal : the journal of the Australian Society of Endodontology Inc",
  ["aust entomol"] = "Austral entomology",
  ["aust entomol (brisb)"] = "The Australian entomologist",
  ["aust fam physician"] = "Australian family physician",
  ["aust fem stud"] = "Australian feminist studies",
  ["aust field ornith"] = "Australian Field Ornithology",
  ["aust for"] = "Australian forestry",
  ["aust gemmol"] = "Australian Gemmologist",
  ["aust geogr"] = "The Australian geographer",
  ["aust geogr stud"] = "Australian geographical studies",
  ["aust geomech j"] = "Australian Geomechanics Journal",
  ["aust health rev"] = "Australian health review : a publication of the Australian Hospital Association",
  ["aust hist stud"] = "Australian historical studies",
  ["aust hosp"] = "Australian hospital",
  ["aust infect control"] = "Australian infection control : official journal of the Australian Infection Control Association Inc",
  ["aust intellect prop j"] = "Australian intellectual property journal",
  ["aust j  learn diffic"] = "Australian journal of learning difficulties",
  ["aust j adv nurs"] = "The Australian journal of advanced nursing : a quarterly publication of the Royal Australian Nursing Federation",
  ["aust j ageing"] = "Australian journal on ageing",
  ["aust j agric econ"] = "Australian Journal of Agricultural Economics",
  ["aust j agric res"] = "Australian journal of agricultural research",
  ["aust j agric resour econ"] = "The Australian journal of agricultural and resource economics",
  ["aust j anthropol"] = "The Australian journal of anthropology",
  ["aust j appl sci"] = "Australian Journal of Applied Science",
  ["aust j basic appl sci"] = "Australian Journal of Basic and Applied Sciences",
  ["aust j biol sci"] = "Australian journal of biological sciences",
  ["aust j biotechnol"] = "Australian journal of biotechnology",
  ["aust j bot"] = "Australian journal of botany",
  ["aust j chem"] = "Australian Journal of Chemistry",
  ["aust j chem eng"] = "Australian Journal of Chemical Engineers",
  ["aust j chin aff"] = "The Australian journal of Chinese affairs = Ao chung",
  ["aust j civ eng"] = "Australian Journal of Civil Engineering",
  ["aust j civil eng"] = "Australian Journal of Civil Engineering",
  ["aust j clin hypnother"] = "The Australian journal of clinical hypnotherapy",
  ["aust j crop sci"] = "Australian Journal of Crop Science",
  ["aust j dent"] = "The Australian journal of dentistry",
  ["aust j dermatol"] = "Australian journal of dermatology",
  ["aust j early child"] = "Australian journal of early childhood",
  ["aust j earth sci"] = "Australian Journal of Earth Sciences",
  ["aust j ecol"] = "Australian Journal of Ecology",
  ["aust j edu chem"] = "Australian Journal of Education in Chemistry",
  ["aust j educ"] = "Australian journal of education",
  ["aust j electr electron eng"] = "Australian Journal of Electrical and Electronics Engineering",
  ["aust j entomol"] = "Australian journal of entomology",
  ["aust j environ educ"] = "Australian Journal of Environmental Education",
  ["aust j exp agric"] = "Australian Journal of Experimental Research",
  ["aust j exp biol med sci"] = "The Australian journal of experimental biology and medical science",
  ["aust j fam law"] = "Australian journal of family law",
  ["aust j forensic sci"] = "The Australian journal of forensic sciences",
  ["aust j fr stud"] = "Australian journal of French studies",
  ["aust j gen pract"] = "Australian journal of general practice",
  ["aust j geod photogramm surv"] = "Australian journal of geodesy, photogrammetry, and surveying",
  ["aust j grape wine res"] = "Australian Journal of Grape and Wine Research",
  ["aust j holist nurs"] = "The Australian journal of holistic nursing",
  ["aust j mar freshwater res"] = "Australian Journal of Marine & Freshwater Research",
  ["aust j marit ocean aff"] = "Australian Journal of Maritime & Ocean Affairs",
  ["aust j marriage fam"] = "Australian journal of marriage & family",
  ["aust j math anal appl"] = "Australian Journal Of Mathematical Analysis And Applications",
  ["aust j mech eng"] = "Australian Journal of Mechanical Engineering",
  ["aust j ment retard"] = "The Australian journal of mental retardation",
  ["aust j midwifery"] = "Australian journal of midwifery : professional journal of the Australian College of Midwives Incorporated",
  ["aust j multi-discip eng"] = "Australian Journal of Multi-Disciplinary Engineering",
  ["aust j ophthalmol"] = "Australian journal of ophthalmology",
  ["aust j optom"] = "The Australian journal of optometry",
  ["aust j phys"] = "Australian Journal of Physics",
  ["aust j physiother"] = "The Australian journal of physiotherapy",
  ["aust j plant physiol"] = "Australian journal of plant physiology",
  ["aust j polit hist"] = "The Australian journal of politics and history",
  ["aust j polit sci"] = "Australian journal of political science",
  ["aust j prim health"] = "Australian journal of primary health",
  ["aust j prof appl ethics"] = "Australian journal of professional and applied ethics",
  ["aust j psychol"] = "Australian journal of psychology",
  ["aust j public health"] = "Australian journal of public health",
  ["aust j rural health"] = "The Australian journal of rural health",
  ["aust j sci"] = "The Australian journal of science",
  ["aust j sci med sport"] = "Australian journal of science and medicine in sport",
  ["aust j sci res (b)"] = "Australian journal of scientific research. Ser. B: Biological sciences",
  ["aust j sci res b"] = "Australian journal of scientific research. Ser. B: Biological sciences",
  ["aust j sex marriage and fam"] = "Australian journal of sex, marriage & family",
  ["aust j soc issues"] = "The Australian journal of social issues",
  ["aust j soil res"] = "Australian Journal of Soil Research",
  ["aust j stat"] = "The Australian journal of statistics",
  ["aust j struct eng"] = "Australian Journal of Structural Engineering",
  ["aust j water resour"] = "Australian Journal of Water Resources",
  ["aust j zool"] = "Australian journal of zoology",
  ["aust law j"] = "The Australian law journal",
  ["aust libr j"] = "The Australian library journal",
  ["aust mammal"] = "Australian mammalogy",
  ["aust med rec j"] = "Australian medical record journal",
  ["aust meteorol mag"] = "Australian Meteorological Magazine",
  ["aust meteorol oceanogr j"] = "Australian Meteorological and Oceanographic Journal",
  ["aust mycol"] = "Australasian mycologist : journal of the Australasian Mycological Society, Inc",
  ["aust n z gen pract"] = "The Australian and New Zealand general practitioner",
  ["aust n z grapegrow winemak"] = "The Australian & New Zealand grapegrower & winemaker",
  ["aust n z j art"] = "Australian and New Zealand journal of art",
  ["aust n z j criminol"] = "The Australian & New Zealand journal of criminology",
  ["aust n z j fam ther"] = "The Australian and New Zealand journal of family therapy",
  ["aust n z j med"] = "Australian and New Zealand journal of medicine",
  ["aust n z j ment health nurs"] = "The Australian and New Zealand journal of mental health nursing",
  ["aust n z j obstet gynaecol"] = "The Australian & New Zealand journal of obstetrics & gynaecology",
  ["aust n z j obstet gynaecol suppl"] = "Australian and New Zealand Journal of Obstetrics and Gynaecology. Supplement",
  ["aust n z j ophthalmol"] = "Australian and New Zealand journal of ophthalmology",
  ["aust n z j psychiatry"] = "The Australian and New Zealand journal of psychiatry",
  ["aust n z j public health"] = "Australian and New Zealand journal of public health",
  ["aust n z j sociol"] = "The Australian and New Zealand journal of sociology",
  ["aust n z j stat"] = "Australian & New Zealand Journal of Statistics",
  ["aust n z j surg"] = "The Australian and New Zealand journal of surgery",
  ["aust new zealand health policy"] = "Australia and New Zealand health policy",
  ["aust nurs j"] = "Australian nursing journal (July 1993)",
  ["aust nurs midwifery j"] = "Australian nursing & midwifery journal",
  ["aust nurses j"] = "The Australian nurses' journal. Royal Australian Nursing Federation",
  ["aust nz j stat"] = "Australian & New Zealand Journal of Statistics",
  ["aust occup ther j"] = "Australian occupational therapy journal",
  ["aust orchid rev"] = "Australian orchid review",
  ["aust orthod j"] = "Australian orthodontic journal",
  ["aust outlook"] = "The Australian outlook",
  ["aust paediatr j"] = "Australian paediatric journal",
  ["aust prescr"] = "Australian prescriber",
  ["aust prosthodont j"] = "Australian prosthodontic journal",
  ["aust prosthodont soc bull"] = "Australian Prosthodontic Society bulletin",
  ["aust psychol"] = "Australian psychologist",
  ["aust q"] = "The Australian quarterly",
  ["aust soc"] = "Australian society",
  ["aust soc prosthodontists bull"] = "Australian Society Of Prosthodontists Bulletin",
  ["aust syst bot"] = "Australian systematic botany",
  ["aust vet j"] = "Australian veterinary journal",
  ["aust zool"] = "The Australian zoologist",
  ["austin addict sci"] = "Austin addiction sciences",
  ["austin alzheimers parkinsons dis"] = "Austin Alzheimer's and Parkinson's disease",
  ["austin biochem"] = "Austin biochemistry",
  ["austin biom biostat"] = "Austin biometrics and biostatistics",
  ["austin j anal pharm chem"] = "Austin journal of analytical and pharmaceutical chemistry",
  ["austin j anesth analg"] = "Austin journal of anesthesia and analgesia",
  ["austin j biomed eng"] = "Austin journal of biomedical engineering",
  ["austin j cancer clin res"] = "Austin journal of cancer and clinical research",
  ["austin j cerebrovasc dis stroke"] = "Austin journal of cerebrovascular disease & stroke",
  ["austin j clin  immunol"] = "Austin journal of clinical immunology",
  ["austin j clin cardiol"] = "Austin journal of clinical cardiology",
  ["austin j clin ophthalmol"] = "Austin journal of clinical ophthalmology",
  ["austin j clin pathol"] = "Austin journal of clinical pathology",
  ["austin j drug abuse addict"] = "Austin journal of drug abuse and addiction",
  ["austin j environ toxicol"] = "Austin journal of environmental toxicology",
  ["austin j hiv aids res"] = "Austin journal of HIV/AIDS research",
  ["austin j med oncol"] = "Austin journal of medical oncology",
  ["austin j nanomed nanotechnol"] = "Austin journal of nanomedicine & nanotechnology",
  ["austin j nephrol hypertens"] = "Austin journal of nephrology and hypertension",
  ["austin j nutr food sci"] = "Austin journal of nutrition and food sciences",
  ["austin j pediatr"] = "Austin journal of pediatrics",
  ["austin j pharmacol ther"] = "Austin journal of pharmacology and therapeutics",
  ["austin j pulm respir med"] = "Austin journal of pulmonary and respiratory medicine",
  ["austin j surg"] = "Austin journal of surgery",
  ["austin j urol"] = "Austin journal of urology",
  ["austin j vasc med"] = "Austin journal of vascular medicine",
  ["austin liver"] = "Austin liver",
  ["austin neurol neurosci"] = "Austin neurology & neurosciences",
  ["austin pharmacol pharm"] = "Austin pharmacology & pharmaceutics",
  ["austin virol retro virol"] = "Austin virology and retro virology",
  ["austral comput sci comm"] = "Australian Computer Science Communications",
  ["austral comput sci commun"] = "Australian Computer Science Communications",
  ["austral ecol"] = "Austral ecology",
  ["austral entomol"] = "Austral Entomology",
  ["austral j phys"] = "Australian Journal of Physics",
  ["austral j statist"] = "The Australian Journal of Statistics",
  ["austral j vet sci"] = "Austral journal of veterinary sciences",
  ["austral math soc gaz"] = "The Australian Mathematical Society. Gazette",
  ["austral math soc lect ser"] = "Australian Mathematical Society Lecture Series",
  ["australas ann med"] = "Australasian annals of medicine",
  ["australas biotechnol"] = "Australasian biotechnology",
  ["australas cathol rec"] = "The Australasian Catholic record",
  ["australas chiropr osteopathy"] = "Australasian chiropractic & osteopathy : journal of the Chiropractic & Osteopathic College of Australasia",
  ["australas emerg care"] = "Australasian emergency care",
  ["australas emerg nurs j"] = "Australasian emergency nursing journal : AENJ",
  ["australas epidemiol"] = "Australasian epidemiologist",
  ["australas hist archaeol"] = "Australasian historical archaeology : journal of the Australasian Society for Historical Archaeology",
  ["australas j ageing"] = "Australasian journal on ageing",
  ["australas j combin"] = "The Australasian Journal of Combinatorics",
  ["australas j dermatol"] = "The Australasian journal of dermatology",
  ["australas j eng educ"] = "Australasian Journal of Engineering Education",
  ["australas j environ educ"] = "Australasian Journal of Environmental Education",
  ["australas j environ manage"] = "Australasian Journal of Environmental Management",
  ["australas j log"] = "The Australasian Journal of Logic",
  ["australas j optom"] = "The Australasian journal of optometry",
  ["australas j pharm"] = "The Australasian journal of pharmacy",
  ["australas j philos"] = "Australasian journal of philosophy",
  ["australas j ultrasound med"] = "Australasian journal of ultrasound in medicine",
  ["australas j water resour"] = "Australasian Journal of Water Resources",
  ["australas lichenol"] = "Australasian lichenology",
  ["australas med j"] = "The Australasian medical journal",
  ["australas nurses j"] = "The Australasian nurses journal",
  ["australas palaeontol mem"] = "Australasian palaeontological memoirs",
  ["australas philos rev"] = "Australasian Philosophical Review",
  ["australas phys eng sci med"] = "Australasian physical & engineering sciences in medicine",
  ["australas plant dis notes"] = "Australasian plant disease notes",
  ["australas plant pathol"] = "Australasian plant pathology : APP",
  ["australas psychiatry"] = "Australasian psychiatry : bulletin of Royal Australian and New Zealand College of Psychiatrists",
  ["australas radiol"] = "Australasian radiology",
  ["australasian j reg stud"] = "Australasian Journal of Regional Studies",
  ["australian bull lab"] = "Australian Bulletin of Labour",
  ["australian commodities forecasts issues"] = "Australian Commodities Forecasts and Issues",
  ["australian econ hist rev"] = "Australian Economic History Review",
  ["australian econ pap"] = "Australian Economic Papers",
  ["australian econ rev"] = "Australian Economic Review",
  ["australian j agr econ"] = "Australian Journal of Agricultural Economics",
  ["australian j agr resource econ"] = "Australian Journal of Agricultural and Resource Economics",
  ["australian j lab econ"] = "Australian Journal of Labour Economics",
  ["australian j manage"] = "Australian Journal of Management",
  ["australian tax forum"] = "Australian Tax Forum",
  ["austrian hist yearb"] = "Austrian history yearbook",
  ["austrian j earth sci"] = "Austrian journal of earth sciences : an international journal of the Austrian Geological Society",
  ["austrian j for sci"] = "Austrian Journal of Forest Science",
  ["austrian med bull"] = "Austrian medical bulletin",
  ["autism adulthood"] = "Autism in adulthood : challenges and management",
  ["autism dev lang impair"] = "Autism & developmental language impairments",
  ["autism open access"] = "Autism-open access",
  ["autism policy pract"] = "Autism policy & practice",
  ["autism res"] = "Autism research : official journal of the International Society for Autism Research",
  ["autism res treat"] = "Autism research and treatment",
  ["auto immun highlights"] = "Auto- immunity highlights",
  ["autobiogr stud"] = "Auto/biography studies : a/b",
  ["autocarto res symp"] = "AutoCarto research symposium",
  ["autoimmun rev"] = "Autoimmunity reviews",
  ["autoimmune dis"] = "Autoimmune diseases",
  ["autoimmune dis ther approaches"] = "Autoimmune diseases and therapeutic approaches : open access",
  ["autoimmune infect dis"] = "Autoimmune and infectious diseases : open access",
  ["autom constr"] = "Automation in Construction",
  ["autom control comput sci"] = "Automatic Control and Computer Sciences",
  ["autom doc math ling"] = "Automatic Documentation and Mathematical Linguistics",
  ["autom exp"] = "Automated experimentation",
  ["autom remote control"] = "Automation and Remote Control",
  ["autom software eng"] = "Automated Software Engineering",
  ["autom-control ind eng ser"] = "Automation-Control and Industrial Engineering Series",
  ["automat comput appl math"] = "Automation Computers Applied Mathematics. Scientific Journal",
  ["automat reason ser"] = "Automated Reasoning Series",
  ["automat remote control"] = "Automation and Remote Control",
  ["automatica (oxf)"] = "Automatica : the journal of IFAC, the International Federation of Automatic Control",
  ["automatica j ifac"] = "Automatica. A Journal of IFAC, the International Federation of Automatic Control",
  ["automot engine technol"] = "Automotive and Engine Technology",
  ["automot ind ai"] = "Automotive Industries AI",
  ["automot innovation"] = "Automotive Innovation",
  ["auton agent multi agent syst"] = "Autonomous agents and multi-agent systems",
  ["auton agents multi-agent syst"] = "Autonomous Agents and Multi-Agent Systems",
  ["auton autacoid pharmacol"] = "Autonomic & autacoid pharmacology",
  ["auton intell syst"] = "Autonomous Intelligent Systems",
  ["auton neurosci"] = "Autonomic neuroscience : basic & clinical",
  ["auton rob"] = "Autonomous Robots",
  ["auton robots"] = "Autonomous robots",
  ["auton syst"] = "Autonomic Systems",
  ["autonomy (birm)"] = "Autonomy (Birmingham, England)",
  ["autophagy rep"] = "Autophagy Reports",
  ["autops case rep"] = "Autopsy & case reports",
  ["autre eur"] = "L'Autre Europe",
  ["aux plaines news"] = "Aux Plaines news",
  ["av (phila)"] = "The A-V",
  ["av commun rev"] = "AV communication review",
  ["av invest"] = "Avance de investigación",
  ["av investig agropecu"] = "Avances en investigación agropecuaria",
  ["av odontoestomatol"] = "Avances en odontoestomatologia",
  ["av periodoncia"] = "Avances en periodoncia",
  ["av piscol clin latinonot"] = "Avances en psicología clínica latinoamericana",
  ["av quim"] = "Avances en Quimica",
  ["ava res rev"] = "AVA research reviews",
  ["avenir med"] = "L'Avenir médical",
  ["aviakosm ekolog med"] = "Aviakosmicheskaia i ekologicheskaia meditsina = Aerospace and environmental medicine",
  ["avian biol res"] = "Avian biology research",
  ["avian conserv ecol"] = "Avian conservation and ecology = Écologie et conservation des oiseaux",
  ["avian dis"] = "Avian diseases",
  ["avian dis dig"] = "Avian diseases digest",
  ["avian pathol"] = "Avian pathology : journal of the W.V.P.A",
  ["avian res"] = "Avian research",
  ["aviat space environ med"] = "Aviation, space, and environmental medicine",
  ["aviat week space technol"] = "Aviation week & space technology",
  ["avic mag"] = "Avicultural Magazine",
  ["avicenna j clin microbiol infect"] = "Avicenna journal of clinical microbiology and infection : AJCMI",
  ["avicenna j med"] = "Avicenna journal of medicine",
  ["avicenna j med biotechnol"] = "Avicenna journal of medical biotechnology",
  ["avicenna j neuropsychophysiol"] = "Avicenna journal of neuropsychophysiology",
  ["avicenna j phytomed"] = "Avicenna journal of phytomedicine",
  ["avs news"] = "AVS news",
  ["avs quantum sci"] = "AVS Quantum Science",
  ["avsc news"] = "AVSC news (Association for Voluntary Surgical Contraception (U.S.))",
  ["avtom svarka"] = "Avtomaticheskaya Svarka",
  ["avtomat i telemekh"] = "Rossiiskaya Akademiya Nauk. Avtomatika i Telemekhanika",
  ["avtomat i vychisl tekhn (minsk)"] = "Minskiy Radiotekhnicheskiy Institut",
  ["awhonn lifelines"] = "AWHONN lifelines",
  ["awhonn voice"] = "AWHONN voice",
  ["awhonns clin issues perinat womens health nurs"] = "AWHONN's clinical issues in perinatal and women's health nursing",
  ["awr bull"] = "A.W.R. Bulletin. Association for the Study of the World Refugee Problem",
  ["awwa water sci"] = "AWWA water science",
  ["azania: archaeol res afr"] = "Azania: Archaeological Research in Africa",
  ["azerb j math"] = "Azerbaijan Journal of Mathematics",
  ["azerb khim zh"] = "Azerbaidzhanskii Khimicheskii Zhurnal",
  ["azerbaidzhanskii meditsinskii zhurnal"] = "Azerbaĭdzhanskiĭ medit︠s︡inskiĭ zhurnal = Aserbaidschanische medizinische Zeitschrift",
  ["aziia afr segodnia"] = "Azii︠a︡ i Afrika segodni︠a︡",
  ["b c med j"] = "British Columbia medical journal",
  ["b c stud"] = "BC studies",
  ["b e j econom anal policy"] = "The B.E. journal of economic analysis & policy",
  ["b e j theor econ"] = "The B. E. Journal of Theoretical Economics",
  ["babiski vestn"] = "Babis̆ki vestnik",
  ["backgr notes ser"] = "Department of State publication. Background notes series",
  ["backgr pap"] = "Background paper (United States. Physician Payment Review Commission)",
  ["bacteriol rev"] = "Bacteriological reviews",
  ["bacteriol virusol parazitol epidemiol"] = "Bacteriologia, virusologia, parazitologia, epidemiologia (Bucharest, Romania : 1990)",
  ["bacteriol virusol parazitol epidemiol (bucur)"] = "Bacteriologia, virusologia, parazitologia, epidemiologia",
  ["baden neujahrsbl"] = "Badener Neujahrsblätter",
  ["baessler arch beitr volkerkd"] = "Baessler-Archiv; Beiträge zur Völkerkunde",
  ["bag j basic appl genet"] = "BAG. Journal of basic and applied genetics",
  ["bahia med"] = "Bahia médica",
  ["bahrain med bull"] = "Bahrain medical bulletin : BMB",
  ["baillieres best pract res clin anaesthesiol"] = "Baillière's best practice & research. Clinical anaesthesiology",
  ["baillieres best pract res clin endocrinol metab"] = "Baillière's best practice & research. Clinical endocrinology & metabolism",
  ["baillieres best pract res clin gastroenterol"] = "Baillière's best practice & research. Clinical gastroenterology",
  ["baillieres best pract res clin haematol"] = "Baillière's best practice & research. Clinical haematology",
  ["baillieres best pract res clin obstet gynaecol"] = "Baillière's best practice & research. Clinical obstetrics & gynaecology",
  ["baillieres best pract res clin rheumatol"] = "Baillière's best practice & research. Clinical rheumatology",
  ["baillieres clin anaesthesiol"] = "Baillière's clinical anaesthesiology",
  ["baillieres clin endocrinol metab"] = "Baillière's clinical endocrinology and metabolism",
  ["baillieres clin gastroenterol"] = "Baillière's clinical gastroenterology",
  ["baillieres clin haematol"] = "Baillière's clinical haematology",
  ["baillieres clin immunol allergy"] = "Baillière's clinical immunology and allergy",
  ["baillieres clin neurol"] = "Baillière's clinical neurology",
  ["baillieres clin obstet gynaecol"] = "Baillière's clinical obstetrics and gynaecology",
  ["baillieres clin oncol"] = "Baillieres Clinical Oncology",
  ["baillieres clin rheumatol"] = "Baillière's clinical rheumatology",
  ["baker cederberg noteb"] = "Baker-Cederberg notebook. Baker-Cederberg Museum & Archives",
  ["baldwins ky revis statut annot ky"] = "Baldwin's Kentucky revised statutes annotated : KRS. Kentucky",
  ["balisage ser markup technol"] = "Balisage series on markup technologies",
  ["balkan j geom appl"] = "Balkan Journal of Geometry and its Applications",
  ["balkan j med genet"] = "Balkan journal of medical genetics : BJMG",
  ["balkan med j"] = "Balkan medical journal",
  ["balkan soc geom monogr textb"] = "Balkan Society of Geometers Monographs and Textbooks",
  ["balt for"] = "Baltic forestry",
  ["balt j health phys act"] = "Baltic journal of health and physical activity",
  ["balt j road bridge eng"] = "Baltic Journal of Road and Bridge Engineering",
  ["balt j sport health sci"] = "Baltic journal of sport & health sciences",
  ["baltim health news"] = "Baltimore health news",
  ["baltim mag"] = "Baltimore magazine",
  ["banach center publ"] = "Banach Center Publications",
  ["banach j math anal"] = "Banach Journal of Mathematical Analysis",
  ["banats j biotechnol"] = "Banat's Journal of Biotechnology",
  ["banbury rep"] = "Banbury report",
  ["banca naz lavoro quart rev"] = "Banca Nazionale del Lavoro Quarterly Review",
  ["banco españa econ bull"] = "Banco de España Economic Bulletin",
  ["bandaoti xuebao"] = "Bandaoti Xuebao",
  ["bangk med j"] = "The Bangkok medical journal",
  ["bangkok post"] = "The Bangkok post",
  ["banglad j microbiol"] = "Bangladesh journal of microbiology",
  ["bangladesh dev stud"] = "Bangladesh Development Studies",
  ["bangladesh devel stud"] = "Bangladesh Development Studies",
  ["bangladesh econ rev"] = "Bangladesh Economic Review",
  ["bangladesh j bot"] = "Bangladesh journal of botany",
  ["bangladesh j pharmacol"] = "Bangladesh journal of pharmacology",
  ["bangladesh j plant taxon"] = "Bangladesh journal of plant taxonomy",
  ["bangladesh j sci res"] = "The Bangladesh journal of scientific research",
  ["bangladesh j zool"] = "Bangladesh journal of zoology",
  ["bangladesh med j"] = "Bangladesh medical journal",
  ["bangladesh med res counc bull"] = "Bangladesh Medical Research Council bulletin",
  ["bangsaseon san eob haghoeji"] = "Bangsaseon san-eob haghoeji",
  ["bank engl quart bull"] = "Bank of England Quarterly Bulletin",
  ["bank israel banking rev"] = "Bank of Israel Banking Review",
  ["bank israel econ rev"] = "Bank of Israel Economic Review",
  ["bank japan monet econ stud"] = "Bank of Japan Monetary and Economic Studies",
  ["bank korea econ pap"] = "Bank of Korea Economic Papers",
  ["bank valletta rev"] = "Bank of Valletta Review",
  ["banko janakari"] = "Banko janakari",
  ["baoj cancer res ther"] = "BAOJ cancer research & therapy",
  ["baoj obes weight loss manag"] = "BAOJ obesity & weight loss management",
  ["baoj pharm sci"] = "BAOJ pharmaceutical sciences",
  ["baptist hist herit"] = "Baptist history and heritage",
  ["baptist q"] = "The Baptist quarterly",
  ["barbados nurs j"] = "Barbados nursing journal",
  ["barbastella (alcala hen)"] = "Barbastella (Alcala de Henares)",
  ["barc quir"] = "Barcelona quirúrgica",
  ["barc respir netw rev"] = "Barcelona respiratory network reviews",
  ["bariatr nurs surg patient care"] = "Bariatric nursing and surgical patient care : official journal of the National Association of Bariatric Nurses",
  ["bariatr surg pract patient care"] = "Bariatric surgical practice and patient care",
  ["barley genet newsl"] = "Barley genetics newsletter",
  ["basal facts"] = "Basal facts",
  ["basal ganglia"] = "Basal ganglia",
  ["basic appl dryland res"] = "Basic and applied dryland research",
  ["basic appl ecol"] = "Basic and Applied Ecology",
  ["basic appl herpetol"] = "Basic and applied herpetology",
  ["basic appl histochem"] = "Basic and applied histochemistry",
  ["basic appl myol"] = "Basic and applied myology : BAM",
  ["basic appl pathol"] = "Basic and applied pathology",
  ["basic appl soc psych"] = "Basic and applied social psychology",
  ["basic clin androl"] = "Basic and clinical andrology",
  ["basic clin neurosci"] = "Basic and clinical neuroscience",
  ["basic clin pharmacol toxicol"] = "Basic & clinical pharmacology & toxicology",
  ["basic clin physiol pharmacol"] = "Basic & Clinical Pharmacology & Toxicology",
  ["basic life sci"] = "Basic life sciences",
  ["basic res cardiol"] = "Basic research in cardiology",
  ["basin res"] = "Basin Research",
  ["basiswissen statist wirtschaftswiss"] = "Basiswissen Statistik fur Wirtschaftswissenschaftler",
  ["basl beitr geogr"] = "Basler Beiträge zur Geographie",
  ["basl beitr geschwiss"] = "Basler Beiträge zur Geschichtswissenschaft",
  ["basl z gesch altertumskd"] = "Basler Zeitschrift für Geschichte und Altertumskunde",
  ["basler veroff gesch med biol"] = "Basler Veröffentlichungen zur Geschichte der Medizin und der Biologie",
  ["batter supercaps"] = "Batteries & supercaps",
  ["batteries electrochem"] = "Batteries and Electrochemistry",
  ["batteries supercaps"] = "Batteries & Supercaps",
  ["battery energy"] = "Battery Energy",
  ["bausch lomb mag"] = "Bausch & Lomb magazine",
  ["baxter health policy rev"] = "The Baxter health policy review",
  ["bayer akad wiss math-natur kl abh (nf)"] = "Bayerische Akademie der Wissenschaften",
  ["bayer akad wiss math-natur kl sitzungsber"] = "Bayerische Akademie der Wissenschaften",
  ["bayer akad wiss philos-hist kl sitzungsber"] = "Bayerische Akademie der Wissenschaften",
  ["bayer aztebl"] = "Bayerisches Ärzteblatt : amtliches Organ der Bayerischen Landesärztekammer und ihrer Bezirksvereine",
  ["bayer cropsci j"] = "Bayer CropScience Journal",
  ["bayesian anal"] = "Bayesian Analysis",
  ["bayesian graph models biomed imaging (2014)"] = "Bayesian and grAphical models for biomedical imaging : first International Workshop, BAMBI 2014, Cambridge, MA, USA, September 18, 2014 ; revised selected papers. BAMBI (Workshop) (1st : 2014 : Cambridge, Mass.)",
  ["bayl law rev"] = "Baylor law review",
  ["baylor dent j"] = "The Baylor dental journal",
  ["baylor nurs educ"] = "Baylor nursing educator",
  ["bayreuth math schr"] = "Bayreuther Mathematische Schriften",
  ["bba clin"] = "BBA clinical",
  ["bbgnt berliner beitr gesch naturwiss tech"] = "BBGNT. Berliner Beitrage zur Geschichte der Naturwissenschaften und der Technik",
  ["bdj open"] = "BDJ open",
  ["beagle (darwin)"] = "Beagle : (Darwin)",
  ["bedside nurse"] = "Bedside nurse",
  ["bee world"] = "Bee World",
  ["begg j orthod theory treat"] = "Begg journal of orthodontic theory and treatment",
  ["behav anal"] = "The Behavior analyst",
  ["behav anal (wash d c)"] = "Behavior analysis (Washington, D.C.)",
  ["behav anal pract"] = "Behavior analysis in practice",
  ["behav anal today"] = "The behavior analyst today",
  ["behav assess"] = "Behavioral assessment",
  ["behav biol"] = "Behavioral biology",
  ["behav brain funct"] = "Behavioral and brain functions : BBF",
  ["behav brain res"] = "Behavioural brain research",
  ["behav brain sci"] = "The Behavioral and brain sciences",
  ["behav change"] = "Behaviour change : journal of the Australian Behaviour Modification Association",
  ["behav cogn neurosci rev"] = "Behavioral and cognitive neuroscience reviews",
  ["behav cogn psychother"] = "Behavioural and cognitive psychotherapy",
  ["behav cognit neurosci rev"] = "Behavioral and Cognitive Neuroscience Reviews",
  ["behav dev bull"] = "Behavioral development bulletin",
  ["behav disord"] = "Behavioral disorders",
  ["behav ecol"] = "Behavioral ecology : official journal of the International Society for Behavioral Ecology",
  ["behav ecol sociobiol"] = "Behavioral ecology and sociobiology",
  ["behav eng"] = "Behavioral engineering",
  ["behav genet"] = "Behavior genetics",
  ["behav healthc"] = "Behavioral healthcare",
  ["behav healthc tomorrow"] = "Behavioral Healthcare Tomorrow",
  ["behav inf technol"] = "Behaviour & information technology",
  ["behav med"] = "Behavioral medicine (Washington, D.C.)",
  ["behav med update"] = "Behavioral medicine update : a publication of the Society of Behavioral Medicine",
  ["behav modif"] = "Behavior modification",
  ["behav neural biol"] = "Behavioral and neural biology",
  ["behav neurol"] = "Behavioural neurology",
  ["behav neuropsychiatry"] = "Behavioral Neuropsychiatry",
  ["behav neurosci"] = "Behavioral neuroscience",
  ["behav pharmacol"] = "Behavioural Pharmacology",
  ["behav processes"] = "Behavioural processes",
  ["behav res methods"] = "Behavior research methods",
  ["behav res methods instrum comput"] = "Behavior research methods, instruments, & computers : a journal of the Psychonomic Society, Inc",
  ["behav res ther"] = "Behaviour research and therapy",
  ["behav sci"] = "Behavioral Science",
  ["behav sci (basel)"] = "Behavioral sciences (Basel, Switzerland)",
  ["behav sci law"] = "Behavioral sciences & the law",
  ["behav sci res"] = "Behavior science research",
  ["behav sleep med"] = "Behavioral sleep medicine",
  ["behav soc action j"] = "Behaviorists for Social Action journal",
  ["behav soc sci librar"] = "Behavioral & social sciences librarian",
  ["behav ther"] = "Behavior therapy",
  ["behav ther (n y n y)"] = "The Behavior therapist",
  ["behavioral res acc"] = "Behavioral Research in Accounting",
  ["behaviormetr quant approaches hum behav"] = "Behaviormetrics: Quantitative Approaches to Human Behavior",
  ["behring inst mitt"] = "Behring Institute Mitteilungen",
  ["beih ann schweiz meteorol anst"] = "Beiheft zu den Annalen der Schweizerischen meteorologischen Anstalt",
  ["beih bot centbl"] = "Beihefte zum Botanischen Centralblatt",
  ["beih jahrb schweiz natforsch ges"] = "Beiheft zum Jahrbuch der Schweizerischen naturforschenden Gesellschaft",
  ["beih nova hedwigia"] = "Beiheft zu Nova Hedwigia",
  ["beih schweiz z forstwes"] = "Beiheft zur Schweizerischen Zeitschrift für Forstwesen",
  ["beih z mykol"] = "Beiheft zur Zeitschrift für Mykologie",
  ["beih z schweiz forstver"] = "Beiheft zu den Zeitschriften des Schweiz. Forstvereins",
  ["beijing da xue xue bao"] = "Beijing Da Xue Xue Bao. Yi Xue Ban (Journal of Peking University. Health Sciences)",
  ["beijing da xue xue bao yi xue ban"] = "Beijing da xue xue bao. Yi xue ban = Journal of Peking University. Health sciences",
  ["beijing daxue xuebao ziran kexue ban"] = "Beijing Daxue Xuebao. Ziran Kexue Ban. Acta Scientiarum Naturalium Universitatis Pekinensis",
  ["beijing j pure appl math"] = "Beijing Journal of Pure and Applied Mathematics",
  ["beijing shifan daxue xuebao"] = "Beijing Shifan Daxue Xuebao. Ziran Kexue Ban. Journal of Beijing Normal University (Natural Science)",
  ["beilstein j nanotechnol"] = "Beilstein Journal of Nanotechnology",
  ["beilstein j org chem"] = "Beilstein journal of organic chemistry",
  ["beitr algebra geom"] = "Beitrage zur Algebra und Geometrie",
  ["beitr avifauna rheinl"] = "Beiträge zur Avifauna des Rheinlandes",
  ["beitr biol pflanzen"] = "Beiträge zur Biologie der Pflanzen",
  ["beitr bolzano-forsch"] = "Beitrage zur Bolzano-Forschung",
  ["beitr chem physiol pathol"] = "Beitraege zur Chemischen Physiologie und Pathologie",
  ["beitr datenverarb unternehmensforsch"] = "Beitrage Datenverarbeitung und Unternehmensforschung",
  ["beitr dtsch philol"] = "Beitrage zur deutschen Philologie",
  ["beitr entomol"] = "Beitraege zur Entomologie",
  ["beitr forschungstech"] = "Beitrage zur Forschungstechnologie",
  ["beitr forstwirtsch"] = "Beiträge für die Forstwirtschaft",
  ["beitr forstwirtsch landschökol"] = "Beiträge für Forstwirtschaft und Landschaftsökologie",
  ["beitr geobot landesaufn schweiz"] = "Beiträge zur geobotanischen Landesaufnahme der Schweiz",
  ["beitr geol schweiz: hydrol"] = "Beiträge zur Geologie der Schweiz: Hydrologie",
  ["beitr gerichtl med"] = "Beitraege zur Gerichtlichen Medizin",
  ["beitr gesch arb"] = "Beitrage zur Geschichte der Arbeiterbewegung : BZG",
  ["beitr gesch med nebengeb"] = "Beiträge zur Geschichte der Medizin und ihrer Nebengebiete",
  ["beitr gesch pharm ihrer nachbargeb"] = "Beitrage zur Geschichte der Pharmazie und ihrer Nachbargebiete",
  ["beitr hochsch wissenschaftsgesch erfurts"] = "Beiträge zur Hochschul- und Wissenschaftsgeschichte Erfurts",
  ["beitr hydrol schweiz"] = "Beiträge zur Hydrologie der Schweiz",
  ["beitr hyg epidemiol"] = "Beitrage zur Hygiene und Epidemiologie",
  ["beitr infusionsther"] = "Beitrage zur Infusionstherapie",
  ["beitr infusionsther transfusionsmed"] = "Beitrage zur Infusionstherapie und Transfusionsmedizin",
  ["beitr infusionther klin ernahr"] = "Beitrage zu Infusionstherapie und Klinische Ernahrung",
  ["beitr jagd- wildforsch"] = "Beiträge zur Jagd- und Wildforschung",
  ["beitr klin chir"] = "Beitrage zur klinischen Chirurgie",
  ["beitr klin erforsch tuberk lungenkr"] = "Beitrage zur Klinik und Erforschung der Tuberkulose und der Lungenkrankheiten",
  ["beitr klin neurol psychiatr"] = "Beitrage zur Klinischen Neurologie und Psychiatrie",
  ["beitr klin tuberk spezif tuberkuloseforsch"] = "Beitrage zur Klinik der Tuberkulose und spezifischen Tuberkulose-Forschung",
  ["beitr krebsforsch"] = "Beitrage zur Krebsforschung",
  ["beitr kryptogamenflora schweiz"] = "Beiträge zur Kryptogamenflora der Schweiz",
  ["beitr kryst mineral"] = "Beitrage zur Krystallographie und Mineralogie",
  ["beitr math"] = "Beitrage zur Mathematik",
  ["beitr natkdl forsch südwestdtschl"] = "Beiträge zur naturkundlichen Forschung in Südwestdeutschland",
  ["beitr natschutz schweiz"] = "Beiträge zum Naturschutz in der Schweiz",
  ["beitr neurochir"] = "Beitrage zur Neurochirurgie",
  ["beitr orthop traumatol"] = "Beitrage zur Orthopadie und Traumatologie",
  ["beitr pathol"] = "Beitrage zur Pathologie",
  ["beitr pathol anat"] = "Beiträge zur pathologischen Anatomie und zur allgemeinen Pathologie",
  ["beitr phys atmos"] = "Beitraege zur Physik der Atmosphaere",
  ["beitr plasmaphys"] = "Beitrage aus der Plasmaphysik",
  ["beitr rheumatol"] = "Beitrage zur Rheumatologie",
  ["beitr sexualforsch"] = "Beitrage zur Sexualforschung",
  ["beitr silikoseforsch"] = "Beitrage zur Silikose-Forschung",
  ["beitr silikoseforsch pneumokoniose"] = "Beiträge zur Silikose-Forschung (Pneumokoniose)",
  ["beitr tab int"] = "Beiträge zur Tabakforschung international",
  ["beitr trop landwirtsch veterinarmed"] = "Beiträge zur tropischen Landwirtschaft und Veterinärmedizin",
  ["beitr wurttemb apothekengesch"] = "Beiträge zur württembergischen Apothekengeschichte",
  ["beitr zur phys atmos"] = "Beiträge Zur Physik der Atmosphäre",
  ["beitr zur phys atmosphäre"] = "Beiträge Zur Physik der Atmosphäre",
  ["beiträge algebra geom"] = "Beiträge zur Algebra und Geometrie",
  ["bekhterev rev psychiatry med psychol"] = "The Bekhterev review of psychiatry and medical psychology = Obozrenie psikhiatrii i medit︠s︡inskoĭ psikhologii imeni V.M. Bekhtereva",
  ["belcast j belizean aff"] = "BELCAST journal of Belizean affairs",
  ["beleid maatsch"] = "Beleid & maatschappij",
  ["belg j bot"] = "Belgian journal of botany : bulletin de la Société royale de botanique de Belgique",
  ["belg j oper res statist comput sci"] = "Belgian Journal of Operations Research Statistics and Computer Science (JORBEL)",
  ["belg j zool"] = "Belgian journal of zoology",
  ["belg tijdschr geneesk"] = "Belgisch tijdschrift voor geneeskunde",
  ["belg tijdschr nieuwste geschied"] = "Revue belge d'histoire contemporaine. Belgisch tijdschrift voor nieuwste geschiedenis",
  ["belitung nurs j"] = "Belitung nursing journal",
  ["belizean stud"] = "Belizean studies",
  ["bell j econ"] = "Bell Journal of Economics",
  ["bell lab rec"] = "Bell Laboratories Record",
  ["bell labs tech j"] = "Bell Labs Technical Journal",
  ["bell syst tech j"] = "Bell System Technical Journal",
  ["benchcouncil trans benchmarks stand eval"] = "BenchCouncil Transactions on Benchmarks, Standards and Evaluations",
  ["benchmark pap hum physiol"] = "Benchmark papers in human physiology",
  ["benchmarking qual manage technol"] = "Benchmarking for Quality Management & Technology",
  ["benchmarking: int j"] = "Benchmarking: An International Journal",
  ["benders health care law mon"] = "Bender's health care law monthly",
  ["benef microbes"] = "Beneficial microbes",
  ["benefic microbes"] = "Beneficial Microbes",
  ["benefits q"] = "Benefits quarterly",
  ["bengal past present"] = "Bengal, past & present : journal of the Calcutta Historical Society",
  ["bengal public health j"] = "The Bengal public health journal",
  ["benha vet med j"] = "Benha veterinary medical journal",
  ["beni suef univ j basic appl sci"] = "Beni-Suef University journal of basic and applied sciences",
  ["beni-suef univ j basic appl sci"] = "Beni-Suef University Journal of Basic and Applied Sciences",
  ["ber akad natschutz landschpfl"] = "Berichte der Akademie für Naturschutz und Landschaftspflege",
  ["ber atwd"] = "Bericht ATWD",
  ["ber bayer landesanst wald forstwirtsch"] = "Berichte aus der Bayerischen Landesanstalt für Wald und Forstwirtschaft",
  ["ber biol"] = "Berita biologi",
  ["ber bonn univ poliklin mund zahn kieferkr"] = "Berichte Aus Der Bonner Universitatsklinik Und Poliklinik Fur Mund-, Zahn- Und Kieferkrankheiten",
  ["ber bot-zool ges liecht-sargans-werdenberg"] = "Berichte der Botanisch-Zoologischen Gesellschaft Liechtenstein- Sargans-Werdenberg",
  ["ber bundesamt wasser geol"] = "Berichte des Bundesamtes für Wasser und Geologie",
  ["ber bunsen ges"] = "Berichte der Bunsen-Gesellschaft",
  ["ber bunsen ges  phys chem"] = "Berichte der Bunsen-Gesellschaft fur Physikalische Chemie",
  ["ber bunsen ges phys chem"] = "Berichte der Bunsen-Gesellschaft fur Physikalische Chemie",
  ["ber bunsen-ges phys chem"] = "Berichte der Bunsen-Gesellschaft Physical Chemistry Chemical Physics",
  ["ber bunsenges phys chem"] = "Berichte der Bunsengesellschaft für Physikalische Chemie",
  ["ber dtsch bot ges"] = "Berichte der Deutschen Botanischen Gesellschaft",
  ["ber dtsch chem ges"] = "Berichte der Deutschen Chemischen Gesellschaft",
  ["ber dtsch ges geol wiss reihe a"] = "Berichte der Deutschen Gesellschaft fur Geologische Wissenschaft Reihe A: Geologie und Palaeontologie",
  ["ber dtsch ges geol wiss reihe b"] = "Berichte der Deutschen Gesellschaft fur Geologische Wissenschaft Reihe B: Mineralogie und Lagerstattenforschung",
  ["ber dtsch keram ges"] = "Berichte der Deutschen Keramischen Gesellschaft",
  ["ber dtsch landeskd"] = "Berichte zur deutschen Landeskunde",
  ["ber dtsch mineral ges"] = "Berichte der Deutschen Mineralogischen Gesellschaft",
  ["ber dtsch pharm ges"] = "Berichte der Deutschen Pharmazeutischen Gesellschaft",
  ["ber dtsch phys ges"] = "Berichte der Deutschen Physikalischen Gesellschaft",
  ["ber dtsch wetterd"] = "Berichte des Deutschen Wetterdienstes",
  ["ber dtsch wetterdienstes"] = "Berichte der Deutschen Wetterdienstes",
  ["ber eidgenöss anst forstl verswes"] = "Berichte Eidg. Anstalt für das forstliche Versuchswesen",
  ["ber eidgenöss forschanst wald schnee landsch"] = "Berichte der Eidg. Forschungsanstalt für Wald, Schnee und Landschaft",
  ["ber forschzent waldökosyst"] = "Berichte des Forschungszentrums Waldökosysteme",
  ["ber freibg forstl forsch"] = "Berichte Freiburger Forstliche Forschung",
  ["ber geobot inst eidgenöss tech hochsch, stift rübel"] = "Berichte des Geobotanischen Institutes der Eidg. Technischen Hochschule, Stiftung Rübel",
  ["ber gesamte physiol exp pharmakol"] = "Berichte uber die Gesamte Physiologie und Experimentalle Pharmakologie",
  ["ber gesellsch math datenverarb"] = "Berichte der Gesellschaft für Mathematik und Datenverarbeitung",
  ["ber landwirtsch"] = "Berichte uber Landwirtschaft",
  ["ber math"] = "Berichte aus der Mathematik",
  ["ber naturforsch ges freibg br"] = "Berichte. Naturforschende Gesellschaft zu Freiburg i. B",
  ["ber phys med ges wurzbg"] = "Berichte. Physikalisch-Medicinische Gesellschaft, Würzburg",
  ["ber raumforsch raumplan"] = "Berichte zur Raumforschung und Raumplanung (Vienna, Austria)",
  ["ber schwyzerische natforsch ges"] = "Berichte der Schwyzerischen Naturforschenden Gesellschaft",
  ["ber st gallen natwiss ges"] = "Berichte der Sankt Gallischen Naturwissenschaftlichen Gesellschaft",
  ["ber versanst wasserbau hydrol glaziol eidgenöss tech hochsch zür"] = "Berichte der Versuchsanstalt für Wasserbau, Hydrologie und Glaziologie an der Eidg. Technischen Hochschule Zürich",
  ["ber wiss"] = "Berichte zur Wissenschaftsgeschichte",
  ["ber wiss-gesch"] = "Berichte zur Wissenschaftsgeschichte",
  ["ber wissenschaftsgesch"] = "Berichte zur Wissenschaftsgeschichte",
  ["ber wissgesch"] = "Berichte zur Wissenschaftsgeschichte",
  ["ber zusammenkunft dtsch ophthalmol ges"] = "Bericht über die Zusammenkunft. Deutsche Ophthalmologische Gesellschaft",
  ["berc bull"] = "BERC bulletin",
  ["bereave care"] = "Bereavement care : for all those who help the bereaved",
  ["beretning oslo"] = "Beretning. Oslo. Helserådet",
  ["berg huttenmannische monatshefte"] = "Berg- und hüttenmännische Monatshefte",
  ["berg- huettenmaenn monatsh"] = "Berg- und Huettenmaennische Monatshefte",
  ["berichte aus dem institut fur statistik und versicherungsmathematik und aus dem institut fur angewandte statistik de"] = "Berichte aus dem Institut fur Statistik und Versicherungsmathematik und aus dem Institut fur Angewandte Statistik der Freien Universitat Berlin",
  ["berichte aus dem institut fur statistik und wirtschaftsmathematik und aus dem institut fur angewandte statistik der"] = "Berichte aus dem Institut fur Statistik und Wirtschaftsmathematik und aus dem Institut fur Angewandte Statistik der Freien Universitat Berlin",
  ["berichte german chapter acm"] = "Berichte des German Chapter of the ACM",
  ["berita jururawat"] = "Berita jururawat",
  ["berkala ilmu kedokteran"] = "Berkala ilmu kedokteran",
  ["berkeley j sociol"] = "Berkeley journal of sociology",
  ["berkeley munic code 1976 berkeley calif"] = "Berkeley municipal code, 1976 : a codification of the general ordinances of the city of Berkeley, California. Berkeley (Calif.)",
  ["berkeley papers hist sci"] = "Berkeley Papers in History of Science",
  ["berkeley symp math stat probab"] = "Berkeley Symposium on Mathematical Statistics and Probability",
  ["berkeley technol law j"] = "Berkeley technology law journal",
  ["berkeley womens law j"] = "Berkeley women's law journal",
  ["berl beitr archaeom"] = "Berliner Beitraege zur Archaeometrie",
  ["berl geogr stud"] = "Berliner geographische Studien",
  ["berl geowiss abh, reihe a"] = "Berliner Geowissenschaftliche Abhandlungen, Reihe A: Geologie und Palaeontologie",
  ["berl hist stud"] = "Berliner historische Studien",
  ["berl j soziol"] = "Berliner journal für Soziologie",
  ["berl med"] = "Berliner Medizin",
  ["berl munch tierarztl wochenschr"] = "Berliner und Münchener tierärztliche Wochenschrift",
  ["berl palaobiologische abh"] = "Berliner paläobiologische Abhandlungen",
  ["berl stat"] = "Berliner Statistik",
  ["berl stat mon schr"] = "Berliner Statistik. Monatsschrift",
  ["berl tierarztl wochenschr"] = "Berliner tierärztliche Wochenschrift",
  ["berlin stud anc world"] = "Berlin Studies of the Ancient World",
  ["berlin-brandenburgische akad wiss ber abh"] = "Berlin-Brandenburgische Akademie der Wissenschaften",
  ["berliner entomol z"] = "Berliner entomologische Zeitschrift",
  ["berliner j soz"] = "Berliner Journal für Soziologie",
  ["berliner studienreihe math"] = "Berliner Studienreihe zur Mathematik",
  ["bern wald"] = "Berner Wald",
  ["bern z gesch heimkd"] = "Berner Zeitschrift für Geschichte und Heimatkunde",
  ["bern ztg"] = "Berner Zeitung",
  ["berner beitr gesch med naturwiss neue folge"] = "Berner Beiträge zur Geschichte der Medizin und der Naturwissenschaften",
  ["bernoulli (andover)"] = "Bernoulli : official journal of the Bernoulli Society for Mathematical Statistics and Probability",
  ["bernstein ser comput neurosci"] = "Bernstein Series in Computational Neuroscience",
  ["berufsdermatosen (monogr)"] = "Berufsdermatosen. Monographien",
  ["best evid health care cochrane colloq 7th 1999 univ s tommaso daquino"] = "The best evidence for health care : the role of the Cochrane Collaboration. Cochrane Colloquium (7th : 1999 : Università S. Tommaso d'Aquino)",
  ["best pract benchmarking healthc"] = "Best practices and benchmarking in healthcare : a practical journal for clinical and management application",
  ["best pract ment health"] = "Best practices in mental health",
  ["best pract res clin anaesthesiol"] = "Best practice & research. Clinical anaesthesiology",
  ["best pract res clin endocrinol metab"] = "Best practice & research. Clinical endocrinology & metabolism",
  ["best pract res clin gastroenterol"] = "Best practice & research. Clinical gastroenterology",
  ["best pract res clin haematol"] = "Best practice & research. Clinical haematology",
  ["best pract res clin obstet gynaecol"] = "Best practice & research. Clinical obstetrics & gynaecology",
  ["best pract res clin rheumatol"] = "Best practice & research. Clinical rheumatology",
  ["bests rev life health insur ed"] = "Best's review. Life-health insurance edition",
  ["beton- stahlbetonbau"] = "Beton- und Stahlbetonbau",
  ["betonwerk fertigteil tech"] = "Betonwerk und Fertigteil Technik",
  ["better health (lond)"] = "Better health",
  ["better homes gard"] = "Better homes and gardens",
  ["between species"] = "Between the species : a journal of ethics",
  ["beverage plant res"] = "Beverage Plant Research",
  ["bevolking gezin"] = "Bevolking en gezin",
  ["beyond relief"] = "Beyond relief",
  ["bfpp bull fr peche prot milieux aquat"] = "BFPP : bulletin français de la pêche et de la protection des milieux aquatiques : bulletin français de la pêche et de la pisciculture",
  ["bhm berg- huttenmann monatsh"] = "BHM Berg- und Huttenmannische Monatshefte",
  ["bi-mon res notes"] = "Bi-monthly Research Notes",
  ["bib lit modern"] = "Bibliotheque De Litterature Moderne",
  ["bib mitt"] = "BIB-Mitteilungen : Informationen aus dem Bundesinstitut für Bevölkerungsforschung",
  ["bibe 2004"] = "BIBE 2004 : proceedings : 19-21 May, 2004, Taichung, Taiwan, ROC. IEEE International Symposium on Bioinformatics and Bioengineering (4th : 2004 : Taichung, Taiwan)",
  ["bibl anat"] = "Bibliotheca anatomica",
  ["bibl avtomat"] = "Biblioteka po Avtomatike",
  ["bibl bibliogr ital"] = "Biblioteca di Bibliografia Italiana",
  ["bibl bot"] = "Bibliotheca Botanica",
  ["bibl cardiol"] = "Bibliotheca cardiologica",
  ["bibl cienc esp"] = "Biblioteca de la Ciencia Española",
  ["bibl ec chartes"] = "Bibliothèque de l'École des chartes",
  ["bibl ekonom"] = "Biblioteka Ekonometryczna",
  ["bibl gastroenterol"] = "Bibliotheca gastroenterologica",
  ["bibl gynaecol"] = "Bibliotheca gynaecologica",
  ["bibl haematol"] = "Bibliotheca haematologica",
  ["bibl hig instituta nr srb"] = "Biblioteka Higijenskog instituta NR Srbije",
  ["bibl hist sci"] = "Bibliothèque d’Histoire des Sciences",
  ["bibl humanisme renaiss"] = "Bibliothèque d'humanisme et renaissance; travaux et documents",
  ["bibl inzh"] = "Biblioteka Inzhenera",
  ["bibl laeger"] = "Bibliotek for laeger",
  ["bibl lichenol"] = "Bibliotheca lichenologica",
  ["bibl mat"] = "\\cyr Biblioteka Matematika",
  ["bibl microbiol"] = "Bibliotheca microbiologica",
  ["bibl mycol"] = "Bibliotheca mycologica",
  ["bibl nuncius studi testi"] = "Biblioteca di Nuncius: Studi e Testi",
  ["bibl nutr dieta"] = "Bibliotheca nutritio et dieta",
  ["bibl ophthalmol"] = "Bibliotheca ophthalmologica : supplementa ad ophthalmologica",
  ["bibl paediatr"] = "Bibliotheca paediatrica",
  ["bibl phycol"] = "Bibliotheca phycologica",
  ["bibl physis"] = "Biblioteca di Physis",
  ["bibl psychiatr"] = "Bibliotheca psychiatrica",
  ["bibl psychiatr neurol"] = "Bibliotheca psychiatrica et neurologica",
  ["bibl radiol"] = "Bibliotheca radiologica",
  ["bibl rev mat iberoamericana"] = "Biblioteca de la Revista Matemática Iberoamericana",
  ["bibl sci"] = "Bibliotheque Scientifique",
  ["bibl stor sci"] = "Biblioteca di storia della scienza",
  ["bibl stor t"] = "Biblioteca Storica Toscana : Sezione Di Storia Del Risorgimento",
  ["bibl storia sci"] = "Biblioteca di Storia della Scienza",
  ["bibl stud-mat"] = "\\cyr Biblioteka Studenta-Matematika",
  ["bibl techn allgemeinwiss"] = "Bibliothek des Technischen Allgemeinwissens",
  ["bibl tuberc"] = "Bibliotheca tuberculosea",
  ["bibl uchitelya mat"] = "Biblioteka Uchitelya Matematiki",
  ["bibl wiss"] = "Bibliothek und Wissenschaft",
  ["bibl “znanie”"] = "Biblioteka “Znanie”",
  ["biblic archaeol"] = "The Biblical archaeologist",
  ["bibliogr hisp"] = "Bibliografía hispánica",
  ["bibliogr hist psychol psychiatry"] = "Bibliographies in the history of psychology and psychiatry",
  ["bibliogr int demogr hist"] = "Bibliographie internationale de la démographie historique",
  ["bibliogr ital stor sci"] = "Bibliografia italiana di storia della scienza",
  ["bibliogr sci nat helv"] = "Bibliographia Scientiae Naturalis Helvetica",
  ["bibliographr reprod"] = "Bibliography of reproduction; a classified monthly list of references compiled from the research literature, vertebrates including man",
  ["biedermanns zentralbl"] = "Biedermanns Zentralblatt",
  ["biedermanns zentralbl abt a"] = "Biedermanns Zentralblatt Abteilung A: Allgemeiner und Referierender Teil",
  ["biedermanns zentralbl abt b"] = "Biedermanns Zentralblatt Abteilung B: Tierernahrung",
  ["bielefelder schr didakt math"] = "Bielefelder Schriften zur Didaktik der Mathematik",
  ["bien naitre"] = "Bien-naître. Association béninoise pour la promotion de la famille",
  ["bienn rev couns psychol"] = "Biennial review of counseling psychology",
  ["big data"] = "Big data",
  ["big data anal"] = "Big data analytics",
  ["big data anal knowl discov (2017)"] = "Big data analytics and knowledge discovery : 19th International Conference, DaWaK 2017, Lyon, France, August 28-31, 2017, Proceedings. DaWaK (Conference) (19th : 2017 : Lyon, France)",
  ["big data inf anal"] = "Big Data and Information Analytics",
  ["big data manag"] = "Big Data Management",
  ["big data min anal"] = "Big Data Mining and Analytics",
  ["big data res"] = "Big Data Research",
  ["big data soc"] = "Big data & society",
  ["big earth data"] = "Big earth data",
  ["big integr artif intell"] = "Big and Integrated Artificial Intelligence",
  ["big quest"] = "The Big Questions",
  ["biharean biol"] = "Biharean biologist",
  ["bijblijven (amst)"] = "Bijblijven (Amsterdam, Netherlands)",
  ["bijdr geschied"] = "Bijdragen tot de geschiedenis",
  ["bijdr meded"] = "Bijdragen en mededelingen (Gelre, Vereeniging tot Beoefening van Geldersche Geschiedenis, Oudheidkunde, en Recht)",
  ["bijdr meded geschied ned"] = "Bijdragen en mededelingen betreffende de geschiedenis der Nederlanden",
  ["bijdr taal land volkenkd"] = "Bijdragen tot de taal-, land- en volkenkunde",
  ["biken j"] = "Biken journal",
  ["bild wiss"] = "Bild der Wissenschaft",
  ["bildmess luftbildwes"] = "Bildmessung und Luftbildwesen",
  ["biling (camb engl)"] = "Bilingualism (Cambridge, England)",
  ["biling res j"] = "Bilingual research journal",
  ["bilt hematol transfuz"] = "Bilten za hematologiju i transfuziju",
  ["bilt mednar fed zob teh"] = "Bilten Mednarodne Federacije Zobotehnikov = Bulletin Of The International Federation Of Dental Technicians. Bulletin De La Federation Internationale Prothesistes Dentaires. Bulletin Der Internationalen Foderation Der Zahntechniker",
  ["bilt udruz ortodonata jugosl"] = "Bilten Udruz̆enja ortodonata Jugoslavije = Bulletin of Orthodontic Society of Yugoslavia",
  ["bing du xue bao"] = "Bing du xue bao = Chinese journal of virology",
  ["binocul vis strabismus q"] = "Binocular vision & strabismus quarterly",
  ["binocul vis strabolog q simms romano"] = "Binocular vision & strabology quarterly, Simms-Romano's",
  ["bio protoc"] = "Bio-protocol",
  ["bio web conf"] = "BIO web of conferences",
  ["bio-des manuf"] = "Bio-Design and Manufacturing",
  ["bio-med mater eng"] = "Bio-Medical Materials and Engineering",
  ["bioact carbohydr diet fibre"] = "Bioactive Carbohydrates and Dietary Fibre",
  ["bioact carbohydr dietary fibre"] = "Bioactive Carbohydrates and Dietary Fibre",
  ["bioact compd health dis"] = "Bioactive compounds in health and disease",
  ["bioact mater"] = "Bioactive materials",
  ["bioanal rev"] = "Bioanalytical reviews",
  ["bioarchaeol int"] = "Bioarchaeology international",
  ["biocatal agric biotechnol"] = "Biocatalysis and agricultural biotechnology",
  ["biocatal biotransform"] = "Biocatalysis and Biotransformation",
  ["biocatal biotransformation"] = "Biocatalysis and biotransformation",
  ["bioceram dev appl"] = "Bioceramics development and applications",
  ["biochem (lond)"] = "The biochemist",
  ["biochem (mosc) suppl ser a membr cell biol"] = "Biochemistry (Moscow) Supplement. Series A, Membrane and cell biology",
  ["biochem anal biochem"] = "Biochemistry and analytical biochemistry : current research",
  ["biochem arch"] = "Biochemical archives",
  ["biochem biophys rep"] = "Biochemistry and biophysics reports",
  ["biochem biophys res commun"] = "Biochemical and Biophysical Research Communications",
  ["biochem biotechnol res"] = "Biochemistry and biotechnology research (Delta)",
  ["biochem cell arch"] = "Biochemical and cellular archives",
  ["biochem cell biol"] = "Biochemistry and Cell Biology",
  ["biochem clin"] = "Biochemical clinics",
  ["biochem educ"] = "Biochemical education",
  ["biochem eng j"] = "Biochemical engineering journal",
  ["biochem exp biol"] = "Biochemistry and experimental biology",
  ["biochem genet"] = "Biochemical Genetics",
  ["biochem insights"] = "Biochemistry insights",
  ["biochem int"] = "Biochemistry international",
  ["biochem j"] = "Biochemical Journal",
  ["biochem med"] = "Biochemical medicine",
  ["biochem med (zagreb)"] = "Biochemia medica",
  ["biochem med metab biol"] = "Biochemical Medicine and Metabolic Biology",
  ["biochem mol biol (n y)"] = "Biochemistry and molecular biology (New York, N.Y.)",
  ["biochem mol biol educ"] = "Biochemistry and molecular biology education : a bimonthly publication of the International Union of Biochemistry and Molecular Biology",
  ["biochem mol biol int"] = "Biochemistry and Molecular Biology International",
  ["biochem mol biol j"] = "Biochemistry & molecular biology journal",
  ["biochem mol med"] = "Biochemical and Molecular Medicine",
  ["biochem mosc suppl b biomed chem"] = "Biochemistry (Moscow) Supplement. Series B, Biomedical chemistry",
  ["biochem pharmacol"] = "Biochemical pharmacology",
  ["biochem pharmacol (amsterdam, neth)"] = "Biochemical Pharmacology (Amsterdam, Netherlands)",
  ["biochem pharmacol (los angel)"] = "Biochemistry & pharmacology : open access",
  ["biochem physiol"] = "Biochemistry & physiology",
  ["biochem physiol pflanz"] = "Biochemie und Physiologie der Pflanzen : BPP",
  ["biochem prep"] = "Biochemical Preparations",
  ["biochem res int"] = "Biochemistry research international",
  ["biochem soc symp"] = "Biochemical Society symposium",
  ["biochem soc trans"] = "Biochemical Society Transactions",
  ["biochem syst ecol"] = "Biochemical Systematics and Ecology",
  ["biochem z"] = "Biochemische Zeitschrift",
  ["biochem: indian j"] = "BioChemistry: An Indian Journal",
  ["biochemica (indianapolis)"] = "Biochemica (Indianapolis, Ind.)",
  ["biochemistry (mosc)"] = "Biochemistry. Biokhimii︠a︡",
  ["biochemistry (rajkot, india)"] = "BioChemistry (Rajkot, India)",
  ["biochim biophys acta"] = "Biochimica et Biophysica Acta",
  ["biochim biophys acta bioenerg"] = "Biochimica et biophysica acta. Bioenergetics",
  ["biochim biophys acta biomembr"] = "Biochimica et biophysica acta. Biomembranes",
  ["biochim biophys acta gen subj"] = "Biochimica et biophysica acta. General subjects",
  ["biochim biophys acta gene regul mech"] = "Biochimica et biophysica acta. Gene regulatory mechanisms",
  ["biochim biophys acta mol basis dis"] = "Biochimica et biophysica acta. Molecular basis of disease",
  ["biochim biophys acta mol cell biol lipids"] = "Biochimica et biophysica acta. Molecular and cell biology of lipids",
  ["biochim biophys acta mol cell res"] = "Biochimica et biophysica acta. Molecular cell research",
  ["biochim biophys acta proteins proteom"] = "Biochimica et biophysica acta. Proteins and proteomics",
  ["biochim biophys acta rev cancer"] = "Biochimica et biophysica acta. Reviews on cancer",
  ["biochim biophys acta, adv"] = "Biochimica et Biophysica Acta, Advances",
  ["biochim biophys acta, bioenerg"] = "Biochimica et Biophysica Acta, Bioenergetics",
  ["biochim biophys acta, biomembr"] = "Biochimica et Biophysica Acta, Biomembranes",
  ["biochim biophys acta, biophys incl photosynth"] = "Biochimica et Biophysica Acta, Biophysics Including Photosynthesis",
  ["biochim biophys acta, biophys incl photsynth"] = "Biochimica et Biophysica Acta (BBA) - Biophysics including Photosynthesis",
  ["biochim biophys acta, enzymol"] = "Biochimica et Biophysica Acta, Enzymology",
  ["biochim biophys acta, enzymol biol oxid"] = "Biochimica et Biophysica Acta, Enzymology and Biological Oxidation",
  ["biochim biophys acta, gen subj"] = "Biochimica et Biophysica Acta, General Subjects",
  ["biochim biophys acta, gene regul mech"] = "Biochimica et Biophysica Acta, Gene Regulatory Mechanisms",
  ["biochim biophys acta, gene struct expression"] = "Biochimica et Biophysica Acta, Gene Structure and Expression",
  ["biochim biophys acta, lipids lipid metab"] = "Biochimica et Biophysica Acta, Lipids and Lipid Metabolism",
  ["biochim biophys acta, mol basis dis"] = "Biochimica et Biophysica Acta, Molecular Basis of Disease",
  ["biochim biophys acta, mol cell biol lipids"] = "Biochimica et Biophysica Acta, Molecular and Cell Biology of Lipids",
  ["biochim biophys acta, mol cell res"] = "Biochimica et Biophysica Acta, Molecular Cell Research",
  ["biochim biophys acta, mucroproteins mucopolysaccharides"] = "Biochimica et Biophysica Acta, Mucoproteins and Mucopolysaccharides",
  ["biochim biophys acta, nucleic acids protein synth"] = "Biochimica et Biophysica Acta, Nucleic Acids and Protein Synthesis",
  ["biochim biophys acta, protein struct"] = "Biochimica et Biophysica Acta, Protein Structure",
  ["biochim biophys acta, protein struct mol enzymol"] = "Biochimica et Biophysica Acta, Protein Structure and Molecular Enzymology",
  ["biochim biophys acta, proteins proteomics"] = "Biochimica et Biophysica Acta, Proteins and Proteomics",
  ["biochim biophys acta, rev bioenerg"] = "Biochimica et Biophysica Acta, Reviews on Bioenergetics",
  ["biochim biophys acta, rev biomembr"] = "Biochimica et Biophysica Acta, Reviews on Biomembranes",
  ["biochim biophys acta, rev cancer"] = "Biochimica et Biophysica Acta, Reviews on Cancer",
  ["biochim biophys acta, spec sect biophys subj"] = "Biochimica et Biophysica Acta (BBA) - Specialized Section on Biophysical Subjects",
  ["biochim biophys acta, spec sect enzymol subj"] = "Biochimica et Biophysica Acta (BBA) - Specialized Section on Enzymological Subjects",
  ["biochim biophys acta, spec sect lipids relat subj"] = "Biochimica et Biophysica Acta (BBA) - Specialized Section on Lipids and Related Subjects",
  ["biochim biophys acta, spec sect mucoproteins"] = "Mucopolysaccharides Biochimica et Biophysica Acta, Specialized Section on Mucoproteins and Mucopolysaccharides",
  ["biochim biophys acta, spec sect mucoproteins mucopolysaccharides"] = "Biochimica et Biophysica Acta (BBA) - Specialized Section on Mucoproteins and Mucopolysaccharides",
  ["biochim biophys acta, spec sect nucleic acids relat"] = "Subj. Biochimica et Biophysica Acta, Specialized Section on Nucleic Acids and Related Subjects",
  ["biochim biophys acta, spec sect nucleic acids relat subj"] = "Biochimica et Biophysica Acta (BBA) - Specialized Section on Nucleic Acids and Related Subjects",
  ["biochim clin"] = "Biochimica clinica",
  ["biochim open"] = "Biochimie open",
  ["biochip j"] = "Biochip journal",
  ["bioconjug chem"] = "Bioconjugate chemistry",
  ["bioconjugate chem"] = "Bioconjugate Chemistry",
  ["biocontrol (dordr)"] = "BioControl (Dordrecht, Netherlands)",
  ["biocontrol sci"] = "Biocontrol science",
  ["biocontrol sci technol"] = "Biocontrol Science and Technology",
  ["biocybern biomed eng"] = "Biocybernetics and biomedical engineering",
  ["biodata min"] = "BioData mining",
  ["biodemogr social biol"] = "Biodemography and Social Biology",
  ["biodemography soc biol"] = "Biodemography and social biology",
  ["biodes manuf"] = "Bio-design and manufacturing",
  ["biodes res"] = "Biodesign Research",
  ["biodivers conserv"] = "Biodiversity and conservation",
  ["biodivers data j"] = "Biodiversity data journal",
  ["biodivers ecol"] = "Biodiversity and ecology = Biodiversität und Ökologie",
  ["biodivers inf sci stand"] = "Biodiversity Information Science and Standards",
  ["biodivers j"] = "Biodiversity journal",
  ["biodivers lett"] = "Biodiversity Letters",
  ["biodivers res conserv"] = "Biodiversity research and conservation",
  ["biodivers sci"] = "Biodiversity Science",
  ["biodiversity (nepean)"] = "Biodiversity (Nepean, Ont.)",
  ["bioelectrochem bioenerg"] = "Bioelectrochemistry and Bioenergetics",
  ["bioelectron med"] = "Bioelectronic medicine",
  ["bioelectron med (lond)"] = "Bioelectronics in medicine",
  ["bioenergy res"] = "Bioenergy research",
  ["bioeng bugs"] = "Bioengineered bugs",
  ["bioeng today"] = "Bioengineering Today",
  ["bioeng transl med"] = "Bioengineering & translational medicine",
  ["bioengineering (basel)"] = "Bioengineering (Basel, Switzerland)",
  ["bioet debat"] = "Bioètica & debat",
  ["bioeth bull"] = "Bioethics bulletin (Washington, D.C.)",
  ["bioeth exam"] = "Bioethics examiner",
  ["bioeth news"] = "Bioethics news",
  ["bioeth res notes"] = "Bioethics research notes",
  ["bioethica belg"] = "Bioethica Belgica",
  ["bioethics dig"] = "Bioethics digest",
  ["bioethics forum"] = "Bioethics forum",
  ["bioethics northwest"] = "Bioethics Northwest",
  ["bioethics q"] = "Bioethics quarterly",
  ["biofeedback self regul"] = "Biofeedback and self-regulation",
  ["bioforum eur"] = "BIOforum Europe : trends and techniques in life science research",
  ["biofuel bioprod biorefin"] = "Biofuels, bioproducts & biorefining : Biofpr",
  ["biofuel res j"] = "Biofuel Research Journal",
  ["biofuels, bioprod biorefin"] = "Biofuels, Bioproducts and Biorefining",
  ["biog amines"] = "Biogenic Amines",
  ["biogeosci discuss"] = "Biogeosciences Discussions",
  ["biogr geneal master index"] = "Biography and genealogy master index",
  ["biogr mem fellows r soc"] = "Biographical memoirs of fellows of the Royal Society. Royal Society (Great Britain)",
  ["biogr mem natl acad sci"] = "Biographical memoirs. National Academy of Sciences (U.S.)",
  ["bioinf adv"] = "Bioinformatics Advances",
  ["bioinf biol insights"] = "Bioinformatics and Biology Insights",
  ["bioinform biol insights"] = "Bioinformatics and biology insights",
  ["bioinform biomed eng (2019)"] = "Bioinformatics and Biomedical Engineering : 7th International Work-Conference, IWBBIO 2019, Granada, Spain, May 8-10, 2019, Proceedings, Parts I and II. IWBBIO (Conference) (7th : 2019 : Granada, Spain)",
  ["bioinform res appl (2018)"] = "Bioinformatics research and applications : 14th International Symposium, ISBRA 2018, Beijing, China, June 8-11, 2018, Proceedings. ISBRA (Conference) (14th : 2018 : Beijing, China)",
  ["bioinorg chem"] = "Bioinorganic chemistry",
  ["bioinorg chem appl"] = "Bioinorganic chemistry and applications",
  ["bioinorg react mech"] = "Bioinorganic reaction mechanisms",
  ["bioinspir biomim"] = "Bioinspiration & biomimetics",
  ["bioinspiration biomimetics"] = "Bioinspiration & Biomimetics",
  ["bioinspired biomimetic nanobiomater"] = "Bioinspired, Biomimetic and Nanobiomaterials",
  ["biointerface res appl chem"] = "Biointerface research in applied chemistry",
  ["bioinvasions rec"] = "Bioinvasions records",
  ["biokhimiya (moscow)"] = "Biokhimiya (Moscow)",
  ["biol agric hortic"] = "Biological Agriculture & Horticulture",
  ["biol ambient"] = "Biologia ambientale",
  ["biol aujourdhui"] = "Biologie aujourd'hui",
  ["biol biomed rep"] = "Biological and biomedical reports",
  ["biol blood marrow transplant"] = "Biology of blood and marrow transplantation : journal of the American Society for Blood and Marrow Transplantation",
  ["biol bull"] = "The Biological bulletin",
  ["biol bull acad sci ussr"] = "Biology bulletin of the Academy of Sciences of the USSR",
  ["biol bull rev"] = "Biology Bulletin Reviews",
  ["biol bull russ acad sci"] = "Biology bulletin of the Russian Academy of Sciences",
  ["biol cell"] = "Biology of the cell",
  ["biol chem"] = "Biological Chemistry",
  ["biol chem hoppe seyler"] = "Biological chemistry Hoppe-Seyler",
  ["biol chem hoppe seyler suppl"] = "Biological chemistry Hoppe-Seyler. Supplement",
  ["biol chem hoppe-seyler"] = "Biological Chemistry Hoppe-Seyler",
  ["biol conserv"] = "Biological conservation",
  ["biol control"] = "Biological Control",
  ["biol cybern"] = "Biological cybernetics",
  ["biol cybernet"] = "Biological Cybernetics",
  ["biol direct"] = "Biology direct",
  ["biol divers conserv"] = "Biological diversity and conservation",
  ["biol eng"] = "Biological Engineering",
  ["biol eng med"] = "Biology, engineering and medicine",
  ["biol eng trans"] = "Biological engineering transactions",
  ["biol environ: proc r ir acad"] = "Biology and Environment: Proceedings of the Royal Irish Academy",
  ["biol fertil soils"] = "Biology and Fertility of Soils",
  ["biol forum"] = "Biological forum : an international journal",
  ["biol futur"] = "Biologia futura",
  ["biol futura"] = "Biologia Futura",
  ["biol gallo-hellenica"] = "Biologia Gallo-hellenica",
  ["biol gastroenterol (paris)"] = "Biologie et gastro-entérologie",
  ["biol hum aff"] = "Biology and human affairs : a British Social Hygiene Council publication",
  ["biol imaging"] = "Biological Imaging",
  ["biol inspired cognit archit"] = "Biologically Inspired Cognitive Architectures",
  ["biol invasions"] = "Biological invasions",
  ["biol j linn soc"] = "Biological Journal of the Linnean Society",
  ["biol j linn soc lond"] = "Biological journal of the Linnean Society. Linnean Society of London",
  ["biol lat"] = "Biologica Latina",
  ["biol lett"] = "Biology letters",
  ["biol listy"] = "Biologické listy",
  ["biol listy suppl"] = "Biologické listy. Supplementum",
  ["biol magn reson"] = "Biological magnetic resonance",
  ["biol mar mediterr"] = "Biologia marina mediterranea",
  ["biol mass spectrom"] = "Biological Mass Spectrometry",
  ["biol med"] = "Biologie et médecine",
  ["biol med (aligarh)"] = "Biology and medicine (Aligarh)",
  ["biol med (niteroi braz)"] = "Biologia medica",
  ["biol med (paris)"] = "Biologie médicale",
  ["biol med phys biomed eng"] = "Biological and Medical Physics, Biomedical Engineering",
  ["biol med res div semiannu rep (argonne national laboratory)"] = "Biological and Medical Research Division semiannual report. Argonne National Laboratory. Division of Biological and Medical Research",
  ["biol mem"] = "Biological memoirs",
  ["biol membr"] = "Biological Membranes",
  ["biol membrany"] = "Biologicheskie membrany",
  ["biol met"] = "Biology of metals",
  ["biol methods protoc"] = "Biology methods & protocols",
  ["biol mood anxiety disord"] = "Biology of mood & anxiety disorders",
  ["biol nauki"] = "Biologicheskie Nauki",
  ["biol nauki (moscow)"] = "Biologicheskie Nauki (Moscow)",
  ["biol neonat"] = "Biologia neonatorum. Neo-natal studies",
  ["biol neonate"] = "Biology of the neonate",
  ["biol nyssana"] = "Biologica nyssana",
  ["biol oceanogr"] = "Biological Oceanography",
  ["biol open"] = "Biology open",
  ["biol pharm bull"] = "Biological & pharmaceutical bulletin",
  ["biol philos"] = "Biology & philosophy",
  ["biol plant"] = "Biologia plantarum",
  ["biol proced online"] = "Biological procedures online",
  ["biol psichiatr psichofarmakol"] = "Biologinė psichiatrija ir psichofarmakologija = Biological psychiatry and psychopharmacology",
  ["biol psychiatry"] = "Biological psychiatry",
  ["biol psychiatry cogn neurosci neuroimaging"] = "Biological psychiatry. Cognitive neuroscience and neuroimaging",
  ["biol psychiatry: cognit neurosci neuroimaging"] = "Biological Psychiatry: Cognitive Neuroscience and Neuroimaging",
  ["biol psychol"] = "Biological psychology",
  ["biol psychol bull"] = "Biological psychology bulletin",
  ["biol reprod"] = "Biology of reproduction",
  ["biol reprod suppl"] = "Biology of reproduction. Supplement",
  ["biol res"] = "Biological research",
  ["biol res nurs"] = "Biological research for nursing",
  ["biol res pregnancy perinatol"] = "Biological research in pregnancy and perinatology",
  ["biol rev"] = "Biological Reviews",
  ["biol rev camb philos soc"] = "Biological reviews of the Cambridge Philosophical Society",
  ["biol rev city coll"] = "The Biological review of the City College",
  ["biol rhythm res"] = "Biological rhythm research",
  ["biol rundsch"] = "Biologische Rundschau",
  ["biol sci space"] = "Uchū Seibutsu Kagaku",
  ["biol sci space 1986"] = "Biological sciences in space 1986 : proceedings of the International Symposium on Biological Sciences in Space, Nagoya (Japan), November 10-12, 1986",
  ["biol sex differ"] = "Biology of sex differences",
  ["biol signals"] = "Biological signals",
  ["biol signals recept"] = "Biological signals and receptors",
  ["biol soc"] = "Biology and society : the journal of the Eugenics Society",
  ["biol sport"] = "Biology of sport",
  ["biol struct morphog"] = "Biological structures and morphogenesis",
  ["biol stud"] = "Bìologìčnì studìï",
  ["biol syst open access"] = "Biological Systems: Open Access",
  ["biol theory"] = "Biological theory",
  ["biol ther"] = "Biologics in therapy",
  ["biol trace elem res"] = "Biological Trace Element Research",
  ["biol unserer zeit"] = "Biologie in unserer Zeit",
  ["biol visnyk melitopolskoho derzhavnoho pedahohichnoho universytetu im bohdana khmelnytskoho"] = "Biolohichnyĭ visnyk Melitopolʹsʹkoho derz︠h︡avnoho pedahohichnoho universytetu imeni Bohdana Khmelʹnyt︠s︡koho",
  ["biol wastes"] = "Biological Wastes",
  ["biol z armen"] = "Biologicheskiĭ zhurnal Armenii = Hayastani kensabanakan handes",
  ["biol zent bl"] = "Biologisches Zentralblatt",
  ["biol-inspired syst"] = "Biologically-Inspired Systems",
  ["biol: targets ther"] = "Biologics: Targets and Therapy",
  ["biolinguistics (nicos)"] = "Biolinguistics",
  ["biologia (bratisl)"] = "Biologia",
  ["biologica (santiago)"] = "Biologica (Santiago, Chile)",
  ["biologija (vilnius)"] = "Biologija (Vilnius, Lithuania)",
  ["biologist (london)"] = "Biologist (London, England)",
  ["biology (basel)"] = "Biology",
  ["biom biostat int j"] = "Biometrics & biostatistics international journal",
  ["biom hum"] = "Biométrie humaine",
  ["biom inform biol med"] = "Biometrie und Informatik in der Biologie und Medizin",
  ["biom j"] = "Biometrical journal. Biometrische Zeitschrift",
  ["biom technol today"] = "Biometric Technology Today",
  ["biom z"] = "Biometrische Zeitschrift",
  ["biomacromol mass spectrom"] = "Biomacromolecular Mass Spectrometry",
  ["biomagn res technol"] = "Biomagnetic research and technology",
  ["biomanuf rev"] = "Biomanufacturing Reviews",
  ["biomark appl"] = "Biomarkers and applications",
  ["biomark cancer"] = "Biomarkers in cancer",
  ["biomark genes"] = "Biomarkers and genes",
  ["biomark insights"] = "Biomarker insights",
  ["biomark j"] = "Biomarkers journal",
  ["biomark med"] = "Biomarkers in medicine",
  ["biomark neuropsychiatry"] = "Biomarkers in neuropsychiatry",
  ["biomark res"] = "Biomarker research",
  ["biomarker insights"] = "Biomarker Insights",
  ["biomarker res"] = "Biomarker Research",
  ["biomarkers genomic med"] = "Biomarkers and Genomic Medicine",
  ["biomarkers med"] = "Biomarkers in Medicine",
  ["biomass bioenergy"] = "Biomass and Bioenergy",
  ["biomass convers biorefin"] = "Biomass Conversion and Biorefinery",
  ["biomater adv"] = "Biomaterials Advances",
  ["biomater artif cells artif organs"] = "Biomaterials, artificial cells, and artificial organs",
  ["biomater artif cells immobilization biotechnol"] = "Biomaterials, artificial cells, and immobilization biotechnology : official journal of the International Society for Artificial Cells and Immobilization Biotechnology",
  ["biomater biosyst"] = "Biomaterials and Biosystems",
  ["biomater invest dent"] = "Biomaterial Investigations in Dentistry",
  ["biomater investig dent"] = "Biomaterial investigations in dentistry",
  ["biomater med devices artif organs"] = "Biomaterials, medical devices, and artificial organs",
  ["biomater res"] = "Biomaterials research",
  ["biomater sci"] = "Biomaterials science",
  ["biomath (sofia)"] = "Biomath (Sofia, Bulgaria)",
  ["biomech model mechanobiol"] = "Biomechanics and modeling in mechanobiology",
  ["biomed biochim acta"] = "Biomedica Biochimica Acta",
  ["biomed biotechnol res j"] = "Biomedical And Biotechnology Research Journal",
  ["biomed bull"] = "Biomedical bulletin",
  ["biomed chromatogr"] = "Biomedical Chromatography",
  ["biomed commun"] = "Biomedical communications",
  ["biomed data manag graph online querying (2015)"] = "Biomedical data management and graph online querying : VLDB 2015 Workshops, Big-O(Q) and DMAH, Waikoloa, HI, USA, August 31-September 4, 2015, revised selected papers. International Conference on Very Large Data Bases (41st : 2015 : Waikoloa, Hawaii)",
  ["biomed digit libr"] = "Biomedical digital libraries",
  ["biomed eng"] = "BioMedical Engineering",
  ["biomed eng (ny)"] = "Biomedical engineering",
  ["biomed eng (singapore)"] = "Biomedical engineering : applications, basis, and communications",
  ["biomed eng adv"] = "Biomedical Engineering Advances",
  ["biomed eng comput biol"] = "Biomedical engineering and computational biology",
  ["biomed eng educ"] = "Biomedical engineering education",
  ["biomed eng lett"] = "Biomedical engineering letters",
  ["biomed eng online"] = "Biomedical engineering online",
  ["biomed eng res"] = "Biomedical engineering research",
  ["biomed eng: appl, basis commun"] = "Biomedical Engineering: Applications, Basis and Communications",
  ["biomed environ mass spectrom"] = "Biomedical & environmental mass spectrometry",
  ["biomed environ sci"] = "Biomedical and environmental sciences : BES",
  ["biomed ethics"] = "Biomedical ethics (Tübingen, Germany)",
  ["biomed genet genom"] = "Biomedical genetics and genomics",
  ["biomed hub"] = "Biomedicine hub",
  ["biomed hum kinet"] = "Biomedical human kinetics",
  ["biomed image registration"] = "Biomedical image registration : second international workshop, WBIR 2003, Philadelphia, PA, USA, June 23-24, 2003 : revised papers. International Workshop on Biomedical Image Registration (2nd : 2003 : Philadelphia, Pa.)",
  ["biomed imaging interv j"] = "Biomedical imaging and intervention journal",
  ["biomed imaging intervention j"] = "Biomedical Imaging and Intervention Journal",
  ["biomed inf insights"] = "Biomedical Informatics Insights",
  ["biomed inform insights"] = "Biomedical informatics insights",
  ["biomed instrum"] = "Bio-Medical instrumentation",
  ["biomed instrum technol"] = "Biomedical instrumentation & technology",
  ["biomed j"] = "Biomedical journal",
  ["biomed j sci tech res"] = "Biomedical journal of scientific & technical research",
  ["biomed khim"] = "Biomedit︠s︡inskai︠a︡ khimii︠a︡",
  ["biomed lib bull"] = "Biomedical Library bulletin (La Jolla, San Diego, Calif.)",
  ["biomed mass spectrom"] = "Biomedical mass spectrometry",
  ["biomed mater"] = "Biomedical materials (Bristol, England)",
  ["biomed mater (bristol, u k)"] = "Biomedical Materials (Bristol, United Kingdom)",
  ["biomed mater devices"] = "Biomedical Materials & Devices",
  ["biomed mater eng"] = "Bio-medical materials and engineering",
  ["biomed microdevices"] = "Biomedical microdevices",
  ["biomed news"] = "Biomedical news; the newspaper for the life scientist",
  ["biomed opt"] = "Biomedical optics (Washington, D.C.)",
  ["biomed opt express"] = "Biomedical optics express",
  ["biomed pap med fac univ palacky olomouc czech repub"] = "Biomedical papers of the Medical Faculty of the University Palacký, Olomouc, Czechoslovakia",
  ["biomed pept proteins nucleic acids"] = "Biomedical peptides, proteins & nucleic acids : structure, synthesis & biological activity",
  ["biomed pharmacother"] = "Biomedicine & pharmacotherapy = Biomédecine & pharmacothérapie",
  ["biomed phys eng express"] = "Biomedical physics & engineering express",
  ["biomed purv"] = "Bio-medical purview",
  ["biomed rep"] = "Biomedical reports",
  ["biomed res"] = "Biomedical research (Tokyo, Japan)",
  ["biomed res (aligarh)"] = "Biomedical research (Aligarh, India)",
  ["biomed res clin pract"] = "Biomedical research and clinical practice",
  ["biomed res int"] = "BioMed research international",
  ["biomed res rev"] = "Biomedical research and reviews",
  ["biomed res trace ele"] = "Biomedical Research on Trace Elements",
  ["biomed res trace elem"] = "Biomedical Research on Trace Elements",
  ["biomed saf stand"] = "Biomedical Safety & Standards",
  ["biomed sci"] = "Biomedical science",
  ["biomed sci instrum"] = "Biomedical Sciences Instrumentation",
  ["biomed sci technol"] = "Biomedical science and technology",
  ["biomed signal process control"] = "Biomedical Signal Processing and Control",
  ["biomed spectrosc imaging"] = "Biomedical spectroscopy and imaging",
  ["biomed tech"] = "Biomedizinische Technik",
  ["biomed tech (berl)"] = "Biomedizinische Technik. Biomedical engineering",
  ["biomed technol"] = "Biomedical Technology",
  ["biomedicine (taipei)"] = "BioMedicine",
  ["biometrical j"] = "Biometrical Journal",
  ["biomimetic intell rob"] = "Biomimetic Intelligence and Robotics",
  ["biomimetics (basel)"] = "Biomimetics (Basel, Switzerland)",
  ["biomol concepts"] = "Biomolecular concepts",
  ["biomol detect quantif"] = "Biomolecular Detection and Quantification",
  ["biomol eng"] = "Biomolecular Engineering",
  ["biomol nmr assign"] = "Biomolecular NMR assignments",
  ["biomol nmr assignments"] = "Biomolecular NMR Assignments",
  ["biomol ther (seoul)"] = "Biomolecules & therapeutics",
  ["bioopt world"] = "BioOptics world",
  ["bioorg chem"] = "Bioorganic chemistry",
  ["bioorg chem app"] = "Bioorganic Chemistry and Applications",
  ["bioorg chem appl"] = "Bioorganic Chemistry and Applications",
  ["bioorg khim"] = "Bioorganicheskaia khimiia",
  ["bioorg med chem"] = "Bioorganic & medicinal chemistry",
  ["bioorg med chem lett"] = "Bioorganic & medicinal chemistry letters",
  ["biopestic int"] = "Biopesticides international",
  ["biopharm drug dispos"] = "Biopharmaceutics & drug disposition",
  ["biopharm int"] = "Biopharm international",
  ["biopharm rep"] = "Biopharmaceutical report",
  ["biophotonics discovery"] = "Biophotonics Discovery",
  ["biophys chem"] = "Biophysical Chemistry",
  ["biophys econ sustainability"] = "Biophysical Economics and Sustainability",
  ["biophys j"] = "Biophysical journal",
  ["biophys physicobiol"] = "Biophysics and physicobiology",
  ["biophys rep"] = "Biophysics reports",
  ["biophys rev"] = "Biophysical reviews",
  ["biophys rev lett"] = "Biophysical Reviews and Letters",
  ["biophys struct mech"] = "Biophysics of structure and mechanism",
  ["biophysicist (rockv)"] = "Biophysicist (Rockville, Md.)",
  ["biophysics (engl transl)"] = "Biophysics (English Translation)",
  ["biophysics (nagoya-shi)"] = "Biophysics (Nagoya-shi, Japan)",
  ["biophysics (oxf)"] = "Biophysics",
  ["biopolim kletka"] = "Biopolimery i kletka",
  ["biopolym cell"] = "Biopolymers & cell",
  ["biopolym symp"] = "Biopolymers symposia",
  ["biopreserv biobank"] = "Biopreservation and biobanking",
  ["biopreserv biobanking"] = "Biopreservation and Biobanking",
  ["bioprocess biosyst eng"] = "Bioprocess and biosystems engineering",
  ["bioprocess eng"] = "Bioprocess Engineering",
  ["bioprocess int"] = "Bioprocess International",
  ["bioprocess j"] = "Bioprocessing Journal",
  ["bioprocess technol"] = "Bioprocess technology",
  ["biopsychosoc med"] = "BioPsychoSocial medicine",
  ["bioquim patol clin"] = "Bioquímica y patología clínica : ByPC : revista de la Asociación Bioquímica Argentina",
  ["biorem j"] = "Bioremediation Journal",
  ["bioremediat j"] = "Bioremediation journal",
  ["biores open access"] = "BioResearch open access",
  ["bioresour bioprocess"] = "Bioresources and bioprocessing",
  ["bioresour technol"] = "Bioresource technology",
  ["bioresour technol rep"] = "Bioresource technology reports",
  ["biorheology suppl"] = "Biorheology. Supplement : the official journal of the International Society of Biorheology",
  ["bios (leverk)"] = "Bios (Leverkusen, Germany)",
  ["biosaf health"] = "Biosafety and health",
  ["biosafety (los angel)"] = "Biosafety",
  ["bioscan (ranchi)"] = "The bioscan (Ranchi)",
  ["biosci bioeng"] = "Bioscience and bioengineering (Boston, Mass.)",
  ["biosci biotechnol biochem"] = "Bioscience, biotechnology, and biochemistry",
  ["biosci biotechnol res asia"] = "Biosciences, Biotechnology Research Asia",
  ["biosci biotechnol res commun"] = "Bioscience biotechnology research communications",
  ["biosci biotechnol, biochem"] = "Bioscience Biotechnology and Biochemistry",
  ["biosci horiz"] = "Bioscience Horizons",
  ["biosci hypotheses"] = "Bioscience hypotheses",
  ["biosci j"] = "Bioscience journal : BJ",
  ["biosci microbiota food health"] = "Bioscience of microbiota, food and health",
  ["biosci microflora"] = "Bioscience and microflora",
  ["biosci proc"] = "Bioscientifica proceedings",
  ["biosci rep"] = "Bioscience reports",
  ["biosci trends"] = "Bioscience trends",
  ["biosci, biotechnol, biochem"] = "Bioscience, Biotechnology, and Biochemistry",
  ["biosecur bioterror"] = "Biosecurity and bioterrorism : biodefense strategy, practice, and science",
  ["biosens bioelectron"] = "Biosensors & bioelectronics",
  ["biosens bioelectron x"] = "Biosensors & bioelectronics: X",
  ["biosens bioelectron: x"] = "Biosensors and Bioelectronics: X",
  ["biosensors (basel)"] = "Biosensors",
  ["biospektrum (heidelb)"] = "Biospektrum : Zeitschrift der Gesellschaft für Biologishe Chemie (GBCH) und der Vereinigung für Allgemeine und Angewandte Mikrobiologie (VAAM)",
  ["biostat bioinforma biomath"] = "Biostatistics, bioinformatics and biomathematics",
  ["biostat biom open access j"] = "Biostatistics and biometrics open access journal",
  ["biostat epidemiol"] = "Biostatistics & epidemiology",
  ["biosurf biotribol"] = "Biosurface and biotribology",
  ["biosyst biorobot"] = "Biosystems & Biorobotics",
  ["biosyst divers"] = "Biosystems diversity",
  ["biosyst eng"] = "Biosystems engineering",
  ["biosystems eng"] = "Biosystems Engineering",
  ["biota colomb"] = "Biota colombiana",
  ["biota neotrop"] = "Biota neotropica",
  ["biotech agr forest"] = "Plant Biotechnology for Sustainable Production of Energy and Co-products",
  ["biotech histochem"] = "Biotechnic & histochemistry : official publication of the Biological Stain Commission",
  ["biotech rapid dispatches"] = "BioTechniques. Rapid dispatches",
  ["biotechnol (rajkot)"] = "Biotechnology : an Indian journal",
  ["biotechnol acta"] = "Biotechnologia acta",
  ["biotechnol adv"] = "Biotechnology advances",
  ["biotechnol agron soc environ"] = "Biotechnologie, Agronomie, Societe et Environnement",
  ["biotechnol alia"] = "Biotechnology et alia",
  ["biotechnol annu rev"] = "Biotechnology annual review",
  ["biotechnol appl biochem"] = "Biotechnology and Applied Biochemistry",
  ["biotechnol bioeng"] = "Biotechnology and bioengineering",
  ["biotechnol bioeng symp"] = "Biotechnology and bioengineering symposium",
  ["biotechnol biofuels"] = "Biotechnology for biofuels",
  ["biotechnol biofuels bioprod"] = "Biotechnology for Biofuels and Bioproducts",
  ["biotechnol bioprocess eng"] = "Biotechnology and Bioprocess Engineering",
  ["biotechnol biotechnol equip"] = "Biotechnology and Biotechnological Equipment",
  ["biotechnol dev monit"] = "Biotechnology and Development Monitor",
  ["biotechnol focus"] = "Biotechnology focus",
  ["biotechnol genet eng rev"] = "Biotechnology & genetic engineering reviews",
  ["biotechnol healthc"] = "Biotechnology healthcare",
  ["biotechnol j"] = "Biotechnology journal",
  ["biotechnol law rep"] = "Biotechnology law report",
  ["biotechnol lett"] = "Biotechnology letters",
  ["biotechnol notes"] = "Biotechnology Notes",
  ["biotechnol prog"] = "Biotechnology progress",
  ["biotechnol progr"] = "Biotechnology Progress",
  ["biotechnol rep (amst)"] = "Biotechnology reports (Amsterdam, Netherlands)",
  ["biotechnol rep,"] = "Biotechnology Reports",
  ["biotechnol res"] = "Biotechnological research",
  ["biotechnol res innovation"] = "Biotechnology Research and Innovation",
  ["biotechnol res int"] = "Biotechnology research international",
  ["biotechnol tech"] = "Biotechnology Techniques",
  ["biotechnol ther"] = "Biotechnology therapeutics",
  ["biotechnol, agron, soc environ"] = "Biotechnologie, Agronomie, Societe et Environnement",
  ["biotechnology (faisalabad)"] = "Biotechnology (Faisalābād, Pakistan)",
  ["biotechnology (n y)"] = "Bio/technology (Nature Publishing Company)",
  ["biotechnology (rajkot, india)"] = "BioTechnology (Rajkot, India)",
  ["biotechnology|shengwu jishu"] = "Shengwu jishu",
  ["biotecnol apl"] = "Biotecnología aplicada : revista de la Sociedad Iberolatinoamericana para Investigaciones sobre Interferón y Biotecnología en Salud",
  ["biotecnol veg"] = "Biotecnología vegetal",
  ["biotecnolog cienc desenvolv"] = "Biotecnologia ciência & desenvolvimento",
  ["biotekhnologiia (moskva)"] = "Biotekhnologiia",
  ["biotelem patient monit"] = "Biotelemetry and patient monitoring",
  ["biotribology (oxf)"] = "Biotribology (Oxford)",
  ["biotropia (bogor)"] = "Biotropia",
  ["bipolar disord"] = "Bipolar disorders",
  ["bird conserv int"] = "Bird conservation international",
  ["bird study"] = "Bird study : the journal of the British Trust for Ornithology",
  ["birkhauser adv texts basler lehrbucher"] = "Birkhauser Advanced Texts: Basler Lehrbucher",
  ["birkhauser skr"] = "Birkhauser Skripten",
  ["birla acad monogr cult sci technol herit india"] = "K. K. Birla Academy Monographs on the Cultural, Scientific and Technological Heritage of India",
  ["birperht publ"] = "BIRPERHT publication",
  ["birt defects res a clin mol teratol"] = "Birth Defects Research. Part A, Clinical and Molecular Teratology",
  ["birth defects orig artic ser"] = "Birth defects original article series",
  ["birth defects res"] = "Birth defects research",
  ["birth defects res a clin mol teratol"] = "Birth defects research. Part A, Clinical and molecular teratology",
  ["birth defects res b dev reprod toxicol"] = "Birth defects research. Part B, Developmental and reproductive toxicology",
  ["birth defects res c embryo today"] = "Birth defects research. Part C, Embryo today : reviews",
  ["birth defects res, part a"] = "Birth Defects Research, Part A: Clinical and Molecular Teratology",
  ["birth defects res, part b"] = "Birth Defects Research, Part B: Developmental and Reproductive Toxicology",
  ["birth defects res, part c"] = "Birth Defects Research, Part C: Embryo Today–Reviews",
  ["birth fam j"] = "Birth and the family journal",
  ["birth gaz"] = "The Birth gazette",
  ["birth psychol bull"] = "Birth Psychology Bulletin",
  ["bisturi (s juan)"] = "El Bisturí",
  ["bit numer math"] = "BIT. Numerical mathematics",
  ["biul bibl jagiellon"] = "Biuletyn-Biblioteki Jagiellońskiej",
  ["biul hist sztuki"] = "Biuletyn historii sztuki",
  ["biul igs"] = "Biuletyn IGS",
  ["biul inst med morsk gdansk"] = "Biuletyn Instytutu Medycyny Morskiej w Gdansku",
  ["biul inst rosl lecz inst przem zielar pozn pol"] = "Biuletyn Instytutu Roślin Leczniczych. Instytut Przemysłu Zielarskiego (Poznań, Poland)",
  ["biul panstw inst med morsk trop j w gdansku"] = "Biuletyn Państwowego Instytutu Medycyny Morskiej i Tropikalnej w Gdańsku. Biulleten' Gosudarstvennogo Instituta morskoĭ i tropicheskoĭ meditsiny v Gdan'ske. Bulletin of the State Institute of Marine and Tropical Medicine in Gdańsk, Poland",
  ["biul zydowskiego inst hist pol"] = "Biuletyn Żydowskiego Instytutu Historycznego",
  ["biull eksp biol med"] = "Biulleten' eksperimental'noĭ biologii i meditsiny",
  ["biull mosk ova ispyt prir (biol)"] = "Bi︠u︡lletenʹ Moskovskogo obshchestva ispytateleĭ prirody. Otdel biologicheskiĭ",
  ["biull vsesoiuznogo kardiol nauchn tsentra amn sssr"] = "Biulleten' Vsesoiuznogo kardiologicheskogo nauchnogo tsentra AMN SSSR",
  ["biulleten russ sfsr minist zdravookhraneniia uchenyi meditsinskii sov"] = "Biulleten'. Russian S.F.S.R. Ministerstvo Zdravookhraneniia. Uchenyĭ meditsinskiĭ sovet",
  ["biulleten sib meditsiny"] = "Bi︠u︡lletenʹ Sibirskoĭ medit︠s︡iny",
  ["biyutiknuluzhii kishavarzi"] = "Biyutiknuluzhī-i kishāvarzī",
  ["bja educ"] = "BJA education",
  ["bjgp open"] = "BJGP open",
  ["bjhs themes"] = "BJHS themes",
  ["bjp rev books"] = "BJP review of books",
  ["bjpsych adv"] = "BJPsych advances",
  ["bjpsych bull"] = "BJPsych bulletin",
  ["bjpsych int"] = "BJPsych international",
  ["bjpsych open"] = "BJPsych open",
  ["bjr case rep"] = "BJR case reports",
  ["bjr open"] = "BJR open",
  ["bjr suppl"] = "BJR Supplement",
  ["bjs open"] = "BJS open",
  ["bju int"] = "BJU international",
  ["bjui compass"] = "BJUI compass",
  ["bl dtsch int polit"] = "Blätter für deutsche und internationale Politik",
  ["bl dtsch landesgesch"] = "Blätter für deutsche Landesgeschichte",
  ["bl gesz"] = "Bleṭer far geshikhṭe",
  ["bl ver pro sihltal"] = "Blätter der Vereinigung Pro Sihltal",
  ["bl wueritemb kirchengesch"] = "Blätter für württembergische Kirchengeschichte",
  ["bl zahnheilkd"] = "Blätter für Zahnheilkunde. Bulletin dentaire",
  ["black aging"] = "Black aging",
  ["black law j"] = "The Black law journal",
  ["black scholar"] = "The Black scholar",
  ["bladder (san franc)"] = "Bladder (San Francisco, Calif.)",
  ["bladder cancer"] = "Bladder cancer (Amsterdam, Netherlands)",
  ["blake stud"] = "Blake studies",
  ["blasting fragm j"] = "Blasting and Fragmentation Journal",
  ["blockchain healthc today"] = "Blockchain in healthcare today",
  ["blockchain: res appl"] = "Blockchain: Research and Applications",
  ["blood adv"] = "Blood advances",
  ["blood cancer discov"] = "Blood cancer discovery",
  ["blood cancer j"] = "Blood cancer journal",
  ["blood cells"] = "Blood cells",
  ["blood cells mol dis"] = "Blood cells, molecules & diseases",
  ["blood coagul fibrinolysis"] = "Blood coagulation & fibrinolysis : an international journal in haemostasis and thrombosis",
  ["blood donor"] = "Blood donor. Bloedskenker",
  ["blood immunol cell ther"] = "Blood Immunology & Cellular Therapy",
  ["blood lymphat cancer"] = "Blood and lymphatic cancer : targets and therapy",
  ["blood press"] = "Blood pressure",
  ["blood press monit"] = "Blood pressure monitoring",
  ["blood press suppl"] = "Blood pressure. Supplement",
  ["blood pressure"] = "Blood Pressure",
  ["blood pressure monit"] = "Blood Pressure Monitoring",
  ["blood purif"] = "Blood purification",
  ["blood red cells iron"] = "Blood Red Cells & Iron",
  ["blood res"] = "Blood research",
  ["blood rev"] = "Blood reviews",
  ["blood sci"] = "Blood Science",
  ["blood transfus"] = "Blood transfusion = Trasfusione del sangue",
  ["blood vessels"] = "Blood vessels",
  ["bloomsbury crit introd contemp epistemol"] = "Bloomsbury Critical Introductions to Contemporary Epistemology",
  ["bloomsbury geogr"] = "Bloomsbury geographer",
  ["bloomsbury stud aristot tradit"] = "Bloomsbury Studies in the Aristotelian Tradition",
  ["blu pagine sci"] = "I Blu—Pagine di Scienza",
  ["blue cross assoc res ser"] = "Blue Cross Association research series",
  ["blue jay"] = "Blue jay",
  ["blue sheet"] = "The Blue sheet",
  ["blue-green syst"] = "Blue-Green Systems",
  ["bmb rep"] = "BMB reports",
  ["bmc anesthesiol"] = "BMC anesthesiology",
  ["bmc biochem"] = "BMC biochemistry",
  ["bmc bioinf"] = "BMC Bioinformatics",
  ["bmc bioinformatics"] = "BMC bioinformatics",
  ["bmc biol"] = "BMC biology",
  ["bmc biomed eng"] = "BMC biomedical engineering",
  ["bmc biophys"] = "BMC biophysics",
  ["bmc biotech"] = "BMC Biotechnology",
  ["bmc biotechnol"] = "BMC biotechnology",
  ["bmc blood disord"] = "BMC blood disorders",
  ["bmc cancer"] = "BMC cancer",
  ["bmc cardiovasc disord"] = "BMC cardiovascular disorders",
  ["bmc cell biol"] = "BMC cell biology",
  ["bmc chem"] = "BMC chemistry",
  ["bmc chem biol"] = "BMC chemical biology",
  ["bmc chem eng"] = "BMC Chemical Engineering",
  ["bmc clin pathol"] = "BMC clinical pathology",
  ["bmc clin pharmacol"] = "BMC clinical pharmacology",
  ["bmc complement altern med"] = "BMC complementary and alternative medicine",
  ["bmc complement med ther"] = "BMC complementary medicine and therapies",
  ["bmc complementary med ther"] = "BMC Complementary Medicine and Therapies",
  ["bmc dermatol"] = "BMC dermatology",
  ["bmc dev biol"] = "BMC developmental biology",
  ["bmc digital health"] = "BMC Digital Health",
  ["bmc ear nose throat disord"] = "BMC ear, nose, and throat disorders",
  ["bmc ecol"] = "BMC ecology",
  ["bmc ecol evol"] = "BMC ecology and evolution",
  ["bmc emerg med"] = "BMC emergency medicine",
  ["bmc endocr disord"] = "BMC endocrine disorders",
  ["bmc energy"] = "BMC Energy",
  ["bmc evol biol"] = "BMC evolutionary biology",
  ["bmc fam pract"] = "BMC family practice",
  ["bmc gastroenterol"] = "BMC gastroenterology",
  ["bmc genet"] = "BMC genetics",
  ["bmc genom data"] = "BMC genomic data",
  ["bmc genomic data"] = "BMC Genomic Data",
  ["bmc genomics"] = "BMC genomics",
  ["bmc geriatr"] = "BMC geriatrics",
  ["bmc health serv res"] = "BMC health services research",
  ["bmc hematol"] = "BMC hematology",
  ["bmc immunol"] = "BMC immunology",
  ["bmc infect dis"] = "BMC infectious diseases",
  ["bmc int health hum rights"] = "BMC international health and human rights",
  ["bmc mater"] = "BMC materials",
  ["bmc mech eng"] = "BMC Mechanical Engineering",
  ["bmc med"] = "BMC medicine",
  ["bmc med educ"] = "BMC medical education",
  ["bmc med ethics"] = "BMC medical ethics",
  ["bmc med genet"] = "BMC medical genetics",
  ["bmc med genomics"] = "BMC medical genomics",
  ["bmc med imaging"] = "BMC medical imaging",
  ["bmc med inf decis making"] = "BMC Medical Informatics and Decision Making",
  ["bmc med inform decis mak"] = "BMC medical informatics and decision making",
  ["bmc med inform decis making"] = "BMC Medical Informatics and Decision Making",
  ["bmc med phys"] = "BMC medical physics",
  ["bmc med res methodol"] = "BMC medical research methodology",
  ["bmc methods"] = "BMC Methods",
  ["bmc microbiol"] = "BMC microbiology",
  ["bmc mol biol"] = "BMC molecular biology",
  ["bmc mol cell biol"] = "BMC molecular and cell biology",
  ["bmc musculoskelet disord"] = "BMC musculoskeletal disorders",
  ["bmc nephrol"] = "BMC nephrology",
  ["bmc neurol"] = "BMC neurology",
  ["bmc neurosci"] = "BMC neuroscience",
  ["bmc nucl med"] = "BMC nuclear medicine",
  ["bmc nurs"] = "BMC nursing",
  ["bmc nutr"] = "BMC nutrition",
  ["bmc obes"] = "BMC obesity",
  ["bmc ophthalmol"] = "BMC ophthalmology",
  ["bmc oral health"] = "BMC oral health",
  ["bmc palliat care"] = "BMC palliative care",
  ["bmc pediatr"] = "BMC pediatrics",
  ["bmc pharmacol"] = "BMC pharmacology",
  ["bmc pharmacol toxicol"] = "BMC pharmacology & toxicology",
  ["bmc physiol"] = "BMC physiology",
  ["bmc plant biol"] = "BMC plant biology",
  ["bmc pregnancy childbirth"] = "BMC pregnancy and childbirth",
  ["bmc proc"] = "BMC proceedings",
  ["bmc psychiatry"] = "BMC psychiatry",
  ["bmc psychol"] = "BMC psychology",
  ["bmc public health"] = "BMC public health",
  ["bmc pulm med"] = "BMC pulmonary medicine",
  ["bmc res notes"] = "BMC research notes",
  ["bmc rheumatol"] = "BMC rheumatology",
  ["bmc sports sci med rehabil"] = "BMC sports science, medicine & rehabilitation",
  ["bmc struct biol"] = "BMC structural biology",
  ["bmc surg"] = "BMC surgery",
  ["bmc syst biol"] = "BMC systems biology",
  ["bmc urol"] = "BMC urology",
  ["bmc vet res"] = "BMC veterinary research",
  ["bmc womens health"] = "BMC women's health",
  ["bmc zool"] = "BMC zoology",
  ["bme front"] = "BME frontiers",
  ["bmj (int ed)"] = "BMJ (International edition)",
  ["bmj case rep"] = "BMJ case reports",
  ["bmj clin evid"] = "BMJ clinical evidence",
  ["bmj connect clin genet genomics"] = "BMJ Connections Clinical Genetics and Genomics",
  ["bmj connect oncol"] = "BMJ Connections Oncology",
  ["bmj digital health ai"] = "BMJ Digital Health & AI",
  ["bmj evid based med"] = "BMJ evidence-based medicine",
  ["bmj glob health"] = "BMJ global health",
  ["bmj health care inform"] = "BMJ health & care informatics",
  ["bmj immunol"] = "BMJ Immunology",
  ["bmj innov"] = "BMJ innovations",
  ["bmj med"] = "BMJ Medicine",
  ["bmj mil health"] = "BMJ military health",
  ["bmj neurol open"] = "BMJ neurology open",
  ["bmj nutr prev health"] = "BMJ nutrition, prevention & health",
  ["bmj open"] = "BMJ open",
  ["bmj open diabetes res care"] = "BMJ open diabetes research & care",
  ["bmj open gastroenterol"] = "BMJ open gastroenterology",
  ["bmj open ophthalmol"] = "BMJ open ophthalmology",
  ["bmj open qual"] = "BMJ open quality",
  ["bmj open respir res"] = "BMJ open respiratory research",
  ["bmj open sci"] = "BMJ Open Science",
  ["bmj open sport exerc med"] = "BMJ open sport & exercise medicine",
  ["bmj open sport exercise med"] = "BMJ Open Sport & Exercise Medicine",
  ["bmj outcomes"] = "BMJ outcomes",
  ["bmj paediatr open"] = "BMJ paediatrics open",
  ["bmj qual improv rep"] = "BMJ quality improvement reports",
  ["bmj qual saf"] = "BMJ quality & safety",
  ["bmj sex reprod health"] = "BMJ sexual & reproductive health",
  ["bmj simul technol enhanc learn"] = "BMJ simulation & technology enhanced learning",
  ["bmj support palliat care"] = "BMJ supportive & palliative care",
  ["bmj surg interv health technol"] = "BMJ surgery, interventions, & health technologies",
  ["bmr j aust geol geophys"] = "BMR Journal of Geology and Geophysics",
  ["bni q"] = "BNI quarterly",
  ["bnwl rep"] = "BNWL [reports]. U.S. Atomic Energy Commission",
  ["bo pu xue za zhi"] = "Bo pu xue za zhi = Chinese journal of microwave & radio-frequency spectroscopy",
  ["boardr rep"] = "Boardroom reports",
  ["bocagiana (funchal)"] = "Bocagiana",
  ["bocconi springer ser"] = "Bocconi & Springer Series",
  ["boch stud philos"] = "Bochumer Studien zur Philosophie",
  ["bodenkd pflanzenernahr"] = "Bodenkunde und Pflanzenernahrung",
  ["bodleian libr rec"] = "The Bodleian Library record",
  ["body forum"] = "The Body forum",
  ["body image"] = "Body image",
  ["body mov dance psychother"] = "Body, movement, and dance in psychotherapy",
  ["body posit"] = "The Body positive",
  ["body soc"] = "Body & society",
  ["boei eisei"] = "[Bōei eisei] Japanese Defense Forces medical journal",
  ["boei ika daigakko zasshi"] = "Bōei Ika Daigakkō zasshi = Journal of the National Defense Medical College",
  ["boethius texte abh gesch math naturwiss"] = "Boethius: Texte und Abhandlungen zur Geschichte der Mathematik und der Naturwissenschaften",
  ["bogens verden"] = "Bogens verden",
  ["bogeon sahoe nonjib"] = "Pogŏn sahoe nonjip = Journal of population, health, and social welfare",
  ["bohemia jahrb coll carolinum"] = "Bohemia (Munich, Germany)",
  ["bois for trop"] = "Bois et Forêts des Tropiques",
  ["boj tuberk"] = "Boj proti tuberkulose",
  ["boj zdr"] = "Boj o zdravie",
  ["bol aaon"] = "Boletin de la AAON",
  ["bol acad nac cienc"] = "Boletín. Academia Nacional de Ciencias (Córdoba, Argentina)",
  ["bol acad nac hist (argent)"] = "Boletín de la Academia Nacional de la Historia. Academia Nacional de la Historia (Argentina)",
  ["bol acad nac hist (caracas)"] = "Boletín de la Academia Nacional de la Historia. Academia Nacional de la Historia (Venezuela)",
  ["bol acad nac med"] = "Boletim (Academia Nacional de Medicina (Brazil))",
  ["bol acad peru cir"] = "Boletín. Academia Peruana de Cirugía",
  ["bol agrup odontol zona cent cap fed b aires"] = "Bolet'in de la Agrupación Odontológica de la Zona Central de la Capital Federal",
  ["bol anal demogr"] = "Boletín de análisis demográfico",
  ["bol antropol"] = "Boletín de antropología",
  ["bol antropol am"] = "Boletín de antropología americana",
  ["bol asoc argent odontol ninos"] = "Boletín de la Asociación Argentina de Odontología para Niños",
  ["bol asoc chil prot fam"] = "Boletn̕ Asociacin̤ Chilena de Proteccin̤ de la Familia",
  ["bol asoc demogr hist"] = "Boletín de la Asociación de Demografía Histórica",
  ["bol asoc esp entomol"] = "Boletin de la Asociación Española de Entomologia",
  ["bol asoc herpetol esp"] = "Boletín de la Asociación Herpetológica Española",
  ["bol asoc mat venez"] = "Boletin de la Asociacion Matematica Venezolana",
  ["bol asoc med p r"] = "Boletín de la Asociación Médica de Puerto Rico",
  ["bol asoc med santiago caballer"] = "Boletín. Asociación Médica de Santiago de los Caballeros",
  ["bol asoc medica nac repub panama"] = "Boletín. Asociación Médica Nacional de la República Panamá",
  ["bol asoc venez enferm prof"] = "Boletín. Asociación Venezolana de Enfermeras Profesionales",
  ["bol b aires univ nac inst clin quir"] = "Boletin. Buenos Aires. Universidad Nacional. Instituto de Clinica Quirurgica",
  ["bol bibl menendez pelayo"] = "Boletín de la Biblioteca de Menéndez Pelayo",
  ["bol bibliogr (sao paulo)"] = "Boletim bibliográfico",
  ["bol bibliogr secr hacienda credito publico"] = "Boletín bibliográfico de la Secretaría de Hacienda y Crédito Público",
  ["bol biol"] = "Boletin biológico",
  ["bol biotecnol"] = "Boletim de biotecnologia",
  ["bol cent biol reprod"] = "Boletim do C.B.R",
  ["bol cent estud hosp serv estado"] = "Boletim do Centro de Estudos, Hospital dos Servidores do Estado",
  ["bol cent estud hosp servidores estado"] = "Boletim do Centro de Estudos, Hospital dos Servidores do Estado",
  ["bol cent estud siglo xviii"] = "Boletín del Centro de Estudios del Siglo XVIII. Universidad de Oviedo. Centro de Estudios del Siglo XVIII",
  ["bol chil parasitol"] = "Boletin Chileno de Parasitologia",
  ["bol cienc geod"] = "Boletim de Ciencias Geodesicas",
  ["bol circ argent odontol"] = "Boletín - Círculo Argentino de Odontología",
  ["bol ciudad trujillo"] = "Boletín. Ciudad Trujillo. Hospital General Padre Billini",
  ["bol clin"] = "Boletín clínico",
  ["bol clin estat"] = "Boletim clínico e de estatística. Lisbon. Hospital Colonial",
  ["bol clin hosp civis lisb"] = "Boletim clínico dos Hospitais Civis de Lisboa",
  ["bol col med camaguey cuba"] = "Boletín. Colegio Médico de Camagüey, Cuba",
  ["bol col med holguin"] = "Boletín. Colegio Médico de Holguín",
  ["bol col med la habana"] = "Boletín del Colegio Médico de la Habana",
  ["bol col med la habana (la habana)"] = "Boletín. Colegio Médico de la Habana (Havana, Cuba)",
  ["bol col prof enferm p r"] = "Boletin - Colegio de Profesionales de la Enfermería de Puerto Rico",
  ["bol colegio bras cir"] = "Boletim. Colégio Brasileiro de Cirurgiões",
  ["bol cons nac poblac"] = "Boletín (Consejo Nacional de Población (Peru))",
  ["bol cons nac tuberc cuba"] = "Boletín. Consejo Nacional de Tuberculosis (Cuba)",
  ["bol cult bibliogr"] = "Boletín cultural y bibliográfico",
  ["bol cult inf cons gen col med esp"] = "Boletin cultural e informativo - Consejo General de Colegios Medicos de España",
  ["bol demogr"] = "Boletim demográfico",
  ["bol dent oper"] = "Boletim de dentística operatória",
  ["bol dent urug"] = "Boletin dental uruguayo",
  ["bol dep med seguro social ecuat"] = "Boletín. Ecuador. Departamento Médico del Seguro Social Ecuatoriano",
  ["bol dir gen odontol (santa fe)"] = "Boletín. Santa Fe (Argentina : Province). Dirección General de Odontología",
  ["bol div nac dermatol sanit"] = "Boletim da Divisao Nacional de Dermatologia Sanitaria",
  ["bol ed col mex"] = "Boletín editorial",
  ["bol educ fis (santiago)"] = "Boletin de educación física",
  ["bol epidemiol (mex city mex)"] = "Boletín epidemiológico (Mexico City, Mexico)",
  ["bol epidemiol aids"] = "Boletim epidemiológico. AIDS",
  ["bol epidemiol chile"] = "Boletín epidemiológico de Chile",
  ["bol equipe odontol sanit"] = "Boletim Da Equipe De Odontologia Sanitaria",
  ["bol estad"] = "Boletín de estadística. Colombia. Ministerio de Higiene. División Técnica de Bio-Estadística",
  ["bol estud latinoam caribe"] = "Boletín de estudios latinoamericanos y del Caribe",
  ["bol estud med biol"] = "Boletín de estudios médicos y biológicos",
  ["bol fac farm odontol ribeirao preto"] = "Boletim da Faculdade de Farmacia e Odontologia de Ribeirao Preto",
  ["bol fac odontol piracicaba"] = "Boletim - Faculdade de Odontologia de Piracicaba",
  ["bol geogr"] = "Boletim geográfico",
  ["bol geogr teor"] = "Boletim de geografia teorética",
  ["bol geol min"] = "Boletin Geologico y Minero",
  ["bol hig ment"] = "Boletim de higiene mental",
  ["bol hist antig"] = "Boletín de historia y antigüedades",
  ["bol hosp"] = "Boletín de los hospitales",
  ["bol hosp mil cuba ejercito hosp mil dr carlos j finlay"] = "Boletín del Hospital Militar. Cuba. Ejército. Hospital Militar Dr. Carlos J. Finlay",
  ["bol hosp oftalmol"] = "Boletín. Mexico City (Mexico). Hospital Oftalmológico de Nuestra Señora de la Luz",
  ["bol hosp san juan dios"] = "Boletín del Hospital San Juan de Dios",
  ["bol hosp univ caracas"] = "Boletín del Hospital Universitario de Caracas",
  ["bol indic coyunt"] = "Boletín de indicadores de coyuntura",
  ["bol inf asoc med argent"] = "Boletin Informativo - Asociación Médica Argentina",
  ["bol inf cent doc pesqui"] = "Boletim informativo do Centro de Documentação e Pesquisa",
  ["bol inf col odontol estomatol (barc)"] = "Boletin de información - Ilustre Colegio Oficial de Odontólogos y Estomatólogos 2. a Región (Cataluña)",
  ["bol inf dent (madr)"] = "Boletiń de información dental",
  ["bol inf hisp am hist med"] = "Boletín informativo hispanoamericano de historia de la medicina",
  ["bol inf parasit chil"] = "Boletin de informaciones parasitarias chilenas",
  ["bol inf soc micol extremena"] = "Boletín informativo (Sociedad Micológica Extremeña)",
  ["bol inf socio econ"] = "Boletim de informação sócio-económica",
  ["bol inst angola"] = "Boletim do Instituto de Angola",
  ["bol inst bot"] = "Boletim do Instituto de Botânica",
  ["bol inst estud asturianos"] = "Boletín del Instituto de Estudios Asturianos. Instituto de Estudios Asturianos (Oviedo, Spain)",
  ["bol inst estud med biol univ nac auton mex"] = "Boletin del Instituto de Estudios Medicos y Biologicos, Universidad Nacional Autonoma de Mexico",
  ["bol inst int am prot infanc"] = "Boletín del Instituto Internacional Americano de Protección a la Infancia. Interamerican Children's Institute",
  ["bol inst interam nino"] = "Boletín del Instituto Interamericano del Niño",
  ["bol inst invest bibliogr (mexico)"] = "Boletin Del Instituto De Investigaciones Bibliograficas",
  ["bol inst pesca"] = "Boletim do Instituto de Pesca. Instituto de Pesca (São Paulo, Brazil : State)",
  ["bol inst psicopedag nac"] = "Boletín. Instituto Psicopedagógico Nacional (Peru)",
  ["bol inst pueric"] = "Boletim do Instituto de Puericultura",
  ["bol inst pueric martagao gesteira"] = "Boletim do Instituto de Puericultura e Pediatria Martagao Gesteira",
  ["bol inst super hig doutor ricardo jorge"] = "Boletim do Instituto Superior de Higiene Doutor Ricardo Jorge",
  ["bol latinoam caribe plant med aromat"] = "Boletin Latinoamericano y del Caribe de Plantas Medicinales y Aromaticas",
  ["bol latinoam caribe plantas med aromat"] = "Boletín Latinoamericano y del Caribe de plantas medicinales y aromáticas",
  ["bol liga argent contra reum b aires"] = "Boletín. Liga Argentina contra el Reumatismo, Buenos Aires",
  ["bol liga contra cancer havana"] = "Boletín de la Liga Contra el Cáncer. Liga Contra el Cáncer (Cuba)",
  ["bol malariol salud ambient"] = "Boletín de malariología y salud ambiental",
  ["bol mat"] = "Boletin de Matematicas. Nueva Serie",
  ["bol mat dent"] = "Boletim de materiais dentários",
  ["bol med"] = "Boletín médico",
  ["bol med cient"] = "Boletin médico científico",
  ["bol med hosp infant mex"] = "Boletín médico del Hospital Infantil de México",
  ["bol med soc"] = "Boletín médico-social",
  ["bol mens estad dane"] = "Boletín mensual de estadísticas (Colombia. Departamento Administrativo Nacional de Estadísticas)",
  ["bol mens estad demogr sanit penins islas adyac"] = "Boletín mensual de estadística demográfico-sanitaria de la península é islas adyacentes / Ministerio de la Gobernación del Reino, Dirección General de Beneficencia y Sanidad",
  ["bol mex city univ nac"] = "Boletín. Universidad Nacional Autónoma de México. Instituto de Química",
  ["bol mex derecho comp"] = "Boletín mexicano de derecho comparado",
  ["bol micolog famcal"] = "Boletín micológico de FAMCAL",
  ["bol mus soc argent"] = "Boletin del Musea Social Argentino",
  ["bol odontol (b aires)"] = "Boletín odontológico",
  ["bol odontol (bogota)"] = "Boletín de odontología",
  ["bol odontol mex"] = "Boletín odontológico mexicano",
  ["bol of estado gac madr spain"] = "Boletín oficial del estado: Gaceta de Madrid. Spain",
  ["bol of sanit panam (engl)"] = "Boletín de la Oficina Sanitaria Panamericana. English ed",
  ["bol oficina sanit panam"] = "Boletín de la Oficina Sanitaria Panamericana. Pan American Sanitary Bureau",
  ["bol patol med"] = "Boletín. Madrid. Hospital Provincial. Instituto de Patología Médica",
  ["bol pesqui (embrapa soja)"] = "Boletim de pesquisa da EMBRAPA Soja",
  ["bol poblac"] = "Boletín de población",
  ["bol popul emprego renda nordeste"] = "Boletim sobre população, emprego e renda no nordeste",
  ["bol prolap"] = "Boletín del PROLAP",
  ["bol protes"] = "Boletín de prótesis",
  ["bol psicol"] = "Boletín de psicología",
  ["bol r soc esp hist nat secc biol"] = "Boletín de la Real Sociedad Española de Historia Natural. Sección biológica : órgano del Instituto de Ciencias Naturales José de Acosta",
  ["bol sanat sao lucas"] = "Boletim do Sanatorio São Lucas. São Paulo, Brazil (City) Sanatorio São Lucas",
  ["bol sanid asist publica"] = "Boletín de sanidad y asistencia pública. Dominican Republic. Secretaría de Sanidad y Asistencia Pública",
  ["bol sanid mil"] = "Boletin de sanidad militar",
  ["bol sanid veg plagas"] = "Boletín de sanidad vegetal, plagas",
  ["bol sao paulo braz state reparticao aguas esgotos"] = "Boletim. São Paulo, Brazil (State) Repartição de Aguas e Esgotos",
  ["bol sea"] = "Boletín de la SEA",
  ["bol serv nac prot matern infant peru"] = "Boletín. Peru. Servicio Nacional de Protección Materno-Infantil",
  ["bol serv odontol sanit (porto alegre)"] = "Boletim Do Servico De Odontologia Sanitaria Da Secretaria Da Saude Do Rio Grande Do Sul",
  ["bol servic med nac empl"] = "Boletín del Servicio Médico Nacional de Empleados. Chile. Servicio Médico Nacional de Empleados",
  ["bol sidema"] = "Boletín del SIDEMA : publicación periódico del Centro de Estudios de Población, CENEP",
  ["bol soc argent bot"] = "Boletín de la Sociedad Argentina de Botánica. Sociedad Argentina de Botánica",
  ["bol soc biol concepc"] = "Boletín de la Sociedad de Biología de Concepción",
  ["bol soc bras med vet"] = "Boletim. Sociedade Brasileira de Medicina Veterinaria",
  ["bol soc brasil mat (ns)"] = "Boletim da Sociedade Brasileira de Matemática",
  ["bol soc castell cult"] = "Boletín de la Sociedad Castellonense de Cultura. Sociedad Castellonense de Cultura",
  ["bol soc catalana pediatr"] = "Boletín de la Sociedad Catalana de Pediatría",
  ["bol soc chil obstet ginecol"] = "Boletin. Sociedad Chilena de Obstetricia y Ginecología",
  ["bol soc chil quim"] = "Boletin de la Sociedad Chilena de Quimica",
  ["bol soc cir rosario"] = "Boletines de la Sociedad de Cirugía de Rosario",
  ["bol soc cir urug"] = "Boletín. Sociedad de Cirugía del Uruguay",
  ["bol soc cuba pediatr"] = "Boletín. Sociedad Cubana de Pediatría",
  ["bol soc cubana dermatol sifilogr"] = "Boletín. Sociedad Cubana de Dermatología y Sifilografía",
  ["bol soc dent guatem"] = "Boletin de la Sociedad Dental de Guatemala",
  ["bol soc esp ceram vidrio"] = "Boletin de la Sociedad Espanola de Ceramica y Vidrio",
  ["bol soc esp hidrol med"] = "Boletín (Sociedad Española de Hidrología Médica)",
  ["bol soc esp hist farm"] = "Boletín de la Sociedad Española de Historia de la Farmacia",
  ["bol soc esp hist med"] = "Boletín de la Sociedad Española de Historia de la Medicina",
  ["bol soc estomatol argent"] = "Boletin De La Sociedad Estomatologica Argentina",
  ["bol soc geogr lisb"] = "Boletim da Sociedade de Geografia de Lisboa",
  ["bol soc geol mex"] = "Boletin De La Sociedad Geologica Mexicana",
  ["bol soc mat mex"] = "Boletín de la Sociedad Matemática Mexicana",
  ["bol soc mat mex (3)"] = "Boletin de la Sociedad Matematica Mexicana. Third Series",
  ["bol soc mat mexicana (3)"] = "Sociedad Matemática Mexicana",
  ["bol soc med cir camp"] = "Boletim. Sociedade de Medicina e Cirurgia de Campinas",
  ["bol soc med mendoza"] = "Boletín. Sociedad Médica de Mendoza",
  ["bol soc mex hist filos med"] = "Boletín de la Sociedad Mexicana de Historia y Filosofía de la Medicina",
  ["bol soc micol madr"] = "Boletín de la Sociedad Micológica de Madrid",
  ["bol soc mutual med farm guadalaj"] = "Boletín de la Sociedad Mutualista Médico-Farmacéutica de Guadalajara (1905)",
  ["bol soc parana mat (2)"] = "Sociedade Paranaense de Matemática",
  ["bol soc parana mat (3)"] = "Boletim da Sociedade Paranaense de Matematica. 3rd Serie",
  ["bol soc paul med vet"] = "Boletim. Sociedade Paulista de Medicina Veterinária",
  ["bol soc port hemorreol microcirc"] = "Boletim da Sociedade Portuguesa de Hemorreologia e Microcirculação",
  ["bol soc port mat"] = "Boletim da Sociedade Portuguesa de Matematica",
  ["bol soc quim peru"] = "Boletín de la Sociedad Química del Péru",
  ["bol soc valencia pediatr"] = "Boletín de la Sociedad Valenciana de Pediatría",
  ["bol soc zool urug"] = "Boletín de la Sociedad Zoológica del Uruguay",
  ["bol socioecon"] = "Boletín socioeconómico (Cali, Colombia : 1985)",
  ["bol tec prod pet"] = "Boletim Tecnico da Producao de Petroleo",
  ["bol tr soc argent cir"] = "Boletines y trabajos - Sociedad Argentina de Cirujanos",
  ["bol trab acad argent cir"] = "Boletines y trabajos de la Academia Argentina de Cirugía. Academia Argentina de Cirugía",
  ["bol trab soc cir b aires"] = "Boletines y trabajos. Sociedad de Cirugía de Buenos Aires (1953)",
  ["bol trimest hosp vina del mar"] = "Boletń trimestral - Hospital de Vin̂a del Mar",
  ["bol univ b aires"] = "Boletín. Universidad de Buenos Aires. Instituto de Clínica Quirúrgica",
  ["bol univ granada"] = "Boletín de la Universidad de Granada. Universidad de Granada",
  ["bol vet"] = "Boletín veterinario",
  ["boletin bogota"] = "Boletín. Bogota (Colombia). Clínica de Marly",
  ["boletin clin luis razetti caracas"] = "Boletín. Clínica Luis Razetti, Caracas. Laboratorio",
  ["boll arch stor mov soc cattol ital"] = "Bollettino dell'Archivio per la storia del movimento sociale cattolico in Italia",
  ["boll assoc micol ecol romana"] = "Bollettino della associazione micologica ed ecologica romana",
  ["boll cent int beltrame stor spazio tempo"] = "Bollettino del Centro internazionale A. Beltrame di storia dello spazio e del tempo",
  ["boll cent micol friul"] = "Bollettino del centro micologico friulano",
  ["boll chim farm"] = "Bollettino chimico farmaceutico",
  ["boll chim ig"] = "Bollettino dei Chimici Igienisti",
  ["boll demogr stor"] = "Bollettino di demografia storica",
  ["boll gruppo micol g bresadola"] = "Bollettino del Gruppo micologico G. Bresadola",
  ["boll inf consoc naz (rome)"] = "Boleettino d'informazioni della Consociazione nazionale infermiere professionali e assistenti sanitarie visitatrici",
  ["boll ist sieroter milan"] = "Bollettino dell'Istituto sieroterapico milanese",
  ["boll mal orecch gola naso"] = "Bollettino delle malattie dell'orecchio, della gola, del naso",
  ["boll malacol"] = "Bollettino malacologico",
  ["boll mem soc piemont chir"] = "Bollettino e memorie della Società piemontese di chirurgia",
  ["boll mem soc tosco umbra chir"] = "Bollettino e memorie della Società tosco-umbra di chirurgia",
  ["boll mens stat (rome)"] = "Bollettino mensile di statistica (Rome, Italy)",
  ["boll metallogr"] = "Bollettino metallografico. Metallurgical bulletin",
  ["boll mus civ stor nat verona bot zool"] = "Bollettino del Museo civico di storia naturale di Verona. Botanica, zoologia",
  ["boll mus ist biol univ genova"] = "Bollettino dei musei e degli istituti biologici dell'Università di Genova",
  ["boll ocul"] = "Bollettino d'oculistica",
  ["boll osp varese"] = "Bollettino dell'Ospedale di Varese",
  ["boll schermogr"] = "Bollettino schermografico",
  ["boll sez reg soc ital dermatol sifilogr"] = "Bollettino delle sezioni regionali",
  ["boll soc ital biol sper"] = "Bollettino della Società italiana di biologia sperimentale",
  ["boll soc ital cardiol"] = "Bollettino della Società italiana di cardiologia",
  ["boll soc italiana sci suolo"] = "Bolletino della Società Italiana della Scienza del Suolo",
  ["boll soc med chir bresciana"] = "Bollettino della Società medico chirurgica bresciana",
  ["boll soc med chir cremona"] = "Bollettino - Società medico chirurgica Cremona",
  ["boll soc med chir modena"] = "Bollettino - Società medico-chirurgica di Modena",
  ["boll soc med chir pisa"] = "Bollettino - Società medico-chirurgica di Pisa",
  ["boll soc paleontol ital"] = "Bollettino della società paleontologica italiana. Società paleontologica italiana",
  ["boll soc ticin sci nat"] = "Bolletino della Società ticinese di Scienze naturali",
  ["boll stor bibliogr subalp"] = "Bollettino storico-bibliografico subalpino",
  ["boll stor sci mat"] = "Bollettino di Storia delle Scienze Matematiche",
  ["boll storia sci mat"] = "Bollettino di Storia delle Scienze Matematiche",
  ["boll un mat ital a (7)"] = "Unione Matematica Italiana",
  ["boll un mat ital b (7)"] = "Unione Matematica Italiana",
  ["boll unione ital chim ig"] = "Bollettino - Unione Italiana Chimici Igienisti",
  ["boll unione mat ital"] = "Bollettino dell’Unione Matematica Italiana",
  ["boll unione mat ital (2008)"] = "Bollettino della Unione matematica italiana (2008)",
  ["boll unione mat ital sez b artic ric mat (8)"] = "Bollettino della Unione Matematica Italiana",
  ["boll zool"] = "Bollettino di zoologia",
  ["boll-soc ital biol sper"] = "Bollettino - Societa Italiana di Biologia Sperimentale",
  ["bologna incontri"] = "Bologna incontri",
  ["bologna medica"] = "Bologna medica",
  ["bolyai soc math stud"] = "Bolyai Society Mathematical Studies",
  ["bombay hosp j"] = "The Bombay Hospital journal",
  ["bone joint j"] = "The bone & joint journal",
  ["bone joint res"] = "Bone & joint research",
  ["bone jt open"] = "Bone & joint open",
  ["bone marrow res"] = "Bone marrow research",
  ["bone marrow transplant"] = "Bone marrow transplantation",
  ["bone miner"] = "Bone and mineral",
  ["bone rep"] = "Bone reports",
  ["bone res"] = "Bone research",
  ["bone tissue regen insights"] = "Bone and tissue regeneration insights",
  ["bonekey osteovision"] = "BoneKEy osteovision",
  ["bonekey rep"] = "BoneKEy reports",
  ["bonn geschichtsbl"] = "Bonner Geschichtsblätter",
  ["bonn zool beitr"] = "Bonner zoologische Beiträge : Herausgeber: Zoologisches Forschungsinstitut und Museum Alexander Koenig, Bonn",
  ["bonn zool bull"] = "Bonn zoological bulletin",
  ["bonn zool monogr"] = "Bonner zoologische Monographien",
  ["bonner math schriften"] = "Bonner Mathematische Schriften",
  ["book abstr health technol assess int meet 4th 2007 barc spain"] = "Book of abstracts : IV Annual Meeting Health Technology Assessment International, Barcelona, Catalonia, Spain 17-20 June 2007. Health Technology Assessment International. Meeting (4th : 2007 : Barcelona, Spain)",
  ["book collect"] = "The Book collector",
  ["book hist"] = "Book history",
  ["book pap group annu"] = "The Book & Paper Group annual",
  ["book suppl j child psychol psychiatr"] = "Book supplement to the Journal of child psychology and psychiatry",
  ["bookmans price index (det)"] = "Bookman's price index : an annual guide to the values of rare and other out-of-print books and sets of periodicals",
  ["books iowa"] = "Books at Iowa",
  ["books print"] = "Books in Print",
  ["bopuxue zazhi"] = "Bopuxue Zazhi",
  ["bord chir"] = "Bordeaux chirurgical",
  ["bord med"] = "Bordeaux médical",
  ["bordens rev nutr res"] = "Borden's review of nutrition research",
  ["borderline personal disord emot dysregul"] = "Borderline personality disorder and emotion dysregulation",
  ["boreal environ res"] = "Boreal Environment Research",
  ["boreas (oslo)"] = "Boreas (Oslo)",
  ["borgyogy venerol sz"] = "Börgyógyászati és venerologiaia szemle",
  ["bosn j basic med sci"] = "Bosnian journal of basic medical sciences",
  ["bosque (valdivia)"] = "Bosque (Valdivia)",
  ["boston bar j"] = "Boston bar journal",
  ["boston coll environ aff law rev"] = "Boston College environmental affairs law review. Boston College. Law School",
  ["boston coll ind commer law rev"] = "Boston College industrial and commercial law review",
  ["boston coll law rev"] = "Boston College law review. Boston College. Law School",
  ["boston coll third world law j"] = "Boston College Third World law journal",
  ["boston college int comp law rev"] = "Boston College international and comparative law review",
  ["boston globe"] = "The Boston globe",
  ["boston med surg j"] = "The Boston medical and surgical journal",
  ["boston stud philos hist sci"] = "Boston Studies in the Philosophy and History of Science",
  ["boston stud philos sci"] = "Boston Studies in the Philosophy of Science",
  ["boston univ int law j"] = "Boston University international law journal",
  ["boston univ j"] = "Boston University journal",
  ["boston univ law rev"] = "Boston University law review. Boston University. School of Law",
  ["boston univ public interest law j"] = "The Boston University public interest law journal",
  ["bot acta"] = "Botanica acta : Berichte der Deutschen Botanischen Gesellschaft = journal of the German Botanical Society",
  ["bot bull acad sinica (taiwan)"] = "Botanical bulletin of Academia Sinica. New series. Zhong yang yan jiu yuan. Zhi wu yan jiu suo",
  ["bot centbl"] = "Botanisches Centralblatt",
  ["bot complut"] = "Botánica complutensis",
  ["bot gaz"] = "Botanical gazette (Chicago, Ill.)",
  ["bot j linn soc"] = "Botanical journal of the Linnean Society. Linnean Society of London",
  ["bot j scotl"] = "Botanical journal of Scotland",
  ["bot jahrb syst pflanzengesch pflanzengeogr"] = "Botanische Jahrbucher fur Systematik, Pflanzengeschichte und Pflanzengeographie",
  ["bot lett"] = "Botany letters",
  ["bot lith"] = "Botanica Lithuanica",
  ["bot mag tokyo"] = "Botanical Magazine Tokyo",
  ["bot mar"] = "Botanica Marina",
  ["bot mus leafl  harv univ"] = "Botanical Museum Leaflets, Harvard University",
  ["bot mus lealf harv univ"] = "Botanical Museum leaflets, Harvard University",
  ["bot pac"] = "Botanica Pacifica : journal of plant science and conservation",
  ["bot rev"] = "The Botanical review; interpreting botanical progress",
  ["bot sci"] = "Botanical sciences",
  ["bot serb"] = "Botanica serbica",
  ["bot stud"] = "Botanical studies",
  ["bot z"] = "Botanicheskiĭ zhurnal",
  ["bot ztg"] = "Botanische Zeitung",
  ["botan acta"] = "Botanica Acta",
  ["botanica (vilnius)"] = "Botanica (Vilnius, Lithuania)",
  ["botanica marina"] = "Botanica marina",
  ["botsw natl health bull"] = "Botswana national health bulletin",
  ["botsw notes rec"] = "Botswana notes and records",
  ["botulinum j"] = "The botulinum journal",
  ["bound value probl"] = "Boundary Value Problems",
  ["bound-layer meteorol"] = "Boundary-Layer Meteorology",
  ["boundary layer meteorol"] = "Boundary-Layer Meteorology",
  ["boundary value probl"] = "Boundary Value Problems",
  ["bov pract (stillwater)"] = "The bovine practitioner",
  ["bp stat zsebkv"] = "Budapest statisztikai zsebkönyve",
  ["bp szekesfov kozeu"] = "Budapest székesföváros közegészségügye",
  ["br archaeol rep, br ser"] = "British Archaeological Reports, British Series",
  ["br biotechnol j"] = "British biotechnology journal",
  ["br birds"] = "British birds; an illustrated magazine devoted to the birds on the British list",
  ["br book news"] = "British book news",
  ["br ceram trans"] = "British Ceramic Transactions",
  ["br chem dig"] = "The British chemical digest",
  ["br chirop j"] = "The British chiropody journal",
  ["br corros j"] = "British Corrosion Journal",
  ["br crop pr"] = "Seed Treatment: Progress and Prospects",
  ["br dent j"] = "British dental journal",
  ["br dent nurs j"] = "British dental nurses' journal (Fleetwood, England : 1994)",
  ["br dent surg assist"] = "The British dental surgery assistant",
  ["br educ res j"] = "British educational research journal",
  ["br food j"] = "British food journal (Croydon, England)",
  ["br food j hyg rev"] = "The British food journal and hygienic review",
  ["br foreign med chir rev"] = "The British and foreign medico-chirurgical review",
  ["br foreign med rev"] = "The British and foreign medical review",
  ["br heart j"] = "British heart journal",
  ["br herit"] = "British heritage",
  ["br hist illus"] = "British history illustrated",
  ["br homeopath j"] = "The British homoeopathic journal",
  ["br ir bot"] = "British & Irish botany",
  ["br ir orthopt j"] = "The British and Irish orthoptic journal",
  ["br j 18th cent stud"] = "The British journal for eighteenth-century studies",
  ["br j addict"] = "British journal of addiction",
  ["br j addict alcohol other drugs"] = "The British journal of addiction to alcohol and other drugs",
  ["br j anaesth"] = "British journal of anaesthesia",
  ["br j anim behav"] = "British Journal of Animal Behaviour",
  ["br j appl phys"] = "British Journal of Applied Physics",
  ["br j appl sci technol"] = "British journal of applied science & technology",
  ["br j audiol"] = "British journal of audiology",
  ["br j audiol suppl"] = "British journal of audiology. Supplement",
  ["br j biomed sci"] = "British journal of biomedical science",
  ["br j can stud"] = "British Journal of Canadian Studies",
  ["br j cancer"] = "British journal of cancer",
  ["br j cancer suppl"] = "The British journal of cancer. Supplement",
  ["br j cardiol"] = "The British journal of cardiology",
  ["br j clin pharmacol"] = "British journal of clinical pharmacology",
  ["br j clin pract"] = "The British journal of clinical practice",
  ["br j clin pract suppl"] = "British journal of clinical practice. Supplement",
  ["br j clin pract symp suppl"] = "British Journal of Clinical Practice. Symposium Supplement",
  ["br j clin psychol"] = "The British journal of clinical psychology",
  ["br j community nurs"] = "British journal of community nursing",
  ["br j criminol"] = "The British journal of criminology",
  ["br j dermatol"] = "The British journal of dermatology",
  ["br j dermatol syph"] = "The British journal of dermatology and syphilis",
  ["br j dev psychol"] = "The British journal of developmental psychology",
  ["br j diabetes vasc dis"] = "The British journal of diabetes & vascular disease",
  ["br j dis chest"] = "British journal of diseases of the chest",
  ["br j disord commun"] = "The British journal of disorders of communication",
  ["br j educ psychol"] = "The British journal of educational psychology",
  ["br j educ technol"] = "British journal of educational technology : journal of the Council for Educational Technology",
  ["br j exp pathol"] = "British journal of experimental pathology",
  ["br j fam plann"] = "The British journal of family planning",
  ["br j forensic pract"] = "British Journal of Forensic Practice",
  ["br j gen pract"] = "The British journal of general practice : the journal of the Royal College of General Practitioners",
  ["br j guid counc"] = "British journal of guidance & counselling",
  ["br j haematol"] = "British journal of haematology",
  ["br j health psychol"] = "British journal of health psychology",
  ["br j hist math"] = "British Journal for the History of Mathematics",
  ["br j hist philos"] = "British journal for the history of philosophy : BJHP : the journal of the British Society for the History of Philosophy",
  ["br j hist sci"] = "British journal for the history of science",
  ["br j hosp med"] = "British journal of hospital medicine",
  ["br j hosp med (lond)"] = "British journal of hospital medicine (London, England : 2005)",
  ["br j ind med"] = "British journal of industrial medicine",
  ["br j ind relat"] = "British journal of industrial relations",
  ["br j int stud"] = "British journal of international studies",
  ["br j law soc"] = "British journal of law and society",
  ["br j learn disabil"] = "British journal of learning disabilities",
  ["br j math stat psychol"] = "The British journal of mathematical and statistical psychology",
  ["br j med educ"] = "British journal of medical education",
  ["br j med hypn"] = "British journal of medical hypnotism : official organ of the British Society of Medical Hypnotists",
  ["br j med med res"] = "British journal of medicine and medical research",
  ["br j med pract"] = "British Journal of Medical Practitioners",
  ["br j med psychol"] = "The British journal of medical psychology",
  ["br j med surg urol"] = "British journal of medical & surgical urology",
  ["br j midwifery"] = "British journal of midwifery",
  ["br j mil hist"] = "British journal for military history",
  ["br j neurosurg"] = "British journal of neurosurgery",
  ["br j nurs"] = "British journal of nursing (Mark Allen Publishing)",
  ["br j nutr"] = "British Journal of Nutrition",
  ["br j obstet gynaecol"] = "British journal of obstetrics and gynaecology",
  ["br j occup ther"] = "The British journal of occupational therapy",
  ["br j ophthalmol"] = "The British journal of ophthalmology",
  ["br j oral maxillofac surg"] = "The British journal of oral & maxillofacial surgery",
  ["br j oral surg"] = "The British journal of oral surgery",
  ["br j orthod"] = "British journal of orthodontics",
  ["br j pain"] = "British journal of pain",
  ["br j perioper nurs"] = "British journal of perioperative nursing : the journal of the National Association of Theatre Nurses",
  ["br j pharm res"] = "British journal of pharmaceutical research",
  ["br j pharmacol"] = "British Journal of Pharmacology",
  ["br j pharmacol chemother"] = "British journal of pharmacology and chemotherapy",
  ["br j philos sci"] = "The British journal for the philosophy of science",
  ["br j phys med"] = "The British journal of physical medicine : including its application to industry",
  ["br j physiol opt"] = "The British journal of physiological optics",
  ["br j physiother"] = "British journal of physiotherapy",
  ["br j plast surg"] = "British journal of plastic surgery",
  ["br j polit int relat"] = "British journal of politics & international relations",
  ["br j polit sci"] = "British journal of political science",
  ["br j prev soc med"] = "British journal of preventive & social medicine",
  ["br j psychiatr soc work"] = "British journal of psychiatric social work",
  ["br j psychiatry"] = "The British journal of psychiatry : the journal of mental science",
  ["br j psychiatry suppl"] = "The British journal of psychiatry. Supplement",
  ["br j psychol"] = "British journal of psychology (London, England : 1953)",
  ["br j psychol gen sect"] = "The British journal of psychology. General section",
  ["br j psychother"] = "British journal of psychotherapy",
  ["br j radiol"] = "The British journal of radiology",
  ["br j radiol suppl"] = "British journal of radiology. Supplement",
  ["br j rheumatol"] = "British journal of rheumatology",
  ["br j sex med"] = "British journal of sexual medicine",
  ["br j soc clin psychol"] = "The British journal of social and clinical psychology",
  ["br j soc med"] = "British journal of social medicine",
  ["br j soc psychol"] = "The British journal of social psychology",
  ["br j soc work"] = "British journal of social work",
  ["br j sociol"] = "The British journal of sociology",
  ["br j sociol educ"] = "British journal of sociology of education",
  ["br j sports hist"] = "The British journal of sports history",
  ["br j sports med"] = "British journal of sports medicine",
  ["br j surg"] = "The British journal of surgery",
  ["br j theatre nurs"] = "The British journal of theatre nursing : NATNews : the official journal of the National Association of Theatre Nurses",
  ["br j tuberc dis chest"] = "The British journal of tuberculosis and diseases of the chest",
  ["br j urol"] = "British journal of urology",
  ["br j vener dis"] = "The British journal of venereal diseases",
  ["br j virol"] = "British journal of virology",
  ["br j vis impair"] = "The British journal of visual impairment",
  ["br libr j"] = "The British Library journal. British Library",
  ["br med bull"] = "British medical bulletin",
  ["br med j"] = "British medical journal",
  ["br med j (clin res ed)"] = "British medical journal (Clinical research ed.)",
  ["br microbiol res j"] = "British microbiology research journal",
  ["br orthopt j"] = "The British orthoptic journal",
  ["br paramed j"] = "British paramedic journal",
  ["br polym j"] = "British Polymer Journa",
  ["br poult sci"] = "British poultry science",
  ["br rev econ issues"] = "British review of economic issues",
  ["br sci news"] = "British science news",
  ["br stud monit"] = "The British studies monitor",
  ["br vet j"] = "The British veterinary journal",
  ["br yearb sci"] = "Britannica yearbook of science and the future",
  ["bradford book"] = "A Bradford Book",
  ["brain (bacau)"] = "Brain : broad research in artificial intelligence and neuroscience",
  ["brain behav"] = "Brain and behavior",
  ["brain behav evol"] = "Brain, behavior and evolution",
  ["brain behav immun"] = "Brain, behavior, and immunity",
  ["brain behav immun health"] = "Brain, behavior, & immunity - health",
  ["brain cell biol"] = "Brain cell biology",
  ["brain circ"] = "Brain circulation",
  ["brain cogn"] = "Brain and cognition",
  ["brain commun"] = "Brain communications",
  ["brain comput interfaces (abingdon)"] = "Brain computer interfaces (Abingdon, England)",
  ["brain connect"] = "Brain connectivity",
  ["brain dev"] = "Brain & development",
  ["brain disord"] = "Brain disorders",
  ["brain disord ther"] = "Brain disorders & therapy",
  ["brain hemorrhages"] = "Brain hemorrhages",
  ["brain imaging behav"] = "Brain imaging and behavior",
  ["brain impair"] = "Brain impairment : a multidisciplinary journal of the Australian Society for the Study of Brain Impairment",
  ["brain impairment"] = "Brain Impairment",
  ["brain inf"] = "Brain Informatics",
  ["brain inform"] = "Brain informatics",
  ["brain inform (2018)"] = "Brain informatics : international conference, BI 2018, Arlington, TX, USA, December 7-9, 2018, proceedings. International Conference on Brain Informatics (2018 : Arlington, Tex.)",
  ["brain inform health (2015)"] = "Brain Informatics and Health : 8th International Conference, BIH 2015, London, UK, August 30-September 2, 2015 : proceedings. BIH (Conference) (8th : 2015 : London, England)",
  ["brain inj"] = "Brain injury",
  ["brain lang"] = "Brain and language",
  ["brain multiphys"] = "Brain multiphysics",
  ["brain nerve"] = "Brain and nerve = Shinkei kenkyū no shinpo",
  ["brain neurosci adv"] = "Brain and neuroscience advances",
  ["brain pathol"] = "Brain pathology (Zurich, Switzerland)",
  ["brain plast"] = "Brain plasticity (Amsterdam, Netherlands)",
  ["brain res"] = "Brain Research",
  ["brain res brain res protoc"] = "Brain research. Brain research protocols",
  ["brain res brain res rev"] = "Brain research. Brain research reviews",
  ["brain res bull"] = "Brain research bulletin",
  ["brain res cogn brain res"] = "Brain research. Cognitive brain research",
  ["brain res dev brain res"] = "Brain research. Developmental brain research",
  ["brain res gene expr patterns"] = "Brain research. Gene expression patterns",
  ["brain res mol brain res"] = "Brain research. Molecular brain research",
  ["brain res protoc"] = "Brain Research Protocols",
  ["brain res rev"] = "Brain research reviews",
  ["brain sci"] = "Brain sciences",
  ["brain sci adv"] = "Brain Science Advances",
  ["brain stimul"] = "Brain stimulation",
  ["brain struct funct"] = "Brain structure & function",
  ["brain topogr"] = "Brain topography",
  ["brain tumor pathol"] = "Brain tumor pathology",
  ["brain tumor res treat"] = "Brain tumor research and treatment",
  ["brain, behav evol"] = "Brain, Behavior and Evolution",
  ["brain-appar commun: j bacomics"] = "Brain-Apparatus Communication: A Journal of Bacomics",
  ["brain-comput interfaces"] = "Brain-Computer Interfaces",
  ["brains minds media"] = "Brains, minds & media : journal of new media in neural and cognitive science and education",
  ["brandeis j fam law"] = "Brandeis journal of family law",
  ["brandeis law j"] = "Brandeis law journal",
  ["brandeis rev"] = "Brandeis review",
  ["bras med"] = "Brasil-médico",
  ["bras med cir"] = "Brasil médico-cirúrgico",
  ["braspen j"] = "BRASPEN journal",
  ["bratisl lek listy"] = "Bratislavské lekárske listy",
  ["braunschw veroff gesch pharm naturwiss"] = "Braunschweiger veröffentlichungen zur geschichte der pharmazie und der naturwissenschaften",
  ["braz arch biol technol"] = "Brazilian archives of biology and technology = (Archivos de biologia e tecnologia)",
  ["braz dent j"] = "Brazilian dental journal",
  ["braz dent sci"] = "Brazilian dental science",
  ["braz econ stud"] = "Brazilian economic studies",
  ["braz j anal chem"] = "Brazilian Journal Of Analytical Chemistry",
  ["braz j anesthesiol"] = "Brazilian journal of anesthesiology (Elsevier)",
  ["braz j anim sci"] = "Brazilian Journal of Animal Science",
  ["braz j aquat sci technol"] = "Brazilian Journal of Aquatic Sciences and Technology",
  ["braz j biol"] = "Brazilian journal of biology = Revista brasleira de biologia",
  ["braz j bot"] = "Brazilian Journal Of Botany",
  ["braz j cardiovasc surg"] = "Brazilian journal of cardiovascular surgery",
  ["braz j chem"] = "Brazilian Journal of Chemistry",
  ["braz j chem eng"] = "Brazilian Journal of Chemical Engineering",
  ["braz j food technol"] = "Brazilian Journal of Food Technology",
  ["braz j geol"] = "Brazilian Journal of Geology",
  ["braz j infect dis"] = "The Brazilian journal of infectious diseases : an official publication of the Brazilian Society of Infectious Diseases",
  ["braz j mater sci eng"] = "Brazilian Journal of Materials Science and Engineering",
  ["braz j med biol res"] = "Brazilian journal of medical and biological research = Revista brasileira de pesquisas médicas e biológicas",
  ["braz j microbiol"] = "Brazilian journal of microbiology : [publication of the Brazilian Society for Microbiology]",
  ["braz j oceanogr"] = "Brazilian Journal of Oceanography",
  ["braz j of sci"] = "Brazilian Journal of Science",
  ["braz j of sci technol"] = "Brazilian Journal of Science and Technology",
  ["braz j oper prod manage"] = "Brazilian Journal Of Operations And Production Management",
  ["braz j oral sci"] = "Brazilian journal of oral sciences",
  ["braz j otorhinolaryngol"] = "Brazilian journal of otorhinolaryngology",
  ["braz j pharm sci"] = "Brazilian Journal Of Pharmaceutical Sciences",
  ["braz j phys"] = "Brazilian Journal of Physics",
  ["braz j phys ther"] = "Brazilian journal of physical therapy",
  ["braz j plant physiol"] = "Revista brasileira de fisiologia vegetal",
  ["braz j popul stud"] = "Brazilian journal of population studies",
  ["braz j poult sci"] = "Brazilian Journal of Poultry Science",
  ["braz j probab stat"] = "Brazilian journal of probability and statistics",
  ["braz j psychiatry"] = "Revista brasileira de psiquiatria (São Paulo, Brazil : 1999)",
  ["braz j vet pathol"] = "Brazilian Journal Of Veterinary Pathology",
  ["braz j vet res anim sci"] = "Brazilian journal of veterinary research and animal science",
  ["braz oral res"] = "Brazilian oral research",
  ["brazilian electronic j econ"] = "Brazilian Electronic Journal of Economics",
  ["brazilian j appl econ"] = "Brazilian Journal of Applied Economics",
  ["brazilian j polit economy"] = "Brazilian Journal of Political Economy",
  ["brazilian rev econometrics"] = "Brazilian Review of Econometrics",
  ["breast cancer"] = "Breast cancer (Tokyo, Japan)",
  ["breast cancer (auckl)"] = "Breast cancer : basic and clinical research",
  ["breast cancer (dove med press)"] = "Breast cancer (Dove Medical Press)",
  ["breast cancer manag"] = "Breast cancer management",
  ["breast cancer manage"] = "Breast Cancer Management",
  ["breast cancer online"] = "Breast cancer online : BCO",
  ["breast cancer res"] = "Breast cancer research : BCR",
  ["breast cancer res treat"] = "Breast cancer research and treatment",
  ["breast care (basel)"] = "Breast care (Basel, Switzerland)",
  ["breast dis"] = "Breast disease",
  ["breast imaging (2012)"] = "Breast imaging : 11th International Workshop, IWDM 2012, Philadelphia, PA, USA, July 8-11, 2012 : proceedings. International Workshop on Breast Imaging (11th : 2012 : Philadelphia, Pa.)",
  ["breast j"] = "The breast journal",
  ["breastfeed med"] = "Breastfeeding medicine : the official journal of the Academy of Breastfeeding Medicine",
  ["breastfeed rev"] = "Breastfeeding review : professional publication of the Nursing Mothers' Association of Australia",
  ["breathe (sheff)"] = "Breathe (Sheffield, England)",
  ["breed sci"] = "Breeding science",
  ["brethr life thought"] = "Brethren life and thought",
  ["brew distill int"] = "Brewer and Distiller International",
  ["brew sci"] = "Brewing science (Nürnberg)",
  ["brewingsci monatsschr brauwiss"] = "BrewingScience - Monatsschrift fur Brauwissenschaft",
  ["bridg wash d c"] = "Bridge (Washington, D.C. : 1992)",
  ["bridge (kans city)"] = "The Bridge",
  ["bridge (wash d c)"] = "Bridge (Washington, D.C. : 1969)",
  ["bridge struct"] = "Bridge Structures",
  ["brief bioinform"] = "Briefings in bioinformatics",
  ["brief funct genomic proteomic"] = "Briefings in functional genomics & proteomics",
  ["brief funct genomics"] = "Briefings in functional genomics",
  ["brief med ethics"] = "Briefings in medical ethics",
  ["brief treat crisis interv"] = "Brief treatment and crisis intervention",
  ["briefings bioinf"] = "Briefings in Bioinformatics",
  ["briefings funct genomics"] = "Briefings in Functional Genomics",
  ["brigh young univ law rev"] = "Brigham Young University law review",
  ["brigham young univ stud"] = "Brigham Young University studies. Brigham Young University",
  ["brills tibet stu lib"] = "Tibetan Buddhist Literature and Praxis: Studies in Its Formative Period, 900-1400",
  ["brill’s companions class recept"] = "Brill’s Companions to Classical Reception",
  ["bristol dig"] = "The Bristol digest",
  ["bristol med chir j"] = "Bristol medico-chirurgical journal (1963)",
  ["bristol med chir j (1883)"] = "Bristol medico-chirurgical journal (1883)",
  ["brit j ind relat"] = "British Journal of Industrial Relations",
  ["brit rev econ issues"] = "British Review of Economic Issues",
  ["british acad classical medieval logic texts"] = "The British Academy Classical and Medieval Logic Texts",
  ["british j hist sci"] = "British Society for the History of Science. British Journal for the History of Science",
  ["british j math statist psych"] = "The British Journal of Mathematical and Statistical Psychology",
  ["british j philos sci"] = "The British Journal for the Philosophy of Science",
  ["brjac–braz j anal chem"] = "BrJAC–Brazilian Journal of Analytical Chemistry",
  ["bromley local hist"] = "Bromley local history",
  ["bronte soc trans"] = "Brontë Society transactions. Brontë Society",
  ["brookhaven symp biol"] = "Brookhaven symposia in biology",
  ["brookings bull"] = "Brookings bulletin (Washington, D.C. : 1962)",
  ["brookings pap econ act"] = "Brookings papers on economic activity",
  ["brookings rev"] = "The Brookings review",
  ["brookings-wharton pap finan services"] = "Brookings-Wharton Papers on Financial Services",
  ["brookings-wharton pap urban aff"] = "Brookings-Wharton Papers on Urban Affairs",
  ["brooklyn hosp j"] = "The Brooklyn Hospital journal",
  ["brooklyn j int law"] = "Brooklyn journal of international law",
  ["brooklyn law rev"] = "Brooklyn law review",
  ["broward rev"] = "Broward review",
  ["brown univ dig addict theory appl"] = "Brown University digest of addiction theory and application : DATA",
  ["brown univ psychopharmacol update"] = "Brown University Psychopharmacology Update",
  ["brucken math"] = "Brucken zur Mathematik",
  ["bruno pini math anal semin"] = "Bruno Pini Mathematical Analysis Seminar",
  ["bruns beitr klin chir"] = "Bruns' Beiträge zur klinischen Chirurgie",
  ["bruns beitr klin chir (1971)"] = "Bruns' Beiträge für klinische Chirurgie",
  ["bruns beitr klinischen chir"] = "Bruns’ Beitrage zur klinischen Chirurgie",
  ["brux med"] = "Bruxelles médical",
  ["bryonora (praha)"] = "Bryonora : zpravodaj Bryologicko-lichenologické sekce ČBS",
  ["bryophyt bibl"] = "Bryophytorum bibliotheca",
  ["bryophyt divers evol"] = "Bryophyte diversity and evolution",
  ["bscs pam"] = "BSCS pamphlets",
  ["bsg proc"] = "BSG Proceedings",
  ["bsgf earth sci bull"] = "BSGF Earth Sciences Bulletin",
  ["bshm bull"] = "BSHM Bulletin. Journal of the British Society for the History of Mathematics",
  ["bshp new texts hist philos"] = "BSHP New Texts in the History of Philosophy",
  ["bt technol j"] = "BT Technology Journal",
  ["bts compt-iut gestion enseign tech super"] = "BTS Comptable-IUT de Gestion. Enseignement Technique Superieur",
  ["btta rev"] = "B.T.T.A. review",
  ["bubble sci eng technol"] = "Bubble science engineering and technology",
  ["buch augenarzt"] = "Bucherei des Augenarztes",
  ["buckwheat newsletter|fagopyrum"] = "Fagopyrum : novosti o ajdi",
  ["buffalo law rev"] = "Buffalo law review",
  ["build acoust"] = "Building Acoustics",
  ["build cities"] = "Buildings and Cities",
  ["build environ"] = "Building and Environment",
  ["build oper manage"] = "Building operating management",
  ["build res inf"] = "Building Research and Information",
  ["build sci"] = "Building Science",
  ["build serv eng res technol"] = "Building Services Engineering Research and Technology",
  ["build simul"] = "Building Simulation",
  ["build syst des"] = "Building systems design",
  ["buildsys15 (2015)"] = "BuildSys'15 : proceedings of the 2nd ACM International Conference on Embedded Systems for Energy-Efficient Buildings : November 4-5, 2015, Seoul, South Korea. ACM Conference on Embedded Systems for Energy-Efficient Buildings (2nd : 2015 : Seoul, Korea)",
  ["built environ"] = "Built environment",
  ["built environ project asset manage"] = "Built Environment Project and Asset Management",
  ["bul acad stiinte repub mold mat"] = "Buletinul Academiei de Stiinte a Republicii Moldova. Matematica. Izvestiya Akademii Nauk Respubliki Moldova. Matematika",
  ["bul inst politeh iaşi secţ i mat mec teor fiz"] = "Buletinul Institutului Politehnic din Iaşi",
  ["bul inst politeh lasi sect stiinte socio umane"] = "Buletinul Institutului Politehnic din Iaşi : Secția Științe socio-umane",
  ["bul keluarga"] = "Buletin keluarga",
  ["bul lab veteriner"] = "Buletin lab veteriner",
  ["bul pol acad sci chem"] = "Bulletin of the Polish Academy of Sciences. Chemistry",
  ["bul stiint sect stiint medicale acad republicii pop romane"] = "Buletin ştiinţific. Secţiunea de ştiinţe medicale. Academia Republicii Populare Romîne",
  ["bul stiint univ politeh timis ser mat fiz"] = "Buletinul Stiintific al Universitatii “Politehnica” din Timisoara. Seria Matematica-Fizica. Scientific Bulletin of the Politehnica University of Timisoara. Transactions on Mathematics & Physics",
  ["bul univ shtet tiranes ser shk mjekesore"] = "Buletin i Universitetit Shtetëror të Tiranës. Seria Shkencat mjekësore",
  ["bul ştiinţ inst politehn cluj-napoca ser mat apl mec"] = "Buletinul Ştiinţific al Institutului Politehnic Cluj-Napoca",
  ["bul ştiinţ univ baia mare ser b"] = "Buletinul Ştiinţific al Universităţii din Baia Mare",
  ["bul ştiinţ univ tehn timișoara mat fiz"] = "Universităţii Tehnice din Timișoara",
  ["bulg astron j"] = "Bulgarian Astronomical Journal",
  ["bulg chem commun"] = "Bulgarian Chemical Communications",
  ["bulg hist rev"] = "Bulgarian historical review",
  ["bulg j agric sci"] = "Bulgarian Journal Of Agricultural Science",
  ["bulg j phys"] = "Bulgarian Journal of Physics",
  ["bulg j vet med"] = "Bulgarian journal of veterinary medicine",
  ["bulg klin"] = "Bŭlgarska klinika",
  ["bulgar j phys"] = "Bulgarian Journal of Physics",
  ["bull (univ minn econ dev cent)"] = "Bulletin (University of Minnesota. Economic Development Center)",
  ["bull aas"] = "Bulletin of the AAS",
  ["bull acad chir dent (paris)"] = "Bulletin de l'Académie de chirurgie dentaire",
  ["bull acad dent (paris)"] = "Bulletin de l'Académie dentaire",
  ["bull acad dent handicap"] = "Bulletin of Academy of Dentistry for the Handicapped",
  ["bull acad gen dent"] = "The Bulletin of the Academy of General Dentistry",
  ["bull acad med"] = "Bulletin de l'Academie de médecine",
  ["bull acad med clevel"] = "Bulletin - Academy of Medicine of Cleveland",
  ["bull acad med n j"] = "Bulletin - Academy of Medicine of New Jersey",
  ["bull acad med tor"] = "Bulletin of the Academy of Medicine, Toronto",
  ["bull acad natl chir dent"] = "Bulletin de l'Académie nationale de chirurgie dentaire (1983)",
  ["bull acad natl med"] = "Bulletin de l'Académie nationale de médecine",
  ["bull acad pol sci [biol]"] = "Bulletin de l’Academie Polonaise des Sciences. Serie des Sciences Biologiques",
  ["bull acad pol sci biol"] = "Bulletin de l'Académie polonaise des sciences. Série des sciences biologiques",
  ["bull acad r med belg"] = "Bulletin de l'Académie royale de médecine de Belgique",
  ["bull acad sci ussr, phys ser"] = "Bulletin of the Academy of Sciences of the USSR, Physical Series [translation of Izvestiya Akademii Nauk SSSR, Seriya Fizicheskaya]",
  ["bull acad serbe sci arts"] = "Bulletin. Srpska akademija nauka i umetnosti. Odeljenje medicinskih nauka",
  ["bull acad serbe sci arts cl sci math natur"] = "Académie Serbe des Sciences et des Arts",
  ["bull acad serbe sci cl sci math nat sci nat"] = "Bulletin - Académie serbe des sciences. Classe des sciences mathématiques et naturelles. Sciences naturelles",
  ["bull acad soc lorraines sci"] = "Bulletin de l'Académie lorraine des sciences",
  ["bull acad vet fr"] = "Bulletin de l'Académie vétérinaire de France",
  ["bull actual med"] = "Bulletin des actualités médicales",
  ["bull acuff clin knoxv tenn"] = "Bulletin. Acuff Clinic, Knoxville, Tenn",
  ["bull adelphi hosp brooklyn"] = "Bulletin. Adelphi Hospital (Brooklyn, New York, N.Y.)",
  ["bull akita res inst food brew"] = "Bulletin of the Akita research institute of food and brewing",
  ["bull akron city hosp"] = "The Bulletin of Akron City Hospital",
  ["bull akron dent soc"] = "Bulletin of the Akron Dental Society",
  ["bull alameda cty dent soc"] = "Bulletin - Alameda County Dental Society",
  ["bull alexandria fac"] = "Bulletin. Kullīyat Ṭibb al-Iskandarīyah",
  ["bull alger carcinol"] = "Bulletin algérien de carcinologie",
  ["bull allahabad math soc"] = "Bulletin of the Allahabad Mathematical Society",
  ["bull allergy"] = "Bulletin of allergy",
  ["bull alloy phase diagr"] = "Bulletin of Alloy Phase Diagrams",
  ["bull allyn mus"] = "Bulletin of the Allyn Museum",
  ["bull am acad arts sci"] = "Bulletin - American Academy of Arts and Sciences. American Academy of Arts and Sciences",
  ["bull am acad psychiatry law"] = "The Bulletin of the American Academy of Psychiatry and the Law",
  ["bull am assoc dent ed"] = "The Bulletin - American Association of Dental Editors",
  ["bull am assoc hist nurs"] = "Bulletin (American Association for the History of Nursing)",
  ["bull am assoc hosp dent"] = "Bulletin - American Association of Hospital Dentists",
  ["bull am assoc med soc workers"] = "Bulletin. American Association of Medical Social Workers",
  ["bull am astron soc"] = "Bulletin of the American Astronomical Society",
  ["bull am cancer soc"] = "Bulletin. American Cancer Society",
  ["bull am coll nurse midwifery"] = "Bulletin of the American College of Nurse-Midwifery",
  ["bull am coll nurse midwives"] = "Bulletin of the American College of Nurse-Midwives",
  ["bull am coll physicians"] = "The Bulletin of the American College of Physicians",
  ["bull am coll surg"] = "Bulletin of the American College of Surgeons",
  ["bull am math soc"] = "Bulletin of the American Mathematical Society",
  ["bull am meteorol soc"] = "Bulletin of the American Meteorological Society",
  ["bull am mus nat hist"] = "Bulletin of the American Museum of Natural History",
  ["bull am pharm assoc"] = "Bulletin. American Pharmaceutical Association. Committee on National Formulary",
  ["bull am phys soc"] = "Bulletin of The American Physical Society",
  ["bull am protestant hosp assoc"] = "Bulletin - American Protestant Hospital Association",
  ["bull am schools orient res"] = "Bulletin of the American Schools of Oriental Research. American Schools of Oriental Research",
  ["bull am soc hosp pharm"] = "Bulletin. American Society of Hospital Pharmacists",
  ["bull am soc inf sci"] = "Bulletin of the American Society for Information Science",
  ["bull am soc overseas res"] = "Bulletin of the American Society of Overseas Research",
  ["bull am soc papyrol"] = "The Bulletin of the American Society of Papyrologists. American Society of Papyrologists",
  ["bull am soc vet clin pathol"] = "Bulletin of the American Society of Veterinary Clinical Pathologists",
  ["bull amer math soc (ns)"] = "American Mathematical Society. Bulletin. New Series",
  ["bull amer meteorol soc"] = "Bulletin of the American Meteorological Society",
  ["bull anesth hist"] = "Bulletin of anesthesia history",
  ["bull angew geol"] = "Bulletin angewandte Geologie",
  ["bull anim behav"] = "Bulletin of Animal Behaviour",
  ["bull anim health prod afr"] = "Bulletin of animal health and production in Africa. Bulletin des santé et production animales en Afrique",
  ["bull annu epidemiol"] = "Bulletin annuel d'épidemiologie",
  ["bull assoc am physicians surg"] = "Bulletin of the Association of American Physicians and Surgeons",
  ["bull assoc anat (nancy)"] = "Bulletin de l'Association des anatomistes",
  ["bull assoc diplomes microbiol fac pharm nancy"] = "Bulletin de l'Association des diplômés de microbiologie de la Faculté de pharmacie de Nancy",
  ["bull assoc déf intérêts jura"] = "Bulletin de l’Association pour la Défense des Intérêts du Jura",
  ["bull assoc fr etud cancer"] = "Bulletin de l'Association française pour l'étude du cancer",
  ["bull assoc geogr fr"] = "Bulletin de l'Association de géographes français",
  ["bull assoc guillaume bude"] = "Bulletin de l'Association Guillaume Budé. Association Guillaume Budé",
  ["bull assoc med haiti"] = "Bulletin. Association médicale haïtienne",
  ["bull assoc med libr"] = "Bulletin of the Association of Medical Librarians",
  ["bull assoc mycol francoph belg"] = "Bulletin de l'Association des mycologues francophones de Belgique",
  ["bull assoc tunis plan fam"] = "Bulletin de l'Association tunisienne du planning familial : publication interne",
  ["bull astron inst neth"] = "Bulletin of the Astronomical Institutes of the Netherlands",
  ["bull astron soc india"] = "Bulletin of the Astronomical Society of India",
  ["bull at sci"] = "The Bulletin of the atomic scientists",
  ["bull atl cape may cty dent soc"] = "Bulletin / Atlantic-Cape May County Dental Society",
  ["bull atmos sci technol"] = "Bulletin of Atmospheric Science and Technology",
  ["bull audiophonol"] = "Bulletin d'audiophonologie",
  ["bull aust math soc"] = "Bulletin of the Australian Mathematical Society",
  ["bull austral math soc"] = "Bulletin of the Australian Mathematical Society",
  ["bull belg math soc simon stevin"] = "Bulletin of the Belgian Mathematical Society - Simon Stevin",
  ["bull bergen cty dent soc"] = "The Bulletin of the Bergen County Dental Society",
  ["bull bibl nat"] = "Bulletin de la Bibliothèque nationale. Bibliothèque nationale (France)",
  ["bull bibl r albert"] = "Bulletin - Bibliothèque royale Albert Ier. Bibliotheque royale Albert Ier",
  ["bull bibliogr"] = "Bulletin of bibliography (Westwood, Mass. : 1979)",
  ["bull biogeogr soc jpn"] = "Bulletin of the Bio-geographical Society of Japan = Nihon Seibutsuchiri Gakkai Kaiho",
  ["bull biol fr belg"] = "Bulletin biologique de la France et de la Belgique",
  ["bull biol soc wash"] = "Bulletin of the Biological Society of Washington",
  ["bull biomath"] = "Bulletin of Biomathematics",
  ["bull bodenkd ges schweiz"] = "Bulletin Bodenkundliche Gesellschaft der Schweiz",
  ["bull br arachnol soc"] = "Bulletin (British Arachnological Society)",
  ["bull br assoc orient"] = "Bulletin of the British Association of Orientalists. British Association of Orientalists",
  ["bull br assoc sport med"] = "Bulletin - British Association of Sport and Medicine",
  ["bull br mus"] = "Bulletin of the British Museum (Natural History). Historical series",
  ["bull br mus nat hist, bot"] = "Bulletin of the British Museum of Natural History, Botany",
  ["bull br mycol soc"] = "Bulletin of the British Mycological Society",
  ["bull br myriap isopod group"] = "Bulletin of the British Myriapod and Isopod Group",
  ["bull br ornithol club"] = "Bulletin of the British Ornithologists’ Club",
  ["bull br psychol soc"] = "Bulletin of the British Psychological Society",
  ["bull br soc middle east stud"] = "Bulletin - British Society for Middle Eastern Studies. British Society for Middle Eastern Studies",
  ["bull braz math soc"] = "Bulletin Of The Brazilian Mathematical Society",
  ["bull braz math soc (ns)"] = "Bulletin of the Brazilian Mathematical Society. New Series. Boletim da Sociedade Brasileira de Matematica",
  ["bull brew sci"] = "Bulletin of Brewing Science",
  ["bull bronx cty dent soc"] = "Bulletin. Bronx County Dental Society",
  ["bull brooklyn"] = "Bulletin. Brooklyn. Jewish Hospital",
  ["bull cairo univ fac agric"] = "Bulletin (Jāmiʻat al-Qāhirah. Kullīyat al-Zirāʻah)",
  ["bull calcutta math soc"] = "Bulletin of the Calcutta Mathematical Society",
  ["bull calcutta sch trop med"] = "Bulletin of the Calcutta School of Tropical Medicine",
  ["bull calif agric exp stn"] = "Bulletin of the California Agricultural Experiment Station",
  ["bull calif diet assoc"] = "Bulletin of the California Dietetic Association",
  ["bull calif state nurses assoc"] = "Bulletin. California State Nurses' Association",
  ["bull can biochem soc"] = "Bulletin of the Canadian Biochemical Society",
  ["bull can energy geosci"] = "Bulletin Of Canadian Energy Geoscience",
  ["bull can hist med"] = "Bulletin canadien d'histoire de la médecine = Canadian bulletin of medical history",
  ["bull can petrol geol"] = "Bulletin of Canadian Petroleum Geology",
  ["bull can tuberc assoc"] = "Bulletin - Canadian Tuberculosis Association",
  ["bull cancer"] = "Bulletin du cancer",
  ["bull cancer (paris)"] = "Bulletin du Cancer",
  ["bull cancer inst okayama univ med sch"] = "Bulletin. Okayama Daigaku. Igakubu. Gangen Kenkyū Shisetsu",
  ["bull cancer radiother"] = "Bulletin du cancer. Radiothérapie : journal de la Société française du cancer : organe de la société française de radiothérapie oncologique",
  ["bull carnegie mus nat hist"] = "Bulletin of Carnegie Museum of Natural History",
  ["bull cent hist econ soc reg lyon"] = "Bulletin du Centre d'histoire économique et sociale de la région lyonnaise. Centre d'histoire économique et sociale de la région lyonnaise",
  ["bull central res inst fukuoka univ"] = "Bulletin of Central Research Institute Fukuoka University",
  ["bull cercle benelux hist pharm"] = "Bulletin - Cercle Benelux d'histoire de la pharmacie",
  ["bull cercle vaud bot"] = "Bulletin du Cercle vaudois de botanique",
  ["bull cercle vaudois bot"] = "Bulletin du Cercle Vaudois de Botanique",
  ["bull cerp"] = "Bulletin du C. E. R. P",
  ["bull charlotte meml hosp"] = "Bulletin. Charlotte, N. C. Charlotte Memorial Hospital",
  ["bull chem soc ethiop"] = "Bulletin of the Chemical Society of Ethiopia",
  ["bull chem soc jpn"] = "Bulletin of the Chemical Society of Japan",
  ["bull chest dis res inst kyoto univ"] = "Bulletin of the Chest Disease Research Institute, Kyoto University",
  ["bull chic heart assoc"] = "Bulletin. Chicago Heart Association",
  ["bull chic med soc"] = "Bulletin. Chicago Medical Society",
  ["bull cincinnati dent soc"] = "Bulletin - Cincinnati Dental Society",
  ["bull cl lett sci morales polit"] = "Bulletin de la Classe des lettres et des sciences morales et politiques. Académie royale des sciences, des lettres et des beaux-arts de Belgique. Classe des lettres et des sciences morales et politiques",
  ["bull cl sci acad r belg"] = "Bulletin de la Classe des Sciences, Academie Royale de Belgique",
  ["bull cl sci math nat sci math"] = "Bulletin. Classe des Sciences Mathematiques et Naturelles. Sciences Mathematiques",
  ["bull cleve dent soc"] = "The Bulletin of the Cleveland Dental Society",
  ["bull clevel med libr"] = "The bulletin of the Cleveland Medical Library. Cleveland Medical Library Association",
  ["bull clin neurosci"] = "Bulletin of clinical neurosciences",
  ["bull coll gen pract can"] = "Bulletin. College of General Practice of Canada",
  ["bull coll lib arts"] = "Wen shi zhe xue bao",
  ["bull college sci univ ryukyus"] = "University of the Ryukyus",
  ["bull colo"] = "Bulletin. Colorado. State Board of Health",
  ["bull commun dis med notes"] = "Bulletin of communicable diseases and medical notes. United Nations Relief and Rehabilitation Administration",
  ["bull comput appl math"] = "Bulletin of Computational Applied Mathematics",
  ["bull concern asian sch"] = "Bulletin of concerned Asian scholars",
  ["bull conn state dent assoc"] = "Bulletin. Connecticut State Dental Association",
  ["bull contra costa dent soc"] = "Bulletin Of The Contra Costa Dental Society",
  ["bull cosm club"] = "Bulletin - Cosmos Club. Cosmos Club (Washington, D.C.)",
  ["bull counc br res levant"] = "Bulletin for the Council for British Research in the Levant\tBull. Counc. Br. Res. Levant",
  ["bull creighton univ"] = "Bulletin. Creighton University. School of Medicine",
  ["bull czechoslovak law"] = "Bulletin of Czechoslovak law",
  ["bull dade cty med assoc inc"] = "Bulletin - Dade County Medical Association",
  ["bull dalnevost malakol obs"] = "Bi︠u︡lletenʹ Dalʹnevostochnogo malakologicheskogo obshchestva",
  ["bull dayton dent soc"] = "Bulletin. Dayton (Ohio) Dental Society",
  ["bull dent guid counc cereb palsy"] = "Bulletin of the Dental Guidance Council for Cerebral Palsy",
  ["bull dep health commonw ky"] = "Bulletin of the Department of Health, Commonwealth of Kentucky. Kentucky. Department of Health",
  ["bull dept gen ed college sci tech nihon univ"] = "Bulletin of Department of General Education",
  ["bull dight inst univ minn"] = "Bulletin - Dight Institute of the University of Minnesota",
  ["bull doc sanit"] = "Bulletin de documentation sanitaire",
  ["bull earthquake eng"] = "Bulletin of Earthquake Engineering",
  ["bull east sib sci cent sib branch russ acad med sci"] = "Bulletin of Eastern-Siberian Scientific Center of Siberian Branch of Russian Academy of Medical Sciences",
  ["bull ec fr extr orient"] = "Bulletin de l'Ecole française d'Extrême-Orient",
  ["bull ecers"] = "Bulletin of ECerS",
  ["bull ecol soc am"] = "Bulletin of the Ecological Society of America",
  ["bull econ res"] = "Bulletin of economic research",
  ["bull econ soc maroc"] = "Bulletin économique et social du Maroc",
  ["bull eighth dist dent soc"] = "Bulletin of the 8th District Dental Society",
  ["bull electrochem"] = "Bulletin of Electrochemistry",
  ["bull eleventh dist dent soc"] = "Bulletin of the Eleventh District Dental Society",
  ["bull emerg trauma"] = "Bulletin of emergency and trauma",
  ["bull endem dis (baghdad)"] = "Bulletin of endemic diseases",
  ["bull eng geol env"] = "Bulletin of Engineering Geology and the Environment",
  ["bull eng geol environ"] = "Bulletin of Engineering Geology and the Environment",
  ["bull entomol res"] = "Bulletin of entomological research",
  ["bull entomol soc am"] = "Bulletin of the Entomological Society of America",
  ["bull entomol soc egypt"] = "Bulletin of the Entomological Society of Egypt : Economic Series",
  ["bull environ contam toxicol"] = "Bulletin of environmental contamination and toxicology",
  ["bull epidemiol hebd (paris)"] = "Bulletin epidémiologique hebdomadaire (Paris, France)",
  ["bull epizoot dis afr"] = "Bulletin of epizootic diseases of Africa. Bulletin des épizooties en Afrique",
  ["bull essex cty dent soc"] = "The Bulletin - Essex County Dental Society",
  ["bull etud orient"] = "Bulletin d'études orientales",
  ["bull etud port bres"] = "Bulletin des études portugaises",
  ["bull etud valeryennes"] = "Bulletin des études Valéryennes",
  ["bull eugen soc"] = "Bulletin (Eugenics Society (London, England))",
  ["bull eur assoc fish pathol"] = "Bulletin of the European Association of Fish Pathologists",
  ["bull eur assoc theor comput sci eatcs"] = "Bulletin of the European Association for Theoretical Computer Science. EATCS",
  ["bull eur ceram soc"] = "Bulletin of the European Ceramic Society",
  ["bull eur physiopathol respir"] = "Bulletin européen de physiopathologie respiratoire",
  ["bull exp biol med"] = "Bulletin of experimental biology and medicine",
  ["bull fac ed kagoshima univ natur sci"] = "Bulletin of the Faculty of Education. Kagoshima University. Natural Science",
  ["bull fac ed utsunomiya univ sect 2"] = "Utsunomiya University",
  ["bull fac ed wakayama univ natur sci"] = "Wakayama University. Faculty of Education. Bulletin. Natural Science",
  ["bull fac educ utsunomiya univ sect 2"] = "Bulletin of the Faculty of Education",
  ["bull fac eng hiroshima univ"] = "Bulletin of the Faculty of Engineering",
  ["bull fac gen ed gifu univ"] = "Gifu University",
  ["bull fac liberal arts chukyo univ"] = "Bulletin of the Faculty of the Liberal Arts",
  ["bull fac liberal arts nagasaki univ"] = "Bulletin of the Faculty of Liberal Arts",
  ["bull fac pharm cairo univ"] = "Bulletin of Faculty of Pharmacy, Cairo University",
  ["bull fac sch educ hiroshima univ part ii"] = "Bulletin of the Faculty of School Education Hiroshima University",
  ["bull fac school ed hiroshima univ part ii"] = "Bulletin of the Faculty of School Education",
  ["bull fac sci alexandria univ"] = "University of Alexandria",
  ["bull fac sci assiut univ a"] = "Bulletin of the Faculty of Science",
  ["bull fac sci assiut univ c"] = "Bulletin of the Faculty of Science",
  ["bull fac sci eng chuo univ ser i math"] = "Bulletin of the Faculty of Science and Engineering",
  ["bull fac sci ibaraki univ ser a"] = "Bulletin of the Faculty of Science",
  ["bull fac sci univ ryukyus"] = "Bulletin of the Faculty of Science. University of the Ryukyus",
  ["bull far east branch russ acad sci"] = "Bulletin of the Far Eastern Branch of the Russian Academy of Sciences",
  ["bull fed assoc mycol mediterr"] = "Bulletin de la Fédération des Associations Mycologiques Méditerranéennes",
  ["bull fed soc gynecol obstet lang fr"] = "Bulletin de la Fédération des sociétés de gynécologie et dóbstétrique de langue française",
  ["bull fifth dist dent soc (fresno)"] = "The Bulletin of the Fifth District Dental Society",
  ["bull fifth dist dent soc state n y"] = "Bulletin Of The Fifth District Dental Society Of The State Of New York",
  ["bull fish biol"] = "Bulletin of fish biology",
  ["bull fla mus nat hist"] = "Bulletin of the Florida Museum of Natural History",
  ["bull for comm"] = "Bulletin Forestry Commission",
  ["bull francoph afr"] = "Bulletin of francophone Africa",
  ["bull fukuoka univ ed iii"] = "Bulletin of Fukuoka University of Education",
  ["bull fulton cty med soc"] = "Bulletin - Fulton County Medical Society",
  ["bull ga acad sci"] = "Bulletin of the Georgia Academy of Science",
  ["bull gandhigram inst rural health fam welf trust"] = "Bulletin (Gandhigram Institute of Rural Health and Family Welfare Trust)",
  ["bull gastrointest endosc"] = "Bulletin of gastrointestinal endoscopy",
  ["bull geisinger"] = "Bulletin - Geisinger Medical Center",
  ["bull gen ed dokkyo univ school medicine"] = "Bulletin of General Education",
  ["bull genesee cty med soc"] = "Bulletin - Genesee County Medical Society",
  ["bull geobot inst eth"] = "Bulletin of the Geobotanical Institute ETH",
  ["bull geodesique"] = "Bulletin Geodesique",
  ["bull geol soc am"] = "Bulletin Geological Society of America",
  ["bull geol soc china"] = "Bulletin of the Geological Society of China",
  ["bull geol soc den"] = "Bulletin of the Geological Society of Denmark",
  ["bull geol soc finl"] = "Bulletin of the Geological Society of Finland",
  ["bull geol surv can"] = "Bulletin Geological Survey of Canada",
  ["bull geol surv finl"] = "Bulletin of The Geological Survey of Finland",
  ["bull georgetown univ med cent"] = "Bulletin. Georgetown University. Medical Center",
  ["bull georgian natl acad sci (ns)"] = "Bulletin of the Georgian National Academy of Sciences. New Series",
  ["bull geosci"] = "Bulletin of geosciences",
  ["bull ghana geogr assoc"] = "Bulletin of the Ghana Geographical Association",
  ["bull glaciol res"] = "Bulletin of Glaciological Research",
  ["bull gloucester cty hist soc"] = "Bulletin of the Gloucester County Historical Society",
  ["bull greek math soc"] = "Bulletin of the Greek Mathematical Society",
  ["bull group eur rech sci stomatol odontol"] = "Bulletin de Groupement européen pour la recher che scientifique en stomatologie & odontologie",
  ["bull group int rech sci stomatol"] = "Bulletin du Groupement international pour la recherche scientifique en stomatologie",
  ["bull group int rech sci stomatol odontol"] = "Bulletin du Groupèment international pour la recherche scientifique en stomatologie & odontologie",
  ["bull gulistan state univ"] = "Bulletin of Gulistan State University",
  ["bull haffkine"] = "Bulletin of Haffkine Institute",
  ["bull hellenic math soc"] = "Bulletin of the Hellenic Mathematical Society",
  ["bull hennepin cty med soc"] = "Bulletin - Hennepin County Medical Society",
  ["bull high inst public health"] = "The Bulletin of the High Institute of Public Health",
  ["bull his sci auvergne"] = "Bulletin historique et scientifique de l'Auvergne",
  ["bull hisp"] = "Bulletin hispanique",
  ["bull hist chem"] = "Bulletin for the History of Chemistry",
  ["bull hist dent"] = "Bulletin of the history of dentistry",
  ["bull hist electr"] = "Bulletin d'histoire de l'électricité",
  ["bull hist med"] = "Bulletin of the history of medicine",
  ["bull hist med suppl"] = "Supplements to the Bulletin of the history of medicine",
  ["bull hist polit"] = "Bulletin d'histoire politique",
  ["bull histol appl physiol pathol tech microsc"] = "Bulletin d'histologie appliquée à la physiologie et à la pathologie et de technique microscopique",
  ["bull histol appl tech microsc"] = "Bulletin d'histologie appliquée et de technique microscopique",
  ["bull honam math soc"] = "Honam Mathematical Society",
  ["bull hong kong math soc"] = "Bulletin of the Hong Kong Mathematical Society",
  ["bull hosp joint dis"] = "Bulletin of the Hospital for Joint Diseases",
  ["bull hosp jt dis"] = "Bulletin (Hospital for Joint Diseases (New York, N.Y.))",
  ["bull hosp jt dis  (2013)"] = "Bulletin of the Hospital for Joint Disease (2013)",
  ["bull hosp jt dis orthop inst"] = "Bulletin of the Hospital for Joint Diseases Orthopaedic Institute",
  ["bull hosp spec surg"] = "Bulletin. Hospital for Special Surgery",
  ["bull hudson cty dent soc"] = "Bulletin. Hudson County Dental Society, Jersey City",
  ["bull hunt inst bot doc"] = "Bulletin of the Hunt Institute for Botanical Documentation. Hunt Institute for Botanical Documentation",
  ["bull hyg"] = "Bulletin d'hygiène. Health bulletin. Montréal (Québec). Department of Health",
  ["bull hyg (lond)"] = "Bulletin of hygiene",
  ["bull iacfs me"] = "Bulletin of the IACFS/ME",
  ["bull ill dent hyg assoc"] = "The bulletin of the Illinois Dental Hygienist's Association. Illinois Dental Hygienists' Association",
  ["bull indian inst hist med hyderabad"] = "Bulletin of the Indian Institute of History of Medicine (Hyderabad)",
  ["bull indian vac soc"] = "Bulletin of the Indian Vacuum Society",
  ["bull indones econ stud"] = "Bulletin of Indonesian economic studies",
  ["bull indonesian econ stud"] = "Bulletin of Indonesian Economic Studies",
  ["bull inf cenaddom"] = "Bulletin d'information - Cenaddom. Centre national de documentation des départements d'outre-mer",
  ["bull inf off natl for"] = "Bulletin d’information Office national des forêts",
  ["bull inf tech"] = "Bulletin d’informations techniques",
  ["bull infirm cathol can"] = "Bulletin des infirmières catholiques du Canada",
  ["bull inform cybernet"] = "Bulletin of Informatics and Cybernetics",
  ["bull inform fed pharm mediterr"] = "Bulletin d'information. Federation pharmaceutique méditerranéenne",
  ["bull insectology"] = "Bulletin of insectology",
  ["bull inst chem res kyoto univ"] = "Bulletin of the Institute for Chemical Research, Kyoto University",
  ["bull inst class stud univ lond"] = "Bulletin. University of London. Institute of Classical Studies",
  ["bull inst combin appl"] = "Bulletin of the Institute of Combinatorics and its Applications",
  ["bull inst egypte"] = "Bulletin de l'Institut d'Egypte",
  ["bull inst fond afr noire ser b"] = "Bulletin de l'Institut fondamental d'Afrique noire. Série B: Sciences humaines",
  ["bull inst fr archeol orient"] = "Bulletin de l'Institut français d'archéologie orientale. Institut français d'archéologie orientale du Caire",
  ["bull inst hist med hyderabad"] = "Bulletin of the Institute of History of Medicine (Hyderabad)",
  ["bull inst hist res"] = "Bulletin of the Institute of Historical Research",
  ["bull inst marit trop med gdynia"] = "Bulletin of the Institute of Maritime and Tropical Medicine in Gdynia",
  ["bull inst math acad sin (ns)"] = "Bulletin of the Institute of Mathematics. Academia Sinica. New Series",
  ["bull inst math acad sinica"] = "Bulletin of the Institute of Mathematics",
  ["bull inst med res kuala lumpur"] = "Bulletin - Institute for Medical Research, Kuala Lumpur",
  ["bull inst med res univ madr"] = "Bulletin of the Institute for Medical Research, University of Madrid",
  ["bull inst min metall"] = "Bulletin Institute of Mining and Metallurgy",
  ["bull inst mod hist acad sin"] = "Zhong yang yan jiu yuan jin dai shi yan jiu suo ji kan. Zhong yang yan jiu yuan. Jin dai shi yan jiu suo",
  ["bull inst natl hyg"] = "Bulletin. Institut national d'hygiène (France)",
  ["bull inst natl sante rech med"] = "Bulletin de l'Institut national de la santé et de la recherche médicale",
  ["bull inst natl sci technol mer"] = "Bulletin de l'Institut national des sciences et technologies de la mer",
  ["bull inst oceanogr"] = "Bulletin de l'Institut océanographique. Institut océanographique",
  ["bull inst pasteur"] = "Bulletin de l’Institut Pasteur",
  ["bull int acad pol sci let cl med"] = "Bulletin international de l'Académie polonaise des sciences et des lettres. Classe de médecine",
  ["bull int assoc eng geol"] = "Bulletin of the International Association of Engineering Geology",
  ["bull int assoc med mus"] = "Bulletin. International Association of Medical Museums",
  ["bull int assoc paleondontol"] = "Bulletin Of The International Association For Paleodontology",
  ["bull int assoc sci hydrol"] = "Bulletin of the International Association of Scientific Hydrology",
  ["bull int comm urgent anthropol ethnol res"] = "Bulletin of the International Committee on Urgent Anthropological and Ethnological Research. International Committee on Urgent Anthropological and Ethnological Research",
  ["bull int dairy assoc"] = "Bulletin of the International Dairy Association",
  ["bull int fiscal doc"] = "Bulletin for International Fiscal Documentation",
  ["bull int math virtual inst"] = "Bulletin of International Mathematical Virtual Institute",
  ["bull int pediatr assoc"] = "Bulletin of the International Pediatric Association. Bulletin de l'Association internationale de pédiatrie. Boletin de la Asociacion Internacional de Pediatria",
  ["bull int serv sante armees terre mer air"] = "Bulletin international des services de santé des armées de terre, de mer et de l'air",
  ["bull int union tuberc"] = "Bulletin of the International Union against Tuberculosis",
  ["bull int union tuberc lung dis"] = "Bulletin of the International Union against Tuberculosis and Lung Disease",
  ["bull iran math soc"] = "Bulletin of the Iranian Mathematical Society",
  ["bull iranian math soc"] = "Bulletin of the Iranian Mathematical Society",
  ["bull islam med"] = "Bulletin of Islamic medicine",
  ["bull j n y state archeol assoc"] = "Bulletin (New York State Archeological Association : 1987)",
  ["bull j rylands univ libr manchester"] = "Bulletin. John Rylands University Library of Manchester",
  ["bull jackson park branch chic med soc"] = "Bulletin - Jackson Park Branch, Chicago Medical Society",
  ["bull jersey city margaret hague mat hosp"] = "Bulletin. Margaret Hague Maternity Hospital (Jersey City, N.J.)",
  ["bull john rylands univ lib"] = "Bulletin. John Rylands Library, Manchester",
  ["bull johns hopkins hosp"] = "Bulletin of the Johns Hopkins Hospital",
  ["bull jpn soc coord chem"] = "Bulletin of Japan Society of Coordination Chemistry",
  ["bull kanagawa dent coll"] = "The Bulletin of the Kanagawa Dental College : BKDC",
  ["bull karaganda univ math ser"] = "Bulletin of the Karaganda University. Mathematics Series",
  ["bull kerala math assoc"] = "Bulletin of Kerala Mathematics Association",
  ["bull king cty med soc"] = "Bulletin - King County Medical Society",
  ["bull korean chem soc"] = "Bulletin of the Korean Chemical Society",
  ["bull korean math soc"] = "Bulletin of the Korean Mathematical Society",
  ["bull koshien univ b"] = "Bulletin of Koshien University",
  ["bull kresge eye inst"] = "Bulletin. Kresge Eye Institute",
  ["bull kyoto univ ed ser b"] = "Bulletin of the Kyoto University of Education",
  ["bull kyushu inst tech math natur sci"] = "Bulletin of the Kyushu Institute of Technology",
  ["bull kyushu inst technol pure appl math"] = "Bulletin of the Kyushu Institute of Technology. Pure and Applied Mathematics",
  ["bull lat am res"] = "Bulletin of Latin American research",
  ["bull leag red cross soc"] = "Bulletin. League of Red Cross Societies",
  ["bull leo baeck inst"] = "Bulletin des Leo Baeck Instituts. Leo Baeck Institute",
  ["bull liaison demogr afr"] = "Bulletin de liaison de démographie africaine",
  ["bull linn cty iowa"] = "Bulletin. Linn County (Iowa) Medical Society",
  ["bull lond math soc"] = "The bulletin of the London Mathematical Society",
  ["bull london math soc"] = "Bulletin of the London Mathematical Society",
  ["bull los angel cty med assoc"] = "Bulletin - Los Angeles County Medical Association",
  ["bull los angel neuro soc"] = "Bulletin of the Los Angeles Neurological Society",
  ["bull los angel surg soc"] = "Bulletin. Los Angeles Surgical Society",
  ["bull los angeles dent soc"] = "Bulletin of the Los Angeles Dental Society",
  ["bull los angeles neurol soc"] = "Bulletin of the Los Angeles neurological societies",
  ["bull louis a weiss mem hosp"] = "The Bulletin of Louis A. Weiss Memorial Hospital",
  ["bull mahoning cty med soc"] = "Bulletin. Mahoning County Medical Society",
  ["bull malays math sci soc"] = "Bulletin of the Malaysian Mathematical Sciences Society",
  ["bull malaysian math soc (2)"] = "Malaysian Mathematical Society",
  ["bull mar sci"] = "Bulletin of Marine Science",
  ["bull mason clin"] = "Bulletin of the Mason Clinic",
  ["bull mass dent soc"] = "Bulletin. Massachusetts Dental Society",
  ["bull mass nurses assoc"] = "Bulletin of the Massachusetts Nurses Association",
  ["bull mater sci"] = "Bulletin of Materials Science",
  ["bull math anal appl"] = "Bulletin of Mathematical Analysis and Applications",
  ["bull math assoc india"] = "Bulletin Mathematical Association of India",
  ["bull math biol"] = "Bulletin of Mathematical Biology",
  ["bull math biophys"] = "The Bulletin of mathematical biophysics",
  ["bull math sci"] = "Bulletin of Mathematical Sciences",
  ["bull math soc sci math roumanie (ns)"] = "Bulletin Mathematique de la Societe des Sciences Mathematiques de Roumanie. Nouvelle Serie",
  ["bull mcguire clin st lukes hosp"] = "Bulletin, McGuire Clinic and St. Luke's Hospital. St. Luke's Hospital (Richmond, Va.)",
  ["bull mcguire clin st lukes hosp richmond st lukes hosp mcguire clin"] = "Bulletin, McGuire Clinic and St. Luke’s Hospital. Richmond. St. Luke’s Hospital. McGuire Clinic",
  ["bull med"] = "Bulletin médical",
  ["bull med ethics"] = "Bulletin of medical ethics",
  ["bull med leg toxicol"] = "Bulletin médecine légale, toxicologie",
  ["bull med leg urgence med cent antipoisons"] = "Bulletin de médecine légale, urgence médicale, centre anti-poisons",
  ["bull med libr assoc"] = "Bulletin of the Medical Library Association",
  ["bull med rev med fr guide prat"] = "Bulletin médical. Revue médicale française. Guide du praticien",
  ["bull med sci hist soc"] = "Bulletin (Medical Sciences Historical Society (Great Britain))",
  ["bull med soc cty erie buffalo acad med"] = "Bulletin [of the] Medical Society, County of Erie [and] the Buffalo Academy of Medicine. Medical Society of the County of Erie (N.Y.)",
  ["bull med soc cty kings kings cty n y"] = "Bulletin. Medical Society of the County of Kings (Kings County, N.Y.)",
  ["bull med soc cty monroe"] = "Bulletin. Medical Society of the County of Monroe",
  ["bull med univ aurore"] = "Bulletin médical. Shanghai. Université l'Aurore",
  ["bull mem acad r med belg"] = "Bulletin et mémoires de l'Académie royale de médecine de Belgique",
  ["bull mem chir marseille"] = "Bulletins et mémoires. Société de chirurgie de Marseille",
  ["bull mem fac mixte med pharm dakar"] = "Bulletins et mémoires de la Faculté mixte de médecine et de pharmacie de Dakar",
  ["bull mem soc anthropol paris"] = "Bulletins et mémoires de la Société d'anthropologie de Paris",
  ["bull mem soc chir paris"] = "Bulletin et mémoires de la Société des chirurgiens de Paris",
  ["bull mem soc fr ophtalmol"] = "Bulletins et mémoires de la Société française d'ophtalmologie",
  ["bull mem soc med hop paris"] = "Bulletins et mémoires de la Société médicale des hôpitaux de Paris",
  ["bull mem soc med paris (1907)"] = "Bulletin et mémoires. Société de médecine de Paris",
  ["bull mem soc med passy ht-savoie"] = "Bulletin et mémoires de la Société médicale de Passy (Haute-Savoie)",
  ["bull menninger clin"] = "Bulletin of the Menninger Clinic",
  ["bull mens off int hyg publique"] = "Bulletin mensuel. Office international d'hygiène publique",
  ["bull mens soc linn lyon"] = "Bulletin mensuel de la Société linnéenne de Lyon",
  ["bull mens soc med mil fr"] = "Bulletin mensuel - Société de médecine militaire française",
  ["bull mens soc vet prat fr"] = "Bulletin mensuel de la Société vétérinaire pratique de France",
  ["bull meriden conn hosp"] = "Bulletin. Meriden, Conn. Hospital",
  ["bull methodol sociol"] = "Bulletin de méthodologie sociologique : BMS",
  ["bull mich dent hyg assoc"] = "The Bulletin of the Michigan Dental Hygienists' Association",
  ["bull mich state dent hyg assoc"] = "The Bulletin of the Michigan State Dental Hygienists' Association",
  ["bull micr appl"] = "Bulletin de microscopie appliquée",
  ["bull midtown dent soc"] = "Bulletin / Midtown Dental Society",
  ["bull mil clin psychol"] = "Bulletin of military clinical psychologists",
  ["bull millard fillmore hosp buffalo"] = "Bulletin. Buffalo (N.Y.). Millard Fillmore Hospital",
  ["bull minn med found"] = "Bulletin. Minnesota Medical Foundation",
  ["bull miss hist soc"] = "Bulletin of the Missouri Historical Society. Missouri Historical Society",
  ["bull mo hist soc"] = "The bulletin - Missouri Historical Society. Missouri Historical Society",
  ["bull mo soc crippl child"] = "Bulletin. Missouri Society for Crippled Children",
  ["bull mol biol med"] = "Bulletin of molecular biology and medicine",
  ["bull monmouth ocean cty dent soc"] = "Bulletin / Monmouth-Ocean County Dental Society",
  ["bull montg bucks dent soc"] = "Bulletin. Montgomery-Bucks Dental Society",
  ["bull moore white med found los angel"] = "Bulletin. Moore-White Medical Foundation, Los Angeles",
  ["bull mt desert isl biol lab salisb cove maine"] = "Bulletin - Mount Desert Island Biological Laboratory, Salisbury Cove, Maine",
  ["bull multnomah cty"] = "Bulletin. Multnomah County (Or.) Medical Society",
  ["bull munic off ville mars"] = "Bulletin municipal officiel de la ville de Marseille",
  ["bull murithienne"] = "Bulletin de la Murithienne",
  ["bull mus comp zool"] = "Bulletin of the Museum of Comparative Zoology at Harvard College",
  ["bull mus hong b arts"] = "Bulletin du Musée hongrois des beaux-arts = O.M. Szépművészeti Múzeum közleményei",
  ["bull mus rov beaux arts belg"] = "Bulletin - Musées royaux des beaux-arts de Belgique. Musées royaux des beaux-arts de Belgique",
  ["bull mycol bot dauphine savoie"] = "Bulletin mycologique et botanique Dauphiné-Savoie",
  ["bull n c dent soc"] = "Bulletin. North Carolina Dental Society",
  ["bull n j coll med dent"] = "Bulletin Of The New Jersey College Of Medicine And Dentistry",
  ["bull n j soc dent child"] = "Bulletin - New Jersey Society of Dentistry for Children",
  ["bull n y acad med"] = "Bulletin of the New York Academy of Medicine",
  ["bull n y public libr"] = "Bulletin of the New York Public Library, Astor, Lenox, and Tilden Foundations. New York Public Library",
  ["bull n y state dent soc anesthesiol"] = "Bulletin Of The New York State Dental Society Of Anesthesiology",
  ["bull n y state soc dent child"] = "Bulletin of the New York State Society of Dentistry for Children",
  ["bull n z soc periodontol"] = "Bulletin - N. Z. Society of Periodontology",
  ["bull nagoya inst tech"] = "Bulletin of Nagoya Institute of Technology",
  ["bull nara univ ed natur sci"] = "Bulletin of Nara University of Education. Natural Science",
  ["bull narc"] = "Bulletin on narcotics",
  ["bull nat fam plan counc vic"] = "Bulletin of the Natural Family Planning Council of Victoria",
  ["bull natl clgh poison control cent"] = "Bulletin (National Clearinghouse for Poison Control Centers (U.S.))",
  ["bull natl guild cathol psychiatr"] = "Bulletin Of The National Guild Of Catholic Psychiatrists",
  ["bull natl inst health"] = "Bulletin. National Institutes of Health (U.S.)",
  ["bull natl med dent assoc natl advocates soc"] = "Bulletin (National Medical and Dental Association (U.S))",
  ["bull natl mus nat sci ser a zool"] = "Bulletin of the National Museum of Nature and Science. Series A, Zoology",
  ["bull natl mus nat sci ser b bot"] = "Bulletin of the National Museum of Nature and Science. Series B, Botany",
  ["bull natl res cent"] = "Bulletin of the National Research Centre",
  ["bull natl sci mus ser b bot"] = "Bulletin of the National Science Museum. Series B, Botany",
  ["bull natl soc med res"] = "Bulletin of the National Society for Medical Research",
  ["bull natl tuberc assoc"] = "Bulletin - National Tuberculosis Association",
  ["bull natl tuberc respir dis assoc"] = "Bulletin - National Tuberculosis and Respiratory Disease Association",
  ["bull new engl med cent"] = "Bulletin. New England Medical Center",
  ["bull new ser am math soc"] = "Bulletin (new series) of the American Mathematical Society",
  ["bull newark dent club"] = "Bulletin of the Newark Dental Club",
  ["bull ninth dist dent soc"] = "Bulletin. Ninth District Dental Society of the State of New York",
  ["bull nor geol unders"] = "Bulletin Norges Geologiske Undersoekels",
  ["bull north dist dent soc"] = "Bulletin - Northern District Dental Society",
  ["bull north shore branch chic med soc"] = "Bulletin - North Shore Branch, Chicago Medical Society",
  ["bull nyu hosp jt dis"] = "Bulletin of the NYU hospital for joint diseases",
  ["bull oak park ill"] = "Bulletin. Oak Park, Ill. West Suburban Hospital",
  ["bull oepp"] = "Bulletin OEPP. EPPO bulletin. European and Mediterranean Plant Protection Organisation",
  ["bull off chambre synd med seine"] = "Bulletin officiel. Chambre syndicale des médecins de la Seine",
  ["bull off int epizoot"] = "Bulletin - Office international des épizooties",
  ["bull okla state dent assoc"] = "Bulletin. Oklahoma State Dental Association",
  ["bull ophthalmol soc egypt"] = "Bulletin of the Ophthalmological Society of Egypt",
  ["bull orange cty med assoc"] = "The bulletin of the Orange County Medical Association",
  ["bull ordre med"] = "Bulletin de l'Ordre des médecins",
  ["bull osaka med coll"] = "Bulletin of the Osaka Medical College",
  ["bull osaka med sch"] = "Bulletin of the Osaka Medical School",
  ["bull osaka med sch suppl"] = "Bulletin of the Osaka Medical School. Supplement",
  ["bull osaka prefect college tech"] = "Bulletin of Osaka Prefectural College of Technology",
  ["bull osaka prefect univ coll technol"] = "Bulletin of Osaka Prefectural University College of Technology",
  ["bull osaka prefect univ ser a eng nat sci"] = "Bulletin of Osaka Prefecture University",
  ["bull osaka prefect univ ser b agric life sci"] = "Bulletin of Osaka Prefecture University. Series B, Agriculture and life sciences",
  ["bull oxford univ inst econ statist"] = "Bulletin of the Oxford University Institute of Economics and Statistics",
  ["bull pac coast soc orthod"] = "Bulletin - Pacific Coast Society of Orthodontists",
  ["bull pan am health organ"] = "Bulletin of the Pan American Health Organization",
  ["bull parenter drug assoc"] = "Bulletin of the Parenteral Drug Association",
  ["bull park ridge cent"] = "Bulletin of the Park Ridge Center",
  ["bull passaic cty dent soc"] = "Bulletin of the Passaic County Dental Society",
  ["bull passaic cty med soc"] = "The Bulletin of the Passaic County Medical Society",
  ["bull peabody mus nat hist"] = "Bulletin (Peabody Museum of Natural History)",
  ["bull peabody mus nat hist, yale univ"] = "Bulletin Peabody Museum of Natural History, Yale University",
  ["bull peace propos"] = "Bulletin of peace proposals",
  ["bull pharm res inst"] = "Bulletin of Pharmaceutical Research Institute",
  ["bull phila assoc psychoanal"] = "Bulletin of the Philadelphia Association for Psychoanalysis",
  ["bull phila cty dent soc"] = "The Bulletin of the Philadelphia County Dental Society",
  ["bull phila pa hosp ayer clin lab"] = "Bulletin. Pennsylvania Hospital (Philadelphia, Pa.). Ayer Clinical Laboratory",
  ["bull philol hist"] = "Bulletin philologique et historique (jusqu'à 1610) du Comité des travaux historiques et scientifiques",
  ["bull philos mediev"] = "Bulletin de philosophie médiévale",
  ["bull physiopathol respir (nancy)"] = "Bulletin de physio-pathologie respiratoire",
  ["bull plainfield dent soc"] = "Bulletin Of The Plainfield Dental Society",
  ["bull pol acad sci chem"] = "Bulletin of the Polish Academy of Sciences. Chemistry",
  ["bull pol acad sci math"] = "Bulletin of the Polish Academy of Sciences. Mathematics",
  ["bull pol acad sci tech sci"] = "Bulletin of the Polish Academy of Sciences Technical Sciences",
  ["bull pol acad sci, chem"] = "Bulletin of the Polish Academy of Sciences Chemistry",
  ["bull pol med sci hist"] = "Bulletin of Polish Medical Science and History",
  ["bull polish acad sci math"] = "Polish Academy of Sciences",
  ["bull popul dev stud cent"] = "Bulletin of the Population and Development Studies Center. Sŏul Taehakkyo. Inʼgu mit Palchŏn Munje Yŏnʼguso",
  ["bull postgrad comm med univ syd"] = "Bulletin of the Post-Graduate Committee in Medicine, University of Sydney",
  ["bull postgrad inst med educ res chandigarh"] = "Bulletin of Postgraduate Institute of Medical Education and Research, Chandigarh",
  ["bull pottawatomie cty okla"] = "Bulletin. Pottawatomie County (Okla.) Medical Society",
  ["bull pract ophthalmol"] = "Bulletin of practical ophthalmology",
  ["bull presse sante fr"] = "Bulletin de presse - Santé de la France",
  ["bull prosthet res"] = "Bulletin of prosthetics research",
  ["bull psychon soc"] = "Bulletin of the Psychonomic Society",
  ["bull publ health soc (kuala lumpur)"] = "Bulletin of the Public Health Society",
  ["bull public health"] = "Bulletins of the public health. United States. Marine Hospital Service",
  ["bull pure appl math"] = "Bulletin of Pure and Applied Mathematics",
  ["bull pure appl sci sect e math stat"] = "Bulletin of Pure & Applied Sciences",
  ["bull r inst inter faith stud"] = "Bulletin of the Royal Institute for Inter-Faith Studies",
  ["bull rech agron gembloux"] = "Bulletin des recherches agronomiques de Gembloux",
  ["bull repr"] = "Bulletin of reprints",
  ["bull res counc isr"] = "Bulletin of the Research Council of Israel",
  ["bull res counc isr sect e exp med"] = "Bulletin of the Research Council of Israel. Section E: Experimental medicine",
  ["bull res humanit"] = "Bulletin of research in the humanities",
  ["bull rheum dis"] = "Bulletin on the rheumatic diseases",
  ["bull romand entomol"] = "Bulletin romand d’entomologie",
  ["bull rosl n y st francis sanat card child"] = "Bulletin. Roslyn, N. Y. St. Francis Sanatorium for Cardiac Children",
  ["bull roswell park meml inst"] = "Bulletin. Roswell Park Memorial Institute",
  ["bull russ acad sci phys"] = "Bulletin of the Russian Academy of Sciences. Physics",
  ["bull russ acad sci: phys"] = "Bulletin of the Russian Academy of Sciences: Physics",
  ["bull s diego cty med soc"] = "Bulletin - San Diego County Medical Society",
  ["bull san diego cty dent soc"] = "Bulletin - San Diego County Dental Society",
  ["bull san franc cty med soc"] = "Bulletin - San Francisco County Medical Society",
  ["bull san mateo cty dent soc"] = "Bulletin - San Mateo County Dental Society",
  ["bull sangamon cty med soc"] = "Bulletin - Sangamon County Medical Society",
  ["bull sanit"] = "Bulletin sanitaire",
  ["bull sante publique (brux )"] = "Bulletin de la santé publique. Belgium. Ministère de la santé publique et de la famille",
  ["bull sch med"] = "Bulletin Of The School Of Medicine",
  ["bull sch med univ md"] = "Bulletin of the School of Medicine (Baltimore, Md.)",
  ["bull sch orient afr stud"] = "Bulletin of the School of Oriental and African Studies. University of London. School of Oriental and African Studies",
  ["bull schweiz akad med wiss"] = "Bulletin der Schweizerischen Akademie der Medizinischen Wissenschaften",
  ["bull sci eng res lab waseda univ"] = "Bulletin of Science and Engineering Research Laboratory",
  ["bull sci instrum soc"] = "Bulletin of the Scientific Instrument Society",
  ["bull sci issue"] = "Bulletin. Scientific issue. Huron Road Hospital",
  ["bull sci math"] = "Bulletin des Sciences Mathematiques",
  ["bull sci med (bologna)"] = "Bullettino delle scienze mediche",
  ["bull sci ser"] = "Bulletin’s Science Series",
  ["bull sci technol soc"] = "Bulletin of science, technology & society",
  ["bull sci, technol soc"] = "Bulletin of Science, Technology & Society",
  ["bull seances acad r sci outre mer"] = "Bulletin des séances- Académie royale des sciences d'outre-mer",
  ["bull second dist dent soc"] = "Bulletin - Second District Dental Society",
  ["bull sect hist mod contemp"] = "Bulletin de la Section d'histoire moderne et contemporaine. France. Comité des travaux historiques et scientifiques. Section d'histoire moderne et contemporaine",
  ["bull sect logic univ lodz"] = "University of Lodz. Department of Logic. Bulletin of the Section of Logic",
  ["bull sect logic univ łódź"] = "University of Łódź",
  ["bull seismol soc am"] = "Bulletin of the Seismological Society of America",
  ["bull seismol soc amer"] = "Bulletin of the Seismological Society of America",
  ["bull sib otd ross akad med nauk"] = "Bi︠u︡lletenʹ Sibirskogo otdelenii︠a︡ Rossiĭskoĭ akademii medit︠s︡inskikh nauk",
  ["bull sinai hosp detroit"] = "Bulletin - Sinai Hospital of Detroit",
  ["bull sloane hosp women columbia presbyt med"] = "Bulletin of the Sloane Hospital for Women in the Columbia-Presbyterian Medical Center, New York City",
  ["bull soc agric sci arts sarthe"] = "Bulletin de la Société d'agriculture, sciences et arts de la Sarthe. Société d'agriculture, sciences et arts de la Sarthe",
  ["bull soc amis sci (med) (poznan)"] = "Bulletin de la Société des amis des sciences et des lettres de Poznań. Séria C: Médecine",
  ["bull soc amis vieux toulon"] = "Bulletin de la Société des amis du vieux Toulon",
  ["bull soc archeol finistere"] = "Bulletin de la Société archéologique du Finistère. Société archéologique du Finistère",
  ["bull soc archeol hist artist vieux pap"] = "Bulletin de la société archéologique, historique et artistique Le vieux papier pour l'étude de la vie et des mœurs d'autrefois",
  ["bull soc archeol hist limousin"] = "Bulletin de la Société archéologique et historique du Limousin",
  ["bull soc archeol hist litt sci gers"] = "Bulletin de la Société archéologique, historique, littéraire & scientifique du Gers. Société archéologique, historique, littéraire et scientifique du Gers",
  ["bull soc belg etud geogr"] = "Bulletin de la Société belge d'études géographiques. Tijdschrift van de Belgische Vereniging voor Aardrijkskundige studies. Société belge d'études géographiques",
  ["bull soc belge etud napoleon"] = "Bulletin (Société belge d'études Napoléoniennes)",
  ["bull soc belge ophtalmol"] = "Bulletin de la Société belge d'ophtalmologie",
  ["bull soc borda"] = "Bulletin de la Société de Borda (Dax, France : 1876)",
  ["bull soc bot fr"] = "Bulletin de la Société Botanique de France",
  ["bull soc chim belg"] = "Bulletin des Societes Chimiques Belges",
  ["bull soc chim biol (paris)"] = "Bulletin de la Société de chimie biologique",
  ["bull soc chim fr"] = "Bulletin de la Societe Chimique de France",
  ["bull soc chir paris"] = "Bulletin. Société de chirurgie de Paris",
  ["bull soc entomol fr"] = "Bulletin de la Société entomologique de France",
  ["bull soc entomol france"] = "Bulletin de la Société entomologique de France",
  ["bull soc entomol mulhouse"] = "Bulletin de la Société entomologique de Mulhouse",
  ["bull soc entomol suisse"] = "Bulletin de la société entomologique Suisse",
  ["bull soc etud his alpes"] = "Bulletin de la Société d'études des Hautes-Alpes. Société d'études des Hautes-Alpes",
  ["bull soc etud oceaniennes"] = "Bulletin de la Société des études océaniennes (Polynésie orientale)",
  ["bull soc fr dermatol syphiligr"] = "Bulletin de la Société française de dermatologie et de syphiligraphie",
  ["bull soc fr hist hop"] = "Bulletin - Société française d'histoire des hôpitaux",
  ["bull soc fr photogramm télédétect"] = "Bulletin de la Société française de photogrammétrie et télédétection",
  ["bull soc fr phys"] = "Bulletin de la Societe Francaise de Physique",
  ["bull soc fribg sci nat"] = "Bulletin de la société fribourgeoise des sciences naturelles",
  ["bull soc geol fr"] = "Bulletin de la Societe Geologique de France",
  ["bull soc géol fr"] = "Bulletin de la Société Géologique de France",
  ["bull soc hist archeol perigord"] = "Bulletin de la Société historique et archéologique du Périgord. Société historique et archéologique du Périgord",
  ["bull soc hist mod"] = "Bulletin de la Société d'histoire moderne. Société d'histoire moderne (Paris, France)",
  ["bull soc hist protestant fr"] = "Bulletin de la Société de l'histoire du protestantisme français (1852)",
  ["bull soc hist sci deux sevres"] = "Bulletin de la Société historique et scientifique des Deux-Sévres",
  ["bull soc ind mulhouse"] = "Bulletin de la Société industrielle de Mulhouse",
  ["bull soc int chir"] = "Bulletin de la Société internationale de chirurgie",
  ["bull soc int his med"] = "Bulletin et mémoires. International Society of the History of Medicine",
  ["bull soc ital med ig trop sez eritrea"] = "Bollettino. Societa italiana di medicina e igiene tropicale. Sezione Eritrea",
  ["bull soc languedoc geogr"] = "Bulletin de la Société languedocienne de géographie",
  ["bull soc lett sci arts correze"] = "Bulletin de la Société des lettres, sciences et arts de la Corrèze. Société des lettres, sciences et arts de la Corrèze",
  ["bull soc liban hist med"] = "Bulletin de la Société libanaise d'histoire de la médecine",
  ["bull soc linn provence"] = "Bulletin de la societé linnéenne de Provence",
  ["bull soc math banja luka"] = "Bulletin of Society of Mathematicians",
  ["bull soc math fr"] = "Bulletin de la Societe Mathematique de France",
  ["bull soc math france"] = "Bulletin de la Societe Mathematique de France",
  ["bull soc math france suppl mem"] = "Bulletin de la Societe Mathematique de France. Supplement Memoire",
  ["bull soc med afr noire lang fr"] = "Bulletin de la Société médicale d'Afrique noire de langue française",
  ["bull soc med history chic"] = "Bulletin. Society of Medical History of Chicago",
  ["bull soc med hyg alger"] = "Bulletin. Société des médecins hygiénistes d'Algérie",
  ["bull soc med pau"] = "Bulletin de la Société médicale de Pau",
  ["bull soc mycol fr"] = "Bulletin de la Société mycologique de France",
  ["bull soc nat antiq fr"] = "Bulletin de la Société nationale des antiquaires de France. Société nationale des antiquaires de France",
  ["bull soc nat luxemb"] = "Bulletin (Société des naturalistes luxembourgeois)",
  ["bull soc neuchâtel sci nat"] = "Bulletin de la société neuchâteloise des sciences naturelles",
  ["bull soc ophtalmol fr"] = "Bulletin des sociétés d'ophtalmologie de France",
  ["bull soc ophtalmol paris soc ophtalmol est lyon ouest"] = "Bulletin de la Société d'ophtalmologie de Paris et des sociétés d'ophtalmologie de l'Est, de Lyon et de l'Ouest. Société d'ophtalmologie de Paris",
  ["bull soc pathol exot"] = "Bulletin de la Société de pathologie exotique (1990)",
  ["bull soc pathol exot filiales"] = "Bulletin de la Société de pathologie exotique et de ses filiales",
  ["bull soc pharm bord"] = "Bulletin de la Société de pharmacie de Bordeaux",
  ["bull soc pharm strasb"] = "Bulletin de la Société de pharmacie de Strasbourg",
  ["bull soc r belge gynecol obstet"] = "Bulletin de la Société royale belge de gynécologie et d'obstétrique",
  ["bull soc r for belg"] = "Bulletin de la Société royale forestière de Belgique",
  ["bull soc ramond"] = "Explorations pyrénéennes : bulletin trimestriel de la Société Ramond",
  ["bull soc roy sci liège"] = "Bulletin de la Société Royale des Sciences de Liège",
  ["bull soc sci hyg aliment aliment ration homme"] = "Bulletin. Société scientifique d'hygiène alimentaire et d'alimentation rationnelle",
  ["bull soc sci lett lodz ser rech deform"] = "Bulletin de la Societe des Sciences et des Lettres de Lodz. Serie: Recherches sur les Deformations",
  ["bull soc sci lett łódź sér rech déform"] = "Bulletin de la Société des Sciences et des Lettres de Łódź",
  ["bull soc sci lettr arts pau"] = "Bulletin. Société des sciences, lettres et arts de Pau",
  ["bull soc sci med grand duche luxemb"] = "Bulletin de la Société des sciences médicales du Grand-Duché de Luxembourg",
  ["bull soc sci nat ouest fr"] = "Bulletin de la Société des sciences naturelles de l'Ouest de la France (1983)",
  ["bull soc univ cartogr"] = "The bulletin of the Society of University Cartographers. Society of University Cartographers",
  ["bull soc vac coaters"] = "Bulletin - Society of Vacuum Coaters",
  ["bull soc vaud sci nat"] = "Bulletin de la Société Vaudoise des sciences naturelles",
  ["bull south calif acad sci"] = "Bulletin (Southern California Academy of Sciences)",
  ["bull spokane cty med soc"] = "Bulletin - Spokane County Medical Society",
  ["bull sri lanka coll microbiol"] = "The bulletin of the Sri Lanka College of microbiologists",
  ["bull st louis dent soc"] = "Bulletin - St. Louis Dental Society",
  ["bull st louis med society"] = "Bulletin. St. Louis Medical Society",
  ["bull st louis park med cen"] = "Bulletin (St. Louis Park Medical Center)",
  ["bull stat inst natl stat"] = "Bulletin De Statistique / Belgium Institut National De Statistique",
  ["bull statec"] = "Bulletin du STATEC. Luxembourg. Service central de la statistique et des études économiques",
  ["bull suffolk cty dent soc"] = "Bulletin of the Suffolk County Dental Society",
  ["bull suffolk cty med soc suffolk acad med"] = "Bulletin - Suffolk County Medical Society",
  ["bull suisse mycol"] = "Bulletin Suisse de Mycologie",
  ["bull switz"] = "Bulletin. Switzerland. Eidgenössisches Gesundheitsamt",
  ["bull symb log"] = "Bulletin of Symbolic Logic",
  ["bull symbolic logic"] = "The Bulletin of Symbolic Logic",
  ["bull tarrant cty med soc"] = "Bulletin. Tarrant County Medical Society",
  ["bull tech comm data eng"] = "Bulletin of the Technical Committee on Data Engineering",
  ["bull tech suisse romande"] = "Bulletin technique de la Suisse romande",
  ["bull tenn nurses assoc"] = "Bulletin - Tennessee Nurses Association",
  ["bull tenth dist dent soc (rockville centre)"] = "Bulletin of the Tenth District Dental Society of the State of New York",
  ["bull tex nurses assoc"] = "Bulletin - Texas Nurses Association",
  ["bull ticmi"] = "Bulletin. Tbilisi International Centre of Mathematics and Informatics",
  ["bull tokyo dent coll"] = "The Bulletin of Tokyo Dental College",
  ["bull tokyo gakugei univ (4)"] = "Bulletin of Tokyo Gakugei University",
  ["bull tokyo gakugei univ nat sci"] = "Bulletin of Tokyo Gakugei University. Natural Sciences",
  ["bull tokyo gakugei univ sect iv math nat sci"] = "Bulletin of Tokyo Gakugei University",
  ["bull tokyo med dent univ"] = "The Bulletin of Tokyo Medical and Dental University",
  ["bull toledo dent soc"] = "The Bulletin of the Toledo Dental Society",
  ["bull tor east med assoc"] = "Bulletin of the Toronto East Medical Association",
  ["bull train"] = "Bulletin on training, trends, techniques, topics",
  ["bull transilv univ brasov ser iii"] = "Bulletin of the Transilvania University of Brasov. Series III. Mathematics, Informatics, Physics",
  ["bull transilv univ brasov ser iii math comput sci"] = "Bulletin of the Transilvania University of Brasov. Series III. Mathematics and Computer Science",
  ["bull transilv univ brașov ser a (ns)"] = "Transilvania University of Brașov",
  ["bull transilv univ brașov ser b (ns)"] = "Transilvania University of Brașov",
  ["bull tri cty dent soc"] = "The bulletin of the Tri-County Dental Society",
  ["bull trimest credit communal belg"] = "Bulletin trimestriel - Crédit communal de Belgique",
  ["bull trimest inst actuaires francais"] = "Bulletin trimestriel de l'Institut des actuaires français. Institut des actuaires français, Paris",
  ["bull trimest plan fam"] = "<Le > Bulletin trimestriel du planning familial",
  ["bull trimest soc mycol fr"] = "Bulletin trimestriel de la Société mycologique de France",
  ["bull tufts n engl med cent"] = "Bulletin. Tufts-New England Medical Center",
  ["bull tulane univ med fac"] = "The Bulletin of the Tulane Medical Faculty",
  ["bull u s army med dep"] = "Bulletin of the U.S. Army Medical Department. United States. Army. Medical Department",
  ["bull unesco reg off educ asia pac"] = "Bulletin of the Unesco Regional Office for Education in Asia and the Pacific",
  ["bull union cty dent soc"] = "Bulletin. Union County (N. J.) Dental Society",
  ["bull univ agric sci vet med cluj napoca"] = "Bulletin of University of Agricultural Sciences and Veterinary Medicine Cluj-Napoca",
  ["bull univ electro-comm"] = "Bulletin of the University of Electro-Communications",
  ["bull univ hosp"] = "Bulletin. Augusta, Ga. University Hospital",
  ["bull univ miami sch med jackson meml hosp"] = "The bulletin [of] the University of Miami School of Medicine and Jackson Memorial Hospital. University of Miami. School of Medicine",
  ["bull univ minn hosp minn med found"] = "Bulletin of the University of Minnesota Hospitals and Minnesota Medical Foundation. University of Minnesota. Hospitals",
  ["bull univ nebr state mus"] = "Bulletin of the University of Nebraska State Museum",
  ["bull univ osaka prefect"] = "Bulletin of University of Osaka Prefecture",
  ["bull univ teach educ fukuoka iii"] = "Bulletin of University of Teacher Education Fukuoka. Part III. Mathematics, Natural Sciences and Technology",
  ["bull utica acad med"] = "Bulletin - Utica Academy of Medicine",
  ["bull va state dent assoc"] = "The Bulletin of the Virginia State Dental Association",
  ["bull val dent soc"] = "Bulletin of the Valley Dental Society",
  ["bull vanc med assoc"] = "Bulletin. Vancouver Medical Association",
  ["bull vener dis"] = "Bulletin of venereal diseases",
  ["bull vener dis control"] = "Bulletin of venereal disease control. Ontario. Department of Health",
  ["bull vet inst pulawy"] = "Bulletin of the Veterinary Institute in Puławy = Biuletyn Instytutu Weterynarii w Puławach",
  ["bull vijnana parishad india"] = "Vijnana Parishad of India",
  ["bull volcanol"] = "Bulletin of Volcanology",
  ["bull vost sib naucn sent"] = "Bi︠u︡lletenʹ Vostochno-Sibirskogo nauchnogo t︠s︡entra",
  ["bull vulg for"] = "Bulletin de vulgarisation forestière",
  ["bull war med"] = "Bulletin of war medicine",
  ["bull world health org suppl"] = "Bulletin of the World Health Organization. Supplement",
  ["bull world health organ"] = "Bulletin of the World Health Organization",
  ["bull world med assoc"] = "Bulletin. World Medical Association",
  ["bull yamagata univ natur sci"] = "Bulletin of the Yamagata University",
  ["bull yamaguchi med sch"] = "The Bulletin of the Yamaguchi Medical School",
  ["bull zool nomencl"] = "The Bulletin of zoological nomenclature",
  ["bull études orient"] = "Bulletin d’études Orientales",
  ["bulletin (ala mus nat hist)"] = "Bulletin (Alabama Museum of Natural History)",
  ["bulletin ny med coll"] = "Bulletin. New York Medical College",
  ["bulletin of the academy of military medical sciences|chun shih i hsueh ko hsueh yuan yuan kan"] = "Jun shi yi xue ke xue yuan yuan kan",
  ["bulletin of the georgian academy of sciences"] = "Bull",
  ["bulletin oilb/srop|iobc/wprs bull"] = "IOBC/WPRS bulletin",
  ["bumed news lett"] = "BuMed news letter",
  ["bumed news lett aviat suppl"] = "BuMed news letter. Aviation supplement",
  ["bundesgesundheitsbl gesundheitsforsch gesundheitsschutz"] = "Bundesgesundheitsblatt, Gesundheitsforschung, Gesundheitsschutz",
  ["bundesgesundheitsblatt gesundheitsforschung gesundheitsschutz"] = "Bundesgesundheitsblatt, Gesundheitsforschung, Gesundheitsschutz",
  ["bunseki kagaku"] = "Bunseki Kagaku",
  ["bur justice stat spec rep"] = "Bureau of Justice Statistics special report",
  ["bur mines inf circ"] = "Bureau of Mines Information Circular",
  ["bur mines rep invest"] = "Bureau of Mines Report of Investigations",
  ["burg monogr in sci"] = "Burg Monographs in Science",
  ["burgenl heimatbl"] = "Burgenländische Heimatblätter",
  ["burlingt mag"] = "The Burlington magazine",
  ["burma med j"] = "Burma medical journal",
  ["burn res"] = "Burnout research",
  ["burnout res"] = "Burnout Research",
  ["burns incl therm inj"] = "Burns, including thermal injury",
  ["burns open"] = "Burns open : an international open access journal for burn injuries",
  ["burns trauma"] = "Burns & trauma",
  ["bus econ"] = "Business economics (Cleveland, Ohio)",
  ["bus econ hist"] = "Business and Economic History",
  ["bus econ j"] = "Business and economics journal",
  ["bus econ rev"] = "Business and economic review (Columbia, S.C.)",
  ["bus ethics"] = "Business ethics (Oxford, England)",
  ["bus ethics q"] = "Business ethics quarterly : the journal of the Society for Business Ethics",
  ["bus health"] = "Business and health",
  ["bus hist"] = "Business history",
  ["bus hist rev"] = "Business history review",
  ["bus horiz"] = "Business horizons",
  ["bus insur"] = "Business insurance",
  ["bus polit"] = "Business and politics",
  ["bus prof ethics"] = "Business & professional ethics",
  ["bus prof ethics j"] = "Business & professional ethics journal",
  ["bus rev"] = "Business review (Federal Reserve Bank of Philadelphia)",
  ["bus soc"] = "Business and society",
  ["bus strategy environ"] = "Business strategy and the environment",
  ["bus thail"] = "Business in Thailand",
  ["bus week"] = "Business week",
  ["butl soc catalana mat"] = "Butlleti de la Societat Catalana de Matematiques",
  ["butll inst catalana hist nat"] = "Butlletí de la Institució Catalana d'Història Natural",
  ["butll soc amics hist cienc farm catalana"] = "Butlletí de la Societat d'Amics de la Història i de la Ciència Farmacèutica Catalana : BSAHCFC",
  ["by bygd"] = "By og bygd",
  ["bygone kent"] = "Bygone Kent",
  ["byogen biseibutsu kenshutsu joho geppo"] = "Byōgen biseibutsu kenshutsu jōhō, geppo = Infectious agents surveillance report (IASR)",
  ["byull eksp biol med"] = "Byulleten Eksperimental’noi Biologii i Meditsiny",
  ["byzantinische forsch"] = "Byzantinische Forschungen",
  ["byzantinische z"] = "Byzantinische Zeitschrift",
  ["bzb bayer zahnarztebl"] = "BZB. Bayerisches Zahnärzteblatt",
  ["bzb bayer zahnarztebl mitt kassenzahnarztl ver bayerns"] = "BZB: Bayerisches Zahnarzteblatt, mit Mitteilungen d. Kassenzahnarztlichen Vereinigung Bayerns",
  ["bündner monbl"] = "Bündner Monatsblatt",
  ["bündnerwald, beih"] = "Bündnerwald, Beiheft",
  ["bürg staat"] = "Bürger im Staat, Der",
  ["c (basel)"] = "C",
  ["c r acad bulg sci"] = "Comptes rendus de l'Académie Bulgare des sciences : sciences mathématiques et naturelles",
  ["c r acad bulgare sci"] = "Dokladi na Blgarskata Akademiya na Naukite. Comptes Rendus de l’Academie Bulgare des Sciences",
  ["c r acad hebd seances acad sci d"] = "Comptes rendus hebdomadaires des séances de l'Académie des sciences. Série D: Sciences naturelles",
  ["c r acad sci"] = "Comptes Rendus de l’Académie des Sciences",
  ["c r acad sci gen"] = "Comptes rendus de l'Académie des sciences. La vie des sciences",
  ["c r acad sci hebd seances acad sci d"] = "Comptes Rendus Hebdomadaires des Seances de l’Academie des Sciences. D, Sciences Naturelles",
  ["c r acad sci ii"] = "Comptes rendus de l'Académie des sciences. Série II, Mécanique, physique, chimie, sciences de l'univers, sciences de la terre",
  ["c r acad sci iib mec"] = "Comptes rendus de l'Académie des sciences. Série II. Fascicule b, Mécanique = Mechanics",
  ["c r acad sci iii"] = "Comptes rendus de l'Académie des sciences. Série III, Sciences de la vie",
  ["c r acad sci paris - math"] = "Comptes Rendus de l’Académie des Sciences",
  ["c r acad sci paris sér i math"] = "Comptes Rendus de l’Académie des Sciences",
  ["c r acad sci paris, ser i"] = "Comptes Rendus de l’Académie des Sciences - Series I - Mathematics",
  ["c r acad sci ser a"] = "Comptes Rendus de l’Académie des Sciences, Serie A: Sciences Mathématiques",
  ["c r acad sci ser b"] = "Comptes Rendus de l’Académie des Sciences, Serie B: Sciences Physiques",
  ["c r acad sci, ser iia: sci terre planets"] = "Comptes Rendus de l’ Academie des Sciences Serie IIa:Sciences de la Terre et des Planets",
  ["c r acad sci, ser iib: mec, phys, chim, astron"] = "Comptes Rendus de l’ Academie des Sciences Serie IIb:Mecanique Physique Chimie Astronomie",
  ["c r acad sci, ser iic: chim"] = "Comptes Rendus de l’ Academie des Sciences Serie IIc:Chemie",
  ["c r acad sci, ser iii"] = "Comptes Rendus de l’ Academie des Sciences Serie III: Sciences de la Vie",
  ["c r assoc anat"] = "Comptes rendus de l'Association des anatomistes",
  ["c r biol"] = "Comptes rendus biologies",
  ["c r chim"] = "Comptes Rendus Chimie",
  ["c r geosci"] = "Comptes Rendus Geoscience",
  ["c r hebd seances acad sci"] = "Comptes rendus hebdomadaires des séances de l'Académie des sciences",
  ["c r hebd séances acad sci"] = "Comptes Rendus Hebdomadaires des Séances de l’Académie des Sciences",
  ["c r math"] = "Comptes rendus mathematique",
  ["c r math acad sci paris"] = "Comptes Rendus Mathematique. Academie des Sciences. Paris",
  ["c r math acad sci soc r can"] = "Comptes Rendus Mathematiques de l’Academie des Sciences. La Societe Royale du Canada. Mathematical Reports of the Academy of Science. The Royal Society of Canada",
  ["c r math rep acad sci canada"] = "La Société Royale du Canada",
  ["c r mec"] = "Comptes Rendus Mecanique",
  ["c r palevol"] = "Comptes rendus. Palévol",
  ["c r phys"] = "Comptes Rendus Physique",
  ["c r physique"] = "Comptes Rendus Physique",
  ["c r seances acad sci d"] = "Comptes rendus des séances de l'Académie des sciences. Série D, Sciences naturelles",
  ["c r seances acad sci iii"] = "Comptes rendus des séances de l'Académie des sciences. Série III, Sciences de la vie",
  ["c r seances soc biol fil"] = "Comptes rendus des séances de la Société de biologie et de ses filiales",
  ["c r soc fr gyncol"] = "Comptes rendus de la Société française de gynécologie",
  ["c r ther pharmacol clin"] = "Comptes rendus de thérapeutique et de pharmacologie clinique",
  ["c r trav lab carlsberg"] = "Comptes-rendus des travaux du Laboratoire Carlsberg",
  ["c r trav lab carlsberg [chim]"] = "Comptes rendus des travaux du Laboratoire Carlsberg. Serie chimique",
  ["c r trav lab carlsberg chim"] = "Comptes rendus des travaux du Laboratoire Carlsberg. Série chimique",
  ["c&en global enterp"] = "C&EN Global Enterprise",
  ["ca cancer j clin"] = "CA: a cancer journal for clinicians",
  ["caai trans intell technol"] = "CAAI Transactions on Intelligence Technology",
  ["cabi agric biosci"] = "CABI Agriculture and Bioscience",
  ["cabi one health"] = "CABI One Health",
  ["cabi rev"] = "CABI Reviews",
  ["cad casa oswaldo cruz"] = "Cadernos da Casa de Oswaldo Cruz",
  ["cad ciênc tecnol"] = "Cadernos de Ciência e Tecnologia",
  ["cad hist filos ciênc (3)"] = "Cadernos de História e Filosofia da Ciência",
  ["cad hist saude"] = "Cadernos de história e saúde",
  ["cad log comput"] = "Cadernos de Logica e Computacao",
  ["cad log filos"] = "Cadernos de Logica e Filosofia",
  ["cad pesqui"] = "Cadernos De Pesquisas",
  ["cad saude colet"] = "Cadernos saúde coletiva",
  ["cad saude publica"] = "Cadernos de saúde pública",
  ["cad saúde pública"] = "Cadernos de Saúde Pública",
  ["cad ter labor"] = "Caderno de terapêutica Labor",
  ["cadth technol overv"] = "CADTH technology overviews",
  ["cafs news"] = "CAFS news : a Centre for African Family Studies (CAFS) newsletter",
  ["cah  homeopath ther comp"] = "Cahiers d'homéopathie et de thérapeutique comparée",
  ["cah acad bretagne"] = "Cahiers de lÁcadémie de Bretagne",
  ["cah afr adm publique"] = "Cahiers africains d'administration publique = African administrative studies",
  ["cah agric"] = "Cahiers Agricultures",
  ["cah alex"] = "Cahiers d'Alexandrie",
  ["cah all isr univers"] = "Les Cahiers de l'Alliance israélite universelle",
  ["cah alsac archeol art hist"] = "Cahiers alsaciens d'archéologie, d'art et d'histoire",
  ["cah am lat"] = "Cahiers des Amériques latines (Paris, France : 1985)",
  ["cah am lat ser sci homme"] = "Série sciences de l'homme",
  ["cah anal donnees"] = "Les Cahiers de l'analyse des données",
  ["cah anesthesiol"] = "Cahiers d'anesthésiologie",
  ["cah ann normandie"] = "Cahiers des annales de Normandie. Annales de Normandie",
  ["cah annee gerontol"] = "Les cahiers de L'année gérontologique",
  ["cah anthropol"] = "Cahiers dánthropologie",
  ["cah archeol hist berry"] = "Cahiers d'archéologie et d'histoire du Berry",
  ["cah bazadais"] = "Les cahiers du Bazadais",
  ["cah biol mar"] = "Cahiers de biologie marine",
  ["cah brux"] = "Cahiers bruxellois",
  ["cah cent rech etud paris lle de france"] = "Cahiers du Centre de recherches et d'études sur Paris et l'Ile-de-France (CREPIF)",
  ["cah cent tech bois"] = "Cahier du Centre technique du bois",
  ["cah civilis alp"] = "Cahiers de civilisation alpine",
  ["cah clio"] = "Cahiers de Clio",
  ["cah coll med hop paris"] = "Les Cahiers du Collège de médecine des hôpitaux de Paris",
  ["cah com consult natl ethique sci sante"] = "Les cahiers du Comité consultatif national d'éthique pour les sciences de la vie et de la santé",
  ["cah confront"] = "Cahiers Confrontation",
  ["cah dix"] = "Les Cahiers des Dix. Les Dix",
  ["cah econ brux"] = "Cahiers économiques de Bruxelles",
  ["cah econ bruxelles"] = "Cahiers Economiques de Bruxelles",
  ["cah econ politique"] = "Cahiers d’Economie Politique",
  ["cah econ soc"] = "Cahiers économiques et sociaux",
  ["cah econ sociologie rurales"] = "Cahiers d’Economie et Sociologie Rurales",
  ["cah etud afr"] = "Cahiers d'études africaines",
  ["cah etud anc"] = "Cahiers des études anciennes",
  ["cah etud hong finl"] = "Cahiers d'études hongroises et finlandaises",
  ["cah etud mediev"] = "Cahiers d'études médiévales (Montréal, Québec)",
  ["cah fanjeaux"] = "Cahiers de fanjeaux",
  ["cah fontenay"] = "Les Cahiers de Fontenay",
  ["cah geogr que"] = "Cahiers de géographie de Québec",
  ["cah geogr rouen"] = "Cahiers geographiques de Rouen",
  ["cah geol"] = "Cahiers géologiques",
  ["cah haut marnais"] = "Les Cahiers haut-marnais",
  ["cah hist"] = "Cahiers d'histoire",
  ["cah hist espaces marx assoc"] = "Cahiers d'histoire (Espaces Marx (Association))",
  ["cah hist mond"] = "Cahiers d'histoire mondiale. Journal of world history",
  ["cah hist second guerr mond"] = "Cahiers d'histoire de la seconde guerre mondiale",
  ["cah inf tech/rev metall"] = "Cahiers d’Informations Techniques / Revue de Metallurgie",
  ["cah int hist econ soc"] = "Cahiers internationaux d'histoire économique et sociale",
  ["cah iroise"] = "Les Cahiers de l'Iroise",
  ["cah laennec"] = "Cahiers Laënnec",
  ["cah leopold delisle"] = "Cahiers Léopold Delisle",
  ["cah log epistemol"] = "Cahiers de Logique et d’Epistemologie",
  ["cah lorrains"] = "Les Cahiers lorrains",
  ["cah med"] = "Les Cahiers de médecine",
  ["cah med auvergne"] = "Cahiers médicaux d'Auvergne",
  ["cah med inter prof"] = "Cahiers de médecine interprofessionnelle",
  ["cah med lyon"] = "Cahiers médicaux lyonnais",
  ["cah med union fr"] = "Cahiers médicaux de l'union française; revue médico-chirurgicale de la France d'outre-mer",
  ["cah med vet"] = "Les Cahiers de médecine vétérinaire",
  ["cah medicaux union fr"] = "Cahiers medicaux de l’union francaise; revue medico-chirurgicale de la France d’outre-mer",
  ["cah mediterr orient monde turco iran"] = "Cahiers d'études sur la Méditerranée orientale et le monde turco-iranien",
  ["cah monde russe"] = "Cahiers du monde russe",
  ["cah monde russe sov"] = "Cahiers du monde russe et soviétique",
  ["cah nurs"] = "Les cahiers du nursing",
  ["cah nutr diet"] = "Cahiers de Nutrition et de Dietetique",
  ["cah o m"] = "Les cahiers d'outre-mer",
  ["cah odontostomatol"] = "Cahiers odonto-stomatologiques",
  ["cah odontostomatol (touraine)"] = "Cahiers d’Odonto-Stomatologie",
  ["cah odontostomatol touraine"] = "Cahiers d'odonto-stomatologie",
  ["cah orient"] = "Les Cahiers de l'Orient",
  ["cah orstom (sci hum)"] = "Cahiers O. R. S. T. O. M. Série sciences humaines",
  ["cah pediatr soc"] = "Cahiers de pédiatrie sociale",
  ["cah phys"] = "Cahiers de Physique",
  ["cah prothese"] = "Les Cahiers de prothèse",
  ["cah psychiatr"] = "Cahiers de psychiatrie",
  ["cah que demogr"] = "Cahiers québécois de démographie",
  ["cah ration"] = "Les Cahiers rationalistes",
  ["cah rech archit"] = "Les Cahiers de la recherche architecturale",
  ["cah rech sociol"] = "Cahiers de recherche sociologique",
  ["cah relig afr"] = "Cahiers des religions africaines",
  ["cah rmf bull med guide pract rev"] = "Cahiers R. M. F.: bulletin médical-guide du praticien-revue médicale française",
  ["cah sci hum"] = "Cahiers des sciences humaines",
  ["cah semin philos"] = "Cahiers du séminaire de philosophie",
  ["cah sexol clin"] = "Cahiers de sexologie clinique",
  ["cah sociol demogr med"] = "Cahiers de sociologie et de démographie médicales",
  ["cah sociol econ cult"] = "Cahiers de sociologie économique et culturelle, ethnopsychologie",
  ["cah tech afro"] = "Cahiers techniques AFRO",
  ["cah topol geom differ categ"] = "Cahiers de Topologie et Geometrie Differentielle Categoriques",
  ["cah tunis"] = "Les cahiers de Tunisie",
  ["cah victor edouardiens"] = "Cahiers victoriens & édouardiens",
  ["cah vilfredo pareto"] = "Cahiers Vilfredo Pareto",
  ["cahiers art sci"] = "Cahiers Art et Science",
  ["cahiers centre logique"] = "Cahiers du Centre de Logique",
  ["cahiers centre études rech opér"] = "Cahiers du Centre d’études de Recherche Opérationnelle",
  ["cahiers math école polytech féd lausanne"] = "Cahiers Mathématiques de l’École Polytechnique Fédérale de Lausanne",
  ["cahiers topologie géom différentielle catég"] = "Cahiers de Topologie et Géométrie Différentielle Catégoriques",
  ["cailiao baohu"] = "Cailiao Baohu",
  ["cailiao rechuli xuebao"] = "Cailiao Rechuli Xuebao",
  ["cairo today"] = "Cairo today",
  ["cais fac newsl"] = "CAIS faculty newsletter",
  ["cal j emerg med"] = "The California journal of emergency medicine",
  ["cal state j med"] = "California state journal of medicine",
  ["cal west med"] = "California and western medicine",
  ["calabar j health sci"] = "Calabar journal of health sciences",
  ["calc var partial differ equ"] = "Calculus of variations and partial differential equations",
  ["calc var partial differ equations"] = "Calculus of Variations and Partial Differential Equations",
  ["calc var partial differential equations"] = "Calculus of Variations and Partial Differential Equations",
  ["calcif tissue int"] = "Calcified tissue international",
  ["calcif tissue res"] = "Calcified tissue research",
  ["calcium bind proteins"] = "Calcium binding proteins",
  ["calcium signal (st clara)"] = "Calcium signaling (Santa Clara, Calif.)",
  ["calcoli sublimi"] = "Calcoli Sublimi—Collana di Analisi Matematica",
  ["calcutta hist j"] = "The Calcutta historical journal",
  ["calcutta med j"] = "Calcutta medical journal",
  ["calcutta med rev"] = "Calcutta medical review",
  ["calcutta stat assoc bull"] = "Calcutta Statistical Association Bulletin",
  ["calcutta statist assoc bull"] = "Calcutta Statistical Association Bulletin",
  ["calif agric"] = "California Agriculture",
  ["calif agric (berkeley)"] = "California agriculture",
  ["calif anthropol"] = "California anthropologist",
  ["calif archaeol"] = "California Archaeology",
  ["calif clin"] = "The California clinician",
  ["calif code regul"] = "California code of regulations",
  ["calif counts"] = "California counts",
  ["calif fish game"] = "California fish and game",
  ["calif health"] = "California's health",
  ["calif hist q"] = "California historical quarterly",
  ["calif hist soc q"] = "California Historical Society quarterly",
  ["calif hosp"] = "California hospitals",
  ["calif j"] = "California journal",
  ["calif j health promot"] = "Californian journal of health promotion",
  ["calif j health syst pharm"] = "The California journal of health-system pharmacy : CJHP",
  ["calif j politics policy"] = "California journal of politics and policy",
  ["calif law rev"] = "California law review",
  ["calif mag"] = "California magazine (Beverly Hills, Calif.)",
  ["calif manage rev"] = "California management review",
  ["calif med"] = "California medicine",
  ["calif nurse"] = "California nurse",
  ["calif public health rep"] = "California public health report. California. Department of Public Health",
  ["calif school psychol"] = "The California school psychologist : CASP",
  ["calif ser public anthropol"] = "California series in public anthropology",
  ["calif sociol"] = "California sociologist",
  ["calif state bar j"] = "California State Bar journal",
  ["calif trial lawyers j"] = "California trial lawyers journal",
  ["calif west int law j"] = "California Western international law journal",
  ["calif west law rev"] = "California Western law review",
  ["californian j health promot"] = "Californian journal of health promotion",
  ["calorim anal therm"] = "Calorimetrie et Analyse Thermique",
  ["calphad: comput coupling phase diagrams thermochem"] = "CALPHAD: Computer Coupling of Phase Diagrams and Thermochemistry",
  ["camb aerosp ser"] = "Cambridge Aerospace Series",
  ["camb anthropol"] = "Cambridge anthropology : a journal of the Department of Social Anthropology, Cambridge University",
  ["camb companions philos"] = "Cambridge Companions to Philosophy",
  ["camb iisc ser"] = "Cambridge-IISc Series",
  ["camb introd philos"] = "Cambridge Introductions to Philosophy",
  ["camb j math"] = "Cambridge Journal of Mathematics",
  ["camb law j"] = "The Cambridge law journal",
  ["camb libr collect"] = "Cambridge Library Collection",
  ["camb monogr at mol chem phys"] = "Cambridge Monographs on Atomic, Molecular and Chemical Physics",
  ["camb monogr part phys nucl phys cosmol"] = "Cambridge Monographs on Particle Physics, Nuclear Physics and Cosmology",
  ["camb munic code 1988 camb mass"] = "Cambridge municipal code 1988 : a codification of the general ordinances of the city of Cambridge, Massachusetts. Cambridge (Mass.)",
  ["camb obs handb res astron"] = "Cambridge Observing Handbooks for Research Astronomers",
  ["camb philos class"] = "Cambridge Philosophy Classics",
  ["camb q healthc ethics"] = "Cambridge quarterly of healthcare ethics : CQ : the international journal of healthcare ethics committees",
  ["camb ser hum comput interact"] = "Cambridge Series on Human-Computer Interaction",
  ["camb ser stat probab math"] = "Cambridge Series in Statistical and Probabilistic Mathematics",
  ["camb texts hist philos"] = "Cambridge Texts in the History of Philosophy",
  ["cambios rev med"] = "Cambios revista médica",
  ["cambodge soir"] = "Cambodge soir",
  ["cambridge archaeol j"] = "Cambridge Archaeological Journal",
  ["cambridge econ pol rev"] = "Cambridge Economic Policy Review",
  ["cambridge elem elem appl categ theory"] = "Cambridge Elements. Elements in Applied Category Theory",
  ["cambridge int ser parallel comput"] = "Cambridge International Series on Parallel Computation",
  ["cambridge j econ"] = "Cambridge journal of economics",
  ["cambridge j jr scientists"] = "Cambridge Journal for Junior Scientists",
  ["cambridge j math"] = "Cambridge Journal of Mathematics",
  ["cambridge lecture notes phys"] = "Cambridge Lecture Notes in Physics",
  ["cambridge math lib"] = "Cambridge Mathematical Library",
  ["cambridge math textbooks"] = "Cambridge Mathematical Textbooks",
  ["cambridge monogr appl comput math"] = "Cambridge Monographs on Applied and Computational Mathematics",
  ["cambridge monogr math phys"] = "Cambridge Monographs on Mathematical Physics",
  ["cambridge monogr mech"] = "Cambridge Monographs on Mechanics",
  ["cambridge monogr phys"] = "Cambridge Monographs on Physics",
  ["cambridge nonlinear sci ser"] = "Cambridge Nonlinear Science Series",
  ["cambridge prisms: coastal futures"] = "Cambridge Prisms: Coastal Futures",
  ["cambridge prisms: extinction"] = "Cambridge Prisms: Extinction",
  ["cambridge prisms: global mental health"] = "Cambridge Prisms: Global Mental Health",
  ["cambridge prisms: plast"] = "Cambridge Prisms: Plastics",
  ["cambridge prisms: precis med"] = "Cambridge Prisms: Precision Medicine",
  ["cambridge prisms: water"] = "Cambridge Prisms: Water",
  ["cambridge q healthcare ethics"] = "Cambridge Quarterly of Healthcare Ethics",
  ["cambridge solid state sci ser"] = "Cambridge Solid State Science Series",
  ["cambridge stud adv math"] = "Cambridge Studies in Advanced Mathematics",
  ["cambridge stud appl econom"] = "Cambridge Studies in Applied Econometrics",
  ["cambridge stud math biol"] = "Cambridge Studies in Mathematical Biology",
  ["cambridge stud modern optics"] = "Cambridge Studies in Modern Optics",
  ["cambridge stud philos"] = "Cambridge Studies in Philosophy",
  ["cambridge stud probab induc decis theory"] = "Cambridge Studies in Probability, Induction, and Decision Theory",
  ["cambridge texts appl math"] = "Cambridge Texts in Applied Mathematics",
  ["cambridge tracts in math"] = "Cambridge Tracts in Mathematics",
  ["cambridge tracts theoret comput sci"] = "Cambridge Tracts in Theoretical Computer Science",
  ["cambridge univ med soc mag"] = "Cambridge University Medical Society Magazine",
  ["campbell law rev"] = "Campbell law review",
  ["campbell syst rev"] = "Campbell Systematic Reviews",
  ["campi sapere filos"] = "Campi del Sapere/Filosofia",
  ["camsi j"] = "CAMSI journal. Journal ACEMI",
  ["can acoust"] = "Canadian acoustics = Acoustique canadienne",
  ["can aeronaut space j"] = "Canadian aeronautics and space journal. Le journal aéronautique et spatial du Canada",
  ["can agric eng"] = "Canadian Agricultural Engineering",
  ["can aids news"] = "Canadian AIDS news",
  ["can air force j"] = "Canadian Air Force Journal",
  ["can alp j"] = "Canadian Alpine Journal",
  ["can am slav stud"] = "Canadian-American Slavic studies. Revue canadienne-américaine d'études slaves",
  ["can anaesth soc j"] = "Canadian Anaesthetists' Society journal",
  ["can appl math q"] = "Canadian Applied Mathematics Quarterly",
  ["can assoc radiol j"] = "Canadian Association of Radiologists journal = Journal l'Association canadienne des radiologistes",
  ["can bar j"] = "Canadian bar journal",
  ["can bar rev"] = "The Canadian bar review",
  ["can bee j"] = "Canadian Bee Journal",
  ["can biosyst eng"] = "Canadian Biosystems Engineering",
  ["can bull cardiovasc nurs"] = "Canadian bulletin of cardiovascular nursing",
  ["can bull med hist"] = "Canadian bulletin of medical history = Bulletin canadien d'histoire de la médecine",
  ["can bus econ"] = "Canadian Business Economics",
  ["can cancer conf"] = "Canadian Cancer Conference",
  ["can ceram q"] = "Canadian Ceramics Quarterly",
  ["can chem news"] = "Canadian Chemical News",
  ["can chem process"] = "Canadian Chemical Processing",
  ["can chem trans"] = "Canadian chemical transactions",
  ["can child adolesc psychiatr rev"] = "The Canadian child and adolescent psychiatry review = La revue canadienne de psychiatrie de l'enfant et de l'adolescent",
  ["can collect"] = "Canadian Collection",
  ["can commun dis rep"] = "Canada communicable disease report = Relevé des maladies transmissibles au Canada",
  ["can commun dis rep wkly"] = "Canada communicable disease report weekly",
  ["can crit care nurs j"] = "Canadian critical care nursing journal",
  ["can dent hyg"] = "The Canadian dental hygienist",
  ["can dimens"] = "Canadian dimension",
  ["can dis wkly rep"] = "Canada diseases weekly report = Rapport hebdomadaire des maladies au Canada",
  ["can doct"] = "Canadian doctor",
  ["can electr eng j"] = "Canadian Electrical Engineering Journal",
  ["can entom"] = "The Canadian entomologist",
  ["can entomol"] = "The Canadian entomologist",
  ["can ethn stud"] = "Canadian ethnic studies",
  ["can fam law q"] = "Canadian family law quarterly",
  ["can fam physician"] = "Canadian family physician Médecin de famille canadien",
  ["can field nat"] = "The Canadian field-naturalist",
  ["can field-nat"] = "Canadian Field-Naturalist",
  ["can for j"] = "Canadian Forestry Journal",
  ["can forces dent serv bull"] = "Canadian Forces Dental Services Bulletin",
  ["can forces dent serv q"] = "Canadian Forces Dental Services quarterly",
  ["can forum"] = "The Canadian forum",
  ["can geogr"] = "The Canadian geographer. Géographe canadien",
  ["can geogr (ottowa)"] = "Canadian geographic",
  ["can geogr j"] = "Canadian Geographical Journal",
  ["can geotech j"] = "Canadian Geotechnical Journal",
  ["can geriatr j"] = "Canadian geriatrics journal : CGJ",
  ["can health welf"] = "Canada's health & welfare",
  ["can hist rev"] = "The Canadian historical review",
  ["can hiv aids policy law newsl"] = "Canadian HIV-AIDS policy & law newsletter",
  ["can hiv aids policy law rev"] = "Canadian HIV/AIDS policy & law review",
  ["can home econ j"] = "Canadian Home Economics Journal",
  ["can hosp"] = "Canadian hospital",
  ["can hum rights advocate"] = "Canadian human rights advocate",
  ["can hum rights yearb"] = "Canadian human rights yearbook = Annuaire canadien des droits de la personne",
  ["can inst food sci technol j"] = "Canadian Institute of Food Science and Technology Journal",
  ["can inst food technol j"] = "Canadian Institute of Food Technology Journal",
  ["can int educ"] = "Canadian and international education. Éducation canadienne et internationale",
  ["can j  econ"] = "Canadian Journal of Economics",
  ["can j action res"] = "Canadian Journal of Action Research",
  ["can j addict"] = "The Canadian journal of addiction",
  ["can j adm sci"] = "Canadian Journal of Adminstrative Sciences",
  ["can j admin sci"] = "Canadian Journal of Administrative Sciences",
  ["can j afr stud"] = "Canadian journal of African studies",
  ["can j aging"] = "Canadian journal on aging = La revue canadienne du vieillissement",
  ["can j agr econ"] = "Canadian Journal of Agricultural Economics",
  ["can j agr sci"] = "Canadian Journal of Agricultural Science",
  ["can j agric econ"] = "Canadian Journal of Agricultural Economics",
  ["can j anaesth"] = "Canadian journal of anaesthesia = Journal canadien d'anesthésie",
  ["can j anal sci spectros"] = "Canadian Journal of Analytical Sciences and Spectroscopy",
  ["can j anesth"] = "Canadian Journal of Anesthesia",
  ["can j anim sci"] = "Canadian Journal of Animal Science",
  ["can j anthropol"] = "Canadian Journal of Anthropology",
  ["can j appl ling"] = "Canadian Journal of Applied Linguistics",
  ["can j appl physiol"] = "Canadian Journal of Applied Physiology",
  ["can j appl sci"] = "Canadian Journal of Applied Sciences",
  ["can j appl spectrosc"] = "Canadian Journal of Applied Spectroscopy",
  ["can j appl sport sci"] = "Canadian Journal of Applied Sport Sciences",
  ["can j archaeol"] = "Canadian journal of archaeology = Journal canadien d'archéologie",
  ["can j art ther"] = "Canadian Journal of Art Therapy",
  ["can j arthropod identif"] = "Canadian journal of arthropod identification",
  ["can j behav sci"] = "Canadian journal of behavioural science. Revue canadienne des sciences du comportement",
  ["can j biochem"] = "Canadian Journal of Biochemistry",
  ["can j biochem cell biol"] = "Canadian Journal of Biochemistry and Cell Biology",
  ["can j biochem physiol"] = "Canadian journal of biochemistry and physiology",
  ["can j bioeth"] = "Canadian journal of bioethics = revue canadienne de bioéthique",
  ["can j bioethics"] = "Canadian Journal of Bioethics",
  ["can j bot"] = "Canadian journal of botany. Journal canadien de botanique",
  ["can j cardiol"] = "The Canadian journal of cardiology",
  ["can j cardiovasc nurs"] = "Canadian Journal of Cardiovascular Nursing",
  ["can j career dev"] = "Canadian Journal of Career Development",
  ["can j cathol hist"] = "Canadian Journal of Catholic History",
  ["can j chem"] = "Canadian Journal of Chemistry",
  ["can j chem eng"] = "Canadian Journal of Chemical Engineering",
  ["can j civ eng"] = "Canadian Journal of Civil Engineering",
  ["can j clin pharmacol"] = "The Canadian journal of clinical pharmacology = Journal canadien de pharmacologie clinique",
  ["can j commun"] = "Canadian Journal of Communication",
  ["can j commun ment health"] = "Canadian journal of community mental health = Revue canadienne de santé mentale communautaire",
  ["can j community dent"] = "The Canadian journal of community dentistry = La Revue canadienne de la dentisterie communautaire",
  ["can j community mental health"] = "Canadian Journal of Community Mental Health",
  ["can j comp med"] = "Canadian journal of comparative medicine (Gardenvale, Québec)",
  ["can j comp med vet sci"] = "Canadian journal of comparative medicine and veterinary science",
  ["can j criminol"] = "Canadian journal of criminology. Revue canadienne de criminologie",
  ["can j criminol crim justice"] = "Canadian Journal of Criminology and Criminal Justice",
  ["can j crit care nurs"] = "The Canadian journal of critical care nursing",
  ["can j dent hyg"] = "Canadian journal of dental hygiene : CJDH = Journal canadien de l'hygiène dentaire : JCHD",
  ["can j dermatol"] = "Canadian Journal of Dermatology",
  ["can j dev stud"] = "Canadian Journal of Development Studies",
  ["can j devel stud"] = "Canadian Journal of Development Studies",
  ["can j diabetes"] = "Canadian journal of diabetes",
  ["can j diet pract res"] = "Canadian journal of dietetic practice and research : a publication of Dietitians of Canada = Revue canadienne de la pratique et de la recherche en diététique : une publication des Diététistes du Canada",
  ["can j earth sci"] = "Canadian Journal of Earth Sciences",
  ["can j econ"] = "The Canadian journal of economics. Revue canadienne d'économique",
  ["can j educ"] = "Canadian Journal of Education",
  ["can j educ adm policy"] = "Canadian Journal of Educational Administration and Policy",
  ["can j electr comput eng"] = "Canadian Journal of Electrical and Computer Engineering",
  ["can j emerg manage"] = "Canadian Journal of Emergency Management",
  ["can j emerg med"] = "Canadian Journal of Emergency Medicine",
  ["can j eur russ stud"] = "Canadian journal of European and Russian Studies",
  ["can j exp psychol"] = "Canadian journal of experimental psychology = Revue canadienne de psychologie expérimentale",
  ["can j explor geophys"] = "Canadian Journal of Exploration Geophysics",
  ["can j fam law"] = "Canadian journal of family law. Revue canadienne de droit familial",
  ["can j fam youth"] = "Canadian Journal of Family and Youth",
  ["can j film stud"] = "Canadian Journal of Film Studies",
  ["can j fish aquat sci"] = "Canadian journal of fisheries and aquatic sciences. Journal canadien des sciences halieutiques et aquatiques",
  ["can j fish aquatsci"] = "Canadian Journal of Fisheries and Aquatic Sciences",
  ["can j for res"] = "Canadian Journal of Forest Research",
  ["can j gastroenterol"] = "Canadian journal of gastroenterology = Journal canadien de gastroenterologie",
  ["can j gastroenterol hepatol"] = "Canadian journal of gastroenterology & hepatology",
  ["can j genet cytol"] = "Canadian Journal of Genetics and Cytology",
  ["can j health technol"] = "Canadian Journal of Health Technologies",
  ["can j heath hist"] = "Canadian Journal of Health History",
  ["can j higher educ"] = "Canadian Journal of Higher Education",
  ["can j hist"] = "Canadian journal of history",
  ["can j hist sport"] = "Canadian journal of history of sport = Revue canadienne de l'histoire des sports",
  ["can j hist sport phys educ"] = "Canadian journal of history of sport and physical education",
  ["can j hosp pharm"] = "The Canadian journal of hospital pharmacy",
  ["can j hum sex"] = "The Canadian journal of human sexuality",
  ["can j inf lib sci"] = "Canadian Journal of Information and Library Science",
  ["can j infect control"] = "Canadian Journal of Infection Control",
  ["can j infect dis"] = "The Canadian journal of infectious diseases = Journal canadien des maladies infectieuses",
  ["can j infect dis med microbiol"] = "The Canadian journal of infectious diseases & medical microbiology = Journal canadien des maladies infectieuses et de la microbiologie médicale",
  ["can j ir stud"] = "Canadian Journal of Irish Studies",
  ["can j ital stud"] = "Canadian journal of Italian studies",
  ["can j kidney health dis"] = "Canadian journal of kidney health and disease",
  ["can j lat am caribb stud"] = "Canadian journal of Latin American and Caribbean studies = Revue canadienne des études latino-américaines et caraïbes",
  ["can j law jurisprud"] = "The Canadian journal of law and jurisprudence",
  ["can j law soc"] = "Canadian journal of law and society = Revue canadienne de droit et société",
  ["can j learn technol"] = "Canadian Journal Of Learning And Technology",
  ["can j ling"] = "Canadian Journal of Linguistics",
  ["can j linguist"] = "Canadian journal of linguistics. La revue canadienne de linguistique",
  ["can j math"] = "Canadian Journal of Mathematics",
  ["can j med lab sci"] = "Canadian Journal of Medical Laboratory Science",
  ["can j med radiat technol"] = "Canadian Journal of Medical Radiation Technology",
  ["can j med sci"] = "Canadian journal of medical sciences",
  ["can j med surg"] = "Canadian Journal of Medicine and Surgery",
  ["can j med technol"] = "Canadian journal of medical technology",
  ["can j microbiol"] = "Canadian Journal of Microbiology",
  ["can j mineral petrol"] = "Canadian Journal of Mineralogy and Petrology",
  ["can j music"] = "Canadian Journal of Music",
  ["can j native stud"] = "The Canadian journal of native studies",
  ["can j neurol"] = "Canadian Journal of Neurology",
  ["can j neurol sci"] = "The Canadian journal of neurological sciences. Le journal canadien des sciences neurologiques",
  ["can j neurosci nurs"] = "Canadian journal of neuroscience nursing",
  ["can j nonprofit social econ res"] = "Canadian Journal of Nonprofit and Social Economy Research",
  ["can j nurs adm"] = "Canadian journal of nursing administration",
  ["can j nurs inf"] = "Canadian Journal of Nursing Informatics",
  ["can j nurs leadersh"] = "Canadian journal of nursing leadership",
  ["can j nurs res"] = "The Canadian journal of nursing research = Revue canadienne de recherche en sciences infirmières",
  ["can j ob gyn womens health care"] = "The Canadian journal of ob/gyn & women's health care",
  ["can j occup ther"] = "Canadian journal of occupational therapy. Revue canadienne d'ergothérapie",
  ["can j oncol"] = "The Canadian journal of oncology",
  ["can j ophthalmol"] = "Canadian journal of ophthalmology. Journal canadien d'ophtalmologie",
  ["can j optom"] = "Canadian Journal of Optometry",
  ["can j otolaryngol"] = "Canadian Journal of Otolaryngology",
  ["can j otolaryngol suppl"] = "Canadian journal of otolaryngology. Supplement",
  ["can j pain"] = "Canadian journal of pain = Revue canadienne de la douleur",
  ["can j pathol"] = "Canadian Journal of Pathology",
  ["can j pharm sci"] = "Canadian Journal of Pharmaceutical Sciences",
  ["can j philos"] = "Canadian journal of philosophy",
  ["can j phys"] = "Canadian Journal of Physics",
  ["can j physiol pharmacol"] = "Canadian Journal of Physiology and Pharmacology",
  ["can j plant pathol"] = "Canadian Journal of Plant Pathology",
  ["can j plant sci"] = "Canadian Journal of Plant Science",
  ["can j plast surg"] = "The Canadian journal of plastic surgery = Journal canadien de chirurgie plastique",
  ["can j polit sci"] = "Canadian journal of political science. Revue canadienne de science politique",
  ["can j program eval"] = "The Canadian journal of program evaluation = La Revue canadienne d'évaluation de programme",
  ["can j psychiatr nurs"] = "The Canadian journal of psychiatric nursing",
  ["can j psychiatry"] = "Canadian journal of psychiatry. Revue canadienne de psychiatrie",
  ["can j psychol"] = "Canadian journal of psychology",
  ["can j public health"] = "Canadian journal of public health = Revue canadienne de santé publique",
  ["can j pure appl sci"] = "Canadian Journal of Pure & Applied Science",
  ["can j radiogr radiother nucl med"] = "The Canadian journal of radiography, radiotherapy, nuclear medicine",
  ["can j reg sci"] = "Canadian Journal of Regional Science",
  ["can j remote sens"] = "Canadian journal of remote sensing = Journal canadien de télédétection",
  ["can j res"] = "Canadian Journal of Research",
  ["can j res e med sci"] = "Canadian journal of research. Section E, Medical sciences",
  ["can j res undergrad res"] = "Canadian Journal of Undergraduate Research",
  ["can j respir crit care sleep med"] = "Canadian Journal of Respiratory, Critical Care, and Sleep Medicine",
  ["can j respir ther"] = "Canadian journal of respiratory therapy : CJRT = Revue canadienne de la thérapie respiratoire : RCTR",
  ["can j rural med"] = "Canadian journal of rural medicine : the official journal of the Society of Rural Physicians of Canada = Journal canadien de la médecine rurale : le journal officiel de la Société de médecine rurale du Canada",
  ["can j sch psychol"] = "Canadian journal of school psychology",
  ["can j school psychol"] = "Canadian Journal of School Psychology",
  ["can j sci math technol educ"] = "Canadian Journal of Science, Mathematics and Technology Education",
  ["can j sociol"] = "Canadian journal of sociology = Cahiers canadiens de sociologie",
  ["can j sociol-cahiers can sociol"] = "Canadian Journal Of Sociology-Cahiers Canadiens De Sociologie",
  ["can j soil sci"] = "Canadian Journal of Soil Science",
  ["can j spectrosc"] = "Canadian Journal of Spectroscopy",
  ["can j speech lang pathol audiol"] = "Canadian journal of speech-language pathology and audiology : CJSLPA = Revue canadienne d'orthophonie et d'audiologie : RCOA",
  ["can j sport sci"] = "Canadian Journal of Sport Science",
  ["can j stat"] = "The Canadian journal of statistics = Revue canadienne de statistique",
  ["can j surg"] = "Canadian journal of surgery. Journal canadien de chirurgie",
  ["can j transp"] = "Canadian Journal of Transportation",
  ["can j urban res"] = "Canadian journal of urban research",
  ["can j urol"] = "The Canadian journal of urology",
  ["can j vet res"] = "Canadian journal of veterinary research = Revue canadienne de recherche vétérinaire",
  ["can j women law"] = "Canadian journal of women and the law = Revue juridique La femme et le droit",
  ["can j womens health care phys addressing womens health issues"] = "The Canadian journal of women's health care for physicians addressing women's health issues",
  ["can j zombie sci"] = "Canadian Journal of Zombie Science",
  ["can j zool"] = "Canadian journal of zoology",
  ["can labour"] = "Canadian labour",
  ["can liver j"] = "Canadian Liver Journal",
  ["can math bull"] = "Canadian Mathematical Bulletin",
  ["can med assoc j"] = "Canadian Medical Association journal",
  ["can med educ j"] = "Canadian medical education journal",
  ["can med j"] = "Canadian Medical Journal",
  ["can ment health"] = "Canada's mental health",
  ["can metall q"] = "Canadian Metallurgical Quarterly",
  ["can min j"] = "Canadian Mining Journal",
  ["can mineral"] = "Canadian Mineralogist",
  ["can mod lang rev"] = "The Canadian modern language review = La revue canadienne des langues vivantes",
  ["can nurse"] = "The Canadian nurse",
  ["can nurse hosp rev"] = "The Canadian nurse and hospital review",
  ["can oncol nurs j"] = "Canadian oncology nursing journal = Revue canadienne de nursing oncologique",
  ["can oper room nurs j"] = "Canadian operating room nursing journal",
  ["can packag"] = "Canadian Packaging",
  ["can pharm j"] = "Canadian pharmaceutical journal",
  ["can pharm j (ott)"] = "Canadian pharmacists journal : CPJ = Revue des pharmaciens du Canada : RPC",
  ["can prosthet orthot j"] = "Canadian Prosthetics And Orthotics Journal",
  ["can psychiatr assoc j"] = "Canadian Psychiatric Association journal",
  ["can psychol"] = "Canadian psychologist",
  ["can psychol rev"] = "Canadian psychological review. Psychologie canadienne",
  ["can public adm"] = "Canadian public administration : Administration publique du Canada",
  ["can public pol"] = "Canadian Public Policy",
  ["can public policy"] = "Canadian public policy. Analyse de politiques",
  ["can res suppl"] = "Cancer research. Supplement",
  ["can respir j"] = "Canadian respiratory journal",
  ["can rev am stud"] = "Canadian review of American studies",
  ["can rev sociol"] = "Canadian review of sociology = Revue canadienne de sociologie",
  ["can rev sociol anthropol"] = "The Canadian review of sociology and anthropology. La Revue canadienne de sociologie et d'anthropologie",
  ["can rev sociol anthropol-rev can sociol anthrol"] = "Canadian Review Of Sociology And Anthropology-Revue Canadienne De Sociologie Et D Anthropologie",
  ["can serv med j"] = "Canadian services medical journal",
  ["can slavon pap"] = "Canadian Slavonic papers",
  ["can soc forensic sci j"] = "Canadian Society of Forensic Science Journal",
  ["can soc trends"] = "Canadian social trends",
  ["can soc work rev"] = "Canadian social work review = Revue canadienne de service social",
  ["can stat rev"] = "Canadian statistical review (Ottawa, Ont.)",
  ["can stud popul"] = "Canadian studies in population",
  ["can supreme court rep can supreme court"] = "Canada Supreme Court reports. Recueil des arrêts de la Cour Suprême du Canada. Canada. Supreme Court",
  ["can tax j"] = "Canadian Tax Journal",
  ["can tech rep fish aquat sci"] = "Canadian technical report of fisheries and aquatic sciences",
  ["can urol assoc j"] = "Canadian Urological Association journal = Journal de l'Association des urologues du Canada",
  ["can vet j"] = "The Canadian veterinary journal = La revue vétérinaire canadienne",
  ["can water resour j"] = "Canadian Water Resources Journal",
  ["can welf"] = "CW, Canadian welfare",
  ["can womens stud"] = "Canadian women's studies = Les cahiers de la femme",
  ["can ye ke xue"] = "Can ye ke xue = Canye kexue = Acta sericologica sinica",
  ["can yearb int law"] = "The Canadian yearbook of international law. Annuaire canadien de droit international",
  ["can young sci j"] = "Canadian Young Scientist Journal",
  ["canad appl math quart"] = "The Canadian Applied Mathematics Quarterly",
  ["canad j math"] = "Canadian Journal of Mathematics. Journal Canadien de Mathematiques",
  ["canad j statist"] = "The Canadian Journal of Statistics. La Revue Canadienne de Statistique",
  ["canad math bull"] = "Canadian Mathematical Bulletin. Bulletin Canadien de Mathematiques",
  ["canad math soc ser monogr adv texts"] = "Canadian Mathematical Society Series of Monographs and Advanced Texts",
  ["canada (tor)"] = "Canada",
  ["canadi j aborig community based hiv/aids res"] = "Canadian journal of Aboriginal community-based HIV/AIDS research",
  ["canadian j elect comput eng"] = "Canadian Journal of Electrical and Computer Engineering",
  ["cancer biochem biophys"] = "Cancer biochemistry biophysics",
  ["cancer biol med"] = "Cancer biology & medicine",
  ["cancer biol ther"] = "Cancer biology & therapy",
  ["cancer biomark"] = "Cancer biomarkers : section A of Disease markers",
  ["cancer biomarkers"] = "Cancer Biomarkers",
  ["cancer biother"] = "Cancer biotherapy",
  ["cancer biother radiopharm"] = "Cancer biotherapy & radiopharmaceuticals",
  ["cancer bull"] = "Cancer bulletin (Houston, Tex.)",
  ["cancer causes control"] = "Cancer causes & control : CCC",
  ["cancer cell"] = "Cancer cell",
  ["cancer cell int"] = "Cancer cell international",
  ["cancer cell microenviron"] = "Cancer cell & microenvironment",
  ["cancer cells"] = "Cancer cells (Cold Spring Harbor, N.Y. : 1989)",
  ["cancer chemother biol response modif"] = "Cancer chemotherapy and biological response modifiers",
  ["cancer chemother pharmacol"] = "Cancer chemotherapy and pharmacology",
  ["cancer chemother rep"] = "Cancer chemotherapy reports",
  ["cancer chemother rep 2"] = "Cancer chemotherapy reports. Part 2",
  ["cancer chemother rep 3"] = "Cancer chemotherapy reports. Part 3",
  ["cancer chemother rep [2]"] = "Cancer Chemotherapy Reports. Part 2",
  ["cancer chemother rep [3]"] = "Cancer Chemotherapy Reports. Part 3",
  ["cancer clin oncol"] = "Cancer and clinical oncology",
  ["cancer clin trials"] = "Cancer clinical trials",
  ["cancer commun"] = "Cancer communications",
  ["cancer commun (lond)"] = "Cancer communications (London, England)",
  ["cancer control"] = "Cancer control : journal of the Moffitt Cancer Center",
  ["cancer converg"] = "Cancer convergence",
  ["cancer cytol"] = "Cancer cytology",
  ["cancer cytopathol"] = "Cancer cytopathology",
  ["cancer detect prev"] = "Cancer detection and prevention",
  ["cancer detect prev suppl"] = "Cancer detection and prevention. Supplement : official publication of the International Society for Preventive Oncology, Inc",
  ["cancer discov"] = "Cancer discovery",
  ["cancer drug deliv"] = "Cancer drug delivery",
  ["cancer drug discov d"] = "Macromolecular Anticancer Therapeutics",
  ["cancer drug resist"] = "Cancer drug resistance (Alhambra, Calif.)",
  ["cancer epidemiol"] = "Cancer epidemiology",
  ["cancer epidemiol biomarkers prev"] = "Cancer epidemiology, biomarkers & prevention : a publication of the American Association for Cancer Research, cosponsored by the American Society of Preventive Oncology",
  ["cancer forum"] = "Cancer forum",
  ["cancer gene ther"] = "Cancer gene therapy",
  ["cancer genet"] = "Cancer genetics",
  ["cancer genet cytogenet"] = "Cancer Genetics and Cytogenetics",
  ["cancer genomics proteomics"] = "Cancer genomics & proteomics",
  ["cancer growth metastasis"] = "Cancer growth and metastasis",
  ["cancer hallm"] = "Cancer hallmarks",
  ["cancer health disparities"] = "Cancer health disparities",
  ["cancer imaging"] = "Cancer imaging : the official publication of the International Cancer Imaging Society",
  ["cancer immun"] = "Cancer immunity",
  ["cancer immunol immunother"] = "Cancer immunology, immunotherapy : CII",
  ["cancer immunol res"] = "Cancer immunology research",
  ["cancer immunol, immunother"] = "Cancer Immunology, Immunotherapy",
  ["cancer inf"] = "Cancer Informatics",
  ["cancer inform"] = "Cancer informatics",
  ["cancer invest"] = "Cancer investigation",
  ["cancer j"] = "Cancer journal (Sudbury, Mass.)",
  ["cancer j sci am"] = "The cancer journal from Scientific American",
  ["cancer lett"] = "Cancer letters",
  ["cancer lett (amsterdam, neth)"] = "Cancer Letters (Amsterdam, Netherlands)",
  ["cancer lett (n y, ny, u s)"] = "Cancer Letters (New York, NY, United States)",
  ["cancer lett (shannon, irel)"] = "Cancer Letters (Shannon, Ireland)",
  ["cancer manag res"] = "Cancer management and research",
  ["cancer med"] = "Cancer medicine",
  ["cancer med j"] = "Cancer medicine journal",
  ["cancer metab"] = "Cancer & metabolism",
  ["cancer metastasis rev"] = "Cancer metastasis reviews",
  ["cancer microenviron"] = "Cancer microenvironment : official journal of the International Cancer Microenvironment Society",
  ["cancer nanotechnol"] = "Cancer Nanotechnology",
  ["cancer news"] = "Cancer news",
  ["cancer nurs"] = "Cancer nursing",
  ["cancer pract"] = "Cancer practice",
  ["cancer prev"] = "Cancer prevention (Philadelphia, Pa.)",
  ["cancer prev control"] = "Cancer prevention & control : CPC = Prévention & contrôle en cancérologie : PCC",
  ["cancer prev res (phila)"] = "Cancer prevention research (Philadelphia, Pa.)",
  ["cancer prog"] = "Cancer progress",
  ["cancer radiother"] = "Cancer radiothérapie : journal de la Société française de radiothérapie oncologique",
  ["cancer rep"] = "Cancer reports",
  ["cancer rep (hoboken)"] = "Cancer reports (Hoboken, N.J.)",
  ["cancer rep rev"] = "Cancer reports and reviews",
  ["cancer res"] = "Cancer research",
  ["cancer res commun"] = "Cancer Research Communications",
  ["cancer res front"] = "Cancer research frontiers",
  ["cancer res j"] = "Cancer research journal",
  ["cancer res treat"] = "Cancer research and treatment",
  ["cancer sci"] = "Cancer science",
  ["cancer stud mol med"] = "Cancer studies and molecular medicine : open journal",
  ["cancer stud ther"] = "Cancer studies and therapeutics",
  ["cancer surv"] = "Cancer surveys",
  ["cancer ther"] = "Cancer therapy",
  ["cancer ther oncol int j"] = "Cancer therapy & oncology international journal",
  ["cancer transl med"] = "Cancer translational medicine",
  ["cancer treat commun"] = "Cancer treatment communications",
  ["cancer treat rep"] = "Cancer treatment reports",
  ["cancer treat res"] = "Cancer treatment and research",
  ["cancer treat res commun"] = "Cancer treatment and research communications",
  ["cancer treat rev"] = "Cancer treatment reviews",
  ["cancers (basel)"] = "Cancers",
  ["cancers head neck"] = "Cancers of the head & neck",
  ["canine genet epidemiol"] = "Canine genetics and epidemiology",
  ["canine med genet"] = "Canine medicine and genetics",
  ["canine pract"] = "Canine practice",
  ["cannabis cannabinoid res"] = "Cannabis and cannabinoid research",
  ["cannt j"] = "CANNT journal = Journal ACITN",
  ["cao ye ke xue"] = "Cao ye ke xue = Pratacultural science = Caoye kexue",
  ["cao ye xue bao"] = "Cao ye xue bao = Acta prataculturae Sinica = Caoye xuebao",
  ["cap chem"] = "The Capital chemist",
  ["cap nurs"] = "Capital nursing",
  ["cap today"] = "CAP today",
  ["cap univ law rev"] = "Capital University law review",
  ["cape breton post"] = "Cape Breton post",
  ["capitation manag rep"] = "Capitation management report",
  ["capitation rates data"] = "Capitation rates & data",
  ["caraka tani: j sustainable agric"] = "Caraka Tani: Journal of Sustainable Agriculture",
  ["carbohydr chem"] = "Carbohydrate Chemistry",
  ["carbohydr lett"] = "Carbohydrate letters",
  ["carbohydr polym"] = "Carbohydrate polymers",
  ["carbohydr polym technol appl"] = "Carbohydrate Polymer Technologies and Applications",
  ["carbohydr res"] = "Carbohydrate Research",
  ["carbon balance manag"] = "Carbon balance and management",
  ["carbon balance manage"] = "Carbon Balance and Management",
  ["carbon capture sci technol"] = "Carbon Capture Science & Technology",
  ["carbon clim law rev"] = "Carbon and Climate Law Review",
  ["carbon energy"] = "Carbon Energy",
  ["carbon lett"] = "Carbon Letters",
  ["carbon manag"] = "Carbon management",
  ["carbon manage"] = "Carbon Management",
  ["carbon mater: chem phys"] = "Carbon Materials: Chemistry and Physics",
  ["carbon n y"] = "Carbon",
  ["carbon neutrality"] = "Carbon Neutrality",
  ["carbon neutralization"] = "Carbon Neutralization",
  ["carbon res"] = "Carbon Research",
  ["carbon resour convers"] = "Carbon Resources Conversion",
  ["carbon trends"] = "Carbon Trends",
  ["carbonates evaporites"] = "Carbonates and Evaporites",
  ["carcinog compr surv"] = "Carcinogenesis; a comprehensive survey",
  ["card electrophysiol clin"] = "Cardiac electrophysiology clinics",
  ["card electrophysiol rev"] = "Cardiac electrophysiology review",
  ["card fail rev"] = "Cardiac failure review",
  ["cardiol angiol"] = "Cardiology and angiology",
  ["cardiol cardiovasc med"] = "Cardiology and cardiovascular medicine",
  ["cardiol clin"] = "Cardiology clinics",
  ["cardiol j"] = "Cardiology journal",
  ["cardiol manage"] = "Cardiology management",
  ["cardiol plus"] = "Cardiology plus",
  ["cardiol prat"] = "Cardiologia pratica",
  ["cardiol res"] = "Cardiology research",
  ["cardiol res cardiovasc med"] = "Cardiology research and cardiovascular medicine",
  ["cardiol res pract"] = "Cardiology research and practice",
  ["cardiol rev"] = "Cardiology in review",
  ["cardiol ther"] = "Cardiology and therapy",
  ["cardiol tunis"] = "Cardiologie tunisienne",
  ["cardiol vasc res (wilmington)"] = "Cardiology & vascular research (Wilmington, Del.)",
  ["cardiol young"] = "Cardiology in the young",
  ["cardiopulm phys ther j"] = "Cardiopulmonary physical therapy journal",
  ["cardiorenal med"] = "Cardiorenal medicine",
  ["cardiovasc clin"] = "Cardiovascular clinics",
  ["cardiovasc diabetol"] = "Cardiovascular diabetology",
  ["cardiovasc diagn ther"] = "Cardiovascular diagnosis and therapy",
  ["cardiovasc digit health j"] = "Cardiovascular digital health journal",
  ["cardiovasc dis"] = "Cardiovascular diseases",
  ["cardiovasc dis med"] = "Cardiovascular disease and medicine",
  ["cardiovasc disord med"] = "Cardiovascular disorder and medicine",
  ["cardiovasc drug rev"] = "Cardiovascular drug reviews",
  ["cardiovasc drugs ther"] = "Cardiovascular drugs and therapy",
  ["cardiovasc endocrinol"] = "Cardiovascular endocrinology",
  ["cardiovasc endocrinol metab"] = "Cardiovascular endocrinology & metabolism",
  ["cardiovasc eng"] = "Cardiovascular engineering (Dordrecht, Netherlands)",
  ["cardiovasc eng technol"] = "Cardiovascular engineering and technology",
  ["cardiovasc hematol agents"] = "Cardiovascular & Hematological Agents",
  ["cardiovasc hematol agents med chem"] = "Cardiovascular & hematological agents in medicinal chemistry",
  ["cardiovasc hematol disord drug targets"] = "Cardiovascular & hematological disorders drug targets",
  ["cardiovasc hematol disord: drug targets"] = "Cardiovascular & Hematological Disorders: Drug Targets",
  ["cardiovasc innov appl"] = "Cardiovascular innovations and applications",
  ["cardiovasc interv ther"] = "Cardiovascular intervention and therapeutics",
  ["cardiovasc intervent radiol"] = "Cardiovascular and interventional radiology",
  ["cardiovasc interventional radiol"] = "CardioVascular and Interventional Radiology",
  ["cardiovasc j afr"] = "Cardiovascular journal of Africa",
  ["cardiovasc j s afr"] = "Cardiovascular journal of South Africa : official journal for Southern Africa Cardiac Society [and] South African Society of Cardiac Practitioners",
  ["cardiovasc med"] = "Cardiovascular medicine",
  ["cardiovasc nurs"] = "Cardio-vascular nursing",
  ["cardiovasc pathobiol"] = "Cardiovascular pathobiology",
  ["cardiovasc pathol"] = "Cardiovascular pathology : the official journal of the Society for Cardiovascular Pathology",
  ["cardiovasc pharm open access"] = "Cardiovascular pharmacology: open access",
  ["cardiovasc psychiatry neurol"] = "Cardiovascular psychiatry and neurology",
  ["cardiovasc radiat med"] = "Cardiovascular radiation medicine",
  ["cardiovasc radiol"] = "Cardiovascular radiology",
  ["cardiovasc regen med"] = "Cardiovascular regenerative medicine",
  ["cardiovasc res"] = "Cardiovascular research",
  ["cardiovasc res cent bull"] = "Cardiovascular Research Center bulletin",
  ["cardiovasc rev rep"] = "Cardiovascular reviews & reports",
  ["cardiovasc revasc med"] = "Cardiovascular revascularization medicine : including molecular interventions",
  ["cardiovasc surg"] = "Cardiovascular surgery (London, England)",
  ["cardiovasc ther"] = "Cardiovascular therapeutics",
  ["cardiovasc thorac open"] = "Cardiovascular and thoracic open",
  ["cardiovasc toxicol"] = "Cardiovascular toxicology",
  ["cardiovasc ultrasound"] = "Cardiovascular ultrasound",
  ["cardozo law rev"] = "Cardozo law review",
  ["care briefs develop isssues"] = "CARE briefs on development issues",
  ["care giver"] = "Care giver : journal of the College of Chaplains",
  ["care manag j"] = "Care management journals : Journal of case management ; The journal of long term home health care",
  ["care wkly"] = "Care weekly",
  ["career dev q"] = "The Career development quarterly",
  ["career dev transit except individ"] = "Career development and transition for exceptional individuals",
  ["caribb aff"] = "Caribbean affairs (Port of Spain, Trinidad and Tobago)",
  ["caribb geogr"] = "Caribbean Geography",
  ["caribb health"] = "Caribbean health",
  ["caribb herpetol"] = "Caribbean herpetology",
  ["caribb j sci"] = "Caribbean journal of science",
  ["caribb med j"] = "Caribbean medical journal",
  ["caribb q"] = "Caribbean quarterly",
  ["caribb rev"] = "Caribbean review",
  ["caribb stud"] = "Caribbean studies (Río Piedras, San Juan, P.R.)",
  ["caribbean j math comput sci"] = "Caribbean Journal of Mathematical and Computing Sciences",
  ["caribe contemp"] = "El Caribe contemporaneo",
  ["caridad cienc arte"] = "Caridad, ciencia y arte",
  ["caries res"] = "Caries research",
  ["carinthia i"] = "Carinthia I : Mittheilungen des Geschichtsvereins für Kärnten",
  ["carl g hempel lect ser"] = "The Carl G. Hempel Lecture Series",
  ["carle sel pap"] = "Carle selected papers",
  ["carlsberg res commun"] = "Carlsberg research communications",
  ["carmarthen antiq"] = "The Carmarthen antiquary",
  ["carmarthensh hist"] = "The Carmarthensire hisorian",
  ["carnegie q"] = "Carnegie quarterly",
  ["carnegie-rochester conf ser public pol"] = "Carnegie-Rochester Conference Series on Public Policy",
  ["carnet econ"] = "Carnet de l'économe",
  ["carnets enfance"] = "Les Carnets de l'enfance = Assignment children",
  ["carney hosp j"] = "Carney Hospital journal",
  ["carol fire rescue ems j"] = "Carolina fire rescue EMS journal",
  ["carol j pharm"] = "The Carolina journal of pharmacy",
  ["carol plann"] = "Carolina planning",
  ["carolina lecture ser"] = "Carolina Lecture Series",
  ["carolinea beih (karlsr)"] = "Carolinea. Beiheft",
  ["carpathian j food sci technol"] = "Carpathian Journal of Food Science and Technology",
  ["carpathian j math"] = "Carpathian Journal of Mathematics",
  ["carpathian math publ"] = "Carpathian Mathematical Publications. Karpatski Matematichni Publikatsii",
  ["carrefour afr"] = "Carrefour africain",
  ["carta econ reg"] = "Carta económica regional : CER",
  ["carta inf"] = "Carta informativa (Asociación Demográfica Salvadoreña)",
  ["cartogr geogr inf sci"] = "Cartography and geographic information science",
  ["cartogr j"] = "The Cartographic journal",
  ["carus math monogr"] = "Carus Mathematical Monographs",
  ["caryologia g citol citosistematica citogenet"] = "Caryologia; giornale di citologia, citosistematica, e citogenetica",
  ["cas cesk lek"] = "Casopis c̆eského lékárnictva",
  ["cas cesk vet"] = "Casopis c̆eskoslovenských veterinár̆ů",
  ["cas lek cesk"] = "Casopís lékar̆ů c̆eských",
  ["cas matice moravske"] = "Časopis Matice moravské (1968)",
  ["cas nar muz"] = "Časopis Národního musea",
  ["cas suvrem povij"] = "Časopis za suvremenu povijest",
  ["cas zgodovino narodop"] = "Časopis za zgodovino in narodopisje. Review for history and ethnography",
  ["casa mujer"] = "Casa de la mujer",
  ["case (phila)"] = "CASE (Philadelphia, Pa.)",
  ["case comment"] = "Case and comment",
  ["case manager"] = "The Case manager",
  ["case orthop j"] = "Case orthopaedic journal",
  ["case rep anesthesiol"] = "Case reports in anesthesiology",
  ["case rep cardiol"] = "Case reports in cardiology",
  ["case rep child meml hosp chic"] = "Case reports. Children's Memorial Hospital (Chicago, Ill.)",
  ["case rep clin pathol"] = "Case reports in clinical pathology",
  ["case rep crit care"] = "Case reports in critical care",
  ["case rep dent"] = "Case reports in dentistry",
  ["case rep dermatol"] = "Case reports in dermatology",
  ["case rep dermatol med"] = "Case reports in dermatological medicine",
  ["case rep emerg med"] = "Case reports in emergency medicine",
  ["case rep endocrinol"] = "Case reports in endocrinology",
  ["case rep gastroenterol"] = "Case reports in gastroenterology",
  ["case rep gastrointest med"] = "Case reports in gastrointestinal medicine",
  ["case rep genet"] = "Case reports in genetics",
  ["case rep hematol"] = "Case reports in hematology",
  ["case rep infect dis"] = "Case reports in infectious diseases",
  ["case rep intern med"] = "Case reports in internal medicine",
  ["case rep med"] = "Case reports in medicine",
  ["case rep nephrol"] = "Case reports in nephrology",
  ["case rep nephrol dial"] = "Case reports in nephrology and dialysis",
  ["case rep nephrol urol"] = "Case reports in nephrology and urology",
  ["case rep neurol"] = "Case reports in neurology",
  ["case rep neurol med"] = "Case reports in neurological medicine",
  ["case rep obstet gynecol"] = "Case reports in obstetrics and gynecology",
  ["case rep oncol"] = "Case reports in oncology",
  ["case rep oncol med"] = "Case reports in oncological medicine",
  ["case rep ophthalmol"] = "Case reports in ophthalmology",
  ["case rep ophthalmol med"] = "Case reports in ophthalmological medicine",
  ["case rep orthop"] = "Case reports in orthopedics",
  ["case rep otolaryngol"] = "Case reports in otolaryngology",
  ["case rep pancreat cancer"] = "Case reports in pancreatic cancer",
  ["case rep pathol"] = "Case reports in pathology",
  ["case rep pediatr"] = "Case reports in pediatrics",
  ["case rep psychiatry"] = "Case reports in psychiatry",
  ["case rep pulmonol"] = "Case reports in pulmonology",
  ["case rep radiol"] = "Case reports in radiology",
  ["case rep rheumatol"] = "Case reports in rheumatology",
  ["case rep surg"] = "Case reports in surgery",
  ["case rep transplant"] = "Case reports in transplantation",
  ["case rep urol"] = "Case reports in urology",
  ["case rep vasc med"] = "Case reports in vascular medicine",
  ["case rep vet med"] = "Case reports in veterinary medicine",
  ["case rep womens health"] = "Case reports in women's health",
  ["case reports hepatol"] = "Case reports in hepatology",
  ["case reports immunol"] = "Case reports in immunology",
  ["case reports plast surg hand surg"] = "Case reports in plastic surgery & hand surgery",
  ["case stud chem environ eng"] = "Case Studies in Chemical and Environmental Engineering",
  ["case stud constr mater"] = "Case Studies in Construction Materials",
  ["case stud eng fail anal"] = "Case Studies in Engineering Failure Analysis",
  ["case stud environ"] = "Case Studies in the Environment",
  ["case stud fire saf"] = "Case Studies in Fire Safety",
  ["case stud health adm"] = "Case studies in health administration",
  ["case stud mech syst signal process"] = "Case Studies in Mechanical Systems and Signal Processing",
  ["case stud nondestrtest eval"] = "Case Studies in Nondestructive Testing and Evaluation",
  ["case stud struct eng"] = "Case Studies in Structural Engineering",
  ["case stud surg"] = "Case studies in surgery",
  ["case stud therm eng"] = "Case Studies in Thermal Engineering",
  ["case stud transp policy"] = "Case Studies in Transport Policy",
  ["case studies bus ind gov stat"] = "Case studies in business, industry and government statistics : CSBIGS",
  ["case west reserve j int law"] = "Case Western Reserve journal of international law",
  ["case west reserve law rev"] = "Case Western Reserve law review",
  ["cases j"] = "Cases journal",
  ["cases public health commun mark"] = "Cases in public health communication and marketing",
  ["casopis hrvatskoga prirodoslovnog muzeja|nat croat"] = "Natura croatica : periodicum Musei Historiae Naturalis Croatici",
  ["casp j math sci"] = "Caspian Journal of Mathematical Sciences",
  ["casp j neurol sci"] = "Caspian journal of neurological sciences",
  ["caspian j intern med"] = "Caspian journal of internal medicine",
  ["cat classif q"] = "Cataloging & classification quarterly",
  ["catal commun"] = "Catalysis Communications",
  ["catal green chem eng"] = "Catalysis in Green Chemistry and Engineering",
  ["catal ind"] = "Catalysis in Industry",
  ["catal lett"] = "Catalysis Letters",
  ["catal letters"] = "Catalysis letters",
  ["catal rev sci eng"] = "Catalysis Reviews - Science and Engineering",
  ["catal rev: sci eng"] = "Catalysis Reviews: Science and Engineering",
  ["catal sci technol"] = "Catalysis Science & Technology",
  ["catal struct react"] = "Catalysis, Structure & Reactivity",
  ["catal surv asia"] = "Catalysis Surveys from Asia",
  ["catal surv jpn"] = "Catalysis Surveys from Japan",
  ["catal sustainable energy"] = "Catalysis for Sustainable Energy",
  ["catal today"] = "Catalysis Today",
  ["catedra clin"] = "Cátedra y clínica",
  ["categ gen algebr struct appl"] = "Categories and General Algebraic Structures with Applications",
  ["catena (amst)"] = "Catena",
  ["catena suppl"] = "Catena supplement",
  ["cater health"] = "Catering & health",
  ["cathet cardiovasc diagn"] = "Catheterization and cardiovascular diagnosis",
  ["catheter cardiovasc interv"] = "Catheterization and cardiovascular interventions : official journal of the Society for Cardiac Angiography & Interventions",
  ["cathol hist rev"] = "The Catholic historical review",
  ["cathol hosp"] = "Catholic hospital",
  ["cathol lawyer"] = "The Catholic lawyer",
  ["cathol med q"] = "The Catholic medical quarterly",
  ["cathol mind"] = "The Catholic mind",
  ["cathol nurse"] = "The Catholic nurse",
  ["cathol stand"] = "Catholic standard (Washington, D.C.)",
  ["cathol univers law rev"] = "Catholic University law review",
  ["cathol update"] = "Catholic update",
  ["cathol woman"] = "Catholic woman",
  ["cathol world"] = "Catholic world",
  ["cato j"] = "The Cato journal",
  ["catrina int j environ sci"] = "Catrina : the international journal of environmental sciences",
  ["cattle pract"] = "Cattle practice : journal of the British Cattle Veterinary Association",
  ["cave res"] = "Cave research",
  ["cbe life sci educ"] = "CBE life sciences education",
  ["cbe views"] = "CBE views : official publication of the Council of Biology Editors",
  ["cbms issues math ed"] = "CBMS Issues in Mathematics Education",
  ["cbms reg conf ser math"] = "CBMS Regional Conference Series in Mathematics",
  ["cbms regional conf ser in math"] = "CBMS Regional Conference Series in Mathematics",
  ["cbms-nsf regional conf ser in appl math"] = "CBMS-NSF Regional Conference Series in Applied Mathematics",
  ["cc ai"] = "CC AI",
  ["ccar j"] = "CCAR journal (New York, N.Y. : 1991)",
  ["ccast wl sw"] = "Surface Physics",
  ["ccf trans high perform comput"] = "CCF Transactions on High Performance Computing",
  ["ccf trans networking"] = "CCF Transactions on Networking",
  ["ccf trans pervasive comput interact"] = "CCF Transactions on Pervasive Computing and Interaction",
  ["ccica annual"] = "CCICA annual : publication of the Catholic Commission on Intellectual and Cultural Affairs",
  ["ccl family found"] = "CCL family foundations",
  ["ccl news"] = "The CCL news",
  ["ccs chem"] = "CCS Chemistry",
  ["cda j"] = "CDA journal",
  ["cdc aids wkly"] = "CDC AIDS weekly",
  ["cdc bull"] = "CDC bulletin. Communicable Disease Center (U.S.)",
  ["cdc nchstp dly news update"] = "CDC NCHSTP daily news update",
  ["cde work pap"] = "CDE working paper. University of Wisconsin--Madison. Center for Demography and Ecology",
  ["cdr (lond engl rev)"] = "CDR (London, England : Review)",
  ["cdr (lond engl wkly)"] = "CDR (London, England : Weekly)",
  ["cdr wkly (online)"] = "CDR weekly : communicable disease report",
  ["cds rev"] = "CDS review",
  ["cdt dig"] = "CDT digest",
  ["ce focus"] = "Ce Focus",
  ["cea13 (2013)"] = "CEA'13 : proceedings of the 5th International Workshop on Multimedia for Cooking & Eating Activities : October 21, 2013, Barcelona, Spain. Workshop on Multimedia for Cooking and Eating Activities (5th : 2013 : Barcelona, Spain)",
  ["ceara med"] = "Ceará médico : orgão do Centro Médico Cearense",
  ["ceas aeronaut j"] = "CEAS aeronautical journal",
  ["ceas space j"] = "CEAS Space Journal",
  ["cedpa netw"] = "CEDPA network (Online)",
  ["cedpa world wide"] = "CEDPA world wide. Centre for Development and Population Activities",
  ["cejor cent eur j oper res"] = "Central European Journal of Operations Research (CEJOR)",
  ["celebr stud"] = "Celebrity studies",
  ["celest mech dyn astron"] = "Celestial mechanics and dynamical astronomy",
  ["celestial mech dyn astron"] = "Celestial Mechanics and Dynamic Astronomy",
  ["celestial mech dynam astronom"] = "Celestial Mechanics & Dynamical Astronomy. An International Journal of Space Dynamics",
  ["cell (cambridge, ma, u s)"] = "Cell (Cambridge, MA, United States)",
  ["cell adh migr"] = "Cell adhesion & migration",
  ["cell adhes commun"] = "Cell adhesion and communication",
  ["cell adhes migr"] = "Cell Adhesion & Migration",
  ["cell biochem biophys"] = "Cell biochemistry and biophysics",
  ["cell biochem funct"] = "Cell Biochemistry and Function",
  ["cell biol (henderson, nv)"] = "Cell biology : research & therapy",
  ["cell biol educ"] = "Cell biology education",
  ["cell biol int"] = "Cell biology international",
  ["cell biol int rep"] = "Cell biology international reports",
  ["cell biol rev"] = "Cell biology reviews : CBR",
  ["cell biol toxicol"] = "Cell biology and toxicology",
  ["cell biophys"] = "Cell biophysics",
  ["cell biosci"] = "Cell & bioscience",
  ["cell calcium"] = "Cell calcium",
  ["cell chem biol"] = "Cell chemical biology",
  ["cell chromosome"] = "Cell & chromosome",
  ["cell commun adhes"] = "Cell communication & adhesion",
  ["cell commun insights"] = "Cell communication insights",
  ["cell commun signal"] = "Cell communication and signaling : CCS",
  ["cell commun signaling"] = "Cell Communication and Signaling",
  ["cell cycle"] = "Cell cycle (Georgetown, Tex.)",
  ["cell death differ"] = "Cell death and differentiation",
  ["cell death dis"] = "Cell death & disease",
  ["cell death discov"] = "Cell death discovery",
  ["cell death discovery"] = "Cell Death Discovery",
  ["cell dev biol"] = "Cell & developmental biology",
  ["cell differ"] = "Cell differentiation",
  ["cell differ dev"] = "Cell Differentiation and Development",
  ["cell discov"] = "Cell discovery",
  ["cell discovery"] = "Cell Discovery",
  ["cell div"] = "Cell division",
  ["cell gene ther insights"] = "Cell & gene therapy insights",
  ["cell genomics"] = "Cell Genomics",
  ["cell growth differ"] = "Cell growth & differentiation : the molecular biology journal of the American Association for Cancer Research",
  ["cell health cytoskelet"] = "Cell health and cytoskeleton",
  ["cell host microbe"] = "Cell host & microbe",
  ["cell immunol"] = "Cellular immunology",
  ["cell j"] = "Cell journal",
  ["cell logist"] = "Cellular logistics",
  ["cell med"] = "Cell medicine",
  ["cell metab"] = "Cell metabolism",
  ["cell microbiol"] = "Cellular microbiology",
  ["cell mol bioeng"] = "Cellular and molecular bioengineering",
  ["cell mol biol"] = "Cellular and molecular biology",
  ["cell mol biol (noisy-le-grand)"] = "Cellular and molecular biology (Noisy-le-Grand, France)",
  ["cell mol biol (omics)"] = "Cellular and molecular biology (OMICS)",
  ["cell mol biol (paris, fr, online)"] = "Cellular and Molecular Biology (Paris, France, Online)",
  ["cell mol biol (paris, fr, print)"] = "Cellular and Molecular Biology (Paris, France, Print)",
  ["cell mol biol (sarreguemines, fr, online)"] = "Cellular and Molecular Biology (Sarreguemines, France, Online)",
  ["cell mol biol (sarreguemines, fr, print)"] = "Cellular and Molecular Biology (Sarreguemines, France, Print)",
  ["cell mol biol incl cyto enzymol"] = "Cellular and molecular biology, including cyto-enzymology",
  ["cell mol biol lett"] = "Cellular & molecular biology letters",
  ["cell mol biol rep"] = "Cell and Molecular Biology Reports",
  ["cell mol biol res"] = "Cellular & molecular biology research",
  ["cell mol gastroenterol hepatol"] = "Cellular and molecular gastroenterology and hepatology",
  ["cell mol immunol"] = "Cellular & molecular immunology",
  ["cell mol life sci"] = "Cellular and Molecular Life Sciences",
  ["cell mol med open access"] = "Cellular & molecular medicine: open access",
  ["cell mol neurobiol"] = "Cellular and molecular neurobiology",
  ["cell motil"] = "Cell motility",
  ["cell motil cytoskeleton"] = "Cell motility and the cytoskeleton",
  ["cell muscle motil"] = "Cell and muscle motility",
  ["cell oncol"] = "Cellular oncology : the official journal of the International Society for Cellular Oncology",
  ["cell oncol (dordr)"] = "Cellular oncology (Dordrecht)",
  ["cell pharmacol"] = "Cellular pharmacology",
  ["cell physiol biochem"] = "Cellular physiology and biochemistry : international journal of experimental cellular physiology, biochemistry, and pharmacology",
  ["cell polym"] = "Cellular Polymers",
  ["cell preserv technol"] = "Cell preservation technology",
  ["cell prolif"] = "Cell proliferation",
  ["cell proliferation"] = "Cell Proliferation",
  ["cell regen"] = "Cell regeneration (London, England)",
  ["cell regener"] = "Cell Regeneration",
  ["cell regul"] = "Cell regulation",
  ["cell rep"] = "Cell reports",
  ["cell rep med"] = "Cell reports. Medicine",
  ["cell rep methods"] = "Cell Reports Methods",
  ["cell rep phys sci"] = "Cell reports. Physical science",
  ["cell reprogram"] = "Cellular reprogramming",
  ["cell reprogramming"] = "Cellular Reprogramming",
  ["cell res"] = "Cell research",
  ["cell signal"] = "Cellular signalling",
  ["cell signalling"] = "Cellular Signalling",
  ["cell stem cell"] = "Cell stem cell",
  ["cell stem cells regen med"] = "Cell, stem cells and regenerative medicine",
  ["cell stress"] = "Cell stress",
  ["cell stress chaperones"] = "Cell stress & chaperones",
  ["cell struct funct"] = "Cell structure and function",
  ["cell surf"] = "Cell surface (Amsterdam, Netherlands)",
  ["cell syst"] = "Cell systems",
  ["cell ther transplant"] = "Cellular therapy and transplantation",
  ["cell tissue bank"] = "Cell and tissue banking",
  ["cell tissue banking"] = "Cell and Tissue Banking",
  ["cell tissue biol"] = "Cell and tissue biology",
  ["cell tissue kinet"] = "Cell and tissue kinetics",
  ["cell tissue res"] = "Cell and tissue research",
  ["cell transplant"] = "Cell transplantation",
  ["cell vis"] = "Cell vision : the journal of analytical morphology",
  ["cellbio (irvine, calif)"] = "CellBio",
  ["cellr4 repair replace regen reprogram"] = "CellR4-- repair, replacement, regeneration, & reprogramming",
  ["cells dev"] = "Cells and Development",
  ["cells tissues organs"] = "Cells, tissues, organs",
  ["cellul chem technol"] = "Cellulose Chemistry and Technology",
  ["cellulose (lond)"] = "Cellulose (London, England)",
  ["cem concr compos"] = "Cement and Concrete Composites",
  ["cem concr res"] = "Cement and Concrete Research",
  ["cemla bol mens"] = "CEMLA Boletin Mensual",
  ["cempi ser"] = "CEMPI Series",
  ["cen case rep"] = "CEN case reports",
  ["cent afr j med"] = "The Central African journal of medicine",
  ["cent asia monit"] = "Central Asia monitor",
  ["cent asian j glob health"] = "Central Asian journal of global health",
  ["cent call"] = "Centre calling",
  ["cent est med"] = "Le Centre-est médical",
  ["cent estud recur odontol nino"] = "Centro de Estudios de Recursos Odontológicos para el Niño : CERON",
  ["cent eur for"] = "Central European Forests",
  ["cent eur geol"] = "Central European Geology",
  ["cent eur hist"] = "Central European history",
  ["cent eur j biol"] = "Central European journal of biology",
  ["cent eur j chem"] = "Central European Journal of Chemistry",
  ["cent eur j comput sci"] = "Central European Journal of Computer Science",
  ["cent eur j energetic mater"] = "Central European Journal of Energetic Materials",
  ["cent eur j eng"] = "Central European Journal of Engineering",
  ["cent eur j geosci"] = "Central European Journal of Geosciences",
  ["cent eur j immunol"] = "Central-European journal of immunology",
  ["cent eur j math"] = "Central European Journal of Mathematics",
  ["cent eur j med"] = "Central European journal of medicine",
  ["cent eur j oper res"] = "Central European journal of operations research",
  ["cent eur j phys"] = "Central European Journal of Physics",
  ["cent eur j public health"] = "Central European journal of public health",
  ["cent eur j sport sci med"] = "Central European Journal of Sport Sciences and Medicine",
  ["cent eur neurosurg"] = "Central European neurosurgery",
  ["cent eur oper res"] = "Central European Journal of Operations Research",
  ["cent european j urol"] = "Central European journal of urology",
  ["cent issues anthropol"] = "Central issues in anthropology : a journal of the Central States Anthropological Society",
  ["cent mag"] = "The Center magazine",
  ["cent med"] = "Centre médical",
  ["cent nerv syst agents med chem"] = "Central nervous system agents in medicinal chemistry",
  ["cent nerv syst trauma"] = "Central nervous system trauma : journal of the American Paralysis Association",
  ["cent phys les houches"] = "Centre de Physique Les Houches",
  ["cent states speech j"] = "Central states speech journal",
  ["cent view"] = "Center view : a publication of the Center for Applied and Professional Ethics, University of Tennessee, Knoxville",
  ["centaur alpha kappa kappa"] = "The Centaur of Alpha Kappa Kappa",
  ["centaur lond engl"] = "Centaur (London, England)",
  ["centbl gesamte forstwes"] = "Centralblatt für das gesamte Forstwesen",
  ["centen rev"] = "The Centennial review",
  ["centr asian surv"] = "Central Asian survey",
  ["central europ j operations res"] = "Central European Journal of Operations Research",
  ["central europ j operations res econ"] = "Central European Journal for Operations Research and Economics",
  ["centre pure appl differential geom (padge)"] = "Centre for Pure and Applied Differential Geometry (PADGE)",
  ["centre rech math proc"] = "Centre de Recherches Mathematiques Proceedings",
  ["centropa (n y n y)"] = "Centropa : a journal of central European architecture and related arts",
  ["cepal rev"] = "CEPAL review",
  ["cephalalgia rep"] = "Cephalalgia reports",
  ["ceram eng sci proc"] = "Ceramic Engineering and Science Proceedings",
  ["ceram int"] = "Ceramics International",
  ["ceram silik"] = "Ceramics - Silikaty",
  ["ceram trans"] = "Ceramic Transactions",
  ["ceramurgia int"] = "Ceramurgia International",
  ["cereal chem"] = "Cereal Chemistry",
  ["cereal res commun"] = "Cereal Research Communications",
  ["cereb circ cogn behav"] = "Cerebral circulation - cognition and behavior",
  ["cereb cortex"] = "Cerebral cortex (New York, N.Y. : 1991)",
  ["cereb cortex comm"] = "Cerebral Cortex Communications",
  ["cereb cortex commun"] = "Cerebral cortex communications",
  ["cereb palsy bull"] = "Cerebral palsy bulletin",
  ["cereb palsy j"] = "The Cerebral palsy journal",
  ["cereb palsy rev"] = "Cerebral palsy review",
  ["cerebellum ataxias"] = "Cerebellum & ataxias",
  ["cerebrospinal fluid res"] = "Cerebrospinal fluid research",
  ["cerebrovasc brain metab rev"] = "Cerebrovascular and brain metabolism reviews",
  ["cerebrovasc dis"] = "Cerebrovascular diseases (Basel, Switzerland)",
  ["cerebrovasc dis extra"] = "Cerebrovascular diseases extra",
  ["cern ideasq j exp innov"] = "CERN ideaSquare journal of experimental innovation",
  ["cerrahpasa tip fak derg"] = "Cerrahpaşa Tip Fakültesi dergisi",
  ["certif dent tech"] = "Certified dental technician",
  ["certif milk"] = "Certified milk",
  ["cervello g nevrol"] = "Il Cervello; giornale di nevrologia",
  ["cervix low female genital tract"] = "The Cervix and the lower female genital tract",
  ["ces odontol"] = "CES odontología",
  ["ces trans electr mach syst"] = "CES Transactions on Electrical Machines and Systems",
  ["cesifo econ stud"] = "CESifo economic studies",
  ["cesk cas hist"] = "Československý časopis historický",
  ["cesk dermatol"] = "Ceskoslovenská dermatologie",
  ["cesk epidemiol mikrobiol imunol"] = "Ceskoslovenská epidemiologie, mikrobiologie, imunologie",
  ["cesk farm"] = "Ceskoslovenská farmacie",
  ["cesk fysiol"] = "Ceskoslovenska Fysiologie",
  ["cesk gastroenterol vyz"] = "Ceskoslovenská gastroenterologie a výz̆iva",
  ["cesk gynekol"] = "Ceskoslovenská gynekologie",
  ["cesk hyg"] = "Ceskoslovenská hygiena",
  ["cesk morfol"] = "Ceskoslovenská morfologie",
  ["cesk neurol"] = "Ceskoslovenská neurologie",
  ["cesk neurol neurochir"] = "Ceskoslovenská neurologie a neurochirurgie",
  ["cesk oftalmol"] = "Ceskoslovenská oftalmologie",
  ["cesk onkol"] = "Ceskoslovenská onkológia",
  ["cesk otolaryngol"] = "Ceskoslovenská otolaryngologie",
  ["cesk patol"] = "Československá patologie",
  ["cesk pediatr"] = "Ceskoslovenská pediatrie",
  ["cesk psychiatr"] = "Ceskoslovenská psychiatrie",
  ["cesk psychol"] = "Ceskoslovenská psychologie",
  ["cesk radiol"] = "Ceskoslovenská radiologie",
  ["cesk rentgenol"] = "Ceskoslovenská rentgenologie",
  ["cesk slov oftalmol"] = "Ceská a slovenská oftalmologie : casopis Ceské oftalmologické spolecnosti a Slovenské oftalmologické spolecnosti",
  ["cesk stomatol"] = "Ceskoslovenská stomatologie",
  ["cesk zdrav"] = "Ceskoslovenské zdravotnictví",
  ["ceska gynekol"] = "Ceská gynekologie",
  ["ceska slov farm"] = "Ceská a Slovenská farmacie : casopis Ceské farmaceutické spolecnosti a Slovenské farmaceutické spolecnosti",
  ["ceska slov psychiatr"] = "Ceská a slovenská psychiatrie",
  ["ceskoslov biol"] = "Ceskoslovenská biologie",
  ["ceskoslov dent"] = "Ceskoslovenský dentista",
  ["ceskoslov nemocnice"] = "Ceskoslovenská nemocnice",
  ["cesra saule"] = "Cesra Säule",
  ["ceur workshop proc"] = "CEUR workshop proceedings",
  ["cex rep civ eff exerc"] = "CEX [reports]; civil effects exercise. U.S. Atomic Energy Commission",
  ["ceylon dent j"] = "Ceylon Dental Journal",
  ["ceylon for"] = "Ceylon Forester",
  ["ceylon j sci biol sci"] = "Ceylon journal of science. Biological sciences",
  ["ceylon med j"] = "The Ceylon medical journal",
  ["cfbs newsl"] = "CFBS newsletter = Bulletin de nouvelles de la FC-SB",
  ["cfw plex"] = "CFW plexus",
  ["cha insight"] = "CHA insight",
  ["cha ye ke xue"] = "Cha ye ke xue = Journal of tea science",
  ["chac rev"] = "C.H.A.C. review",
  ["chaeoe hanin yon gu"] = "Chaeoe Hanin yŏn'gu = Studies of Koreans abroad",
  ["chain drug rev"] = "Chain drug review",
  ["chall adv comput chem phys"] = "Challenges and Advances in Computational Chemistry and Physics",
  ["challenge (atlanta ga)"] = "Challenge (Atlanta, Ga.)",
  ["challenge (white plains)"] = "Challenge; the magazine of economic affairs",
  ["challenges sustainability"] = "Challenges in Sustainability",
  ["chamchuri j math"] = "Chamchuri Journal of Mathematics",
  ["chance (n y)"] = "Chance (New York, N.Y.)",
  ["chang gaozhuan xuebao"] = "Changshu Gao-Zhuan xuebao = Journal of Changshu College",
  ["chang gung med j"] = "Chang Gung medical journal",
  ["chang jiang da xue xue bao zi ran ke xue ban"] = "Chang jiang da xue xue bao. Zi ran ke xue ban = Journal of Yangtze University. Natural science edition",
  ["chang keng i hsueh"] = "Chang Keng I Hsueh Chang Gung Medical Journal",
  ["chang men"] = "Changing men (Madison, Wis.)",
  ["change adapt socio-ecol syst"] = "Change and Adaptation in Socio-Ecological Systems",
  ["changgeng yi xue za zhi"] = "Changgeng yi xue za zhi",
  ["changing times"] = "Changing times",
  ["channels (austin)"] = "Channels (Austin, Tex.)",
  ["chaos complex lett"] = "Chaos and complexity letters",
  ["chaos complexity lett"] = "Chaos and Complexity Letters",
  ["chaos soliton fract"] = "Chaos, Solitons & Fractals",
  ["chaos solitons fractals"] = "Chaos, solitons, and fractals",
  ["chaos, solitons  fractals"] = "Chaos, Solitons & Fractals",
  ["chaos, solitons  fractals:x"] = "Chaos, Solitons & Fractals:X",
  ["chaos, solitons fractals"] = "Chaos, Solitons & Fractals",
  ["chapman & hall probab ser"] = "Chapman & Hall Probability Series",
  ["chapman & hall/crc big data ser"] = "Chapman & Hall/CRC Big Data Series",
  ["chapman & hall/crc biostat ser"] = "Chapman & Hall/CRC Biostatistics Series",
  ["chapman & hall/crc comput biol ser"] = "Chapman & Hall/CRC Computational Biology Series",
  ["chapman & hall/crc comput inf sci ser"] = "Chapman & Hall/CRC Computer and Information Science Series",
  ["chapman & hall/crc comput sci data anal ser"] = "Chapman & Hall/CRC Computer Science and Data Analysis Series",
  ["chapman & hall/crc comput sci ser"] = "Chapman & Hall/CRC Computational Science Series",
  ["chapman & hall/crc cryptogr netw secur"] = "Chapman & Hall/CRC Cryptography and Network Security",
  ["chapman & hall/crc data min knowl discov ser"] = "Chapman & Hall/CRC Data Mining and Knowledge Discovery Series",
  ["chapman & hall/crc financ math ser"] = "Chapman & Hall/CRC Financial Mathematics Series",
  ["chapman & hall/crc handb mod stat methods"] = "Chapman & Hall/CRC Handbooks of Modern Statistical Methods",
  ["chapman & hall/crc interdiscip stat ser"] = "Chapman & Hall/CRC Interdisciplinary Statistics Series",
  ["chapman & hall/crc mach learn pattern recogn ser"] = "Chapman & Hall/CRC Machine Learning & Pattern Recognition Series",
  ["chapman & hall/crc math"] = "Chapman & Hall/CRC Mathematics",
  ["chapman & hall/crc math comput biol ser"] = "Chapman & Hall/CRC Mathematical and Computational Biology Series",
  ["chapman & hall/crc math comput imaging sci"] = "Chapman & Hall/CRC Mathematical and Computational Imaging Sciences",
  ["chapman & hall/crc numer anal sci comput"] = "Chapman & Hall/CRC Numerical Analysis and Scientific Computing",
  ["chapman & hall/crc ser actuar sci"] = "Chapman & Hall/CRC Series in Actuarial Science",
  ["chapman & hall/crc ser oper res"] = "Chapman & Hall/CRC Series in Operations Research",
  ["chapman & hall/crc texts stat sci ser"] = "Chapman & Hall/CRC Texts in Statistical Science Series",
  ["chapman and hall math ser"] = "Chapman and Hall Mathematics Series",
  ["chapman hall crc handb mod stat methods"] = "Chapman & Hall/CRC handbooks of modern statistical methods",
  ["chapman hall crc math comput biol ser"] = "Chapman & Hall/CRC mathematical & computational biology series",
  ["chapman hall/crc math biol ser"] = "Chapman & Hall/CRC Mathematical Biology Series",
  ["chapman math notes"] = "Chapman Mathematical Notes",
  ["character pers"] = "Character and personality",
  ["charadrius (bonn)"] = "Charadrius : Zeitschrift für Vogelkunde, Vogelschutz und Naturschutz in Rheinland und in Westfalen",
  ["chartwell-bratt student text"] = "A Chartwell-Bratt Student Text",
  ["chaucer rev"] = "The Chaucer review",
  ["chebyshevskii sb"] = "Chebyshevskii Sbornik",
  ["check list"] = "Check list (Luís Felipe Toledo)",
  ["chekh fiziol"] = "Chekhoslovatskaia fiziologiia",
  ["chekhoslovatskaia biol"] = "Chekhoslovatskaia biologiia",
  ["chelonian conserv biol"] = "Chelonian conservation and biology : journal of the IUCN/SSC Tortoise and Freshwater Turtle Specialist Group and international bulletin of chelonian research",
  ["chelyab fiz-mat zh"] = "Chelyabinskii Fiziko-Matematicheskii Zhurnal",
  ["chem - asian j"] = "Chemistry - An Asian Journal",
  ["chem - eur j"] = "Chemistry - A European Journal",
  ["chem - methods"] = "Chemistry - Methods",
  ["chem afr"] = "Chemistry Africa",
  ["chem anal"] = "Chemist-Analyst",
  ["chem anal (warsaw)"] = "Chemia Analityczna",
  ["chem anal (warsaw, pol)"] = "Chemia Analityczna (Warsaw, Poland)",
  ["chem analityczna"] = "Chemia analityczna",
  ["chem anlagen verfahren"] = "Chemie Anlagen und Verfahren",
  ["chem asian j"] = "Chemistry, an Asian journal",
  ["chem aust"] = "Chemistry in Australia",
  ["chem ber"] = "Chemische Berichte",
  ["chem ber recl"] = "Chemische Berichte-Recueil",
  ["chem biochem eng q"] = "Chemical and Biochemical Engineering Quarterly",
  ["chem biodivers"] = "Chemistry & biodiversity",
  ["chem biodiversity"] = "Chemistry & Biodiversity",
  ["chem biol"] = "Chemistry & biology",
  ["chem biol (cambridge, ma, u s)"] = "Chemistry & Biology (Cambridge, MA, United States)",
  ["chem biol (oxford, u k)"] = "Chemistry & Biology (Oxford, United Kingdom)",
  ["chem biol drug des"] = "Chemical biology & drug design",
  ["chem biol interact"] = "Chemico-biological interactions",
  ["chem biol interface"] = "Chemistry & Biology Interface",
  ["chem biol technol agric"] = "Chemical and Biological Technologies in Agriculture",
  ["chem biomed imaging"] = "Chemical & Biomedical Imaging",
  ["chem br"] = "Chemistry in Britain",
  ["chem can"] = "Chemistry in Canada",
  ["chem catal"] = "Chem Catalysis",
  ["chem cent j"] = "Chemistry Central journal",
  ["chem chron"] = "Chemika Chronika",
  ["chem commun"] = "Chemical Communications",
  ["chem commun (camb)"] = "Chemical communications (Cambridge, England)",
  ["chem commun (cambridge, u k)"] = "Chemical Communications (Cambridge, United Kingdom)",
  ["chem commun camb engl"] = "Chemical Communications (Cambridge, England)",
  ["chem data collect"] = "Chemical Data Collections",
  ["chem depend"] = "Chemical dependencies",
  ["chem drug"] = "The Chemist and druggist",
  ["chem drug export rev"] = "The Chemist and druggist export review",
  ["chem ecol"] = "Chemistry and Ecology",
  ["chem educ"] = "Chemical Educator",
  ["chem educ res pract"] = "Chemical Education Research and Practice",
  ["chem educ res pract eur"] = "Chemical Education Research and Practice in Europe",
  ["chem eng (london)"] = "Chemical Engineer (London)",
  ["chem eng (n y)"] = "Chemical Engineering (New York)",
  ["chem eng (new york)"] = "Chemical Engineering (New York)",
  ["chem eng commun"] = "Chemical Engineering Communications",
  ["chem eng educ"] = "Chemical Engineering Education",
  ["chem eng j"] = "Chemical Engineering Journal",
  ["chem eng j (amsterdam, neth)"] = "Chemical Engineering Journal (Amsterdam, Netherlands)",
  ["chem eng j (lausanne)"] = "Chemical Engineering Journal (Lausanne)",
  ["chem eng j adv"] = "Chemical Engineering Journal Advances",
  ["chem eng news"] = "Chemical and Engineering News",
  ["chem eng process"] = "Chemical engineering and processing = Génie des procédés = Verfahrenstechnik",
  ["chem eng process process intensif"] = "Chemical Engineering and Processing: Process Intensification",
  ["chem eng prog"] = "Chemical Engineering Progress",
  ["chem eng res des"] = "Chemical Engineering Research and Design",
  ["chem eng sci"] = "Chemical Engineering Science",
  ["chem eng sci: x"] = "Chemical Engineering Science: X",
  ["chem eng ser"] = "Chemical Engineering Series",
  ["chem eng technol"] = "Chemical Engineering and Technology",
  ["chem eng trans"] = "Chemical engineering transactions",
  ["chem erde"] = "Chemie der Erde - Geochemistry",
  ["chem eur j"] = "Chemistry A European Journal",
  ["chem express"] = "Chemistry Express",
  ["chem fibers int"] = "Chemical Fibers International",
  ["chem geol"] = "Chemical Geology",
  ["chem health saf"] = "Chemical Health and Safety",
  ["chem herit"] = "Chemical heritage",
  ["chem heterocycl compd"] = "Chemistry of Heterocyclic Compounds",
  ["chem heterocycl compd (n y)"] = "Chemistry of heterocyclic compounds",
  ["chem heterocycl compd (n y, ny, u s)"] = "Chemistry of Heterocyclic Compounds (New York, NY, United States)",
  ["chem immunol"] = "Chemical immunology",
  ["chem immunol allergy"] = "Chemical immunology and allergy",
  ["chem ind"] = "Chemistry & industry",
  ["chem ind (chichester, u k)"] = "Chemistry & Industry (Chichester, United Kingdom)",
  ["chem ind (duesseldorf)"] = "Chemische Industrie (Duesseldorf)",
  ["chem ind (london)"] = "Chemistry and Industry",
  ["chem ind (london, u k)"] = "Chemistry & Industry (London, United Kingdom)",
  ["chem ind chem eng q"] = "Chemical Industry & Chemical Engineering Quarterly",
  ["chem inform"] = "Chemical informatics (Wilmington, Del.)",
  ["chem ing tech"] = "Chemie Ingenieur Technik",
  ["chem int"] = "Chemistry International",
  ["chem j armen"] = "Chemical journal of Armenia",
  ["chem j mold"] = "Chemistry Journal of Moldova",
  ["chem lett"] = "Chemistry Letters",
  ["chem lett (jpn)"] = "Chemistry letters",
  ["chem listy"] = "Chemicke Listy",
  ["chem mater"] = "Chemistry of Materials",
  ["chem methods"] = "Chemistry Methods",
  ["chem modif surf symp chem modif surf"] = "Chemically modified surfaces : recent developments. Symposium on Chemically Modified Surfaces (6th : 1995 : San Jose, Calif.)",
  ["chem nat compd"] = "Chemistry of Natural Compounds",
  ["chem pap"] = "Chemical Papers",
  ["chem pap - chem zvesti"] = "Chemical Papers - Chemicke Zvesti",
  ["chem pet eng"] = "Chemical and Petroleum Engineering",
  ["chem pharm bull"] = "Chemical and Pharmaceutical Bulletin",
  ["chem pharm bull (tokyo)"] = "Chemical & pharmaceutical bulletin",
  ["chem phys"] = "Chemical Physics",
  ["chem phys carbon"] = "Chemistry and Physics of Carbon",
  ["chem phys impact"] = "Chemical Physics Impact",
  ["chem phys lett"] = "Chemical Physics Letters",
  ["chem phys lett: x"] = "Chemical Physics Letters: X",
  ["chem phys lipids"] = "Chemistry and Physics of Lipids",
  ["chem phys processes combust"] = "Chemical and Physical Processes in Combustion",
  ["chem phys res j"] = "Chemical Physics Research Journal",
  ["chem phys rev"] = "Chemical Physics Reviews",
  ["chem process"] = "Chemical Processing",
  ["chem prod process model"] = "Chemical Product and Process Modeling",
  ["chem prum"] = "Chemicky Prumysl",
  ["chem rec"] = "Chemical record (New York, N.Y.)",
  ["chem res chin univ"] = "Chemical Research in Chinese Universities",
  ["chem res toxicol"] = "Chemical research in toxicology",
  ["chem rev"] = "Chemical reviews",
  ["chem rev (washington, dc, u s)"] = "Chemical Reviews (Washington, DC, United States)",
  ["chem rev fett- harz-ind"] = "Chemische Revue uber die Fett- und Harz-Industrie",
  ["chem sci"] = "Chemical science",
  ["chem sci j"] = "Chemical sciences journal",
  ["chem sci trans"] = "Chemical Science Transactions",
  ["chem scr"] = "Chemica scripta",
  ["chem senses"] = "Chemical senses",
  ["chem sin"] = "Chemica Sinica",
  ["chem soc rev"] = "Chemical Society reviews",
  ["chem speciation bioavailability"] = "Chemical Speciation and Bioavailability",
  ["chem sq"] = "Chemistry squared",
  ["chem teach int"] = "Chemistry Teacher International : best practices in chemistry education",
  ["chem teach int best pract chem educ"] = "Chemistry Teacher International: Best Practices in Chemistry Education",
  ["chem tech"] = "ChemTech",
  ["chem tech (leipzig)"] = "Chemische Technik",
  ["chem technol (kaunas, lith)"] = "Chemine Technologija (Kaunas, Lithuania)",
  ["chem technol fuels oils"] = "Chemistry and Technology of Fuels and Oils",
  ["chem thermodyn therm anal"] = "Chemical Thermodynamics and Thermal Analysis",
  ["chem umsch geb fette, oele, wachse harze"] = "Chemische Umschau auf dem Gebiete der Fette, Oele, Wachse und Harze",
  ["chem unserer zeit"] = "Chemie in unserer Zeit",
  ["chem vap deposition"] = "Chemical Vapor Deposition",
  ["chem week"] = "Chemical week",
  ["chem weekbl"] = "Chemisch weekblad",
  ["chem zentralbl"] = "Chemisches Zentralblatt",
  ["chem ztg"] = "Chemiker-Zeitung",
  ["chem zvesti"] = "Chemické zvesti",
  ["chem-bio inf j"] = "Chem-Bio Informatics Journal",
  ["chem-biol interact"] = "Chemico-Biological Interactions",
  ["chembioeng rev"] = "ChemBioEng Reviews",
  ["chemistry (easton)"] = "Chemistry",
  ["chemistry (rajkot, india)"] = "Chemistry (Rajkot, India)",
  ["chemom intell lab syst"] = "Chemometrics and Intelligent Laboratory Systems",
  ["chemometr intell lab syst"] = "Chemometrics and intelligent laboratory systems : an international journal sponsored by the Chemometrics Society",
  ["chemometrics res stud ser"] = "Chemometrics Research Studies Series",
  ["chemosens percept"] = "Chemosensory Perception",
  ["chemosensors (basel)"] = "Chemosensors (Basel, Switzerland)",
  ["chemother res pract"] = "Chemotherapy research and practice",
  ["chemotherapia (basel)"] = "Chemotherapia",
  ["chemotherapy (basel, switz)"] = "Chemotherapy (Basel, Switzerland)",
  ["chemotherapy (tokyo)"] = "Chemotherapy (Tokyo)",
  ["chemtracts: inorg chem"] = "Chemtracts: Inorganic Chemistry",
  ["chemtracts: macromol chem"] = "Chemtracts: Macromolecular Chemistry",
  ["chemtracts: org chem"] = "Chemtracts: Organic Chemistry",
  ["chem–asian j"] = "Chemistry–An Asian Journal",
  ["chem–eur j"] = "Chemistry–A European Journal",
  ["chen tzu yen chiu"] = "Chen Tzu Yen Chiu (Acupuncture Research)",
  ["cheng shi she ji (2015)"] = "Cheng shi she ji (2015)",
  ["chest surg clin n am"] = "Chest surgery clinics of North America",
  ["chesterton rev"] = "The Chesterton review",
  ["chic hist"] = "Chicago history",
  ["chic j int law"] = "Chicago journal of international law",
  ["chic j theoret comput sci"] = "Chicago Journal of Theoretical Computer Science",
  ["chic kent law rev"] = "Chicago-Kent law review",
  ["chic med"] = "Chicago medicine",
  ["chic med rec"] = "The Chicago medical recorder",
  ["chic med sch q"] = "The Chicago Medical School quarterly",
  ["chic pol rev"] = "Chicago policy review",
  ["chic stud"] = "Chicago studies",
  ["chic today"] = "Chicago today (University of Chicago)",
  ["chicago bar rec"] = "Chicago bar record",
  ["chicago j theoret comput sci"] = "Chicago Journal of Theoretical Computer Science",
  ["chicago lectures in math"] = "Chicago Lectures in Mathematics",
  ["chicago lectures phys"] = "Chicago Lectures in Physics",
  ["chicano law rev"] = "Chicano law review",
  ["chief inf off j"] = "Chief information officer journal",
  ["chieh fang chun i hsueh tsa chih"] = "Jie fang jun yi xue za zhi. [Liberation army medical journal]",
  ["chikusan sochi kenkyujo kenkyu hokoku"] = "Chikusan Sōchi Kenkyūjo kenkyū hōkoku",
  ["chikwa kijae hakhoe chi"] = "Chikwa Kijae Hakhoe Chi (Journal of the Korea Research Society for Dental Materials)",
  ["chil j agric anim sci"] = "Chilean journal of agricultural & animal sciences : ex Agro-ciencia",
  ["chil j agric res"] = "Chilean journal of agricultural research",
  ["chil j stat"] = "Chilean journal of statistics",
  ["child abuse negl"] = "Child abuse & neglect",
  ["child abuse rev"] = "Child abuse review (Chichester, England : 1992)",
  ["child adolesc ment health"] = "Child and adolescent mental health",
  ["child adolesc obes"] = "Child and adolescent obesity (Abingdon, England)",
  ["child adolesc psychiatr clin n am"] = "Child and adolescent psychiatric clinics of North America",
  ["child adolesc psychiatry ment health"] = "Child and adolescent psychiatry and mental health",
  ["child adolesc psychopharmacol  news"] = "Child & adolescent psychopharmacology news",
  ["child adolesc social work j"] = "Child & adolescent social work journal : C & A",
  ["child care health dev"] = "Child: care, health and development",
  ["child care pract"] = "Child care in practice : Northern Ireland journal of multi-disciplinary child care practice",
  ["child care q"] = "Child care quarterly",
  ["child dev"] = "Child development",
  ["child dev perspect"] = "Child development perspectives",
  ["child dev res"] = "Child development research",
  ["child educ"] = "Childhood education",
  ["child fam"] = "Child and family",
  ["child fam behav ther"] = "Child & family behavior therapy",
  ["child fam law q"] = "Child and family law quarterly",
  ["child fam policy pract rev"] = "Child and family policy and practice review",
  ["child fam soc work"] = "Child & family social work",
  ["child geogr"] = "Children's geographies",
  ["child health alert"] = "Child health alert",
  ["child health care"] = "Children's health care : journal of the Association for the Care of Children's Health",
  ["child health dialogue"] = "Child health dialogue",
  ["child indic res"] = "Child indicators research",
  ["child lang teach ther"] = "Child language teaching and therapy",
  ["child leg rights j"] = "Children's legal rights journal",
  ["child libr"] = "Children & libraries",
  ["child maltreat"] = "Child maltreatment",
  ["child nephrol urol"] = "Child nephrology and urology",
  ["child neurol open"] = "Child neurology open",
  ["child neuropsychol"] = "Child neuropsychology : a journal on normal and abnormal development in childhood and adolescence",
  ["child obes"] = "Childhood obesity (Print)",
  ["child psychiatry hum dev"] = "Child psychiatry and human development",
  ["child psychiatry q"] = "Child psychiatry quarterly",
  ["child psychol psychiatry review"] = "Child psychology & psychiatry review",
  ["child sch"] = "Children & schools",
  ["child soc"] = "Children & society",
  ["child stud asia pac context"] = "Child studies in Asia-Pacific context",
  ["child study"] = "Child study; a journal of parent education",
  ["child surviv action news"] = "Child survival action news",
  ["child teenagers"] = "Children and teenagers",
  ["child today"] = "Children today",
  ["child trop"] = "Children in the tropics",
  ["child welfare"] = "Child welfare",
  ["child worldw"] = "Children worldwide",
  ["child youth care forum"] = "Child & youth care forum",
  ["child youth environ"] = "Children, youth and environments",
  ["child youth serv"] = "Child & youth services",
  ["child youth serv rev"] = "Children and youth services review",
  ["children (basel)"] = "Children (Basel, Switzerland)",
  ["childs brain"] = "Child's brain",
  ["childs nerv syst"] = "Child's nervous system : ChNS : official journal of the International Society for Pediatric Neurosurgery",
  ["chilton’s food eng"] = "Chilton’s Food Eng.",
  ["chilwa kijae hakhoe chi"] = "Ch'ikwa Kijae Hakhoe chi. The Journal of the Korea Research Society for Dental Materials",
  ["chim acta turc"] = "Chimica Acta Turcica",
  ["chim actual"] = "Chimie Actualites",
  ["chim ind"] = "Chimica e l’Industria",
  ["chim ind (milan, italy)"] = "Chimica e l’Industria (Milan, Italy)",
  ["chim ind genie chim"] = "Chimie et Industrie, Genie Chimique",
  ["chim nouv"] = "Chimie Nouvelle",
  ["chim oggi"] = "Chimica Oggi",
  ["chim ther"] = "Chimica therapeutica",
  ["chimia (aarau)"] = "Chimia",
  ["chin ann math"] = "Chinese Annals of Mathematics",
  ["chin ann math ser b"] = "Chinese Annals of Mathematics, Series B",
  ["chin astron astrophy"] = "Chinese Astronomy and Astrophysics",
  ["chin birds"] = "Chinese birds : the international journal of ornithology",
  ["chin chem lett"] = "Chinese Chemical Letters",
  ["chin clin oncol"] = "Chinese clinical oncology",
  ["chin cult"] = "Chinese culture",
  ["chin econ stud"] = "Chinese economic studies",
  ["chin environ dev"] = "Chinese environment & development (Armonk, N.Y.)",
  ["chin geogr sci"] = "Chinese geographical science",
  ["chin ger j clin oncol"] = "Chinese-German Journal of Clinical Oncology",
  ["chin herb med"] = "Chinese herbal medicines",
  ["chin j  electr eng"] = "Chinese Journal of Electrical Engineering",
  ["chin j acad radiol"] = "Chinese journal of academic radiology",
  ["chin j aeronaut"] = "Chinese Journal of Aeronautics",
  ["chin j agric biotechnol"] = "Chinese Journal of Agricultural Biotechnology",
  ["chin j agric biotechol"] = "Chinese journal of agricultural biotechnology",
  ["chin j anal chem"] = "Chinese Journal of Analytical Chemistry",
  ["chin j anim infect dis"] = "Zhongguo dong wu chuan ran bing xue bao",
  ["chin j artif intell"] = "Chinese Journal of Artificial Intelligence",
  ["chin j astron astrophys"] = "Chinese Journal of Astronomy and Astrophysics",
  ["chin j atmos sci"] = "Chinese Journal of Atmospheric Sciences",
  ["chin j biochem biophys"] = "Chinese journal of biochemistry and biophysics = Acta biochimica et biophysica Sinica = Sheng wu hua xue yu sheng wu wu li xue bao",
  ["chin j biotechnol"] = "Chinese Journal of Biotechnology",
  ["chin j bot"] = "Chinese journal of botany",
  ["chin j cancer"] = "Chinese journal of cancer",
  ["chin j cancer res"] = "Chinese journal of cancer research = Chung-kuo yen cheng yen chiu",
  ["chin j catal"] = "Chinese Journal of Catalysis",
  ["chin j chem"] = "Chinese Journal of Chemistry",
  ["chin j chem eng"] = "Chinese Journal of Chemical Engineering",
  ["chin j chem phys"] = "Chinese Journal of Chemical Physics",
  ["chin j chromatogr"] = "Chinese Journal of Chromatography",
  ["chin j dent res"] = "The Chinese journal of dental research : the official journal of the Scientific Section of the Chinese Stomatological Association (CSA)",
  ["chin j dig dis"] = "Chinese journal of digestive diseases",
  ["chin j electron"] = "Chinese Journal of Electronics",
  ["chin j genet"] = "Chinese journal of genetics = Acta genetica Sinica",
  ["chin j geochem"] = "Chinese Journal of Geochemistry",
  ["chin j geophys"] = "Chinese Journal of Geophysics",
  ["chin j inorg chem"] = "Chinese Journal of Inorganic Chemistry",
  ["chin j integr med"] = "Chinese journal of integrative medicine",
  ["chin j mech eng"] = "Chinese Journal of Mechanical Engineering",
  ["chin j mech eng: addit manuf front"] = "Chinese Journal of Mechanical Engineering: Additive Manufacturing Frontiers",
  ["chin j met sci technol"] = "Chinese Journal of Metal Science & Technology",
  ["chin j nat med"] = "Chinese journal of natural medicines",
  ["chin j neurosci"] = "Zhongguo shen jing ke xue za zhi = Chinese journal of neuroscience",
  ["chin j nucl phys"] = "Chinese Journal of Nuclear Physics",
  ["chin j oceanol limnol"] = "Chinese Journal of Oceanology and Limnology",
  ["chin j org chem"] = "Chinese Journal of Organic Chemistry",
  ["chin j phys"] = "Chinese Journal of Physics",
  ["chin j physiol"] = "The Chinese journal of physiology",
  ["chin j polym sci"] = "Chinese Journal of Polymer Science",
  ["chin j popul resour environ"] = "Chinese Journal of Population, Resources and Environment",
  ["chin j popul sci"] = "Chinese journal of population science",
  ["chin j prev vet med"] = "Chinese journal of preventive veterinary medicine",
  ["chin j sociol"] = "Chinese journal of sociology",
  ["chin j struct chem"] = "Chinese Journal of Structural Chemistry",
  ["chin j traumatol"] = "Chinese journal of traumatology = Zhonghua chuang shang za zhi",
  ["chin j urban environ stud"] = "Chinese Journal of Urban and Environmental Studies",
  ["chin j vet sci"] = "Zhongguo shou yi xue bao = Chinese journal of veterinary science",
  ["chin j wood sci technol"] = "Chinese Journal of Wood Science and Technology",
  ["chin med"] = "Chinese medicine",
  ["chin med j"] = "Chinese medical journal (Peking, China : 1932)",
  ["chin med j (engl)"] = "Chinese medical journal",
  ["chin med sci j"] = "Chinese medical sciences journal = Chung-kuo i hsüeh k'o hsüeh tsa chih",
  ["chin neurosurg j"] = "Chinese neurosurgical journal",
  ["chin opt lett"] = "Chinese optics letters : COL",
  ["chin phys"] = "Chinese Physics",
  ["chin phys b"] = "Chinese Physics B",
  ["chin phys c"] = "Chinese Physics C",
  ["chin phys lett"] = "Chinese Physics Letters",
  ["chin rev trop med"] = "The Chinese review of tropical medicine",
  ["chin sci"] = "Chinese science",
  ["chin sci bull"] = "Chinese Science Bulletin",
  ["chin sociol anthropol"] = "Chinese sociology and anthropology",
  ["chin sociol rev"] = "Chinese sociological review",
  ["chin tob sci"] = "Chinese tobacco science",
  ["china (singap)"] = "China (National University of Singapore. East Asian Institute)",
  ["china aktuell"] = "China aktuell",
  ["china bus rev"] = "The China business review",
  ["china cdc wkly"] = "China CDC weekly",
  ["china center adv sci tech world lab sympos workshop proc"] = "China Center of Advanced Science and Technology (World Laboratory) Symposium/Workshop Proceedings",
  ["china commun"] = "China Communications",
  ["china econ rev"] = "China Economic Review",
  ["china economic j"] = "China economic journal",
  ["china environ sci"] = "China Environmental Science",
  ["china foundry"] = "China Foundry",
  ["china geol"] = "China Geology",
  ["china health rev"] = "China health review",
  ["china herbivores|zhongguo cao shi dong wu"] = "Zhongguo cao shi dong wu",
  ["china j"] = "China journal (Canberra, A.C.T.)",
  ["china law pract"] = "China law and practice",
  ["china natl j new gastroenterol"] = "China national journal of new gastroenterology = Zhongguo xin xiao hua bing xue za zhi",
  ["china newsl"] = "China newsletter",
  ["china ocean eng"] = "China Ocean Engineering",
  ["china popul dev stud"] = "China population and development studies",
  ["china popul headl"] = "China population headliners",
  ["china popul newsl"] = "China population newsletter",
  ["china popul res leads"] = "China population research leads",
  ["china popul res newsl"] = "China population research newsletter",
  ["china popul today"] = "China population today",
  ["china public health|zhongguo gong gong wei sheng"] = "Zhongguo gong gong wei sheng",
  ["china q"] = "The China quarterly",
  ["china quart"] = "China Quarterly",
  ["china reconstr"] = "China reconstructs",
  ["china rep"] = "China report (New Delhi, India)",
  ["china trop med"] = "Zhongguo re dai yi xue",
  ["chinas med"] = "China's medicine",
  ['chinese agricultural science bulletin / "zhongguo nong xue tong bao" bian ji bu|zhongguo nong xue tong bao'] = "Zhongguo nong xue tong bao",
  ["chinese ann hist sci tech"] = "Chinese Annals of History of Science and Technology",
  ["chinese ann math ser a"] = "Chinese Annals of Mathematics. Series A. Shuxue Niankan. A Ji",
  ["chinese ann math ser b"] = "Chinese Annals of Mathematics. Series B",
  ["chinese astronom astrophys"] = "Chinese Astronomy and Astrophysics",
  ["chinese biochemical journal / zhongguo sheng wu hua xue hui zhu ban|sheng wu hua hsueh tsa chih"] = "Sheng wu hua xue za zhi",
  ["chinese biodiversity|sheng wu duo yang xing"] = "Sheng wu duo yang xing",
  ["chinese econ stud"] = "Chinese Economic Studies",
  ["chinese economy"] = "Chinese Economy",
  ["chinese j appl mech"] = "Chinese Journal of Applied Mechanics",
  ["chinese j appl probab statist"] = "Chinese Journal of Applied Probability and Statistics. Yingyong Gailu Tongji",
  ["chinese j comput"] = "Chinese Journal of Computers. Jisuanji Xuebao",
  ["chinese j contemp math"] = "Chinese Journal of Contemporary Mathematics",
  ["chinese j numer math appl"] = "Chinese Journal of Numerical Mathematics and Applications",
  ["chinese j phys"] = "Chinese Journal of Physics",
  ["chinese journal of animal quaratine|zhongguo dongwu jianyi"] = "Zhongguo dongwu jianyi",
  ["chinese journal of antibiotics|zhongguo kang sheng su za zhi"] = "Zhongguo kang sheng su za zhi",
  ["chinese journal of applied and environmental biology / zhongguo ke xue yuan chengdu sheng wu yan jiu suo zhu ban|ying yong yu huan jing sheng wu xue bao"] = "Ying yong yu huan jing sheng wu xue bao",
  ["chinese journal of biochemistry and molecular biology|zhongguo sheng wu hua xue yu fen zi sheng wu xue bao"] = "Zhongguo sheng wu hua xue yu fen zi sheng wu xue bao",
  ["chinese journal of gastroenterology and hepatology|weichangbingxue he ganbingxue zazhi"] = "Weichangbingxue he ganbingxue zazhi",
  ["chinese journal of immunology|zhongguo mian yi xue za zhi"] = "Zhongguo mian yi xue za zhi",
  ["chinese journal of neuroscience|chin j neurosci"] = "Zhongguo shen jing ke xue za zhi",
  ["chinese journal of nosocomiology|zhonghua yi yuan gan ran xue za zhi"] = "Zhonghua yi yuan gan ran xue za zhi",
  ["chinese journal of oil crop sciences / zhongguo nong ye ke xue yuan you liao zuo wu yan jiu suo zhu ban|zhongguo you liao zuo wu xue bao"] = "Zhongguo you liao zuo wu xue bao",
  ["chinese journal of vaccines and immunization|zhongguo ji hua mian yi"] = "Zhongguo ji hua mian yi",
  ["chinese journal of veterinary medicine|zhongguo shou yi za zhi"] = "Zhongguo shou yi za zhi",
  ["chinese journal of veterinary science and technology|zhongguo shouyi ke-ji"] = "Zhongguo Shouyi Ke-ji",
  ["chinese journal of veterinary science|chin j vet sci"] = "Zhongguo shou yi xue bao",
  ["chinese journal of zoonoses|zhongguo ren shou gong huan bing za zhi"] = "Zhongguo ren shou gong huan bing za zhi",
  ["chinese phys lett"] = "Chinese Physics Letters",
  ["chinese quart j math"] = "Chinese Quarterly Journal of Mathematics. Shuxue Jikan",
  ["chinese sci bull"] = "Chinese Science Bulletin",
  ["chinese traditional and herbal drugs|zhong cao yao"] = "Zhong cao yao",
  ["chinese vegetables|zhongguo shu cai"] = "Zhongguo shu cai",
  ["chir dent fr"] = "Le Chirurgien-dentiste de France",
  ["chir forum exp klin forsch"] = "Chirurgisches Forum für experimentelle und klinische Forschung",
  ["chir ital"] = "Chirurgia italiana",
  ["chir main"] = "Chirurgie de la main",
  ["chir maxillofac plast"] = "Chirurgia maxillofacialis & plastica",
  ["chir narzadow ruchu ortop pol"] = "Chirurgia narzadów ruchu i ortopedia polska",
  ["chir organi mov"] = "La Chirurgia degli organi di movimento",
  ["chir patol sper"] = "Chirurgia e patologia sperimentale",
  ["chir pediatr"] = "Chirurgie pédiatrique",
  ["chir urol"] = "Chirurgia urologica",
  ["chirigaku hyoron"] = "Chirigaku hyōron",
  ["chirop rec"] = "Chiropody record",
  ["chiropodist (lond)"] = "Chiropodist (Society of Chiropodists)",
  ["chiropr hist"] = "Chiropractic history : the archives and journal of the Association for the History of Chiropractic",
  ["chiropr man therap"] = "Chiropractic & manual therapies",
  ["chiropr osteopat"] = "Chiropractic & osteopathy",
  ["chirurgia (bucur)"] = "Chirurgia (Bucharest, Romania : 1990)",
  ["chittagong univ stud part ii sci"] = "Chittagong University Studies",
  ["chittys law j"] = "Chitty's law journal",
  ["chnia geogr"] = "The China geographer",
  ["cho-sŏn in-min kong-hwa-kuk kwa-hak-wŏn t’ong-bo"] = "Cho-sŏn Min-ju-ju-ŭi In-min Kong-hwa-kuk Kwa-hak-wŏn",
  ["choesin uihak new med j"] = "[Ch'oesin ŭihak] The new medical journal",
  ["choice (middletown)"] = "Choice (Chicago, Ill.)",
  ["choices (ames)"] = "Choices : the magazine of food, farm, and resource issues",
  ["choices respir manage"] = "Choices in respiratory management",
  ["choix trav etrang"] = "Choix de travaux étrangers",
  ["chongqing shifan daxue xuebao zi ran ke xue ban"] = "Chongqing shi fan da xue xue bao. Zi ran ke xue ban",
  ["chongqing yi xue"] = "Chongqing yi xue",
  ["chongsin pogon kwa sahoe saop"] = "Chŏngsin pogŏn kwa sahoe saŏp",
  ["chongsonyonhak yongu"] = "Ch'ŏngsonyŏnhak yŏngu",
  ["chonnam med j"] = "Chonnam medical journal",
  ["chonnam uidae chapchi"] = "Chŏnnam Uidae chapchi",
  ["chornomorskyi bot zhurnal"] = "Chornomorsʹkyĭ botanichnyĭ z︠h︡urnal = Chornomorski botanical journal",
  ["chosen ibo"] = "Chosen i-bo. The Korean medical journal",
  ["choson uihak"] = "[Chosŏn ŭihak] [Korean medicine]",
  ["chot mai het kan phayaban"] = "Chot mai hēt kān phayābān",
  ["christ bioeth"] = "Christian bioethics",
  ["christ century"] = "Christian century (Chicago, Ill. : 1902)",
  ["christ crisis"] = "Christianity and crisis",
  ["christ hist (worcest pa)"] = "Christian history (Worcester, Pa. : Township)",
  ["christ hosp med bull"] = "Christ Hospital medical bulletin",
  ["christ med dent soc j"] = "Christian Medical Dental Society journal",
  ["christ med soc j"] = "Christian Medical Society journal",
  ["christ nurse (mysore)"] = "Christian nurse",
  ["christ nurse int"] = "Christian nurse international",
  ["christ sch rev"] = "Christian scholar's review",
  ["christ sci monitor"] = "Christian Science monitor (Boston, Mass. : 1983)",
  ["christ sci monitor (east ed)"] = "Christian Science monitor (Boston, Mass. : Eastern ed.)",
  ["christ soc action"] = "Christian social action (Washington, D.C.)",
  ["christ today"] = "Christianity today",
  ["christ world"] = "Christ to the world",
  ["christiana albertina kiel univ z"] = "Christiana Albertina; Kieler Universitaẗs-Zeitschrift",
  ["chromatogr res int"] = "Chromatography Research International",
  ["chromatogr rev"] = "Chromatographic reviews",
  ["chromatogr sci ser"] = "Chromatographic Science Series",
  ["chromatography (basel)"] = "Chromatography (Basel)",
  ["chromosom bot"] = "Chromosome botany",
  ["chromosome res"] = "Chromosome research : an international journal on the molecular, supramolecular and evolutionary aspects of chromosome biology",
  ["chron actual sedeis"] = "Chroniques d'actualité de la S.E.D.E.I.S",
  ["chron bot"] = "Chronica botanica",
  ["chron ceped"] = "La chronique du CEPED",
  ["chron egypte"] = "Chronique d'Égypte; bulletin périodique de la Fondation égyptologique reine Élisabeth",
  ["chron health impact assess"] = "Chronicles of health impact assessment",
  ["chron high educ"] = "The Chronicle of higher education",
  ["chron horticult"] = "Chronica horticulturae",
  ["chron mentor coach"] = "The chronicle of mentoring & coaching",
  ["chron new res"] = "Chronicle of the New Researcher",
  ["chron okla"] = "Chronicles of Oklahoma",
  ["chron r coll physicians edinb"] = "Chronicle (Royal College of Physicians of Edinburgh)",
  ["chron respir dis"] = "Chronic respiratory disease",
  ["chron world health organ"] = "Chronicle of the World Health Organization",
  ["chronic dis can"] = "Chronic diseases in Canada",
  ["chronic dis inj can"] = "Chronic diseases and injuries in Canada",
  ["chronic dis transl med"] = "Chronic diseases and translational medicine",
  ["chronic illn"] = "Chronic illness",
  ["chronic obstr pulm dis"] = "Chronic obstructive pulmonary diseases (Miami, Fla.)",
  ["chronic pain manag"] = "Chronic pain & management",
  ["chronic stress (thousand oaks)"] = "Chronic stress (Thousand Oaks, Calif.)",
  ["chronicle (phila)"] = "Chronicle (Philadelphia, Pa. : 1990)",
  ["chronobiol int"] = "Chronobiology international",
  ["chudoku kenkyu"] = "Chūdoku kenkyū : Chūdoku Kenkyūkai jun kikanshi = The Japanese journal of toxicology",
  ["chulalongkorn wetchasan"] = "Chulālongkōn wētchasān = Medical journal of Chulalongkorn Hospital Medical School (Thailand)",
  ["chun shih i hsueh ko hsueh yuan yuan kan"] = "Jun shi yi xue ke xue yuan yuan kan = Bulletin of the Academy of Military Medical Sciences",
  ["chuncui shuxue yu yingyong shuxue zhuanzhu"] = "Chuncui Shuxue yu Yingyong Shuxue Zhuanzhu",
  ["chung ang ui dai chi"] = "Chungang Ŭidae chi = The Chung-Ang journal of medicine",
  ["chung hua chieh ho ho hu hsi tsa chih"] = "Chung-Hua Chieh Ho Ho Hu Hsi Tsa Chih Chinese Journal of Tuberculosis and Respiratory Diseases",
  ["chung hua chung liu tsa chih"] = "Chung-Hua Chung Liu Tsa Chih [Chinese Journal of Oncology]",
  ["chung hua erh pi yen hou ko tsa chih"] = "Chung-Hua Erh Pi Yen Hou Ko Tsa Chih Chinese Journal Of Otorhinolaryngology",
  ["chung hua fu chan ko tsa chih"] = "Chung-Hua Fu Chan Ko Tsa Chih [Chinese Journal of Obstetrics AndGynecology]",
  ["chung hua hsiao hua tsa chih"] = "Zhonghua xiao hua za zhi = Chinese journal of digestion",
  ["chung hua hsueh yeh hsueh tsa chi"] = "Chung-Hua Hsueh Yeh Hsueh Tsa Chih Chinese Journal of Hematology",
  ["chung hua i hsueh tsa chih"] = "Chung-Hua I Hsueh Tsa Chih [Chinese Medical Journal]",
  ["chung hua i hsueh tsa chih (taipei)"] = "Chung-Hua I Hsueh Tsa Chih [Chinese Medical Journal]",
  ["chung hua kou chiang hsueh tsa chih"] = "Chung-Hua Kou Chiang I Hsueh Tsa Chih Chinese Journal of Stomatology",
  ["chung hua liu hsing ping hsueh tsa chih"] = "Chung-Hua Liu Hsing Ping Hsueh Tsa Chih Chinese Journal of Epidemiology (Pei-Ching",
  ["chung hua min kuo fu chan ko i hsueh hui"] = "Zhonghua Minguo fu chan ke yi xue hui za zhi. The Journal of obstetrics and gynecology of the Republic of China",
  ["chung hua nei ko tsa chih"] = "Chung-Hua Nei Ko Tsa Chih Chinese Journal of Internal Medicine",
  ["chung hua ping li hsueh tsa chih"] = "Chung-Hua Ping Li Hsueh Tsa Chih [Chinese Journal of Pathology]",
  ["chung hua wai ko tsa chih"] = "Chung-Hua Wai Ko Tsa Chih [Chinese Journal of Surgery]",
  ["chung hua yen ko tsa chih"] = "Chung-Hua Yen Ko Tsa Chih [Chinese Journal of Ophthalmology]",
  ["chung hua yu fang i hsueh tsa chih"] = "Chung-Hua Yu Fang I Hsueh Tsa Chih [Chinese Journal Of Preventive Medicine]",
  ["chung kuo chung hsi i chieh ho tsa chih"] = "Chung-Kuo Chung Hsi I Chieh Ho Tsa Chih (Pei-Ching",
  ["chung kuo chung yao tsa chih"] = "Chung-Kuo Chung Yao Tsa Chih China Journal of Chinese Materia Medica (Pei-Ching",
  ["chung kuo i hsueh ko hsueh yuan hsueh pao"] = "Chung-Kuo I Hsueh Ko Hsueh Yuan Hsueh Pao Acta Academiae Medicinae Sinicae",
  ["chung kuo i liao chi hsieh tsa chih"] = "Chung-Kuo I Liao Chi Hsieh Tsa Chih Chinese Journal of Medical Instrumentation (Shang-Hai Shih",
  ["chung kuo yao hsueh tsa chih"] = "Zhongguo yao xue za zhi (Zhongguo yao xue hui : 1989)",
  ["chung kuo yao li hsueh pao"] = "Chung-Kuo Yao Li Hsueh Pao [Acta Pharmacologica Sinica]",
  ["chung kuo ying yung sheng li hsueh tsa chih"] = "Chung-Kuo Ying Yung Sheng Li Hsueh Tsa Chih Chinese Journal of Applied Physiology",
  ["chung yao tsai"] = "Chung Yao Tsai Journal of Chinese Medicinal Materials (Kuang-Chou Shih",
  ["chung-kuo yen cheng yen chiu|chin j cancer res"] = "Chinese journal of cancer research",
  ["chungang uihak"] = "Chungang ŭihak = The Korean central journal of medicine",
  ["chunlasan phantusart"] = "Chunlasan phantusart",
  ["church hist"] = "Church history",
  ["church soc"] = "Church and society",
  ["churchill/brown ser"] = "Churchill/Brown Series",
  ["chuzhou xue yuan xue bao"] = "Chuzhou xue yuan xue bao",
  ["ci ji yi xue za zhi"] = "Ci ji yi xue za zhi = Tzu-chi medical journal",
  ["ciba clin symp"] = "Ciba clinical symposia",
  ["ciba found study group"] = "Ciba Foundation study group",
  ["ciba found symp"] = "Ciba Foundation symposium",
  ["ciba symp"] = "Ciba symposia",
  ["ciba z (basel)"] = "Ciba-Zeitschrift",
  ["ciciams nouv"] = "C.I.C.I.A.M.S. nouvelles",
  ["ciclos hist econom soc"] = "Ciclos en la historia, la economía y la sociedad",
  ["ciel terre"] = "Ciel et terre",
  ["cien saude colet"] = "Ciência & saúde coletiva",
  ["cienc agron"] = "Ciência agronômica",
  ["cienc agrotecnol"] = "Ciencia e Agrotecnologia",
  ["cienc biol mol cell biol"] = "Ciencia Biologia Molecular and Cellular Biology",
  ["cienc cienc (mex city mex)"] = "Ciencia (Mexico City, Mexico)",
  ["cienc conducta"] = "Ciencias de la conducta : revista del Centro Caribeño de Estudios Post-Graduados",
  ["cienc cult"] = "Ciência e cultura",
  ["cienc desarro"] = "Ciencia y desarrollo",
  ["cienc econ (san jose)"] = "Ciencias económicas (San José, Costa Rica)",
  ["cienc florestal"] = "Ciencia Florestal",
  ["cienc innov salud"] = "Ciencia e innovación en salud",
  ["cienc invest"] = "Ciencia e investigación",
  ["cienc investig agrar"] = "Ciencia e investigación agraria",
  ["cienc mar"] = "Ciencias marinas",
  ["cienc mat (havana)"] = "Revista Ciencias Matemáticas",
  ["cienc med hisp am"] = "Ciencias médicas hispano-americanas",
  ["cienc rural"] = "Ciência rural (Online)",
  ["cienc soc"] = "Ciencia Y Sociedad",
  ["cienc soc relig"] = "Ciencias sociales y religión = Ciências sociais e religião",
  ["cienc tec vitivinic"] = "Ciencia e Tecnica Vitivinicola",
  ["cienc tecn"] = "Ciencia y Tecnología",
  ["cienc tecnol aliment"] = "Ciencia y Tecnologia Alimentaria",
  ["cienc trab"] = "Ciencia & trabajo : C&T",
  ["cienc téc"] = "Ciencia y Técnica",
  ["cienc vet (heredia)"] = "Ciencias veterinarias (Heredia, Costa Rica)",
  ["cim bull"] = "CIM Bulletin",
  ["cim j"] = "CIM Journal",
  ["cim ser math sci"] = "CIM Series in Mathematical Sciences",
  ["cim14 conf interdiscip musicol (2014)"] = "CIM14, Conference on Interdisciplinary Musicology : proceedings. Conference on Interdisciplinary Musicology (9th : 2014 : Berlin, Germany)",
  ["cime found subser"] = "CIME Foundation Subseries",
  ["cime summer sch"] = "Centro Internazionale Matematico Estivo (C.I.M.E.) Summer Schools",
  ["cin plus"] = "CIN plus",
  ["cinci hist soc bull"] = "The Cincinnati Historical Society bulletin",
  ["cinci j med"] = "Cincinnati journal of medicine",
  ["cincinnati dent soc bull"] = "Cincinnati Dental Society bulletin (1979)",
  ["cine j"] = "Cinema journal",
  ["cir bucal"] = "Cirugia Bucal",
  ["cir cir"] = "Cirugía y cirujanos",
  ["cir esp"] = "Cirugía española",
  ["cir ginecol urol"] = "Cirugía, ginecología y urología",
  ["cir lett u s army"] = "Circular letter. United States. Army. Forces, Pacific. Office of the Chief Surgeon",
  ["cir ortop traumatol"] = "Cirugía ortopédica y traumatología",
  ["cir pediatr"] = "Cirugía pediátrica : organo oficial de la Sociedad Española de Cirugía Pediátrica",
  ["circ arrhythm electrophysiol"] = "Circulation. Arrhythmia and electrophysiology",
  ["circ cardiovasc genet"] = "Circulation. Cardiovascular genetics",
  ["circ cardiovasc imaging"] = "Circulation. Cardiovascular imaging",
  ["circ cardiovasc interv"] = "Circulation. Cardiovascular interventions",
  ["circ cardiovasc qual outcomes"] = "Circulation. Cardiovascular quality and outcomes",
  ["circ econ sustainability"] = "Circular Economy and Sustainability",
  ["circ farm"] = "Circular farmacéutica",
  ["circ genom precis med"] = "Circulation. Genomic and precision medicine",
  ["circ heart fail"] = "Circulation. Heart failure",
  ["circ j"] = "Circulation journal : official journal of the Japanese Circulation Society",
  ["circ lett u s far east command"] = "Circular letter. United States. Far East Command. Medical Section",
  ["circ odontol san martin tres febr"] = "Circulo Odontologico De San Martin Y Tres De Febrero",
  ["circ rep"] = "Circulation reports",
  ["circ res"] = "Circulation research",
  ["circ shock"] = "Circulatory shock",
  ["circ shock suppl"] = "Circulatory shock. Supplement",
  ["circ: arrhythmia electrophysiol"] = "Circulation: Arrhythmia and Electrophysiology",
  ["circ: cardiovasc imaging"] = "Circulation: Cardiovascular Imaging",
  ["circ: cardiovasc interventions"] = "Circulation: Cardiovascular Interventions",
  ["circ: cardiovasc qual outcomes"] = "Circulation: Cardiovascular Quality & Outcomes",
  ["circ: genomic precis med"] = "Circulation: Genomic and Precision Medicine",
  ["circ: heart failure"] = "Circulation: Heart Failure",
  ["circuit world"] = "Circuit World",
  ["circuits syst signal process"] = "Circuits, systems, and signal processing",
  ["circuits systems signal process"] = "Circuits, Systems, and Signal Processing",
  ["cirdap dev dig"] = "CIRDAP development digest",
  ["cired open access proc j"] = "CIRED Open Access Proceedings Journal",
  ["cires cah ivoir rech econ soc"] = "CIRES, Cahiers ivoiriens de recherche économique et sociale. Centre ivoirien de recherches économiques et sociales",
  ["cirm jean-morlet ser"] = "CIRM Jean-Morlet Series",
  ["cirp ann"] = "CIRP Annals",
  ["cirp j manuf sci technol"] = "CIRP Journal of Manufacturing Science and Technology",
  ["cism courses and lect"] = "CISM International Centre for Mechanical Sciences. Courses and Lectures",
  ["cism courses and lectures"] = "CISM Courses and Lectures",
  ["cities environ"] = "Cities and the environment",
  ["cities health"] = "Cities & health",
  ["citiz sci"] = "Citizen science : theory and practice",
  ["citiz sci: theory pract"] = "Citizen Science: Theory and Practice",
  ["citizensh stud"] = "Citizenship studies",
  ["city (lond)"] = "City (London, England)",
  ["city built environ"] = "City and Built Environment",
  ["city community"] = "City & community",
  ["city environ interact"] = "City and Environment Interactions",
  ["city j"] = "City journal (New York, N.Y.)",
  ["city soc (wash)"] = "City & society (Washington, D.C.)",
  ["cityscape: j pol devel res"] = "Cityscape: A Journal of Policy Development and Research",
  ["civ cattol"] = "La Civiltà cattolica",
  ["civ code annot state calif calif"] = "Civil code annotated of the State of California. California",
  ["civ eng"] = "Civil engineering (New York, N.Y. : 1983)",
  ["civ eng des"] = "Civil Engineering Design",
  ["civ eng environ syst"] = "Civil Engineering and Environmental Systems",
  ["civ eng mag arch"] = "Civil Engineering Magazine Archive",
  ["civ eng sci"] = "Civil Engineering Science",
  ["civ liberties rev"] = "The Civil liberties review",
  ["civ war hist"] = "Civil War history",
  ["civ war times illus"] = "Civil War times illustrated",
  ["civil lib"] = "Civil liberties",
  ["cjc open"] = "CJC open",
  ["clao j"] = "The CLAO journal : official publication of the Contact Lens Association of Ophthalmologists, Inc",
  ["clarendon aristotle ser"] = "Clarendon Aristotle Series",
  ["clarendon lib logic philos"] = "Clarendon Library of Logic and Philosophy",
  ["class antiq"] = "Classical antiquity",
  ["class j"] = "The classical journal",
  ["class outlook"] = "The Classical outlook",
  ["class philol"] = "Classical philology",
  ["class philos argum"] = "Classic Philosophical Arguments",
  ["class probl collect"] = "Classic Problem Collections",
  ["class q"] = "Classical quarterly",
  ["class quantum gravity"] = "Classical and quantum gravity",
  ["class texts sci"] = "Classic Texts in the Sciences",
  ["class world"] = "Classical world",
  ["classical quantum gravity"] = "Classical and Quantum Gravity",
  ["classical rev"] = "The Classical review",
  ["classici econom"] = "Classici dell’Economia",
  ["classics appl math"] = "Classics in Applied Mathematics",
  ["classics math"] = "Classics in Mathematics",
  ["classics soviet math"] = "Classics of Soviet Mathematics",
  ["classr companion econ"] = "Classroom Companion. Economics",
  ["classr res mater ser"] = "Classroom Resource Materials Series",
  ["classr resour mater"] = "Classroom Resource Materials",
  ["clay lect notes"] = "Clay Lecture Notes",
  ["clay math proc"] = "Clay Mathematics Proceedings",
  ["clay miner"] = "Clay Minerals",
  ["clay res"] = "Clay Research",
  ["clay technol"] = "Clay Technology",
  ["clays clay miner"] = "Clays and Clay Minerals",
  ["clean (weinh)"] = "Clean : soil, air, water",
  ["clean - soil air water"] = "CLEAN - Soil, Air, Water",
  ["clean air"] = "Clean Air",
  ["clean energy"] = "Clean Energy",
  ["clean eng technol"] = "Cleaner engineering and technology",
  ["clean technol"] = "Clean Technologies",
  ["clean technol environ policy"] = "Clean Technologies and Environmental Policy",
  ["clean-soil air water"] = "Clean-Soil Air Water",
  ["clean: soil, air, water"] = "Clean: Soil, Air, Water",
  ["cleaner chem eng"] = "Cleaner Chemical Engineering",
  ["cleaner circ bioecon"] = "Cleaner and Circular Bioeconomy",
  ["cleaner energy syst"] = "Cleaner Energy Systems",
  ["cleaner eng technol"] = "Cleaner Engineering and Technology",
  ["cleaner environ syst"] = "Cleaner Environmental Systems",
  ["cleaner logist supply chain"] = "Cleaner Logistics and Supply Chain",
  ["cleaner mater"] = "Cleaner Materials",
  ["cleaner prod lett"] = "Cleaner Production Letters",
  ["cleaner responsible consumption"] = "Cleaner and Responsible Consumption",
  ["cleaner waste syst"] = "Cleaner Waste Systems",
  ["clearing house"] = "Clearing house (Menasha, Wis.)",
  ["cleft palate craniofac j"] = "The Cleft palate-craniofacial journal : official publication of the American Cleft Palate-Craniofacial Association",
  ["cleft palate j"] = "The Cleft palate journal",
  ["clergy rev"] = "The Clergy review",
  ["cleve clin j med"] = "Cleveland Clinic journal of medicine",
  ["cleve clin q"] = "Cleveland Clinic quarterly",
  ["clevel state law rev"] = "Cleveland state law review",
  ["clgh rev"] = "Clearinghouse review",
  ["clim action"] = "Climate Action",
  ["clim chang econ (singap)"] = "Climate change economics",
  ["clim change"] = "Climatic Change",
  ["clim change ecol"] = "Climate Change Ecology",
  ["clim change econ"] = "Climate Change Economics",
  ["clim change res"] = "Climate Change Research",
  ["clim dev"] = "Climate and development",
  ["clim dyn"] = "Climate Dynamics",
  ["clim energy"] = "Climate and Energy",
  ["clim law"] = "Climate Law",
  ["clim past"] = "Climate of the Past",
  ["clim past discuss"] = "Climate of the Past Discussion",
  ["clim policy"] = "Climate Policy",
  ["clim res"] = "Climate research",
  ["clim resilience sustainability"] = "Climate Resilience and Sustainability",
  ["clim risk manag"] = "Climate risk management",
  ["clim risk manage"] = "Climate Risk Management",
  ["clim serv"] = "Climate services",
  ["climate (basel)"] = "Climate (Basel, Switzerland)",
  ["climatol bull"] = "Climatological Bulletin",
  ["clin adv hematol oncol"] = "Clinical advances in hematology & oncology : H&O",
  ["clin adv periodontics"] = "Clinical advances in periodontics",
  ["clin allergy"] = "Clinical allergy",
  ["clin allergy immunol"] = "Clinical allergy and immunology",
  ["clin anat"] = "Clinical anatomy (New York, N.Y.)",
  ["clin anesth"] = "Clinical anesthesia",
  ["clin appl immunol rev"] = "Clinical and applied immunology reviews",
  ["clin appl thromb hemost"] = "Clinical and applied thrombosis/hemostasis : official journal of the International Academy of Clinical and Applied Thrombosis/Hemostasis",
  ["clin arch bone jt dis"] = "Clinical archives of bone and joint diseases",
  ["clin arch commun disord"] = "Clinical archives of communication disorders",
  ["clin auton res"] = "Clinical autonomic research : official journal of the Clinical Autonomic Research Society",
  ["clin biochem"] = "Clinical biochemistry",
  ["clin biochem rev"] = "The Clinical biochemist. Reviews",
  ["clin biomech"] = "Clinical Biomechanics",
  ["clin biomech (bristol, avon)"] = "Clinical biomechanics (Bristol, Avon)",
  ["clin breast cancer"] = "Clinical breast cancer",
  ["clin bull"] = "Clinical bulletin",
  ["clin bull sch med west reserve univ assoc hosp"] = "The clinical bulletin of the School of Medicine of Western Reserve University and its associated hospitals. Western Reserve University. School of Medicine",
  ["clin bull univ hosp clevel (ohio)"] = "The clinical bulletin. University Hospitals of Cleveland (Ohio)",
  ["clin calcium"] = "Clinical calcium",
  ["clin cancer drugs"] = "Clinical cancer drugs",
  ["clin cancer investig j"] = "Clinical cancer investigation journal",
  ["clin cancer res"] = "Clinical cancer research : an official journal of the American Association for Cancer Research",
  ["clin cardiol"] = "Clinical cardiology",
  ["clin case rep"] = "Clinical case reports",
  ["clin case rep rev"] = "Clinical case reports and reviews",
  ["clin case stud"] = "Clinical case studies",
  ["clin cases miner bone metab"] = "Clinical cases in mineral and bone metabolism : the official journal of the Italian Society of Osteoporosis, Mineral Metabolism, and Skeletal Diseases",
  ["clin chem"] = "Clinical chemistry",
  ["clin chem (washington, dc, u s)"] = "Clinical Chemistry (Washington, DC, United States)",
  ["clin chem lab med"] = "Clinical Chemistry and Laboratory Medicine",
  ["clin chest med"] = "Clinics in chest medicine",
  ["clin child fam psychol rev"] = "Clinical child and family psychology review",
  ["clin child psychol psychiatry"] = "Clinical child psychology and psychiatry",
  ["clin chim acta"] = "Clinica Chimica Acta",
  ["clin colon rectal surg"] = "Clinics in colon and rectal surgery",
  ["clin colorectal cancer"] = "Clinical colorectal cancer",
  ["clin commun disord"] = "Clinics in communication disorders",
  ["clin contemp"] = "Clínica contemporânea; revista de medicina e cirurgia",
  ["clin cornerstone"] = "Clinical cornerstone",
  ["clin cosmet investig dent"] = "Clinical, cosmetic and investigational dentistry",
  ["clin cosmet investig dermatol"] = "Clinical, cosmetic and investigational dermatology",
  ["clin cytol"] = "Clinical cytology",
  ["clin dermatol"] = "Clinics in dermatology",
  ["clin dermatol (wilmington)"] = "Clinical dermatology (Wilmington, Del.)",
  ["clin dev immunol"] = "Clinical & developmental immunology",
  ["clin dev med"] = "Clinics in developmental medicine",
  ["clin diabetes"] = "Clinical diabetes : a publication of the American Diabetes Association",
  ["clin diabetes endocrinol"] = "Clinical diabetes and endocrinology",
  ["clin diagn lab immunol"] = "Clinical and diagnostic laboratory immunology",
  ["clin diagn pathol"] = "Clinical and diagnostic pathology",
  ["clin diagn ultrasound"] = "Clinics in diagnostic ultrasound",
  ["clin diagn virol"] = "Clinical and diagnostic virology",
  ["clin drug invest"] = "Clinical Drug Investigation",
  ["clin drug investig"] = "Clinical drug investigation",
  ["clin dysmorphol"] = "Clinical dysmorphology",
  ["clin eeg neurosci"] = "Clinical EEG and neuroscience",
  ["clin eff nurs"] = "Clinical effectiveness in nursing",
  ["clin electroencephalogr"] = "Clinical EEG (electroencephalography)",
  ["clin endocrinol"] = "Clinical Endocrinology",
  ["clin endocrinol (oxf)"] = "Clinical endocrinology",
  ["clin endocrinol metab"] = "Clinics in endocrinology and metabolism",
  ["clin endosc"] = "Clinical endoscopy",
  ["clin eng"] = "Clinical engineering",
  ["clin eng inf serv"] = "Clinical engineering information service",
  ["clin eng news"] = "Clinical engineering news",
  ["clin epidemiol"] = "Clinical epidemiology",
  ["clin epidemiol glob health"] = "Clinical epidemiology and global health",
  ["clin epigenet"] = "Clinical Epigenetics",
  ["clin epigenet: epigenet commun"] = "Clinical Epigenetics: Epigenetics Communications",
  ["clin epigenetics"] = "Clinical epigenetics",
  ["clin ethics"] = "Clinical ethics",
  ["clin ethics rep"] = "Clinical ethics report",
  ["clin eur"] = "Clinica europea",
  ["clin evid"] = "Clinical evidence",
  ["clin excell nurse pract"] = "Clinical excellence for nurse practitioners : the international journal of NPACE",
  ["clin excerpts j devoted ther"] = "Clinical excerpts; a journal devoted to therapeutics",
  ["clin exerc physiol"] = "Clinical exercise physiology",
  ["clin exp allergy"] = "Clinical and experimental allergy : journal of the British Society for Allergy and Clinical Immunology",
  ["clin exp dent res"] = "Clinical and experimental dental research",
  ["clin exp dermatol"] = "Clinical and experimental dermatology",
  ["clin exp dial apheresis"] = "Clinical and experimental dialysis and apheresis",
  ["clin exp emerg med"] = "Clinical and experimental emergency medicine",
  ["clin exp gastroenterol"] = "Clinical and experimental gastroenterology",
  ["clin exp gastroenterol hepatol"] = "Clinical & experimental gastroenterology & hepatology",
  ["clin exp hepatol"] = "Clinical and experimental hepatology",
  ["clin exp hypertens"] = "Clinical and experimental hypertension (New York, N.Y. : 1993)",
  ["clin exp hypertens a"] = "Clinical and experimental hypertension. Part A, Theory and practice",
  ["clin exp hypertens b"] = "Clinical and experimental hypertension. Part B, Hypertension in pregnancy",
  ["clin exp immunol"] = "Clinical and experimental immunology",
  ["clin exp med"] = "Clinical and experimental medicine",
  ["clin exp med sci"] = "Clinical and Experimental Medical Sciences",
  ["clin exp metastasis"] = "Clinical & experimental metastasis",
  ["clin exp nephrol"] = "Clinical and experimental nephrology",
  ["clin exp neuroimmunol"] = "Clinical & experimental neuroimmunology",
  ["clin exp neurol"] = "Clinical and experimental neurology",
  ["clin exp obstet gynecol"] = "Clinical and experimental obstetrics & gynecology",
  ["clin exp ophthalmol"] = "Clinical & experimental ophthalmology",
  ["clin exp optom"] = "Clinical & experimental optometry",
  ["clin exp otorhinolaryngol"] = "Clinical and experimental otorhinolaryngology",
  ["clin exp pathol"] = "Clinical and experimental pathology",
  ["clin exp pediatr"] = "Clinical and experimental pediatrics",
  ["clin exp pharmacol"] = "Clinical & experimental pharmacology",
  ["clin exp pharmacol physiol"] = "Clinical and experimental pharmacology & physiology",
  ["clin exp pharmacol physiol suppl"] = "Clinical and experimental pharmacology & physiology. Supplement",
  ["clin exp psychol"] = "Clinical and experimental psychology",
  ["clin exp reprod med"] = "Clinical and experimental reproductive medicine",
  ["clin exp rheumatol"] = "Clinical and experimental rheumatology",
  ["clin exp vaccine res"] = "Clinical and experimental vaccine research",
  ["clin experiment ophthalmol"] = "Clinical and Experimental Ophthalmology",
  ["clin eye vis care"] = "Clinical eye and vision care",
  ["clin gastroenterol"] = "Clinics in gastroenterology",
  ["clin gastroenterol hepatol"] = "Clinical gastroenterology and hepatology : the official clinical practice journal of the American Gastroenterological Association",
  ["clin genet"] = "Clinical genetics",
  ["clin genitourin cancer"] = "Clinical genitourinary cancer",
  ["clin geriatr"] = "Clinical geriatrics",
  ["clin geriatr med"] = "Clinics in geriatric medicine",
  ["clin gerontol"] = "Clinical gerontologist",
  ["clin ginecol"] = "La Clinica ginecologica",
  ["clin gov"] = "Clinical governance",
  ["clin haematol"] = "Clinics in haematology",
  ["clin hematol int"] = "Clinical hematology international",
  ["clin hemorheol"] = "Clinical hemorheology",
  ["clin hemorheol microcirc"] = "Clinical hemorheology and microcirculation",
  ["clin hig hidrol"] = "Clinica, higiene e hidrologia",
  ["clin hisp"] = "Clinica hispánica",
  ["clin hypertens"] = "Clinical hypertension",
  ["clin imaging"] = "Clinical imaging",
  ["clin immunol"] = "Clinical immunology (Orlando, Fla.)",
  ["clin immunol (amsterdam, neth)"] = "Clinical Immunology (Amsterdam, Netherlands)",
  ["clin immunol (san diego, ca, u s)"] = "Clinical Immunology (San Diego, CA, United States)",
  ["clin immunol immunopathol"] = "Clinical immunology and immunopathology",
  ["clin immunol rev"] = "Clinical immunology reviews",
  ["clin immunother"] = "Clinical immunotherapeutics",
  ["clin implant dent relat res"] = "Clinical implant dentistry and related research",
  ["clin infect dis"] = "Clinical infectious diseases : an official publication of the Infectious Diseases Society of America",
  ["clin infect pract"] = "Clinical infection in practice",
  ["clin intensive care"] = "Clinical intensive care : international journal of critical & coronary care medicine",
  ["clin interv aging"] = "Clinical interventions in aging",
  ["clin invest ginecol obstet"] = "Clínica e investigación en ginecología y obstetricia",
  ["clin invest med"] = "Clinical and investigative medicine. Médecine clinique et experimentale",
  ["clin investig"] = "The Clinical investigator",
  ["clin investig (lond)"] = "Clinical investigation",
  ["clin investig arterioscler"] = "Clínica e investigación en arteriosclerosis : publicación oficial de la Sociedad Española de Arteriosclerosis",
  ["clin j"] = "The Clinical journal",
  ["clin j am soc nephrol"] = "Clinical journal of the American Society of Nephrology : CJASN",
  ["clin j gastroenterol"] = "Clinical journal of gastroenterology",
  ["clin j oncol nurs"] = "Clinical journal of oncology nursing",
  ["clin j pain"] = "The Clinical journal of pain",
  ["clin j sport med"] = "Clinical journal of sport medicine : official journal of the Canadian Academy of Sport Medicine",
  ["clin j womens health"] = "Clinical journal of women's health",
  ["clin kidney j"] = "Clinical kidney journal",
  ["clin lab"] = "Clinical laboratory",
  ["clin lab (zaragoza)"] = "Clinica y laboratorio",
  ["clin lab haematol"] = "Clinical and laboratory haematology",
  ["clin lab int"] = "Clinical laboratory international",
  ["clin lab manage rev"] = "Clinical laboratory management review : official publication of the Clinical Laboratory Management Association",
  ["clin lab med"] = "Clinics in laboratory medicine",
  ["clin lab sci"] = "Clinical laboratory science : journal of the American Society for Medical Technology",
  ["clin lact (amarillo)"] = "Clinical lactation (Amarillo, Tex.)",
  ["clin laser mon"] = "Clinical laser monthly",
  ["clin lat"] = "Clinica latina",
  ["clin leadersh manag rev"] = "Clinical leadership & management review : the journal of CLMA",
  ["clin linguist phon"] = "Clinical linguistics & phonetics",
  ["clin lipidol"] = "Clinical lipidology",
  ["clin lipidol metab disord"] = "Clinical Lipidology and Metabolic Disorders",
  ["clin liver dis"] = "Clinics in liver disease",
  ["clin liver dis (hoboken)"] = "Clinical liver disease",
  ["clin lung cancer"] = "Clinical lung cancer",
  ["clin lymphoma"] = "Clinical lymphoma",
  ["clin lymphoma myeloma"] = "Clinical lymphoma & myeloma",
  ["clin lymphoma myeloma leuk"] = "Clinical lymphoma, myeloma & leukemia",
  ["clin mass spectrom"] = "Clinical mass spectrometry (Del Mar, Calif.)",
  ["clin mater"] = "Clinical Materials",
  ["clin med"] = "Clinical Medicine",
  ["clin med (lond)"] = "Clinical medicine (London, England)",
  ["clin med (northfield il)"] = "Clinical medicine",
  ["clin med (northfield)"] = "Clinical medicine (Northfield, Ill.)",
  ["clin med arthritis musculoskelet disord"] = "Clinical medicine. Arthritis and musculoskeletal disorders",
  ["clin med cardiol"] = "Clinical medicine. Cardiology",
  ["clin med case rep"] = "Clinical medicine. Case reports",
  ["clin med circ respirat pulm med"] = "Clinical medicine. Circulatory, respiratory and pulmonary medicine",
  ["clin med dermatol"] = "Clinical medicine. Dermatology",
  ["clin med gastroenterol"] = "Clinical Medicine. Gastroenterology",
  ["clin med geriatr"] = "Clinical medicine. Geriatrics",
  ["clin med insights arthritis musculoskelet disord"] = "Clinical medicine insights. Arthritis and musculoskeletal disorders",
  ["clin med insights blood disord"] = "Clinical medicine insights. Blood disorders",
  ["clin med insights cardiol"] = "Clinical Medicine Insights. Cardiology",
  ["clin med insights case rep"] = "Clinical medicine insights. Case reports",
  ["clin med insights circ respir pulm med"] = "Clinical medicine insights. Circulatory, respiratory and pulmonary medicine",
  ["clin med insights ear nose throat"] = "Clinical medicine insights. Ear, nose and throat",
  ["clin med insights endocrinol diabetes"] = "Clinical medicine insights. Endocrinology and diabetes",
  ["clin med insights gastroenterol"] = "Clinical medicine insights. Gastroenterology",
  ["clin med insights oncol"] = "Clinical Medicine Insights. Oncology",
  ["clin med insights pathol"] = "Clinical medicine insights. Pathology",
  ["clin med insights pediatr"] = "Clinical medicine insights. Pediatrics",
  ["clin med insights reprod health"] = "Clinical medicine insights. Reproductive health",
  ["clin med insights ther"] = "Clinical medicine insights. Therapeutics",
  ["clin med insights urol"] = "Clinical medicine insights. Urology",
  ["clin med insights womens health"] = "Clinical medicine insights. Women's health",
  ["clin med oncol"] = "Clinical medicine. Oncology",
  ["clin med pathol"] = "Clinical medicine. Pathology",
  ["clin med pediatr"] = "Clinical medicine. Pediatrics",
  ["clin med reprod health"] = "Clinical medicine. Reproductive health",
  ["clin med res"] = "Clinical medicine & research",
  ["clin med res (n y)"] = "Clinical medicine research",
  ["clin med rev case rep"] = "Clinical medical reviews and case reports",
  ["clin med rev oncol"] = "Clinical medicine reviews in oncology",
  ["clin med rev ther"] = "Clinical medicine reviews in therapeutics",
  ["clin med rev vasc health"] = "Clinical medicine reviews in vascular health",
  ["clin med ther"] = "Clinical medicine. Therapeutics",
  ["clin med urol"] = "Clinical medicine. Urology [electronic resource]",
  ["clin microbiol"] = "Clinical microbiology (Los Angeles, Calif.)",
  ["clin microbiol case rep"] = "Clinical microbiology & case reports",
  ["clin microbiol infect"] = "Clinical microbiology and infection : the official publication of the European Society of Clinical Microbiology and Infectious Diseases",
  ["clin microbiol infect dis"] = "Clinical microbiology and infectious diseases",
  ["clin microbiol newsl"] = "Clinical microbiology newsletter",
  ["clin microbiol rev"] = "Clinical microbiology reviews",
  ["clin mol allergy"] = "Clinical and molecular allergy : CMA",
  ["clin mol hepatol"] = "Clinical and molecular hepatology",
  ["clin mol med"] = "Clinical molecular medicine",
  ["clin mol pathol"] = "Clinical molecular pathology",
  ["clin mother child health"] = "Clinics in mother and child health",
  ["clin nephrol"] = "Clinical nephrology",
  ["clin nephrol case stud"] = "Clinical nephrology. Case studies",
  ["clin nephrol res"] = "Clinical nephrology and research",
  ["clin neurol neurosurg"] = "Clinical neurology and neurosurgery",
  ["clin neuropathol"] = "Clinical neuropathology",
  ["clin neuropharmacol"] = "Clinical neuropharmacology",
  ["clin neurophysiol"] = "Clinical neurophysiology : official journal of the International Federation of Clinical Neurophysiology",
  ["clin neurophysiol pract"] = "Clinical neurophysiology practice",
  ["clin neuropsychiatry"] = "Clinical neuropsychiatry",
  ["clin neuropsychol"] = "The Clinical neuropsychologist",
  ["clin neuroradiol"] = "Clinical neuroradiology",
  ["clin neurosci"] = "Clinical neuroscience (New York, N.Y.)",
  ["clin neurosci res"] = "Clinical neuroscience research",
  ["clin neurosurg"] = "Clinical neurosurgery",
  ["clin notes respir dis"] = "Clinical notes on respiratory diseases",
  ["clin nucl med"] = "Clinical nuclear medicine",
  ["clin nuova rass prog med int"] = "Clinica nuova; rassegna del progresso medico internazionale",
  ["clin nuova rass prog med internazionale"] = "Clinica nuova; rassegna del progresso medico internazionale",
  ["clin nurs pract epilepsy"] = "Clinical nursing practice in epilepsy",
  ["clin nurs res"] = "Clinical nursing research",
  ["clin nurs stud"] = "Clinical nursing studies",
  ["clin nurse spec"] = "Clinical nurse specialist CNS",
  ["clin nutr"] = "Clinical nutrition (Edinburgh, Scotland)",
  ["clin nutr espen"] = "Clinical nutrition ESPEN",
  ["clin nutr exp"] = "Clinical nutrition experimental",
  ["clin nutr open sci"] = "Clinical Nutrition Open Science",
  ["clin nutr res"] = "Clinical nutrition research",
  ["clin obes"] = "Clinical obesity",
  ["clin obstet gynaecol"] = "Clinics in obstetrics and gynaecology",
  ["clin obstet gynecol"] = "Clinical obstetrics and gynecology",
  ["clin obstet gynecol reprod med"] = "Clinical obstetrics, gynecology and reproductive medicine",
  ["clin occup environ med"] = "Clinics in occupational and environmental medicine",
  ["clin odontoiatr rev mens oper dent chir orale ortod protesi"] = "Clinica odontoiatrica; revista mensile di operativa dentaria, chirurgia orale, ortodontia e protesi",
  ["clin odontol"] = "Clinic odontologia : la revue de l'omnipraticien",
  ["clin oncol"] = "Clinical oncology",
  ["clin oncol (belmont)"] = "Clinics in oncology (Belmont, Calif.)",
  ["clin oncol (las vegas)"] = "Clinics of oncology",
  ["clin oncol (r coll radiol)"] = "Clinical oncology (Royal College of Radiologists (Great Britain))",
  ["clin oncol adolesc young adults"] = "Clinical oncology in adolescents and young adults",
  ["clin oncol cancer res"] = "Clinical Oncology and Cancer Research",
  ["clin oncol res"] = "Clinical oncology and research",
  ["clin ophthalmol"] = "Clinical ophthalmology (Auckland, N.Z.)",
  ["clin optom (auckl)"] = "Clinical optometry",
  ["clin oral implants res"] = "Clinical oral implants research",
  ["clin oral investig"] = "Clinical oral investigations",
  ["clin orthod res"] = "Clinical orthodontics and research",
  ["clin orthop"] = "Clinical orthopaedics",
  ["clin orthop relat res"] = "Clinical orthopaedics and related research",
  ["clin orthop surg"] = "Clinics in orthopedic surgery",
  ["clin ortop"] = "La Clinica ortopedica",
  ["clin ostet ginecol"] = "La Clinica ostetrica e ginecologica",
  ["clin otolaryngol"] = "Clinical otolaryngology : official journal of ENT-UK ; official journal of Netherlands Society for Oto-Rhino-Laryngology & Cervico-Facial Surgery",
  ["clin otolaryngol allied sci"] = "Clinical otolaryngology and allied sciences",
  ["clin otorinolaringoiatr"] = "La Clinica otorinolaringoiatrica",
  ["clin ovarian cancer other gynecol malig"] = "Clinical ovarian cancer and other gynecologic malignancies",
  ["clin park relat disord"] = "Clinical parkinsonism & related disorders",
  ["clin pathol"] = "Clinical pathology (Thousand Oaks, Ventura County, Calif.)",
  ["clin pediatr (bologna)"] = "La Clinica pediatrica",
  ["clin pediatr (phila)"] = "Clinical pediatrics",
  ["clin pediatr dermatol"] = "Clinical pediatric dermatology (Wilmington. Del.)",
  ["clin pediatr emerg med"] = "Clinical pediatric emergency medicine",
  ["clin pediatr endocrinol"] = "Clinical pediatric endocrinology : case reports and clinical investigations : official journal of the Japanese Society for Pediatric Endocrinology",
  ["clin perform qual health care"] = "Clinical performance and quality health care",
  ["clin perinatol"] = "Clinics in perinatology",
  ["clin pharm"] = "Clinical pharmacy",
  ["clin pharmacokinet"] = "Clinical pharmacokinetics",
  ["clin pharmacol"] = "Clinical pharmacology : advances and applications",
  ["clin pharmacol biopharm"] = "Clinical pharmacology & biopharmaceutics",
  ["clin pharmacol drug dev"] = "Clinical pharmacology in drug development",
  ["clin pharmacol ther"] = "Clinical pharmacology and therapeutics",
  ["clin pharmacol ther (st louis)"] = "Clinical Pharmacology and Therapeutics (St. Louis)",
  ["clin pharmacol transl med"] = "Clinical pharmacology and translational medicine",
  ["clin pharmacol: adv appl"] = "Clinical Pharmacology: Advances and Applications",
  ["clin phys"] = "Clinical Physiology Series",
  ["clin phys physiol meas"] = "Clinical Physics and Physiological Measurement",
  ["clin physiol"] = "Clinical physiology (Oxford, England)",
  ["clin physiol biochem"] = "Clinical physiology and biochemistry",
  ["clin physiol funct imaging"] = "Clinical physiology and functional imaging",
  ["clin phytosci"] = "Clinical Phytoscience",
  ["clin plasma med"] = "Clinical plasma medicine",
  ["clin plast surg"] = "Clinics in plastic surgery",
  ["clin podiatr med surg"] = "Clinics in podiatric medicine and surgery",
  ["clin podiatry"] = "Clinics in podiatry",
  ["clin positron imaging"] = "Clinical positron imaging : official journal of the Institute for Clinical P.E.T",
  ["clin pract"] = "Clinics and practice",
  ["clin pract (lond)"] = "Clinical practice (London, England)",
  ["clin pract cases emerg med"] = "Clinical practice and cases in emergency medicine",
  ["clin pract epidemiol ment health"] = "Clinical practice and epidemiology in mental health : CP & EMH",
  ["clin pract epidemol ment health"] = "Clinical practice and epidemiology in mental health : CP & EMH",
  ["clin pract guidel quick ref guide clin"] = "Clinical practice guideline. Quick reference guide for clinicians",
  ["clin pract pediatr psychol"] = "Clinical practice in pediatric psychology",
  ["clin prev dent"] = "Clinical preventive dentistry",
  ["clin privil white pap"] = "Clinical privilege white paper",
  ["clin proc"] = "Clinical proceedings",
  ["clin proc child hosp dist columbia"] = "Clinical proceedings - Children's Hospital of the District of Columbia",
  ["clin proc child hosp natl med cent"] = "Clinical proceedings - Children's Hospital National Medical Center",
  ["clin prostate cancer"] = "Clinical prostate cancer",
  ["clin prostate kidney cancer"] = "Clinical Prostate and Kidney Cancer",
  ["clin proteomics"] = "Clinical proteomics",
  ["clin psychiatry (wilmington)"] = "Clinical psychiatry (Wilmington, Del.)",
  ["clin psychol"] = "The Clinical psychologist",
  ["clin psychol (aust psychol soc)"] = "Clinical psychologist (Australian Psychological Society)",
  ["clin psychol (new york)"] = "Clinical psychology : a publication of the Division of Clinical Psychology of the American Psychological Association",
  ["clin psychol psychother"] = "Clinical psychology & psychotherapy",
  ["clin psychol rev"] = "Clinical psychology review",
  ["clin psychol sci"] = "Clinical psychological science : a journal of the Association for Psychological Science",
  ["clin psychopharmacol neurosci"] = "Clinical psychopharmacology and neuroscience : the official scientific journal of the Korean College of Neuropsychopharmacology",
  ["clin pulm med"] = "Clinical pulmonary medicine",
  ["clin radiol"] = "Clinical radiology",
  ["clin radiol imaging j"] = "Clinical radiology & imaging journal",
  ["clin rehabil"] = "Clinical rehabilitation",
  ["clin rep"] = "Clinical reports",
  ["clin reprod fertil"] = "Clinical reproduction and fertility",
  ["clin res"] = "Clinical research",
  ["clin res (alex)"] = "Clinical researcher (Alexandria, Va.)",
  ["clin res cardiol"] = "Clinical research in cardiology : official journal of the German Cardiac Society",
  ["clin res cardiol suppl"] = "Clinical research in cardiology supplements",
  ["clin res diabetes endocrinol"] = "Clinical research in diabetes and endocrinology",
  ["clin res foot ankle"] = "Clinical research on foot & ankle",
  ["clin res hepatol gastroenterol"] = "Clinics and research in hepatology and gastroenterology",
  ["clin res hiv aids"] = "Clinical research in HIV/AIDS",
  ["clin res infect dis"] = "Clinical research in infectious diseases",
  ["clin res pr drug regul aff"] = "Clinical research practices and drug regulatory affairs",
  ["clin res pulmonol"] = "Clinical research in pulmonology",
  ["clin res regul aff"] = "Clinical research and regulatory affairs",
  ["clin res trials"] = "Clinical research and trials",
  ["clin resour manag"] = "Clinical resource management",
  ["clin respir j"] = "The clinical respiratory journal",
  ["clin rev allergy"] = "Clinical reviews in allergy",
  ["clin rev allergy immunol"] = "Clinical reviews in allergy & immunology",
  ["clin rev bone miner metab"] = "Clinical reviews in bone and mineral metabolism",
  ["clin rheum dis"] = "Clinics in rheumatic diseases",
  ["clin rheumatol"] = "Clinical rheumatology",
  ["clin risk"] = "Clinical risk",
  ["clin salud"] = "Clínica y salud",
  ["clin sarcoma res"] = "Clinical sarcoma research",
  ["clin sch rev"] = "Clinical scholars review",
  ["clin schizophr relat psychoses"] = "Clinical schizophrenia & related psychoses",
  ["clin sci"] = "Clinical science",
  ["clin sci (colch)"] = "Clinical Science",
  ["clin sci (lond)"] = "Clinical science (London, England : 1979)",
  ["clin sci mol med"] = "Clinical science and molecular medicine",
  ["clin sci mol med suppl"] = "Clinical science and molecular medicine. Supplement",
  ["clin shoulder elb"] = "Clinics in shoulder and elbow",
  ["clin simul nurs"] = "Clinical simulation in nursing",
  ["clin skin cancer"] = "Clinical skin cancer",
  ["clin soc j"] = "Clinical Society journal",
  ["clin soc work j"] = "Clinical social work journal",
  ["clin spectrosc"] = "Clinical Spectroscopy",
  ["clin spine surg"] = "Clinical spine surgery",
  ["clin sports med"] = "Clinics in sports medicine",
  ["clin superv"] = "The Clinical supervisor",
  ["clin surg"] = "Clinics in surgery",
  ["clin surg (las vegas)"] = "Clinics of surgery",
  ["clin symp"] = "Clinical symposia (Summit, N.J. : 1957)",
  ["clin symp (summit)"] = "Clinical symposia (Summit, N.J. : 1948)",
  ["clin teach"] = "The clinical teacher",
  ["clin tech small anim pract"] = "Clinical techniques in small animal practice",
  ["clin ter"] = "La Clinica terapeutica",
  ["clin term"] = "La Clinica termale",
  ["clin ther"] = "Clinical therapeutics",
  ["clin thyroidol"] = "Clinical thyroidology",
  ["clin tisiol"] = "Clinica tisiológica",
  ["clin torax"] = "Clinica del tórax",
  ["clin toxicol"] = "Clinical toxicology",
  ["clin toxicol (phila)"] = "Clinical toxicology (Philadelphia, Pa.)",
  ["clin transl allergy"] = "Clinical and translational allergy",
  ["clin transl gastroenterol"] = "Clinical and translational gastroenterology",
  ["clin transl imaging"] = "Clinical and translational imaging",
  ["clin transl immunol"] = "Clinical & Translational Immunology",
  ["clin transl immunology"] = "Clinical & translational immunology",
  ["clin transl med"] = "Clinical and translational medicine",
  ["clin transl neurosci"] = "Clinical and Translational Neurocience",
  ["clin transl oncol"] = "Clinical & translational oncology : official publication of the Federation of Spanish Oncology Societies and of the National Cancer Institute of Mexico",
  ["clin transl radiat oncol"] = "Clinical and translational radiation oncology",
  ["clin transl sci"] = "Clinical and translational science",
  ["clin transpl"] = "Clinical transplants",
  ["clin transplant"] = "Clinical transplantation",
  ["clin trials"] = "Clinical trials (London, England)",
  ["clin trials degener dis"] = "Clinical trials in degenerative diseases",
  ["clin trials j"] = "Clinical trials journal",
  ["clin trials metaanal"] = "Clinical trials and meta-analysis",
  ["clin trials regul sci cardiol"] = "Clinical trials and regulatory science in cardiology",
  ["clin va mason hosp"] = "Clinics of the Virginia Mason Hospital",
  ["clin vaccine immunol"] = "Clinical and vaccine immunology : CVI",
  ["clin vet (milano)"] = "La Clinica veterinaria",
  ["clinic (paris)"] = "Clinic",
  ["clinica (valladolid)"] = "Clínica (Valladolid, Spain)",
  ["clinician (goa)"] = "The Clinician",
  ["clinicoecon outcomes res"] = "ClinicoEconomics and outcomes research : CEOR",
  ["clinics (sao paulo)"] = "Clinics (São Paulo, Brazil)",
  ["clinique (paris)"] = "Clinique (Paris, France)",
  ["clio (roma)"] = "Clio (Roma)",
  ["clio (toulouse)"] = "Clio (Toulouse, France)",
  ["clio med"] = "Clio medica (Amsterdam, Netherlands)",
  ["cliometrica (berl)"] = "Cliometrica",
  ["clocks sleep"] = "Clocks & sleep",
  ["cloning stem cells"] = "Cloning and stem cells",
  ["cloning transgenes"] = "Cloning & transgenesis",
  ["clothing text res j"] = "Clothing and Textiles Research Journal",
  ["clujul med"] = "Clujul medical (1957)",
  ["cluster comput"] = "Cluster computing",
  ["clustering phenom nuclei"] = "Clustering Phenomena in Nuclei",
  ["cmaj open"] = "CMAJ open",
  ["cmc-comput mater continua"] = "CMC-Computers Materials & Continua",
  ["cmd curr med dig"] = "CMD; current medical digest",
  ["cme (berl)"] = "CME (Berlin, Germany)",
  ["cme j geriatr med"] = "CME journal. Geriatric medicine",
  ["cmes-comp model eng sci"] = "CMES Computer Modeling in Engineering and Sciences",
  ["cmi lect ser math"] = "CMI Lecture Series in Mathematics",
  ["cms books math/ouvrages math smc"] = "CMS Books in Mathematics/Ouvrages de Mathematiques de la SMC",
  ["cms conf proc"] = "CMS Conference Proceedings",
  ["cms/caims books math"] = "CMS/CAIMS Books in Mathematics",
  ["cna bull"] = "CNA bulletin",
  ["cnl nucl rev"] = "CNL Nuclear Review",
  ["cns drug rev"] = "CNS drug reviews",
  ["cns drugs"] = "CNS drugs",
  ["cns neurol disord drug targets"] = "CNS & neurological disorders drug targets",
  ["cns neurol disord: drug targets"] = "CNS & Neurological Disorders: Drug Targets",
  ["cns neurosci ther"] = "CNS neuroscience & therapeutics",
  ["cns oncol"] = "CNS oncology",
  ["cns spectr"] = "CNS spectrums",
  ["coal prep (philadelphia, pa, u s)"] = "Coal Preparation (Philadelphia, PA, United States)",
  ["coast manage"] = "Coastal management : an international journal of marine environment, resources, law, and society",
  ["coast mar sci"] = "Coastal marine science",
  ["coastal eng"] = "Coastal Engineering",
  ["coastal eng j"] = "Coastal Engineering Journal",
  ["coastal estuarine stud"] = "Coastal and Estuarine Studies",
  ["coastal manage"] = "Coastal Management",
  ["coatings (basel)"] = "Coatings (Basel, Switzerland)",
  ["cochlear implants int"] = "Cochlear implants international",
  ["cochrane database syst rev"] = "The Cochrane database of systematic reviews",
  ["cochrane evidence synth methods"] = "Cochrane Evidence Synthesis and Methods",
  ["cochrane libr"] = "The Cochrane library",
  ["code ala 1975 ala"] = "Code of Alabama, 1975 : with provision for subsequent pocket parts. Alabama",
  ["code fed regul"] = "Code of Federal regulations",
  ["code fed regul public welfare"] = "Code of Federal regulations. 45: Public welfare",
  ["code fed regul shipping"] = "Code of federal regulations. 46, Shipping",
  ["code ga annot ga"] = "Code of Georgia annotated : including Code of 1981; including the Code of 1933, the Code of 1981 and all laws of general application subsequently enacted, with editorial notes and complete annotations of the decisions of the Supreme Court and Court of Appeals of Georgia and of all the Federal courts construing Georgia laws. Georgia",
  ["code laws s c 1976 annot s c"] = "Code of laws of South Carolina, 1976, annotated : containing permanent public statutes of general application to the end of the 1975 Legislative session : prepared under the supervision of the Legislative Council and Code Commissioner and adopted as the official code of the State by the 1977 session of the General Assembly. South Carolina",
  ["code va 1950 va"] = "Code of Virginia, 1950 : With provisions for subsequent pocket parts. Annotated. Virginia",
  ["coeur med interne"] = "Coeur et médecine interne",
  ["coevol q"] = "The Coevolution quarterly",
  ["coffee sci"] = "Coffee Science",
  ["cogener distrib gener j"] = "Cogeneration & Distributed Generation Journal",
  ["cogent arts humanit"] = "Cogent arts & humanities",
  ["cogent biol"] = "Cogent biology",
  ["cogent chem"] = "Cogent chemistry",
  ["cogent eng"] = "Cogent engineering",
  ["cogent environ sci"] = "Cogent environmental science",
  ["cogent food agric"] = "Cogent food & agriculture",
  ["cogent geosci"] = "Cogent geoscience",
  ["cogent math stat"] = "Cogent mathematics & statistics",
  ["cogent med"] = "Cogent medicine",
  ["cogent phys"] = "Cogent Physics",
  ["cogent psychol"] = "Cogent psychology",
  ["cogent soc sci"] = "Cogent social sciences",
  ["cogent social sci"] = "Cogent Social Science",
  ["cogn affect behav neurosci"] = "Cognitive, affective & behavioral neuroscience",
  ["cogn behav neurol"] = "Cognitive and behavioral neurology : official journal of the Society for Behavioral and Cognitive Neurology",
  ["cogn behav pract"] = "Cognitive and behavioral practice",
  ["cogn behav ther"] = "Cognitive behaviour therapy",
  ["cogn behav therap"] = "Cognitive behaviour therapist",
  ["cogn brain behav"] = "Cognition, brain, behavior : an interdisciplinary journal",
  ["cogn dev"] = "Cognitive development",
  ["cogn emot"] = "Cognition & emotion",
  ["cogn instr"] = "Cognition and instruction",
  ["cogn intell robot"] = "Cognitive Intelligence and Robotics",
  ["cogn linguist"] = "Cognitive linguistics",
  ["cogn neurodyn"] = "Cognitive neurodynamics",
  ["cogn neuropsychiatry"] = "Cognitive neuropsychiatry",
  ["cogn neuropsychol"] = "Cognitive neuropsychology",
  ["cogn neurosci"] = "Cognitive neuroscience",
  ["cogn process"] = "Cognitive processing",
  ["cogn psychol"] = "Cognitive psychology",
  ["cogn res princ implic"] = "Cognitive research: principles and implications",
  ["cogn sci"] = "Cognitive science",
  ["cogn sci (hauppauge)"] = "Cognitive sciences",
  ["cogn sci technol"] = "Cognitive Science and Technology",
  ["cogn syst monogr"] = "Cognitive Systems Monographs",
  ["cogn syst res"] = "Cognitive systems research",
  ["cogn technol work"] = "Cognition, technology & work (Online)",
  ["cognit comput"] = "Cognitive Computation",
  ["cognit comput syst"] = "Cognitive Computation and Systems",
  ["cognit neurodyn"] = "Cognitive Neurodynamics",
  ["cognit neuropsychiatry"] = "Cognitive neuropsychiatry",
  ["cognit neurosci"] = "Cognitive Neuroscience",
  ["cognit process"] = "Cognitive Processing",
  ["cognit psychol"] = "Cognitive Psychology",
  ["cognit res: princ implic"] = "Cognitive Research: Principles and Implications",
  ["cognit rob"] = "Cognitive Robotics",
  ["cognit sci"] = "Cognitive Science",
  ["cognit syst res"] = "Cognitive Systems Research",
  ["cognit technol work"] = "Cognition, Technology & Work",
  ["cognit ther res"] = "Cognitive therapy and research",
  ["coimbra math texts"] = "Coimbra Mathematical Texts",
  ["coimbra med"] = "Coimbra médica",
  ["coin world"] = "Coin world",
  ["coke chem"] = "Coke and Chemistry",
  ["colby libr q"] = "The Colby Library quarterly",
  ["cold reg sci technol"] = "Cold Regions Science and Technology",
  ["cold spring harb mol case stud"] = "Cold Spring Harbor molecular case studies",
  ["cold spring harb perspect biol"] = "Cold Spring Harbor perspectives in biology",
  ["cold spring harb perspect med"] = "Cold Spring Harbor perspectives in medicine",
  ["cold spring harb protoc"] = "Cold Spring Harbor protocols",
  ["cold spring harb symp quant biol"] = "Cold Spring Harbor symposia on quantitative biology",
  ["cold spring harbor mol case stud"] = "Cold Spring Harbor Molecular Case Studies",
  ["cold spring harbor perspect biol"] = "Cold Spring Harbor Perspectives in Biology",
  ["cold spring harbor symp quant biol"] = "Cold Spring Harbor Symposia on Quantitative Biology",
  ["cold war hist"] = "Cold War history",
  ["colec enrique pérez arbeláez"] = "Colección Enrique Pérez Arbeláez",
  ["colec mat apl"] = "Colecao Matematica e Aplicacoes",
  ["colec mat univ"] = "Colecao Matematica Universitaria",
  ["colec mem"] = "Colección Memorias",
  ["colecc congr"] = "Colección: Congresos",
  ["colecc estud"] = "Coleccion Estudio",
  ["colecc estud cieplan"] = "Colección estudios CIEPLAN",
  ["colecc hist cienc salud"] = "Colección histórica de ciencias de la salud",
  ["colecc julio carrizosa valenzuela"] = "Coleccion Julio Carrizosa Valenzuela",
  ["colecc obra sel"] = "Coleccion Obra Selecta",
  ["colección estud cieplan"] = "Colección Estudios CIEPLAN",
  ["colect univ (cluj-napoca)"] = "Colectia Universitaria (Cluj-Napoca)",
  ["colecţ univ"] = "Colecţia Universitaria",
  ["coleopt bull"] = "The Coleopterists' bulletin",
  ["coleopterists bull"] = "Coleopterists Bulletin",
  ["coleç cle"] = "Coleção CLE",
  ["coll antropol"] = "Collegium antropologicum",
  ["coll logicum ann kurt-gödel-soc"] = "Collegium Logicum",
  ["coll math j"] = "College Mathematics Journal",
  ["coll nabla"] = "Collection NABLA",
  ["coll relat res"] = "Collagen and related research",
  ["coll res libr"] = "College and research libraries",
  ["coll rev"] = "College review (Denver, Colo.)",
  ["coll sci mat"] = "Collana di Scienze Matematiche",
  ["coll stud j"] = "College student journal",
  ["coll studi"] = "Collana di Studi",
  ["coll univ"] = "College and university : the journal of the American Association of Collegiate Registrars",
  ["coll works cardiopulm dis"] = "Collected works on cardio-pulmonary disease",
  ["collab anthropol"] = "Collaborative anthropologies",
  ["collaborations (coral gables)"] = "Collaborations (Coral Gables, Fla.)",
  ["collabra psychol"] = "Collabra. Psychology",
  ["collana monogr gazz med sicil"] = "Collana monografica. Gazzetta medica siciliana",
  ["collect aléa-saclay monogr texts statist phys"] = "Collection Aléa-Saclay: Monographs and Texts in Statistical Physics",
  ["collect autom base"] = "Collection Automatique de Base",
  ["collect blain faye martin"] = "Collection Blain Faye Martin",
  ["collect bot"] = "Collectanea botánica",
  ["collect c r geosci"] = "Collection Comptes rendus. Géoscience",
  ["collect czech chem commun"] = "Collection of Czechoslovak Chemical Communications",
  ["collect czechoslov chem commun"] = "Collection of Czechoslovak chemical communications",
  ["collect didact math"] = "Collection “Didactiques”. Mathematiques",
  ["collect droit adm locale"] = "Collection Droit de l’Administration Locale",
  ["collect dyn"] = "Collective dynamics",
  ["collect essais"] = "Collection Essais",
  ["collect etud class"] = "Collection d'études classiques",
  ["collect hist pensée"] = "Collection Histoire de la Pensée",
  ["collect idees"] = "Collection Idees",
  ["collect ime"] = "Collection de l’I.M.E.",
  ["collect indust mec"] = "Collection des Industries Mecaniques",
  ["collect info sup"] = "Collection Info Sup",
  ["collect inform"] = "Collection Informatique",
  ["collect inform prepas-univ"] = "Collection Informatique Prepas-Universite",
  ["collect math"] = "Collectanea Mathematica",
  ["collect meta"] = "Collection META",
  ["collect méthodes"] = "Collection Méthodes",
  ["collect n s hist soc"] = "Collections of the Nova Scotia Historical Society",
  ["collect notes internes dir etudes rech math inform telecomm"] = "Collection de Notes Internes de la Direction des Etudes et Recherches: Mathematiques, Informatique, Telecommunications",
  ["collect papers mayo clinic mayo found"] = "Collected papers of the Mayo Clinic and the Mayo Foundation. Mayo Clinic",
  ["collect points ser sci"] = "Collection Points: Serie Sciences",
  ["collect regards sci"] = "Collection Regards sur la Science",
  ["collect sci hist"] = "Collection Sciences dans l’Histoire",
  ["collect see"] = "Collection SEE",
  ["collect sem inria"] = "Collection Seminaires INRIA",
  ["collect studies ser"] = "Collected Studies Series",
  ["collect trav acad int hist sci"] = "Collection de Travaux de l’Académie Internationale d’Histoire des Sciences",
  ["collect univ france"] = "Collection des Universités de France",
  ["collect vis sci"] = "Collection Visions des Sciences",
  ["collect works"] = "Collected Works",
  ["collect études"] = "Collection études",
  ["college math j"] = "The College Mathematics Journal",
  ["colloid interface sci commun"] = "Colloid and Interface Science Communications",
  ["colloid j"] = "Colloid Journal",
  ["colloid polym sci"] = "Colloid and Polymer Science",
  ["colloids interface sci commun"] = "Colloids and Interface Science Communications",
  ["colloids surf"] = "Colloids and Surfaces",
  ["colloids surf a physicochem eng asp"] = "Colloids and surfaces. A, Physicochemical and engineering aspects",
  ["colloids surf b biointerfaces"] = "Colloids and surfaces. B, Biointerfaces",
  ["colloids surf, a"] = "Colloids and Surfaces A: Physicochemical and Engineering Aspects",
  ["colloids surf, b"] = "Colloids and Surfaces B: Biointerfaces",
  ["colloq int cnrs"] = "Colloques internationaux. Centre national de la recherche scientifique (France)",
  ["colloq math"] = "Colloquium Mathematicum",
  ["colloq math soc jános bolyai"] = "Colloquia Mathematica Societatis János Bolyai",
  ["colloq nat cnrs"] = "Colloques Nationaux du Centre National de la Recherche Scientifique",
  ["colloq natx cent natl rech sci"] = "Colloques nationaux. Centre national de la recherche scientifique (France)",
  ["colloq semin inst fr rech sci dev coop"] = "Colloques et sʹeminaires - Institut franc̦ais de recherche scientifique pour le dʹeveloppement en coopʹeration",
  ["coln latin am hist rev"] = "Colonial Latin American historical review : CLAHR",
  ["colo herit"] = "Colorado heritage : the journal of the Colorado Historical Society",
  ["colo j int environ law policy"] = "Colorado journal of international environmental law and policy",
  ["colo lawyer"] = "Colorado lawyer",
  ["colo mag"] = "The Colorado magazine",
  ["colo med"] = "Colorado medicine",
  ["colo nurse"] = "Colorado nurse",
  ["colo q"] = "Colorado quarterly",
  ["colo revis statut 1973 colo"] = "Colorado revised statutes, 1973. Colorado",
  ["colomb appl linguist j"] = "Colombian applied linguistics journal",
  ["colomb med"] = "Colombia médica",
  ["colomb med (cali)"] = "Colombia médica (Cali, Colombia)",
  ["colon latin am rev"] = "Colonial Latin American review",
  ["colon plant anim prod"] = "Colonial Plant and Animal Products",
  ["color res appl"] = "Color research and application",
  ["color technol"] = "Coloration Technology",
  ["colorectal cancer"] = "Colorectal cancer",
  ["colorectal dis"] = "Colorectal disease : the official journal of the Association of Coloproctology of Great Britain and Ireland",
  ["columbia forum"] = "Columbia forum",
  ["columbia human rights law rev"] = "Columbia human rights law review",
  ["columbia j gend law"] = "Columbia journal of gender and law",
  ["columbia j law soc probl"] = "Columbia journal of law and social problems",
  ["columbia j rev"] = "Columbia journalism review",
  ["columbia j world bus"] = "The Columbia journal of world business",
  ["columbia law rev"] = "Columbia law review",
  ["columbia optom"] = "The Columbia optometrist",
  ["columbia sci technol law rev"] = "The Columbia science and technology law review",
  ["columbus dent soc bull"] = "Bulletin. Columbus Dental Society",
  ["comb chem"] = "Combinatorial Chemistry - an Online Journal",
  ["comb chem high throughput screen"] = "Combinatorial chemistry & high throughput screening",
  ["comb chem high throughput screening"] = "Combinatorial Chemistry and High Throughput Screening",
  ["comb number theory"] = "Combinatorics and Number Theory",
  ["comb optim"] = "Combinatorial Optimization",
  ["comb probab comput"] = "Combinatorics, Probability and Computing",
  ["comb theory"] = "Combinatorial Theory",
  ["combin probab comput"] = "Combinatorics, Probability and Computing",
  ["combust explos shock waves"] = "Combustion, explosion, and shock waves",
  ["combust flame"] = "Combustion and Flame",
  ["combust sci technol"] = "Combustion Science and Technology",
  ["combust theor model"] = "Combustion Theory and Modelling",
  ["combust theory model"] = "Combustion Theory and Modelling",
  ["comercio exterior"] = "Comercio Exterior",
  ["comertul mod"] = "Comertul Modern",
  ["comm algebra"] = "Communications in Algebra",
  ["comm amer math soc"] = "Communications of the American Mathematical Society",
  ["comm anal geom"] = "Communications in Analysis and Geometry",
  ["comm anal theory contin fractions"] = "Communications in the Analytic Theory of Continued Fractions",
  ["comm app math comp sci"] = "Communications in applied mathematics and computational science",
  ["comm appl nonlinear anal"] = "Communications on Applied Nonlinear Analysis",
  ["comm cogn"] = "Communication and Cognition",
  ["comm cognition monogr"] = "Communication & Cognition Monographies",
  ["comm control eng ser"] = "Communications and Control Engineering Series",
  ["comm control engrg ser"] = "Communications and Control Engineering Series",
  ["comm fac sci univ ankara ser a1 math statist"] = "University of Ankara",
  ["comm math phys"] = "Communications in Mathematical Physics",
  ["comm numer methods eng"] = "Communications in Numerical Methods in Engineering",
  ["comm partial differential equations"] = "Communications in Partial Differential Equations",
  ["comm pure appl math"] = "Communications on Pure and Applied Mathematics",
  ["comm statist simulation comput"] = "Communications in Statistics. Simulation and Computation",
  ["comm statist stochastic models"] = "Communications in Statistics",
  ["comm statist theory methods"] = "Communications in Statistics. Theory and Methods",
  ["comm theoret phys"] = "Communications in Theoretical Physics",
  ["comm theoret phys (allahabad)"] = "Communications in Theoretical Physics (Allahabad)",
  ["commagene j biol"] = "Commagene journal of biology",
  ["command hist rep"] = "Command historical report. Naval Medical Research Institute (U.S.)",
  ["comment math"] = "Commentationes Mathematicae. Annals of the Polish Mathematical Society. Series I",
  ["comment math helv"] = "Commentarii Mathematici Helvetici",
  ["comment math prace mat"] = "Polskiego Towarzystwa Matematycznego",
  ["comment math univ carolin"] = "Commentationes Mathematicae Universitatis Carolinae",
  ["comment math univ st paul"] = "Commentarii Mathematici Universitatis Sancti Pauli",
  ["comment math univ st pauli"] = "Commentarii Mathematici Universitatis Sancti Pauli",
  ["comments astrophys"] = "Comments on Astrophysics",
  ["comments astrophys space phys"] = "Comments on Astrophysics and Space Physics",
  ["comments at mol phys"] = "Comments on Atomic and Molecular Physics",
  ["comments condens matter phys"] = "Comments on Condensed Matter Physics",
  ["comments inorg chem"] = "Comments on Inorganic Chemistry",
  ["comments mod chem a comments inorg chem"] = "Comments on modern chemistry. Part A, Comments on inorganic chemistry : a journal of critical discussion of the current literature",
  ["comments mod phys"] = "Comments on Modern Physics",
  ["comments nucl part phys"] = "Comments on Nuclear and Particle Physics",
  ["comments plasma phys controlled fusion"] = "Comments on Plasma Physics and Controlled Fusion",
  ["comments solid state phys"] = "Comments on Solid State Physics",
  ["commer law eur"] = "Commercial laws of Europe",
  ["commer law j"] = "Commercial law journal",
  ["common factor"] = "Common factor (Stoughton, Mass.)",
  ["common law world rev"] = "Common law world review",
  ["common place"] = "A common place",
  ["common sense pest control q"] = "Common Sense Pest Control Quarterly",
  ["commonhealth (phila)"] = "Commonhealth (Philadelphia, Pa.)",
  ["commonw comp polit"] = "Commonwealth & comparative politics",
  ["commonw for rev"] = "Commonwealth Forestry Review, The",
  ["commun acm"] = "Communications of the ACM",
  ["commun agric appl biol sci"] = "Communications in Agricultural and Applied Biological Sciences",
  ["commun algebra"] = "Communications in algebra",
  ["commun am math soc"] = "Communications of the American Mathematical Society",
  ["commun anal comput"] = "Communications on Analysis and Computation",
  ["commun anal geom"] = "Communications in Analysis and Geomtry",
  ["commun anal mech"] = "Communications in Analysis and Mechanics",
  ["commun appl anal"] = "Communications in applied analysis",
  ["commun appl ind math"] = "Communications in Applied and Industrial Mathematics",
  ["commun appl math comput"] = "Communications on Applied Mathematics and Computation",
  ["commun appl math comput sci"] = "Communications in Applied Mathematics and Computational Science",
  ["commun behav biol"] = "Communications in behavioral biology. Part A: [Original articles]",
  ["commun biol"] = "Communications biology",
  ["commun chem"] = "Communications chemistry",
  ["commun comb optim"] = "Communications in Combinatorics and Optimization",
  ["commun comput inf sci"] = "Communications in Computer and Information Science",
  ["commun comput phys"] = "Communications in Computational Physics",
  ["commun contemp math"] = "Communications in Contemporary Mathematics",
  ["commun cult crit"] = "Communication, culture & critique",
  ["commun cybern syst sci eng"] = "Communications in Cybernetics, Systems Science and Engineering",
  ["commun dis intell"] = "Communicable diseases intelligence",
  ["commun dis intell (2018)"] = "Communicable diseases intelligence (2018)",
  ["commun dis intell q rep"] = "Communicable diseases intelligence quarterly report",
  ["commun dis public health"] = "Communicable disease and public health",
  ["commun dis rep cdr rev"] = "Communicable disease report. CDR review",
  ["commun dis rep cdr suppl"] = "Communicable disease report. CDR supplement",
  ["commun dis rep cdr wkly"] = "Communicable disease report. CDR weekly",
  ["commun disord q"] = "Communication disorders quarterly",
  ["commun disord rev"] = "Communicative disorders review",
  ["commun earth environ"] = "Communications earth & environment",
  ["commun educ"] = "Communication education",
  ["commun eng"] = "Communications Engineering",
  ["commun fac sci univ ank ser a1 math stat"] = "Communications. Faculty of Sciences. University of Ankara. Series A1. Mathematics and Statistics",
  ["commun fac sci univ ank sér a1 math stat"] = "Communications",
  ["commun hist artis med suppl"] = "Communicationes de historia artis medicinae. Suppl",
  ["commun inf syst"] = "Communications in information and systems",
  ["commun inst for fenn"] = "Communicationes Instituti Forestalis Fenniae",
  ["commun integr biol"] = "Communicative & integrative biology",
  ["commun korean math soc"] = "Korean Mathematical Society. Communications",
  ["commun mater"] = "Communications materials",
  ["commun math"] = "Communications in Mathematics",
  ["commun math anal"] = "Communications in Mathematical Analysis",
  ["commun math anal appl"] = "Communications in Mathematical Analysis and Applications",
  ["commun math phys"] = "Communications in Mathematical Physics",
  ["commun math res"] = "Communications in Mathematical Research",
  ["commun math sci"] = "Communications in Mathematical Sciences",
  ["commun math stat"] = "Communications in Mathematics and Statistics",
  ["commun med"] = "Communication & medicine",
  ["commun methods meas"] = "Communication methods and measures",
  ["commun monogr"] = "Communication monographs",
  ["commun newsl"] = "Communique Newsletter",
  ["commun nonlinear sci numer simul"] = "Communications in Nonlinear Science and Numerical Simulation",
  ["commun number theory phys"] = "Communications in Number Theory and Physics",
  ["commun numer anal"] = "Communications in Numerical Analysis",
  ["commun numer methods eng"] = "Communications in Numerical Methods in Engineering",
  ["commun nurs res"] = "Communicating nursing research",
  ["commun partial differ equations"] = "Communications in Partial Differential Equations",
  ["commun phys"] = "Communications physics",
  ["commun psychopharmacol"] = "Communications in psychopharmacology",
  ["commun pure appl anal"] = "Communications on Pure and Applied Analysis",
  ["commun pure appl math"] = "Communications on Pure and Applied Mathematics",
  ["commun q"] = "Communication quarterly",
  ["commun res rep"] = "Communication research reports : CRR",
  ["commun sci disord"] = "Communication sciences & disorders (Seoul, Korea)",
  ["commun soil sci plant anal"] = "Communications in Soil Science and Plant Analysis",
  ["commun stat appl methods"] = "Communications for statistical applications and methods",
  ["commun stat case stud data anal appl"] = "Communications in statistics. Case studies, data analysis and applications",
  ["commun stat simul comput"] = "Communications in statistics: Simulation and computation",
  ["commun stat theory methods"] = "Communications in statistics: theory and methods",
  ["commun stat- simul comput"] = "Communications in Statistics - Simulation and Computation",
  ["commun stat- theory methods"] = "Communications in Statistics - Theory and Methods",
  ["commun stat: case stud data anal appl"] = "Communications in Statistics: Case Studies, Data Analysis and Applications",
  ["commun stoch anal"] = "Communications on Stochastic Analysis",
  ["commun stud"] = "Communication studies",
  ["commun surveys tuts"] = "Communications Surveys and Tutorials",
  ["commun theor phys"] = "Communications in Theoretical Physics",
  ["commun theor phys (beijing)"] = "Communications in Theoretical Physics",
  ["commun theory"] = "Communication theory : CT : a journal of the International Communication Association",
  ["commun transp res"] = "Communications in Transportation Research",
  ["commun yearb"] = "Communication yearbook",
  ["communal soc"] = "Communal societies",
  ["communic res"] = "Communication research",
  ["communications strategies"] = "Communications and Strategies",
  ["communique (wash dc)"] = "Communique",
  ["communist econ"] = "Communist economies",
  ["communist econ econ transform"] = "Communist economies and economic transformation",
  ["communist economies econ transformation"] = "Communist Economies and Economic Transformation",
  ["communist post-communist stud"] = "Communist and post-communist studies",
  ["community based public health policy pract"] = "Community-based public health policy & practice",
  ["community coll j res pract"] = "Community college journal of research and practice",
  ["community coll rev"] = "Community college review",
  ["community dent health"] = "Community dental health",
  ["community dent oral epidemiol"] = "Community dentistry and oral epidemiology",
  ["community dev (columb)"] = "Community development (Columbus, Ohio)",
  ["community dev j"] = "Community development journal",
  ["community ecol"] = "Community ecology : CE",
  ["community eye health"] = "Community eye health",
  ["community genet"] = "Community genetics",
  ["community health (bristol)"] = "Community health (Bristol, England)",
  ["community health stud"] = "Community health studies",
  ["community jr coll j"] = "Community and junior college journal",
  ["community lit j"] = "Community literacy journal",
  ["community med"] = "Community medicine",
  ["community ment health j"] = "Community mental health journal",
  ["community ment health rev"] = "Community mental health review",
  ["community nurs"] = "Community nursing",
  ["community nurse"] = "Community nurse",
  ["community oncol"] = "Community oncology",
  ["community outlook"] = "Community outlook",
  ["community pract"] = "Community practitioner : the journal of the Community Practitioners' & Health Visitors' Association",
  ["community psychol"] = "Community psychologist",
  ["community psychol glob perspect"] = "Community psychology in global perspective",
  ["community work fam"] = "Community, work & family",
  ["comp biochem physiol"] = "Comparative biochemistry and physiology",
  ["comp biochem physiol a"] = "Comparative Biochemistry and Physiology. A, Comparative Physiology",
  ["comp biochem physiol a comp physiol"] = "Comparative biochemistry and physiology. A, Comparative physiology",
  ["comp biochem physiol a mol integr physiol"] = "Comparative biochemistry and physiology. Part A, Molecular & integrative physiology",
  ["comp biochem physiol a physiol"] = "Comparative biochemistry and physiology. Part A, Physiology",
  ["comp biochem physiol a: mol integr physiol"] = "Comparative Biochemistry and Physiology Part A: Molecular & Integrative Physiology",
  ["comp biochem physiol a: physiol"] = "Comparative Biochemistry and Physiology Part A: Physiology",
  ["comp biochem physiol b"] = "Comparative biochemistry and physiology. B, Comparative biochemistry",
  ["comp biochem physiol b biochem mol biol"] = "Comparative biochemistry and physiology. Part B, Biochemistry & molecular biology",
  ["comp biochem physiol b: biochem mol biol"] = "Comparative Biochemistry and Physiology Part B: Biochemistry and Molecular Biology",
  ["comp biochem physiol b: comp biochem"] = "Comparative Biochemistry and Physiology Part B: Comparative Biochemistry",
  ["comp biochem physiol biochem mol biol"] = "Comparative biochemistry and physiology. Biochemistry and molecular biology",
  ["comp biochem physiol c"] = "Comparative Biochemistry and Physiology. C, Comparative Pharmacology",
  ["comp biochem physiol c comp pharmacol"] = "Comparative biochemistry and physiology. C: Comparative pharmacology",
  ["comp biochem physiol c comp pharmacol toxicol"] = "Comparative biochemistry and physiology. C, Comparative pharmacology and toxicology",
  ["comp biochem physiol c pharmacol toxicol endocrinol"] = "Comparative biochemistry and physiology. Part C, Pharmacology, toxicology & endocrinology",
  ["comp biochem physiol c toxicol pharmacol"] = "Comparative biochemistry and physiology. Toxicology & pharmacology : CBP",
  ["comp biochem physiol c: pharmacol toxicol endocrinol"] = "Comparative Biochemistry and Physiology Part C: Pharmacology Toxicology and Endocrinology",
  ["comp biochem physiol c: toxicol pharmacol"] = "Comparative Biochemistry and Physiology Part C: Toxicology & Pharmacology",
  ["comp biochem physiol comp physiol"] = "Comparative biochemistry and physiology. Comparative physiology",
  ["comp biochem physiol d: genomics proteomics"] = "Comparative Biochemistry and Physiology Part D: Genomics and Proteomics",
  ["comp biochem physiol part d genomics proteomics"] = "Comparative biochemistry and physiology. Part D, Genomics & proteomics",
  ["comp biochem physiol pharmacol toxicol endocrinol"] = "Comparative biochemistry and physiology. Pharmacology, toxicology and endocrinology",
  ["comp biochem physiol physiol"] = "Comparative biochemistry and physiology. Physiology",
  ["comp biochem physiol, a: comp physiol"] = "Comparative Biochemistry and Physiology, A: Comparative Physiology",
  ["comp biochem physiol, b: comp biochem"] = "Comparative Biochemistry and Physiology, B: Comparative Biochemistry",
  ["comp biochem physiol, c: comp pharmacol toxicol"] = "Comparative Biochemistry and Physiology, C: Comparative Pharmacology and Toxicology",
  ["comp biochem physiol, part a: mol integr physiol"] = "Comparative Biochemistry and Physiology, Part A: Molecular & Integrative Physiology",
  ["comp biochem physiol, part b: biochem mol biol"] = "Comparative Biochemistry and Physiology, Part B: Biochemistry & Molecular Biology",
  ["comp biochem physiol, part c: toxicol pharmacol"] = "Comparative Biochemistry and Physiology, Part C: Toxicology & Pharmacology",
  ["comp biochem physiol, part d: genomics proteomics"] = "Comparative Biochemistry and Physiology, Part D: Genomics & Proteomics",
  ["comp clin path"] = "Comparative clinical pathology",
  ["comp clin pathol"] = "Comparative Clinical Pathology",
  ["comp cogn behav rev"] = "Comparative cognition & behavior reviews",
  ["comp cogn neurosci"] = "Comparative Cognition and Neuroscience",
  ["comp cytogenet"] = "Comparative cytogenetics",
  ["comp econ stud"] = "Comparative economic studies",
  ["comp educ"] = "Comparative education",
  ["comp educ rev"] = "Comparative education review",
  ["comp eff res (auckl)"] = "Comparative effectiveness research (Auckland, N.Z.)",
  ["comp exerc physiol"] = "Comparative exercise physiology",
  ["comp exercise physiol"] = "Comparative Exercise Physiology",
  ["comp exptl methods"] = "Moving Boundaries Vii: Computational Modelling of Free and Moving Boundary Problems",
  ["comp funct genomics"] = "Comparative and functional genomics",
  ["comp gen pharmacol"] = "Comparative and general pharmacology",
  ["comp hepatol"] = "Comparative hepatology",
  ["comp immunol microbiol infect dis"] = "Comparative immunology, microbiology and infectious diseases",
  ["comp int law j south afr"] = "The Comparative and international law journal of southern Africa = Tydskrif vir regsvergelyking en internasionale reg van Suidelik Afrika = Jornal de direito comparativo e internacional para os paises do sul da Africa = Journal de droit comparé et international des pays de l'Afrique Australe = Zeitschrift für Rechtsvergleichung und internationales Recht des Südlichen Afrika",
  ["comp labor law j"] = "Comparative labor law journal : a publication of the U.S. National Branch of the International Society for Labor Law and Social Security [and] the Wharton School, and the Law School of the University of Pennsylvania",
  ["comp med"] = "Comparative medicine",
  ["comp med east west"] = "Comparative medicine East and West",
  ["comp migr stud"] = "Comparative migration studies",
  ["comp model eng sci"] = "Computer Modeling in Engineering and Sciences",
  ["comp parasitol"] = "Comparative parasitology",
  ["comp polit"] = "Comparative politics",
  ["comp polit stud"] = "Comparative political studies",
  ["comp soc res"] = "Comparative social research",
  ["comp stud soc hist"] = "Comparative studies in society and history",
  ["comp synth creat sys"] = "Computational Synthesis and Creative Systems",
  ["comp urban res"] = "Comparative urban research",
  ["compact textb math"] = "Compact Textbooks in Mathematics",
  ["companion (glos)"] = "Companion",
  ["companion anim"] = "Companion animal",
  ["compar econ stud"] = "Comparative Economic Studies",
  ["comparativ (leipz)"] = "Comparativ : Leipziger Beiträge zur Universalgeschichte und vergleichenden Gesellschaftsforschung",
  ["compend contin educ dent"] = "Compendium of continuing education in dentistry (Jamesburg, N.J. : 1995)",
  ["compend contin educ dent suppl"] = "Compendium of continuing education in dentistry. (Jamesburg, N.J. : 1995). Supplement",
  ["compend contin educ gen dent"] = "The Compendium on continuing education in general dentistry",
  ["compend contin educ vet"] = "Compendium (Yardley, PA)",
  ["compend suppl"] = "Compendium (Newtown, Pa.). Supplement",
  ["compens benefits rev"] = "Compensation and benefits review",
  ["compens med"] = "Compensation medicine",
  ["compens rev"] = "Compensation review",
  ["compet change"] = "Competition & change",
  ["complement health pract rev"] = "Complementary health practice review",
  ["complement inflamm"] = "Complement and inflammation",
  ["complement med res"] = "Complementary medicine research",
  ["complement ther clin pract"] = "Complementary therapies in clinical practice",
  ["complement ther med"] = "Complementary therapies in medicine",
  ["complement ther nurs midwifery"] = "Complementary therapies in nursing & midwifery",
  ["complementi mat indirizzo didàtt"] = "Complementi di Matematica per l’Indirizzo Didàttico",
  ["complex adapt syst model"] = "Complex Adaptive Systems Modeling",
  ["complex adapt systems"] = "Complex Adaptive Systems",
  ["complex anal oper theory"] = "Complex Analysis and Operator Theory",
  ["complex anal synerg"] = "Complex Analysis and its Synergies",
  ["complex anal synergies"] = "Complex Analysis and its Synergies",
  ["complex int"] = "Complexity International",
  ["complex intell syst"] = "Complex & Intelligent Systems",
  ["complex met"] = "Complex Metals",
  ["complex netw appl vii (2018)"] = "Complex networks and their applications VII : proceedings the 7th International Conference on Complex Networks and Their Applications COMPLEX NETWORKS 2018. International Conference on Complex Networks and Their Applications (7th : 2018 : Cambridge, England)",
  ["complex netw dyn syst"] = "Complex Networks and Dynamic Systems",
  ["complex netw xi (2020)"] = "Complex networks XI : Proceedings of the 11th Conference on Complex Networks CompleNet 2020. CompleNet (Workshop) (11th : 2020 : Exeter, England)",
  ["complex psychiatry"] = "Complex psychiatry",
  ["complex syst"] = "Complex Systems",
  ["complex systems"] = "Complex Systems",
  ["complex var elliptic equ"] = "Complex Variables and Elliptic Equations. An International Journal",
  ["complex var elliptic equations"] = "Complex Variables and Elliptic Equations",
  ["complex variables theory appl"] = "Complex Variables",
  ["complexity int"] = "Complexity International",
  ["complicat card patient"] = "Complicated Cardiac Patient",
  ["compos adv mater"] = "Composites and Advanced Materials",
  ["compos b eng"] = "Composites. Part B, Engineering",
  ["compos bus anal"] = "Composites Business Analyst",
  ["compos commun"] = "Composites Communications",
  ["compos eng"] = "Composites Engineering",
  ["compos interfaces"] = "Composite Interfaces",
  ["compos manuf"] = "Composites Manufacturing",
  ["compos math"] = "Compositio Mathematica",
  ["compos newsl"] = "Compositae newsletter",
  ["compos part a appl sci manuf"] = "Composites. Part A, Applied science and manufacturing",
  ["compos sci technol"] = "Composites Science and Technology",
  ["compos struct"] = "Composite Structures",
  ["compos theory pract"] = "Composites Theory and Practice",
  ["compos: mech comput appl: int j"] = "Composites: Mechanics, Computations, Applications: An International Journal",
  ["composites part a"] = "Composites Part A Applied Science and Manufacturing",
  ["composites part b"] = "Composites Part B Engineering",
  ["composites, part a"] = "Composites Part A: Applied Science and Manufacturing",
  ["composites, part b"] = "Composites Part B: Engineering",
  ["composites, part c: open access"] = "Composites Part C: Open Access",
  ["compositio math"] = "Compositio Mathematica",
  ["compost sci util"] = "Compost science & utilization",
  ["compr child adolesc nurs"] = "Comprehensive child and adolescent nursing",
  ["compr course anal"] = "A Comprehensive Course in Analysis",
  ["compr gerontol [a]"] = "Comprehensive Gerontology. Section A, Clinical and Laboratory Sciences",
  ["compr gerontol [b]"] = "Comprehensive Gerontology. Section B, Behavioural, Social and Applied Sciences",
  ["compr gerontol [c]"] = "Comprehensive Gerontology. Section C, Interdisciplinary Topics",
  ["compr gerontol a"] = "Comprehensive gerontology. Section A, Clinical and laboratory sciences",
  ["compr gerontol b"] = "Comprehensive gerontology. Section B, Behavioural, social, and applied sciences",
  ["compr gerontol c"] = "Comprehensive gerontology. Section C, Interdisciplinary topics",
  ["compr immunol"] = "Comprehensive immunology",
  ["compr ophthalmol update"] = "Comprehensive ophthalmology update",
  ["compr physiol"] = "Comprehensive Physiology",
  ["compr psychiatry"] = "Comprehensive psychiatry",
  ["compr psychoneuroendocrinol"] = "Comprehensive psychoneuroendocrinology",
  ["compr results soc psychol"] = "Comprehensive results in social psychology",
  ["compr rev food sci food saf"] = "Comprehensive reviews in food science and food safety",
  ["compr ser photochem photobiol"] = "Comprehensive Series in Photochemistry & Photobiology",
  ["compr ther"] = "Comprehensive therapy",
  ["compre ser photochem photobiol"] = "Comprehensive Series in Photochemistry & Photobiology",
  ["comptes rendus académie sci - ser - math"] = "Comptes Rendus de l’Académie des Sciences - Series I - Mathematics",
  ["comptes rendus académie sci - ser i"] = "Comptes Rendus de l’Académie des Sciences - Series I - Mathematics",
  ["comptes rendus académie sci - ser iv"] = "Comptes Rendus de l’Académie des Sciences - Series IV - Physique et Astrophysique",
  ["comptes rendus hebd séances acad sci paris"] = "Comptes rendus hebdomadaires des séances de l’académie des sciences, Paris",
  ["comptes rendus phys"] = "Comptes Rendus Physique",
  ["comptes rendus séance acad agric fr"] = "Comptes rendus des séances de l’Académie d’Agriculture de France",
  ["comput & fluids"] = "Computers & Fluids. An International Journal",
  ["comput & structures"] = "Computers & Structures",
  ["comput aided des"] = "Computer aided design",
  ["comput aided des appl"] = "Computer-aided design and applications",
  ["comput aided geom des"] = "Computer Aided Geometric Design",
  ["comput aided geom design"] = "Computer Aided Geometric Design",
  ["comput aided surg"] = "Computer aided surgery : official journal of the International Society for Computer Aided Surgery",
  ["comput aided verif"] = "Computer-aided verification : proceedings. CAV (Conference)",
  ["comput anim virtual worlds"] = "Computer Animation and Virtual Worlds",
  ["comput animat"] = "Computer animation. ACM SIGGRAPH Symposium on Computer Animation",
  ["comput animat virtual worlds"] = "Computer animation and virtual worlds",
  ["comput appl biosci"] = "Computer applications in the biosciences : CABIOS",
  ["comput appl eng educ"] = "Computer Applications in Engineering Education",
  ["comput appl math"] = "Computational and Applied Mathematics",
  ["comput archit des methodol"] = "Computer Architecture and Design Methodologies",
  ["comput archit news"] = "Computer architecture news",
  ["comput artif intell"] = "Computers and Artificial Intelligence",
  ["comput assist lang learn"] = "Computer assisted language learning",
  ["comput assist robot endosc (2014)"] = "Computer-assisted and robotic endoscopy : first International Workshop, CARE 2014, held in conjunction with MICCAI 2014, Boston, MA, USA, September 18, 2014 : revised selected papers. CARE (Workshop) (1st : 2014 : Boston, Mass.)",
  ["comput assist robot endosc clin image based proced (2017)"] = "Computer assisted and robotic endoscopy and clinical image-based procedures : 4th International Workshop, CARE 2017, and 6th International Workshop, CLIP 2017, held in conjunction with MICCAI 2017 Québec City, QC, Canada, September 14, 2017, proceedings. CARE (Workshop) (4th : 2017 : Québec, Québec)",
  ["comput assist surg (abingdon)"] = "Computer assisted surgery (Abingdon, England)",
  ["comput assisted lang learn"] = "Computer Assisted Language Learning",
  ["comput assisted surg"] = "Computer Assisted Surgery",
  ["comput astrophys cosmol"] = "Computational astrophysics and cosmology",
  ["comput biol"] = "Computational Biology",
  ["comput biol chem"] = "Computational Biology and Chemistry",
  ["comput biol j"] = "Computational biology journal",
  ["comput biol med"] = "Computers in biology and medicine",
  ["comput biomech med (2017)"] = "Computational Biomechanics for Medicine : Measurements, Models, and Predictions",
  ["comput biomech med algorithms models appl (2017)"] = "Computational biomechanics for medicine. From algorithms to models and applications",
  ["comput biomech med personal valid ther (2019)"] = "Computational biomechanics for medicine. Personalisation, validation and therapy",
  ["comput biomech med soft tiss musculoskelet syst"] = "Computational biomechanics for medicine. Soft tissues and the musculoskeletal system",
  ["comput biomed res"] = "Computers and Biomedical Research",
  ["comput brain behav"] = "Computational brain & behavior",
  ["comput cardiol"] = "Computers in cardiology",
  ["comput cardiol (2010)"] = "Computing in cardiology",
  ["comput chem"] = "Computers & chemistry",
  ["comput chem (oxford)"] = "Computers and Chemistry",
  ["comput chem eng"] = "Computers & Chemical Engineering",
  ["comput cogn sci"] = "Computational cognitive science",
  ["comput cognit sci"] = "Computational Cognitive Science",
  ["comput commun"] = "Computer Communications",
  ["comput commun netw"] = "Computer Communications and Networks",
  ["comput complexity"] = "Computational Complexity",
  ["comput compos"] = "Computers and composition",
  ["comput concr"] = "Computers and Concrete",
  ["comput condens matter"] = "Computational Condensed Matter",
  ["comput diffus mri (2014)"] = "Computational diffusion MRI : MICCAI Workshop, Boston, MA, USA, September 2014. CDMRI (Workshop) (6th : 2014 : Boston, Mass.)",
  ["comput diffus mri (2015)"] = "Computational diffusion MRI : MICCAI Workshop, Munich, Germany, October 9th, 2015. CDMRI (Workshop) (7th : 2015 : Munich, Germany)",
  ["comput diffus mri (2016)"] = "Computational Diffusion MRI : MICCAI workshop, Athens, Greece, October 2016. CDMRI (Workshop) (8th : 2016 : Athens, Greece)",
  ["comput diffus mri brain connect (2013)"] = "Computational Diffusion MRI and Brain Connectivity : MICCAI Workshops, Nagoya, Japan, September 22nd, 2013. MICCAI Workshop on Computation Diffusion MRI (5th : 2013 : Nagoya-shi, Japan)",
  ["comput econ"] = "Computational economics",
  ["comput educ"] = "Computers & education",
  ["comput educ open"] = "Computers and Education Open",
  ["comput educ: artif intell"] = "Computers and Education: Artificial Intelligence",
  ["comput electr eng"] = "Computers and Electrical Engineering",
  ["comput electron agric"] = "Computers and Electronics in Agriculture",
  ["comput eng ser"] = "Computer Engineering Series",
  ["comput environ urban syst"] = "Computers, environment and urban systems",
  ["comput exp methods struct"] = "Computational and Experimental Methods in Structures",
  ["comput fluids"] = "Computers & Fluids",
  ["comput games j"] = "Computer Games Journal",
  ["comput geom"] = "Computational geometry : theory and applications",
  ["comput geom theory appl"] = "Computational Geometry: Theory and Applications",
  ["comput geom topol"] = "Computing in Geometry and Topology (CGT)",
  ["comput geosci"] = "Computers & geosciences",
  ["comput geotech"] = "Computers and Geotechnics",
  ["comput graph"] = "Computers & graphics",
  ["comput graph (acm)"] = "Computer graphics",
  ["comput graph forum"] = "Computer graphics forum : journal of the European Association for Computer Graphics",
  ["comput graph x"] = "Computers & graphics: X",
  ["comput graphics"] = "Computers & Graphics",
  ["comput graphics forum"] = "Computer Graphics Forum",
  ["comput graphics: x"] = "Computers & Graphics: X",
  ["comput healthc"] = "Computers in healthcare",
  ["comput hosp"] = "Computers in hospitals",
  ["comput hum"] = "Computers and the humanities",
  ["comput hum behav"] = "Computers in Human Behavior",
  ["comput hum behav rep"] = "Computers in Human Behavior Reports",
  ["comput human behav"] = "Computers in human behavior",
  ["comput imaging vision"] = "Computational Imaging and Vision",
  ["comput ind"] = "Computers in Industry",
  ["comput ind eng"] = "Computers & Industrial Engineering",
  ["comput inform"] = "Computing and Informatics",
  ["comput inform nurs"] = "Computers, informatics, nursing : CIN",
  ["comput integr manuf syst"] = "Computer Integrated Manufacturing Systems",
  ["comput intell"] = "Computational Intelligence",
  ["comput intell eng probl solving"] = "Computational Intelligence in Engineering Problem Solving",
  ["comput intell methods appl"] = "Computational Intelligence Methods and Applications",
  ["comput intell methods bioinform biostat (2016)"] = "Computational intelligence methods for bioinformatics and biostatistics : 13th International Meeting, CIBB 2016, Stirling, UK, September 1-3, 2016, revised selected papers. CIBB (Meeting) (13th : 2016 : Stirling, England)",
  ["comput intell neurosci"] = "Computational intelligence and neuroscience",
  ["comput intelligence"] = "Computational Intelligence",
  ["comput j"] = "The computer journal",
  ["comput lang syst struct"] = "Computer languages, systems & structures",
  ["comput law j"] = "Computer/law journal",
  ["comput lett"] = "Computing Letters",
  ["comput life"] = "Computer Life",
  ["comput linguist"] = "Computational Linguistics",
  ["comput linguist assoc comput linguist"] = "Computational linguistics (Association for Computational Linguistics)",
  ["comput manag sci"] = "Computational Management Science",
  ["comput mater continua"] = "Computers, Materials & Continua",
  ["comput mater sci"] = "Computational Materials Science",
  ["comput math anal ser"] = "Computational Mathematics and Analysis Series",
  ["comput math appl"] = "Computers & mathematics with applications (Oxford, England : 1987)",
  ["comput math appl (athens)"] = "Hellenic Mathematical Society",
  ["comput math biophys"] = "Computational and mathematical biophysics",
  ["comput math math phys"] = "Computational Mathematics and Mathematical Physics",
  ["comput math methods"] = "Computational and Mathematical Methods",
  ["comput math methods med"] = "Computational and mathematical methods in medicine",
  ["comput math model"] = "Computational Mathematics and Modeling",
  ["comput math organ theory"] = "Computational and mathematical organization theory",
  ["comput mech"] = "Computational Mechanics",
  ["comput med imaging graph"] = "Computerized medical imaging and graphics : the official journal of the Computerized Medical Imaging Society",
  ["comput med imaging graphics"] = "Computerized Medical Imaging and Graphics",
  ["comput med public health biotechnol (1994)"] = "Computational medicine, public health, and biotechnology : building a man in the machine",
  ["comput methods"] = "Computational Methods",
  ["comput methods appl math"] = "Computational Methods in Applied Mathematics",
  ["comput methods appl mech eng"] = "Computer Methods in Applied Mechanics and Engineering",
  ["comput methods appl mech engrg"] = "Computer Methods in Applied Mechanics and Engineering",
  ["comput methods appl sci"] = "Computational methods in applied sciences (Springer)",
  ["comput methods biomech biomed eng"] = "Computer Methods in Biomechanics and Biomedical Engineering",
  ["comput methods biomech biomed eng imaging vis"] = "Computer methods in biomechanics and biomedical engineering. Imaging & visualization",
  ["comput methods biomech biomed eng: imaging visualization"] = "Computer Methods in Biomechanics and Biomedical Engineering: Imaging and Visualization",
  ["comput methods biomech biomed engin"] = "Computer methods in biomechanics and biomedical engineering",
  ["comput methods clin appl musculoskelet imaging (2017)"] = "Computational methods and clinical applications in musculoskeletal imaging : 5th International Workshop, MSKI 2017, held in conjunction with MICCAI 2017, Quebec City, QC, Canada, September 10, 2017, revised selected papers. MSKI (Workshop) (5th : 2017 : Québec, Québec)",
  ["comput methods differ equ"] = "Computational Methods for Differential Equations",
  ["comput methods eng sci"] = "Computational Methods in Engineering & the Sciences",
  ["comput methods funct theory"] = "Computational Methods and Function Theory",
  ["comput methods programs biomed"] = "Computer methods and programs in biomedicine",
  ["comput model eng sci"] = "Computer modeling in engineering & sciences : CMES",
  ["comput model objects present images (2014)"] = "Computational modeling of objects presented in images : fundamentals, methods, and applications : 4th International Conference, CompIMAGE 2014, Pittsburgh, PA, USA, September 3-5, 2014 : proceedings. CompIMAGE (Conference) (4th : 2014 : Pittsburgh, Pa.)",
  ["comput model simul eng"] = "Computer Modeling and Simulation in Engineering",
  ["comput mol biosci"] = "Computational molecular bioscience",
  ["comput music j"] = "Computer Music Journal",
  ["comput music sci"] = "Computational Music Science",
  ["comput networks"] = "Computer Networks",
  ["comput nurs"] = "Computers in nursing",
  ["comput open"] = "Computing Open",
  ["comput oper res"] = "Computers & Operations Research",
  ["comput optim appl"] = "Computational Optimization and Applications",
  ["comput part mech"] = "Computational particle mechanics",
  ["comput pathol ophthalmic med image anal (2018)"] = "Computational pathology and ophthalmic medical image analysis : first International Workshop, COMPAY 2018, and 5th International Workshop, OMIA 2018, held in conjunction with MICCAI 2018, Granada, Spain, September 16-20, 2018, proceedings. COMPAY (Workshop) (1st : 2018 : Granada, Spain)",
  ["comput people ser"] = "Computers and People Series",
  ["comput phys"] = "Computers in Physics",
  ["comput phys comm"] = "Computer Physics Communications",
  ["comput phys commun"] = "Computer Physics Communications",
  ["comput programs biomed"] = "Computer programs in biomedicine",
  ["comput psychiatr"] = "Computational psychiatry (Cambridge, Mass.)",
  ["comput psychiatry"] = "Computational Psychiatry",
  ["comput radiol"] = "Computerized radiology : official journal of the Computerized Tomography Society",
  ["comput sch"] = "Computers in the schools : interdisciplinary journal of practice, theory, and applied research",
  ["comput sci"] = "Computer Science",
  ["comput sci (berl)"] = "Computer science (Berlin, Germany)",
  ["comput sci discov"] = "Computational science & discovery",
  ["comput sci discovery"] = "Computational Science & Discovery",
  ["comput sci educ"] = "Computer Science Education",
  ["comput sci eng"] = "Computing in Science & Engineering",
  ["comput sci found appl log"] = "Computer Science Foundations and Applied Logic",
  ["comput sci j moldova"] = "Academy of Sciences of Moldova. Institute of Mathematics and Computer Science. Computer Science Journal of Moldova",
  ["comput sci lib"] = "The Computer Science Library",
  ["comput sci monogr"] = "Computer Science Monographs",
  ["comput sci rev"] = "Computer Science Review",
  ["comput sci rev trends"] = "Computer Science Reviews and Trends",
  ["comput sci sci comput"] = "Computer Science and Scientific Computing",
  ["comput sci: comput archit design"] = "Computer Science: Computer Architecture and Design",
  ["comput sci: systems programming"] = "Computer Science: Systems Programming",
  ["comput secur"] = "Computers & security",
  ["comput soc netw"] = "Computational social networks",
  ["comput soc sci"] = "Computational Social Sciences",
  ["comput social networks"] = "Computational Social Networks",
  ["comput softw big sci"] = "Computing and software for big science",
  ["comput software big sci"] = "Computing and Software for Big Science",
  ["comput speech lang"] = "Computer Speech & Language",
  ["comput stand interfaces"] = "Computer standards & interfaces",
  ["comput stat"] = "Computational statistics",
  ["comput stat data anal"] = "Computational Statistics and Data Analysis",
  ["comput statist"] = "Computational Statistics",
  ["comput statist data anal"] = "Computational Statistics & Data Analysis",
  ["comput struct"] = "Computers and Structures",
  ["comput struct biotechnol j"] = "Computational and structural biotechnology journal",
  ["comput suppl"] = "Computing Supplementum",
  ["comput support coop work"] = "Computer supported cooperative work : CSCW : an international journal",
  ["comput syst bioinformatics conf"] = "Computational systems bioinformatics. Computational Systems Bioinformatics Conference",
  ["comput syst oncol"] = "Computational and systems oncology",
  ["comput syst sci eng"] = "Computer Systems Science and Engineering",
  ["comput technol appl"] = "Computer technology and application",
  ["comput theor chem"] = "Computational & theoretical chemistry",
  ["comput theor polym sci"] = "Computational and Theoretical Polymer Science",
  ["comput therm sci: int j"] = "Computational Thermal Sciences: An International Journal",
  ["comput tomogr"] = "Computerized tomography",
  ["comput topol image context (2016)"] = "Computational topology in image context : 6th International Workshop, CTIC 2016, Marseille, France, June 15-17, 2016, Proceedings. CTIC (Conference) (6th : 2016 : Marseille, France)",
  ["comput toxicol"] = "Computational toxicology (Amsterdam, Netherlands)",
  ["comput urban sci"] = "Computational Urban Science",
  ["comput vis biomed image appl (2005)"] = "Computer vision for biomedical image applications : first international workshop, CVBIA 2005, Beijing, China, October 21, 2005 : proceedings. CVBIA 2005 (2005 : Beijing, China)",
  ["comput vis graph image process"] = "Computer vision, graphics, and image processing",
  ["comput vis image underst"] = "Computer vision and image understanding : CVIU",
  ["comput vis media (beijing)"] = "Computational visual media",
  ["comput vis sci"] = "Computing And Visualization In Science",
  ["comput vision image understanding"] = "Computer Vision and Image Understanding",
  ["comput visual media"] = "Computational Visual Media",
  ["comput visualization sci"] = "Computing and Visualization in Science",
  ["comput-aided civ infrastruct eng"] = "Computer-Aided Civil and Infrastructure Engineering",
  ["comput-aided des"] = "Computer-Aided Design",
  ["comput-aided des applic"] = "Computer-Aided Design and Applications",
  ["computation (basel)"] = "Computation (Basel, Switzerland)",
  ["computational econ"] = "Computational Economics",
  ["computer (long beach calif)"] = "Computer",
  ["computer sci econ manage"] = "Computer Science in Economics and Management",
  ["comsig rev"] = "COMSIG review",
  ["comun dis rep cdr suppl"] = "Communicable Disease Report. CDR Supplement",
  ["comun jorn com esp deterg"] = "Comunicaciones presentadas a las Jornadas del Comite Espanol de la Detergencia",
  ["comut front conf"] = "Computer Frontiers Conference : [proceedings]. Computer Frontiers Conference",
  ["concepts immunopathol"] = "Concepts in immunopathology",
  ["concepts magn reson"] = "Concepts in Magenetic Resonance",
  ["concepts magn reson part a"] = "Concepts in Magnetic Resonance Part A",
  ["concepts magn reson part a bridg educ res"] = "Concepts in magnetic resonance. Part A, Bridging education and research",
  ["concepts magn reson part b"] = "Concepts in Magnetic Resonance Part B",
  ["concepts magn reson part b magn reson eng"] = "Concepts in magnetic resonance. Part B, Magnetic resonance engineering",
  ["concern (anaheim)"] = "Concern",
  ["concern care aging"] = "Concern in care of the aging",
  ["concern dying"] = "Concern for dying",
  ["concerned demogr"] = "Concerned demography",
  ["concienc latinoam"] = "Conciencia latinoamericana : boletín periódico de opiniones de Católicas por el Derecho a Decidir",
  ["concordia hist inst q"] = "Concordia Historical Institute quarterly. Concordia Historical Institute",
  ["concordia j"] = "Concordia journal",
  ["concordia law rev"] = "Concordia law review",
  ["concours med"] = "Concours médical",
  ["concr eng int"] = "Concrete Engineering International",
  ["concr int"] = "Concrete International",
  ["concr oper"] = "Concrete Operators",
  ["concr sci eng"] = "Concrete Science and Engineering",
  ["concurr comput"] = "Concurrency and computation : practice & experience",
  ["concurr eng res appl"] = "Concurrent engineering, research, and applications",
  ["concurrency comput pract exper"] = "Concurrency and Computation: Practice and Experience",
  ["concurrent eng"] = "Concurrent Engineering",
  ["concurrent eng: res appl"] = "Concurrent Engineering: Research and Applications",
  ["cond med"] = "Conditioning medicine",
  ["cond reflex"] = "Conditional reflex",
  ["condens matter"] = "Condensed matter",
  ["condens matter phys"] = "Condensed Matter Physics",
  ["condens matter res t"] = "Physical Properties of The Low-dimensional A3b6 and A3b3c62 Compounds",
  ["condens matter theor"] = "Condensed Matter Theories",
  ["conditio jud"] = "Conditio Judaica",
  ["condorcet stud"] = "Condorcet studies",
  ["conf adv aids vaccine dev 1997 conf adv aids vaccine dev 1997 bethesda md"] = "Conference on Advances in AIDS Vaccine Development, 1997 : Ninth Annual Meeting of the National Cooperative Vaccine Development Groups for AIDS, May 4-7, 1997 : program book. Conference on Advances in AIDS Vaccine Development (1997 : Bethesda, Md.)",
  ["conf adv aids vaccine dev natl coop vaccine dev groups aids meet 8th 1996 bethesda md"] = "Conference on advances in AIDS vaccine development : program book. National Cooperative Vaccine Development Groups for AIDS. Meeting (8th : 1996 : Bethesda, Md.)",
  ["conf board rec"] = "The Conference Board record",
  ["conf comput vis pattern recognit workshops"] = "Conference on Computer Vision and Pattern Recognition Workshops. IEEE Computer Society Conference on Computer Vision and Pattern Recognition. Workshops",
  ["conf estud hist organ cienc"] = "Conferencias y estudios de historia y organización de la ciencia",
  ["conf hist med"] = "Conférences d'histoire de la médecine : [fascicule]. Conférences d'histoire de la médecine",
  ["conf human syst interact"] = "Conference on Human System Interactions : [proceedings]. International Conference on Human System Interaction",
  ["conf lasers electro optics"] = "Conference on Lasers and Electro-optics : (CLEO). Conference on Lasers and Electro-optics",
  ["conf lyon ophtalmol"] = "Conférences lyonnaises d'ophtalmologie",
  ["conf proc ethnogr prax ind conf"] = "Conference proceedings. Ethnographic Praxis in Industry Conference",
  ["conf proc front educ conf"] = "Conference proceedings. Frontiers in Education Conference",
  ["conf proc ieee int conf signal image process appl"] = "Conference proceedings. IEEE International Conference on Signal and Image Processing Applications",
  ["conf proc ieee int conf syst man cybern"] = "Conference proceedings. IEEE International Conference on Systems, Man, and Cybernetics",
  ["conf proc int conf image form xray comput tomogr"] = "Conference proceedings. International Conference on Image Formation in X-Ray Computed Tomography",
  ["conf proc int res counc biomech inj"] = "Conference proceedings. International Research Council on Biomechanics of Injury",
  ["conf proc lect not g"] = "Geometry, Topology & Physics",
  ["conf proc lecture notes appl phys"] = "Conference Proceedings and Lecture Notes in Applied Physics",
  ["conf proc lecture notes geom topology"] = "Conference Proceedings and Lecture Notes in Geometry and Topology",
  ["conf proc lecture notes phys"] = "Conference Proceedings and Lecture Notes in Physics",
  ["conf proc soc exp mech ser"] = "Conference Proceedings of the Society for Experimental Mechanics Series",
  ["conf proc soc plast eng"] = "Conference proceedings. Society of Plastics Engineers. Technical Conference",
  ["conf proc theory comput"] = "Conference Proceedings on Theory of Computing",
  ["conf rec asilomar conf signals syst comput"] = "Conference record. Asilomar Conference on Signals, Systems & Computers",
  ["conf rec ieee photovoltaic spec conf"] = "Conference Record of the IEEE Photovoltaic Specialists Conference",
  ["conf semin mat univ bari"] = "Conferenze del Seminario di Matematica dell’Università di Bari",
  ["confed aust crit care nurses j"] = "Confederation of Australian Critical Care Nurses journal",
  ["confed state med surg j"] = "Confederate States medical and surgical journal",
  ["confin neurol"] = "Confinia neurologica",
  ["confin psychiatr"] = "Confinia psychiatrica. Borderland of psychiatry. Grenzgebiete der Psychiatrie. Les Confins de la psychiatrie",
  ["confl health"] = "Conflict and health",
  ["conflict manage peace sci"] = "Conflict Management and Peace Science",
  ["confluentes math"] = "Confluentes Mathematici",
  ["conform geom dyn"] = "Conformal Geometry and Dynamics. An Electronic Journal of the American Mathematical Society",
  ["confront radio anat clin"] = "Confrontations Radio-Anatomo-Cliniques",
  ["congenit anom (kyoto)"] = "Congenital anomalies",
  ["congenit heart dis"] = "Congenital heart disease",
  ["congest heart fail"] = "Congestive heart failure (Greenwich, Conn.)",
  ["congr int stomatol"] = "Congrès international de stomatologie",
  ["congr mon"] = "Congress monthly",
  ["congr numer"] = "Congressus Numerantium. A Conference Journal on Numerical Themes",
  ["congr q wkly rep"] = "Congressional quarterly weekly report",
  ["congr soc cir chile"] = "Congreso. Sociedad de Cirujanos de Chile",
  ["conn dent stud j"] = "Connecticut dental student journal",
  ["conn gen statut annot conn"] = "Connecticut General statutes annotated. Connecticut",
  ["conn health bull"] = "Connecticut health bulletin",
  ["conn hist soc bull"] = "The Connecticut Historical Society bulletin",
  ["conn j int law"] = "Connecticut journal of international law",
  ["conn law rev"] = "Connecticut law review",
  ["conn med"] = "Connecticut medicine",
  ["conn nurs news"] = "Connecticut nursing news (Meriden, Conn. : 1980)",
  ["conn probate law j"] = "The Connecticut probate law journal",
  ["conn sci"] = "Connection science",
  ["conn state med j"] = "Connecticut state medical journal",
  ["connaiss arts"] = "Connaissance des arts",
  ["connaiss egypte anc"] = "Connaissance de l’Egypte Ancienne",
  ["connect (tor)"] = "Connections (Toronto, Ont.)",
  ["connect neuroimaging (2018)"] = "Connectomics in neuroImaging : second international workshop, CNI 2018, held in conjunction with MICCAI 2018, Granada, Spain, September 20, 2018 : proceedings. CNI (Workshop) (2nd : 2018 : Granada, Spain)",
  ["connect neuroimaging (2019)"] = "Connectomics in neuroImaging : third International Workshop, CNI 2019, held in conjunction with MICCAI 2019, Shenzhen, China, October 13, 2019, Proceedings. CNI (Workshop) (3rd : 2019 : Shenzhen Shi, China)",
  ["connect sci"] = "Connection Science",
  ["connect tissue res"] = "Connective tissue research",
  ["connectomics neuroimaging (2017)"] = "Connectomics in neuroimaging : First International Workshop, CNI 2017, held in conjunction with MICCAI 2017, Quebec City, QC, Canada, September 14, 2017, proceedings. CNI (Workshop) (1st : 2017 : Québec, Québec)",
  ["conscious cogn"] = "Consciousness and cognition",
  ["consens dev conf summ natl inst health"] = "Consensus development conference summaries",
  ["consens statement"] = "Consensus statement. National Institutes of Health Consensus Development Conference",
  ["conserv biol"] = "Conservation biology : the journal of the Society for Conservation Biology",
  ["conserv biol ser"] = "Conservation biology series",
  ["conserv genet"] = "Conservation genetics (Print)",
  ["conserv genet resour"] = "Conservation genetics resources",
  ["conserv jud"] = "Conservative Judaism",
  ["conserv lett"] = "Conservation letters",
  ["conserv manage archaeol sites"] = "Conservation and Management of Archaeological Sites",
  ["conserv physiol"] = "Conservation physiology",
  ["conserv sci pract"] = "Conservation science and practice",
  ["const comment"] = "Constitutional commentary",
  ["const law j"] = "Constitutional law journal (Newark, N.J. : 1990)",
  ["constit polit economy"] = "Constitutional Political Economy",
  ["constr approx"] = "Constructive approximation",
  ["constr build mater"] = "Construction and Building Materials",
  ["constr hist"] = "Construction history : journal of the Construction History Group",
  ["constr innovation"] = "Construction Innovation",
  ["constr math anal"] = "Constructive Mathematical Analysis",
  ["construct manage econ"] = "Construction Management and Economics",
  ["construct res innovation"] = "Construction Research and Innovation",
  ["construct rob"] = "Construction Robotics",
  ["constructive approximation"] = "Constructive Approximation",
  ["consult pharm"] = "The Consultant pharmacist : the journal of the American Society of Consultant Pharmacists",
  ["consult psychol j"] = "Consulting psychology journal",
  ["consult specif eng"] = "Consulting-specifying engineer",
  ["consum health perspect"] = "Consumer health perspectives",
  ["consum mark abroad"] = "Consumer markets abroad",
  ["consum rep"] = "Consumer reports",
  ["consum rep health"] = "Consumer reports on health",
  ["cont lens anterior eye"] = "Contact lens & anterior eye : the journal of the British Contact Lens Association",
  ["cont philos rev"] = "Continental philosophy review",
  ["cont shelf res"] = "Continental Shelf Research",
  ["contact (thousand oaks)"] = "Contact (Thousand Oaks (Ventura County, Calif.))",
  ["contact context"] = "Contact in context",
  ["contact dermatitis"] = "Contact dermatitis",
  ["contact intraocul lens med j"] = "Contact and intraocular lens medical journal",
  ["contact point"] = "Contact point",
  ["contam control"] = "Contamination control",
  ["contam control biomed environ"] = "Contamination control-biomedical environments",
  ["contemp adm"] = "Contemporary administrator",
  ["contemp adm long term care"] = "Contemporary administrator for long-term care",
  ["contemp adv sci technol"] = "Contemporary Advances in Science and Technology",
  ["contemp anal appl math"] = "Contemporary Analysis and Applied Mathematics",
  ["contemp anesth pract"] = "Contemporary anesthesia practice",
  ["contemp behav health care"] = "Contemporary behavioral health care",
  ["contemp br hist"] = "Contemporary British history",
  ["contemp clin dent"] = "Contemporary clinical dentistry",
  ["contemp clin trials"] = "Contemporary clinical trials",
  ["contemp clin trials commun"] = "Contemporary clinical trials communications",
  ["contemp crisis"] = "Contemporary crises",
  ["contemp drug probl"] = "Contemporary drug problems",
  ["contemp econ policy"] = "Contemporary economic policy",
  ["contemp educ psychol"] = "Contemporary educational psychology",
  ["contemp educ technol"] = "Contemporary educational technology",
  ["contemp eur hist"] = "Contemporary European history",
  ["contemp evaluation res"] = "Contemporary Evaluation Research",
  ["contemp fam ther"] = "Contemporary family therapy",
  ["contemp fr civiliz"] = "Contemporary French civilization",
  ["contemp gerontol"] = "Contemporary gerontology",
  ["contemp hist con ser"] = "Contemporary History in Context Series",
  ["contemp hypn"] = "Contemporary hypnosis : the journal of the British Society of Experimental and Clinical Hypnosis",
  ["contemp hypn integr ther"] = "Contemporary hypnosis & integrative therapy",
  ["contemp intern med"] = "Contemporary internal medicine",
  ["contemp issues clin biochem"] = "Contemporary issues in clinical biochemistry",
  ["contemp issues commun sci disord"] = "Contemporary issues in communication science and disorders : CICSD",
  ["contemp issues educ res (littleton)"] = "Contemporary issues in education research (Littleton, Colo.)",
  ["contemp jew"] = "Contemporary Jewry",
  ["contemp longterm care"] = "Contemporary longterm care",
  ["contemp marx"] = "Contemporary Marxism",
  ["contemp mater"] = "Contemporary materials",
  ["contemp math"] = "Contemporary Mathematics",
  ["contemp math appl monogr expo lect notes"] = "Contemporary Mathematics and Its Applications: Monographs, Expositions and Lecture Notes",
  ["contemp mathematicians"] = "Contemporary Mathematicians",
  ["contemp neurol ser"] = "Contemporary neurology series",
  ["contemp neurosurg"] = "Contemporary neurosurgery",
  ["contemp nurse"] = "Contemporary nurse",
  ["contemp ob gyn"] = "Contemporary ob/gyn",
  ["contemp oncol"] = "Contemporary oncology",
  ["contemp oncol (pozn)"] = "Contemporary oncology (Poznań, Poland)",
  ["contemp org synth"] = "Contemporary Organic synthesis",
  ["contemp orthop"] = "Contemporary orthopaedics",
  ["contemp pac"] = "The Contemporary Pacific",
  ["contemp pediatr"] = "Contemporary pediatrics",
  ["contemp pharm pract"] = "Contemporary pharmacy practice",
  ["contemp philos"] = "Contemporary philosophy",
  ["contemp philos new s"] = "Contemporary Philosophy-new Survery",
  ["contemp phys"] = "Contemporary Physics",
  ["contemp pol issues"] = "Contemporary Policy Issues",
  ["contemp policy issues"] = "Contemporary policy issues",
  ["contemp probl ecol"] = "Contemporary problems of ecology",
  ["contemp psychiatry (hagerstown md)"] = "Contemporary psychiatry (Hagerstown, Md.)",
  ["contemp psychoanal"] = "Contemporary psychoanalysis",
  ["contemp rev"] = "Contemporary review (London, England)",
  ["contemp rev obstet gynaecol"] = "Contemporary reviews in obstetrics and gynaecology",
  ["contemp sch psychol"] = "Contemporary school psychology",
  ["contemp secur policy"] = "Contemporary security policy",
  ["contemp soc sci"] = "Contemporary social science",
  ["contemp sociol"] = "Contemporary sociology",
  ["contemp sociol-j rev"] = "Contemporary Sociology-A Journal Of Reviews",
  ["contemp south asia"] = "Contemporary South Asia",
  ["contemp southeast asia"] = "Contemporary Southeast Asia",
  ["contemp surg"] = "Contemporary surgery",
  ["contemp syst think"] = "Contemporary Systems Thinking",
  ["contemp ther"] = "Contemporary therapy",
  ["contemp top immunobiol"] = "Contemporary topics in immunobiology",
  ["contemp top lab anim sci"] = "Contemporary topics in laboratory animal science",
  ["contemp top mol immunol"] = "Contemporary topics in molecular immunology",
  ["contemp urol"] = "Contemporary urology",
  ["contemporary econ pol"] = "Contemporary Economic Policy",
  ["contests math"] = "Contests in Mathematics",
  ["contexts (berkeley calif)"] = "Contexts (Berkeley, Calif.)",
  ["contin care"] = "Continuing care",
  ["contin chang"] = "Continuity and change",
  ["contin mech thermodyn"] = "Continuum Mechanics and Thermodynamics",
  ["continuum (minneap minn)"] = "Continuum (Minneapolis, Minn.)",
  ["continuum (mount lawley)"] = "Continuum (Mount Lawley, W.A.)",
  ["continuum (n y)"] = "Continuum",
  ["continuum mech thermodyn"] = "Continuum Mechanics and Thermodynamics",
  ["continuum soc soc work leadersh health care"] = "Continuum (Society for Social Work Leadership in Health Care)",
  ["contr am hi"] = "Contributions in American History",
  ["contr med hist"] = "Contributions in medical history",
  ["contra costa dent bull"] = "Contra Costa dental bulletin",
  ["contracept deliv syst"] = "Contraceptive delivery systems",
  ["contracept fertil sex"] = "Contraception, fertilité, sexualité (1992)",
  ["contracept fertil sex (paris)"] = "Contraception, fertilité, sexualité",
  ["contracept rep"] = "Contraception Report",
  ["contracept reprod med"] = "Contraception and reproductive medicine",
  ["contracept technol"] = "Contraceptive technology",
  ["contracept technol update"] = "Contraceptive technology update",
  ["contracept x"] = "Contraception: X",
  ["contract des"] = "Contract design",
  ["contract healthc"] = "Contract healthcare",
  ["contract inter"] = "Contract interiors",
  ["contrast media mol imaging"] = "Contrast media & molecular imaging",
  ["contrat soc"] = "Le Contrat social",
  ["contrib algebra geom"] = "Contributions to Algebra and Geometry",
  ["contrib appl math stat"] = "Contributions to Applied and Mathematical Statistics",
  ["contrib asian stud"] = "Contributions to Asian studies",
  ["contrib astron obs skalnate pleso"] = "Contributions of the Astronomical Observatory Skalnate Pleso",
  ["contrib atmos phys"] = "Contributions to Atmospheric Physics",
  ["contrib biol geol"] = "Contributions in biology and geology",
  ["contrib cent res inst agric bogor"] = "Contributions from the Central Research Institute for Agriculture Bogor",
  ["contrib centro linceo interdiscip sci mat appl"] = "Contributi del Centro Linceo Interdisciplinare di Scienze Matematiche e loro Applicazioni",
  ["contrib discrete math"] = "Contributions to Discrete Mathematics",
  ["contrib econ"] = "Contributions to Economics",
  ["contrib econ analysis policy"] = "Contributions to economic analysis & policy",
  ["contrib econom"] = "Contributions to Economics",
  ["contrib econom anal"] = "Contributions to Economic Analysis",
  ["contrib embryol"] = "Contributions to embryology",
  ["contrib entomol"] = "Contributions to Entomology",
  ["contrib epidemiol biostat"] = "Contributions to epidemiology and biostatistics",
  ["contrib geophys geod"] = "Contributions to Geophysics & Geodesy",
  ["contrib gynecol obstet"] = "Contributions to gynecology and obstetrics",
  ["contrib hist econ soc"] = "Contributions à l'histoire économique et sociale",
  ["contrib hum dev"] = "Contributions to human development",
  ["contrib india sociol"] = "Contributions To Indian Sociology",
  ["contrib indian sociol"] = "Contributions to Indian sociology",
  ["contrib manag sci"] = "Contributions to Management Science",
  ["contrib mar sci"] = "Contributions in marine science",
  ["contrib math comput sci"] = "Contributions in Mathematical and Computational Sciences",
  ["contrib med stud"] = "Contributions in medical studies",
  ["contrib meteorit soc"] = "Contributions of the Meteoritical Society",
  ["contrib microbiol"] = "Contributions to microbiology",
  ["contrib microbiol immunol"] = "Contributions to microbiology and immunology",
  ["contrib mineral petrol"] = "Contributions to Mineralogy and Petrology",
  ["contrib nat hist"] = "Contributions to natural history",
  ["contrib nepalese stud"] = "Contributions to Nepalese studies",
  ["contrib nephrol"] = "Contributions to nephrology",
  ["contrib phenomenol"] = "Contributions to Phenomenology",
  ["contrib philos"] = "Contributions in Philosophy",
  ["contrib plasma phys"] = "Contributions to Plasma Physics",
  ["contrib polit econ"] = "Contributions to political economy",
  ["contrib primatol"] = "Contributions to primatology",
  ["contrib psychol"] = "Contributions in psychology",
  ["contrib sens physiol"] = "Contributions to sensory physiology",
  ["contrib soc res meteorites"] = "Contributions of the Society for Research on Meteorites",
  ["contrib stat"] = "Contributions to Statistics",
  ["contrib statist"] = "Contributions to Statistics",
  ["contrib to polit economy"] = "Contributions to Political Economy",
  ["contrib u s natl herb"] = "Contributions from the United States National Herbarium",
  ["contrib univ mich herb"] = "Contributions from the University of Michigan Herbarium",
  ["contrib zool"] = "Contributions to zoology (Amsterdam, Netherlands : 1995)",
  ["control autom"] = "Control & Automation",
  ["control clin trials"] = "Controlled clinical trials",
  ["control cybernet"] = "Control and Cybernetics",
  ["control dyn syst"] = "Control and Dynamic Systems",
  ["control eng"] = "Control Engineering",
  ["control eng pract"] = "Control engineering practice",
  ["control intell syst"] = "Control and Intelligent Systems",
  ["control nat"] = "The Control of Nature",
  ["control plagas"] = "Control de plagas",
  ["control robot sens ser"] = "Control, Robotics and Sensors Series",
  ["control theory appl"] = "Control Theory & Applications",
  ["control theory technol"] = "Control Theory and Technology",
  ["controll release newsl"] = "Controlled release newsletter",
  ["converg sci phys oncol"] = "Convergent science physical oncology",
  ["convergence (lond)"] = "Convergence (London, England)",
  ["convergent sci phys oncol"] = "Convergent Science Physical Oncology",
  ["convuls ther"] = "Convulsive therapy",
  ["coo rep"] = "COO [reports]. U.S. Atomic Energy Commission",
  ["cooley law rev"] = "Cooley law review",
  ["coop confl"] = "Cooperation and conflict",
  ["coop dent (b aires)"] = "El Cooperador dental; cooperativismo, información y ciencia odontológica",
  ["cooper rowan med j"] = "Cooper Rowan medical journal",
  ["coord chem rev"] = "Coordination Chemistry Reviews",
  ["coord noteb"] = "Coordinators' notebook : an international resource for early childhood development",
  ["copernic books"] = "Copernicus Books",
  ["cor vasa"] = "Cor et vasa",
  ["coral reefs"] = "Coral reefs (Online)",
  ["core books in adv math"] = "Core Books in Advanced Mathematics",
  ["core evid"] = "Core evidence",
  ["core j pediatr"] = "Core journals in pediatrics",
  ["cornell east asia ser"] = "Cornell East Asia Series",
  ["cornell hotel restaur adm q"] = "The Cornell hotel and restaurant administration quarterly",
  ["cornell int law j"] = "Cornell international law journal",
  ["cornell j law public policy"] = "Cornell journal of law and public policy",
  ["cornell j soc relat"] = "The Cornell journal of social relations",
  ["cornell law rev"] = "Cornell law review",
  ["cornell mag"] = "Cornell magazine (Ithaca, N.Y. : 1993)",
  ["cornell med j"] = "Cornell medical journal",
  ["cornell plant"] = "Cornell Plantations",
  ["cornell vet"] = "The Cornell veterinarian",
  ["coron artery dis"] = "Coronary artery disease",
  ["corp comment"] = "Corporate commentary",
  ["corp soc responsib environ manag"] = "Corporate social responsibility and environmental management",
  ["corps med (ettelbruck)"] = "Le Corps médical",
  ["corpus med graecorum"] = "Corpus medicorum Graecorum",
  ["corpus med graecorum suppl lipsiae"] = "Corpus medicorum Graecorum. Supplementum",
  ["corpus med graecorum suppl orient"] = "Supplementum orientale. Corpus medicorum Graecorum",
  ["corpus pragmat"] = "Corpus pragmatics : international journal of corpus linguistics and pragmatics",
  ["corr farm"] = "Il Corriere dei farmacisti",
  ["correct compend"] = "Corrections compendium",
  ["correct soc psych j behav tech methods ther"] = "Corrective and social psychiatry and journal of behavior technology methods and therapy",
  ["correct today"] = "Corrections today",
  ["correo poblac salud"] = "Correo poblacional y de la salud",
  ["corros commun"] = "Corrosion Communications",
  ["corros eng sci technol"] = "Corrosion Engineering Science and Technology",
  ["corros eng, sci technol"] = "Corrosion Engineering, Science and Technology",
  ["corros mater"] = "Corrosion & Materials",
  ["corros rev"] = "Corrosion Reviews",
  ["corros sci"] = "Corrosion Science",
  ["corrosion (houston, tx, u s)"] = "Corrosion (Houston, TX, United States)",
  ["cosmic res"] = "Cosmic Research",
  ["cospar colloq ser"] = "COSPAR Colloquia Series",
  ["cospar inf bull"] = "COSPAR Information Bulletin",
  ["cost containment"] = "Cost containment",
  ["cost eff resour alloc"] = "Cost effectiveness and resource allocation : C/E",
  ["cost qual"] = "Cost & quality : CQ",
  ["cost qual q j"] = "Cost & quality quarterly journal : CQ",
  ["cotinga (sandy)"] = "Cotinga",
  ["counc undergrad res q"] = "Council on Undergraduate Research quarterly",
  ["couns psychol"] = "The Counseling psychologist",
  ["couns psychol q"] = "Counselling psychology quarterly",
  ["couns psychother res"] = "Counselling and psychotherapy research",
  ["couns spiritual"] = "Counseling et spiritualité = Counselling and spirituality",
  ["couns values"] = "Counseling and values",
  ["counselor (deerfield beach)"] = "Counselor (Deerfield Beach, Fla.)",
  ["countdown (athens, ohio)"] = "Countdown (Athens, Ohio)",
  ["countdown istanb"] = "Countdown to Istanbul",
  ["couple family psychol"] = "Couple & family psychology",
  ["cour unesco"] = "Courier (Paris, France : 1984)",
  ["courant lect notes math"] = "Courant Lecture Notes in Mathematics",
  ["courr environ inra"] = "Courrier de l’environnement de l’INRA, Le",
  ["courr exploit scieur"] = "Courrier de l’exploitant et du scieur",
  ["courr pays est"] = "Le courrier des pays de l'Est",
  ["courr unesco"] = "Le Courrier de l'Unesco",
  ["cours (publ univ rouen havre)"] = "Cours",
  ["cours cirm"] = "Les Cours du CIRM",
  ["cours géom fac sci"] = "Cours de Géométrie de la Faculté des Sciences",
  ["cours spec"] = "Cours Specialises",
  ["cours spéc"] = "Cours Spécialisés",
  ["courts health sci law"] = "Courts, health science & the law",
  ["covertaction q"] = "Covertaction quarterly",
  ["coyunt demogr"] = "Coyuntura demográfica",
  ["coyunt econ"] = "Coyuntura económica (Bogotá, Colombia)",
  ["cpc res rev"] = "Cpc Research Review",
  ["cphc/bcs disting diss"] = "CPHC/BCS Distinguished Dissertations",
  ["cpp chem plants and process"] = "CPP Chemical Plants and Processing",
  ["cpq cancer"] = "CPQ cancer",
  ["cpq neurol psychol"] = "CPQ neurology and psychology",
  ["cpr popul res"] = "CPR population research",
  ["cpss power electron ser"] = "CPSS Power Electronics Series",
  ["cpss trans power electron appl"] = "CPSS Transactions on Power Electronics and Applications",
  ["cpt pharmacometrics syst pharmacol"] = "CPT: pharmacometrics & systems pharmacology",
  ["cpt: pharmacometrics syst pharmacol"] = "CPT: Pharmacometrics & Systems Pharmacology",
  ["cr acad sci, ser iia: sci terre planets"] = "Comptes Rendus de l’ Academie des Sciences Serie IIa:Sciences de la Terre et des Planets",
  ["cr acad sci, ser iib: mec, phys, chim, astron"] = "Comptes Rendus de l’ Academie des Sciences Serie IIb:Mecanique Physique Chimie Astronomie",
  ["cr acad sci, ser iic: chim"] = "Comptes Rendus de l’ Academie des Sciences Serie IIc:Chemie",
  ["cr acad sci, ser iii"] = "Comptes Rendus de l’ Academie des Sciences Serie III: Sciences de la Vie",
  ["cr biol"] = "Comptes Rendus Biologies",
  ["cr chim"] = "Comptes Rendus Chimie",
  ["cr geosci"] = "Comptes Rendus Geoscience",
  ["cr hebd seances acad sci"] = "Comptes Rendus Hebdomadaires des Seances de l’Academie des Sciences",
  ["cr math"] = "Comptes Rendus Mathematique",
  ["cr mec"] = "Comptes Rendus Mecanique",
  ["cr palevol"] = "Comptes Rendus Palevol",
  ["cr phys"] = "Comptes Rendus Physique",
  ["cr trav lab carlsberg ser physiol"] = "Comptes rendus des travaux du Laboratoire Carlsberg. Série physiologique",
  ["cranio clin int"] = "Cranio clinics international",
  ["craniofac growth ser"] = "Craniofacial growth series",
  ["craniomaxillofac trauma reconstr"] = "Craniomaxillofacial trauma & reconstruction",
  ["crc crit rev biochem"] = "CRC critical reviews in biochemistry",
  ["crc crit rev bioeng"] = "CRC critical reviews in bioengineering",
  ["crc crit rev clin lab sci"] = "CRC critical reviews in clinical laboratory sciences",
  ["crc crit rev clin neurobiol"] = "CRC critical reviews in clinical neurobiology",
  ["crc crit rev clin radiol nucl med"] = "CRC critical reviews in clinical radiology and nuclear medicine",
  ["crc crit rev diagn imaging"] = "CRC critical reviews in diagnostic imaging",
  ["crc crit rev food sci nutr"] = "CRC critical reviews in food science and nutrition",
  ["crc crit rev immunol"] = "CRC Critical reviews in immunology",
  ["crc crit rev microbiol"] = "CRC critical reviews in microbiology",
  ["crc crit rev plant sci"] = "Critical reviews in plant sciences",
  ["crc crit rev radiol sci"] = "CRC critical reviews in radiological sciences",
  ["crc crit rev toxicol"] = "CRC critical reviews in toxicology",
  ["crc math model ser"] = "CRC Mathematical Modelling Series",
  ["crc press ser discrete math appl"] = "CRC Press Series on Discrete Mathematics and its Applications",
  ["crc press/chapman hall handb math ser"] = "CRC Press/Chapman and Hall Handbooks in Mathematics Series",
  ["crc ser applied comput mech"] = "CRC Series in Applied and Computational Mechanics",
  ["crc ser comput mech appl anal"] = "CRC Series in Computational Mechanics and Applied Analysis",
  ["creat educ"] = "Creative education",
  ["creat math inform"] = "Creative Mathematics and Informatics",
  ["creat nurs"] = "Creative nursing",
  ["creat res j"] = "Creativity research journal",
  ["creighton law rev"] = "Creighton law review",
  ["cretac res"] = "Cretaceous research",
  ["cretaceous res"] = "Cretaceous Research",
  ["crhcs news"] = "CRHCS news : the newsletter of the Commonwealth Regional Health Community Secretariat for East, Central and Southern Africa. Commonwealth Regional Health Community for East, Central, and Southern Africa. Secretariat",
  ["crianca port"] = "A Criança portuguesa",
  ["crim behav ment health"] = "Criminal behaviour and mental health : CBMH",
  ["crim justice behav"] = "Criminal justice and behavior",
  ["crim justice ethics"] = "Criminal justice ethics",
  ["crim justice hist"] = "Criminal justice history",
  ["crim justice j"] = "Criminal justice journal",
  ["crim justice policy rev"] = "Criminal justice policy review",
  ["crim justice rev"] = "Criminal justice review",
  ["crim justice stud (abingdon)"] = "Criminal justice studies (Abingdon, England)",
  ["crim law bull"] = "Criminal law bulletin",
  ["crim law forum"] = "Criminal law forum",
  ["crim law j"] = "Criminal law journal",
  ["crim law philos"] = "Criminal law and philosophy",
  ["crim law rev"] = "Criminal law review (London, England)",
  ["crim rep can new ser"] = "Criminal reports [Canada] new series, annotated; a series of reports with annotations and practice notes on criminal cases arising in the courts of the various provinces in Canada",
  ["crime delinq"] = "Crime and delinquency",
  ["crime hist soc"] = "Crime, histoire & sociétés = Crime, history & societies",
  ["crime justice"] = "Crime and justice (Chicago, Ill.)",
  ["crime law soc change"] = "Crime, law, and social change",
  ["crime media cult"] = "Crime, media, culture",
  ["crime sci"] = "Crime science",
  ["crime soc justice"] = "Crime and social justice",
  ["criminol crim justice"] = "Criminology & criminal justice : the international journal of policy and practice",
  ["criminol public policy"] = "Criminology & public policy",
  ["criminologie (montr)"] = "Criminologie",
  ["crippl child"] = "The Crippled child",
  ["crispr j"] = "The CRISPR journal",
  ["crit afr stud"] = "Critical African studies",
  ["crit anthropol"] = "Critique of anthropology",
  ["crit arts"] = "Critical arts",
  ["crit asian stud"] = "Critical Asian studies",
  ["crit care"] = "Critical care (London, England)",
  ["crit care (houten)"] = "Critical care (Houten, Netherlands)",
  ["crit care clin"] = "Critical care clinics",
  ["crit care explor"] = "Critical care explorations",
  ["crit care med"] = "Critical care medicine",
  ["crit care nurs clin north am"] = "Critical care nursing clinics of North America",
  ["crit care nurs q"] = "Critical care nursing quarterly",
  ["crit care nurse"] = "Critical care nurse",
  ["crit care res pract"] = "Critical care research and practice",
  ["crit care resusc"] = "Critical care and resuscitation : journal of the Australasian Academy of Critical Care Medicine",
  ["crit care shock"] = "Critical care and shock",
  ["crit care update"] = "Critical care update",
  ["crit connect"] = "Critical connections : the complete news source for critical care professionals",
  ["crit criminol"] = "Critical criminology",
  ["crit inq"] = "Critical inquiry",
  ["crit insights aquacult"] = "Critical Insights in Aquaculture",
  ["crit insights clim change"] = "Critical Insights in Climate Change",
  ["crit insights environ sci technol"] = "Critical Insights in Environmental Science and Technology",
  ["crit insights geochem geophys"] = "Critical Insights in Geochemistry and Geophysics",
  ["crit mil stud"] = "Critical military studies",
  ["crit path aids proj"] = "Critical Path AIDS project",
  ["crit pathw cardiol"] = "Critical pathways in cardiology",
  ["crit policy stud"] = "Critical policy studies",
  ["crit public health"] = "Critical public health",
  ["crit q"] = "The Critical quarterly",
  ["crit radic soc work"] = "Critical and radical social work",
  ["crit rev (n y)"] = "Critical review (New York, N.Y.)",
  ["crit rev anal chem"] = "Critical reviews in analytical chemistry",
  ["crit rev biochem mol biol"] = "Critical reviews in biochemistry and molecular biology",
  ["crit rev bioeng"] = "Critical reviews in bioengineering",
  ["crit rev biomed eng"] = "Critical reviews in biomedical engineering",
  ["crit rev biotechnol"] = "Critical reviews in biotechnology",
  ["crit rev bone miner metab"] = "Critical Reviews in Bone and Mineral Metabolism",
  ["crit rev clin lab sci"] = "Critical reviews in clinical laboratory sciences",
  ["crit rev comput tomogr"] = "Critical reviews in computed tomography",
  ["crit rev diagn imaging"] = "Critical reviews in diagnostic imaging",
  ["crit rev env sci technol"] = "Critical Reviews in Environment Science and Technology",
  ["crit rev environ sci technol"] = "Critical reviews in environmental science and technology",
  ["crit rev eukaryot gene expr"] = "Critical reviews in eukaryotic gene expression",
  ["crit rev eukaryotic gene express"] = "Critical Reviews in Eukaryotic Gene Expression",
  ["crit rev food sci nutr"] = "Critical reviews in food science and nutrition",
  ["crit rev immunol"] = "Critical reviews in immunology",
  ["crit rev int soc political philos"] = "Critical review of international social and political philosophy",
  ["crit rev med inform"] = "Critical reviews in medical informatics",
  ["crit rev microbiol"] = "Critical reviews in microbiology",
  ["crit rev neurobiol"] = "Critical reviews in neurobiology",
  ["crit rev neurosurg"] = "Critical reviews in neurosurgery : CR",
  ["crit rev oncog"] = "Critical reviews in oncogenesis",
  ["crit rev oncol hematol"] = "Critical reviews in oncology/hematology",
  ["crit rev oral biol med"] = "Critical reviews in oral biology and medicine : an official publication of the American Association of Oral Biologists",
  ["crit rev phys rehabil med"] = "Critical reviews in physical and rehabilitation medicine",
  ["crit rev plant sci"] = "Critical Reviews in Plant Sciences",
  ["crit rev solid state mater sci"] = "Critical Reviews in Solid State and Materials Sciences",
  ["crit rev ther drug carrier syst"] = "Critical reviews in therapeutic drug carrier systems",
  ["crit rev toxicol"] = "Critical reviews in toxicology",
  ["crit soc policy"] = "Critical social policy",
  ["crit soc res"] = "Critical social research",
  ["crit sociol"] = "La Critica sociologica",
  ["crit sociol (eugene)"] = "Critical sociology",
  ["crit stor"] = "Critica storica",
  ["crit stud mass commun"] = "Critical studies in mass communication : CSMC : a publication of the Speech Communication Association",
  ["crit stud media commun"] = "Critical studies in media communication",
  ["crit stud secur"] = "Critical studies on security",
  ["crit stud terror"] = "Critical studies on terrorism",
  ["crit ultrasound j"] = "Critical ultrasound journal",
  ["critic rev"] = "Critical Review",
  ["crm advocate"] = "CRM advocate",
  ["crm monogr ser"] = "CRM Monograph Series",
  ["crm proc lecture notes"] = "CRM Proceedings & Lecture Notes",
  ["crm ser math phys"] = "CRM Series in Mathematical Physics",
  ["crm series"] = "Centro di Ricerca Matematica Ennio De Giorgi (CRM) Series",
  ["croat chem acta"] = "Croatica Chemica Acta",
  ["croat med j"] = "Croatian medical journal",
  ["croat oper res rev crorr"] = "Croatian Operational Research Review. CRORR",
  ["crohns colitis 360"] = "Crohn's & colitis 360",
  ["cron idi"] = "Cronache dell'IDI",
  ["cron med"] = "La Crónica médica",
  ["cron pompeiane"] = "Cronache pompeiane",
  ["croom helm ser geogr environment"] = "Croom Helm Series in Geography and Environment",
  ["crop breed appl biotechnol"] = "Crop Breeding and Applied Biotechnology",
  ["crop breed genet genom"] = "Crop breeding, genetics and genomics",
  ["crop breed j"] = "Crop breeding journal",
  ["crop des"] = "Crop Design",
  ["crop environ"] = "Crop and Environment",
  ["crop forage turfgrass manage"] = "Crop, Forage & Turfgrass Management",
  ["crop j"] = "The Crop journal",
  ["crop pasture sci"] = "Crop & pasture science",
  ["crop prot"] = "Crop Protection",
  ["crop sci"] = "Crop science",
  ["crops soils"] = "Crops & Soils",
  ["cross cont"] = "Cultural Approaches to Parenting",
  ["cross cult res"] = "Cross-cultural research : official journal of the Society for Cross-Cultural Research",
  ["cross curr"] = "Cross currents",
  ["cross curr (ann arbor mi)"] = "Cross currents (Ann Arbor, Mich.)",
  ["crossref hum resour manage"] = "Cross-reference on human resources management",
  ["crossroads (de kalb)"] = "Crossroads (De Kalb, Ill.)",
  ["crown agents rev"] = "The Crown agents review",
  ["crustac issues"] = "Crustacean issues",
  ["cruz roja cuba"] = "Cruz Roja Cubana",
  ["cryo letters"] = "Cryo letters",
  ["cryogenic materials ser"] = "Cryogenic Materials Series",
  ["cryogenics (guildf)"] = "Cryogenics",
  ["cryosphere discuss"] = "Cryosphere Discussions",
  ["cryptogam algol"] = "Cryptogamie. Algologie",
  ["cryptogam bot"] = "Cryptogamic Botany",
  ["cryptogam bryol"] = "Cryptogamie. Bryologie",
  ["cryptogam bryol lichenol"] = "Cryptogamie. Bryologie, lichénologie",
  ["cryptogam helv"] = "Cryptogamica Helvetica",
  ["cryptogam mycol"] = "Cryptogamie. Mycologie",
  ["cryptogr biodivers assess"] = "Cryptogam biodiversity and assessment",
  ["cryptogr commun"] = "Cryptography and Communications",
  ["cryst eng"] = "Crystal Engineering",
  ["cryst growth des"] = "Crystal Growth and Design",
  ["cryst res technol"] = "Crystal Research and Technology",
  ["crystallogr rep"] = "Crystallography Reports",
  ["crystallogr rev"] = "Crystallography Reviews",
  ["crystals (basel)"] = "Crystals",
  ["csa j (denver)"] = "CSA journal (Denver, Colo.)",
  ["cscw conf comput support coop work"] = "CSCW : proceedings of the Conference on Computer-Supported Cooperative Work. Conference on Computer-Supported Cooperative Work",
  ["csd commun"] = "CSD Communications",
  ["csee j power energy syst"] = "CSEE Journal of Power and Energy Systems",
  ["csh protoc"] = "CSH protocols",
  ["csiam trans appl math"] = "CSIAM Transactions on Applied Mathematics",
  ["csli lecture notes"] = "CSLI Lecture Notes",
  ["ct li lun yu ying yong yan jiu"] = "CT Li lun yu ying yong yan jiu",
  ["cte j"] = "The CTE journal",
  ["ctm class top math"] = "CTM. Classical Topics in Mathematics",
  ["ctry demogr profiles"] = "Country demographic profiles",
  ["ctry life"] = "Country life (London, England)",
  ["ctry profiles"] = "Country profiles",
  ["ctwatch q"] = "CTWatch quarterly : cyberinfrastructure technology watch",
  ["cuad am"] = "Cuadernos americanos",
  ["cuad aragon hist med cienc a"] = "Cuadernos aragoneses de historia de la medicina y de la ciencia. Serie A, Monografías",
  ["cuad arquit urban"] = "Cuadernos de arquitectura y urbanismo",
  ["cuad bioet"] = "Cuadernos de bioética : revista oficial de la Asociación Española de Bioética y Ética Médica",
  ["cuad cendes"] = "Cuadernos del CENDES",
  ["cuad claeh"] = "Cuadernos del CLAEH",
  ["cuad complut hist med cienc"] = "Cuadernos Complutenses de historia de la medicina y de la ciencia",
  ["cuad comun amidep"] = "Cuadernos de comunicación AMIDEP",
  ["cuad econ"] = "Cuadernos de economía (Santiago, Chile)",
  ["cuad econ soc"] = "Cuadernos de economía social",
  ["cuad estud gallegos"] = "Cuadernos de estudios gallegos",
  ["cuad farmacol"] = "Cuadernos de farmacología",
  ["cuad filol cl"] = "Cuadernos de filología clásica",
  ["cuad hisp hist med cienc"] = "Cuadernos hispánicos de historia de la medicina y de la ciencia",
  ["cuad hispanoam"] = "Cuadernos hispanoamericanos",
  ["cuad hist"] = "Cuadernos de historia (Santiago, Chile)",
  ["cuad hist contemp"] = "Cuadernos de historia contemporánea",
  ["cuad hist derecho"] = "Cuadernos de historia del derecho (Madrid, Spain)",
  ["cuad hist esp"] = "Cuadernos de historia de España",
  ["cuad hist med esp"] = "Cuadernos de historia de la medicina española",
  ["cuad hist med esp monogr"] = "Cuadernos de historia de la medicina española. Monografías",
  ["cuad hist mod"] = "Cuadernos de historia moderna",
  ["cuad hist salud publica"] = "Cuadernos de historia de la salud pública",
  ["cuad invest hist"] = "Cuadernos de investigación histórica",
  ["cuad investig educ"] = "Cuaderno de investigación en la educación : serie especial de temas especiales. University of Puerto Rico. Recinto de Río Piedras. Centro de Investigaciones Educativas",
  ["cuad log epistemol leng"] = "Cuadernos de Logica, Epistemologia y Lenguaje",
  ["cuad med divulg cient"] = "Cuadernos médicos y de divulgación científica",
  ["cuad med soc"] = "Cuadernos médicos-sociales",
  ["cuad nutr"] = "Cuadernos de nutrición",
  ["cuad uned"] = "Cuadernos de la UNED",
  ["cuad valencia hist med cienc"] = "Cuadernos valencianos de historia de la medicina y de la ciencia",
  ["cuadernos de la cvf"] = "Cuadernos de la C.V.F.",
  ["cuadernos econ"] = "Cuadernos de Economia",
  ["cuadernos econ ice"] = "Cuadernos Economicos de I.C.E.",
  ["cuban stud"] = "Cuban studies",
  ["cuihua xuebao"] = "Cuihua Xuebao",
  ["cult agric food environ"] = "Culture, agriculture, food and environment",
  ["cult anthropol"] = "Cultural anthropology : journal of the Society for Cultural Anthropology",
  ["cult brain"] = "Culture and brain",
  ["cult civiliz middle east"] = "Culture and Civilization in the Middle East",
  ["cult confl"] = "Cultures et conflits = Culturas y conflictos = Cultures and conflicts",
  ["cult dev"] = "Cultures et développement",
  ["cult divers ment health"] = "Cultural diversity and mental health",
  ["cult geogr"] = "Cultural geographies",
  ["cult health sex"] = "Culture, health & sexuality",
  ["cult med"] = "Cultura médica",
  ["cult med (rome)"] = "Cultura medica",
  ["cult med psychiatry"] = "Culture, medicine and psychiatry",
  ["cult psychol"] = "Culture & psychology",
  ["cult relig"] = "Culture and religion",
  ["cult sc"] = "Cultura e scuola",
  ["cult sci"] = "La Cultura Scientifica",
  ["cult soc hist"] = "Cultural and social history : the journal of the Social History Society",
  ["cult sociol"] = "Cultural sociology",
  ["cult stud crit methodol"] = "Cultural studies <=> critical methodologies",
  ["cult stud sci educ"] = "Cultural studies of science education",
  ["cult surv q"] = "Cultural survival quarterly",
  ["cult tech"] = "Culture technique",
  ["culta bononia"] = "Culta Bononia",
  ["cultur divers ethnic minor psychol"] = "Cultural diversity & ethnic minority psychology",
  ["cultura (iași)"] = "Cultura (Iași, Romania)",
  ["cumberl cty hist"] = "Cumberland County history. Cumberland County Historical Society and Hamilton Library Association",
  ["cumberland law rev"] = "Cumberland law review",
  ["cumberland samford law rev"] = "Cumberland-Samford law review",
  ["cuore circ"] = "Cuore e circolazione",
  ["cur popul rep farm popul"] = "Current population reports. Series P-27, Farm population",
  ["curator (n y)"] = "Curator : a quarterly publication of the American Museum of Natural History",
  ["curity res notes"] = "Curity research notes",
  ["curq web"] = "CURQ on the Web",
  ["curr addict rep"] = "Current addiction reports",
  ["curr adv ophthalmol"] = "Current advances in ophthalmology",
  ["curr aff bull"] = "Current affairs bulletin",
  ["curr aging sci"] = "Current aging science",
  ["curr alcohol"] = "Currents in alcoholism",
  ["curr allergy asthma rep"] = "Current allergy and asthma reports",
  ["curr allergy rep"] = "Current allergy reports",
  ["curr altern energy"] = "Current Alternative Energy",
  ["curr alzheimer res"] = "Current Alzheimer research",
  ["curr anaesth crit care"] = "Current anaesthesia and critical care",
  ["curr anal chem"] = "Current Analytical Chemistry",
  ["curr anesthesiol rep"] = "Current anesthesiology reports",
  ["curr angiogenes"] = "Current angiogenesis",
  ["curr anthropol"] = "Current anthropology",
  ["curr appl mater"] = "Current Applied Materials",
  ["curr appl phys"] = "Current Applied Physics",
  ["curr appl polym sci"] = "Current Applied Polymer Science",
  ["curr appl sci technol"] = "Current applied science and technology",
  ["curr artif intell"] = "Current Artificial Intelligence",
  ["curr atheroscler rep"] = "Current atherosclerosis reports",
  ["curr behav neurosci rep"] = "Current behavioral neuroscience reports",
  ["curr bibliogr afr aff"] = "A Current bibliography on African affairs",
  ["curr bioact compd"] = "Current Bioactive Compounds",
  ["curr bioinf"] = "Current Bioinformatics",
  ["curr bioinform"] = "Current bioinformatics",
  ["curr biol"] = "Current biology : CB",
  ["curr biomark find"] = "Current biomarker findings",
  ["curr biotechnol"] = "Current biotechnology",
  ["curr bladder dysfunct rep"] = "Current bladder dysfunction reports",
  ["curr breast cancer rep"] = "Current breast cancer reports",
  ["curr cancer drug targets"] = "Current cancer drug targets",
  ["curr cancer rep"] = "Current cancer reports",
  ["curr cancer ther rev"] = "Current cancer therapy reviews",
  ["curr cardiol rep"] = "Current cardiology reports",
  ["curr cardiol rev"] = "Current cardiology reviews",
  ["curr cardiovasc imaging rep"] = "Current cardiovascular imaging reports",
  ["curr cardiovasc risk rep"] = "Current cardiovascular risk reports",
  ["curr catal"] = "Current Catalysis",
  ["curr cell biochem"] = "Current cellular biochemistry",
  ["curr chem biol"] = "Current chemical biology",
  ["curr chem genom transl med"] = "Current chemical genomics and translational medicine",
  ["curr chem genomics"] = "Current chemical genomics",
  ["curr chromatogr"] = "Current chromatography",
  ["curr clim change rep"] = "Current climate change reports",
  ["curr clin microbiol rep"] = "Current clinical microbiology reports",
  ["curr clin pharmacol"] = "Current clinical pharmacology",
  ["curr clin top infect dis"] = "Current clinical topics in infectious diseases",
  ["curr colorectal cancer rep"] = "Current colorectal cancer reports",
  ["curr commun cell mol biol"] = "Current communications in cell & molecular biology",
  ["curr comput aided drug des"] = "Current computer-aided drug design",
  ["curr comput-aided drug des"] = "Current Computer-Aided Drug Design",
  ["curr conc cerebrovasc dis stroke"] = "Current concepts of cerebrovascular disease: Stroke",
  ["curr concepts hosp pharm manage"] = "Current concepts in hospital pharmacy management",
  ["curr concepts nutr"] = "Current concepts in nutrition",
  ["curr contents clin med"] = "Current contents. Clinical medicine",
  ["curr control trials cardiovasc med"] = "Current controlled trials in cardiovascular medicine",
  ["curr dermatol rep"] = "Current dermatology reports",
  ["curr dev disord rep"] = "Current developmental disorders reports",
  ["curr dev math"] = "Current Developments in Mathematics",
  ["curr dev math sci"] = "Current Developments in Mathematical Sciences",
  ["curr dev nutr"] = "Current developments in nutrition",
  ["curr dev psychopharmacol"] = "Current developments in psychopharmacology",
  ["curr diab rep"] = "Current diabetes reports",
  ["curr diabetes rev"] = "Current diabetes reviews",
  ["curr diagn pathol"] = "Current diagnostic pathology",
  ["curr dig post sov press"] = "The Current digest of the post-Soviet press",
  ["curr dig sov press"] = "The Current digest of the Soviet press",
  ["curr dir autoimmun"] = "Current directions in autoimmunity",
  ["curr dir psychol sci"] = "Current directions in psychological science",
  ["curr drug abuse rev"] = "Current Drug Abuse Reviews",
  ["curr drug deliv"] = "Current drug delivery",
  ["curr drug delivery"] = "Current Drug Delivery",
  ["curr drug discov technol"] = "Current drug discovery technologies",
  ["curr drug metab"] = "Current drug metabolism",
  ["curr drug res rev"] = "Current drug research reviews",
  ["curr drug saf"] = "Current drug safety",
  ["curr drug targets"] = "Current drug targets",
  ["curr drug targets cardiovasc haematol disord"] = "Current drug targets. Cardiovascular & haematological disorders",
  ["curr drug targets cns neurol disord"] = "Current drug targets. CNS and neurological disorders",
  ["curr drug targets immune endocr metabol disord"] = "Current drug targets. Immune, endocrine and metabolic disorders",
  ["curr drug targets infect disord"] = "Current drug targets. Infectious disorders",
  ["curr drug targets inflamm allergy"] = "Current drug targets. Inflammation and allergy",
  ["curr drug targets: cardiovasc & haematol disord"] = "Current Drug Targets: Cardiovascular & Haematological Disorders",
  ["curr drug targets: cns neurol disord"] = "Current Drug Targets: CNS & Neurological Disorders",
  ["curr drug targets: immune, endocr metab disord"] = "Current Drug Targets: Immune, Endocrine and Metabolic Disorders",
  ["curr drug targets: infect disord"] = "Current Drug Targets: Infectious Disorders",
  ["curr drug targets: inflammation allergy"] = "Current Drug Targets: Inflammation & Allergy",
  ["curr drug ther"] = "Current drug therapy",
  ["curr emerg hosp med rep"] = "Current emergency and hospital medicine reports",
  ["curr eng lett"] = "Current Engineering Letters",
  ["curr eng lett rev"] = "Current Engineering Letters and Reviews",
  ["curr environ health rep"] = "Current environmental health reports",
  ["curr enzym inhib"] = "Current enzyme inhibition",
  ["curr enzyme inhib"] = "Current Enzyme Inhibition",
  ["curr epidemiol rep"] = "Current epidemiology reports",
  ["curr eye res"] = "Current eye research",
  ["curr find infect dis"] = "Current findings of infectious diseases",
  ["curr food sci technol rep"] = "Current Food Science and Technology Reports",
  ["curr for rep"] = "Current Forestry Reports",
  ["curr forensic sci"] = "Current Forensic Science",
  ["curr funct foods"] = "Current Functional Foods",
  ["curr fungal infect rep"] = "Current fungal infection reports",
  ["curr gastroenterol rep"] = "Current gastroenterology reports",
  ["curr gene ther"] = "Current gene therapy",
  ["curr genet"] = "Current genetics",
  ["curr genet med rep"] = "Current genetic medicine reports",
  ["curr genomics"] = "Current genomics",
  ["curr geriatr rep"] = "Current geriatrics reports",
  ["curr gerontol geriatr res"] = "Current gerontology and geriatrics research",
  ["curr green chem"] = "Current Green Chemistry",
  ["curr health sci j"] = "Current health sciences journal",
  ["curr heart fail rep"] = "Current heart failure reports",
  ["curr hematol malig rep"] = "Current hematologic malignancy reports",
  ["curr hematol rep"] = "Current hematology reports",
  ["curr hepat rep"] = "Current hepatitis reports",
  ["curr hepatol rep"] = "Current hepatology reports",
  ["curr herpetol"] = "Current herpetology",
  ["curr hist"] = "Current history (New York, N.Y. : 1941)",
  ["curr hiv res"] = "Current HIV research",
  ["curr hiv/aids rep"] = "Current HIV/AIDS reports",
  ["curr hypertens rep"] = "Current hypertension reports",
  ["curr hypertens rev"] = "Current hypertension reviews",
  ["curr immunol rev"] = "Current immunology reviews",
  ["curr infect dis rep"] = "Current infectious disease reports",
  ["curr inorg chem"] = "Current Inorganic Chemistry",
  ["curr interv cardiol rep"] = "Current interventional cardiology reports",
  ["curr issues educ (tempe)"] = "Current issues in education (Tempe, Ariz.)",
  ["curr issues emerg elearn"] = "Current issues in emerging elearning",
  ["curr issues intest microbiol"] = "Current issues in intestinal microbiology",
  ["curr issues mol biol"] = "Current issues in molecular biology",
  ["curr issues public health"] = "Current issues in public health",
  ["curr j neurol"] = "Current journal of neurology",
  ["curr landscape ecol rep"] = "Current Landscape Ecology Reports",
  ["curr law statut annot gb"] = "Current law statutes annotated. Great Britain",
  ["curr leg probl"] = "Current legal problems",
  ["curr leg thought lawyers dig law rev"] = "Current legal thought; the lawyers' digest of law reviews",
  ["curr mater sci"] = "Current Materials Science",
  ["curr med atty"] = "Current medicine for attorneys",
  ["curr med chem"] = "Current medicinal chemistry",
  ["curr med chem anti inflamm anti allergy agents"] = "Current medicinal chemistry. Anti-inflammatory & anti-allergy agents",
  ["curr med chem anticancer agents"] = "Current medicinal chemistry. Anti-cancer agents",
  ["curr med chem cardiovasc hematol agents"] = "Current medicinal chemistry. Cardiovascular and hematological agents",
  ["curr med chem immunol endocr metab agents"] = "Current medicinal chemistry. Immunology, endocrine & metabolic agents",
  ["curr med chem: anti-cancer agents"] = "Current Medicinal Chemistry: Anti-Cancer Agents",
  ["curr med chem: anti-infect agents"] = "Current Medicinal Chemistry: Anti-Infective Agents",
  ["curr med chem: anti-inflammatory anti-allergy agents"] = "Current Medicinal Chemistry: Anti-Inflammatory & Anti-Allergy Agents",
  ["curr med chem: cardiovasc hematol agents"] = "Current Medicinal Chemistry: Cardiovascular & Hematological Agents",
  ["curr med chem: cent nerv syst agents"] = "Current Medicinal Chemistry: Central Nervous System Agents",
  ["curr med chem: immunol, endocr metab agents"] = "Current Medicinal Chemistry: Immunology, Endocrine & Metabolic Agents",
  ["curr med dig"] = "Current medical digest",
  ["curr med drugs"] = "Current medicine and drugs",
  ["curr med imaging"] = "Current medical imaging",
  ["curr med imaging rev"] = "Current medical imaging reviews",
  ["curr med lit cardiol"] = "Current medical literature. Cardiology",
  ["curr med lit dermatol"] = "Current medical literature. Dermatology",
  ["curr med lit rheumatol"] = "Current medical literature. Rheumatology",
  ["curr med mycol"] = "Current medical mycology",
  ["curr med pract"] = "Current medical practice",
  ["curr med res opin"] = "Current medical research and opinion",
  ["curr med res pract"] = "Current medicine research and practice",
  ["curr med sci"] = "Current medical science",
  ["curr metabolomics"] = "Current Metabolomics",
  ["curr microbiol"] = "Current microbiology",
  ["curr microsc rep"] = "Current Microscopy Reports",
  ["curr microwave chem"] = "Current Microwave Chemistry",
  ["curr mod biol"] = "Currents in modern biology",
  ["curr mol biol rep"] = "Current molecular biology reports",
  ["curr mol imaging"] = "Current molecular imaging",
  ["curr mol med"] = "Current molecular medicine",
  ["curr mol pharmacol"] = "Current molecular pharmacology",
  ["curr nanomater"] = "Current Nanomaterials",
  ["curr nanomed"] = "Current Nanomedicine",
  ["curr nanosci"] = "Current nanoscience",
  ["curr nanotoxic prev"] = "Current Nanotoxicity and Prevention",
  ["curr nat sci"] = "Current Natural Sciences",
  ["curr neurobiol"] = "Current neurobiology",
  ["curr neurol neurosci rep"] = "Current neurology and neuroscience reports",
  ["curr neuropharmacol"] = "Current neuropharmacology",
  ["curr neurovasc res"] = "Current neurovascular research",
  ["curr nutr food sci"] = "Current nutrition and food science",
  ["curr nutr rep"] = "Current nutrition reports",
  ["curr obes rep"] = "Current obesity reports",
  ["curr obstet gynaecol"] = "Current obstetrics & gynaecology",
  ["curr obstet gynecol rep"] = "Current obstetrics and gynecology reports",
  ["curr oncol"] = "Current oncology (Toronto, Ont.)",
  ["curr oncol rep"] = "Current oncology reports",
  ["curr ophthalmol rep"] = "Current ophthalmology reports",
  ["curr opin allergy clin immunol"] = "Current opinion in allergy and clinical immunology",
  ["curr opin anaesthesiol"] = "Current opinion in anaesthesiology",
  ["curr opin behav sci"] = "Current opinion in behavioral sciences",
  ["curr opin biomed eng"] = "Current opinion in biomedical engineering",
  ["curr opin biotechnol"] = "Current opinion in biotechnology",
  ["curr opin cardiol"] = "Current opinion in cardiology",
  ["curr opin cell biol"] = "Current opinion in cell biology",
  ["curr opin chem biol"] = "Current opinion in chemical biology",
  ["curr opin chem eng"] = "Current opinion in chemical engineering",
  ["curr opin clin nutr metab care"] = "Current opinion in clinical nutrition and metabolic care",
  ["curr opin colloid interface sci"] = "Current Opinion in Colloid and Interface Science",
  ["curr opin cosmet dent"] = "Current opinion in cosmetic dentistry",
  ["curr opin crit care"] = "Current opinion in critical care",
  ["curr opin dent"] = "Current opinion in dentistry",
  ["curr opin drug discov devel"] = "Current opinion in drug discovery & development",
  ["curr opin drug discovery dev"] = "Current Opinion in Drug Discovery & Development",
  ["curr opin electrochem"] = "Current opinion in electrochemistry",
  ["curr opin endocr metab res"] = "Current opinion in endocrine and metabolic research",
  ["curr opin endocrinol diabetes"] = "Current opinion in endocrinology & diabetes",
  ["curr opin endocrinol diabetes obes"] = "Current opinion in endocrinology, diabetes, and obesity",
  ["curr opin environ sci health"] = "Current opinion in environmental science & health",
  ["curr opin environ sustain"] = "Current opinion in environmental sustainability",
  ["curr opin environ sustainability"] = "Current Opinion in Environment Sustainability",
  ["curr opin food sci"] = "Current Opinion in Food Science",
  ["curr opin gastroenterol"] = "Current opinion in gastroenterology",
  ["curr opin gen surg"] = "Current opinion in general surgery",
  ["curr opin genet dev"] = "Current opinion in genetics & development",
  ["curr opin green sustain chem"] = "Current opinion in green and sustainable chemistry",
  ["curr opin green sustainable chem"] = "Current Opinion in Green and Sustainable Chemistry",
  ["curr opin gynecol obstet"] = "Current opinion in gynecology and obstetrics",
  ["curr opin hematol"] = "Current opinion in hematology",
  ["curr opin hiv aids"] = "Current opinion in HIV and AIDS",
  ["curr opin immunol"] = "Current opinion in immunology",
  ["curr opin infect dis"] = "Current opinion in infectious diseases",
  ["curr opin insect sci"] = "Current opinion in insect science",
  ["curr opin invest drugs"] = "Current Opinion in Investigational Drugs",
  ["curr opin investig drugs"] = "Current opinion in investigational drugs (London, England : 2000)",
  ["curr opin lipidol"] = "Current opinion in lipidology",
  ["curr opin microbiol"] = "Current opinion in microbiology",
  ["curr opin mol ther"] = "Current Opinion in Molecular Therapeutics",
  ["curr opin nephrol hypertens"] = "Current opinion in nephrology and hypertension",
  ["curr opin neurobiol"] = "Current opinion in neurobiology",
  ["curr opin neurol"] = "Current opinion in neurology",
  ["curr opin neurol neurosurg"] = "Current opinion in neurology and neurosurgery",
  ["curr opin obstet gynecol"] = "Current opinion in obstetrics & gynecology",
  ["curr opin oncol"] = "Current opinion in oncology",
  ["curr opin ophthalmol"] = "Current opinion in ophthalmology",
  ["curr opin organ transplant"] = "Current opinion in organ transplantation",
  ["curr opin orthop"] = "Current opinion in orthopaedics",
  ["curr opin otolaryngol head neck surg"] = "Current opinion in otolaryngology & head and neck surgery",
  ["curr opin pediatr"] = "Current opinion in pediatrics",
  ["curr opin periodontol"] = "Current opinion in periodontology",
  ["curr opin pharmacol"] = "Current opinion in pharmacology",
  ["curr opin physiol"] = "Current opinion in physiology",
  ["curr opin plant biol"] = "Current opinion in plant biology",
  ["curr opin psychiatry"] = "Current opinion in psychiatry",
  ["curr opin psychol"] = "Current opinion in psychology",
  ["curr opin pulm med"] = "Current opinion in pulmonary medicine",
  ["curr opin radiol"] = "Current opinion in radiology",
  ["curr opin rheumatol"] = "Current opinion in rheumatology",
  ["curr opin solid state mater sci"] = "Current Opinion in Solid State and Materials Science",
  ["curr opin struct biol"] = "Current opinion in structural biology",
  ["curr opin support palliat care"] = "Current opinion in supportive and palliative care",
  ["curr opin syst biol"] = "Current opinion in systems biology",
  ["curr opin toxicol"] = "Current opinion in toxicology",
  ["curr opin urol"] = "Current opinion in urology",
  ["curr opin virol"] = "Current opinion in virology",
  ["curr opt photonics"] = "Current Optics and Photonics",
  ["curr oral health rep"] = "Current oral health reports",
  ["curr org chem"] = "Current Organic Chemistry",
  ["curr org synth"] = "Current organic synthesis",
  ["curr organocatal"] = "Current Organocatalysis",
  ["curr orthop"] = "Current orthopaedics",
  ["curr orthop pract"] = "Current orthopaedic practice",
  ["curr osteoporos rep"] = "Current osteoporosis reports",
  ["curr otorhinolaryngol rep"] = "Current otorhinolaryngology reports",
  ["curr pain headache rep"] = "Current pain and headache reports",
  ["curr pathobiol rep"] = "Current pathobiology reports",
  ["curr pediatr rep"] = "Current pediatrics reports",
  ["curr pediatr rev"] = "Current pediatric reviews",
  ["curr pharm anal"] = "Current pharmaceutical analysis",
  ["curr pharm biotechnol"] = "Current pharmaceutical biotechnology",
  ["curr pharm des"] = "Current pharmaceutical design",
  ["curr pharm teach learn"] = "Currents in pharmacy teaching & learning",
  ["curr pharmacogenomics"] = "Current Pharmacogenomics",
  ["curr pharmacogenomics pers med"] = "Current Pharmacogenomics and Personalized Medicine",
  ["curr pharmacogenomics person med"] = "Current pharmacogenomics and personalized medicine",
  ["curr pharmacol rep"] = "Current pharmacology reports",
  ["curr phys"] = "Current Physics",
  ["curr phys chem"] = "Current physical chemistry",
  ["curr phys med rehabil rep"] = "Current physical medicine and rehabilitation reports",
  ["curr plant biol"] = "Current Plant Biology",
  ["curr plant sci biotechnol agric"] = "Current Plant Science in Biotechnology and Agriculture",
  ["curr pollut rep"] = "Current pollution reports",
  ["curr popul rep 23 spec stud"] = "Current population reports. Series P-23. Special studies",
  ["curr popul rep [23] tech stud"] = "Current population reports. Series P-23, Technical studies",
  ["curr popul rep [p-26]"] = "Current population reports. Series P-26, Federal-State Cooperative Program for Population Estimates",
  ["curr popul rep [spec censuses]"] = "Current Population Reports. Special Censuses. Series P-28",
  ["curr popul rep consum income"] = "Current population reports. Series P-60, Consumer income",
  ["curr popul rep p-26"] = "Current population reports. Series P-26, Local population estimates",
  ["curr popul rep popul charact"] = "Current population reports. Series P-20, Population characteristics",
  ["curr popul rep popul estim proj"] = "Current population reports. Series P-25, Population estimates and projections",
  ["curr popul rep serp-70 househ econ stud"] = "Current population reports. Series P-70, Household economic studies",
  ["curr popul rep spec censuses"] = "Current population reports. Series P-28, Special censuses",
  ["curr pract gerontol nurs"] = "Current practice in gerontological nursing",
  ["curr pract obstet gynecol nurs"] = "Current practice in obstetric and gynecologic nursing",
  ["curr pract orthop surg"] = "Current practice in orthopaedic surgery",
  ["curr pract pediatr nurs"] = "Current practice in pediatric nursing",
  ["curr prescr"] = "Current prescribing",
  ["curr probiotics"] = "Current Probiotics",
  ["curr probl cancer"] = "Current problems in cancer",
  ["curr probl cardiol"] = "Current problems in cardiology",
  ["curr probl clin biochem"] = "Current problems in clinical biochemistry",
  ["curr probl dermatol"] = "Current problems in dermatology",
  ["curr probl diagn radiol"] = "Current problems in diagnostic radiology",
  ["curr probl obstet gynecol"] = "Current problems in obstetrics and gynecology",
  ["curr probl obstet gynecol fertil"] = "Current problems in obstetrics, gynecology and fertility",
  ["curr probl pediatr"] = "Current problems in pediatrics",
  ["curr probl pediatr adolesc health care"] = "Current problems in pediatric and adolescent health care",
  ["curr probl sov med"] = "Current problems of Soviet medicine",
  ["curr probl surg"] = "Current problems in surgery",
  ["curr prostate rep"] = "Current prostate reports",
  ["curr protein pept sci"] = "Current protein & peptide science",
  ["curr proteomics"] = "Current Proteomics",
  ["curr protoc"] = "Current protocols",
  ["curr protoc bioinformatics"] = "Current protocols in bioinformatics",
  ["curr protoc cell biol"] = "Current protocols in cell biology",
  ["curr protoc chem biol"] = "Current protocols in chemical biology",
  ["curr protoc cytom"] = "Current protocols in cytometry",
  ["curr protoc essent lab tech"] = "Current protocols essential laboratory techniques",
  ["curr protoc hum genet"] = "Current protocols in human genetics",
  ["curr protoc immunol"] = "Current protocols in immunology",
  ["curr protoc microbiol"] = "Current protocols in microbiology",
  ["curr protoc mol biol"] = "Current protocols in molecular biology",
  ["curr protoc mouse biol"] = "Current protocols in mouse biology",
  ["curr protoc neurosci"] = "Current protocols in neuroscience",
  ["curr protoc nucleic acid chem"] = "Current protocols in nucleic acid chemistry",
  ["curr protoc pharmacol"] = "Current protocols in pharmacology",
  ["curr protoc plant biol"] = "Current protocols in plant biology",
  ["curr protoc protein sci"] = "Current protocols in protein science",
  ["curr protoc stem cell biol"] = "Current protocols in stem cell biology",
  ["curr protoc toxicol"] = "Current protocols in toxicology",
  ["curr psychiatr"] = "Current psychiatry",
  ["curr psychiatr ther"] = "Current psychiatric therapies",
  ["curr psychiatry rep"] = "Current psychiatry reports",
  ["curr psychiatry rev"] = "Current psychiatry reviews",
  ["curr psychol"] = "Current psychology (New Brunswick, N.J.)",
  ["curr psychol cogn"] = "Current psychology of cognition = Cahiers de psychologie cognitive : CPC",
  ["curr psychol lett"] = "Current psychology letters : behaviour, brain & cognition : CPL",
  ["curr psychopharmacol"] = "Current psychopharmacology",
  ["curr psychos ther rep"] = "Current psychosis & therapeutics reports",
  ["curr pulmonol rep"] = "Current pulmonology reports",
  ["curr radiol rep"] = "Current radiology reports",
  ["curr radiopharm"] = "Current radiopharmaceuticals",
  ["curr res anesth analg"] = "Current researches in anesthesia & analgesia",
  ["curr res bacteriol"] = "Current research in bacteriology",
  ["curr res biotechnol"] = "Current research in biotechnology",
  ["curr res cell biol"] = "Current research in cell biology",
  ["curr res chem biol"] = "Current Research in Chemical Biology",
  ["curr res diabetes obes j"] = "Current research in diabetes & obesity journal",
  ["curr res ecol social psychol"] = "Current Research in Ecological and Social Psychology",
  ["curr res environ appl mycol j fungal biol"] = "Current research in environmental & applied mycology. Journal of fungal biology",
  ["curr res environ sustainability"] = "Current Research in Environmental Sustainability",
  ["curr res food sci"] = "Current research in food science",
  ["curr res green sustainable chem"] = "Current Research in Green and Sustainable Chemistry",
  ["curr res immunol"] = "Current Research in Immunology",
  ["curr res insect sci"] = "Current Research in Insect Science",
  ["curr res microb sci"] = "Current research in microbial sciences",
  ["curr res microbiol biotechnol"] = "Current research in microbiology and biotechnology",
  ["curr res neurobiol"] = "Current Research in Neurobiology",
  ["curr res occup prof"] = "Current research on occupations and professions",
  ["curr res pharmacol drug discovery"] = "Current Research in Pharmacology and Drug Discovery",
  ["curr res physiol"] = "Current research in physiology",
  ["curr res psychol"] = "Current research in psychology",
  ["curr res struct biol"] = "Current research in structural biology",
  ["curr res toxicol"] = "Current research in toxicology",
  ["curr res transl med"] = "Current research in translational medicine",
  ["curr res virol sci"] = "Current Research in Virological Science",
  ["curr respir care rep"] = "Current respiratory care reports",
  ["curr respir med rev"] = "Current respiratory medicine reviews",
  ["curr rev clin exp pharmacol"] = "Current Reviews in Clinical and Experimental Pharmacology",
  ["curr rev musculoskelet med"] = "Current reviews in musculoskeletal medicine",
  ["curr rev pain"] = "Current review of pain",
  ["curr rheumatol rep"] = "Current rheumatology reports",
  ["curr rheumatol rev"] = "Current rheumatology reviews",
  ["curr rob rep"] = "Current Robotics Reports",
  ["curr sci"] = "Current science",
  ["curr sci int"] = "Current science international",
  ["curr sep"] = "Current separations",
  ["curr sep drug dev"] = "Current Separations and Drug Development",
  ["curr sex health rep"] = "Current sexual health reports",
  ["curr signal transduct ther"] = "Current signal transduction therapy",
  ["curr signal transduction ther"] = "Current Signal Transduction Therapy",
  ["curr sleep med rep"] = "Current sleep medicine reports",
  ["curr sociol"] = "Current sociology. La Sociologie contemporaine",
  ["curr sports med rep"] = "Current sports medicine reports",
  ["curr stem cell rep"] = "Current stem cell reports",
  ["curr stem cell res ther"] = "Current stem cell research & therapy",
  ["curr stud hematol blood transfus"] = "Current studies in hematology and blood transfusion",
  ["curr stud nat brain funct"] = "Current studies on the nature of brain function",
  ["curr surg"] = "Current surgery",
  ["curr surg rep"] = "Current surgery reports",
  ["curr sustainable/renewable energy rep"] = "Current Sustainable/Renewable Energy Reports",
  ["curr swed archaeol"] = "Current Swedish Archaeology",
  ["curr theol mission"] = "Currents in theology and mission",
  ["curr ther (seaforth)"] = "Current therapeutics",
  ["curr ther endocrinol metab"] = "Current therapy in endocrinology and metabolism",
  ["curr ther res clin exp"] = "Current therapeutic research, clinical and experimental",
  ["curr tissue eng"] = "Current tissue engineering",
  ["curr tissue microenviron rep"] = "Current tissue microenvironment reports",
  ["curr tissue microenvironrep"] = "Current Tissue Microenvironment Reports",
  ["curr top behav neurosci"] = "Current topics in behavioral neurosciences",
  ["curr top biochem res"] = "Current topics in biochemical research",
  ["curr top cell regul"] = "Current topics in cellular regulation",
  ["curr top comp pathobiol"] = "Current topics in comparative pathobiology",
  ["curr top dev biol"] = "Current topics in developmental biology",
  ["curr top electrochem"] = "Current topics in electrochemistry",
  ["curr top exp endocrinol"] = "Current topics in experimental endocrinology",
  ["curr top eye res"] = "Current topics in eye research",
  ["curr top genet"] = "Current topics in genetics",
  ["curr top hematol"] = "Current topics in hematology",
  ["curr top med chem"] = "Current topics in medicinal chemistry",
  ["curr top med chem (sharjah, united arab emirates)"] = "Current Topics in Medicinal Chemistry (Sharjah, United Arab Emirates)",
  ["curr top med mycol"] = "Current topics in medical mycology",
  ["curr top membr"] = "Current topics in membranes",
  ["curr top membr transp"] = "Current topics in membranes and transport",
  ["curr top microbiol immunol"] = "Current topics in microbiology and immunology",
  ["curr top mol endocrinol"] = "Current topics in molecular endocrinology",
  ["curr top neuroendocrinol"] = "Current Topics In Neuroendocrinology",
  ["curr top nutraceutical res"] = "Current topics in nutraceutical research",
  ["curr top pathol"] = "Current topics in pathology. Ergebnisse der Pathologie",
  ["curr top pept protein res"] = "Current topics in peptide & protein research",
  ["curr top pharmacol"] = "Current topics in pharmacology",
  ["curr top phytochem"] = "Current topics in phytochemistry",
  ["curr top plant biol"] = "Current topics in plant biology",
  ["curr top radiat res q"] = "Current topics in radiation research quarterly",
  ["curr top surg res"] = "Current topics in surgical research",
  ["curr top toxicol"] = "Current topics in toxicology",
  ["curr top virol"] = "Current Topics In Virology",
  ["curr topics plant physiol"] = "Current topics in plant physiology",
  ["curr transl geriatr exp gerontol rep"] = "Current translational geriatrics and experimental gerontology reports",
  ["curr transplant rep"] = "Current transplantation reports",
  ["curr trauma rep"] = "Current trauma reports",
  ["curr treat options allergy"] = "Current treatment options in allergy",
  ["curr treat options cardiovasc med"] = "Current treatment options in cardiovascular medicine",
  ["curr treat options gastroenterol"] = "Current treatment options in gastroenterology",
  ["curr treat options infect dis"] = "Current treatment options in infectious diseases",
  ["curr treat options neurol"] = "Current treatment options in neurology",
  ["curr treat options oncol"] = "Current treatment options in oncology",
  ["curr treat options pediatr"] = "Current treatment options in pediatrics",
  ["curr treat options psychiatry"] = "Current treatment options in psychiatry",
  ["curr treatm opt rheumatol"] = "Current treatment options in rheumatology",
  ["curr trends biomed eng biosci"] = "Current trends in biomedical engineering & biosciences",
  ["curr trends biotechnol pharm"] = "Current trends in biotechnology and pharmacy",
  ["curr trends clin med imaging"] = "Current trends in clinical & medical imaging",
  ["curr trends endocinol"] = "Current trends in endocrinology",
  ["curr trends immunol"] = "Current trends in immunology",
  ["curr trends med chem"] = "Current trends in medicinal chemistry",
  ["curr trends med diagn methods"] = "Current trends in medical diagnostic methods",
  ["curr trends microbiol"] = "Current trends in microbiology",
  ["curr trends neurol"] = "Current trends in neurology",
  ["curr trends polym sci"] = "Current trends in polymer science",
  ["curr trop med rep"] = "Current tropical medicine reports",
  ["curr urol"] = "Current urology",
  ["curr urol rep"] = "Current urology reports",
  ["curr vasc pharmacol"] = "Current vascular pharmacology",
  ["curr womens health rep"] = "Current women's health reports",
  ["curr womens health rev"] = "Current women's health reviews",
  ["curr work hist med"] = "Current work in the history of medicine",
  ["curr world lead"] = "Current world leaders",
  ["curr zool"] = "Current zoology",
  ["current pharm anal"] = "Current Pharmaceutical Analysis",
  ["current pharm biotechnol"] = "Current Pharmaceutical Biotechnology",
  ["current pharmacol rep"] = "Current Pharmacology Reports",
  ["current sci"] = "Current Science",
  ["curric j"] = "The curriculum journal",
  ["curso int oftalmol"] = "Curso Internacional de Oftalmología. Curso Internacional de Oftalmología",
  ["cursos congr univ santiago de compostela"] = "Cursos e Congresos da Universidade de Santiago de Compostela",
  ["curtis’s bot mag"] = "Curtis’s Botanical Magazine",
  ["curved layered struct"] = "Curved And Layered Structures",
  ["cut edge psychiatry pract"] = "Cutting edge psychiatry in practice",
  ["cutan ocul toxicol"] = "Cutaneous and ocular toxicology",
  ["cutaneous ocul toxicol"] = "Cutaneous and Ocular Toxicology",
  ["cvd prev"] = "CVD prevention : the journal of the International Society and Federation of Cardiology",
  ["cvd prev control"] = "CVD prevention and control",
  ["cvi forum"] = "CVI forum",
  ["cvi newswatch"] = "CVI newswatch",
  ["cvir endovasc"] = "CVIR endovascular",
  ["cwi quarterly"] = "Centrum voor Wiskunde en Informatica",
  ["cwi syllabi"] = "CWI Syllabi",
  ["cwi tract"] = "CWI Tract",
  ["cyber secur appl"] = "Cyber Security and Applications",
  ["cyber-phys syst"] = "Cyber-Physical Systems",
  ["cyber-phys syst ser"] = "The Cyber-Physical Systems Series",
  ["cybern inf technol"] = "Cybernetics and Information Technologies",
  ["cybern syst"] = "Cybernetics and systems",
  ["cybern syst anal"] = "Cybernetics and systems analysis",
  ["cybernet systems anal"] = "Cybernetics and Systems Analysis",
  ["cyberpsychol behav"] = "Cyberpsychology & behavior : the impact of the Internet, multimedia and virtual reality on behavior and society",
  ["cyberpsychol behav soc netw"] = "Cyberpsychology, behavior and social networking",
  ["cyberpsychol behav social networking"] = "Cyberpsychology, Behavior, and Social Networking",
  ["cyberpsychology (brno)"] = "Cyberpsychology",
  ["cyborg bionic syst"] = "Cyborg and Bionic Systems",
  ["cylchgrawn llyfrgell genedlaethol cymru"] = "Cylchgrawn Llyfrgell Genedlaethol Cymru. The National Library of Wales journal. National Library of Wales",
  ["cyprus j econ"] = "Cyprus Journal of Economics",
  ["cyprus med j"] = "Cyprus medical journal",
  ["cyprus rev"] = "Cyprus Review",
  ["cyta j food"] = "CyTA Journal of Food",
  ["cytogene genome res"] = "Cytogenetic and Genome Research",
  ["cytogenet cell genet"] = "Cytogenetics and cell genetics",
  ["cytogenet genome res"] = "Cytogenetic and genome research",
  ["cytokine (philadelphia)"] = "Cytokine (Philadelphia)",
  ["cytokine growth factor rev"] = "Cytokine & growth factor reviews",
  ["cytokine x"] = "Cytokine: X",
  ["cytokines cell mol ther"] = "Cytokines, cellular & molecular therapy",
  ["cytokines mol ther"] = "Cytokines and molecular therapy",
  ["cytol genet"] = "Cytology and genetics",
  ["cytologia (tokyo)"] = "Cytologia",
  ["cytometry a"] = "Cytometry. Part A : the journal of the International Society for Analytical Cytology",
  ["cytometry b clin cytom"] = "Cytometry. Part B, Clinical cytometry",
  ["cytometry suppl"] = "Cytometry. Supplement : the journal of the Society for Analytical Cytology",
  ["cytometry, part a"] = "Cytometry, Part A",
  ["cytometry, part b"] = "Cytometry, Part B: Clinical Cytometry",
  ["cytoskeleton (hoboken)"] = "Cytoskeleton (Hoboken, N.J.)",
  ["czas geogr"] = "Czasopismo geograficzne : kwartalnik Zrzeszenia Pol. Nauczycieli Geografji, Towarzystwa Geograficznego we Lwowie i Towarzystwa Geograficznego w Poznaniu",
  ["czas praw hist"] = "Czasopismo prawno-historyczne",
  ["czas stomatol"] = "Czasopismo stomatologiczne",
  ["czech econ dig"] = "Czechoslovak economic digest",
  ["czech econ digest"] = "Czechoslovak Economic Digest",
  ["czech econ pap"] = "Czechoslovak Economic Papers",
  ["czech j anim sci"] = "Czech journal of animal science = Zivočišná výroba",
  ["czech j food sci"] = "Czechoslovak Journal of Food Sciences",
  ["czech j genet plant breed"] = "Czech journal of genetics and plant breeding = Genetika a éslechtéenâi",
  ["czech j operations res"] = "Czechoslovak Journal for Operations Research",
  ["czech j phys"] = "Czechoslovak Journal of Physics",
  ["czech math j"] = "Czechoslovak Mathematical Journal",
  ["czech med"] = "Czechoslovak medicine",
  ["czech mycol"] = "Czech mycology",
  ["czech polar rep"] = "Czech polar reports",
  ["czech sociol rev"] = "Czech sociological review",
  ["czechoslov j phys"] = "Czechoslovak Journal of Physics",
  ["czechoslov sociol rev"] = "Czechoslovak sociological review",
  ["czechoslovak j phys"] = "Czechoslovak Journal of Physics",
  ["czechoslovak math j"] = "Czechoslovak Mathematical Journal",
  ["d cent am"] = "Diario de Centro América (Guatemala, Guatemala : 1880)",
  ["d yucatan"] = "Diario de Yucatán",
  ["da asia rep"] = "DA, Asia report",
  ["da west asia rep"] = "DA, West Asia report",
  ["da zhong ri bao"] = "Da zhong ri bao",
  ["dacca univ stud a"] = "The Dacca University studies: Part A. University of Dacca",
  ["daehan hwangyeong gonghag hoeji"] = "Daehan hwan'gyeong gonghag hoeji",
  ["dagstuhl rep"] = "Dagstuhl reports",
  ["dairy sci technol"] = "Dairy science & technology",
  ["dakar med"] = "Dakar Medical",
  ["dalhous law j"] = "Dalhousie law journal",
  ["dalhousie dent j"] = "Dalhousie dental journal",
  ["dalhousie rev"] = "The Dalhousie review",
  ["dalian hai yang da xue xue bao"] = "Dalian hai yang da xue xue bao",
  ["dallas med j"] = "The Dallas medical journal",
  ["dalnevost mat zh"] = "Dalnevostochnyi Matematicheskii Zhurnal. Far Eastern Mathematical Journal",
  ["dalnevost z infekc patol"] = "Dalʹnevostochnyĭ zhurnal infekt︠s︡ionnoĭ patologii",
  ["dalton trans"] = "Dalton Transactions",
  ["dan amtsrad"] = "Danmarks amtsrad",
  ["dan j archaeol"] = "Danish Journal of Archaeology",
  ["dan kemi"] = "Danski Kemi",
  ["dan med bull"] = "Danish medical bulletin",
  ["dan med j"] = "Danish medical journal",
  ["dan medicinhist arbog"] = "Dansk medicinhistorisk årbog",
  ["dan tidsskr farm"] = "Dansk tidsskrift for farmaci",
  ["dang dai zhongguo yan jiu"] = "Dang dai Zhongguo yan jiu = Modern China studies",
  ["dangerous prop ind mater rep"] = "Dangerous Properties of Industrial Materials Report",
  ["dapim refuiim"] = "Dapim refuiim. Folia medica",
  ["dar mag"] = "Daughters of the American Revolution magazine",
  ["daru j pharm sci"] = "DARU Journal of Pharmaceutical Sciences",
  ["dasc picom datacom cyberscitech 2017 (2017)"] = "DASC-PICom-DataCom-CyberSciTech 2017 : 2017 IEEE 15th International Conference on Dependable, Autonomic and Secure Computing ; 2017 IEEE 15th International Conference on Pervasive Intelligence and Computing ; 2017 IEEE 3rd International Conference on Big Data Intelligence and Computing ; 2017 IEEE Cyber Science and Technology Congress : 6-11 November 2017, Orlando, Florida. IEEE International Symposium on Dependable, Autonomic and Secure Computing (15th : 2017 : Orlando, Fla.)",
  ["data (basel)"] = "Data",
  ["data appl secur priv xxxii (2018)"] = "Data and applications security and privacy XXXII : 32nd Annual IFIP WG 11.3 Conference, DBSec 2018, Bergamo, Italy, July 16-18, 2018, Proceedings. Annual IFIP WG 11.3 Working Conference on Data and Applications Security (32nd : 2018 : Bergamo, Italy)",
  ["data asia"] = "Data Asia",
  ["data basics"] = "Data basics : a publication supported by and for the members of The Society for Clinical Data Management, Inc",
  ["data brief"] = "Data in brief",
  ["data bull (cent stud health syst change)"] = "Data bulletin (Center for Studying Health System Change)",
  ["data inf manag"] = "Data and information management",
  ["data inf manage"] = "Data and Information Management",
  ["data intell"] = "Data intelligence",
  ["data knowl eng"] = "Data and Knowledge Engineering",
  ["data min big data (2017)"] = "Data Mining and Big Data : second International Conference, DMBD 2017, Fukuoka, Japan, July 27-August 1, 2017. Proceedings. DMBD (Conference) (2nd : 2017 : Fukuoka, Japan)",
  ["data min knowl discov"] = "Data mining and knowledge discovery",
  ["data min knowl discovery"] = "Data Mining and Knowledge Discovery",
  ["data policy"] = "Data & policy",
  ["data sci"] = "Data Science",
  ["data sci digital eng"] = "Data Science and Digital Engineering",
  ["data sci eng"] = "Data Science and Engineering",
  ["data sci j"] = "Data Science Journal",
  ["data sci manage"] = "Data Science and Management",
  ["data sci sci"] = "Data Science in Science",
  ["data strateg benchmarks"] = "Data strategies & benchmarks : the monthly advisory for health care executives",
  ["data strategies benchmarks"] = "Data Strategies and Benchmarks",
  ["data technol appl"] = "Data Technologies and Applications",
  ["data user news"] = "Data user news",
  ["data-centric eng"] = "Data-Centric Engineering",
  ["data-driven modell"] = "Data-Driven Modelling",
  ["database (oxford)"] = "Database : the journal of biological databases and curation",
  ["dataset pap biol"] = "Dataset papers in biology",
  ["datenbank spektrum"] = "Datenbank-Spektrum : Zeitschrift für Datenbanktechnologie : Organ der Fachgruppe Datenbanken der Gesellschaft für Informatik e.V",
  ["dauphin cty rep"] = "Dauphin County reports",
  ["david hilbert’s lect found math phys 1891–1933"] = "David Hilbert’s Lectures on the Foundations of Mathematics and Physics 1891–1933",
  ["davis nurs surv"] = "The Davis nursing survey",
  ["davos rev"] = "Davoser Revue",
  ["daxue huaxue"] = "Daxue Huaxue",
  ["de economia"] = "De economía",
  ["de economist"] = "De Economist",
  ["de economía"] = "De Economía",
  ["de gruyter exp math"] = "De Gruyter Expositions in Mathematics",
  ["de gruyter grad"] = "De Gruyter Graduate",
  ["de gruyter proc"] = "De Gruyter Proceedings",
  ["de gruyter proc math"] = "De Gruyter Proceedings in Mathematics",
  ["de gruyter ser appl math eng inf sci"] = "De Gruyter Series on the Applications of Mathematics in Engineering and Information Sciences",
  ["de gruyter ser appl numer math"] = "De Gruyter Series in Applied and Numerical Mathematics",
  ["de gruyter ser comput sci eng"] = "De Gruyter Series in Computational Science and Engineering",
  ["de gruyter ser discrete math appl"] = "De Gruyter Series in Discrete Mathematics and Applications",
  ["de gruyter ser log appl"] = "De Gruyter Series in Logic and its Applications",
  ["de gruyter ser math life sci"] = "De Gruyter Series in Mathematics and Life Sciences",
  ["de gruyter ser nonlinear anal appl"] = "De Gruyter Series in Nonlinear Analysis and Applications",
  ["de gruyter ser probab stoch"] = "De Gruyter Series in Probability and Stochastics",
  ["de gruyter stud"] = "De Gruyter Studium",
  ["de gruyter stud math"] = "De Gruyter Studies in Mathematics",
  ["de gruyter stud math phys"] = "De Gruyter Studies in Mathematical Physics",
  ["de gruyter textbook"] = "de Gruyter Textbook",
  ["de j dent eng"] = "DE; the journal of dental engeering",
  ["de paul law rev"] = "De Paul law review",
  ["deaconess hosp med bull"] = "Deaconess Hospital Medical Bulletin",
  ["deafness educ int"] = "Deafness & education international : the journal of the British Association of Teachers of the Deaf",
  ["deans list"] = "Deans List",
  ["deans notes"] = "Deans Notes",
  ["death educ"] = "Death education",
  ["death stud"] = "Death studies",
  ["debates ner"] = "Debates do NER : publicação do Núcleo de Estudos da Religião do Programa de Pós-Graduação em Antropologia Social da Universidade Federal do Rio Grande do Sul",
  ["deccan geogr"] = "The Deccan geographer",
  ["dechema biotechnol conf"] = "DECHEMA Biotechnology Conferences",
  ["decis anal"] = "Decision analysis : a journal of the Institute for Operations Research and the Management Sciences",
  ["decis econ finance"] = "Decisions in Economics and Finance. A Journal of Applied Mathematics",
  ["decis eng"] = "Decision Engineering",
  ["decis mak manuf serv"] = "Decision Making in Manufacturing and Services",
  ["decis sci"] = "Decision Sciences",
  ["decis support syst"] = "Decision support systems",
  ["decision (wash d c )"] = "Decision (Washington, D.C.)",
  ["decision sci"] = "Decision Science",
  ["deep dev educ exch pap"] = "Development education exchange papers : DEEP",
  ["deep learn data label med appl (2016)"] = "Deep learning and data labeling for medical applications : First International Workshop, LABELS 2016, and Second International Workshop, DLMIA 2016, held in conjunction with MICCAI 2016, Athens, Greece, October 21, 2016, proceedings",
  ["deep learn med image anal multimodal learn clin decis support (2017)"] = "Deep learning in medical image analysis and multimodal learning for clinical decision support : Third International Workshop, DLMIA 2017, and 7th International Workshop, ML-CDS 2017, held in conjunction with MICCAI 2017 Québec City, QC, Canada, September 14, 2017, proceedings",
  ["deep learn med image anal multimodal learn clin decis support (2018)"] = "Deep Learning in Medical Image Analysis and Multimodal Learning for Clinical Decision Support : 4th International Workshop, DLMIA 2018, and 8th International Workshop, ML-CDS 2018, held in conjunction with MICCAI 2018, Granada, Spain, September 20, 2018 : proceedings. DLMIA (Workshop) (4th : 2018 : Granada, Spain)",
  ["deep sea res 1 oceanogr res pap"] = "Deep-sea research. Part I, Oceanographic research papers",
  ["deep sea res 2 top stud oceanogr"] = "Deep-sea research. Part II, Topical studies in oceanography",
  ["deep sea res a"] = "Deep-sea research. Part A, Oceanographic research papers",
  ["deep sea res part a"] = "Deep Sea Research Part A: Oceanographic Research Papers",
  ["deep sea res part b"] = "Deep Sea Research Part B: Oceanographic Literature Review",
  ["deep sea res part i"] = "Deep Sea Research Part I: Oceanographic Research Papers",
  ["deep sea res part ii"] = "Deep Sea Research Part II: Topical Studies in Oceanography",
  ["deep underground sci eng"] = "Deep Underground Science & Engineering",
  ["deep-sea res"] = "Deep-Sea Research",
  ["def acquis res j"] = "Defense acquisition research journal",
  ["def couns j"] = "Defense counsel journal",
  ["def law j"] = "Defense law journal",
  ["def life sci j"] = "Defence life science journal",
  ["def natl"] = "Défense nationale",
  ["def odontol"] = "Defensa odontológica",
  ["def sci j"] = "Defence science journal",
  ["def secur anal"] = "Defense & Security Analysis",
  ["def stud"] = "Defence Studies",
  ["def technol"] = "Defence Technology",
  ["defence econ"] = "Defence Economics",
  ["defence peace econ"] = "Defence and Peace Economics",
  ["degener neurol neuromuscul dis"] = "Degenerative neurological and neuromuscular disease",
  ["dei j sci eng res"] = "DEI Journal of Science & Engineering Research",
  ["dej mat/hist math"] = "Dejiny Matematiky/History of Mathematics",
  ["dejiny ved tech"] = "DVT, Dějiny věd a techniky",
  ["del code annot del"] = "Delaware code annotated. Delaware",
  ["del hist"] = "Delaware history",
  ["del med j"] = "Delaware medical journal",
  ["del nurse"] = "Delaware nurse",
  ["dela j public health"] = "Delaware journal of public health",
  ["delft prog rep"] = "Delft progress report",
  ["delt hell mikrobiol hygieinol hetair"] = "Deltion Hellenikes Mikrobiologikes kai Hygieinologikes Hetaireias",
  ["delt paidiatr klin panepistem athenon"] = "Deltion tes Paidiatrikes Klinikes tou Panepistemiou Athenon",
  ["delta j sci"] = "Delta Journal of Science",
  ["deltion iatrocheirourgike hetaireia athenon"] = "Deltion. Iatrocheirourgikē Hetaireia Athēnōn",
  ["dement geriatr cogn dis extra"] = "Dementia and geriatric cognitive disorders extra",
  ["dement geriatr cogn disord"] = "Dementia and geriatric cognitive disorders",
  ["dement neurocogn disord"] = "Dementia and neurocognitive disorders",
  ["dement neuropsychol"] = "Dementia & neuropsychologia",
  ["dementia (london)"] = "Dementia (London, England)",
  ["demogr afr"] = "Démographie africaine",
  ["demogr bull"] = "Demographic bulletin (Wellington, N.Z.)",
  ["demogr econ"] = "Demografía y economía",
  ["demogr hist bull inf soc"] = "Dh : bulletin d'information",
  ["demogr india"] = "Demography India",
  ["demogr inf"] = "Demographische Informationen",
  ["demogr issled"] = "Demograficheskie issledovanii︠a︡ (Kiev, Ukraine)",
  ["demogr res"] = "Demographic research",
  ["demogr res monogr"] = "Demographic Research Monographs",
  ["demogr sveske"] = "Demografske sveske",
  ["demohrafichni doslidzhennia"] = "Demohrafichni doslidz︠h︡enni︠a︡",
  ["demonstr math"] = "Demonstratio Mathematica",
  ["demonstratio math"] = "Demonstratio Mathematica",
  ["den jyske hist"] = "Den jyske historiker",
  ["dendrochronologia (verona)"] = "Dendrochronologia",
  ["dengue bull"] = "Dengue bulletin",
  ["denki kagaku oyobi kogyo butsuri kagaku"] = "Denki Kagaku oyobi Kogyo Butsuri Kagaku",
  ["denkschr osterr akad wiss philos-hist kl"] = "Denkschriften Osterreichische Akademie der Wissenschaften, Philosophisch-Historische Klasse",
  ["denkschr schweiz natforsch ges"] = "Denkschriften der Schweizerischen naturforschenden Gesellschaft",
  ["dens (curitiba)"] = "Dens",
  ["dens sapiens"] = "Dens Sapiens",
  ["dent 3000"] = "Dentistry 3000",
  ["dent abstr"] = "Dental abstracts; a selection of world dental literature",
  ["dent anaesth sedat"] = "Dental anaesthesia and sedation",
  ["dent angles"] = "Dental Angles",
  ["dent assist"] = "The Dental assistant",
  ["dent assist (waco tx)"] = "Dental assisting",
  ["dent assist j"] = "The Dental assistant journal : journal of the American Dental Assistants Association",
  ["dent cadmos"] = "Dental Cadmos",
  ["dent caries fluor"] = "Dental caries and fluorine. American Association for the Advancement of Science. Section on Medical Sciences. Subsection on Dentistry",
  ["dent clin north am"] = "Dental clinics of North America",
  ["dent concepts"] = "Dental concepts",
  ["dent delin"] = "Dental delineator",
  ["dent dialogue"] = "Dental dialogue (Bombay, India)",
  ["dent dienst"] = "Dental Dienst; Fachzeitschrift für den Dental-Markt; technisches Fachblatt für Prothetik",
  ["dent dig"] = "Dental digest",
  ["dent dimens"] = "Dental dimensions",
  ["dent discourse"] = "Dental Discourse",
  ["dent echo (heidelb)"] = "Dental echo",
  ["dent econ"] = "Dental economics - oral hygiene",
  ["dent gaz"] = "The Dental gazette",
  ["dent health"] = "Dental health",
  ["dent health (london)"] = "Dental Health",
  ["dent hist"] = "Dental historian : Lindsay Club newsletter",
  ["dent hyg (chic)"] = "Dental hygiene",
  ["dent hyg (san franc)"] = "Dental hygienist (San Francisco, Calif.)",
  ["dent hypotheses"] = "Dental hypotheses",
  ["dent images"] = "Dental images",
  ["dent implantol update"] = "Dental implantology update",
  ["dent items interest"] = "Dental items of interest",
  ["dent j"] = "Dental journal",
  ["dent j (basel)"] = "Dentistry journal",
  ["dent j aust"] = "The Dental journal of Australia",
  ["dent j malays"] = "Dental journal of Malaysia",
  ["dent j malaysia singapore"] = "The Dental journal of Malaysia & Singapore",
  ["dent j zamb"] = "Dental journal of Zambia : official publication of the Zambia Dental Association",
  ["dent jpn (tokyo)"] = "Dentistry in Japan",
  ["dent lab bl"] = "Dental laboratorie bladet",
  ["dent lab manage today"] = "Dental lab management today",
  ["dent lab rev"] = "Dental laboratory review",
  ["dent labor (munch)"] = "Das Dental-Labor. Le Laboratoire dentaire. The Dental laboratory",
  ["dent mag oral top"] = "The Dental magazine and oral topics",
  ["dent manage"] = "Dental management",
  ["dent mater"] = "Dental materials : official publication of the Academy of Dental Materials",
  ["dent mater j"] = "Dental materials journal",
  ["dent med probl"] = "Dental and medical problems",
  ["dent mirror (atlanta)"] = "Dental Mirror",
  ["dent mirror (quezon city)"] = "Dental Mirror",
  ["dent news"] = "Dentist news",
  ["dent news (lond)"] = "Dental news",
  ["dent off"] = "Dental office",
  ["dent oral craniofac res"] = "Dental, oral, and craniofacial research",
  ["dent outlook"] = "The Dental outlook",
  ["dent pract"] = "Dental practice",
  ["dent pract (cincinnati)"] = "Dental practice; views, trends and news of dentistry",
  ["dent pract (ewell)"] = "DP. Dental practice",
  ["dent pract dent rec"] = "The Dental practitioner and dental record",
  ["dent pract manage"] = "Dental practice management",
  ["dent press"] = "Dental press",
  ["dent prog (chic)"] = "Dental progress",
  ["dent radiogr photogr"] = "Dental radiography and photography",
  ["dent rec (london)"] = "The Dental record",
  ["dent res grad study"] = "Dental research and graduate study",
  ["dent res j (isfahan)"] = "Dental research journal",
  ["dent rev"] = "Dental-revue (Solothurn, Switzerland)",
  ["dent rundsch"] = "Dentistische Rundschau",
  ["dent sch q"] = "Dental School quarterly",
  ["dent stud"] = "Dental student",
  ["dent surv"] = "Dental survey",
  ["dent team"] = "Dental Team",
  ["dent teamwork"] = "Dental teamwork",
  ["dent tech"] = "The Dental technician",
  ["dent ther newsl"] = "Dental Therapeutics Newsletter",
  ["dent today"] = "Dentistry today",
  ["dent traumatol"] = "Dental traumatology : official publication of International Association for Dental Traumatology",
  ["dent update"] = "Dental update",
  ["dent world"] = "Dental world (London, England)",
  ["dental press j orthod"] = "Dental press journal of orthodontics",
  ["dentistry (lisle)"] = "Dentistry (Lisle, Ill.)",
  ["dentistry (loma linda)"] = "Dentistry (Loma Linda, Calif.)",
  ["dentistry (sunnyvale)"] = "Dentistry (Sunnyvale, Calif.)",
  ["dentomaxillofac radiol"] = "Dento maxillo facial radiology",
  ["dentomaxillofac radiol suppl"] = "Dento-maxillo-facial radiology. Supplement",
  ["dentoral (istanbul)"] = "Dentoral",
  ["denver j int law policy"] = "Denver journal of international law and policy",
  ["denver law j"] = "Denver law journal",
  ["denver univ law rev"] = "Denver University law review",
  ["denver west roundup"] = "The Denver Westerners roundup",
  ["dep health soc serv q mag"] = "Department of Health and Social Services quarterly magazine",
  ["dep state bull"] = "The Department of State bulletin",
  ["depaul j health care law"] = "DePaul journal of health care law",
  ["depend model"] = "Dependence Modeling",
  ["depos rec"] = "The depositional record : a journal of biological, physical and geochemical sedimentary processes",
  ["depositional rec"] = "Depositional Record",
  ["depress anxiety"] = "Depression and anxiety",
  ["depress causes diagn"] = "Depression-causes Diagnosis and Treatment",
  ["depress mind body"] = "Depression, mind and body",
  ["depress res treat"] = "Depression research and treatment",
  ["der donaurum"] = "Der Donaurum",
  ["derm beruf umwelt"] = "Dermatosen in Beruf und Umwelt. Occupation and environment",
  ["dermatol case rep"] = "Dermatology case reports",
  ["dermatol clin"] = "Dermatologic clinics",
  ["dermatol iber lat am"] = "Dermatologia ibero latino-americana",
  ["dermatol int"] = "Dermatologia internationalis",
  ["dermatol monatsschr"] = "Dermatologische Monatschrift",
  ["dermatol nurs"] = "Dermatology nursing",
  ["dermatol nurs (lond)"] = "Dermatological nursing : the journal of the British Dermatological Nursing Group",
  ["dermatol online j"] = "Dermatology online journal",
  ["dermatol pract concept"] = "Dermatology practical & conceptual",
  ["dermatol reports"] = "Dermatology reports",
  ["dermatol res pract"] = "Dermatology research and practice",
  ["dermatol rev mex"] = "Dermatologia; revista mexicana",
  ["dermatol surg"] = "Dermatologic surgery : official publication for American Society for Dermatologic Surgery [et al.]",
  ["dermatol ther"] = "Dermatologic therapy",
  ["dermatol ther (heidelb)"] = "Dermatology and therapy",
  ["dermatol times"] = "Dermatology Times",
  ["dermatol trop ecol geogr"] = "Dermatologia tropica et ecologica geographica",
  ["dermatol wochenschr"] = "Dermatologische Wochenschrift",
  ["dermatopathology (basel)"] = "Dermatopathology (Basel, Switzerland)",
  ["dermosifilografo (torino)"] = "Il Dermosifilografo",
  ["des autom embedded syst"] = "Design Automation for Embedded Systems",
  ["des codes cryptogr"] = "Designs, codes, and cryptography",
  ["des eng can"] = "Design Engineering Canada",
  ["des issues"] = "Design Issues",
  ["des manage j"] = "Design Management Journal",
  ["des manage rev"] = "Design management review",
  ["des monomers polym"] = "Designed monomers and polymers",
  ["des sci"] = "Design Science",
  ["des stud"] = "Design studies",
  ["des technol educ"] = "Design and technology education : an international journal",
  ["des user exp usability"] = "Design, user experience, and usability : design philosophy, methods, and tools : second international Conference, DUXU 2013, held as part of HCI International 2013, Las Vegas, NV, USA, July 21-26, 2013, Proceedings. Part I. DUXU (Conference) (2nd : 2013 : Las Vegas, Nev.)",
  ["desalin water treat"] = "Desalination and Water Treatment",
  ["desalination water treat"] = "Desalination and water treatment",
  ["desarro base"] = "Desarrollo de base : revista de la Fundación Interamericana",
  ["desarro rural amer"] = "Desarrollo rural en las Américas",
  ["desarro soc"] = "Desarrollo y sociedad",
  ["desarrollo econ"] = "Desarrollo económico",
  ["descr ectomycorrhizae"] = "Descriptions of ectomycorrhizae",
  ["deseret news"] = "Deseret news (Salt Lake City, Utah : 1964)",
  ["desertif control"] = "Desertification control",
  ["design health (abingdon)"] = "Design for health (Abingdon, England)",
  ["designs (basel)"] = "Designs",
  ["detroit coll law rev"] = "Detroit College of Law review",
  ["detroit dent bull"] = "Detroit dental bulletin",
  ["detroit med news"] = "Detroit medical news",
  ["detroit perspect"] = "Detroit in perspective",
  ["deutsche hochschulschrift"] = "Deutsche Hochschulschriften",
  ["dev bank jpn res ser"] = "Development Bank of Japan Research Series",
  ["dev biol"] = "Developmental biology",
  ["dev biol (amsterdam, neth)"] = "Developmental Biology (Amsterdam, Netherlands)",
  ["dev biol (basel)"] = "Developments in biologicals",
  ["dev biol (basel, switz)"] = "Developments in Biologicals (Basel, Switzerland)",
  ["dev biol (n y 1985)"] = "Developmental biology (New York, N.Y. : 1985)",
  ["dev biol (san diego, ca, u s)"] = "Developmental Biology (San Diego, CA, United States)",
  ["dev biol stand"] = "Developments in biological standardization",
  ["dev brief"] = "Development brief",
  ["dev built environ"] = "Developments in the Built Environment",
  ["dev bull"] = "Development bulletin (Australian Development Studies Network)",
  ["dev cell"] = "Developmental cell",
  ["dev change"] = "Development and change",
  ["dev child welf"] = "Developmental child welfare",
  ["dev civilis"] = "Développement & civilisations",
  ["dev clay sci"] = "Developments in clay science",
  ["dev cogn neurosci"] = "Developmental cognitive neuroscience",
  ["dev cognit neurosci"] = "Developmental Cognitive Neuroscience",
  ["dev commun rep"] = "Development communication report",
  ["dev comp immunol"] = "Developmental and comparative immunology",
  ["dev dialogue"] = "Development dialogue",
  ["dev dig"] = "Development digest",
  ["dev dir"] = "Development directions",
  ["dev disabil res rev"] = "Developmental disabilities research reviews",
  ["dev dyn"] = "Developmental dynamics : an official publication of the American Association of Anatomists",
  ["dev econ"] = "The Developing economies",
  ["dev endocrinol"] = "Developments in endocrinology",
  ["dev eng"] = "Development engineering",
  ["dev forum"] = "Development forum",
  ["dev gend brief"] = "Development and gender in brief : a quarterly update from BRIDGE, raising gende awareness among policy-makers and practitioners",
  ["dev genes evol"] = "Development genes and evolution",
  ["dev genet"] = "Developmental Genetics",
  ["dev growth differ"] = "Development, growth & differentiation",
  ["dev health econ public policy"] = "Developments in health economics and public policy",
  ["dev heat transf"] = "Developments in Heat Transfer",
  ["dev immunol"] = "Developmental immunology",
  ["dev int"] = "Development international",
  ["dev math"] = "Developments in Mathematics",
  ["dev med child neurol"] = "Developmental medicine and child neurology",
  ["dev med child neurol suppl"] = "Developmental medicine and child neurology. Supplement",
  ["dev network"] = "Development network",
  ["dev neurobiol"] = "Developmental neurobiology",
  ["dev neuropsychol"] = "Developmental neuropsychology",
  ["dev neurorehabil"] = "Developmental neurorehabilitation",
  ["dev neurosci"] = "Developmental neuroscience",
  ["dev neurosci (basel, switz)"] = "Developmental Neuroscience (Basel, Switzerland)",
  ["dev ophthalmol"] = "Developments in ophthalmology",
  ["dev outreach"] = "Development outreach : putting knowledge to work for development",
  ["dev peace"] = "Development and peace",
  ["dev period med"] = "Developmental period medicine",
  ["dev pharmacol ther"] = "Developmental pharmacology and therapeutics",
  ["dev policy rev"] = "Development policy review : the journal of the Overseas Development Institute",
  ["dev pract"] = "Development in practice",
  ["dev psychobiol"] = "Developmental psychobiology",
  ["dev psychol"] = "Developmental psychology",
  ["dev psychopathol"] = "Development and psychopathology",
  ["dev que"] = "Développement-Québec",
  ["dev reprod"] = "Development & reproduction",
  ["dev rev"] = "Developmental review : DR",
  ["dev sante"] = "Développement et santé : revue de perfectionnement médical et sanitaire en pays tropical",
  ["dev sci"] = "Developmental science",
  ["dev soc"] = "Development and society (Sŏul Taehakkyo. Institute for Social Devdelopment and Policy Research)",
  ["dev south afr"] = "Development Southern Africa",
  ["dev suppl"] = "Development (Cambridge, England). Supplement",
  ["dev toxicol environ sci"] = "Developments in toxicology and environmental science",
  ["dev volta"] = "Le Développement voltaïque",
  ["dev world bioeth"] = "Developing world bioethics",
  ["dev world bioethics"] = "Developing World Bioethics",
  ["devel change"] = "Development and Change",
  ["devel southern africa"] = "Development Southern Africa",
  ["develop petrol"] = "Developments in Petrology",
  ["developing economies"] = "Developing Economies",
  ["development (cambridge, u k)"] = "Development (Cambridge, United Kingdom)",
  ["development (rome)"] = "Development (Society for International Development)",
  ["developnet news"] = "DevelopNet news",
  ["deviant behav"] = "Deviant behavior",
  ["devon assoc adv sci"] = "Report and transactions - The Devonshire Association for the Advancement of Science, Literature and Art. Devonshire Association for the Advancement of Science, Literature and Art",
  ["devon hist"] = "The Devon historian",
  ["dfi j"] = "DFI Journal",
  ["dgds differ geom dyn syst monogr"] = "DGDS. Differential Geometry—Dynamical Systems. Monographs",
  ["dghs chron"] = "Dghs Chronicle",
  ["dh15 (2015)"] = "DH'15: proceedings of the 5th International Conference on Digital Health 2015 : May 18-20, 2015, Florence, Italy. International Conference on Digital Health (5th : 2015 : Florence, Italy)",
  ["dhaka univ j pharm sci"] = "The Dhaka University journal of pharmaceutical sciences",
  ["dhaka univ stud b biol stud"] = "The Dhaka University studies. Part B",
  ["dhs dimens"] = "DHS+ dimensions : a biannual newsletter of the Demographic and Health Surveys project",
  ["di 3 jun yi da xue xue bao"] = "Di 3 jun yi da xue xue bao = Disanjunyidaxue xuebao = Acta Acadiemiae [i.e. Academiae] Medicinae Militaris Tertiae",
  ["di 4 jun yi da xue xue bao"] = "Di 4 jun yi da xue xue bao = Journal of the Fourth Military Medical University = Disi junyi daxue xuebao",
  ["di fang bing tong bao"] = "Di fang bing tong bao = Endemic disease bulletin",
  ["di yi jun yi da xue xue bao"] = "Di 1 jun yi da xue xue bao = Academic journal of the first medical college of PLA",
  ["dia med"] = "El Día médico",
  ["dia med urug"] = "Dia medico uruguayo",
  ["diab vasc dis res"] = "Diabetes & vascular disease research",
  ["diabet foot ankle"] = "Diabetic foot & ankle",
  ["diabet med"] = "Diabetic medicine : a journal of the British Diabetic Association",
  ["diabet nephrop"] = "Diabetic nephropathy : DN",
  ["diabete metab"] = "Diabète & métabolisme",
  ["diabetes care"] = "Diabetes care",
  ["diabetes care educ newsl"] = "Diabetes care & education newsletter",
  ["diabetes educ"] = "The Diabetes educator",
  ["diabetes forecast"] = "Diabetes forecast",
  ["diabetes manag (lond)"] = "Diabetes management (London, England)",
  ["diabetes metab"] = "Diabetes & metabolism",
  ["diabetes metab j"] = "Diabetes & metabolism journal",
  ["diabetes metab res rev"] = "Diabetes/metabolism research and reviews",
  ["diabetes metab rev"] = "Diabetes/metabolism reviews",
  ["diabetes metab syndr"] = "Diabetes & metabolic syndrome",
  ["diabetes metab syndr obes"] = "Diabetes, metabolic syndrome and obesity : targets and therapy",
  ["diabetes nutr metab"] = "Diabetes, nutrition & metabolism",
  ["diabetes obes int j"] = "Diabetes & obesity international journal",
  ["diabetes obes metab"] = "Diabetes, obesity & metabolism",
  ["diabetes res"] = "Diabetes research (Edinburgh, Scotland)",
  ["diabetes res (fairfax)"] = "Diabetes research (Fairfax, Va.)",
  ["diabetes res clin pract"] = "Diabetes research and clinical practice",
  ["diabetes res clin pract suppl"] = "Diabetes research and clinical practice. Supplement",
  ["diabetes rev (alex)"] = "Diabetes reviews (Alexandria, Va.)",
  ["diabetes self manag"] = "Diabetes self-management",
  ["diabetes spectr"] = "Diabetes spectrum : a publication of the American Diabetes Association",
  ["diabetes technol ther"] = "Diabetes technology & therapeutics",
  ["diabetes ther"] = "Diabetes therapy : research, treatment and education of diabetes and related disorders",
  ["diabetol croat"] = "Diabetologia Croatica",
  ["diabetol int"] = "Diabetology international",
  ["diabetol metab syndr"] = "Diabetology & metabolic syndrome",
  ["diagn afr"] = "Diagnostics in Africa",
  ["diagn clin immunol"] = "Diagnostic and Clinical Immunology",
  ["diagn cytopathol"] = "Diagnostic cytopathology",
  ["diagn gynecol obstet"] = "Diagnostic gynecology and obstetrics",
  ["diagn histopathol"] = "Diagnostic histopathology",
  ["diagn histopathol (oxf)"] = "Diagnostic histopathology (Oxford, England)",
  ["diagn imaging"] = "Diagnostic imaging",
  ["diagn imaging clin med"] = "Diagnostic imaging in clinical medicine",
  ["diagn imaging eur"] = "Diagnostic imaging Europe",
  ["diagn immunol"] = "Diagnostic immunology",
  ["diagn interv imaging"] = "Diagnostic and interventional imaging",
  ["diagn interv radiol"] = "Diagnostic and interventional radiology (Ankara, Turkey)",
  ["diagn interventional imaging"] = "Diagnostic and Interventional Imaging",
  ["diagn microbiol infect dis"] = "Diagnostic microbiology and infectious disease",
  ["diagn mol pathol"] = "Diagnostic Molecular Pathology",
  ["diagn pathol"] = "Diagnostic pathology",
  ["diagn pathol open access"] = "Diagnostic pathology: open access",
  ["diagn progn res"] = "Diagnostic and prognostic research",
  ["diagn ther endosc"] = "Diagnostic and therapeutic endoscopy",
  ["diagn traitements"] = "Diagnostics & traitements",
  ["diagnosi (napoli)"] = "La Diagnosi; rivista mensile di medicina pratica",
  ["diagnosis (berl)"] = "Diagnosis (Berlin, Germany)",
  ["diagnostics (basel)"] = "Diagnostics (Basel, Switzerland)",
  ["dial transplant"] = "Dialysis & transplantation",
  ["dialect anthropol"] = "Dialectical anthropology",
  ["dialect hum"] = "Dialectics and humanism",
  ["dialectica (bern)"] = "Dialectica (Bern, Switzerland)",
  ["dializ transplant yanik"] = "Dializ, transplantasyon ve yanik : Türkiye Organ Nakli ve Yanik Tedavi Vakfi, tip bilimleri dergisi = Dialysis, transplantation & burn : medical journal of the Turkish Transplantation and Burn Foundation",
  ["dialog cardiovasc med"] = "Dialogues in cardiovascular medicine : DCM",
  ["dialog fairleigh dickinson univ sch dent"] = "The Dialog",
  ["dialog games log"] = "Dialogues and Games of Logic",
  ["dialogue (lyon)"] = "Dialogue (Saint-Etienne, Loire, France)",
  ["dialogue can philos assoc"] = "Dialogue",
  ["dialogue diarrhoea"] = "Dialogue on diarrhoea",
  ["dialogues clim change"] = "Dialogues on Climate Change",
  ["dialogues clin neurosci"] = "Dialogues in clinical neuroscience",
  ["dialogues contracept"] = "Dialogues in contraception",
  ["dialogues hum geogr"] = "Dialogues in human geography",
  ["diam relat mater"] = "Diamond and related materials",
  ["diamond films technol"] = "Diamond Films and Technology",
  ["diamond relat mater"] = "Diamond and Related Materials",
  ["dian hua xue"] = "Dian hua xue",
  ["dian zi xian wei xue bao"] = "Dian zi xian wei xue bao = Journal of Chinese Electron Microscopy Society",
  ["dianzi keji daxue xuebao"] = "Dianzi Keji Daxue Xuebao",
  ["diarrhoea dialogue"] = "Diarrhoea Dialogue",
  ["diatom res"] = "Diatom research : the journal of the International Society for Diatom Research",
  ["dickinson law rev"] = "Dickinson law review",
  ["dicle univ tip fakul derg"] = "Dicle Üniversitesi Tip Fakültesi dergisi = The journal of the Faculty of Medicine, University of Dicle",
  ["didatt propos esper"] = "Didattica: Proposte ed Esperienze",
  ["die grundlehren der mathematischen wissenschaften in einzeldarstellungen mit besonderer berucksichtigung der anwendung"] = "Die Grundlehren der mathematischen Wissenschaften in Einzeldarstellungen mit besonderer Berucksichtigung der Anwendungsgebiete",
  ["die grundlehren der mathematischen wissenschaften in einzeldarstellungen mit besonderer berucksightigung der anwendung"] = "Die Grundlehren der mathematischen Wissenschaften in Einzeldarstellungen mit besonderer Berucksightigung der Anwendungsgebiete",
  ["dif soc"] = "Difesa sociale",
  ["differ equ"] = "Differential Equations",
  ["differ equ appl"] = "Differential Equations & Applications",
  ["differ equ dyn syst"] = "Differential equations and dynamical systems",
  ["differ equations"] = "Differential Equations",
  ["differ equations dyn syst"] = "Differential Equations and Dynamical Systems",
  ["differ geom appl"] = "Differential geometry and its applications",
  ["differ geom dyn syst"] = "Differential Geometry—Dynamical Systems",
  ["differ integr equations"] = "Differential and Integral Equations",
  ["differ uravn"] = "Differentsial\\cprime nye Uravneniya",
  ["differ uravn protsessy upr"] = "Differentsialnye Uravneniya i Protsessy Upravleniya. Differential Equations and Control Processes",
  ["differ-algebr equ forum"] = "Differential-Algebraic Equations Forum",
  ["differential equations"] = "Differential Equations",
  ["differential equations dynam systems"] = "Differential Equations and Dynamical Systems",
  ["differential geom appl"] = "Differential Geometry and its Applications",
  ["differential integral equations"] = "Differential and Integral Equations. An International Journal for Theory & Applications",
  ["differentiation (malden, ma, u s)"] = "Differentiation (Malden, MA, United States)",
  ["differentiation (oxford, u k)"] = "Differentiation (Oxford, United Kingdom)",
  ["differentsial\\cprime naya geom mnogoobraz figur"] = "Differentsial\\cprime naya Geometriya Mnogoobraziy Figur",
  ["differentsialnaya geom mnogoobraz figur"] = "Differentsialnaya Geometriya Mnogoobrazii Figur",
  ["diffus defect data, pt a"] = "Diffusion and Defect Data–Solid State Data, Pt. A: Defect and Diffusion Forum",
  ["diffus defect data, pt b"] = "Diffusion and Defect Data–Solid State Data, Pt. B: Solid State Phenomena",
  ["dig antibiot"] = "Digesta antibiotica",
  ["dig dis"] = "Digestive diseases (Basel, Switzerland)",
  ["dig dis interv"] = "Digestive disease interventions",
  ["dig dis sci"] = "Digestive diseases and sciences",
  ["dig endosc"] = "Digestive endoscopy : official journal of the Japan Gastroenterological Endoscopy Society",
  ["dig j nanomater biostruct"] = "Digest Journal of Nanomaterials and Biostructures",
  ["dig liver dis"] = "Digestive and liver disease : official journal of the Italian Society of Gastroenterology and the Italian Association for the Study of the Liver",
  ["dig med res"] = "Digestive medicine research",
  ["dig neurol psychiatr"] = "Digest of neurology and psychiatry",
  ["dig surg"] = "Digestive surgery",
  ["dig tech pap"] = "Digest of technical papers. SID International Symposium",
  ["dig tech pap ieee int solid state circuits conf"] = "Digest of technical papers. IEEE International Solid-State Circuits Conference",
  ["dig tech papers"] = "Digest of technical papers. International Conference on Solid-State Sensors, Actuators, and Microsystems",
  ["digest (wash d c)"] = "The digest",
  ["digest public gen bills"] = "Digest of public general bills and selected resolutions, with index",
  ["digit biomark"] = "Digital biomarkers",
  ["digit cult educ"] = "Digital culture & education",
  ["digit finance"] = "Digital finance",
  ["digit health"] = "Digital health",
  ["digit investig"] = "Digital investigation",
  ["digit j ophthalmol"] = "Digital journal of ophthalmology : DJO",
  ["digit journal (abingdon)"] = "Digital journalism (Abingdon, England)",
  ["digit libr perspect"] = "Digital library perspectives",
  ["digit libraries cult herit knowl dissem future creat (2011)"] = "Digital libraries for cultural heritage, knowledge dissemination, and future creation : 13th International Conference on Asia-Pacific Digital Libraries, ICADL 2011, Beijing, China, October 24-27, 2011. International Conference on Asia-Pacific Digital Libraries (13th : 2011 : Beijing, China)",
  ["digit med"] = "Digital medicine",
  ["digit pathol (2019)"] = "Digital pathology : 15th European Congress, ECDP 2019, Warwick, UK, April 10-13, 2019, Proceedings. European Congress on Digital Pathology (15th : 2019 : Warwick, United Kingdom)",
  ["digit signal process"] = "Digital signal processing",
  ["digital appl archaeol cult heritage"] = "Digital Applications in Archaeology and Cultural Heritage",
  ["digital biomarkers"] = "Digital Biomarkers",
  ["digital chem eng"] = "Digital Chemical Engineering",
  ["digital commun networks"] = "Digital Communications and Networks",
  ["digital discovery"] = "Digital Discovery",
  ["digital eng digital twin"] = "Digital Engineering and Digital Twin",
  ["digital exper math educ"] = "Digital Experiences in Mathematics Education",
  ["digital geogr soc"] = "Digital Geography and Society",
  ["digital gov: res pract"] = "Digital Government: Research and Practice",
  ["digital health"] = "DIGITAL HEALTH",
  ["digital invest"] = "Digital Investigation",
  ["digital signal process"] = "Digital Signal Processing",
  ["digital threats: res pract"] = "Digital Threats: Research and Practice",
  ["digital twins appl"] = "Digital Twins and Applications",
  ["digital war"] = "Digital War",
  ["digitalbiomarkers 17 (2017)"] = "DigitalBiomarkers'17 : proceedings of the 1st Workshop on Digital Biomarkers : June 23, 2017, Niagara Falls, NY, USA. Workshop on Digital Biomarkers (1st : 2017 : Niagara Falls, N.Y.)",
  ["digitale bilddiagn"] = "Digitale Bilddiagnostik",
  ["dignitas (mex)"] = "Dignitas",
  ["dimacs ser discrete math theoret comput sci"] = "DIMACS Series in Discrete Mathematics and Theoretical Computer Science",
  ["dimens crit care nurs"] = "Dimensions of critical care nursing : DCCN",
  ["dimens dent hyg"] = "Dimensions of dental hygiene",
  ["dimens health serv"] = "Dimensions in health service",
  ["dimens oncol nurs"] = "Dimensions in oncology nursing : journal of the Division of Nursing",
  ["dimensions (n y n y)"] = "Dimensions (New York, N.Y.)",
  ["dimensions (wash)"] = "Dimensions/NBS",
  ["dinamika i prochnost\\cprime mashin"] = "Dinamika i Prochnost\\cprime Mashin",
  ["dinamika sistem"] = "Nizhegorodskiy Gosudavstvennyy Universitet",
  ["dinamika sploshn sredy"] = "Rossiiskaya Akademiya Nauk. Sibirskoe Otdelenie. Institut Gidrodinamiki. Dinamika Sploshnoi Sredy",
  ["dip metodi modelli mat sci appl"] = "Dipartimento di Metodi e Modelli Matematici per le Scienze Applicate",
  ["dipend patologiche"] = "Dipendenze patologiche",
  ["dipl statecraft"] = "Diplomacy and statecraft",
  ["diqiu huaxue"] = "Diqiu Huaxue",
  ["dir arch manuscr repos"] = "Directory Of Archives And Manuscript Repositories",
  ["dir boards"] = "Directors & boards",
  ["dir chaos"] = "Directions in Chaos",
  ["dir lav"] = "Il Diritto del lavoro",
  ["dir psychiatry"] = "Directions in psychiatry",
  ["dirasat b pure appl sci"] = "Dirasat. Series B, Pure and applied sciences",
  ["dirasat med biol sci"] = "Dirasat. Medical and biological sciences",
  ["dirasat nat eng sci"] = "Dirasat",
  ["dirasat sukkaniyah"] = "Dirāsāt sukkānīyah",
  ["dis (des interact syst conf)"] = "DIS. Designing Interactive Systems (Conference)",
  ["dis aquat org"] = "Diseases of Aquatic Organisms",
  ["dis aquat organ"] = "Diseases of aquatic organisms",
  ["dis asian aquac"] = "Diseases in Asian aquaculture",
  ["dis chest"] = "Diseases of the chest",
  ["dis colon rectum"] = "Diseases of the colon and rectum",
  ["dis esophagus"] = "Diseases of the esophagus : official journal of the International Society for Diseases of the Esophagus",
  ["dis manag"] = "Disease management : DM",
  ["dis manag advis"] = "Disease management advisor",
  ["dis manag clin outcomes"] = "Disease management and clinical outcomes",
  ["dis markers"] = "Disease markers",
  ["dis model mech"] = "Disease models & mechanisms",
  ["dis models mech"] = "Disease Models & Mechanisms",
  ["dis mon"] = "Disease-a-month : DM",
  ["dis nerv syst"] = "Diseases of the nervous system",
  ["disabil handicap soc"] = "Disability, handicap & society",
  ["disabil health j"] = "Disability and health journal",
  ["disabil rehabil"] = "Disability and rehabilitation",
  ["disabil rehabil assist technol"] = "Disability and rehabilitation. Assistive technology",
  ["disabil soc"] = "Disability & society",
  ["disabil stud q"] = "Disability studies quarterly",
  ["disaster health"] = "Disaster health",
  ["disaster manag response"] = "Disaster management & response : DMR : an official publication of the Emergency Nurses Association",
  ["disaster med public health prep"] = "Disaster medicine and public health preparedness",
  ["disaster mil med"] = "Disaster and military medicine",
  ["disaster prev manag"] = "Disaster prevention and management",
  ["disaster prev manage"] = "Disaster Prevention and Management",
  ["disch plann update"] = "Discharge planning update",
  ["discourse process"] = "Discourse processes",
  ["discourse stud"] = "Discourse studies",
  ["discourses math appl"] = "Discourses in Mathematics and its Applications",
  ["discov innov"] = "Discovery and innovation",
  ["discov mater"] = "Discover materials",
  ["discov med"] = "Discovery medicine",
  ["discov subtleties sugars (2013)"] = "Discovering the subtleties of sugars : proceedings of the 3rd Beilstein Glyco-Bioinformatics Symposium : June 10th - 14th, 2013, Potsdam, Germany. International Beilstein Symposium on Glyco-Bioinformatics (3rd : 2013 : Potsdam, Germany)",
  ["discover artif intell"] = "Discover Artificial Intelligence",
  ["discover chem eng"] = "Discover Chemical Engineering",
  ["discover energy"] = "Discover Energy",
  ["discover food"] = "Discover Food",
  ["discover internet things"] = "Discover Internet of Things",
  ["discover mater"] = "Discover Materials",
  ["discover mech eng"] = "Discover Mechanical Engineering",
  ["discover sustainability"] = "Discover Sustainability",
  ["discover water"] = "Discover Water",
  ["discoveries (craiova)"] = "Discoveries (Craiova, Romania)",
  ["discovery immunother"] = "Discovery Immunotherapy",
  ["discret event dyn syst"] = "Discrete event dynamic systems",
  ["discret math theor comput sci"] = "Discrete mathematics and theoretical computer science",
  ["discrete anal"] = "Discrete Analysis",
  ["discrete appl math"] = "Discrete Applied Mathematics",
  ["discrete appl math (1979)"] = "Discrete applied mathematics",
  ["discrete comput geom"] = "Discrete & Computational Geometry",
  ["discrete contin dyn syst"] = "Discrete and Continuous Dynamical Systems",
  ["discrete contin dyn syst - ser b"] = "Discrete and Continuous Dynamical Systems Series B",
  ["discrete contin dyn syst - ser s"] = "Discrete and Continuous Dynamical Systems - Series S",
  ["discrete contin dyn syst ser a"] = "Discrete and continuous dynamical systems. Series A",
  ["discrete contin dyn syst ser b"] = "Discrete and Continuous Dynamical Systems. Series B. A Journal Bridging Mathematics and Sciences",
  ["discrete contin dyn syst ser s"] = "Discrete and Continuous Dynamical Systems. Series S",
  ["discrete contin dynam systems"] = "Discrete and Continuous Dynamical Systems",
  ["discrete continuous dyn syst ser b"] = "Discrete and continuous dynamical systems. Series B",
  ["discrete dyn nat soc"] = "Discrete dynamics in nature and society",
  ["discrete event dyn syst"] = "Discrete Event Dynamic Systems",
  ["discrete event dyn syst theory appl"] = "Discrete Event Dynamic Systems Theory and Applications",
  ["discrete math"] = "Discrete Mathematics",
  ["discrete math algorithms appl"] = "Discrete mathematics, algorithms, and applications",
  ["discrete math appl"] = "Discrete Mathematics and Applications",
  ["discrete math appl (boca raton)"] = "Discrete Mathematics and its Applications (Boca Raton)",
  ["discrete math lett"] = "Discrete Mathematics Letters",
  ["discrete math theor comput sci"] = "Discrete Mathematics and Theoretical Computer Science",
  ["discrete math theor comput sci proc"] = "Discrete Mathematics & Theoretical Computer Science Proceedings",
  ["discrete math, algorithms appl"] = "Discrete Mathematics, Algorithms and Applications",
  ["discrete optim"] = "Discrete Optimization",
  ["discuss faraday soc"] = "Discussions of the Faraday Society",
  ["discuss math algebra stochastic methods"] = "Discussiones Mathematicae",
  ["discuss math differ incl control optim"] = "Discussiones Mathematicae. Differential Inclusions, Control and Optimization",
  ["discuss math differential incl"] = "Discussiones Mathematicae",
  ["discuss math gen algebra appl"] = "Discussiones Mathematicae. General Algebra and Applications",
  ["discuss math graph theory"] = "Discussiones Mathematicae. Graph Theory",
  ["discuss math probab stat"] = "Discussiones Mathematicae. Probability and Statistics",
  ["discuss pap univ tor dep geogr"] = "Discussion paper (University of Toronto. Dept. of Geography)",
  ["diskret anal issled oper"] = "Diskretnyy Analiz i Issledovanie Operatsiy",
  ["diskret mat"] = "Diskretnaya Matematika",
  ["diskretn anal issled oper"] = "Diskretnyi Analiz i Issledovanie Operatsii",
  ["diskretn anal issled oper ser 1"] = "Diskretnyy Analiz i Issledovanie Operatsiy",
  ["diskretn anal issled oper ser 2"] = "Diskretnyy Analiz i Issledovanie Operatsiy",
  ["diskussionsforum med ethik"] = "Diskussionsforum medizinische Ethik",
  ["disrupt sci technol"] = "Disruptive science and technology",
  ["diss abstr"] = "Dissertation abstracts",
  ["diss abstr int"] = "Dissertation abstracts international",
  ["diss bot"] = "Dissertationes botanicae",
  ["diss eth"] = "Dissertation ETH",
  ["diss math univ tartu"] = "Dissertationes Mathematicae Universitatis Tartuensis",
  ["diss summ math"] = "Dissertation Summaries in Mathematics",
  ["dissertationes math"] = "Dissertationes Mathematicae",
  ["dissertationes math (rozprawy mat)"] = "Dissertationes Mathematicae (Rozprawy Matematyczne)",
  ["dissolut technol"] = "Dissolution technologies",
  ["dist nurs"] = "District nursing",
  ["disting diss comput sci"] = "Distinguished Dissertations in Computer Science",
  ["distrib comput"] = "Distributed Computing",
  ["distrib gener altern energy j"] = "Distributed Generation & Alternative Energy Journal",
  ["distrib ledger technol: res pract"] = "Distributed Ledger Technologies: Research and Practice",
  ["distrib parallel databases"] = "Distributed and parallel databases",
  ["distrib sys eng"] = "Distributed Systems Engineering",
  ["distrib syst eng"] = "Distributed Systems Engineering",
  ["divers distrib"] = "Diversity & distributions",
  ["divers equal health care"] = "Diversity and equality in health and care",
  ["diversity (basel)"] = "Diversity",
  ["diving hyperb med"] = "Diving and hyperbaric medicine",
  ["divulg cult odontol"] = "Divulgacion cultural odontológica",
  ["divulg mat"] = "Revista Matematica de la Universidad del Zulia. Divulgaciones Matematicas",
  ["divulg med dent"] = "Divulgación médico-dental; boletín de educación médico-dental y especialidades",
  ["dix huit siecle"] = "Dix-huitième siècle",
  ["dlib mag"] = "D-Lib magazine : the magazine of the Digital Library Forum",
  ["dlr nachr"] = "DLR-Nachrichten : Mitteilungsblatt der Deutschen Forschungsanstalt für Luft- und Raumfahrt",
  ["dly mail guard"] = "Daily Mail and Guardian",
  ["dly rep (u s foreign broadcast inf serv)"] = "Daily report (United States. Foreign Broadcast Information Service)",
  ["dly rep china"] = "Daily report. China",
  ["dly rep peoples repub china"] = "Daily report. People's Republic of China",
  ["dly stat can"] = "Daily (Statistics Canada)",
  ["dly times"] = "Daily times (Lagos, Nigeria)",
  ["dly wash law report"] = "The Daily Washington law reporter",
  ["dmv sem"] = "DMV Seminar",
  ["dna barcodes (berlin)"] = "DNA barcodes (Berlin)",
  ["dna cell biol"] = "DNA and cell biology",
  ["dna repair"] = "DNA Repair",
  ["dna repair (amst)"] = "DNA repair",
  ["dna res"] = "DNA research : an international journal for rapid publication of reports on genes and genomes",
  ["dna seq"] = "DNA sequence : the journal of DNA sequencing and mapping",
  ["dobutsu iden ikushu kenkyu"] = "Dōbutsu iden ikushu kenkyū = Journal of animal genetics",
  ["doc anal geogr"] = "Documents d'anàlisi geogràfica",
  ["doc eur cent"] = "Documentation sur l'Europe centrale",
  ["doc faun helv"] = "Documenta Faunistica Helvetiae",
  ["doc geigy acta psychosom (dtsch ausg)"] = "Documenta Geigy. Acta psychosomatica. [Deutsche Ausg]",
  ["doc geigy acta rheumatol"] = "Documenta Geigy. Acta rheumatologica. [English ed.]",
  ["doc geigy ser chir"] = "Documenta Geigy. Series chirurgica",
  ["doc hist vocab sci"] = "Documents pour l'histoire du vocabulaire scientifique",
  ["doc ist ital idrobiol dott marco de marchi"] = "Documenta dell'Istituto italiano di idrobiologia. Istituto italiano di idrobiologia (Pallanza, Italy)",
  ["doc math"] = "Documenta Mathematica",
  ["doc math (paris)"] = "Documents Mathematiques (Paris)",
  ["doc med ethics"] = "Documentation in medical ethics",
  ["doc med geogr trop"] = "Documenta de medicina geographica et tropica",
  ["doc mycol"] = "Documents mycologiques",
  ["doc neerl indones morbis trop"] = "Documenta Neerlandica et Indonesica de morbis tropicis; quarterly journal of tropical medicine and hygiene",
  ["doc ophthalmol"] = "Documenta ophthalmologica. Advances in ophthalmology",
  ["doc ophthalmol proc ser"] = "Documenta ophthalmologica. Proceedings series",
  ["doc phytosociol, nouv sér"] = "Documents phytosociologiques, nouvelle série",
  ["doc rheumatol"] = "Documenta rheumatologica. [Deutsche Ausg.]",
  ["doc rheumatol (1956)"] = "Documenta rheumatologica. [English ed.]",
  ["doctor thesis"] = "Doctoral Thesis",
  ["dok bodenkd ges schweiz"] = "Dokument Bodenkundliche Gesellschaft der Schweiz",
  ["dok hoechst arch"] = "Dokumente aus Hoechster Archiven",
  ["dok inf schweiz orts- reg- landesplan, disp"] = "Dokumente und Informationen zur schweizerischen Orts-, Regional- und Landesplanung, DISP",
  ["dokl akad nauk"] = "Doklady Akademii Nauk",
  ["dokl akad nauk armen"] = "Doklady",
  ["dokl akad nauk az ssr"] = "Doklady - Akademiya Nauk Azerbaidzhana",
  ["dokl akad nauk azerbay dzhana"] = "Akademiya Nauk Azerbay dzhana",
  ["dokl akad nauk belarusi"] = "Doklady Akademii Nauk Belarusi",
  ["dokl akad nauk respub tadzhikistan"] = "Akademiya Nauk Respubliki Tadzhikistan",
  ["dokl akad nauk soiuza sov sotsialisticheskikh resp"] = "Doklady Akademii nauk Soiuza Sovetskikh Sotsialisticheskikh Respublik",
  ["dokl akad nauk sssr"] = "Doklady Akademii nauk SSSR",
  ["dokl akad nauk sssr [sov phys dokl]"] = "Doklady Akademii Nauk SSSR [Soviet Physics–Doklady]",
  ["dokl akad nauk tadzh ssr"] = "Doklady Akademii Nauk Tadzhikskoi SSR",
  ["dokl akad nauk ukr ssr"] = "Doklady Akademii Nauk Ukrainy",
  ["dokl akad nauk uzssr"] = "Doklady Akademii Nauk UzSSR",
  ["dokl biochem"] = "Doklady biochemistry : proceedings of the Academy of Sciences of the USSR, Biochemistry section",
  ["dokl biochem biophys"] = "Doklady. Biochemistry and biophysics",
  ["dokl biol sci"] = "Doklady biological sciences : proceedings of the Academy of Sciences of the USSR, Biological sciences sections",
  ["dokl biophys"] = "Doklady Biophysics",
  ["dokl bolg akad nauk"] = "Doklady Bolgarskoi Akademii Nauk",
  ["dokl bot sci"] = "Doklady Botanical Sciences",
  ["dokl bulg acad nauk"] = "Doklady Bolgarskoi akademii nauk",
  ["dokl bulg akad nauk"] = "Dokladi na Bulgarskata Akademiya na Naukite",
  ["dokl chem"] = "Doklady Chemistry",
  ["dokl chem technol"] = "Doklady Chemical Technology",
  ["dokl earth sci"] = "Doklady Earth Sciences",
  ["dokl math"] = "Doklady Mathematics",
  ["dokl nats akad nauk armen"] = "Doklady. Natsionalnaya Akademiya Nauk Armenii. Reports. National Academy of Sciences of Armenia",
  ["dokl nats akad nauk azerb"] = "Doklady. Natsionalnaya Akademiya Nauk Azerbaidzhana. Reports. National Academy of Sciences of Azerbaijan",
  ["dokl nats akad nauk belarusi"] = "Doklady Natsionalnoi Akademii Nauk Belarusi",
  ["dokl natsionalnoi akad nauk belarusi"] = "Doklady Natsionalʹnoĭ akademii nauk Belarusi",
  ["dokl phys"] = "Doklady Physics",
  ["dokl phys chem"] = "Doklady Physical Chemistry",
  ["dokl ross akademii selskokhoziaistvennykh nauk"] = "Doklady Rossiĭskoĭ akademii selʹskokhozi︠a︡ĭstvennykh nauk",
  ["dokl semin inst prikl mat im i n vekua"] = "Doklady",
  ["dokumente gesch math"] = "Dokumente zur Geschichte der Mathematik",
  ["dolciani math exp"] = "The Dolciani Mathematical Expositions",
  ["dolentium hominum"] = "Dolentium Hominum",
  ["dolomites res notes approx"] = "Dolomites Research Notes on Approximation",
  ["dom law rep"] = "Dominion law reports",
  ["domain adapt represent transf med image learn less labels imperfect data (2019)"] = "Domain Adaptation and Representation Transfer and Medical Image Learning with Less Labels and Imperfect Data : first MICCAI Workshop, DART 2019, and first International Workshop, MIL3ID 2019, Shenzhen, Held in Conjunction with MICCAI 2019, Shenzhen, China, October 13 and 17, 2019, Proceedings. DART (Workshop) (1st : 2019 : Shenzhen Shi, China)",
  ["domain decompos method sci eng xx (2011)"] = "Domain decomposition methods in science and engineering XX. International Conference on Domain Decomposition Methods for Partial Differential Equations (20th : 2011 : San Diego, Calif.)",
  ["domest anim endocrinol"] = "Domestic animal endocrinology",
  ["domus med"] = "Domus medici",
  ["dong bei nong ye da xue xue bao"] = "Dong bei nong ye da xue xue bao : Journal of Northeast Agricultural University",
  ["dong wu fen lei xue bao"] = "Dong wu fen lei xue bao = Acta zootaxonomica Sinica",
  ["dong wu xue bao"] = "Dong wu xue bao. [Acta zoologica Sinica]",
  ["dong wu ying yang xue bao"] = "Dong wu ying yang xue bao = Acta Zoonutrimenta Sinica",
  ["dongbei lin ye da xue xue bao"] = "Dongbei lin ye da xue xue bao = Journal of North-East Forestry University",
  ["dongbei shida xuebao"] = "Dongbei Shida Xuebao",
  ["dongmul jawon yeongu"] = "Dongmul jawon yeon-gu",
  ["dongwu yixue jinzhan"] = "Dongwu yixue jinzhan",
  ["dongwuxue yanjiu"] = "Dong wu xue yan jiu = Zoological research",
  ["dongwuxue zazhi"] = "Tung wu hsüeh tsa chih : Dongwuxue zazhi",
  ["dongwuyuan xuebao"] = "Dòngwùyuán xuébào",
  ["dopov akad nauk ukr rsr b"] = "Dopovidi Akademiī nauk Ukraīns'koī RSR. Seriia B: Heolohichni, khimichni ta biolohichni nauky",
  ["dopov nats akad nauk ukr"] = "Dopovidi Natsional’noi Akademii Nauk Ukraini",
  ["dopov nats akad nauk ukr mat prirodozn tekh nauki"] = "Dopovidi Natsionalnoi Akademii Nauk Ukraini. Matematika. Prirodoznavstvo. Tekhnichni Nauki",
  ["dopov natsionalnoi akad nauk ukr matemat pryrodoznavstvo tekhnichni nauky"] = "Dopovidi Nat︠s︡ionalʹnoï akademiï nauk Ukraïny. Matematyka, pryrodoznavstvo, tekhnichni nauky",
  ["doriana (genova)"] = "Doriana",
  ["dose response"] = "Dose-response : a publication of International Hormesis Society",
  ["douleurs: eval diagn trait"] = "Douleurs : Evaluation - Diagnostic - Traitement",
  ["down earth"] = "Down to earth",
  ["downs syndr res pract"] = "Down's syndrome, research and practice : the journal of the Sarah Duffen Centre",
  ["dp rep"] = "DP [reports]. U.S. Atomic Energy Commission",
  ["dr interact newsl"] = "Drug interactions newsletter",
  ["dr med penny mag"] = "The Doctor; a medical penny magazine",
  ["draco med"] = "Draco pro medico",
  ["drake law rev"] = "Drake law review",
  ["draper fund rep"] = "Draper Fund report",
  ["draper world popul fund rep"] = "Draper World Population Fund report",
  ["drg monit"] = "DRG monitor",
  ["driem tijdschr gemeentekrediet belg"] = "Driemaandelijks tijdschrift - Gemeentekrediet van Belgie",
  ["drink water eng sci"] = "Drinking water engineering and science",
  ["drinking water eng sci"] = "Drinking Water Engineering and Science",
  ["drinking water eng sci discuss"] = "Drinking Water Engineering and Science Discussions",
  ["dritte welt"] = "Die Dritte Welt",
  ["droit soc"] = "Droit social",
  ["drone syst appl"] = "Drone Systems and Applications",
  ["drosoph inf serv"] = "Drosophila information service",
  ["drug alcohol depend"] = "Drug and alcohol dependence",
  ["drug alcohol depend rep"] = "Drug and Alcohol Dependence Reports",
  ["drug alcohol rev"] = "Drug and alcohol review",
  ["drug chem toxicol"] = "Drug and chemical toxicology",
  ["drug chem toxicol (1977)"] = "Drug and Chemical Toxicology (1977)",
  ["drug cosmet ind"] = "Drug and cosmetic industry",
  ["drug court rev"] = "Drug court review",
  ["drug deliv"] = "Drug delivery",
  ["drug deliv (lond)"] = "Drug delivery (London, England. 2007)",
  ["drug deliv lett"] = "Drug delivery letters",
  ["drug deliv transl res"] = "Drug delivery and translational research",
  ["drug delivery"] = "Drug Delivery",
  ["drug delivery transl res"] = "Drug Delivery and Translation Research",
  ["drug des"] = "Drug designing : open access",
  ["drug des deliv"] = "Drug design and delivery",
  ["drug des dev ther"] = "Drug Design Development and Therapy",
  ["drug des devel ther"] = "Drug design, development and therapy",
  ["drug des discov"] = "Drug design and discovery",
  ["drug des discovery"] = "Drug Design and Discovery",
  ["drug dev ind pharm"] = "Drug development and industrial pharmacy",
  ["drug dev res"] = "Drug development research",
  ["drug discov ther"] = "Drug discoveries & therapeutics",
  ["drug discov today"] = "Drug discovery today",
  ["drug discov today biosilico"] = "Drug discovery today. Biosilico",
  ["drug discov today dis mech"] = "Drug discovery today. Disease mechanisms",
  ["drug discov today dis models"] = "Drug discovery today. Disease models",
  ["drug discov today technol"] = "Drug discovery today. Technologies",
  ["drug discov today ther strateg"] = "Drug discovery today. Therapeutic strategies",
  ["drug discovery dev"] = "Drug Discovery and Development",
  ["drug discovery today"] = "Drug Discovery Today",
  ["drug discovery today: technol"] = "Drug Discovery Today: Technologies",
  ["drug forum"] = "Drug forum",
  ["drug healthc patient saf"] = "Drug, healthcare and patient safety",
  ["drug inf j"] = "Drug information journal",
  ["drug intell clin pharm"] = "Drug intelligence & clinical pharmacy",
  ["drug investig"] = "Drug investigation",
  ["drug merch"] = "Drug merchandising",
  ["drug metab dispos"] = "Drug metabolism and disposition: the biological fate of chemicals",
  ["drug metab drug interact"] = "Drug Metabolism and Drug Interactions",
  ["drug metab lett"] = "Drug metabolism letters",
  ["drug metab pers ther"] = "Drug metabolism and personalized therapy",
  ["drug metab pharmacokinet"] = "Drug metabolism and pharmacokinetics",
  ["drug metab rev"] = "Drug metabolism reviews",
  ["drug metabol drug interact"] = "Drug metabolism and drug interactions",
  ["drug monit"] = "The drug monitor",
  ["drug news perspect"] = "Drug news & perspectives",
  ["drug nutr interact"] = "Drug-nutrient interactions",
  ["drug repurp rescue reposition"] = "Drug repurposing, rescue, and repositioning",
  ["drug repurposing rescue repositioning"] = "Drug Repurposing, Rescue, and Repositioning",
  ["drug res"] = "Drug research",
  ["drug res (stuttg)"] = "Drug research",
  ["drug res (stuttgart, ger)"] = "Drug Research (Stuttgart, Germany)",
  ["drug res rep"] = "Drug research reports",
  ["drug resist updat"] = "Drug resistance updates : reviews and commentaries in antimicrobial and anticancer chemotherapy",
  ["drug saf"] = "Drug safety",
  ["drug saf case rep"] = "Drug safety - case reports",
  ["drug sci policy law"] = "Drug science, policy and law",
  ["drug stand"] = "Drug standards",
  ["drug target insights"] = "Drug target insights",
  ["drug target rev"] = "Drug target review",
  ["drug test  anal"] = "Drug testing and analysis",
  ["drug test anal"] = "Drug Testing and Analysis",
  ["drug ther"] = "Drug therapeutics",
  ["drug ther (ny)"] = "Drug therapy",
  ["drug ther bull"] = "Drug and therapeutics bulletin",
  ["drug ther hosp ed"] = "Drug therapy. [Hospital edition]",
  ["drug topics"] = "Drug topics",
  ["drug trade news"] = "Drug trade news",
  ["drugs (abingdon engl)"] = "Drugs (Abingdon, England)",
  ["drugs aging"] = "Drugs & aging",
  ["drugs alcohol today"] = "Drugs and alcohol today",
  ["drugs context"] = "Drugs in context",
  ["drugs exp clin res"] = "Drugs under experimental and clinical research",
  ["drugs future"] = "Drugs of the future",
  ["drugs made ger"] = "Drugs made in Germany",
  ["drugs r d"] = "Drugs in R&D",
  ["drugs r&d"] = "Drugs in R&D",
  ["drugs real world outcome"] = "Drugs - Real World Outcome",
  ["drugs real world outcomes"] = "Drugs - real world outcomes",
  ["drugs soc (new york)"] = "Drugs & society (New York, N.Y.)",
  ["drugs ther perspect"] = "Drugs & therapy perspectives : for rational drug selection and use",
  ["drugs today (barc)"] = "Drugs of today (Barcelona, Spain : 1998)",
  ["drus istraz"] = "Drustvena Istrazivanja",
  ["drying technol"] = "Drying Technology",
  ["dtsch apoth"] = "Der Deutsche Apotheker",
  ["dtsch apoth ztg"] = "Deutsche Apotheker-Zeitung",
  ["dtsch arch klin med"] = "Deutsches Archiv für klinische Medizin",
  ["dtsch arztebl"] = "Deutsches Ärzteblatt",
  ["dtsch arztebl ausg a"] = "Deutsches Ärzteblatt (Ausg. A)",
  ["dtsch arztebl int"] = "Deutsches Ärzteblatt international",
  ["dtsch aussenpolit"] = "Deutsche Aussenpolitik",
  ["dtsch baumsch"] = "Deutsche Baumschule",
  ["dtsch dent z"] = "Deutsche dentistische Zeitschrift",
  ["dtsch drog ztg"] = "Deutsche Drogistenzeitung",
  ["dtsch entomol z"] = "Deutsche Entomologische Zeitschrift",
  ["dtsch gesundheit"] = "Deutscher Gesundheitskalender",
  ["dtsch gesundheitspolit"] = "Deutsche Gesundheitspolitik",
  ["dtsch gesundheitsw"] = "Das Deutsche Gesundheitswesen",
  ["dtsch gesundheitswes"] = "Deutsche Gesundheitswesen",
  ["dtsch gewässerkdl mitt"] = "Deutsche gewässerkundliche Mitteilungen",
  ["dtsch hydrogr z"] = "Deutsche Hydrographische Zeitschrift",
  ["dtsch jahrb volkskd"] = "Deutsches Jahrbuch für Volkskunde",
  ["dtsch krankenpflegez"] = "Deutsche Krankenpflegezeitschrift",
  ["dtsch lebensm- rundsch"] = "Deutsche Lebensmittel Rundschau",
  ["dtsch lebensmitt rundsch"] = "Deutsche Lebensmittel-Rundschau : Zeitschrift für Lebensmittelkunde und Lebensmittelrecht",
  ["dtsch med j"] = "Deutsches medizinisches Journal",
  ["dtsch med rundsch"] = "Deutsche medizinische Rundschau; Monatsschrift mit ärztlicher Akademie",
  ["dtsch med wochenschr"] = "Deutsche medizinische Wochenschrift (1946)",
  ["dtsch rentenversicher"] = "Deutsche Rentenversicherung",
  ["dtsch schiffahrtsarch"] = "Deutsches Schiffahrtsarchiv",
  ["dtsch schwesternztg"] = "Deutsche Schwesternzeitung",
  ["dtsch stomatol"] = "Deutsche Stomatologie",
  ["dtsch tierarztl wochenschr"] = "Deutsche tierärztliche Wochenschrift",
  ["dtsch vierteljahresschr litt wiss geistesgesch"] = "Deutsche Vierteljahrsschrift für Literaturwissenschaft und Geistesgeschichte",
  ["dtsch z gesamte gerichtl med"] = "Deutsche Zeitschrift fur die Gesamte Gerichtliche Medizin",
  ["dtsch z mund kiefer gesichtschir"] = "Deutsche Zeitschrift für Mund-, Kiefer- und Gesichts-Chirurgie",
  ["dtsch z nervenheilkd"] = "Deutsche Zeitschrift fur Nervenheilkunde",
  ["dtsch z philos"] = "Deutsche Zeitschrift für Philosophie",
  ["dtsch z sportmed"] = "Deutsche Zeitschrift fur Sportmedizin",
  ["dtsch z verdau stoffwechselkr"] = "Deutsche Zeitschrift fur Verdauungs- und Stoffwechselkrankheiten",
  ["dtsch z zahnarztl hypn"] = "Deutsche Zeitschrift für zahnärztliche Hypnose",
  ["dtsch zahn mund kieferheilkd zentralbl"] = "Deutsche Zahn-, Mund-, und Kieferheilkunde mit Zentralblatt",
  ["dtsch zahn mund kieferheilkd zentralbl gesamte"] = "Deutsche Zahn-, Mund-, und Kieferheilkunde mit Zentralblatt für die gesamte Zahn-, Mund-, und Kieferheilkunde",
  ["dtsch zahnarztekal"] = "Deutscher Zahnärztekalender",
  ["dtsch zahnarztl z"] = "Deutsche zahnärztliche Zeitschrift",
  ["dtsch zentralbl krankenpfl"] = "Deutsches Zentralblatt fur Krankenpflege",
  ["dtschl arch"] = "Deutschland Archiv",
  ["dtw dtsch tierarztl wochenschr"] = "DTW: Deutsche Tierarztliche Wochenschrift",
  ["du bois rev"] = "Du Bois review : social science research on race",
  ["dual diagn (foster city)"] = "Dual diagnosis (Foster City)",
  ["dubl univ law j"] = "Dublin University law journal",
  ["dublin hist rec"] = "Dublin historical record",
  ["dublin j med chem sci"] = "Dublin Journal of Medical and Chemical Science",
  ["dublin j med sci"] = "Dublin Journal of Medical Science",
  ["duke j gend law policy"] = "Duke journal of gender law & policy",
  ["duke law j"] = "Duke law journal",
  ["duke law technol rev"] = "Duke law and technology review",
  ["duke math j"] = "Duke Mathematical Journal",
  ["duke med health news"] = "DukeMedicine healthnews",
  ["dunel notes"] = "Duneland notes",
  ["duodecim suppl"] = "Duodecim; lääketieteellinen aikakauskirja. Supplementum",
  ["duquesne law rev"] = "Duquesne law review",
  ["durham thomas harriot sem"] = "The Durham Thomas Harriot Seminar. Occasional Paper",
  ["durham univ j"] = "The Durham University journal. University of Durham",
  ["dusseld arb gesch med"] = "Düsseldorfer Arbeiten zur Geschichte der Medizin",
  ["dusseld arb gesch med beih"] = "Düsseldorfer Arbeiten zur Geschichte der Medizin. Beiheft",
  ["dusseld geogr schr"] = "Düsseldorfer geographische Schriften",
  ["dusunen adam"] = "Düşünen adam : Bakırköy Ruh ve Sinir Hastalıkları Hastanesi yayın organı",
  ["dutch bird"] = "Dutch birding",
  ["duv math"] = "DUV Mathematik",
  ["duxbury adv ser"] = "Duxbury Advanced Series",
  ["duxbury adv ser statist decis sci"] = "The Duxbury Advanced Series in Statistics and Decision Sciences",
  ["dvs ber"] = "DVS Berichte",
  ["dvt—dějiny věd tech"] = "Dějiny Věd a Techniky",
  ["dyes pigm"] = "Dyes and Pigments",
  ["dyn atmos oceans"] = "Dynamics of Atmospheres and Oceans",
  ["dyn biochem process biotechnol mol biol"] = "Dynamic biochemistry, process biotechnology & molecular biology",
  ["dyn contin discrete impuls syst ser a math anal"] = "Dynamics of Continuous, Discrete & Impulsive Systems. Series A. Mathematical Analysis",
  ["dyn contin discrete impuls syst ser b appl algorithms"] = "Dynamics of Continuous, Discrete & Impulsive Systems. Series B. Applications & Algorithms",
  ["dyn control"] = "Dynamics and Control",
  ["dyn games appl"] = "Dynamic Games and Applications",
  ["dyn med"] = "Dynamic medicine : DM",
  ["dyn model econom econ finance"] = "Dynamic Modeling and Econometrics in Economics and Finance",
  ["dyn partial differ equ"] = "Dynamics of Partial Differential Equations",
  ["dyn partial differ equations"] = "Dynamics of Partial Differential Equations",
  ["dyn psychiatr"] = "Dynamische Psychiatrie. Dynamic psychiatry",
  ["dyn stat clim syst"] = "Dynamics and Statistics of the Climate System",
  ["dyn syst"] = "Dynamical Systems. An International Journal",
  ["dyn syst appl"] = "Dynamic Systems and Applications",
  ["dyn syst int j"] = "Dynamical Systems An International Journal",
  ["dyna (medellin)"] = "Dyna",
  ["dynam contin discrete impuls systems"] = "Dynamics of Continuous, Discrete and Impulsive Systems",
  ["dynam control"] = "Dynamics and Control",
  ["dynam econom theory and appl"] = "Dynamic Economics: Theory and Applications",
  ["dynam report expositions dynam systems (ns)"] = "Dynamics Reported: Expositions in Dynamical Systems (New Series)",
  ["dynam stability systems"] = "Dynamics and Stability of Systems",
  ["dynam systems appl"] = "Dynamic Systems and Applications",
  ["dyny ysrl"] = "Dine Yiśraʼel : shanaton le-mishpaṭ ʻIvri ule-dine mishpaḥah be-Yiśraʼel",
  ["dysmorphol clin genet"] = "Dysmorphology and clinical genetics : official publication of the Center for Birth Defects Information Services, Inc",
  ["e (norwalk)"] = "E (Norwalk, Conn.)",
  ["e j appl psychol"] = "E-journal of applied psychology : clinical and social issues",
  ["e law"] = "E law : Murdoch University electronic journal of law",
  ["e n e"] = "E N E",
  ["e spen eur e j clin nutr metab"] = "e-SPEN, the European e-journal of clinical nutrition and metabolism",
  ["e&g quat sci j"] = "E&G Quaternary Science Journal",
  ["e-j chem"] = "E-Journal of Chemistry",
  ["e-j reservoir eng"] = "E-Journal of Reservoir Engineering",
  ["e-j surf sci nanotechnol"] = "e-Journal of Surface Science and Nanotechnology",
  ["e-prime adv electr eng electron energy"] = "e-Prime Advances in Electrical Engineering, Electronics and Energy",
  ["eaa ser"] = "European Actuarial Academy (EAA) Series",
  ["eai endorsed trans pervasive health technol"] = "EAI endorsed transactions on pervasive health and technology",
  ["eai/springer innov commun comput"] = "EAI/Springer Innovations in Communication and Computing",
  ["ear hear"] = "Ear and hearing",
  ["ear nose throat j"] = "Ear, nose, & throat journal",
  ["early am lit"] = "Early American literature",
  ["early child dev care"] = "Early child development and care",
  ["early child educ j"] = "Early childhood education journal",
  ["early child res q"] = "Early childhood research quarterly",
  ["early child serv (san diego)"] = "Early childhood services (San Diego, Calif.)",
  ["early dev parent"] = "Early development & parenting",
  ["early educ dev"] = "Early education and development",
  ["early engl books (cross index)"] = "English books, 1475-1640; consolidated cross index by STC numbers",
  ["early hum dev"] = "Early human development",
  ["early interv psychiatry"] = "Early intervention in psychiatry",
  ["early medieval eur"] = "Early Medieval Europe",
  ["early pregnancy"] = "Early pregnancy (Online)",
  ["early sci med"] = "Early science and medicine",
  ["early sources aphasia relat disord"] = "Early sources in aphasia and related disorders",
  ["early years (stoke-on-trent)"] = "Early years : an international journal of research and development",
  ["earth (waukesha)"] = "Earth (Waukesha, Wis.)",
  ["earth environ sci trans"] = "Earth and Environmental Science Transactions",
  ["earth environ sci trans r soc edinb"] = "Earth and environmental science transactions of the Royal Society of Edinburgh",
  ["earth environ sci trans r soc edinburgh"] = "Earth and Environmental Science Transactions of the Royal Society of Edinburgh",
  ["earth interact"] = "Earth Interactions",
  ["earth miner sci"] = "Earth and mineral sciences",
  ["earth moon planets"] = "Earth, moon, and planets",
  ["earth negot bull"] = "Earth Negotiations Bulletin",
  ["earth planet phys"] = "Earth and Planetary Physics",
  ["earth planet sci lett"] = "Earth and Planetary Science Letters",
  ["earth planets space"] = "Earth, Planets and Space",
  ["earth sci front"] = "Earth Science Frontiers",
  ["earth sci hist"] = "Earth sciences history : journal of the History of the Earth Sciences Society",
  ["earth sci inf"] = "Earth Science Informatics",
  ["earth sci inform"] = "Earth science informatics",
  ["earth sci res j"] = "Earth Sciences Research Journal",
  ["earth sci rev"] = "Earth-Science Reviews",
  ["earth sci syst soc"] = "Earth Science, Systems and Society",
  ["earth space"] = "Earth in space",
  ["earth space rev"] = "Earth space review",
  ["earth space sci"] = "Earth and Space Science",
  ["earth surf dyn"] = "Earth Surface Dynamics",
  ["earth surf dyn discuss"] = "Earth Surface Dynamics Discussions",
  ["earth surf process landf"] = "Earth surface processes and landforms",
  ["earth surf processes"] = "Earth Surface Processes",
  ["earth surf processes landf"] = "Earth Surface Processes and Landforms",
  ["earth surf processes landforms"] = "Earth Surface Processes and Landforms",
  ["earth syst dyn"] = "Earth System Dynamics",
  ["earth syst dyn discuss"] = "Earth System Dynamics Discussions",
  ["earth syst environ"] = "Earth Systems and Environment",
  ["earth syst governance"] = "Earth System Governance",
  ["earth syst sci data"] = "Earth System Science Data",
  ["earth syst sci data discuss"] = "Earth System Science Data Discussions",
  ["earth times (n y)"] = "The Earth times",
  ["earthq eng struct dyn"] = "Earthquake engineering & structural dynamics",
  ["earthq spectra"] = "Earthquake spectra : the professional journal of the Earthquake Engineering Research Institute",
  ["earthquake eng eng vibr"] = "Earthquake Engineering and Engineering Vibration",
  ["earthquake eng resilience"] = "Earthquake Engineering and Resilience",
  ["earthquake eng struct dyn"] = "Earthquake Engineering and Structural Dynamics",
  ["earthquake res adv"] = "Earthquake Research Advances",
  ["earthquake sci"] = "Earthquake Science",
  ["earthquake spectra"] = "Earthquake Spectra",
  ["earthquakes struct"] = "Earthquakes and Structures",
  ["earths future"] = "Earth's future",
  ["earth’s future"] = "Earth’s Future",
  ["east afr"] = "East African (Online)",
  ["east afr agric for j"] = "East African Agricultural and Forestry Journal",
  ["east afr econ rev"] = "Eastern Africa economic review",
  ["east afr geogr rev"] = "The East African geographical review",
  ["east afr health res j"] = "The East African health research journal",
  ["east afr j appl health monitor eval"] = "East African journal of applied health monitoring and evaluation",
  ["east afr j public health"] = "East African journal of public health",
  ["east afr j rural dev"] = "Eastern Africa journal of rural development",
  ["east afr j sci technol innovation"] = "East African Journal of Science, Technology and Innovation",
  ["east afr med j"] = "East African Medical Journal",
  ["east afr soc sci res rev"] = "Eastern Africa social science research review",
  ["east afr stand"] = "East African standard (Nairobi, Kenya : Daily : 1995)",
  ["east afri agric for j"] = "East African agricultural and forestry journal",
  ["east african j rural devel"] = "East African Journal of Rural Development",
  ["east anthropol"] = "The Eastern anthropologist",
  ["east asia (piscataway)"] = "East Asia (Piscataway, N.J.)",
  ["east asian arch psychiatry"] = "East Asian archives of psychiatry : official journal of the Hong Kong College of Psychiatrists = Dong Ya jing shen ke xue zhi : Xianggang jing shen ke yi xue yuan qi kan",
  ["east asian hist (canberra)"] = "East Asian history",
  ["east asian j appl math"] = "East Asian Journal on Applied Mathematics",
  ["east asian j applied math"] = "East Asian journal on applied mathematics",
  ["east asian sci technol med"] = "East Asian science, technology, and medicine",
  ["east asian sci technol soc"] = "East Asian science, technology and society",
  ["east asian sci technol soc: int j"] = "East Asian Science, Technology and Society: An International Journal",
  ["east biol"] = "Eastern biologist",
  ["east cent afr j surg"] = "East and Central African journal of surgery",
  ["east cent eur"] = "East central Europe. L'Europe du centre-est",
  ["east china norm univ sci rep"] = "East China Normal University Scientific Reports",
  ["east econ j"] = "Eastern economic journal",
  ["east eur jew aff"] = "East European Jewish affairs",
  ["east eur polit soc"] = "Eastern European politics and societies : EEPS",
  ["east eur q"] = "East European quarterly",
  ["east europ econ"] = "Eastern European economics",
  ["east horiz"] = "Eastern horizon",
  ["east j approx"] = "East Journal on Approximations",
  ["east j med"] = "Eastern journal of medicine : EJM",
  ["east mediterr health j"] = "Eastern Mediterranean health journal = La revue de santé de la Méditerranée orientale = al-Majallah al-ṣiḥḥīyah li-sharq al-mutawassiṭ",
  ["east south afr geogr j"] = "Eastern and southern Africa geographical journal",
  ["east tenn soc publ"] = "The East Tennessee Historical Society's publications. East Tennessee Historical Society",
  ["east west perspect"] = "East-West perspectives : journal of the East-West Center",
  ["east-west j econ bus"] = "East-West Journal of Economics and Business",
  ["east-west j math"] = "East-West Journal of Mathematics",
  ["east-west j numer math"] = "East-West Journal of Numerical Mathematics",
  ["east-west ser econ"] = "East-West Series in Economics, Business, and the Environment",
  ["eastern africa econ rev"] = "Eastern Africa Economic Review",
  ["eastern econ j"] = "Eastern Economic Journal",
  ["eastern europ econ"] = "Eastern European Economics",
  ["eat behav"] = "Eating behaviors",
  ["eat disord"] = "Eating disorders",
  ["eat weight disord"] = "Eating and weight disorders : EWD",
  ["ebook abstr health technol assess int meet 5th 2008 montr can"] = "E-book of abstracts : V Annual Meeting HTAi 2008, July 6-9, 2008, Montréal, Canada. Health Technology Assessment International. Meeting (5th : 2008 : Montréal, Canada)",
  ["ebp briefs"] = "EBP briefs",
  ["ebri issue brief"] = "EBRI issue brief",
  ["ec anaesth"] = "EC anaesthesia",
  ["ec clin exp anat"] = "EC clinical and experimental anatomy",
  ["ec clin med case rep"] = "EC clinical and medical case reports",
  ["ec dent sci"] = "EC dental science",
  ["ec ete probab st-flour"] = "Ecole d’Ete de Probabilites de Saint-Flour",
  ["ec gastroenterol dig syst"] = "EC gastroenterology and digestive system",
  ["ec gynaecol"] = "EC gynaecology",
  ["ec microbiol"] = "EC microbiology",
  ["ec nutr"] = "EC nutrition",
  ["ec ophthalmol"] = "EC ophthalmology",
  ["ec orthop"] = "EC orthopaedics",
  ["ec pharmacol toxicol"] = "EC pharmacology and toxicology",
  ["ec psychol psychiatr"] = "EC psychology and psychiatry",
  ["ec pulmonol respir med"] = "EC pulmonology and respiratory medicine",
  ["echo med cevennes"] = "L'Echo médical des Cévennes",
  ["echo med nord"] = "L'echo médical du nord",
  ["echo res pract"] = "Echo research and practice",
  ["ecletica quim"] = "Ecletica Quimica",
  ["eclogae geol helv"] = "Eclogae Geologicae Helvetiae",
  ["eco mont"] = "Eco.mont",
  ["eco-effic ind sci"] = "Eco-Efficiency in Industry and Science",
  ["eco-environ health"] = "Eco-Environment & Health",
  ["ecohydrol hydrobiol"] = "Ecohydrology & Hydrobiology",
  ["ecol appl"] = "Ecological applications : a publication of the Ecological Society of America",
  ["ecol chem eng"] = "Ecological Chemistry and Engineering",
  ["ecol chem eng a"] = "Ecological Chemistry and Engineering A",
  ["ecol chem eng s"] = "Ecological Chemistry and Engineering S",
  ["ecol complexity"] = "Ecological Complexity",
  ["ecol conserv"] = "Ecological Conservation",
  ["ecol dis"] = "Ecology of disease",
  ["ecol econ"] = "Ecological economics : the journal of the International Society for Ecological Economics",
  ["ecol eng"] = "Ecological engineering",
  ["ecol eng: x"] = "Ecological Engineering: X",
  ["ecol entomol"] = "Ecological entomology",
  ["ecol evol"] = "Ecology and evolution",
  ["ecol food nutr"] = "Ecology of food and nutrition",
  ["ecol freshw fish"] = "Ecology of freshwater fish",
  ["ecol freshwater fish"] = "Ecology of Freshwater Fish",
  ["ecol genet genom"] = "Ecological genetics and genomics",
  ["ecol genet genomics"] = "Ecological Genetics and Genomics",
  ["ecol indic"] = "Ecological Indicators",
  ["ecol inf"] = "Ecological Informatics",
  ["ecol inform"] = "Ecological informatics",
  ["ecol law q"] = "Ecology law quarterly",
  ["ecol lett"] = "Ecology letters",
  ["ecol manage restor"] = "Ecological Management and Restoration",
  ["ecol mediterr"] = "Ecologia Mediterranea : revue d'écologie terrestre et limnique",
  ["ecol model"] = "Ecological Modelling",
  ["ecol modell"] = "Ecological modelling",
  ["ecol monogr"] = "Ecological monographs",
  ["ecol montenegrina"] = "Ecologica montenegrina",
  ["ecol process"] = "Ecological processes",
  ["ecol processes"] = "Ecological Processes",
  ["ecol psychol"] = "Ecological psychology : a publication of the International Society for Ecological Psychology",
  ["ecol res"] = "Ecological Research",
  ["ecol res monogr"] = "Ecological Research Monographs",
  ["ecol sci"] = "Ecologic science",
  ["ecol soc"] = "Ecology and society : a journal of integrative science for resilience and sustainability",
  ["ecol solutions evidence"] = "Ecological Solutions and Evidence",
  ["ecol stud"] = "Ecological Studies",
  ["ecolog econ"] = "Ecological Economics",
  ["econ aff"] = "Economics Affairs (Calcutta)",
  ["econ anal"] = "Economic Analysis",
  ["econ anal pol"] = "Economic Analysis and Policy, New Series",
  ["econ anal policy"] = "Economic analysis and policy",
  ["econ anal workers’ manage"] = "Economic Analysis and Workers’ Management",
  ["econ anthropol"] = "Economic anthropology (Hoboken, N.J.)",
  ["econ aplicada"] = "Economia Aplicada",
  ["econ appl"] = "Économie appliquée",
  ["econ bot"] = "Economic botany",
  ["econ bull asia pac"] = "Economic bulletin for Asia and the Pacific",
  ["econ bull eur"] = "Economic bulletin for Europe",
  ["econ bull ghana"] = "The Economic bulletin of Ghana",
  ["econ bull natl bank egypt"] = "Economic bulletin - National Bank of Egypt",
  ["econ bus bull"] = "Economic and Business Bulletin",
  ["econ bus educ"] = "Economics and Business Education",
  ["econ bus rev"] = "Economic and Business Review",
  ["econ chilena"] = "Economía Chilena",
  ["econ colomb"] = "Economía colombiana : revista de la Contraloría General de la República",
  ["econ comput econ cybern stud res"] = "Economic computation and economic cybernetics studies and research",
  ["econ computat econ cybern stud res"] = "Economic Computation and Economic Cybernetics Studies and Research",
  ["econ credito"] = "Economia e credito",
  ["econ desarro"] = "Economía y desarrollo",
  ["econ design"] = "Economic Design",
  ["econ dev cult change"] = "Economic development and cultural change",
  ["econ devel cult change"] = "Economic Development and Cultural Change",
  ["econ devel quart"] = "Economic Development Quarterly",
  ["econ dirit agroaliment"] = "Economia & diritto agroalimentare",
  ["econ disaster clim chang"] = "Economics of disasters and climate change",
  ["econ disasters clim change"] = "Economics of Disasters and Climate Change",
  ["econ educ rev"] = "Economics of education review",
  ["econ eye"] = "Economic eye",
  ["econ fam"] = "L'Économie familiale",
  ["econ finan modelling"] = "Economic and Financial Modelling",
  ["econ finan rev"] = "Economic and Financial Review",
  ["econ fonti energ ambiente"] = "Economia delle fonti di energia",
  ["econ forum"] = "Economic forum",
  ["econ geogr"] = "Economic geography",
  ["econ geol"] = "Economic geology and the bulletin of the Society of Economic Geologists",
  ["econ geol bull soc econ geol"] = "Economic Geology and the Bulletin of the Society of Economic Geologists",
  ["econ governance"] = "Economics of Governance",
  ["econ hist (sweden)"] = "Economy and history",
  ["econ hist rev"] = "The Economic history review",
  ["econ hum biol"] = "Economics and human biology",
  ["econ humanisme"] = "Economie et humanisme",
  ["econ ind"] = "Economia Industrial",
  ["econ ind democracy"] = "Economic and Industrial Democracy",
  ["econ inf commun entertain"] = "The Economics of Information, Communication and Entertainment",
  ["econ innovation new tech"] = "Economics of Innovation and New Technology",
  ["econ inq"] = "Economic inquiry",
  ["econ inquiry"] = "Economic Inquiry",
  ["econ int"] = "Economia Internazionale",
  ["econ issues"] = "Economic Issues",
  ["econ issues probl perspect"] = "Economic Issues, Problems and Perspectives",
  ["econ j"] = "Economic Journal",
  ["econ j (london)"] = "Economic journal (London, England)",
  ["econ j nepal"] = "The Economic journal of Nepal",
  ["econ j watch"] = "Econ journal watch",
  ["econ lab relat rev"] = "Economic and Labour Relations Review",
  ["econ lav"] = "Economia & lavoro",
  ["econ lavoro"] = "Economia e Lavoro",
  ["econ lett"] = "Economics letters",
  ["econ letters"] = "Economics Letters",
  ["econ med anim"] = "Economie et médecine animales",
  ["econ merid"] = "Economie méridionale",
  ["econ mexicana"] = "Economia Mexicana, Nueva Epoca",
  ["econ model"] = "Economic modelling",
  ["econ modelling"] = "Economic Modelling",
  ["econ notes"] = "Economic Notes",
  ["econ outlook usa"] = "Economic outlook USA",
  ["econ perspect"] = "Economic Perspectives",
  ["econ philos"] = "Economics and philosophy",
  ["econ plann"] = "Economics of planning",
  ["econ planning"] = "Economics of Planning",
  ["econ pol rev"] = "Economic Policy Review",
  ["econ pol: a europ forum"] = "Economic Policy: A European Forum",
  ["econ policy"] = "Economic policy",
  ["econ polit stud"] = "Economic and political studies",
  ["econ polit wkly"] = "Economic and political weekly",
  ["econ politica"] = "Economia Politica",
  ["econ politics"] = "Economics and Politics",
  ["econ previs"] = "Economie & prévision. Statistiques & études financières. Série économique",
  ["econ prospect int"] = "Economie prospective internationale",
  ["econ prévision"] = "Economie et Prévision",
  ["econ pubblica"] = "Economia pubblica : mensile di studi e d'informazione del Ciriec",
  ["econ rec"] = "The Economic record",
  ["econ res"] = "Economic Research",
  ["econ res j"] = "Economic research journal",
  ["econ reun"] = "L'Economie de la Réunion",
  ["econ rev"] = "Economic review (Federal Reserve Bank of Atlanta)",
  ["econ rev (dallas)"] = "Economic review (Federal Reserve Bank of Dallas)",
  ["econ rev (keizai kenkyu)"] = "Economic Review (Keizai Kenkyu)",
  ["econ rurale"] = "Economie Rurale",
  ["econ scelte pubbliche"] = "Economia delle Scelte Pubbliche",
  ["econ sci technol innov"] = "Economics of Science, Technology and Innovation",
  ["econ soc"] = "Economy and society",
  ["econ soc (paris)"] = "Économies et sociétés",
  ["econ soc aff"] = "Economic and social affairs",
  ["econ soc hist jaarb"] = "Economisch- en sociaal-historisch jaarboek",
  ["econ soc rev"] = "Economic and Social Review",
  ["econ soc rev (irel)"] = "The Economic and social review",
  ["econ soc stud natl inst econ soc res"] = "Economic and Social Studies. National Institute of Economic and Social Research",
  ["econ soc tijdschr"] = "Economisch en Sociaal Tijdschrift",
  ["econ social"] = "Economia e socialismo",
  ["econ sociol"] = "Economia e sociologia (Evora, Portugal)",
  ["econ sociol trab"] = "Economía y sociología del trabajo",
  ["econ stat"] = "Economie & statistique",
  ["econ stat ber"] = "Economisch-statistische Berichten",
  ["econ stor"] = "Economia e storia",
  ["econ stud"] = "Economic studies (Calcutta, India)",
  ["econ stud inequal soc exclusion well-being"] = "Economic Studies in Inequality, Social Exclusion and Well-Being",
  ["econ stud quart"] = "Economic Studies Quarterly",
  ["econ syst res"] = "Economic systems research : journal of the International Input-Output Association",
  ["econ systems"] = "Economic Systems",
  ["econ systems res"] = "Economic Systems Research",
  ["econ theory"] = "Econometric theory",
  ["econ theory bull"] = "Economic Theory Bulletin",
  ["econ transition"] = "Economics of Transition",
  ["econ trends"] = "Economic trends (New Delhi, India)",
  ["econ wkly"] = "The Economic weekly",
  ["econ: biann collec recent ger stud"] = "Economics: A Biannual Collection of Recent German Studies",
  ["econ: j econ assoc"] = "Economics: The Journal of the Economics Association",
  ["econom comput econom cybernet stud res"] = "Economic Computation and Economic Cybernetics Studies and Research",
  ["econom inf age theory pract meas"] = "Econometrics in the Information Age: Theory and Practice of Measurement",
  ["econom j"] = "The econometrics journal",
  ["econom lett"] = "Economics Letters",
  ["econom rev"] = "Econometric reviews",
  ["econom soc monogr"] = "Econometric Society Monographs",
  ["econom stat"] = "Econometrics and statistics",
  ["econom theory"] = "Economic Theory",
  ["econom theory econometrics math econom"] = "Economic Theory, Econometrics, and Mathematical Economics",
  ["econom tinbergen inst lect"] = "Econometric and Tinbergen Institutes Lectures",
  ["econometric rev"] = "Econometric Reviews",
  ["econometric theory"] = "Econometric Theory",
  ["econometrics j"] = "Econometrics Journal",
  ["economia (peru)"] = "Economia (Pontifical Catholic University of Peru)",
  ["economia (portugal)"] = "Economia (Portuguese Catholic University)",
  ["economia umana rassegna medica internazionale"] = "L'Economia umana; rassegna medica internazionale",
  ["economica (ns)"] = "Economica. New Series",
  ["economist (leiden)"] = "De Economist",
  ["economy hist"] = "Economy and History",
  ["economy society"] = "Economy and Society",
  ["económica (la plata)"] = "Económica (National University of La Plata)",
  ["ecosal plus"] = "EcoSal Plus",
  ["ecosyst health sustainability"] = "Ecosystem Health and Sustainability",
  ["ecosyst people"] = "Ecosystems and People",
  ["ecosyst people (abingdon)"] = "Ecosystems and people (Abingdon, England)",
  ["ecosyst serv"] = "Ecosystem services",
  ["ecotoxicol environ saf"] = "Ecotoxicology and environmental safety",
  ["ecrits paris"] = "Écrits de Paris: revue des questions actuelles",
  ["ecs adv"] = "ECS Advances",
  ["ecs electrochem lett"] = "ECS Electrochemistry Letters",
  ["ecs j solid state sci technol"] = "ECS Journal of Solid State Science and Technology",
  ["ecs sens plus"] = "ECS Sensors Plus",
  ["ecs solid state lett"] = "ECS Solid State Letters",
  ["ecs trans"] = "ECS Transactions",
  ["ecuad debate"] = "Ecuador debate",
  ["ecum rev"] = "The Ecumenical review",
  ["ed angew"] = "Edition Angewandte",
  ["ed belin"] = "Aux Editions Belin",
  ["ed gutenbergplatz leipz (eagle)"] = "Edition am Gutenbergplatz Leipzig (EAGLE)",
  ["ed manag"] = "ED management : the monthly update on emergency department management",
  ["ed media"] = "ED-MEDIA. ED-MEDIA (Conference)",
  ["ed naz opera mat francesco maurolico"] = "Edizione Nazionale dell’Opera Matematica di Francesco Maurolico",
  ["ed open sources"] = "Edition Open Sources",
  ["edinb adv textb linguist"] = "Edinburgh Advanced Textbooks in Linguistics",
  ["edinb dent hosp gaz"] = "Edinburgh Dental Hospital gazette",
  ["edinb j bot"] = "Edinburgh journal of botany",
  ["edinb med j"] = "Edinburgh medical journal",
  ["edinb med surg j"] = "Edinburgh medical and surgical journal",
  ["edinb mon j med sci"] = "The Edinburgh monthly journal of medical science",
  ["edinburgh j bot"] = "Edinburgh Journal of Botany",
  ["edinburgh med j"] = "Edinburgh medical journal",
  ["editor res rep"] = "Editorial research reports",
  ["edntna erca j"] = "EDTNA/ERCA Journal",
  ["edorium j disabil rehabil"] = "Edorium journal of disability and rehabilitation",
  ["edorium j surg"] = "Edorium journal of surgery",
  ["eds mag"] = "EDS magazine : official journal of the E.E.C. Dental Students Committee",
  ["edtna erca j"] = "EDTNA/ERCA journal (English ed.)",
  ["educ 3 13"] = "Education 3-13",
  ["educ action res"] = "Educational action research",
  ["educ adm q"] = "Educational administration quarterly : EAQ",
  ["educ assess eval account"] = "Educational assessment, evaluation and accountability",
  ["educ broadcast int"] = "Educational broadcasting international",
  ["educ chem"] = "Education in Chemistry",
  ["educ chem eng"] = "Education for Chemical Engineers",
  ["educ citizsh soc justice"] = "Education, citizenship and social justice",
  ["educ dent (ica)"] = "Educación dental",
  ["educ dir dent aux"] = "Educational directions for dental auxiliaries",
  ["educ dir dent hyg"] = "Educational directions in dental hygiene",
  ["educ doc inf"] = "Educational documentation and information : Bulletin of the International Bureau of Education",
  ["educ econ"] = "Education economics",
  ["educ eval policy anal"] = "Educational evaluation and policy analysis",
  ["educ finance policy"] = "Education finance and policy",
  ["educ focus"] = "Educational focus",
  ["educ form"] = "Education et Formation",
  ["educ forum"] = "The Educational forum",
  ["educ francoph"] = "Éducation et francophonie",
  ["educ gerontol"] = "Educational gerontology",
  ["educ health (abingdon)"] = "Education for health (Abingdon, England)",
  ["educ horiz"] = "Educational horizons",
  ["educ ind telev"] = "Educational & industrial television",
  ["educ inf"] = "Education for Information",
  ["educ inf technol (dordr)"] = "Education and information technologies",
  ["educ leadersh"] = "Educational leadership : journal of the Department of Supervision and Curriculum Development, N.E.A",
  ["educ med salud"] = "Educación médica y salud",
  ["educ neurosci"] = "Educational Neuroscience",
  ["educ north"] = "Education in the north",
  ["educ policy (los altos calif)"] = "Educational policy (Los Altos, Calif.)",
  ["educ policy anal arch"] = "Education policy analysis archives",
  ["educ prim care"] = "Education for primary care : an official publication of the Association of Course Organisers, National Association of GP Tutors, World Organisation of Family Doctors",
  ["educ psychol"] = "Educational psychologist",
  ["educ psychol (lond)"] = "Educational psychology",
  ["educ psychol meas"] = "Educational and psychological measurement",
  ["educ psychol pract"] = "Educational psychology in practice",
  ["educ psychol rev"] = "Educational psychology review",
  ["educ quim"] = "Educacion Quimica",
  ["educ real"] = "Educação e realidade",
  ["educ rec"] = "The Educational record",
  ["educ res"] = "Educational researcher (Washington, D.C. : 1972)",
  ["educ res (windsor)"] = "Educational research; a review for teachers and all concerned with progress in education",
  ["educ res eval"] = "Educational research and evaluation : an international journal on theory and practice",
  ["educ res int"] = "Education research international",
  ["educ res rev"] = "Educational research review",
  ["educ rev (birm)"] = "Educational review",
  ["educ sci"] = "Education in science : the bulletin of the Association for Science Education",
  ["educ sci (basel)"] = "Education sciences",
  ["educ soc"] = "Educação & sociedade",
  ["educ stud"] = "Educational studies",
  ["educ stud math"] = "Educational Studies in Mathematics",
  ["educ technol res dev"] = "Educational technology research and development : ETR & D",
  ["educ theory"] = "Educational theory",
  ["educ train autism dev disabil"] = "Education and training in autism and developmental disabilities",
  ["educ train dev disabil"] = "Education and training in developmental disabilities",
  ["educ train ment retard"] = "Education and training of the mentally retarded",
  ["educ treat children"] = "Education & treatment of children",
  ["educ update"] = "Educator's update",
  ["educ urban soc"] = "Education and urban society",
  ["educ week"] = "Education week",
  ["educafrica (ed fr)"] = "Educafrica (Version française)",
  ["education (chula vista)"] = "Education",
  ["eecs tech rep ser"] = "EECS technical report series",
  ["eeg emg z elektroenzephalogr elektromyogr verwandte geb"] = "EEG-EMG Zeitschrift für Elektroenzephalographie, Elektromyographie und verwandte Gebiete",
  ["eej trans fundam mater"] = "IEEJ Transactions On Fundamentals And Materials",
  ["ees catal"] = "EES Catalysis",
  ["eesti nsv tead akad toim biol"] = "Eesti NSV Teaduste Akadeemia toimetised. Bioloogia = Izvestiia Akademii nauk Estonskoĭ SSR. Biologiia",
  ["eesti teaduste akadeemia toimetised bioloogia, okoloogia|proc estonian acad sci biol ecol"] = "Proceedings of the Estonian Academy of Sciences. Biology, ecology",
  ["eff clin pract"] = "Effective clinical practice : ECP",
  ["eff health care"] = "Effective health care",
  ["effluent water treat j"] = "Effluent and water treatment journal",
  ["efort open rev"] = "EFORT open reviews",
  ["efsa j"] = "EFSA journal. European Food Safety Authority",
  ["efsa supporting publ"] = "EFSA Supporting Publications",
  ["egems (wash dc)"] = "EGEMS (Washington, DC)",
  ["egit psikol olcme deger derg"] = "Eğitimde ve psikolojide ölçme ve değerlendirme dergisi",
  ["egypt acad j biol sci b zool"] = "Egyptian academic journal of biological sciences. B, Zoology",
  ["egypt acad j biol sci c physiol mol biol"] = "Egyptian academic journal of biological sciences. C, Physiology & molecular biology",
  ["egypt dent j"] = "Egyptian dental journal",
  ["egypt heart j"] = "The Egyptian heart journal : (EHJ) : official bulletin of the Egyptian Society of Cardiology",
  ["egypt inf j"] = "Egyptian Informatics Journal",
  ["egypt j agron"] = "Egyptian Journal of Agronomy",
  ["egypt j anaesth"] = "Egyptian journal of anaesthesia",
  ["egypt j anal chem"] = "Egyptian Journal of Analytical Chemistry",
  ["egypt j aquat biol fish"] = "Egyptian journal of aquatic biology and fisheries",
  ["egypt j aquat res"] = "Egyptian Journal of Aquatic Research",
  ["egypt j archaeol restor stud"] = "Egyptian Journal of Archaeological and Restoration Studies",
  ["egypt j basic appl sci"] = "Egyptian Journal of Basic and Applied Sciences",
  ["egypt j basic clin pharmacol"] = "Egyptian journal of basic and clinical pharmacology",
  ["egypt j bilharz"] = "Egyptian journal of bilharziasis",
  ["egypt j biochem mol biol"] = "Egyptian Journal of Biochemistry and Molecular Biology",
  ["egypt j biol pest control"] = "Egyptian journal of biological pest control",
  ["egypt j biotechnol"] = "Egyptian Journal of Biotechnology",
  ["egypt j bot"] = "Egyptian Journal of Botany",
  ["egypt j chem"] = "Egyptian journal of chemistry",
  ["egypt j chest dis tuberc"] = "The Egyptian journal of chest diseases and tuberculosis",
  ["egypt j exp biol zool"] = "Egyptian Journal of experimental biology. Zoology",
  ["egypt j forensic sci"] = "Egyptian journal of forensic sciences",
  ["egypt j genet cytol"] = "Egyptian journal of genetics and cytology",
  ["egypt j geol"] = "Egyptian Journal of Geology",
  ["egypt j haematol"] = "The Egyptian journal of haematology : the official journal of the Egyptian Society of Haematology",
  ["egypt j hosp med"] = "Egyptian Journal of Hospital Medicine",
  ["egypt j immunol"] = "The Egyptian journal of immunology",
  ["egypt j intern med"] = "The Egyptian journal of internal medicine",
  ["egypt j med hum genet"] = "Egyptian Journal of Medical Human Genetics",
  ["egypt j med microbiol"] = "Egyptian journal of medical microbiology",
  ["egypt j microbiol"] = "Egyptian journal of microbiology",
  ["egypt j neurol psychiatr neurosurg"] = "The Egyptian journal of neurology, psychiatry and neurosurgery",
  ["egypt j occup med"] = "Egyptian journal of occupational medicine",
  ["egypt j pet"] = "Egyptian Journal of Petroleum",
  ["egypt j phytopathol"] = "Egyptian journal of phytopathology",
  ["egypt j psychiatry"] = "The Egyptian journal of psychiatry : official journal of the Egyptian Psychiatric Association",
  ["egypt j radiol nucl med"] = "Egyptian Journal of Radiology and Nuclear Medicine",
  ["egypt j remote sens space sci"] = "Egyptian Journal of Remote Sensing and Space Science",
  ["egypt j vet sci"] = "Egyptian journal of veterinary science",
  ["egypt j virol"] = "Egyptian journal of virology",
  ["egypt med j"] = "Egyptian medical journal",
  ["egypt popul fam plann rev"] = "The Egyptian population and family planning review",
  ["egypt today"] = "Egypt today (International Business Associates (Cairo, Egypt))",
  ["egypte contemp"] = "L'Egypte contemporaine",
  ["egypte monde arabe"] = "Egypte/monde arabe",
  ["egyptian statist j"] = "The Egyptian Statistical Journal",
  ["ehealth int"] = "Ehealth international",
  ["ehealth telecommun syst netw"] = "E-health telecommunication systems and networks",
  ["ehp toxicogenomics"] = "EHP toxicogenomics : journal of the National Institute of Environmental Health Sciences",
  ["ehs law bull ser japan"] = "EHS law bulletin series, Japan. Japan",
  ["eib pap"] = "EIB Papers",
  ["eichsfelder heimath"] = "Eichsfelder Heimathefte",
  ["eighteenth century (lubbock)"] = "Eighteenth century (Lubbock, Tex.)",
  ["eighteenth century curr bibliogr"] = "Eighteenth Century: A Current Bibliography",
  ["eighteenth century irel"] = "Eighteenth-century Ireland",
  ["eighteenth century life"] = "Eighteenth century life",
  ["eighteenth century stud"] = "Eighteenth-century studies",
  ["einblicke wiss"] = "Einblicke in die Wissenschaft",
  ["einstein (sao paulo)"] = "Einstein (São Paulo, Brazil)",
  ["einstein j biol med"] = "Einstein Journal of Biology and Medicine",
  ["einstein q j biol med"] = "The Einstein quarterly journal of biology and medicine",
  ["einstein stud"] = "Einstein Studies",
  ["eire irel"] = "Éire-Ireland; a journal of Irish studies",
  ["eisei dobutsu"] = "[Eisei dōbutsu] Japanese journal of sanitary zoology",
  ["eisei shikenjo hokoku"] = "Eisei Shikenjo hōkoku. Bulletin of National Institute of Hygienic Sciences",
  ["eiszeitalt ggw"] = "Eiszeitalter und Gegenwart",
  ["eiszeitalter ggw"] = "Eiszeitalter und Gegenwart",
  ["eiyo to shokuryo"] = "[Eiyŏ to shokuryŏ] Journal of the Japanese Society of Food and Nutrition",
  ["ej neonatol res"] = "The e-journal of neonatology research",
  ["ejnmmi phys"] = "EJNMMI physics",
  ["ejnmmi radiopharm chem"] = "EJNMMI radiopharmacy and chemistry",
  ["ejnmmi res"] = "EJNMMI research",
  ["ejournal biol sci"] = "eJournal of biological sciences",
  ["ejves short rep"] = "EJVES short reports",
  ["ejves vasc forum"] = "EJVES vascular forum",
  ["eklem hastalik cerrahisi"] = "Eklem hastalıkları ve cerrahisi = Joint diseases & related surgery",
  ["ekol cssr"] = "Ekologia (Cssr)",
  ["ekol genet"] = "Ėkologicheskai︠a︡ genetika",
  ["ekologia (cssr)"] = "Ekologia (CSSR)",
  ["ekologija (liet moksl akad (spausd))"] = "Ekologija (Vilnius, Lithuania)",
  ["ekon cas"] = "Ekonomický c̆asopis",
  ["ekon gaz"] = "Ėkonomicheskai︠a︡ gazeta",
  ["ekon istrazivanja"] = "Ekonomska Istrazivanja",
  ["ekon keuangan indones"] = "Ekonomi dan Keuangan Indonesia",
  ["ekon misao"] = "Ekonomska misao",
  ["ekon misao praksa"] = "Ekonomska Misao i Praksa",
  ["ekon pregl"] = "Ekonomski pregled",
  ["ekon revija"] = "Ekonomska revija",
  ["ekon samf tidskr"] = "Ekonomiska samfundets tidskrift",
  ["ekon samfundets tidskr"] = "Ekonomiska Samfundets Tidskrift",
  ["ekon selsk khoz"] = "Ėkonomika selʹskogo khozi︠a︡ĭstva : ėkonomicheskiĭ zhurnal Ministerstva selʹskogo khozi︠a︡ĭstva SSSR",
  ["ekon sov ukr"] = "Ėkonomika Sovetskoĭ Ukrainy",
  ["ekon-mat obzor"] = "Ekonomicko-Matematicky Obzor",
  ["ekonom upravl"] = "Ekonomika i Upravlenie",
  ["ekonomika radvanskoyi ukhayeny"] = "Ekonomika Radvanskoyi Ukhayeny",
  ["ekotekhnol resursosberezhenie"] = "Ekotekhnologii i Resursosberezhenie",
  ["eksp khir anesteziol"] = "Eksperimental'naia khirurgiia i anesteziologiia",
  ["eksp khirurgiia"] = "Eksperimental'naia khirurgiia",
  ["eksp klin farmakol"] = "Eksperimental'naia i klinicheskaia farmakologiia",
  ["eksp klin gastroenterol"] = "Ėksperimental'nai︠a︡ i klinicheskai︠a︡ gastroėnterologii︠a︡ = Experimental & clinical gastroenterology",
  ["eksp med morfol"] = "Eksperimentalna meditsina i morfologiia",
  ["eksp onkol"] = "Eksperimentalʹnai︠a︡ onkologii︠a︡",
  ["el peruano d of"] = "El Peruano. Diario oficial. Peru",
  ["elder care"] = "Elderly care",
  ["elder law j"] = "The elder law journal",
  ["elder law rev (penrith south)"] = "Elder law review (Penrith South, N.S.W.)",
  ["elec eng electron"] = "Electrical Engineering and Electronics",
  ["elect law j"] = "Election law journal",
  ["elect stud"] = "Electoral studies",
  ["electr eng"] = "Electrical Engineering",
  ["electr eng appl signal process ser"] = "Electrical Engineering and Applied Signal Processing Series",
  ["electr eng handbook ser"] = "The Electrical Engineering Handbook Series",
  ["electr eng jpn"] = "Electrical Engineering in Japan",
  ["electr engrg commun signal process ser"] = "Electrical Engineering Communications and Signal Processing Series",
  ["electr j"] = "Electricity Journal",
  ["electr mater appl"] = "Electrical Materials and Applications",
  ["electr power compon syst"] = "Electric Power Components & Systems",
  ["electr power syst res"] = "Electric Power Systems Reserch",
  ["electro- magnetobiol"] = "Electro- and Magnetobiology",
  ["electrocatalysis (n y)"] = "Electrocatalysis (New York, N.Y.)",
  ["electrochem commun"] = "Electrochemistry Communications",
  ["electrochem energy rev"] = "Electrochemical Energy Reviews",
  ["electrochem energy technol"] = "Electrochemical Energy Technology",
  ["electrochem sci adv"] = "Electrochemical Science Advances",
  ["electrochem soc interface"] = "Electrochemical Society Interface",
  ["electrochem solid-state lett"] = "Electrochemical and Solid-State Letters",
  ["electrochemistry (tokyo)"] = "Electrochemistry (Tokyo, Japan)",
  ["electrochemistry (tokyo, jpn)"] = "Electrochemistry (Tokyo, Japan)",
  ["electrochim acta"] = "Electrochimica Acta",
  ["electrocomponent sci technol"] = "Electrocomponent Science and Technology",
  ["electrodeposition surf treat"] = "Electrodeposition and Surface Treatment",
  ["electrodiagn ther"] = "Electrodiagnostic-Therapie",
  ["electroencephalogr clin neurophysiol"] = "Electroencephalography and clinical neurophysiology",
  ["electroencephalogr clin neurophysiol suppl"] = "Electroencephalography and clinical neurophysiology. Supplement",
  ["electrofisiol arritm"] = "Electrofisiología & arritmias",
  ["electrolyte blood press"] = "Electrolyte & blood pressure : E & BP",
  ["electromagn biol med"] = "Electromagnetic biology and medicine",
  ["electromagn waves (camb)"] = "Electromagnetic waves (Cambridge, Mass.)",
  ["electromagn wirel radar microw"] = "Electromagnetics, Wireless, Radar, and Microwaves",
  ["electromyogr clin neurophysiol"] = "Electromyography and clinical neurophysiology",
  ["electron backscatter diffr mater sci"] = "Electron Backscatter Diffraction in Materials Science",
  ["electron comm probab"] = "Electronic Communications in Probability",
  ["electron commer res appl"] = "Electronic commerce research and applications",
  ["electron commun jpn"] = "Electronics and Communications in Japan",
  ["electron commun probab"] = "Electronic Communications in Probability",
  ["electron educ"] = "Electronics Education",
  ["electron electric engrg res stud electromagnet appl ser"] = "Electronic & Electrical Engineering Research Studies: Electromagnetic Applications Series",
  ["electron eng syst ser"] = "Electronic Engineering Systems Series",
  ["electron int j time use res"] = "Electronic international journal of time use research",
  ["electron j appl stat anal"] = "Electronic Journal of Applied Statistical Analysis",
  ["electron j aust n z hist"] = "The electronic journal of Australian and New Zealand history",
  ["electron j biotechnol"] = "Electronic Journal of Biotechnology",
  ["electron j bound elem"] = "Electronic Journal of Boundary Elements",
  ["electron j boundary elem"] = "Electronic Journal of Boundary Elements",
  ["electron j comb"] = "Electronic journal of combinatorics",
  ["electron j combin"] = "Electronic Journal of Combinatorics",
  ["electron j differ equ conf"] = "Electronic Journal of Differential Equations Conference",
  ["electron j differ equations"] = "Electronic Journal of Differential Equations",
  ["electron j differential equations"] = "Electronic Journal of Differential Equations",
  ["electron j geotech eng"] = "Electronic Journal of Geotechnical Engineering",
  ["electron j graph theory appl (ejgta)"] = "Electronic Journal of Graph Theory and Applications. EJGTA",
  ["electron j inf syst dev countries"] = "Electronic Journal of Information Systems in Developing Countries",
  ["electron j linear algebra"] = "Electronic Journal of Linear Algebra",
  ["electron j math anal appl"] = "Electronic journal of mathematical analysis and applications",
  ["electron j oper meteorol"] = "Electronic Journal of Operational Meteorology",
  ["electron j probab"] = "Electronic Journal of Probability",
  ["electron j qual theory differ equ"] = "Electronic Journal of Qualitative Theory of Differential Equations",
  ["electron j qual theory differ equations"] = "Electronic Journal of Qualitative Theory of Differential Equations",
  ["electron j sci educ"] = "The electronic journal of science education",
  ["electron j severe storms metereol"] = "Electronic Journal of Severe Storms Metereology",
  ["electron j stat"] = "Electronic Journal of Statistics",
  ["electron j theor chem"] = "Electronic Journal of Theoretical Chemistry",
  ["electron lett"] = "Electronics Letters",
  ["electron magn photonic mater div (empmd) monogr ser"] = "Electronic, Magnetic and Photonic Materials Division (EMPMD) Monograph Series",
  ["electron mater lett"] = "Electronic Materials Letters",
  ["electron med dig"] = "Electronic medical digest",
  ["electron microsc"] = "Electron Microscopy",
  ["electron microsc rev"] = "Electron microscopy reviews",
  ["electron news"] = "Electronic News",
  ["electron notes discrete math"] = "Electronic Notes in Discrete Mathematics",
  ["electron notes theor comput sci"] = "Electronic Notes in Theoretical Computer Science",
  ["electron notes theor inform comput sci"] = "Electronic Notes in Theoretical Informatics and Computer Science",
  ["electron physician"] = "Electronic physician",
  ["electron power"] = "Electronics and Power",
  ["electron proc theor comput sci"] = "Electronic proceedings in theoretical computer science",
  ["electron proc theor comput sci (eptcs)"] = "Electronic Proceedings in Theoretical Computer Science (EPTCS)",
  ["electron res announc amer math soc"] = "Electronic Research Announcements of the American Mathematical Society",
  ["electron res announc math sci"] = "Electronic Research Announcements in Mathematical Sciences",
  ["electron res arch"] = "Electronic Research Archive",
  ["electron struct"] = "Electronic Structure",
  ["electron syst news"] = "Electronic Systems News",
  ["electron technol"] = "Electronic Technology",
  ["electron texts engrs sci"] = "Electronics Texts for Engineers and Scientists",
  ["electron trans artif intell"] = "Electronic Transactions on Artificial Intelligence",
  ["electron trans numer anal"] = "Electronic Transactions on Numerical Analysis",
  ["electron workshops comput"] = "Electronic Workshops in Computing",
  ["electronic j hum sex"] = "The electronic journal of human sexuality",
  ["electronics (basel)"] = "Electronics",
  ["electrophoresis (weinheim, fed repub ger)"] = "Electrophoresis (Weinheim, Federal Republic of Germany)",
  ["elektromed biomed tech"] = "Elektro Medizin; Biomedizin und Technik",
  ["elektron elektrotech"] = "Elektronika ir Elektrotechnika",
  ["elektrotech inf"] = "Elektrotechnik und Informationstechnik",
  ["elektrotech inftech"] = "Elektrotechnik und Informationstechnik",
  ["elektrotech maschinenbau"] = "Elektrotechnik und Maschinenbau",
  ["elektrotech z"] = "Elektrotechnische Zeitschrift",
  ["elem appl categ theory"] = "Elements in Applied Category Theory",
  ["elem decis theory philos"] = "Elements in Decision Theory and Philosophy",
  ["elem math"] = "Elemente der Mathematik",
  ["elem non-local data interact: found appl"] = "Elements in Non-local Data Interactions: Foundations and Applications",
  ["elem philos biol"] = "Elements in the Philosophy of Biology",
  ["elem philos math"] = "Elements in the Philosophy of Mathematics",
  ["elem sch j"] = "The Elementary school journal",
  ["elem sci anth"] = "Elementa: Science of the Anthropocene",
  ["elementa (wash d c)"] = "Elementa (Washington, D.C.)",
  ["elements (que)"] = "Elements (Québec, Québec)",
  ["ellis horwood ser phys appl"] = "Ellis Horwood Series in Physics and its Applications",
  ["elsevier oceanogr ser"] = "Elsevier Oceanography Series",
  ["eltex studientexte elektrotech"] = "ELTEX Studientexte Elektrotechnik",
  ["em (pittsburgh pa)"] = "EM (Pittsburgh, Pa.)",
  ["embalmers mon natl funer dir"] = "The Embalmers' monthly and national funeral director",
  ["embnet j"] = "EMBnet.journal",
  ["embo j"] = "The EMBO journal",
  ["embo mol med"] = "EMBO molecular medicine",
  ["embo rep"] = "EMBO reports",
  ["embryologia (nagoya)"] = "Embryologia",
  ["emc pediatr"] = "EMC. Pediatría",
  ["eme eme estud domin"] = "Eme eme; estudios dominicanos",
  ["emed res"] = "eMedical research",
  ["emerg (tehran)"] = "Emergency (Tehran, Iran)",
  ["emerg adulthood"] = "Emerging adulthood (Print)",
  ["emerg complex comput"] = "Emergence, Complexity and Computation",
  ["emerg contam"] = "Emerging contaminants",
  ["emerg dep news"] = "Emergency department news : EDN",
  ["emerg health serv q"] = "Emergency health services quarterly",
  ["emerg health serv rev"] = "Emergency health services review",
  ["emerg health threats j"] = "Emerging health threats journal",
  ["emerg infect dis"] = "Emerging infectious diseases",
  ["emerg med"] = "Emergency medicine (Glendale, Calif.) [sound recording]",
  ["emerg med (fremantle)"] = "Emergency medicine (Fremantle, W.A.)",
  ["emerg med (los angel)"] = "Emergency medicine : open access",
  ["emerg med (n y)"] = "Emergency medicine",
  ["emerg med australas"] = "Emergency medicine Australasia : EMA",
  ["emerg med clin north am"] = "Emergency medicine clinics of North America",
  ["emerg med int"] = "Emergency medicine international",
  ["emerg med investig"] = "Emergency medicine investigations",
  ["emerg med j"] = "Emergency medicine journal : EMJ",
  ["emerg med pract"] = "Emergency medicine practice",
  ["emerg med rep"] = "Emergency medicine reports",
  ["emerg med serv"] = "Emergency medical services",
  ["emerg methodol appl model identif control"] = "Emerging Methodologies and Applications in Modelling, Identification and Control",
  ["emerg microbes infect"] = "Emerging microbes & infections",
  ["emerg nurse"] = "Emergency nurse : the journal of the RCN Accident and Emergency Nursing Association",
  ["emerg plann dig"] = "Emergency planning digest",
  ["emerg radiol"] = "Emergency radiology",
  ["emerg themes epidemiol"] = "Emerging themes in epidemiology",
  ["emerg top life sci"] = "Emerging topics in life sciences",
  ["emerg top stat biostat"] = "Emerging Topics in Statistics and Biostatistics",
  ["emerg trends mechatron"] = "Emerging Trends in Mechatronics",
  ["emergent mater"] = "Emergent materials",
  ["emergent sci"] = "Emergent scientist",
  ["emerging anim species"] = "Emerging Animal Species",
  ["emerging contam"] = "Emerging Contaminants",
  ["emerging mater res"] = "Emerging Materials Research",
  ["emerging microbes infect"] = "Emerging Microbes & Infections",
  ["emerging top life sci"] = "Emerging Topics in Life Sciences",
  ["emi educ media int"] = "Educational media international",
  ["emir j food agric"] = "Emirates journal of food and agriculture",
  ["emisor demogr"] = "Emisor demográfico",
  ["emiss control sci technol"] = "Emission Control Science and Technology",
  ["emj gastroenterol"] = "EMJ. Gastroenterology",
  ["emory int law rev"] = "Emory international law review",
  ["emory law j"] = "Emory law journal",
  ["emory mag"] = "Emory magazine",
  ["emot behav diffic"] = "Emotional and behavioural difficulties : the journal of the Association of Workers for Children with Emotional and Behavioural Difficulties",
  ["emot rev"] = "Emotion review : journal of the International Society for Research on Emotion",
  ["emot space soc"] = "Emotion, space and society",
  ["emp state rep"] = "Empire state report",
  ["emphasis nurs"] = "Emphasis, nursing",
  ["empir econ"] = "Empirical economics",
  ["empir musicol rev"] = "Empirical musicology review : EMR",
  ["empir softw eng"] = "Empirical software engineering",
  ["empirical econ"] = "Empirical Economics",
  ["empirical software eng"] = "Empirical Software Engineering",
  ["empl benefits j"] = "Employee benefits journal",
  ["empl desempl"] = "Empleo y desempleo",
  ["empl health fit"] = "Employee health & fitness",
  ["employ relat today"] = "Employment relations today",
  ["employee benefit plan rev"] = "Employee benefit plan review",
  ["employee health fitness newsl"] = "Employee Health and Fitness Newsletter",
  ["employee relat law j"] = "Employee relations law journal",
  ["emporia state res stud"] = "The Emporia State research studies",
  ["ems cancer sci j"] = "EMS cancer science journal",
  ["ems mag"] = "EMS magazine",
  ["ems monogr math"] = "EMS Monographs in Mathematics",
  ["ems ser congr rep"] = "EMS Series of Congress Reports",
  ["ems ser ind appl math"] = "EMS Series in Industrial and Applied Mathematics",
  ["ems ser lect math"] = "EMS Series of Lectures in Mathematics",
  ["ems surv math sci"] = "EMS Surveys in Mathematical Sciences",
  ["ems textbk math"] = "EMS Textbooks in Mathematics",
  ["ems tracts math"] = "EMS Tracts in Mathematics",
  ["ems world"] = "EMS world",
  ["emt j"] = "The EMT journal",
  ["emu - austral ornithol"] = "Emu - Austral Ornithology",
  ["en familia"] = "En familia. Venezuela. Ministerio de Agricultura y Cría. Dirección de Extensión. División de Divulgación",
  ["encycl complex syst sci"] = "Encyclopedia of Complexity and Systems Science",
  ["encycl semant comput rob intell"] = "Encyclopedia with Semantic Computing and Robotic Intelligence",
  ["encycl soc work"] = "Encyclopedia of social work",
  ["encyclopaedia math sci"] = "Encyclopaedia of Mathematical Sciences",
  ["encyclopedia math appl"] = "Encyclopedia of Mathematics and its Applications",
  ["endanger species res"] = "Endangered species research",
  ["endocr connect"] = "Endocrine connections",
  ["endocr dev"] = "Endocrine development",
  ["endocr disruptors"] = "Endocrine Disruptors",
  ["endocr disruptors (austin)"] = "Endocrine disruptors (Austin, Tex.)",
  ["endocr j"] = "Endocrine journal",
  ["endocr metab immune disord drug targets"] = "Endocrine, metabolic & immune disorders drug targets",
  ["endocr metab sci"] = "Endocrine and Metabolic Science",
  ["endocr oncol"] = "Endocrine Oncology",
  ["endocr pathol"] = "Endocrine pathology",
  ["endocr pract"] = "Endocrine practice : official journal of the American College of Endocrinology and the American Association of Clinical Endocrinologists",
  ["endocr regul"] = "Endocrine regulations",
  ["endocr relat cancer"] = "Endocrine-related cancer",
  ["endocr res"] = "Endocrine research",
  ["endocr res commun"] = "Endocrine research communications",
  ["endocr rev"] = "Endocrine reviews",
  ["endocr terap"] = "Endocrinologia y terapéutica",
  ["endocr, metab immune disord: drug targets"] = "Endocrine, Metabolic & Immune Disorders: Drug Targets",
  ["endocrinol diabetes metab"] = "Endocrinology, diabetes & metabolism",
  ["endocrinol diabetes metab case rep"] = "Endocrinology, diabetes & metabolism case reports",
  ["endocrinol diabetes metab j"] = "Endocrinology, diabetes and metabolism journal",
  ["endocrinol diabetes nutr"] = "Endocrinología, diabetes y nutrición",
  ["endocrinol exp"] = "Endocrinologia experimentalis",
  ["endocrinol jpn"] = "Endocrinologia japonica",
  ["endocrinol metab"] = "Endocrinology and Metabolism",
  ["endocrinol metab (seoul)"] = "Endocrinology and metabolism (Seoul, Korea)",
  ["endocrinol metab clin north am"] = "Endocrinology and metabolism clinics of North America",
  ["endocrinol metab syndr"] = "Endocrinology & metabolic syndrome : current research",
  ["endocrinol nutr"] = "Endocrinología y nutrición : órgano de la Sociedad Española de Endocrinología y Nutrición",
  ["endocrinol sci cost"] = "Endocrinologia e scienza della costituzione",
  ["endocrinology (baltimore)"] = "Endocrinology (Baltimore)",
  ["endocytobiosis cell res"] = "Endocytobiosis and cell research",
  ["endod dent traumatol"] = "Endodontics & dental traumatology",
  ["endod prac"] = "Endodontic practice",
  ["endod pract"] = "Endodontic Practice",
  ["endod rep"] = "The Endodontic report",
  ["endod topics"] = "Endodontic topics",
  ["endodoncia (mex)"] = "Endodoncia",
  ["endokrynol diabetol chor przemiany materii wieku rozw"] = "Endokrynologia, diabetologia i choroby przemiany materii wieku rozwojowego : organ Polskiego Towarzystwa Endokrynologów Dziecięcych",
  ["endokrynol pol"] = "Endokrynologia Polska",
  ["endoplasmic reticulum stress dis"] = "Endoplasmic reticulum stress in diseases",
  ["endosc int open"] = "Endoscopy international open",
  ["endosc surg allied technol"] = "Endoscopic surgery and allied technologies",
  ["endosc ultrasound"] = "Endoscopic ultrasound",
  ["energ wasser prax"] = "Energie Wasser Praxis",
  ["energetic mater front"] = "Energetic Materials Frontiers",
  ["energies (basel)"] = "Energies",
  ["energotekhnol resursosberezhenie"] = "Energotekhnologii i Resursosberezhenie",
  ["energy (oxf)"] = "Energy (Oxford, England)",
  ["energy adv"] = "Energy Advances",
  ["energy ai"] = "Energy and AI",
  ["energy build"] = "Energy and Buildings",
  ["energy built environ"] = "Energy and Built Environment",
  ["energy clim change"] = "Energy and Climate Change",
  ["energy convers econ"] = "Energy Conversion and Economics",
  ["energy convers manag"] = "Energy conversion and management",
  ["energy convers manage"] = "Energy Conversion and Management",
  ["energy convers manage: x"] = "Energy Conversion and Management: X",
  ["energy ecol environ"] = "Energy, ecology & environment",
  ["energy econ"] = "Energy Economics",
  ["energy effic"] = "Energy Efficiency",
  ["energy emiss control technol"] = "Energy and emission control technologies",
  ["energy eng"] = "Energy Engineering",
  ["energy environ"] = "Energy & Environment",
  ["energy environ mater"] = "Energy & Environmental Materials",
  ["energy environ sci"] = "Energy & Environmental Science",
  ["energy explor exploit"] = "Energy Exploration & Exploitation",
  ["energy fuels"] = "Energy & fuels : an American Chemical Society journal",
  ["energy geosci"] = "Energy Geoscience",
  ["energy harvesting syst"] = "Energy Harvesting Systems",
  ["energy inf"] = "Energy Informatics",
  ["energy internet"] = "Energy Internet",
  ["energy j"] = "Energy Journal",
  ["energy mater"] = "Energy Materials",
  ["energy mater adv"] = "Energy Material Advances",
  ["energy minimization methods comput vis pattern recognit"] = "Energy minimization methods in computer vision and pattern recognition. International Workshop on Energy Minimization Methods in Computer Vision and Pattern Recognition",
  ["energy policy"] = "Energy policy",
  ["energy policy res"] = "Energy and Policy Research",
  ["energy power eng"] = "Energy and power engineering",
  ["energy procedia"] = "Energy Procedia",
  ["energy rep"] = "Energy Reports",
  ["energy res soc sci"] = "Energy research & social science",
  ["energy res social sci"] = "Energy Research & Social Science",
  ["energy rev"] = "Energy Reviews",
  ["energy sci"] = "Energy Science",
  ["energy sci eng"] = "Energy science & engineering",
  ["energy sources part a"] = "Energy Sources, Part A: Recovery, Utilization, and Environmental Effects",
  ["energy sources part b"] = "Energy Sources, Part B: Economics, Planning and Policy",
  ["energy storage"] = "Energy Storage",
  ["energy storage mater"] = "Energy storage materials",
  ["energy storage sav"] = "Energy Storage and Saving",
  ["energy strategy rev"] = "Energy Strategy Reviews",
  ["energy stud rev"] = "Energy Studies Review",
  ["energy sustain dev"] = "Energy for sustainable development : the journal of the International Energy Initiative",
  ["energy sustain soc"] = "Energy, sustainability and society",
  ["energy sustainability soc"] = "Energy, Sustainability and Society",
  ["energy sustainable dev"] = "Energy for Sustainable Development",
  ["energy syst"] = "Energy Systems",
  ["energy syst electr eng"] = "Energy Systems in Electrical Engineering",
  ["energy technol"] = "Energy Technology",
  ["energy technol (weinh)"] = "Energy technology (Weinheim, Germany)",
  ["energy wasser prax"] = "Energie Wasser Praxis",
  ["enfant milieu trop"] = "L'enfant en milieu tropical",
  ["enferm clin"] = "Enfermería clínica",
  ["enferm infecc microbiol clin"] = "Enfermedades infecciosas y microbiología clínica",
  ["enferm intensiva"] = "Enfermería intensiva",
  ["enferm torax"] = "Enfermedades del tórax",
  ["enferma argent"] = "La Enfermera argentina",
  ["enfoque fem"] = "Enfoque feminista",
  ["enfoques aten prim"] = "Enfoques en atención primaria",
  ["eng agric environ food"] = "Engineering in Agriculture, Environment and Food",
  ["eng anal bound elem"] = "Engineering analysis with boundary elements",
  ["eng anal boundary elem"] = "Engineering Analysis with Boundary Elements",
  ["eng appl artif intell"] = "Engineering Applications of Artificial Intelligence",
  ["eng appl comput fluid mech"] = "Engineering Applications of Computational Fluid Mechanics",
  ["eng appl comput methods"] = "Engineering Applications of Computational Methods",
  ["eng biol"] = "Engineering Biology",
  ["eng cast solutions"] = "Engineered Casting Solutions",
  ["eng comput"] = "Engineering Computations",
  ["eng comput (swansea)"] = "Engineering computations",
  ["eng constr archit manage"] = "Engineering Construction and Architectural Management",
  ["eng costs prod econ"] = "Engineering Costs and Production Economics",
  ["eng econ"] = "Engineering Economist",
  ["eng fail anal"] = "Engineering Failure Analysis",
  ["eng fract mech"] = "Engineering Fracture Mechanics",
  ["eng geol"] = "Engineering Geology",
  ["eng j"] = "The engineers journal",
  ["eng j am inst steel constr"] = "Engineering Journal American Institute of Steel Construction",
  ["eng life sci"] = "Engineering in life sciences",
  ["eng manage j"] = "Engineering Management Journal",
  ["eng mater"] = "Engineering Materials",
  ["eng mater process"] = "Engineering Materials and Processes",
  ["eng med"] = "Engineering in medicine",
  ["eng microbiol"] = "Engineering Microbiology",
  ["eng min j"] = "Engineering and Mining Journal",
  ["eng min j press"] = "Engineering and Mining Journal Press",
  ["eng optim"] = "Engineering Optimization",
  ["eng optim methods app"] = "Engineering Optimization. Methods and Applications",
  ["eng process econ"] = "Engineering and Process Economics",
  ["eng project organ j"] = "Engineering Project Organization Journal",
  ["eng regener"] = "Engineered Regeneration",
  ["eng rep"] = "Engineering reports : open access",
  ["eng res express"] = "Engineering Research Express",
  ["eng sci"] = "Engineering and science",
  ["eng sci educ j"] = "Engineering Science and Education Journal",
  ["eng sci technol"] = "Engineering Science and Technology",
  ["eng sci technol int j"] = "Engineering Science and Technology, an International Journal",
  ["eng struct"] = "Engineering Structures",
  ["eng struct technol"] = "Engineering Structures and Technologies",
  ["eng stud"] = "Engineering Studies",
  ["eng syst sustain ser"] = "Engineering Systems and Sustainability Series",
  ["eng technol"] = "Engineering & Technology",
  ["eng trans"] = "Engineering Transactions",
  ["engag sci technol soc"] = "Engaging science, technology, and society",
  ["engage soc action"] = "Engage/social action",
  ["engei gakkai zasshi"] = "Engei Gakkai zasshi. Engei Gakkai (Japan)",
  ["engineering (beijing)"] = "Engineering (Beijing, China)",
  ["engl hist rev"] = "The English historical review",
  ["engl lang notes"] = "English language notes",
  ["engl lit renaiss"] = "English literary renaissance",
  ["engl stud (amst)"] = "English studies (Amsterdam, Netherlands)",
  ["enke reihe ao[a]"] = "Enke Reihe zur AO[A]",
  ["enlb emerg nurse leg bull"] = "ENLB. Emergency nurse legal bulletin",
  ["enliven chall cancer detect ther"] = "Enliven. Challenges in cancer detection and therapy",
  ["enliven j stem cell res regen med"] = "Enliven. Journal of stem cell research & regenerative medicine",
  ["eno fo"] = "Eno Fo",
  ["enquetes mus vie wallonne"] = "Enquêtes du Musée de la vie wallonne",
  ["ensaios mat"] = "Ensaios Matematicos",
  ["enseign assiste ordinateur"] = "Enseignement Assiste par Ordinateur",
  ["enseign math"] = "Enseignement des Mathematiques",
  ["enseign math (2)"] = "L’Enseignement Mathématique",
  ["enseign sup-maths"] = "Enseignement SUP-Maths",
  ["ent audiol news"] = "ENT & audiology news",
  ["enteric pathogens monit group rep"] = "Enteric Pathogens Monitoring Group report",
  ["enterp inf syst"] = "Enterprise information systems",
  ["enterp soc"] = "Enterprise & society",
  ["enterpr innovation manage stud"] = "Enterprise and Innovation Management Studies",
  ["enterprise society"] = "Enterprise and Society",
  ["entertain comput"] = "Entertainment computing",
  ["entomol abh"] = "Entomologische Abhandlungen",
  ["entomol am"] = "Entomologica Americana",
  ["entomol ber luzern"] = "Entomologische Berichte Luzern",
  ["entomol bl biol syst kafer"] = "Entomologische blätter für biologie und systematik der käfer",
  ["entomol exp appl"] = "Entomologia experimentalis et applicata",
  ["entomol fenn"] = "Entomologica Fennica",
  ["entomol gaz"] = "Entomologist's gazette",
  ["entomol gen"] = "Entomologia Generalis",
  ["entomol hell"] = "Entomologia Hellenica",
  ["entomol i parazitol issled v povolzhe"] = "Ėntomologicheskie i parazitologicheskie issledovanii︠a︡ v Povolzh'e",
  ["entomol nachr ber"] = "Entomologische Nachrichten und Berichte",
  ["entomol nachr bl"] = "Entomologisches Nachrichtenblatt (Vienna, Austria : 1994)",
  ["entomol news"] = "Entomological news",
  ["entomol ornithol herpetol"] = "Entomology, ornithology, & herpetology : current research",
  ["entomol res"] = "Entomological Research",
  ["entomol rev"] = "Entomological review",
  ["entomol rom"] = "Entomologica Romanica",
  ["entomol sci"] = "Entomological Science",
  ["entomol sin"] = "Entomologia Sinica = Zhongguo kun chong xue",
  ["entomol tidskr"] = "Entomologisk tidskrift",
  ["entomol vect"] = "Entomologia y Vectores",
  ["entomol z"] = "Entomologische Zeitschrift",
  ["entomol z insektenborse"] = "Entomologische Zeitschrift mit Insektenbörse",
  ["entomolog ber"] = "Entomologische berichten",
  ["entomologia (pavia)"] = "Entomologia",
  ["entomologica (bari)"] = "Entomologica (Bari, Italy)",
  ["entomologiste (paris)"] = "L' Entomologiste",
  ["entomon int z gesamte insektenkunde"] = "Entomon; internationale Zeitschrift für die gesamte Insektenkunde",
  ["entomotaxonomia|kun chong fen lei xue bao"] = "Kun chong fen lei xue bao",
  ["entre nous cph den"] = "Entre nous (Copenhagen, Denmark)",
  ["entrep hist"] = "Entreprises et histoire",
  ["entrepreneurship reg devel"] = "Entrepreneurship and Regional Development",
  ["entret bichat med entret bichat"] = "Les entretiens de Bichat. Médecine. Entretiens de Bichat",
  ["entret bichat pitie salpetriere ther"] = "Thérapeutique. Entretiens de Bichat Pitié-Salpêtrière",
  ["entropy (basel)"] = "Entropy (Basel, Switzerland)",
  ["entropy prod everyday live"] = "Entropy Production in Everyday Live",
  ["entscheid bundesgerichtshofes zivilsachen"] = "Entscheidungen des Bundesgerichtshofes in Zivilsachen. Germany (West). Bundesgerichtshof",
  ["entwicklungsgeschichte und systematik der pflanzen supplementum|plant sys evol, suppl"] = "Plant systematics and evolution. Supplementum",
  ["entwicklungsgeschichte und systematik der pflanzen|plant syst evol"] = "Plant systematics and evolution",
  ["enumer comb appl"] = "Enumerative Combinatorics and Applications",
  ["env polit"] = "Environmental politics",
  ["environ  eng res"] = "Environmental Engineering Research",
  ["environ aciton"] = "Environmental action",
  ["environ adv"] = "Environmental Advances",
  ["environ aff"] = "Environmental affairs",
  ["environ afr"] = "Environnement africain",
  ["environ anal health toxicol"] = "Environmental analysis, health and toxicology",
  ["environ archaeol"] = "Environmental Archaeology",
  ["environ behav"] = "Environment and behavior",
  ["environ bioindic"] = "Environmental Bioindicators",
  ["environ biol fishes"] = "Environmental Biology of Fishes",
  ["environ biol med"] = "Environmental biology and medicine",
  ["environ biosaf res"] = "Environmental Biosafety Research",
  ["environ biosafety res"] = "Environmental biosafety research",
  ["environ biotechnol"] = "Environmental biotechnology",
  ["environ carcinog ecotoxicol rev"] = "Environment Carcinogenesis and Ecotoxicology Reviews",
  ["environ challenges"] = "Environmental Challenges",
  ["environ change secur proj rep"] = "Environmental Change and Security Project report",
  ["environ chem"] = "Environmental Chemistry",
  ["environ chem ecotoxicol"] = "Environmental Chemistry and Ecotoxicology",
  ["environ chem lett"] = "Environmental Chemistry Letters",
  ["environ claims j"] = "Environmental Claims Journal",
  ["environ clim technol"] = "Environmental and Climate Technologies",
  ["environ commun"] = "Environmental communication",
  ["environ conserv"] = "Environmental conservation",
  ["environ conserv j"] = "Environment conservation journal",
  ["environ control biol"] = "Environmental Control in Biology",
  ["environ data sci"] = "Environmental Data Science",
  ["environ dev"] = "Environmental development",
  ["environ dev econ"] = "Environment and development economics",
  ["environ dev sustain"] = "Environment, development and sustainability",
  ["environ dev sustainability"] = "Environment, Development and Sustainability",
  ["environ devel econ"] = "Environment and Development Economics",
  ["environ dis"] = "Environmental disease",
  ["environ dna"] = "Environmental DNA (Hoboken, N.J.)",
  ["environ earth sci"] = "Environmental earth sciences",
  ["environ ecol stat"] = "Environmental and Ecological Statistics",
  ["environ econ pol stud"] = "Environmental Economics and Policy Studies",
  ["environ econ policy stud"] = "Environmental Economics and Policy Studies",
  ["environ educ inf"] = "Environmental education and information",
  ["environ educ res"] = "Environmental education research",
  ["environ educ res,"] = "Environmental Education Research",
  ["environ endocrinol"] = "Environmental Endocrinology",
  ["environ energy policy econ"] = "Environmental and energy policy and the economy",
  ["environ eng geosci"] = "Environmental and Engineering Geoscience",
  ["environ eng manag j"] = "Environmental engineering and management journal",
  ["environ eng manage j"] = "Environmental Engineering and Management Journal",
  ["environ eng policy"] = "Environmental Engineering and Policy",
  ["environ eng res"] = "Environmental Engineering Research",
  ["environ eng sci"] = "Environmental Engineering Science",
  ["environ entomol"] = "Environmental entomology",
  ["environ epidemiol"] = "Environmental epidemiology (Philadelphia, Pa.)",
  ["environ epigenet"] = "Environmental epigenetics",
  ["environ ethics"] = "Environmental ethics",
  ["environ evid"] = "Environmental evidence",
  ["environ evidence"] = "Environmental Evidence",
  ["environ exp biol"] = "Environmental and Experimental Biology",
  ["environ exp bot"] = "Environmental and experimental botany",
  ["environ fluid mech"] = "Environmental Fluid Mechanics",
  ["environ fluid mech (dordr)"] = "Environmental fluid mechanics (Dordrecht, Netherlands : 2001)",
  ["environ folio"] = "Environment folio",
  ["environ forensics"] = "Environmental Forensics",
  ["environ geochem health"] = "Environmental geochemistry and health",
  ["environ geol"] = "Environmental Geology",
  ["environ geosci"] = "Environmental Geosciences",
  ["environ hazards"] = "Environmental Hazards",
  ["environ health"] = "Environmental health : a global access science source",
  ["environ health insights"] = "Environmental health insights",
  ["environ health perspect"] = "Environmental health perspectives",
  ["environ health prev med"] = "Environmental health and preventive medicine",
  ["environ health preventative med"] = "Environmental Health and Preventive Medicine",
  ["environ health ser [radiol health]"] = "Environmental Health Series. Radiological Health",
  ["environ health ser radiol health"] = "Environmental health series. RH, Radiological health",
  ["environ health toxicol"] = "Environmental health and toxicology",
  ["environ hist"] = "Environmental History",
  ["environ hist camb"] = "Environment and history",
  ["environ hist durh n c"] = "Environmental history",
  ["environ hist newsl"] = "Environmental History Newsletter",
  ["environ hist rev"] = "Environmental history review : EHR : a publication of the American Society for Environmental History",
  ["environ humanit"] = "Environmental humanities",
  ["environ impact assess rev"] = "Environmental impact assessment review",
  ["environ innov soc transit"] = "Environmental innovation and societal transitions",
  ["environ innovation manage"] = "Environment, Innovation and Management",
  ["environ innovation societal transitions"] = "Environmental Innovation and Societal Transitions",
  ["environ int"] = "Environment international",
  ["environ justice"] = "Environmental justice (Print)",
  ["environ law"] = "Environmental law (Northwestern School of Law)",
  ["environ law manage"] = "Environmental Law & Management",
  ["environ lett"] = "Environmental letters",
  ["environ manage"] = "Environmental management",
  ["environ manage health"] = "Environmental Management and Health",
  ["environ med"] = "Environmental medicine : annual report of the Research Institute of Environmental Medicine, Nagoya University",
  ["environ microbiol"] = "Environmental microbiology",
  ["environ microbiol rep"] = "Environmental microbiology reports",
  ["environ microbiome"] = "Environmental microbiome",
  ["environ microbiomes"] = "Environmental Microbiomes",
  ["environ model assess"] = "Environmental Modeling and Assessment",
  ["environ model softw"] = "Environmental modelling & software : with environment data news",
  ["environ modell software"] = "Environmental Modelling and Software",
  ["environ mol mutagen"] = "Environmental and Molecular Mutagenesis",
  ["environ monit assess"] = "Environmental monitoring and assessment",
  ["environ mutagen"] = "Environmental mutagenesis",
  ["environ nanotechnol monit manag"] = "Environmental nanotechnology, monitoring & management",
  ["environ nanotechnol monit manage"] = "Environmental Nanotechnology, Monitoring & Management",
  ["environ nat resour j"] = "Environment and natural resources journal",
  ["environ physiol biochem"] = "Environmental physiology & biochemistry",
  ["environ plan a"] = "Environment & planning A",
  ["environ plan b urban anal city sci"] = "Environment and planning. B, urban analytics and city science",
  ["environ plan d"] = "Environment and planning. D, Society & space",
  ["environ plan e nat space"] = "Environment and planning. E, Nature and space",
  ["environ plann a: econ space"] = "Environment and Planning A: Economy and Space",
  ["environ plann b plann des"] = "Environment and planning. B, Planning & design",
  ["environ plann b: urban anal city sci"] = "Environment and Planning B: Urban Analytics and City Science",
  ["environ plann c gov policy"] = "Environment and planning. C, Government & policy",
  ["environ plann c: polit spaces"] = "Environment and Planning C: Politics and Spaces",
  ["environ plann d: soc space"] = "Environment and Planning D: Society and Space",
  ["environ plann e: nat space"] = "Environment and Planning E: Nature and Space",
  ["environ plann f"] = "Environment and Planning F",
  ["environ planning a"] = "Environment and Planning A",
  ["environ planning c: govt pol"] = "Environment and Planning C: Government and Policy",
  ["environ policy governance"] = "Environmental Policy and Governance",
  ["environ policy law"] = "Environmental Policy and Law",
  ["environ pollut"] = "Environmental pollution (Barking, Essex : 1987)",
  ["environ pollut (amsterdam, neth)"] = "Environmental Pollution (Amsterdam, Netherlands)",
  ["environ pollut (oxford, u k)"] = "Environmental Pollution (Oxford, United Kingdom)",
  ["environ pollut (tor)"] = "Environment and pollution (Toronto, Ont.)",
  ["environ pollut bioavailability"] = "Environmental Pollutants and Bioavailability",
  ["environ pract"] = "Environmental Practice",
  ["environ processes"] = "Environmental Processes",
  ["environ prog"] = "Environmental Progress",
  ["environ prog sustain energy"] = "Environmental progress & sustainable energy",
  ["environ prog sustainable energy"] = "Environmental Progress & Sustainable Energy",
  ["environ prot eng"] = "Environment Protection Engineering",
  ["environ psychol res"] = "Environmental Psychology Research",
  ["environ qual annu rep counc environ qual"] = "Environmental quality; annual report of the Council on Environmental Quality",
  ["environ qual manage"] = "Environmental Quality Management",
  ["environ qual saf"] = "Environmental quality and safety",
  ["environ qual saf suppl"] = "Environmental quality and safety. Supplement",
  ["environ res"] = "Environmental Research",
  ["environ res commun"] = "Environmental Research Communications",
  ["environ res j"] = "Environment research journal",
  ["environ res lett"] = "Environmental Research Letters",
  ["environ res technol"] = "Environmental Research and Technology",
  ["environ res: clim"] = "Environmental Research: Climate",
  ["environ res: ecol"] = "Environmental Research: Ecology",
  ["environ res: energy"] = "Environmental Research: Energy",
  ["environ res: food syst"] = "Environmental Research: Food Systems",
  ["environ res: health"] = "Environmental Research: Health",
  ["environ res: infrastruct sustainability"] = "Environmental Research: Infrastructure and Sustainability",
  ["environ resour econ"] = "Environmental and Resource Economics",
  ["environ resour econ (dordr)"] = "Environmental & resource economics",
  ["environ resource econ"] = "Environmental and Resource Economics",
  ["environ rev"] = "Environmental Review",
  ["environ sci"] = "Environmental sciences : an international journal of environmental physiology and toxicology",
  ["environ sci (camb)"] = "Environmental science : water research & technology",
  ["environ sci (ruse)"] = "Environmental sciences",
  ["environ sci ecotechnol"] = "Environmental Science and Ecotechnology",
  ["environ sci eur"] = "Environmental Sciences Europe",
  ["environ sci nano"] = "Environmental science. Nano",
  ["environ sci policy"] = "Environmental Science & Policy",
  ["environ sci pollut res"] = "Environmental Science and Pollution Research",
  ["environ sci pollut res int"] = "Environmental science and pollution research international",
  ["environ sci process impacts"] = "Environmental science. Processes & impacts",
  ["environ sci processes impacts"] = "Environmental Science: Processes & Impacts",
  ["environ sci technol"] = "Environmental science & technology",
  ["environ sci technol lett"] = "Environmental Science & Technology Letters",
  ["environ sci water res technol"] = "Environmental Science: Water Research & Technology",
  ["environ sci: adv"] = "Environmental Science: Advances",
  ["environ sci: atmos"] = "Environmental Science: Atmospheres",
  ["environ sci: nano"] = "Environmental Science: Nano",
  ["environ sci: processes impacts"] = "Environmental Science: Processes & Impacts",
  ["environ secur"] = "Environment and Security",
  ["environ sociol"] = "Environmental sociology",
  ["environ sustainability"] = "Environmental Sustainability",
  ["environ sustainability indic"] = "Environmental and Sustainability Indicators",
  ["environ syst dec"] = "Environment Systems and Decisions",
  ["environ syst decis"] = "Environment systems & decisions",
  ["environ syst res"] = "Environmental Systems Research",
  ["environ technol"] = "Environmental technology",
  ["environ technol innov"] = "Environmental technology & innovation",
  ["environ technol innovation"] = "Environmental Technology & Innovation",
  ["environ technol rev"] = "Environmental Technology Reviews",
  ["environ toxicol"] = "Environmental toxicology",
  ["environ toxicol chem"] = "Environmental Toxicology and Chemistry",
  ["environ toxicol pharmacol"] = "Environmental toxicology and pharmacology",
  ["environ toxicol water qual"] = "Environmental Toxicology and Water Quality",
  ["environ urban"] = "Environment and urbanization",
  ["environ values"] = "Environmental values",
  ["environ: sci policy sustainable dev"] = "Environment: Science and Policy for Sustainable Development",
  ["environmental monogr sympos"] = "Environmental Monographs & Symposia",
  ["environments (basel)"] = "Environments (Basel, Switzerland)",
  ["enzyme microb technol"] = "Enzyme and microbial technology",
  ["enzyme protein"] = "Enzyme and Protein",
  ["enzyme res"] = "Enzyme research",
  ["enzymes (essen)"] = "Enzymes",
  ["enzymol biol clin (basel)"] = "Enzymologia biologica et clinica",
  ["enzymol mol biol carbonyl metab"] = "Enzymology and molecular biology of carbonyl metabolism",
  ["eos (washington dc)"] = "Eos",
  ["eos trans trans am geophys union"] = "EOS, Transactions, American Geophysical Union",
  ["eos, trans amer geophys union"] = "Eos, Transactions, American Geophysical Union (EOS)",
  ["ep lab dig"] = "EP lab digest",
  ["epa j"] = "EPA journal",
  ["epe j"] = "EPE Journal",
  ["epeirotike hestia"] = "Ēpeirōtikē hestia",
  ["epet etaireias stereoelladikon meleton"] = "Epetēris Hetaireias Stereoelladikōn Meletōn",
  ["epi newsl"] = "EPI newsletter",
  ["epic ser comput"] = "EPiC Series in Computing",
  ["epidemiol biostat public health"] = "Epidemiology, biostatistics and public health",
  ["epidemiol bull"] = "Epidemiological bulletin",
  ["epidemiol bull (taipei taiwan)"] = "Epidemiology bulletin (Taipei, Taiwan)",
  ["epidemiol community health"] = "Epidemiology and community health",
  ["epidemiol health"] = "Epidemiology and health",
  ["epidemiol infect"] = "Epidemiology and infection",
  ["epidemiol infekc bolezni (mosc : 2011)"] = "Ėpidemiologii︠a︡ i infekt︠s︡ionnye bolezni (Moscow, Russia : 2011)",
  ["epidemiol methods"] = "Epidemiologic methods",
  ["epidemiol mikrobiol imunol"] = "Epidemiologie, mikrobiologie, imunologie : casopis Spolecnosti pro epidemiologii a mikrobiologii Ceské lékarské spolecnosti J.E. Purkyne",
  ["epidemiol perspect innov"] = "Epidemiologic perspectives & innovations : EP+I",
  ["epidemiol prev"] = "Epidemiologia e prevenzione",
  ["epidemiol psichiatr soc"] = "Epidemiologia e psichiatria sociale",
  ["epidemiol psychiatr sci"] = "Epidemiology and psychiatric sciences",
  ["epidemiol res int"] = "Epidemiology Research International",
  ["epidemiol rev"] = "Epidemiologic reviews",
  ["epidemiol serv saude"] = "Epidemiologia e serviços de saúde : revista do Sistema Unico de Saúde do Brasil",
  ["epidemiol vakcinoprofil"] = "Ėpidemiologii︠a︡ i vakt︠s︡inoprofilaktika",
  ["epidemiol vital stat rep"] = "Epidemiological and vital statistics report. Rapport épidémiologique et démographique",
  ["epidemiology (sunnyvale)"] = "Epidemiology (Sunnyvale, Calif.)",
  ["epigenet chromatin"] = "Epigenetics & Chromatin",
  ["epigenet commun"] = "Epigenetics Communications",
  ["epigenet diagn ther"] = "Epigenetic diagnosis & therapy",
  ["epigenet insights"] = "Epigenetics insights",
  ["epigenetics chromatin"] = "Epigenetics & chromatin",
  ["epigraph stud"] = "Epigraphische Studien",
  ["epijournal geom algebrique"] = "Epijournal de Geometrie Algebrique. EPIGA",
  ["epilepsia open"] = "Epilepsia open",
  ["epilepsy behav"] = "Epilepsy & behavior : E&B",
  ["epilepsy behav case rep"] = "Epilepsy & behavior case reports",
  ["epilepsy behav rep"] = "Epilepsy & behavior reports",
  ["epilepsy curr"] = "Epilepsy currents",
  ["epilepsy j"] = "Epilepsy journal",
  ["epilepsy res"] = "Epilepsy research",
  ["epilepsy res suppl"] = "Epilepsy research. Supplement",
  ["epilepsy res treat"] = "Epilepsy research and treatment",
  ["epileptic disord"] = "Epileptic disorders : international epilepsy journal with videotape",
  ["epistemata würzburg wissensch schrift reihe philos"] = "Epistemata",
  ["episteme (edinb)"] = "Episteme (Edinburgh)",
  ["epithelial cell biol"] = "Epithelial cell biology",
  ["epitheor klin farmakol farmakokinet int ed"] = "Epitheorese Klinikes Farmakologias kai Farmakokinetikes, International Edition",
  ["epitheor koin ereun"] = "Epitheorese Koinonikon Ereunon",
  ["epizoot anim health west afr"] = "Epizootiology and animal health in West Africa",
  ["epj appl metamater"] = "EPJ Applied Metamaterials",
  ["epj appl phys"] = "EPJ Applied Physics",
  ["epj data sci"] = "EPJ Data Science",
  ["epj n nucl sci technol"] = "EPJ N Nuclear Sciences & Technologies",
  ["epj nonlinear biomed phys"] = "EPJ Nonlinear Biomedical Physics",
  ["epj nucl sci technol"] = "EPJ Nuclear Sciences and Technologies",
  ["epj photovolt"] = "EPJ Photovoltaics",
  ["epj photovoltaics"] = "EPJ Photovoltaics",
  ["epj plus"] = "EPJ Plus",
  ["epj quantum technol"] = "EPJ quantum technology",
  ["epj special topics"] = "EPJ Special Topics",
  ["epj tech instrum"] = "EPJ Techniques and Instrumentation",
  ["epj web conf"] = "EPJ web of conferences",
  ["epma j"] = "The EPMA journal",
  ["eppo bull"] = "EPPO Bulletin",
  ["equilib res"] = "Heikō shinkei kagaku",
  ["equine comp exercise physiol"] = "Equine and Comparative Exercise Physiology",
  ["equine vet educ"] = "Equine veterinary education",
  ["equine vet j"] = "Equine veterinary journal",
  ["equine vet j suppl"] = "Equine veterinary journal. Supplement",
  ["equity excell educ"] = "Equity & excellence in education : University of Massachusetts School of Education journal",
  ["era social"] = "Era socialistă",
  ["ercoftac ser"] = "ERCOFTAC Series",
  ["erdkd wissen"] = "Erdkundliches Wissen",
  ["erdoel, erdgas, kohle"] = "Erdoel, Erdgas, Kohle",
  ["erdoel, kohle, erdgas, petrochem"] = "Erdoel & Kohle Erdgas, Petrochemie",
  ["erevan gos univ uchen zap estestv nauki"] = "Erevanskiy Gosudarstvennyy Universitet",
  ["ergeb allg pathol pathol anat"] = "Ergebnisse der allgemeinen Pathologie und pathologischen Anatomie",
  ["ergeb anat entwicklungsgesch"] = "Ergebnisse der Anatomie und Entwicklungsgeschichte",
  ["ergeb biol"] = "Ergebnisse der Biologie",
  ["ergeb chir orthop"] = "Ergebnisse der Chirurgie und Orthopadie",
  ["ergeb enzymforsch"] = "Ergebnisse der Enzymforschung",
  ["ergeb gesamten tuberkuloseforsch"] = "Ergebnisse der gesamten Tuberkuloseforschung",
  ["ergeb hyg bakteriol immunitatsforsch exp ther"] = "Ergebnisse der Hygiene, Bakteriologie, Immunitätsforschung und experimentellen Therapie",
  ["ergeb inn med kinderheilkd"] = "Ergebnisse der inneren Medizin und Kinderheilkunde",
  ["ergeb math grenzgeb (3)"] = "Ergebnisse der Mathematik und ihrer Grenzgebiete. 3. Folge. A Series of Modern Surveys in Mathematics",
  ["ergeb mikrobiol immunitatsforsch exp ther"] = "Ergebnisse der Mikrobiologie, Immunitätsforschung und experimentellen Therapie",
  ["ergeb physiol"] = "Ergebnisse der Physiologie, biologischen Chemie und experimentellen Pharmakologie",
  ["ergeb wiss unters schweiz natlpark"] = "Ergebnisse der wissenschaftlichen Untersuchungen im schweizerischen Nationalpark",
  ["ergo (ann arbor)"] = "Ergo (Ann Arbor, Mich.)",
  ["ergod theory dyn syst"] = "Ergodic Theory and Dynamical Systems",
  ["ergodic theory dynam systems"] = "Ergodic Theory and Dynamical Systems",
  ["ergon des"] = "Ergonomics in design : the magazine of human factors applications",
  ["ergon int j"] = "Ergonomics international journal",
  ["erj open res"] = "ERJ open research",
  ["ernahrung wiss prax"] = "Ernährung - Wissenschaft und Praxis",
  ["ernahrungsforsch ber mitt"] = "Ernährungsforschung. Berichte und Mitteilungen",
  ["ernst schering found symp proc"] = "Ernst Schering Foundation symposium proceedings",
  ["ernst schering res found workshop"] = "Ernst Schering Research Foundation workshop",
  ["ers spectr"] = "ERS spectrum",
  ["ertel’s collect pap"] = "Ertel’s Collected Papers",
  ["erzieh unterr"] = "Erziehung und Unterricht; österreichische pädagogische Zeitschrift",
  ["es +t"] = "Es + T",
  ["esa bull"] = "ESA bulletin. Bulletin ASE. European Space Agency",
  ["esa j"] = "ESA journal",
  ["esa sci tech rev"] = "ESA Scientific and Technical Review",
  ["esaim control optim calc var"] = "ESAIM. Control, Optimisation and Calculus of Variations",
  ["esaim math model numer anal"] = "ESAIM. Mathematical Modelling and Numerical Analysis",
  ["esaim probab stat"] = "ESAIM. Probability and statistics = Probabilités et statistique : P & S",
  ["esaim probab statist"] = "European Series in Applied and Industrial Mathematics",
  ["esaim proc"] = "ESAIM Proceedings",
  ["esaim proc surv"] = "ESAIM. Proceedings and surveys",
  ["esaim proc surveys"] = "ESAIM: Proceedings and Surveys",
  ["esaim: proc surv"] = "ESAIM: Proceedings and Surveys",
  ["esc farm"] = "La Escuela de farmacia",
  ["esc heart fail"] = "ESC heart failure",
  ["escr psicol"] = "Escritos de psicología",
  ["esencia odontol"] = "Esencia Odontologica",
  ["esercit mat"] = "Esercitazioni di Matematica",
  ["esi lect math phys"] = "ESI Lectures in Mathematics and Physics",
  ["esis publ"] = "ESIS Publication",
  ["esmo open"] = "ESMO open",
  ["espac abierto"] = "Espacio abierto",
  ["espace geogr"] = "L'Espace géographique",
  ["espace popul soc"] = "Espace, populations, sociétés",
  ["espaces soc"] = "Espaces et sociétés (Paris, France)",
  ["espen j"] = "e-SPEN journal",
  ["esper dermatol"] = "Esperienze dermatologiche",
  ["esper ric"] = "Esperienze e ricerche",
  ["esquisses math"] = "Esquisses Mathematiques",
  ["esrc data arch bull"] = "ESRC Data Archive bulletin",
  ["essays arts sci"] = "Essays in arts and sciences",
  ["essays biochem"] = "Essays in biochemistry",
  ["essays cogn psych"] = "Essays in Cognitive Psychology",
  ["essays fundam immunol"] = "Essays in fundamental immunology",
  ["essays neurochem neuropharmacol"] = "Essays in neurochemistry and neuropharmacology",
  ["essays stud (lond)"] = "Essays and studies (London, England : 1950)",
  ["essence (downsview)"] = "Essence",
  ["essence comput ser"] = "The Essence of Computing Series",
  ["essent drugs monit"] = "Essential drugs monitor",
  ["essent number theory"] = "Essential Number Theory",
  ["essent psychopharmacol"] = "Essential psychopharmacology",
  ["essent textb math"] = "Essential Textbooks in Mathematics",
  ["essex inst hist collect"] = "Essex Institute historical collections",
  ["essex j"] = "Essex journal",
  ["est j archaeol"] = "Estonian Journal of Archaeology",
  ["est j earth sci"] = "Estonian Journal Of Earth Sciences",
  ["est j eng"] = "Estonian Journal of Engineering",
  ["estad panamena"] = "Estadística panameña. Boletín",
  ["estad panamena situac demogr secc 231 migr migr int"] = "Estadística panameña. Situación demográfica, sección 231: Migración internacional. Panama. Dirección de Estadística y Censo",
  ["estad venez"] = "Estadística venezolana",
  ["estadíst española"] = "Instituto Nacional de Estadística",
  ["este pais"] = "Este país (Mexico City, Mexico)",
  ["estodont press"] = "Estodont/press",
  ["estomatol cult"] = "Estomatologia e cultura",
  ["estuar coast shelf sci"] = "Estuarine, coastal and shelf science",
  ["estuaries coast"] = "Estuaries and coasts : journal of the Estuarine Research Federation",
  ["estuaries coasts"] = "Estuaries and Coasts",
  ["estuarine coastal mar sci"] = "Estuarine and Coastal Marine Science",
  ["estuarine coastal shelf sci"] = "Estuarine, Coastal and Shelf Science",
  ["estud asia afr"] = "Estudios de Asia y Africa",
  ["estud catedra hist farm legis farm univ granada"] = "Estudios de la Cátedra de Historia de la Farmacia y Legislación Farmacéutica de la Universidad de Granada",
  ["estud cebrap"] = "Estudos CEBRAP. Centro Brasileiro de Análise e Planejamento",
  ["estud cent am"] = "ECA : estudios centro americanos",
  ["estud cl"] = "Estudios clásicos",
  ["estud cult nahuatl"] = "Estudios de cultura náhuatl",
  ["estud de econ"] = "Estudios de Economia",
  ["estud demogr"] = "Estudos demográficos : revista do Gabinete de Estudos Demográficos do Instituto Nacional de Estatística",
  ["estud demogr urbanos col mex"] = "Estudios demográficos y urbanos",
  ["estud dep hist mod"] = "Estudios del Departamento de Historia Moderna. Universidad de Zaragoza. Departamento de Historia moderna",
  ["estud econ"] = "Estudios de economía",
  ["estud econ (mex)"] = "Estudios económicos (México City, Mexico)",
  ["estud econ (sao paulo)"] = "Estudos econômicos",
  ["estud empresariales"] = "Estudios Empresariales",
  ["estud fem"] = "Estudos feministas",
  ["estud geogr"] = "Estudios geográficos",
  ["estud hist"] = "Estudos históricos (Rio de Janeiro, Brazil)",
  ["estud hist doc arxius protoc"] = "Estudis històrics i documents dels Arxius de Protocols",
  ["estud hist med vaca"] = "Estudios de historia de la medicina vasca",
  ["estud hist novohisp"] = "Estudios de historia novohispana",
  ["estud hist saude"] = "Estudos de história e saúde",
  ["estud hist soc"] = "Estudios de historia social (Madrid, Spain)",
  ["estud hist soc econ am"] = "Estudios de historia social y económica de América",
  ["estud int"] = "Estudios Internacionales",
  ["estud latinoam"] = "Estudios latinoamericanos (Mexico City, Mexico)",
  ["estud migr"] = "Estudios migratorios",
  ["estud migr latinoam"] = "Estudios migratorios latinoamericanos",
  ["estud parag"] = "Estudios paraguayos",
  ["estud poblac"] = "Estudios de población",
  ["estud psicol"] = "Estudios de psicología",
  ["estud psiq"] = "Estudos psíquicos",
  ["estud publicos"] = "Estudios públicos",
  ["estud rom (barcelona)"] = "Estudis romànics",
  ["estud rurales latinoam"] = "Estudios rurales latinoamericanos",
  ["estud segov"] = "Estudios segovianos",
  ["estud soc"] = "Estudios sociales (Santo Domingo, Dominican Republic)",
  ["estud soc centroam"] = "Estudios sociales centroamericanos",
  ["estud sociol"] = "Estudios sociológicos (Mexico City, Mexico)",
  ["estudos econ"] = "Estudos Economicos",
  ["et j"] = "ET journal",
  ["eta sigma gamman"] = "The Eta Sigma Gamman",
  ["etfrn newsletter"] = "ETFRN (European Tropical Forest Research Network) Newsletter",
  ["ethical currents"] = "Ethical currents : a publication of the Center for Healthcare Ethics, St. Joseph Health System, and California Association of Catholic Hospitals",
  ["ethical hum psychol psychiatry"] = "Ethical human psychology and psychiatry",
  ["ethical hum sci serv"] = "Ethical human sciences and services : an international journal of critical inquiry",
  ["ethical perspect"] = "Ethical perspectives",
  ["ethical theory moral pract"] = "Ethical theory and moral practice : an international forum",
  ["ethics behav"] = "Ethics & behavior",
  ["ethics biol eng med"] = "Ethics in biology, engineering & medicine",
  ["ethics biol eng med: int j"] = "Ethics in Biology, Engineering and Medicine: An International Journal",
  ["ethics comm newsl"] = "Ethics committee newsletter",
  ["ethics environ"] = "Ethics & the Environment",
  ["ethics hum res"] = "Ethics & human research",
  ["ethics inf technol"] = "Ethics and information technology",
  ["ethics int aff"] = "Ethics & international affairs",
  ["ethics intellect disabil"] = "Ethics and intellectual disability",
  ["ethics law aging rev"] = "Ethics, law, and aging review",
  ["ethics med"] = "Ethics & medicine : a Christian perspective on issues in bioethics",
  ["ethics med public health"] = "Ethics, medicine, and public health",
  ["ethics medics"] = "Ethics and medics",
  ["ethics policy environ"] = "Ethics, policy & environment",
  ["ethics sci environ polit"] = "Ethics in science and environmental politics",
  ["ethics sci med"] = "Ethics in science & medicine",
  ["ethics soc welf"] = "Ethics and social welfare",
  ["ethik med"] = "Ethik in der Medizin : Organ der Akademie für Ethik in der Medizin",
  ["ethiop herald"] = "The Ethiopian herald",
  ["ethiop j dev res"] = "Ethiopian journal of development research",
  ["ethiop j health dev"] = "The Ethiopian journal of health development = YaʼItyoṗyā ṭénā lemāt maṣḥét",
  ["ethiop j health sci"] = "Ethiopian journal of health sciences",
  ["ethiop j public health nutr"] = "Ethiopian journal of public health and nutrition",
  ["ethiop med j"] = "Ethiopian medical journal",
  ["ethiop obs"] = "Ethiopia observer",
  ["ethique sante"] = "Éthique & santé",
  ["ethn dis"] = "Ethnicity & disease",
  ["ethn forum"] = "Ethnic forum : bulletin of ethnic studies and ethnic bibliography",
  ["ethn health"] = "Ethnicity & health",
  ["ethn inequal health soc care"] = "Ethnicity and inequalities in health and social care",
  ["ethn racial stud"] = "Ethnic and racial studies",
  ["ethnic groups"] = "Ethnic Groups",
  ["ethnol eur"] = "Ethnologia Europaea",
  ["ethnol fr"] = "Ethnologie française",
  ["ethnol scand"] = "Ethnologia Scandinavica",
  ["ethnologies (que)"] = "Ethnologies",
  ["ethol ecol evol"] = "Ethology, ecology & evolution",
  ["ethol sociobiol"] = "Ethology and sociobiology",
  ["etnogr pol"] = "Etnografia polska",
  ["etnol stud"] = "Etnologiska studier",
  ["etri j"] = "ETRI Journal",
  ["ettore majorana int sci ser: phys sci"] = "Ettore Majorana International Science Series: Physical Sciences",
  ["etud angl"] = "Etudes anglaises (Periodical)",
  ["etud balk"] = "Etudes balkaniques",
  ["etud can"] = "Études canadiennes. Canadian studies",
  ["etud class"] = "Études classiques (Namur, Belgium)",
  ["etud conjoncture"] = "Etudes et Conjoncture",
  ["etud doc groupe demogr afr"] = "Etudes & documents (Groupe de démographie africaine (France))",
  ["etud expans"] = "Etudes & expansion",
  ["etud fr"] = "Etudes françaises (Montréal, Québec)",
  ["etud ger"] = "Etudes germaniques",
  ["etud gest sols"] = "Etude et Gestion des Sols",
  ["etud hist afr"] = "Études d'histoire africaine",
  ["etud hist med"] = "Etudes d'histoire de la médecine",
  ["etud int"] = "Etudes internationales",
  ["etud inuit"] = "Études inuit. Inuit studies",
  ["etud irl"] = "Etudes irlandaises",
  ["etud limousines"] = "Études limousines",
  ["etud napoleon"] = "Etudes napoléoniennes : bulletin historique de la Société de sauvegarde du château impérial de Pont-de-Briques",
  ["etud polemol"] = "Etudes polémologiques",
  ["etud rurales"] = "Etudes rurales",
  ["etud soins serv infirm"] = "Etudes Sur Les Soins Et Le Service Infirmier",
  ["etud togol popul"] = "Etudes togolaises de population",
  ["etud vauclus"] = "Etudes vauclusiennes",
  ["etud zair"] = "Etudes zaïroises",
  ["etude popul afr"] = "Etude de la population africaine = African population studies",
  ["etudes (paris)"] = "Etudes",
  ["etudes cent rech etudes demogr"] = "Études du Centre de recherches et d'études démographiques. Centre de recherche et d'études démographiques",
  ["etudes freud"] = "Etudes freudiennes",
  ["etudes mali"] = "Études maliennes",
  ["etudes neonatales"] = "Etudes néo-natales. Neo-natal studies",
  ["etudes rwandaises"] = "Etudes rwandaises",
  ["etudes stat inst natl stat"] = "Etudes Statistiques",
  ["eubios j asian int bioeth"] = "Eubios journal of Asian and international bioethics : EJAIB",
  ["eugen news"] = "The Eugenical news",
  ["eugen q"] = "Eugenics quarterly",
  ["eugen rev"] = "The Eugenics review",
  ["eugen soc symp"] = "Eugenics Society symposia",
  ["eukaryot cell"] = "Eukaryotic cell",
  ["eukaryotic cell"] = "Eukaryotic Cell",
  ["eupa open proteom"] = "EuPA open proteomics",
  ["eupa open proteomics"] = "EuPA Open Proteomics",
  ["eur acad res"] = "European academic research",
  ["eur account rev"] = "The European accounting review",
  ["eur actuar j"] = "European actuarial journal",
  ["eur actuarial j"] = "European Actuarial Journal",
  ["eur addict res"] = "European addiction research",
  ["eur ann allergy clin immunol"] = "European annals of allergy and clinical immunology",
  ["eur ann otorhinolaryngol head neck dis"] = "European annals of otorhinolaryngology, head and neck diseases",
  ["eur arch otorhinolaryngol"] = "European archives of oto-rhino-laryngology : official journal of the European Federation of Oto-Rhino-Laryngological Societies (EUFOS) : affiliated with the German Society for Oto-Rhino-Laryngology - Head and Neck Surgery",
  ["eur arch otorhinolaryngol suppl"] = "European archives of oto-rhino-laryngology. Supplement",
  ["eur arch paediatr dent"] = "European archives of paediatric dentistry : official journal of the European Academy of Paediatric Dentistry",
  ["eur arch psychiatry clin neurosci"] = "European archives of psychiatry and clinical neuroscience",
  ["eur arch psychiatry neurol sci"] = "European archives of psychiatry and neurological sciences",
  ["eur asia stud"] = "Europe-Asia studies",
  ["eur biophys j"] = "European biophysics journal : EBJ",
  ["eur cardiol"] = "European cardiology",
  ["eur cell mater"] = "European cells & materials",
  ["eur cells mater"] = "European Cells and Materials",
  ["eur chem bull"] = "European Chemical Bulletin",
  ["eur child adolesc psychiatry"] = "European child & adolescent psychiatry",
  ["eur clin respir j"] = "European clinical respiratory journal",
  ["eur consort math ind (berl)"] = "European Consortium for Mathematics in Industry (Berlin)",
  ["eur cytokine netw"] = "European cytokine network",
  ["eur cytokine network"] = "European Cytokine Network",
  ["eur eat disord rev"] = "European eating disorders review : the journal of the Eating Disorders Association",
  ["eur econ rev"] = "European economic review",
  ["eur endocrinol"] = "European endocrinology",
  ["eur endod j"] = "European endodontic journal",
  ["eur ethn"] = "Europa ethnica",
  ["eur food res technol"] = "European Food Research and Technology",
  ["eur foreign aff rev"] = "European foreign affairs review",
  ["eur gastroenterol hepatol rev"] = "European gastroenterology & hepatology review",
  ["eur geriatr med"] = "European geriatric medicine",
  ["eur heart j"] = "European heart journal",
  ["eur heart j acute cardiovasc care"] = "European heart journal. Acute cardiovascular care",
  ["eur heart j cardiovasc imaging"] = "European heart journal cardiovascular Imaging",
  ["eur heart j cardiovasc pharmacother"] = "European heart journal. Cardiovascular pharmacotherapy",
  ["eur heart j case rep"] = "European heart journal. Case reports",
  ["eur heart j digital health"] = "European Heart Journal - Digital Health",
  ["eur heart j imaging methods pract"] = "European Heart Journal - Imaging Methods and Practice",
  ["eur heart j open"] = "European Heart Journal Open",
  ["eur heart j qual care clin outcomes"] = "European heart journal. Quality of care & clinical outcomes",
  ["eur heart j suppl"] = "European heart journal supplements : journal of the European Society of Cardiology",
  ["eur hist q"] = "European history quarterly",
  ["eur hochschulschr vii med b gesch med"] = "Europäische Hochschulschriften. Reihe VII, Medizin. Abt. B, Geschichte der Medizin = Publications universitaires européennes. Série VII, Médicine. Div. B, Histoire de la médecine = European university papers. Series VII, Health sciences. Div. B, History of medicine",
  ["eur hum rights law rev"] = "European human rights law review",
  ["eur infect dis"] = "European infectious disease",
  ["eur intellect prop rev"] = "European intellectual property review",
  ["eur j adv res biol life sci"] = "European journal of advanced research in biological and life sciences",
  ["eur j ageing"] = "European journal of ageing",
  ["eur j agron"] = "European Journal of Agronomy",
  ["eur j anaesthesiol"] = "European journal of anaesthesiology",
  ["eur j anaesthesiol suppl"] = "European journal of anaesthesiology. Supplement",
  ["eur j anat"] = "European journal of anatomy : official journal of the Spanish Society of Anatomy",
  ["eur j appl math"] = "European Journal of Applied Mathematics",
  ["eur j appl microbiol"] = "European Journal of Applied Microbiology",
  ["eur j appl microbiol biotechnol"] = "European Journal of Applied Microbiology and Biotechnology",
  ["eur j appl physiol"] = "European journal of applied physiology",
  ["eur j appl physiol occup physiol"] = "European journal of applied physiology and occupational physiology",
  ["eur j archaeol"] = "European journal of archaeology",
  ["eur j arrhythm electrophysiol"] = "European journal of arrhythmia & electrophysiology",
  ["eur j artif intell"] = "The European Journal on Artificial Intelligence",
  ["eur j basic appl histochem"] = "European journal of basic and applied histochemistry",
  ["eur j behav anal"] = "European journal of behavior analysis",
  ["eur j biochem"] = "European Journal of Biochemistry",
  ["eur j biol res"] = "European journal of biological research",
  ["eur j biol sci"] = "European journal of biological sciences",
  ["eur j biomed inform (praha)"] = "European journal for biomedical informatics",
  ["eur j breast health"] = "European journal of breast health",
  ["eur j cancer"] = "European journal of cancer (Oxford, England : 1990)",
  ["eur j cancer b oral oncol"] = "European journal of cancer. Part B, Oral oncology",
  ["eur j cancer care"] = "European Journal of Cancer Care",
  ["eur j cancer care (engl)"] = "European journal of cancer care",
  ["eur j cancer clin oncol"] = "European journal of cancer & clinical oncology",
  ["eur j cancer prev"] = "European journal of cancer prevention : the official journal of the European Cancer Prevention Organisation (ECP)",
  ["eur j cardiol"] = "European journal of cardiology",
  ["eur j cardiothorac surg"] = "European journal of cardio-thoracic surgery : official journal of the European Association for Cardio-thoracic Surgery",
  ["eur j cardiovasc med"] = "The European journal of cardiovascular medicine",
  ["eur j cardiovasc nurs"] = "European journal of cardiovascular nursing : journal of the Working Group on Cardiovascular Nursing of the European Society of Cardiology",
  ["eur j cardiovasc prev rehabil"] = "European journal of cardiovascular prevention and rehabilitation : official journal of the European Society of Cardiology, Working Groups on Epidemiology & Prevention and Cardiac Rehabilitation and Exercise Physiology",
  ["eur j cardiovasc res"] = "European journal of cardiovascular research",
  ["eur j case rep intern med"] = "European journal of case reports in internal medicine",
  ["eur j cell biol"] = "European Journal of Cell Biology",
  ["eur j cell biol suppl"] = "European journal of cell biology. Supplement",
  ["eur j chem"] = "European journal of chemistry (Print)",
  ["eur j clin chem clin biochem"] = "European Journal of Clinical Chemistry and Clinical Biochemistry",
  ["eur j clin invest"] = "European journal of clinical investigation",
  ["eur j clin microbiol"] = "European journal of clinical microbiology",
  ["eur j clin microbiol infect dis"] = "European journal of clinical microbiology & infectious diseases : official publication of the European Society of Clinical Microbiology",
  ["eur j clin nutr"] = "European journal of clinical nutrition",
  ["eur j clin pharmacol"] = "European journal of clinical pharmacology",
  ["eur j cogn psychol"] = "The European journal of cognitive psychology",
  ["eur j comb"] = "European journal of combinatorics = Journal européen de combinatoire = Europäische Zeitschrift für Kombinatorik",
  ["eur j combin"] = "European Journal of Combinatorics",
  ["eur j commun"] = "European journal of communication",
  ["eur j comput mech"] = "European Journal of Computational Mechanics",
  ["eur j contracept reprod health care"] = "The European journal of contraception & reproductive health care : the official journal of the European Society of Contraception",
  ["eur j control"] = "European Journal of Control",
  ["eur j crim pol res"] = "European journal on criminal policy and research",
  ["eur j criminol"] = "European journal of criminology",
  ["eur j cult polit sociol"] = "European journal of cultural and political sociology",
  ["eur j dent"] = "European journal of dentistry",
  ["eur j dent educ"] = "European journal of dental education : official journal of the Association for Dental Education in Europe",
  ["eur j dermatol"] = "European journal of dermatology : EJD",
  ["eur j dev psychol"] = "The European journal of developmental psychology",
  ["eur j dev res"] = "The European journal of development research",
  ["eur j dev sci"] = "European journal of developmental science",
  ["eur j disord commun"] = "European journal of disorders of communication : the journal of the College of Speech and Language Therapists, London",
  ["eur j drug metab pharmacokinet"] = "European journal of drug metabolism and pharmacokinetics",
  ["eur j east asian stud"] = "European journal of East Asian studies",
  ["eur j echocardiogr"] = "European journal of echocardiography : the journal of the Working Group on Echocardiography of the European Society of Cardiology",
  ["eur j ecol"] = "European journal of ecology",
  ["eur j educ"] = "European journal of education",
  ["eur j emerg med"] = "European journal of emergency medicine : official journal of the European Society for Emergency Medicine",
  ["eur j endocrinol"] = "European journal of endocrinology",
  ["eur j eng educ"] = "European Journal of Engineering Educatino",
  ["eur j engl stud"] = "European journal of English studies",
  ["eur j entomol"] = "European Journal of Entomology",
  ["eur j environ civ eng"] = "European Journal of Environmental and Civil Engineering",
  ["eur j environ public health"] = "European journal of environment and public health",
  ["eur j epidemiol"] = "European journal of epidemiology",
  ["eur j esthet dent"] = "The European journal of esthetic dentistry : official journal of the European Academy of Esthetic Dentistry",
  ["eur j exp biol"] = "European journal of experimental biology",
  ["eur j for pathol"] = "European Journal of Forest Pathology",
  ["eur j for res"] = "European journal of forest research",
  ["eur j forest pathol"] = "European journal of forest pathology. Journal européen de pathologie forestière. Europäische Zeitschrift für Forstpathologie",
  ["eur j futures res"] = "European Journal of Futures Research",
  ["eur j gastroenterol hepatol"] = "European journal of gastroenterology & hepatology",
  ["eur j gen pract"] = "The European journal of general practice",
  ["eur j genet mol toxicol"] = "The European journal of genetic toxicology",
  ["eur j genet soc"] = "European journal of genetics in society : an ethical approach to genetics",
  ["eur j gynaecol oncol"] = "European journal of gynaecological oncology",
  ["eur j haematol"] = "European journal of haematology",
  ["eur j haematol suppl"] = "European journal of haematology. Supplementum",
  ["eur j health econ"] = "The European journal of health economics : HEPAC : health economics in prevention and care",
  ["eur j health law"] = "European journal of health law",
  ["eur j health psychol"] = "European journal of health psychology",
  ["eur j heart fail"] = "European journal of heart failure",
  ["eur j hist econ thought"] = "The European journal of the history of economic thought",
  ["eur j histochem"] = "European journal of histochemistry : EJH",
  ["eur j homelessness"] = "European journal of homelessness",
  ["eur j hortic sci"] = "European journal of horticultural science",
  ["eur j hosp pharm"] = "European journal of hospital pharmacy : science and practice",
  ["eur j hum genet"] = "European journal of human genetics : EJHG",
  ["eur j hybrid imaging"] = "European journal of hybrid imaging",
  ["eur j immunogenet"] = "European Journal of Immunogenetics",
  ["eur j immunol"] = "European journal of immunology",
  ["eur j ind eng"] = "European Journal of Industrial Engineering",
  ["eur j inf syst"] = "European journal of information systems : an official journal of the Operational Research Society",
  ["eur j inflamm"] = "European journal of inflammation",
  ["eur j innovation manage"] = "European Journal Of Innovation Management",
  ["eur j inorg chem"] = "European Journal of Inorganic Chemistry",
  ["eur j int aff"] = "The European journal of international affairs",
  ["eur j int relat"] = "European journal of international relations",
  ["eur j integr med"] = "European journal of integrative medicine",
  ["eur j intensive care med"] = "European journal of intensive care medicine",
  ["eur j intern med"] = "European journal of internal medicine",
  ["eur j invest health psychol educ"] = "European Journal of Investigation in Health, Psychology and Education",
  ["eur j investig health psychol educ"] = "European journal of investigation in health, psychology and education",
  ["eur j lipid sci technol"] = "European Journal of Lipid Science and Technology",
  ["eur j mark"] = "European journal of marketing",
  ["eur j mass spectrom"] = "European Journal of Mass Spectrometry",
  ["eur j mass spectrom (chichester)"] = "European journal of mass spectrometry (Chichester, England)",
  ["eur j mass spectrom (chichester, eng)"] = "European journal of mass spectrometry (Chichester, England)",
  ["eur j mater"] = "European Journal of Materials",
  ["eur j math"] = "European Journal of Mathematics",
  ["eur j mech a solids"] = "European Journal of Mechanics - A/Solids",
  ["eur j mech b fluids"] = "European Journal of Mechanics - B/Fluids",
  ["eur j med"] = "The European journal of medicine",
  ["eur j med chem"] = "European journal of medicinal chemistry",
  ["eur j med chem rep"] = "European Journal of Medicinal Chemistry Reports",
  ["eur j med genet"] = "European journal of medical genetics",
  ["eur j med res"] = "European journal of medical research",
  ["eur j microbiol immunol"] = "European Journal of Microbiology and Immunology",
  ["eur j microbiol immunol (bp)"] = "European journal of microbiology & immunology",
  ["eur j midwifery"] = "European journal of midwifery",
  ["eur j mineral"] = "European Journal of Mineralogy",
  ["eur j morphol"] = "European journal of morphology",
  ["eur j nanomed"] = "European journal of nanomedicine",
  ["eur j neurodegener dis"] = "European journal of neurodegenerative disease",
  ["eur j neurol"] = "European journal of neurology",
  ["eur j neurosci"] = "The European journal of neuroscience",
  ["eur j nucl med"] = "European journal of nuclear medicine",
  ["eur j nucl med mol imaging"] = "European journal of nuclear medicine and molecular imaging",
  ["eur j nutr"] = "European journal of nutrition",
  ["eur j nutr food saf"] = "European journal of nutrition & food safety",
  ["eur j obstet gynecol"] = "European journal of obstetrics and gynecology",
  ["eur j obstet gynecol reprod biol"] = "European journal of obstetrics, gynecology, and reproductive biology",
  ["eur j obstet gynecol reprod biol x"] = "European journal of obstetrics & gynecology and reproductive biology: X",
  ["eur j oncol nurs"] = "European journal of oncology nursing : the official journal of European Oncology Nursing Society",
  ["eur j oncol pharm"] = "European Journal of Oncology Pharmacy",
  ["eur j oper res"] = "European Journal of Operational Research",
  ["eur j ophthalmol"] = "European journal of ophthalmology",
  ["eur j oral implantol"] = "European journal of oral implantology",
  ["eur j oral sci"] = "European journal of oral sciences",
  ["eur j org chem"] = "European Journal of Organic Chemistry",
  ["eur j orthod"] = "European journal of orthodontics",
  ["eur j orthop surg traumatol"] = "European journal of orthopaedic surgery & traumatology : orthopédie traumatologie",
  ["eur j paediatr dent"] = "European journal of paediatric dentistry",
  ["eur j paediatr neurol"] = "European journal of paediatric neurology : EJPN : official journal of the European Paediatric Neurology Society",
  ["eur j pain"] = "European journal of pain (London, England)",
  ["eur j pain suppl"] = "European journal of pain supplements",
  ["eur j palliat care"] = "European journal of palliative care : the journal of the European Association for Palliative Care",
  ["eur j pediatr"] = "European journal of pediatrics",
  ["eur j pediatr surg"] = "European Journal of Pediatric Surgery",
  ["eur j pers"] = "European journal of personality",
  ["eur j pers cent healthc"] = "European journal for person centered healthcare",
  ["eur j pharm biopharm"] = "European journal of pharmaceutics and biopharmaceutics : official journal of Arbeitsgemeinschaft für Pharmazeutische Verfahrenstechnik e.V",
  ["eur j pharm med res"] = "European journal pharmaceutical and medical research",
  ["eur j pharm sci"] = "European journal of pharmaceutical sciences : official journal of the European Federation for Pharmaceutical Sciences",
  ["eur j pharmacol"] = "European Journal of Pharmacology",
  ["eur j pharmacol environ toxicol pharm"] = "European Journal of Pharmacology: Environmental Toxicology and Pharmacology",
  ["eur j pharmacol mol pharmacol"] = "European Journal of Pharmacology: Molecular Pharmacology",
  ["eur j pharmacol, mol pharmacol sect"] = "European Journal of Pharmacology, Molecular Pharmacology Section",
  ["eur j philos"] = "European journal of philosophy",
  ["eur j philos sci"] = "European Journal for Philosophy of Science",
  ["eur j phycol"] = "European Journal of Phycology",
  ["eur j phys"] = "European Journal of Physics",
  ["eur j phys rehabil med"] = "European journal of physical and rehabilitation medicine",
  ["eur j physiother"] = "European journal of physiotherapy",
  ["eur j plant pathol"] = "European journal of plant pathology",
  ["eur j plast surg"] = "European journal of plastic surgery",
  ["eur j polit econ"] = "European journal of political economy",
  ["eur j polit res"] = "European journal of political research",
  ["eur j political theory"] = "European journal of political theory : EJPT",
  ["eur j popul"] = "European journal of population = Revue européenne de démographie",
  ["eur j prev cardiol"] = "European journal of preventive cardiology",
  ["eur j prosthodont restor dent"] = "The European journal of prosthodontics and restorative dentistry",
  ["eur j protistol"] = "European journal of protistology",
  ["eur j psychiatry"] = "The European journal of psychiatry",
  ["eur j psychol"] = "Europe's journal of psychology",
  ["eur j psychol assess"] = "European journal of psychological assessment : official organ of the European Association of Psychological Assessment",
  ["eur j psychother couns"] = "European journal of psychotherapy & counselling",
  ["eur j psychotraumatol"] = "European journal of psychotraumatology",
  ["eur j public health"] = "European journal of public health",
  ["eur j pure appl math"] = "European Journal of Pure and Applied Mathematics",
  ["eur j radiol"] = "European journal of radiology",
  ["eur j radiol open"] = "European journal of radiology open",
  ["eur j remote sens"] = "European journal of remote sensing",
  ["eur j respir dis"] = "European journal of respiratory diseases",
  ["eur j respir dis suppl"] = "European journal of respiratory diseases. Supplement",
  ["eur j rheumatol"] = "European journal of rheumatology",
  ["eur j rheumatol inflamm"] = "European journal of rheumatology and inflammation",
  ["eur j risk regul"] = "European journal of risk regulation : EJRR",
  ["eur j sci res"] = "European journal of scientific research = revue européenne de la recherche scientifique",
  ["eur j secur res"] = "European Journal for Security Research",
  ["eur j sex transm dis"] = "European journal of sexually transmitted diseases",
  ["eur j soc psychol"] = "European journal of social psychology",
  ["eur j soc secur"] = "European journal of social security",
  ["eur j soil biol"] = "European Journal of Soil Biology",
  ["eur j soil sci"] = "European Journal of Soil Science",
  ["eur j solid state inorg chem"] = "European Journal of Solid State and Inorganic Chemistry",
  ["eur j spec needs educ"] = "European journal of special needs education",
  ["eur j sport sci"] = "European journal of sport science",
  ["eur j sport soc"] = "European Journal for Sport and Society",
  ["eur j sports exerc sci"] = "European journal of sports & exercise science",
  ["eur j surg"] = "European Journal of Surgery",
  ["eur j surg oncol"] = "European journal of surgical oncology : the journal of the European Society of Surgical Oncology and the British Association of Surgical Oncology",
  ["eur j surg suppl"] = "The European journal of surgery. Supplement. : = Acta chirurgica. Supplement",
  ["eur j taxon"] = "European journal of taxonomy",
  ["eur j toxicol"] = "European journal of toxicology",
  ["eur j toxicol environ hyg"] = "European Journal of Toxicology and Environmental Hygiene",
  ["eur j toxicol hyg environ"] = "European journal of toxicology and hygiene of environment",
  ["eur j transl clin med"] = "European journal of translational and clinical medicine",
  ["eur j transl myol"] = "European journal of translational myology",
  ["eur j trauma emerg surg"] = "European journal of trauma and emergency surgery : official publication of the European Trauma Society",
  ["eur j ultrasound"] = "European Journal of Ultrasound",
  ["eur j vasc endovasc surg"] = "European journal of vascular and endovascular surgery : the official journal of the European Society for Vascular Surgery",
  ["eur j vasc surg"] = "European journal of vascular surgery",
  ["eur j wildl res"] = "European journal of wildlife research",
  ["eur j womens stud"] = "European journal of women's studies",
  ["eur j wood wood prod"] = "European Journal of Wood and Wood Products",
  ["eur j zool res"] = "European journal of zoological research",
  ["eur law rev"] = "European law review",
  ["eur leg towar new paradig"] = "The European legacy, toward new paradigms : journal of the International Society for the Study of European Ideas",
  ["eur mass spectrom"] = "European Mass Spectrometry",
  ["eur math soc mag"] = "European Mathematical Society Magazine",
  ["eur math soc newsl"] = "European Mathematical Society. Newsletter",
  ["eur med (paris)"] = "Europa medica. [Edición española]",
  ["eur med j (chelmsf)"] = "European Medical Journal (Chelmsford, England)",
  ["eur med j diabetes"] = "European medical journal. Diabetes",
  ["eur med j hepatol"] = "European medical journal. Hepatology",
  ["eur med j oncol"] = "European medical journal. Oncology",
  ["eur med j reprod health"] = "European Medical Journal. Reproductive health",
  ["eur med j respir"] = "European medical journal. Respiratory",
  ["eur med j rheumatol"] = "European medical journal. Rheumatology",
  ["eur med j urol"] = "European medical Journal. Urology",
  ["eur mem reihe i stud"] = "Europaea Memoria. Reihe I. Studien",
  ["eur mol biol organ j"] = "European Molecular Biology Organization Journal",
  ["eur musculoskelet rev"] = "European musculoskeletal review",
  ["eur neurol"] = "European neurology",
  ["eur neurol j"] = "European neurological journal",
  ["eur neurol rev"] = "European neurological review",
  ["eur neuropsychopharmacol"] = "European neuropsychopharmacology : the journal of the European College of Neuropsychopharmacology",
  ["eur om"] = "Europe, outremer",
  ["eur oncol"] = "European oncology",
  ["eur oncol haematol"] = "European oncology & haematology",
  ["eur oral res"] = "European oral research",
  ["eur orthop traumatol"] = "European orthopaedics and traumatology",
  ["eur paediatr rev"] = "European paediatrics review",
  ["eur phy educ rev"] = "European physical education review",
  ["eur phys educ rev"] = "European Physical Education Review",
  ["eur phys j a"] = "European Physical Journal A: Hadrons and Nuclei",
  ["eur phys j a hadron nucl"] = "The European physical journal. A, Hadrons and nuclei",
  ["eur phys j am"] = "European Physical Journal AM: Applied Metamaterials",
  ["eur phys j ap"] = "The European Physical Journal AP",
  ["eur phys j appl phys"] = "European Physical Journal Applied Physics",
  ["eur phys j b"] = "European Physical Journal B: Condensed Matter and Complex Systems",
  ["eur phys j b-condens matter complex syst"] = "European Physical Journal B: Condensed Matter and Complex Systems",
  ["eur phys j c"] = "European Physical Journal C: Particles and Fields",
  ["eur phys j c part fields"] = "The European physical journal. C, Particles and fields",
  ["eur phys j d"] = "European Physical Journal D: Atomic, Molecular, Optical and Plasma Physics",
  ["eur phys j d at mol opt phys"] = "The European physical journal. D, Atomic, molecular, and optical physics",
  ["eur phys j ds"] = "European Physical Journal DS: Data Science",
  ["eur phys j e"] = "European Physical Journal E: Soft Matter",
  ["eur phys j e soft matter"] = "The European physical journal. E, Soft matter",
  ["eur phys j e: soft matter biol phys"] = "European Physical Journal E: Soft Matter and Biological Physics",
  ["eur phys j h"] = "European Physical Journal H: Historical Perspectives on Contemporary Physics",
  ["eur phys j n"] = "European Physical Journal N: Nuclear Sciences and Technologies",
  ["eur phys j nbp"] = "European Physical Journal NBP: Nonlinear Biomedical Physics",
  ["eur phys j plus"] = "European Physical Journal Plus",
  ["eur phys j pv"] = "European Physical Journal PV: Photovoltaics",
  ["eur phys j qt"] = "European Physical Journal QT: Quantum Technology",
  ["eur phys j spec top"] = "European Physical Journal Special Topics",
  ["eur phys j special topics"] = "The European Physical Journal Special Topics",
  ["eur phys j st"] = "European Physical Journal ST: Special Topics",
  ["eur phys j ti"] = "European Physical Journal TI: Techniques and Instrumentation",
  ["eur phys j woc"] = "European Physical Journal WOC: Web of Conferences",
  ["eur polym j"] = "European Polymer Journal",
  ["eur potato j"] = "European Potato Journal",
  ["eur poult sci"] = "European Poultry Science",
  ["eur psychiatry"] = "European psychiatry : the journal of the Association of European Psychiatrists",
  ["eur psychol"] = "European psychologist",
  ["eur qual assur netw newsl"] = "European Quality Assurance Network newsletter",
  ["eur radio"] = "European Radiology",
  ["eur radio exp"] = "European Radiology Experimental",
  ["eur radio suppl"] = "European Radiology Supplements",
  ["eur radiol"] = "European radiology",
  ["eur radiol exp"] = "European radiology experimental",
  ["eur reg"] = "Europa Regional",
  ["eur respir j"] = "The European respiratory journal",
  ["eur respir j suppl"] = "The European respiratory journal. Supplement",
  ["eur respir rev"] = "European respiratory review : an official journal of the European Respiratory Society",
  ["eur rev"] = "European review (Chichester, England)",
  ["eur rev aging phys act"] = "European review of aging and physical activity : official journal of the European Group for Research into Elderly and Physical Activity",
  ["eur rev agric econ"] = "European Review of Agricultural Economics",
  ["eur rev appl psychol"] = "European review of applied psychology = Revue européenne de psychologie appliquée",
  ["eur rev econ hist"] = "European review of economic history",
  ["eur rev hist"] = "European review of history = Revue européene d'histoire",
  ["eur rev med pharmacol sci"] = "European review for medical and pharmacological sciences",
  ["eur rev soc psychol"] = "European review of social psychology",
  ["eur sci j"] = "European scientific journal",
  ["eur soc"] = "European Societies",
  ["eur sociol rev"] = "European sociological review",
  ["eur space agency spec publ sp"] = "European Space Agency Special Publication SP",
  ["eur spine j"] = "European spine journal : official publication of the European Spine Society, the European Spinal Deformity Society, and the European Section of the Cervical Spine Research Society",
  ["eur sport hist rev"] = "The European sports history review",
  ["eur stroke j"] = "European stroke journal",
  ["eur stud j"] = "The European studies journal",
  ["eur stud philos sci"] = "European Studies in Philosophy of Science",
  ["eur stud rev"] = "European studies review",
  ["eur surg"] = "European surgery : ACA : Acta chirurgica Austriaca",
  ["eur surg res"] = "European surgical research. Europäische chirurgische Forschung. Recherches chirurgicales européennes",
  ["eur thyroid j"] = "European thyroid journal",
  ["eur trans electr power"] = "European Transactions on Electrical Power",
  ["eur trans telecommun"] = "European Transactions on Telecommunications",
  ["eur transport res rev"] = "European Transport Research Review",
  ["eur union polit"] = "European Union politics",
  ["eur univ inst ser d: econom"] = "European University Institute. Series D: Economics",
  ["eur urban reg stud"] = "European urban and regional studies",
  ["eur urol"] = "European urology",
  ["eur urol focus"] = "European urology focus",
  ["eur urol oncol"] = "European urology oncology",
  ["eur urol open sci"] = "European urology open science",
  ["eur urol rev"] = "European urological review",
  ["eur water pollut control"] = "European water pollution control : official publication of the European Water Pollution Control Association (EWPCA)",
  ["eur zool j"] = "The European zoological journal",
  ["eura medicophys"] = "Europa medicophysica",
  ["eurasian chem technol j"] = "Eurasian Chemico-Technological Journal",
  ["eurasian entomol j"] = "Euroasian Entomological Journal",
  ["eurasian geogr econ"] = "Eurasian geography and economics",
  ["eurasian j anal chem"] = "Eurasian Journal of Analytical Chemistry",
  ["eurasian j biosci"] = "Eurasian journal of biosciences",
  ["eurasian j med"] = "The Eurasian journal of medicine",
  ["eurasian j med oncol"] = "Eurasian journal of medicine and oncology",
  ["eurasian math j"] = "Eurasian Mathematical Journal",
  ["eurasian min"] = "Euroasian Mining",
  ["eurasian soil sci"] = "Eurasian Soil Science",
  ["eurasip j adv signal process"] = "EURASIP journal on advances in signal processing",
  ["eurasip j appl signal processing"] = "EURASIP journal on applied signal processing",
  ["eurasip j audio speech music process"] = "EURASIP journal on audio, speech, and music processing",
  ["eurasip j bioinf syst biol"] = "EURASIP Journal on Bioinformatics & Systems Biology",
  ["eurasip j bioinform syst biol"] = "EURASIP journal on bioinformatics & systems biology",
  ["eurasip j embedded syst"] = "EURASIP Journal on Embedded Systems",
  ["eurasip j image video process"] = "EURASIP journal on image and video processing",
  ["eurasip j inf secur"] = "EURASIP journal on information security",
  ["eurasip j wirel commun netw"] = "EURASIP journal on wireless communications and networking",
  ["eurasip j wireless commun networking"] = "EURASIP Journal on Wireless Communications and Networking",
  ["euro adv tutor oper res"] = "EURO Advanced Tutorials on Operational Research",
  ["euro j comput optim"] = "EURO Journal on Computational Optimization",
  ["euro j decis process"] = "EURO Journal on Decision Processes",
  ["euro j transp logist"] = "EURO Journal on Transportation and Logistics",
  ["euro med j innov"] = "European medical journal. Innovations",
  ["euro rep stud"] = "EURO reports and studies",
  ["euro surveill"] = "Euro surveillance : bulletin Européen sur les maladies transmissibles = European communicable disease bulletin",
  ["euro-mediterr j environ integr"] = "Euro-Mediterranean Journal for Environmental Integration",
  ["euroasian j hepatogastroenterol"] = "Euroasian journal of hepato-gastroenterology",
  ["eurobiotech j"] = "The EuroBiotech journal",
  ["eurocourses comput inform sci"] = "Eurocourses. Computer and Information Science",
  ["eurograph ieee vgtc symp vis"] = "Eurographics/IEEE VGTC Symposium on Visualization : EUROVIS : [proceedings]. Eurographics/IEEE VGTC Symposium on Visualization",
  ["eurographics symp parallel graph vis"] = "Eurographics Symposium on Parallel Graphics and Visualization : EG PGV : [proceedings]. Eurographics Symposium on Parallel Graphics and Visualization",
  ["eurographics workshop vis comput biomed"] = "Eurographics Workshop on Visual Computing for Biomedicine",
  ["eurohealth (lond)"] = "Eurohealth",
  ["euromath bull"] = "Euromath Bulletin",
  ["euromediterr j environ integr"] = "Euro-Mediterranean journal for environmental integration",
  ["europ demogr inf bull"] = "European demographic information bulletin",
  ["europ econ rev"] = "European Economic Review",
  ["europ finan manage"] = "European Financial Management",
  ["europ finance rev"] = "European Finance Review",
  ["europ j devel res"] = "European Journal of Development Research",
  ["europ j finance"] = "European Journal of Finance",
  ["europ j hist econ thought"] = "European Journal of the History of Economic Thought",
  ["europ j housing pol"] = "European Journal of Housing Policy",
  ["europ j ind relat"] = "European Journal of Industrial Relations",
  ["europ j int relat"] = "European Journal of International Relations",
  ["europ j law econ"] = "European Journal of Law and Economics",
  ["europ j polit economy"] = "European Journal of Political Economy",
  ["europ rev agr econ"] = "European Review of Agricultural Economics",
  ["europ rev econ hist"] = "European Review of Economic History",
  ["europa archiv"] = "Europa-Archiv",
  ["european consort math indust"] = "European Consortium for Mathematics in Industry",
  ["european j appl math"] = "European Journal of Applied Mathematics",
  ["european j clin med oncol"] = "European journal of clinical & medical oncology",
  ["european j combin"] = "European Journal of Combinatorics",
  ["european j gen dent"] = "European journal of general dentistry",
  ["european j mech a solids"] = "European Journal of Mechanics",
  ["european j mech b fluids"] = "European Journal of Mechanics",
  ["european j med plants"] = "European journal of medicinal plants",
  ["european j oper res"] = "European Journal of Operational Research",
  ["european j org chem"] = "European journal of organic chemistry",
  ["european j pediatr surg rep"] = "European journal of pediatric surgery reports",
  ["european j phys"] = "European Journal of Physics",
  ["european univ inst ser d: econom"] = "European University Institute. Series D: Economics",
  ["europhys lett"] = "Europhysics Letters",
  ["europhys news"] = "Europhysics News",
  ["europäische hochschulschrift reihe v volks- betriebswirtsch"] = "Europäische Hochschulschriften",
  ["europäische hochschulschrift reihe xx philos"] = "Europäische Hochschulschriften",
  ["eut report-wsk"] = "EUT Report-WSK",
  ["euthan rev"] = "The Euthanasia review",
  ["euthanasia news"] = "Euthanasia news",
  ["eval health prof"] = "Evaluation & the health professions",
  ["eval j australas"] = "Evaluation journal of Australasia",
  ["eval pract"] = "Evaluation practice",
  ["eval program plann"] = "Evaluation and program planning",
  ["eval q"] = "Evaluation quarterly",
  ["eval rev"] = "Evaluation review",
  ["evaluation (lond)"] = "Evaluation (London, England : 1995)",
  ["evening sun"] = "Evening sun (Baltimore, Md.)",
  ["everybody's health"] = "Everybody's health",
  ["evid action int cochrane colloq 8th 2000 cape town s afr"] = "Evidence for action : challenges for the Cochrane Collaboration in the 21st century : abstract book. International Cochrane Colloquium (8th : 2000 : Cape Town, South Africa)",
  ["evid based cardiovasc med"] = "Evidence-based cardiovascular medicine",
  ["evid based child health"] = "Evidence-based child health : a Cochrane review journal",
  ["evid based commun assess interv"] = "Evidence-based communication assessment and intervention",
  ["evid based complement alternat med"] = "Evidence-based complementary and alternative medicine : eCAM",
  ["evid based dent"] = "Evidence-based dentistry",
  ["evid based libr inf pract"] = "Evidence based library and information practice",
  ["evid based med"] = "Evidence-based medicine",
  ["evid based med public health"] = "Evidence-based medicine & public health",
  ["evid based ment health"] = "Evidence-based mental health",
  ["evid based nurs"] = "Evidence-based nursing",
  ["evid based pract child adolesc ment health"] = "Evidence-based practice in child and adolescent mental health",
  ["evid based preclin med"] = "Evidence-based preclinical medicine",
  ["evid based spine care j"] = "Evidence-based spine-care journal",
  ["evid policy"] = "Evidence & policy : a journal of research, debate and practice",
  ["evid rep technol assess (full rep)"] = "Evidence report/technology assessment",
  ["evid rep technol assess (summ)"] = "Evidence report/technology assessment (Summary)",
  ["evidence-based integr med"] = "Evidence-Based Integrative Medicine",
  ["evidence-based toxicol"] = "Evidence-Based Toxicology",
  ["evol anthropol"] = "Evolutionary anthropology",
  ["evol anthropol: news issues rev"] = "Evolutionary Anthropology: Issues, News, and Reviews",
  ["evol appl"] = "Evolutionary applications",
  ["evol behav sci"] = "Evolutionary behavioral sciences",
  ["evol bioinf"] = "Evolutionary Bioinformatics",
  ["evol bioinform online"] = "Evolutionary bioinformatics online",
  ["evol biol"] = "Evolutionary biology",
  ["evol comput"] = "Evolutionary computation",
  ["evol comput mach learn data min bioinform"] = "Evolutionary computation, machine learning and data mining in bioinformatics. EvoBIO (Conference)",
  ["evol dev"] = "Evolution & development",
  ["evol ecol"] = "Evolutionary ecology",
  ["evol ecol res"] = "Evolutionary ecology research",
  ["evol econ soc complex sci"] = "Evolutionary Economics and Social Complexity Science",
  ["evol equ control theory"] = "Evolution Equations and Control Theory",
  ["evol equations control theory"] = "Evolution Equations and Control Theory",
  ["evol hum behav"] = "Evolution and human behavior : official journal of the Human Behavior and Evolution Society",
  ["evol hum sci"] = "Evolutionary human sciences",
  ["evol intell"] = "Evolutionary intelligence",
  ["evol j linn soc"] = "Evolutionary Journal of the Linnean Society",
  ["evol lett"] = "Evolution letters",
  ["evol med"] = "L' Evolution médicale",
  ["evol med public health"] = "Evolution, medicine, and public health",
  ["evol psychiatr (paris)"] = "L' Evolution psychiatrique",
  ["evol psychol"] = "Evolutionary psychology : an international journal of evolutionary approaches to psychology and behavior",
  ["evol psychol sci"] = "Evolutionary psychological science",
  ["evol stud imaginative cult"] = "Evolutionary studies in imaginative culture",
  ["evol syst"] = "Evolutionary systematics",
  ["evol: educ outreach"] = "Evolution: Education and Outreach",
  ["evolution (n y)"] = "Evolution",
  ["evolution int j org evolution"] = "Evolution; international journal of organic evolution",
  ["evraziat entomol z"] = "Evraziatskiĭ ėntomologicheskiĭ zhurnal = Euroasian entomological journal",
  ["evraziiskii soiuz uchenykh"] = "Evraziĭskiĭ soi︠u︡z uchenykh",
  ["ex counterex"] = "Examples and Counterexamples",
  ["ex countex"] = "Examples and Counterexamples",
  ["examples counterexamples"] = "Examples and Counterexamples",
  ["except child"] = "Exceptional children",
  ["excerpta med"] = "Excerpta medica",
  ["excerpta med (dermatol)"] = "Excerpta medica. Section 13, Dermatology and venereology",
  ["excerpta med (ophthalmol)"] = "Excerpta medica. Section 12, Ophthalmology",
  ["excerpta med (urol)"] = "Excerpta medica. Section 28, Urology and nephrology",
  ["excerpta med 9 surg"] = "Excerpta medica. Section 9, Surgery",
  ["excerpta medica 15 chest dis"] = "Excerpta medica. Section 15, Chest diseases",
  ["excerpta medica 20 gerontol geriatr"] = "Excerpta medica. Section 20, Gerontology and geriatrics",
  ["excli j"] = "EXCLI journal",
  ["excursions rec math ser"] = "Excursions in Recreational Mathematics Series",
  ["exec housekeep today"] = "Executive housekeeping today",
  ["exec housekeeper"] = "Executive housekeeper",
  ["exec intell rev"] = "Executive intelligence review",
  ["exec solut healthc manag"] = "Executive solutions for healthcare management",
  ["exerc immunol rev"] = "Exercise immunology review",
  ["exerc sport sci rev"] = "Exercise and sport sciences reviews",
  ["exercise sport sci rev"] = "Exercise and Sport Sciences Reviews",
  ["exercise, sport, movement"] = "Exercise, Sport, and Movement",
  ["exeter papers econ hist"] = "Exeter papers in economic history",
  ["exhib guide wellcome hist med libr"] = "Exhibition guide. Wellcome Historical Medical Library",
  ["exobiol future mars missions"] = "Exobiology and future Mars missions",
  ["exosomes microvesicles"] = "Exosomes and microvesicles",
  ["exp aging res"] = "Experimental aging research",
  ["exp agric"] = "Experimental Agriculture",
  ["exp anal hum behav bull"] = "Experimental analysis of human behavior bulletin [electronic resource]",
  ["exp anim"] = "Experimental animals",
  ["exp appl acarol"] = "Experimental & applied acarology",
  ["exp astron"] = "Experimental Astronomy",
  ["exp astron (dordr)"] = "Experimental astronomy",
  ["exp biol"] = "Experimental biology",
  ["exp biol med"] = "Experimental biology and medicine",
  ["exp biol med (london, u k)"] = "Experimental Biology and Medicine (London, United Kingdom)",
  ["exp biol med (maywood)"] = "Experimental biology and medicine (Maywood, N.J.)",
  ["exp biol med (maywood, nj, u s)"] = "Experimental Biology and Medicine (Maywood, NJ, United States)",
  ["exp biol online"] = "Experimental Biology Online",
  ["exp bot"] = "Experimental Botany",
  ["exp brain res"] = "Experimental brain research",
  ["exp cell biol"] = "Experimental cell biology",
  ["exp cell res"] = "Experimental Cell Research",
  ["exp cell res suppl"] = "Experimental cell research. Supplement",
  ["exp clin cardiol"] = "Experimental and clinical cardiology",
  ["exp clin endocrinol"] = "Experimental and clinical endocrinology",
  ["exp clin endocrinol diabetes"] = "Experimental and clinical endocrinology & diabetes : official journal, German Society of Endocrinology [and] German Diabetes Association",
  ["exp clin immunogenet"] = "Experimental and clinical immunogenetics",
  ["exp clin psychopharmacol"] = "Experimental and clinical psychopharmacology",
  ["exp clin transplant"] = "Experimental and clinical transplantation : official journal of the Middle East Society for Organ Transplantation",
  ["exp comput multiph flow"] = "Experimental and computational multiphase flow",
  ["exp comput multiphase flow"] = "Experimental and Computational Multiphase Flow",
  ["exp dermatol"] = "Experimental dermatology",
  ["exp diabesity res"] = "Experimental diabesity research",
  ["exp diabetes res"] = "Experimental diabetes research",
  ["exp econ"] = "Experimental economics",
  ["exp embryol teratol"] = "Experimental embryology and teratology",
  ["exp eye res"] = "Experimental eye research",
  ["exp fluids"] = "Experiments in Fluids",
  ["exp gerontol"] = "Experimental gerontology",
  ["exp heat transfer"] = "Experimental Heat Transfer",
  ["exp hematol"] = "Experimental Hematology",
  ["exp hematol oncol"] = "Experimental hematology & oncology",
  ["exp ir meets multilinguality multimodality interact (2017)"] = "Experimental IR meets multilinguality, multimodality, and interaction : 8th International Conference of the CLEF Association, CLEF 2017, Dublin, Ireland, September 11-14, 2017, Proceedings. Cross-Language Evaluation Forum. Conference (8th : 2017 : Dublin, Ireland)",
  ["exp lung res"] = "Experimental lung research",
  ["exp math"] = "Experimental Mathematics",
  ["exp mech"] = "Experimental Mechanics",
  ["exp med"] = "Experimental medicine",
  ["exp med pathol klin"] = "Experimentelle Medizin, Pathologie und Klinik",
  ["exp med surg"] = "Experimental medicine and surgery",
  ["exp mol med"] = "Experimental & molecular medicine",
  ["exp mol pathol"] = "Experimental and Molecular Pathology",
  ["exp mol pathol suppl"] = "Experimental and molecular pathology. Supplement",
  ["exp mycol"] = "Experimental Mycology",
  ["exp nephrol"] = "Experimental nephrology",
  ["exp neurobiol"] = "Experimental neurobiology",
  ["exp neurol"] = "Experimental neurology",
  ["exp neurol suppl"] = "Experimental neurology. Supplement",
  ["exp oncol"] = "Experimental oncology",
  ["exp parasitol"] = "Experimental parasitology",
  ["exp pathol"] = "Experimental pathology",
  ["exp pathol (jena)"] = "Experimentelle Pathologie",
  ["exp pathol suppl"] = "Experimentelle Pathologie. Supplement = Experimental pathology. Supplement",
  ["exp pathol suppl (1981)"] = "Experimental pathology. Supplement",
  ["exp physiol"] = "Experimental physiology",
  ["exp psychol"] = "Experimental psychology",
  ["exp results"] = "Experimental results",
  ["exp strahlenther klin strahlenbiol"] = "Experimentelle Strahlentherapie und Klinische Strahlenbiologie",
  ["exp stud psychol pedagog"] = "Experimental studies in psychology and pedagogy",
  ["exp suppl"] = "Experientia supplementum (2012)",
  ["exp tech"] = "Experimental Techniques",
  ["exp tech phys"] = "Experimentelle Technik der Physik",
  ["exp ther med"] = "Experimental and therapeutic medicine",
  ["exp therm fluid sci"] = "Experimental Thermal and Fluid Science",
  ["exp toxicol pathol"] = "Experimental and Toxicologic Pathology",
  ["exp transl stroke med"] = "Experimental & translational stroke medicine",
  ["exper econ"] = "Experimental Economics",
  ["experientia suppl"] = "Experientia. Supplementum",
  ["experiment math"] = "Experimental Mathematics",
  ["expert opin biol ther"] = "Expert opinion on biological therapy",
  ["expert opin drug deliv"] = "Expert opinion on drug delivery",
  ["expert opin drug delivery"] = "Expert Opinion on Drug Delivery",
  ["expert opin drug discov"] = "Expert opinion on drug discovery",
  ["expert opin drug discovery"] = "Expert Opinion on Drug Discovery",
  ["expert opin drug metab toxicol"] = "Expert opinion on drug metabolism & toxicology",
  ["expert opin drug metabol toxicol"] = "Expert Opinion on Drug Metabolsim & Toxicology",
  ["expert opin drug saf"] = "Expert opinion on drug safety",
  ["expert opin emerg drugs"] = "Expert opinion on emerging drugs",
  ["expert opin emerging drugs"] = "Expert Opinion on Emerging Drugs",
  ["expert opin environ biol"] = "Expert opinion on environmental biology",
  ["expert opin invest drugs"] = "Expert Opinion on Investigational Drugs",
  ["expert opin investig drugs"] = "Expert opinion on investigational drugs",
  ["expert opin med diagn"] = "Expert opinion on medical diagnostics",
  ["expert opin orphan drugs"] = "Expert Opinion on Orphan Drugs",
  ["expert opin pharmacother"] = "Expert opinion on pharmacotherapy",
  ["expert opin ther pat"] = "Expert opinion on therapeutic patents",
  ["expert opin ther targets"] = "Expert opinion on therapeutic targets",
  ["expert rev anti infect ther"] = "Expert review of anti-infective therapy",
  ["expert rev anti-infect ther"] = "Expert Review of Anti-infective Therapy",
  ["expert rev anticancer ther"] = "Expert review of anticancer therapy",
  ["expert rev cardiovasc ther"] = "Expert review of cardiovascular therapy",
  ["expert rev clin immunol"] = "Expert review of clinical immunology",
  ["expert rev clin pharmacol"] = "Expert review of clinical pharmacology",
  ["expert rev dermatol"] = "Expert review of dermatology",
  ["expert rev endocrinol metab"] = "Expert review of endocrinology & metabolism",
  ["expert rev gastroenterol hepatol"] = "Expert review of gastroenterology & hepatology",
  ["expert rev hematol"] = "Expert review of hematology",
  ["expert rev med devices"] = "Expert review of medical devices",
  ["expert rev mol diagn"] = "Expert review of molecular diagnostics",
  ["expert rev mol med"] = "Expert reviews in molecular medicine",
  ["expert rev neurother"] = "Expert review of neurotherapeutics",
  ["expert rev obstet gynecol"] = "Expert review of obstetrics & gynecology",
  ["expert rev ophthalmol"] = "Expert review of ophthalmology",
  ["expert rev pharmacoecon outcomes res"] = "Expert review of pharmacoeconomics & outcomes research",
  ["expert rev pharmacoeconomics outcomes res"] = "Expert Review of Pharmacoeconomics & Outcomes Research",
  ["expert rev precis med drug dev"] = "Expert review of precision medicine and drug development",
  ["expert rev proteomics"] = "Expert review of proteomics",
  ["expert rev qual life cancer care"] = "Expert review of quality of life in cancer care",
  ["expert rev respir med"] = "Expert review of respiratory medicine",
  ["expert rev vaccines"] = "Expert review of vaccines",
  ["expert syst"] = "Expert systems",
  ["expert syst appl"] = "Expert Systems with Applications",
  ["expert syst appl: x"] = "Expert Systems with Applications: X",
  ["explicit description language automat process text"] = "Explicit Description of Language and Automatic Processing of Text",
  ["explor complex"] = "Exploring Complexity",
  ["explor early am culture"] = "Explorations in early American culture",
  ["explor econ hist"] = "Explorations in economic history",
  ["explor ethn stud"] = "Explorations in ethnic studies : the journal of the National Association of Interdisciplinary Ethnic Studies",
  ["explor geophys"] = "Exploration Geophysics",
  ["explor j"] = "Explorers journal",
  ["explor knowl"] = "Explorations in knowledge",
  ["explor med"] = "Exploration of medicine",
  ["explor min geol"] = "Exploration and Mining Geology",
  ["explor renaiss cult"] = "Explorations in Renaissance culture",
  ["explor res hypothesis med"] = "Exploratory research and hypothesis in medicine",
  ["explor sights sounds"] = "Explorations in sights and sounds",
  ["explor target antitumor ther"] = "Exploration of targeted anti-tumor therapy",
  ["explor: cours contrib sci ed"] = "Exploration: Cours et Contributions pour les Sciences de l’Education",
  ["explorations econ hist"] = "Explorations in Economic History",
  ["explore (ny)"] = "Explore (New York, N.Y.)",
  ["explorer (hayward)"] = "Explorer",
  ["explorer (kansas city)"] = "The Explorer : the journal of the School of Dentistry of the University of Missouri at Kansas City. University of Missouri at Kansas City. School of Dentistry",
  ["expo health"] = "Exposure and health",
  ["expo math"] = "Expositiones mathematicae",
  ["expo times"] = "The Expository times",
  ["expos annu biochim med"] = "Exposés annuels de biochimie médicale",
  ["exposition math"] = "Expositiones Mathematicae",
  ["expositiones math"] = "Expositiones Mathematicae",
  ["exposure health"] = "Exposure and Health",
  ["express polym lett"] = "EXPRESS polymer letters",
  ["ext abstr hum factors computing syst"] = "Extended abstracts on Human factors in computing systems. CHI Conference",
  ["extr ind soc"] = "Extractive Industries and Society",
  ["extracell vesicles circ nucl acids"] = "Extracellular vesicles and circulating nucleic acids",
  ["extracta math"] = "Extracta Mathematicae",
  ["extrem physiol med"] = "Extreme physiology & medicine",
  ["extreme life biospeology astrobiol"] = "Extreme life, biospeology & astrobiology",
  ["extreme mech lett"] = "Extreme Mechanics Letters",
  ["extreme physiol med"] = "Extreme Physiology & Medicine",
  ["extremes (boston)"] = "Extremes",
  ["exxon monogr"] = "Exxon Monographs",
  ["eye (lond)"] = "Eye (London, England)",
  ["eye brain"] = "Eye and brain",
  ["eye contact lens"] = "Eye & contact lens",
  ["eye ear nose throat mon"] = "Eye, ear, nose & throat monthly",
  ["eye rep"] = "Eye reports",
  ["eye sci"] = "Eye science",
  ["eye vis (lond)"] = "Eye and vision (London, England)",
  ["eye vision"] = "Eye and Vision",
  ["f s rep"] = "F&S reports",
  ["f s rev"] = "F&S reviews",
  ["f s sci"] = "F&S science",
  ["f1000 biol rep"] = "F1000 biology reports",
  ["f1000 fac rev"] = "F1000 faculty reviews",
  ["f1000 med rep"] = "F1000 medicine reports",
  ["f1000prime rep"] = "F1000prime reports",
  ["fa i hsueh tsa chih"] = "Fa I Hsueh Tsa Chih Journal of Forensic Medicine (Shang-Hai",
  ["fa yi xue za zhi"] = "Fa yi xue za zhi",
  ["fac cathol lille"] = "Annuaire - Facultés catholiques de Lille (1972)",
  ["fac notes (new orleans la)"] = "Faculty notes (New Orleans, La.)",
  ["fac rev"] = "Faculty reviews",
  ["facets (ott)"] = "Facets (Ottawa)",
  ["fachber simul"] = "Fachberichte Simulation",
  ["fachbl schweiz anst"] = "Fachblatt für schweizerisches Anstaltswesen. Revue suisse des établissements hospitaliers",
  ["fachbl schweiz heime anst rev suisse etabl hosp educ"] = "Fachblatt für schweizerische Heime und Anstalten. Revue suisse des établissements hospitaliers et d'éducation",
  ["fachbuchreihe stud fortbild prax"] = "Fachbuchreihe fur Studium Fortbildung Praxis",
  ["facial orthop temporomandibular arthrol"] = "Facial orthopedics and temporomandibular arthrology",
  ["facial plast surg"] = "Facial plastic surgery : FPS",
  ["facial plast surg aesthet med"] = "Facial plastic surgery & aesthetic medicine",
  ["facial plast surg clin north am"] = "Facial plastic surgery clinics of North America",
  ["fact sheet (cent home care policy res)"] = "Fact sheet (Center for Home Care Policy and Research (U.S.))",
  ["fact sheets swed"] = "Fact sheets on Sweden",
  ["facta univ - ser electron energ"] = "Facta universitatis - series: Electronics and Energetics",
  ["facta univ ser autom control robot"] = "Facta Universitatis. Series Automatic Control and Robotics",
  ["facta univ ser math inform"] = "Facta Universitatis. Series: Mathematics and Informatics",
  ["facta univ ser mech autom control robot"] = "Facta universitatis. Series, Mechanics, automatic control and robotics",
  ["facta univ ser mech automat control robot"] = "University of Niš",
  ["factor odontol"] = "Factor odontológico",
  ["facts infant feed"] = "Facts about infant feeding",
  ["facts views vis obgyn"] = "Facts, views & vision in ObGyn",
  ["fag tidsskr sykepleien"] = "Fag tidsskriftet sykepleien",
  ["fait tend"] = "Faits et tendances",
  ["faith philos"] = "Faith and philosophy : journal of the Society of Christian Philosophers",
  ["faits lang"] = "Faits de langues",
  ["fam cancer"] = "Familial cancer",
  ["fam circle"] = "Family circle (Mount Morris, Ill.)",
  ["fam community health"] = "Family & community health",
  ["fam community hist"] = "Family & community history : journal of the Family and Community Historical Research Society",
  ["fam concil courts rev"] = "Family and conciliation courts review",
  ["fam consum sci res j"] = "Family and consumer sciences research journal",
  ["fam coord"] = "Family coordinator",
  ["fam court rev"] = "Family court review",
  ["fam dev"] = "Famille et développement",
  ["fam econ rev"] = "Family economics review",
  ["fam health"] = "Family health",
  ["fam j alex va"] = "Family journal (Alexandria, Va.)",
  ["fam law q"] = "Family law quarterly",
  ["fam law rep"] = "Family law reports",
  ["fam law rep (wash d c)"] = "The Family law reporter",
  ["fam law rev"] = "Family law review",
  ["fam life coord"] = "The Family life coordinator",
  ["fam life educ"] = "Family life educator",
  ["fam life matters"] = "Family life matters",
  ["fam matters"] = "Family matters (Melbourne, Vic.)",
  ["fam med"] = "Family medicine",
  ["fam med community health"] = "Family medicine and community health",
  ["fam med med sci res"] = "Family medicine & medical science research",
  ["fam med prim care open access"] = "Family medicine and primary care -- open access",
  ["fam plan manag"] = "Family planning manager",
  ["fam plann"] = "Family Planning",
  ["fam plann (palo alto)"] = "The Family planner",
  ["fam plann dig"] = "Family planning digest",
  ["fam plann inf serv"] = "Family planning information service",
  ["fam plann news (tokyo)"] = "Family planning news",
  ["fam plann perspect"] = "Family planning perspectives",
  ["fam plann popul rep"] = "Family planning/population reporter; a review of State laws and policies",
  ["fam plann q"] = "Family planning quarterly",
  ["fam plann resume"] = "Family planning résumé",
  ["fam plann today"] = "Family planning today",
  ["fam pract"] = "Family practice",
  ["fam pract manag"] = "Family practice management",
  ["fam pract news"] = "Family practice news",
  ["fam pract res j"] = "Family practice research journal",
  ["fam process"] = "Family process",
  ["fam psychol"] = "The Family psychologist : bulletin of the Division of Family Psychology (43)",
  ["fam relat"] = "Family relations",
  ["fam relatsh soc"] = "Families, relationships and societies : in international journal of research and debate",
  ["fam sci"] = "Family science",
  ["fam soc"] = "Families in society : the journal of contemporary human services",
  ["fam syst health"] = "Families, systems & health : the journal of collaborative family healthcare",
  ["fam syst med"] = "Family systems medicine",
  ["fam ther"] = "Family therapy",
  ["fam urol"] = "Family Urology",
  ["fam viol prev health pract"] = "Family violence prevention and health practice : an e-journal of the Family Violence Prevention Fund",
  ["families child health"] = "Families and child health",
  ["family econ nutrition rev"] = "Family Economics and Nutrition Review",
  ["family econ rev"] = "Family Economics Review",
  ["family law"] = "Family law (Chichester)",
  ["family plan world"] = "Family planning world",
  ["famplanco news"] = "Famplanco news",
  ["fao fish circ"] = "FAO fisheries circular = FAO circulaire sur les pêches = FAO circular de pesca",
  ["fao food nutr pap"] = "FAO food and nutrition paper",
  ["fao food nutr ser"] = "FAO food and nutrition series",
  ["fao for dev pap"] = "FAO Forestry Development Paper",
  ["fao nutr meet rep ser"] = "FAO nutrition meetings report series",
  ["fao nutr stud"] = "FAO nutritional studies",
  ["far east aff"] = "Far Eastern affairs",
  ["far east econ rev"] = "Far Eastern economic review",
  ["far east entomol"] = "Far Eastern entomologist = Dalʹnevostochnyĭ ėntomolog",
  ["far east j appl math"] = "Far east journal of applied mathematics",
  ["far east j math sci"] = "Far East Journal of Mathematical Sciences",
  ["faraday discuss"] = "Faraday Discussions",
  ["faraday discuss chem soc"] = "Faraday Discussions of the Chemical Society",
  ["faraday discuss r soc chem"] = "Faraday Discussions of the Royal Society of Chemistry",
  ["faraday symp chem soc"] = "Faraday Symposium of the Chemical Society",
  ["faraday symp r soc chem"] = "Faraday Symposium of the Royal Society of Chemistry",
  ["faraday trans"] = "Faraday Transactions",
  ["farday spec discuss chem soc"] = "Faraday Special Discussions of the Chemical Society",
  ["farm clin"] = "Farmacia clínica",
  ["farm glas"] = "Farmaceutski glasnik",
  ["farm hosp"] = "Farmacia hospitalaria : órgano oficial de expresión científica de la Sociedad Española de Farmacia Hospitalaria",
  ["farm nueva"] = "Farmacia nueva",
  ["farm obz"] = "Farmaceutický obzor",
  ["farm pol"] = "Farmacja polska",
  ["farm quim"] = "Farmacia y química",
  ["farm revy"] = "Farmacevtisk revy",
  ["farm tid"] = "Farmaceutisk tidende",
  ["farm vestn"] = "Farmacevtski vestnik",
  ["farm zh"] = "Farmatsevtychnyĭ zhurnal",
  ["farmacihist sallsk ars"] = "Farmacihistoriska sallskapets °arsskrift",
  ["farmaco [prat]"] = "Farmaco. Edizione Pratica",
  ["farmaco [sci]"] = "Farmaco. Edizione Scientifica",
  ["farmaco prat"] = "Il Farmaco; edizione pratica",
  ["farmaco sci"] = "Il Farmaco; edizione scientifica",
  ["farmaco sci tec"] = "Il Farmaco, scienza e tecnica",
  ["farmacoter actual"] = "Farmacoterapia actual",
  ["farmakol toksikol"] = "Farmakologiia i toksikologiia",
  ["farmatsiia (mosk)"] = "Farmatsiia",
  ["farmatsiia (sofia)"] = "Farmatsiia",
  ["farmatsiya (moscow)"] = "Farmatsiya (Moscow)",
  ["farriers mag"] = "The Farrier's magazine : or, the archives of veterinary science",
  ["fas prof bull"] = "F.A.S. professional bulletin",
  ["fas public interest rep"] = "F.A.S. public interest report",
  ["fasc math"] = "Polytechnica Posnaniensis. Institutum Mathematicum. Fasciculi Mathematici",
  ["faseb bioadv"] = "FASEB bioAdvances",
  ["faseb j"] = "FASEB Journal",
  ["fashion text"] = "Fashion and Textiles",
  ["faslnamahi kumish"] = "Faṣlnāmah-i kūmish",
  ["faslnamahi takhassusii ulumi zisti"] = "Faṣlnāmah-i takhaṣṣuṣī-í ulūm-i zīstī",
  ["fat science technology|fett wiss technol"] = "Fett Wissenschaft Technologie : Organ der Deutschen Gesellschaft fur Fettwissenschaft e.V.",
  ["fatigue fract eng mater struct"] = "Fatigue and Fracture of Engineering Materials and Structures",
  ["faulkner grays med health"] = "Faulkner & Gray's medicine & health",
  ["fauna flora"] = "Fauna och flora",
  ["fauna n z"] = "Fauna of New Zealand",
  ["fauna norv"] = "Fauna norvegica",
  ["fauna ryukyuana"] = "Fauna Ryukyuana",
  ["fave secc cienc vet"] = "FAVE. Seccion ciencias veterinarias",
  ["fda consum"] = "FDA consumer",
  ["fda drug bull"] = "FDA drug bulletin",
  ["fda med bull"] = "FDA medical bulletin : important information for health professionals from the U.S. Food & Drug Administration",
  ["fdc rep drugs cosmet"] = "FDC reports. Drugs and cosmetics",
  ["fdi world"] = "FDI world",
  ["fdic banking rev"] = "FDIC Banking Review",
  ["fdlis food drug policy forum"] = "FDLI's food and drug policy forum",
  ["featur rev math rev"] = "Featured Reviews in Mathematical Reviews",
  ["febs j"] = "The FEBS journal",
  ["febs lett"] = "FEBS Letters",
  ["febs open bio"] = "FEBS open bio",
  ["fed bar j"] = "Federal bar journal : the official publication of the Federal Bar Association",
  ["fed bull"] = "Federation bulletin",
  ["fed court rep"] = "The Federal Court reports. Australia. Federal Court",
  ["fed eur biochem soc lett"] = "Federation of European Biochemical Societies Letters",
  ["fed insur corp couns q"] = "Federation of Insurance & Corporate Counsel quarterly",
  ["fed insur couns q"] = "Federation of Insurance Counsel quarterly",
  ["fed law rep"] = "The federal law reports : being reports of cases decided by the federal courts (other than the high court), state courts exercising federal jurisdiction, and courts of territories",
  ["fed oper dent"] = "Federation of operative dentistry",
  ["fed pract"] = "Federal practitioner : for the health care professionals of the VA, DoD, and PHS",
  ["fed probat"] = "Federal probation",
  ["fed proc"] = "Federation proceedings",
  ["fed proc transl suppl"] = "Federation proceedings. Translation supplement; selected translations from medical-related science",
  ["fed regist"] = "Federal Register",
  ["fed report"] = "The Federal reporter",
  ["fed reserve bank atlanta econ rev"] = "Federal Reserve Bank of Atlanta Economic Review",
  ["fed reserve bank atlanta rev"] = "Federal Reserve Bank of Atlanta Monthly Review",
  ["fed reserve bank boston new eng econ rev"] = "Federal Reserve Bank of Boston New England Economic Review",
  ["fed reserve bank bus rev phila"] = "Federal Reserve Bank Business Review of Philadelphia",
  ["fed reserve bank chicago econ perspect"] = "Federal Reserve Bank of Chicago Economic Perspectives",
  ["fed reserve bank cleveland econ rev"] = "Federal Reserve Bank of Cleveland Economic Review",
  ["fed reserve bank dallas econ finan rev"] = "Federal Reserve Bank of Dallas Economic and Financial Review",
  ["fed reserve bank dallas econ rev"] = "Federal Reserve Bank of Dallas Economic Review",
  ["fed reserve bank kansas city econ rev"] = "Federal Reserve Bank of Kansas City Economic Review",
  ["fed reserve bank kansas city rev"] = "Federal Reserve Bank of Kansas City Monthly Review",
  ["fed reserve bank minneapolis quart rev"] = "Federal Reserve Bank of Minneapolis Quarterly Review",
  ["fed reserve bank new york econ pol rev"] = "Federal Reserve Bank of New York Economic Policy Review",
  ["fed reserve bank new york quart rev"] = "Federal Reserve Bank of New York Quarterly Review",
  ["fed reserve bank new york rev"] = "Federal Reserve Bank of New York Monthly Review",
  ["fed reserve bank philadelphia bus rev"] = "Federal Reserve Bank of Philadelphia Business Review",
  ["fed reserve bank richmond econ quart"] = "Federal Reserve Bank of Richmond Economic Quarterly",
  ["fed reserve bank richmond econ rev"] = "Federal Reserve Bank of Richmond Economic Review",
  ["fed reserve bank san francisco econ rev"] = "Federal Reserve Bank of San Francisco Economic Review",
  ["fed reserve bank san francisco rev"] = "Federal Reserve Bank of San Francisco Monthly Review",
  ["fed reserve bank st louis rev"] = "Federal Reserve Bank of St. Louis Review",
  ["fed reserve bull"] = "Federal Reserve Bulletin",
  ["fed rules decis"] = "Federal rules decisions",
  ["fed suppl"] = "Federal supplement",
  ["feddes repert"] = "Feddes repertorium",
  ["feddes repertorium"] = "Feddes Repertorium",
  ["feit fict"] = "Feit & fictie : tijdschrift voor de geschiedenis van de representatie",
  ["feldsher akush"] = "Felʹdsher i akusherka",
  ["feline pract"] = "Feline practice",
  ["fem criminol"] = "Feminist criminology",
  ["fem econ"] = "Feminist economics",
  ["fem issues"] = "Feminist issues",
  ["fem leg stud"] = "Feminist legal studies",
  ["fem media stud"] = "Feminist media studies",
  ["fem psychol"] = "Feminism & psychology",
  ["fem rev"] = "Feminist review",
  ["fem stud"] = "Feminist studies : FS",
  ["female health top diagn report"] = "Female health topics & diagnostic reporter",
  ["female patient"] = "Female patient",
  ["female patient (parsippany)"] = "The female patient",
  ["female patient (prim care ed)"] = "Female patient (Primary care ed.)",
  ["female pelvic med reconstr surg"] = "Female pelvic medicine & reconstructive surgery",
  ["feminist econ"] = "Feminist Economics",
  ["femnet news"] = "FEMNET news",
  ["fems immunol med microbiol"] = "FEMS immunology and medical microbiology",
  ["fems microbes"] = "FEMS Microbes",
  ["fems microbiol ecol"] = "FEMS microbiology ecology",
  ["fems microbiol immunol"] = "FEMS microbiology immunology",
  ["fems microbiol lett"] = "FEMS microbiology letters",
  ["fems microbiol rev"] = "FEMS microbiology reviews",
  ["fems yeast res"] = "FEMS yeast research",
  ["fen derg"] = "Fen Dergisi. Journal of Science",
  ["fen zi xi bao sheng wu xue bao"] = "Fen zi xi bao sheng wu xue bao = Journal of molecular cell biology",
  ["fen zi zhi wu yu zhong|mol plant breed"] = "Molecular plant breeding",
  ["fenetre cours"] = "Fenetre sur Cours",
  ["fenxi ceshi tongbao"] = "Fenxi Ceshi Tongbao",
  ["fenxi huaxue"] = "Fenxi Huaxue",
  ["fenxi shiyanshi"] = "Fenxi Shiyanshi",
  ["fenzi cuihua"] = "Fenzi Cuihua",
  ["ferment technol"] = "Fermentation technology",
  ["fermentation (basel)"] = "Fermentation (Basel, Switzerland)",
  ["fern gaz"] = "The Fern gazette",
  ["fernerkund photogramm"] = "Fernerkundung und Photogrammetrie",
  ["ferroelectr lett"] = "Ferroelectrics Letters",
  ["ferroelectr lett sect"] = "Ferroelectric Letters Section",
  ["ferroelectr rev"] = "Ferroelectrics Review",
  ["ferroelectr, lett sect"] = "Ferroelectrics, Letters Section",
  ["fertil contracept"] = "Fertility and contraception",
  ["fertil contracept sex"] = "Fertilité, contraception, sexualité",
  ["fertil determ res notes"] = "Fertility determinants research notes",
  ["fertil orthog"] = "Fertilité, orthogénie",
  ["fertil res pract"] = "Fertility research and practice",
  ["fertil steril"] = "Fertility and sterility",
  ["fetal diagn ther"] = "Fetal diagnosis and therapy",
  ["fetal infant ophthalmic med image anal (2017)"] = "Fetal, Infant and Ophthalmic Medical Image Analysis : International Workshop, FIFI 2017 and 4th International Workshop, OMIA 2017, Held in Conjunction with MICCAI 2017, Québec City, QC, Canada, September 14, 2017. FIFI (Workshop) (2017 : Québec, Québec)",
  ["fetal matern med rev"] = "Fetal and maternal medicine review",
  ["fetal pediatr pathol"] = "Fetal and pediatric pathology",
  ["fetal ther"] = "Fetal therapy",
  ["fett - lipid"] = "Fett - Lipid",
  ["fett wiss technol"] = "Fett Wissenschaft Technologie : Organ der Deutschen Gesellschaft für Fettwissenschaft e.V. = Fat science technology",
  ["fettchem umsch"] = "Fettchemische Umschau",
  ["fette seifen"] = "Fette und Seifen",
  ["feuill biol"] = "Feuillets de biologie",
  ["feuill prat"] = "Feuillets du praticien",
  ["few body syst"] = "Few-body systems",
  ["few-body syst"] = "Few-Body Systems",
  ["few-body syst, suppl"] = "Few-Body Systems, Supplementum",
  ["few-body systems"] = "Few-Body Systems",
  ["fias interdiscip sci ser"] = "FIAS Interdisciplinary Science Series",
  ["fiber integr opt"] = "Fiber and Integrated Optics",
  ["fibers (basel)"] = "Fibers (Basel, Switzerland)",
  ["fibers polym"] = "Fibers and Polymers",
  ["fibonacci quart"] = "The Fibonacci Quarterly. The Official Journal of the Fibonacci Association",
  ["fibre chem"] = "Fibre Chemistry",
  ["fibre sci technol"] = "Fibre Science and Technology",
  ["fibres text east eur"] = "Fibres & Textiles in Eastern Europe",
  ["fibrogenesis tissue repair"] = "Fibrogenesis & tissue repair",
  ["fich med ter puriss"] = "Fichero médico terapéutico",
  ["fides et hist"] = "Fides et historia : official publication of the Conference on Faith and History",
  ["field actions sci rep"] = "Field actions science reports",
  ["field anal chem technol"] = "Field Analytical Chemistry and Technology.",
  ["field crops res"] = "Field crops research",
  ["field exch"] = "Field exchange",
  ["field methods"] = "Field methods",
  ["field mycol"] = "Field mycology : a magazine for the study and identification of wild fungi",
  ["field staff rep"] = "Field staff reports",
  ["field stud"] = "Field Studies",
  ["fieldiana anthropol"] = "Fieldiana Anthropology",
  ["fieldiana bot"] = "Fieldiana. botany",
  ["fieldiana botan"] = "Fieldiana Botany",
  ["fieldiana geol"] = "Fieldiana Geology",
  ["fieldiana life earth sci"] = "Fieldiana. Life and earth sciences",
  ["fieldiana zool"] = "Fieldiana Zoology",
  ["fields inst commun"] = "Fields Institute Communications",
  ["fields inst monogr"] = "Fields Institute Monographs",
  ["fields math educ j"] = "Fields Mathematics Education Journal",
  ["fields quanta"] = "Fields and Quanta",
  ["fieldstaff rep"] = "Fieldstaff reports. Asia. Southeast Asia series",
  ["fieldstaff rep northeast afr ser"] = "Fieldstaff reports. Africa. Northeast Africa series",
  ["fieldstaff rep west eur ser"] = "Fieldstaff reports. West Europe series",
  ["fight against"] = "The Fight against disease",
  ["fiji med j"] = "Fiji medical journal",
  ["filaria j"] = "Filaria journal",
  ["film hist"] = "Film History",
  ["film int"] = "Film international",
  ["filosofia (london)"] = "Filosofia—Filosofia Contemporanea e Historia da Filosofia (London)",
  ["filson club hist q"] = "The Filson Club history quarterly",
  ["filtr ind anal"] = "Filtration Industry Analyst",
  ["filtr sep"] = "Filtration & Separation",
  ["fin lakaresallsk handl"] = "Finska läkaresällskapets handlingar",
  ["finan eng japanese markets"] = "Financial Engineering and the Japanese Markets",
  ["finan hist rev"] = "Financial History Review",
  ["finan industry stud"] = "Financial Industry Studies",
  ["finan manage"] = "Financial Management",
  ["finan markets"] = "Financial Markets, Institutions and Instruments",
  ["finan practice educ"] = "Financial Practice and Education",
  ["finan rev"] = "Financial Review",
  ["financ eng explain"] = "Financial Engineering Explained",
  ["financ manage"] = "Financial management",
  ["financ math"] = "Financial Mathematics",
  ["financ math appl"] = "Financial Mathematics and Applications",
  ["financ math fintech"] = "Financial Mathematics and Fintech",
  ["financ res lett"] = "Finance research letters",
  ["financ stat"] = "Financial Statistics",
  ["financ times"] = "Financial times (North American edition)",
  ["finance dev"] = "Finance & development",
  ["finance devel"] = "Finance and Development",
  ["finance india"] = "Finance India",
  ["finance prof"] = "Finance for Professionals",
  ["finance stoch"] = "Finance and Stochastics",
  ["finance stochastics"] = "Finance and Stochastics",
  ["finance trade rev"] = "Finance and Trade Review",
  ["finance úver"] = "Finance a Úver",
  ["find brief"] = "Findings brief : health care financing & organization",
  ["fine focus"] = "Fine focus",
  ["finite elem anal des"] = "Finite Elements in Analysis and Design",
  ["finite fields appl"] = "Finite Fields and Their Applications",
  ["finnish econ pap"] = "Finnish Economic Papers",
  ["fire (basel)"] = "Fire (Basel, Switzerland)",
  ["fire ecol"] = "Fire Ecology",
  ["fire j"] = "Fire journal (Boston, Mass.)",
  ["fire mater"] = "Fire and Materials",
  ["fire saf j"] = "Fire Safety Journal",
  ["fire saf sci"] = "Fire Safety Science",
  ["fire sci rev"] = "Fire Science Reviews",
  ["fire technol"] = "Fire Technology",
  ["first annu conf vaccine res conf vaccine res 1st 1998 wash d c"] = "The First Annual Conference on Vaccine Research : basic science, product development, clinical & field studies : final program and abstract book : May 30-June 1, 1998, Marriott at Metro Center, Washington, D.C. Conference on Vaccine Research (1st : 1998 : Washington, D.C.)",
  ["first annu ieee conf control technol appl (2017)"] = "First Annual IEEE Conference on Control Technology and Applications : CCTA 2017 : Kohala Coast, Hawai'i, August 27-30, 2017. IEEE Conference on Control Technology and Applications (1st : 2017 : Waimea, Hawaii Island, Hawaii)",
  ["first int press lect ser"] = "First International Press Lecture Series",
  ["first lang"] = "First language",
  ["first people"] = "The First people",
  ["first peoples child fam rev"] = "First peoples child & family review",
  ["first things"] = "First things (New York, N.Y.)",
  ["fisc stud"] = "Fiscal studies",
  ["fish aquat sci"] = "Fisheries and Aquatic Science",
  ["fish aquatic sci"] = "Fisheries and aquatic sciences",
  ["fish bull (wash d c)"] = "Fishery bulletin (Washington, D.C. : 1971)",
  ["fish fish"] = "Fish and Fisheries",
  ["fish fish (oxf)"] = "Fish and fisheries (Oxford, England)",
  ["fish manag ecol"] = "Fisheries management and ecology",
  ["fish manage ecol"] = "Fisheries Management and Ecology",
  ["fish oceanogr"] = "Fisheries oceanography",
  ["fish pathol"] = "Fish pathology",
  ["fish physiol biochem"] = "Fish physiology and biochemistry",
  ["fish res"] = "Fisheries research",
  ["fish sahul"] = "Fishes of Sahul : journal of the Australia New Guinea Fishes Association",
  ["fish sci"] = "Fisheries science : FS",
  ["fish sci (carlton, aust)"] = "Fisheries Science (Carlton, Australia)",
  ["fish sci (tokyo, jpn)"] = "Fisheries Science (Tokyo, Japan)",
  ["fish shellfish immunol"] = "Fish & shellfish immunology",
  ["fish shellfish immunol rep"] = "Fish and Shellfish Immunology Reports",
  ["fisheries (bethesda)"] = "Fisheries",
  ["fisiol med"] = "Fisiologia e medicina",
  ["fission prod inhal proj"] = "Fission product inhalation project [technical progress report]. Lovelace Foundation for Medical Education and Research",
  ["fitopatol bras"] = "Fitopatologia brasileira",
  ["fixed point theory"] = "Fixed Point Theory",
  ["fixed point theory algorithms sci eng"] = "Fixed Point Theory and Algorithms for Sciences and Engineering",
  ["fixed point theory appl"] = "Fixed Point Theory and Applications",
  ["fiz atmos okeana"] = "Rossíiskaya Akademiya Nauk",
  ["fiz elem chastits at yadra [sov j part nucl]"] = "Fizika Elementarnykh Chastits i Atomnogo Yadra [Soviet Journal of Particles and Nuclei]",
  ["fiz goreniya vzryva"] = "Fizika Goreniya i Vzryva",
  ["fiz khim obrab mater"] = "Fizika i Khimiya Obrabotki Materialov",
  ["fiz khim stekla"] = "Fizika i Khimiya Stekla",
  ["fiz met metalloved"] = "Fizika Metallov i Metallovedenie",
  ["fiz met metalloved [phys met metallogr (ussr)]"] = "Fizika Metallov i Metallovedenie [Physics of Metals and Metallography (USSR)]",
  ["fiz nizk temp (kiev)"] = "Fizika Nizkikh Temperatur (Kiev)",
  ["fiz nizk temp [sov j low temp phys]"] = "Fizika Nizkikh Temperatur [Soviet Journal of Low Temperature Physics]",
  ["fiz plazmy"] = "Fizika Plazmy",
  ["fiz plazmy (moscow)"] = "Fizika Plazmy (Moscow)",
  ["fiz plazmy [sov j plasma phys]"] = "Fizika Plazmy [Soviet Journal of Plasma Physics]",
  ["fiz tekh poluprovodn (leningrad)"] = "Fizika i Tekhnika Poluprovodnikov (S. -Peterburg)",
  ["fiz tekh poluprovodn [sov phys semicond]"] = "Fizika i Tekhnika Poluprovodnikov [Soviet Physics–Semiconductors]",
  ["fiz tverd tela (leningrad)"] = "Fizika Tverdogo Tela (S. -Peterburg)",
  ["fiz tverd tela (leningrad) [sov phys solid state]"] = "Fizika Tverdogo Tela (Leningrad) [Soviet Physics–Solid State]",
  ["fiz zemli"] = "Rossíiskaya Akademiya Nauk",
  ["fiz èlementar chastits i atom yadra"] = "Fizika Èlementarnykh Chastits i Atomnogo Yadra",
  ["fiz-khim mekh mater"] = "Fiziko-Khimicheskaya Mekhanika Materialov",
  ["fiz-mat nasled"] = "Fiziko-Matematicheskoe Nasledie",
  ["fiz-mat spis"] = "Fiziko-Matematichesko Spisanie",
  ["fiziol akt rechovini"] = "Fiziologichno Aktivni Rechovini",
  ["fiziol biokhim kul’t rast"] = "Fiziologiya i Biokhimiya Kul’turnykh Rastenii",
  ["fiziol cheloveka"] = "Fiziologiia cheloveka",
  ["fiziol norm patol"] = "Fiziologia normală şi patologică",
  ["fiziol rast (moscow)"] = "Fiziologiya Rastenii (Moscow)",
  ["fiziol zh"] = "Fiziologicheskiĭ zhurnal",
  ["fiziol zh im i m sechenova"] = "Fiziologicheskiĭ zhurnal imeni I.M. Sechenova",
  ["fiziol zh sssr im i m sechenova"] = "Fiziologicheskiĭ zhurnal SSSR imeni I. M. Sechenova",
  ["fj-lmi subser"] = "FJ-LMI Subseries",
  ["fla bar j"] = "The Florida Bar journal",
  ["fla dent j"] = "Florida dental journal",
  ["fla entomol"] = "The Florida entomologist",
  ["fla entomolog"] = "Florida Entomoloigist",
  ["fla geogr"] = "The Florida geographer",
  ["fla health notes"] = "Florida health notes",
  ["fla hist q"] = "The Florida historical quarterly",
  ["fla j environ health"] = "Florida journal of environmental health",
  ["fla j int law"] = "Florida journal of international law",
  ["fla law rev"] = "Florida law review",
  ["fla nurse"] = "The Florida nurse",
  ["fla optom"] = "Florida optometrist",
  ["fla public health rev"] = "Florida public health review",
  ["fla sci"] = "Florida Scientist",
  ["fla state univ law rev"] = "Florida State University law review. Florida State University. College of Law",
  ["fla suppl"] = "Florida supplement : decisions of lower courts of record and state commissions",
  ["flame retard therm stab mater"] = "Flame Retardancy and Thermal Stability of Materials",
  ["flash inf"] = "Flash-Informations",
  ["flatchem chem flat mater"] = "FlatChem Chemistry of Flat Materials",
  ["flavins flavoproteins"] = "Flavins and flavoproteins",
  ["flavour fragance j"] = "Flavour and Fragance Journal",
  ["flavour fragr j"] = "Flavour and fragrance journal",
  ["fletcher forum"] = "The Fletcher forum",
  ["fletcher forum world aff"] = "The Fletcher forum of world affairs",
  ["flex print electron"] = "Flexible and Printed Electronics",
  ["flex serv manuf j"] = "Flexible services and manufacturing journal",
  ["flexible printed electron"] = "Flexible and Printed Electronics",
  ["flexible serv manuf j"] = "Flexible Services and Manufacturing Journal",
  ["flight saf dig"] = "Flight safety digest (Arlington, Va. : 1988)",
  ["flinders j hist polit"] = "Flinders journal of history and politics",
  ["flintknappers exch"] = "Flintknappers' exchange",
  ["flora allg bot ztg"] = "Flora oder Allgemeine Botanische Zeitung",
  ["flora allg bot ztg abt a, physiol biochem"] = "Flora oder Allgemeine Botanische Zeitung. Abt. A, Physiologie und Biochemie",
  ["flora allg bot ztg abt b, morphol geobot"] = "Flora oder Allgemeine Botanische Zeitung. Abt. B, Morphologie und Geobotanik",
  ["florence nightingale j nurs"] = "Florence Nightingale journal of nursing",
  ["florist rdbr"] = "Floristische Rundbriefe",
  ["flow meas instrum"] = "Flow Measurement and Instrumentation",
  ["flow turbul combust"] = "Flow, Turbulence and Combustion",
  ["flow: appl fluid mech"] = "Flow: Applications of Fluid Mechanics",
  ["fluctuation noise lett"] = "Fluctuations and Noise Letters",
  ["fluid dyn"] = "Fluid Dynamics",
  ["fluid dyn mater process"] = "Fluid Dynamics & Materials Processing",
  ["fluid dyn res"] = "Fluid Dynamics Research",
  ["fluid dynam"] = "Fluid Dynamics",
  ["fluid dynam res"] = "Japan Society of Fluid Mechanics",
  ["fluid mech appl"] = "Fluid Mechanics and its Applications",
  ["fluid mech res"] = "Fluid Mechanics Research",
  ["fluid phase equilib"] = "Fluid Phase Equilibria",
  ["fluid/part sep j"] = "Fluid/Particle Separation Journal",
  ["fluids barriers cns"] = "Fluids and barriers of the CNS",
  ["fluss obst"] = "Flussiges Obst",
  ["fly (austin)"] = "Fly",
  ["flying saf"] = "Flying safety (Washington, D.C. : 1981)",
  ["fmda abt i: texte"] = "Forschungen und Materialien zur Deutschen Aufklärung (FMDA)",
  ["fmda abt ii: monogr"] = "Forschungen und Materialien zur Deutschen Aufklarung (FMDA). Abteilung II: Monographien",
  ["fnib info"] = "FNIB info",
  ["focus (am psychiatr publ)"] = "Focus (American Psychiatric Publishing)",
  ["focus (madison)"] = "Focus",
  ["focus aacn"] = "Focus On Aacn",
  ["focus altern complement ther"] = "Focus on alternative and complementary therapies : FACT",
  ["focus autism other dev disabl"] = "Focus on autism and other developmental disabilities",
  ["focus autom control ser"] = "Focus. Automation and Control Series",
  ["focus catal"] = "Focus on Catalysts",
  ["focus comput graph"] = "Focus on Computer Graphics",
  ["focus crit care"] = "Focus on critical care",
  ["focus except child"] = "Focus on exceptional children",
  ["focus gend"] = "Focus on gender",
  ["focus mda"] = "Focus MDA : the official publication of the Missouri Dental Association",
  ["focus ohio dent"] = "Focus on Ohio dentistry",
  ["focus pigm"] = "Focus on Pigments",
  ["focus polyvinyl chloride"] = "Focus on Polyvinyl Chloride",
  ["focus powder coat"] = "Focus on Powder Coatings",
  ["focus ser digit signal image process"] = "Focus Series in Digital Signal and Image Processing",
  ["focus ser finance bus manag"] = "Focus Series in Finance, Business and Management",
  ["focus surfactants"] = "Focus on Surfactants",
  ["focus tech coop"] = "Focus: technical cooperation. Focus. Cooperación técnica. Focus. Coopération technique",
  ["fogorv sz"] = "Fogorvosi szemle",
  ["foi lang"] = "Foi et langage",
  ["fol ophthalmol"] = "Folia ophthalmologica",
  ["fold des"] = "Folding and Design",
  ["foldr ert"] = "Földrajzi értesítő",
  ["folha med"] = "Folha medica",
  ["folia allergol (roma)"] = "Folia allergologica",
  ["folia biol"] = "Folia Biologica",
  ["folia biol (krakow)"] = "Folia Biologica",
  ["folia biol (praha)"] = "Folia biologica",
  ["folia biotheor"] = "Folia biotheoretica",
  ["folia cardiol"] = "Folia cardiologica",
  ["folia clin biol (sao paulo)"] = "Folia clinica et biologica",
  ["folia clin int (barc)"] = "Folia clínica internacional",
  ["folia cryptogam est"] = "Folia cryptogamica Estonica",
  ["folia endocrinol"] = "Folia endocrinologica",
  ["folia endocrinol mens incretologia incretoterapia"] = "Folia endocrinologica; mensile di incretologia e incretoterapia",
  ["folia entomol hung (1970)"] = "Folia entomologica Hungarica (Budapest, Hungary : 1970)",
  ["folia fac sci natur univ masaryk brun math"] = "Folia Facultatis Scientiarium Naturalium Universitatis Masarykianae Brunensis",
  ["folia for pol"] = "Folia Forestalia Polonica",
  ["folia for pol ser a for"] = "Folia forestalia Polonica. Series A, Forestry",
  ["folia geobot"] = "Folia geobotanica",
  ["folia geobot phytotaxon"] = "Folia Geobotanica et Phytotaxonomica",
  ["folia geogr ser geogr oecon"] = "Folia geographica. Series geographica-oeconomica",
  ["folia gynaecol (1908)"] = "Folia gynaecologica",
  ["folia haematol (frankf)"] = "Folia haematologica; internationales Magazin für Blutforschung",
  ["folia haematol int mag klin morphol blutforsch"] = "Folia haematologica (Leipzig, Germany : 1928)",
  ["folia hered pathol (milano)"] = "Folia hereditaria et pathologica",
  ["folia histochem cytobiol"] = "Folia histochemica et cytobiologica",
  ["folia histochem cytochem (krakow)"] = "Folia histochemica et cytochemica",
  ["folia humanist"] = "Folia humanistica",
  ["folia malacol"] = "Folia malacologica",
  ["folia math"] = "Folia Mathematica",
  ["folia med (napoli)"] = "Folia medica. Folia medica (Naples, Italy)",
  ["folia med (plovdiv)"] = "Folia medica",
  ["folia med cracov"] = "Folia medica Cracoviensia",
  ["folia med neerl"] = "Folia medica Neerlandica",
  ["folia mendeliana"] = "Folia mendeliana",
  ["folia microbiol"] = "Folia Microbiologica",
  ["folia microbiol (dordrecht, neth)"] = "Folia Microbiologica (Dordrecht, Netherlands)",
  ["folia microbiol (prague, czech repub)"] = "Folia Microbiologica (Prague, Czech Republic)",
  ["folia microbiol (praha)"] = "Folia microbiologica",
  ["folia morphol (praha)"] = "Folia Morphologica (Prague)",
  ["folia morphol (warsz)"] = "Folia morphologica",
  ["folia neuropathol"] = "Folia neuropathologica",
  ["folia oecol"] = "Folia Oecologica",
  ["folia orient"] = "Folia orientalia",
  ["folia parasitol (praha)"] = "Folia parasitologica",
  ["folia phoniatr (basel)"] = "Folia phoniatrica",
  ["folia phoniatr logop"] = "Folia phoniatrica et logopaedica : official organ of the International Association of Logopedics and Phoniatrics (IALP)",
  ["folia primatol"] = "Folia Primatologica",
  ["folia primatol (basel)"] = "Folia primatologica; international journal of primatology",
  ["folia psychiatr"] = "Folia psychiatrica",
  ["folia psychiatr neurol jpn"] = "Folia psychiatrica et neurologica japonica",
  ["folia psychiatr neurol neurochir neerl"] = "Folia psychiatrica, neurologica et neurochirurgica Neerlandica",
  ["folia stomatol"] = "Folia stomatologica",
  ["folia vet"] = "Folia veterinaria",
  ["folia vet lat"] = "Folia veterinaria Latina",
  ["folia zool"] = "Folia Zoologica",
  ["folia zool brno"] = "Folia zoologica",
  ["folkl am"] = "Folklore americano",
  ["folkl brabancon"] = "Folklore brabançon",
  ["folkl champagne"] = "Folklore de Champagne",
  ["folkl suisse"] = "Folklore suisse",
  ["folli publ log lang inf"] = "FoLLI Publications on Logic, Language and Information",
  ["fond cime"] = "Fondazione C",
  ["fond cime/cime found subser"] = "Fondazione CIME/CIME Foundation Subseries",
  ["fonderie fondeur aujourd’hui"] = "Fonderie Fondeur d’Aujourd’hui",
  ["fonti stor sanita"] = "Fonti per la storia della sanità : collana del CISO Veneto patrocinata dalla regione Veneto",
  ["fontilles rev leprol"] = "Fontilles",
  ["food addit contam"] = "Food Additives and Contaminants",
  ["food addit contam part a chem anal control expo risk assess"] = "Food additives & contaminants. Part A, Chemistry, analysis, control, exposure & risk assessment",
  ["food addit contam part b surveill"] = "Food additives & contaminants. Part B, Surveillance",
  ["food addit contam, part a"] = "Food Additives and Contaminants, Part A: Chemistry Analysis Control Exposure & Risk Assessment",
  ["food addit contam, part b"] = "Food Additives and Contaminants, Part B: Surveillance",
  ["food agric immunol"] = "Food and agricultural immunology",
  ["food anal methods"] = "Food Analytical Methods",
  ["food bioeng"] = "Food Bioengineering",
  ["food biophys"] = "Food Biophysics",
  ["food bioproc tech"] = "Food and bioprocess technology",
  ["food bioprocess technol"] = "Food and Bioprocess Technology",
  ["food bioprod process"] = "Food and Bioproducts Processing",
  ["food biosci"] = "Food bioscience",
  ["food biotechnol"] = "Food biotechnology",
  ["food chem"] = "Food chemistry",
  ["food chem adv"] = "Food Chemistry Advances",
  ["food chem toxicol"] = "Food and Chemical Toxicology",
  ["food chem x"] = "Food chemistry: X",
  ["food chem: mol sci"] = "Food Chemistry: Molecular Sciences",
  ["food chem: x"] = "Food Chemistry: X",
  ["food control"] = "Food control",
  ["food cosmet toxicol"] = "Food and Cosmetics Toxicology",
  ["food cult soc"] = "Food, culture, & society",
  ["food dig"] = "Food digestion",
  ["food digest: res curr opin"] = "Food Digestion: Research and Current Opinion",
  ["food drug cosmet law j"] = "Food, drug, cosmetic law journal",
  ["food drug cosmet law q"] = "Food, Drug, cosmetic law quarterly",
  ["food drug cosmet med device law dig"] = "Food, drug, cosmetic, and medical device law digest : a publication of the Food, Drug, and Cosmetic Law Section",
  ["food drug law j"] = "Food and drug law journal",
  ["food energy secur"] = "Food and Energy Security",
  ["food eng rev"] = "Food Engineering Reviews",
  ["food environ virol"] = "Food and environmental virology",
  ["food ethics"] = "Food ethics",
  ["food foods ingredients j jpn"] = "Food & Foods Ingredients Journal of Japan",
  ["food foodways"] = "Food & foodways",
  ["food front"] = "Food frontiers",
  ["food funct"] = "Food & function",
  ["food hydrocoll"] = "Food hydrocolloids",
  ["food hydrocolloids"] = "Food Hydrocolloids",
  ["food ind"] = "Food industries",
  ["food manage"] = "Food management",
  ["food mater equip"] = "Food materials and equipment",
  ["food microbiol"] = "Food microbiology",
  ["food modell j"] = "Food Modelling Journal",
  ["food nutr (roma)"] = "Food and nutrition",
  ["food nutr bull"] = "Food and nutrition bulletin",
  ["food nutr res"] = "Food & nutrition research",
  ["food nutr sci"] = "Food and nutrition sciences",
  ["food packag shelf life"] = "Food packaging and shelf life",
  ["food pol"] = "Food Policy",
  ["food policy"] = "Food policy",
  ["food prod process nutr"] = "Food Production, Processing and Nutrition",
  ["food prot trends"] = "Food protection trends",
  ["food public health"] = "Food and public health",
  ["food qual"] = "Food Quality",
  ["food qual prefer"] = "Food quality and preference",
  ["food qual preference"] = "Food Quality and Preference",
  ["food qual saf"] = "Food Quality and Safety",
  ["food res"] = "Food research",
  ["food res inst stud"] = "Food Research Institute studies (Stanford, Calif. : 1975)",
  ["food res inst stud agric econ trade dev"] = "Food Research Institute studies in agricultural economics, trade, and development",
  ["food res int"] = "Food research international (Ottawa, Ont.)",
  ["food rev int"] = "Food Reviews International",
  ["food saf (tokyo)"] = "Food safety (Tokyo, Japan)",
  ["food sci agric chem"] = "Food science and agricultural chemistry",
  ["food sci anim resour"] = "Food science of animal resources",
  ["food sci appl biotechnol"] = "Food Science And Applied Biotechnology",
  ["food sci biotechnol"] = "Food science and biotechnology",
  ["food sci hum wellness"] = "Food Science and Human Wellness",
  ["food sci nutr"] = "Food science & nutrition",
  ["food sci nutr cases"] = "Food Science and Nutrition Cases",
  ["food sci technol"] = "Food Science and Technology",
  ["food sci technol braz"] = "Food Science And Technology Brazil",
  ["food sci technol bull"] = "Food science and technology bulletin",
  ["food sci technol int"] = "Food science and technology international = Ciencia y tecnología de los alimentos internacional",
  ["food sci technol res"] = "Food Science and Technology Research",
  ["food secur"] = "Food security",
  ["food struct"] = "Food Structure",
  ["food stud"] = "Food studies",
  ["food technol"] = "Food technology",
  ["food technol biotechnol"] = "Food technology and biotechnology",
  ["food waterborne parasitol"] = "Food and waterborne parasitology",
  ["food webs"] = "Food Webs",
  ["foodborne pathog dis"] = "Foodborne pathogens and disease",
  ["fookien times yearb"] = "The Fookien times yearbook",
  ["foot (edinb)"] = "Foot (Edinburgh, Scotland)",
  ["foot ankle"] = "Foot & ankle",
  ["foot ankle clin"] = "Foot and ankle clinics",
  ["foot ankle int"] = "Foot & ankle international",
  ["foot ankle orthop"] = "Foot & ankle orthopaedics",
  ["foot ankle spec"] = "Foot & ankle specialist",
  ["foot ankle surg"] = "Foot and ankle surgery : official journal of the European Society of Foot and Ankle Surgeons",
  ["footwear sci"] = "Footwear Science",
  ["for abstr"] = "Forestry Abstracts",
  ["for can nor"] = "Forestry Canda Modeling Working Group : Proceedings of The Fifth Annual Workshop",
  ["for chron"] = "Forestry Chronicle",
  ["for conserv hist"] = "Forest & Conservation History",
  ["for ecol manage"] = "Forest ecology and management",
  ["for ecosyst"] = "Forest ecosystems",
  ["for fr"] = "Forêts de France",
  ["for fr action for"] = "Forêts de France et action forestière",
  ["for genet"] = "Forest genetics : FG : international journal of forest genetics",
  ["for ideas"] = "Forestry ideas",
  ["for immunopathol dis therap"] = "Forum on immunopathological diseases and therapeutics",
  ["for ind"] = "Forest Industries",
  ["for insect dis leafl"] = "Forest Insect & Disease Leaflet",
  ["for landsc res"] = "Forest and Landscape Research",
  ["for méditerr"] = "Forêt méditerranéenne",
  ["for pathol"] = "Forest pathology",
  ["for policy econ"] = "Forest policy and economics",
  ["for prod j"] = "Forest products journal",
  ["for res"] = "Forestry Research",
  ["for sci"] = "Forest Science",
  ["for sci pract"] = "Forest Science and Practice",
  ["for sci technol"] = "Forest Science and Technology",
  ["for snow landsc res"] = "Forest Snow and Landscape Research",
  ["for stud"] = "Forestry Studies",
  ["for stud china"] = "Forestry studies in China",
  ["for syst"] = "Forest systems",
  ["for tech rep"] = "Forestry Technical Report",
  ["for your inf (inst health rec inf manag)"] = "For your information (Institute of Health Record & Information Management)",
  ["forage grazinglands"] = "Forage & Grazinglands",
  ["forced migr rev"] = "Forced migration review",
  ["forces mech"] = "Forces in Mechanics",
  ["ford found rep"] = "Ford Foundation report (New York, N.Y. : 1992)",
  ["fordham int law j"] = "Fordham international law journal",
  ["fordham law rev"] = "Fordham law review",
  ["fordham urban law j"] = "The Fordham urban law journal",
  ["foreign aff"] = "Foreign affairs (Council on Foreign Relations)",
  ["foreign lang ann"] = "Foreign language annals",
  ["foreign policy"] = "Foreign policy",
  ["forensic chem"] = "Forensic chemistry (Amsterdam, Netherlands)",
  ["forensic eng"] = "Forensic Engineering",
  ["forensic genomics"] = "Forensic Genomics",
  ["forensic imaging"] = "Forensic Imaging",
  ["forensic med anat res"] = "Forensic medicine and anatomy research",
  ["forensic sci"] = "Forensic science",
  ["forensic sci addict res"] = "Forensic science & addiction research",
  ["forensic sci criminol"] = "Forensic science and criminology",
  ["forensic sci int"] = "Forensic science international",
  ["forensic sci int genet"] = "Forensic science international. Genetics",
  ["forensic sci int genet suppl ser"] = "Forensic science international. Genetics supplement series",
  ["forensic sci int mind law"] = "Forensic science international. Mind and law",
  ["forensic sci int suppl ser"] = "Forensic Science International Supplement Series",
  ["forensic sci int: digital invest"] = "Forensic Science International: Digital Investigation",
  ["forensic sci int: genet"] = "Forensic Science International: Genetics",
  ["forensic sci int: genet suppl ser"] = "Forensic Science International: Genetics Supplement Series",
  ["forensic sci int: mind law"] = "Forensic Science International: Mind and Law",
  ["forensic sci int: rep"] = "Forensic Science International: Reports",
  ["forensic sci int: synergy"] = "Forensic Science International: Synergy",
  ["forensic sci med pathol"] = "Forensic science, medicine, and pathology",
  ["forensic sci policy manage"] = "Forensic Science Policy and Management",
  ["forensic sci prog"] = "Forensic Science Progress",
  ["forensic sci res"] = "Forensic sciences research",
  ["forensic sci rev"] = "Forensic science review",
  ["forensic toxicol"] = "Forensic Toxicology",
  ["forensische psychiatr psychol kriminol"] = "Forensische Psychiatrie, Psychologie, Kriminologie",
  ["foresight (colch)"] = "Foresight (Colchester, Vt.)",
  ["forest sci technol"] = "Forest science and technology",
  ["forestry (lond)"] = "Forestry (London, England)",
  ["forests trees livelihoods"] = "Forests, Trees and Livelihoods",
  ["form aggiorn mat insegn"] = "Formazione e Aggiornamento in Matematica degli Insegnanti",
  ["form asp comput"] = "Formal Aspects of Computing",
  ["form methods syst des"] = "Formal methods in system design",
  ["formal aspects comput"] = "Formal Aspects of Computing",
  ["formal aspects comput: appl formal methods"] = "Formal Aspects of Computing: Applicable Formal Methods",
  ["formal methods syst des"] = "Formal Methods in System Design",
  ["formos entomol"] = "Formosan entomologist",
  ["formos j med humanit"] = "Formosan journal of medical humanities",
  ["foro int"] = "Foro internacional",
  ["foro mund salud"] = "Foro mundial de la salud",
  ["forsch fortbild chir bewegungsappar"] = "Forschung und Fortbildung in der Chirurgie des Bewegungsapparates",
  ["forsch fortbild chir bewegungsapparates"] = "Forschung und Fortbildung in der Chirurgie des Bewegungsapparates",
  ["forsch fortschr"] = "Forschungen und Fortschritte",
  ["forsch geb ingenieurwes"] = "Forschung auf dem Gebiete des Ingenieurwesens",
  ["forsch ingenieurwes"] = "Forschung im Ingenieurwesen",
  ["forsch inst realienkd mittelalt fruhen neuzeit diskuss mater"] = "Forschungen des Instituts fur Realienkunde des Mittelalters und der Fruhen Neuzeit. Diskussionen und Materialien",
  ["forsch j neue soz beweg"] = "Forschungsjournal neue soziale Bewegungen",
  ["forsch komplement med (2006)"] = "Forschende Komplementarmedizin (2006)",
  ["forsch komplementarmed"] = "Forschende Komplementarmedizin",
  ["forsch komplementarmed klass naturheilkd"] = "Forschende Komplementärmedizin und klassische Naturheilkunde = Research in complementary and natural classical medicine",
  ["forsch komplementmed"] = "Forschende Komplementärmedizin (2006)",
  ["forsch volks landeskd"] = "Forschungen zur Volks- und Landeskunde",
  ["forschber forstl forschanst münch"] = "Forschungsbericht forstliche Forschungsanstalt München",
  ["forschergeb inform"] = "Forschungsergebnisse zur Informatik",
  ["forschungsberichte atb"] = "Forschungsberichte des ATB = Research reports of ATB",
  ["forst holz"] = "Forst und Holz",
  ["forst- holzwirt"] = "Forst- und Holzwirt, Der",
  ["forstl forschber münch"] = "Forstliche Forschungsberichte München",
  ["forstl forsøgsvaes dan"] = "forstlige forsøgsvaesen i Danmark, Det",
  ["forstl rundsch"] = "Forstliche Rundschau",
  ["forstl umsch"] = "Forstliche Umschau",
  ["forstl-natwiss z"] = "Forstlich-naturwissenschaftliche Zeitschrift",
  ["forstschutz bull"] = "Forstschutz Bulletin",
  ["forsttech inf"] = "Forsttechnische Informationen",
  ["forstwiss centralbl"] = "Forstwissenschaftliches Centralblatt",
  ["fort concho rep"] = "Fort Concho report",
  ["fortn rev chic dent soc"] = "The Fortnightly review of the Chicago Dental Society",
  ["fortpflanz besamung aufzucht haustiere"] = "Fortpflanzung, Besamung, und Aufzucht der Haustiere",
  ["fortran numer recipes"] = "Fortran Numerical Recipes",
  ["fortschr allerg"] = "Fortschritte der Allergielehre",
  ["fortschr androl"] = "Fortschritte der Andrologie",
  ["fortschr angew radioisot grenzgeb"] = "Fortschritte der angewandten Radioisotopie und Grenzgebiete",
  ["fortschr arzneimittelforsch"] = "Fortschritte der Arzneimittelforschung. Progress in drug research. Progrès des recherches pharmaceutiques",
  ["fortschr chem forsch"] = "Fortschritte der Chemischen Forschung",
  ["fortschr chem org naturst"] = "Fortschritte der Chemie organischer Naturstoffe = Progress in the chemistry of organic natural products. Progrès dans la chimie des substances organiques naturelles",
  ["fortschr chem phys phys chem"] = "Fortschritte der Chemie, Physik und Physikalischen Chemie",
  ["fortschr geb rontgenstr"] = "Fortschritte auf dem Gebiete der Rontgenstrahlen",
  ["fortschr geb rontgenstr nuklearmed"] = "Fortschritte auf dem Gebiete der Rontgenstrahlen und Nuklearmedizine",
  ["fortschr geb rontgenstrahlen erganzungsbd"] = "Fortschritte auf dem Gebiete der Röntgenstrahlen. Ergänzungsband",
  ["fortschr geb rontgenstrahlen neuen bildgeb verfahr erganzungsbd"] = "Fortschritte auf dem Gebiete der Röntgenstrahlen und der neuen bildgebenden Verfahren. Ergänzungsband",
  ["fortschr geb rontgenstrahlen nuklearmed erganzungsbd"] = "Fortschritte auf dem Gebiete der Rontgenstrahlen und der Nuklearmedizin. Erganzungsband",
  ["fortschr geburtshilfe gynakol"] = "Fortschritte der Geburtshilfe und Gynakologie",
  ["fortschr hals nasen ohrenheilkd"] = "Fortschritte der Hals-Nasen-Ohrenheilkunde",
  ["fortschr hochpolym forsch"] = "Fortschritte der Hochpolymoren-Forschung",
  ["fortschr immunitatsforsch"] = "Fortschritte der Immunitatsforschung",
  ["fortschr kardiol"] = "Fortschritte der Kardiologie",
  ["fortschr kiefer gesichtschir"] = "Fortschritte der Kiefer- und Gesichts-Chirurgie",
  ["fortschr kieferorthop"] = "Fortschritte der Kieferorthopadie",
  ["fortschr med"] = "Fortschritte der Medizin",
  ["fortschr med monogr"] = "Fortschritte der Medizin. Monographie",
  ["fortschr med orig"] = "Fortschritte der Medizin. Originalien",
  ["fortschr med suppl"] = "Fortschritte der Medizin. Supplement : die Kongressinformation für die Praxis",
  ["fortschr mineral"] = "Fortschritte der Mineralogie",
  ["fortschr mineral kristallog petrogr"] = "Fortschritte der Mineralogie, Kristallograhie, und Petrographie",
  ["fortschr neurol psychiatr"] = "Fortschritte der Neurologie-Psychiatrie",
  ["fortschr neurol psychiatr grenzgeb"] = "Fortschritte der Neurologie, Psychiatrie, und ihrer Grenzgebiete",
  ["fortschr ophthalmol"] = "Fortschritte der Ophthalmologie : Zeitschrift der Deutschen Ophthalmologischen Gesellschaft",
  ["fortschr phys"] = "Fortschritte der Physik",
  ["fortschr phys chem"] = "Fortschritte der Physikalischen Chemie",
  ["fortschr psychosom med"] = "Fortschritte der psychosomatischen Medizin. Advances in psychosomatic medicine",
  ["fortschr strahlenschutz"] = "Fortschritte im Strahlenschutz",
  ["fortschr teerfarbenfabr verw industriezweige"] = "Fortschritte der Teerfarben fabrikation und Verwandter Industriezweige",
  ["fortschr tierphysiol tierernahr"] = "Fortschritte in der Tierphysiologie und Tierernahrung",
  ["fortschr verfahrenstech"] = "Fortschritte der Verfahrenstechnik",
  ["fortschr verfahrenstech abt a"] = "Fortschritte der Verfahrenstechnik Abteilung A: Grundlagen der Verfahrenstechnik",
  ["fortschr verfahrenstech abt b"] = "Fortschritte der Verfahrenstechnik Abteilung B: Mechanische Verfahrenstechnik",
  ["fortschr verfahrenstech abt c"] = "Fortschritte der Verfahrenstechnik Abteilung C: Thermische Verfahrenstechnik",
  ["fortschr verfahrenstech abt d"] = "Fortschritte der Verfahrenstechnik Abteilung D: Reaktionstechnik",
  ["fortschr verfahrenstech abt e"] = "Fortschritte der Verfahrenstechnik Abteilung E: Planung und Betrieb Von Anlagen",
  ["fortschr verfahrenstech abt f"] = "Fortschritte der Verfahrenstechnik Abteilung F: Spezielle Technologien",
  ["fortschr verhaltensforsch"] = "Fortschritte der Verhaltensforschung",
  ["fortschr zool"] = "Fortschritte der Zoologie",
  ["fortschrittsber dtschen keram ges"] = "Fortschriftsberiche der Deutschen Keramischen Gesellschaft",
  ["fortschrittsber kolloide polym"] = "Fortschrittsberichte uber Kolloide und Polymere",
  ["fortschrittsber landwirtsch nahrungsguterwirtsch"] = "Fortschrittsberichte fur die Landwirtschaft und Nahrungsguterwirtschaft",
  ["fortuna vitrea"] = "Fortuna vitrea",
  ["forum (genova)"] = "Forum (Genoa, Italy)",
  ["forum (wash)"] = "Forum (Washington, D.C. : 1977)",
  ["forum appl res public policy"] = "Forum for applied research and public policy",
  ["forum clin oncol"] = "Forum of clinical oncology",
  ["forum dev stud"] = "Forum for development studies",
  ["forum devel stud"] = "Forum for Development Studies",
  ["forum fam plan west hemisph"] = "Forum",
  ["forum geom"] = "Forum Geometricorum. A Journal on Classical Euclidean Geometry and Related Areas",
  ["forum health econ policy"] = "Forum for health economics & policy",
  ["forum immunopathol dis ther"] = "Forum on Immunopathological Diseases and Therapeutics",
  ["forum implantol"] = "Forum implantologicum",
  ["forum interdiscip math"] = "Forum for Interdisciplinary Mathematics",
  ["forum math"] = "Forum Mathematicum",
  ["forum math pi"] = "Forum of Mathematics Pi",
  ["forum math sigma"] = "Forum of Mathematics Sigma",
  ["forum med"] = "Forum on medicine",
  ["forum mond sante"] = "Forum mondial de la santé",
  ["forum nutr"] = "Forum of nutrition",
  ["forum public policy"] = "Forum on public policy",
  ["forum qual soc res"] = "Forum, qualitative social research",
  ["forum rev invat super"] = "Forum; revista invatamintului superior",
  ["forum soc econ"] = "Forum for Social Economics",
  ["forum stat"] = "Forum statisticum",
  ["forêt suisse"] = "Forêt Suisse",
  ["fossil rec"] = "Fossil Record",
  ["fottea (praha)"] = "Fottea (Praha)",
  ["found chem"] = "Foundations of Chemistry",
  ["found comm cogn"] = "Foundations of Communication and Cognition",
  ["found comput decision sci"] = "Poznań University of Technology",
  ["found comput math"] = "Foundations of Computational Mathematics",
  ["found comput ser"] = "Foundations of Computing Series",
  ["found comut math"] = "Foundations of computational mathematics (New York, N.Y.)",
  ["found data sci"] = "Foundations of Data Science",
  ["found eng mech"] = "Foundations of Engineering Mechanics",
  ["found mol model simul (2015)"] = "Foundations of Molecular Modeling and Simulation : select papers from FOMMS 2015. International Conference on Foundations of Molecular Modeling and Simulation (6th : 2015 : Mount Hood, Or.)",
  ["found news"] = "Foundation news",
  ["found news comment"] = "Foundation news & commentary",
  ["found phys"] = "Foundations of Physics",
  ["found phys lett"] = "Foundation of Physics Letters",
  ["found quest sci"] = "Foundational Questions in Science",
  ["found sci"] = "Foundations of Science",
  ["found signal process commun netw"] = "Foundations in Signal Processing, Communications and Networking",
  ["found trends mach learn"] = "Foundations and Trends in Machine Learning",
  ["found trends signal process"] = "Foundations and Trends${}^\\\\circledR$ in Signal Processing",
  ["found trends syst control"] = "Foundations and Trends in Systems and Control",
  ["found trends theor comput sci"] = "Foundations and Trends${}^\\\\circledR$ in Theoretical Computer Science",
  ["found undergrad res math"] = "Foundations for Undergraduate Research in Mathematics",
  ["fp essent"] = "FP essentials",
  ["fpm bull"] = "Federal personnel manual system. FPM bulletin. United States. Office of Personnel Management",
  ["fpop bull"] = "Fpop Bulletin",
  ["fr am rev"] = "FAR, The French-American review",
  ["fr colon hist"] = "French colonial history",
  ["fr hist"] = "French history",
  ["fr hist stud"] = "French historical studies",
  ["fr med"] = "France Medecine",
  ["fr stud"] = "French studies",
  ["fra sundhedsstyr"] = "Fra Sundhedsstyrelsen",
  ["fract calc appl anal"] = "Fractional Calculus and Applied Analysis",
  ["fract calc appl sci eng"] = "Fractional Calculus in Applied Sciences and Engineering",
  ["fract differ calc"] = "Fractional Differential Calculus",
  ["fractal (niterói)"] = "Fractal : revista de psicologia",
  ["fractals dyn math sci arts theory appl"] = "Fractals and Dynamics in Mathematics, Science, and the Arts: Theory and Applications",
  ["fragance j"] = "Fragrance Journal",
  ["fragm entomol"] = "Fragmenta entomologica",
  ["fragm florist geobot pol"] = "Fragmenta floristica et geobotanica Polonica",
  ["frank j fabozzi ser"] = "The Frank J. Fabozzi Series",
  ["frankf beitr gesch theor ethik med"] = "Frankfurter Beitrage zur Geschichte, Theorie und Ethik der Medizin",
  ["frankf hefte"] = "Frankfurter Hefte",
  ["frankf z pathol"] = "Frankfurter Zeitschrift fur Pathologie",
  ["franklin open"] = "Franklin Open",
  ["frater psi omega"] = "The Frater of Psi Omega",
  ["free china rev"] = "Free China review",
  ["free clin res collect"] = "Free clinic research collective",
  ["free inq"] = "Free inquiry (Buffalo, N.Y.)",
  ["free inq creat sociol"] = "Free inquiry in creative sociology",
  ["free neuropathol"] = "Free neuropathology",
  ["free radic antioxid"] = "Free Radicals and Antioxidants",
  ["free radic biol med"] = "Free radical biology & medicine",
  ["free radic res"] = "Free radical research",
  ["free radic res commun"] = "Free radical research communications",
  ["free radical biol med"] = "Free Radical Biology and Medicine",
  ["free radical res"] = "Free Radical Research",
  ["free radicals antioxid"] = "Free Radicals and Antioxidants",
  ["free world horiz"] = "Free world horizons",
  ["freedom issue"] = "Freedom at issue",
  ["freedom rev"] = "Freedom review",
  ["freib forsch medizingesch"] = "Freiburger Forschungen zur Medizingeschichte",
  ["freiberg forschungsh a"] = "Freiberger Forschungshefte A.",
  ["freibg bodenkd abh"] = "Freiburger bodenkundliche Abhandlungen",
  ["freibg geogr hefte"] = "Freiburger Geographische Hefte",
  ["freibg schr hydrol"] = "Freiburger Schriften zur Hydrologie",
  ["freibg stud frühen neuzeit"] = "Freiburger Studien zur Frühen Neuzeit",
  ["freie zahnarzt"] = "Der Freie Zahnarzt",
  ["fresenius environ bull"] = "Fresenius Environment bulletin",
  ["fresenius j anal chem"] = "Fresenius Journal of Analytical Chemistry",
  ["freshw biol"] = "Freshwater biology",
  ["freshw crayfish"] = "Freshwater crayfish",
  ["freshw mollusk biol conserv"] = "Freshwater mollusk biology and conservation : the journal of the Freshwater Mollusk Conservation Society",
  ["freshw rev"] = "Freshwater reviews : a journal of the Freshwater Biological Association",
  ["freshw sci"] = "Freshwater science (Print)",
  ["freshwater biol"] = "Freshwater Biology",
  ["freshwater mollusk biol conserv"] = "Freshwater Mollusk Biology and Conservation",
  ["freshwater rev"] = "Freshwater Reviews",
  ["freshwater sci"] = "Freshwater Science",
  ["friends p i nixon med hist libr"] = "Friends of the P.I. Nixon Medical Historical Library",
  ["friends women newsl"] = "Friends of Women newsletter",
  ["fritschiana (graz)"] = "Fritschiana : Veröffentlichungen aus dem Herbarium des Instituts für Botanik der Karl-Franzens-Universität Graz (GZU)",
  ["friuli med"] = "Il Friuli medico",
  ["front aerosp eng"] = "Frontiers in Aerospace Enginering",
  ["front aging"] = "Frontiers in Aging",
  ["front aging neurosci"] = "Frontiers in aging neuroscience",
  ["front aging ser"] = "Frontiers in aging series",
  ["front agric china"] = "Frontiers of Agriculture in China",
  ["front agric sci eng"] = "Frontiers of agricultural science and engineering",
  ["front agron"] = "Frontiers in Agronomy",
  ["front allergy"] = "Frontiers in Allergy",
  ["front amphib reptile sci"] = "Frontiers in Amphibian and Reptile Science",
  ["front anal sci"] = "Frontiers in Analytical Science",
  ["front anesthesiol"] = "Frontiers in Anesthesiology",
  ["front anim sci"] = "Frontiers in Animal Science",
  ["front antibiot"] = "Frontiers in Antibiotics",
  ["front appl dyn syst rev tutor"] = "Frontiers in Applied Dynamical Systems: Reviews and Tutorials",
  ["front appl math stat"] = "Frontiers in applied mathematics and statistics",
  ["front aquacult"] = "Frontiers in Aquacult.",
  ["front aquat physiol"] = "Frontiers in Aquatic Physiology",
  ["front arachnid sci"] = "Frontiers in Arachnid Science",
  ["front archit res"] = "Frontiers of Architectural Research",
  ["front artif intell"] = "Frontiers in artificial intelligence",
  ["front artif intell appl"] = "Frontiers in Artificial Intelligence and Applications",
  ["front astron space sci"] = "Frontiers in Astronomy and Space Sciences",
  ["front behav neurosci"] = "Frontiers in behavioral neuroscience",
  ["front big data"] = "Frontiers in big data",
  ["front bioeng biotechnol"] = "Frontiers in bioengineering and biotechnology",
  ["front biogeogr"] = "Frontiers of biogeography",
  ["front bioinf"] = "Frontiers in Bioinformatics",
  ["front biol"] = "Frontiers of biology",
  ["front biol (beijing)"] = "Frontiers in biology",
  ["front biol china"] = "Frontiers of biology in China : selected publications from Chinese universities",
  ["front biomater sci"] = "Frontiers in Biomaterials Science",
  ["front biomech"] = "Frontiers of Biomechanics",
  ["front biosci"] = "Frontiers in bioscience : a journal and virtual library",
  ["front biosci (elite ed)"] = "Frontiers in bioscience (Elite edition)",
  ["front biosci (landmark ed)"] = "Frontiers in bioscience (Landmark edition)",
  ["front biosci (schol ed)"] = "Frontiers in bioscience (Scholar edition)",
  ["front bird sci"] = "Frontiers in Bird Science",
  ["front blockchain"] = "Frontiers in Blockchain",
  ["front built environ"] = "Frontiers in built environment",
  ["front carbon"] = "Frontiers in Carbon",
  ["front cardiovasc med"] = "Frontiers in cardiovascular medicine",
  ["front catal"] = "Frontiers in Catalysis",
  ["front cell death"] = "Frontiers in Cell Death",
  ["front cell dev biol"] = "Frontiers in cell and developmental biology",
  ["front cell infect microbiol"] = "Frontiers in cellular and infection microbiology",
  ["front cell neurosci"] = "Frontiers in cellular neuroscience",
  ["front chem"] = "Frontiers in chemistry",
  ["front chem china"] = "Frontiers of Chemistry in China",
  ["front chem eng"] = "Frontiers in Chemical Engineering",
  ["front chem eng china"] = "Frontiers of Chemical Engineering in China",
  ["front chem sci eng"] = "Frontiers of chemical science and engineering",
  ["front clim"] = "Frontiers in Climate",
  ["front cognit"] = "Frontiers in Cognition",
  ["front coll"] = "The Frontiers Collection",
  ["front commun"] = "Frontiers in Communication",
  ["front commun (lausanne)"] = "Frontiers in communication",
  ["front commun networks"] = "Frontiers in Communications and Networks",
  ["front comput neurosci"] = "Frontiers in computational neuroscience",
  ["front comput sci"] = "Frontiers in computer science",
  ["front comput sci china"] = "Frontiers of computer science in China",
  ["front conserv sci"] = "Frontiers in Conservation Science",
  ["front control eng"] = "Frontiers in Control Engineering",
  ["front dent"] = "Frontiers in dentistry",
  ["front dent med"] = "Frontiers in Dental Medicine",
  ["front digit health"] = "Frontiers in digital health",
  ["front digital health"] = "Frontiers in Digital Health",
  ["front drug delivery"] = "Frontiers in Drug Delivery",
  ["front drug des discovery"] = "Frontiers in Drug Design and Discovery",
  ["front drug discovery"] = "Frontiers in Drug Discovery",
  ["front drug saf regul"] = "Frontiers in Drug Safety and Regulation",
  ["front earth sci"] = "Frontiers in Earth Science",
  ["front earth sci (lausanne)"] = "Frontiers in earth science",
  ["front earth sci chin"] = "Frontiers of Earth Science in China",
  ["front earth sci china"] = "Frontiers of Earth Science in China",
  ["front ecol environ"] = "Frontiers in ecology and the environment",
  ["front ecol evol"] = "Frontiers in ecology and evolution",
  ["front educ (lausanne)"] = "Frontiers in education",
  ["front electr electron eng"] = "Frontiers of Electrical and Electronic Engineering",
  ["front electr electron eng china"] = "Frontiers of Electrical and Electronic Engineering in China",
  ["front electron"] = "Frontiers in Electronics",
  ["front electron mater"] = "Frontiers in Electronic Materials",
  ["front elliptic parabol probl"] = "Frontiers in Elliptic and Parabolic Problems",
  ["front endocrinol"] = "Frontiers in Endocrinology",
  ["front endocrinol (lausanne)"] = "Frontiers in endocrinology",
  ["front energy"] = "Frontiers in Energy",
  ["front energy power eng china"] = "Frontiers of Energy and Power Engineering in China",
  ["front energy res"] = "Frontiers in Energy Research",
  ["front eng built environ"] = "Frontiers in Engineering and Built Environment",
  ["front eng manage"] = "Frontiers of Engineering Management",
  ["front environ archaeol"] = "Frontiers in Environmental Archaeology",
  ["front environ chem"] = "Frontiers in Environmental Chemistry",
  ["front environ econ"] = "Frontiers in Environmental Economics",
  ["front environ sci"] = "Frontiers in Environmental Science",
  ["front environ sci eng"] = "Frontiers of environmental science & engineering",
  ["front environ sci eng china"] = "Frontiers of Environmental Science & Engineering in China",
  ["front epidemiol"] = "Frontiers in Epidemiology",
  ["front evol neurosci"] = "Frontiers in evolutionary neuroscience",
  ["front evolut neurosci"] = "Frontiers in Evolutionary Neuroscience",
  ["front food sci technol"] = "Frontiers in Food Science and Technology",
  ["front for global change"] = "Frontiers in Forests and Global Change",
  ["front fuels"] = "Frontiers in Fuels",
  ["front fungal biol"] = "Frontiers in Fungal Biology",
  ["front future transp"] = "Frontiers in Future Transportation",
  ["front gastroenterol"] = "Frontiers in Gastroenterology",
  ["front gastrointest res"] = "Frontiers of gastrointestinal research",
  ["front genet"] = "Frontiers in genetics",
  ["front genome ed"] = "Frontiers in genome editing",
  ["front geochem"] = "Frontiers in Geochemistry",
  ["front glob womens health"] = "Frontiers in global women's health",
  ["front global women’s health"] = "Frontiers in Global Women’s Health",
  ["front health policy res"] = "Frontiers in health policy research",
  ["front health serv"] = "Frontiers in Health Services",
  ["front health serv manage"] = "Frontiers of health services management",
  ["front heat mass transfer"] = "Frontiers in Heat and Mass Transfer",
  ["front hematol"] = "Frontiers in Hematology",
  ["front high perform comput"] = "Frontiers in High Performance Computing",
  ["front hist china"] = "Frontiers of history in China",
  ["front hist sci"] = "Frontiers in the History of Science",
  ["front horm res"] = "Frontiers of hormone research",
  ["front hortic"] = "Frontiers in Horticulture",
  ["front hum dyn"] = "Frontiers in Human Dynamics",
  ["front hum neurosci"] = "Frontiers in human neuroscience",
  ["front ict"] = "Frontiers in ICT (Lausanne, Switzerland)",
  ["front imaging"] = "Frontiers in Imaging",
  ["front immunol"] = "Frontiers in immunology",
  ["front inf technol electron eng"] = "Frontiers of Information Technology & Electronic Engineering",
  ["front insect sci"] = "Frontiers in Insect Science",
  ["front integr neurosci"] = "Frontiers in integrative neuroscience",
  ["front internet things"] = "Frontiers in the Internet of Things",
  ["front lab chip technol"] = "Frontiers in Lab on a Chip Technologies",
  ["front lang sci"] = "Frontiers in Language Sciences",
  ["front life sci"] = "Frontiers in life science",
  ["front lines"] = "Front Lines",
  ["front lines res"] = "Front lines research",
  ["front malar"] = "Frontiers in Malaria",
  ["front manuf technol"] = "Frontiers in Manufacturing Technology",
  ["front mar sci"] = "Frontiers in Marine Science",
  ["front mater"] = "Frontiers in Materials",
  ["front mater sci"] = "Frontiers of materials science",
  ["front mater sci china"] = "Frontiers of Materials Science in China",
  ["front math"] = "Frontiers of Mathematics",
  ["front math china"] = "Frontiers of mathematics in China : selected papers from Chinese universities",
  ["front math finance"] = "Frontiers of Mathematical Finance",
  ["front mech eng"] = "Frontiers in Mechanical Engineering",
  ["front mech eng china"] = "Frontiers of Mechanical Engineering in China",
  ["front med"] = "Frontiers of medicine",
  ["front med (lausanne)"] = "Frontiers in medicine",
  ["front med biol eng"] = "Frontiers of medical and biological engineering : the international journal of the Japan Society of Medical Electronics and Biological Engineering",
  ["front med china"] = "Frontiers of medicine in China",
  ["front med technol"] = "Frontiers in medical technology",
  ["front membr sci technol"] = "Frontiers in Membrane Science and Technology",
  ["front met alloys"] = "Frontiers in Metals and Alloys",
  ["front microbiol"] = "Frontiers in microbiology",
  ["front microbiomes"] = "Frontiers in Microbiomes",
  ["front mol biosci"] = "Frontiers in molecular biosciences",
  ["front mol med"] = "Frontiers in Molecular Medicine",
  ["front mol neurosci"] = "Frontiers in molecular neuroscience",
  ["front nanotechnol"] = "Frontiers in Nanotechnology",
  ["front nat prod"] = "Frontiers in Natural Products",
  ["front nephrol"] = "Frontiers in Nephrology",
  ["front network physiol"] = "Frontiers in Network Physiology",
  ["front neural circuits"] = "Frontiers in neural circuits",
  ["front neuroanat"] = "Frontiers in neuroanatomy",
  ["front neuroendocr sci"] = "Frontiers in Neuroendocrine Science",
  ["front neuroendocrinol"] = "Frontiers in neuroendocrinology",
  ["front neuroenerg"] = "Frontiers in Neuroenergetics",
  ["front neuroenergetics"] = "Frontiers in neuroenergetics",
  ["front neuroeng"] = "Frontiers in neuroengineering",
  ["front neuroergon"] = "Frontiers in Neuroergonomics",
  ["front neuroimaging"] = "Frontiers in Neuroimaging",
  ["front neuroinf"] = "Frontiers in Neuroinformatics",
  ["front neuroinform"] = "Frontiers in neuroinformatics",
  ["front neurol"] = "Frontiers in neurology",
  ["front neurol neurosci"] = "Frontiers of neurology and neuroscience",
  ["front neurol neurosci res"] = "Frontiers in neurology and neuroscience research",
  ["front neurorob"] = "Frontiers in Neurorobotics",
  ["front neurorobot"] = "Frontiers in neurorobotics",
  ["front neurosci"] = "Frontiers in neuroscience",
  ["front nucl eng"] = "Frontiers in Nuclear Engineering",
  ["front nucl med"] = "Frontiers in Nuclear Medicine",
  ["front nurs serv q bull"] = "Frontier Nursing Service quarterly bulletin",
  ["front nutr"] = "Frontiers in nutrition",
  ["front oncol"] = "Frontiers in oncology",
  ["front opt"] = "Frontiers in optics. Annual Meeting of the Optical Society of America",
  ["front optoelectron"] = "Frontiers of Optoelectronics",
  ["front oral biol"] = "Frontiers of oral biology",
  ["front oral health"] = "Frontiers in Oral Health",
  ["front oral maxillofac med"] = "Frontiers of oral and maxillofacial medicine",
  ["front oral physiol"] = "Frontiers of oral physiology",
  ["front pain res"] = "Frontiers in Pain Research",
  ["front parasitol"] = "Frontiers in Parasitology",
  ["front pediatr"] = "Frontiers in pediatrics",
  ["front pharmacol"] = "Frontiers in pharmacology",
  ["front photonics"] = "Frontiers in Photonics",
  ["front phys"] = "Frontiers in physics",
  ["front phys (beijing)"] = "Frontiers of physics",
  ["front phys china"] = "Frontiers of physics in China",
  ["front physiol"] = "Frontiers in physiology",
  ["front plant sci"] = "Frontiers in plant science",
  ["front polit sci"] = "Frontiers in political science",
  ["front probab stat sci"] = "Frontiers in Probability and the Statistical Sciences",
  ["front psychiatry"] = "Frontiers in psychiatry",
  ["front psychol"] = "Frontiers in psychology",
  ["front psychol behav sci"] = "Frontiers in psychological and behavioral science",
  ["front public health"] = "Frontiers in public health",
  ["front public health serv syst res"] = "Frontiers in public health services & systems research",
  ["front quantum sci technol"] = "Frontiers in Quantum Science and Technology",
  ["front radiat ther oncol"] = "Frontiers of radiation therapy and oncology",
  ["front radiol"] = "Frontiers in Radiology",
  ["front rehabil sci"] = "Frontiers in Rehabilitation Sciences",
  ["front remote sens"] = "Frontiers in Remote Sensing",
  ["front res comput mech mater biol"] = "Frontier Research in Computation and Mechanics of Materials and Biology",
  ["front res metr anal"] = "Frontiers in research metrics and analytics",
  ["front res metrics anal"] = "Frontiers in Research Metrics and Analytics",
  ["front rob ai"] = "Frontiers in Robotics and AI",
  ["front robot ai"] = "Frontiers in robotics and AI",
  ["front sci ser"] = "Frontier Science Series",
  ["front sens"] = "Frontiers in Sensors",
  ["front signal process"] = "Frontiers in Signal Processing",
  ["front sleep"] = "Frontiers in Sleep",
  ["front smart grids"] = "Frontiers in Smart Grids",
  ["front soc psychol"] = "Social Cognition: The Basis of Human Interaction",
  ["front soft matter"] = "Frontiers in Soft Matter",
  ["front soil sci"] = "Frontiers in Soil Science",
  ["front space technol"] = "Frontiers in Space Technologies",
  ["front sports act living"] = "Frontiers in sports and active living",
  ["front sports active living"] = "Frontiers in Sports and Active Living",
  ["front stat qual control"] = "Frontiers in Statistical Quality Control",
  ["front struct civ eng"] = "Frontiers of Structural and Civil Engineering",
  ["front surg"] = "Frontiers in surgery",
  ["front sustain food syst"] = "Frontiers in sustainable food systems",
  ["front sustainability"] = "Frontiers in Sustainability",
  ["front sustainable cities"] = "Frontiers in Sustainable Cities",
  ["front sustainable energy policy"] = "Frontiers in Sustainable Energy Policy",
  ["front sustainable food syst"] = "Frontiers in Sustainable Food Systems",
  ["front sustainable resour manage"] = "Frontiers in Sustainable Resource Management",
  ["front sustainable tourism"] = "Frontiers in Sustainable Tourism",
  ["front synaptic neurosci"] = "Frontiers in synaptic neuroscience",
  ["front syst biol"] = "Frontiers in Systems Biology",
  ["front syst neurosci"] = "Frontiers in systems neuroscience",
  ["front theor phys"] = "Frontiers in Theoretical Physics",
  ["front therm eng"] = "Frontiers in Thermal Engineering",
  ["front toxicol"] = "Frontiers in Toxicology",
  ["front transplant"] = "Frontiers in Transplantation",
  ["front urol"] = "Frontiers in Urology",
  ["front vet sci"] = "Frontiers in veterinary science",
  ["front virol"] = "Frontiers in Virology",
  ["front virtual real"] = "Frontiers in virtual reality",
  ["front virtual reality"] = "Frontiers in Virtual Reality",
  ["front water"] = "Frontiers in Water",
  ["front womens health"] = "Frontiers in women's health",
  ["front young minds"] = "Frontiers for young minds",
  ["front zool"] = "Frontiers in zoology",
  ["frontiers (boulder)"] = "Frontiers",
  ["frontiers appl math"] = "Frontiers in Applied Mathematics",
  ["frontiers artif intell appl"] = "Frontiers in Artif. Intell. and Applications",
  ["frontiers artificial intelligence appl"] = "Frontiers in Artificial Intelligence and Applications",
  ["frontiers econom res"] = "Frontiers of Economic Research",
  ["frontiers oper res appl systems anal"] = "Frontiers of Operational Research and Applied Systems Analysis",
  ["frontline gastroenterol"] = "Frontline gastroenterology",
  ["frontline learn res"] = "Frontline learning research",
  ["fronttrop dis"] = "Frontiers in Tropical Diseases",
  ["fruhmittelalt stud"] = "Frühmittelalterliche Studien",
  ["fruhneuz info"] = "Frühneuzeit-Info",
  ["fu dan xue bao zi ran ke xue ban"] = "Fu dan xue bao. Zi ran ke xue ban = Journal of Fudan University. Natural science",
  ["fu jian shui chan"] = "Fu jian shui chan = Fujian Fisheries",
  ["fu jian yi yao za zhi"] = "Fujian yi yao za zhi = Fujian medical journal",
  ["fuel (lond)"] = "Fuel (London, England)",
  ["fuel cells"] = "Fuel Cells",
  ["fuel cells (weinh)"] = "Fuel cells (Weinheim)",
  ["fuel cells (weinheim, ger)"] = "Fuel Cells (Weinheim, Germany)",
  ["fuel cells bull"] = "Fuel Cells Bulletin",
  ["fuel commun"] = "Fuel Communications",
  ["fuel process technol"] = "Fuel Processing Technology",
  ["fuel sci technol int"] = "Fuel Science and Technology International",
  ["fugitive leaves hist collect"] = "Fugitive leaves from the historical collections. College of Physicians of Philadelphia. Library",
  ["fujian linxueyuan xuebao"] = "Fujian linxueyuan xuebao",
  ["fujian medical journal|fu jian yi yao za zhi"] = "Fujian yi yao za zhi",
  ["fujian nong ye ke ji"] = "Fujian nong ye ke ji",
  ["fujian nong ye xue bao"] = "Fujian nong ye xue bao = Fujian journal of agricultural sciences",
  ["fujian shifan daxue xuebao ziran kexue ban"] = "Fujian Shifan Daxue Xuebao",
  ["fujian yike daxue xuebao shehui kexue ban"] = "Fujian yike daxue xuebao. Shehui kexue ban",
  ["fujitsu sci tech j"] = "Fujitsu Scientific & Technical Journal",
  ["fukui ken nogyo shikenjo kenkyu hokoku"] = "Fukui-ken Nogyo Shikenjo kenkyu hokoku = Bulletin of the Fukui Agricultural Experiment Station",
  ["fukuoka igaku zasshi"] = "Fukuoka igaku zasshi = Hukuoka acta medica",
  ["fukuoka shika daigaku gakkai zasshi"] = "Fukuoka Shika Daigaku Gakkai zasshi",
  ["fukuoka univ sci rep"] = "Fukuoka University Science Reports",
  ["fukushima igaku zasshi"] = "[Fukushima igaku zasshi] Fukushima medical journal",
  ["fukushima j med sci"] = "Fukushima journal of medical science",
  ["full circ"] = "Full Circle",
  ["fullerene sci technol"] = "Fullerene Science and Technology",
  ["fullerenes nanotubes carbon nanostruct"] = "Fullerence, Nanotubes, and Carbon Nanostructures",
  ["fullerenes, nanotubes, carbon nanostruct"] = "Fullerenes, Nanotubes, and Carbon Nanostructures",
  ["funct anal appl"] = "Functional Analysis and its Applications",
  ["funct anal approx comput"] = "Functional Analysis, Approximation and Computation",
  ["funct anal other math"] = "Functional Analysis and Other Mathematics",
  ["funct approx comment math"] = "Uniwersytet im. Adama Mickiewicza w Poznaniu. Wydzial Matematyki i Informatyki. Functiones et Approximatio Commentarii Mathematici",
  ["funct approximatio comment math"] = "Functiones et Approximatio Commentarii Mathematici",
  ["funct compos mater"] = "Functional Composite Materials",
  ["funct compos struct"] = "Functional Composites and Structures",
  ["funct dev morphol"] = "Functional and developmental morphology",
  ["funct diamond"] = "Functional Diamond",
  ["funct differ equ"] = "Functional Differential Equations",
  ["funct ecol"] = "Functional ecology",
  ["funct food rev"] = "Functional food reviews (Print)",
  ["funct funct struct mater"] = "Functional and functionally structured materials : selected, peer reviewed papers from the Chinese Materials Conference 2015, July 10-14, 2015, Guiyang, China. Chinese Materials Conference (2015 : Guiyang, China)",
  ["funct integr genomics"] = "Functional & integrative genomics",
  ["funct mater"] = "Functional Materials",
  ["funct mater lett"] = "Functional Materials Letters",
  ["funct neurol"] = "Functional neurology",
  ["funct orthod"] = "The Functional orthodontist",
  ["funct plant biol"] = "Functional plant biology : FPB",
  ["funct plant sci biotechnol"] = "Functional plant science & biotechnology",
  ["function (oxf)"] = "Function (Oxford, England)",
  ["fund inform"] = "Fundamenta Informaticae",
  ["fund math"] = "Fundamenta Mathematicae",
  ["fund raising manage"] = "Fund raising management",
  ["fund stud comput sci"] = "Fundamental Studies in Computer Science",
  ["fund theories phys"] = "Fundamental Theories of Physics",
  ["fundam algorithms"] = "Fundamentals of Algorithms",
  ["fundam appl limnol"] = "Fundamental and applied limnology : official journal of the International Association of Theoretical and Applied Limnology",
  ["fundam appl nematol"] = "Fundamental and applied nematology",
  ["fundam appl nucl phys ser"] = "Fundamental and Applied Nuclear Physics Series",
  ["fundam appl toxicol"] = "Fundamental and Applied Toxicology",
  ["fundam clin pharmacol"] = "Fundamental & clinical pharmacology",
  ["fundam inform"] = "Fundamenta informaticae",
  ["fundam math"] = "Fundamenta Mathematicae",
  ["fundam plasma phys"] = "Fundamental Plasma Physics",
  ["fundam prikl mat"] = "Fundamentalnaya i Prikladnaya Matematika",
  ["fundam res"] = "Fundamental Research",
  ["fundam sci"] = "Fundamenta scientiae",
  ["fundam theor phys"] = "Fundamental Theories of Physics",
  ["fundam toxicol sci"] = "Fundamental toxicological sciences",
  ["fungal biol"] = "Fungal biology",
  ["fungal biol biotechnol"] = "Fungal biology and biotechnology",
  ["fungal biol rev"] = "Fungal biology reviews",
  ["fungal divers"] = "Fungal diversity",
  ["fungal diversity"] = "Fungal Diversity",
  ["fungal ecol"] = "Fungal ecology",
  ["fungal genet biol"] = "Fungal genetics and biology : FG & B",
  ["fungal genet ecol"] = "Fungal Genetics and Ecology",
  ["fungal genet newsl"] = "Fungal genetics newsletter",
  ["fungal genet rep"] = "Fungal genetics reports",
  ["fungal genom biol"] = "Fungal genomics & biology",
  ["fungal planet"] = "Fungal planet",
  ["fungal sci"] = "Fungal science",
  ["fungal syst evol"] = "Fungal systematics and evolution",
  ["fungi iber"] = "Fungi iberici",
  ["fungi non delineati"] = "Fungi non delineati : raro vel haud perspecte et explorate descripti aut definite picti",
  ["funkcial ekvac"] = "Funkcialaj Ekvacioj. Serio Internacia",
  ["funktsional anal i prilozhen"] = "Funktsionalnyi Analiz i ego Prilozheniya",
  ["funtai oyobi funmatsu yakin"] = "Funtai oyobi Funmatsu Yakin",
  ["fushe fanghu"] = "Fushe Fanghu",
  ["fusion eng des"] = "Fusion Engineering and Design",
  ["fusion sci technol"] = "Fusion Science and Technology",
  ["fusion technol"] = "Fusion Technology",
  ["fut rheumatol"] = "Future rheumatology",
  ["futruibles 2000"] = "Futuribles 2000 [i.e. Deux mille]",
  ["futur anter"] = "Futur antérieur",
  ["futur hiv ther"] = "Future HIV therapy",
  ["futur j pharm sci"] = "Future journal of pharmaceutical sciences",
  ["future bus finance"] = "Future of Business and Finance",
  ["future cardiol"] = "Future cardiology",
  ["future child"] = "Future of Children",
  ["future cities environ"] = "Future Cities and Environment",
  ["future comput inf j"] = "Future Computing and Informatics Journal",
  ["future dent"] = "Future dentistry",
  ["future drug discov"] = "Future drug discovery",
  ["future drug discovery"] = "Future Drug Discovery",
  ["future foods"] = "Future Foods",
  ["future gener comput syst"] = "Future generations computer systems : FGCS",
  ["future healthc j"] = "Future healthcare journal",
  ["future hosp j"] = "Future hospital journal",
  ["future internet"] = "Future internet",
  ["future j pharm sci"] = "Future Journal of Pharmaceutical Sciences",
  ["future lipidol"] = "Future lipidology",
  ["future med ai"] = "Future Medicine AI",
  ["future med chem"] = "Future medicinal chemistry",
  ["future microbiol"] = "Future microbiology",
  ["future neurol"] = "Future neurology",
  ["future oncol"] = "Future oncology (London, England)",
  ["future sci oa"] = "Future science OA",
  ["future virol"] = "Future virology",
  ["fuzzy inf eng"] = "Fuzzy Information and Engineering",
  ["fuzzy manag methods"] = "Fuzzy Management Methods",
  ["fuzzy optim dec making"] = "Fuzzy Optimization and Decision Making",
  ["fuzzy optim decis mak"] = "Fuzzy Optimization and Decision Making. A Journal of Modeling and Computation Under Uncertainty",
  ["fuzzy sets and systems"] = "Fuzzy Sets and Systems. An International Journal in Information Science and Engineering",
  ["fuzzy sets syst"] = "Fuzzy Sets and Systems",
  ["fysiatr revmatol vestn"] = "Fysiatrický a reumatologický vestník",
  ["fysiatr vestn cesk fysiatr spol"] = "Fysiatrický vĕstník. Československá fysiatrická společnost v Praze",
  ["g accad med torino"] = "Giornale della Accademia di medicina di Torino",
  ["g anest stomatol"] = "Giornale di anestesia stomatologica = Journal of dental anaesthesia",
  ["g batteriol immunol"] = "Giornale di batteriologia e immunologia",
  ["g batteriol virol immunol"] = "Giornale di batteriologia, virologia ed immunologia",
  ["g batteriol virol immunol [clin]"] = "Giornale di Batteriologia, Virologia ed Immunologia ed Annali dell Ospedale Maria Vittoria di Torino. Parte 2, Sezione Clinica",
  ["g batteriol virol immunol clin"] = "Giornale di batteriologia, virologia ed immunologia ed annali dell'Ospedale Maria Vittoria di Torino. Parte 2: Sezione clinica",
  ["g batteriol virol immunol microbiol"] = "Giornale di batteriologia, virologia ed immunologia ed annali dell'Ospedale Maria Vittoria di Torino. Parte 1: Sezione microbiologica",
  ["g biochim"] = "Giornale di biochimica",
  ["g bot ital (florence italy)"] = "Giornale botanico italiano (Florence, Italy : 1962)",
  ["g chir"] = "Il Giornale di chirurgia",
  ["g clin med"] = "Giornale di clinica medica",
  ["g crit filos ital"] = "Giornale critico della filosofia italiana",
  ["g e n"] = "G.E.N",
  ["g econ ann econ"] = "Giornale degli economisti e annali di economia",
  ["g endodonzia"] = "Giornale di endodonzia",
  ["g fis sanit prot radiaz"] = "Giornale di fisica sanitaria e protezione contro le radiazioni. The journal of health physics and radiation protection",
  ["g gerontol"] = "Giornale di gerontologia",
  ["g gerontol suppl"] = "Giornale di gerontologia. Supplemento",
  ["g ig med prev"] = "Giornale di igiene e medicina preventiva",
  ["g ital anestesiol"] = "Giornale italiano di anestesiologia",
  ["g ital arterioscler"] = "Giornale italiano dell'arteriosclerosi",
  ["g ital cardiol"] = "Giornale italiano di cardiologia",
  ["g ital cardiol (rome)"] = "Giornale italiano di cardiologia (2006)",
  ["g ital chemioter"] = "Giornale italiano di chemioterapia",
  ["g ital chir"] = "Giornale italiano di chirurgia",
  ["g ital della tuberc"] = "Giornale italiano della tubercolosi",
  ["g ital dermatol"] = "Giornale italiano di dermatologia",
  ["g ital dermatol minerva dermatol"] = "Giornale italiano di dermatolotia. Minerva dermatologica",
  ["g ital dermatol venereol"] = "Giornale italiano di dermatologia e venereologia : organo ufficiale, Società italiana di dermatologia e sifilografia",
  ["g ital endod"] = "Giornale italiano di endodonzia",
  ["g ital endodonzia"] = "Giornale Italiano di Endodonzia",
  ["g ital filol"] = "Giornale italiano di filologia",
  ["g ital mal torace"] = "Giornale italiano delle malattie del torace",
  ["g ital med lav"] = "Giornale italiano di medicina del lavoro",
  ["g ital med lav ergon"] = "Giornale italiano di medicina del lavoro ed ergonomia",
  ["g ital nefrol"] = "Giornale italiano di nefrologia : organo ufficiale della Società italiana di nefrologia",
  ["g ital oftalmol"] = "Giornale italiano di oftalmologia",
  ["g ital oncol"] = "Giornale italiano di oncologia",
  ["g ital patol"] = "Giornale italiano di patologia e di scienze affini",
  ["g ital tuberc mal torace"] = "Giornale italiano della tubercolosi e delle malattie del torace",
  ["g mal infett parassit"] = "Giornale di malattie infettive e parassitarie",
  ["g med mil"] = "Giornale di medicina militare",
  ["g pneumol"] = "Giornale di pneumologia",
  ["g pop"] = "Giornale del Popolo",
  ["g psichiatr neuropatol"] = "Giornale di psichiatria e di neuropatologia",
  ["g sci mediche"] = "Giornale di scienze mediche",
  ["g stomatol ortognatodonzia"] = "Giornale di stomatologia e di ortognatodonzia",
  ["g tec nefrol dial"] = "Giornale de techniche nefrologiche & dialitiche",
  ["g veneto sci med"] = "Giornale veneto di scienze mediche",
  ["g3 (bethesda)"] = "G3 (Bethesda, Md.)",
  ["g3: genes, genomes, genet"] = "G3: Genes, Genomes, Genetics",
  ["ga epidemiol rep"] = "Georgia epidemiology report",
  ["ga hist q"] = "The Georgia historical quarterly",
  ["ga hosp today"] = "Georgia hospitals today",
  ["ga j sci"] = "Georgia journal of science : official publication of the Georgia Academy of Science",
  ["ga nurse"] = "Georgia nursing",
  ["ga pharm q"] = "Georgia pharmacist quarterly",
  ["ga rev"] = "The Georgia review",
  ["ga state univ law rev"] = "Georgia State University law review",
  ["gabi j"] = "GaBI journal",
  ["gac conasida"] = "Gaceta CONASIDA",
  ["gac med (guayaquil)"] = "Gaceta medica",
  ["gac med caracas"] = "Gaceta médica de Caracas",
  ["gac med centroam"] = "Gaceta médica centroamericana",
  ["gac med esp"] = "Gaceta médica española",
  ["gac med lima"] = "Gaceta médica de Lima",
  ["gac med mex"] = "Gaceta médica de México",
  ["gac med norte"] = "Gaceta médica del Norte",
  ["gac med occidente"] = "Gaceta médica de occidente",
  ["gac med quir boliv"] = "Gaceta médico-quirúrgica de Bolivia",
  ["gac med villahermosa"] = "Gaceta médica",
  ["gac méd méx"] = "Gaceta Médica de México",
  ["gac peru cir med"] = "Gaceta peruana de cirugía y medicina",
  ["gac r soc mat esp"] = "La Gaceta de la Real Sociedad Matematica Espanola",
  ["gac sanit"] = "Gaceta sanitaria",
  ["gaea an soc argent estud geogr"] = "Gaea; anales de la Sociedad Argentina de Estudios Geográficos",
  ["gait posture"] = "Gait & posture",
  ["gakuto int ser math sci appl"] = "GAKUTO International Series",
  ["gakuto internat ser math sci appl"] = "GAKUTO International Series. Mathematical Sciences and Applications",
  ["galapagos res"] = "Galapagos Research",
  ["galaxea (tokyo)"] = "Galaxea",
  ["galen acta"] = "Galénica acta",
  ["galen cas med farm hem srod nauke"] = "Galenika; c̆asopis za medicinu, farmaciju, hemiju i srodne nauke",
  ["galenika cas med farm hem srod nauke"] = "Galenika; casopis za medicinu, farmaciju, hemiju i srodne nauke",
  ["galicia clin"] = "Galicia-clinica",
  ["gall biol acta"] = "Gallica biologica acta",
  ["gallup opin index"] = "The Gallup opinion index",
  ["gallup rep"] = "Gallup report (Princeton, N.J. : 1981)",
  ["game theory appl"] = "Game Theory and Applications",
  ["games (basel)"] = "Games",
  ["games cult"] = "Games and culture",
  ["games econ behav"] = "Games and economic behavior",
  ["games econom behav"] = "Games and Economic Behavior",
  ["games health j"] = "Games for health journal",
  ["games: res pract"] = "Games: Research and Practice",
  ["gamete res"] = "Gamete research",
  ["gaming law rev econ"] = "Gaming law review and economics",
  ["gamm mitt ges angew math mech"] = "Mitteilungen der Gesellschaft für Angewandte Mathematik und Mechanik",
  ["gams j math math biosci"] = "GAMS Journal of Mathematics and Mathemaical Biosciences",
  ["gan no rinsho"] = "Gan no rinsho. Japan journal of cancer clinics",
  ["gan to kagaku ryoho"] = "Gan to kagaku ryoho. Cancer & chemotherapy",
  ["ganguang kexue yu guang huaxue"] = "Ganguang Kexue Yu Guang Huaxue",
  ["ganita bharati"] = "Ganita Bharati. Bulletin of the Indian Society for History of Mathematics",
  ["ganita sandesh"] = "Ganita Sandesh",
  ["gansu nong ye da xue xue bao"] = "Gansu nong ye da xue xue bao = Gansu nongye daxue xuebao",
  ["gansu nong ye ke ji"] = "Gansu nong ye ke ji = Gansu nongye keji",
  ["gansu nongye daxue xuebao|gansu nong ye da xue xue bao"] = "Gansu nong ye da xue xue bao",
  ["gao fen zi cai liao ke xue yu gong cheng"] = "Gao fen zi cai liao ke xue yu gong cheng = Polymeric materials science & engineering",
  ["gaodeng xuexiao huaxue xuebao"] = "Gaodeng Xuexiao Huaxue Xuebao",
  ["gaofenzi cailiao kexue yu gongcheng"] = "Gaofenzi Cailiao Kexue Yu Gongcheng",
  ["gaofenzi xuebao"] = "Gaofenzi Xuebao",
  ["gaojishu tongxun"] = "Gaojishu tongxun",
  ["gaoneng wuli yu hewuli"] = "Gaoneng Wuli Yu Hewuli",
  ["gaoxiao huaxue gongcheng xuebao"] = "Gaoxiao Huaxue Gongcheng Xuebao",
  ["gaoxiao yingyong shuxue xuebao ser a"] = "Gaoxiao Yingyong Shuxue Xuebao",
  ["gaoxiong yi xue ke xue za zhi"] = "Gaoxiong yi xue ke xue za zhi = The Kaohsiung journal of medical sciences",
  ["gard bull (singapore)"] = "The Gardens' bulletin, Singapore",
  ["gard hist"] = "Garden history",
  ["garland ref libr soc sci"] = "Garland Reference Library of Social Science",
  ["gas sci eng"] = "Gas Science and Engineering",
  ["gas sep purif"] = "Gas Separation and Purification",
  ["gas wasser abwasser"] = "Gas, Wasser, Abwasser",
  ["gastric cancer"] = "Gastric cancer : official journal of the International Gastric Cancer Association and the Japanese Gastric Cancer Association",
  ["gastroenterol bohema"] = "Gastroenterologia bohema",
  ["gastroenterol clin biol"] = "Gastroentérologie clinique et biologique",
  ["gastroenterol clin north am"] = "Gastroenterology clinics of North America",
  ["gastroenterol fortbildungskurse prax"] = "Gastroenterologische Fortbildungskurse fur die Praxis",
  ["gastroenterol hepatol"] = "Gastroenterología y hepatología",
  ["gastroenterol hepatol (bartlesville)"] = "Gastroenterology & hepatology (Bartlesville, Okla.)",
  ["gastroenterol hepatol (n y)"] = "Gastroenterology & hepatology",
  ["gastroenterol hepatol (qué)"] = "Gastroenterology and hepatology",
  ["gastroenterol hepatol bed bench"] = "Gastroenterology and hepatology from bed to bench",
  ["gastroenterol insights"] = "Gastroenterology insights",
  ["gastroenterol j"] = "Gastroenterologisches Journal : Organ der Gesellschaft für Gastroenterologie der DDR",
  ["gastroenterol jpn"] = "Gastroenterologia Japonica",
  ["gastroenterol nurs"] = "Gastroenterology nursing : the official journal of the Society of Gastroenterology Nurses and Associates",
  ["gastroenterol pol"] = "Gastroenterologia Polska : organ Polskiego Towarzystwa Gastroenterologii",
  ["gastroenterol rep"] = "The gastroenterology report",
  ["gastroenterol rep (oxf)"] = "Gastroenterology report",
  ["gastroenterol res pract"] = "Gastroenterology research and practice",
  ["gastroenterology res"] = "Gastroenterology research",
  ["gastrointest cancer"] = "Gastrointestinal cancer : targets and therapy",
  ["gastrointest cancer (jersey city)"] = "Gastrointestinal cancer : research & therapy",
  ["gastrointest cancer res"] = "Gastrointestinal cancer research : GCR",
  ["gastrointest disord (basel)"] = "Gastrointestinal disorders (Basel, Switzerland)",
  ["gastrointest endosc"] = "Gastrointestinal endoscopy",
  ["gastrointest endosc clin n am"] = "Gastrointestinal endoscopy clinics of North America",
  ["gastrointest radiol"] = "Gastrointestinal radiology",
  ["gastrointest tumors"] = "Gastrointestinal tumors",
  ["gastronomica (berkeley calif)"] = "Gastronomica : the journal of food and culture",
  ["gaswaerme int"] = "Gaswaerme International",
  ["gates open res"] = "Gates open research",
  ["gateway herit"] = "Gateway heritage : quarterly journal of the Missouri Historical Society",
  ["gauss-ges göttingen mitt"] = "Gauss-Gesellschaft e",
  ["gavroche (evreux)"] = "Gavroche",
  ["gayana (concepc)"] = "Gayana",
  ["gayana bot"] = "Gayana. Botánica",
  ["gaz arch"] = "La Gazette des archives",
  ["gaz egypt paediatr assoc"] = "The Gazette of the Egyptian Paediatric Association",
  ["gaz grolier club"] = "Gazette of the Grolier Club. Grolier Club",
  ["gaz hop civ mil empire ottoman"] = "Gazette des hôpitaux civils et militaires de l'Empire ottoman",
  ["gaz law soc upper can"] = "Gazette - The Law Society of Upper Canada. Law Society of Upper Canada",
  ["gaz math"] = "Gazette des Mathematiciens",
  ["gaz med (paris)"] = "Gazette médicale (Paris, France)",
  ["gaz med fr"] = "Gazette médicale de France",
  ["gaz med limousine (1908)"] = "Gazette médicale limousine",
  ["gaz med lomb"] = "Gazzetta medica lombarda",
  ["gaz med port"] = "Gazeta médica portuguesa",
  ["gaz woda tech"] = "Gaz, woda; technika sanitarna",
  ["gazodinam i teploobmen"] = "Gazodinamika i Teploobmen",
  ["gazz chim ital"] = "Gazzetta chimica Italiana",
  ["gazz int med chir"] = "Gazzetta internazionale di medicina e chirurgia",
  ["gazz med ital"] = "Gazzetta medica italiana",
  ["gazz osp clin"] = "Gazzetta degli ospedali e delle cliniche",
  ["gazz sanit"] = "Gazzetta sanitaria",
  ["gbf monogr"] = "GBF Monographs",
  ["gcb bioenergy"] = "GCB Bioenergy",
  ["gd voix rech"] = "Les Grandes Voix de la Recherche",
  ["ge mag"] = "Gé-magazine",
  ["ge port j gastroenterol"] = "GE Portuguese journal of gastroenterology",
  ["geburtshilfe frauenheilkd"] = "Geburtshilfe und Frauenheilkunde",
  ["gedrag gezond"] = "Gedrag & gezondheid",
  ["gefahrst reinhalt luft"] = "Gefahrstoffe, Reinhaltung der Luft = Air quality control",
  ["gefahrstoffe - reinhalt luft"] = "Gefahrstoffe Reinhaltung der Luft",
  ["gefied  welt"] = "Gefiederte Welt, Die",
  ["gegenbaurs morphol jahrb"] = "Gegenbaurs morphologisches Jahrbuch",
  ["geka chiryo"] = "Geka chiryo. Surgical therapy",
  ["gelfand math sem"] = "The Gelfand Mathematical Seminars",
  ["gelis derg orta dogu tek univ"] = "Gelişme dergisi = Studies in development",
  ["geloof wet"] = "Geloof en wetenschap",
  ["gem int j geomath"] = "GEM - International Journal on Geomathematics",
  ["gem state rn news lett"] = "The Gem State RN news letter",
  ["gematol transfuziol"] = "Gematologiia i transfuziologiia",
  ["gems gemol"] = "Gems & Gemology",
  ["gen comp endocrinol"] = "General and comparative endocrinology",
  ["gen cytochem methods"] = "General cytochemical methods",
  ["gen dent"] = "General dentistry",
  ["gen diagn pathol"] = "General & diagnostic pathology",
  ["gen hosp psychiatry"] = "General hospital psychiatry",
  ["gen inf programme unisistnewsl"] = "General Information Programme, Unisist newsletter",
  ["gen laws r i 1956 r i"] = "General laws of Rhode Island, 1956 : completely annotated. Rhode Island",
  ["gen med (los angel)"] = "General medicine (Los Angeles, Calif.)",
  ["gen pharmacol"] = "General pharmacology",
  ["gen physiol biophys"] = "General physiology and biophysics",
  ["gen pract"] = "General practice",
  ["gen pract clin"] = "General practice clinics",
  ["gen psychiatr"] = "General psychiatry",
  ["gen psychol"] = "The general psychologist",
  ["gen relativ gravit"] = "General relativity and gravitation",
  ["gen relativ gravitation"] = "General Relativity and Gravitation",
  ["gen relativity gravitation"] = "General Relativity and Gravitation",
  ["gen surg laparosc news"] = "General surgery & laparoscopy news",
  ["gen syst"] = "General systems : yearbook of the Society for the Advancement of General Systems Theory",
  ["gen tech rep usda for serv"] = "General Technical Report USDA Forest Service",
  ["gen thorac cardiovasc surg"] = "General thoracic and cardiovascular surgery",
  ["gen topol appl"] = "General Topology and its Applications",
  ["gencho hiroshima igaku"] = "[Gencho Hiroshima igaku] The Hiroshima medical journal",
  ["gend action"] = "Gender action : a newsletter of the USAID Office of Women in Development",
  ["gend cg newsl"] = "The gender cg newsletter",
  ["gend dev"] = "Gender and development",
  ["gend educ"] = "Gender and education",
  ["gend genome"] = "Gender and the genome",
  ["gend hist"] = "Gender & history",
  ["gend manag"] = "Gender in management",
  ["gend med"] = "Gender medicine",
  ["gend place cult"] = "Gender, place and culture : a journal of feminist geography",
  ["gend soc"] = "Gender & society : official publication of Sociologists for Women in Society",
  ["gend technol dev"] = "Gender, technology and development",
  ["gend work organ"] = "Gender, work, and organization",
  ["gendai igaku"] = "[Gendai igaku] The current medicine",
  ["gendaishi kenkyu"] = "Gendaishi kenkyū",
  ["gender issues"] = "Gender issues",
  ["genders (austin tex)"] = "Genders",
  ["gene amplif anal"] = "Gene amplification and analysis",
  ["gene anal tech"] = "Gene analysis techniques",
  ["gene cell tissue"] = "Gene, cell and tissue",
  ["gene expr"] = "Gene expression",
  ["gene expr patterns"] = "Gene expression patterns : GEP",
  ["gene expression patterns"] = "Gene Expression Patterns",
  ["gene funct dis"] = "Gene function & disease",
  ["gene genome ed"] = "Gene and Genome Editing",
  ["gene geogr"] = "Gene geography : a computerized bulletin on human gene frequencies",
  ["gene regul syst bio"] = "Gene regulation and systems biology",
  ["gene regul syst biol"] = "Gene Regulation and Systems Biology",
  ["gene rep"] = "Gene reports",
  ["gene technol"] = "Gene technology",
  ["gene ther"] = "Gene therapy",
  ["gene ther mol biol"] = "Gene therapy & molecular biology",
  ["gene ther regul"] = "Gene Therapy and Regulation",
  ["gene transl bioinform"] = "Gene & translational bioinformatics",
  ["gene x"] = "Gene: X",
  ["gene x environ"] = "Gene x Environment",
  ["geneeskd bl"] = "Geneeskundige bladen uit kliniek en laboratorium voor de praktijk",
  ["geneeskd gids"] = "Geneeskundige gids",
  ["geneeskd sport"] = "Geneeskunde en sport",
  ["genes (basel)"] = "Genes",
  ["genes brain behav"] = "Genes, brain, and behavior",
  ["genes cancer"] = "Genes & cancer",
  ["genes cells"] = "Genes to cells : devoted to molecular & cellular mechanisms",
  ["genes chromosomes cancer"] = "Genes, chromosomes & cancer",
  ["genes dev"] = "Genes & development",
  ["genes dis"] = "Genes & diseases",
  ["genes environ"] = "Genes and environment : the official journal of the Japanese Environmental Mutagen Society",
  ["genes funct"] = "Genes and function",
  ["genes genet syst"] = "Genes & genetic systems",
  ["genes genomics"] = "Genes & genomics",
  ["genes immun"] = "Genes and immunity",
  ["genes nutr"] = "Genes & nutrition",
  ["genes to cells"] = "Genes to Cells",
  ["genes, brain behav"] = "Genes, Brain and Behavior",
  ["genes, chromosomes cancer"] = "Genes, Chromosomes and Cancer",
  ["genet anal"] = "Genetic analysis : biomolecular engineering",
  ["genet anal - biomol eng"] = "Genetic Analysis - Biomolecular Engineering",
  ["genet anal tech appl"] = "Genetic analysis, techniques and applications",
  ["genet aquat org"] = "Genetics of aquatic organisms",
  ["genet couns"] = "Genetic counseling (Geneva, Switzerland)",
  ["genet eng"] = "Genetic engineering (Academic Press)",
  ["genet eng (n y)"] = "Genetic engineering",
  ["genet eng biosaf j"] = "Genetic engineering and biosafety journal",
  ["genet eng biotechnol j"] = "Genetic engineering and biotechnology journal",
  ["genet eng news"] = "Genetic engineering news : GEN",
  ["genet epidemiol"] = "Genetic epidemiology",
  ["genet epidemiol suppl"] = "Genetic epidemiology. Supplement",
  ["genet epigenet"] = "Genetics & epigenetics",
  ["genet evol comput conf"] = "Genetic and Evolutionary Computation Conference : [proceedings]. Genetic and Evolutionary Computation Conference",
  ["genet evol comput ser"] = "Genetic and Evolutionary Computation Series",
  ["genet genet eng"] = "Genetics and genetic engineering",
  ["genet iber"] = "Genetica iberica",
  ["genet med"] = "Genetics in medicine : official journal of the American College of Medical Genetics",
  ["genet mol biol"] = "Genetics and molecular biology",
  ["genet mol res"] = "Genetics and molecular research : GMR",
  ["genet plant physiol"] = "Genetics and plant physiology",
  ["genet program evolvable mach"] = "Genetic programming and evolvable machines",
  ["genet psychol monogr"] = "Genetic psychology monographs",
  ["genet res"] = "Genetical research",
  ["genet res (camb)"] = "Genetics research",
  ["genet res int"] = "Genetics research international",
  ["genet resour crop evol"] = "Genetic resources and crop evolution",
  ["genet sel evol"] = "Genetics, selection, evolution : GSE",
  ["genet soc gen psychol monogr"] = "Genetic, social, and general psychology monographs",
  ["genet test"] = "Genetic testing",
  ["genet test mol biomarkers"] = "Genetic testing and molecular biomarkers",
  ["genet vaccines ther"] = "Genetic vaccines and therapy",
  ["genethics news"] = "GenEthics news : genetic engineering, ethics and the environment",
  ["genetic resour"] = "The Genetic resource",
  ["genetika (moscow)"] = "Genetika (Moscow)",
  ["geneva afr"] = "Genève-Afrique. Geneva-Africa",
  ["geneva pap risk ins theory"] = "Geneva Papers on Risk and Insurance Theory",
  ["geneva pap risk ins: issues practice"] = "Geneva Papers on Risk and Insurance: Issues and Practice",
  ["geneva pap risk insur issues pract"] = "The Geneva papers on risk and insurance. Issues and practice",
  ["geneva risk insur rev"] = "The Geneva risk and insurance review",
  ["genitourin med"] = "Genitourinary medicine",
  ["genom comput biol"] = "Genomics and computational biology",
  ["genom data"] = "Genomics data",
  ["genom discov"] = "Genomics discovery",
  ["genom genet"] = "Genomics and genetics : the official journal of Genetics Society of Thailand",
  ["genom soc policy"] = "Genomics, society, and policy",
  ["genome announc"] = "Genome announcements",
  ["genome biol"] = "Genome biology",
  ["genome biol evol"] = "Genome biology and evolution",
  ["genome dyn"] = "Genome dynamics",
  ["genome dyn stab"] = "Genome dynamics and stability",
  ["genome inform"] = "Genome informatics. International Conference on Genome Informatics",
  ["genome inform ser workshop genome inform"] = "Genome informatics. Workshop on Genome Informatics",
  ["genome instab dis"] = "Genome instability & disease",
  ["genome integr"] = "Genome integrity",
  ["genome lett"] = "Genome letters",
  ["genome med"] = "Genome medicine",
  ["genome res"] = "Genome research",
  ["genome sci technol"] = "Genome science & technology",
  ["genomic med"] = "Genomic medicine",
  ["genomic med, biomarkers, heath sci"] = "Genomic Medicine, Biomarkers, and Health Science",
  ["genomics appl biol"] = "Genomics and Applied Biology",
  ["genomics data"] = "Genomics Data",
  ["genomics inform"] = "Genomics & informatics",
  ["genomics insights"] = "Genomics insights",
  ["genomics proteomics bioinf"] = "Genomics Proteomics & Bioinformatcs",
  ["genomics proteomics bioinformatics"] = "Genomics, proteomics & bioinformatics",
  ["gensei dobutsugaku zasshi"] = "Gensei Dobutsugaku Zasshi",
  ["genshikaku kenkyu"] = "Genshikaku Kenkyu",
  ["geo eco trop"] = "Geo-Eco-Trop",
  ["geo spat inf sci"] = "Geo-spatial information science = Diqui kongjian xinxi kexue xuebao",
  ["geo-mar lett"] = "Geo-Marine Letters",
  ["geo-spatial inf sci"] = "Geo-spatial Information Science",
  ["geobios mem spec"] = "Geobios (Lyon, France). Mémoire spécial",
  ["geobot sel"] = "Geobotanica selecta",
  ["geocarto int"] = "Geocarto international",
  ["geochem explor environ anal"] = "Geochemistry: Exploration, Environment, Analysis",
  ["geochem geophys geosyst"] = "Geochemistry, Geophysics, Geosystems",
  ["geochem int"] = "Geochemistry International",
  ["geochem j"] = "Geochemical Journal",
  ["geochem perspect"] = "Geochemical Perspectives",
  ["geochem perspect lett"] = "Geochemical perspectives letters",
  ["geochem trans"] = "Geochemical transactions",
  ["geochem: explor environ anal"] = "Geochemistry: Exploration, Environment, Analysis",
  ["geochim cosmochim acta"] = "Geochimica et Cosmochimica Acta",
  ["geod cartogr"] = "Geodesy and Cartography",
  ["geod geodyn"] = "Geodesy and geodynamics",
  ["geodermal reg"] = "Geodermal Regional",
  ["geodin acta"] = "Geodinamica Acta",
  ["geoenergy sci eng"] = "Geoenergy Science and Engineering",
  ["geoenviron disasters"] = "Geoenvironmental Disasters",
  ["geoenvironmental disasters"] = "Geoenvironmental disasters",
  ["geofis int"] = "Geofisica Internacional",
  ["geofis pura appl"] = "Geofisica pura e applicata",
  ["geofys publ"] = "Geofysiske Publikasjoner",
  ["geofãs int"] = "GeofÃsica Internacional",
  ["geofã­s int"] = "GeofÃ­sica Internacional",
  ["geog glas"] = "Geografski glasnik = Bulletin de géographie",
  ["geoge mason univ civ rights law j"] = "George Mason University civil rights law journal",
  ["geogr anal"] = "Geographical analysis",
  ["geogr ann"] = "Geografiska Annaler",
  ["geogr ann ser a  phys geogr"] = "Geografiska Annaler: Series A, Physical Geography",
  ["geogr ann ser b"] = "Geografiska annaler. Series B, Human geography",
  ["geogr ann ser b  hum geogr"] = "Geografiska Annaler: Series B, Human Geography",
  ["geogr ber"] = "Geographische Berichte",
  ["geogr bern"] = "Geographica Bernensia",
  ["geogr bull"] = "Geography bulletin",
  ["geogr cas"] = "Geografický časopis = Geograficheskiĭ zhurnal = Geographical review",
  ["geogr compass"] = "Geography compass",
  ["geogr environ"] = "Geography and Environment",
  ["geogr environ sustainability"] = "Geography, Environment, Sustainability",
  ["geogr hel"] = "Geographica Helvetica",
  ["geogr helv"] = "Geographica helvetica",
  ["geogr j"] = "The Geographical journal",
  ["geogr mag"] = "The Geographical magazine",
  ["geogr med"] = "Geographia medica",
  ["geogr med suppl"] = "Geographia medica. Supplement = Geographia medica. Sonderband",
  ["geogr nat resour"] = "Geography and Natural Resources",
  ["geogr obs"] = "The Geographical observer",
  ["geogr perspect"] = "Geographical perspectives",
  ["geogr phys quat"] = "Geographie Physique et Quaternaire",
  ["geogr pol"] = "Geographia Polonica",
  ["geogr pregl"] = "Geografski pregled. Revue de géographie",
  ["geogr rech"] = "Géographie et recherche",
  ["geogr res"] = "Geographical Research",
  ["geogr rev"] = "Geographical review",
  ["geogr rev (oxf)"] = "Geography review",
  ["geogr rev india"] = "Geographical review of India",
  ["geogr rev jpn"] = "Geographical review of Japan",
  ["geogr rundsch"] = "Geographische Rundschau",
  ["geogr shkole"] = "Geografii︠a︡ v shkole (Moscow, Russia)",
  ["geogr surv"] = "Geographical survey",
  ["geogr sustainability"] = "Geography and Sustainability",
  ["geogr tidsskr"] = "Geografisk tidskrift",
  ["geogr tidsskr - dan j geogr"] = "Geografisk Tidsskrift - Danish Journal of Geography",
  ["geogr tydschr"] = "Geografisch tijdschrift",
  ["geogr vestn"] = "Geografski vestnik = Bulletin de la Société de géographie de Ljubljana",
  ["geogr z"] = "Geographische Zeitschrift",
  ["geoj tourism geosites"] = "Geojournal of Tourism and Geosites",
  ["geol belg"] = "Geologica Belgica",
  ["geol carpath"] = "Geologica Carpathica : the journal of Geological Institute of Slovak Academy of Sciences",
  ["geol ecol landscapes"] = "Geology, Ecology, and Landscapes",
  ["geol geofiz"] = "Geologiya i Geofizika",
  ["geol j"] = "Geological Journal",
  ["geol jahrb"] = "Geologisches Jahrbuch",
  ["geol mag"] = "Geological Magazine",
  ["geol mijnbouw"] = "Geologie en Mijnbouw",
  ["geol ore deposits"] = "Geology of Ore Deposits",
  ["geol rundsch"] = "Geologische Rundschau",
  ["geol saxonica"] = "Geologica Saxonica",
  ["geol soc am bull"] = "Geological Society of America Bulletin",
  ["geol soc london spec publ"] = "Geological Society, London, Special Publications",
  ["geol soc spec pap"] = "Geological Society of America Special Papers",
  ["geol soc spec pub"] = "Geological Society Special Publication",
  ["geol soc spec publ"] = "Geological Society Special Publication",
  ["geol soc specpubl"] = "Geological Society Special Publications",
  ["geol surv finl spec pap"] = "Geological Survey of Finland Special Paper",
  ["geol today"] = "Geology Today",
  ["geol ultraiectina (gedrukt)"] = "Geologica ultraiectina (Gedrukt). Rijksuniversiteit faculteit aardwetenschappen",
  ["geom comput"] = "Geometry and Computing",
  ["geom dedic"] = "Geometriae dedicata",
  ["geom dedicata"] = "Geometriae Dedicata",
  ["geom des publ"] = "Geometric Design Publications",
  ["geom flows"] = "Geometric Flows",
  ["geom funct anal"] = "Geometric and Functional Analysis",
  ["geom imaging comput"] = "Geometry, Imaging and Computing",
  ["geom integrability quantization"] = "Geometry, Integrability and Quantization",
  ["geom integrability quantization proc ser"] = "Geometry, Integrability and Quantization, Proceedings Series",
  ["geom mech"] = "Geometric Mechanics",
  ["geom sci inf (2013)"] = "Geometric science of information : first international conference, GSI 2013, Paris, France, August 28-30, 2013 : proceedings. Geometric Science of Information (1st : 2013 : Paris, France)",
  ["geom topol"] = "Geometry & Topology",
  ["geom topol monogr"] = "Geometry & Topology Monographs",
  ["geomag aeron"] = "Geomagnetism and Aeronomy",
  ["geomatics nat hazards risk"] = "Geomatics, Natural Hazards And Risk",
  ["geomech energy environ"] = "Geomechanics for Energy and the Environment",
  ["geomech eng"] = "Geomechanics and Engineering",
  ["geomech geoeng"] = "Geomechanics and Geoengineering",
  ["geomech geophys geo-energy geo-resour"] = "Geomechanics and Geophysics for Geo-Energy and Geo-Resources",
  ["geomech tunnelling"] = "Geomechanics and Tunnelling",
  ["geomicrobiol j"] = "Geomicrobiology journal",
  ["geomorphol relef processus environ"] = "Geomorphologie Relief Processus Environnement",
  ["geomorphology (amst)"] = "Geomorphology (Amsterdam, Netherlands)",
  ["geophys astrophys fluid dyn"] = "Geophysical and Astrophysical Fluid Dynamics",
  ["geophys astrophys fluid dynam"] = "Geophysical and Astrophysical Fluid Dynamics",
  ["geophys fluid dyn"] = "Geophysical Fluid Dynamics",
  ["geophys j"] = "Geophysical Journal",
  ["geophys j int"] = "Geophysical Journal International",
  ["geophys j r astron soc"] = "Geophysical Journal of the Royal Astronomical Society",
  ["geophys mag"] = "Geophysical Magazine",
  ["geophys monogr"] = "Geophysical Monograph",
  ["geophys prospect"] = "Geophysical Prospecting",
  ["geophys res lett"] = "Geophysical Research Letters",
  ["geophys suppl mon not r astron soc"] = "Geophysical Supplements to the Monthly Notices of the Royal Astronomical Society",
  ["geoplanet earth planet sci"] = "GeoPlanet: Earth and Planetary Sciences",
  ["george mason univ law rev"] = "George Mason University law review",
  ["george washington j int law econ"] = "The George Washington journal of international law and economics",
  ["george washington law rev"] = "The George Washington law review",
  ["george wright forum"] = "The George Wright forum",
  ["georget immgr law j"] = "Georgetown immigration law journal",
  ["georget j gend law"] = "The Georgetown journal of gender and the law",
  ["georget j leg ethics"] = "The Georgetown journal of legal ethics",
  ["georget mag"] = "Georgetown Magazine",
  ["georgetown dent j"] = "Georgetown dental journal",
  ["georgetown law j"] = "The Georgetown law journal",
  ["georgetown med bull"] = "Georgetown medical bulletin",
  ["georgetown univ right to life j"] = "The Georgetown University right-to-life journal",
  ["georgetown univ sch dent mirror"] = "Georgetown University School of Dentistry Mirror",
  ["georgia law rev"] = "Georgia law review (Athens, Ga. : 1966)",
  ["georgia state bar j"] = "Georgia State Bar journal",
  ["georgian math j"] = "Georgian Mathematical Journal",
  ["georgian med news"] = "Georgian medical news",
  ["georisk: assess manage risk eng syst geohazards"] = "Georisk: Assessment and Management of Risk for Engineered Systems and Geohazards",
  ["geosci can"] = "Geoscience Canada",
  ["geosci comm"] = "Geoscience Communication",
  ["geosci data j"] = "Geoscience Data Journal",
  ["geosci front"] = "Geoscience Frontiers",
  ["geosci instrum methods data syst"] = "Geoscientific Instrumentation, Methods and Data Systems",
  ["geosci instrum methods data syst discuss"] = "Geoscientific Instrumentation, Methods and Data Systems Discussions",
  ["geosci j"] = "Geosciences Journal",
  ["geosci lett"] = "Geoscience letters",
  ["geosci model dev"] = "Geoscientific model development",
  ["geosci model dev discuss"] = "Geoscientific Model Development Discussions",
  ["geosciences (basel)"] = "Geosciences",
  ["geospat health"] = "Geospatial health",
  ["geospatial health"] = "Geospatial health",
  ["geosphere (boulder)"] = "Geosphere (Boulder, Colo.)",
  ["geostand geoanal res"] = "Geostandards and Geoanalytical Research",
  ["geosynth int"] = "Geosynthetics international",
  ["geosyst eng"] = "Geosystem Engineering",
  ["geosyst geoenviron"] = "Geosystems and Geoenvironment",
  ["geosyst math"] = "Geosystems Mathematics",
  ["geotech geol eng"] = "Geotechnical and Geological Engineering",
  ["geotech test j"] = "Geotechnical Testing Journal",
  ["geotext geomembr"] = "Geotextiles and Geomembranes",
  ["geotherm energy"] = "Geothermal Energy",
  ["geotherm energy sci"] = "Geothermal Energy Science",
  ["gep news"] = "GEP News",
  ["ger comments"] = "German comments",
  ["ger econ rev"] = "German economic review",
  ["ger hist"] = "German history : the journal of the German History Society",
  ["ger j agric econ"] = "German Journal Of Agricultural Economics",
  ["ger j exercise sport res"] = "German Journal of Exercise and Sport Research",
  ["ger j ophthalmol"] = "German journal of ophthalmology",
  ["ger j vet res"] = "German Journal Of Veterinary Research",
  ["ger life lett"] = "German life and letters",
  ["ger med"] = "German medicine",
  ["ger med mon"] = "German medical monthly",
  ["ger med sci"] = "German medical science : GMS e-journal",
  ["ger polit"] = "German politics",
  ["ger res"] = "German Research",
  ["ger rom monatsschr"] = "Germanisch-Romanische Monatsschrift",
  ["ger stud rev"] = "German studies review",
  ["ger yearb int law"] = "Jahrbuch für internationales Recht. German yearbook of international law",
  ["geriatr aging"] = "Geriatrics & aging",
  ["geriatr gerontol int"] = "Geriatrics & gerontology international",
  ["geriatr nephrol urol"] = "Geriatric nephrology and urology",
  ["geriatr nur (lond)"] = "Geriatric Nursing",
  ["geriatr nur home care"] = "Geriatric Nursing and Home Care",
  ["geriatr nurs"] = "Geriatric nursing (New York, N.Y.)",
  ["geriatr nurs (lond)"] = "Geriatric nursing (London, England)",
  ["geriatr nurs (minneap)"] = "Geriatric nursing",
  ["geriatr nurs home care"] = "Geriatric nursing and home care",
  ["geriatr orthop surg rehabil"] = "Geriatric orthopaedic surgery & rehabilitation",
  ["geriatr psychol neuropsychiatr vieil"] = "Gériatrie et psychologie neuropsychiatrie du vieillissement",
  ["geriatrics (basel)"] = "Geriatrics (Basel, Switzerland)",
  ["gerlands beitr geophys"] = "Gerlands Beitrage zur Geophysik",
  ["germant crier"] = "Germantowne crier",
  ["gerontol clin (basel)"] = "Gerontologia clinica",
  ["gerontol geriatr educ"] = "Gerontology & geriatrics education",
  ["gerontol geriatr med"] = "Gerontology & geriatric medicine",
  ["gerontol soc"] = "Gérontologie et société : cahiers de la Fondation nationale de gérontologie",
  ["geropsych (bern)"] = "GeroPsych",
  ["gesch ges"] = "Geschichte und Gesellschaft (Vandenhoeck & Ruprecht)",
  ["gesch ggw (graz)"] = "Geschichte und Gegenwart",
  ["gesch hosp"] = "Geschichte des Hospitals",
  ["gesch reg"] = "Geschichte und Region = Storia e regione",
  ["gesch wiss unterr"] = "Geschichte in Wissenschaft und Unterricht",
  ["gesetzgeb bundes kantone"] = "Gesetzgebung des Bundes und der Kantone",
  ["gesnerus suppl"] = "Gesnerus. Supplement",
  ["gestalt theory"] = "Gestalt theory",
  ["gesture (amst)"] = "Gesture (Amsterdam, Netherlands)",
  ["gesund ing"] = "Gesundheits-Ingenieur",
  ["gesund umwelttech"] = "Gesundheits- und Umwelttechnik",
  ["gesund wohlfahrt"] = "Gesundheit und Wohlfahrt. Revue suisse d'hygiène",
  ["gesunde pflanzen"] = "Gesunde Pflanz.",
  ["gesundheit derending"] = "Gesundheit (Derendingen, Switzerland)",
  ["gesundheitsschutz umwelttech"] = "Gesundsheitsschutz und Umwelttechnik",
  ["geus bull"] = "GEUS Bulletin",
  ["gewaesserschutz, wasser, abwasser"] = "Gewaesserschutz, Wasser, Abwasser",
  ["gha today"] = "GHA today",
  ["ghaa j"] = "GHAA journal",
  ["ghana j sociol"] = "Ghana journal of sociology",
  ["ghana med j"] = "Ghana medical journal",
  ["ghana nurse"] = "The Ghanaian nurse",
  ["ghana off news bull"] = "Ghana official news bulletin",
  ["ghana soc sci j"] = "Ghana social science journal",
  ["gi ed proc"] = "GI-Edition. Proceedings",
  ["gibier faune sauvage"] = "Gibier, faune sauvage",
  ["gibrid vychisl mashiny kompleksy"] = "Akademiya Nauk Ukrainy",
  ["gidrokhim mater"] = "Gidrokhimicheskie Materialy",
  ["gidromekh (kiev)"] = "Natsional\\cprime naya Akademiya Nauk Ukrainy",
  ["giess geogr schr"] = "Giessener Geographische Schriften",
  ["gift child q"] = "The gifted child quarterly",
  ["gifted talent int"] = "Gifted and talented international",
  ["gifu shika gakkai zasshi"] = "Gifu Shika Gakkai zasshi = The Journal of Gifu Dental Society",
  ["gig sanit"] = "Gigiena i sanitariia",
  ["gig tr prof zabol"] = "Gigiena truda i professional'nye zabolevaniia",
  ["ginecol clin"] = "Ginecologia clinica",
  ["ginecol obstet (lima)"] = "Ginecologia y obstetricia",
  ["ginecol obstet mex"] = "Ginecología y obstetricia de México",
  ["ginekol pol"] = "Ginekologia polska",
  ["giorn econ"] = "Giornale degli Economisti e Annali di Economia",
  ["gisci remote sens"] = "GIScience & remote sensing",
  ["git lab j eur"] = "G.I.T. laboratory journal Europe",
  ["git labor fachz"] = "GIT Labor-Fachzeitschrift",
  ["git, fachz lab"] = "GIT, Fachzeitschrift für das Laboratorium",
  ["giyah pizishki"] = "Giyāh/pizishkī",
  ["giyah pizishki (ahvaz)"] = "Giyāh/pizishkī (Ahvāz, Iran)",
  ["glamorgan hist"] = "Stewart William's Glamorgan historian",
  ["gland surg"] = "Gland surgery",
  ["glas belgrad hig inst nr srb"] = "Glasnik. Belgrad. Higijenski institut NR Srbije",
  ["glas hem drus beogr"] = "Glasnik Hemijskog drustva Beograd = Bulletin de la Société chimique Beograd",
  ["glas mat"] = "Glasnik matematički",
  ["glas mat ser iii"] = "Glasnik Matematicki. Serija III",
  ["glas srp akad nauka [med]"] = "Glasnik, Srpska Akademija Nauka i Umetnosti, Odeljenje Medicinskih Nauka",
  ["glas srp akad nauka med"] = "Glas. Srpska akademija nauka i umetnosti. Odeljenje medicinskih nauka",
  ["glas zavodzdrav"] = "Glasnik. Republicki zavod za zdravstvenu zastitu SR Srbije",
  ["glasg dent j"] = "Glasgow dental journal",
  ["glasg math j"] = "Glasgow Mathematical Journal",
  ["glasgow math j"] = "Glasgow Mathematical Journal",
  ["glasgow med j"] = "Glasgow medical journal",
  ["glass ceram"] = "Glass and Ceramics",
  ["glass phys chem"] = "Glass Physics and Chemistry",
  ["glass res"] = "Glass Research",
  ["glass sci technol"] = "Glass Science and Technology",
  ["glass sci technol (offenbach, ger)"] = "Glass Science and Technology (Offenbach, Germany)",
  ["glass struct eng"] = "Glass Structures & Engineering",
  ["glass technol"] = "Glass Technology",
  ["glass technol: eur j glass sci technol, part a"] = "Glass Technology: European Journal of Glass Science and Technology, Part A",
  ["glaucoma update"] = "Glaucoma update",
  ["glaxo vol occas contrib sci art med"] = "The Glaxo volume; an occasional contribution to the science and art of medicine",
  ["glendale law rev"] = "Glendale law review",
  ["glob access std diagn"] = "Global access to STD diagnostics",
  ["glob adv health med"] = "Global advances in health and medicine",
  ["glob aidsnews"] = "Global AIDSnews : the newsletter of the World Health Organization Global Programme on AIDS",
  ["glob anesth perioper med"] = "Global anesthesia and perioperative medicine",
  ["glob bioeth"] = "Global bioethics = Problemi di bioetica",
  ["glob biogeochem cycles"] = "Global Biogeochemical Cycles",
  ["glob cardiol sci pract"] = "Global cardiology science & practice",
  ["glob cent news"] = "Global center news",
  ["glob chall"] = "Global challenges (Hoboken, NJ)",
  ["glob chang biol"] = "Global change biology",
  ["glob change biol"] = "Global Change Biology",
  ["glob change biol bioenergy"] = "Global change biology. Bioenergy",
  ["glob child health new rev"] = "Global child health news & review",
  ["glob dermatol"] = "Global dermatology",
  ["glob drugs ther"] = "Global drugs and therapeutics",
  ["glob ecol biogeogr"] = "Global ecology and biogeography : a journal of macroecology",
  ["glob ecol conserv"] = "Global ecology and conservation",
  ["glob environ change"] = "Global environmental change : human and policy dimensions",
  ["glob environ polit"] = "Global environmental politics",
  ["glob epidemiol"] = "Global epidemiology",
  ["glob food hist"] = "Global food history",
  ["glob food sec"] = "Global food security",
  ["glob health action"] = "Global health action",
  ["glob health commun"] = "Global health communication",
  ["glob health epidemiol genom"] = "Global health, epidemiology and genomics",
  ["glob health gov"] = "Global health governance : the scholarly journal for the new health security paradigm",
  ["glob health innov"] = "Global health innovation",
  ["glob health j"] = "Global health journal (Amsterdam, Netherlands)",
  ["glob health med"] = "Global health & medicine",
  ["glob health promot"] = "Global health promotion",
  ["glob health res policy"] = "Global health research and policy",
  ["glob health sci pract"] = "Global health, science and practice",
  ["glob heart"] = "Global heart",
  ["glob imaging insights"] = "Global imaging insights",
  ["glob impacts"] = "Global impacts : research for development review",
  ["glob issues"] = "Global issues (Washington, D.C.)",
  ["glob issues water policy"] = "Global Issues in Water Policy",
  ["glob j addict rehabil med"] = "Global journal of addiction & rehabilitation medicine",
  ["glob j adv eng technol sci"] = "Global journal of advanced engineering technologies and sciences",
  ["glob j adv res class mod geom"] = "Global Journal of Advanced Research on Classical and Modern Geometries",
  ["glob j anim breed genet"] = "Global journal of animal breeding and genetics",
  ["glob j biochem"] = "Global journal of biochemistry",
  ["glob j chem"] = "Global journal of chemistry",
  ["glob j community psychol pract"] = "Global journal of community psychology practice",
  ["glob j environ sci"] = "Global Journal of Environmental Sciences",
  ["glob j geol sci"] = "Global Journal of Geological Sciences",
  ["glob j health educ promot"] = "Global journal of health education and promotion",
  ["glob j health sci"] = "Global journal of health science",
  ["glob j med res"] = "Global journal of medical research",
  ["glob j obes diabetes metab syndr"] = "Global journal of obesity, diabetes and metabolic syndrome",
  ["glob j orthop res"] = "Global journal of orthopedics research",
  ["glob j pediatr neonatal care"] = "Global journal of pediatrics & neonatal care",
  ["glob j perioper med"] = "Global journal of perioperative medicine",
  ["glob j pure appl sci"] = "Global Journal of Pure and Applied Sciences",
  ["glob j reprod med"] = "Global journal of reproductive medicine",
  ["glob j res anal"] = "Global journal for research analysis",
  ["glob j surg"] = "Global journal of surgery",
  ["glob med genet"] = "Global medical genetics",
  ["glob med ther"] = "Global medicine and therapeutics",
  ["glob ment health (camb)"] = "Global mental health (Cambridge, England)",
  ["glob netw"] = "Global Networks-A Journal Of Transnational Affairs",
  ["glob netw (oxf)"] = "Global networks (Oxford, England)",
  ["glob pediatr health"] = "Global pediatric health",
  ["glob planet change"] = "Global and planetary change",
  ["glob policy"] = "Global policy",
  ["glob psychiatry"] = "Global psychiatry",
  ["glob public health"] = "Global public health",
  ["glob qual nurs res"] = "Global qualitative nursing research",
  ["glob reprod health"] = "Global reproductive health",
  ["glob soc"] = "Global society : journal of interdisciplinary international relations",
  ["glob soc policy"] = "Global social policy",
  ["glob soc welf"] = "Global social welfare : research, policy & practice",
  ["glob soc work"] = "Global social work",
  ["glob surg"] = "Global surgery (London)",
  ["glob transit"] = "Global transitions",
  ["glob vaccines immunol"] = "Global vaccines and immunology",
  ["glob vet"] = "Global veterinaria",
  ["global biogeochem cycles"] = "Global biogeochemical cycles",
  ["global bus econ rev"] = "Global Business and Economics Review",
  ["global challenges"] = "Global Challenges",
  ["global change biol"] = "Global Change Biology",
  ["global ecol biogeogr"] = "Global Ecology and Biogeography",
  ["global ecol conserv"] = "Global Ecology and Conservation",
  ["global econ rev"] = "Global Economic Review",
  ["global economy quart"] = "Global Economy Quarterly",
  ["global energy interconnect"] = "Global Energy Interconnection",
  ["global environ change"] = "Global Environmental Change",
  ["global environ change part b: environ hazards"] = "Global Environmental Change Part B: Environmental Hazards",
  ["global environ politics"] = "Global Environmental Politics",
  ["global finance j"] = "Global Finance Journal",
  ["global food secur"] = "Global Food Security",
  ["global futures dig"] = "Global futures digest",
  ["global health"] = "Globalization and health",
  ["global health epidemiol genomics"] = "Global Health, Epidemiology and Genomics",
  ["global health res policy"] = "Global Health Research and Policy",
  ["global implementation res appl"] = "Global Implementation Research and Applications",
  ["global j breast cancer res"] = "Global journal of breast cancer research",
  ["global j environ sci"] = "Global Journal of Environmental Sciences",
  ["global j flexible syst manage"] = "Global Journal of Flexible Systems Management",
  ["global j geol sci"] = "Global Journal of Geological Sciences",
  ["global j pure appl sci"] = "Global Journal of Pure and Applied Sciences",
  ["global planet change"] = "Global and Planetary Change",
  ["global secur: health, sci policy"] = "Global Security: Health, Science and Policy",
  ["global spine j"] = "Global spine journal",
  ["global stud q"] = "Global Studies Quarterly",
  ["global sustainability"] = "Global Sustainability",
  ["global transitions"] = "Global Transitions",
  ["global transitions proc"] = "Global Transitions Proceedings",
  ["glomerular dis"] = "Glomerular diseases",
  ["glycobiol insights"] = "Glycobiology insights",
  ["glycoconj j"] = "Glycoconjugate journal",
  ["glycoconjugate j"] = "Glycoconjugate Journal",
  ["gm crops"] = "GM crops",
  ["gm crops food"] = "GM crops & food",
  ["gmd res ser"] = "GMD Research Series",
  ["gmda bull"] = "GMDA bulletin",
  ["gmhc treat issues"] = "GMHC treatment issues : the Gay Men's Health Crisis newsletter of experimental AIDS therapies",
  ["gms curr top comput robot assist surg"] = "GMS current topics in computer and robot assisted surgery",
  ["gms curr top otorhinolaryngol head neck surg"] = "GMS current topics in otorhinolaryngology, head and neck surgery",
  ["gms health innov technol"] = "GMS health innovation and technologies",
  ["gms health technol assess"] = "GMS health technology assessment",
  ["gms hyg infect control"] = "GMS hygiene and infection control",
  ["gms infect dis"] = "GMS infectious diseases",
  ["gms interdiscip plast reconstr surg dgpw"] = "GMS Interdisciplinary plastic and reconstructive surgery DGPW",
  ["gms j med educ"] = "GMS journal for medical education",
  ["gms krankenhhyg interdiszip"] = "GMS Krankenhaushygiene interdisziplinär",
  ["gms ophthalmol cases"] = "GMS ophthalmology cases",
  ["gms z med ausbild"] = "GMS Zeitschrift für medizinische Ausbildung",
  ["gmu law rev"] = "GMU law review",
  ["gnu radio"] = "GNU radio",
  ["god univ arkhit stroit geod sofiya svit\\cdprime k ii mat mekh"] = "Universitet po Arkhitektura, Stroitelstvo i Geodeziya Sofiya",
  ["god vissh khim-tekhnol inst, sofia"] = "Godishnik na Visshiya Khimiko-Tekhnologicheski Institut, Sofiya",
  ["god vojnomed akad"] = "Godisnjak Vojnomedicinske akademije = Annual of the Military Medical Academy",
  ["god zb inst mat prir-mat fak univ kiril metodij skopje"] = "Godishen Zbornik na Institutot za Matematika. Prirodno-Matematichki. Fakultet Universitet “Sv. Kiril i Metodij”. Skopje. Annuaire de l’Institut des Mathematiques. Faculte des Sciences de l’Universite “Sts. Kiril et Metodij”. Skopje",
  ["god zb med fak skopje"] = "Godis̆en zbornik na Medicinskiot fakultet vo Skopje",
  ["godishnik vissh uchebn zaved prilozhna mat"] = "Godishnik na Visshite Uchebni Zavedeniya",
  ["godishnik vissh uchebn zaved tekhn fiz"] = "Godishnik na Visshite Uchebni Zavedeniya",
  ["godishnik vissh uchebn zaved tekhn mekh"] = "Godishnik na Visshite Uchebni Zavedeniya",
  ["gold bull"] = "Gold bulletin",
  ["gold gate law rev"] = "Golden Gate law review",
  ["gold gate univ law rev"] = "Golden Gate University law review. Golden Gate University. School of Law",
  ["gondwana res"] = "Gondwana Research",
  ["gongcheng shuxue xuebao"] = "Gongcheng Shuxue Xuebao. Chinese Journal of Engineering Mathematics",
  ["gongneng gaofenzi xuebao"] = "Gongneng Gaofenzi Xuebao",
  ["gonzaga law rev"] = "Gonzaga law review",
  ["good clin pract j"] = "Good clinical practice journal : GCP",
  ["good housekeeping"] = "Good housekeeping (New York, N.Y.)",
  ["gorman lect econ"] = "Gorman Lectures in Economics",
  ["gosp delo"] = "Gospital'noe delo",
  ["gospodarstvo istre"] = "Gospodarstvo Istre",
  ["goteb tandlak sallsk arsb"] = "Goteborgs Tandlakare-Sallskaps Arsbok",
  ["goteb univ dep socio"] = "Goteborg University - Department of Sociology Monograph",
  ["gothenbg pap theor linguist"] = "Gothenburg papers in theoretical linguistics",
  ["gott jahrb"] = "Göttinger Jahrbuch",
  ["gov inf q"] = "Government information quarterly",
  ["gov publ rev"] = "Government publications review (New York, N.Y. : 1982)",
  ["gov relat note"] = "Government relations note",
  ["governance (oxf)"] = "Governance (Oxford, England)",
  ["gps solutions"] = "GPS Solutions",
  ["grace hosp bul"] = "The Grace Hospital bulletin",
  ["grad fac philos j"] = "Graduate faculty philosophy journal",
  ["grad j math"] = "The Graduate Journal of Mathematics",
  ["grad stud math"] = "Graduate Studies in Mathematics",
  ["grad student ser phys"] = "Graduate Student Series in Physics",
  ["grad texts comput sci"] = "Graduate Texts in Computer Science",
  ["grad texts contemp phys"] = "Graduate Texts in Contemporary Physics",
  ["grad texts in math"] = "Graduate Texts in Mathematics",
  ["grad texts oper res"] = "Graduate Texts in Operations Research",
  ["grad texts phys"] = "Graduate Texts in Physics",
  ["graefes arch clin exp ophthalmol"] = "Graefe's archive for clinical and experimental ophthalmology = Albrecht von Graefes Archiv für klinische und experimentelle Ophthalmologie",
  ["graft (georget tex)"] = "Graft (Georgetown, Tex.)",
  ["grain oil sci technol"] = "Grain & Oil Science and Technology",
  ["grand rounds urol"] = "Grand rounds in urology",
  ["grands class gauthier-villars"] = "Les Grands Classiques Gauthier-Villars",
  ["grands noms sci tech"] = "Les Grands Noms de la Science et de la Technique",
  ["grants mag"] = "Grants magazine",
  ["grantsmanship cent news"] = "Grantsmanship Center News",
  ["granul matter"] = "Granular matter",
  ["granular comput"] = "Granular Computing",
  ["granular matter"] = "Granular Matter",
  ["graph gems ser"] = "Graphics Gems Series",
  ["graph learn med imaging (2019)"] = "Graph Learning in Medical Imaging : First International Workshop, GLMI 2019, Held in Conjunction with MICCAI 2019, Shenzhen, China, October 17, 2019, Proceedings",
  ["graph models"] = "Graphical models",
  ["graph scr"] = "Graphis scripta",
  ["graphene 2d mater"] = "Graphene and 2D Materials",
  ["graphene technol"] = "Graphene technology",
  ["graphical models"] = "Graphical Models",
  ["graphics visual comput"] = "Graphics and Visual Computing",
  ["graphs biomed image anal comput anat imaging genet (2017)"] = "Graphs in biomedical image analysis, computational anatomy and imaging genetics : First International Workshop, GRAIL 2017, 6th International Workshop, MFCA 2017, and Third International Workshop, MICGen 2017, held in conjunction with MICCAI 2017, Québec City, QC, Canada, September 10-14, 2017, proceedings. GRAIL (Workshop) (1st : 2017 : Québec, Québec)",
  ["graphs biomed image anal integr med imaging nonimaging modalities (2018)"] = "Graphs in biomedical image analysis and integrating medical imaging and non-imaging modalities : second International Workshop, GRAIL 2018 and first International Workshop, Beyond MIC 2018, held in conjunction with MICCAI 2018, Granada, Spain, September 20, 2018, Proceedings. GRAIL (Workshop) (2nd : 2018 : Granada, Spain)",
  ["graphs comb"] = "Graphs and combinatorics",
  ["graphs combin"] = "Graphs and Combinatorics",
  ["grasas aceites"] = "Grasas y Aceites",
  ["grasas aceites (seville)"] = "Grasas y Aceites (Seville)",
  ["grass forage sci"] = "Grass and Forage Science",
  ["grassl sci"] = "Grassland science",
  ["grassland res"] = "Grassland Research",
  ["grassland sci"] = "Grassland Science",
  ["grassroots dev"] = "Grassroots Development",
  ["gratz coll annual jew stud"] = "Gratz College annual of Jewish studies. Gratz College",
  ["gravit cosmol"] = "Gravitation and Cosmology",
  ["gravit space biol bull"] = "Gravitational and space biology bulletin : publication of the American Society for Gravitational and Space Biology",
  ["gravit space res"] = "Gravitational and space research : publication of the American Society for Gravitational and Space Research",
  ["gravitation cosmol"] = "Gravitation and Cosmology",
  ["gravitatsiya i teor otnositel\\cprime nosti"] = "Kazanskiy Gosudarstvennyy Universitet",
  ["grazer math ber"] = "Grazer Mathematische Berichte",
  ["great basin nat"] = "The Great Basin naturalist",
  ["great circle"] = "The Great circle : journal of the Australian Association for Maritime History",
  ["great ideas today"] = "The Great ideas today",
  ["great lakes entomol"] = "The Great Lakes entomologist",
  ["great ormond st j"] = "Great Ormond Street journal",
  ["great plains res"] = "Great Plains Research",
  ["greater milw dent bull"] = "The Greater Milwaukee dental bulletin",
  ["greater st louis dent soc bull"] = "The Greater St. Louis Dental Society bulletin",
  ["gredleriana (bolzano)"] = "Gredleriana",
  ["greek econ rev"] = "Greek economic review",
  ["greek roman byz stud (cambridge mass)"] = "Greek, Roman and Byzantine studies",
  ["green anal chem"] = "Green Analytical Chemistry",
  ["green chem"] = "Green Chemistry",
  ["green chem eng"] = "Green Chemical Engineering",
  ["green chem lett rev"] = "Green Chemistry Letters and Reviews",
  ["green chem process"] = "Green Chemical Processing",
  ["green energy environ"] = "Green Energy & Environment",
  ["green energy environ technol"] = "Green Energy and Environment Technology",
  ["green energy intell transp"] = "Green Energy and Intelligent Transportation",
  ["green energy resour"] = "Green Energy and Resources",
  ["green energy technol"] = "Green Energy and Technology",
  ["green mater"] = "Green Materials",
  ["green process synth"] = "Green Processing and Synthesis",
  ["green process synth,"] = "Green Processing and Synthesis",
  ["green schools catal q"] = "Green schools catalyst quarterly",
  ["green sustainable chem"] = "Green and Sustainable Chemistry",
  ["green synth catal"] = "Green Synthesis and Catalysis",
  ["green technol resilience sustainability"] = "Green Technology, Resilience, and Sustainability",
  ["green technol sustainability"] = "Green Technologies and Sustainability",
  ["greenhouse gases sci technol"] = "Greenhouse Gases: Science and Technology",
  ["greenl geosci"] = "Greenland Geoscience",
  ["grenoble sci"] = "Grenoble Sciences",
  ["grenzgeb med"] = "Grenzgebiete der Medizin",
  ["grenzgeb wiss"] = "Grenzgebiete der Wissenschaft",
  ["grey syst theory appl"] = "Grey Systems Theory and Application",
  ["grief matters"] = "Grief matters",
  ["griffin’s statist monogr ser"] = "Griffin’s Statistical Monograph Series",
  ["grma news"] = "GRMA news",
  ["groene witte kruis"] = "Het Groene en het Witte Kruis",
  ["groningen orient stud"] = "Groningen Oriental Studies",
  ["ground water"] = "Ground water",
  ["ground water monit rem"] = "Ground Water Monitoring and Remediation",
  ["ground water monit remediat"] = "Ground water monitoring & remediation",
  ["ground water monit rev"] = "Ground Water Monitoring Review",
  ["groundw sustain dev"] = "Groundwater for sustainable development",
  ["groundwater monit rem"] = "Groundwater Monitoring and Remediation",
  ["groundwater sustainable dev"] = "Groundwater for Sustainable Development",
  ["group (new york)"] = "Group",
  ["group decis negot"] = "Group decision and negotiation",
  ["group dyn"] = "Group dynamics : theory, research, and practice : the official journal of Division 49, Group Psychology and Group Psychotherapy of the American Psychological Association",
  ["group facil"] = "Group facilitation : a research & applications journal",
  ["group health j"] = "The Group health journal",
  ["group organ manag"] = "Group & organization management",
  ["group pract"] = "Group practice",
  ["group pract j"] = "Group practice journal",
  ["group process intergroup relat"] = "Group processes & intergroup relations : GPIR",
  ["groups complex cryptol"] = "Groups. Complexity. Cryptology",
  ["groups complexity cryptology"] = "Groups, Complexity, Cryptology",
  ["groups geom dyn"] = "Groups, Geometry, and Dynamics",
  ["growth change"] = "Growth and change",
  ["growth dev aging"] = "Growth, development, and aging : GDA",
  ["growth factors"] = "Growth factors (Chur, Switzerland)",
  ["growth form"] = "Growth and Form",
  ["growth genet horm"] = "Growth, genetics & hormones",
  ["growth horm igf res"] = "Growth hormone & IGF research : official journal of the Growth Hormone Research Society and the International IGF Research Society",
  ["growth hormone & igf research"] = "Growth Horm. IGF Res.",
  ["growth regul"] = "Growth regulation",
  ["grud serdechnososudistaia khir"] = "Grudnaia i serdechno-sosudistaia khirurgiia",
  ["grudn khir"] = "Grudnaia khirurgiia (Moscow, Russia)",
  ["grundfragen silikoseforsch"] = "Grundfragen aus der Silikoseforschung",
  ["grundkurs math"] = "Grundkurs Mathematik",
  ["grundkurs phys"] = "Grundkurs Physik",
  ["grundlagen math inform"] = "Grundlagen der Mathematik und Informatik",
  ["grundlehren math wiss"] = "Grundlehren der Mathematischen Wissenschaften",
  ["grundlehren text ed"] = "Grundlehren Text Editions",
  ["grundwasser (berl)"] = "Grundwasser (Berlin)",
  ["grundwissen math"] = "Grundwissen Mathematik",
  ["gsa today"] = "GSA today : a publication of the Geological Society of America",
  ["gstf int j bioinforma biotechnol"] = "GSTF international journal on bioinformatics & biotechnology",
  ["gstf int j comput"] = "GSTF international journal on computing",
  ["gstf j adv med res"] = "GSTF journal of advances in medical research",
  ["gstf j chem sci"] = "GSTF Journal of Chemical Sciences",
  ["gstf j media commun"] = "GSTF journal on media & communications",
  ["gstf j nurs health care"] = "GSTF journal of nursing and health care",
  ["gt concern"] = "The great concern",
  ["gu sheng wu xue bao"] = "Gu sheng wu xue bao = Acta palaeontologica Sinica",
  ["guam med assoc j"] = "The Guam Medical Association journal",
  ["guam rec"] = "Guam recorder",
  ["guang ming ri bao"] = "Guang ming ri bao",
  ["guang pu xue yu guang pu fen xi"] = "Guang pu xue yu guang pu fen xi = Guang pu",
  ["guangdong hai yang da xue xue bao"] = "Guangdong hai yang da xue xue bao = Journal of Guangdong ocean university",
  ["guangdong nong ye ke xue"] = "Guangdong nong ye ke xue = Guangdong agricultural sciences",
  ["guangdong yi xue"] = "Guangdong yi xue = Guangdong medical journal",
  ["guangpuxue yu guangpu fenxi"] = "Guangpuxue Yu Guangpu Fenxi",
  ["guangxi nongye shengwu kexue"] = "Guangxi nongye shengwu kexue = Journal of agricultural and biological science",
  ["guangxi sci"] = "Guangxi Sciences",
  ["guangxi xu mu shou yi"] = "Guangxi xu mu shou yi",
  ["guangxi yi ke da xue xue bao"] = "Guangxi yi ke da xue xue bao = Guangxi yike daxue xuebao = Acta academiae medicinae Guangxi",
  ["guangxi zhi wu"] = "Guangxi zhi wu",
  ["guangxue xuebao"] = "Guangxue Xuebao",
  ["guatem indig"] = "Guatemala indígena",
  ["guatem med"] = "Guatemala médica",
  ["guelph ichthyol rev"] = "Guelph ichthyology reviews",
  ["guerr mond conflits contemp"] = "Guerres mondiales et conflits contemporains",
  ["guid, navig, control"] = "Guidance, Navigation, and Control",
  ["guide microforms print author title"] = "Guide To Microforms In Print. Author, Title",
  ["guide prat"] = "Guide du praticien; édition du district parisien",
  ["guide repr"] = "Guide to reprints",
  ["guigoz sci rev"] = "Scientific review",
  ["guildhall misc"] = "The Guildhall miscellany",
  ["guildhall stud lond hist"] = "Guildhall studies in London history",
  ["guisuanyan xuebao"] = "Guisuanyan Xuebao",
  ["guizhou nong ye ke xue"] = "Guizhou nong ye ke xue = Journal of Guizhou agricultural science",
  ["guizhou shifan daxue xuebao ziran kexue ban"] = "Guizhou Shifan Daxue Xuebao",
  ["gujarat stat rev"] = "Gujarat statistical review",
  ["gujarat statist rev"] = "Gujarat Statistical Review",
  ["gulf caribb res"] = "Gulf and Caribbean research",
  ["gulf coast hist rev"] = "Gulf Coast historical review",
  ["gulf j math"] = "Gulf Journal of Mathematics",
  ["gulf j oncolog"] = "The Gulf journal of oncology",
  ["gulf mex sci"] = "Gulf of Mexico science",
  ["gunma j med sci"] = "The Gunma journal of medical sciences",
  ["guo ji bing du xue za zhi"] = "Guo ji bing du xue za zhi",
  ["guo ji fang she yi xue he yi xue za zhi"] = "Guo ji fang she yi xue he yi xue za zhi = International journal of radiation medicine and nuclear medicine",
  ["guo ji yi xue ji sheng chong bing za zhi"] = "Guo ji yi xue ji sheng chong bing za zhi",
  ["guo li taiwan shi fan da xue di li yan jiu suo di li yan jiu bao gaob"] = "Guo li Taiwan shi fan da xue Di li yan jiu suo di li yan jiu bao gao. Kuo li Tʻai-wan shih fan ta hsüeh. Ti li yen chiu so",
  ["guocheng gongcheng xuebao"] = "Guocheng Gongcheng Xuebao",
  ["guoli taiwan daxue zhiwu bing chonghai xeukan"] = "Guʹoli̕ Tʹaiwan da̕xuʹe zhʹiwu̕ bi̕ng cʹongha̕i xʹeukan = NTU phytopathologist & entomologist",
  ["guoshu xuebao"] = "Guoshu xuebao = Journal of fruit science",
  ["guru nanak j sociol"] = "Guru Nanak journal of sociology",
  ["gut liver"] = "Gut and liver",
  ["gut microbes"] = "Gut microbes",
  ["gut microbiome"] = "Gut Microbiome",
  ["gut pathog"] = "Gut pathogens",
  ["gut pathogens"] = "Gut Pathogens",
  ["gutenberg jahrb"] = "Gutenberg Jahrbuch",
  ["guthrie bull"] = "The Guthrie bulletin of the Donald Guthrie Medical Center",
  ["guthrie clin bull"] = "Guthrie Clinic bulletin",
  ["guthrie j donald guthrie found med res"] = "The Guthrie journal of the Donald Guthrie Foundation for Medical Research",
  ["guttmacher rep public policy"] = "The Guttmacher report on public policy",
  ["guys hosp gaz"] = "Guy's Hospital gazette",
  ["guys hosp rep"] = "Guy's Hospital reports",
  ["gyenbug uidai jabji"] = "Kyŏngbuk Uidae chapchi",
  ["gynakol endokrinol"] = "Gynäkologische endokrinologie",
  ["gynakol geburtshilfliche rundsch"] = "Gynakologisch-Geburtshilfliche Rundschau",
  ["gynakol prax"] = "Gynäkologische Praxis",
  ["gynakol rundsch"] = "Gynakologische Rundschau",
  ["gynecol endocrinol"] = "Gynecological endocrinology : the official journal of the International Society of Gynecological Endocrinology",
  ["gynecol invest"] = "Gynecologic investigation",
  ["gynecol minim invasive ther"] = "Gynecology and minimally invasive therapy",
  ["gynecol obstet (paris)"] = "Gynécologie et obstétrique",
  ["gynecol obstet (sunnyvale)"] = "Gynecology & obstetrics (Sunnyvale, Calif.)",
  ["gynecol obstet fertil"] = "Gynécologie, obstétrique & fertilité",
  ["gynecol obstet fertil senol"] = "Gynécologie, obstétrique, fertilité & sénologie",
  ["gynecol obstet invest"] = "Gynecologic and obstetric investigation",
  ["gynecol obstet res"] = "Gynecology and obstetrics research : open journal",
  ["gynecol oncol"] = "Gynecologic oncology",
  ["gynecol oncol case rep"] = "Gynecologic oncology case reports",
  ["gynecol oncol rep"] = "Gynecologic oncology reports",
  ["gynecol oncol res pract"] = "Gynecologic oncology research and practice",
  ["gynecol pelvic med"] = "Gynecology and pelvic medicine",
  ["gynecol prat"] = "Gynécologie pratique",
  ["gynecol reprod endocrinol"] = "Gynecology and reproductive endocrinology",
  ["gynecol reprod health"] = "Gynecology & reproductive health",
  ["gynecol rev gynecol"] = "Gynécologie (Paris, France : 1993)",
  ["gynecol surg"] = "Gynecological surgery",
  ["gyoruigaku zasshi"] = "Gyoruigaku Zasshi",
  ["géogr phys quat"] = "Géographie Physique et Quaternaire",
  ["göttinger universitätsschr ser a schr"] = "Göttinger Universitätsschriften",
  ["h2open j"] = "H2Open Journal",
  ["habitat debate"] = "Habitat debate",
  ["habitat int"] = "Habitat international",
  ["habitation (elmsford)"] = "Habitation (Elmsford, N.Y.)",
  ["hacet bull nat sci eng ser b"] = "Hacettepe Bulletin of Natural Sciences and Engineering",
  ["hacet j math stat"] = "Hacettepe Journal of Mathematics and Statistics",
  ["hacettepe bull med surg"] = "Hacettepe bulletin of medicine/surgery",
  ["hacettepe sosyal bilimler derg"] = "Hacettepe sosyal ve beşerî bilimler dergisi",
  ["hadassah mag"] = "Hadassah magazine",
  ["hadronic j"] = "Hadronic Journal",
  ["hadronic j suppl"] = "Hadronic Journal Supplement",
  ["hadronic press monogr math"] = "Hadronic Press Monographs in Mathematics",
  ["hadtort kozl"] = "Hadtörténelmi közlemények",
  ["haematol blood transfus"] = "Haematology and blood transfusion",
  ["haematol lat"] = "Haematologica Latina",
  ["haematol meet rep"] = "Haematology meeting reports",
  ["haematol pol"] = "Haematologica Polonica",
  ["haematol rep"] = "Haematologica reports",
  ["haematologia (budap)"] = "Haematologia",
  ["haerbin yi ke da xue xue bao"] = "Ha'erbin yi ke da xue xue bao = Journal of Harbin Medical University = Harbin yike daxue xuebao",
  ["hahnemann hosp tidings"] = "Hahnemann Hospital tidings",
  ["hahnemann mon"] = "Hahnemann monthly",
  ["hai yang hu zhao tong bao"] = "Hai yang hu zhao tong bao = Transactions of oceanology and limnology",
  ["hai yang xue bao"] = "Acta oceanologica Sinica = Hai yang xue bao",
  ["haiyang kexue"] = "Haiyang kexue = Marine sciences",
  ["haiyang yu huzhao"] = "Hǎiyáng yǔ húzhǎo = Oceanologia et limnologia sinica",
  ["haiyang yuye"] = "Haiyang yuye = Marine fisheries",
  ["hak-sul chi"] = "Kŏn’guk Taehakkyo, Haksul Yŏn’guwon",
  ["halls j health"] = "Hall's journal of health",
  ["hallym int j aging hija"] = "Hallym international journal of aging : HIJA",
  ["halsburys statut engl g b"] = "Halsbury's statutes of England. Great Britain",
  ["halve maen"] = "De Halve maen",
  ["hamatol bluttransfus"] = "Hamatologie und Bluttransfusion",
  ["hamb arztebl"] = "Hamburger Ärzteblatt",
  ["hamdard islam"] = "Hamdard Islamicus : quarterly journal of the Hamdard National Foundation, Pakistan",
  ["hamdard med"] = "Hamdard medicus",
  ["hamline law rev"] = "Hamline law review",
  ["hamyln symp med robot (2016)"] = "The Hamlyn Symposium on Medical Robotics : Proceedings. Hamlyn Symposium on Medical Robotics (9th : 2016 : London, England)",
  ["han gug eoryu hag hoeji"] = "Han-gug eoryu hag-hoeji = Korean Journal of Ichthyology",
  ["han xue yan jiu"] = "Han xue yan jiu = Chinese studies",
  ["hand (n y)"] = "Hand (New York, N.Y.)",
  ["hand clin"] = "Hand clinics",
  ["hand microsurg"] = "Hand & microsurgery",
  ["hand surg"] = "Hand surgery : an international journal devoted to hand and upper limb surgery and related research : journal of the Asia-Pacific Federation of Societies for Surgery of the Hand",
  ["hand surg rehabil"] = "Hand surgery & rehabilitation",
  ["hand ther"] = "Hand therapy",
  ["handb behav neurosci"] = "Handbook of behavioral neuroscience",
  ["handb clin neurol"] = "Handbook of clinical neurology",
  ["handb exp pharmacol"] = "Handbook of experimental pharmacology",
  ["handb graph gramm comput graph transform"] = "Handbook of Graph Grammars and Computing by Graph Transformation",
  ["handb health technol assess int"] = "Handbook : 3rd annual meeting. Health Technology Assessment International. Meeting (3rd : 2006 : Adelaide, S. Aust.)",
  ["handb hist log"] = "Handbook of the History of Logic",
  ["handb log comput sci"] = "Handbook of Logic in Computer Science",
  ["handb numer anal"] = "Handbook of Numerical Analysis",
  ["handb philos log"] = "Handbook of Philosophical Logic",
  ["handb philos sci"] = "Handbook of the Philosophy of Science",
  ["handb veg sci"] = "Handbook of Vegetation Science",
  ["handbook internat doc info"] = "Handbook of International Documentation and Information",
  ["handbook of statist"] = "Handbook of Statistics",
  ["handbooks in econom"] = "Handbooks in Economics",
  ["handbooks oper res management sci"] = "Handbooks in Operations Research and Management Science",
  ["handbuch phys"] = "Handbuch der Physik",
  ["handchir mikrochir plast chir"] = "Handchirurgie, Mikrochirurgie, plastische Chirurgie : Organ der Deutschsprachigen Arbeitsgemeinschaft für Handchirurgie : Organ der Deutschsprachigen Arbeitsgemeinschaft für Mikrochirurgie der Peripheren Nerven und Gefässe : Organ der Vereinigung der Deutschen Plastischen Chirurgen",
  ["handel maatsch geschied oudheidk gent"] = "Handelingen der Maatschappij voor Geschiedenis en Oudheidkunde te Gent. Société d'histoire et d'archéologie de Gand",
  ["handeln entscheid komplexen ökon situat"] = "Handeln und Entscheiden in Komplexen Ökonomischen Situationen",
  ["handicap (washington)"] = "Handicap; the news magazine for the nation's handicapped",
  ["hangug eobyeong haghoeji"] = "Hangug eobyeong haghoeji",
  ["hangug haeyang haghoeji"] = "Han'gug haeyang haghoeji. Korean society of oceanography",
  ["hangug hwangyeong saengmul haghoeji"] = "Hang̕ug hwang̕yeong saengmul haghoeji",
  ["hangug hwangyeong saengtae haghoeji"] = "Han'gug hwan'gyeong saengtae haghoeji = Korean journal of environment and ecology",
  ["hangug jaweon sigmul haghoeji"] = "Han'gug jaweon sig'mul haghoeji = Korean Journal of Plant Resources",
  ["hangug nimhag hoi ji"] = "Hangug n̊imhag hoi ji",
  ["hangug nonghwahag hoeji"] = "Han'guk Nonghwa Hakhoe chi",
  ["hangug sigmul byeongri haghoeji"] = "Korean journal of plant pathology",
  ["hangug sigpum wisaeng anjeonseong haghoeji"] = "Han'gug sigpum wi'saeng anjeonseong haghoeji",
  ["hangug susan haghoi ji"] = "Han'guk Susan Hakhoe chi = Bulletin of the Korean Fisheries Society",
  ["hangug yangbong haghoeji"] = "Han'gug yangbong haghoeji",
  ["hangug yangsig haghoeji"] = "Han-gug yangsig hag-hoeji",
  ["hangug yugsu haghoeji"] = "Han'gug yugsu haghoeji",
  ["hanguk chuksan hakhoe chi"] = "Han'guk Ch'uksan Hakhoe chi = Korean journal of animal science",
  ["hanguk haeyang paio hakhoe"] = "Han'guk Haeyang Paio Hakhoe",
  ["hanguk hwangyong nonghakhoe chi"] = "Han'guk Hwan'gyŏng Nonghakhoe chi = Korean journal of environmental agriculture",
  ["hanguk jakmul hakhoe chi"] = "Hang'uk jakmul hakhoe chi",
  ["hanguk kachuk wisaeng hakhoe chi"] = "Han'guk Kach'uk Wisaeng Hakhoe chi",
  ["hanguk kwahaksa hakhoeji"] = "Han'guk Kwahaksa Hakhoeji",
  ["hanguk kyun hakoe chi"] = "Hanʼguk Kyunhakhoe chi",
  ["hanguk misaengmul saengmyong konghakhoe chi"] = "Han'guk Misaengmul, Saengmyŏng Konghakhoe chi = Korean journal of microbiology and biotechnology",
  ["hanguk nonghwa hakhoe chi"] = "Han’guk Nonghwa Hakhoe chi",
  ["hanguk nonyonhak"] = "Han'guk nonyŏnhak = Journal of Korea Gerontological Society",
  ["hanguk saenghwahakhoe chi"] = "Hanʼguk Saenghwahakhoe chi = the Korean biochemical journal",
  ["hanguk sikpum chori kwahakhoe chi"] = "Han’guk Sikp’um Chori Kwahakhoe chi",
  ["hanguk sikpum jeojang yutong hakoeji"] = "Han-guk sikpum jeojang yutong hakoeji",
  ["hanguk sikpum yongyang kwahakhoe chi"] = "Han'guk Sikp'um Yŏngyang Kwahakhoe chi",
  ["hanguk simni hakhoe chi kongang"] = "Han'guk Simni Hakhoe chi. Kŏn'gang = The Korean journal of health psychology",
  ["hanguk tongmul pullyu hakhoe chi"] = "Han'guk Tongmul Pullyu Hakhoe chi",
  ["hanguk toyang piryo hakhoe chi"] = "Han'guk T'oyang Piryo Hakhoe chi. Han'guk T'oyang Piryo Hakhoe",
  ["hanguk uiyak"] = "[Hanguk uiyak] Korean journal of medicine",
  ["hanguk ungyong konchung hakhoe chi"] = "Han'guk Ŭngyong Konchʻung Hakhoe chi = Korean journal of applied entomology",
  ["hanguk ungyong saengmyong hwahakhoe chi"] = "Han'guk Ŭngyong Saengmyŏng Hwahakhoe chi = Journal of the Korean Society for Applied Biological Chemistry",
  ["hanguk wonye hakhoe chi"] = "Han'guk Wŏnye Hakhoe chi. Han'guk Wŏnye Hakhoe",
  ["hanguk yakyong changmul hakhoe chi"] = "Han'guk Yakyong Changmul Hakhoe chi = Korean journal of medicinal crop science",
  ["hanguk yangseo pachungnyu hakoeji"] = "Han-guk yang-seo pachungnyu hakoeji",
  ["hann abh gesch med natwiss"] = "Hannoversche Abhandlungen zur Geschichte der Medizin und der Naturwissenschaften",
  ["hansenol int"] = "Hansenologia internationalis",
  ["han’guk saenghwa hakhoechi"] = "Han’guk Saenghwa Hakhoechi",
  ["han’guk t’oyang piryo hakhoechi"] = "Han’guk T’oyang Piryo Hakhoechi",
  ["haptics (2010)"] = "Haptics : generating and perceiving tangible sensations : international conference, EuroHaptics 2010, Amsterdam, July 8-10, 2010 : proceedings. EuroHaptics 2010 (2010 : Amsterdam, Netherlands)",
  ["haptics (2014)"] = "Haptics : neuroscience, devices, modeling, and applications : 9th International Conference, EuroHaptics 2014, Versailles, France, June 24-26, 2014 : proceedings",
  ["haptics (2018)"] = "Haptics : science, technology, and applications : 11th International Conference, EuroHaptics 2018, Pisa, Italy, June 13-16, 2018, proceedings. EuroHaptics Conference (11th : 2018 : Pisa, Italy)",
  ["harb dent log"] = "Harbor dental log",
  ["hard tissue"] = "Hard tissue",
  ["hardy-ramanujan j"] = "Hardy-Ramanujan Journal",
  ["harefuah (tel aviv)"] = "Harefuah. Foreign ed",
  ["harlem hosp bull"] = "The Harlem Hospital bulletin",
  ["harlem hosp bull (n y)"] = "The Harlem Hospital bulletin",
  ["harm reduct j"] = "Harm reduction journal",
  ["harmful algae"] = "Harmful algae",
  ["harmful algae 2002 (2002)"] = "Harmful algae 2002 : proceedings of the Xth International Conference on Harmful Algae, St. Pete Beach, Florida, USA, October 21-25, 2002. International Conference on Harmful Algae (10th : 2002 : St. Pete Beach, Florida)",
  ["harmful algae 2012 (2012)"] = "Harmful algae 2012 : proceedings of the 15th International Conference on Harmful Algae : October 29 - November 2, 2012, CECO, Changwon, Gyeongnam, Korea. International Conference on Harmful Algae (15th : 2012 : Changwon, Gyeongnam, Korea)",
  ["harmful algae news"] = "Harmful algae news",
  ["harmonia (etagnac)"] = "Harmonia : coccinellidae du monde",
  ["harofe haivri heb med j"] = "Harofé haivri. The Hebrew medical journal",
  ["harper hosp bull"] = "Harper Hospital bulletin",
  ["harpers (n y n y)"] = "Harper's",
  ["hartford hosp bull"] = "Hartford Hospital bulletin",
  ["harv aids rev"] = "Harvard AIDS review",
  ["harv blacklett j"] = "The Harvard blackletter journal",
  ["harv blacklett law j"] = "Harvard blackletter law journal",
  ["harv bus rev"] = "Harvard business review",
  ["harv civ rights-civil lib law rev"] = "Harvard civil rights-civil liberties law review",
  ["harv data sci rev"] = "Harvard data science review",
  ["harv dent alumni bull"] = "Harvard dental alumni bulletin",
  ["harv dent bull"] = "Harvard dental bulletin",
  ["harv educ rev"] = "Harvard educational review",
  ["harv health lett"] = "Harvard health letter",
  ["harv heart lett"] = "Harvard heart letter : from Harvard Medical School",
  ["harv hum rights j"] = "Harvard human rights journal",
  ["harv j asiat stud"] = "Harvard journal of Asiatic studies",
  ["harv j law public policy"] = "Harvard journal of law & public policy",
  ["harv j law technol"] = "Harvard journal of law & technology",
  ["harv j minor public health"] = "The Harvard journal of minority public health",
  ["harv law rev"] = "Harvard law review",
  ["harv libr bull"] = "Harvard library bulletin",
  ["harv mag"] = "Harvard magazine",
  ["harv med alumni bull"] = "Harvard medical alumni bulletin",
  ["harv med stud rev"] = "Harvard medical student review",
  ["harv mens health watch"] = "Harvard men's health watch",
  ["harv ment health lett"] = "The Harvard mental health letter",
  ["harv pap bot"] = "Harvard papers in botany",
  ["harv public health alumni bull"] = "Harvard public health alumni bulletin",
  ["harv public health rev (camb)"] = "Harvard public health review (Cambridge, Mass.)",
  ["harv rev psychiatry"] = "Harvard review of psychiatry",
  ["harv s ukrain st"] = "Harvard Series in Ukrainian Studies",
  ["harv stud classic philol"] = "Harvard studies in classical philology",
  ["harv theol rev"] = "The Harvard theological review",
  ["harv ukr stud"] = "Harvard Ukrainian studies",
  ["harv univ cent math sci appl ser math"] = "Harvard University Center of Mathematical Sciences and Applications (CMSA) Series in Mathematics",
  ["harv women's law j"] = "Harvard women's law journal",
  ["harv womens health watch"] = "Harvard women's health watch",
  ["harv women’s law j"] = "Harvard women’s law journal",
  ["harvard bus rev"] = "Harvard Business Review",
  ["harvard data sci rev"] = "Harvard Data Science Review",
  ["harvard health policy rev"] = "Harvard health policy review : a student publication of the Harvard Interfaculty Initiative in Health Policy",
  ["harvard int law j"] = "Harvard international law journal",
  ["harvard int rev"] = "Harvard international review",
  ["harvard j legis"] = "Harvard journal on legislation",
  ["harvard pap botan"] = "Harvard Papers in Botany",
  ["harvey lect"] = "Harvey lectures",
  ["harz z"] = "Harz-Zeitschrift",
  ["hasl rep"] = "HASL [reports]. U.S. Atomic Energy Commission",
  ["hastings cent rep"] = "The Hastings Center report",
  ["hastings cent stud"] = "Hastings Center Studies",
  ["hastings constit law q"] = "Hastings constitutional law quarterly",
  ["hastings int comp law rev"] = "Hastings international and comparative law review",
  ["hastings law j"] = "The Hastings law journal",
  ["hastings womens law j"] = "Hastings women's law journal",
  ["hattori shokubutsu kenkyujo hokoku|j hattori bot lab"] = "The Journal of the Hattori Botanical Laboratory",
  ["havemeyer found monogr ser"] = "Havemeyer Foundation Monograph Series. Dorothy Russell Havemeyer Foundation",
  ["haw bar j"] = "Hawaii bar journal (Honolulu, Hawaii : 1963)",
  ["hawaii acad sci honol"] = "Proceedings of the annual meeting. Hawaiian Academy of Science, Honolulu",
  ["hawaii dent assoc j"] = "Hawaii Dental Association journal",
  ["hawaii dent j"] = "Hawaii dental journal",
  ["hawaii j health soc welf"] = "Hawai'i journal of health & social welfare",
  ["hawaii j hist"] = "The Hawaiian journal of history",
  ["hawaii j med public health"] = "Hawai'i journal of medicine & public health : a journal of Asia Pacific Medicine & Public Health",
  ["hawaii j public health"] = "Hawaii journal of public health",
  ["hawaii med j"] = "Hawaii medical journal",
  ["hawaii nurse"] = "The Hawaii nurse : the official monthly newsletter of the Hawaii Nurses' Association",
  ["hawaii nurse (honol)"] = "Hawaii nurse (Honolulu, Hawaii : 1994)",
  ["hawaii nurses pipeline"] = "Hawaii nurses pipeline",
  ["hawaii revis statut hawaii"] = "Hawaii revised statutes : comprising the statutes of the State of Hawaii, consolidated, revised, and annotated. Hawaii",
  ["hayati j biosci"] = "HAYATI Journal of Biosciences",
  ["hayes hist j"] = "Hayes historical journal",
  ["hazard control qual rep"] = "Hazards control quarterly report",
  ["hazard waste hazard mater"] = "Hazardous Waste and Hazardous Materials",
  ["hba lect notes math"] = "HBA Lecture Notes in Mathematics",
  ["hbi rep"] = "HBI report",
  ["hci int 2018 posters ext abstr (2018)"] = "HCI International 2018 - posters' extended abstracts : 20th International Conference, HCI International 2018, Las Vegas, NV, USA, July 15-20, 2018, Proceedings, Part I. International Conference on Human-Computer Interaction (20th : 2018 : Las Vegas, Nev.)",
  ["hci int 2019 posters (2019)"] = "HCI International 2019 - Posters : 21st international conference",
  ["hci int 2020 late break posters (2020)"] = "HCI International 2020 - late breaking posters : 22nd International Conference, HCII 2020, Copenhagen, Denmark, July 19-24, 2020, Proceedings, Part I",
  ["hda now"] = "HDA now",
  ["he nong xue bao"] = "He nong xue bao = Acta agriculturae nucleatae Sinica",
  ["head face med"] = "Head & face medicine",
  ["head neck"] = "Head & neck",
  ["head neck oncol"] = "Head & neck oncology",
  ["head neck pathol"] = "Head and neck pathology",
  ["head neck surg"] = "Head & neck surgery",
  ["head neck tumor segm (2020)"] = "Head and Neck Tumor Segmentation : First Challenge, HECKTOR 2020, Held in Conjunction with MICCAI 2020, Lima, Peru, October 4, 2020, Proceedings",
  ["headache: j head face pain"] = "Headache: The Journal of Head and Face Pain",
  ["heal light"] = "Healing Light",
  ["health (london)"] = "Health (London, England : 1997)",
  ["health (n y)"] = "Health (Family Media, Inc.)",
  ["health aff"] = "Health affairs",
  ["health aff (millwood)"] = "Health affairs (Project Hope)",
  ["health after 50 sci am consum health"] = "Health after 50 with Scientific American Consumer Health",
  ["health army"] = "Health of the Army. United States. Surgeon-General's Office",
  ["health beat"] = "Health beat (Manila, Philippines)",
  ["health behav policy rev"] = "Health behavior and policy review",
  ["health behav res"] = "Health behavior research",
  ["health bull"] = "Health bulletin. Kern Co., Calif. Department of Public Health",
  ["health bull (edinb)"] = "Health Bulletin",
  ["health bull (melb)"] = "Health bulletin",
  ["health bull (raleigh)"] = "The Health bulletin",
  ["health bull teach"] = "Health bulletin for teachers",
  ["health can soc"] = "Health and Canadian society (Winnipeg, Man.)",
  ["health care (don mills)"] = "Health care",
  ["health care anal"] = "Health care analysis : HCA : journal of health philosophy and policy",
  ["health care can"] = "Health care in Canada",
  ["health care cost reengineering rep"] = "Health care cost reengineering report",
  ["health care curr rev"] = "Health care. Current reviews",
  ["health care dimen"] = "Health care dimensions",
  ["health care educ"] = "Health care education",
  ["health care ethics usa"] = "Health care ethics USA : a publication of the Center for Health Care Ethics",
  ["health care financ rev"] = "Health care financing review",
  ["health care financ rev annu suppl"] = "Health care financing review. Annual supplement",
  ["health care financ rev stat suppl"] = "Health care financing review. Statistical supplement",
  ["health care financ trends"] = "Health care financing trends",
  ["health care food nutr focus"] = "Health care food & nutrition focus",
  ["health care innov"] = "Health care innovations : the journal of the American Association of Preferred Provider Organizations",
  ["health care law mon"] = "Health care law monthly",
  ["health care law newsl"] = "Health care law newsletter",
  ["health care manag"] = "Health care management (Philadelphia, Pa.)",
  ["health care manag (frederick)"] = "The health care manager",
  ["health care manag sci"] = "Health care management science",
  ["health care manage rev"] = "Health care management review",
  ["health care manage sci"] = "Health Care Management Science",
  ["health care mark"] = "Health care marketer",
  ["health care mark target market"] = "Health care marketer & target market",
  ["health care newsl"] = "Health care newsletter",
  ["health care plann mark"] = "Health care planning & marketing : HCP&M",
  ["health care reform week"] = "Health care reform week",
  ["health care secur saf manage"] = "Health care security and safety management",
  ["health care strateg manage"] = "Health care strategic management",
  ["health care superv"] = "The Health care supervisor",
  ["health care syst"] = "Health care systems",
  ["health care week"] = "Health care week",
  ["health care women int"] = "Health care for women international",
  ["health cent j"] = "The Health center journal",
  ["health commun"] = "Health communication",
  ["health commun informatics"] = "Health communications and informatics",
  ["health cost manage"] = "Health cost management",
  ["health data manag"] = "Health data management",
  ["health data sci"] = "Health Data Science",
  ["health devices"] = "Health devices",
  ["health econ"] = "Health economics",
  ["health econ outcome res"] = "Health economics & outcome research : Open access",
  ["health econ policy law"] = "Health economics, policy, and law",
  ["health econ rev"] = "Health economics review",
  ["health educ"] = "Health education",
  ["health educ (lond)"] = "Health education (Bradford, West Yorkshire, England)",
  ["health educ (muncie)"] = "Health educator : journal of Eta Sigma Gamma",
  ["health educ behav"] = "Health education & behavior : the official publication of the Society for Public Health Education",
  ["health educ behav suppl"] = "Health Education and Behavior. Supplement",
  ["health educ bull"] = "Health education bulletin",
  ["health educ care"] = "Health education and care",
  ["health educ j"] = "Health education journal",
  ["health educ monogr"] = "Health education monographs",
  ["health educ public health"] = "Health education and public health",
  ["health educ q"] = "Health education quarterly",
  ["health educ q suppl"] = "Health education quarterly. Supplement",
  ["health educ rep"] = "Health education reports",
  ["health educ res"] = "Health education research",
  ["health educ welf indic"] = "Health, education, and welfare indicators",
  ["health educ work"] = "Health educators at work",
  ["health equity"] = "Health equity",
  ["health estate"] = "Health estate",
  ["health estate j"] = "Health estate journal : journal of the Institute of Hospital Engineering",
  ["health expect"] = "Health expectations : an international journal of public participation in health care and health policy",
  ["health expectations"] = "Health Expectations",
  ["health facil manage"] = "Health facilities management",
  ["health forum j"] = "Health Forum journal",
  ["health history"] = "Health and history",
  ["health horiz"] = "Health horizon",
  ["health hum rights"] = "Health and human rights",
  ["health ind today"] = "Health industry today",
  ["health inf j"] = "Health Informatics Journal",
  ["health inf manag"] = "Health information management : journal of the Health Information Management Association of Australia",
  ["health inf manage j"] = "Health Information Management Journal",
  ["health inf sci"] = "Health Information Science",
  ["health inf sci syst"] = "Health information science and systems",
  ["health inf syst telemed"] = "Health information systems and telemedicine",
  ["health info libr j"] = "Health information and libraries journal",
  ["health informatics j"] = "Health informatics journal",
  ["health insur stat cms"] = "Health insurance statistics. CMS; current Medicare survey report",
  ["health justice"] = "Health & justice",
  ["health lab sci"] = "Health laboratory science",
  ["health law can"] = "Health law in Canada",
  ["health law j"] = "Health law journal",
  ["health law news"] = "Health law news. University of Houston. Health Law and Policy Institute",
  ["health law proj libr bull"] = "Health Law Project library bulletin",
  ["health law rev"] = "Health law review",
  ["health law vigil"] = "Health law vigil",
  ["health libr rev"] = "Health libraries review",
  ["health lit res pract"] = "Health literacy research and practice",
  ["health manag technol"] = "Health management technology",
  ["health manage"] = "Health management",
  ["health manage forum"] = "Health management forum",
  ["health manage q"] = "Health management quarterly : HMQ",
  ["health manpow lit"] = "Health manpower literature",
  ["health manpow manage"] = "Health manpower management",
  ["health manpow rep"] = "Health manpower report",
  ["health mark q"] = "Health marketing quarterly",
  ["health marketing quart"] = "Health Marketing Quarterly",
  ["health matrix"] = "Health matrix",
  ["health matrix clevel"] = "Health matrix (Cleveland, Ohio : 1991)",
  ["health med"] = "Health & medicine : journal of the Health and Medicine Policy Research Group",
  ["health med care serv rev"] = "Health & medical care services review",
  ["health millions"] = "Health for the millions",
  ["health n hav"] = "Health",
  ["health news"] = "Health news (Waltham, Mass.)",
  ["health news (rochester)"] = "Health news",
  ["health news views"] = "Health news and views (Gaborone, Botswana)",
  ["health organ bull"] = "Bulletin. Health Organisation",
  ["health outcomes res med"] = "Health outcomes research in medicine",
  ["health pac bull"] = "Health PAC bulletin",
  ["health people"] = "The Health of the people",
  ["health perspect"] = "Health perspectives",
  ["health phys"] = "Health Physics",
  ["health place"] = "Health & place",
  ["health plann manpow rep"] = "Health planning & manpower report",
  ["health plann manpow rep (1976)"] = "Health planning & manpower reports",
  ["health policy"] = "Health policy (Amsterdam, Netherlands)",
  ["health policy (new york)"] = "Health policy",
  ["health policy educ"] = "Health policy and education",
  ["health policy open"] = "Health policy OPEN",
  ["health policy plan"] = "Health policy and planning",
  ["health policy q"] = "Health policy quarterly",
  ["health policy technol"] = "Health policy and technology",
  ["health policy week"] = "Health policy week : HPW",
  ["health popul perspect issues"] = "Health and population; perspectives and issues",
  ["health pract physician assist"] = "Health practitioner. Physician assistant",
  ["health press"] = "The health press",
  ["health prog"] = "Health progress (Saint Louis, Mo.)",
  ["health promot"] = "Health promotion (Oxford, England)",
  ["health promot chronic dis prev can"] = "Health promotion and chronic disease prevention in Canada : research, policy and practice",
  ["health promot int"] = "Health promotion international",
  ["health promot j austr"] = "Health promotion journal of Australia : official journal of Australian Association of Health Promotion Professionals",
  ["health promot perspect"] = "Health promotion perspectives",
  ["health promot pract"] = "Health promotion practice",
  ["health psychol"] = "Health psychology : official journal of the Division of Health Psychology, American Psychological Association",
  ["health psychol behav med"] = "Health psychology and behavioral medicine",
  ["health psychol bull"] = "Health psychology bulletin",
  ["health psychol open"] = "Health psychology open",
  ["health psychol rep"] = "Health psychology report",
  ["health psychol res"] = "Health psychology research",
  ["health psychol rev"] = "Health psychology review",
  ["health qual life outcomes"] = "Health and quality of life outcomes",
  ["health rays"] = "Health Rays",
  ["health reform prior serv"] = "Health reform and priority services",
  ["health rep"] = "Health reports",
  ["health res policy syst"] = "Health research policy and systems",
  ["health risk soc"] = "Health, risk & society",
  ["health sa"] = "Health SA = SA Gesondheid",
  ["health saf code annot state calif adopt april 7 1939 calif"] = "Health and Safety Code annotated of the State of California, adopted April 7, 1939. California",
  ["health sci rep"] = "Health science reports",
  ["health secur"] = "Health security",
  ["health serv insights"] = "Health services insights",
  ["health serv j"] = "The Health service journal",
  ["health serv manage"] = "Health services management",
  ["health serv manage res"] = "Health services management research",
  ["health serv manager"] = "Health services manager",
  ["health serv manpow rev"] = "Health services manpower review",
  ["health serv outcomes res methodol"] = "Health services & outcomes research methodology",
  ["health serv rep"] = "Health services reports",
  ["health serv res"] = "Health services research",
  ["health serv res manag epidemiol"] = "Health services research and managerial epidemiology",
  ["health services res"] = "Health Services Research",
  ["health sex"] = "Health and sexuality",
  ["health soc care community"] = "Health & social care in the community",
  ["health soc serv j"] = "Health and social service journal",
  ["health soc welf"] = "Health and social welfare",
  ["health soc work"] = "Health & social work",
  ["health sociol rev"] = "Health sociology review : the journal of the Health Section of the Australian Sociological Association",
  ["health stat nord ctries"] = "Health statistics in the Nordic countries = Helsestatistikk i de nordiske land",
  ["health stat q"] = "Health statistics quarterly",
  ["health syst (basingstoke)"] = "Health systems (Basingstoke, England)",
  ["health syst lead"] = "Health system leader",
  ["health syst manage"] = "Health systems management",
  ["health syst policy res"] = "Health systems and policy research",
  ["health syst reform"] = "Health systems and reform",
  ["health syst rev"] = "Health systems review",
  ["health syst transit"] = "Health systems in transition",
  ["health technol"] = "Health technology",
  ["health technol (berl)"] = "Health and technology",
  ["health technol assess"] = "Health technology assessment (Winchester, England)",
  ["health technol assess (rockv)"] = "Health technology assessment",
  ["health technol assess (winch eng)"] = "Health Technology Assessment",
  ["health technol assess rep"] = "Health technology assessment reports",
  ["health technol dir"] = "Health technology directions",
  ["health transit rev"] = "Health transition review : the cultural, social, and behavioural determinants of health",
  ["health trends"] = "Health trends",
  ["health values"] = "Health values",
  ["health visit"] = "Health visitor",
  ["healthc (amst)"] = "Healthcare (Amsterdam, Netherlands)",
  ["healthc ala"] = "Healthcare Alabama",
  ["healthc benchmarks"] = "Healthcare benchmarks",
  ["healthc bottom line"] = "Healthcare bottom line",
  ["healthc comput commun"] = "Healthcare computing & communications",
  ["healthc demand dis manag"] = "Healthcare demand & disease management",
  ["healthc exec"] = "Healthcare executive",
  ["healthc exec curr"] = "Healthcare executive currents",
  ["healthc facil manag ser"] = "Healthcare facilities management series",
  ["healthc financ manage"] = "Healthcare financial management : journal of the Healthcare Financial Management Association",
  ["healthc foodserv"] = "Healthcare foodservice",
  ["healthc foodserv mag"] = "Healthcare foodservice magazine : the international trade publication for the healthcare foodservice industry",
  ["healthc forum"] = "Healthcare forum",
  ["healthc forum j"] = "The Healthcare Forum journal",
  ["healthc hazard manage monit"] = "Healthcare hazard management monitor : HHMM : the newsletter of the Center for Healthcare Environmental Management",
  ["healthc hazard mater manage"] = "Healthcare hazardous materials management : HHMM",
  ["healthc hum resour"] = "Healthcare human resources",
  ["healthc hum resour spec rep"] = "Healthcare human resources. Special report",
  ["healthc inf manage"] = "Healthcare information management : journal of the Healthcare Information and Management Systems Society of the American Hospital Association",
  ["healthc infect"] = "Healthcare infection",
  ["healthc inform"] = "Healthcare informatics : the business magazine for information and communication systems",
  ["healthc inform res"] = "Healthcare informatics research",
  ["healthc leadersh manag rep"] = "Healthcare leadership & management report",
  ["healthc leadersh rep"] = "Healthcare leadership report",
  ["healthc low resour settings"] = "Healthcare in low-resource settings",
  ["healthc manage forum"] = "Healthcare management forum",
  ["healthc pap"] = "HealthcarePapers",
  ["healthc philanthr"] = "Healthcare philanthropy",
  ["healthc policy"] = "Healthcare policy = Politiques de santé",
  ["healthc prot manage"] = "Healthcare protection management",
  ["healthc q"] = "Healthcare quarterly (Toronto, Ont.)",
  ["healthc strateg"] = "The healthcare strategist",
  ["healthc syst strategy rep"] = "Healthcare systems strategy report",
  ["healthc technol lett"] = "Healthcare technology letters",
  ["healthc transform"] = "Healthcare transformation",
  ["healthc trends transit"] = "Healthcare trends & transition",
  ["healthcare (basel)"] = "Healthcare (Basel, Switzerland)",
  ["healthcare anal"] = "Healthcare Analytics",
  ["healthcare benchmarks qual improv"] = "Healthcare benchmarks and quality improvement",
  ["healthcare technol lett"] = "Healthcare Technology Letters",
  ["healthcare transform: artif intell autom rob"] = "Healthcare Transformation: Artificial Intelligence, Automation, and Robotics",
  ["healthy aging clin care elder"] = "Healthy aging & clinical care in the elderly",
  ["healthy aging res"] = "Healthy aging research",
  ["healthy people 2000 stat notes"] = "Healthy People 2000 statistical notes",
  ["healthy people 2000 stat surveill"] = "Healthy People 2000 statistics and surveillance",
  ["healthy people 2010 stat notes"] = "Healthy People 2010 statistical notes : from the Centers for Disease Control and Prevention/National Center for Health Statistics",
  ["healthy people stat notes"] = "Healthy people statistical notes",
  ["hear j"] = "The Hearing journal",
  ["hear news (washington)"] = "Hearing news",
  ["hear res"] = "Hearing research",
  ["hearing balance commun"] = "Hearing, balance and communication",
  ["heart advis"] = "Heart advisor",
  ["heart asia"] = "Heart Asia",
  ["heart bull"] = "The Heart bulletin",
  ["heart cent bull (roslyn)"] = "Bulletin. Roslyn, N. Y. St. Francis Hospital and Sanatorium for Cardiac Children",
  ["heart dis"] = "Heart disease (Hagerstown, Md.)",
  ["heart dis stroke"] = "Heart disease and stroke : a journal for primary care physicians",
  ["heart fail"] = "Heart failure",
  ["heart fail clin"] = "Heart failure clinics",
  ["heart fail monit"] = "Heart failure monitor",
  ["heart fail rev"] = "Heart failure reviews",
  ["heart int"] = "Heart international",
  ["heart lung"] = "Heart & lung : the journal of critical care",
  ["heart lung circ"] = "Heart, lung & circulation",
  ["heart lung vessel"] = "Heart, lung and vessels",
  ["heart metab"] = "Heart and metabolism : management of the coronary patient",
  ["heart mind (mumbai)"] = "Heart and mind (Mumbai, India)",
  ["heart rhythm"] = "Heart rhythm",
  ["heart rhythm o2"] = "Heart rhythm O2",
  ["heart surg forum"] = "The heart surgery forum",
  ["heart transplant"] = "Heart transplantation",
  ["heart vessels"] = "Heart and vessels",
  ["heart vessels suppl"] = "Heart and vessels. Supplement",
  ["heart views"] = "Heart views : the official journal of the Gulf Heart Association",
  ["heartbeat (los angeles)"] = "Heartbeat",
  ["heartrhythm case rep"] = "HeartRhythm case reports",
  ["heat exch des updates"] = "Heat Exchanger Design Updates",
  ["heat mass transf"] = "Heat and mass transfer = Wärme- und Stoffübertragung",
  ["heat mass transfer"] = "Heat and Mass Transfer",
  ["heat pipe sci technol int j"] = "Heat Pipe Science and Technology, An International Journal",
  ["heat piping air cond"] = "Heating, piping, and air conditioning",
  ["heat recovery syst chp"] = "Heat Recovery Systems and CHP",
  ["heat transf res"] = "Heat transfer research",
  ["heat transfer"] = "Heat Transfer",
  ["heat transfer asian res"] = "Heat Transfer - Asian Research",
  ["heat transfer eng"] = "Heat Transfer Engineering",
  ["heat transfer res"] = "Heat Transfer Research",
  ["heat transfer: soviet rev"] = "Heat Transfer: Soviet Reviews",
  ["heat treat met"] = "Heat Treatment of Metals",
  ["heat treat surf eng"] = "Heat Treatment and Surface Engineering",
  ["heb union coll annu"] = "Hebrew Union College annual. Hebrew Union College",
  ["hebammen wissen"] = "Hebammen Wissen",
  ["hebei nong ye da xue xue bao"] = "Hebei nong ye da xue xue bao = Hebei nongye daxue xuebao",
  ["hec forum"] = "HEC forum : an interdisciplinary journal on hospitals' ethical and legal issues",
  ["hecheng xiangjiao gongye"] = "Hecheng Xiangjiao Gongye",
  ["hedgehog rev"] = "The Hedgehog review",
  ["hefte unfallheilkd"] = "Hefte zur Unfallheilkunde",
  ["heidelb jahrb"] = "Heidelberger Jahrbücher",
  ["heilkd heilwege"] = "Heilkunde - Heilwege",
  ["heilongjiang bayi nongken daxue xuebao"] = "Heilongjiang bayi nongken daxue xuebao",
  ["heilongjiang daxue ziran kexue xuebao"] = "Heilongjiang Daxue Ziran Kexue Xuebao",
  ["heilongjiang xumu shouyi"] = "Heilongjiang xumu shouyi",
  ["heilpadagog werkbl"] = "Heilpädagogische Werkblätter",
  ["hektoen int"] = "Hektoen international : a journal of medical humanities",
  ["helgol mar res"] = "Helgoland marine research",
  ["hell adelphe"] = "Hellenis adelphe",
  ["hell cheirourgike"] = "Hellēnikē cheirourgikē. Acta chirurgica Hellenica",
  ["hell iatr"] = "Hellenike iatrike",
  ["hell j nucl med"] = "Hellenic journal of nuclear medicine",
  ["hell period stomat gnathopathoprosopike cheir"] = "To Hellēniko periodiko gia stomatikē & gnathoprosōpikē cheirourgikē",
  ["hell stomatol chron"] = "Hellenika stomatologika chronika. Hellenic stomatological annals",
  ["hellenic j cardiol"] = "Hellenic journal of cardiology : HJC = Hellēnikē kardiologikē epitheōrēsē",
  ["helr harvard environ law rev"] = "The Harvard environmental law review : HELR",
  ["helsinki univ tech digit systems lab ser a res rep"] = "Helsinki University of Technology",
  ["helsinki univ tech inst math res rep"] = "Helsinki University of Technology",
  ["helv chim acta"] = "Helvetica Chimica Acta",
  ["helv chir acta"] = "Helvetica chirurgica acta",
  ["helv chir acta suppl"] = "Helvetica chirurgica acta. Supplementum",
  ["helv med acta"] = "Helvetica medica acta",
  ["helv med acta suppl"] = "Helvetica medica acta. Supplementum",
  ["helv odontol acta"] = "Helvetica odontologica acta",
  ["helv paediatr acta"] = "Helvetica paediatrica acta",
  ["helv paediatr acta suppl"] = "Helvetica paediatrica acta. Supplementum",
  ["helv phys acta"] = "Helvetica Physica Acta",
  ["helv physiol pharmacol acta"] = "Helvetica physiologica et pharmacologica acta",
  ["helv physiol pharmacol acta suppl 1"] = "Helvetica physiologica et pharmacologica acta. Supplementum",
  ["helvetica phys acta"] = "Helvetica Physica Acta",
  ["hem ind"] = "Hemijska industrija",
  ["hematol cell ther"] = "Hematology and cell therapy",
  ["hematol educ"] = "Hematology education. European Hematology Association. Congress. Education Program",
  ["hematol j"] = "The hematology journal : the official journal of the European Haematology Association",
  ["hematol med oncol"] = "Hematology & medical oncology",
  ["hematol oncol"] = "Hematological oncology",
  ["hematol oncol clin north am"] = "Hematology/oncology clinics of North America",
  ["hematol oncol stem cell ther"] = "Hematology/oncology and stem cell therapy",
  ["hematol pathol"] = "Hematologic pathology",
  ["hematol rep"] = "Hematology reports",
  ["hematol rev"] = "Hematology reviews",
  ["hematol transfus cell ther"] = "Hematology, transfusion and cell therapy",
  ["hematology am soc hematol educ program"] = "Hematology. American Society of Hematology. Education Program",
  ["hematopathol mol hematol"] = "Hematopathology and molecular hematology",
  ["hemingway rev"] = "The Hemingway review",
  ["hemodial int"] = "Hemodialysis international. International Symposium on Home Hemodialysis",
  ["hemonc today"] = "HemOnc today",
  ["henan nongye daxue xuebao"] = "Henan nongye daxue xuebao",
  ["henan nongye kexue"] = "Henan nongye kexue",
  ["hengyang shi fan xue yuan xue bao"] = "Hengyang shi fan xue yuan xue bao = Journal of Heng Yang Teachers' College",
  ["henry e sigerist suppl bull hist med"] = "The Henry E. Sigerist supplements to the Bulletin of the history of medicine",
  ["henry ford hosp med bull"] = "Henry Ford Hospital medical bulletin",
  ["henry ford hosp med j"] = "Henry Ford Hospital medical journal",
  ["henry holt ref book"] = "A Henry Holt Reference Book",
  ["henry j kaiser fam found reprod health rep"] = "Henry J. Kaiser Family Foundation: Reproductive Health Report",
  ["hepat med"] = "Hepatic medicine : evidence and research",
  ["hepat mon"] = "Hepatitis monthly",
  ["hepat oncol"] = "Hepatic oncology",
  ["hepat res treat"] = "Hepatitis research and treatment",
  ["hepatitis wkly"] = "Hepatitis Weekly",
  ["hepatobiliary pancreat dis int"] = "Hepatobiliary & pancreatic diseases international : HBPD INT",
  ["hepatobiliary surg nutr"] = "Hepatobiliary surgery and nutrition",
  ["hepatol commun"] = "Hepatology communications",
  ["hepatol int"] = "Hepatology international",
  ["hepatol med policy"] = "Hepatology, medicine and policy",
  ["hepatol res"] = "Hepatology research : the official journal of the Japan Society of Hepatology",
  ["hepatoma res"] = "Hepatoma research",
  ["her geneal"] = "Heraldique et généalogie",
  ["her russ acad sci"] = "Herald of the Russian Academy of Sciences",
  ["herbal rev"] = "The Herbal review",
  ["herd: health environ res des j"] = "HERD: Health Environments Research & Design Journal",
  ["hered cancer clin pract"] = "Hereditary cancer in clinical practice",
  ["hereditary genet"] = "Hereditary genetics : current research",
  ["heredity (edinb)"] = "Heredity",
  ["herit eur math"] = "Heritage of European Mathematics",
  ["herit rev"] = "Heritage review (Bismarck, N.D.)",
  ["herit sci"] = "Heritage science",
  ["herit zimb"] = "Heritage of Zimbabwe",
  ["heritage sci"] = "Heritage Science",
  ["hermes (wiesb)"] = "Hermes",
  ["heroin addict relat clin probl"] = "Heroin addiction and related clinical problems",
  ["herpetol bull"] = "The herpetological bulletin",
  ["herpetol conser biol"] = "Herpetolgical Conservation and Biology",
  ["herpetol conserv biol"] = "Herpetological conservation and biology",
  ["herpetol j"] = "The Herpetological journal",
  ["herpetol monogr"] = "Herpetological Monographs",
  ["herpetol notes"] = "Herpetology notes",
  ["herpetol rev"] = "Herpetological review",
  ["herpetological monographs"] = "Herpetological monograph",
  ["herz kreislauf"] = "Herz-Kreislauf",
  ["herzschr elektrophysiol"] = "Herzschrittmachertherapie + Elektrophysiologie",
  ["herzschrittmacherther elektrophysiol"] = "Herzschrittmachertherapie & Elektrophysiologie",
  ["hespa mitt"] = "HESPA Mitteilungen",
  ["hesperian found news"] = "Hesperian Foundation news",
  ["hesperis tamuda"] = "Hespéris tamuda",
  ["hess arztebl"] = "Hessisches Ärzteblatt",
  ["hess familienkd"] = "Hessische Familienkunde",
  ["hess florist briefe"] = "Hessische Floristische Briefe",
  ["heteroat chem"] = "Heteroatom Chemistry",
  ["heteroc sumatr"] = "Heterocera sumatrana",
  ["heterocycl comm"] = "Heterocyclic communications",
  ["heterocycl commun"] = "Heterocyclic Communications",
  ["heterocycl lett"] = "Heterocyclic letters",
  ["heterog data manag polystores anal healthc (2018)"] = "Heterogeneous data management, polystores, and analytics for healthcare : VLDB 2018 workshops, Poly and DMAH, Rio de Janeiro, Brazil, August 31, 2018, revised selected papers",
  ["heterogen chem rev"] = "Heterogeneous Chemistry Reviews",
  ["heuresis v ling log sci"] = "Heuresis V",
  ["hexagon roche [engl]"] = "Hexagon Roche",
  ["hexagon roche engl"] = "Hexagon Roche",
  ["hfsp j"] = "HFSP journal",
  ["hgg adv"] = "HGG advances",
  ["hifu to hinyo dermatol urol"] = "Hifu to hinyo. The Dermatology and urology",
  ["hifuka kiyo"] = "Hifuka kiyo. Acta dermatologica",
  ["hifuka no rinsho"] = "Hifuka no rinsho. Rinsho derma (Tokyo)",
  ["hig cas hig mikrobiol epidemiol sanit teh"] = "Higijena; c̆asopis za higijenu, mikrobiologiju, epidemiologiju i sanitarnu tehniku",
  ["hig salubr"] = "Higiene y salubridad",
  ["hig sanit"] = "Higiene sanitaria",
  ["hig zdraveopaz"] = "Khigiena i zdraveopazvane",
  ["higashi nihon shigaku zasshi"] = "Higashi Nihon shigaku zasshi",
  ["higashi nippon shigaku zasshi"] = "Higashi Nippon Shigaku Zasshi",
  ["high alt med biol"] = "High altitude medicine & biology",
  ["high blood press cardiovasc prev"] = "High blood pressure & cardiovascular prevention : the official journal of the Italian Society of Hypertension",
  ["high dimens probab"] = "High dimensional probability",
  ["high educ (dordr)"] = "Higher education",
  ["high educ pedagog"] = "Higher education pedagogies",
  ["high energy chem"] = "High Energy Chemistry",
  ["high energy density phys"] = "High Energy Density Physics",
  ["high energy phys nuclear phys"] = "High Energy Physics and Nuclear Physics",
  ["high entropy alloys mater"] = "High Entropy Alloys & Materials",
  ["high freq"] = "High Frequency",
  ["high perform comput (2016)"] = "High performance computing : 31st International Conference, ISC High Performance 2016, Frankfurt, Germany, June 19-23, 2016, Proceedings. ISC High Performance (Conference) (31st : 2016 : Frankfurt, Germany)",
  ["high perform polym"] = "High Performance Polymers",
  ["high power laser sci eng"] = "High Power Laser Science and Engineering",
  ["high press res"] = "High pressure research",
  ["high pressure res"] = "High Pressure Research",
  ["high sch j"] = "High school journal (Chapel Hill, N.C.)",
  ["high struct"] = "Higher Structures",
  ["high technol (boston)"] = "High technology (Boston, Mass.)",
  ["high technol law j"] = "High technology law journal",
  ["high temp"] = "High Temperature",
  ["high temp (ussr)"] = "High Temperature (USSR) [translation of Teplofizika Vysokikh Temperatur]",
  ["high temp - high pressures"] = "High Temperatures - High Pressures",
  ["high temp corros mater"] = "High Temperature Corrosion of Materials",
  ["high temp high press"] = "High temperatures - high pressures",
  ["high temp mater processes (berlin, ger)"] = "High Temperature Materials and Processes (Berlin, Germany)",
  ["high temp mater processes (london)"] = "High Temperature Materials and Processes",
  ["high temp mater processes (london, u k)"] = "High Temperature Materials and Processes (London, United Kingdom)",
  ["high temp mater processes (new york)"] = "High Temperature Material Processes",
  ["high temp mater processes (tel aviv, israel)"] = "High Temperature Materials and Processes (Tel Aviv, Israel)",
  ["high temp mater processes: int q high-technol plasma processes"] = "High Temperature Materials Processes: An International Quarterly of High-Technology Plasma Processes",
  ["high temp-high pressures"] = "High Temperatures-High Pressures",
  ["high throughput"] = "High-throughput",
  ["high voltage"] = "High Voltage",
  ["high-confid comput"] = "High-Confidence Computing",
  ["higher-order symb comput"] = "Higher-Order and Symbolic Computation",
  ["highl ophthalmol"] = "Highlights of ophthalmology",
  ["highlights astron"] = "Highlights of Astronomy",
  ["higiene y seguridad (mexico 1936)"] = "Higiene y seguridad",
  ["hildesheimer inform-ber"] = "Hildesheimer Informatik-Bericht",
  ["hillside j clin psychiatry"] = "The Hillside journal of clinical psychiatry",
  ["him prir soedin"] = "Khimiia prirodnykh soedineniĭ",
  ["himalayan chem pharm bull"] = "Himalayan Chemical and Pharmaceutical Bulletin",
  ["himalayan geol"] = "Himalayan Geology",
  ["himalayan j environ zool"] = "Himalayan Journal of Environment and Zoology",
  ["hims 2016 (2016)"] = "HIMS 2016 : proceedings of the 2016 International Conference on Health Informatics and Medical Systems. HIMS (Conference) (2016 : Las Vegas, Nev.)",
  ["hindu astronom math text ser"] = "Hindu Astronomical and Mathematical Text Series",
  ["hindustan antibiot bull"] = "Hindustan antibiotics bulletin",
  ["hinyokika kiyo"] = "Hinyokika kiyo. Acta urologica Japonica",
  ["hip int"] = "Hip international : the journal of clinical and experimental research on hip pathology and therapy",
  ["hip pelvis"] = "Hip & pelvis",
  ["hipertens riesgo vasc"] = "Hipertensión y riesgo vascular",
  ["hippocrates (sausalito)"] = "Hippocrates (Sausalito, Calif.)",
  ["hippokrates (helsinki)"] = "Hippokrates (Helsinki, Finland)",
  ["hirosaki daigaku nogakubu gakujutsu hokoku"] = "Hirosaki Daigaku Nōgakubu Gakujutsu hōkoku = Bulletin of the Faculty of Agriculture, Hirosaki University",
  ["hirosaki igaku"] = "Hirosaki igaku = Hirosaki medical journal",
  ["hiroshima daigaku shigaku zasshi"] = "Hiroshima Daigaku shigaku zasshi. The Journal of Hiroshima University Dental Society",
  ["hiroshima j math ed"] = "Hiroshima Journal of Mathematics Education",
  ["hiroshima j med sci"] = "Hiroshima journal of medical sciences",
  ["hiroshima math j"] = "Hiroshima Mathematical Journal",
  ["hisp am hist rev"] = "The Hispanic American historical review",
  ["hisp health care int"] = "Hispanic health care international : the official journal of the National Association of Hispanic Nurses",
  ["hisp j behav sci"] = "Hispanic journal of behavioral sciences",
  ["hisp med"] = "Hispalis médica; revista sevillana de medicina y cirugia",
  ["hisp rev"] = "Hispanic review",
  ["hisp sacra"] = "Hispania sacra",
  ["hist (hist assoc s afr)"] = "Historia (Historical Association of South Africa)",
  ["hist abstr"] = "Historical Abstracts",
  ["hist abstr [a]"] = "Historical abstracts. Part A, Modern History abstracts 1450-1914",
  ["hist accid trav"] = "Histoire des accidents du travail",
  ["hist afr"] = "History in Africa",
  ["hist agrar"] = "Historia agraria",
  ["hist agric"] = "History of agriculture",
  ["hist aikak"] = "Historiallinen aikakauskirja",
  ["hist anaesth soc proc"] = "The History of Anaesthesia Society proceedings",
  ["hist anal philos"] = "History of Analytic Philosophy",
  ["hist anthropol"] = "History of anthropology",
  ["hist anthropol chur"] = "History and anthropology",
  ["hist archaeol"] = "Historical archaeology",
  ["hist ark"] = "Historiallinen arkisto. Suomen Historiallinen Seura",
  ["hist biol"] = "Historical biology",
  ["hist bull (calgary)"] = "Historical bulletin; notes and abstracts dealing with medical history",
  ["hist casopis"] = "Historický časopis",
  ["hist chem"] = "History of Chemistry",
  ["hist child q"] = "History of childhood quarterly",
  ["hist cienc saude manguinhos"] = "História, ciências, saúde--Manguinhos",
  ["hist compass"] = "History compass",
  ["hist comput"] = "History & computing",
  ["hist comun soc"] = "Historia y comunicación social",
  ["hist contemp ser univers pais vasco"] = "Historia contemporánea (Series) (Universidad del País Vasco)",
  ["hist crit"] = "Historia crítica (Bogotá, Colombia)",
  ["hist cult"] = "Historia y cultura",
  ["hist cult astron"] = "Historical & Cultural Astronomy",
  ["hist demogr"] = "Historická demografie",
  ["hist demogr mitt"] = "Historisch-demographische Mitteilungen. Communicatión de démographie historique",
  ["hist econ ideas"] = "History of Economic Ideas",
  ["hist econ rev"] = "History of Economics Review",
  ["hist econ soc"] = "Histoire, économie et société",
  ["hist educ"] = "History of education",
  ["hist educ (paris)"] = "Histoire de l'éducation",
  ["hist educ q"] = "History of education quarterly",
  ["hist eur ideas"] = "History of European ideas",
  ["hist fam"] = "The history of the family : an international quarterly",
  ["hist geo space sci"] = "History of Geo- and Space Science",
  ["hist geogr"] = "Historical geography",
  ["hist geom"] = "Histoires de Geometries",
  ["hist graf"] = "Historia y grafía",
  ["hist hist ustav akad ved ces repub"] = "Historica (Historický ústav (Akademie věd Česke republiky))",
  ["hist hosp"] = "Historia hospitalium",
  ["hist human sci"] = "History of the human sciences",
  ["hist ind"] = "Historia industrial",
  ["hist inst sci"] = "Histoires des Institutions Scientifiques",
  ["hist int public health"] = "History of international public health",
  ["hist irel"] = "History Ireland",
  ["hist j"] = "Historical journal (Cambridge, England)",
  ["hist j auckl waikato"] = "Historical journal: Auckland-Waikato",
  ["hist j film radio telev"] = "Historical journal of film, radio, and television",
  ["hist j mass"] = "Historical journal of Massachusetts",
  ["hist j west mass"] = "Historical journal of western Massachusetts",
  ["hist jahrb"] = "Historisches Jahrbuch",
  ["hist jahrb stadt linz"] = "Historisches Jahrbuch der Stadt Linz",
  ["hist life course stud"] = "Historical life course studies",
  ["hist lima"] = "Histórica",
  ["hist litteraturhist stud"] = "Historiska och litteraturhistoriska studier",
  ["hist mag"] = "Histoire magazine",
  ["hist mag protestant episcop church"] = "Historical magazine of the Protestant Episcopal Church",
  ["hist math"] = "History of Mathematics",
  ["hist math educ"] = "History of Mathematics Education",
  ["hist math subser"] = "History of Mathematics Subseries",
  ["hist mech mach sci"] = "History of Mechanism and Machine Science",
  ["hist med"] = "History of medicine",
  ["hist med sante"] = "Histoire, médecine et santé",
  ["hist med univ south ala coll med"] = "History of medicine (Mobile, Ala.)",
  ["hist med vet"] = "Historia medicinae veterinariae",
  ["hist mem"] = "History & memory",
  ["hist mes"] = "Histoire & mesure",
  ["hist methods"] = "Historical methods",
  ["hist methods newsl"] = "Historical methods newsletter",
  ["hist mex"] = "Historia mexicana",
  ["hist n h"] = "Historical New Hampshire",
  ["hist nat"] = "Histoire et nature",
  ["hist nat (mendoza)"] = "Historia natural (Mendoza)",
  ["hist nat bulg"] = "Historia naturalis Bulgarica",
  ["hist nauk biol med"] = "Historia nauk biologicznych i medycznych",
  ["hist neurosci"] = "History of neuroscience",
  ["hist news"] = "Historical news",
  ["hist now (christch)"] = "History now (Christchurch, N.Z.)",
  ["hist nurs bull"] = "History of nursing bulletin",
  ["hist nurs soc j"] = "History of Nursing Society journal",
  ["hist ophthalmol int"] = "Historia ophthalmologica internationalis",
  ["hist pap can hist assoc"] = "Historical papers. Canadian Historical Association. Meeting",
  ["hist philos life sci"] = "History and philosophy of the life sciences",
  ["hist philos logic"] = "History and Philosophy of Logic",
  ["hist philos psychol"] = "History & philosophy of psychology",
  ["hist philos sci"] = "History and Philosophy of Science—Heresy, Crossroads, and Intersections",
  ["hist philos sci (paris)"] = "Histoire et Philosophie des Sciences (Paris)",
  ["hist photogr"] = "History of photography",
  ["hist phys"] = "History of Physics",
  ["hist polit"] = "Historia y política (Madrid, Spain : 1999)",
  ["hist polit econ"] = "History of political economy",
  ["hist polit economy"] = "History of Political Economy",
  ["hist polit thought"] = "History of political thought",
  ["hist present (champaign)"] = "History of the present (Champaign, Ill.)",
  ["hist preserv"] = "Historic preservation (Washington, D.C.)",
  ["hist psychiatry"] = "History of psychiatry",
  ["hist psychoanal"] = "History of psychoanalysis",
  ["hist psychol"] = "History of psychology",
  ["hist questoes debates"] = "História, questões & debates : revista da Associação Paranaense de História",
  ["hist rec aust sci"] = "Historical records of Australian science",
  ["hist reflect"] = "Historical reflections. Réflexions historiques",
  ["hist relig"] = "History of religions",
  ["hist res"] = "Historical research : the bulletin of the Institute of Historical Research",
  ["hist sci"] = "History of science",
  ["hist sci (paris)"] = "Histoire des Sciences (Paris)",
  ["hist sci (tokyo)"] = "Historia scientiarum : international journal of the History of Science Society of Japan",
  ["hist sci med"] = "Histoire des sciences médicales",
  ["hist sci tech china"] = "History of Science and Technology in China",
  ["hist sci tech reprint ser"] = "History of Science and Technology Reprint Series",
  ["hist scotl"] = "History Scotland",
  ["hist soc"] = "Historia y sociedad (Río Piedras, San Juan, P.R.)",
  ["hist soc rurales"] = "Histoire & sociétés rurales",
  ["hist soz forsch"] = "Historische Sozialforschung = Historical social research",
  ["hist steps"] = "Historic Steps",
  ["hist stud"] = "Historical studies (Melbourne, Vic.)",
  ["hist stud nat sci"] = "Historical studies in the natural sciences",
  ["hist stud phys biol sci"] = "Historical studies in the physical and biological sciences : HSPS",
  ["hist teacher"] = "The History teacher",
  ["hist technol"] = "History and technology",
  ["hist theory"] = "History and theory",
  ["hist tidskr"] = "Historisk tidskrift (Stockholm, Sweden)",
  ["hist tidskr finl"] = "Historisk tidskrift för Finland",
  ["hist tidsskr"] = "Historisk tidsskrift : udgivet af Den norske historiske forening",
  ["hist today"] = "History today",
  ["hist univ"] = "History of universities",
  ["hist urbaine"] = "Histoire urbaine",
  ["hist vida"] = "Historia y vida",
  ["hist workshop"] = "History workshop",
  ["hist workshop j"] = "History workshop journal : HWJ",
  ["hist z"] = "Historische Zeitschrift",
  ["hist-anal stud nat mind action"] = "Historical-Analytical Studies on Nature, Mind and Action",
  ["histochem cell biol"] = "Histochemistry and cell biology",
  ["histochem j"] = "The Histochemical journal",
  ["histoire soc"] = "Histoire sociale. Social history",
  ["histol histopathol"] = "Histology and histopathology",
  ["historia (argentina)"] = "Historia",
  ["historia antropol fuentes orales"] = "Historia antropología y fuentes orales",
  ["historia math"] = "Historia Mathematica",
  ["historia santiago"] = "Historia",
  ["historia sci (2)"] = "Historia Scientiarum",
  ["historia unisinos"] = "História unisinos : revista do Programa de Pós-Graduação em Historía da Universidade do Vale do Rio dos Sinos",
  ["historian (lond)"] = "Historian (London, England)",
  ["historias (mex)"] = "Historias : revista de la Dirección de Estudios Históricos del Instituto Nacional de Antiopología e Historia",
  ["historie (arhus)"] = "Historie (Århus, Denmark)",
  ["historiel foren arsskr"] = "Historielärarnas förenings årsskrift",
  ["historiens geogr"] = "Historiens et géographes",
  ["historiogr linguist"] = "Historiographia linguistica",
  ["history (lond)"] = "History (Historical Association (Great Britain))",
  ["hitotsubashi j arts sci"] = "Hitotsubashi Journal of Arts & Sciences",
  ["hitotsubashi j com manage"] = "Hitotsubashi Journal of Commerce and Management",
  ["hitotsubashi j econ"] = "Hitotsubashi Journal of Economics",
  ["hitotsubashi j law polit"] = "Hitotsubashi journal of law & politics",
  ["hiv adv res dev"] = "HIV advance research and development : open access",
  ["hiv aids (auckl)"] = "HIV/AIDS (Auckland, N.Z.)",
  ["hiv aids policy law rev"] = "HIV/AIDS policy & law review",
  ["hiv aids surveill rep"] = "HIV/AIDS surveillance report (Atlanta, Ga.)",
  ["hiv capsule rep"] = "HIV capsule report",
  ["hiv clin"] = "HIV clinician",
  ["hiv clin trials"] = "HIV clinical trials",
  ["hiv hotline"] = "HIV hotline",
  ["hiv inside"] = "HIV inside : a newsletter for correctional professionals",
  ["hiv med"] = "HIV medicine",
  ["hiv prev plus"] = "HIV prevention plus!",
  ["hiv res clin pract"] = "HIV research & clinical practice",
  ["hiv seq compend"] = "HIV sequence compendium",
  ["hiv spec"] = "HIV specialist",
  ["hiv ther"] = "HIV therapy",
  ["hiv/aids res treat"] = "HIV/AIDS research and treatment : open journal",
  ["hiyoshi rev natur sci"] = "Keio University",
  ["hkie trans"] = "HKIE Transactions",
  ["hladnikia (ljubl)"] = "Hladnikia",
  ["hmo pract"] = "HMO practice",
  ["hmong stud j"] = "Hmong studies journal",
  ["hno nachr"] = "HNO-Nachrichten",
  ["ho kyushu byogaichu kenkyukai"] = "Ho. Kyushu Byogaichu Kenkyukai",
  ["hoaj biol"] = "HOAJ biology",
  ["hochschuldidakt material"] = "Hochschuldidaktische Materialien",
  ["hofstra law rev"] = "Hofstra law review",
  ["hohenh volkswirtsch schr"] = "Hohenheimer Volkswirtschaftliche Schriften",
  ["hoja tisiol"] = "Hoja tisiológica",
  ["hokenfu zasshi"] = "[Hokenfu zasshi] The Japanese journal for public health nurse",
  ["hokkaido daigaku nogakubu kiyo|j fac agric hokkaido univ"] = "Journal of the Faculty of Agriculture, Hokkaido University",
  ["hokkaido igaku zasshi"] = "[Hokkaido igaku zasshi] The Hokkaido journal of medical science",
  ["hokkaido math j"] = "Hokkaido Mathematical Journal",
  ["hokkaido shika ishikai shi"] = "Hokkaido Shika Ishikai shi",
  ["holdsworth law rev"] = "Holdsworth law review",
  ["holist nurs pract"] = "Holistic nursing practice",
  ["holistic assertive nurse"] = "Holistic Assertive Nurse",
  ["holocaust genocide stud"] = "Holocaust and genocide studies",
  ["holz roh werkst"] = "Holz als Roh und Werkstoff",
  ["holz roh- werkst"] = "Holz als Roh- und Werkstoff",
  ["holz, schweiz holzztg"] = "Holz, Schweizerische Holzzeitung",
  ["holzforsch schweiz"] = "Holzforschung Schweiz",
  ["holzpreisstat – rohholz"] = "Holzpreisstatistik – Rohholz",
  ["home care econ"] = "Home care economics",
  ["home care manag"] = "Home care manager",
  ["home care provid"] = "Home care provider",
  ["home health care manag pract"] = "Home health care management & practice",
  ["home health care serv q"] = "Home health care services quarterly",
  ["home health j"] = "Home health journal",
  ["home health rev"] = "Home health review",
  ["home healthc now"] = "Home healthcare now",
  ["home healthc nurse"] = "Home healthcare nurse",
  ["home healthc nurse manag"] = "Home healthcare nurse manager",
  ["home hemodial int"] = "Home hemodialysis international. International Symposium on Home Hemodialysis",
  ["home hemodial int (1997)"] = "Home hemodialysis international. International Symposium on Daily Home Hemodialysis",
  ["homemak mag"] = "Homemaker's magazine",
  ["homeopath bull"] = "Homoeopathic bulletin",
  ["homeopath fr"] = "L' Homéopathie française",
  ["homeopath her"] = "Homoeopathic herald",
  ["homeost health dis"] = "Homeostasis in health and disease : international journal devoted to integrative brain functions and homeostatic systems",
  ["homicide stud"] = "Homicide studies",
  ["homilet pastor rev"] = "The Homiletic and pastoral review",
  ["homme (vienna aust)"] = "Homme (Vienna, Austria)",
  ["hommes migr"] = "Hommes & migrations",
  ["hommes migr doc"] = "Hommes et migrations. Documents",
  ["hommes terres nord"] = "Hommes et terres du Nord",
  ["homoeopath physician"] = "The Homoeopathic physician",
  ["homoeopath rec"] = "The Homoeopathic recorder",
  ["homol homotopy appl"] = "Homology Homotopy and Applications\t.",
  ["homology homotopy appl"] = "Homology, Homotopy and Applications",
  ["honam math j"] = "Honam Mathematical Journal",
  ["honan i ko ta hsueh hsueh pao"] = "Henan yi ke da xue xue bao = Journal of Henan Medical University = Henan yikedaxue xuebao",
  ["hong kong econ pap"] = "Hong Kong Economic Papers",
  ["hong kong j occup ther"] = "Hong Kong journal of occupational therapy : HKJOT",
  ["hong kong j ophthalmol"] = "Hong Kong journal of ophthalmology : the official publication of the College of Ophthalmologists of Hong Kong = Xianggang yan ke xue kan : Xianggang yan ke yi xue yuan",
  ["hong kong j paediatr"] = "The Hong Kong journal of paediatrics : the journal of Hong Kong Paediatric Society = Hsiang-kang erh k'o i hsueh hui hui k'an",
  ["hong kong j soc work"] = "The Hong Kong journal of social work = Xiang Gang she hui gong zuo xue bao",
  ["hong kong law j"] = "Hong Kong law journal",
  ["hong kong med j"] = "Hong Kong medical journal = Xianggang yi xue za zhi",
  ["hong kong mon dig stat"] = "Hong Kong monthly digest of statistics",
  ["hong kong pharm j"] = "Hong Kong pharmaceutical journal",
  ["hong kong physiother j"] = "Hong Kong physiotherapy journal : official publication of the Hong Kong Physiotherapy Association Limited = Wu li chih liao",
  ["hong qi"] = "Hong qi",
  ["honyurui kagaku"] = "Honyūrui kagaku",
  ["hoosier health her"] = "Hoosier health herald",
  ["hopkins hiv rep"] = "The Hopkins HIV report : a bimonthly newsletter for healthcare providers",
  ["hoppe seylers z physiol chem"] = "Hoppe-Seyler's Zeitschrift für physiologische Chemie",
  ["hoppe-seyler’s z physiol chem"] = "Hoppe-Seyler’s Zeitschrift fur Physiologische Chemie",
  ["hora med"] = "Hora médica",
  ["horiz biochem biophys"] = "Horizons in biochemistry and biophysics",
  ["horiz cancer res"] = "Horizons in cancer research",
  ["horiz enferm"] = "Horizonte de enfermería",
  ["horiz med"] = "Horizons médicaux",
  ["horiz med (barcelona)"] = "Horizonte médico",
  ["horizon (ny)"] = "Horizon",
  ["horm behav"] = "Hormones and behavior",
  ["horm cancer"] = "Hormones & cancer",
  ["horm metab res"] = "Hormone and metabolic research = Hormon- und Stoffwechselforschung = Hormones et métabolisme",
  ["horm metab res suppl"] = "Hormone and metabolic research. Supplement series",
  ["horm mol biol clin invest"] = "Hormone Molecular Biology and Clinical Investigation",
  ["horm mol biol clin investig"] = "Hormone molecular biology and clinical investigation",
  ["horm res"] = "Hormone research",
  ["horm res paediatr"] = "Hormone research in pædiatrics",
  ["hormone metab res"] = "Hormone and Metabolic Research",
  ["hormone mol biol clin invest"] = "Hormone Molecular Biology and Clinical Investigation",
  ["hormones (athens)"] = "Hormones (Athens, Greece)",
  ["hort j"] = "The horticulture journal",
  ["hortic bras"] = "Horticultura brasileira",
  ["hortic environ biotechnol"] = "Horticulture, Environment and Biotechnology",
  ["hortic plant j"] = "Horticultural plant journal",
  ["hortic res"] = "Horticulture research",
  ["hortic rev (am soc hortic sci)"] = "Horticultural reviews",
  ["horumon to rinsho"] = "Horumon to rinshō. Clinical endocrinology",
  ["horwood ser math appl"] = "Horwood Series in Mathematics & Applications",
  ["hosp (lond)"] = "The Hospital",
  ["hosp adm (chic)"] = "Hospital Administration",
  ["hosp adm (new delhi)"] = "Hospital administration",
  ["hosp adm can"] = "Hospital administration in Canada",
  ["hosp admin curr"] = "Hospital administration currents",
  ["hosp admitting mon"] = "Hospital admitting monthly",
  ["hosp aviat"] = "Hospital aviation",
  ["hosp bond rev"] = "Hospital bond review",
  ["hosp cap finance"] = "Hospital capital finance American Hospital Association, Division of Hospital Planning and Capital Finance",
  ["hosp case manag"] = "Hospital case management : the monthly update on hospital-based care planning and critical paths",
  ["hosp chron"] = "Hospital chronicles = Nosokomeiaka chronika",
  ["hosp community psychiatry"] = "Hospital & community psychiatry",
  ["hosp corps q"] = "Hospital Corps quarterly",
  ["hosp cost manag account"] = "Hospital cost management and accounting",
  ["hosp counc bull"] = "Hospital Council bulletin",
  ["hosp dev"] = "Hospital development",
  ["hosp employee health"] = "Hospital employee health",
  ["hosp eng"] = "Hospital engineering",
  ["hosp entrep newsl"] = "Hospital entrepreneurs' newsletter",
  ["hosp equip supplies"] = "Hospital equipment & supplies",
  ["hosp ethics"] = "Hospital ethics",
  ["hosp financ manage"] = "Hospital financial management",
  ["hosp food nutr focus"] = "Hospital food & nutrition focus",
  ["hosp formul"] = "Hospital formulary",
  ["hosp forum"] = "Hospital forum",
  ["hosp gauzette"] = "Hospital gauzette",
  ["hosp gen (madr)"] = "Hospital general",
  ["hosp gift shop manage"] = "Hospital gift shop management",
  ["hosp guest relations rep"] = "Hospital guest relations report",
  ["hosp hazard mater manage"] = "Hospital hazardous materials management",
  ["hosp health care newsl"] = "Hospital Health Care Newsletter",
  ["hosp health manag"] = "Hospital and health management",
  ["hosp health netw"] = "Hospitals and Health Networks",
  ["hosp health netw 360"] = "Hospitals & health networks 360° : H&HN 360°",
  ["hosp health rev"] = "The Hospital and health review",
  ["hosp health serv adm"] = "Hospital & health services administration",
  ["hosp health serv rev"] = "The Hospital and health services review",
  ["hosp infect control"] = "Hospital infection control",
  ["hosp j"] = "The Hospice journal",
  ["hosp law newsl"] = "Hospital law newsletter",
  ["hosp libr"] = "Hospital libraries",
  ["hosp mag"] = "The Hospital magazine",
  ["hosp manage"] = "Hospital management",
  ["hosp manage commun"] = "Hospital management communications",
  ["hosp manage q"] = "Hospital management quarterly : HMQ",
  ["hosp manager"] = "The Hospital manager",
  ["hosp mater manage"] = "Hospital material[dollar sign] management",
  ["hosp mater manage q"] = "Hospital materiel management quarterly",
  ["hosp med"] = "Hospital medicine (London, England : 1998)",
  ["hosp med clin"] = "Hospital medicine clinics",
  ["hosp med staff"] = "The Hospital medical staff",
  ["hosp natl assoc inpatient physicians"] = "Hospitalist",
  ["hosp outlook"] = "Hospital outlook",
  ["hosp patient relat rep"] = "Hospital patient relations report",
  ["hosp pediatr"] = "Hospital pediatrics",
  ["hosp peer rev"] = "Hospital peer review",
  ["hosp pharm"] = "Hospital pharmacy",
  ["hosp physician"] = "Hospital physician",
  ["hosp pract"] = "Hospital practice",
  ["hosp pract (1995)"] = "Hospital practice (1995)",
  ["hosp pract (hosp ed)"] = "Hospital practice (Hospital ed.)",
  ["hosp pract (minneap)"] = "Hospital practice (1995)",
  ["hosp pract (off ed)"] = "Hospital practice (Office ed.)",
  ["hosp pract res"] = "Hospital practices and research",
  ["hosp prog"] = "Hospital progress",
  ["hosp prog directory issue"] = "Hospital progress. Directory issue",
  ["hosp purch manage"] = "Hospital purchasing management",
  ["hosp q"] = "Hospital quarterly",
  ["hosp rec study"] = "Hospital record study : a joint study by CPHA and IMS America Ltd",
  ["hosp revenue rep"] = "Hospital revenue report",
  ["hosp risk manage"] = "Hospital risk management",
  ["hosp secur saf manage"] = "Hospital security and safety management",
  ["hosp strategy rep"] = "Hospital strategy report",
  ["hosp superv"] = "Hospital supervision",
  ["hosp superv bull"] = "Hospital supervisor's bulletin",
  ["hosp technol ser"] = "Hospital technology series",
  ["hosp top"] = "Hospital topics",
  ["hosp trib"] = "Hospital tribune",
  ["hosp trustee"] = "Hospital trustee",
  ["hospital (lond 1886)"] = "The Hospital",
  ["hospital (rio j)"] = "Hospital",
  ["hospitals (lond)"] = "Hospitals",
  ["hotel j"] = "Hotel Journal",
  ["hotel tour rev"] = "Hotel + Touristik Revue",
  ["hotetsu rinsho"] = "Hotetsu rinsho. Practice in prosthodontics",
  ["houille blanche"] = "Houille Blanche",
  ["hous policy debate"] = "Housing policy debate",
  ["hous soc"] = "Housing and society",
  ["hous stud"] = "Housing studies",
  ["household pers prod ind"] = "Household & Personal Products Industry",
  ["housing finance rev"] = "Housing Finance Review",
  ["housing pol debate"] = "Housing Policy Debate",
  ["housing stud"] = "Housing Studies",
  ["houst j health law policy"] = "Houston journal of health law & policy",
  ["houst j int law"] = "Houston journal of international law",
  ["houst law rev"] = "Houston law review",
  ["houst lawyer"] = "Houston lawyer : official monthly publication of the Houston Bar Association",
  ["houston j math"] = "Houston Journal of Mathematics",
  ["houston rev"] = "The Houston review: history and culture of the Gulf Coast",
  ["how eval health programs"] = "How to: Evaluate health programs",
  ["howard j commun"] = "The Howard journal of communications",
  ["howard law j"] = "Howard law journal",
  ["hpa resour"] = "HPA resource",
  ["hpb (oxford)"] = "HPB : the official journal of the International Hepato Pancreato Biliary Association",
  ["hpb surg"] = "HPB surgery : a world journal of hepatic, pancreatic and biliary surgery",
  ["hpn hosp purch news"] = "Hospital purchasing news : HPN",
  ["hrb open res"] = "HRB open research",
  ["hrc cc j high resolut chromatogr chromatogr commun"] = "HRC & CC. Journal of high resolution chromatography & chromatography communications",
  ["hrc j high resolut chromatogr"] = "HRC Journal of High Resolution Chromatography",
  ["hrsa careaction"] = "HRSA careaction",
  ["hsmha health rep"] = "HSMHA health reports",
  ["hsoa j addict addict disord"] = "HSOA journal of addiction & addictive disorders",
  ["hsoa j addict neuropharmacol"] = "HSOA journal of addiction & neuropharmacology",
  ["hsoa j aids clin res stds"] = "HSOA journal of AIDS clinical research & STDs",
  ["hsoa j biotech res biochem"] = "HSOA journal of biotech research & biochemistry",
  ["hsoa j cell biol cell metabol"] = "HSOA journal of cell biology & cell metabolism",
  ["hsoa j community med public health care"] = "HSOA journal of community medicine & public health care",
  ["hsoa j food sci nutr"] = "HSOA journal of food science & nutrition",
  ["hsoa j gastroenterol hepatol res"] = "HSOA journal of gastroenterology & hepatology research",
  ["hsoa j hum endocrinol"] = "HSOA journal of human endocrinology",
  ["hsoa j infect non infect dis"] = "HSOA journal of infectious & non infectious diseases",
  ["hsoa j nephrol ren ther"] = "HSOA journal of nephrology & renal therapy",
  ["hsoa j ophthalmol clin res"] = "HSOA journal of ophthalmology & clinical research",
  ["hsoa j psychiatry depress anxiety"] = "HSOA journal of psychiatry, depression & anxiety",
  ["hsoa j reprod med gynaecol obstet"] = "HSOA journal of reproductive medicine gynaecology & obstetrics",
  ["hsoa j stem cells res dev ther"] = "HSOA journal of stem cells research, development & therapy",
  ["hsoa j toxicol"] = "HSOA journal of toxicology : current research",
  ["hsoa trends anat physiol"] = "HSOA trends in anatomy and physiology",
  ["hsr proc intensive care cardiovasc anesth"] = "HSR proceedings in intensive care & cardiovascular anesthesia",
  ["hss j"] = "HSS journal : the musculoskeletal journal of Hospital for Special Surgery",
  ["hstc bull"] = "HSTC bulletin",
  ["hsueh tsa chih"] = "Chung-Hua Min Kuo Wei Sheng Wu Chi Mien I Hsueh Tsa Chih Chinese JournalOf Microbiology and Immunology Chung Hua Min Kuo Wei Sheng Wu Chi Mien",
  ["htd (am soc mech eng)"] = "HTD (American Society of Mechanical Engineers)",
  ["htm haerteri tech mitt"] = "HTM, Haerterei Technische Mitteilungen",
  ["htm j heat treat mater"] = "HTM Journal of Heat Treatment and Materials",
  ["htm, haerterei tech mit"] = "HTM, Haerterei-Technische Mitteilungen",
  ["htm, haerterei tech mitt"] = "HTM, Haerterei-Technische Mitteilungen",
  ["hu li yan jiu"] = "Hu li yan jiu = Nursing research",
  ["hu li za zhi"] = "Hu li za zhi The journal of nursing",
  ["hu po ke xue"] = "Hu po ke xue = Journal of lake sciences",
  ["hua bei nong xue bao"] = "Hua bei nong xue bao = Acta agriculturae boreali-sinica",
  ["hua dong li gong da xue xue bao zi ran ke xue ban"] = "Hua dong li gong da xue xue bao. Zi ran ke xue ban = Journal of East China University of Science and Technology. Natural science edition",
  ["hua gong xue bao"] = "Hua gong xue bao = Chemical industry and engineering",
  ["hua hsi i ko ta hsueh hsueh pao"] = "Hua-Hsi I Ko Ta Hsueh Hsueh Pao [Journal of West China University Of Medical Sciences]",
  ["hua hsi kou chiang i hsueh tsa chih"] = "Hua Hsi Kou Chiang I Hsueh Tsa Chih West China Journal Of Stomatology (Cheng-Tu",
  ["hua nan nong ye da xue xue bao"] = "Hua nan nong ye da xue xue bao = Journal of South China Agricultural University",
  ["hua sheng xue bao"] = "Hua sheng xue bao = Journal of peanut science",
  ["hua xi kou qiang yi xue za zhi"] = "Hua xi kou qiang yi xue za zhi = Huaxi kouqiang yixue zazhi = West China journal of stomatology",
  ["hua xi yi ke da xue xue bao"] = "Hua xi yi ke da xue xue bao = Journal of West China University of Medical Sciences = Huaxi yike daxue xuebao",
  ["hua zhong nong ye da xue xue bao"] = "Hua zhong nong ye da xue xue bao = Journal Huazhong (Central China) Agricultural University",
  ["hua zhong shi fan da xue xue bao zi ran ke xue ban"] = "Hua zhong shi fan da xue xue bao. Zi ran ke xue ban = Journal of Central China Normal University. Natural sciences edition",
  ["huadong huagong xueyuan xuebao"] = "Huadong Huagong Xueyuan Xuebao",
  ["huadong ligong daxue xuebao"] = "Huadong Ligong Daxue Xuebao",
  ["huadong ligong daxue xuebao, ziran kexueban"] = "Huadong Ligong Daxue Xuebao, Ziran Kexueban",
  ["huadong shifan daxue xuebao"] = "Huadong shifan daxue xuebao = Journal of East China Normal University",
  ["huagong jinzhan"] = "Hua gong jin zhan = Chemical industry and engineering progress : CIEP",
  ["huagong xuebao (chin ed)"] = "Huagong Xuebao (Chinese Edition)",
  ["huaihai gongxueyuan xuebao"] = "Huaihai gongxueyuan xuebao : ziran kexue ban",
  ["huaihua shizhuan xuebao"] = "Huaihua Shizhuan Xuebao",
  ["huan jing ke xue"] = "Huan jing ke xue= Huanjing kexue",
  ["huan jing kun chong xue bao"] = "Huan jing kun chong xue bao",
  ["huan jing wu ran zhi li ji shu yu she bei"] = "Huan jing wu ran zhi li ji shu yu she bei = Techniques and equipment for environmental pollution control",
  ["huan jing yu jian kang za zhi"] = "Huan jing yu jian kang za zhi = Journal of environment and health",
  ["huanan shifan daxue xuebao"] = "Huanan Shifan Daxue xuebao. Ziran kexue ban. Huanan Shifan Daxue",
  ["huanjing huaxue"] = "Huan jing hua xue = Environmental chemistry",
  ["huanjing kexue"] = "Huanjing Kexue",
  ["huanjing kexue xuebao"] = "Huan jing ke xue xue bao = Acta scientiae circumstantiae",
  ["huaxue fanying gongcheng yu gongyi"] = "Huaxue Fanying Gongcheng Yu Gongyi",
  ["huaxue shiji"] = "Huaxue Shiji",
  ["huaxue shijie"] = "Huaxue Shijie",
  ["huaxue tongbao"] = "Huaxue Tongbao",
  ["huaxue wuli xuebao"] = "Huaxue Wuli Xuebao",
  ["huaxue xuebao"] = "Hua xue xue bao. Acta chimica Sinica",
  ["hubei nong ye ke xue"] = "Hubei nong ye ke xue",
  ["hudson cty health rec"] = "Hudson County health record [monthly]",
  ["hugo j"] = "The HUGO journal",
  ["hugoye (wash d c)"] = "Hugoye : journal of Syriac studies",
  ["huisarts wet"] = "Huisarts en wetenschap",
  ["hum affairs (bratisl)"] = "Human affairs (Bratislava, Slovakia)",
  ["hum anim interact bull"] = "Human-animal interaction bulletin",
  ["hum antibodies"] = "Human antibodies",
  ["hum antibodies hybridomas"] = "Human antibodies and hybridomas",
  ["hum asp it aged popul (2015)"] = "Human Aspects of IT for the Aged Population : Design for Aging : First International Conference, ITAP 2015, held as part of HCI International 2015, Los Angeles, CA, USA, August 2-7, 2015. Proceedings. Part I. ITAP (Conference) (1st : 2015 : Los Angeles, Calif.)",
  ["hum behav"] = "Human behavior; the newsmagazine of the social sciences",
  ["hum behav emerg technol"] = "Human behavior and emerging technologies",
  ["hum behav emerging technol"] = "Human Behavior and Emerging Technologies",
  ["hum biol"] = "Human biology",
  ["hum biol oceania"] = "Human biology in Oceania",
  ["hum brain artif intell (2019)"] = "Human brain and artificial intelligence : first International Workshop, HBAI 2019, held in conjunction with IJCAI 2019, Macao, China, August 12, 2019, revised selected papers. HBAI (Conference) (1st : 2019 : Macau, China)",
  ["hum brain mapp"] = "Human brain mapping",
  ["hum cell"] = "Human cell",
  ["hum commun res"] = "Human communication research",
  ["hum comput (fairfax)"] = "Human computation (Fairfax, Va.)",
  ["hum comput interact"] = "Human-computer interaction",
  ["hum dev"] = "Human development",
  ["hum dimens wildl"] = "Human Dimensions of Wildlife",
  ["hum ecol"] = "Human ecology",
  ["hum ecol interdiscip j"] = "Human ecology: an interdisciplinary journal",
  ["hum ecol risk assess"] = "Human & Ecological Risk Assessment",
  ["hum ecol risk assess: int j"] = "Human and Ecological Risk Assessment: An International Journal",
  ["hum ethol bull"] = "Human ethology bulletin",
  ["hum events"] = "Human events",
  ["hum evol"] = "Human evolution",
  ["hum exp toxicol"] = "Human & experimental toxicology",
  ["hum factors"] = "Human factors",
  ["hum factors ergon manuf"] = "Human factors and ergonomics in manufacturing",
  ["hum factors ergon manuf serv ind"] = "Human Factors and Ergonomics in Manufacturing & Service Industries",
  ["hum factors healthcare"] = "Human Factors in Healthcare",
  ["hum factors mech eng def saf"] = "Human Factors and Mechanical Engineering for Defense and Safety",
  ["hum fertil"] = "Human fertility",
  ["hum fertil (camb)"] = "Human fertility (Cambridge, England)",
  ["hum gene ther"] = "Human Gene Therapy",
  ["hum gene ther clin dev"] = "Human Gene Therapy Clinical Development",
  ["hum gene ther methods"] = "Human Gene Therapy Methods",
  ["hum gene ther: clin dev"] = "Human Gene Therapy: Clinical Development",
  ["hum gene ther: methods"] = "Human Gene Therapy: Methods",
  ["hum genet"] = "Human genetics",
  ["hum genet embryol"] = "Human genetics & embryology : current research",
  ["hum genet genomics adv"] = "Human Genetics and Genomics Advances",
  ["hum genet suppl"] = "Human genetics. Supplement",
  ["hum genome news"] = "Human genome news",
  ["hum genome var"] = "Human genome variation",
  ["hum genomics"] = "Human genomics",
  ["hum genomics proteomics"] = "Human genomics and proteomics : HGP",
  ["hum health care"] = "Humane health care",
  ["hum health care int"] = "Humane health care international",
  ["hum hered"] = "Human heredity",
  ["hum immunol"] = "Human immunology",
  ["hum life rev"] = "The Human life review",
  ["hum microb j"] = "Human microbiome journal",
  ["hum microbiome j"] = "Human Microbiome Journal",
  ["hum mol genet"] = "Human molecular genetics",
  ["hum mov sci"] = "Human movement science",
  ["hum mutat"] = "Human mutation",
  ["hum nat"] = "Human nature (Hawthorne, N.Y.)",
  ["hum neurobiol"] = "Human neurobiology",
  ["hum nutr appl nutr"] = "Human nutrition. Applied nutrition",
  ["hum nutr clin nutr"] = "Human nutrition. Clinical nutrition",
  ["hum nutr metab"] = "Human Nutrition & Metabolism",
  ["hum organ"] = "Human organization",
  ["hum parasit dis (auckl)"] = "Human parasitic diseases (Auckland, N.Z.)",
  ["hum pathol"] = "Human pathology",
  ["hum pathol (n y)"] = "Human pathology (New York)",
  ["hum pathol: case rep"] = "Human Pathology: Case Reports",
  ["hum perf extrem environ"] = "Human performance in extreme environments : the journal of the Society for Human Performance in Extreme Environments",
  ["hum perform"] = "Human performance",
  ["hum perspect health sci tech"] = "Human Perspectives in Health Sciences and Technology",
  ["hum physiol"] = "Human physiology",
  ["hum psychopharmacol"] = "Human psychopharmacology",
  ["hum psychopharmacol - clin exp"] = "Human Psychopharmacology - Clinical and Experimental",
  ["hum psychopharmacol clin exp"] = "Human Psychopharmacology: Clinical and Experimental",
  ["hum relat"] = "Human relations; studies towards the integration of the social sciences",
  ["hum reprod"] = "Human reproduction (Oxford, England)",
  ["hum reprod genet ethics"] = "Human reproduction and genetic ethics",
  ["hum reprod open"] = "Human reproduction open",
  ["hum reprod update"] = "Human reproduction update",
  ["hum res rep"] = "Human research report",
  ["hum resour dev q"] = "Human resource development quarterly",
  ["hum resour health"] = "Human resources for health",
  ["hum resour manage"] = "Human resource management",
  ["hum rights q"] = "Human rights quarterly",
  ["hum serv organ manag leadersh gov"] = "Human service organizations, management, leadership & governance",
  ["hum stud"] = "Human studies",
  ["hum syst manage"] = "Human Systems Management",
  ["hum toxicol"] = "Human toxicology",
  ["hum vaccin"] = "Human vaccines",
  ["hum vaccin immunother"] = "Human vaccines & immunotherapeutics",
  ["hum vaccines immunother"] = "Human Vaccines & Immunotherapeutics",
  ["hum-anim interact"] = "Humanâ€”Animal Interactions",
  ["hum-centric comput inf sci"] = "Human-centric Computing and Information Sciences",
  ["hum-centric intell syst"] = "Human-Centric Intelligent Systems",
  ["hum-comput interact"] = "Human-Computer Interaction",
  ["hum-comput interact ser"] = "Human-Computer Interaction Series",
  ["hum-intell syst integr"] = "Human-Intelligent Systems Integration",
  ["hum-wildl interact"] = "Human-Wildlife Interactions",
  ["human asp it aged popul (2016)"] = "Human Aspects of IT for the Aged Population : healthy and active aging : second International Conference, ITAP 2016, held as part of HCI International 2016 Toronto, ON, Canada, July 17-22, 2016, Proceedings. Part II. ITAP (Conference) (2nd : 2016 : Toronto, Ont.)",
  ["human comm res ser"] = "Human Communication Research Series",
  ["human geogr"] = "Human geographies",
  ["human resource devel quart"] = "Human Resource Development Quarterly",
  ["human rights"] = "Human rights (Chicago, Ill.)",
  ["humane med"] = "Humane medicine",
  ["humanismus tech"] = "Humanismus und Technik Jahrbuch",
  ["humanit rep"] = "Humanities report",
  ["humanit soc sci commun"] = "Humanities & social sciences communications",
  ["humanitas (monterey n l)"] = "Humanitas",
  ["humanity soc"] = "Humanity & society",
  ["humboldt j soc relat"] = "Humboldt journal of social relations",
  ["humic subst environ"] = "Humic Substances in the Environment",
  ["hunan agric sci technol"] = "Hu'nan agricultural science & technology",
  ["hunan ann math"] = "Hunan Annals of Mathematics",
  ["hunan daxue xuebao"] = "Hunan Daxue Xuebao",
  ["hunan i ko ta hsueh hsueh pao"] = "Hu-Nan I Ko Ta Hsueh Hsueh Pao Bulletin of Hunan Medical University (Chang-Sha",
  ["hunan jiaoyu xueyuan xuebao (ziran kexue)"] = "Hunan Jiaoyu Xueyuan",
  ["hunan keji xueyuan xuebao"] = "Hu'nan Keji Xueyuan xuebao",
  ["hunan nong ye da xue xue bao"] = "Hunan nong ye da xue xue bao = Journal of Hunan Agricultural University",
  ["hunan yi ke da xue xue bao"] = "Hunan yi ke da xue xue bao = Hunan yike daxue xuebao = Bulletin of Hunan Medical University",
  ["hung acta phys"] = "Hungarica Acta Physica",
  ["hung acta physiol"] = "Hungarica acta physiologica",
  ["hung j ind chem"] = "Hungarian Journal of Industrial Chemistry",
  ["hung stud rev"] = "Hungarian studies review : HSR",
  ["huntingt libr q"] = "The Huntington Library quarterly",
  ["husserl stud"] = "Husserl studies",
  ["hutn listy"] = "Hutnicke Listy",
  ["hvac&r res"] = "HVAC&R research",
  ["hw sa us at energy comm"] = "HW-SA [reports]. U.S. Atomic Energy Commission",
  ["hwahak konghak"] = "Hwahak konghak",
  ["hwahak kwa kongop ui chinbo"] = "Hwahak Kwa Kongop Ui Chinbo",
  ["hybrid adv"] = "Hybrid Advances",
  ["hybrid hybridomics"] = "Hybridoma and hybridomics",
  ["hybrid methods eng"] = "Hybrid Methods in Engineering",
  ["hybridoma (larchmt)"] = "Hybridoma (2005)",
  ["hydrobiol j"] = "Hydrobiological journal. American Fisheries Society",
  ["hydrocarbon eng"] = "Hydrocarbon Engineering",
  ["hydrocarbon process, int ed"] = "Hydrocarbon Processing, International Edition",
  ["hydroecol appl"] = "Hydroecologie Appliquee",
  ["hydrogeol j"] = "Hydrogeology journal",
  ["hydrol earth syst sci"] = "Hydrology and Earth System Sciences",
  ["hydrol earth syst sci discuss"] = "Hydrology and Earth System Sciences Discussions",
  ["hydrol j"] = "Hydrogeology Journal",
  ["hydrol process"] = "Hydrological Processes",
  ["hydrol processes"] = "Hydrological Processes",
  ["hydrol res"] = "Hydrology Research",
  ["hydrol res lett"] = "Hydrological Research Letters",
  ["hydrol sci bull"] = "Hydrological Sciences Bulletin",
  ["hydrol sci j"] = "Hydrological Sciences Journal",
  ["hydrol wasserbewirtsch"] = "Hydrologie und Wasserbewirtschaftung",
  ["hydrother physiother"] = "Hydrotherapie-Physiotherapie",
  ["hyg ment"] = "L'Hygiéne mentale",
  ["hyg revy"] = "Hygienisk revy",
  ["hygiea int"] = "Hygiea internationalis",
  ["hyogo univ j"] = "Hyogo University Journal",
  ["hyomen gijutsu"] = "Hyomen Gijutsu",
  ["hyperfine interact"] = "Hyperfine Interactions",
  ["hypertens pregnancy"] = "Hypertension in pregnancy",
  ["hypertens res"] = "Hypertension research : official journal of the Japanese Society of Hypertension",
  ["hypertension (dallas)"] = "Hypertension (Dallas)",
  ["hypothesis (macon)"] = "Hypothesis : the newsletter of the Research Section of MLA",
  ["hypothesis (tor)"] = "Hypothesis (University of Toronto. Dept. of Medical Biophysics)",
  ["hypoxia (auckl)"] = "Hypoxia (Auckland, N.Z.)",
  ["i chuan hsueh pao"] = "I Chuan Hsueh Pao. Acta Genetica Sinica",
  ["i e"] = "I.e. (Wheeling, Ill.)",
  ["iads newsl"] = "IADS newsletter",
  ["iaea bull"] = "IAEA Bulletin",
  ["iaeng int j appl math"] = "IAENG International Journal of Applied Mathematics",
  ["iaeng int j comput sci"] = "IAENG international journal of computer science",
  ["iaes int j artif intell"] = "IAES International Journal on Artificial Intelligence",
  ["iahs publ"] = "IAHS Publication",
  ["ial news"] = "IAL news",
  ["iale bull"] = "IALE (International Association for Landscape Ecology) Bulletin",
  ["iapac mon"] = "IAPAC monthly",
  ["iapqr trans"] = "IAPQR Transactions",
  ["iarc monogr eval carcinog risk chem hum"] = "IARC monographs on the evaluation of the carcinogenic risk of chemicals to humans",
  ["iarc monogr eval carcinog risk chem hum suppl"] = "IARC monographs on the evaluation of the carcinogenic risk of chemicals to humans. Supplement",
  ["iarc monogr eval carcinog risk chem man"] = "IARC monographs on the evaluation of the carcinogenic risk of chemicals to man",
  ["iarc monogr eval carcinog risks hum"] = "IARC monographs on the evaluation of carcinogenic risks to humans",
  ["iarc monogr eval carcinog risks hum suppl"] = "IARC monographs on the evaluation of carcinogenic risks to humans. Supplement",
  ["iarc publ"] = "IARC publications",
  ["iarc sci publ"] = "IARC scientific publications",
  ["ias/park city math ser"] = "IAS/Park City Mathematics Series",
  ["ias/park city math subser"] = "IAS/Park City Mathematical Subseries",
  ["ias/pcmi teach program ser"] = "IAS/PCMI—The Teacher Program Series",
  ["iassist q"] = "IASSIST quarterly",
  ["iasted int conf comput syst biol (2006)"] = "IASTED International Conference on Computational and Systems Biology : November 13-14, 2006, Dallas, Texas, USA (CASB 2006)",
  ["iatr epitheor enoplon dyn"] = "Iatrike epitheoresis enoplon dynameon. Hellenic Armed Forces medical review",
  ["iatss res"] = "IATSS Research",
  ["iau symp"] = "IAU Symposium",
  ["iavi rep"] = "IAVI report : newsletter on international AIDS vaccine research",
  ["iawa j"] = "IAWA journal",
  ["ib sci j sci"] = "IB Scientific Journal of Science",
  ["ibero americana (stockholm)"] = "Ibero-americana (Stockholm, Sweden : 1971)",
  ["iberus (madr)"] = "Iberus : revista de la Sociedad Española de Malacología",
  ["ibis (lond 1859)"] = "The Ibis",
  ["ibm j res dev"] = "IBM Journal of Research and Development",
  ["ibm syst j"] = "IBM Systems Journal",
  ["ibm tech discl bull"] = "IBM Technical Disclosure Bulletin",
  ["ibms bonekey"] = "IBMS boneKEy",
  ["ibro neurosci rep"] = "IBRO neuroscience reports",
  ["ibro rep"] = "IBRO reports",
  ["ibs j sci"] = "IBS journal of science",
  ["ibs materi"] = "IBS-Materialien",
  ["ibsci j sci"] = "IBScientific Journal of Science",
  ["icao j"] = "ICAO journal",
  ["icase/larc interdiscip ser sci eng"] = "ICASE/LaRC Interdisciplinary Series in Science and Engineering",
  ["iccad ieee acm int conf comput aided des"] = "ICCAD. IEEE/ACM International Conference on Computer-Aided Design",
  ["iccm not"] = "ICCM Notices. Notices of the International Congress of Chinese Mathematicians",
  ["iccw j"] = "ICCW journal",
  ["iccw news bull"] = "ICCW news bulletin",
  ["icd sci educ j"] = "Icd Scientific And Educational Journal",
  ["icea news"] = "ICEA news",
  ["icel agric sci"] = "Icelandic agricultural sciences",
  ["ices j mar sci"] = "ICES Journal of Marine Science",
  ["icfai univ j genet evol"] = "ICFAI University journal of genetics and evolution",
  ["ichthyol explor freshw"] = "Ichthyological exploration of freshwaters",
  ["ichthyol herpetol"] = "Ichthyology & Herpetology",
  ["ichthyol res"] = "Ichthyological research",
  ["ici j"] = "ICI Journal",
  ["iciam 2019 sema simai springer ser"] = "ICIAM 2019 SEMA SIMAI Springer Series",
  ["icis chem bus"] = "ICIS Chemical Business",
  ["icis chem bus am"] = "ICIS Chemical Business Americas",
  ["icm ser"] = "ICM Series",
  ["icme-13 monogr"] = "ICME-13 Monographs",
  ["icmh newsl"] = "IMCH newsletter",
  ["icmr 17 (2017)"] = "ICMR'17 : proceedings of the 2017 ACM International Conference on Multimedia Retrieval : June 6-9, 2017, Bucharest, Romania. ACM International Conference on Multimedia Retrieval (2017 : Bucharest, Romania)",
  ["icmr ann"] = "ICMR annals : annual reports of the International Center for Medical Research",
  ["icmr bull"] = "ICMR bulletin",
  ["icon (lond)"] = "Icon (London, England)",
  ["iconogr mycol"] = "Iconographia Mycologica",
  ["icp sel pap"] = "ICP Selected Papers",
  ["icpd 94"] = "ICPD 94 : newsletter of the International Conference on Population and Development",
  ["icrs j int res commun"] = "IRCS journal of international research communications",
  ["icrs med rep"] = "I. C. R. S. medical reports",
  ["icsa book ser stat"] = "ICSA Book Series in Statistics",
  ["icssr newsl"] = "ICSSR newsletter. Indian Council of Social Science Research",
  ["icsu rev"] = "ICSU review",
  ["icsu short rep"] = "ICSU Short Reports",
  ["ict express"] = "ICT Express",
  ["ict syst secur priv prot (2017)"] = "ICT systems security and privacy protection : 32nd IFIP TC 11 International Conference, SEC 2017, Rome, Italy, May 29-31, 2017, Proceedings. IFIP TC11 International Information Security Conference (32nd : 2017 : Rome, Italy)",
  ["ictp ser theoret phys"] = "The ICTP Series in Theoretical Physics",
  ["icu dir"] = "ICU director",
  ["icu manag pract"] = "ICU management & practice",
  ["ida code ida"] = "Idaho code : containing the general laws of Idaho annotated,. Idaho",
  ["ida j desalin water reuse"] = "IDA Journal of Desalination and Water Reuse",
  ["ida law rev"] = "Idaho law review",
  ["idaa commun"] = "I.D.A.A. communique",
  ["idea j law technol"] = "IDEA The Journal of Law and Technology",
  ["ideas context"] = "Ideas in Context",
  ["ideas prod"] = "Ideas & production",
  ["ideggyogy sz"] = "Ideggyógyászati szemle",
  ["identities (yverdon)"] = "Identities (Yverdon, Switzerland)",
  ["identity (mahwah, n j)"] = "Identity (Mahwah, N.J.)",
  ["idesia (arica)"] = "Idesia",
  ["idm materialien studien"] = "IDM Materialien und Studien",
  ["ido rep"] = "IDO [reports]. U.S. Atomic Energy Commission",
  ["idrc rep"] = "The IDRC reports",
  ["ids bull"] = "IDS bulletin",
  ["iee electromagnet waves ser"] = "IEE Electromagnetic Waves Series",
  ["iee hist tech ser"] = "IEE History of Technology Series",
  ["iee j comput digital tech"] = "IEE Journal on Computers and Digital Techniques",
  ["iee j electr power appl"] = "IEE Journal on Electric Power Applications",
  ["iee j electron circuits syst"] = "IEEE Journal on Electronic Circuits and Systems",
  ["iee j microwaves antennas propag"] = "IEE Journal on Microwaves, Antennas & Propagation",
  ["iee proc nanobiotechnol"] = "IEE proceedings. Nanobiotechnology",
  ["iee proc syst biol"] = "IEE proceedings. Systems biology",
  ["iee proc: electr power appl"] = "IEE Proceedings: Electric Power Applications",
  ["iee proc: nanobiotechnol"] = "IEE Proceedings: Nanobiotechnology",
  ["iee proc: optoelectron"] = "IEE Proceedings: Optolectronics",
  ["iee proc: sci meas technol"] = "IEE Proceedings: Science, Measurement & Technology",
  ["iee trans emerg topics comput intell"] = "IEEE Transactions on Emerging Topics in Computational Intelligence",
  ["iee-iere proc - india"] = "IEE-IERE Proceedings - India",
  ["ieee access"] = "IEEE Access",
  ["ieee acm int symp clust cloud grid comput"] = "IEEE/ACM International Symposium on Cluster, Cloud and Grid Computing. IEEE/ACM International Symposium on Cluster, Cloud, and Grid Computing",
  ["ieee acm trans netw"] = "IEEE/ACM transactions on networking : a joint publication of the IEEE Communications Society, the IEEE Computer Society, and the ACM with its Special Interest Group on Data Communication",
  ["ieee aerosp electron syst mag"] = "IEEE Aerospace and Electronic Systems Magazine",
  ["ieee ann hist comput"] = "IEEE Annals of the History of Computing",
  ["ieee antennas propag mag"] = "IEEE Antennas and Propagation Magazine",
  ["ieee antennas wirel propag lett"] = "IEEE Antennas and Wireless Propagation Letters",
  ["ieee antennas wireless propag lett"] = "IEEE Antennas and Wireless Propagation Letters",
  ["ieee appl imag pattern recognit workshop"] = "IEEE Applied Imagery Pattern Recognition Workshop : [proceedings]. IEEE Applied Imagery Pattern Recognition Workshop",
  ["ieee asme int conf adv intell mechatron"] = "IEEE/ASME International Conference on Advanced Intelligent Mechatronics : [proceedings]. IEEE/ASME International Conference on Advanced Intelligent Mechatronics",
  ["ieee asme trans mechatron"] = "IEEE/ASME transactions on mechatronics : a joint publication of the IEEE Industrial Electronics Society and the ASME Dynamic Systems and Control Division",
  ["ieee assp mag"] = "IEEE ASSP Magazine (1984-1990)",
  ["ieee audio, speech, language process"] = "IEEE Transactions on Audio, Speech, and Language Processing",
  ["ieee biom compend"] = "IEEE Biometrics Compendium",
  ["ieee biomed circuits syst conf"] = "IEEE Biomedical Circuits and Systems Conference : healthcare technology : [proceedings]. IEEE Biomedical Circuits and Systems Conference",
  ["ieee biometrics compend"] = "IEEE Biometrics Compendium",
  ["ieee can j electr comput eng"] = "IEEE Canadian Journal of Electrical and Computer Engineering",
  ["ieee china summit int conf signal inf process"] = "IEEE China Summit & International Conference on Signal and Information Processing. IEEE China Summit & International Conference on Signal and Information Processing",
  ["ieee circuits devices mag"] = "IEEE Circuits and Devices Magazine",
  ["ieee circuits syst mag"] = "IEEE Circuits and Systems Magazine",
  ["ieee cloud comput"] = "IEEE Cloud Computing",
  ["ieee cogsima (2020)"] = "IEEE CogSIMA : 2020 IEEE International Conference on Cognitive and Computational Aspects of Situation Management (CogSIMA) : proceedings : Virtual Conference, 24-28 August 2020. IEEE Conference on Cognitive and Computational Aspects of Situation Management (2020 : Online)",
  ["ieee commun lett"] = "IEEE Communications Letters",
  ["ieee commun mag"] = "IEEE Communications Magazine",
  ["ieee commun soc mag"] = "IEEE Communications Society Magazine (through 1978)",
  ["ieee commun stand mag"] = "IEEE Communications Standards Magazine",
  ["ieee commun surv tutorials"] = "IEEE Communication Surveys and Tutorials",
  ["ieee commun surveys tuts"] = "IEEE Communications Surveys and Tutorials",
  ["ieee comput appl power"] = "IEEE Computer Applications in Power",
  ["ieee comput archit lett"] = "IEEE Computer Architecture Letters",
  ["ieee comput graph appl"] = "IEEE computer graphics and applications",
  ["ieee comput graphics appl"] = "IEEE Computer Graphics and Applications",
  ["ieee comput intell mag"] = "IEEE Computational Intelligence Magazine",
  ["ieee comput sci eng"] = "IEEE Computing in Science and Engineering",
  ["ieee comput sci eng mag"] = "IEEE Computation in Science and Engineering Magazine",
  ["ieee computer"] = "IEEE Computer",
  ["ieee concurrency"] = "IEEE Concurrency",
  ["ieee conf games 2019 (2019)"] = "IEEE Conference on Games 2019 : London, United Kingdom, 20-23 August 2019",
  ["ieee consum electron mag"] = "IEEE Consumer Electronics Magazine",
  ["ieee control syst"] = "IEEE Control Systems",
  ["ieee control syst lett"] = "IEEE Control Systems Letters",
  ["ieee control syst mag"] = "IEEE Control Systems Magazine",
  ["ieee control syst mag n y"] = "IEEE control systems magazine (New York, N.Y. : 1985)",
  ["ieee data descr"] = "IEEE Data Descriptions",
  ["ieee des test"] = "IEEE Design & Test",
  ["ieee des test comput"] = "IEEE Design and Test of Computers",
  ["ieee distrib syst online"] = "IEEE Distributed Systems Online",
  ["ieee electr insul mag"] = "IEEE Electrical Insulation Magazine",
  ["ieee electrif mag"] = "IEEE Electrification Magazine",
  ["ieee electrific mag"] = "IEEE Electrification Magazine",
  ["ieee electromagn compat"] = "IEEE Electromagnetic Compatibility Magazine",
  ["ieee electromagn compat mag"] = "IEEE electromagnetic compatibility magazine",
  ["ieee electron device lett"] = "IEEE Electron Device Letters",
  ["ieee electron devices mag"] = "IEEE Electron Devices Magazine",
  ["ieee electrotechnol rev"] = "IEEE ElectroTechnology Review",
  ["ieee embed syst lett"] = "IEEE embedded systems letters",
  ["ieee embedded sys lett"] = "IEEE Embedded Systems Letters",
  ["ieee embedded syst lett"] = "IEEE Embedded Systems Letters",
  ["ieee eng manag rev"] = "IEEE Engineering Management Review",
  ["ieee eng manage rev"] = "IEEE Engineering Management Review",
  ["ieee eng med biol mag"] = "IEEE Engineering in Medicine and Biology Magazine",
  ["ieee expert"] = "IEEE Expert (through 1997)",
  ["ieee geosci remote sens lett"] = "IEEE Geoscience and Remote Sensing Letters",
  ["ieee geosci remote sens mag"] = "IEEE Geoscience and Remote Sensing Magazine",
  ["ieee haptics symp"] = "IEEE Haptics Symposium : [proceedings]. IEEE Haptics Symposium",
  ["ieee ind appl mag"] = "IEEE Industry Applications Magazine",
  ["ieee ind electron mag"] = "IEEE Industrial Electronics Magazine",
  ["ieee instrum meas mag"] = "IEEE Instrumentation and Measurement Magazine",
  ["ieee int conf autom face gesture recognit workshops"] = "IEEE International Conference on Automatic Face & Gesture Recognition and Workshops",
  ["ieee int conf automation sci eng (case)"] = "IEEE International Conference on Automation Science and Engineering (CASE) : [proceedings]. IEEE Conference on Automation Science and Engineering",
  ["ieee int conf bioinform biomed workshops"] = "IEEE International Conference on Bioinformatics and Biomedicine workshops. IEEE International Conference on Bioinformatics and Biomedicine",
  ["ieee int conf control autom"] = "IEEE International Conference on Control and Automation : [proceedings]. International Conference on Control and Automation",
  ["ieee int conf dev learn"] = "IEEE International Conference on Development and Learning : [proceedings]. IEEE International Conference on Development and Learning",
  ["ieee int conf dev learn epigenetic robot"] = "IEEE International Conference on Development and Learning and Epigenetic Robotics : [proceedings]. IEEE International Conference on Development and Learning and Epigenetic Robotics",
  ["ieee int conf electro inf technol"] = "IEEE International Conference on Electro Information Technology : [proceedings]. IEEE International Conference on Electro Information Technology",
  ["ieee int conf healthc inform"] = "IEEE International Conference on Healthcare Informatics. IEEE International Conference on Healthcare Informatics",
  ["ieee int conf rfid"] = "IEEE International Conference on RFID. IEEE International Conference on RFID",
  ["ieee int conf robot autom"] = "IEEE International Conference on Robotics and Automation : ICRA : [proceedings]. IEEE International Conference on Robotics and Automation",
  ["ieee int conf softw qual reliab secur"] = "IEEE International Conference on Software Quality, Reliability and Security : proceedings. IEEE International Conference on Software Quality, Reliability and Security",
  ["ieee int conf systems biol"] = "IEEE International Conference on Systems Biology : [proceedings]. IEEE International Conference on Systems Biology",
  ["ieee int geosci remote sens symp proc"] = "IEEE International Geoscience and Remote Sensing Symposium proceedings. International Geoscience and Remote Sensing Symposium",
  ["ieee int symp circuits syst proc"] = "IEEE International Symposium on Circuits and Systems proceedings. IEEE International Symposium on Circuits and Systems",
  ["ieee int symp parallel distrib process workshops phd forum"] = "IEEE International Symposium on Parallel & Distributed Processing, Workshops and Phd Forum : [proceedings]. IEEE International Symposium on Parallel & Distributed Processing, Workshops and Phd Forum",
  ["ieee int ultrason symp"] = "IEEE International Ultrasonics Symposium : [proceedings]. IEEE International Ultrasonics Symposium",
  ["ieee int vac electron conf"] = "IEEE International Vacuum Electronics Conference. International Vacuum Electronics Conference",
  ["ieee int workshop biomed circuits syst"] = "IEEE International Workshop on Biomedical Circuits & Systems",
  ["ieee int workshop genomic signal process stat"] = "IEEE International Workshop on Genomic Signal Processing and Statistics : [proceedings]. IEEE International Workshop on Genomic Signal Processing and Statistics",
  ["ieee int workshop mach learn signal process"] = "IEEE International Workshop on Machine Learning for Signal Processing : [proceedings]. IEEE International Workshop on Machine Learning for Signal Processing",
  ["ieee intell inform bull"] = "The IEEE intelligent informatics bulletin",
  ["ieee intell syst"] = "IEEE Intelligent Systems",
  ["ieee intell transp syst mag"] = "IEEE Intelligent Transportation Systems Magazine",
  ["ieee internet comput"] = "IEEE Internet Computing",
  ["ieee internet things j"] = "IEEE Internet of Things Journal",
  ["ieee internet things mag"] = "IEEE Internet of Things Magazine",
  ["ieee it prof"] = "IEEE IT Professional",
  ["ieee j biomed health inf"] = "IEEE Journal of Biomedical and Health Informatics",
  ["ieee j biomed health inform"] = "IEEE journal of biomedical and health informatics",
  ["ieee j electromagn"] = "IEEE Journal of Electromagnetics",
  ["ieee j electromagn rf microw med biol"] = "IEEE journal of electromagnetics, RF and microwaves in medicine and biology",
  ["ieee j electromagn rf microwaves med biol"] = "IEEE Journal of Electromagnetics, RF and Microwaves in Medicine and Biology",
  ["ieee j electron devices soc"] = "IEEE Journal of the Electron Devices Society",
  ["ieee j emerg sel top circuits syst"] = "IEEE journal on emerging and selected topics in circuits and systems",
  ["ieee j emerg sel top power electron"] = "IEEE journal of emerging and selected topics in power electronics",
  ["ieee j emerging sel top circuits syst"] = "IEEE Journal on Emerging and Selected Topics in Circuits and Systems",
  ["ieee j emerging sel top ind electron"] = "IEEE Journal of Emerging and Selected Topics in Industrial Electronics",
  ["ieee j emerging sel top power electron"] = "IEEE Journal of Emerging and Selected Topics in Power Electronics",
  ["ieee j explor solid-state comput devices circuits"] = "IEEE Journal on Exploratory Solid-State Computational Devices and Circuits",
  ["ieee j explor solid-state computat devices circuits"] = "IEEE Journal on Exploratory Solid-State Computational Devices and Circuits",
  ["ieee j indoor seamless positioning navig"] = "IEEE Journal on Indoor and Seamless Positioning and Navigation",
  ["ieee j microwaves"] = "IEEE Journal of Microwaves",
  ["ieee j miniaturization air space syst"] = "IEEE Journal on Miniaturization for Air and Space Systems",
  ["ieee j multiscale and multiphys comput techn"] = "IEEE Journal on Multiscale and Multiphysics Computational Techniques",
  ["ieee j multiscale multiphys comput tech"] = "IEEE journal on multiscale and multiphysics computational techniques",
  ["ieee j ocean eng"] = "IEEE Journal of Oceanic Engineering",
  ["ieee j oceanic eng"] = "IEEE Journal of Oceanic Engineering",
  ["ieee j of sel top quantum electron"] = "IEEE Journal of Selected Topics in Quantum Electronics",
  ["ieee j opt commun netw"] = "IEEE Journal of Optical Communications and Networking",
  ["ieee j photovolt"] = "IEEE journal of photovoltaics",
  ["ieee j photovoltaics"] = "IEEE Journal of Photovoltaics",
  ["ieee j product safety eng"] = "IEEE Journal of Product Safety Engineering",
  ["ieee j quantum electron"] = "IEEE Journal of Quantum Electronics",
  ["ieee j radio freq identif (rfid)"] = "IEEE Journal of Radio Frequency Identification (RFID)",
  ["ieee j robot autom"] = "IEEE Journal of Robotics and Automation",
  ["ieee j sel areas commun"] = "IEEE Journal on Selected Areas in Communications",
  ["ieee j sel areas inf theory"] = "IEEE Journal on Selected Areas in Information Theory",
  ["ieee j sel areas sens"] = "IEEE Journal on Selected Areas in Sensors",
  ["ieee j sel top appl earth obs remote sens"] = "IEEE journal of selected topics in applied earth observations and remote sensing",
  ["ieee j sel top quantum electron"] = "IEEE Journal of Selected Topics in Quantum Electronics",
  ["ieee j sel top sign proces"] = "IEEE Journal on Selected Topics in Signal Processing",
  ["ieee j sel top signal process"] = "IEEE journal of selected topics in signal processing",
  ["ieee j sel topics appl earth observ remote sens"] = "IEEE Journal of Selected Topics in Applied Earth Observations and Remote Sensing",
  ["ieee j sel topics quantum electron"] = "IEEE Journal of Selected Topics in Quantum Electronics",
  ["ieee j sel topics signal process"] = "IEEE Journal of Selected Topics in Signal Processing",
  ["ieee j solid-state circuits"] = "IEEE Journal of Solid-State Circuits",
  ["ieee j technol comput aided design"] = "IEEE Journal on Technology in Computer Aided Design",
  ["ieee j transl eng health med"] = "IEEE journal of translational engineering in health and medicine",
  ["ieee lat am trans"] = "IEEE Latin America Transactions",
  ["ieee learn technol"] = "IEEE Learning Technology",
  ["ieee lett comput soc"] = "IEEE letters of the Computer Society",
  ["ieee lett electromagn compat pract appl"] = "IEEE Letters on Electromagnetic Compatibility Practice and Applications",
  ["ieee life sci conf"] = "IEEE Life Sciences Conference. Life Sciences Conference",
  ["ieee life sci lett"] = "IEEE Life Sciences Letters",
  ["ieee magn lett"] = "IEEE Magnetics Letters",
  ["ieee micro"] = "IEEE Micro",
  ["ieee microw guided wave lett"] = "IEEE Microwave and Guided Wave Letters",
  ["ieee microw mag"] = "IEEE microwave magazine",
  ["ieee microw wirel compon lett"] = "IEEE microwave and wireless components letters : a publication of the IEEE Microwave Theory and Techniques Society",
  ["ieee microw wireless compon lett"] = "IEEE Microwave and Wireless Components Letters",
  ["ieee microwave mag"] = "IEEE Microwave Magazine",
  ["ieee microwave wireless compon lett"] = "IEEE Microwave and Wireless Components Letters",
  ["ieee mtt-s 2019 int microw biomed conf imbioc 2019 (2019)"] = "The IEEE MTT-S 2019 International Microwave Biomedical Conference (IMBioC 2019) : proceedings : May 06-08, 2019, International Conference Hotel of Nanjing, Nanjing, China. IEEE MTT-S International Microwave Bio Conference (2019 : Nanjing Shi, China)",
  ["ieee mtts int microw symp"] = "IEEE MTT-S International Microwave Symposium digest. IEEE MTT-S International Microwave Symposium",
  ["ieee multimedia"] = "IEEE Multimedia",
  ["ieee nanatechnol mag"] = "IEEE Nanotechnology Magazine",
  ["ieee nanotechnol express"] = "IEEE Nanotechnology Express",
  ["ieee nanotechnol mag"] = "IEEE nanotechnology magazine",
  ["ieee netw"] = "IEEE network",
  ["ieee network"] = "IEEE Network",
  ["ieee networking lett"] = "IEEE Networking Letters",
  ["ieee nih life sci syst appl workshop"] = "IEEE/NIH Life Science Systems and Applications Workshop. IEEE/NIH Life Science Systems and Applications Workshop",
  ["ieee npss real time conf"] = "IEEE NPSS Real Time Conference : conference record. Conference on Real-Time Computer Applications in Nuclear, Particle, and Plasma Physics",
  ["ieee nucl sci symp conf rec (1997)"] = "IEEE Nuclear Science Symposium conference record. Nuclear Science Symposium",
  ["ieee open access j power energy"] = "IEEE Open Access Journal of Power and Energy",
  ["ieee open j antennas propag"] = "IEEE Open Journal of Antennas and Propagation",
  ["ieee open j circuits syst"] = "IEEE Open Journal of Circuits and Systems",
  ["ieee open j commun soc"] = "IEEE Open Journal of the Communications Society",
  ["ieee open j comput soc"] = "IEEE Open Journal of the Computer Society",
  ["ieee open j control syst"] = "IEEE Open Journal of Control Systems",
  ["ieee open j eng med biol"] = "IEEE open journal of engineering in medicine and biology",
  ["ieee open j immersive disp"] = "IEEE Open Journal on Immersive Displays",
  ["ieee open j ind appl"] = "IEEE Open Journal of Industry Applications",
  ["ieee open j ind electron soc"] = "IEEE Open Journal of the Industrial Electronics Society",
  ["ieee open j instrum meas"] = "IEEE Open Journal of Instrumentation and Measurement",
  ["ieee open j intell transp syst"] = "IEEE Open Journal of Intelligent Transportation Systems",
  ["ieee open j microwave theory tech soc"] = "IEEE Open Journal of the Microwave Theory and Techniques Society",
  ["ieee open j nanotechnol"] = "IEEE Open Journal of Nanotechnology",
  ["ieee open j power electron"] = "IEEE Open Journal of Power Electronics",
  ["ieee open j signal process"] = "IEEE Open Journal of Signal Processing",
  ["ieee open j solid-state circuits"] = "IEEE Open Journal of Solid-State Circuits",
  ["ieee open j syst eng"] = "IEEE Open Journal of Systems Engineering",
  ["ieee open j ultrason ferroelectr freq control"] = "IEEE Open Journal on Ultrasonics, Ferroelectrics, and Frequency Control",
  ["ieee open j veh technol"] = "IEEE Open Journal of Vehicular Technology",
  ["ieee pac vis symp"] = "IEEE Pacific Visualization Symposium : [proceedings]. IEEE Pacific Visualisation Symposium",
  ["ieee pers commun"] = "IEEE Personal Communications",
  ["ieee personal commun mag"] = "IEEE Personal Communications Magazine",
  ["ieee pervasive comput"] = "IEEE Pervasive Computing",
  ["ieee photon j"] = "IEEE Photonics Journal",
  ["ieee photon technol lett"] = "IEEE Photonics Technology Letters",
  ["ieee photonics conf"] = "IEEE Photonics Conference : [proceedings]. IEEE Photonics Conference",
  ["ieee photonics j"] = "IEEE Photonics Journal",
  ["ieee photonics technol lett"] = "IEEE Photonics Technology Letters",
  ["ieee photonicsglob singap"] = "IEEE PhotonicsGlobal[at]Singapore",
  ["ieee potentials"] = "IEEE Potentials",
  ["ieee power electron lett"] = "IEEE Power Electronics Letters",
  ["ieee power electron mag"] = "IEEE Power Electronics Magazine",
  ["ieee power energ mag"] = "IEEE Power and Energy Magazine",
  ["ieee power energy mag"] = "IEEE Power and Energy Magazine",
  ["ieee power energy technol syst j"] = "IEEE Power and Energy Technology Systems Journal",
  ["ieee power eng rev"] = "IEEE Power Engineering Review",
  ["ieee press ser comput intell"] = "IEEE Press Series on Computational Intelligence",
  ["ieee press ser power eng"] = "IEEE Press Series on Power Engineering",
  ["ieee press ser rf microw technol"] = "IEEE Press Series on RF and Microwave Technology",
  ["ieee press underst sci technol ser"] = "IEEE Press Understanding Science & Technology Series",
  ["ieee proc solid state electron devices"] = "IEEE Proceedings I - Solid-State and Electron Devices",
  ["ieee pulse"] = "IEEE pulse",
  ["ieee reliab mag"] = "IEEE Reliability Magazine",
  ["ieee rev biomed eng"] = "IEEE reviews in biomedical engineering",
  ["ieee rev iberoam tecnol aprendizaje"] = "IEEE Revista Iberoamericana de Tecnologias del Aprendizaje",
  ["ieee revista iberoamericana de technolog’ias del aprendizaje"] = "IEEE Revista Iberoamericana de Technolog’ias del Aprendizaje",
  ["ieee rfic j"] = "IEEE RFIC Journal",
  ["ieee rfic virtual j"] = "IEEE RFIC Virtual Journal",
  ["ieee rfid j"] = "IEEE RFID Journal",
  ["ieee rfid virtual j"] = "IEEE RFID Virtual Journal",
  ["ieee rob autom lett"] = "IEEE Robotics and Automation Letters",
  ["ieee rob autom mag"] = "IEEE Robotics and Automation Magazine",
  ["ieee rob autom pract"] = "IEEE Robotics and Automation Practice",
  ["ieee robot autom lett"] = "IEEE robotics and automation letters",
  ["ieee robot autom mag"] = "IEEE robotics & automation magazine",
  ["ieee secur priv"] = "IEEE security & privacy",
  ["ieee secur privacy"] = "IEEE Security and Privacy",
  ["ieee security privacy"] = "IEEE Security and Privacy",
  ["ieee sens j"] = "IEEE sensors journal",
  ["ieee sens lett"] = "IEEE sensors letters",
  ["ieee sens rev"] = "IEEE Sensors Reviews",
  ["ieee sensors j"] = "IEEE Sensors Journal",
  ["ieee sice rsj int conf multisens fusion integr intell syst"] = "IEEE/SICE/RSJ International Conference on Multisensor Fusion and Integration for Intelligent Systems. IEEE/SICE/RSJ International Conference on Multisensor Fusion and Integration for Intelligent Systems",
  ["ieee signal process lett"] = "IEEE Signal Processing Letters",
  ["ieee signal process mag"] = "IEEE Signal Processing Magazine",
  ["ieee softw"] = "IEEE Software",
  ["ieee software"] = "IEEE Software",
  ["ieee solid state circuits lett"] = "IEEE solid-state circuits letters",
  ["ieee solid-state circuits lett"] = "IEEE Solid-State Circuits Letters",
  ["ieee solid-state circuits mag"] = "IEEE Solid-State Circuits Magazine",
  ["ieee spectr"] = "IEEE Spectrum",
  ["ieee sustain comput"] = "IEEE Sustainable Computing",
  ["ieee symp comput intell bioinforma comput biol proc"] = "IEEE Symposium on Computational Intelligence in Bioinformatics and Computational Biology proceedings. IEEE Symposium on Computational Intelligence in Bioinformatics and Computational Biology",
  ["ieee symp comput intell data min"] = "IEEE Symposium on Computational Intelligence and Data Mining. IEEE Symposium on Computational Intelligence and Data Mining",
  ["ieee syst j"] = "IEEE Systems Journal",
  ["ieee syst man cybern lett"] = "IEEE Systems, Man, and Cybernetics Letters",
  ["ieee syst man cybern mag"] = "IEEE systems, man, and cybernetics magazine",
  ["ieee syst, man, cybern mag"] = "IEEE Systems, Man, and Cybernetics Magazine",
  ["ieee technol soc mag"] = "IEEE Technology and Society Magazine",
  ["ieee trans acoust"] = "IEEE transactions on acoustics, speech, and signal processing",
  ["ieee trans acoust speech signal process"] = "IEEE Transactions on Acoustics, Speech, and Signal Processing",
  ["ieee trans acoust, speech, signal process"] = "IEEE Transactions on Acoustics, Speech, and Signal Processing",
  ["ieee trans adv packag"] = "IEEE Transactions on Advanced Packaging",
  ["ieee trans aeronaut navig electron"] = "IEEE Transactions on Aeronautical and Navigational Electronics",
  ["ieee trans aerosp"] = "IEEE Transactions on Aerospace",
  ["ieee trans aerosp electron syst"] = "IEEE Transactions on Aerospace and Electronic Systems",
  ["ieee trans aerosp navig electron"] = "IEEE Transactions on Aerospace and Navigation Electronics",
  ["ieee trans affect comput"] = "IEEE transactions on affective computing",
  ["ieee trans affective comput"] = "IEEE Transactions on Affective Computing",
  ["ieee trans agrifood electron"] = "IEEE Transactions on AgriFood Electronics",
  ["ieee trans airborne electron"] = "IEEE Transactions on Airborne Electronics",
  ["ieee trans antennas and propagation"] = "Institute of Electrical and Electronics Engineers. Transactions on Antennas and Propagation",
  ["ieee trans antennas propag"] = "IEEE Transactions on Antennas and Propagation",
  ["ieee trans appl ind"] = "IEEE Transactions on Applications and Industry",
  ["ieee trans appl indus"] = "IEEE Transactions on Applications and Industry",
  ["ieee trans appl supercond"] = "IEEE Transactions on Applied Superconductivity",
  ["ieee trans artif intell"] = "IEEE Transactions on Artificial Intelligence",
  ["ieee trans audio"] = "IEEE Transactions on Audio",
  ["ieee trans audio electroacoust"] = "IEEE Transactions on Audio and Electroacoustics",
  ["ieee trans audio speech lang process"] = "IEEE Transactions on Audio, Speech and Language Processing",
  ["ieee trans autom control"] = "IEEE Transactions on Automatic Control",
  ["ieee trans autom sci eng"] = "IEEE Transactions on Automation Science and Engineering",
  ["ieee trans automat contr"] = "IEEE transactions on automatic control",
  ["ieee trans automat control"] = "Institute of Electrical and Electronics Engineers. Transactions on Automatic Control",
  ["ieee trans auton ment dev"] = "IEEE Transactions on Autonomous Mental Development",
  ["ieee trans auton mental develop"] = "IEEE Transactions on Autonomous Mental Development",
  ["ieee trans big data"] = "IEEE Transactions on Big Data",
  ["ieee trans bio-med electron"] = "IEEE Transactions on Bio-Medical Electronics",
  ["ieee trans bio-med eng"] = "IEEE Transactions on Bio-Medical Engineering",
  ["ieee trans biom behav identity sci"] = "IEEE transactions on biometrics, behavior, and identity science",
  ["ieee trans biomed circuits syst"] = "IEEE transactions on biomedical circuits and systems",
  ["ieee trans biomed eng"] = "IEEE Transactions on Biomedical Engineering",
  ["ieee trans broadcast"] = "IEEE Transactions on Broadcasting",
  ["ieee trans broadcast telev receivers"] = "IEEE Transactions on Broadcast and Television Receivers",
  ["ieee trans broadcast television receivers"] = "IEEE Transactions on Broadcast and Television Receivers",
  ["ieee trans cable telev"] = "IEEE Transactions on Cable Television",
  ["ieee trans circuit theory"] = "IEEE Transactions on Circuit Theory",
  ["ieee trans circuits syst"] = "IEEE transactions on circuits and systems",
  ["ieee trans circuits syst artif intell"] = "IEEE Transactions on Circuits and Systems for Artificial Intelligence",
  ["ieee trans circuits syst express briefs"] = "IEEE Transactions on Circuits and Systems II: Express Briefs",
  ["ieee trans circuits syst i"] = "IEEE Transactions on Circuits and Systems—Part I: Fundamental Theory and Applications",
  ["ieee trans circuits syst i fundam theory appl"] = "IEEE transactions on circuits and systems. I, Fundamental theory and applications : a publication of the IEEE Circuits and Systems Society",
  ["ieee trans circuits syst i regul pap"] = "IEEE transactions on circuits and systems. I, Regular papers : a publication of the IEEE Circuits and Systems Society",
  ["ieee trans circuits syst ii"] = "IEEE Transactions on Circuits and Systems—Part II: Analog and Digital Signal Processing",
  ["ieee trans circuits syst ii express briefs"] = "IEEE transactions on circuits and systems. II, Express briefs : a publication of the IEEE Circuits and Systems Society",
  ["ieee trans circuits syst regul pap"] = "IEEE Transactions on Circuits and Systems I: Regular Papers",
  ["ieee trans circuits syst video technol"] = "IEEE Transactions on Circuits and Systems for Video Technology",
  ["ieee trans circuits systems i fund theory appl"] = "IEEE Transactions on Circuits and Systems",
  ["ieee trans cloud comput"] = "IEEE Transactions on Cloud Computing",
  ["ieee trans cogn commun netw"] = "IEEE transactions on cognitive communications and networking",
  ["ieee trans cogn dev syst"] = "IEEE transactions on cognitive and developmental systems",
  ["ieee trans cogn develop syst"] = "IEEE Transactions on Cognitive and Developmental Systems",
  ["ieee trans cognit commun networking"] = "IEEE Transactions on Cognitive Communications and Networking",
  ["ieee trans cognit dev syst"] = "IEEE Transactions on Cognitive and Developmental Systems",
  ["ieee trans commun"] = "IEEE Transactions on Communications",
  ["ieee trans commun syst"] = "IEEE Transactions on Communications Systems",
  ["ieee trans commun technol"] = "IEEE Transactions on Communication Technology",
  ["ieee trans comp, packag, manufact technol a"] = "IEEE Transactions on Components, Packaging and Manufacturing Technology—Part A",
  ["ieee trans comp, packag, manufact technol b"] = "IEEE Transactions on Components, Packaging and Manufacturing Technology—Part B: Advanced Packaging",
  ["ieee trans comp, packag, manufact technol c"] = "IEEE Transactions on Components, Packaging and Manufacturing Technology—Part C: Manufacturing",
  ["ieee trans compon packag manuf technol"] = "IEEE Transactions on Components, Packaging and Manufacturing Technology",
  ["ieee trans compon packag manuf technol part a:"] = "IEEE Transactions on Components, Packaging, and Manufacturing Technology, Part A:",
  ["ieee trans compon packag manuf technol part b:"] = "IEEE Transactions on Components, Packaging, and Manufacturing Technology, Part B:",
  ["ieee trans compon packag manuf technol part c:"] = "IEEE Transactions on Components, Packaging, and Manufacturing Technology, Part C:",
  ["ieee trans compon packag technol"] = "IEEE Transactions on Components and Packaging Technologies",
  ["ieee trans compon packaging manuf technol"] = "IEEE transactions on components, packaging, and manufacturing technology",
  ["ieee trans compon parts"] = "IEEE Transactions on Component Parts",
  ["ieee trans compon, hybrids, manuf technol"] = "IEEE Transactions on Components, Hybrids and Manufacturing Technology",
  ["ieee trans compon, packag, manuf technol a"] = "IEEE Transactions on Components, Packaging and Manufacturing Technology A",
  ["ieee trans compon, packag, manuf technol b"] = "IEEE Transactions on Components, Packaging and Manufacturing Technology B: Advanced Packaging",
  ["ieee trans compon, packag, manuf technol c"] = "IEEE Transactions on Components, Packaging and Manufacturing Technology C: Manufacturing",
  ["ieee trans comput"] = "IEEE Transactions on Computers",
  ["ieee trans comput aided des integr circuits syst"] = "IEEE Transactions on Computer-Aided Design of Integrated Circuits and Systems",
  ["ieee trans comput imaging"] = "IEEE Transactions on Computational Imaging",
  ["ieee trans comput intell ai games"] = "IEEE transactions on computational intelligence and AI in games",
  ["ieee trans comput intell ai in games"] = "IEEE Transactions on Computational Intelligence and AI in Games",
  ["ieee trans comput soc syst"] = "IEEE transactions on computational social systems",
  ["ieee trans comput social syst"] = "IEEE Transactions on Computational Social Systems",
  ["ieee trans comput-aided design integr circuits syst"] = "IEEE Transactions on Computer-Aided Design of Integrated Circuits and Systems",
  ["ieee trans consum electron"] = "IEEE Transactions on Consumer Electronics",
  ["ieee trans control netw syst"] = "IEEE Transactions on Control of Network Systems",
  ["ieee trans control network syst"] = "IEEE Transactions on Control of Network Systems",
  ["ieee trans control syst technol"] = "IEEE Transactions on Control Systems Technology",
  ["ieee trans cybern"] = "IEEE transactions on cybernetics",
  ["ieee trans depend sec comput"] = "IEEE Transactions on Dependable and Secure Computing",
  ["ieee trans dependable secure comput"] = "IEEE Transactions on Dependable and Secure Computing",
  ["ieee trans device mater rel"] = "IEEE Transactions on Device and Materials Reliability",
  ["ieee trans device mater reliab"] = "IEEE Transactions on Device and Materials Reliability",
  ["ieee trans dielectr electr insul"] = "IEEE Transactions on Dielectrics and Electrical Insulation",
  ["ieee trans educ"] = "IEEE Transactions on Education",
  ["ieee trans electr insul"] = "IEEE Transactions on Electrical Insulation",
  ["ieee trans electromagn compat"] = "IEEE Transactions on Electromagnetic Compatibility",
  ["ieee trans electron comput"] = "IEEE Transactions on Electronic Computers",
  ["ieee trans electron devices"] = "IEEE Transactions on Electron Devices",
  ["ieee trans electron packag manuf"] = "IEEE Transactions on Electronics Packaging Manufacturing",
  ["ieee trans emerg sel topics circuits syst"] = "IEEE Transactions on Emerging and Selected Topics in Circuits and Systems",
  ["ieee trans emerg sel topics power electron"] = "IEEE Transactions on Emerging and Selected Topics in Power Electronics",
  ["ieee trans emerg top comput"] = "IEEE transactions on emerging topics in computing",
  ["ieee trans emerg top comput intell"] = "IEEE transactions on emerging topics in computational intelligence",
  ["ieee trans emerg topics comput"] = "IEEE Transactions on Emerging Topics in Computing",
  ["ieee trans emerging top comput"] = "IEEE Transactions on Emerging Topics in Computing",
  ["ieee trans emerging top comput intell"] = "IEEE Transactions on Emerging Topics in Computational Intelligence",
  ["ieee trans energy convers"] = "IEEE Transactions on Energy Conversion",
  ["ieee trans energy mark policy regul"] = "IEEE Transactions on Energy Markets, Policy and Regulation",
  ["ieee trans eng manag"] = "IEEE transactions on engineering management",
  ["ieee trans eng manage"] = "IEEE Transactions on Engineering Management",
  ["ieee trans eng writing speech"] = "IEEE Transactions on Engineering Writing and Speech",
  ["ieee trans evol comput"] = "IEEE Transactions on Evolutionary Computation",
  ["ieee trans field rob"] = "IEEE Transactions on Field Robotics",
  ["ieee trans fuzzy syst"] = "IEEE Transactions on Fuzzy Systems",
  ["ieee trans games"] = "IEEE transactions on games",
  ["ieee trans geosci electron"] = "IEEE Transactions on Geoscience Electronics",
  ["ieee trans geosci remote sens"] = "IEEE Transactions on Geoscience and Remote Sensing",
  ["ieee trans green commun netw"] = "IEEE Transactions on Green Communications and Networking",
  ["ieee trans green commun networking"] = "IEEE Transactions on Green Communications and Networking",
  ["ieee trans haptic"] = "IEEE Transactions on Haptics",
  ["ieee trans haptics"] = "IEEE transactions on haptics",
  ["ieee trans hum factors electron"] = "IEEE Transactions on Human Factors in Electronics",
  ["ieee trans hum mach syst"] = "IEEE transactions on human-machine systems",
  ["ieee trans hum-mach syst"] = "IEEE Transactions on Human-Machine Systems",
  ["ieee trans human-mach syst"] = "IEEE Transactions on Human-Machine Systems",
  ["ieee trans image process"] = "IEEE Transactions on Image Processing",
  ["ieee trans ind appl"] = "IEEE Transactions on Industry Applications",
  ["ieee trans ind cyber-phys syst"] = "IEEE Transactions on Industrial Cyber-Physical Systems",
  ["ieee trans ind electron"] = "IEEE Transactions on Industrial Electronics",
  ["ieee trans ind electron contr instrum"] = "IEEE Transactions on Industrial Electronics and Control Instrumentation",
  ["ieee trans ind gen appl"] = "IEEE Transactions on Industry and General Applications",
  ["ieee trans ind inf"] = "IEEE Transactions on Industrial Informatics",
  ["ieee trans ind informat"] = "IEEE Transactions on Industrial Informatics",
  ["ieee trans industr inform"] = "IEEE transactions on industrial informatics",
  ["ieee trans inf forensics secur"] = "IEEE Transactions on Information Forensics and Security",
  ["ieee trans inf forensics security"] = "IEEE Transactions on Information Forensics and Security",
  ["ieee trans inf technol biomed"] = "IEEE Transactions on Information Technology in Biomedicine",
  ["ieee trans inf theory"] = "IEEE Transactions on Information Theory",
  ["ieee trans inform theory"] = "Institute of Electrical and Electronics Engineers. Transactions on Information Theory",
  ["ieee trans instrum meas"] = "IEEE Transactions on Instrumentation and Measurement",
  ["ieee trans intell transp syst"] = "IEEE Transactions on Intelligent Transportation Systems",
  ["ieee trans intell veh"] = "IEEE Transactions on Intelligent Vehicles",
  ["ieee trans knowl data eng"] = "IEEE Transactions on Knowledge and Data Engineering",
  ["ieee trans learn technol"] = "IEEE Transactions on Learning Technologies",
  ["ieee trans mach learn commun networking"] = "IEEE Transactions on Machine Learning in Communications and Networking",
  ["ieee trans magn"] = "IEEE Transactions on Magnetics",
  ["ieee trans man-mach syst"] = "IEEE Transactions on Man-Machine Systems",
  ["ieee trans manuf technol"] = "IEEE Transactions on Manufacturing Technology",
  ["ieee trans mater electron devices"] = "IEEE Transactions on Materials for Electron Devices",
  ["ieee trans med electron"] = "IEEE Transactions on Medical Electronics",
  ["ieee trans med imag"] = "IEEE Transactions on Medical Imaging",
  ["ieee trans med imaging"] = "IEEE transactions on medical imaging",
  ["ieee trans med rob bionics"] = "IEEE Transactions on Medical Robotics and Bionics",
  ["ieee trans med robot bionics"] = "IEEE transactions on medical robotics and bionics",
  ["ieee trans microw theory tech"] = "IEEE transactions on microwave theory and techniques",
  ["ieee trans microw theory techn"] = "IEEE Transactions on Microwave Theory and Techniques",
  ["ieee trans microwave theory tech"] = "IEEE Transactions on Microwave Theory and Techniques",
  ["ieee trans mil electron"] = "IEEE Transactions on Military Electronics",
  ["ieee trans mob comput"] = "IEEE Transactions on Mobile Computing",
  ["ieee trans mobile comput"] = "IEEE Transactions on Mobile Computing",
  ["ieee trans mol biol multi-scale commun"] = "IEEE Transactions on Molecular, Biological, and Multi-Scale Communications",
  ["ieee trans mol biol multiscale commun"] = "IEEE transactions on molecular, biological, and multi-scale communications",
  ["ieee trans multi-scale comput syst"] = "IEEE Transactions on Multi-Scale Computing Systems",
  ["ieee trans multimedia"] = "IEEE Transactions on Multimedia",
  ["ieee trans nanobiosci"] = "IEEE Transactions on Nanobioscience",
  ["ieee trans nanobioscience"] = "IEEE transactions on nanobioscience",
  ["ieee trans nanotechnol"] = "IEEE Transactions on Nanotechnology",
  ["ieee trans netw sci eng"] = "IEEE transactions on network science and engineering",
  ["ieee trans netw serv manage"] = "IEEE Transactions on Network and Service Management",
  ["ieee trans netw service manag"] = "IEEE Transactions on Network and Service Management",
  ["ieee trans network sci eng"] = "IEEE Transactions on Network Science and Engineering",
  ["ieee trans network serv manage"] = "IEEE Transactions on Network and Service Management",
  ["ieee trans neural netw"] = "IEEE transactions on neural networks",
  ["ieee trans neural netw learn syst"] = "IEEE transactions on neural networks and learning systems",
  ["ieee trans neural networks"] = "IEEE Transactions on Neural Networks",
  ["ieee trans neural networks learn syst"] = "IEEE Transactions on Neural Networks and Learning Systems",
  ["ieee trans neural syst rehabil eng"] = "IEEE transactions on neural systems and rehabilitation engineering : a publication of the IEEE Engineering in Medicine and Biology Society",
  ["ieee trans nucl sci"] = "IEEE Transactions on Nuclear Science",
  ["ieee trans on cloud comput"] = "IEEE Transactions on Cloud Computing",
  ["ieee trans on cogn commun netw"] = "IEEE Transactions on Cognitive Communications and Networking",
  ["ieee trans parallel distrib syst"] = "IEEE Transactions on Parallel and Distributed Systems",
  ["ieee trans parts, hybrids, packag"] = "IEEE Transactions on Parts, Hybrids and Packaging",
  ["ieee trans parts, mater, packag"] = "IEEE Transactions on Parts, Materials and Packaging",
  ["ieee trans pattern anal mach intell"] = "IEEE transactions on pattern analysis and machine intelligence",
  ["ieee trans plasma sci"] = "IEEE Transactions on Plasma Science",
  ["ieee trans plasma sci ieee nucl plasma sci soc"] = "IEEE transactions on plasma science. IEEE Nuclear and Plasma Sciences Society",
  ["ieee trans power app syst"] = "IEEE Transactions on Power Apparatus and Systems",
  ["ieee trans power del"] = "IEEE Transactions on Power Delivery",
  ["ieee trans power delivery"] = "IEEE Transactions on Power Delivery",
  ["ieee trans power electron"] = "IEEE Transactions on Power Electronics",
  ["ieee trans power syst"] = "IEEE Transactions on Power Systems",
  ["ieee trans privacy"] = "IEEE Transactions on Privacy",
  ["ieee trans prof commun"] = "IEEE Transactions on Professional Communication",
  ["ieee trans quantum eng"] = "IEEE Transactions on Quantum Engineering",
  ["ieee trans radar syst"] = "IEEE Transactions on Radar Systems",
  ["ieee trans radiat plasma med sci"] = "IEEE transactions on radiation and plasma medical sciences",
  ["ieee trans radio freq interference"] = "IEEE Transactions on Radio Frequency Interference",
  ["ieee trans rehabil eng"] = "IEEE Transactions on Rehabilitation Engineering",
  ["ieee trans rel"] = "IEEE Transactions on Reliability",
  ["ieee trans reliab"] = "IEEE Transactions on Reliability",
  ["ieee trans rob"] = "IEEE Transactions on Robotics",
  ["ieee trans rob autom"] = "IEEE transactions on robotics and automation : a publication of the IEEE Robotics and Automation Society",
  ["ieee trans robot"] = "IEEE transactions on robotics : a publication of the IEEE Robotics and Automation Society",
  ["ieee trans robot autom"] = "IEEE Transactions on Robotics and Automation",
  ["ieee trans semicond manuf"] = "IEEE Transactions on Semiconductor Manufacturing",
  ["ieee trans semicond technol model simul"] = "IEEE Transactions on Semiconductor Technology Modeling and Simulation",
  ["ieee trans serv comput"] = "IEEE Transactions on Services Computing",
  ["ieee trans signal inf process netw"] = "IEEE transactions on signal and information processing over networks",
  ["ieee trans signal inf process networks"] = "IEEE Transactions on Signal and Information Processing over Networks",
  ["ieee trans signal inform process netw"] = "IEEE Transactions on Signal and Information Processing over Networks",
  ["ieee trans signal power integr"] = "IEEE Transactions on Signal and Power Integrity",
  ["ieee trans signal process"] = "IEEE Transactions on Signal Processing",
  ["ieee trans smart grid"] = "IEEE Transactions on Smart Grid",
  ["ieee trans softw eng"] = "IEEE Transactions on Software Engineering",
  ["ieee trans software eng"] = "IEEE Transactions on Software Engineering",
  ["ieee trans sonics ultrason"] = "IEEE Transactions on Sonics and Ultrasonics",
  ["ieee trans space electron telem"] = "IEEE Transactions on Space Electronics and Telemetry",
  ["ieee trans speech audio process"] = "IEEE transactions on speech and audio processing : a publication of the IEEE Signal Processing Society",
  ["ieee trans sustain comput"] = "IEEE Transactions on Sustainable Computing",
  ["ieee trans sustain energy"] = "IEEE transactions on sustainable energy",
  ["ieee trans sustainable comput"] = "IEEE Transactions on Sustainable Computing",
  ["ieee trans sustainable energy"] = "IEEE Transactions on Sustainable Energy",
  ["ieee trans syst man cybern"] = "IEEE transactions on systems, man, and cybernetics",
  ["ieee trans syst man cybern a syst hum"] = "IEEE transactions on systems, man, and cybernetics. Part A, Systems and humans : a publication of the IEEE Systems, Man, and Cybernetics Society",
  ["ieee trans syst man cybern b cybern"] = "IEEE transactions on systems, man, and cybernetics. Part B, Cybernetics : a publication of the IEEE Systems, Man, and Cybernetics Society",
  ["ieee trans syst man cybern c appl rev"] = "IEEE transactions on systems, man and cybernetics. Part C, Applications and reviews : a publication of the IEEE Systems, Man, and Cybernetics Society",
  ["ieee trans syst man cybern part a syst humans"] = "IEEE Transactions on Systems, Man, and Cybernetics Part A: Systems and Humans",
  ["ieee trans syst man cybern part b cybern"] = "IEEE Transactions on Systems, Man, and Cybernetics Part B: Cybernetics",
  ["ieee trans syst man cybern part c appl rev"] = "IEEE Transactions on Systems, Man, and Cybernetics Part C: Applications and Reviews",
  ["ieee trans syst man cybern syst"] = "IEEE transactions on systems, man, and cybernetics. Systems",
  ["ieee trans syst man cybern: syst"] = "IEEE Transactions on Systems, Man, and Cybernetics: Systems",
  ["ieee trans syst sci cybern"] = "IEEE Transactions on Systems Science and Cybernetics",
  ["ieee trans syst, man, cybern"] = "IEEE Transactions on Systems, Man, and Cybernetics",
  ["ieee trans syst, man, cybern a"] = "IEEE Transactions on Systems, Man, and Cybernetics—Part A: Systems and Humans",
  ["ieee trans syst, man, cybern b"] = "IEEE Transactions on Systems, Man, and Cybernetics—Part B: Cybernetics",
  ["ieee trans syst, man, cybern c"] = "IEEE Transactions on Systems, Man, and Cybernetics—Part C: Applications and Reviews",
  ["ieee trans syst, man, cybern, syst"] = "IEEE Transactions on Systems, Man, and Cybernetics: Systems",
  ["ieee trans technol soc"] = "IEEE Transactions on Technology and Society",
  ["ieee trans terahertz sci technol"] = "IEEE Transactions on Terahertz Science and Technology",
  ["ieee trans thz sci technol"] = "IEEE Transactions on Terahertz Science and Technology",
  ["ieee trans transp electrif"] = "IEEE Transactions on Transportation Electrification",
  ["ieee trans transport electrific"] = "IEEE Transactions on Transportation Electrification",
  ["ieee trans ultrason eng"] = "IEEE Transactions on Ultrasonics Engineering",
  ["ieee trans ultrason ferroelectr freq control"] = "IEEE transactions on ultrasonics, ferroelectrics, and frequency control",
  ["ieee trans ultrason, ferroelectr, freq control"] = "IEEE Transactions on Ultrasonics, Ferroelectrics, and Frequency Control",
  ["ieee trans veh commun"] = "IEEE Transactions on Vehicular Communications",
  ["ieee trans veh technol"] = "IEEE Transactions on Vehicular Technology",
  ["ieee trans very large scale integr vlsi syst"] = "IEEE Transactions on Very Large Scale Integration (VLSI) Systems",
  ["ieee trans vis comput graph"] = "IEEE transactions on visualization and computer graphics",
  ["ieee trans vis comput graphics"] = "IEEE Transactions on Visualization and Computer Graphics",
  ["ieee trans visual comput graphics"] = "IEEE Transactions on Visualization and Computer Graphics",
  ["ieee trans vlsi syst"] = "IEEE Transactions on Very Large Scale Integration (VLSI) Systems",
  ["ieee trans wirel commun"] = "IEEE transactions on wireless communications",
  ["ieee trans wireless commun"] = "IEEE Transactions on Wireless Communications",
  ["ieee transl j magn jpn"] = "IEEE Translation Journal on Magnetics in Japan",
  ["ieee veh technol mag"] = "IEEE Vehicular Technology Magazine",
  ["ieee winter conf appl comput vis"] = "IEEE Winter Conference on Applications of Computer Vision. IEEE Winter Conference on Applications of Computer Vision",
  ["ieee wirel commun"] = "IEEE wireless communications",
  ["ieee wireless commun"] = "IEEE Wireless Communications",
  ["ieee wireless commun lett"] = "IEEE Wireless Communications Letters",
  ["ieee wireless commun mag"] = "IEEE Wireless Communications Magazine",
  ["ieee women eng mag"] = "IEEE Women in Engineering Magazine",
  ["ieee workshop microelectron electron devices"] = "IEEE Workshop on Microelectronics and Electron Devices : [proceedings]. IEEE Workshop on Microelectronics and Electron Devices",
  ["ieee/acm int conf autom softw eng"] = "IEEE/ACM International Conference on Automated Software Engineering : [proceedings]. IEEE/ACM International Conference on Automated Software Engineering",
  ["ieee/acm int conf autom softw eng workshops"] = "IEEE/ACM International Conference on Automated Software Engineering workshops. IEEE/ACM International Conference on Automated Software Engineering",
  ["ieee/acm trans audio speech lang process"] = "IEEE/ACM transactions on audio, speech, and language processing",
  ["ieee/acm trans audio, speech, language process"] = "IEEE/ACM Transactions on Audio, Speech, and Language Processing",
  ["ieee/acm trans comput biol bioinf"] = "IEEE/ACM Transactions on Computational Biology and Bioinformatics",
  ["ieee/acm trans comput biol bioinform"] = "IEEE/ACM transactions on computational biology and bioinformatics",
  ["ieee/acm trans comput biol bioinformatics"] = "IEEE/ACM Transactions on Computational Biology and Bioinformatics",
  ["ieee/acm trans netw"] = "IEEE/ACM Transactions on Networking",
  ["ieee/acm trans networking"] = "IEEE/ACM Transactions on Networking",
  ["ieee/asme trans mechatron"] = "IEEE/ASME Transactions on Mechatronics",
  ["ieee/asme trans mechatronics"] = "IEEE/ASME Transactions on Mechatronics",
  ["ieee/caa j autom sin"] = "IEEE/CAA Journal of Automatica Sinica",
  ["ieee/caa j autom sinica"] = "IEEE/CAA Journal of Automatica Sinica",
  ["ieee/ecs electrochem solid-state lett"] = "IEEE/ECS Electrochemical and Solid-State Letters",
  ["ieee/osa j disp technol"] = "IEEE/OSA Journal of Display Technology",
  ["ieee/osa j display technol"] = "IEEE/OSA Journal of Display Technology",
  ["ieee/osa j opt commun networking"] = "IEEE/OSA Journal of Optical Communications and Networking",
  ["ieee/oup class reissue"] = "An IEEE/OUP Classic Reissue",
  ["ieee/oup ser electromagn wave theory"] = "IEEE/OUP Series on Electromagnetic Wave Theory",
  ["ieee/tms j electron mater"] = "IEEE/TMS Journal of Electronic Materials",
  ["ieej j ind appl"] = "IEEJ Journal of Industry Applications",
  ["ieej trans electr electron eng"] = "IEEJ Transactions on Electrical and Electronic Engineering",
  ["ieej trans sens micromachines"] = "IEEJ Transactions on Sensors and Micromachines",
  ["ieice electron express"] = "IEICE Electronics Express",
  ["ieice trans commun"] = "IEICE Transactions on Communications",
  ["ieice trans electron"] = "IEICE Transactions on Electronics",
  ["ieice trans fundam electron commun comput sci"] = "IEICE Transactions on Fundamentals of Electronics, Communications and Computer Science",
  ["ieice trans inf syst"] = "IEICE Transactions on Information and Systems",
  ["ieri procedia"] = "IERI Procedia",
  ["ies j part a: civ struct eng"] = "IES Journal Part A: Civil & Structural Engineering",
  ["iet biom"] = "IET Biometrics",
  ["iet blockchain"] = "IET Blockchain",
  ["iet circuits devices syst"] = "IET Circuits, Devices and Systems",
  ["iet collab intell manuf"] = "IET Collaborative Intelligent Manufacturing",
  ["iet commun"] = "IET Communications",
  ["iet commun eng"] = "IET Communications Engineer",
  ["iet comput digit tech"] = "IET computers & digital techniques",
  ["iet comput digital tech"] = "IET Computers and Digital Techniques",
  ["iet comput vision"] = "IET Computer Vision",
  ["iet control eng ser"] = "IET Control Engineering Series",
  ["iet control theory appl"] = "IET Control Theory and Applications",
  ["iet cyber-phys syst: theor appl"] = "IET Cyber-Physical Systems: Theory & Applications",
  ["iet cyber-syst robot"] = "IET Cyber-Systems and Robotics",
  ["iet electr power appl"] = "IET Electric Power Applications",
  ["iet electr syst transp"] = "IET Electrical Systems in Transportation",
  ["iet electron syst softw"] = "IET Electronics Systems and Software",
  ["iet electron syst software"] = "IET Electronics Systems and Software",
  ["iet energy syst integr"] = "IET Energy Systems Integration",
  ["iet eng manage"] = "IET Engineering Management",
  ["iet gener transm distrib"] = "IET Generation, Transmission and Distribution",
  ["iet image proc"] = "IET Image Processing",
  ["iet image process"] = "IET image processing",
  ["iet inf secur"] = "IET Information Security",
  ["iet intel transport syst"] = "IET Intelligent Transport Systems",
  ["iet manuf"] = "IET Manufacturing",
  ["iet micro nano lett"] = "IET Micro & Nano Letters",
  ["iet microwaves antennas propag"] = "IET Microwaves, Antennas and Propagation",
  ["iet nanobiotechnol"] = "IET nanobiotechnology",
  ["iet nanodielectr"] = "IET Nanodielectrics",
  ["iet networks"] = "IET Networks",
  ["iet optoelectron"] = "IET Optoelectronics",
  ["iet power electron"] = "IET Power Electronics",
  ["iet power energy ser"] = "IET Power and Energy Series",
  ["iet power eng"] = "IET Power Engineer",
  ["iet proc: sci meas technol"] = "IET Proceedings: Science, Measurement & Technology",
  ["iet quantum commun"] = "IET Quantum Communication",
  ["iet radar sonar navig"] = "IET Radar, Sonar and Navigation",
  ["iet renew power gener"] = "IET Renewable Power Generation",
  ["iet renewable power gener"] = "IET Renewable Power Generation",
  ["iet sci meas technol"] = "IET Science, Measurement and Technology",
  ["iet signal proc"] = "IET Signal Processing",
  ["iet signal process"] = "IET Signal Processing",
  ["iet smart cities"] = "IET Smart Cities",
  ["iet smart grid"] = "IET Smart Grid",
  ["iet software"] = "IET Software",
  ["iet synth biol"] = "IET Synthetic Biology",
  ["iet syst biol"] = "IET systems biology",
  ["iet wireless sens syst"] = "IET Wireless Sensor Systems",
  ["iete j educ"] = "IETE Journal of Education",
  ["iete j res"] = "IETE journal of research",
  ["iete tech rev"] = "IETE Technical Review",
  ["ifac j syst control"] = "IFAC Journal of Systems and Control",
  ["ifac proc vol"] = "IFAC Proceedings Volumes",
  ["ifac sympos ser"] = "IFAC Symposia Series",
  ["ifac workshop ser"] = "IFAC Workshop Series",
  ["ifcolog j log appl"] = "The IfCoLog Journal of Logics and their Applications",
  ["ifip adv inf commun technol"] = "IFIP advances in information and communication technology",
  ["ifip congress ser"] = "IFIP Congress Series",
  ["ifip trans"] = "IFIP Transactions",
  ["ifip trans a comput sci tech"] = "IFIP Transactions A: Computer Science and Technology",
  ["ifip trans c comm systems"] = "IFIP Transactions C: Communication Systems",
  ["ifmbe proc"] = "IFMBE proceedings",
  ["ifo beitr quant wirtschaftsforsch"] = "Ifo Beitrage zur Quantitativen Wirtschaftsforschung",
  ["ifo stud"] = "IFO-Studien",
  ["ifo studien"] = "Ifo Studien",
  ["iforest - biogeosc for"] = "iForest - Biogeosciences and Forestry",
  ["ifsr int ser syst sci syst eng"] = "IFSR International Series in Systems Science and Systems Engineering",
  ["ifsr internat ser systems sci engrg"] = "IFSR International Series on Systems Science and Engineering",
  ["ig mod"] = "L'igiene moderna",
  ["ig sanita pubbl"] = "Igiene e sanità pubblica",
  ["ig sanita pubblica collana monogr"] = "Igiene e sanitá pubblica. Collana di monografie",
  ["igaku butsuri"] = "Igaku butsuri : Nihon Igaku Butsuri Gakkai kikanshi = Japanese journal of medical physics : an official journal of Japan Society of Medical Physics",
  ["igaku kenkyu"] = "Igaku kenkyu. Acta medica",
  ["igaku to seibutsugaku"] = "Igaku to seibutsugaku. Medicine and biology",
  ["igakushi kenkyu"] = "[Igakushi kenkyu] [Studies on history of medicine]",
  ["igcc news"] = "IGCC news",
  ["ihag nonjib"] = "Ihag nonjib",
  ["iheringia ser zool"] = "Iheringia. Série zoologia",
  ["ihs prim care provid"] = "The IHS primary care provider",
  ["iiasa proc ser"] = "IIASA Proceedings Series",
  ["iiasa rep"] = "IIASA reports",
  ["iic int rev ind prop copyr law"] = "IIC; international review of industrial property and copyright law",
  ["iie trans"] = "IIE Transactions",
  ["iie trans healthc syst eng"] = "IIE transactions on healthcare systems engineering",
  ["iie trans occup"] = "IIE transactions on occupational ergonomics and human factors",
  ["iie trans occup ergon hum factors"] = "IIE Transactions on Occupational Ergonomics & Human Factors",
  ["iioab j"] = "IIOAB journal",
  ["iips newsl"] = "IIPS newsletter",
  ["iisc lect notes ser"] = "IISc Lecture Notes Series",
  ["iise trans"] = "IISE transactions",
  ["iise trans healthc syst eng"] = "IISE transactions on healthcare systems engineering",
  ["iise trans healthcare syst eng"] = "IISE Transactions on Healthcare Systems Engineering",
  ["iise trans occup ergon hum factors"] = "IISE transactions on occupational ergonomics and human factors",
  ["iium j econ manage"] = "IIUM Journal of Economics and Management",
  ["ijc metab endocr"] = "IJC metabolic & endocrine",
  ["ijcai (u s)"] = "IJCAI : proceedings of the conference",
  ["iju case rep"] = "IJU case reports",
  ["ikushugaku kenkyu"] = "Ikushugaku kenkyu",
  ["il filarete"] = "Il Filarete: Pubblicazioni della Facolta di Lettere e Filosofia dell’Universita degli Studi di Milano",
  ["il pensiero econ italiano"] = "Il Pensiero Economico Italiano",
  ["il politico"] = "Il Politico",
  ["il risparmio"] = "Il Risparmio",
  ["ilar j"] = "ILAR journal",
  ["ilar news"] = "ILAR news",
  ["ill agr econ"] = "Illinois Agricultural Economics",
  ["ill bar j"] = "Illinois bar journal",
  ["ill cancer bull"] = "The Illinois cancer bulletin",
  ["ill classic stud"] = "Illinois classical studies",
  ["ill dent j"] = "Illinois dental journal",
  ["ill health messenger"] = "Illinois health messenger",
  ["ill j math"] = "Illinois Journal of Mathematics",
  ["ill med j"] = "The Illinois medical journal",
  ["ill res"] = "Illinois research",
  ["ill revis statut"] = "Illinois revised statutes. Illinois",
  ["ill revis statut ill ill"] = "Illinois revised statutes / compiled and edited under Smith-Hurd classification. Illinois",
  ["illinois j math"] = "Illinois Journal of Mathematics",
  ["illn crises loss"] = "Illness, crises, and loss",
  ["illum eng"] = "Illuminating engineering",
  ["ilmu kelaut"] = "Ilmu kelautan : Indonesian journal of marine sciences",
  ["ilsa j int law"] = "ILSA journal of international law",
  ["im dienste wald landsch"] = "Im Dienste von Wald und Landschaft",
  ["ima fungus"] = "IMA fungus",
  ["ima j appl math"] = "IMA Journal of Applied Mathematics",
  ["ima j manag math"] = "IMA Journal of Management Mathematics",
  ["ima j manage math"] = "IMA Journal of Management Mathematics",
  ["ima j math appl bus indust"] = "IMA Journal of Mathematics Applied in Business and Industry",
  ["ima j math appl med biol"] = "IMA journal of mathematics applied in medicine and biology",
  ["ima j math control inf"] = "IMA Journal of Mathematical Control and Information",
  ["ima j math control inform"] = "IMA Journal of Mathematical Control and Information",
  ["ima j numer anal"] = "IMA Journal of Numerical Analysis",
  ["ima monogr ser"] = "The Institute of Mathematics & its Applications Monograph Series",
  ["ima vol math appl"] = "The IMA Volumes in Mathematics and its Applications",
  ["image (in)"] = "Image",
  ["image anal mov organ breast thorac images (2018)"] = "Image analysis for moving organ, breast, and thoracic images : third International Workshop, RAMBO 2018, fourth International Workshop, BIA 2018, and first International Workshop, TIA 2018, held in conjunction with MICCAI 2018, Granada, Spain, September 16 and 20, 2018, Proceedings. RAMBO (Workshop) (3rd : 2018 : Granada, Spain)",
  ["image anal stereol"] = "Image Analysis and Stereology",
  ["image j nurs sch"] = "Image--the journal of nursing scholarship",
  ["image vis comput"] = "Image and vision computing",
  ["image vision comput"] = "Image and Vision Computing",
  ["images marquette univ dent reflections"] = "Images; Marquette University dental reflections",
  ["images paediatr cardiol"] = "Images in paediatric cardiology",
  ["imagin cogn pers"] = "Imagination, cognition and personality",
  ["imaging decis (berl)"] = "Imaging decisions (Berlin, Germany)",
  ["imaging med"] = "Imaging in medicine",
  ["imaging neurosci"] = "Imaging Neuroscience",
  ["imaging patient cust simul syst point care ultrasound (2017)"] = "Imaging for patient-customized simulations and systems for point-of-care ultrasound : International Workshops, BIVPCS 2017 and POCUS 2017, held in conjunction with MICCAI 2017, Québec City, QC, Canada, September 14, 2017, Proceedings. BIVPCS (Workshop) (2017 : Québec, Québec)",
  ["imaging sci dent"] = "Imaging science in dentistry",
  ["imaging sci j"] = "Imaging Science Journal",
  ["imago mundi"] = "Imago mundi",
  ["imanagers j nurs"] = "i-Manager's journal on nursing",
  ["ime bull"] = "Bulletin (Institute of Medical Ethics (Great Britain))",
  ["imf staff pap"] = "IMF Staff Papers",
  ["imf surv"] = "IMF survey. International Monetary Fund",
  ["imhotep j afr math pures appl"] = "IMHOTEP. Journal Africain de Mathematiques Pures et Appliquees. African Journal of Pure and Applied Mathematics",
  ["imhotep math proc"] = "Imhotep Mathematical Proceedings",
  ["imj ill med j"] = "IMJ. Illinois medical journal",
  ["immigr minor"] = "Immigrants & minorities",
  ["immun ageing"] = "Immunity & ageing : I & A",
  ["immun infekt"] = "Immunitat und Infektion",
  ["immun inflamm dis"] = "Immunity, inflammation and disease",
  ["immun inflammation dis"] = "Immunity, Inflammation and Disease",
  ["immune netw"] = "Immune network",
  ["immune syst"] = "Immune System",
  ["immunobiol suppl"] = "Immunobiology. Supplement",
  ["immunodefic rev"] = "Immunodeficiency reviews",
  ["immunoendocrinology (houst)"] = "Immunoendocrinology (Houston, Tex.)",
  ["immunol allergy clin north am"] = "Immunology and allergy clinics of North America",
  ["immunol cell biol"] = "Immunology and cell biology",
  ["immunol clin"] = "Immunologia clinica",
  ["immunol commun"] = "Immunological communications",
  ["immunol disord immunother"] = "Immunological disorders & immunotherapy",
  ["immunol endocr metab agents med chem"] = "Immunology, endocrine & metabolic agents in medicinal chemistry",
  ["immunol immunogenet  insights"] = "Immunology and immunogenetics insights",
  ["immunol infect dis"] = "Immunology and infectious diseases",
  ["immunol innov"] = "Immunology innovation",
  ["immunol invest"] = "Immunological investigations",
  ["immunol lett"] = "Immunology letters",
  ["immunol med"] = "Immunological medicine",
  ["immunol res"] = "Immunologic research",
  ["immunol rev"] = "Immunological reviews",
  ["immunol ser"] = "Immunology series",
  ["immunol suppl"] = "Immunology. Supplement",
  ["immunol today"] = "Immunology today",
  ["immunol, endocr metab agents med chem"] = "Immunology, Endocrine & Metabolic Agents in Medicinal Chemistry",
  ["immunome res"] = "Immunome research",
  ["immunopharmacol immunotoxicol"] = "Immunopharmacology and immunotoxicology",
  ["immunotargets ther"] = "ImmunoTargets and therapy",
  ["immunother adv"] = "Immunotherapy Advances",
  ["immunotherapy (los angel)"] = "Immunotherapy (Los Angeles, Calif.)",
  ["impa monogr"] = "IMPA Monographs",
  ["impact  mag"] = "IMPACT magazine",
  ["impact (am coll physicians)"] = "IMpact (American College of Physicians)",
  ["impact assess project appraisal"] = "Impact Assessment and Project Appraisal",
  ["impact comput sci eng"] = "IMPACT of Computing in Science and Engineering",
  ["impact hiv"] = "Impact on HIV",
  ["impact sci soc"] = "Impact of science on society",
  ["impact sci society"] = "Impact of Science on Society",
  ["impan lect notes"] = "IMPAN Lecture Notes",
  ["impfstoffe frankf a m"] = "Arbeiten Aus Dem Paul-Ehrlich-Institut (Bundesamt Fur Sera UndImpfstoffe) Zu Frankfurt A.M. Arb Paul Ehrlich Inst Bundesamt Ser",
  ["implant dent"] = "Implant dentistry",
  ["implant soc"] = "The Implant Society : [periodical]",
  ["implement res pract"] = "Implementation research and practice",
  ["implement sci"] = "Implementation science : IS",
  ["implement sci commun"] = "Implementation science communications",
  ["implementation sci commun"] = "Implementation Science Communications",
  ["implicit relig"] = "Implicit religion : journal of the Centre for the Study of Implicit Religion and Contemporary Spirituality",
  ["important adv oncol"] = "Important advances in oncology",
  ["imprensa medica"] = "Imprensa médica",
  ["impressions (orange)"] = "Impressions",
  ["imprint (n y n y 1976)"] = "Imprint (American Historical Print Collectors Society)",
  ["improv hum performance q"] = "Improving human performance quarterly",
  ["impulse (columbia)"] = "Impulse (Columbia, S.C.)",
  ["impulse (sydney)"] = "Impulse",
  ["ims ind med surg"] = "IMS, Industrial medicine and surgery",
  ["ims lecture notes monogr ser"] = "Institute of Mathematical Statistics Lecture Notes—Monograph Series",
  ["ims rep"] = "IMS Report",
  ["ims textb"] = "Institute of Mathematical Statistics Textbooks",
  ["imsc lect notes math"] = "IMSc Lecture Notes in Mathematics",
  ["in fire ethics"] = "In/fire ethics : newsletter of the International Network of Feminists Interested in Reproductive Health",
  ["in health"] = "In health",
  ["in mind"] = "In mind : the inquisitive mind, social psychology for you",
  ["in point fact"] = "In point of fact",
  ["in pract"] = "In practice",
  ["in silico biol"] = "In silico biology",
  ["in silico cell tissue sci"] = "In silico cell and tissue science",
  ["in silico pharmacol"] = "In silico pharmacology",
  ["in silico plants"] = "In silico plants",
  ["in vitr mol toxicol"] = "In vitro & molecular toxicology",
  ["in vitro"] = "In vitro",
  ["in vitro cell dev biol"] = "In vitro cellular & developmental biology : journal of the Tissue Culture Association",
  ["in vitro cell dev biol anim"] = "In vitro cellular & developmental biology. Animal",
  ["in vitro cell dev biol plant"] = "In vitro cellular & developmental biology. Plant : journal of the Tissue Culture Association",
  ["in vitro models"] = "In vitro models",
  ["in vitro monogr"] = "In vitro. Monograph",
  ["in vitro toxicol"] = "In vitro toxicology",
  ["in vivo"] = "In vivo (Athens, Greece)",
  ["in vivo (brooklyn)"] = "In vivo",
  ["inclusion (wash)"] = "Inclusion (Washington, D.C.)",
  ["ind appl math"] = "Industrial and Applied Mathematics",
  ["ind archaeol rev"] = "Industrial archaeology review",
  ["ind biotechnol"] = "Industrial Biotechnology",
  ["ind biotechnol (new rochelle n y)"] = "Industrial biotechnology (New Rochelle, N.Y.)",
  ["ind chem mater"] = "Industrial Chemistry & Materials",
  ["ind corp change"] = "Industrial and Corporate Change",
  ["ind crops prod"] = "Industrial Crops and Products",
  ["ind crops res j"] = "Industrial crops research journal",
  ["ind diamond rev"] = "Industrial Diamond Review",
  ["ind eng"] = "Industrial engineering (American Institute of Industrial Engineers)",
  ["ind eng chem"] = "Industrial and engineering chemistry",
  ["ind eng chem anal ed"] = "Industrial and Engineering Chemistry, Analytical Edition",
  ["ind eng chem chem eng data series"] = "Industrial & Engineering Chemistry Chemical & Engineering Data Series",
  ["ind eng chem fundam"] = "Industrial and Engineering Chemistry Fundamentals",
  ["ind eng chem process des dev"] = "Industrial & Engineering Chemistry Process Design and Development",
  ["ind eng chem prod res dev"] = "Industrial & Engineering Chemistry Product Research and Development",
  ["ind eng chem res"] = "Industrial & engineering chemistry research",
  ["ind free china"] = "Zi you Zhongguo zhi gong ye",
  ["ind health"] = "Industrial health",
  ["ind health bull"] = "Industrial health bulletin",
  ["ind health care"] = "Industry and health care",
  ["ind health care (cambridge ma)"] = "Industry and health care (Ballinger Publishing Co.)",
  ["ind health mon"] = "Industrial health monthly",
  ["ind health rev"] = "Industrial health review",
  ["ind hyg newsl"] = "Industrial hygiene newsletter",
  ["ind innov"] = "Industry and innovation",
  ["ind lab relat rev"] = "Industrial and Labor Relations Review",
  ["ind labor relat rev"] = "Industrial & labor relations review",
  ["ind lubr tribol"] = "Industrial Lubrication and Tribology",
  ["ind manage"] = "Industrial Management",
  ["ind med gaz"] = "The Indian medical gazette",
  ["ind med surg"] = "Industrial medicine & surgery",
  ["ind metrol"] = "Industrial Metrology",
  ["ind nurs"] = "Industrial nursing",
  ["ind organ"] = "Industrielle Organisation",
  ["ind organ psychol"] = "Industrial and organizational psychology",
  ["ind organ rev"] = "Industrial Organization Review",
  ["ind photogr"] = "Industrial Photography",
  ["ind psychiatry j"] = "Industrial psychiatry journal",
  ["ind relat"] = "Industrial Relations",
  ["ind relat (berkeley)"] = "Industrial relations",
  ["ind rob"] = "The Industrial robot",
  ["ind robot"] = "Industrial Robot",
  ["ind week"] = "Industry week",
  ["indag math (ns)"] = "Koninklijke Nederlandse Akademie van Wetenschappen. Indagationes Mathematicae. New Series",
  ["indagationes math"] = "Indagationes Mathematicae",
  ["indent [engl]"] = "Indent; Journal of International Dentistry; English Edition",
  ["indent engl"] = "Indent; Journal Of International Dentistry; English Edition",
  ["indep pract (lutterworth)"] = "Independent practitioner (Lutterworth, England)",
  ["independent rev"] = "Independent Review",
  ["index fungorum"] = "Index fungorum : published numbers",
  ["india j med sci"] = "The India journal of medical science",
  ["india policy forum"] = "India Policy Forum : [papers]. India Policy Forum. Conference",
  ["india q"] = "India quarterly",
  ["indian arch"] = "Indian archives",
  ["indian chem eng"] = "Indian Chemical Engineer",
  ["indian concr j"] = "Indian Concrete Journal",
  ["indian dent j"] = "The Indian dental journal",
  ["indian dermatol online j"] = "Indian dermatology online journal",
  ["indian drugs"] = "Indian Drugs",
  ["indian econ j"] = "The Indian economic journal : the quarterly journal of the Indian Economic Association",
  ["indian econ rev"] = "Indian economic review",
  ["indian econ soc hist rev"] = "The Indian economic and social history review",
  ["indian fern j"] = "Indian fern journal : international journal of pteridology published by the Indian Fern Society",
  ["indian foundry j"] = "Indian Foundry Journal",
  ["indian geotech j"] = "Indian Geotechnical Journal",
  ["indian heart j"] = "Indian heart journal",
  ["indian heart j teach ser"] = "Indian heart journal: teaching series",
  ["indian hist"] = "The Indian historian",
  ["indian hist rev"] = "The Indian historical review",
  ["indian j  agric biochem"] = "Indian Journal Of Agricultural Biochemistry",
  ["indian j  agric econ"] = "Indian Journal Of Agricultural Economics",
  ["indian j  agric res"] = "Indian Journal Of Agricultural Research",
  ["indian j  agric sci"] = "Indian Journal Of Agricultural Sciences",
  ["indian j agric chem"] = "Indian Journal of Agricultural Chemistry",
  ["indian j agric environ bio-technol"] = "Indian Journal of Agriculture, Environment and Bio-Technology",
  ["indian j agric res"] = "Indian journal of agricultural research",
  ["indian j agron"] = "Indian Journal of Agronomy",
  ["indian j am stud"] = "Indian journal of American studies",
  ["indian j anaesth"] = "Indian journal of anaesthesia",
  ["indian j anim res"] = "Indian journal of animal research",
  ["indian j anim sci"] = "The Indian journal of animal sciences",
  ["indian j appl econ"] = "Indian Journal of Applied Economics",
  ["indian j appl res"] = "Indian journal of applied research",
  ["indian j behav"] = "Indian journal of behaviour",
  ["indian j biochem"] = "Indian Journal of Biochemistry",
  ["indian j biochem biophys"] = "Indian Journal of Biochemistry and Biophysics",
  ["indian j biotechnol"] = "Indian journal of biotechnology",
  ["indian j cancer"] = "Indian journal of cancer",
  ["indian j cardiovasc dis women wincars"] = "Indian journal of cardiovascular disease in women WINCARS",
  ["indian j chem"] = "Indian journal of chemistry",
  ["indian j chem technol"] = "Indian Journal of Chemical Technology",
  ["indian j chem, sect a"] = "Indian Journal of Chemistry Section A: Inorganic, Bio-inorganic, Physical, Theoretical and Analytical Chemistry",
  ["indian j chem, sect a: inorg, bio-inorg, phys, theor"] = "Anal. Chem. Indian Journal of Chemistry, Section A: Inorganic, Bio-inorganic, Physical, Theoretical & Analytical Chemistry",
  ["indian j chem, sect a: inorg, bio-inorg, phys, theor anal chem"] = "Indian Journal of Chemistry, Section A: Inorganic, Bio-inorganic, Physical, Theoretical & Analytical Chemistry",
  ["indian j chem, sect b"] = "Indian Journal of Chemistry Section B: Organic Chemistry including Medicinal Chemistry",
  ["indian j chem, sect b: org chem incl med chem"] = "Indian Journal of Chemistry, Section B: Organic Chemistry Including Medicinal Chemistry",
  ["indian j chest dis"] = "The Indian journal of chest diseases",
  ["indian j chest dis allied sci"] = "The Indian journal of chest diseases & allied sciences",
  ["indian j child health"] = "The Indian journal of child health",
  ["indian j child health (bhopal)"] = "Indian journal of child health",
  ["indian j clin anat physiol"] = "Indian journal of clinical anatomy and physiology",
  ["indian j clin biochem"] = "Indian journal of clinical biochemistry : IJCB",
  ["indian j clin cardiol"] = "Indian Journal of Clinical Cardiology",
  ["indian j clin psychol"] = "Indian journal of clinical psychology",
  ["indian j community health"] = "Indian journal of community health",
  ["indian j community med"] = "Indian journal of community medicine : official publication of Indian Association of Preventive & Social Medicine",
  ["indian j comput sci eng"] = "Indian Journal Of Computer Science And Engineering",
  ["indian j crit care med"] = "Indian journal of critical care medicine : peer-reviewed, official publication of Indian Society of Critical Care Medicine",
  ["indian j dent"] = "Indian journal of dentistry",
  ["indian j dent res"] = "Indian journal of dental research : official publication of Indian Society for Dental Research",
  ["indian j dermatol"] = "Indian journal of dermatology",
  ["indian j dermatol venereol"] = "Indian journal of dermatology and venereology",
  ["indian j dermatol venereol leprol"] = "Indian journal of dermatology, venereology and leprology",
  ["indian j discrete math"] = "Indian Journal of Discrete Mathematics",
  ["indian j ecol"] = "Indian Journal Of Ecology",
  ["indian j econ"] = "Indian journal of economics",
  ["indian j econ dev"] = "Indian journal of economics and development",
  ["indian j endocrinol metab"] = "Indian journal of endocrinology and metabolism",
  ["indian j eng mater sci"] = "Indian Journal of Engineering and Materials Science",
  ["indian j entomol"] = "The Indian journal of entomology",
  ["indian j environ ecoplann"] = "Indian Journal of Environment and Ecoplanning",
  ["indian j environ health"] = "Indian journal of environmental health",
  ["indian j environ prot"] = "Indian Journal of Environmental Protection",
  ["indian j exp biol"] = "Indian journal of experimental biology",
  ["indian j fibre text res"] = "Indian Journal of Fibre and Textile Research",
  ["indian j fish"] = "Indian Journal Of Fisheries",
  ["indian j forensic med pathol"] = "Indian Journal of Forensic Medicine and Pathology",
  ["indian j forensic med toxicol"] = "Indian Journal of Forensic Medicine and Toxicology",
  ["indian j forensic sci"] = "Indian Journal of Forensic Sciences",
  ["indian j gastroenterol"] = "Indian journal of gastroenterology : official journal of the Indian Society of Gastroenterology",
  ["indian j gend stud"] = "Indian journal of gender studies",
  ["indian j gender stud"] = "Indian Journal of Gender Studies",
  ["indian j genet plant breed"] = "The Indian journal of genetics & plant breeding : official publication of the Indian Society of Genetics & Plant Breeding",
  ["indian j geo mar sci"] = "Indian Journal Of Geo Marine Sciences",
  ["indian j geochem"] = "Indian Journal of Geochemistry",
  ["indian j geomarine sci"] = "Indian journal of (geo) marine sciences",
  ["indian j geosci"] = "Indian Journal Of Geosciences",
  ["indian j gynecol oncol"] = "Indian journal of gynecologic oncology",
  ["indian j helminthol"] = "Indian journal of helminthology",
  ["indian j hematol blood transfus"] = "Indian journal of hematology & blood transfusion : an official journal of Indian Society of Hematology and Blood Transfusion",
  ["indian j heterocycl chem"] = "Indian Journal of Heterocyclic Chemistry",
  ["indian j hist med"] = "Indian journal of the history of medicine",
  ["indian j hist sci"] = "Indian Journal of History of Science",
  ["indian j hortic"] = "The Indian journal of horticulture",
  ["indian j hosp pharm"] = "The Indian journal of hospital pharmacy",
  ["indian j hum genet"] = "Indian journal of human genetics",
  ["indian j ind appl math"] = "Indian Journal of Industrial and Applied Mathematics",
  ["indian j ind relat"] = "Indian journal of industrial relations",
  ["indian j lab econ"] = "Indian Journal of Labour Economics",
  ["indian j labour econ"] = "The Indian journal of labour economics : the quarterly journal of the Indian Society of Labour Economics",
  ["indian j lepr"] = "Indian journal of leprosy",
  ["indian j malariol"] = "Indian journal of malariology",
  ["indian j mar sci"] = "Indian journal of marine sciences",
  ["indian j matern child health"] = "Indian journal of maternal and child health : official publication of Indian Maternal and Child Health Association",
  ["indian j math"] = "Indian Journal of Mathematics",
  ["indian j med educ"] = "Indian journal of medical education",
  ["indian j med ethics"] = "Indian journal of medical ethics",
  ["indian j med inform"] = "Indian journal of medical informatics",
  ["indian j med microbiol"] = "Indian journal of medical microbiology",
  ["indian j med paediatr oncol"] = "Indian journal of medical and paediatric oncology : official journal of Indian Society of Medical & Paediatric Oncology",
  ["indian j med res"] = "The Indian journal of medical research",
  ["indian j med sci"] = "Indian journal of medical sciences",
  ["indian j med surg"] = "The Indian journal of medicine & surgery",
  ["indian j microbiol"] = "Indian journal of microbiology",
  ["indian j nat prod"] = "Indian Journal of Natural Products",
  ["indian j nat prod resour"] = "Indian journal of natural products and resources",
  ["indian j nematol"] = "Indian journal of nematology : official publication of the Nematological Society of India",
  ["indian j nephrol"] = "Indian journal of nephrology",
  ["indian j nucl med"] = "Indian journal of nuclear medicine : IJNM : the official journal of the Society of Nuclear Medicine, India",
  ["indian j nutr diet"] = "The Indian journal of nutrition and dietetics",
  ["indian j occup environ med"] = "Indian journal of occupational and environmental medicine",
  ["indian j ophthalmol"] = "Indian journal of ophthalmology",
  ["indian j orthop"] = "Indian journal of orthopaedics",
  ["indian j otolaryngol"] = "Indian journal of otolaryngology",
  ["indian j otolaryngol head neck surg"] = "Indian journal of otolaryngology and head and neck surgery : official publication of the Association of Otolaryngologists of India",
  ["indian j palliat care"] = "Indian journal of palliative care",
  ["indian j pathol bacteriol"] = "Indian journal of pathology & bacteriology",
  ["indian j pathol microbiol"] = "Indian journal of pathology & microbiology",
  ["indian j pediatr"] = "Indian journal of pediatrics",
  ["indian j pharm"] = "The Indian journal of pharmacy",
  ["indian j pharm educ res"] = "Indian Journal Of Pharmaceutical Education And Research",
  ["indian j pharm sci"] = "Indian journal of pharmaceutical sciences",
  ["indian j pharmacol"] = "Indian journal of pharmacology",
  ["indian j phys"] = "Indian Journal of Physics",
  ["indian j phys anthropol hum genet"] = "Indian journal of physical anthropology and human genetics",
  ["indian j phys proc indian assoc cultiv sci (2004)"] = "Indian journal of physics and proceedings of the Indian Association for the Cultivation of Science (2004)",
  ["indian j phys, a"] = "Indian Journal of Physics, A",
  ["indian j phys, b"] = "Indian Journal of Physics, B",
  ["indian j physiol pharmacol"] = "Indian journal of physiology and pharmacology",
  ["indian j physiother occup ther"] = "Indian journal of physiotherapy and occupational therapy",
  ["indian j plant physiol"] = "Indian journal of plant physiology",
  ["indian j plant prot"] = "Indian journal of plant protection",
  ["indian j plast surg"] = "Indian journal of plastic surgery : official publication of the Association of Plastic Surgeons of India",
  ["indian j prev soc med"] = "Indian journal of preventive and social medicine",
  ["indian j psychiatry"] = "Indian journal of psychiatry",
  ["indian j psychol"] = "Indian journal of psychology",
  ["indian j psychol med"] = "Indian journal of psychological medicine",
  ["indian j public adm"] = "The Indian journal of public administration : quarterly journal of the Indian Institute of Public Administration",
  ["indian j public health"] = "Indian journal of public health",
  ["indian j public health res dev"] = "Indian journal of public health research & development",
  ["indian j pure appl math"] = "Indian Journal of Pure and Applied Mathematics",
  ["indian j pure appl phys"] = "Indian Journal of Pure and Applied Physics",
  ["indian j quant econ"] = "Indian journal of quantitative economics",
  ["indian j radio space phys"] = "Indian Journal of Radio & Space Physics",
  ["indian j radiol imaging"] = "The Indian journal of radiology & imaging",
  ["indian j rheumatol"] = "Indian journal of rheumatology",
  ["indian j sci"] = "Indian journal of science",
  ["indian j sci res"] = "Indian journal of scientific research",
  ["indian j sci technol"] = "Indian journal of science and technology",
  ["indian j sex transm dis"] = "Indian journal of sexually transmitted diseases",
  ["indian j sex transm dis aids"] = "Indian journal of sexually transmitted diseases and AIDS",
  ["indian j sleep med"] = "Indian journal of sleep medicine",
  ["indian j small ruminants"] = "Indian Journal Of Small Ruminants",
  ["indian j soc psychiatry"] = "Indian journal of social psychiatry",
  ["indian j soc res"] = "Indian journal of social research",
  ["indian j soc sci"] = "Indian journal of social science",
  ["indian j soc work"] = "The Indian journal of social work",
  ["indian j surg"] = "The Indian journal of surgery",
  ["indian j surg oncol"] = "Indian journal of surgical oncology",
  ["indian j technol"] = "Indian Journal of Technology",
  ["indian j thorac cardiovasc surg"] = "Indian journal of thoracic and cardiovascular surgery",
  ["indian j tradit knowl"] = "Indian Journal Of Traditional Knowledge",
  ["indian j tuberc"] = "The Indian journal of tuberculosis",
  ["indian j urol"] = "Indian journal of urology : IJU : journal of the Urological Society of India",
  ["indian j vener dis dermatol"] = "Indian journal of venereal diseases and dermatology",
  ["indian j vet pathol"] = "Indian journal of veterinary pathology",
  ["indian j virol"] = "Indian journal of virology : an official organ of Indian Virological Society",
  ["indian med j"] = "Indian medical journal",
  ["indian med rec"] = "Indian medical record",
  ["indian med trib"] = "Indian medical tribune",
  ["indian obstet gynaecol"] = "Indian obstetrics & gynaecology",
  ["indian pacing electrophysiol j"] = "Indian pacing and electrophysiology journal",
  ["indian pediatr"] = "Indian pediatrics",
  ["indian physician"] = "The Indian physician",
  ["indian phytopathol"] = "Indian phytopathology",
  ["indian pol sci rev"] = "The Indian political science review",
  ["indian pract"] = "The Indian practitioner",
  ["indian psychol rev"] = "Indian psychological review",
  ["indian sci cruiser"] = "Indian Science Cruiser",
  ["indian sociol bull"] = "Indian sociological bulletin",
  ["indian stat inst ser"] = "Indian Statistical Institute Series",
  ["indian vet j"] = "The Indian veterinary journal",
  ["indiana bus rev"] = "Indiana Business Review",
  ["indiana code indiana"] = "Indiana code : comprising all laws of a general and permanent nature including statutes enacted through the 1982 regular session of the Indiana General Assembly. Indiana",
  ["indiana folkl"] = "Indiana folklore",
  ["indiana health law rev"] = "Indiana health law review",
  ["indiana int comp law rev"] = "Indiana international & comparative law review",
  ["indiana law j"] = "Indiana law journal (Indianapolis, Ind. : 1926)",
  ["indiana law rev"] = "Indiana law review",
  ["indiana mag hist"] = "Indiana magazine of history",
  ["indiana med"] = "Indiana medicine : the journal of the Indiana State Medical Association",
  ["indiana med hist q"] = "Indiana medical history quarterly",
  ["indiana nurse"] = "The Indiana nurse",
  ["indiana pharm"] = "The Indiana pharmacist",
  ["indiana soc stud q"] = "Indiana social studies quarterly",
  ["indiana univ j undergrad res"] = "Indiana University journal of undergraduate research",
  ["indiana univ math j"] = "Indiana University Mathematics Journal",
  ["indices naturwiss med period 1850"] = "Indices naturwissenschaftlich-medizinischer Periodica bis 1850",
  ["individ differ res"] = "Individual differences research : IDR",
  ["individ psychol"] = "Individual psychology",
  ["individ psychol bull"] = "Individual psychology bulletin",
  ["indo iran j"] = "Indo-Iranian journal",
  ["indo koten kenkyu"] = "Indo koten kenkyū = Acta Indologica",
  ["indones fish res j"] = "Indonesian fisheries research journal",
  ["indones j biotechnol"] = "Indonesian journal of biotechnology",
  ["indones j chem"] = "Indonesian Journal of Chemistry",
  ["indones j geogr"] = "The Indonesian journal of geography",
  ["indones malay world"] = "Indonesia and the Malay world",
  ["indones q"] = "The Indonesian quarterly",
  ["indoor air"] = "Indoor air",
  ["indoor built environ"] = "Indoor and Built Environment",
  ["inductive log program"] = "Inductive logic programming. ILP",
  ["indust math"] = "Industrial Mathematics",
  ["industrial microbiology|gong ye wei sheng wu"] = "Gong ye wei sheng wu",
  ["industry devel"] = "Industry and Development",
  ["industry innovation"] = "Industry and Innovation",
  ["inf & ing vorles inform ingenieurstud"] = "Inf & Ing: Vorlesungen zum Informatik- und Ingenieurstudium",
  ["inf agrar"] = "L'Informatore agrario",
  ["inf behav"] = "Information and behavior",
  ["inf bot ital"] = "Informatore botanico italiano",
  ["inf commun soc"] = "Information, communication and society",
  ["inf comput"] = "Information and computation",
  ["inf contemp soc (2019)"] = "Information in Contemporary Society : 14th International Conference, iConference 2019, Washington, DC, USA, March 31-April 3, 2019, Proceedings. iConference (Conference) (14th : 2019 : Washington, D.C.)",
  ["inf demogr"] = "Informe demográfico (São Paulo (Brazil : State). Fundação Sistema Estadual de Análise de Dados)",
  ["inf dent"] = "L' Information dentaire",
  ["inf discov deliv"] = "Information discovery and delivery",
  ["inf disp (1975)"] = "Information display",
  ["inf fitopatol"] = "Informatore fitopatologico",
  ["inf fusion"] = "An international journal on information fusion",
  ["inf fusion data sci"] = "Information Fusion and Data Science",
  ["inf geogr"] = "L'Information géographique",
  ["inf geom"] = "Information Geometry",
  ["inf health social care"] = "Informatics for Health and Social Care",
  ["inf hist"] = "L'Information historique",
  ["inf hist art"] = "L'Information d'histoire de l'art",
  ["inf inference"] = "Information and inference : a journal of the IMA",
  ["inf inference j ima"] = "Information and Inference: A Journal of the IMA",
  ["inf knowl syst manage"] = "Information, knowledge, systems management",
  ["inf litt"] = "L'Information littéraire",
  ["inf manage"] = "Information management (PTN Publishing Corporation)",
  ["inf med"] = "L'Informatore medico",
  ["inf med paramed (montreal)"] = "L' Information médicale et paramédicale",
  ["inf med unlocked"] = "Informatics in Medicine Unlocked",
  ["inf medicas"] = "Informaciones médicas",
  ["inf natschutz landschpfl nordwestdtschl"] = "Informationen zu Naturschutz und Landschaftspflege in Nordwestdeutschland",
  ["inf odontostomatol"] = "Informatore Odonto-Stomatologico",
  ["inf organ"] = "Information and Organization",
  ["inf orthod kieferorthop"] = "Informationen aus Orthodontie und Kieferorthopädie : mit Beiträgen aus der internationalen Literatur",
  ["inf priv"] = "Information privacy",
  ["inf process comput assist interv (2012)"] = "Information processing in computer-assisted interventions : third International Conference, IPCAI 2012, Pisa, Italy, June 27, 2012 proceedings. IPCAI (Conference) (3rd : 2012 : Pisa, Italy)",
  ["inf process lett"] = "Information processing letters",
  ["inf process manag"] = "Information processing & management",
  ["inf process manage"] = "Information Processing & Management",
  ["inf prof"] = "Information Professional",
  ["inf psiquiatr"] = "Informação psiquiátrica : IP : órgão oficial do Serviço de Psiquiatria e Psicologia Médica do Hospital de Clínicas da UERJ",
  ["inf psychiatr"] = "L'Information psychiatrique",
  ["inf raumentwickl"] = "Informationen zur Raumentwicklung",
  ["inf rep north for res cent"] = "Information Report Northern Forest Research Centre",
  ["inf retr boston"] = "Information retrieval",
  ["inf retr j"] = "Information Retrieval Journal",
  ["inf retr ser"] = "The Information Retrieval Series",
  ["inf reuse integr iri ieee int conf"] = "Information Reuse and Integration (IRI), IEEE International Conference on",
  ["inf sci"] = "Informing science",
  ["inf sci (n y)"] = "Information sciences",
  ["inf sci (ny)"] = "Information sciences",
  ["inf sci it educ conf"] = "Informing Science and IT Education Conference : proceedings and schedule. Informing Science and IT Education Conference",
  ["inf secur comput fraud"] = "Information security and computer fraud",
  ["inf secur cryptography"] = "Information Security and Cryptography",
  ["inf secur j: global perspect"] = "Information Security Journal: A Global Perspective",
  ["inf serv use"] = "Information services & use",
  ["inf sobre enferm vener"] = "Información sobre enfermedades venéreas",
  ["inf softw technol"] = "Information and software technology",
  ["inf software technol"] = "Information and Software Technology",
  ["inf spektrum"] = "Informatik Spektrum",
  ["inf summ"] = "Information summary (International Clearinghouse on Adolescent Fertility)",
  ["inf syst"] = "Information systems",
  ["inf syst front"] = "Information systems frontiers : a journal of research and innovation",
  ["inf syst j"] = "Information Systems Journal",
  ["inf syst manage"] = "Information Systems Management",
  ["inf syst res"] = "Information systems research : ISR",
  ["inf technol dev"] = "Information technology for development",
  ["inf technol manage"] = "Information Technology and Management",
  ["inf vis"] = "Information visualization",
  ["inf visualization"] = "Information Visualization",
  ["inf zür wald"] = "Information Zürcher Wald",
  ["inf, commun soc"] = "Information, Communication & Society",
  ["infant behav dev"] = "Infant behavior & development",
  ["infant child adolesc nutr"] = "Infant, child & adolescent nutrition",
  ["infant child dev"] = "Infant and child development",
  ["infant ment health j"] = "Infant mental health journal",
  ["infants young child"] = "Infants and young children",
  ["infanz anorm"] = "Infanzia anormale",
  ["infber  bayer  landesamtes  wasserwirtsch"] = "Informationsberichte des Bayerischen Landesamtes für Wasserwirtschaft",
  ["infbl forschbereiches landsch"] = "Informationsblatt des Forschungsbereiches Landschaft",
  ["infbl forschbereiches landschökol"] = "Informationsblatt des Forschungsbereiches Landschaftsökologie",
  ["infdienst natschutz niedersachs"] = "Informationsdienst Naturschutz Niedersachsen",
  ["infect agent cancer"] = "Infectious agents and cancer",
  ["infect agents dis"] = "Infectious agents and disease",
  ["infect chemother"] = "Infection & chemotherapy",
  ["infect control"] = "Infection control : IC",
  ["infect control can"] = "Infection control Canada",
  ["infect control dig"] = "Infection control digest",
  ["infect control hosp epidemiol"] = "Infection control and hospital epidemiology",
  ["infect control rounds"] = "Infection control rounds",
  ["infect control urol care"] = "Infection control & urological care",
  ["infect control wkly"] = "Infection control weekly",
  ["infect dis"] = "Infectious diseases",
  ["infect dis (auckl)"] = "Infectious diseases",
  ["infect dis (lond)"] = "Infectious diseases (London, England)",
  ["infect dis clin north am"] = "Infectious disease clinics of North America",
  ["infect dis clin pract (baltim md)"] = "Infectious diseases in clinical practice (Baltimore, Md.)",
  ["infect dis diagn treat"] = "Infectious diseases diagnosis & treatment",
  ["infect dis health"] = "Infection, disease & health",
  ["infect dis j"] = "Infectious diseases journal : IDJ",
  ["infect dis model"] = "Infectious Disease Modelling",
  ["infect dis newsl (n y)"] = "Infectious diseases newsletter (New York, N.Y.)",
  ["infect dis now"] = "Infectious diseases now",
  ["infect dis obstet gynecol"] = "Infectious diseases in obstetrics and gynecology",
  ["infect dis poverty"] = "Infectious diseases of poverty",
  ["infect dis rep"] = "Infectious disease reports",
  ["infect dis ther"] = "Infectious diseases and therapy",
  ["infect disord drug targets"] = "Infectious disorders drug targets",
  ["infect disord: drug targets"] = "Infectious Disorders: Drug Targets",
  ["infect drug resist"] = "Infection and drug resistance",
  ["infect ecol epidemiol"] = "Infection ecology & epidemiology",
  ["infect epidemiol med"] = "Infection epidemiology and medicine",
  ["infect epidemiol microbiol"] = "Infection, epidemiology and microbiology",
  ["infect genet evol"] = "Infection, genetics and evolution : journal of molecular epidemiology and evolutionary genetics in infectious diseases",
  ["infect immun"] = "Infection and immunity",
  ["infect immun pharmacol"] = "Infection, immunity, and pharmacology",
  ["infect med"] = "Infections in medicine",
  ["infect microbes dis"] = "Infectious Microbes & Diseases",
  ["infect, genet evol"] = "Infection, Genetics and Evolution",
  ["infectious agents surveillance report (iasr)|byogen biseibutsu kenshutsu joho geppo"] = "Byogen biseibutsu kenshutsu joho, geppo",
  ["infektsiia immun"] = "Infekt︠s︡ii︠a︡ i immunitet",
  ["infez med"] = "Le infezioni in medicina",
  ["infin dimens anal quantum probab relat top"] = "Infinite Dimensional Analysis, Quantum Probability and Related Topics",
  ["infin dimension anal quantum probab relat top"] = "Infinite Dimensional Analysis Quantum Probability and Related Topics",
  ["infirm aux"] = "L'Infirmière auxiliaire : revue de la Corporation professionnelle des infirmières et infirmiers auxiliaires du Québec",
  ["infirm can"] = "L' Infirmière canadienne",
  ["infirm fr"] = "L'Infirmière francaise",
  ["infirm haiti"] = "L' Infirmiere haïtienne",
  ["infirm que"] = "L'Infirmière du Québec : revue officielle de l'Ordre des infirmières et infirmiers du Québec",
  ["inflamm allergy drug targets"] = "Inflammation & allergy drug targets",
  ["inflamm allergy: drug targets"] = "Inflammation & Allergy: Drug Targets",
  ["inflamm bowel dis"] = "Inflammatory bowel diseases",
  ["inflamm cell signal"] = "Inflammation and cell signaling",
  ["inflamm intest dis"] = "Inflammatory intestinal diseases",
  ["inflamm regen"] = "Inflammation and regeneration",
  ["inflamm res"] = "Inflammation Research",
  ["influenza other respir viruses"] = "Influenza and other respiratory viruses",
  ["influenza res treat"] = "Influenza research and treatment",
  ["info comercial española revista econ"] = "Informacion Comercial Española Revista de Economia",
  ["info econ pol"] = "Information Economics and Policy",
  ["info trends"] = "Info trends : medicine, law & ethics",
  ["infor inf syst oper res"] = "INFOR. Information Systems and Operational Research",
  ["infor: inf syst oper res"] = "INFOR: Information Systems and Operational Research",
  ["inform (champaign)"] = "Inform (Champaign, Ill.)",
  ["inform and comput"] = "Information and Computation",
  ["inform and control (shenyang)"] = "Information and Control",
  ["inform bull"] = "Information Bulletin",
  ["inform health soc care"] = "Informatics for health & social care",
  ["inform med unlocked"] = "Informatics in medicine unlocked",
  ["inform prim care"] = "Informatics in primary care",
  ["inform process lett"] = "Information Processing Letters",
  ["inform sci"] = "Information Sciences. An International Journal",
  ["inform secur appl (2013)"] = "Information security applications : 14th International Workshop, WISA 2013, Jeju Island, Korea, August 19-21, 2013 : revised selected papers. WISA (Workshop : Information security) (14th : 2013 : Cheju-do, Korea)",
  ["informatica (ljubl)"] = "Informatica. An International Journal of Computing and Informatics",
  ["informatica (vilnius)"] = "Lithuanian Academy of Sciences",
  ["informatics (mdpi)"] = "Informatics (MDPI)",
  ["information (basel)"] = "Information (Basel)",
  ["informationstech datenverarb"] = "Informationstechnik und Datenverarbeitung",
  ["informs j comput"] = "INFORMS Journal on Computing",
  ["informs j optim"] = "INFORMS Journal on Optimization",
  ["infort traumatol lav"] = "Infortunistica e traumatologia del lavoro",
  ["infosys sci found ser"] = "Infosys Science Foundation Series",
  ["infosys sci found ser math sci"] = "Infosys Science Foundation Series in Mathematical Sciences",
  ["infrared phys"] = "Infrared Physics",
  ["infrared phys technol"] = "Infrared Physics and Technology",
  ["infrastruct complexity"] = "Infrastructure Complexity",
  ["infrastructures (basel)"] = "Infrastructures",
  ["infusionsther klin ernahr"] = "Infusionstherapie und klinische Ernährung",
  ["infusionsther klin ernahr sonderh"] = "Infusionstherapie und klinische Ernährung. Sonderheft",
  ["infusionsther transfusionsmed"] = "Infusionstherapie und Transfusionsmedizin",
  ["ing agua"] = "IngenierÃ­a del agua",
  ["ing arch"] = "Ingenieur Archiv",
  ["ing cienc"] = "Ingeniería y ciencia",
  ["ing compet"] = "Ingenieria y Competitividad",
  ["ing invest tecnol"] = "IngenierÃ­a, InvestigaciÃ³n y TecnologÃ­a",
  ["ing quim"] = "Ingeniera Quimica (Madrid)",
  ["ing rech biomed"] = "Ingénierie et recherche biomédicale : IRBM = Biomedical engineering and research",
  ["ingan kwahak"] = "In'gan kwahak. Human science",
  ["ingbiol mittbl"] = "Ingenieurbiologie Mitteilungsblatt",
  ["ingu munje nonjip"] = "In'gu munje nonjip = Journal of population studies",
  ["ingu pogon nonjip"] = "In'gu pogŏn nonjip = Journal of population and health studies",
  ["inhal toxicol"] = "Inhalation toxicology",
  ["inhalation toxicol"] = "Inhalation Toxicology",
  ["inhaled part"] = "Inhaled particles",
  ["initiat reprod health policy"] = "Initiatives in reproductive health policy",
  ["initiatives popul"] = "Initiatives in population",
  ["inj control saf promot"] = "Injury control and safety promotion",
  ["inj epidemiol"] = "Injury epidemiology",
  ["inj med"] = "Shang hai yi xue = Injury medicine",
  ["inj prev"] = "Injury prevention : journal of the International Society for Child and Adolescent Injury Prevention",
  ["inland seas"] = "Inland seas",
  ["inland water biol"] = "Inland water biology",
  ["inland waters"] = "Inland waters : journal of the International Society of Limnology",
  ["inn med"] = "Innere Medizin",
  ["innate immun"] = "Innate immunity",
  ["innov aging"] = "Innovation in aging",
  ["innov appl math"] = "Innovations in Applied Mathematics",
  ["innov clin neurosci"] = "Innovations in clinical neuroscience",
  ["innov entrep health"] = "Innovation and entrepreneurship in health",
  ["innov food sci emerg technol"] = "Innovative food science & emerging technologies : IFSET : the official scientific journal of the European Federation of Food Science and Technology",
  ["innov graph theory"] = "Innovations in Graph Theory",
  ["innov high educ"] = "Innovative higher education",
  ["innov incidence geom"] = "Innovations in Incidence Geometry. Algebraic, Topological and Combinatorial",
  ["innov j"] = "The innovation journal : the public sector innovation journal",
  ["innov oncol nurs"] = "Innovations in oncology nursing",
  ["innov pharm"] = "Innovations in pharmacy",
  ["innov res clin serv community support rehabil"] = "Innovation and research in clinical services, community support, and rehabilitation",
  ["innov rom food biotechnol"] = "Innovative Romanian food biotechnology",
  ["innov surg sci"] = "Innovative surgical sciences",
  ["innov syst softw eng"] = "Innovations in systems and software engineering",
  ["innovate (north miami beach)"] = "Innovate (North Miami Beach, Fla.)",
  ["innovation (abingdon)"] = "Innovation (Abingdon, England)",
  ["innovation (n y)"] = "Innovation (New York, N.Y.)",
  ["innovation (north syd)"] = "Innovation (North Sydney, N.S.W.)",
  ["innovation emerging technol"] = "Innovation and Emerging Technologies",
  ["innovation green dev"] = "Innovation and Green Development",
  ["innovations (phila)"] = "Innovations (Philadelphia, Pa.)",
  ["innovations incidence geom- algebraic topol comb"] = "Innovations in Incidence Geometry â€” Algebraic, Topological and Combinatorial",
  ["innovations phila pa"] = "Innovations (Philadelphia, Pa.)",
  ["innovations phys: x"] = "Innovations in Physics: X",
  ["innovations syst software eng"] = "Innovations in Systems and Software Engineering",
  ["innovative biosyst bioeng"] = "Innovative Biosystems and Bioengineering",
  ["innovative food sci emerg technol"] = "Innovative Food Science & Emerging Technologies",
  ["innovative infrastruct solutions"] = "Innovative Infrastructure Solutions",
  ["inorg chem"] = "Inorganic chemistry",
  ["inorg chem (washington, dc, u s)"] = "Inorganic Chemistry (Washington, DC, United States)",
  ["inorg chem commun"] = "Inorganic Chemistry Communications",
  ["inorg chem front"] = "Inorganic Chemistry Frontiers",
  ["inorg chim acta"] = "Inorganica Chimica Acta",
  ["inorg chim acta rev"] = "Inorganica Chimica Acta Reviews",
  ["inorg mater"] = "Inorganic Materials",
  ["inorg mater (ussr)"] = "Inorganic Materials (USSR) [translation of Izvestiya Akademii Nauk SSSR, Neorganicheskie Materialy]",
  ["inorg mater appl res"] = "Inorganic Materials: Applied Research",
  ["inorg nano-metal chem"] = "Inorganic and Nano-Metal Chemistry",
  ["inorg nucl chem lett"] = "Inorganic and Nuclear Chemistry Letters",
  ["inorg react mech"] = "Inorganic Reaction Mechanisms",
  ["inorg react mech (philadelphia, pa, u s)"] = "Inorganic Reaction Mechanisms (Philadelphia, PA, United States)",
  ["inorg synth"] = "Inorganic Syntheses",
  ["inorganica chim acta"] = "Inorganica chimica acta",
  ["inorganics (basel)"] = "Inorganics",
  ["inquiry (oslo)"] = "Inquiry (Oslo, Norway)",
  ["ins rep"] = "INS reporter",
  ["ins: mathematics econ"] = "Insurance: Mathematics and Economics",
  ["insect biochem"] = "Insect biochemistry",
  ["insect biochem mol biol"] = "Insect Biochemistry and Molecular Biology",
  ["insect conserv divers"] = "Insect conservation and diversity",
  ["insect conserv diversity"] = "Insect Conservation and Diversity",
  ["insect mol biol"] = "Insect Molecular Biology",
  ["insect sci"] = "Insect science",
  ["insect syst divers"] = "Insect systematics and diversity",
  ["insect syst diversity"] = "Insect Systematics and Diversity",
  ["insect syst evol"] = "Insect systematics & evolution",
  ["insecta matsumurana"] = "Insecta matsumurana",
  ["insecta mundi"] = "Insecta mundi",
  ["insectes soc"] = "Insectes Sociaux",
  ["insectes sociaux"] = "Insectes sociaux",
  ["inside cell"] = "Inside the cell",
  ["insight (lawrence)"] = "Insight (Lawrence, Kan.)",
  ["insight bioinforma"] = "Insight bioinformatics",
  ["insight j"] = "The insight journal",
  ["insights allergy asthma bronchitis"] = "Insights in allergy, asthma & bronchitis",
  ["insights biomed"] = "Insights in biomedicine",
  ["insights biomed res"] = "Insights of biomedical research",
  ["insights cancer res"] = "Insights in cancer research",
  ["insights enzym res"] = "Insights in enzyme research",
  ["insights imaging"] = "Insights into imaging",
  ["insights into imaging"] = "Insights into Imaging",
  ["insights neurosurg"] = "Insights in neurosurgery",
  ["insights nutr metab"] = "Insights in nutrition and metabolism",
  ["inst angew wirtschaftsforsch tubingen schriftenreihe"] = "Institut fur Angewandte Wirtschaftsforschung Tubingen Schriftenreihe",
  ["inst child explor beyond"] = "Institutionalised children exploration and beyond",
  ["inst civ eng constr law q"] = "Institution of Civil Engineers Construction Law Quarterly",
  ["inst hautes études sci publ math"] = "Institut des Hautes études Scientifiques",
  ["inst int pedagog report"] = "Institutionen für Internationell Pedagogik",
  ["inst math appl conf ser new ser"] = "The Institute of Mathematics and its Applications Conference Series",
  ["inst math sci lect notes"] = "Institute of Mathematical Sciences Lecture Notes",
  ["inst math stat (ims) collect"] = "Institute of Mathematical Statistics (IMS) Collections",
  ["inst math stat collect"] = "Institute of Mathematical Statistics collections",
  ["inst math stat monogr"] = "Institute of Mathematical Statistics (IMS) Monographs",
  ["inst nonlinear sci"] = "Institute for Nonlinear Science",
  ["inst transp eng ite j"] = "Institute of Transportation Engineers. ITE Journal",
  ["inst vol feed"] = "Institutions/volume feeding",
  ["inst élie cartan"] = "Institut Élie Cartan",
  ["instr course lect"] = "Instructional course lectures",
  ["instr sci"] = "Instructional science",
  ["instraw news"] = "INSTRAW news : women and development",
  ["instrum exp tech"] = "Instruments and Experiment Techniques",
  ["instrum exp tech (ussr)"] = "Instruments and Experimental Techniques (USSR) [translation of Pribory i Tekhnika Eksperimenta]",
  ["instrum sci technol"] = "Instrumentation Science and Technology",
  ["insur couns j"] = "Insurance counsel journal",
  ["insur law j"] = "Insurance law journal (Chicago, Ill.)",
  ["insur math econ"] = "Insurance, mathematics & economics",
  ["insurance math econom"] = "Insurance: Mathematics & Economics",
  ["int  j interferon cytokine mediat res"] = "International journal of interferon, cytokine and mediator research",
  ["int  j mater sci"] = "International journal of material science",
  ["int abstr surg"] = "International abstracts of surgery",
  ["int adv surg oncol"] = "International advances in surgical oncology",
  ["int advances econ res"] = "International Advances in Economic Research",
  ["int aff"] = "International affairs",
  ["int agrophys"] = "International Agrophysics",
  ["int anesthesiol clin"] = "International anesthesiology clinics",
  ["int angiol"] = "International angiology : a journal of the International Union of Angiology",
  ["int annu conf ict"] = "International Annual Conference of ICT",
  ["int appl mech"] = "International Applied Mechanics",
  ["int aquat res"] = "International aquatic research",
  ["int aquat res,"] = "International Aquatic Research",
  ["int arab j antimicrob agents"] = "The international Arabic journal of antimicrobial agents",
  ["int arch allergy appl immunol"] = "International archives of allergy and applied immunology",
  ["int arch allergy immunol"] = "International archives of allergy and immunology",
  ["int arch arbeitsmed"] = "Internationales Archiv fur Arbeitsmedizin",
  ["int arch gewerbepathol gewerbehyg"] = "Internationales Archiv für Gewerbepathologie und Gewerbehygiene",
  ["int arch hist ideas"] = "International Archives of the History of Ideas/Archives internationales d’histoire des idees",
  ["int arch med"] = "International archives of medicine",
  ["int arch nurs health care"] = "International archives of nursing and health care",
  ["int arch occup environ health"] = "International archives of occupational and environmental health",
  ["int arch otorhinolaryngol"] = "International archives of otorhinolaryngology",
  ["int arch photogramm remote sens"] = "International Archives of Photogrammetry and Remote Sensing",
  ["int arch transl med"] = "International archives of translational medicine",
  ["int asianforum"] = "Internationales Asienforum",
  ["int assoc math geosci stud math geosci"] = "International Association for Mathematical Geosciences. Studies in Mathematical Geosciences",
  ["int astron union colloq"] = "International Astronomical Union Colloquium",
  ["int bibliogr hist sci"] = "International bibliography of historical sciences",
  ["int biodeterior biodegrad"] = "International Biodeterioration & Biodegradation",
  ["int biodeterior biodegradation"] = "International biodeterioration & biodegradation",
  ["int biol biomed j"] = "International biological and biomedical journal",
  ["int biol rev"] = "International biology review",
  ["int biomech"] = "International biomechanics",
  ["int braz j urol"] = "International braz j urol : official journal of the Brazilian Society of Urology",
  ["int breastfeed j"] = "International breastfeeding journal",
  ["int bus rev"] = "International business review (Oxford, England)",
  ["int cancer conf j"] = "International cancer conference journal",
  ["int cardiovasc res j"] = "International cardiovascular research journal",
  ["int child health"] = "International child health",
  ["int clin pathol j"] = "International clinical pathology journal",
  ["int clin psychopharmacol"] = "International clinical psychopharmacology",
  ["int commun gaz"] = "International communication gazette",
  ["int commun heat mass transfer"] = "International Communications in Heat and Mass Transfer",
  ["int comp law q"] = "The International and comparative law quarterly",
  ["int conf adapt struct technol"] = "International Conference on Adaptive Structures and Technologies : [proceedings]. International Conference on Adaptive Structures and Technologies",
  ["int conf affect comput intell interact workshops"] = "International Conference on Affective Computing and Intelligent Interaction and workshops : [proceedings]. ACII (Conference)",
  ["int conf aids"] = "International Conference On Aids",
  ["int conf audit vis speech process"] = "International Conference on Auditory-Visual Speech Processing",
  ["int conf bioinform biomed eng"] = "International Conference on Bioinformatics and Biomedical Engineering : [proceedings]. International Conference on Bioinformatics and Biomedical Engineering",
  ["int conf bioinspired comput theor appl"] = "International Conference on Bio-inspired Computing, Theories and Applications : [proceedings]. International Conference on Bio-inspired Computing, Theories and Applications",
  ["int conf collab comput"] = "International conference on collaborative computing : networking, applications and worksharing (CollaborateCom). International Conference on Collaborative Computing: Networking, Applications, and Worksharing",
  ["int conf commun syst netw"] = "International Conference on Communication Systems and Networks : [proceedings]. International Conference on Communication Systems and Networks",
  ["int conf comput math biomed eng"] = "International Conference on Computational & Mathematical Biomedical Engineering",
  ["int conf comput netw commun"] = "International Conference on Computing, Networking, and Communications : [proceedings]. International Conference on Computing, Networking and Communications",
  ["int conf dependable syst netw workshops"] = "International Conference on Dependable Systems and Networks workshops : [proceedings]. International Conference on Dependable Systems and Networks",
  ["int conf digit signal process proc"] = "International Conference on Digital Signal Processing proceedings : DSP. International Conference on Digital Signal Processing",
  ["int conf electromagn adv appl"] = "International Conference on Electromagnetics in Advanced Applications : proceedings : ICEAA. International Conference on Electromagnetics in Advanced Applications",
  ["int conf environ syst"] = "International Conference on Environmental Systems. American Institute of Aeronautics and Astronautics",
  ["int conf geoinform"] = "International Conference on Geoinformatics : [proceedings]. International Conference on Geoinformatics",
  ["int conf gisci short pap proc"] = "International Conference on GIScience short paper proceedings",
  ["int conf infrared millim terahertz waves"] = "International Conference on Infrared, Millimeter, and Terahertz Waves : [proceedings]. International Conference on Infrared, Millimeter, and Terahertz Waves",
  ["int conf pattern recognit"] = "International conference on pattern recognition",
  ["int conf pervasive comput technol healthc"] = "International Conference on Pervasive Computing Technologies for Healthcare : [proceedings]. International Conference on Pervasive Computing Technologies for Healthcare",
  ["int conf sci stat database manag"] = "International Conference on Scientific and Statistical Database Management : [proceedings]. International Conference on Scientific and Statistical Database Management",
  ["int conf signal process commun"] = "International Conference on Signal Processing & Communications (SPCOM). Conference on Signal Processing, and Communications",
  ["int conf signal process proc"] = "International conference on signal processing proceedings. International Conference on Signal Processing",
  ["int congr ser"] = "International congress series",
  ["int congr ser - excerpta med"] = "International Congress Series - Excerpta Medica",
  ["int contact lens clin"] = "International contact lens clinic (New York, N.Y.)",
  ["int contrib to lab stud"] = "International Contributions to Labour Studies",
  ["int corros conf ser"] = "International Corrosion Conference Series",
  ["int crim justice rev"] = "International criminal justice review",
  ["int dairy j"] = "International Dairy Journal",
  ["int data ser, sel data mixtures, ser a"] = "International DATA Series, Selected Data on Mixtures, Series A",
  ["int demogr"] = "International demographics",
  ["int dent j"] = "International dental journal",
  ["int desalin water reuse q"] = "International Desalination & Water Reuse Quarterly",
  ["int dev rev"] = "International development review",
  ["int dig health legis"] = "International digest of health legislation",
  ["int disabil stud"] = "International disability studies",
  ["int drug discov"] = "International drug discovery",
  ["int drug ther newsl"] = "International drug therapy newsletter",
  ["int econ"] = "International Economics",
  ["int econ insights"] = "International economic insights",
  ["int econ j"] = "International economic journal",
  ["int econ rev"] = "International Economic Review",
  ["int econ rev (philadelphia)"] = "International economic review",
  ["int econom rev"] = "International Economic Review",
  ["int economy"] = "International Economy",
  ["int educ res j"] = "International education and research journal",
  ["int electron j algebra"] = "International Electronic Journal of Algebra",
  ["int electron j geom"] = "International Electronic Journal of Geometry",
  ["int electron j health educ"] = "The international electronic journal of health education",
  ["int electron j pure appl math"] = "International Electronic Journal of Pure and Applied Mathematics",
  ["int emerg nurs"] = "International emergency nursing",
  ["int endod j"] = "International endodontic journal",
  ["int energy j"] = "International Energy Journal",
  ["int environ agreem"] = "International environmental agreements : politics, law and economics",
  ["int fam plan perspect"] = "International family planning perspectives",
  ["int fam plann dig"] = "International family planning digest",
  ["int fam plann persp"] = "International family planning perspectives and digest",
  ["int fem j polit"] = "International feminist journal of politics",
  ["int finance"] = "International Finance",
  ["int fire serv j leadersh manag"] = "International fire service journal of leadership and management",
  ["int food agribus manage rev"] = "International Food and Agribusiness Management Review",
  ["int food res j"] = "International food research journal",
  ["int for fire news"] = "International Forest Fire News",
  ["int for rev"] = "International Forestry Review",
  ["int forum allergy rhinol"] = "International forum of allergy & rhinology",
  ["int gambl stud"] = "International gambling studies",
  ["int game theory rev"] = "International Game Theory Review",
  ["int gaz (for hill)"] = "International gazette (Forest Hill, Md.)",
  ["int geol rev"] = "International Geology Review",
  ["int guide classical stud"] = "ABS international guide to classical studies",
  ["int handb inf syst"] = "International Handbooks on Information Systems",
  ["int health"] = "International health",
  ["int health dev"] = "International Health And Development",
  ["int health hist newsl"] = "International Health History Newsletter",
  ["int health news"] = "International health news",
  ["int heart j"] = "International heart journal",
  ["int heat treat surf eng"] = "International Heat Treatment and Surface Engineering",
  ["int hist nurs j"] = "International history of nursing journal : IHNJ",
  ["int hist rev"] = "The International history review",
  ["int ieee embs conf neural eng"] = "International IEEE/EMBS Conference on Neural Engineering : [proceedings]. International IEEE EMBS Conference on Neural Engineering",
  ["int immunol"] = "International immunology",
  ["int immunopharmacol"] = "International immunopharmacology",
  ["int indig policy j"] = "International indigenous policy journal",
  ["int innov"] = "International innovation : disseminating science, research and technology",
  ["int j"] = "International journal (Toronto, Ont.)",
  ["int j  acarol"] = "International Journal of Acarology",
  ["int j  adv comput sci"] = "International journal of advanced computer science",
  ["int j  des nat ecodyn"] = "International journal of design & nature and ecodynamics : a transdisciplinary journal relating to nature, science and the humanities",
  ["int j  sustainable econ"] = "International Journal of Sustainable Economy",
  ["int j  transit justice"] = "The international journal of transitional justice",
  ["int j abras technol"] = "International Journal of Abrasive Technology",
  ["int j acad med"] = "International journal of academic medicine",
  ["int j acad res"] = "International journal of academic research",
  ["int j acad res dev"] = "International journal of academic research and development",
  ["int j acarol"] = "International journal of acarology",
  ["int j acoust vibr"] = "International Journal of Acoustics and Vibrations",
  ["int j ad hoc ubiquitous comput"] = "International Journal of Ad Hoc and Ubiquitous Computing",
  ["int j adapt control signal process"] = "International Journal of Adaptive Control and Signal Processing",
  ["int j adapt innovative syst"] = "International Journal of Adaptive and Innovative Systems",
  ["int j addict"] = "The International journal of the addictions",
  ["int j addit subtract manuf"] = "International Journal of Additive and Subtractive Materials Manufacturing",
  ["int j adhes adhes"] = "International Journal of Adhesion and Adhesives",
  ["int j adolesc med health"] = "International journal of adolescent medicine and health",
  ["int j adolesc youth"] = "International journal of adolescence and youth",
  ["int j adult orthodon orthognath surg"] = "The International journal of adult orthodontics and orthognathic surgery",
  ["int j adv appl math mech"] = "International Journal of Advances in Applied Mathematics and Mechanics",
  ["int j adv biotechnol res"] = "International journal of advanced biotechnology and research",
  ["int j adv comput sci appl"] = "International journal of advanced computer science and applications : IJACSA",
  ["int j adv comput technol"] = "International journal of advancements in computing technology",
  ["int j adv couns"] = "International journal for the advancement of counseling",
  ["int j adv eng sci appl math"] = "International Journal of Advances in Engineering Sciences and Applied Mathematics",
  ["int j adv eng technol"] = "International journal of advances in engineering and technology",
  ["int j adv innov res"] = "International journal of advance & innovative research",
  ["int j adv intell paradigms"] = "International Journal of Advanced Intelligence Paradigms",
  ["int j adv life sci"] = "International journal on advances in life sciences",
  ["int j adv life sci (coimbatore)"] = "International journal of advanced life sciences",
  ["int j adv logist"] = "International Journal of Advanced Logistics",
  ["int j adv manuf syst"] = "International journal of advanced manufacturing systems",
  ["int j adv manuf technol"] = "International Journal of Advanced Manufacturing Technology",
  ["int j adv mechatron syst"] = "International Journal of Advanced Mechatronic Systems",
  ["int j adv nucl reactor des technol"] = "International Journal of Advanced Nuclear Reactor Design and Technology",
  ["int j adv res (indore)"] = "International journal of advanced research",
  ["int j adv res sci eng technol"] = "International journal of advanced research in science, engineering and technology",
  ["int j adv rob syst"] = "International Journal of Advanced Robotic Systems",
  ["int j adv robot syst"] = "International journal of advanced robotic systems",
  ["int j adv sci eng inf technol"] = "International journal on advanced science, engineering and information technology",
  ["int j adv sci eng technol"] = "International journal of advances in science, engineering and technology",
  ["int j adv struct eng"] = "International Journal of Advanced Structural Engineering",
  ["int j advert"] = "International journal of advertising",
  ["int j aeroacoust"] = "International journal of aeroacoustics",
  ["int j aerodyn"] = "International Journal of Aerodynamics",
  ["int j aeronaut space sci"] = "International Journal of Aeronautical and Space Sciences",
  ["int j aerosp eng"] = "International Journal of Aerospace Engineering",
  ["int j aerosp psychol"] = "International Journal of Aerospace Psychology",
  ["int j aerosp syst sci eng"] = "International Journal of Aerospace System Science and Engineering",
  ["int j afr hist stud"] = "The International journal of African historical studies",
  ["int j afr nurs sci"] = "International journal of Africa nursing sciences",
  ["int j agent technol syst"] = "International journal of agent technologies and systems",
  ["int j agent-oriented softw eng"] = "International Journal of Agent-Oriented Software Engineering",
  ["int j agent-oriented software eng"] = "International Journal of Agent-Oriented Software Engineering",
  ["int j agile extreme software dev"] = "International Journal of Agile and Extreme Software Development",
  ["int j agile manuf"] = "International Journal of Agile Manufacturing",
  ["int j agile syst manage"] = "International Journal of Agile Systems and Management",
  ["int j aging hum dev"] = "International journal of aging & human development",
  ["int j agric biol"] = "International journal of agriculture and biology",
  ["int j agric biol eng"] = "International Journal of Agricultural and Biological Engineering",
  ["int j agric for"] = "International journal of agriculture and forestry (Print)",
  ["int j agric innov res"] = "International journal of agriculture innovations and research",
  ["int j agric innovation technol globalisation"] = "International Journal of Agriculture Innovation, Technology and Globalisation",
  ["int j agric resour governance ecol"] = "International Journal of Agricultural Resources, Governance and Ecology",
  ["int j agric sci res (chennai)"] = "International journal of agricultural science and research (IJASR)",
  ["int j agric sustain"] = "International journal of agricultural sustainability",
  ["int j agric sustainability"] = "International Journal of Agricultural Sustainability",
  ["int j agrisci"] = "International journal of agriscience",
  ["int j agron"] = "International Journal of Agronomy",
  ["int j air pollut"] = "International journal of air pollution",
  ["int j air water pollut"] = "International Journal of Air and Water Pollution",
  ["int j air-cond refrig"] = "International Journal of Air-Conditioning and Refrigeration",
  ["int j alcohol alcohol"] = "International journal on alcohol and alcoholism",
  ["int j alcohol drug res"] = "International journal of alcohol and drug research",
  ["int j algae"] = "International Journal on Algae",
  ["int j algebra comput"] = "International Journal of Algebra and Computation",
  ["int j altern propul"] = "International Journal of Alternative Propulsion",
  ["int j alzheimers dis"] = "International journal of Alzheimer's disease",
  ["int j ambient energy"] = "International Journal of Ambient Energy",
  ["int j anal"] = "International Journal of Analysis",
  ["int j anal chem"] = "International journal of analytical chemistry",
  ["int j anal mass spectrom cromatogr"] = "International journal of analytical mass spectrometry and chromatography",
  ["int j anal pharm biomed sci"] = "International journal of analytical, pharmaceutical and biomedical sciences",
  ["int j anat var"] = "International journal of anatomical variations",
  ["int j androl"] = "International journal of andrology",
  ["int j anesth"] = "International journal of anesthesia",
  ["int j anesth anesth"] = "International journal of anesthetics and anesthesiology",
  ["int j anesthesiol pain med"] = "International journal of anesthesiology & pain medicine",
  ["int j anesthesiol res"] = "International journal of anesthesiology & research",
  ["int j angiol"] = "The International journal of angiology : official publication of the International College of Angiology, Inc",
  ["int j anim vet adv"] = "International journal of animal and veterinary advances",
  ["int j antennas propag"] = "International journal of antennas and propagation",
  ["int j anthropol"] = "International journal of anthropology",
  ["int j anthropol ethnol"] = "International journal of anthropology and ethnology",
  ["int j antimicrob agents"] = "International Journal of Antimicrobial Agents",
  ["int j appl basic med res"] = "International journal of applied & basic medical research",
  ["int j appl biol pharm"] = "International journal of applied biology and pharmaceutical technology",
  ["int j appl biotechnol biochem"] = "International journal of applied biotechnology and biochemistry",
  ["int j appl ceram technol"] = "International Journal of Applied Ceramic Technology",
  ["int j appl chem"] = "International Journal of Applied Chemistry",
  ["int j appl comput math"] = "International Journal of Applied and Computational Mathematics",
  ["int j appl comput topol"] = "Journal of Applied and Computational Topology",
  ["int j appl cryptogr"] = "International Journal of Applied Cryptography",
  ["int j appl decis sci"] = "International Journal of Applied Decision Sciences",
  ["int j appl earth obs geoinf"] = "International Journal of Applied Earth Observation and Geoinformation",
  ["int j appl econ econometrics"] = "International Journal of Applied Economics and Econometrics",
  ["int j appl electromagn mech"] = "International Journal of Applied Electromagnetics and Mechanics",
  ["int j appl electromagnet mech"] = "International Journal of Applied Electromagnetics and Mechanics",
  ["int j appl eng res"] = "International journal of applied engineering research : IJAER",
  ["int j appl glass sci"] = "International Journal of Applied Glass Science",
  ["int j appl inf syst"] = "International journal of applied information systems",
  ["int j appl math"] = "International Journal of Applied Mathematics",
  ["int j appl math (sofia)"] = "International journal of applied mathematics",
  ["int j appl math anal appl"] = "International Journal of Applied Mathematical Analysis and Applications",
  ["int j appl math comput"] = "International Journal of Applied Mathematics and Computation",
  ["int j appl math comput sci"] = "International Journal of Applied Mathematics and Computer Science",
  ["int j appl math sci"] = "International Journal of Applied Mathematical Sciences",
  ["int j appl mech"] = "International journal of applied mechanics",
  ["int j appl nonlinear sci"] = "International Journal of Applied Nonlinear Science",
  ["int j appl pattern recognit"] = "International Journal of Applied Pattern Recognition",
  ["int j appl philos"] = "The International journal of applied philosophy",
  ["int j appl posit psychol"] = "International journal of applied positive psychology",
  ["int j appl psychol"] = "International journal of applied psychology",
  ["int j appl pure sci agric"] = "International journal of applied and pure science and agriculture",
  ["int j appl radiat isot"] = "International Journal of Applied Radiation and Isotopes",
  ["int j appl res vet med"] = "International journal of applied research in veterinary medicine",
  ["int j appl sci biotechnol"] = "International journal of applied sciences and biotechnology",
  ["int j appl sci comput"] = "International Journal of Applied Science and Computations",
  ["int j appl sci technol"] = "International Journal of Applied Science and Technology",
  ["int j appl syst stud"] = "International Journal of Applied Systemic Studies",
  ["int j approx reason"] = "International journal of approximate reasoning : official publication of the North American Fuzzy Information Processing Society",
  ["int j approximate reasoning"] = "International Journal of Approximate Reasoning",
  ["int j aquac fish sci"] = "International journal of aquaculture and fishery sciences",
  ["int j aquat res environ stud"] = "International Journal of Aquatic Research and Environmental Studies",
  ["int j archit comput"] = "International Journal of Architectural Computing",
  ["int j archit heritage"] = "International Journal of Architectural Heritage",
  ["int j aromather"] = "International Journal of Aromatherapy",
  ["int j artif intell educ"] = "International journal of artificial intelligence in education",
  ["int j artif intell rob res"] = "International Journal of Artificial Intelligence and Robotics Research",
  ["int j artif intell soft comput"] = "International Journal of Artificial Intelligence and Soft Computing",
  ["int j artif intell tools"] = "International Journal on Artificial Intelligence Tools",
  ["int j artif organs"] = "The International journal of artificial organs",
  ["int j asian lang process"] = "International Journal of Asian Language Processing",
  ["int j astrobiol"] = "International Journal of Astrobiology",
  ["int j audiol"] = "International journal of audiology",
  ["int j autom comput"] = "International Journal of Automation and Computing",
  ["int j autom control"] = "International Journal of Automation and Control",
  ["int j automot compos"] = "International Journal of Automotive Composites",
  ["int j automot eng"] = "International journal of automotive engineering",
  ["int j automot technol"] = "International Journal of Automotive Technology",
  ["int j automot technol manage"] = "International Journal of Automotive Technology and Management",
  ["int j auton adapt commun syst"] = "International journal of autonomous and adaptive communications systems",
  ["int j auton comput"] = "International Journal of Autonomic Computing",
  ["int j aviat psychol"] = "The International journal of aviation psychology",
  ["int j ayurveda res"] = "International journal of Ayurveda research",
  ["int j bacteriol"] = "International journal of bacteriology",
  ["int j basic appl physiol"] = "International journal of basic and applied physiology",
  ["int j basic clin immunol"] = "International journal of basic and clinical immunology",
  ["int j basic clin pharmacol"] = "International journal of basic and clinical pharmacology",
  ["int j behav consult ther"] = "International journal of behavioral and consultation therapy",
  ["int j behav dev"] = "International journal of behavioral development",
  ["int j behav healthc res"] = "International journal of behavioural & healthcare research",
  ["int j behav med"] = "International journal of behavioral medicine",
  ["int j behav nutr phys act"] = "The international journal of behavioral nutrition and physical activity",
  ["int j bifur chaos appl sci eng"] = "International Journal of Bifurcation and Chaos in Applied Sciences and Engineering",
  ["int j bifurcat chaos"] = "International journal of bifurcation and chaos in applied sciences and engineering",
  ["int j bifurcation chaos"] = "International Journal of Bifurcation and Chaos",
  ["int j bifurcation chaos appl sci eng"] = "International Journal of Bifurcation and Chaos in Applied Sciences and Engineering",
  ["int j big data"] = "International journal of big data",
  ["int j big data intel"] = "International Journal of Big Data Intelligence",
  ["int j big data manage"] = "International Journal of Big Data Management",
  ["int j big data min global warming"] = "International Journal of Big Data Mining for Global Warming",
  ["int j biling educ biling"] = "International journal of bilingual education and bilingualism",
  ["int j billing"] = "The international journal of bilingualism : cross-disciplinary, cross-linguistic studies of language behavior",
  ["int j bio-chromatogr"] = "International Journal of Biochromatography",
  ["int j bio-inspired comput"] = "International Journal of Bio-Inspired Computation",
  ["int j bio-med comput"] = "International Journal of Bio-Medical Computing",
  ["int j bioassays"] = "International journal of bioassays",
  ["int j biobased plast"] = "International Journal of Biobased Plastics",
  ["int j biochem"] = "International Journal of Biochemistry",
  ["int j biochem biotechnol"] = "International journal of biochemistry and biotechnology",
  ["int j biochem cell biol"] = "The international journal of biochemistry & cell biology",
  ["int j biochem mol biol"] = "International journal of biochemistry and molecular biology",
  ["int j biochem res rev"] = "International journal of biochemistry research & review",
  ["int j biochemiphysics"] = "International Journal of BioChemiPhysics",
  ["int j biochromatogr"] = "International journal of bio-chromatography",
  ["int j bioclimatol biometeorol"] = "International Journal of Bioclimatology and Biometeorology",
  ["int j biodivers conserv"] = "International journal of biodiversity and conservation",
  ["int j biodivers sci ecosyst serv manag"] = "International journal of biodiversity science, ecosystem services & management",
  ["int j biodivers sci manage"] = "International Journal of Biodiversity Science & Management",
  ["int j bioelectromagn"] = "International journal of bioelectromagnetism",
  ["int j bioinf res appl"] = "International Journal of Bioinformatics Research and Applications",
  ["int j bioinform res appl"] = "International journal of bioinformatics research and applications",
  ["int j biol"] = "International journal of biology",
  ["int j biol biotech"] = "International Journal of Biology and Biotechnology",
  ["int j biol chem"] = "International Journal of Biological Chemistry",
  ["int j biol chem sci"] = "International journal of biological and chemical sciences",
  ["int j biol macromol"] = "International Journal of Biological Macromolecules",
  ["int j biol markers"] = "The International journal of biological markers",
  ["int j biol med res"] = "International journal of biological and medical research",
  ["int j biol pharm allied sci"] = "International journal of biology, pharmacy and allied sciences",
  ["int j biol res pregnancy"] = "International journal of biological research in pregnancy",
  ["int j biol sci"] = "International journal of biological sciences",
  ["int j biom"] = "International journal of biometrics",
  ["int j biom bioinformatics"] = "International journal of biometrics and bioinformatics : IJBB",
  ["int j biomater"] = "International journal of biomaterials",
  ["int j biomater res eng"] = "International journal of biomaterials research and engineering",
  ["int j biomath"] = "International Journal of Biomathematics",
  ["int j biomechatron biomed rob"] = "International Journal of Biomechatronics and Biomedical Robotics",
  ["int j biomed comput"] = "International journal of bio-medical computing",
  ["int j biomed data min"] = "International journal of biomedical data mining",
  ["int j biomed eng"] = "International Journal of Biomedical Engineering",
  ["int j biomed eng clin sci"] = "International journal of biomedical engineering and clinical science",
  ["int j biomed eng technol"] = "International Journal of Biomedical Engineering and Technology",
  ["int j biomed imaging"] = "International journal of biomedical imaging",
  ["int j biomed investig"] = "International journal of biomedical investigation",
  ["int j biomed nanosci nanotechnol"] = "International Journal of Biomedical Nanoscience and",
  ["int j biomed res"] = "International journal of biomedical research",
  ["int j biomed sci"] = "International Journal of Biomedical Sciences",
  ["int j biometeorol"] = "International journal of biometeorology",
  ["int j bioorganic chem mol biol"] = "International journal of bioorganic chemistry & molecular biology",
  ["int j biopharm sci"] = "International journal of biopharmaceutical sciences",
  ["int j bioprint"] = "International journal of bioprinting",
  ["int j biosci"] = "International journal of biosciences",
  ["int j biosci biochem bioinforma"] = "International journal of bioscience, biochemistry and bioinformatics",
  ["int j biosci psychiatr technol ijbspt"] = "International journal of biosciences, psychiatry, and technology (IJBSPT)",
  ["int j biosens bioelectron"] = "International journal of biosensors & bioelectronics",
  ["int j biostat"] = "The international journal of biostatistics",
  ["int j biotech trends technol"] = "International journal of biotech trends and technology",
  ["int j biotechnol"] = "International Journal of Biotechnology",
  ["int j biotechnol res (chennai)"] = "International journal of bio-technology and research (Chennai)",
  ["int j biotechnol wellness ind"] = "International journal of biotechnology for wellness industries",
  ["int j bipolar disord"] = "International journal of bipolar disorders",
  ["int j birth parent educ"] = "International journal of birth and parent education",
  ["int j blood res disord"] = "International journal of blood research and disorders",
  ["int j body compos res"] = "International journal of body composition research",
  ["int j bone marrow res"] = "International journal of bone marrow research",
  ["int j bot stud"] = "International journal of botany studies",
  ["int j botany"] = "International journal of botany : IJB",
  ["int j brain cogn sci"] = "International journal of brain and cognitive sciences",
  ["int j brain disord treat"] = "International journal of brain disorders and treatment",
  ["int j breast cancer"] = "International journal of breast cancer",
  ["int j build pathol adapt"] = "International Journal of Building Pathology and Adaptation",
  ["int j bullying prev"] = "International journal of bullying prevention : an official publication of the International Bullying Prevention Association",
  ["int j burns trauma"] = "International journal of burns and trauma",
  ["int j bus"] = "International Journal of Business",
  ["int j bus inf syst"] = "International journal of business information systems",
  ["int j canc prev"] = "International journal of cancer prevention",
  ["int j cancer"] = "International Journal of Cancer",
  ["int j cancer clin res"] = "International journal of cancer and clinical research",
  ["int j cancer manag"] = "International journal of cancer management",
  ["int j cancer oncol"] = "International journal of cancer and oncology",
  ["int j cancer res"] = "International journal of cancer research",
  ["int j cancer res mol mech"] = "International journal of cancer research and molecular mechanisms",
  ["int j cancer res prev"] = "International journal of cancer research and prevention",
  ["int j cancer res ther"] = "International journal of cancer research & therapy",
  ["int j cancer sci ther"] = "International journal of cancer science & therapy",
  ["int j cancer suppl"] = "International Journal of Cancer. Supplement",
  ["int j card imaging"] = "International journal of cardiac imaging",
  ["int j cardiol"] = "International journal of cardiology",
  ["int j cardiol heart vasc"] = "International journal of cardiology. Heart & vasculature",
  ["int j cardiol heart vessel"] = "International journal of cardiology. Heart & vessels",
  ["int j cardiol heart vessels"] = "International Journal of Cardiology-Heart & Vessels",
  ["int j cardiol hypertens"] = "International Journal of Cardiology. Hypertension",
  ["int j cardiovasc imaging"] = "The international journal of cardiovascular imaging",
  ["int j cardiovasc intervent"] = "International journal of cardiovascular interventions",
  ["int j cardiovasc res"] = "International journal of cardiovascular research",
  ["int j cardol metab endocr"] = "International Journal of Cardiology-Metabolic & Endocrine",
  ["int j care caring"] = "International journal of care and caring",
  ["int j care coord"] = "International journal of care coordination",
  ["int j care pathw"] = "International journal of care pathways",
  ["int j caring sci"] = "International journal of caring sciences",
  ["int j cartogr"] = "International Journal of Cartography",
  ["int j case rep imag"] = "International journal of case reports and images",
  ["int j cast met res"] = "International Journal of Cast Metals Research",
  ["int j cell biol"] = "International journal of cell biology",
  ["int j cell biol physiol"] = "International journal of cell biology and physiology",
  ["int j cell cloning"] = "International journal of cell cloning",
  ["int j cell sci mol biol"] = "International journal of cell science & molecular biology",
  ["int j cell tissue eng artif cells regener med"] = "International Journal of Cell and Tissue Engineering, Artificial Cells and Regenerative Medicine",
  ["int j cem compos lightweight concr"] = "International Journal of Cement Composites and Lightweight Concrete",
  ["int j cem compos lightweight concrete"] = "International Journal of Cement Composites and Lightweight Concrete",
  ["int j cent bank"] = "International journal of central banking",
  ["int j ceram eng sci"] = "International Journal of Ceramic Engineering and Science",
  ["int j cerebrovasc dis stroke"] = "International journal of cerebrovascular disease and stroke",
  ["int j chem"] = "International journal of chemistry",
  ["int j chem anal sci"] = "International Journal of Chemical and Analytical Science",
  ["int j chem biochem sci"] = "International Journal Of Chemical And Biochemical Sciences",
  ["int j chem eng"] = "International Journal of Chemical Engineering",
  ["int j chem eng appl"] = "International Journal of Chemical Engineering and",
  ["int j chem kinet"] = "International Journal of Chemical Kinetics",
  ["int j chem model"] = "International Journal of Chemical Modeling",
  ["int j chem reactor eng"] = "International Journal of Chemical Reactor Engineering",
  ["int j chem sc"] = "International Journal of Chemical Sciences",
  ["int j chem stud"] = "International journal of chemical studies",
  ["int j chemtech res"] = "International journal of chemtech research",
  ["int j child adolesc health"] = "International journal of child and adolescent health",
  ["int j child adolesc resil"] = "International journal of child and adolescent resilience",
  ["int j child comput interact"] = "International journal of child-computer interaction",
  ["int j child dev ment health"] = "International journal of child development and mental health",
  ["int j child health hum dev"] = "International journal of child health and human development : IJCHD",
  ["int j child health nutr"] = "International journal of child health and nutrition",
  ["int j child maltreat"] = "International journal on child maltreatment : research, policy and practice",
  ["int j child psychother"] = "International journal of child psychotherapy",
  ["int j child youth family stud"] = "International journal of child, youth & family studies : IJCYFS",
  ["int j child-comput interact"] = "International Journal of Child-Computer Interaction",
  ["int j childbirth"] = "International journal of childbirth",
  ["int j childbirth educ"] = "The International journal of childbirth education : the official publication of the International Childbirth Education Association",
  ["int j chin cult manag"] = "International journal of Chinese culture and management",
  ["int j chron obstruct pulmon dis"] = "International journal of chronic obstructive pulmonary disease",
  ["int j chronic dis"] = "International journal of chronic diseases",
  ["int j chronic dis ther"] = "International journal of chronic diseases & therapy",
  ["int j chronobiol"] = "International journal of chronobiology",
  ["int j circuit theory appl"] = "International Journal of Circuit Theory and Applications",
  ["int j circumpolar health"] = "International journal of circumpolar health",
  ["int j circumpolar health suppl"] = "International Journal of Circumpolar Health. Supplement",
  ["int j civ eng"] = "International Journal of Civil Engineering",
  ["int j civ engagem soc change"] = "International journal of civic engagement and social change",
  ["int j clim chang strateg manag"] = "International journal of climate change strategies and management",
  ["int j clim change strategies manage"] = "International Journal of Climate Change Strategies and Management",
  ["int j clim change: impacts responses"] = "International Journal of Climate Change: Impacts and Responses",
  ["int j climatol"] = "International Journal of Climatology",
  ["int j clin anesthesiol"] = "International journal of clinical anesthesiology",
  ["int j clin biostat biom"] = "International journal of clinical biostatistics and biometrics",
  ["int j clin case rep rev"] = "International journal of clinical case reports and reviews : open access",
  ["int j clin endocrinol metab"] = "International journal of clinical endocrinology and metabolism",
  ["int j clin exp hypn"] = "The International journal of clinical and experimental hypnosis",
  ["int j clin exp med"] = "International Journal of Clinical and Experimental Medicine",
  ["int j clin exp path"] = "International Journal of Clinical and Experimental Pathology",
  ["int j clin exp pathol"] = "International journal of clinical and experimental pathology",
  ["int j clin exp physiol"] = "International journal of clinical and experimental physiology",
  ["int j clin health psychol"] = "International journal of clinical and health psychology : IJCHP",
  ["int j clin lab res"] = "International Journal of Clinical and Laboratory Research",
  ["int j clin med"] = "International journal of clinical medicine",
  ["int j clin monit comput"] = "International journal of clinical monitoring and computing",
  ["int j clin oncol"] = "International journal of clinical oncology",
  ["int j clin oral maxillofac surg"] = "International journal of clinical oral and maxillofacial surgery",
  ["int j clin pediatr dent"] = "International journal of clinical pediatric dentistry",
  ["int j clin pharm"] = "International journal of clinical pharmacy",
  ["int j clin pharmacol"] = "International journal of clinical pharmacology, therapy and toxicology",
  ["int j clin pharmacol biopharm"] = "International journal of clinical pharmacology and biopharmacy",
  ["int j clin pharmacol res"] = "International journal of clinical pharmacology research",
  ["int j clin pharmacol ther"] = "International journal of clinical pharmacology and therapeutics",
  ["int j clin pharmacol ther toxicol"] = "International journal of clinical pharmacology, therapy, and toxicology",
  ["int j clin pharmacol toxicol"] = "International journal of clinical pharmacology & toxicology",
  ["int j clin pract"] = "International journal of clinical practice",
  ["int j clin pract suppl"] = "International journal of clinical practice. Supplement",
  ["int j clin psychiatry ment health"] = "International journal of clinical psychiatry and mental health",
  ["int j clin res trials"] = "International journal of clinical research & trials",
  ["int j clin rheumtol"] = "International journal of clinical rheumatology",
  ["int j clin tranfus med"] = "International Journal of Clinical Transfusion Medicine",
  ["int j clin trials"] = "International journal of clinical trials",
  ["int j clothing sci technol"] = "International Journal of Clothing Science and Technology",
  ["int j coal geol"] = "International Journal of Coal Geology",
  ["int j coal prep util"] = "International Journal of Coal Preparation and Utilization",
  ["int j coal sci technol"] = "International journal of coal science & technology",
  ["int j cogn ergon"] = "International journal of cognitive ergonomics",
  ["int j cogn linguist"] = "International journal of cognitive linguistics",
  ["int j cogn technol"] = "The international journal of cognitive technology : the official journal of the Practical Memory Institute",
  ["int j cogn ther"] = "International journal of cognitive therapy",
  ["int j cognit biom"] = "International Journal of Cognitive Biometrics",
  ["int j cognit comput eng"] = "International Journal of Cognitive Computing in Engineering",
  ["int j collab res intern med public health"] = "International journal of collaborative research on internal medicine & public health",
  ["int j colorectal dis"] = "International journal of colorectal disease",
  ["int j comadem"] = "International Journal of COMADEM",
  ["int j commons"] = "International journal of the commons",
  ["int j commun"] = "International journal of communication",
  ["int j commun antenna propag"] = "International journal on communications antenna and propagation",
  ["int j commun health"] = "International journal of communication and health",
  ["int j commun netw syst sci"] = "International journal of communications, network and systems sciences",
  ["int j commun networks distrib syst"] = "International Journal of Communication Networks and Distributed Systems",
  ["int j commun syst"] = "International Journal of Communication Systems",
  ["int j community based nurs midwifery"] = "International journal of community based nursing and midwifery",
  ["int j community curr res"] = "International journal of community currency research",
  ["int j community med public health"] = "International journal of community medicine and public health",
  ["int j comp appl crim justice"] = "International journal of comparative and applied criminal justice",
  ["int j comp psychol"] = "International journal of comparative psychology",
  ["int j comp sci"] = "International journal of computer science",
  ["int j comp sociol"] = "International journal of comparative sociology",
  ["int j complement altern med"] = "International journal of complementary & alternative medicine",
  ["int j comput appl"] = "International Journal of Computers and Applications",
  ["int j comput appl technol"] = "International Journal of Computer Applications in Technology",
  ["int j comput assist radiol surg"] = "International journal of computer assisted radiology and surgery",
  ["int j comput assisted radiol surg"] = "International Journal of Computer Assisted Radiology and Surgery",
  ["int j comput biol drug des"] = "International journal of computational biology and drug design",
  ["int j comput dent"] = "International journal of computerized dentistry",
  ["int j comput econ econom"] = "International Journal of Computational Economics and Econometrics",
  ["int j comput eng sci"] = "International Journal of Computational Engineering Science",
  ["int j comput fluid dyn"] = "International Journal of Computational Fluid Dynamics",
  ["int j comput games technol"] = "International Journal of Computer Games Technology",
  ["int j comput geom appl"] = "International journal of computational geometry & applications",
  ["int j comput healthc"] = "International journal of computers in healthcare",
  ["int j comput integr manuf"] = "International journal of computer integrated manufacturing",
  ["int j comput intell appl"] = "International Journal of Computational Intelligence and Applications",
  ["int j comput intell syst"] = "International Journal of Computational Intelligence Systems",
  ["int j comput mater sci eng"] = "International journal of computational materials science and engineering",
  ["int j comput mater sci surf eng"] = "International Journal of Computational Materials Science and Surface Engineering",
  ["int j comput math"] = "International Journal of Computer Mathematics",
  ["int j comput math comput syst theory"] = "International Journal of Computer Mathematics. Computer Systems Theory",
  ["int j comput math learn"] = "International Journal of Computers for Mathematical Learning",
  ["int j comput math: comput syst theory"] = "International Journal of Computer Mathematics: Computer Systems Theory",
  ["int j comput methods"] = "International Journal of Computational Methods",
  ["int j comput methods eng sci mech"] = "International Journal for Computational Methods in Engineering Science and Mechanics",
  ["int j comput methods exp meas"] = "International Journal of Computational Methods and Experimental Measurements",
  ["int j comput models algorithms med"] = "International journal of computational models and algorithms in medicine",
  ["int j comput process lang"] = "International Journal of Computer Processing of Languages",
  ["int j comput sci"] = "International journal of computational science",
  ["int j comput sci (rabat)"] = "International journal of computer science (Rabat)",
  ["int j comput sci eng"] = "International Journal of Computational Science and Engineering",
  ["int j comput sci math"] = "International Journal of Computing Science and Mathematics",
  ["int j comput sci sport"] = "International journal of computer science in sport",
  ["int j comput support collab learn"] = "International journal of computer-supported collaborative learning",
  ["int j comput vis"] = "International journal of computer vision",
  ["int j comput vis biomech"] = "International journal for computational vision and biomechanics",
  ["int j comput vis robot"] = "International journal of computational vision and robotics",
  ["int j comput vision"] = "International Journal of Computer Vision",
  ["int j comput-supported collab learn"] = "International Journal of Computer-Supported Collaborative Learning",
  ["int j computer integr manuf"] = "International Journal of Computer Integrated Manufacturing",
  ["int j comut fluid dyn"] = "International journal of computational fluid dynamics",
  ["int j concr struct mater"] = "International journal of concrete structures & materials",
  ["int j condens matter res"] = "International Journal of Condensed Matter Research",
  ["int j conf violence"] = "International journal of conflict and violence",
  ["int j const law"] = "International journal of constitutional law",
  ["int j constr educ res"] = "International journal of construction education and research",
  ["int j construct educ res"] = "International Journal of Construction Education and Research",
  ["int j construct manage"] = "International Journal of Construction Management",
  ["int j construct project manage"] = "International Journal of Construction Project Management",
  ["int j consum stud"] = "International journal of consumer studies",
  ["int j contemp pediatrics"] = "International journal of contemporary pediatrics",
  ["int j contemp sociol"] = "International journal of contemporary sociology",
  ["int j contin eng educ life long learn"] = "International journal of continuing engineering education and life-long learning",
  ["int j control"] = "International Journal of Control",
  ["int j control autom syst"] = "International Journal of Control, Automation and Systems",
  ["int j coop inf syst"] = "International journal of cooperative information systems",
  ["int j cooperative inf syst"] = "International Journal of Cooperative Information Systems",
  ["int j corros"] = "International Journal of Corrosion",
  ["int j cosmet sci"] = "International journal of cosmetic science",
  ["int j crashworthiness"] = "International Journal of Crashworthiness",
  ["int j criminol sociol"] = "International journal of criminology and sociology",
  ["int j crit illn inj sci"] = "International journal of critical illness and injury science",
  ["int j crit infrastruct"] = "International Journal of Critical Infrastructures",
  ["int j crit infrastruct prot"] = "International Journal of Critical Infrastructure Protection",
  ["int j crowd sci"] = "International Journal of Crowd Science",
  ["int j crude drug res"] = "International journal of crude drug research",
  ["int j cult ment health"] = "International journal of culture and mental health",
  ["int j curr adv res"] = "International journal of current advanced research",
  ["int j curr chem"] = "International journal of current chemistry",
  ["int j curr eng sci res"] = "International Journal of Current Engineering and Scientific Research",
  ["int j curr microbiol appl sci"] = "International journal of current microbiology and applied sciences",
  ["int j curr multidiscip stud"] = "International journal of current multidisciplinary studies",
  ["int j curr pharm res"] = "International journal of current pharmaceutical research",
  ["int j curr res"] = "International journal of current research",
  ["int j curr res acad rev"] = "International journal of current research and academic review",
  ["int j curr res biosci plant biol"] = "International journal of current research in biosciences and plant biology",
  ["int j curr res rev"] = "International journal of current research and review",
  ["int j curr trends sci technol"] = "International Journal of Current Trends in Science and Technology",
  ["int j dairy technol"] = "International Journal of Dairy Technology",
  ["int j damage mech"] = "International Journal of Damage Mechanics",
  ["int j data min bioinf"] = "International Journal of Data Mining and Bioinformatics",
  ["int j data min bioinform"] = "International journal of data mining and bioinformatics",
  ["int j data min model manage"] = "International Journal Of Data Mining Modelling And Management",
  ["int j data sci anal"] = "International journal of data science and analytics",
  ["int j data sci math sci"] = "International Journal of Data Science in the Mathematical Sciences",
  ["int j data warehouse min"] = "International Journal of Data Warehousing and Mining",
  ["int j dent"] = "International journal of dentistry",
  ["int j dent hyg"] = "International journal of dental hygiene",
  ["int j dent oral health"] = "International journal of dentistry and oral health",
  ["int j dent oral sci"] = "International journal of dentistry and oral science",
  ["int j dent sci res"] = "International Journal of Dental Science and Research",
  ["int j dent symp"] = "International journal of dental symposia",
  ["int j dermatol"] = "International journal of dermatology",
  ["int j dermatol venereol"] = "International journal of dermatology and venereology",
  ["int j des creativity innovation"] = "International Journal of Design Creativity and Innovation",
  ["int j des educ"] = "The international journal of design education",
  ["int j des learn"] = "International journal of designs for learning",
  ["int j des nat"] = "International Journal of Design and Nature",
  ["int j des nat ecodyn"] = "International Journal of Design & Nature and Ecodynamics",
  ["int j dev biol"] = "The International journal of developmental biology",
  ["int j dev disabil"] = "International journal of developmental disabilities",
  ["int j dev neurosci"] = "International journal of developmental neuroscience : the official journal of the International Society for Developmental Neuroscience",
  ["int j dev res"] = "International journal of developmental research",
  ["int j dev sci"] = "International journal of developmental science",
  ["int j diabetes clin res"] = "International journal of diabetes and clinical research",
  ["int j diabetes dev ctries"] = "International journal of diabetes in developing countries",
  ["int j diabetes mellit"] = "International journal of diabetes mellitus",
  ["int j diabetes metab syndr"] = "International journal of diabetes & metabolic syndrome",
  ["int j diabetes res"] = "International journal of diabetes research",
  ["int j diabetol vasc dis res"] = "International journal of diabetology & vascular disease research",
  ["int j differ equ"] = "International Journal of Differential Equations",
  ["int j differ equations"] = "International Journal of Differential Equations",
  ["int j difference equ"] = "International Journal of Difference Equations",
  ["int j digit curation"] = "International journal of digital curation",
  ["int j digit earth"] = "International journal of digital earth",
  ["int j digital earth"] = "International Journal of Digital Earth",
  ["int j digital multimedia broadcast"] = "International Journal of Digital Multimedia Broadcasting",
  ["int j disabil hum dev"] = "International journal on disability and human development : IJDHD",
  ["int j disaster resil built environ"] = "International journal of disaster resilience in the built environment",
  ["int j disaster resilience built environ"] = "International Journal of Disaster Resilience in the Built Environment",
  ["int j disaster risk reduct"] = "International Journal of Disaster Risk Reduction",
  ["int j disaster risk sci"] = "International Journal of Disaster Risk Science",
  ["int j distrib sens netw"] = "International Journal of Distributed Sensor Networks",
  ["int j divers organ communities nations"] = "The international journal of diversity in organisations, communities and nations",
  ["int j doc anal recogn"] = "International Journal on Document Analysis and Recognition",
  ["int j doc anal recognit"] = "International journal on document analysis and recognition (Online)",
  ["int j drug policy"] = "The International journal on drug policy",
  ["int j dyn control"] = "International Journal of Dynamics and Control",
  ["int j dyn syst differ equ"] = "International Journal of Dynamical Systems and Differential Equations",
  ["int j dyn syst differ equations"] = "International Journal of Dynamical Systems and Differential Equations",
  ["int j e-navig marit econ"] = "International Journal of e-Navigation and Maritime Economy",
  ["int j early child"] = "International journal of early childhood = Revue internationale de l'enfance préscolaire = Revista internacional de la infancia pre-escolar",
  ["int j early years educ"] = "International journal of early years education",
  ["int j earth sci"] = "International Journal of Earth Sciences",
  ["int j eat disord"] = "The International journal of eating disorders",
  ["int j ecodyn"] = "International Journal of Ecodynamics",
  ["int j ecol"] = "International Journal of Ecology",
  ["int j ecol econ stat"] = "International journal of ecological economics and statistics",
  ["int j econ bus"] = "International journal of the economics of business",
  ["int j econ finance"] = "International journal of economics and finance",
  ["int j econ res"] = "International journal of economic research",
  ["int j econ theory"] = "International Journal of Economic Theory",
  ["int j ecotoxicol ecobiol"] = "International journal of ecotoxicology and ecobiology",
  ["int j educ dev"] = "International journal of educational development",
  ["int j educ dev using inf commun technol"] = "International journal of education and development using information and communication technology",
  ["int j educ method"] = "International journal of educational methodology",
  ["int j educ psychol assess"] = "The International journal of educational and psychological assessment",
  ["int j educ res"] = "International journal of educational research",
  ["int j educ soc sci"] = "International journal of education and social science",
  ["int j educ vocat guid"] = "International journal for educational and vocational guidance",
  ["int j electr comput syst eng"] = "International Journal of Electrical, Computer, and Systems Engineering",
  ["int j electr eng"] = "International Journal of Electrical Engineering",
  ["int j electr eng educ"] = "International Journal of Electrical Engineering Education",
  ["int j electr hybrid veh"] = "International Journal of Electric and Hybrid Vehicles",
  ["int j electr power energy syst"] = "International Journal of Electrical Power & Energy Systems",
  ["int j electrochem"] = "International Journal of Electrochemistry",
  ["int j electrochem sci"] = "International Journal of Electrochemical Science",
  ["int j electron"] = "International Journal of Electronics",
  ["int j electron commun"] = "International journal of electronics and communications = Archiv fur Elektronik und Ubertragungstechnik : AEU",
  ["int j electron healthc"] = "International journal of electronic healthcare",
  ["int j electron lett"] = "International Journal of Electronics Letters",
  ["int j electron secur digit forensics"] = "International Journal of Electronic Security and Digital Forensics",
  ["int j embed syst"] = "International journal of embedded systems",
  ["int j embedded syst"] = "International Journal of Embedded Systems",
  ["int j emerg electr power syst"] = "International Journal of Emerging Electric Power Systems",
  ["int j emerg manage"] = "International Journal of Emergency Management",
  ["int j emerg med"] = "International journal of emergency medicine",
  ["int j emerg ment health"] = "International journal of emergency mental health",
  ["int j emerg multidiscip fluid sci"] = "International journal of emerging multidisciplinary fluid sciences",
  ["int j emergency manage"] = "International Journal of Emergency Management",
  ["int j emerging electr power syst"] = "International Journal of Emerging Electric Power Systems",
  ["int j emot educ"] = "The international journal of emotional education",
  ["int j empirical econ"] = "International Journal of Empirical Economics",
  ["int j endocr oncol"] = "International journal of endocrine oncology",
  ["int j endocrinol"] = "International journal of endocrinology",
  ["int j endocrinol metab"] = "International journal of endocrinology and metabolism",
  ["int j endocrinol metab disord"] = "International journal of endocrinology and metabolic disorders",
  ["int j energetic mater chem propul"] = "International Journal of Energetic Materials and Chemical Propulsion",
  ["int j energy clean environ"] = "International Journal of Energy for a Clean Environment",
  ["int j energy convers"] = "International Journal On Energy Conversion",
  ["int j energy environ"] = "International Journal of Energy and Environment",
  ["int j energy environ eng"] = "International Journal of Energy and Environmental Engineering",
  ["int j energy prod manage"] = "International Journal of Energy Production and Management",
  ["int j energy res"] = "International Journal of Energy Research",
  ["int j energy sect manage"] = "International Journal of Energy Sector Management",
  ["int j energy water resour"] = "International Journal of Energy and Water Resources",
  ["int j eng"] = "International Journal of Engineering",
  ["int j eng adv technol"] = "International journal of engineering and advanced technology",
  ["int j eng bus manage"] = "International Journal of Engineering Business Management",
  ["int j eng educ"] = "International Journal of Engineering Education",
  ["int j eng math"] = "International Journal of Engineering Mathematics",
  ["int j eng model"] = "International Journal for Engineering Modelling",
  ["int j eng res afr"] = "International Journal of Engineering Research in Africa",
  ["int j eng res appl"] = "International journal of engineering research and applications",
  ["int j eng res technol (ahmedabad)"] = "International journal of engineering research & technology (Ahmedabad)",
  ["int j eng sci"] = "International Journal of Engineering Science",
  ["int j eng sci (ghaziabad)"] = "The international journal of engineering and science",
  ["int j eng sci res"] = "International journal of engineering and scientific research",
  ["int j eng sci res technol"] = "International journal of engineering sciences & research technology",
  ["int j eng syst modell simul"] = "International Journal of Engineering Systems Modelling and Simulation",
  ["int j eng technol"] = "International journal of engineering and technology : IJET",
  ["int j eng technol sci innov"] = "International journal of engineering technology and scientific innovation",
  ["int j eng trans a"] = "International Journal of Engineering, Transactions A: Basics",
  ["int j eng trans b"] = "International Journal of Engineering, Transactions B: Applications",
  ["int j eng, trans a:"] = "International Journal of Engineering, Transactions A:",
  ["int j engine res"] = "International Journal of Engine Research",
  ["int j engl linguist"] = "International journal of English linguistics",
  ["int j enhanc res sci technol eng"] = "International journal of enhanced research in science technology and engineering",
  ["int j enteric pathog"] = "International journal of enteric pathogens",
  ["int j enterp inf syst"] = "International Journal of Enterprise Information Systems",
  ["int j entomol res"] = "International Journal of Entomological Research",
  ["int j entrep innov manag"] = "International journal of entrepreneurship and innovation management",
  ["int j environ agric res"] = "International journal of environmental & agriculture research",
  ["int j environ anal chem"] = "International Journal of Environmental Analytical Chemistry",
  ["int j environ bioremediat biodegrad"] = "International journal of environmental bioremediation & biodegradation",
  ["int j environ health eng"] = "International journal of environmental health engineering",
  ["int j environ health res"] = "International journal of environmental health research",
  ["int j environ impacts"] = "International Journal of Environmental Impacts",
  ["int j environ pollut"] = "International Journal of Environment and Pollution",
  ["int j environ pollut res"] = "International journal of environment and pollution research",
  ["int j environ res"] = "International journal of environmental research",
  ["int j environ res public health"] = "International journal of environmental research and public health",
  ["int j environ resour res"] = "International journal of environmental resources research",
  ["int j environ sci"] = "International journal of environmental sciences",
  ["int j environ sci educ"] = "International journal of environmental and science education : IJESE",
  ["int j environ sci eng"] = "International journal of environmental science and engineering (Pairie View, Tex.)",
  ["int j environ sci technol"] = "International Journal of Environmental Science and Technology",
  ["int j environ sci technol (tehran)"] = "International journal of environmental science and technology : IJEST",
  ["int j environ stud"] = "The International journal of environmental studies",
  ["int j environ sustainability"] = "International Journal of  Environmental Sustainability",
  ["int j environ sustainable dev"] = "International Journal of Environment And Sustainable Development",
  ["int j environ technol manage"] = "International Journal of Environmental Technology and Management",
  ["int j environ waste manag"] = "International journal of environment and waste management",
  ["int j environ waste manage"] = "International Journal of Environment and Waste Management",
  ["int j epidemiol"] = "International journal of epidemiology",
  ["int j epidemiol res"] = "International journal of epidemiologic research",
  ["int j epilepsy"] = "International journal of epilepsy",
  ["int j equilib res"] = "International journal of equilibrium research",
  ["int j equity health"] = "International journal for equity in health",
  ["int j essent oil ther"] = "International Journal of Essential Oil Therapeutics",
  ["int j esthet dent"] = "The international journal of esthetic dentistry",
  ["int j evid based healthc"] = "International journal of evidence-based healthcare",
  ["int j evidence-based healthcare"] = "International Journal of Evidence-Based Healthcare",
  ["int j evol"] = "International journal of evolution",
  ["int j evol biol"] = "International journal of evolutionary biology",
  ["int j evol equ"] = "International journal of evolution equations",
  ["int j exerc sci"] = "International journal of exercise science",
  ["int j exergy"] = "International Journal of Exergy",
  ["int j exp comput biomech"] = "International journal of experimental and computational biomechanics",
  ["int j exp diabesity res"] = "International journal of experimental diabesity research",
  ["int j exp diabetes res"] = "International journal of experimental diabetes research",
  ["int j exp path"] = "International Journal of Experimental Pathology",
  ["int j exp pathol"] = "International journal of experimental pathology",
  ["int j extensive res"] = "International journal of extensive research",
  ["int j extrem manuf"] = "International Journal of Extreme Manufacturing",
  ["int j extreme manuf"] = "International Journal of Extreme Manufacturing",
  ["int j family med"] = "International journal of family medicine",
  ["int j fatigue"] = "International Journal of Fatigue",
  ["int j fem approaches bioeth"] = "International journal of feminist approaches to bioethics",
  ["int j fermented foods"] = "International journal of fermented foods",
  ["int j fertil"] = "International journal of fertility",
  ["int j fertil menopausal stud"] = "International journal of fertility and menopausal studies",
  ["int j fertil steril"] = "International journal of fertility & sterility",
  ["int j fertil womens med"] = "International journal of fertility and women's medicine",
  ["int j financ eng"] = "International Journal of Financial Engineering",
  ["int j finance econ"] = "International Journal of Finance and Economics",
  ["int j finite vol"] = "International Journal on Finite Volumes",
  ["int j fish aquat stud"] = "International journal of fisheries and aquatic studies",
  ["int j fit"] = "International journal of fitness",
  ["int j flexible manuf syst"] = "International Journal of Flexible Manufacturing Systems",
  ["int j fluid eng"] = "International Journal of Fluid Engineering",
  ["int j fluid mech res"] = "International Journal of Fluid Mechanics Research",
  ["int j fluid power"] = "International Journal of Fluid Power",
  ["int j food agric econ"] = "International journal of food and agricultural economics",
  ["int j food contam"] = "International Journal of Food Contamination",
  ["int j food eng"] = "International Journal of Food Engineering",
  ["int j food microbiol"] = "International journal of food microbiology",
  ["int j food prop"] = "International Journal of Food Properties",
  ["int j food sci"] = "International journal of food science",
  ["int j food sci nutr"] = "International journal of food sciences and nutrition",
  ["int j food sci nutr diet"] = "International journal of food science, nutrition and dietetics",
  ["int j food sci technol"] = "International Journal of Food Science and Technology",
  ["int j food stud"] = "International Journal of Food Studies",
  ["int j for eng"] = "International Journal of Forest Engineering",
  ["int j for res"] = "International Journal of Forestry Research",
  ["int j forecast"] = "International journal of forecasting",
  ["int j forecasting"] = "International Journal of Forecasting",
  ["int j forensic dent"] = "The International journal of forensic dentistry",
  ["int j forensic ment health"] = "International journal of forensic mental health",
  ["int j forensic mental health"] = "International Journal of Forensic Mental Health",
  ["int j forensic sci pathol"] = "International journal of forensic science & pathology",
  ["int j form processes"] = "International Journal of Forming Processes",
  ["int j found comput sci"] = "International Journal of Foundations of Computer Science",
  ["int j fract"] = "International Journal of Fracture",
  ["int j fractmech"] = "International Journal of Fracture Mechanics",
  ["int j fruit sci"] = "International Journal of Fruit Science",
  ["int j funct inform personal med"] = "International journal of functional informatics and personalised medicine",
  ["int j funct nutr"] = "International journal of functional nutrition",
  ["int j fuzzy syst"] = "International Journal of Fuzzy Systems",
  ["int j game theory"] = "International Journal of Game Theory",
  ["int j gaming comput mediat simul"] = "International journal of gaming and computer-mediated simulations",
  ["int j gastrointest cancer"] = "International journal of gastrointestinal cancer",
  ["int j gastron food sci"] = "International journal of gastronomy and food science",
  ["int j gen med"] = "International journal of general medicine",
  ["int j gen syst"] = "International Journal of General Systems",
  ["int j gend sci technol"] = "International journal of gender, science and technology",
  ["int j genet mol biol"] = "International journal of genetics and molecular biology",
  ["int j genom data min"] = "International journal of genomics and data mining",
  ["int j genomics"] = "International journal of genomics",
  ["int j genomics proteomics"] = "International journal of genomics and proteomics",
  ["int j geo-eng"] = "International Journal of Geo-Engineering",
  ["int j geogr inf sci"] = "International journal of geographical information science : IJGIS",
  ["int j geoheritage parks"] = "International Journal of Geoheritage and Parks",
  ["int j geom"] = "International Journal of Geometry",
  ["int j geom methods mod phys"] = "International Journal of Geometric Methods in Modern Physics",
  ["int j geomag aeron"] = "International Journal of Geomagnetism and Aeronomy",
  ["int j geomatics"] = "International Journal of Geomatics",
  ["int j geomech"] = "International Journal of Geomechanics",
  ["int j geophys"] = "International Journal of Geophysics",
  ["int j geosynth ground eng"] = "International Journal of Geosynthetics and Ground Engineering",
  ["int j geotech eng"] = "International Journal of Geotechnical Engineering",
  ["int j geriatr psychiatry"] = "International journal of geriatric psychiatry",
  ["int j gerontol"] = "International journal of gerontology",
  ["int j global energy issues"] = "International Journal of Global Energy Issues",
  ["int j global warming"] = "International Journal of Global Warming",
  ["int j green energy"] = "International Journal of Green Energy",
  ["int j green nanotechnol"] = "International Journal of Green Nanotechnology",
  ["int j green nanotechnol biomed"] = "International Journal of Green Nanotechnology: Biomedicine",
  ["int j green nanotechnol mater sci eng"] = "International Journal of Green Nanotechnology: Materials Science and Engineering",
  ["int j green nanotechnol phys chem"] = "International Journal of Green Nanotechnology: Physics and Chemistry",
  ["int j greenhouse gas control"] = "International Journal of Greenhouse Gas Control",
  ["int j grid util comput"] = "International Journal of Grid and Utility Computing",
  ["int j group psychother"] = "International journal of group psychotherapy",
  ["int j group theory"] = "International Journal of Group Theory",
  ["int j gynaecol obstet"] = "International journal of gynaecology and obstetrics: the official organ of the International Federation of Gynaecology and Obstetrics",
  ["int j gynecol cancer"] = "International journal of gynecological cancer : official journal of the International Gynecological Cancer Society",
  ["int j gynecol obstet reprod med res"] = "International journal of gynecological obstetrical and reproductive medicine research",
  ["int j gynecol pathol"] = "International journal of gynecological pathology : official journal of the International Society of Gynecological Pathologists",
  ["int j head neck surg"] = "International journal of head and neck surgery",
  ["int j health care finance econ"] = "International journal of health care finance and economics",
  ["int j health care qual assur"] = "International journal of health care quality assurance",
  ["int j health care qual assur inc leadersh health serv"] = "International journal of health care quality assurance incorporating Leadership in health services",
  ["int j health econ manag"] = "International journal of health economics and management",
  ["int j health educ"] = "International journal of health education",
  ["int j health geogr"] = "International journal of health geographics",
  ["int j health geographics"] = "International Journal of Health Geographics",
  ["int j health nutr"] = "International journal of health and nutrition",
  ["int j health plann manage"] = "The International journal of health planning and management",
  ["int j health policy manag"] = "International journal of health policy and management",
  ["int j health prof"] = "International journal of health professions",
  ["int j health promot educ"] = "International journal of health promotion and education",
  ["int j health sci"] = "International journal of health sciences",
  ["int j health sci (qassim)"] = "International journal of health sciences",
  ["int j health sci educ"] = "International journal of health sciences education",
  ["int j health sci res"] = "International journal of health sciences and research",
  ["int j health serv"] = "International journal of health services : planning, administration, evaluation",
  ["int j health wellness soc"] = "The international journal of health, wellness & society",
  ["int j healthc"] = "International journal of healthcare",
  ["int j healthc inf syst inform"] = "International journal of healthcare information systems and informatics : official publication of the Information Resources Management Association",
  ["int j healthc manag"] = "International journal of healthcare management",
  ["int j healthc med sci"] = "International journal of healthcare and medical sciences",
  ["int j heat exch"] = "International Journal of Heat Exchangers",
  ["int j heat fluid flow"] = "International Journal of Heat and Fluid Flow",
  ["int j heat mass transf"] = "International journal of heat and mass transfer",
  ["int j heat mass transfer"] = "International Journal of Heat and Mass Transfer",
  ["int j heat technol"] = "International Journal of Heat and Technology",
  ["int j heavy veh syst"] = "International Journal of Heavy Vehicle Systems",
  ["int j hematol"] = "International journal of hematology",
  ["int j hematol oncol"] = "International journal of hematologic oncology",
  ["int j hematol oncol stem cell res"] = "International journal of hematology-oncology and stem cell research",
  ["int j hematol res"] = "International journal of hematology research",
  ["int j hematol ther"] = "International journal of hematology & therapy",
  ["int j hepatol"] = "International journal of hepatology",
  ["int j herb med"] = "International journal of herbal medicine",
  ["int j hig technol ceram"] = "International Journal of High Technology Ceramics",
  ["int j high perform comput appl"] = "International Journal of High Performance Computing Applications",
  ["int j high perform comput networking"] = "International Journal of High Performance Computing and Networking",
  ["int j high perform syst archit"] = "International Journal of High Performance Systems Architecture",
  ["int j high risk behav addict"] = "International journal of high risk behaviors & addiction",
  ["int j high speed comput"] = "International Journal of High Speed Computing",
  ["int j high speed electron syst"] = "International Journal of High Speed Electronics and Systems",
  ["int j high throughput screen"] = "International journal of high throughput screening",
  ["int j hisp psychol"] = "International journal of Hispanic psychology",
  ["int j hist archaeol"] = "International journal of historical archaeology",
  ["int j hist eng technol"] = "International Journal for the History of Engineering & Technology",
  ["int j hist sport"] = "The International journal of the history of sport",
  ["int j hiv aids res"] = "International journal of HIV/AIDS and research",
  ["int j hortic sci"] = "International journal of horticultural science",
  ["int j hortic sci technol"] = "International journal of horticultural science and technology",
  ["int j hosp manag"] = "International journal of hospitality management",
  ["int j hous sci appl"] = "International Journal for Housing Science and Its Applications",
  ["int j hr"] = "International journal of HR : humanoid robotics",
  ["int j hum caring"] = "International journal for human caring",
  ["int j hum comput interact"] = "International journal of human-computer interaction",
  ["int j hum comput stud"] = "International journal of human-computer studies",
  ["int j hum ecol"] = "International journal of human ecology",
  ["int j hum factors ergon"] = "International journal of human factors and ergonomics",
  ["int j hum genet"] = "International journal of human genetics",
  ["int j hum genet genet disord"] = "International journal of human genetics and genetic disorders",
  ["int j hum rights healthc"] = "International journal of human rights in healthcare",
  ["int j hum-comput interact"] = "International Journal of Human-Computer Interaction",
  ["int j hum-comput stud"] = "International Journal of Human-Computer Studies",
  ["int j humanit soc sci"] = "International journal of humanities and social science",
  ["int j humanoid rob"] = "International Journal of Humanoid Robotics",
  ["int j hybrid intell syst"] = "International journal of hybrid intelligent systems",
  ["int j hybrid microelectron"] = "International Journal of Hybrid Microelectronics",
  ["int j hydrocarbon eng"] = "International Journal of Hydrocarbon Engineering",
  ["int j hydrog energy"] = "International Journal of Hydrogen Energy",
  ["int j hydrogen energy"] = "International Journal of Hydrogen Energy",
  ["int j hydrol sci technol"] = "International Journal of Hydrology Science and Technology",
  ["int j hydromechatron"] = "International Journal of Hydromechatronics",
  ["int j hydropower dams"] = "International Journal on Hydropower and Dams",
  ["int j hyg environ health"] = "International journal of hygiene and environmental health",
  ["int j hypertens"] = "International journal of hypertension",
  ["int j hyperthermia"] = "International journal of hyperthermia : the official journal of European Society for Hyperthermic Oncology, North American Hyperthermia Group",
  ["int j image data fusion"] = "International journal of image and data fusion",
  ["int j image graph"] = "International journal of image and graphics",
  ["int j image graphics"] = "International Journal of Image and Graphics",
  ["int j imaging syst technol"] = "International Journal of Imaging Systems and Technology",
  ["int j immunogenet"] = "International journal of immunogenetics",
  ["int j immunol immunother"] = "International journal of immunology and immunotherapy",
  ["int j immunol stud"] = "International journal of immunological studies",
  ["int j immunopathol pharmacol"] = "International journal of immunopathology and pharmacology",
  ["int j immunopharmacol"] = "International Journal of Immunopharmacology",
  ["int j immunother cancer res"] = "International journal of immunotherapy and cancer research",
  ["int j impact eng"] = "International journal of impact engineering",
  ["int j implant dent"] = "International journal of implant dentistry",
  ["int j impot res"] = "International journal of impotence research",
  ["int j ind chem"] = "International Journal of Industrial Chemistry",
  ["int j ind eng oper manage"] = "International Journal of Industrial Engineering and Operations Management",
  ["int j ind engtheory appl pract"] = "International Journal of Industrial Engineering: Theory Applications and Practice",
  ["int j ind ergon"] = "International journal of industrial ergonomics",
  ["int j ind organ"] = "International Journal of Industrial Organization",
  ["int j ind organiz"] = "International Journal of Industrial Organization",
  ["int j ind syst eng"] = "International Journal of Industrial and Systems Engineering",
  ["int j indig health"] = "International journal of indigenous health",
  ["int j indust entomol"] = "International journal of industrial entomology",
  ["int j inf acquis"] = "International Journal of Information Acquisition",
  ["int j inf coding theory"] = "International Journal of Information and Coding Theory. IJICOT",
  ["int j inf commun technol"] = "International Journal of Information and Communication Technology",
  ["int j inf commun technol educ"] = "International journal of information and communication technology education : an official publication of the Information Resources Management Association",
  ["int j inf comput secur"] = "International Journal of Information and Computer Security",
  ["int j inf educ technol"] = "International journal of information and education technology (IJIET)",
  ["int j inf manage"] = "International Journal of Information Management",
  ["int j inf manage data insights"] = "International Journal of Information Management Data Insights",
  ["int j inf manage sci"] = "International Journal of Information and Management Sciences",
  ["int j inf qual"] = "International Journal of Information Quality",
  ["int j inf secur"] = "International Journal of Information Security",
  ["int j inf syst change manag"] = "International journal of information systems and change management",
  ["int j inf syst change mange"] = "International Journal of Information Systems and Change Management",
  ["int j inf syst sci"] = "International Journal of Information & Systems Sciences",
  ["int j inf technol"] = "International journal of information technology : an official journal of Bharati Vidyapeeth's Institute of Computer Applications and Management",
  ["int j inf technol decis mak"] = "International journal of information technology & decision making",
  ["int j inf technol decis making"] = "International Journal of Information Technology & Decision Making",
  ["int j inf technol manage"] = "International Journal of Information Technology and Management",
  ["int j inf technol web eng"] = "International Journal of Information Technology and Web Engineering",
  ["int j infect control"] = "International journal of infection control",
  ["int j infect dis"] = "International journal of infectious diseases : IJID : official publication of the International Society for Infectious Diseases",
  ["int j inflam"] = "International journal of inflammation",
  ["int j inform management sci"] = "International Journal of Information and Management Sciences",
  ["int j infrared millimeter waves"] = "International journal of infrared and millimeter waves",
  ["int j inj contr saf promot"] = "International journal of injury control and safety promotion",
  ["int j inj control saf promot"] = "International Journal of Injury Control and Safety Promotion",
  ["int j innov appl stud"] = "International journal of innovation and applied studies",
  ["int j innov educ res"] = "International journal for innovation education and research",
  ["int j innov med health sci"] = "International journal of innovative medicine and health science",
  ["int j innov res (patuakhali)"] = "International journal of innovative research",
  ["int j innov res sci eng technol"] = "International journal of innovative research in science, engineering and technology",
  ["int j innov res stud"] = "International journal of innovative research and studies",
  ["int j innov sci res technol"] = "International journal of innovative science and research technology",
  ["int j innovation manage"] = "International Journal of Innovation Management",
  ["int j innovation sci"] = "International Journal of Innovation Science",
  ["int j innovation stud"] = "International Journal of Innovation Studies",
  ["int j innovation sustainable dev"] = "International Journal of Innovation and Sustainable Development",
  ["int j innovation technol manage"] = "International Journal of Innovation and Technology Management",
  ["int j innovative comput appl"] = "International Journal of Innovative Computing and Applications",
  ["int j inorg chem"] = "International journal of inorganic chemistry",
  ["int j inorg mater"] = "International Journal of Inorganic Materials",
  ["int j insect morphol embryol"] = "International Journal of Insect Morphology and Embryology",
  ["int j insect sci"] = "International journal of insect science",
  ["int j instr media"] = "International journal of instructional media",
  ["int j integr biol"] = "International journal of integrative biology",
  ["int j integr care"] = "International journal of integrated care",
  ["int j integr med"] = "International journal of integrative medicine",
  ["int j integr pediatr environ med"] = "International journal of integrative pediatrics and environmental medicine",
  ["int j intell comput cybern"] = "International Journal of Intelligent Computing and Cybernetics",
  ["int j intell control syst"] = "The International Journal of Intelligent Control and Systems",
  ["int j intell counterintell"] = "International Journal of Intelligence and CounterIntelligence",
  ["int j intell eng inf"] = "International Journal of Intelligent Engineering Informatics",
  ["int j intell inf database syst"] = "International Journal of Intelligent Information and Database Systems",
  ["int j intell inf technol"] = "International Journal of Intelligent Information Technologies",
  ["int j intell networks"] = "International Journal of Intelligent Networks",
  ["int j intell rob appl"] = "International Journal of Intelligent Robotics and Applications",
  ["int j intell robot appl"] = "International journal of intelligent robotics and applications",
  ["int j intell sci"] = "International journal of intelligence science",
  ["int j intell syst"] = "International Journal of Intelligent Systems",
  ["int j intell transp syst res"] = "International Journal of Intelligent Transportation Systems Research",
  ["int j interact des manuf"] = "International Journal on Interactive Design and Manufacturing",
  ["int j interact multimedia artif intell"] = "International Journal of Interactive Multimedia and Artificial Intelligence",
  ["int j intercult relat"] = "International journal of intercultural relations : IJIR",
  ["int j internet manuf serv"] = "International Journal of Internet Manufacturing and Services",
  ["int j internet protoc technol"] = "International Journal of Internet Protocol Technology",
  ["int j ion mobil spectrom"] = "International journal for ion mobility spectrometry : official publication of the International Society for Ion Mobility Spectrometry",
  ["int j ion mobility spectrom"] = "International Journal of Ion Mobility Spectrometry",
  ["int j jpn soc precis eng"] = "International Journal of the Japan Society for Precision Engineering",
  ["int j keratoconus ectatic corneal dis"] = "International journal of keratoconus and ectatic corneal diseases",
  ["int j knowl eng soft data paradig"] = "International journal of knowledge engineering and soft data paradigms",
  ["int j knowledge-based intell eng syst"] = "International Journal of Knowledge-Based and Intelligent Engineering Systems",
  ["int j lab hematol"] = "International journal of laboratory hematology",
  ["int j labor hematol"] = "International Journal of Laboratory Hematology",
  ["int j lang commun disord"] = "International journal of language & communication disorders",
  ["int j latest trends eng technol"] = "International journal of latest trends in engineering & technology : IJLTET",
  ["int j law built environ"] = "International Journal of Law in the Built Environment",
  ["int j law context"] = "International journal of law in context",
  ["int j law crime justice"] = "International journal of law, crime and justice",
  ["int j law fam"] = "International journal of law and the family",
  ["int j law policy family"] = "International journal of law, policy, and the family",
  ["int j law psychiatry"] = "International journal of law and psychiatry",
  ["int j learn chang"] = "International journal of learning and change",
  ["int j leg med"] = "International Journal of Legal Medicine",
  ["int j legal med"] = "International journal of legal medicine",
  ["int j lepr"] = "International journal of Leprosy",
  ["int j lepr other mycobact dis"] = "International journal of leprosy and other mycobacterial diseases : official organ of the International Leprosy Association",
  ["int j lib arts soc sci"] = "International journal of liberal arts and social science",
  ["int j life cycle assess"] = "International Journal of Life Cycle Assessment",
  ["int j life sci"] = "International journal of life sciences (Amravati)",
  ["int j life sci biotechnol pharma res"] = "International journal of life sciences biotechnology and pharma research",
  ["int j life sci pharma res"] = "International journal of life science & pharma research",
  ["int j life sci res"] = "International journal of life sciences research",
  ["int j life sci technol"] = "International journal of life sciences and technology",
  ["int j lightweight mater manuf"] = "International Journal of Lightweight Materials and Manufacture",
  ["int j limnol"] = "International Journal of Limnology",
  ["int j logist res appl"] = "International Journal of Logistics Research and Applications",
  ["int j low carbon technol"] = "International Journal of Low Carbon Technologies",
  ["int j low extrem wounds"] = "The international journal of lower extremity wounds",
  ["int j low radiat"] = "International journal of low radiation",
  ["int j low-carbon technol"] = "International Journal of Low-Carbon Technologies",
  ["int j mach conscious"] = "International Journal of Machine Consciousness",
  ["int j mach learn comput"] = "International journal of machine learning and computing",
  ["int j mach learn cybern"] = "International Journal of Machine Learning and Cybernetics",
  ["int j mach mach mater"] = "International Journal of Machining and Machinability of Materials",
  ["int j mach tool des res"] = "International Journal of Machine Tool Design and Research",
  ["int j mach tools manuf"] = "International Journal of Machine Tools and Manufacture",
  ["int j magn"] = "International Journal of Magnetism",
  ["int j magn part imaging"] = "International journal on magnetic particle imaging",
  ["int j malacol"] = "International Journal of Malacology",
  ["int j man mach stud"] = "International Journal of Man-Machine Studies",
  ["int j manag bus"] = "International journal of management and business",
  ["int j manage sci eng manage"] = "International Journal of Management Science and Engineering Management",
  ["int j manage sustainability"] = "International Journal of Management and Sustainability",
  ["int j manpow"] = "International journal of manpower",
  ["int j manpower"] = "International Journal of Manpower",
  ["int j manuf mater mech eng"] = "International Journal of Manufacturing, Materials, and Mechanical Engineering",
  ["int j manuf res"] = "International Journal of Manufacturing Research",
  ["int j manuf technol manage"] = "International Journal of Manufacturing Technology and Management",
  ["int j maps math"] = "International Journal of Maps in Mathematics",
  ["int j mar energy"] = "International Journal of Marine Energy",
  ["int j mar sci"] = "International Journal of Marine Science",
  ["int j marit eng"] = "International Journal of Maritime Engineering",
  ["int j marit hist"] = "International journal of maritime history",
  ["int j maritime econ"] = "International Journal of Maritime Economics",
  ["int j mark stud"] = "International journal of marketing studies",
  ["int j masonry res innov"] = "International Journal of Masonry Research and Innovation",
  ["int j mass emerg disasters"] = "International journal of mass emergencies and disasters",
  ["int j mass spectrom"] = "International Journal of Mass Spectrometry",
  ["int j mass spectrom ion phys"] = "International Journal of Mass Spectrometry and Ion Physics",
  ["int j mass spectrom ion process"] = "International journal of mass spectrometry and ion processes",
  ["int j mass spectrom ion processes"] = "International Journal of Mass Spectrometry and Ion Processes",
  ["int j mater eng appl"] = "International Journal of Materials in Engineering Applications",
  ["int j mater eng innovation"] = "International Journal of Materials Engineering Innovation",
  ["int j mater form"] = "International Journal of Material Forming",
  ["int j mater prod technol"] = "International Journal of Materials and Product Technology",
  ["int j mater res"] = "International Journal of Materials Research",
  ["int j mater struct integrity"] = "International Journal of Materials and Structural Integrity",
  ["int j math"] = "International Journal of Mathematics",
  ["int j math anal (ns)"] = "International Journal of Mathematics and Analysis. New Series",
  ["int j math comput phys electr comput eng"] = "International Journal of Mathematical, Computational, Physical, Electrical and Computer Engineering",
  ["int j math comput sci"] = "International journal of mathematics and computer science",
  ["int j math ed sci tech"] = "International Journal of Mathematical Education in Science and Technology",
  ["int j math educ sci technol"] = "International Journal of Mathematical Education in Science and Technology",
  ["int j math game theory algebra"] = "International Journal of Mathematics, Game Theory, and Algebra",
  ["int j math ind"] = "International Journal of Mathematics for Industry",
  ["int j math math sci"] = "International Journal of Mathematics and Mathematical Sciences",
  ["int j math oper res"] = "International Journal of Mathematics in Operational Research",
  ["int j math stat"] = "International Journal of Mathematics and Statistics",
  ["int j math stat sci"] = "International Journal of Mathematical and Statistical Sciences",
  ["int j mch aids"] = "International journal of MCH and AIDS",
  ["int j mech eng educ"] = "International Journal of Mechanical Engineering Education",
  ["int j mech mater des"] = "International Journal of Mechanics and Materials in Design",
  ["int j mech mater eng"] = "International Journal of Mechanical and Materials Engineering",
  ["int j mech sci"] = "International Journal of Mechanical Sciences",
  ["int j mech syst dyn"] = "International Journal of Mechanical Systems Dynamics",
  ["int j mech syst sci eng"] = "International Journal of Mechanical Systems Science and Engineering",
  ["int j mechatron manuf syst"] = "International Journal of Mechatronics and Manufacturing Systems",
  ["int j med (dubai)"] = "International journal of medicine",
  ["int j med biochem"] = "International Journal of Medical Biochemistry",
  ["int j med biol front"] = "International journal of medical and biological frontiers",
  ["int j med biomed sci"] = "International journal of medical and biomedical sciences",
  ["int j med biomed stud"] = "International journal of medical and biomedical studies",
  ["int j med chem"] = "International journal of medicinal chemistry",
  ["int j med educ"] = "International journal of medical education",
  ["int j med eng inf"] = "International Journal of Medical Engineering and Informatics",
  ["int j med eng inform"] = "International journal of medical engineering and informatics",
  ["int j med inf"] = "International Journal of Medical Informatics",
  ["int j med inform"] = "International journal of medical informatics",
  ["int j med informatics"] = "International Journal of Medical Informatics",
  ["int j med law"] = "The International journal of medicine and law : IJML",
  ["int j med microbiol"] = "International journal of medical microbiology : IJMM",
  ["int j med microbiol virol parasitol infect dis"] = "International Journal of Medical Microbiology, Virology, Parasitology and Infectious Diseases",
  ["int j med mol adv sci"] = "International Journal of Molecular Medicine and Advance Sciences",
  ["int j med mushrooms"] = "International journal of medicinal mushrooms",
  ["int j med nano res"] = "International journal of medical nano research",
  ["int j med pharm case reports"] = "International journal of medical and pharmaceutical case reports",
  ["int j med phys clin eng radiat oncol"] = "International journal of medical physics, clinical engineering and radiation oncology",
  ["int j med public health"] = "International journal of medicine and public health",
  ["int j med rob comput assisted surg"] = "International Journal of Medical Robotics and Computer Assisted Surgery",
  ["int j med robot"] = "The international journal of medical robotics + computer assisted surgery : MRCAS",
  ["int j med sci"] = "International journal of medical sciences",
  ["int j med sci clin invent"] = "International journal of medical science and clinical invention",
  ["int j med sci public health"] = "International journal of medical science and public health",
  ["int j med toxicol forensic med"] = "International Journal Of Medical Toxicology And Forensic Medicine",
  ["int j med toxicol legal med"] = "International Journal Of Medical Toxicology And Legal Medicine",
  ["int j mens health"] = "International journal of men's health",
  ["int j ment health"] = "International journal of mental health",
  ["int j ment health addict"] = "International journal of mental health and addiction",
  ["int j ment health nurs"] = "International journal of mental health nursing",
  ["int j ment health promot"] = "The international journal of mental health promotion",
  ["int j ment health psychiatry"] = "International journal of mental health & psychiatry",
  ["int j ment health syst"] = "International journal of mental health systems",
  ["int j metadata semant ontol"] = "International Journal of Metadata, Semantics and Ontologies",
  ["int j metalcast"] = "International Journal of Metalcasting",
  ["int j methods psychiatr res"] = "International journal of methods in psychiatric research",
  ["int j metrol qual eng"] = "International Journal of Metrology and Quality Engineering",
  ["int j micro air veh"] = "International Journal of Micro Air Vehicles",
  ["int j microbiol"] = "International journal of microbiology",
  ["int j microbiol biotechnol"] = "International journal of microbiology and biotechnology",
  ["int j microbiol curr res"] = "International journal of microbiology and current research",
  ["int j microbiol immunol res"] = "International journal of microbiology and immunology research",
  ["int j microbiol res"] = "International journal of microbiology research",
  ["int j microcirc clin exp"] = "International journal of microcirculation, clinical and experimental",
  ["int j microcircuits electron packag"] = "International Journal of Microcircuits and Electronic Packaging",
  ["int j micronano scale transp"] = "International journal of micro-nano scale transport",
  ["int j microsimul"] = "International Journal of Microsimulation",
  ["int j microstruct mater prop"] = "International Journal of Microstructure and Materials Properties",
  ["int j microw wirel technol"] = "International journal of microwave and wireless technologies",
  ["int j microwave wireless technolog"] = "International Journal of Microwave and Wireless Technologies",
  ["int j middle east stud"] = "International journal of Middle East studies",
  ["int j migr bord stud"] = "International journal of migration and border studies",
  ["int j migr health soc care"] = "International journal of migration, health, and social care",
  ["int j min geo-eng"] = "International Journal Of Mining And Geo-Engineering",
  ["int j min miner eng"] = "International Journal of Mining and Mineral Processing Engineering",
  ["int j min reclam environ"] = "International Journal of Mining, Reclamation and Environment",
  ["int j min sci technol"] = "International journal of mining science and technology",
  ["int j mine water"] = "International Journal of Mine Water",
  ["int j miner metall mater"] = "International Journal of Minerals, Metallurgy, and Materials",
  ["int j miner process"] = "International Journal of Mineral Processing",
  ["int j mobile commun"] = "International Journal of Mobile Communications",
  ["int j mobile network des innov"] = "International Journal of Mobile Network Design and Innovation",
  ["int j mod chem"] = "International Journal of Modern Chemistry",
  ["int j mod org chem"] = "International Journal of Modern Organic Chemistry",
  ["int j mod phys a"] = "International journal of modern physics. A, Particles and fields, gravitation, cosmology",
  ["int j mod phys b"] = "International Journal of Modern Physics B: Condensed Matter Physics, Statistical Physics, Applied Physics",
  ["int j mod phys c"] = "International Journal of Modern Physics C: Computational Physics and Physical Computation",
  ["int j mod phys conf ser"] = "International journal of modern physics. Conference series",
  ["int j mod phys d"] = "International Journal of Modern Physics D: Gravitation; Astrophysics and Cosmology",
  ["int j mod phys d gravit astrophys cosmol"] = "International journal of modern physics. D, Gravitation, astrophysics and cosmology",
  ["int j mod phys e"] = "International Journal of Modern Physics E: Nuclear Physics",
  ["int j mod sci technol (kumbakonam)"] = "International journal of modern science and technology (Kumbakonam)",
  ["int j model ident control"] = "International Journal of Modelling, Identification and Control",
  ["int j model identif control"] = "International journal of modelling identification and control",
  ["int j model simul"] = "International Journal of Modelling and Simulation",
  ["int j model, simul, sci comput"] = "International Journal of Modeling, Simulation, and Scientific Computing",
  ["int j modell identif control"] = "International Journal of Modelling, Identification and Control",
  ["int j modell simul"] = "International Journal of Modelling and Simulation",
  ["int j modern phys a"] = "International Journal of Modern Physics A",
  ["int j modern phys b"] = "International Journal of Modern Physics B",
  ["int j modern phys c"] = "International Journal of Modern Physics C",
  ["int j modern phys conf ser"] = "International Journal of Modern Physics: Conference Series",
  ["int j modern phys d"] = "International Journal of Modern Physics",
  ["int j modern phys e"] = "International Journal of Modern Physics",
  ["int j mol biol (edmond)"] = "International journal of molecular biology (Edmond, Okla.)",
  ["int j mol cell med"] = "International journal of molecular and cellular medicine",
  ["int j mol clin microbiol"] = "International journal of molecular and clinical microbiology",
  ["int j mol epidemiol genet"] = "International journal of molecular epidemiology and genetics",
  ["int j mol imaging"] = "International journal of molecular imaging",
  ["int j mol med"] = "International journal of molecular medicine",
  ["int j mol med adv sci"] = "International Journal of Molecular Medicine and Advance Sciences",
  ["int j mol med sci"] = "International journal of molecular medical science",
  ["int j mol sci"] = "International journal of molecular sciences",
  ["int j monit surveill technol res"] = "International journal of monitoring and surveillance technologies research",
  ["int j morphol"] = "International journal of morphology = Revista internacional de morfología",
  ["int j mosq res"] = "International journal of mosquito research",
  ["int j ms care"] = "International journal of MS care",
  ["int j mult res approaches"] = "International journal of multiple research approaches",
  ["int j multiling"] = "International journal of multilingualism",
  ["int j multimed inf retr"] = "International journal of multimedia information retrieval",
  ["int j multiph flow"] = "International Journal of Multiphase Flow",
  ["int j multiphase flow"] = "International Journal of Multiphase Flow",
  ["int j multiphys"] = "International Journal of Multiphysics",
  ["int j multiscale comput eng"] = "International Journal for Multiscale Computational Engineering",
  ["int j mycobacteriol"] = "International journal of mycobacteriology",
  ["int j myriap"] = "International journal of myriapodology",
  ["int j nano biomater"] = "International Journal of Nano and Biomaterials",
  ["int j nano stud technol"] = "International journal of nano studies & technology",
  ["int j nanodimens"] = "International journal of nanodimension",
  ["int j nanoelectron mater"] = "International Journal Of Nanoelectronics And Materials",
  ["int j nanomanuf"] = "International Journal of Nanomanufacturing",
  ["int j nanomed"] = "International Journal of Nanomedicine",
  ["int j nanomed nanosurg"] = "International journal of nanomedicine and nanosurgery",
  ["int j nanomedicine"] = "International journal of nanomedicine",
  ["int j nanopart"] = "International Journal Of Nanoparticles",
  ["int j nanosci"] = "International Journal of Nanoscience",
  ["int j nanosci nanotechnol"] = "International Journal of Nanoscience and Nanotechnology",
  ["int j nanotechnol"] = "International Journal of Nanotechnology",
  ["int j nanotechnol appl"] = "International journal of nanotechnology and applications",
  ["int j nanotechnol eng med"] = "International journal of nanotechnology in medicine & engineering",
  ["int j nanotechnol nanomed"] = "International journal of nanotechnology and nanomedicine",
  ["int j nat eng sci"] = "International Journal of Natural and Engineering Sciences",
  ["int j naut archaeol"] = "International Journal of Nautical Archaeology",
  ["int j nav archit ocean eng"] = "International Journal of Naval Architecture and Ocean Engineering",
  ["int j nematol"] = "International journal of nematology",
  ["int j neonatal screen"] = "International journal of neonatal screening",
  ["int j nephrol"] = "International journal of nephrology",
  ["int j nephrol kidney fail"] = "International journal of nephrology and kidney failure",
  ["int j nephrol renovasc dis"] = "International journal of nephrology and renovascular disease",
  ["int j nephrol urol"] = "International journal of nephrology & urology",
  ["int j network manage"] = "International Journal of Network Management",
  ["int j networked distrib comput"] = "International Journal of Networked and Distributed Computing",
  ["int j networking virtual organ"] = "International Journal of Networking and Virtual Organisations",
  ["int j neural syst"] = "International Journal of Neural Systems",
  ["int j neurodegener dis"] = "International journal of neurodegenerative disorders",
  ["int j neurol"] = "International journal of neurology",
  ["int j neurol brain disord"] = "International journal of neurology & brain disorders",
  ["int j neurol neurother"] = "International journal of neurology and neurotherapy",
  ["int j neuropharmacol"] = "International journal of neuropharmacology",
  ["int j neuroprot neuroregener"] = "International journal of neuroprotection and neuroregeneration",
  ["int j neuropsychiatry"] = "International journal of neuropsychiatry",
  ["int j neuropsychopharmacol"] = "The international journal of neuropsychopharmacology",
  ["int j neurorehabil"] = "International journal of neurorehabilitation",
  ["int j neurosci"] = "The International journal of neuroscience",
  ["int j neurosci neuroinf"] = "International Journal of Neuroscience and Neuroinformatics",
  ["int j non equilibr process"] = "International Journal of Non-Equilibrium Processing",
  ["int j non linear mech"] = "International Journal of Non-Linear Mechanics",
  ["int j non-linear mech"] = "International Journal of Non-Linear Mechanics",
  ["int j noncommun dis"] = "International journal of noncommunicable diseases",
  ["int j nondestr test"] = "International Journal of Nondestructive Testing",
  ["int j nonferrous metall"] = "International Journal of Nonferrous Metallurgy",
  ["int j nonlinear model sci eng"] = "International Journal of Nonlineare Modelling in Science and Engineering",
  ["int j nonlinear modell sci eng"] = "International Journal of Nonlinear Modelling in Science and Engineering",
  ["int j nonlinear sci"] = "International Journal of Nonlinear Science",
  ["int j nonlinear sci numer simul"] = "International Journal of Nonlinear Sciences and Numerical Simulation",
  ["int j nucl desalin"] = "International Journal of Nuclear Desalination",
  ["int j nucl energy sci technol"] = "International Journal of Nuclear Energy Science and Technology",
  ["int j nucl med biol"] = "International Journal of Nuclear Medicine and Biology",
  ["int j number theory"] = "International Journal of Number Theory",
  ["int j numer anal methods geomech"] = "International Journal for Numerical and Analytical Methods in Geomechanics",
  ["int j numer anal model"] = "International Journal of Numerical Analysis and Modeling",
  ["int j numer anal model b"] = "International journal of numerical analysis & modeling. Series B",
  ["int j numer anal model ser b"] = "International Journal of Numerical Analysis and Modeling. Series B",
  ["int j numer method biomed eng"] = "International journal for numerical methods in biomedical engineering",
  ["int j numer methods biomed eng"] = "International Journal of Numerical Methods in Biomedical Engineering",
  ["int j numer methods eng"] = "International Journal for Numerical Methods in Engineering",
  ["int j numer methods fluids"] = "International Journal for Numerical Methods in Fluids",
  ["int j numer methods heat fluid flow"] = "International Journal of Numerical Methods for Heat and Fluid Flow",
  ["int j numer model electron networks devices fields"] = "International Journal of Numerical Modelling: Electronic Networks, Devices and Fields",
  ["int j numer modell electron networks devices fields"] = "International Journal of Numerical Modelling: Electronic Networks, Devices and Fields",
  ["int j nurs (n y)"] = "International journal of nursing (New York, N.Y.)",
  ["int j nurs clin pract"] = "International journal of nursing & clinical practices",
  ["int j nurs educ scholarsh"] = "International journal of nursing education scholarship",
  ["int j nurs knowl"] = "International journal of nursing knowledge",
  ["int j nurs pract"] = "International journal of nursing practice",
  ["int j nurs res"] = "International journal of nursing research",
  ["int j nurs sci"] = "International journal of nursing sciences",
  ["int j nurs stud"] = "International journal of nursing studies",
  ["int j nurs stud adv"] = "International journal of nursing studies advances",
  ["int j nurs terminol classif"] = "International journal of nursing terminologies and classifications : the official journal of NANDA International",
  ["int j nutr"] = "International journal of nutrition",
  ["int j nutr metab"] = "International journal of nutrition and metabolism",
  ["int j nutr pharmacol neurol dis"] = "International journal of nutrition, pharmacology, neurological diseases",
  ["int j obes"] = "International journal of obesity",
  ["int j obes (lond)"] = "International journal of obesity (2005)",
  ["int j obes nutr sci"] = "International journal of obesity and nutritional science",
  ["int j obes relat metab disord"] = "International journal of obesity and related metabolic disorders : journal of the International Association for the Study of Obesity",
  ["int j obes suppl"] = "International journal of obesity supplements",
  ["int j obstet anesth"] = "International journal of obstetric anesthesia",
  ["int j occup environ health"] = "International journal of occupational and environmental health",
  ["int j occup environ med"] = "The international journal of occupational and environmental medicine",
  ["int j occup health saf"] = "The International journal of occupational health & safety",
  ["int j occup med environ health"] = "International journal of occupational medicine and environmental health",
  ["int j occup saf ergon"] = "International journal of occupational safety and ergonomics : JOSE",
  ["int j occup saf health"] = "International journal of occupational safety and health",
  ["int j ocean coastal eng"] = "International Journal of Ocean and Coastal Engineering",
  ["int j oceanol limnol"] = "International Journal of Oceanology and Limnology",
  ["int j odonatol"] = "International journal of odonatology : official organ of the Worldwide Dragonfly Association",
  ["int j offender ther comp criminol"] = "International journal of offender therapy and comparative criminology",
  ["int j offshore polar eng"] = "International Journal of Offshore and Polar Engineering",
  ["int j oil gas coal technol"] = "International Journal of Oil, Gas and Coal Technology",
  ["int j older people nurs"] = "International journal of older people nursing",
  ["int j oncol"] = "International Journal of Oncology",
  ["int j one health"] = "International journal of one health",
  ["int j oper prod manage"] = "International Journal of Operations & Production Management",
  ["int j oper res"] = "International Journal of Operational Research",
  ["int j oper res (taichung)"] = "International Journal of Operations Research",
  ["int j oper res nepal"] = "International Journal of Operational Research Nepal. IJORN",
  ["int j ophthalmic pathol"] = "International Journal of Ophthalmic Pathology",
  ["int j ophthalmic res"] = "International journal of ophthalmic research",
  ["int j ophthalmol"] = "International journal of ophthalmology",
  ["int j ophthalmol clin res"] = "International journal of ophthalmology and clinical research",
  ["int j ophthalmol eye sci"] = "International journal of ophthalmology & eye science",
  ["int j opt"] = "International journal of optics",
  ["int j optim control theor appl ijocta"] = "An International Journal of Optimization and Control. Theories & Applications. IJOCTA",
  ["int j optim theory methods appl"] = "International Journal of Optimization. Theory Methods and Applications",
  ["int j optomechatronics"] = "International Journal of Optomechatronics",
  ["int j oral biol"] = "International journal of oral biology : official journal of the Korean Academy of Oral Biology and the UCLA Dental Research Institute",
  ["int j oral implantol"] = "The International journal of oral implantology : implantologist",
  ["int j oral implantol (berl)"] = "International journal of oral implantology (Berlin, Germany)",
  ["int j oral maxillofac implants"] = "The International journal of oral & maxillofacial implants",
  ["int j oral maxillofac surg"] = "International journal of oral and maxillofacial surgery",
  ["int j oral myol"] = "The International journal of oral myology",
  ["int j oral sci"] = "International journal of oral science",
  ["int j oral surg"] = "International journal of oral surgery",
  ["int j org chem"] = "International Journal of Organic Chemistry",
  ["int j org chem (irvine)"] = "International journal of organic chemistry",
  ["int j organ transplant med"] = "International journal of organ transplantation medicine",
  ["int j orofacial myology"] = "The International journal of orofacial myology : official publication of the International Association of Orofacial Myology",
  ["int j orthod"] = "International journal of orthodontics",
  ["int j orthod milwaukee"] = "International journal of orthodontics (Milwaukee, Wis.)",
  ["int j orthop"] = "International journal of orthopedics",
  ["int j orthop (hong kong)"] = "International journal of orthopaedics (Hong Kong)",
  ["int j orthop trauma nurs"] = "International journal of orthopaedic and trauma nursing",
  ["int j osteoarchaeol"] = "International journal of osteoarchaeology",
  ["int j osteopath med"] = "International journal of osteopathic medicine : IJOM",
  ["int j otolaryngol"] = "International journal of otolaryngology",
  ["int j otorhinolaryngol"] = "International journal of otorhinolaryngology",
  ["int j paediatr dent"] = "International journal of paediatric dentistry",
  ["int j paleopathol"] = "International journal of paleopathology",
  ["int j palliat nurs"] = "International journal of palliative nursing",
  ["int j pancreatol"] = "International journal of pancreatology : official journal of the International Association of Pancreatology",
  ["int j parallel emergent distrib syst"] = "International Journal of Parallel, Emergent and Distributed Systems",
  ["int j parallel program"] = "International Journal of Parallel Programming",
  ["int j parasitol"] = "International journal for parasitology",
  ["int j parasitol drugs drug resist"] = "International journal for parasitology. Drugs and drug resistance",
  ["int j parasitol parasites wildl"] = "International journal for parasitology. Parasites and wildlife",
  ["int j parasitol: drugs drug resist"] = "International Journal of Parasitology: Drugs and Drug Resistance",
  ["int j parasitol: parasites wildl"] = "International Journal of Parasitology: Parasites and Wildlife",
  ["int j part ther"] = "International journal of particle therapy",
  ["int j partial hosp"] = "International journal of partial hospitalization",
  ["int j pattern recognit artif intell"] = "International Journal of Pattern Recognition and Artificial Intelligence",
  ["int j pavement eng"] = "International Journal of Pavement Engineering",
  ["int j pavement res technol"] = "International Journal of Pavement Research and Technology",
  ["int j pediatr"] = "International journal of pediatrics",
  ["int j pediatr adolesc med"] = "International journal of pediatrics & adolescent medicine",
  ["int j pediatr child health"] = "International journal of pediatrics and child health",
  ["int j pediatr endocrinol"] = "International journal of pediatric endocrinology",
  ["int j pediatr nephrol"] = "The International journal of pediatric nephrology",
  ["int j pediatr obes"] = "International journal of pediatric obesity : IJPO : an official journal of the International Association for the Study of Obesity",
  ["int j pediatr otorhinolaryngol"] = "International journal of pediatric otorhinolaryngology",
  ["int j pediatr otorhinolaryngol extra"] = "International journal of pediatric otorhinolaryngology extra",
  ["int j pept"] = "International journal of peptides",
  ["int j pept protein res"] = "International Journal of Peptide and Protein Research",
  ["int j pept res ther"] = "International Journal of Peptide Research and Therapeutics",
  ["int j perform anal sport"] = "International journal of performance analysis in sport",
  ["int j perform arts digit media"] = "International journal of performance arts and digital media",
  ["int j periodontics restorative dent"] = "The International journal of periodontics & restorative dentistry",
  ["int j pers cent med"] = "International journal of person centered medicine",
  ["int j personal psychol"] = "International journal of personality psychology",
  ["int j pervasive comput commun"] = "International Journal of Pervasive Computing and Communications",
  ["int j pest manag"] = "International journal of pest management",
  ["int j pest manage"] = "International Journal of Pest Management",
  ["int j pet sci technol"] = "International journal of petroleum science and technology",
  ["int j pharm"] = "International journal of pharmaceutics",
  ["int j pharm (amsterdam, neth)"] = "International Journal of Pharmaceutics (Amsterdam, Netherlands)",
  ["int j pharm biol arch"] = "International journal of pharmaceutical and biological archive",
  ["int j pharm biol sci"] = "International journal of pharmacy and biological sciences",
  ["int j pharm compd"] = "International journal of pharmaceutical compounding",
  ["int j pharm healthc mark"] = "International journal of pharmaceutical and healthcare marketing",
  ["int j pharm investig"] = "International journal of pharmaceutical investigation",
  ["int j pharm pharm res"] = "International journal of pharmacy and pharmaceutical research",
  ["int j pharm pharm sci"] = "International Journal of Pharmacy and Pharmaceutical Sciences",
  ["int j pharm pract"] = "The International journal of pharmacy practice",
  ["int j pharm sci invent"] = "International journal of pharmaceutical science invention",
  ["int j pharm sci nanotechnol"] = "International Journal of Pharmaceutical Sciences and Nanotechnology",
  ["int j pharm sci res"] = "International journal of pharmaceutical sciences and research",
  ["int j pharm sci rev res"] = "International journal of pharmaceutical sciences review and research",
  ["int j pharm technol prod manuf"] = "International Journal of Pharmaceutical Technology and Product Manufacture",
  ["int j pharm x"] = "International journal of pharmaceutics: X",
  ["int j pharm: x"] = "International Journal of Pharmaceutics: X",
  ["int j pharma bio sci"] = "International journal of pharma and bio sciences",
  ["int j pharma inf j biotechnol biother"] = "International journal of pharma informa : journal of biotechnology and biotherapeutics",
  ["int j pharma res rev"] = "International journal of pharma research & review",
  ["int j pharma sci"] = "International journal of pharma sciences",
  ["int j pharmacokinet"] = "International journal of pharmacokinetics",
  ["int j pharmacol"] = "International journal of pharmacology : IJP",
  ["int j pharmacol toxicol"] = "International journal of phramcology and toxicology",
  ["int j pharmtech res"] = "International journal of pharmtech research",
  ["int j philipp sci technol"] = "International journal of Philippine science and technology",
  ["int j philos stud"] = "International journal of philosophical studies",
  ["int j photoenergy"] = "International Journal of Photoenergy",
  ["int j phycol phycochem"] = "International Journal of Phycology and Phycochemistry",
  ["int j phys med rehabil"] = "International journal of physical medicine & rehabilitation",
  ["int j physiol"] = "International journal of physiology",
  ["int j physiol pathophysiol"] = "International Journal of Physiology and Pathophysiology",
  ["int j physiol pathophysiol pharmacol"] = "International journal of physiology, pathophysiology and pharmacology",
  ["int j phytorem"] = "International Journal of Phytoremediation",
  ["int j phytoremediation"] = "International journal of phytoremediation",
  ["int j pixe"] = "International Journal of PIXE",
  ["int j plant biol res"] = "International journal of plant biology & research",
  ["int j plant genomics"] = "International journal of plant genomics",
  ["int j plant pathol"] = "International journal of plant pathology",
  ["int j plant prod"] = "International journal of plant production",
  ["int j plant prot"] = "International journal of plant protection",
  ["int j plant sci"] = "International journal of plant sciences",
  ["int j plant soil sci"] = "International journal of plant & soil science",
  ["int j plasma sci eng"] = "International Journal of Plasma Science and Engineering",
  ["int j plast"] = "International Journal of Plasticity",
  ["int j plast technol"] = "International Journal of Plastics Technology",
  ["int j play"] = "International journal of play",
  ["int j polit"] = "International journal of politics",
  ["int j polit cult soc"] = "International journal of politics, culture, and society",
  ["int j pollut res"] = "International journal of pollution research",
  ["int j polym anal charact"] = "International Journal of Polymer Analysis and Characterization",
  ["int j polym mater"] = "International Journal of Polymeric Materials",
  ["int j polym mater polym biomater"] = "International Journal of Polymeric Materials and Polymeric Biomaterials",
  ["int j polym sci"] = "International journal of polymer science",
  ["int j popul data sci"] = "International journal of population data science",
  ["int j popul geogr"] = "International journal of population geography : IJPG",
  ["int j popul res"] = "International journal of population research",
  ["int j popul stud"] = "International journal of population studies",
  ["int j poult sci"] = "International journal of poultry science",
  ["int j powder metall"] = "International Journal of Powder Metallurgy",
  ["int j powder metall (princeton, n j)"] = "International Journal of Powder Metallurgy (Princeton, New Jersey)",
  ["int j power energy syst"] = "International Journal of Power and Energy Systems",
  ["int j precis eng"] = "International Journal Precision Engineering",
  ["int j precis eng manuf"] = "International Journal of Precision Engineering and Manufacturing",
  ["int j precis eng manuf green technol"] = "International Journal of Precision Engineering and Manufacturing - Green Technology",
  ["int j press polit"] = "The international journal of press/politics",
  ["int j press vessels pip"] = "International Journal of Pressure Vessels and Piping",
  ["int j prev med"] = "International journal of preventive medicine",
  ["int j primatol"] = "International journal of primatology",
  ["int j prison health"] = "International journal of prisoner health",
  ["int j priv health inf manag"] = "International journal of privacy and health information management",
  ["int j probiotics prebiotics"] = "International journal of probiotics & prebiotics",
  ["int j prod econ"] = "International Journal of Production Economics",
  ["int j prod lifecycle manag"] = "International journal of product lifecycle management",
  ["int j prod res"] = "International Journal of Production Research",
  ["int j product econ"] = "International Journal of Production Economics",
  ["int j progn health manag"] = "International journal of prognostics and health management",
  ["int j project manage"] = "International Journal of Project Management",
  ["int j prosthodont"] = "The International journal of prosthodontics",
  ["int j prot struct"] = "International Journal of Protective Structures",
  ["int j protein res"] = "International journal of protein research",
  ["int j proteomics"] = "International journal of proteomics",
  ["int j psychiatr nurs res"] = "The international journal of psychiatric nursing research",
  ["int j psychiatry"] = "International journal of psychiatry",
  ["int j psychiatry (overl park)"] = "International journal of psychiatry (Overland Park)",
  ["int j psychiatry clin pract"] = "International journal of psychiatry in clinical practice",
  ["int j psychiatry med"] = "International journal of psychiatry in medicine",
  ["int j psychoanal"] = "The International journal of psycho-analysis",
  ["int j psychoanal psychother"] = "International journal of psychoanalytic psychotherapy",
  ["int j psychoanal self psychol"] = "International journal of psychoanalytic self psychology",
  ["int j psychol"] = "International journal of psychology : Journal international de psychologie",
  ["int j psychol behav sci"] = "International journal of psychology and behavioral sciences",
  ["int j psychol neurosci"] = "International journal of psychology and neuroscience",
  ["int j psychol psychoanal"] = "International journal of psychology and psychoanalysis",
  ["int j psychol relig"] = "The International journal for the psychology of religion",
  ["int j psychol res (medellin)"] = "International journal of psychological research",
  ["int j psychol stud"] = "International journal of psychological studies",
  ["int j psychophysiol"] = "International journal of psychophysiology : official journal of the International Organization of Psychophysiology",
  ["int j psychosom"] = "International journal of psychosomatics : official publication of the International Psychosomatics Institute",
  ["int j publ health sci"] = "International journal of public health science",
  ["int j public health"] = "International journal of public health",
  ["int j public health res"] = "International journal of public health research",
  ["int j public ment health neurosci"] = "International journal of public mental health and neurosciences",
  ["int j public opin res"] = "International journal of public opinion research",
  ["int j public pol"] = "International journal of public policy",
  ["int j punjab stud"] = "International journal of Punjab studies",
  ["int j pure appl biosci"] = "International journal of pure and applied bioscience",
  ["int j pure appl math"] = "International journal of pure and applied mathematics : IJPAM",
  ["int j pure appl sci technol"] = "International journal of pure and applied sciences and technology",
  ["int j qual health care"] = "International journal for quality in health care : journal of the International Society for Quality in Health Care",
  ["int j qual methods"] = "International journal of qualitative methods",
  ["int j qual reliab manage"] = "International Journal of Quality & Reliability Management",
  ["int j qual sci"] = "International Journal of Quality Science",
  ["int j qual stud educ"] = "International journal of qualitative studies in education : QSE",
  ["int j qual stud health well-being"] = "International journal of qualitative studies on health and well-being",
  ["int j quant res educ"] = "International journal of quantitative research in education",
  ["int j quantum chem"] = "International Journal of Quantum Chemistry",
  ["int j quantum chem 1"] = "International Journal of Quantum Chemistry, Part 1",
  ["int j quantum chem 2"] = "International Journal of Quantum Chemistry, Part 2",
  ["int j quantum chem quantum biol symp"] = "International journal of quantum chemistry. Quantum biology symposium : proceedings of the International Symposium on Quantum Biology and Quantum Pharmacology. International Symposium on Quantum Biology and Quantum Pharmacology",
  ["int j quantum inf"] = "International Journal of Quantum Information",
  ["int j rad appl instrum [a]"] = "International Journal of Radiation Applications and Instrumentation. Part A, Applied Radiation and Isotopes",
  ["int j rad appl instrum a"] = "International journal of radiation applications and instrumentation. Part A, Applied radiation and isotopes",
  ["int j rad appl instrum b"] = "International journal of radiation applications and instrumentation. Part B, Nuclear medicine and biology",
  ["int j rad appl instrum d"] = "International Journal Of Radiation Applications And Instrumentation. Part D, Nuclear Tracks And Radiation Measurements",
  ["int j radiat appl instrum part a"] = "International Journal of Radiation Applications and Instrumentation. Part A. Applied Radiation and Isotopes",
  ["int j radiat appl instrum part b"] = "International Journal of Radiation Applications and Instrumentation. Part B. Nuclear Medicine and Biology",
  ["int j radiat appl instrum part c"] = "International Journal of Radiation Applications and Instrumentation. Part C. Radiation Physics and Chemistry",
  ["int j radiat appl instrum part d"] = "International Journal of Radiation Applications and Instrumentation. Part D. Nuclear Tracks and Radiation Measurements",
  ["int j radiat appl instrum part e"] = "International Journal of Radiation Applications and Instrumentation. Part E. Nuclear Geophysics",
  ["int j radiat biol"] = "International Journal of Radiation Biology",
  ["int j radiat biol relat stud phys chem med"] = "International journal of radiation biology and related studies in physics, chemistry, and medicine",
  ["int j radiat med"] = "International Journal of Radiation Medicine",
  ["int j radiat oncol biol phys"] = "International journal of radiation oncology, biology, physics",
  ["int j radiat phys"] = "International Journal of Radiation Physics",
  ["int j radiat phys chem"] = "International Journal for Radiation Physics and Chemistry",
  ["int j radio freq identif technol appl"] = "International Journal of Radio Frequency Identification Technology and Applications",
  ["int j rail transp"] = "International Journal of Rail Transportation",
  ["int j recent sci res"] = "International journal of recent scientific research",
  ["int j reconfigurable comput"] = "International Journal of Reconfigurable Computing",
  ["int j recyl org waste agric"] = "International Journal of Recycling of Organic Waste in Agriculture",
  ["int j refract hard met"] = "International journal of refractory & hard metals",
  ["int j refract met hard mater"] = "International Journal of Refractory Metals and Hard Materials",
  ["int j refract metals hard mater"] = "International journal of refractory metals & hard materials",
  ["int j refrig"] = "International Journal of Refrigeration",
  ["int j rehabil res"] = "International Journal of Rehabilitation Research",
  ["int j reliab qual saf eng"] = "International Journal of Reliability, Quality and Safety Engineering",
  ["int j reliab saf"] = "International Journal of Reliability and Safety",
  ["int j religion spiritual soc"] = "The international journal of religion and spirituality in society",
  ["int j remote sens"] = "International Journal of Remote Sensing",
  ["int j renewable energy dev"] = "International Journal of Renewable Energy Development",
  ["int j renewable energy res"] = "International Journal of Renewable Energy Research",
  ["int j reprod biomed"] = "International journal of reproductive biomedicine",
  ["int j reprod contracept obstet gynecol"] = "International journal of reproduction, contraception, obstetrics and gynecology",
  ["int j reprod med"] = "International journal of reproductive medicine",
  ["int j res appl sci eng technol"] = "International journal for research in applied science and engineering technology",
  ["int j res ayurveda pharm"] = "International journal of research in ayurveda and pharmacy",
  ["int j res biosci agric technol"] = "International journal of researches in biosciences, agriculture & technology",
  ["int j res eng technol"] = "International journal of research in engineering and technology",
  ["int j res learn disabil"] = "International journal for research in learning disabilities",
  ["int j res med sci"] = "International journal of research in medical sciences",
  ["int j res nurs"] = "International journal of research in nursing",
  ["int j res orthop"] = "International journal of research in orthopaedics",
  ["int j res serv learn community engagem"] = "The international journal of research on service-learning and community engagement",
  ["int j res stud biosci"] = "International journal of research studies in biosciences",
  ["int j res undergrad math educ"] = "International Journal of Research in Undergraduate Mathematics Education",
  ["int j res zool"] = "International journal of research in zoology",
  ["int j respir pulm med"] = "International journal of respiratory and pulmonary medicine",
  ["int j retina vitreous"] = "International journal of retina and vitreous",
  ["int j rf microwave comput aided eng"] = "International Journal of RF and Microwave Computer-Aided Engineering",
  ["int j rf technol res appl"] = "International Journal of RF Technologies: Research and Aoplications",
  ["int j rheum dis"] = "International journal of rheumatic diseases",
  ["int j rheumatol"] = "International journal of rheumatology",
  ["int j risk assess manag"] = "International journal of risk assessment and management",
  ["int j risk conting manag"] = "International journal of risk and contingency management",
  ["int j risk saf med"] = "The International journal of risk & safety in medicine",
  ["int j river basin manage"] = "International Journal of River Basin Management",
  ["int j rob autom"] = "International Journal of Robotics and Automation",
  ["int j rob res"] = "International Journal of Robotics Research",
  ["int j robust nonlinear control"] = "International Journal of Robust and Nonlinear Control",
  ["int j rock mech min sci"] = "International Journal of Rock Mechanics and Mining Sciences",
  ["int j rock mech min sci (1997)"] = "International journal of rock mechanics and mining sciences (Oxford, England : 1997)",
  ["int j rotating mach"] = "International Journal of Rotating Machinery",
  ["int j salt lake res"] = "International journal of salt lake research",
  ["int j satell commun networking"] = "International Journal of Satellite Communications and Networking",
  ["int j sch educ psychol"] = "International journal of school & educational psychology",
  ["int j school disaffect"] = "The international journal on school disaffection",
  ["int j sci"] = "International journal of sciences",
  ["int j sci adv res technol"] = "International journal for science and advance research in technology",
  ["int j sci appl res"] = "International journal of sciences and applied research",
  ["int j sci appl technol"] = "International journal of science and applied technology",
  ["int j sci basic appl res"] = "International journal of sciences, basic and applied research",
  ["int j sci commer humanit"] = "International journal of science, commerce and humanities",
  ["int j sci comput"] = "International journal of scientific computing",
  ["int j sci educ"] = "International journal of science education",
  ["int j sci educ b commun public engagem"] = "International journal of science education. Part B. Communication and public engagement",
  ["int j sci educ,"] = "International Journal of Science Education",
  ["int j sci educ, part b"] = "International Journal of Science Education, Part B",
  ["int j sci eng res"] = "International journal of scientific and engineering research",
  ["int j sci environ technol"] = "International journal of science, environment and technology",
  ["int j sci math educ"] = "International journal of science and mathematics education",
  ["int j sci nat"] = "International journal of science and nature",
  ["int j sci res"] = "International journal of scientific research",
  ["int j sci res (ahmedabad)"] = "International journal of scientific research (Ahmedabad, India)",
  ["int j sci res (raipur)"] = "International journal of science and research (Raipur, India)",
  ["int j sci res comput sci appl manag stud"] = "International journal of scientific research in computer science applications and management studies",
  ["int j sci res methodol"] = "International journal of science and research methodology",
  ["int j sci res sci eng technol"] = "International journal of scientific research in science, engineering and technology",
  ["int j sci res sci technol"] = "International journal of scientific research in science and technology",
  ["int j sci soc"] = "The international journal of science in society",
  ["int j sci study"] = "International journal of scientific study",
  ["int j secur netw"] = "International Journal of Security and Networks",
  ["int j secur softw eng"] = "International journal of secure software engineering",
  ["int j sediment res"] = "International Journal of Sediment Research",
  ["int j self help self care"] = "International journal of self help & self care",
  ["int j self propag high temp synth"] = "International Journal of Self-Propagating High-Temperature Synthesis",
  ["int j self-propag high-temp synth"] = "International Journal of Self-Propagating High-Temperature Synthesis",
  ["int j semant comput"] = "International journal of semantic computing",
  ["int j semant web inf syst"] = "International journal on Semantic Web and information systems",
  ["int j semiot law"] = "International journal for the semiotics of law = Revue internationale de sémiotique juridique",
  ["int j sens netw"] = "International Journal of Sensor Networks",
  ["int j sens netw data commun"] = "International journal of sensor networks and data communications",
  ["int j serv oper inf"] = "International Journal of Services Operations and Informatics",
  ["int j serv technol manage"] = "International Journal of Services, Technology and Management",
  ["int j sex econ orgone res"] = "International journal of sex-economy and orgone-research",
  ["int j sex health"] = "International journal of sexual health : official journal of the World Association for Sexual Health",
  ["int j sexol"] = "International journal of sexology",
  ["int j shape model"] = "International Journal of Shape Modeling",
  ["int j shaping model"] = "International Journal of Shape Modeling",
  ["int j shoulder surg"] = "International journal of shoulder surgery",
  ["int j simul multi des optim"] = "International Journal for Simulation and Multidisciplinary Design Optimization",
  ["int j simul process model"] = "International Journal of Simulation and Process Modelling",
  ["int j simul syst sci technol"] = "International Journal of Simulation: Systems, Science and Technology",
  ["int j smart nano mater"] = "International journal of smart and nano materials",
  ["int j smart sens intell syst"] = "International Journal of Smart Sensing and Intelligent Systems",
  ["int j smart sustainable cities"] = "International Journal on Smart and Sustainable Cities",
  ["int j smart technol"] = "International Journal of Smart Technologies",
  ["int j soc econ"] = "International journal of social economics",
  ["int j soc educ"] = "The International journal of social education : official journal of the Indiana Council for the Social Studies",
  ["int j soc lang"] = "International journal of the sociology of language",
  ["int j soc mater eng resour"] = "International Journal of the Society of Material Engineering for Resources",
  ["int j soc psychiatry"] = "The International journal of social psychiatry",
  ["int j soc res methodol"] = "International journal of social research methodology",
  ["int j soc robot"] = "International journal of social robotics",
  ["int j soc sci humanit"] = "International journal of social science and humanity : IJSSH",
  ["int j soc sci res"] = "International journal of social science research",
  ["int j soc sci stud"] = "International journal of social science studies",
  ["int j soc syst sci"] = "International journal of society systems science",
  ["int j soc welf"] = "International journal of social welfare",
  ["int j social ecol sustainable dev"] = "International Journal of Social Ecology and Sustainable Development",
  ["int j social rob"] = "International Journal Of Social Robotics",
  ["int j sociol"] = "International journal of sociology",
  ["int j sociol fam"] = "International journal of sociology of the family",
  ["int j sociol law"] = "International Journal Of The Sociology Of Law",
  ["int j sociol soc policy"] = "The International journal of sociology and social policy",
  ["int j soft comput"] = "International Journal of Soft Computing",
  ["int j softw tools technol trans"] = "International Journal on Software Tools for Technology Transfer",
  ["int j softw tools technol transf"] = "International journal on software tools for technology transfer : STTT",
  ["int j software eng knowl eng"] = "International Journal of Software Engineering and Knowledge Engineering",
  ["int j software eng knowledge eng"] = "International Journal of Software Engineering and Knowledge Engineering",
  ["int j software tools technol transfer"] = "International Journal on Software Tools for Technology Transfer",
  ["int j soil dyn earthquake eng"] = "International Journal of Soil Dynamics and Earthquake Engineering",
  ["int j soil sci"] = "International Journal of Soil Science",
  ["int j soil sediment water"] = "The international journal of soil, sediment and water : documenting the cutting edge of environmental stewardship",
  ["int j solids struct"] = "International Journal of Solids and Structures",
  ["int j solids structures"] = "International Journal of Solids and Structures",
  ["int j south asian stud (new delhi)"] = "International journal of South Asian studies",
  ["int j south asian stud (pondicherry)"] = "International journal of South Asian studies (Pondicherry, India)",
  ["int j space struct"] = "International Journal of Space Structures",
  ["int j spec educ"] = "International journal of special education",
  ["int j spectro"] = "International Journal of Spectroscopy",
  ["int j spectrosc"] = "International journal of spectroscopy",
  ["int j speech lang pathol"] = "International journal of speech-language pathology",
  ["int j speech lang pathol audiol"] = "International journal of speech & language pathology and audiology",
  ["int j speech technol"] = "International Journal of Speech Technology",
  ["int j speleol"] = "International Journal of Speleology",
  ["int j spine surg"] = "International journal of spine surgery",
  ["int j sport exerc psychol"] = "International journal of sport and exercise psychology",
  ["int j sport exercise psychol"] = "International Journal of Sport and Exercise Psychology",
  ["int j sport nutr"] = "International Journal of Sport Nutrition",
  ["int j sport nutr exerc metab"] = "International journal of sport nutrition and exercise metabolism",
  ["int j sport nutr exercise metab"] = "International Journal of Sport Nutrition and Exercise Metabolism",
  ["int j sport psychol"] = "International journal of sport psychology",
  ["int j sports exerc med"] = "International journal of sports and exercise medicine",
  ["int j sports med"] = "International journal of sports medicine",
  ["int j sports phys ther"] = "International journal of sports physical therapy",
  ["int j sports physiol perform"] = "International journal of sports physiology and performance",
  ["int j sports sci coach"] = "International journal of sports science & coaching",
  ["int j spray combust dyn"] = "International Journal of Spray and Combustion Dynamics",
  ["int j stat appl"] = "International journal of statistics and applications",
  ["int j stat manag syst"] = "International journal of statistics and management system",
  ["int j stat med res"] = "International journal of statistics in medical research",
  ["int j stat probab"] = "International journal of statistics and probability",
  ["int j std aids"] = "International journal of STD & AIDS",
  ["int j steel struct"] = "International Journal of Streel Structures",
  ["int j stem cell res ther"] = "International journal of stem cell research and therapy",
  ["int j stem cells"] = "International journal of stem cells",
  ["int j stem educ"] = "International journal of STEM education",
  ["int j stem educ res"] = "International Journal of STEM Education Research",
  ["int j stoch anal"] = "International Journal of Stochastic Analysis",
  ["int j stomatol occlusion med"] = "International journal of stomatology & occlusion medicine",
  ["int j stress manag"] = "International journal of stress management",
  ["int j stroke"] = "International journal of stroke : official journal of the International Stroke Society",
  ["int j struct changes sol"] = "The international journal of structural changes in solids : mechanics and applications",
  ["int j struct eng"] = "International Journal of Structural Engineering",
  ["int j struct integrity"] = "International Journal of Structural Integrity",
  ["int j struct stab dyn"] = "International Journal of Structural Stability and Dynamics",
  ["int j stud scholarsh phys ther"] = "International journal of student scholarship in physical therapy",
  ["int j study anim probl"] = "International journal for the study of animal problems",
  ["int j sulfur chem part a:"] = "International Journal of Sulfur Chemistry Part A: Original Experimental and Theoretical Studies",
  ["int j sulfur chem part b:"] = "International Journal of Sulfur Chemistry Part B: Quartler Reports on Sulfur Chemistry",
  ["int j supercomput appl"] = "The international journal of supercomputer applications",
  ["int j surf eng interdiscip mater sci"] = "International Journal Of Surface Engineering and Interdisciplinary Materials Science",
  ["int j surf min reclam environ"] = "International Journal of Surface Mining, Reclamation, and Environment",
  ["int j surf sci eng"] = "International Journal of Surface Science and Engineering",
  ["int j surg"] = "International journal of surgery (London, England)",
  ["int j surg case rep"] = "International journal of surgery case reports",
  ["int j surg investig"] = "International journal of surgical investigation",
  ["int j surg oncol"] = "International journal of surgical oncology",
  ["int j surg oncol (n y)"] = "International journal of surgery. Oncology",
  ["int j surg pathol"] = "International journal of surgical pathology",
  ["int j surg protoc"] = "International journal of surgery protocols",
  ["int j surg res"] = "International journal of surgery and research",
  ["int j sustain transp"] = "International journal of sustainable transportation",
  ["int j sustainability econ social cult context"] = "International Journal of Sustainability in Economic Social and Cultural Context",
  ["int j sustainability higher educ"] = "International Journal of Sustainability in Higher Education",
  ["int j sustainable agric manage inf"] = "International Journal of Sustainable Agricultural Management and Informatics",
  ["int j sustainable aviat"] = "International Journal of Sustainable Aviation",
  ["int j sustainable build technol urban dev"] = "International Journal of Sustainable Building Technology and Urban Development",
  ["int j sustainable built environ"] = "International Journal of Sustainable Built Environment",
  ["int j sustainable constr eng technol"] = "International Journal of Sustainable Construction Engineering and Technology",
  ["int j sustainable dev"] = "International Journal of Sustainable Development",
  ["int j sustainable dev plann"] = "International Journal of Sustainable Development and Planning",
  ["int j sustainable dev world ecol"] = "International Journal of Sustainable Development and World Ecology",
  ["int j sustainable energy"] = "International Journal of Sustainable Energy",
  ["int j sustainable energy plann manage"] = "International Journal of Sustainable Energy Planning and Management",
  ["int j sustainable eng"] = "International Journal of Sustainable Engineering",
  ["int j sustainable oceans coasts"] = "International Journal of Sustainable Oceans and Coasts",
  ["int j sustainable soc"] = "International Journal of Sustainable Society",
  ["int j sustainable transp"] = "International Journal of Sustainable Transportation",
  ["int j syst assur eng manage"] = "International Journal Of Systems Assurance Engineering And Management",
  ["int j syst bacteriol"] = "International journal of systematic bacteriology",
  ["int j syst evol microbiol"] = "International journal of systematic and evolutionary microbiology",
  ["int j syst sci"] = "International Journal of Systems Science",
  ["int j syst sci: oper logist"] = "International Journal of Systems Science: Operations and Logistics",
  ["int j syst synth biol"] = "International journal of systems and synthetic biology",
  ["int j target ther cancer"] = "The International journal of targeted therapies in cancer",
  ["int j technoethics"] = "International journal of technoethics",
  ["int j technol aging"] = "International journal of technology & aging",
  ["int j technol assess health care"] = "International journal of technology assessment in health care",
  ["int j technol des educ"] = "International Journal of Technology and Design Education",
  ["int j technol hum interact"] = "International Journal of Technology and Human Interaction",
  ["int j technol intell plann"] = "International Journal of Technology Intelligence and Planning",
  ["int j technol knowl soc"] = "International journal of technology, knowledge and society",
  ["int j technol manag"] = "International journal of technology management = Journal international de la gestion technologique",
  ["int j technol manage"] = "International Journal of Technology Management",
  ["int j technol manage sustainable dev"] = "International Journal of Technology Management and Sustainable Development",
  ["int j technol policy manage"] = "International Journal of Technology, Policy and Management",
  ["int j telemed appl"] = "International journal of telemedicine and applications",
  ["int j telerehabil"] = "International journal of telerehabilitation",
  ["int j test"] = "International journal of testing",
  ["int j theor appl finance"] = "International Journal of Theoretical and Applied Finance",
  ["int j theor phys"] = "International Journal of Theoretical Physics",
  ["int j theor phys group theory nonlinear opt"] = "International Journal of Theoretical Physics, Group Theory, and Nonlinear Optics",
  ["int j theoret phys"] = "International Journal of Theoretical Physics",
  ["int j theoretical appl finance"] = "International Journal of Theoretical and Applied Finance",
  ["int j ther massage bodywork"] = "International journal of therapeutic massage & bodywork",
  ["int j ther rehabil"] = "International journal of therapy and rehabilitation",
  ["int j therm sci"] = "International Journal of Thermal Sciences",
  ["int j thermodyn"] = "International Journal of Thermodynamics",
  ["int j thermofluids"] = "International Journal of Thermofluids",
  ["int j thermophys"] = "International Journal of Thermophysics",
  ["int j tissue eng"] = "International journal of tissue engineering",
  ["int j tissue react"] = "International journal of tissue reactions",
  ["int j toxicol"] = "International journal of toxicology",
  ["int j train dev"] = "International journal of training and development",
  ["int j transgend"] = "The international journal of transgenderism",
  ["int j transgend health"] = "International journal of transgender health",
  ["int j transl sci"] = "International journal of translational science",
  ["int j transp dev integr"] = "International Journal of Transport Development and Integration",
  ["int j transp phenom"] = "International Journal of Transport Phenomena",
  ["int j transp sci technol"] = "International Journal of Transportation Science and Technology",
  ["int j transport econ"] = "International Journal of Transport Economics",
  ["int j trauma nurs"] = "International journal of trauma nursing",
  ["int j travel med glob health"] = "International journal of travel medicine and global health",
  ["int j trichology"] = "International journal of trichology",
  ["int j trop agric"] = "International journal of tropical agriculture = Antarrāshtriya ūshnakatibandhīya kr̥shi-śodha patrikā",
  ["int j trop dis health"] = "International journal of tropical disease & health",
  ["int j trop insect sci"] = "International Journal of Tropical Insect Science",
  ["int j trop subtrop hortic"] = "International Journal of Tropical and Subtropical Horticulture",
  ["int j tryptophan res"] = "International journal of tryptophan research : IJTR",
  ["int j tuberc lung dis"] = "The international journal of tuberculosis and lung disease : the official journal of the International Union against Tuberculosis and Lung Disease",
  ["int j turbo jet engines"] = "International Journal Of Turbo And Jet Engines",
  ["int j uncertain fuzziness knowledge-based systems"] = "International Journal of Uncertainty, Fuzziness and Knowledge-Based Systems",
  ["int j uncertain quantif"] = "International journal for uncertainty quantification",
  ["int j uncertainty fuzziness knowledge based syst"] = "International Journal of Uncertainty, Fuzziness and Knowlege-Based Systems",
  ["int j uncertainty quantif"] = "International Journal for Uncertainty Quantification",
  ["int j univ teach fac dev"] = "International journal of university teaching and faculty development",
  ["int j univers pharm life sci"] = "International journal of universal pharmacy and life sciences",
  ["int j unmanned syst"] = "International Journal of Unmanned Systems",
  ["int j urban reg res"] = "International journal of urban and regional research",
  ["int j urban sci"] = "International Journal of Urban Sciences",
  ["int j urban sustainable dev"] = "International Journal of Urban Sustainable Development",
  ["int j urol"] = "International journal of urology : official journal of the Japanese Urological Association",
  ["int j urol nurs"] = "International journal of urological nursing : the journal of the BAUN",
  ["int j vasc med"] = "International journal of vascular medicine",
  ["int j veg sci"] = "International Journal of Vegetable Science",
  ["int j veh auton syst"] = "International Journal of Vehicle Autonomous Systems",
  ["int j veh des"] = "International Journal of Vehicle Design",
  ["int j veh inf commun syst"] = "International Journal of Vehicle Information and Communication Systems",
  ["int j veh noise vib"] = "International Journal of Vehicle Noise and Vibration",
  ["int j veh saf"] = "International Journal of Vehicle Safety",
  ["int j veh syst model test"] = "International Journal of Vehicle Systems Modelling and Testing",
  ["int j vent"] = "International Journal of Ventilation",
  ["int j vet sci"] = "International journal of veterinary science",
  ["int j vet sci med"] = "International journal of veterinary science and medicine",
  ["int j virol"] = "International journal of virology",
  ["int j virol mol biol"] = "International journal of virology and molecular biology",
  ["int j virol stud res"] = "International journal of virology studies & research",
  ["int j virtual communities soc netw"] = "International journal of virtual communities and social networking",
  ["int j virtual real"] = "The international journal of virtual reality : a multimedia publication for professionals",
  ["int j vitam nutr res"] = "International journal for vitamin and nutrition research. Internationale Zeitschrift für Vitamin- und Ernährungsforschung. Journal international de vitaminologie et de nutrition",
  ["int j vitam nutr res suppl"] = "International journal for vitamin and nutrition research. Supplement = Internationale Zeitschrift für Vitamin- und Ernährungsforschung. Supplement",
  ["int j waste resour"] = "International journal of waste resources",
  ["int j water"] = "International Journal of Water",
  ["int j water resour dev"] = "International Journal of Water Resources Development",
  ["int j wavelets multiresolut inf process"] = "International journal of wavelets, multiresolution and information processing",
  ["int j wavelets multiresolution inf process"] = "International Journal of Wavelets, Multiresolution and Information Processing",
  ["int j web based communities"] = "International Journal of Web Based Communities",
  ["int j web eng technol"] = "International Journal of Web Engineering and Technology",
  ["int j web grid serv"] = "International Journal of Web and Grid Services",
  ["int j web serv res"] = "International Journal of Web Services Research",
  ["int j wild silkmoth silk"] = "International journal of wild silkmoth & silk",
  ["int j wildland fire"] = "International journal of wildland fire",
  ["int j wine res"] = "International Journal of Wine Research",
  ["int j wirel inf netw"] = "International journal of wireless information networks",
  ["int j wireless inf networks"] = "International Journal of Wireless Information Networks",
  ["int j wireless mobile comput"] = "International Journal of Wireless and Mobile Computing",
  ["int j wireless opt commun"] = "International Journal on Wireless & Optical Communications",
  ["int j womens dermatol"] = "International journal of women's dermatology",
  ["int j womens health"] = "International journal of women's health",
  ["int j womens health wellness"] = "International journal of women's health and wellness",
  ["int j womens stud"] = "International journal of women's studies",
  ["int j workplace health manag"] = "International journal of workplace health management",
  ["int j wrestling sci"] = "International Journal of Wrestling Science",
  ["int j yoga"] = "International journal of yoga",
  ["int j yoga therap"] = "International journal of yoga therapy",
  ["int j zool"] = "International journal of zoology",
  ["int j zool res"] = "International journal of zoology research",
  ["int j zoonoses"] = "International journal of zoonoses",
  ["int lab rev"] = "International Labour Review",
  ["int labor work class hist"] = "International labor and working class history",
  ["int labour rev"] = "International labour review",
  ["int law econ"] = "International Law and Economics",
  ["int law news"] = "The international law news",
  ["int libr diabetes metab"] = "International library of diabetes & metabolism",
  ["int libr rev"] = "International library review",
  ["int lichenol newsl"] = "International Lichenological Newsletter",
  ["int mar energy j"] = "International Marine Energy Journal",
  ["int marit health"] = "International maritime health",
  ["int mater rev"] = "International Materials Reviews",
  ["int math news"] = "International Mathematical News",
  ["int math res not"] = "International Mathematics Research Notices",
  ["int math res not imrn"] = "International Mathematics Research Notices. IMRN",
  ["int math res notices"] = "International Mathematics Research Notices",
  ["int math res pap"] = "International Mathematics Research Papers",
  ["int math res surv"] = "International Mathematics Research Surveys",
  ["int mech eng congress expo"] = "International Mechanical Engineering Congress and Exposition : [proceedings]. International Mechanical Engineering Congress and Exposition",
  ["int med abstr rev"] = "International medical abstracts & reviews",
  ["int med case rep j"] = "International medical case reports journal",
  ["int microbiol"] = "International microbiology : the official journal of the Spanish Society for Microbiology",
  ["int migr"] = "International migration (Geneva, Switzerland)",
  ["int migr rev"] = "The International migration review",
  ["int mitt bodenkd"] = "Internationale Mitteilungen für Bodenkunde",
  ["int monet fund staff pap"] = "International Monetary Fund Staff Papers",
  ["int monit rheumatol"] = "The International monitor. Rheumatology",
  ["int ms j"] = "International MS journal",
  ["int multidiscip res j"] = "International multidisciplinary research journal",
  ["int multiling res j"] = "International multilingual research journal",
  ["int musculoskelet med"] = "International musculoskeletal medicine",
  ["int nano lett"] = "International nano letters",
  ["int neuropsychiatr dis j"] = "International neuropsychiatric disease journal",
  ["int neurotrauma lett"] = "International neurotrauma letter",
  ["int neurourol j"] = "International neurourology journal",
  ["int nonwovens j"] = "International Nonwovens Journal",
  ["int nurs rev"] = "International nursing review",
  ["int online j educ teach"] = "International online journal of education & teaching",
  ["int ophthalmol"] = "International ophthalmology",
  ["int ophthalmol clin"] = "International ophthalmology clinics",
  ["int organ"] = "International organization",
  ["int orthod"] = "International orthodontics",
  ["int orthop"] = "International orthopaedics",
  ["int peacekeep"] = "International peacekeeping (London, England)",
  ["int peat j"] = "International Peat Journal",
  ["int pediatr"] = "International pediatrics : the journal of the Miami Children's Hospital",
  ["int perspect psychol"] = "International perspectives in psychology : research, practice, consultation",
  ["int perspect sex reprod health"] = "International perspectives on sexual and reproductive health",
  ["int perspect vict"] = "International perspectives in victimology",
  ["int pharmacopsychiatry"] = "International pharmacopsychiatry",
  ["int philos q"] = "International philosophical quarterly : IPQ",
  ["int physiol j"] = "International physiology journal",
  ["int plan stud"] = "International planning studies",
  ["int plann parenthood news"] = "International planned parenthood news",
  ["int pol rev"] = "International Policy Review",
  ["int political sociol"] = "International political sociology : IPS",
  ["int polym proc"] = "International Polymer Processing",
  ["int polym process"] = "International polymer processing : the journal of the Polymer Processing Society",
  ["int proc chem biol environ eng"] = "International Proceedings of Chemical, Biological & Environmental Engineering",
  ["int proj newsl"] = "International project newsletter",
  ["int psychiatry"] = "International psychiatry : bulletin of the Board of International Affairs of the Royal College of Psychiatrists",
  ["int psychiatry clin"] = "International psychiatry clinics",
  ["int psychogeriatr"] = "International psychogeriatrics",
  ["int public health j"] = "International public health journal",
  ["int public manag j"] = "International public management journal : IPMJ",
  ["int q community health educ"] = "International quarterly of community health education",
  ["int rec med"] = "International record of medicine",
  ["int rec med gen pract clin"] = "International record of medicine and general practice clinics",
  ["int reg sci rev"] = "International regional science review",
  ["int rehabil med"] = "International rehabilitation medicine",
  ["int relat (david davies mem inst int stud)"] = "International relations (David Davies Memorial Institute of International Studies)",
  ["int res geogr environ educ"] = "International Research in Geographical and Environmental Education",
  ["int res j eng technol"] = "International Research Journal of Engineering and Technology",
  ["int res j environ sci"] = "International research journal of environmental sciences",
  ["int res j nat appl sci"] = "International research journal of natural and applied sciences",
  ["int res j pure appl chem"] = "International research journal of pure and applied chemistry",
  ["int res med sci"] = "International research on medical sciences",
  ["int rescuer"] = "International rescuer",
  ["int rev"] = "International review (Steubenville, Ohio)",
  ["int rev appl econ"] = "International review of applied economics",
  ["int rev appl sci eng"] = "International Review of Applied Sciences and Engineering",
  ["int rev cell mol biol"] = "International review of cell and molecular biology",
  ["int rev connect tissue res"] = "International review of connective tissue research",
  ["int rev cytol"] = "International review of cytology",
  ["int rev cytol suppl"] = "International review of cytology. Supplement",
  ["int rev econ"] = "International review of economics",
  ["int rev econ finance"] = "International Review of Economics and Finance",
  ["int rev educ"] = "International review of education. Internationale Zeitschrift für Erziehungswissenschaft. Revue internationale de pédagogie",
  ["int rev electr eng"] = "International Review of Electrical Engineering",
  ["int rev exp pathol"] = "International review of experimental pathology",
  ["int rev finan anal"] = "International Review of Financial Analysis",
  ["int rev gesamten hydrobiol"] = "Internationale Revue der Gesamten Hydrobiologie",
  ["int rev gesamten hydrobiol hydrogr"] = "Internationale Revue der Gesamten Hydrobiologie und Hydrographie",
  ["int rev hydrobiol"] = "International Review of Hydrobiology",
  ["int rev immunol"] = "International reviews of immunology",
  ["int rev law econ"] = "International review of law and economics",
  ["int rev missions"] = "International review of missions",
  ["int rev mod sociol"] = "International Review Of Modern Sociology",
  ["int rev nat fam plann"] = "International review of natural family planning",
  ["int rev neurobiol"] = "International review of neurobiology",
  ["int rev neurobiol suppl"] = "International review of neurobiology. Supplement",
  ["int rev phys chem"] = "International Reviews in Physical Chemistry",
  ["int rev physiol"] = "International review of physiology",
  ["int rev poult sci"] = "International Review of Poultry Science",
  ["int rev psychiatry"] = "International review of psychiatry (Abingdon, England)",
  ["int rev psychoanal"] = "The International review of psycho-analysis",
  ["int rev qual res"] = "International review of qualitative research : IRQR",
  ["int rev res dev disabil"] = "International review of research in developmental disabilities",
  ["int rev res ment retard"] = "International review of research in mental retardation",
  ["int rev soc hist"] = "International review of social history",
  ["int rev soc res"] = "International Review of Social Research",
  ["int rev sociol sport"] = "International review for the sociology of sport",
  ["int rev spatial plann sustainable dev"] = "International Review for Spatial Planning and Sustainable Development",
  ["int rev sport exerc psychol"] = "International review of sport and exercise psychology",
  ["int rev sport exercise psychol"] = "International Review of Sport and Exercise Psychology",
  ["int rev thromb"] = "International review of thrombosis",
  ["int rev trop med"] = "International review of tropical medicine",
  ["int rev vict"] = "International review of victimology",
  ["int rice res notes"] = "International rice research notes",
  ["int sampe symp exhib"] = "International SAMPE Symposium and Exhibition : [proceedings]. International SAMPE Symposium and Exhibition",
  ["int sch res notices"] = "International scholarly research notices",
  ["int schools comput sci"] = "International Schools for Computer Scientists",
  ["int sci rev ser"] = "International science review series",
  ["int secur"] = "International security",
  ["int semin surg oncol"] = "International seminars in surgical oncology : ISSO",
  ["int ser actuar sci"] = "International Series on Actuarial Science",
  ["int ser comput eng"] = "International Series on Computational Engineering",
  ["int ser intell tech"] = "International Series in Intelligent Technologies",
  ["int ser monogr chem"] = "International Series of Monographs on Chemistry",
  ["int ser monogr oral biol"] = "International series of monographs on oral biology",
  ["int ser monogr phys"] = "International Series of Monographs on Physics",
  ["int ser monogr pure appl math"] = "International Series of Monographs on Pure and Applied Mathematics",
  ["int ser numer math"] = "International Series of Numerical Mathematics",
  ["int ser oper res management sci"] = "International Series in Operations Research & Management Science",
  ["int ser quant mark"] = "International Series in Quantitative Marketing",
  ["int shipbuild prog"] = "International Shipbuilding Progress",
  ["int soc anal appl comput"] = "International Society for Analysis, Applications and Computation",
  ["int soc dev rev"] = "International social development review",
  ["int soc sci j"] = "International social science journal",
  ["int soc sci rev"] = "International social science review",
  ["int soc secur rev"] = "International social security review",
  ["int soc work"] = "International social work",
  ["int social rev"] = "International Socialist review",
  ["int sociol"] = "International sociology : journal of the International Sociological Association",
  ["int soil water conserv res"] = "International Soil and Water Conservation Research",
  ["int solid state sens actuators microsyst conf"] = "International Solid-State Sensors, Actuators and Microsystems Conference : [proceedings]. International Conference on Solid-State Sensors, Actuators, and Microsystems",
  ["int spect"] = "Internationale spectator",
  ["int spect (rome)"] = "The international spectator : a quarterly journal of the Istituto affari internazionali",
  ["int sport coach j"] = "International sport coaching journal",
  ["int stat rev"] = "International statistical review = Revue internationale de statistique",
  ["int std res rev"] = "International STD research & reviews",
  ["int stud hist math teach"] = "International Studies in the History of Mathematics and its Teaching",
  ["int stud philos sci"] = "International Studies in the Philosophy of Science",
  ["int stud q"] = "International studies quarterly : a publication of the International Studies Association",
  ["int sugar j"] = "International Sugar Journal",
  ["int surg"] = "International surgery",
  ["int symp fluorid prev dent"] = "International Symposium On Fluoridation And Preventive Dentistry",
  ["int symp inf theory appl"] = "International Symposium on Information Theory and its Applications. International Symposium on Information Theory and its Applications",
  ["int symp mix augment real"] = "International Symposium on Mixed and Augmented Reality : (ISMAR) [proceedings]. IEEE and ACM International Symposium on Mixed and Augmented Reality",
  ["int symp process syst eng"] = "International symposium on process systems engineering",
  ["int symp reprod physiol fish"] = "International Symposium on Reproductive Physiology of Fish",
  ["int sympos econom theory econometrics"] = "International Symposia in Economic Theory and Econometrics",
  ["int tax public financ"] = "International tax and public finance",
  ["int tax public finance"] = "International Tax and Public Finance",
  ["int tinnitus j"] = "The international tinnitus journal",
  ["int trade j"] = "International Trade Journal",
  ["int tran oper res"] = "International Transactions in Operational Research",
  ["int trans electr energy syst"] = "International Transactions on Electrical Energy Systems",
  ["int trans j eng manag appl sci technol"] = "International transaction journal of engineering, management, & applied sciences & technologies",
  ["int trans oper res"] = "International transactions in operational research : a journal of The International Federation of Operational Research Societies",
  ["int tree crops j"] = "International Tree Crops Journal",
  ["int trends immun"] = "International trends in immunity",
  ["int turfgrass soc res j"] = "International Turfgrass Society research journal",
  ["int undergrad j serv learn leadersh soc change"] = "The international undergraduate journal for service-learning, leadership, and social change",
  ["int union cryst cryst sympos"] = "International Union of Crystallography Crystallographic Symposia",
  ["int urogynecol j"] = "International urogynecology journal",
  ["int urogynecol j pelvic floor dysfunct"] = "International urogynecology journal and pelvic floor dysfunction",
  ["int urol nephrol"] = "International urology and nephrology",
  ["int wildl"] = "International wildlife",
  ["int wood prod j"] = "International Wood Products Journal",
  ["int work conf interp nat artif comput"] = "International Work-Conference on the Interplay between Natural and Artificial Computation",
  ["int workshop databases expert syst appl"] = "International Workshop on Databases and Expert Systems Applications : proceedings",
  ["int wound j"] = "International wound journal",
  ["int year-book game theory appl"] = "International Year-Book of Game Theory and Applications",
  ["int z angew physiol"] = "Internationale Zeitschrift für angewandte Physiologie, einschliesslich Arbeitsphysiologie",
  ["int z klin pharmakol ther toxikol"] = "Internationale Zeitschrift für klinische Pharmakologie, Therapie, und Toxikologie. International journal of clinical pharmacology, therapy, and toxicology",
  ["int z phys med rehabil"] = "Internationale Zeitschrift für physikalische Medizin und Rehabilitation, mit Zentralblatt für die gesamte Rehabilitation",
  ["int z vitam ernahrungsforsch beih"] = "Internationale Zeitschrift fur Vitamin- und Ernahrungsforschung. Beiheft",
  ["int z vitaminforsch"] = "Internationale Zeitschrift für Vitaminforschung. Internătional journal of vitamin research. Journal international de vitaminologie",
  ["int z vitaminforsch beih"] = "Internationale Zeitschrift für Vitaminforschung. Beiheft",
  ["int zoo yearb"] = "The International zoo yearbook",
  ["inted proc"] = "INTED proceedings",
  ["integr assess"] = "Integrated Assessment",
  ["integr biol"] = "Integrative Biology",
  ["integr biol (camb)"] = "Integrative biology : quantitative biosciences from nano to macro",
  ["integr blood press control"] = "Integrated blood pressure control",
  ["integr blood pressure control"] = "Integrated Blood Pressure Control",
  ["integr cancer biol res"] = "Integrative cancer biology & research",
  ["integr cancer sci ther"] = "Integrative cancer science and therapeutics",
  ["integr cancer ther"] = "Integrative cancer therapies",
  ["integr clin med"] = "Integrative clinical medicine",
  ["integr comp biol"] = "Integrative and comparative biology",
  ["integr comput aided eng"] = "Integrated computer-aided engineering",
  ["integr comput-aided eng"] = "Integrated Computer-Aided Engineering",
  ["integr converv"] = "Integrative Conservation",
  ["integr environ assess manag"] = "Integrated environmental assessment and management",
  ["integr environ assess manage"] = "Integrated Environmental Assessment and Manaement",
  ["integr equations oper theory"] = "Integral Equations and Operator Theory",
  ["integr ferroelectr"] = "Integrated Ferroelectrics",
  ["integr food nutr metab"] = "Integrative food, nutrition and metabolism",
  ["integr gynecol obstet j"] = "Integrative gynecology and obstetrics journal",
  ["integr healthc rep"] = "Integrated healthcare report",
  ["integr manuf syst"] = "Integrated Manufacturing Systems",
  ["integr mater manuf innov"] = "Integrating materials and manufacturing innovation",
  ["integr mater manuf innovation"] = "Integrating Materials and Manufacturing Innovation",
  ["integr med"] = "Integrative medicine : integrating conventional and alternative medicine",
  ["integr med (encinitas)"] = "Integrative medicine (Encinitas, Calif.)",
  ["integr med insights"] = "Integrative medicine insights",
  ["integr med int"] = "Integrative medicine international",
  ["integr med res"] = "Integrative medicine research",
  ["integr mol med"] = "Integrative molecular medicine",
  ["integr obes diabetes"] = "Integrative obesity and diabetes",
  ["integr org biol"] = "Integrative organismal biology (Oxford, England)",
  ["integr pest manage rev"] = "Integrated Pest Management Reviews",
  ["integr pharm res pract"] = "Integrated pharmacy research & practice",
  ["integr physiol behav sci"] = "Integrative physiological and behavioral science : the official journal of the Pavlovian Society",
  ["integr psychiatry"] = "Integrative psychiatry : IP",
  ["integr psychol behav sci"] = "Integrative psychological & behavioral science",
  ["integr ser inf syst"] = "Integrated Series in Information Systems",
  ["integr syst"] = "Integrative systematics : Stuttgart contributions to natural history",
  ["integr syst: stuttgart contrib nat hist"] = "Integrative Systematics: Stuttgart Contributions to Natural History",
  ["integr transforms special funct"] = "Integral Transforms and Special Functions",
  ["integr zool"] = "Integrative zoology",
  ["integral equations operator theory"] = "Integral Equations and Operator Theory",
  ["integral transform spec funct"] = "Integral transforms and special functions",
  ["integral transforms spec funct"] = "Integral Transforms and Special Functions. An International Journal",
  ["integration (amst)"] = "Integration (Amsterdam, Netherlands)",
  ["integration trade"] = "Integration and Trade",
  ["intel robotics aton agents"] = "Intelligent Robotics Atonomous Agents",
  ["intel robotics auton agent"] = "Intelligent Robotics Autonomous Agents",
  ["intel robotics auton agents"] = "Intelligent Robotics Autonomous Agents",
  ["intell autom soft comput"] = "Intelligent Automation & Soft Computing",
  ["intell based med"] = "Intelligence-based medicine",
  ["intell build int"] = "Intelligent Buildings International",
  ["intell comput"] = "Intelligent Computing",
  ["intell control learn syst"] = "Intelligent Control and Learning Systems",
  ["intell converged networks"] = "Intelligent and Converged Networks",
  ["intell data anal"] = "Intelligent Data Analysis",
  ["intell decis technol"] = "Intelligent Decision Technologies",
  ["intell ind syst"] = "Intelligent Industrial Systems",
  ["intell inf manag"] = "Intelligent information management",
  ["intell med"] = "Intelligent Medicine",
  ["intell sci big data eng (2013)"] = "Intelligence science and big data engineering : 4th international conference, IScIDE 2013, Beijing, China, July 31-August 2, 2013, revised selected papers. IScIDE (4th : 2013 : Beijing, China)",
  ["intell serv rob"] = "Intelligent Service Robotics",
  ["intell serv robot"] = "Intelligent service robotics",
  ["intell surg"] = "Intelligent Surgery",
  ["intell syst appl"] = "Intelligent Systems with Applications",
  ["intell syst control autom sci eng"] = "Intelligent Systems, Control and Automation: Science and Engineering",
  ["intell syst eng"] = "Intelligent Systems Engineering",
  ["intell syst ref libr"] = "Intelligent Systems Reference Library",
  ["intellect dev disabil"] = "Intellectual and developmental disabilities",
  ["intellect dig"] = "Intellectual digest",
  ["intellect hist islam world"] = "Intellectual history of the Islamicate world",
  ["intellect hist rev"] = "Intellectual history review",
  ["intellect prop j"] = "Intellectual property journal",
  ["intensive care med"] = "Intensive care medicine",
  ["intensive care med exp"] = "Intensive care medicine experimental",
  ["intensive care nurs"] = "Intensive care nursing",
  ["intensive care world"] = "Intensive care world",
  ["intensive crit care nurs"] = "Intensive & critical care nursing",
  ["intensivmed notfallmed"] = "Intensivmedizin + Notfallmedizin : Organ der Deutschen und der Österreichischen Gesellschaft für internistische Intensivmedizin, der Sektion Neurologie der DGIM und der Sektion Intensivmedizin im Berufsverband Deutscher Internisten e.V",
  ["intensivmed prax"] = "Intensivmedizinische Praxis",
  ["inter am econ aff"] = "Inter-American economic affairs",
  ["inter des"] = "Interior design (New York, N.Y.)",
  ["inter econ"] = "Inter economics",
  ["inter nord"] = "Inter-Nord",
  ["interact cardiovasc thorac surg"] = "Interactive cardiovascular and thoracic surgery",
  ["interact comput"] = "Interacting with computers",
  ["interact j med res"] = "Interactive journal of medical research",
  ["interact mech math"] = "Interaction of Mechanics and Mathematics",
  ["interact stud"] = "Interaction studies",
  ["interactions (n y)"] = "Interactions (New York, N.Y.)",
  ["interam j psychol"] = "Revista interamericana de psicología = Interamerican journal of psychology",
  ["intercamb urol corresp"] = "Intercambio urológico por correspondencia",
  ["interceram - int ceram rev"] = "Interceram - International Ceramic Review",
  ["interchange (tor : 1984)"] = "Interchange (Toronto, Ont. : 1984)",
  ["intercom (des moines)"] = "Intercom",
  ["interdiscip appl math"] = "Interdisciplinary Applied Mathematics",
  ["interdiscip archaeol"] = "Interdisciplinaria archaeologica",
  ["interdiscip contrib archaeol"] = "Interdisciplinary Contributions to Archaeology",
  ["interdiscip environ rev"] = "Interdisciplinary Environmental Review",
  ["interdiscip inf sci"] = "Interdisciplinary information sciences",
  ["interdiscip inform sci"] = "Interdisciplinary Information Sciences",
  ["interdiscip j best pract glob dev"] = "Interdisciplinary journal of best practices in global development",
  ["interdiscip j partnersh stud"] = "Interdisciplinary journal of partnership studies",
  ["interdiscip j probl based learn"] = "The interdisciplinary journal of problem-based learning",
  ["interdiscip j res relig"] = "Interdisciplinary journal of research on religion",
  ["interdiscip j sustainable oceans coasts"] = "Interdisciplinary Journal of Sustainable Oceans and Coasts",
  ["interdiscip mater"] = "Interdisciplinary Materials",
  ["interdiscip math"] = "Interdisciplinary Mathematics",
  ["interdiscip math sci"] = "Interdisciplinary Mathematical Sciences",
  ["interdiscip neurosurg"] = "Interdisciplinary neurosurgery : Advanced techniques and case management",
  ["interdiscip neurosurg adv tech case manage"] = "Interdisciplinary Neurosurgery: Advanced Techniques and Case Management",
  ["interdiscip perspect infect dis"] = "Interdisciplinary perspectives on infectious diseases",
  ["interdiscip sci"] = "Interdisciplinary sciences, computational life sciences",
  ["interdiscip sci rev"] = "Interdisciplinary science reviews : ISR",
  ["interdiscip sci-comput life sci"] = "Interdisciplinary Sciences-Computational LIfe Sciences",
  ["interdiscip sci: comput life sci"] = "Interdisciplinary Sciences: Computational LIfe Sciences",
  ["interdiscip statist"] = "Interdisciplinary Statistics",
  ["interdiscip stud econ manag"] = "Interdisciplinary Studies in Economics and Management",
  ["interdiscip stud nat math"] = "Interdisciplinary Studies on the Nature of Mathematics",
  ["interdiscip systems res"] = "Interdisciplinary Systems Research",
  ["interdiscip top gerontol"] = "Interdisciplinary topics in gerontology",
  ["interdiscip top gerontol geriatr"] = "Interdisciplinary topics in gerontology and geriatrics",
  ["interdiscip toxicol"] = "Interdisciplinary toxicology",
  ["interface (maynooth)"] = "Interface : a journal for and about social movements",
  ["interface focus"] = "Interface focus",
  ["interface sci"] = "Interface Science",
  ["interfaces (providence)"] = "Interfaces",
  ["interfaces free bound"] = "Interfaces and Free Boundaries. Mathematical Analysis, Computation and Applications",
  ["interfaces free boundaries"] = "Interfaces and Free Boundaries",
  ["interfacial phenom heat transf"] = "Interfacial phenomena and heat transfer",
  ["interfacial phenom heat transfer"] = "Interfacial Phenomena and Heat Transfer",
  ["intermetallics (barking)"] = "Intermetallics",
  ["intermountain econ rev"] = "Intermountain Economic Review",
  ["intermt j sci"] = "Intermountain journal of sciences : IJS",
  ["intern emerg med"] = "Internal and emergency medicine",
  ["intern j pattern recognit artif intell"] = "International journal of pattern recognition and artificial intelligence",
  ["intern med"] = "Internal medicine (Tokyo, Japan)",
  ["intern med j"] = "Internal medicine journal",
  ["intern med open access"] = "Internal medicine: open access",
  ["intern med rev (wash d c)"] = "Internal medicine review (Washington, D.C. : Online)",
  ["internat appl mech"] = "International Applied Mechanics",
  ["internat cryogenics monogr ser"] = "The International Cryogenics Monograph Series",
  ["internat econom rev"] = "International Economic Review",
  ["internat geophys ser"] = "International Geophysics Series",
  ["internat j adapt control signal process"] = "International Journal of Adaptive Control and Signal Processing",
  ["internat j algebra comput"] = "International Journal of Algebra and Computation",
  ["internat j appl sci comput"] = "International Journal of Applied Science and Computations",
  ["internat j approx reason"] = "International Journal of Approximate Reasoning",
  ["internat j bifur chaos appl sci engrg"] = "International Journal of Bifurcation and Chaos in Applied Sciences and Engineering",
  ["internat j comput geom appl"] = "International Journal of Computational Geometry & Applications",
  ["internat j control"] = "International Journal of Control",
  ["internat j engrg sci"] = "International Journal of Engineering Science",
  ["internat j found comput sci"] = "International Journal of Foundations of Computer Science",
  ["internat j game theory"] = "International Journal of Game Theory",
  ["internat j math"] = "International Journal of Mathematics",
  ["internat j math ed sci tech"] = "International Journal of Mathematical Education in Science and Technology",
  ["internat j modern phys a"] = "International Journal of Modern Physics A. Particles and Fields. Gravitation. Cosmology",
  ["internat j modern phys b"] = "International Journal of Modern Physics B",
  ["internat j modern phys c"] = "International Journal of Modern Physics C. Computational Physics and Physical Computation",
  ["internat j modern phys d"] = "International Journal of Modern Physics. D. Gravitation, Astrophysics, Cosmology",
  ["internat j numer methods engrg"] = "International Journal for Numerical Methods in Engineering",
  ["internat j numer methods fluids"] = "International Journal for Numerical Methods in Fluids",
  ["internat j numer methods heat fluid flow"] = "International Journal of Numerical Methods for Heat & Fluid Flow",
  ["internat j robust nonlinear control"] = "International Journal of Robust and Nonlinear Control",
  ["internat j systems sci"] = "International Journal of Systems Science. Principles and Applications of Systems and Integration",
  ["internat j theoret phys"] = "International Journal of Theoretical Physics",
  ["internat j uncertain fuzziness knowledge-based systems"] = "International Journal of Uncertainty, Fuzziness and Knowledge-Based Systems",
  ["internat lib crit writ econ"] = "The International Library of Critical Writings in Economics",
  ["internat ser appl systems anal"] = "International Series on Applied Systems Analysis",
  ["internat ser monogr phys"] = "International Series of Monographs on Physics",
  ["internat ser numer math"] = "International Series of Numerical Mathematics",
  ["internat ser oper res management sci"] = "International Series in Operations Research & Management Science",
  ["internat ser pure appl phys"] = "International Series in Pure and Applied Physics",
  ["internat ser systems control"] = "International Series on Systems and Control",
  ["internat stud econom model"] = "International Studies in Economic Modelling",
  ["internat trends manufacturing tech"] = "International Trends in Manufacturing Technology",
  ["internat union cryst cryst sympos"] = "International Union of Crystallography Crystallographic Symposia",
  ["internet comput"] = "Internet Computing",
  ["internet healthc strateg"] = "Internet healthcare strategies",
  ["internet high educ"] = "The Internet and higher education",
  ["internet hist"] = "Internet Histories",
  ["internet interv"] = "Internet interventions",
  ["internet interventions"] = "Internet Interventions",
  ["internet j allied health sci pract"] = "The Internet journal of allied health sciences and practice",
  ["internet j chem"] = "Internet Journal of Chemistry",
  ["internet j hematol"] = "The Internet journal of hematology",
  ["internet j ment health"] = "The Internet journal of mental health",
  ["internet j microbiol"] = "The Internet journal of microbiology",
  ["internet j nephrol"] = "The Internet journal of nephrology",
  ["internet j nutr wellness"] = "The Internet journal of nutrition and wellness",
  ["internet j pediatr neonatol"] = "The Internet journal of pediatrics and neonatology",
  ["internet j vib spectro"] = "Internet Journal of Vibrational Spectroscopy",
  ["internet math"] = "Internet Mathematics",
  ["internet technol lett"] = "Internet Technology Letters",
  ["internet things"] = "Internet of Things",
  ["internet things cyber-phys syst"] = "Internet of Things and Cyber-Physical Systems",
  ["internist (berl)"] = "Der Internist",
  ["internist prax"] = "Internistische Praxis",
  ["interpers int j pers relatsh"] = "Interpersona : an international journal of personal relationships",
  ["interpret transl train"] = "The Interpreter and translator trainer",
  ["interpreting (amst)"] = "Interpreting : international journal of research and practice in interpreting",
  ["interserv ind train simul educ conf"] = "The Interservice/Industry Training, Simulation, and Education Conference : I/ITSEC. Interservice/Industry Training, Simulation, and Education Conference",
  ["interstudy qual edge"] = "The Interstudy quality edge : measurement and management of clinical outcomes",
  ["interuniv fac work conf"] = "Inter-University Faculty Work Conference",
  ["interv cardiol"] = "Interventional cardiology (London, England)",
  ["interv cardiol (lond)"] = "Interventional cardiology",
  ["interv cardiol clin"] = "Interventional cardiology clinics",
  ["interv med appl sci"] = "Interventional medicine & applied science",
  ["interv neurol"] = "Interventional neurology",
  ["interv neuroradiol"] = "Interventional neuroradiology : journal of peritherapeutic neuroradiology, surgical procedures and related neurosciences",
  ["interv oncol 360"] = "Interventional oncology 360",
  ["interv psicosoc"] = "Intervención psicosocial",
  ["interv sch clin"] = "Intervention in school and clinic",
  ["intervention (amstelveen)"] = "Intervention (Amstelveen, Netherlands)",
  ["interventional neuroradiol"] = "Interventional Neuroradiology",
  ["interventions (lond)"] = "Interventions (London, England)",
  ["intest res"] = "Intestinal research",
  ["intj metrol qual eng"] = "International Journal of Metrology and Quality Engineering",
  ["intl j disabil dev educ"] = "International journal of disability, development, and education",
  ["intractable rare dis res"] = "Intractable & rare diseases research",
  ["intrinsically disord proteins"] = "Intrinsically disordered proteins",
  ["invariant theory algebr transform groups"] = "Invariant Theory and Algebraic Transformation Groups",
  ["invasion metastasis"] = "Invasion & metastasis",
  ["invasive plant sci manag"] = "Invasive plant science and management",
  ["invasive plant sci manage"] = "Invasive Plant Science and Management",
  ["invent math"] = "Inventiones Mathematicae",
  ["inventions (basel)"] = "Inventions (Basel, Switzerland)",
  ["inventory optim"] = "Inventory Optimization",
  ["inverse ill-posed probl ser"] = "Inverse and Ill-posed Problems Series",
  ["inverse prob"] = "Inverse Problems",
  ["inverse prob eng"] = "Inverse Problems in Engineerign",
  ["inverse prob sci eng"] = "Inverse Problems in Science and Engineering",
  ["inverse probl"] = "Inverse problems",
  ["inverse probl eng"] = "Inverse Problems in Engineering",
  ["inverse probl imaging"] = "Inverse Problems and Imaging",
  ["inverse probl imaging (springfield)"] = "Inverse problems and imaging (Springfield, Mo.)",
  ["inverse probl sci eng"] = "Inverse Problems in Science and Engineering",
  ["inverse problems"] = "Inverse Problems. An International Journal on the Theory and Practice of Inverse Problems, Inverse Methods and Computerized Inversion of Data",
  ["invert neurosci"] = "Invertebrate neuroscience : IN",
  ["invertebr biol"] = "Invertebrate biology : a quarterly journal of the American Microscopical Society and the Division of Invertebrate Zoology/ASZ",
  ["invertebr neurosci"] = "Invertebrate Neuroscience",
  ["invertebr reprod dev"] = "Invertebrate reproduction & development",
  ["invertebr syst"] = "Invertebrate systematics",
  ["invertebrate surviv j"] = "Invertebrate survival journal : ISJ",
  ["invest cell pathol"] = "Investigative & cell pathology",
  ["invest clin"] = "Investigacion Clinica",
  ["invest econ"] = "Investigación económica",
  ["invest educ enferm"] = "Investigación y educación en enfermería",
  ["invest genet"] = "Investigative Genetics",
  ["invest hist"] = "Investigaciones históricas (Universidad de Valladolid. Departamento de Historia Moderna)",
  ["invest math learn"] = "Investigations in Mathematics Learning",
  ["invest med int"] = "Investigación médica internacional",
  ["invest new drugs"] = "Investigational new drugs",
  ["invest ophthalmol"] = "Investigative ophthalmology",
  ["invest ophthalmol vis sci"] = "Investigative ophthalmology & visual science",
  ["invest ophthalmol visual sci"] = "Investigative Ophthalmology & Visual Science",
  ["invest pol"] = "Investment Policy",
  ["invest radiol"] = "Investigative radiology",
  ["invest urol"] = "Investigative urology",
  ["investig cienc"] = "Investigacion y ciencia",
  ["investig clin urol"] = "Investigative and clinical urology",
  ["investig genet"] = "Investigative genetics",
  ["investig magn reson imaging"] = "Investigative magnetic resonance imaging",
  ["investig mar cicimar"] = "Investigaciones marinas CICIMAR",
  ["investig pract pediatr pathol"] = "Investigation and Practice in Pediatric Pathology",
  ["investig urol (berl)"] = "Investigative urology (Berlin, Germany)",
  ["investigacion oper"] = "Investigacion Operacional",
  ["investigaciones econ"] = "Investigaciones Economicas",
  ["investigación econ"] = "Investigación Económica",
  ["investigación oper"] = "Investigación Operacional",
  ["investor owned hosp rev"] = "The Investor-owned hospital review",
  ["inz chem procesowa"] = "Inzynieria Chemiczna i Procesowa",
  ["inzh-fiz zh"] = "Inzhenerno-Fizicheskii Zhurnal",
  ["iobc wprs bull"] = "IOBC/WPRS bulletin = Bulletin OILB/SROP",
  ["iom lett"] = "IOM letters",
  ["ion channels"] = "Ion channels",
  ["ion exch membr"] = "Ion exchange and membranes",
  ["ion exch solvent extr"] = "Ion Exchange and Solvent Extraction",
  ["ionics (kiel)"] = "Ionics",
  ["iop concise phys"] = "IOP Concise Physics",
  ["iop conf ser earth environ sci"] = "IOP conference series. Earth and environmental science",
  ["iop conf ser mater sci eng"] = "IOP conference series. Materials science and engineering",
  ["iop conf ser: earth environ sci"] = "IOP Conference Series: Earth and Environmental Science",
  ["iop conf ser: mater sci eng"] = "IOP Conference Series: Materials Science and Engineering",
  ["iop expand phys"] = "IOP Expanding Physics",
  ["iop scinotes"] = "IOP SciNotes",
  ["iop short meet ser"] = "IOP Short Meetings Series",
  ["iosr j agric vet sci"] = "IOSR journal of agriculture and veterinary science",
  ["iosr j biotechnol biochem"] = "IOSR journal of biotechnology and biochemistry",
  ["iosr j comput eng"] = "IOSR journal of computer engineering",
  ["iosr j environ sci toxicol food technol"] = "IOSR journal of environmental science, toxicology and food technology",
  ["iosr j pharm"] = "IOSR journal of pharmacy",
  ["iosr j pharm biol sci"] = "IOSR journal of pharmacy and biological sciences",
  ["iotdi 2017 (2017)"] = "IoTDI 2017 : 2017 IEEE/ACM Second International Conference on Internet-of-Things Design and Implementation : proceedings : 18-20 April 2017, Pittsburgh, Pennsylvania, USA. IoTDI (Conference) (2nd : 2017 : Pittsburgh, Pa.)",
  ["iowa agric exp stn res bull"] = "IOWA Agricultural Experiment Station Research Bulletin",
  ["iowa code annot iowa"] = "Iowa code annotated. Iowa",
  ["iowa dent bull"] = "Iowa Dental Bulletin",
  ["iowa dent j"] = "The Iowa dental journal",
  ["iowa law rev"] = "Iowa law review",
  ["iowa med"] = "Iowa medicine : journal of the Iowa Medical Society",
  ["iowa orthop j"] = "The Iowa orthopaedic journal",
  ["iowa state coll j sci"] = "Iowa State College journal of science",
  ["iowa state univ vet"] = "Iowa State University veterinarian",
  ["ipa diss ser"] = "IPA Dissertation Series",
  ["ipol j image process online"] = "IPOL Journal. Image Processing Online",
  ["ippa newsl"] = "Ippa - Newsletter",
  ["ippf eur reg inf"] = "Planned Parenthood in Europe regional information bulletin = Planning familial en Europe bulletin d'information régional = Familienplanung in Europe regionale informationen",
  ["ippf med bull"] = "IPPF medical bulletin",
  ["ippf news"] = "IPPF news",
  ["ippf open file"] = "IPPF open file",
  ["ippf whr news serv"] = "IPPF/WHR news service. International Planned Parenthood Federation. Western Hemisphere Region",
  ["ippologia (cremona)"] = "Ippologia : rivista ufficiale SIVE",
  ["ippta j"] = "IPPTA Journal",
  ["ir bank rev"] = "The Irish banking review",
  ["ir econ soc hist"] = "Irish economic and social history : journal of the Economic and Social History Society of Ireland",
  ["ir geogr"] = "Irish geography : bulletin of the Geographical Society of Ireland",
  ["ir hist stud"] = "Irish historical studies : joint journal of the Irish Historical Society and the Ulster Society for Irish Historical Studies",
  ["ir j agric food res"] = "Irish Journal of Agricultural and Food Research",
  ["ir j earth sci"] = "Irish Journal of Earth Sciences",
  ["ir j med sci"] = "Irish journal of medical science",
  ["ir j occup ther"] = "Irish Journal of Occupational Therapy",
  ["ir j psychol"] = "The Irish journal of psychology",
  ["ir j psychol med"] = "Irish journal of psychological medicine",
  ["ir law rep mon"] = "Irish law reports monthly",
  ["ir med j"] = "Irish medical journal",
  ["ir nat j"] = "The Irish naturalists' journal",
  ["ir nurs hosp world"] = "Irish nursing and hospital world",
  ["ir nurs news"] = "The Irish nursing news",
  ["ir nurses j"] = "Irish nurses' journal",
  ["ir stud rev"] = "Irish studies review",
  ["ir sword"] = "The Irish sword : the journal of the Military History Society of Ireland",
  ["ir univ rev"] = "Irish University Review",
  ["ir vet j"] = "Irish veterinary journal",
  ["iral int rev appl linguist lang teach"] = "IRAL, International review of applied linguistics in language teaching : Revue internationale de linguistique appliquée enseignement des langues. Internationale Zeitschrift für angewandte Linguistik in der Spracherziehung",
  ["iran biomed j"] = "Iranian biomedical journal",
  ["iran endod j"] = "Iranian endodontic journal",
  ["iran j allergy asthma immunol"] = "Iranian journal of allergy, asthma, and immunology",
  ["iran j anim biosyst"] = "Iranian journal of animal biosystematics",
  ["iran j appl anim sci"] = "Iranian journal of applied animal science",
  ["iran j arthropod borne dis"] = "Iranian journal of arthropod-borne diseases",
  ["iran j basic med sci"] = "Iranian journal of basic medical sciences",
  ["iran j biotechnol"] = "Iranian journal of biotechnology",
  ["iran j bot"] = "The Iranian journal of botany",
  ["iran j cancer prev"] = "Iranian journal of cancer prevention",
  ["iran j chem chem eng"] = "Iranian Journal of Chemistry & Chemical Engineering",
  ["iran j child neurol"] = "Iranian journal of child neurology",
  ["iran j fish sci"] = "Iranian journal of fisheries sciences",
  ["iran j fuzzy syst"] = "Iranian Journal of Fuzzy Systems",
  ["iran j genet plant breed"] = "Iranian journal of genetics and plant breeding",
  ["iran j ichthyol"] = "Iranian journal of ichthyology",
  ["iran j immunol"] = "Iranian journal of immunology : IJI",
  ["iran j kidney dis"] = "Iranian journal of kidney diseases",
  ["iran j mater sci eng"] = "Iranian Journal of Materials Science and Engineering",
  ["iran j math sci inform"] = "Iranian Journal of Mathematical Sciences and Informatics",
  ["iran j med microbiol"] = "Majallah-i mīkrub/shināsī-i pizishkī-i Īrān",
  ["iran j med sci"] = "Iranian journal of medical sciences",
  ["iran j microbiol"] = "Iranian journal of microbiology",
  ["iran j neurol"] = "Iranian journal of neurology",
  ["iran j nucl med"] = "Majallah-i pizishkī-i hastahī̄ Īrān",
  ["iran j nurs midwifery res"] = "Iranian journal of nursing and midwifery research",
  ["iran j ophthalmol"] = "Iranian journal of ophthalmology",
  ["iran j otorhinolaryngol"] = "Iranian journal of otorhinolaryngology",
  ["iran j parasitol"] = "Iranian journal of parasitology",
  ["iran j pathol"] = "Iranian journal of pathology",
  ["iran j ped hematol oncol"] = "Iranian journal of pediatric hematology and oncology",
  ["iran j pediatr"] = "Iranian journal of pediatrics",
  ["iran j pharm res"] = "Iranian journal of pharmaceutical research : IJPR",
  ["iran j plant pathol"] = "Iranian journal of plant pathology",
  ["iran j plant physiol"] = "Iranian Journal of Plant Physiology",
  ["iran j polym sci technol"] = "Iranian Journal of Polymer Science & Technology",
  ["iran j psychiatry"] = "Iranian journal of psychiatry",
  ["iran j psychiatry behav sci"] = "Iranian journal of psychiatry and behavioral sciences",
  ["iran j public health"] = "Iranian journal of public health",
  ["iran j radiol"] = "Iranian journal of radiology : a quarterly journal published by the Iranian Radiological Society",
  ["iran j reprod med"] = "Iranian journal of reproductive medicine",
  ["iran j sci"] = "Iran Journal of Science",
  ["iran j sci technol trans a sci"] = "Iranian journal of science and technology. Transaction A, Science",
  ["iran j sci technol trans civ eng"] = "Iranian Journal of Science & Technology Transactions of Civil Engineering",
  ["iran j sci technol trans electr eng"] = "Iranian Journal of Science & Technology Transactions of Electrical Engineering",
  ["iran j sci technol trans mech eng"] = "Iranian Journal of Science & Technology Transactions of Mechanical Engineering",
  ["iran j toxicol"] = "Iran Journal of Toxicology",
  ["iran j vet med"] = "Iranian journal of veterinary medicine",
  ["iran j vet res"] = "Iranian journal of veterinary research",
  ["iran j vet sci technol"] = "Iranian journal of veterinary science and technology",
  ["iran j vet surg"] = "Iranian journal of veterinary surgery",
  ["iran j virol"] = "Iranian journal of virology",
  ["iran jcomput sci"] = "Iran Journal of Computer Science",
  ["iran polym j"] = "Iranian Polymer Journal",
  ["iran red crescent med j"] = "Iranian Red Crescent medical journal",
  ["iran stud"] = "Iranian studies : bulletin of the Society for Iranian Cultural and Social Studies",
  ["iranian j environ health sci eng"] = "Iranian journal of environmental health science & engineering",
  ["iranian j sci tech"] = "Iranian Journal of Science and Technology",
  ["iranian journal of basic medical sciences / mashad university of medical sciences|majallah iulum i payah i pizishki i iran"] = "Majallah-i’ulum-i payah-i pizishki-i Iran",
  ["iraqi dent j"] = "Iraqi Dental Journal",
  ["iraqi j med sci"] = "Iraqi journal of medical sciences",
  ["irc med sci reprod obstet gynecol"] = "Ircs Medical Science: Reproduction, Obstetrics And Gynecology",
  ["ircs j med sci"] = "IRCS journal of medical science",
  ["ircs med sci"] = "IRCS medical science",
  ["ircs med sci biomed technol"] = "IRCS medical science: Biomedical technology",
  ["ircs med sci cardiovasc syst"] = "IRCS medical science: Cardiovascular system",
  ["ircs med sci clin med"] = "IRCS medical science. Clinical medicine",
  ["ire trans aeronaut navig electron"] = "IRE Transactions on Aeronautical and Navigational Electronics",
  ["ire trans aerosp navig electron"] = "IRE Transactions on Aerospace and Navigational Electronics",
  ["ire trans anntenas propag"] = "IRE Transactions on Antennas and Propagation",
  ["ire trans audio"] = "IRE Transactions on Audio",
  ["ire trans autom control"] = "IRE Transactions on Automatic Control",
  ["ire trans bio-med electron"] = "IRE Transactions on Bio-Medical Electronics",
  ["ire trans biomed electron"] = "IRE transactions on bio-medical electronics",
  ["ire trans broadcast"] = "IRE Transactions on Broadcasting",
  ["ire trans broadcast telev receivers"] = "IRE Transactions on Broadcast and Television Receivers",
  ["ire trans broadcast transm syst"] = "IRE Transactions on Broadcast Transmission Systems",
  ["ire trans circuit theory"] = "IRE Transactions on Circuit Theory",
  ["ire trans commun syst"] = "IRE Transactions on Communication Systems",
  ["ire trans compon parts"] = "IRE Transactions on Component Parts",
  ["ire trans electron comput"] = "IRE Transactions on Electronic Computers",
  ["ire trans electron devices"] = "IRE Transactions on Electron Devices",
  ["ire trans eng manage"] = "IRE Transactions on Engineering Management",
  ["ire trans hum factors electron"] = "IRE Transactions on Human Factors in Electronics",
  ["ire trans ind electron"] = "IRE Transactions on Industrial Electronics",
  ["ire trans inf theory"] = "IRE Transactions on Information Theory",
  ["ire trans instrum"] = "IRE Transactions on Instrumentation",
  ["ire trans med electron"] = "IRE transactions on medical electronics",
  ["ire trans microwave theory tech"] = "IRE Transactions on Microwave Theory and Techniques",
  ["ire trans mil electron"] = "IRE Transactions on Military Electronics",
  ["ire trans nucl sci"] = "IRE Transactions on Nuclear Science",
  ["ire trans prod eng prod"] = "IRE Transactions on Product Engineering and Production",
  ["ire trans prod tech"] = "IRE Transactions on Production Techniques",
  ["ire trans radio freq interference"] = "IRE Transactions on Radio Frequency Interference",
  ["ire trans reliab qual control"] = "IRE Transactions on Reliability and Quality Control",
  ["ire trans space electron telem"] = "IRE Transactions on Space Electronics and Telemetry",
  ["ire trans telem remote control"] = "IRE Transactions on Telemetry and Remote Control",
  ["ire trans ultrason eng"] = "IRE Transactions on Ultrasonic Engineering",
  ["ire trans veh commun"] = "IRE Transactions on Vehicular Communications",
  ["irish banking rev"] = "Irish Banking Review",
  ["irish j agr econ rural sociology"] = "Irish Journal of Agricultural Economics and Rural Sociology",
  ["irish math soc bull"] = "Irish Mathematical Society Bulletin",
  ["irish theol q"] = "The Irish theological quarterly",
  ["irish times"] = "The Irish times",
  ["irma lect math theor phys"] = "IRMA Lectures in Mathematics and Theoretical Physics",
  ["iron steel technol"] = "Iron & Steel Technology",
  ["ironmaking steelmaking"] = "Ironmaking and Steelmaking",
  ["irrig drain"] = "Irrigation and drainage (International Commission on Irrigation and Drainage)",
  ["irrig drain syst"] = "Irrigation and Drainage Systens",
  ["irrig sci"] = "Irrigation science",
  ["is&t int symp electron imaging"] = "IS&T International Symposium on Electronic Imaging",
  ["isa trans"] = "ISA transactions",
  ["isaac newton inst ser lectures"] = "The Isaac Newton Institute Series of Lectures",
  ["isbt sci ser"] = "ISBT science series",
  ["iscb community j"] = "ISCB Community Journal",
  ["isci notes"] = "iScience notes",
  ["ise rev"] = "ISE Review",
  ["ish j hydraul eng"] = "ISH Journal of Hydraulic Engineering",
  ["ishikawaken hakusan shizen hogo senta kenkyu hokoku"] = "Ishikawa-ken Hakusan Shizen Hogo Sentā kenkyū hōkoku",
  ["isij int"] = "ISIJ International",
  ["isl arc"] = "Island Arc",
  ["isl law rev"] = "ISL law review. International School of Law",
  ["isl mag"] = "The Island magazine",
  ["islam philos theol sci"] = "Islamic Philosophy, Theology, and Science",
  ["islam q"] = "The Islamic quarterly",
  ["islam stud"] = "Islamic studies",
  ["islamic econ stud"] = "Islamic Economic Studies",
  ["isme commun"] = "ISME Communications",
  ["isme glomis electron j"] = "ISME/GLOMIS electronic journal",
  ["isme j"] = "The ISME journal",
  ["isokinet exerc sci"] = "Isokinetics and exercise science",
  ["isokinet exercise sci"] = "Isokinetics and Exercise Science",
  ["isot environ health stud"] = "Isotopes in Environmental and Health Studies",
  ["isot geosci"] = "Isotope Geoscience",
  ["isot radiat technol"] = "Isotopes and Radiation Technology",
  ["isotopes environ health stud"] = "Isotopes in environmental and health studies",
  ["isozyme bull"] = "Isozyme Bulletin",
  ["isozymes curr top biol med res"] = "Isozymes",
  ["ispra courses materials, engrg mech sci"] = "Ispra Courses on Materials, Engineering and Mechanical Science",
  ["ispra courses nuclear engrg tech"] = "Ispra Courses on Nuclear Engineering and Technology",
  ["ispra courses reliab risk anal"] = "Ispra Courses on Reliability and Risk Analysis",
  ["isprs ann photogramm remote sens spatial inf sci"] = "ISPRS Annals of the Photogrammetry, Remote Sensing and Spatial Information Sciences",
  ["isprs int j geo-inf"] = "ISPRS International Journal of Geo-Information",
  ["isprs int j geoinf"] = "ISPRS international journal of geo-information",
  ["isprs j photogramm remote sens"] = "ISPRS Journal of Photogrammetry and Remote Sensing",
  ["isprs open j photogramm remote sens"] = "ISPRS Open Journal of Photogrammetry and Remote Sensing",
  ["isr ann psychiatr relat discip"] = "The Israel annals of psychiatry and related disciplines",
  ["isr j aquac"] = "The Israeli journal of aquaculture = Bamidgeh",
  ["isr j chem"] = "Israel Journal of Chemistry",
  ["isr j dent med"] = "Israel Journal of Dental Medicine",
  ["isr j dent sci"] = "Israel journal of dental sciences",
  ["isr j ecol evol"] = "Israel journal of ecology & evolution",
  ["isr j entomol"] = "Israel journal of entomology",
  ["isr j exp med"] = "Israel journal of experimental medicine",
  ["isr j health policy res"] = "Israel journal of health policy research",
  ["isr j math"] = "Israel Journal of Mathematics",
  ["isr j med sci"] = "Israel journal of medical sciences",
  ["isr j plant sci"] = "Israel journal of plant sciences",
  ["isr j psychiatry relat sci"] = "The Israel journal of psychiatry and related sciences",
  ["isr j zool"] = "Israel journal of zoology",
  ["isr law rev"] = "Israel law review",
  ["isr med assoc j"] = "The Israel Medical Association journal : IMAJ",
  ["isr med j"] = "Israel medical journal",
  ["isr orient stud"] = "Israel oriental studies",
  ["isr soc sci res"] = "Israel social science research",
  ["isra 2013 (2013)"] = "ISRA 2013",
  ["israel j math"] = "Israel Journal of Mathematics",
  ["israel law rep"] = "Israel law reports. Israel. Bet ha-mishpaṭ ha-ʻelyon",
  ["israel math conf proc"] = "Israel Mathematical Conference Proceedings",
  ["israel stud (bloomingt ind)"] = "Israel studies (Bloomington, Ind.)",
  ["isrn addict"] = "ISRN Addiction",
  ["isrn aids"] = "ISRN AIDS",
  ["isrn allergy"] = "ISRN allergy",
  ["isrn anat"] = "ISRN anatomy",
  ["isrn anesthesiol"] = "ISRN anesthesiology",
  ["isrn biochem"] = "ISRN biochemistry",
  ["isrn bioinform"] = "ISRN bioinformatics",
  ["isrn biomath"] = "ISRN biomathematics",
  ["isrn biotechnol"] = "ISRN biotechnology",
  ["isrn cardiol"] = "ISRN cardiology",
  ["isrn cell biol"] = "ISRN cell biology",
  ["isrn comput biol"] = "ISRN computational biology",
  ["isrn dent"] = "ISRN dentistry",
  ["isrn dermatol"] = "ISRN dermatology",
  ["isrn ecol"] = "ISRN ecology",
  ["isrn endocrinol"] = "ISRN endocrinology",
  ["isrn entomol"] = "ISRN entomology",
  ["isrn family med"] = "ISRN family medicine",
  ["isrn gastroenterol"] = "ISRN gastroenterology",
  ["isrn genetics"] = "ISRN Genetics",
  ["isrn geriatr"] = "ISRN geriatrics",
  ["isrn hematol"] = "ISRN hematology",
  ["isrn hepatol"] = "ISRN hepatology",
  ["isrn infect dis"] = "ISRN infectious diseases",
  ["isrn inflamm"] = "ISRN inflammation",
  ["isrn microbiol"] = "ISRN microbiology",
  ["isrn minim invasive surg"] = "ISRN minimally invasive surgery",
  ["isrn mol biol"] = "ISRN molecular biology",
  ["isrn mol imaging"] = "ISRN Molecular Imaging",
  ["isrn nanomater"] = "ISRN nanomaterials",
  ["isrn nephrol"] = "ISRN nephrology",
  ["isrn neurol"] = "ISRN neurology",
  ["isrn neurosci"] = "ISRN neuroscience",
  ["isrn nurs"] = "ISRN nursing",
  ["isrn nutr"] = "ISRN nutrition",
  ["isrn obes"] = "ISRN obesity",
  ["isrn obstet gynecol"] = "ISRN obstetrics and gynecology",
  ["isrn oncol"] = "ISRN oncology",
  ["isrn ophthalmol"] = "ISRN ophthalmology",
  ["isrn opt"] = "ISRN Optics",
  ["isrn org chem"] = "ISRN organic chemistry",
  ["isrn orthop"] = "ISRN orthopedics",
  ["isrn otolaryngol"] = "ISRN otolaryngology",
  ["isrn oxidative med"] = "ISRN oxidative medicine",
  ["isrn pain"] = "ISRN Pain",
  ["isrn parasitol"] = "ISRN parasitology",
  ["isrn pathol"] = "ISRN pathology",
  ["isrn pediatr"] = "ISRN pediatrics",
  ["isrn pharm"] = "ISRN pharmaceutics",
  ["isrn pharmacol"] = "ISRN pharmacology",
  ["isrn physiol"] = "ISRN Physiology",
  ["isrn prev med"] = "ISRN preventive medicine",
  ["isrn psychiatry"] = "ISRN psychiatry",
  ["isrn public health"] = "ISRN public health",
  ["isrn pulmonol"] = "ISRN pulmonology",
  ["isrn radiol"] = "ISRN radiology",
  ["isrn rehabil"] = "ISRN Rehabilitation",
  ["isrn rheumatol"] = "ISRN rheumatology",
  ["isrn stroke"] = "ISRN stroke",
  ["isrn struct biol"] = "ISRN structural biology",
  ["isrn surg"] = "ISRN surgery",
  ["isrn tissue eng"] = "ISRN tissue engineering",
  ["isrn toxicol"] = "ISRN toxicology",
  ["isrn trop med"] = "ISRN tropical medicine",
  ["isrn urol"] = "ISRN urology",
  ["isrn vet sci"] = "ISRN veterinary science",
  ["isrn virol"] = "ISRN virology",
  ["isrn zool"] = "ISRN zoology",
  ["issbd bull"] = "ISSBD bulletin",
  ["isscs 2013 (2013)"] = "ISSCS 2013 : International Symposium on Signals, Circuits and Systems : 11-12 July, 2013, Iasi, Romania : program. International Symposium on Signals, Circuits, and Systems (12th : 2013 : Iași, Romania)",
  ["issled genet"] = "Issledovaniia po genetike",
  ["issled operatsiy i asu"] = "Kievskiy Gosudarstvennyy Universitet",
  ["issled prikl mat"] = "Kazanskiy Universitet",
  ["issled prikl mat inform"] = "Issledovaniya po Prikladnoi Matematike i Informatike",
  ["issled teor plastin i obolochek"] = "Kazanskiy Universitet",
  ["isss j micro smart syst"] = "ISSS journal of micro and smart systems",
  ["issue brief (commonw fund)"] = "Issue brief (Commonwealth Fund)",
  ["issue brief (george wash univ med cent ensuring solut alcohol probl)"] = "Issue brief (George Washington University. Medical Center. Ensuring Solutions to Alcohol Problems)",
  ["issue brief (grantmakers health)"] = "Issue brief (Grantmakers in Health)",
  ["issue brief (inst health care costs solut)"] = "Issue brief (Institute on Health Care Costs and Solutions)",
  ["issue brief (mass health policy forum)"] = "Issue brief (Massachusetts Health Policy Forum)",
  ["issue brief (public policy inst (am assoc retired pers))"] = "Issue Brief (Public Policy Institute (American Association of Retired Persons))",
  ["issue brief cent medicare educ"] = "Issue brief (Center for Medicare Education)",
  ["issue brief cent stud health syst change"] = "Issue brief (Center for Studying Health System Change)",
  ["issue brief george wash univ cent health serv res policy"] = "Issue brief (George Washington University. Center for Health Services Research and Policy)",
  ["issue brief george wash univ natl health policy forum"] = "Issue brief (George Washington University. National Health Policy Forum : 2005)",
  ["issue brief health policy track serv"] = "Issue brief (Health Policy Tracking Service)",
  ["issue brief natl health policy forum"] = "Issue brief (George Washington University. National Health Policy Forum)",
  ["issues (st louis mo)"] = "Issues (Saint Louis, Mo.)",
  ["issues brief (alan guttmacher inst)"] = "Issues in brief (Alan Guttmacher Institute)",
  ["issues ciminol"] = "Issues in criminology",
  ["issues compr pediatr nurs"] = "Issues in comprehensive pediatric nursing",
  ["issues emerg health technol"] = "Issues in emerging health technologies",
  ["issues ethics"] = "Issues in ethics",
  ["issues health care"] = "Issues in health care (Philadelphia, Pa.)",
  ["issues health care women"] = "Issues in health care of women",
  ["issues law med"] = "Issues in law & medicine",
  ["issues math ed"] = "Issues in Mathematics Education",
  ["issues med ethics"] = "Issues in medical ethics",
  ["issues ment health nurs"] = "Issues in mental health nursing",
  ["issues race soc"] = "Issues in race & society : an interdisciplinary global journal",
  ["issues reprod genet eng"] = "Issues in reproductive and genetic engineering : journal of international feminist analysis",
  ["issues sci technol"] = "Issues in science and technology",
  ["issues stud"] = "Issues & studies",
  ["issues subj teach ser"] = "Issues in Subject Teaching Series",
  ["ist int surf technol"] = "IST International Surface Technology",
  ["istanb univ sci fac j math phys astronom"] = "Istanbul University. Science Faculty. The Journal of Mathematics, Physics and Astronomy",
  ["istanbul tip fak mecmuasi"] = "Istanbul Tıp Fakültesi mecmuasi",
  ["istanbul univ dishekim fak derg"] = "İstanbul Üniversitesi Dişhekimliği Fakültesi dergisi = The journal of the Dental Faculty of Istanbul",
  ["istit lombardo accad sci lett rend a"] = "Istituto Lombardo",
  ["istit ric base ser monogr adv math"] = "Istituto per la Ricerca di Base",
  ["istit veneto sci lett arti atti cl sci fis mat natur"] = "Istituto Veneto di Scienze, Lettere ed Arti",
  ["istor arh (mosk 1992)"] = "Istoricheskiĭ arkhiv (Moscow, Russia : 1992)",
  ["istor biol issled"] = "Istoriko-biologicheskie issledovaniia",
  ["istor cas"] = "Istorijski časopis : organ Istorijskog instituta Srpske akademije nauka i umetnosti",
  ["istor metodol estestv nauk"] = "Istoriya i Metodologiya Estestvennykh Nauk",
  ["istor nauki ross xx veka"] = "\\cyr Istoriki Nauki Rossii XX \\cyr Veka",
  ["istor pregl"] = "Istoricheski pregled",
  ["istor sssr"] = "Istorii︠a︡ SSSR (Moscow, Russia : 1957)",
  ["istor zapisi"] = "Istorijski zapisi : organ Istoriskog instituta i Društva istoričara SR Crne Gore",
  ["istor-astronom issled"] = "\\cyr Istoriko-Astronomicheskie Issledovaniya",
  ["istor-mat issled"] = "Istoriko-Matematicheskie Issledovaniya",
  ["isv sarat univ nov ser ser him biol ekol"] = "Izvestii︠a︡ Saratovskogo universiteta. Novai︠a︡ serii︠a︡. Serii︠a︡ khimii︠a︡. Biologii︠a︡. Ėkologii︠a︡",
  ["it prof"] = "IT Professional",
  ["itab corfu greece (2010)"] = "ITAB 2010 Corfu, Greece : 10th International Conference on Information Technology and Applications in Biomedicine : Emerging Technologies for Patient Specific Healthcare : 2-5 November 2010, Aquis Corfu Holiday Palace Hotel, Greece. International Conference on Information Technology and Applications in Biomedicine (10th : 2010 : Corfu, Greece)",
  ["ital botan"] = "Italian Botanist",
  ["ital contemp"] = "Italia contemporanea",
  ["ital for mont"] = "Italia forestale e montana, L’",
  ["ital gen rev dermatol"] = "Italian general review of dermatology",
  ["ital gen rev otorhinolaryngol"] = "Italian general review of oto-rhino-laryngology. Revue générale italienne d'oto-rhino-laryngologie",
  ["ital heart j"] = "Italian heart journal : official journal of the Italian Federation of Cardiology",
  ["ital heart j suppl"] = "Italian heart journal. Supplement : official journal of the Italian Federation of Cardiology",
  ["ital j aerosp med"] = "Italian journal of aerospace medicine",
  ["ital j agrometeorol"] = "Italian Journal of Agrometeorology",
  ["ital j agron"] = "Italian Journal of Agronomy",
  ["ital j anat embryol"] = "Italian Journal of Anatomy and Embryology",
  ["ital j anim sci"] = "Italian journal of animal science",
  ["ital j biochem"] = "Italian Journal of Biochemistry",
  ["ital j eng geol environ"] = "Italian Journal Of Engineering Geology And Environment",
  ["ital j food saf"] = "Italian journal of food safety",
  ["ital j food sci"] = "Italian Journal of Food Science",
  ["ital j gastroenterol"] = "The Italian journal of gastroenterology",
  ["ital j gastroenterol hepatol"] = "Italian journal of gastroenterology and hepatology",
  ["ital j geosci"] = "Italian Journal of Geoscience",
  ["ital j gynaecol obstet"] = "Italian journal of gynaecology & obstetrics : official publication of the Società italiana di ginecologia e ostetricia (SIGO)",
  ["ital j lab med"] = "Italian Journal of Laboratory Medicine",
  ["ital j mycol"] = "Italian journal of mycology",
  ["ital j neurol sci"] = "Italian journal of neurological sciences",
  ["ital j orthop traumatol"] = "Italian journal of orthopaedics and traumatology",
  ["ital j orthop traumatol suppl"] = "Italian journal of orthopaedics and traumatology. Supplementum",
  ["ital j pediatr"] = "Italian journal of pediatrics",
  ["ital j pure appl math"] = "Italian Journal of Pure and Applied Mathematics",
  ["ital j surg sci"] = "The Italian journal of surgical sciences",
  ["ital j zool (modena)"] = "The Italian journal of zoology",
  ["ital medioev um"] = "Italia medioevale e umanistica",
  ["ital rev agric econ"] = "Italian Review of Agricultural Economics",
  ["ital stud"] = "Italian studies",
  ["italy doc notes"] = "Italy; documents and notes",
  ["itch (phila)"] = "Itch (Philadelphia, Pa.)",
  ["ite lett batteries new technol med"] = "ITE Letters on Batteries New Technologies & Medicine",
  ["ite-iba lett batteries new technol med"] = "ITE-IBA Letters New Technologies & Medicine",
  ["items issues"] = "Items & issues",
  ["itogi nauk"] = "Itogi nauki",
  ["itogi nauki i tekhniki"] = "\\cyr Itogi Nauki i Tekhniki",
  ["itogi nauki tekh ser sovrem mat prilozh temat obz"] = "Itogi Nauki i Tekhniki. Seriya Sovremennaya Matematika i ee Prilozheniya. Tematicheskie Obzory",
  ["itogi nauki yug ross ser mat monogr"] = "Itogi Nauki. Yug Rossii. Matematicheskaya Monografiya",
  ["itu j (geneva)"] = "ITU journal : ICT discoveries",
  ["iubmb life"] = "IUBMB life",
  ["iucr monogr crystallogr"] = "IUCr Monographs on Crystallography",
  ["iucr texts crystallogr"] = "IUCr Texts on Crystallography",
  ["iugosl physiol pharmacol acta"] = "Iugoslavica physiologica et pharmacologica acta",
  ["iup j sci technol"] = "IUP journal of science and technology",
  ["iussp pap"] = "IUSSP papers",
  ["iutam bookser"] = "IUTAM Bookseries",
  ["ivd technol"] = "IVD technology : for in vitro diagnostics development and manufacturing",
  ["ivy j ethics"] = "Ivy journal of ethics",
  ["iwanami ser mod math"] = "Iwanami Series in Modern Mathematics",
  ["iwk int wiss korresp gesch dtsch arb beweg"] = "Internationale wissenschaftliche Korrespondenz zur Geschichte der deutschen Arbeiterbewegung",
  ["iyakuhin kenkyu"] = "Iyakuhin Kenkyu",
  ["iyodenshi to seitai kogaku"] = "Iyō denshi to seitai kōgaku. Japanese journal of medical electronics and biological engineering",
  ["iza j labor econ"] = "IZA journal of labor economics",
  ["iza j labor policy"] = "IZA journal of labor policy",
  ["iza world labor"] = "IZA world of labor : evidence-based policy making",
  ["izdan bibl"] = "Izdanja Biblioteke",
  ["izmer tekh"] = "Izmeritel’naya Tekhnika",
  ["izv acad nauk sssr, fiz atmos okeana [izv acad sci ussr, atmos oceanic phys]"] = "Izvestiya Akademii Nauk SSSR, Fizika Atmosfery i Okeana [Izvestiya, Academy of Sciences, USSR, Atmospheric and Oceanic Physics]",
  ["izv acad nauk sssr, fiz zemli [izv acad sci ussr, phys solid earth]"] = "Izvestiya Akademii Nauk SSSR, Fizika Zemli [Izvestiya, Academy of Sciences, USSR, Physics of the Solid Earth]",
  ["izv acad nauk sssr, ser fiz [bull acad sci ussr, phys ser]"] = "Izvestiya Akademii Nauk SSSR, Seriya Fizicheskaya [Bulletin of the Academy of Sciences of the USSR, Physical Series]",
  ["izv acad sci ussr, atmos oceanic phys"] = "Izvestiya, Academy of Sciences, USSR, Atmospheric and Oceanic Physics [translation of Izvestiya Akademii Nauk SSSR, Fizika Atmosfery i Okeana]",
  ["izv acad sci ussr, phys solid earth"] = "Izvestiya, Academy of Sciences, USSR, Physics of the Solid Earth [translation of Izvestiya Akademii Nauk SSSR, Fizika Zemli]",
  ["izv akad gruz ssr ser biol"] = "Izvestiia Akademii nauk Gruzinskoĭ SSR. Seriia biologicheskaia",
  ["izv akad nauk gruz ssr, ser khim"] = "Izvestiya Akademii Nauk Gruzii, Seriya Khimicheskaya",
  ["izv akad nauk kaz ssr, ser khim"] = "Izvestiya Akademii Nauk Respubliki Kazakhstan, Seriya Khimicheskaya",
  ["izv akad nauk kirg ssr [biol]"] = "Izvestiia Akademii nauk Kirgizskoi SSR. Seriia biologicheskikh nauk",
  ["izv akad nauk kirg ssr biol"] = "Izvestiia Akademii nauk Kirgizskoĭ SSR. Seriia biologicheskikh nauk",
  ["izv akad nauk respub armeniya mekh"] = "Akademiya Nauk Respubliki Armeniya",
  ["izv akad nauk respub kyrgyzstan"] = "Akademiya Nauk Respubliki Kyrgyzstan",
  ["izv akad nauk respub moldova fiz tekhn"] = "Akademiya Nauk Respubliki Moldova",
  ["izv akad nauk respub moldova mat"] = "Akademiya Nauk Respubliki Moldova",
  ["izv akad nauk ser biol"] = "Izvestiia Akademii Nauk. Seriia Biologicheskaia",
  ["izv akad nauk sssr biol"] = "Izvestiia Akademii nauk SSSR. Seriia biologicheskaia",
  ["izv akad nauk sssr otdelenie khim nauk"] = "Izvestiia Akademii nauk SSSR, Otdelenie khimicheskikh nauk",
  ["izv akad nauk sssr ser geogr"] = "Izvestii︠a︡ Akademii nauk SSSR. Serii︠a︡ geograficheskai︠a︡",
  ["izv akad nauk sssr ser khim"] = "Izvestiia Akademii nauk SSSR. Seriia khimicheskaia",
  ["izv akad nauk sssr, neorg mater [inorg mater (ussr)]"] = "Izvestiya Akademii Nauk SSSR, Neorganicheskie Materialy [Inorganic Materials (USSR)]",
  ["izv akad nauk sssr, ser geol"] = "Izvestiya Akademii Nauk, Seriya Geologicheskaya",
  ["izv akad nauk, ser fiz"] = "Izvestiya Akademii Nauk, Seriya Fizicheskaya",
  ["izv atmos ocean phys"] = "Izvestiya, Atmospheric and Oceanic Physics",
  ["izv atmos oceanic phys"] = "Izvestiya Atmospheric and Oceanic Physics",
  ["izv bulg akad naukite sofia"] = "Izvestiia. Bŭlgarska akademiia na naukite, Sofia. Institut za sotsialna meditsina",
  ["izv bulg akad naukite sofia nauchnoizsledovatelski inst protivorak antibiot"] = "Izvestiia. Nauchnoizsledovatelski institut za protivorakovi antibiotit︠s︡i (Bulgarska akademii︠a︡ na naukite)",
  ["izv chelyabinsk nauchn tsentra"] = "Izvestiya Chelyabinskogo Nauchnogo Tsentra. Proceedings of the Chelyabinsk Scientific Center",
  ["izv inst fiziol (sofiia)"] = "Izvestiia na Instituta po fiziologiia",
  ["izv inst istor bkp"] = "Izvestii︠a︡ na Instituta po istorii︠a︡ na BKP",
  ["izv inst mat inform"] = "Izvestiya Instituta Matematiki i Informatiki",
  ["izv irkutsk gos univ ser mat"] = "Izvestiya Irkutskogo Gosudarstvennogo Universiteta. Seriya Matematika",
  ["izv math"] = "Izvestiya. Mathematics",
  ["izv meditsinskite inst bulg akad naukite sofia otd biol meditsinski nauki"] = "Izvestiia na meditsinskite instituti. [Bulletin des instituts de médecine]. Bŭlgarska akademii︠a︡ na naukite. Otdelenie za biologicheski i medit︠s︡inski nauki",
  ["izv mikrobiol inst (sofiia)"] = "Izvestiia na Mikrobiologicheskiia institut",
  ["izv minister nauki akad nauk resp kaz ser fiz-mat"] = "Izvestiya Ministerstva Nauki–Akademii Nauk Respubliki Kazakhstan",
  ["izv nats akad nauk armen mekh"] = "Izvestiya Natsionalnoi Akademii Nauk Armenii. Mekhanika. Proceedings of National Academy of Sciences of Armenia. Mechanics",
  ["izv nats akad nauk armenii mat"] = "Natsionalnaya Akademiya Nauk Armenii. Izvestiya. Matematika",
  ["izv nats akad nauk resp kaz ser fiz-mat"] = "Izvestiya Natsionalnoi Akademii Nauk Respubliki Kazakhstan. Seriya Fiziko-Matematicheskaya",
  ["izv nats akad nauk resp kaz, ser khim"] = "Izvestiya Natsional’noi Akademii Nauk Respubliki Kazakhstan, Seriya Khimicheskaya",
  ["izv nats akad nauk resp kaz, ser khim tekhnol"] = "Izvestiya Natsional’noi Akademii Nauk Respubliki Kazakhstan, Seriya Khimii i Tekhnologii",
  ["izv phys solid earth"] = "Izvestiya Physics of the Solid Earth",
  ["izv ross akad nauk mekh zhidk gaza"] = "Rossiiskaya Akademiya Nauk. Izvestiya. Mekhanika Zhidkosti i Gaza",
  ["izv ross akad nauk ser mat"] = "Rossiiskaya Akademiya Nauk. Izvestiya. Seriya Matematicheskaya",
  ["izv ross akad nauk teor sist upr"] = "Izvestiya Rossiiskoi Akademii Nauk. Teoriya i Sistemy Upravleniya",
  ["izv ross akad nauk, ser fiz"] = "Izvestiya Rossiiskoi Akademii Nauk, Seriya Fizicheskaya",
  ["izv russ geogr obshchestva"] = "Izvestii︠a︡ Russkogo geograficheskogo obshchestva (1992)",
  ["izv samar naucn centra ross akad nauk"] = "Izvestii︠a︡ Samarskogo nauchnogo t︠s︡entra Rossiĭskoĭ akademii nauk",
  ["izv sarat univ (ns) ser mat mekh inform"] = "Izvestiya Saratovskogo Universiteta. Novaya Seriya. Seriya. Matematika. Mekhanika. Informatika",
  ["izv seriia fiziol meditsiny qazaq ssr ghylum akad"] = "Izvestiia. Seriia fiziologii i meditsiny. Qazaq SSR Ghylym Akademiiasy",
  ["izv tekhn univ plovdiv fund nauk prilozhen"] = "Tekhnicheskiya Universitet v Plovdiv",
  ["izv timirjazevsk selskokhoziaistvennoi akad"] = "Izvestii︠a︡ Timiri︠a︡zevskoĭ selʹskokhozi︠a︡ĭstvennoĭ akademii",
  ["izv vses geogr obshchestva"] = "Izvestii︠a︡ Vsesoi︠u︡znogo geograficheskogo obshchestva. Geograficheskoe obshchestvo Soi︠u︡za SSR",
  ["izv vyssh uchebn zavad fiz [sov phys j]"] = "Izvestiya Vysshikh Uchebnykh Zavedenii, Fizika [Soviet Physics Journal]",
  ["izv vyssh uchebn zaved fiz"] = "Izvestiya Vysshikh Uchebnykh Zavedeniy",
  ["izv vyssh uchebn zaved mat"] = "Izvestiya Vysshikh Uchebnykh Zavedenii. Matematika. Kazanskii Gosudarstvennyi Universitet",
  ["izv vyssh uchebn zaved prikl neliney n dinamika"] = "Izvestiya Vysshikh Uchebnykh Zavedeniy",
  ["izv vyssh uchebn zaved radiofiz"] = "Izvestiya Vysshikh Uchebnykh Zavedeniy",
  ["izv vyssh uchebn zaved radiofiz [sov radiophys]"] = "Izvestiya Vysshikh Uchebnykh Zavedenii, Radiofizika [Soviet Radiophysics]",
  ["izv vyssh uchebn zaved sev-kavk reg estestv nauki"] = "Izvestiya Vysshikh Uchebnykh Zavedeniy",
  ["izv vyssh uchebn zaved severo-kavkaz reg estestv nauk"] = "Izvestiya Vysshikh Uchebnykh Zavedeniy",
  ["izv vyssh uchebn zaved, chern metall"] = "Izvestiya Vysshikh Uchebnykh Zavedenii, Chernaya Metallurgiya",
  ["izv vyssh uchebn zaved, fiz"] = "Izvestiya Vysshikh Uchebnykh Zavedenii, Fizika",
  ["izv vyssh uchebn zaved, khim khim tekhnol"] = "Izvestiya Vysshikh Uchebnykh Zavedenii, Khimiya I Khimicheskaya Tekhnologiya",
  ["izv vyssh uchebn zaved, tsvetn metall"] = "Izvestiya Vysshikh Uchebnykh Zavedenii, Tsvetnaya Metallurgiya",
  ["j (basel)"] = "J",
  ["j (inst health rec inf manag)"] = "Journal (Institute of Health Record & Information Management : 2008)",
  ["j 3d print med"] = "Journal of 3D printing in medicine",
  ["j aapos"] = "Journal of AAPOS : the official publication of the American Association for Pediatric Ophthalmology and Strabismus",
  ["j abbasid stud"] = "Journal of Abbasid studies",
  ["j abdom surg"] = "The Journal of abdominal surgery",
  ["j abnorm child psychol"] = "Journal of abnormal child psychology",
  ["j abnorm psychol"] = "Journal of abnormal psychology",
  ["j abnorm soc psychol"] = "Journal of abnormal and social psychology",
  ["j aborig health"] = "Journal of aboriginal health",
  ["j abstr differ equ appl"] = "Journal of Abstract Differential Equations and Applications",
  ["j acad consult liaison psychiatry"] = "Journal of the Academy of Consultation-Liaison Psychiatry",
  ["j acad ethics"] = "Journal of academic ethics",
  ["j acad gen dent"] = "Journal - Academy of General Dentistry",
  ["j acad hosp adm"] = "Journal (Academy of Hospital Administration (India))",
  ["j acad mark sci"] = "Journal of the Academy of Marketing Science",
  ["j acad nutr diet"] = "Journal of the Academy of Nutrition and Dietetics",
  ["j acad ophthalmol"] = "Journal of academic ophthalmology (2017)",
  ["j acad rehabil audiol"] = "Journal of the Academy of Rehabilitative Audiology",
  ["j acc"] = "Journal of Accounting, Auditing and Finance",
  ["j acc econ"] = "Journal of Accounting and Economics",
  ["j acc res"] = "Journal of Accounting Research",
  ["j access serv"] = "Journal of access services",
  ["j accid emerg med"] = "Journal of accident & emergency medicine",
  ["j achiev mater manuf eng"] = "Journal Of Achievements In Materials and Manufacturing Engineering",
  ["j acm"] = "Journal of the ACM",
  ["j acoust soc am"] = "The Journal of the Acoustical Society of America",
  ["j acquir immune defic syndr"] = "Journal of acquired immune deficiency syndromes (1999)",
  ["j acquir immune defic syndr (1988)"] = "Journal of acquired immune deficiency syndromes",
  ["j acquir immune defic syndr hum retrovirol"] = "Journal of acquired immune deficiency syndromes and human retrovirology : official publication of the International Retrovirology Association",
  ["j act sedentary sleep behav"] = "Journal of Activity, Sedentary and Sleep Behaviors",
  ["j acupunct meridian stud"] = "Journal of acupuncture and meridian studies",
  ["j acute care phys ther"] = "Journal of acute care physical therapy",
  ["j acute med"] = "Journal of acute medicine",
  ["j addict"] = "Journal of addiction",
  ["j addict behav ther"] = "Journal of addictive behaviors and therapy",
  ["j addict behav ther rehabil"] = "Journal of addictive behaviors, therapy & rehabilitation",
  ["j addict depend"] = "Journal of addiction and dependence",
  ["j addict dis"] = "Journal of addictive diseases",
  ["j addict med"] = "Journal of addiction medicine",
  ["j addict med ther sci"] = "Journal of addiction medicine and therapeutic science",
  ["j addict nurs"] = "Journal of addictions nursing",
  ["j addict offender couns"] = "Journal of addictions & offender counseling",
  ["j addict prev"] = "Journal of addiction & prevention",
  ["j addict prev med"] = "Journal of addiction and preventive medicine",
  ["j addict res (opast group)"] = "Journal of addiction research (OPAST Group)",
  ["j addict res ther"] = "Journal of Addiction Research & Therapy",
  ["j addict sci"] = "Journal of addiction science",
  ["j adel bot gard"] = "Journal of the Adelaide Botanic Gardens",
  ["j adenocarcinoma"] = "Journal of adenocarcinoma",
  ["j adhd relat disord"] = "Journal of ADHD & related disorders",
  ["j adhes"] = "Journal of Adhesion",
  ["j adhes dent"] = "The journal of adhesive dentistry",
  ["j adhes sci technol"] = "Journal of Adhesion Science and Technology",
  ["j adolesc"] = "Journal of adolescence",
  ["j adolesc adult lit"] = "Journal of adolescent & adult literacy : a journal from the International Reading Association",
  ["j adolesc health"] = "The Journal of adolescent health : official publication of the Society for Adolescent Medicine",
  ["j adolesc health care"] = "Journal of adolescent health care : official publication of the Society for Adolescent Medicine",
  ["j adolesc res"] = "Journal of adolescent research",
  ["j adolesc young adult oncol"] = "Journal of adolescent and young adult oncology",
  ["j adult dev"] = "Journal of adult development",
  ["j adv acad"] = "Journal of advanced academics",
  ["j adv biol"] = "Journal of advances in biology",
  ["j adv biol biotechnol"] = "Journal of advances in biology & biotechnology",
  ["j adv biotechnol exp ther"] = "Journal Of Advanced Biotechnology And Experimental Therapeutics",
  ["j adv ceram"] = "Journal of Advanced Ceramics",
  ["j adv chem phys"] = "Journal of Advances in Chemical Physics",
  ["j adv concr technol"] = "Journal of Advanced Concrete Technology",
  ["j adv dairy res"] = "Journal of advances in dairy research",
  ["j adv dielectr"] = "Journal of Advanced Dielectrics",
  ["j adv joining processes"] = "Journal of Advanced Joining Processes",
  ["j adv manuf process"] = "Journal of advanced manufacturing and processing",
  ["j adv manuf syst"] = "Journal of Advanced Manufacturing Systems",
  ["j adv mater"] = "Journal of Advanced Materials",
  ["j adv math stud"] = "Journal of Advanced Mathematical Studies",
  ["j adv med educ prof"] = "Journal of advances in medical education & professionalism",
  ["j adv med life sci"] = "Journal of advancement in medical and life sciences",
  ["j adv med med res"] = "Journal of advances in medicine and medical research",
  ["j adv med pharm sci"] = "Journal of advances in medical and pharmaceutical sciences",
  ["j adv med surg nurs"] = "Journal of advanced medical-surgical nursing",
  ["j adv microbiol"] = "Journal of advances in microbiology",
  ["j adv microsc res"] = "Journal of Advanced Microscopy Research",
  ["j adv model earth syst"] = "Journal of advances in modeling earth systems",
  ["j adv nurs"] = "Journal of advanced nursing",
  ["j adv nutr hum metab"] = "Journal of advanced nutritional and human metabolism",
  ["j adv oral res"] = "Journal of advanced oral research",
  ["j adv oxid technol"] = "Journal of Advanced Oxidation Technologies",
  ["j adv pharm technol res"] = "Journal of advanced pharmaceutical technology & research",
  ["j adv pract oncol"] = "Journal of the advanced practitioner in oncology",
  ["j adv prosthodont"] = "The journal of advanced prosthodontics",
  ["j adv res"] = "Journal of advanced research",
  ["j adv sci res"] = "Journal of advanced scientific research",
  ["j adv smart convergence"] = "Journal of advanced smart convergence",
  ["j adv stud topol"] = "Journal of Advanced Studies in Topology",
  ["j adv transp"] = "Journal of advanced transportation",
  ["j adv vet anim res"] = "Journal of advanced veterinary and animal research",
  ["j adv vet res"] = "Journal of advanced veterinary research",
  ["j advert"] = "Journal of advertising",
  ["j advert res"] = "Journal of advertising research",
  ["j aeronaut sci"] = "Journal of the Aeronautical Sciences",
  ["j aerosol med"] = "Journal of aerosol medicine : the official journal of the International Society for Aerosols in Medicine",
  ["j aerosol med pulm drug deliv"] = "Journal of aerosol medicine and pulmonary drug delivery",
  ["j aerosol med pulm drug delivery"] = "Journal of Aerosol Medicine and Pulmonary Drug Delivery",
  ["j aerosol sci"] = "Journal of Aerosol Science",
  ["j aerosp eng"] = "Journal of aerospace engineering",
  ["j aerosp oper"] = "Journal of Aerospace Operations",
  ["j affect disord"] = "Journal of affective disorders",
  ["j affect disord rep"] = "Journal of affective disorders reports",
  ["j afr am stud (new brunsw)"] = "Journal of African American studies (New Brunswick, N.J.)",
  ["j afr assoc physiol sci"] = "Journal of African Association of Physiological Sciences",
  ["j afr earth sci"] = "Journal of African Earth Sciences",
  ["j afr econ"] = "Journal of African economies",
  ["j afr hist"] = "Journal of African history",
  ["j afr law"] = "Journal of African law",
  ["j afr stud"] = "Journal of African studies",
  ["j african bus"] = "Journal of African Business",
  ["j african economies"] = "Journal of African Economies",
  ["j african finance econ devel"] = "Journal of African Finance and Economic Development",
  ["j afro am hist geneal soc"] = "Journal of the Afro-American Historical and Genealogical Society. Afro-American Historical and Genealogical Society (Washington, D.C.)",
  ["j ageing res healthc"] = "Journal of ageing research and healthcare",
  ["j aggress confl peace res"] = "Journal of aggression, conflict and peace research",
  ["j aggress maltreat trauma"] = "Journal of aggression, maltreatment & trauma",
  ["j aging gerontol"] = "Journal of aging and gerontology",
  ["j aging health"] = "Journal of aging and health",
  ["j aging jud"] = "Journal Of Aging And Judaism",
  ["j aging phys act"] = "Journal of aging and physical activity",
  ["j aging res"] = "Journal of aging research",
  ["j aging res clin pract"] = "The journal of aging research & clinical practice",
  ["j aging sci"] = "Journal of aging science",
  ["j aging soc policy"] = "Journal of aging & social policy",
  ["j aging stud"] = "Journal of aging studies",
  ["j agr appl econ"] = "Journal of Agricultural and Applied Economics",
  ["j agr econ"] = "Journal of Agricultural Economics",
  ["j agr econ res"] = "Journal of Agricultural Economics Research",
  ["j agr resource econ"] = "Journal of Agricultural and Resource Economics",
  ["j agribus dev emerg econ"] = "Journal of agribusiness in developing and emerging economies",
  ["j agric appl econ"] = "Journal of Agricultural and Applied Economics",
  ["j agric appl econ assoc"] = "Journal of Agricultural and Applied Economics Association",
  ["j agric biol environ stat"] = "Journal of agricultural, biological, and environmental statistics",
  ["j agric biol sci"] = "Journal of agricultural and biological science",
  ["j agric chem environ"] = "Journal of agricultural chemistry and environment",
  ["j agric econ"] = "Journal of agricultural economics",
  ["j agric educ"] = "Journal of agricultural education",
  ["j agric educ ext"] = "Journal of Agricultural Education and Extension",
  ["j agric eng res"] = "Journal of Agricultural Engineering Research",
  ["j agric entomol"] = "Journal of Agricultural Entomology",
  ["j agric environ ethics"] = "Journal of agricultural & environmental ethics",
  ["j agric ethics"] = "Journal of agricultural ethics",
  ["j agric food chem"] = "Journal of agricultural and food chemistry",
  ["j agric food ind organ"] = "Journal of agricultural & food industrial organization",
  ["j agric food inf"] = "Journal of Agricultural & Food Information",
  ["j agric food res"] = "Journal of agriculture and food research",
  ["j agric food sys community dev"] = "Journal of Agriculture, Food Systems, and Community Development",
  ["j agric food syst community dev"] = "Journal of agriculture, food systems, and community development",
  ["j agric food technol"] = "Jounal of agriculture and food technology",
  ["j agric for"] = "Journal of agriculture and forestry (New York, N.Y.)",
  ["j agric meteorol"] = "Journal of Agricultural Meteorology",
  ["j agric res"] = "Journal of agricultural research",
  ["j agric saf health"] = "Journal of agricultural safety and health",
  ["j agric sci"] = "Journal of Agricultural Science",
  ["j agric sci (belihuloya)"] = "Journal of agricultural sciences",
  ["j agric sci (tor)"] = "Journal of agricultural science (Toronto, Ont.)",
  ["j agric sci technol"] = "Journal of agricultural science and technology : JAST",
  ["j agric sci technol a"] = "Journal of agricultural science and technology. A",
  ["j agric sci technol b"] = "Journal of agricultural science and technology. B",
  ["j agric tradit bot appl"] = "Journal d'agriculture traditionnelle et de botanique appliquée : JATBA",
  ["j agric trop bot appl"] = "Journal d'agriculture tropicale et de botanique appliquée",
  ["j agric urban entomol"] = "Journal of agricultural and urban entomology",
  ["j agroaliment proc technol"] = "Journal of Agroalimentary Processes and Technologies",
  ["j agrobiol"] = "Journal of agrobiology",
  ["j agrobiotechnol"] = "Journal of agrobiotechnology",
  ["j agromed"] = "Journal of Agromedicine",
  ["j agromedicine"] = "Journal of agromedicine",
  ["j agrometeorol"] = "Journal of Agrometeorology",
  ["j agron crop sci"] = "Journal of agronomy and crop science",
  ["j agron educ"] = "Journal of Agronomic Education",
  ["j ahima"] = "Journal of AHIMA",
  ["j aids clin res"] = "Journal of AIDS & clinical research",
  ["j aids hiv infect"] = "Journal of AIDS and HIV infections",
  ["j aids hiv res"] = "Journal of AIDS and HIV research (Online)",
  ["j aids hiv treat"] = "Journal of AIDS and HIV treatment",
  ["j aids immune res"] = "Journal of AIDS and immune research",
  ["j air med transp"] = "The Journal of air medical transport",
  ["j air pollut control assoc"] = "Journal of the Air Pollution Control Association",
  ["j air transp (rest)"] = "Journal of air transportation (Reston, Va.)",
  ["j air transp manag"] = "Journal of air transport management",
  ["j air transp manage"] = "Journal of Air Transport Management",
  ["j air transp res soc"] = "Journal of the Air Transport Research Society",
  ["j air waste manag assoc"] = "Journal of the Air & Waste Management Association (1995)",
  ["j air waste manage assoc"] = "Journal of the Air & Waste Management Association",
  ["j aircr"] = "Journal of aircraft",
  ["j ala acad sci"] = "The Journal of the Alabama Academy of Science",
  ["j ala dent assoc"] = "Journal - Alabama Dental Association",
  ["j albert einstein med cent (phila)"] = "The Journal of the Albert Einstein Medical Center, Philadelphia",
  ["j alcohol drug depend"] = "Journal of alcoholism and drug dependence",
  ["j alcohol drug educ"] = "Journal of alcohol and drug education",
  ["j algal biomass util"] = "Journal of algal biomass utilization",
  ["j algebr hyperstruct log algebras"] = "Journal of Algebraic Hyperstructures and Logical Algebras",
  ["j algebr stat"] = "Journal of algebraic statistics",
  ["j algebr syst"] = "Journal of Algebraic Systems",
  ["j algebra"] = "Journal of Algebra",
  ["j algebra appl"] = "Journal of algebra and its applications",
  ["j algebra comb discrete struct appl"] = "Journal of Algebra Combinatorics Discrete Structures and Applications",
  ["j algebra comput appl"] = "Journal of Algebra and Computational Applications",
  ["j algebra relat topics"] = "Journal of Algebra and Related Topics",
  ["j algebraic comb"] = "Journal of Algebraic Combinatorics",
  ["j algebraic combin"] = "Journal of Algebraic Combinatorics. An International Journal",
  ["j algebraic geom"] = "Journal of Algebraic Geometry",
  ["j algorithm comput technol"] = "Journal of algorithms & computational technology",
  ["j algorithms"] = "Journal of Algorithms",
  ["j algorithms comput technol"] = "Journal of Algorithms & Computational Technology",
  ["j all india dent assoc"] = "The Journal of the All India Dental Association",
  ["j all india ophthalmol soc"] = "Journal of the All-India Ophthalmological Society",
  ["j allergy"] = "The Journal of allergy",
  ["j allergy (cairo)"] = "Journal of allergy",
  ["j allergy clin immunol"] = "The Journal of allergy and clinical immunology",
  ["j allergy clin immunol pract"] = "The journal of allergy and clinical immunology. In practice",
  ["j allergy infect dis"] = "Journal of allergy and infectious diseases",
  ["j allergy ther"] = "Journal of allergy & therapy",
  ["j allied health"] = "Journal of allied health",
  ["j alloys"] = "Journal of Alloys and Compounds",
  ["j alloys compd"] = "Journal of Alloys and Compounds",
  ["j alloys compd commun"] = "Journal of Alloys and Compounds Communications",
  ["j altern complement med"] = "Journal of alternative and complementary medicine (New York, N.Y.)",
  ["j alzheimers dis"] = "Journal of Alzheimer's disease : JAD",
  ["j alzheimers dis parkinsonism"] = "Journal of Alzheimer's disease & Parkinsonism",
  ["j alzheimers dis rep"] = "Journal of Alzheimer's disease reports",
  ["j alzheimers parkinsonism dement"] = "Journal of Alzheimer's parkinsonism & dementia",
  ["j am acad appl nutr"] = "Journal of the American Academy of Applied Nutrition. American Academy of Applied Nutrition",
  ["j am acad audiol"] = "Journal of the American Academy of Audiology",
  ["j am acad child adolesc psychiatry"] = "Journal of the American Academy of Child and Adolescent Psychiatry",
  ["j am acad child psychiatry"] = "Journal of the American Academy of Child Psychiatry",
  ["j am acad dermatol"] = "Journal of the American Academy of Dermatology",
  ["j am acad gnathol orthop"] = "Journal (American Academy of Gnathologic Orthopedics)",
  ["j am acad gold foil oper"] = "The Journal of the American Academy of Gold Foil Operators",
  ["j am acad nurse pract"] = "Journal of the American Academy of Nurse Practitioners",
  ["j am acad orthop surg"] = "The Journal of the American Academy of Orthopaedic Surgeons",
  ["j am acad orthop surg glob res rev"] = "Journal of the American Academy of Orthopaedic Surgeons. Global research & reviews",
  ["j am acad physician assist"] = "Journal of the American Academy of Physician Assistants",
  ["j am acad psychiatry law"] = "The journal of the American Academy of Psychiatry and the Law",
  ["j am acad psychoanal"] = "The Journal of the American Academy of Psychoanalysis",
  ["j am acad psychoanal dyn psychiatry"] = "The journal of the American Academy of Psychoanalysis and Dynamic Psychiatry",
  ["j am acad relig"] = "Journal of the American Academy of Religion. American Academy of Religion",
  ["j am acad spec educ prof"] = "Journal of the American Academy of Special Education Professionals",
  ["j am aging assoc"] = "Journal of the American Aging Association",
  ["j am anim hosp assoc"] = "Journal of the American Animal Hospital Association",
  ["j am assoc gynecol laparosc"] = "The Journal of the American Association of Gynecologic Laparoscopists",
  ["j am assoc instr invest poult husb"] = "Journal of the American Association of Instructors and Investigators of Poultry Husbandry",
  ["j am assoc lab anim sci"] = "Journal of the American Association for Laboratory Animal Science : JAALAS",
  ["j am assoc med rec libr"] = "Journal. American Association of Medical Record Librarians",
  ["j am assoc med transcr"] = "Journal (American Association for Medical Transcription)",
  ["j am assoc nephrol nurses tech"] = "The Journal of the American Association of Nephrology Nurses & Technicians",
  ["j am assoc nurse anesth"] = "The Journal of the American Association of Nurse Anesthetists",
  ["j am assoc nurse pract"] = "Journal of the American Association of Nurse Practitioners",
  ["j am aud soc"] = "Journal of the American Auditory Society",
  ["j am audiol soc"] = "Journal of the American Audiology Society",
  ["j am board fam med"] = "Journal of the American Board of Family Medicine : JABFM",
  ["j am board fam pract"] = "The Journal of the American Board of Family Practice",
  ["j am ceram soc"] = "Journal of the American Ceramic Society",
  ["j am chem soc"] = "Journal of the American Chemical Society",
  ["j am col certif wound spec"] = "The journal of the American College of Certified Wound Specialists",
  ["j am coll cardiol"] = "Journal of the American College of Cardiology",
  ["j am coll clin pharm"] = "Journal of the American College of Clinical Pharmacy : JACCP",
  ["j am coll clin wound spec"] = "The journal of the American College of Clinical Wound Specialists",
  ["j am coll dent"] = "The Journal of the American College of Dentists",
  ["j am coll emerg physicians open"] = "Journal of the American College of Emergency Physicians open",
  ["j am coll health"] = "Journal of American college health : J of ACH",
  ["j am coll health assoc"] = "Journal of the American College Health Association",
  ["j am coll neuropsychiatr"] = "Journal of the American College of Neuropsychiatrists",
  ["j am coll nutr"] = "Journal of the American College of Nutrition",
  ["j am coll radiol"] = "Journal of the American College of Radiology : JACR",
  ["j am coll surg"] = "Journal of the American College of Surgeons",
  ["j am coll toxicol"] = "Journal of the American College of Toxicology",
  ["j am cult"] = "Journal of American culture",
  ["j am cult (malden)"] = "Journal of American culture (Malden, Mass. : 2003)",
  ["j am dent assoc"] = "Journal of the American Dental Association (1939)",
  ["j am dent assoc (ed ital)"] = "Journal Of The American Dental Association (Edizione Italiana)",
  ["j am dent hyg assoc"] = "The Journal of the American Dental Hygienists' Association",
  ["j am dent soc anesthesiol"] = "Journal of the American Dental Society of Anesthesiology",
  ["j am diet assoc"] = "Journal of the American Dietetic Association",
  ["j am ethn hist"] = "Journal of American ethnic history",
  ["j am folk"] = "The Journal of American folk-lore",
  ["j am geriatr soc"] = "Journal of the American Geriatrics Society",
  ["j am health care"] = "The journal of American health care",
  ["j am health care assoc"] = "Journal - American Health Care Association",
  ["j am health policy"] = "The Journal of American health policy",
  ["j am heart assoc"] = "Journal of the American Heart Association",
  ["j am helicopter soc"] = "Journal of the American Helicopter Society. American Helicopter Society",
  ["j am hist"] = "Journal of American history (Bloomington, Ind.)",
  ["j am hist soc ger russ"] = "Journal of the American Historical Society of Germans from Russia. American Historical Society of Germans from Russia",
  ["j am indian educ"] = "Journal of American Indian education",
  ["j am inst conserv"] = "Journal of the American Institute for Conservation",
  ["j am inst electr eng"] = "Journal of the American Institute of Electrical Engineers",
  ["j am inst homeopath"] = "Journal of the American Institute of Homeopathy",
  ["j am inst plann"] = "Journal of the American Institute of Planners",
  ["j am insur"] = "Journal of American insurance",
  ["j am intraocul implant soc"] = "Journal - American Intra-Ocular Implant Society",
  ["j am killifish assoc"] = "The Journal of the American Killifish Association",
  ["j am leather chem assoc"] = "Journal of the American Leather Chemists Association",
  ["j am math soc"] = "Journal of the American Mathematical Society",
  ["j am med assoc"] = "Journal of the American Medical Association",
  ["j am med dir assoc"] = "Journal of the American Medical Directors Association",
  ["j am med inf assoc"] = "Journal of the American Medical Informatics Association",
  ["j am med inform assoc"] = "Journal of the American Medical Informatics Association : JAMIA",
  ["j am med rec assoc"] = "Journal of the American Medical Record Association",
  ["j am med technol"] = "Journal of the American Medical Technologists",
  ["j am med wom assoc"] = "Journal of the American Medical Womens Association",
  ["j am med womens assoc"] = "Journal of the American Medical Women's Association",
  ["j am med womens assoc (1972)"] = "Journal of the American Medical Women's Association (1972)",
  ["j am mosq control assoc"] = "Journal of the American Mosquito Control Association",
  ["j am mosq control assoc suppl"] = "Journal of the American Mosquito Control Association. Supplement",
  ["j am nutr assoc"] = "Journal of the American Nutrition Association",
  ["j am oil chem soc"] = "Journal of the American Oil Chemists’ Society",
  ["j am optom assoc"] = "Journal of the American Optometric Association",
  ["j am orient soc"] = "Journal of the American Oriental Society. American Oriental Society",
  ["j am osteopath assoc"] = "The Journal of the American Osteopathic Association",
  ["j am osteopath coll radiol"] = "Journal of the American Osteopathic College of Radiology",
  ["j am paraplegia soc"] = "The Journal of the American Paraplegia Society",
  ["j am pharm assoc"] = "Journal of the American Pharmaceutical Association",
  ["j am pharm assoc (2003)"] = "Journal of the American Pharmacists Association : JAPhA",
  ["j am pharm assoc (wash)"] = "Journal of the American Pharmaceutical Association (Washington, D.C. : 1996)",
  ["j am pharm assoc am pharm assoc (baltim)"] = "Journal of the American Pharmaceutical Association. American Pharmaceutical Association",
  ["j am phys ther assoc"] = "Journal of the American Physical Therapy Association",
  ["j am plann assoc"] = "Journal of the American Planning Association. American Planning Association",
  ["j am podiatr med assoc"] = "Journal of the American Podiatric Medical Association",
  ["j am podiatry assoc"] = "Journal of the American Podiatry Association",
  ["j am pomol soc"] = "Journal of the American Pomological Society",
  ["j am psychiatr nurses assoc"] = "Journal of the American Psychiatric Nurses Association",
  ["j am psychoanal assoc"] = "Journal of the American Psychoanalytic Association",
  ["j am public health assoc"] = "Journal. American Public Health Association",
  ["j am sci"] = "The journal of American science",
  ["j am sci affil"] = "Journal of the American Scientific Affiliation. American Scientific Affiliation",
  ["j am soc agron"] = "Journal - American Society of Agronomy",
  ["j am soc brew chem"] = "Journal of the American Society of Brewing Chemists",
  ["j am soc cytopathol"] = "Journal of the American Society of Cytopathology",
  ["j am soc echocardiogr"] = "Journal of the American Society of Echocardiography : official publication of the American Society of Echocardiography",
  ["j am soc geriatr dent"] = "The Journal of the American Society for Geriatric Dentistry",
  ["j am soc hortic sci"] = "Journal of the American Society for Horticultural Science. American Society for Horticultural Science",
  ["j am soc hypertens"] = "Journal of the American Society of Hypertension : JASH",
  ["j am soc inf sci"] = "Journal of the American Society for Information Science. American Society for Information Science",
  ["j am soc inf sci technol"] = "Journal of the American Society for Information Science and Technology : JASIST",
  ["j am soc mass spectrom"] = "Journal of the American Society for Mass Spectrometry",
  ["j am soc nav eng"] = "Journal of the American Society of Naval Engineers",
  ["j am soc naval eng"] = "Journal of the American Society of Naval Engineers",
  ["j am soc nephrol"] = "Journal of the American Society of Nephrology : JASN",
  ["j am soc prev dent"] = "The Journal of the American Society for Preventive Dentistry",
  ["j am soc psych res"] = "The Journal of the American Society for Psychical Research",
  ["j am soc psychosom dent med"] = "The Journal of the American Society of Psychosomatic Dentistry and Medicine",
  ["j am soc study orthod"] = "Journal of the American Society for the Study of Orthodontics. American Society for the Study of Orthodontics",
  ["j am soc sugar cane technol"] = "Journal. American Society of Sugar Cane Technologists. Florida Division",
  ["j am stat assoc"] = "Journal of the American Statistical Association",
  ["j am stud"] = "Journal of American studies",
  ["j am vener dis assoc"] = "Journal of the American Venereal Disease Association",
  ["j am vet med assoc"] = "Journal of the American Veterinary Medical Association",
  ["j am water resour assoc"] = "Journal of the American Water Resources Association",
  ["j am water works assn"] = "Journal American Water Works Association",
  ["j am water works assoc"] = "Journal - American Water Works Association",
  ["j amat sport"] = "Journal of amateur sport",
  ["j ambient intell hum comput"] = "Journal of Ambient Intelligence and Humanized Computing",
  ["j ambient intell humaniz comput"] = "Journal of ambient intelligence and humanized computing",
  ["j ambient intell smart environ"] = "Journal of Ambient Intelligence and Smart Environments",
  ["j ambul care manage"] = "The Journal of ambulatory care management",
  ["j ambul care mark"] = "Journal of ambulatory care marketing",
  ["j ambulatory care manage"] = "Journal of Ambulatory Care Management",
  ["j amer math soc"] = "Journal of the American Mathematical Society",
  ["j amer real estate urban econ assoc"] = "Journal of the American Real Estate and Urban Economics Association",
  ["j amer statist assoc"] = "Journal of the American Statistical Association",
  ["j amer statistical assoc"] = "Journal of the American Statistical Association",
  ["j amino acids"] = "Journal of amino acids",
  ["j anaesthesiol clin pharmacol"] = "Journal of anaesthesiology, clinical pharmacology",
  ["j anal"] = "Journal of Analysis",
  ["j anal appl"] = "Journal of Analysis and Applications",
  ["j anal appl pyrolysis"] = "Journal of Analytical and Applied Pyrolysis",
  ["j anal at spectrom"] = "Journal of Analytical Atomic Spectrometry",
  ["j anal bioanal tech"] = "Journal of analytical & bioanalytical techniques",
  ["j anal chem"] = "Journal of Analytical Chemistry",
  ["j anal math"] = "Journal d’Analyse Mathematique",
  ["j anal methods chem"] = "Journal of analytical methods in chemistry",
  ["j anal mol tech"] = "Journal of analytical & molecular techniques",
  ["j anal oncol"] = "Journal of analytical oncology",
  ["j anal pharm res"] = "Journal of analytical & pharmaceutical research",
  ["j anal psychol"] = "The Journal of analytical psychology",
  ["j anal sci technol"] = "Journal of Analytical Science and Technology",
  ["j anal test"] = "Journal of analysis and testing",
  ["j anal toxicol"] = "Journal of Analytical Toxicology",
  ["j anat"] = "Journal of anatomy",
  ["j anat physiol"] = "Journal of anatomy and physiology",
  ["j anat soc india"] = "Journal of the Anatomical Society of India",
  ["j androl"] = "Journal of andrology",
  ["j anesth"] = "Journal of anesthesia",
  ["j anesth clin res"] = "Journal of anesthesia & clinical research",
  ["j anesth crit care"] = "Journal of anesthesia & critical care : open access",
  ["j anesth hist"] = "Journal of anesthesia history",
  ["j anesth perioper med"] = "Journal of anesthesia and perioperative medicine",
  ["j anesth res pain med"] = "Journal of anesthetic research and pain medicine",
  ["j angiogenes res"] = "Journal of angiogenesis research",
  ["j anglocont dent soc"] = "Journal (Anglo-Continental Dental Society)",
  ["j anhui univ nat sci"] = "Journal of Anhui University. Natural Science Edition. Anhui Daxue Xuebao. Ziran Kexue Ban",
  ["j anim breed genet"] = "Journal of animal breeding and genetics = Zeitschrift für Tierzüchtung und Züchtungsbiologie",
  ["j anim divers"] = "Journal of animal diversity",
  ["j anim ecol"] = "The Journal of animal ecology",
  ["j anim feed sci"] = "Journal of animal and feed sciences",
  ["j anim physiol anim nutr"] = "Journal of Animal Physiology and Animal Nutrition",
  ["j anim physiol anim nutr (berl)"] = "Journal of animal physiology and animal nutrition",
  ["j anim physiol nutr"] = "Journal of Animal Physiology and Nutrition",
  ["j anim plant sci"] = "Journal of animal and plant sciences",
  ["j anim prod"] = "Journal of animal production",
  ["j anim sci"] = "Journal of animal science",
  ["j anim sci (champaign, il, u s)"] = "Journal of Animal Science (Champaign, IL, United States)",
  ["j anim sci (savoy, il, u s)"] = "Journal of Animal Science (Savoy, IL, United States)",
  ["j anim sci biotechnol"] = "Journal of animal science and biotechnology",
  ["j anim sci technol"] = "Journal of animal science and technology",
  ["j anim vet adv"] = "Journal of animal and veterinary advances : JAVA",
  ["j anthropol archaeol"] = "Journal of anthropological archaeology",
  ["j anthropol n am"] = "Journal for the anthropology of North America",
  ["j anthropol res"] = "Journal of anthropological research",
  ["j anthropol sci"] = "Journal of anthropological sciences = Rivista di antropologia : JASS",
  ["j anthropol surv india"] = "The Journal of the Anthropological Survey of India",
  ["j anti aging med"] = "Journal of anti-aging medicine",
  ["j antibiot"] = "Journal of Antibiotics",
  ["j antibiot (tokyo)"] = "The Journal of antibiotics",
  ["j antibiot [b]"] = "Journal of Antibiotics. Ser. B.",
  ["j antibiot b"] = "The Journal of antibiotics. Ser. B",
  ["j antimicrob agents"] = "Journal of antimicrobial agents",
  ["j antimicrob chemother"] = "Journal of Antimicrobial Chemotherapy",
  ["j antivir antiretrovir"] = "Journal of antivirals & antiretrovirals",
  ["j anus rectum colon"] = "Journal of the anus, rectum and colon",
  ["j anxiety disord"] = "Journal of anxiety disorders",
  ["j aoac int"] = "Journal of AOAC International",
  ["j apdsa (tokyo)"] = "Journal Of Apdsa; Asian Pacific Dental Student Association",
  ["j apic res"] = "Journal of Apicultural Research",
  ["j apic sci"] = "Journal of apicultural science",
  ["j appalach health"] = "Journal of Appalachian health",
  ["j appalach stud"] = "Journal of Appalachian studies",
  ["j appl anal"] = "Journal of Applied Analysis",
  ["j appl anal comput"] = "Journal of Applied Analysis and Computation",
  ["j appl anim nutr"] = "Journal of Applied Animal Nutrition",
  ["j appl anim res"] = "Journal of applied animal research",
  ["j appl anim welf sci"] = "Journal of applied animal welfare science : JAAWS",
  ["j appl anim welfare res"] = "Journal of Applied Animal Welfare Research",
  ["j appl anim welfare sci"] = "Journal of Applied Animal Welfare Science",
  ["j appl aquacult"] = "Journal of Applied Aquaculture",
  ["j appl arts health"] = "Journal of applied arts & health",
  ["j appl bacteriol"] = "Journal of Applied Bacteriology",
  ["j appl behav anal"] = "Journal of applied behavior analysis",
  ["j appl behav sci"] = "The Journal of applied behavioral science",
  ["j appl bioanal"] = "Journal of applied bioanalysis",
  ["j appl biobehav res"] = "Journal of applied biobehavioral research",
  ["j appl biochem"] = "Journal of applied biochemistry",
  ["j appl bioinforma comput biol"] = "Journal of applied bioinformatics & computational biology",
  ["j appl biol biotechnol"] = "Journal of applied biology & biotechnology",
  ["j appl biol chem"] = "Journal of Applied Biological Chemistry",
  ["j appl biol sci"] = "Journal of Applied Biological Sciences",
  ["j appl biomater"] = "Journal of Applied Biomaterials",
  ["j appl biomater biomech"] = "Journal of applied biomaterials & biomechanics : JABB",
  ["j appl biomater funct mater"] = "Journal of applied biomaterials & functional materials",
  ["j appl biomech"] = "Journal of applied biomechanics",
  ["j appl biomed"] = "Journal of applied biomedicine",
  ["j appl biosci"] = "Journal of applied biosciences",
  ["j appl biotechnol rep"] = "Journal Of Applied Biotechnology Reports",
  ["j appl bot food qual"] = "Journal of applied botany and food quality = Angewandte Botanik",
  ["j appl bus res"] = "Journal of Applied Business Research",
  ["j appl cardiol"] = "Journal of applied cardiology",
  ["j appl chem"] = "Journal of Applied Chemistry",
  ["j appl chem biotech"] = "Journal of Applied Chemistry & Biotechnology",
  ["j appl clin med phys"] = "Journal of applied clinical medical physics",
  ["j appl commun res"] = "Journal of applied communication research : JACR",
  ["j appl comput sci methods"] = "Journal of Applied Computer Science Methods",
  ["j appl comput topol"] = "Journal of applied and computational topology",
  ["j appl crystallogr"] = "Journal of Applied Crystallography",
  ["j appl dev psychol"] = "Journal of applied developmental psychology",
  ["j appl ecol"] = "The Journal of applied ecology",
  ["j appl econ"] = "Journal of applied economics",
  ["j appl econ (chichester engl)"] = "Journal of applied econometrics (Chichester, England)",
  ["j appl econometrics"] = "Journal of Applied Econometrics",
  ["j appl electrochem"] = "Journal of Applied Electrochemistry",
  ["j appl entomol"] = "Journal of Applied Entomology",
  ["j appl environ microbiol"] = "Journal of applied & environmental microbiology",
  ["j appl funct anal"] = "Journal of Applied Functional Analysis",
  ["j appl genet"] = "Journal of applied genetics",
  ["j appl geochem"] = "Journal of Applied Geochemistry",
  ["j appl geod"] = "Journal of Applied Geodesy",
  ["j appl geophy"] = "Journal of applied geophysics",
  ["j appl geophys"] = "Journal of Applied Geophysics",
  ["j appl gerontol"] = "Journal of applied gerontology : the official journal of the Southern Gerontological Society",
  ["j appl glycosci"] = "Journal of Applied Glycoscience",
  ["j appl glycosci (1999)"] = "Journal of applied glycoscience",
  ["j appl hist"] = "Journal of applied history",
  ["j appl ichthyol"] = "Zeitschrift für angewandte Ichthyologie = Journal of applied ichthyology",
  ["j appl ind math"] = "Journal of Applied and Industrial Mathematics",
  ["j appl juv justice serv"] = "Journal of applied juvenile justice services",
  ["j appl lab med"] = "The journal of applied laboratory medicine",
  ["j appl learn technol"] = "Journal of applied learning technology",
  ["j appl log"] = "Journal of Applied Logic",
  ["j appl logic"] = "Journal of Applied Logic",
  ["j appl logics"] = "Journal of Applied Logics. The IfCoLog Journal of Logics and their Applications",
  ["j appl manag entrep"] = "The journal of applied management and entrepreneurship",
  ["j appl math"] = "Journal of Applied Mathematics",
  ["j appl math anal appl"] = "Journal of Applied Mathematical Analysis and Applications",
  ["j appl math comput"] = "Journal of Applied Mathematics and Computing",
  ["j appl math comput mech"] = "Journal of Applied Mathematics and Computational Mechanics",
  ["j appl math decis sci"] = "Journal of Applied Mathematics & Decision Sciences",
  ["j appl math inform"] = "Journal of Applied Mathematics & Informatics",
  ["j appl math mech"] = "Journal of Applied Mathematics and Mechanics",
  ["j appl math stat inform"] = "Journal of Applied Mathematics, Statistics and Informatics",
  ["j appl meas"] = "Journal of applied measurement",
  ["j appl mech"] = "Journal of Applied Mechanics",
  ["j appl mech eng"] = "Journal of applied mechanical engineering",
  ["j appl mech tech phys"] = "Journal of Applied Mechanics and Technical Physics",
  ["j appl mech trans asme"] = "ASME Journal of Applied Mechanics",
  ["j appl med sci"] = "Journal of applied medical sciences",
  ["j appl metalwork"] = "Journal of Applied Metalworking",
  ["j appl meteorol"] = "Journal of Applied Meteorology",
  ["j appl meteorol climatol"] = "Journal of applied meteorology and climatology",
  ["j appl microbiol"] = "Journal of applied microbiology",
  ["j appl microbiol biochem"] = "Journal of applied microbiology and biochemistry",
  ["j appl non-class log"] = "Journal of Applied Non-Classical Logics",
  ["j appl non-classical logics"] = "Journal of Applied Non-Classical Logics",
  ["j appl nonlinear dyn"] = "Journal of Applied Nonlinear Dynamics",
  ["j appl numer optim"] = "Journal of Applied and Numerical Optimization",
  ["j appl nutr"] = "Journal of Applied Nutrition",
  ["j appl oral sci"] = "Journal of applied oral science : revista FOB",
  ["j appl pharm sci"] = "Journal of applied pharmaceutical science",
  ["j appl philos"] = "Journal of applied philosophy",
  ["j appl photogr eng"] = "Journal of Applied Photographic Engineering",
  ["j appl phycol"] = "Journal of Applied Phycology",
  ["j appl phys"] = "Journal of Applied Physics",
  ["j appl phys (melville, ny, u s)"] = "Journal of Applied Physics (Melville, NY, United States)",
  ["j appl physiol"] = "Journal of Applied Physiology",
  ["j appl physiol (1985)"] = "Journal of applied physiology (Bethesda, Md. : 1985)",
  ["j appl physiol respir environ exerc physiol"] = "Journal of applied physiology: respiratory, environmental and exercise physiology",
  ["j appl polym sci"] = "Journal of Applied Polymer Science",
  ["j appl polym sci symp"] = "Journal of applied polymer science. Applied polymer symposium",
  ["j appl polym sci: appl polym symp"] = "Journal of Applied Polymer Science: Applied Polymer Symposium",
  ["j appl poult res"] = "Journal of Applied Poultry Research",
  ["j appl probab"] = "Journal of Applied Probability",
  ["j appl probab stat"] = "Journal of applied probability and statistics",
  ["j appl psychol"] = "The Journal of applied psychology",
  ["j appl rehabil couns"] = "Journal of applied rehabilitation counseling",
  ["j appl remote sens"] = "Journal of Applied Remote Sensing",
  ["j appl res"] = "The journal of applied research",
  ["j appl res child"] = "The journal of applied research on children : informing policy for children at risk",
  ["j appl res community coll"] = "Journal of Applied Research in the Community College",
  ["j appl res intellect disabil"] = "Journal of applied research in intellectual disabilities : JARID",
  ["j appl res med aromat plants"] = "Journal of Applied Research on Medicinal and Aromatic Plants",
  ["j appl res mem cogn"] = "Journal of applied research in memory and cognition",
  ["j appl res technol"] = "Journal of Applied Research and Technology",
  ["j appl sch psychol"] = "Journal of applied school psychology",
  ["j appl sci (faisalabad)"] = "Journal of applied sciences (Faisalabad, Pakistan)",
  ["j appl sci (thailand)"] = "The Journal of applied science",
  ["j appl sci res"] = "Journal of applied sciences research",
  ["j appl sci south afr"] = "Journal of applied science in Southern Africa : JASSA : the journal of the University of Zimbabwe",
  ["j appl soc psychol"] = "Journal of applied social psychology",
  ["j appl soc sci"] = "The Journal of applied social sciences",
  ["j appl soc sci (boulder)"] = "Journal of applied social science",
  ["j appl soc sci stud"] = "Journal of Applied Social Science Studies",
  ["j appl solution chem model"] = "Journal of Applied Solution Chemistry and Modeling",
  ["j appl spectrosc"] = "Journal of Applied Spectroscopy",
  ["j appl spectrosc (ussr)"] = "Journal of Applied Spectroscopy (USSR) [translation of Zhurnal Prikladnoi Spektroskopii]",
  ["j appl sport psychol"] = "Journal of applied sport psychology",
  ["j appl sports sci"] = "Journal of Applied Sports Sciences",
  ["j appl stat"] = "Journal of applied statistics",
  ["j appl statist"] = "Journal of Applied Statistics",
  ["j appl statist sci"] = "Journal of applied statistical science",
  ["j appl syst anal"] = "Journal of Applied Systems Analysis",
  ["j appl toxicol"] = "Journal of Applied Toxicology",
  ["j appl volcanol"] = "Journal of Applied Volcanology",
  ["j appl water eng res"] = "Journal of Applied Water Engineering and Research",
  ["j approx theory"] = "Journal of Approximation Theory",
  ["j approximation theory"] = "Journal of Approximation Theory",
  ["j aquac res dev"] = "Journal of aquaculture research & development",
  ["j aquat anim health"] = "Journal of aquatic animal health",
  ["j aquat ecosyst stress recovery"] = "Journal of Aquatic Ecosystem Stress and Recovery",
  ["j aquat food prod technol"] = "Journal of Aquatic Food Product Technology",
  ["j aquat plant manag"] = "Journal of aquatic plant management",
  ["j aquat res mar sci"] = "Journal of aquatic research and marine sciences",
  ["j arab aff"] = "Journal of Arab affairs",
  ["j arachnol"] = "The Journal of arachnology",
  ["j archaeol method theory"] = "Journal of archaeological method and theory",
  ["j archaeol res"] = "Journal of Archaeological Research",
  ["j archaeol sci"] = "Journal of archaeological science",
  ["j archaeol sci rep"] = "Journal of archaeological science, reports",
  ["j archaeol sci: rep"] = "Journal of Archaeological Science: Reports",
  ["j archit conserv"] = "Journal of Architectural Conservation",
  ["j archit eng"] = "Journal of Architectural Engineering",
  ["j archit plann res"] = "Journal of architectural and planning research",
  ["j argent chem soc"] = "Journal of the Argentine Chemical Society",
  ["j arid environ"] = "Journal of arid environments",
  ["j arid land"] = "Journal of arid land",
  ["j ariz hist"] = "The Journal of Arizona history",
  ["j ariz-nev acad sci"] = "Journal of the Arizona-Nevada Academy of Science",
  ["j ark acad sci"] = "Journal of the Arkansas Academy of Science",
  ["j ark med soc"] = "The Journal of the Arkansas Medical Society",
  ["j arnold arbor"] = "Journal of the Arnold Arboretum. Arnold Arboretum",
  ["j arrhythm"] = "Journal of arrhythmia",
  ["j arthritis"] = "Journal of arthritis",
  ["j arthroplasty"] = "The Journal of arthroplasty",
  ["j arthropod borne dis"] = "Journal of arthropod-borne diseases",
  ["j artif intell"] = "Journal on Artificial Intelligence",
  ["j artif intell conscious"] = "Journal of Artificial Intelligence and Consciousness",
  ["j artif intell res"] = "The journal of artificial intelligence research",
  ["j artif organs"] = "Journal of artificial organs : the official journal of the Japanese Society for Artificial Organs",
  ["j artif soc soc simul"] = "Journal of artificial societies and social simulation : JASSS",
  ["j artificial intelligence res"] = "Journal of Artificial Intelligence Research",
  ["j asahikawa nat college tech"] = "Journal of the Asahikawa National College of Technology",
  ["j asean fed endocr soc"] = "Journal of the ASEAN Federation of Endocrine Societies",
  ["j asia pac biodivers"] = "Journal of Asia-Pacific biodiversity",
  ["j asia pac econ"] = "Journal of the Asia Pacific economy",
  ["j asia pac entomol"] = "Journal of Asia-Pacific entomology",
  ["j asia pacific economy"] = "Journal of the Asia Pacific Economy",
  ["j asia-pac biodivers"] = "Journal of Asia-Pacific Biodiversity",
  ["j asia-pac entomol"] = "Journal of Asia-Pacific Entomology",
  ["j asia-pacific bus"] = "Journal of Asia-Pacific Business",
  ["j asian afr stud"] = "Journal of Asian and African studies",
  ["j asian archit build eng"] = "Journal of Asian Architecture and Building Engineering",
  ["j asian ceram soc"] = "Journal of Asian Ceramic Societies",
  ["j asian earth sci"] = "Journal of Asian Earth Sciences",
  ["j asian earth sci: x"] = "Journal of Asian Earth Sciences: X",
  ["j asian econ"] = "Journal of Asian economics",
  ["j asian fed obstet gynaecol"] = "The Journal of the Asian Federation of Obstetrics and Gynaecology",
  ["j asian nat prod res"] = "Journal of Asian natural products research",
  ["j asian public policy"] = "Journal of Asian public policy",
  ["j asian stud"] = "The Journal of Asian studies",
  ["j asiat"] = "Journal asiatique",
  ["j asiat soc"] = "Journal of the Asiatic Society",
  ["j asiat soc bangladesh humanit"] = "Journal of the Asiatic Society of Bangladesh. Humanities",
  ["j assam acad math"] = "Journal of the Assam Academy of Mathematics (JAAM)",
  ["j assam sci soc"] = "Journal of the Assam Science Society",
  ["j assam science society"] = "Journal. Assam Science Society",
  ["j assist reprod genet"] = "Journal of assisted reproduction and genetics",
  ["j assist technol"] = "Journal of assistive technologies",
  ["j assistive rehabil ther technol"] = "Journal of Assistive, Rehabilitative and Therapeutic Technologies",
  ["j assoc acad minor phys"] = "Journal of the Association for Academic Minority Physicians : the official publication of the Association for Academic Minority Physicians",
  ["j assoc adv med instrum"] = "JAAMI : journal of the Association for the Advancement of Medical Instrumentation",
  ["j assoc am med coll"] = "Journal. Association of American Medical Colleges",
  ["j assoc asphalt paving technol"] = "Journal of the Association of Asphalt Paving Technologists",
  ["j assoc care child health"] = "Journal of the Association for the Care of Children’s Health",
  ["j assoc care child hosp"] = "Journal of the Association for the Care of Children in Hospitals",
  ["j assoc consum res"] = "Journal of the Association for Consumer Research",
  ["j assoc crime scene reconstr"] = "Journal of the Association for Crime Scene Reconstruction",
  ["j assoc environ resour econ"] = "Journal of the Association of Environmental and Resource Economists",
  ["j assoc food drug off"] = "Journal of the Association of Food and Drug Officials",
  ["j assoc genet technol"] = "Journal of the Association of Genetic Technologists",
  ["j assoc healthc philanthr"] = "Journal (Association for Healthcare Philanthropy (U.S.))",
  ["j assoc hosp med educ"] = "Journal - Association for Hospital Medical Education",
  ["j assoc inf sci technol"] = "Journal of the Association for Information Science and Technology",
  ["j assoc inf syst"] = "Journal of the Association for Information Systems",
  ["j assoc math res"] = "Journal of the Association for Mathematical Research",
  ["j assoc med illus"] = "Journal of the Association of Medical Illustrators",
  ["j assoc med microbiol infect dis can"] = "Journal of the Association of Medical Microbiology and Infectious Disease Canada = Journal officiel de l’Association pour la microbiologie médicale et l’infectiologie Canada",
  ["j assoc med women india"] = "The Journal of the Association of Medical Women in India",
  ["j assoc nurses aids care"] = "The Journal of the Association of Nurses in AIDS Care : JANAC",
  ["j assoc off agric chem"] = "Journal of the Association of Official Agricultural Chemists",
  ["j assoc off anal chem"] = "Journal - Association of Official Analytical Chemists",
  ["j assoc pediatr oncol nurses"] = "Journal of the Association of Pediatric Oncology Nurses",
  ["j assoc pers sev handicaps"] = "The Journal of the Association for Persons with Severe Handicaps : official publication of the Association for Persons with Severe Handicaps",
  ["j assoc phys ment rehabil"] = "Journal of the Association for Physical and Mental Rehabilitation",
  ["j assoc physicians india"] = "The Journal of the Association of Physicians of India",
  ["j assoc res otolaryngol"] = "Journal of the Association for Research in Otolaryngology : JARO",
  ["j assoc univ technol manag"] = "Journal of the Association of University Technology Managers",
  ["j asthet chir"] = "Journal für ästhetische Chirurgie",
  ["j asthma"] = "The Journal of asthma : official journal of the Association for the Care of Asthma",
  ["j asthma allergy"] = "Journal of asthma and allergy",
  ["j asthma allergy educ"] = "Journal of asthma & allergy educators",
  ["j asthma res"] = "The Journal of asthma research",
  ["j astm int"] = "Journal of ASTM International",
  ["j astron hist heritage"] = "Journal of Astronomical History and Heritage",
  ["j astron instrum"] = "Journal of Astronomical Instrumentation",
  ["j astron telesc instrum syst"] = "Journal of Astronomical Telescopes, Instruments, and Systems",
  ["j astronaut sci"] = "Journal of Astronautical Sciences",
  ["j astrophys astron"] = "Journal of Astrophysics and Astronomy",
  ["j at mol opt phys"] = "Journal of atomic, molecular, and optical physics",
  ["j at mol phys"] = "Journal of atomic and molecular physics",
  ["j atheroscler res"] = "Journal of atherosclerosis research",
  ["j atheroscler thromb"] = "Journal of atherosclerosis and thrombosis",
  ["j athl enhanc"] = "Journal of athletic enhancement",
  ["j athl train"] = "Journal of athletic training",
  ["j atmos chem"] = "Journal of Atmospheric Chemistry",
  ["j atmos ocean sci"] = "Journal of Atmospheric & Ocean Science",
  ["j atmos ocean technol"] = "Journal of atmospheric and oceanic technology",
  ["j atmos oceanic technol"] = "Journal of Atmospheric and Oceanic Technology",
  ["j atmos sci"] = "Journal of Atmospheric Science",
  ["j atmos sol terr phys"] = "Journal of Atmospheric and Solar-Terrestrial Physics",
  ["j atmos terr phys"] = "Journal of Atmospheric and Terrestrial Physics",
  ["j atmospheric sci"] = "Journal of the Atmospheric Sciences",
  ["j atr fibrillation"] = "Journal of atrial fibrillation",
  ["j atten disord"] = "Journal of attention disorders",
  ["j aud res"] = "The Journal of auditory research",
  ["j aud res suppl"] = "Journal of auditory research. Supplement",
  ["j audio eng soc"] = "Journal of the Audio Engineering Society. Audio Engineering Society",
  ["j audiol otol"] = "Journal of audiology & otology",
  ["j audiov media med"] = "The Journal of audiovisual media in medicine",
  ["j aust ceram soc"] = "Journal of the Australian Ceramic Society",
  ["j aust coll midwives"] = "Journal, Australian College of Midwives",
  ["j aust inst surg dent tech"] = "Journal Of The Australian Institute Of Surgical And Dental Technicians (Melbourne)",
  ["j aust math soc"] = "Journal of the Australian Mathematical Society",
  ["j aust popul assoc"] = "Journal of the Australian Population Association",
  ["j austral math soc ser a"] = "Australian Mathematical Society",
  ["j austral math soc ser b"] = "Australian Mathematical Society",
  ["j australas ceram soc"] = "Journal of the Australasian Ceramic Society",
  ["j australas coll road saf"] = "Journal of the Australasian College of Road Safety",
  ["j australian polit economy"] = "Journal of Australian Political Economy",
  ["j autism child schizophr"] = "Journal of autism and childhood schizophrenia",
  ["j autism dev disord"] = "Journal of autism and developmental disorders",
  ["j autoimmun"] = "Journal of autoimmunity",
  ["j autoimmune dis"] = "Journal of autoimmune diseases",
  ["j autoimmune dis rheumatol"] = "Journal of autoimmune diseases and rheumatology",
  ["j autoimmune disord"] = "Journal of autoimmune disorders",
  ["j autom chem"] = "Journal of Automatic Chemistry",
  ["j autom inf sci"] = "Journal of Automation and Information Sciences",
  ["j autom intell"] = "Journal of Automation and Intelligence",
  ["j autom lang comb"] = "Journal of Automata, Languages and Combinatorics",
  ["j autom methods manag chem"] = "Journal of automated methods & management in chemistry",
  ["j autom methods manage chem"] = "Journal of Automated Methods & Management in Chemistry",
  ["j autom reason"] = "Journal of automated reasoning",
  ["j autom reasoning"] = "Journal of Automated Reasoning",
  ["j automat chem"] = "The Journal of automatic chemistry",
  ["j automat inform sci"] = "Journal of Automation and Information Sciences",
  ["j automat reason"] = "Journal of Automated Reasoning",
  ["j auton intell"] = "Journal of Autonomous Intelligence",
  ["j auton nerv syst"] = "Journal of the autonomic nervous system",
  ["j auton pharmacol"] = "Journal of Autonomic Pharmacology",
  ["j auton veh syst"] = "Journal of Autonomous Vehicles and Systems",
  ["j aux odontol"] = "Journal Das Auxiliares Odontologicas",
  ["j avian biol"] = "Journal of avian biology",
  ["j avian med surg"] = "Journal of avian medicine and surgery",
  ["j aviat med"] = "The Journal of aviation medicine",
  ["j awwa"] = "Journal AWWA",
  ["j ayub med coll abbottabad"] = "Journal of Ayub Medical College, Abbottabad : JAMC",
  ["j ayurveda integr med"] = "Journal of Ayurveda and integrative medicine",
  ["j back musculoskelet rehabil"] = "Journal of back and musculoskeletal rehabilitation",
  ["j back musculoskeletal rehabil"] = "Journal of back and musculoskeletal rehabilitation",
  ["j bacteriol"] = "Journal of bacteriology",
  ["j bacteriol mycol (monroe township)"] = "Journal of bacteriology and mycology (Monroe (Middlesex County, N.J. : Township))",
  ["j bacteriol parasitol"] = "Journal of bacteriology & parasitology",
  ["j bacteriol virol"] = "Journal of bacteriology and virology : JBV",
  ["j bahrain med soc"] = "Journal of the Bahrain Medical Society = Majallat Jam'īyat al-Atibbā' al-Bahraynīyah",
  ["j balt stud"] = "Journal of Baltic studies",
  ["j baltimore coll dent surg"] = "Journal of the Baltimore College of Dental Surgery",
  ["j bangladesh acad sci"] = "Journal of Bangladesh Academy of Sciences. Bangladesh Academy of Sciences",
  ["j bangladesh coll phys surg"] = "Journal of Bangladesh College of Physicians & Surgeons",
  ["j bank financ"] = "Journal of banking & finance",
  ["j bank res"] = "Journal of Bank Research",
  ["j banking finance"] = "Journal of Banking and Finance",
  ["j baoji college arts sci nat sci"] = "Baoji College of Arts and Science",
  ["j basic appl mycol egypt"] = "Journal of Basic and Applied Mycology, Egypt",
  ["j basic appl zool"] = "Journal of basic & applied zoology",
  ["j basic clin med"] = "Journal of basic and clinical medicine",
  ["j basic clin pharm"] = "Journal of basic and clinical pharmacy",
  ["j basic clin physiol pharmacol"] = "Journal of basic and clinical physiology and pharmacology",
  ["j basic microbiol"] = "Journal of basic microbiology",
  ["j behav addict"] = "Journal of behavioral addictions",
  ["j behav anal health sports fit med"] = "Journal of behavior analysis in health, sports, fitness and medicine",
  ["j behav assess"] = "Journal of behavioral assessment",
  ["j behav brain sci"] = "Journal of behavioral and brain science",
  ["j behav cogn ther"] = "Journal of behavioral and cognitive therapy",
  ["j behav decis mak"] = "Journal of behavioral decision making",
  ["j behav econ"] = "The Journal of behavioral economics",
  ["j behav educ"] = "Journal of behavioral education",
  ["j behav exp econ"] = "Journal of behavioral and experimental economics",
  ["j behav exp finance"] = "Journal of behavioral and experimental finance",
  ["j behav financ"] = "The journal of behavioral finance",
  ["j behav health"] = "Journal of behavioral health",
  ["j behav health serv res"] = "The journal of behavioral health services & research",
  ["j behav med"] = "Journal of behavioral medicine",
  ["j behav neurosci res"] = "Journal of behavioral and neuroscience research",
  ["j behav soc sci (cedarv)"] = "Journal of behavioral and social sciences (Cedarville, Ohio)",
  ["j behav ther exp psychiatry"] = "Journal of behavior therapy and experimental psychiatry",
  ["j behavioral econ"] = "Journal of Behavioral Economics",
  ["j beijing inst tech"] = "Beijing Institute of Technology",
  ["j beijing inst technol (chin ed)"] = "Journal of Beijing Institute of Technology (Chinese Edition)",
  ["j beijing univ technol"] = "Journal of Beijing University of Technology. Beijing Gongye Daxue Xuebao",
  ["j belg soc radiol"] = "Journal of the Belgian Society of Radiology",
  ["j belge med phys"] = "Journal belge de médecine physique = Belgisch tijdschrift voor fysische geneeskunde",
  ["j belge med phys rehabil"] = "Journal belge de médecine physique et de réhabilitation. = Belgisch tijdschrift voor fysische geneeskunde en rehabilitatie",
  ["j belge med phys rhumatol"] = "Journal belge de médecine physique et de rhumatologie. Belgisch tijdschrift voor fysische geneeskunde en reumatologie",
  ["j belge neurol psychiatr"] = "Journal belge de neurologie et de psychiatrie",
  ["j belge radiol"] = "Journal belge de radiologie",
  ["j belge rhumatol med phys"] = "Journal belge de rhumatologie et de médecine physique = Belgisch tijdschrift voor reumatologie en fysische geneeskunde",
  ["j belge urol"] = "Journal belge d'urologie",
  ["j benefit cost anal"] = "Journal of benefit-cost analysis",
  ["j benefit-cost anal"] = "Journal of Benefit-Cost Analysis",
  ["j bengal tuberc assoc"] = "Journal. Bengal Tuberculosis Association",
  ["j bergen cty dent soc"] = "The journal of the Bergen County Dental Society. Bergen County Dental Society",
  ["j berry res"] = "Journal of berry research",
  ["j best pract health prof divers"] = "Journal of best practices in health professions diversity : research, education and policy",
  ["j biblic ethics med"] = "Journal of Biblical ethics in medicine",
  ["j big data"] = "Journal of big data",
  ["j big data anal transp"] = "Journal of Big Data Analytics in Transportation",
  ["j bihar math soc"] = "Bihar Mathematical Society. Journal",
  ["j binocul vis ocul motil"] = "Journal of binocular vision and ocular motility",
  ["j bio tribocorros"] = "Journal of bio- and tribo-corrosion",
  ["j bio- tribo-corros"] = "Journal of Bio- and Tribo-Corrosion",
  ["j bio-x res"] = "Journal of Bio-X Research",
  ["j bioact compat polym"] = "Journal of bioactive and compatible polymers",
  ["j bioanal biomed"] = "Journal of bioanalysis & biomedicine",
  ["j biobased mater bioenergy"] = "Journal of Biobased Materials and Bioenergy",
  ["j biochem"] = "Journal of Biochemistry",
  ["j biochem (tokyo)"] = "Journal of Biochemistry",
  ["j biochem (tokyo, jpn)"] = "Journal of Biochemistry (Tokyo, Japan)",
  ["j biochem anal stud"] = "Journal of biochemistry and analytical studies",
  ["j biochem bioph methods"] = "Journal of Biochemical and Biophysical Methods",
  ["j biochem biophys methods"] = "Journal of biochemical and biophysical methods",
  ["j biochem microbiol technol eng"] = "Journal of Biochemical and Microbiological Technology and Engineering",
  ["j biochem mol biol"] = "Journal of biochemistry and molecular biology",
  ["j biochem mol biol biophys"] = "Journal of biochemistry, molecular biology, and biophysics : JBMBB : the official journal of the Federation of Asian and Oceanian Biochemists and Molecular Biologists (FAOBMB)",
  ["j biochem mol toxicol"] = "Journal of Biochemical and Molecular Toxicology",
  ["j biochem pharmacol res"] = "Journal of biochemical and pharmacological research",
  ["j biochem physiol"] = "Journal of biochemistry and physiology",
  ["j biochem technol"] = "Journal of biochemical technology",
  ["j biochem toxicol"] = "Journal of biochemical toxicology",
  ["j biocommun"] = "The Journal of biocommunication",
  ["j biodivers"] = "Journal of Biodiversity",
  ["j biodivers environ sci"] = "Journal of biodiversity and environmental sciences",
  ["j bioecon"] = "Journal of Bioeconomics",
  ["j bioenerg"] = "Journal of bioenergetics",
  ["j bioenerg biomembr"] = "Journal of Bioenergetics and Biomembranes",
  ["j bioeng"] = "Journal of Bioengineering",
  ["j bioeng biomed sci"] = "Journal of bioengineering & biomedical science",
  ["j bioequivalence bioavailab"] = "Journal of bioequivalence & bioavailability",
  ["j bioequivalence bioavailability"] = "Journal of Bioequivalence & Bioavailability",
  ["j bioeth"] = "Journal of bioethics",
  ["j bioeth inq"] = "Journal of bioethical inquiry",
  ["j bioethical inq"] = "Journal of Bioethical Inquiry",
  ["j biofertil biopestic"] = "Journal of biofertilizers & biopesticides",
  ["j biogeogr"] = "Journal of biogeography",
  ["j bioinf comput biol"] = "Journal of Bioinformatics and Computational Biology",
  ["j bioinform"] = "Journal of bioinformatics",
  ["j bioinform comput biol"] = "Journal of bioinformatics and computational biology",
  ["j bioinform diabetes"] = "Journal of bioinformatics and diabetes",
  ["j bioinform genom"] = "Journal of bioinformatics and genomics",
  ["j bioinform proteom rev"] = "Journal of bioinformatics and proteomics review",
  ["j biol"] = "Journal of biology",
  ["j biol (denpasar)"] = "Jurnal biologi",
  ["j biol act prod nat"] = "Journal of Biologically Active Products from Nature",
  ["j biol agric healthc"] = "Journal of biology, agriculture and healthcare",
  ["j biol buccale"] = "Journal de biologie buccale",
  ["j biol chem"] = "The Journal of biological chemistry",
  ["j biol chem sci"] = "Journal of biological and chemical sciences",
  ["j biol dyn"] = "Journal of biological dynamics",
  ["j biol educ"] = "Journal of biological education",
  ["j biol eng"] = "Journal of biological engineering",
  ["j biol inorg chem"] = "Journal of biological inorganic chemistry : JBIC : a publication of the Society of Biological Inorganic Chemistry",
  ["j biol med sci"] = "Journal of biological and medical sciences",
  ["j biol methods"] = "Journal of biological methods",
  ["j biol nat"] = "Journal of biology and nature",
  ["j biol pharm chem res"] = "Journal of biological, pharmaceutical and chemical research",
  ["j biol photogr"] = "Journal of biological photography",
  ["j biol photogr assoc"] = "Journal of the Biological Photographic Association",
  ["j biol phys"] = "Journal of biological physics",
  ["j biol phys chem"] = "Journal of biological physics and chemistry : JBPC",
  ["j biol regul homeost agents"] = "Journal of biological regulators and homeostatic agents",
  ["j biol res (thessalon)"] = "Journal of biological research (Thessalonikē, Greece)",
  ["j biol res italy"] = "Journal Of Biological Research Italy",
  ["j biol res thessaloniki"] = "Journal of Biological Research-Thessaloniki",
  ["j biol response mod"] = "Journal of biological response modifiers",
  ["j biol rhythms"] = "Journal of biological rhythms",
  ["j biol sci"] = "The Journal of biological sciences",
  ["j biol sci (faisalabad)"] = "Journal of biological sciences (Faisalābād, Pakistan)",
  ["j biol stand"] = "Journal of biological standardization",
  ["j biol syst"] = "Journal of biological systems",
  ["j biol systems"] = "Journal of Biological Systems",
  ["j biolaw bus"] = "The journal of biolaw & business",
  ["j biologi indones"] = "Jurnal biologi Indonesia",
  ["j biolumin chemilumin"] = "Journal of Bioluminescence and Chemiluminescence",
  ["j biom biostat"] = "Journal of biometrics & biostatistics",
  ["j biomark"] = "Journal of biomarkers",
  ["j biomark drug dev"] = "Journal of biomarkers in drug development",
  ["j biomater appl"] = "Journal of biomaterials applications",
  ["j biomater dent"] = "Journal de biomatériaux dentaires : [publication du Collège français de biomatériaux dentaires]",
  ["j biomater nanobiotechnol"] = "Journal of biomaterials and nanobiotechnology",
  ["j biomater sci polym ed"] = "Journal of biomaterials science. Polymer edition",
  ["j biomater sci, polym ed"] = "Journal of Biomaterials Science Polymer Edition",
  ["j biomater tissue eng"] = "Journal of biomaterials and tissue engineering",
  ["j biomath"] = "Journal of Biomathematics. Shengwu Shuxue Xuebao",
  ["j biomech"] = "Journal of biomechanics",
  ["j biomech eng"] = "Journal of biomechanical engineering",
  ["j biomed (syd)"] = "Journal of Biomedicine (Sydney, NSW)",
  ["j biomed allied res"] = "Journal of biomedical and allied research",
  ["j biomed biotechnol"] = "Journal of biomedicine & biotechnology",
  ["j biomed discov collab"] = "Journal of biomedical discovery and collaboration",
  ["j biomed discovery collab"] = "Journal of Biomedical Discovery and Collaboration",
  ["j biomed educ"] = "Journal of biomedical education",
  ["j biomed eng"] = "Journal of Biomedical Engineering",
  ["j biomed eng inform"] = "Journal of biomedical engineering and informatics",
  ["j biomed eng med imaging"] = "Journal of biomedical engineering and medical imaging",
  ["j biomed graph comput"] = "Journal of biomedical graphics and computing",
  ["j biomed inf"] = "Journal of Biomedical Informatics",
  ["j biomed inf: x"] = "Journal of Biomedical Informatics: X",
  ["j biomed inform"] = "Journal of biomedical informatics",
  ["j biomed inform x"] = "Journal of biomedical informatics: X",
  ["j biomed lab sci"] = "Journal of biomedical and laboratory sciences",
  ["j biomed mater res"] = "Journal of Biomedical Materials Research",
  ["j biomed mater res a"] = "Journal of biomedical materials research. Part A",
  ["j biomed mater res b appl biomater"] = "Journal of biomedical materials research. Part B, Applied biomaterials",
  ["j biomed mater res part a"] = "Journal of Biomedial Materials Research Part A",
  ["j biomed mater res part b"] = "Journal of Biomedial Materials Research Part B Applied Biomaterials",
  ["j biomed mater res, part a"] = "Journal of Biomedical Materials Research, Part A",
  ["j biomed mater res, part b"] = "Journal of Biomedical Materials Research, Part B: Applied Biomaterials",
  ["j biomed nanotechnol"] = "Journal of biomedical nanotechnology",
  ["j biomed opt"] = "Journal of biomedical optics",
  ["j biomed photonics eng"] = "Journal of biomedical photonics & engineering",
  ["j biomed phys eng"] = "Journal of biomedical physics & engineering",
  ["j biomed res"] = "Journal of biomedical research",
  ["j biomed sci"] = "Journal of biomedical science",
  ["j biomed sci eng"] = "Journal of Biomedical Science and Engineering",
  ["j biomed semant"] = "Journal of Biomedical Semantics",
  ["j biomed semantics"] = "Journal of biomedical semantics",
  ["j biomed technol res"] = "Journal of biomedical technology and research",
  ["j biomim biomater tissue eng"] = "Journal of biomimetics, biomaterials, and tissue engineering",
  ["j biomimetics biomater tissue eng"] = "Journal of Biomimetics Biomaterials and Tissue Engineering",
  ["j biomol nmr"] = "Journal of Biomolecular NMR",
  ["j biomol res ther"] = "Journal of biomolecular research & therapeutics",
  ["j biomol screen"] = "Journal of biomolecular screening",
  ["j biomol screening"] = "Journal of Biomolecular Screening",
  ["j biomol struct dyn"] = "Journal of Biomolecular Structure and Dynamics",
  ["j biomol tech"] = "Journal of biomolecular techniques : JBT",
  ["j bionanosci"] = "Journal of Bionanoscience",
  ["j bionic eng"] = "Journal of Bionic Engineering",
  ["j biopharm stat"] = "Journal of biopharmaceutical statistics",
  ["j biophotonics"] = "Journal of biophotonics",
  ["j biophys"] = "Journal of biophysics (Hindawi Publishing Corporation : Online)",
  ["j biophys biochem cytol"] = "The Journal of biophysical and biochemical cytology",
  ["j bioprocess biotech"] = "Journal of bioprocessing & biotechniques",
  ["j biorem biodegrad"] = "Journal of Bioremediation & Biodegradation",
  ["j bioremediat biodegrad"] = "Journal of bioremediation & biodegradation",
  ["j bioresour bioprod"] = "Journal of Bioresources and Bioproducts",
  ["j biorheol"] = "Journal of Biorheology",
  ["j bios logos"] = "Jurnal bios logos",
  ["j biosaf biosecur"] = "Journal of biosafety and biosecurity",
  ["j biosci"] = "Journal of biosciences",
  ["j biosci (rajshari)"] = "Journal of bio-science",
  ["j biosci bioeng"] = "Journal of bioscience and bioengineering",
  ["j biosci biotechnol discov"] = "Journal of bioscience and biotechnology discovery",
  ["j biosci med"] = "The journal of bioscience and medicine",
  ["j biosci med (irvine)"] = "Journal of biosciences and medicines",
  ["j biosens bioelectron"] = "Journal of biosensors & bioelectronics",
  ["j biosens biomark diagn"] = "Journal of biosensors, biomarkers and diagnostics",
  ["j biosoc sci"] = "Journal of biosocial science",
  ["j biosoc sci suppl"] = "Journal of biosocial science. Supplement",
  ["j biosocial sci"] = "Journal of Biosocial Science",
  ["j biosyst eng"] = "Journal of Biosystems Engineering",
  ["j biota"] = "Jurnal biota",
  ["j biotech res"] = "Journal of biotech research",
  ["j biotechnol"] = "Journal of biotechnology",
  ["j biotechnol biomater"] = "Journal of biotechnology & biomaterials",
  ["j biotechnol: x"] = "Journal of Biotechnology: X",
  ["j bioterror biodef"] = "Journal of bioterrorism & biodefense",
  ["j biox res"] = "Journal of bio-X research",
  ["j bisex"] = "Journal of bisexuality",
  ["j black health perspect"] = "Journal of Black health perspectives",
  ["j black psychol"] = "The Journal of black psychology",
  ["j black sea mediterr environ"] = "Journal of the Black Sea/Mediterranean environment",
  ["j black sea/mediterr environ"] = "Journal of the Black Sea/Mediterranean Environment",
  ["j black sex relatsh"] = "Journal of black sexuality and relationships",
  ["j black stud"] = "Journal of black studies",
  ["j blockchain intell comput"] = "Journal of Blockchain and Intelligent Computing",
  ["j blockchain res"] = "Journal of Blockchain Research",
  ["j blood disord transfus"] = "Journal of blood disorders & transfusion",
  ["j blood lymph"] = "Journal of blood & lymph",
  ["j blood med"] = "Journal of blood medicine",
  ["j blood transfus"] = "Journal of blood transfusion",
  ["j board greenkeep res"] = "Journal of the Board of Greenkeeping Research",
  ["j bodyw mov ther"] = "Journal of bodywork and movement therapies",
  ["j bombay nat hist soc"] = "The Journal of the Bombay Natural History Society",
  ["j bone joint surg"] = "The Journal of bone and joint surgery",
  ["j bone joint surg am"] = "The Journal of bone and joint surgery. American volume",
  ["j bone joint surg br"] = "The Journal of bone and joint surgery. British volume",
  ["j bone jt infect"] = "Journal of bone and joint infection",
  ["j bone marrow res"] = "Journal of bone marrow research",
  ["j bone metab"] = "Journal of bone metabolism",
  ["j bone miner metab"] = "Journal of bone and mineral metabolism",
  ["j bone miner res"] = "Journal of bone and mineral research : the official journal of the American Society for Bone and Mineral Research",
  ["j bone oncol"] = "Journal of bone oncology",
  ["j bone rep recomm"] = "Journal of bone reports & recommendations",
  ["j bone res"] = "Journal of bone research",
  ["j borderl stud"] = "Journal of borderlands studies",
  ["j boston soc med sci"] = "Journal. Boston Society of Medical Sciences",
  ["j bot"] = "Journal of botany",
  ["j bot (faisalabad)"] = "Journal of botany (Faisalabad)",
  ["j bot res inst tex"] = "Journal of the Botanical Research Institute of Texas",
  ["j bot soc bengal"] = "Journal of the Botanical Society of Bengal",
  ["j bowman gray sch med wake for coll"] = "Journal. Bowman Gray School of Medicine",
  ["j br acad"] = "Journal of the British Academy",
  ["j br archaeol assoc"] = "Journal of the British Archaeological Association",
  ["j br endod soc"] = "Journal of the British Endodontic Society",
  ["j br fer soc"] = "Journal of the British Fertility Society",
  ["j br inst radio eng"] = "Journal of the British Institution of Radio Engineers",
  ["j br interplanet soc"] = "Journal of the British Interplanetary Society",
  ["j br menopause soc"] = "The journal of the British Menopause Society",
  ["j br stud"] = "The Journal of British studies",
  ["j br waterworks assoc"] = "Journal of the British Waterworks Association",
  ["j brachial plex peripher nerve inj"] = "Journal of brachial plexus and peripheral nerve injury",
  ["j brain dis"] = "Journal of brain disease",
  ["j bras doencas torac"] = "Jornal brasileiro de doenças torácicas",
  ["j bras ginecol"] = "Jornal brasileiro de ginecologia",
  ["j bras med"] = "Jornal brasileiro de medicina",
  ["j bras nefrol"] = "Jornal brasileiro de nefrologia : ʹorgão oficial de Sociedades Brasileira e Latino-Americana de Nefrologia",
  ["j bras neurol"] = "Jornal brasileiro de neurologia",
  ["j bras patol med lab"] = "Jornal brasileiro de patologia e medicina laboratorial",
  ["j bras pneumol"] = "Jornal brasileiro de pneumologia : publicaça̋o oficial da Sociedade Brasileira de Pneumologia e Tisilogia",
  ["j bras psiquiatr"] = "Jornal brasileiro de psiquiatria",
  ["j bras reprod assist"] = "Jornal brasileiro de reprodução assistida",
  ["j braz chem soc"] = "Journal of the Brazilian Chemical Society",
  ["j braz comput sci"] = "Journal of the Brazialian Computer Society",
  ["j braz comput soc"] = "Journal of the Brazilian Computer Society",
  ["j braz soc mech sci eng"] = "Journal of the Brazilian Society of Mechanical Sciences and Engineering",
  ["j breast cancer"] = "Journal of breast cancer",
  ["j breast health"] = "The journal of breast health",
  ["j breast imaging"] = "Journal of breast imaging",
  ["j breath res"] = "Journal of breath research",
  ["j bridge eng"] = "Journal of Bridge Engineering",
  ["j broadcast electron media"] = "Journal of broadcasting & electronic media",
  ["j bronchology interv pulmonol"] = "Journal of bronchology & interventional pulmonology",
  ["j bryol"] = "Journal of bryology",
  ["j buddh ethics"] = "Journal of Buddhist ethics",
  ["j buffalo sci"] = "Journal of buffalo science",
  ["j build eng"] = "Journal of Building Engineering",
  ["j build pathol rehabil"] = "Journal of Building Pathology and Rehabilitation",
  ["j build perform simul"] = "Journal of Building Performance Simulation",
  ["j build phys"] = "Journal of Building Physics",
  ["j buon"] = "Journal of B.U.ON. : official journal of the Balkan Union of Oncology",
  ["j burma stud"] = "The journal of Burma studies",
  ["j burn care rehabil"] = "The Journal of burn care & rehabilitation",
  ["j burn care res"] = "Journal of burn care & research : official publication of the American Burn Association",
  ["j burns wounds"] = "Journal of burns and wounds",
  ["j bus"] = "Journal of business (Chicago, Ill.)",
  ["j bus contin emer plan"] = "Journal of business continuity & emergency planning",
  ["j bus econ stat"] = "Journal of business & economic statistics : a publication of the American Statistical Association",
  ["j bus econ statist"] = "Journal of Business and Economic Statistics",
  ["j bus econom statist"] = "Journal of Business & Economic Statistics",
  ["j bus ethics"] = "Journal of business ethics : JBE",
  ["j bus finance account"] = "Journal of business finance & accounting",
  ["j bus psychol"] = "Journal of business and psychology",
  ["j bus res"] = "Journal of business research",
  ["j bus soc stud"] = "The Journal of business and social studies",
  ["j bus strategy"] = "The Journal of business strategy",
  ["j bus tech commun"] = "Journal of business and technical communication",
  ["j bus univ chic"] = "The Journal of business of the University of Chicago",
  ["j bus ventur"] = "Journal of business venturing",
  ["j cachexia sarcopenia muscle"] = "Journal of cachexia, sarcopenia and muscle",
  ["j caffeine adenosine res"] = "Journal of caffeine and adenosine research",
  ["j caffeine res"] = "Journal of caffeine research",
  ["j calif alliance ment ill"] = "The journal of the California Alliance for the Mentally Ill",
  ["j calif dent assoc"] = "Journal of the California Dental Association",
  ["j calif gt basin anthropol"] = "Journal of California and Great Basin anthropology",
  ["j calif state dent assoc"] = "Journal of the California State Dental Association",
  ["j camel pract res"] = "Journal of Camel Practice and Research",
  ["j can acad child adolesc psychiatry"] = "Journal of the Canadian Academy of Child and Adolescent Psychiatry = Journal de l'Académie canadienne de psychiatrie de l'enfant et de l'adolescent",
  ["j can assoc gastroenterol"] = "Journal of the Canadian Association of Gastroenterology",
  ["j can assoc radiol"] = "Journal of the Canadian Association of Radiologists",
  ["j can chiropr assoc"] = "The Journal of the Canadian Chiropractic Association",
  ["j can church hist soc"] = "The Journal of the Canadian Church Historical Society",
  ["j can dent assoc"] = "Journal (Canadian Dental Association)",
  ["j can dent assoc (tor)"] = "Journal of the Canadian Dental Association",
  ["j can diet assoc"] = "Journal of the Canadian Dietetic Association",
  ["j can health lib assoc"] = "Journal of the Canadian Health Libraries Association",
  ["j can health libr assoc"] = "The journal of the Canadian Health Libraries Association",
  ["j can med serv"] = "The Journal of the Canadian medical services",
  ["j can pet technol"] = "Journal of Canadian Petroleum Technology",
  ["j can res updates"] = "Journal of cancer research updates",
  ["j can sci terre"] = "Journal Canadien des Sciences de la Terre",
  ["j can stud"] = "Journal of Canadian studies. Revue d'études canadiennes",
  ["j cancer"] = "Journal of Cancer",
  ["j cancer biol"] = "Journal of cancer biology",
  ["j cancer biol res"] = "Journal of cancer biology & research",
  ["j cancer clin oncol"] = "Journal of cancer and clinical oncology",
  ["j cancer clin trials"] = "Journal of cancer clinical trials",
  ["j cancer educ"] = "Journal of cancer education : the official journal of the American Association for Cancer Education",
  ["j cancer epidemiol"] = "Journal of cancer epidemiology",
  ["j cancer epidemiol prev"] = "Journal of cancer epidemiology and prevention",
  ["j cancer epidemiol prev (imedpub)"] = "Journal of cancer epidemiology and prevention (iMedPub)",
  ["j cancer epidemiol treat"] = "Journal of cancer epidemiology & treatment",
  ["j cancer immunol (wilmington)"] = "Journal of cancer immunology",
  ["j cancer metastasis treat"] = "Journal of cancer metastasis and treatment",
  ["j cancer mol"] = "Journal of Cancer Molecules",
  ["j cancer policy"] = "Journal of cancer policy",
  ["j cancer prev"] = "Journal of cancer prevention",
  ["j cancer prev curr res"] = "Journal of cancer prevention & current research",
  ["j cancer rehabil"] = "Journal of cancer rehabilitation",
  ["j cancer res"] = "The Journal of cancer research",
  ["j cancer res cell ther"] = "Journal of cancer research and cellular therapeutics",
  ["j cancer res clin oncol"] = "Journal of cancer research and clinical oncology",
  ["j cancer res forecast"] = "Journal of cancer research forecast",
  ["j cancer res oncobiol"] = "Journal of cancer research and oncobiology",
  ["j cancer res ther"] = "Journal of cancer research and therapeutics",
  ["j cancer res ther (manch)"] = "Journal of cancer research & therapy",
  ["j cancer res ther oncol"] = "Journal of cancer research and therapeutic oncology",
  ["j cancer sci"] = "Journal of cancer sciences",
  ["j cancer sci clin ther"] = "Journal of cancer science and clinical therapeutics",
  ["j cancer sci ther"] = "Journal of cancer science & therapy",
  ["j cancer stem cell res"] = "Journal of cancer stem cell research",
  ["j cancer surviv"] = "Journal of cancer survivorship : research and practice",
  ["j cancer ther"] = "Journal of cancer therapy",
  ["j cancer ther res"] = "Journal of cancer therapeutics & research",
  ["j cancer treat res"] = "Journal of cancer treatment and research",
  ["j cancer treatment diagn"] = "Journal of cancer treatment & diagnosis",
  ["j cannabis res"] = "Journal of cannabis research",
  ["j cannabis ther"] = "Journal of Cannabis Therapeutics",
  ["j cannt"] = "Le Journal CANNT = CANNT journal : the journal of the Canadian Association of Nephrology Nurses and Technicians",
  ["j capill electrophor microchip technol"] = "Journal of capillary electrophoresis and microchip technology",
  ["j capillary electrophor"] = "Journal of capillary electrophoresis",
  ["j carbohydr chem"] = "Journal of Carbohydrate Chemistry",
  ["j carcinog"] = "Journal of carcinogenesis",
  ["j carcinog mutagen"] = "Journal of carcinogenesis & mutagenesis",
  ["j card fail"] = "Journal of cardiac failure",
  ["j card surg"] = "Journal of cardiac surgery",
  ["j cardiobiol"] = "Journal of cardiobiology",
  ["j cardiogr"] = "Journal of cardiography",
  ["j cardiogr suppl"] = "Journal of cardiography. Supplement",
  ["j cardiol"] = "Journal of cardiology",
  ["j cardiol cardiovasc med"] = "Journal of cardiology and cardiovascular medicine",
  ["j cardiol cardiovasc sci"] = "Journal of cardiology and cardiovascular sciences",
  ["j cardiol cardiovasc ther"] = "Journal of cardiology & cardiovascular therapy",
  ["j cardiol cases"] = "Journal of cardiology cases",
  ["j cardiol clin res"] = "Journal of cardiology & clinical research",
  ["j cardiol curr res"] = "Journal of cardiology & current research",
  ["j cardiol suppl"] = "Journal of cardiology. Supplement",
  ["j cardiol ther"] = "Journal of cardiology and therapeutics",
  ["j cardiol ther (hong kong)"] = "Journal of cardiology and therapy",
  ["j cardiometab syndr"] = "Journal of the cardiometabolic syndrome",
  ["j cardiopulm rehabil"] = "Journal of cardiopulmonary rehabilitation",
  ["j cardiopulm rehabil prev"] = "Journal of cardiopulmonary rehabilitation and prevention",
  ["j cardiorespir dis"] = "Journal of cardio-respiratory diseases",
  ["j cardiothorac anesth"] = "Journal of cardiothoracic anesthesia",
  ["j cardiothorac surg"] = "Journal of cardiothoracic surgery",
  ["j cardiothorac trauma"] = "The journal of cardiothoracic trauma",
  ["j cardiothorac vasc anesth"] = "Journal of cardiothoracic and vascular anesthesia",
  ["j cardiovasc comput tomogr"] = "Journal of cardiovascular computed tomography",
  ["j cardiovasc dev dis"] = "Journal of cardiovascular development and disease",
  ["j cardiovasc diagn proced"] = "Journal of cardiovascular diagnosis and procedures",
  ["j cardiovasc dis"] = "Journal of cardiovascular disease",
  ["j cardiovasc dis diagn"] = "Journal of cardiovascular diseases & diagnosis",
  ["j cardiovasc dis res"] = "Journal of cardiovascular disease research",
  ["j cardiovasc echogr"] = "Journal of cardiovascular echography",
  ["j cardiovasc electrophysiol"] = "Journal of cardiovascular electrophysiology",
  ["j cardiovasc imaging"] = "Journal of cardiovascular imaging",
  ["j cardiovasc magn reson"] = "Journal of cardiovascular magnetic resonance : official journal of the Society for Cardiovascular Magnetic Resonance",
  ["j cardiovasc manag"] = "The Journal of cardiovascular management : the official journal of the American College of Cardiovascular Administrators",
  ["j cardiovasc med"] = "The Journal of cardiovascular medicine",
  ["j cardiovasc med (hagerstown)"] = "Journal of cardiovascular medicine (Hagerstown, Md.)",
  ["j cardiovasc nurs"] = "The Journal of cardiovascular nursing",
  ["j cardiovasc pharmacol"] = "Journal of Cardiovascular Pharmacology",
  ["j cardiovasc pharmacol ther"] = "Journal of cardiovascular pharmacology and therapeutics",
  ["j cardiovasc risk"] = "Journal of cardiovascular risk",
  ["j cardiovasc surg (torino)"] = "The Journal of cardiovascular surgery",
  ["j cardiovasc thorac res"] = "Journal of cardiovascular and thoracic research",
  ["j cardiovasc transl res"] = "Journal of cardiovascular translational research",
  ["j cardiovasc ultrasound"] = "Journal of cardiovascular ultrasound",
  ["j cardpulm rehabil"] = "Journal of Cardiopulmonary Rehabilitation",
  ["j care serv manag"] = "Journal of care services management",
  ["j career assess"] = "Journal of career assessment",
  ["j career dev"] = "Journal of career development",
  ["j caribb hist"] = "The Journal of Caribbean history",
  ["j caribb ornithol"] = "The journal of Caribbean ornithology",
  ["j caring sci"] = "Journal of caring sciences",
  ["j case manag"] = "Journal of case management",
  ["j case rep med"] = "Journal of case reports in medicine",
  ["j catal"] = "Journal of Catalysis",
  ["j cataract refract surg"] = "Journal of cataract and refractive surgery",
  ["j cathol nurses guild engl wales"] = "Journal of the Catholic Nurses Guild of England and Wales",
  ["j causal inference"] = "Journal of causal inference",
  ["j caves karst stud"] = "Journal of cave and karst studies : the National Speleological Society bulletin",
  ["j cell autom"] = "Journal of Cellular Automata",
  ["j cell biochem"] = "Journal of Cellular Biochemistry",
  ["j cell biochem suppl"] = "Journal of cellular biochemistry. Supplement",
  ["j cell biol"] = "The Journal of cell biology",
  ["j cell biol histol"] = "Journal of cell biology and histology",
  ["j cell biol mol sci"] = "Journal of cellular biology & molecular sciences",
  ["j cell biotechnol"] = "Journal of cellular biotechnology",
  ["j cell commun signal"] = "Journal of cell communication and signaling",
  ["j cell commun signaling"] = "Journal of Cell Communication and Signaling",
  ["j cell comp physiol"] = "Journal of cellular and comparative physiology",
  ["j cell death"] = "Journal of cell death",
  ["j cell immunol"] = "Journal of cellular immunology",
  ["j cell immunother"] = "Journal of Cellular Immunotherapy",
  ["j cell mol med"] = "Journal of cellular and molecular medicine",
  ["j cell physiol"] = "Journal of Cellular Physiology",
  ["j cell physiol suppl"] = "Journal of cellular physiology. Supplement",
  ["j cell plast"] = "Journal of Cellular Plastics",
  ["j cell sci"] = "Journal of Cell Science",
  ["j cell sci suppl"] = "Journal of cell science. Supplement",
  ["j cell sci ther"] = "Journal of cell science & therapy",
  ["j cell signal"] = "Journal of cell signaling",
  ["j cent china norm univ nat sci"] = "Journal of Central China Normal University. Natural Sciences",
  ["j cent nerv syst dis"] = "Journal of central nervous system disease",
  ["j cent south univ"] = "Journal of Central South University",
  ["j central china normal univ natur sci"] = "Journal of Central China Normal University",
  ["j ceram"] = "Journal of Ceramics",
  ["j ceram process res"] = "Journal of Ceramic Processing Research",
  ["j ceram sci technol"] = "Journal of Ceramic Science and Technology",
  ["j ceram soc jpn"] = "Journal of the Ceramic Society of Japan",
  ["j cereal sci"] = "Journal of Cereal Science",
  ["j cereb blood flow metab"] = "Journal of cerebral blood flow and metabolism : official journal of the International Society of Cerebral Blood Flow and Metabolism",
  ["j cerebrovasc endovasc neurosurg"] = "Journal of cerebrovascular and endovascular neurosurgery",
  ["j cetacean res manag"] = "The journal of cetacean research and management",
  ["j changsha comm univ"] = "Changsha Communications University",
  ["j changsha univ electr power nat sci ed"] = "Changsha University of Electric Power",
  ["j character educ"] = "Journal of character education",
  ["j charles h tweed int found"] = "Journal of the Charles H. Tweed International Foundation",
  ["j chem"] = "Journal of chemistry",
  ["j chem biol"] = "Journal of chemical biology",
  ["j chem biol phys sci"] = "Journal of chemical, biological, and physical sciences",
  ["j chem chem eng"] = "Journal of chemistry and chemical engineering",
  ["j chem crystallogr"] = "Journal of Chemical Crystallography",
  ["j chem doc"] = "Journal of Chemical Documentation",
  ["j chem ecol"] = "Journal of chemical ecology",
  ["j chem educ"] = "Journal of Chemical Education",
  ["j chem eng chem res"] = "Journal of chemical engineering and chemistry research",
  ["j chem eng data"] = "Journal of Chemical and Engineering Data",
  ["j chem eng jpn"] = "Journal of Chemical Engineering of Japan",
  ["j chem health saf"] = "Journal of Chemical Heath and Safetyu",
  ["j chem inf comput sci"] = "Journal of Chemical Information and Computer Sciences",
  ["j chem inf model"] = "Journal of chemical information and modeling",
  ["j chem metrol"] = "Journal of Chemical Metrology",
  ["j chem neuroanat"] = "Journal of chemical neuroanatomy",
  ["j chem pharm res"] = "Journal of chemical and pharmaceutical research",
  ["j chem phys"] = "The Journal of chemical physics",
  ["j chem res"] = "Journal of Chemical Research",
  ["j chem res synop"] = "Journal of chemical research. Synopses",
  ["j chem res, synop "] = "Journal of Chemical Research Synopsis",
  ["j chem sci"] = "Journal of Chemical Sciences",
  ["j chem sci (bangalore)"] = "Journal of chemical sciences (Bangalore, India)",
  ["j chem sci (bangalore, india)"] = "Journal of Chemical Sciences (Bangalore, India)",
  ["j chem sci (berlin, ger)"] = "Journal of Chemical Sciences (Berlin, Germany)",
  ["j chem soc"] = "Journal of the Chemical Society",
  ["j chem soc [perkin 1]"] = "Journal of the Chemical Society. Perkin Transactions 1",
  ["j chem soc a"] = "Journal of the Chemical Society A: Inorganic, Physical, Theoretical",
  ["j chem soc abstr"] = "Journal of the Chemical Society, Abstracts",
  ["j chem soc b"] = "Journal of the Chemical Society B: Physical Organic",
  ["j chem soc c"] = "Journal of the Chemical Society C: Organic",
  ["j chem soc chem commun"] = "Journal of the Chemical Society. Chemical communications",
  ["j chem soc d"] = "Journal of the Chemical Society D: Chemical Communications",
  ["j chem soc pak"] = "Journal of the Chemical Society of Pakistan",
  ["j chem soc perkin 1"] = "Journal of the Chemical Society. Perkin transactions 1",
  ["j chem soc trans"] = "Journal of the Chemical Society, Transactions",
  ["j chem soc, chem commun"] = "Journal of the Chemical Society, Chemical Communications",
  ["j chem soc, dalton trans"] = "Journal of the Chemical Society, Dalton Transactions",
  ["j chem soc, faraday trans"] = "Journal of the Chemical Society, Faraday Transactions",
  ["j chem soc, faraday trans 1 f"] = "Journal of the Chemical Society, Faraday Transactions 1: Physical Chemistry in Condensed Phases",
  ["j chem soc, faraday trans 2"] = "Journal of the Chemical Society, Faraday Transactions 2: Molecular and Chemical Physics",
  ["j chem soc, perkin trans 1"] = "Journal of the Chemical Society, Perkin Transactions 1",
  ["j chem soc, perkin trans 2"] = "Journal of the Chemical Society, Perkin Transactions 2",
  ["j chem software"] = "Journal of Chemical Software",
  ["j chem technol biotechnol"] = "Journal of Chemical Technology and Biotechnology",
  ["j chem technol biotechnol biotechnol"] = "Journal of Chemical Technology and Biotechnology. Biotechnology",
  ["j chem technol biotechnol chem technol"] = "Journal of Chemical Technology and Biotechnology. Chemical Technology",
  ["j chem technol metall"] = "Journal Of Chemical Technology And Metallurgy",
  ["j chem theory comput"] = "Journal of chemical theory and computation",
  ["j chem thermodyn"] = "Journal of Chemical Thermodynamics",
  ["j cheminf"] = "Journal of Cheminformatics",
  ["j cheminform"] = "Journal of cheminformatics",
  ["j chemom"] = "Journal of Chemometrics",
  ["j chemother"] = "Journal of chemotherapy (Florence, Italy)",
  ["j chengdu univ natur sci"] = "Journal of Chengdu University",
  ["j chengdu univ sci tech"] = "Journal of Chengdu University of Science and Technology",
  ["j cherokee stud"] = "Journal of Cherokee studies",
  ["j chest surg"] = "Journal of chest surgery",
  ["j chester archaeol soc"] = "Journal of the Chester Archaeological Society. Chester Archaeological Society",
  ["j chiba med soc"] = "Chiba Igakkai zasshi. The Journal of Chiba Medical Society",
  ["j chil chem soc"] = "Journal of the Chilean Chemical Society",
  ["j child adolesc behav"] = "Journal of child and adolescent behavior",
  ["j child adolesc couns"] = "Journal of child and adolescent counseling",
  ["j child adolesc ment health"] = "Journal of child and adolescent mental health",
  ["j child adolesc psychiatr ment health nurs"] = "Journal of child and adolescent psychiatric and mental health nursing",
  ["j child adolesc psychiatr nurs"] = "Journal of child and adolescent psychiatric nursing : official publication of the Association of Child and Adolescent Psychiatric Nurses, Inc",
  ["j child adolesc psychopharmacol"] = "Journal of child and adolescent psychopharmacology",
  ["j child adolesc subst abuse"] = "Journal of child & adolescent substance abuse",
  ["j child adolesc trauma"] = "Journal of child & adolescent trauma",
  ["j child asthma res inst hosp denver"] = "Journal. Children's Asthma Research Institute and Hospital (Denver, Colo.)",
  ["j child custody"] = "Journal of child custody",
  ["j child dev disord"] = "Journal of childhood & developmental disorders",
  ["j child fam nurs"] = "Journal of child and family nursing",
  ["j child fam stud"] = "Journal of child and family studies",
  ["j child health care"] = "Journal of child health care : for professionals working with children in the hospital and community",
  ["j child lang"] = "Journal of child language",
  ["j child lang acquis dev"] = "Journal of child language acquisition and development",
  ["j child media"] = "Journal of children and media",
  ["j child neurol"] = "Journal of child neurology",
  ["j child nutr manag"] = "Journal of child nutrition & management (Online)",
  ["j child obes"] = "Journal of childhood obesity",
  ["j child orthop"] = "Journal of children's orthopaedics",
  ["j child poverty"] = "Journal of children & poverty",
  ["j child psychiatry"] = "The Journal of child psychiatry",
  ["j child psychol psychiatry"] = "Journal of child psychology and psychiatry, and allied disciplines",
  ["j child sci"] = "Journal of Child Science",
  ["j child serv"] = "Journal of children's services",
  ["j child sex abus"] = "Journal of child sexual abuse",
  ["j chim phys"] = "Journal de Chimie Physique",
  ["j chim phys phys- chim biol"] = "Journal de Chimie Physique et de Physico-Chimie Biologique",
  ["j chim phys physicochim biol"] = "Journal de chimie physique et de physicochimie biologique",
  ["j chin adv mater soc"] = "Journal of the Chinese Advanced Materials Society",
  ["j chin chem soc"] = "Journal of the Chinese Chemical Society",
  ["j chin chem soc (taipei, taiwan)"] = "Journal of the Chinese Chemical Society (Taipei, Taiwan)",
  ["j chin chem soc (weinheim, ger)"] = "Journal of the Chinese Chemical Society (Weinheim, Germany)",
  ["j chin inst chem eng,"] = "Journal of the Chinese Institute of Chemical Engineers",
  ["j chin inst eng"] = "Journal of the Chinese Institute of Engineers",
  ["j chin med"] = "The Journal of Chinese medicine",
  ["j chin med assoc"] = "Journal of the Chinese Medical Association : JCMA",
  ["j chin polit sci"] = "Journal of Chinese political science",
  ["j chin soc anim sci"] = "Zhongguo xu mu xue hui hui zhi = Journal of the Chinese Society of Animal Science",
  ["j chin univ hong kong"] = "Xianggang Zhong wen da xue xue bao. Chinese University of Hong Kong",
  ["j china univ geosci"] = "Journal of the China University of Geosciences",
  ["j china univ min technol"] = "Journal of China University of Mining and Technology",
  ["j china univ sci tech"] = "Journal of China University of Science and Technology",
  ["j chinese inst engrs"] = "Chinese Institute of Engineers",
  ["j chir (paris)"] = "Journal de chirurgie",
  ["j chir visc"] = "Journal de chirurgie viscérale",
  ["j chiropr educ"] = "The Journal of chiropractic education",
  ["j chiropr humanit"] = "Journal of chiropractic humanities",
  ["j chiropr med"] = "Journal of chiropractic medicine",
  ["j christ med assoc india"] = "The Journal of the Christian Medical Association of India",
  ["j christ nurs"] = "Journal of Christian nursing : a quarterly publication of Nurses Christian Fellowship",
  ["j chromatogr"] = "Journal of Chromatography",
  ["j chromatogr a"] = "Journal of chromatography. A",
  ["j chromatogr b"] = "Journal of Chromatography B",
  ["j chromatogr b analyt technol biomed life sci"] = "Journal of chromatography. B, Analytical technologies in the biomedical and life sciences",
  ["j chromatogr b biomed appl"] = "Journal of chromatography. B, Biomedical applications",
  ["j chromatogr b biomed sci app"] = "Journal of Chromatography. B, Biomedical Sciences and Applications",
  ["j chromatogr b biomed sci appl"] = "Journal of chromatography. B, Biomedical sciences and applications",
  ["j chromatogr b: anal technol biomed life sci"] = "Journal of Chromatography B: Analytical Technologies in the Biomedical and Life Sciences",
  ["j chromatogr open"] = "Journal of Chromatography Open",
  ["j chromatogr sci"] = "Journal of Chromatographic Science",
  ["j chromatogr sep tech"] = "Journal of Chromatography and Separation Techniques",
  ["j chromatogr suppl"] = "Journal of chromatography. Supplementary volume",
  ["j chromatogr, a"] = "Journal of Chromatography, A",
  ["j chromatogr, b: anal technol biomed life sci"] = "Journal of Chromatography, B: Analytical Technologies in the Biomedical and Life Sciences",
  ["j chronic dis"] = "Journal of chronic diseases",
  ["j chronic dis manag"] = "Journal of chronic diseases and management",
  ["j chronic fatigue syndr"] = "Journal of chronic fatigue syndrome",
  ["j chungcheong math soc"] = "Journal of the Chungcheong Mathematical Society",
  ["j church state"] = "A Journal of Church and State",
  ["j circ biomark"] = "Journal of circulating biomarkers",
  ["j circadian rhythms"] = "Journal of circadian rhythms",
  ["j circuits, syst comput"] = "Journal of Circuits, Systems and Computers",
  ["j citrus pathol"] = "Journal of citrus pathology",
  ["j civ eng educ"] = "Journal of Civl Engineering Education",
  ["j civ eng manage"] = "Journal of Civil Engineering and Management",
  ["j civ struct health monit"] = "Journal of civil structural health monitoring",
  ["j civ war med"] = "The journal of Civil War medicine : official quarterly publication of the Society of Civil War Surgeons, Inc",
  ["j class anal"] = "Journal of Classical Analysis",
  ["j classif"] = "Journal of classification",
  ["j classification"] = "Journal of Classification",
  ["j clean energy energy storage"] = "Journal of Clean Energy and Energy Storage",
  ["j clean prod"] = "Journal of cleaner production",
  ["j cleaner prod"] = "Journal of Cleaner Production",
  ["j cleft lip palate craniofacial anomalies"] = "Journal of cleft lip palate and craniofacial anomalies",
  ["j clim"] = "Journal of climate",
  ["j clim action res policy"] = "Journal of Climate Action, Research, and Policy",
  ["j clim appl meteorol"] = "Journal of Climate and Applied Meteorology",
  ["j clim change"] = "Journal of Climate Change",
  ["j clim change health"] = "Journal of Climate Change and Health",
  ["j clim meteorol"] = "Journal of Climate Meteorology",
  ["j clim resilience clim justice"] = "Journal of Climate Resilience & Climate Justice",
  ["j climate"] = "Journal of Climate",
  ["j climatol"] = "Journal of Climatology",
  ["j clin"] = "Jornal dos clinicos",
  ["j clin aesthet dermatol"] = "The Journal of clinical and aesthetic dermatology",
  ["j clin anat pathol (jcap)"] = "Journal of clinical & anatomic pathology (JCAP)",
  ["j clin anesth"] = "Journal of clinical anesthesia",
  ["j clin anesth manag"] = "Journal of clinical anesthesia and management",
  ["j clin anesth pain manag"] = "Journal of clinical anesthesia and pain management",
  ["j clin apher"] = "Journal of clinical apheresis",
  ["j clin apheresis"] = "Journal of Clinical Apheresis",
  ["j clin biochem nutr"] = "Journal of clinical biochemistry and nutrition",
  ["j clin bioinf"] = "Journal of Clinical Bioinformatics",
  ["j clin bioinforma"] = "Journal of clinical bioinformatics",
  ["j clin cardiol cardiovasc ther"] = "Journal of clinical cardiology and cardiovascular therapy",
  ["j clin case rep"] = "Journal of clinical case reports",
  ["j clin cases rep"] = "Journal of clinical cases & reports",
  ["j clin cell immunol"] = "Journal of clinical & cellular immunology",
  ["j clin chem clin biochem"] = "Journal of Clinical Chemistry and Clinical Biochemistry",
  ["j clin child adolesc psychol"] = "Journal of clinical child and adolescent psychology : the official journal for the Society of Clinical Child and Adolescent Psychology, American Psychological Association, Division 53",
  ["j clin child psychol"] = "Journal of clinical child psychology",
  ["j clin comput"] = "Journal of clinical computing",
  ["j clin densitom"] = "Journal of clinical densitometry : the official journal of the International Society for Clinical Densitometry",
  ["j clin dent"] = "The Journal of clinical dentistry",
  ["j clin dermatol"] = "Journal of clinical dermatology",
  ["j clin diagn res"] = "Journal of clinical and diagnostic research : JCDR",
  ["j clin dysmorphol"] = "The Journal of clinical dysmorphology",
  ["j clin endocrinol"] = "The Journal of clinical endocrinology",
  ["j clin endocrinol diabetes"] = "Journal of clinical endocrinology and diabetes",
  ["j clin endocrinol metab"] = "Journal of Clinical Endocrinology and Metabolism",
  ["j clin eng"] = "Journal of clinical engineering",
  ["j clin epidemiol"] = "Journal of clinical epidemiology",
  ["j clin epigenet"] = "Journal of clinical epigenetics",
  ["j clin ethics"] = "The Journal of clinical ethics",
  ["j clin exerc physiol"] = "Journal of clinical exercise physiology",
  ["j clin exp cardiolog"] = "Journal of clinical & experimental cardiology",
  ["j clin exp dent"] = "Journal of clinical and experimental dentistry",
  ["j clin exp dermatol res"] = "Journal of clinical & experimental dermatology research",
  ["j clin exp hematop"] = "Journal of clinical and experimental hematopathology : JCEH",
  ["j clin exp hepatol"] = "Journal of clinical and experimental hepatology",
  ["j clin exp immunol"] = "Journal of clinical & experimental immunology",
  ["j clin exp nephrol"] = "Journal of clinical & experimental nephrology",
  ["j clin exp neuropsychol"] = "Journal of clinical and experimental neuropsychology",
  ["j clin exp oncol"] = "Journal of clinical & experimental oncology",
  ["j clin exp ophthalmol"] = "Journal of clinical & experimental ophthalmology",
  ["j clin exp pathol"] = "Journal of clinical & experimental pathology",
  ["j clin exp psychopathol"] = "Journal of clinical and experimental psychopathology",
  ["j clin exp psychopathol q rev psychiatry neurol"] = "Journal of clinical and experimental psychopathology & quarterly review of psychiatry and neurology",
  ["j clin exp res cardiol"] = "Journal of clinical and experimental research in cardiology",
  ["j clin forensic med"] = "Journal of clinical forensic medicine",
  ["j clin gastroenterol"] = "Journal of clinical gastroenterology",
  ["j clin gastroenterol hepatol"] = "Journal of clinical gastroenterology and hepatology",
  ["j clin genom"] = "Journal of clinical genomics",
  ["j clin gynecol obstet"] = "Journal of clinical gynecology and obstetrics",
  ["j clin haematol"] = "Journal of clinical haematology",
  ["j clin hosp pharm"] = "Journal of clinical and hospital pharmacy",
  ["j clin hypertens"] = "Journal of clinical hypertension",
  ["j clin hypertens (greenwich)"] = "Journal of clinical hypertension (Greenwich, Conn.)",
  ["j clin imaging sci"] = "Journal of clinical imaging science",
  ["j clin immunol"] = "Journal of clinical immunology",
  ["j clin insights neonatol"] = "Journal of Clinical Insights in Neonatology",
  ["j clin invest"] = "The Journal of clinical investigation",
  ["j clin investig dermatol"] = "Journal of clinical and investigative dermatology",
  ["j clin issues psychol"] = "Journal of clinical issues in psychology",
  ["j clin lab anal"] = "Journal of clinical laboratory analysis",
  ["j clin lab immunol"] = "Journal of clinical & laboratory immunology",
  ["j clin lab med"] = "Journal of clinical and laboratory medicine",
  ["j clin laser med surg"] = "Journal of clinical laser medicine & surgery",
  ["j clin lipidol"] = "Journal of clinical lipidology",
  ["j clin med"] = "Journal of clinical medicine",
  ["j clin med res"] = "Journal of clinical medicine research",
  ["j clin med ther"] = "Journal of clinical medicine and therapeutics",
  ["j clin metab diabetes"] = "Journal of clinical metabolism & diabetes",
  ["j clin microbiol"] = "Journal of clinical microbiology",
  ["j clin mol endocrinol"] = "Journal of clinical and molecular endocrinology",
  ["j clin monit"] = "Journal of clinical monitoring",
  ["j clin monit comput"] = "Journal of clinical monitoring and computing",
  ["j clin mov disord"] = "Journal of clinical movement disorders",
  ["j clin neonatol"] = "Journal of clinical neonatology",
  ["j clin nephrol ren care"] = "Journal of clinical nephrology and renal care",
  ["j clin nephrol res"] = "Journal of clinical nephrology and research",
  ["j clin neurol"] = "Journal of clinical neurology (Seoul, Korea)",
  ["j clin neuromuscul dis"] = "Journal of clinical neuromuscular disease",
  ["j clin neuroophthalmol"] = "Journal of clinical neuro-ophthalmology",
  ["j clin neurophysiol"] = "Journal of clinical neurophysiology : official publication of the American Electroencephalographic Society",
  ["j clin neuropsychol"] = "Journal of clinical neuropsychology",
  ["j clin neurosci"] = "Journal of clinical neuroscience : official journal of the Neurosurgical Society of Australasia",
  ["j clin nurs"] = "Journal of clinical nursing",
  ["j clin nutr"] = "The Journal of clinical nutrition",
  ["j clin nutr food sci"] = "Journal of clinical nutrition and food science",
  ["j clin nutr metab"] = "Journal of clinical nutrition and metabolism",
  ["j clin oncol"] = "Journal of clinical oncology : official journal of the American Society of Clinical Oncology",
  ["j clin oncol res"] = "Journal of clinical oncology and research",
  ["j clin oncol: clin cancer inf"] = "Journal of Clinical Oncology: Clinical Cancer Informatics",
  ["j clin oncol: precis oncol"] = "Journal of Clinical Oncology: Precision Oncology",
  ["j clin ophthalmol eye disord"] = "Journal of clinical ophthalmology & eye disorders",
  ["j clin orthod"] = "Journal of clinical orthodontics : JCO",
  ["j clin orthop trauma"] = "Journal of clinical orthopaedics and trauma",
  ["j clin outcomes manag"] = "Journal of clinical outcomes management : JCOM",
  ["j clin pathol"] = "Journal of clinical pathology",
  ["j clin pathol suppl (assoc clin pathol)"] = "Journal of clinical pathology. Supplement (Association of Clinical Pathologists)",
  ["j clin pathol suppl (r coll pathol)"] = "Journal of clinical pathology. Supplement (Royal College of Pathologists)",
  ["j clin pathol suppl coll pathol"] = "Journal of clinical pathology. Supplement (College of Pathologists)",
  ["j clin pathw"] = "Journal of clinical pathways : the foundation of value-based care",
  ["j clin pediatr dent"] = "The Journal of clinical pediatric dentistry",
  ["j clin periodontol"] = "Journal of clinical periodontology",
  ["j clin pharm ther"] = "Journal of clinical pharmacy and therapeutics",
  ["j clin pharmacol"] = "Journal of clinical pharmacology",
  ["j clin pharmacol j new drugs"] = "The Journal of clinical pharmacology and the journal of new drugs",
  ["j clin pharmacol new drugs"] = "The Journal of clinical pharmacology and new drugs",
  ["j clin prev cardiol"] = "Journal of clinical and preventive cardiology",
  ["j clin psychiatry"] = "The Journal of clinical psychiatry",
  ["j clin psychol"] = "Journal of clinical psychology",
  ["j clin psychol med settings"] = "Journal of clinical psychology in medical settings",
  ["j clin psychopharmacol"] = "Journal of clinical psychopharmacology",
  ["j clin res best pract"] = "Journal of clinical research best practices",
  ["j clin res bioeth"] = "Journal of clinical research & bioethics",
  ["j clin res drug dev"] = "Journal of clinical research and drug development",
  ["j clin res hiv aids prev"] = "Journal of clinical research in HIV AIDS and prevention",
  ["j clin res oncol"] = "Journal of clinical research in oncology",
  ["j clin res pediatr endocrinol"] = "Journal of clinical research in pediatric endocrinology",
  ["j clin res rep"] = "Journal of clinical research and reports",
  ["j clin respir med"] = "Journal of clinical respiratory medicine",
  ["j clin rheumatol"] = "Journal of clinical rheumatology : practical reports on rheumatic & musculoskeletal diseases",
  ["j clin rheumatol musculoskelet med"] = "Journal of clinical rheumatology & musculoskeletal medicine",
  ["j clin sleep med"] = "Journal of clinical sleep medicine : JCSM : official publication of the American Academy of Sleep Medicine",
  ["j clin sport psychol"] = "Journal of clinical sport psychology",
  ["j clin stomatol conf"] = "Journal of clinical stomatology conferences",
  ["j clin toxicol"] = "Journal of clinical toxicology",
  ["j clin transl endocrinol"] = "Journal of clinical & translational endocrinology",
  ["j clin transl endocrinol case rep"] = "Journal of clinical and translational endocrinology case reports",
  ["j clin transl endocrinol: case rep"] = "Journal of Clinical & Translational Endocrinology: Case Reports",
  ["j clin transl hepatol"] = "Journal of clinical and translational hepatology",
  ["j clin transl res"] = "Journal of clinical and translational research",
  ["j clin transl sci"] = "Journal of clinical and translational science",
  ["j clin trials"] = "Journal of clinical trials",
  ["j clin tuberc other mycobact dis"] = "Journal of clinical tuberculosis and other mycobacterial diseases",
  ["j clin ultrasound"] = "Journal of clinical ultrasound : JCU",
  ["j clin urol"] = "Journal of clinical urology",
  ["j clin virol"] = "Journal of clinical virology : the official publication of the Pan American Society for Clinical Virology",
  ["j clinimmunol"] = "Journal of Clinical Immunology",
  ["j cloud comput"] = "Journal of Cloud Computing",
  ["j cloud comput (heidelb)"] = "Journal of cloud computing (Heidelberg, Germany)",
  ["j clust sci"] = "Journal of cluster science",
  ["j cluster sci"] = "Journal of Cluster Science",
  ["j cmsc"] = "Journal of the CMSC",
  ["j co2 util"] = "Journal of CO2 Utilization",
  ["j coagul disord"] = "Journal of coagulation disorders",
  ["j coal sci eng"] = "Journal of Coal Science and Engineering",
  ["j coast conserv"] = "Journal of coastal conservation",
  ["j coast res"] = "Journal of coastal research",
  ["j coastal conserv"] = "Journal of Coastal Conservation",
  ["j coastal res"] = "Journal of Coastal Research",
  ["j coat technol"] = "Journal of Coatings Technology",
  ["j coat technol res"] = "Journal of Coatings Technology and Research",
  ["j cogn"] = "Journal of cognition",
  ["j cogn behav psychother"] = "Journal of cognitive and behavioral psychotherapies : the official journal of the International Institute for the Advanced Studies of Psychotherapy and Applied Mental Health",
  ["j cogn cult"] = "Journal of cognition and culture",
  ["j cogn dev"] = "Journal of cognition and development : official journal of the Cognitive Development Society",
  ["j cogn educ psychol"] = "Journal of cognitive education and psychology : JCEP",
  ["j cogn eng decis mak"] = "Journal of cognitive engineering and decision making",
  ["j cogn enhanc"] = "Journal of cognitive enhancement : towards the integration of theory and practice",
  ["j cogn neurosci"] = "Journal of cognitive neuroscience",
  ["j cogn psychol (hove)"] = "Journal of cognitive psychology (Hove, England)",
  ["j cogn psychother"] = "Journal of cognitive psychotherapy",
  ["j cogn sci (seoul)"] = "Journal of cognitive science",
  ["j cognit eng decis making"] = "Journal of Cognitive Engineering and Decision Making",
  ["j cognit enhancement"] = "Journal of Cognitive Enhancement",
  ["j cognit neurosci"] = "Journal of Cognitive Neuroscience",
  ["j cold reg eng"] = "Journal of Cold Regions Engineering",
  ["j coll character"] = "Journal of college and character",
  ["j coll gen pract"] = "The Journal of the College of General Practitioners",
  ["j coll gen pract can"] = "Journal - College of General Practice of Canada",
  ["j coll gen pract res newsl"] = "Journal of the College of General Practitioners and research newsletter. College of General Practitioners (Great Britain)",
  ["j coll physicians surg pak"] = "Journal of the College of Physicians and Surgeons--Pakistan : JCPSP",
  ["j coll radiol australas"] = "Journal of the College of Radiologists of Australasia",
  ["j coll sci teach"] = "Journal of college science teaching",
  ["j coll stud dev"] = "Journal of college student development",
  ["j coll stud ret"] = "Journal of college student retention : research, theory & practice",
  ["j coll teach learn"] = "Journal of college teaching and learning",
  ["j coll univ law"] = "The Journal of college and university law",
  ["j collect negotiations public sect"] = "Journal of collective negotiations in the public sector",
  ["j college stud psychother"] = "Journal of college student psychotherapy",
  ["j colloid interface sci"] = "Journal of Colloid and Interface Science",
  ["j colloid sci"] = "Journal of colloid science",
  ["j colo dent assoc"] = "Journal of the Colorado Dental Association",
  ["j colo state dent assoc"] = "Journal. Colorado State Dental Association",
  ["j colon colon hist"] = "Journal of colonialism & colonial history",
  ["j comb"] = "Journal of Combinatorics",
  ["j comb algebra"] = "Journal of Combinatorial Algebra",
  ["j comb chem"] = "Journal of Combinatorial Chemistry",
  ["j comb des"] = "Journal of Combinatorial Designs",
  ["j comb inf syst sci"] = "Journal of Combinatorics, Information & System Sciences",
  ["j comb number theory"] = "Journal of Combinatorics and Number Theory",
  ["j comb optim"] = "Journal of Combinatorial Optimization",
  ["j comb theory ser a"] = "Journal of combinatorial theory. Series A",
  ["j comb theory ser b"] = "Journal of Combinatorial Theory Series B",
  ["j combin des"] = "Journal of Combinatorial Designs",
  ["j combin inform system sci"] = "Journal of Combinatorics, Information & System Sciences",
  ["j combin math combin comput"] = "Journal of Combinatorial Mathematics and Combinatorial Computing",
  ["j combin theory ser a"] = "Journal of Combinatorial Theory. Series A",
  ["j combin theory ser b"] = "Journal of Combinatorial Theory. Series B",
  ["j combust"] = "Journal of Combustion",
  ["j combust soc jpn"] = "Journal of the Combustion Society of Japan",
  ["j commer biotechnol"] = "Journal of commercial biotechnology",
  ["j commod sci technol qual"] = "Journal of Commodity Science, Technology and Quality",
  ["j common mark stud"] = "Journal of common market studies",
  ["j common market stud"] = "Journal of Common Market Studies",
  ["j commonw comp polit"] = "The Journal of Commonwealth & comparative politics",
  ["j commonw lit"] = "The journal of Commonwealth literature",
  ["j commonw polit stud"] = "Journal of Commonwealth political studies",
  ["j commun"] = "The Journal of communication",
  ["j commun dis"] = "The Journal of communicable diseases",
  ["j commun disord"] = "Journal of communication disorders",
  ["j commun healthc"] = "Journal of communication in healthcare",
  ["j commun inf networks"] = "Journal of Communications and Information Networks",
  ["j commun inq"] = "The Journal of communication inquiry",
  ["j commun networks"] = "Journal of Communications and Networks",
  ["j commun technol electron"] = "Journal of Communications Technology and Electronics",
  ["j community appl soc psychol"] = "Journal of community & applied social psychology",
  ["j community dev soc"] = "Journal of the Community Development Society",
  ["j community engagem scholarsh"] = "Journal of community engagement and scholarship",
  ["j community genet"] = "Journal of community genetics",
  ["j community health"] = "Journal of community health",
  ["j community health nurs"] = "Journal of community health nursing",
  ["j community health res"] = "Journal of community health research",
  ["j community hosp intern med perspect"] = "Journal of community hospital internal medicine perspectives",
  ["j community med (reno)"] = "Journal of community medicine (Reno, Nev.)",
  ["j community med health educ"] = "Journal of community medicine & health education",
  ["j community med public health"] = "Journal of community medicine & public health",
  ["j community pract"] = "Journal of community practice",
  ["j community psychol"] = "Journal of community psychology",
  ["j community public health nurs"] = "Journal of community & public health nursing",
  ["j community support oncol"] = "The Journal of community and supportive oncology",
  ["j commut algebra"] = "Journal of Commutative Algebra",
  ["j comorb"] = "Journal of comorbidity",
  ["j comp econ"] = "Journal of comparative economics",
  ["j comp eff res"] = "Journal of comparative effectiveness research",
  ["j comp fam stud"] = "Journal of comparative family studies",
  ["j comp neurol"] = "The Journal of comparative neurology",
  ["j comp pathol"] = "Journal of comparative pathology",
  ["j comp pathol ther"] = "The Journal of comparative pathology and therapeutics",
  ["j comp physiol"] = "Journal of comparative physiology",
  ["j comp physiol [a]"] = "Journal of Comparative Physiology. A, Sensory, Neural, and Behavioral Physiology",
  ["j comp physiol [b]"] = "Journal of Comparative Physiology. B, Biochemical, Systemic, and Environmental Physiology",
  ["j comp physiol a"] = "Journal of comparative physiology. A, Sensory, neural, and behavioral physiology",
  ["j comp physiol a neuroethol sens neural behav physiol"] = "Journal of comparative physiology. A, Neuroethology, sensory, neural, and behavioral physiology",
  ["j comp physiol b"] = "Journal of comparative physiology. B, Biochemical, systemic, and environmental physiology",
  ["j comp physiol psychol"] = "Journal of comparative and physiological psychology",
  ["j comp psychol"] = "Journal of comparative psychology (Washington, D.C. : 1983)",
  ["j compar econ"] = "Journal of Comparative Economics",
  ["j compassionate health care"] = "Journal of compassionate health care",
  ["j complement integr med"] = "Journal of complementary & integrative medicine",
  ["j complex"] = "Journal of complexity",
  ["j complex anal"] = "Journal of Complex Analysis",
  ["j complex netw"] = "Journal of complex networks",
  ["j complex networks"] = "Journal of Complex Networks",
  ["j complexity"] = "Journal of Complexity",
  ["j compliance health care"] = "The Journal of compliance in health care : JCHC",
  ["j compos"] = "Journal of composites",
  ["j compos constr"] = "Journal of Composites for Construction",
  ["j compos mater"] = "Journal of Composite Materials",
  ["j compos sci"] = "Journal Of Composites Science",
  ["j compos tech res"] = "Journal of Composites Technology and Research",
  ["j comput"] = "Journal of computing",
  ["j comput (taipei)"] = "Journal of computers",
  ["j comput acoust"] = "Journal of Computational Acoustics",
  ["j comput aided mol des"] = "Journal of computer-aided molecular design",
  ["j comput algebra"] = "Journal of Computational Algebra",
  ["j comput anal appl"] = "Journal of Computational Analysis and Applications",
  ["j comput appl archaeol"] = "Journal of Computer Applications in Archaeology",
  ["j comput appl math"] = "Journal of Computational and Applied Mathematics",
  ["j comput assist learn"] = "Journal of computer assisted learning",
  ["j comput assist microsc"] = "Journal of computer-assisted microscopy",
  ["j comput assist tomogr"] = "Journal of computer assisted tomography",
  ["j comput assisted learn"] = "Journal of Computer Assisted Learning",
  ["j comput assisted tomogr"] = "Journal of Computer Assisted Tomography",
  ["j comput biol"] = "Journal of Computational Biology",
  ["j comput biol bioinform res"] = "Journal of computational biology and bioinformatics research",
  ["j comput biophys chem"] = "Journal of Computational Biophysics and Chemistry",
  ["j comput chem"] = "Journal of computational chemistry",
  ["j comput chem jpn"] = "Journal of Computer Chemistry, Japan",
  ["j comput chem, jpn"] = "Journal of Computer Chemistry, Japan",
  ["j comput civ eng"] = "Journal of Computing in Civil Engineering",
  ["j comput civil eng"] = "Journal of Computing in Civil Engineering",
  ["j comput des eng"] = "Journal of Computational Design and Engineering",
  ["j comput dyn"] = "Journal of Computational Dynamics",
  ["j comput electron"] = "Journal of computational electronics",
  ["j comput eng math"] = "Journal of Computational and Engineering Mathematics",
  ["j comput finance"] = "Journal of Computational Finance",
  ["j comput geom"] = "Journal of Computational Geometry",
  ["j comput graph stat"] = "Journal of computational and graphical statistics : a joint publication of American Statistical Association, Institute of Mathematical Statistics, Interface Foundation of North America",
  ["j comput graph statist"] = "Journal of Computational and Graphical Statistics",
  ["j comput graphical stat"] = "Journal of Computational and Graphical Statistics",
  ["j comput high educ"] = "Journal of computing in higher education",
  ["j comput inf"] = "Journal of Computing and Information",
  ["j comput inf sci eng"] = "Journal of Computing and Information Science in Engineering",
  ["j comput inf syst"] = "Journal of Computer Information Systems",
  ["j comput inform"] = "Journal of Computing and Information",
  ["j comput interdiscip sci"] = "Journal of computational interdisciplinary sciences",
  ["j comput lang"] = "Journal of computer languages",
  ["j comput math"] = "Journal of Computational Mathematics",
  ["j comput math data sci"] = "Journal of Computational Mathematics and Data Science",
  ["j comput mediat commun"] = "Journal of computer-mediated communication : JCMC",
  ["j comput methods appl math"] = "Journal of computational methods in applied mathematics",
  ["j comput methods mol des"] = "Journal of Computational Methods in Molecular Design",
  ["j comput methods sci eng"] = "Journal of Computational Methods in Science and Engineering",
  ["j comput multiph flows"] = "Journal of Computational Multiphase Flows",
  ["j comput multiphase flows"] = "Journal of Computational Multiphase Flows",
  ["j comput networks commun"] = "Journal of Computer Networks and Communications",
  ["j comput neurosci"] = "Journal of computational neuroscience",
  ["j comput nonlinear dyn"] = "Journal of Computational and Nonlinear Dynamics",
  ["j comput phys"] = "Journal of Computational Physics",
  ["j comput phys x"] = "Journal of Computational Physics: X",
  ["j comput phys: x"] = "Journal of Computational Physics X",
  ["j comput sci"] = "Journal of Computational Science",
  ["j comput sci appl math"] = "Journal of Computer Science and Applied Mathematics",
  ["j comput sci eng"] = "Journal of Computing Science and Engineering",
  ["j comput sci syst biol"] = "Journal of computer science and systems biology",
  ["j comput sci tech"] = "Journal of Computer Science and Technology",
  ["j comput sci technol"] = "Journal of computer science and technology",
  ["j comput sci technol argent"] = "Journal of Computer Science And Technology Argentina",
  ["j comput secur"] = "Journal of computer security",
  ["j comput soc sci"] = "Journal of computational social science",
  ["j comput social sci"] = "Journal of Computational Social Science",
  ["j comput surg"] = "Journal of computational surgery",
  ["j comput syst sci"] = "Journal of computer and system sciences",
  ["j comput syst sci int"] = "Journal of Computer and System Sciences International",
  ["j comput system sci"] = "Journal of Computer and System Sciences",
  ["j comput theor nanosci"] = "Journal of Computational and Theoretical Nanoscience",
  ["j comput theor transp"] = "Journal of Computational and Theoretical Transport",
  ["j comput tomogr"] = "The Journal of computed tomography",
  ["j comput virol hacking tech"] = "Journal of Computer Virology and Hacking Techniques",
  ["j comput vis image process"] = "Journal of computer vision and image processing",
  ["j comput-aided mater des"] = "Journal of Computer-Aided Materials Design",
  ["j comput-aided mol des"] = "Journal of Computer-Aided Molecular Design",
  ["j comput-assisted microsc"] = "Journal of Computer-Assisted Microscopy",
  ["j comput-mediated commun"] = "Journal of Computer-Mediated Communication",
  ["j conchol"] = "Journal of conchology",
  ["j concr appl math"] = "Journal of Concrete and Applicable Mathematics",
  ["j concussion"] = "Journal of concussion",
  ["j conf workshop"] = "Journalism Conference And Workshop",
  ["j conflict resolut"] = "The Journal of conflict resolution",
  ["j conflict resolution"] = "Journal of Conflict Resolution",
  ["j conn state dent assoc"] = "Journal - Connecticut State Dental Association",
  ["j cons aff"] = "Journal of Consumer Affairs",
  ["j cons pol"] = "Journal of Consumer Policy",
  ["j cons res"] = "Journal of Consumer Research",
  ["j conscious stud"] = "Journal of consciousness studies : controversies in science & the humanities",
  ["j conserv dent"] = "Journal of conservative dentistry : JCD",
  ["j conserv mus stud"] = "Journal of conservation & museum studies",
  ["j constr eng"] = "Journal of Construction Engineering",
  ["j constr eng manag"] = "Journal of construction engineering and management",
  ["j constr eng manage"] = "Journal of Construction Engineering Management",
  ["j constr psychol"] = "Journal of constructivist psychology",
  ["j constr res"] = "Journal of Construction Research",
  ["j constr steel res"] = "Journal of Constructional Steel Research",
  ["j consult clin psychol"] = "Journal of consulting and clinical psychology",
  ["j consult psychol"] = "Journal of consulting psychology",
  ["j consum aff"] = "The Journal of consumer affairs",
  ["j consum health internet"] = "Journal of consumer health on the Internet",
  ["j consum mark"] = "The journal of consumer marketing",
  ["j consum policy (dordr)"] = "Journal of consumer policy",
  ["j consum prot, food saf"] = "Journal of Consumer Protection and Food Safety",
  ["j consum psychol"] = "Journal of consumer psychology : the official journal of the Society for Consumer Psychology",
  ["j consum res"] = "The Journal of consumer research",
  ["j contam hydrol"] = "Journal of contaminant hydrology",
  ["j contemp afr stud"] = "Journal of contemporary African studies : JCAS",
  ["j contemp asia"] = "Journal of contemporary Asia",
  ["j contemp brachytherapy"] = "Journal of contemporary brachytherapy",
  ["j contemp china"] = "The journal of contemporary China = Dang dai Zhongguo",
  ["j contemp crim justice"] = "Journal of contemporary criminal justice",
  ["j contemp dent pract"] = "The journal of contemporary dental practice",
  ["j contemp diabetes res"] = "Journal of contemporary diabetes research",
  ["j contemp ethnogr"] = "Journal of contemporary ethnography",
  ["j contemp health law policy"] = "The Journal of contemporary health law and policy",
  ["j contemp hist"] = "Journal of contemporary history",
  ["j contemp law"] = "Journal of contemporary law",
  ["j contemp math anal"] = "Journal of Contemporary Mathematical Analysis",
  ["j contemp med educ"] = "Journal of contemporary medical education",
  ["j contemp phys"] = "Journal of Contemporary Physics",
  ["j contemp psychother"] = "Journal of contemporary psychotherapy",
  ["j contemp water res educ"] = "Journal of contemporary water research & education",
  ["j contextual behav sci"] = "Journal of contextual behavioral science",
  ["j contin educ health prof"] = "The Journal of continuing education in the health professions",
  ["j contin educ nurs"] = "Journal of continuing education in nursing",
  ["j contra costa dent soc"] = "Journal / Contra Costa Dental Society",
  ["j contracept"] = "The Journal of contraception",
  ["j contracept stud"] = "Journal of contraceptive studies",
  ["j control autom electr syst"] = "Journal of Control, Automation and Electrical Systems",
  ["j control decis"] = "Journal of Control and Decision",
  ["j control release"] = "Journal of controlled release : official journal of the Controlled Release Society",
  ["j control sci eng"] = "Journal of Control Science and Engineering",
  ["j controlled release"] = "Journal of Controlled Release",
  ["j convex anal"] = "Journal of Convex Analysis",
  ["j coord chem"] = "Journal of Coordination Chemistry",
  ["j corp finan: contracting"] = "Journal of Corporate Finance: Contracting, Governance and Organization",
  ["j corp law"] = "The Journal of corporation law",
  ["j correct educ (glen mills)"] = "Journal of correctional education (Glen Mills, Pa.)",
  ["j correct health care"] = "Journal of correctional health care : the official journal of the National Commission on Correctional Health Care",
  ["j corros sci eng"] = "Journal of Corrosion Science and Engineering",
  ["j cosmet dermatol"] = "Journal of cosmetic dermatology",
  ["j cosmet laser ther"] = "Journal of cosmetic and laser therapy : official publication of the European Society for Laser Dermatology",
  ["j cosmet sci"] = "Journal of cosmetic science",
  ["j cosmol astropart phys"] = "Journal of Cosmology and Astroparticle Physics",
  ["j cosmology & astro-part phys"] = "Journal of Cosmology and Astro-Particle Physics",
  ["j costmet laser ther"] = "Journal of Cosmetic and Laser Therapy",
  ["j cotton res"] = "Journal of Cotton Research",
  ["j cotton sci"] = "Journal of cotton science",
  ["j counc sci ind res"] = "Journal. Council for Scientific and Industrial Research (Australia)",
  ["j couns dev"] = "Journal of counseling and development : JCD",
  ["j couns psychol"] = "Journal of counseling psychology",
  ["j couns sexol sex wellness"] = "Journal of counseling sexology & sexual wellness : research, practice, and education",
  ["j couple relatsh ther"] = "Journal of couple & relationship therapy",
  ["j coupled syst multiscale dyn"] = "Journal of coupled systems and multiscale dynamics",
  ["j craniofac genet dev biol"] = "Journal of craniofacial genetics and developmental biology",
  ["j craniofac genet dev biol suppl"] = "Journal of craniofacial genetics and developmental biology. Supplement",
  ["j craniofac surg"] = "The Journal of craniofacial surgery",
  ["j craniomandib disord"] = "Journal of craniomandibular disorders : facial & oral pain",
  ["j craniomandibular pract"] = "The Journal of cranio-mandibular practice",
  ["j craniomaxillofac surg"] = "Journal of cranio-maxillo-facial surgery : official publication of the European Association for Cranio-Maxillo-Facial Surgery",
  ["j craniomaxillofac trauma"] = "The Journal of cranio-maxillofacial trauma",
  ["j craniovertebr junction spine"] = "Journal of craniovertebral junction & spine",
  ["j crash prev inj control"] = "Journal of Crash Prevention and Injury Control",
  ["j creat behav"] = "The Journal of creative behavior",
  ["j creat ment health"] = "Journal of creativity in mental health",
  ["j crim justice"] = "Journal of criminal justice",
  ["j crim law criminol"] = "The Journal of criminal law & criminology",
  ["j crim law criminol (1931)"] = "The Journal of criminal law and criminology, including the American journal of police science",
  ["j crim law criminol police sci"] = "The Journal of criminal law, criminology, and police science",
  ["j crim psychol"] = "Journal of criminal psychology",
  ["j crime justice"] = "Journal of crime and justice",
  ["j crit care"] = "Journal of critical care",
  ["j crit care med"] = "Journal of critical care medicine",
  ["j crit care med (targu mures)"] = "Journal of critical care medicine (Universitatea de Medicină și Farmacie din Târgu-Mureș)",
  ["j crit illn"] = "The Journal of critical illness",
  ["j crit realism"] = "Journal of critical realism",
  ["j crohns colitis"] = "Journal of Crohn's & colitis",
  ["j crop improv"] = "Journal of crop improvement",
  ["j crop prot"] = "Journal of crop protection",
  ["j crop sci biotechnol"] = "Journal of Crop Science and Biotechnology",
  ["j cross cult gerontol"] = "Journal of cross-cultural gerontology",
  ["j cross cult psychol"] = "Journal of cross-cultural psychology",
  ["j crustac biol"] = "Journal of crustacean biology : a quarterly of the Crustacean Society for the publication of research on any aspect of the biology of crustacea",
  ["j crustacean biol"] = "Journal of Crustacean Biology",
  ["j cryptogr"] = "Journal of Cryptography",
  ["j cryptogr eng"] = "Journal of cryptographic engineering",
  ["j cryptology"] = "Journal of Cryptology. The Journal of the International Association for Cryptologic Research",
  ["j cryst growth"] = "Journal of Crystal Growth",
  ["j cult cogn sci"] = "Journal of cultural cognitive science",
  ["j cult divers"] = "Journal of cultural diversity",
  ["j cult econ"] = "Journal of cultural economy",
  ["j cult evol psychol"] = "Journal of cultural and evolutionary psychology",
  ["j cult herit"] = "Journal of cultural heritage",
  ["j cult heritage"] = "Journal of Cultural Heritage",
  ["j cult heritage manage sustainable dev"] = "Journal of Cultural Heritage Management and Sustainable Development",
  ["j curr adolesc med"] = "The Journal of current adolescent medicine",
  ["j curr chin aff"] = "Journal of current Chinese affairs",
  ["j curr glaucoma pract"] = "Journal of current glaucoma practice",
  ["j curr ophthalmol"] = "Journal of current ophthalmology",
  ["j curr sci technol"] = "Journal of current science and technology",
  ["j curr soc issues"] = "Journal of current social issues",
  ["j curr surg"] = "Journal of current surgery",
  ["j cutan aesthet surg"] = "Journal of cutaneous and aesthetic surgery",
  ["j cutan laser ther"] = "Journal of cutaneous laser therapy",
  ["j cutan med surg"] = "Journal of cutaneous medicine and surgery",
  ["j cutan pathol"] = "Journal of cutaneous pathology",
  ["j cyber secur"] = "Journal of Cyber Security",
  ["j cyber secur technol"] = "Journal of Cyber Security Technology",
  ["j cyber ther rehabil"] = "Journal of cyber therapy and rehabilitation",
  ["j cybersecur"] = "Journal of cybersecurity",
  ["j cyclic nucleotide protein phosphor res"] = "Journal of cyclic nucleotide and protein phosphorylation research",
  ["j cyclic nucleotide res"] = "Journal of cyclic nucleotide research",
  ["j cyst fibros"] = "Journal of cystic fibrosis : official journal of the European Cystic Fibrosis Society",
  ["j cytokine biol"] = "Journal of cytokine biology",
  ["j cytol"] = "Journal of cytology",
  ["j cytol histol"] = "Journal of cytology & histology",
  ["j cytol mol biol"] = "Journal of cytology & molecular biology",
  ["j czech geol soc"] = "Journal of the Czech Geological Society",
  ["j dairy foods home sci"] = "Journal of dairying, foods & home sciences",
  ["j dairy res"] = "Journal of Dairy Research",
  ["j dairy sci"] = "Journal of Dairy Science",
  ["j dairy sci biotechnol"] = "Journal of dairy science and biotechnology",
  ["j dalian univ tech"] = "Journal of Dalian University of Technology",
  ["j dalian univ technol"] = "Journal of Dalian University of Technology. Dalian Ligon Daxue Xuebao",
  ["j dance med sci"] = "Journal of dance medicine & science : official publication of the International Association for Dance Medicine & Science",
  ["j dance somat pract"] = "Journal of dance & somatic practices",
  ["j data dyn syst"] = "Journal of Data and Dynamic Systems",
  ["j data inf manage"] = "Journal of Data, Information and Management",
  ["j data inf qual"] = "Journal of Data and Information Quality",
  ["j data inf sci"] = "Journal of data and information science (Warsaw, Poland)",
  ["j data mining genomics proteomics"] = "Journal of data mining in genomics & proteomics",
  ["j data sci"] = "Journal of data science : JDS",
  ["j data semant"] = "Journal on data semantics",
  ["j deaf stud deaf educ"] = "Journal of deaf studies and deaf education",
  ["j decis syst"] = "Journal of decision systems",
  ["j deep found inst"] = "Journal of the Deep Foundations Institute",
  ["j def anal logist"] = "Journal of Defense Analytics and Logistics",
  ["j def model simul"] = "Journal of Defense Modeling and Simulation",
  ["j degraded min lands manage"] = "Journal of Degraded and Mining Lands Management",
  ["j demogr economics"] = "Journal of demographic economics",
  ["j dent"] = "Journal of dentistry",
  ["j dent (shiraz)"] = "Journal of dentistry (Shīrāz, Iran)",
  ["j dent (tehran)"] = "Journal of dentistry (Tehran, Iran)",
  ["j dent anesth pain med"] = "Journal of dental anesthesia and pain medicine",
  ["j dent assoc s afr"] = "The Journal of the Dental Association of South Africa = Die Tydskrif van die Tandheelkundige Vereniging van Suid-Afrika",
  ["j dent assoc thai"] = "The Journal of the Dental Association of Thailand",
  ["j dent aux"] = "The Journal of the dental auxiliaries, Malaya, Malaysia",
  ["j dent belge"] = "Le Journal dentaire belge. Belgisch blad voor tandheelkunde",
  ["j dent biomater"] = "Journal of dental biomaterials",
  ["j dent biomech"] = "Journal of dental biomechanics",
  ["j dent child"] = "Journal of dentistry for children",
  ["j dent child (chic)"] = "Journal of dentistry for children (Chicago, Ill.)",
  ["j dent craniofac res"] = "Journal of dental and craniofacial research",
  ["j dent dent med"] = "Journal of dentistry and dental medicine",
  ["j dent educ"] = "Journal of dental education",
  ["j dent guid counc handicap"] = "Journal of the Dental Guidance Council on the Handicapped",
  ["j dent handicap"] = "Journal of dentistry for the handicapped",
  ["j dent health oral disord ther"] = "Journal of dental health, oral disorders & therapy",
  ["j dent hyg"] = "Journal of dental hygiene : JDH",
  ["j dent maxillofac surg"] = "Journal of dental and maxillofacial surgery",
  ["j dent med"] = "The Journal of dental medicine",
  ["j dent oral biol"] = "Journal of dentistry and oral biology",
  ["j dent oral craniofac epidemiol"] = "Journal of dental, oral and craniofacial epidemiology",
  ["j dent oral hyg"] = "Journal of dentistry and oral hygiene",
  ["j dent oral sci"] = "Journal of dentistry and oral sciences",
  ["j dent pract adm"] = "Journal of dental practice administration : JDPA : official publication of American Academy of Dental Practice Administration, Organization of Teachers of Dental Practice Administration, American Academy of Dental Group Practice",
  ["j dent que"] = "Le Journal dentaire du Québec",
  ["j dent res"] = "Journal of dental research",
  ["j dent res dent clin dent prospects"] = "Journal of dental research, dental clinics, dental prospects",
  ["j dent sch natl univ iran"] = "Journal of the Dental School, National University of Iran",
  ["j dent sci"] = "Journal of dental sciences",
  ["j dent sleep med"] = "Journal of dental sleep medicine",
  ["j dent soc state  n y"] = "The Journal of the Dental Society of the State of New York",
  ["j dent symp"] = "The Journal of dental symposia",
  ["j dent technol"] = "Journal of dental technology : the peer-reviewed publication of the National Association of Dental Laboratories",
  ["j depress anxiety"] = "Journal of depression & anxiety",
  ["j derivatives"] = "Journal of Derivatives",
  ["j dermatol"] = "The Journal of dermatology",
  ["j dermatol case rep"] = "Journal of dermatological case reports",
  ["j dermatol nurses assoc"] = "Journal of the Dermatology Nurses' Association",
  ["j dermatol sci"] = "Journal of dermatological science",
  ["j dermatol surg"] = "The Journal of dermatologic surgery",
  ["j dermatol surg oncol"] = "The Journal of dermatologic surgery and oncology",
  ["j dermatolog treat"] = "The Journal of dermatological treatment",
  ["j des hist"] = "Journal of design history",
  ["j des manuf autom"] = "Journal of Design and Manufacturing Automation",
  ["j des sci"] = "Journal of Design and Science",
  ["j dev admin stud"] = "The Journal of development and administrative studies",
  ["j dev agric econ"] = "Journal of development and agricultural economics",
  ["j dev areas"] = "Journal Of Developing Areas",
  ["j dev behav pediatr"] = "Journal of developmental and behavioral pediatrics : JDBP",
  ["j dev biol"] = "Journal of developmental biology",
  ["j dev biol regen med"] = "Journal of developmental biology & regenerative medicine",
  ["j dev biol tissue eng"] = "Journal of developmental biology and tissue engineering",
  ["j dev comm"] = "The Journal of development communication",
  ["j dev disabl"] = "Journal on developmental disabilities = Le journal sur les handicaps du dévelopment [sic.]",
  ["j dev drugs"] = "Journal of developing drugs",
  ["j dev econ"] = "Journal of development economics",
  ["j dev effect"] = "Journal of development effectiveness",
  ["j dev entrep"] = "Journal of developmental entrepreneurship : JDE",
  ["j dev leadersh"] = "Journal for development and leadership. Nelson Mandela Metropolitan University",
  ["j dev learn disord"] = "The journal of developmental and learning disorders",
  ["j dev life course criminol"] = "Journal of developmental and life-course criminology",
  ["j dev orig health dis"] = "Journal of developmental origins of health and disease",
  ["j dev origins health dis"] = "Journal of Developmental Origins of Health and Disease",
  ["j dev phys disabil"] = "Journal of developmental and physical disabilities",
  ["j dev physiol"] = "Journal of developmental physiology",
  ["j dev process"] = "The journal of developmental processes",
  ["j dev soc"] = "Journal of developing societies",
  ["j dev stud"] = "The journal of development studies",
  ["j devel econ"] = "Journal of Development Economics",
  ["j devel econ pol"] = "Journal of Development and Economic Policies",
  ["j devel stud"] = "Journal of Development Studies",
  ["j developing areas"] = "Journal of Developing Areas",
  ["j diabet complications"] = "The Journal of diabetic complications",
  ["j diabetes"] = "Journal of diabetes",
  ["j diabetes clin res"] = "Journal of diabetes and clinical research",
  ["j diabetes complications"] = "Journal of diabetes and its complications",
  ["j diabetes investig"] = "Journal of diabetes investigation",
  ["j diabetes mellit metab syndr"] = "Journal of diabetes mellitus and metabolic syndrome",
  ["j diabetes mellitus"] = "Journal of diabetes mellitus",
  ["j diabetes metab"] = "Journal of diabetes & metabolism",
  ["j diabetes metab disord"] = "Journal of diabetes and metabolic disorders",
  ["j diabetes metab disord control"] = "Journal of diabetes, metabolic disorders & control",
  ["j diabetes nurs"] = "Journal of diabetes nursing",
  ["j diabetes obes"] = "Journal of diabetes and obesity",
  ["j diabetes relat disord"] = "Journal of diabetes and related disorders",
  ["j diabetes res"] = "Journal of diabetes research",
  ["j diabetes res clin metab"] = "Journal of diabetes research & clinical metabolism",
  ["j diabetes res ther"] = "Journal of diabetes research and therapy",
  ["j diabetes sci technol"] = "Journal of diabetes science and technology",
  ["j diabetes treat"] = "Journal of diabetes and treatment",
  ["j diagn imaging ther"] = "Journal of diagnostic imaging in therapy",
  ["j diagn med sonogr"] = "Journal of diagnostic medical sonography : JDMS",
  ["j diagn radiogr imaging"] = "Journal of Diagnostic Radiography and Imaging",
  ["j diagn tech biomed anal"] = "Journal of diagnostic techniques and biomedical analysis",
  ["j dial"] = "Journal of dialysis",
  ["j diarrhoeal dis res"] = "Journal of diarrhoeal diseases research",
  ["j diet suppl"] = "Journal of dietary supplements",
  ["j differ equ"] = "Journal of differential equations",
  ["j differ equations"] = "Journal of Differential Equations",
  ["j differ equations appl"] = "Journal of Difference Equations and Applications",
  ["j differ geom"] = "Journal of Differential Geometry",
  ["j difference equ appl"] = "Journal of Difference Equations and Applications",
  ["j differential equations"] = "Journal of Differential Equations",
  ["j differential geom"] = "Journal of Differential Geometry",
  ["j dig cancer rep"] = "Journal of digestive cancer reports",
  ["j dig dis"] = "Journal of digestive diseases",
  ["j dig dis hepatol"] = "Journal of digestive diseases and hepatology",
  ["j digit imaging"] = "Journal of digital imaging",
  ["j digit landsc archit"] = "Journal of digital landscape architecture : JoDLA",
  ["j digit media manag"] = "Journal of digital media management",
  ["j digital imaging"] = "Journal of Digital Imaging",
  ["j dis glob health"] = "Journal of disease and global health",
  ["j disabil policy stud"] = "Journal of disability policy studies",
  ["j discrete algorithms"] = "Journal of Discrete Algorithms",
  ["j discrete algorithms (amst)"] = "Journal of discrete algorithms (Amsterdam, Netherlands)",
  ["j discrete algoritms"] = "Journal of Discrete Algorithms",
  ["j discrete math sci cryptogr"] = "Journal of Discrete Mathematical Sciences and Cryptography",
  ["j discrete math sci cryptography"] = "Journal of Discrete Mathematical Sciences & Cryptography",
  ["j disp technol"] = "Journal of Display Technology",
  ["j dispers sci technol"] = "Journal of dispersion science and technology",
  ["j dispersion sci technol"] = "Journal of Dispersion Science and Technology",
  ["j dist columbia dent soc"] = "District of Columbia Dental Society journal. District of Columbia Dental Society",
  ["j divers high educ"] = "Journal of diversity in higher education",
  ["j divorce"] = "Journal of divorce",
  ["j divorce & remarriage"] = "Journal of divorce & remarriage",
  ["j divorce remarriage"] = "Journal of divorce & remarriage",
  ["j doc"] = "The Journal of documentation; devoted to the recording, organization and dissemination of specialized knowledge",
  ["j dow univ health sci"] = "Journal of the Dow University of Health Sciences",
  ["j down syndr chromosom abnorm"] = "Journal of Down Syndrome & chromosome abnormalities",
  ["j dr ntr univ health sci"] = "Journal of Dr. NTR University of Health Sciences",
  ["j dr nurs pract"] = "Journal of doctoral nursing practice",
  ["j drug abuse"] = "Journal of drug abuse",
  ["j drug addict educ erad"] = "Journal of drug addiction, education, and eradication",
  ["j drug alcohol res"] = "Journal of drug and alcohol research",
  ["j drug assess"] = "Journal of drug assessment",
  ["j drug deliv"] = "Journal of drug delivery",
  ["j drug deliv sci technol"] = "Journal of drug delivery science and technology",
  ["j drug delivery"] = "Journal of Drug Delivery",
  ["j drug delivery sci technol"] = "Journal of Drug Delivery Science and Technology",
  ["j drug des res"] = "Journal of drug design and research",
  ["j drug educ"] = "Journal of drug education",
  ["j drug issues"] = "Journal of drug issues",
  ["j drug metab toxicol"] = "Journal of drug metabolism & toxicology",
  ["j drug policy anal"] = "Journal of drug policy analysis",
  ["j drug res"] = "Journal of drug research",
  ["j drug target"] = "Journal of drug targeting",
  ["j drug targeting"] = "Journal of Drug Targeting",
  ["j drugs dermatol"] = "Journal of drugs in dermatology : JDD",
  ["j dtsch dermatol ges"] = "Journal der Deutschen Dermatologischen Gesellschaft = Journal of the German Society of Dermatology : JDDG",
  ["j dual diagn"] = "Journal of dual diagnosis",
  ["j dyn behav mater"] = "Journal of Dynamic Behavior of Materials",
  ["j dyn control syst"] = "Journal of Dynamical and Control Systems",
  ["j dyn differ equ"] = "Journal of dynamics and differential equations",
  ["j dyn differ equations"] = "Journal of Dynamics and Differential Equations",
  ["j dyn games"] = "Journal of Dynamics and Games",
  ["j dyn syst geom theor"] = "Journal of Dynamical Systems and Geometric Theories",
  ["j dyn syst geom theories"] = "Journal of Dynamical Systems and Geometric Theories",
  ["j dyn syst meas contr"] = "Journal of Dynamic Systems, Measurement, and Control",
  ["j dyn syst meas control"] = "Journal of dynamic systems, measurement, and control",
  ["j dynam control systems"] = "Journal of Dynamical and Control Systems",
  ["j dynam differential equations"] = "Journal of Dynamics and Differential Equations",
  ["j early adolesc"] = "The Journal of early adolescence",
  ["j early child infant psychol"] = "Journal of early childhood and infant psychology",
  ["j early child res"] = "Journal of early childhood research : ECR",
  ["j early child teach educ"] = "Journal of early childhood teacher education",
  ["j early hear detect interv"] = "Journal of early hearing detection and intervention",
  ["j early intensive behav interv"] = "Journal of early and intensive behavior intervention : JEIBI",
  ["j early interv"] = "Journal of early intervention",
  ["j early mod hist"] = "Journal of early modern history",
  ["j early mod stud (bucur)"] = "Journal of early modern studies",
  ["j early repub"] = "Journal of the early Republic",
  ["j earth environ sci"] = "Journal of earth and environmental sciences",
  ["j earth sci"] = "Journal of Earth Science",
  ["j earth sci clim change"] = "Journal of earth science & climatic change",
  ["j earth space phys"] = "Journal of the Earth and Space Physics",
  ["j earth syst sci"] = "Journal of Earth System Science",
  ["j earthquake eng"] = "Journal of Earthquake Engineering",
  ["j earthquake tsunami"] = "Journal of Earthquake and Tsunami",
  ["j east afr nat hist"] = "Journal of East African natural history",
  ["j east afr res dev"] = "Journal of Eastern African research & development",
  ["j east afr stud"] = "Journal of Eastern African studies : the journal of the British Institute in Eastern Africa",
  ["j east asian aff"] = "The Journal of East Asian affairs",
  ["j east asian ling"] = "Journal of East Asian linguistics",
  ["j east china norm univ natur sci ed"] = "Journal of East China Normal University. Natural Science Edition. Huadong Shifan Daxue Xuebao. Ziran Kexue Ban",
  ["j east west stud"] = "Journal of East and West studies",
  ["j east-west bus"] = "Journal of East-West Business",
  ["j eat disord"] = "Journal of eating disorders",
  ["j ec polytech math"] = "Journal de l’Ecole polytechnique. Mathematiques",
  ["j ecclesiast hist"] = "The Journal of ecclesiastical history",
  ["j echocardiogr"] = "Journal of echocardiography",
  ["j ecohydraul"] = "Journal of Ecohydraulics",
  ["j ecol"] = "The Journal of ecology",
  ["j ecol environ"] = "Journal of ecology and environment",
  ["j ecol health environ"] = "Journal of ecology of health & environment",
  ["j econ (mvea)"] = "Journal of Economics (MVEA)",
  ["j econ (vienna)"] = "Journal of economics (Vienna, Austria : Online)",
  ["j econ (z nationalökon)"] = "Journal of Economics (Zeitschrift für Nationalökonomie)",
  ["j econ ageing"] = "Journal of the economics of ageing",
  ["j econ asymmetries"] = "Journal of economic asymmetries",
  ["j econ behav organ"] = "Journal of economic behavior & organization",
  ["j econ bus"] = "Journal of economics and business",
  ["j econ cooperation islamic countries"] = "Journal of Economic Cooperation among Islamic Countries",
  ["j econ dev"] = "Journal of economic development",
  ["j econ devel"] = "Journal of Economic Development",
  ["j econ dyn control"] = "Journal of economic dynamics & control",
  ["j econ dynam control"] = "Journal of Economic Dynamics and Control",
  ["j econ educ"] = "The Journal of economic education",
  ["j econ entomol"] = "Journal of economic entomology",
  ["j econ finance"] = "Journal of Economics and Finance",
  ["j econ geogr"] = "Journal of economic geography",
  ["j econ growth"] = "Journal of Economic Growth",
  ["j econ growth (boston)"] = "Journal of economic growth (Boston, Mass.)",
  ["j econ hist"] = "The journal of economic history",
  ["j econ inequal"] = "Journal of economic inequality",
  ["j econ integration"] = "Journal of Economic Integration",
  ["j econ interact coord"] = "Journal of economic interaction and coordination",
  ["j econ issues"] = "Journal of economic issues",
  ["j econ lit"] = "Journal of economic literature",
  ["j econ manag strategy"] = "Journal of economics & management strategy",
  ["j econ manage strategy"] = "Journal of Economics and Management Strategy",
  ["j econ methodology"] = "Journal of Economic Methodology",
  ["j econ of bus"] = "Journal of the Economics of Business",
  ["j econ perspect"] = "The journal of economic perspectives : a journal of the American Economic Association",
  ["j econ psych"] = "Journal of Economic Psychology",
  ["j econ psychol"] = "Journal Of Economic Psychology",
  ["j econ public financ"] = "Journal of economics and public finance",
  ["j econ res"] = "Journal of Economic Research",
  ["j econ sci assoc"] = "Journal of the Economic Science Association",
  ["j econ soc hist orient"] = "Journal of economic and social history of the orient",
  ["j econ soc meas"] = "Journal of economic and social measurement",
  ["j econ soc pol"] = "Journal of Economic and Social Policy",
  ["j econ soc res"] = "Journal of Economic and Social Research",
  ["j econ struct"] = "Journal of economic structures",
  ["j econ stud"] = "Journal of economic studies (Glasgow, Scotland)",
  ["j econ surv"] = "Journal of economic surveys",
  ["j econ surveys"] = "Journal of Economic Surveys",
  ["j econ theory"] = "Journal of economic theory",
  ["j econ theory econometrics"] = "Journal of Economic Theory and Econometrics",
  ["j econ trade mark manag"] = "Journal of economics, trade and marketing management",
  ["j econom"] = "Journal of econometrics",
  ["j econom dynam control"] = "Journal of Economic Dynamics & Control",
  ["j econom method"] = "Journal of econometric methods",
  ["j econom methods"] = "Journal of Econometric Methods",
  ["j econom theory"] = "Journal of Economic Theory",
  ["j econometrics"] = "Journal of Econometrics",
  ["j ecosyst ecography"] = "Journal of ecosystem & ecography",
  ["j ecotourism"] = "Journal of Ecotourism",
  ["j ect"] = "The journal of ECT",
  ["j ecumenical stud"] = "Journal of ecumenical studies",
  ["j educ (boston)"] = "Journal of education (Boston, Mass.)",
  ["j educ adm hist"] = "Journal of educational administration and history",
  ["j educ audiol"] = "Journal of educational audiology : official journal of the Educational Audiology Association",
  ["j educ behav stat"] = "Journal of educational and behavioral statistics : a quarterly publication sponsored by the American Educational Research Association and the American Statistical Association",
  ["j educ comput res"] = "Journal of Educational Computing Research",
  ["j educ cult stud"] = "Journal of education and culture studies",
  ["j educ develop psychol"] = "Journal of educational and developmental psychology",
  ["j educ elearn res"] = "Journal of education and e-learning research",
  ["j educ environ sci health"] = "Journal of education in science, environment and health",
  ["j educ eval health prof"] = "Journal of educational evaluation for health professions",
  ["j educ finance"] = "Journal of Education Finance",
  ["j educ health promot"] = "Journal of education and health promotion",
  ["j educ meas"] = "Journal of educational measurement",
  ["j educ perioper med"] = "The journal of education in perioperative medicine : JEPM",
  ["j educ psychol"] = "Journal of educational psychology",
  ["j educ psychol consult"] = "Journal of educational and psychological consultation : the official journal of the Association for Educational and Psychological Consultants",
  ["j educ res"] = "The Journal of educational research",
  ["j educ soc work"] = "Journal of education for social work",
  ["j educ stud placed risk"] = "Journal of education for students placed at risk",
  ["j educ sustain dev"] = "Journal of education for sustainable development",
  ["j educ teach"] = "Journal of education for teaching : international research and pedagogy",
  ["j educ teach soc stud"] = "Journal of education, teaching and social studies",
  ["j educ techno soc"] = "Journal of educational technology & society",
  ["j educ technol syst"] = "Journal of Educational Technology Systems",
  ["j educ train stud"] = "Journal of education and training studies",
  ["j egypt archaeol"] = "The Journal of Egyptian archaeology",
  ["j egypt math soc"] = "Journal of the Egyptian Mathematical Society",
  ["j egypt med assoc"] = "The Journal of the Egyptian Medical Association",
  ["j egypt natl canc inst"] = "Journal of the Egyptian National Cancer Institute",
  ["j egypt public health assoc"] = "The Journal of the Egyptian Public Health Association",
  ["j egypt soc obstet gynecol"] = "Journal Of The Egyptian Society Of Obstetrics And Gynecology",
  ["j egypt soc parasitol"] = "Journal of the Egyptian Society of Parasitology",
  ["j egyptian math soc"] = "Journal of the Egyptian Mathematical Society",
  ["j ehealth technol appl"] = "Journal of ehealth technology and application",
  ["j eighteenth century stud"] = "Journal for eighteenth-century studies",
  ["j elast"] = "Journal of Elasticity",
  ["j elasticity"] = "Journal of Elasticity. The Physical and Mathematical Science of Solids",
  ["j elastomers plast"] = "Journal of Elastomers and Plastics",
  ["j elder abuse negl"] = "Journal of elder abuse & neglect",
  ["j elect public opin parties"] = "Journal of elections, public opinion and parties",
  ["j electr bioimpedance"] = "Journal of electrical bioimpedance",
  ["j electr comput eng"] = "Journal of Electrical and Computer Engineering",
  ["j electr electron eng aust"] = "Journal of electrical and electronics engineering, Australia",
  ["j electr eng technol"] = "Journal of Electrical Engineering and Technology",
  ["j electr propul"] = "Journal of Electric Propulsion",
  ["j electr syst inf technol"] = "Journal of Electrical Systems and Information Technology",
  ["j electroanal chem"] = "Journal of Electroanalytical Chemistry",
  ["j electroanal chem (lausanne)"] = "Journal of electroanalytical chemistry (Lausanne, Switzerland)",
  ["j electroanal chem interfacial electrochem"] = "Journal of Electroanalytical Chemistry and Interfacial Electrochemistry",
  ["j electrocardiol"] = "Journal of electrocardiology",
  ["j electroceram"] = "Journal of Electroceramics",
  ["j electroceramics"] = "Journal of Electroceramics",
  ["j electrochem energy convers storage"] = "Journal of Electrochemical Energy Conversion and Storage",
  ["j electrochem sci technol"] = "Journal of Electrochemical Science and Technology",
  ["j electrochem soc"] = "Journal of the Electrochemical Society",
  ["j electrochem soc india"] = "Journal of the Electrochemical Society of India",
  ["j electromagn anal appl"] = "Journal of Electromagnetic Analysis and Applications",
  ["j electromagn waves appl"] = "Journal of electromagnetic waves and applications",
  ["j electromyogr kinesiol"] = "Journal of electromyography and kinesiology : official journal of the International Society of Electrophysiological Kinesiology",
  ["j electron"] = "Journal of Electronics (China)",
  ["j electron imaging"] = "Journal of electronic imaging",
  ["j electron manuf"] = "Journal of Electronics Manufacturing",
  ["j electron mater"] = "Journal of Electronic Materials",
  ["j electron microsc"] = "Journal of Electron Microscopy",
  ["j electron microsc (tokyo)"] = "Journal of electron microscopy",
  ["j electron microsc tech"] = "Journal of electron microscopy technique",
  ["j electron packag"] = "Journal of Electronic Packaging",
  ["j electron publ"] = "The journal of electronic publishing : JEP",
  ["j electron resour librariansh"] = "Journal of electronic resources librarianship",
  ["j electron sci technol"] = "Journal of Electronic Science and Technology",
  ["j electron spectros relat phenomena"] = "Journal of electron spectroscopy and related phenomena",
  ["j electron spectrosc relat phenom"] = "Journal of Electron Spectroscopy and Related Phenomena",
  ["j electron test"] = "Journal of Electronic Testing",
  ["j electronic mater"] = "Journal of Electronic Materials",
  ["j electrostat"] = "Journal of Electrostatics",
  ["j electrotech math"] = "Journal of Electrotechnics and Mathematics",
  ["j elem"] = "Journal of elementology",
  ["j elementol"] = "Journal of Elementology",
  ["j elisha mitchell sci soc"] = "Journal of the Elisha Mitchell Scientific Society",
  ["j elisha mitchell sci soc chapel hill n c"] = "Journal. Elisha Mitchell Scientific Society (Chapel Hill, N.C.)",
  ["j elliptic parabol equ"] = "Journal of elliptic and parabolic equations",
  ["j elliptic parabolic equations"] = "Journal of Elliptic and Parabolic Equations",
  ["j embryol exp morphol"] = "Journal of embryology and experimental morphology",
  ["j emerg crit care med"] = "Journal of emergency and critical care medicine (Hong Kong, China)",
  ["j emerg dis virol"] = "Journal of emerging diseases and virology",
  ["j emerg infect dis"] = "Journal of emerging infectious diseases",
  ["j emerg investig"] = "Journal of emerging investigators",
  ["j emerg manag"] = "Journal of emergency management (Weston, Mass.)",
  ["j emerg manage disaster commun"] = "Journal of Emergency Management and Disaster Communications",
  ["j emerg med"] = "The Journal of emergency medicine",
  ["j emerg med forecast"] = "Journal of emergency medicine forecast",
  ["j emerg med serv jems"] = "Journal of Emergency Medical Services",
  ["j emerg nurs"] = "Journal of emergency nursing",
  ["j emerg technol innov res"] = "Journal of emerging technologies and innovative research",
  ["j emerg trauma shock"] = "Journal of emergencies, trauma, and shock",
  ["j emerging markets"] = "Journal of Emerging Markets",
  ["j emot behav disord"] = "Journal of emotional and behavioral disorders",
  ["j empir finance"] = "Journal of empirical finance",
  ["j empir leg stud"] = "Journal of empirical legal studies",
  ["j empir res hum res ethics"] = "Journal of empirical research on human research ethics : JERHRE",
  ["j empirical finance"] = "Journal of Empirical Finance",
  ["j empirical res hum res ethics"] = "Journal of Empirical Research on Human Research Ethics",
  ["j employ couns"] = "Journal of employment counseling",
  ["j enabling technol"] = "Journal of enabling technologies",
  ["j endocr genet"] = "The journal of endocrine genetics : JEG",
  ["j endocr soc"] = "Journal of the Endocrine Society",
  ["j endocrinol"] = "Journal of Endocrinology",
  ["j endocrinol diabetes"] = "Journal of endocrinology and diabetes",
  ["j endocrinol diabetes mellit"] = "Journal of endocrinology and diabetes mellitus",
  ["j endocrinol diabetes obes"] = "Journal of endocrinology, diabetes & obesity",
  ["j endocrinol invest"] = "Journal of endocrinological investigation",
  ["j endocrinol metab"] = "Journal of endocrinology and metabolism",
  ["j endocrinol reprod"] = "Journal of endocrinology and reproduction : JER",
  ["j endocrinol thyroid res"] = "Journal of endocrinology and thyroid research",
  ["j endod"] = "Journal of endodontics",
  ["j endometr"] = "Journal of endometriosis",
  ["j endometr pelvic pain disord"] = "Journal of endometriosis and pelvic pain disorders",
  ["j endotoxin res"] = "Journal of endotoxin research",
  ["j endourol"] = "Journal of endourology",
  ["j endourol b videourol"] = "Journal of endourology. Part B, Videourology",
  ["j endourol case rep"] = "Journal of endourology case reports",
  ["j endovasc resusc trauma manag"] = "Journal of endovascular resuscitation and trauma management",
  ["j endovasc surg"] = "Journal of endovascular surgery : the official journal of the International Society for Endovascular Surgery",
  ["j endovasc ther"] = "Journal of endovascular therapy : an official journal of the International Society of Endovascular Specialists",
  ["j energ mater"] = "Journal of Energetic Materials",
  ["j energetic mater"] = "Journal of Energetic Materials",
  ["j energy"] = "Journal of Energy",
  ["j energy chem"] = "Journal of Energy Chemistry",
  ["j energy devel"] = "Journal of Energy and Development",
  ["j energy eng"] = "Journal of Energy Engineering",
  ["j energy finance devel"] = "Journal of Energy Finance and Development",
  ["j energy inst"] = "Journal of the Energy Institute",
  ["j energy lit"] = "Journal of Energy Literature",
  ["j energy res technol"] = "Journal of Energy Resources Technology",
  ["j energy resour technol"] = "Journal of energy resources technology",
  ["j energy south afr"] = "Journal of Energy in Southern Africa",
  ["j energy storage"] = "Journal of energy storage",
  ["j eng"] = "Journal of Engineering",
  ["j eng (stevenage)"] = "Journal of engineering (Stevenage, England)",
  ["j eng appl sci"] = "Journal of Engineering and Applied Science",
  ["j eng des"] = "Journal of Engineering Design",
  ["j eng educ"] = "Journal of Engineering Education",
  ["j eng fiber fabr"] = "Journal of engineered fibers and fabrics",
  ["j eng fibers fabr"] = "Journal of Engineered Fibers and Fabrics",
  ["j eng gas turbine power"] = "Journal of engineering for gas turbines and power",
  ["j eng gas turbines power"] = "Journal of Engineering for Gas Turbines and Power",
  ["j eng linguist"] = "Journal of English linguistics",
  ["j eng mater technol"] = "Journal of Engineering Materials and Technology",
  ["j eng math"] = "Journal of Engineering Mathematics",
  ["j eng mech"] = "Journal of Engineering Mechanics",
  ["j eng phys thermophys"] = "Journal of Engineering Physics and Thermophysics",
  ["j eng psychol"] = "Journal of engineering psychology",
  ["j eng sci med diagn ther"] = "Journal of engineering and science in medical diagnostics and therapy",
  ["j eng sustainable build cities"] = "Journal of Engineering for Sustainable Buildings and Cities",
  ["j eng tech manage"] = "Journal of Engineering and Technology Management",
  ["j eng thermophys"] = "Journal of Engineering Thermophysics",
  ["j eng, des technol"] = "Journal of Engineering, Design and Technology",
  ["j engl acad purp"] = "Journal of English for academic purposes",
  ["j engrg math"] = "Journal of Engineering Mathematics",
  ["j enhanced heat transfer"] = "Journal of Enhanced Heat Transfer",
  ["j enterostomal ther"] = "Journal of enterostomal therapy",
  ["j enterp transform"] = "Journal of Enterprise Transformation",
  ["j entomol"] = "Journal of entomology (Online)",
  ["j entomol acarol res"] = "Journal Of Entomological And Acarological Research",
  ["j entomol indones"] = "Jurnal entomologi Indonesia",
  ["j entomol nematol"] = "Journal of entomology and nematology",
  ["j entomol res"] = "Journal of Entomological Research",
  ["j entomol res soc"] = "Journal of the Entomological Research Society",
  ["j entomol sci"] = "Journal of entomological science",
  ["j entomol ser a gen entomol"] = "Journal of Entomology Series A, General Entomology",
  ["j entomol ser b taxon"] = "Journal of Entomology Series B, Taxonomy",
  ["j entomol soc b c"] = "Journal of the Entomological Society of British Columbia. Entomological Society of British Columbia",
  ["j entomol soc ont"] = "Journal of the Entomological Society of Ontario",
  ["j entomol soc south afr"] = "Journal of the Entomological Society of Southern Africa",
  ["j entomol zool stud"] = "Journal of entomology and zoology studies",
  ["j entrep educ"] = "Journal of entrepreneurship education",
  ["j entwickl polit"] = "Journal für Entwicklungspolitik : JEP",
  ["j environ anal toxicol"] = "Journal of environmental & analytical toxicology",
  ["j environ assess policy manage"] = "Journal of Environmental Assessment Policy and Management",
  ["j environ biol"] = "Journal of Environmental Biology",
  ["j environ chem eng"] = "Journal of Environmental Chemical Engineering",
  ["j environ dev"] = "The Journal of environment & development",
  ["j environ devel"] = "Journal of Environment and Development",
  ["j environ econ manage"] = "Journal of environmental economics and management",
  ["j environ educ"] = "The Journal of environmental education",
  ["j environ eng"] = "Journal of Environmental Engineering",
  ["j environ eng (new york)"] = "Journal of environmental engineering (New York, N.Y.)",
  ["j environ eng (reston, va, us)"] = "Journal of Environmental Engineering (Reston, VA, United States)",
  ["j environ eng ecol sci"] = "Journal of environmental engineering & ecological science",
  ["j environ eng geophys"] = "Journal of Environmental and Engineering Geophysics",
  ["j environ eng landscape manage"] = "Journal of Environmental Engineering and Landscape Management",
  ["j environ eng sci"] = "Journal of Environmental Engineering and Science",
  ["j environ health"] = "Journal of Environmental Health",
  ["j environ health sci"] = "Journal of environment and health sciences",
  ["j environ health sci eng"] = "Journal of environmental health science & engineering",
  ["j environ health sustainable dev"] = "Journal of Environmental Health and Sustainable Development",
  ["j environ hortic"] = "Journal of environmental horticulture",
  ["j environ immunol toxicol"] = "Journal of environmental immunology and toxicology",
  ["j environ inf"] = "Journal of Environmental Informatics",
  ["j environ manage"] = "Journal of environmental management",
  ["j environ mater sustainable energy"] = "Journal of Environmental Materials and Sustainable Energy",
  ["j environ monit"] = "Journal of Environmental Monitoring",
  ["j environ nanotechnol"] = "Journal of Environmental Nanotechnology",
  ["j environ occup sci"] = "Journal of environmental and occupational science",
  ["j environ pathol toxicol"] = "Journal of Environmental Pathology and Toxicology",
  ["j environ pathol toxicol oncol"] = "Journal of environmental pathology, toxicology and oncology : official organ of the International Society for Environmental Toxicology and Cancer",
  ["j environ plann manage"] = "Journal of Environmental Planning and Management",
  ["j environ plann policy manage"] = "Journal of Environmental Policy & Planning",
  ["j environ planning manage"] = "Journal of Environmental Planning and Management",
  ["j environ polym degrad"] = "Journal of environmental polymer degradation",
  ["j environ prot (irvine, calif)"] = "Journal of environmental protection",
  ["j environ prot ecol"] = "Journal of Environmental Protection and Ecology",
  ["j environ prot sci"] = "Journal of environmental protection science",
  ["j environ psychol"] = "Journal of environmental psychology",
  ["j environ public health"] = "Journal of environmental and public health",
  ["j environ qual"] = "Journal of environmental quality",
  ["j environ radioact"] = "Journal of environmental radioactivity",
  ["j environ sci"] = "Journal of Environmental Sciences",
  ["j environ sci (china)"] = "Journal of environmental sciences (China)",
  ["j environ sci eng"] = "Journal of environmental science & engineering",
  ["j environ sci eng (nagpur, india)"] = "Journal of Environmental Science & Engineering (Nagpur, India)",
  ["j environ sci health [c]"] = "Journal of Environmental Science and Health. Part C, Environmental Health Sciences",
  ["j environ sci health a"] = "Journal of Environmental Science and Health, Part A Environmental Science",
  ["j environ sci health a environ sci eng"] = "Journal of environmental science and health. Part A, Environmental science and engineering",
  ["j environ sci health a environ sci eng toxic hazard subst control"] = "Journal of environmental science and health. Part A, Environmental science and engineering & toxic and hazardous substance control",
  ["j environ sci health a environ sci health part a: environ sci eng"] = "Journal of Environmental Science and Health, Part A: Environmental Science and Engineering",
  ["j environ sci health a tox hazard subst environ eng"] = "Journal of environmental science and health. Part A, Toxic/hazardous substances & environmental engineering",
  ["j environ sci health a: environ sci engtoxic hazard subst control"] = "Journal of Environmental Science and Health, Part A: Environmental Science and Engineering & Toxic and Hazardous Substance Control",
  ["j environ sci health a: toxic/hazard subst environ eng"] = "Journal of Environmental Science and Health, Part A: Toxic/Hazardous Substances & Environmental Engineering",
  ["j environ sci health b"] = "Journal of environmental science and health. Part. B, Pesticides, food contaminants, and agricultural wastes",
  ["j environ sci health c"] = "Journal of environmental science and health. Part C: Environmental health sciences",
  ["j environ sci health c environ carcinog ecotoxicol rev"] = "Journal of environmental science and health. Part C, Environmental carcinogenesis & ecotoxicology reviews",
  ["j environ sci health c toxicol carcinog"] = "Journal of environmental science and health. Part C, Toxicology and carcinogenesis",
  ["j environ sci health c: environ carcinog ecotoxicol rev"] = "Journal of Environmental Science and Health, Part C: Environmental Carcinogenesis & Ecotoxicology Reviews",
  ["j environ sci health part a environ sci health part a environ sci eng"] = "Journal of Environmental Science and Health, Part A: Environmental Science and Engineering",
  ["j environ sci health part a tox hazard subst environ eng"] = "Journal of Environmental Science and Health. Part A, Toxic/Hazardous Substances and Environmental Engineering",
  ["j environ sci health part a toxic/hazard subst environ eng"] = "Journal of Environmental Science and Health, Part A: Toxic/Hazardous Substances & Environmental Engineering",
  ["j environ sci health, part a"] = "Journal of Environmental Science and Health, Part A Environmental Science",
  ["j environ sci health, part a environ sci engtoxic hazard subst control"] = "Journal of Environmental Science and Health, Part A: Environmental Science and Engineering & Toxic and Hazardous Substance Control",
  ["j environ sci health, part b"] = "Journal of Environmental Science and Health, Part B: Pesticides, Food Contaminants, and Agricultural Wastes",
  ["j environ sci health, part c environ carcinog ecotoxicol rev"] = "Journal of Environmental Science and Health, Part C: Environmental Carcinogenesis & Ecotoxicology Reviews",
  ["j environ sci health, part c environ carcinog rev"] = "Journal of Environmental Science and Health, Part C: Environmental Carcinogenesis Reviews",
  ["j environ sci health, part c environ health sci"] = "Journal of Environmental Science and Health, Part C: Environmental Health Sciences",
  ["j environ sci health: toxic/hazard subst"] = "Environ. Eng. Journal of Environmental Science and Health, Part A: Toxic/Hazardous Substances & Environmental Engineering",
  ["j environ sci technol"] = "Journal of Environmental Science and Technology",
  ["j environ solutions oil gas min"] = "Journal of Environmental Solutions for Oil, Gas, and Mining",
  ["j environ stud (northborough)"] = "Journal of environmental studies (Northborough, Mass.)",
  ["j environ stud sci"] = "Journal of environmental studies and sciences",
  ["j environ toxicol public health"] = "Journal of environmental toxiology and public health",
  ["j environ toxicol stud"] = "Journal of environmental and toxicological studies",
  ["j enzym inhib"] = "Journal of Enzyme Inhibition",
  ["j enzyme inhib"] = "Journal of Enzyme Inhibition",
  ["j enzyme inhib med chem"] = "Journal of enzyme inhibition and medicinal chemistry",
  ["j enzymol metab"] = "Journal of enzymology and metabolism",
  ["j epidemiol"] = "Journal of epidemiology",
  ["j epidemiol biostat"] = "Journal of epidemiology and biostatistics",
  ["j epidemiol community health"] = "Journal of epidemiology and community health",
  ["j epidemiol community health (1978)"] = "Journal of epidemiology and community health",
  ["j epidemiol glob health"] = "Journal of epidemiology and global health",
  ["j epidemiol global health"] = "Journal of Epidemiology and Global Health",
  ["j epidemiol prev med"] = "Journal of epidemiology and preventive medicine",
  ["j epidemiol public health rev"] = "Journal of epidemiology and public health reviews",
  ["j epidemiol res"] = "Journal of epidemiological research",
  ["j epilepsy"] = "Journal of epilepsy",
  ["j epilepsy res"] = "Journal of epilepsy research",
  ["j epithel biol pharmacol"] = "Journal of epithelial biology & pharmacology",
  ["j equine sci"] = "Journal of equine science",
  ["j equine vet sci"] = "Journal of equine veterinary science",
  ["j equity health"] = "The journal of equity in health",
  ["j erie stud"] = "The Journal of Erie studies",
  ["j escience librariansh"] = "Journal of escience librarianship",
  ["j essent oil bear plants"] = "Journal of Essential Oil-Bearing Plants",
  ["j essent oil res"] = "Journal of Essential Oil Research",
  ["j esthet dent"] = "Journal of esthetic dentistry",
  ["j et nurs"] = "Journal of ET nursing : official publication, International Association for Enterostomal Therapy",
  ["j ethics"] = "The journal of ethics",
  ["j ethics law aging"] = "Journal of ethics, law, and aging",
  ["j ethics ment health"] = "Journal of ethics in mental health",
  ["j ethiop stud"] = "Journal of Ethiopian studies",
  ["j ethn crim justice"] = "Journal of ethnicity in criminal justice",
  ["j ethn cult divers soc work"] = "Journal of ethnic & cultural diversity in social work",
  ["j ethn migr stud"] = "Journal of ethnic and migration studies",
  ["j ethn stud"] = "The Journal of ethnic studies",
  ["j ethn subst abuse"] = "Journal of ethnicity in substance abuse",
  ["j ethnic foods"] = "Journal of Ethnic Foods",
  ["j ethnobiol"] = "Journal of ethnobiology",
  ["j ethnobiol ethnomed"] = "Journal of ethnobiology and ethnomedicine",
  ["j ethnogr qual res"] = "Journal of ethnographic and qualitative research",
  ["j ethnopharmacol"] = "Journal of ethnopharmacology",
  ["j ethol"] = "Journal of ethology",
  ["j eukaryot microbiol"] = "The Journal of eukaryotic microbiology",
  ["j eukaryotic microbiol"] = "Journal of Eukaryotic Microbiology",
  ["j eur acad dermatol venereol"] = "Journal of the European Academy of Dermatology and Venereology : JEADV",
  ["j eur assoc health inf libr"] = "Journal of the European Association for Health Information and Libraries",
  ["j eur ceram soc"] = "Journal of the European Ceramic Society",
  ["j eur cme"] = "Journal of European CME",
  ["j eur econ assoc"] = "Journal of the European Economic Association",
  ["j eur econ hist"] = "The Journal of European economic history",
  ["j eur hydrol"] = "Journal Europeen d’Hydrologie",
  ["j eur integr"] = "Journal of European integration",
  ["j eur landscape"] = "Journal of European Landscape",
  ["j eur math soc"] = "Journal of the European Mathematical Society",
  ["j eur math soc (jems)"] = "Journal of the European Mathematical Society (JEMS)",
  ["j eur mosq control assoc"] = "Journal of the European Mosquito Control Association",
  ["j eur opt soc part a"] = "Journal of the European Optical Society Part A",
  ["j eur opt soc part b"] = "Journal of the European Optical Society Part B",
  ["j eur opt soc rapid publ"] = "Journal of the European Optical Society. Rapid publications",
  ["j eur opt soc-rapid publ"] = "Journal of the European Optical Society-Rapid Publications",
  ["j eur public policy"] = "Journal of European public policy",
  ["j eur soc policy"] = "Journal of European social policy",
  ["j eur stud"] = "Journal of European studies",
  ["j europ econ hist"] = "Journal of European Economic History",
  ["j eval clin pract"] = "Journal of evaluation in clinical practice",
  ["j evang theol soc"] = "Journal of the Evangelical Theological Society. Evangelical Theological Society",
  ["j evid based complementary altern med"] = "Journal of evidence-based complementary & alternative medicine",
  ["j evid based dent pract"] = "The journal of evidence-based dental practice",
  ["j evid based integr med"] = "Journal of evidence-based integrative medicine",
  ["j evid based med"] = "Journal of evidence-based medicine",
  ["j evid based psychother"] = "Journal of evidence-based psychotherapies",
  ["j evid based soc work"] = "Journal of evidence-based social work",
  ["j evid based soc work (2019)"] = "Journal of evidence-based social work (2019)",
  ["j evid inf soc work"] = "Journal of evidence-informed social work",
  ["j evidence qual health care"] = "Journal of Evidence and Quality in Health Care",
  ["j evidence-based integr med"] = "Journal of Evidence-Based Integrative Medicine",
  ["j evidence-based med"] = "Journal of Evidence-Based Medicine",
  ["j evol biochem physiol"] = "Journal of evolutionary biochemistry and physiology",
  ["j evol biol"] = "Journal of evolutionary biology",
  ["j evol econ"] = "Journal of evolutionary economics",
  ["j evol equ"] = "Journal of evolution equations",
  ["j evol equations"] = "Journal of Evolution Equations",
  ["j evol health"] = "Journal of evolution and health",
  ["j evol med dent sci"] = "Journal of evolution of medical and dental sciences",
  ["j evol stem cell res"] = "Journal of evolving stem cell research",
  ["j evol technol"] = "Journal of evolution and technology",
  ["j evolutionary econ"] = "Journal of Evolutionary Economics",
  ["j excell coll teach"] = "Journal on excellence in college teaching",
  ["j except child"] = "Journal of exceptional children",
  ["j exerc nutrition biochem"] = "Journal of exercise nutrition & biochemistry",
  ["j exerc physiol online"] = "Journal of exercise physiology online",
  ["j exerc rehabil"] = "Journal of exercise rehabilitation",
  ["j exerc sci fit"] = "Journal of exercise science and fitness",
  ["j exerc sports orthop"] = "Journal of exercise, sports & orthopedics",
  ["j exercise physiol"] = "Journal Of Exercise Physiology",
  ["j existent"] = "Journal of existentialism",
  ["j exot pet med"] = "Journal of exotic pet medicine",
  ["j exp anal behav"] = "Journal of the experimental analysis of behavior",
  ["j exp anim sci"] = "Journal of experimental animal science",
  ["j exp appl anim sci"] = "Journal of experimental and applied animal sciences",
  ["j exp biol"] = "The Journal of experimental biology",
  ["j exp bot"] = "Journal of experimental botany",
  ["j exp child psychol"] = "Journal of experimental child psychology",
  ["j exp clin assist reprod"] = "Journal of experimental & clinical assisted reproduction",
  ["j exp clin cancer res"] = "Journal of experimental & clinical cancer research : CR",
  ["j exp clin med"] = "Journal of experimental and clinical medicine",
  ["j exp criminol"] = "Journal of experimental criminology",
  ["j exp educ"] = "Journal of experimental education",
  ["j exp integr med"] = "Journal of experimental and integrative medicine",
  ["j exp life sci"] = "Journal of experimental life science",
  ["j exp mar bio ecol"] = "Journal of experimental marine biology and ecology",
  ["j exp mar biol ecol"] = "Journal of Experimental Marine Biology and Ecology",
  ["j exp med"] = "The Journal of experimental medicine",
  ["j exp med sci"] = "Journal of experimental medical sciences",
  ["j exp nanosci"] = "Journal of Experimental Nanoscience",
  ["j exp nephrol"] = "Journal of experimental nephrology",
  ["j exp neurol"] = "Journal of experimental neurology",
  ["j exp neurosci"] = "Journal of experimental neuroscience",
  ["j exp orthop"] = "Journal of experimental orthopaedics",
  ["j exp pathol"] = "Journal of experimental pathology",
  ["j exp pathol (oxford)"] = "Journal of experimental pathology (Oxford, England)",
  ["j exp pathol (wilmington)"] = "Journal of Experimental Pathology",
  ["j exp pharmacol"] = "Journal of experimental pharmacology",
  ["j exp psychol"] = "Journal of experimental psychology",
  ["j exp psychol [hum learn]"] = "Journal of Experimental Psychology: Human Learning and Memory",
  ["j exp psychol anim behav process"] = "Journal of experimental psychology. Animal behavior processes",
  ["j exp psychol anim learn cogn"] = "Journal of experimental psychology. Animal learning and cognition",
  ["j exp psychol anim learn cognit"] = "Journal of Experimental Psychology. Animal Learning and Cognition",
  ["j exp psychol appl"] = "Journal of experimental psychology. Applied",
  ["j exp psychol gen"] = "Journal of experimental psychology. General",
  ["j exp psychol hum learn"] = "Journal of experimental psychology. Human learning and memory",
  ["j exp psychol hum percept perform"] = "Journal of experimental psychology. Human perception and performance",
  ["j exp psychol learn mem cogn"] = "Journal of experimental psychology. Learning, memory, and cognition",
  ["j exp psychopathol"] = "Journal of experimental psychopathology",
  ["j exp res hum growth aging"] = "Journal of experimental research on human growth & aging",
  ["j exp second sci"] = "The journal of experimental secondary science",
  ["j exp soc psychol"] = "Journal of experimental social psychology",
  ["j exp stroke transl med"] = "Journal of experimental stroke & translational medicine",
  ["j exp theor artif intell"] = "Journal of Experimental & Theoretical Artificial Intelligence",
  ["j exp theor phys"] = "Journal of Experimental and Theoretical Physics",
  ["j exp ther oncol"] = "Journal of experimental therapeutics & oncology",
  ["j exp zool"] = "Journal of Experimental Zoology",
  ["j exp zool a comp exp biol"] = "Journal of experimental zoology. Part A, Comparative experimental biology",
  ["j exp zool a ecol genet physiol"] = "Journal of experimental zoology. Part A, Ecological genetics and physiology",
  ["j exp zool a ecol integr physiol"] = "Journal of experimental zoology. Part A, Ecological and integrative physiology",
  ["j exp zool b mol dev evol"] = "Journal of experimental zoology. Part B, Molecular and developmental evolution",
  ["j exp zool part a ecol genet physiol"] = "Journal of experimental zoology. Part A, Ecological genetics and physiology",
  ["j exp zool part a: comp exp biol"] = "Journal of Experimental Zoology Part A, Comparative Experimental Biology",
  ["j exp zool part a: ecol genet physiol"] = "Journal of Experimental Zoology Part A: Ecological Genetics and Physiology",
  ["j exp zool part a: ecol integr physiol"] = "Journal of Experimental Zoology Part A, Ecological and Integrative Physiology",
  ["j exp zool part b"] = "Journal of Experimental Zoology Part B Molecular and Developmental Evolution",
  ["j exp zool suppl"] = "Journal of Experimental Zoology. Supplement",
  ["j exp zoolog a comp exp biol"] = "Journal of Experimental Zoology. Part A, Comparative Experimental Biology",
  ["j exp zoolog b mol dev evol"] = "Journal of Experimental Zoology. Part B, Molecular and Developmental Evolution",
  ["j exp zoology india"] = "Journal of experimental zoology India",
  ["j experiment theoret phys"] = "Journal of Experimental and Theoretical Physics",
  ["j explor res pharmacol"] = "Journal of exploratory research in pharmacology",
  ["j explos eng"] = "Journal of Explosives Engineering",
  ["j expo anal environ epidemiol"] = "Journal of exposure analysis and environmental epidemiology",
  ["j expo sci environ epidemiol"] = "Journal of exposure science & environmental epidemiology",
  ["j exposure anal environ epidemiol"] = "Journal of Exposure Analysis and Environment Epidemiology",
  ["j exposure sci environ epidemiol"] = "Journal of Exposure Science & Environmental Epidemiology",
  ["j ext"] = "Journal of extension",
  ["j extra corpor technol"] = "The journal of extra-corporeal technology",
  ["j extracell biol"] = "Journal of Extracellular Biology",
  ["j extracell vesicles"] = "Journal of extracellular vesicles",
  ["j extreme events"] = "Journal of Extreme Events",
  ["j eye mov res"] = "Journal of eye movement research",
  ["j fac agric hokkaido univ"] = "Journal of the Faculty of Agriculture, Hokkaido University = Hokkaido Daigaku Nōgakubu kiyo",
  ["j fac agric, shinshu univ"] = "Journal of the Faculty of Agriculture, Shinshu University",
  ["j fac dev"] = "The journal of faculty development",
  ["j fac eng chiba univ"] = "Journal of the Faculty of Engineering",
  ["j fac eng univ tokyo ser b"] = "Journal of the Faculty of Engineering",
  ["j fac int stud cult kyushu sangyo univ"] = "Kyushu Sangyo University",
  ["j fac liberal arts yamaguchi univ natur sci"] = "Journal of the Faculty of Liberal Arts",
  ["j fac med baghdad"] = "Journal of the Faculty of Medicine, Baghdad",
  ["j fac radiol"] = "The Journal of the Faculty of Radiologists. Faculty of Radiologists (Great Britain)",
  ["j fac sci eng kinki univ"] = "Journal of the Faculty of Science and Engineering",
  ["j fac sci shinshu univ"] = "Journal of the Faculty of Science. Shinshu University",
  ["j fac sci tech kinki univ"] = "Kinki University",
  ["j facade des eng"] = "Journal of Facade Design and Engineering",
  ["j fail anal prev"] = "Journal of Failure Analysis and Prevention",
  ["j fam commun"] = "Journal of family communication",
  ["j fam consum sci"] = "Journal of family and consumer sciences",
  ["j fam cult"] = "The Journal of family and culture",
  ["j fam econ issues"] = "Journal of family and economic issues",
  ["j fam health"] = "Journal of family health",
  ["j fam health care"] = "The journal of family health care",
  ["j fam hist"] = "Journal of family history",
  ["j fam issues"] = "Journal of family issues",
  ["j fam law"] = "Journal of family law",
  ["j fam med"] = "Journal of family medicine",
  ["j fam med community health"] = "Journal of family medicine & community health",
  ["j fam med dis prev"] = "Journal of family medicine and disease prevention",
  ["j fam nurs"] = "Journal of family nursing",
  ["j fam plann reprod health care"] = "The journal of family planning and reproductive health care",
  ["j fam pract"] = "The Journal of family practice",
  ["j fam psychol"] = "Journal of family psychology : JFP : journal of the Division of Family Psychology of the American Psychological Association (Division 43)",
  ["j fam psychother"] = "Journal of family psychotherapy",
  ["j fam soc work"] = "Journal of family social work",
  ["j fam stud"] = "Journal of family studies",
  ["j fam theory rev"] = "Journal of family theory & review",
  ["j fam ther"] = "Journal of family therapy",
  ["j fam violence"] = "Journal of family violence",
  ["j fam welf"] = "The Journal of family welfare",
  ["j famil health train"] = "Journal of family health training",
  ["j family community med"] = "Journal of family & community medicine",
  ["j family econ issues"] = "Journal of Family and Economic Issues",
  ["j family med prim care"] = "Journal of family medicine and primary care",
  ["j family reprod health"] = "Journal of family & reproductive health",
  ["j family strengths"] = "Journal of family strengths",
  ["j far eastern bus"] = "Journal of Far Eastern Business",
  ["j farm"] = "Jornal dos farmacêuticos",
  ["j farm econ"] = "Journal of Farm Economics",
  ["j fasting health"] = "Journal of fasting and health",
  ["j feline med surg"] = "Journal of feline medicine and surgery",
  ["j feline med surg open rep"] = "Journal of Feline Medicine and Surgery Open Reports",
  ["j fem fam ther"] = "Journal of feminist family therapy",
  ["j ferment bioeng"] = "Journal of Fermentation and Bioengineering",
  ["j ferment technol"] = "Journal of Fermentation Technology",
  ["j fertil in vitro ivf worldw reprod med genet stem cell biol"] = "Journal of fertilization. In vitro IVF worldwide reproductive medicine genetics & stem cell biology",
  ["j fertili in vitro"] = "Journal of fertilization in vitro",
  ["j fiber bioeng inf"] = "Journal of Fiber Bioengineering and Informatics",
  ["j fiber sci technol"] = "Journal of Fiber Science and Technology",
  ["j field archaeol"] = "Journal of field archaeology",
  ["j field ornith"] = "Journal of Field Ornithology",
  ["j field ornithol"] = "Journal of field ornithology",
  ["j field rob"] = "Journal of Field Robotics",
  ["j field robot"] = "Journal of field robotics",
  ["j finan econ"] = "Journal of Financial Economics",
  ["j finan intermediation"] = "Journal of Financial Intermediation",
  ["j finan manage anal"] = "Journal of Financial Management and Analysis",
  ["j finan quant anal"] = "Journal of Financial and Quantitative Analysis",
  ["j finan res"] = "Journal of Financial Research",
  ["j finan services res"] = "Journal of Financial Services Research",
  ["j financ crime"] = "Journal of financial crime",
  ["j financ econ"] = "Journal of financial economics",
  ["j finance"] = "The Journal of finance",
  ["j finsler geom appl"] = "Journal of Finsler Geometry and its Applications",
  ["j fintech"] = "Journal of FinTech",
  ["j fire prot eng"] = "Journal of Fire Protection Engineering",
  ["j fire sci"] = "Journal of Fire Sciences",
  ["j fish aquac"] = "Journal of fisheries and aquaculture",
  ["j fish aquat sci"] = "Journal of Fisheries and Aquatic Science",
  ["j fish biol"] = "Journal of fish biology",
  ["j fish dis"] = "Journal of fish diseases",
  ["j fish environ"] = "Journal of fisheries and environment",
  ["j fish res board can"] = "Journal of the Fisheries Research Board of Canada",
  ["j fish sci technol"] = "Journal of fisheries science and technology",
  ["j fish wildl manag"] = "Journal of fish and wildlife management",
  ["j fish wildl manage"] = "Journal of Fish and Wildlife Management",
  ["j fishscicom"] = "Journal of FisheriesSciences.com",
  ["j fitopatologi"] = "Jurnal Fitopatologi = Indonesian journal of phytopathology",
  ["j fixed point theory appl"] = "Journal of Fixed Point Theory and Applications",
  ["j fla med assoc"] = "The Journal of the Florida Medical Association",
  ["j fla state dent soc"] = "The Journal of the Florida State Dental Society",
  ["j flood risk manag"] = "Journal of flood risk management",
  ["j flood risk manage"] = "Journal of Flood Risk Management",
  ["j flow chem"] = "Journal of flow chemistry",
  ["j flow visualization image process"] = "Journal of Flow Visualization and Image Processing",
  ["j fluency disord"] = "Journal of fluency disorders",
  ["j fluid flow heat mass transf"] = "Journal of fluid flow, heat and mass transfer",
  ["j fluid mech"] = "Journal of Fluid Mechanics",
  ["j fluids eng"] = "Journal of Fluids Engineering",
  ["j fluids struct"] = "Journal of Fluids and Structures",
  ["j fluor chem"] = "Journal of fluorine chemistry",
  ["j fluoresc"] = "Journal of fluorescence",
  ["j fluorine chem"] = "Journal of Fluorine Chemistry",
  ["j foetal med"] = "Journal of foetal medicine",
  ["j food agric environ"] = "Journal of Food, Agriculture and Environment",
  ["j food bioact"] = "Journal of food bioactives : an official scientific publication of the International Society of Nutraceuticals and Functional Foods (ISNFF)",
  ["j food biochem"] = "Journal of food biochemistry",
  ["j food chem nanotechnol"] = "Journal Of Food Chemistry And Nanotechnology",
  ["j food compos anal"] = "Journal of Food Composition and Analysis",
  ["j food compost anal"] = "Journal of food composition and analysis : an official publication of the United Nations University, International Network of Food Data Systems",
  ["j food drug anal"] = "Journal of food and drug analysis",
  ["j food eng"] = "Journal of Food Engineering",
  ["j food ind microbiol"] = "Journal of food & industrial microbiology",
  ["j food lipids"] = "Journal of Food Lipids",
  ["j food meas charact"] = "Journal of Food Measurement and Characterization",
  ["j food nutr disord"] = "Journal of food & nutritional disorders",
  ["j food nutr res (newark)"] = "Journal of food and nutrition research (Newark, Del.)",
  ["j food process eng"] = "Journal of food process engineering",
  ["j food process preserv"] = "Journal of Food Processing and Preservation",
  ["j food process technol"] = "Journal of food processing & technology",
  ["j food prod mark"] = "Journal of Food Products Marketing",
  ["j food prot"] = "Journal of food protection",
  ["j food qual"] = "Journal of food quality",
  ["j food qual hazards control"] = "Journal of food quality & hazards control",
  ["j food res"] = "Journal of food research",
  ["j food saf"] = "Journal of Food Safety",
  ["j food saf food qual"] = "Journal of Food Safety and Food Quality",
  ["j food sci"] = "Journal of food science",
  ["j food sci educ"] = "Journal of Food Science Education",
  ["j food sci eng"] = "Journal of food science and engineering",
  ["j food sci nutr res"] = "Journal of food science and nutrition research",
  ["j food sci technol"] = "Journal of food science and technology",
  ["j food technol"] = "Journal of Food Technology",
  ["j foot ankle res"] = "Journal of foot and ankle research",
  ["j foot ankle surg"] = "The Journal of foot and ankle surgery : official publication of the American College of Foot and Ankle Surgeons",
  ["j foot surg"] = "The Journal of foot surgery",
  ["j for"] = "Journal of forestry",
  ["j for econ"] = "Journal of forest economics",
  ["j for res"] = "Journal of forest research : open access",
  ["j for res (harbin)"] = "Journal of forestry research",
  ["j for sci (prague)"] = "Journal of forest science",
  ["j for suisse"] = "Journal forestier suisse",
  ["j foramineferal res"] = "Journal of Foramineferal Research",
  ["j foraminifer res"] = "Journal of foraminiferal research",
  ["j forecast"] = "Journal of forecasting",
  ["j forecasting"] = "Journal of Forecasting",
  ["j forens psychiatry psychol"] = "The journal of forensic psychiatry & psychology",
  ["j forensic dent sci"] = "Journal of forensic dental sciences",
  ["j forensic econ"] = "Journal of Forensic Economics",
  ["j forensic identif"] = "Journal of forensic identification",
  ["j forensic leg med"] = "Journal of forensic and legal medicine",
  ["j forensic legal med"] = "Journal of Forensic and Legal Medicine",
  ["j forensic med"] = "Journal of Forensic Medicine",
  ["j forensic med toxicol"] = "Journal of Forensic Medicine and Toxicology",
  ["j forensic neuropsychol"] = "Journal of Forensic Neuropsychology",
  ["j forensic nurs"] = "Journal of forensic nursing",
  ["j forensic odontostomatol"] = "The Journal of forensic odonto-stomatology",
  ["j forensic pract"] = "Journal of Forensic Practice",
  ["j forensic psychiatry psychol"] = "Journal of Forensic Psychiatry & Psychology",
  ["j forensic psychol pract"] = "Journal of forensic psychology practice",
  ["j forensic psychol res pract"] = "Journal of forensic psychology research and practice",
  ["j forensic radiol imaging"] = "Journal of Forensic Radiology and Imaging",
  ["j forensic sci"] = "Journal of forensic sciences",
  ["j forensic sci crim investig"] = "Journal of forensic sciences & criminal investigation",
  ["j forensic sci criminol"] = "Journal of forensic science & criminology",
  ["j forensic sci soc"] = "Journal - Forensic Science Society",
  ["j forensic soc work"] = "Journal of forensic social work",
  ["j forensics res"] = "Journal of forensics research",
  ["j form des learn"] = "Journal of formative design in learning",
  ["j formaliz reason"] = "Journal of formalized reasoning",
  ["j formos med assoc"] = "Journal of the Formosan Medical Association = Taiwan yi zhi",
  ["j fourier anal appl"] = "The Journal of Fourier Analysis and Applications",
  ["j fr med chir thorac"] = "Journal français de médecine et chirurgie thoraciques",
  ["j fr ophtalmol"] = "Journal français d'ophtalmologie",
  ["j fr otorhinolaryngol audiophonol chir maxillofac"] = "Journal français d'oto-rhino-laryngologie, audio-phonologie et chirurgie maxillo-faciale",
  ["j fr otorhinolaryngol chir maxillofac"] = "Journal français d'oto-rhino-laryngologie et chirurgie maxillo-faciale",
  ["j fract calc"] = "Journal of Fractional Calculus",
  ["j fract calc appl"] = "Journal of Fractional Calculus and Applications",
  ["j fractal geom"] = "Journal of Fractal Geometry",
  ["j frailty aging"] = "The Journal of frailty & aging",
  ["j frailty sarcopenia falls"] = "Journal of frailty, sarcopenia and falls",
  ["j franklin inst"] = "Journal of the Franklin Institute",
  ["j franklin inst b"] = "Journal of the Franklin Institute",
  ["j franklin inst- eng appl math"] = "Journal of the Franklin Institute-Engineering and Applied Mathematics",
  ["j free radic biol med"] = "Journal of free radicals in biology & medicine",
  ["j free radicals biol med"] = "Journal of Free Radicals in Biology & Medicine",
  ["j freshw ecol"] = "Journal of freshwater ecology",
  ["j freshwater ecol"] = "Journal of Freshwater Ecology",
  ["j freshwater fish"] = "Journal of Freshwater Fish",
  ["j frict wear"] = "Journal of Friction and Wear",
  ["j friends hist soc"] = "The Journal of the Friends' Historical Society",
  ["j fudan univ nat sci"] = "Fudan University. Journal. Natural Science. Fudan Xuebao. Ziran Kexue Ban",
  ["j fuel cell sci technol"] = "Journal of Fuel Cell Science and Technology",
  ["j fuel chem technol"] = "Journal of Fuel Chemistry and Technology",
  ["j funct anal"] = "Journal of Functional Analysis",
  ["j funct biomater"] = "Journal of functional biomaterials",
  ["j funct foods"] = "Journal of Functional Foods",
  ["j funct logic programming"] = "The Journal of Functional and Logic Programming",
  ["j funct morphol kinesiol"] = "Journal of functional morphology and kinesiology",
  ["j funct program"] = "Journal of Functional Programming",
  ["j funct programming"] = "Journal of Functional Programming",
  ["j funct spaces"] = "Journal of Function Spaces",
  ["j fungi"] = "Journal of Fungi",
  ["j fungi (basel)"] = "Journal of fungi (Basel, Switzerland)",
  ["j furth high educ"] = "Journal of further and higher education",
  ["j fusion energy"] = "Journal of Fusion Energy",
  ["j future foods"] = "Journal of Future Foods",
  ["j future robot life"] = "Journal of Future Robot Life",
  ["j futures markets"] = "Journal of Futures Markets",
  ["j fuzhou univ nat sci ed"] = "Journal of Fuzhou University. Natural Science Edition. Fuzhou Daxue Xuebao. Ziran Kexue Ban",
  ["j fuzzy math"] = "Journal of Fuzzy Mathematics",
  ["j ga dent assoc"] = "Journal of the Georgia Dental Association",
  ["j ga public health assoc"] = "Journal of the Georgia Public Health Association",
  ["j gambl issues"] = "Journal of gambling issues",
  ["j gambl stud"] = "Journal of gambling studies",
  ["j gansu univ technol"] = "Journal of Gansu University of Technology",
  ["j gansu univ technol (engl ed)"] = "Journal of Gansu University of Technology",
  ["j gard hist"] = "Journal of garden history",
  ["j gasbeleucht verw beleuchtungsarten wasserversorg"] = "Journal fur Gasbeleuchtung und Verwandte Beleuchtungsarten sowie fur Wasserversorgung",
  ["j gastric cancer"] = "Journal of gastric cancer",
  ["j gastroenterol"] = "Journal of gastroenterology",
  ["j gastroenterol hepatol"] = "Journal of gastroenterology and hepatology",
  ["j gastroenterol hepatol erkrank"] = "Journal für gastroenterologische und hepatologische Erkrankungen",
  ["j gastroenterol hepatol res"] = "Journal of gastroenterology and hepatology research",
  ["j gastroenterol pancreatol liver disord"] = "Journal of gastroenterology, pancreatology & liver disorders",
  ["j gastrointest cancer"] = "Journal of gastrointestinal cancer",
  ["j gastrointest dig syst"] = "Journal of gastrointestinal & digestive system",
  ["j gastrointest oncol"] = "Journal of gastrointestinal oncology",
  ["j gastrointest surg"] = "Journal of gastrointestinal surgery : official journal of the Society for Surgery of the Alimentary Tract",
  ["j gastrointestin liver dis"] = "Journal of gastrointestinal and liver diseases : JGLD",
  ["j gay lesbian med assoc"] = "Journal of the Gay and Lesbian Medical Association",
  ["j gay lesbian ment health"] = "Journal of gay & lesbian mental health",
  ["j gay lesbian soc serv"] = "Journal of gay & lesbian social services",
  ["j gemmol"] = "Journal of Gemmology",
  ["j gen appl microbiol"] = "The Journal of general and applied microbiology",
  ["j gen educ"] = "The Journal of general education",
  ["j gen fam med"] = "Journal of general and family medicine",
  ["j gen intern med"] = "Journal of general internal medicine",
  ["j gen lie theory appl"] = "Journal of Generalized Lie Theory and Applications",
  ["j gen med (dover)"] = "Journal of general medicine (Dover, Del.)",
  ["j gen microbiol"] = "Journal of General Microbiology",
  ["j gen mol virol"] = "Journal of general and molecular virology",
  ["j gen orthod"] = "Journal of general orthodontics",
  ["j gen philos sci"] = "Journal for General Philosophy of Science",
  ["j gen physiol"] = "The Journal of general physiology",
  ["j gen plant pathol"] = "Journal of general plant pathology : JGPP",
  ["j gen pract (los angel)"] = "Journal of general practice (Los Angeles, Calif.)",
  ["j gen psychol"] = "The Journal of general psychology",
  ["j gen virol"] = "The Journal of general virology",
  ["j gend specif med"] = "The journal of gender-specific medicine : JGSM : the official journal of the Partnership for Women's Health at Columbia",
  ["j gend stud"] = "Journal of gender studies",
  ["j gene med"] = "The journal of gene medicine",
  ["j gene ther"] = "Journal of gene therapy",
  ["j generic med"] = "Journal of generic medicines",
  ["j genes proteins"] = "Journal of genes and proteins",
  ["j genet"] = "Journal of genetics",
  ["j genet breed"] = "Journal of genetics & breeding",
  ["j genet couns"] = "Journal of genetic counseling",
  ["j genet disord dis inf"] = "Journal of genetic disorders & disease information",
  ["j genet disord genet rep"] = "Journal of genetic disorders & genetic reports",
  ["j genet eng biotechnol"] = "Journal, genetic engineering & biotechnology",
  ["j genet geneal"] = "The journal of genetic genealogy",
  ["j genet genome res"] = "Journal of genetics and genome research",
  ["j genet genomics"] = "Journal of genetics and genomics = Yi chuan xue bao",
  ["j genet hum"] = "Journal de génétique humaine",
  ["j genet psychol"] = "The Journal of genetic psychology",
  ["j genet study"] = "Journal of genetics study",
  ["j genet syndr gene ther"] = "Journal of genetic syndromes & gene therapy",
  ["j genit syst disord"] = "Journal of genital system & disorders",
  ["j genitourin disord"] = "Journal of genitourinary disorders",
  ["j genocide res"] = "Journal of genocide research",
  ["j genome sci technol"] = "Journal of Genome Science and Technology",
  ["j genomics"] = "Journal of genomics",
  ["j geo-eng sci"] = "Journal of Geo-Engineering Sciences",
  ["j geochem explor"] = "Journal of Geochemical Exploration",
  ["j geod"] = "Journal of Geodesy",
  ["j geodetic sci"] = "Journal of Geodetic Science",
  ["j geodyn"] = "Journal of Geodynamics",
  ["j geog"] = "The Journal of geography",
  ["j geog sci"] = "Journal of Geographical Sciences",
  ["j geogr"] = "Journal of Geography",
  ["j geogr environ earth sci int"] = "Journal of Geography, Environment and Earth Science International",
  ["j geogr sci"] = "Journal of Geographical Sciences",
  ["j geogr syst"] = "Journal of geographical systems",
  ["j geograph systems"] = "Journal of Geographical Systems",
  ["j geol"] = "Journal of Geology",
  ["j geol educ"] = "Journal of Geological Education",
  ["j geol res"] = "Journal of Geological Research",
  ["j geol soc (london, uk)"] = "Journal of the Geological Society (London, United Kingdom)",
  ["j geol soc aust"] = "Journal of the Geological Society of Australia",
  ["j geol soc china"] = "Journal of the Geological Society of China",
  ["j geol soc india"] = "Journal of the Geological Society of India",
  ["j geol soc jam"] = "Journal of the Geological Society of Jamaica",
  ["j geol soc london"] = "Journal of the Geological Society of London",
  ["j geom"] = "Journal of Geometry",
  ["j geom anal"] = "Journal of Geometric Analysis",
  ["j geom graph"] = "Journal for Geometry and Graphics",
  ["j geom mech"] = "Journal of Geometric Mechanics",
  ["j geom phys"] = "Journal of Geometry and Physics",
  ["j geom symmetry phys"] = "Journal of Geometry and Symmetry in Physics",
  ["j geomagn geoelec"] = "Journal of Geomagnetism and Geoelectricity",
  ["j geophys"] = "Journal of Geophysics",
  ["j geophys eng"] = "Journal of Geophysics and Engineering",
  ["j geophys res"] = "Journal of Geophysical Research",
  ["j geophys res [atmos]"] = "Journal of Geophysical Research, [Atmospheres]",
  ["j geophys res [biogeosci]"] = "Journal of Geophysical Research, [Biogeosciences]",
  ["j geophys res [earth surface]"] = "Journal of Geophysical Research, [Earth Surface]",
  ["j geophys res [oceans]"] = "Journal of Geophysical Research, [Oceans]",
  ["j geophys res [planets]"] = "Journal of Geophysical Research, [Planets]",
  ["j geophys res [solid earth]"] = "Journal of Geophysical Research, [Solid Earth]",
  ["j geophys res [space phys]"] = "Journal of Geophysical Research, [Space Physics]",
  ["j geophys res a: space phys"] = "Journal of Geophysical Research, A: Space Physics",
  ["j geophys res atmos"] = "Journal of geophysical research. Atmospheres : JGR",
  ["j geophys res b: solid earth"] = "Journal of Geophysical Research, B: Solid Earth",
  ["j geophys res biogeosci"] = "Journal of geophysical research. Biogeosciences",
  ["j geophys res c: oceans"] = "Journal of Geophysical Research, C: Oceans",
  ["j geophys res d: atmos"] = "Journal of Geophysical Research, D: Atmospheres",
  ["j geophys res e: planets"] = "Journal of Geophysical Research, E: Planets",
  ["j geophys res earth surf"] = "Journal of Geophysical Research, Earth Surface",
  ["j geophys res f: earth surf"] = "Journal of Geophysical Research, F: Earth Surface",
  ["j geophys res g: biogeosci"] = "Journal of Geophysical Research, G: Biogeosciences",
  ["j geophys res oceans"] = "Journal of Geophysical Research, Oceans",
  ["j geophys res planets"] = "Journal of geophysical research. Planets",
  ["j geophys res solid earth"] = "Journal of Geophysical Research, Solid Earth",
  ["j geophys res space phys"] = "Journal of Geophysical Research, Space Physics",
  ["j geophys res: atmos"] = "Journal of Geophysical Research: Atmospheres",
  ["j geophys res: biogeosci"] = "Journal of Geophysical Research: Biogeosciences",
  ["j geophys res: earth surf"] = "Journal of Geophysical Research: Earth Surface",
  ["j geophys res: mach learn comput"] = "Journal of Geophysical Research: Machine Learning and Computation",
  ["j geophys res: oceans"] = "Journal of Geophysical Research: Oceans",
  ["j geophys res: planets"] = "Journal of Geophysical Research: Planets",
  ["j geophys res: solid earth"] = "Journal of Geophysical Research: Solid Earth",
  ["j geophys res: space phys"] = "Journal of Geophysical Research: Space Physics",
  ["j geosci"] = "Journal of Geosciences",
  ["j geosci educ"] = "Journal of Geoscience Education",
  ["j geotech geoenviron eng"] = "Journal of Geotechnical and Geoenvironmental Engineering",
  ["j geovisualization spatial anal"] = "Journal of Geovisualization and Spatial Analysis",
  ["j geriatr"] = "Journal of geriatrics",
  ["j geriatr cardiol"] = "Journal of geriatric cardiology : JGC",
  ["j geriatr care res"] = "Journal of geriatric care and research",
  ["j geriatr med gerontol"] = "Journal of geriatric medicine and gerontology",
  ["j geriatr oncol"] = "Journal of geriatric oncology",
  ["j geriatr palliat care"] = "Journal of geriatrics and palliative care",
  ["j geriatr phys ther"] = "Journal of geriatric physical therapy (2001)",
  ["j geriatr psychiatry"] = "Journal of geriatric psychiatry",
  ["j geriatr psychiatry neurol"] = "Journal of geriatric psychiatry and neurology",
  ["j german statistical society"] = "Journal of the German Statistical Society",
  ["j germantown hosp"] = "Journal of the Germantown Hospital",
  ["j gerontol"] = "Journal of gerontology",
  ["j gerontol a biol sci med sci"] = "The journals of gerontology. Series A, Biological sciences and medical sciences",
  ["j gerontol b psychol sci soc sci"] = "The journals of gerontology. Series B, Psychological sciences and social sciences",
  ["j gerontol geriatr res"] = "Journal of gerontology & geriatric research",
  ["j gerontol nurs"] = "Journal of gerontological nursing",
  ["j gerontol soc work"] = "Journal of gerontological social work",
  ["j gesch"] = "Journal für Geschichte",
  ["j ghana sci assoc"] = "Journal of the Ghana Science Association",
  ["j ginseng res"] = "Journal of ginseng research",
  ["j glaciol"] = "Journal of Glaciology",
  ["j glaciol geocryol"] = "Journal of glaciology and geocryology",
  ["j glass stud"] = "Journal of glass studies",
  ["j glaucoma"] = "Journal of glaucoma",
  ["j glbt fam stud"] = "Journal of GLBT family studies",
  ["j glob antimicrob resist"] = "Journal of global antimicrobial resistance",
  ["j glob drug policy pract"] = "The journal of global drug policy and practice",
  ["j glob econ anal"] = "Journal of global economic analysis",
  ["j glob ethics"] = "Journal of global ethics",
  ["j glob health"] = "Journal of global health",
  ["j glob health perspect"] = "Journal of global health perspectives",
  ["j glob health rep"] = "Journal of global health reports",
  ["j glob hist"] = "Journal of global history",
  ["j glob infect dis"] = "Journal of global infectious diseases",
  ["j glob oncol"] = "Journal of global oncology",
  ["j glob optim"] = "Journal of global optimization : an international journal dealing with theoretical and computational aspects of seeking global optima and their applications in science, management and engineering",
  ["j glob radiol"] = "The journal of global radiology",
  ["j global antimicrob resist"] = "Journal of Global Antimicrobial Resistance",
  ["j global change data discovery"] = "Journal of Global Change Data & Discovery",
  ["j global inf technol manage"] = "Journal of Global Information Technology Management",
  ["j global oncol"] = "Journal of Global Oncology",
  ["j global optim"] = "Journal of Global Optimization",
  ["j glycobiol"] = "Journal of glycobiology",
  ["j glycomics lipidomics"] = "Journal of glycomics & lipidomics",
  ["j gnathol"] = "The Journal of gnathology",
  ["j gokova geom topol ggt"] = "Journal of Gokova Geometry Topology. GGT",
  ["j gov inf"] = "Journal of government information : an international review of policy, issues and resources",
  ["j grad med educ"] = "Journal of graduate medical education",
  ["j grad sch fac eng univ tokyo ser b"] = "Journal of the Graduate School and Faculty of Engineering",
  ["j graph algorithms appl"] = "Journal of Graph Algorithms and Applications",
  ["j graph label"] = "Journal of Graph Labeling. A Graph Theoretical Research Journal",
  ["j graph theory"] = "Journal of Graph Theory",
  ["j gravit physiol"] = "Journal of gravitational physiology : a journal of the International Society for Gravitational Physiology",
  ["j great lakes res"] = "Journal of Great Lakes Research",
  ["j grid comput"] = "Journal of grid computing",
  ["j group psychother psychodrama sociom"] = "Journal of group psychotherapy, psychodrama and sociometry",
  ["j group theory"] = "Journal of Group Theory",
  ["j groups addict recover"] = "Journal of groups in addiction & recovery",
  ["j groups complex cryptol"] = "Journal of Groups, Complexity, Cryptology",
  ["j gt houst dent soc"] = "The Journal of the Greater Houston Dental Society",
  ["j guid control dyn"] = "Journal of guidance, control, and dynamics : a publication of the American Institute of Aeronautics and Astronautics devoted to the technology of dynamics and control",
  ["j gynaecol endocrinol"] = "Journal of gynaecological endocrinology",
  ["j gynakol endokrinol"] = "Journal für gynäkologische Endokrinologie (Österreichische Ausg.)",
  ["j gynecol endosc surg"] = "Journal of gynecological endoscopy and surgery",
  ["j gynecol obstet biol reprod (paris)"] = "Journal de gynécologie, obstétrique et biologie de la reproduction",
  ["j gynecol obstet hum reprod"] = "Journal of gynecology obstetrics and human reproduction",
  ["j gynecol oncol"] = "Journal of gynecologic oncology",
  ["j gynecol res"] = "Journal of gynecology research",
  ["j gynecol surg"] = "Journal of gynecologic surgery",
  ["j halacha contemporary society"] = "Journal of halacha and contemporary society",
  ["j hand microsurg"] = "Journal of hand and microsurgery",
  ["j hand surg [am]"] = "Journal of Hand Surgery. American Volume",
  ["j hand surg [br]"] = "Journal of Hand Surgery. British Volume",
  ["j hand surg am"] = "The Journal of hand surgery",
  ["j hand surg asian pac vol"] = "The journal of hand surgery Asian-Pacific volume",
  ["j hand surg br"] = "Journal of hand surgery (Edinburgh, Scotland)",
  ["j hand surg eur vol"] = "The Journal of hand surgery, European volume",
  ["j hand surg glob online"] = "Journal of hand surgery global online",
  ["j hand ther"] = "Journal of hand therapy : official journal of the American Society of Hand Therapists",
  ["j hangzhou univ natur sci ed"] = "Journal of Hangzhou University",
  ["j happiness stud"] = "Journal of happiness studies",
  ["j harbin inst tech"] = "Journal of Harbin Institute of Technology. Harbin Gongye Daxue Xuebao",
  ["j hard tissue biol"] = "Journal of hard tissue biology",
  ["j hardware syst secur"] = "Journal of Hardware and Systems Security",
  ["j harmon res med health sci"] = "Journal of harmonized research in medical and health sciences",
  ["j hattori bot lab"] = "The Journal of the Hattori Botanical Laboratory = Hattori Shokubutsu Kenkyūjo hōkoku",
  ["j hawaii dent assoc"] = "Journal of the Hawaii Dental Association",
  ["j hawaii state dent assoc"] = "Journal of the Hawaii State Dental Association",
  ["j hazard mater"] = "Journal of hazardous materials",
  ["j hazard mater adv"] = "Journal of Hazardous Materials Advances",
  ["j hazard mater lett"] = "Journal of Hazardous Materials Letters",
  ["j hazard toxic radioact waste"] = "Journal of hazardous, toxic and radioactive waste",
  ["j head neck spine surg"] = "Journal of head, neck & spine surgery",
  ["j head trauma rehabil"] = "The Journal of head trauma rehabilitation",
  ["j headache pain"] = "The journal of headache and pain",
  ["j health adm educ"] = "The Journal of health administration education",
  ["j health behav public health"] = "Journal of health behavior and public health",
  ["j health biomed law"] = "Journal of health & biomedical law",
  ["j health care benefits"] = "Journal of health care benefits",
  ["j health care chaplain"] = "Journal of health care chaplaincy",
  ["j health care finance"] = "Journal of health care finance",
  ["j health care inter des"] = "Journal of Health Care Interior Design",
  ["j health care law policy"] = "Journal of health care law & policy",
  ["j health care mark"] = "Journal of health care marketing",
  ["j health care poor underserved"] = "Journal of health care for the poor and underserved",
  ["j health care technol"] = "Journal of health care technology",
  ["j health commun"] = "Journal of health communication",
  ["j health dispar res pract"] = "Journal of health disparities research and practice",
  ["j health econ"] = "Journal of health economics",
  ["j health econ dev"] = "Journal of health economics and development",
  ["j health econ outcomes res"] = "Journal of health economics and outcomes research",
  ["j health educ"] = "Journal of health education",
  ["j health educ res dev"] = "Journal of health education research & development",
  ["j health educ teach"] = "Journal of health education teaching",
  ["j health environ educ"] = "The journal of health, environment & education",
  ["j health hosp law"] = "Journal of health and hospital law : a publication of the American Academy of Hospital Attorneys of the American Hospital Association",
  ["j health hum behav"] = "Journal of health and human behavior",
  ["j health hum resour adm"] = "Journal of health and human resources administration",
  ["j health hum serv adm"] = "Journal of health and human services administration",
  ["j health inform"] = "Journal of health informatics",
  ["j health inform dev ctries"] = "Journal of health informatics in developing countries",
  ["j health law"] = "Journal of health law",
  ["j health life sci law"] = "Journal of health & life sciences law",
  ["j health manag"] = "Journal of health management",
  ["j health mass commun"] = "Journal of health & mass communication",
  ["j health med econ"] = "Journal of health & medical economics",
  ["j health med inform"] = "Journal of health & medical informatics",
  ["j health organ manag"] = "Journal of health organization and management",
  ["j health phys educ"] = "The Journal of health and physical education",
  ["j health polit policy law"] = "Journal of health politics, policy and law",
  ["j health politics"] = "Journal of Health Politics, Policy and Law",
  ["j health pollut"] = "Journal of health & pollution",
  ["j health popul dev ctries"] = "Journal of health & population in developing countries",
  ["j health popul nutr"] = "Journal of health, population, and nutrition",
  ["j health psychol"] = "Journal of health psychology",
  ["j health res"] = "Journal of health research",
  ["j health res rev dev ctries"] = "Journal of health research and reviews in developing countries",
  ["j health sci"] = "Journal of health science (Tokyo, Japan)",
  ["j health sci (el monte)"] = "Journal of health science (El Monte)",
  ["j health sci educ"] = "Journal of health science & education",
  ["j health sci surveill syst"] = "Journal of health sciences and surveillance system",
  ["j health serv psychol"] = "Journal of health service psychology",
  ["j health serv res policy"] = "Journal of health services research & policy",
  ["j health soc behav"] = "Journal of health and social behavior",
  ["j health soc policy"] = "Journal of health & social policy",
  ["j health soc sci"] = "Journal of health and social sciences",
  ["j healthc commun"] = "Journal of healthcare communications",
  ["j healthc des"] = "Journal of Healthcare Design",
  ["j healthc educ train"] = "Journal of healthcare education and training : the journal of the American Society for Healthcare Education and Training",
  ["j healthc eng"] = "Journal of healthcare engineering",
  ["j healthc inf manag"] = "Journal of healthcare information management : JHIM",
  ["j healthc inform res"] = "Journal of healthcare informatics research",
  ["j healthc leadersh"] = "Journal of healthcare leadership",
  ["j healthc manag"] = "Journal of healthcare management / American College of Healthcare Executives",
  ["j healthc mater manage"] = "Journal of healthcare materiel management",
  ["j healthc prot manage"] = "Journal of healthcare protection management : publication of the International Association for Hospital Security",
  ["j healthc qual"] = "Journal for healthcare quality : official publication of the National Association for Healthcare Quality",
  ["j healthc qual res"] = "Journal of healthcare quality research",
  ["j healthc resour manag"] = "Journal of healthcare resource management",
  ["j healthc risk manag"] = "Journal of healthcare risk management : the journal of the American Society for Healthcare Risk Management",
  ["j healthc sci humanit"] = "Journal of healthcare, science and the humanities",
  ["j healthcare eng"] = "Journal of Healthcare Engineering",
  ["j healthcare inf res"] = "Journal of Healthcare Informatics Research",
  ["j hear sci"] = "Journal of hearing science",
  ["j heart cardiol"] = "Journal of heart and cardiology",
  ["j heart fail"] = "The Journal of heart failure : the official journal of the International Society of Heart Failure",
  ["j heart health"] = "Journal of heart health",
  ["j heart lung transplant"] = "The Journal of heart and lung transplantation : the official publication of the International Society for Heart Transplantation",
  ["j heart transplant"] = "The Journal of heart transplantation",
  ["j heart valve dis"] = "The Journal of heart valve disease",
  ["j heat recovery syst"] = "Journal of Heat Recovery Systems",
  ["j heat transfer"] = "Journal of Heat Transfer",
  ["j heat treat"] = "Journal of Heat Treating",
  ["j hefei univ technol nat sci"] = "Journal of Hefei University of Technology. Natural Science. Hefei Gongye Daxue Xuebao. Ziran Kexue Ban",
  ["j hell stud"] = "The Journal of Hellenic studies",
  ["j helminthol"] = "Journal of helminthology",
  ["j hematol"] = "Journal of hematology",
  ["j hematol malig"] = "Journal of hematological malignancies",
  ["j hematol oncol"] = "Journal of hematology & oncology",
  ["j hematol oncol pharm"] = "Journal of hematology oncology pharmacy",
  ["j hematol res"] = "Journal of hematology research",
  ["j hematol thromb"] = "Journal of hematology & thrombosis",
  ["j hematol transfus"] = "Journal of hematology & transfusion",
  ["j hematop"] = "Journal of hematopathology",
  ["j hematopathol"] = "Journal of Hematopathology",
  ["j hematother"] = "Journal of hematotherapy",
  ["j hematother stem cell res"] = "Journal of hematotherapy & stem cell research",
  ["j hepatobiliary pancreat sci"] = "Journal of hepato-biliary-pancreatic sciences",
  ["j hepatobiliary pancreat surg"] = "Journal of hepato-biliary-pancreatic surgery",
  ["j hepatocell carcinoma"] = "Journal of hepatocellular carcinoma",
  ["j hepatol"] = "Journal of hepatology",
  ["j hepatol suppl"] = "Journal of hepatology. Supplement",
  ["j herb med"] = "Journal of herbal medicine",
  ["j herb pharmacother"] = "Journal of herbal pharmacotherapy",
  ["j herbs spices med plants"] = "Journal of Herbs Spices & Medicinal Plants",
  ["j hered"] = "The Journal of heredity",
  ["j heritage manage"] = "Journal of Heritage Management",
  ["j herpetol"] = "Journal of herpetology",
  ["j herpetol med surg"] = "Journal of herpetological medicine and surgery",
  ["j heterocycl chem"] = "Journal of Heterocyclic Chemistry",
  ["j heuristics"] = "Journal of Heuristics",
  ["j high educ outreach engagem"] = "Journal of higher education outreach and engagement",
  ["j high energy astrophys"] = "Journal of High Energy Astrophysics",
  ["j high energy phys"] = "Journal of High Energy Physics",
  ["j high resolut chromatogr"] = "Journal of high resolution chromatography : HRC",
  ["j high school res clin sci"] = "Journal of high school research in the clinical sciences",
  ["j high speed networks"] = "Journal of High Speed Networks",
  ["j high technol manage res"] = "Journal of High Technology Management Research",
  ["j higher educ"] = "The Journal of higher education",
  ["j higher educ pol manage"] = "Journal of Higher Education Policy and Management",
  ["j highway transp res dev"] = "Journal of Highway and Transportation Research and Development (English Edition)",
  ["j hill res"] = "Journal of hill research",
  ["j hillside hosp"] = "Journal of the Hillside Hospital",
  ["j hip preserv surg"] = "Journal of hip preservation surgery",
  ["j hirnforsch"] = "Journal fur Hirnforschung",
  ["j hispanic high educ"] = "Journal of Hispanic higher education",
  ["j hist arabic sci"] = "Journal for the history of Arabic science",
  ["j hist astron"] = "Journal for the History of Astronomy",
  ["j hist astronom"] = "Journal for the History of Astronomy",
  ["j hist behav sci"] = "Journal of the history of the behavioral sciences",
  ["j hist biol"] = "Journal of the history of biology",
  ["j hist child youth"] = "The journal of the history of childhood and youth",
  ["j hist dent"] = "Journal of the history of dentistry",
  ["j hist econ thought"] = "Journal of the history of economic thought : JHET",
  ["j hist geogr"] = "Journal of historical geography",
  ["j hist ideas"] = "Journal of the history of ideas",
  ["j hist med"] = "Jornal de historia da medicina",
  ["j hist med allied sci"] = "Journal of the history of medicine and allied sciences",
  ["j hist neurosci"] = "Journal of the history of the neurosciences",
  ["j hist philos"] = "Journal of the history of philosophy",
  ["j hist sex"] = "Journal of the history of sexuality",
  ["j hist sociol"] = "Journal of historical sociology",
  ["j histochem cytochem"] = "The journal of histochemistry and cytochemistry : official journal of the Histochemistry Society",
  ["j histotechnol"] = "Journal of histotechnology",
  ["j hiv aids"] = "Journal of HIV and AIDS",
  ["j hiv aids prev child youth"] = "Journal of HIV/AIDS prevention in children & youth",
  ["j hiv aids soc serv"] = "Journal of HIV/AIDS & social services",
  ["j hiv aids surveill epidemiol"] = "Journal of HIV AIDS surveillance & epidemiology",
  ["j hiv clin sci res"] = "Journal of HIV for clinical and scientific research",
  ["j hiv retro virus"] = "Journal of HIV & retro virus",
  ["j hiv ther"] = "Journal of HIV therapy",
  ["j hiv/aids prev educ adolesc child"] = "Journal of HIV/AIDS prevention & education for adolescents & children",
  ["j hohai univ"] = "Journal of Hohai University",
  ["j hokkaido univ ed sect ii a"] = "Journal of Hokkaido University of Education",
  ["j holist nurs"] = "Journal of holistic nursing : official journal of the American Holistic Nurses' Association",
  ["j hologr speckle"] = "Journal of Holography and Speckle",
  ["j home econ"] = "Journal of home economics",
  ["j homel secur emerg manag"] = "Journal of homeland security and emergency management",
  ["j homosex"] = "Journal of homosexuality",
  ["j homotopy relat struct"] = "Journal of Homotopy and Related Structures",
  ["j hong kong branch r asiat soc"] = "Journal of the Hong Kong Branch of the Royal Asiatic Society",
  ["j hortic sci"] = "The Journal of horticultural science",
  ["j hortic sci biotechnol"] = "The journal of horticultural science & biotechnology",
  ["j hosp adm"] = "Journal of hospital administration",
  ["j hosp admit manage"] = "The Journal for hospital admitting management : official publication of the National Association of Hospital Admitting Managers",
  ["j hosp dent pract"] = "Journal of hospital dental practice",
  ["j hosp infect"] = "The Journal of hospital infection",
  ["j hosp leis sport tour educ"] = "Journal of hospitality, leisure, sport & tourism education",
  ["j hosp librariansh"] = "Journal of hospital librarianship",
  ["j hosp manag health policy"] = "Journal of hospital management and health policy",
  ["j hosp mark"] = "Journal of hospital marketing",
  ["j hosp mark public relations"] = "Journal of hospital marketing & public relations",
  ["j hosp med"] = "Journal of hospital medicine",
  ["j hosp palliat nurs"] = "Journal of hospice and palliative nursing : JHPN : the official journal of the Hospice and Palliative Nurses Association",
  ["j hosp supply process distrib"] = "Journal of hospital supply, processing, and distribution",
  ["j hotel dieu montreal"] = "Journal. Hôtel-Dieu de Montréal",
  ["j hous built  environ"] = "Journal of Housing and the Built Environment",
  ["j hous built environ"] = "Journal of housing and the built environment : HBE",
  ["j hous econ"] = "Journal of housing economics",
  ["j hous elderly"] = "Journal of housing for the elderly",
  ["j hous res"] = "Journal of Housing Research",
  ["j housing econ"] = "Journal of Housing Economics",
  ["j housing res"] = "Journal of Housing Research",
  ["j houston dist dent soc"] = "Journal. Houston District Dental Society",
  ["j huaqiao univ nat sci ed"] = "Journal of Huaqiao University",
  ["j huazhong univ sci tech"] = "Huazhong (Central China) University of Science and Technology",
  ["j huazhong univ sci technol nat sci"] = "Journal of Huazhong University of Science and Technology. Natural Science Edition. Huazhong Keji Daxue Xuebao. Ziran Kexue Ban",
  ["j huazhong univ sci technolog med sci"] = "Journal of Huazhong University of Science and Technology. Medical sciences = Hua zhong ke ji da xue xue bao. Yi xue Ying De wen ban = Huazhong keji daxue xuebao. Yixue Yingdewen ban",
  ["j hum behav soc environ"] = "Journal Of Human Behavior In The Social Environment",
  ["j hum cap"] = "Journal of human capital",
  ["j hum clin genet"] = "Journal of human & clinical genetics",
  ["j hum ecol"] = "Journal of human ecology (Delhi, India)",
  ["j hum environ syst"] = "Journal of the human-environment system",
  ["j hum ergol (tokyo)"] = "Journal of human ergology",
  ["j hum evol"] = "Journal of human evolution",
  ["j hum genet"] = "Journal of human genetics",
  ["j hum hypertens"] = "Journal of human hypertension",
  ["j hum immun"] = "Journal of Human Immunity",
  ["j hum kinet"] = "Journal of human kinetics",
  ["j hum lact"] = "Journal of human lactation : official journal of International Lactation Consultant Association",
  ["j hum nutr"] = "Journal of human nutrition",
  ["j hum nutr (carson city)"] = "Journal of human nutrition (Carson City, Nev.)",
  ["j hum nutr diet"] = "Journal of human nutrition and dietetics : the official journal of the British Dietetic Association",
  ["j hum nutr food sci"] = "Journal of human nutrition and food science",
  ["j hum reprod sci"] = "Journal of human reproductive sciences",
  ["j hum resour"] = "The Journal of human resources",
  ["j hum rights"] = "Journal of human rights",
  ["j hum rights environ"] = "Journal of human rights and the environment",
  ["j hum rights pract"] = "Journal of human rights practice",
  ["j hum rights soc work"] = "Journal of human rights and social work",
  ["j hum robot interact"] = "Journal of human-robot interaction",
  ["j hum secur"] = "Journal of human security",
  ["j hum traffick"] = "Journal of human trafficking",
  ["j hum transcr"] = "Journal of human transcriptome",
  ["j hum transcriptome"] = "Journal of Human Transcriptome",
  ["j hum virol"] = "Journal of human virology",
  ["j hum virol retrovirol"] = "Journal of human virology & retrovirology",
  ["j human dev capabil"] = "Journal of human development and capabilities",
  ["j human res"] = "Journal of Human Resources",
  ["j human stress"] = "Journal of human stress",
  ["j humanist educ dev"] = "The Journal of humanistic education and development",
  ["j humanist math"] = "Journal of Humanistic Mathematics",
  ["j humanist psychol"] = "Journal of humanistic psychology",
  ["j hunan univ nat sci"] = "Journal of Hunan University. Natural Sciences. Hunan Daxue Xuebao. Ziran Kexue Ban",
  ["j hunger environ nutr"] = "Journal of hunger & environmental nutrition",
  ["j huntingtons dis"] = "Journal of Huntington's disease",
  ["j huntington’s dis"] = "Journal of Huntington’s Disease",
  ["j hydraul eng"] = "Journal of Hydraulic Engineering",
  ["j hydraul res"] = "Journal of Hydraulic Research",
  ["j hydro-environ res"] = "Journal of Hydro-environment Research",
  ["j hydrodyn"] = "Journal of Hydrodynamics",
  ["j hydrodyn ser b"] = "Journal of Hydrodynamics, Series B",
  ["j hydrodyn ser b (english ed)"] = "Journal of Hydrodynamics",
  ["j hydrodynam b"] = "Journal of hydrodynamics. Ser. B",
  ["j hydroinf"] = "Journal of Hydroinformatics",
  ["j hydrol"] = "Journal of Hydrology",
  ["j hydrol (amst)"] = "Journal of hydrology",
  ["j hydrol eng"] = "Journal of Hydrologic Engineering",
  ["j hydrol hydromech"] = "Journal Of Hydrology And Hydromechanics",
  ["j hydrol nz"] = "Journal Of Hydrology New Zealand",
  ["j hydrol reg stud"] = "Journal of hydrology. Regional studies",
  ["j hydrol x"] = "Journal of hydrology: X",
  ["j hydrol: reg stud"] = "Journal of Hydrology: Regional Studies",
  ["j hydrometeorol"] = "Journal of hydrometeorology",
  ["j hyg (lond)"] = "The Journal of hygiene",
  ["j hyg epidemiol microbiol immunol"] = "Journal of hygiene, epidemiology, microbiology, and immunology",
  ["j hymen res"] = "Journal of Hymenoptera Research",
  ["j hymenopt res"] = "Journal of Hymenoptera research",
  ["j hyperbolic differ equ"] = "Journal of Hyperbolic Differential Equations",
  ["j hyperbolic differ equations"] = "Journal of Hyperbolic Differential Equations",
  ["j hyperstruct"] = "Journal of Hyperstructures",
  ["j hypertens"] = "Journal of Hypertension",
  ["j hypertens (los angel)"] = "Journal of hypertension : open access",
  ["j hypertens manag"] = "Journal of hypertension and management",
  ["j hypertens suppl"] = "Journal of hypertension. Supplement : official journal of the International Society of Hypertension",
  ["j iber geol"] = "Journal of Iberian Geology",
  ["j iber lat am stud"] = "Journal of Iberian and Latin American studies",
  ["j ichthyol"] = "Journal of ichthyology",
  ["j icru"] = "Journal of the ICRU",
  ["j idaho acad sci"] = "Journal of the Idaho Academy of Science",
  ["j iktiologi indones"] = "Jurnal Iktiologi Indonesia",
  ["j ill state hist soc"] = "Journal of the Illinois State Historical Society. Illinois State Historical Society",
  ["j illum eng soc"] = "Journal of the Illuminating Engineering Society",
  ["j ima"] = "The Journal of IMA",
  ["j image guid surg"] = "Journal of image guided surgery",
  ["j imaging"] = "Journal of imaging",
  ["j imaging sci"] = "Journal of imaging science",
  ["j imaging sci technol"] = "Journal of Imaging Science and Technology",
  ["j immigr health"] = "Journal of immigrant health",
  ["j immigr minor health"] = "Journal of immigrant and minority health",
  ["j immigr refug stud"] = "Journal of immigrant & refugee studies",
  ["j immune based ther vaccines"] = "Journal of immune based therapies and vaccines",
  ["j immune based ther vaccines antimicrob"] = "Journal of immune based therapies, vaccines and antimicrobials",
  ["j immune disord ther"] = "Journal of immune disorders & therapy",
  ["j immunoassay"] = "Journal of immunoassay",
  ["j immunoassay immunochem"] = "Journal of immunoassay & immunochemistry",
  ["j immunobiol"] = "Journal of immunobiology",
  ["j immunodefic disord"] = "Journal of immunodeficiency & disorders",
  ["j immunogenet"] = "Journal of immunogenetics",
  ["j immunol"] = "Journal of immunology (Baltimore, Md. : 1950)",
  ["j immunol clin res"] = "Journal of immunology and clinical research",
  ["j immunol immunotech"] = "Journal of immunology and immuno-techniques",
  ["j immunol infect dis"] = "Journal of immunology and infectious diseases",
  ["j immunol methods"] = "Journal of immunological methods",
  ["j immunol regen med"] = "Journal of immunology and regenerative medicine",
  ["j immunol regener med"] = "Journal of Immunology and Regenerative Medicine",
  ["j immunol res"] = "Journal of immunology research",
  ["j immunol res ther"] = "Journal of immunology research and therapy",
  ["j immunol sci"] = "Journal of immunological sciences",
  ["j immunol tech infect dis"] = "Journal of immunological techniques in infectious diseases",
  ["j immunopharmacol"] = "Journal of immunopharmacology",
  ["j immunother"] = "Journal of immunotherapy (Hagerstown, Md. : 1997)",
  ["j immunother (1991)"] = "Journal of immunotherapy : official journal of the Society for Biological Therapy",
  ["j immunother (1997)"] = "Journal of immunotherapy (Hagerstown, Md. : 1997)",
  ["j immunother appl"] = "Journal of immunotherapy applications",
  ["j immunother cancer"] = "Journal for immunotherapy of cancer",
  ["j immunother emphasis tumor immunol"] = "Journal of immunotherapy with emphasis on tumor immunology : official journal of the Society for Biological Therapy",
  ["j immunother precis oncol"] = "Journal of immunotherapy and precision oncology",
  ["j immunotoxicol"] = "Journal of immunotoxicology",
  ["j imp commonw hist"] = "The Journal of imperial and commonwealth history",
  ["j in vitro fert embryo transf"] = "Journal of in vitro fertilization and embryo transfer : IVF",
  ["j inborn errors metab screen"] = "Journal of inborn errors of metabolism and screening",
  ["j incl phenom macrocycl chem"] = "Journal of inclusion phenomena and macrocyclic chemistry",
  ["j inclusion phenom macrocyclic chem"] = "Journal of Inclusion Phenomena and Macrocyclic Chemistry",
  ["j inclusion phenom mol recognit chem"] = "Journal of Inclusion Phenomena and Molecular Recognition in Chemistry",
  ["j inclusion phenommacrocyclic chem"] = "Journal of Inclusion Phenomena and Macrocyclic Chemistry",
  ["j income distribution"] = "Journal of Income Distribution",
  ["j ind ecol"] = "Journal of Industrial Ecology",
  ["j ind econ"] = "The Journal of industrial economics",
  ["j ind eng chem"] = "Journal of Industrial and Engineering Chemistry",
  ["j ind eng chem (amsterdam, neth)"] = "Journal of Industrial and Engineering Chemistry (Amsterdam, Netherlands)",
  ["j ind eng chem (seoul, repub korea)"] = "Journal of Industrial and Engineering Chemistry (Seoul, Republic of Korea)",
  ["j ind eng int"] = "Journal of Industrial Engineering International",
  ["j ind hyg toxicol"] = "The Journal of industrial hygiene and toxicology",
  ["j ind inf integr"] = "Journal of Industrial Information Integration",
  ["j ind integr manage"] = "Journal of Industrial Integration and Management",
  ["j ind manag optim"] = "Journal of Industrial and Management Optimization",
  ["j ind manage optim"] = "Journal of Industrial and Management Optimization",
  ["j ind microbiol"] = "Journal of Industrial Microbiology",
  ["j ind microbiol biotechnol"] = "Journal of industrial microbiology & biotechnology",
  ["j ind pollut control"] = "Journal of Industrial Pollution Control",
  ["j ind prod eng"] = "Journal of Industrial and Production Engineering",
  ["j ind relat"] = "The Journal of industrial relations",
  ["j ind text"] = "Journal of Industrial Textiles",
  ["j indian acad dent"] = "Journal of the Indian Academy of Dentistry",
  ["j indian acad dent spec res"] = "Journal of Indian Academy of Dental Specialist Researchers",
  ["j indian acad forensic med"] = "Journal Of Indian Academy Of Forensic Medicine",
  ["j indian acad forensic sci"] = "Journal of the Indian Academy of Forensic Sciences",
  ["j indian acad math"] = "Indian Academy of Mathematics",
  ["j indian acad wood sci"] = "Journal of the Indian Academy of Wood Science",
  ["j indian anthropol soc"] = "Journal of the Indian Anthropological Society",
  ["j indian assoc commun dis"] = "Journal of Indian Association for Communicable Diseases",
  ["j indian assoc pediatr surg"] = "Journal of Indian Association of Pediatric Surgeons",
  ["j indian chem soc"] = "Journal of the Indian Chemical Society",
  ["j indian counc chem"] = "Journal of Indian Council of Chemists",
  ["j indian dent assoc"] = "Journal of the Indian Dental Association",
  ["j indian forensic sci"] = "Journal of the Indian Academy of Forensic Sciences",
  ["j indian inst sci"] = "Journal of the Indian Institute of Science",
  ["j indian law inst"] = "Journal of the Indian Law Institute",
  ["j indian leather technol assoc"] = "Journal of the Indian Leather Technologists’ Association",
  ["j indian math soc"] = "Journal of The Indian Mathematical Society",
  ["j indian math soc (ns)"] = "The Journal of the Indian Mathematical Society. New Series",
  ["j indian med assoc"] = "Journal of the Indian Medical Association",
  ["j indian med prof"] = "Journal of the Indian medical profession",
  ["j indian ocean reg"] = "Journal of the Indian Ocean Region",
  ["j indian orthod soc"] = "The Journal of Indian Orthodontic Society",
  ["j indian philos"] = "Journal of Indian philosophy",
  ["j indian prosthodont soc"] = "Journal of Indian Prosthodontic Society",
  ["j indian soc agric stat"] = "Journal of the Indian Society of Agricultural Statistics. Indian Society of Agricultural Statistics",
  ["j indian soc agricultural statist"] = "Journal of the Indian Society of Agricultural Statistics",
  ["j indian soc pedod prev dent"] = "Journal of the Indian Society of Pedodontics and Preventive Dentistry",
  ["j indian soc periodontol"] = "Journal of Indian Society of Periodontology",
  ["j indian soc probab stat"] = "Journal of the Indian Society for Probability and Statistics",
  ["j indian soc remote sens"] = "Journal of the Indian Society of Remote Sensing",
  ["j indian soc soil sci"] = "Journal of the Indian Society of Soil Science",
  ["j indian soc statist oper res"] = "Journal of Indian Society of Statistics and Operations Research",
  ["j indian statist assoc"] = "Journal of the Indian Statistical Association",
  ["j indiana dent assoc"] = "Journal (Indiana Dental Association)",
  ["j indiana state dent assoc"] = "The Journal of the Indiana State Dental Association",
  ["j indiana state med assoc"] = "The Journal of the Indiana State Medical Association",
  ["j indianap dist dent soc"] = "Journal. Indianapolis District Dental Society",
  ["j indig soc dev"] = "Journal of indigenous social development",
  ["j indig wellbeing"] = "Journal of indigenous wellbeing : Te mauri-Pimatisiwin",
  ["j individ differ"] = "Journal of individual differences",
  ["j individ psychol"] = "Journal of individual psychology",
  ["j individ psychol (1998)"] = "Journal of individual psychology (1998)",
  ["j indones math soc"] = "Journal of the Indonesian Mathematical Society",
  ["j indones nat hist"] = "Journal of Indonesian natural history",
  ["j indones trop anim agric"] = "Journal of the Indonesian tropical animal agriculture",
  ["j industry stud"] = "Journal of Industry Studies",
  ["j inequal appl"] = "Journal of Inequalities and Applications",
  ["j inequal spec funct"] = "Journal of Inequalities and Special Functions",
  ["j inf"] = "Journal of Informatics",
  ["j inf disp"] = "Journal of Information Display",
  ["j inf image manage"] = "The Journal of information and image management : JIIM",
  ["j inf intell"] = "Journal of Information and Intelligence",
  ["j inf knowl manage"] = "Journal of Information & Knowledge Management",
  ["j inf optim sci"] = "Journal of Information and Optimization Sciences",
  ["j inf process syst"] = "Journal of Information Processing Systems",
  ["j inf sci"] = "Journal of information science",
  ["j inf secur appl"] = "Journal of Information Security and Applications",
  ["j inf technol"] = "Journal of Information Technology",
  ["j inf technol case appl res"] = "Journal of Information Technology Case and Application Research",
  ["j inf technol construct"] = "Journal of Information Technology in Construction",
  ["j inf technol softw eng"] = "Journal of information technology & software engineering",
  ["j inf telecommun"] = "Journal of Information and Telecommunication",
  ["j infant child adolesc psychother"] = "Journal of infant, child, and adolescent psychotherapy : JICAP",
  ["j infect"] = "The Journal of infection",
  ["j infect (grand rapids)"] = "Journal of infectiology",
  ["j infect chemother"] = "Journal of infection and chemotherapy : official journal of the Japan Society of Chemotherapy",
  ["j infect dev ctries"] = "Journal of infection in developing countries",
  ["j infect dis"] = "The Journal of infectious diseases",
  ["j infect dis med"] = "Journal of infectious diseases and medicine",
  ["j infect dis ther"] = "Journal of infectious disease and therapy",
  ["j infect prev"] = "Journal of infection prevention",
  ["j infect public health"] = "Journal of infection and public health",
  ["j infect pulm dis"] = "Journal of infectious pulmonary diseases",
  ["j inflamm"] = "Journal of inflammation",
  ["j inflamm (lond)"] = "Journal of inflammation (London, England)",
  ["j inflamm bowel dis disord"] = "Journal of inflammatory bowel diseases & disorders",
  ["j inflamm res"] = "Journal of inflammation research",
  ["j inflammation res"] = "Journal of Inflammation Research",
  ["j infor ethics"] = "Journal of information ethics",
  ["j inform nurs"] = "Journal of informatics nursing",
  ["j inform optim sci"] = "Journal of Information & Optimization Sciences",
  ["j informetr"] = "Journal of informetrics",
  ["j infrared millim terahertz waves"] = "Journal of infrared, millimeter and terahertz waves",
  ["j infrared millimeter waves"] = "Journal of Infrared and Millimeter Waves",
  ["j infrared, millimeter, terahertz waves"] = "Journal of Infrared, Millimeter, and Terahertz Waves",
  ["j infrastruct dev"] = "Journal of Infrastructure Development",
  ["j infrastruct intell resilience"] = "Journal of Infrastructure Intelligence and Resilience",
  ["j infrastruct preserv resilience"] = "Journal of Infrastructure Preservation and Resilience",
  ["j infrastruct syst"] = "Journal of Infrastructure Systems",
  ["j infus chemother"] = "The Journal of infusional chemotherapy",
  ["j infus nurs"] = "Journal of infusion nursing : the official publication of the Infusion Nurses Society",
  ["j inherit metab dis"] = "Journal of inherited metabolic disease",
  ["j inherited metab dis"] = "Journal of Inherited Metabolic Disease",
  ["j inj violence res"] = "Journal of injury & violence research",
  ["j inn mong norm univ nat sci"] = "Journal of Inner Mongolia Normal University. Nei Mongol Shifan Daxue Xuebao. Ziran Kexue. Hanwen Ban",
  ["j inn mong univ nat sci"] = "Journal of Inner Mongolia University. Natural Science Edition",
  ["j innate immun"] = "Journal of innate immunity",
  ["j innov biol"] = "Journal of innovative biology",
  ["j innov card rhythm manag"] = "The Journal of innovations in cardiac rhythm management",
  ["j innov entrep"] = "Journal of innovation and entrepreneurship",
  ["j innov health inform"] = "Journal of innovation in health informatics",
  ["j innov opt health sci"] = "Journal of innovative optical health sciences",
  ["j innov pharm biol sci"] = "Journal of innovations in pharmaceutical and biological sciences",
  ["j innovation digital ecosyst"] = "Journal of Innovation in Digital Ecosystems",
  ["j innovation knowl"] = "Journal of Innovation and Knowledge",
  ["j innovative opt health sci"] = "Journal of Innovative Optical Health Sciences",
  ["j inorg biochem"] = "Journal of Inorganic Biochemistry",
  ["j inorg nucl chem"] = "Journal of Inorganic and Nuclear Chemistry",
  ["j inorg organomet polym"] = "Journal of Inorganic and Organometallic Polymers",
  ["j inorg organomet polym mater"] = "Journal of Inorganic and Organometallic Polymers and Materials",
  ["j insect behav"] = "Journal of insect behavior",
  ["j insect biodivers"] = "Journal of insect biodiversity",
  ["j insect biodivers syst"] = "Journal of insect biodiversity and systematics",
  ["j insect biotechnol sericology"] = "Journal of insect biotechnology and sericology",
  ["j insect conserv"] = "Journal of insect conservation",
  ["j insect physiol"] = "Journal of insect physiology",
  ["j insect sci"] = "Journal of insect science (Online)",
  ["j insects"] = "Journal of insects",
  ["j insects food feed"] = "Journal of Insects as Food and Feed",
  ["j inst actuar"] = "Journal of the Institute of Actuaries",
  ["j inst agric anim sci"] = "Journal of the Institute of Agriculture and Animal Science",
  ["j inst brew"] = "Journal of the Institute of Brewing",
  ["j inst chart account sri lanka"] = "Journal of the Institute of Chartered Accountants of Sri Lanka",
  ["j inst chem"] = "Journal of Institute of Chemists (India)",
  ["j inst chem (india)"] = "Journal of the Institution of Chemists (India)",
  ["j inst conserv"] = "Journal of the Institute for Conservation",
  ["j inst econ res"] = "Journal of Institute of Economic Research. J.S.S. Institute of Economic Research",
  ["j inst electr eng"] = "Journal of the Institution of Electrical Engineers",
  ["j inst electron radio eng"] = "Journal of the Institution of Electronic and Radio Engineers",
  ["j inst eng (india) chem eng div"] = "Journal of the Institution of Engineers (India) Chemical Engineering Division",
  ["j inst eng (india) metall mater eng div"] = "Journal of the Institution of Engineers (India) Metallurgical and Materials Engineering Division",
  ["j inst eng india ser a"] = "Journal of The Institution of Engineers (India): Series A",
  ["j inst eng india ser b"] = "Journal of The Institution of Engineers (India): Series B",
  ["j inst eng india ser c"] = "Journal of The Institution of Engineers (India): Series C",
  ["j inst eng india ser d"] = "Journal of The Institution of Engineers (India): Series D",
  ["j inst eng india ser e"] = "Journal of The Institution of Engineers (India): Series E",
  ["j inst environ sci"] = "Journal of the Institute of Environment Sciences",
  ["j inst hosp eng"] = "The Journal of the Institute of Hospital Engineering",
  ["j inst locomot eng"] = "Journal of the Institution of Locomotive Engineers",
  ["j inst math comput sci comput sci ser"] = "Institute of Mathematics & Computer Sciences. Journal. (Computer Science Series)",
  ["j inst math comput sci math ser"] = "Institute of Mathematics & Computer Sciences",
  ["j inst math jussieu"] = "Journal of the Institute of Mathematics of Jussieu",
  ["j inst med"] = "Journal of the Institute of Medicine",
  ["j inst met"] = "Journal of the Institute of Metals",
  ["j inst pet"] = "Journal of the Institute of Petroleum",
  ["j inst prod eng"] = "Journal of the Institution of Production Engineers",
  ["j inst sanit eng lond"] = "Journal. Institution of Sanitary Engineers (Great Britain)",
  ["j inst sterile serv manage"] = "Journal (Institute of Sterile Services Management)",
  ["j inst theor econ"] = "Journal of institutional and theoretical economics : JITE = Zeitschrift für die gesamte Staatswissenschaft",
  ["j inst theoretical econ"] = "Journal of Institutional and Theoretical Economics",
  ["j inst water environ manage"] = "Journal of the Institution of Water and Environment Management",
  ["j instrum"] = "Journal of Instrumentation",
  ["j instrum soc india"] = "Journal of the Instrument Society of India",
  ["j insur med"] = "Journal of insurance medicine (New York, N.Y.)",
  ["j int acad periodontol"] = "Journal of the International Academy of Periodontology",
  ["j int acad phys sci"] = "Journal of International Academy of Physical Sciences",
  ["j int adv otol"] = "The journal of international advanced otology",
  ["j int aff"] = "Journal of international affairs",
  ["j int aids soc"] = "Journal of the International AIDS Society",
  ["j int assoc dent child"] = "Journal of the International Association of Dentistry for Children",
  ["j int assoc hist psychoanal"] = "Journal / International Association For The History Of Psychoanalysis",
  ["j int assoc med sci educ"] = "The journal of the International Association of Medical Science Educators : JIAMSE",
  ["j int assoc physicians aids care"] = "Journal of the International Association of Physicians in AIDS Care",
  ["j int assoc physicians aids care (chic ill)"] = "Journal of the International Association of Physicians in AIDS Care (Chicago, Ill. : 2002)",
  ["j int assoc physicians aids care (chic)"] = "Journal of the International Association of Physicians in AIDS Care (Chicago, Ill. : 2002)",
  ["j int assoc provid aids care"] = "Journal of the International Association of Providers of AIDS Care",
  ["j int bioethique"] = "Journal international de bioéthique = International journal of bioethics",
  ["j int bioethique ethique sci"] = "Journal international de bioéthique et d'éthique des sciences",
  ["j int bus stud"] = "Journal of international business studies",
  ["j int chir"] = "Journal international de chirurgie",
  ["j int coll dent (jpn)"] = "Journal Of The International College Of Dentists; Japan Section",
  ["j int coll surg"] = "The Journal of the International College of Surgeons",
  ["j int compar econ"] = "Journal of International and Comparative Economics",
  ["j int consum mark"] = "Journal of international consumer marketing",
  ["j int coop"] = "Journal of international cooperation",
  ["j int counc electr eng"] = "Journal of International Council on Electrical Engineering",
  ["j int dev"] = "Journal of international development",
  ["j int devel"] = "Journal of International Development",
  ["j int econ"] = "Journal of international economics",
  ["j int econ law"] = "Journal of International Economic Law",
  ["j int fed clin chem"] = "Journal of the International Federation of Clinical Chemistry",
  ["j int fed gynaecol obstet"] = "Journal of the International Federation of Gynaecology and Obstetrics",
  ["j int finan markets"] = "Journal of International Financial Markets, Institutions and Money",
  ["j int law econ"] = "The Journal of international law and economics",
  ["j int life sci res"] = "Journal of International Life Sciences Research",
  ["j int marit saf environ aff ship"] = "Journal of International Maritime Safety, Environmental Affairs, and Shipping",
  ["j int math virtual inst"] = "Journal of International Mathematical Virtual Institute",
  ["j int med res"] = "The Journal of international medical research",
  ["j int migr integr"] = "Journal of international migration and integration",
  ["j int money finance"] = "Journal of international money and finance",
  ["j int neuropsychol soc"] = "Journal of the International Neuropsychological Society : JINS",
  ["j int oral health"] = "Journal of international oral health : JIOH",
  ["j int phon assoc"] = "Journal of the International Phonetic Association",
  ["j int relat dev (ljubl)"] = "Journal of international relations and development",
  ["j int res commun"] = "Journal of International Research Communications",
  ["j int soc history islamic med"] = "Journal of the International Society for the History of Islamic Medicine (ISHIM) : JISHIM",
  ["j int soc phys rehabil med"] = "The journal of the International Society of Physical and Rehabilitation Medicine",
  ["j int soc prev community dent"] = "Journal of International Society of Preventive & Community Dentistry",
  ["j int soc respir prot"] = "Journal of the International Society for Respiratory Protection",
  ["j int soc sports nutr"] = "Journal of the International Society of Sports Nutrition",
  ["j int soc telemed ehealth"] = "Journal of the International Society for Telemedicine and eHealth",
  ["j int stud dev"] = "Journal of international studies and development",
  ["j int trade econ dev"] = "The journal of international trade & economic development",
  ["j int trade econ devel"] = "Journal of International Trade and Economic Development",
  ["j int vict"] = "Journal international de victimologie",
  ["j int wildl law policy"] = "Journal of International Wildlife Law & Policy",
  ["j int wine food soc"] = "Journal of the International Wine & Food Society (London : 1989)",
  ["j int womens stud"] = "Journal of international women's studies",
  ["j integer seq"] = "Journal of integer sequences",
  ["j integr agric"] = "Journal of integrative agriculture",
  ["j integr bioinform"] = "Journal of integrative bioinformatics",
  ["j integr cardiol"] = "Journal of integrative cardiology",
  ["j integr cardiol open access"] = "Journal of integrative cardiology open access",
  ["j integr care (brighton)"] = "Journal of integrated care (Brighton, England)",
  ["j integr des process sci"] = "Journal of Integrated Design and Process Science",
  ["j integr environ sci"] = "Journal of Integrative Environmental Sciences",
  ["j integr equations appl"] = "Journal of Integral Equations and Applications",
  ["j integr med"] = "Journal of integrative medicine",
  ["j integr neurosci"] = "Journal of integrative neuroscience",
  ["j integr omics"] = "Journal of integrated OMICS",
  ["j integr oncol"] = "Journal of integrative oncology",
  ["j integr pest manag"] = "Journal of integrated pest management",
  ["j integr pest manage"] = "Journal of Integrated Pest Management",
  ["j integr plant biol"] = "Journal of integrative plant biology",
  ["j integrable syst"] = "Journal of Integrable Systems",
  ["j integral equations appl"] = "Journal of Integral Equations and Applications",
  ["j integral equations math phys"] = "Journal of Integral Equations and Mathematical Physics",
  ["j intell"] = "Journal of Intelligence",
  ["j intell connected veh"] = "Journal of Intelligent and Connected Vehicles",
  ["j intell fuzzy syst"] = "Journal of Intelligent & Fuzzy Systems",
  ["j intell hist"] = "Journal of Intelligence History",
  ["j intell inf syst"] = "Journal of intelligent information systems",
  ["j intell manuf"] = "Journal of intelligent manufacturing",
  ["j intell manuf spec equip"] = "Journal of Intelligent Manufacturing and Special Equipment",
  ["j intell mater syst struct"] = "Journal of Intelligent Material Systems and Structures",
  ["j intell med healthcare"] = "Journal of Intelligent Medicine and Healthcare",
  ["j intell rob syst"] = "Journal of Intelligent and Robotic Systems",
  ["j intell robot syst"] = "Journal of intelligent & robotic systems",
  ["j intell transp syst"] = "Journal of intelligent transportation systems",
  ["j intell transp syst technol plann oper"] = "Journal of Intelligent Transportation Systems: Technology, Planning, and Operations",
  ["j intellect dev disabil"] = "Journal of intellectual & developmental disability",
  ["j intellect disabil"] = "Journal of intellectual disabilities : JOID",
  ["j intellect disabil offending behav"] = "Journal of intellectual disabilities and offending behaviour",
  ["j intellect disabil res"] = "Journal of intellectual disability research : JIDR",
  ["j intensive care"] = "Journal of intensive care",
  ["j intensive care med"] = "Journal of intensive care medicine",
  ["j intensive care soc"] = "Journal of the Intensive Care Society",
  ["j intensive crit care"] = "Journal of intensive and critical care",
  ["j inter am stud world aff"] = "Journal of interamerican studies and world affairs",
  ["j inter des"] = "Journal of interior design",
  ["j interact res commun disord"] = "Journal of interactional research in communication disorders",
  ["j interact sci"] = "Journal of interaction science",
  ["j interact technol pedagog"] = "The journal of interactive technology and pedagogy",
  ["j interamer stud world aff"] = "Journal of Interamerican Studies and World Affairs",
  ["j interconnection networks"] = "Journal of Interconnection Networks",
  ["j intercult commun res"] = "Journal of intercultural communication research",
  ["j intercult ethnopharmacol"] = "Journal of intercultural ethnopharmacology",
  ["j intercult stud"] = "Journal of intercultural studies (Melbourne, Vic.)",
  ["j interdenominatl theol cent"] = "The journal of the Interdenominational Theological Center. Interdenominational Theological Center (Atlanta, Ga.)",
  ["j interdiscip biol"] = "Journal of interdisciplinary biology",
  ["j interdiscip hist"] = "The Journal of interdisciplinary history",
  ["j interdiscip histopathol"] = "Journal of interdisciplinary histopathology",
  ["j interdiscip math"] = "Journal of Interdisciplinary Mathematics",
  ["j interdiscip nanomed"] = "Journal of interdisciplinary nanomedicine",
  ["j interdiscipl cycle res"] = "Journal of interdisciplinary cycle research",
  ["j interdisciplinary econ"] = "Journal of Interdisciplinary Economics",
  ["j interferon cytokine res"] = "Journal of interferon & cytokine research : the official journal of the International Society for Interferon and Cytokine Research",
  ["j interferon res"] = "Journal of Interferon Research",
  ["j intergener relatsh"] = "Journal of intergenerational relationships",
  ["j intern med"] = "Journal of internal medicine",
  ["j intern med suppl"] = "Journal of internal medicine. Supplement",
  ["j internet serv appl"] = "Journal of Internet Services and Applications",
  ["j internet things"] = "Journal on Internet of Things",
  ["j interpers violence"] = "Journal of interpersonal violence",
  ["j interpolat approx sci comput"] = "Journal of Interpolation and Approximation in Scientific Computing",
  ["j interprof care"] = "Journal of interprofessional care",
  ["j interprof educ pract"] = "Journal of interprofessional education & practice",
  ["j interv card electrophysiol"] = "Journal of interventional cardiac electrophysiology : an international journal of arrhythmias and pacing",
  ["j interv cardiol"] = "Journal of interventional cardiology",
  ["j interv gastroenterol"] = "Journal of interventional gastroenterology",
  ["j interv oncol"] = "Journal of interventional oncology",
  ["j intervent cardiol"] = "Journal of Interventional Cardiology",
  ["j interventional card electrophysiol"] = "Journal of Interventional Cardiac Electrophysiology",
  ["j intraven nurs"] = "Journal of intravenous nursing : the official publication of the Intravenous Nurses Society",
  ["j invasive cardiol"] = "The Journal of invasive cardiology",
  ["j invasive fungal infect"] = "The journal of invasive fungal infections",
  ["j inverse ill posed probl"] = "Journal of inverse and ill-posed problems",
  ["j inverse ill-posed probl"] = "Journal of Inverse and Ill-Posed Problems",
  ["j invertebr pathol"] = "Journal of invertebrate pathology",
  ["j invest clin dent"] = "Journal of Investigative and Clinical Dentistry",
  ["j invest dermatol"] = "The Journal of investigative dermatology",
  ["j invest med"] = "Journal of Investigative Medicine",
  ["j invest med high impact case rep"] = "Journal of Investigative Medicine High Impact Case Reports",
  ["j invest surg"] = "Journal of investigative surgery : the official journal of the Academy of Surgical Research",
  ["j investig allergol clin immunol"] = "Journal of investigational allergology & clinical immunology",
  ["j investig clin dent"] = "Journal of investigative and clinical dentistry",
  ["j investig dermatol symp proc"] = "The journal of investigative dermatology. Symposium proceedings",
  ["j investig med"] = "Journal of investigative medicine : the official publication of the American Federation for Clinical Research",
  ["j investig med high impact case rep"] = "Journal of investigative medicine high impact case reports",
  ["j ionic liq"] = "Journal of Ionic Liquids",
  ["j iowa med soc"] = "Journal of the Iowa Medical Society",
  ["j iowa state med soc"] = "Journal. Iowa State Medical Society",
  ["j ir coll physicians surg"] = "Journal of the Irish Colleges of Physicians and Surgeons",
  ["j ir dent assoc"] = "Journal of the Irish Dental Association",
  ["j ir med assoc"] = "Journal of the Irish Medical Association",
  ["j iran chem soc"] = "Journal of the Iranian Chemical Society",
  ["j iran dent assoc"] = "Journal of the Iran Dental Association",
  ["j iran math soc"] = "Journal of the Iranian Mathematical Society",
  ["j iran stat soc (jirss)"] = "Journal of the Iranian Statistical Society. (JIRSS)",
  ["j iron steel res int"] = "Journal of Iron and Steel Research International",
  ["j irrig drain eng"] = "Journal of Irrigation and Drainage Engineering",
  ["j isakos"] = "Journal of ISAKOS : joint disorders & orthopaedic sports medicine",
  ["j isl coastal archaeol"] = "Journal of Island and Coastal Archaeology",
  ["j islam dent assoc iran"] = "Journal of Islamic Dental Association of Iran",
  ["j isr hist"] = "The journal of Israeli history",
  ["j issues intercoll athl"] = "Journal of issues in intercollegiate athletics",
  ["j istanb univ fac dent"] = "Journal of Istanbul University Faculty of Dentistry",
  ["j japan statist soc"] = "Journal of the Japan Statistical Society (Nihon Tokei Gakkai Kaiho)",
  ["j japanese int economies"] = "Journal of the Japanese and International Economies",
  ["j japanese soc comput statist"] = "Journal of the Japanese Society of Computational Statistics",
  ["j jec"] = "Journal des JEC",
  ["j jew communal serv"] = "Journal of Jewish communal service",
  ["j jew med ethics halacha"] = "A journal of Jewish medical ethics and halacha",
  ["j jiangxi norm univ nat sci ed"] = "Journal of Jiangxi Normal University",
  ["j jilin univ sci"] = "Journal of Jilin University. Science Edition. Jilin Daxue Xuebao. Lixue Ban",
  ["j jinan univ nat sci med"] = "Journal of Jinan University. Natural Science & Medicine Edition. Jinan Daxue Xuebao. Ziran Kexue yu Yixue Ban",
  ["j jishou univ nat sci ed"] = "Jishou University",
  ["j jj group hosp grant med coll"] = "Journal of J. J. Group of Hospitals and Grant Medical College",
  ["j jpn bot"] = "Journal of Japanese Botany",
  ["j jpn inst met"] = "Journal of the Japan Institute of Metals",
  ["j jpn int econ"] = "Journal of the Japanese and international economies",
  ["j jpn obstet gynecol soc"] = "Journal of the Japanese Obstetrical & Gynecological Society",
  ["j jpn pet inst"] = "Journal of the Japan Petroleum Institute",
  ["j jpn phys ther assoc"] = "Journal of the Japanese Physical Therapy Association = Rigaku ryōhō",
  ["j jpn soc powder powder metall"] = "Journal Of The Japan Society Of Powder And Powder Metallurgy",
  ["j jpn stat soc jpn issue"] = "Journal of the Japan Statistical Society. Japanese Issue",
  ["j jpn stud"] = "The Journal of Japanese studies",
  ["j juv justice"] = "Journal of juvenile justice",
  ["j kans bar assoc"] = "The Journal of the Kansas Bar Association. Kansas Bar Association",
  ["j kans city mo southwest clin soc"] = "Journal. Kansas City (Mo.) Southwest Clinical Society",
  ["j kans dent assoc"] = "Journal of the Kansas Dental Association",
  ["j kans entomol soc"] = "Journal of the Kansas Entomological Society",
  ["j kans med soc"] = "The Journal of the Kansas Medical Society",
  ["j kans state dent assoc"] = "The Journal of the Kansas State Dental Association",
  ["j kansas entomol soc"] = "Journal of the Kansas Entomological Society",
  ["j kedokt hewan (banda aceh)"] = "Jurnal kedokteran hewan",
  ["j kelaut"] = "Jurnal kelautan",
  ["j kenya chem soc"] = "Journal of the Kenya Chemical Society",
  ["j kidney"] = "Journal of kidney",
  ["j kidney cancer vhl"] = "Journal of kidney cancer and VHL",
  ["j king abdulaziz univ: islamic econ"] = "Journal of King Abdulaziz University: Islamic Economics",
  ["j king saud univ agric sci"] = "Journal of King Saud University, Agricultural Sciences",
  ["j king saud univ comput inf sci"] = "Journal of King Saud University, Computer and Information Sciences",
  ["j king saud univ eng sci"] = "Journal of King Saud University, Engineering Sciences",
  ["j king saud univ sci"] = "Journal of King Saud University. Science",
  ["j klin endokrinol stoffwechs"] = "Journal für klinische Endokrinologie und Stoffwechsel",
  ["j klin endokrinolog stoffwechsel"] = "Journal fur Klinische Endokrinologie und Stoffwechsel",
  ["j knee surg"] = "The journal of knee surgery",
  ["j knot theory ramif"] = "Journal of knot theory and its ramifications",
  ["j knot theory ramifications"] = "Journal of Knot Theory and its Ramifications",
  ["j knowl econ"] = "Journal of the Knowledge Economy",
  ["j korea inst appl supercond cryog inst"] = "Journal of the Korea Institute of Applied Superconductivity and Cryogenics",
  ["j korea soc math educ ser b pure appl math"] = "Journal of the Korea Society of Mathematical Education",
  ["j korean acad nurs"] = "Journal of Korean Academy of Nursing",
  ["j korean am med assoc"] = "Journal of Korean American Medical Association",
  ["j korean assoc oral maxillofac surg"] = "Journal of the Korean Association of Oral and Maxillofacial Surgeons",
  ["j korean astron soc"] = "Ch'ŏnmun Hakhoe chi. The journal of the Korean Astronomical Society. Han'guk Ch'ŏnmun Hakhoe",
  ["j korean ceram soc"] = "Journal of the Korean Ceramic Society",
  ["j korean chem soc"] = "Journal of the Korean Chemical Society",
  ["j korean inst met mater"] = "Journal of Korean Institute of Metals and Materials",
  ["j korean math soc"] = "Journal of the Korean Mathematical Society",
  ["j korean med sci"] = "Journal of Korean medical science",
  ["j korean nat"] = "Journal of Korean Nature",
  ["j korean neurosurg soc"] = "Journal of Korean Neurosurgical Society",
  ["j korean phys soc"] = "Journal of the Korean Physical Society",
  ["j korean res soc dent hypn"] = "Journal Of The Korean Research Society For Dental Hypnosis",
  ["j korean soc appl biol chem"] = "Journal of the Korean Society for Applied Biological Chemistry",
  ["j korean soc atmos environ"] = "Journal Of Korean Society For Atmospheric Environment",
  ["j korean soc coloproctol"] = "Journal of the Korean Society of Coloproctology",
  ["j korean soc hypertens"] = "Journal of the Korean Society of Hypertension",
  ["j korean soc ind appl math"] = "Journal of the Korean Society for Industrial and Applied Mathematics",
  ["j korean soc math educ ser b pure appl math"] = "Journal of the Korean Society of Mathematical Education. Series B. The Pure and Applied Mathematics",
  ["j korean soc math educ ser b theor math pedag math"] = "Journal of the Korean Society of Mathematical Education. Series B. Theoretical Mathematics and Pedagogical Mathematics",
  ["j korean soc microbiol"] = "The journal of the Korean Society for Microbiology",
  ["j korean stat soc"] = "Journal of the Korean Statistical Society",
  ["j korean statist soc"] = "Journal of the Korean Statistical Society",
  ["j korean surg soc"] = "Journal of the Korean Surgical Society",
  ["j kulturpflanz"] = "Journal fur Kulturpflanzen",
  ["j kulturpflanzen"] = "Journal für Kulturpflanzen = Journal of cultivated plants",
  ["j kurd stud"] = "The journal of Kurdish studies",
  ["j kuwait med assoc"] = "The Journal of the Kuwait Medical Association",
  ["j ky acad sci"] = "Journal of the Kentucky Academy of Science",
  ["j ky dent assoc"] = "Journal of the Kentucky Dental Association",
  ["j ky med assoc"] = "The Journal of the Kentucky Medical Association",
  ["j ky state med assoc"] = "The Journal of the Kentucky State Medical Association",
  ["j la dent assoc"] = "Journal of the Louisiana Dental Association",
  ["j la state med soc"] = "The Journal of the Louisiana State Medical Society : official organ of the Louisiana State Medical Society",
  ["j la state univ agric mech coll"] = "Journal. Louisiana. State University and Agricultural and Mechanical College. School of Medicine",
  ["j lab autom"] = "Journal of laboratory automation",
  ["j lab chem educ"] = "Journal of laboratory chemical education",
  ["j lab clin med"] = "Journal of Laboratory and Clinical Medicine",
  ["j lab econ"] = "Journal of Labor Economics",
  ["j lab med"] = "Journal of Laboratory Medicine",
  ["j lab physicians"] = "Journal of laboratory physicians",
  ["j lab precis med"] = "Journal of laboratory and precision medicine",
  ["j lab res"] = "Journal of Labor Research",
  ["j labelled comp radiopharm"] = "Journal of labelled compounds & radiopharmaceuticals",
  ["j labelled compd"] = "Journal of labelled compounds",
  ["j labelled compd radiopharm"] = "Journal of Labelled Compounds and Radiopharmaceuticals",
  ["j labor econ"] = "Journal of labor economics",
  ["j labor res"] = "Journal of labor research",
  ["j labour mark res"] = "Journal for labour market research",
  ["j lanc cty hist soc"] = "Journal of the Lancaster County Historical Society. Lancaster County Historical Society (Pa.)",
  ["j lancet"] = "The Journal-lancet",
  ["j land use sci"] = "Journal of land use science",
  ["j landscape archit"] = "Journal of Landscape Architecture",
  ["j landscape ecol"] = "Journal of Landscape Ecology",
  ["j landwirtsch"] = "Journal fur Landwirtschaft",
  ["j landwirtschaft"] = "Journal fur Landwirtschaft",
  ["j lang evol"] = "Journal of language evolution",
  ["j lang soc psychol"] = "Journal of language and social psychology",
  ["j lanzhou railway inst"] = "Journal of Lanzhou Railway Institute",
  ["j lanzhou univ nat sci"] = "Journal of Lanzhou University",
  ["j lao stud"] = "Journal of Lao studies",
  ["j laparoendosc adv surg tech a"] = "Journal of laparoendoscopic & advanced surgical techniques. Part A",
  ["j laparoendosc adv surg tech part b videoscop"] = "Journal of laparoendoscopic & advanced surgical techniques. Part B, Videoscopy",
  ["j laparoendosc surg"] = "Journal of laparoendoscopic surgery",
  ["j laryngol otol"] = "The Journal of laryngology and otology",
  ["j laryngol otol suppl"] = "The Journal of laryngology and otology. Supplement",
  ["j laser appl"] = "Journal of Laser Applications",
  ["j laser micro/nanoeng"] = "Journal of Laser Micro/nanoengineering",
  ["j lasers med sci"] = "Journal of lasers in medical sciences",
  ["j lasers opt photonics"] = "Journal of lasers, optics & photonics",
  ["j lat am caribb anthropol"] = "The journal of Latin American and Caribbean anthropology",
  ["j lat am geriatr med"] = "Journal of Latin American geriatric medicine",
  ["j lat am lore"] = "Journal of Latin American lore",
  ["j lat am stud"] = "Journal of Latin American studies",
  ["j lat educ"] = "Journal of Latinos and education",
  ["j lat psychol"] = "Journal of Latina/o psychology",
  ["j law"] = "Journal of Law, Economics, and Organization",
  ["j law biosci"] = "Journal of law and the biosciences",
  ["j law econ"] = "The Journal of law & economics",
  ["j law econ organ"] = "Journal of law, economics, & organization",
  ["j law enforc leadersh ethics"] = "Journal of law enforcement leadership and ethics",
  ["j law ethics dent"] = "Journal of law and ethics in dentistry",
  ["j law health"] = "Journal of law and health",
  ["j law med"] = "Journal of law and medicine",
  ["j law med ethics"] = "The Journal of law, medicine & ethics : a journal of the American Society of Law, Medicine & Ethics",
  ["j law policy"] = "Journal of law and policy",
  ["j law polit"] = "The Journal of law & politics",
  ["j law relig"] = "The Journal of law and religion",
  ["j law soc"] = "Journal of law and society",
  ["j law technol"] = "The Journal of law and technology",
  ["j leadersh organ stud"] = "Journal of leadership & organizational studies",
  ["j learn disabil"] = "Journal of learning disabilities",
  ["j leather chem technol"] = "Journal of Leather and Chemical Technology",
  ["j leather sci eng"] = "Journal of Leather Science and Engineering",
  ["j leg aff dispute resolut eng constr"] = "Journal of Legal Affairs and Dispute Resolution in Engineering and Construction",
  ["j leg med"] = "The Journal of legal medicine",
  ["j leg med (n y)"] = "The Journal of legal medicine",
  ["j legal econ"] = "Journal of Legal Economics",
  ["j legal educ"] = "Journal of legal education",
  ["j legal hist"] = "The Journal of legal history",
  ["j legal med"] = "Journal of Legal Medicine",
  ["j legal stud"] = "The Journal of legal studies",
  ["j legis"] = "Journal of legislation",
  ["j leis res"] = "Journal of leisure research",
  ["j lepid soc"] = "Journal of the Lepidopterists' Society",
  ["j lesbian stud"] = "Journal of lesbian studies",
  ["j less-common met"] = "Journal of Less-Common Metals",
  ["j leuk (los angel)"] = "Journal of leukemia (Los Angeles, Calif.)",
  ["j leukoc biol"] = "Journal of leukocyte biology",
  ["j leukoc biol suppl"] = "Journal of leukocyte biology. Supplement",
  ["j leukocyte biol"] = "Journal of Leukocyte Biology",
  ["j lgbt health res"] = "Journal of LGBT health research",
  ["j lgbt issues couns"] = "Journal of LGBT Issues in counseling",
  ["j lgbt youth"] = "Journal of LGBT youth",
  ["j libert stud"] = "The Journal of libertarian studies",
  ["j libr"] = "Journal of librarianship",
  ["j libr adm"] = "Journal of library administration",
  ["j libr hist"] = "Journal of library history (Tallahassee, Fla. : 1974)",
  ["j libr metadata"] = "Journal of library metadata",
  ["j libr sch commun"] = "Journal of librarianship and scholarly communication",
  ["j lie theory"] = "Journal of Lie Theory",
  ["j life care plan"] = "Journal of life care planning",
  ["j life sci"] = "Journal of life science",
  ["j life sci (libertyville)"] = "Journal of life sciences (Libertyville, Ill.)",
  ["j life sci (westlake village)"] = "Journal of life sciences (Westlake Village, Calif.)",
  ["j lifestyle med"] = "Journal of lifestyle medicine",
  ["j light met"] = "Journal of Light Metals",
  ["j light technol"] = "Journal of Lightwave Technology",
  ["j light visual environ"] = "Journal of Light & Visual Environment",
  ["j lightw technol"] = "IEEE/OSA Journal of Lightwave Technology",
  ["j lightwave technol"] = "Journal of Lightwave Technology",
  ["j limnol"] = "Journal of limnology",
  ["j linear topol algebra"] = "Journal of Linear and Topological Algebra",
  ["j ling geogr"] = "Journal of Linguistic Geography",
  ["j linguist"] = "Journal of linguistics",
  ["j linn soc london, bot"] = "Journal of the Linnean Society of London, Botany",
  ["j linn soc london, zool"] = "Journal of the Linnean Society of London, Zoology",
  ["j lipid atheroscler"] = "Journal of lipid and atherosclerosis",
  ["j lipid mediat"] = "Journal of lipid mediators",
  ["j lipid mediat cell signal"] = "Journal of Lipid Mediators and Cell Signalling",
  ["j lipid mediators"] = "Journal of Lipid Mediators",
  ["j lipid mediators cell signalling"] = "Journal of Lipid Mediators and Cell Signalling",
  ["j lipid res"] = "Journal of lipid research",
  ["j lipid sci technol"] = "Journal of Lipid Science and Technology",
  ["j lipids"] = "Journal of lipids",
  ["j liposome res"] = "Journal of liposome research",
  ["j liq chromatogr"] = "Journal of liquid chromatography",
  ["j liq chromatogr relat technol"] = "Journal of liquid chromatography & related technologies",
  ["j liq chromatogr related technol"] = "Journal of Liquid Chromatography and Related Technologies",
  ["j lit res"] = "Journal of literacy research : JLR",
  ["j lit sci"] = "Journal of literature and science",
  ["j lithotr stone dis"] = "The Journal of lithotripsy & stone disease",
  ["j liver"] = "Journal of liver",
  ["j liver disease transplant"] = "Journal of liver: disease & transplantation",
  ["j liver res disord ther"] = "Journal of liver research, disorders & therapy",
  ["j locat based serv"] = "Journal of Location Based Services",
  ["j log algebr methods program"] = "Journal of Logical and Algebraic Methods in Programming",
  ["j log algebr program"] = "The journal of logic and algebraic programming",
  ["j log algebraic program"] = "Journal of Logic and Algebraic Programming",
  ["j log anal"] = "Journal of Logic and Analysis",
  ["j log comput"] = "Journal of Logic and Computation",
  ["j log lang inf"] = "Journal of Logic, Language and Information",
  ["j log program"] = "Journal of Logic Programming",
  ["j logic comput"] = "Journal of Logic and Computation",
  ["j logic lang inf"] = "Journal of logic, language, and information",
  ["j logic lang inform"] = "Journal of Logic, Language and Information",
  ["j logic programming"] = "Journal of Logic Programming",
  ["j logical algebraic methods program"] = "Journal of Logical and Algebraic Methods in Programming",
  ["j lond math soc"] = "Journal of the London Mathematical Society",
  ["j lond math soc (2)"] = "Journal of the London Mathematical Society. Second Series",
  ["j london math soc"] = "Journal of the London Mathematical Society",
  ["j london math soc (2)"] = "The Journal of the London Mathematical Society",
  ["j long isl hist"] = "The Journal of Long Island history",
  ["j long term care adm"] = "The Journal of long term care administration",
  ["j long term eff med implants"] = "Journal of long-term effects of medical implants",
  ["j long term home health care"] = "The Journal of long term home health care : The PRIDE Institute journal",
  ["j long-term eff med implants"] = "Journal of Long-Term Effects of Medical Implants",
  ["j loss prev process ind"] = "Journal of Loss Prevention in the Process Industries",
  ["j loss trauma"] = "Journal of loss & trauma",
  ["j low freq noise vib act control"] = "Journal of Low Frequency Noise, Vibration, and Active Control",
  ["j low freq noise vibr act control"] = "Journal of Low Frequency Noise, Vibration and Active Control",
  ["j low genit tract dis"] = "Journal of lower genital tract disease",
  ["j low power electron"] = "Journal of Low Power Electronics",
  ["j low temp phys"] = "Journal of Low Temperature Physics",
  ["j lubr technol"] = "Journal of Lubrication Technology",
  ["j lumin"] = "Journal of Luminescence",
  ["j lung cancer"] = "Journal of lung cancer",
  ["j lung health dis"] = "Journal of lung health and diseases",
  ["j lung pulm respir res"] = "Journal of lung, pulmonary & respiratory research",
  ["j luoyang univ"] = "Luoyang University",
  ["j lymphoedema"] = "Journal of lymphoedema",
  ["j mach forming technol"] = "Journal of Machining and Forming Technologies",
  ["j mach learn"] = "Journal of Machine Learning",
  ["j mach learn model comput"] = "Journal of Machine Learning for Modeling and Computing",
  ["j mach learn res"] = "Journal of machine learning research : JMLR",
  ["j macomb dent soc"] = "Journal Of The Macomb Dental Society",
  ["j macroecon"] = "Journal of macroeconomics",
  ["j macromol sci a pure appl chem"] = "Journal of Macromolecular Science, Part A Pure and Applied Chemistry",
  ["j macromol sci b phys"] = "Journal of Macromolecular Science, Part B Physics",
  ["j macromol sci part a pure appl chem"] = "Journal of Macromolecular Science, Part A Pure and Applied Chemistry",
  ["j macromol sci part b phys"] = "Journal of Macromolecular Science, Part B Physics",
  ["j macromol sci phys"] = "Journal of macromolecular science. Physics",
  ["j macromol sci, part a: pure appl chem"] = "Journal of Macromolecular Science, Part A: Pure and Applied Chemistry",
  ["j macromol sci, part b: phys"] = "Journal of Macromolecular Science, Part B: Physics",
  ["j macromol sci, part c: polym rev"] = "Journal of Macromolecular Science, Part C: Polymer Reviews",
  ["j macromol sci, phys"] = "Journal of Macromolecular Science - Physics",
  ["j macromol sci, polym rev"] = "Journal of Macromolecular Science Polymer Reviews",
  ["j macromol sci, pure appl chem"] = "Journal of Macromolecular Science Pure and Applied Chemistry",
  ["j macromol sci, rev macromol chem phys"] = "Journal of Macromolecular Science - Reviews in Macromolecular Chemistry and Physics",
  ["j macrotrends health med"] = "The journal of macrotrends in health and medicine",
  ["j magn"] = "Journal of Magnetics",
  ["j magn magn mater"] = "Journal of Magnetism and Magnetic Materials",
  ["j magn reson"] = "Journal of Magnetic Resonance",
  ["j magn reson a"] = "Journal of magnetic resonance. Series A",
  ["j magn reson b"] = "Journal of magnetic resonance. Series B",
  ["j magn reson imaging"] = "Journal of magnetic resonance imaging : JMRI",
  ["j magn reson open"] = "Journal of Magnetic Resonance Open",
  ["j magn reson, ser a"] = "Journal of Magnetic Resonance Series A",
  ["j magn reson, ser b"] = "Journal of Magnetic Resonance Series B",
  ["j magn soc jpn"] = "Journal of the Magnetics Society of Japan",
  ["j magnesium alloys"] = "Journal of Magnesium and Alloys",
  ["j maine dent assoc"] = "Journal / Maine Dental Association",
  ["j maine med assoc"] = "The Journal of the Maine Medical Association",
  ["j maint addict"] = "Journal of maintenance in the addictions",
  ["j mal vasc"] = "Journal des maladies vasculaires",
  ["j malar inst india"] = "Journal of the Malaria institute of India",
  ["j malays branch r asiat soc"] = "Journal of the Malaysian Branch of the Royal Asiatic Society",
  ["j mamm evol"] = "Journal of mammalian evolution",
  ["j mamm ova res"] = "Journal of mammalian ova research",
  ["j mammal"] = "Journal of mammalogy",
  ["j mammal soc jpn"] = "Journal of the Mammalogical Society of Japan",
  ["j mammary gland biol neoplasia"] = "Journal of mammary gland biology and neoplasia",
  ["j man manip ther"] = "The Journal of manual & manipulative therapy",
  ["j manag care med"] = "Journal of managed care medicine : JMCM",
  ["j manag care pharm"] = "Journal of managed care pharmacy : JMCP",
  ["j manag care spec pharm"] = "Journal of managed care & specialty pharmacy",
  ["j manag inf syst"] = "Journal of management information systems : JMIS",
  ["j manag marketing healthc"] = "Journal of management & marketing in healthcare",
  ["j manag med"] = "Journal of management in medicine",
  ["j manage"] = "Journal of management",
  ["j manage acc res"] = "Journal of Management Accounting Research",
  ["j manage eng"] = "Journal of Management in Engineering",
  ["j manage governance"] = "Journal of Management and Governance",
  ["j manage sci eng"] = "Journal of Management Science and Engineering",
  ["j manage sci rep"] = "Journal of Management Scientific Reports",
  ["j manipulative physiol ther"] = "Journal of manipulative and physiological therapeutics",
  ["j manuf mater process"] = "Journal Of Manufacturing And Materials Processing",
  ["j manuf process"] = "Journal of manufacturing processes",
  ["j manuf processes"] = "Journal of Manufacturing Processes",
  ["j manuf sci eng"] = "Journal of Manufacturing Science and Engineering",
  ["j manuf sci prod"] = "Journal of Manufacturing Science and Production",
  ["j manuf syst"] = "Journal of Manufacturing Systems",
  ["j manuf technol manage"] = "Journal of Manufacturing Technology Management",
  ["j manuf technol res"] = "Journal of Manufacturing Technology Research",
  ["j map geogr lib"] = "Journal of Map & Geography Libraries",
  ["j map geogr libr"] = "Journal of map & geography libraries",
  ["j maps"] = "Journal of maps",
  ["j mar biol"] = "Journal of marine biology",
  ["j mar biol ass uk"] = "Journal of the Marine Biology Association of the United Kingdom",
  ["j mar biol assoc u k"] = "Journal of the Marine Biological Association of the UK",
  ["j mar biotechnol"] = "Journal of marine biotechnology",
  ["j mar eng technol"] = "Journal of Marine Engineering & Technology",
  ["j mar isl cult"] = "Journal of Marine and Island Cultures",
  ["j mar res"] = "Journal of Marine Research",
  ["j mar sci"] = "Journal of Marine Sciences",
  ["j mar sci appl"] = "Journal of Marine Science and Application",
  ["j mar sci eng"] = "Journal of marine science and engineering",
  ["j mar sci technol"] = "Journal of Marine Science and Technology",
  ["j mar sys"] = "Journal of Marine Systems",
  ["j mar syst"] = "Journal of Marine Systems",
  ["j mar technol soc"] = "Journal of the Marine Technology Society",
  ["j marit archaeol"] = "Journal of Maritime Archaeology",
  ["j marit res"] = "Journal for maritime research",
  ["j marital fam ther"] = "Journal of marital and family therapy",
  ["j mark"] = "Journal of marketing",
  ["j mark access health policy"] = "Journal of market access & health policy",
  ["j mark ment health"] = "Journal of marketing for mental health",
  ["j mark prof"] = "Journal of marketing for professions",
  ["j mark res"] = "JMR, Journal of marketing research",
  ["j mark res soc"] = "Journal of the Market Research Society. Market Research Society",
  ["j marketing"] = "Journal of Marketing",
  ["j marketing res"] = "Journal of Marketing Research",
  ["j markets morality"] = "Journal of Markets and Morality",
  ["j marmara univ dent fac"] = "Journal of Marmara University Dental Faculty",
  ["j marriage fam"] = "Journal of marriage and the family",
  ["j marriage fam couns"] = "Journal of marriage and family counseling",
  ["j mass assoc boards health"] = "Journal. Massachusetts Association of Boards of Health",
  ["j mass dent soc"] = "Journal of the Massachusetts Dental Society",
  ["j mass spectrom"] = "Journal of Mass Spectrometry",
  ["j mass spectrom adv clin lab"] = "Journal of Mass Spectrometry and Advances in the Clinical Lab",
  ["j mass spectrom soc jpn"] = "Journal of the Mass Spectrometry Society of Japan",
  ["j mater"] = "Journal of materials",
  ["j mater (n y, ny, u s)"] = "Journal of Materials (New York, NY, United States)",
  ["j mater chem"] = "Journal of Materials Chemistry",
  ["j mater chem\ta"] = "Journal of Materials Chemistry A",
  ["j mater chem a"] = "Journal of Materials Chemistry A: Materials for Energy and Sustainability",
  ["j mater chem a mater"] = "Journal of materials chemistry. A",
  ["j mater chem b"] = "Journal of materials chemistry. B",
  ["j mater chem c"] = "Journal of Materials Chemistry C",
  ["j mater chem c mater"] = "Journal of materials chemistry. C",
  ["j mater chema"] = "Journal of Materials Chemistry A",
  ["j mater civ eng"] = "Journal of Materials in Civil Engineering",
  ["j mater cycles waste manag"] = "Journal of material cycles and waste management",
  ["j mater cycles waste manage"] = "Journal of Material Cycles and Waste Management",
  ["j mater educ"] = "Journal of Materials Education",
  ["j mater energy syst"] = "Journal of Materials for Energy Systems",
  ["j mater eng"] = "Journal of Materials Engineering",
  ["j mater eng perform"] = "Journal of Materials Engineering and Performance",
  ["j mater eng struct"] = "Journal of Materials and Engineering Structures",
  ["j mater environ sci"] = "Journal of Materials and Environmental Science",
  ["j mater process manuf sci"] = "Journal of Materials Processing and Manufacturing Science",
  ["j mater process technol"] = "Journal of Materials Processing Technology",
  ["j mater res"] = "Journal of Materials Research",
  ["j mater res technol"] = "Journal of Materials Research and Technology",
  ["j mater sci"] = "Journal of Materials Science",
  ["j mater sci - mater electron"] = "Journal of Materials Science - Materials in Electronics",
  ["j mater sci - mater med"] = "Journal of Materials Science Materials in Medicine",
  ["j mater sci eng"] = "Journal of Materials Science and Engineering",
  ["j mater sci eng a"] = "Journal of materials science and engineering. A",
  ["j mater sci lett"] = "Journal of materials science letters",
  ["j mater sci mater electron"] = "Journal of Materials Science: Materials in Electronics",
  ["j mater sci mater med"] = "Journal of materials science. Materials in medicine",
  ["j mater sci technol"] = "Journal of Materials Science and Technology",
  ["j mater sci: mater electron"] = "Journal of Materials Science: Materials in Electronics",
  ["j mater sci: mater med"] = "Journal of Materials Science: Materials in Medicine",
  ["j mater shaping technol"] = "Journal of Materials Shaping Technology",
  ["j mater synth process"] = "Journal of Materials Synthesis and Processing",
  ["j materiomics"] = "Journal of Materiomics",
  ["j matern fetal med"] = "The Journal of maternal-fetal medicine",
  ["j matern fetal neonatal med"] = "The journal of maternal-fetal & neonatal medicine : the official journal of the European Association of Perinatal Medicine, the Federation of Asia and Oceania Perinatal Societies, the International Society of Perinatal Obstetricians",
  ["j math"] = "Journal of Mathematics",
  ["j math (wuhan)"] = "Journal of Mathematics",
  ["j math anal"] = "Journal of Mathematical Analysis",
  ["j math anal appl"] = "Journal of Mathematical Analysis and Applications",
  ["j math appl"] = "Journal of Mathematics and Applications",
  ["j math arts"] = "Journal of Mathematics and the Arts",
  ["j math behav"] = "Journal of Mathematical Behavior",
  ["j math biol"] = "Journal of Mathematical Biology",
  ["j math chem"] = "Journal of Mathematical Chemistry",
  ["j math cryptol"] = "Journal of Mathematical Cryptology",
  ["j math econ"] = "Journal of Mathematical Economics",
  ["j math econom"] = "Journal of Mathematical Economics",
  ["j math fluid mech"] = "Journal of Mathematical Fluid Mechanics",
  ["j math fundam sci"] = "Journal of Mathematical and Fundamental Sciences",
  ["j math imaging vis"] = "Journal of mathematical imaging and vision",
  ["j math imaging vision"] = "Journal of Mathematical Imaging and Vision",
  ["j math ind"] = "Journal of Mathematics in Industry",
  ["j math inequal"] = "Journal of Mathematical Inequalities",
  ["j math kyoto univ"] = "Journal of Mathematics of Kyoto University",
  ["j math log"] = "Journal of Mathematical Logic",
  ["j math logic"] = "Journal of Mathematical Logic",
  ["j math model"] = "Journal of Mathematical Modeling",
  ["j math modell algorithms oper res"] = "Journal of Mathematical Modelling and Algorithms in Operations Research",
  ["j math music"] = "Journal of Mathematics and Music",
  ["j math neurosci"] = "Journal of mathematical neuroscience",
  ["j math phys"] = "Journal of Mathematical Physics",
  ["j math phys (ny)"] = "Journal of Mathematical Physics (New York)",
  ["j math phys anal geom"] = "Journal of Mathematical Physics Analysis Geometry",
  ["j math phys sci"] = "Journal of Mathematical and Physical Sciences",
  ["j math psych"] = "Journal of Mathematical Psychology",
  ["j math psychol"] = "Journal of mathematical psychology",
  ["j math pures appl"] = "Journal de mathématiques pures et appliquées",
  ["j math pures appl (9)"] = "Journal de Mathematiques Pures et Appliquees. Neuvieme Serie",
  ["j math res appl"] = "Journal of Mathematical Research with Applications",
  ["j math res exposition"] = "Journal of Mathematical Research and Exposition",
  ["j math sci"] = "Journal of Mathematical Sciences",
  ["j math sci (calcutta)"] = "Journal of Mathematical Sciences (Calcutta)",
  ["j math sci (dattapukur)"] = "Journal of Mathematical Sciences (Dattapukur)",
  ["j math sci (new york)"] = "Journal of Mathematical Sciences (New York)",
  ["j math sci (ny)"] = "Journal of Mathematical Sciences (New York)",
  ["j math sci adv appl"] = "Journal of Mathematical Sciences. Advances and Applications",
  ["j math sci univ tokyo"] = "The University of Tokyo. Journal of Mathematical Sciences",
  ["j math soc japan"] = "Journal of the Mathematical Society of Japan",
  ["j math soc jpn"] = "Journal of the Mathematical Society of Japan",
  ["j math sociol"] = "The Journal of mathematical sociology",
  ["j math stat"] = "Journal of mathematics and statistics",
  ["j math study"] = "Journal of Mathematical Study. Shuxue Yanjiu",
  ["j math systems estim control"] = "Journal of Mathematical Systems, Estimation, and Control",
  ["j math teach educ"] = "Journal of Mathematics Teacher Education",
  ["j math tokushima univ"] = "Journal of Mathematics. Tokushima University",
  ["j math-didakt"] = "Journal fur Mathematik-Didaktik",
  ["j maulana azad college tech"] = "Maulana Azad College of Technology",
  ["j maxillofac oral surg"] = "Journal of maxillofacial and oral surgery",
  ["j maxillofac orthop"] = "Journal of maxillofacial orthopedics",
  ["j maxillofac surg"] = "Journal of maxillofacial surgery",
  ["j maxillofac trauma"] = "Journal of maxillofacial trauma",
  ["j md state dent assoc"] = "Journal of the Maryland State Dental Association",
  ["j meas phys behav"] = "Journal for the measurement of physical behaviour",
  ["j mech"] = "Journal of Mechanics",
  ["j mech behav biomed mater"] = "Journal of the mechanical behavior of biomedical materials",
  ["j mech behav mater"] = "Journal of the mechanical behavior of materials",
  ["j mech des"] = "Journal of Mechanical Design",
  ["j mech des n y"] = "Journal of mechanical design (New York, N.Y. : 1990)",
  ["j mech eng sci"] = "Journal of Mechanical Engineering Science",
  ["j mech mater struct"] = "Journal of Mechanics of Materials and Structures",
  ["j mech med biol"] = "Journal of mechanics in medicine and biology",
  ["j mech phys solids"] = "Journal of the Mechanics and Physics of Solids",
  ["j mech rob"] = "Journal of Mechanisms and Robotics",
  ["j mech robot"] = "Journal of mechanisms and robotics",
  ["j mech sci technol"] = "Journal of Mechanical Science and Technology",
  ["j mech work technol"] = "Journal of Mechanical Working Technology",
  ["j mechanochem cell motil"] = "Journal of mechanochemistry & cell motility",
  ["j med"] = "Journal of medicine",
  ["j med (oporto)"] = "Jornal do médico",
  ["j med artif intell"] = "Journal of medical artificial intelligence",
  ["j med assoc eire"] = "Journal. Medical Association of Eire",
  ["j med assoc ga"] = "Journal of the Medical Association of Georgia",
  ["j med assoc prev war"] = "Journal of the Medical Association for Prevention of War",
  ["j med assoc state ala"] = "Journal of the Medical Association of the State of Alabama",
  ["j med assoc thai"] = "Journal of the Medical Association of Thailand = Chotmaihet thangphaet",
  ["j med bacteriol"] = "Journal of medical bacteriology",
  ["j med besancon"] = "Journal de médecine de Besançon",
  ["j med biochem"] = "Journal of medical biochemistry",
  ["j med bioeng"] = "Journal of medical and bioengineering",
  ["j med biogr"] = "Journal of medical biography",
  ["j med biol eng"] = "Journal of medical and biological engineering",
  ["j med bord"] = "Journal de médecine de Bordeaux et du Sud-Ouest",
  ["j med caen"] = "Journal de médecine de Caen",
  ["j med case rep"] = "Journal of medical case reports",
  ["j med case rep rev"] = "Journal of medical case reports and reviews",
  ["j med case reports"] = "Journal of medical case reports",
  ["j med cases"] = "Journal of medical cases",
  ["j med chem"] = "Journal of Medicinal Chemistry",
  ["j med chem biol radiol def"] = "The journal of medical, chemical, biological, and radiological defense",
  ["j med chem drug des"] = "Journal of medicinal chemistry and drug design",
  ["j med chir prat"] = "Journal de médecine et de chirurgie pratiques",
  ["j med cuneif"] = "Le journal des médecines cunéiformes",
  ["j med dent sci"] = "Journal of medical and dental sciences",
  ["j med device"] = "Journal of medical devices",
  ["j med devices"] = "Journal of Medical Devices",
  ["j med diagn"] = "Journal of Medical Diagnostics",
  ["j med diagn methods"] = "Journal of medical diagnostic methods",
  ["j med econ"] = "Journal of medical economics",
  ["j med educ"] = "Journal of medical education",
  ["j med educ curric dev"] = "Journal of medical education and curricular development",
  ["j med educ res"] = "Journal of Medical Education Research",
  ["j med educ train"] = "Journal of medical education and training",
  ["j med eng"] = "Journal of medical engineering",
  ["j med eng technol"] = "Journal of medical engineering & technology",
  ["j med entomol"] = "Journal of medical entomology",
  ["j med entomol suppl"] = "Journal of medical entomology. Supplement",
  ["j med ethics"] = "Journal of medical ethics",
  ["j med ethics hist med"] = "Journal of medical ethics and history of medicine",
  ["j med ext reality"] = "Journal of Medical Extended Reality",
  ["j med food"] = "Journal of medicinal food",
  ["j med genet"] = "Journal of medical genetics",
  ["j med humanit"] = "The Journal of medical humanities",
  ["j med humanit bioeth"] = "The Journal of medical humanities and bioethics",
  ["j med hypotheses ideas"] = "Journal of Medical Hypotheses & Ideas",
  ["j med imaging"] = "Journal of Medical Imaging",
  ["j med imaging (bellingham)"] = "Journal of medical imaging (Bellingham, Wash.)",
  ["j med imaging health inf"] = "Journal of Medical Imaging and Health Informatics",
  ["j med imaging health inform"] = "Journal of medical imaging and health informatics",
  ["j med imaging radiat oncol"] = "Journal of medical imaging and radiation oncology",
  ["j med imaging radiat sci"] = "Journal of medical imaging and radiation sciences",
  ["j med internet res"] = "Journal of medical Internet research",
  ["j med invest"] = "The journal of medical investigation : JMI",
  ["j med lab technol"] = "The Journal of medical laboratory technology",
  ["j med law"] = "Journal of medicine and law",
  ["j med law ethics"] = "Journal of medical law and ethics",
  ["j med leg droit med"] = "Journal de médecine légale droit médical",
  ["j med liban"] = "Le Journal médical libanais. The Lebanese medical journal",
  ["j med libr assoc"] = "Journal of the Medical Library Association : JMLA",
  ["j med life"] = "Journal of medicine and life",
  ["j med lyon"] = "Journal de médecine de Lyon",
  ["j med microbiol"] = "Journal of medical microbiology",
  ["j med microbiol diagn"] = "Journal of medical microbiology & diagnosis",
  ["j med nantes"] = "Journal de médecine de Nantes",
  ["j med nucl biophys"] = "Journal de Medecine Nucleaire et Biophysique",
  ["j med oncol ther"] = "Journal of medical oncology and therapeutics",
  ["j med paris"] = "Journal de médecine de Paris",
  ["j med pernamb"] = "Jornal de medicina de Pernambuco",
  ["j med person"] = "Journal of medicine and the person",
  ["j med pest control"] = "Journal of Medical Pest Control",
  ["j med pharm chem"] = "Journal of medicinal and pharmaceutical chemistry",
  ["j med philos"] = "The Journal of medicine and philosophy",
  ["j med phys"] = "Journal of medical physics",
  ["j med plant res"] = "Journal of medicinal plant research",
  ["j med plants econ dev"] = "Journal of Medicinal Plants for Economic Development",
  ["j med pract manage"] = "The Journal of medical practice management : MPM",
  ["j med primatol"] = "Journal of medical primatology",
  ["j med radiat sci"] = "Journal of medical radiation sciences",
  ["j med regul"] = "Journal of medical regulation",
  ["j med res"] = "The Journal of medical research",
  ["j med res innov"] = "Journal of medical research and innovation",
  ["j med rob res"] = "Journal of Medical Robotics Research",
  ["j med robot res"] = "Journal of medical robotics research",
  ["j med sci"] = "Journal of medical science",
  ["j med sci (faisalabad, pakistan)"] = "Journal of medical sciences (Faisalābād, Pakistan)",
  ["j med screen"] = "Journal of medical screening",
  ["j med screening"] = "Journal of Medical Screening",
  ["j med signals sens"] = "Journal of medical signals and sensors",
  ["j med soc n j"] = "The Journal of the Medical Society of New Jersey",
  ["j med speech lang pathol"] = "Journal of medical speech-language pathology",
  ["j med stat inform"] = "Journal of medical statistics and informatics",
  ["j med strasb"] = "Journal de médecine de Strasbourg",
  ["j med surg pathol"] = "Journal of medical & surgical pathology",
  ["j med syst"] = "Journal of medical systems",
  ["j med technol"] = "Journal of medical technology : official publication of American Medical Technologists [and] American Society for Medical Technology",
  ["j med toxicol"] = "Journal of medical toxicology : official journal of the American College of Medical Toxicology",
  ["j med trop"] = "Journal of medicine in the tropics",
  ["j med ultrason"] = "Journal of Medical Ultrasonics",
  ["j med ultrason (2001)"] = "Journal of medical ultrasonics (2001)",
  ["j med ultrasound"] = "Journal of medical ultrasound",
  ["j med vasc"] = "Journal de médecine vasculaire",
  ["j med vet mycol"] = "Journal of medical and veterinary mycology : bi-monthly publication of the International Society for Human and Animal Mycology",
  ["j med virol"] = "Journal of medical virology",
  ["j media bus stud"] = "Journal of media business studies",
  ["j media lit educ"] = "The journal of media literacy education",
  ["j media psychol"] = "Journal of media psychology",
  ["j medicaid manage"] = "Journal for Medicaid management",
  ["j mediev hist"] = "Journal of medieval history",
  ["j mediev iber stud"] = "Journal of medieval Iberian studies",
  ["j mediev renaiss stud"] = "The journal of medieval and Renaissance studies",
  ["j medieval hist"] = "Journal of Medieval History",
  ["j mediterr med"] = "Journal méditerranée médical",
  ["j mega  infrastruct sustainable dev"] = "Journal of Mega Infrastructure & Sustainable Development",
  ["j melittology"] = "Journal of melittology",
  ["j mem lang"] = "Journal of memory and language",
  ["j memb sci"] = "Journal of membrane science",
  ["j membr biol"] = "Journal of Membrane Biology",
  ["j membr comput"] = "Journal of Membrane Computing",
  ["j membr sci"] = "Journal of Membrane Science",
  ["j membr sci lett"] = "Journal of Membrane Science Letters",
  ["j membr sci technol"] = "Journal of membrane science & technology",
  ["j membr sep technol"] = "Journal of Membrane and Separation Technology",
  ["j meningitis"] = "Journal of meningitis",
  ["j menopausal med"] = "Journal of menopausal medicine",
  ["j mens health"] = "Journal of men's health",
  ["j mens health gend"] = "The journal of men's health & gender : the official journal of the International Society for Men's Health & Gender",
  ["j mens stud"] = "The Journal of men's studies",
  ["j ment defic res"] = "Journal of mental deficiency research",
  ["j ment defic res [monogr ser]"] = "Journal of mental deficiency research. Monograph series",
  ["j ment health"] = "Journal of mental health (Abingdon, England)",
  ["j ment health adm"] = "Journal of mental health administration",
  ["j ment health aging"] = "Journal of mental health and aging",
  ["j ment health clin psychol"] = "Journal of mental health & clinical psychology",
  ["j ment health couns"] = "Journal of mental health counseling",
  ["j ment health policy econ"] = "The journal of mental health policy and economics",
  ["j ment health res intellect disabil"] = "Journal of mental health research in intellectual disabilities",
  ["j ment sci"] = "The Journal of mental science",
  ["j ment subnorm"] = "The Journal of mental subnormality",
  ["j mercer dent soc"] = "Journal. Mercer Dental Society, Trenton",
  ["j met mater miner"] = "Journal Of Metals, Materials and Minerals",
  ["j metab syndr"] = "Journal of Metabolic Syndrome",
  ["j metabolomics"] = "Journal of metabolomics",
  ["j metabolomics syst biol"] = "Journal of metabolomics and systems biology",
  ["j metabonomics metab"] = "Journal of metabonomics & metabolites",
  ["j metall mater sci"] = "Journal of Metallurgy and Materials Science",
  ["j metamorph geol"] = "Journal of Metamorphic Geology",
  ["j metastable nanocryst mater"] = "Journal of Metastable and Nanocrystalline Materials",
  ["j meteorol"] = "Journal of Meteorology",
  ["j meteorol res japan"] = "Journal of Meteorological Research, Japan",
  ["j meteorol soc jpn"] = "Journal of the Meteorological Society of Japan",
  ["j meteorolog res"] = "Journal of Meteorological Research",
  ["j meteorolog soc jpn"] = "Journal of the Meteorological Society of Japan",
  ["j methods meas soc sci"] = "Journal of methods and measurement in the social sciences",
  ["j mex am hist"] = "The Journal of Mexican American history",
  ["j mex chem soc"] = "Journal of the Mexican Chemical Society",
  ["j mich dent assoc"] = "The Journal of the Michigan Dental Association",
  ["j mich state dent assoc"] = "The Journal of the Michigan State Dental Association",
  ["j mich state dent soc"] = "Journal. Michigan State Dental Society",
  ["j mich state med soc"] = "Journal - Michigan State Medical Society",
  ["j micro - nano mechatron"] = "Journal of Micro - Nano Mechatronics",
  ["j micro nano-manuf"] = "Journal of Micro and Nano-Manufacturing",
  ["j micro nanolithogr mems moems"] = "Journal of micro/nanolithography, MEMS, and MOEMS : JM3",
  ["j micro nanomanuf"] = "Journal of micro- and nano-manufacturing",
  ["j micro-bio rob"] = "Journal of Micro-Bio Robotics",
  ["j micro/nanolithog microfab microsys"] = "Journal of Micro/Nanolithography, Microfabrican, and Microsystems",
  ["j micro/nanolithogr mems moems"] = "Journal of Micro/Nanolithography, MEMS and MOEMS",
  ["j micro/nanopatterning mater metrol"] = "Journal Of Micro/Nanopatterning Materials And Metrology",
  ["j microb biochem technol"] = "Journal of microbial & biochemical technology",
  ["j microbio robot"] = "Journal of micro-bio robotics",
  ["j microbiol"] = "Journal of microbiology (Seoul, Korea)",
  ["j microbiol antimicrob"] = "Journal of microbiology and antimicrobials",
  ["j microbiol biol educ"] = "Journal of microbiology & biology education",
  ["j microbiol biotechnol"] = "Journal of microbiology and biotechnology",
  ["j microbiol biotechnol food sci"] = "The journal of microbiology, biotechnology and food sciences",
  ["j microbiol biotechnol res"] = "Journal of microbiology and biotechnology research",
  ["j microbiol epidemiol immunobiol"] = "Journal of microbiology, epidemiology and immunobiology. Zhurnal mikrobiologii, epidemiologii i immunobiologii",
  ["j microbiol exp"] = "Journal of microbiology & experimentation",
  ["j microbiol immunol infect"] = "Journal of microbiology, immunology, and infection = Wei mian yu gan ran za zhi",
  ["j microbiol methods"] = "Journal of microbiological methods",
  ["j microbiol res (rosemead, calif)"] = "Journal of microbiology research",
  ["j microcolumn sep"] = "The journal of microcolumn separations : JMS",
  ["j microelectromech syst"] = "Journal of microelectromechanical systems : a joint IEEE and ASME publication on microstructures, microactuators, microsensors, and microsystems",
  ["j microelectron electron packag"] = "Journal of Microelectronics and Electronic Packaging",
  ["j microencapsul"] = "Journal of microencapsulation",
  ["j microencapsulation"] = "Journal of Microencapsulation",
  ["j microgr"] = "The Journal of micrographics",
  ["j microlith microfab microsys"] = "Journal of Microlithography, Microfabrication, and Microsystems",
  ["j micromanuf"] = "Journal of Micromanufacturing",
  ["j micromech microeng"] = "Journal of Micromechanics and Microengineering",
  ["j micromech mol phys"] = "Journal of micromechanics and molecular physics",
  ["j micromechanics microengineering"] = "Journal of Micromechanics and Microengineering",
  ["j micropalaeontol"] = "Journal of micropalaeontology",
  ["j micropaleontol"] = "Journal of Micropaleontology",
  ["j microsc"] = "Journal of microscopy",
  ["j microsc (paris)"] = "Journal de microscopie",
  ["j microsc biol cell"] = "Journal de microscopie et de biologie cellulaire",
  ["j microsc ultrastruct"] = "Journal of microscopy and ultrastructure",
  ["j microsurg"] = "Journal of microsurgery",
  ["j microw power"] = "The Journal of microwave power",
  ["j microw power electromagn energy"] = "The Journal of microwave power and electromagnetic energy : a publication of the International Microwave Power Institute",
  ["j microwave power"] = "Journal of Microwave Power",
  ["j microwave power electromagn energy"] = "Journal of Microwave Power and Electromagnetic Energy",
  ["j middle east afr"] = "Journal of the Middle East and Africa",
  ["j midlife health"] = "Journal of mid-life health",
  ["j midwifery womens health"] = "Journal of midwifery & women's health",
  ["j migr health"] = "Journal of migration and health",
  ["j migr hist"] = "Journal of migration history",
  ["j migr hum secur"] = "Journal on migration and human security",
  ["j migr refug issues"] = "Journal of migration and refugee issues",
  ["j mikol indones"] = "Jurnal mikologi Indonesia",
  ["j mil gov couns"] = "Journal of military and government counseling",
  ["j mil hist"] = "The Journal of military history",
  ["j mil med pac"] = "The journal of military medicine in the Pacific",
  ["j mil veteran fam health"] = "Journal of military, veteran and family health",
  ["j mil veterans health"] = "Journal of military and veterans' health",
  ["j milk food technol"] = "Journal of Milk and Food Technology",
  ["j min environ"] = "Journal Of Mining And Environment",
  ["j min inst"] = "Journal Of Mining Institute",
  ["j min mater processinst jpn"] = "Journal of the Mining and Materials Processing Institute of Japan",
  ["j min metall sect a"] = "Journal of Mining and Metallurgy, Section A: Mining",
  ["j min metall sect b"] = "Journal of Mining and Metallurgy, Section B: Metallurgy",
  ["j min sci"] = "Journal of Mining Science",
  ["j min sci technol"] = "Journal Of Mining Science And Technology",
  ["j mine vent soc s afr"] = "Journal of the Mine Ventilation Society of South Africa. Mine Ventilation Society of South Africa",
  ["j mineral petrol sci"] = "Journal of Mineralogical and Petrological Sciences",
  ["j mines met fuels"] = "Journal of Mines, Metals and Fuels",
  ["j minim access surg"] = "Journal of minimal access surgery",
  ["j minim invasive gynecol"] = "Journal of minimally invasive gynecology",
  ["j minn acad sci"] = "Journal of the Minnesota Academy of Science",
  ["j miss acad sci"] = "Journal of the Mississippi Academy of Sciences. Mississippi Academy of Sciences",
  ["j miss dent assoc"] = "Journal - Mississippi Dental Association",
  ["j miss hist"] = "The Journal of Mississippi history",
  ["j miss state med assoc"] = "Journal of the Mississippi State Medical Association",
  ["j mitochondria, plastids endosymbiosis"] = "Journal of Mitochondria, Plastids and Endosymbiosis",
  ["j mix methods res"] = "Journal of mixed methods research",
  ["j mixed methods res"] = "Journal of Mixed Methods Research",
  ["j mmij"] = "Journal of MMIJ",
  ["j mo bar"] = "Journal of the Missouri Bar",
  ["j mo dent assoc"] = "Journal of the Missouri Dental Association",
  ["j mo state dent assoc"] = "Journal. Missouri State Dental Association",
  ["j mob technol med"] = "Journal of mobile technology in medicine",
  ["j mod afr stud"] = "The Journal of modern African studies",
  ["j mod appl stat methods"] = "Journal of modern applied statistical methods : JMASM",
  ["j mod dyn"] = "Journal of Modern Dynamics",
  ["j mod eur hist"] = "Journal of modern European history = Zeitschrift für moderne europäische Geschichte = Revue d'histoire européenne contemporaine",
  ["j mod hist"] = "The Journal of modern history",
  ["j mod ital stud"] = "Journal of modern Italian studies",
  ["j mod jew stud"] = "Journal of modern Jewish studies",
  ["j mod math stat"] = "Journal of Modern Mathematics and Statistics",
  ["j mod methods numer math"] = "Journal of Modern Methods in Numerical Mathematics",
  ["j mod opt"] = "Journal of Modern Optics",
  ["j mod physiol res"] = "Journal of modern physiological research",
  ["j mod power syst clean energy"] = "Journal of Modern Power Systems and Clean Energy",
  ["j model ophthalmol"] = "Journal for modeling in ophthalmology",
  ["j modell manage"] = "Journal of Modelling in Management",
  ["j modern opt"] = "Journal of Modern Optics",
  ["j mol appl genet"] = "Journal of molecular and applied genetics",
  ["j mol biochem"] = "Journal of molecular biochemistry",
  ["j mol biol"] = "Journal of Molecular Biology",
  ["j mol biol methods"] = "Journal of molecular biology and methods",
  ["j mol biol mol imaging"] = "Journal of molecular biology and molecular imaging",
  ["j mol biol res"] = "Journal of molecular biology research",
  ["j mol biol ther"] = "Journal of molecular biology & therapeutics",
  ["j mol biomark diagn"] = "Journal of molecular biomarkers & diagnosis",
  ["j mol catal"] = "Journal of Molecular Catalysis",
  ["j mol catal a chem"] = "Journal of molecular catalysis. A, Chemical",
  ["j mol catal a: chem"] = "Journal of Molecular Catalysis A: Chemical",
  ["j mol catal b enzym"] = "Journal of molecular catalysis. B, Enzymatic",
  ["j mol catal b: enzym"] = "Journal of Molecular Catalysis B: Enzymatic",
  ["j mol cell biol"] = "Journal of molecular cell biology",
  ["j mol cell cardiol"] = "Journal of molecular and cellular cardiology",
  ["j mol cell immunol"] = "The Journal of molecular and cellular immunology : JMCI",
  ["j mol cell pathol"] = "Journal of Molecular and Cellular Pathology",
  ["j mol cloning genet recomb"] = "Journal of molecular cloning & genetic recombination",
  ["j mol diagn"] = "The Journal of molecular diagnostics : JMD",
  ["j mol endocrinol"] = "Journal of molecular endocrinology",
  ["j mol eng mater"] = "Journal of molecular and engineering materials",
  ["j mol evol"] = "Journal of molecular evolution",
  ["j mol genet"] = "Journal of molecular genetics",
  ["j mol genet (isleworth)"] = "Journal of molecular genetics (Isleworth, London, England)",
  ["j mol genet med"] = "Journal of Molecular and Genetic Medicine",
  ["j mol graph"] = "Journal of molecular graphics",
  ["j mol graph model"] = "Journal of molecular graphics & modelling",
  ["j mol graphics"] = "Journal of Molecular Graphics",
  ["j mol graphics model"] = "Journal of Molecular Graphics and Modelling",
  ["j mol graphics modell"] = "Journal of Molecular Graphics and Modelling",
  ["j mol histol"] = "Journal of molecular histology",
  ["j mol imaging dyn"] = "Journal of molecular imaging & dynamics",
  ["j mol immunol"] = "Journal of molecular immunology",
  ["j mol liq"] = "Journal of Molecular Liquids",
  ["j mol med"] = "Journal of Molecular Medicine",
  ["j mol med (berl)"] = "Journal of molecular medicine (Berlin, Germany)",
  ["j mol med clin appl"] = "Journal of molecular medicine and clinical applications",
  ["j mol microbiol biotechnol"] = "Journal of Molecular Microbiology and Biotechnology",
  ["j mol model"] = "Journal of molecular modeling",
  ["j mol neurosci"] = "Journal of molecular neuroscience : MN",
  ["j mol pathol"] = "Journal of Molecular Pathology",
  ["j mol pharm org process res"] = "Journal of molecular pharmaceutics & organic process research",
  ["j mol psychiatry"] = "Journal of molecular psychiatry",
  ["j mol recognit"] = "Journal of Molecular Recognition",
  ["j mol signal"] = "Journal of molecular signaling",
  ["j mol signaling"] = "Journal of Molecular Signaling",
  ["j mol spectrosc"] = "Journal of Molecular Spectroscopy",
  ["j mol struct"] = "Journal of Molecular Structure",
  ["j mol struct theochem"] = "Journal of Molecular Structure: THEOCHEM",
  ["j molluscan stud"] = "The Journal of molluscan studies",
  ["j monet econ"] = "Journal of monetary economics",
  ["j money"] = "Journal of Money, Credit, and Banking",
  ["j money credit bank"] = "Journal of money, credit, and banking",
  ["j mong math soc"] = "Journal of the Mongolian Mathematical Society",
  ["j moral educ"] = "Journal of moral education",
  ["j moral philos"] = "Journal of moral philosophy",
  ["j mormon hist"] = "Journal of Mormon history",
  ["j morphol"] = "Journal of morphology",
  ["j moscow phys soc"] = "Moscow Physical Society",
  ["j mosq res"] = "Journal of mosquito research",
  ["j mot behav"] = "Journal of motor behavior",
  ["j mot learn dev"] = "Journal of motor learning and development",
  ["j mother child"] = "Journal of mother and child",
  ["j mountain sci"] = "Journal of Mountain Science",
  ["j mov disord"] = "Journal of movement disorders",
  ["j mt agric balk"] = "Journal of mountain agriculture on the Balkans",
  ["j mt sci"] = "Journal of mountain science",
  ["j mt sinai hosp n y"] = "Journal of the Mount Sinai Hospital, New York",
  ["j mucosal immunol res"] = "Journal of mucosal immunology research",
  ["j mult scler (foster city)"] = "Journal of multiple sclerosis",
  ["j mult-valued logic soft comput"] = "Journal of Multiple-Valued Logic and Soft Computing",
  ["j multi-criter decis anal"] = "Journal of Multi-Criteria Decision Analysis",
  ["j multicult couns devel"] = "Journal of multicultural counseling and development",
  ["j multicult nurs health"] = "The journal of multicultural nursing & health : official journal of the Center for the Study of Multiculturalism and Health Care, Inc",
  ["j multicult soc work"] = "Journal of multicultural social work",
  ["j multidiscip eval"] = "Journal of multidisciplinary evaluation",
  ["j multidiscip healthc"] = "Journal of multidisciplinary healthcare",
  ["j multiling commun disord"] = "Journal of multilingual communication disorders",
  ["j multiling multicult dev"] = "Journal of multilingual and multicultural development",
  ["j multimed"] = "Journal of multimedia",
  ["j multimodal user interface"] = "Journal on Multimodal User Interfaces",
  ["j multimorb comorb"] = "Journal of Multimorbidity and Comorbidity",
  ["j multinat finan manage"] = "Journal of Multinational Financial Management",
  ["j multiscale model"] = "Journal of Multiscale Modeling",
  ["j multiscale modell"] = "Journal of Multiscale Modelling",
  ["j multivar anal"] = "Journal of multivariate analysis",
  ["j multivariate anal"] = "Journal of Multivariate Analysis",
  ["j mus educ"] = "Journal of museum education : roundtable reports",
  ["j muscle res cell motil"] = "Journal of muscle research and cell motility",
  ["j musculoskelet disord treat"] = "Journal of musculoskeletal disorders and treatment",
  ["j musculoskelet med"] = "The Journal of musculoskeletal medicine",
  ["j musculoskelet neuronal interact"] = "Journal of musculoskeletal & neuronal interactions",
  ["j musculoskelet pain"] = "Journal of musculoskeletal pain",
  ["j musculoskelet res"] = "Journal of musculoskeletal research",
  ["j musculoskeletal res"] = "Journal of Musculoskeletal Research",
  ["j music ther"] = "Journal of music therapy",
  ["j muslim ment health"] = "Journal of Muslim mental health",
  ["j muslim minor aff"] = "Journal of Muslim minority affairs",
  ["j mycol"] = "Journal of mycology",
  ["j mycol infect"] = "Journal of mycology and infection",
  ["j mycol med"] = "Journal de mycologie médicale",
  ["j mycol plant pathol"] = "Journal of mycology and plant pathology",
  ["j mycopathol res"] = "Journal of mycopathological research",
  ["j n c acad sci"] = "Journal of the North Carolina Academy of Science",
  ["j n c dent soc"] = "The Journal of the North Carolina Dental Society",
  ["j n engl water environ assoc"] = "Journal of New England Water Environment Association",
  ["j n engl water works assoc"] = "Journal of New England Water Works Association",
  ["j n h dent soc"] = "Journal (New Hampshire Dental Society)",
  ["j n j dent assoc"] = "Journal of the New Jersey Dental Association",
  ["j n j dent hyg assoc"] = "Journal Of The New Jersey Dental Hygienists Association",
  ["j n j state dent soc"] = "The Journal of the New Jersey State Dental Society",
  ["j n y state nurses assoc"] = "The Journal of the New York State Nurses' Association",
  ["j n y state sch nurse teach assoc"] = "The Journal of the New York State School Nurse Teachers Association",
  ["j n z soc periodontol"] = "Journal of the New Zealand Society of Periodontology",
  ["j nal assoc"] = "Journal of NAL Associates",
  ["j nanjing norm univ nat sci ed"] = "Journal of Nanjing Normal University. Natural Science Edition. Nanjing Shida Xuebao. Ziran Kexue Ban",
  ["j nanjing univ nat sci"] = "Journal of Nanjing University. Natural Sciences",
  ["j nano educ"] = "Journal of Nano Education",
  ["j nano electron phys"] = "Journal Of Nano And Electronic Physics",
  ["j nano energy power res"] = "Journal of Nano Energy and Power Research",
  ["j nano res"] = "Journal of Nano Research",
  ["j nanobiotechnol"] = "Journal of Nanbiotechnology",
  ["j nanobiotechnology"] = "Journal of nanobiotechnology",
  ["j nanoelectron optoelectron"] = "Journal of Nanoelectronics and Optoelectronics",
  ["j nanoeng nanomanuf"] = "Journal of Nanoengineering and Nanomanufacturing",
  ["j nanoeurosci"] = "Journal of Nanoeuroscience",
  ["j nanomater"] = "Journal of nanomaterials",
  ["j nanomech micromech"] = "Journal of nanomechanics & micromechanics",
  ["j nanomed nanotechnol"] = "Journal of nanomedicine & nanotechnology",
  ["j nanomed res"] = "Journal of nanomedicine research",
  ["j nanomedine biotherapeutic discov"] = "Journal of nanomedicine & biotherapeutic discovery",
  ["j nanoneurosci"] = "Journal of nanoneuroscience",
  ["j nanopart"] = "Journal of Nanoparticles",
  ["j nanopart res"] = "Journal of Nanoparticle Research",
  ["j nanoparticle res interdiscip forum nanoscale sci technol"] = "Journal of Nanoparticle Research",
  ["j nanopharm drug deliv"] = "Journal of nanopharmaceutics and drug delivery",
  ["j nanophoton"] = "Journal of Nanophotonics",
  ["j nanophotonics"] = "Journal of Nanophotonics",
  ["j nanosci lett"] = "Journal of nanoscience letters",
  ["j nanosci nanomed"] = "Journal of nanoscience and nanomedicine",
  ["j nanosci nanotechnol"] = "Journal of nanoscience and nanotechnology",
  ["j nanostruct"] = "Journal of Nanostructures",
  ["j nanostruct chem"] = "Journal of Nanostructure in Chemistry",
  ["j nanostructure chem"] = "Journal of nanostructure in chemistry",
  ["j nanotechnol"] = "Journal of nanotechnology",
  ["j nanotechnol eng med"] = "Journal of nanotechnology in engineering and medicine",
  ["j nanotechnol mater sci"] = "Journal of nanotechnology and materials science",
  ["j nanotoxicol nanomed"] = "Journal of nanotoxicology and nanomedicine",
  ["j nat conserv"] = "Journal of Nature Conservation",
  ["j nat fibers"] = "Journal of Natural Fibers",
  ["j nat gas chem"] = "Journal of Natural Gas Chemistry",
  ["j nat gas geosci"] = "Journal of Natural Gas Geoscience",
  ["j nat gas sci eng"] = "Journal of Natural Gas Science and Engineering",
  ["j nat hist"] = "Journal of Natural History",
  ["j nat hist mus"] = "Journal of natural history museum",
  ["j nat hist mus inst chiba"] = "Chiba Kenritsu Chūō Hakubutsukan shizenshi kenkyū hōkoku = Journal of the Natural History Museum and Institute, Chiba",
  ["j nat indones"] = "Jurnal natur Indonesia : wacana sains Indonesia",
  ["j nat med"] = "Journal of natural medicines",
  ["j nat pestic res"] = "Journal of Natural Pesticide Research",
  ["j nat prod"] = "Journal of Natural Products",
  ["j nat resour life sci educ"] = "Journal of Natural Resources and Life Sciences Education",
  ["j nat resour policy res"] = "Journal of natural resources policy research",
  ["j nat sci"] = "Journal of nature and science",
  ["j nat sci biol med"] = "Journal of natural science, biology, and medicine",
  ["j nat sci hunan norm univ"] = "Journal of Natural Science of Hunan Normal University. Hunan Shifan Daxue Ziran Kexue Xuebao",
  ["j nat sci nanjing norm univ"] = "Journal of Natural Science. Nanjing Normal University",
  ["j nat sci sustainable technol"] = "Journal of Nature Science and Sustainable Technology",
  ["j nat toxins"] = "Journal of Natural Toxins",
  ["j natal zulu hist"] = "Journal of Natal and Zulu history",
  ["j natl analg soc"] = "Journal. National Analgesia Society",
  ["j natl assoc chirop"] = "Journal of the National Association of Chiropodists",
  ["j natl assoc chiropodists"] = "Journal of the National Association of Chiropodists",
  ["j natl assoc hosp dev"] = "Journal (National Association for Hospital Development (U.S.))",
  ["j natl assoc priv psychiatr hosp"] = "Journal - National Association of Private Psychiatric Hospitals",
  ["j natl assoc seventh day advent dent"] = "Journal Of The National Association Of Seventh-Day Adventist Dentists",
  ["j natl black nurses assoc"] = "Journal of National Black Nurses' Association : JNBNA",
  ["j natl cancer inst"] = "Journal of the National Cancer Institute",
  ["j natl cancer inst monogr"] = "Journal of the National Cancer Institute. Monographs",
  ["j natl compr canc netw"] = "Journal of the National Comprehensive Cancer Network : JNCCN",
  ["j natl malar soc"] = "Journal. National Malaria Society (U.S.)",
  ["j natl med assoc"] = "Journal of the National Medical Association",
  ["j natl proctol assoc"] = "Journal. National Proctologic Association",
  ["j natl sci found"] = "Journal of the National Science Foundation of Sri Lanka",
  ["j natl sci found sri lanka"] = "Journal of the National Science Foundation of Sri Lanka",
  ["j natur geom"] = "Journal of Natural Geometry",
  ["j natur phys sci"] = "Journal of Natural & Physical Sciences",
  ["j natur sci math"] = "The Journal of Natural Sciences and Mathematics",
  ["j navig"] = "Journal of Navigation",
  ["j nc acad sci"] = "Journal of the North Carolina Academy of Science",
  ["j near east stud"] = "Journal of Near Eastern studies",
  ["j near infrared spectrosc"] = "Journal of Near Infrared Spectroscopy",
  ["j nebr dent assoc"] = "The Journal of the Nebraska Dental Association",
  ["j negat results"] = "Journal of negative results : ecology & evolutionary biology",
  ["j negat results biomed"] = "Journal of negative results in biomedicine",
  ["j negro educ"] = "The Journal of Negro education",
  ["j negro hist"] = "The Journal of Negro history",
  ["j nematode morphol system"] = "Journal of nematode morphology and systematics",
  ["j nematol"] = "Journal of nematology",
  ["j neonatal biol"] = "Journal of neonatal biology",
  ["j neonatal nurs"] = "Journal of neonatal nursing : JNN",
  ["j neonatal perinatal med"] = "Journal of neonatal-perinatal medicine",
  ["j neonatal surg"] = "Journal of neonatal surgery",
  ["j nepal agric res counc"] = "Journal of Nepal Agricultural Research Council",
  ["j nepal chem soc"] = "Journal of Nepal Chemical Society",
  ["j nepal health res counc"] = "Journal of Nepal Health Research Council",
  ["j nephrol"] = "Journal of nephrology",
  ["j nephrol nurs"] = "Journal of nephrology nursing",
  ["j nephrol ren transplant"] = "Journal of nephrology and renal transplantation",
  ["j nephrol res"] = "Journal of nephrology research",
  ["j nephrol soc work"] = "The journal of nephrology social work",
  ["j nephrol ther"] = "Journal of nephrology & therapeutics",
  ["j nephropathol"] = "Journal of nephropathology",
  ["j nephropharmacol"] = "Journal of nephropharmacology",
  ["j nerv ment dis"] = "The Journal of nervous and mental disease",
  ["j network comput appl"] = "Journal of Network and Computer Applications",
  ["j network syst manage"] = "Journal of Network and Systems Management",
  ["j neural eng"] = "Journal of neural engineering",
  ["j neural transm"] = "Journal of Neural Transmission",
  ["j neural transm (vienna)"] = "Journal of neural transmission (Vienna, Austria : 1996)",
  ["j neural transm gen sect"] = "Journal of neural transmission. General section",
  ["j neural transm park dis dement sect"] = "Journal of neural transmission. Parkinson's disease and dementia section",
  ["j neural transm suppl"] = "Journal of neural transmission. Supplementum",
  ["j neural transm: gen sect"] = "Journal of Neural Transmission: General Section",
  ["j neural transplant"] = "Journal of neural transplantation",
  ["j neural transplant plast"] = "Journal of neural transplantation & plasticity",
  ["j neuro-oncol"] = "Journal of Neuro-Oncology",
  ["j neuroaids"] = "Journal of neuro-AIDS",
  ["j neuroanaesth crit care"] = "Journal of neuroanaesthesiology and critical care",
  ["j neurobiol"] = "Journal of Neurobiology",
  ["j neurobiol physiol"] = "Journal of neurobiology and physiology",
  ["j neurochem"] = "Journal of Neurochemistry",
  ["j neurocytol"] = "Journal of neurocytology",
  ["j neurodegener dis"] = "Journal of neurodegenerative diseases",
  ["j neurodegener regen"] = "Journal of neurodegeneration & regeneration",
  ["j neurodev disord"] = "Journal of neurodevelopmental disorders",
  ["j neuroendocrinol"] = "Journal of neuroendocrinology",
  ["j neuroeng rehabil"] = "Journal of neuroengineering and rehabilitation",
  ["j neurogastroenterol motil"] = "Journal of neurogastroenterology and motility",
  ["j neurogenet"] = "Journal of neurogenetics",
  ["j neuroimaging"] = "Journal of neuroimaging : official journal of the American Society of Neuroimaging",
  ["j neuroimaging psychiatry neurol"] = "Journal of neuroimaging in psychiatry & neurology",
  ["j neuroimmune pharmacol"] = "Journal of neuroimmune pharmacology : the official journal of the Society on NeuroImmune Pharmacology",
  ["j neuroimmunol"] = "Journal of neuroimmunology",
  ["j neuroimmunol suppl"] = "Journal of neuroimmunology. Supplement",
  ["j neuroinfect dis"] = "Journal of neuroinfectious diseases",
  ["j neuroinflamm neurodegener dis"] = "Journal of neuroinflammation and neurodegenerative diseases",
  ["j neuroinflammation"] = "Journal of neuroinflammation",
  ["j neurointerv surg"] = "Journal of neurointerventional surgery",
  ["j neurol"] = "Journal of neurology",
  ["j neurol disord"] = "Journal of neurological disorders",
  ["j neurol disord stroke"] = "Journal of neurological disorders & stroke",
  ["j neurol exp neurosci"] = "Journal of neurology and experimental neuroscience",
  ["j neurol neurobiol"] = "Journal of neurology and neurobiology",
  ["j neurol neurol disord"] = "Journal of neurology and neurological disorders",
  ["j neurol neurol sci disord"] = "Journal of neurology, neurological science and disorders",
  ["j neurol neuromedicine"] = "Journal of neurology & neuromedicine",
  ["j neurol neurophysiol"] = "Journal of neurology & neurophysiology",
  ["j neurol neurorehabilit res"] = "Journal of neurology and neurorehabilitation research",
  ["j neurol neurosci"] = "Journal of neurology and neuroscience",
  ["j neurol neurosurg"] = "Journal of neurology and neurosurgery",
  ["j neurol neurosurg psychiatry"] = "Journal of neurology, neurosurgery, and psychiatry",
  ["j neurol neurosurg psychiatry res"] = "Journal of neurology, neurosurgery & psychiatry research",
  ["j neurol neurosurg spine"] = "Journal of neurology, neurosurgery and spine",
  ["j neurol phys ther"] = "Journal of neurologic physical therapy : JNPT",
  ["j neurol psychiatry"] = "Journal of neurology and psychiatry",
  ["j neurol psychol"] = "Journal of neurology and psychology",
  ["j neurol psychopathol"] = "The Journal of neurology and psychopathology",
  ["j neurol rehabil"] = "Journal of neurologic rehabilitation",
  ["j neurol res"] = "Journal of neurology research",
  ["j neurol res ther"] = "Journal of neurological research and therapy",
  ["j neurol sci"] = "Journal of the neurological sciences",
  ["j neurol sci psychiatry"] = "Journal of neurological sciences and psychiatry",
  ["j neurol sci turk"] = "Journal of neurological sciences (Turkish)",
  ["j neurol surg a cent eur neurosurg"] = "Journal of neurological surgery. Part A, Central European neurosurgery",
  ["j neurol surg b skull base"] = "Journal of neurological surgery. Part B, Skull base",
  ["j neurol surg rep"] = "Journal of neurological surgery reports",
  ["j neurol transl neurosci"] = "Journal of neurology & translational neuroscience",
  ["j neurolinguistics"] = "Journal of neurolinguistics",
  ["j neuromuscul dis"] = "Journal of neuromuscular diseases",
  ["j neuromusculoskelet syst"] = "Journal of the neuromusculoskeletal system : JNMS : a journal of the American Chiropractic Association, Inc",
  ["j neurooncol"] = "Journal of neuro-oncology",
  ["j neuroophthalmol"] = "Journal of neuro-ophthalmology : the official journal of the North American Neuro-Ophthalmology Society",
  ["j neuroparasitology"] = "Journal of neuroparasitology",
  ["j neuropathic pain symptom palliation"] = "Journal of neuropathic pain & symptom palliation",
  ["j neuropathol clin neurol"] = "Journal of neuropathology and clinical neurology",
  ["j neuropathol exp neurol"] = "Journal of neuropathology and experimental neurology",
  ["j neurophysiol"] = "Journal of neurophysiology",
  ["j neurophysiol neurol disord"] = "Journal of neurophysiology and neurological disorders",
  ["j neuropsychiatr"] = "Journal of neuropsychiatry",
  ["j neuropsychiatry (foster city)"] = "Journal of neuropsychiatry (Foster City, Calif.)",
  ["j neuropsychiatry clin neurosci"] = "The Journal of neuropsychiatry and clinical neurosciences",
  ["j neuropsychol"] = "Journal of neuropsychology",
  ["j neuroradiol"] = "Journal of neuroradiology = Journal de neuroradiologie",
  ["j neurosci"] = "The Journal of neuroscience : the official journal of the Society for Neuroscience",
  ["j neurosci methods"] = "Journal of neuroscience methods",
  ["j neurosci neuroeng"] = "Journal of neuroscience and neuroengineering",
  ["j neurosci neuropharmacol"] = "Journal of neuroscience and neuropharmacology",
  ["j neurosci nurs"] = "The Journal of neuroscience nursing : journal of the American Association of Neuroscience Nurses",
  ["j neurosci psychol econ"] = "Journal of neuroscience, psychology, and economics",
  ["j neurosci rehabil"] = "Journal of neuroscience and rehabilitation",
  ["j neurosci res"] = "Journal of Neuroscience Research",
  ["j neurosci rural pract"] = "Journal of neurosciences in rural practice",
  ["j neurosurg"] = "Journal of neurosurgery",
  ["j neurosurg anesthesiol"] = "Journal of neurosurgical anesthesiology",
  ["j neurosurg nurs"] = "Journal of neurosurgical nursing",
  ["j neurosurg pediatr"] = "Journal of neurosurgery. Pediatrics",
  ["j neurosurg sci"] = "Journal of neurosurgical sciences",
  ["j neurosurg spine"] = "Journal of neurosurgery. Spine",
  ["j neurother"] = "Journal of neurotherapy",
  ["j neurotrauma"] = "Journal of neurotrauma",
  ["j neurovirol"] = "Journal of neurovirology",
  ["j neurovisc relat"] = "Journal of neuro-visceral relations",
  ["j neutron res"] = "Journal of Neutron Research",
  ["j new biol rep"] = "Journal on new biological reports",
  ["j new drugs"] = "The Journal of new drugs",
  ["j new mater electrochem syst"] = "Journal of New Materials for Electrochemical Systems",
  ["j new music res"] = "Journal of new music research",
  ["j newark beth isr hosp"] = "Journal - Newark Beth Israel Hospital",
  ["j niger assoc dent stud"] = "Journal Of The Nigeria Association Of Dental Students",
  ["j niger math soc"] = "Journal of the Nigerian Mathematical Society",
  ["j niger soc phys sci"] = "Journal of the Nigerian Society of Physical Sciences",
  ["j nigeria med assoc"] = "Journal of the Nigeria Medical Association",
  ["j nigerian math soc"] = "Journal of the Nigerian Mathematical Society",
  ["j nih res"] = "The Journal of NIH research : life sciences research and news about the National Institutes of Health and the Alcohol, Drug Abuse, and Mental Health Administration",
  ["j nihon univ sch dent"] = "The Journal of Nihon University School of Dentistry",
  ["j ningxia univ nat sci ed"] = "Journal of Ningxia University",
  ["j nippon med sch"] = "Journal of Nippon Medical School = Nippon Ika Daigaku zasshi",
  ["j non cryst solids"] = "Journal of non-crystalline solids",
  ["j non-cryst solids"] = "Journal of Non-Crystalline Solids",
  ["j non-cryst solids x"] = "Journal of Non-Crystalline Solids X",
  ["j non-equilib thermodyn"] = "Journal of Non-Equilibrium Thermodynamics",
  ["j non-newton fluid mech"] = "Journal of Non-Newtonian Fluid Mechanics",
  ["j non-newtonian fluid mech"] = "Journal of Non-Newtonian Fluid Mechanics",
  ["j noncommut geom"] = "Journal of Noncommutative Geometry",
  ["j nondestr eval"] = "Journal of Nondestructive Evaluation",
  ["j nondestr eval diagn progn eng syst"] = "Journal of Nondestructive Evaluation, Diagnostics & Prognostics of Engineering Systems",
  ["j nondestruct eval diagn progn eng syst"] = "Journal of nondestructive evaluation, diagnostics and prognostics of engineering systems",
  ["j nonlinear anal optim"] = "Journal of Nonlinear Analysis and Optimization. Theory and Applications",
  ["j nonlinear complex data sci"] = "Journal of Nonlinear, Complex and Data Science",
  ["j nonlinear convex anal"] = "Journal of nonlinear and convex analysis",
  ["j nonlinear evol equ appl"] = "Journal of Nonlinear Evolution Equations and Applications",
  ["j nonlinear math phys"] = "Journal of Nonlinear Mathematical Physics",
  ["j nonlinear opt phys mater"] = "Journal of Nonlinear Optical Physics and Materials",
  ["j nonlinear sci"] = "Journal of Nonlinear Science",
  ["j nonlinear sci appl"] = "Journal of Nonlinear Sciences and Applications. JNSA",
  ["j nonlinear syst appl"] = "Journal of Nonlinear Systems and Applications",
  ["j nonnewton fluid mech"] = "Journal of non-Newtonian fluid mechanics",
  ["j nonparam stat"] = "Journal of Nonparametric Statistics",
  ["j nonparametr stat"] = "Journal of nonparametric statistics",
  ["j nonparametr statist"] = "Journal of Nonparametric Statistics",
  ["j nonsmooth anal optim"] = "Journal of Nonsmooth Analysis and Optimization",
  ["j nonverbal behav"] = "Journal of nonverbal behavior",
  ["j north am benthol soc"] = "Journal of the North American Benthological Society",
  ["j north atl"] = "Journal of the North Atlantic",
  ["j north la hist assoc"] = "Journal (North Louisiana Historical Association)",
  ["j northeast agric univ"] = "The journal of Northeast Agricultural University",
  ["j northeast univ nat sci"] = "Journal of Northeastern University. Natural Science. Dongbei Daxue Xuebao. Ziran Kexue Ban",
  ["j northwest atl fish sci"] = "Journal of Northwest Atlantic Fishery Science",
  ["j northwest polytech univ"] = "Journal of the Northwestern Polytechnical University",
  ["j northwest univ"] = "Journal of Northwest University",
  ["j northwest univ nat sci"] = "Journal of Northwest University. Natural Science Edition. Xibei Daxue Xuebao. Ziran Kexue Ban",
  ["j nov physiother"] = "Journal of novel physiotherapies",
  ["j nov physiother phys rehabil"] = "Journal of novel physiotherapy and physical rehabilitation",
  ["j nucl biol med"] = "Journal of Nuclear Biology and Medicine",
  ["j nucl cardiol"] = "Journal of nuclear cardiology : official publication of the American Society of Nuclear Cardiology",
  ["j nucl energy"] = "Journal of Nuclear Energy",
  ["j nucl energy a"] = "Journal of Nuclear Energy, Part A: Reactor Science",
  ["j nucl energy b"] = "Journal of Nuclear Energy, Part B: Reactor Technology",
  ["j nucl energy c"] = "Journal of Nuclear Energy. Part C, Plasma Physics, Accelerators, Thermonuclear Research",
  ["j nucl energy part a:"] = "Journal of Nuclear Energy, Part A: Reactor Science",
  ["j nucl energy part b:"] = "Journal of Nuclear Energy, Part B: Reactor Technology",
  ["j nucl energy part c:"] = "Journal of Nuclear Energy. Part C, Plasma Physics, Accelerators, Thermonuclear Research",
  ["j nucl eng radiat sci"] = "Journal of Nuclear Engineering and Radiation Science",
  ["j nucl mater"] = "Journal of Nuclear Materials",
  ["j nucl med"] = "Journal of nuclear medicine : official publication, Society of Nuclear Medicine",
  ["j nucl med allied sci"] = "The Journal of nuclear medicine and allied sciences",
  ["j nucl med radiat ther"] = "Journal of nuclear medicine & radiation therapy",
  ["j nucl med technol"] = "Journal of nuclear medicine technology",
  ["j nucl radiochem sci"] = "Journal of Nuclear and Radiochemical Sciences",
  ["j nucl sci technol"] = "Journal of Nuclear Science and Technology",
  ["j nucl sci technol (abingdon, u k)"] = "Journal of Nuclear Science and Technology (Abingdon, United Kingdom)",
  ["j nucl sci technol (tokyo, jpn)"] = "Journal of Nuclear Science and Technology (Tokyo, Japan)",
  ["j nucleic acids"] = "Journal of nucleic acids",
  ["j nucleic acids investig"] = "Journal of nucleic acids investigation",
  ["j number theory"] = "Journal of Number Theory",
  ["j numer anal approx theory"] = "Journal of Numerical Analysis and Approximation Theory",
  ["j numer cogn"] = "Journal of numerical cognition",
  ["j numer math"] = "Journal of Numerical Mathematics",
  ["j numer math stoch"] = "Journal of Numerical Mathematics and Stochastics",
  ["j numer methods comput appl"] = "Journal on Numerical Methods and Computer Applications",
  ["j nurs (luton)"] = "Journal of nursing (Luton, England)",
  ["j nurs adm"] = "The Journal of nursing administration",
  ["j nurs care"] = "The Journal of nursing care",
  ["j nurs care qual"] = "Journal of nursing care quality",
  ["j nurs dr stud scholarsh"] = "Journal of nursing doctoral students scholarship",
  ["j nurs educ"] = "The Journal of nursing education",
  ["j nurs educ pract"] = "Journal of nursing education and practice",
  ["j nurs ethics"] = "Journal Of Nursing Ethics",
  ["j nurs health"] = "Journal of nursing and health",
  ["j nurs healthc chronic illn"] = "Journal of nursing and healthcare of chronic illness",
  ["j nurs hist"] = "Journal of nursing history : a publication of the Nursing Archives Associates at Boston University",
  ["j nurs home res sci"] = "The journal of nursing home research sciences",
  ["j nurs law"] = "Journal of nursing law",
  ["j nurs manag"] = "Journal of nursing management",
  ["j nurs meas"] = "Journal of nursing measurement",
  ["j nurs qual assur"] = "Journal of nursing quality assurance",
  ["j nurs regul"] = "Journal of nursing regulation",
  ["j nurs res"] = "The journal of nursing research : JNR",
  ["j nurs scholarsh"] = "Journal of nursing scholarship : an official publication of Sigma Theta Tau International Honor Society of Nursing",
  ["j nurs staff dev"] = "Journal of nursing staff development : JNSD",
  ["j nurse midwifery"] = "Journal of nurse-midwifery",
  ["j nurse pract"] = "The journal for nurse practitioners : JNP",
  ["j nurses prof dev"] = "Journal for nurses in professional development",
  ["j nurses staff dev"] = "Journal for nurses in staff development : JNSD : official journal of the National Nursing Staff Development Organization",
  ["j nutr"] = "Journal of Nutrition",
  ["j nutr biochem"] = "The Journal of nutritional biochemistry",
  ["j nutr biol"] = "Journal of nutritional biology",
  ["j nutr diet"] = "Journal of nutrition and dietetics",
  ["j nutr disord ther"] = "Journal of nutritional disorders & therapy",
  ["j nutr ecol food res"] = "Journal of nutritional ecology and food research",
  ["j nutr educ"] = "Journal of nutrition education",
  ["j nutr educ behav"] = "Journal of nutrition education and behavior",
  ["j nutr elder"] = "Journal of nutrition for the elderly",
  ["j nutr environ med"] = "Journal of Nutritional & Environmental Medicine",
  ["j nutr food sci"] = "Journal of nutrition & food sciences",
  ["j nutr gerontol geriatr"] = "Journal of nutrition in gerontology and geriatrics",
  ["j nutr health"] = "Journal of nutrition & health (Northborough)",
  ["j nutr health aging"] = "The journal of nutrition, health & aging",
  ["j nutr immunol"] = "Journal of Nutritional Immunology",
  ["j nutr intermed metab"] = "Journal of nutrition & intermediary metabolism",
  ["j nutr med diet care"] = "Journal of nutritional medicine and diet care",
  ["j nutr metab"] = "Journal of nutrition and metabolism",
  ["j nutr oncol"] = "Journal of Nutritional Oncology",
  ["j nutr physiol"] = "Journal of Nutritional Physiology",
  ["j nutr sci"] = "Journal of nutritional science",
  ["j nutr sci vitaminol"] = "Journal of Nutritional Science and Vitaminology",
  ["j nutr sci vitaminol (tokyo)"] = "Journal of nutritional science and vitaminology",
  ["j nutr, health aging"] = "Journal of Nutrition, Health & Aging",
  ["j nutrigenet nutrigenomics"] = "Journal of nutrigenetics and nutrigenomics",
  ["j nutrit health food sci"] = "Journal of nutritional health & food science",
  ["j obes"] = "Journal of obesity",
  ["j obes metab syndr"] = "Journal of obesity & metabolic syndrome",
  ["j obes overweight"] = "Journal of Obesity and Overweight",
  ["j obes weight loss medicat"] = "Journal of obesity and weight-loss medication",
  ["j obes weight loss ther"] = "Journal of obesity & weight loss therapy",
  ["j obsessive compuls relat disord"] = "Journal of obsessive-compulsive and related disorders",
  ["j obstet gynaecol"] = "Journal of obstetrics and gynaecology : the journal of the Institute of Obstetrics and Gynaecology",
  ["j obstet gynaecol (lahore)"] = "Journal of obstetrics and gynaecology",
  ["j obstet gynaecol (tokyo 1995)"] = "Journal of obstetrics and gynaecology (Tokyo, Japan)",
  ["j obstet gynaecol br commonw"] = "The Journal of obstetrics and gynaecology of the British Commonwealth",
  ["j obstet gynaecol br emp"] = "The Journal of obstetrics and gynaecology of the British Empire",
  ["j obstet gynaecol can"] = "Journal of obstetrics and gynaecology Canada : JOGC = Journal d'obstétrique et gynécologie du Canada : JOGC",
  ["j obstet gynaecol east cent africa"] = "Journal of obstetrics & gynaecology of Eastern and Central Africa",
  ["j obstet gynaecol india"] = "Journal of obstetrics and gynaecology of India",
  ["j obstet gynaecol res"] = "The journal of obstetrics and gynaecology research",
  ["j obstet gynaecol soc north india"] = "Journal. Obstetric and Gynaecological Society of Northern India",
  ["j obstet gynecol neonatal nurs"] = "Journal of obstetric, gynecologic, and neonatal nursing : JOGNN",
  ["j occup accid"] = "Journal of Occupational Accidents",
  ["j occup environ hyg"] = "Journal of occupational and environmental hygiene",
  ["j occup environ med"] = "Journal of occupational and environmental medicine",
  ["j occup health"] = "Journal of occupational health",
  ["j occup health psychol"] = "Journal of occupational health psychology",
  ["j occup med"] = "Journal of occupational medicine. : official publication of the Industrial Medical Association",
  ["j occup med toxicol"] = "Journal of occupational medicine and toxicology (London, England)",
  ["j occup organ psychol"] = "Journal of occupational and organizational psychology",
  ["j occup rehabil"] = "Journal of occupational rehabilitation",
  ["j occup sci"] = "Journal of occupational science",
  ["j occup ther sch early interv"] = "Journal of occupational therapy, schools & early intervention",
  ["j ocean clim"] = "Journal of Ocean and Climate",
  ["j ocean coast econ"] = "Journal of ocean and coastal economics",
  ["j ocean eng mar energy"] = "Journal of ocean engineering and marine energy",
  ["j ocean eng sci"] = "Journal of Ocean Engineering and Science",
  ["j ocean sci found"] = "Journal of the ocean science foundation",
  ["j ocean technol"] = "Journal of Ocean Technology",
  ["j ocean univ china"] = "Journal of Ocean University of China (English Edition)",
  ["j ocean univ qingdao"] = "Ocean University of Qingdao",
  ["j oceanogr"] = "Journal of Oceanography",
  ["j oceanogr soc jpn"] = "Journal of the Oceanographical Society of Japan",
  ["j oceanol limnol"] = "Journal of oceanology and limnology",
  ["j ocul biol"] = "Journal of ocular biology",
  ["j ocul biol dis infor"] = "Journal of ocular biology, diseases, and informatics",
  ["j ocul dis ther"] = "Journal of ocular diseases and therapeutics",
  ["j ocul pharmacol"] = "Journal of ocular pharmacology",
  ["j ocul pharmacol ther"] = "Journal of ocular pharmacology and therapeutics : the official journal of the Association for Ocular Pharmacology and Therapeutics",
  ["j odontol conserv"] = "Journal d'odontologie conservatrice",
  ["j of southwest univ"] = "Journal of Southwest University (Natural Science Edition)",
  ["j off repub fr ed lois decrets"] = "Journal Officiel De La Republique Francaise, Edition Des Lois Et Decrets",
  ["j off stat"] = "Journal of official statistics",
  ["j offender rehabil"] = "Journal of offender rehabilitation",
  ["j offshore mech arct eng"] = "Journal of Offshore Mechanics and Arctic Engineering",
  ["j ohio state dent soc"] = "Journal. Ohio State Dental Society",
  ["j oil gas petrochem sci"] = "Journal of oil, gas and petrochemical sciences",
  ["j oil palm res"] = "Journal of oil palm research",
  ["j oilseeds res"] = "Journal of oilseeds research",
  ["j okla dent assoc"] = "Journal - Oklahoma Dental Association",
  ["j okla state dent assoc"] = "Journal - Oklahoma State Dental Association",
  ["j okla state med assoc"] = "The Journal of the Oklahoma State Medical Association",
  ["j oleo sci"] = "Journal of oleo science",
  ["j omaha midwest clin soc"] = "Journal. Omaha Mid-West Clinical Society",
  ["j oncobiomarkers"] = "Journal of oncobiomarkers",
  ["j oncol"] = "Journal of oncology",
  ["j oncol cancer res"] = "Journal of oncology and cancer research",
  ["j oncol manag"] = "The Journal of oncology management : the official journal of the American College of Oncology Administrators",
  ["j oncol navig surviv"] = "Journal of oncology navigation & survivorship",
  ["j oncol pharm pract"] = "Journal of oncology pharmacy practice : official publication of the International Society of Oncology Pharmacy Practitioners",
  ["j oncol pract"] = "Journal of oncology practice",
  ["j oncol res ther"] = "Journal of oncology research and therapy",
  ["j oncol research"] = "Journal of oncology research",
  ["j oncol transl res"] = "Journal of oncology translational research",
  ["j oncopathol"] = "The journal of oncopathology",
  ["j oncopathol clin res"] = "Journal of oncopathology and clinical research",
  ["j online learn teach"] = "Journal of online learning and teaching",
  ["j ont dent assoc"] = "The Journal of the Ontario Dental Association",
  ["j open archaeol data"] = "Journal of Open Archaeology Data",
  ["j open hardware"] = "Journal of Open Hardware",
  ["j open res softw"] = "Journal of open research software",
  ["j open res software"] = "Journal of Open Research Software",
  ["j open source educ"] = "The Journal of open source education",
  ["j open source softw"] = "Journal of open source software",
  ["j open source software"] = "Journal of Open Source Software",
  ["j oper manage"] = "Journal of Operations Management",
  ["j oper meteorol"] = "Journal of Operational Meteorology",
  ["j oper oceanogr"] = "Journal of Operational Oceanography",
  ["j oper res soc"] = "The Journal of the Operational Research Society",
  ["j oper res soc china"] = "Journal of the Operations Research Society of China",
  ["j oper res soc japan"] = "Journal of the Operations Research Society of Japan",
  ["j oper room res inst"] = "The Journal of the Operating Room Research Institute : JORRI",
  ["j oper strategic plann"] = "Journal of Operations and Strategic Planning",
  ["j operat meteorol"] = "Journal of operational meteorology",
  ["j operator theory"] = "Journal of Operator Theory",
  ["j ophthalmic inflamm infect"] = "Journal of ophthalmic inflammation and infection",
  ["j ophthalmic nurs technol"] = "Journal of ophthalmic nursing & technology",
  ["j ophthalmic photogr"] = "The Journal of ophthalmic photography",
  ["j ophthalmic vis res"] = "Journal of ophthalmic & vision research",
  ["j ophthalmol"] = "Journal of ophthalmology",
  ["j opioid manag"] = "Journal of opioid management",
  ["j opt"] = "Journal of Optics",
  ["j opt a: pure appl opt"] = "Journal of Optics A: Pure and Applied Optics",
  ["j opt b quantum semiclassical opt"] = "Journal of optics. B, Quantum and semiclassical optics : journal of the European Optical Society",
  ["j opt b: quantum semiclassical opt"] = "Journal of Optics B: Quantum and Semiclassical Optics",
  ["j opt commun"] = "Journal of Optical Communication",
  ["j opt commun networking"] = "Journal of Optical Communications and Networking",
  ["j opt microsyst"] = "Journal of Optical Microsystems",
  ["j opt soc am"] = "Journal of the Optical Society of America",
  ["j opt soc am a"] = "Journal of the Optical Society of America. A, Optics and image science",
  ["j opt soc am a opt image sci vis"] = "Journal of the Optical Society of America. A, Optics, image science, and vision",
  ["j opt soc am a:"] = "Journal of the Optical Society of America A: Optics, Image Science, and Vision",
  ["j opt soc am b"] = "Journal of the Optical Society of America. B, Optical physics",
  ["j opt soc am b: opt phys"] = "Journal of the Optical Society of America B: Optical Physics",
  ["j opt soc amer a"] = "Journal of the Optical Society of America A",
  ["j opt soc korea"] = "Journal of the Optical Society of Korea",
  ["j optim"] = "Journal of Optimization",
  ["j optim differ equ appl"] = "Journal of Optimization, Differential Equations and their Applications",
  ["j optim theory"] = "Journal of Optimization Theory",
  ["j optim theory appl"] = "Journal of Optimization Theory and Applications",
  ["j optoelectron adv mater"] = "Journal of Optoelectronics and Advanced Materials",
  ["j optom"] = "Journal of optometry",
  ["j or state dent assoc"] = "Journal. Oregon State Dental Association",
  ["j oral biol (northborough)"] = "Journal of oral biology (Northborough, Mass.)",
  ["j oral biol craniofac res"] = "Journal of oral biology and craniofacial research",
  ["j oral biosci"] = "Journal of oral biosciences",
  ["j oral dis"] = "Journal of oral diseases",
  ["j oral facial pain  headache"] = "Journal of oral & facial pain and headache",
  ["j oral hyg health"] = "Journal of oral hygiene & health",
  ["j oral implant transplant surg"] = "The Journal of oral implant and transplant surgery",
  ["j oral implantol"] = "The Journal of oral implantology",
  ["j oral maxillofac pathol"] = "Journal of oral and maxillofacial pathology : JOMFP",
  ["j oral maxillofac res"] = "Journal of oral & maxillofacial research",
  ["j oral maxillofac surg"] = "Journal of oral and maxillofacial surgery : official journal of the American Association of Oral and Maxillofacial Surgeons",
  ["j oral maxillofac surg med pathol"] = "Journal of oral and maxillofacial surgery, medicine, and pathology",
  ["j oral med"] = "Journal of oral medicine",
  ["j oral med pain"] = "Journal of oral medicine and pain",
  ["j oral microbiol"] = "Journal of oral microbiology",
  ["j oral pathol"] = "Journal of oral pathology",
  ["j oral pathol med"] = "Journal of oral pathology & medicine : official publication of the International Association of Oral Pathologists and the American Academy of Oral Pathology",
  ["j oral rehabil"] = "Journal of oral rehabilitation",
  ["j oral sci"] = "Journal of oral science",
  ["j oral surg"] = "Journal of oral surgery (American Dental Association : 1965)",
  ["j oral surg (chic)"] = "Journal of oral surgery",
  ["j oral surg anesth hosp dent serv"] = "Journal of oral surgery, anesthesia, and hospital dental service",
  ["j oral ther pharmacol"] = "Journal of oral therapeutics and pharmacology",
  ["j oreg dent assoc"] = "The Journal of the Oregon Dental Association",
  ["j org biomol simul"] = "Journal of organic and biomolecular simulations",
  ["j org chem"] = "The Journal of organic chemistry",
  ["j org semicond"] = "Journal of Organic Semiconductors",
  ["j organ behav"] = "Journal of organizational behavior",
  ["j organ behav manage"] = "Journal of organizational behavior management",
  ["j organ dysfunct"] = "Journal of organ dysfunction",
  ["j organ end user comput"] = "Journal of organizational and end user computing : an official publication of the Information Resources Management Association",
  ["j organ psychol"] = "Journal of organizational psychology",
  ["j organomet chem"] = "Journal of Organometallic Chemistry",
  ["j orissa math soc"] = "Journal of the Orissa Mathematical Society",
  ["j ornithol"] = "Journal of ornithology",
  ["j orofac orthop"] = "Journal of orofacial orthopedics = Fortschritte der Kieferorthopädie : Organ/official journal Deutsche Gesellschaft für Kieferorthopädie",
  ["j orofac pain"] = "Journal of orofacial pain",
  ["j orofac sci"] = "Journal of orofacial sciences",
  ["j ort res"] = "Journal of Ortoptera Research",
  ["j orthod"] = "Journal of orthodontics",
  ["j orthod sci"] = "Journal of orthodontic science",
  ["j orthomol med"] = "Journal of orthomolecular medicine : official journal of the Academy of Orthomolecular Medicine",
  ["j orthop"] = "Journal of orthopaedics",
  ["j orthop case rep"] = "Journal of orthopaedic case reports",
  ["j orthop res"] = "Journal of orthopaedic research : official publication of the Orthopaedic Research Society",
  ["j orthop res ther"] = "Journal of orthopedic research and therapy",
  ["j orthop rheumatol"] = "Journal of orthopedics & rheumatology",
  ["j orthop sci"] = "Journal of orthopaedic science : official journal of the Japanese Orthopaedic Association",
  ["j orthop sports phys ther"] = "The Journal of orthopaedic and sports physical therapy",
  ["j orthop surg"] = "Journal of orthopaedic surgery and research",
  ["j orthop surg (hong kong)"] = "Journal of orthopaedic surgery (Hong Kong)",
  ["j orthop surg res"] = "Journal of orthopaedic surgery and research",
  ["j orthop translat"] = "Journal of orthopaedic translation",
  ["j orthop trauma"] = "Journal of orthopaedic trauma",
  ["j orthop traumatol"] = "Journal of orthopaedics and traumatology : official journal of the Italian Society of Orthopaedics and Traumatology",
  ["j orthoptera res"] = "Journal of orthoptera research",
  ["j osaka dent univ"] = "Journal of Osaka Dental University",
  ["j osaka univ dent sch"] = "The Journal of Osaka University Dental School",
  ["j oslo city hosp"] = "Journal of the Oslo city hospitals",
  ["j osteopath (kirksvill)"] = "The Journal of osteopathy",
  ["j osteoporos"] = "Journal of osteoporosis",
  ["j osteoporos phys act"] = "Journal of osteoporosis and physical activity",
  ["j otol"] = "Journal of otology",
  ["j otol rhinol"] = "Journal of otology & rhinology",
  ["j otolaryngol"] = "The Journal of otolaryngology",
  ["j otolaryngol head neck surg"] = "Journal of otolaryngology - head & neck surgery = Le Journal d'oto-rhino-laryngologie et de chirurgie cervico-faciale",
  ["j otolaryngol soc aust"] = "Journal of the Oto-laryngological Society of Australia",
  ["j otolaryngol suppl"] = "The Journal of otolaryngology. Supplement",
  ["j otto rank assoc"] = "Journal of the Otto Rank Association. Otto Rank Association",
  ["j outcome meas"] = "Journal of outcome measurement",
  ["j outdoor environ educ"] = "Journal of Outdoor and Environmental Education",
  ["j ovarian res"] = "Journal of ovarian research",
  ["j ovonic res"] = "Journal of Ovonic Research",
  ["j pa acad sci"] = "Journal of the Pennsylvania Academy of Science",
  ["j pac hist"] = "The Journal of Pacific history",
  ["j packag technol res"] = "Journal of packaging technology and research",
  ["j paediatr child health"] = "Journal of paediatrics and child health",
  ["j paediatr dent"] = "Journal of paediatric dentistry",
  ["j pain"] = "The journal of pain",
  ["j pain manag"] = "Journal of pain management",
  ["j pain manag med"] = "Journal of pain management & medicine",
  ["j pain palliat care pharmacother"] = "Journal of pain & palliative care pharmacotherapy",
  ["j pain palliative care pharmacother"] = "Journal of Pain & Palliative Care Pharmacotherapy",
  ["j pain relief"] = "Journal of pain & relief",
  ["j pain res"] = "Journal of pain research",
  ["j pain symptom manage"] = "Journal of pain and symptom management",
  ["j pak med assoc"] = "JPMA. The Journal of the Pakistan Medical Association",
  ["j pak psychiatr soc"] = "Journal of Pakistan Psychiatric Society",
  ["j palaeogeogr"] = "Journal of Palaeogeography",
  ["j paleolimnol"] = "Journal of Paleolimnology",
  ["j paleolithic archaeol"] = "Journal of Paleolithic Archaeology",
  ["j paleontol"] = "Journal of Paleontology",
  ["j palest arab med assoc"] = "Journal. Palestine Arab Medical Association",
  ["j palest stud"] = "Journal of Palestine studies",
  ["j palliat care"] = "Journal of palliative care",
  ["j palliat care med"] = "Journal of palliative care & medicine",
  ["j palliat med"] = "Journal of palliative medicine",
  ["j palynol"] = "Journal of Palynology",
  ["j pan afr stud"] = "The journal of Pan African studies",
  ["j pancreat cancer"] = "Journal of pancreatic cancer",
  ["j pancreatol"] = "Journal of pancreatology",
  ["j pap conserv"] = "Journal of Paper Conservation",
  ["j parallel distrib comput"] = "Journal of Parallel and Distributed Computing",
  ["j paramed sci"] = "Journal of paramedical sciences",
  ["j parapsychol"] = "The Journal of parapsychology",
  ["j parasit dis"] = "Journal of parasitic diseases : official organ of the Indian Society for Parasitology",
  ["j parasitol"] = "The Journal of parasitology",
  ["j parasitol (faisalabad)"] = "Journal of parasitology (Faisalabad)",
  ["j parasitol res"] = "Journal of parasitology research",
  ["j parenter drug assoc"] = "Journal of the Parenteral Drug Association",
  ["j parenter sci technol"] = "Journal of parenteral science and technology : a publication of the Parenteral Drug Association",
  ["j parenter ther"] = "The Journal of parenteral therapy",
  ["j park recreat admi"] = "Journal of park and recreation administration",
  ["j parkinsons dis"] = "Journal of Parkinson's disease",
  ["j parkinsons dis alzheimers dis"] = "Journal of Parkinson's disease and Alzheimer's disease",
  ["j parodontol"] = "Journal de parodontologie",
  ["j partial differ equ"] = "Journal of Partial Differential Equations",
  ["j partial differential equations"] = "Journal of Partial Differential Equations",
  ["j particip med"] = "Journal of participatory medicine",
  ["j particip res methods"] = "Journal of participatory research methods",
  ["j pastoral care"] = "Journal of pastoral care",
  ["j pastoral care counsel"] = "The journal of pastoral care & counseling : JPCC",
  ["j pat off soc"] = "Journal of the Patent Office Society. Patent Office Society (U.S.)",
  ["j pat trademark off soc"] = "Journal of the Patent and Trademark Office Society",
  ["j pathog"] = "Journal of pathogens",
  ["j pathol"] = "The Journal of pathology",
  ["j pathol bacteriol"] = "The Journal of pathology and bacteriology",
  ["j pathol clin res"] = "The journal of pathology. Clinical research",
  ["j pathol inf"] = "Journal of Pathology Informatics",
  ["j pathol inform"] = "Journal of pathology informatics",
  ["j pathol transl med"] = "Journal of pathology and translational medicine",
  ["j pathol: clin res"] = "Journal of Pathology: Clinical Research",
  ["j patient acc manage"] = "Journal of patient account management",
  ["j patient cent res rev"] = "Journal of patient-centered research and reviews",
  ["j patient exp"] = "Journal of patient experience",
  ["j patient rep outcomes"] = "Journal of patient-reported outcomes",
  ["j patient saf"] = "Journal of patient safety",
  ["j patient saf risk manag"] = "Journal of patient safety and risk management",
  ["j peace res"] = "Journal of peace research",
  ["j peace sci"] = "Journal of Peace Science",
  ["j peasant stud"] = "The Journal of peasant studies",
  ["j pediatr"] = "The Journal of pediatrics",
  ["j pediatr (rio j)"] = "Jornal de pediatria",
  ["j pediatr adolesc gynecol"] = "Journal of pediatric and adolescent gynecology",
  ["j pediatr biochem"] = "Journal of pediatric biochemistry",
  ["j pediatr child health care"] = "Journal of pediatrics & child health care",
  ["j pediatr child nutr"] = "Journal of pediatrics and child nutrition",
  ["j pediatr endocrinol"] = "The Journal of pediatric endocrinology",
  ["j pediatr endocrinol (jersey city)"] = "Journal of pediatric endocrinology (Jersey City, N.J.)",
  ["j pediatr endocrinol metab"] = "Journal of pediatric endocrinology & metabolism : JPEM",
  ["j pediatr epilepsy"] = "Journal of pediatric epilepsy",
  ["j pediatr gastroenterol nutr"] = "Journal of pediatric gastroenterology and nutrition",
  ["j pediatr genet"] = "Journal of pediatric genetics",
  ["j pediatr health care"] = "Journal of pediatric health care : official publication of National Association of Pediatric Nurse Associates & Practitioners",
  ["j pediatr hematol oncol"] = "Journal of pediatric hematology/oncology",
  ["j pediatr infect dis"] = "Journal of pediatric infectious diseases",
  ["j pediatr intensive care"] = "Journal of pediatric intensive care",
  ["j pediatr mother care"] = "Journal of pediatrics and mother care",
  ["j pediatr neonatal care"] = "Journal of pediatrics & neonatal care",
  ["j pediatr neurol"] = "Journal of pediatric neurology : JPN",
  ["j pediatr neuropsychol"] = "Journal of pediatric neuropsychology",
  ["j pediatr neuroradiol"] = "Journal of pediatric neuroradiology",
  ["j pediatr neurosci"] = "Journal of pediatric neurosciences",
  ["j pediatr nurs"] = "Journal of pediatric nursing",
  ["j pediatr oncol"] = "Journal of pediatric oncology",
  ["j pediatr oncol nurs"] = "Journal of pediatric oncology nursing : official journal of the Association of Pediatric Oncology Nurses",
  ["j pediatr ophthalmol"] = "Journal of pediatric ophthalmology",
  ["j pediatr ophthalmol strabismus"] = "Journal of pediatric ophthalmology and strabismus",
  ["j pediatr orthop"] = "Journal of pediatric orthopedics",
  ["j pediatr orthop b"] = "Journal of pediatric orthopedics. Part B",
  ["j pediatr pediatr med"] = "Journal of pediatrics and pediatric medicine",
  ["j pediatr perinat nutr"] = "Journal of pediatric & perinatal nutrition",
  ["j pediatr perinatol child health"] = "Journal of pediatrics, perinatology and child health",
  ["j pediatr pharmacol ther"] = "The journal of pediatric pharmacology and therapeutics : JPPT : the official journal of PPAG",
  ["j pediatr prenatal nutr"] = "Journal of Pediatric & Prenatal Nutrition",
  ["j pediatr psychol"] = "Journal of pediatric psychology",
  ["j pediatr pueric"] = "Journal de pédiatrie et de puériculture",
  ["j pediatr rehabil med"] = "Journal of pediatric rehabilitation medicine",
  ["j pediatr rev"] = "Journal of pediatrics review",
  ["j pediatr surg"] = "Journal of pediatric surgery",
  ["j pediatr surg case rep"] = "Journal of pediatric surgery case reports",
  ["j pediatr surg nurs"] = "Journal of pediatric surgical nursing",
  ["j pediatr urol"] = "Journal of pediatric urology",
  ["j pediatr x"] = "The journal of pediatrics: X",
  ["j pediatric infect dis soc"] = "Journal of the Pediatric Infectious Diseases Society",
  ["j pedod"] = "The Journal of pedodontics",
  ["j pelvic med surg"] = "Journal of pelvic medicine & surgery",
  ["j penelit tanam ind"] = "Jurnal penelitian tanaman industri = Industrial crops research journal",
  ["j pengolah has perikan indones"] = "Jurnal Pengolahan Hasil Perikanan Indonesia",
  ["j pension econ financ"] = "Journal of pension economics & finance",
  ["j pept res"] = "Journal of Peptide Research",
  ["j pept sci"] = "Journal of Peptide Science",
  ["j perform constr facil"] = "Journal of Performance of Constructed Facilities",
  ["j perform health res"] = "Journal of performance health research",
  ["j perianesth nurs"] = "Journal of perianesthesia nursing : official journal of the American Society of PeriAnesthesia Nurses",
  ["j peridyn nonlocal model"] = "Journal of Peridynamics and Nonlocal Modeling",
  ["j perinat educ"] = "The Journal of perinatal education",
  ["j perinat med"] = "Journal of perinatal medicine",
  ["j perinat neonatal nurs"] = "The Journal of perinatal & neonatal nursing",
  ["j perinatol"] = "Journal of perinatology : official journal of the California Perinatal Association",
  ["j periodontal implant sci"] = "Journal of periodontal & implant science",
  ["j periodontal res"] = "Journal of periodontal research",
  ["j periodontal res suppl"] = "Journal of periodontal research. Supplement",
  ["j periodontol"] = "Journal of periodontology",
  ["j perioper crit intensiv care nurs"] = "Journal of perioperative & critical intensive care nursing",
  ["j perioper pract"] = "Journal of perioperative practice",
  ["j peripher nerv syst"] = "Journal of the peripheral nervous system : JPNS",
  ["j pers"] = "Journal of personality",
  ["j pers assess"] = "Journal of personality assessment",
  ["j pers disord"] = "Journal of personality disorders",
  ["j pers eval educ"] = "Journal of personnel evaluation in education",
  ["j pers med"] = "Journal of personalized medicine",
  ["j pers nanomed"] = "Journal of personalized nanomedicine",
  ["j pers oriented res"] = "Journal for person-oriented research",
  ["j pers soc psychol"] = "Journal of personality and social psychology",
  ["j personal disord"] = "Journal of Personality Disorders",
  ["j pest sci"] = "Journal of Pest Science",
  ["j pest sci (2004)"] = "Journal of pest science",
  ["j pestic sci"] = "Journal of pesticide science",
  ["j pet environ biotechnol"] = "Journal of petroleum & environmental biotechnology",
  ["j pet explor prod technol"] = "Journal of petroleum exploration and production technology",
  ["j pet geol"] = "Journal of Petroleum Geology",
  ["j pet sci eng"] = "Journal of petroleum science & engineering",
  ["j pet sci technol"] = "Journal of Petroleum Science and Technology",
  ["j pet technol"] = "Journal of Petroleum Technology",
  ["j petrol"] = "Journal of Petrology",
  ["j pharm (cairo)"] = "Journal of pharmaceutics",
  ["j pharm anal"] = "Journal of pharmaceutical analysis",
  ["j pharm belg"] = "Journal de pharmacie de Belgique",
  ["j pharm bioallied sci"] = "Journal of pharmacy & bioallied sciences",
  ["j pharm biomed anal"] = "Journal of Pharmaceutical and Biomedical Analysis",
  ["j pharm biomed sci"] = "Journal of pharmaceutical and biomedical sciences",
  ["j pharm chem"] = "Journal of Pharmacy and Chemistry",
  ["j pharm drug deliv res"] = "Journal of pharmaceutics & drug delivery research",
  ["j pharm drug res"] = "Journal of pharmaceutics and drug research",
  ["j pharm health care sci"] = "Journal of pharmaceutical health care and sciences",
  ["j pharm health serv res"] = "Journal of pharmaceutical health services research : an official journal of the Royal Pharmaceutical Society of Great Britain",
  ["j pharm innov"] = "Journal of pharmaceutical innovation",
  ["j pharm innovation"] = "Journal of Pharmaceutical Innovation",
  ["j pharm invest"] = "Journal of Pharmaceutical Investigation",
  ["j pharm investig"] = "Journal of pharmaceutical investigation",
  ["j pharm law"] = "The Journal of pharmacy & law",
  ["j pharm mark manage"] = "Journal of pharmaceutical marketing & management",
  ["j pharm negat results"] = "Journal of pharmaceutical negative results",
  ["j pharm pharm"] = "Journal of pharmacy and pharmaceutics",
  ["j pharm pharm sci"] = "Journal of pharmacy & pharmaceutical sciences : a publication of the Canadian Society for Pharmaceutical Sciences, Société canadienne des sciences pharmaceutiques",
  ["j pharm pharmacogn res"] = "Journal of pharmacy & pharmacognosy research",
  ["j pharm pharmacol"] = "Journal of Pharmacy and Pharmacology",
  ["j pharm pharmacol (los angel)"] = "Journal of pharmaceutics & pharmacology",
  ["j pharm policy pract"] = "Journal of pharmaceutical policy and practice",
  ["j pharm pract"] = "Journal of pharmacy practice",
  ["j pharm pract res"] = "Journal of Pharmacy Practice and Research",
  ["j pharm res"] = "Journal of Pharmacy Research",
  ["j pharm res int"] = "Journal of pharmaceutical research international",
  ["j pharm sci"] = "Journal of Pharmaceutical Sciences",
  ["j pharm sci exp pharmacol"] = "Journal of pharmaceutical sciences and experimental pharmacology",
  ["j pharm sci pharmacol"] = "Journal of pharmaceutical sciences and pharmacology",
  ["j pharm sci technol"] = "Journal of pharmaceutical science and technology : the official journal of PDA",
  ["j pharm soc wis"] = "The journal of the Pharmacy Society of Wisconsin : JPSW",
  ["j pharm stat"] = "Journal of Biopharmaceutical Statistics",
  ["j pharm technol"] = "The Journal of pharmacy technology : jPT : official publication of the Association of Pharmacy Technicians",
  ["j pharm technol drug res"] = "Journal of pharmaceutical technology & drug research",
  ["j pharmacobio-dyn"] = "Journal of Pharmacobio-Dynamics",
  ["j pharmacobiodyn"] = "Journal of pharmacobio-dynamics",
  ["j pharmacogenomics pharmacoproteomics"] = "Journal of pharmacogenomics & pharmacoproteomics",
  ["j pharmacogn nat prod"] = "Journal of pharmacognosy & natural products",
  ["j pharmacogn phytochem"] = "Journal of pharmacognosy and phytochemistry",
  ["j pharmacokinet biopharm"] = "Journal of pharmacokinetics and biopharmaceutics",
  ["j pharmacokinet pharmacodyn"] = "Journal of pharmacokinetics and pharmacodynamics",
  ["j pharmacol"] = "Journal de pharmacologie",
  ["j pharmacol biomed anal"] = "Journal of pharmacological & biomedical analysis",
  ["j pharmacol clin toxicol"] = "Journal of pharmacology & clinical toxicology",
  ["j pharmacol clin trials"] = "Journal of pharmacology and clinical trials",
  ["j pharmacol exp ther"] = "Journal of Pharmacology and Experimental Therapeutics",
  ["j pharmacol methods"] = "Journal of pharmacological methods",
  ["j pharmacol pharmacother"] = "Journal of pharmacology & pharmacotherapeutics",
  ["j pharmacol rep"] = "Journal of pharmacological reports",
  ["j pharmacol sci"] = "Journal of pharmacological sciences",
  ["j pharmacol sci (tokyo, jpn)"] = "Journal of Pharmacological Sciences (Tokyo, Japan)",
  ["j pharmacol toxicol methods"] = "Journal of Pharmacological and Toxicological Methods",
  ["j pharmacopuncture"] = "Journal of pharmacopuncture",
  ["j pharmacother"] = "Journal of pharmacotherapy",
  ["j pharmacovigil"] = "Journal of pharmacovigilance",
  ["j phase equilib"] = "Journal of Phase Equilibrium",
  ["j phase equilib diffus"] = "Journal of Phase Equilibrium and Diffusion",
  ["j phase equilibria"] = "Journal of Phase Equilibria",
  ["j phase equilibria diffus"] = "Journal of phase equilibria and diffusion",
  ["j phila cty dent soc"] = "The journal of the Philadelphia County Dental Society",
  ["j philipp dent assoc"] = "The Journal of the Philippine Dental Association",
  ["j philipp dev"] = "Journal of Philippine development : a technical journal of the National Economic and Development Authority",
  ["j philipp fed priv med pract"] = "Journal of the Philippine Federation of Private Medical Practitioners",
  ["j philipp med assoc"] = "Journal of the Philippine Medical Association",
  ["j philipp stat"] = "Journal of Philippine statistics",
  ["j philos"] = "The journal of philosophy",
  ["j philos logic"] = "Journal of Philosophical Logic",
  ["j philos sci law"] = "The journal of philosophy, science & law",
  ["j philos sport"] = "Journal of the philosophy of sport",
  ["j phon"] = "Journal of phonetics",
  ["j photocatal"] = "Journal Of Photocatalysis",
  ["j photochem"] = "Journal of Photochemistry",
  ["j photochem photobiol"] = "Journal of photochemistry and photobiology",
  ["j photochem photobiol a chem"] = "Journal of photochemistry and photobiology. A, Chemistry",
  ["j photochem photobiol b"] = "Journal of photochemistry and photobiology. B, Biology",
  ["j photochem photobiol c photochem rev"] = "Journal of Photochemistry and Photobiology C: Photochemistry Reviews",
  ["j photochem photobiol chem"] = "Journal of Photochemistry and Photobiology",
  ["j photochem photobiol, a"] = "Journal of Photochemistry and Photobiology A: Chemistry",
  ["j photochem photobiol, b"] = "Journal of Photochemistry and Photobiology B: Biology",
  ["j photochem photobiol, c"] = "Journal of Photochemistry and Photobiology C: Photochemistry Reviews",
  ["j photogramm remote sens"] = "Journal of Photogrammetry and Remote Sensing",
  ["j photonics energy"] = "Journal of Photonics for Energy",
  ["j photopolym sci technol"] = "Journal of Photopolymer Science and Technology",
  ["j phycol"] = "Journal of phycology",
  ["j phylogenetics evol biol"] = "Journal of phylogenetics & evolutionary biology",
  ["j phys"] = "Journal de Physique",
  ["j phys (moscow)"] = "Journal of Physics (Moscow)",
  ["j phys (paris)"] = "Journal de Physique (Paris)",
  ["j phys [e]"] = "Journal of Physics. E, Scientific Instruments",
  ["j phys a"] = "Journal of Physics A",
  ["j phys a math gen"] = "Journal of physics A: Mathematical and general",
  ["j phys a math theor"] = "Journal of physics. A, Mathematical and theoretical",
  ["j phys a: gen phys"] = "Journal of Physics A: General Physics",
  ["j phys a: math gen"] = "Journal of Physics A: Mathematical and General",
  ["j phys a: math nucl gen"] = "Journal of Physics A: Mathematical, Nuclear and General",
  ["j phys a: math theor"] = "Journal of Physics A: Mathematical and Theoretical",
  ["j phys act health"] = "Journal of physical activity & health",
  ["j phys act nutr rehabil"] = "Journal of physical activity, nutrition and rehabilitation",
  ["j phys act res"] = "Journal of physical activity research",
  ["j phys appl phys"] = "Journal of Physics D: Applied Physics",
  ["j phys b"] = "Journal of Physics B",
  ["j phys b at mol opt phys"] = "Journal of physics. B, Atomic, molecular, and optical physics : an Institute of Physics journal",
  ["j phys b: at mol opt phys"] = "Journal of Physics B: Atomic, Molecular and Optical Physics",
  ["j phys b: at mol phys"] = "Journal of Physics B: Atomic and Molecular Physics",
  ["j phys c"] = "Journal of Physics C",
  ["j phys c solid state phys"] = "Journal of Physics C: Solid State Physics",
  ["j phys c: solid state phys"] = "Journal of Physics C: Solid State Physics",
  ["j phys chem"] = "Journal of Physical Chemistry",
  ["j phys chem a"] = "The journal of physical chemistry. A",
  ["j phys chem b"] = "The journal of physical chemistry. B",
  ["j phys chem biophys"] = "Journal of physical chemistry & biophysics",
  ["j phys chem c"] = "Journal of Physical Chemistry C",
  ["j phys chem c nanomater interfaces"] = "The journal of physical chemistry. C, Nanomaterials and interfaces",
  ["j phys chem lett"] = "The journal of physical chemistry letters",
  ["j phys chem ref data"] = "Journal of Physical and Chemical Reference Data",
  ["j phys chem solids"] = "Journal of Physics and Chemistry of Solids",
  ["j phys colloid chem"] = "The Journal of physical and colloid chemistry",
  ["j phys colloq"] = "Journal de Physique Colloques",
  ["j phys commun"] = "Journal of Physics Communications",
  ["j phys complexity"] = "Journal of Physics: Complexity",
  ["j phys condens matter"] = "Journal of physics. Condensed matter : an Institute of Physics journal",
  ["j phys conf ser"] = "Journal of physics. Conference series",
  ["j phys conf series"] = "Journal of Physics: Conference series",
  ["j phys d"] = "Journal of Physics D",
  ["j phys d appl phys"] = "Journal of physics D: Applied physics",
  ["j phys d: appl phys"] = "Journal of Physics D: Applied Physics",
  ["j phys e"] = "Journal of physics E: Scientific instruments",
  ["j phys e: sci instrum"] = "Journal of Physics E: Scientific Instruments",
  ["j phys educ"] = "Journal of physical education",
  ["j phys educ recreat dance"] = "Journal of physical education, recreation & dance",
  ["j phys energy"] = "Journal of Physics: Energy",
  ["j phys f"] = "Journal of Physics F",
  ["j phys f: met phys"] = "Journal of Physics F: Metal Physics",
  ["j phys fit sports med"] = "The journal of physical fitness and sports medicine",
  ["j phys g"] = "Journal of Physics G",
  ["j phys g nucl part phys"] = "Journal of physics. G, Nuclear and particle physics : an Institute of Physics journal",
  ["j phys g: nucl part phys"] = "Journal of Physics G: Nuclear and Particle Physics",
  ["j phys g: nucl phys"] = "Journal of Physics G: Nuclear Physics",
  ["j phys i"] = "Journal de Physique I",
  ["j phys ii"] = "Journal de Physique II",
  ["j phys iii"] = "Journal de Physique III",
  ["j phys iv"] = "Journal de Physique IV",
  ["j phys iv fr"] = "Journal de Physique IV",
  ["j phys iv proc"] = "Journal de Physique IV (Proceedings)",
  ["j phys mater"] = "Journal of Physics: Materials",
  ["j phys math"] = "Journal of Physical Mathematics",
  ["j phys math theor"] = "Journal of Physics A: Mathematical and Theoretical",
  ["j phys med rehabil"] = "Journal of physical medicine and rehabilitation (Wilmington, Del.)",
  ["j phys oceanogr"] = "Journal of Physical Oceanography",
  ["j phys org chem"] = "Journal of Physical Organic Chemistry",
  ["j phys photonics"] = "Journal of Physics: Photonics",
  ["j phys radium"] = "Journal de Physique et le Radium",
  ["j phys sci"] = "Journal of Physical Science",
  ["j phys soc japan"] = "Journal of the Physical Society of Japan",
  ["j phys soc jpn"] = "Journal of the Physical Society of Japan",
  ["j phys stud"] = "Journal of Physical Studies. Zhurnal Fizichnykh Doslidzhen",
  ["j phys ther educ"] = "Journal, physical therapy education",
  ["j phys ther health promot"] = "Journal of physical therapy and health promotion",
  ["j phys ther sci"] = "Journal of physical therapy science",
  ["j phys: complexity"] = "Journal of Physics: Complexity",
  ["j phys: condens matter"] = "Journal of Physics: Condensed Matter",
  ["j phys: conf ser"] = "Journal of Physics: Conference Series",
  ["j phys: energy"] = "Journal of Physics: Energy",
  ["j phys: mater"] = "Journal of Physics: Materials",
  ["j phys: photonics"] = "Journal of Physics: Photonics",
  ["j physic educ sport manag"] = "Journal of physical education and sport management",
  ["j physical soc japan"] = "Journal Of The Physical Society Of Japan",
  ["j physician assist educ"] = "The journal of physician assistant education : the official journal of the Physician Assistant Education Association",
  ["j physicians assoc aids care"] = "Journal of the Physicians Association for AIDS Care",
  ["j physiol"] = "The Journal of physiology",
  ["j physiol (lond)"] = "Journal of Physiology",
  ["j physiol (oxford, u k)"] = "Journal of Physiology (Oxford, United Kingdom)",
  ["j physiol (paris)"] = "Journal de physiologie",
  ["j physiol anthropol"] = "Journal of physiological anthropology",
  ["j physiol anthropol appl human sci"] = "Journal of physiological anthropology and applied human science",
  ["j physiol biochem"] = "Journal of physiology and biochemistry",
  ["j physiol paris"] = "Journal of physiology, Paris",
  ["j physiol pathol gen"] = "Journal de physiologie et de pathologie générale",
  ["j physiol pharmacol"] = "Journal of physiology and pharmacology : an official journal of the Polish Physiological Society",
  ["j physiol sci"] = "The journal of physiological sciences : JPS",
  ["j physiol suppl (paris)"] = "Journal de physiologie. Supplément",
  ["j physiol-paris"] = "Journal of Physiology-Paris",
  ["j physiother"] = "Journal of physiotherapy",
  ["j physiother phys rehabil"] = "Journal of physiotherapy & physical rehabilitation",
  ["j physique i"] = "Journal de Physique I",
  ["j phytopathol"] = "Phytopathologische Zeitschrift. Journal of phytopathology",
  ["j phytopathol (1986)"] = "Journal of phytopathology (1986)",
  ["j pierre fauchard acad"] = "Journal of Pierre Fauchard Academy (Pierre Fauchard Academy. India Section)",
  ["j pigment disord"] = "Journal of pigmentary disorders",
  ["j pineal res"] = "Journal of pineal research",
  ["j pipeline sci eng"] = "Journal of Pipeline Science and Engineering",
  ["j pipeline syst eng pract"] = "Journal of Pipeline Systems Engineering and Practice",
  ["j place manage dev"] = "Journal of Place Management and Development",
  ["j plan educ res"] = "Journal of planning education and research",
  ["j plan hist"] = "Journal of planning history",
  ["j plan lit"] = "Journal of planning literature",
  ["j planar chromatogr - mod tlc"] = "Journal of Planar Chromatography - Modern TLC",
  ["j planar chromatogr–mod tlc"] = "Journal of Planar Chromatography–Modern TLC",
  ["j plankton res"] = "Journal of plankton research",
  ["j plant biochem biotechnol"] = "Journal of Plant Biochemistry and Biotechnology",
  ["j plant biochem physiol"] = "Journal of plant biochemistry & physiology",
  ["j plant biol"] = "Journal of plant biology = Singmul Hakhoe chi",
  ["j plant biotechnol"] = "Journal of plant biotechnology",
  ["j plant breed crop sci"] = "Journal of plant breeding and crop science",
  ["j plant dis prot"] = "Journal of Plant Diseases and Protection",
  ["j plant dis prot (2006)"] = "Journal of plant diseases and protection : scientific journal of the German Phytomedical Society (DPG)",
  ["j plant ecol"] = "Journal of plant ecology : JPE",
  ["j plant growth regul"] = "Journal of Plant Growth Regulation",
  ["j plant hydraul"] = "The journal of plant hydraulics",
  ["j plant interact"] = "Journal of Plant Interactions",
  ["j plant mol biol biotechnol"] = "Journal of Plant Molecular Biology and Biotechnology",
  ["j plant nutr"] = "Journal of Plant Nutrition",
  ["j plant nutr soil sci"] = "Journal of Plant Nutrition and Soil Science",
  ["j plant nutr soil sci (1999)"] = "Journal of plant nutrition and soil science = Zeitschrift für Pflanzenernährung und Bodenkunde",
  ["j plant pathol"] = "Journal of plant pathology : an international journal of the Italian Phytopathological Society",
  ["j plant pathol microbiol"] = "Journal of plant pathology & microbiology",
  ["j plant physiol"] = "Journal of plant physiology",
  ["j plant physiol pathol"] = "Journal of plant physiology & pathology",
  ["j plant prot res"] = "Journal of Plant Protection Research",
  ["j plant regist"] = "Journal of plant registrations",
  ["j plant res"] = "Journal of plant research",
  ["j plant stud"] = "Journal of Plant Studies",
  ["j plasma phys"] = "Journal of Plasma Physics",
  ["j plast film sheeting"] = "Journal of Plastic Film & Sheeting",
  ["j plast reconstr aesthet surg"] = "Journal of plastic, reconstructive & aesthetic surgery : JPRAS",
  ["j plast reconstr surg nurs"] = "The Journal of plastic and reconstructive surgical nursing : official organ of the American Society of Plastic and Reconstructive Surgical Nurses",
  ["j plast surg hand surg"] = "Journal of plastic surgery and hand surgery",
  ["j poet ther"] = "Journal of poetry therapy",
  ["j pol anal manage"] = "Journal of Policy Analysis and Management",
  ["j pol modeling"] = "Journal of Policy Modeling",
  ["j pol reform"] = "Journal of Policy Reform",
  ["j police crim psychol"] = "Journal of police and criminal psychology",
  ["j police crisis negot"] = "Journal of police crisis negotiations : an international journal",
  ["j policy anal manage"] = "Journal of policy analysis and management : [the journal of the Association for Public Policy Analysis and Management]",
  ["j policy hist"] = "Journal of policy history : JPH",
  ["j policy model"] = "Journal of policy modeling",
  ["j policy pract"] = "Journal of policy practice",
  ["j policy pract intellect disabil"] = "Journal of policy and practice in intellectual disabilities",
  ["j polit"] = "The journal of politics",
  ["j polit ecol"] = "Journal of political ecology",
  ["j polit econ"] = "The journal of political economy",
  ["j polit economy"] = "Journal of Political Economy",
  ["j polit mil soc"] = "Journal of political & military sociology : JPMS",
  ["j polit philos"] = "The journal of political philosophy",
  ["j polit sci"] = "Journal of political science",
  ["j politics law"] = "Journal of politics and law",
  ["j pollinat ecol"] = "Journal of pollination ecology",
  ["j polym"] = "Journal of polymers",
  ["j polym eng"] = "Journal of Polymer Engineering",
  ["j polym environ"] = "Journal of Polymers and the Environment",
  ["j polym mater"] = "Journal of Polymer Materials",
  ["j polym res"] = "Journal of Polymer Research",
  ["j polym sci"] = "Journal of Polymer Science",
  ["j polym sci (2020)"] = "Journal of polymer science (2020)",
  ["j polym sci [a1]"] = "Journal of Polymer Science. Part A-1, Polymer Chemistry",
  ["j polym sci [b]"] = "Journal of Polymer Science. Part B, Polymer Letters",
  ["j polym sci a"] = "Journal of polymer science: Part A, General papers",
  ["j polym sci a polym chem"] = "Journal of polymer science. Part A, Polymer chemistry",
  ["j polym sci a1"] = "Journal of polymer science. Part A-1, Polymer chemistry",
  ["j polym sci b"] = "Journal of polymer science. Part B: Polymer letters",
  ["j polym sci b polym phys"] = "Journal of polymer science. Part B, Polymer physics",
  ["j polym sci polym lett ed"] = "Journal of Polymer Science, Polymer Letters Edition",
  ["j polym sci polym phys ed"] = "Journal of Polymer Science, Polymer Physics Edition",
  ["j polym sci polym symp"] = "Journal of Polymer Science Polymer Symposium",
  ["j polym sci, part a: polym chem"] = "Journal of Polymer Science Part A: Polymer Chemistry",
  ["j polym sci, part b: polym phys"] = "Journal of Polymer Science Part B: Polymer Physics",
  ["j polym sci, part c: polym lett"] = "Journal of Polymer Science Part C: Polymer Letters",
  ["j polyn soc"] = "The journal of the Polynesian Society. Polynesian Society (N.Z.)",
  ["j pop cult"] = "Journal of popular culture",
  ["j pop film telev"] = "The journal of popular film and television : JPF & T",
  ["j popul"] = "Journal of population",
  ["j popul ageing"] = "Journal of population ageing",
  ["j popul assoc korea"] = "Han'guk In'gu Hakhoe chi = The journal of the Population Association of Korea",
  ["j popul econ"] = "Journal of population economics",
  ["j popul res"] = "Journal of population research",
  ["j popul res (canberra)"] = "Journal of population research (Canberra, A.C.T.)",
  ["j popul ther clin pharmacol"] = "Journal of population therapeutics and clinical pharmacology = Journal de la thérapeutique des populations et de la pharmacologie clinique",
  ["j population econ"] = "Journal of Population Economics",
  ["j porous mater"] = "Journal of Porous Materials",
  ["j porous media"] = "Journal of porous media",
  ["j porphyr phthalocyanines"] = "Journal of porphyrins and phthalocyanines",
  ["j porphyrins phthalocyanines"] = "Journal of Porphyrins and Phthalocyanines",
  ["j portfol manage"] = "Journal of Portfolio Management",
  ["j posit behav interv"] = "Journal of positive behavior interventions",
  ["j posit psychol"] = "The journal of positive psychology",
  ["j post anesth nurs"] = "Journal of post anesthesia nursing",
  ["j post keynes econ"] = "Journal of post Keynesian economics",
  ["j post keynesian econ"] = "Journal of Post Keynesian Economics",
  ["j postcolon writ"] = "Journal of postcolonial writing",
  ["j postgrad med"] = "Journal of postgraduate medicine",
  ["j posthum stud"] = "Journal of posthuman studies : philosophy, technology, media",
  ["j postsecond educ disabil"] = "Journal of postsecondary education and disability",
  ["j poult sci"] = "The journal of poultry science",
  ["j poverty"] = "Journal of poverty",
  ["j poverty alleviation int dev"] = "Journal of poverty alleviation and international development",
  ["j poverty soc justice"] = "The journal of poverty and social justice : research, policy, practice",
  ["j power electron"] = "Journal of Power Electronics",
  ["j power sources"] = "Journal of Power Sources",
  ["j power sources adv"] = "Journal of Power Sources Advances",
  ["j pract nurs"] = "The Journal of practical nursing",
  ["j pragmat"] = "Journal of pragmatics",
  ["j prakt chem"] = "Journal fur Praktische Chemie",
  ["j prat rev gen clin ther"] = "Journal des praticiens; revue générale de clinique et de thérapeutique",
  ["j precis med"] = "The Journal of precision medicine",
  ["j precis med: health dis"] = "Journal of Precision Medicine: Health and Disease",
  ["j precis respir med"] = "Journal of precision respiratory medicine",
  ["j precoll eng educ res"] = "Journal of pre-college engineering education research",
  ["j pregnancy"] = "Journal of pregnancy",
  ["j pregnancy child health"] = "Journal of pregnancy and child health",
  ["j prenat med"] = "Journal of prenatal medicine",
  ["j prenat perinat psychol health"] = "Journal of prenatal & perinatal psychology & health",
  ["j presbyt hist"] = "Journal of Presbyterian history",
  ["j press vessel technol"] = "Journal of pressure vessel technology",
  ["j pressure vessel technol"] = "Journal of Pressure Vessel Technology",
  ["j prev"] = "Journal of Prevention",
  ["j prev alzheimers dis"] = "The journal of prevention of Alzheimer's disease",
  ["j prev dent"] = "The Journal of preventive dentistry",
  ["j prev health promot"] = "Journal of prevention and health promotion",
  ["j prev interv community"] = "Journal of prevention & intervention in the community",
  ["j prev med (wilmington)"] = "Journal of preventive medicine",
  ["j prev med hyg"] = "Journal of preventive medicine and hygiene",
  ["j prev med pub health"] = "Journal of Preventive Medicine and Public Health",
  ["j prev med public health"] = "Journal of preventive medicine and public health = Yebang Ŭihakhoe chi",
  ["j prev psychiatry"] = "Journal of preventive psychiatry",
  ["j prev soc med"] = "Journal of preventive and social medicine : JOPSOM : a bi-annual journal of the National Institute of Preventive and Social Medicine",
  ["j preventive cardiol"] = "Journal of preventive cardiology",
  ["j prim care community health"] = "Journal of primary care & community health",
  ["j prim health care"] = "Journal of primary health care",
  ["j prim prev"] = "The journal of primary prevention",
  ["j primatol"] = "Journal of primatology",
  ["j prime res math"] = "Journal of Prime Research in Mathematics",
  ["j prison health"] = "Journal of prison health",
  ["j prison jail health"] = "Journal of prison & jail health",
  ["j prison prisons"] = "Journal of prisoners on prisons",
  ["j priv confid"] = "The Journal of privacy and confidentiality",
  ["j private enterprise"] = "Journal of Private Enterprise",
  ["j probab stat"] = "Journal of Probability and Statistics",
  ["j probiotics health"] = "Journal of probiotics & health",
  ["j probl solving"] = "The journal of problem solving",
  ["j proc inst chem gb irel"] = "Journal and Proceedins of the Institute of Chemistry of Great Britain and Ireland",
  ["j proc linn soc london bot"] = "Journal of the Proceedings of the Linnean Society of London. Botany",
  ["j proc linn soc london zool"] = "Journal of the Proceedings of the Linnean Society of London. Zoology",
  ["j proc r inst chem"] = "Journal and Proceedings of the Royal Institute of Chemistry",
  ["j proc r inst gb irel"] = "Journal and Proceedings of the Royal Institute of Chemistry of Great Britain and Ireland",
  ["j process control"] = "Journal of Process Control",
  ["j prod agric"] = "Journal of Production Agriculture",
  ["j prod anal"] = "Journal of Productivity Analysis",
  ["j prod innov manage"] = "The Journal of product innovation management",
  ["j prod innovation manage"] = "Journal of Product Innovation Management",
  ["j prod liability"] = "Journal of products liability",
  ["j productiv anal"] = "Journal of Productivity Analysis",
  ["j prof issues eng educ pract"] = "Journal of Professional Issues in Engineering Education and Practice",
  ["j prof nurs"] = "Journal of professional nursing : official journal of the American Association of Colleges of Nursing",
  ["j prof serv mark"] = "Journal of professional services marketing",
  ["j progress hum serv"] = "Journal of progressive human services",
  ["j progress res soc sci"] = "Journal of progressive research in social sciences",
  ["j proj tech"] = "Journal of projective techniques",
  ["j proj tech pers assess"] = "Journal of projective techniques & personality assessment",
  ["j propul power"] = "Journal of Propulsion and Power",
  ["j propuls power"] = "Journal of propulsion and power",
  ["j prosthet dent"] = "The Journal of prosthetic dentistry",
  ["j prosthet orthot"] = "Journal of prosthetics and orthotics : JPO",
  ["j prosthodont"] = "Journal of prosthodontics : official journal of the American College of Prosthodontists",
  ["j prosthodont res"] = "Journal of prosthodontic research",
  ["j protein chem"] = "Journal of Protein Chemistry",
  ["j proteins proteom"] = "Journal of proteins and proteomics",
  ["j proteins proteomics"] = "Journal of Proteins and Proteomics",
  ["j proteolysis"] = "Journal of proteolysis",
  ["j proteom genom res"] = "Journal of proteomics and genomics research",
  ["j proteome res"] = "Journal of proteome research",
  ["j proteome sci comput biol"] = "Journal of proteome science and computational biology",
  ["j proteomics"] = "Journal of proteomics",
  ["j proteomics bioinform"] = "Journal of proteomics & bioinformatics",
  ["j proton ther"] = "Journal of proton therapy",
  ["j protozool"] = "The Journal of protozoology",
  ["j protozool res"] = "The Journal of protozoology research",
  ["j pseudo-differ oper appl"] = "Journal of Pseudo-Differential Operators and Applications",
  ["j pseudodiffer oper appl"] = "Journal of pseudo-differential operators and applications",
  ["j psicol"] = "Jornal de psicologia",
  ["j psoriasis psoriatic arthritis"] = "Journal of psoriasis and psoriatic arthritis",
  ["j psychedelic drugs"] = "Journal of Psychedelic Drugs",
  ["j psychedelic stud"] = "Journal of psychedelic studies",
  ["j psychiatr brain sci"] = "Journal of psychiatry and brain science",
  ["j psychiatr educ"] = "Journal of psychiatric education",
  ["j psychiatr ment health nurs"] = "Journal of psychiatric and mental health nursing",
  ["j psychiatr nurs"] = "Journal of psychiatric nursing",
  ["j psychiatr nurs ment health serv"] = "Journal of psychiatric nursing and mental health services",
  ["j psychiatr pract"] = "Journal of psychiatric practice",
  ["j psychiatr res"] = "Journal of psychiatric research",
  ["j psychiatr soc work"] = "Journal of psychiatric social work",
  ["j psychiatr treat eval"] = "Journal of psychiatric treatment and evaluation",
  ["j psychiatry"] = "Journal of psychiatry",
  ["j psychiatry behav sci"] = "Journal of psychiatry and behavioral science",
  ["j psychiatry cogn behav"] = "Journal of psychiatry and cognitive behaviour",
  ["j psychiatry law"] = "The Journal of psychiatry & law",
  ["j psychiatry neurosci"] = "Journal of psychiatry & neuroscience : JPN",
  ["j psychiatry psychiatr disord"] = "Journal of psychiatry and psychiatric disorders",
  ["j psychoact drugs"] = "Journal of Psychoactive Drugs",
  ["j psychoactive drugs"] = "Journal of psychoactive drugs",
  ["j psychoeduc assess"] = "Journal of psychoeducational assessment",
  ["j psychohist"] = "The Journal of psychohistory",
  ["j psychol"] = "The Journal of psychology",
  ["j psychol abnorm child"] = "Journal of psychological abnormalities in children",
  ["j psychol afr"] = "Journal of psychology in Africa (south of the Sahara, the Caribbean, and Afro-Latin America)",
  ["j psychol ai"] = "Journal of Psychology and AI",
  ["j psychol behav res"] = "Journal of psychology & behavior research",
  ["j psychol cogn"] = "Journal of psychology and cognition",
  ["j psychol couns sch"] = "Journal of psychologists and counsellors in schools",
  ["j psychol human sex"] = "Journal of psychology & human sexuality",
  ["j psychol judaism"] = "Journal of psychology and Judaism",
  ["j psychol med ment pathol"] = "Journal of psychological medicine and mental pathology (London, England : 1848)",
  ["j psychol med ment pathol (lond)"] = "Journal of psychological medicine and mental pathology (London, England : 1875)",
  ["j psychol neuropsychiatr disord brain stimul"] = "Journal of psychology, neuropsychiatric disorders and brain stimulation",
  ["j psychol norm pathol (paris)"] = "Journal de psychologie normale et pathologique",
  ["j psychol psychiatr"] = "Journal of psychology and psychiatry",
  ["j psychol psychother"] = "Journal of psychology & psychotherapy",
  ["j psychol theol"] = "Journal of psychology and theology",
  ["j psycholinguist res"] = "Journal of psycholinguistic research",
  ["j psychopathol behav assess"] = "Journal of psychopathology and behavioral assessment",
  ["j psychopharmacol"] = "Journal of psychopharmacology (Oxford, England)",
  ["j psychopharmacol (oxf)"] = "Journal of Psychopharmacology",
  ["j psychophysiol"] = "Journal of psychophysiology",
  ["j psychosoc nurs ment health serv"] = "Journal of psychosocial nursing and mental health services",
  ["j psychosoc oncol"] = "Journal of psychosocial oncology",
  ["j psychosoc oncol res pract"] = "Journal of psychosocial oncology research and practice",
  ["j psychosoc rehabil ment health"] = "Journal of psychosocial rehabilitation and mental health",
  ["j psychosom obstet gynaecol"] = "Journal of psychosomatic obstetrics and gynaecology",
  ["j psychosom res"] = "Journal of psychosomatic research",
  ["j psychother integr"] = "Journal of psychotherapy integration",
  ["j psychother pract res"] = "The Journal of psychotherapy practice and research",
  ["j pubjab acad forensic med toxicol"] = "Journal Of Punjab Academy Of Forensic Medicine And Toxicology",
  ["j public adm res theory"] = "Journal of public administration research and theory : J-PART",
  ["j public aff"] = "Journal of public affairs",
  ["j public child welf"] = "Journal of public child welfare",
  ["j public econ"] = "Journal of public economics",
  ["j public econ theory"] = "Journal of public economic theory",
  ["j public finance public choice"] = "Journal of Public Finance and Public Choice",
  ["j public health"] = "Journal of Public Health",
  ["j public health (bangkok)"] = "Journal of public health",
  ["j public health (oxf)"] = "Journal of public health (Oxford, England)",
  ["j public health afr"] = "Journal of public health in Africa",
  ["j public health dent"] = "Journal of public health dentistry",
  ["j public health emerg"] = "Journal of public health and emergency",
  ["j public health epidemiol"] = "Journal of public health and epidemiology",
  ["j public health issues pract"] = "Journal of public health issues and practices",
  ["j public health manag pract"] = "Journal of public health management and practice : JPHMP",
  ["j public health med"] = "Journal of public health medicine",
  ["j public health policy"] = "Journal of public health policy",
  ["j public health res"] = "Journal of public health research",
  ["j public health sanit rev"] = "Journal of public health, and sanitary review",
  ["j public int aff"] = "Journal of Public and International Affairs",
  ["j public ment health"] = "Journal of public mental health",
  ["j public policy"] = "Journal of public policy",
  ["j public policy mark"] = "Journal of public policy & marketing : JPP&M : an annual publication of the Division of Research, Graduate School of Business Administration, the University of Michigan",
  ["j public trans"] = "Journal of public transportation",
  ["j public transp"] = "Journal of Public Transportation",
  ["j pulm respir med"] = "Journal of pulmonary & respiratory medicine",
  ["j pulp pap sci"] = "Journal of Pulp and Paper Science",
  ["j punjab acad forensic med toxicol"] = "Journal of Punjab Academy of Forensic Medicine & Toxicology",
  ["j pure appl algebra"] = "Journal of Pure and Applied Algebra",
  ["j pure appl math adv appl"] = "Journal of Pure and Applied Mathematics. Advances and Applications",
  ["j pure appl microbiol"] = "Journal of pure & applied microbiology",
  ["j pure math"] = "Journal of Pure Mathematics",
  ["j qingdao univ nat sci ed"] = "Journal of Qingdao University",
  ["j qual assur"] = "Journal of quality assurance : a publication of the National Association of Quality Assurance Professionals",
  ["j qual clin pract"] = "Journal of quality in clinical practice",
  ["j qual crim justice criminol"] = "Journal of qualitative criminal justice & criminology : JQCJC",
  ["j qual maint eng"] = "Journal of quality in maintenance engineering",
  ["j qual technol"] = "Journal of Quality Technology",
  ["j quant anal sports"] = "Journal of quantitative analysis in sports",
  ["j quant criminol"] = "Journal of quantitative criminology",
  ["j quant econ"] = "Journal of quantitative economics : journal of the Indian Econometric Society",
  ["j quant linguist"] = "Journal of quantitative linguistics",
  ["j quant spectrosc radiat transf"] = "Journal of quantitative spectroscopy & radiative transfer",
  ["j quant spectrosc radiat transfer"] = "Journal of Quantitative Spectroscopy and Radiative Transfer",
  ["j quant trait loci"] = "Journal of quantitative trait loci : JQTL",
  ["j quantum comput"] = "Journal of Quantum Computing",
  ["j quat sci"] = "Journal of quaternary science",
  ["j r aeronaut soc"] = "Journal of the Roy Aeronautical Society",
  ["j r anthropol inst"] = "The journal of the Royal Anthropological Institute",
  ["j r army med corps"] = "Journal of the Royal Army Medical Corps",
  ["j r army vet corps"] = "The Journal of the Royal Army Veterinary Corps",
  ["j r asiat soc gb irel"] = "Journal of the Royal Asiatic Society of Great Britain & Ireland. Royal Asiatic Society of Great Britain and Ireland",
  ["j r astron soc can"] = "Journal of the Royal Astronomical Society of Canada",
  ["j r aust hist soc"] = "Journal of the Royal Australian Historical Society",
  ["j r coll gen pract"] = "The Journal of the Royal College of General Practitioners",
  ["j r coll gen pract occas pap"] = "The Journal of the Royal College of General Practitioners. Occasional paper",
  ["j r coll physicians edinb"] = "The journal of the Royal College of Physicians of Edinburgh",
  ["j r coll physicians lond"] = "Journal of the Royal College of Physicians of London",
  ["j r coll surg edinb"] = "Journal of the Royal College of Surgeons of Edinburgh",
  ["j r fac med iraq"] = "The Journal of the Royal Faculty of Medicine of Iraq",
  ["j r i state dent soc"] = "Journal of the Rhode Island State Dental Society",
  ["j r inst chem"] = "Journal of the Royal Institute of Chemistry",
  ["j r inst public health"] = "The Journal of the Royal Institute of Public Health and Hygiene",
  ["j r microsc soc"] = "Journal of the Royal Microscopical Society",
  ["j r n s hist soc"] = "Journal of the Royal Nova Scotia Historical Society",
  ["j r nav med serv"] = "Journal of the Royal Naval Medical Service",
  ["j r sanit inst"] = "Journal. Royal Sanitary Institute (Great Britain)",
  ["j r soc antiq irel"] = "Journal of the Royal Society of Antiquaries of Ireland. Royal Society of Antiquaries of Ireland",
  ["j r soc arts"] = "Journal of the Royal Society of Arts",
  ["j r soc health"] = "Journal of the Royal Society of Health",
  ["j r soc interface"] = "Journal of the Royal Society, Interface",
  ["j r soc med"] = "Journal of the Royal Society of Medicine",
  ["j r soc n z"] = "Journal of the Royal Society of New Zealand",
  ["j r soc promot health"] = "The journal of the Royal Society for the Promotion of Health",
  ["j r soc west aust"] = "Journal of the Royal Society of Western Australia",
  ["j r stat soc"] = "Journal of the Royal Statistical Society",
  ["j r stat soc [ser a]"] = "Journal of the Royal Statistical Society. Series A (General)",
  ["j r stat soc a"] = "Journal of the Royal Statistical Society Series A (Statistics in Society)",
  ["j r stat soc b"] = "Journal of the Royal Statistical Society Series B (Statistical Methodology)",
  ["j r stat soc c"] = "Journal of the Royal Statistical Society Series C (Applied Statistics)",
  ["j r stat soc d"] = "Journal of the Royal Statistical Society Series D (The Statistician)",
  ["j r stat soc ser a"] = "Journal of the Royal Statistical Society. Series A (General)",
  ["j r stat soc ser a stat soc"] = "Journal of the Royal Statistical Society. Series A, (Statistics in Society)",
  ["j r stat soc ser b stat methodol"] = "Journal of the Royal Statistical Society. Series B. Statistical Methodology",
  ["j r stat soc ser c appl stat"] = "Journal of the Royal Statistical Society. Series C, Applied statistics",
  ["j r stat soc series b stat methodol"] = "Journal of the Royal Statistical Society. Series B, Statistical methodology",
  ["j race ethn polit"] = "Journal of race, ethnicity, and politics",
  ["j racial ethn health disparities"] = "Journal of racial and ethnic health disparities",
  ["j radars"] = "Journal of Radars",
  ["j radiat oncol"] = "Journal of radiation oncology",
  ["j radiat oncol res"] = "Journal of radiation oncology and research",
  ["j radiat res"] = "Journal of radiation research",
  ["j radiat res (tokyo)"] = "Journal of Radiation Research",
  ["j radiat res appl sci"] = "Journal of Radiation Research and Applied Sciences",
  ["j radio electron"] = "Journal of Radio Electronics",
  ["j radioanal nucl chem"] = "Journal of Radioanalytical and Nuclear Chemistry",
  ["j radioanal nucl chem art"] = "Journal of Radioanalytical and Nuclear Chemistry Articles",
  ["j radioanal nucl chem lett"] = "Journal of Radioanalytical and Nuclear Chemistry Letters",
  ["j radiol"] = "Journal de radiologie",
  ["j radiol case rep"] = "Journal of radiology case reports",
  ["j radiol electrol arch electr medicale"] = "Journal de radiologie, d'électrologie & archives d'électricité médicale",
  ["j radiol electrol med nucl"] = "Journal de radiologie, d'électrologie, et de médecine nucléaire",
  ["j radiol nurs"] = "Journal of radiology nursing",
  ["j radiol prot"] = "Journal of Radiological Protection",
  ["j radiol radiat ther"] = "Journal of radiology and radiation therapy",
  ["j radiosurg sbrt"] = "Journal of radiosurgery and SBRT",
  ["j radiother pract"] = "Journal of Radiotherapy in Practice",
  ["j rail rapid transit"] = "Journal of Rail and Rapid Transit",
  ["j rail transp plann manage"] = "Journal of Rail Transport Planning and Management",
  ["j rajasthan acad phys sci"] = "Journal of Rajasthan Academy of Physical Sciences",
  ["j raman spectrosc"] = "Journal of Raman Spectroscopy",
  ["j ramanujan math soc"] = "Journal of the Ramanujan Mathematical Society",
  ["j ramanujan soc math math sci"] = "Journal of Ramanujan Society of Mathematics and Mathematical Sciences",
  ["j rapid methods autom microbiol"] = "Journal of Rapid Methods and Automation in Microbiology",
  ["j raptor res"] = "The journal of raptor research",
  ["j rare dis res treat"] = "Journal of rare diseases research & treatment",
  ["j rare disord"] = "The Journal of rare disorders",
  ["j rare earths"] = "Journal of Rare Earths",
  ["j ration emot cogn behav ther"] = "Journal of rational-emotive and cognitive-behavior therapy : RET",
  ["j real estate finance econ"] = "Journal of Real Estate Finance and Economics",
  ["j real estate lit"] = "Journal of Real Estate Literature",
  ["j real estate portfol manage"] = "Journal of Real Estate Portfolio Management",
  ["j real estate practice educ"] = "Journal of Real Estate Practice and Education",
  ["j real estate res"] = "Journal of Real Estate Research",
  ["j real time image process"] = "Journal of real-time image processing",
  ["j real-time image process"] = "Journal of Real-Time Image Processing",
  ["j recept res"] = "Journal of receptor research",
  ["j recept signal transduct res"] = "Journal of receptor and signal transduction research",
  ["j recept signal transduction"] = "Journal of Receptors and Signal Transduction",
  ["j receptor ligand channel res"] = "Journal of receptor, ligand and channel research",
  ["j rech atmos"] = "Journal de Recherches Atmospheriques",
  ["j reconstr microsurg"] = "Journal of reconstructive microsurgery",
  ["j reconstr microsurg open"] = "Journal of reconstructive microsurgery open",
  ["j recreational math"] = "Journal of Recreational Mathematics",
  ["j reform jud"] = "Journal of Reform Judaism",
  ["j refract corneal surg"] = "Journal of refractive and corneal surgery",
  ["j refract surg"] = "Journal of refractive surgery (Thorofare, N.J. : 1995)",
  ["j refug stud"] = "Journal of refugee studies",
  ["j reg anal pol"] = "Journal of Regional Analysis and Policy",
  ["j reg med campuses"] = "Journal of regional medical campuses",
  ["j reg policy"] = "Journal of regional policy",
  ["j reg sci"] = "Journal of regional science",
  ["j regen med"] = "Journal of regenerative medicine",
  ["j regen med tissue eng"] = "Journal of regenerative medicine & tissue engineering",
  ["j registry manag"] = "Journal of registry management",
  ["j regul econ"] = "Journal of regulatory economics",
  ["j rehabil"] = "Journal of rehabilitation",
  ["j rehabil assist technol eng"] = "Journal of rehabilitation and assistive technologies engineering",
  ["j rehabil assistive technol eng"] = "Journal of Rehabilitation and Assistive Technologies Engineering",
  ["j rehabil med"] = "Journal of rehabilitation medicine",
  ["j rehabil med clin commun"] = "Journal of rehabilitation medicine. Clinical communications",
  ["j rehabil med suppl"] = "Journal of rehabilitation medicine. Supplement",
  ["j rehabil r d"] = "Journal of rehabilitation R&D",
  ["j rehabil res dev"] = "Journal of rehabilitation research and development",
  ["j rehabil res dev clin suppl"] = "Journal of rehabilitation research and development. Clinical supplement",
  ["j reine angew math"] = "Journal fur die Reine und Angewandte Mathematik",
  ["j reinf plast compos"] = "Journal of Reinforced Plastics and Composites",
  ["j relatsh res"] = "Journal of relationships research",
  ["j reliab intell environ"] = "Journal of reliable intelligent environments",
  ["j reliab sci eng"] = "Journal of Reliability Science and Engineering",
  ["j relig"] = "The Journal of religion",
  ["j relig afr"] = "Journal of religion in Africa. Religion en Afrique",
  ["j relig aging"] = "Journal of religion & aging",
  ["j relig ethics"] = "The Journal of religious ethics",
  ["j relig gerontol"] = "Journal of religious gerontology",
  ["j relig health"] = "Journal of religion and health",
  ["j relig hist"] = "The Journal of religious history",
  ["j relig spiritual aging"] = "Journal of religion, spirituality & aging",
  ["j relig spiritual soc work"] = "Journal of religion & spirituality in social work",
  ["j remanuf"] = "Journal of Remanufacturing",
  ["j remote sens"] = "Journal of Remote Sensing",
  ["j ren cancer"] = "Journal of renal cancer",
  ["j ren care"] = "Journal of renal care",
  ["j ren med"] = "Journal of renal medicine",
  ["j ren nutr"] = "Journal of renal nutrition : the official journal of the Council on Renal Nutrition of the National Kidney Foundation",
  ["j renal inj prev"] = "Journal of renal injury prevention",
  ["j renew mater"] = "Journal of renewable materials",
  ["j renew sustain energy"] = "Journal of Renewable & Sustainable Energy",
  ["j renewable energy"] = "Journal of Renewable Energy",
  ["j renewable energy environ"] = "Journal of Renewable Energy and Environment",
  ["j renewable mater"] = "Journal of Renewable Materials",
  ["j renewable sustainable energy"] = "Journal of Renewable and Sustainable Energy",
  ["j renin angiotensin aldosterone syst"] = "Journal of the renin-angiotensin-aldosterone system : JRAAS",
  ["j renin-angiotensin-aldosterone syst"] = "Journal of the Renin-Angiotensin-Aldosterone System",
  ["j reprod biol health"] = "Journal of reproductive biology and health",
  ["j reprod biotechnol fertil"] = "Journal of reproductive biotechnology & fertility",
  ["j reprod dev"] = "The Journal of reproduction and development",
  ["j reprod fertil"] = "Journal of reproduction and fertility",
  ["j reprod fertil abstr ser"] = "Journal of reproduction & fertility. Abstract series",
  ["j reprod fertil suppl"] = "Journal of reproduction and fertility. Supplement",
  ["j reprod health med"] = "Journal of reproductive health and medicine",
  ["j reprod immunol"] = "Journal of reproductive immunology",
  ["j reprod infant psychol"] = "Journal of reproductive and infant psychology",
  ["j reprod infertil"] = "Journal of reproduction & infertility",
  ["j reprod med"] = "The Journal of reproductive medicine",
  ["j res adm"] = "The journal of research administration",
  ["j res adolesc"] = "Journal of research on adolescence : the official journal of the Society for Research on Adolescence",
  ["j res appl sci"] = "Journal of research in applied sciences",
  ["j res biol sci"] = "Journal of research in biological sciences",
  ["j res child educ"] = "Journal of research in childhood education : JRCE",
  ["j res crime delinq"] = "The Journal of research in crime and delinquency",
  ["j res educ eff"] = "Journal of research on educational effectiveness",
  ["j res educ indian med"] = "The Journal of research and education in Indian medicine",
  ["j res for wildl environ"] = "Journal of research in forestry, wildlife and environment",
  ["j res health sci"] = "Journal of research in health sciences",
  ["j res indian med"] = "The Journal of research in Indian medicine",
  ["j res interprof pract educ"] = "Journal of research in interprofessional practice and education",
  ["j res islamic econ"] = "Journal of Research in Islamic Economics",
  ["j res lepid"] = "The journal of research on the Lepidoptera",
  ["j res math educ"] = "Journal for research in mathematics education",
  ["j res med dent sci"] = "Journal of research in medical and dental science",
  ["j res med sci"] = "Journal of research in medical sciences : the official journal of Isfahan University of Medical Sciences",
  ["j res nat bur stand"] = "Journal of Research of the National Bureau of Standards",
  ["j res nat inst stand technol"] = "Journal of Research of the National Institute of Standards and Technology",
  ["j res natl bur stand"] = "Journal of Research of the National Bureau of Standards",
  ["j res natl bur stand (1934)"] = "Journal of research of the National Bureau of Standards",
  ["j res natl bur stand (1977)"] = "Journal of research of the National Bureau of Standards (1977)",
  ["j res natl bur stand a phys chem"] = "Journal of research of the National Bureau of Standards. Section A, Physics and chemistry",
  ["j res natl bur stand sec a"] = "Journal of Research of the National Bureau of Standards, Section A: Physics and Chemistry",
  ["j res natl bur stand sec b"] = "Journal of Research of the National Bureau of Standards, Section B: Mathematical Sciences",
  ["j res natl bur stand sec c"] = "Journal of Research of the National Bureau of Standards, Section C: Engineering and Instrumentation",
  ["j res natl inst stand technol"] = "Journal of research of the National Institute of Standards and Technology",
  ["j res nurs"] = "Journal of research in nursing : JRN",
  ["j res pers"] = "Journal of research in personality",
  ["j res pharm pract"] = "Journal of research in pharmacy practice",
  ["j res pharmaceut econ"] = "Journal of Research in Pharmaceutical Economics",
  ["j res pract"] = "Journal of research practice",
  ["j res pract adult lit second basic educ"] = "Journal of research and practice for adult literacy, secondary, and basic education",
  ["j res read"] = "Journal of research in reading",
  ["j res sci teach"] = "Journal of research in science teaching",
  ["j res spec educ needs"] = "Journal of research in special educational needs : JORSEN",
  ["j res updates polym sci"] = "Journal of Research Updates in Polymer Science",
  ["j residuals sci technol"] = "Journal of Residuals Science and Technology",
  ["j resour ecol"] = "Journal of resources and ecology",
  ["j resour energy dev"] = "Journal of Resources, Energy and Development",
  ["j respir dis"] = "The Journal of respiratory diseases",
  ["j responsible innov"] = "Journal of responsible innovation",
  ["j responsible technol"] = "Journal of Responsible Technology",
  ["j restor med"] = "Journal of restorative medicine",
  ["j reticuloendothel soc"] = "Journal of the Reticuloendothelial Society",
  ["j retire"] = "The journal of retirement",
  ["j reward defic syndr"] = "Journal of reward deficiency syndrome",
  ["j reward defic syndr addict sci"] = "Journal of reward deficiency syndrome and addiction science",
  ["j rheol"] = "Journal of Rheology",
  ["j rheol (melville, ny, u s)"] = "Journal of Rheology (Melville, NY, United States)",
  ["j rheol (n y n y)"] = "Journal of rheology",
  ["j rheol (n y, ny, u s)"] = "Journal of Rheology (New York, NY, United States)",
  ["j rheum dis treat"] = "Journal of rheumatic diseases and treatment",
  ["j rheumatol"] = "The Journal of rheumatology",
  ["j rheumatol res"] = "Journal of rheumatology research",
  ["j rheumatol suppl"] = "The Journal of rheumatology. Supplement",
  ["j rice res"] = "Journal of rice research",
  ["j risk ins"] = "Journal of Risk and Insurance",
  ["j risk insur"] = "The Journal of risk and insurance",
  ["j risk issues"] = "The Journal of at-risk issues : a joint publication of the National Dropout Prevention Center and Network",
  ["j risk res"] = "Journal of risk research",
  ["j risk uncertain"] = "Journal of risk and uncertainty",
  ["j risk uncertainty"] = "Journal of Risk and Uncertainty",
  ["j risk uncertainty eng syst part a: civ eng"] = "Journal of Risk and Uncertainty in Engineering Systems, Part A: Civil Engineering",
  ["j risk uncertainty eng syst part b: mech eng"] = "Journal of Risk and Uncertainty in Engineering Systems, Part B: Mechanical Engineering",
  ["j rnai gene silencing"] = "Journal of RNAi and gene silencing : an international journal of RNA and gene targeting research",
  ["j road eng"] = "Journal of Road Engineering",
  ["j road saf"] = "Journal of Road Safety",
  ["j rob"] = "Journal of Robotics",
  ["j rob surg"] = "Journal of Robotic Surgery",
  ["j rob syst"] = "Journal of Robotic Systems",
  ["j rob, networking artif life"] = "Journal of Robotics, Networking and Artificial Life",
  ["j robert h lurie cancer cent northwest univ"] = "The journal of the Robert H. Lurie Cancer Center of Northwestern University",
  ["j robot netw artif life"] = "Journal of Robotics, Networking and Artificial Life",
  ["j robot surg"] = "Journal of robotic surgery",
  ["j robot syst"] = "Journal of robotic systems",
  ["j rock mech geotech eng"] = "Journal of Rock Mechanics and Geotechnical Engineering",
  ["j rocky mt analg soc"] = "Journal of the Rocky Mountain Analgesia Society",
  ["j rocky mt mediev renaiss assoc"] = "Journal of the Rocky Mountain Medieval and Renaissance Association. Rocky Mountain Medieval and Renaissance Association",
  ["j rom lit stud"] = "Journal of Romanian literary studies [electronic resource]",
  ["j roy statist soc ser a"] = "Journal of the Royal Statistical Society. Series A. Statistics in Society",
  ["j roy statist soc ser b"] = "Journal of the Royal Statistical Society",
  ["j roy statist soc ser c"] = "Journal of the Royal Statistical Society",
  ["j roy statistical society"] = "Journal of the Royal Statistical Society, Series A",
  ["j rubber res"] = "Journal of rubber research (Kuala Lumpur, Malaysia)",
  ["j rural community dev"] = "Journal of rural and community development",
  ["j rural community psychiatr"] = "Journal of rural and community psychiatry",
  ["j rural community psychol"] = "Journal of rural community psychology (Online)",
  ["j rural dev"] = "Journal Of Rural Development",
  ["j rural health"] = "The Journal of rural health : official journal of the American Rural Health Association and the National Rural Health Care Association",
  ["j rural med"] = "Journal of rural medicine : JRM",
  ["j rural soc sci"] = "Journal of rural social sciences",
  ["j rural stud"] = "Journal of rural studies",
  ["j russ laser res"] = "Journal of Russian Laser Research",
  ["j russ phytopathol soc"] = "Journal of Russian Phytopathological Society",
  ["j russell soc"] = "Journal of the Russell Society",
  ["j rutgers univ libr"] = "The Journal of the Rutgers University Library",
  ["j s afr chem inst"] = "Journal of the South African Chemical Institute. South African Chemical Institute",
  ["j s afr inst civ eng"] = "Journal of the South African Institution of Civil Engineering",
  ["j s afr inst min metall"] = "Journal of the South African Institute of Mining and Metallurgy",
  ["j s afr logop soc"] = "Journal of the South African Logopedic Society",
  ["j s afr speech hear assoc"] = "Journal of the South African Speech and Hearing Association",
  ["j s afr vet assoc"] = "Journal of the South African Veterinary Association",
  ["j s afr vet med assoc"] = "Journal of the South African Veterinary Medical Association",
  ["j s c acad sci"] = "Journal of the South Carolina Academy of Science",
  ["j s c med assoc"] = "Journal of the South Carolina Medical Association (1975)",
  ["j saf health environ res"] = "Journal of safety, health and environmental research",
  ["j saf res"] = "Journal of Safety Research",
  ["j saf sci manage"] = "Journal of Safety Science and Management",
  ["j saf sci resilience"] = "Journal of Safety Science and Resilience",
  ["j safety res"] = "Journal of safety research",
  ["j san antonio dent soc"] = "Journal. San Antonio District Dental Society",
  ["j san diego hist"] = "The Journal of San Diego history",
  ["j sandwich struct mater"] = "Journal of Sandwich Structures and Materials",
  ["j satisf boolean model comput"] = "Journal on Satisfiability, Boolean Modeling and Computation",
  ["j saudi chem soc"] = "Journal of Saudi Chemical Society",
  ["j saudi heart assoc"] = "Journal of the Saudi Heart Association",
  ["j saudi soc agric sci"] = "Journal of the Saudi Society of Agricultural Sciences",
  ["j scand stud criminol crime prev"] = "Journal of Scandinavian studies in criminology and crime prevention",
  ["j sch choice"] = "Journal of school choice",
  ["j sch health"] = "The Journal of school health",
  ["j sch leadersh"] = "Journal of school leadership",
  ["j sch nurs"] = "The Journal of school nursing : the official publication of the National Association of School Nurses",
  ["j sch psychol"] = "Journal of school psychology",
  ["j sch violence"] = "Journal of school violence",
  ["j sched"] = "Journal of Scheduling",
  ["j schizophr res"] = "Journal of schizophrenia research",
  ["j sci"] = "Journal of science",
  ["j sci adv mater devices"] = "Journal of Science: Advanced Materials and Devices",
  ["j sci comput"] = "Journal of Scientific Computing",
  ["j sci conf proc"] = "Journal of Scientific Conference Proceedings",
  ["j sci dent"] = "Journal of scientific dentistry",
  ["j sci educ technol"] = "Journal of Science Education and Technology",
  ["j sci eng technol"] = "Journal of science, engineering and technology",
  ["j sci explor"] = "Journal of scientific exploration : a publication of the Society for Scientific Exploration",
  ["j sci food agric"] = "Journal of the science of food and agriculture",
  ["j sci ind res"] = "Journal of Scientific and Industrial Research",
  ["j sci ind res (1942)"] = "Journal of scientific and industrial research (New Delhi, India : 1942)",
  ["j sci ind res (c)"] = "Journal of scientific & industrial research. C, Biological sciences",
  ["j sci ind res (india)"] = "Journal of scientific and industrial research",
  ["j sci instrum"] = "Journal of Scientific Instruments",
  ["j sci islam repub iran"] = "Islamic Republic of Iran. Journal of Sciences",
  ["j sci med"] = "Journal of science & medicine",
  ["j sci med lille"] = "Journal des sciences médicales de Lille",
  ["j sci med sport"] = "Journal of science and medicine in sport",
  ["j sci policy gov"] = "The journal of science policy & governance",
  ["j sci res chulalongkorn univ"] = "Wārsān wičhai witthayāsāt",
  ["j sci res med sci"] = "Journal for scientific research. Medical sciences",
  ["j sci res rep"] = "Journal of scientific research and reports",
  ["j sci soc thailand"] = "Journal of the Science Society of Thailand",
  ["j sci sport exercise"] = "Journal of Science in Sport and Exercise",
  ["j sci stud relig"] = "Journal For The Scientific Study Of Religion",
  ["j sci study relig"] = "Journal for the scientific study of religion",
  ["j sci teach educ"] = "Journal of Science Teacher Education",
  ["j sci teacher educ"] = "Journal of science teacher education",
  ["j sci technol environ"] = "Journal of science, technology and environment",
  ["j sci technol policy manage"] = "Journal of Science and Technology Policy Management",
  ["j sci technol tropics"] = "Journal of science & technology in the tropics",
  ["j sci univ tehran int ed"] = "Journal of Science of the University of Tehran",
  ["j sci: adv mater devices"] = "Journal of Science: Advanced Materials and Devices",
  ["j scleroderma relat disord"] = "Journal of scleroderma and related disorders",
  ["j scott hist stud"] = "Journal Of Scottish Historical Studies",
  ["j scott labour hist soc"] = "Journal - Scottish Labour History Society. Scottish Labour History Society",
  ["j sea res"] = "Journal of Sea Research",
  ["j seattle dist dent soc"] = "Journal of the Seattle District Dental Society",
  ["j seattle king cty dent soc"] = "Journal - Seattle-King County Dental Society",
  ["j second dist dent soc state n y"] = "Journal. Second District Dental Society of the State of New York",
  ["j second lang pronunciation"] = "Journal of second language pronunciation",
  ["j second lang writ"] = "Journal of second language writing",
  ["j sediment petrol"] = "Journal of Sedimentary Petrology",
  ["j sediment res"] = "Journal of Sedimentary Research",
  ["j sediment res a sediment petrol process"] = "Journal of sedimentary research. Section A, Sedimentary petrology and processes : an international journal of SEPM (Society for Sedimentary Geology)",
  ["j seed sci"] = "Journal of Seed Science",
  ["j seismic explor"] = "Journal of Seismic Exploration",
  ["j seismol"] = "Journal of seismology",
  ["j seismolog"] = "Journal of Seismology",
  ["j semant"] = "Journal of semantics",
  ["j semicond"] = "Journal of Semiconductors",
  ["j semicond technol sci"] = "Journal of Semiconductor Technology and Science",
  ["j semit stud"] = "Journal of Semitic studies",
  ["j sens"] = "Journal of sensors",
  ["j sens sens syst"] = "Journal of Sensors and Sensor Systems",
  ["j sens stud"] = "Journal of sensory studies",
  ["j sens technol"] = "Journal of sensor technology",
  ["j sep sci"] = "Journal of separation science",
  ["j serb chem soc"] = "Journal of the Serbian Chemical Society",
  ["j serbian soc comput mech"] = "Journal of the Serbian Society for Computational Mechanics",
  ["j serv res"] = "Journal of service research",
  ["j sex aggress"] = "The journal of sexual aggression",
  ["j sex educ"] = "The Journal of sex education",
  ["j sex educ ther"] = "Journal of sex education and therapy",
  ["j sex marital ther"] = "Journal of sex & marital therapy",
  ["j sex med"] = "The journal of sexual medicine",
  ["j sex res"] = "Journal of sex research",
  ["j sex transm dis"] = "Journal of sexually transmitted diseases",
  ["j sfds"] = "Journal de la SFdS. Journal de la Societe Francaise de Statistique",
  ["j shaanxi normal univ nat sci ed"] = "Shaanxi Normal University. Journal. Natural Science Edition. Shaanxi Shifan Daxue Xuebao. Ziran Kexue Ban",
  ["j shandong univ nat sci"] = "Journal of Shandong University. Natural Science. Shandong Daxue Xuebao. Lixue Ban",
  ["j shanghai inst railway tech"] = "Shanghai Institute of Railway Technology",
  ["j shanghai jiaotong univ (chin ed)"] = "Journal of Shanghai Jiaotong University. Chinese Edition. Shanghai Jiaotong Daxue Xuebao",
  ["j shanghai jiaotong univ (engl ed)"] = "Journal of Shanghai Jiaotong University",
  ["j shanghai jiaotong univ sci"] = "Journal of Shanghai Jiaotong University (science)",
  ["j shanghai univ"] = "Journal of Shanghai University",
  ["j shanghai univ nat sci"] = "Journal of Shanghai University. Natural Science Edition. Shanghai Daxue Xuebao. Ziran Kexue Ban",
  ["j shellfish res"] = "Journal of shellfish research",
  ["j ship prod des"] = "Journal of Ship Production and Design",
  ["j ship res"] = "Journal of Ship Research",
  ["j shoulder elb arthroplast"] = "Journal of shoulder and elbow arthroplasty",
  ["j shoulder elbow surg"] = "Journal of shoulder and elbow surgery",
  ["j siam soc"] = "The Journal of the Siam Society",
  ["j sichuan union univ eng sci ed"] = "Journal of Sichuan Union University",
  ["j sichuan univ eng sci ed"] = "Journal of Sichuan University. Engineering Science Edition. Sichuan Daxue Xuebao. Gongcheng Kexue Ban",
  ["j sick cell dis hemoglobinopathies"] = "Journal of sickle cell disease and hemoglobinopathies",
  ["j signal process syst"] = "Journal of Signal Processing Systems",
  ["j signal transduct"] = "Journal of signal transduction",
  ["j simul"] = "Journal of simulation : JOS",
  ["j sing"] = "Journal of singing : the official journal of the National Association of Teachers of Singing",
  ["j singapore paediatr soc"] = "The Journal of the Singapore Paediatric Society",
  ["j singul"] = "Journal of Singularities",
  ["j skin cancer"] = "Journal of skin cancer",
  ["j skyscape archaeol"] = "Journal of Skyscape Archaeology",
  ["j sleep disord manag"] = "Journal of sleep disorders and management",
  ["j sleep disord ther"] = "Journal of sleep disorders & therapy",
  ["j sleep disord treat care"] = "Journal of sleep disorders-- treatment & care",
  ["j sleep med disord"] = "Journal of sleep medicine and disorders",
  ["j sleep res"] = "Journal of sleep research",
  ["j small anim pract"] = "The Journal of small animal practice",
  ["j smart cities soc"] = "Journal of Smart Cities and Society",
  ["j smok cess"] = "Journal of Smoking Cessation",
  ["j smok cessat"] = "Journal of smoking cessation",
  ["j smooth muscle res"] = "Journal of smooth muscle research = Nihon Heikatsukin Gakkai kikanshi",
  ["j soc action couns psychol"] = "Journal for social action in counseling and psychology",
  ["j soc adm pharm"] = "Journal of social and administrative pharmacy : JSAP",
  ["j soc alg chim"] = "Journal de la Societe Algerienne de Chimie",
  ["j soc am"] = "Journal de la Société des américanistes",
  ["j soc arch"] = "Journal of the Society of Archivists. Society of Archivists (Great Britain)",
  ["j soc army hist res"] = "Journal of the Society for Army Historical Research. Society for Army Historical Research (London, England)",
  ["j soc behav health sci"] = "Journal of social, behavioral and health sciences",
  ["j soc behav pers"] = "Journal of social behavior and personality",
  ["j soc biol"] = "Journal de la Societe de Biologie",
  ["j soc biol struct"] = "Journal of social and biological structures",
  ["j soc bras fonoaudiol"] = "Jornal da Sociedade Brasileira de Fonoaudiologia",
  ["j soc casework"] = "Journal of social casework",
  ["j soc chem ind london"] = "Journal of the Society of Chemical Industry, London",
  ["j soc chim tunis"] = "Journal de la Societe Chimique de Tunisie",
  ["j soc christ ethics"] = "Journal of the Society of Christian Ethics",
  ["j soc cienc med lisb"] = "Jornal da Sociedade das Ciencias Medicas de Lisboa",
  ["j soc clin psychol"] = "Journal of social and clinical psychology",
  ["j soc cosmet chem"] = "Journal of the Society of Cosmetic Chemists",
  ["j soc dev afr"] = "Journal of social development in Africa",
  ["j soc distress homeless"] = "Journal of social distress and the homeless",
  ["j soc dyers colour"] = "Journal of the Society of Dyers and Colourists",
  ["j soc econ devel"] = "Journal of Social and Economic Development",
  ["j soc econ stud"] = "Journal of social and economic studies",
  ["j soc evol cult psychol"] = "Journal of social, evolutionary & cultural psychology : JSEC",
  ["j soc fr statistique (2009)"] = "Journal de la Société française de statistique (2009)",
  ["j soc glass technol"] = "Journal of the Society of Glass Technology",
  ["j soc gynecol investig"] = "Journal of the Society for Gynecologic Investigation",
  ["j soc health syst"] = "Journal of the Society for Health Systems",
  ["j soc hist"] = "Journal of social history",
  ["j soc hyg"] = "Journal of social hygiene",
  ["j soc inf disp"] = "Journal of the Society for Information Display",
  ["j soc inorg mater, jpn"] = "Journal of the Society of Inorganic Materials, Japan",
  ["j soc integr oncol"] = "Journal of the Society for Integrative Oncology",
  ["j soc issues"] = "The Journal of social issues",
  ["j soc laparoendosc surg"] = "Journal of the Society of Laparoendoscopic Surgeons",
  ["j soc leather technol chem"] = "Journal of the Society of Leather Technologists and Chemists",
  ["j soc mark"] = "Journal of social marketing",
  ["j soc mater sci, jpn"] = "Journal of the Society of Materials Science, Japan",
  ["j soc mech eng int j sera:"] = "Japan Society of Mechanical Engineers International Journal Series A: Solid Mechanics and Material Engineering",
  ["j soc mech eng int j serb:"] = "Japan Society of Mechanical Engineers International Journal Series B: Fluids and Thermal Engineering",
  ["j soc mech eng int j serc:"] = "Japan Society of Mechanical Engineers International Journal Series C: Mechanical Systems, Machine Elements and Manufacturing",
  ["j soc occup med"] = "The Journal of the Society of Occupational Medicine",
  ["j soc ocean"] = "Journal de la Société des océanistes",
  ["j soc ouest afr chim"] = "Journal de la Societe Ouest Africaine de Chimie",
  ["j soc pediatr nurs"] = "Journal of the Society of Pediatric Nurses : JSPN",
  ["j soc pers relat"] = "Journal of social and personal relationships",
  ["j soc philos"] = "Journal of social philosophy",
  ["j soc policy"] = "Journal of social policy",
  ["j soc polit aff"] = "The Journal of social and political affairs",
  ["j soc polit econ stud"] = "The Journal of social, political, and economic studies",
  ["j soc polit stud"] = "The Journal of social and political studies",
  ["j soc psychol"] = "The Journal of social psychology",
  ["j soc radiol prot"] = "Journal of the Society for Radiological Protection",
  ["j soc res adm"] = "Journal of the Society of Research Administrators. Society of Research Administrators",
  ["j soc sci"] = "Journal of the Social Sciences",
  ["j soc serv res"] = "Journal of social service research",
  ["j soc soc policy"] = "Journal of societal & social policy",
  ["j soc social work res"] = "Journal of the Society for Social Work and Research",
  ["j soc stat paris"] = "Journal de la Société de statistique de Paris",
  ["j soc struct"] = "Journal of social structure : JOSS",
  ["j soc stud"] = "The Journal of social studies",
  ["j soc telegraph eng"] = "Journal of the Society of Telegraph Engineers",
  ["j soc welf fam law"] = "The Journal of social welfare & family law",
  ["j soc welfare law"] = "The Journal of social welfare law",
  ["j soc work (lond)"] = "Journal of social work (London, England)",
  ["j soc work disabil rehabil"] = "Journal of social work in disability & rehabilitation",
  ["j soc work educ"] = "Journal of social work education",
  ["j soc work end life palliat care"] = "Journal of social work in end-of-life & palliative care",
  ["j soc work glob community"] = "Journal of social work in the global community",
  ["j soc work hum sex"] = "Journal of social work & human sexuality",
  ["j soc work pract"] = "Journal of social work practice",
  ["j soc work pract addict"] = "Journal of social work practice in the addictions",
  ["j soc work values ethics"] = "Journal of social work values and ethics",
  ["j social archaeol"] = "Journal of Social Archaeology",
  ["j social comput"] = "Journal of Social Computing",
  ["j socio econ"] = "The Journal of socio-economics",
  ["j socio-econ"] = "Journal of Socio-Economics",
  ["j sociol"] = "Journal Of Sociology",
  ["j sociol (melb)"] = "Journal of sociology (Melbourne, Vic.)",
  ["j sociol soc welf"] = "Journal of sociology and social welfare",
  ["j sociol stud"] = "The Journal of sociological studies",
  ["j softw"] = "Journal of Software. Ruanjian Xuebao",
  ["j softw (malden)"] = "Journal of software (Malden, MA)",
  ["j softw algebra geom"] = "Journal of Software for Algebra and Geometry",
  ["j software eng res dev"] = "Journal of Software Engineering Research and Development",
  ["j software: evol process"] = "Journal of Software: Evolution and Process",
  ["j sogc"] = "Journal SOGC : journal of the Society of Obstetricians and Gynaecologists of Canada",
  ["j soil sci"] = "Journal of Soil Science",
  ["j soil sci plant nutr"] = "Journal of soil science and plant nutrition",
  ["j soil water conserv"] = "Journal of Soil and Water Conservation",
  ["j soils sediments"] = "Journal of Soils and Sediments",
  ["j sol energy eng"] = "Journal of Solar Energy Engineering",
  ["j sol energy soc india"] = "Journal of the Solar Energy Society of India",
  ["j sol-gel sci technol"] = "Journal of Sol-Gel Science and Technology",
  ["j solgel sci technol"] = "Journal of sol-gel science and technology",
  ["j solid state chem"] = "Journal of Solid State Chemistry",
  ["j solid state electrochem"] = "Journal of Solid State Electrochemistry",
  ["j solid tumors"] = "Journal of solid tumors",
  ["j solution chem"] = "Journal of Solution Chemistry",
  ["j sound vib"] = "Journal of Sound and Vibration",
  ["j sound vibration"] = "Journal of Sound and Vibration",
  ["j south afr inst civ eng"] = "Journal of the South African Institute of Civil Engineering",
  ["j south afr inst civil eng"] = "Journal of the South African Institute of Civil Engineering",
  ["j south afr inst min metall"] = "Journal of the Southern African Institute of Mining and Metallurgy",
  ["j south afr stud"] = "Journal of southern African studies",
  ["j south am earth sci"] = "Journal of South American earth sciences",
  ["j south amer earth sci"] = "Journal of South American Earth Sciences",
  ["j south asian middle east stud"] = "Journal of South Asian and Middle Eastern studies",
  ["j south calif dent assistants assoc"] = "Journal (Southern California Dental Assistants Association)",
  ["j south calif dent assoc"] = "Journal - Southern California Dental Association",
  ["j south calif state dent assoc"] = "Journal - Southern California State Dental Association",
  ["j south calif state dent hyg assoc"] = "Journal of the Southern California Dental Hygienists' Association",
  ["j south china normal univ natur sci ed"] = "Journal of South China Normal University. Natural Science Edition. Huanan Shifan Daxue Xuebao. Ziran Kexue Ban",
  ["j south hemisph earth syst sci"] = "Journal of Southern Hemisphere Earth Systems Science",
  ["j south hist"] = "The Journal of southern history",
  ["j south orthop assoc"] = "Journal of the Southern Orthopaedic Association",
  ["j southeast asian am educ adv"] = "Journal of Southeast Asian American education & advancement : JSAAEA",
  ["j southeast asian earth sci"] = "Journal of Southeast Asian Earth Sciences",
  ["j southeast asian stud"] = "Journal of Southeast Asian studies",
  ["j southeast univ"] = "Journal of Southeast University",
  ["j southeast univ (english ed)"] = "Journal of Southeast University. English Edition. Dongnan Daxue Xuebao. Yingwen Ban",
  ["j southeast univ nat sci"] = "Journal of Southeast University. Natural Science Edition. Dongnan Daxue Xuebao. Ziran Kexue Ban",
  ["j southwest"] = "Journal of the Southwest",
  ["j sov natly"] = "Journal of Soviet nationalities : a quarterly publication of the Center on East-West Trade, Investment, and Communications",
  ["j soz forsch"] = "Journal für Sozialforschung",
  ["j space saf eng"] = "The journal of space safety engineering",
  ["j space weather space clim"] = "Journal of Space Weather and Space Climate",
  ["j spacecr rockets"] = "Journal of spacecraft and rockets",
  ["j spat sci"] = "Journal of spatial science",
  ["j spatial econom"] = "Journal of Spatial Econometrics",
  ["j spatial sci"] = "Journal of Spatial Science",
  ["j spec educ"] = "The Journal of special education",
  ["j spec educ technol"] = "Journal of special education technology : a publication of Utah State University, the Association for Special Education Technology, and the Technology and Media Division of the Council for Exceptional Children",
  ["j spec oper med"] = "Journal of special operations medicine : a peer reviewed journal for SOF medical professionals",
  ["j spec pediatr nurs"] = "Journal for specialists in pediatric nursing : JSPN",
  ["j special oper med"] = "Journal of Special Operations Medicine",
  ["j species res"] = "Journal of species research",
  ["j spectr theory"] = "Journal of Spectral Theory",
  ["j spectral theory"] = "Journal of Spectral Theory",
  ["j spectro"] = "Journal of Spectroscopy",
  ["j spectrosc (hindawi)"] = "Journal of spectroscopy (Hindawi)",
  ["j spectrosc (n y, ny, u s)"] = "Journal of Spectroscopy (New York, NY, United States)",
  ["j spectrosc dyn"] = "Journal of spectroscopy and dynamics",
  ["j speech disord"] = "The Journal of speech disorders",
  ["j speech hear disord"] = "The Journal of speech and hearing disorders",
  ["j speech hear disord monogr suppl"] = "Journal of speech and hearing disorders. Monograph supplement",
  ["j speech hear res"] = "Journal of speech and hearing research",
  ["j speech lang hear res"] = "Journal of speech, language, and hearing research : JSLHR",
  ["j speech lang pathol appl behav anal"] = "The journal of speech and language pathology, applied behavior analysis",
  ["j speech pathol ther"] = "Journal of speech pathology & therapy",
  ["j spinal cord med"] = "The journal of spinal cord medicine",
  ["j spinal disord"] = "Journal of spinal disorders",
  ["j spinal disord tech"] = "Journal of spinal disorders & techniques",
  ["j spine"] = "Journal of spine",
  ["j spine neurosurg"] = "Journal of spine & neurosurgery",
  ["j spine surg"] = "Journal of spine surgery (Hong Kong)",
  ["j spiritual ment health"] = "Journal of spirituality in mental health",
  ["j sport behav"] = "Journal of sport behavior",
  ["j sport dev"] = "Journal of sport for development",
  ["j sport exerc psychol"] = "Journal of sport & exercise psychology",
  ["j sport health sci"] = "Journal of sport and health science",
  ["j sport hist"] = "Journal of sport history",
  ["j sport psychol action"] = "Journal of sport psychology in action",
  ["j sport rehabil"] = "Journal of sport rehabilitation",
  ["j sport soc issues"] = "Journal of sport and social issues",
  ["j sports anal"] = "Journal of Sports Analytics",
  ["j sports econ"] = "Journal of Sports Economics",
  ["j sports econom"] = "Journal of sports economics",
  ["j sports med"] = "The Journal of sports medicine",
  ["j sports med (hindawi publ corp)"] = "Journal of sports medicine (Hindawi Publishing Corporation)",
  ["j sports med doping stud"] = "Journal of sports medicine & doping studies",
  ["j sports med phys fitness"] = "The Journal of sports medicine and physical fitness",
  ["j sports sci"] = "Journal of sports sciences",
  ["j sports sci med"] = "Journal of sports science & medicine",
  ["j sri lanka branch r asiat soc"] = "Journal of the Sri Lanka Branch of the Royal Asiatic Society",
  ["j stat appl probab"] = "Journal of statistics applications & probability",
  ["j stat comput simul"] = "Journal of Statistical Computation and Simulation",
  ["j stat data sci educ"] = "Journal of Statistics and Data Science Education",
  ["j stat distrib appl"] = "Journal of statistical distributions and applications",
  ["j stat educ"] = "Journal of statistics education : an international journal on the teaching and learning of statistics",
  ["j stat manag syst"] = "Journal of Statistics & Management Systems",
  ["j stat manage syst"] = "Journal of Statistics and Management Systems",
  ["j stat mech"] = "Journal of statistical mechanics (Online)",
  ["j stat mech theory exp"] = "Journal of Statistical Mechanics: Theory and Experiment",
  ["j stat mech: theory exp"] = "Journal of Statistical Mechanics: Theory and Experiment",
  ["j stat phys"] = "Journal of Statistical Physics",
  ["j stat plan inference"] = "Journal of statistical planning and inference",
  ["j stat plann inference"] = "Journal of Statistical Planning and Inference",
  ["j stat res"] = "Journal of statistical research",
  ["j stat sci appl"] = "Journal of statistical science and application",
  ["j stat softw"] = "Journal of statistical software",
  ["j stat stud"] = "Journal of Statistical Studies",
  ["j stat theory appl"] = "Journal of Statistical Theory and Applications",
  ["j stat theory pract"] = "Journal of Statistical Theory and Practice",
  ["j state gov"] = "The Journal of state government",
  ["j statist comput simulation"] = "Journal of Statistical Computation and Simulation",
  ["j statist phys"] = "Journal of Statistical Physics",
  ["j statist plann inference"] = "Journal of Statistical Planning and Inference",
  ["j statist res"] = "Journal of Statistical Research",
  ["j statistical soc inquiry society ireland"] = "Journal of the Statistical and Social Inquiry Society of Ireland",
  ["j stem  educ"] = "Journal of STEM education : innovations and research",
  ["j stem cell regen biol"] = "Journal of stem cell and regenerative biology",
  ["j stem cell res (overl park)"] = "Journal of stem cell research",
  ["j stem cell res ther"] = "Journal of stem cell research & therapy",
  ["j stem cell res ther (edmond)"] = "Journal of stem cell research & therapeutics",
  ["j stem cell res transplant"] = "Journal of stem cell research and transplantation",
  ["j stem cell ther transplant"] = "Journal of stem cell therapy and transplantation",
  ["j stem cell transplant biol"] = "Journal of stem cell and transplantation biology",
  ["j stem cells"] = "Journal of stem cells",
  ["j stem cells regen med"] = "Journal of stem cells & regenerative medicine",
  ["j stem educ res"] = "Journal for STEM education research",
  ["j stem outreach"] = "Journal of STEM outreach",
  ["j sterile serv manage"] = "Journal of sterile services management",
  ["j steroid biochem"] = "Journal of Steroid Biochemistry",
  ["j steroid biochem mol biol"] = "Journal of Steroid Biochemistry and Molecular Biology",
  ["j steroids horm sci"] = "Journal of steroids & hormonal science",
  ["j stoch anal"] = "Journal of Stochastic Analysis. JOSA",
  ["j stomatol belg"] = "Journal de stomatologie de Belgique",
  ["j stomatol oral maxillofac surg"] = "Journal of stomatology, oral and maxillofacial surgery",
  ["j stone dis"] = "The Journal of stone disease",
  ["j stored prod res"] = "Journal of Stored Products Research",
  ["j strain anal eng des"] = "Journal of Strain Analysis for Engineering Design",
  ["j strategic inf syst"] = "Journal of Strategic Information Systems",
  ["j strength cond res"] = "Journal of strength and conditioning research",
  ["j stroke"] = "Journal of stroke",
  ["j stroke cerebrovasc dis"] = "Journal of stroke and cerebrovascular diseases : the official journal of National Stroke Association",
  ["j struct biol"] = "Journal of structural biology",
  ["j struct biol x"] = "Journal of structural biology: X",
  ["j struct biol: x"] = "Journal of Structural Biology: X",
  ["j struct chem"] = "Journal of Structural Chemistry",
  ["j struct control"] = "Journal of Structural Control",
  ["j struct des"] = "Journal of Structural Design",
  ["j struct eng"] = "Journal of Structural Engineering",
  ["j struct eng (n y n y)"] = "Journal of structural engineering (New York, N.Y.)",
  ["j struct fire eng"] = "Journal of Structural Fire Engineering",
  ["j struct funct genomics"] = "Journal of structural and functional genomics",
  ["j struct geol"] = "Journal of Structural Geology",
  ["j struct heart dis"] = "Journal of structural heart disease",
  ["j struct integrity maint"] = "Journal of Structural Integrity and Maintenance",
  ["j stud aff res pract"] = "Journal of student affairs research and practice",
  ["j stud alcohol"] = "Journal of studies on alcohol",
  ["j stud alcohol drugs"] = "Journal of studies on alcohol and drugs",
  ["j stud alcohol drugs suppl"] = "Journal of studies on alcohol and drugs. Supplement",
  ["j stud alcohol suppl"] = "Journal of studies on alcohol. Supplement",
  ["j stud econ econometrics"] = "Journal for Studies in Economics and Econometrics",
  ["j stud int educ"] = "Journal of studies in international education",
  ["j stud phys ther res"] = "Journal of student physical therapy research",
  ["j stud soc sci"] = "Journal of studies in social sciences",
  ["j study new testam"] = "Journal for the study of the New Testament",
  ["j study relig exp"] = "Journal for the study of religious experience",
  ["j stutt ther advocacy res"] = "The journal of stuttering therapy, advocacy, and research",
  ["j submicrosc cytol"] = "Journal of submicroscopic cytology",
  ["j submicrosc cytol pathol"] = "Journal of submicroscopic cytology and pathology",
  ["j subst abus alcohol"] = "Journal of substance abuse and alcoholism",
  ["j subst abuse"] = "Journal of Substance Abuse",
  ["j subst abuse treat"] = "Journal of substance abuse treatment",
  ["j subst use"] = "Journal of substance use",
  ["j sugar beet res"] = "Journal of sugar beet research",
  ["j sulfur chem"] = "Journal of Sulfur Chemistry",
  ["j sulphur chem"] = "Journal of sulphur chemistry",
  ["j supercomput"] = "The Journal of supercomputing",
  ["j supercond"] = "Journal of Superconductivity",
  ["j supercond nov magn"] = "Journal of superconductivity and novel magnetism",
  ["j supercond novel magn"] = "Journal of Superconductivity and Novel Magnetism",
  ["j supercrit fluids"] = "Journal of Supercritical Fluids",
  ["j superhard mater"] = "Journal of Superhard Materials",
  ["j support oncol"] = "The journal of supportive oncology",
  ["j supramol chem"] = "Journal of Supramolecular Chemistry",
  ["j supramol struct"] = "Journal of supramolecular structure",
  ["j supramol struct cell biochem"] = "Journal of supramolecular structure and cellular biochemistry",
  ["j supramol struct cell biochem suppl"] = "Journal of supramolecular structure and cellular biochemistry. Supplement",
  ["j supramol struct suppl"] = "Journal of supramolecular structure. Supplement",
  ["j surf anal"] = "Journal of surface analysis (Online)",
  ["j surf eng mater adv technol"] = "Journal of surface engineered materials and advanced technology",
  ["j surf invest"] = "Journal of Surface Investigation: X-ray, Synchrotron and Neutron Techniques",
  ["j surf invest: x-ray, synchrotron neutron tech"] = "Journal of Surface Investigation: X-Ray, Synchrotron and Neutron Techniques",
  ["j surf investig x-ray synchrotron neutron tech"] = "Journal of Surface Investigation: X-ray, Synchrotron and Neutron Techniques",
  ["j surfactants deterg"] = "Journal of Surfactants and Detergents",
  ["j surg (lisle)"] = "Journal of surgery (Lisle, IL)",
  ["j surg (n y n y)"] = "Journal of surgery (New York, N.Y.)",
  ["j surg (northborough)"] = "Journal of surgery (Northborough, Mass.)",
  ["j surg case rep"] = "Journal of surgical case reports",
  ["j surg educ"] = "Journal of surgical education",
  ["j surg emerg med"] = "Journal of surgery and emergency medicine",
  ["j surg oncol"] = "Journal of surgical oncology",
  ["j surg oncol (tallinn)"] = "Journal of surgical oncology (Tallinn, Estonia)",
  ["j surg oncol suppl"] = "Journal of surgical oncology. Supplement",
  ["j surg orthop adv"] = "Journal of surgical orthopaedic advances",
  ["j surg radiol"] = "Journal of surgical radiology",
  ["j surg res"] = "The Journal of surgical research",
  ["j surg res (houst)"] = "Journal of surgery and research",
  ["j surg sci"] = "Journal of surgery and science",
  ["j surg tech case rep"] = "Journal of surgical technique and case report",
  ["j surg urol"] = "Journal of surgical urology",
  ["j surv eng"] = "Journal of Surveying Engineering",
  ["j surv fish sci"] = "Journal of survey in fisheries sciences",
  ["j surv stat methodol"] = "Journal of survey statistics and methodology",
  ["j sustain bioenergy syst"] = "Journal of sustainable bioenergy systems",
  ["j sustain cem based mater"] = "Journal of sustainable cement-based materials",
  ["j sustain dev"] = "Journal of sustainable development",
  ["j sustain sci manag"] = "Journal of sustainability science and management",
  ["j sustain water built environ"] = "Journal of sustainable water in the built environment",
  ["j sustainability res"] = "Journal of Sustainability Research",
  ["j sustainability sci manage"] = "Journal of Sustainability Science and Management",
  ["j sustainable agric environ"] = "Journal of Sustainable Agriculture and Environment",
  ["j sustainable archit civ eng"] = "Journal of Sustainable Architecture and Civil Engineering",
  ["j sustainable cem-based mater"] = "Journal of Sustainable Cement-Based Materials",
  ["j sustainable dev energy water environ syst"] = "Journal of Sustainable Development of Energy, Water and Environment Systems",
  ["j sustainable energy eng"] = "Journal of Sustainable Energy Engineering",
  ["j sustainable for"] = "Journal of Sustainable Forestry",
  ["j sustainable mater sci eng"] = "Journal of Sustainable Materials Science and Engineering",
  ["j sustainable metall"] = "Journal of Sustainable Metallurgy",
  ["j sustainable min"] = "Journal of Sustainable Mining",
  ["j sustainable prod des"] = "Journal of Sustainable Product Design",
  ["j sustainable sci manage"] = "Journal of Sustainable Science and Management",
  ["j sustainable tourism"] = "Journal of Sustainable Tourism",
  ["j sustainable water built environ"] = "Journal of Sustainable Water in the Built Environment",
  ["j swim res"] = "The journal of swimming research",
  ["j sykepleien"] = "Journalen sykepleien",
  ["j symb comput"] = "Journal of symbolic computation",
  ["j symb log"] = "Journal of Symbolic Logic",
  ["j symbolic comput"] = "Journal of Symbolic Computation",
  ["j symbolic logic"] = "The Journal of Symbolic Logic",
  ["j symplectic geom"] = "Journal of Symplectic Geometry",
  ["j symptoms signs"] = "Journal of symptoms and signs",
  ["j synchrotron radiat"] = "Journal of synchrotron radiation",
  ["j syndr"] = "Journal of syndromes",
  ["j synth lubr"] = "Journal of Synthetic Lubrication",
  ["j synth org chem jpn"] = "Journal of Synthetic Organic Chemistry, Japan",
  ["j syst archit"] = "Journal of Systems Architecture",
  ["j syst chem"] = "Journal of Systems Chemistry",
  ["j syst cybern inf"] = "Journal of systemics, cybernetics and informatics",
  ["j syst eng"] = "Journal of Systems Engineering",
  ["j syst eng electron"] = "Journal of Systems Engineering and Electronics",
  ["j syst evol"] = "Journal of systematics and evolution",
  ["j syst integr"] = "Journal of Systems Integration",
  ["j syst integr neurosci"] = "Journal of systems and integrative neuroscience",
  ["j syst palaeontol"] = "Journal of systematic palaeontology",
  ["j syst paleontol"] = "Journal of Systematic Palaeontology",
  ["j syst sci complex"] = "Journal of systems science and complexity",
  ["j syst sci complexity"] = "Journal of Systems Science and Complexity",
  ["j syst sci syst eng"] = "Journal of Systems Science and Systems Engineering",
  ["j syst softw"] = "The Journal of systems and software",
  ["j syst software"] = "Journal of Systems and Software",
  ["j systems sci math sci"] = "Journal of Systems Science and Mathematical Sciences. Xitong Kexue yu Shuxue",
  ["j taibah univ med sci"] = "Journal of Taibah University Medical Sciences",
  ["j taibah univ sci"] = "Journal of Taibah University for Science : JTUSCI",
  ["j taiwan inst chem eng"] = "Journal of the Taiwan Institute of Chemical Engineers",
  ["j target ther cancer"] = "The journal of targeted therapies in cancer",
  ["j taxation"] = "Journal of Taxation",
  ["j teach action res"] = "Journal of teacher action research",
  ["j teach educ"] = "Journal of teacher education",
  ["j teach phys educ"] = "Journal of teaching in physical education : JTPE",
  ["j teach res chem"] = "Journal of Teaching and Research in Chemistry",
  ["j teach soc work"] = "Journal of teaching in social work",
  ["j tech phys"] = "Journal of Technical Physics",
  ["j tech transfer"] = "Journal of Technology Transfer",
  ["j tech univ plovdiv fundam sci appl ser a pure appl math"] = "Journal of the Technical University at Plovdiv",
  ["j technol behav sci"] = "Journal of technology in behavioral science",
  ["j technol comput aided des tcad"] = "Journal of Technology Computer Aided Design TCAD",
  ["j technol hum serv"] = "Journal of technology in human services",
  ["j technol sci educ"] = "Journal of technology and science education",
  ["j technol transf"] = "The Journal of technology transfer",
  ["j technol transfer"] = "Journal of Technology Transfer",
  ["j tehran heart cent"] = "The journal of Tehran Heart Center",
  ["j teknol"] = "Jurnal teknologi",
  ["j telemed telecare"] = "Journal of telemedicine and telecare",
  ["j tenn acad sci"] = "Journal. Tennessee Academy of Science",
  ["j tenn dent assoc"] = "The Journal of the Tennessee Dental Association",
  ["j tenn med assoc"] = "Journal of the Tennessee Medical Association",
  ["j tenn state dent assoc"] = "Journal - Tennessee State Dental Association",
  ["j tensor soc"] = "Journal of the Tensor Society",
  ["j terramech"] = "Journal of Terramechanics",
  ["j terror res"] = "Journal of terrorism research",
  ["j test eval"] = "Journal of Testing and Evaluation",
  ["j texas dent hyg assoc"] = "The Journal of the Texas Dental Hygienists' Association. Texas Dental Hygienists' Association",
  ["j text des res pract"] = "Journal of Textile Design Research and Practice",
  ["j text fibrous mater"] = "Journal of Textiles and Fibrous Materials",
  ["j text inst"] = "Journal of the Textile Institute",
  ["j text sci eng"] = "Journal of textile science & engineering",
  ["j texture stud"] = "Journal of texture studies",
  ["j thai assoc volunt steriliz"] = "Journal of Thai Association for Voluntary Sterilization",
  ["j thanatol"] = "Journal of thanatology",
  ["j theol south afr"] = "Journal of theology for Southern Africa",
  ["j theor appl comput sci"] = "Journal of theoretical and applied computer science",
  ["j theor appl inf technol"] = "Journal of Theoretical and Applied Information Technology",
  ["j theor appl mech"] = "Journal Of Theoretical And Applied Mechanics",
  ["j theor appl phys"] = "Journal Of Theoretical And Applied Physics",
  ["j theor biol"] = "Journal of Theoretical Biology",
  ["j theor comput acoust"] = "Journal of Theoretical and Computational Acoustics",
  ["j theor comput chem"] = "Journal of theoretical & computational chemistry",
  ["j theor nombres bordeaux"] = "Journal de Theorie des Nombres de Bordeaux",
  ["j theor polit"] = "Journal of theoretical politics",
  ["j theor probab"] = "Journal of theoretical probability",
  ["j theoret appl mech"] = "Journal of Theoretical and Applied Mechanics",
  ["j theoret biol"] = "Journal of Theoretical Biology",
  ["j theoret probab"] = "Journal of Theoretical Probability",
  ["j theory constr test"] = "Journal of theory construction & testing",
  ["j theory pract dent public health"] = "Journal of theory and practice of dental public health",
  ["j theory soc behav"] = "Journal for the theory of social behaviour",
  ["j ther ultrasound"] = "Journal of therapeutic ultrasound",
  ["j therm anal"] = "Journal of Thermal Analysis",
  ["j therm anal calorim"] = "Journal of Thermal Analysis and Calorimetry",
  ["j therm biol"] = "Journal of thermal biology",
  ["j therm envelope build sci"] = "Journal of Thermal Envelope and Building Science",
  ["j therm insul"] = "Journal of Thermal Insulation",
  ["j therm insul build envelopes"] = "Journal of Thermal Insulation and Building Envelopes",
  ["j therm sci"] = "Journal of Thermal Science",
  ["j therm sci eng appl"] = "Journal of thermal science and engineering applications",
  ["j therm sci technol"] = "Journal of Thermal Science and Technology",
  ["j therm spray technol"] = "Journal of Thermal Spray Technology",
  ["j therm stresses"] = "Journal of Thermal Stresses",
  ["j thermal stresses"] = "Journal of Thermal Stresses",
  ["j thermodyn catal"] = "Journal of Thermodynamics & Catalysis",
  ["j thermophys heat trans"] = "Journal of thermophysics and heat transfer",
  ["j thermophys heat transfer"] = "Journal of Thermophysics and Heat Transfer",
  ["j thermoplast compos mater"] = "Journal of Thermoplastic Composite Materials",
  ["j third world stud"] = "Journal of third world studies",
  ["j thorac cardiovasc surg"] = "The Journal of thoracic and cardiovascular surgery",
  ["j thorac dis"] = "Journal of thoracic disease",
  ["j thorac imaging"] = "Journal of thoracic imaging",
  ["j thorac oncol"] = "Journal of thoracic oncology : official publication of the International Association for the Study of Lung Cancer",
  ["j thorac surg"] = "The Journal of thoracic surgery",
  ["j thought"] = "Journal of thought",
  ["j threat taxa"] = "Journal of threatened taxa",
  ["j thromb haemost"] = "Journal of thrombosis and haemostasis : JTH",
  ["j thromb thrombolysis"] = "Journal of thrombosis and thrombolysis",
  ["j thyroid disord ther"] = "Journal of thyroid disorders & therapy",
  ["j thyroid res"] = "Journal of thyroid research",
  ["j théor nombres bordeaux"] = "Journal de Théorie des Nombres de Bordeaux",
  ["j tianjin norm univ nat sci ed"] = "Journal of Tianjin Normal University",
  ["j tianjin univ"] = "Journal of Tianjin University",
  ["j time ser anal"] = "Journal of Time Series Analysis",
  ["j time ser econom"] = "Journal of Time Series Econometrics",
  ["j time series anal"] = "Journal of Time Series Analysis",
  ["j time use res"] = "Journal of time use research",
  ["j tissue cult methods"] = "Journal of tissue culture methods : Tissue Culture Association manual of cell, tissue, and organ culture procedures",
  ["j tissue eng"] = "Journal of tissue engineering",
  ["j tissue eng regen med"] = "Journal of tissue engineering and regenerative medicine",
  ["j tissue eng regener med"] = "Journal of Tissue Engineering and Regenerative Medicine",
  ["j tissue sci eng"] = "Journal of tissue science & engineering",
  ["j tissue viability"] = "Journal of tissue viability",
  ["j tn state med assoc"] = "The Journal of the Tennessee State Medical Association. Tennessee State Medical Association",
  ["j tongji med univ"] = "Journal of Tongji Medical University",
  ["j topol"] = "Journal of Topology",
  ["j topol anal"] = "Journal of Topology\tand Analysis",
  ["j torrey bot soc"] = "The journal of the Torrey Botanical Society",
  ["j toxicol"] = "Journal of toxicology",
  ["j toxicol clin exp"] = "Journal de Toxicologie Clinique et Experimentale",
  ["j toxicol clin toxicol"] = "Journal of toxicology. Clinical toxicology",
  ["j toxicol cutaneous ocul toxicol"] = "Journal of toxicology. Cutaneous and ocular toxicology",
  ["j toxicol educ"] = "Journal of toxicological education",
  ["j toxicol environ health"] = "Journal of toxicology and environmental health",
  ["j toxicol environ health a"] = "Journal of toxicology and environmental health. Part A",
  ["j toxicol environ health b crit rev"] = "Journal of toxicology and environmental health. Part B, Critical reviews",
  ["j toxicol environ health part a"] = "Journal of Toxicology & Environmental Health Part A: Current Issues",
  ["j toxicol environ health part b"] = "Journal of Toxicology & Environmental Health Part B: Critical Reviews",
  ["j toxicol environ health sci"] = "Journal of toxicology and environmental health sciences",
  ["j toxicol environ health suppl"] = "Journal of toxicology and environmental health. Supplement",
  ["j toxicol environ health, part a"] = "Journal of Toxicology and Environmental Health, Part A",
  ["j toxicol environ health, part b"] = "Journal of Toxicology and Environmental Health, Part B: Critical Reviews",
  ["j toxicol med"] = "Journal de Toxicologie Medicale",
  ["j toxicol pathol"] = "Journal of toxicologic pathology",
  ["j toxicol pharmacol"] = "Journal of toxicology and pharmacology",
  ["j toxicol sci"] = "The Journal of toxicological sciences",
  ["j toxicol toxin rev"] = "Journal of toxicology. Toxin reviews",
  ["j trace elem electrolytes health dis"] = "Journal of trace elements and electrolytes in health and disease",
  ["j trace elem exp med"] = "Journal of Trace Elements in Experimental Medicine",
  ["j trace elem med biol"] = "Journal of trace elements in medicine and biology : organ of the Society for Minerals and Trace Elements (GMS)",
  ["j trace microprobe tech"] = "Journal of Trace and Microprobe Techniques",
  ["j tradit chin med"] = "Journal of traditional Chinese medicine = Chung i tsa chih ying wen pan",
  ["j tradit complement med"] = "Journal of traditional and complementary medicine",
  ["j tradit med clin naturop"] = "Journal of traditional medicine & clinical naturopathy",
  ["j traffic transp eng"] = "Journal of Traffic and Transportation Engineering",
  ["j transcult nurs"] = "Journal of transcultural nursing : official journal of the Transcultural Nursing Society",
  ["j transfus"] = "Journal of transfusion",
  ["j transl autoimmun"] = "Journal of translational autoimmunity",
  ["j transl genet genom"] = "Journal of translational genetics and genomics",
  ["j transl int med"] = "Journal of translational internal medicine",
  ["j transl med"] = "Journal of translational medicine",
  ["j transl med epidemiol"] = "Journal of translational medicine & epidemiology",
  ["j transl neurosci"] = "Journal of translational neurosciences",
  ["j transl neurosci (beijing)"] = "Journal of translational neuroscience",
  ["j transl sci"] = "Journal of translational science",
  ["j transnat manage devel"] = "Journal of Transnational Management Development",
  ["j transnatl law policy"] = "Journal of transnational law & policy",
  ["j transp econ pol"] = "Journal of Transport Economics and Policy",
  ["j transp eng"] = "Journal of Transportation Engineering",
  ["j transp eng a syst"] = "Journal of transportation engineering. Part A, Systems",
  ["j transp eng part a syst"] = "Journal of Transportation Engineering Part A. Systems",
  ["j transp eng part b pavements"] = "Journal of Transportation Engineering Part B. Pavements",
  ["j transp geogr"] = "Journal of transport geography",
  ["j transp health"] = "Journal of transport & health",
  ["j transp land use"] = "Journal of transport and land use",
  ["j transp saf secur"] = "Journal of Transportation Safety & Security",
  ["j transp secur"] = "Journal of Transportation Security",
  ["j transp technol"] = "Journal of transportation technologies",
  ["j transpl coord"] = "Journal of transplant coordination : official publication of the North American Transplant Coordinators Organization (NATCO)",
  ["j transplant"] = "Journal of transplantation",
  ["j trauma"] = "The Journal of trauma",
  ["j trauma acute care surg"] = "The journal of trauma and acute care surgery",
  ["j trauma dissociation"] = "Journal of trauma & dissociation : the official journal of the International Society for the Study of Dissociation (ISSD)",
  ["j trauma manag outcomes"] = "Journal of trauma management & outcomes",
  ["j trauma nurs"] = "Journal of trauma nursing : the official journal of the Society of Trauma Nurses",
  ["j trauma stress"] = "Journal of traumatic stress",
  ["j trauma stress disord treat"] = "Journal of Traumatic Stress Disorders & Treatment",
  ["j trauma treat"] = "Journal of trauma & treatment",
  ["j travel med"] = "Journal of travel medicine",
  ["j travel res"] = "Journal of travel research",
  ["j tribol"] = "Journal of Tribology",
  ["j trop agric"] = "Journal of tropical agriculture",
  ["j trop agric food sci"] = "Journal of tropical agriculture and food science",
  ["j trop biodivers biotechnol"] = "Journal of tropical biodiversity and biotechnology",
  ["j trop biol conserv"] = "Journal of tropical biology & conservation",
  ["j trop crop sci"] = "Journal of tropical crop science",
  ["j trop dis"] = "Journal of tropical diseases",
  ["j trop dis public health"] = "Journal of tropical diseases & public health",
  ["j trop ecol"] = "Journal of tropical ecology",
  ["j trop ecology"] = "Journal of Tropical Ecology",
  ["j trop for sci"] = "Journal of Tropical Forest Science",
  ["j trop geogr"] = "The Journal of tropical geography",
  ["j trop life sci"] = "Journal of tropical life science",
  ["j trop med"] = "Journal of tropical medicine",
  ["j trop med hyg"] = "The Journal of tropical medicine and hygiene",
  ["j trop meteorol"] = "Journal of Tropical Meteorology",
  ["j trop microbiol biotechnol"] = "Journal of Tropical Microbiology and Biotechnology",
  ["j trop pediatr"] = "Journal of tropical pediatrics",
  ["j trop pediatr (1967)"] = "Journal of tropical pediatrics (1967)",
  ["j trop pediatr (lond)"] = "Journal of tropical pediatrics (London, England : 1955)",
  ["j trop pediatr afr child health"] = "The Journal of tropical pediatrics and African child health",
  ["j trop pediatr environ child health"] = "The Journal of tropical pediatrics and environmental child health",
  ["j trop subtrop botany"] = "Re dai ya re dai zhi wu xue bao = Journal of tropical and subtropical botany",
  ["j tsinghua univ"] = "Journal of Tsinghua University. Science and Technology. Qinghua Daxue Xuebao. Ziran Kexue Ban",
  ["j tsuda college"] = "Journal of Tsuda College",
  ["j tuberc res"] = "Journal of tuberculosis research",
  ["j turbomach"] = "Journal of Turbomachinery",
  ["j turbul"] = "Journal of Turbulence",
  ["j turfgrass manage"] = "Journal of Turfgrass Management",
  ["j turk chem soc sect a chem"] = "Journal Of The Turkish Chemical Society Section A Chemistry",
  ["j turk ger gynecol assoc"] = "Journal of the Turkish German Gynecological Association",
  ["j ultrafine grained nanostruct mater"] = "Journal Of Ultrafine Grained And Nanostructured Materials",
  ["j ultrason"] = "Journal of ultrasonography",
  ["j ultrasound"] = "Journal of ultrasound",
  ["j ultrasound med"] = "Journal of ultrasound in medicine : official journal of the American Institute of Ultrasound in Medicine",
  ["j ultrastruct mol struct res"] = "Journal of Ultrastructure and Molecular Structure Research",
  ["j ultrastruct res"] = "Journal of Ultrastructure Research",
  ["j ultrastruct res suppl"] = "Journal of ultrastructure research. Supplement",
  ["j umm al-qura univ appl sci"] = "Journal of Umm Al-Qura University for Applied Sciences",
  ["j umm al-qura univ eng archit"] = "Journal of Umm Al-Qura University for Engineering and Architecture",
  ["j uncertain syst"] = "Journal of Uncertain Systems",
  ["j uncertainty anal appl"] = "Journal of Uncertainty Analysis and Applications",
  ["j unconv oil gas resour"] = "Journal of Unconventional Oil and Gas Resources",
  ["j undergrad chem res"] = "Journal of undergraduate chemistry research",
  ["j undergrad neurosci educ"] = "Journal of undergraduate neuroscience education : JUNE : a publication of FUN, Faculty for Undergraduate Neuroscience",
  ["j undergrad rep phys"] = "Journal of Undergraduate Reports in Physics",
  ["j undergrad res (gainesv)"] = "Journal of undergraduate research (Gainesville, Fla.)",
  ["j united reform church hist soc"] = "Journal - United Reformed Church History Society. United Reformed Church History Society",
  ["j univ babylon eng sci"] = "Journal of University of Babylon for Engineering Sciences",
  ["j univ babylon pure appl sci"] = "Journal of University of Babylon for Pure and Applied Sciences",
  ["j univ sci technol beijing, miner metall mater"] = "Journal of University of Science and Technology Beijing, Mineral, Metallurgy, Material",
  ["j univ sci technol china"] = "Journal of University of Science and Technology of China. Zhongguo Kexue Jishu Daxue Xuebao",
  ["j univers comput sci"] = "Journal of universal computer science : J. UCS",
  ["j unmanned veh syst"] = "Journal of unmanned vehicle systems",
  ["j uoeh"] = "Journal of UOEH",
  ["j urban"] = "Journal of urbanism",
  ["j urban aff"] = "Journal of urban affairs",
  ["j urban des (abingdon)"] = "Journal of urban design",
  ["j urban ecol"] = "Journal of Urban Ecology",
  ["j urban econ"] = "Journal of urban economics",
  ["j urban futures"] = "Journal of Urban Futures",
  ["j urban health"] = "Journal of urban health : bulletin of the New York Academy of Medicine",
  ["j urban hist"] = "Journal of urban history",
  ["j urban law"] = "Journal of urban law",
  ["j urban mobility"] = "Journal of Urban Mobility",
  ["j urban plan dev"] = "Journal of urban planning and development",
  ["j urban plann dev"] = "Journal of Urban Planning and Development",
  ["j urban psychiatry"] = "The Journal of urban psychiatry",
  ["j urban reg inf syst assoc"] = "Journal of the Urban and Regional Information Systems Association",
  ["j urban technol"] = "Journal of Urban Technology",
  ["j urol"] = "The Journal of urology",
  ["j urol (paris)"] = "Journal d'urologie",
  ["j urol medicale chir"] = "Journal d'urologie médicale et chirurgicale",
  ["j urol nephrol (paris)"] = "Journal d'urologie et de néphrologie",
  ["j urol nurs"] = "Journal of urological nursing",
  ["j urol res"] = "Journal of urology and research",
  ["j urol urogynakologie"] = "Journal für Urologie und Urogynäkologie : Zeitschrift für Urologie und Urogynäkologie in Klinik und Praxis",
  ["j us china med sci"] = "Journal of US-China medical science",
  ["j usability stud"] = "Journal of usability studies",
  ["j vac sci technol"] = "Journal of Vacuum Science and Technology",
  ["j vac sci technol a"] = "Journal of vacuum science & technology. A, Vacuum, surfaces, and films : an official journal of the American Vacuum Society",
  ["j vac sci technol a: vac surf films"] = "Journal of Vacuum Science & Technology A: Vacuum, Surfaces, and Films",
  ["j vac sci technol b"] = "Journal of Vacuum Science & Technology B",
  ["j vac sci technol b microelectron nanometer struct process meas phenom"] = "Journal of vacuum science & technology. B, Microelectronics and nanometer structures : processing, measurement, and phenomena : an official journal of the American Vacuum Society",
  ["j vac sci technol b nanotechnol microelectron"] = "Journal of vacuum science and technology. B, Nanotechnology & microelectronics : materials, processing, measurement, & phenomena : JVST B",
  ["j vac sci technol b: microelectron nanometer struct"] = "Journal of Vacuum Science & Technology B: Microelectronics and Nanometer Structures",
  ["j vac sci technol b: microelectron process phenom"] = "Journal of Vacuum Science & Technology B: Microelectronics Processing and Phenomena",
  ["j vac sci technol, a"] = "Journal of Vacuum Science and Technology A",
  ["j vac sci technol, b"] = "Journal of Vacuum Science and Technology B",
  ["j vac sci technol, b: microelectron nanometer"] = "Struct.–Process., Meas., Phenom. Journal of Vacuum Science & Technology, B: Microelectronics and Nanometer Structures–Processing, Measurement, and Phenomena",
  ["j vac sci technol, b: microelectron nanometer struct–process, meas, phenom"] = "Journal of Vacuum Science & Technology, B: Microelectronics and Nanometer Structures–Processing, Measurement, and Phenomena",
  ["j vac sci technol, b: nanotechnol microelectron: mater, process, meas, phenom"] = "Journal of Vacuum Science & Technology, B: Nanotechnology & Microelectronics: Materials, Processing, Measurement, & Phenomena",
  ["j vaccines immun"] = "Journal of vaccines & immunization",
  ["j vaccines vaccin"] = "Journal of vaccines & vaccination",
  ["j value inq"] = "The Journal of value inquiry",
  ["j vasc access"] = "The journal of vascular access",
  ["j vasc bras"] = "Jornal vascular brasileiro",
  ["j vasc diagn"] = "Journal of vascular diagnostics",
  ["j vasc diagn interv"] = "Journal of vascular diagnostics and interventions",
  ["j vasc endovasc surg"] = "Journal of vascular and endovascular surgery",
  ["j vasc interv neurol"] = "Journal of vascular and interventional neurology",
  ["j vasc interv radiol"] = "Journal of vascular and interventional radiology : JVIR",
  ["j vasc med surg"] = "Journal of vascular medicine & surgery",
  ["j vasc nurs"] = "Journal of vascular nursing : official publication of the Society for Peripheral Vascular Nursing",
  ["j vasc res"] = "Journal of vascular research",
  ["j vasc surg"] = "Journal of vascular surgery",
  ["j vasc surg cases"] = "Journal of vascular surgery cases",
  ["j vasc surg cases innov tech"] = "Journal of vascular surgery cases and innovative techniques",
  ["j vasc surg venous lymphat disord"] = "Journal of vascular surgery. Venous and lymphatic disorders",
  ["j vasc ultrasound"] = "Journal for Vascular Ultrasound",
  ["j vector borne dis"] = "Journal of vector borne diseases",
  ["j vector ecol"] = "Journal of vector ecology : journal of the Society for Vector Ecology",
  ["j veg sci"] = "Journal of vegetation science : official organ of the International Association for Vegetation Science",
  ["j vener dis inf"] = "The Journal of venereal disease information",
  ["j venom anim toxins incl trop dis"] = "The journal of venomous animals and toxins including tropical diseases",
  ["j venom res"] = "Journal of venom research",
  ["j venomous anim toxins incl trop dis"] = "Journal of Venomous Animals Including Tropical Diseases",
  ["j verbal learning verbal behav"] = "Journal of verbal learning and verbal behavior",
  ["j verbrauch lebensm"] = "Journal für Verbraucherschutz und Lebensmittelsicherheit = Journal of consumer protection and food safety",
  ["j verbraucherschutz lebensmittelsicherh"] = "Journal fur Verbraucherschutz und Lebensmittelsicherheit",
  ["j verif valid uncertain quantif"] = "Journal of verification, validation, and uncertainty quantification",
  ["j verif validation uncertainty quantif"] = "Journal of Verification, Validation, and Uncertainty Quantification",
  ["j vertebr  paleontol"] = "Journal of Vertebrate Paleontology",
  ["j vertebr biol"] = "Journal of vertebrate biology",
  ["j vertebr paleontol"] = "Journal of vertebrate paleontology",
  ["j vestib res"] = "Journal of vestibular research : equilibrium & orientation",
  ["j vestibular res"] = "Journal of Vestibular Research",
  ["j vet behav"] = "Journal of veterinary behavior : clinical applications and research : official journal of : Australian Veterinary Behaviour Interest Group, International Working Dog Breeding Association",
  ["j vet cardiol"] = "Journal of veterinary cardiology : the official journal of the European Society of Veterinary Cardiology",
  ["j vet dent"] = "Journal of veterinary dentistry",
  ["j vet diagn invest"] = "Journal of veterinary diagnostic investigation : official publication of the American Association of Veterinary Laboratory Diagnosticians, Inc",
  ["j vet educ res"] = "Journal of Veterinary Education Research",
  ["j vet emerg crit care"] = "Journal of Veterinary Emergency and Critical Care",
  ["j vet emerg crit care (san antonio)"] = "Journal of veterinary emergency and critical care (San Antonio, Tex. : 2001)",
  ["j vet intern med"] = "Journal of veterinary internal medicine",
  ["j vet med"] = "Journal of veterinary medicine",
  ["j vet med a physiol pathol clin med"] = "Journal of veterinary medicine. A, Physiology, pathology, clinical medicine",
  ["j vet med b infect dis vet public health"] = "Journal of veterinary medicine. B, Infectious diseases and veterinary public health",
  ["j vet med educ"] = "Journal of veterinary medical education",
  ["j vet med res"] = "Journal of veterinary medicine and research",
  ["j vet med sci"] = "The Journal of veterinary medical science",
  ["j vet med ser a"] = "Journal of Veterinary Medicine Series A",
  ["j vet med ser b"] = "Journal of Veterinary Medicine Series B",
  ["j vet pharmacol ther"] = "Journal of veterinary pharmacology and therapeutics",
  ["j vet res"] = "Journal of veterinary research",
  ["j vet sci"] = "Journal of veterinary science",
  ["j vet sci med diagn"] = "Journal of veterinary science & medical diagnosis",
  ["j vet sci technol"] = "Journal of veterinary science & technology",
  ["j vib acoust"] = "Journal of Vibration and Acoustics",
  ["j vib control"] = "Journal of Vibration and Control",
  ["j vib eng technol"] = "Journal of Vibration Engineering & Technologies",
  ["j vibroeng"] = "Journal Of Vibroengineering",
  ["j vic cult"] = "Journal of Victorian culture : JVC",
  ["j vinyl add tech"] = "Journal of Vinyl and Additive Technology",
  ["j vinyl addit technol"] = "Journal of Vinyl and Additive Technology",
  ["j vinyl tech"] = "Journal of Vinyl Technology",
  ["j viral dis"] = "The Journal of viral diseases",
  ["j viral hepat"] = "Journal of viral hepatitis",
  ["j virol"] = "Journal of virology",
  ["j virol antivir res"] = "Journal of virology & antiviral research",
  ["j virol methods"] = "Journal of virological methods",
  ["j virtual worlds res"] = "Journal of virtual worlds research",
  ["j virus erad"] = "Journal of virus eradication",
  ["j viruses"] = "Journal of viruses",
  ["j vis"] = "Journal of vision",
  ["j vis (tokyo)"] = "Journal of visualization",
  ["j vis art pract"] = "Journal of visual art practice",
  ["j vis commun image represent"] = "Journal of visual communication and image representation",
  ["j vis commun med"] = "Journal of visual communication in medicine",
  ["j vis exp"] = "Journal of visualized experiments : JoVE",
  ["j vis impair blind"] = "Journal of visual impairment & blindness",
  ["j vis lang comput"] = "Journal of visual languages and computing",
  ["j vis surg"] = "Journal of visualized surgery",
  ["j visc surg"] = "Journal of visceral surgery",
  ["j visual commun image represent"] = "Journal of Visual Communication and Image Representation",
  ["j visualization"] = "Journal of Visualization",
  ["j visualized exp"] = "Journal of Visualized Experiments",
  ["j vitaminol (kyoto)"] = "The Journal of vitaminology",
  ["j vitic enol"] = "Journal of Viticulture and Enology",
  ["j vitreoretin dis"] = "Journal of vitreoretinal diseases",
  ["j vlsi sig proc"] = "Journal of VLSI Signal Processing",
  ["j vlsi signal process"] = "Journal of VLSI Signal Processing",
  ["j vlsi signal process syst signal image video technol"] = "Journal of VLSI signal processing systems for signal, image, and video technology",
  ["j vocat behav"] = "Journal of vocational behavior",
  ["j vocat rehabil"] = "Journal of vocational rehabilitation",
  ["j voice"] = "Journal of voice : official journal of the Voice Foundation",
  ["j volcanol geotherm res"] = "Journal of Volcanology and Geothermal Research",
  ["j volcanol seismolog"] = "Journal of Volcanology and Seismology",
  ["j volunt action res"] = "Journal of voluntary action research",
  ["j volunt adm"] = "The Journal of volunteer administration",
  ["j w va hist assoc"] = "The journal of the West Virginia Historical Association. West Virginia Historical Association",
  ["j war cult stud"] = "Journal of war & culture studies",
  ["j warburg courtauld inst"] = "Journal of the Warburg and Courtauld Institutes",
  ["j wash acad sci"] = "Journal. Washington Academy of Sciences, Washington, D. C",
  ["j watch aids clin care"] = "Journal watch. AIDS clinical care",
  ["j water chem technol"] = "Journal of Water Chemistry and Technology",
  ["j water clim"] = "Journal of Water and Climate",
  ["j water clim change"] = "Journal of Water and Climate Change",
  ["j water environ nanotechnol"] = "Journal Of Water And Environmental Nanotechnology",
  ["j water environ technol"] = "Journal of water and environment technology",
  ["j water health"] = "Journal of water and health",
  ["j water land dev"] = "Journal Of Water And Land Development",
  ["j water manage model"] = "Journal Of Water Management Modeling",
  ["j water pollut control fed"] = "Journal - Water Pollution Control Federation",
  ["j water process eng"] = "Journal of Water Process Engineering",
  ["j water resour plan manag"] = "Journal of water resources planning and management",
  ["j water resour plann manage"] = "Journal of Water Resources Planning and Management",
  ["j water resour prot"] = "Journal of Water Resource and Protection",
  ["j water reuse desalin"] = "Journal of Water Reuse and Desalination",
  ["j water sanit hyg dev"] = "Journal of water, sanitation, and hygiene for development : a journal of the International Water Association",
  ["j water supply res technol aqua"] = "Journal of Water Supply: Research and Technology - AQUA",
  ["j water technol"] = "Journal of Water Technology",
  ["j waterw port coast ocean eng"] = "Journal of waterway, port, coastal, and ocean engineering",
  ["j waterw port coastal ocean eng"] = "Journal of Waterway Port, Coastal, and Ocean Engineering",
  ["j web semant"] = "Journal of Web Semantics",
  ["j west"] = "Journal of the West",
  ["j west afr coll surg"] = "Journal of the West African College of Surgeons",
  ["j west afr lang"] = "The Journal of West African languages",
  ["j west aust nurses"] = "Journal of the West Australian Nurses",
  ["j west soc periodontal periodontal abstr"] = "Journal of the Western Society of Periodontology / Periodontal Abstracts",
  ["j west soc periodontol periodontal abstr"] = "The Journal of the Western Society of Periodontology/Periodontal abstracts",
  ["j west va philo soc"] = "Journal of the West Virginia Philosophical Society",
  ["j wetland archaeol"] = "Journal of Wetland Archaeology",
  ["j wetlands ecol"] = "Journal of Wetlands Ecology",
  ["j wide bandgap mater"] = "Journal of Wide Bandgap Materials",
  ["j wilderness med"] = "Journal of wilderness medicine",
  ["j wildl biodivers"] = "Journal of wildlife and biodiversity",
  ["j wildl dis"] = "Journal of wildlife diseases",
  ["j wildl manage"] = "The Journal of wildlife management",
  ["j wildl res"] = "Jounal of Wildlife Research",
  ["j wind eng ind aerodyn"] = "Journal of Wind Engineering and Industrial Aerodynamics",
  ["j wine econ"] = "Journal of Wine Economics",
  ["j wine res"] = "Journal of wine research",
  ["j wirel mob netw ubiquitous comput dependable appl"] = "Journal of wireless mobile networks, ubiquitous computing, and dependable applications",
  ["j wis dent assoc"] = "The Journal of the Wisconsin Dental Association",
  ["j wis state dent soc"] = "The Journal of the Wisconsin State Dental Society",
  ["j women aging"] = "Journal of women & aging",
  ["j women high educ"] = "Journal about women in higher education",
  ["j women minor sci eng"] = "Journal of women and minorities in science and engineering",
  ["j women minorities sci eng"] = "Journal of Women and Minorities in Science and Engineering",
  ["j women polit policy"] = "Journal of women, politics & policy",
  ["j womens heal"] = "Journal of women's health and law",
  ["j womens health"] = "Journal of women's health",
  ["j womens health (larchmt)"] = "Journal of women's health (2002)",
  ["j womens health care"] = "Journal of women's health care",
  ["j womens health dev"] = "Journal of women's health and development",
  ["j womens health gend based med"] = "Journal of women's health & gender-based medicine",
  ["j womens health issues care"] = "Journal of women's health, issues & care",
  ["j womens health phys therap"] = "Journal of women's health physical therapy",
  ["j womens hist"] = "Journal of women's history",
  ["j wood chem technol"] = "Journal of Wood Chemistry and Technology",
  ["j wood sci"] = "Journal of Wood Science",
  ["j workplace behav health"] = "Journal of workplace behavioral health",
  ["j workplace learn"] = "The journal of workplace learning",
  ["j world aquac soc"] = "Journal of the World Aquaculture Society",
  ["j world aquacult soc"] = "Journal of the World Aquaculture Society",
  ["j world bus"] = "Journal of World Business",
  ["j world fed orthod"] = "Journal of the World federation of orthodontists",
  ["j world hist"] = "Journal of world history : official journal of the World History Association",
  ["j world intellect prop"] = "The Journal of world intellectual property",
  ["j world prehist"] = "Journal of world prehistory",
  ["j world trade"] = "Journal of World Trade",
  ["j world trade law"] = "Journal of World Trade Law",
  ["j worlds poult res"] = "Journal of world's poultry research",
  ["j wound care"] = "Journal of wound care",
  ["j wound ostomy continence nurs"] = "Journal of wound, ostomy, and continence nursing : official publication of The Wound, Ostomy and Continence Nurses Society",
  ["j wrist surg"] = "Journal of wrist surgery",
  ["j writ anal"] = "The journal of writing analytics",
  ["j writ res"] = "Journal of writing research",
  ["j wscg (plzen)"] = "Journal of WSCG",
  ["j wuhan univ natur sci ed"] = "Journal of Wuhan University. Natural Science Edition. Wuhan Daxue Xuebao. Lixue Ban",
  ["j wuhan univ technol\tmater sci ed"] = "Journal of Wuhan University of Technology Materials Science Edition",
  ["j wuhan univ technol mater sci ed"] = "Journal of Wuhan University of Technology. Materials science edition",
  ["j x-ray sci technol"] = "Journal of X-Ray Science and Technology",
  ["j xenobiot"] = "Journal of xenobiotics",
  ["j xiangya med"] = "Journal of Xiangya medicine",
  ["j xinjiang univ natur sci"] = "Xinjiang University",
  ["j xray sci technol"] = "Journal of X-ray science and technology",
  ["j yangzhou univ nat sci ed"] = "Journal of Yangzhou University",
  ["j yeast fungal res"] = "Journal of yeast and fungal research",
  ["j yiyang teachers college"] = "Journal of Yiyang Teachers’ College",
  ["j yoga phys ther"] = "Journal of yoga & physical therapy",
  ["j young investig"] = "Journal of young investigators",
  ["j young pharm"] = "Journal of young pharmacists : JYP",
  ["j youth adolesc"] = "Journal of youth and adolescence",
  ["j youth dev"] = "Journal of youth development : bridging research and practice",
  ["j youth stud"] = "Journal of youth studies",
  ["j yunnan univ nat sci"] = "Journal of Yunnan University. Natural Sciences. Yunnan Daxue Xuebao. Ziran Kexue Ban",
  ["j zhejiang univ sci"] = "Journal of Zhejiang University. Science",
  ["j zhejiang univ sci b"] = "Journal of Zhejiang University. Science. B",
  ["j zhejiang univ sci ed"] = "Journal of Zhejiang University. Science Edition. Zhejiang Daxue Xuebao. Lixue Ban",
  ["j zhejiang univ-sci a"] = "Journal of Zhejiang University-SCIENCE A Applied Physics & Engineering",
  ["j zhejiang univ-sci b"] = "Journal of Zhejiang University-SCIENCE B Biomedicine & Biotechnology",
  ["j zhengzhou univ eng sci"] = "Journal of Zhengzhou University. Engineering Science. Zhengzhou Daxue Xuebao. Gongxue Ban",
  ["j zhengzhou univ nat sci ed"] = "Journal of Zhengzhou University. Natural Science Edition. Zhengzhou Daxue Xuebao. Lixue Ban",
  ["j zoo aquar res"] = "Journal of zoo and aquarium research",
  ["j zoo aquarium res"] = "Journal of Zoo and Aquarium Research",
  ["j zoo biol"] = "Journal of zoo biology",
  ["j zoo wildl med"] = "Journal of zoo and wildlife medicine : official publication of the American Association of Zoo Veterinarians",
  ["j zool"] = "Journal of zoology",
  ["j zool (1987)"] = "Journal of zoology (London, England : 1987)",
  ["j zool (london)"] = "Journal of Zoology (London)",
  ["j zool syst evol res"] = "Journal of zoological systematics and evolutionary research = Zeitschrift für zoologische Systematik und Evolutionsforschung",
  ["ja clin rep"] = "JA clinical reports",
  ["jaad case rep"] = "JAAD case reports",
  ["jaad int"] = "JAAD international",
  ["jaakko hintikka sel pap"] = "Jaakko Hintikka Selected Papers",
  ["jaarb kankeronderz kankerbestrijd ned"] = "Jaarboek van kankeronderzoek en kankerbestrijding in Nederland. Yearbook for cancer research and fight against cancer in the Netherlands",
  ["jaarb numaga"] = "Jaarboek Numaga : gewijd aan heden en verleden van nijmegen en omgeving",
  ["jac antimicrob resist"] = "JAC-antimicrobial resistance",
  ["jac-antimicrob resist"] = "JAC-Antimicrobial Resistance",
  ["jacc basic transl sci"] = "JACC. Basic to translational science",
  ["jacc cardiooncol"] = "JACC. CardioOncology",
  ["jacc cardiovasc imaging"] = "JACC. Cardiovascular imaging",
  ["jacc cardiovasc interv"] = "JACC. Cardiovascular interventions",
  ["jacc case rep"] = "JACC. Case reports",
  ["jacc clin electrophysiol"] = "JACC. Clinical electrophysiology",
  ["jacc heart fail"] = "JACC. Heart failure",
  ["jacc: adv"] = "JACC: Advances",
  ["jacc: asia"] = "JACC: Asia",
  ["jacc: basic transl sci"] = "JACC: Basic to Translational Science",
  ["jacc: cardiooncol"] = "JACC: CardioOncology",
  ["jacc: cardiovasc imaging"] = "JACC: Cardiovascular Imaging",
  ["jacc: cardiovasc interventions"] = "JACC: Cardiovascular Interventions",
  ["jacc: case rep"] = "JACC: Case Reports",
  ["jacc: clin electrophysiol"] = "JACC: Clinical Electrophysiology",
  ["jacc: heart failure"] = "JACC: Heart Failure",
  ["jackson clin bull"] = "The Jackson Clinic bulletin",
  ["jacobs j addict ther"] = "Jacobs journal of addiction and therapy",
  ["jacobs j aids hiv"] = "Jacobs journal of AIDS/HIV",
  ["jacobs j allergy immunol"] = "Jacobs journal of allergy and immunology",
  ["jacobs j biotechnol bioeng"] = "Jacobs journal of biotechnology and bioengineering",
  ["jacobs j community med"] = "Jacobs journal of community medicine",
  ["jacobs j diabetes endocrinol"] = "Jacobs journal of diabetes and endocrinology",
  ["jacobs j environ sci"] = "Jacobs journal of environmental sciences",
  ["jacobs j food nutr"] = "Jacobs journal of food and nutrition",
  ["jacobs j genet"] = "Jacobs journal of genetics",
  ["jacobs j gerontol"] = "Jacobs journal of gerontology",
  ["jacobs j med diagn med imaging"] = "Jacobs journal of medical diagnosis and medical imaging",
  ["jacobs j microbiol pathol"] = "Jacobs journal of microbiology and pathology",
  ["jacobs j mol transl med"] = "Jacobs journal of molecular and translational medicine",
  ["jacobs j ophthalmol"] = "Jacobs journal of ophthalmology",
  ["jacobs j pediatr"] = "Jacobs journal of pediatrics",
  ["jacobs j psychiatry behav sci"] = "Jacobs journal of psychiatry and behavioral science",
  ["jacobs j pulmonol"] = "Jacobs journal of pulmonology",
  ["jacobs j radiat oncol"] = "Jacobs journal of radiation oncology",
  ["jacs au"] = "JACS Au",
  ["jaen j approx"] = "Jaen Journal on Approximation",
  ["jagd nat"] = "Jagd und Natur",
  ["jahr nationalökon statist"] = "Jahrbücher für Nationalökonomie und Statistik",
  ["jahr regionalwissen"] = "Jahrbuch für Regionalwissenschaft",
  ["jahr wirtsch osteuropas"] = "Jahrbuch der Wirtschaft Osteuropas",
  ["jahr wiss ethik"] = "Jahrbuch für Wissenschaft und Ethik",
  ["jahrb albertus univ koenigsb"] = "Jahrbuch der Albertus-Universität zu Königsberg/Pr",
  ["jahrb antike christentum"] = "Jahrbuch für Antike und Christentum",
  ["jahrb auktionspreise bucher"] = "Jahrbuch der Auktionspreise für Bücher, Handschriften und Autographen; Ergebnisse der Auktionen in Deutschland, Holland, Oesterreich und der Schweiz",
  ["jahrb brandenbg landesgesch"] = "Jahrbuch für brandenburgische Landesgeschichte",
  ["jahrb coburg landesstift"] = "Jahrbuch der Coburger Landesstiftung. Coburger Landesstiftung",
  ["jahrb dtsch luftfahrtforsch"] = "Jahrbuch der Deutschen Luftfahrtforschung",
  ["jahrb geol bundesanst"] = "Jahrbuch der Geologischen Bundesanstalt",
  ["jahrb geol bundesanst (1945)"] = "Jahrbuch der Geologischen Bundesanstalt (Vienna, Austria : 1945)",
  ["jahrb gesch"] = "Jahrbuch für Geschichte",
  ["jahrb gesch lat am"] = "Jahrbuch für Geschichte Lateinamerikas",
  ["jahrb gesch osteur"] = "Jahrbücher für Geschichte Osteuropas",
  ["jahrb gesch staat wirtsch ges lateinam"] = "Jahrbuch für Geschichte von Staat, Wirtschaft und Gesellschaft Lateinamerikas",
  ["jahrb hist forsch bundesrepub dtschl"] = "Jahrbuch der historischen Forschung in der Bundesrepublik Deutschland",
  ["jahrb hist ver furst liechtenstein"] = "Jahrbuch des Historischen Vereins für das Fürstentum Liechtenstein. Historischer Verein für das Fürstentum Liechtenstein",
  ["jahrb inst dtsch gesch"] = "Jahrbuch des Instituts für Deutsche Geschichte. Universiṭat Tel-Aviv. Makhon le-hisṭoryah Germanit",
  ["jahrb inst gesch med robert bosch stift"] = "Jahrbuch des Instituts für Geschichte der Medizin der Robert Bosch Stiftung",
  ["jahrb landeskd niederosterr"] = "Jahrbuch für Landeskunde von Niederösterreich",
  ["jahrb max planck gesell"] = "Jahrbuch Max Planck Gesellschaft",
  ["jahrb natl okon stat"] = "Jahrbücher für Nationalökonomie und Statistik",
  ["jahrb natschutz landschpfl"] = "Jahrbuch für Naturschutz und Landschaftspflege",
  ["jahrb oberaargau"] = "Jahrbuch des Oberaargaus",
  ["jahrb oberflachentech"] = "Jahrbuch Oberflachentechnik",
  ["jahrb oberoesterr musealver"] = "Jahrbuch des Oberösterreichischen Musealvereines. Oberösterreichischer Musealverein",
  ["jahrb preuss geol landesanst"] = "Jahrbuch der Preussischen Geologischen Landesanstalt",
  ["jahrb schlesisch friedrich wilhelms univ breslau"] = "Jahrbuch der Schlesischen Friedrich-Wilhelms-Universität zu Breslau",
  ["jahrb schweiz akad natwiss"] = "Jahrbuch Schweizerische Akademie der Naturwissenschaften",
  ["jahrb schweiz natforsch ges"] = "Jahrbuch der Schweizerischen naturforschenden Gesellschaft",
  ["jahrb schweiz wald- holzwirtsch"] = "Jahrbuch der schweizerischen Wald- und Holzwirtschaft",
  ["jahrb schweiz wasserwirtschafts-verb"] = "Jahrbuch des Schweizerischen Wasserwirtschaftsverbandes",
  ["jahrb sozialwiss"] = "Jahrbuch für Sozialwissenschaft",
  ["jahrb st gallische natwiss ges"] = "Jahrbuch der St. Gallischen Naturwissenschaftlichen Gesellschaft",
  ["jahrb thuner- brienzersee"] = "Jahrbuch vom Thuner- und Brienzersee",
  ["jahrb univ duesseld"] = "Jahrbuch der Universität Düsseldorf",
  ["jahrb ver schutz alppflanzen -tiere"] = "Jahrbuch des Vereins zum Schutze der Alpenpflanzen und -Tiere",
  ["jahrb vereins gesch stadt wien"] = "Jahrbuch des Vereins für Geschichte der Stadt Wien. Verein für Geschichte der Stadt Wien",
  ["jahrb vers lehranst brau berlin"] = "Jahrbuch der Versuchs und Lehranstalt fur Brauerei in Berlin",
  ["jahrb volksgesundh"] = "Jahrbuch Volksgesundheit",
  ["jahrb wirtschaftsgesch"] = "Jahrbuch für Wirtschaftsgeschichte",
  ["jahrb wiss bot"] = "Jahrbucher fur Wissenschaftliche Botanik",
  ["jahrb zürsee"] = "Jahrbuch vom Zürichsee",
  ["jahresber deutsch math-verein"] = "Jahresbericht der Deutschen Mathematiker-Vereinigung",
  ["jahresber dtsch math-ver"] = "Jahresbericht der Deutschen Mathematiker-Vereinigung",
  ["jahresber lignum"] = "Jahresbericht Lignum",
  ["jahresber lit zentralbl dtschl math astron"] = "Jahresberichte des Literarischen Zentralblattes fur Deutschland Mathematik Astronomie",
  ["jahresber natforsch ges graubünden"] = "Jahresbericht der Naturforschenden Gesellschaft Graubünden",
  ["jahresber schweiz akad med wiss"] = "Jahresbericht (Schweizerische Akademie der Medizinischen Wissenschaften : 1988)",
  ["jahresh ges naturkd wurtt"] = "Jahreshefte der Gesellschaft für Naturkunde in Württemberg. Gesellschaft für Naturkunde in Württemberg",
  ["jahresh ver vaterl natkd württ"] = "Jahreshefte des Vereins für vaterländische Naturkunde in Württemberg",
  ["jala charlottesv va"] = "JALA (Charlottesville, Va.)",
  ["jam j"] = "Jamaica journal",
  ["jam public health"] = "Jamaica public health",
  ["jama cardiol"] = "JAMA cardiology",
  ["jama dermatol"] = "JAMA dermatology",
  ["jama facial plast surg"] = "JAMA facial plastic surgery",
  ["jama health forum"] = "JAMA health forum",
  ["jama intern med"] = "JAMA internal medicine",
  ["jama j am med assoc"] = "JAMA The Journal Of The American Medical Association",
  ["jama netw open"] = "JAMA network open",
  ["jama neurol"] = "JAMA neurology",
  ["jama oncol"] = "JAMA oncology",
  ["jama ophthalmol"] = "JAMA ophthalmology",
  ["jama otolaryngol head neck surg"] = "JAMA otolaryngology-- head & neck surgery",
  ["jama pediatr"] = "JAMA pediatrics",
  ["jama psychiatry"] = "JAMA psychiatry",
  ["jama surg"] = "JAMA surgery",
  ["jama, j am med assoc"] = "JAMA, the Journal of the American Medical Association",
  ["jamaican nurse"] = "The Jamaican nurse",
  ["james arthur lect"] = "James Arthur lecture on the evolution of the human brain",
  ["james joyce q"] = "James Joyce quarterly",
  ["jamia open"] = "JAMIA open",
  ["janasamkhya siksha mukhapatra"] = "Janasaṃkhyā śikshā mukhapatra",
  ["japan annual rev electron comput telecom 1982"] = "Japan Annual Reviews in Electronics, Computers & Telecommunications 1982",
  ["japan j indust appl math"] = "Japan Journal of Industrial and Applied Mathematics",
  ["japan med assoc j"] = "Japan Medical Association journal : JMAJ",
  ["japan res quart"] = "Japan Research Quarterly",
  ["japan world econ"] = "Japan and the world economy",
  ["japan world economy"] = "Japan and the World Economy",
  ["japanese econ rev"] = "Japanese Economic Review",
  ["japanese econ stud"] = "Japanese economic studies",
  ["japanese economy"] = "Japanese Economy",
  ["japanese j fuzzy theory systems"] = "Japanese Journal of Fuzzy Theory and Systems",
  ["japanese journal of nematology|nihon senchu gakkai shi"] = "Nihon Senchu Gakkai shi",
  ["japanese poultry science|nippon kakin gakkaishi"] = "Nihon Kakin Gakkaishi",
  ["jasa express lett"] = "JASA express letters",
  ["jawra j am water resour assoc"] = "JAWRA Journal of the American Water Resources Association",
  ["jb js open access"] = "JB & JS open access",
  ["jbi database system rev implement rep"] = "JBI database of systematic reviews and implementation reports",
  ["jbi evid implement"] = "JBI evidence implementation",
  ["jbi evid synth"] = "JBI evidence synthesis",
  ["jbi evidence implementation"] = "JBI Evidence Implementation",
  ["jbi evidence synth"] = "JBI Evidence Synthesis",
  ["jbi libr syst rev"] = "JBI library of systematic reviews",
  ["jbic, j biol inorg chem"] = "JBIC, Journal of Biological Inorganic Chemistry",
  ["jbjs case connect"] = "JBJS case connector",
  ["jbjs essent surg tech"] = "JBJS essential surgical techniques",
  ["jbjs rev"] = "JBJS reviews",
  ["jbmr plus"] = "JBMR plus",
  ["jbra assist reprod"] = "JBRA assisted reproduction",
  ["jcah perspect"] = "JCAH perspectives. Joint Commission on Accreditation of Hospitals",
  ["jcca j can chiropr assoc"] = "JCCA. Journal of the Canadian Chiropractic Association. Journal de l’Association chiropratique canadienne",
  ["jcem case rep"] = "JCEM Case Reports",
  ["jci insight"] = "JCI insight",
  ["jcis open"] = "JCIS Open",
  ["jco clin cancer inf"] = "JCO Clinical Cancer Informatics",
  ["jco clin cancer inform"] = "JCO clinical cancer informatics",
  ["jco glob oncol"] = "JCO global oncology",
  ["jco oncol pract"] = "JCO oncology practice",
  ["jco precis oncol"] = "JCO precision oncology",
  ["jcp: biochem phys"] = "JCP: Biochemical Physics",
  ["jcrs online case rep"] = "JCRS online case reports",
  ["jcsm clin rep"] = "JCSM clinical reports",
  ["jcsm rapid commun"] = "JCSM rapid communications",
  ["jdr clin trans res"] = "JDR clinical and translational research",
  ["jdr clin transl res"] = "JDR Clinical & Translational Research",
  ["jednota annu furdek"] = "Jednota annual Furdek",
  ["jefferson-hillman hosp bull"] = "The Jefferson-Hillman hospital bulletin",
  ["jena z med naturwiss"] = "Jenaische Zeitschrift für Medicin und Naturwissenschaft",
  ["jerus stud philos hist sci"] = "Jerusalem Studies in Philosophy and History of Science",
  ["jerusalem q"] = "The Jerusalem quarterly",
  ["jesuit stud"] = "Jesuit Studies. Modernity through the Prism of Jesuit History",
  ["jetp lett"] = "JETP Letters",
  ["jeune afr"] = "Jeune Afrique (Paris, France : 1980)",
  ["jew aff"] = "Jewish affairs",
  ["jew cult hist"] = "Jewish culture and history",
  ["jew dig"] = "The Jewish digest",
  ["jew j sociol"] = "Jewish Journal Of Sociology",
  ["jew life"] = "Jewish life (Union of Orthodox Jewish Congregations of America : 1975)",
  ["jew mem hosp bull"] = "The Jewish Memorial Hospital bulletin",
  ["jew obs middle east rev"] = "Jewish observer and Middle East review",
  ["jew q rev"] = "Jewish quarterly review (Philadelphia, Pa.)",
  ["jew soc stud"] = "Jewish social studies",
  ["jfe j four electr"] = "JFE Journal du Four Electrique",
  ["jfe j four electr ind electrochim"] = "JFE Journal du Four Electrique & des Industries Electrochimiques",
  ["jfms open rep"] = "JFMS open reports",
  ["jforl j fr otorhinolaryngol audiophonol chir maxillofac"] = "JFORL. Journal français d'oto-rhino-laryngologie; audiophonologie et chirurgie maxillo-faciale",
  ["jfourier anal appl"] = "Journal of Fourier Analysis and Applications",
  ["jgh open"] = "JGH open : an open access journal of gastroenterology and hepatology",
  ["jhep rep"] = "JHEP reports : innovation in hepatology",
  ["ji bing jian ce"] = "Ji bing jian ce",
  ["ji chu yi xue yu lin chuang"] = "Ji chu yi xue yu lin chuang = Jichu yixue yu linchuang = Basic medical sciences and clinics",
  ["ji di yan jiu"] = "Ji di yan jiu = Chinese journal of polar research",
  ["ji guang sheng wu xue bao"] = "Ji guang sheng wu xue bao = Acta laser biology sinica = Jiguang shengwu xuebao",
  ["ji sheng chong xue yu ji sheng chong bing za zhi"] = "Ji sheng chong xue yu ji sheng chong bing za zhi = Journal of parasitology & parasitic diseases",
  ["ji sheng chong yu yi xue kun chong xue bao"] = "Ji sheng chong yu yi xue kun chong xue bao",
  ["ji shou da xue xue bao zi ran ke xue ban"] = "Ji shou da xue xue bao. Zi ran ke xue ban",
  ["ji yin zu xue yu ying yong sheng wu xue"] = "Ji yin zu xue yu ying yong sheng wu xue",
  ["jian kang bao"] = "Jian kang bao",
  ["jiangsu gongye xueyuan xuebao"] = "Jiangsu gongye xueyuan xuebao = Journal of Jiangsu Polytechnic University",
  ["jiangsu nong ye ke xue"] = "Jiangsu nong ye ke xue = Jiangsu nongye kexue",
  ["jiangsu nong ye xue bao"] = "Jiangsu nong ye xue bao = Journal of agricultural sciences",
  ["jiangsu nongye kexue / jiangsu sheng nong ye ke xue yuan bian|jiangsu nong ye ke xue"] = "Jiangsu nong ye ke xue",
  ["jiangxi nong ye xue bao"] = "Jiangxi nong ye xue bao",
  ["jiangxi nongye daxue xuebao"] = "Jiangxi nongye daxue xuebao = Acta agriculturae universitatis jiangxiensis",
  ["jiangxi yixueyuan xuebao"] = "Jiangxi yixueyuan xuebao = Acta academiae medicinae jiangxi",
  ["jica newsl"] = "JICA newsletter",
  ["jid innov"] = "JID innovations",
  ["jie fang jun yi xue za zhi"] = "Jie fang jun yi xue za zhi",
  ["jiegou huaxue"] = "Jiegou Huaxue",
  ["jika shinryo"] = "[Jika shinryō] Journal for pediatric praxis",
  ["jikeikai med j"] = "Jikeikai medical journal",
  ["jikken dobutsu"] = "Jikken dobutsu. Experimental animals",
  ["jikoketsu yuketsu"] = "Jikoketsu yuketsu : Nihon Jikoketsu Yuketsu Gakkai kaishi = Journal of Japanese Society of Autologous Blood Transfusion",
  ["jilin da xue xue bao"] = "Jilin da xue xue bao. Li xue ban = Journal of Jilin University. Science edition",
  ["jilin nong ye da xue xue bao"] = "Jilin nong ye da xue xue bao = Acta agriculturae universitatis jilinensis",
  ["jilin xu mu shou yi"] = "Jilin xu mu shou yi",
  ["jimd rep"] = "JIMD reports",
  ["jimlar mutane"] = "Jimlar mutane",
  ["jin dai shi yan jiu"] = "Jin dai shi yan jiu",
  ["jin dai zhong shi yan jiu tong xun"] = "Jin dai Zhongguo shi yan jiu tong xun : Newsletter for modern Chinese history",
  ["jin dai zhongguo fu nu shi yan jiu"] = "Jin dai Zhongguo fu nü shi yan jiu",
  ["jinan da xue xue bao"] = "Jinan da xue xue bao. Zi ran ke xue yu yi xue ban",
  ["jing ji lin yan jiu"] = "Jing ji lin yan jiu = Economic Forest Researches",
  ["jing ji lun wen"] = "Jing ji lun wen",
  ["jinko mondai kenkyu"] = "Jinko mondai kenkyu. [Journal of population problems]",
  ["jinko mondai kenkyusho nenpo"] = "Nenpo : Annual reports. Jinkō Mondai Kenkyūjo (Japan)",
  ["jinkogaku kenkyu"] = "Jinkōgaku kenkyū",
  ["jinrui idengaku zasshi"] = "Jinrui idengaku zasshi. The Japanese journal of human genetics",
  ["jinruigaku zasshi"] = "Jinruigaku zasshi = The Journal of the Anthropological Society of Nihon",
  ["jinruigaku zasshi|anthropol sci"] = "Anthropological science : journal of the Anthropological Society of Nippon",
  ["jinshu rechuli"] = "Jinshu Rechuli",
  ["jinshu xuebao"] = "Jinshu Xuebao",
  ["jircas j sci pap"] = "JIRCAS journal for scientific papers",
  ["jise j inf sci eng"] = "JISE. Journal of Information Science and Engineering",
  ["jisuanji yu yingyong huaxue"] = "Jisuanji Yu Yingyong Huaxue",
  ["jk pract"] = "JK practitioner : a journal of current clinical medicine & surgery",
  ["jma j"] = "JMA journal",
  ["jme pract bioethics"] = "JME Practical Bioethics",
  ["jmir aging"] = "JMIR aging",
  ["jmir ai"] = "JMIR AI",
  ["jmir bioinf biotechnol"] = "JMIR Bioinformatics and Biotechnology",
  ["jmir bioinform biotech"] = "JMIR bioinformatics and biotechnology",
  ["jmir biomed eng"] = "JMIR biomedical engineering",
  ["jmir cancer"] = "JMIR cancer",
  ["jmir cardio"] = "JMIR cardio",
  ["jmir data"] = "JMIR Data",
  ["jmir dermatol"] = "JMIR dermatology",
  ["jmir diabetes"] = "JMIR diabetes",
  ["jmir form res"] = "JMIR formative research",
  ["jmir hum factors"] = "JMIR human factors",
  ["jmir infodemiol"] = "JMIR Infodemiology",
  ["jmir med educ"] = "JMIR medical education",
  ["jmir med inf"] = "JMIR Medical Informatics",
  ["jmir med inform"] = "JMIR medical informatics",
  ["jmir ment health"] = "JMIR mental health",
  ["jmir mhealth uhealth"] = "JMIR mHealth and uHealth",
  ["jmir nurs"] = "JMIR nursing",
  ["jmir pediatr parent"] = "JMIR pediatrics and parenting",
  ["jmir perioper med"] = "JMIR perioperative medicine",
  ["jmir public health surveill"] = "JMIR public health and surveillance",
  ["jmir public health surveillance"] = "JMIR public health and surveillance",
  ["jmir rehabil assist technol"] = "JMIR rehabilitation and assistive technologies",
  ["jmir rehabil assistive technol"] = "JMIR Rehabilitation and Assistive Technologies",
  ["jmir res protoc"] = "JMIR research protocols",
  ["jmir serious games"] = "JMIR serious games",
  ["jmlr workshop conf proc"] = "JMLR workshop and conference proceedings",
  ["jmm case rep"] = "JMM case reports",
  ["jmst adv"] = "JMST Advances",
  ["jnaiam j numer anal ind appl math"] = "JNAIAM. Journal of Numerical Analysis, Industrial and Applied Mathematics",
  ["jnci cancer spectr"] = "JNCI cancer spectrum",
  ["jnma j nepal med assoc"] = "JNMA; journal of the Nepal Medical Association",
  ["jnul publ"] = "JNUL Publications",
  ["joannea geol palaontol"] = "Joannea--Geologie und Paläontologie",
  ["joensuun yliop luonnont julk"] = "Joensuun Yliopiston Luonnontieteellisiä Julkaisuja",
  ["jogn nurs"] = "JOGN nursing; journal of obstetric, gynecologic, and neonatal nursing",
  ["john marshall j comput inf law"] = "The John Marshall journal of computer & information law",
  ["john marshall j pract proced"] = "The John Marshall journal of practice and procedure",
  ["john marshall law rev"] = "The John Marshall law review",
  ["johns hopkins alumni mag"] = "The Johns Hopkins alumni magazine",
  ["johns hopkins apl tech dig"] = "Johns Hopkins APL technical digest",
  ["johns hopkins mag"] = "Johns Hopkins magazine",
  ["johns hopkins med j"] = "The Johns Hopkins medical journal",
  ["johns hopkins med j suppl"] = "Johns Hopkins medical journal. Supplement",
  ["johns hopkins med lett health after 50"] = "The Johns Hopkins medical letter health after 50",
  ["johns hopkins nurses alumnae mag"] = "The Johns Hopkins nurses alumnae magazine",
  ["johns hopkins nurses alumni mag"] = "The Johns Hopkins Nurses' alumni magazine",
  ["johns hopkins stud hist math"] = "Johns Hopkins Studies in the History of Mathematics",
  ["johns hopkins stud math sci"] = "Johns Hopkins Studies in the Mathematical Sciences",
  ["johnson matthey technol rev"] = "Johnson Matthey Technology Review",
  ["joicfp news"] = "Joicfp News",
  ["joicfp rev"] = "JOICFP review",
  ["joint bone spine"] = "Joint bone spine",
  ["joint eurohaptics conf symp haptic interfaces virtual environ teleoper syst"] = "Joint EuroHaptics Conference and Symposium on Haptic Interfaces for Virtual Environment and Teleoperator Systems : World Haptics Conference. World Haptics Conference",
  ["joj ophthalmol"] = "JOJ ophthalmology",
  ["jom (1989)"] = "JOM (Warrendale, Pa. : 1989)",
  ["jomec j"] = "JOMEC journal : journalism, media and cultural studies",
  ["jonas healthc law ethics regul"] = "JONA'S healthcare law, ethics and regulation",
  ["jones bartlett books math comput sci"] = "Jones and Bartlett Books in Mathematics and Computer Science",
  ["jones bartlett ser comput sci"] = "Jones and Bartlett Series in Computer Science",
  ["jor spine"] = "JOR spine",
  ["jordan dent j"] = "The Jordan dental journal",
  ["jordan j biol sci"] = "Jordan journal of biological sciences",
  ["jordan j chem"] = "Jordan Journal of Chemistry",
  ["jordan j math stat"] = "Jordan Journal of Mathematics and Statistics",
  ["jordan med j"] = "Jordan medical journal",
  ["jorn med"] = "Jornada médica",
  ["josai shika daigaku kiyo"] = "Jōsai Shika Daigaku kiyō. The Bulletin of the Josai Dental University",
  ["josanpu zasshi"] = "Josanpu zasshi = The Japanese journal for midwife",
  ["joshi eiyo daigaku kiyo"] = "Jōshi Eiyō Daigaku kiyō",
  ["jot, j oberflaechentech"] = "JOT Journal fur Oberflaechentechnik",
  ["journ annu diabetol hotel dieu"] = "Journées annuelles de diabétologie de l'Hôtel-Dieu",
  ["journ equ deriv partielles"] = "Journees Equations aux Derives Partielles",
  ["journ sci cent nat coord etud rech nutr aliment"] = "Les Journées scientifiques du Centre national de coordination des études et recherches sur la nutrition et l'alimentation",
  ["journal (can soc forensic sci)"] = "Journal (Canadian Society of Forensic Science)",
  ["journal (inst health rec inf manag)"] = "Journal (Institute of Health Record Information and Management)",
  ["journal commun monogr"] = "Journalism & communication monographs",
  ["journal conf workshop ada"] = "Journalism Conference : [proceedings]. Journalism Conference",
  ["journal hist"] = "Journalism history",
  ["journal huazhong (central china) agricultural university|hua zhong nong ye da xue xue bao"] = "Hua zhong nong ye da xue xue bao",
  ["journal mass commun q"] = "Journalism & mass communication quarterly",
  ["journal of agricultural and biological science|guangxi nongye shengwu kexue"] = "Guangxi nongye shengwu kexue",
  ["journal of agricultural biotechnology|nong ye sheng wu ji shu xue bao"] = "Nong ye sheng wu ji shu xue bao",
  ["journal of analytical bio-science|seibutsu shiryo bunseki"] = "Seibutsu shiryo bunseki",
  ["journal of anhui agricultural college|anhui nongxueyuan xuebao"] = "Anhui nongxueyuan xuebao",
  ["journal of animal genetics|dobutsu iden ikushu kenkyu"] = "Dobutsu iden ikushu kenkyu",
  ["journal of applied ichthyology|j appl ichthyol"] = "Zeitschrift fur angewandte Ichthyologie",
  ["journal of beijing medical university|beijing yi ke da xue xue bao"] = "Beijing yi ke da xue xue bao",
  ["journal of china agricultural university|zhongguo nong ye da xue xue bao"] = "Zhongguo nong ye da xue xue bao",
  ["journal of chinese biotechnology|zhongguo sheng wu gong cheng za zhi"] = "Zhongguo sheng wu gong cheng za zhi",
  ["journal of east china normal university|huadong shifan daxue xuebao"] = "Huadong shifan daxue xuebao",
  ["journal of fishery sciences of china / zhongguo shui chan ke xue yan jiu yuan bian ji chu ban|zhongguo shui chan ke xue"] = "Zhongguo shui chan ke xue",
  ["journal of fruit science|guoshu xuebao"] = "Guoshu xuebao",
  ["journal of fudan university natural science|fu dan xue bao zi ran ke xue ban"] = "Fu dan xue bao. Zi ran ke xue ban",
  ["journal of heng yang teachers’ college|hengyang shi fan xue yuan xue bao"] = "Hengyang shi fan xue yuan xue bao",
  ["journal of hunan agricultural university|hunan nong ye da xue xue bao"] = "Hunan nong ye da xue xue bao",
  ["journal of immunology|mian yi xue za zhi"] = "Mian yi xue za zhi",
  ["journal of japan soy sauce research institute|nihon shoyu kenkyujo zasshi"] = "Nihon Shoyu Kenkyujo zasshi",
  ["journal of kanazawa medical university|kanazawa ika daigaku zasshi"] = "Kanazawa Ika Daigaku zasshi",
  ["journal of pathogen biology|zhongguo bing yuan sheng wu xue za zhi"] = "Zhongguo bing yuan sheng wu xue za zhi",
  ["journal of peanut science|hua sheng xue bao"] = "Hua sheng xue bao",
  ["journal of shandong university natural science|shandong da xue xue bao"] = "Shandong Da Xue xue bao. Li xue ban",
  ["journal of shenyang pharmaceutical university|shenyang yao ke da xue xue bao"] = "Shenyang yao ke da xue xue bao",
  ["journal of snake|she zhi"] = "She zhi",
  ["journal of south china agricultural university|hua nan nong ye da xue xue bao"] = "Hua nan nong ye da xue xue bao",
  ["journal of the chinese society of animal science|j chin soc anim sci"] = "Zhongguo xu mu xue hui hui zhi",
  ["journal of tianjin medical university|tianjin yi ke da xue xue bao"] = "Tianjin yi ke da xue xue bao",
  ["journal of tianjin university|tianjin daxue xuebao"] = "Tianjin da xue xue bao",
  ["journal of tropical medicine|re dai yi xue za zhi"] = "Re dai yi xue za zhi",
  ["journal of wuhan university natural science edition|wuhan da xue xue bao li xue ban"] = "Wuhan da xue xue bao. Li xue ban",
  ["journal of xiamen university natural science|xiamen da xue xue bao zi ran ke xue ban"] = "Xiamen da xue xue bao. Zi ran ke xue ban",
  ["journal of yunnan agricultural university|yunnan nong ye da xue xue bao"] = "Yunnan nong ye da xue xue bao",
  ["journal q"] = "The journalism quarterly",
  ["journal stud"] = "Journalism studies",
  ["journalism (lond)"] = "Journalism (London, England)",
  ["journalism monogr"] = "Journalism monographs (Austin, Tex.)",
  ["jove chem"] = "JoVE Chemistry",
  ["jp j biostat"] = "JP journal of biostatistics",
  ["jpc j planar chromatogr - mod tlc"] = "JPC Journal of Planar Chromatography Modern TLC",
  ["jpen j parenter enteral nutr"] = "JPEN. Journal of parenteral and enteral nutrition",
  ["jpgn rep"] = "JPGN reports",
  ["jphys energy"] = "JPhys Energy",
  ["jphys mater"] = "JPhys Materials",
  ["jphys photonics"] = "JPhys photonics",
  ["jpma j pak med assoc"] = "JPMA: Journal of the Pakistan Medical Association",
  ["jpn agric res q"] = "Japan agricultural research quarterly",
  ["jpn circ j"] = "Japanese circulation journal",
  ["jpn clin med"] = "Japanese clinical medicine",
  ["jpn dent j"] = "Japanese dental journal",
  ["jpn dent sci rev"] = "The Japanese dental science review",
  ["jpn econ rev"] = "The Japanese Economic Review. The Journal of the Japanese Economic Association",
  ["jpn econ rev (oxf)"] = "Japanese economic review (Oxford, England)",
  ["jpn heart j"] = "Japanese heart journal",
  ["jpn hosp"] = "Japan-hospitals : the journal of the Japan Hospital Association",
  ["jpn j antibiot"] = "The Japanese journal of antibiotics",
  ["jpn j appl entomol zool"] = "Japanese Journal of Applied Entomology and Zoology",
  ["jpn j appl phys"] = "Japanese Journal of Applied Physics",
  ["jpn j appl phys (2008)"] = "Japanese journal of applied physics (2008)",
  ["jpn j appl phys part 2 lett express lett"] = "Japanese Journal of Applied Physics, Part 2: Letters and Express Letters",
  ["jpn j appl phys pt 1"] = "Japanese journal of applied physics. Part 1, Regular papers & short notes",
  ["jpn j appl phys, part 1"] = "Japanese Journal of Applied Physics Part 1",
  ["jpn j appl phys, part 2"] = "Japanese Journal of Applied Physics Part 2",
  ["jpn j behaviormetrics"] = "The Japanese Journal of Behaviormetrics",
  ["jpn j cancer chemother"] = "Japanese Journal of Cancer and Chemotherapy",
  ["jpn j cancer res"] = "Japanese journal of cancer research : Gann",
  ["jpn j chemother"] = "Japanese Journal of Chemotherapy",
  ["jpn j clin chem"] = "Japanese Journal of Clinical Chemistry",
  ["jpn j clin oncol"] = "Japanese journal of clinical oncology",
  ["jpn j clin pharmacol ther"] = "Japanese Journal of Clinical Pharmacology and Therapeutics",
  ["jpn j crop sci"] = "Nihon Sakumotsu Gakkai kiji",
  ["jpn j dermatol b"] = "The Japanese journal of dermatology. Ser. B",
  ["jpn j ecol"] = "Japanese Journal of Ecology",
  ["jpn j electrophor"] = "Japanese journal of electrophoresis",
  ["jpn j exp med"] = "The Japanese journal of experimental medicine",
  ["jpn j forensic toxicol"] = "Japanese Journal of Forensic Toxicology",
  ["jpn j gastroenterol"] = "Japanese journal of gastroenterology",
  ["jpn j genet"] = "Idengaku zasshi",
  ["jpn j health phys"] = "Japanese Journal of Health Physics",
  ["jpn j helicobacter res"] = "Japanese journal of helicobacter research",
  ["jpn j hum genet"] = "The Japanese journal of human genetics",
  ["jpn j hum geogr"] = "Japanese Journal of Human Geography",
  ["jpn j ind appl math"] = "Japan Journal of Industrial and Applied Mathematics",
  ["jpn j infect dis"] = "Japanese journal of infectious diseases",
  ["jpn j legal med"] = "Japanese Journal of Legal Medicine",
  ["jpn j limnol"] = "Japanese Journal of Limnnology",
  ["jpn j math"] = "Japanese Journal of Mathematics",
  ["jpn j math (ns)"] = "Japanese Journal of Mathematics",
  ["jpn j med"] = "Japanese journal of medicine",
  ["jpn j med (lond)"] = "Japan journal of medicine",
  ["jpn j med sci biol"] = "Japanese journal of medical science & biology",
  ["jpn j microbiol"] = "Japanese journal of microbiology",
  ["jpn j nurs sci"] = "Japan journal of nursing science : JJNS",
  ["jpn j obstet gynecol"] = "The Japanese journal of obstetrics and gynecology",
  ["jpn j ophthalmol"] = "Japanese journal of ophthalmology",
  ["jpn j pharmacol"] = "Japanese Journal of Pharmacology",
  ["jpn j phycol"] = "Japanese Journal of Phycology",
  ["jpn j phys"] = "Japanese Journal of Physics",
  ["jpn j physiol"] = "The Japanese journal of physiology",
  ["jpn j popul"] = "The Japanese journal of population",
  ["jpn j psychiatry neurol"] = "The Japanese journal of psychiatry and neurology",
  ["jpn j psychon sci"] = "Kiso shinrigaku kenkyū",
  ["jpn j radiol"] = "Japanese journal of radiology",
  ["jpn j sports med phys fitness"] = "Japanese Journal of Physical Fitness and Sports Medicine",
  ["jpn j stat data sci"] = "Japanese Journal of Statistics and Data Science",
  ["jpn j surg"] = "The Japanese journal of surgery",
  ["jpn j syst entomol"] = "Japanese journal of systematic entomology = Nihon Konchū Bunrui Gakkai kaihō",
  ["jpn j thorac cardiovasc surg"] = "The Japanese journal of thoracic and cardiovascular surgery : official publication of the Japanese Association for Thoracic Surgery = Nihon Kyōbu Geka Gakkai zasshi",
  ["jpn j toxicol environ health"] = "Japanese Journal of toxicology and Environment health",
  ["jpn j tuberc"] = "The Japanese journal of tuberculosis",
  ["jpn j tuberc chest dis"] = "The Japanese journal of tuberculosis and chest diseases",
  ["jpn j vet res"] = "Japanese Journal of Veterinary Research",
  ["jpn med j"] = "[Nihon iji shinpō] = Japanese medical journal",
  ["jpn med j (natl inst health jpn)"] = "Japanese medical journal",
  ["jpn psychol res"] = "The Japanese psychological research",
  ["jpn q"] = "Japan quarterly (Asahi Shinbunsha)",
  ["jpn soc mech eng int j sera:"] = "Japan Society of Mechanical Engineers International Journal Series A: Solid Mechanics and Material Engineering",
  ["jpn soc mech eng int j serb:"] = "Japan Society of Mechanical Engineers International Journal Series B: Fluids and Thermal Engineering",
  ["jpn soc mech eng int j serc:"] = "Japan Society of Mechanical Engineers International Journal Series C: Mechanical Systems, Machine Elements and Manufacturing",
  ["jpn stud hist sci"] = "Japanese studies in the history of science",
  ["jpo j pract orthod"] = "JPO: the journal of practical orthodontics",
  ["jpras open"] = "JPRAS open",
  ["jps conf proc"] = "JPS conference proceedings",
  ["jpss j probab stat sci"] = "JPSS. Journal of Probability and Statistical Science",
  ["jr off advis group"] = "Junior Officer Advisory Group : a newsletter for junior officers by junior officers",
  ["jrsm cardiovasc dis"] = "JRSM cardiovascular disease",
  ["jrsm open"] = "JRSM open",
  ["jrsm short rep"] = "JRSM short reports",
  ["jsac grapevine"] = "JSAC grapevine. Joint Strategy and Action Committee (New York, N.Y.)",
  ["jses int"] = "JSES international",
  ["jses open access"] = "JSES open access",
  ["jsiam lett"] = "JSIAM Letters",
  ["jsm allergy asthma"] = "JSM allergy and asthma",
  ["jsm alzheimers dis relat dement"] = "JSM Alzheimer's disease and related dementia",
  ["jsm atheroscler"] = "JSM atherosclerosis",
  ["jsm biochem mol biol"] = "JSM biochemistry and molecular biology",
  ["jsm bioinform genom proteom"] = "JSM Bioinformatics, genomics and preteomics",
  ["jsm biomark"] = "JSM biomarkers",
  ["jsm biomed imaging data pap"] = "JSM biomedical imaging data papers",
  ["jsm biotechnol biomed eng"] = "JSM biotechnology & biomedical engineering",
  ["jsm cell dev biol"] = "JSM cell & developmental biology",
  ["jsm chem"] = "JSM Chemistry",
  ["jsm dent surg"] = "JSM dental surgery",
  ["jsm microbiol"] = "JSM microbiology",
  ["jsm neurosurg spine"] = "JSM neurosurgery and spine",
  ["jsm trop med res"] = "JSM tropical medicine and research",
  ["jsme int j, ser a"] = "JSME International Journal Series A: Mechanics and Material Engineering",
  ["jsme int j, ser b"] = "JSME International Journal Series B: Fluids and Thermal Engineering",
  ["jsme int j, ser c"] = "JSME International Journal Series C: Mechanical Systems Machine Elements and Manufacturing",
  ["jss res ser stat"] = "JSS Research Series in Statistics",
  ["jt comm j qual improv"] = "The Joint Commission journal on quality improvement",
  ["jt comm j qual patient saf"] = "Joint Commission journal on quality and patient safety",
  ["jt comm j qual saf"] = "Joint Commission journal on quality and safety",
  ["jt comm perspect"] = "Joint Commission perspectives. Joint Commission on Accreditation of Healthcare Organizations",
  ["jt dis relat surg"] = "Joint diseases and related surgery",
  ["jt meet abstr am dairy sci assoc"] = "Joint meeting abstracts. American Dairy Science Association. Meeting",
  ["jtcvs tech"] = "JTCVS techniques",
  ["jto clin res rep"] = "JTO clinical and research reports",
  ["jud librariansh"] = "Judaica librarianship",
  ["judges j"] = "The Judges' journal",
  ["judgm decis mak"] = "Judgment and decision making",
  ["jugosl ginekol opstet"] = "Jugoslavenska ginekologija i opstetricija",
  ["jugosl ginekol perinatol"] = "Jugoslavenska ginekologija i perinatologija",
  ["jugosl pregl"] = "Jugoslovenski pregled; informativno dokumentarni priručnik o Jugoslaviji",
  ["julk suom naishammaslaak ryhma"] = "Julkaisu. Suomen Naishammaslääkärit ryhmä, Helsingfors",
  ["jun wu xi tong"] = "Jun wu xi tong = Mycosystema",
  ["jun wu xue bao"] = "Jun wu xue bao = Mycosystema",
  ["jundishapur j microbiol"] = "Jundishapur journal of microbiology",
  ["jundishapur j nat pharm prod"] = "Jundishapur journal of natural pharmaceutical products",
  ["juniper online j mater sci"] = "Juniper online journal material science",
  ["juniper online j public health"] = "Juniper online journal of public health",
  ["junwu yanjiu"] = "Junwu yanjiu",
  ["jurid rev"] = "The Juridical review",
  ["juris dr"] = "Juris doctor",
  ["justice q"] = "Justice quarterly : JQ",
  ["justice res policy"] = "Justice research and policy : journal of the Justice Research and Statistics Association : JRP",
  ["justus liebigs ann chem"] = "Justus Liebigs Annalen der Chemie",
  ["juv fam court j"] = "Juvenile & family court journal",
  ["juv med"] = "La Juventud médica",
  ["jvs vasc sci"] = "JVS-vascular science",
  ["jvst a"] = "Journal of Vacuum Science and Technology A",
  ["jvst b"] = "Journal of Vacuum Science and Technology B",
  ["jyske saml"] = "Jyske samlinger",
  ["k dan vidensk selsk mat fys medd"] = "Kongelige Danske Videnskabernes Selskab, Matematisk-Fysiske Meddelelser",
  ["k dan vidensk selsk mater fys medd"] = "Kongelige Danske Videnskabernes Selskab, Matematisk-Fysiske Meddelelser",
  ["k krigsvetenskapakad handlingar tidskr"] = "Kungliga Krigsvetenskapsakademiens handlingar och tidskrift",
  ["k ned akad wet verh afd lett (n s)"] = "Koninklijke Nederlandse Akademie van Wetenschappen",
  ["k rep"] = "K [reports]. U.S. Atomic Energy Commission",
  ["kaen kaset"] = "Kǣn kasēt = Khon Kaen agriculture journal",
  ["kafkas univ vet fak derg"] = "Kafkas Üniversitesi veteriner fakültesi dergisi",
  ["kagaku (kyoto)"] = "Kagaku (Kyoto)",
  ["kagaku kogaku"] = "Kagaku Kogaku",
  ["kagaku kogaku ronbunshu"] = "Kagaku Kogaku Ronbunshu",
  ["kagaku to seibutsu"] = "Kagaku to seibutsu. [Chemistry and biology]",
  ["kagakushi kenkyu"] = "Kagakushi kenkyu. [Journal of the history of science, Japan",
  ["kagoshima daigaku igaku zasshi"] = "[Igaku zasshi] Medical journal. Kagoshima Daigaku. Igakubu",
  ["kagoshima daigaku shigakubu kiyo"] = "Kagoshima Daigaku Shigakubu kiyō = Annals of Kagoshima University Dental School",
  ["kaibogaku zasshi"] = "Kaibogaku zasshi. Journal of anatomy",
  ["kaiin dayori nihon kontakuto renzu gakkai"] = "Kaiin dayori. [Membership news]. Nippon Kontakuto Renzu Gakkai",
  ["kaiin dayori nippon kontakuto renzu gakkai"] = "Kaiin dayori. [Membership news]. Nippon Kontakuto Renzu Gakkai",
  ["kajok kyehoek nonjip"] = "Kajok kyehoek nonjip. Journal of family planning studies",
  ["kako gijutsu (osaka)"] = "Kako Gijutsu (Osaka)",
  ["kaku igaku"] = "Kaku igaku. The Japanese journal of nuclear medicine",
  ["kalpa publ comput"] = "Kalpa Publications in Computing",
  ["kampener alm"] = "Kamper almanak",
  ["kan, tan, sui"] = "Kan, Tan, Sui",
  ["kanagawa kenritsu hakubutsukan kenkyu hokoku shizen kagaku"] = "Kanagawa Kenritsu Hakubutsukan kenkyū hōkoku. Shizen kagaku = Bulletin of the Kanagawa Prefectural Museum. Natural science",
  ["kanagawa shigaku"] = "Kanagawa shigaku. The Journal of the Kanagawa Odontological Society",
  ["kanazawa ika daigaku zasshi"] = "Kanazawa Ika Daigaku zasshi = Journal of Kanazawa Medical University",
  ["kango gijutsu"] = "[Kango gijutsu] : [Nursing technique]",
  ["kango kenkyu"] = "Kango kenkyu. The Japanese journal of nursing research",
  ["kango kyoiku"] = "[Kango kyōiku] Japanese journal of nurses' education",
  ["kango kyoshitsu"] = "Kango kyoshitsu. [Nursing classroom]",
  ["kango tenbo"] = "Kango tenbō. The Japanese journal of nursing science",
  ["kangogaku zasshi"] = "Kangogaku zasshi",
  ["kanho hakhoe chi"] = "Kanho Hakhoe chi [The Journal of Nurses Academic Society]",
  ["kanhohak tamgu"] = "Kanhohak t'amgu",
  ["kankyo kagaku sogo kenkyujo nenpo"] = "Kankyo Kagaku Sogo Kenkyujo nenpo",
  ["kans geogr"] = "The Kansas geographer",
  ["kans hist"] = "Kansas history",
  ["kans hist q"] = "The Kansas historical quarterly",
  ["kans j med"] = "Kansas journal of medicine",
  ["kans med"] = "Kansas medicine : the journal of the Kansas Medical Society",
  ["kans nurse"] = "The Kansas nurse",
  ["kans statut annot kans"] = "Kansas statutes annotated : duly arranged, numbered, annotated, and indexed with histories, notes, and references as required by law. Kansas",
  ["kansai byochugai kenkyukaiho"] = "Kansai Byōchūgai Kenkyūkai hō",
  ["kansai daigaku keizai ronshu"] = "Kansai Daigaku keizai ronshū",
  ["kansant aikak"] = "Kansantaloudellinen Aikakauskirja",
  ["kansenshogaku zasshi"] = "Kansenshōgaku zasshi. The Journal of the Japanese Association for Infectious Diseases",
  ["kantou tousan byougaichuu kenkyuukaihou"] = "Kantou Tousan Byougaichuu Kenkyuukaihou",
  ["kao hsiung i hsueh ko hsueh tsa chih"] = "Kao-Hsiung I Hsueh Ko Hsueh Tsa Chih [Kaohsiung Journal Of Medical Sciences]",
  ["kao ku jen lei hsueh kan"] = "Kao gu ren lei xue kan",
  ["kaohsiung j med sci"] = "The Kaohsiung journal of medical sciences",
  ["karachi univ j sci"] = "Karachi University",
  ["karbala int j mod sci"] = "Karbala International Journal of Modern Science",
  ["kardiochir torakochirurgia pol"] = "Kardiochirurgia i torakochirurgia polska = Polish journal of cardio-thoracic surgery",
  ["kardiol pol"] = "Kardiologia polska",
  ["kasetsart j soc sci"] = "Witthayāsān Kasētsāt. Sākhā sangkhommasāt = The Kasetsart journal. Social sciences",
  ["kassenzahnarzt colloq med dent"] = "Der Kassenzahnarzt; Colloquium med. dent",
  ["katedrah betoldot erets yisrael veyishuvah"] = "Ḳatedrah be-toldot Erets-Yiśraʼel ṿe-yishuvah",
  ["kathmandu univ med j"] = "Kathmandu University Medical Journal",
  ["kathmandu univ med j (kumj)"] = "Kathmandu University medical journal (KUMJ)",
  ["katollik taehak uihakpu nonmunjip"] = "Kat'ollik Taehak Uihakpu nonmunjip",
  ["kauch rezina"] = "Kauchuk i Rezina",
  ["kautsch gummi kunstst"] = "Kautschuk Gummi Kunststoffe",
  ["kavk entomol bjul"] = "Kavkazskij ėntomologičeskij bjulleten = Caucasian entomological bulletin",
  ["kazan med zh"] = "Kazanskiĭ meditsinskiĭ zhurnal",
  ["ke ji dao bao"] = "Ke ji dao bao = Science & technology review",
  ["ke xue fa zhan"] = "Ke xue fa zhan",
  ["keele cognition sem"] = "Keele Cognition Seminars",
  ["keep posted"] = "Keeping posted",
  ["keio econ stud"] = "Keio economic studies",
  ["keio j med"] = "The Keio journal of medicine",
  ["keio sci tech rep"] = "Keio Science and Technology Reports",
  ["keisei geka"] = "Keisei geka. Plastic & reconstructive surgery",
  ["kek proc"] = "KEK Proceedings",
  ["kem ind"] = "Kemija u Industriji",
  ["kemikaru enjiniyaringu"] = "Kemikaru Enjiniyaringu",
  ["kendall’s lib statist"] = "Kendall’s Library of Statistics",
  ["kennedy inst ethics j"] = "Kennedy Institute of Ethics journal",
  ["kenya nurs j"] = "Kenya nursing journal",
  ["kenya stat dig"] = "Kenya statistical digest",
  ["keram z"] = "Keramische Zeitschrift",
  ["kertesz szolesz"] = "Kerteszet es szoleszet",
  ["kew bull"] = "Kew Bulletin",
  ["kexue tongbao"] = "Ko Hsueh Tung Pao = Kexue Tongbao [English Language Edition]",
  ["kexue tongbao (chinese)"] = "Kexue Tongbao",
  ["kexue tongbao [english language edition]|kexue tongbao"] = "Ko Hsueh Tung Pao",
  ["kexue tongbao|chin sci bull"] = "Chinese science bulletin",
  ["key eng mater"] = "Key Engineering Materials",
  ["key report"] = "The Key reporter",
  ["key texts classic stud hist ideas"] = "Key Texts: Classic Studies in the History of Ideas",
  ["keystone folkl q"] = "Keystone folklore quarterly",
  ["kgk, kautsch gummi kunstst"] = "KGK, Kautschuk Gummi Kunststoffe",
  ["khayyam j math"] = "Khayyam Journal of Mathematics",
  ["khim drev"] = "Khimiya Drevesiny",
  ["khim fiz"] = "Khimicheskaya Fizika",
  ["khim interesakh ustoich razvit"] = "Khimiya v Interesakh Ustoichivogo Razvitiya",
  ["khim neft mashinostr"] = "Khimicheskoe i Neftyanoe Mashinostroenie",
  ["khim prom-st (st petersburg, russ fed)"] = "Khimicheskaya Promyshlennost (St. Petersburg, Russian Federation)",
  ["khim prom-st segodnya"] = "Khimicheskaya Promyshlennost Segodnya",
  ["khim sel’sk khoz"] = "Khimizatsiya Sel’skogo Khozyaistva",
  ["khim tekhnol topl masel"] = "Khimiya i Tekhnologiya Topliv i Masel",
  ["khim tekhnol vody"] = "Khimiya i Tekhnologiya Vody",
  ["khim tverd topl (moscow, russ fed)"] = "Khimiya Tverdogo Topliva (Moscow, Russian Federation)",
  ["khim volokna"] = "Khimicheskie Volokna",
  ["khim vys energ"] = "Khimiya Vysokikh Energii",
  ["khimiia meditsina"] = "Khimiia i meditsina",
  ["khirurgiia (mosk)"] = "Khirurgiia",
  ["khirurgiia (sofiia)"] = "Khirurgiia",
  ["khoa hoc ky thuat thu y"] = "Khoa học Kỹ thuật Thú y",
  ["khoa hoc va cong nghe"] = "Khoa học và công nghệ",
  ["ki kunstliche intell"] = "KI Kunstliche Intelligenz",
  ["kias springer ser math"] = "KIAS Springer Series in Mathematics",
  ["kibern neogranich vozmozhn vozmozhn ogranich"] = "\\cyr Kibernetika—Neogranichennye Vozmozhnosti i Vozmozhnye Ogranicheniya",
  ["kibernet i vychisl tekhn"] = "Akademiya Nauk Ukrainy",
  ["kibernet sistem anal"] = "Natsionalnaya Akademiya Nauk Ukrainy. Institut Kibernetiki im. V. M. Glushkova. Kibernetika i Sistemnyi Analiz",
  ["kichin kitosan kenkyu"] = "Kichin, kitosan kenkyu = Chitin and chitosan research",
  ["kidney blood press res"] = "Kidney & blood pressure research",
  ["kidney blood pressure res"] = "Kidney and Blood Pressure Research",
  ["kidney cancer"] = "Kidney cancer (Clifton, Va.)",
  ["kidney cancer j"] = "Kidney cancer journal : official journal of the Kidney Cancer Association",
  ["kidney dis (basel)"] = "Kidney diseases (Basel, Switzerland)",
  ["kidney int"] = "Kidney international",
  ["kidney int rep"] = "Kidney international reports",
  ["kidney int suppl"] = "Kidney international. Supplement",
  ["kidney int suppl (2011)"] = "Kidney international supplements",
  ["kidney med"] = "Kidney medicine",
  ["kidney res clin pract"] = "Kidney research and clinical practice",
  ["kiel geogr schr"] = "Kieler geographische Schriften",
  ["kiel stud"] = "Kieler Studien",
  ["kieler beitr gesch med pharm"] = "Kieler Beiträge zur Geschichte der Medizin und Pharmazie",
  ["kikan riron keizaigaku"] = "Kikan riron-keizaigaku",
  ["killidata ser"] = "Killi-data series",
  ["kinderarztl prax"] = "Kinderarztliche Praxis",
  ["kinematics phys celestial bodies"] = "Kinematics and Physics of Celestial Bodies",
  ["kinesiol rev (champaign)"] = "Kinesiology review (Champaign, Ill.)",
  ["kinesiology (zagreb)"] = "Kinesiology (Zagreb, Croatia)",
  ["kinet catal"] = "Kinetics and Catalysis",
  ["kinet katal"] = "Kinetika i Kataliz",
  ["kinet rel models"] = "Kinetic and Related Models",
  ["kinet relat models"] = "Kinetic and Related Models",
  ["kinetoplastid biol dis"] = "Kinetoplastid biology and disease",
  ["king faisal spec hosp med j"] = "The King Faisal Specialist Hospital medical journal",
  ["kings coll hosp gaz"] = "King's College Hospital gazette",
  ["kings coll hosp gaz clin suppl"] = "King's College Hospital gazette. Clinical supplement",
  ["kings law j"] = "King's law journal : KLJ",
  ["kinjin kenkyusho kenkyu hokoku"] = "Kinjin Kenkyusho kenkyu hokoku",
  ["kinki univ ser quantum comput"] = "Kinki University Series on Quantum Computing",
  ["kino zairyo"] = "Kino Zairyo",
  ["kiph bull"] = "KIPH bulletin",
  ["kirk-othmer encycl chem technol"] = "Kirk-Othmer Encyclopedia of Chemical Technology",
  ["kiryat sefer"] = "Kiryat sefer",
  ["kisaengchunghak chapchi"] = "Kisaengch'unghak chapchi. The Korean journal of parasitology",
  ["kisechugaku zasshi"] = "Kiseichugaku zasshi. Japanese journal of parasitology",
  ["kiseichugaku zasshi"] = "Kiseichugaku zasshi. Japanese journal of parasitology",
  ["kiserl orvostud"] = "Kísérletes orvostudomány",
  ["kita kanto igaku"] = "Kita Kantō igaku. Kitakanto medical journal",
  ["kitasato arch exp med"] = "The Kitasato archives of experimental medicine",
  ["kitasato med j"] = "The Kitasato medical journal",
  ["klagenf beitr didakt math"] = "Klagenfurter Beitrage zur Didaktik der Mathematik",
  ["klass texte wiss"] = "Klassische Texte der Wissenschaft",
  ["kleine senckenbr"] = "Kleine Senckenbergreihe",
  ["klimik derg"] = "KLIMIK dergisi",
  ["klin anasthesiol intensivther"] = "Klinische Anasthesiologie und Intensivtherapie",
  ["klin imunol alergol"] = "Klinická imunológia a alergológia",
  ["klin khir"] = "Klinicheskaia khirurgiia",
  ["klin lab diagn"] = "Klinicheskaia laboratornaia diagnostika",
  ["klin lech zlokach novoobraz"] = "Klinika i lechenie zlokachestvennykh novoobrazovaniĭ",
  ["klin med (mosk)"] = "Klinicheskaia meditsina",
  ["klin med osterr z wiss prakt med"] = "Klinische Medizin; österreichische Zeitschrift für wissenschaftliche und praktische Medizin",
  ["klin mikrobiol infekc lek"] = "Klinická mikrobiologie a infekc̆ní lékar̆ství",
  ["klin monatsbl augenheilkd"] = "Klinische Monatsblatter fur Augenheilkunde",
  ["klin monatsblatter augenheilkd augenarztl fortbild"] = "Klinische Monatsblatter fur Augenheilkunde und fur augenarztliche Fortbildung",
  ["klin monbl augenheilkd"] = "Klinische Monatsblätter für Augenheilkunde",
  ["klin monbl augenheilkd augenarztl fortbild"] = "Klinische Monatsblätter für Augenheilkunde und für augenärztliche Fortbildung",
  ["klin neuroradiol"] = "Klinische Neuroradiologie",
  ["klin oczna"] = "Klinika oczna",
  ["klin onkol"] = "Klinická onkologie : casopis Ceské a Slovenské onkologické spolecnosti",
  ["klin padiatr"] = "Klinische Pädiatrie",
  ["klin prax"] = "Klinik und Praxis",
  ["klin sygepleje"] = "Klinisk sygepleje",
  ["klin wochenschr"] = "Klinische Wochenschrift",
  ["klinik psikofarmakol bülteni"] = "Klinik psikofarmakoloji bülteni = Bulletin of clinical psychopharmacology",
  ["klitynna orhanna transplant"] = "Klitynna ta orhanna transplantolohii︠a︡",
  ["kluwer int handb educ"] = "Kluwer International Handbooks of Education",
  ["kluwer int ser discrete event dyn syst"] = "The Kluwer International Series on Discrete Event Dynamic Systems",
  ["kluwer int ser eng comput sci"] = "The Kluwer International Series in Engineering and Computer Science",
  ["kluwer int ser softw eng"] = "The Kluwer International Series in Software Engineering",
  ["kluwer internat ser engrg comput sci knowledge represent, learn expert systems"] = "The Kluwer International Series in Engineering and Computer Science. Knowledge Representation, Learning and Expert Systems",
  ["kluwer texts math sci"] = "Kluwer Texts in the Mathematical Sciences",
  ["km itl sci technol j"] = "KM ITL science and technology journal",
  ["kn j cartogr geogr inf"] = "KN - journal of cartography and geographic information",
  ["knee surg relat res"] = "Knee surgery & related research",
  ["knee surg sports traumatol arthrosc"] = "Knee surgery, sports traumatology, arthroscopy : official journal of the ESSKA",
  ["knjiga (med)"] = "Knjiga. Jugoslavenska akademija znanosti i umjetnosti. Odjel za medicinske nauke",
  ["knowl based syst"] = "Knowledge-based systems",
  ["knowl eng rev"] = "The Knowledge engineering review",
  ["knowl inf syst"] = "Knowledge and information systems",
  ["knowl manag aquat ecosyst"] = "Knowledge and management of aquatic ecosystems",
  ["knowl manage aquat syst"] = "Knowledge and Management of Aquatic Ecosystems",
  ["knowl organ sustain world chall perspect cult sci technol shar connect soc (2016)"] = "Knowledge organization for a sustainable world: challenges and perspectives for cultural, scientific, and technological sharing in a connected society : proceedings of the fourteenth international ISKO conference 27-29 September 2016 Rio de Janeiro, Brazil. Organized by the International Society for Knowledge Organization (ISKO) ISKO-Brazil São Paulo State University",
  ["knowl process manage"] = "Knowledge and Process Management",
  ["knowledge-based syst"] = "Knowledge-Based Systems",
  ["ko hsueh tung pao"] = "[Ke xue tong bao] [Science bulletin]",
  ["kobe daigaku igakubu kiyo"] = "Kobe Daigaku Igakubu kiyo. Medical journal of Kobe University",
  ["kobe econ bus rev"] = "Kobe Economic and Business Review",
  ["kobe ika daigaku kiyo"] = "Kobe Ika Daigaku kiyo",
  ["kobe j math"] = "Kobe Journal of Mathematics",
  ["kobe j med sci"] = "The Kobe journal of medical sciences",
  ["kobe univ econ rev"] = "Kobe University Economic Review",
  ["kobunshi ronbunshu"] = "Kobunshi Ronbunshu",
  ["kocatepe vet derg"] = "Kocatepe veteriner dergisi",
  ["kodai math j"] = "Kodai Mathematical Journal",
  ["kogyo zairyo"] = "Kogyo Zairyo",
  ["koks khim"] = "Koks i Khimiya",
  ["koku eisei gakkai zasshi"] = "Koku Eisei Gakkai zasshi",
  ["kokubyo gakkai zasshi"] = "Kōkūbyō Gakkai zasshi. The Journal of the Stomatological Society, Japan",
  ["kokumin-keizai zasshi"] = "Kokumin-Keizai Zasshi",
  ["kokuritsu iyakuhin shokuhin eisei kenkyusho hokoku"] = "Kokuritsu Iyakuhin Shokuhin Eisei Kenkyūjo hōkoku = Bulletin of National Institute of Health Sciences",
  ["kokuritsu kagaku hakubutsukan senpo"] = "Kokuritsu Kagaku Hakubutsukan senpō : Memoirs of the National Science Museum. Kokuritsu Kagaku Hakubutsukan (Japan)",
  ["kokuritsu kogai kenkyusho kenkyu hokoku"] = "Kokuritsu Kogai Kenkyusho Tokubetsu Kenkyu Hokoku",
  ["kokusai kyoryoku kenkyushi"] = "Kokusai kyōryoku kenkyūshi",
  ["kokusaigaku revyu"] = "Kokusaigaku revyū = Obirin review of international studies",
  ["kokyu to junkan"] = "Kokyu to junkan. Respiration & circulation",
  ["kokyuki shinryo"] = "[Kokyūki shinryō] Clinics of respiratory organs",
  ["koleopterol rundsch"] = "Koleopterologische Rundschau",
  ["kolleg philos"] = "Kolleg Philosophie",
  ["kolloid z"] = "Kolloid Zeitschrift",
  ["kolloid z z polym"] = "Kolloid Zeitschrift & Zeitschrift fur Polymere",
  ["kolloidchem beih"] = "Kolloidchemische Beihefte",
  ["kolloidn zh"] = "Kolloidnyi Zhurnal",
  ["koln med beitr"] = "Kölner medizinhistorische Beiträge",
  ["koln z soziol sozialpsych"] = "Kölner Zeitschrift für Soziologie und Sozialpsychologie",
  ["kolner z soz sozialpsychol (aufl)"] = "Kolner Zeitschrift fur Soziologie und Sozialpsychologie",
  ["kolner z soz sozpsychol"] = "Kölner Zeitschrift für Soziologie und Sozialpsychologie",
  ["kompass nutr diet"] = "Kompass Nutrition & Dietetics",
  ["kompet-zent holz"] = "Kompetenz-Zentrum Holz",
  ["kompleksn ispol’z miner syr’ya"] = "Kompleksnoe Ispol’zovanie Mineral’nogo Syr’ya",
  ["komun mazur warm"] = "Komunikaty mazursko-warmińskie",
  ["kona powder part j"] = "Kona Powder and Particle Journal",
  ["kong dansk vidensk"] = "Kongelige Danske Videnskabernes Selskab: Matematisk-fysiske Meddelelser",
  ["kongressbd dtsch ges chir kongr"] = "Kongressband. Deutsche Gesellschaft für Chirurgie. Kongress",
  ["konink nederl akad wetensch verh afd natuurk eerste reeks"] = "Koninklijke Nederlandse Akademie van Wetenschappen",
  ["konuralp j math"] = "Konuralp Journal of Mathematics",
  ["koord khim"] = "Koordinatsionnaya Khimiya (Russian Journal of Coordination Chemistry)",
  ["kora ber"] = "KORA Bericht",
  ["korea (south)"] = "Korea & world affairs",
  ["korea j"] = "Korea journal",
  ["korea j popul dev"] = "Korea journal of population and development",
  ["korea-aust rheol j"] = "Korea-Australia Rheology Journal",
  ["korean ann math"] = "Korean Annals of Mathematics",
  ["korean circ j"] = "Korean circulation journal",
  ["korean diabetes j"] = "Korean diabetes journal",
  ["korean econ rev"] = "Korean economic review",
  ["korean j anesthesiol"] = "Korean journal of anesthesiology",
  ["korean j appl statist"] = "Korean Journal of Applied Statistics",
  ["korean j audiol"] = "Korean journal of audiology",
  ["korean j biochem"] = "The Korean journal of biochemistry : official journal of the Korean Biochemical Society",
  ["korean j biol sci"] = "Korean journal of biological sciences",
  ["korean j chem eng"] = "Korean Journal of Chemical Engineering",
  ["korean j clin pathol"] = "Taehan Imsang Pyŏngni Hakhoe chi = Korean journal of clinical pathology",
  ["korean j comput appl math"] = "The Korean Journal of Computational & Applied Mathematics",
  ["korean j crit care med"] = "Korean journal of critical care medicine",
  ["korean j def anal"] = "Korean Journal Of Defense Analysis",
  ["korean j entomology"] = "Korean journal of entomology",
  ["korean j environ agric"] = "Korean Journal of Environmental Agriculture",
  ["korean j fam med"] = "Korean journal of family medicine",
  ["korean j food sci anim resour"] = "Korean journal for food science of animal resources",
  ["korean j gastroenterol"] = "The Korean journal of gastroenterology = Taehan Sohwagi Hakhoe chi",
  ["korean j genet"] = "The Korean journal of genetics",
  ["korean j hematol"] = "The Korean journal of hematology",
  ["korean j hepatobiliary pancreat surg"] = "Korean journal of hepato-biliary-pancreatic surgery",
  ["korean j hepatol"] = "The Korean journal of hepatology",
  ["korean j intern med"] = "The Korean journal of internal medicine",
  ["korean j lab med"] = "The Korean journal of laboratory medicine",
  ["korean j malacol"] = "The Korean journal of malacology",
  ["korean j mater res"] = "Korean Journal of Materials Research",
  ["korean j math"] = "The Korean Journal of Mathematics",
  ["korean j med educ"] = "Korean journal of medical education",
  ["korean j met mater"] = "Korean Journal of Metals and Materials",
  ["korean j neurotrauma"] = "Korean journal of neurotrauma",
  ["korean j ophthalmol"] = "Korean journal of ophthalmology : KJO",
  ["korean j orthod"] = "Korean journal of orthodontics",
  ["korean j pain"] = "The Korean journal of pain",
  ["korean j parasitol"] = "The Korean journal of parasitology",
  ["korean j pathol"] = "Korean journal of pathology",
  ["korean j pediatr"] = "Korean journal of pediatrics",
  ["korean j physiol pharmacol"] = "The Korean journal of physiology & pharmacology : official journal of the Korean Physiological Society and the Korean Society of Pharmacology",
  ["korean j radiol"] = "Korean journal of radiology",
  ["korean j sericult sci"] = "The Korean journal of sericultural science",
  ["korean j soil sci fert"] = "Korean Journal of Soil Science and Fertilizer",
  ["korean j spine"] = "Korean Journal of Spine",
  ["korean j syst zool"] = "Korean journal of systematic zoology",
  ["korean j thorac cardiovasc surg"] = "The Korean journal of thoracic and cardiovascular surgery",
  ["korean j urol"] = "Korean journal of urology",
  ["korean j vet res"] = "Korean journal of veterinary research",
  ["korean journal of animal science|hanguk chuksan hakhoe chi"] = "Han’guk Ch’uksan Hakhoe chi",
  ["korean journal of applied entomology|hanguk ungyong konchung hakhoe chi"] = "Han’guk Ungyong Konch’ung Hakhoe chi",
  ["korean journal of clinical microbiology|taehan imsang misaengmul hakhoe chi"] = "Taehan Imsang Misaengmul Hakhoe chi",
  ["korean journal of clinical pathology|korean j clin pathol"] = "Taehan Imsang Pyongni Hakhoe chi",
  ["korean journal of medicinal crop science|hanguk yakyong changmul hakhoe chi"] = "Han’guk Yakyong Changmul Hakhoe chi",
  ["korean journal of microbiology and biotechnology|hanguk misaengmul saengmyong konghakhoe chi"] = "Han’guk Misaengmul, Saengmyong Konghakhoe chi",
  ["korean journal of plant biotechnology|sigmul saengmyeong gong haghoeji"] = "Sigmul saengmyeong gong haghoeji",
  ["korean journal of plant taxonomy|sigmul bunryu hag-hoeji"] = "Sigmul bunryu hag-hoeji",
  ["korean korean am stud bull"] = "Korean and Korean-American studies bulletin",
  ["koryo insam hakhoe chi"] = "Koryŏ Insam Hakhoe chi",
  ["kosei kagaku jpn j public health"] = "[Kōsei kagaku] The Japanese journal of public health",
  ["koshu eisei in kenkyu hokoku"] = "Kōshū Eiseiin kenkyū hōkoku. Bulletin of the Institute of Public Health",
  ["kosm biol aviakosm med"] = "Kosmicheskaia biologiia i aviakosmicheskaia meditsina",
  ["kosm biol med"] = "Kosmicheskaia biologiia i meditsina",
  ["kosm ser a biol"] = "Kosmos. Seria A, Biologia",
  ["kou qiang yi xue yan jui"] = "Kou qiang yi xue yan jiu = Journal of oral science research",
  ["kovove mater"] = "Kovove Materialy Metallic Materials",
  ["kozh-obuvn prom-st"] = "Kozhevenno-Obuvnaya Promyshlennost",
  ["kragujevac j math"] = "Kragujevac Journal of Mathematics",
  ["krankenhaushyg infektionsverhut"] = "Krankenhaus-Hygiene + Infektionsverhütung",
  ["krankenpfl j"] = "Krankenpflege Journal",
  ["krankenpfl soins infirm"] = "Krankenpflege. Soins infirmiers",
  ["krankenpflege (frankf)"] = "Krankenpflege (Frankfurt am Main, Germany)",
  ["kratk soobshch fiz"] = "Kratkie Soobshcheniya po Fizike",
  ["krebsforsch krebsbekampf"] = "Krebsforschung und Krebsbekampfung",
  ["kredit kapital"] = "Kredit und Kapital",
  ["krim forensische wiss"] = "Kriminalistik und Forensische Wissenschaft",
  ["krim z eksp klin med"] = "Krimskiĭ zhurnal ėksperimentalʹnoĭ i klinicheskoĭ medit︠s︡iny = Krymsʹkyĭ z︠h︡urnal eksperymentalʹnoï ta klinichnoï medyt︠s︡yny = Crimean journal of experimental and clinical medicine",
  ["krist tech"] = "Kristall und Technik",
  ["kristallografiya [sov phys crystallogr]"] = "Kristallografiya [Soviet Physics–Crystallography]",
  ["krit jahrb philos"] = "Kritisches Jahrbuch der Philosophie",
  ["kroc found ser"] = "Kroc Foundation series",
  ["ksbb j"] = "KSBB journal",
  ["ksce j civ eng"] = "KSCE Journal of Civil Engineering",
  ["ksii trans internet inf syst"] = "KSII Transactions on Internet and Information Systems",
  ["ku chi chui tung wu yu ku jen lei"] = "[Ku chi chui tung wu yü ku jên lei] : Vertebrata Palasiatica",
  ["kuang pu hseuh yu kuang pu fen hsi"] = "Kuang Pu Hseuh Yu Kuang Pu Fen Hsi Spectroscopy and Spectral Analysis (Pei-Ching",
  ["kula kula"] = "Kula kula",
  ["kulak burun bogaz ihtis derg"] = "Kulak burun boğaz ihtisas dergisi : KBB = Journal of ear, nose, and throat",
  ["kult fiz"] = "Kultura fizyczna",
  ["kult spolecz"] = "Kultura i społeczeństwo",
  ["kult tech"] = "Kultur & Technik : Zeitschrift des Deutschen Museums München",
  ["kult zahnheilkd einzeldarst"] = "Kulturgeschichte der Zahnheilkunde in Einzeldarstellungen",
  ["kumamoto igakkai zasshi"] = "Kumamoto Igakkai zasshi. The Journal of the Kumamoto Medical Society",
  ["kumamoto j math"] = "Kumamoto Journal of Mathematics",
  ["kumamoto med j"] = "The Kumamoto medical journal",
  ["kun chong fen lei xue bao"] = "Kun chong fen lei xue bao = Entomotaxonomia",
  ["kun chong xue bao"] = "Kun chong xue bao. Acta entomologica Sinica",
  ["kun chong zhi shi"] = "Kun chong zhi shi = Kunchong zhishi",
  ["kunchong zhishi / [zhongguo kun chong xue hui]|kun chong zhi shi"] = "Kun chong zhi shi",
  ["kunst orient"] = "Kunst des Orients",
  ["kunst ther"] = "Kunst & Therapie",
  ["kunstliche intell (oldenbourg)"] = "Künstliche intelligenz",
  ["kupr spoud"] = "Kypriakai spoudai : deltion tēs Hetaireias Kypriakōn Spoudōn",
  ["kurinikaru sutadi"] = "Kurinikaru sutadī = Clinical study",
  ["kurume med j"] = "The Kurume medical journal",
  ["kusuri no chishiki"] = "[Kusuri no chishiki] Pharmaceutical review",
  ["kuwait j sci"] = "Kuwait Journal of Science",
  ["kuwait j sci eng"] = "Kuwait Journal of Science & Engineering",
  ["kuwait med j"] = "The Kuwait medical journal : KMJ : the official journal of the Kuwait Medical Association",
  ["kvant elektron (moscow) [sov j quantum electron]"] = "Kvantovaya Elektronika (Moscow) [Soviet Journal of Quantum Electronics]",
  ["kvantovaya elektron (moscow)"] = "Kvantovaya Elektronika (Moscow)",
  ["kvartalsskr sven nationalforen hjart lungsjukd"] = "Kvartalsskrift. Svenska nationalföreningen mot hjärt- och lungsjukdomar",
  ["kwansei gakuin univ nat sci rev"] = "Kwansei Gakuin University",
  ["kwart hist"] = "Kwartalnik historyczny",
  ["kwart hist kult mater"] = "Kwartalnik historii kultury materialnej",
  ["kwart hist nauk tech"] = "Kwartalnik Historii Nauki i Techniki",
  ["kwart hist nauki tech"] = "Kwartalnik historii nauki i techniki : Kwartal'nyĭ zhurnal istorii nauki i tekhniki -",
  ["kwart hist ruchu zaw"] = "Kwartalnik historii ruchu zawodowego",
  ["kwart hist zydow"] = "Kwartalnik historii żydów : Jewish history quarterly",
  ["ky bench bar"] = "Kentucky bench & bar",
  ["ky dent j"] = "Kentucky dental journal",
  ["ky folkl rec"] = "Kentucky folklore record",
  ["ky hosp mag"] = "Kentucky hospitals magazine",
  ["ky j commun"] = "Kentucky journal of communication",
  ["ky law j"] = "Kentucky law journal (Lexington, Ky.)",
  ["ky med j"] = "Kentucky medical journal",
  ["ky nurse"] = "Kentucky nurse",
  ["ky nurses assoc newsl"] = "Kentucky Nurses Association Newsletter",
  ["ky rep"] = "KY [reports]. U.S. Atomic Energy Commission",
  ["kybernetika (prague)"] = "Kybernetika",
  ["kybernetika (prague) suppl"] = "Kybernetika",
  ["kyklos (oxford)"] = "Kyklos (Online)",
  ["kyobu geka"] = "Kyobu geka. The Japanese journal of thoracic surgery",
  ["kyorin igakkai zasshi"] = "Kyōrin Igakkai zasshi. Journal of the Kyorin Medical Society",
  ["kyoto daigaku kekkaku kenkyusho kiyo"] = "Kiyo. [Reports]. Kyōto Daigaku. Kekkaku Kenkyūjo",
  ["kyoto daigaku kokukagaku kiyo"] = "Kyoto Daigaku kokukagaku kiyo. Bulletin of stomatology, Kyoto University",
  ["kyoto furitsu ika daigaku zasshi"] = "[Zasshi] Journal. Kyoto Furitsu Ika Daigaku",
  ["kyoto j math"] = "Kyoto Journal of Mathematics",
  ["kyoto univ econ rev"] = "Kyoto University Economic Review",
  ["kyrkohist arsskr"] = "Kyrkohistorisk årsskrift",
  ["kyungpook math j"] = "Kyungpook Mathematical Journal",
  ["kyushu j math"] = "Kyushu Journal of Mathematics",
  ["kyushu j med sci"] = "Kyushu journal of medical science",
  ["la hist"] = "Louisiana history",
  ["la law rev"] = "Louisiana law review",
  ["la mat per il 3+2"] = "La Matematica per il 3+2",
  ["la rep"] = "LA [reports]. U.S. Atomic Energy Commission",
  ["la stud"] = "Louisiana studies",
  ["lab anim"] = "Laboratory animals",
  ["lab anim (ny)"] = "Lab animal",
  ["lab anim care"] = "Laboratory animal care",
  ["lab anim res"] = "Laboratory animal research",
  ["lab anim sci"] = "Laboratory animal science",
  ["lab animal sci prof"] = "Laboratory animal science professional",
  ["lab autom inf manage"] = "Laboratory Automation and Information Management",
  ["lab chip"] = "Lab on a chip",
  ["lab clin"] = "Laboratorio clinico",
  ["lab delo"] = "Laboratornoe delo",
  ["lab dig"] = "The Laboratory digest",
  ["lab econ"] = "Labour Economics",
  ["lab hematol"] = "Laboratory hematology : official publication of the International Society for Laboratory Hematology",
  ["lab hist"] = "Labor History",
  ["lab invest"] = "Laboratory investigation; a journal of technical methods and pathology",
  ["lab matters"] = "Lab matters : analysis, answers, actions",
  ["lab med"] = "Laboratory medicine",
  ["lab phonol"] = "Laboratory phonology",
  ["lab pract"] = "Laboratory practice",
  ["lab res methods biol med"] = "Laboratory and research methods in biology and medicine",
  ["lab rob autom"] = "Laboratory Robotics and Automation",
  ["lab world"] = "Lab world",
  ["lab world mag"] = "Lab world magazine",
  ["labciencia not tec lab"] = "LabCiencia con Noticias Tecnicas del Laboratorio",
  ["labor (durh)"] = "Labor (Durham, N.C. : Online)",
  ["labor hist"] = "Labor history",
  ["labor law j"] = "Labor law journal",
  ["labor lawyer"] = "The Labor lawyer",
  ["labor stud j"] = "Labor studies journal",
  ["labour (rome)"] = "Labour (Rome, Italy : Online)",
  ["labour cap soc"] = "Labour capital and society. Travail capital et société",
  ["labour econ"] = "Labour economics",
  ["labour hist"] = "Labour history",
  ["labour hist rev"] = "Labour history review",
  ["labour soc"] = "Labour and society",
  ["lactarius (jaén)"] = "Lactarius : boletín de la Asociación Micológica",
  ["lactation rev"] = "The Lactation review",
  ["ladies home j"] = "The Ladies' home journal",
  ["lahey clin bull"] = "The Lahey Clinic bulletin",
  ["lahey clin found bull"] = "Lahey Clinic Foundation bulletin",
  ["lahore j econ"] = "Lahore Journal of Economics",
  ["laimburg j"] = "Laimburg-Journal",
  ["lake line"] = "Lake line",
  ["lake reserv manag"] = "Lake and reservoir management",
  ["lake reservoir manage"] = "Lake and Reservoir Management",
  ["lakes & reservoirs: science, policy and management for sustainable use"] = "Lakes & Reservoirs: Science, Policy and Management for Sustainable Use",
  ["lakes reserv"] = "Lakes & reservoirs : research and management",
  ["lakes reservoirs res manage"] = "Lakes & Reservoirs: Research and Management",
  ["lakokras mater ikh primen"] = "Lakokrasochnye Materialy i Ikh Primenenie",
  ["lancet child adolesc health"] = "The Lancet. Child & adolescent health",
  ["lancet diabetes endocrinol"] = "The lancet. Diabetes & endocrinology",
  ["lancet digit health"] = "The Lancet. Digital health",
  ["lancet gastroenterol hepatol"] = "The lancet. Gastroenterology & hepatology",
  ["lancet glob health"] = "The Lancet. Global health",
  ["lancet haematol"] = "The Lancet. Haematology",
  ["lancet healthy longev"] = "The Lancet. Healthy longevity",
  ["lancet hiv"] = "The lancet. HIV",
  ["lancet infect dis"] = "The Lancet. Infectious diseases",
  ["lancet microbe"] = "The Lancet. Microbe",
  ["lancet neurol"] = "The Lancet. Neurology",
  ["lancet oncol"] = "The Lancet. Oncology",
  ["lancet planet health"] = "The Lancet. Planetary health",
  ["lancet planet heath"] = "Lancet Planetary Health",
  ["lancet psychiatry"] = "The lancet. Psychiatry",
  ["lancet public health"] = "The Lancet. Public health",
  ["lancet reg health - southeast asia"] = "Lancet Regional Health - Southeast Asia",
  ["lancet reg health west pac"] = "The Lancet Regional Health. Western Pacific",
  ["lancet respir med"] = "The Lancet. Respiratory medicine",
  ["lancet rheumatol"] = "The Lancet. Rheumatology",
  ["land (basel)"] = "Land",
  ["land aalst"] = "Het Land van Aalst",
  ["land degrad dev"] = "Land Degradation and Development",
  ["land econ"] = "Land economics",
  ["land use policy"] = "Land use policy",
  ["land water law rev"] = "Land and water law review",
  ["landbauforschung (braunschw)"] = "Landbauforschung",
  ["landeshydrol -geol, mitt"] = "Landeshydrologie und -geologie, Mitteilung",
  ["landsc archit front"] = "Landscape architecture frontiers",
  ["landsc ecol"] = "Landscape ecology",
  ["landsc hist"] = "Landscape history : journal of the Society for Landscape Studies",
  ["landsc j"] = "Landscape journal",
  ["landsc log"] = "Landscapes in Logic",
  ["landsc res"] = "Landscape research",
  ["landsc res rec"] = "Landscape research record",
  ["landsc urban plan"] = "Landscape and urban planning",
  ["landscape ecol"] = "Landscape Ecology",
  ["landscape ecol eng"] = "Landscape and Ecological Engineering",
  ["landscape hist"] = "Landscape History",
  ["landscape res"] = "Landscape Research",
  ["landscape urban plann"] = "Landscape and Urban Planning",
  ["landsch aargau"] = "Landschaft Aargau",
  ["landschentwickl umweltforsch"] = "Landschaftsentwicklung und Umweltforschung",
  ["landschgenes landschökol"] = "Landschaftsgenese und Landschaftsökologie",
  ["landschpfl vegkd"] = "Landschaftspflege und Vegetationskunde",
  ["landschökol umweltforsch"] = "Landschaftsökologie und Umweltforschung",
  ["landwirtsch jahrb schweiz"] = "Landwirtschaftliches Jahrbuch der Schweiz",
  ["lang acquis"] = "Language acquisition",
  ["lang assess q"] = "Language assessment quarterly",
  ["lang cogn"] = "Language and cognition",
  ["lang cogn neurosci"] = "Language, cognition and neuroscience",
  ["lang cogn process"] = "Language and cognitive processes",
  ["lang commun"] = "Language & communication",
  ["lang commun disord"] = "Language and Communication Disorders",
  ["lang ed soc"] = "Language, Education and Society",
  ["lang learn"] = "Language learning",
  ["lang learn dev"] = "Language learning and development : the official journal of the Society for Language Development",
  ["lang learn technol"] = "Language learning & technology : LL & T",
  ["lang linguist (taipei)"] = "Language and linguistics",
  ["lang linguist compass"] = "Language and linguistics compass",
  ["lang lit (harlow)"] = "Language and literature (Harlow, England)",
  ["lang policy"] = "Language policy",
  ["lang res"] = "Language research",
  ["lang resour eval"] = "Language resources and evaluation",
  ["lang sci"] = "Language sciences (Oxford, England)",
  ["lang soc"] = "Language In Society",
  ["lang sociocult theory"] = "Language and sociocultural theory",
  ["lang speech"] = "Language and speech",
  ["lang speech commun"] = "Language, Speech, and Communication",
  ["lang speech hear serv sch"] = "Language, speech, and hearing services in schools",
  ["lang var change"] = "Language variation and change",
  ["langenbecks arch chir"] = "Langenbecks Archiv fur Chirurgie",
  ["langenbecks arch chir suppl ii verh dtsch ges chir"] = "Langenbecks Archiv für Chirurgie. Supplement II, Verhandlungen der Deutschen Gesellschaft für Chirurgie. Deutsche Gesellschaft für Chirurgie. Kongress",
  ["langenbecks arch chir suppl kongressbd"] = "Langenbecks Archiv für Chirurgie. Supplement. Kongressband. Deutsche Gesellschaft für Chirurgie. Kongress",
  ["langenbecks arch surg"] = "Langenbeck's archives of surgery",
  ["language (baltim)"] = "Language",
  ["languages (basel)"] = "Languages (Basel, Switzerland)",
  ["languedoc med"] = "Languedoc médical",
  ["laparosc endosc robot surg"] = "Laparoscopic, endoscopic, and robotic surgery",
  ["laparosc surg"] = "Laparoscopic surgery",
  ["lar do med"] = "O Lar do médico",
  ["larc med"] = "LARC médical",
  ["large anim rev"] = "Large Animal Review",
  ["large scale annot biomed data export label synth hardw aware learn med imaging comput assist interv (2019)"] = "Large-Scale Annotation of Biomedical Data and Expert Label Synthesis and Hardware Aware Learning for Medical Imaging and Computer Assisted Intervention : International Workshops, LABELS 2019, HAL-MICCAI 2019, and CuRIOUS 2019, held in conjunction with MICCAI 2019, Shenzhen, China, October 13 and 17, 2019, Proceedings. LABELS (Workshop) (4th : 2019 : Shenzhen Shi, China)",
  ["large scale assess educ"] = "Large-scale assessments in education",
  ["laryngol rhinol otol (stuttg)"] = "Laryngologie, Rhinologie, Otologie",
  ["laryngoscope investig otolaryngol"] = "Laryngoscope investigative otolaryngology",
  ["laser chem"] = "Laser Chemistry",
  ["laser focus world"] = "Laser focus world",
  ["laser nurs"] = "Laser nursing",
  ["laser part beams"] = "Laser and Particle Beams",
  ["laser photon rev"] = "Laser & photonics reviews",
  ["laser photonics rev"] = "Laser & Photonics Reviews",
  ["laser phys"] = "Laser Physics",
  ["laser phys lett"] = "Laser Physics Letters",
  ["laser tech j"] = "Laser Technik Journal",
  ["laser ther"] = "Laser therapy",
  ["lasers dent sci"] = "Lasers in dental science",
  ["lasers eng"] = "Lasers in Engineering",
  ["lasers life sci"] = "Lasers in the life sciences",
  ["lasers manuf mater process"] = "Lasers in Manufacturing and Materials Processing",
  ["lasers med sci"] = "Lasers in medical science",
  ["lasers surg med"] = "Lasers in surgery and medicine",
  ["lasers surg med suppl"] = "Lasers in surgery and medicine. Supplement",
  ["lat am"] = "Latinskai︠a︡ Amerika = América Latina",
  ["lat am appl res"] = "Latin America Applied Research",
  ["lat am econ rev"] = "Latin American economic review",
  ["lat am j aquat mamm"] = "The Latin American journal of aquatic mammals : (LAJAM)",
  ["lat am j aquat res"] = "Latin American journal of aquatic research",
  ["lat am j econ"] = "Latin American journal of economics",
  ["lat am j pharm"] = "Latin American Journal of Pharmacy",
  ["lat am j solids struct"] = "Latin American Journal of Solids and Structures",
  ["lat am perspect"] = "Latin American perspectives",
  ["lat am policy"] = "Latin American policy",
  ["lat am policy j"] = "Latin American Policy Journal",
  ["lat am polit soc"] = "Latin American politics and society",
  ["lat am popul hist bull"] = "Latin American population history bulletin",
  ["lat am popul hist news"] = "Latin American population history newsletter",
  ["lat am res rev"] = "Latin American research review",
  ["lat amer math ser"] = "Latin American Mathematics Series",
  ["lat amer math ser ufscar subser"] = "Latin American Mathematics Series—UFSCar Subseries",
  ["lat stud"] = "Latino studies",
  ["late imp china"] = "Late imperial China = Chʻing shih wen tʻi",
  ["latin amer politics society"] = "Latin American Politics and Society",
  ["latv j phys tech sci"] = "Latvian Journal of Physics and Technical Sciences",
  ["latv kim z"] = "Latvijas Kimijas Zurnals",
  ["latv mat ezhegodnik"] = "Latviy skaya Akademiya Nauk",
  ["latv zinat akad vestis a humanit zinat"] = "Latvijas Zinātn̦u akadēmijas vēstis. A dal̦a, Sociālās un humanitārās zinātnes = Proceedings of the Latvian Academy of Sciences. Section A",
  ["laufener seminbeitr"] = "Laufener Seminarbeiträge",
  ["laund news"] = "Laundry news",
  ["laundry dry clean j can"] = "Laundry and Dry Cleaning Journal of Canada",
  ["lav ist anat istol patol univ studi perugia"] = "Lavori dell'Istituto di anatomia e istologia patologica, Università degli studi di Perugia",
  ["lav um"] = "Lavoro umano",
  ["laval med"] = "Laval médical",
  ["law contemp prob"] = "Law and Contemporary Problems",
  ["law contemp probl"] = "Law and contemporary problems",
  ["law context"] = "Law in context (Bundoora, Vic.)",
  ["law enforc exec forum"] = "Law enforcement executive forum",
  ["law file"] = "Law file",
  ["law hum behav"] = "Law and human behavior",
  ["law hum genome rev"] = "Law and the Human Genome Review",
  ["law inequal"] = "Law & inequality",
  ["law innov technol"] = "Law, innovation and technology",
  ["law inst j"] = "Law Institute journal : the official organ of the Law Institute of Victoria",
  ["law justice"] = "Law & justice",
  ["law lib"] = "Law & liberty",
  ["law libr j"] = "Law library journal",
  ["law med health care"] = "Law, medicine & health care : a publication of the American Society of Law & Medicine",
  ["law med news"] = "Law-medicine news",
  ["law ment health"] = "Law and mental health",
  ["law philos"] = "Law and philosophy",
  ["law policy"] = "Law & policy",
  ["law policy int bus"] = "Law and policy in international business",
  ["law policy q"] = "Law & policy quarterly",
  ["law probab risk"] = "Law, probability & risk : a journal of reasoning under uncertainty",
  ["law psychol rev"] = "Law and psychology review",
  ["law q rev"] = "The Law quarterly review",
  ["law rep house lords"] = "The Law reports. House of Lords, and Judicial Committee of the Privy Council, and peerage cases",
  ["law rev"] = "Law review",
  ["law soc inq"] = "Law & social inquiry : journal of the American Bar Foundation",
  ["law soc rev"] = "Law & society review",
  ["law teach"] = "The Law teacher",
  ["laws ill"] = "Laws. Illinois",
  ["laws p r annot p r"] = "Laws of Puerto Rico annotated. Puerto Rico",
  ["laws state n m n m"] = "Laws of the state of New Mexico. New Mexico",
  ["laws state or or"] = "Laws of the state of Oregon. Oregon",
  ["lc gc"] = "LC GC : magazine of liquid and gas chromatography",
  ["lc gc eur"] = "LC GC Europe",
  ["lc gc n am"] = "LC GC North America",
  ["lc gc north am"] = "LC GC North America",
  ["lcgc north am"] = "LCGC North America",
  ["lda j"] = "LDA journal",
  ["ldi issue brief"] = "LDI issue brief",
  ["lead edge"] = "Leading edge (Tulsa, Okla. : Online)",
  ["leadersh health serv"] = "Leadership in health services = Leadership dans les services de santé",
  ["leadersh health serv (bradf engl)"] = "Leadership in health services (Bradford, England)",
  ["leadersh manage eng"] = "Leadership and Management in Engineering",
  ["leadersh policy sch"] = "Leadership and policy in schools",
  ["leadersh q"] = "The leadership quarterly",
  ["leadership (lond)"] = "Leadership (London)",
  ["leading edge"] = "Leading Edge",
  ["leafl agric res organ, div for"] = "Leaflet Agricultural Research Organization, Division of Forestry",
  ["league exch"] = "League exchange",
  ["league lines"] = "League lines",
  ["lean construct j"] = "Lean Construction Journal",
  ["learn behav"] = "Learning & behavior",
  ["learn cult soc interact"] = "Learning, culture and social interaction",
  ["learn curve"] = "Learning Curve",
  ["learn disabil"] = "Learning disabilities (Weston, Mass.)",
  ["learn disabil (pittsbg)"] = "Learning disabilities (Pittsburgh, Pa.)",
  ["learn disabil q"] = "Learning disability quarterly : journal of the Division for Children with Learning Disabilities",
  ["learn disabil res pract"] = "Learning disabilities research & practice : a publication of the Division for Learning Disabilities, Council for Exceptional Children",
  ["learn doing soc cogn comput perspect"] = "Learning in Doing: Social, Cognitive, and Computational Perspectives",
  ["learn environ res"] = "Learning environments research",
  ["learn health syst"] = "Learning health systems",
  ["learn individ differ"] = "Learning and individual differences",
  ["learn instr"] = "Learning and instruction",
  ["learn landsc"] = "LEARNing landscapes",
  ["learn mater biosci"] = "Learning Materials in Biosciences",
  ["learn media technol"] = "Learning, media and technology",
  ["learn mem"] = "Learning & memory (Cold Spring Harbor, N.Y.)",
  ["learn motiv"] = "Learning and motivation",
  ["learn percept"] = "Learning & perception",
  ["learn publ"] = "Learned publishing : journal of the Association of Learned and Professional Society Publishers",
  ["learned publ"] = "Learned Publishing",
  ["leban pharm j"] = "The Lebanese pharmaceutical journal. La Revue pharmaceutique libanaise",
  ["lebens kunst cademario nachr"] = "Lebens-Kunst; Cademario-Nachrichten",
  ["lebensm wiss technol"] = "Lebensmittel-Wissenschaft und Technologie",
  ["lebensm-wiss technol"] = "Lebensmittel-Wissenschaft und -Technologie",
  ["lebenson wiss technol"] = "Lebensmittel-Wissenschaft + [i.e. und] Technologie. Food science + technology. Science + technologie alimentaire",
  ["lebensversicher med"] = "Lebensversicherungs Medizin",
  ["leber magen darm"] = "Leber, Magen, Darm",
  ["lebniz int proc inform"] = "LIPIcs : Leibniz international proceedings in informatics",
  ["lect mat"] = "Lecturas Matematicas",
  ["lect notes appl comput mech"] = "Lecture Notes in Applied and Computational Mechanics",
  ["lect notes appl math mech"] = "Lecture Notes in Applied Mathematics and Mechanics",
  ["lect notes appl numer harmon anal"] = "Lecture Notes in Applied and Numerical Harmonic Analysis",
  ["lect notes bioeng"] = "Lecture Notes in Bioengineering",
  ["lect notes biomath"] = "Lecture Notes in Biomathematics",
  ["lect notes bus inf process"] = "Lecture Notes in Business Information Processing",
  ["lect notes chem"] = "Lecture Notes in Chemistry",
  ["lect notes civ eng"] = "Lecture Notes in Civil Engineering",
  ["lect notes comput sci"] = "Lecture Notes in Computer Science",
  ["lect notes comput sci eng"] = "Lecture Notes in Computational Science and Engineering",
  ["lect notes comput vis biomech"] = "Lecture Notes in Computational Vision and Biomechanics",
  ["lect notes control inf sci"] = "Lecture Notes in Control and Information Sciences",
  ["lect notes control inf sci proc"] = "Lecture Notes in Control and Information Sciences. Proceedings",
  ["lect notes data eng commun technol"] = "Lecture Notes on Data Engineering and Communications Technologies",
  ["lect notes earth sci"] = "Lecture Notes in Earth Sciences",
  ["lect notes earth syst sci"] = "Lecture Notes in Earth System Sciences",
  ["lect notes electr eng"] = "Lecture Notes in Electrical Engineering",
  ["lect notes energy"] = "Lecture Notes in Energy",
  ["lect notes geosyst math comput"] = "Lecture Notes in Geosystems Mathematics and Computing",
  ["lect notes in bioinform"] = "Lecture Notes in Bioinformatics",
  ["lect notes inst comput sci soc inform telecommun eng"] = "Lecture Notes of the Institute for Computer Sciences, Social Informatics and Telecommunications Engineering",
  ["lect notes intell transp infrastruct"] = "Lecture Notes in Intelligent Transportation and Infrastructure",
  ["lect notes log"] = "Lecture Notes in Logic",
  ["lect notes logist"] = "Lecture Notes in Logistics",
  ["lect notes math"] = "Lecture Notes in Mathematics",
  ["lect notes math fluid mech"] = "Lecture Notes in Mathematical Fluid Mechanics",
  ["lect notes math model life sci"] = "Lecture Notes on Mathematical Modelling in the Life Sciences",
  ["lect notes mech"] = "Lecture Notes in Mechanics",
  ["lect notes mech eng"] = "Lecture Notes in Mechanical Engineering",
  ["lect notes med inform"] = "Lecture notes in medical informatics",
  ["lect notes mobil"] = "Lecture Notes in Mobility",
  ["lect notes monogr ser"] = "Lecture notes-monograph series",
  ["lect notes morphog"] = "Lecture Notes in Morphogenesis",
  ["lect notes netw syst"] = "Lecture Notes in Networks and Systems",
  ["lect notes nonlinear anal"] = "Lecture Notes in Nonlinear Analysis",
  ["lect notes numer methods eng sci"] = "Lecture Notes on Numerical Methods in Engineering and Sciences",
  ["lect notes oper res"] = "Lecture Notes in Operations Research",
  ["lect notes phys"] = "Lecture Notes in Physics",
  ["lect notes phys monogr"] = "Lecture Notes in Physics Monographs",
  ["lect notes semin interdiscip mat"] = "Lecture Notes of Seminario Interdisciplinare di Matematica",
  ["lect notes ser inst math sci natl univ singap"] = "Lecture Notes Series. Institute for Mathematical Sciences. National University of Singapore",
  ["lect notes soc netw"] = "Lecture Notes in Social Networks",
  ["lect notes stat"] = "Lecture Notes in Statistics",
  ["lect notes stat proc"] = "Lecture Notes in Statistics—Proceedings",
  ["lect notes ticmi"] = "Lecture Notes of Tbilisi International Centre of Mathematics and Informatics",
  ["lect notes unione mat ital"] = "Lecture Notes of the Unione Matematica Italiana",
  ["lect pure appl math"] = "Lectures in Pure and Applied Mathematics",
  ["lect rev ser"] = "Lecture and review series. Naval Medical Research Institute (U.S.)",
  ["lect sci basis med"] = "Lectures on the scientific basis of medicine",
  ["lecturas econ"] = "Lecturas de Economia",
  ["lecture notes first braz school comput algebra"] = "Lecture Notes from the First Brazilian School on Computer Algebra",
  ["lecture notes in artif intell"] = "Lecture Notes in Artif. Intell.",
  ["lecture notes in biomath"] = "Lecture Notes in Biomathematics",
  ["lecture notes in chem"] = "Lecture Notes in Chemistry",
  ["lecture notes in comput sci"] = "Lecture Notes in Computer Science",
  ["lecture notes in control and inform sci"] = "Lecture Notes in Control and Information Sciences",
  ["lecture notes in econom and math systems"] = "Lecture Notes in Economics and Mathematical Systems",
  ["lecture notes in math"] = "Lecture Notes in Mathematics",
  ["lecture notes in phys"] = "Lecture Notes in Physics",
  ["lecture notes in phys new ser m monogr"] = "Lecture Notes in Physics",
  ["lecture notes in pure and appl math"] = "Lecture Notes in Pure and Applied Mathematics",
  ["lecture notes in statist"] = "Lecture Notes in Statistics",
  ["lecture notes logic"] = "Lecture Notes in Logic",
  ["lecture notes numer appl anal"] = "Lecture Notes in Numerical and Applied Analysis",
  ["lecture notes ser"] = "Lecture Notes Series",
  ["lecture notes ser comput"] = "Lecture Notes Series on Computing",
  ["lectures in appl math"] = "Lectures in Applied Mathematics",
  ["lectures math eth zurich"] = "Lectures in Mathematics ETH Zurich",
  ["lectures math eth zürich"] = "Lectures in Mathematics ETH Zürich",
  ["leech (wales)"] = "The Leech",
  ["leeds dent j"] = "Leeds Dental Journal",
  ["leg aspects med pract"] = "Legal aspects of medical practice",
  ["leg ethics"] = "Legal ethics (Oxford, England)",
  ["leg med"] = "Legal medicine",
  ["leg med (tokyo)"] = "Legal medicine (Tokyo, Japan)",
  ["leg med annu"] = "Legal medicine annual",
  ["leg med q"] = "Legal medical quarterly",
  ["leg stud"] = "Legal studies (Society of Public Teachers of Law)",
  ["leg stud (soc leg scholars)"] = "Legal studies (Society of Legal Scholars)",
  ["legal criminol psychol"] = "Legal and criminological psychology",
  ["legal med"] = "Legal Medicine",
  ["legal ref serv q"] = "Legal reference services quarterly",
  ["lege artis med"] = "Lege artis medicinae : új magyar orvosi hírmondó",
  ["legume genomics genet"] = "Legume genomics and genetics",
  ["legume res"] = "Legume Research",
  ["legume sci"] = "Legume Science",
  ["lehr- handb geld borse bank versicher"] = "Lehr- und Handbucher zu Geld, Borse, Bank und Versicherung",
  ["lehrbuch math"] = "Lehrbuch Mathematik",
  ["lehrbücher monogr didakt math"] = "Lehrbücher und Monographien zur Didaktik der Mathematik",
  ["leibniz oxford"] = "Leibniz from Oxford",
  ["leis sci"] = "Leisure sciences",
  ["leisure (waterloo)"] = "Leisure (Waterloo, Ont.)",
  ["leisure sci"] = "Leisure Science",
  ["leitfaden angew inform"] = "Leitfaden der Angewandten Informatik",
  ["leitfäden angew math mech"] = "Leitfäden der Angewandten Mathematik und Mechanik",
  ["leitfäden inform"] = "Leitfäden der Informatik",
  ["leitfäden monogr inform"] = "Leitfäden und Monographien der Informatik",
  ["leitz, mitt wiss tech"] = "Leitz, Mitteilungen für Wissenschaft und Technik",
  ["lek list"] = "Lékařské listy",
  ["lek obz"] = "Lekársky obzor",
  ["lek pr"] = "Lekárske práce",
  ["lek veda zahr"] = "Lékar̆ská vĕda v zahranic̆í",
  ["lek wojsk"] = "Lekarz wojskowy",
  ["lemur news"] = "Lemur news",
  ["leningrad gos univ uchen zap ser mat nauk"] = "Leningradskiy Gosudarstvennyy Ordena Lenina Universitet imeni A",
  ["lens eye toxic res"] = "Lens and eye toxicity research",
  ["lens res"] = "Lens research",
  ["leonardo (oxf)"] = "Leonardo (Oxford, England)",
  ["leonhardi euleri opera omnia ser 4a"] = "Leonhardi Euleri Opera Omnia, Series Quarta A: Commercium Epistolicum",
  ["lepid novae"] = "Lepidoptera novae",
  ["lepr india"] = "Leprosy in India",
  ["lepr rev"] = "Leprosy review",
  ["les nouv"] = "LES nouvelles. Licensing Executives Society (U.S.A.)",
  ["let hoje"] = "Letras de hoje",
  ["lett appl microbiol"] = "Letters in applied microbiology",
  ["lett biomath"] = "Letters in Biomathematics",
  ["lett drug des discov"] = "Letters in drug design & discovery",
  ["lett drug des discovery"] = "Letters in Drug Design & Discovery",
  ["lett ford found"] = "Ford Foundation letter. Ford Foundation",
  ["lett funct foods"] = "Letters in Functional Foods",
  ["lett health biol sci"] = "Letters in health & biological sciences",
  ["lett heat mass transfer"] = "Letters in Heat and Mass Transfer",
  ["lett high energy phys"] = "Letters in High Energy Physics",
  ["lett ital"] = "Lettere italiane",
  ["lett mat"] = "Lettera Matematica",
  ["lett mat pristem"] = "Lettera matematica pristem",
  ["lett mater"] = "Letters On Materials",
  ["lett math phys"] = "Letters in Mathematical Physics",
  ["lett nuovo cimento"] = "Lettere al Nuovo Cimento",
  ["lett org chem"] = "Letters in Organic Chemistry",
  ["lett orient class"] = "Lettres Orientales et Classiques",
  ["lett pept sci"] = "Letters in Peptide Science",
  ["lett spat resour sci"] = "Letters in spatial and resource sciences",
  ["lett spatial resour sci"] = "Letters in Spatial and Resource Sciences",
  ["leuk lymphoma"] = "Leukemia & lymphoma",
  ["leuk res"] = "Leukemia research",
  ["leuk res rep"] = "Leukemia research reports",
  ["leuk res treatment"] = "Leukemia research and treatment",
  ["leuk suppl"] = "Leukemia supplements",
  ["leuven notes math theoret phys ser b theoret particle phys"] = "Leuven Notes in Mathematical and Theoretical Physics",
  ["leuvense bijdr"] = "Leuvensche bijdragen",
  ["leveltari kozl"] = "Levéltári közlemények",
  ["leveltari sz"] = "Levéltári szemle",
  ["lex sci"] = "Lex et scientia",
  ["lexington books ser comput sci"] = "Lexington Books Series in Computer Science",
  ["lexis dc code dc"] = "LEXIS District of Columbia code. District of Columbia",
  ["leyte samar stud"] = "Leyte-Samar studies",
  ["lezioni lincee"] = "Lezioni Lincee",
  ["lgbt health"] = "LGBT health",
  ["lhb hydrosci j"] = "LHB Hydroscience Journal",
  ["li shi dang an"] = "Li shi dang an",
  ["li shi yu yan yan jiu suo ji kan"] = "Li shi yu yan yan jiu suo ji kan",
  ["liang you shi pin ke ji"] = "Liang you shi pin ke ji",
  ["lib congr acquis manuscr div"] = "Library of Congress acquisitions. Manuscript Division. Library of Congress. Manuscript Division",
  ["lib eng math"] = "Library of Engineering Mathematics",
  ["lib engrg math"] = "Library of Engineering Mathematics",
  ["lib life fam"] = "Liberty, life, and family",
  ["lib math"] = "Libertas mathematica",
  ["lib math (ns)"] = "Libertas Mathematica. New Series",
  ["lib sci slant doc"] = "Library science with a slant to documentation",
  ["liberian stud j"] = "Liberian studies journal",
  ["libertas math"] = "Libertas Mathematica",
  ["libr assoc rec"] = "Library Association record. Library Association",
  ["libr bull univ lond"] = "Education libraries bulletin",
  ["libr chron"] = "Library chronicle",
  ["libr chron univ tex austin"] = "The Library chronicle of the University of Texas at Austin",
  ["libr cult"] = "Libraries & culture",
  ["libr cult rec"] = "Libraries & the cultural record",
  ["libr hist"] = "Library history",
  ["libr inf sci res"] = "Library & information science research",
  ["libr j"] = "Library journal",
  ["libr q"] = "The Library quarterly",
  ["libr resour tech serv"] = "Library resources & technical services",
  ["libr resources tech serv"] = "Library Resources and Technical Services",
  ["libr rev (lond)"] = "Library review. Great Britain. Forestry Commission",
  ["libr technol rep"] = "Library technology reports",
  ["libr theor"] = "Library of Theoria",
  ["libr trends"] = "Library trends",
  ["library (lond)"] = "The Library",
  ["libyan econ bus rev"] = "Libyan Economic and Business Review",
  ["libyan j med"] = "The Libyan journal of medicine",
  ["lichenologist (lond)"] = "Lichenologist (London, England)",
  ["lick obs bull"] = "Lick Observatory Bulletin",
  ["liebigs ann chem"] = "Liebigs Annalen der Chemie",
  ["liebigs ann recl"] = "Liebigs Annalen - Recueil",
  ["liet fiz rinkinys"] = "Lietuvos Fizikos Rinkinys",
  ["liet istor metrast"] = "Lietuvos istorijos metraštis",
  ["liet mat rink"] = "Matematikos ir Informatikos Institutas",
  ["liet tautos praeit"] = "Lietuvių tautos praeitis. Lithuanian historical review",
  ["liet tsr aukst mokyklu mokslo darb"] = "Lietuvos TSR auks̆tu̧ju̧ mokyklu̧ mokslo darbai: Biologija",
  ["life (basel)"] = "Life (Basel, Switzerland)",
  ["life (jaipur)"] = "Life (Jaipur, India)",
  ["life cycle reliab saf eng"] = "Life Cycle Reliability and Safety Engineering",
  ["life excit biol"] = "Life: the excitement of biology",
  ["life extreme environ"] = "Life in Extreme Environments",
  ["life health"] = "Life and health",
  ["life learn univ fac life conf"] = "Life and learning. University Faculty for Life. Conference",
  ["life sci"] = "Life Sciences",
  ["life sci (1962)"] = "Life sciences (1962)",
  ["life sci adv"] = "Life science advances",
  ["life sci adv biochem"] = "Life science advances. Biochemistry : a journal of the Council of Scientific Research Integration",
  ["life sci adv exp clin endocrinol"] = "Life science advances. Experimental and clinical endocrinology : a journal of the Council of Scientific Research Integration",
  ["life sci adv genet"] = "Life science advances. Genetics : a journal of the Council of Scientific Research Integration",
  ["life sci adv plant physiol"] = "Life science advances. Plant physiology : a journal of the Council of Scientific Research Integration",
  ["life sci adv steroid biochem"] = "Life science advances. Steroid biochemistry : a journal of the Council of Scientific Research Integration",
  ["life sci alliance"] = "Life science alliance",
  ["life sci i"] = "Life sciences. Pt. 1: Physiology and pharmacology",
  ["life sci ii"] = "Life sciences. Pt. 2: Biochemistry, general and molecular biology",
  ["life sci j"] = "Life science journal",
  ["life sci soc policy"] = "Life sciences, society and policy",
  ["life sci space res"] = "Life sciences and space research",
  ["life sci space res (amst)"] = "Life sciences in space research",
  ["life science research / hunan normal university|sheng ming ke xue yan jiu"] = "Sheng ming ke xue yan jiu / Hunan shi fan da xue zhu ban",
  ["life span disabil"] = "Life span and disability",
  ["life support biosph sci"] = "Life support & biosphere science : international journal of earth space",
  ["life support syst"] = "Life support systems : the journal of the European Society for Artificial Organs",
  ["life threat behav"] = "Life-threatening behavior",
  ["lifecycle j"] = "Lifecycle Journal",
  ["lifelong learn adult years"] = "Lifelong learning, the adult years",
  ["lifestyle genom"] = "Lifestyle genomics",
  ["lifestyle genomics"] = "Lifestyle Genomics",
  ["lifetime data anal"] = "Lifetime data analysis",
  ["liga urug contra tuberc"] = "Liga Uruguaya contra la Tuberculosis",
  ["light j"] = "Lighting Journal",
  ["light met (warrendale, pa)"] = "Light Metals (Warrendale, Pennsylvania)",
  ["light res technol"] = "Lighting Research and Technology",
  ["light sci appl"] = "Light, science & applications",
  ["light: sci appl"] = "Light: Science & Applications",
  ["lightweight des"] = "Lightweight Design",
  ["lightweight des worldwide"] = "Lightweight Design Worldwide",
  ["lihua jianyan, huaxue fence"] = "Lihua Jianyan, Huaxue Fence",
  ["liiketaloudellinen aikak"] = "Liiketaloudellinen Aikakauskirja",
  ["lijec vjesn"] = "Lijecnicki Vjesnik",
  ["lik sprava"] = "Likarska Sprava",
  ["lille chir"] = "Lille chirurgical",
  ["lille med"] = "Lille médical : journal de la Faculté de médecine et de pharmacie de l'Université de Lille",
  ["limnol akt"] = "Limnologie Aktuell",
  ["limnol freshw biol"] = "Limnology and freshwater biology",
  ["limnol oceanogr"] = "Limnology and oceanography",
  ["limnol oceanogr bull"] = "Limnology and Oceanography Bulletin",
  ["limnol oceanogr e-lect"] = "Limnology and Oceanography e-Lectures",
  ["limnol oceanogr fluids environ"] = "Limnology and Oceanography: Fluids and Environments",
  ["limnol oceanogr lett"] = "Limnology and oceanography letters",
  ["limnol oceanogr methods"] = "Limnology and Oceanography: Methods",
  ["limnology (tokyo)"] = "Limnology",
  ["lin chuang er bi yan hou ke za zhi"] = "Lin chuang er bi yan hou ke za zhi = Journal of clinical otorhinolaryngology",
  ["lin chuang erh pi yen hou ko tsa chih"] = "Lin Chuang Erh Pi Yen Hou Ko Tsa Chih Journal of Clinical Otorhinolaryngology (Wu-Han Shih",
  ["lin chuang gan dan bing za zhi"] = "Lin chuang gan dan bing za zhi = Chinese journal of clinical hepatology",
  ["lin chuang jian yan za zhi"] = "[Lin chuang jian yan za zhi] [Journal of clinical laboratory examination]",
  ["lin chuang yi xue za zhi"] = "Lin chuang yi xue za zhi = Journal of clinical medicine",
  ["lin chung er bi yan hou tou jing wai ke za zhi"] = "Lin chuang er bi yan hou tou jing wai ke za zhi = Journal of clinical otorhinolaryngology, head, and neck surgery",
  ["lin ye ke xue (1979)"] = "Lin ye ke xue (Beijing, China : 1979)",
  ["lin ye ke xue yan jiu"] = "Lin ye ke xue yan jiu = Forest research",
  ["lin ye yu huan jing ke xue"] = "Lin ye yu huan jing ke xue = Forestry and Environmental Science",
  ["linacre q"] = "The Linacre quarterly",
  ["linc her"] = "Lincoln herald",
  ["linc law rev"] = "Lincoln law review (San Francisco, Calif.)",
  ["linchan huaxue yu gongye"] = "Linchan Huaxue Yu Gongye",
  ["lincoln rev"] = "Lincoln review (Washington, D.C.)",
  ["lincom textb math"] = "LINCOM Textbooks in Mathematics",
  ["linea ecol"] = "Linea Ecologica",
  ["linear algebra appl"] = "Linear Algebra and its Applications",
  ["linear and multilinear algebra"] = "Linear and Multilinear Algebra",
  ["linear multilinear algebra"] = "Linear and Multilinear Algebra",
  ["linear nonlinear anal"] = "Linear and Nonlinear Analysis. An International Journal",
  ["linear oper linear syst"] = "Linear Operators and Linear Systems",
  ["linear topol spaces complex anal"] = "Linear Topological Spaces and Complex Analysis",
  ["linen supply news"] = "Linen supply news",
  ["ling inq"] = "Linguistic Inquiry",
  ["lingua nostra"] = "Lingua nostra",
  ["linguist approaches biling"] = "Linguistic approaches to bilingualism",
  ["linguist inq"] = "Linguistic inquiry (Online)",
  ["linguist philos"] = "Linguistics and philosophy",
  ["linguist vanguard"] = "Linguistics vanguard : multimodal online journal",
  ["linguist var"] = "Linguistic variation",
  ["lingvist investig suppl"] = "Lingvisticae Investigationes: Supplementa",
  ["links (oxford)"] = "Links : a newsletter on gender for Oxfam GB staff and partners",
  ["linköping stud sci tech diss"] = "Linköping Studies in Science and Technology",
  ["linn soc symp ser"] = "Linnean Society Symposium Series",
  ["linneana belg"] = "Linneana belgica",
  ["linz biol beitr"] = "Linzer biologische Beiträge",
  ["lipics leibniz int proc inform"] = "LIPIcs. Leibniz International Proceedings in Informatics",
  ["lipid insights"] = "Lipid insights",
  ["lipid technol"] = "Lipid technology",
  ["lipids health dis"] = "Lipids in health and disease",
  ["lipp mitt gesch landeskd"] = "Lippische Mitteilungen aus Geschichte und Landeskunde",
  ["lippincott health promot lett"] = "The Lippincott health promotion letter",
  ["lippincotts case manag"] = "Lippincott's case management : managing the process of patient care",
  ["lippincotts prim care pract"] = "Lippincott's primary care practice",
  ["lipsey lect"] = "Lipsey Lectures",
  ["liq cryst"] = "Liquid Crystals",
  ["liq cryst rev"] = "Liquid Crystals Reviews",
  ["liq cryst today"] = "Liquid Crystal Today",
  ["liq waste recovery"] = "Liquid Waste Recovery",
  ["lisb med"] = "Lisboa médica",
  ["lishi yanjiu"] = "Lishi yanjiu",
  ["liste abr mots titres publ ser suppl"] = "Liste d’Abreviations de Mots des Titres de Publications en Serie. Supplement",
  ["listing res cardiovasc field"] = "A listing of research in the cardiovascular field",
  ["lit discuss"] = "Literacy discussion",
  ["lit hist"] = "Literature & history",
  ["lit inf comput educ j"] = "Literacy information and computer education journal",
  ["lit med"] = "Literature and medicine",
  ["lit numer stud"] = "Literacy & numeracy studies : an international journal in the education and training of adults",
  ["lit res instr"] = "Literacy research and instruction",
  ["lit research"] = "Literacy research : theory, method, and practice",
  ["lit teach learn"] = "Literacy, teaching and learning",
  ["liteinoe proizvod"] = "Liteinoe Proizvodstvo",
  ["lith hist stud"] = "Lithuanian historical studies",
  ["lith j phys"] = "Lithuanian Journal of Physics",
  ["lith math j"] = "Lithuanian Mathematical Journal",
  ["lithic technol"] = "Lithic Technology",
  ["lithium (edinb)"] = "Lithium (Edinburgh, Scotland)",
  ["lithol min resour"] = "Lithology and Mineral Resources",
  ["lithuanian math j"] = "Lithuanian Mathematical Journal",
  ["litt med soc"] = "Littérature, médecine, société",
  ["litt prag"] = "Litteraria pragensia",
  ["little hist"] = "Little Histories",
  ["liturg condenda"] = "Liturgy and Muse",
  ["liupanshui shi zhuan xuebao"] = "Liupanshui shi-zhuan xuebao",
  ["liver cancer"] = "Liver cancer",
  ["liver int"] = "Liver international : official journal of the International Association for the Study of the Liver",
  ["liver res"] = "Liver research",
  ["liver transpl"] = "Liver transplantation : official publication of the American Association for the Study of Liver Diseases and the International Liver Transplantation Society",
  ["liver transpl surg"] = "Liver transplantation and surgery : official publication of the American Association for the Study of Liver Diseases and the International Liver Transplantation Society",
  ["liverp classical mon"] = "Liverpool classical monthly : LCM",
  ["liverp law rev"] = "The Liverpool law review",
  ["liverp med chir j"] = "The Liverpool medico-chirurgical journal",
  ["livest prod sci"] = "Livestock production science",
  ["livest sci"] = "Livestock science",
  ["living 100 monogr"] = "Living to 100 monograph",
  ["living j comput mol sci"] = "Living journal of computational molecular science",
  ["living rev comput astrophys"] = "Living reviews in computational astrophysics",
  ["living rev landscape res"] = "Living Reviews in Landscape Research",
  ["living rev relativ"] = "Living Reviews in Relativity",
  ["living rev sol phys"] = "Living Reviews in Solar Physics",
  ["livre estampe"] = "Le Livre et l'estampe",
  ["lj spec rep"] = "LJ special report",
  ["llano estac heritage"] = "Llano Estacado heritage",
  ["lloyds bank rev"] = "Lloyds Bank Review",
  ["lms j comput math"] = "LMS Journal of Computation and Mathematics",
  ["lobachevskii j math"] = "Lobachevskii Journal of Mathematics",
  ["lobund reports"] = "Lobund reports",
  ["local econ"] = "Local economy",
  ["local economy"] = "Local Economy",
  ["local environ"] = "Local environment",
  ["local hist"] = "The Local historian",
  ["local land soil news"] = "Local land and soil news",
  ["local popul stud"] = "Local population studies",
  ["local reg anesth"] = "Local and regional anesthesia",
  ["locus juiz fora braz"] = "Locus (Juiz de Fora, Brazil)",
  ["loescher universita manuali"] = "Loescher Universita Manueli",
  ["log anal"] = "Logic and Analysis",
  ["log anal (ns)"] = "Logique et Analyse. Nouvelle Serie",
  ["log argum"] = "Logic and Argumentation",
  ["log argum reason"] = "Logic, Argumentation & Reasoning",
  ["log asia stud log libr"] = "Logic in Asia: Studia Logica Library",
  ["log cogn syst"] = "Logic and Cognitive Systems",
  ["log epistemol unity sci"] = "Logic, Epistemology, and the Unity of Science",
  ["log issled"] = "Logicheskie Issledovaniya",
  ["log j igpl"] = "Logic journal of the IGPL",
  ["log log philos"] = "Logic and Logical Philosophy",
  ["log methods comput sci"] = "Logical Methods in Computer Science",
  ["log mystic seapt"] = "The Log of Mystic seaport",
  ["log phds"] = "Logic PhDs",
  ["log trianguli"] = "Logica Trianguli",
  ["log univers"] = "Logica Universalis",
  ["logic comput philos"] = "Logic and Computation in Philosophy",
  ["logic j igpl"] = "Logic Journal of the IGPL",
  ["logic log philos"] = "Logic and Logical Philosophy",
  ["logica nova"] = "Logica Nova",
  ["logique et anal (ns)"] = "Logique et Analyse",
  ["logist inf manage"] = "Logistics Information Management",
  ["logist res"] = "Logistics Research",
  ["logist transp rev"] = "Logistics and Transportation Review",
  ["logist world"] = "Logistics World",
  ["logoped phoniatr vocol"] = "Logopedics, phoniatrics, vocology",
  ["logos (santa clara)"] = "Logos (Santa Clara, Calif.)",
  ["loma linda univ dent mag"] = "Loma Linda University Dentist Magazine",
  ["lond archaeol"] = "The London archaeologist",
  ["lond clin med j"] = "London Clinic medical journal",
  ["lond edinb dublin philos mag j sci"] = "The London, Edinburgh and Dublin philosophical magazine and journal of science",
  ["lond edinb mon j med sci"] = "London and Edinburgh monthly journal of medical science",
  ["lond hosp gaz"] = "The London Hospital gazette",
  ["lond j"] = "The London journal",
  ["lond j med"] = "London journal of medicine",
  ["lond math soc newsl"] = "London Mathematical Society. Newsletter",
  ["lond med j"] = "The London medical journal",
  ["lond med phys j"] = "The London medical and physical journal",
  ["london j prim care (abingdon)"] = "London journal of primary care",
  ["london math soc lecture note ser"] = "London Mathematical Society Lecture Note Series",
  ["london math soc monogr (ns)"] = "London Mathematical Society Monographs",
  ["london math soc stud texts"] = "London Mathematical Society Student Texts",
  ["long isl forum"] = "Long Island forum",
  ["long island hist j"] = "The Long Island historical journal",
  ["long range plann"] = "Long range planning",
  ["long term care (don mills)"] = "Long term care (Don Mills, Ont.)",
  ["long term care health serv adm q"] = "Long term care and health services administration quarterly",
  ["long term care q"] = "Long-term care quarterly",
  ["longev healthspan"] = "Longevity & healthspan",
  ["longit life course stud"] = "Longitudinal and life course studies",
  ["los alamos nat lab prepr arch astrophys"] = "Los Alamos National Laboratory, Preprint Archive, Astrophysics",
  ["los alamos nat lab prepr arch condens matter"] = "Los Alamos National Laboratory, Preprint Archive, Condensed Matter",
  ["los alamos nat lab prepr arch gen relativ quantum cosmol"] = "Los Alamos National Laboratory, Preprint Archive, General Relativity and Quantum Cosmology",
  ["los alamos nat lab prepr arch high energy phys exp"] = "Los Alamos National Laboratory, Preprint Archive, High Energy Physics-Experiment",
  ["los alamos nat lab prepr arch high energy phys lattice"] = "Los Alamos National Laboratory, Preprint Archive, High Energy Physics-Lattice",
  ["los alamos nat lab prepr arch high energy phys nucl exp"] = "Los Alamos National Laboratory, Preprint Archive, Nuclear Experiment",
  ["los alamos nat lab prepr arch high energy phys phenomenol"] = "Los Alamos National Laboratory, Preprint Archive, High Energy Physics-Phenomenology",
  ["los alamos nat lab prepr arch high energy phys theory"] = "Los Alamos National Laboratory, Preprint Archive, High Energy Physics-Theory",
  ["los alamos nat lab prepr arch nucl theory"] = "Los Alamos National Laboratory, Preprint Archive, Nuclear Theory",
  ["los alamos nat lab prepr arch phys"] = "Los Alamos National Laboratory, Preprint Archive, Physics",
  ["los alamos nat lab prepr arch quant biol"] = "Los Alamos National Laboratory, Preprint Archive, Quantitative Biology",
  ["los alamos nat lab prepr arch quantum phys"] = "Los Alamos National Laboratory, Preprint Archive, Quantum Physics",
  ["los alamos natl lab prepr arch astrophys"] = "Los Alamos National Laboratory, Preprint Archive, Astrophysics",
  ["los alamos natl lab prepr arch condens matter"] = "Los Alamos National Laboratory, Preprint Archive, Condensed Matter",
  ["los alamos natl lab prepr arch gen relativ quantum cosmol"] = "Los Alamos National Laboratory, Preprint Archive, General Relativity and Quantum Cosmology",
  ["los alamos natl lab prepr arch high energy phys exp"] = "Los Alamos National Laboratory, Preprint Archive, High Energy Physics-Experiment",
  ["los alamos natl lab prepr arch high energy phys lattice"] = "Los Alamos National Laboratory, Preprint Archive, High Energy Physics-Lattice",
  ["los alamos natl lab prepr arch high energy phys nucl exp"] = "Los Alamos National Laboratory, Preprint Archive, Nuclear Experiment",
  ["los alamos natl lab prepr arch high energy phys phenomenol"] = "Los Alamos National Laboratory, Preprint Archive, High Energy Physics-Phenomenology",
  ["los alamos natl lab prepr arch high energy phys theory"] = "Los Alamos National Laboratory, Preprint Archive, High Energy Physics-Theory",
  ["los alamos natl lab prepr arch nucl theory"] = "Los Alamos National Laboratory, Preprint Archive, Nuclear Theory",
  ["los alamos natl lab prepr arch phys"] = "Los Alamos National Laboratory, Preprint Archive, Physics",
  ["los alamos natl lab prepr arch quant biol"] = "Los Alamos National Laboratory, Preprint Archive, Quantitative Biology",
  ["los alamos natl lab prepr arch quantum phys"] = "Los Alamos National Laboratory, Preprint Archive, Quantum Physics",
  ["los alamos sci"] = "Los Alamos science",
  ["los angel lawyer"] = "Los Angeles lawyer",
  ["los angeles times"] = "The Los Angeles times",
  ["loss grief care"] = "Loss, grief & care",
  ["loteria (panama)"] = "Revista lotería",
  ["lotta contro tuberc"] = "Lotta Contro la Tubercolosi",
  ["lotta contro tuberc malat polm sociali"] = "Lotta contro la tubercolosi e le malattie polmonari sociali",
  ["lotta tuberc"] = "Lotta contro la tubercolosi",
  ["louv med"] = "Louvain médical",
  ["louv stud"] = "Louvain studies",
  ["louvain econ rev"] = "Louvain Economic Review",
  ["lovelace clin rev"] = "The Lovelace Clinic review",
  ["lovtid k dan a"] = "Lovtidende. A, Almindelige love, anordninger og bekendtgoerelser. Denmark",
  ["low carbon econ"] = "Low carbon economy",
  ["low temp phys"] = "Low Temperature Physics",
  ["low urin tract symptoms"] = "Lower urinary tract symptoms",
  ["loyola consum law rev"] = "Loyola consumer law review",
  ["loyola law rev"] = "Loyola law review",
  ["loyola los angel int comp law j"] = "Loyola of Los Angeles international and comparative law journal",
  ["loyola los angel law rev"] = "Loyola of Los Angeles law review",
  ["loyola univ chic law j"] = "Loyola University of Chicago law journal. Loyola University Chicago. School of Law",
  ["loyola univ chicago law j"] = "Loyola University of Chicago law journal. Loyola University of Chicago. School of Law",
  ["lso work pap linguist"] = "LSO working papers in linguistics",
  ["ltc regul risk liabil advis"] = "LTC regulatory risk & liability advisor",
  ["ltcc adv math ser"] = "LTCC Advanced Mathematics Series",
  ["lubr eng"] = "Lubrication Engineering",
  ["lubr sci"] = "Lubrication Science",
  ["luft raumfahrt"] = "Luft- und Raumfahrt",
  ["lumiere class"] = "Lumiere Classique",
  ["lundellia (austin tex)"] = "Lundellia : journal of the Plant Resources Center of the University of Texas at Austin",
  ["lundqua rep"] = "Lundqua Report",
  ["lung cancer"] = "Lung cancer (Amsterdam, Netherlands)",
  ["lung cancer (auckl)"] = "Lung Cancer (Auckland, N.Z.)",
  ["lung cancer int"] = "Lung cancer international",
  ["lung cancer manag"] = "Lung cancer management",
  ["lung cancer manage"] = "Lung Cancer Management",
  ["lung health prof mag"] = "Lung health professional magazine",
  ["lung india"] = "Lung India : official organ of Indian Chest Society",
  ["lungs breath"] = "Lungs and breathing",
  ["lupus (los angel)"] = "Lupus (Los Angeles)",
  ["lupus sci med"] = "Lupus science & medicine",
  ["luso-braz rev"] = "Luso-Brazilian review",
  ["lutheran q"] = "Lutheran quarterly (Gettysburg, Pa. : 1949)",
  ["lutheran theol j"] = "Lutheran theological journal",
  ["lutte cancer"] = "La Lutte contre le cancer",
  ["luzif amor"] = "Luzifer-Amor : Zeitschrift zur Geschichte der Psychoanalyse",
  ["lwt food sci technol"] = "LWT Food Science and Technology",
  ["lwt–food sci technol"] = "LWT–Food Science and Technology",
  ["lychnos lardomshist samf arsb"] = "Lychnos : Lärdomshistoriska samfundets årsbok = annual of the Swedish History of Science Society",
  ["lying in"] = "Lying-in",
  ["lymphat res biol"] = "Lymphatic research and biology",
  ["lymphokine cytokine res"] = "Lymphokine and cytokine research",
  ["lymphokine res"] = "Lymphokine research",
  ["lyon chir"] = "Lyon chirurgical",
  ["lyon med"] = "Lyon médical",
  ["lyon mediterr med med sud est"] = "Lyon Méditerranée médical. Médecine du Sud-Est",
  ["lyon pharm"] = "Lyon pharmaceutique",
  ["löbf-mitt, landesanst ökol bodenordn forsten, landesamt agrarordn nordrh-westfal"] = "LÖBF-Mitteilungen, Landesanstalt für Ökologie, Bodenordnung und Forsten / Landesamt für Agrarordnung, Nordrhein-Westfalen",
  ["löbf-schrreihe, landesanst ökol bodenordn forsten nordrh-westfal"] = "LÖBF-Schriftenreihe, Landesanstalt für Ökologie, Bodenordnung und Forsten, Nordrhein-Westfalen",
  ["lölf-mitt, landesanst ökol landschentwickl  forstplan, nordrh-westfal"] = "LÖLF-Mitteilungen, Landesanstalt für Ökologie, Landschaftsentwicklung und Forstplanung, Nordrhein-Westfalen",
  ["l’actual econ"] = "L’Actualité Economique",
  ["l’evolution de l’humanite, synthese collective premiere section, serie complementaire: la science dans l’an"] = "L’Evolution de l’Humanite, Synthese Collective. Premiere Section, Serie Complementaire: La Science dans l’Antiquite",
  ["m b pharm bull"] = "M & B pharmaceutical bulletin",
  ["m b porter lectures"] = "M",
  ["m s c vet"] = "M. S. C. veterinarian",
  ["m&t ser"] = "M&T Series",
  ["ma zui xue za zhi"] = "Ma zui xue za zhi = Anaesthesiologica Sinica",
  ["ma zui yu jian hu lun tan"] = "Ma zui yu jian hu lun tan = Forum of anesthesia and monitoring",
  ["maa notes"] = "MAA Notes",
  ["maa probl books ser"] = "MAA Problem Books Series",
  ["maa spectrum"] = "MAA Spectrum",
  ["maa textb"] = "MAA Textbooks",
  ["maandschr cent bur stat"] = "Maandschrift - Centraal Bureau voor de Statistiek. Netherlands. Centraal Bureau voor de Statistiek",
  ["maandschr kindergeneeskd"] = "Maandschrift voor kindergeneeskunde",
  ["maandstat bevolking"] = "Maandstatistiek van de bevolking (Hague, Netherlands : 1982)",
  ["maandstat bevolking volksgezond"] = "Maandstatistiek van bevolking en volksgezondheid. Netherlands. Centraal Bureau voor de Statistiek",
  ["maas j islam sci"] = "MAAS journal of Islamic science",
  ["maastrich j eur comp law"] = "Maastricht journal of European and comparative law",
  ["maced j chem chem eng"] = "Macedonian Journal of Chemistry and Chemical Engineering",
  ["maced vet rev"] = "Macedonian veterinary review",
  ["mach intell res"] = "Machine Intelligence Research",
  ["mach intelligence pattern recogn"] = "Machine Intelligence and Pattern Recognition",
  ["mach learn"] = "Machine Learning",
  ["mach learn appl"] = "Machine Learning with Applications",
  ["mach learn found methodol appl"] = "Machine Learning: Foundations, Methodologies, and Applications",
  ["mach learn interpret neuroimaging (2011)"] = "Machine learning and interpretation in neuroimaging : international workshop, MLINI 2011, held at NIPS 2011, Sierra Nevada, Spain, December 16-17, 2011 : revised selected and invited contributions. MLINI (Workshop) (2011 : Sierra Nevada, Spain)",
  ["mach learn interpret neuroimaging (2014)"] = "Machine learning and interpretation in neuroimaging : 4th international workshop, MLINI 2014, held at NIPS 2014, Montreal QC, Canada, December 13, 2014 : revised selected papers. MLINI (Workshop) (4th : 2014 : Montréal, Québec)",
  ["mach learn knowl extr"] = "Machine learning and knowledge extraction",
  ["mach learn knowl extr (2019)"] = "Machine learning and knowledge extraction : third IFIP TC 5, TC 12, WG 8.4, WG 8.9, WG 12.9 international cross-domain conference, CD-MAKE 2019, Canterbury, UK, August 26-29, 2019 : proceedings. International Cross-Domain Conference for Machine Learning and Knowledge Extraction (3rd : 2019 : Canterbury, England)",
  ["mach learn med eng cardiovasc health intravasc imaging comput assist stenting (2019)"] = "Machine learning and medical engineering for cardiovascular health and intravascular imaging and computer assisted stenting : first International Workshop, MLMECH 2019, and 8th Joint International Workshop, CVII-STENT 2019, held in conjunction with MICCAI 2019, Shenzhen, China, October 13, 2019, Proceedings. MLMECH (Workshop) (1st : 2019 : Shenzhen Shi, China)",
  ["mach learn med imaging"] = "Machine learning in medical imaging. MLMI (Workshop)",
  ["mach learn sci technol"] = "Machine learning: science and technology",
  ["mach learn: earth"] = "Machine Learning: Earth",
  ["mach learn: eng"] = "Machine Learning: Engineering",
  ["mach learn: health"] = "Machine Learning: Health",
  ["mach learn: sci technol"] = "Machine Learning: Science and Technology",
  ["mach sci technol"] = "Machining Science and Technology",
  ["mach transl"] = "Machine translation : MT",
  ["mach vis appl"] = "Machine vision and applications",
  ["mach vision appl"] = "Machine Vision and Applications",
  ["macmillan new electron ser"] = "Macmillan New Electronics Series",
  ["macroecon dyn"] = "Macroeconomic dynamics",
  ["macroecon dynam"] = "Macroeconomic Dynamics",
  ["macromol biosci"] = "Macromolecular bioscience",
  ["macromol chem phys"] = "Macromolecular Chemistry and Physics",
  ["macromol chem phys suppl"] = "Macromolecular Chemistry and Physics Supplement",
  ["macromol mater eng"] = "Macromolecular Materials and Engineering",
  ["macromol rapid commun"] = "Macromolecular rapid communications",
  ["macromol react eng"] = "Macromolecular Reaction Engineering",
  ["macromol res"] = "Macromolecular Research",
  ["macromol symp"] = "Macromolecular Symposia",
  ["macromol theory simul"] = "Macromolecular Theory and Simulations",
  ["macromolecules (washington, dc, u s)"] = "Macromolecules (Washington, DC, United States)",
  ["macrophage (houst)"] = "Macrophage",
  ["madag conserv dev"] = "Madagascar conservation and development",
  ["madima16 (2016)"] = "MADiMa'16 : proceedings of the 2nd International Workshop on Multimedia Assisted Dietary Management : October 16, 2016, Amsterdam, The Netherlands. International Workshop on Multimedia Assisted Dietary Management (2nd : 2016 : Amsterdam, Netherlands)",
  ["madj persat dokt gigi indones"] = "Madjalah PDGI",
  ["madjalah kedokt indones"] = "Madjalah Kedokteran Indonesia",
  ["madridge j cancer study res"] = "Madridge journal of cancer study & research",
  ["madridge j dent oral surg"] = "Madridge journal of dentistry and oral surgery",
  ["madridge j diabetes"] = "Madridge journal of diabetes",
  ["madridge j pharm res"] = "Madridge journal of pharmaceutical research",
  ["maedica (bucur)"] = "Mædica",
  ["maejo int j sci technol"] = "Maejo International Journal of Science and Technology",
  ["mag albemarle cty hist"] = "The Magazine of Albemarle County history",
  ["mag antiq"] = "Magazine antiques (New York, N.Y. : 1971)",
  ["mag camb univ med soc"] = "Magazine. Cambridge University Medical Society",
  ["mag concr res"] = "Magazine of Concrete Research",
  ["mag istor"] = "Magazin istoric",
  ["mag litt"] = "Magazine littéraire",
  ["mag r free hosp (lond engl)"] = "Magazine. Royal Free Hospital (London, England). School of Medicine",
  ["magallania (punta arenas)"] = "Magallania",
  ["magallat albasrat liabhat nahlat altamr"] = "Maǧallaẗ al-baṣraẗ li-abḥāt̲ nahlaẗ al-tamr",
  ["magallat albasrat lilabhat albaytariyyat"] = "Maǧallaẗ al-baṣraẗ li-l-abḥāt̲ al-bayṭariyyat̲",
  ["maghreb machrek"] = "Maghreb, Machrek",
  ["maghreb math rev"] = "Maghreb Mathematical Review",
  ["maghreb med"] = "Maghreb médical = al-Majallah al-ṭibbīyah lil-Maghrib al-ʻArabī",
  ["maghreb rev"] = "The Maghreb review. Majallat al-Maghrib",
  ["magic ritual witch"] = "Magic, ritual, and witchcraft",
  ["magn reson"] = "Magnetic Resonance",
  ["magn reson (gott)"] = "Magnetic resonance (Göttingen, Germany)",
  ["magn reson annu"] = "Magnetic resonance annual",
  ["magn reson chem"] = "Magnetic resonance in chemistry : MRC",
  ["magn reson imaging"] = "Magnetic resonance imaging",
  ["magn reson imaging clin n am"] = "Magnetic resonance imaging clinics of North America",
  ["magn reson insights"] = "Magnetic resonance insights",
  ["magn reson lett"] = "Magnetic Resonance Letters",
  ["magn reson mater phys, biol med"] = "Magnetic Resonance Materials in Physics, Biology and Medicine",
  ["magn reson med"] = "Magnetic resonance in medicine",
  ["magn reson med sci"] = "Magnetic resonance in medical sciences : MRMS : an official journal of Japan Society of Magnetic Resonance in Medicine",
  ["magn reson q"] = "Magnetic resonance quarterly",
  ["magnes res"] = "Magnesium research",
  ["magnes trace elem"] = "Magnesium and trace elements",
  ["magnesium res"] = "Magnesium Research",
  ["magnit gidrodinamika"] = "Akademiya Nauk Latviiskoy SSR",
  ["magy allatorvosok lapja"] = "Magyar állatorvosok lapja",
  ["magy belorv arch"] = "Magyar belorvosi archivum",
  ["magy belorv arch (1955)"] = "Magyar belorvosi archivum (1955)",
  ["magy belorv arch ideggyogy szle"] = "Magyar belorvosi archivum és ideggyógyászati szemle",
  ["magy kem foly"] = "Magyar Kemiai Folyoirat",
  ["magy kem foly, kem kozl"] = "Magyar Kemiai Folyoirat, Kemiai Kozlemenyek",
  ["magy noorv lapja"] = "Magyar nöorvosok lapja",
  ["magy onkol"] = "Magyar onkologia",
  ["magy radiol"] = "Magyar radiologia",
  ["magy seb"] = "Magyar sebészet",
  ["magy traumatol orthop helyreallito seb"] = "Magyar traumatológia, orthopaedia és helyreállító sebészet",
  ["magy traumatol ortop kezseb plasztikai seb"] = "Magyar traumatológia, ortopédia, kézsebészet, plasztikai sebészet",
  ["magy tud"] = "Magyar tudomány : [a Magyar Tudományos Akadémia Értesítője]",
  ["magyar konyvsz"] = "Magyar Könyvszemle",
  ["mahapanjikara kipatrika"] = "Mahāpañjīkāra kī patrikā = Registrar General's news letter. India. Office of the Registrar General",
  ["maharashtra med j"] = "Maharashtra medical journal",
  ["mahidol popul q gaz"] = "Mahidol population gazette",
  ["mahidol university journal of pharmaceutical sciences / mahawitthayalai mahidon|warasan phesatchasat"] = "Warasan phesatchasat",
  ["mai lei zuo wu xue bao"] = "Mai lei zuo wu xue bao",
  ["main group chem"] = "Main Group Chemistry",
  ["main group met chem"] = "Main Group Metal Chemistry",
  ["maine hist soc q"] = "Maine Historical Society quarterly",
  ["maine law rev"] = "Maine law review",
  ["maine nurse"] = "Maine nurse",
  ["maine revis statut annot 1964 maine"] = "Maine revised statutes annotated, 1964 : being the tenth revision of the Revised statutes of the state of Maine, 1964. Maine",
  ["mainsl haul"] = "Mains'l haul",
  ["maison poincare regards math"] = "Maison Poincare. Regards Mathematiques",
  ["maize genet coop news lett"] = "Maize genetics cooperation news letter",
  ["maj danishkadahi dampizishki"] = "Majallah-i dānishkadah-i dāmpizishkī",
  ["maj iptek inst teknol  sepuluh nop 1945 surabaya"] = "Majalah IPTEK Institut Teknologi Sepuluh Nopember 1945 Surabaya",
  ["majalah demografi indones"] = "Majalah demografi Indonesia",
  ["majalah kedokt indones"] = "Majalah Kedokteran Indonesia",
  ["majalah obstet ginekol indones"] = "Majalah obstetri dan ginekologi Indonesia",
  ["majallah iulum i payah i pizishki i iran"] = "Majallah-i'ulūm-i pāyah-i pizishkī-i Īrān = Iranian journal of basic medical sciences",
  ["majallah jamia dandan pazshki"] = "Majallah-i Jāmiah-i Dandānpizishkī-i Īrān = The journal of the Iranian Dental Association",
  ["majallahi danishgahi ulumi pizishkii mazandaran"] = "Majallah-i Dānishgāh-i ʻUlūm-i Pizishkī-i Māzandarān",
  ["majallahi danishi giyah pizishkii iran"] = "Majallah-i dānish-i giyāh/pizishkī-i Īrān",
  ["majallahi danishkadahi pizishkii isfahan"] = "Majallāh-i dānishkadah-i pizishkī-i iṣfahān",
  ["majallahi hifazati giyahan"] = "Majallah-i hifāzat-i giyāhān",
  ["majallahi ilmi pizhuhishii danishgahi ulumi pizishki khadamati bihdashtii darmanii zanjan"] = "Majallah-i ʻilmī pizhūhishī-i Dānishgāh-i ʻUlūm-i Pizishkī va Khadamāt-i Bihdāshtī-i Darmānī-i Zanjān",
  ["majallahi ilmipizhuhishii danishgahilumi pizishki va khadamati bihdashti darmanii arak"] = "Majallah-iʻ ilmī-pizhūhishī-i dānishgāh-iʻlūm-i pizishkī va khadamāt-i bihdāshtī darmānī-i Arāk",
  ["majallahi ulum fununi daryayi"] = "Majallah-i ulūm va funūn-i daryāyī",
  ["majallahi zist shinasii iran"] = "Majallah-i zīst/shināsī-i Īrān",
  ["majallat albuhuth waaldirasat alarabiyah"] = "Majallat al-buḥūth wa-al-dirāsāt al-ʻArabīyah",
  ["majallat aldiwan alqawmi lilusrah waal umran lbashari"] = "Majāllat al-Dīwān al-Qawmī lil-Usrah wa-al-ʻUmrān al-Basharī",
  ["majallat alsharq alawasat"] = "Majallat al-Sharq al-Awsaṭ",
  ["majallat ma had al makhtutat al arabiy"] = "Majallat ma had al-makhtūtāt al-Arabiyah",
  ["majallat niqabat attiba alasnnan alurduniyah"] = "Majallat Niqābat Attibā' al-Asnān al-Urdunīyah",
  ["majallat tibb alasnan alsuriyah"] = "Majallat tibb al-asnān al-Sūrīyah",
  ["majallat tibb alfamm alsuriyah"] = "Majallat tibb al-famm al-Sūrīyah",
  ["major amer univ phd qualif questions solut"] = "Major American Universities Ph",
  ["major probl clin pediatr"] = "Major problems in clinical pediatrics",
  ["major probl clin surg"] = "Major problems in clinical surgery",
  ["major probl intern med"] = "Major problems in internal medicine",
  ["major probl obstet gynecol"] = "Major problems in obstetrics and gynecology",
  ["major probl pathol"] = "Major problems in pathology",
  ["mak rounds health faith ethics"] = "Making the rounds in health, faith, & ethics",
  ["makara j sci"] = "Makara journal of science",
  ["makedon akad nauk umet oddel mat-tehn nauk prilozi"] = "Makedonska Akademija na Naukite i Umetnostite",
  ["makedon med pregl"] = "Makedonski medicinski pregled. Revue médicale macedonienne",
  ["makinoa new ser"] = "Makinoa new series : bulletin of the Makino Botanical Garden",
  ["makromol chem"] = "Makromolekulare Chemie",
  ["makromol chem\trapid commun"] = "Makromolekulare Chemie, Rapid Communications",
  ["makromol chem macromol symp"] = "Makromolekulare Chemie, Macromolecular Symposia",
  ["makromol chem rapid commun"] = "Makromolekulare Chemie, Rapid Communications",
  ["makromol chem suppl"] = "Makromolekulare Chemie, Supplement",
  ["makromol chem theory simul"] = "Makromolekulare Chemie, Theory and Simulations",
  ["mal cardiovasc"] = "Malattie cardiovascolari",
  ["mala jazyk kn"] = "Mala Jazykova Kniznice",
  ["malacol bohemoslov"] = "Malacologica Bohemoslovaca : journal about molluscs in Central Europe",
  ["malagasy nat"] = "Malagasy nature",
  ["malar chemother control elimin"] = "Malaria chemotherapy, control & elimination",
  ["malar control elimin"] = "Malaria control & elimination",
  ["malar j"] = "Malaria journal",
  ["malar res treat"] = "Malaria research and treatment",
  ["malar wkly"] = "Malaria weekly",
  ["malariaworld j"] = "MalariaWorld journal",
  ["malawi med j"] = "Malawi medical journal : the journal of Medical Association of Malawi",
  ["malawian geogr"] = "The Malawian geographer",
  ["malay econ rev"] = "The Malayan economic review : the journal of the Economic Society of Singapore, the Department of Economics and Statistics and the Economic Research Centre of the University of Singapore",
  ["malaya j biosci"] = "Malaya journal of biosciences",
  ["malaya j mat"] = "Malaya Journal of Matematik",
  ["malayan econ rev"] = "Malayan Economic Review",
  ["malayan j trop geogr"] = "Malayan Journal of Tropical Geography",
  ["malayan nat j"] = "Malayan Nature Journal, The",
  ["malays appl biol"] = "Malaysian applied biology = Biologi gunaan Malaysia",
  ["malays dent j"] = "Malaysian dental journal",
  ["malays fam physician"] = "Malaysian family physician : the official journal of the Academy of Family Physicians of Malaysia",
  ["malays for"] = "Malaysian Forester, The",
  ["malays j biochem mol biol"] = "Malaysian journal of biochemistry and molecular biology",
  ["malays j econ studies"] = "Malaysian journal of economic studies : journal of the Malaysian Economic Association and the Faculty of Economics and Administration, University of Malaya",
  ["malays j math sci"] = "Malaysian Journal of Mathematical Sciences",
  ["malays j med sci"] = "The Malaysian journal of medical sciences : MJMS",
  ["malays j microbiol"] = "Malaysian journal of microbiology",
  ["malays j nutr"] = "Malaysian journal of nutrition",
  ["malays j pathol"] = "The Malaysian journal of pathology",
  ["malays j reprod health"] = "Malaysian journal of reproductive health : a publication of the Reproductive Research Centre of the National Population and Family Development Board, Malaysia",
  ["malays j trop geogr"] = "Malaysian journal of tropical geography",
  ["malays j vet res (putrajaya)"] = "Malaysian journal of veterinary research",
  ["malays orthop j"] = "Malaysian orthopaedic journal",
  ["male monogr pwn tech"] = "Male Monografie PWN. Technika",
  ["male nurses j"] = "The male nurses' journal",
  ["mali med"] = "Le Mali médical",
  ["malloch room newsl"] = "The Malloch room newsletter",
  ["malpract dig"] = "Malpractice digest",
  ["mamm biol"] = "Mammalian biology = Zeitschrift für Säugetierkunde",
  ["mamm biol z saugetierkd"] = "Mammalian Biology - Zeitschrift fur Saugetierkunde",
  ["mamm genome"] = "Mammalian genome : official journal of the International Mammalian Genome Society",
  ["mamm res"] = "Mammal research",
  ["mamm rev"] = "Mammal review",
  ["mamm species"] = "Mammalian Species",
  ["mammal res"] = "Mammal Research",
  ["mammal rev"] = "Mammal Review",
  ["mammal study"] = "Mammal study",
  ["man (lond)"] = "Man; a monthly record of anthropological science",
  ["man hist"] = "Manitoba history",
  ["man india"] = "Man in India",
  ["man med"] = "Man and medicine",
  ["man ther"] = "Manual therapy",
  ["man-made text india"] = "Man-Made Textiles in India",
  ["manag biol invasion"] = "Management of biological invasions : international journal of applied research on biological invasions",
  ["manag care"] = "Managed care (Langhorne, Pa.)",
  ["manag care interface"] = "Managed care interface",
  ["manag care q"] = "Managed care quarterly",
  ["manag care strateg"] = "Managed care strategies (Atlanta, Ga.)",
  ["manag commun q"] = "Management communication quarterly",
  ["manag compliance ser"] = "Management and compliance series",
  ["manag ind eng"] = "Management and Industrial Engineering",
  ["manag int dev"] = "Managing international development : MID",
  ["manag int rev"] = "Management international review : MIR : journal of international business",
  ["manag learn"] = "Management learning",
  ["manag medicare medicaid news"] = "Managed medicare & medicaid news",
  ["manag organ rev"] = "Management and organization review : the journal of the International Association for Chinese Management Research",
  ["manag prof"] = "Management for Professionals",
  ["manage acc"] = "Management Accounting",
  ["manage biol invasions"] = "Management of Biological Invasions",
  ["manage environ qual"] = "Management of Environmental Quality",
  ["manage focus"] = "Management focus",
  ["manage rev"] = "Management review",
  ["manage sci"] = "Management science",
  ["manage sustainability: arab rev"] = "Management & Sustainability: An Arab Review",
  ["manage train rev"] = "Management Training Review",
  ["manage world"] = "Management world",
  ["management and inform systems"] = "Management and Information Systems",
  ["managerial dec econ"] = "Managerial and Decision Economics",
  ["manch med gaz"] = "The Manchester medical gazette",
  ["manch men"] = "Manchester memoirs",
  ["manch univ med sch gaz"] = "Manchester University Medical School gazette",
  ["manchester phys ser"] = "The Manchester Physics Series",
  ["manchester sch"] = "Manchester School",
  ["manchester sch econ soc stud"] = "The Manchester school of economic and social studies",
  ["manedsskr prakt laegegern"] = "Maanedsskrift for praktisk lægegerning og social Medicin",
  ["manglar (tumbes)"] = "Manglar (Tumbes)",
  ["mangroves salt marshes"] = "Mangroves and Salt Marshes",
  ["manit law j"] = "Manitoba law journal",
  ["manit med rev"] = "Manitoba medical review",
  ["mankind q"] = "The mankind quarterly",
  ["manshur atmajallat dirasat alkhalij waal jazirah alarabiyah"] = "Manshūrāt Majallat dirāsāt al-Khalīj wa-al-Jazīrah al-ʻArabīyah",
  ["mansoura j chem"] = "Mansoura Journal of Chemistry",
  ["manter (linc)"] = "Manter (Lincoln, Neb.)",
  ["manuf chem"] = "Manufacturing Chemist",
  ["manuf chem aerosol news"] = "Manufacturing chemist and aerosol news",
  ["manuf eng mater proc"] = "Manufacturing Engineering and Materials Processing",
  ["manuf eng mater process"] = "Manufacturing Engineering and Materials Processing",
  ["manuf lett"] = "Manufacturing letters",
  ["manuf rev"] = "Manufacturing Review",
  ["manuf rev (les ulis)"] = "Manufacturing review",
  ["manuf serv oper manag"] = "Manufacturing & service operations management : M & SOM",
  ["manuscr math"] = "Manuscripta mathematica",
  ["manuscripta math"] = "Manuscripta Mathematica",
  ["manuscripts (n y)"] = "Manuscripts (New York, N.Y.)",
  ["maple tech newsl"] = "Maple Technical Newsletter, The",
  ["mar behav physiol"] = "Marine behaviour and physiology",
  ["mar biodivers"] = "Marine Biodiversity",
  ["mar biodivers rec"] = "Marine biodiversity records",
  ["mar biol"] = "Marine Biology",
  ["mar biol res"] = "Marine Biology Research",
  ["mar biotechnol"] = "Marine Biotechnology",
  ["mar biotechnol (ny)"] = "Marine biotechnology (New York, N.Y.)",
  ["mar chem"] = "Marine Chemistry",
  ["mar coast fish"] = "Marine and coastal fisheries : dynamics, management , and ecosystem science",
  ["mar coastal fish"] = "Marine and Coastal Fisheries",
  ["mar corps gaz"] = "The Marine Corps gazette",
  ["mar drugs"] = "Marine drugs",
  ["mar ecol"] = "Marine Ecology",
  ["mar ecol (berl)"] = "Marine ecology (Berlin, Germany)",
  ["mar ecol prog ser"] = "Marine ecology progress series",
  ["mar econ manage"] = "Marine Economics and Management",
  ["mar environ res"] = "Marine environmental research",
  ["mar fish rev"] = "Marine Fisheries Review",
  ["mar freshw behav physiol"] = "Marine and freshwater behaviour and physiology",
  ["mar freshw harmful algae (2014)"] = "Marine and freshwater harmful algae : proceedings of the 16th International Conference on Harmful Algae, 27th-31st October 2014, Wellington, New Zealand. International Conference on Harmful Algae (16th : 2014 : Wellington, N.Z.), author",
  ["mar freshw res"] = "Marine & freshwater research",
  ["mar freshwater behav physiol"] = "Marine and Freshwater Behaviour and Physiology",
  ["mar freshwater res"] = "Marine and Freshwater Research",
  ["mar genomics"] = "Marine genomics",
  ["mar geod"] = "Marine Geodesy",
  ["mar geol"] = "Marine Geology",
  ["mar geonomics"] = "Marine Genomics",
  ["mar geophys res"] = "Marine Geophysical Researches",
  ["mar geophys res (dordr)"] = "Marine geophysical researches",
  ["mar georesour geotechnol"] = "Marine Georesources and Geotechnology",
  ["mar geotechnol"] = "Marine Geotechnology",
  ["mar life sci technol"] = "Marine life science & technology",
  ["mar mamm sci"] = "Marine Mammal Science",
  ["mar mammal sci"] = "Marine Mammal Science",
  ["mar microb food webs"] = "Marine Microbial Food Webs",
  ["mar micropaleontol"] = "Marine Micropaleontology",
  ["mar min"] = "Marine Mining",
  ["mar mirror"] = "The Mariner's mirror",
  ["mar obs"] = "Marine Observer",
  ["mar ornithol"] = "Marine ornithology",
  ["mar pet geol"] = "Marine and Petroleum Geology",
  ["mar policy"] = "Marine policy",
  ["mar pollut bull"] = "Marine pollution bulletin",
  ["mar resour econ"] = "Marine Resource Economics",
  ["mar rundsch"] = "Marine-rundschau",
  ["mar sci biol"] = "Marine science and biology",
  ["mar sci technol bull"] = "Marine science and technology bulletin",
  ["mar struct"] = "Marine Structures",
  ["mar syst ocean technol"] = "Marine Systems & Ocean Technology",
  ["mar technol"] = "Marine Technology",
  ["mar technol soc j"] = "Marine Technology Society Journal",
  ["mar wea log"] = "Mariners Weather Log",
  ["marbg geogr schr"] = "Marburger geographische Schriften",
  ["marbg schrift medgesch"] = "Marburger Schriften zur Medizingeschichte",
  ["marg otolaryngol"] = "Marginalia otolaryngologica",
  ["margin j appl econ res"] = "Margin - the journal of applied economic research",
  ["marginalia dermatol"] = "Marginalia dermatologica",
  ["margins (baltim)"] = "Margins (Baltimore, Md.)",
  ["marine resource econ"] = "Marine Resource Economics",
  ["mario boella ser electromagn inf commun"] = "Mario Boella Series on Electromagnetism in Information and Communication",
  ["marit aff: j natl marit found india"] = "Maritime Affairs: Journal of the National Maritime Foundation of India",
  ["marit occup health"] = "Maritime Occupational Health",
  ["marit policy manage"] = "Maritime Policy & Management",
  ["marit transport res"] = "Maritime Transport Research",
  ["maritime pol manage"] = "Maritime Policy and Management",
  ["mark forces"] = "Market forces",
  ["mark health serv"] = "Marketing health services",
  ["mark lett"] = "Marketing letters",
  ["mark public policy conf proc"] = "Marketing and Public Policy Conference proceedings. Marketing and Public Policy Conference",
  ["marketing sci"] = "Marketing Science",
  ["markov process relat fields"] = "Markov processes and related fields",
  ["markov process related fields"] = "Markov Processes and Related Fields",
  ["marmara pharm j"] = "Marmara pharmaceutical journal",
  ["maroc med"] = "al-Maghrib al-ṭibbī. Maroc médical",
  ["marquette bus rev"] = "Marquette Business Review",
  ["marquette intellect prop law rev"] = "Marquette intellectual property law review",
  ["marquette law rev"] = "Marquette law review",
  ["marquette med rev"] = "Marquette medical review",
  ["marriage fam living"] = "Marriage and family living",
  ["marriage fam newsl"] = "Marriage and family newsletter",
  ["marriage fam rev"] = "Marriage & family review",
  ["mars (los angel)"] = "Mars (Los Angeles, Calif.)",
  ["mars chir"] = "Marseille chirurgical",
  ["mars med"] = "Marseille médical",
  ["marshall j med"] = "Marshall journal of medicine",
  ["martin gardner’s math games"] = "Martin Gardner’s Mathematical Games",
  ["marx perspect"] = "Marxist perspectives",
  ["masa rev"] = "MASA Review",
  ["mass commun soc"] = "Mass communication & society",
  ["mass gen laws annot mass"] = "Massachusetts general laws annotated : under arrangement of the official General laws of Massachusetts. Massachusetts",
  ["mass health j"] = "The Massachusetts health journal",
  ["mass hist soc proc"] = "Proceedings of the Massachusetts Historical Society. Massachusetts Historical Society",
  ["mass journal ment health"] = "Massachusetts journal of mental health",
  ["mass law rev"] = "Massachusetts law review",
  ["mass nurse"] = "The Massachusetts nurse",
  ["mass physician"] = "Massachusetts physician",
  ["mass rep mass supreme judic court"] = "Massachusetts reports : cases argued and determined in the Supreme Judicial Court of Massachusetts. Massachusetts. Supreme Judicial Court",
  ["mass rev"] = "The Massachusetts review",
  ["mass spectrom (tokyo)"] = "Mass spectrometry (Tokyo, Japan)",
  ["mass spectrom purif tech"] = "Mass spectrometry & purification techniques",
  ["mass spectrom rev"] = "Mass spectrometry reviews",
  ["mass spektrom"] = "Mass-spektrometria",
  ["mastering math finance"] = "Mastering Mathematical Finance",
  ["masterkey ind lore hist"] = "The Masterkey for Indian lore and history",
  ["masters abstr int"] = "Masters abstracts international",
  ["masters modern phys"] = "Masters of Modern Physics",
  ["mastozool neotrop"] = "Mastozoología neotropical",
  ["masui to sosei"] = "Masui to sosei. Hiroshima journal of anesthesia",
  ["masz elektr zesz probl"] = "Maszyny Elektryczne - Zeszyty Problemowe",
  ["masz przepływ"] = "Maszyny Przepływowe",
  ["mat apl comput"] = "Sociedade Brasiliera de Matemática Applicada e Computacional",
  ["mat bilten"] = "Matematichki Bilten. Bulletin Mathematique de la Societe des Mathematiciens de la Republique de Macedoine",
  ["mat complement"] = "Matematiche Complementari",
  ["mat contemp"] = "Matematica Contemporanea",
  ["mat cult soc riv unione mat ital (i)"] = "Matematica, Cultura e Societa. Rivista dell’Unione Matematica Italiana. Serie I",
  ["mat fiz anal geom"] = "Matematicheskaya Fizika, Analiz, Geometriya",
  ["mat fiz kompyut model"] = "Matematicheskaya Fizika i Kompyuternoe Modelirovanie. Mathematical Physics and Computer Simulation",
  ["mat fiz neliney n mekh"] = "Akademiya Nauk Ukrainy",
  ["mat lapok (ns)"] = "Matematikai Lapok. New Series",
  ["mat mat model"] = "Matematika i Matematicheskoe Modelirovanie",
  ["mat medd danske vid selsk"] = "Matematisk-fysiske Meddelelser",
  ["mat metody i fiz-mekh polya"] = "Akademiya Nauk Ukrainy",
  ["mat model"] = "Rossiiskaya Akademiya Nauk. Matematicheskoe Modelirovanie",
  ["mat prilozh"] = "\\cyr Matematika i ee Prilozheniya",
  ["mat res soc symp proc"] = "Materials Research Society Symposium Proceedings",
  ["mat sb"] = "Matematicheskii Sbornik",
  ["mat sem sntl"] = "Matematicky Seminar SNTL",
  ["mat ser a conf semin trab mat"] = "MAT. Serie A: Conferencias, Seminarios y Trabajos de Matematica",
  ["mat stos"] = "Roczniki Polskiego Towarzystwa Matematycznego",
  ["mat stud"] = "Matematichni Studii. Pratsi Lvivskogo Matematichnogo Tovaristva",
  ["mat teor igr prilozh"] = "Matematicheskaya Teoriya Igr i ee Prilozheniya",
  ["mat tr"] = "Matematicheskie Trudy",
  ["mat vesnik"] = "Matematichki Vesnik",
  ["mat vopr kriptografii"] = "Matematicheskie Voprosy Kriptografii. Mathematical Aspects of Cryptography",
  ["mat zametki"] = "Matematicheskie Zametki",
  ["match (mulh)"] = "Match (Mülheim an der Ruhr, Germany)",
  ["match commun math comput chem"] = "MATCH. Communications in Mathematical and in Computer Chemistry",
  ["match-commun math comput chem"] = "Match-Communications in Mathematical and in Computer Chemistry",
  ["matematiche (catania)"] = "Le Matematiche",
  ["matematika (johor)"] = "Matematika. Malaysian Journal of Industrial and Applied Mathematics",
  ["mater adv"] = "Materials advances",
  ["mater aust"] = "Materials Australia",
  ["mater bad inst meteor gospod wod ser hydrol oceanol"] = "Materiały badawcze. Instytut meteorologii i gospodarki wodnej. Seria, hydrologia i oceanologia",
  ["mater bevolkwiss"] = "Materialien zur Bevölkerungswissenschaft",
  ["mater charact"] = "Materials Characterization",
  ["mater chem front"] = "Materials chemistry frontiers",
  ["mater chem phys"] = "Materials Chemistry and Physics",
  ["mater circ econ"] = "Materials Circular Economy",
  ["mater constr"] = "Materiales de Construccion",
  ["mater corros"] = "Materials and Corrosion",
  ["mater cult"] = "Material culture",
  ["mater des"] = "Materials & design",
  ["mater des process commun"] = "Material Design & Processing Communications",
  ["mater discover"] = "Materials Discovery",
  ["mater emerging technol sustainability"] = "Materials and Emerging Technologies for Sustainability",
  ["mater eng (modena, italy)"] = "Materials Engineering (Modena, Italy)",
  ["mater eval"] = "Materials Evaluation",
  ["mater express"] = "Materials Express",
  ["mater focus"] = "Materials Focus",
  ["mater form mach tribol"] = "Materials Forming, Machining and Tribology",
  ["mater forum"] = "Materials Forum",
  ["mater high temp"] = "Materials at High Temperatures",
  ["mater hist bull"] = "Material history bulletin. Bulletin d'histoire de la culture matérielle",
  ["mater horiz"] = "Materials Horizons",
  ["mater horiz from nat nanomater"] = "Materials Horizons: From Nature to Nanomaterials",
  ["mater inf data sci"] = "Materials Informatics and Data Science",
  ["mater lett"] = "Materials Letters",
  ["mater lett: x"] = "Materials Letters: X",
  ["mater manag health care"] = "Materials management in health care",
  ["mater manuf processes"] = "Materials and Manufacturing Processes",
  ["mater med greca"] = "Materia medica greca",
  ["mater med nordmark"] = "Materia medica Nordmark",
  ["mater med pol"] = "Materia medica Polona. Polish journal of medicine and pharmacy",
  ["mater open"] = "Materials Open",
  ["mater org"] = "Materials and Organisms",
  ["mater perform"] = "Materials performance",
  ["mater perform charact"] = "Materials Performance and Characterization",
  ["mater performance"] = "Materials Performance",
  ["mater phys mech"] = "Materials Physics and Mechanics",
  ["mater pr antropol"] = "Materiały i prace antropologiczne",
  ["mater prot"] = "Materials Protection",
  ["mater quantum technol"] = "Materials for Quantum Technology",
  ["mater relig"] = "Material religion : the journal of objects, art and belief",
  ["mater renew sustain energy"] = "Materials for renewable and sustainable energy",
  ["mater renewable sustainable energy"] = "Materials for Renewable and Sustainable Energy",
  ["mater rep: energy"] = "Materials Reports: Energy",
  ["mater res"] = "Materials Research",
  ["mater res bull"] = "Materials Research Bulletin",
  ["mater res express"] = "Materials Research Express",
  ["mater res innovations"] = "Materials Research Innovations",
  ["mater res lett"] = "Materials Research Letters",
  ["mater res soc symp proc"] = "Materials Research Society Symposium Proceedings",
  ["mater res soc symp proc mater res soc"] = "Materials Research Society symposia proceedings. Materials Research Society",
  ["mater sci"] = "Materials Science",
  ["mater sci energy technol"] = "Materials science for energy technologies",
  ["mater sci eng"] = "Materials Science and Engineering",
  ["mater sci eng a struct mater"] = "Materials science & engineering. A, Structural materials : properties, microstructure and processing",
  ["mater sci eng b"] = "Materials Science and Engineering B",
  ["mater sci eng b solid state mater adv technol"] = "Materials science & engineering. B, Solid-state materials for advanced technology",
  ["mater sci eng b solid-state mater adv technol"] = "Materials Science & Engineering B: Solid-State Materials for Advanced Technology",
  ["mater sci eng c"] = "Materials Science and Engineering: C",
  ["mater sci eng c biomim mater sens syst"] = "Materials science & engineering. C, Biomimetic materials, sensors and systems",
  ["mater sci eng c biomim supramol syst"] = "Materials science & engineering. C, Biomimetic and supramolecular systems",
  ["mater sci eng c mater biol appl"] = "Materials science & engineering. C, Materials for biological applications",
  ["mater sci eng r rep"] = "Materials science & engineering. R, Reports : a review journal",
  ["mater sci eng, a"] = "Materials Science and Engineering A",
  ["mater sci eng, b"] = "Materials Science and Engineering B",
  ["mater sci eng, c"] = "Materials Science and Engineering C",
  ["mater sci eng, r"] = "Materials Science and Engineering Reports",
  ["mater sci eng: r: rep"] = "Materials Science and Engineering: R: Reports",
  ["mater sci forum"] = "Materials Science Forum",
  ["mater sci rep"] = "Materials Science Reports",
  ["mater sci res int"] = "Materials Science Research International",
  ["mater sci semicond process"] = "Materials Science in Semiconductor Processing",
  ["mater sci ser"] = "Materials Science Series",
  ["mater sci technol"] = "Materials Science and Technology",
  ["mater sci-pol"] = "Materials Science-Poland",
  ["mater sociomed"] = "Materia socio-medica",
  ["mater struct"] = "Materials and Structures",
  ["mater tech"] = "Materiaux & Techniques",
  ["mater technol"] = "Materials Technology",
  ["mater technol (n y n y)"] = "Materials technology (New York, N.Y.)",
  ["mater tehnol"] = "Materiali in Tehnologije",
  ["mater test"] = "Materials Testing",
  ["mater theory"] = "Materials Theory",
  ["mater ther"] = "Materia therapeutica : aus Klinik und Praxis für Klinik und Praxis",
  ["mater ther dr kutiak"] = "Materia therapeutica Dr. Kutiak : aus Klinik und Praxis für Klinik und Praxis",
  ["mater today"] = "Materials Today",
  ["mater today (kidlington)"] = "Materials today (Kidlington, England)",
  ["mater today adv"] = "Materials today. Advances",
  ["mater today bio"] = "Materials today. Bio",
  ["mater today chem"] = "Materials today. Chemistry",
  ["mater today commun"] = "Materials Today Communications",
  ["mater today electron"] = "Materials Today Electronics",
  ["mater today energy"] = "Materials today. Energy",
  ["mater today nano"] = "Materials today. Nano",
  ["mater today phys"] = "Materials Today Physics",
  ["mater today proc"] = "Materials today. Proceedings",
  ["mater today sustainability"] = "Materials Today Sustainability",
  ["mater today: proc"] = "Materials Today: Proceedings",
  ["mater trans"] = "Materials Transactions",
  ["mater trans, jim"] = "Materials Transactions, JIM",
  ["mater world"] = "Materials World",
  ["materialia (oxf)"] = "Materialia",
  ["materials (basel)"] = "Materials (Basel, Switzerland)",
  ["materials sci res"] = "Materials Science Research",
  ["materialwiss werkstofftech"] = "Materialwissenschaft und Werkstofftechnik",
  ["matern child health j"] = "Maternal and child health journal",
  ["matern child nurs j"] = "Maternal-child nursing journal",
  ["matern child nutr"] = "Maternal & child nutrition",
  ["matern fetal med"] = "Maternal-fetal medicine (Wolters Kluwer Health, Inc.)",
  ["matern health neonatol perinatol"] = "Maternal health, neonatology and perinatology",
  ["matern infanc (sao paulo)"] = "Maternidade e infância; arquivos médicos-sociais",
  ["matern infanz (1926)"] = "Maternita e infanzia",
  ["matern pediatr nutr"] = "Maternal and pediatric nutrition",
  ["materwiss werksttech"] = "Materialwissenschaft und Werkstofftechnik",
  ["math anal appl"] = "Mathematical Analysis and Its Applications",
  ["math anal tech appl eng"] = "Mathematical and Analytical Techniques with Applications to Engineering",
  ["math ann"] = "Mathematische Annalen",
  ["math anwend"] = "Mathematik fur Anwendungen",
  ["math appl"] = "Mathematics for applications",
  ["math appl (berlin)"] = "Mathematiques & Applications (Berlin)",
  ["math appl (brno)"] = "Mathematics for Applications",
  ["math appl (china ser)"] = "Mathematics and its Applications (China Series)",
  ["math appl (soviet ser)"] = "Mathematics and its Applications (Soviet Series)",
  ["math appl (warsaw)"] = "Mathematica Applicanda. Matematyka Stosowana",
  ["math appl (wuhan)"] = "Mathematica Applicata. Yingyong Shuxue",
  ["math appl model engin soc sci"] = "Mathematics and Its Applications: Modelling, Engineering, and Social Sciences",
  ["math balkanica (ns)"] = "Mathematica Balkanica. New Series",
  ["math beisp"] = "Mathematik in Beispielen",
  ["math biosci"] = "Mathematical Biosciences",
  ["math biosci eng"] = "Mathematical biosciences and engineering : MBE",
  ["math biosci inst lect ser stoch biol syst"] = "Mathematical Biosciences Institute Lecture Series. Stochastics in Biological Systems",
  ["math biosci interact"] = "Mathematics and Biosciences in Interaction",
  ["math biosci subser"] = "Mathematical Biosciences Subseries",
  ["math bohem"] = "Academy of Sciences of the Czech Republic. Mathematical Institute. Mathematica Bohemica",
  ["math built environ"] = "Mathematics and the Built Environment",
  ["math chem ser"] = "Mathematical Chemistry Series",
  ["math clim weather forecast"] = "Mathematics of Climate and Weather Forecasting",
  ["math clim weather forecasting"] = "Mathematics of Climate and Weather Forecasting",
  ["math cognit"] = "Mathematical Cognition",
  ["math commun"] = "Mathematical Communications",
  ["math comp"] = "Mathematics of Computation",
  ["math comput"] = "Mathematics of Computation",
  ["math comput appl"] = "Mathematical and Computational Applications",
  ["math comput biol numer anal"] = "Mathematical and Computational Biology and Numerical Analysis",
  ["math comput geom data"] = "Mathematics, Computation and Geometry of Data",
  ["math comput model"] = "Mathematical and computer modelling",
  ["math comput model dyn syst"] = "Mathematical and Computer Modelling of Dynamical Systems",
  ["math comput modell"] = "Mathematical and Computer Modelling",
  ["math comput modell dyn syst"] = "Mathematical and Computer Modelling of Dynamical Systems",
  ["math comput modelling"] = "Mathematical and Computer Modelling",
  ["math comput sci"] = "Mathematics in Computer Science",
  ["math comput simul"] = "Mathematics and computers in simulation",
  ["math comput simulation"] = "Mathematics and Computers in Simulation",
  ["math concepts methods sci eng"] = "Mathematical Concepts and Methods in Science and Engineering",
  ["math control relat fields"] = "Mathematical Control and Related Fields",
  ["math control signals syst"] = "Mathematics of Control Signals and Systems",
  ["math control signals systems"] = "Mathematics of Control, Signals, and Systems",
  ["math cult arts"] = "Mathematics, Culture, and the Arts",
  ["math data"] = "Mathematics of Data",
  ["math ed"] = "Mathematical Education",
  ["math ed (siwan)"] = "The Mathematics Education",
  ["math ed lib"] = "Mathematics Education Library",
  ["math eng"] = "Mathematics in Engineering",
  ["math eng indust"] = "Mathematical Engineering in Industry",
  ["math eng manuf manag sci"] = "Mathematical Engineering, Manufacturing, and Management Sciences",
  ["math financ"] = "Mathematical finance",
  ["math financ econ"] = "Mathematics and Financial Economics",
  ["math finance"] = "Mathematical Finance",
  ["math forum"] = "Mathematical Forum",
  ["math found comput"] = "Mathematical Foundations of Computing",
  ["math found comput vis mach learn robot"] = "Mathematical Foundations for Computer Vision, Machine Learning, and Robotics",
  ["math gaz"] = "Mathematical Gazette",
  ["math geol"] = "Mathematical Geology",
  ["math geosci"] = "Mathematical Geosciences",
  ["math horiz"] = "Math Horizons",
  ["math ind"] = "Mathematics in Industry",
  ["math ind (phila)"] = "Mathematics in Industry (Philadelphia)",
  ["math ind (tokyo)"] = "Mathematics for Industry (Tokyo)",
  ["math ind case stud"] = "Mathematics-in-industry case studies",
  ["math inequal appl"] = "Mathematical Inequalities & Applications",
  ["math inform sci humaines"] = "Mathématiques Informatique et Sciences Humaines",
  ["math inform sci rep ser b"] = "Mathematical and Information Sciences Report. Series B",
  ["math ingen naturwiss"] = "Mathematik für Ingenieure und Naturwissenschaftler",
  ["math ingen naturwiss okonom sonstige anwendungsorient berufe"] = "Mathematik fur Ingenieure, Naturwissenschaftler, Okonomen und Sonstige Anwendungsorientierte Berufe",
  ["math intell"] = "Mathematical Intelligencer",
  ["math intelligencer"] = "The Mathematical Intelligencer",
  ["math j"] = "The Mathematica journal",
  ["math j ibaraki univ"] = "Mathematical Journal of Ibaraki University",
  ["math j okayama univ"] = "Mathematical Journal of Okayama University",
  ["math j toyama univ"] = "Mathematics Journal of Toyama University",
  ["math japon"] = "Mathematica Japonica",
  ["math kompakt"] = "Mathematik Kompakt",
  ["math kontext"] = "Mathematik im Kontext",
  ["math lect peking univ"] = "Mathematical Lectures from Peking University",
  ["math leitfäden"] = "Mathematische Leitfäden",
  ["math log q"] = "Mathematical Logic Quarterly",
  ["math logic found"] = "Mathematical Logic and Foundations",
  ["math logic quart"] = "Mathematical Logic Quarterly",
  ["math mag"] = "Mathematics Magazine",
  ["math mech complex syst"] = "Mathematics and Mechanics of Complex Systems",
  ["math mech solids"] = "Mathematics and Mechanics of Solids",
  ["math med biol"] = "Mathematical medicine and biology : a journal of the IMA",
  ["math medley"] = "Mathematical Medley",
  ["math methods appl sci"] = "Mathematical Methods in the Applied Sciences",
  ["math methods econ finance"] = "Mathematical Methods in Economics and Finance",
  ["math methods model"] = "Mathematical Methods and Modeling",
  ["math methods oper res"] = "Mathematical Methods of Operations Research",
  ["math methods oper res (heidelb)"] = "Mathematical methods of operations research (Heidelberg, Germany)",
  ["math methods operations res"] = "Mathematical Methods of Operations Research",
  ["math methods stat"] = "Mathematical Methods of Statistics",
  ["math methods statist"] = "Mathematical Methods of Statistics",
  ["math mind"] = "Mathematics in Mind",
  ["math model"] = "Mathematical Modeling",
  ["math model anal"] = "Mathematical Modelling and Analysis",
  ["math model comput"] = "Mathematical Modeling and Computation",
  ["math model nat phenom"] = "Mathematical modelling of natural phenomena",
  ["math model systems"] = "Mathematical Modelling of Systems",
  ["math model theory appl"] = "Mathematical Modelling: Theory and Applications",
  ["math modeling comput experiment"] = "Mathematical Modeling and Computational Experiment",
  ["math modell"] = "Mathematical Modelling",
  ["math modell anal"] = "Mathematical Modelling and Analysis",
  ["math modell nat phenom"] = "Mathematical Modelling of Natural Phenomena",
  ["math modell theory appl"] = "Mathematical Modelling: Theory and Applications",
  ["math modelling sci comput"] = "Mathematical Modelling and Scientific Computing",
  ["math models comput simul"] = "Mathematical Models and Computer Simulations",
  ["math models method appl sci"] = "Mathematical Models and Methods in Applied Sciences",
  ["math models methods appl sci"] = "Mathematical Models and Methods in Applied Sciences",
  ["math montisnigri"] = "Mathematica Montisnigri. Matematika Tsrne Gore",
  ["math morav"] = "Mathematica Moravica",
  ["math nachr"] = "Mathematische Nachrichten",
  ["math nat sci"] = "Mathematics and Natural Sciences",
  ["math naturwiss unterr"] = "Mathematische und Naturwissenschaftliche Unterricht",
  ["math natwiss"] = "Mathematik und Naturwissenschaften",
  ["math neurosci appl"] = "Mathematical Neuroscience and Applications",
  ["math newsl"] = "Mathematics Newsletter",
  ["math notae"] = "Mathematicae Notae",
  ["math notes"] = "Mathematical Notes",
  ["math numer sin"] = "Mathematica Numerica Sinica. Jisuan Shuxue",
  ["math numer sinica"] = "Mathematica Numerica Sinica",
  ["math olymp ser"] = "Mathematical Olympiad Series",
  ["math online first collect"] = "Mathematics Online First Collections",
  ["math open"] = "Mathematics Open",
  ["math oper res"] = "Mathematics of Operations Research",
  ["math optim wirtschmath/math optim economath"] = "Mathematische Optimierung und Wirtschaftsmathematik/Mathematical Optimization and Economathematics",
  ["math pannon"] = "Mathematica Pannonica",
  ["math pannon (n s)"] = "Mathematica Pannonica. New Series",
  ["math phys anal geom"] = "Mathematical Physics Analysis and Geometry",
  ["math phys appl math"] = "Mathematical Physics and Applied Mathematics",
  ["math phys electron j"] = "Mathematical Physics Electronic Journal",
  ["math phys sci technol"] = "Mathematics and Physics for Science and Technology",
  ["math phys stud"] = "Mathematical Physics Studies",
  ["math planet earth"] = "Mathematics of Planet Earth",
  ["math popul stud"] = "Mathematical population studies",
  ["math practice theory"] = "Mathematics in Practice and Theory",
  ["math probl comput sci"] = "Mathematical Problems of Computer Science",
  ["math probl eng"] = "Mathematical Problems in Engineering",
  ["math proc cambridge philos soc"] = "Mathematical Proceedings of the Cambridge Philosophical Society",
  ["math proc r ir acad"] = "Mathematical Proceedings of the Royal Irish Academy",
  ["math program"] = "Mathematical Programming",
  ["math program comput"] = "Mathematical Programming Computation",
  ["math programming"] = "Mathematical Programming",
  ["math rep (bucur)"] = "Mathematical Reports (Bucuresti)",
  ["math res"] = "Mathematical Research",
  ["math res dev"] = "Mathematics Research Developments",
  ["math res dev ser"] = "Mathematics Research Developments Series",
  ["math res lett"] = "Mathematical Research Letters",
  ["math res rep"] = "Mathematics Research Reports",
  ["math scand"] = "Mathematica Scandinavica",
  ["math schulerbucherei"] = "Mathematische Schulerbucherei",
  ["math sci"] = "The Mathematical Scientist",
  ["math sci (springer)"] = "Mathematical Sciences",
  ["math sci appl e-notes"] = "Mathematical Sciences & Applications E-Notes. An International Electronic Journal of Mathematics",
  ["math sci eng"] = "Mathematics in Science and Engineering",
  ["math sci prof dir"] = "Mathematical Sciences Professional Directory",
  ["math sci ref ser"] = "Mathematical Sciences Reference Series",
  ["math sci res hot-line"] = "Mathematical Sciences Research Hot-Line",
  ["math sci res inst publ"] = "Mathematical Sciences Research Institute Publications",
  ["math semesterber"] = "Mathematische Semesterberichte",
  ["math slovaca"] = "Mathematica Slovaca",
  ["math soc sci"] = "Mathematical social sciences",
  ["math social sci"] = "Mathematical Social Sciences",
  ["math stat learn"] = "Mathematical Statistics and Learning",
  ["math stat ser"] = "Mathematics and Statistics Series",
  ["math struct comput sci"] = "Mathematical Structures in Computer Science",
  ["math structures comput sci"] = "Mathematical Structures in Computer Science. A Journal in the Applications of Categorical, Algebraic and Geometric Methods in Computer Science",
  ["math stud (tartu)"] = "Mathematics Studies (Tartu)",
  ["math stud monogr ser"] = "Mathematical Studies Monograph Series",
  ["math student"] = "The Mathematics Student",
  ["math studienanfanger"] = "Mathematik fur Studienanfanger",
  ["math study resour"] = "Mathematic Study Resources",
  ["math super spec"] = "Mathematiques Superieures et Speciales",
  ["math surveys monogr"] = "Mathematical Surveys and Monographs",
  ["math teach probl-based approach"] = "Mathematics for Teaching: A Problem-Based Approach",
  ["math teach:  learn teach prek-12"] = "Mathematics Teacher: Learning and Teaching PreK-12",
  ["math textb sci eng"] = "Mathematics Textbooks for Science and Engineering",
  ["math texte"] = "Mathematische Texte",
  ["math theory appl (changsha)"] = "Mathematical Theory and Applications",
  ["math think learn"] = "Mathematical thinking & learning",
  ["math thinking learn"] = "Mathematical Thinking and Learning",
  ["math today"] = "Mathematics Today",
  ["math today (ahmedabad)"] = "Mathematics Today (Ahmedabad)",
  ["math today (southend-on-sea)"] = "Mathematics Today",
  ["math top"] = "Mathematical Topics",
  ["math vis"] = "Mathematics and visualization",
  ["math wirtschaftswiss"] = "Mathematik und Wirtschaftswissenschaft",
  ["math world"] = "Mathematical World",
  ["math z"] = "Mathematische Zeitschrift",
  ["math, graeca antiq"] = "Mathematica Graeca Antiqua",
  ["math- ind case stud"] = "Mathematics-in-Industry Case Studies",
  ["math-astronom blätter—nf"] = "Mathematisch-Astronomische Blätter—Neue Folge",
  ["math-for-ind (mi) lect note ser"] = "Math-for-Industry (MI) Lecture Note Series",
  ["mathematica (cluj)"] = "Académie Roumaine",
  ["mathematics (basel)"] = "Mathematics (Basel, Switzerland)",
  ["mathematisch centrum, afdeling mathematische besliskunde in samenwerking met het interuniversitair instituut bedrijfskun"] = "Mathematisch Centrum, Afdeling Mathematische Besliskunde in samenwerking met het Interuniversitair Instituut Bedrijfskunde Delft/Rotterdam",
  ["mathematische lehrbucher und monographien, herausgegeben von der deutschen akademie der wissenschaften zu berlin, fors"] = "Mathematische Lehrbucher und Monographien, herausgegeben von der Deutschen Akademie der Wissenschaften zu Berlin, Forschungsinstitut fur Mathematik. I. Abteilung, Mathematische Lehrbucher",
  ["mathews j case rep"] = "Mathews journal of case reports",
  ["mathews j ophthalmol"] = "Mathews journal of ophthalmology",
  ["maths in action"] = "MathematicS in Action",
  ["mathware soft comput"] = "Mathware & Soft Computing",
  ["matimyas mat"] = "Matimyas Matematika",
  ["matimyás mat"] = "Matimyás Matematika",
  ["matrix biol"] = "Matrix Biology",
  ["matrix biol plus"] = "Matrix biology plus",
  ["matrix book ser"] = "MATRIX Book Series",
  ["matrix suppl"] = "Matrix (Stuttgart, Germany). Supplement",
  ["matter radiat at extremes"] = "Matter and Radiation at Extremes",
  ["matter radiat extremes"] = "Matter and Radiation at Extremes",
  ["matters (zur)"] = "Matters",
  ["matters sel"] = "Matters select",
  ["matér carte géol suisse"] = "Matériaux pour la Carte Géologique de la Suisse, n.s.",
  ["matér levé géobot suisse"] = "Matériaux pour le levé géobotanique de la Suisse",
  ["max planck res libr hist dev knowl, textb"] = "Max Planck Research Library for the History and Development of Knowledge. Textbooks",
  ["maxillofac plast reconstr surg"] = "Maxillofacial plastic and reconstructive surgery",
  ["maximum entropy bayesian methods"] = "Maximum entropy and Bayesian methods. International Workshop on Maximum Entropy and Bayesian Methods",
  ["maxwell rev"] = "Maxwell review",
  ["mayo clin health lett"] = "Mayo Clinic health letter (English ed.)",
  ["mayo clin proc"] = "Mayo Clinic proceedings",
  ["mayo clin proc innov qual outcomes"] = "Mayo Clinic proceedings. Innovations, quality & outcomes",
  ["mayo clin womens healthsource"] = "Mayo Clinic women's healthsource",
  ["mbm met bull mon"] = "MBM Metal Bulletin Monthly",
  ["mcgeorge law rev"] = "McGeorge law review",
  ["mcgill dent rev"] = "The McGill dental review",
  ["mcgill j educ"] = "McGill journal of education",
  ["mcgill j law health"] = "McGill journal of law and health : MJLH = Revue de droit et santé de McGill : RDSM",
  ["mcgill j med"] = "McGill journal of medicine : MJM : an international forum for the advancement of medical sciences by students",
  ["mcgill law j"] = "McGill law journal. Revue de droit de McGill",
  ["mcgill med j"] = "McGill medical journal",
  ["mcgill news"] = "The McGill news",
  ["mcgill queens hannah inst stud hist med health soc"] = "McGill-Queen's/Hannah Institute studies in the history of medicine, health, and society",
  ["mcgill sci undergrad res j"] = "McGill Science undergraduate research journal : MSURJ",
  ["mcgill-queen stud hist ideas"] = "McGill-Queen’s Studies in the History of Ideas",
  ["mcgraw hills med health"] = "McGraw-Hill's medicine & health",
  ["mcgraw hills wash rep med health"] = "McGraw-Hill's Washington report on medicine & health",
  ["mcgraw-hill ser indust engrg management sci"] = "McGraw-Hill Series in Industrial Engineering and Management Science",
  ["mcgregor clin bull"] = "McGregor Clinic bulletin",
  ["mch news pac"] = "MCH news PAC",
  ["mckinneys consol laws n y annot n y state"] = "McKinney's Consolidated laws of New York annotated : with annotations from state and federal courts and state agencies. New York (State)",
  ["mcn am j matern child nurs"] = "MCN. The American journal of maternal child nursing",
  ["md (chic)"] = "MD (Chicago, Ill.)",
  ["md advis"] = "MD advisor : a journal for New Jersey medical community",
  ["md anderson solid tu"] = "Md Anderson Solid Tumor Oncology Series",
  ["md bar j"] = "The Maryland bar journal",
  ["md comput"] = "M.D. computing : computers in medical practice",
  ["md fam dr"] = "The Maryland family doctor",
  ["md geneal soc bull"] = "Maryland Genealogical Society bulletin",
  ["md health bull"] = "Maryland health bulletin",
  ["md hist"] = "The Maryland historian",
  ["md hist mag"] = "Maryland historical magazine",
  ["md j contemp leg issues"] = "Maryland journal of contemporary legal issues",
  ["md j int law trade"] = "Maryland journal of international law and trade",
  ["md law forum"] = "Maryland law forum",
  ["md law rev"] = "Maryland law review (Baltimore, Md. : 1936)",
  ["md med"] = "Maryland medicine : MM : a publication of MEDCHI, the Maryland State Medical Society",
  ["md med j"] = "Maryland medical journal (Baltimore, Md. : 1985)",
  ["md nurs news"] = "Maryland nursing news",
  ["md nurse"] = "The Maryland nurse",
  ["md state med j"] = "Maryland state medical journal",
  ["mde manage decis econ"] = "Managerial and decision economics : MDE",
  ["mdm policy pract"] = "MDM policy & practice",
  ["mead johnson symp perinat dev med"] = "Mead Johnson Symposium on Perinatal and Developmental Medicine",
  ["meas control"] = "Measurement and Control",
  ["meas eval couns dev"] = "Measurement and evaluation in counseling and development : the official publication of the Association for Measurement and Evaluation in Counseling and Development",
  ["meas phys educ exerc sci"] = "Measurement in physical education and exercise science",
  ["meas phys educ exercise sci"] = "Measurement in Physical Education and Exercise Science",
  ["meas sci technol"] = "Measurement Science and Technology",
  ["meas tech"] = "Measurement Techniques",
  ["meas: food"] = "Measurement: Food",
  ["meas: interdiscip res perspect"] = "Measurement: Interdisciplinary Research and Perspectives",
  ["meas: sens"] = "Measurement: Sensors",
  ["measurement ( mahwah n j)"] = "Measurement : interdisciplinary research and perspectives",
  ["measurement (lond)"] = "Measurement : journal of the International Measurement Confederation",
  ["meat sci"] = "Meat science",
  ["mec dent"] = "El Mecanico dental",
  ["mec ind"] = "Mecanique & Industries",
  ["mech adv mater mod process"] = "Mechanics of advanced materials and modern processes",
  ["mech adv mater mod processes"] = "Mechanics of Advanced Materials and Modern Processes",
  ["mech adv mater struct"] = "Mechanics of Advanced Materials and Structures",
  ["mech ageing dev"] = "Mechanisms of ageing and development",
  ["mech based des struct mach"] = "Mechanics Based Design of Structures and Machines",
  ["mech chem biosyst"] = "Mechanics & chemistry of biosystems : MCB",
  ["mech cohesive-frict mater"] = "Mechanics of Cohesive-frictional Materials",
  ["mech compos mater"] = "Mechanics of Composite Materials",
  ["mech compos mater struct"] = "Mechanics of Composite Materials and Structures",
  ["mech dev"] = "Mechanisms of development",
  ["mech elastic inelastic solids"] = "Mechanics of Elastic and Inelastic Solids",
  ["mech eng"] = "Mechanical engineering (New York, N.Y. : 1919)",
  ["mech eng ser"] = "Mechanical Engineering Series",
  ["mech eng solid mech ser"] = "Mechanical Engineering and Solid Mechanics Series",
  ["mech engrg res stud"] = "Mechanical Engineering Research Studies",
  ["mech engrg ser"] = "Mechanical Engineering Series",
  ["mech ind"] = "Mechanics & Industry",
  ["mech mach sci"] = "Mechanisms and Machine Science",
  ["mech mach theory"] = "Mechanism and Machine Theory",
  ["mech mater"] = "Mechanics of Materials",
  ["mech math methods ser ii: thermal stresses"] = "Mechanics and Mathematical Methods. A Series of Handbooks. Series II: Thermal Stresses",
  ["mech of fracture"] = "Mechanics of Fracture",
  ["mech res comm"] = "Mechanics Research Communications",
  ["mech res commun"] = "Mechanics Research Communications",
  ["mech sci"] = "Mechanical Sciences",
  ["mech soft mater"] = "Mechanics of soft materials",
  ["mech solids"] = "Mechanics of Solids",
  ["mech structures mach"] = "Mechanics of Structures and Machines",
  ["mech syst sig process"] = "Mechanical Systems and Signal Processing",
  ["mech syst signal process"] = "Mechanical systems and signal processing",
  ["mech time depend mater"] = "Mechanics of time-dependent materials",
  ["mech time-depend mater"] = "Mechanics of Time Dependent Materials",
  ["mech werkst konstr bauwes"] = "Mechanik, Werkstoffe und Konstruktion im Bauwesen",
  ["mechatron syst control"] = "Mechatronic Systems and Control",
  ["mechatronics (oxf)"] = "Mechatronics : the science of intelligent machines",
  ["med (n y)"] = "Med (New York, N.Y.)",
  ["med 2 0"] = "Medicine 2.0",
  ["med abstr"] = "Medical abstracts",
  ["med actual"] = "Medicamentos de actualidad. Drugs of today",
  ["med acupunct"] = "Medical acupuncture",
  ["med aeronaut"] = "La Médecine aéronautique",
  ["med aeronaut spat"] = "Médecine aéronautique et spatiale",
  ["med aff"] = "Medical affairs",
  ["med afr noire"] = "Médecine d'Afrique noire",
  ["med akad z"] = "Medit︠s︡inskiĭ akademicheskiĭ zhurnal",
  ["med ann dist columbia"] = "The Medical annals of the District of Columbia",
  ["med annu"] = "The Medical annual",
  ["med anthropol"] = "Medical anthropology",
  ["med anthropol newsl"] = "Medical anthropology newsletter",
  ["med anthropol q"] = "Medical anthropology quarterly",
  ["med anthropol theory"] = "Medicine anthropology theory",
  ["med arch"] = "Medical archives (Sarajevo, Bosnia and Herzegovina)",
  ["med arh"] = "Medicinski arhiv",
  ["med armees"] = "Medécine et armées",
  ["med aromat plant sci biotechnol"] = "Medicinal and aromatic plant science and biotechnology",
  ["med aromat plants (los angel)"] = "Medicinal & aromatic plants",
  ["med art"] = "Medical art",
  ["med arts sci"] = "Medical arts and sciences",
  ["med aspects hum sex"] = "Medical aspects of human sexuality",
  ["med audio vis"] = "Médecine et audio vision",
  ["med bild dienst"] = "Medizinischer Bild-Dienst",
  ["med biochem"] = "Medical biochemistry",
  ["med biol"] = "Medical biology",
  ["med biol eng"] = "Medical & biological engineering",
  ["med biol eng comput"] = "Medical & biological engineering & computing",
  ["med biol illus"] = "Medical & biological illustration",
  ["med bogota colomb"] = "Medicina (Universidad de los Andes)",
  ["med bohringer (overseas)"] = "Medico Boehringer: overseas ed",
  ["med bookm hist"] = "Medical bookman and historian",
  ["med br (lond)"] = "Medica Britannica",
  ["med bull (ann arbor)"] = "Medical bulletin (Ann Arbor, Mich.)",
  ["med bull (n y)"] = "The Medical bulletin",
  ["med bull harrisburg polyclin hosp"] = "Medical bulletin. Harrisburg Polyclinic Hospital",
  ["med bull istanbul med fac"] = "Medical bulletin of Istanbul Medical Faculty",
  ["med bull st louis univ"] = "Medical bulletin. St. Louis University",
  ["med bull u s army army 1st"] = "Medical bulletin. United States. Army. Army, 1st",
  ["med bull u s army eur command med div"] = "Medical bulletin. United States. Army. European Command. Medical Division",
  ["med bull u s army force europe theater off theater chief surg"] = "Medical bulletin. United States. Army. Forces in the European Theater. Office of the Theater Chief Surgeon",
  ["med bull u s army mediterr theater oper"] = "Medical bulletin. United States. Army. Mediterranean Theater of Operations",
  ["med bull uganda"] = "Medical Bulletin Of Uganda",
  ["med bull us"] = "The Medical bulletin of the U. S. Army Far East",
  ["med bull us army eur"] = "Medical bulletin of the U. S. Army, Europe. United States. Army, Europe. Medical Division",
  ["med bull vet adm"] = "Medical bulletin. United States. Veterans Administration. Department of Medicine and Surgery",
  ["med cannabis cannabinoids"] = "Medical cannabis and cannabinoids",
  ["med care"] = "Medical care",
  ["med care res rev"] = "Medical care research and review : MCRR",
  ["med care rev"] = "Medical care review",
  ["med case rep (wilmington)"] = "Medical case reports (Wilmington, Del.)",
  ["med case rep short rev"] = "Medical case reports and short reviews",
  ["med chem"] = "Medicinal chemistry (Shāriqah (United Arab Emirates))",
  ["med chem (los angeles)"] = "Medicinal chemistry",
  ["med chem (sharjah, united arab emirates)"] = "Medicinal Chemistry (Sharjah, United Arab Emirates)",
  ["med chem res"] = "Medicinal Chemistry Research",
  ["med chir dig"] = "Médecine & chirurgie digestives",
  ["med chir j"] = "Medico-chirurgical journal : or, Quarterly register of medical and surgical science",
  ["med chir j rev"] = "The Medico-chirurgical journal and review",
  ["med chir rev"] = "The Medico-chirurgical review",
  ["med chir rev j med sci anal ser"] = "The Medico-chirurgical review and journal of medical science. Analytical series",
  ["med chir trans"] = "Medico-chirurgical transactions",
  ["med cir (bogota)"] = "Medicina y cirugía",
  ["med cir farm"] = "Medicina, cirurgia, farmácia",
  ["med cir guerra"] = "Medicina y cirugía de guerra",
  ["med claims manag"] = "Medical claims management",
  ["med clin (barc)"] = "Medicina clínica",
  ["med clin (engl ed)"] = "Medicina clínica (English ed.)",
  ["med clin north am"] = "The Medical clinics of North America",
  ["med clin res"] = "Medical & clinical research",
  ["med clin sper"] = "Medicina clinica e sperimentale",
  ["med colon"] = "La Medicina colonial",
  ["med comment"] = "The Medical comment",
  ["med commun (n j)"] = "Medical communications",
  ["med comput vis (2013)"] = "Medical computer vision : large data in medical imaging : third international MICCAI workshop, MCV 2013, Nagoya, Japan, September 26, 2013 : revised selected papers. MCV (Workshop) (3rd : 2013 : Nagoya-shi, Japan)",
  ["med comput vis bayesian graph models biomed imaging (2016)"] = "Medical computer vision and Bayesian and graphical models for biomedical imaging : MICCAI 2016 international workshop, MCV and BAMBI, Athens, Greece, October 21, 2016 : revised selected papers",
  ["med confl surviv"] = "Medicine, conflict, and survival",
  ["med contact (bussum)"] = "Medisch contact",
  ["med contemp"] = "Medicina contemporânea (Lisbon, Portugal)",
  ["med counterpoint"] = "Medical counterpoint",
  ["med crit psychol j"] = "Medical critic and psychological journal",
  ["med cutan ibero lat am"] = "Medicina cutánea ibero-latino-americana",
  ["med decis making"] = "Medical decision making : an international journal of the Society for Medical Decision Making",
  ["med dent bull"] = "The Medical and dental bulletin",
  ["med dent j"] = "Medical/dental journal",
  ["med deporte trab"] = "Medicina del deporte y del trabajo",
  ["med des mater"] = "Medical design and material",
  ["med desarrollo"] = "Medicina y desarrollo",
  ["med device technol"] = "Medical device technology",
  ["med devices (auckl)"] = "Medical devices (Auckland, N.Z.)",
  ["med devices sens"] = "Medical devices & sensors",
  ["med devices: evidence res"] = "Medical Devices: Evidence and Research",
  ["med dig"] = "Medical digest",
  ["med dimens"] = "Medical dimensions",
  ["med dir chic"] = "Medical directory of Chicago (including Cook Co.)",
  ["med dok"] = "Medizinische Dokumentation",
  ["med dosim"] = "Medical dosimetry : official journal of the American Association of Medical Dosimetrists",
  ["med dosw"] = "Medycyna doświadczalna",
  ["med dosw mikrobiol"] = "Medycyna doświadczalna i mikrobiologia",
  ["med drug discov"] = "Medicine in drug discovery",
  ["med drug discovery"] = "Medicine in Drug Discovery",
  ["med econ"] = "Medical economics",
  ["med educ"] = "Medical education",
  ["med educ (chicago ill)"] = "Medical education (Chicago, Ill.)",
  ["med educ online"] = "Medical education online",
  ["med egypte"] = "Médecine d'Egypte",
  ["med electron"] = "Medical electronics",
  ["med electron biol eng"] = "Medical electronics & biological engineering",
  ["med electron microsc"] = "Medical electron microscopy : official journal of the Clinical Electron Microscopy Society of Japan",
  ["med emory"] = "Medicine at Emory",
  ["med encount"] = "Medical encounter",
  ["med eng phys"] = "Medical engineering & physics",
  ["med entomol zool"] = "Medical Entomology and Zoology",
  ["med epigenet"] = "Medical epigenetics",
  ["med ernahr"] = "Medizin und Ernährung",
  ["med esp"] = "Medicina española",
  ["med ethics"] = "Medical ethics : journal of Forum for Medical Ethics Society",
  ["med ethics (burlingt mass)"] = "Medical ethics (Burlington, Mass.)",
  ["med ethics (burlingt, mass)"] = "Medical ethics (Burlington, Mass.)",
  ["med ethics pract royal pharm soc g b"] = "Medicines, ethics and practice : a guide for pharmacists. Royal Pharmaceutical Society of Great Britain",
  ["med etika bioet"] = "Medicínska etika a bioetika : časopis Ústavu medicínskej etiky a bioetiky = Medical ethics & bioethics : journal of the Institute of Medical Ethics & Bioethics",
  ["med exp int j exp med"] = "Medicina experimentalis : International journal of experimental medicine",
  ["med facts obs"] = "Medical facts and observations",
  ["med farm meded"] = "Medisch-farmaceutische mededelingen",
  ["med fis rehabil"] = "Medicina física y rehabilitación",
  ["med foren tidskr"] = "Medicinska föreningens tidskrift",
  ["med forum"] = "Medicinsk Forum",
  ["med fr (bombay)"] = "La Médecine en France",
  ["med gas res"] = "Medical gas research",
  ["med gen"] = "Medizinische Genetik",
  ["med gen fr"] = "Le Médecin généraliste de France",
  ["med genet"] = "Medizinische Genetik : Mitteilungsblatt des Berufsverbandes Medizinische Genetik e.V",
  ["med geogr herit"] = "Medicine's geographic heritage",
  ["med ges"] = "Medizin und Gesellschaft",
  ["med ges gesch"] = "Medizin, Gesellschaft, und Geschichte : Jahrbuch des Instituts für Geschichte der Medizin der Robert Bosch Stiftung",
  ["med ges gesch beih"] = "Medizin, Gesellschaft, und Geschichte. Beiheft : Jahrbuch des Instituts für Geschichte der Medizin der Robert Bosch Stiftung",
  ["med gesch kult"] = "Medizin in Geschichte und Kultur",
  ["med glas"] = "Medicinski glasnik",
  ["med glas (zenica)"] = "Medicinski glasnik : official publication of the Medical Association of Zenica-Doboj Canton, Bosnia and Herzegovina",
  ["med glob surviv"] = "Medicine and global survival : M & GS",
  ["med group manage"] = "Medical group management",
  ["med group manage j"] = "Medical group management journal",
  ["med gynaecol androl sociol"] = "Medical gynaecology, andrology, and sociology",
  ["med gynaecol sociol"] = "Medical gynaecology and sociology",
  ["med health"] = "Medicine & health",
  ["med health care philos"] = "Medicine, health care, and philosophy",
  ["med health r i"] = "Medicine and health, Rhode Island",
  ["med herit"] = "Medical heritage",
  ["med hist"] = "Medical history",
  ["med hist (barc)"] = "Medicina e historia",
  ["med hist aust"] = "Medical history Australia",
  ["med hist suppl"] = "Medical history. Supplement",
  ["med hoje"] = "Medicina De Hoje",
  ["med humanit"] = "Medical humanities",
  ["med humanit rev"] = "Medical humanities review",
  ["med hyg (geneve)"] = "Médecine et hygiène",
  ["med hypotheses"] = "Medical hypotheses",
  ["med hypotheses res"] = "Medical hypotheses and research : MHR",
  ["med hypothesis discov innov ophthalmol"] = "Medical hypothesis, discovery & innovation ophthalmology journal",
  ["med illus"] = "Medicine illustrated",
  ["med image anal"] = "Medical image analysis",
  ["med image underst anal conf (2017)"] = "Medical Image Understanding and Analysis : 21st Annual Conference, MIUA 2017, Edinburgh, UK, July 11-13, 2017, Proceedings. Medical Image Understanding and Analysis (Conference) (21st : 2017 : Edinburgh, Scotland)",
  ["med imaging augment real (2016)"] = "Medical imaging and augmented reality : 7th International Conference, MIAR 2016, Bern, Switzerland, August 24-26, 2016 : proceedings. MIAR (Workshop) (7th : 2016 : Bern, Switzerland)",
  ["med immunol"] = "Medical immunology (London, England)",
  ["med inf (lond)"] = "Medical Informatics",
  ["med infant"] = "Medicina infantil",
  ["med infant (paris)"] = "La Médecine infantile",
  ["med inform (lond)"] = "Medical informatics = Médecine et informatique",
  ["med inform internet med"] = "Medical informatics and the Internet in medicine",
  ["med inform statist"] = "Medizinische Informatik und Statistik",
  ["med innov bus"] = "Medical innovation & business",
  ["med insight"] = "Medical insight",
  ["med instrum"] = "Medical instrumentation",
  ["med instrum (luton)"] = "Medical instrumentation (Luton, England)",
  ["med int"] = "Medicine international (UK ed.)",
  ["med int (milano)"] = "La Medicina internazionale",
  ["med int rev"] = "Medicina internacia revuo",
  ["med intensiva"] = "Medicina intensiva",
  ["med interface"] = "Medical interface",
  ["med interna"] = "Medicina Interna",
  ["med interna (bucur)"] = "Medicină internă",
  ["med interne"] = "Médecine interne",
  ["med investig louisville"] = "Medical investigator (Louisville, Ky.)",
  ["med ital"] = "La Medicina italiana",
  ["med j (ft sam houst tex)"] = "Medical journal (Fort Sam Houston, Tex.)",
  ["med j armed forces india"] = "Medical journal, Armed Forces India",
  ["med j aust"] = "The Medical journal of Australia",
  ["med j cairo univ"] = "The Medical journal of Cairo University",
  ["med j egypt armed forces"] = "The Medical journal of the Egyptian Armed Forces",
  ["med j islam repub iran"] = "Medical journal of the Islamic Republic of Iran",
  ["med j malaya"] = "The Medical journal of Malaya",
  ["med j malaysia"] = "The Medical journal of Malaysia",
  ["med j nutrition metab"] = "Mediterranean journal of nutrition and metabolism",
  ["med j obstet gynecol"] = "Medical journal of obstetrics and gynecology",
  ["med j osaka univ"] = "Medical journal of Osaka University",
  ["med j southwest"] = "The Medical journal of the South-West",
  ["med j zambia"] = "Medical journal of Zambia",
  ["med klin"] = "Medizinische Klinik",
  ["med klin (munich)"] = "Medizinische Klinik (Munich, Germany : 1983)",
  ["med klin [klin]"] = "Medizinische Klinik (Klinik-Ausg.)",
  ["med klin [prax]"] = "Medizinische Klinik (Praxis-Ausg.)",
  ["med klin intensivmed notfmed"] = "Medizinische Klinik, Intensivmedizin und Notfallmedizin",
  ["med klin klin"] = "Medizinische Klinik (Klinik-Ausg.)",
  ["med klin prax"] = "Medizinische Klinik (Praxis-Ausg.)",
  ["med klin suppl"] = "Medizinische Klinik (Munich, Germany : 1983). Supplement",
  ["med konink acad wetensch belgië"] = "Mededelingen van de Koninklijke Academie voor Wetenschappen, Letteren en Schone Kunsten van België",
  ["med lab"] = "Médecine et laboratoire",
  ["med lab (stuttg)"] = "Das Medizinische Laboratorium",
  ["med lab sci"] = "Medical laboratory sciences",
  ["med lab technol"] = "Medical laboratory technology",
  ["med laser appl"] = "Medical Laser Application",
  ["med lat"] = "Medicina latina; revista médica de colaboración científica",
  ["med lav"] = "La Medicina del lavoro",
  ["med law"] = "Medicine and law",
  ["med law int"] = "Medical law international",
  ["med law rev"] = "Medical law review",
  ["med leg bull"] = "Medico-legal bulletin",
  ["med leg criminol rev"] = "The Medico-legal and criminological review",
  ["med leg dommage corpor"] = "Médecine légale et dommage corporel",
  ["med leg j"] = "The Medico-legal journal",
  ["med letop"] = "Meditsinski letopisi",
  ["med lett drugs ther"] = "The Medical letter on drugs and therapeutics",
  ["med lib"] = "Medical libraries",
  ["med library hist j"] = "Medical library and historical journal",
  ["med libre"] = "Le Médecin libre",
  ["med maandbl"] = "Medisch maandblad",
  ["med mal infect"] = "Médecine et maladies infectieuses",
  ["med malpract cost containment j"] = "The Medical malpractice cost containment journal",
  ["med manag netw"] = "Medical management network",
  ["med mark media"] = "Medical marketing & media",
  ["med microbiol immunol"] = "Medical microbiology and immunology",
  ["med microbiol immunol (berl)"] = "Medical Microbiology and Immunology",
  ["med microbiol lett"] = "Medical microbiology letters : an international journal for rapid communications on all aspects of medical and clinical microbiology",
  ["med microecol"] = "Medicine in Microecology",
  ["med mission austria"] = "Medical mission to Austria, July 1-August 8, 1947. American Unitarian Service Committee in cooperation with World Health Organization Interim Commission. [Abridged report submitted by Erwin Kohn]. Unitarian Service Committee",
  ["med mol morphol"] = "Medical molecular morphology",
  ["med monatsschr"] = "Medizinische Monatsschrift",
  ["med monatsschr beih"] = "Medizinische Monatsschrift. Beihefte",
  ["med monatsschr pharm"] = "Medizinische Monatsschrift fur Pharmazeuten",
  ["med monde"] = "Médecine dans le monde",
  ["med moral newsl"] = "The Medical-moral newsletter",
  ["med mycol"] = "Medical mycology",
  ["med mycol case rep"] = "Medical mycology case reports",
  ["med mycol j"] = "Medical mycology journal",
  ["med mycol open access"] = "Medical mycology: open access",
  ["med nachr ver staaten"] = "Medizinische Nachrichten aus den Vereinigten Staaten",
  ["med netw strategy rep"] = "Medical network strategy report",
  ["med news lett aviat suppl"] = "Medical news letter. Aviation supplement. United States. Navy Department Department of Aviation Medicine",
  ["med news schering"] = "Medical news Schering",
  ["med newsl (lond)"] = "Medical newsletter",
  ["med newsl (london)"] = "Medical Newsletter",
  ["med nov"] = "Medizinische Novitäten",
  ["med nov technol devices"] = "Medicine in novel technology and devices",
  ["med novel technol devices"] = "Medicine in Novel Technology and Devices",
  ["med nowozytna"] = "Medycyna nowozytna : studia nad historia medycyny",
  ["med off"] = "The Medical officer",
  ["med omics"] = "Medicine in Omics",
  ["med oncol"] = "Medical oncology (Northwood, London, England)",
  ["med oncol tumor pharmacother"] = "Medical oncology and tumor pharmacotherapy",
  ["med one"] = "Med one",
  ["med opin (n y)"] = "Medical opinion (New York)",
  ["med opin rev"] = "Medical opinion & review",
  ["med oral"] = "Medicina oral : órgano oficial de la Sociedad Española de Medicina Oral y de la Academia Iberoamericana de Patología y Medicina Bucal",
  ["med oral patol oral cir bucal"] = "Medicina oral, patología oral y cirugía bucal",
  ["med panam"] = "Medicina panamericana",
  ["med parazitol (mosk)"] = "Meditsinskaia parazitologiia i parazitarnye bolezni",
  ["med pediatr oncol"] = "Medical and pediatric oncology",
  ["med pediatr oncol suppl"] = "Medical and pediatric oncology. Supplement",
  ["med periskop ingelheim"] = "Das Medizinische Periskop Ingelheim",
  ["med pharm rep"] = "Medicine and pharmacy reports",
  ["med pharmacol exp int j exp med"] = "Medicina et pharmacologia experimentalis. International journal of experimental medicine",
  ["med photonics"] = "Medical Photonics",
  ["med phys"] = "Medical physics",
  ["med phys j"] = "The Medical and physical journal",
  ["med plants"] = "Medicinal Plants",
  ["med pr"] = "Medycyna pracy",
  ["med pract"] = "El Medico práctico",
  ["med pract (zaragoza)"] = "Medicina práctica",
  ["med pravo"] = "Medychne pravo",
  ["med pregl"] = "Medicinski pregled",
  ["med press"] = "The Medical press",
  ["med press egypt"] = "The Medical press of Egypt",
  ["med princ pract"] = "Medical principles and practice : international journal of the Kuwait University, Health Science Centre",
  ["med probl perform art"] = "Medical problems of performing artists",
  ["med proc (johannesb)"] = "Medical proceedings : a South African journal for the advancement of medical science = Mediese bydraes : 'n Suid-Afrikaanse tydskrif vir die bevordering van die geneeskunde",
  ["med prod sales"] = "Medical products sales : MPS : the official journal of the American Surgical Trade Association",
  ["med prog"] = "Medical progress; a review of medical advances",
  ["med prog technol"] = "Medical progress through technology",
  ["med prom sssr"] = "Meditsinskaia promyshlennost' SSSR",
  ["med psicosom"] = "Medicina psicosomatica",
  ["med q rev"] = "The Medical quarterly review",
  ["med rab"] = "Meditsinskiĭ rabotnik",
  ["med radiogr"] = "Medical radiographer",
  ["med radiogr photogr"] = "Medical radiography and photography",
  ["med radiol"] = "Medical radiology",
  ["med radiol (mosk)"] = "Meditsinskaia radiologiia",
  ["med razgl"] = "Medicinski razgledi",
  ["med rec"] = "Medical record (New York : 1934)",
  ["med rec (reading)"] = "Medical record (Reading, Pa.)",
  ["med rec ann"] = "Medical record and annals",
  ["med rec health care inf j"] = "Medical record and health care information journal",
  ["med rec news"] = "Medical record news",
  ["med ref serv q"] = "Medical reference services quarterly",
  ["med res arch"] = "Medical research archives",
  ["med res counc spec rep ser (lond)"] = "Medical Research Council Special Report Series",
  ["med res eng"] = "Medical research engineering",
  ["med res innov"] = "Medical research and innovations",
  ["med res rep"] = "Medical Research report. United States. Navy. Mine Defense Laboratory. Medical Research Dept",
  ["med res rev"] = "Medicinal research reviews",
  ["med reserve (paris)"] = "Le Médecin de réserve",
  ["med rev"] = "Medicina em revista",
  ["med sachverstand"] = "Der Medizinische Sachverständige",
  ["med sante trop"] = "Médecine et santé tropicales",
  ["med sci"] = "Medical science",
  ["med sci (basel)"] = "Medical sciences (Basel, Switzerland)",
  ["med sci (paris)"] = "Médecine sciences : M/S",
  ["med sci (turkey)"] = "Medicine science",
  ["med sci educ"] = "Medical science educator",
  ["med sci law"] = "Medicine, science, and the law",
  ["med sci monit"] = "Medical science monitor : international medical journal of experimental and clinical research",
  ["med sci monit basic res"] = "Medical science monitor basic research",
  ["med sci res"] = "Medical science research",
  ["med sci sports"] = "Medicine and science in sports",
  ["med sci sports exerc"] = "Medicine and science in sports and exercise",
  ["med sci sports exercise"] = "Medicine & Science in Sports & Exercise",
  ["med secoli"] = "Medicina nei secoli",
  ["med sect proc"] = "Medical Section Proceedings",
  ["med segur trab (madr)"] = "Medicina y seguridad del trabajo",
  ["med self care"] = "Medical self-care",
  ["med sens imaging"] = "Medical Sensors & Imaging",
  ["med serv"] = "Medical service",
  ["med serv j can"] = "Medical services journal, Canada",
  ["med sestra"] = "Medit︠s︡inskai︠a︡ sestra",
  ["med soc (berkeley)"] = "Medicine and society",
  ["med soc report"] = "Medical Society reporter",
  ["med sport"] = "Medizin und Sport",
  ["med sport (berl)"] = "Medizin und Sport",
  ["med sport (roma)"] = "Medicina dello sport; rivista di fisiopatologia dello sport",
  ["med sport sci"] = "Medicine and sport science",
  ["med staff couns"] = "The Medical staff counselor",
  ["med stud"] = "Medicine studies",
  ["med student res j"] = "Medical student research journal",
  ["med surg suggest"] = "Medico-surgical suggestions",
  ["med teach"] = "Medical teacher",
  ["med tech (stuttg)"] = "Medizinische Technik",
  ["med techn bull"] = "Medical technicians bulletin",
  ["med tekh"] = "Meditsinskaia tekhnika",
  ["med ther med reprod gynecol endocrinol"] = "Médecine thérapeutique médecine de la reproduction, gynécologie, endocrinologie",
  ["med thorac"] = "Medicina thoracalis",
  ["med times"] = "Medical times",
  ["med today"] = "Medicine today",
  ["med today tomorrow"] = "Medicine today & tomorrow",
  ["med toxicol"] = "Medical toxicology",
  ["med toxicol adverse drug exp"] = "Medical toxicology and adverse drug experience",
  ["med tr prom ekol"] = "Meditsina truda i promyshlennaia ekologiia",
  ["med tradic"] = "Medicina tradicional",
  ["med trial tech q"] = "Medical trial technique quarterly",
  ["med trib med news"] = "Medical tribune and medical news",
  ["med trop (madr)"] = "La Medicina tropical",
  ["med trop (mars)"] = "Medecine Tropicale",
  ["med ultrason"] = "Medical ultrasonography",
  ["med ultrasound"] = "Medical ultrasound",
  ["med ultrasound preterm perinat paediatr image anal (2020)"] = "Medical ultrasound, and preterm, perinatal and paediatric image analysis",
  ["med usine rev hyg ind mal prof"] = "Le Médecin d'usine; revue d'hygiène industrielle et des maladies professionnelles",
  ["med vet braz"] = "Medicina Veterinaria Brazil",
  ["med vet entomol"] = "Medical and veterinary entomology",
  ["med virol pathog dis control"] = "Medical Virology from Pathogenesis to Disease Control",
  ["med vjesn (osijek)"] = "Medicinski vjesnik (Osijek, Croatia)",
  ["med war"] = "Medicine and war",
  ["med waste anal"] = "Medical waste analyst",
  ["med way"] = "The Medical way",
  ["med welt"] = "Die Medizinische Welt",
  ["med weter"] = "Medycyna weterynaryjna",
  ["med wieku rozwoj"] = "Medycyna wieku rozwojowego",
  ["med womans j"] = "Medical woman's journal",
  ["med world"] = "Medical world",
  ["med world (new york)"] = "The Medical world",
  ["med world news"] = "Medical world news",
  ["med writ"] = "Medical Writing",
  ["med year"] = "Medicine of the year",
  ["med zh uzb"] = "Meditsinskii zhurnal Uzbekistana",
  ["med zhurnal"] = "Medit︠s︡inskiĭ zhurnal",
  ["medcomm (beijing)"] = "MedComm",
  ["medcomm - biomater appl"] = "MedComm - Biomaterials and Applications",
  ["medcomm - future med"] = "MedComm - Future Medicine",
  ["medd fam hist proj hist inst upps univ"] = "Meddelande från Familjehistoriska projektet, Historiska institutionen, Uppsala universitet",
  ["medd flyg navalmed namnd statens namnd for flyg navalmed forsk forsoksvererksamhet"] = "Meddelanden från Flyg- och navalmedicinska nämnden. Statens nämnd för flyg- och navalmedicinsk forsknings- och försöksverksamhet",
  ["medd groenl: geosc"] = "Meddelelser om Groenland: Geoscience",
  ["medd nor farm selsk"] = "Meddelelser fra Norsk farmaceutisk selskap",
  ["medd sundhedsstyr beredskabsafdelingen"] = "Meddelelser fra Sundhedsstyrelsen. Denmark. Sundhedsstyrelsen. Beredskabsafdelingen",
  ["meded fac landbouwwet, rijksuniv gent"] = "Mededelingen van de Faculteit Landbouwwetenschappen, Universiteit Gent",
  ["meded k acad wet lett schone kunsten belg kl wet"] = "Mededelingen van de Koninklijke Academie voor wetenschappen, letteren en schone kunsten van Belgie. Klasse der wetenschappen",
  ["meded landbouwhoogesch wagening"] = "Mededelingen van de Landbouwhoogeschool te Wageningen",
  ["meded rijksuniv gent fak landbouwkd toegep biol wet"] = "Mededelingen (Rijksuniversiteit te Gent. Fakulteit van de Landbouwkundige en Toegepaste Biologische Wetenschappen)",
  ["meded wiskd genoot"] = "Mededelingen van het Wiskundig Genootschap",
  ["medeni med j"] = "Medeniyet medical journal",
  ["medhist mater"] = "Medizinhistorische Materialien",
  ["medi-cal policy inst issue brief"] = "Medi-Cal Policy Institute issue brief",
  ["media & gend monit"] = "Media & gender monitor / An occasional publication by the Women’s Programme at the World Association for Christian Communication",
  ["media asia"] = "Media Asia",
  ["media commun"] = "Media and communication",
  ["media cult soc"] = "Media, culture, and society",
  ["media dev"] = "Media development",
  ["media gend monit"] = "Media & gender monitor",
  ["media hist"] = "Media history",
  ["media peternakan fakultas peternakan institut pertanian bogor"] = "Media Peternakan",
  ["media psychol"] = "Media psychology",
  ["mediaev j"] = "The Mediæval journal",
  ["mediaev philos pol"] = "Mediaevalia philosophica polonorum",
  ["mediaev stud"] = "Mediaeval studies",
  ["mediators inflamm"] = "Mediators of inflammation",
  ["mediators inflammation"] = "Mediators of Inflammation",
  ["medicam hist soc"] = "Medicamento, história e sociedade",
  ["medicamenta (madr)"] = "Medicamenta",
  ["medicare brief"] = "Medicare brief",
  ["medicare medicaid res rev"] = "Medicare & medicaid research review",
  ["medicc rev"] = "MEDICC review",
  ["medicina (b aires)"] = "Medicina",
  ["medicina (firenze)"] = "Medicina (Florence, Italy)",
  ["medicina (kaunas)"] = "Medicina (Kaunas, Lithuania)",
  ["medicina (mex)"] = "Medicina",
  ["medicine (abingdon)"] = "Medicine (Abingdon, England : UK ed.)",
  ["medicine (baltimore)"] = "Medicine",
  ["medicines (basel)"] = "Medicines (Basel, Switzerland)",
  ["medico (eur)"] = "Medico; Europa Ausg",
  ["medico (int)"] = "Medico; international ed",
  ["medico (porto)"] = "O Medico; semanario de assuntos médicos e paramédicos",
  ["medicoleg bioeth"] = "Medicolegal and bioethics",
  ["medicoleg news"] = "Medicolegal news",
  ["mediev archaeol"] = "Medieval archaeology",
  ["mediev humanist"] = "Medievalia et humanistica",
  ["mediev span med text ser"] = "Medieval Spanish medical text series",
  ["medieval archaeol"] = "Medieval Archaeology",
  ["medika (zagreb)"] = "Medika; informativni c̆asopis farmaceutiskih poduzeća NRH",
  ["medioev romanzo"] = "Medioevo romanzo",
  ["medioevo lat"] = "Medioevo latino",
  ["mediterr agric sci"] = "Mediterranean agricultural sciences",
  ["mediterr bot"] = "Mediterranean Botany",
  ["mediterr conf control automation"] = "Mediterranean Conference on Control & Automation : [proceedings]. IEEE Mediterranean Conference on Control & Automation",
  ["mediterr geosci rev"] = "Mediterranean Geoscience Reviews",
  ["mediterr hist rev"] = "Mediterranean historical review",
  ["mediterr j hematol infect dis"] = "Mediterranean journal of hematology and infectious diseases",
  ["mediterr j math"] = "Mediterranean Journal of Mathematics",
  ["mediterr j nutr metab"] = "Mediterranean Journal of Nutrition and Metabolism",
  ["mediterr j rheumatol"] = "Mediterranean journal of rheumatology",
  ["mediterr j rheumatol ell reumatol"] = "Mediterranean journal of rheumatology - Ellīnikī reumatologia",
  ["mediterr j soc sci"] = "Mediterranean journal of social sciences",
  ["mediterr mar sci"] = "Mediterranean marine science",
  ["mediterr med"] = "Méditerranée médicale",
  ["mediterr peoples"] = "Peuples méditerranéens. Mediterranean peoples",
  ["mediterr stud (kirksv mo)"] = "[Mediterranean studies]",
  ["medium aevum"] = "Medium aevum",
  ["medizinhist j"] = "Medizinhistorisches Journal",
  ["medlab mag"] = "MedLab magazine",
  ["medscape j med"] = "Medscape journal of medicine",
  ["medscape womens health"] = "Medscape women's health",
  ["medsci law"] = "Medicine, Science and the Law",
  ["medsurg nurs"] = "Medsurg nursing : official journal of the Academy of Medical-Surgical Nurses",
  ["medtube sci"] = "MEDtube science",
  ["meharri dent"] = "The Meharri-Dent",
  ["meiji seika kenkyu nenpo yakuhin bumon"] = "[Meiji Seika kenkyū nenpō; yakuhin bumon] Scientific reports of Meiji Seika Kaisha; pharmaceutical division",
  ["meikai daigaku shigaku zasshi"] = "Meikai Daigaku shigaku zasshi = The Journal of Meikai University School of Dentistry",
  ["meisheimer entomol ser"] = "Melsheimer entomological series",
  ["mekh kompoz mater (zinatne)"] = "Mekhanika Kompositnykh Materialov (Zinatne)",
  ["mekh tverd tela"] = "Natsionalnaya Akademiya Nauk Ukrainy. Institut Prikladnoi Matematiki i Mekhaniki. Mekhanika Tverdogo Tela",
  ["melanges casa velazquez"] = "Mélanges de la Casa de Velázquez",
  ["melanges ec fr rome moyen age temps mod"] = "Mélanges de l'Ecole française de Rome. Moyen âge, temps modernes",
  ["melanoma manag"] = "Melanoma management",
  ["melanoma res"] = "Melanoma research",
  ["melb hist j"] = "Melbourne historical journal",
  ["melb univ law rev"] = "Melbourne University law review",
  ["meld nor landbrukshoegsk"] = "Meldinger fra Norges Landbrukshoegskole",
  ["melita theol"] = "Melita theologica : the review of the Royal University Students' Theological Association, Malta",
  ["mellen hist med"] = "Mellen history of medicine",
  ["melliand textilber"] = "Melliand Textilberichte",
  ["mem acad chir (paris)"] = "Mémoires. Académie de chirurgie (France)",
  ["mem acad cienc lisb cl cienc"] = "Memórias da Academia das Ciências de Lisboa, Classe de Ciências",
  ["mem acad r med belg"] = "Mémoires de l'Académie royale de médecine de Belgique",
  ["mem acad sci inscr b-lett toulouse"] = "Mémoires de l'Académie des sciences, inscriptions et belles-lettres de Toulouse",
  ["mem accad sci torino cl sci fis mat natur (5)"] = "Accademia delle Scienze di Torino",
  ["mem am assoc pet geol"] = "Memoir - American Association of Petroleum Geologists",
  ["mem am math soc"] = "Memoirs of the American Mathematical Society",
  ["mem amer math soc"] = "Memoirs of the American Mathematical Society",
  ["mem amer philos soc"] = "Memoirs of the American Philosophical Society",
  ["mem biospeol"] = "Mémoires de biospéologie",
  ["mem cognit"] = "Memory & cognition",
  ["mem col nac (mex)"] = "Memoria de el Colegio Nacional. Colegio Nacional (Mexico)",
  ["mem college ed akita univ natur sci"] = "Memoirs of the College of Education Akita University",
  ["mem differ equ math phys"] = "Memoirs on Differential Equations and Mathematical Physics",
  ["mem differential equations math phys"] = "Georgian Academy of Sciences",
  ["mem eur math soc"] = "Memoirs of the European Mathematical Society",
  ["mem fac ed kumamoto univ natur sci"] = "Kumamoto University",
  ["mem fac educ shiga univ nat sci"] = "Memoirs of the Faculty of Education",
  ["mem fac eng design kyoto inst tech ser sci tech"] = "Memoirs of the Faculty of Engineering and Design Kyoto Institute of Technology",
  ["mem fac eng hiroshima univ"] = "Memoirs of the Faculty of Engineering Hiroshima University",
  ["mem fac eng miyazaki univ"] = "Memoirs of the Faculty of Engineering Miyazaki University",
  ["mem fac gen ed kumamoto univ natur sci"] = "Kumamoto University",
  ["mem fac sci eng shimane univ ser b math sci"] = "Memoirs of the Faculty of Science and Engineering Shimane University",
  ["mem fac sci kochi univ ser a math"] = "Kochi University",
  ["mem fac sci kochi univ ser d biol"] = "Memoirs of the Faculty of Science, Kochi University. Series D, Biology",
  ["mem fac sci kyushu univ, ser c"] = "Memoirs of the Faculty of Science, Kyushu University, Series C: Chemistry",
  ["mem grad sch sci eng shimane univ ser b math"] = "Memoirs of the Graduate School of Science and Engineering. Shimane University. Series B. Mathematics",
  ["mem inst butantan"] = "Memorias do Instituto Butantan",
  ["mem inst oswaldo cruz"] = "Memórias do Instituto Oswaldo Cruz",
  ["mem inst sci eng ritsumeikan univ"] = "Memoirs of the Institute of Science and Engineering",
  ["mem inst sci engrg ritsumeikan univ"] = "Memoirs of the Institute of Science and Engineering. Ritsumeikan University",
  ["mem insts sci ind res, osaka univ"] = "Memoirs of the Institute of Scientific and Industrial Research, Osaka University",
  ["mem muroran inst technol"] = "Memoirs of the Muroran Institute of Technology",
  ["mem mus civ stor nat verona iia monogr nat"] = "Memorie del Museo civico di storia naturale di Verona. IIa Serie. Monografie naturalistiche",
  ["mem mus natl hist nat (1993)"] = "Mémoires du Muséum national d'histoire naturelle (1993)",
  ["mem mus vic"] = "Memoirs of Museum Victoria",
  ["mem n y bot gard"] = "Memoirs of the New York Botanical Garden",
  ["mem nat defense acad"] = "Memoirs of the National Defense Academy",
  ["mem osaka inst tech"] = "Memoirs of Osaka Institute of Technology",
  ["mem osaka inst tech ser a"] = "Memoirs of the Osaka Institute of Technology",
  ["mem osaka kyoiku univ iii natur sci appl sci"] = "Memoirs of Osaka Kyoiku University",
  ["mem osaka kyoiku univ ser iii nat sci appl sci"] = "Memoirs of Osaka Kyoiku University. Ser. III. Natural Science and Applied Science",
  ["mem proc chem soc"] = "Memoirs and Proceedings, Chemical Society",
  ["mem proc manch lit philos soc"] = "Memoirs and proceedings of the Manchester Literary & Philosophical Society. Manchester Literary and Philosophical Society",
  ["mem queensl mus"] = "Memoirs of the Queensland Museum",
  ["mem r astron soc"] = "Memoirs of the Royal Astronomical Society",
  ["mem real acad cienc artes barcelona"] = "Memorias de la Real Academia de Ciencias y Artes de Barcelona",
  ["mem real acad cienc exact fís natur madrid"] = "Memorias de la Real Academia de Ciencias Exactas, Físicas y Naturales de Madrid",
  ["mem school sci eng waseda univ"] = "Memoirs of the School of Science and Engineering",
  ["mem sect stiint acad romana ser iv"] = "Academia Romana. Memoriile Sectiilor Stiintifice. Seria IV. Memoirs of the Scientific Sections. Romanian Academy. Series IV",
  ["mem secţ ştiinţ acad română ser iv"] = "Academia Română",
  ["mem soc"] = "Memoria y sociedad : revista del Departamento de Historia y Geografía",
  ["mem soc agric commer sci arts marne"] = "Mémoires de la Société d'agriculture, commerce, sciences et arts du département de la Marne. Société d'agriculture, commerce, sciences et arts du département de la Marne, Châlons-sur-Marne",
  ["mem soc mat mex"] = "Memorias de la Sociedad Matematica Mexicana",
  ["mem soc math fr (ns)"] = "Memoires de la Societe Mathematique de France. Nouvelle Serie",
  ["mem soc r sci liege"] = "Memoires de la Societe Royale des Sciences de Liege",
  ["mem soc urug urol"] = "Memoria. Sociedad Uruguaya de Urología",
  ["mem stud"] = "Memory studies",
  ["memai heiko igaku"] = "Memai heikō igaku",
  ["membr biochem"] = "Membrane biochemistry",
  ["membr cell biol"] = "Membrane & cell biology",
  ["membr membr technol"] = "Membranes and Membrane Technologies",
  ["membr technol"] = "Membrane Technology",
  ["membr water treat"] = "Membrane and Water Treatment",
  ["membranes (basel)"] = "Membranes",
  ["membranes (basel, switz)"] = "Membranes (Basel, Switzerland)",
  ["memet comput"] = "Memetic computing",
  ["memo med res counc"] = "Medical Research Council memorandum",
  ["memo rep nav med res inst (us)"] = "Memorandum report. Naval Medical Research Institute (U.S.)",
  ["memo rm"] = "Memorandum RM-. Rand Corporation",
  ["memo soc fauna flora fenn"] = "Memoranda Societatis pro Fauna et Flora Fennica",
  ["memorias congreso med penins"] = "Memorias. Congreso Médico Peninsular",
  ["memories - mater devices circuits syst"] = "Memories- Materials, Devices, Circuits and Systems",
  ["memphis med j"] = "Memphis medical journal",
  ["memphis mid south med j"] = "Memphis and Mid-South medical journal",
  ["memphis state univ law rev"] = "Memphis State University law review",
  ["men masc"] = "Men and masculinities",
  ["men's reprod health"] = "Men's reproductive health",
  ["menara perkeb"] = "Menara perkebunan",
  ["mendel newsl"] = "The Mendel newsletter; archival resources for the history of genetics & allied sciences",
  ["mendeleev commun"] = "Mendeleev Communications",
  ["menemui mat"] = "Menemui Matematik",
  ["menninger perspect"] = "Menninger perspective",
  ["mennon q rev"] = "The Mennonite quarterly review",
  ["menopausal med"] = "Menopausal medicine : for clinicians who provide care for women",
  ["menopause int"] = "Menopause international",
  ["menopause manag"] = "Menopause management",
  ["mens agit—colloquia"] = "Mens Agitat—Colloq.",
  ["mens ondernem"] = "Mens en onderneming",
  ["mens sana monogr"] = "Mens sana monographs",
  ["ment disabil law rep"] = "Mental disability law reporter",
  ["ment health (lond)"] = "Mental health (London)",
  ["ment health (montr)"] = "Mental health",
  ["ment health addict res"] = "Mental health and addiction research",
  ["ment health bull (danville)"] = "Mental health bulletin",
  ["ment health care"] = "Mental health care",
  ["ment health clin"] = "The mental health clinician",
  ["ment health fam med"] = "Mental health in family medicine",
  ["ment health nurs"] = "Mental health nursing : journal of the Psychiatric Nurses Association",
  ["ment health phys act"] = "Mental health and physical activity",
  ["ment health prev"] = "Mental health & prevention",
  ["ment health relig cult"] = "Mental health, religion & culture",
  ["ment health rev (brighton)"] = "Mental health review (Brighton, England)",
  ["ment health serv res"] = "Mental health services research",
  ["ment health soc"] = "Mental health and society",
  ["ment health stat note"] = "Mental health statistical note",
  ["ment health subst use"] = "Mental health and substance use : dual diagnosis",
  ["ment health today"] = "Mental health today (Brighton, England)",
  ["ment hosp"] = "Mental hospitals",
  ["ment hyg"] = "Mental Hygiene",
  ["ment hyg (lond)"] = "Mental hygiene",
  ["ment hyg surv"] = "Mental hygiene survey",
  ["ment illn"] = "Mental illness",
  ["ment lex"] = "The mental lexicon",
  ["ment phys disabil law rep"] = "Mental and physical disability law reporter",
  ["ment retard"] = "Mental retardation",
  ["ment retard abstr"] = "Mental retardation abstracts",
  ["ment retard dev disabil res rev"] = "Mental retardation and developmental disabilities research reviews",
  ["ment retard law"] = "Mental retardation and the law",
  ["ment welf"] = "Mental welfare",
  ["mental health spec interest sect q"] = "Mental health special interest section quarterly",
  ["mentor tutoring"] = "Mentoring & tutoring",
  ["men’s reprod health"] = "Men’s reproductive health",
  ["mercer dent soc newsl"] = "Mercer Dental Society Newsletter",
  ["mercer law rev"] = "Mercer law review",
  ["merck rep"] = "The Merck report",
  ["meridian crossing aesthet"] = "Meridian: Crossing Aesthetics",
  ["merkbl forstprakt eidgenöss anst forstl verswes"] = "Merkblatt für den Forstpraktiker Eidg. Anstalt für das forstliche Versuchswesen",
  ["merkbl prax"] = "Merkblatt für die Praxis",
  ["merrill palmer q"] = "Merrill-Palmer quarterly",
  ["merrill palmer q (wayne state univ press)"] = "Merrill-Palmer quarterly (Wayne State University. Press)",
  ["merrill palmer q behav dev"] = "Merrill-Palmer quarterly of behavior and development",
  ["mes eval educ"] = "Mesure et évaluation en éducation",
  ["mese sanit"] = "Mese sanitario",
  ["mesic ciba"] = "Mĕsíc̆ník Ciba",
  ["mesoam herpetol"] = "Mesoamerican herpetology",
  ["mesoamerica (antigua guatem)"] = "Mesoamérica (Antigua, Guatemala)",
  ["mesoporous biomater"] = "Mesoporous biomaterials",
  ["messenger (los angel)"] = "Messenger (Los Angeles, Calif. : Print)",
  ["met based drugs"] = "Metal-based drugs",
  ["met finish"] = "Metal Finishing",
  ["met ions biol syst"] = "Metal Ions in Biological Systems",
  ["met ions life sci"] = "Metal ions in life sciences",
  ["met mater"] = "Metals and Materials",
  ["met mater int"] = "Metals and Materials International",
  ["met powder rep"] = "Metal Powder Report",
  ["met sci heat treat"] = "Metal Science and Heat Treatment",
  ["met-based drugs"] = "Metal-Based Drugs",
  ["meta gene"] = "Meta gene",
  ["metab bone dis relat res"] = "Metabolic bone disease & related research",
  ["metab brain dis"] = "Metabolic brain disease",
  ["metab clin exp"] = "Metabolism Clinical and Experimental",
  ["metab eng"] = "Metabolic engineering",
  ["metab eng commun"] = "Metabolic engineering communications",
  ["metab nutr oncol"] = "Metabolism and Nutrition in Oncology",
  ["metab ophthalmol"] = "Metabolic ophthalmology",
  ["metab pediatr ophthalmol"] = "Metabolic and pediatric ophthalmology",
  ["metab pediatr syst ophthalmol"] = "Metabolic, pediatric, and systemic ophthalmology",
  ["metab pediatr syst ophthalmol (1985)"] = "Metabolic, pediatric, and systemic ophthalmology (New York, N.Y. : 1985)",
  ["metab syndr relat disord"] = "Metabolic syndrome and related disorders",
  ["metab, clin exp"] = "Metabolism, Clinical and Experimental",
  ["metabarcoding metagenom"] = "Metabarcoding and metagenomics",
  ["metabol open"] = "Metabolism open",
  ["metabolomics (los angel)"] = "Metabolomics : open access",
  ["metacogn learn"] = "Metacognition and learning",
  ["metagenomics (cairo)"] = "Metagenomics (Cairo, Egypt)",
  ["metall (berlin)"] = "Metall (Berlin)",
  ["metall ital"] = "Metallurgia Italiana",
  ["metall mater eng"] = "Metallurgical and Materials Engineering",
  ["metall mater trans a"] = "Metallurgical and Materials Transactions A: Physical Metallurgy and Materials Science",
  ["metall mater trans a phys metall mater sci"] = "Metallurgical and materials transactions. A. Physical metallurgy and materials science",
  ["metall mater trans b"] = "Metallurgical and Materials Transactions B: Process Metallurgy and Materials Processing Science",
  ["metall mater trans b process metall mater process sci"] = "Metallurgical And Materials Transactions B-Process Metallurgy And Materials Processing Science",
  ["metall mater trans e"] = "Metallurgical and Materials Transactions E, Materials for Energy Systems",
  ["metall met forming"] = "Metallurgia and Metal Forming",
  ["metall microstruct anal"] = "Metallography, Microstructure, and Analysis",
  ["metall res technol"] = "Metallurgical Research & Technology",
  ["metall trans a"] = "Metallurgical Transactions A",
  ["metall trans b"] = "Metallurgical Transactions B",
  ["metallges period rev"] = "Metallgesellschaft Periodic Review",
  ["metallofiz noveishie tekhnol"] = "Metallofizika i Noveishie Tekhnologii",
  ["metallofizika (akad nauk ukr ssr, otd fiz astron)"] = "Metallofizika (Kiev)",
  ["metalloorg khim"] = "Metalloorganicheskaya Khimiya",
  ["metalloproteinases med"] = "Metalloproteinases in medicine",
  ["metalloved term obrab met"] = "Metallovedenie i Termicheskaya Obrabotka Metallov",
  ["metalogicon (ns)"] = "Metalogicon. International Journal of Pure and Applied Logic, Linguistics and Philosophy. New Series",
  ["metals (basel)"] = "Metals",
  ["metals (basel, switz)"] = "Metals (Basel, Switzerland)",
  ["metaphor symb"] = "Metaphor and symbol",
  ["meteor gidrol"] = "Meteorologiya i Gidrologiya",
  ["meteor rundsch"] = "Meteorologische Rundschau",
  ["meteor z"] = "Meteorologische Zeitschrift",
  ["meteorit planet sci"] = "Meteoritics & Planetary Science",
  ["meteorol appl"] = "Meteorological Applications",
  ["meteorol atmos phys"] = "Meteorology and Atmospheric Physics",
  ["meteorol mag"] = "Meteorological Magazine",
  ["meteorol monogr"] = "Meteorological Monographs",
  ["meteorol technol int"] = "Meteorological technology international",
  ["meteorol z"] = "Meteorologische Zeitschrift",
  ["method cell biol"] = "Neurons: Methods and Applications for The Cell Biologist",
  ["method innov"] = "Methodological innovations",
  ["method regelungstech"] = "Methoden der Regelungstechnik",
  ["methoden daten anal"] = "Methoden, daten, analysen",
  ["methoden psych"] = "Methoden der Psychologie",
  ["methoden regel autom"] = "Methoden der Regelungs- und Automatisierungstechnik",
  ["methoden verfahren math phys"] = "Methoden und Verfahren der Mathematischen Physik",
  ["methodist debakey cardiovasc j"] = "Methodist DeBakey cardiovascular journal",
  ["methodist hist"] = "Methodist history",
  ["methodol comput appl probab"] = "Methodology and Computing in Applied Probability",
  ["methodol innovations"] = "Methodological Innovations",
  ["methodol music res"] = "Methodology of Music Research",
  ["methodol sci"] = "Methodology and science",
  ["methodol soc sci"] = "Methodology in the Social Sciences",
  ["methodology (gott)"] = "Methodology : European journal of research methods for the behavioral & social sciences",
  ["methodos ser"] = "Methodos Series",
  ["methods (amsterdam, neth)"] = "Methods (Amsterdam, Netherlands)",
  ["methods (oxford, u k)"] = "Methods (Oxford, United Kingdom)",
  ["methods (san diego, ca, u s)"] = "Methods (San Diego, CA, United States)",
  ["methods achiev exp pathol"] = "Methods and achievements in experimental pathology",
  ["methods appl anal"] = "Methods and Applications of Analysis",
  ["methods appl fluoresc"] = "Methods and applications in fluorescence",
  ["methods biochem anal"] = "Methods of Biochemical Analysis",
  ["methods cell biol"] = "Methods in cell biology",
  ["methods cell sci"] = "Methods in cell science : an official journal of the Society for In Vitro Biology",
  ["methods ecol evol"] = "Methods in ecology and evolution",
  ["methods enzymol"] = "Methods in enzymology",
  ["methods experiment phys"] = "Methods of Experimental Physics",
  ["methods find exp clin pharmacol"] = "Methods and findings in experimental and clinical pharmacology",
  ["methods funct anal topology"] = "Methods of Functional Analysis and Topology",
  ["methods geochem geophys"] = "Methods in Geochemistry and Geophysics",
  ["methods inf med"] = "Methods of information in medicine",
  ["methods inf med suppl"] = "Methods of information in medicine. Supplement",
  ["methods med res"] = "Methods in medical research",
  ["methods mol biol"] = "Methods in molecular biology (Clifton, N.J.)",
  ["methods mol cell biol"] = "Methods in molecular and cellular biology",
  ["methods mol med"] = "Methods in molecular medicine",
  ["methods oceanogr"] = "Methods in Oceanography",
  ["methods protoc"] = "Methods and protocols",
  ["methods rep rti press"] = "Methods report (RTI Press)",
  ["methods stat ecol"] = "Methods in Statistical Ecology",
  ["metode sci stud j"] = "Mètode science studies journal",
  ["metody diskret anal"] = "Metody Diskretnogo Analiza",
  ["metody mat model"] = "\\cyr Metody Matematicheskogo Modelirovaniya",
  ["metody optim reshenii"] = "Metody Optimalnykh Reshenii",
  ["metody vychisl"] = "Metody Vychislenii",
  ["metrop univ"] = "Metropolitan universities",
  ["metu stud devel"] = "Middle East Technical University Studies in Development",
  ["mex farm"] = "México farmacético",
  ["mex j biotechnol"] = "Mexican Journal Of Biotechnology",
  ["mex j sci res"] = "Mexican journal of scientific research",
  ["mex stud"] = "Mexican studies = Estudios mexicanos",
  ["mezzog eur"] = "Mezzogiorno d'Europa",
  ["mgh news"] = "MGH news",
  ["mgma connex"] = "MGMA connexion",
  ["mhq (ny ny)"] = "MHQ : the quarterly journal of military history",
  ["mian hua xue bao"] = "Mian hua xue bao = Acta gossypii Sinica",
  ["mian yi xue za zhi"] = "Mian yi xue za zhi = Journal of immunology",
  ["mich acad"] = "Michigan academician",
  ["mich academician"] = "Michigan Academician",
  ["mich appeals rep mich court appeals"] = "Michigan appeals reports : cases decided in the Michigan Court of Appeals. Michigan. Court of Appeals",
  ["mich compil laws annot mich"] = "Michigan compiled laws annotated. Michigan",
  ["mich health hosp"] = "Michigan health & hospitals",
  ["mich hist"] = "Michigan history",
  ["mich hist rev"] = "The Michigan historical review",
  ["mich hosp"] = "Michigan hospitals",
  ["mich j community serv learn"] = "Michigan journal of community service learning",
  ["mich j gend law"] = "Michigan journal of gender & law",
  ["mich jew hist"] = "Michigan Jewish history",
  ["mich law rev"] = "Michigan law review",
  ["mich math j"] = "Michigan Mathematical Journal",
  ["mich med"] = "Michigan medicine",
  ["mich nurse"] = "The Michigan nurse",
  ["mich public health"] = "Michigan public health",
  ["mich q rev"] = "Michigan quarterly review",
  ["mich state int law rev"] = "Michigan State international law review",
  ["mich statut annot mich"] = "Michigan statutes annotated. Michigan",
  ["michel serres mater futur"] = "Michel Serres and Material Futures",
  ["michies annot code public gen laws md md"] = "Michie's annotated code of the public general laws of Maryland. Maryland",
  ["michies w va code annot w va"] = "Michie's West Virginia code, annotated. West Virginia",
  ["michigan math j"] = "Michigan Mathematical Journal",
  ["micol aplicada int"] = "Micologia aplicada international",
  ["micol ital"] = "Micologia italiana",
  ["micol veg mediterr"] = "Micologia e vegetazione mediterranea",
  ["micologia toscana"] = "Micologia Toscana",
  ["micro nano eng"] = "Micro and Nano Engineering",
  ["micro nano lett"] = "Micro & nano letters",
  ["micro nano syst lett"] = "Micro and Nano Systems Letters",
  ["micro nanostruct"] = "Micro and Nanostructures",
  ["micro nanosyst"] = "Micro and Nanosystems",
  ["microarrays (basel)"] = "Microarrays (Basel, Switzerland)",
  ["microb bioact"] = "Microbial bioactives",
  ["microb biotechnol"] = "Microbial biotechnology",
  ["microb cell"] = "Microbial cell (Graz, Austria)",
  ["microb cell fact"] = "Microbial cell factories",
  ["microb comp genomics"] = "Microbial and Comparative Genomics",
  ["microb drug resist"] = "Microbial drug resistance (Larchmont, N.Y.)",
  ["microb ecol"] = "Microbial ecology",
  ["microb ecol health dis"] = "Microbial ecology in health and disease",
  ["microb genom"] = "Microbial genomics",
  ["microb genomics"] = "Microbial Genomics",
  ["microb inf exp"] = "Microbial Informatics and Experimentation",
  ["microb inform exp"] = "Microbial informatics and experimentation",
  ["microb pathog"] = "Microbial pathogenesis",
  ["microb pathogen"] = "Microbial Pathogenesis",
  ["microb physiol"] = "Microbial physiology",
  ["microb releases"] = "Microbial releases : viruses, bacteria, fungi",
  ["microb risk anal"] = "Microbial risk analysis",
  ["microbe wash dc"] = "Microbe (Washington, D.C.)",
  ["microbes environ"] = "Microbes and environments",
  ["microbes infect"] = "Microbes and infection",
  ["microbiol aust"] = "Microbiology Australia",
  ["microbiol biotechnol lett"] = "Microbiology and Biotechnology Letters",
  ["microbiol cult collect"] = "Microbiology And Culture Collections",
  ["microbiol educ"] = "Microbiology education",
  ["microbiol esp"] = "Microbiología española",
  ["microbiol immunol"] = "Microbiology and immunology",
  ["microbiol indones"] = "Microbiology Indonesia",
  ["microbiol insights"] = "Microbiology insights",
  ["microbiol mol biol rev"] = "Microbiology and molecular biology reviews : MMBR",
  ["microbiol open"] = "Microbiology Open",
  ["microbiol outlooks"] = "Microbiology Outlooks",
  ["microbiol parazitol epidemiol (bucur)"] = "Microbiologia, parazitologia, epidemiologia",
  ["microbiol res"] = "Microbiological research",
  ["microbiol res (pavia)"] = "Microbiology research",
  ["microbiol res j int"] = "Microbiology research journal international",
  ["microbiol resour announc"] = "Microbiology resource announcements",
  ["microbiol resour announce"] = "Microbiology Resource Announcements",
  ["microbiol rev"] = "Microbiological reviews",
  ["microbiol sci"] = "Microbiological sciences",
  ["microbiol spectr"] = "Microbiology spectrum",
  ["microbiol spectrum"] = "Microbiology Spectrum",
  ["microbiology (london, u k)"] = "Microbiology (London, United Kingdom)",
  ["microbiology (moscow, russ fed)"] = "Microbiology (Moscow, Russian Federation)(Translation of Mikrobiologiya)",
  ["microbiology (n y)"] = "Microbiology",
  ["microbiology (n y, ny, u s)"] = "Microbiology (New York, NY, United States)",
  ["microbiology (reading)"] = "Microbiology (Reading, England)",
  ["microbiology (reading, u k)"] = "Microbiology (Reading, United Kingdom)",
  ["microbiology (road town, v i (br))"] = "Microbiology (Road Town, Virgin Islands (British))",
  ["microchem j"] = "Microchemical Journal",
  ["microchim acta"] = "Microchimica Acta",
  ["microcirc endothelium lymphatics"] = "Microcirculation, endothelium, and lymphatics",
  ["microelectron eng"] = "Microelectronic Engineering",
  ["microelectron int"] = "Microelectronics International",
  ["microelectron j"] = "Microelectronics Journal",
  ["microelectron reliab"] = "Microelectronics Reliability",
  ["microelectronics j"] = "Microelectronics journal",
  ["microfluid nanofluid"] = "Microfluidics and Nanofluidics",
  ["microfluid nanofluidics"] = "Microfluidics and nanofluidics",
  ["microgram j"] = "Microgram Journal",
  ["microgravity q"] = "Microgravity quarterly : MGQ",
  ["microgravity sci technol"] = "Microgravity science and technology",
  ["micromachines (basel)"] = "Micromachines",
  ["micron microsc acta"] = "Micron and Microscopica Acta",
  ["microphysiol syst"] = "Microphysiological systems",
  ["microplast nanoplast"] = "Microplastics and Nanoplastics",
  ["microporous mater"] = "Microporous Materials",
  ["microporous mesoporous mater"] = "Microporous and Mesoporous Materials",
  ["microprocess microsyst"] = "Microprocessors and Microsystems",
  ["micropubl biol"] = "microPublication biology",
  ["microrna diagn ther"] = "microRNA diagnostics and therapeutics",
  ["micros today"] = "Microscopy today",
  ["microsc acta"] = "Microscopica acta",
  ["microsc acta suppl"] = "Microscopica acta. Supplement",
  ["microsc anal"] = "Microscopy and Analysis",
  ["microsc anal (am ed)"] = "Microscopy and analysis (Americas ed.)",
  ["microsc electron biol celular"] = "Microscopía electrónica y biología celular : organo oficial de las Sociedades Latinoamericana de Microscopía Electrónica e Iberoamericana de Biología Celular",
  ["microsc j struct rec"] = "Microscopic Journal and Structural Record",
  ["microsc microanal"] = "Microscopy and microanalysis : the official journal of Microscopy Society of America, Microbeam Analysis Society, Microscopical Society of Canada",
  ["microsc res tech"] = "Microscopy research and technique",
  ["microsc today"] = "Microscopy Today",
  ["microscale thermophys eng"] = "Microscale Thermophysical Engineering",
  ["microscopy (oxf)"] = "Microscopy (Oxford, England)",
  ["microsyst nanoeng"] = "Microsystems & nanoengineering",
  ["microsyst technol"] = "Microsystem Technologies",
  ["microvasc res"] = "Microvascular research",
  ["microw j (int ed)"] = "Microwave journal",
  ["microw opt technol lett"] = "Microwave and optical technology letters",
  ["microwave opt technol lett"] = "Microwave and Optical Technology Letters",
  ["mic—model identif control"] = "Modeling, Identification and Control",
  ["mid am"] = "Mid-America; an historical review",
  ["mid am rev sociol"] = "Mid-American review of sociology",
  ["midas j"] = "MIDAS journal",
  ["midcont j archaeol"] = "Midcontinental journal of archaeology, MCJA",
  ["middle east afr j ophthalmol"] = "Middle East African journal of ophthalmology",
  ["middle east bus econ rev"] = "Middle East Business and Economic Review",
  ["middle east crit"] = "Middle East critique",
  ["middle east dent oral health"] = "Middle East dentistry & oral health",
  ["middle east dev j"] = "Middle East development journal",
  ["middle east econ pap"] = "Middle East Economic Papers",
  ["middle east fertil soc j"] = "Middle East Fertility Society journal",
  ["middle east health"] = "Middle East health",
  ["middle east j"] = "The Middle East journal",
  ["middle east j anaesthesiol"] = "Middle East journal of anaesthesiology",
  ["middle east j anesthesiol"] = "Middle East Journal of Anesthesiology",
  ["middle east j cancer"] = "Middle East journal of cancer",
  ["middle east j dig dis"] = "Middle East journal of digestive diseases",
  ["middle east j sci res"] = "Middle East journal of scientific research",
  ["middle east law gov"] = "Middle East law and governance",
  ["middle east med j"] = "The Middle East medical journal",
  ["middle east rep"] = "Middle East report (New York, N.Y.)",
  ["middle east stud"] = "Middle Eastern studies",
  ["middle east stud assoc bull"] = "Middle East Studies Association bulletin",
  ["middx hosp j"] = "The Middlesex Hospital journal",
  ["middx vet"] = "The Middlesex veterinarian",
  ["midl hist"] = "Midland history",
  ["midstream (n y)"] = "Midstream; a monthly Jewish review",
  ["midwest alliance nurs j"] = "Midwest Alliance in Nursing journal",
  ["midwest dent"] = "Midwestern dentist",
  ["midwest drug"] = "Midwestern druggist",
  ["midwest educ res (akron)"] = "Mid-western educational researcher (Akron, Ohio : 1991)",
  ["midwest med ethics"] = "Midwest medical ethics : a publication of the Midwest Bioethics Center",
  ["midwest q (pittsb)"] = "The Midwest quarterly",
  ["midwest stud philos"] = "Midwest studies in philosophy",
  ["midwife health visit"] = "Midwife and health visitor",
  ["midwife health visit community nurse"] = "Midwife, health visitor & community nurse",
  ["midwifery today childbirth educ"] = "Midwifery today and childbirth education",
  ["midwifery today int midwife"] = "Midwifery today with international midwife",
  ["midwives (1995)"] = "Midwives : official journal of the Royal College of Midwives",
  ["midwives chron"] = "Midwives chronicle",
  ["mie med j"] = "Mie medical journal",
  ["mieux compr"] = "Mieux Comprende",
  ["migr dev"] = "Migration and development",
  ["migr news"] = "Migration news",
  ["migr stud"] = "Migration studies",
  ["migr teme"] = "Migracijske teme",
  ["migr today"] = "Migration today (Staten Island, New York, N.Y.)",
  ["migr world mag"] = "Migration world magazine",
  ["migrat lett"] = "Migration letters : an international journal of migration studies",
  ["mikol fitopatol"] = "Mikologiia i fitopatologiia",
  ["mikrobiol biotehnol"] = "Mìkrobìologìâ ì bìotehnologìâ",
  ["mikrobiol z"] = "Mikrobiolohichnyĭ zhurnal (Kiev, Ukraine : 1993)",
  ["mikrobiol zh"] = "Mikrobiologicheskiĭ zhurnal",
  ["mikrobiyol bul"] = "Mikrobiyoloji bülteni",
  ["mikrobn biotekhnologii"] = "Mikrobnye biotekhnologii",
  ["mikrochim acta"] = "Mikrochimica acta",
  ["mikrochim ichnoanal acta"] = "Mikrochimica et ichnoanalytica acta",
  ["mil aff"] = "Military affairs : journal of the American Military Institute",
  ["mil behav health"] = "Military behavioral health",
  ["mil eng"] = "The Military engineer",
  ["mil law rev"] = "Military law review",
  ["mil med"] = "Military medicine",
  ["mil med res"] = "Military Medical Research",
  ["mil oper res"] = "Military Operations Research",
  ["mil psychol"] = "Military psychology : the official journal of the Division of Military Psychology, American Psychological Association",
  ["mil rev"] = "Military review",
  ["mil surg"] = "Military surgeon",
  ["milan j math"] = "Milan Journal of Mathematics",
  ["milbank mem fund q"] = "The Milbank Memorial Fund quarterly",
  ["milbank mem fund q health soc"] = "The Milbank Memorial Fund quarterly. Health and society",
  ["milbank q"] = "The Milbank quarterly",
  ["milit collect hist"] = "Military collector & historian",
  ["milit hist tex southwest"] = "Military history of Texas and the Southwest",
  ["militargesch z"] = "Militärgeschichtliche Zeitschrift",
  ["milw hist"] = "Milwaukee history",
  ["milwaukee med times"] = "Milwaukee medical times",
  ["mimeo ser"] = "Mimeo series (University of North Carolina (System). Institute of Statistics)",
  ["mims mag"] = "Mims magazine weekly : the journal of quality patient care",
  ["min eng"] = "Mining Engineering",
  ["min guo dang an"] = "Min guo dang an",
  ["min inf anal bull"] = "Mining Informational and Analytical Bulletin",
  ["min intell knowl explor (2015)"] = "Mining intelligence and knowledge exploration : third International Conference, MIKE 2015, Hyderabad, India, December 9-11, 2015. Proceeedings. MIKE (Conference) (3rd : 2015 : Hyderabad, India)",
  ["min j"] = "Mining Journal",
  ["min mag"] = "Mining Magazine",
  ["min metall explor"] = "Mining, metallurgy & exploration",
  ["min miner deposits"] = "Mining Of Mineral Deposits",
  ["min resour eng"] = "Mineral Resources Engineering",
  ["min sci press"] = "Mining and Scientific Press",
  ["min sci technol"] = "Mining Science and Technology",
  ["min technol"] = "Mining Technology",
  ["mind assoc occas ser"] = "Mind Association Occasional Series",
  ["mind brain"] = "Mind & brain : the journal of psychiatry",
  ["mind brain educ"] = "Mind, brain and education : the official journal of the International Mind, Brain, and Education Society",
  ["mind cult act"] = "Mind, culture, and activity",
  ["mind lang"] = "Mind & language",
  ["mind mean metaphys"] = "Mind, Meaning and Metaphysics",
  ["mind ment health mag"] = "Mind and mental health magazine",
  ["mindfulness (n y)"] = "Mindfulness",
  ["mindfulness compassion"] = "Mindfulness & compassion",
  ["minds mach (dordr)"] = "Minds and machines",
  ["mine water environ"] = "Mine Water and the Environment",
  ["miner depos"] = "Mineralium deposita",
  ["miner deposita"] = "Mineralium Deposita",
  ["miner econ"] = "Mineral Economics",
  ["miner electrolyte metab"] = "Mineral and electrolyte metabolism",
  ["miner eng"] = "Minerals Engineering",
  ["miner metall process"] = "Minerals and Metallurgical Processing",
  ["miner process extr metall"] = "Transactions. Section C, Mineral processing & extractive metallurgy",
  ["miner process extr metall rev"] = "Mineral Processing and Extractive Metallurgy Review",
  ["miner slovaca"] = "Mineralia Slovaca",
  ["mineral mag"] = "Mineralogical Magazine",
  ["mineral petrol"] = "Mineralogy and Petrology",
  ["mineral rec"] = "Mineralogical Record",
  ["mineral zh"] = "Mineralogicheskii Zhurnal",
  ["minerals (basel)"] = "Minerals (Basel, Switzerland)",
  ["minerva (arlingt va)"] = "Minerva (Arlington, Va.)",
  ["minerva anestesiol"] = "Minerva anestesiologica",
  ["minerva biotechnol biomol res"] = "Minerva Biotechnology and Biomolecular Research",
  ["minerva biotecnol"] = "Minerva biotecnologica",
  ["minerva cardioangiol"] = "Minerva cardioangiologica",
  ["minerva cardioangiol eur"] = "Minerva cardioangiologica Europea",
  ["minerva chir"] = "Minerva chirurgica",
  ["minerva dermatol"] = "Minerva dermatologica",
  ["minerva dietol"] = "Minerva dietologica",
  ["minerva dietol gastroenterol"] = "Minerva dietologica e gastroenterologica",
  ["minerva ecol idroclimatol fis sanit"] = "Minerva ecologica idroclimatologica fisicosanitaria",
  ["minerva ecol idroclimatol fisiconucl"] = "Minerva ecologica idroclimatologica e fisiconucleare",
  ["minerva endocrinol"] = "Minerva endocrinologica",
  ["minerva farm"] = "Minerva farmaceutica",
  ["minerva fisiconucl"] = "Minerva fisiconucleare",
  ["minerva fisioter radiobiol"] = "Minerva fisioterapica e radiobiologica",
  ["minerva gastroenterol"] = "Minerva gastroenterologica",
  ["minerva gastroenterol dietol"] = "Minerva gastroenterologica e dietologica",
  ["minerva ginecol"] = "Minerva ginecologica",
  ["minerva med"] = "Minerva medica",
  ["minerva medicoleg"] = "Minerva medicolegale; archivio di antropologia criminale, psichiatria, e medicina legale",
  ["minerva nefrol"] = "Minerva nefrologica",
  ["minerva neurochir"] = "Minerva neurochirurgica",
  ["minerva nipiol"] = "Minerva nipiologica",
  ["minerva nucl"] = "Minerva nucleare",
  ["minerva oftalmol"] = "Minerva oftalmologica",
  ["minerva ortognatod"] = "Minerva ortognatodontica",
  ["minerva ortop"] = "Minerva ortopedica",
  ["minerva ortop traumatol"] = "Minerva ortopedica e traumatologica : organo ufficiale della Società piemontese-ligure-lombarda di ortopedia e traumatologia",
  ["minerva otorinolaringol"] = "Minerva otorinolaringologica",
  ["minerva pediatr"] = "Minerva pediatrica",
  ["minerva pneumol"] = "Minerva pneumologica",
  ["minerva psichiatr"] = "Minerva psichiatrica",
  ["minerva radiol"] = "Minerva radiologica",
  ["minerva radiol fisioter radiobiol"] = "Minerva radiologica, fisioterapica e radiobiologica",
  ["minerva stomatol"] = "Minerva stomatologica",
  ["minerva urol"] = "Minerva urologica",
  ["minerva urol nefrol"] = "Minerva urologica e nefrologica = The Italian journal of urology and nephrology",
  ["mines mach"] = "Minds and Machines",
  ["ming stud"] = "Ming studies",
  ["mini rev med chem"] = "Mini reviews in medicinal chemistry",
  ["mini rev org chem"] = "Mini-reviews in organic chemistry",
  ["mini-rev med chem"] = "Mini-Reviews in Medicinal Chemistry",
  ["mini-rev org chem"] = "Mini-Reviews in Organic Chemistry",
  ["minia j agric res dev"] = "Minia journal of agricultural research and development",
  ["minim invasive neurosurg"] = "Minimally invasive neurosurgery : MIN",
  ["minim invasive surg"] = "Minimally invasive surgery",
  ["minim invasive surg nurs"] = "Minimally invasive surgical nursing",
  ["minim invasive ther allied technol"] = "Minimally invasive therapy & allied technologies : MITAT : official journal of the Society for Minimally Invasive Therapy",
  ["minimally invasive ther allied technol"] = "Minimally Invasive Therapy & Allied Technologies",
  ["minimax theory appl"] = "Minimax Theory and its Applications",
  ["minist ed sci cult japan rep grant aid sci res prior areas"] = "Ministry of Education, Science and Culture, Japan. Report of Grant-in-Aid for Scientific Research on Priority Areas",
  ["minn hist"] = "Minnesota history",
  ["minn j law sci technol"] = "Minnesota journal of law, science & technology",
  ["minn law rev"] = "Minnesota law review",
  ["minn med"] = "Minnesota medicine",
  ["minn nurs accent"] = "Minnesota nursing accent",
  ["minn pharm"] = "Minnesota pharmacist",
  ["minn statut annot minn"] = "Minnesota statutes annotated. Minnesota",
  ["minn stud philos sci"] = "Minnesota Studies in The Philosophy of Science",
  ["minn symp child psychol ser"] = "Minnesota symposia on child psychology (Series)",
  ["minn welfare"] = "Minnesota welfare",
  ["minneap dist dent j"] = "The Minneapolis district dental journal",
  ["minnetesol j (2014)"] = "MinneTESOL journal (2014)",
  ["minor nurse"] = "Minority nurse",
  ["minor nurse newsl"] = "Minority nurse newsletter",
  ["minor planet bull"] = "The Minor planet bulletin",
  ["minufiya j agric res"] = "Minufiya journal of agricultural research",
  ["minzoku eisei"] = "[Minzoku eisei] Race hygiene",
  ["mir nauki kult obraz"] = "Mir nauki, kulʻtury, obrazovaniâ",
  ["mircen j appl microbiol biotechnol"] = "MIRCEN journal of applied microbiology and biotechnology",
  ["mires peat"] = "Mires and Peat",
  ["mirovaia ekon mezdunar otnosheiia"] = "Mirovai︠a︡ ėkonomika i mezhdunarodnye otnoshenii︠a︡",
  ["mirror (wash d c)"] = "Mirror",
  ["mis q"] = "MIS quarterly : management information systems",
  ["misaengmul gwa san eob"] = "Misaengmul kwa sanŏp = The microorganisms and industry",
  ["misaengmul hakhoe chi"] = "Misaengmul Hakhoe chi (Seoul, Korea : 1997)",
  ["misainmurhag hoiji"] = "Misaengmul Hakhoe chi",
  ["misc estud arabes hebraicos"] = "Miscelánea de estudios árabes y hebraicos",
  ["misc pap"] = "Miscellaneous papers (Centre for Entomological Studies (Ankara, Turkey))",
  ["misc publ mus zool univ mich"] = "Miscellaneous publications (University of Michigan. Museum of Zoology)",
  ["miscelanea mat"] = "Miscelanea Matematica",
  ["miscelánea mat"] = "Miscelánea Matemática",
  ["mises point accouch pediatre"] = "Mises au point de l'accoucheur et du pédiatre",
  ["miskolc math notes"] = "Miskolc Mathematical Notes. A Publication of the University of Miskolc",
  ["miss coll law rev"] = "Mississippi College law review. Mississippi College. School of Law",
  ["miss dent assoc j"] = "Mississippi Dental Association journal",
  ["miss doct"] = "The Mississippi doctor",
  ["miss law j"] = "Mississippi law journal",
  ["miss law rev"] = "Missouri law review",
  ["miss q"] = "The Mississippi quarterly",
  ["miss rn"] = "Mississippi RN",
  ["miss val j bus econ"] = "Mississippi Valley Journal of Business and Economics",
  ["miss valley med j"] = "Mississippi Valley medical journal (Quincy, Ill)",
  ["mission hisp"] = "Missionalia hispánica",
  ["missouri j math sci"] = "Missouri Journal of Mathematical Sciences",
  ["mit linc lab ser"] = "MIT Lincoln Laboratory Series",
  ["mit press essent knowl ser"] = "MIT Press Essential Knowledge Series",
  ["mit press ser artif intell"] = "MIT Press Series in Artif. Intell.",
  ["mit press ser econ learn soc evol"] = "MIT Press Series on Economic Learning and Social Evolution",
  ["mit press ser logic program"] = "MIT Press Series in Logic Programming",
  ["mit press ser structural mech"] = "The MIT Press Series in Structural Mechanics",
  ["mit sloan manag rev"] = "MIT Sloan management review",
  ["mitig adapt strateg glob chang"] = "Mitigation and adaptation strategies for global change",
  ["mitigation adapt strategies global change"] = "Mitigation and Adaptation Strategies for Global Change",
  ["mitochondrial commun"] = "Mitochondrial Communications",
  ["mitochondrial dna"] = "Mitochondrial DNA",
  ["mitochondrial dna a dna mapp seq anal"] = "Mitochondrial DNA. Part A, DNA mapping, sequencing, and analysis",
  ["mitochondrial dna b resour"] = "Mitochondrial DNA. Part B, Resources",
  ["mitochondrial dna part a"] = "Mitochondrial DNA Part A DNA Mapping, Sequencing, and Analysis",
  ["mitochondrial dna part b"] = "Mitochondrial DNA Part B Resources",
  ["mits technol rev"] = "MIT's technology review",
  ["mitt aargau natforsch ges"] = "Mitteilungen der Aargauischen naturforschenden Gesellschaft",
  ["mitt abt bot landesmus joanneum graz"] = "Mitteilungen der Abteilung für Botanik am Landesmuseum Joanneum in Graz",
  ["mitt abt forstl biom univ freibg"] = "Mitteilungen der Abteilung für Forstliche Biometrie der Universität Freiburg",
  ["mitt arbeitsmarkt berufsforsch"] = "Mitteilungen aus der Arbeitsmarkt- und Berufsforschung",
  ["mitt arbgem geobot schlesw-holst hambg"] = "Mitteilungen der Arbeitsgemeinschaft Geobotanik in Schleswig-Holstein und Hamburg",
  ["mitt arbkr wald wasser"] = "Mitteilungen des Arbeitskreises Wald und Wasser",
  ["mitt bad landesver natkd natschutz"] = "Mitteilungen des Badischen Landesvereins für Naturkunde und Naturschutz",
  ["mitt biol bundesanst land- forstwirtsch, berl-dahl"] = "Mitteilungen aus der Biologischen Bundesanstalt für Land- und Forstwirtschaft, Berlin-Dahlem",
  ["mitt bundesforschanst forst-  holzwirtsch"] = "Mitteilungen der Bundesforschungsanstalt für Forst- und Holzwirtschaft",
  ["mitt dtsch archaeol inst abt kairo"] = "Mitteilungen des Deutschen Archäologischen Instituts. Deutsches Archäologisches Institut. Abteilung Kairo",
  ["mitt dtsch bodenkdl ges"] = "Mitteilungen der Deutschen Bodenkundlichen Gesellschaft",
  ["mitt dtsch dendrol ges"] = "Mitteilungen der Deutschen dendrologischen Gesellschaft",
  ["mitt dtsch ges allg angew entomol"] = "Mitteilungen der Deutschen Gesellschaft für Allgemeine und Angewandte Entomologie",
  ["mitt dtsch malakozool ges"] = "Mitteilungen der Deutschen Malakozoologischen Gesellschaft",
  ["mitt dtsch math-ver"] = "Mitteilungen der Deutschen Mathematiker-Vereinigung",
  ["mitt dtsch pharm ges pharm ges ddr"] = "Mitteilungen. Deutsche Pharmazeutische Gesellschaft",
  ["mitt eidgenöss anst forstl verswes"] = "Mitteilungen Eidg. Anstalt für das forstliche Versuchswesen",
  ["mitt eidgenöss forschanst wald schnee landsch"] = "Mitteilungen der Eidg. Forschungsanstalt für Wald, Schnee und Landschaft",
  ["mitt eidgenöss inst schnee- lawinenforsch"] = "Mitteilungen des Eidg. Institutes für Schnee- und Lawinenforschung",
  ["mitt entomol ges basel"] = "Mitteilungen der Entomologischen Gesellschaft Basel",
  ["mitt florist-soziol arbgem"] = "Mitteilungen der Floristisch-soziologischen Arbeitsgemeinschaft",
  ["mitt forschungsbeitr cusanus ges"] = "Mitteilungen und Forschungsbeiträge der Cusanus-Gesellschaft",
  ["mitt forstl bundes-versanst wien"] = "Mitteilungen der Forstlichen Bundes-Versuchsanstalt Wien",
  ["mitt forstl vers- forschanst baden-württ"] = "Mitteilungen der Forstlichen Versuchs- und Forschungsanstalt Baden-Württemberg",
  ["mitt geb lebensmittelunters hyg"] = "Mitteilungen aus dem Gebiete der Lebensmittel-untersuchung un Hygiene = Travaux de chimie alimentaire et d'hygiène",
  ["mitt geogr-ethnogr ges zür"] = "Mitteilungen der Geographisch-Ethnographischen Gesellschaft Zürich",
  ["mitt ges angew math mech"] = "Mitteilungen der Gesellschaft für Angewandte Mathematik und Mechanik",
  ["mitt ges biblwes dok landbaues"] = "Mitteilungen der Gesellschaft für Bibliothekswesen und Dokumentation des Landbaues",
  ["mitt ges salzburger landeskd"] = "Mitteilungen der Gesellschaft für Salzburger Landeskunde",
  ["mitt gesell dtscher chem fachgruppe gesch chem"] = "Mitteilungen - Gesellschaft Deutscher Chemiker - Fachgruppe Geschichte der Chemie",
  ["mitt hess landesforstverwalt"] = "Mitteilungen der Hessischen Landesforstverwaltung",
  ["mitt hist ver kanton schwyz"] = "Mitteilungen des Historischen Vereins des Kantons Schwyz",
  ["mitt inst allg bot hamb"] = "Mitteilungen aus dem Institut für Allgemeine Botanik Hamburg. Institut für Allgemeine Botanik Hamburg",
  ["mitt inst orientforsch"] = "Mitteilungen des Instituts für Orientforschung. Deutsche Akademie der Wissenschaften zu Berlin. Institut für Orientforschung",
  ["mitt int bodenkdl ges"] = "Mitteilungen der Internationalen bodenkundlichen Gesellschaft",
  ["mitt int ver theor angew limnol"] = "Mitteilungen Internationale Vereinigung fur Theoretische und Angewandte Limnologie",
  ["mitt klosterneuburg rebe wein obstb fruchteverwert"] = "Mitteilungen Klosterneuburg : Rebe und Wein, Obstbau und Früchteverwertung",
  ["mitt krems stadtarch"] = "Mitteilungen des Kremser Stadtarchivs. Krems an der Donau (Austria). Stadtarchiv",
  ["mitt kurat waldarb forsttech"] = "Mitteilungen des Kuratoriums für Waldarbeit und Forsttechnik",
  ["mitt lebensmittelunters hyg"] = "Mitteilungen aus Lebensmitteluntersuchung und Hygiene = Travaux de chimie alimentaire et d'hygiène",
  ["mitt math ges hamburg"] = "Mitteilungen der Mathematischen Gesellschaft in Hamburg",
  ["mitt math sem giessen"] = "Mitteilungen aus dem Mathematischen Seminar Giessen",
  ["mitt med abt schweiz unfallversicherunst"] = "Mitteilungen. Schweizerische Unfallversicherungsanstalt. Medizinische Abteilung",
  ["mitt munch entomol ges"] = "Mitteilungen der Münchner Entomologischen Gesellschaft. Münchner Entomologische Gesellschaft",
  ["mitt mus nat berl foss rec"] = "Mitteilungen aus dem Museum für Naturkunde in Berlin. Fossil record",
  ["mitt mus naturkunde berl dtsch entomol z"] = "Mitteilungen aus dem museum für naturkunde in Berlin. Deutsche entomologische zeitschrift",
  ["mitt mus naturkunde berl zoolog reihe"] = "Mitteilungen aus dem Museum für Naturkunde in Berlin. Zoologische Reihe",
  ["mitt natforsch ges bern"] = "Mitteilungen der Naturforschenden Gesellschaft in Bern",
  ["mitt natforsch ges kanton glarus"] = "Mitteilungen der Naturforschenden Gesellschaft des Kantons Glarus",
  ["mitt natforsch ges kanton solothurn"] = "Mitteilungen der Naturforschenden Gesellschaft des Kantons Solothurn",
  ["mitt natforsch ges luzern"] = "Mitteilungen der Naturforschenden Gesellschaft Luzern",
  ["mitt natforsch ges schaffhausen"] = "Mitteilungen der Naturforschenden Gesellschaft Schaffhausen",
  ["mitt naturforsch ges bern"] = "Mitteilungen der Naturforschenden Gesellschaft in Bern. Naturforschende Gesellschaft in Bern",
  ["mitt natwiss ges thun"] = "Mitteilungen der Naturwissenschaftlichen Gesellschaft Thun",
  ["mitt natwiss ges winterthur"] = "Mitteilungen der Naturwissenschaftlichen Gesellschaft Winterthur",
  ["mitt niedersächs forstl versanst"] = "Mitteilungen der Niedersächsischen forstlichen Versuchsanstalt",
  ["mitt norddtsch natschutzakad"] = "Mitteilungen aus der Norddeutschen Naturschutzakademie (NNA)",
  ["mitt osterr geogr ges"] = "Mitteilungen der Österreichischen Geographischen Gesellschaft. Österreichische Geographische Gesellschaft",
  ["mitt osterr ges tropenmed parasitol"] = "Mitteilungen der Osterreichischen Gesellschaft fur Tropenmedizin und Parasitologie",
  ["mitt osterr ges tropmed parasitol"] = "Mitteilungen der Österreichischen Gesellschaft für Tropenmedizin und Parasitologie",
  ["mitt osterr sanitatsverwalt"] = "Mitteilungen der österreichischen Sanitätsverwaltung",
  ["mitt pollichia"] = "Mitteilungen der Pollichia",
  ["mitt region rechenzentrums erlangen"] = "Mitteilungsblatt des Regionalen Rechenzentrums Erlangen",
  ["mitt schweiz entomol ges"] = "Mitteilungen der Schweizerischen entomologischen Gesellschaft = Bulletin de la Société entomologique suisse. Schweizerische Entomologische Gesellschaft",
  ["mitt staatsforstverwalt bayerns"] = "Mitteilungen aus der Staatsforstverwaltung Bayerns",
  ["mitt thurgau natforsch ges"] = "Mitteilungen der Thurgauischen Naturforschenden Gesellschaft",
  ["mitt ver forstl standortskd forstpflanzenzücht"] = "Mitteilungen des Vereins für forstliche Standortskunde und Forstpflanzenzüchtung",
  ["mitt versanst wasserbau hydrol glaziol eidgenöss tech hochsch zür"] = "Mitteilungen der Versuchsanstalt für Wasserbau, Hydrologie und Glaziologie an der Eidg. Technischen Hochschule Zürich",
  ["mittbl osterr ges stat inform"] = "Mitteilungsblatt. Österreichische Gesellschaft für Statistik und Informatik",
  ["mitteilungen arbeitsmarkt- berufsforschung"] = "Mitteilungen aus der Arbeitsmarkt- und Berufsforschung",
  ["mitteilungsbl - chem ges d d r"] = "Mitteilungsblatt - Chemische Gesellschaft der Deutschen Demokratischen Republik",
  ["mitteldtsch famkd"] = "Mitteldeutsche Familienkunde",
  ["mittellat jahrb"] = "Mittellateinisches Jahrbuch",
  ["mixhs 12 (2012)"] = "MIX-HS'12 : proceedings of the 2nd International Workshop on Managing Interoperability and Complexity in Health Systems October 29, 2012, Maui, Hawaii, USA. International Workshop on Managing Interoperability and Complexity in Health Systems (2nd : 2012 : Maui, Hawaii)",
  ["mizunamishi kaseki hakubutsukan kenkyu hokoku"] = "Mizunami-shi Kaseki Hakubutsukan kenkyū hōkoku. Bulletin of the Mizunami Fossil Museum. Mizunami-shi Kaseki Hakubutsukan",
  ["mla dir period"] = "Mla Directory Of Periodicals",
  ["mlm rep"] = "MLM [reports]. U.S. Atomic Energy Commission",
  ["mln bull"] = "MLN bulletin",
  ["mlo med lab obs"] = "MLO: medical laboratory observer",
  ["mlq math log q"] = "MLQ. Mathematical Logic Quarterly",
  ["mmhealth16 (2016)"] = "MMHealth'16 : proceedings of the 2016 ACM Workshop on Multimedia for Personal Health and Health Care : October 16, 2016, Amsterdam, The Netherlands. ACM Workshop on Multimedia for Personal Health and Health Care (1st : 2016 : Amsterdam, Netherlands)",
  ["mmhealth17 (2017)"] = "MMHealth'17 : proceedings of the 2nd International Workshop on Multimedia for Personal Health and Health Care : October 23, 2017, Mountain View, CA, USA. ACM Workshop on Multimedia for Personal Health and Health Care (2nd : 2017 : Mountain View, Calif.)",
  ["mmw fortschr med"] = "MMW Fortschritte der Medizin",
  ["mmw munch med wochenschr"] = "MMW, Münchener medizinische Wochenschrift",
  ["mmwr cdc surveill summ"] = "MMWR. CDC surveillance summaries : Morbidity and mortality weekly report. CDC surveillance summaries",
  ["mmwr morb mortal wkly rep"] = "MMWR. Morbidity and mortality weekly report",
  ["mmwr recomm rep"] = "MMWR. Recommendations and reports : Morbidity and mortality weekly report. Recommendations and reports",
  ["mmwr suppl"] = "MMWR supplements",
  ["mmwr surveill summ"] = "Morbidity and mortality weekly report. Surveillance summaries (Washington, D.C. : 2002)",
  ["mna accent"] = "MNA accent",
  ["mnemosyne bibl class batava suppl"] = "Mnemosyne. Bibliotheca Classica Batava. Supplementum",
  ["mni open res"] = "MNI open research",
  ["mo dent j"] = "Missouri dental journal (Jefferson City, Mo.)",
  ["mo hist rev"] = "Missouri historical review",
  ["mo j math sci"] = "Missouri Journal of Mathematical Sciences",
  ["mo med"] = "Missouri medicine",
  ["mo nurse"] = "The Missouri nurse",
  ["mob dna"] = "Mobile DNA",
  ["mob genet elements"] = "Mobile genetic elements",
  ["mob media commun"] = "Mobile media & communication",
  ["mobile dna"] = "Mobile DNA",
  ["mobile inf syst"] = "Mobile Information Systems",
  ["mobile networks appl"] = "Mobile Networks and Applications",
  ["moct-most: econ pol transitional economies"] = "MOCT-MOST: Economic Policy in Transitional Economies",
  ["mod  pathol"] = "Modern Pathology",
  ["mod age"] = "Modern age (Chicago, Ill.)",
  ["mod agric"] = "Modern Agriculture",
  ["mod appl sci"] = "Modern applied science",
  ["mod asian stud"] = "Modern Asian studies",
  ["mod birkhauser class"] = "Modern Birkhauser Classics",
  ["mod cell biol"] = "Modern cell biology",
  ["mod china"] = "Modern China",
  ["mod churchm"] = "The Modern churchman",
  ["mod clin med res"] = "Modern clinical medicine research",
  ["mod concepts cardiovasc dis"] = "Modern concepts of cardiovascular disease",
  ["mod contemp fr"] = "Modern & contemporary France",
  ["mod dent (n y n y 1931)"] = "Modern dentistry",
  ["mod dent pract"] = "Modern dental practice",
  ["mod deutung aristot log"] = "Zur Modernen Deutung der Aristotelischen Logik",
  ["mod drug discovery"] = "Modern Drug Discovery",
  ["mod electron mater"] = "Modern Electronic Materials",
  ["mod fiction stud"] = "Modern fiction studies",
  ["mod french identities"] = "Modern French Identities",
  ["mod genet"] = "Modern genetics",
  ["mod greek stud yearb"] = "Modern Greek studies yearbook",
  ["mod healthc"] = "Modern healthcare",
  ["mod healthc (short term care)"] = "Modern healthcare. [Short-term care ed.]",
  ["mod hosp"] = "Modern hospital",
  ["mod italy"] = "Modern Italy : journal of the Association for the Study of Modern Italy",
  ["mod jud"] = "Modern Judaism",
  ["mod lang notes"] = "Modern language notes",
  ["mod law rev"] = "The Modern law review",
  ["mod med"] = "Modern medicine",
  ["mod med (chic)"] = "Modern medicine",
  ["mod med asia"] = "Modern medicine of Asia",
  ["mod med can"] = "Modern medicine of Canada",
  ["mod midwife"] = "Modern midwife",
  ["mod nurs home"] = "Modern nursing home",
  ["mod pathol"] = "Modern pathology : an official journal of the United States and Canadian Academy of Pathology, Inc",
  ["mod pharm"] = "Modern pharmacy",
  ["mod philol"] = "Modern philology",
  ["mod phys lett a"] = "Modern Physics Letters A",
  ["mod phys lett b"] = "Modern Physics Letters B",
  ["mod probab stat"] = "Modern Probability and Statistics",
  ["mod probl ophthalmol"] = "Modern problems in ophthalmology",
  ["mod probl paediatr"] = "Modern problems in paediatrics",
  ["mod probl pharmacopsychiatry"] = "Modern problems of pharmacopsychiatry",
  ["mod res catal"] = "Modern Research in Catalysis",
  ["mod res inflamm"] = "Modern research in inflammation",
  ["mod rheumatol"] = "Modern rheumatology",
  ["mod rheumatol case rep"] = "Modern rheumatology case reports",
  ["mod schoolman"] = "The Modern schoolman",
  ["mod stoch theory appl"] = "Modern Stochastics. Theory and Applications",
  ["mod stud lit cult"] = "Modernist studies: literature & culture 1920-1940",
  ["mod theol"] = "Modern theology",
  ["mod treat"] = "Modern treatment",
  ["mod treat yearb"] = "Modern treatment yearbook",
  ["mod trends financ eng"] = "Modern Trends in Financial Engineering",
  ["mod trends hum reprod physiol"] = "Modern trends in human reproductive physiology",
  ["mod trends immunol"] = "Modern trends in immunology",
  ["mod trends med virol"] = "Modern trends in medical virology",
  ["mod trends neurol"] = "Modern trends in neurology",
  ["mod trends orthop"] = "Modern trends in orthopaedics",
  ["mod trends pharmacopsychiatry"] = "Modern trends in pharmacopsychiatry",
  ["mod trends plast surg"] = "Modern trends in plastic surgery",
  ["mod trends psychiatry"] = "Modern trends in psychiatry",
  ["mod trends radiother"] = "Modern trends in radiotherapy",
  ["mod trends rheumatol"] = "Modern trends in rheumatology",
  ["mod trends surg"] = "Modern trends in surgery",
  ["mod vet pract"] = "Modern veterinary practice",
  ["model anal inf sist"] = "Modelirovanie i Analiz Informatsionnykh Sistem. Modeling and Analysis of Information Systems",
  ["model assist stat appl"] = "Model assisted statistics and applications : an international journal",
  ["model assisted stat appl"] = "Model Assisted Statistics and Applications",
  ["model dyn syst"] = "Modeling Dynamic Systems",
  ["model earth syst environ"] = "Modeling earth systems and environment",
  ["model mekh"] = "Rossíiskaya Akademiya Nauk",
  ["model optim sci technol"] = "Modeling and Optimization in Science and Technologies",
  ["model simul mat sci eng"] = "Modelling and simulation in materials science and engineering",
  ["model simul mater sci eng"] = "Modelling and Simulation in Materials Science and Engineering",
  ["model simul sci eng technol"] = "Modeling and Simulation in Science, Engineering and Technology",
  ["modell meas control c"] = "Modelling Measurement & Control C",
  ["modell simul eng"] = "Modelling and Simulation in Engineering",
  ["modell simul mater sci eng"] = "Modelling and Simulation in Materials Science and Engineering",
  ["modern acoust signal process"] = "Modern Acoustics and Signal Processing",
  ["modern approaches geophys"] = "Modern Approaches in Geophysics",
  ["modern logic"] = "Modern Logic",
  ["modern phys lett a"] = "Modern Physics Letters A. Gravitation, Cosmology, Astrophysics, Nuclear Physics, Particles and Fields, Accelerator Physics, Quantum Information",
  ["modern phys lett b"] = "Modern Physics Letters B. Condensed Matter Physics, Statistical Physics, Atomic, Molecular and Optical Physics",
  ["modmod (baltim md)"] = "Modernism/modernity",
  ["modular math ser"] = "Modular Mathematics Series",
  ["modules life sci"] = "Modules in Life Sciences",
  ["mohu xitong yu shuxue"] = "Mohu Xitong yu Shuxue",
  ["mois afr"] = "Mois en Afrique (Paris, France : 1980)",
  ["moj anat physiol"] = "MOJ anatomy & physiology",
  ["moj immunol"] = "MOJ immunology",
  ["moj lymphology phlebol"] = "MOJ lymphology & phlebology",
  ["moj orthop rheumatol"] = "MOJ orthopedics & rheumatology",
  ["moj proteom bioinform"] = "MOJ proteomics & bioinformatics",
  ["moj sol photoenergy syst"] = "MOJ solar and photoenergy systems",
  ["moj sports med"] = "MOJ sports medicine",
  ["moj toxicol"] = "MOJ toxicology",
  ["mokuzai gakkaishi"] = "Mokuzai gakkai shi",
  ["mol aspects med"] = "Molecular aspects of medicine",
  ["mol astrophys"] = "Molecular Astrophysics",
  ["mol autism"] = "Molecular autism",
  ["mol based math biol"] = "Molecular based mathematical biology",
  ["mol biochem parasitol"] = "Molecular and biochemical parasitology",
  ["mol biol"] = "Molecular biology",
  ["mol biol (los angel)"] = "Molecular biology (Los Angeles, Calif.)",
  ["mol biol (moscow)"] = "Molekulyarnaya Biologiya (Moscow)",
  ["mol biol (moscow, russ fed, engl ed)"] = "Molecular Biology (Moscow, Russian Federation, English Edition) (Translation of Molekulyarnaya Biologiya)",
  ["mol biol (mosk)"] = "Molekuliarnaia biologiia",
  ["mol biol (n y, ny, u s, engl ed)"] = "Molecular Biology (New York, NY, United States, English Edition)",
  ["mol biol biochem biophys"] = "Molecular biology, biochemistry, and biophysics",
  ["mol biol cell"] = "Molecular biology of the cell",
  ["mol biol evol"] = "Molecular biology and evolution",
  ["mol biol int"] = "Molecular biology international",
  ["mol biol med"] = "Molecular biology & medicine",
  ["mol biol rep"] = "Molecular biology reports",
  ["mol biol res commun"] = "Molecular biology research communications",
  ["mol biol today"] = "Molecular biology today : the online journal of molecular biology",
  ["mol biomed"] = "Molecular Biomedicine",
  ["mol biosyst"] = "Molecular Biosystems",
  ["mol biotechnol"] = "Molecular Biotechnology",
  ["mol biother"] = "Molecular biotherapy",
  ["mol brain"] = "Molecular brain",
  ["mol brain res"] = "Molecular Brain Research",
  ["mol breed"] = "Molecular breeding : new strategies in plant improvement",
  ["mol cancer"] = "Molecular cancer",
  ["mol cancer biol"] = "Molecular cancer biology",
  ["mol cancer res"] = "Molecular cancer research : MCR",
  ["mol cancer ther"] = "Molecular cancer therapeutics",
  ["mol carcinog"] = "Molecular carcinogenesis",
  ["mol catal"] = "Molecular Catalysis",
  ["mol cell"] = "Molecular cell",
  ["mol cell biochem"] = "Molecular and Cellular Biochemistry",
  ["mol cell biol"] = "Molecular and cellular biology",
  ["mol cell biol hum dis ser"] = "Molecular and cell biology of human diseases series",
  ["mol cell biol res commun"] = "Molecular cell biology research communications : MCBRC",
  ["mol cell biomech"] = "Molecular & cellular biomechanics : MCB",
  ["mol cell differ"] = "Molecular and cellular differentiation",
  ["mol cell endocrinol"] = "Molecular and Cellular Endocrinology",
  ["mol cell epilepsy"] = "Molecular & cellular epilepsy",
  ["mol cell neurosci"] = "Molecular and cellular neurosciences",
  ["mol cell oncol"] = "Molecular & cellular oncology",
  ["mol cell pediatr"] = "Molecular and cellular pediatrics",
  ["mol cell pharm"] = "Molecular and Cellular Pharmacology",
  ["mol cell pharmacol"] = "Molecular and cellular pharmacology",
  ["mol cell probes"] = "Molecular and Cellular Probes",
  ["mol cell proteomics"] = "Molecular & cellular proteomics : MCP",
  ["mol cell ther"] = "Molecular and Cellular Therapies",
  ["mol cell toxicol"] = "Molecular & Cellular Toxicology",
  ["mol cells"] = "Molecules and cells",
  ["mol chem neuropathol"] = "Molecular and Chemical Neuropathology",
  ["mol clin oncol"] = "Molecular and clinical oncology",
  ["mol cryst liq cryst"] = "Molecular Crystals and Liquid Crystals",
  ["mol cryst liq cryst lett"] = "Molecular Crystals and Liquid Crystals Lett.",
  ["mol cryst liq cryst sci technol, sect a"] = "Molecular Crystals and Liquid Crystals Science and Technology Section A",
  ["mol cryst liq cryst sci technol, sect b"] = "Molecular Crystals and Liquid Crystals Science and Technology Section B",
  ["mol cryst liq cryst sci technol, sect c"] = "Molecular Crystals and Liquid Crystals Science and Technology Section C",
  ["mol cytogenet"] = "Molecular cytogenetics",
  ["mol diagn"] = "Molecular diagnosis : a journal devoted to the understanding of human disease through the clinical application of molecular biology",
  ["mol diagn ther"] = "Molecular diagnosis & therapy",
  ["mol divers"] = "Molecular diversity",
  ["mol diversity"] = "Molecular Diversity",
  ["mol ecol"] = "Molecular ecology",
  ["mol ecol notes"] = "Molecular ecology notes",
  ["mol ecol resour"] = "Molecular ecology resources",
  ["mol endocrinol"] = "Molecular Endocrinology",
  ["mol entomol"] = "Molecular entomology",
  ["mol front j"] = "Molecular Frontiers journal",
  ["mol gen genet"] = "Molecular & general genetics : MGG",
  ["mol gen microbiol virol"] = "Molecular genetics, microbiology and virology : Molekulyarnaya genetika, mikrobiologiya i virusologiya",
  ["mol gen mikrobiol virusol"] = "Molekuliarnaia genetika, mikrobiologiia i virusologiia",
  ["mol genet genomic med"] = "Molecular genetics & genomic medicine",
  ["mol genet genomics"] = "Molecular genetics and genomics : MGG",
  ["mol genet med"] = "Molecular genetic medicine",
  ["mol genet metab"] = "Molecular genetics and metabolism",
  ["mol genet metab rep"] = "Molecular genetics and metabolism reports",
  ["mol genet microbiol virol"] = "Molecular Genetics, Microbiology and Virology",
  ["mol genet, mikrobiol virusol"] = "Molekulyarnaya Genetika, Mikrobiologiya i Virusologiya",
  ["mol hortic"] = "Molecular Horticulture",
  ["mol hum reprod"] = "Molecular human reproduction",
  ["mol imag"] = "Molecular Imaging",
  ["mol imag biol"] = "Molecular Imaging & Biology",
  ["mol imaging"] = "Molecular imaging",
  ["mol imaging biol"] = "Molecular imaging and biology",
  ["mol imaging radionucl ther"] = "Molecular imaging and radionuclide therapy",
  ["mol imaging reconstr anal mov body organs stroke imaging treat (2017)"] = "Molecular imaging, reconstruction and analysis of moving body organs, and stroke imaging and treatment : Fifth International Workshop, CMMI 2017, Second International Workshop, RAMBO 2017, and First International Workshop, SWITCH 2017, held in conjunction with MICCAI 2017, Québec City, QC, Canada, September 14, 2017, proceedings. CMMI (Workshop) (5th : 2017 : Québec, Québec)",
  ["mol immunol"] = "Molecular immunology",
  ["mol imprint"] = "Molecular Imprinting",
  ["mol inf"] = "Molecular Informatics",
  ["mol inform"] = "Molecular informatics",
  ["mol interv"] = "Molecular interventions",
  ["mol interventions"] = "Molecular Interventions",
  ["mol life"] = "Molecular life",
  ["mol mar biol biotech"] = "Molecular Marine Biology and Biotechnology",
  ["mol mar biol biotechnol"] = "Molecular Marine Biology and Biotechnology",
  ["mol med"] = "Molecular medicine (Cambridge, Mass.)",
  ["mol med (manhasset, ny, u s)"] = "Molecular Medicine (Manhasset, NY, United States)",
  ["mol med (sofia)"] = "Molekuliarna Meditsina",
  ["mol med rep"] = "Molecular medicine reports",
  ["mol med ther"] = "Molecular Medicine & Therapeutics",
  ["mol med today"] = "Molecular medicine today",
  ["mol membr biol"] = "Molecular Membrane Biology",
  ["mol metab"] = "Molecular metabolism",
  ["mol microbiol"] = "Molecular Microbiology",
  ["mol model simul appl perspect"] = "Molecular Modeling and Simulation. Applications and Perspectives",
  ["mol neurobiol"] = "Molecular neurobiology",
  ["mol neurodegener"] = "Molecular neurodegeneration",
  ["mol neuropsychiatry"] = "Molecular neuropsychiatry",
  ["mol nutr food res"] = "Molecular nutrition & food research",
  ["mol omics"] = "Molecular omics",
  ["mol oncol"] = "Molecular oncology",
  ["mol online"] = "Molecules Online",
  ["mol oral microbiol"] = "Molecular oral microbiology",
  ["mol pain"] = "Molecular pain",
  ["mol pathol"] = "Molecular pathology : MP",
  ["mol pharm"] = "Molecular pharmaceutics",
  ["mol pharmaceutics"] = "Molecular Pharmaceutics",
  ["mol pharmacol"] = "Molecular Pharmacology",
  ["mol phylogenet evol"] = "Molecular phylogenetics and evolution",
  ["mol phys"] = "Molecular Physics",
  ["mol plant"] = "Molecular plant",
  ["mol plant breed"] = "Molecular plant breeding",
  ["mol plant microbe interact"] = "Molecular plant-microbe interactions : MPMI",
  ["mol plant pathol"] = "Molecular plant pathology",
  ["mol plant-microbe interact"] = "Molecular Plant-Microbe Interactions",
  ["mol psychiatry"] = "Molecular psychiatry",
  ["mol reprod dev"] = "Molecular reproduction and development",
  ["mol screen news"] = "Molecular Screening News",
  ["mol simul"] = "Molecular Simulation",
  ["mol supramol photochem"] = "Molecular and Supramolecular Photochemistry",
  ["mol syndromol"] = "Molecular syndromology",
  ["mol syst biol"] = "Molecular systems biology",
  ["mol syst des eng"] = "Molecular systems design & engineering",
  ["mol ther"] = "Molecular therapy : the journal of the American Society of Gene Therapy",
  ["mol ther methods clin dev"] = "Molecular therapy. Methods & clinical development",
  ["mol ther nucleic acids"] = "Molecular therapy. Nucleic acids",
  ["mol ther oncolytics"] = "Molecular therapy oncolytics",
  ["mol ther–methods clin dev"] = "Molecular Therapy–Methods & Clinical Development",
  ["mol ther–nucleic acids"] = "Molecular Therapy–Nucleic Acids",
  ["mol toxicol"] = "Molecular toxicology",
  ["mol urol"] = "Molecular urology",
  ["mol vis"] = "Molecular vision",
  ["mol vision"] = "Molecular Vision",
  ["molluscan res"] = "Molluscan research",
  ["momenti probl storia pensiero"] = "Momenti e Problemi della Storia del Pensiero",
  ["momento econ"] = "Momento Económico",
  ["mon bull bur health"] = "Monthly bulletin of the Bureau of Health",
  ["mon bull g b emerg public health lab serv"] = "Monthly bulletin. Great Britain. Emergency Public Health Laboratory Service",
  ["mon bull indiana state board health"] = "Monthly bulletin. Indiana State Board of Health",
  ["mon bull minist health emerg public health lab serv"] = "Monthly bulletin of the Ministry of Health and the Emergency Public Health Laboratory Service",
  ["mon bull minist health public health lab serv"] = "Monthly bulletin of the Ministry of Health and the Public Health Laboratory Service",
  ["mon bull stat u n stat off"] = "Monthly bulletin of statistics (United Nations. Statistical Office)",
  ["mon health rep u s army"] = "Monthly health report. United States. Army. Military District of Washington",
  ["mon heath bull"] = "Monthly health bulletin. United States. Army. Army, 2nd",
  ["mon hefte math"] = "Monatshefte für Mathematik",
  ["mon j med (edinb)"] = "Monthly journal of medicine (Edinburgh, Scotland)",
  ["mon j med sci"] = "Monthly journal of medical science",
  ["mon lab rev"] = "Monthly Labor Review",
  ["mon labor rev"] = "Monthly labor review",
  ["mon microsc j"] = "Monthly Microscopical Journal",
  ["mon not r astron soc"] = "Monthly notices of the Royal Astronomical Society",
  ["mon not r astron soc lett"] = "Monthly Notices of the Royal Astronomical Society Letters",
  ["mon rev"] = "Monthly review (New York, N.Y. : 1949)",
  ["mon rev new york state dep labor div ind hyg"] = "Monthly review - New York State Department of Labor, Division of Industrial Hygiene",
  ["mon vital stat rep"] = "Monthly vital statistics report",
  ["mon wea rev"] = "Monthly Weather Review",
  ["mon weather rev"] = "Monthly weather review",
  ["monaldi arch chest dis"] = "Monaldi archives for chest disease = Archivio Monaldi per le malattie del torace",
  ["monash bioeth rev"] = "Monash bioethics review",
  ["monash univ law rev"] = "Monash University law review. Monash University. Faculty of Law",
  ["monastic stud"] = "Monastic studies",
  ["monatsh chem"] = "Monatshefte fur Chemie",
  ["monatsh math"] = "Monatshefte fur Mathematik",
  ["monatsh tierheilkd"] = "Monatshefte für Tierheilkunde",
  ["monatsh veterinarmed"] = "Monatshefte fur Veterinarmedizin",
  ["monatskurse arztl fortbild [bildbeil]"] = "Monatskurse fur die arztliche Fortbildung. Bildbeilage; Farbige Medizin",
  ["monatskurse arztl fortbild bildbeil"] = "Monatskurse für die ärztliche Fortbildung. Bildbeilage; Farbige Medizin",
  ["monatsschr brauwiss"] = "Monatsschrift fur Brauwissenschaft",
  ["monatsschr dtsch zahnarzte freie zahnarzt"] = "Monatsschrift deutscher Zahnärzte: der freie Zahnarzt",
  ["monatsschr geburtshilfe gynakol"] = "Monatsschrift für Geburtshülfe und Gynäkologie",
  ["monatsschr kinderheilkd"] = "Monatsschrift fur Kinderheilkunde",
  ["monatsschr kriminol"] = "Monatsschrift für Kriminologie und Strafrechtsreform",
  ["monatsschr ohrenheilkd laryngorhinol"] = "Monatsschrift fur Ohrenheilkunde und Laryngo-Rhinologie",
  ["monatsschr psychiatr neurol"] = "Monatsschrift für Psychiatrie und Neurologie",
  ["monatsschr tuberkulosebekampf"] = "Monatsschrift für Tuberkulosebekämpfung",
  ["monatsschr unfallheilkd"] = "Monatsschrift fur Unfallheilkunde",
  ["monatsschr unfallheilkd versicher versorg verkehrsmed"] = "Monatsschrift fur Unfallheilkunde, Versicherungs-, Versorgungs- und Verkehrsmedizin",
  ["monatsschr unfallheilkd versicherungsmed"] = "Monatsschrift für Unfallheilkunde und Versicherungsmedizin",
  ["monatsschrift brauwiss"] = "Monatsschrift für brauwissenschaft (Internet)",
  ["mondai to kenkyu"] = "Mondai to kenkyū : Chūgoku jijō senmonshi",
  ["monday dev"] = "Monday developments : a biweekly publication of InterAction",
  ["monde alp rhodan"] = "Le Monde alpin et rhodanien",
  ["monde dent"] = "Le monde dentaire",
  ["monde med"] = "Le Monde médical",
  ["mondes cult"] = "Mondes et cultures : comptes rendus trimestriels des séances de l'Académie des sciences d'outre-mer",
  ["mondes dev"] = "Mondes En Developpement",
  ["mondo aperto"] = "Mondo Aperto",
  ["mondo cin"] = "Mondo cinese",
  ["mondo odontostomatol"] = "Mondo odontostomatologico",
  ["mondo ortod"] = "Mondo ortodontico",
  ["moneda crédito"] = "Moneda y Crédito",
  ["monet econ stud"] = "Monetary and Economic Studies",
  ["moneta credito"] = "Moneta e Credito",
  ["mong j biol sci"] = "Mongolian journal of biological sciences",
  ["mong j demogr"] = "Mongolian journal of demography : journal of Population Teaching and Research Center",
  ["mong math j"] = "Mongolian Mathematical Journal",
  ["monit farm ter"] = "Monitor de la farmacía y de la terapéutica",
  ["monit med farm"] = "El Monitor médico-farmacéutico",
  ["monit ostet ginecol endocrinol metab"] = "Monitore ostetrico-ginecologico di endocrinologia e del metabolismo",
  ["monit ostet ginecolog"] = "Monitore ostetrico-ginecologico",
  ["monit pharm lab"] = "Le Moniteur des pharmacies et des laboratoires",
  ["monitor (assoc clin pharmacol)"] = "Monitor (Associates of Clinical Pharmacology)",
  ["monkhmer stud"] = "Mon-Khmer studies",
  ["monoclon antib immunodiagn immunother"] = "Monoclonal antibodies in immunodiagnosis and immunotherapy",
  ["monoclonal antibodies immunodiagn immunother"] = "Monoclonal Antibodies in Immunodiagnosis and Immunotherapy",
  ["monogr acad ci exact fís-quím nat zaragoza"] = "Monografías de la Academia de Ciencias Exactas, Físicas, Químicas y Naturales de Zaragoza",
  ["monogr allergy"] = "Monographs in allergy",
  ["monogr am assoc ment defic"] = "Monograph of the American Association of Mental Deficiency",
  ["monogr am assoc ment retard"] = "Monographs of the American Association on Mental Retardation",
  ["monogr atheroscler"] = "Monographs on atherosclerosis",
  ["monogr boll"] = "Monografia del Bollettino. Societa italiana di medicina e igiene tropicale. Sezione Eritrea",
  ["monogr cienc tecnol"] = "Monografías Ciencia y Tecnología",
  ["monogr clin cytol"] = "Monographs in clinical cytology",
  ["monogr clin neurosci"] = "Monographs in clinical neuroscience",
  ["monogr comput sci"] = "Monographs in Computer Science",
  ["monogr contemp math"] = "Monographs in Contemporary Mathematics",
  ["monogr corp cent reg poblac"] = "Monografías de la Corporación Centro Regional de la Población",
  ["monogr dev biol"] = "Monographs in developmental biology",
  ["monogr endocrinol"] = "Monographs on endocrinology",
  ["monogr gesamtgeb neurol psychiatr"] = "Monographien aus dem Gesamtgebiete der Neurologie und Psychiatrie",
  ["monogr gesamtgeb neurol psychiatr (berlin)"] = "Monographien aus dem Gesamtgebiete der Neurologie und Psychiatrie",
  ["monogr gesamtgeb psychiatr psychiatry ser"] = "Monographien aus dem Gesamtgebiete der Psychiatrie",
  ["monogr hum genet"] = "Monographs in human genetics",
  ["monogr inequal"] = "Monographs in Inequalities",
  ["monogr linguist math"] = "Monographies de Linguistique Mathematique",
  ["monogr mat"] = "Monografie Matematyczne",
  ["monogr mat garcia galdeano"] = "Monografias Matematicas “Garcia de Galdeano”",
  ["monogr math"] = "Monographs in Mathematics",
  ["monogr math econ"] = "Monographs in Mathematical Economics",
  ["monogr modernen math"] = "Monographien zur Modernen Mathematik",
  ["monogr neoplast dis var sites"] = "Neoplastic disease at various sites",
  ["monogr neural sci"] = "Monographs in neural sciences",
  ["monogr number theory"] = "Monographs in Number Theory",
  ["monogr numer anal"] = "Monographs on Numerical Analysis",
  ["monogr oral sci"] = "Monographs in oral science",
  ["monogr paediatr"] = "Monographs in paediatrics",
  ["monogr pagine micol"] = "Monografie di pagine di micologia",
  ["monogr palaeontogr soc"] = "Monographs of the Palaeontographical Society",
  ["monogr pathol"] = "Monographs in pathology",
  ["monogr phys chem mater"] = "Monographs on the Physics and Chemistry of Materials",
  ["monogr physiol soc"] = "Monographs of the Physiological Society",
  ["monogr popul biol"] = "Monographs in population biology",
  ["monogr quad int stor med sanita"] = "Monografie di quaderni internazionali di storia della medicina e della sanità",
  ["monogr real acad ci exact fis-quim nat zaragoza"] = "Monografias de la Real Academia de Ciencias Exactas, Fisicas, Quimicas y Naturales de Zaragoza",
  ["monogr res notes math"] = "Monographs and Research Notes in Mathematics",
  ["monogr ser"] = "American Indian and Alaska Native Mental Health Research. MonographSeries Am Indian Alsk Native Ment Health Re",
  ["monogr ser parana’s math soc"] = "Monograph Series of the Parana’s Mathematical Society",
  ["monogr ser world health organ"] = "Monograph series. World Health Organization",
  ["monogr soc res child dev"] = "Monographs of the Society for Research in Child Development",
  ["monogr sov med sci"] = "Monographs in Soviet medical sciences",
  ["monogr statist appl probab"] = "Monographs on Statistics and Applied Probability",
  ["monogr surg sci"] = "Monographs in the surgical sciences",
  ["monogr textbooks pure appl math"] = "Monographs and Textbooks in Pure and Applied Mathematics",
  ["monogr theoret comput sci eatcs ser"] = "Monographs in Theoretical Computer Science. An EATCS Series",
  ["monogr virol"] = "Monographs in virology",
  ["monogr west n am nat"] = "Monographs of the Western North American naturalist",
  ["monografii ngtu"] = "Monografii. Novosibirskii Gosudartstvennyi Tekhnicheskii Universitet",
  ["monographs surveys water res engrg"] = "Monographs and Surveys in Water Resource Engineering",
  ["monschr württ forstwes"] = "Monatsschrift für das württembergische Forstwesen",
  ["monspel hippocrates"] = "Monspeliensis Hippocrates",
  ["mont code annot mont"] = "Montana code annotated. Montana",
  ["mont law rev"] = "Montana law review",
  ["mont wildl"] = "Montana wildlife (Bozeman, Mont.)",
  ["monte carlo methods appl"] = "Monte Carlo Methods and Applications",
  ["monte verità"] = "Monte Verità",
  ["monterrey med"] = "Monterrey médico",
  ["monti boschi"] = "Monti e boschi",
  ["montp med"] = "Montpellier médical",
  ["montreal med"] = "Montréal médical",
  ["montroll memorial lecture ser math phys"] = "Montroll Memorial Lecture Series in Mathematical Physics",
  ["monum hist"] = "Monuments historiques : Mh",
  ["monum nihon"] = "Monumenta Nipponica",
  ["monum nippon"] = "Monumenta Nipponica",
  ["moral philos politics"] = "Moral philosophy and politics",
  ["morb mortal"] = "Morbidity and mortality",
  ["morb mortal wkly rep surveill summ"] = "Morbidity and mortality weekly report. Surveillance summaries : MMWR / Centers for Disease Control",
  ["morgan kaufmann ser comput graph geom model"] = "The Morgan Kaufmann Series in Computer Graphics and Geometric Modeling",
  ["morgan kaufmann ser data management systems"] = "The Morgan Kaufmann Series in Data Management Systems",
  ["morningside lect math"] = "Morningside Lectures in Mathematics",
  ["moroc j algebra geom appl"] = "Moroccan Journal of Algebra and Geometry with Applications",
  ["morphol embryol (bucur)"] = "Morphologie et embryologie",
  ["morphol igazsagugyi orv sz"] = "Morphologiai és igazságügyi orvosi szemle",
  ["morphol med"] = "Morphologia medica",
  ["morphology (dordr)"] = "Morphology (Dordrecht, Netherlands)",
  ["morrocan j chem"] = "Moroccan Journal of Chemistry",
  ["morrocan j pure appl anal"] = "Moroccan Journal of Pure and Applied Analysis",
  ["mortality (abingdon)"] = "Mortality (Abingdon, England)",
  ["mos-siam ser optim"] = "MOS-SIAM Series on Optimization",
  ["mosc j comb number theory"] = "Moscow Journal of Combinatorics and Number Theory",
  ["mosc lect"] = "Moscow Lectures",
  ["mosc math j"] = "Moscow Mathematical Journal",
  ["moscow j comb number theory"] = "Moscow Journal of Combinatorics and Number Theory",
  ["moscow math j"] = "Moscow Mathematical Journal",
  ["moscow univ biol sci bull"] = "Moscow University biological sciences bulletin",
  ["moscow univ chem bull"] = "Moscow University Chemistry Bulletin",
  ["moscow univ comput math cybern"] = "Moscow University Computational Mathematics and Cybernetics",
  ["moscow univ comput math cybernet"] = "Moscow University Computational Mathematics and Cybernetics",
  ["moscow univ geol bull"] = "Moscow University Geology Bulletin",
  ["moscow univ math bull"] = "Moscow University Mathematics Bulletin",
  ["moscow univ mech bull"] = "Moscow University Mechanics Bulletin",
  ["moscow univ phys bull"] = "Moscow University Physics Bulletin",
  ["moscow univ soil sci bull"] = "Moscow University Soil Science Bulletin",
  ["mosq news"] = "Mosquito news",
  ["most newsl"] = "MOST newsletter",
  ["mot z"] = "Motortechnische Zeitschrift",
  ["mother child"] = "Mother and child",
  ["mother jones"] = "Mother Jones",
  ["mothers child"] = "Mothers and children",
  ["motiv emot"] = "Motivation and emotion",
  ["motiv interviewing"] = "Motivational interviewing : training, research, implementation, practice",
  ["motiv sci"] = "Motivation science",
  ["motor control"] = "Motor control",
  ["mouse genome"] = "Mouse genome",
  ["mouth mirror"] = "The Mouth mirror",
  ["mouv soc"] = "Le Mouvement social",
  ["mov disord"] = "Movement disorders : official journal of the Movement Disorder Society",
  ["mov disord clin pract"] = "Movement disorders clinical practice",
  ["mov ecol"] = "Movement ecology",
  ["mov operaio soc"] = "Movimento operaio e socialista",
  ["mov sport sci"] = "Movement & Sport Science",
  ["mov worlds"] = "Moving worlds",
  ["mp med prakt pediatr"] = "MP Medycyna praktyczna. Pediatria",
  ["mpt metall plant technol"] = "MPT Metallurgical Plant and Technology",
  ["mpt metall plant technol int"] = "MPT Metallurgical Plant and Technology International",
  ["mrc news"] = "MRC news",
  ["mrs adv"] = "MRS Advances",
  ["mrs bull"] = "MRS Bulletin",
  ["mrs commun"] = "MRS Communications",
  ["mrs energy & sustainability"] = "MRS Energy & Sustainability",
  ["mrs internet j nitride semicond res"] = "MRS Internet Journal of Nitride Semiconductor Research",
  ["mrs online proc lib"] = "MRS Online Proceedings Library",
  ["mrs online proc libr"] = "MRS Online Proceedings Library",
  ["mrs proc"] = "MRS Proceedings",
  ["ms&a model simul appl"] = "MS&A. Modeling, Simulation and Applications",
  ["msda j"] = "MSDA journal : journal of the Maryland State Dental Association",
  ["msj mem"] = "MSJ Memoirs",
  ["msn mon sci news"] = "MSN; monthly science news",
  ["msri math circ libr"] = "MSRI Mathematical Circles Library",
  ["mt geol"] = "Mountain Geologist",
  ["mt res dev"] = "Mountain research and development",
  ["mt sinai j med"] = "The Mount Sinai journal of medicine, New York",
  ["mucosal immunol"] = "Mucosal immunology",
  ["mull abfall"] = "Mull und Abfall",
  ["mult criteria decis mak"] = "Multiple Criteria Decision Making",
  ["mult scler"] = "Multiple sclerosis (Houndmills, Basingstoke, England)",
  ["mult scler demyelinating disord"] = "Multiple sclerosis and demyelinating disorders",
  ["mult scler int"] = "Multiple sclerosis international",
  ["mult scler j exp transl clin"] = "Multiple sclerosis journal - experimental, translational and clinical",
  ["mult scler relat disord"] = "Multiple sclerosis and related disorders",
  ["multiagent grid syst"] = "Multiagent and Grid Systems",
  ["multibody sysdyn"] = "Multibody System Dynamics",
  ["multibody syst dyn"] = "Multibody system dynamics",
  ["multidimens syst signal process"] = "Multidimensional systems and signal processing",
  ["multidimens systems signal process"] = "Multidimensional Systems and Signal Processing",
  ["multidimension syst signal process"] = "Multidimensional Systems and Signal Processing",
  ["multidiscip appl opt"] = "Multidisciplinary and Applied Optics",
  ["multidiscip j educ research"] = "Multidisciplinary journal of educational research",
  ["multidiscip model mater struct"] = "Multidiscipline Modeling in Materials and Structures",
  ["multidiscip respir med"] = "Multidisciplinary respiratory medicine",
  ["multifunct mater"] = "Multifunctional Materials",
  ["multimed man cardiothorac surg"] = "Multimedia manual of cardiothoracic surgery : MMCTS",
  ["multimed syst"] = "Multimedia systems",
  ["multimed tools appl"] = "Multimedia tools and applications",
  ["multimedia syst"] = "Multimedia Systems",
  ["multimedia tools appl"] = "Multimedia Tools and Applications",
  ["multimodal brain image anal (2011)"] = "Multimodal brain image analysis : first international workshop, MBIA 2011, held in conjunction with MICCAI 2011, Toronto, Canada, September 18, 2011 : proceedings",
  ["multimodal brain image anal (2012)"] = "Multimodal brain image analysis : second International Workshop, MBIA 2012, held in conjunction with MICCAI 2012, Nice, France, October 1-5, 2012 : proceedings. MBIA (Workshop) (2nd : 2012 : Nice, France)",
  ["multimodal brain image anal (2013)"] = "Multimodal brain image analysis : third International Workshop, MBIA 2013, held in conjunction with MICCAI 2013, Nagoya, Japan, September 22, 2013 : proceedings. MBIA (Workshop) (3rd : 2013 : Nagoya-shi, Japan)",
  ["multimodal brain image anal math found comput anat (2019)"] = "Multimodal Brain Image Analysis and Mathematical Foundations of Computational Anatomy : 4th International Workshop, MBIA 2019, and 7th International Workshop, MFCA 2019, held in conjunction with MICCAI 2019, Shenzhen, China, October 17, 2019, Proceedings. MBIA (Workshop) (4th : 2019 : Shenzhen Shi, China)",
  ["multimodal learn clin decis support clin image based proc (2020)"] = "Multimodal learning for clinical decision support and clinical image-based procedures : 10th International Workshop, ML-CDS 2020, and 9th International Workshop, CLIP 2020, held in conjunction with MICCAI 2020, Lima, Peru, October 4-8, 2020, Proceedings. ML-CDS (Workshop) (10th : 2020 : Online)",
  ["multinat finance j"] = "Multinational Finance Journal",
  ["multiphase sci technol"] = "Multiphase Science and Technology",
  ["multiphys model"] = "Multiphysics Modeling",
  ["multiscale model simul"] = "Multiscale Modeling and Simulation",
  ["multiscale multidiscip model exp des"] = "Multiscale and Multidisciplinary Modeling, Experiments and Design",
  ["multiscale sci eng"] = "Multiscale Science and Engineering",
  ["multisens res"] = "Multisensory research",
  ["multivar behav res"] = "Multivariate Behavioral Research",
  ["multivariate behav res"] = "Multivariate behavioral research",
  ["munca sanit"] = "Munca sanitară",
  ["munch med wochenschr"] = "Münchener medizinische Wochenschrift (1950)",
  ["mund kiefer gesichtschir"] = "Mund-, Kiefer- und Gesichtschirurgie : MKG",
  ["mundo med"] = "Mundo médico",
  ["mundo saude (1995)"] = "Mundo da saúde (1995)",
  ["munger afr libr notes"] = "Munger Africana Library notes",
  ["munst lect math"] = "Munster Lectures in Mathematics",
  ["munster j math"] = "Munster Journal of Mathematics",
  ["munstersche beitr gesch theor med"] = "Münstersche Beiträge zur Geschichte und Theorie der Medizin",
  ["mus geneve"] = "Musées de Genève",
  ["mus helv"] = "Museum helveticum",
  ["mus hist j"] = "Museum history journal",
  ["mus north ariz bull"] = "Museum of Northern Arizona bulletin",
  ["mus soc issues"] = "Museums & social issues : a journal of reflective discourse",
  ["muscle biol"] = "Muscle biology",
  ["muscle nerve"] = "Muscle & nerve",
  ["muscle nerve suppl"] = "Muscle & nerve. Supplement",
  ["muscles ligaments tendons j"] = "Muscles, ligaments and tendons journal",
  ["musculoskelet regen"] = "Musculoskeletal regeneration",
  ["musculoskelet sci pract"] = "Musculoskeletal science & practice",
  ["musculoskelet surg"] = "Musculoskeletal surgery",
  ["musculoskeletal care"] = "Musculoskeletal care",
  ["mushroom res"] = "Mushroom research",
  ["music lett"] = "Music & letters",
  ["music med"] = "Music and medicine",
  ["music percept"] = "Music perception",
  ["music sci"] = "Musicae scientiae : the journal of the European Society for the Cognitive Sciences of Music",
  ["music sci (lond)"] = "Music & science",
  ["music ther perspect"] = "Music therapy perspectives",
  ["muslim world"] = "The Muslim world",
  ["muslim world (hartford)"] = "Muslim world (Hartford, Conn.)",
  ["musmat braz j music math"] = "MusMat. Brazilian Journal of Music and Mathematics",
  ["musmat brazil j music math"] = "MusMat. Brazilian Journal of Music and Mathematics",
  ["mutat res"] = "Mutation research",
  ["mutat res  dnaging: genet instab aging"] = "Mutation Research, DNAging: Genetic Instability and Aging",
  ["mutat res  mutat res genomics"] = "Mutation Research, Mutation Research Genomics",
  ["mutat res dna repair"] = "Mutation Research, DNA Repair",
  ["mutat res dna repair rep"] = "Mutation Research, DNA Repair Reports",
  ["mutat res dnaging: genet instab aging"] = "Mutation Research, DNAging: Genetic Instability and Aging",
  ["mutat res environ mutagen relat subj"] = "Mutation Research, Environmental Mutagenesis and Related Subjects",
  ["mutat res fundam mol mech mutagen"] = "Mutation Research, Fundamental and Molecular Mechanisms of Mutagenesis",
  ["mutat res genet toxicol"] = "Mutation Research, Genetic Toxicology",
  ["mutat res genet toxicol environ mutagen"] = "Mutation research. Genetic toxicology and environmental mutagenesis",
  ["mutat res genet toxicol test"] = "Mutation Research, Genetic Toxicology Testing",
  ["mutat res lett"] = "Mutation Research Letters",
  ["mutat res mutat res genomics"] = "Mutation Research, Mutation Research Genomics",
  ["mutat res rev genet toxicol"] = "Mutation Research, Reviews in Genetic Toxicology",
  ["mutat res rev mutat res"] = "Mutation research. Reviews in mutation research",
  ["muthanna j pure sci"] = "Muthanna journal of pure science",
  ["mvp mod vet pract"] = "MVP. Modern veterinary practice",
  ["mycobact dis"] = "Mycobacterial diseases : tuberculosis & leprosy",
  ["mycol balc"] = "Mycologia balcanica",
  ["mycol bavarica"] = "Mycologia Bavarica = Bavarian journal of mycology",
  ["mycol helv"] = "Mycologia Helvetica",
  ["mycol iran"] = "Mycologia Iranica",
  ["mycol monten"] = "Mycologia Montenegrina",
  ["mycol prog"] = "Mycological progress",
  ["mycol res"] = "Mycological research",
  ["mycopathol mycol appl"] = "Mycopathologia et mycologia applicata",
  ["mycosystema / zhongguo ke xue yuan wei sheng wu yan jiu suo zhongguo jun wu xue hui zhu ban|jun wu xi tong"] = "Jun wu xi tong",
  ["mycotoxin res"] = "Mycotoxin research",
  ["mykosen suppl"] = "Mykosen. Supplement",
  ["mymensingh med j"] = "Mymensingh medical journal : MMJ",
  ["myrmecol nachr"] = "Myrmecologische nachrichten",
  ["myrmecol news"] = "Myrmecological news",
  ["mysore j agric sci"] = "Mysore journal of agricultural sciences",
  ["mém soc math fr (ns)"] = "Mémoires de la Société Mathématique de France",
  ["mém soc math france (ns)"] = "Mémoires de la Société Mathématique de France",
  ["mém soc vaud sci nat"] = "Mémoires de la Société vaudoise des Sciences Naturelles",
  ["méthodes math ingr"] = "Méthodes Mathématiques pour l’Ingénieur",
  ["méthodes prat ingén"] = "Méthodes et Pratiques de l’Ingénieur",
  ["n am actuar j"] = "North American actuarial journal : NAAJ",
  ["n am actuarial j"] = "North American Actuarial Journal",
  ["n am archaeol"] = "North American Archaeologist",
  ["n am fungi"] = "North American fungi",
  ["n am j aquac"] = "North American journal of aquaculture",
  ["n am j fish manag"] = "North American journal of fisheries management",
  ["n am j med sci"] = "North American journal of medical sciences",
  ["n am j med sci (boston)"] = "North American journal of medicine & science",
  ["n am j psychol"] = "North American journal of psychology",
  ["n am j sports phys ther"] = "North American journal of sports physical therapy : NAJSPT",
  ["n amer actuarial j"] = "North American Actuarial Journal",
  ["n amer j econ finance"] = "North American Journal of Economics and Finance",
  ["n biotechnol"] = "New biotechnology",
  ["n c dent gaz"] = "North Carolina dental gazette : a publication of North Carolina Dental Society",
  ["n c dent j"] = "North Carolina dental journal",
  ["n c folklor"] = "North Carolina folklore",
  ["n c hist rev"] = "The North Carolina historical review",
  ["n c j law technol"] = "North Carolina journal of law & technology",
  ["n c j ment health"] = "North Carolina journal of mental health",
  ["n c med j"] = "North Carolina medical journal",
  ["n d hist"] = "North Dakota history",
  ["n d law rev"] = "North Dakota law review",
  ["n engl dent j"] = "New England dental journal",
  ["n engl galaxy"] = "The New-England galaxy",
  ["n engl hist geneal regist"] = "The New England historical and genealogical register",
  ["n engl j crim civ confin"] = "New England journal on criminal and civil confinement",
  ["n engl j hum serv"] = "New England journal of human services",
  ["n engl j med"] = "The New England journal of medicine",
  ["n engl j med overseas ed"] = "New England journal of medicine (Overseas ed.)",
  ["n engl j public policy"] = "New England journal of public policy : a journal of the John W. McCormack Institute of Public Affairs, University of Massachusetts at Boston",
  ["n engl reg allergy proc"] = "New England and regional allergy proceedings",
  ["n h health news"] = "New Hampshire health news",
  ["n h revis statut annot n h"] = "New Hampshire revised statutes annotated, 1955. New Hampshire",
  ["n hc perspect community"] = "N & HC perspectives on community : official publication of the National League for Nursing",
  ["n j healthc"] = "New Jersey healthCare",
  ["n j hist"] = "New Jersey history",
  ["n j hist comm newsl"] = "New Jersey Historical Commission newsletter. New Jersey Historical Commission",
  ["n j j pharm"] = "New Jersey journal of pharmacy",
  ["n j league nurs news"] = "New Jersey League for Nursing news",
  ["n j med"] = "New Jersey medicine : the journal of the Medical Society of New Jersey",
  ["n j nurse"] = "New Jersey nurse",
  ["n j statut annot n j"] = "New Jersey statutes annotated : Permanent ed. Laws of a general and permanent nature under arrangement of Revised statutes, 1937, including laws of the 163rd Legislature, 1939; with annotations from cases construing or applying the laws, tables, and index. New Jersey",
  ["n m dent j"] = "New Mexico dental journal",
  ["n m health off"] = "New Mexico health officer",
  ["n m hist rev"] = "New Mexico historical review",
  ["n m law rev"] = "New Mexico law review",
  ["n m nurse"] = "New Mexico nurse",
  ["n m statut 1978 annot n m"] = "New Mexico statutes, 1978 annotated. New Mexico",
  ["n s hist rev"] = "Nova Scotia historical review",
  ["n s med bull"] = "The Nova Scotia medical bulletin",
  ["n s w public health bull"] = "New South Wales public health bulletin",
  ["n y aff"] = "New York affairs",
  ["n y folk"] = "New York folklore",
  ["n y hist"] = "New York history. New York State Historical Association",
  ["n y j dent"] = "The New York journal of dentistry",
  ["n y law forum"] = "New York law forum",
  ["n y law j"] = "New York law journal",
  ["n y law sch j hum rights"] = "New York Law School journal of human rights",
  ["n y law sch j int comp law"] = "New York Law School journal of international and comparative law",
  ["n y law school hum rights annu"] = "New York Law School human rights annual",
  ["n y med coll news notes"] = "New York Medical College news & notes",
  ["n y med q"] = "The New York medical quarterly",
  ["n y physician"] = "The New York physician",
  ["n y state bar j"] = "New York state bar journal",
  ["n y state dent j"] = "The New York state dental journal",
  ["n y state j med"] = "New York state journal of medicine",
  ["n y state nurse"] = "New York state nurse",
  ["n y state pharm"] = "New York state pharmacist",
  ["n y state psychol"] = "New York State psychologist",
  ["n y suppl second ser"] = "New York supplement. Second series",
  ["n y times"] = "New York Times",
  ["n y times mag"] = "The New York times magazine",
  ["n y times web"] = "The New York times on the Web",
  ["n y univ j dent"] = "New York University journal of dentistry",
  ["n y univ j int law polit"] = "New York University journal of international law & politics. New York University. International Law Society",
  ["n y univ j legis public policy"] = "New York University journal of legislation and public policy",
  ["n y univ law rev"] = "New York University law review (1950)",
  ["n y univ med q"] = "New York University medical quarterly",
  ["n z bioeth j"] = "New Zealand bioethics journal",
  ["n z cartogr geogr inf sys"] = "New Zealand cartography and geographic information systems : the journal of the New Zealand Cartographic Society",
  ["n z dent j"] = "The New Zealand dental journal",
  ["n z entomol"] = "The New Zealand entomologist",
  ["n z fam physician"] = "The New Zealand family physician",
  ["n z geog"] = "New Zealand geographer",
  ["n z geogr"] = "New Zealand Geographer",
  ["n z health hospital"] = "New Zealand health & hospital",
  ["n z hosp"] = "New Zealand hospital",
  ["n z j  mar freshwater res"] = "New Zealand Journal of Marine and Freshwater Research",
  ["n z j agric res"] = "New Zealand Journal of Agricultural Research",
  ["n z j bot"] = "New Zealand journal of botany",
  ["n z j botan"] = "New Zealand Journal of Botany",
  ["n z j crop hortic sci"] = "New Zealand journal of crop and horticultural science",
  ["n z j ecol"] = "New Zealand journal of ecology",
  ["n z j fam plann"] = "New Zealand journal of family planning",
  ["n z j for sci"] = "New Zealand journal of forestry science",
  ["n z j geogr"] = "New Zealand journal of geography",
  ["n z j geol geophys"] = "New Zealand Journal of Geology and Geophysics",
  ["n z j hist"] = "The New Zealand journal of history",
  ["n z j mar freshwater res"] = "New Zealand journal of marine and freshwater research",
  ["n z j math"] = "New Zealand Journal of Mathematics",
  ["n z j med lab sci"] = "New Zealand Journal of Medical Laboratory Science",
  ["n z j sci technol sect b"] = "The New Zealand journal of science and technology. B. General section",
  ["n z j zool"] = "New Zealand journal of zoology",
  ["n z law j"] = "The New Zealand law journal : NZLJ",
  ["n z law rep"] = "The New Zealand law reports",
  ["n z libr"] = "New Zealand libraries",
  ["n z med j"] = "The New Zealand medical journal",
  ["n z nurs forum"] = "New Zealand nursing forum",
  ["n z nurs j"] = "The New Zealand nursing journal. Kai tiaki",
  ["n z plant prot"] = "New Zealand Plant Protection. New Zealand plant protection society",
  ["n z popul newsl"] = "New Zealand population newsletter",
  ["n z popul rev"] = "New Zealand population review",
  ["n z sch dent serv gaz"] = "New Zealand School Dental Service gazette",
  ["n z slav j"] = "New Zealand Slavonic journal",
  ["n z vet j"] = "New Zealand veterinary journal",
  ["naacog newsl"] = "NAACOG newsletter",
  ["naacog tech bull"] = "NAACOG technical bulletin",
  ["naacogs clin issu perinat womens health nurs"] = "NAACOG's clinical issues in perinatal and women's health nursing",
  ["naber rep"] = "NABER report",
  ["nacdl j"] = "NACDL journal",
  ["nachr akad wiss gott philol hist kl"] = "Nachrichten von der Akademie der Wissenschaften in Göttingen. Philologisch-Historische Klasse",
  ["nachr akad wiss gottingen ii math phys kl"] = "Nachrichten der Akademie der Wissenschaften in Gottingen II: Mathematisch-Physikalische Klasse",
  ["nachr akad wiss gottingen ii: math phys kl 1983"] = "Nachrichten der Akademie der Wissenschaften in Gottingen II: Mathematisch-Physikalische Klasse 1983",
  ["nachr akad wiss göttingen math-phys kl ii"] = "Nachrichten der Akademie der Wissenschaften in Göttingen",
  ["nachr bayer entomol"] = "Nachrichtenblatt der Bayerischen Entomologen",
  ["nachr chem"] = "Nachrichten aus der Chemie",
  ["nachr chem tech lab"] = "Nachrichten aus Chemie Technik und Laboratorium",
  ["nachr dtsch pflanzenschutzd"] = "Nachrichtenblatt des deutschen Pflanzenschutzdienstes",
  ["nachr entomol ver apollo"] = "Nachrichten des Entomologischen Vereins Apollo",
  ["nachr inst gesch naturwiss hamburg"] = "Nachrichten aus dem Institut für Geschichte der Naturwissenschaften Hamburg",
  ["nachrbl dtsch pflanzenschutzd"] = "Nachrichtenblatt des Deutschen Pflanzenschutzdienstes",
  ["nachrichtenbl dtsch ges gesch med naturwiss tech"] = "Nachrichtenblatt. Deutsche Gesellschaft für Geschichte der Medizin, der Naturwissenschaften und der Technik",
  ["nachrichtenbl dtsch pflanzenschutzdienst"] = "Nachtrichtenblatt des Deutschen Pflanzenschutzdienstes",
  ["nacta j"] = "NACTA Journal",
  ["nacubo bus off"] = "NACUBO business officer : magazine of the National Association of College and University Business Officers. National Association of College and University Business Officers",
  ["nadc ma united states nav air dev cen johnsville pa aviat med accel lab"] = "NADC-MA-. United States. Naval Air Development Center, Johnsville, Pa. Aviation Medical Acceleration Laboratory",
  ["nadc-ml rep"] = "NADC-ML [reports]. United States. Naval Air Development Center, Johnsville, Pa. Aviation Medical Acceleration Laboratory",
  ["nadc-mr rep"] = "NADC-MR [reports]. United States. Naval Air Development Center, Johnsville, Pa. Aerospace Medical Research Department",
  ["nadl j"] = "NADL journal",
  ["nagasaki igakkai zasshi"] = "Nagasaki Igakkai zasshi = Nagasaki medical journal",
  ["nagoya gakuin daigaku ronshu igaku kenko kagaku supotsu kagakuhen"] = "Nagoya Gakuin Daigaku ronshu. Igaku, kenko kagaku, supotsu kagakuhen",
  ["nagoya j med sci"] = "Nagoya journal of medical science",
  ["nagoya math j"] = "Nagoya Mathematical Journal",
  ["nagoya med j"] = "Nagoya medical journal",
  ["naham access manag j"] = "NAHAM access management journal",
  ["naham manage j"] = "The NAHAM management journal",
  ["naika hokan"] = "Naika hokan. Japanese archives of internal medicine",
  ["nakadori byoin iho"] = "Nakadori Byoin iho. Nakadori Hospital medical bulletin]",
  ["namahyi anjumani hasharahshinasani iran"] = "Nāmah-yi anjuman-i ḥasharahshināsān-i Īrān",
  ["nan fang yi ke da xue xue bao"] = "Nan fang yi ke da xue xue bao = Journal of Southern Medical University",
  ["nan kai da xue xue bao zi ran ke xue ban"] = "Nan kai da xue xue bao. Zi ran ke xue ban = Acta scientiarum naturalium Universitatis Nankaiensis",
  ["nan nu"] = "Nan nü : men, women, and gender in early and Imperial China",
  ["nan xing xue za zhi"] = "Nan xing xue za zhi = Chinese journal of andrology",
  ["nan yang wen ti"] = "Nan yang wen ti",
  ["nanching ta hsueh hsueh pao"] = "Nanjing da xue xue bao. Zhe xue, ren wen ke xue, she hui ke xue = Journal of Nanjing University",
  ["nanfang nongye xuebao"] = "Nanfang nongye xuebao = Journal of Southern Agriculture",
  ["nanfang shuichan kexue"] = "Nanfang shuichan kexue",
  ["nanhua daxue xuebao ziran kexue ban"] = "Nanhua daxue xuebao. Ziran kexue ban = Journal of Nanhua University Science and Technology",
  ["nanjing daxue xuebao shuxue bannian kan"] = "Nanjing Daxue Xuebao. Shuxue Bannian Kan. Nanjing University. Journal. Mathematical Biquarterly",
  ["nanjing daxue xuebao ziran kexue ban"] = "Nanjing Daxue Xuebao",
  ["nanjing nong ye da xue xue bao"] = "Nanjing nong ye da xue xue bao = Journal of Nanjing Agricultural University",
  ["nanjing shi da xue bao"] = "Nanjing shi da xue bao. Zi ran ke xue ban = Nanjing shida xuebao",
  ["nanjing shida xuebao|nanjing shi da xue bao"] = "Nanjing shi da xue bao. Zi ran ke xue ban",
  ["nankai ser pure appl math theoret phys"] = "Nankai Series in Pure, Applied Mathematics and Theoretical Physics",
  ["nankai tracts math"] = "Nankai Tracts in Mathematics",
  ["nano biomed eng"] = "Nano biomedicine and engineering",
  ["nano commun netw"] = "Nano communication networks",
  ["nano commun networks"] = "Nano Communication Networks",
  ["nano converg"] = "Nano convergence",
  ["nano convergence"] = "Nano Convergence",
  ["nano energy"] = "Nano Energy",
  ["nano express"] = "Nano Express",
  ["nano futur"] = "Nano Futures",
  ["nano futures"] = "Nano Futures",
  ["nano lett"] = "Nano Letters",
  ["nano life"] = "Nano LIFE",
  ["nano mater sci"] = "Nano Materials Science",
  ["nano micro lett"] = "Nano Micro Letters",
  ["nano micro mech rev"] = "Nano Micro Mechanics Review",
  ["nano res"] = "Nano Research",
  ["nano rev"] = "Nano reviews",
  ["nano rev exp"] = "Nano reviews & experiments",
  ["nano sci nano technol"] = "Nano Science and Nano Technology",
  ["nano sci nano technol: indian j"] = "Nano Science and Nano Technology: An Indian Journal",
  ["nano select"] = "Nano Select",
  ["nano today"] = "Nano Today",
  ["nano-micro lett"] = "Nano-Micro Letters",
  ["nano-struct nano-objects"] = "Nano-Structures & Nano-Objects",
  ["nanobiomedicine (rij)"] = "Nanobiomedicine",
  ["nanobiotechnol rep"] = "Nanobiotechnology Reports",
  ["nanomanuf metrol"] = "Nanomanufacturing And Metrology",
  ["nanomater energy"] = "Nanomaterials and Energy",
  ["nanomater environ"] = "Nanomaterials and the environment",
  ["nanomater nanotechnol"] = "Nanomaterials and Nanotechnology",
  ["nanomaterials (basel)"] = "Nanomaterials (Basel, Switzerland)",
  ["nanomech sci technol: int j"] = "Nanomechanics Science and Technology: An International Journal",
  ["nanomed j"] = "Nanomedicine journal",
  ["nanomed nanosci res"] = "Nanomedicine & nanoscience research",
  ["nanomed nanotechnol biol med"] = "Nanomedicine: Nanotechnology, Biology and Medicine",
  ["nanomed res j"] = "Nanomedicine Research Journal",
  ["nanomedicine (chichester)"] = "Nanomedicine : design and applications of magnetic nanomaterials, nanosensors, and nanosystems",
  ["nanomedicine (lond)"] = "Nanomedicine (London, England)",
  ["nanomedicine (london, u k)"] = "Nanomedicine (London, United Kingdom)",
  ["nanomedicine (n y, ny, u s)"] = "Nanomedicine (New York, NY, United States)",
  ["nanomicro lett"] = "Nano-micro letters",
  ["nanoscale adv"] = "Nanoscale Advances",
  ["nanoscale horiz"] = "Nanoscale horizons",
  ["nanoscale microscale thermophys eng"] = "Nanoscale and Microscale Thermophysical Engineering",
  ["nanoscale res lett"] = "Nanoscale research letters",
  ["nanoscale syst math model theory appl"] = "Nanoscale Systems. Mathematical Modeling, Theory and Applications",
  ["nanoscale syst: math model theory appl"] = "Nanoscale Systems: Mathematical Modeling, Theory and Applications",
  ["nanosci nanoeng"] = "Nanoscience and nanoengineering",
  ["nanosci nanotechnol - asia"] = "Nanoscience and Nanotechnology - Asia",
  ["nanosci nanotechnol lett"] = "Nanoscience and Nanotechnolog Letters",
  ["nanosci technol"] = "NanoScience and Technology",
  ["nanosci technol: int j"] = "Nanoscience and Technology: An International Journal",
  ["nanosist nanomater nanotehnol"] = "Nanosistemi Nanomateriali Nanotehnologii",
  ["nanostruct mater"] = "Nanostructured Materials",
  ["nanostruct sci and tech"] = "Nanostructure Science and Technology",
  ["nanotechnol construct"] = "Nanotechnologies In Construction",
  ["nanotechnol environ eng"] = "Nanotechnology For Environmental Engineering",
  ["nanotechnol percept"] = "Nanotechnology perceptions",
  ["nanotechnol precis eng"] = "Nanotechnology and Precision Engineering",
  ["nanotechnol rev"] = "Nanotechnology reviews",
  ["nanotechnol russ"] = "Nanotechnologies in Russia",
  ["nanotechnol sci appl"] = "Nanotechnology, science and applications",
  ["nanoworld j"] = "NanoWorld journal",
  ["nanr news"] = "Nanr News",
  ["nanyang wenti yenchiu"] = "Nan yang wen ti yan qiu = Southeast Asian affairs",
  ["napa bull"] = "NAPA bulletin",
  ["napht news"] = "NAPHT news",
  ["napravna pedagog"] = "Nápravná pedagogika",
  ["napt bull"] = "NAPT bulletin",
  ["nar azii afr"] = "Narody Azii i Afriki",
  ["nar cancer"] = "NAR cancer",
  ["nar genom bioinform"] = "NAR genomics and bioinformatics",
  ["nar genomics bioinf"] = "NAR Genomics and Bioinformatics",
  ["nar zdrav"] = "Narodno zdravlje",
  ["naresa mongr"] = "NARESA monograph",
  ["narrat inq"] = "Narrative inquiry : NI",
  ["narrat inq bioeth"] = "Narrative inquiry in bioethics",
  ["nasa contract rep nasa cr"] = "NASA contractor report. NASA CR. United States. National Aeronautics and Space Administration",
  ["nasa monogr syst softw eng"] = "NASA Monographs in Systems and Software Engineering",
  ["nasa tech memo"] = "NASA technical memorandum",
  ["nasn sch nurse"] = "NASN school nurse (Print)",
  ["naspcp newsl"] = "NASPCP newsletter. National AIDS/STD Prevention and Control Program (Philippines)",
  ["nassau med news"] = "Nassau medical news",
  ["nasu suteshon"] = "Nāsu sutēshon : NS = Nurses' station quarterly",
  ["nasza przeslosc"] = "Nasza przeszłość",
  ["nat acad sci lett"] = "National Academy Science Letters",
  ["nat afr"] = "Nature Africa",
  ["nat aging"] = "Nature aging",
  ["nat areas j"] = "Natural Areas Journal",
  ["nat astron"] = "Nature astronomy",
  ["nat belg"] = "Naturalistes Belges, Les",
  ["nat biomed eng"] = "Nature biomedical engineering",
  ["nat biotechnol"] = "Nature Biotechnology",
  ["nat can (qué)"] = "Naturaliste Canadien (Québec)",
  ["nat cancer"] = "Nature cancer",
  ["nat cardiovasc res"] = "Nature Cardiovascular Research",
  ["nat catal"] = "Nature Catalysis",
  ["nat cell biol"] = "Nature cell biology",
  ["nat chem"] = "Nature chemistry",
  ["nat chem biol"] = "Nature chemical biology",
  ["nat chem eng"] = "Nature Chemical Engineering",
  ["nat cities"] = "Nature Cities",
  ["nat clim chang"] = "Nature climate change",
  ["nat clim change"] = "Nature Climate Change",
  ["nat clin pract cardiovasc med"] = "Nature clinical practice. Cardiovascular medicine",
  ["nat clin pract endocrinol metab"] = "Nature clinical practice. Endocrinology & metabolism",
  ["nat clin pract gastroenterol hepatol"] = "Nature clinical practice. Gastroenterology & hepatology",
  ["nat clin pract nephrol"] = "Nature clinical practice. Nephrology",
  ["nat clin pract neurol"] = "Nature clinical practice. Neurology",
  ["nat clin pract oncol"] = "Nature clinical practice. Oncology",
  ["nat clin pract rheumatol"] = "Nature clinical practice. Rheumatology",
  ["nat clin pract urol"] = "Nature clinical practice. Urology",
  ["nat commun"] = "Nature communications",
  ["nat comput"] = "Natural computing",
  ["nat comput sci"] = "Nature computational science",
  ["nat comput ser"] = "Natural Computing Series",
  ["nat conserv"] = "Nature conservation",
  ["nat conserv bulg"] = "Nature Conservation Bulgaria",
  ["nat conserv res"] = "Nature conservation research = Zapovednai︠a︡ nauka",
  ["nat croat"] = "Natura croatica : periodicum Musei Historiae Naturalis Croatici = časopis Hrvatskoga prirodoslovnog muzeja",
  ["nat cult"] = "Shizen to bunka (Fukuoka, Japan)",
  ["nat dig"] = "Nature Digest",
  ["nat ecol evol"] = "Nature ecology & evolution",
  ["nat electron"] = "Nature electronics",
  ["nat energy"] = "Nature Energy",
  ["nat eng sci"] = "Natural and engineering sciences",
  ["nat environ"] = "Nature and environment",
  ["nat environ pollut technol"] = "Nature, Environment and Pollution Technology",
  ["nat food"] = "Nature food",
  ["nat gas electr"] = "Natural Gas & Electricity",
  ["nat gas ind b"] = "Natural Gas Industry B",
  ["nat genet"] = "Nature Genetics",
  ["nat geogr mag"] = "National Geographic Magazine",
  ["nat geosci"] = "Nature Geoscience",
  ["nat hazard"] = "Natural Hazards",
  ["nat hazard res"] = "Natural Hazards Research",
  ["nat hazard rev"] = "Natural Hazards Review",
  ["nat hazards (dordr)"] = "Natural hazards (Dordrecht, Netherlands)",
  ["nat hazards earth syst sci"] = "Natural Hazards and Earth System Sciences",
  ["nat hazards earth syst sci discuss"] = "Natural Hazards and Earth System Sciences Discussions",
  ["nat hazards rev"] = "Natural hazards review",
  ["nat health"] = "Nature Health",
  ["nat heim (munst)"] = "Natur und heimat (Münster)",
  ["nat hist"] = "Natural history",
  ["nat hist bull siam soc"] = "The Natural history bulletin of the Siam Society",
  ["nat hist collect museomics"] = "Natural History Collections and Museomics",
  ["nat hist res spec issue"] = "Natural history research. Special issue",
  ["nat hist sci"] = "Natural history sciences : atti della Società italiana di scienze naturali e del Museo civico di storia naturale in Milano",
  ["nat hum activities"] = "Nature and human activities",
  ["nat hum behav"] = "Nature human behaviour",
  ["nat immun"] = "Natural immunity",
  ["nat immun cell growth regul"] = "Natural immunity and cell growth regulation",
  ["nat immunol"] = "Nature immunology",
  ["nat ind"] = "Nature India",
  ["nat inf"] = "Nature Information",
  ["nat inst econ rev"] = "National Institute Economic Review",
  ["nat italy"] = "Nature Italy",
  ["nat j (opole)"] = "Nature Journal (Opole)",
  ["nat landsch"] = "Natur und Landschaft",
  ["nat lang eng"] = "Natural language engineering",
  ["nat lang linguist theory"] = "Natural language & linguistic theory",
  ["nat lang process j"] = "Natural Language Processing Journal",
  ["nat law forum"] = "Natural law forum",
  ["nat life southeast asia"] = "Nature and Life in Southeast Asia",
  ["nat mach intell"] = "Nature machine intelligence",
  ["nat mater"] = "Nature materials",
  ["nat med"] = "Nature medicine",
  ["nat med (n y, ny, u s)"] = "Nature Medicine (New York, NY, United States)",
  ["nat mensch"] = "Natur und Mensch",
  ["nat ment health"] = "Nature Mental Health",
  ["nat metab"] = "Nature metabolism",
  ["nat methods"] = "Nature methods",
  ["nat microbiol"] = "Nature microbiology",
  ["nat monspel ser bot"] = "Naturalia Monspeliensia. Série botanique",
  ["nat mus"] = "Natur und Museum",
  ["nat nano"] = "Nature Nano",
  ["nat nanotechnol"] = "Nature nanotechnology",
  ["nat nanotechnology"] = "Nature Nanotechnology",
  ["nat natschutz mecklenbg"] = "Natur und Naturschutz in Mecklenburg",
  ["nat neurosci"] = "Nature neuroscience",
  ["nat new biol"] = "Nature: New biology",
  ["nat on line"] = "Natureza on line",
  ["nat philos"] = "Natur und Philosophie",
  ["nat photonics"] = "Nature Photonics",
  ["nat phys"] = "Nature Physics",
  ["nat plants"] = "Nature plants",
  ["nat prod bioprospect"] = "Natural products and bioprospecting",
  ["nat prod chem res"] = "Natural products chemistry & research",
  ["nat prod commun"] = "Natural Products Communications",
  ["nat prod j"] = "Natural Products Journal",
  ["nat prod lett"] = "Natural product letters",
  ["nat prod rep"] = "Natural Product Reports",
  ["nat prod res"] = "Natural product research",
  ["nat prod res part b"] = "Natural Product Research Part B: Bioactive Natural Products",
  ["nat prod res, part a"] = "Natural Product Research Part A: Structure and Synthesis",
  ["nat prod res, part b"] = "Natural Product Research, Part B: Bioactive Natural Products",
  ["nat prod sci"] = "Natural Product Sciences",
  ["nat prog"] = "Nature et Progrès",
  ["nat protoc"] = "Nature protocols",
  ["nat rep stem cells"] = "Nature reports. stem cells",
  ["nat resour eng"] = "Natural Resources & Engineering",
  ["nat resour forum"] = "Natural resources forum",
  ["nat resour j"] = "Natural resources journal",
  ["nat resour model"] = "Natural Resource Modeling",
  ["nat resour res"] = "Natural Resources Research",
  ["nat resources"] = "Nature and Resources",
  ["nat rev biodivers"] = "Nature Reviews Biodiversity",
  ["nat rev bioeng"] = "Nature Reviews Bioengineering",
  ["nat rev cancer"] = "Nature reviews. Cancer",
  ["nat rev cardiol"] = "Nature reviews. Cardiology",
  ["nat rev chem"] = "Nature Reviews Chemistry",
  ["nat rev clean technol"] = "Nature Reviews Clean Technology",
  ["nat rev clin oncol"] = "Nature reviews. Clinical oncology",
  ["nat rev dis primers"] = "Nature reviews. Disease primers",
  ["nat rev drug discov"] = "Nature reviews. Drug discovery",
  ["nat rev drug discovery"] = "Nature Reviews Drug Discovery",
  ["nat rev earth environ"] = "Nature reviews. Earth & environment",
  ["nat rev electr eng"] = "Nature Reviews Electrical Engineering",
  ["nat rev endocrinol"] = "Nature reviews. Endocrinology",
  ["nat rev gastroenterol hepatol"] = "Nature reviews. Gastroenterology & hepatology",
  ["nat rev genet"] = "Nature reviews. Genetics",
  ["nat rev immunol"] = "Nature reviews. Immunology",
  ["nat rev mater"] = "Nature Reviews Materials",
  ["nat rev methods primers"] = "Nature Reviews Methods Primers",
  ["nat rev microbiol"] = "Nature reviews. Microbiology",
  ["nat rev mol cell biol"] = "Nature reviews. Molecular cell biology",
  ["nat rev nephrol"] = "Nature reviews. Nephrology",
  ["nat rev neurol"] = "Nature reviews. Neurology",
  ["nat rev neurosci"] = "Nature reviews. Neuroscience",
  ["nat rev phys"] = "Nature Reviews Physics",
  ["nat rev psychol"] = "Nature Reviews Psychology",
  ["nat rev rheumatol"] = "Nature reviews. Rheumatology",
  ["nat rev urol"] = "Nature reviews. Urology",
  ["nat rubber res"] = "Natural Rubber Research",
  ["nat sci"] = "Natural Sciences",
  ["nat sci (east lansing)"] = "Nature and science",
  ["nat sci (irvine)"] = "Natural science",
  ["nat sci educ"] = "Natural Sciences Education",
  ["nat sci j"] = "Natural Science Journal",
  ["nat sci sleep"] = "Nature and science of sleep",
  ["nat sci soc"] = "Natures Sciences Societes",
  ["nat sens"] = "Nature Sensors",
  ["nat struct biol"] = "Nature Structural Biology",
  ["nat struct mol biol"] = "Nature structural & molecular biology",
  ["nat sustain"] = "Nature sustainability",
  ["nat sustainability"] = "Nature Sustainability",
  ["nat synth"] = "Nature Synthesis",
  ["nat syst"] = "Nature and system",
  ["nat tax j"] = "National Tax Journal",
  ["nat tech"] = "Natur und Technik",
  ["nat toxins"] = "Natural Toxins",
  ["nat water"] = "Nature Water",
  ["nat westminster bank quart rev"] = "National Westminster Bank Quarterly Review",
  ["nat- landschschutzgeb baden-württ"] = "Natur- und Landschaftsschutzgebiete Baden-Württembergs",
  ["nationalism ethn polit"] = "Nationalism & ethnic politics",
  ["nationalokon tidsskr"] = "Nationaløkonomisk tidsskrift : maanedskrift for samfundsspørgsmaal, økonomi og handel",
  ["nationaløkon tidsskr"] = "Nationaløkonomisk Tidsskrift",
  ["nations bus"] = "Nation's business",
  ["nations health"] = "Nation's health (Washington, D.C.)",
  ["nations natl"] = "Nations and nationalism",
  ["natl acad sci lett"] = "National Academy science letters. National Academy of Sciences, India",
  ["natl acad sci lett (india)"] = "National Academy Science Letters (India)",
  ["natl aids bull"] = "National AIDS bulletin",
  ["natl black law j"] = "National Black law journal",
  ["natl bur econ res bull aging health"] = "National Bureau of Economic Research bulletin on aging and health",
  ["natl bur stand circ (us)"] = "National Bureau of Standards (U.S.), Circular",
  ["natl bur stand misc publ (us)"] = "National Bureau of Standards (U.S.), Miscellaneous Publication",
  ["natl bur stand spec publ (us)"] = "National Bureau of Standards (U.S.), Special Publication",
  ["natl cancer inst carcinog tech rep ser"] = "National Cancer Institute carcinogenesis technical report series",
  ["natl cancer inst monogr"] = "National Cancer Institute monograph",
  ["natl cancer inst res rep"] = "National Cancer Institute research report",
  ["natl cathol bioeth q"] = "The national Catholic bioethics quarterly",
  ["natl cathol report"] = "National Catholic reporter",
  ["natl civ rev"] = "National civic review",
  ["natl clgh poison control cent bull"] = "National Clearinghouse for Poison Control Centers Bulletin",
  ["natl conf dent public relat"] = "National Conference on Dental Public Relations",
  ["natl conf women hiv natl conf women hiv innov care policy prev 1997 los angel calif"] = "National Conference on Women & HIV : Innovation for Care, Policy and Prevention. National Conference on Women & HIV: Innovation for Care, Policy and Prevention (1997 : Los Angeles, Calif.)",
  ["natl dent assoc j"] = "National Dental Association journal",
  ["natl dent health conf"] = "National Dental Health Conference",
  ["natl elect med q"] = "The National eclectic medical quarterly",
  ["natl eng"] = "The National engineer",
  ["natl fam health surv bull"] = "National Family Health Survey bulletin",
  ["natl fed cathol physicians guild newsl"] = "National Federation of Catholic Physicians' Guilds newsletter",
  ["natl forum"] = "National forum",
  ["natl forum hosp health aff"] = "National Forum on Hospital and Health Affairs",
  ["natl geogr mag"] = "National geographic",
  ["natl geogr res"] = "National geographic research",
  ["natl health"] = "National health; a monthly journal devoted to the interests of public health and hygiene, sanitation and sewage purification, town planning, housing, parks and gardens",
  ["natl health stat report"] = "National health statistics reports",
  ["natl hosp health care"] = "National hospital and health care",
  ["natl inst anim health q (tokyo)"] = "National Institute of Animal Health quarterly",
  ["natl inst drug abuse res monogr ser"] = "National Institute on Drug Abuse research monograph series",
  ["natl inst econ rev"] = "National Institute economic review",
  ["natl inst health consens dev conf consens statement"] = "National Institutes of Health consensus development conference consensus statement",
  ["natl inst health consens dev conf summ"] = "National Institutes of Health consensus development conference summaries",
  ["natl j (wash)"] = "National journal",
  ["natl j integr res med"] = "National journal of integrated research in medicine",
  ["natl j maxillofac surg"] = "National journal of maxillofacial surgery",
  ["natl j physiol pharm pharmacol"] = "National journal of physiology, pharmacy and pharmacology",
  ["natl j sociol"] = "National journal of sociology : NJS",
  ["natl jew law rev"] = "National Jewish law review",
  ["natl jew mon"] = "The National Jewish monthly",
  ["natl law j"] = "The National law journal",
  ["natl med care util expend surv b"] = "National Medical Care Utilization and Expenditure Survey (Series). Series B, Descriptive report",
  ["natl med care util expend surv c"] = "National Medical Care Utilization and Expenditure Survey (Series). Series C, Analytical report",
  ["natl med j india"] = "The National medical journal of India",
  ["natl med leg j"] = "National medical-legal journal",
  ["natl negro health news"] = "National Negro health news",
  ["natl netw"] = "National network (Dallas, Tex.)",
  ["natl nosocomial infect study rep"] = "National Nosocomial Infections Study report",
  ["natl obs"] = "National observer (Washington, D.C.)",
  ["natl pap"] = "Nationalities papers",
  ["natl popul news bull"] = "National population news bulletin : newsletter of the National Population Council",
  ["natl rep subacute care"] = "National report on subacute care",
  ["natl rev"] = "National review",
  ["natl saf congr trans"] = "National Safety Congress Transactions",
  ["natl saf news"] = "National safety news",
  ["natl sci open"] = "National Science Open",
  ["natl sci rev"] = "National Science Review",
  ["natl tax j"] = "National tax journal",
  ["natl toxicol program genet modif model rep"] = "National Toxicology Program genetically modified model report",
  ["natl toxicol program tech rep ser"] = "National Toxicology Program technical report series",
  ["natl underwrit [life health]"] = "National Underwriter",
  ["natl underwrit life health"] = "The National underwriter. Life & health insurance edition",
  ["natl union cat manuscr collect"] = "The National union catalog of manuscript collections",
  ["natl univ law rev"] = "National University law review",
  ["natl vital stat rep"] = "National vital statistics reports : from the Centers for Disease Control and Prevention, National Center for Health Statistics, National Vital Statistics System",
  ["natl wea dig"] = "National Weather Digest",
  ["natl widl"] = "National wildlife",
  ["natl womens health rep"] = "National women's health report",
  ["natlpark-forsch schweiz"] = "Nationalpark-Forschung in der Schweiz",
  ["nato adv sci inst ser a life sci"] = "NATO Advanced Science Institutes Series A: Life Sciences",
  ["nato adv sci inst ser b phys"] = "NATO Advanced Science Institutes Series B: Physics",
  ["nato adv sci inst ser c math phys sci"] = "NATO Advanced Science Institutes Series C: Mathematical and Physical Sciences",
  ["nato adv sci inst ser e appl sci"] = "NATO Advanced Science Institutes Series E: Applied Sciences",
  ["nato adv sci inst ser f comput systems sci"] = "NATO Advanced Science Institutes Series F: Computer and Systems Sciences",
  ["nato adv stud inst ser a life sci"] = "NATO advanced study institutes series. Series A, Life sciences",
  ["nato asi ser a life sci"] = "NATO ASI series. Series A, Life sciences",
  ["nato asi ser h cell biol"] = "NATO ASI series. Series H, Cell biology",
  ["nato asi ser ser i glob environ change"] = "NATO Advanced Science Institutes Series, Series I: Global Environmental Change",
  ["nato asi ser, ser a"] = "NATO ASI Series, Series A: Life Sciences",
  ["nato asi ser, ser b"] = "NATO ASI Series, Series B: Physics",
  ["nato asi ser, ser c"] = "NATO ASI Series, Series C: Mathematical and Physical Sciences",
  ["nato asi ser, ser e"] = "NATO ASI Series, Series E: Applied Physics",
  ["nato asi ser, ser g"] = "NATO ASI Series, Series G: Ecological Sciences",
  ["nato asi ser, ser h"] = "NATO ASI Series, Series H: Cell Biology",
  ["nato asi ser, ser i"] = "NATO ASI Series, Series I: Global Environmental Change",
  ["nato challenges modern soc"] = "NATO Challenges of Modern Society",
  ["nato conf ser iv: marine sci"] = "NATO Conference Series IV: Marine Sciences",
  ["nato conf ser vi: materials sci"] = "NATO Conference Series VI: Materials Science",
  ["nato sci peace secur ser d inf commun secur"] = "NATO Science for Peace and Security Series D: Information and Communication Security",
  ["natschutz heute"] = "Naturschutz heute",
  ["natschutz landschpfl hambg"] = "Naturschutz und Landschaftspflege in Hamburg",
  ["natschutz landschpfl niedersachs"] = "Naturschutz und Landschaftspflege in Niedersachsen",
  ["natur mus"] = "Natur und Museum",
  ["natur recht"] = "Natur und Recht",
  ["natur resource modeling"] = "Natural Resource Modeling",
  ["natur sci j harbin normal univ"] = "Natural Sciences Journal of Harbin Normal University",
  ["natur sci j xiangtan univ"] = "Natural Science Journal of Xiangtan University",
  ["natur sci rep ochanomizu univ"] = "Natural Science Report of the Ochanomizu University",
  ["natural comm comput"] = "Natural Communications with Computers",
  ["natural res j"] = "Natural Resources Journal",
  ["natural res modeling"] = "Natural Resource Modeling",
  ["nature (london)"] = "Nature (London)",
  ["nature (london, u k)"] = "Nature (London, United Kingdom)",
  ["nature (paris)"] = "Nature, Paris",
  ["nature (paris, france)"] = "Nature (Paris, France)",
  ["nature new biol"] = "Nature. New Biology",
  ["naturschutz landschaftsplan"] = "Naturschutz und Landschaftsplanung",
  ["naturwiss rundsch"] = "Naturwissenschaftliche Rundschau",
  ["naturwiss rundschau"] = "Naturwissenschaftliche Rundschau",
  ["naturwiss unterr chem"] = "Naturwissenschaft im Unterricht Chemie",
  ["naturwiss unterr phys"] = "Naturwissenschaft im Unterricht Physik",
  ["naturwiss wochenschr"] = "Naturwissenschaftliche Wochenschrift",
  ["nauchn obz vopr org zdravookhr istor med sssr"] = "Nauchnyĭ obzor: voprosy organizatsii zdravookhraneniia i istorii meditsiny v SSSR",
  ["nauchni tr nauchnoizsled stomatol inst (sofiia)"] = "Nauchni trudove. Sofia. Nauchno-izsledovatelski stomatologichen institut",
  ["nauchni tr vissh med inst sofiia"] = "Nauchni trudove na Visshiia meditsinski institut, Sofiia",
  ["nauchno-biograf lit"] = "\\cyr Nauchno-Biograficheskaya Literatura",
  ["nauchnye doki vyss shkoly biol nauki"] = "Nauchnye doklady vyssheĭ shkoly. Biologicheskie nauki",
  ["nauka i tekhn progress"] = "Nauka i Tekhnicheskii Progress",
  ["nauka tehnol"] = "Nauka i tehnologii",
  ["naukovi dopovidi natsionalnoho universytetu bioresursiv pryrodokorystuvannia ukr"] = "Naukovi dopovidi Nat︠s︡ionalʹnoho universytetu bioresursiv i pryrodokorystuvanni︠a︡ Ukrainy",
  ["naukovij visn nac medicnogo univ im o o bogomolca"] = "Naukovij vìsnik Nacìonalʹnogo medičnogo unìversitetu ìmenì O.O. Bogomolʹcâ",
  ["naunyn schmiedebergs arch exp pathol pharmakol"] = "Naunyn-Schmiedebergs Archiv für experimentelle Pathologie und Pharmakologie",
  ["naunyn schmiedebergs arch pharmacol"] = "Naunyn-Schmiedeberg's archives of pharmacology",
  ["naunyn schmiedebergs arch pharmakol"] = "Naunyn-Schmiedebergs Archiv für Pharmakologie",
  ["naunyn schmiedebergs arch pharmakol exp pathol"] = "Naunyn-Schmiedebergs Archiv für Pharmakologie und experimentelle Pathologie",
  ["naunyn-schmiedeberg’s arch exp pathol pharmacol"] = "Naunyn-Schmiedebergs Archiv fur Experimentelle Pathologie und Pharmacologie",
  ["naunyn-schmiedeberg’s arch pharmacol"] = "Naunyn-Schmiedeberg’s Archives of Pharmacology",
  ["nautilus (philadelphia)"] = "The Nautilus",
  ["nav eng j"] = "Naval Engineers Journal",
  ["nav res logist"] = "Naval Research Logistics",
  ["nav res logist q"] = "Naval Research Logistics Quarterly",
  ["nav rev"] = "Naval review (Annapolis, Md.)",
  ["naval res logist"] = "Naval Research Logistics",
  ["navig- j inst navig"] = "Navigation- Journal of the Institute of Navigation",
  ["navy med"] = "Navy medicine",
  ["nber macroecon annu"] = "NBER macroeconomics annual",
  ["nber monogr"] = "A National Bureau of Economic Research Monograph",
  ["nber work pap ser"] = "Working paper series (National Bureau of Economic Research)",
  ["nbu j plant sci"] = "NBU journal of plant sciences",
  ["nc state data cent newsl"] = "North Carolina State Data Center newsletter",
  ["ncbhr commun"] = "NCBHR communiqué = Communiqué CNBRH",
  ["ncce news"] = "NCCE news",
  ["ncehr commun"] = "NCEHR communiqué = Communiqué CNERH",
  ["nchs data brief"] = "NCHS data brief",
  ["nci monogr"] = "NCI monographs : a publication of the National Cancer Institute",
  ["ncp bull"] = "NCP bulletin",
  ["ncsdha dent hyg"] = "The NCSDHA dental hygienist",
  ["ncsl int meas"] = "NCSL International measure",
  ["ncsl legisbrief"] = "NCSL legisbrief",
  ["ncsli meas"] = "NCSLI Measure",
  ["nda j"] = "NDA journal",
  ["ndds symp"] = "NDSS symposium",
  ["ndt & e int"] = "NDT & E International",
  ["ndt and e int"] = "NDT and E International",
  ["ndt e int"] = "NDT & E international : independent nondestructive testing and evaluation",
  ["ndt int"] = "NDT International",
  ["ndt plus"] = "NDT plus",
  ["nea rhome"] = "Νέα Ῥώμη. Rivista di ricerche bizantinistiche",
  ["near east archaeol"] = "Near Eastern Archaeology",
  ["near surf geophys"] = "Near Surface Geophysics",
  ["neb law rev"] = "Nebraska law review",
  ["nebr hist"] = "Nebraska history",
  ["nebr j econ bus"] = "Nebraska Journal of Economics and Business",
  ["nebr med j"] = "The Nebraska medical journal",
  ["nebr nurse"] = "Nebraska nurse",
  ["nebr state med j"] = "The Nebraska state medical journal",
  ["nebr symp motiv"] = "Nebraska Symposium on Motivation. Nebraska Symposium on Motivation",
  ["nec j adv technol"] = "NEC Journal of Advanced Technology",
  ["nec tech j"] = "NEC Technical Journal",
  ["necas center ser"] = "Necas Center Series",
  ["ned arch kerkgeschied"] = "Nederlandsch archief voor kerkgeschiedenis",
  ["ned bosb-tijdschr"] = "Nederlands Bosbouw-Tijdschrift",
  ["ned milit geneeskd tijdschr"] = "Nederlands militair geneeskundig tijdschrift",
  ["ned tandartsenbl"] = "Nederlands tandartsenblad",
  ["ned tijdschr geneeskd"] = "Nederlands tijdschrift voor geneeskunde",
  ["ned tijdschr gerontol"] = "Nederlands tijdschrift voor gerontologie",
  ["ned tijdschr hyg, microbiol serol"] = "Nederlandsch Tijdschrift voor Hygiene, Microbiologie en Serologie",
  ["ned tijdschr natuurkd"] = "Nederlands Tijdschrift voor Natuurkunde",
  ["ned tijdschr natuurkd (den haag 1934)"] = "Nederlands tijdschrift voor natuurkunde",
  ["ned tijdschr psychol"] = "Nederlands tijdschrift voor de psychologie en haar grensgebieden",
  ["ned tijdschr tandheelkd"] = "Nederlands tijdschrift voor tandheelkunde",
  ["ned tijdschr verloskd gynaecol"] = "Nederlandsch tijdschrift voor verloskunde en gynaecologie",
  ["ned tijdshr evidence based pract"] = "Nederlandsch Tijdschrift voor Evidence Based Practice",
  ["needham res inst stud"] = "Needham Research Institute Studies",
  ["neerl volksleven"] = "Neerlands volksleven",
  ["neft khoz"] = "Neftyanoe Khozyaistvo (Petroleum Industry)",
  ["negro educ rev"] = "The Negro educational review",
  ["negro hist bull"] = "Negro history bulletin",
  ["nehw health watch"] = "NEHW health watch",
  ["nei menggu nong ye da xue xue bao"] = "Nei menggu nong ye da xue xue bao : Zi ran ke xue ban",
  ["neijiang shifan xueyuan xuebao"] = "Neijiang shifan xueyuan xuebao",
  ["neimenggu daxue xuebao ziran kexue"] = "Neimenggu Daxue Xuebao",
  ["neimenggu shida xuebao ziran kexue hanwen ban"] = "Neimenggu Shida Xuebao",
  ["nejm ai"] = "NEJM AI",
  ["nejm catal"] = "NEJM catalyst",
  ["nejm catal innov care deliv"] = "NEJM catalyst innovations in care delivery",
  ["nejm evidence"] = "NEJM Evidence",
  ["nelinein din"] = "Nelineinaya Dinamika",
  ["neliniini koliv"] = "Neliniini Kolivannya",
  ["nematol bras"] = "Nematologia Brasileira",
  ["nematol mediterr"] = "Nematologia mediterranea",
  ["neonatal intensive care"] = "Neonatal intensive care : the journal of perinatology-neonatology",
  ["neonatal netw"] = "Neonatal network : NN",
  ["neonatal pharmacol q"] = "Neonatal pharmacology quarterly",
  ["neopsichiatria (pisa)"] = "Neopsichiatria : rassegna di psichiatria, neurologia, endocrinologia",
  ["neorg mater"] = "Neorganicheskie Materialy",
  ["neotrop biodivers"] = "Neotropical biodiversity",
  ["neotrop biol conserv"] = "Neotropical Biology & Conservation",
  ["neotrop entomol"] = "Neotropical entomology",
  ["neotrop helminthol"] = "Neotropical helminthology",
  ["neotrop ichthyol"] = "Neotropical ichthyology : official journal of the Sociedade Brasileira de Ictiologia",
  ["neotrop primates"] = "Neotropical primates : a newsletter of the Neotropical Section of the IUCN/SSC Primate Specialist Group",
  ["nepal j agric econ"] = "Nepalese journal of agricultural economics",
  ["nepal j epidemiol"] = "Nepal journal of epidemiology",
  ["nepal j ophthalmol"] = "Nepalese journal of ophthalmology : a biannual peer-reviewed academic journal of the Nepal Ophthalmic Society : NEPJOPH",
  ["nepal j sci technol"] = "Nepal journal of science and technology",
  ["nepal med coll j"] = "Nepal Medical College journal : NMCJ",
  ["nepal popul dev j"] = "Nepal population and development journal",
  ["nepali math sci rep"] = "The Nepali Mathematical Sciences Report",
  ["nephrol dial transplant"] = "Nephrology, dialysis, transplantation : official publication of the European Dialysis and Transplant Association - European Renal Association",
  ["nephrol news issues"] = "Nephrology news & issues",
  ["nephrol nurs j"] = "Nephrology nursing journal : journal of the American Nephrology Nurses' Association",
  ["nephrol nurse"] = "Nephrology nurse",
  ["nephrol self assess program"] = "Nephrology self-assessment program : NephSAP",
  ["nephrol ther"] = "Néphrologie & thérapeutique",
  ["nephrology (carlton)"] = "Nephrology (Carlton, Vic.)",
  ["nephron clin pract"] = "Nephron. Clinical practice",
  ["nephron exp nephrol"] = "Nephron. Experimental nephrology",
  ["nephron extra"] = "Nephron extra",
  ["nephron physiol"] = "Nephron. Physiology",
  ["nephrourol mon"] = "Nephro-urology monthly",
  ["nerv child"] = "The Nervous child",
  ["nerv sist"] = "Nervnaia sistema",
  ["nestle nutr inst workshop ser"] = "Nestlé Nutrition Institute workshop series",
  ["nestle nutr workshop ser"] = "Nestlé Nutrition workshop series",
  ["nestle nutr workshop ser clin perform programme"] = "Nestlé Nutrition workshop series. Clinical & performance programme",
  ["nestle nutr workshop ser pediatr program"] = "Nestlé Nutrition workshop series. Paediatric programme",
  ["neth heart j"] = "Netherlands heart journal : monthly journal of the Netherlands Society of Cardiology and the Netherlands Heart Foundation",
  ["neth j geosci"] = "Netherlands Journal of Geosciences",
  ["neth j med"] = "The Netherlands journal of medicine",
  ["neth j psychol"] = "Netherlands journal of psychology",
  ["neth j sea res"] = "Netherlands Journal of Sea Research",
  ["neth j soc sci"] = "The Netherlands' journal of social sciences : a publication of the Netherlands' Sociological and Anthropological Society",
  ["neth j sociol"] = "The Netherlands journal of sociology : Sociologia Neerlandica",
  ["neth j surg"] = "The Netherlands journal of surgery",
  ["neth milk dairy j"] = "Netherlands milk and dairy journal",
  ["netw fr"] = "Network en français",
  ["netw heterog media"] = "Networks and Heterogeneous Media",
  ["netw int fertil res program"] = "Network - International Fertility Research Program",
  ["netw model anal health inform bioinform"] = "Network modeling and analysis in health informatics and bioinformatics",
  ["netw neurosci"] = "Network neuroscience (Cambridge, Mass.)",
  ["netw res triangle park n c"] = "Network (Research Triangle Park, N.C.)",
  ["netw sci (camb univ press)"] = "Network science (Cambridge University Press)",
  ["netw spat econ"] = "Networks and spatial economics",
  ["netw syst med"] = "Network and systems medicine",
  ["netw telecommun ser"] = "Networks and Telecommunications Series",
  ["network model anal health inf bioinf"] = "Network Modeling Analysis in Health Informatics and Bioinformatics",
  ["network neurosci"] = "Network Neuroscience",
  ["network sci"] = "Network Science",
  ["network: comput neural syst"] = "Network: Computation in Neural Systems",
  ["networking sci"] = "Networking Science",
  ["networks (n y)"] = "Networks (New York, NY)",
  ["networks heterogen media"] = "Networks and Heterogeneous Media",
  ["networks spatial econ"] = "Networks & Spatial Economics",
  ["neue heimat"] = "Neue Heimat",
  ["neue jurist wochenschr"] = "Neue juristische Wochenschrift",
  ["neue landsch"] = "Neue Landschaft",
  ["neue munch beitr gesch med naturwiss"] = "Neue Münchner Beiträge zur Geschichte der Medizin und Naturwissenschaften. Naturwissenschaftshistorische Reihe",
  ["neue munch beitr gesch med naturwiss medizinhist reihe"] = "Neue Munchner Beitrage zur Geschichte der Medizin und Naturwissenschaften. Medizinhistorische Reihe",
  ["neue osterr z kinderheilkd"] = "Neue österreichische Zeitschrift für Kinderheilkunde",
  ["neue polit lit"] = "Neue politische Literatur",
  ["neue rundsch"] = "Die Neue Rundschau",
  ["neue zür ztg"] = "Neue Zürcher Zeitung",
  ["neuere med wiss quellen stud"] = "Neuere Medizin- und Wissenschaftsgeschichte",
  ["neues jahrb geol palaontol abh"] = "Neues Jahrbuch für Geologie und Paläontologie. Abhandlungen",
  ["neues jahrb mineral abh"] = "Neues Jahrbuch fur Mineralogie, Abhandlungen",
  ["neues jahrb mineral geol palaeontol abh abt a"] = "Neues Jahrbuch fur Mineralogie, Geologie und Palaeontologie, Abhandlungen Abteiling A: Mineralogie, Petrographie",
  ["neues jahrb mineral geol palaeontol abh abt b"] = "Neues Jahrbuch fur Mineralogie, Geologie und Palaeontologie, Abhandlungen Abteiling B: Geologie, Palaeontologie",
  ["neues jahrb mineral geol palaeontol ref"] = "Neues Jahrbuch fur Mineralogie, Geologie und Palaeontologie, Referate",
  ["neues jahrb mineral monatsh"] = "Neues Jahrbuch fur Mineralogie, Monatshefte",
  ["neues jahrb mineral, abh"] = "Neues Jahrbuch für Mineralogie, Abhandlungen",
  ["neues jahrb mineral, monatsh"] = "Neues Jahrbuch für Mineralogie, Monatshefte",
  ["neujahrsbl leseges wädenswil"] = "Neujahrsblatt der Lesegesellschaft Wädenswil",
  ["neujahrsbl natforsch ges schaffhausen"] = "Neujahrsblatt der Naturforschenden Gesellschaft Schaffhausen",
  ["neujahrsbl natforsch ges zür"] = "Neujahrsblatt der Naturforschenden Gesellschaft in Zürich",
  ["neujahrsbote glarner hinterland"] = "Neujahrsbote für das Glarner Hinterland",
  ["neumol cir torax"] = "Neumologia y cirugía de tórax",
  ["neuphilol mitt"] = "Neuphilologische Mitteilungen",
  ["neural comput"] = "Neural computation",
  ["neural comput appl"] = "Neural computing & applications",
  ["neural dev"] = "Neural development",
  ["neural develop"] = "Neural development",
  ["neural inf process ser"] = "Neural Information Processing Series",
  ["neural netw"] = "Neural networks : the official journal of the International Neural Network Society",
  ["neural netw syst tech appl"] = "Neural Network Systems Techniques and Applications",
  ["neural networks"] = "Neural Networks",
  ["neural parallel sci comput"] = "Neural, Parallel & Scientific Computations",
  ["neural plast"] = "Neural plasticity",
  ["neural process lett"] = "Neural processing letters",
  ["neural regen res"] = "Neural regeneration research",
  ["neural regener res"] = "Neural Regeneration Research",
  ["neural syst circuits"] = "Neural systems & circuits",
  ["neurips 18 compet (2018)"] = "The NeurIPS '18 Competition : from Machine Learning to Intelligent Conversations",
  ["neuro endocrinol lett"] = "Neuro endocrinology letters",
  ["neuro oncol"] = "Neuro-oncology",
  ["neuro-oncol adv"] = "Neuro-Oncology Advances",
  ["neuro-oncol pract"] = "Neuro-Oncology Practice",
  ["neurobehav hiv med"] = "Neurobehavioral HIV medicine",
  ["neurobehav toxicol"] = "Neurobehavioral toxicology",
  ["neurobehav toxicol teratol"] = "Neurobehavioral toxicology and teratology",
  ["neurobiol aging"] = "Neurobiology of aging",
  ["neurobiol aging sci"] = "Neurobiology of Aging Science",
  ["neurobiol circadian rhythms"] = "Neurobiology of Circadian Rhythms",
  ["neurobiol dis"] = "Neurobiology of disease",
  ["neurobiol lang"] = "Neurobiology of Language",
  ["neurobiol lang (camb)"] = "Neurobiology of language (Cambridge, Mass.)",
  ["neurobiol learn mem"] = "Neurobiology of learning and memory",
  ["neurobiol lipids"] = "Neurobiology of lipids",
  ["neurobiol pain"] = "Neurobiology of pain (Cambridge, Mass.)",
  ["neurobiol sleep circadian rhythms"] = "Neurobiology of sleep and circadian rhythms",
  ["neurobiol stress"] = "Neurobiology of stress",
  ["neurobiology (bp)"] = "Neurobiology (Budapest, Hungary)",
  ["neurochem int"] = "Neurochemistry International",
  ["neurochem j"] = "Neurochemical Journal",
  ["neurochem neuropharmacol"] = "Neurochemistry & neuropharmacology : open access",
  ["neurochem pathol"] = "Neurochemical pathology",
  ["neurochem res"] = "Neurochemical Research",
  ["neurochirurgia (stuttg)"] = "Neurochirurgia",
  ["neurocirugia (astur)"] = "Neurocirugía (Asturias, Spain)",
  ["neurocrit care"] = "Neurocritical care",
  ["neurodegener dis"] = "Neuro-degenerative diseases",
  ["neurodegener dis manag"] = "Neurodegenerative disease management",
  ["neurodiagn j"] = "The Neurodiagnostic journal",
  ["neuroembryology aging"] = "Neuroembryology and aging",
  ["neurogastroenterol motil"] = "Neurogastroenterology and motility : the official journal of the European Gastrointestinal Motility Society",
  ["neurogenesis (austin)"] = "Neurogenesis (Austin, Tex.)",
  ["neurographics (2011)"] = "Neurographics (2011)",
  ["neuroimage clin"] = "NeuroImage. Clinical",
  ["neuroimaging clin n am"] = "Neuroimaging clinics of North America",
  ["neuroimmunol neuroinflamm"] = "Neuroimmunology and neuroinflammation",
  ["neuroimmunol rep"] = "Neuroimmunology Reports",
  ["neurol asia"] = "Neurology Asia",
  ["neurol clin"] = "Neurologic clinics",
  ["neurol clin neurophysiol"] = "Neurology & clinical neurophysiology : NCN",
  ["neurol clin neurosci"] = "Neurology and clinical neuroscience",
  ["neurol clin pract"] = "Neurology. Clinical practice",
  ["neurol croat"] = "Neurologia croatica : glasilo Udruzenja neurologa Jugoslavije = official journal of Yugoslav Neurological Association",
  ["neurol disord epilepsy j"] = "Neurological disorders & epilepsy journal",
  ["neurol genet"] = "Neurology. Genetics",
  ["neurol india"] = "Neurology India",
  ["neurol int"] = "Neurology international",
  ["neurol med chir (tokyo)"] = "Neurologia medico-chirurgica",
  ["neurol neurobiol"] = "Neurology and neurobiology",
  ["neurol neurobiol (tallinn)"] = "Neurology and neurobiology (Tallinn, Estonia)",
  ["neurol neurochir pol"] = "Neurologia i neurochirurgia polska",
  ["neurol neurochir psychiatr pol"] = "Neurologia, neurochirurgia i psychiatria polska",
  ["neurol neurocir psiquiatr"] = "Neurología, neurocirugía, psiquiatría",
  ["neurol neuroimmunol neuroinflamm"] = "Neurology® neuroimmunology & neuroinflammation",
  ["neurol neurophysiol neurosci"] = "Neurology, neurophysiology, and neuroscience",
  ["neurol psihiatr neurochir"] = "Neurologia, psihiatria, neurochirurgia",
  ["neurol psychiatr (bucur)"] = "Neurologie et psychiatrie",
  ["neurol psychiatr ceskoslov"] = "Neurologie a psychiatrie c̆eskoslovenská",
  ["neurol psychiatry brain res"] = "Neurology, psychiatry, and brain research",
  ["neurol res"] = "Neurological research",
  ["neurol res int"] = "Neurology research international",
  ["neurol res pract"] = "Neurological research and practice",
  ["neurol res ther"] = "Neurological research and therapy",
  ["neurol sci"] = "Neurological sciences : official journal of the Italian Neurological Society and of the Italian Society of Clinical Neurophysiology",
  ["neurol sci neurophysiol"] = "Neurological Sciences and Neurophysiology",
  ["neurol sci neurosurg"] = "Neurological sciences and neurosurgery",
  ["neurol ther"] = "Neurology and therapy",
  ["neurologisch (wien)"] = "Neurologisch : fachmagazin für neurologie",
  ["neurologist (hyderabad)"] = "The neurologist (Hyderabad, India)",
  ["neurology (ecronicon)"] = "Neurology (E-Cronicon)",
  ["neuromodulation: technol neural interface"] = "Neuromodulation: Technology at the Neural Interface",
  ["neuromol med"] = "NeuroMolecular Medicine",
  ["neuromolecular med"] = "Neuromolecular medicine",
  ["neuromorph comput eng"] = "Neuromorphic Computing and Engineering",
  ["neuromuscul disord"] = "Neuromuscular disorders : NMD",
  ["neuron behav data anal theory"] = "Neurons, behavior, data analysis and theory",
  ["neuron glia biol"] = "Neuron glia biology",
  ["neuronal signal"] = "Neuronal signaling",
  ["neuronal signaling"] = "Neuronal Signaling",
  ["neurooncol adv"] = "Neuro-oncology advances",
  ["neurooncol pract"] = "Neuro-oncology practice",
  ["neuropathol  appl neurobiol"] = "Neuropathology and Applied Neurobiology",
  ["neuropathol appl neurobiol"] = "Neuropathology and applied neurobiology",
  ["neuropatol pol"] = "Neuropatologia polska",
  ["neuropeptides (amsterdam, neth)"] = "Neuropeptides (Amsterdam, Netherlands)",
  ["neuropeptides (oxford, u k)"] = "Neuropeptides (Oxford, United Kingdom)",
  ["neurophysiol clin"] = "Neurophysiologie clinique = Clinical neurophysiology",
  ["neuropsychiatr dis treat"] = "Neuropsychiatric disease and treatment",
  ["neuropsychiatr electrophysiol"] = "Neuropsychiatric electrophysiology",
  ["neuropsychiatr enfance adolesc"] = "Neuropsychiatrie de l'enfance et de l'adolescence",
  ["neuropsychiatry neuropsychol behav neurol"] = "Neuropsychiatry, neuropsychology, and behavioral neurology",
  ["neuropsychol dev cogn b aging neuropsychol cogn"] = "Neuropsychology, development, and cognition. Section B, Aging, neuropsychology and cognition",
  ["neuropsychol rehabil"] = "Neuropsychological rehabilitation",
  ["neuropsychol rev"] = "Neuropsychology review",
  ["neuropsychopharmacol hung"] = "Neuropsychopharmacologia Hungarica : a Magyar Pszichofarmakológiai Egyesület lapja = official journal of the Hungarian Association of Psychopharmacology",
  ["neuropsychopharmacol rep"] = "Neuropsychopharmacology reports",
  ["neuroradiol j"] = "The neuroradiology journal",
  ["neurorehabil neural repair"] = "Neurorehabilitation and neural repair",
  ["neurosci appl"] = "Neuroscience Applied",
  ["neurosci behav physiol"] = "Neuroscience and Behavioral Physiology",
  ["neurosci biobehav rev"] = "Neuroscience and biobehavioral reviews",
  ["neurosci bull"] = "Neuroscience bulletin",
  ["neurosci commun (houst)"] = "Neuroscience communications",
  ["neurosci conscious"] = "Neuroscience of consciousness",
  ["neurosci inf"] = "Neuroscience Informatics",
  ["neurosci insights"] = "Neuroscience insights",
  ["neurosci j"] = "Neuroscience journal",
  ["neurosci lett"] = "Neuroscience Letters",
  ["neurosci lett suppl"] = "Neuroscience letters. Supplement",
  ["neurosci med"] = "Neuroscience and medicine",
  ["neurosci neuroecon"] = "Neuroscience and neuroeconomics",
  ["neurosci news"] = "Neuroscience news",
  ["neurosci res"] = "Neuroscience research",
  ["neurosci res (n y)"] = "Neurosciences research",
  ["neurosci res commun"] = "Neuroscience research communications",
  ["neurosci res program bull"] = "Neurosciences Research Program bulletin",
  ["neurosci res suppl"] = "Neuroscience research. Supplement : the official journal of the Japan Neuroscience Society",
  ["neuroscience (amsterdam, neth)"] = "Neuroscience (Amsterdam, Netherlands)",
  ["neuroscience (oxford, u k)"] = "Neuroscience (Oxford, United Kingdom)",
  ["neuroscience (san diego, ca, u s)"] = "Neuroscience (San Diego, CA, United States)",
  ["neurosciences (riyadh)"] = "Neurosciences (Riyadh, Saudi Arabia)",
  ["neurosurg clin n am"] = "Neurosurgery clinics of North America",
  ["neurosurg focus"] = "Neurosurgical focus",
  ["neurosurg q"] = "Neurosurgery quarterly",
  ["neurosurg rev"] = "Neurosurgical review",
  ["neurotox res"] = "Neurotoxicity research",
  ["neurotoxic res"] = "Neurotoxicity Research",
  ["neurotoxicol teratol"] = "Neurotoxicology and teratology",
  ["neurotransmitter (houst)"] = "Neurotransmitter (Houston, Tex.)",
  ["neurotrauma rep"] = "Neurotrauma reports",
  ["neurourol urodyn"] = "Neurourology and urodynamics",
  ["neurovasc imaging"] = "Neurovascular imaging (London)",
  ["neuss jahrb"] = "Neusser Jahrbuch",
  ["neutron news"] = "Neutron News",
  ["nev hist soc q"] = "Nevada Historical Society quarterly (1961)",
  ["nev nurses assoc q newsl"] = "Nevada Nurses Association Quarterly Newsletter",
  ["nev revis statut nev"] = "Nevada revised statutes : containing all statute laws of Nevada of a general, public, and permanent nature. Nevada",
  ["nev rnformation"] = "Nevada RNformation",
  ["nevrone period neurol psichiatr sci affin"] = "Nevrone; periodico di neurologia, psichiatria e scienze affini",
  ["nevropatol psikhiatriia"] = "Nevropatologiia i psikhiatriia",
  ["new afr"] = "New African",
  ["new am rev"] = "New American review",
  ["new astron"] = "New Astronomy",
  ["new astron rev"] = "New Astronomy Reviews",
  ["new atlantis"] = "New Atlantis (Washington, D.C.)",
  ["new bioeth"] = "The New bioethics : a multidisciplinary journal of biotechnology and the body",
  ["new biol"] = "The New biologist",
  ["new biotechnol"] = "New Biotechnology",
  ["new blackfriars"] = "New Blackfriars",
  ["new carbon mater"] = "New Carbon Materials",
  ["new cathol world"] = "New Catholic world",
  ["new community"] = "New community",
  ["new compr biochem"] = "New Comprehensive Biochemistry",
  ["new crit idiom"] = "New Critical Idiom",
  ["new dent"] = "The New dentist : the official journal of the American Student Dental Association",
  ["new dev quant trading invest"] = "New Developments in Quantitative Trading and Investment",
  ["new diamond front carbon technol"] = "New Diamond and Frontier Carbon Technology",
  ["new dir child adolesc dev"] = "New directions for child and adolescent development",
  ["new dir child dev"] = "New directions for child development",
  ["new dir eval"] = "New directions for evaluation",
  ["new dir ment health serv"] = "New directions for mental health services",
  ["new dir philos sci"] = "New Directions in the Philosophy of Science",
  ["new dir stud leadersh"] = "New directions for student leadership",
  ["new dir youth dev"] = "New directions for youth development",
  ["new dis rep"] = "New disease reports",
  ["new econ windows"] = "New Economic Windows",
  ["new economy"] = "New Economy",
  ["new egypt j med"] = "New Egyptian journal of medicine",
  ["new egypt j microbiol"] = "New Egyptian Journal of Microbiology",
  ["new engl j hist"] = "The New England journal of history",
  ["new engl j prison law"] = "New England journal on prison law",
  ["new engl law rev"] = "New England law review",
  ["new engl q"] = "The New England quarterly",
  ["new england econ rev"] = "New England Economic Review",
  ["new era nurs image int"] = "New era nursing image international : NENI",
  ["new for"] = "New Forests",
  ["new for (dordr)"] = "New forests",
  ["new front ophthalmol"] = "New frontiers in ophthalmology (London)",
  ["new front reg sci asian perspect"] = "New Frontiers in Regional Science: Asian Perspectives",
  ["new gener comput"] = "New generation computing",
  ["new genet soc"] = "New genetics and society",
  ["new ger crit"] = "New German critique, NGC",
  ["new guard"] = "New guard",
  ["new health"] = "New health",
  ["new hibernia rev"] = "New hibernia review = Iris éireannach nua",
  ["new horiz"] = "New horizons (Baltimore, Md.)",
  ["new horiz transl med"] = "New Horizons in Translational Medicine",
  ["new humanist"] = "New humanist (London, England)",
  ["new hung q"] = "The New Hungarian quarterly",
  ["new icmi stud ser"] = "New ICMI Studies Series",
  ["new ideas psychol"] = "New ideas in psychology",
  ["new int"] = "New internationalist",
  ["new iraqi j med"] = "The new Iraqi journal of medicine",
  ["new istanbul contrib clin sci"] = "New Istanbul contribution to clinical science",
  ["new j (inst health rec inf manag)"] = "New journal (Institute of Health Record & Information Management)",
  ["new j bot"] = "New Journal of Botany",
  ["new j botany"] = "New journal of botany",
  ["new j chem"] = "New Journal of Chemistry",
  ["new j phys"] = "New Journal of Physics",
  ["new j sci"] = "New journal of science",
  ["new law j"] = "The New law journal",
  ["new lead"] = "New leader (New York, N.Y. : 1935)",
  ["new left rev"] = "New Left review",
  ["new lit hist"] = "New literary history",
  ["new martin gardner math libr"] = "New Martin Gardner Mathematical Library",
  ["new mater compd appl"] = "New Materials, Compounds and Applications",
  ["new math library"] = "New Mathematical Library",
  ["new math methods syst appl set"] = "New Mathematical Methods, Systems and Applications Set",
  ["new math monogr"] = "New Mathematical Monographs",
  ["new math nat comput"] = "New Mathematics and Natural Computation",
  ["new media soc"] = "New media & society",
  ["new mediev lit (turnhout)"] = "New medieval literatures",
  ["new mexico bus"] = "New Mexico Business",
  ["new microbes new infect"] = "New microbes and new infections",
  ["new microbiol"] = "The new microbiologica",
  ["new nano"] = "New Nano",
  ["new negat plant sci"] = "New Negatives in Plant Science",
  ["new orleans med surg j"] = "The New Orleans medical and surgical journal",
  ["new perspect q"] = "New perspectives quarterly : NPQ",
  ["new perspect turk"] = "New perspectives on Turkey",
  ["new philipp"] = "New Philippines",
  ["new philos"] = "The New philosophy",
  ["new physician"] = "The New physician",
  ["new phytol"] = "The New phytologist",
  ["new polit economy"] = "New Political Economy",
  ["new polym mater"] = "New Polymeric Materials",
  ["new repub"] = "New republic (New York, N.Y.)",
  ["new rev bioeth"] = "New review of bioethics",
  ["new rev hypermedia multimed"] = "The new review of hypermedia and multimedia",
  ["new scholasticism"] = "The New scholasticism",
  ["new school psychol bull"] = "The New School psychology bulletin",
  ["new sci"] = "New Scientist",
  ["new sci (1956)"] = "New scientist",
  ["new sci sci j"] = "New scientist and science journal",
  ["new ser"] = "New Series",
  ["new soc"] = "New society",
  ["new solut"] = "New solutions : a journal of environmental and occupational health policy : NS",
  ["new south"] = "New South (Atlanta, Ga.)",
  ["new space"] = "New space",
  ["new statesman"] = "New statesman (London, England : 1957)",
  ["new stud am intellect cult hist"] = "New Studies in American Intellectual and Cultural History",
  ["new technol work employ"] = "New technology, work and employment",
  ["new theol rev"] = "New theology review",
  ["new times"] = "New times",
  ["new trends czech econ"] = "New Trends Czech. Econ.",
  ["new trends image anal process iciap 2015 workshops (2015)"] = "New trends in image analysis and processing -- ICIAP 2015 Workshops : ICIAP 2015 International Workshops, BioFor, CTMR, RHEUMA, ISCA, MADiMa, SBMI, and QoEM, Genoa, Italy, September 7-8, 2015, Proceedings",
  ["new trends math sci"] = "New Trends in Mathematical Sciences",
  ["new univ q"] = "The New universities quarterly",
  ["new vis"] = "New vision (Kampala, Uganda)",
  ["new voices psychol"] = "New voices in psychology",
  ["new west"] = "New West",
  ["new world"] = "New world (New Orleans, La.)",
  ["new world ir nurs"] = "The New world of Irish nursing : official journal of Irish Nurses Organisation and National Council of Nurses",
  ["new york"] = "New York",
  ["new york j math"] = "New York Journal of Mathematics",
  ["new york rev books"] = "The New York review of books",
  ["new yorker"] = "New Yorker (New York, N.Y. : 1925)",
  ["new zealand econ pap"] = "New Zealand Economic Papers",
  ["new zealand geographer"] = "New Zealand Geographer",
  ["new zealand j math"] = "New Zealand Journal of Mathematics",
  ["new zealand math soc newslett"] = "New Zealand Mathematical Society Newsletter",
  ["newborn infant nurs rev"] = "Newborn and infant nursing reviews : NAINR",
  ["newcastle med j"] = "Newcastle medical journal",
  ["newfoundl med assoc j"] = "Newfoundland Medical Association journal",
  ["newport hist"] = "Newport history : bulletin of the Newport Historical Society",
  ["news bull indian dent assoc"] = "News Bulletin - Indian Dental Association",
  ["news for hist"] = "News of Forest History",
  ["news lett aaf rheum fever control program"] = "News letter (AAF Rheumatic Fever Control Program (U.S.))",
  ["news lett am assoc psychiatr soc work"] = "News-letter. American Association of Psychiatric Social Workers",
  ["news lett florence nightingale int nurs assoc"] = "News Letter; Florence Nightingale International Nurses Association",
  ["news lett maine state nurses assoc"] = "News letter. Maine State Nurses' Association",
  ["news lett natl counc rehabil"] = "News letter. National Council on Rehabilitation",
  ["news lett popul cent bangalore"] = "News letter (Population Centre Bangalore)",
  ["news nor"] = "News of Norway",
  ["news notes ohio dent j"] = "News Notes / Ohio Dental Journal",
  ["news phi delta epsil"] = "News. Phi Delta Epsilon",
  ["news physiol sci"] = "News in physiological sciences : an international journal of physiology produced jointly by the International Union of Physiological Sciences and the American Physiological Society",
  ["news views"] = "News & Views : Rhode Island Nursing Magazine",
  ["news views bangladesh inst res promot essent reprod health technol"] = "News and views - Bangladesh Institute of Research for Promotion of Essential & Reproductive Health and Technologies",
  ["newsl am acad health adm"] = "Newsletter - American Academy of Health Administration. American Academy of Health Administration",
  ["newsl am acad implant dent"] = "Newsletter. American Academy of Implant Dentistry",
  ["newsl am acad pediatr"] = "Newsletter. American Academy of Pediatrics",
  ["newsl am acad psychiatry law"] = "Newsletter (American Academy of Psychiatry and the Law)",
  ["newsl am assoc tissue banks"] = "Newsletter / American Association Of Tissue Banks",
  ["newsl am epilepsy leag"] = "Newsletter. American Epilepsy League",
  ["newsl am soc anesthesiol"] = "Newsletter - American Society of Anesthesiologists",
  ["newsl biomed saf stand"] = "The Newsletter of biomedical safety & standards",
  ["newsl br soc hist sci"] = "Newsletter. British Society for the History of Science",
  ["newsl can soc hist med"] = "Newsletter (Canadian Society for the History of Medicine)",
  ["newsl fam plan assoc hong kong"] = "Newsletter - Family Planning Association of Hong Kong. Family Planning Association of Hong Kong",
  ["newsl fam plan int assist"] = "Newsletter - Family Planning International Assistance. Planned Parenthood Federation of America. Family Planning International Assistance",
  ["newsl feminism philos"] = "Newsletter on feminism and philosophy",
  ["newsl group use psychol hist"] = "Newsletter - Group for the Use of Psychology in History. Group for the Use of Psychology in History",
  ["newsl hist anthropol"] = "Newsletter, history of anthropology",
  ["newsl infant feed action coalit"] = "Newsletter (Infant Feeding Action Coalition)",
  ["newsl inst interam aff"] = "Newsletter. Institute of Inter-American Affairs (U.S.). Health and Sanitation Division",
  ["newsl int acad periodontol"] = "Newsletter (International Academy of Periodontology)",
  ["newsl int coll dent india sect"] = "Newsletter. International College of Dentists. India Section",
  ["newsl int union sci study popul"] = "Newsletter. International Union for the Scientific Study of Population",
  ["newsl inter afr comm tradit pract affect health women child"] = "Newsletter (Inter-African Committee on Traditional Practices Affecting the Health of Women and Children)",
  ["newsl ky nurses assoc"] = "Newsletter. Kentucky Nurses Association",
  ["newsl macro syst inst resour dev demogr health surv"] = "Newsletter (Macro Systems. Institute for Resource Development. Demographic and Health Surveys)",
  ["newsl med sci hist soc"] = "Newsletter (Medical Sciences Historical Society (Great Britain))",
  ["newsl nepal math soc"] = "Newsletter. Nepal Mathematical Society",
  ["newsl oral hist assoc"] = "Newsletter. Oral History Association",
  ["newsl scand soc forensic odontol"] = "News letter. Scandinavian Society of Forensic Odontology",
  ["newsl sci technol human values"] = "Newsletter on science, technology & human values",
  ["newsl soc anc med"] = "Newsletter (Society for Ancient Medicine)",
  ["newsl soc anc med pharm"] = "Newsletter (Society for Ancient Medicine & Pharmacy)",
  ["newsl soc welf hist group"] = "Newsletter-Social Welfare History Group. Social Welfare History Group",
  ["newsl springfield dent soc"] = "Newsletter - Springfield Dental Society",
  ["newsl stratigr"] = "Newsletters on stratigraphy",
  ["newsl suffolk cty med soc"] = "News letter. Suffolk County Medical Society",
  ["newsl wis league nurs"] = "Newsletter. Wisconsin League for Nursing",
  ["newsl womens glob netw reprod rights"] = "Newsletter (Women's Global Network on Reproductive Rights)",
  ["newsline people aids coalit n y"] = "Newsline (People with AIDS Coalition of New York)",
  ["newsp res j"] = "Newspaper research journal",
  ["next gener seq appl"] = "Next generation, sequencing & applications",
  ["nexus netw j"] = "Nexus network journal",
  ["nexus network j"] = "Nexus Network Journal",
  ["nfais newsl"] = "NFAIS newsletter",
  ["nfi bull"] = "NFI bulletin",
  ["nfpa j"] = "NFPA journal : the official magazine of the National Fire Protection Association",
  ["nfs j"] = "NFS Journal",
  ["nghien cuu lich su"] = "Nghiên cứu lịch sử",
  ["nhpf issue brief"] = "NHPF issue brief",
  ["nhsa dialog"] = "NHSA dialog",
  ["ni 2012 (2012)"] = "NI 2012 : 11th International Congress on Nursing Informatics, June 23-27, 2012, Montreal, Canada. International Congress in Nursing Informatics (11th : 2012 : Montréal, Québec)",
  ["niagara frontier"] = "Niagara frontier",
  ["niaid aids agenda"] = "NIAID AIDS agenda",
  ["nicar med"] = "Nicaragua médica",
  ["nicar odontol"] = "Nicaragua Odontologica",
  ["nice hist"] = "Nice historique",
  ["nice med"] = "Nice médical",
  ["nichibunken newsl"] = "Nichibunken newsletter = Nihon Bunka",
  ["nichidai igaku zasshi"] = "Nichidai igaku zasshi",
  ["nichidai koko kagaku"] = "Nichidai kōkū kagaku = Nihon University journal of oral science",
  ["nicotin tob res"] = "Nicotine & Tobacco Research",
  ["nicotine tob res"] = "Nicotine & tobacco research : official journal of the Society for Research on Nicotine and Tobacco",
  ["nida res monogr"] = "NIDA research monograph",
  ["niederdtsch mitt"] = "Niederdeutsche Mitteilungen",
  ["niederhein jahrb"] = "Niederrheinisches Jahrbuch",
  ["niedersachs arztebl"] = "Niedersächsisches Aerzteblatt",
  ["niedersachs zahnarztebl"] = "Niedersachsisches Zahnarzteblatt",
  ["niedersächs akad geowiss veröff"] = "Niedersächsische Akademie der Geowissenschaften Veröffentlichungen",
  ["nieuw arch wisk (4)"] = "Nieuw Archief voor Wiskunde",
  ["nieuw arch wiskd (5)"] = "Nieuw Archief voor Wiskunde. Vijfde Serie",
  ["nieuwe ned bijdr geschied geneeskd natuurwet"] = "Nieuwe Nederlandse bijdragen tot de geschiedenis der geneeskunde en der natuurwetenschappen",
  ["nieuwe west indische gids"] = "Nieuwe West-Indische gids",
  ["nig q j hosp med"] = "Nigerian quarterly journal of hospital medicine",
  ["niger behav sci j"] = "Nigerian behavioral sciences journal",
  ["niger classics"] = "Nigeria and the classics",
  ["niger dent j"] = "Nigerian Dental Journal",
  ["niger food j"] = "Nigerian Food Journal",
  ["niger j biotechnol"] = "Nigerian journal of biotechnology",
  ["niger j clin pract"] = "Nigerian journal of clinical practice",
  ["niger j econ soc stud"] = "The Nigerian journal of economic and social studies",
  ["niger j med"] = "Nigerian journal of medicine : journal of the National Association of Resident Doctors of Nigeria",
  ["niger j nat prod med"] = "Nigerian journal of natural products and medicine",
  ["niger j nutr sci"] = "Nigerian Journal of Nutrional Sciences",
  ["niger j paediatr"] = "Nigerian journal of paediatrics",
  ["niger j parasitol"] = "Nigerian Journal of Parasitology",
  ["niger j pharm sci"] = "Nigerian journal of pharmaceutical sciences",
  ["niger j physiol sci"] = "Nigerian journal of physiological sciences : official publication of the Physiological Society of Nigeria",
  ["niger j surg"] = "Nigerian journal of surgery : official publication of the Nigerian Surgical Research Society",
  ["niger j surg sci"] = "Nigerian journal of surgical sciences : official journal of the Nigerian Section of International College of Surgeons",
  ["niger j technol dev"] = "Nigerian Journal of Technological Development",
  ["niger mag"] = "Nigeria magazine",
  ["niger med j"] = "Nigerian medical journal : journal of the Nigeria Medical Association",
  ["niger med pract"] = "The Nigerian medical practitioner",
  ["niger nurse"] = "The Nigerian nurse",
  ["niger pop"] = "Nigeria's population : quarterly journal of population activities in Nigeria",
  ["niger postgrad med j"] = "The Nigerian postgraduate medical journal",
  ["niger vet j"] = "Nigerian veterinary journal",
  ["nigerian j econ soc stud"] = "Nigerian Journal of Economic and Social Studies",
  ["nih catal"] = "The NIH catalyst : a publication for NIH intramural scientists",
  ["nih conf retroviral integr nih conf retroviral integr 1995 bethesda md"] = "NIH Conference on Retroviral Integrase : molecular biology and pharmacology : a novel target for the treatment of AIDS, January 19-20, 1995. NIH Conference on Retroviral Integrase (1995 : Bethesda, Md.)",
  ["nih consens dev conf summ"] = "NIH consensus development conference summary",
  ["nih consens state sci statements"] = "NIH consensus and state-of-the-science statements",
  ["nih consens statement"] = "NIH consensus statement",
  ["nih guide grants contracts"] = "NIH guide for grants and contracts (Online)",
  ["nihae bull"] = "NIHAE bulletin",
  ["nihon ago kansetsu gakkai zasshi"] = "Nihon Ago Kansetsu Gakkai Zasshi",
  ["nihon arukoru yakubutsu igakkai zasshi"] = "Nihon Arukōru Yakubutsu Igakkai zasshi = Japanese journal of alcohol studies & drug dependence",
  ["nihon bentosu gakkai shi"] = "Nihon Bentosu Gakkai shi = Benthos research",
  ["nihon biseibutsu seitai gakkaiho"] = "Nihon Biseibutsu Seitai Gakkaihō = Bulletin of Japanese Society of Microbial Ecology",
  ["nihon byori gakkai kaishi"] = "Nihon Byori Gakkai kaishi",
  ["nihon chikusan gakkaiho|anim sci j"] = "Animal science journal",
  ["nihon chokucho komonbyo"] = "Nihon Chokucho Komonbyo Gakkai zasshi",
  ["nihon daicho komonbyo gakkai zasshi"] = "Nihon Daicho Komonbyo Gakkai zasshi",
  ["nihon dani gakkai shi"] = "Nihon Dani Gakkai shi = Journal of the Acarological Society of Japan",
  ["nihon densenbyo gakkai zasshi"] = "Nihon Densenbyo Gakkai zasshi",
  ["nihon dojo hiryogaku zasshi"] = "Nihon dojō hiryōgaku zasshi : dojō, hiryō, shokubutsu eiyō",
  ["nihon eiseigaku zasshi"] = "Nihon eiseigaku zasshi. Japanese journal of hygiene",
  ["nihon eiyo shokuryo gakkai shi"] = "Nihon Eiyō Shokuryō Gakkai shi = Nippon Eiyō Shokuryō Gakkaishi = Journal of Japanese Society of Nutrition and Food Science",
  ["nihon funin gakkai zasshi"] = "Nihon Funin Gakkai zasshi",
  ["nihon gan chiryo gakkai shi"] = "Nihon Gan Chiryo Gakkai shi",
  ["nihon ganka kiyo"] = "Nihon ganka kiyo",
  ["nihon geka gakkai zasshi"] = "Nihon Geka Gakkai zasshi",
  ["nihon geka hokan"] = "Nihon geka hokan. Archiv für japanische Chirurgie",
  ["nihon hansenbyo gakkai zasshi"] = "Nihon Hansenbyō Gakkai zasshi = Japanese journal of leprosy : official organ of the Japanese Leprosy Association",
  ["nihon heikatsukin gakkai zasshi"] = "Nihon Heikatsukin Gakkai zasshi",
  ["nihon hifuka gakkai zasshi"] = "Nihon Hifuka Gakkai zasshi. The Japanese journal of dermatology",
  ["nihon hinyokika gakkai zasshi"] = "Nihon Hinyōkika Gakkai zasshi. The japanese journal of urology",
  ["nihon hoigaku zasshi"] = "Nihon hōigaku zasshi = The Japanese journal of legal medicine",
  ["nihon hoshasen gijutsu gakkai zasshi"] = "Nihon Hōshasen Gijutsu Gakkai zasshi",
  ["nihon hotetsu shika gakkai zasshi"] = "Nihon Hotetsu Shika Gakkai zasshi",
  ["nihon igaku hoshasen gakkai zasshi"] = "Nihon Igaku Hōshasen Gakkai zasshi. Nippon acta radiologica",
  ["nihon ika daigaku zasshi"] = "Nihon Ika Daigaku zasshi",
  ["nihon ishigaku zasshi"] = "Nihon ishigaku zasshi. [Journal of Japanese history of medicine]",
  ["nihon ishikai zasshi"] = "Nihon Ishikai zasshi. Journal of the Japan Medical Association",
  ["nihon ishinkin gakkai zasshi"] = "Nihon Ishinkin Gakkai zasshi = Japanese journal of medical mycology",
  ["nihon jibiinkoka gakkai kaiho"] = "Nihon Jibiinkoka Gakkai kaiho",
  ["nihon jinzo gakkai shi"] = "Nihon Jinzo Gakkai shi",
  ["nihon juigaku zasshi"] = "Nihon juigaku zasshi. The Japanese journal of veterinary science",
  ["nihon kango kagakkaishi"] = "Nihon Kango Kagakkai shi = Journal of Japan Academy of Nursing Science",
  ["nihon kessen shiketsu gakkai shi"] = "Nihon Kessen Shiketsu Gakkai shi = Japanese journal of thrombosis and hemostasis",
  ["nihon ketsueki gakkai zasshi"] = "Nihon Ketsueki Gakkai zasshi : journal of Japan Haematological Society",
  ["nihon kingakkai kaiho"] = "Nihon Kingakkai Kaiho",
  ["nihon koku geka gakkai zasshi"] = "Nihon Koku Geka Gakkai zasshi",
  ["nihon kokyuki gakkai zasshi"] = "Nihon Kokyūki Gakkai zasshi = the journal of the Japanese Respiratory Society",
  ["nihon koshu eisei zasshi"] = "[Nihon kōshū eisei zasshi] Japanese journal of public health",
  ["nihon kyobu geka gakkai zasshi"] = "[Zasshi] [Journal]. Nihon Kyōbu Geka Gakkai",
  ["nihon kyobu shikkan gakkai zasshi"] = "Nihon Kyōbu Shikkan Gakkai zasshi",
  ["nihon kyosei shika gakkai zasshi"] = "Nihon Kyōsei Shika Gakkai zasshi = The journal of Japan Orthodontic Society",
  ["nihon masu sukuriningu gakkai shi"] = "Nihon Masu Sukurīningu Gakkai shi = Journal of Japanese Society for Mass-Screening",
  ["nihon naibunpi gakkai zasshi"] = "Nihon Naibunpi Gakkai zasshi",
  ["nihon naika gakkai zasshi"] = "Nihon Naika Gakkai zasshi. The Journal of the Japanese Society of Internal Medicine",
  ["nihon rai gakkai zasshi"] = "Nihon Rai Gakkai zasshi",
  ["nihon rekishi"] = "Nihon rekishi = The Nihon-rekishi",
  ["nihon reoroji gakkaishi"] = "Nihon Reoroji Gakkaishi",
  ["nihon rinsho"] = "Nihon rinsho. Japanese journal of clinical medicine",
  ["nihon rinsho kekkaku"] = "[Nihon rinshō kekkaku] The Japanese journal of clinical tuberculosis",
  ["nihon rinsho meneki gakkai kaishi"] = "Nihon Rinshō Men'eki Gakkai kaishi = Japanese journal of clinical immunology",
  ["nihon rinsho shishubyo danwakai kaishi"] = "Nihon Rinshō Shishūbyō Danwakai kaishi",
  ["nihon ronen igakkai zasshi"] = "Nihon Ronen Igakkai zasshi. Japanese journal of geriatrics",
  ["nihon saikingaku zasshi"] = "Nihon saikingaku zasshi. Japanese journal of bacteriology",
  ["nihon sanka fujinka gakkai zasshi"] = "Nihon Sanka Fujinka Gakkai zasshi",
  ["nihon sanshigaku zasshi"] = "Nihon sanshigaku zasshi = The journal of sericultural science",
  ["nihon seikeigeka gakkai zasshi"] = "Nihon Seikeigeka Gakkai zasshi",
  ["nihon seirigaku zasshi"] = "Nihon seirigaku zasshi. Journal of the Physiological Society of Japan",
  ["nihon senchu gakkai shi"] = "Nihon Senchū Gakkai shi = Japanese journal of nematology",
  ["nihon shika hyoron"] = "[Nihon shika hyōron] The Nippon dental review",
  ["nihon shika ishikai zasshi"] = "Nihon Shika Ishikai zasshi",
  ["nihon shika zairyo kikai gakkai zasshi"] = "Nihon Shika Zairyo Kikai Gakkai zasshi. The Journal of the Japan Research Society of Dental Materials & Appliances",
  ["nihon shinkei seishin yakurigaku zasshi"] = "Nihon shinkei seishin yakurigaku zasshi = Japanese journal of psychopharmacology",
  ["nihon shishubyo gakkai kaishi"] = "Nihon Shishubyo Gakkai kaishi",
  ["nihon shokaki geka gakkai zasshi"] = "Nihon Shokaki Geka Gakkai zasshi",
  ["nihon shokakibyo gakkai zasshi"] = "Nihon Shokakibyo Gakkai zasshi = The Japanese journal of gastro-enterology",
  ["nihon shoni junkanki gakkai zasshi"] = "Nihon Shōni Junkanki Gakkai zasshi = Pediatric cardiology and cardiac surgery",
  ["nihon shonika gakkai zasshi"] = "Nihon Shonika Gakkai zasshi. Acta paediatrica Japonica",
  ["nihon shoyu kenkyujo zasshi"] = "Nihon Shōyu Kenkyūjo zasshi = Journal of Japan Soy Sauce Research Institute",
  ["nihon sutoma rihabiriteshon gakkaishi j jpn soc stoma rehabil"] = "Nihon Sutoma Rihabiriteshon Gakkaishi [Journal Of Japanese Society Of Stoma Rehabilitation]",
  ["nihon univ comp law"] = "Comparative law",
  ["nihon univ j med"] = "The Nihon University journal of medicine",
  ["nihon yakubutsugaku zasshi"] = "[Nihon yakubutsugaku zasshi] [Japanese journal of pharmacology]",
  ["nihon yakurigaku zasshi"] = "Nihon yakurigaku zasshi. Folia pharmacologica Japonica",
  ["nihon yuketsu gakkai zasshi"] = "Nihon Yuketsu Gakkai zasshi = Journal of the Japan Society of Blood Transfusion",
  ["nihonkai math j"] = "Nihonkai Mathematical Journal",
  ["nihonshi kenkyu"] = "Nihon shi kenkyū",
  ["niigata igakkai zasshi"] = "Niigata Igakkai zasshi. Niigata medical journal",
  ["nina rapp"] = "NINA rapport",
  ["ninchi shinkei kagaku"] = "Ninchi shinkei kagaku",
  ["ninet century"] = "Nineteenth century (Philadelphia, Pa.)",
  ["ninet century contexts"] = "Nineteenth-century contexts",
  ["ninet century fr stud"] = "Nineteenth-century French studies",
  ["ninet century lit"] = "Nineteenth-century literature",
  ["nineteenth century fict"] = "Nineteenth-century fiction",
  ["ninety nine"] = "Ninety nine",
  ["ningxia yixueyuan xuebao"] = "Ningxia Yixueyuan xuebao. Ningxia medical college",
  ["niph ann"] = "NIPH annals",
  ["nippon byori gakkai kaishi"] = "Nippon Byori Gakkai Kaishi (Transactiones Societatis Pathologicae Japonicae)",
  ["nippon chokucho komonbyo"] = "Nippon Chokucho Komonbyo Gakkai zasshi",
  ["nippon daicho komonbyo gakkai zasshi"] = "Nippon Daicho Komonbyo Gakkai Zasshi (Journal of Japan Society of Colo-Proctology)",
  ["nippon densenbyo gakkai zasshi"] = "Nippon Densenbyo Gakkai Zasshi",
  ["nippon eiseigaku zasshi"] = "Nippon Eiseigaku Zasshi (Japanese Journal of Hygiene)",
  ["nippon funin gakkai zasshi"] = "Nippon Funin Gakkai Zasshi (Japanese Journal of Fertility and Sterility)",
  ["nippon gan chiryo gakkai shi"] = "Nippon Gan Chiryo Gakkai Shi (Journal of Japan Society for Cancer Therapy)",
  ["nippon ganka gakkai zasshi"] = "Nippon Ganka Gakkai zasshi",
  ["nippon ganka kiyo"] = "Nippon Ganka Kiyo (Folia Ophthalmologica Japonica. Bulletin of Japanese Ophthalmology)",
  ["nippon geka gakkai zasshi"] = "Nippon Geka Gakkai Zasshi (Journal of Japan Surgical Society)",
  ["nippon geka hokan"] = "Nippon Geka Hokan (Archiv fur Japanische Chirurgie)",
  ["nippon heikatsukin gakkai zasshi"] = "Nippon Heikatsukin Gakkai Zasshi",
  ["nippon hifuka gakkai zasshi"] = "Nippon Hifuka Gakkai Zasshi (Japanese Journal of Dermatology)",
  ["nippon hinyokika gakkai zasshi"] = "Nippon Hinyokika Gakkai Zasshi (Japanese Journal of Urology)",
  ["nippon hoigaku zasshi"] = "Nippon Hoigaku Zasshi (Japanese Journal of Legal Medicine)",
  ["nippon hosenkin gakkaishi"] = "Nihon Hōsenkin Gakkai shi = Actinomycetologica",
  ["nippon hoshasen gijutsu gakkai zasshi"] = "Nippon Hoshasen Gijutsu Gakkai zasshi",
  ["nippon hotetsu shika gakkai zasshi"] = "Nippon Hotetsu Shika Gakkai Zasshi",
  ["nippon igaku hoshasen gakkai zasshi"] = "Nippon Igaku Hoshasen Gakkai Zasshi (Nippon Acta Radiologica)",
  ["nippon ika daigaku zasshi"] = "Nippon Ika Daigaku Zasshi (Journal of the Nippon Medical School)",
  ["nippon ishigaku zasshi"] = "Nihon ishigaku zasshi. [Journal of Japanese history of medicine]",
  ["nippon ishikai zasshi"] = "Nippon Ishikai Zasshi (Journal of the Japan Medical Association)",
  ["nippon ishinkin gakkai zasshi"] = "Nippon Ishinkin Gakkai Zasshi (Japanese Journal of Medical Mycology)",
  ["nippon jibiinkoka gakkai kaiho"] = "Nippon Jibiinkoka Gakkai Kaiho",
  ["nippon jinzo gakkai shi"] = "Nippon Jinzo Gakkai Shi",
  ["nippon juigaku zasshi"] = "Nippon Juigaku Zasshi (Japanese Journal of Veterinary Science)",
  ["nippon juishikai zasshi"] = "Nippon Juishikai zasshi. Journal of the Japan Veterinary Medical Association",
  ["nippon kagaku ryoho gakkai zasshi"] = "Nippon Kagaku Ryoho Gakkai zasshi. Chemotherapy",
  ["nippon kakin gakkaishi"] = "Nihon Kakin Gakkaishi = Japanese poultry science",
  ["nippon ketsueki gakkai zasshi"] = "Nippon Ketsueki Gakkai Zasshi (Acta Haematologica Japonica)",
  ["nippon kikai gakkai ronbunshu, a-hen"] = "Nippon Kikai Gakkai Ronbunshu, A-hen",
  ["nippon kikai gakkai ronbunshu, b-hen"] = "Nippon Kikai Gakkai Ronbunshu, B-hen",
  ["nippon kinzoku gakkaishi"] = "Nippon Kinzoku Gakkaishi",
  ["nippon koku geka gakkai zasshi"] = "Nippon Koku Geka Gakkai Zasshi",
  ["nippon kontakuto renzu gakkai kaiin dayori"] = "Nippon Kontakuto Renzu Gakkai Kaiin Dayori (Journal of Japan Contact Lens Society)",
  ["nippon koshu eisei zasshi"] = "Nippon Koshu Eisei Zasshi (Japanese journal of Public Health)",
  ["nippon kyobu geka gakkai zasshi"] = "Nippon Kyobu Geka Gakkai Zasshi (Journal of the Japanese Association for Thoracic Surgery)",
  ["nippon kyobu shikkan gakkai zasshi"] = "Nippon Kyobu Shikkan Gakkai Zasshi (Japanese Journal of Thoracic Diseases)",
  ["nippon kyosei shika gakkai zasshi"] = "Nippon Kyosei Shika Gakkai Zasshi (Journal of Japan Orthodontic Society)",
  ["nippon naibunpi gakkai zasshi"] = "Nippon Naibunpi Gakkai Zasshi (Folia Endocrinologica Japonica)",
  ["nippon naika gakkai zasshi"] = "Nippon Naika Gakkai Zasshi (Journal of Japanese Society of Internal Medicine)",
  ["nippon nettai igakkai zasshi"] = "Nippon Nettai Igakkai zasshi. Japanese journal of tropical medicine and hygiene",
  ["nippon nogeikagaku kaishi"] = "Nippon Nôgeikagaku Kaishi",
  ["nippon oyo jiki gakkaishi"] = "Nippon Oyo Jiki Gakkaishi",
  ["nippon rai gakkai zasshi"] = "Nippon Rai Gakkai Zasshi (Japanese Journal of Leprosy)",
  ["nippon rinsho"] = "Nippon Rinsho (Japanese Journal of Clinical Medicine)",
  ["nippon ronen igakkai zasshi"] = "Nippon Ronen Igakkai Zasshi (Japanese Journal of Geriatrics)",
  ["nippon saikingaku zasshi"] = "Nippon Saikingaku Zasshi (Japanese Journal of Bacteriology)",
  ["nippon sanka fujinka gakkai zasshi"] = "Nippon Sanka Fujinka Gakkai Zasshi (Acta Obstetrica et Gynaecologica Japonica)",
  ["nippon seikeigeka gakkai zasshi"] = "Nippon Seikeigeka Gakkai Zasshi (Journal of the Japanese Orthopaedic Association)",
  ["nippon seirigaku zasshi"] = "Nippon Seirigaku Zasshi (Journal of the Physiological Society of Japan)",
  ["nippon seramikkusu kyokai gakujutsu ronbunshi"] = "Nippon Seramikkusu Kyokai Gakujutsu Ronbunshi",
  ["nippon shika hyoron"] = "Nippon Shika Hyoron (Nippon Dental Review)",
  ["nippon shika ishikai zasshi"] = "Nippon Shika Ishikai Zasshi (Journal of the Japan Dental Association)",
  ["nippon shika zairyo kikai gakkai zasshi"] = "Nippon Shika Zairyo Kikai Gakkai Zasshi (Journal of the Japan Research Society of Dental Materials and Appliances)",
  ["nippon shishubyo gakkai kaishi"] = "Nippon Shishubyo Gakkai Kaishi (Journal of the Japanese Association of Periodontology)",
  ["nippon shokakibyo gakkai zasshi"] = "Nippon Shokakibyo Gakkai Zasshi (Japanese Journal of Gastroenterology)",
  ["nippon shokubutsu byori gakkaiho"] = "Nihon Shokubutsu Byōri Gakkaiho = Annals of the Phytopathological Society of Japan",
  ["nippon shokuhin kogyo gakkaishi"] = "Nippon Shokuhin Kogyo Gakkaishi",
  ["nippon shonika gakkai zasshi"] = "Nippon Shonika Gakkai Zasshi (Acta Paediatrica Japonica)",
  ["nippon suisan gakkai shi"] = "Nippon Suisan Gakkai Shi. Bulletin Of The Japanese Society Of Scientific Fisheries",
  ["nippon yakurigaku zasshi"] = "Nippon Yakurigaku Zasshi",
  ["nir news"] = "NIR News",
  ["niscl soc"] = "Discourse & Society",
  ["nishi nihon hinyokika"] = "Nishi Nihon hinyokika = [West Japan journal of urology]",
  ["nisshin igaku jpn j med prog"] = "[Nisshin igaku] The Japanese journal of medical progress",
  ["nist spec publ"] = "NIST Special Publication",
  ["nitric oxide"] = "Nitric Oxide",
  ["niwa inf ser"] = "NIWA information series",
  ["njas wageningen j life sci"] = "NJAS - Wageningen Journal of Life Sciences",
  ["njas: impact agric life sci"] = "NJAS: Impact in Agricultural and Life Sciences",
  ["njpc bull"] = "NJPC bulletin",
  ["njsna newsl"] = "Njsna Newsletter",
  ["nlm gatew"] = "NLM gateway",
  ["nln conv pap"] = "NLN convention papers",
  ["nln publ"] = "NLN publications",
  ["nln update"] = "NLN update : connecting members of the National League for Nursing",
  ["nmc case rep j"] = "NMC case report journal",
  ["nmos bull"] = "N.M.O.S. bulletin",
  ["nmr biomed"] = "NMR in biomedicine",
  ["no junkan taisha"] = "No junkan taisha = Cerebral blood flow and metabolism",
  ["no shinkei geka"] = "No shinkei geka. Neurological surgery",
  ["no to hattatsu"] = "No to hattatsu = Brain and development",
  ["no to shinkei"] = "Nō to shinkei = Brain and nerve",
  ["noaa tech memo nmfs swfsc"] = "NOAA technical memorandum NMFS-SWFSC",
  ["noaa tech rep nmfs"] = "NOAA technical report NMFS",
  ["nobel laureate ser"] = "Nobel Laureate Series",
  ["nobel symp"] = "Nobel Symposium",
  ["nodea nonlinear differential equations appl"] = "NoDEA. Nonlinear Differential Equations and Applications",
  ["nodycon conf proc ser"] = "NODYCON Conference Proceedings Series",
  ["nogyogyaszati szuleszeti tovabbkepzo szemle"] = "Nőgyógyászati és szülészeti továbbképző szemle",
  ["noise control eng j"] = "Noise control engineering journal",
  ["noise health"] = "Noise & health",
  ["noise vibr worldwide"] = "Noise & Vibration Worldwide",
  ["nomad people"] = "Nomadic peoples",
  ["nomura res inst quart econ rev"] = "Nomura Research Institute Quarterly Economic Review",
  ["non-coding rna res"] = "Non-coding RNA Research",
  ["non-destr test"] = "Non-Destructive Testing",
  ["nonauton dyn syst"] = "Nonautonomous Dynamical Systems",
  ["noncoding rna"] = "Non-coding RNA",
  ["noncoding rna investig"] = "Non-coding RNA investigation",
  ["noncoding rna res"] = "Non-coding RNA research",
  ["nonconvex optim appl"] = "Nonconvex Optimization and its Applications",
  ["nondestrtest eval"] = "Nondestructive Testing and Evaluation",
  ["nong nghiep nong thon"] = "Nông nghiệp & Nông thôn",
  ["nong ye huan jing ke xue xue bao"] = "Nong ye huan jing ke xue xue bao = Journal of agro-environment science",
  ["nong ye sheng wu ji shu xue bao"] = "Nong ye sheng wu ji shu xue bao = Journal of agricultural biotechnology",
  ["nongyag gwahag hoeji"] = "Nong'yag gwahag hoeji",
  ["nongyaoxue xuebao"] = "Nongyaoxue xuebao",
  ["nonlinear anal"] = "Nonlinear Analysis",
  ["nonlinear anal forum"] = "Nonlinear Analysis Forum. An International Journal",
  ["nonlinear anal hybrid syst"] = "Nonlinear Analysis: Hybrid Systems",
  ["nonlinear anal model control"] = "Nonlinear Analysis. Modelling and Control",
  ["nonlinear anal real world appl"] = "Nonlinear Analysis: Real World Applications",
  ["nonlinear anal theory methods appl"] = "Nonlinear Analysis: Theory Methods & Applications",
  ["nonlinear anal-model control"] = "Nonlinear Analysis: Modelling and Control",
  ["nonlinear biomed phys"] = "Nonlinear biomedical physics",
  ["nonlinear differ equations appl"] = "Nonlinear Differential Equations and Applications",
  ["nonlinear dyn"] = "Nonlinear Dynamics",
  ["nonlinear dyn psychol life sci"] = "Nonlinear Dynamics, Psychology, and Life Sciences",
  ["nonlinear dyn syst theory"] = "Nonlinear Dynamics and Systems Theory. An International Journal of Research and Surveys",
  ["nonlinear dynam"] = "Nonlinear Dynamics",
  ["nonlinear dynamics psychol life sci"] = "Nonlinear dynamics, psychology, and life sciences",
  ["nonlinear opt (mclc) sect b"] = "Nonlinear Optics (Mclc) Section B",
  ["nonlinear opt quantum opt"] = "Nonlinear Optics Quantum Optics",
  ["nonlinear oscillations"] = "Nonlinear Oscillations",
  ["nonlinear phenom complex systems"] = "Nonlinear Phenomena and Complex Systems",
  ["nonlinear phys sci"] = "Nonlinear Physical Science",
  ["nonlinear process geophys"] = "Nonlinear processes in geophysics",
  ["nonlinear processes geophys"] = "Nonlinear Processes in Geophysics",
  ["nonlinear processes geophys discuss"] = "Nonlinear Processes in Geophysics Discussions",
  ["nonlinear sci today"] = "Nonlinear Science Today",
  ["nonlinear stud"] = "Nonlinear Studies. The International Journal",
  ["nonlinear syst complex"] = "Nonlinear Systems and Complexity",
  ["nonlinear time ser chaos"] = "Nonlinear Time Series and Chaos",
  ["nonlinear world"] = "Nonlinear World",
  ["nonlinearity biol toxicol med"] = "Nonlinearity in biology, toxicology, medicine",
  ["nonpharmacol ther dement"] = "Non-pharmacological therapies in dementia",
  ["nonprofit manag leadersh"] = "Nonprofit management & leadership",
  ["nonprofit manage leadership"] = "Nonprofit Management and Leadership",
  ["nonprofit volunt sect q"] = "Nonprofit and voluntary sector quarterly",
  ["nor epidemiol"] = "Norsk epidemiologi = Norwegian journal of epidemiology",
  ["nor geogr tidsskr"] = "Norsk geografisk tidsskrift. Norwegian journal of geography",
  ["nor geol tidsskr"] = "Norsk Geologisk Tidsskrift",
  ["nor j entomol (1999)"] = "Norwegian journal of entomology (Oslo, Norway : 1999)",
  ["nor tannlaegeforen tid"] = "Den Norske tannlaegeforenings tidende",
  ["nor teol tidsskr"] = "Norsk theologisk tidsskrift",
  ["nord folkmedmus medd"] = "Nordiskt folkmedicinmuseum meddelanden",
  ["nord hyg tidskr"] = "Nordisk hygienisk tidskrift",
  ["nord hyg tidskr suppl"] = "Nordisk hygienisk tidskrift. Supplementum",
  ["nord j bot"] = "Nordic journal of botany",
  ["nord j freshw res"] = "Nordic journal of freshwater research",
  ["nord j hum rights"] = "Nordic journal of human rights",
  ["nord j migr res"] = "Nordic journal of migration research",
  ["nord j music ther"] = "Nordic journal of music therapy",
  ["nord j nurs res"] = "Nordic journal of nursing research",
  ["nord j psychiatry"] = "Nordic journal of psychiatry",
  ["nord j sci technol stud"] = "Nordic journal of science and technology studies",
  ["nord j sov east eur stud"] = "Nordic journal of Soviet and East European studies",
  ["nord med"] = "Nordisk medicin",
  ["nord medicinhist arsb"] = "Nordisk medicinhistorisk årsbok",
  ["nord medicinhist arsb suppl"] = "Nordisk medicinhistorisk årsbok. Supplementum",
  ["nord notes"] = "Nordic notes",
  ["nord psychol"] = "Nordic psychology",
  ["nord psyk medl"] = "Nordisk psykiatrisk medlemsblad",
  ["nord psykiatr tidsskr"] = "Nordisk psykiatrisk tidsskrift. Nordic journal of psychiatry",
  ["nord psykol"] = "Nordisk psykologi",
  ["nord pulp pap res j"] = "Nordic Pulp & Paper Research Journal",
  ["nord pulp paper res j"] = "Nordic pulp & paper research journal",
  ["nord sexol"] = "Nordisk sexologi",
  ["nord vet med"] = "Nordisk veterinaermedicin",
  ["norden news"] = "The Norden news",
  ["nordic j comput"] = "Nordic Journal of Computing",
  ["nordic j philos logic"] = "Nordic Journal of Philosophical Logic",
  ["nordic j polit economy"] = "Nordic Journal of Political Economy",
  ["nordic pulp pap res j"] = "Nordic Pulp and Paper Research Journal",
  ["nordic wittgenstein stud"] = "Nordic Wittgenstein Studies",
  ["nordisk alkohol nark"] = "Nordisk alkohol- & narkotikatidskrift : NAT",
  ["nordisk ost forum"] = "Nordisk øst-forum",
  ["nordost arch"] = "Nordost-Archiv",
  ["norm pathol anat (stuttg)"] = "Normale und pathologische Anatomie",
  ["norma int j masc stud"] = "Norma, international journal for masculinity studies",
  ["norman surg ser"] = "Norman surgery series",
  ["noro psikiyatr ars"] = "Nöro psikiyatri arşivi",
  ["norsk antropol tidsskr"] = "Norsk Antropologisk Tidsskrift",
  ["norsk geogr tidsskr - norw j geogr"] = "Norsk Geografisk Tidsskrift - Norwegian Journal of Geography",
  ["north afr middle east epilepsy j"] = "North African and Middle East epilepsy journal",
  ["north am dialogue"] = "North American dialogue : newsletter of the Society for the Anthropology of North America",
  ["north am j aquacult"] = "North American Journal of Aquaculture",
  ["north am j fish manage"] = "North American Journal of Fisheries Management",
  ["north am rev"] = "The North American review",
  ["north am vet"] = "The North American veterinarian",
  ["north carol centr law j"] = "North Carolina Central law journal",
  ["north carol law rev"] = "North Carolina law review",
  ["north clin istanb"] = "Northern clinics of Istanbul",
  ["north east nat"] = "North-eastern naturalist",
  ["north east rep second ser"] = "North eastern reporter. Second series",
  ["north hist"] = "Northern history",
  ["north ill univ law rev"] = "Northern Illinois University law review",
  ["north irel law rep"] = "The Northern Ireland law reports",
  ["north irel leg q"] = "The Northern Ireland legal quarterly",
  ["north j appl for"] = "Northern Journal of Applied Forestry",
  ["north ky law rev"] = "Northern Kentucky law review",
  ["north ky state law forum"] = "Northern Kentucky State law forum",
  ["north mar"] = "The Northern mariner : journal of the Canadian Nautical Research Society = Le marin du nord : revue de Société canadienne pour la recherche nautique",
  ["north n y med j"] = "Northern New York medical journal",
  ["north scotl"] = "Northern Scotland",
  ["north territ nat"] = "Northern Territory naturalist",
  ["north va herit"] = "Northern Virginia heritage",
  ["north west j zool"] = "North-western journal of zoology",
  ["north west rep second ser"] = "North western reporter. Second series",
  ["north wing"] = "The north wing; the magazine of the Sheffield medical and dental schools",
  ["north-holland math library"] = "North-Holland Mathematical Library",
  ["north-holland math stud"] = "North-Holland Mathematics Studies",
  ["north-holland ser appl math mech"] = "North-Holland Series in Applied Mathematics and Mechanics",
  ["north-west eur j math"] = "North-Western European Journal of Mathematics",
  ["northeast afr stud"] = "Northeast African Studies",
  ["northeast folk"] = "Northeast folklore",
  ["northeast hist archaeology"] = "Northeast historical archaeology",
  ["northeast j contemp res"] = "Northeast journal of contemporary research",
  ["northeast math j"] = "Northeastern Mathematical Journal",
  ["northeast nat"] = "Northeastern Naturalist",
  ["northeast nat (steuben)"] = "Northeastern naturalist",
  ["northeast nev hist soc q"] = "Quarterly - Northeastern Nevada Historical Society. Northeastern Nevada Historical Society",
  ["northwest dent"] = "Northwest dentistry",
  ["northwest dent res"] = "Northwestern dental research",
  ["northwest j law soc policy"] = "Northwestern journal of law and social policy",
  ["northwest med"] = "Northwest medicine",
  ["northwest nat"] = "Northwestern Naturalist",
  ["northwest nat (olymp wash)"] = "Northwestern naturalist : a journal of vertebrate biology",
  ["northwest ohio q"] = "Northwest Ohio quarterly",
  ["northwest public health"] = "Northwest public health",
  ["northwest rep"] = "Northwest report : a newsletter of the Northwest Area Foundation",
  ["northwest sci"] = "Northwest Science",
  ["northwest univ bull dent res grad study"] = "Northwestern University Bulletin, Dental Research and Graduate Study",
  ["northwest univ law rev"] = "Northwestern University law review",
  ["norw archaeol rev"] = "Norwegian Archaeological Review",
  ["norw j entomol"] = "Norwegian Journal Of Entomology",
  ["norw j geol"] = "Norwegian Journal of Geology",
  ["nos oiseaux"] = "Nos Oiseaux",
  ["noshuyo byori"] = "Nōshuyō byōri = Brain tumor pathology",
  ["nosokom chron"] = "Nosokomeiaka chronika",
  ["nostro paese"] = "nostro paese, Il",
  ["not algarum"] = "Notulae algarum",
  ["not am math soc"] = "Notices of the American Mathematical Society. American Mathematical Society",
  ["not ammin sanit (1948)"] = "Notiziario dell'amministrazione sanitaria. Italy. Alto commissariato per l'igiene e la sanità pubblica",
  ["not bot horti agrobot cluj napoca"] = "Notulae botanicae Horti Agrobotanici Cluj-Napoca",
  ["not clin"] = "Noticias clínicas",
  ["not farm"] = "Notiziario farmaceutico",
  ["not isms"] = "Notices from the ISMS",
  ["not man apart"] = "Not Man Apart",
  ["not med esp"] = "Noticiero médico español",
  ["not nat acad nat sci philadelphia"] = "Notulae naturae of the Academy of Natural Sciences of Philadelphia",
  ["not odonatologicae"] = "Notulae Odonatologicae",
  ["not polit"] = "Notizie di Politeia",
  ["not prat"] = "Notice pour le praticien",
  ["not s afr math soc"] = "Notices of the South African Mathematical Society. Mededelings van die Suid-Afrikaanse Wiskundevereniging",
  ["not sci biol"] = "Notulae scientia biologicae",
  ["nota lepidopterol"] = "Nota lepidopterologica",
  ["notas algebra anal"] = "Notas de Algebra y Analisis",
  ["notas comun mat"] = "Notas e Comunicacoes de Matematica",
  ["notas inst mat estatist univ sao paulo ser mat"] = "Notas do Instituto de Matematica e Estatistica da Universidade de Sao Paulo, Serie Matematica",
  ["notas lógica mat"] = "Notas de Lógica Matemática",
  ["notas mat apl"] = "Notas em Matematica Aplicada",
  ["notas mat discreta"] = "Notas de Matemática Discreta",
  ["notas poblacion"] = "Notas de población",
  ["notas quir sanat deschamps"] = "Notas quirúrgicas del Sanatorio Deschamps",
  ["notas salud publica"] = "Notas de salud pública",
  ["notas soc mat chile"] = "Notas de la Sociedad de Matemática de Chile",
  ["notatki plockie"] = "Notatki płockie",
  ["note mat"] = "Note di Matematica",
  ["note riv psichiatr"] = "Note e riviste di psichiatria",
  ["note tech group tech for"] = "Note technique Groupement technique forestier",
  ["notes and rec"] = "Notes and Records. The Royal Society Journal of the History of Science",
  ["notes and records roy soc london"] = "Notes and Records of the Royal Society of London",
  ["notes czechoslov med assoc g b"] = "Notes. Czechoslovak Medical Association in Great Britain",
  ["notes ifs"] = "Notes on Intuitionistic Fuzzy Sets",
  ["notes number theory discret math"] = "Notes on number theory and discrete mathematics",
  ["notes number theory discrete math"] = "Notes on Number Theory and Discrete Mathematics",
  ["notes numer fluid mech"] = "Notes on Numerical Fluid Mechanics",
  ["notes numer fluid mech multidiscip des"] = "Notes on Numerical Fluid Mechanics and Multidisciplinary Design",
  ["notes queries"] = "Notes and queries",
  ["notes rec"] = "Notes and Records",
  ["notes rec r soc lond"] = "Notes and records of the Royal Society of London",
  ["notes tech cent écol for rural"] = "Notes techniques du Centre d’écologie forestière et rurale",
  ["notes undergr"] = "Notes from the underground (New York, N.Y.)",
  ["notf rett med"] = "Notfall & rettungsmedizin",
  ["notices amer math soc"] = "Notices of the American Mathematical Society",
  ["notre dame j form log"] = "Notre Dame Journal of Formal Logic",
  ["notre dame j formal logic"] = "Notre Dame Journal of Formal Logic",
  ["notre dame j law ethics public policy"] = "Notre Dame journal of law, ethics & public policy",
  ["notre dame law rev"] = "The Notre Dame law review",
  ["notre dame lawyer"] = "The Notre Dame lawyer",
  ["notre dame math lect"] = "Notre Dame Mathematical Lectures",
  ["notulae odonatologicae"] = "Notulae odonatologicae",
  ["notulae sci biol"] = "Notulae Scientia Biologicae",
  ["nouson kougaku kenkyuujo gihou"] = "Nouson kougaku kenkyuujo gihou = Technical report of the National Institute for Rural Engineering",
  ["nouv bibl math"] = "Nouvelle Bibliotheque Mathematique",
  ["nouv com int cathol infirm assist med soc"] = "Nouvelles - Comité international catholique des infirmières et assistantes médico-sociales. News - International Committee of Catholic Nurses and Medico-Social Workers",
  ["nouv ec"] = "Nouvelle école",
  ["nouv j chim"] = "Nouveau journal de chimie",
  ["nouv landeshydrol -geol"] = "Nouvelles der Landeshydrologie und -geologie",
  ["nouv prat vet canine feline"] = "Nouveau Praticien Veterinaire (Le) canine & feline",
  ["nouv prat vet elevages sante"] = "Nouveau Praticien Veterinaire (Le) elevages & sante",
  ["nouv prat vet equine"] = "Nouveau Praticien Veterinaire (Le) equine",
  ["nouv presse med"] = "La Nouvelle presse médicale",
  ["nouv rendez-vous archimede"] = "Les Nouveaux Rendez-vous d’Archimede",
  ["nouv repub lett"] = "Nouvelles de la république des lettres (Naples, Italy)",
  ["nouv rev 16e siecle"] = "Nouvelle revue du XVIe siècle",
  ["nouv rev deux mondes"] = "La Nouvelle revue des deux mondes",
  ["nouv rev fr hematol"] = "Nouvelle revue française d'hématologie",
  ["nouv rev fr hematol blood cells"] = "Nouvelle revue française d'hématologie; blood cells",
  ["nouv rev opt"] = "Nouvelle Revue d’Optique",
  ["nouv rev opt appl"] = "Nouvelle Revue d’Optique AppliquTe",
  ["nouv vis sci"] = "Nouvelles Visions des Sciences",
  ["nov approaches cancer study"] = "Novel approaches in cancer study",
  ["nov khir"] = "Novaia khirurgiia",
  ["nov med priborostr"] = "Novosti meditsinskogo priborostroeniia",
  ["nov med tekh"] = "Novosti meditsinskoĭ tekhniki",
  ["nov novejsaja istor"] = "Novai︠a︡ i noveĭshai︠a︡ istorii︠a︡",
  ["nov sist nizs rast"] = "Novosti sistematiki nizshikh rasteniĭ",
  ["nova acta cient compostel biol"] = "Nova acta científica compostelana. Bioloxía : NACC. Bioloxía",
  ["nova acta leopold"] = "Nova acta Leopoldina : Abhandlungen der Kaiserlich Leopoldinisch-Carolinisch Deutschen Akademie der Naturforscher",
  ["nova acta leopoldina (nf)"] = "Deutsche Akademie der Naturforscher Leopoldina",
  ["nova acta stomatol"] = "Nova acta stomatologica",
  ["nova biotechnol chim"] = "Nova Biotechnologica et Chimica",
  ["nova econ"] = "Nova economia (Belo Horizonte, Brazil)",
  ["nova hedwig beih"] = "Nova Hedwigia. Beiheft",
  ["nova hedwigia"] = "Nova Hedwigia",
  ["nova j math game theory algebra"] = "Nova Journal of Mathematics, Game Theory and Algebra",
  ["nova j theor phys"] = "Nova Journal of Theoretical Physics",
  ["nova law j"] = "Nova law journal",
  ["nova law rev"] = "Nova law review",
  ["novartis found symp"] = "Novartis Foundation symposium",
  ["noved med"] = "Novedades medicas",
  ["novenyvedelem (bp 1965)"] = "Növényvédelem",
  ["novi sad j math"] = "Novi Sad Journal of Mathematics",
  ["novice varst gozdov"] = "Novice iz varstva gozdov",
  ["novit bot univ carol"] = "Novitates botanicae Universitatis Carolinae",
  ["novit caribaea"] = "Novitates Caribaea",
  ["novoe global anal"] = "\\cyr Novoe v Global\\cprime nom Analize",
  ["novon (st louis)"] = "Novon : a journal for botanical nomenclature",
  ["novon: j bot nomencl"] = "Novon: A Journal for Botanical Nomenclature",
  ["novyi khirurgicheskii arkhiv"] = "Novyĭ khirurgicheskiĭ arkhiv",
  ["now lekarskie organ wydz lekarskiego tow przyj nauk posnanskiego"] = "Nowiny lekarskie : organ Wydziału Lekarskiego Towarzystwa Przyjaciół Nauk Posnańskiego",
  ["nowe drogi"] = "Nowe drogi",
  ["np news"] = "NP news (Bellevue, Wash.)",
  ["npg asia mater"] = "NPG Asia Materials",
  ["npg forum ser"] = "NPG forum series",
  ["npj 2d mater appl"] = "NPJ 2D materials and applications",
  ["npj aging mech dis"] = "NPJ aging and mechanisms of disease",
  ["npj biodivers"] = "npj Biodiversity",
  ["npj biofilms microbiomes"] = "NPJ biofilms and microbiomes",
  ["npj breast cancer"] = "NPJ breast cancer",
  ["npj clean water"] = "NPJ clean water",
  ["npj clim atmos sci"] = "NPJ climate and atmospheric science",
  ["npj comput mater"] = "npj Computational Materials",
  ["npj digit med"] = "NPJ digital medicine",
  ["npj digital med"] = "npj Digital Medicine",
  ["npj flexible electron"] = "npj Flexible Electronics",
  ["npj genom med"] = "NPJ genomic medicine",
  ["npj genomic med"] = "npj Genomic Medicine",
  ["npj mater degrad"] = "Npj Materials degradation",
  ["npj microgravity"] = "NPJ microgravity",
  ["npj mol phenomics"] = "npj Molecular Phenomics",
  ["npj ocean sustainability"] = "npj Ocean Sustainability",
  ["npj parkinsons dis"] = "NPJ Parkinson's disease",
  ["npj parkinson’s dis"] = "npj Parkinson’s Disease",
  ["npj pollut control"] = "npj Pollution Control",
  ["npj precis oncol"] = "NPJ precision oncology",
  ["npj prim care respir med"] = "NPJ primary care respiratory medicine",
  ["npj primary care respir med"] = "npj Primary Care Respiratory Medicine",
  ["npj quantum inf"] = "npj Quantum Information",
  ["npj quantum mater"] = "npj quantum materials",
  ["npj regen med"] = "NPJ Regenerative medicine",
  ["npj regener med"] = "npj Regenerative Medicine",
  ["npj rob"] = "npj Robotics",
  ["npj schizophr"] = "NPJ schizophrenia",
  ["npj sci food"] = "NPJ science of food",
  ["npj sci learn"] = "NPJ science of learning",
  ["npj syst biol appl"] = "NPJ systems biology and applications",
  ["npj vaccines"] = "NPJ vaccines",
  ["npn med"] = "NPN médecine",
  ["nriag j astron geophys"] = "NRIAG Journal of Astronomy and Geophysics",
  ["nrl rev"] = "NRL review (1988)",
  ["ns hist q"] = "The Nova Scotia historical quarterly",
  ["nsf-cbms regional conf ser probab statist"] = "NSF-CBMS Regional Conference Series in Probability and Statistics",
  ["nt learn curve"] = "NT learning curve",
  ["ntm j hist sci technol med"] = "NTM Journal of the History of Science, Technology and Medicine",
  ["ntm z gesch wissen tech med"] = "NTM Zeitschrift fur Geschichte der Wissenschaften, Technik und Medizin",
  ["ntp cerhr mon"] = "NTP CERHR MON",
  ["ntp monogr"] = "NTP monograph",
  ["ntq new theatre q"] = "New theatre quarterly : NTQ",
  ["nucl anal"] = "Nuclear Analysis",
  ["nucl chem waste manage"] = "Nuclear and Chemical Waste Management",
  ["nucl data sect a"] = "Nuclear Data Section A",
  ["nucl data sect b"] = "Nuclear Data Section B",
  ["nucl data sheets"] = "Nuclear Data Sheets",
  ["nucl data, sec a"] = "Nuclear Data, Section A",
  ["nucl data, sec b"] = "Nuclear Data, Section B",
  ["nucl energy technol"] = "Nuclear Energy and Technology",
  ["nucl eng des"] = "Nuclear Engineering and Design",
  ["nucl eng des/fusion"] = "Nuclear engineering and design/fusion : an international journal devoted to the thermal, mechanical, materials, structural, and design problems of fusion energy",
  ["nucl eng technol"] = "Nuclear Engineering and Technology",
  ["nucl fusion"] = "Nuclear Fusion",
  ["nucl geophys"] = "Nuclear Geophysics",
  ["nucl instrum"] = "Nuclear Instruments",
  ["nucl instrum methods"] = "Nuclear Instruments and Methods",
  ["nucl instrum methods phys res"] = "Nuclear Instruments and Methods in Physics Research",
  ["nucl instrum methods phys res a"] = "Nuclear instruments & methods in physics research. Section A, Accelerators, spectrometers, detectors and associated equipment",
  ["nucl instrum methods phys res b"] = "Nuclear instruments & methods in physics research. Section B, Beam interactions with materials and atoms",
  ["nucl instrum methods phys res sect b"] = "Nuclear Instruments and Methods in Physics Research Section B",
  ["nucl instrum methods phys res sect b beam interact mater at"] = "Nuclear Instruments and Methods in Physics Research Section B: Beam Interactions with Materials and Atoms",
  ["nucl instrum methods phys res, sect a"] = "Nuclear Instruments and Methods in Physics Research Section A: Accelerators, Spectrometers, Detectors and Associated Equipment",
  ["nucl instrum methods phys res, sect b"] = "Nuclear Instruments and Methods in Physics Research Section B: Beam Interactions with Materials and Atoms",
  ["nucl mater energy"] = "Nuclear Materials and Energy",
  ["nucl med (stuttg)"] = "Nuclear-Medizin",
  ["nucl med biol"] = "Nuclear Medicine and Biology",
  ["nucl med biomed imaging"] = "Nuclear medicine and biomedical imaging",
  ["nucl med commun"] = "Nuclear medicine communications",
  ["nucl med mol imaging"] = "Nuclear medicine and molecular imaging",
  ["nucl med rev cent east eur"] = "Nuclear medicine review. Central & Eastern Europe",
  ["nucl part phys proc"] = "Nuclear and Particle Physics Proceedings",
  ["nucl phys"] = "Nuclear Physics",
  ["nucl phys a"] = "Nuclear Physics A",
  ["nucl phys at energy"] = "Nuclear Physics And Atomic Energy",
  ["nucl phys b"] = "Nuclear Physics B",
  ["nucl phys b proc suppl"] = "Nuclear Physics B - Proceedings Supplements",
  ["nucl phys b, proc suppl"] = "Nuclear Physics B, Proceedings Supplements",
  ["nucl phys news"] = "Nuclear Physics News",
  ["nucl recept"] = "Nuclear receptor",
  ["nucl recept signal"] = "Nuclear receptor signaling",
  ["nucl recept signaling"] = "Nuclear Receptor Signaling",
  ["nucl receptor res"] = "Nuclear receptor research",
  ["nucl sci abstr"] = "Nuclear science abstracts",
  ["nucl sci eng"] = "Nuclear science and engineering : the journal of the American Nuclear Society",
  ["nucl sci tech"] = "Nuclear Science and Techniques",
  ["nucl struct eng"] = "Nuclear Structural Engineering",
  ["nucl technol"] = "Nuclear technology",
  ["nucl track detect"] = "Nuclear Track Detection",
  ["nucl tracks"] = "Nuclear Tracks",
  ["nucl tracks radiat meas"] = "Nuclear Tracks And Radiation Measurements",
  ["nuclear part phys proc"] = "Nuclear and Particle Physics Proceedings",
  ["nuclear phys b"] = "Nuclear Physics. B. Theoretical, Phenomenological, and Experimental High Energy Physics. Quantum Field Theory and Statistical Systems",
  ["nuclear phys b proc suppl"] = "Nuclear Physics B",
  ["nucleic acid ther"] = "Nucleic acid therapeutics",
  ["nucleic acids mol biol"] = "Nucleic acids and molecular biology",
  ["nucleic acids res"] = "Nucleic acids research",
  ["nucleic acids res spec publ"] = "Nucleic acids research. Special publication",
  ["nucleic acids res suppl"] = "Nucleic acids research. Supplement (2001)",
  ["nucleic acids symp ser"] = "Nucleic Acids Symposium Series",
  ["nucleic acids symp ser (oxf)"] = "Nucleic acids symposium series (2004)",
  ["nucleosides nucleotides"] = "Nucleosides & nucleotides",
  ["nucleosides nucleotides nucleic acids"] = "Nucleosides, nucleotides & nucleic acids",
  ["nucleosides, nucleotides nucleic acids"] = "Nucleosides, Nucleotides & Nucleic Acids",
  ["nucleus (calcutta)"] = "The Nucleus : an international journal of cytology and allied topics",
  ["nucleus (camb)"] = "Nucleus (Cambridge, Mass.)",
  ["nuestra hist"] = "Nuestra historia",
  ["nueva enferm"] = "Nueva enfermeria",
  ["nuevas ideas rev med"] = "Nuevas ideas; revista de medicina",
  ["nuevos libros"] = "Nuevos libros",
  ["nufusbil derg"] = "Nüfusbilim dergisi",
  ["num cient accion sind"] = "Acción sindical. Número científico",
  ["numer algebra control optim"] = "Numerical Algebra, Control and Optimization",
  ["numer algebra, control optim"] = "Numerical Algebra, Control and Optimization",
  ["numer algorithms"] = "Numerical Algorithms",
  ["numer anal appl"] = "Numerical Analysis and Applications",
  ["numer funct anal optim"] = "Numerical Functional Analysis and Optimization",
  ["numer heat transfer, part a"] = "Numerical Heat Transfer Part A",
  ["numer heat transfer, part b"] = "Numerical Heat Transfer Part B",
  ["numer insights"] = "Numerical Insights",
  ["numer linear algebra appl"] = "Numerical Linear Algebra with Applications",
  ["numer math"] = "Numerische Mathematik",
  ["numer math (heidelb)"] = "Numerische mathematik",
  ["numer math j chinese univ"] = "Numerical Mathematics. A Journal of Chinese Universities. Gaodeng Xuexiao Jisuan Shuxue Xuebao",
  ["numer math j chinese univ (english ser)"] = "Numerical Mathematics",
  ["numer math sci comput"] = "Numerical Mathematics and Scientific Computation",
  ["numer math theory methods appl"] = "Numerical Mathematics Theory Methods and Applications",
  ["numer methods eng ser"] = "Numerical Methods in Engineering Series",
  ["numer methods partial differ equ"] = "Numerical methods for partial differential equations",
  ["numer methods partial differ equations"] = "Numerical Methods for Partial Differential Equations",
  ["numer methods partial differential equations"] = "Numerical Methods for Partial Differential Equations. An International Journal",
  ["numeri heat transf a appl"] = "Numerical heat transfer. Part A, Applications",
  ["numerical algorithms"] = "Numerical Algorithms",
  ["nuncius ann storia sci"] = "Istituto e Museo di Storia della Scienza Firenze",
  ["nunt radiol"] = "Nuntius radiologicus",
  ["nuova antologia"] = "Nuova antologia",
  ["nuova riv neurol"] = "Nuova rivista di neurologia",
  ["nuova riv stor"] = "Nuova rivista storica",
  ["nuova vet"] = "La Nuova veterinaria",
  ["nuovi ann ig microbiol"] = "Nuovi annali d'igiene e microbiologia",
  ["nuovo cimento"] = "Nuovo Cimento",
  ["nuovo cimento a"] = "Nuovo Cimento A",
  ["nuovo cimento a (11)"] = "Società Italiana di Fisica",
  ["nuovo cimento b"] = "Nuovo Cimento B",
  ["nuovo cimento b (11)"] = "Società Italiana di Fisica",
  ["nuovo cimento c"] = "Il Nuovo cimento della Società italiana di fisica. C",
  ["nuovo cimento d"] = "Nuovo Cimento D",
  ["nuovo cimento soc ital fis a (12)"] = "Il Nuovo Cimento della Società Italiana di Fisica",
  ["nuovo cimento soc ital fis b (12)"] = "Il Nuovo Cimento della Società Italiana di Fisica",
  ["nuovo cimento soc ital fis, a"] = "Nuovo Cimento della Societa Italiana di Fisica A",
  ["nuovo cimento soc ital fis, b"] = "Nuovo Cimento della Societa Italiana di Fisica B",
  ["nuovo cimento soc ital fis, c"] = "Nuovo Cimento della Societa Italiana di Fisica C",
  ["nuovo cimento soc ital fis, d"] = "Nuovo Cimento della Societa Italiana di Fisica D",
  ["nuovo g bot ital"] = "Nuovo giornale botanico italiano",
  ["nur al-mantiq wal-ma\\rasp rifah"] = "Nur al-Mantiq wal-Ma\\rasp rifah",
  ["nurs adm q"] = "Nursing administration quarterly",
  ["nurs bc"] = "Nursing BC",
  ["nurs care"] = "Nursing care",
  ["nurs careers"] = "Nursing careers",
  ["nurs case manag"] = "Nursing case management : managing the process of patient care",
  ["nurs child young people"] = "Nursing children and young people",
  ["nurs clin north am"] = "The Nursing clinics of North America",
  ["nurs crit care"] = "Nursing in critical care",
  ["nurs diagn"] = "Nursing diagnosis : ND : the official journal of the North American Nursing Diagnosis Association",
  ["nurs dig"] = "Nursing digest",
  ["nurs dyn"] = "Nursing dynamics",
  ["nurs econ"] = "Nursing economic$",
  ["nurs educ microworld"] = "Nursing educators microworld",
  ["nurs educ monogr"] = "Nursing education monographs",
  ["nurs educ perspect"] = "Nursing education perspectives",
  ["nurs elder"] = "Nursing the elderly : in hospital, homes and the community",
  ["nurs ethics"] = "Nursing ethics",
  ["nurs focus"] = "Nursing focus",
  ["nurs forum"] = "Nursing forum",
  ["nurs forum (auckl)"] = "Nursing forum",
  ["nurs health (alhambra)"] = "Nursing and health (Alhambra, Calif.)",
  ["nurs health care"] = "Nursing & health care : official publication of the National League for Nursing",
  ["nurs health care (winfield)"] = "Nursing and health care (Winfield, Ill.)",
  ["nurs health care perspect"] = "Nursing and health care perspectives",
  ["nurs health sci"] = "Nursing & health sciences",
  ["nurs hist rev"] = "Nursing history review : official journal of the American Association for the History of Nursing",
  ["nurs homes"] = "Nursing homes",
  ["nurs homes sr citiz care"] = "Nursing homes and senior citizen care",
  ["nurs inq"] = "Nursing inquiry",
  ["nurs j (manila)"] = "Nursing journal",
  ["nurs j india"] = "The Nursing journal of India",
  ["nurs j singapore"] = "The Nursing journal of Singapore",
  ["nurs law ethics"] = "Nursing law & ethics",
  ["nurs law regan rep"] = "Nursing law's Regan report",
  ["nurs leadersh"] = "Nursing leadership",
  ["nurs leadersh (tor ont)"] = "Nursing leadership (Toronto, Ont.)",
  ["nurs leadersh forum"] = "Nursing leadership forum",
  ["nurs life"] = "NursingLife",
  ["nurs manag (harrow)"] = "Nursing management (Harrow, London, England : 1994)",
  ["nurs manage"] = "Nursing management",
  ["nurs midwifery stud"] = "Nursing and midwifery studies",
  ["nurs mirror"] = "Nursing mirror",
  ["nurs mirror midwives j"] = "Nursing mirror and midwives journal",
  ["nurs montreal"] = "Nursing Montreal",
  ["nurs n z"] = "Nursing New Zealand",
  ["nurs news"] = "Nursing news",
  ["nurs news (meriden)"] = "Nursing News",
  ["nurs older people"] = "Nursing older people",
  ["nurs open"] = "Nursing open",
  ["nurs outlook"] = "Nursing outlook",
  ["nurs palliat care"] = "Nursing and palliative care",
  ["nurs pap"] = "Nursing papers. Perspectives en nursing",
  ["nurs philos"] = "Nursing philosophy : an international journal for healthcare professionals",
  ["nurs pract"] = "Nursing practice (Edinburgh, Scotland)",
  ["nurs prax n z"] = "Nursing praxis in New Zealand inc",
  ["nurs qual connect"] = "Nursing quality connection",
  ["nurs que"] = "Nursing Quebec",
  ["nurs rep"] = "Nursing reports",
  ["nurs res"] = "Nursing research",
  ["nurs res conf"] = "Nursing Research Conference",
  ["nurs res pract"] = "Nursing research and practice",
  ["nurs res rep"] = "Nursing research report",
  ["nurs rev"] = "Nursing review (North Sydney, N.S.W.)",
  ["nurs rsa"] = "Nursing RSA = Verpleging RSA",
  ["nurs sci"] = "Nursing science",
  ["nurs sci q"] = "Nursing science quarterly",
  ["nurs spectr (fla ed)"] = "Nursing spectrum (Florida ed.)",
  ["nurs spectr (gt chic ne ill nw indiana ed)"] = "Nursing spectrum (Illinois ed.)",
  ["nurs spectr (gt phila tri state ed)"] = "Nursing spectrum (Greater Philadelphia/Tri-state ed.)",
  ["nurs spectr (n engl ed)"] = "Nursing spectrum (New England ed.)",
  ["nurs spectr (wash d c)"] = "Nursing spectrum (D.C./Baltimore metro ed.)",
  ["nurs staff dev insid"] = "Nursing staff development insider",
  ["nurs stand"] = "Nursing standard (Royal College of Nursing (Great Britain) : 1987)",
  ["nurs stand spec suppl"] = "Nursing standard (Royal College of Nursing (Great Britain) : 1987). Special supplement",
  ["nurs success today"] = "Nursing success today",
  ["nurs times"] = "Nursing times",
  ["nurs times nurs homes"] = "Nursing times. Nursing homes",
  ["nurs womens health"] = "Nursing for women's health",
  ["nurs world"] = "Nursing world",
  ["nurse anesth"] = "Nurse anesthesia",
  ["nurse author ed"] = "Nurse author & editor",
  ["nurse educ"] = "Nurse educator",
  ["nurse educ oppor innov"] = "Nurse Educators Opportunities And Innovations",
  ["nurse educ pract"] = "Nurse education in practice",
  ["nurse educ today"] = "Nurse education today",
  ["nurse lead"] = "Nurse leader",
  ["nurse managers bookshelf"] = "Nurse managers' bookshelf",
  ["nurse pract"] = "The Nurse practitioner",
  ["nurse pract forum"] = "Nurse practitioner forum",
  ["nurse res"] = "Nurse researcher",
  ["nurses lamp"] = "Nurses Lamp",
  ["nursing (auckl)"] = "Nursing (Auckland, N.Z.)",
  ["nursing (brux)"] = "Nursing",
  ["nursing (lond)"] = "Nursing",
  ["nutr abstr rev"] = "Nutrition abstracts and reviews",
  ["nutr abstr rev ser hum exp"] = "Nutrition abstracts and reviews. Series A: Human and experimental",
  ["nutr aging"] = "Nutrition and Aging",
  ["nutr aging (amst)"] = "Nutrition and aging (Amsterdam, Netherlands)",
  ["nutr bull"] = "Nutrition bulletin",
  ["nutr cancer"] = "Nutrition and cancer",
  ["nutr clin care"] = "Nutrition in clinical care : an official publication of Tufts University",
  ["nutr clin pract"] = "Nutrition in clinical practice : official publication of the American Society for Parenteral and Enteral Nutrition",
  ["nutr cycl agroecosyst"] = "Nutrient cycling in agroecosystems",
  ["nutr cycling agroecosyst"] = "Nutrient Cycling in Agroecosystems",
  ["nutr diabetes"] = "Nutrition & diabetes",
  ["nutr diet"] = "Nutrition & dietetics: the journal of the Dietitians Association of Australia",
  ["nutr diet suppl"] = "Nutrition and dietary supplements",
  ["nutr dieta eur rev nutr diet"] = "Nutritio et dieta; European review of nutrition and dietetics",
  ["nutr dietary suppl"] = "Nutrition and Dietary Supplements",
  ["nutr food sci"] = "Nutrition and food science",
  ["nutr health"] = "Nutrition and health",
  ["nutr healthy aging"] = "Nutrition and healthy aging",
  ["nutr healthy living"] = "Nutrition for healthy living",
  ["nutr hist notes"] = "Nutrition history notes",
  ["nutr hosp"] = "Nutrición hospitalaria",
  ["nutr j"] = "Nutrition journal",
  ["nutr life"] = "Nutrition for life",
  ["nutr metab"] = "Nutrition and metabolism",
  ["nutr metab (lond)"] = "Nutrition & metabolism",
  ["nutr metab cardiovasc dis"] = "Nutrition, metabolism, and cardiovascular diseases : NMCD",
  ["nutr metab insights"] = "Nutrition and metabolic insights",
  ["nutr neurosci"] = "Nutritional neuroscience",
  ["nutr news (rosemt)"] = "Nutrition news",
  ["nutr rep int"] = "Nutrition reports international",
  ["nutr res"] = "Nutrition research (New York, N.Y.)",
  ["nutr res (n y, ny, u s)"] = "Nutrition Research (New York, NY, United States)",
  ["nutr res pract"] = "Nutrition research and practice",
  ["nutr res rev"] = "Nutrition research reviews",
  ["nutr rev"] = "Nutrition reviews",
  ["nutr rev mother child"] = "Nutrition review. Mother and child",
  ["nutr today"] = "Nutrition today",
  ["nutr update"] = "Nutrition update",
  ["nutr value major nutr irradiat food apprais toxic irradiat food"] = "On the nutritive value of the major nutrients of irradiated foods and appraisal of the toxicity of irradiated foods; progress report. University of Illinois (Urbana-Champaign campus). Division of Animal Nutrition",
  ["nutzholzverbrauch schweiz"] = "Nutzholzverbrauch in der Schweiz",
  ["nwb-studienbuch wirtschaftswissenschaft"] = "NWB-Studienbucher Wirtschaftswissenschaften",
  ["nwsa j"] = "NWSA journal : a publication of the National Women's Studies Association",
  ["ny dent j"] = "The New York dental journal",
  ["ny folkl q"] = "New York folklore quarterly",
  ["ny hist soc q"] = "New York Historical Society quarterly. New-York Historical Society",
  ["ny law sch law rev"] = "New York Law School law review. New York Law School",
  ["ny med"] = "New York medicine",
  ["ny times"] = "New York Times",
  ["ny times (print)"] = "The New York times",
  ["nyjm monogr"] = "New York Journal of Mathematics. NYJM Monographs",
  ["nyngjon hag hoi ji"] = "Han'guk Yukchong Hakhoe chi",
  ["nyo rep"] = "NYO [reports]. U.S. Atomic Energy Commission",
  ["nyssnta j"] = "NYSSNTA journal",
  ["nz eng"] = "New Zealand Engineering",
  ["nz entomol"] = "New Zealand Entomologist",
  ["nz environ"] = "New Zealand Environment",
  ["nz fert j"] = "New Zealand Fertilizer Journal",
  ["nz geol surv bull"] = "New Zealand Geological Survey Bulletin",
  ["nz j agric"] = "New Zealand Journal of Agriculture",
  ["nz j agric res"] = "New Zealand Journal of Agricultural Research",
  ["nz j dairy sci technol"] = "New Zealand Journal of Dairy Science and Technology",
  ["nz j for"] = "New Zealand Journal of Forestry",
  ["nz j for sci"] = "New Zealand Journal of Forest Science",
  ["nz j geol geophys"] = "New Zealand Journal of Geology and Geophysics",
  ["nz j mar freshwater res"] = "New Zealand Journal of Marine and Freshwater Research",
  ["nz j psychol"] = "New Zealand journal of psychology",
  ["nz j sci"] = "New Zealand Journal of Science",
  ["nz j sci technol sect: a"] = "New Zealand Journal of Science and Technology Section A",
  ["nz j sci technol sect: b"] = "New Zealand Journal of Science and Technology Section B",
  ["nz j technol"] = "New Zealand Journal of Technology",
  ["nz j zool"] = "New Zealand Journal of Zoology",
  ["nz psychol"] = "The New Zealand psychologist",
  ["oa alcohol"] = "OA alcohol",
  ["oa autism"] = "OA autism",
  ["oa fam med"] = "OA family medicine",
  ["oa genet"] = "OA genetics",
  ["oa j pregnancy child care"] = "OA journal of pregnancy and child care",
  ["oa microbiol"] = "OA microbiology",
  ["oa minim invasive surg"] = "OA minimally invasive surgery",
  ["oa mol oncol"] = "OA molecular oncology",
  ["oa womens health"] = "OA women's health",
  ["oalma j"] = "The OALMA journal",
  ["oasics openaccess ser inform"] = "OASIcs: OpenAccess Series in Informatics",
  ["ob gyn dig"] = "Ob/gyn digest",
  ["ob gyn news"] = "Ob. gyn. news",
  ["ob/gyn world"] = "Ob/gyn World",
  ["obe metab"] = "Obesity and metabolism",
  ["obere extrem"] = "Obere extremität",
  ["oberwolfach rep"] = "Oberwolfach Reports",
  ["oberwolfach semin"] = "Oberwolfach Seminars",
  ["obes control ther"] = "Obesity & control therapies : open access",
  ["obes endocrinol"] = "Obesity and Endocrinology",
  ["obes facts"] = "Obesity facts",
  ["obes manag"] = "Obesity management",
  ["obes med"] = "Obesity medicine",
  ["obes open access"] = "Obesity, open access",
  ["obes res"] = "Obesity research",
  ["obes res clin pract"] = "Obesity research & clinical practice",
  ["obes rev"] = "Obesity reviews : an official journal of the International Association for the Study of Obesity",
  ["obes sci pract"] = "Obesity science & practice",
  ["obes surg"] = "Obesity surgery",
  ["obes weight  manag"] = "Obesity and weight management",
  ["obesity (silver spring)"] = "Obesity (Silver Spring, Md.)",
  ["objets mondes"] = "Objets et mondes : la revue du Musée de l'homme, Muséum national d'histoire naturelle",
  ["obm genet"] = "OBM genetics",
  ["obm geriat"] = "OBM geriatrics",
  ["obm integr compliment med"] = "OBM integrative and complimentary medicine",
  ["obm neurobiol"] = "OBM neurobiology",
  ["obm transplant"] = "OBM transplantation",
  ["obradoiro hist mod"] = "Obradoiro de historia moderna",
  ["obs diagn econ"] = "Observations et diagnostics économiques. Revue de l'OFCE",
  ["obs stud"] = "Observational studies",
  ["observer (lond)"] = "Observer (London, England)",
  ["obstet ginecol (bucur)"] = "Obstetrică şi ginecologia",
  ["obstet ginecol lat am"] = "Obstetricia y ginecología latino-americanas",
  ["obstet gynaecol"] = "The obstetrician & gynaecologist : the journal for continuing professional development from the Royal College of Obstetricians & Gynaecologists",
  ["obstet gynaecol cases rev"] = "Obstetrics and gynaecology cases - reviews",
  ["obstet gynaecol reprod med"] = "Obstetrics, gynaecology and reproductive medicine",
  ["obstet gynecol"] = "Obstetrics and gynecology",
  ["obstet gynecol annu"] = "Obstetrics and gynecology annual",
  ["obstet gynecol clin north am"] = "Obstetrics and gynecology clinics of North America",
  ["obstet gynecol int"] = "Obstetrics and gynecology international",
  ["obstet gynecol int j"] = "Obstetrics & gynecology international journal",
  ["obstet gynecol sci"] = "Obstetrics & gynecology science",
  ["obstet gynecol surv"] = "Obstetrical & gynecological survey",
  ["obstet med"] = "Obstetric medicine",
  ["obstet prat"] = "Obstetricia prática",
  ["obzornik mat fiz"] = "Drustvo Matematikov, Fizikov in Astronomov Slovenije. Obzornik za Matematiko in Fiziko. Izdaja",
  ["occ p inst men stud"] = "Occasional Papers of The Institute of Mennonite Studies",
  ["occas newsl lindsay club"] = "Occasional newsletter (Lindsay Club)",
  ["occas pap med hist aust"] = "Occasional papers on medical history Australia",
  ["occas pap mus zool univ mich"] = "Occasional papers of the Museum of Zoology, University of Michigan",
  ["occas pap r coll gen pract"] = "Occasional paper (Royal College of General Practitioners)",
  ["occas pap soc hist fed gov"] = "Occasional papers (Society for History in the Federal Government (U.S.))",
  ["occas pap tex tech univ mus"] = "Occasional papers (Texas Tech University. Museum)",
  ["occup environ med"] = "Occupational and environmental medicine",
  ["occup ergon"] = "Occupational ergonomics : the journal of the International Society for Occupational Ergonomics and Safety",
  ["occup health (auckl)"] = "Occupational Health",
  ["occup health (lond)"] = "Occupational health; a journal for occupational health nurses",
  ["occup health (wash d c)"] = "Occupational health (Washington, D.C.)",
  ["occup health nurs"] = "Occupational health nursing",
  ["occup health nurse (auckl)"] = "The Occupational health nurse",
  ["occup health rev"] = "Occupational health review",
  ["occup health saf"] = "Occupational health & safety (Waco, Tex.)",
  ["occup health sci"] = "Occupational health science",
  ["occup med"] = "Occupational medicine (Philadelphia, Pa.)",
  ["occup med (chic ill)"] = "Occupational medicine",
  ["occup med (lond)"] = "Occupational medicine (Oxford, England)",
  ["occup med (oxf)"] = "Occupational Medicine",
  ["occup med health aff"] = "Occupational medicine & health affairs",
  ["occup outlook q"] = "Occupational outlook quarterly",
  ["occup psychol"] = "Occupational psychology",
  ["occup ther health care"] = "Occupational therapy in health care",
  ["occup ther int"] = "Occupational therapy international",
  ["occup ther ment health"] = "Occupational therapy in mental health",
  ["occup ther rehabil"] = "Occupational therapy and rehabilitation",
  ["ocds bull"] = "Ocds Bulletin",
  ["ocean coast manag"] = "Ocean & coastal management",
  ["ocean coast manage"] = "Ocean and Coastal Management",
  ["ocean coastal manage"] = "Ocean and Coastal Management",
  ["ocean coastal res"] = "Ocean and Coastal Research",
  ["ocean dev int law"] = "Ocean Development & International Law",
  ["ocean dyn"] = "Ocean Dynamics",
  ["ocean eng"] = "Ocean Engineering",
  ["ocean eng oceanogr"] = "Ocean Engineering & Oceanography",
  ["ocean manage"] = "Ocean Manage.",
  ["ocean model"] = "Ocean Modelling",
  ["ocean model (oxf)"] = "Ocean modelling",
  ["ocean modell"] = "Ocean Modelling",
  ["ocean phys eng"] = "Ocean Physics and Engineering",
  ["ocean polar res"] = "Ocean and Polar Research",
  ["ocean sci"] = "Ocean Science",
  ["ocean sci discuss"] = "Ocean Science Discussions",
  ["ocean sci j"] = "Ocean Science Journal",
  ["ocean sci technol"] = "Ocean Science and Technology",
  ["ocean shoreline manage"] = "Ocean and Shoreline Management",
  ["ocean-land-atmos res"] = "Ocean-Land-Atmosphere Research",
  ["oceanogr fish open access j"] = "Oceanography & fisheries open access journal",
  ["oceanogr mag"] = "Oceanographic Magazine",
  ["oceanogr mar biol"] = "Oceanographic and Marine Biology",
  ["oceanogr meteorol"] = "Oceanography and Meteorology",
  ["oceanography (wash d c)"] = "Oceanography (Washington, D.C.)",
  ["oceanol hydrobiol stud"] = "Oceanological and hydrobiological studies",
  ["oceanolog acta"] = "Oceanologica Acta",
  ["oceanology (wash d c)"] = "Oceanology",
  ["ochanomizu igaku zasshi"] = "[Ochanomizu igaku zasshi] : the Ochanomizu medical journal",
  ["ocherki istor estestvoznan tekhn"] = "Akademiya Nauk Ukrainy",
  ["ochrana rostlin / czech academy of agricultural sciences|plant prot sci"] = "Plant protection science",
  ["ochsner clin rep"] = "Ochsner Clinic reports",
  ["ochsner j"] = "The Ochsner journal",
  ["octogon math mag"] = "Octogon Mathematical Magazine",
  ["ocul fr"] = "L'Oculiste français",
  ["ocul immunol inflamm"] = "Ocular immunology and inflammation",
  ["ocul oncol pathol"] = "Ocular oncology and pathology",
  ["ocul surf"] = "The ocular surface",
  ["odontoiatr prat"] = "Odontoiatria pratica",
  ["odontoiatr protesi dent"] = "Odontoiatria e protesi dentaria",
  ["odontoiatr rev iberoam med boca"] = "Odontoiatría; revista ibero-americana de medicina de la boca",
  ["odontol am"] = "Odontologia de América",
  ["odontol atual"] = "Odontologia Atual",
  ["odontol bull"] = "Odontological bulletin",
  ["odontol capixaba"] = "Odontologia capixaba",
  ["odontol chil"] = "Odontología chilena",
  ["odontol clin"] = "Odontologia Clinica",
  ["odontol conserv"] = "Odontologie Conservatrice",
  ["odontol din"] = "Odontologia dinámica",
  ["odontol foren tidskr"] = "Odontologiska föreningens tidskrift",
  ["odontol infant"] = "Odontología infantil",
  ["odontol jalisc"] = "Odontologia Jalisciense",
  ["odontol mexico"] = "Odontologia",
  ["odontol peru"] = "Odontologia Peruana",
  ["odontol postgrado"] = "Odontología de postgrado",
  ["odontol pract"] = "Odontologia practica",
  ["odontol revy"] = "Odontologisk revy",
  ["odontol revy suppl"] = "Odontologisk revy. Supplement",
  ["odontol samf finl arsb"] = "Odontologiska Samfundet i Finland Arsbok",
  ["odontol tidskr"] = "Odontologisk tidskrift",
  ["odontol urug"] = "Odontologia uruguaya",
  ["odontologia (lima)"] = "Odontología",
  ["odontologo (belo horizonte)"] = "Odontologo",
  ["odontostomatol implantoprotesi"] = "Odontostomatologia E Implantoprotesi",
  ["odontostomatol proodos"] = "Odontostomatologike proodos",
  ["odontostomatol trop"] = "Odonto-stomatologie tropicale = Tropical dental journal",
  ["oecd econ stud"] = "OECD Economic Studies",
  ["oecd obs"] = "The OECD observer. Organisation for Economic Co-operation and Development",
  ["oecon pol"] = "Oeconomica Polona",
  ["oekol plant"] = "Oekologia Plantarum",
  ["oekol vögel"] = "Oekologie der Vögel",
  ["oeno one"] = "OENO one",
  ["oesterr wasser abfallwirtsch"] = "Oesterreichische Wasser- und Abfallwirtschaft",
  ["oeuvres complet d’alembert ser iii opuscules et mem math 1757–1783"] = "OEuvres Completes de d’Alembert. Serie III. Opuscules et memoires mathematiques, 1757–1783",
  ["off bull am hosp assoc"] = "Official bulletin. American Hospital Association",
  ["off comm roster fellows"] = "Officers, committees, and roster of fellows. Medical Society of the State of North Carolina",
  ["off j can assoc crit care nurs"] = "Official journal of the Canadian Association of Critical Care Nurses",
  ["off our backs"] = "Off our backs",
  ["off publ dent hyg assoc state n y"] = "Official publication. Dental Hygienists' Association of the State of New York",
  ["off rec world health organ"] = "Official records. World Health Organization",
  ["offender programs rep"] = "Offender programs report",
  ["offentl gesundheitsdienst"] = "Der Öffentliche Gesundheitsdienst",
  ["offentl gesundheitswes"] = "Das Offentliche Gesundheitswesen",
  ["oftalmol zh"] = "Oftalmologicheskiĭ zhurnal",
  ["ohio dent j"] = "The Ohio dental journal",
  ["ohio hist"] = "Ohio history",
  ["ohio j public health"] = "Ohio journal of public health",
  ["ohio j sci"] = "The Ohio journal of science",
  ["ohio med"] = "Ohio medicine : journal of the Ohio State Medical Association",
  ["ohio north univ law rev"] = "Ohio Northern University law review",
  ["ohio nurses rev"] = "Ohio nurses review",
  ["ohio public health"] = "Ohio public health",
  ["ohio state law j"] = "Ohio State law journal",
  ["ohio state med j"] = "The Ohio State medical journal",
  ["ohio state u bull bus res"] = "Ohio State University Bulletin of Business Research",
  ["ohio state univ math res inst publ"] = "Ohio State University Mathematical Research Institute Publications",
  ["ohms dig"] = "OHMS digest",
  ["oil chem pollut"] = "Oil and Chemical Pollution",
  ["oil crop sci"] = "Oil Crop Science",
  ["oil energy trends"] = "Oil and Energy Trends",
  ["oil energy trends: annu stat rev"] = "Oil and Energy Trends: Annual Statistical Review",
  ["oil gas int"] = "Oil and Gas International",
  ["oil gas j"] = "Oil and Gas Journal",
  ["oil gas sci technol"] = "Oil & Gas Science and Technology",
  ["oil paint drug report"] = "Oil, paint, and drug reporter",
  ["oil petrochem pollut"] = "Oil and Petrochemical Pollution",
  ["oil, gas (hamburg, ger)"] = "Oil, Gas (Hamburg, Germany)",
  ["oilseeds fats, crops lipids"] = "Oilseeds and fats, Crops and Lipids",
  ["okajimas folia anat jpn"] = "Okajimas folia anatomica Japonica",
  ["okayama igakkai zasshi"] = "Okayama Igakkai zasshi",
  ["okinawan j am stud"] = "The Okinawan journal of American studies",
  ["okla nurse"] = "The Oklahoma nurse",
  ["okla statut annot okla"] = "Oklahoma statutes annotated. Oklahoma",
  ["oklahoma city univ law rev"] = "Oklahoma City University law review. Oklahoma City University",
  ["oklahoma law rev"] = "Oklahoma law review",
  ["okon anal rechts"] = "Okonomische Analyse des Rechts",
  ["okon polit"] = "Okonomi og Politik",
  ["ol corps gras, lipides"] = "Oleagineux, Corps Gras, Lipides",
  ["old engl newsl"] = "Old English newsletter",
  ["old northwest"] = "The Old Northwest",
  ["oldenburg familienkd"] = "Oldenburgische Familienkunde",
  ["omaly anio"] = "Omaly sy anio. Hier et aujourd'hui",
  ["oman j ophthalmol"] = "Oman journal of ophthalmology",
  ["oman med j"] = "Oman medical journal",
  ["omega (westport)"] = "Omega",
  ["omics j radiol"] = "OMICS journal of radiology",
  ["omics: j integr biol"] = "OMICS: A Journal of Integrative Biology",
  ["omnia med"] = "Omnia medica",
  ["omnia med suppl"] = "Omnia medica. Supplemento",
  ["omnia med ther"] = "Omnia medica et therapeutica",
  ["omnia ther"] = "Omnia therapeutica",
  ["omnia ther suppl"] = "Omnia therapeutica. Supplemento",
  ["on state public health (1921)"] = "On the state of the public health. Great Britain. Ministry of Health",
  ["on think"] = "On Thinking",
  ["ona j"] = "ONA journal",
  ["onco targets ther"] = "OncoTargets and therapy",
  ["onco ther"] = "Onco Therapeutics",
  ["oncodev biol med"] = "Oncodevelopmental biology and medicine : the journal of the International Society for Oncodevelopmental Biology and Medicine",
  ["oncogen (westerville)"] = "Oncogen",
  ["oncogene res"] = "Oncogene research",
  ["oncol discov"] = "Oncology discovery",
  ["oncol fellows"] = "Oncology fellows",
  ["oncol hematol rev"] = "Oncology & hematology review",
  ["oncol issues"] = "Oncology issues",
  ["oncol lett"] = "Oncology letters",
  ["oncol nurs forum"] = "Oncology nursing forum",
  ["oncol nurs news"] = "Oncology nursing news",
  ["oncol nurse"] = "The Oncology nurse",
  ["oncol nurse advis"] = "Oncology nurse advisor",
  ["oncol nurse ed"] = "Oncology nurse edition",
  ["oncol nutr connect"] = "Oncology nutrition connection : a publication of the ON DPG, Oncology Nutrition Dietetic Practice Group",
  ["oncol pharm"] = "The oncology pharmacist",
  ["oncol rep"] = "Oncology reports",
  ["oncol res"] = "Oncology research",
  ["oncol res treat"] = "Oncology research and treatment",
  ["oncol rev"] = "Oncology reviews",
  ["oncol signal"] = "Oncology signaling",
  ["oncol ther"] = "Oncology and therapy",
  ["oncologie (paris)"] = "Oncologie (Paris, France)",
  ["oncology (huntingt)"] = "Oncology",
  ["oncology (williston park)"] = "Oncology (Williston Park, N.Y.)",
  ["oncolytic virother"] = "Oncolytic virotherapy",
  ["oncotargets ther"] = "OncoTargets and Therapy",
  ["onderstepoort j vet res"] = "The Onderstepoort journal of veterinary research",
  ["onderstepoort j vet sci anim ind"] = "The Onderstepoort journal of veterinary science and animal industry",
  ["ondontol bonaer"] = "Odontología bonaerense",
  ["one earth"] = "One earth (Cambridge, Mass.)",
  ["one ecosyst"] = "One Ecosystem",
  ["one health"] = "One health (Amsterdam, Netherlands)",
  ["one health outlook"] = "One health outlook",
  ["one one"] = "One on one",
  ["onkologe (berl)"] = "Der Onkologe : Organ der Deutschen Krebsgesellschaft e.V",
  ["online int interdiscip res j"] = "Online international interdisciplinary research journal : an international multidisciplinary journal",
  ["online j anal comb"] = "Online Journal of Analytic Combinatorics",
  ["online j bioinform"] = "Online journal of bioinformatics : OJB",
  ["online j biol sci"] = "OnLine journal of biological sciences",
  ["online j commun media technol"] = "Online journal of communication and media technologies",
  ["online j complement altern med"] = "Online journal of complementary & alternative medicine",
  ["online j cult competence nurs healthc"] = "Online journal of cultural competence in nursing and healthcare",
  ["online j curr clin trials"] = "The Online journal of current clinical trials",
  ["online j distance educ elearn"] = "The online journal of distance education and e-learning : TOJDEL",
  ["online j health allied sci"] = "Online journal of health and allied sciences : OJHAS",
  ["online j health ethics"] = "Online journal of health ethics",
  ["online j issues nurs"] = "Online journal of issues in nursing",
  ["online j knowl synth nurs"] = "The online journal of knowledge synthesis for nursing",
  ["online j nurs inform"] = "On-line journal of nursing informatics",
  ["online j public health inform"] = "Online journal of public health informatics",
  ["online j rural nurs health care"] = "Online journal of rural nursing and health care : the official journal of the Rural Nurse Organization",
  ["online j rural res policy"] = "Online journal of rural research and policy",
  ["online j rural urban res"] = "Online journal of rural and urban research",
  ["online j vet res"] = "Online journal of veterinary research : OJVR",
  ["online learn"] = "Online learning : the official journal of the Online Learning Consortium",
  ["online soc netw media"] = "Online social networks and media",
  ["online, adl-nachr"] = "Online, ADL-Nachrichten",
  ["ons amst"] = "Ons Amsterdam",
  ["ons connect"] = "ONS connect",
  ["ons heem"] = "Ons heem",
  ["ons news"] = "ONS news",
  ["onsen kagaku"] = "[Onsen kagaku] Journal of the Balneological Society of Japan",
  ["ont dent"] = "Ontario dentist",
  ["ont health technol assess ser"] = "Ontario health technology assessment series",
  ["ont hist"] = "Ontario history",
  ["ont med rev"] = "Ontario medical review",
  ["ont psychol"] = "The Ontario psychologist",
  ["ontos math log"] = "Ontos Mathematical Logic",
  ["ooq occup outlook q"] = "OOQ; Occupational Outlook Quarterly",
  ["oostvlaam zanten"] = "Oostvlaamsche zanten",
  ["opec energy rev"] = "OPEC Energy Review",
  ["opec rev"] = "OPEC Review",
  ["open access anim physiol"] = "Open access animal physiology",
  ["open access bioinformatics"] = "Open access bioinformatics",
  ["open access emerg med"] = "Open access emergency medicine : OAEM",
  ["open access insect physiol"] = "Open access insect physiology",
  ["open access j behav sci psychol"] = "Open access journal of behavioural science & psychology",
  ["open access j biomed eng appl"] = "Open access journal of biomedical engineering and its applications",
  ["open access j biomed eng biosci"] = "Open access journal of biomedical engineering and biosciences",
  ["open access j clin trials"] = "Open access journal of clinical trials",
  ["open access j contracept"] = "Open access journal of contraception",
  ["open access j gerontol geriatr med"] = "Open access journal of gerontology & geriatric medicine",
  ["open access j neurol neurosurg"] = "Open access journal of neurology & neurosurgery",
  ["open access j ophthalmol"] = "Open access journal of ophthalmology",
  ["open access j sports med"] = "Open access journal of sports medicine",
  ["open access j toxicol"] = "Open access journal of toxicology",
  ["open access j transl med res"] = "Open access journal of translational medicine & research",
  ["open access j urol"] = "Open access journal of urology",
  ["open access maced j med sci"] = "Open access Macedonian journal of medical sciences",
  ["open access med stat"] = "Open access medical statistics",
  ["open access rheumatol"] = "Open access rheumatology : research and reviews",
  ["open addict j"] = "The open addiction journal",
  ["open agric"] = "Open agriculture",
  ["open agric j"] = "The open agriculture journal",
  ["open aids j"] = "The open AIDS journal",
  ["open allergy j"] = "The open allergy journal",
  ["open anal chem j"] = "The open analytical chemistry journal",
  ["open anthropol j"] = "The open anthropology journal",
  ["open appl math j"] = "Open Applied Mathematics Journal",
  ["open astron"] = "Open Astronomy",
  ["open atmos sci j"] = "Open Atmospheric Science Journal",
  ["open autoimmun j"] = "The Open autoimmunity journal",
  ["open biochem j"] = "Open Biochemistry Journal",
  ["open biochem lett"] = "Open Biochemistry Letters",
  ["open bioinform j"] = "Open Bioinformatics Journal",
  ["open bioinforma j"] = "The open bioinformatics journal",
  ["open biol"] = "Open biology",
  ["open biol j"] = "The open biology journal",
  ["open biomark j"] = "The open biomarkers journal",
  ["open biomater j"] = "Open Biomaterials Journal",
  ["open biomed eng j"] = "Open Biomedical Engineering Journal",
  ["open biomed englett"] = "Open Biomedical Engineering Letters",
  ["open biotechnol j"] = "Open Biotechnology Journal",
  ["open bone j"] = "The Open bone journal",
  ["open book ser"] = "The Open Book Series",
  ["open cancer immunol j"] = "The Open cancer immunology journal",
  ["open cancer j"] = "The open cancer journal",
  ["open cardiovasc med j"] = "The open cardiovascular medicine journal",
  ["open cell dev biol j"] = "The open cell development & biology journal",
  ["open cell signal j"] = "The open cell signaling journal",
  ["open ceram"] = "Open Ceramics",
  ["open chem"] = "Open chemistry",
  ["open chem eng j"] = "Open Chemical Engineering Journal",
  ["open circ vasc j"] = "The open circulation & vascular journal",
  ["open clin cancer j"] = "The open clinical cancer journal",
  ["open commun nonlinear math phys"] = "Open Communications in Nonlinear Mathematical Physics",
  ["open complement med j"] = "The open complementary medicine journal",
  ["open conf proc j"] = "The open conference proceedings journal",
  ["open crit care med j"] = "The open critical care medicine journal",
  ["open cybern syst j"] = "Open Cybernetics and Systemics Journal",
  ["open dent j"] = "The open dentistry journal",
  ["open dermatol j"] = "The open dermatology journal",
  ["open diabetes j"] = "The open diabetes journal",
  ["open drug delivery j"] = "Open Drug Delivery Journal",
  ["open drug delivery rev"] = "Open Drug Delivery Reviews",
  ["open drug discov j"] = "The open drug discovery journal",
  ["open drug metab j"] = "The open drug metabolism journal",
  ["open economies rev"] = "Open Economies Review",
  ["open endocrinol j"] = "The open endocrinology journal",
  ["open eng"] = "Open Engineering",
  ["open entomol j"] = "The open entomology journal",
  ["open enzym inhib j"] = "The open enzyme inhibition journal",
  ["open epidemiol j"] = "The open epidemiology journal",
  ["open fam stud j"] = "The open family studies journal",
  ["open food sci j"] = "Open Food Science Journal",
  ["open forensic sci j"] = "The open forensic science journal",
  ["open forum infect dis"] = "Open forum infectious diseases",
  ["open gene ther j"] = "The open gene therapy journal",
  ["open genomics j"] = "The open genomics journal",
  ["open geosci"] = "Open Geosciences",
  ["open geospatial data software stand"] = "Open Geospatial Data, Software and Standards",
  ["open geriatr med j"] = "The open geriatric medicine journal",
  ["open glycosci"] = "Open glycoscience",
  ["open health data"] = "Open health data",
  ["open health serv policy j"] = "The open health services and policy journal",
  ["open heart"] = "Open heart",
  ["open hypertens j"] = "The open hypertension journal",
  ["open immunol j"] = "The open immunology journal",
  ["open inf syst j"] = "The open information systems journal",
  ["open infect dis j"] = "The open infectious diseases journal",
  ["open inflamm j"] = "The open inflammation journal",
  ["open inorg chem j"] = "Open Inorganic Chemistry Journal",
  ["open inorg chem rev"] = "Open Inorganic Chemistry Reviews",
  ["open j anesthesiol"] = "Open Journal of Anesthesiology",
  ["open j anim sci"] = "Open journal of animal sciences",
  ["open j apoptosis"] = "Open journal of apoptosis",
  ["open j biophys"] = "Open journal of biophysics",
  ["open j bioresour"] = "Open journal of bioresources",
  ["open j cardiovasc surg"] = "Open journal of cardiovascular surgery",
  ["open j clin diagn"] = "Open journal of clinical diagnostics",
  ["open j clin med case rep"] = "Open journal of clinical & medical case reports",
  ["open j discrete appl math"] = "Open Journal of Discrete Applied Mathematics. ODAM",
  ["open j ecol"] = "Open journal of ecology",
  ["open j endocr metab dis"] = "Open journal of endocrine and metabolic diseases",
  ["open j epidemiol"] = "Open journal of epidemiology",
  ["open j for"] = "Open journal of forestry",
  ["open j gastroenterol"] = "Open journal of gastroenterology",
  ["open j genet"] = "Open journal of genetics",
  ["open j hematol"] = "Open journal of hematology",
  ["open j immunol"] = "Open journal of immunology",
  ["open j inorg chem"] = "Open journal of inorganic chemistry",
  ["open j inorg non-met mater"] = "Open Journal of Inorganic Non-Metallic Materials",
  ["open j math optim"] = "Open Journal of Mathematical Optimization (OJMO)",
  ["open j med chem"] = "Open journal of medicinal chemistry",
  ["open j med microbiol"] = "Open journal of medical microbiology",
  ["open j med psychol"] = "Open journal of medical psychology",
  ["open j met"] = "Open journal of metal",
  ["open j mod linguist"] = "Open journal of modern linguistics",
  ["open j mol integr physiol"] = "Open journal of molecular and integrative physiology",
  ["open j nephrol"] = "Open journal of nephrology",
  ["open j neurosci"] = "Open journal of neuroscience",
  ["open j nurs"] = "Open journal of nursing",
  ["open j obstet gynecol"] = "Open journal of obstetrics and gynecology",
  ["open j occup ther"] = "The open journal of occupational therapy",
  ["open j ophthalmol"] = "Open journal of ophthalmology",
  ["open j org polym mater"] = "Open Journal of Organic Polymer Materials",
  ["open j orthop"] = "Open journal of orthopedics",
  ["open j pediatr"] = "Open journal of pediatrics",
  ["open j philos"] = "Open journal of philosophy",
  ["open j phys chem"] = "Open journal of physical chemistry",
  ["open j polit sci"] = "Open journal of political science",
  ["open j polym chem"] = "Open Journal of Polymer Chemistry",
  ["open j prev med"] = "Open journal of preventive medicine",
  ["open j psychiatr"] = "Open journal of psychiatry",
  ["open j psychiatry allied sci"] = "Open journal of psychiatry & allied sciences",
  ["open j radiol"] = "Open journal of radiology",
  ["open j rheumatol autoimmune dis"] = "Open journal of rheumatology and autoimmune diseases",
  ["open j soc sci"] = "Open journal of social sciences",
  ["open j soil sci"] = "Open Journal of Soil Science",
  ["open j stat"] = "Open journal of statistics",
  ["open j stomatol"] = "Open journal of stomatology",
  ["open j stud new sci res"] = "Open Journal for Students and New Scientists Research OJSSR",
  ["open j synth theory appl"] = "Open Journal of Synthesis Theory and Applications",
  ["open j vet med"] = "Open journal of veterinary medicine",
  ["open life sci"] = "Open life sciences",
  ["open linguist"] = "Open linguistics",
  ["open longev sci"] = "Open Longevity Science",
  ["open magn reson j"] = "Open Magnetic Resonance Journal",
  ["open magn reson rev"] = "The open magnetic resonance reviews",
  ["open mar biol j"] = "The open marine biology journal",
  ["open math"] = "Open Mathematics",
  ["open math j"] = "Open Mathematics Journal",
  ["open med"] = "Open medicine : a peer-reviewed, independent, open-access journal",
  ["open med (wars)"] = "Open medicine (Warsaw, Poland)",
  ["open med chem j"] = "Open Medicinal Chemistry Journal",
  ["open med imaging j"] = "The open medical imaging journal",
  ["open med inform j"] = "The open medical informatics journal",
  ["open med j"] = "Open medicine journal",
  ["open memo natl health med res counc aust"] = "Open memo. National Health and Medical Research Council (Australia). Fatigue Laboratory",
  ["open microbiol j"] = "The open microbiology journal",
  ["open mind"] = "Open Mind",
  ["open mind (camb)"] = "Open mind : discoveries in cognitive science",
  ["open mycol j"] = "The open mycology journal",
  ["open nanomed j"] = "The open nanomedicine journal",
  ["open nanosci j"] = "Open Nanoscience Journal",
  ["open neuroendocrinol j"] = "Open neuroendocrinology journal (Online)",
  ["open neuroimag j"] = "The open neuroimaging journal",
  ["open neurol j"] = "The open neurology journal",
  ["open neuropsychopharmacol j"] = "The open neuropsychopharmacology journal",
  ["open neurosci j"] = "The open neuroscience journal",
  ["open nucl med j"] = "The open nuclear medicine journal",
  ["open numer methods j"] = "Open Numerical Methods Journal",
  ["open nurs j"] = "The open nursing journal",
  ["open nutr j"] = "The open nutrition journal",
  ["open nutraceuticals j"] = "The open nutraceuticals journal",
  ["open obes j"] = "The open obesity journal",
  ["open ophthalmol j"] = "The open ophthalmology journal",
  ["open org chem j"] = "Open Organic Chemistry Journal",
  ["open ornithol j"] = "The open ornithology journal",
  ["open orthop j"] = "The open orthopaedics journal",
  ["open pacing electrophysiol ther j"] = "The open pacing, electrophysiology & therapy journal",
  ["open pain j"] = "The open pain journal",
  ["open parasitol j"] = "The open parasitology journal",
  ["open pathol j"] = "The open pathology journal",
  ["open pediatr med journal"] = "The open pediatric medicine journal",
  ["open pharmacol j"] = "The open pharmacology journal",
  ["open phys"] = "Open Physics",
  ["open phys chem j"] = "Open Physical Chemistry Journal",
  ["open phys chem rev"] = "Open Physical Chemistry Reviews",
  ["open proc life sci"] = "Open Proceedings in Life Sciences",
  ["open prost cancer j"] = "The open prostate cancer journal",
  ["open proteomics j"] = "The open proteomics journal",
  ["open psychiatr j"] = "The open psychiatry journal",
  ["open psychol"] = "Open psychology",
  ["open psychol j"] = "The open psychology journal",
  ["open public health j"] = "The open public health journal",
  ["open quat"] = "Open quaternary",
  ["open rehabil j"] = "The open rehabilitation journal",
  ["open reprod sci j"] = "The open reproductive science journal",
  ["open res europe"] = "Open Research Europe",
  ["open respir med j"] = "The open respiratory medicine journal",
  ["open rheumatol j"] = "The open rheumatology journal",
  ["open sleep j"] = "The open sleep journal",
  ["open spectrosc j"] = "Open Spectroscopy Journal",
  ["open sports sci j"] = "The open sports sciences journal",
  ["open stem cell j"] = "The open stem cell journal",
  ["open surg oncol j"] = "Open surgical oncology journal (Online)",
  ["open syst biol j"] = "The open systems biology journal",
  ["open syst inf dyn"] = "Open Systems and Information Dynamics",
  ["open tissue eng regen med j"] = "The open tissue engineering and regenerative medicine journal",
  ["open toxicol j"] = "The open toxicology journal",
  ["open toxinology j"] = "The open toxinology journal",
  ["open transl med j"] = "The open translational medicine journal",
  ["open transplant j"] = "The open transplantation journal",
  ["open trop med j"] = "The open tropical medicine journal",
  ["open urol nephrol j"] = "The open urology & nephrology journal",
  ["open vaccine j"] = "The open vaccine journal",
  ["open vet j"] = "Open veterinary journal",
  ["open vet sci j"] = "The open veterinary science journal",
  ["open virol j"] = "The open virology journal",
  ["open water"] = "Open water",
  ["open womens health j"] = "The open women's health journal",
  ["open zool j"] = "The open zoology journal",
  ["oper dent"] = "Operative dentistry",
  ["oper dent suppl"] = "Operative dentistry. Supplement",
  ["oper manage res"] = "Operations Management Research",
  ["oper matrices"] = "Operators and Matrices",
  ["oper neurosurg (hagerstown)"] = "Operative neurosurgery (Hagerstown, Md.)",
  ["oper orthop traumatol"] = "Operative Orthopädie und Traumatologie",
  ["oper program systems ser"] = "Operating and Programming Systems Series",
  ["oper res"] = "Operations Research",
  ["oper res forum"] = "Operations Research Forum",
  ["oper res health care"] = "Operations research for health care",
  ["oper res int j"] = "Operational Research An International Journal",
  ["oper res lett"] = "Operations Research Letters",
  ["oper res perspect"] = "Operations Research Perspectives",
  ["oper res proc"] = "Operations Research Proceedings",
  ["oper res q"] = "Operational research quarterly",
  ["oper res ser"] = "Operations Research Series",
  ["oper res trans"] = "Operations Research Transactions. Yunchouxue Xuebao",
  ["oper res/comput sci interfaces ser"] = "Operations Research/Computer Science Interfaces Series",
  ["oper tech orthop"] = "Operative techniques in orthopaedics",
  ["oper tech otolayngol head neck surg"] = "Operative techniques in otolaryngology--head and neck surgery",
  ["oper tech sports med"] = "Operative techniques in sports medicine",
  ["oper tech thorac cardiovasc surg"] = "Operative techniques in thoracic and cardiovascular surgery : a comparative atlas : an official publication of the American Association for Thoracic Surgery",
  ["oper theory adv appl"] = "Operator Theory: Advances and Applications",
  ["opera bot"] = "Opera Botanica",
  ["opera med physiol"] = "Opera medica et physiologica",
  ["opera pharm rariora"] = "Opera pharmaceutica rariora",
  ["opera q"] = "The opera quarterly",
  ["ophthal plast reconstr surg"] = "Ophthalmic Plastic and Reconstructive Surgery",
  ["ophthalmic epidemiol"] = "Ophthalmic epidemiology",
  ["ophthalmic genet"] = "Ophthalmic genetics",
  ["ophthalmic lit"] = "Ophthalmic literature",
  ["ophthalmic med image anal (2019)"] = "Ophthalmic medical image analysis : 6th International Workshop, OMIA 2019, held in conjunction with MICCAI 2019, Shenzhen, China, October 17, Proceedings. OMIA (Workshop) (6th : 2019 : Shenzhen Shi, China)",
  ["ophthalmic paediatr genet"] = "Ophthalmic paediatrics and genetics",
  ["ophthalmic physiol opt"] = "Ophthalmic & physiological optics : the journal of the British College of Ophthalmic Opticians (Optometrists)",
  ["ophthalmic plast reconstr surg"] = "Ophthalmic plastic and reconstructive surgery",
  ["ophthalmic res"] = "Ophthalmic research",
  ["ophthalmic semin"] = "Ophthalmic seminars",
  ["ophthalmic surg"] = "Ophthalmic surgery",
  ["ophthalmic surg lasers"] = "Ophthalmic surgery and lasers",
  ["ophthalmic surg lasers imaging"] = "Ophthalmic surgery, lasers & imaging : the official journal of the International Society for Imaging in the Eye",
  ["ophthalmic surg lasers imaging retina"] = "Ophthalmic surgery, lasers & imaging retina",
  ["ophthalmol clin north am"] = "Ophthalmology clinics of North America",
  ["ophthalmol eye dis"] = "Ophthalmology and eye diseases",
  ["ophthalmol glaucoma"] = "Ophthalmology. Glaucoma",
  ["ophthalmol ibero am"] = "Ophthalmologia ibero americana",
  ["ophthalmol int"] = "Ophthalmology international",
  ["ophthalmol res rep"] = "Ophthalmology research and reports",
  ["ophthalmol retina"] = "Ophthalmology. Retina",
  ["ophthalmol ther"] = "Ophthalmology and therapy",
  ["opin atty gen state md md atty gen off"] = "Opinions of the Attorney General--State of Maryland. Maryland. Attorney General's Office",
  ["opt - int j light electron opt"] = "Optik - International Journal for Light and Electron Optics",
  ["opt acta"] = "Optica Acta",
  ["opt acta (lond)"] = "Optica acta",
  ["opt commun"] = "Optics Communications",
  ["opt continuum"] = "Optics Continuum",
  ["opt dev"] = "Optical developments",
  ["opt eng"] = "Optical Engineering",
  ["opt express"] = "Optics express",
  ["opt fiber technol"] = "Optical Fiber Technology",
  ["opt j rev optom"] = "Optical journal and review of optometry",
  ["opt laser technol"] = "Optics & Laser Technology",
  ["opt lasers eng"] = "Optics and Lasers in Engineering",
  ["opt lett"] = "Optics letters",
  ["opt mater"] = "Optical Materials",
  ["opt mater (amst)"] = "Optical Materials",
  ["opt mater express"] = "Optical Materials Express",
  ["opt mater: x"] = "Optical Materials: X",
  ["opt mem neural networks"] = "Optical Memory and Neural Networks",
  ["opt nanosc"] = "Optical Nanoscopy",
  ["opt nanoscopy"] = "Optical nanoscopy",
  ["opt news"] = "Optics News",
  ["opt photonics lett"] = "Optics and Photonics Letters",
  ["opt photonics news"] = "Optics & Photonics News",
  ["opt photonik"] = "Optik & Photonik",
  ["opt quantum electron"] = "Optical and Quantum Electronics",
  ["opt rev"] = "Optical Review",
  ["opt spectrosc"] = "Optics and Spectroscopy",
  ["opt spektrosk"] = "Optika i Spektroskopiya",
  ["opt spektrosk [opt spectrosc (ussr)]"] = "Optika i Spektroskopiya [Optics and Spectroscopy (USSR)]",
  ["opt switching networking"] = "Optical Switching and Networking",
  ["opt zh"] = "Opticheskii Zhurnal",
  ["opt-mekh prom-st"] = "Optiko-Mekhanicheskaya Promyshlennost",
  ["optica quantum"] = "Optica Quantum",
  ["optik (stuttg)"] = "Optik",
  ["optik (stuttgart)"] = "Optik (Stuttgart)",
  ["optim control appl methods"] = "Optimal control applications & methods",
  ["optim eng"] = "Optimization and Engineering",
  ["optim issled oper"] = "\\cyr Optimizatsiya i Issledovanie Operatsiy",
  ["optim lett"] = "Optimization Letters",
  ["optim methods softw"] = "Optimization methods & software",
  ["optim methods software"] = "Optimization Methods and Software",
  ["optimal control appl methods"] = "Optimal Control Applications & Methods",
  ["options (van nuys)"] = "Options (Van Nuys, Calif.)",
  ["options policy pract"] = "Options for policy and practice",
  ["opto-electron rev"] = "Opto-Electronics Review",
  ["optoelectron adv mater rapid commun"] = "Optoelectronics and Advanced Materials, Rapid Communications",
  ["optoelectron commun conf"] = "Optoelectronics and communications conference",
  ["optoelectron instrum data process"] = "Optoelectronics, Instrumentation and Data Processing",
  ["optoelectron lett"] = "Optoelectronics Letters",
  ["optoelectron mater devices"] = "Optoelectronic materials and devices",
  ["optom clin"] = "Optometry clinics : the official publication of the Prentice Society",
  ["optom glaucoma soc e j"] = "Optometric Glaucoma Society e-journal",
  ["optom open access"] = "Optometry, open access",
  ["optom vis perform"] = "Optometry & visual performance",
  ["optom vis sci"] = "Optometry and vision science : official publication of the American Academy of Optometry",
  ["optom wkly"] = "The Optometric weekly",
  ["optomechatron technol isot 2010 int symp (2010)"] = "Optomechatronic Technologies (ISOT), 2010 International Symposium on : 25-27 Oct. 2010 : [Toronto, ON]. International Symposium on Optomechatronic Technologies (2010 : Toronto, Ont.)",
  ["opt–mekh prom [sov j opt technol]"] = "Optiko–Mekhanicheskaya Promyshlennost [Soviet Journal of Optical Technology]",
  ["opusc med"] = "Opuscula medica",
  ["opusc philolichenum"] = "Opuscula philolichenum",
  ["opusc zool inst zoosyst oecol univ bp"] = "Opuscula Zoologica Instituti Zoosystematici et Oecologici Universitatis Budapestinensis",
  ["opuscula math"] = "Opuscula Mathematica",
  ["or 20 context aware oper theaters comput assist robot endosc clin image based proced skin image anal (2018)"] = "OR 2.0 context-aware operating theaters, computer assisted robotic endoscopy, clinical image-based procedures, and skin image analysis : first international workshop, OR 2.0 2018, 5th international workshop, CARE 2018, 7th international workshop, CLIP 2018, third international workshop, ISIC 2018, held in conjunction with MICCAI 2018, Granada, Spain, September 16-20, 2018 : proceedings. OR 2.0 (Workshop) (1st : 2018 : Granada, Spain)",
  ["or health bull"] = "Oregon health bulletin",
  ["or manager"] = "OR manager",
  ["or spectr"] = "OR spectrum : quantitative approaches in management",
  ["or spectrum"] = "OR Spectrum",
  ["or spektrum"] = "Operations Research Spektrum",
  ["or tech"] = "OR tech",
  ["oral biosci med"] = "Oral biosciences & medicine : OBM",
  ["oral dis"] = "Oral diseases",
  ["oral health"] = "Oral health",
  ["oral health case rep"] = "Oral health case reports",
  ["oral health dent manag"] = "Oral health and dental management",
  ["oral health dent stud"] = "Oral health and dental studies",
  ["oral health prev dent"] = "Oral health & preventive dentistry",
  ["oral hist (colch)"] = "Oral history",
  ["oral hist rev"] = "The Oral history review",
  ["oral hyg"] = "Oral hygiene",
  ["oral implantol"] = "Oral implantology",
  ["oral implantol (rome)"] = "ORAL & implantology",
  ["oral maxillofac surg"] = "Oral and maxillofacial surgery",
  ["oral maxillofac surg cases"] = "Oral and maxillofacial surgery cases",
  ["oral maxillofac surg clin north am"] = "Oral and maxillofacial surgery clinics of North America",
  ["oral microbiol immunol"] = "Oral microbiology and immunology",
  ["oral oncol"] = "Oral oncology",
  ["oral radiol"] = "Oral radiology",
  ["oral sci int"] = "Oral science international",
  ["oral sci rev"] = "Oral sciences reviews",
  ["oral surg"] = "Oral surgery",
  ["oral surg oral diagn"] = "Oral surgery, oral diagnosis : OSD",
  ["oral surg oral med oral pathol"] = "Oral surgery, oral medicine, and oral pathology",
  ["oral surg oral med oral pathol oral radiol"] = "Oral surgery, oral medicine, oral pathology and oral radiology",
  ["oral surg oral med oral pathol oral radiol endod"] = "Oral surgery, oral medicine, oral pathology, oral radiology, and endodontics",
  ["orale implantol"] = "Orale Implantologie",
  ["orange cty dent soc bull"] = "Orange County Dental Society Bulletin",
  ["orbis econ"] = "Orbis Economicus",
  ["orbita math"] = "Orbita Mathematicae",
  ["order (dordr)"] = "Order (Dordrecht, Netherlands)",
  ["ordre natl chir dent cons natl bull off"] = "Bulletin officiel du Conseil national de l'Ordre",
  ["ore energy resour geol"] = "Ore and Energy Resource Geology",
  ["ore geol rev"] = "Ore Geology Reviews",
  ["oreg hist q"] = "Oregon historical quarterly. Oregon Historical Society",
  ["oreg nurse"] = "The Oregon nurse",
  ["oreg state dent j"] = "Oregon state dental journal",
  ["oregon bus rev"] = "Oregon Business Review",
  ["oregon law rev"] = "Oregon law review",
  ["org agric"] = "Organic Agriculture",
  ["org biomol chem"] = "Organic & biomolecular chemistry",
  ["org chem front"] = "Organic Chemistry Frontiers",
  ["org chem int"] = "Organic chemistry International",
  ["org divers evol"] = "Organisms, diversity & evolution",
  ["org electron"] = "Organic Electronics",
  ["org geochem"] = "Organic Geochemistry",
  ["org lett"] = "Organic letters",
  ["org magn reson"] = "Organic Magnetic Resonance",
  ["org mass spectrom"] = "Organic Mass Spectrometry",
  ["org med chem lett"] = "Organic and medicinal chemistry letters",
  ["org prep proced int"] = "Organic Preparations and Procedures International",
  ["org process res dev"] = "Organic Process Research and Development",
  ["org react"] = "Organic Reactions",
  ["org react mech"] = "Organic Reaction Mechanics",
  ["org synth"] = "Organic Syntheses",
  ["organ behav hum decis process"] = "Organizational behavior and human decision processes",
  ["organ behav hum perform"] = "Organizational behavior and human performance",
  ["organ cybersecur j: pract process people"] = "Organizational Cybersecurity Journal: Practice, Process and People",
  ["organ dyn"] = "Organizational dynamics",
  ["organ environ"] = "Organization & environment",
  ["organ ethic"] = "Organizational ethics : healthcare, business, and policy : OE",
  ["organ res methods"] = "Organizational research methods",
  ["organic synth"] = "Organic syntheses; an annual publication of satisfactory methods for the preparation of organic chemicals",
  ["organiz vychisl struktur process"] = "Organizatsiya Vychislitelnykh Struktur i Protsessov",
  ["organiz vycisl struktur process"] = "Organizatsiya Vychislitelnykh Struktur i Protsessov",
  ["organization (lond)"] = "Organization (London, England)",
  ["organohalogen compd"] = "Organohalogen Compounds",
  ["organomet chem"] = "Organometallic Chemistry",
  ["organon f (bratisl)"] = "Organon F",
  ["oriens extremus"] = "Oriens extremus",
  ["orient insects"] = "Oriental insects",
  ["orient j chem"] = "Oriental Journal of Chemistry",
  ["orient pharm exp med"] = "Oriental pharmacy and experimental medicine",
  ["orientac med"] = "Orientación médica",
  ["oriente crist"] = "Oriente Cristiano",
  ["oriente mod"] = "Oriente moderno",
  ["orig life"] = "Origins of life",
  ["orig life evol biosph"] = "Origins of life and evolution of the biosphere : the journal of the International Society for the Study of the Origin of Life",
  ["origins life evol biosphere"] = "Origins of Life and Evolution of the Biosphere",
  ["orins rep us at energy comm"] = "ORINS [reports]. U.S. Atomic Energy Commission",
  ["orizz ortop odie riabil"] = "Orizzonti della ortopedia odierna e della riabilitazione",
  ["orl head neck nurs"] = "ORL-head and neck nursing : official journal of the Society of Otorhinolaryngology and Head-Neck Nurses",
  ["orl j otorhinolaryngol relat spec"] = "ORL; journal for oto-rhino-laryngology and its related specialties",
  ["ornac j"] = "ORNAC journal",
  ["ornis fenn"] = "Ornis fennica",
  ["ornis norv"] = "Ornis norvegica",
  ["ornis svec"] = "Ornis Svecica",
  ["ornith res"] = "Ornithology Research",
  ["ornithol abh"] = "Ornithologische Abhandlungen",
  ["ornithol appl"] = "Ornithological Applications",
  ["ornithol beob"] = "ornithologische Beobachter, Der",
  ["ornithol jahresh baden-württ"] = "Ornithologische Jahreshefte für Baden-Württemberg",
  ["ornithol monogr"] = "Ornithological monographs",
  ["ornithol sci"] = "Ornithological science",
  ["ornitol neotrop"] = "Ornitología neotropical",
  ["ornl-nsic rep"] = "ORNL-NSIC [reports]. U.S. Atomic Energy Commission",
  ["ornl-tm rep"] = "ORNL-TM [reports]. U.S. Atomic Energy Commission",
  ["oro rep"] = "ORO [reports]. U.S. Atomic Energy Commission",
  ["orphanet j rare dis"] = "Orphanet journal of rare diseases",
  ["orthod craniofac res"] = "Orthodontics & craniofacial research",
  ["orthod epitheor"] = "Orthodontikē epitheōrēsē : epiotēmoniko periodiko tēs Orthodontikēs Etaireias tēs Ellados",
  ["orthod epitheorese"] = "Orthodontike Epitheorese",
  ["orthod fr"] = "L' Orthodontie française",
  ["orthod rev"] = "Orthodontic review",
  ["orthod waves"] = "Orthodontic waves (English ed.)",
  ["orthodontics (chic)"] = "Orthodontics : the art and practice of dentofacial enhancement",
  ["orthop audiosynop contin med educ"] = "Orthopaedic audio-synopsis continuing medical education",
  ["orthop clin north am"] = "The Orthopedic clinics of North America",
  ["orthop j harv med sch"] = "The orthopaedic journal at Harvard Medical School",
  ["orthop j sports med"] = "Orthopaedic journal of sports medicine",
  ["orthop knowl online"] = "Orthopaedic knowledge online",
  ["orthop muscular syst"] = "Orthopedic & muscular system : current research",
  ["orthop nurs"] = "Orthopedic nursing",
  ["orthop phys ther clin n am"] = "Orthopaedic physical therapy clinics of North America",
  ["orthop res rev"] = "Orthopedic research and reviews",
  ["orthop rev"] = "Orthopaedic review",
  ["orthop rev (pavia)"] = "Orthopedic reviews",
  ["orthop sports med"] = "Orthopedics and sports medicine : open access journal",
  ["orthop surg"] = "Orthopaedic surgery",
  ["orthop tech (engl ed)"] = "Orthopädie-Technik (English ed.)",
  ["orthop trauma"] = "Orthopaedics and trauma",
  ["orthop traumatol surg res"] = "Orthopaedics & traumatology, surgery & research : OTSR",
  ["ortop maxilar"] = "Ortopedia maxilar",
  ["ortop traumatol appar mot"] = "Ortopedia e traumatologia dell'apparato motore",
  ["ortop traumatol rehabil"] = "Ortopedia, traumatologia, rehabilitacja",
  ["ortop travmatol protez"] = "Ortopediia travmatologiia i protezirovanie",
  ["orv hetil"] = "Orvosi hetilap",
  ["orvosok lapja nepeu"] = "Orovosok lapja (Magyar Orvosok Szabad Szakszervezete)",
  ["orvostort kozl"] = "Orvostörténeti közlemények",
  ["oryza (cuttack)"] = "Oryza",
  ["osa contin"] = "OSA continuum",
  ["osa continuum"] = "OSA Continuum",
  ["osa tech dig online"] = "OSA Technical Digest (online)",
  ["osaka city med j"] = "Osaka city medical journal",
  ["osaka daigaku shigaku zasshi"] = "[Osaka Daigaku shigaku zasshi] The journal of Osaka University Dental Society",
  ["osaka econ pap"] = "Osaka Economic Papers",
  ["osaka j math"] = "Osaka Journal of Mathematics",
  ["osaka univ law rev"] = "Osaka University law review",
  ["osgoode hall law j"] = "Osgoode Hall law journal",
  ["osiris (2)"] = "Osiris",
  ["osler libr newsl"] = "Osler Library newsletter",
  ["osler libr stud hist med"] = "Osler Library studies in the history of medicine",
  ["osnabruck schrift math reihe d math-didakt manuskr"] = "Osnabrucker Schriften zur Mathematik. Reihe D: Mathematisch-Didaktische Manuskripte",
  ["osnabruck schrift math reihe i inform"] = "Osnabrucker Schriften zur Mathematik. Reihe I: Informatik",
  ["osnabruck schrift math reihe p preprints"] = "Osnabrucker Schriften zur Mathematik. Reihe P: Preprints",
  ["osnabruck schrift math reihe u mater mathematikunterr"] = "Osnabrucker Schriften zur Mathematik. Reihe U: Materialien zum Mathematikunterricht",
  ["osnabrück stud math"] = "Osnabrücker Studien zur Mathematik",
  ["osong public health res perspect"] = "Osong public health and research perspectives",
  ["osp ital chir"] = "Ospedali d'Italia - chirurgia",
  ["osp maggiore"] = "L' Ospedale maggiore",
  ["osp psichiatr"] = "L'Ospedale psichiatrico",
  ["oss rom"] = "Osservatore romano (English ed.)",
  ["osteoarthr cartil open"] = "Osteoarthritis and cartilage open",
  ["osteoarthritis cartilage"] = "Osteoarthritis and cartilage",
  ["osteopath hosp"] = "The Osteopathic hospital",
  ["osteopath hosp leadersh"] = "Osteopathic hospital leadership",
  ["osteopath mag"] = "The Osteopathic magazine",
  ["osteopath med prim care"] = "Osteopathic medicine and primary care",
  ["osteopath med surg"] = "Osteopathic medicine and surgery",
  ["osteopath prof"] = "The Osteopathic profession",
  ["osteoporos int"] = "Osteoporosis international : a journal established as result of cooperation between the European Foundation for Osteoporosis and the National Osteoporosis Foundation of the USA",
  ["osteoporos sarcopenia"] = "Osteoporosis and sarcopenia",
  ["osterr apoth ztg"] = "Oesterreichische Apotheker Zeitung",
  ["osterr arzteztg"] = "Österreichische Ärztezeitung",
  ["osterr bot z"] = "Osterreichische Botanische Zeitschrift",
  ["osterr chem z"] = "Osterreichische Chemie Zeitschrift",
  ["osterr dent rev"] = "Österreichische Dental-Revue",
  ["osterr dent z"] = "Oesterreichische Dentisten Zeitschrift",
  ["osterr gesch lit geogr"] = "Österreich in Geschichte und Literatur mit Geographie",
  ["osterr hebammenztg"] = "Osterreichische Hebammenzeitung",
  ["osterr kneipp mag"] = "Osterreichische Kneipp-Magazin",
  ["osterr krankenpflegez"] = "Osterreichische Krankenpflegezeitschrift",
  ["osterr mil z"] = "Österreichische Militärische Zeitschrift",
  ["osterr osth"] = "Österreichische Osthefte",
  ["osterr pflegezeitschrift"] = "Österreichische Pflegezeitschrift : Organ des Österreichischen Gesundheits- und Krankenpflegeverbands",
  ["osterr schwesternztg"] = "Osterreichische Schwesternzeitung",
  ["osterr z erforsch bekampf krebskr"] = "Osterreichische Zeitschrift fur Erforschung und Bekampfung der Krebskrankheit",
  ["osterr z geschwiss"] = "Österreichische Zeitschrift für Geschichtswissenschaften : ÖZG",
  ["osterr z kinderheilkd kinderfuersorge"] = "Österreichische Zeitschrift für Kinderheilkunde und Kinderfürsorge",
  ["osterr z onkol"] = "Österreichische Zeitschrift für Onkologie. Austrian journal of oncology",
  ["osterr z pilzkd"] = "Österreichische Zeitschrift für Pilzkunde",
  ["osterr z stat inform"] = "Österreichische Zeitschrift für Statistik und Informatik",
  ["osterr z stomatol"] = "Osterreichische Zeitschrift fur Stomatologie",
  ["osterr zahnarzteztg"] = "Österreichische Zahnärzts-Zeitung; Mitteilungen der Bundesfachgruppe für Zahnheilkunde der Österreichischen Ärztekammer",
  ["osterr zahnprothet"] = "Osterreichische Zahnprothetik",
  ["osterr zahntech handw"] = "Das Österreichische Zahntechniker Handwerk : offizielles Organ der Bundesinnung der Zahntechniker",
  ["osterr zahntech handwerk"] = "Osterreichische Zahntechniker Handwerk",
  ["osterreichische akademie der wissenschaften, philosophisch-historische klasse, sitzungsberichte, 252 band, 5 abhandl"] = "Osterreichische Akademie der Wissenschaften, Philosophisch-Historische Klasse, Sitzungsberichte, 252. Band, 5. Abhandlung; Veroffentlichungen der Kommission fur Geschichte der Erziehung und des Unterrichts",
  ["osterreichische akademie der wissenschaftenkommission fur byzantinistikinstitut fur by"] = "Osterreichische Akademie der Wissenschaften.Kommission fur Byzantinistik.Institut fur Byzantinistik der Universitat Wien",
  ["osterreichische pflegezeitschrift"] = "Osterreichische Pflegezeitschrift : Organ des Osterreichischen Gesundheits- und Krankenpflegeverbands",
  ["ostomy wound manage"] = "Ostomy/wound management",
  ["ot pract"] = "OT practice",
  ["ota int"] = "OTA international : the open access journal of orthopaedic trauma",
  ["otago law rev"] = "Otago law review",
  ["otbor i obrabotka informatsii"] = "Akademiya Nauk Ukrainy",
  ["otecestv arh"] = "Otechestvennye arkhivy",
  ["otechestvennaia istor"] = "Otechestvennai︠a︡ istorii︠a︡",
  ["otjr (thorofare n j)"] = "OTJR : occupation, participation and health",
  ["oto open"] = "OTO open",
  ["otol neurotol"] = "Otology & neurotology : official publication of the American Otological Society, American Neurotology Society [and] European Academy of Otology and Neurotology",
  ["otolaryngol (sunnyvale)"] = "Otolaryngology (Sunnyvale, Calif.)",
  ["otolaryngol clin north am"] = "Otolaryngologic clinics of North America",
  ["otolaryngol head neck surg"] = "Otolaryngology--head and neck surgery : official journal of American Academy of Otolaryngology-Head and Neck Surgery",
  ["otolaryngol head neck surg (1979)"] = "Otolaryngology and head and neck surgery",
  ["otolaryngol pol"] = "Otolaryngologia polska = The Polish otolaryngology",
  ["otorhinolaryngol danub"] = "Oto-rhino-laryngologia Danubiana",
  ["otorhinolaryngol head neck surg"] = "Otorhinolaryngology-head and neck surgery",
  ["otorhinolaryngolog int"] = "L'Oto-rhino-laryngologie internationale",
  ["otorinolaringol ital"] = "L'Oto-rino-laringologia italiana",
  ["otorinolaringol oftalmol"] = "Oto-rino-laringologia şi oftalmologia",
  ["ottawa law rev"] = "Ottawa law review",
  ["ou daigaku shigakushi"] = "Ōu Daigaku shigakushi",
  ["ou mei yan jiu"] = "Ou Mei yan jiu = EurAmerica",
  ["oudheidkd meded rijksmus oudheiden leiden"] = "Oudheidkundige mededeelingen van het Rijksmuseum van Oudheden te Leiden = Nuntii ex Museo Antiquario Leidensi",
  ["ouest med"] = "Ouest-médical",
  ["our planet"] = "Our planet : the magazine of the United Nations Environment Programme",
  ["outbreak surveill investig rep"] = "Outbreak, surveillance and investigation reports",
  ["outcomes manag"] = "Outcomes management",
  ["outcomes manag nurs pract"] = "Outcomes management for nursing practice",
  ["outlook agric"] = "Outlook on agriculture",
  ["outlook blind teach forum"] = "Outlook for the blind and the teachers forum",
  ["outlook bull south dent soc n j"] = "Outlook and bulletin. Southern Dental Society of New Jersey",
  ["outlook mag"] = "Outlook magazine",
  ["outst contrib log"] = "Outstanding Contributions to Logic",
  ["ouvert philos"] = "Ouverture Philosophique",
  ["overseas postgrad med j"] = "The overseas post-graduate medical journal",
  ["oxf bull econ stat"] = "Oxford bulletin of economics and statistics",
  ["oxf class texts phys sci"] = "Oxford Classic Texts in the Physical Sciences",
  ["oxf econ pap"] = "Oxford economic papers",
  ["oxf ger stud"] = "Oxford German studies",
  ["oxf grad texts"] = "Oxford Graduate Texts",
  ["oxf grad texts math"] = "Oxford Graduate Texts in Mathematics",
  ["oxf j leg stud"] = "Oxford journal of legal studies",
  ["oxf master ser phys"] = "Oxford Master Series in Physics",
  ["oxf master ser stat comput theor phys"] = "Oxford Master Series in Statistical Computational, and Theoretical Physics",
  ["oxf med case reports"] = "Oxford medical case reports",
  ["oxf med sch gaz"] = "Oxford Medical School gazette",
  ["oxf open immunol"] = "Oxford open immunology",
  ["oxf philos monogr"] = "Oxford Philosophical Monographs",
  ["oxf read philos"] = "Oxford Readings in Philosophy",
  ["oxf rev econ policy"] = "Oxford review of economic policy",
  ["oxf rev educ"] = "Oxford review of education",
  ["oxf rev reprod biol"] = "Oxford reviews of reproductive biology",
  ["oxf ser mater model"] = "Oxford Series on Materials Modelling",
  ["oxf ser soc cogn soc neurosci"] = "Oxford series in social cognition and social neuroscience",
  ["oxf slav pap"] = "Oxford Slavonic papers",
  ["oxf stud anc philos"] = "Oxford studies in ancient philosophy",
  ["oxf stud philos sci"] = "Oxford Studies in Philosophy of Science",
  ["oxf stud theor linguist"] = "Oxford Studies in Theoretical Linguistics",
  ["oxf surv eukaryot genes"] = "Oxford surveys on eukaryotic genes",
  ["oxford appl math comput sci ser"] = "Oxford Applied Mathematics and Computing Science Series",
  ["oxford bull econ statist"] = "Oxford Bulletin of Economics and Statistics",
  ["oxford devel stud"] = "Oxford Development Studies",
  ["oxford econ pap"] = "Oxford Economic Papers",
  ["oxford eng sci ser"] = "Oxford Engineering Science Series",
  ["oxford j archaeol"] = "Oxford Journal of Archaeology",
  ["oxford lecture ser math appl"] = "Oxford Lecture Series in Mathematics and its Applications",
  ["oxford logic guides"] = "Oxford Logic Guides",
  ["oxford math monogr"] = "Oxford Mathematical Monographs",
  ["oxford open clim change"] = "Oxford Open Climate Change",
  ["oxford open digital health"] = "Oxford Open Digital Health",
  ["oxford open energy"] = "Oxford Open Energy",
  ["oxford open immunol"] = "Oxford Open Immunology",
  ["oxford open infrastruct health"] = "Oxford Open Infrastructure and Health",
  ["oxford open mater sci"] = "Oxford Open Materials Science",
  ["oxford open neurosci"] = "Oxford Open Neuroscience",
  ["oxford philos concepts"] = "Oxford Philosophical Concepts",
  ["oxford rev econ pol"] = "Oxford Review of Economic Policy",
  ["oxford sci publ"] = "Oxford Science Publications",
  ["oxford ser realism quantum phys"] = "Oxford Series on Realism in Quantum Physics",
  ["oxford statist sci ser"] = "Oxford Statistical Science Series",
  ["oxford stud probab"] = "Oxford Studies in Probability",
  ["oxid antioxid med sci"] = "Oxidants and antioxidants in medical science",
  ["oxid commun"] = "Oxidation Communications",
  ["oxid med cell longev"] = "Oxidative medicine and cellular longevity",
  ["oxid med cell longevity"] = "Oxidative Medicine and Cellular Longevity",
  ["oxid met"] = "Oxidation of Metals",
  ["oyo butsuri"] = "Oyo Butsuri",
  ["oyo yakuri"] = "Oyo Yakuri",
  ["ozone sci eng"] = "Ozone Science and Engineering",
  ["ozone: sci eng"] = "Ozone: Science and Engineering",
  ["ozs osterr z soziol"] = "ÖZS, Österreichische Zeitschrift für Soziologie",
  ["o’zb kim j"] = "O’zbekiston Kimyo Jurnali",
  ["p n g med j"] = "Papua and New Guinea medical journal",
  ["p r enferm"] = "Puerto Rico y su enfermera",
  ["p r health sci j"] = "Puerto Rico health sciences journal",
  ["p s q"] = "P & S quarterly",
  ["p t"] = "P & T : a peer-reviewed journal for formulary management",
  ["p v expo discuss congr soc int chir"] = "Procès-verbaux, exposés et discussions [du] congrès de la Société internationale de chirurgie. International Society of Surgery",
  ["p-adic numbers ultrametric anal appl"] = "p-Adic Numbers, Ultrametric Analysis, and Applications",
  ["pa archaeol"] = "Pennsylvania archaeologist",
  ["pa bar assoc q"] = "Pennsylvania Bar Association quarterly. Pennsylvania Bar Association",
  ["pa dent j (harrisb)"] = "Pennsylvania dental journal",
  ["pa dist cty rep"] = "Pennsylvania district and county reports",
  ["pa folklife"] = "Pennsylvania folklife",
  ["pa geogr"] = "The Pennsylvania geographer",
  ["pa health"] = "Pennsylvania's health : official journal of the Pennsylvania State Dept. of Health",
  ["pa health you"] = "Pennsylvania health & you",
  ["pa herit"] = "Pennsylvania heritage (1974)",
  ["pa hist"] = "Pennsylvania history",
  ["pa hosp bull"] = "The Pennsylvania Hospital bulletin",
  ["pa j"] = "The P. A. journal",
  ["pa mag hist biogr"] = "The Pennsylvania magazine of history and biography",
  ["pa med"] = "Pennsylvania medicine",
  ["pa med j"] = "Pennsylvania medical journal (1928)",
  ["pa mennon herit"] = "Pennsylvania Mennonite heritage",
  ["pa nurse"] = "The Pennsylvania nurse",
  ["pa pharm"] = "The Pennsylvania pharmacist",
  ["paa pattern anal appl"] = "PAA. Pattern Analysis and Applications",
  ["pac aff"] = "Pacific affairs",
  ["pac aids alert bull"] = "Pacific AIDS alert bulletin",
  ["pac asia inq"] = "Pacific Asia inquiry : multidisciplinary perspectives",
  ["pac conserv biol"] = "Pacific conservation biology : a journal devoted to conservation and land management in the Pacific region",
  ["pac health dialog"] = "Pacific health dialog",
  ["pac hist"] = "The Pacific historian",
  ["pac hist rev"] = "Pacific historical review",
  ["pac j appl math"] = "Pacific Journal of Applied Mathematics",
  ["pac j math"] = "Pacific journal of mathematics",
  ["pac j math ind"] = "Pacific Journal of Mathematics for Industry",
  ["pac j med sci"] = "Pacific journal of medical sciences",
  ["pac j optim"] = "Pacific Journal of Optimization. An International Journal",
  ["pac law j"] = "Pacific law journal (Sacramento, Calif.)",
  ["pac med surg"] = "Pacific medicine and surgery",
  ["pac northwest"] = "The Pacific northwesterner",
  ["pac northwest fungi"] = "Pacific northwest fungi",
  ["pac northwest q"] = "Pacific Northwest quarterly",
  ["pac philos q"] = "Pacific philosophical quarterly",
  ["pac rep second ser"] = "Pacific reporter. Second series",
  ["pac rim int j nurs res thail"] = "Pacific Rim international journal of nursing research",
  ["pac sci"] = "Pacific science",
  ["pac sci rev"] = "Pacific Science Review",
  ["pac sci rev a: nat sci eng"] = "Pacific Science Review A: Natural Science and Engineering",
  ["pac sociol rev"] = "Pacific sociological review",
  ["pac stud"] = "Pacific studies",
  ["pac symp biocomput"] = "Pacific Symposium on Biocomputing. Pacific Symposium on Biocomputing",
  ["pac viewp"] = "Pacific viewpoint",
  ["pace law rev"] = "Pace law review",
  ["pachart hist astronom ser"] = "Pachart History of Astronomy Series",
  ["pacific asian j energy"] = "Pacific and Asian Journal of Energy",
  ["pacific econ bull"] = "Pacific Economic Bulletin",
  ["pacific econ rev"] = "Pacific Economic Review",
  ["pacific j math"] = "Pacific Journal of Mathematics",
  ["pacific-basin finance j"] = "Pacific-Basin Finance Journal",
  ["pacing clin electrophysiol"] = "Pacing and clinical electrophysiology : PACE",
  ["packag technol sci"] = "Packaging Technology and Science",
  ["packag, transp, storage secur radioact mater"] = "Packaging, Transport, Storage & Security of Radioactive Material",
  ["paddy water environ"] = "Paddy and Water Environment",
  ["padiatr grenzgeb"] = "Padiatrie und Grenzgebiete",
  ["padiatr padol"] = "Padiatrie und Padologie",
  ["padiatr padol suppl"] = "Padiatrie und Padologie. Supplementum",
  ["paedagog hist"] = "Paedagogica historica",
  ["paediatr anaesth"] = "Paediatric anaesthesia",
  ["paediatr child health"] = "Paediatrics & child health",
  ["paediatr child health (oxford)"] = "Paediatrics and child health",
  ["paediatr child health can"] = "Paediatrics And Child Health Canada",
  ["paediatr drugs"] = "Paediatric drugs",
  ["paediatr indones"] = "Paediatrica Indonesiana",
  ["paediatr int child health"] = "Paediatrics and international child health",
  ["paediatr jpn"] = "Paediatria Japonica",
  ["paediatr nurs"] = "Paediatric nursing",
  ["paediatr perinat drug ther"] = "Paediatric and perinatal drug therapy",
  ["paediatr perinat epidemiol"] = "Paediatric and perinatal epidemiology",
  ["paediatr respir rev"] = "Paediatric respiratory reviews",
  ["paediatr univ tokyo"] = "Paediatria Universitatis Tokyo",
  ["pag bull"] = "PAG bulletin",
  ["pageoph top vol"] = "Pageoph Topical Volumes",
  ["pages doc"] = "Pages documentaires",
  ["pagine stor med"] = "Pagine di storia della medicina",
  ["pagine stor med [collana]"] = "Pagine di storia della medicina. Collana",
  ["pagine stor med collana"] = "Pagine di storia della medicina. Collana",
  ["pahlavi med j"] = "Pahlavi medical journal",
  ["paho today"] = "PAHO today : the newsletter of the Pan American Health Organization",
  ["paideuma (wiesb)"] = "Paideuma",
  ["pain clin updates"] = "Pain clinical updates",
  ["pain headache"] = "Pain and headache",
  ["pain manag"] = "Pain management",
  ["pain manag nurs"] = "Pain management nursing : official journal of the American Society of Pain Management Nurses",
  ["pain med"] = "Pain medicine (Malden, Mass.)",
  ["pain physician"] = "Pain physician",
  ["pain pract"] = "Pain practice : the official journal of World Institute of Pain",
  ["pain rep"] = "Pain reports",
  ["pain res manag"] = "Pain research & management",
  ["pain res manage"] = "Pain Research and Management",
  ["pain res treat"] = "Pain research and treatment",
  ["pain suppl"] = "Pain. Supplement",
  ["pain ther"] = "Pain and therapy",
  ["pak dent rev"] = "The Pakistan dental review",
  ["pak dev rev"] = "Pakistan development review",
  ["pak econ soc rev"] = "Pakistan economic and social review",
  ["pak gulf econ"] = "Pakistan & Gulf economist",
  ["pak j agric sci"] = "Pakistan journal of agricultural sciences",
  ["pak j anal chem"] = "Pakistan Journal of Analytical Chemistry",
  ["pak j anal environ chem"] = "Pakistan Journal of Analytical & Environmental Chemistry",
  ["pak j biochem mol biol"] = "Pakistan Journal of Biochemistry and Molecular Biology",
  ["pak j biol sci"] = "Pakistan journal of biological sciences : PJBS",
  ["pak j bot"] = "Pakistan Journal of Botany",
  ["pak j fam plann"] = "Pakistan journal of family planning",
  ["pak j health"] = "Pakistan journal of health",
  ["pak j life soc sci"] = "Pakistan journal of life and social sciences",
  ["pak j med res"] = "Pakistan journal of medical research",
  ["pak j med sci"] = "Pakistan journal of medical sciences",
  ["pak j nematol"] = "Pakistan journal of nematology : an official publication of Pakistan Society of Nematologists",
  ["pak j nutr"] = "Pakistan journal of nutrition : PJN",
  ["pak j pharm sci"] = "Pakistan journal of pharmaceutical sciences",
  ["pak j pharmacol"] = "Pakistan Journal of Pharmacology",
  ["pak j psychol"] = "Pakistan journal of psychology",
  ["pak j sci"] = "Pakistan journal of science",
  ["pak j sci ind res"] = "Pakistan journal of scientific and industrial research",
  ["pak j sci res"] = "Pakistan Journal of Scientific Research",
  ["pak j stat oper res"] = "Pakistan Journal of Statistics and Operation Research",
  ["pak j surg gynaecol obstet"] = "The Pakistan journal of surgery, gynaecology, and obstetrics",
  ["pak j zool"] = "Pakistan journal of zoology",
  ["pak nurs health rev"] = "The Pakistan nursing and health review",
  ["pak pediatr j"] = "Pakistan pediatric journal",
  ["pak popul rev"] = "Pakistan population review",
  ["pak vet j"] = "Pakistan veterinary journal",
  ["pakistan devel rev"] = "Pakistan Development Review",
  ["pakistan econ soc rev"] = "Pakistan Economic and Social Review",
  ["pakistan j appl econ"] = "Pakistan Journal of Applied Economics",
  ["pakistan j statist"] = "Pakistan Journal of Statistics",
  ["palaentolog z"] = "Palaeontologische Zeitschrift",
  ["palaeobiodivers palaeoenviron"] = "Palaeobiodiversity and Palaeoenvironments",
  ["palaeodiversity (stuttg)"] = "Palaeodiversity",
  ["palaeogeogr palaeoclimatol palaeoecol"] = "Palaeogeography, palaeoclimatology, palaeoecology",
  ["palaeontogr abt a palazool stratigr"] = "Palaeontographica Abteilung A Palaozoologie Stratigraphie",
  ["palaeontogr abt b palaeophytol"] = "Palaeontographica Abteilung B Palaeophytologie",
  ["palaeontol electronica"] = "Palaeontologia electronica (Online)",
  ["palaontol z"] = "Paläontologische zeitschrift",
  ["palarch’s j  vertebr palaeontol"] = "PalArch’s Journal of Vertebrate Palaeontology",
  ["palarch’s j vertebr palaeontol"] = "PalArch’s Journal of Vertebrate Palaeontology",
  ["palawan sci"] = "The Palawan scientist",
  ["paleobiodivers paleoenviron"] = "Palaeobiodiversity and palaeoenvironments",
  ["paleoceanogr paleoclimatol"] = "Paleoceanography and paleoclimatology",
  ["paleoclim res"] = "Paleoclimate Research",
  ["paleontol bull"] = "Paleontological Bulletin",
  ["paleontol contrib"] = "Paleontological Contributions",
  ["paleontol j"] = "Paleontological Journal",
  ["paleontol mex"] = "Paleontologia Mexicana",
  ["paleontol res"] = "Paleontological Research",
  ["paleontol soc pap"] = "Paleontological Society Papers",
  ["paleontolog j"] = "Paleontological Journal",
  ["paleopathol assoc monogr"] = "Paleopathology Association monograph",
  ["paleopathol newsl"] = "Paleopathology newsletter",
  ["palest explor q"] = "Palestine exploration quarterly",
  ["palest j math"] = "Palestine Journal of Mathematics",
  ["palgrave commun"] = "Palgrave communications",
  ["palgrave philos today"] = "Palgrave Philosophy Today",
  ["palgrave pivots sports econ"] = "Palgrave Pivots in Sports Economics",
  ["palgrave stud educ philos theory"] = "Palgrave Studies in Educational Philosophy and Theory",
  ["palgrave stud hist sci technol"] = "Palgrave Studies in the History of Science and Technology",
  ["palgrave stud lit sci med"] = "Palgrave Studies in Literature, Science and Medicine",
  ["palgrave stud sustain environ macroecon"] = "Palgrave Studies in Sustainability, Environment and Macroeconomics",
  ["palgrave texts econom"] = "Palgrave Texts in Econometrics",
  ["palimpsest (iowa city)"] = "The Palimpsest",
  ["palliat care"] = "Palliative care",
  ["palliat care soc pract"] = "Palliative care and social practice",
  ["palliat med"] = "Palliative medicine",
  ["palliat med care"] = "Palliative medicine & care : open access",
  ["palliat med rep"] = "Palliative medicine reports",
  ["palliat support care"] = "Palliative & supportive care",
  ["pam lewis carroll"] = "The Pamphlets of Lewis Carroll",
  ["pamm proc appl math mech"] = "PAMM. Proceedings in Applied Mathematics and Mechanics",
  ["pan afr j"] = "Pan-African journal",
  ["pan afr med j"] = "The Pan African medical journal",
  ["pan am health"] = "Pan American health",
  ["pan am med womans j"] = "Pan American medical woman's journal",
  ["pan-pac entomol"] = "The Pan-Pacific entomologist",
  ["panam j aquat sci"] = "Pan-American journal of aquatic sciences",
  ["panamer math j"] = "PanAmerican Mathematical Journal",
  ["pancreas (fairfax)"] = "Pancreas (Fairfax, Va.)",
  ["pancreat disord ther"] = "Pancreatic disorders & therapy",
  ["panhand-plains hist rev"] = "Panhandle-Plains historical review",
  ["panminerva med"] = "Panminerva medica",
  ["panor syntheses"] = "Panoramas et Syntheses",
  ["panor synthèses"] = "Panoramas et Synthèses",
  ["panstwo prawo"] = "Państwo i prawo : organ Zrzeszenia Prawników Demokratów w Polsce",
  ["pap - geol surv can"] = "Paper - Geological Survey of Canada",
  ["pap appl geogr"] = "Papers in Applied Geography",
  ["pap avulsos zool"] = "Papéis avulsos de zoologia",
  ["pap bibliogr soc am"] = "Papers. Bibliographical Society of America",
  ["pap bibliogr soc can"] = "Papers of the Bibliographical Society of Canada. Cahiers de la Société bibliographique du Canada. Bibliographical Society of Canada",
  ["pap br sch rome"] = "Papers of the British School at Rome",
  ["pap celul"] = "Papir a Celuloza",
  ["pap econ española"] = "Papeles de Economía Española",
  ["pap far east hist"] = "Papers on Far Eastern history",
  ["pap fr seventeenth century lit"] = "Papers on French seventeenth century literature",
  ["pap kroeber anthropol soc"] = "Papers. Kroeber Anthropological Society",
  ["pap lab tree ring res"] = "Papers of the Laboratory of Tree Ring Research",
  ["pap meteorol geophys"] = "Papers in Meteorology and Geophysics",
  ["pap mich acad sci arts lett"] = "Papers of the Michigan Academy of Science, Arts and Letters",
  ["pap natl conf prof nurses physicians"] = "Papers. National Conference for Professional Nurses and Physicians",
  ["pap non-market dec making"] = "Papers on Non-Market Decision Making",
  ["pap palaeontol"] = "Papers in palaeontology",
  ["pap phys oceanogr meteorol"] = "Papers in Physical Oceanography and Meteorology",
  ["pap present workshop evol martian atmos"] = "Papers presented to the Workshop on the Evolution of the Martian Atmosphere : Kona, Hawaii, June 29-July 1, 1992",
  ["pap proc"] = "Papers and proceedings. Royal Society of Tasmania",
  ["pap puu"] = "Paperi Ja Puu",
  ["pap reg sci"] = "Papers in regional science : the journal of the Regional Science Association International",
  ["pap reg sci assoc"] = "Papers. Regional Science Association. Meeting",
  ["pap ser united hosp fund n y"] = "Paper series (United Hospital Fund of New York)",
  ["pap slov stud"] = "Papers in Slovene studies",
  ["papeles poblac"] = "Papeles de población",
  ["papillomavirus res"] = "Papillomavirus research (Amsterdam, Netherlands)",
  ["paracelsus arch prakt med"] = "Paracelsus; Archiv der praktischen Medizin",
  ["paragraph (mod crit theory group)"] = "Paragraph (Cambridge, England)",
  ["parallel algorithms appl"] = "Parallel Algorithms and Applications",
  ["parallel comput"] = "Parallel Computing",
  ["parallel distrib comput ser"] = "Parallel and Distributed Computing Series",
  ["parallel process lett"] = "Parallel Processing Letters",
  ["parallel robots: theory appl"] = "Parallel Robots: Theory and Applications",
  ["paramed int"] = "Paramedics international",
  ["parasit vectors"] = "Parasites & vectors",
  ["parasite epidemiol control"] = "Parasite epidemiology and control",
  ["parasite immunol"] = "Parasite immunology",
  ["parasites vectors"] = "Parasites and Vectors",
  ["parasitol día"] = "Parasitología al día",
  ["parasitol int"] = "Parasitology international",
  ["parasitol latinoam"] = "Parasitología latinoamericana",
  ["parasitol open"] = "Parasitology open",
  ["parasitol res"] = "Parasitology research",
  ["parasitol today"] = "Parasitology today (Personal ed.)",
  ["parent sci pract"] = "Parenting, science and practice",
  ["parents mag"] = "Parents' magazine (Bergenfield, N.J.)",
  ["parents mag (ny)"] = "Parents magazine (New York, N.Y. : 1985)",
  ["paris med"] = "Paris médical",
  ["parisankhyan samikkha"] = "Parisankhyan Samikkha",
  ["parivar ayojan"] = "Parivar ayojan",
  ["park sci"] = "Park science",
  ["parkinsonism relat disord"] = "Parkinsonism & related disorders",
  ["parkinsons dis"] = "Parkinson's disease",
  ["parliam aff"] = "Parliamentary affairs",
  ["parma med"] = "Parma medica",
  ["parnass arch"] = "Parnassiana archives",
  ["parodontol acad rev"] = "Parodontologie and academy review",
  ["parodontol stomatol (nuova)"] = "Parodontologia e stomatologia (nuova) : organo ufficiale dell'ARPA italiana, della Società italiana Jonoforesi stomatologica e della Accademia ligustica di stomatologia",
  ["paroi arterielle"] = "Paroi artérielle",
  ["parola passato"] = "La parola del passato",
  ["part a news"] = "Part A news : independent news to legally maximize Medicare part A dollars",
  ["part b news"] = "Part B news",
  ["part fibre toxicol"] = "Particle and fibre toxicology",
  ["part j"] = "Participant journal",
  ["part part syst char"] = "Particle and Particle Systems Characterization",
  ["part part syst charact"] = "Particle & particle systems characterization : measurement and description of particle properties and behavior in powders and other disperse systems",
  ["part sci technol"] = "Particulate Science and Technology",
  ["part technol ser"] = "Particle Technology Series",
  ["partial differ equ appl"] = "Partial Differential Equations and Applications",
  ["partial differ equ meas theory"] = "Partial Differential Equations and Measure Theory",
  ["partial differ equations appl"] = "Partial Differential Equations and Applications",
  ["partial differ equations appl math"] = "Partial Differential Equations in Applied Mathematics",
  ["particles and fields subser"] = "Particles and Fields Subseries",
  ["partis rev"] = "Partisan review (New York, N.Y. : 1936)",
  ["partner abuse"] = "Partner abuse",
  ["parttort kozl"] = "Párttörténeti közlemények. Magyar Szocialista Munkáspárt. Központi Bizottság. Párttörténeti Intézet",
  ["party politics"] = "Party politics",
  ["pas rep"] = "PAS reporter",
  ["pasonariti kenkyu"] = "Pāsonariti kenkyū",
  ["passato presente"] = "Passato e presente (Florence, Italy)",
  ["passau schr psychologiegesch"] = "Passauer Schriften zur Psychologiegeschichte",
  ["past global changes mag"] = "Past Global Changes Magazine",
  ["past imperfect"] = "Past imperfect (Edmonton, Alta.)",
  ["past present"] = "Past & present",
  ["pastor care educ"] = "Pastoral care in education",
  ["pastoral psychol"] = "Pastoral psychology",
  ["patch based tech med imaging (2015)"] = "Patch-based techniques in medical imaging : First International Workshop, Patch-MI 2015, held in conjunction with MICCAI 2015, Munich, Germany, October 9, 2015, revised selected papers. Patch-MI (Workshop) (1st : 2015 : Munich, Germany)",
  ["patch based tech med imaging (2016)"] = "Patch-based techniques in medical imaging : Second International Workshop, Patch-MI 2016, held in conjunction with MICCAI 2016, Athens, Greece, October 17, 2016 : proceedings. Patch-MI (Workshop) (2nd : 2016 : Athens, Greece)",
  ["patch based tech med imaging (2017)"] = "Patch-based techniques in medical imaging : third International Workshop, Patch-MI 2017, held in conjunction with MICCAI 2017, Quebec City, QC, Canada, September 14, 2017, Proceedings. Patch-MI (Workshop) (3rd : 2017 : Québec, Québec)",
  ["patch based tech med imaging (2018)"] = "Patch-based techniques in medical imaging : 4th international workshop, Patch-MI 2018, held in conjunction with MICCAI 2018, Granada, Spain, September 20, 2018 : proceedings. Patch-MI (Workshop) (4th : 2018 : Granada, Spain)",
  ["pathobiol aging age relat dis"] = "Pathobiology of aging & age related diseases",
  ["pathobiol aging age-relat dis"] = "Pathobiology of Aging & Age-Related Diseases",
  ["pathobiol annu"] = "Pathobiology annual",
  ["pathog dis"] = "Pathogens and disease",
  ["pathog glob health"] = "Pathogens and global health",
  ["pathog global health"] = "Pathogens and Global Health",
  ["pathog immun"] = "Pathogens & immunity",
  ["pathogenesis (amst)"] = "Pathogenesis : the journal of mechanisms in disease processes",
  ["pathol annu"] = "Pathology annual",
  ["pathol biol"] = "Pathologie et biologie",
  ["pathol biol (paris)"] = "Pathologie-biologie",
  ["pathol case rev"] = "Pathology case reviews",
  ["pathol discov"] = "Pathology discovery",
  ["pathol eur"] = "Pathologia Europaea",
  ["pathol gen"] = "La Pathologie générale",
  ["pathol immunopathol res"] = "Pathology and immunopathology research",
  ["pathol int"] = "Pathology international",
  ["pathol lab med int"] = "Pathology and laboratory medicine international",
  ["pathol microbiol (basel)"] = "Pathologia et microbiologia",
  ["pathol oncol res"] = "Pathology oncology research : POR",
  ["pathol res pract"] = "Pathology, research and practice",
  ["pathol vet"] = "Pathologia veterinaria",
  ["patholog res int"] = "Pathology research international",
  ["pathology (phila)"] = "Pathology (Philadelphia, Pa.)",
  ["pathophysiol haemost thromb"] = "Pathophysiology of haemostasis and thrombosis",
  ["pathw math"] = "Pathways in Mathematics",
  ["patient acc"] = "Patient accounts",
  ["patient care"] = "Patient care",
  ["patient care manag"] = "Patient care management",
  ["patient couns health educ"] = "Patient counselling and health education",
  ["patient educ couns"] = "Patient education and counseling",
  ["patient educ newsl"] = "Patient education newsletter",
  ["patient exp j"] = "Patient experience journal",
  ["patient focus care"] = "Patient-focused care : the health care executive's guide to organizational restructuring",
  ["patient focus care satisf"] = "Patient-focused care and satisfaction",
  ["patient intell"] = "Patient intelligence",
  ["patient prefer adherence"] = "Patient preference and adherence",
  ["patient relat outcome meas"] = "Patient related outcome measures",
  ["patient saf surg"] = "Patient safety in surgery",
  ["patma banasirakan handes"] = "Patma-banasirakan handes",
  ["patna j med"] = "The Patna journal of medicine",
  ["patol clin ostet ginecol"] = "Patologia e clinica ostetrica e ginecologica",
  ["patol fiziol eksp ter"] = "Patologicheskaia fiziologiia i èksperimental'naia terapiia",
  ["patol pol"] = "Patologia polska",
  ["patologia (mex)"] = "Patología",
  ["pattern anal appl"] = "Pattern analysis and applications : PAA",
  ["pattern recognit"] = "Pattern recognition",
  ["pattern recognit image anal"] = "Pattern Recognition and Image Analysis",
  ["pattern recognit lett"] = "Pattern Recognition Letters",
  ["pattern recognit phys"] = "Pattern Recognition in Physics",
  ["patterns (n y)"] = "Patterns (New York, N.Y.)",
  ["patterns prejudice"] = "Patterns of prejudice",
  ["pavlov j biol sci"] = "The Pavlovian journal of biological science",
  ["pavlov j high nerv act"] = "Pavlov journal of higher nervous activity",
  ["pawlow z hohere nerventatigkeit"] = "Pawlow-Zeitschrift für höhere Nerventätigkeit. Deutsche Ausg",
  ["pci j"] = "PCI Journal",
  ["pcr methods appl"] = "PCR methods and applications",
  ["pcr rep"] = "The PCR reporter",
  ["pda j pharm sci technol"] = "PDA Journal of Pharmaceutical Science and Technology",
  ["pdpta 19 (2019)"] = "PDPTA '19 : proceedings of the 2019 International Conference on Parallel & Distributed Processing Techniquess & Applications. International Conference on Parallel and Distributed Processing Techniques and Applications (2019 : Las Vegas, Nev.)",
  ["peabody j educ"] = "PJE. Peabody journal of education",
  ["peace confl"] = "Peace and conflict : journal of peace psychology : the journal of the Division of Peace Psychology of the American Psychological Association",
  ["peace res society int pap"] = "Peace Research Society International Papers",
  ["peace sci society"] = "Peace Science Society (International) Papers",
  ["peasant stud"] = "Peasant studies",
  ["ped health"] = "Pediatric health",
  ["pedacta (cluj-napoca)"] = "PedActa",
  ["pedagog semin j genet psychol"] = "The Pedagogical seminary and journal of genetic psychology",
  ["pedagogy health promot"] = "Pedagogy in health promotion",
  ["pediatr adolesc gynecol"] = "Pediatric and adolescent gynecology : official journal of the International Federation of Infantile and Juvenile Gynecology",
  ["pediatr aids hiv infect"] = "Pediatric AIDS and HIV infection",
  ["pediatr akus ginekol"] = "Pediatrii︠a︡ akusherstvo i ginekologii︠a︡",
  ["pediatr akush ginekol"] = "Pediatriia, Akusherstvo i Ginekologiia",
  ["pediatr alert"] = "Pediatric alert",
  ["pediatr allergy immunol"] = "Pediatric allergy and immunology : official publication of the European Society of Pediatric Allergy and Immunology",
  ["pediatr allergy immunol pulmonol"] = "Pediatric allergy, immunology, and pulmonology",
  ["pediatr am"] = "Pediatria de las Américas",
  ["pediatr ann"] = "Pediatric annals",
  ["pediatr asthma allergy immunol"] = "Pediatric asthma, allergy & immunology",
  ["pediatr blood cancer"] = "Pediatric blood & cancer",
  ["pediatr cardiol"] = "Pediatric cardiology",
  ["pediatr care (wilmington)"] = "Pediatric care (Wilmington, Del.)",
  ["pediatr case rev"] = "Pediatric case reviews (Print)",
  ["pediatr cerrahi derg"] = "Pediatrik cerrahi dergisi",
  ["pediatr clin india"] = "Pediatric clinics of India",
  ["pediatr clin north am"] = "Pediatric clinics of North America",
  ["pediatr crit care med"] = "Pediatric critical care medicine : a journal of the Society of Critical Care Medicine and the World Federation of Pediatric Intensive and Critical Care Societies",
  ["pediatr dent"] = "Pediatric dentistry",
  ["pediatr dent j"] = "Pediatric dental journal : international journal of Japanese Society of Pediatric Dentistry",
  ["pediatr dermatol"] = "Pediatric dermatology",
  ["pediatr dev pathol"] = "Pediatric and developmental pathology : the official journal of the Society for Pediatric Pathology and the Paediatric Pathology Society",
  ["pediatr diabetes"] = "Pediatric diabetes",
  ["pediatr dimens"] = "Pediatric dimensions",
  ["pediatr drugs"] = "Pediatric Drugs",
  ["pediatr emerg care"] = "Pediatric emergency care",
  ["pediatr emerg med pract"] = "Pediatric emergency medicine practice",
  ["pediatr endocrinol diabetes metab"] = "Pediatric endocrinology, diabetes, and metabolism",
  ["pediatr endocrinol rev"] = "Pediatric endocrinology reviews : PER",
  ["pediatr exerc sci"] = "Pediatric exercise science",
  ["pediatr gastroenterol hepatol nutr"] = "Pediatric gastroenterology, hepatology & nutrition",
  ["pediatr health, med ther"] = "Pediatric Health, Medicine and Therapeutics",
  ["pediatr hematol oncol"] = "Pediatric hematology and oncology",
  ["pediatr infect dis"] = "Pediatric infectious disease",
  ["pediatr infect dis j"] = "The Pediatric infectious disease journal",
  ["pediatr int"] = "Pediatrics international : official journal of the Japan Pediatric Society",
  ["pediatr int (roma)"] = "Pediatria internazionale",
  ["pediatr investig"] = "Pediatric investigation",
  ["pediatr listy"] = "Pediatrické listy",
  ["pediatr med"] = "Pediatric medicine (Hong Kong, China)",
  ["pediatr med chir"] = "La Pediatria medica e chirurgica : Medical and surgical pediatrics",
  ["pediatr med prat"] = "La Pediatria del medico pratico",
  ["pediatr neonatal nurs"] = "Pediatrics and neonatal nursing : open journal",
  ["pediatr neonatol"] = "Pediatrics and neonatology",
  ["pediatr nephrol"] = "Pediatric nephrology (Berlin, Germany)",
  ["pediatr neurol"] = "Pediatric neurology",
  ["pediatr neurol briefs"] = "Pediatric neurology briefs",
  ["pediatr neurosci"] = "Pediatric neuroscience",
  ["pediatr neurosurg"] = "Pediatric neurosurgery",
  ["pediatr news"] = "Pediatric news",
  ["pediatr nurs"] = "Pediatric nursing",
  ["pediatr obes"] = "Pediatric obesity",
  ["pediatr oncall"] = "Pediatric oncall",
  ["pediatr panam"] = "Pediatria panamericana",
  ["pediatr pathol"] = "Pediatric pathology",
  ["pediatr pathol lab med"] = "Pediatric pathology & laboratory medicine : journal of the Society for Pediatric Pathology, affiliated with the International Paediatric Pathology Association",
  ["pediatr pathol mol med"] = "Pediatric pathology & molecular medicine",
  ["pediatr pharmacol (new york)"] = "Pediatric pharmacology (New York, N.Y.)",
  ["pediatr phys ther"] = "Pediatric physical therapy : the official publication of the Section on Pediatrics of the American Physical Therapy Association",
  ["pediatr pol"] = "Pediatria polska",
  ["pediatr prat"] = "Pediatria prática",
  ["pediatr pulmonol"] = "Pediatric pulmonology",
  ["pediatr pulmonol suppl"] = "Pediatric pulmonology. Supplement",
  ["pediatr qual saf"] = "Pediatric quality & safety",
  ["pediatr radiol"] = "Pediatric radiology",
  ["pediatr rehabil"] = "Pediatric rehabilitation",
  ["pediatr rep"] = "Pediatric reports",
  ["pediatr res"] = "Pediatric research",
  ["pediatr rev"] = "Pediatrics in review",
  ["pediatr rheumatol online j"] = "Pediatric rheumatology online journal",
  ["pediatr surg int"] = "Pediatric surgery international",
  ["pediatr ther"] = "Pediatrics & therapeutics : current research",
  ["pediatr transplant"] = "Pediatric transplantation",
  ["pediatria (bucur)"] = "Pediatria",
  ["pediatria (napoli)"] = "La Pediatria",
  ["pediatria (rio)"] = "A Pediatria; revista mensal de clinica, patologia e higiene infantil",
  ["pediatria (santiago)"] = "Pediatria",
  ["pediatria (sofia)"] = "Pediatriia",
  ["pediatric health med ther"] = "Pediatric health, medicine and therapeutics",
  ["pediatrie (bucur)"] = "Pediatrie (Bucharest, Romania)",
  ["pedobiologia (jena)"] = "Pedobiologia",
  ["pedod fr"] = "Pedodontie Francaise",
  ["peer peer netw appl"] = "Peer-to-peer networking and applications",
  ["peer rev"] = "Peer review : emerging trends and key debates in undergraduate education",
  ["peer-to-peer networking appl"] = "Peer-to-Peer Networking and Applications",
  ["peerj anal chem"] = "PeerJ Analytical Chemistry",
  ["peerj comput sci"] = "PeerJ Computer Science",
  ["peerj inorg chem"] = "PeerJ Inorganic Chemistry",
  ["peerj life environ"] = "PeerJ Life and Environment",
  ["peerj mater sci"] = "PeerJ Materials Science",
  ["peerj open adv ecol"] = "PeerJ Open Advances in Ecology",
  ["peerj open adv mar biol"] = "PeerJ Open Advances in Marine Biology",
  ["peerj open adv plant sci"] = "PeerJ Open Advances in Plant Science",
  ["peerj open adv zool"] = "PeerJ Open Advances in Zoology",
  ["peerj org chem"] = "PeerJ Organic Chemistry",
  ["peerj phys chem"] = "PeerJ physical chemistry",
  ["peerj prepr"] = "PeerJ preprints",
  ["peking math j"] = "Peking Mathematical Journal",
  ["peking univ ser math"] = "Peking University Series in Mathematics",
  ["pelican news"] = "Pelican news",
  ["pemb pers pap"] = "Pembroke Persian Papers",
  ["pendik vet mikrobiyol derg"] = "Pendik veteriner mikrobiyoloji dergisi",
  ["penn bioeth j"] = "Penn bioethics journal",
  ["penn dent j (phila)"] = "The Penn dental journal",
  ["penn gse perspect urban educ"] = "Penn GSE perspectives on urban education",
  ["penn med"] = "Penn medicine",
  ["pensa iberamer: revista econ política"] = "Pensamiento Iberoamericano: Revista de Economía Política",
  ["pensam educ"] = "Pensamiento educativo : revista de investigación educacional Latinoamericana",
  ["pensee hommes"] = "La Pensée et les hommes",
  ["pensiero med (milano)"] = "Pensiero medico",
  ["pensiero polit"] = "Il Pensiero politico",
  ["people (chicago)"] = "People weekly",
  ["people (wash dc)"] = "People",
  ["people count"] = "People Count",
  ["people nat"] = "People and Nature",
  ["people nat (hoboken)"] = "People and nature (Hoboken, N.J.)",
  ["people place"] = "People and place",
  ["people planet"] = "People & the planet",
  ["pepperdine law rev"] = "Pepperdine law review",
  ["pept res"] = "Peptide research",
  ["pept sci"] = "Peptide Science",
  ["pept sci (hoboken)"] = "Peptide science (Hoboken, N.J.)",
  ["peptide protein rev"] = "Peptide and protein reviews",
  ["peptides (amsterdam, neth)"] = "Peptides (Amsterdam, Netherlands)",
  ["peptides (n y, ny, u s)"] = "Peptides (New York, NY, United States)",
  ["peptides (ny)"] = "Peptides",
  ["per med"] = "Personalized medicine",
  ["percept mot skills"] = "Perceptual and motor skills",
  ["percept psychophys"] = "Perception & psychophysics",
  ["perf latinoam"] = "Perfiles latinoamericanos : revista de la Sede Académica de México de la Facultad Latinoamericana de Ciencias Sociales",
  ["perform enhanc health"] = "Performance enhancement & health",
  ["perform eval"] = "Performance Evaluation",
  ["perform improv"] = "Performance Improvement",
  ["perform improv advis"] = "Performance improvement advisor",
  ["perform res"] = "Performance research",
  ["pergamon stud neurosci"] = "Pergamon Studies in Neuroscience",
  ["perinat care"] = "Perinatal care",
  ["perinatol reprod hum"] = "Perinatología y reproducción humana",
  ["period biol"] = "Periodicum biologorum",
  ["period counc am mil past"] = "Periodical (Council on America's Military Past)",
  ["period math hung"] = "Periodica mathematica Hungarica",
  ["period math hungar"] = "Periodica Mathematica Hungarica. Journal of the Janos Bolyai Mathematical Society",
  ["period polytech mech eng"] = "Periodica Polytechnica",
  ["period polytech, chem eng"] = "Periodica Polytechnica, Chemical Engineering",
  ["period polytech, civ eng"] = "Periodica Polytechnica, Civil Engineering",
  ["period polytech, eng"] = "Periodica Polytechnica, Engineering",
  ["period polytech, mech eng"] = "Periodica Polytechnica, Mechanical Engineering",
  ["periodica polytech, chem eng"] = "Periodica Polytechnica, Chemical Engineering",
  ["periodontal abstr"] = "Periodontal abstracts",
  ["periodontal case rep"] = "Periodontal case reports : a publication of the Northeastern Society of Periodontists",
  ["periodontal clin investig"] = "Periodontal clinical investigations : official publication of the Northeastern Society of Periodontists",
  ["periodontol 2000"] = "Periodontology 2000",
  ["perioper care oper room manag"] = "Perioperative care and operating room management",
  ["perioper med (lond)"] = "Perioperative medicine (London, England)",
  ["perioper nurs q"] = "Perioperative nursing quarterly",
  ["perit dial int"] = "Peritoneal dialysis international : journal of the International Society for Peritoneal Dialysis",
  ["perkins j"] = "Perkins journal",
  ["perm found med bull"] = "Permanente Foundation medical bulletin",
  ["perm j"] = "The Permanente journal",
  ["permafr periglac process"] = "Permafrost and periglacial processes",
  ["permafrost periglacial processes"] = "Permafrost and Periglacial Processes",
  ["permian hist annu"] = "Permian historical annual",
  ["pers adm"] = "The Personnel administrator",
  ["pers bioet"] = "Persona y bioética",
  ["pers derecho"] = "Persona y derecho",
  ["pers guid j"] = "The Personnel and guidance journal",
  ["pers individ dif"] = "Personality and individual differences",
  ["pers j"] = "The Personnel journal",
  ["pers med"] = "Personalized Medicine",
  ["pers med oncol"] = "Personalized medicine in oncology",
  ["pers med psychiatry"] = "Personalized medicine in psychiatry",
  ["pers psychol"] = "Personnel psychology",
  ["pers relatsh"] = "Personal Relationships",
  ["pers soc psychol bull"] = "Personality & social psychology bulletin",
  ["pers soc psychol rev"] = "Personality and social psychology review : an official journal of the Society for Personality and Social Psychology, Inc",
  ["pers ubiquitous comput"] = "Personal and ubiquitous computing",
  ["personal disord"] = "Personality disorders",
  ["personal ment health"] = "Personality and mental health",
  ["personal neurosci"] = "Personality neuroscience",
  ["perspect (am probat parole assoc)"] = "Perspectives (American Probation and Parole Association)",
  ["perspect accredit"] = "Perspectives on accreditation",
  ["perspect addict nurs"] = "Perspectives on addictions nursing : a publication of the National Nurses Society on Addictions",
  ["perspect am hist"] = "Perspectives in American history",
  ["perspect asha spec interest groups"] = "Perspectives of the ASHA special interest groups",
  ["perspect augment altern commun"] = "Perspectives on augmentative and alternative communication",
  ["perspect behav sci"] = "Perspectives on behavior science",
  ["perspect bioet am"] = "Perspectivas bioéticas en las américas",
  ["perspect biol med"] = "Perspectives in biology and medicine",
  ["perspect biosecur"] = "Perspectives in biosecurity",
  ["perspect clin res"] = "Perspectives in clinical research",
  ["perspect dev neurobiol"] = "Perspectives on developmental neurobiology",
  ["perspect drug discov des"] = "Perspectives in drug discovery and design : PD3",
  ["perspect drug discovery des"] = "Perspectives in Drug Discovery and Design",
  ["perspect earth space sci"] = "Perspectives of Earth and Space Scientists",
  ["perspect ecol conserv"] = "Perspectives in ecology and conservation",
  ["perspect fluen fluen disord"] = "Perspectives on fluency and fluency disorders",
  ["perspect form induction revis evol"] = "Perspectives in Formal Induction, Revision and Evolution",
  ["perspect gerontol"] = "Perspectives on gerontology",
  ["perspect health"] = "Perspectives in health : the magazine of the Pan American Health Organization",
  ["perspect health inf manag"] = "Perspectives in health information management",
  ["perspect healthc risk manage"] = "Perspectives in healthcare risk management",
  ["perspect hear hear disord child"] = "Perspectives on hearing and hearing disorders in childhood",
  ["perspect hear hear disord res res diagn"] = "Perspectives on hearing and hearing disorders. Research and research diagnostics",
  ["perspect infirm"] = "Perspective infirmière : revue officielle de l'Ordre des infirmières et infirmiers du Québec",
  ["perspect int planif fam"] = "Perspectivas internacionales en planificación familiar",
  ["perspect internation plan fam"] = "Perspectives internationales du planning familial",
  ["perspect lab income"] = "Perspectives on Labour and Income",
  ["perspect lang learn educ"] = "Perspectives on language learning and education",
  ["perspect lang lit"] = "Perspectives on language and literacy",
  ["perspect log"] = "Perspectives in Logic",
  ["perspect math"] = "Perspectives in Mathematics",
  ["perspect math logic"] = "Perspectives in Mathematical Logic",
  ["perspect med chem"] = "Perspectives in Medicinal Chemistry",
  ["perspect med educ"] = "Perspectives on medical education",
  ["perspect med virol"] = "Perspectives in medical virology",
  ["perspect medicaid manage"] = "Perspectives on medicaid management",
  ["perspect medicaid medicare manage"] = "Perspectives on Medicaid and Medicare management",
  ["perspect medicin chem"] = "Perspectives in medicinal chemistry",
  ["perspect mex am stud"] = "Perspectives in Mexican American studies",
  ["perspect nephrol hypertens"] = "Perspectives in nephrology and hypertension",
  ["perspect neural comput"] = "Perspectives in Neural Computing",
  ["perspect neurophysiol neurogenic speech lang disord"] = "Perspectives on neurophysiology and neurogenic speech and language disorders",
  ["perspect nutr hum"] = "Perspectivas en nutrición humana : órgano de divulgación academica de la Escuela de Nutrición y Dietética de la Universidad de Antioquia",
  ["perspect pediatr pathol"] = "Perspectives in pediatric pathology",
  ["perspect plant ecol evol syst"] = "Perspectives in Plant Ecology, Evolution and Systematics",
  ["perspect psychiatr care"] = "Perspectives in psychiatric care",
  ["perspect psychol sci"] = "Perspectives on psychological science : a journal of the Association for Psychological Science",
  ["perspect public health"] = "Perspectives in public health",
  ["perspect public manag gov"] = "Perspectives on public management and governance",
  ["perspect respir nurs"] = "Perspectives in respiratory nursing : a publication of the Respiratory Nursing Society",
  ["perspect sci"] = "Perspectives in Science",
  ["perspect sci (neth)"] = "Perspectives in science",
  ["perspect sci christ faith"] = "Perspectives on science and Christian faith : journal of the American Scientific Affiliation",
  ["perspect sex reprod health"] = "Perspectives on sexual and reproductive health",
  ["perspect soc work"] = "Perspectives in social work",
  ["perspect soc work (houst)"] = "Perspectives on social work : the journal of the doctoral students of the University of Houston Graduate School of Social Work",
  ["perspect speech sci orofac disord"] = "Perspectives on speech science and orofacial disorders",
  ["perspect undergrad res mentor"] = "Perspectives on undergraduate research and mentoring : PURM",
  ["perspect vasc surg endovasc ther"] = "Perspectives in vascular surgery and endovascular therapy",
  ["perspectives (montclair)"] = "Perspectives",
  ["perspectives (wash)"] = "Perspectives (Washington, D.C. : 1984)",
  ["perspectives prof"] = "Perspectives on the professions",
  ["perspekt mathdidakt"] = "Perspektiven der Mathematikdidaktik",
  ["perspekt razvit vychisl tekh"] = "Perspektivy Razvitiya Vychislitelnoi Tekhniki",
  ["pertanika j sci technol"] = "Pertanika journal of science & technology",
  ["pertanika j trop agric sci"] = "Pertanika Journal of Tropical Agricultural Science",
  ["peruano normas legales"] = "El Peruano : Normas legales :. Peru",
  ["pervasive disp 2017 (2017)"] = "Pervasive Displays 2017 : the 6th ACM International Symposium on Pervasive Displays : proceedings : Lugano, Switzerland, June 7-9, 2017. International Symposium on Pervasive Displays (6th : 2017 : Lugano, Switzerland)",
  ["pervasive mob comput"] = "Pervasive and Mobile Computing",
  ["pesqui agropecu bras"] = "Pesquisa Agropecuária Brasileira",
  ["pesqui agropecu trop"] = "Pesquisa agropecuária tropical",
  ["pesqui bras odontopediatria clin integr"] = "Pesquisa brasileira em odontopediatria e clínica integrada",
  ["pesqui florest bras"] = "Pesquisa florestal brasileira",
  ["pesqui odontol bras"] = "Pesquisa odontológica brasileira = Brazilian oral research",
  ["pesqui planej econ"] = "Pesquisa e planejamento econômico",
  ["pesqui vet bras"] = "Pesquisa veterinária brasileira : revista do Colégio Brasileiro de Patologia Animal = Brazilian journal of veterinary research",
  ["pesquisa planejamento econ"] = "Pesquisa e Planejamento Econômico",
  ["pest control"] = "Pest control",
  ["pest manag sci"] = "Pest management science",
  ["pest manage hortic ecosyst"] = "Pest Management in Horticultural Ecosystems",
  ["pest manage sci"] = "Pest Management Science",
  ["pestic biochem physiol"] = "Pesticide Biochemistry and Physiology",
  ["pestic fitomed"] = "Pesticidi i fitomedicina = Pesticides and phytomedicine",
  ["pestic monit j"] = "Pesticides monitoring journal",
  ["pestic outlook"] = "Pesticide Outlook",
  ["pestic people nat"] = "Pesticides, People and Nature",
  ["pestic sci"] = "Pesticide science",
  ["pet chem"] = "Petroleum Chemistry",
  ["pet clin"] = "PET clinics",
  ["pet coal"] = "Petroleum and Coal",
  ["pet eng"] = "Petroleum Engineering",
  ["pet explor dev"] = "Petroleum Exploration and Development",
  ["pet geosci"] = "Petroleum Geoscience",
  ["pet res"] = "Petroleum Research",
  ["pet sci"] = "Petroleum Science",
  ["pet sci technol"] = "Petroleum Science and Technology",
  ["petermanns geogr mitt"] = "Petermanns geographische Mitteilungen",
  ["petria (roma)"] = "Petria (Roma)",
  ["pezcoller found j"] = "The Pezcoller Foundation journal : news from the Pezcoller Foundation world",
  ["pfca rev"] = "PFCA review",
  ["pfg- j photogramm remote sens geoinf sci"] = "PFG-Journal of Photogrammetry Remote Sensing and Geoinformation Science",
  ["pflanzenschutz-nachr bayer"] = "Pflanzenschutz-Nachrichten Bayer",
  ["pflege aktuell"] = "Pflege aktuell",
  ["pflege z"] = "Pflege Zeitschrift",
  ["pfluegers arch"] = "Pfluegers Archiv",
  ["pfluegers arch gesame physiol menschen tiere"] = "Fluegers Archiv fur die Gesamte Physiologie des Menschen und der Tiere",
  ["pflugers arch"] = "Pflügers Archiv : European journal of physiology",
  ["pflugers arch - eur j physiol"] = "Pflugers Archiv - European Journal of Physiology",
  ["pflugers arch gesamte physiol menschen tiere"] = "Pflügers Archiv für die gesamte Physiologie des Menschen und der Tiere",
  ["pga proc geol assoc"] = "PGA Proceedings of the Geologists’ Association",
  ["pgrsa q"] = "Quarterly (Plant Growth Regulator Society of America)",
  ["phage (new rochelle)"] = "PHAGE (New Rochelle, N.Y.)",
  ["phage: ther appl res"] = "PHAGE: Therapy, Applications, and Research",
  ["pharm acta helv"] = "Pharmaceutica Acta Helvetiae",
  ["pharm anal acta"] = "Pharmaceutica analytica acta",
  ["pharm beih"] = "Die Pharmazie. Beihefte",
  ["pharm biol"] = "Pharmaceutical biology",
  ["pharm bioprocess"] = "Pharmaceutical bioprocessing",
  ["pharm biotechnol"] = "Pharmaceutical biotechnology",
  ["pharm bull"] = "Pharmaceutical bulletin",
  ["pharm chem j"] = "Pharmaceutical Chemistry Journal",
  ["pharm crop"] = "Pharmaceutical crops",
  ["pharm dev regul"] = "Pharmaceutical development and regulation",
  ["pharm dev technol"] = "Pharmaceutical development and technology",
  ["pharm diagn innov"] = "Pharmaceutical & diagnostic innovation",
  ["pharm fr"] = "Pharmacien de France",
  ["pharm front"] = "Pharmaceutical frontiers",
  ["pharm hist"] = "Pharmacy in history",
  ["pharm hist (lond)"] = "Pharmaceutical historian",
  ["pharm hist aust"] = "Pharmacy history Australia : the newsletter of the Australian Academy of the History of Pharmacy",
  ["pharm ind"] = "La Pharmacie industrielle",
  ["pharm int"] = "Pharmacy international",
  ["pharm j"] = "The Pharmaceutical journal",
  ["pharm j n z"] = "The Pharmaceutical journal of New Zealand",
  ["pharm lett"] = "Der Pharmacia Lettre",
  ["pharm manage comb am j pharm"] = "Pharmacy management combined with the American journal of pharmacy : PM",
  ["pharm med"] = "Pharmaceutical Medicine",
  ["pharm methods"] = "Pharmaceutical Methods",
  ["pharm nanotechnol"] = "Pharmaceutical nanotechnology",
  ["pharm outsourcing"] = "Pharmaceutical outsourcing",
  ["pharm pat anal"] = "Pharmaceutical patent analyst",
  ["pharm pharmacol commun"] = "Pharmacy and Pharmacology Communications",
  ["pharm pharmacol int j"] = "Pharmacy & pharmacology international journal",
  ["pharm pharmacol lett"] = "Pharmaceutical and Pharmacological Letters",
  ["pharm pract (granada)"] = "Pharmacy practice",
  ["pharm pract manag q"] = "Pharmacy practice management quarterly",
  ["pharm prax"] = "Pharmazeutische Praxis",
  ["pharm regul aff"] = "Pharmaceutical regulatory affairs : open access",
  ["pharm res"] = "Pharmaceutical Research",
  ["pharm rev"] = "Pharmaceutical reviews",
  ["pharm rundsch"] = "Pharmazeutische Rundschau",
  ["pharm sci technol today"] = "Pharmaceutical Science & Technology Today",
  ["pharm sci technolo today"] = "Pharmaceutical science & technology today",
  ["pharm sin"] = "Pharmacia Sinica",
  ["pharm stat"] = "Pharmaceutical statistics",
  ["pharm technol"] = "Pharmaceutical Technology",
  ["pharm technol hosp pharm"] = "Pharmaceutical technology in hospital pharmacy",
  ["pharm times"] = "Pharmacy times",
  ["pharm today"] = "Pharmacy Today",
  ["pharm unserer zeit"] = "Pharmazie in unserer Zeit",
  ["pharm update"] = "Pharmacy update",
  ["pharm weekbl"] = "Pharmaceutisch weekblad",
  ["pharm weekbl sci"] = "Pharmaceutisch weekblad. Scientific edition",
  ["pharm world sci"] = "Pharmacy world & science : PWS",
  ["pharm zentralhalle dtschl"] = "Pharmazeutische Zentralhalle fur Deutschland",
  ["pharm ztg (1856)"] = "Pharmazeutische Zeitung",
  ["pharma chem"] = "Pharma Chemica",
  ["pharma focus asia"] = "Pharma focus asia",
  ["pharma innov"] = "The pharma innovation",
  ["pharmacal adv"] = "Pharmacal advance",
  ["pharmaceut med"] = "Pharmaceutical medicine",
  ["pharmaceuticals (basel)"] = "Pharmaceuticals (Basel, Switzerland)",
  ["pharmacoecon open"] = "PharmacoEconomics - open",
  ["pharmacoecon outcomes news"] = "PharmacoEconomics & outcomes news",
  ["pharmacoepidemiol drug saf"] = "Pharmacoepidemiology and drug safety",
  ["pharmacogenet genomics"] = "Pharmacogenetics and genomics",
  ["pharmacogenomics j"] = "The pharmacogenomics journal",
  ["pharmacogenomics pers med"] = "Pharmacogenomics and Personalized Medicine",
  ["pharmacogn mag"] = "Pharmacognosy magazine",
  ["pharmacogn res"] = "Pharmacognosy Research",
  ["pharmacogn rev"] = "Pharmacognosy Reviews",
  ["pharmacognosy res"] = "Pharmacognosy research",
  ["pharmacol  biochem behav"] = "Pharmacology, Biochemistry and Behavior",
  ["pharmacol biochem behav"] = "Pharmacology, biochemistry, and behavior",
  ["pharmacol commun"] = "Pharmacology communications",
  ["pharmacol drug dev ther"] = "Pharmacology, drug development & therapeutics",
  ["pharmacol pharm"] = "Pharmacology & pharmacy",
  ["pharmacol physicians"] = "Pharmacology for physicians",
  ["pharmacol rep"] = "Pharmacological reports : PR",
  ["pharmacol res"] = "Pharmacological Research",
  ["pharmacol res commun"] = "Pharmacological Research Communications",
  ["pharmacol res mod chin med"] = "Pharmacological Research - Modern Chinese Medicine",
  ["pharmacol res perspect"] = "Pharmacology research & perspectives",
  ["pharmacol rev"] = "Pharmacological reviews",
  ["pharmacol ther"] = "Pharmacology & therapeutics",
  ["pharmacol ther [b]"] = "Pharmacology and Therapeutics. Part B, General and Systematic Pharmacology",
  ["pharmacol ther b"] = "Pharmacology & therapeutics. Part B: General & systematic pharmacology",
  ["pharmacol ther dent"] = "Pharmacology and therapeutics in dentistry",
  ["pharmacol ther part a"] = "Pharmacology & Therapeutics Part A: Chemotherapy, Toxicology and Metabolic Inhibitors",
  ["pharmacol ther part b"] = "Pharmacology & Therapeutics Part B: General and Systemic Pharmacology",
  ["pharmacol ther part c"] = "Pharmacology & Therapeutics Part C: Clinical Pharmacology and Therapeutics",
  ["pharmacol ther, part a"] = "Pharmacology & Therapeutics, Part A: Chemotherapy, Toxicology and Metabolic Inhibitors",
  ["pharmacol ther, part b"] = "Pharmacology & Therapeutics, Part B: General & Systematic Pharmacology",
  ["pharmacol ther, part c"] = "Pharmacology & Therapeutics, Part C: Clinical Pharmacology and Therapeutics",
  ["pharmacol toxicol"] = "Pharmacology & toxicology",
  ["pharmacol toxicol (copenhagen)"] = "Pharmacology & Toxicology (Copenhagen)",
  ["pharmacol, biochem behav"] = "Pharmacology, Biochemistry and Behavior",
  ["pharmacy (basel)"] = "Pharmacy (Basel, Switzerland)",
  ["pharmakopsychiatr neuropsychopharmakol"] = "Pharmakopsychiatrie, Neuro-Psychopharmakologie",
  ["pharmaziehist bibliogr"] = "Pharmaziehistorische Bibliographie : PhB",
  ["pharmaziehist forsch"] = "Pharmaziehistorische Forschungen",
  ["pharmeur bio sci notes"] = "Pharmeuropa bio & scientific notes",
  ["pharmeur sci notes"] = "Pharmeuropa scientific notes",
  ["pharmeuropa bio"] = "Pharmeuropa bio",
  ["pharmeuropa spec issue biol"] = "Pharmeuropa. Special issue biologicals",
  ["pharmgenomics pers med"] = "Pharmacogenomics and personalized medicine",
  ["pharos alpha omega alpha honor med soc"] = "The Pharos of Alpha Omega Alpha-Honor Medical Society. Alpha Omega Alpha",
  ["phase transit"] = "Phase Transitions",
  ["phase transit crit phenom"] = "Phase Transition and Critical Phenomena",
  ["phase transitions"] = "Phase Transitions",
  ["phc4 fyi"] = "PHC4 FYI",
  ["phenomenol cogn sci"] = "Phenomenology and the cognitive sciences",
  ["phi delta kappan"] = "Phi Delta Kappan",
  ["phi kappa phi j"] = "Phi Kappa Phi journal",
  ["phi lambda kappa q"] = "The Phi Lambda Kappa quarterly",
  ["phil trans r soc a"] = "Philosophical Transactions of the Royal Society A",
  ["phil trans r soc b"] = "Philosophical Transactions of the Royal Society B",
  ["phila health bull"] = "Philadelphia health bulletin",
  ["phila inq"] = "Philadelphia inquirer (Philadelphia, Pa. : 1969)",
  ["phila med"] = "Philadelphia medicine",
  ["phila soc innov j"] = "Philadelphia social innovations journal",
  ["philipp agric sci"] = "The Philippine agricultural scientist",
  ["philipp dev"] = "Philippine development",
  ["philipp econ j"] = "The Philippine economic journal",
  ["philipp entomol"] = "Philippine entomologist",
  ["philipp geogr j"] = "Philippine geographical journal",
  ["philipp j cancer"] = "Philippine journal of cancer",
  ["philipp j cardiol"] = "Philippine journal of cardiology",
  ["philipp j crop sci"] = "The Philippine journal of crop science",
  ["philipp j intern med"] = "Philippine journal of internal medicine",
  ["philipp j ment health"] = "Philippine journal of mental health",
  ["philipp j nurs"] = "The Philippine journal of nursing",
  ["philipp j nutr"] = "Philippine journal of nutrition",
  ["philipp j obstet gynecol"] = "Philippine journal of obstetrics & gynecology : official publication, Philippine Obstetrical and Gynecological Society",
  ["philipp j ophthalmol"] = "Philippine journal of ophthalmology",
  ["philipp j ophthalmol otolaryngol"] = "Philippine journal of ophthalmology and otolaryngology",
  ["philipp j pediatr"] = "The Philippine journal of pediatrics",
  ["philipp j psychol"] = "Philippine journal of psychology",
  ["philipp j public adm"] = "Philippine journal of public administration",
  ["philipp j sci"] = "Philippine journal of science",
  ["philipp j surg"] = "Philippine journal of surgery",
  ["philipp j surg spec"] = "Philippine journal of surgical specialties",
  ["philipp j surg surg spec"] = "Philippine journal of surgery and surgical specialties",
  ["philipp j syst biol"] = "Philippine journal of systematic biology",
  ["philipp j vet anim sci"] = "Philippine journal of veterinary and animal sciences",
  ["philipp law j"] = "Philippine law journal",
  ["philipp med world"] = "The Philippine medical world",
  ["philipp popul j"] = "Philippine population journal",
  ["philipp popul newsl"] = "Philippine Population Newsletter",
  ["philipp q cult soc"] = "Philippine quarterly of culture and society",
  ["philipp rev econ"] = "The Philippine review of economics",
  ["philipp rev econ bus"] = "The Philippine review of economics & business",
  ["philipp sci lett"] = "Philippine science letters",
  ["philipp soc sci humanit rev"] = "Philippine social sciences and humanities review",
  ["philipp sociol rev"] = "Philippine sociological review",
  ["philipp stud"] = "Philippine studies",
  ["philippine econ j"] = "Philippine Economic Journal",
  ["philippine j bus econ"] = "Philippine Journal of Business and Economics",
  ["philippine j bus finance"] = "Philippine Journal of Business and Finance",
  ["philippine rev econ"] = "Philippine Review of Economics",
  ["philippine rev econ bus"] = "Philippine Review of Economics and Business",
  ["philips j res"] = "Philips Journal of Research",
  ["philips res rep"] = "Philips Research Reports",
  ["phillip j"] = "Phillip Journal",
  ["phillip j restaur zahnmed"] = "Phillip Journal für restaurative Zahnmedizin",
  ["philos abhandlungen"] = "Philosophische Abhandlungen",
  ["philos anal"] = "Philosophische Analyse/Philosophical Analysis",
  ["philos antiq"] = "Philosophia Antiqua",
  ["philos bibliothek"] = "Philosophische Bibliothek",
  ["philos compass"] = "Philosophy compass",
  ["philos context"] = "Philosophy in context",
  ["philos cult stud revisit/hist-genet stud philos kultgesch"] = "Philosophy and Cultural Studies Revisited/Historisch-genetische Studien zur Philosophie und Kulturgeschichte",
  ["philos east west"] = "Philosophy East & West",
  ["philos eng technol"] = "Philosophy of Engineering and Technology",
  ["philos ethics humanit med"] = "Philosophy, ethics, and humanities in medicine : PEHM",
  ["philos exch"] = "Philosophic exchange",
  ["philos explor"] = "Philosophical explorations : an international journal for the philosophy of mind and action",
  ["philos forum"] = "The Philosophical forum",
  ["philos j"] = "The Philosophical journal",
  ["philos mag"] = "Philosophical Magazine",
  ["philos mag (abingdon)"] = "Philosophical magazine (Abingdon, England)",
  ["philos mag a"] = "Philosophical Magazine A",
  ["philos mag abingdon engl"] = "Philosophical Mag. Abingdon England",
  ["philos mag b"] = "Philosophical Magazine B",
  ["philos mag lett"] = "Philosophical Magazine Letters",
  ["philos manag"] = "Philosophy of management",
  ["philos math"] = "Philosophia Mathematica",
  ["philos math (3)"] = "Philosophia Mathematica. Series III",
  ["philos mediev"] = "Philosophes Medievaux",
  ["philos nat"] = "Philosophia naturalis",
  ["philos natur"] = "Philosophia Naturalis",
  ["philos perspect"] = "Philosophical Perspectives. A Supplement to Nous",
  ["philos phenomenol res"] = "Philosophy and phenomenological research",
  ["philos psychiatr psychol"] = "Philosophy, psychiatry, & psychology : PPP",
  ["philos psychol"] = "Philosophical psychology",
  ["philos public aff"] = "Philosophy & public affairs",
  ["philos public policy q"] = "Philosophy & public policy quarterly",
  ["philos q"] = "The Philosophical quarterly",
  ["philos quart"] = "The Philosophical Quarterly",
  ["philos res analysis"] = "Philosophic research and analysis",
  ["philos res arch"] = "Philosophy research archives (Bowling Green, Ohio : 1982)",
  ["philos rev"] = "The Philosophical review",
  ["philos sci"] = "Philosophy of science",
  ["philos sci (paris)"] = "Philosophia Scientiae. Travaux d’Histoire des Sciences et de Philosophie",
  ["philos sci eur perspect"] = "The Philosophy of Science in a European Perspective",
  ["philos soc crit"] = "Philosophy & social criticism",
  ["philos soc sci"] = "Philosophy of the social sciences",
  ["philos stud"] = "Philosophical studies",
  ["philos stud ser"] = "Philosophical Studies Series",
  ["philos technol"] = "Philosophy & technology",
  ["philos texte stud"] = "Philosophische Texte und Studien",
  ["philos theol"] = "Philosophy, theology",
  ["philos trans a math phys eng sci"] = "Philosophical transactions. Series A, Mathematical, physical, and engineering sciences",
  ["philos trans phys sci eng"] = "Philosophical transactions. Physical sciences and engineering",
  ["philos trans r soc"] = "Philosophical Transactions of the Royal Society",
  ["philos trans r soc a"] = "Philosophical Transactions of the Royal Society A",
  ["philos trans r soc a: math phys eng sci"] = "Philosophical Transactions of the Royal Society A: Mathematical, Physical and Engineering Sciences",
  ["philos trans r soc lond"] = "Philosophical transactions of the Royal Society of London",
  ["philos trans r soc lond a"] = "Philosophical transactions of the Royal Society of London. A",
  ["philos trans r soc lond b biol sci"] = "Philosophical transactions of the Royal Society of London. Series B, Biological sciences",
  ["philos trans r soc lond ser a math phys eng sci"] = "Philosophical Transactions of the Royal Society of London. Series A. Mathematical, Physical and Engineering Sciences",
  ["philos trans r soc london"] = "Philosophical Transactions of the Royal Society of London",
  ["philos trans r soc london, ser a"] = "Philosophical Transactions of the Royal Society of London, Series A: Mathematical, Physical and Engineering Sciences",
  ["philos trans r soc london, ser b"] = "Philosophical Transactions of the Royal Society of London, Series B: Biological Sciences",
  ["philos trans r soc math phys eng sci"] = "Philosophical Transactions of the Royal Society A: Mathematical, Physical and Engineering Sciences",
  ["philos trans roy soc a"] = "Philosophical Transactions of the Royal Society A. Mathematical, Physical and Engineering Sciences",
  ["philos trans roy soc london"] = "Philosophical Transactions of the Royal Society of London",
  ["philos transact a math phys eng sci"] = "Philosophical transactions. Series A, Mathematical, physical, and engineering sciences",
  ["philosophia (mendoza)"] = "Philosophia",
  ["philosophia (ramat gan)"] = "Philosophia (Ramat-Gan, Israel)",
  ["phispc monogr ser hist philos sci cult india"] = "PHISPC Monograph Series on History of Philosophy, Science and Culture in India",
  ["phlebitis bull"] = "Phlebitis bulletin",
  ["phoebe (oneonta n y)"] = "Phoebe (Oneonta, N.Y.)",
  ["phosphorus sulfur rel elem"] = "Phosphorus Sulfur and Related Elements",
  ["phosphorus sulfur silicon relat elem"] = "Phosphorus, sulfur, and silicon and the related elements",
  ["phosphorus, sulfur silicon relat elem"] = "Phosphorus Sulfur Silicon and the Related Elements",
  ["photobiomodul photomed laser surg"] = "Photobiomodulation, photomedicine, and laser surgery",
  ["photobiomodulation photomed laser surg"] = "Photobiomodulation, Photomedicine, and Laser Surgery",
  ["photochem photobiol"] = "Photochemistry and Photobiology",
  ["photochem photobiol sci"] = "Photochemical & photobiological sciences : Official journal of the European Photochemistry Association and the European Society for Photobiology",
  ["photochem-spec perio"] = "Photochemistry-a Specialist Periodical Report",
  ["photodermatol photoimmunol photomed"] = "Photodermatology, photoimmunology & photomedicine",
  ["photodiagn photodyn ther"] = "Photodiagnosis and Photodynamic Therapy",
  ["photodiagnosis photodyn ther"] = "Photodiagnosis and photodynamic therapy",
  ["photogr sci eng"] = "Photographic Science and Engineering",
  ["photogramm eng remote sens"] = "Photogrammetric Engineering and Remote Sensing",
  ["photogramm eng remote sensing"] = "Photogrammetric engineering and remote sensing",
  ["photogramm rec"] = "Photogrammetric Record",
  ["photomed laser surg"] = "Photomedicine and laser surgery",
  ["photonic network commun"] = "Photonic Network Communications",
  ["photonic sensors"] = "Photonic Sensors",
  ["photonics insights"] = "Photonics Insights",
  ["photonics lasers med"] = "Photonics & lasers in medicine",
  ["photonics lett pol"] = "Photonics letters of Poland",
  ["photonics nanostruct"] = "Photonics and nanostructures : fundamentals and applications",
  ["photonics nanostruct fundam appl"] = "Photonics and Nanostructures - Fundamentals and Applications",
  ["photonics res"] = "Photonics Research",
  ["photonics spectra"] = "Photonics spectra",
  ["photosynth res"] = "Photosynthesis research",
  ["photovoltaics bull"] = "Photovoltaics Bulletin",
  ["phycol res"] = "Phycological Research",
  ["phycological res"] = "Phycological research",
  ["phys a"] = "Physica A",
  ["phys act health"] = "Physical Activity and Health",
  ["phys act nutr"] = "Physical activity and nutrition",
  ["phys at nucl"] = "Physics of Atomic Nuclei",
  ["phys atomic nuclei"] = "Physics of Atomic Nuclei",
  ["phys atoms and molecules"] = "Physics of Atoms and Molecules",
  ["phys b"] = "Physica B",
  ["phys b (amsterdam, neth)"] = "Physica B: Condensed Matter (Amsterdam, Netherlands)",
  ["phys b condens matter"] = "Physica B: Condensed Matter",
  ["phys b: condens matter"] = "Physica B: Condensed Matter",
  ["phys ber"] = "Physikalische Berichte",
  ["phys biol"] = "Physical biology",
  ["phys bl"] = "Physikalische Blätter",
  ["phys c"] = "Physica C",
  ["phys c (amsterdam, neth)"] = "Physica C: Superconductivity and Its Applications (Amsterdam, Netherlands)",
  ["phys c supercond its appl"] = "Physica C: Superconductivity and its Applications",
  ["phys c: supercond its appl"] = "Physica C: Superconductivity and its Applications",
  ["phys chem chem phys"] = "Physical chemistry chemical physics : PCCP",
  ["phys chem chem phys pccp"] = "Physical Chemistry Chemical Physics",
  ["phys chem earth"] = "Physics and Chemistry of the Earth",
  ["phys chem earth (2002)"] = "Physics and chemistry of the earth (2002)",
  ["phys chem earth a/b/c/"] = "Physics and Chemistry of the Earth A/B/C",
  ["phys chem earth b"] = "Physics and Chemistry of the Earth B",
  ["phys chem earth part a"] = "Physics and Chemistry of the Earth, Part A: Solid Earth and Geodesy",
  ["phys chem earth part b"] = "Physics and Chemistry of the Earth, Part B: Hydrology, Oceans and Atmosphere",
  ["phys chem earth part c"] = "Physics and Chemistry of the Earth, Part C: Solar, Terrestrial & Planetary Science",
  ["phys chem glasses"] = "Physics and Chemistry of Glasses",
  ["phys chem glasses: eur j glass sci technol, part b"] = "Physics and Chemistry of Glasses: European Journal of Glass Science and Technology, Part B",
  ["phys chem liq"] = "Physics and Chemistry of Liquids",
  ["phys chem liquids"] = "Physics and chemistry of liquids",
  ["phys chem miner"] = "Physics and Chemistry of Minerals",
  ["phys chem res"] = "Physical Chemistry Research",
  ["phys chem solids"] = "Physics and Chemistry of Solids",
  ["phys commun"] = "Physical Communication",
  ["phys d"] = "Physica D",
  ["phys d nonlinear phenom"] = "Physica D: Nonlinear Phenomena",
  ["phys d: nonlinear phenom"] = "Physica D: Nonlinear Phenomena",
  ["phys dark universe"] = "Physics of the Dark Universe",
  ["phys disabil"] = "Physical disabilities : education and related services",
  ["phys e"] = "Physica E",
  ["phys e: low-dimens syst nanostructures"] = "Physica E: Low-dimensional Systems and Nanostructures",
  ["phys earth planet in"] = "Physics of the Earth and Planetary Interiors",
  ["phys earth planet inter"] = "Physics of the Earth and Planetary Interiors",
  ["phys earth space environ"] = "Physics of Earth and Space Environments",
  ["phys east"] = "Physician east",
  ["phys educ"] = "Physics Education",
  ["phys educ res"] = "Physics Education Research",
  ["phys educ sport pedagogy"] = "Physical education and sport pedagogy",
  ["phys eng sci med"] = "Physical and engineering sciences in medicine",
  ["phys essays"] = "Physics Essays",
  ["phys fluids"] = "Physics of Fluids",
  ["phys fluids (1994)"] = "Physics of fluids (Woodbury, N.Y. : 1994)",
  ["phys fluids a"] = "Physics of Fluids A",
  ["phys fluids b"] = "Physics of Fluids B",
  ["phys geogr"] = "Physical geography",
  ["phys imaging radiat oncol"] = "Physics and imaging in radiation oncology",
  ["phys j"] = "Physik Journal",
  ["phys kondens mater"] = "Physik der Kondensierten Materie",
  ["phys lett"] = "Physics Letters",
  ["phys lett a"] = "Physics Letters A",
  ["phys lett b"] = "Physics Letters B",
  ["phys life rev"] = "Physics of life reviews",
  ["phys meas"] = "Physiological Measurement",
  ["phys med"] = "Physica medica : PM : an international journal devoted to the applications of physics to medicine and biology : official journal of the Italian Association of Biomedical Physics (AIFB)",
  ["phys med biol"] = "Physics in medicine and biology",
  ["phys med rehabil clin n am"] = "Physical medicine and rehabilitation clinics of North America",
  ["phys med rehabil int"] = "Physical medicine and rehabilitation international",
  ["phys med rehabil res"] = "Physical medicine and rehabilitation research",
  ["phys mesomech"] = "Physical Mesomechanics",
  ["phys met"] = "Physics of Metals",
  ["phys met metall"] = "Physics of Metals and Metallography",
  ["phys met metallogr (ussr)"] = "Physics of Metals and Metallography (USSR) [translation of Fizika Metallovi Metallovedenie]",
  ["phys monogr"] = "Physikalische Monographien",
  ["phys nonlinear phenom"] = "Physica D: Nonlinear Phenomena",
  ["phys notes"] = "Physics Notes",
  ["phys occup ther geriatr"] = "Physical & occupational therapy in geriatrics",
  ["phys occup ther pediatr"] = "Physical & occupational therapy in pediatrics",
  ["phys oceanogr"] = "Physical Oceanography",
  ["phys open"] = "Physics Open",
  ["phys part nucl"] = "Physics of Particles and Nuclei",
  ["phys part nucl lett"] = "Physics of Particles and Nuclei Letters",
  ["phys particles nuclei"] = "Physics of Particles and Nuclei",
  ["phys perspect"] = "Physics in perspective",
  ["phys phys fiz"] = "Physics Physique Fizika",
  ["phys plasma"] = "Physics of Plasmas",
  ["phys plasmas"] = "Physics of Plasmas",
  ["phys procedia"] = "Physics procedia",
  ["phys rep"] = "Physics Reports",
  ["phys rep kumamoto univ"] = "Physics Reports of Kumamoto University",
  ["phys rev"] = "Physical Review",
  ["phys rev a"] = "Physical Review A",
  ["phys rev a  (coll park)"] = "Physical review. A",
  ["phys rev a gen phys"] = "Physical review. A, General physics",
  ["phys rev a: at mol opt phys"] = "Physical Review A: Atomic, Molecular, and Optical Physics",
  ["phys rev a: at, mol, opt phys"] = "Physical Review A: Atomic, Molecular, and Optical Physics",
  ["phys rev accel beams"] = "Physical Review Accelerators and Beams",
  ["phys rev appl"] = "Physical Review Applied",
  ["phys rev applied"] = "Physical Review Applied",
  ["phys rev b"] = "Physical review. B",
  ["phys rev b condens matter"] = "Physical review. B, Condensed matter",
  ["phys rev b condens matter mater phys"] = "Physical review. B, Condensed matter and materials physics",
  ["phys rev b: condens matter"] = "Physical Review B: Condensed Matter",
  ["phys rev b: condens matter mater phys"] = "Physical Review B: Condensed Matter and Materials Physics",
  ["phys rev c"] = "Physical review. C",
  ["phys rev c nucl phys"] = "Physical review. C, Nuclear physics",
  ["phys rev c: nucl phys"] = "Physical Review C: Nuclear Physics",
  ["phys rev d"] = "Physical Review D",
  ["phys rev d part fields"] = "Physical review. D, Particles and fields",
  ["phys rev d: part fields"] = "Physical Review D: Particles and Fields",
  ["phys rev d: part, fields, gravitation, cosmol"] = "Physical Review D: Particles, Fields, Gravitation, and Cosmology",
  ["phys rev e"] = "Physical Review E",
  ["phys rev e stat nonlin soft matter phys"] = "Physical review. E, Statistical, nonlinear, and soft matter physics",
  ["phys rev e stat phys plasmas fluids relat interdiscip topics"] = "Physical review. E, Statistical physics, plasmas, fluids, and related interdisciplinary topics",
  ["phys rev e: stat nonlinear soft matter phys"] = "Physical Review E: Statistical, Nonlinear, and Soft Matter Physics",
  ["phys rev e: stat phys plasmas fluids relat interdiscip top"] = "Physical Review E: Statistical Physics, Plasmas, Fluids, and Related Interdisciplinary Topics",
  ["phys rev e: stat phys plasmas fluids relat interdisciplin top"] = "Physical Review E: Statistical Physics, Plasmas, Fluids, and Related Interdisciplinary Topics",
  ["phys rev e: stat, nonlinear, soft matter phys"] = "Physical Review E: Statistical, Nonlinear, and Soft Matter Physics",
  ["phys rev fluids"] = "Physical Review Fluids",
  ["phys rev lett"] = "Physical Review Letters",
  ["phys rev mater"] = "Physical review materials",
  ["phys rev materials"] = "Physical Review Materials",
  ["phys rev phys educ res"] = "Physical Review Physics Education Research",
  ["phys rev res"] = "Physical review research",
  ["phys rev research"] = "Physical Review Research",
  ["phys rev spec top accel beams"] = "Physical Review Special Topics - Accelerators and Beams",
  ["phys rev spec top phys educ res"] = "Physical Review Special Topics - Physics Education Research",
  ["phys rev x"] = "Physical Review X",
  ["phys scr"] = "Physica Scripta",
  ["phys scr t"] = "Physica Scripta, T.",
  ["phys scr, t"] = "Physica Scripta, T",
  ["phys scripta"] = "Physica Scripta",
  ["phys sep sci eng"] = "Physical Separaion in Science and Engineering",
  ["phys solid state"] = "Physics of the Solid State",
  ["phys sportsmed"] = "The Physician and sportsmedicine",
  ["phys status solidi"] = "Physica Status Solidi",
  ["phys status solidi a"] = "physica status solidi (a)",
  ["phys status solidi a: appl res"] = "Physica Status Solidi A: Applied Research",
  ["phys status solidi appl mater sci"] = "Physica Status Solidi A: Applications and Materials Science",
  ["phys status solidi b"] = "physica status solidi (b)",
  ["phys status solidi b basic res"] = "Physica status solidi. B, Basic research",
  ["phys status solidi b basic solid state phys"] = "Physica status solidi. B, Basic solid state physics : PSS",
  ["phys status solidi b: basic res"] = "Physica Status Solidi B: Basic Research",
  ["phys status solidi b: basic solid state phys"] = "Physica Status Solidi B: Basic Solid State Physics",
  ["phys status solidi c"] = "physica status solidi (c)",
  ["phys status solidi c: curr top solid state phys"] = "Physica Status Solidi C: Current Topics in Solid State Physics",
  ["phys status solidi rapid res lett"] = "Physica status solidi. Rapid research letters : PSS-RRL",
  ["phys status solidi rrl"] = "Physica Status Solidi Rapid Research Letters",
  ["phys status solidi rrl - rapid res lett"] = "physica status solidi (RRL) – Rapid Research Letters",
  ["phys teach"] = "Physics Teacher",
  ["phys technol"] = "Physics in Technology",
  ["phys ther"] = "Physical therapy",
  ["phys ther rehabil"] = "Physical therapy and rehabilitation",
  ["phys ther res"] = "Physical therapy research",
  ["phys ther rev"] = "The Physical therapy review",
  ["phys ther sport"] = "Physical therapy in sport : official journal of the Association of Chartered Physiotherapists in Sports Medicine",
  ["phys today"] = "Physics Today",
  ["phys unserer zeit"] = "Physik in Unserer Zeit",
  ["phys usp"] = "Physics-Uspekhi",
  ["phys wave phenom"] = "Physics of Wave Phenomenon",
  ["phys world"] = "Physics World",
  ["phys world discov"] = "Physics World Discovery",
  ["phys z"] = "Physikalishce Zeitschrift",
  ["phys z sowjetunion"] = "Physikalische Zeitschrift der Sowjetunion",
  ["physica (utrecht)"] = "Physica (Utrecht)",
  ["physica a"] = "Physica A: Statistical Mechanics and its Applications",
  ["physica a (amsterdam, neth)"] = "Physica A: Statistical Mechanics and Its Applications (Amsterdam, Netherlands)",
  ["physica b"] = "Physica B: Condensed Matter",
  ["physica b condens matter"] = "Physica. B, Condensed matter",
  ["physica c"] = "Physica C: Superconductivity",
  ["physica c supercond"] = "Physica. C, Superconductivity",
  ["physica d"] = "Physica D: Nonlinear Phenomena",
  ["physica d (amsterdam, neth)"] = "Physica D: Nonlinear Phenomena (Amsterdam, Netherlands)",
  ["physica e"] = "Physica E: Low-dimensional Systems and Nanostructures",
  ["physica e (amsterdam, neth)"] = "Physica E: Low-dimensional Systems and Nanostructures (Amsterdam, Netherlands)",
  ["physica e low dimens syst nanostruct"] = "Physica. E, Low-dimensional systems & nanostructures",
  ["physica med"] = "Physica Medica",
  ["physica status solidi a appl res"] = "Physica status solidi (A): Applied research",
  ["physician assist"] = "Physician assistant",
  ["physician assist clin"] = "Physician assistant clinics",
  ["physician assist health pract"] = "Physician assistant. Health practitioner",
  ["physician exec"] = "Physician executive",
  ["physician leadersh j"] = "Physician leadership journal",
  ["physician perform paym rep"] = "Physician performance & payment report",
  ["physician relat update"] = "Physician relations update",
  ["physician's bull"] = "Physician's bulletin",
  ["physicians manage"] = "Physician's management",
  ["physicians world"] = "Physician's world",
  ["physicochem probl miner process"] = "Physicochemical Problems of Mineral Processing",
  ["physics (college park md)"] = "Physics",
  ["physics (ny)"] = "Physics (New York)",
  ["physics procedia"] = "Physics Procedia",
  ["physiol behav"] = "Physiology & behavior",
  ["physiol biochem zool"] = "Physiological and biochemical zoology : PBZ",
  ["physiol bohemoslov"] = "Physiologia Bohemoslovaca",
  ["physiol chem phys"] = "Physiological chemistry and physics",
  ["physiol chem phys med nmr"] = "Physiological Chemistry and Physics and Medical NMR",
  ["physiol comp ocol int j comp physiol ecol"] = "Physiologia comparata et oecologia; an international journal of comparative physiology and ecology",
  ["physiol entomol"] = "Physiological entomology",
  ["physiol genomics"] = "Physiological genomics",
  ["physiol int"] = "Physiology international",
  ["physiol j"] = "Physiology journal",
  ["physiol meas"] = "Physiological measurement",
  ["physiol mol biol plants"] = "Physiology and molecular biology of plants : an international journal of functional plant biology",
  ["physiol mol plant pathol"] = "Physiological and Molecular Plant Pathology",
  ["physiol pharmacol"] = "Physiology and Pharmacology",
  ["physiol pharmacol physicians"] = "Physiology & pharmacology for physicians",
  ["physiol physicians"] = "Physiology for physicians",
  ["physiol plant"] = "Physiologia plantarum",
  ["physiol plant pathol"] = "Physiological Plant Pathology",
  ["physiol rep"] = "Physiological reports",
  ["physiol res"] = "Physiological research",
  ["physiol rev"] = "Physiological reviews",
  ["physiol rev suppl"] = "Physiological reviews. Supplement",
  ["physiol teach"] = "The Physiology teacher",
  ["physiol veg"] = "Physiologie végétale",
  ["physiol zool"] = "Physiological zoology",
  ["physiology (bethesda)"] = "Physiology (Bethesda, Md.)",
  ["physiother can"] = "Physiotherapy Canada. Physiothérapie Canada",
  ["physiother pract res"] = "Physiotherapy practice and research",
  ["physiother res int"] = "Physiotherapy research international : the journal for researchers and clinicians in physical therapy",
  ["physiother theory pract"] = "Physiotherapy theory and practice",
  ["physiotherap rev"] = "The Physiotherapy review",
  ["physis riv int stor sci"] = "Physis; rivista internazionale di storia della scienza",
  ["physis riv internaz storia sci (ns)"] = "Physis—Rivista Internazionale di Storia della Scienza. Nuova Serie",
  ["phytobiomes j"] = "Phytobiomes journal",
  ["phytochem anal"] = "Phytochemical analysis : PCA",
  ["phytochem lett"] = "Phytochemistry Letters",
  ["phytochem rev"] = "Phytochemistry Reviews",
  ["phytochemistry (elsevier)"] = "Phytochemistry (Elsevier)",
  ["phytol balc (sofia)"] = "Phytologia balcanica",
  ["phytol balcan"] = "Phytologia Balcanica",
  ["phytomed plus"] = "Phytomedicine Plus",
  ["phyton (b aires)"] = "Phyton",
  ["phyton (buenos aires)"] = "Phyton",
  ["phyton-int j exp bot"] = "Phyton-International Journal of Experimental Botany",
  ["phytopathogenic mollicutes"] = "Phytopathogenic mollicutes",
  ["phytopathol mediterr"] = "Phytopathologia mediterranea",
  ["phytopathol pol"] = "Phytopathologia polonica",
  ["phytopathol res"] = "Phytopathology Research",
  ["phytopathol z"] = "Phytopathologische Zeitschrift",
  ["phytother res"] = "Phytotherapy research : PTR",
  ["phytotherapie (paris)"] = "Phytothérapie (Paris, France)",
  ["pi mu epsilon j"] = "Pi Mu Epsilon Journal",
  ["pi perspect"] = "PI perspective",
  ["piact pap"] = "PIACT papers",
  ["piact prod news"] = "PIACT product news. Program for the Introduction and Adaptation of Contraceptive Technology (U.S.)",
  ["piact prod newsl"] = "PIACT product newsletter. Program for the Introduction and Adaptation of Contraceptive Technology (U.S.)",
  ["pidsa abstr"] = "PIDSA abstracts",
  ["pieleg polozna"] = "Pielȩgniarka i połozna",
  ["pigm cell melanoma res"] = "Pigment Cell & Melanoma Research",
  ["pigm cell res"] = "Pigment Cell Research",
  ["pigm resin technol"] = "Pigment and Resin Technology",
  ["pigment cell melanoma res"] = "Pigment cell & melanoma research",
  ["pigment cell res"] = "Pigment Cell Research",
  ["pilot evanst hosp"] = "The Pilot-Evanston Hospital",
  ["pilot feasibility stud"] = "Pilot and feasibility studies",
  ["ping tu hsueh tsa chih"] = "Bing du xue za zhi = Virologica Sinica",
  ["pinheiros ter"] = "Pinheiros terapêutico",
  ["pioneer am soc trans"] = "Pioneer America Society transactions : P.A.S.T. Pioneer America Society",
  ["pioneers arts humanit sci eng pract"] = "Pioneers in Arts, Humanities, Science, Engineering, Practice",
  ["pis\\cprime ma zh èksper teoret fiz"] = "Pis\\cprime ma v Zhurnal Èksperimental\\cprime noy i Teoreticheskoy Fiziki",
  ["pishch prom-st (moscow)"] = "Pishchevaya Promyshlennost (Moscow)",
  ["pisum genet"] = "Pisum genetics",
  ["pis’ma astron zh"] = "Pis’ma v Astronomicheskii Zhurnal",
  ["pis’ma astron zh [sov astron lett]"] = "Pis’ma v Astronomicheskii Zhurnal [Soviet Astronomy Letters]",
  ["pis’ma zh eksp teor fiz"] = "Pis’ma v Zhurnal Eksperimental’noi i Teoreticheskoi Fiziki",
  ["pis’ma zh eksp teor fiz [jetp lett]"] = "Pis’ma v Zhurnal Eksperimentalnoi i Teoreticheskoi Fiziki [JETP Letters]",
  ["pis’ma zh tekh fiz"] = "Pis’ma v Zhurnal Tekhnicheskoi Fiziki",
  ["pis’ma zh tekh fiz [sov tech phys lett]"] = "Pis’ma v Zhurnal Tekhnicheskoi Fiziki [Soviet Technical Physics Letters]",
  ["pitman monogr surveys pure appl math"] = "Pitman Monographs and Surveys in Pure and Applied Mathematics",
  ["pitman res notes math ser"] = "Pitman Research Notes in Mathematics Series",
  ["pittsbg med bull"] = "Pittsburgh medical bulletin",
  ["pizhuhish va sazandidi"] = "Pizhuhish va sazandidi = Research & reconstruction",
  ["pizhuhishhayi karburdi giyahpizishki"] = "Pizhūhish/hā-yi kārburdī dar giyāhpizishkī",
  ["placenta suppl"] = "Placenta. Supplement",
  ["plain dealer"] = "Plain dealer (Cleveland, Ohio : 1961)",
  ["plains anthropol"] = "Plains anthropologist",
  ["plan arb"] = "Plan og arbeid",
  ["plan parent chall"] = "Planned parenthood challenges",
  ["plan parent eur"] = "Planned parenthood in Europe = Planning familial en Europe",
  ["plan parent rev"] = "Planned parenthood review",
  ["plan parent world popul wash memo"] = "Planned Parenthood-World Population Washington memo",
  ["plan perspect"] = "Planning perspectives : PP",
  ["planej agora"] = "Planejamento Agora",
  ["planet rep"] = "The Planetary report",
  ["planet sci"] = "Planetary science",
  ["planet sci j"] = "The planetary science journal",
  ["planet space sci"] = "Planetary and Space Science",
  ["planfed news"] = "Planfed news : newsletter of the Planned Parenthood Federation of Nigeria",
  ["plankton benthos res"] = "Plankton and Benthos Research",
  ["plann hist"] = "Planning history",
  ["planned parent (india)"] = "Planned parenthood",
  ["plant arch"] = "Plant archives",
  ["plant biol"] = "Plant biology",
  ["plant biol (stuttg)"] = "Plant biology (Stuttgart, Germany)",
  ["plant biosyst"] = "Plant biosystems",
  ["plant biosyst - int j dealing aspects plant biosyst"] = "Plant Biosystems - An International Journal Dealing with all Aspects of Plant Biology",
  ["plant biotechnol"] = "Plant biotechnology (Sheffield, England)",
  ["plant biotechnol (tokyo)"] = "Plant biotechnology (Tokyo, Japan)",
  ["plant biotechnol j"] = "Plant biotechnology journal",
  ["plant biotechnol rep"] = "Plant biotechnology reports",
  ["plant breed"] = "Plant breeding = Zeitschrift für Pflanzenzüchtung",
  ["plant breed biotechnol"] = "Plant breeding and biotechnology",
  ["plant breed rev"] = "Plant breeding reviews",
  ["plant cell"] = "The Plant cell",
  ["plant cell biotechnol mol biol"] = "Plant cell biotechnology and molecular biology",
  ["plant cell environ"] = "Plant, cell & environment",
  ["plant cell physiol"] = "Plant & cell physiology",
  ["plant cell rep"] = "Plant cell reports",
  ["plant cell tissue organ cult"] = "Plant cell, tissue and organ culture",
  ["plant commun"] = "Plant communications",
  ["plant direct"] = "Plant direct",
  ["plant dis"] = "Plant disease",
  ["plant dis manag rep"] = "Plant disease management reports : PDMR",
  ["plant divers"] = "Plant diversity",
  ["plant divers evol"] = "Plant diversity and evolution",
  ["plant diversity resour"] = "Plant diversity and resources",
  ["plant ecol"] = "Plant ecology",
  ["plant ecol divers"] = "Plant ecology & diversity",
  ["plant ecol evol"] = "Plant ecology and evolution",
  ["plant ecolog"] = "Plant Ecology",
  ["plant ecolog divers"] = "Plant Ecology and Diversity",
  ["plant foods hum nutr"] = "Plant foods for human nutrition (Dordrecht, Netherlands)",
  ["plant fungal syst"] = "Plant and fungal systematics",
  ["plant gene"] = "Plant gene",
  ["plant gene trait"] = "Plant gene & trait",
  ["plant genet resour"] = "Plant genetic resources : characterization and utilization",
  ["plant genet resour newsl"] = "Plant genetic resources newsletter (Rome, Italy : 1979)",
  ["plant genome"] = "The plant genome",
  ["plant growth regul"] = "Plant growth regulation",
  ["plant health"] = "Plantation health",
  ["plant health cases"] = "Plant Health Cases",
  ["plant health prog"] = "Plant health progress",
  ["plant j"] = "The Plant journal : for cell and molecular biology",
  ["plant methods"] = "Plant methods",
  ["plant mol biol"] = "Plant molecular biology",
  ["plant mol biol rep"] = "Plant Molecular Biology Reporter",
  ["plant mol biol report"] = "Plant molecular biology reporter",
  ["plant omics"] = "Plant omics",
  ["plant pathol"] = "Plant pathology",
  ["plant pathol j"] = "The plant pathology journal",
  ["plant pathol j (faisalabad)"] = "Plant pathology journal",
  ["plant pathol quar j fungal biolog"] = "Plant Pathology & Quarantine Journal of Fungal Biology",
  ["plant pathology bulletin|zhi wu bing li xue hui kan"] = "Zhi wu bing li xue hui kan",
  ["plant phenome j"] = "Plant Phenome Journal",
  ["plant phenomics"] = "Plant phenomics (Washington, D.C.)",
  ["plant physiol"] = "Plant physiology",
  ["plant physiol biochem"] = "Plant physiology and biochemistry : PPB",
  ["plant physiol biochem (amsterdam, neth)"] = "Plant Physiology and Biochemistry (Amsterdam, Netherlands)",
  ["plant physiol biochem (issy les moulineaux, fr)"] = "Plant Physiology and Biochemistry (Issy les Moulineaux, France)",
  ["plant physiol biochem (issy-les-moulineaux, fr)"] = "Plant Physiology and Biochemistry (Issy-les-Moulineaux, France)",
  ["plant physiol rep"] = "Plant Physiology Reports",
  ["plant physiol suppl (bethesda)"] = "Plant physiology. Supplement",
  ["plant prod sci"] = "Plant production science",
  ["plant prot q"] = "Plant protection quarterly",
  ["plant prot sci"] = "Plant protection science = Ochrana rostlin",
  ["plant reprod"] = "Plant reproduction",
  ["plant sci"] = "Plant science : an international journal of experimental plant biology",
  ["plant sci (amsterdam, neth)"] = "Plant Science (Amsterdam, Netherlands)",
  ["plant sci (limerick, irel)"] = "Plant Science (Limerick, Ireland)",
  ["plant sci (shannon, irel)"] = "Plant Science (Shannon, Ireland)",
  ["plant sci j"] = "Plant science journal",
  ["plant sci lett"] = "Plant science letters",
  ["plant signal behav"] = "Plant signaling & behavior",
  ["plant signaling behav"] = "Plant Signaling & Behavior",
  ["plant sociol"] = "Plant Sociology",
  ["plant soil"] = "Plant and soil",
  ["plant soil environ"] = "Plant, soil and environment",
  ["plant species biol"] = "Plant species biology",
  ["plant stress"] = "Plant Stress",
  ["plant sys evol suppl"] = "Plant systematics and evolution. Supplementum = Entwicklungsgeschichte und Systematik der Pflanzen. Supplementum",
  ["plant syst evol"] = "Plant systematics and evolution = Entwicklungsgeschichte und Systematik der Pflanzen",
  ["plant tissue cult biotechnol"] = "Plant tissue culture & biotechnology",
  ["plant viruses"] = "Plant viruses",
  ["plant, cell environ"] = "Plant, Cell and Environment",
  ["plant-environ interact"] = "Plant-Environment Interactions",
  ["plant/oper prog"] = "Plant/Operations Progress",
  ["planta med"] = "Planta Medica",
  ["plants (basel)"] = "Plants (Basel, Switzerland)",
  ["plants people planet"] = "Plants, People, Planet,",
  ["plasma chem plasma process"] = "Plasma Chemistry and Plasma Processing",
  ["plasma devices oper"] = "Plasma Devices and Operations",
  ["plasma med"] = "Plasma medicine",
  ["plasma phys"] = "Plasma Physics",
  ["plasma phys control fusion"] = "Plasma physics and controlled fusion",
  ["plasma phys controlled fusion"] = "Plasma Physics and Controlled Fusion",
  ["plasma phys controlled nucl fusion res"] = "Plasma Physics and Controlled Nuclear Fusion Research",
  ["plasma phys rep"] = "Plasma Physics Reports",
  ["plasma process polym"] = "Plasma processes and polymers (Print)",
  ["plasma processes polym"] = "Plasma Processes and Polymers",
  ["plasma res express"] = "Plasma Research Express",
  ["plasma sci technol"] = "Plasma Science and Technology",
  ["plasma sources sci technol"] = "Plasma Sources Science and Technology",
  ["plasmas polym"] = "Plasmas and Polymers",
  ["plast aesthet res"] = "Plastic and aesthetic research",
  ["plast eng"] = "Plastics Engineering",
  ["plast massy"] = "Plasticheskie Massy",
  ["plast reconstr surg"] = "Plastic and reconstructive surgery",
  ["plast reconstr surg (1946)"] = "Plastic and reconstructive surgery (1946)",
  ["plast reconstr surg glob open"] = "Plastic and reconstructive surgery. Global open",
  ["plast reconstr surg transplant bull"] = "Plastic and reconstructive surgery and the transplantation bulletin",
  ["plast rubber compos process appl"] = "Plastics Rubber and Composites Processing and Applications",
  ["plast surg (oakv)"] = "Plastic surgery (Oakville, Ont.)",
  ["plast surg int"] = "Plastic surgery international",
  ["plast surg nurs"] = "Plastic surgical nursing : official journal of the American Society of Plastic and Reconstructive Surgical Nurses",
  ["plast, rubber compos"] = "Plastics, Rubber and Composites",
  ["plat surf finish"] = "Plating and Surface Finishing",
  ["platin met rev"] = "Platinum metals review",
  ["platinum met rev"] = "Platinum Metals Review",
  ["platon (athens)"] = "Platōn",
  ["pleura (thousand oaks)"] = "Pleura (Thousand Oaks, Ventura County, Calif.)",
  ["pleura peritoneum"] = "Pleura and peritoneum",
  ["pliska stud math"] = "Pliska Studia Mathematica",
  ["plos biol"] = "PLoS biology",
  ["plos clim"] = "PLOS Climate",
  ["plos clin trials"] = "PLoS clinical trials",
  ["plos complex syst"] = "PLOS Complex Systems",
  ["plos comput biol"] = "PLoS computational biology",
  ["plos curr"] = "PLoS currents",
  ["plos curr:  evidence genomic tests"] = "PLOS Currents: Evidence on Genomic Tests",
  ["plos curr: disasters"] = "PLOS Currents: Disasters",
  ["plos curr: evidence genomic tests"] = "PLOS Currents: Evidence on Genomic Tests",
  ["plos curr: huntington dis"] = "PLOS Currents: Huntington Disease",
  ["plos curr: muscular dystrophy"] = "PLOS Currents: Muscular Dystrophy",
  ["plos curr: outbreaks"] = "PLOS Currents: Outbreaks",
  ["plos curr: tree life"] = "PLOS Currents: Tree of Life",
  ["plos digital health"] = "PLOS Digital Health",
  ["plos genet"] = "PLoS genetics",
  ["plos global public health"] = "PLOS Global Public Health",
  ["plos med"] = "PLoS medicine",
  ["plos negl trop dis"] = "PLoS neglected tropical diseases",
  ["plos negltrop dis"] = "PLOS Neglected Tropical Diseases",
  ["plos one"] = "PloS one",
  ["plos pathog"] = "PLoS pathogens",
  ["plos sustainability transform"] = "PLOS Sustainability and Transformation",
  ["plos water"] = "PLOS Water",
  ["plovdiv univ paisii khilendarski nauchn trud mat"] = "Plovdivski Universitet “Paisii Khilendarski”. Nauchni Trudove. Matematika",
  ["plucne bolesti"] = "Plućne bolesti : casopis Udruzenja pneumoftiziologa Jugoslavije = the journal of Yugoslav Association of Phthisiology and Pneumology",
  ["plucne bolesti tuberk"] = "Plućne bolesti i tuberkuloza",
  ["plural soc"] = "Plural societies",
  ["plus lumière"] = "Plus de Lumière",
  ["plzen lek sb"] = "Plzen̆ský lékar̆ský sborník",
  ["pm r"] = "PM & R : the journal of injury, function, and rehabilitation",
  ["pmc biophys"] = "PMC biophysics",
  ["pmc phys a"] = "PMC Physics A",
  ["pmc phys b"] = "PMC Physics B",
  ["pmse prepr"] = "PMSE Preprints",
  ["pmse preprints"] = "PMSE preprints. American Chemical Society. Division of Polymeric Materials: Science and Engineering. Meeting",
  ["pneumo news"] = "Pneumo news",
  ["pneumologe (berl)"] = "Der Pneumologe",
  ["pneumonia (nathan)"] = "Pneumonia (Nathan Qld.)",
  ["pneumonol alergol pol"] = "Pneumonologia i alergologia polska",
  ["pneumonol danub"] = "Pneumonologia Danubiana",
  ["pneumonol phymatiologike epitheor"] = "Pneumonologike kai phymatiologike epitheoresis",
  ["pneumonol pol"] = "Pneumonologia polska",
  ["pnlde subser control"] = "PNLDE Subseries in Control",
  ["poblac desarro"] = "Población y desarrollo (1991)",
  ["poblac salud mesoam"] = "Población y salud en mesoamérica",
  ["podiplomski sem mat"] = "Podiplomski Seminar iz Matematike",
  ["podstaw algorytmy numer"] = "Podstawowe Algorytmy Numeryczne",
  ["podstaw nauki tech monogr"] = "Podstawowe Nauki Techniczne",
  ["poetics (amst)"] = "Poetics (Hague, Netherlands)",
  ["poggendorff’s ann phys"] = "Poggendorff’s Annalen der Physik",
  ["pogon sahoe yongu"] = "Pogŏn sahoe yŏn'gu",
  ["poiesis prax"] = "Poiesis & praxis : international journal of ethics of science and technology assessment",
  ["poincare j anal appl"] = "Poincare Journal of Analysis and Applications",
  ["point care"] = "Point of care",
  ["poj nurs pract res"] = "POJ nursing practice & research",
  ["pol am stud"] = "Polish American studies",
  ["pol anal"] = "Policy Analysis",
  ["pol arch intern med"] = "Polish archives of internal medicine",
  ["pol arch med wewn"] = "Polskie Archiwum Medycyny Wewnetrznej",
  ["pol arch weter"] = "Polskie archiwum weterynaryjne",
  ["pol bot j"] = "Polish botanical journal",
  ["pol contemp philos philos humanit"] = "Polish Contemporary Philosophy and Philosophical Humanities",
  ["pol ecol stud"] = "Polish ecological studies",
  ["pol hyperb res"] = "Polish hyperbaric research",
  ["pol j chem"] = "Polish journal of chemistry",
  ["pol j chem technol"] = "Polish Journal of Chemical Technology",
  ["pol j ecol"] = "Polish journal of ecology",
  ["pol j entomol"] = "Polish Journal of Entomology",
  ["pol j environ stud"] = "Polish journal of environmental studies",
  ["pol j food nutr sci"] = "Polish journal of food and nutrition sciences",
  ["pol j med phys eng"] = "Polish Journal Of Medical Physics And Engineering",
  ["pol j microbiol"] = "Polish journal of microbiology",
  ["pol j nat sci"] = "Polish Journal of Natural Science",
  ["pol j occup med"] = "Polish journal of occupational medicine",
  ["pol j occup med environ health"] = "Polish journal of occupational medicine and environmental health",
  ["pol j pathol"] = "Polish journal of pathology : official journal of the Polish Society of Pathologists",
  ["pol j pharmacol"] = "Polish journal of pharmacology",
  ["pol j pharmacol pharm"] = "Polish journal of pharmacology and pharmacy",
  ["pol j radiol"] = "Polish journal of radiology",
  ["pol j soil sci"] = "Polish Journal Of Soil Science",
  ["pol j vet sci"] = "Polish journal of veterinary sciences",
  ["pol j zool"] = "Polish Journal of Zoology",
  ["pol marit res"] = "Polish Maritime Research",
  ["pol med hist sci bull"] = "Polish medical history and science bulletin",
  ["pol med j"] = "Polish medical journal",
  ["pol med sci hist bull"] = "Polish medical science and history bulletin",
  ["pol merkur lekarski"] = "Polski merkuriusz lekarski : organ Polskiego Towarzystwa Lekarskiego",
  ["pol merkuriusz lek"] = "Polski Merkuriusz Lekarski",
  ["pol orthop traumatol"] = "Polish orthopedics and traumatology",
  ["pol perspect"] = "Polish perspectives",
  ["pol polar res"] = "Polish polar research",
  ["pol popul rev"] = "Polish population review",
  ["pol prz radiol"] = "Polski przegla̧d radiologiczny",
  ["pol przegl chir"] = "Polski przeglad chirurgiczny",
  ["pol przegl radiol"] = "Polski przeglad radiologii",
  ["pol przegl radiol med nukl"] = "Polski przegla̜d radiologii i medycyny nuklearnej",
  ["pol rev"] = "The Polish review",
  ["pol sci"] = "Policy Sciences",
  ["pol stud"] = "Policy Studies",
  ["pol tyg lek"] = "Polski tygodnik lekarski (Warsaw, Poland : 1960)",
  ["pol tyg lek (wars)"] = "Polski tygodnik lekarski",
  ["pol west aff"] = "Polish Western affairs",
  ["polar biol"] = "Polar biology",
  ["polar biosci"] = "Polar bioscience",
  ["polar geogr"] = "Polar Geography",
  ["polar geogr (palm beach)"] = "Polar geography (Palm Beach, Fla.)",
  ["polar j"] = "The polar journal",
  ["polar rec"] = "Polar Record",
  ["polar rec (gr brit)"] = "The Polar record",
  ["polar res"] = "Polar research",
  ["polar sci"] = "Polar Science",
  ["polarnet tech rep"] = "Polarnet technical report",
  ["police chief"] = "The Police chief",
  ["police j"] = "The police journal",
  ["police pract res"] = "Police practice & research : an international journal",
  ["police q"] = "Police quarterly",
  ["policing soc"] = "Policing & society",
  ["policlin infant"] = "Policlinico infantile",
  ["policlinico [chir]"] = "Policlinico. Sezione Chirurgica",
  ["policlinico [med]"] = "Policlinico. Sezione Medica",
  ["policlinico [prat]"] = "Policlinico. Sezione Pratica",
  ["policlinico chir"] = "Il Policlinico. Sezione chirurgica",
  ["policlinico med"] = "Il Policlinico. Sezione medica",
  ["policlinico prat"] = "Il Policlinico. Sezione pratica",
  ["policy anal"] = "Policy analysis",
  ["policy anal brief h ser"] = "Policy analysis brief. H series",
  ["policy anal brief w ser"] = "Policy analysis brief. W series",
  ["policy brief (cent home care policy res)"] = "Policy brief (Center for Home Care Policy and Research (U.S.))",
  ["policy brief commonw fund"] = "Policy brief (Commonwealth Fund)",
  ["policy brief george wash univ cent health serv res policy"] = "Policy brief (George Washington University. Center for Health Services Research and Policy)",
  ["policy brief health care technol inst"] = "Policy brief (Alexandria, Va.)",
  ["policy brief ucla cent health policy res"] = "Policy brief (UCLA Center for Health Policy Research)",
  ["policy insights behav brain sci"] = "Policy insights from the behavioral and brain sciences",
  ["policy internet"] = "Policy & internet",
  ["policy matters inst res public policy"] = "Policy matters",
  ["policy polit"] = "Policy and politics",
  ["policy polit nurs pract"] = "Policy, politics & nursing practice",
  ["policy rev"] = "Policy review",
  ["policy sci"] = "Policy sciences",
  ["policy soc"] = "Policy & society",
  ["policy statement r coll gen pract"] = "Policy statement. Royal College of General Practitioners",
  ["policy stud"] = "Policy studies (Policy Studies Institute)",
  ["policy stud j"] = "Policy studies journal: the journal of the Policy Studies Organization",
  ["policy stud rev"] = "Policy studies review",
  ["polim med"] = "Polimery w medycynie",
  ["polimery (warsaw)"] = "Polimery (Warsaw)",
  ["polio news"] = "Polio news",
  ["polit aff"] = "Political affairs",
  ["polit afr"] = "Politique africaine (Paris, France : 1981)",
  ["polit anal"] = "Political analysis : an annual publication of the Methodology Section of the American Political Science Association",
  ["polit behav"] = "Political behavior",
  ["polit commun"] = "Political communication",
  ["polit ekon"] = "Politická ekonomie",
  ["polit etrang"] = "Politique étrangère",
  ["polit geogr"] = "Political geography",
  ["polit groups identities"] = "Politics, groups & identities",
  ["polit int"] = "Politica internazionale",
  ["polit intern"] = "Politique internationale",
  ["polit leg anthropol rev"] = "Political and legal anthropology review : PoLAR",
  ["polit life sci"] = "Politics and the Life Sciences",
  ["polit mein"] = "Die Politische Meinung",
  ["polit philos econ"] = "Politics philosophy & economics",
  ["polit policy"] = "Politics & policy (Statesboro, Ga.)",
  ["polit psychol"] = "Political psychology",
  ["polit q"] = "The Political quarterly",
  ["polit res q"] = "Political research quarterly",
  ["polit sci"] = "Political science (Wellington, N.Z.)",
  ["polit sci q"] = "Political science quarterly",
  ["polit sci quart"] = "Political Science Quarterly",
  ["polit soc"] = "Politics & society",
  ["polit soc (madrid)"] = "Política y sociedad (Madrid, Spain)",
  ["polit stud"] = "Politische Studien",
  ["polit stud (oxf)"] = "Political studies",
  ["polit theory"] = "Political theory",
  ["polit today"] = "Politics today",
  ["polit vierteljahresschr"] = "Politische Vierteljahresschrift",
  ["polit zeitgesch"] = "Aus Politik und Zeitgeschichte",
  ["polit ökol"] = "Politische Ökologie",
  ["politehn univ bucharest sci bull ser a appl math phys"] = "“Politehnica” University of Bucharest. Scientific Bulletin. Series A. Applied Mathematics and Physics",
  ["politica econ"] = "Politica Economica",
  ["political sci res methods"] = "Political science research and methods",
  ["politická ekon"] = "Politická Ekonomie",
  ["politico (pavia)"] = "Politico (Pavia, Italy)",
  ["politics (syd)"] = "Politics (Kensington, N.S.W.)",
  ["politics life sci"] = "Politics and the life sciences : the journal of the Association for Politics and the Life Sciences",
  ["politics relig"] = "Politics and religion",
  ["politiq popul"] = "Politiques de population",
  ["polito springer ser"] = "PoliTO Springer Series",
  ["poljopr znan smotra"] = "Poljoprivredna znanstvena smotra = Agriculturae conspectus scientificus : ACS",
  ["pollen spores"] = "Pollen et Spores",
  ["pollut atmos"] = "Pollution Atmospherique",
  ["pollut res"] = "Pollution research",
  ["polycycl aromat compd"] = "Polycyclic aromatic compounds",
  ["polycyclic aromat compd"] = "Polycyclic Aromatic Compounds",
  ["polym adv technol"] = "Polymers for Advanced Technologies",
  ["polym bull"] = "Polymer Bulletin",
  ["polym bull (berl)"] = "Polymer bulletin (Berlin, Germany)",
  ["polym bull (berlin)"] = "Polymer Bulletin (Berlin)",
  ["polym bull (heidelberg, ger)"] = "Polymer Bulletin (Heidelberg, Germany)",
  ["polym chem"] = "Polymer Chemistry",
  ["polym compos"] = "Polymer Composites",
  ["polym cryst"] = "Polymer Crystallization",
  ["polym degrad stab"] = "Polymer Degradation and Stability",
  ["polym eng sci"] = "Polymer Engineering and Science",
  ["polym gels networks"] = "Polymer Gels and Networks",
  ["polym int"] = "Polymer International",
  ["polym j"] = "Polymer Journal",
  ["polym j (tokyo, jpn)"] = "Polymer Journal (Tokyo, Japan)",
  ["polym mater sci eng"] = "Polymeric Materials Science and Engineering",
  ["polym networks blends"] = "Polymer Networks and Blends",
  ["polym news"] = "Polymer News",
  ["polym photochem"] = "Polymer Photochemistry",
  ["polym plast technol eng"] = "Polymer - Plastics Technology and Engineering",
  ["polym polym compos"] = "Polymers and Polymer Composites",
  ["polym prepr (am chem soc, div polym chem)"] = "Polymeric Preprints (American Chemical Society, Division of Polymer Chemistry)",
  ["polym react eng"] = "Polymer Reaction Engineering",
  ["polym recycl"] = "Polymer Recycling",
  ["polym renewable resour"] = "Polymers from Renewable Resources",
  ["polym rev"] = "Polymer Reviews",
  ["polym rev (phila pa)"] = "Polymer reviews (Philadelphia, Pa.)",
  ["polym rev (philadelphia, pa, u s)"] = "Polymer Reviews (Philadelphia, PA, United States)",
  ["polym sci ser a chem phys"] = "Polymer science. Series A, Chemistry, physics",
  ["polym sci ser a polym phys"] = "Polymer Science Series A - Polymer Physics",
  ["polym sci ser b polym chem"] = "Polymer Science Series B - Polymer Chemistry",
  ["polym sci ser c sel top"] = "Polymer Science Series C - Selected Topics",
  ["polym sci ser d glues sealing mater"] = "Polymer Science Series D - Glues and Sealing Materials",
  ["polym sci, ser a"] = "Polymer Science, Series A",
  ["polym sci, ser b"] = "Polymer Science, Series B",
  ["polym sci, ser c"] = "Polymer Science, Series C",
  ["polym sci, ser d"] = "Polymer Science, Series D",
  ["polym test"] = "Polymer Testing",
  ["polym-plast technol eng"] = "Polymer-Plastics Technology and Engineering",
  ["polym-plast technol mater"] = "Polymer-Plastics Technology and Materials",
  ["polymer (guildf)"] = "Polymer",
  ["polymer (korea)"] = "Polymer (Korea)",
  ["polymer sci tech"] = "Polymer Science and Technology",
  ["polymers (basel)"] = "Polymers",
  ["polymers (basel, switz)"] = "Polymers (Basel, Switzerland)",
  ["pomeranian j life sci"] = "Pomeranian journal of life sciences",
  ["pomoshch\\cprime abitur"] = "\\cyr V Pomoshch\\cprime Abiturientu",
  ["pontif acad sci scr varia"] = "Pontificae Academiae Scientiarum Scripta Varia",
  ["pop gov"] = "Popular government",
  ["pop sahel"] = "Pop Sahel : bulletin d'information sur la population et le développement",
  ["popcen newsl"] = "POPCEN news letter. Population Centre (Lucknow, India)",
  ["pope speaks"] = "The Pope speaks",
  ["popin bull"] = "POPIN bulletin",
  ["popul avenir"] = "Population et avenir",
  ["popul briefs"] = "Population briefs : reports on Population Council research",
  ["popul bull"] = "Population bulletin",
  ["popul bull ecwa"] = "Population bulletin of ECWA",
  ["popul bull escwa"] = "Population bulletin of ESCWA",
  ["popul bull u n econ comm west asia"] = "Population bulletin of the United Nations Economic Commission for Western Asia",
  ["popul bull un"] = "Population bulletin of the United Nations",
  ["popul chron"] = "Population chronicle",
  ["popul commun tech doc"] = "Population communication: technical documentation",
  ["popul concern news"] = "Population Concern news",
  ["popul data inf serv"] = "Population Data Information Service",
  ["popul desenvolv"] = "População & desenvolvimento",
  ["popul dev rev"] = "Population and development review",
  ["popul dyn q"] = "Population dynamics quarterly",
  ["popul ecol"] = "Population ecology",
  ["popul econ"] = "Population and economics",
  ["popul educ asia ocean newsl"] = "Population education in Asia and Oceania newsletter",
  ["popul educ asia pac newsl forum"] = "Population education in Asia and the Pacific newsletter",
  ["popul educ interchange"] = "Population education interchange",
  ["popul environ"] = "Population and environment",
  ["popul environ psychol newsl"] = "Population And Environmental Psychology Newsletter",
  ["popul famille"] = "Population et famille",
  ["popul forum"] = "Population forum : monthly newsletter of the Commission on Population",
  ["popul geogr"] = "Population geography : a journal of the Association of Population Geographers of India",
  ["popul headl"] = "Population headliners",
  ["popul health  metrics"] = "Population Health Metrics",
  ["popul health manag"] = "Population health management",
  ["popul health manage"] = "Population Health Management",
  ["popul health metr"] = "Population health metrics",
  ["popul index"] = "Population index",
  ["popul manag"] = "Population manager : ICOMP review",
  ["popul med"] = "Population medicine",
  ["popul newsl"] = "Population newsletter",
  ["popul notes"] = "Population notes (University of Wisconsin--Madison. Applied Population Laboratory)",
  ["popul policy compend"] = "Population policy compendium",
  ["popul rep"] = "Population report (Honolulu, Hawaii)",
  ["popul rep a"] = "Population reports. Series A, Oral contraceptives",
  ["popul rep b"] = "Population reports. Series B, Intrauterine devices",
  ["popul rep c"] = "Population reports. Series C: Sterilization [Female]",
  ["popul rep d"] = "Population reports (Washington, D.C.). Series D, Male sterilization",
  ["popul rep e"] = "Population reports. Series E, Law and policy",
  ["popul rep f"] = "Population reports. Series F, Pregnancy termination",
  ["popul rep g"] = "Population reports. Series G, Prostaglandins",
  ["popul rep h"] = "Population reports. Series H, Barrier methods",
  ["popul rep i"] = "Population reports. Series I, Periodic abstinence",
  ["popul rep j"] = "Population reports. Series J, Family planning programs",
  ["popul rep k"] = "Population reports. Series K, Injectables and implants",
  ["popul rep l"] = "Population reports. Series L, Issues in world health",
  ["popul rep m"] = "Population Reports. Series M, Special Topics",
  ["popul rep spec top monogr"] = "Population reports. Special topic monographs",
  ["popul res"] = "Population research (Peking, China)",
  ["popul res abstr"] = "Population research abstract : a contribution to research on India's population",
  ["popul res bull"] = "Population research bulletin. Pakistan Institute of Development Economics",
  ["popul res leads"] = "Population research leads",
  ["popul res policy rev"] = "Population research and policy review",
  ["popul res stud"] = "Population researches and studies",
  ["popul rev"] = "Population review",
  ["popul sci"] = "Population sciences (Cairo, Egypt)",
  ["popul soc (paris)"] = "Population et sociétés; bulletin mensuel d'informations démographiques, économiques, sociales",
  ["popul space place"] = "Population, space and place",
  ["popul stud (camb)"] = "Population studies",
  ["popul stud (ny)"] = "Population studies",
  ["popul stud japan"] = "Population Studies of Japan",
  ["popul stud transl ser"] = "Population studies translation series",
  ["popul talk point"] = "Population talk point",
  ["popul today"] = "Population today",
  ["popul trends"] = "Population trends",
  ["popul trends public policy"] = "Population trends and public policy",
  ["popul zpr"] = "Populační zprávy",
  ["population (ny)"] = "Population (New York, N.Y.)",
  ["population (paris)"] = "Population",
  ["population (wash dc)"] = "Population",
  ["population bull"] = "Population Bulletin",
  ["population community biol"] = "Population and Community Biology",
  ["population devel rev"] = "Population and Development Review",
  ["population res pol rev"] = "Population Research and Policy Review",
  ["population rev"] = "Population Review",
  ["population stud"] = "Population Studies",
  ["porcine health manag"] = "Porcine health management",
  ["porcine health manage"] = "Porcine Health Management",
  ["porcine res"] = "Porcine research",
  ["poromechanics v (2013)"] = "Poromechanics V : proceedings of the fifth Biot Conference on Poromechanics, July 10-12, 2013, Vienna, Austria. Biot Conference on Poromechanics (5th : 2013 : Vienna, Austria)",
  ["poroshk metall (kiev)"] = "Poroshkovaya Metallurgiya (Kiev)",
  ["port acta biol ser b sist ecol biogeogr paleontol"] = "Portugaliae acta biologica. Série B, Sistemática, ecologia, biogeografia e paleontologia",
  ["port electrochim acta"] = "Portugaliae Electrochimica Acta",
  ["port math"] = "Portugaliae Mathematica",
  ["port med"] = "Portugal médico",
  ["portland clin bull"] = "Bulletin. Portland, Or. Clinic",
  ["porto biomed j"] = "Porto biomedical journal",
  ["portugal math"] = "Portugaliae Mathematica",
  ["portuguese stud rev"] = "Portuguese studies review",
  ["pos hist"] = "Pós-história",
  ["pos proc sci"] = "PoS, proceedings of science",
  ["poseb izd od prir mat nauka"] = "Posebna izdanja. Odjeljenje prirodnih i matematiickih nauka",
  ["posebna izdan"] = "Posebna Izdanja",
  ["posit aware"] = "Positively aware : the monthly journal of the Test Positive Aware Network",
  ["posit dir news"] = "Positive Directions news : a support and information network of people with HIV/AIDS, their families, friends and providers",
  ["posit health news"] = "Positive health news",
  ["posit living"] = "Positive living (Los Angeles, Calif.)",
  ["posit outlook"] = "Positive outlook : the newsletter of the AIDS Alternative Health Project",
  ["positions (durh n c)"] = "Positions (Durham, N.C.)",
  ["positivity (dordr)"] = "Positivity",
  ["post hist"] = "Postal history journal",
  ["post reprod health"] = "Post reproductive health",
  ["post sov aff"] = "Post-Soviet affairs",
  ["post sov geogr"] = "Post-Soviet geography",
  ["post sov geogr econ"] = "Post-Soviet geography and economics",
  ["post-communist economies"] = "Post-Communist Economies",
  ["post-medieval archaeol"] = "Post-Medieval Archaeology",
  ["post-soviet aff"] = "Post-Soviet Affairs",
  ["post-soviet geogr econ"] = "Post-Soviet Geography and Economics",
  ["post\\polhk epy cybernet"] = "Polskie Towarzystwo Cybernetyczne",
  ["postal hist int"] = "Postal history international",
  ["postcommunist econ"] = "Post-communist economies",
  ["postdigital sci educ"] = "Postdigital Science and Education",
  ["postdiplom sem fiz"] = "Postdiplomski Seminar iz Fizike",
  ["postdoc j"] = "Postdoc journal : a journal of postdoctoral research and postdoctoral affairs",
  ["postep chir"] = "Postȩpy chirurgii",
  ["postep neurol neurochir psychiatr"] = "Postȩpy neurologii, neurochirurgii, i psychiatrii",
  ["postep reumatol"] = "Postȩpy reumatologii",
  ["postep wiedzy med"] = "Postȩpy wiedzy medycznej",
  ["postepy biochem"] = "Postepy biochemii",
  ["postepy dermatol alergol"] = "Postȩpy dermatologii i alergologii",
  ["postepy hig med dosw"] = "Postȩpy higieny i medycyny doświadczalnej",
  ["postepy hig med dosw (online)"] = "Postȩpy higieny i medycyny doświadczalnej (Online)",
  ["postepy kardiol interwencyjnej"] = "Postępy w kardiologii interwencyjnej = Advances in interventional cardiology",
  ["postgrad dent"] = "Postgraduate dentistry",
  ["postgrad doct afr"] = "Postgraduate doctor. Africa",
  ["postgrad med"] = "Postgraduate medicine",
  ["postgrad med j"] = "Postgraduate medical journal",
  ["postgrad semin am urol assoc north cent"] = "Postgraduate seminar. American Urological Association. North Central Section",
  ["postgraduate course ophthalmol"] = "Post-graduate course in ophthalmology; proceedings",
  ["postharvest biol technol"] = "Postharvest biology and technology",
  ["postmediev archaeol"] = "Post-medieval archaeology",
  ["potato res"] = "Potato Research",
  ["potential anal"] = "Potential Analysis",
  ["potravinarstvo slovak j food sci"] = "Potravinarstvo Slovak Journal of Food Sciences",
  ["potsdamer forsch naturwiss reihe b"] = "Potsdamer Forschungen. Naturwissenschaftliche Reihe B.",
  ["poult sci"] = "Poultry science",
  ["poumon coeur"] = "Le Poumon et le coeur",
  ["pour sci"] = "Pour la science",
  ["poverty public policy"] = "Poverty & public policy",
  ["povij pril"] = "Povijesni prilozi",
  ["povolz ekol z"] = "Povolzhskiĭ ėkologicheskiĭ zhurnal",
  ["powder diffr"] = "Powder Diffraction",
  ["powder metall"] = "Powder Metallurgy",
  ["powder metall met ceram"] = "Powder Metallurgy and Metal Ceramics",
  ["powder metall technol"] = "Powder Metallurgy Technology",
  ["powder technol"] = "Powder Technology",
  ["power electron devices compon"] = "Power Electronic Devices and Components",
  ["power electron power syst"] = "Power Electronics and Power Systems",
  ["power eng"] = "Power Engineer",
  ["power syst"] = "Power Systems",
  ["poznań stud philos sci humanities"] = "Poznań Studies in the Philosophy of the Sciences and the Humanities",
  ["pp news"] = "PP news",
  ["ppar res"] = "PPAR research",
  ["ppo spec rep"] = "PPO-special report",
  ["pr dejin prir ved"] = "Práce z dějin přírodních věd",
  ["pr inst mat nats akad nauk ukr mat zastos"] = "Pratsi Institutu Matematiki Natsionalnoi Akademii Nauk Ukraini. Matematika ta ii Zastosuvannya",
  ["pr j public health trop med"] = "The Puerto Rico journal of public health and tropical medicine",
  ["pr kom med dosw"] = "Prace Komisji Medycyny Doświadczalnej",
  ["pr lodz tow nauk [iv]"] = "Prace, Lodzkie Towarzystwo Naukowe, Wydzial IV, Nauk Lekarskich",
  ["pr lodz tow nauk iv"] = "Prace. Lódzkie Towarzystwo Naukowe. Wydział IV, Nauk Lekarskich",
  ["pr nauk inst technol nieorg nawozow miner politech wroclaw"] = "Prace Naukowe Instytutu Technologii Nieorganicznej i Nawozow Mineralnych Politechniki Wroclawskiej",
  ["pr nauk uniw śl katow"] = "Prace Naukowe Uniwersytetu Śląskiego w Katowicach",
  ["prac lek"] = "Pracovní lékar̆stvi",
  ["prac studi geograf"] = "Prace i studia geograficzne",
  ["prace ippt/iftr rep"] = "Prace IPPT/IFTR Reports",
  ["prace nauk inst cybernet tech politech wrocław"] = "Prace Naukowe Instytutu Cybernetyki Technicznej Politechniki Wrocławskiej",
  ["prace nauk inst cybernet tech politech wrocław ser monograf"] = "Politechnika Wrocławska",
  ["prace nauk inst inform stosow politech wroc"] = "Prace Naukowe Instytutu Informatyki Stosowanej Politechniki Wroclawskiej",
  ["prace nauk inst mat politech wrocław ser monograf"] = "Wrocław",
  ["pract anthropol"] = "Practicing anthropology",
  ["pract dent monogr"] = "Practical dental monographs",
  ["pract dermatol"] = "Practical dermatology",
  ["pract diabetes int"] = "Practical diabetes international : the journal for diabetes care teams worldwide",
  ["pract dig"] = "Practice digest",
  ["pract exp adv res comput 2018 (2018)"] = "Practice and Experience in Advanced Research Computing 2018 : Seamless Creativity : July 22-26 2017, Pittsburgh, Pennsylvania. Practice and Experience in Advanced Research Computing (Conference) (2018 : Pittsburgh, Pa.)",
  ["pract fail anal"] = "Practical Failure Analysis",
  ["pract farm"] = "Practica farmaceutică",
  ["pract gastroenterol"] = "Practical gastroenterology",
  ["pract innov (wash d c)"] = "Practice innovations (Washington, D.C.)",
  ["pract lab med"] = "Practical laboratory medicine",
  ["pract lawyer"] = "The Practical lawyer",
  ["pract midwife"] = "The practising midwife",
  ["pract neurol"] = "Practical neurology",
  ["pract neurol (fort wash pa)"] = "Practical neurology (Fort Washington, Pa.)",
  ["pract odontol"] = "Práctica odontológica",
  ["pract otorhinolaryngol (basel)"] = "Practica oto-rhino-laryngologica",
  ["pract pain manag"] = "Practical pain management",
  ["pract period hazard toxic radioact waste manage"] = "Practice Periodical of Hazardous, Toxic, Waste Management",
  ["pract period struct des constr"] = "Practical Periodical on Structural Design and Construction",
  ["pract periodontics aesthet dent"] = "Practical periodontics and aesthetic dentistry : PPAD",
  ["pract proced aesthet dent"] = "Practical procedures & aesthetic dentistry : PPAD",
  ["pract proteomics"] = "Practical Proteomics",
  ["pract psychol physicians"] = "Practical psychology for physicians",
  ["pract radiat oncol"] = "Practical radiation oncology",
  ["pract resour"] = "Practice resource",
  ["pract ser"] = "Practitioner Series",
  ["pract spectrosc"] = "Practical Spectroscopy",
  ["practice (birm)"] = "Practice (Birmingham, England)",
  ["prader willi perspect"] = "Prader-Willi perspectives : a journal of Prader-Willi Syndrome information for parents & professionals",
  ["praehist z"] = "Praehistorische Zeitschrift",
  ["pragmat case stud psychother"] = "Pragmatic case studies in psychotherapy : PCSP",
  ["pragmat obs res"] = "Pragmatic and observational research",
  ["prague econ pap"] = "Prague Economic Papers",
  ["prague med rep"] = "Prague medical report",
  ["prairie rose"] = "The Prairie rose",
  ["prakt anaesth"] = "Praktische Anästhesie, Wiederbelebung und Intensivtherapie",
  ["prakt arzt"] = "Der Praktische Arzt",
  ["prakt forstwirt schweiz"] = "praktische Forstwirt für die Schweiz, Der",
  ["prakt kieferorthop"] = "Praktische Kieferorthopadie",
  ["prakt lek"] = "Praktický lékar̆",
  ["prakt metallogr"] = "Praktische Metallographie",
  ["prakt tierarzt"] = "Der Praktische Tierarzt",
  ["prakt zubn lek"] = "Praktické zubní lékar̆ství",
  ["pramana - j phys"] = "Pramana – Journal of Physics",
  ["pras open"] = "PRAS open",
  ["prat anesth reanim"] = "Le praticien en anesthésie réanimation",
  ["prat organ soins"] = "Pratiques et organisation des soins",
  ["prax kinderpsychol kinderpsychiatr"] = "Praxis der Kinderpsychologie und Kinderpsychiatrie",
  ["prax kinderpsychol kinderpsychiatr beih"] = "Praxis der Kinderpsychologie und Kinderpsychiatrie. Beiheft",
  ["prax klin pneumol"] = "Praxis und Klinik der Pneumologie",
  ["prax kur"] = "Praxis-Kurier",
  ["prax naturwiss biol sch"] = "Praxis der Naturwissenschaften, Biologie in der Schule",
  ["prax naturwiss chem"] = "Praxis der Naturwissenschaften, Chemie",
  ["prax naturwiss chem sch"] = "Praxis der Naturwissenschaften, Chemie in der Schule",
  ["prax naturwiss phys"] = "Praxis der Naturwissenschaften, Physik",
  ["prax pneumol"] = "Praxis der Pneumologie",
  ["prax psychother"] = "Praxis der Psychotherapie",
  ["prax psychother psychosom"] = "Praxis der Psychotherapie und Psychosomatik",
  ["praxis (bern 1994)"] = "Praxis",
  ["praxis (oneonta n y)"] = "Praxis (Oneonta, N.Y.)",
  ["prb rep"] = "PRB report",
  ["pre- peri-nat psychol j"] = "Pre- and peri-natal psychology journal : official journal of the Pre and Peri-natal Psychology Association of North America",
  ["precambrian res"] = "Precambrian Research",
  ["precis agric"] = "Precision agriculture",
  ["precis anal trav"] = "Précis analytique des travaux. Académie des sciences, belles-lettres et arts de Rouen",
  ["precis cancer med"] = "Precision cancer medicine",
  ["precis chem"] = "Precision Chemistry",
  ["precis clin med"] = "Precision clinical medicine",
  ["precis eng"] = "Precision Engineering",
  ["precis med"] = "Precision medicine",
  ["precis med sci"] = "Precision Medicine Sciences",
  ["precis met"] = "Precision Metal",
  ["precis nanomed"] = "Precision nanomedicine",
  ["precis nutr"] = "Precision Nutrition",
  ["precis radiat\toncol"] = "Precision Radiation Oncology",
  ["precis radiat oncol"] = "Precision radiation oncology",
  ["preconf papers natl dent health conf (u s)"] = "[Preconference papers]. National Dental Health Conference",
  ["predict intell med"] = "PRedictive Intelligence in MEdicine. PRIME (Workshop)",
  ["pregnancy hypertens"] = "Pregnancy hypertension",
  ["prehosp disaster med"] = "Prehospital and disaster medicine",
  ["prehosp emerg care"] = "Prehospital emergency care : official journal of the National Association of EMS Physicians and the National Association of State EMS Directors",
  ["prehospital disaster med"] = "Prehospital and Disaster Medicine",
  ["prem cycle etud med"] = "Premier Cycle des Etudes Medicales",
  ["premed j columbia univ"] = "Pre-medical journal of Columbia University",
  ["prenat diagn"] = "Prenatal diagnosis",
  ["prenat neonatal med"] = "Prenatal and neonatal medicine : the international journal of basic and clinical research and practice",
  ["prensa med"] = "La Prensa médica",
  ["prensa med argent"] = "Prensa médica argentina",
  ["prensa med mex"] = "La Prensa médica mexicana",
  ["prensa pediatr rev am puericu pediatr"] = "Prensa pediátrica; revista americana de puericultura y pediatría",
  ["prentice hall inform system sci ser"] = "Prentice Hall Information and System Sciences Series",
  ["prentice hall int ser comput sci"] = "Prentice Hall International Series in Computer Science",
  ["prentice hall int ser indust systems eng"] = "Prentice Hall International Series in Industrial and Systems Engineering",
  ["prentice hall internat ser comput systems sci engrg"] = "Prentice Hall International Series in Computer Systems Science and Engineering",
  ["prentice hall ser adv commun technol"] = "Prentice Hall Series in Advanced Communications Technologies",
  ["prentice hall ser comput sci"] = "Prentice Hall Series in Computer Science",
  ["prentice hall ser environ intell manuf syst"] = "Prentice Hall Series on Environmental and Intelligent Manufacturing Systems",
  ["prentice hall ser statist"] = "Prentice Hall Series in Statistics",
  ["prentice hall ser tech math"] = "Prentice Hall Series in Technical Mathematics",
  ["prentice hall software ser"] = "Prentice Hall Software Series",
  ["prentice pract"] = "Prentice Practica",
  ["prep biochem"] = "Preparative biochemistry",
  ["prep biochem biotechnol"] = "Preparative biochemistry & biotechnology",
  ["prepod ist sk"] = "Prepodavanie istorii v shkole",
  ["prepr - am chem soc, div energy fuels"] = "Preprints - American Chemical Society, Division of Energy & Fuels",
  ["prepr - am chem soc, div pet chem"] = "Preprints - American Chemical Society, Division of Petroleum Chemistry",
  ["prepr ext abstr acs natl meet, am chem soc, div"] = "Environ. Chem. Preprints of Extended Abstracts presented at the ACS National Meeting, American Chemical Society, Division of Environmental Chemistry",
  ["prepr ext abstr acs natl meet, am chem soc, div environ chem"] = "Preprints of Extended Abstracts presented at the ACS National Meeting, American Chemical Society, Division of Environmental Chemistry",
  ["prepr pap - am chem soc, div fuel chem"] = "Preprint Papers - American Chemical Society, Division of Fuel Chemistry",
  ["prepr symp - am chem soc, div fuel chem"] = "Preprints of Symposia - American Chemical Society, Division of Fuel Chemistry",
  ["preprints (basel)"] = "Preprints",
  ["pres counc phys fit sports res dig"] = "President's Council on Physical Fitness and Sports research digest",
  ["pres stud q"] = "Presidential studies quarterly",
  ["presbyt st lukes hosp med bull"] = "Presbyterian-St. Luke's Hospital medical bulletin",
  ["prescr j"] = "Prescribers' journal",
  ["prescrire int"] = "Prescrire international",
  ["presence (camb)"] = "Presence (Cambridge, Mass.)",
  ["presence afr"] = "Présence africaine (Paris, France : 1967)",
  ["presence normande"] = "Présence normande",
  ["present tense"] = "Present tense (New York, N.Y.)",
  ["presse med"] = "Presse médicale (Paris, France : 1983)",
  ["presse therm clim"] = "La Presse thermale et climatique",
  ["prev assist dent"] = "Prevenzione & assistenza dentale",
  ["prev cardiol"] = "Preventive cardiology",
  ["prev chronic dis"] = "Preventing chronic disease",
  ["prev control"] = "Prevention and control : the official journal of the World Heart Federation",
  ["prev couns psychol"] = "Prevention in counseling psychology : theory, research, practice and training",
  ["prev hum serv"] = "Prevention in human services",
  ["prev med"] = "Preventive medicine",
  ["prev med (baltim)"] = "Preventive medicine",
  ["prev med community health"] = "Preventive medicine and community health",
  ["prev med rep"] = "Preventive medicine reports",
  ["prev nutr food sci"] = "Preventive nutrition and food science",
  ["prev oggi"] = "Prevenzione Oggi. ISPESL",
  ["prev res"] = "The prevention researcher",
  ["prev sch fail"] = "Preventing school failure : alternative education for children and youth",
  ["prev sci"] = "Prevention science : the official journal of the Society for Prevention Research",
  ["prev soc"] = "Prevencion social",
  ["prev stomatol"] = "Prevenzione stomatologica",
  ["prev strateg"] = "Prevention strategist",
  ["prev vet med"] = "Preventive veterinary medicine",
  ["previdenza soc"] = "Previdenza sociale",
  ["previs soc"] = "Prevision social",
  ["prib tekh e ksp [instrum exp tech (ussr)]"] = "Pribory i Tekhnika Eksperimenta [Instruments and Experimental Techniques (USSR)]",
  ["pride inst j long term home health care"] = "Pride Institute journal of long term home health care",
  ["prikl biokhim mikrobiol"] = "Prikladnaia biokhimiia i mikrobiologiia",
  ["prikl diskretn mat"] = "Prikladnaya Diskretnaya Matematika",
  ["prikl mat mekh"] = "Rossiiskaya Akademiya Nauk. Prikladnaya Matematika i Mekhanika",
  ["prikl mat model"] = "\\cyr Prikladnye Matematicheskie Modeli",
  ["prikl problemy proch i plast"] = "Gor\\cprime kovskiy Gosudarstvennyy Universitet imeni N",
  ["pril (makedon akad nauk umet odd med nauki)"] = "Prilozi (Makedonska akademija na naukite i umetnostite. Oddelenie za medicinski nauki)",
  ["pril odd prir-mat biotekhnichki nauki"] = "Prilozi. Oddelenie za Prirodno-Matematichki i Biotekhnichki Nauki. Contributions. Section of Natural, Mathematical and Biotechnical Sciences",
  ["prim care"] = "Primary care",
  ["prim care community psychiatr"] = "Primary care & community psychiatry",
  ["prim care companion cns disord"] = "The primary care companion for CNS disorders",
  ["prim care companion j clin psychiatry"] = "Primary care companion to the Journal of clinical psychiatry",
  ["prim care diabetes"] = "Primary care diabetes",
  ["prim care respir j"] = "Primary care respiratory journal : journal of the General Practice Airways Group",
  ["prim care update ob gyns"] = "Primary care update for Ob/Gyns",
  ["prim dent care"] = "Primary dental care : journal of the Faculty of General Dental Practitioners (UK)",
  ["prim dent j"] = "Primary dental journal",
  ["prim health care"] = "Primary health care : open access",
  ["prim health care res dev"] = "Primary health care research & development",
  ["prim prev insights"] = "Primary prevention insights",
  ["prim psychiatry"] = "Primary psychiatry",
  ["prim sens neuron"] = "Primary sensory neuron : the international interdisciplinary journal reporting basic and clinical research on sensory receptors and primary afferent neurons",
  ["primary cardiol"] = "Primary cardiology",
  ["primate biol"] = "Primate biology",
  ["primate conserv"] = "Primate conservation : the newsletter and journal of the IUCN/SSC Primate Specialist Group",
  ["primate rep"] = "Primate report",
  ["primates med"] = "Primates in medicine",
  ["primena radioakt izot joniz zrac med"] = "Primena radioaktivnih izotopa i jonizujućih zrac̆enja u medicini; bilten",
  ["primers complex syst"] = "Primers in Complex Systems",
  ["primus (terre ht)"] = "PRIMUS : problems, resources, and issues in mathematics undergraduate studies",
  ["princ pract clin res"] = "Principles and practice of clinical research (2015)",
  ["princ viana"] = "Príncipe de Viana",
  ["princess takamatsu symp"] = "Princess Takamatsu symposia",
  ["princet found contemp philos"] = "Princeton Foundations of Contemporary Philosophy",
  ["princet hist"] = "Princeton history",
  ["princet j bioeth"] = "Princeton journal of bioethics",
  ["princet leg libr"] = "Princeton Legacy Library",
  ["princet lifesav study guide"] = "Princeton Lifesaver Study Guide",
  ["princet ser finance"] = "Princeton Series in Finance",
  ["princeton landmarks math"] = "Princeton Landmarks in Mathematics",
  ["princeton landmarks phys"] = "Princeton Landmarks in Physics",
  ["princeton math ser"] = "Princeton Mathematical Series",
  ["princeton paperbacks"] = "Princeton Paperbacks",
  ["princeton sci lib"] = "Princeton Science Library",
  ["princeton ser appl math"] = "Princeton Series in Applied Mathematics",
  ["princeton ser astrophys"] = "Princeton Series in Astrophysics",
  ["princeton ser comput sci"] = "Princeton Series in Computer Science",
  ["princeton ser phys"] = "Princeton Series in Physics",
  ["princeton ser theor comput biol"] = "Princeton Series in Theoretical and Computational Biology",
  ["princeton univ libr chron"] = "The Princeton University library chronicle",
  ["print q"] = "Print quarterly",
  ["prir clovek zdravje"] = "Priroda, c̆lovek in zdravje",
  ["priroda (moscow)"] = "Priroda (Moscow)",
  ["prison j"] = "The Prison journal",
  ["prison serv j"] = "Prison service journal",
  ["prispev novejso zgod"] = "Prispevki za novejšo zgodovino",
  ["priv pact"] = "Private practice",
  ["priv stat databases"] = "Privacy in statistical databases. PSD (Conference : 2004-)",
  ["prn forum"] = "PRN forum",
  ["pro fam inf"] = "Pro Familia Informationen",
  ["pro fono"] = "Pró-fono : revista de atualização científica",
  ["pro infirm"] = "Pro Infirmis",
  ["pro math"] = "Pro Mathematica",
  ["pro med arztl ref"] = "Pro medico; ärztliches Referatenblatt",
  ["pro re nata"] = "Pro re nata PRN : the official publication of the Utah Nurses' Association",
  ["prob desarrollo"] = "Problemas del Desarrollo",
  ["prob econ"] = "Problems of Economics",
  ["prob econ transition"] = "Problems of Economic Transition",
  ["probab appl"] = "Probability and its Applications",
  ["probab appl (n y)"] = "Probability and its Applications (New York)",
  ["probab eng inf sci"] = "Probability in the Engineering and Informational Sciences",
  ["probab eng inform sci"] = "Probability in the Engineering and Informational Sciences",
  ["probab eng mech"] = "Probabilistic Engineering Mechanics",
  ["probab engrg inform sci"] = "Probability in the Engineering and Informational Sciences",
  ["probab math phys"] = "Probability and Mathematical Physics",
  ["probab math statist"] = "Probability and Mathematical Statistics",
  ["probab stochastics ser"] = "Probability and Stochastics Series",
  ["probab surv"] = "Probability Surveys",
  ["probab theory relat fields"] = "Probability theory and related fields",
  ["probab theory related fields"] = "Probability Theory and Related Fields",
  ["probab theory stoch model"] = "Probability Theory and Stochastic Modelling",
  ["probab uncertain quant risk"] = "Probability, uncertainty and quantitative risk",
  ["probab, uncertainty quant risk"] = "Probability, Uncertainty and Quantitative Risk",
  ["probate law j"] = "Probate law journal",
  ["probate prop"] = "Probate and property (Chicago, Ill. : 1987)",
  ["probe (adelaide)"] = "Probe",
  ["probe (lond)"] = "The Probe",
  ["probiotics antimicrob proteins"] = "Probiotics and antimicrobial proteins",
  ["probl actuels endocrinol nutr"] = "Problèmes actuels d'endocrinologie et de nutrition",
  ["probl actuels otorhinolaryngol"] = "Problèmes actuels d'oto-rhino-laryngologie",
  ["probl am lat"] = "Problèmes d'Amérique latine",
  ["probl anal issues anal"] = "Problemy Analiza. Issues of Analysis",
  ["probl biokhimii michurinskoi biol"] = "Problemy biokhimii v michurinskoĭ biologii",
  ["probl books in math"] = "Problem Books in Mathematics",
  ["probl cardiol"] = "Probleme de cardiologie",
  ["probl desarro"] = "Problemas del desarrollo",
  ["probl econ"] = "Problems of economics",
  ["probl econ transit"] = "Problems of economic transition",
  ["probl endokrinol (mosk)"] = "Problemy ėndokrinologii",
  ["probl endokrinol gormonoter"] = "Problemy ėndokrinologii i gormonoterapii",
  ["probl fiz mat tekh"] = "Problemy Fiziki, Matematiki i Tekhniki",
  ["probl fiziol akust"] = "Problemy fiziologicheskoi akustiki",
  ["probl fizol opt"] = "Problemy fiziologicheskoĭ optiki",
  ["probl gematol pereliv krovi"] = "Problemy gematologii i perelivaniia krovi",
  ["probl geogr"] = "Problemi na geografii︠a︡ta",
  ["probl globale omen"] = "Probleme Globale ale Omenirii",
  ["probl hematol blood transfus"] = "Problems of hematology and blood transfusion. Problemy gematologii i perelivaniia krovi",
  ["probl inf transm"] = "Problems of Information Transmission",
  ["probl istor demogr sssr"] = "Problemy istoricheskoĭ demografii SSSR",
  ["probl khig"] = "Problemi na khigienata",
  ["probl kosm biol"] = "Problemy Kosmicheskoi Biologii",
  ["probl lesoved i lesovod"] = "Problemy lesovedeniia i lesovodstva : sbornik nauchnykh trudov IL NAN Belarusi",
  ["probl mat fiz ser pentru liceu"] = "Probleme de Matematica si Fizica: Serie Pentru Liceu",
  ["probl math écrit capes 1991–1996"] = "Problèmes de Mathématiques",
  ["probl med mikol"] = "Problemy medicinskoj mikologii",
  ["probl med wieku rozwoj"] = "Problemy medycyny wieku rozwojowego",
  ["probl oncol"] = "Problems of oncology. Voprosy onkologii",
  ["probl onkol"] = "Problemi na onkologiiata",
  ["probl osobo opas infekc"] = "Problemy osobo opasnykh infektsiĭ",
  ["probl patol infect"] = "Probleme de patologie infecţioasă [din literatura sovietică de specialitate]",
  ["probl prochn"] = "Problemy Prochnosti",
  ["probl program"] = "Problemy Programmirovaniya. Problems in Programming",
  ["probl radiac med radiobiol"] = "Problemy radiat︠s︡iĭnoï medyt︠s︡yny ta radiobiolohiï",
  ["probl reumatol"] = "Probleme de reumatologie",
  ["probl rodziny"] = "Problemy rodziny",
  ["probl sev"] = "Problemy Severa",
  ["probl solving math beyond"] = "Problem Solving in Mathematics and Beyond",
  ["probl sotsialnoi gig istor med"] = "Problemy sot︠s︡ialʹnoĭ gigieny i istorii︠a︡ medit︠s︡iny",
  ["probl sotsialnoi gig zdravookhranenniiai istor med"] = "Problemy sot︠s︡ialʹnoĭ gigieny, zdravookhranenii︠a︡ i istorii medit︠s︡iny",
  ["probl sovrem neirokhirurgii"] = "Problemy sovremennoĭ neĭrokhirurgii",
  ["probl symp aktuell ther"] = "Probleme. Symposien aktueller therapeutischer",
  ["probl ter"] = "Probleme de terapeutică",
  ["probl tuberculoza"] = "Probleme de tuberculoză",
  ["probl tuberk"] = "Problemy tuberkuleza",
  ["probl tuberk bolezn legk"] = "Problemy tuberkuleza i bolezneĭ legkikh",
  ["probl vet med"] = "Problems in veterinary medicine",
  ["probl virol"] = "Problems of virology",
  ["problem books in math"] = "Problem Books in Mathematics",
  ["problemi tekhn kibernet robot"] = "Problemi na Tekhnicheskata Kibernetika i Robotikata",
  ["problems inform transmission"] = "Problems of Information Transmission",
  ["problemy peredachi informatsii"] = "Rossiiskaya Akademiya Nauk. Problemy Peredachi Informatsii",
  ["problemy teor gravitatsii i element chastits"] = "Problemy Teorii Gravitatsii i Èlementarnykh Chastits",
  ["problemy upravlen inform"] = "Natsionalnaya Akademiya Nauk Ukrainy. Problemy Upravleniya i Informatiki",
  ["problemy yadern fiz i kosm luchey"] = "Problemy Yadernoy Fiziki i Kosmicheskikh Luchey",
  ["proc (bayl univ med cent)"] = "Proceedings (Baylor University. Medical Center)",
  ["proc (ieee int conf multimed expo)"] = "Proceedings. IEEE International Conference on Multimedia and Expo",
  ["proc (usnc ursi radio sci meet)"] = "Proceedings. USNC-URSI Radio Science Meeting",
  ["proc - aesf annu tech conf"] = "Proceedings of the AESF Annual Technical Conference",
  ["proc - electrochem soc"] = "Proceedings - Electrochemical Society",
  ["proc - electron compon conf"] = "Proceedings - Electronic Components & Technology Conference",
  ["proc - indian acad sci, chem sci"] = "Proceedings - Indian Academy of Sciences, Chemical Sciences",
  ["proc 17th int web all conf (2020)"] = "Proceedings of the 17th International Web for All Conference",
  ["proc 18th acm int conf interact des child (2019)"] = "Proceedings of the 18th ACM International Conference on Interaction Design and Children",
  ["proc 1st acm sigspatial workshop geospat humanit (2017)"] = "Proceedings of the 1st ACM SIGSPATIAL Workshop on Geospatial Humanities. ACM SIGSPATIAL Workshop on Geospatial Humanities (1st : 2017 : Redondo Beach, Calif.)",
  ["proc 1st workshop sch web min (2017)"] = "Proceedings of the 1st Workshop on Scholarly Web Mining. Workshop on Scholarly Web Mining (1st : 2017 : Cambridge, England)",
  ["proc 2012 beliv workshop (2012)"] = "Proceedings of the 2012 BELIV Workshop : beyond time and errors - novel evaluation methods for visualization : Seattle WA, USA - October 14-15, 2012. BELIV (Conference) (2012 : Seattle, Wash.)",
  ["proc 2014 6th int adv res workshop in silico oncol cancer investig (2014)"] = "Proceedings of the 2014 6th International Advanced Research Workshop on In Silico Oncology and Cancer Investigation : the CHIC Project Workshop (IARWISOCI) : Athens, Greece, 3-4 November 2014. International Advanced Research Workshop on In Silico Oncology and Cancer Investigation (6th : 2014 : Athens, Greece)",
  ["proc 2014 acm southeast reg conf"] = "Proceedings of the 2014 ACM Southeast Regional Conference",
  ["proc 2015 resil week rsw (2015)"] = "Proceedings, 2015 Resilience Week (RSW) : Hilton Philadelphia at Penn's Landing, Philadelphia, PA, 18-20 August, 2015. Resilience Week (2015 : Philadelphia, Pa.)",
  ["proc 2015 workshop vis anal healthc (2015)"] = "Proceedings of the 2015 Workshop on Visual Analytics in Healthcare",
  ["proc 2nd workshop hum loop data anal (2017)"] = "Proceedings of the 2nd Workshop on Human-In-the-Loop Data Analytics. Workshop on Human-In-the-Loop Data Analytics (2nd : 2017 : Chicago, Ill.)",
  ["proc 5th int workshop explor search databases web (2018)"] = "Proceedings of the 5th International Workshop on Exploratory Search in Databases and the Web. International Workshop on Exploratory Search in Databases and the Web (5th : 2018 : Houston, Tex.)",
  ["proc a"] = "Proceedings A",
  ["proc a razmadze math inst"] = "Georgian Academy of Sciences",
  ["proc aaai artif intell interact digit enterain conf"] = "Proceedings. AAAI Artificial Intelligence and Interactive Digital Entertainment Conference",
  ["proc acad nat sci phila"] = "Proceedings. Academy of Natural Sciences of Philadelphia",
  ["proc acad nat sci philadelphia"] = "Proceedings of the Academy of Natural Sciences of Philadelphia",
  ["proc acad polit sci"] = "Proceedings of the Academy of Political Science",
  ["proc acm comput graphics comput syst"] = "Proceedings of the ACM on Computer Graphics and Interactive Techniques",
  ["proc acm conf health inference learn (2020)"] = "Proceedings of the ACM Conference on Health, Inference, and Learning",
  ["proc acm hum comput interact"] = "Proceedings of the ACM on human-computer interaction",
  ["proc acm interact mob wearable ubiquitous technol"] = "Proceedings of the ACM on interactive, mobile, wearable and ubiquitous technologies",
  ["proc acm interact mobile wearable ubiquitous technol"] = "Proceedings of the ACM on Interactive, Mobile, Wearable and Ubiquitous Technologies",
  ["proc acm manage data"] = "Proceedings of the ACM on Management of Data",
  ["proc acm meas anal comput syst"] = "Proceedings of the ACM on Measurement and Analysis of Computing Systems",
  ["proc acm networking"] = "Proceedings of the ACM on Networking",
  ["proc acm program lang"] = "Proceedings of the ACM on Programming Languages",
  ["proc acm siggraph symp interact 3d graph games"] = "Proceedings. ACM SIGGRAPH Symposium on Interactive 3D Graphics and Games",
  ["proc acm sigmod int conf manag data"] = "Proceedings. ACM-Sigmod International Conference on Management of Data",
  ["proc acm software eng"] = "Proceedings of the ACM on Software Engineering",
  ["proc acm symp doc eng"] = "Proceedings of the ACM Symposium on Document Engineering. ACM Symposium on Document Engineering",
  ["proc acm symp user interface softw tech"] = "Proceedings of the ACM Symposium on User Interface Software and Technology. ACM Symposium on User Interface Software and Technology",
  ["proc acm symp virtual real softw technol"] = "Proceedings of the ACM Symposium on Virtual Reality Software and Technology. ACM Symposium on Virtual Reality Software and Technology",
  ["proc air waste manage assoc meet"] = "Annual meeting & exhibition proceedings CD-ROM. Air & Waste Management Association. Meeting",
  ["proc am antiq soc"] = "Proceedings of the American Antiquarian Society. American Antiquarian Society",
  ["proc am assoc cancer res"] = "Proceedings of the American Association for Cancer Research",
  ["proc am cathol philos assoc"] = "Proceedings of the American Catholic Philosophical Association",
  ["proc am fed clin res"] = "Proceedings. American Federation for Clinical Research",
  ["proc am math soc"] = "Proceedings of the American Mathematical Society. American Mathematical Society",
  ["proc am philos soc"] = "Proceedings of the American Philosophical Society",
  ["proc am soc agron"] = "Proceedings of the American Society of Agronomy",
  ["proc am soc brew chem"] = "Proceedings of the American Society of Brewing Chemists",
  ["proc am soc inf sci technol"] = "Proceedings of the American Society for Information Science & Technology",
  ["proc am stat assoc"] = "Proceedings. American Statistical Association. Annual Meeting",
  ["proc am thorac soc"] = "Proceedings of the American Thoracic Society",
  ["proc amer math soc"] = "Proceedings of the American Mathematical Society",
  ["proc amer math soc ser b"] = "Proceedings of the American Mathematical Society. Series B",
  ["proc amia annu fall symp"] = "Proceedings : a conference of the American Medical Informatics Association. AMIA Fall Symposium",
  ["proc amia symp"] = "Proceedings. AMIA Symposium",
  ["proc anal div chem soc"] = "Proceedings of the Analytical Division of the Chemical Society",
  ["proc anim care panel"] = "Proceedings. Animal Care Panel (U.S.)",
  ["proc annu biol colloq"] = "Proceedings. Colloquium in Biological Sciences",
  ["proc annu clin spinal cord inj conf"] = "Proceedings. Clinical Spinal Cord Injury Conference",
  ["proc annu conf progn health manag soc"] = "Proceedings of the Annual Conference of the Prognostics and Health Management Society. Prognostics and Health Management Society. Conference",
  ["proc annu conf res med educ"] = "Proceedings of the Annual Conference on Research in Medical Education",
  ["proc annu manage conf am dent assoc"] = "Proceedings, Annual Management Conference - American Dental Association",
  ["proc annu meet"] = "Proceedings of annual meeting. Industrial Hygiene Foundation of America",
  ["proc annu meet am assoc cancer res"] = "Proceedings. American Association for Cancer Research. Annual Meeting",
  ["proc annu meet am psychopathol assoc"] = "Proceedings of the annual meeting of the American Psychopathological Association",
  ["proc annu meet am stud health assoc"] = "Proceedings of the annual meeting. American Student Health Association",
  ["proc annu meet am stud health assoc pac coast  sect"] = "Proceedings [of the] annual meeting. American Student Health Association. Pacific Coast Section",
  ["proc annu meet cent soc clin res u s"] = "Proceedings [of the] annual meeting. Central Society for Clinical Research (U.S.)",
  ["proc annu meet med sect am counc life insur"] = "Proceedings, the annual meeting of the Medical Section of the American Council of Life Insurance",
  ["proc annu meet med sect am life conv"] = "Proceedings, annual meeting of the Medical Section of the American Life Convention",
  ["proc annu meet med sect am life insur assoc"] = "Proceedings, the annual meeting of the Medical Section of the American Life Insurance Association",
  ["proc annu meet n j  mosqu exterm assoc"] = "Proceedings, annual meeting - New Jersey Mosquito Extermination Association",
  ["proc annu meet u s anim health assoc"] = "Proceedings, annual meeting of the United States Animal Health Association",
  ["proc annu meet u s livest sanit assoc"] = "Proceedings, annual meeting of the United States Livestock Sanitary Association",
  ["proc annu symp comput appl med care"] = "Proceedings. Symposium on Computer Applications in Medical Care",
  ["proc annu symp eugen soc"] = "Proceedings of the annual symposium of the Eugenics Society",
  ["proc appl math mech"] = "Proceedings in Applied Mathematics and Mechanics",
  ["proc asme conf smart mater adapt struct intell syst"] = "Proceedings of the ASME Conference on Smart Materials, Adaptive Structures, and Intelligent Systems. ASME Conference on Smart Materials, Adaptive Structures, and Intelligent Systems",
  ["proc asme dyn syst control conf"] = "Proceedings of the ASME Dynamic Systems and Control Conference. ASME Dynamic Systems and Control Conference",
  ["proc asme int conf manuf sci eng"] = "Proceedings of the ASME International Conference on Manufacturing Science and Engineering. ASME International Conference on Manufacturing Science and Engineering",
  ["proc asme micro nanoscale heat mass transf int conf (2012)"] = "Proceedings of the ASME Micro/Nanoscale Heat and Mass Transfer International Conference -- 2012 : presented at the ASME 2012 3rd Micro/Nanoscale Heat and Mass Transfer International Conference, March 3-6, 2012, Atlanta, Georgia. Micro/Nanoscale Heat and Mass Transfer International Conference (2012 : Atlanta, Georgia)",
  ["proc asme verification valid symp 2019 (2019)"] = "Proceedings of the ASME Verification and Validation Symposium -2019- : VVS2019 : presented at ASME 2019 Verification and Validation Symposium, May 15-17, 2019, Las Vegas, Nevada, USA. ASME Verification and Validation Symposium (2019 : Las Vegas, Nev.)",
  ["proc assoc am physicians"] = "Proceedings of the Association of American Physicians",
  ["proc assoc inf sci technol"] = "Proceedings of the Association for Information Science and Technology. Association for Information Science and Technology",
  ["proc aust assoc neurol"] = "Proceedings of the Australian Association of Neurologists",
  ["proc aust physiol pharmacol soc"] = "Proceedings of the Australian Physiological and Pharmacological Society",
  ["proc biol sci"] = "Proceedings. Biological sciences",
  ["proc biol soc wash"] = "Proceedings. Biological Society of Washington",
  ["proc bot soc br isles"] = "Proceedings of the Botanical Society of the British Isles. Botanical Society of the British Isles",
  ["proc br acad"] = "Proceedings of the British Academy",
  ["proc br inst radio eng"] = "Proceedings of the British Institution of Radio Engineers",
  ["proc br paedod soc"] = "Proceedings of the British Paedodontic Society",
  ["proc br soc dent maxillofac radiol"] = "Proceedings of the British Society of Dental and Maxillofacial Radiology",
  ["proc calif acad sci"] = "Proceedings. California Academy of Sciences",
  ["proc camb philol soc"] = "Proceedings Of The Cambridge Philosophical Society",
  ["proc camb philos soc"] = "Proceedings Of The Cambridge Philosophical Society",
  ["proc cambridge philos soc"] = "Proceedings of the Cambridge Philosophical Society",
  ["proc can cancer conf"] = "Proceedings. Canadian Cancer Conference",
  ["proc can conf comput robot vis"] = "Proceedings. Canadian Conference on Computer and Robot Vision",
  ["proc cardiff med soc"] = "The Proceedings of the Cardiff Medical Society",
  ["proc cent math appl"] = "Proceedings of the Centre for Mathematics and its Applications",
  ["proc centre math appl austral nat univ"] = "Proceedings of the Centre for Mathematics and its Applications, Australian National University",
  ["proc chem soc"] = "Proceedings. Chemical Society (Great Britain)",
  ["proc chin acad med sci peking union med coll"] = "Proceedings of the Chinese Academy of Medical Sciences and the Peking Union Medical College = Chung-kuo i hsüeh k'o hsüeh yüan, Chung-kuo hsieh ho i k'o ta hsüeh hsüeh pao",
  ["proc clin dial transplant forum"] = "Proceedings of the Clinical Dialysis and Transplant Forum",
  ["proc combust inst"] = "Proceedings of the Combustion Institute",
  ["proc compsac"] = "Proceedings : Annual International Computer Software and Applications Conference. COMPSAC",
  ["proc conf"] = "Proceedings of the conference. Association for Computational Linguistics. North American Chapter. Meeting",
  ["proc conf army physicians cent mediterr forces"] = "Proceedings of the Conference of Army Physicians, Central Mediterranean Forces : held at the Institute superiore di sanità : Rome, 29th January to 3rd February 1945. Conference of Army Physicians, Central Mediterranean Forces (1945 : Istituto superiore di sanità)",
  ["proc conf assoc comput linguist meet"] = "Proceedings of the conference. Association for Computational Linguistics. Meeting",
  ["proc conf empir methods nat lang process"] = "Proceedings of the Conference on Empirical Methods in Natural Language Processing. Conference on Empirical Methods in Natural Language Processing",
  ["proc conf hazard waste res"] = "Proceedings of the Conference on Hazardous Waste Research. Conference on Hazardous Waste Research (Kansas State University)",
  ["proc congress c m f army surg"] = "Proceedings of the Congress of C.M.F. Army Surgeons. Congress of C.M.F. Army Surgeons (1945 : Rome, Italy)",
  ["proc consort revolut eur"] = "Proceedings. Consortium on Revolutionary Europe, 1750-1850",
  ["proc counc mass med soc"] = "Proceedings of the council - Massachusetts Medical Society",
  ["proc data compress conf"] = "Proceedings. Data Compression Conference",
  ["proc des autom conf"] = "Proceedings. Design Automation Conference",
  ["proc dorset nat hist archeol soc"] = "Proceedings of the Dorset Natural History and Archaeological Society. Dorset Natural History and Archaeological Society",
  ["proc edinb math soc (2)"] = "Proceedings of the Edinburgh Mathematical Society. Series II",
  ["proc edinburgh math soc"] = "Proceedings of the Edinburgh Mathematical Society",
  ["proc edinburgh math soc (2)"] = "Proceedings of the Edinburgh Mathematical Society",
  ["proc edinburgh math soc sect a: math"] = "Proceedings of the Edinburgh Mathematical Society Section A: Mathematics",
  ["proc eighth wirel stud stud stud workshop (2016)"] = "Proceedings of the Eighth Wireless of the Students, by the Students, and for the Students Workshop. Workshop on Wireless of the Students, by the Students, for the Students (8th : 2016 : New York, N.Y.)",
  ["proc entomol soc b c"] = "Proceedings of the Entomological Society of British Columbia",
  ["proc entomol soc london"] = "Proceedings of the Entomological Society of London",
  ["proc entomol soc wash"] = "Proceedings of the Entomological Society of Washington",
  ["proc est acad sci"] = "Proceedings of the Estonian Academy of Sciences",
  ["proc est acad sci eng"] = "Proceedings of the Estonian Academy of Sciences",
  ["proc est acad sci, chem"] = "Proceedings of the Estonian Academy of Sciences, Chemistry",
  ["proc estonian acad sci biol ecol"] = "Proceedings of the Estonian Academy of Sciences. Biology, ecology = Eesti Teaduste Akadeemia toimetised. Bioloogia, ökoloogia",
  ["proc estonian acad sci phys math"] = "Estonian Academy of Sciences",
  ["proc eur conf antennas propag"] = "Proceedings of the European Conference on Antennas and Propagation. European Conference on Antennas and Propagation",
  ["proc eur dial transplant assoc"] = "Proceedings of the European Dialysis and Transplant Association. European Dialysis and Transplant Association",
  ["proc eur dial transplant assoc eur ren assoc"] = "Proceedings of the European Dialysis and Transplant Association - European Renal Association. European Dialysis and Transplant Association - European Renal Association. Congress",
  ["proc eur prosthodontic assoc"] = "Proceedings of the European Prosthodontic Association",
  ["proc euromicro int conf parallel distrib netw based process"] = "Proceedings. Euromicro International Conference on Parallel, Distributed, and Network-based Processing",
  ["proc exploredb17 (2017)"] = "Proceedings of the ExploreDB'17. International Workshop on Exploratory Search in Databases and the Web (4th : 2017 : Chicago, Ill.)",
  ["proc eye track res appl symp"] = "Proceedings. Eye Tracking Research & Applications Symposium",
  ["proc finn dent soc"] = "Proceedings of the Finnish Dental Society. Suomen Hammaslääkäriseuran toimituksia",
  ["proc first acm sigcas conf comput sustain soc compass 2018 (2018)"] = "Proceedings of the First ACM SIGCAS Conference on Computing and Sustainable Societies (COMPASS) 2018 : Tech Museum, San Jose, CA, USA, Facebook, Menlo Park, CA, USA, June 20-22, 2018. ACM SIGCAS Conference on Computing and Sustainable Societies (1st : 2018 : San Jose, Calif. ; Menlo Park, Calif.)",
  ["proc first jt nasa cardiopulm workshop"] = "Proceedings of the First Joint NASA Cardiopulmonary Workshop. Joint NASA Cardiopulmonary Workshop (1st : 1990 : Houston, Tex.)",
  ["proc fla conf recent adv robot"] = "Proceedings. Florida Conference on Recent Advances in Robotics",
  ["proc forum acust"] = "Proceedings of Forum Acusticum",
  ["proc found orthod res"] = "Proceedings of the Foundation for Orthodontic Research",
  ["proc fourth workshop data anal scale danac 2015 (2015)"] = "Proceedings of the Fourth Workshop on Data analytics at sCale (DanaC 2015) : May 31st, 2015, Melbourne, Australia. Workshop on Data Analytics in the Cloud (4th : 2015 : Melbourne, Vic.)",
  ["proc front converg biosci inf technol (2007)"] = "Proceedings of the Frontiers in the Convergence of Bioscience and Information Technologies : Jeju Island, Korea, October 11-13, 2007. Frontiers in the Convergence of Bioscience and Information Technologies (2007 : Cheju-do, Korea)",
  ["proc genet evol comput conf"] = "Proceedings of the Genetic and Evolutionary Computation Conference. Genetic and Evolutionary Computation Conference",
  ["proc geol assoc"] = "Proceedings of the Geologists’ Association",
  ["proc glasgow math assoc"] = "Proceedings of the Glasgow Mathematical Association",
  ["proc gov technol inf policies (2010)"] = "Proceedings, Governance of technology, information and policies : addressing the challenges of worldwide interconnectivity. Workshop on Governance of Technology, Information and Policies (2010 : Austin, Tex.)",
  ["proc gulf caribb fish inst univ miami"] = "Proceedings of the annual session - Gulf and Caribbean Fisheries Institute. Gulf and Caribbean Fisheries Institute",
  ["proc hawaii entomol soc"] = "Proceedings of the Hawaiian Entomological Society. Hawaiian Entomological Society",
  ["proc health policy forum"] = "Proceedings of the Health Policy Forum. Health Policy Forum",
  ["proc helminthol soc wash"] = "Proceedings of the Helminthological Society of Washington",
  ["proc huguenot soc lond"] = "Proceedings of the Huguenot Society of London. Huguenot Society of London",
  ["proc hum factors ergon soc annu meet"] = "Proceedings of the Human Factors and Ergonomics Society Annual Meeting",
  ["proc i vekua inst appl math"] = "Proceedings of Ilia Vekua Institute of Applied Mathematics",
  ["proc ibm med symp"] = "Proceedings. IBM Medical Symposium",
  ["proc ieee"] = "Proceedings of the IEEE",
  ["proc ieee comput soc bioinform conf"] = "Proceedings. IEEE Computer Society Bioinformatics Conference",
  ["proc ieee comput soc conf comput vis pattern recognit"] = "Proceedings. IEEE Computer Society Conference on Computer Vision and Pattern Recognition",
  ["proc ieee comput syst bioinform conf"] = "Proceedings. IEEE Computational Systems Bioinformatics Conference",
  ["proc ieee conf intell transp syst"] = "Proceedings. IEEE Conference on Intelligent Transportation Systems",
  ["proc ieee conf virtual real 3d user interfaces"] = "Proceedings. IEEE Conference on Virtual Reality and 3D User Interfaces",
  ["proc ieee glob humanit technol conf"] = "Proceedings. IEEE Global Humanitarian Technology Conference",
  ["proc ieee inst electr electron eng"] = "Proceedings of the IEEE. Institute of Electrical and Electronics Engineers",
  ["proc ieee int autom softw eng conf"] = "Proceedings. IEEE International Automated Software Engineering Conference",
  ["proc ieee int conf  semant comput"] = "Proceedings. IEEE International Conference on Semantic Computing",
  ["proc ieee int conf cloud eng"] = "Proceedings of the IEEE International Conference on Cloud Engineering. IEEE International Conference on Cloud Engineering",
  ["proc ieee int conf clust comput"] = "Proceedings. IEEE International Conference on Cluster Computing",
  ["proc ieee int conf comput vis"] = "Proceedings. IEEE International Conference on Computer Vision",
  ["proc ieee int conf data min"] = "Proceedings. IEEE International Conference on Data Mining",
  ["proc ieee int conf healthc inform imaging syst biol"] = "Proceedings. IEEE International Conference on Healthcare Informatics, Imaging and Systems Biology",
  ["proc ieee int conf micro electro mech syst"] = "Proceedings. IEEE International Conference on Micro Electro Mechanical Systems",
  ["proc ieee int conf robot autom"] = "Proceedings. IEEE International Conference on Robotics and Automation",
  ["proc ieee int congr big data"] = "Proceedings. IEEE International Congress on Big Data",
  ["proc ieee int requir eng conf"] = "Proceedings. IEEE International Requirements Engineering Conference",
  ["proc ieee int symp bioinformatics bioeng"] = "Proceedings. IEEE International Symposium on Bioinformatics and Bioengineering",
  ["proc ieee int symp biomed imaging"] = "Proceedings. IEEE International Symposium on Biomedical Imaging",
  ["proc ieee int symp comput based med syst"] = "Proceedings. IEEE International Symposium on Computer-Based Medical Systems",
  ["proc ieee int symp field program cust comput mach"] = "Proceedings IEEE International Symposium on Field-Programmable Custom Computing Machines : FCCM 2011 : 1-3 May 2011, Salt Lake City, Utah, USA. IEEE Symposium on FPGAs for Custom Computing Machines (19th : 2011 : Salt Lake City, Utah)",
  ["proc ieee int symp high assur syst eng"] = "Proceedings. IEEE International Symposium on High-Assurance Systems Engineering",
  ["proc ieee int symp info theory"] = "Proceedings. IEEE International Symposium on Information Theory",
  ["proc ieee int symp policies distrib syst netw"] = "Proceedings. IEEE International Symposium on Policies for Distributed Systems and Networks",
  ["proc ieee micro electro mech syst"] = "Proceedings, IEEE micro electro mechanical systems. IEEE Workshop on Micro Electro Mechanical Systems",
  ["proc ieee radio wirel symp"] = "Proceedings. IEEE Radio and Wireless Symposium",
  ["proc ieee sens"] = "Proceedings of IEEE Sensors. IEEE International Conference on Sensors",
  ["proc ieee southeastcon"] = "Proceedings of IEEE Southeastcon. IEEE Southeastcon",
  ["proc ieee southwest symp image anal interpret"] = "Proceedings. IEEE Southwest Symposium on Image Analysis and Interpretation",
  ["proc ieee symp 3d user interfaces"] = "Proceedings. IEEE Symposium on 3D User Interfaces",
  ["proc ieee symp biol data vis"] = "Proceedings. IEEE Symposium on Biological Data Visualization",
  ["proc ieee symp comput based med syst"] = "Proceedings. IEEE Symposium on Computer-Based Medical Systems",
  ["proc ieee symp comput commun"] = "Proceedings. IEEE Symposium on Computers and Communications",
  ["proc ieee symp large scale data anal vis"] = "Proceedings. IEEE Symposium on Large-Scale Data Analysis and Visualization",
  ["proc ieee symp secur priv"] = "Proceedings. IEEE Symposium on Security and Privacy",
  ["proc ieee symp secur priv workshops"] = "Proceedings. IEEE Symposium on Security and Privacy. Workshops",
  ["proc ieee symp vis anal sci technol"] = "Proceedings. IEEE Symposium on Visual Analytics Science and Technology",
  ["proc ieee ultrason symp"] = "Proceedings. IEEE Ultrasonics Symposium",
  ["proc ieee virtual real conf"] = "Proceedings. IEEE Virtual Reality Conference",
  ["proc ieee west n y image signal process workshop"] = "Proceedings. IEEE Western New York Image and Signal Processing Workshop",
  ["proc ieee wic acm int conf web intell intell agent technol"] = "Proceedings. IEEE/WIC/ACM International Conference on Web Intelligence and Intelligent Agent Technology",
  ["proc ieee workshop appl comput vis"] = "Proceedings. IEEE Workshop on Applications of Computer Vision",
  ["proc ieee workshop autom speech recognit underst"] = "Proceedings. IEEE Workshop on Automatic Speech Recognition and Understanding",
  ["proc ieee workshop mob comput syst appl"] = "Proceedings. IEEE Workshop on Mobile Computing Systems and Applications",
  ["proc ifac world congress"] = "Proceedings of the IFAC World Congress. International Federation of Automatic Control. World Congress",
  ["proc ifsa world congr"] = "Proceedings. IFSA World Congress",
  ["proc indian acad sci math sci"] = "Proceedings of The Indian Academy Of Sciences Mathematical Sciences",
  ["proc indian nat sci acad"] = "Proceedings of the Indian National Science Academy",
  ["proc indian natl sci acad"] = "Proceedings of the Indian National Science Academy",
  ["proc indian natl sci acad b biol sci"] = "Proceedings of the Indian National Science Academy. Part B, Biological sciences",
  ["proc indiana acad sci"] = "Proceedings of the Indiana Academy of Science",
  ["proc inf theory workshop"] = "Proceedings. Information Theory Workshop",
  ["proc inst acoust"] = "Proceedings of the Institute of Acoustics. Institute of Acoustics (Great Britain)",
  ["proc inst civ eng"] = "Proceedings of the Institution of Civil Engineers",
  ["proc inst civ eng bridge eng"] = "Proceedings of the Institution of Civil Engineers: Bridge Engineering",
  ["proc inst civ eng civ eng"] = "Proceedings of the Institution of Civil Engineers: Civil Engineering",
  ["proc inst civ eng constr mater"] = "Proceedings of the Institution of Civil Engineers: Construction Materials",
  ["proc inst civ eng energy"] = "Proceedings of the Institution of Civil Engineers: Energy",
  ["proc inst civ eng eng comput mech"] = "Proceedings of the Institution of Civil Engineers: Engineering and Computational Mechanics",
  ["proc inst civ eng eng hist heritage"] = "Proceedings of the Institution of Civil Engineers: Engineering History and Heritage",
  ["proc inst civ eng eng sustainability"] = "Proceedings of the Institution of Civil Engineers: Engineering Sustainability",
  ["proc inst civ eng forensic eng"] = "Proceedings of the Institution of Civil Engineers: Forensic Engineering",
  ["proc inst civ eng geotech eng"] = "Proceedings of the Institution of Civil Engineers: Geotechnical Engineering",
  ["proc inst civ eng ground improv"] = "Proceedings of the Institution of Civil Engineers: Ground Improvement",
  ["proc inst civ eng manage procure law"] = "Proceedings of the Institution of Civil Engineers: Management, Procurement and Law",
  ["proc inst civ eng marit eng"] = "Proceedings of the Institution of Civil Engineers: Maritime Engineering",
  ["proc inst civ eng munic eng"] = "Proceedings of the Institution of Civil Engineers: Municipal Engineer",
  ["proc inst civ eng smart infrastruct constr"] = "Proceedings of the Institution of Civil Engineers: Smart Infrastructure and Construction",
  ["proc inst civ eng struct build"] = "Proceedings of the Institution of Civil Engineers: Structures and Buildings",
  ["proc inst civ eng transp"] = "Proceedings of the Institution of Civil Engineers: Transport",
  ["proc inst civ eng urban des plan"] = "Proceedings of the Institution of Civil Engineers. Urban design and planning",
  ["proc inst civ eng urban plann des"] = "Proceedings of the Institution of Civil Engineers: Urban Planning and Design",
  ["proc inst civ eng waste resour manag"] = "Proceedings of the Institution of Civil Engineers: Waste Resource Management",
  ["proc inst civ eng water manage"] = "Proceedings of the Institution of Civil Engineers: Water Management",
  ["proc inst civ eng water marit energy"] = "Proceedings of the Institution of Civil Engineers: Water, Maritime and Energy",
  ["proc inst civ eng water marit eng"] = "Proceedings of the Institution of Civil Engineers: Water and Maritime Engineering",
  ["proc inst electr eng"] = "Proceedings of the Institution of Electrical Engineers",
  ["proc inst electron radio eng"] = "Proceedings of the Institution of Electronic and Radio Engineers",
  ["proc inst math mech natl acad sci azerb"] = "Proceedings of Institute of Mathematics and Mechanics. National Academy of Sciences of Azerbaijan",
  ["proc inst math natl acad sci ukr math appl"] = "Proceedings of the Institute of Mathematics of the National Academy of Sciences of Ukraine. Mathematics and its Applications",
  ["proc inst mech eng [h]"] = "Proceedings of the Institution of Mechanical Engineers. Part H, Journal of Engineering in Medicine",
  ["proc inst mech eng b j eng manuf"] = "Proceedings of the Institution of Mechanical Engineers. Part B, Journal of engineering manufacture",
  ["proc inst mech eng c"] = "Proceedings of the Institution of Mechanical Engineers, Part C: Journal of Mechanical Engineering Science",
  ["proc inst mech eng c j mech eng sci"] = "Proceedings of the Institution of Mechanical Engineers. Part C. Journal of mechanical engineering science",
  ["proc inst mech eng f j rail rapid transit"] = "Proceedings of the Institution of Mechanical Engineers. Part F, Journal of rail and rapid transit",
  ["proc inst mech eng g j aerosp eng"] = "Proceedings of the Institution of Mechanical Engineers. Part G, Journal of aerospace engineering",
  ["proc inst mech eng h"] = "Proceedings of the Institution of Mechanical Engineers. Part H, Journal of engineering in medicine",
  ["proc inst mech eng o j risk reliab"] = "Proceedings of the Institution of Mechanical Engineers. Part O, Journal of risk and reliability",
  ["proc inst mech eng p j sport eng technol"] = "Proceedings of the Institution of Mechanical Engineers. Part P, Journal of sports engineering and technology",
  ["proc inst mech eng part a j power energy"] = "Proceedings of the Institution of Mechanical Engineers, Part A: Journal of Power and Energy",
  ["proc inst mech eng part b j eng manuf"] = "Proceedings of the Institution of Mechanical Engineers, Part B: Journal of Engineering Manufacture",
  ["proc inst mech eng part c j mech eng sci"] = "Proceedings of the Institution of Mechanical Engineers, Part C: Journal of Mechanical Engineering Science",
  ["proc inst mech eng part d j automob eng"] = "Proceedings of the Institution of Mechanical Engineers, Part D: Journal of Automobile Engineering",
  ["proc inst mech eng part e j process mech eng"] = "Proceedings of the Institution of Mechanical Engineers, Part E: Journal of Process Mechanical Engineering",
  ["proc inst mech eng part f j rail rapid transit"] = "Proceedings of the Institution of Mechanical Engineers, Part F: Journal of Rail and Rapid Transit",
  ["proc inst mech eng part g j aerosp eng"] = "Proceedings of the Institution of Mechanical Engineers, Part G: Journal of Aerospace Engineering",
  ["proc inst mech eng part h j eng med"] = "Proceedings of the Institution of Mechanical Engineers, Part H: Journal of Engineering in Medicine",
  ["proc inst mech eng part i j syst control eng"] = "Proceedings of the Institution of Mechanical Engineers. Part I: Journal of Systems and Control Engineering",
  ["proc inst mech eng part j j eng tribol"] = "Proceedings of the Institution of Mechanical Engineers, Part J: Journal of Engineering Tribology",
  ["proc inst mech eng part k j multi-body dyn"] = "Proceedings of the Institution of Mechanical Engineers, Part K: Journal of Multi-body Dynamics",
  ["proc inst mech eng part l j mater des appl"] = "Proceedings of the Institution of Mechanical Engineers, Part L: Journal of Materials: Design and Applications",
  ["proc inst mech eng part m j eng marit environ"] = "Proceedings of the Institution of Mechanical Engineers, Part M: Journal of Engineering for the Maritime Environment",
  ["proc inst mech eng, imeche conf"] = "Proceedings of the Institution of Mechanical Engineers, IMechE Conference",
  ["proc inst mech eng, part a: j power energy"] = "Proceedings of the Institution of Mechanical Engineers, Part A: Journal of Power and Energy",
  ["proc inst mech eng, part b: j eng manuf"] = "Proceedings of the Institution of Mechanical Engineers, Part B: Journal of Engineering Manufacture",
  ["proc inst mech eng, part c: j mech eng sci"] = "Proceedings of the Institution of Mechanical Engineers, Part C: Journal of Mechanical Engineering Science",
  ["proc inst mech eng, part d: j automob eng"] = "Proceedings of the Institution of Mechanical Engineers, Part D: Journal of Automobile Engineering",
  ["proc inst mech eng, part e: j process mech eng"] = "Proceedings of the Institution of Mechanical Engineers, Part E: Journal of Process Mechanical Engineering",
  ["proc inst mech eng, part f: j rail rapid transit"] = "Proceedings of the Institution of Mechanical Engineers, Part F: Journal of Rail and Rapid Transit",
  ["proc inst mech eng, part g: j aerosp eng"] = "Proceedings of the Institution of Mechanical Engineers, Part G: Journal of Aerospace Engineering",
  ["proc inst mech eng, part h: j eng med"] = "Proceedings of the Institution of Mechanical Engineers, Part H: Journal of Engineering in Medicine",
  ["proc inst mech eng, part i: j syst control eng"] = "Proceedings of the Institution of Mechanical Engineers, Part I: Journal of Systems and Control Engineering",
  ["proc inst mech eng, part j: j eng tribol"] = "Proceedings of the Institution of Mechanical Engineers, Part J: Journal of Engineering Tribology",
  ["proc inst mech eng, part k: j multi-body dyn"] = "Proceedings of the Institution of Mechanical Engineers, Part K: Journal of Multi-body Dynamics",
  ["proc inst mech eng, part l: j mater: des appl"] = "Proceedings of the Institution of Mechanical Engineers, Part L: Journal of Materials: Design and Applications",
  ["proc inst mech eng, part m: j eng marit environ"] = "Proceedings of the Institution of Mechanical Engineers, Part M: Journal of Engineering for the Maritime Environment",
  ["proc inst mech eng, part n: j nanomater nanoeng nanosyst"] = "Proceedings of the Institution of Mechanical Engineers, Part N: Journal of Nanomaterials, Nanoengineering and Nanosystems",
  ["proc inst mech eng, part o: j risk reliab"] = "Proceedings of the Institution of Mechanical Engineers, Part O: Journal of Risk and Reliability",
  ["proc inst mech eng, part p: j sports eng technol"] = "Proceedings of the Institution of Mechanical Engineers. Part P: Journal of Sports Engineering and Technology",
  ["proc inst med chic"] = "The Proceedings of the Institute of Medicine of Chicago",
  ["proc inst statist math"] = "Proceedings of the Institute of Statistical Mathematics",
  ["proc int acad oral pathol"] = "Proceedings Of The International Academy Of Oral Pathology",
  ["proc int assoc hydrol sci"] = "Proceedings of the International Association of Hydrological Sciences",
  ["proc int astron congr"] = "Proceedings of the International Astronomical Congress",
  ["proc int astron union"] = "Proceedings of the International Astronomical Union",
  ["proc int conf 3d vis"] = "Proceedings. International Conference on 3D Vision",
  ["proc int conf automot user interfaces interact veh appl (2014)"] = "Proceedings : AutomotiveUI 2014 : 6th International Conference on Automotive User Interfaces and Interactive Vehicular Applications : September 17-19, 2014, Seattle, Washington, USA. International Conference on Automotive User Interfaces and Interactive Vehicular Applications (6th : 2014 : Seattle, Wash.)",
  ["proc int conf availab reliab secur"] = "Proceedings. International Conference on Availability, Reliability and Security",
  ["proc int conf comput ling"] = "Proceedings of COLING. International Conference on Computational Linguistics",
  ["proc int conf data eng"] = "Proceedings. International Conference on Data Engineering",
  ["proc int conf distrib comput syst"] = "Proceedings. International Conference on Distributed Computing Systems",
  ["proc int conf health care syst eng  (2013)"] = "Proceedings of the International Conference on Health Care Systems Engineering. International Conference on Health Care Systems Engineering (2013 : Milan, Italy)",
  ["proc int conf image proc"] = "Proceedings. International Conference on Image Processing",
  ["proc int conf intell syst mol biol"] = "Proceedings. International Conference on Intelligent Systems for Molecular Biology",
  ["proc int conf lasers"] = "Proceedings of the International Conference on Lasers",
  ["proc int conf mach learn cybern"] = "Proceedings. International Conference on Machine Learning and Cybernetics",
  ["proc int conf qual softw"] = "Proceedings. International Conference on Quality Software",
  ["proc int conf softw eng"] = "Proceedings - International Conference on Software Engineering. International Conference on Software Engineering",
  ["proc int conf spok lang process"] = "Proceedings : ICSLP. International Conference on Spoken Language Processing",
  ["proc int congr ecol"] = "Proceedings of the International Congress of Ecology",
  ["proc int congr yak"] = "Proceedings. International Congress on Yak",
  ["proc int coral reef symp"] = "Proceedings. International Coral Reef Symposium",
  ["proc int fall workshop vis model vis"] = "Proceedings. International Fall Workshop Vision, Modeling, and Visualization",
  ["proc int geom cent"] = "Proceedings of the International Geometry Center",
  ["proc int ircobi conf biomech impacts"] = "Proceedings. International IRCOBI Conference on the Biomechanics of Impacts",
  ["proc int meshing roundtable"] = "Proceedings. International Meshing Roundtable",
  ["proc int power sources symp"] = "Proceedings of the International Power Sources Symposium",
  ["proc int pyrotech semin"] = "Proceedings of the International Pyrotechnics Seminar",
  ["proc int symp boreal for"] = "Proceedings of the International Symposium on Boreal Forest",
  ["proc int symp comput archit"] = "Proceedings. International Symposium on Computer Architecture",
  ["proc int symp hum factors ergon health care"] = "Proceedings of the International Symposium on Human Factors and Ergonomics in Health Care",
  ["proc int symp hum factors ergon healthc"] = "Proceedings of the International Symposium of Human Factors and Ergonomics in Healthcare. International Symposium of Human Factors and Ergonomics in Healthcare",
  ["proc int symp symb numer algorithms sci comput"] = "Proceedings. International Symposium on Symbolic and Numeric Algorithms for Scientific Computing",
  ["proc int symp wearable comput"] = "Proceedings. International Symposium on Wearable Computers",
  ["proc int workshop soc inform"] = "Proceedings. International Workshop on Social Informatics",
  ["proc integr ocean drill program"] = "Proceedings of the Integrated Ocean Drilling Program. Integrated Ocean Drilling Program",
  ["proc internat centre heat mass transfer"] = "Proceedings of the International Centre for Heat and Mass Transfer",
  ["proc intersoc energy convers eng conf"] = "Proceedings of the Intersociety Energy Conversion Engineering Conference",
  ["proc ipdps (conf)"] = "Proceedings. IPDPS (Conference)",
  ["proc ire"] = "Proceedings of the IRE",
  ["proc jangjeon math soc"] = "Proceedings of the Jangjeon Mathematical Society. Memoirs of the Jangjeon Mathematical Society",
  ["proc japan acad ser a math sci"] = "Japan Academy. Proceedings. Series A. Mathematical Sciences",
  ["proc jpn acad"] = "Proceedings of the Japan Academy",
  ["proc jpn acad ser a, math sci"] = "Proceedings of the Japan Academy, Series A, Mathematical Sciences",
  ["proc jpn acad ser b phys biol sci"] = "Proceedings of the Japan Academy. Series B, Physical and biological sciences",
  ["proc k ned akad wet"] = "Proceedings of the Koninklijke Nederlandse Akademie van Wetenschappen",
  ["proc k ned akad wet c"] = "Proceedings of the Koninklijke Nederlandse Akademie van Wetenschappen. Series C. Biological and medical sciences",
  ["proc konink nederl akad wetensch"] = "Koninklijke Nederlandse Akademie van Wetenschappen",
  ["proc koninklijke nederlandse akademie wetenschappen"] = "Proceedings of the Koninklijke Nederlandse Akademie van Wetenschappen. Biological, chemical, geological, physical, and medical sciences",
  ["proc latv acad sci b nat exact appl sci"] = "Proceedings of the Latvian Academy of Sciences. Section B, Natural, exact, and applied sciences = Latvijas Zinātnu akadēmijas vēstis. B dala, Dabaszinātnes",
  ["proc latv acad sci sect b"] = "Proceedings of the Latvian Academy of Sciences, Section B: Natural, Exact and Applied Sciences",
  ["proc lebedev phys inst"] = "Proceedings of the P. N. Lebedev Physics Institute",
  ["proc linn soc london"] = "Proceedings of the Linnean Society of London",
  ["proc linn soc n s w"] = "Proceedings of the Linnean Society of New South Wales. Linnean Society of New South Wales",
  ["proc lond math soc (3)"] = "Proceedings of the London Mathematical Society. Third Series",
  ["proc london math soc"] = "Proceedings of the London Mathematical Society",
  ["proc london math soc (3)"] = "Proceedings of the London Mathematical Society",
  ["proc lunar planet sci"] = "Proceedings of lunar and planetary science",
  ["proc mach learn res"] = "Proceedings of machine learning research",
  ["proc mach learn res (pmlr)"] = "Proceedings of Machine Learning Research (PMLR)",
  ["proc math phys eng sci"] = "Proceedings. Mathematical, physical, and engineering sciences",
  ["proc math phys soc egypt"] = "Proceedings of the Mathematical and Physical Society of Egypt",
  ["proc math sci"] = "Proceedings - Mathematical Sciences",
  ["proc math soc"] = "Proceedings of the Mathematical Society",
  ["proc meas behav 2016 (2016)"] = "Proceedings of Measuring Behavior 2016 : 10th International Conference on Methods and Techniques in Behavioral Research. International Conference on Methods and Techniques in Behavioral Research (10th : 2016 : Dublin, Ireland)",
  ["proc meet acoust"] = "Proceedings of meetings on acoustics. Acoustical Society of America",
  ["proc mine med off assoc"] = "Proceedings of the Mine Medical Officers' Association",
  ["proc mine med off assoc sa"] = "Proceedings of the Mine Medical Officers' Association of S. A",
  ["proc n s inst sci"] = "Proceedings of the Nova Scotian Institute of Science",
  ["proc n y state assoc public health lab"] = "Proceedings of the New York State Association of Public Health Laboratories",
  ["proc nat acad sci india"] = "Proceedings of the National Academy of Sciences, India",
  ["proc nat acad sci india sect a"] = "Proceedings of the National Academy of Sciences, India. Section A. Physical Sciences",
  ["proc nat acad sci usa"] = "Proceedings of the National academy of Sciences of the United States of America",
  ["proc natl acad sci"] = "Proceedings of the National Academy of Sciences",
  ["proc natl acad sci india sect b biol sci"] = "Proceedings of the National Academy of Sciences, India. Section B",
  ["proc natl acad sci u s a"] = "Proceedings of the National Academy of Sciences of the United States of America",
  ["proc natl acad sci usa"] = "Proceedings of the National Academy of Sciences of the United States of America",
  ["proc natl acad sci usa, early ed"] = "Proceedings of the National Academy of Sciences of the United States of America, Early Edition",
  ["proc natl acad sci, india, sect a"] = "Proceedings of the National Academy of Sciences, India, Section A: Physical Sciences",
  ["proc natl acad sci, india, sect a phys sci"] = "Proceedings of the National Academy of Sciences, India Section A: Physical Sciences",
  ["proc natl acad sci, india, sect b biol sci"] = "Proceedings of the National Academy of Sciences, India Section B: Biological Sciences",
  ["proc natl cancer conf"] = "Proceedings. National Cancer Conference",
  ["proc natl conf methadone treat"] = "Proceedings. National Conference on Methadone Treatment",
  ["proc natl inst ecol repub korea"] = "Proceedings of the National Institute of Ecology of the Republic of Korea",
  ["proc natl sci counc repub china b"] = "Proceedings of the National Science Council, Republic of China. Part B, Life sciences",
  ["proc nec res sympos"] = "Proceedings of the NEC Research Symposia",
  ["proc neuropsychiatr conf sixth serv command"] = "Proceedings of the Neuropsychiatric Conference of the Sixth Service Command : John B. Murphy Memorial Auditorium, American College of Surgeons, 50 East Erie Street, Chicago, Illinois, 16-17 November 1945. Neuropsychiatric Conference of the Sixth Service Command (1945 : Chicago, Ill.)",
  ["proc nipr symp antarct meteorites"] = "Proceedings of the NIPR Symposium on Antarctic Meteorites. NIPR Symposium on Antarctic Meteorites",
  ["proc nurs theory conf"] = "Proceedings. Nursing Theory Conference (University of Kansas)",
  ["proc nutr soc"] = "The Proceedings of the Nutrition Society",
  ["proc nutr soc india"] = "Proceedings of the Nutrition Society of India",
  ["proc ocean drill prog sci results"] = "Proceedings of the Ocean Drilling Program. Scientific results. Ocean Drilling Program",
  ["proc ocean drill program part b sci results"] = "Proceedings of the Ocean Drilling Program, Part B: Scientific Results",
  ["proc ohio collab conf bioinform"] = "Proceedings. Ohio Collaborative Conference on Bioinformatics",
  ["proc okla acad sci"] = "Proceedings of the Oklahoma Academy of Science",
  ["proc one hta health technol assess int meet 1st 2004 krakow pol"] = "Proceedings one HTA : yesterday, today & tomorrow, HTAi 2004, May 30-June 2, 2004, Krakow, Poland. Health Technology Assessment International. Meeting (1st : 2004 : Krakow, Poland)",
  ["proc pac conf comput graph appl"] = "Proceedings. Pacific Conference on Computer Graphics and Applications",
  ["proc pak acad sci a"] = "Proceedings of the Pakistan Academy of Sciences. A. Physical and Computational Sciences",
  ["proc pakistan acad sci"] = "Proceedings of the Pakistan Academy of Sciences",
  ["proc pap annu conf calif mosq control assoc"] = "Proceedings and papers of the annual conference of the California Mosquito Control Association, inc",
  ["proc pap ga assoc hist"] = "Journal of the Georgia Association of Historians",
  ["proc pharm soc egypt"] = "The Proceedings of the Pharmaceutical Society of Egypt",
  ["proc phys soc"] = "Proceedings of the Physical Society",
  ["proc phys soc b"] = "Proceedings of the Physical Society B",
  ["proc phys soc london"] = "Proceedings of the Physical Society of London",
  ["proc phys soc london, sec a"] = "Proceedings of the Physical Society, London, Section A",
  ["proc phys soc london, sec b"] = "Proceedings of the Physical Society, London, Section B",
  ["proc phys soc london, sect a"] = "Proceedings of the Physical Society. Section A",
  ["proc phys soc london, sect b"] = "Proceedings of the Physical Society. Section B",
  ["proc phys soc, sect a"] = "Proceedings of the Physical Society, Section A",
  ["proc phys soc, sect b"] = "Proceedings of the Physical Society, Section B",
  ["proc physiol soc"] = "Proceedings of the Physiological Society",
  ["proc prehist soc"] = "Proceedings of the Prehistoric Society",
  ["proc prof staff"] = "Proceedings of the professional staff, Brooke General Hospital. Brooke General Hospital",
  ["proc pure appl math"] = "Proceedings in Pure and Applied Mathematics",
  ["proc r australas coll physicians"] = "Proceedings. Royal Australasian College of Physicians",
  ["proc r coll physicians edinb"] = "Proceedings of the Royal College of Physicians of Edinburgh",
  ["proc r entomol soc london"] = "Proceedings of the Royal Entomological Society of London",
  ["proc r entomol soc london ser b taxon"] = "Proceedings of the Royal Entomological Society of London Series B Taxonomy",
  ["proc r inst g b"] = "Proceedings of the Royal Institution of Great Britain",
  ["proc r ir acad [b]"] = "Proceedings of the Royal Irish Academy. Section B: Biological, Geological and Chemical Science",
  ["proc r ir acad b"] = "Proceedings of the Royal Irish Academy. Section B: Biological, geological, and chemical science",
  ["proc r ir acad c archaeol celt stud hist linguist lit"] = "Proceedings of the Royal Irish Academy. Section C. Archaeology, Celtic studies, history, linguistics and literature. Royal Irish Academy",
  ["proc r soc a"] = "Proceedings of the Royal Society A",
  ["proc r soc b"] = "Proceedings of the Royal Society B",
  ["proc r soc edinb [biol]"] = "Proceedings, Royal Society of Edinburgh. Section B: Biology",
  ["proc r soc edinb [nat environ]"] = "Proceedings, Royal Society of Edinburgh. Section B: Natural Environment",
  ["proc r soc edinb biol"] = "Proceedings of the Royal Society of Edinburgh. Section B: Biology",
  ["proc r soc edinb nat environ"] = "Proceedings - Royal Society of Edinburgh. Section B: Natural environment",
  ["proc r soc lond a"] = "Proceedings of the Royal Society of London A",
  ["proc r soc lond a math phys sci"] = "Proceedings of the Royal Society of London. Series A, Mathematical and physical sciences",
  ["proc r soc lond b biol sci"] = "Proceedings of the Royal Society of London. Series B, Biological sciences",
  ["proc r soc lond ser math phys sci"] = "Proceedings of the Royal Society of London A",
  ["proc r soc london"] = "Proceedings of the Royal Society of London",
  ["proc r soc london a"] = "Proceedings of the Royal Society of London. Series A, Containing Papers of a Mathematical and Physical Character",
  ["proc r soc london a - math phys sci"] = "Proceedings of the Royal Society of London. Series A - Mathematical and Physical Sciences",
  ["proc r soc london b"] = "Proceedings of the Royal Society of London. Series B",
  ["proc r soc london, ser a"] = "Proceedings of the Royal Society of London Series A",
  ["proc r soc london, ser a: math phys sci"] = "Proceedings of the Royal Society of London, Series A: Mathematical and Physical Sciences",
  ["proc r soc london, ser b"] = "Proceedings of the Royal Society of London Series B",
  ["proc r soc math phys eng sci"] = "Proceedings of the Royal Society of London A",
  ["proc r soc med"] = "Proceedings of the Royal Society of Medicine",
  ["proc r soc ser b biol sci"] = "Proceedings of the Royal Society Series B Biological Sciences",
  ["proc r soc victoria"] = "Proceedings of the Royal Society of Victoria",
  ["proc rmrs"] = "Proceedings RMRS",
  ["proc rom acad sci ser a math phys tech sci inf sci"] = "Proceedings Of The Romanian Academy Series A Mathematics Physics Technical Sciences Information Science",
  ["proc rom acad ser a math phys tech sci inf sci"] = "Proceedings of the Romanian Academy. Series A. Mathematics, Physics, Technical Sciences, Information Science",
  ["proc roy irish acad"] = "Proceedings of the Royal Irish Academy",
  ["proc roy soc edinburgh"] = "Proceedings of the Royal Society of Edinburgh",
  ["proc roy soc edinburgh sect a"] = "Proceedings of the Royal Society of Edinburgh. Section A. Mathematics",
  ["proc roy soc london"] = "Proceedings of the Royal Society",
  ["proc rudolf virchow med soc city n y"] = "Proceedings of the Rudolf Virchow Medical Society in the City of New York",
  ["proc s c hist assoc"] = "The proceedings of the South Carolina Historical Association. South Carolina Historical Association",
  ["proc s dak acad sci"] = "Proceedings of the South Dakota Academy of Science",
  ["proc sch sci tokai univ"] = "Proceedings of the School of Science of Tokai University",
  ["proc school sci tokai univ"] = "Proceedings of the School of Science of Tokai University",
  ["proc second workshop data manag end end mach learn (2018)"] = "Proceedings of the Second Workshop on Data Management for End-to-End Machine Learning. Workshop on Data Management for End-to-End Machine Learning (2nd : 2018 : Houston, Tex.)",
  ["proc sigchi conf hum factor comput syst"] = "Proceedings of the SIGCHI conference on human factors in computing systems. CHI Conference",
  ["proc singapore natl acad sci"] = "Proceedings of the Singapore National Academy of Science",
  ["proc site"] = "Proceedings of SITE. Society for Information Technology and Teacher Education. International Conference",
  ["proc soc agric bacteriol"] = "Proceedings of the Society of Agricultural Bacteriologists",
  ["proc soc anal chem"] = "Proceedings of the Society for Analytical Chemistry",
  ["proc soc appl bacteriol"] = "Proceedings of the Society for Applied Bacteriology",
  ["proc soc exp biol med"] = "Proceedings of the Society for Experimental Biology and Medicine",
  ["proc soc photo opt instrum eng"] = "Proceedings - Society of Photo-Optical Instrumentation Engineers",
  ["proc soc stat sect"] = "Proceedings of the Social Statistics Section. American Statistical Association. Social Statistics Section",
  ["proc soc study fertil"] = "Proceedings. Society for the Study of Fertility",
  ["proc soil sci soc am"] = "Proceedings Soil Science Society of America",
  ["proc spi annu tech/mark conf"] = "Proceedings of the SPI Annual Technical/Marketing Conference",
  ["proc spie"] = "Proceedings of SPIE",
  ["proc spie int soc opt eng"] = "Proceedings of SPIE The International Society for Optical Engineering",
  ["proc spie-int soc opt eng"] = "Proceedings of SPIE - The International Society for Optical Engineering",
  ["proc staff meet honol clin"] = "Proceedings of the staff meetings. Honolulu. Clinic",
  ["proc staff meet mayo clin"] = "Proceedings of the staff meetings. Mayo Clinic",
  ["proc staff meet pethah tiqva isr beilinson hosp"] = "Proceedings of the staff meetings. Pethah-Tiqva, Israel. Beilinson Hospital",
  ["proc staff meet tulsa okla hillcrest meml hosp"] = "Proceedings of staff meeting. Tulsa, Okla. Hillcrest Memorial Hospital",
  ["proc state secr manage conf"] = "Proceedings. State Secretaries Management Conference",
  ["proc state secr manage conf am dent assoc"] = "Proceedings, State Secretaries Management Conference, American Dental Association",
  ["proc steklov inst math"] = "Proceedings of the Steklov Institute of Mathematics",
  ["proc stekov inst math"] = "Proceedings of the Stekov Institute of Mathematics",
  ["proc stem cell res oncog"] = "Proceedings of stem cell research and oncogenesis",
  ["proc symp comput archit high perform comput"] = "Proceedings. Symposium on Computer Architecture and High Performance Computing",
  ["proc symp haptic interface virtual env teleoperator syst"] = "Proceedings. Symposium on Haptic Interfaces for Virtual Environment and Teleoperator Systems",
  ["proc sympos appl math"] = "Proceedings of Symposia in Applied Mathematics",
  ["proc sympos pure math"] = "Proceedings of Symposia in Pure Mathematics",
  ["proc trans n s inst sci"] = "Proceedings and Transactions of the Nova Scotian Institute of Science",
  ["proc transvaal mine med off assoc"] = "Proceedings. Transvaal Mine Medical Officers' Association",
  ["proc u s nav inst"] = "Proceedings - United States Naval Institute. United States Naval Institute",
  ["proc univ otago med sch"] = "Proceedings of the University of Otago Medical School",
  ["proc us nav inst"] = "Proceedings - United States Naval Institute",
  ["proc usenix symp oper syst des implement (osdi)"] = "Proceedings of the -- USENIX Symposium on Operating Systems Design and Implementation (OSDI). USENIX Symposium on Operating Systems Design and Implementation",
  ["proc veterans adm spinal cord inj conf"] = "Proceedings. Veterans Administration Spinal Cord Injury Conference",
  ["proc virchow pirquet med soc"] = "Proceedings of the Virchow-Pirquet Medical Society",
  ["proc vldb endow"] = "Proceedings of the VLDB Endowment",
  ["proc w va acad sci"] = "Proceedings of the West Virginia Academy of Science",
  ["proc wesley hist soc"] = "Proceedings of the Wesley Historical Society. Wesley Historical Society",
  ["proc west mark ed assoc conf"] = "Proceedings",
  ["proc west pharmacol soc"] = "Proceedings of the Western Pharmacology Society",
  ["proc wild trout wild trout symp"] = "Proceedings of Wild Trout. Wild Trout Symposium (Yellowstone National Park)",
  ["proc wkly semin neurol"] = "Proceedings of the Weekly Seminar in Neurology",
  ["proc workshop math methods biomed image analysis"] = "Proceedings. Workshop on Mathematical Methods in Biomedical Image Analysis",
  ["proc workshop vitam d"] = "Proceedings of the Workshop on Vitamin D",
  ["proc yorkshire geol soc"] = "Proceedings of the Yorkshire Geological Society",
  ["proc zool soc"] = "Proceedings of the Zoological Society",
  ["proc: plant sci"] = "Proceedings: Plant Science",
  ["procedia chem"] = "Procedia chemistry",
  ["procedia cirp"] = "Procedia CIRP",
  ["procedia comput sci"] = "Procedia Computer Science",
  ["procedia earth planet sci"] = "Procedia Earth and Planetary Science",
  ["procedia eng"] = "Procedia engineering",
  ["procedia environ sci"] = "Procedia Environmental Sciences",
  ["procedia food sci"] = "Procedia Food Science",
  ["procedia iutam"] = "Procedia IUTAM",
  ["procedia manuf"] = "Procedia manufacturing",
  ["procedia mater sci"] = "Procedia Materials Science",
  ["procedia soc behav sci"] = "Procedia, social and behavioral sciences",
  ["procedia struct integrity"] = "Procedia Structural Integrity",
  ["procedia technol"] = "Procedia Technology",
  ["procedia vaccinol"] = "Procedia in vaccinology",
  ["proceed canad soc hist philos math"] = "Proceedings of the Canadian Society for History and Philosophy of Mathematics/La Societe Canadienne d’Histoire et de Philosophie des Mathematiques",
  ["proceed rijeka sch econ: j econ bus"] = "Proceedings of Rijeka School of Economics: Journal of Economics and Business",
  ["proceedings (ieee int conf bioinformatics biomed)"] = "Proceedings. IEEE International Conference on Bioinformatics and Biomedicine",
  ["proceedings (mdpi)"] = "Proceedings (MDPI)",
  ["proceedings of the japanese association of mycotoxicology|maikotokishin"] = "Maikotokishin : Maikotokishin Kenkyukai kaiho",
  ["proceedings vldb endowment"] = "Proceedings of the VLDB Endowment. International Conference on Very Large Data Bases",
  ["process appl ceram"] = "Processing and Application of Ceramics",
  ["process biochem"] = "Process biochemistry (Barking, London, England)",
  ["process biochem (amsterdam, neth)"] = "Process Biochemistry (Amsterdam, Netherlands)",
  ["process biochem (oxford, u k)"] = "Process Biochemistry (Oxford, United Kingdom)",
  ["process control qual"] = "Process Control and Quality",
  ["process hist"] = "Procesos históricos",
  ["process integr optim sustainability"] = "Process Integration and Optimization for Sustainability",
  ["process saf environ prot"] = "Process Safety and Environment Protection",
  ["process saf prog"] = "Process Safety Progress",
  ["processes (basel)"] = "Processes (Basel, Switzerland)",
  ["proclins cardiol"] = "ProClinS cardiology",
  ["prod eng"] = "Production Engineering",
  ["prod leiterplatten sys"] = "Produktion von Leiterplatten und Systemen",
  ["prod manuf res"] = "Production & manufacturing research",
  ["prod oper manag"] = "Production and operations management",
  ["prod oper manage"] = "Production and Operations Management",
  ["prod pharm"] = "Produits pharmaceutiques",
  ["prod probl pharm"] = "Produits & problêmes pharmaceutiques",
  ["prof anim sci"] = "The professional animal scientist",
  ["prof care mother child"] = "Professional care of mother and child",
  ["prof case manag"] = "Professional case management",
  ["prof dev (phila)"] = "Professional development (Philadelphia, Pa.)",
  ["prof dev ser (chic ill)"] = "Professional development series (Chicago, Ill.)",
  ["prof ethics"] = "Professional ethics (Gainesville, Fla.)",
  ["prof ethics rep"] = "Professional ethics report : newsletter of the American Association for the Advancement of Science, Committee on Scientific Freedom & Responsibility, Professional Society Ethics Group",
  ["prof flashes"] = "Professional flashes",
  ["prof geogr"] = "The Professional geographer : the journal of the Association of American Geographers",
  ["prof indust comput ser"] = "Professional and Industrial Computing Series",
  ["prof inferm"] = "Professioni infermieristiche",
  ["prof nurs home"] = "Professional nursing home",
  ["prof nurse"] = "Professional nurse (London, England)",
  ["prof psychol"] = "Professional psychology",
  ["prof psychol res pr"] = "Professional psychology, research and practice",
  ["prof saf"] = "Professional safety",
  ["prof sanit manage"] = "Professional sanitation management",
  ["profil klin med"] = "Profilakticheskai︠a︡ i klinicheskai︠a︡ medit︠s︡ina",
  ["profile med pract"] = "Profile of medical practice",
  ["profiles drug subst excip relat methodol"] = "Profiles of drug substances, excipients, and related methodology",
  ["profiles healthc commun"] = "Profiles in healthcare communications",
  ["profiles healthc mark"] = "Profiles in healthcare marketing",
  ["profiles hosp mark"] = "Profiles in hospital marketing",
  ["prog addit manuf"] = "Progress in Additive Manufacturing",
  ["prog adv comput intell eng"] = "Progress in advanced computing and intelligent engineering : proceedings of ICACIE 2016. International Conference on Advanced Computing and Intelligent Engineering (2016 : Bhubaneswar, India)",
  ["prog aerosp sci"] = "Progress in Aerospace Sciences",
  ["prog agric eng sci"] = "Progess in Agricultural Engineering Sciences",
  ["prog aids pathol"] = "Progress in AIDS pathology",
  ["prog allergy"] = "Progress in allergy",
  ["prog arch"] = "Progressive architecture",
  ["prog archit"] = "Progressive Architecture",
  ["prog artif intell"] = "Progress in Artificial Intelligence",
  ["prog at med"] = "Progress in atomic medicine",
  ["prog behav modif"] = "Progress in behavior modification",
  ["prog biochem pharmacol"] = "Progress in biochemical pharmacology",
  ["prog biocybern"] = "Progress in biocybernetics",
  ["prog biol sci"] = "Progress in biological sciences",
  ["prog biomater"] = "Progress in biomaterials",
  ["prog biomed eng"] = "Progress in Biomedical Engineering",
  ["prog biomed eng (bristol)"] = "Progress in biomedical engineering (Bristol, England)",
  ["prog biometeorol"] = "Progress in biometeorology. Division A: Progress in human biometeorology",
  ["prog biophys biophys chem"] = "Progress in biophysics and biophysical chemistry",
  ["prog biophys mol biol"] = "Progress in biophysics and molecular biology",
  ["prog biotechnol"] = "Progress in Biotechnology",
  ["prog bot"] = "Progress in botany. Fortschritte der Botanik",
  ["prog brain res"] = "Progress in brain research",
  ["prog cardiovasc dis"] = "Progress in cardiovascular diseases",
  ["prog cardiovasc nurs"] = "Progress in cardiovascular nursing",
  ["prog cell cycle res"] = "Progress in cell cycle research",
  ["prog chem appl chitin deriv"] = "Progress on chemistry and application of chitin and its derivatives",
  ["prog chem fats other lipds"] = "Progress in the Chemistry of Fats and Other Lipids",
  ["prog chem fats other lipids"] = "Progress in the chemistry of fats and other lipids",
  ["prog chem org nat prod"] = "Progress in the chemistry of organic natural products",
  ["prog chem toxicol"] = "Progress in chemical toxicology",
  ["prog clin biol res"] = "Progress in clinical and biological research",
  ["prog clin cancer"] = "Progress in clinical cancer",
  ["prog clin immunol"] = "Progress in clinical immunology",
  ["prog clin parasitol"] = "Progress in clinical parasitology",
  ["prog clin pathol"] = "Progress in clinical pathology",
  ["prog colloid polym sci"] = "Progress in Colloid and Polymer Science",
  ["prog community health partnersh"] = "Progress in community health partnerships : research, education, and action",
  ["prog comput fluid dyn"] = "Progress in Computational Fluid Dynamics",
  ["prog cryst growth charact"] = "Progress in Crystal Growth and Characterization",
  ["prog cryst growth charact mater"] = "Progress in Crystal Growth and Characterization of Materials",
  ["prog disaster sci"] = "Progress in Disaster Science",
  ["prog drug res"] = "Progress in drug research. Fortschritte der Arzneimittelforschung. Progrès des recherches pharmaceutiques",
  ["prog earth planet sci"] = "Progress in Earth and Planetary Science",
  ["prog electromagn res b pier b"] = "Progress in electromagnetics research B. Pier B",
  ["prog electromagn res c pier c"] = "Progress in electromagnetics research C. Pier C",
  ["prog electromagn res m pier m"] = "Progress in electromagnetics research M. Pier M",
  ["prog electromagn res symp"] = "Progress in Electromagnetics Research Symposium : [proceedings]. Progress in Electromagnetics Research Symposium",
  ["prog energy"] = "Progress in Energy",
  ["prog energy combust sci"] = "Progress in Energy and Combustion Science",
  ["prog environ geogr"] = "Progress in Environmental Geography",
  ["prog exp pers psychopathol res"] = "Progress in experimental personality & psychopathology research",
  ["prog exp pers res"] = "Progress in experimental personality research",
  ["prog exp tumor res"] = "Progress in experimental tumor research",
  ["prog food nutr sci"] = "Progress in food & nutrition science",
  ["prog growth factor res"] = "Progress in growth factor research",
  ["prog gynecol"] = "Progress in gynecology",
  ["prog hematol"] = "Progress in hematology",
  ["prog hemost thromb"] = "Progress in hemostasis and thrombosis",
  ["prog heterocycl chem"] = "Progress in Heterocyclic Chemistry",
  ["prog histochem cytochem"] = "Progress in Histochemistry and Cytochemistry",
  ["prog hum geogr"] = "Progress in human geography",
  ["prog hum reprod res"] = "Progress in human reproduction research",
  ["prog immunobiol stand"] = "Progress in immunobiological standardization",
  ["prog ind ecol"] = "Progress in Industrial Ecology",
  ["prog ind microbiol"] = "Progress in Industrial Microbiology",
  ["prog inorg biochem biophys"] = "Progress in inorganic biochemistry and biophysics",
  ["prog inorg chem"] = "Progress in Inorganic Chemistry",
  ["prog is"] = "Progress in IS",
  ["prog lipid res"] = "Progress in lipid research",
  ["prog liver dis"] = "Progress in liver diseases",
  ["prog mater sci"] = "Progress in Materials Science",
  ["prog math phys"] = "Progress in Mathematical Physics",
  ["prog med"] = "Progressos da medicina",
  ["prog med (napoli)"] = "Il Progresso medico",
  ["prog med (paris)"] = "Le Progrés médical",
  ["prog med chem"] = "Progress in medicinal chemistry",
  ["prog med genet"] = "Progress in medical genetics",
  ["prog med rev asoc med latinoam"] = "El Progreso médico; revista de la Asociación Médica Latinoamericana",
  ["prog med virol"] = "Progress in medical virology. Fortschritte der medizinischen Virusforschung. Progrès en virologie médicale",
  ["prog met phys"] = "Progress in Metal Physics",
  ["prog mol biol transl sci"] = "Progress in molecular biology and translational science",
  ["prog mol subcell biol"] = "Progress in molecular and subcellular biology",
  ["prog nat sci"] = "Progress in Natural Science",
  ["prog nat sci mater int"] = "Progress in Natural Science: Materials International",
  ["prog nat sci: mater int"] = "Progress in Natural Science: Materials International",
  ["prog neuro-psychopharmacol"] = "Progress in Neuro-Psychopharmacology",
  ["prog neuro-psychopharmacol biol psychiatry"] = "Progress in Neuro-Psychopharmacology & Biological Psychiatry",
  ["prog neurobiol"] = "Progress in neurobiology",
  ["prog neurol psychiatry"] = "Progress in neurology and psychiatry",
  ["prog neurol surg"] = "Progress in neurological surgery",
  ["prog neuropsychopharmacol"] = "Progress in neuro-psychopharmacology",
  ["prog neuropsychopharmacol biol psychiatry"] = "Progress in neuro-psychopharmacology & biological psychiatry",
  ["prog nucl energy"] = "Progress in Nuclear Energy",
  ["prog nucl energy 6 biol sci"] = "Biological sciences",
  ["prog nucl energy 7 med sci"] = "Medical sciences (London, England)",
  ["prog nucl magn reson spectrosc"] = "Progress in nuclear magnetic resonance spectroscopy",
  ["prog nucl med"] = "Progress in nuclear medicine",
  ["prog nucleic acid res mol biol"] = "Progress in Nucleic Acid Research and Molecular Biology",
  ["prog oceanogr"] = "Progress in Oceanography",
  ["prog odontoiatr"] = "Progresso odontoiatrico",
  ["prog odontostomatol"] = "Le Progrés odonto-stomatologique",
  ["prog opt sci photonics"] = "Progress in Optical Science and Photonics",
  ["prog org coat"] = "Progress in Organic Coatings",
  ["prog orthod"] = "Progress in orthodontics",
  ["prog pain res manag"] = "Progress in pain research and management",
  ["prog palliat care"] = "Progress in palliative care",
  ["prog pap recycl"] = "Progress in Paper Recycling",
  ["prog part nucl phys"] = "Progress in Particle and Nuclear Physics",
  ["prog patol clin"] = "Progresos de patología y clínica",
  ["prog pediat study"] = "Progress pediatric study. American Academy of Pediatrics. Committee for the Study of Child Health Services",
  ["prog pediatr cardiol"] = "Progress in pediatric cardiology",
  ["prog pediatr pueric"] = "Progresos de pediatría y puericultura",
  ["prog pediatr surg"] = "Progress in pediatric surgery",
  ["prog pharmacol clin pharmacol"] = "Progress in pharmacology and clinical pharmacology",
  ["prog photovolt"] = "Progress in photovoltaics",
  ["prog photovoltaics res appl"] = "Progress in Photovoltaics: Research and Applications",
  ["prog phys geogr"] = "Progress in Physical Geography",
  ["prog phys geogr: earth environ"] = "Progress in Physical Geography: Earth and Environment",
  ["prog phys ther"] = "Progress in physical therapy",
  ["prog plann"] = "Progress in planning",
  ["prog polym sci"] = "Progress in Polymer Science",
  ["prog prev med (n y)"] = "Progress in preventive medicine (New York, N.Y.)",
  ["prog psychobiol physiol psychol"] = "Progress in psychobiology and physiological psychology",
  ["prog psychother"] = "Progress in psychotherapy",
  ["prog quantum electron"] = "Progress in Quantum Electronics",
  ["prog radiat ther"] = "Progress in radiation therapy",
  ["prog react kinet"] = "Progress in Reaction Kinetics",
  ["prog react kinet mech"] = "Progress in Reaction Kinetics and Mechanism",
  ["prog rehabil med"] = "Progress in rehabilitation medicine",
  ["prog rep cent popul res natl inst child health hum dev"] = "Progress report of the Center for Population Research, National Institute of Child Health and Human Development. Center for Population Research (National Institute of Child Health and Human Development (U.S.))",
  ["prog rep health dev south afr"] = "Progress reports on health & development in Southern Africa",
  ["prog respir res"] = "Progress in Respiratory Research",
  ["prog retin eye res"] = "Progress in retinal and eye research",
  ["prog rubber plast recycl technol"] = "Progress in Rubber, Plastics and Recycling Technology",
  ["prog solid state chem"] = "Progress in Solid State Chemistry",
  ["prog stem cell"] = "Progress in stem cell",
  ["prog stereochem"] = "Progress in stereochemistry",
  ["prog struct eng mater"] = "Progress in Structural Engineering and Materials",
  ["prog struct mater eng"] = "Progress in Structural Engineering and Materials",
  ["prog surf sci"] = "Progress in Surface Science",
  ["prog surg"] = "Progress in surgery",
  ["prog ter"] = "Il Progresso terapeutico",
  ["prog ter clin"] = "Progresos de terapéutica clínica",
  ["prog theor exp phys"] = "Progress of Theoretical and Experimental Physics",
  ["prog theor phys"] = "Progress of Theoretical Physics",
  ["prog theor phys suppl"] = "Progress of Theoretical Physics Supplements",
  ["prog transplant"] = "Progress in transplantation (Aliso Viejo, Calif.)",
  ["prog tumor res"] = "Progress in tumor research",
  ["prog urol"] = "Progrès en urologie : journal de l'Association française d'urologie et de la Société française d'urologie",
  ["prog vet microbiol immunol"] = "Progress in veterinary microbiology and immunology",
  ["progr comput sci appl logic"] = "Progress in Computer Science and Applied Logic",
  ["progr creep fracture"] = "Progress in Creep and Fracture",
  ["progr math"] = "Progress in Mathematics",
  ["progr math (varanasi)"] = "Progress of Mathematics",
  ["progr med"] = "Progress in medicine",
  ["progr med (napoli)"] = "Progresso Medico",
  ["progr med (paris)"] = "Progres Medical",
  ["progr natur sci (english ed)"] = "Progress in Natural Science",
  ["progr nonlinear differential equations appl"] = "Progress in Nonlinear Differential Equations and their Applications",
  ["progr pediatr puericult"] = "Progresos de pediatria y puericultura",
  ["progr phys"] = "Progress in Physics",
  ["progr probab"] = "Progress in Probability",
  ["progr pure appl discrete math"] = "Progress in Pure and Applied Discrete Mathematics",
  ["progr systems control theory"] = "Progress in Systems and Control Theory",
  ["progr ter clin"] = "Progresos de Terapeutica Clinica",
  ["progr theoret comput sci"] = "Progress in Theoretical Computer Science",
  ["progr theoret phys"] = "Progress of Theoretical Physics",
  ["progr theoret phys suppl"] = "Progress of Theoretical Physics",
  ["program abstr 3rd conf retrovir oppor infect conf retrovir oppor infect 3rd 1996 wash d c"] = "Program and abstracts of the 3rd Conference on Retroviruses and Opportunistic Infections : Sheraton Washington Hotel, January 28-February 2 [i.e. 1], 1996, Washington, DC. Conference on Retroviruses and Opportunistic Infections (3rd : 1996 : Washington, D.C.)",
  ["program abstr 4th conf retrovir oppor infect conf retrovir oppor infect 4th 1997 wash dc"] = "Program and abstracts of the 4th Conference on Retroviruses and Opportunistic Infections : Sheraton Washington Hotel, January 22-26, 1997, Washington, DC. Conference on Retroviruses and Opportunistic Infections (4th : 1997 : Washington, D.C.)",
  ["program abstr 5th conf retrovir oppor infect conf retrovir oppor infect 5th 1998 chic ill"] = "Program and abstracts of the 5th Conference on Retroviruses and Opportunistic Infections : Sheraton Chicago Hotel and Towers, February 1-5 1998, Chicago, IL. Conference on Retroviruses and Opportunistic Infections (5th : 1998 : Chicago, Ill.)",
  ["program abstr 6th conf retrovir oppor infect conf retrovir oppor infect 6th 1999 chic ill"] = "Program and abstracts of the 6th Conference on Retroviruses and Opportunistic Infections : Sheraton Chicago Hotel & Towers, January 31 - February 4, 1999. Conference on Retroviruses and Opportunistic Infections (6th : 1999 : Chicago, Ill.)",
  ["program abstr first natl conf hum retrovir relat infect natl conf hum retrovir relat infect 1st 1993 wash dc"] = "Program and abstracts of the First National Conference on Human Retroviruses and Related Infections : Sheraton Washington Hotel, 12-16 December 1993, Washington, DC. National Conference on Human Retroviruses and Related Infections (1st : 1993 : Washington, D.C.)",
  ["program abstr hiv drp symp underst antivir drug resist 1st 2000 chantilly va"] = "Program & abstracts : first HIV DRP Symposium Understanding Antiviral Drug Resistance, December 3-6, 2000, Westfields Internaitonal Conference Center, Chantilly, Virginia. HIV DRP Symposium Understanding Antiviral Drug Resistance (1st : 2000 : Chantilly, Virginia)",
  ["program abstr hiv infect women conf hiv infect women conf 1995 wash dc"] = "Program and abstracts of the HIV Infection in Women Conference : setting a new agenda : February 22-24, 1995, Sheraton Washington Hotel, Washington, D.C. HIV Infection in Women Conference (1995 : Washington, D.C.)",
  ["program abstr second natl conf hum retrovir relat infect natl conf hum retrovir relat infect 2nd 1995 wash dc"] = "Program and abstracts of the Second National Conference on Human Retroviruses and Related Infections : Sheraton Washington Hotel, 29 January-2 February 1995, Washington, DC. National Conference on Human Retroviruses and Related Infections (2nd : 1995 : Washington, D.C.)",
  ["program comput softw"] = "Programming and Computer Software",
  ["program notes assoc univ programs health adm"] = "Program notes (Association of University Programs.)",
  ["program q asia found"] = "Program quarterly - The Asia Foundation. The Asia Foundation",
  ["programme abstr aids vaccine 2004"] = "Final programme & abstracts : AIDS vaccine 04, Lausanne, Switzerland, August 30th-September 1st, 2004. AIDS Vaccine International Conference (2004 : Lausanne, Switzerland)",
  ["progress nurse"] = "The Progressive nurse",
  ["progress rei bot"] = "Progressus Rei Botanicae",
  ["progress ter"] = "Progressi di terapia",
  ["proj inf perspect"] = "Project Inform perspective",
  ["proj math!"] = "Project MATHEMATICS!",
  ["proj rep usaf sch aviat med"] = "Project report. USAF School of Aviation Medicine",
  ["prologue j natl arch"] = "Prologue (Washington, D.C.)",
  ["promed komplement"] = "Promed komplementär",
  ["promega notes"] = "Promega notes magazine",
  ["prometeo (milan)"] = "Prometeo (Milan, Italy : 1983)",
  ["promot dent"] = "La Promotion dentaire",
  ["promot educ"] = "Promotion & education",
  ["promot health"] = "Promoting health",
  ["propellants explos pyrotech"] = "Propellants, Explosives, Pyrotechnics",
  ["propellants, explos, pyrotech"] = "Propellants, Explosives, Pyrotechnics",
  ["prophyl antivenerienne"] = "La Prophylaxie antivénérienne",
  ["prophyl sanit morale"] = "La Prophylaxie sanitaire et morale",
  ["propul power res"] = "Propulsion and Power Research",
  ["prose stud"] = "Prose studies",
  ["prospects (paris)"] = "Prospects",
  ["prostaglandins leukot essent fatty acids"] = "Prostaglandins, leukotrienes, and essential fatty acids",
  ["prostaglandins leukot med"] = "Prostaglandins, leukotrienes, and medicine",
  ["prostaglandins med"] = "Prostaglandins and medicine",
  ["prostaglandins other lipid mediat"] = "Prostaglandins & other lipid mediators",
  ["prostaglandins other lipid mediators"] = "Prostaglandins and Other Lipid Mediators",
  ["prostaglandins, leukotrienes essent fatty acids"] = "Prostaglandins, Leukotrienes and Essential Fatty Acids",
  ["prostate cancer"] = "Prostate cancer",
  ["prostate cancer imaging (2011)"] = "Prostate cancer imaging : image analysis and image-guided interventions : international workshop held in conjunction with MICCAI 2011, Toronto, Canada, September 22, 2011 : proceedings",
  ["prostate cancer prostatic dis"] = "Prostate cancer and prostatic diseases",
  ["prostate int"] = "Prostate international",
  ["prostate suppl"] = "The Prostate. Supplement",
  ["prosthet orthot int"] = "Prosthetics and orthotics international",
  ["prot child"] = "Protecting children",
  ["prot control mod power syst"] = "Protection and Control of Modern Power Systems",
  ["prot hum subj"] = "Protecting human subjects",
  ["prot met"] = "Protection of Metals",
  ["prot met phys chem"] = "Protection of Metals and Physical Chemistry of Surfaces",
  ["prot met phys chem surf"] = "Protection of Metals and Physial Chemistry of Surfaces",
  ["prot nat"] = "Protection de la nature",
  ["prot soc"] = "Protección social",
  ["protein alter erythroc low temp"] = "Protein alterations in erythrocytes at low temperatures. Bjorksten Research Foundation",
  ["protein cell"] = "Protein & cell",
  ["protein eng"] = "Protein Engineering",
  ["protein eng des sel"] = "Protein engineering, design & selection : PEDS",
  ["protein eng, des sel"] = "Protein Engineering, Design & Selection",
  ["protein expr purif"] = "Protein expression and purification",
  ["protein expression purif"] = "Protein Expression and Purification",
  ["protein j"] = "The protein journal",
  ["protein pept lett"] = "Protein and peptide letters",
  ["protein profile"] = "Protein profile",
  ["protein rev web"] = "Protein Reviews on the Web",
  ["protein sci"] = "Protein science : a publication of the Protein Society",
  ["protein seq data anal"] = "Protein sequences & data analysis",
  ["proteins struct funct bioinf"] = "Proteins: Structure, FUnction, and Bioinformatics",
  ["proteins: struct, funct, bioinf"] = "Proteins: Structure, Function, and Bioinformatics",
  ["proteins: struct, funct, genet"] = "Proteins: Structure, Function, and Genetics",
  ["proteoglycan res"] = "Proteoglycan Research",
  ["proteom bioinform"] = "Proteomics & bioinformatics : current research",
  ["proteome sci"] = "Proteome science",
  ["proteomics clin appl"] = "Proteomics. Clinical applications",
  ["proteomics insights"] = "Proteomics insights",
  ["proteomics: clin appl"] = "Proteomics: Clinical Applications",
  ["protes clin"] = "Protesis clínica; periódico técnico informativo para la profesión dental",
  ["protes dent"] = "El Protesista dental",
  ["protesis (b aires)"] = "Protesis",
  ["protet stomatol"] = "Protetyka stomatologiczna",
  ["protides biol fluid proc colloq"] = "Protides of the biological fluids; proceedings of the colloquium",
  ["protist genomics"] = "Protist genomics",
  ["protoc exch"] = "Protocol exchange",
  ["prov china"] = "Provincial China",
  ["prov med"] = "La Provence médicale",
  ["prov med j retrosp med sci"] = "Provincial medical journal and retrospect of the medical sciences",
  ["prov med surg j"] = "Provincial medical & surgical journal",
  ["prov med surg j (1840)"] = "Provincial medical & surgical journal",
  ["provence hist"] = "Provence historique; revue trimestrielle",
  ["provid stud res note agency health care policy res"] = "Provider studies research note",
  ["prx energy"] = "PRX Energy",
  ["prx life"] = "PRX Life",
  ["prx quantum"] = "PRX Quantum",
  ["prz gastroenterol"] = "Przegla̜d gastroenterologiczny",
  ["prz geogr"] = "Przegląd geograficzny. Polśkiĭ geograficheskiĭ obzor. Polish geographical review. Revue polonaise de géographie",
  ["prz hist osw"] = "Przegląd historyczno-oświatowy",
  ["prz menopauzalny"] = "Przegla̜d menopauzalny = Menopause review",
  ["prz orient"] = "Przegląd orientalistyczny",
  ["prz zach"] = "Przegląd zachodni",
  ["przegl dermatol"] = "Przegla̧d dermatologiczny",
  ["przegl epidemiol"] = "Przegla̧d epidemiologiczny",
  ["przegl hist"] = "Przegląd historyczny",
  ["przegl lek"] = "Przeglad Lekarski",
  ["przegl pol"] = "Przegląd polonijny",
  ["przem chem"] = "Przemysl Chemiczny",
  ["przeszlosc demograf pol"] = "Przeszłość demograficzna Polski",
  ["prépubl inst rech math av"] = "Prépublications de l’Institut de Recherche Mathématique Avancée",
  ["ps (wash dc)"] = "PS",
  ["ps polit sci polit"] = "PS, political science & politics",
  ["psd commun newsl"] = "PSD communication newsletter",
  ["pseudo diff oper"] = "Pseudo-Differential Operators. Theory and Applications",
  ["psi chi j undergrad res"] = "Psi Chi journal of undergraduate research",
  ["psi newsl"] = "PSI newsletter",
  ["psichiatr gen evol"] = "Psichiatria generale e dell' età evolutiva",
  ["psichologija (vilniaus univ)"] = "Psichologija",
  ["psicoanal appl med pedagog sociol lett arte"] = "Psicoanalisi applicata alla medicina, pedagogia, sociologia, letteratura ed arte",
  ["psicol conductual"] = "Psicología conductual",
  ["psicol educ (madr)"] = "Psicología Educativa",
  ["psicol estud"] = "Psicologia em estudo",
  ["psicol reflex crit"] = "Psicologia, reflexão e crítica : revista semestral do Departamento de Psicologia da UFRGS",
  ["psicol soc (bologna)"] = "Psicologia sociale",
  ["psicologica (valencia)"] = "Psicológica : revista de metodología y psicología experimental",
  ["psicooncologia (pozuelo de alarcon)"] = "Psicooncología",
  ["psikholog zh"] = "Psikhologicheskiĭ zhurnal",
  ["psiquis (mexico)"] = "Psiquis",
  ["psoriasis (auckl)"] = "Psoriasis (Auckland, N.Z.)",
  ["psoriasis forum"] = "Psoriasis forum",
  ["psr q"] = "PSR quarterly (Baltimore, Md.)",
  ["psych j"] = "PsyCh journal",
  ["psyche (camb mass)"] = "Psyche; a journal of entomology",
  ["psyche (stuttg)"] = "Psyche",
  ["psyche rev int sci homme psychanal"] = "Psyché; revue internationale des sciences de l'homme et de psychanalyse",
  ["psyche: j entomol"] = "Psyche: A Journal of Entomology",
  ["psychedelic med"] = "Psychedelic Medicine",
  ["psychiatr ann"] = "Psychiatric annals",
  ["psychiatr bull (2014)"] = "Psychiatric bulletin (2014)",
  ["psychiatr bull r coll psychiatr"] = "Psychiatric bulletin (London, England)",
  ["psychiatr clin (basel)"] = "Psychiatria clinica",
  ["psychiatr clin north am"] = "The Psychiatric clinics of North America",
  ["psychiatr commun"] = "Psychiatric communications",
  ["psychiatr danub"] = "Psychiatria Danubina",
  ["psychiatr dev"] = "Psychiatric developments",
  ["psychiatr enfant"] = "La Psychiatrie de l'enfant",
  ["psychiatr forum"] = "Psychiatric forum",
  ["psychiatr genet"] = "Psychiatric genetics",
  ["psychiatr hosp"] = "The Psychiatric hospital",
  ["psychiatr hung"] = "Psychiatria Hungarica : A Magyar Pszichiátriai Társaság tudományos folyóirata",
  ["psychiatr j univ ott"] = "Psychiatric journal of the University of Ottawa : Revue de psychiatrie de l'Université d'Ottawa",
  ["psychiatr med"] = "Psychiatric medicine",
  ["psychiatr med update mass gen hosp rev physicians"] = "Psychiatric medicine update: Massachusetts General Hospital reviews for physicians",
  ["psychiatr neurol (basel)"] = "Psychiatria et neurologia",
  ["psychiatr neurol med psychol (leipz)"] = "Psychiatrie, Neurologie, und medizinische Psychologie",
  ["psychiatr neurol med psychol beih"] = "Psychiatrie, Neurologie und medizinische Psychologie. Beihefte",
  ["psychiatr neurol neurochir"] = "Psychiatria, neurologia, neurochirurgia",
  ["psychiatr news"] = "Psychiatric news",
  ["psychiatr opin"] = "Psychiatric opinion",
  ["psychiatr pol"] = "Psychiatria polska",
  ["psychiatr prax"] = "Psychiatrische Praxis",
  ["psychiatr psychol law"] = "Psychiatry, psychology, and law : an interdisciplinary journal of the Australian and New Zealand Association of Psychiatry, Psychology and Law",
  ["psychiatr q"] = "The Psychiatric quarterly",
  ["psychiatr q suppl"] = "The Psychiatric quarterly. Supplement",
  ["psychiatr rehabil j"] = "Psychiatric rehabilitation journal",
  ["psychiatr res rep am psychiatr assoc"] = "Psychiatric research reports",
  ["psychiatr serv"] = "Psychiatric services (Washington, D.C.)",
  ["psychiatr times"] = "The Psychiatric times",
  ["psychiatrie (stuttg)"] = "Psychiatrie (Stuttgart, Germany)",
  ["psychiatry (abingdon)"] = "Psychiatry (Abingdon, England)",
  ["psychiatry (edgmont)"] = "Psychiatry (Edgmont (Pa. : Township))",
  ["psychiatry clin neurosci"] = "Psychiatry and clinical neurosciences",
  ["psychiatry clin psychopharmacol"] = "Psychiatry and Clinical Psychopharmacology",
  ["psychiatry dig"] = "Psychiatry digest",
  ["psychiatry investig"] = "Psychiatry investigation",
  ["psychiatry j"] = "Psychiatry journal",
  ["psychiatry med"] = "Psychiatry in medicine",
  ["psychiatry res"] = "Psychiatry research",
  ["psychiatry res neuroimaging"] = "Psychiatry research. Neuroimaging",
  ["psychoanal dialogues"] = "Psychoanalytic dialogues",
  ["psychoanal hist"] = "Psychoanalysis and history",
  ["psychoanal inq"] = "Psychoanalytic inquiry",
  ["psychoanal psychoanal rev"] = "Psychoanalysis and the psychoanalytic review",
  ["psychoanal psychol"] = "Psychoanalytic psychology : the official journal of the Division of Psychoanalysis, American Psychological Association, Division 39",
  ["psychoanal psychother"] = "Psychoanalytic psychotherapy",
  ["psychoanal q"] = "The Psychoanalytic quarterly",
  ["psychoanal rev"] = "Psychoanalytic review",
  ["psychoanal soc sci"] = "Psychoanalysis and the social sciences",
  ["psychoanal study child"] = "The Psychoanalytic study of the child",
  ["psychobiology (austin, tex)"] = "Psychobiology (Austin, Tex.)",
  ["psychodyn couns"] = "Psychodynamic counselling",
  ["psychodyn pract"] = "Psychodynamic practice",
  ["psychodyn psychiatry"] = "Psychodynamic psychiatry",
  ["psychohist rev"] = "The Psychohistory review",
  ["psychol abstr"] = "Psychological abstracts",
  ["psychol addict behav"] = "Psychology of addictive behaviors : journal of the Society of Psychologists in Addictive Behaviors",
  ["psychol aesthet creat arts"] = "Psychology of aesthetics, creativity, and the arts",
  ["psychol aging"] = "Psychology and aging",
  ["psychol assess"] = "Psychological assessment",
  ["psychol behav sci"] = "Psychology and behavioral sciences (New York, N.Y. 2012)",
  ["psychol belg"] = "Psychologica Belgica",
  ["psychol bull"] = "Psychological bulletin",
  ["psychol clin"] = "The Psychological clinic",
  ["psychol cogn sci"] = "Psychology and cognitive sciences : open journal",
  ["psychol conscious (wash d c)"] = "Psychology of consciousness (Washington, D.C.)",
  ["psychol crime law"] = "Psychology, crime & law : PC & L",
  ["psychol dev soc j"] = "Psychology and developing societies",
  ["psychol educ"] = "Psychology and education",
  ["psychol forsch"] = "Psychologische Forschung",
  ["psychol gezondh"] = "Psychologie & gezondheid",
  ["psychol health"] = "Psychology & health",
  ["psychol health med"] = "Psychology, health & medicine",
  ["psychol inj law"] = "Psychological injury and law",
  ["psychol inq"] = "Psychological inquiry",
  ["psychol issues"] = "Psychological issues",
  ["psychol learn motiv"] = "The psychology of learning and motivation",
  ["psychol manag j"] = "The psychologist manager journal",
  ["psychol mark"] = "Psychology & marketing",
  ["psychol med"] = "Psychological medicine",
  ["psychol med (paris)"] = "Psychologie médicale",
  ["psychol med monogr suppl"] = "Psychological medicine. Monograph supplement",
  ["psychol men masc"] = "Psychology of men & masculinity",
  ["psychol ment health care"] = "Psychology and mental health care : open access",
  ["psychol methods"] = "Psychological methods",
  ["psychol monogr"] = "Psychological monographs",
  ["psychol music"] = "Psychology of music",
  ["psychol neuropsychiatr vieil"] = "Psychologie & neuropsychiatrie du vieillissement",
  ["psychol neurosci"] = "Psychology & neuroscience",
  ["psychol pop media cult"] = "Psychology of popular media culture",
  ["psychol prax"] = "Psychologische Praxis; Schriftenreihe für Erziehung und Jugendpflege",
  ["psychol psychother"] = "Psychology and psychotherapy",
  ["psychol public policy law"] = "Psychology, public policy, and law : an official law review of the University of Arizona College of Law and the University of Miami School of Law",
  ["psychol rec"] = "The Psychological record",
  ["psychol rep"] = "Psychological reports",
  ["psychol res"] = "Psychological research",
  ["psychol res (libertyville)"] = "Psychology research (Libertyville, Ill.)",
  ["psychol res behav manag"] = "Psychology research and behavior management",
  ["psychol rev"] = "Psychological review",
  ["psychol rundsch"] = "Psychologische Rundschau; Ueberblick über die Fortschritte der Psychologie in Deutschland, Oesterreich, und der Schweiz",
  ["psychol sch"] = "Psychology in the schools",
  ["psychol sci"] = "Psychological science",
  ["psychol sci agenda"] = "Psychological science agenda",
  ["psychol sci public interest"] = "Psychological science in the public interest : a journal of the American Psychological Society",
  ["psychol sci q"] = "Psychology science quarterly",
  ["psychol serv"] = "Psychological services",
  ["psychol sex"] = "Psychology and sexuality",
  ["psychol sex orientat gend divers"] = "Psychology of sexual orientation and gender diversity",
  ["psychol sex rev"] = "Psychology of sexualities review",
  ["psychol sport exerc"] = "Psychology of sport and exercise",
  ["psychol stud (mysore)"] = "Psychological studies",
  ["psychol stud popul fam plan"] = "Psychological studies in population/family planning",
  ["psychol test assess model"] = "Psychological test and assessment modeling",
  ["psychol today"] = "Psychology today",
  ["psychol trauma"] = "Psychological trauma : theory, research, practice and policy",
  ["psychol violence"] = "Psychology of violence",
  ["psychol well being"] = "Psychology of well-being",
  ["psychol women q"] = "Psychology of women quarterly",
  ["psycholog relig spiritual"] = "Psychology of religion and spirituality",
  ["psychology (irvine)"] = "Psychology (Irvine, Calif.)",
  ["psychon bull rev"] = "Psychonomic bulletin & review",
  ["psychon sci"] = "Psychonomic science",
  ["psychopathol rev"] = "Psychopathology review",
  ["psychopharm rev"] = "Psychopharm review : timely reports in psychopharmacology and device-based therapies",
  ["psychopharmacol bull"] = "Psychopharmacology bulletin",
  ["psychopharmacol commun"] = "Psychopharmacology communications",
  ["psychopharmacol ser"] = "Psychopharmacology series",
  ["psychopharmacol serv cent bull"] = "Psychopharmacology Service Center bulletin",
  ["psychopharmacology (berl)"] = "Psychopharmacology",
  ["psychopharmacology (berlin, ger)"] = "Psychopharmacology (Berlin, Germany)",
  ["psychopharmacology (heidelberg, ger)"] = "Psychopharmacology (Heidelberg, Germany)",
  ["psychopharmacology suppl"] = "Psychopharmacology. Supplementum",
  ["psychosoc med"] = "Psycho-social medicine",
  ["psychosoc rehabil j"] = "Psychosocial rehabilitation journal",
  ["psychosom konsiliarpsychiatr"] = "Psychosomatik und Konsiliarpsychiatrie",
  ["psychosom med"] = "Psychosomatic medicine",
  ["psychother med psychol (stuttg)"] = "Psychotherapie, medizinische Psychologie",
  ["psychother psychosom"] = "Psychotherapy and psychosomatics",
  ["psychother psychosom med psychol"] = "Psychotherapie, Psychosomatik, medizinische Psychologie",
  ["psychother res"] = "Psychotherapy research : journal of the Society for Psychotherapy Research",
  ["psychotherapeut (berl)"] = "Psychotherapeut",
  ["psychotherapies (geneva)"] = "Psychothérapies (Geneva, Switzerland)",
  ["psychotherapy (chic)"] = "Psychotherapy (Chicago, Ill.)",
  ["psychotr (paris)"] = "Psychotropes (Paris, France)",
  ["pszichol tanulman"] = "Pszichológiai tanulmányok",
  ["ptep prog theor exp phys"] = "PTEP. Progress of Theoretical and Experimental Physics",
  ["ptolemaeus arab lat, stud"] = "Ptolemaeus Arabus et Latinus. Studies",
  ["pubbl cent sper agric for"] = "Pubblicazioni del Centro di sperimentazione agricola e forestale",
  ["pubbl stn zool napoli ii"] = "Pubblicazioni della Stazione Zoologica di Napoli. Section II: History and Philosophy of the Life Sciences",
  ["pubbl univ ferrara"] = "Pubblicazioni dell’Università di Ferrara",
  ["publ am inst hist pharm"] = "Publication - American Institute of the History of Pharmacy",
  ["publ arch henri-poincare"] = "Publications des Archives Henri-Poincare",
  ["publ astron soc aust"] = "Publications of the Astronomical Society of Australia : PASA",
  ["publ astron soc jpn"] = "Publications of the Astronomical Society of Japan",
  ["publ astron soc jpn nihon tenmon gakkai"] = "Publications of the Astronomical Society of Japan. Nihon Tenmon Gakkai",
  ["publ astron soc pac"] = "Publications of the Astronomical Society of the Pacific",
  ["publ australas inst min metall"] = "Publications of the Australasian Institute of Mining and Metallurgy",
  ["publ austrian ludwig wittgenstein soc new ser"] = "Publications of the Austrian Ludwig Wittgenstein Society. New Series",
  ["publ b aires cent investig"] = "Publicaciones. Buenos Aires (Argentina). Centro de Investigaciones Tisiológicas",
  ["publ cent rech math pures sér i"] = "Publications du Centre de Recherches en Mathématiques Pures, Séries I",
  ["publ centre rech math pures sér ii"] = "Publications du Centre de Recherches en Mathématiques Pures, Série II",
  ["publ child dev univ calif"] = "Publications in child development. University of California, Berkeley",
  ["publ cient alter"] = "Publicaciones científicas Alter",
  ["publ cient oficina sanit panam"] = "Publicación científica - Oficina Sanitaria Panamericana",
  ["publ cour eur droits homme sr a arrets decis"] = "Publications de la Cour européenne des droits de l'homme. Série A, Arrêts et décisions = Publications of the European Court of Human Rights. Series A, Judgments and decisions. European Court of Human Rights",
  ["publ dep mat"] = "Publicacoes do Departamento de Matematica",
  ["publ dep mat univ extremadura"] = "Publicaciones del Departamento de Matemáticas, Universidad de Extremadura",
  ["publ electron paleontol asoc"] = "Publicacion Electronica de la Asociacion Paleontologica Argentina",
  ["publ fac electr eng ser automat control"] = "Publications of the Faculty of Electrical Engineering",
  ["publ farm (sao paulo)"] = "Publicações farmacêuticas",
  ["publ ger hist inst"] = "Publications of The German Historical Institute",
  ["publ group adv psychiatry"] = "Publication - Group for the Advancement of Psychiatry",
  ["publ henri-poincaré-arch"] = "Publikationen des Henri-Poincaré-Archivs",
  ["publ hochiminh city math soc"] = "Publications of the HoChiMinh City Mathematical Society",
  ["publ ind med hist"] = "Publications in Indiana medical history",
  ["publ inst math (beograd) (ns)"] = "Institut Mathematique. Publications. Nouvelle Serie",
  ["publ inst pasteur guyane fr inini"] = "Publication. Cayenne, French Guiana. Institut Pasteur de la Guyane française et de l'Inini",
  ["publ inst rech math av"] = "Publication de l’Institut de Recherche Mathématique Avancée",
  ["publ inst rech math rennes"] = "Publications de l’Institut de Recherche Mathématique de Rennes",
  ["publ int res cent salzbg austria"] = "Publications of the International Research Center, Salzburg, Austria",
  ["publ mat"] = "Publicacions MatemÃ tiques",
  ["publ mat impa"] = "Publicacoes Matematicas do IMPA",
  ["publ mat urug"] = "Publicaciones Matematicas del Uruguay",
  ["publ math besancon algebre theorie nr"] = "Publications Mathematiques de Besancon. Algebre et Theorie des Nombres",
  ["publ math debrecen"] = "Publicationes Mathematicae Debrecen",
  ["publ math fac sci besançon"] = "Publications Mathématiques de la Faculté des Sciences de Besançon",
  ["publ math ihes"] = "Publications Mathematiques de l’IHES",
  ["publ math inst hautes etudes sci"] = "Publications Mathematiques. Institut de Hautes Etudes Scientifiques",
  ["publ math orsay"] = "Publications Mathématiques d’Orsay",
  ["publ math univ paris 7—denis diderot"] = "Publications Mathématiques de l’Université Paris 7—Denis Diderot",
  ["publ math univ pierre et marie curie"] = "Publications Mathématiques de l’Université Pierre et Marie Curie",
  ["publ med"] = "Publicaciones médicas",
  ["publ mus farm"] = "Publicaciones del Museo de la Farmacia",
  ["publ neth interuniv demogr inst nidi popul fam study cent cbgs"] = "Publications of the Netherlands Interdisciplinary Demographic Institute (N.I.D.I) and the Population and Family Study Centre (C.B.G.S.)",
  ["publ newton inst"] = "Publications of the Newton Institute",
  ["publ psychol"] = "Publications in psychology",
  ["publ public health univ calif"] = "Publications in public health. University of California (1868-1952)",
  ["publ res inst math sci"] = "Publications of the Research Institute for Mathematical Sciences",
  ["publ res q"] = "Publishing Research Quarterly",
  ["publ sec mat"] = "Publicacions de la Seccio de Mathematiques",
  ["publ soc hist archeol limbg"] = "Publications de la Société historique et archéologique dans le Limbourg. Limburgs Geschied- en Oudheidkundig Genootschap",
  ["publ soc lit sci"] = "Publication of the Society for Literature and Science: PSLS",
  ["publ thoresby soc"] = "The Publications of the Thoresby Society",
  ["publ univ chile"] = "Publicaciones. Universidad de Chile. Laboratorio de Medicina Experimental",
  ["publ univ east finl rep stud for nat sci"] = "Publications of the University of Eastern Finland. Reports and Studies in Forestry and Natural Sciences",
  ["publ univ joensuu, ser bi"] = "Publications of the University of Joensuu, Series BI",
  ["public adm"] = "Public administration",
  ["public adm dev"] = "Public administration and development : a journal of the Royal Institute of Public Administration",
  ["public adm rev"] = "Public administration review",
  ["public admin rev"] = "Public Administration Review",
  ["public aff q"] = "Public affairs quarterly",
  ["public aff rep"] = "Public affairs report : bulletin of the Bureau of Public Administration",
  ["public budg finance"] = "Public Budgeting and Finance",
  ["public choice"] = "Public choice",
  ["public citiz"] = "Public citizen (Washington, D.C.)",
  ["public finan q"] = "Public finance quarterly",
  ["public financ"] = "Public finance = Finances publiques",
  ["public finance"] = "Public Finance",
  ["public finance quart"] = "Public Finance Quarterly",
  ["public finance rev"] = "Public finance review : PFR",
  ["public health"] = "Public health",
  ["public health (fairfax)"] = "Public health (Fairfax, Va.)",
  ["public health action"] = "Public health action",
  ["public health bulletin (wash d c)"] = "Public health bulletin",
  ["public health ethics"] = "Public health ethics",
  ["public health front"] = "Public health frontier",
  ["public health genomics"] = "Public health genomics",
  ["public health lab"] = "The Public health laboratory",
  ["public health monogr"] = "Public health monograph",
  ["public health news"] = "Public health news",
  ["public health nurs"] = "Public health nursing (Boston, Mass.)",
  ["public health nutr"] = "Public health nutrition",
  ["public health panor"] = "Public health panorama : journal of the WHO Regional Office for Europe = Panorama obshchestvennogo zdravookhranenii︠a︡",
  ["public health pap"] = "Public health papers",
  ["public health pap rep"] = "Public health papers and reports",
  ["public health pract (oxf)"] = "Public health in practice (Oxford, England)",
  ["public health rep"] = "Public health reports (Washington, D.C. : 1974)",
  ["public health res (southampt)"] = "Public health research (Southampton, England)",
  ["public health res pract"] = "Public health research & practice",
  ["public health rev"] = "Public health reviews",
  ["public health tech monogr us"] = "Public health technical monograph. United States. Public Health Service",
  ["public hist"] = "The Public historian",
  ["public hist news"] = "Public history news",
  ["public interest"] = "The Public interest",
  ["public law"] = "Public law",
  ["public law forum"] = "Public law forum",
  ["public management sources"] = "Public Management Sources",
  ["public opin q"] = "Public opinion quarterly",
  ["public pers manage"] = "Public personnel management",
  ["public pol"] = "Public Policy",
  ["public policy"] = "Public policy",
  ["public policy adm"] = "Public policy and administration",
  ["public policy aging rep"] = "The Public policy and aging report",
  ["public relat j"] = "The Public relations journal",
  ["public relat rev"] = "Public relations review",
  ["public sect contract rep"] = "Public sector contracting report : the monthly guide to Medicare and Medicaid managed care",
  ["public sector health care risk manage"] = "Public sector, health care risk management",
  ["public serv rev uk sci technol"] = "Public service review. UK science & technology",
  ["public transp"] = "Public Transport",
  ["public underst sci"] = "Public understanding of science (Bristol, England)",
  ["public welf"] = "Public welfare",
  ["public welf indiana"] = "Public welfare in Indiana",
  ["public works manage policy"] = "Public Works Management & Policy",
  ["publicaciones del observatorio de santiago, ii consejo superior de investigaciones cientificas, instituto nacional de g"] = "Publicaciones del Observatorio de Santiago, II. Consejo Superior de Investigaciones Cientificas, Instituto Nacional de Geofisica",
  ["publicacoes med"] = "Publicações médicas",
  ["publications de l’institut de mathematique de l’universite de strasbourg, xiv actualites scientifiques et industr"] = "Publications de l’Institut de Mathematique de l’Universite de Strasbourg, XIV. Actualites Scientifiques et Industrielles",
  ["publications de l’institut de mathematique de l’universite de strasbourg, xv actualites scientifiques et industri"] = "Publications de l’Institut de Mathematique de l’Universite de Strasbourg, XV. Actualites Scientifiques et Industrielles",
  ["publico priv"] = "O público e o privado : caderno dos núcleos e grupos de pesquisa vinculados ao Mestrado Acadêmico em Políticas Públicas e Sociedade da Universidade Estadual do Ceará",
  ["pulaski cty hist rev"] = "Pulaski County historical review",
  ["pulm circ"] = "Pulmonary circulation",
  ["pulm crit care med"] = "Pulmonary and critical care medicine",
  ["pulm med"] = "Pulmonary medicine",
  ["pulm pharmacol"] = "Pulmonary pharmacology",
  ["pulm pharmacol ther"] = "Pulmonary pharmacology & therapeutics",
  ["pulm res respir med"] = "Pulmonary research and respiratory medicine : open journal",
  ["pulm ther"] = "Pulmonary therapy",
  ["pulse (basel)"] = "Pulse (Basel, Switzerland)",
  ["pump j undergrad res"] = "The PUMP Journal of Undergraduate Research",
  ["punishm soc"] = "Punishment & society",
  ["punjab j polit"] = "Punjab journal of politics",
  ["punjab med j"] = "The Punjab medical journal",
  ["punjab med j (banga)"] = "The Punjab medical journal",
  ["punjab univ econ"] = "Punjab University Economist",
  ["punjab univ j math (lahore)"] = "The Punjab University. Journal of Mathematics",
  ["purch adm"] = "Purchasing administration",
  ["purdons pa statut annot pa"] = "Purdon's Pennsylvania statutes, annotated. Pennsylvania",
  ["purdue pharm"] = "The Purdue pharmacist",
  ["pure appl anal"] = "Pure and Applied Analysis",
  ["pure appl chem"] = "Pure and Applied Chemistry",
  ["pure appl funct anal"] = "Pure and Applied Functional Analysis",
  ["pure appl geophys"] = "Pure and Applied Geophysics",
  ["pure appl math"] = "Pure and Applied Mathematics",
  ["pure appl math (amst)"] = "Pure and Applied Mathematics (Amsterdam)",
  ["pure appl math (boca raton)"] = "Pure and Applied Mathematics (Boca Raton)",
  ["pure appl math (hoboken)"] = "Pure and Applied Mathematics (Hoboken)",
  ["pure appl math q"] = "Pure and Applied Mathematics Quarterly",
  ["pure appl math sci"] = "Pure and Applied Mathematika Sciences",
  ["pure appl opt"] = "Pure and Applied Optics",
  ["pure appl phys"] = "Pure and Applied Physics",
  ["pure appl undergrad texts"] = "Pure and Applied Undergraduate Texts",
  ["pure cult study bact"] = "Pure culture study of bacteria",
  ["pure math appl"] = "Pure Mathematics and Applications",
  ["pure math appl (puma)"] = "Pure Mathematics and Applications. PU.M.A.",
  ["pure math manuscript"] = "Pure Mathematics Manuscript",
  ["purinergic signal"] = "Purinergic signalling",
  ["purinergic signalling"] = "Purinergic Signalling",
  ["pus media pelatih aids untuk war newsl"] = "Pusat Media Pelatihan AIDS untuk Wartawan Newsletter",
  ["pyrethrum post"] = "Pyrethrum Post",
  ["pz wiss"] = "PZ Wissenschaft : Pharmazeutische Zeitung, Wissenschaftsausgabe",
  ["q appl math"] = "Quarterly of Applied Mathematics",
  ["q bull"] = "Quarterly bulletin. New York (N.Y.). Department of Health",
  ["q bull chin bibliogr"] = "Quarterly bulletin of Chinese bibliography",
  ["q bull commonw med assoc"] = "Quarterly bulletin (Commonwealth Medical Association)",
  ["q bull indiana univ med cent"] = "The Quarterly bulletin of the Indiana University Medical Center",
  ["q bull int assoc agric inf spec"] = "Quarterly Bulletin of the International Association of Agricultural Information Specialists",
  ["q bull la board health"] = "Quarterly bulletin. Louisiana. Board of Health",
  ["q bull northwest univ med sch"] = "Quarterly bulletin. Northwestern University (Evanston, Ill.). Medical School",
  ["q bull s afr libr"] = "Quarterly bulletin of the South African Library = Kwartaalblad van die Suid-Afrikaanse Biblioteek",
  ["q bull sea view hosp"] = "Quarterly bulletin of Sea View Hospital. New York. Sea View Hospital, Staten Island. Clinical Society",
  ["q check list classical stud"] = "ABS quarterly check-list of classical studies",
  ["q demogr bull"] = "Quarterly demographic bulletin",
  ["q dent rev"] = "Quarterly Dental Review",
  ["q econ comment"] = "Quarterly economic commentary",
  ["q j adm"] = "The quarterly journal of administration",
  ["q j chem soc"] = "Quarterly Journal of the Chemical Society",
  ["q j crude drug res"] = "Quarterly journal of crude drug research = Vierteljährliche Zietschrift für Drogen-Forschung = Revue trimestrielle des recherches sur les matières premières",
  ["q j econ"] = "The quarterly journal of economics",
  ["q j eng geol"] = "Quarterly Journal of Engineering Geology",
  ["q j eng geol hydrogeol"] = "Quarterly Journal of Engineering Geology and Hydrogeology",
  ["q j exp physiol"] = "Quarterly journal of experimental physiology (Cambridge, England)",
  ["q j exp physiol cogn med sci"] = "Quarterly journal of experimental physiology and cognate medical sciences",
  ["q j exp psychol"] = "The Quarterly journal of experimental psychology",
  ["q j exp psychol (colchester)"] = "Quarterly journal of experimental psychology (2006)",
  ["q j exp psychol (hove)"] = "Quarterly journal of experimental psychology (2006)",
  ["q j exp psychol a"] = "The Quarterly journal of experimental psychology. A, Human experimental psychology",
  ["q j exp psychol b"] = "The Quarterly journal of experimental psychology. B, Comparative and physiological psychology",
  ["q j finance"] = "The quarterly journal of finance",
  ["q j geol soc london"] = "Quarterly Journal of the Geological Society of London",
  ["q j int agric"] = "Quarterly Journal of International Agriculture",
  ["q j libr cong"] = "Quarterly journal. Library of Congress",
  ["q j math"] = "Quarterly Journal of Mathematics",
  ["q j mech appl math"] = "Quarterly Journal of Mechanics and Applied Mathematics",
  ["q j med"] = "The Quarterly journal of medicine",
  ["q j ment health"] = "Quarterly journal of mental health",
  ["q j microsc sci"] = "Quarterly Journal of Microscopical Science",
  ["q j nucl med"] = "The quarterly journal of nuclear medicine : official publication of the Italian Association of Nuclear Medicine (AIMN) [and] the International Association of Radiopharmacology (IAR)",
  ["q j nucl med mol imaging"] = "The quarterly journal of nuclear medicine and molecular imaging : official publication of the Italian Association of Nuclear Medicine (AIMN) [and] the International Association of Radiopharmacology (IAR), [and] Section of the Society of Radiopharmaceutical Chemistry and Biology",
  ["q j pharm pharmacol"] = "Quarterly journal of pharmacy and pharmacology",
  ["q j r astron soc"] = "Quarterly Journal of the Royal Astronomical Society",
  ["q j r meteorol soc"] = "Quarterly journal of the Royal Meteorological Society. Royal Meteorological Society (Great Britain)",
  ["q j r meteorolog soc"] = "Quarterly Journal of the Royal Meteorological Society",
  ["q j sci lit arts"] = "Quarterly Journal of Science, Literature, and the Arts",
  ["q j soc aff"] = "The Quarterly journal of social affairs",
  ["q j speech"] = "The Quarterly journal of speech",
  ["q j stud alcohol"] = "Quarterly Journal of Studies on Alcohol",
  ["q magn reson biol med"] = "Quarterly of Magnetic Resonance in Biology and Medicine",
  ["q med rev"] = "Quarterly medical review",
  ["q natl dent assoc"] = "The Quarterly of the National Dental Association, Inc",
  ["q popul bull"] = "Quarterly population bulletin. New Zealand. Department of Statistics",
  ["q prog rep united states air force radiat lab univ chic"] = "Quarterly progress report. United States. Air Force. Radiation Laboratory, University of Chicago",
  ["q res rep"] = "Quarterly research report. USAF School of Aviation Medicine",
  ["q rev"] = "Quarterly review (London, England)",
  ["q rev allergy appl immunol"] = "Quarterly review of allergy and applied immunology",
  ["q rev biol"] = "The Quarterly review of biology",
  ["q rev biophys"] = "Quarterly reviews of biophysics",
  ["q rev br red cross soc"] = "Quarterly review. British Red Cross Society",
  ["q rev chem soc"] = "Quarterly Reviews, Chemical Society",
  ["q rev dc nurses assoc"] = "The Quarterly review",
  ["q rev distance educ"] = "Quarterly review of distance education",
  ["q rev drug metab drug interact"] = "Quarterly reviews on drug metabolism and drug interactions",
  ["q rev econ bus"] = "The Quarterly review of economics and business",
  ["q rev econ finance"] = "The Quarterly review of economics and finance : journal of the Midwest Economics Association",
  ["q rev med"] = "Quarterly review of medicine",
  ["q rev obstet gynecol"] = "Quarterly review of obstetrics and gynecology",
  ["q rev otorhinolaryngol bronchoesophagol"] = "Quarterly review of otorhinolaryngology and broncho-esophagology",
  ["q rev pediatr"] = "Quarterly review of pediatrics",
  ["q rev psychiatry neurol"] = "Quarterly review of psychiatry and neurology",
  ["q rev surg"] = "Quarterly review of surgery",
  ["q rev surg obstet gynecol"] = "Quarterly review of surgery, obstetrics and gynecology",
  ["q rev u s veterans adm"] = "Quarterly review. United States. Veterans Administration. Coordination Service. Research Division",
  ["qa brief"] = "Q.A. brief",
  ["qa rev"] = "QA review : quality assurance news and views",
  ["qatar med j"] = "Qatar medical journal",
  ["qatar univ sci j"] = "Qatar University Science Journal",
  ["qing shi yan jiu"] = "Qing shi yan jiu = Qingshi yanjiu",
  ["qld dent j"] = "The Queensland dental journal",
  ["qld nurse"] = "The Queensland nurse",
  ["qld nurses j"] = "Queensland nurses journal",
  ["qmw maths notes"] = "QMW Maths Notes",
  ["qp–pq: quantum probab white noise anal"] = "QP–PQ: Quantum Probability and White Noise Analysis",
  ["qrb discov"] = "QRB discovery",
  ["qrb discovery"] = "QRB Discovery",
  ["qrb qual rev bull"] = "QRB. Quality review bulletin",
  ["qrc advis"] = "QRC advisor",
  ["qsar comb sci"] = "QSAR & Combinatorial Science",
  ["qsci proc"] = "Qscience proceedings",
  ["quad accad pontaniana"] = "Quaderni dell’ Accademia Pontaniana",
  ["quad accad sci torino"] = "Quaderni",
  ["quad antibiot"] = "Quaderni di antibiotica",
  ["quad azione soc"] = "Quaderni Di Azione Sociale",
  ["quad clin ostet ginecol"] = "Quaderni di clinica ostetrica e ginecologica",
  ["quad criminol clin"] = "Quaderni di criminologia clinica",
  ["quad ing chim ital"] = "Quaderni dell’Ingegnere Chimico Italiano",
  ["quad int stor med sanita"] = "Quaderni internazionali di storia della medicina e della sanità",
  ["quad mat"] = "Quaderni di Matematica",
  ["quad merceol"] = "Quaderni di merceologia",
  ["quad nutr"] = "Quaderni della nutrizione",
  ["quad ostet"] = "Il Quaderno dell'ostetrica",
  ["quad radiol"] = "Quaderni di radiologia",
  ["quad raggruppamento toscoumbroemiliano stor med"] = "Quaderni del Raggruppamento tosco-umbro-emiliano di storia della medicina",
  ["quad ric didatt"] = "Quaderni di Ricerca in Didattica",
  ["quad sc norm super di pisa (n s)"] = "Quaderni. Scuola Normale Superiore di Pisa (Nuova Serie)",
  ["quad sc norm super pisa sel"] = "Quaderni. Scuola Normale Superiore di Pisa. Selections",
  ["quad sclavo diagn"] = "Quaderni Sclavo di diagnostica clinica e di laboratorio",
  ["quad stat"] = "Quaderni di statistica",
  ["quad stor"] = "Quaderni storici",
  ["quad stor med sci"] = "Quaderni di storia della medicina e della scienza",
  ["quad stor univ padova"] = "Quaderni per la storia dell'Università di Padova",
  ["quadern  med"] = "Quaderni medici",
  ["quaderni econ finanza"] = "Quaderni di Economia e Finanza",
  ["quaderni storici delle marche"] = "Quaderni Storici delle Marche",
  ["quaest math"] = "Quaestiones mathematicae : journal of the South African Mathematical Society",
  ["quaestiones math"] = "Quaestiones Mathematicae",
  ["quaker hist"] = "Quaker history",
  ["qual ageing"] = "Quality in ageing : policy, practice and research",
  ["qual ageing older adults"] = "Quality in ageing and older adults",
  ["qual assur"] = "Quality assurance (San Diego, Calif.)",
  ["qual assur health care"] = "Quality assurance in health care : the official journal of the International Society for Quality Assurance in Health Care",
  ["qual assur saf crops food"] = "Quality Assurance & Safety of Crops and Food",
  ["qual assur util rev"] = "Quality assurance and utilization review : official journal of the American College of Utilization Review Physicians",
  ["qual connect"] = "Quality connection (Brookline, Mass.)",
  ["qual eng"] = "Quality Engineering",
  ["qual health care"] = "Quality in health care : QHC",
  ["qual health res"] = "Qualitative health research",
  ["qual inq"] = "Qualitative inquiry : QI",
  ["qual lett healthc lead"] = "The Quality letter for healthcare leaders",
  ["qual life cardiovasc care"] = "Quality of life and cardiovascular care",
  ["qual life res"] = "Quality of life research : an international journal of quality of life aspects of treatment, care and rehabilitation",
  ["qual manag health care"] = "Quality management in health care",
  ["qual prim care"] = "Quality in primary care",
  ["qual prog"] = "Quality progress",
  ["qual psychol"] = "Qualitative psychology (Washington, D.C.)",
  ["qual quant"] = "Quality & quantity",
  ["qual quant anal sci sch commun"] = "Qualitative and Quantitative Analysis of Scientific and Scholarly Communication",
  ["qual reliab"] = "Quality and Reliability",
  ["qual reliab eng int"] = "Quality and Reliability Engineering International",
  ["qual rep"] = "Qualitative report (Online)",
  ["qual res"] = "Qualitative research : QR",
  ["qual res j"] = "Qualitative Research Journal",
  ["qual res med healthc"] = "Qualitative research in medicine & healthcare",
  ["qual res psychol"] = "Qualitative research in psychology",
  ["qual res sport exerc"] = "Qualitative research in sport and exercise",
  ["qual res sport exerc health"] = "Qualitative research in sport, exercise and health",
  ["qual res sport exercise health"] = "Qualitative Research in Sport, Exercise and Health",
  ["qual saf health care"] = "Quality & safety in health care",
  ["qual soc work"] = "Qualitative social work : QSW : research and practice",
  ["qual sociol"] = "Qualitative sociology",
  ["qual technol quant manag"] = "Quality technology & quantitative management",
  ["qual technol quant manage"] = "Quality Technology & Quantitative Management",
  ["qual theory dyn syst"] = "Qualitative Theory of Dynamical Systems",
  ["qual user exp"] = "Quality and user experience",
  ["quant biol"] = "Quantitative biology (Beijing, China)",
  ["quant biosci"] = "Quantitative bio-science",
  ["quant econ"] = "Quantitative Economics",
  ["quant econom"] = "Quantitative economics",
  ["quant finance"] = "Quantitative Finance",
  ["quant geol geostat"] = "Quantitative Geology and Geostatistics",
  ["quant imaging med surg"] = "Quantitative imaging in medicine and surgery",
  ["quant infrared thermogr j"] = "Quantitative infrared thermography journal",
  ["quant method psychol"] = "The quantitative methods for psychology",
  ["quant methods humanit soc sci"] = "Quantitative Methods in the Humanities and Social Sciences",
  ["quant microbiol"] = "Quantitative Microbiology",
  ["quant netw biol"] = "Quantitative and Network Biology",
  ["quant perspect behav econ finance"] = "Quantitative Perspectives on Behavioral Economics and Finance",
  ["quant plant biol"] = "Quantitative Plant Biology",
  ["quant sci stud"] = "Quantitative Science Studies",
  ["quant struct-act relat"] = "Quantitative Structure-Activity Relationships",
  ["quantitative appl in the social sciences"] = "Quantitative Applications in the Social Sciences",
  ["quantitative method unternehmungsplanung"] = "Quantitative Methoden der Unternehmungsplanung",
  ["quantitative stud social relations"] = "Quantitative Studies in Social Relations",
  ["quantum beam sci"] = "Quantum Beam Science",
  ["quantum bioinform iv (2010)"] = "Quantum bioinformatics IV : from quantum information to bio-informatics : Tokyo University of Science, Japan, 10-13 March 2010",
  ["quantum bioinform v (2011)"] = "Quantum bio-informatics V : proceedings of the quantum bio-informatics 2011, Tokyo University of Science, Japan, 7-12 March 2011. Quantum Bio-Informatics (Conference) (5th : 2011 : Tokyo, Japan)",
  ["quantum elec (woodbury)"] = "Quantum electronics",
  ["quantum electron"] = "Quantum Electronics",
  ["quantum electron (uk) or quantum electron (usa)"] = "Quantum Electronics",
  ["quantum electron—principles appl"] = "Quantum Electronics—Principles and Applications",
  ["quantum eng"] = "Quantum Engineering",
  ["quantum front"] = "Quantum Frontiers",
  ["quantum inf comput"] = "Quantum Information and Computation",
  ["quantum inf process"] = "Quantum Information Processing",
  ["quantum mach intell"] = "Quantum machine intelligence",
  ["quantum opt"] = "Quantum Optics",
  ["quantum rep"] = "Quantum Reports",
  ["quantum sci technol"] = "Quantum Science and Technology",
  ["quantum semiclass optics"] = "Quantum and Semiclassical Optics",
  ["quantum semiclassical opt"] = "Quantum and Semiclassical Optics",
  ["quantum stud math found"] = "Quantum Studies. Mathematics and Foundations",
  ["quantum stud: math found"] = "Quantum Studies: Mathematics and Foundations",
  ["quantum topol"] = "Quantum Topology",
  ["quart appl math"] = "Quarterly of Applied Mathematics",
  ["quart int"] = "Quarternary International",
  ["quart j austrian econ"] = "Quarterly Journal of Austrian Economics",
  ["quart j bus econ"] = "Quarterly Journal of Business and Economics",
  ["quart j econ"] = "Quarterly Journal of Economics",
  ["quart j indian econ stud soc sci"] = "Quarterly Journal of Indian Economic Studies in Social Science",
  ["quart j math oxford ser (2)"] = "The Quarterly Journal of Mathematics",
  ["quart j mech appl math"] = "The Quarterly Journal of Mechanics and Applied Mathematics",
  ["quart j polit sci"] = "Quarterly journal of political science",
  ["quart j roy meteorol soc"] = "Quarterly Journal of the Royal Meteorological Society",
  ["quart rev agr econ"] = "Quarterly Review of Agricultural Economics",
  ["quart rev econ bus"] = "Quarterly Review of Economics and Business",
  ["quart rev econ finance"] = "Quarterly Review of Economics and Finance",
  ["quart rev rural economy"] = "Quarterly Review of the Rural Economy",
  ["quarterly phi beta pi med fratern"] = "Quarterly. Phi Beta Pi Medical Fraternity",
  ["quasigroups related systems"] = "Quasigroups and Related Systems",
  ["quat geochronol"] = "Quaternary Geochronology",
  ["quat int"] = "Quaternary International",
  ["quat res"] = "Quaternary Research",
  ["quat sci adv"] = "Quaternary Science Advances",
  ["quat sci rev"] = "Quaternary Science Reviews",
  ["que sais-je?"] = "Que Sais-Je?",
  ["queen's q"] = "Queen's quarterly",
  ["queens gazette"] = "Queen's gazette",
  ["queens law j"] = "Queen's law journal",
  ["queens med mag (1972)"] = "Queen's medical magazine",
  ["queens nurs j"] = "Queen's nursing journal",
  ["queen’s papers in pure and appl math"] = "Queen’s Papers in Pure and Applied Mathematics",
  ["queen’s q"] = "Queen’s quarterly",
  ["quellen beitr gesch univ jena"] = "Quellen und Beitrage zur Geschichte der Universitat Jena",
  ["quellen forsch ital arch bibl"] = "Quellen und Forschungen aus italienischen Archiven und Bibliotheken",
  ["quellen leben wirk adam ries sohne"] = "Quellen zum Leben und Wirken Adam Ries’ und seiner Sohne",
  ["quellen stud gesch pharm"] = "Quellen und Studien zur Geschichte der Pharmazie",
  ["quercy rech"] = "Quercy recherche",
  ["quest d int"] = "Questions [d'internat]",
  ["questions answers gen topology"] = "Questions and Answers in General Topology",
  ["queueing syst"] = "Queueing systems",
  ["queueing systems theory appl"] = "Queueing Systems",
  ["qufu shifan daxue xuebao ziran kexue ban"] = "Qufu Shifan Daxue Xuebao",
  ["quim anal (barcelona)"] = "Quimica Analitica",
  ["quim bras"] = "Quimica no Brasil",
  ["quim nova"] = "Quimica Nova",
  ["quinnipiac health law j"] = "Quinnipiac health law journal",
  ["quintessence dent technol"] = "Quintessence of dental technology",
  ["quintessence int"] = "Quintessence international (Berlin, Germany : 1985)",
  ["quintessence int (berl)"] = "Quintessence international",
  ["quintessence int [fr]"] = "Quintessence International. Edition Francaise",
  ["quintessence int dent dig"] = "Quintessence international, dental digest",
  ["quintessence int fr"] = "Quintessence International. Edition Francaise",
  ["quintessencia protese lab"] = "Quintessência de prótese de laboratório",
  ["quintessenz impulse"] = "Quintessenz-Impulse : magazin für Kommunikation & Praxismanagement",
  ["quintessenz j"] = "Quintessenz Journal",
  ["quintessenz zahntech"] = "Die Quintessenz der Zahntechnik",
  ["quinto cent"] = "Quinto centenario",
  ["qüestiió (2)"] = "Qüestiió",
  ["r can dent corps q"] = "The Royal Canadian Dental Corps quarterly",
  ["r dent hosp mag"] = "The Royal Dental Hospital magazine",
  ["r i dent j"] = "Rhode Island dental journal",
  ["r i hist"] = "Rhode Island history",
  ["r i med"] = "Rhode Island medicine",
  ["r i med j"] = "R.I. medical journal",
  ["r i med j (2013)"] = "Rhode Island medical journal (2013)",
  ["r inst chem lect monogr rep"] = "Royal Institute of Chemistry, Lectures, Monographs, and Reports",
  ["r inst chem lect ser"] = "Royal Institute of Chemistry, Lecture Series",
  ["r inst chem rev"] = "Royal Institute of Chemistry, Reviews",
  ["r inst public health hyg j"] = "The Royal Institute of Public Health and Hygiene journal",
  ["r j"] = "The R journal",
  ["r melb hosp clin rep"] = "Royal Melbourne Hospital clinical reports",
  ["r nav med bull"] = "Royal naval medical bulletin",
  ["r s rep"] = "R & S report. Hawaii. Department of Health. Research and Statistics Office",
  ["r sitios"] = "Reales sitios",
  ["r soc health j"] = "Royal Society of Health journal",
  ["r soc lond philos trans ser a math phys eng sci"] = "The Royal Society of London",
  ["r soc lond proc ser a math phys eng sci"] = "The Royal Society of London",
  ["r soc open sci"] = "Royal Society open science",
  ["rabies bull eur"] = "Rabies bulletin Europe",
  ["race cl"] = "Race & class",
  ["race ethn educ"] = "Race, ethnicity and education",
  ["race gend cl"] = "Race, gender & class (Towson, Md.)",
  ["race gend sci"] = "Race, Gender, and Science",
  ["race justice"] = "Race and justice",
  ["race soc"] = "Race & society : Official journal of the Association of Black Sociologists",
  ["race soc probl"] = "Race and social problems",
  ["rad conf proc"] = "RAD Conference Proceedings",
  ["rad hrvat akad znan umjet mat znan"] = "Rad Hrvatske Akademije Znanosti i Umjetnosti. Matematicke Znanosti",
  ["rad hrvatske akad znan umjet"] = "Rad Hrvatske Akademije Znanosti i Umjetnosti",
  ["rad jugosl akad znan umjet razred med znan"] = "Rad Jugoslavenske akademije znanosti i umjetnosti. Razred za medicinske znanosti",
  ["rad mat"] = "Radovi Matematički",
  ["rad med fak zagrebu"] = "Radovi Medicinskog faculteta u Zagrebu",
  ["radc tdr"] = "RADC-TDR-. Rome Air Development Center",
  ["radiat biol"] = "Radiation Biology",
  ["radiat data rep"] = "Radiation data and reports",
  ["radiat detect technol methods"] = "Radiation Detection Technology and Methods",
  ["radiat eff"] = "Radiation effects",
  ["radiat eff defects solids"] = "Radiation Effects and Defects in Solids",
  ["radiat environ biophys"] = "Radiation and environmental biophysics",
  ["radiat meas"] = "Radiation Measurements",
  ["radiat med"] = "Radiation medicine",
  ["radiat med prot"] = "Radiation medicine and protection",
  ["radiat oncol"] = "Radiation oncology (London, England)",
  ["radiat oncol investig"] = "Radiation oncology investigations",
  ["radiat oncol j"] = "Radiation oncology journal",
  ["radiat phys chem"] = "Radiation Physics and Chemistry",
  ["radiat phys chem oxf engl 1993"] = "Radiation physics and chemistry (Oxford, England : 1993)",
  ["radiat prot dosim"] = "Radiation Protection Dosimetry",
  ["radiat prot dosimetry"] = "Radiation protection dosimetry",
  ["radiat res"] = "Radiation Research",
  ["radiat res suppl"] = "Radiation research. Supplement",
  ["radiats bezop zashch aes"] = "Radiatsionnaya Bezopasnost i Zashchita AES",
  ["radiats biol radioecol"] = "Radiatsionnaia Biologiia, Radioecologiia",
  ["radic am"] = "Radical America",
  ["radic hist rev"] = "Radical history review",
  ["radio electron eng"] = "Radio and Electronic Engineer",
  ["radio eng electron (ussr)"] = "Radio Engineering and Electronics (USSR) [translation of Radiotekhnika i Elektronika]",
  ["radio eng electron phys (ussr)"] = "Radio Engineering and Electronic Physics (USSR) [translation of Radiotekhnika i Elektronika]",
  ["radio phys radio astron"] = "Radio Physics and Radio Astronomy",
  ["radio sci"] = "Radio Science",
  ["radioactiv nat strontium hum bone"] = "Radioactive and natural strontium in human bone; U. K. results",
  ["radiobiol  radiother"] = "Radiobiologia, Radiotherapia",
  ["radiobiol radioter fis med"] = "Radiobiologia, radioterapia, e fisica medica",
  ["radiobiol radiother"] = "Radiobiologia, Radiotherapia",
  ["radiobiol radiother (berl)"] = "Radiobiologia, radiotherapia",
  ["radiochemistry (moscow, russ fed)"] = "Radiochemistry (Moscow, Russian Federation)(Translation of Radiokhimiya)",
  ["radiochemistry (n y, ny, u s)"] = "Radiochemistry (New York, NY, United States)",
  ["radiochim acta"] = "Radiochimica Acta",
  ["radioelectron commun syst"] = "Radioelectronics and Communications Systems",
  ["radiogr clin photogr"] = "Radiography and clinical photography",
  ["radiogr today"] = "Radiography today",
  ["radiography (lond)"] = "Radiography (London, England : 1995)",
  ["radiol artif intell"] = "Radiology. Artificial intelligence",
  ["radiol austriaca"] = "Radiologia Austriaca",
  ["radiol bras"] = "Radiologia brasileira",
  ["radiol cardiothorac imaging"] = "Radiology. Cardiothoracic imaging",
  ["radiol case rep"] = "Radiology case reports",
  ["radiol clin"] = "Radiologia clinica",
  ["radiol clin (basel)"] = "Radiologia Clinica",
  ["radiol clin biol"] = "Radiologia clinica et biologica",
  ["radiol clin north am"] = "Radiologic clinics of North America",
  ["radiol diagn (berl)"] = "Radiologia diagnostica",
  ["radiol health data rep"] = "Radiological health data and reports",
  ["radiol imaging cancer"] = "Radiology. Imaging cancer",
  ["radiol infect dis"] = "Radiology of Infectious Diseases",
  ["radiol interam"] = "Radiología interamericana",
  ["radiol manage"] = "Radiology management",
  ["radiol med"] = "La Radiologia medica",
  ["radiol med (torino)"] = "Radiologia Medica",
  ["radiol oncol"] = "Radiology and oncology",
  ["radiol open j"] = "Radiology - open journal",
  ["radiol phys technol"] = "Radiological physics and technology",
  ["radiol prat"] = "Radiologia pratica",
  ["radiol res pract"] = "Radiology research and practice",
  ["radiol sper"] = "Radiologia sperimentale",
  ["radiol technol"] = "Radiologic technology",
  ["radiologia (panama)"] = "Radiologia",
  ["radiologia (roma)"] = "Radiologia; rassegna internazionale trimestrale di radiobiologia, radioterapia, radiodiagnostica, terapia fisica e fisica applicata alla medicina",
  ["radiophys and quantum electronics"] = "Radiophysics and Quantum Electronics",
  ["radiophys quantum electron"] = "Radiophysics and Quantum Electronics",
  ["radiotekh elektron (moscow)"] = "Radiotekhnika i Elektronika (Moscow)",
  ["radiotekh elektron [radio eng electron (ussr)]"] = "Radiotekhnika i Elektronika [Radio Engineering and Electronics (USSR)]",
  ["radiotekh elektron [radio eng electron phys (ussr)]"] = "Radiotekhnika i Elektronika [Radio Engineering and Electronic Physics (USSR)]",
  ["radioter radiobiol fis medica"] = "Radioterapia, radiobiologia e fisica medica",
  ["radiother oncol"] = "Radiotherapy and oncology : journal of the European Society for Therapeutic Radiology and Oncology",
  ["radon ser comput appl math"] = "Radon Series on Computational and Applied Mathematics",
  ["raffles bull zool"] = "The Raffles bulletin of zoology",
  ["railway eng sci"] = "Railway Engineering Science",
  ["railway sci"] = "Railway Sciences",
  ["rairo inform théor appl"] = "RAIRO Informatique Théorique et Applications",
  ["rairo modél math anal numér"] = "RAIRO Modélisation Mathématique et Analyse Numérique",
  ["rairo oper res"] = "RAIRO - Operations Research",
  ["rairo rech opér"] = "RAIRO Recherche Opérationnelle",
  ["rairo theor inf appl"] = "RAIRO - Theoretical Informatics and Applications",
  ["rairo theor inform appl"] = "RAIRO Theoretical Informatics and Applications. Informatique Theorique et Applications",
  ["rairo theor inform appl (rairo:ita)"] = "RAIRO Theoretical Informatics and Applications (RAIRO: ITA)",
  ["rajasthan med j"] = "Rajasthan medical journal",
  ["ramanujan j"] = "The Ramanujan journal",
  ["ramanujan math soc collect works ser"] = "Ramanujan Mathematical Society Collected Works Series",
  ["ramanujan math soc lect notes ser"] = "Ramanujan Mathematical Society Lecture Notes Series",
  ["rambam maimonides med j"] = "Rambam Maimonides medical journal",
  ["rand health q"] = "Rand health quarterly",
  ["rand j econ"] = "The Rand journal of economics",
  ["random comput dynam"] = "Random & Computational Dynamics",
  ["random matrices theory appl"] = "Random Matrices. Theory and Applications",
  ["random matrices: theory appl"] = "Random Matrices: Theory and Application",
  ["random oper stoch equ"] = "Random Operators and Stochastic Equations",
  ["random oper stochastic equations"] = "Random Operators and Stochastic Equations",
  ["random struct algorithms"] = "Random structures & algorithms",
  ["random structures algorithms"] = "Random Structures & Algorithms",
  ["ranf rev"] = "RANF review",
  ["range manage agrofor"] = "Range Management and Agroforestry",
  ["rangel ecol manag"] = "Rangeland ecology & management",
  ["rangeland ecol manage"] = "Rangeland Ecology & Management",
  ["rangeland j"] = "Rangeland Journal",
  ["ranliao huaxue xuebao"] = "Ranliao Huaxue Xuebao",
  ["rap bull biol assess"] = "RAP Bulletin of Biological Assessment",
  ["rapid commun mass spectrom"] = "Rapid communications in mass spectrometry : RCM",
  ["rapid prototyp j"] = "Rapid prototyping journal",
  ["rapid prototyping j"] = "Rapid Prototyping Journal",
  ["rapp congr natl tuberc"] = "Rapports. Congrès national de la tuberculose",
  ["rapp fonct tec inst pasteur guyane fr"] = "Rapport sur le fonctionnement technique. Cayenne, French Guiana. Institut Pasteur de la Guyane française et de l'Inini",
  ["rapp stat sentralbyra"] = "Rapporter Fra Statistisk Sentralbyra",
  ["rapra rev rep"] = "RAPRA Review Reports",
  ["rare cancers ther"] = "Rare cancers and therapy",
  ["rare dis"] = "Rare diseases (Austin, Tex.)",
  ["rare met"] = "Rare Metals",
  ["rare met (beijing, china)"] = "Rare Metals (Beijing, China)",
  ["rare met mater eng"] = "Rare Metals Materials And Engineering",
  ["rare metals"] = "Rare metals",
  ["rare tumors"] = "Rare tumors",
  ["ras tech instrum"] = "RAS Techniques and Instruments",
  ["rasayan j chem"] = "Rasayan Journal of Chemistry",
  ["raspr grada povij znan"] = "Rasprave i građa za povijest znanosti",
  ["rass arch stato"] = "Rassegna degli archivi di Stato",
  ["rass clin sci"] = "Rassegna clinico-scientifica",
  ["rass clin ter"] = "La Rassegna di clinica, terapia e scienze affini",
  ["rass dermatol sifilogr"] = "Rassegna di dermatologia e di sifilografia",
  ["rass fisiopatol clin ter"] = "Rassegna di fisiopatologia clinica e terapeutica",
  ["rass giuliana med"] = "Rassegna giuliana di medicina",
  ["rass int clin ter"] = "Rassegna internazionale di clinica e terapia",
  ["rass int stomatol prat"] = "Rassegna internazionale di stomatologia pratica",
  ["rass ital chir med"] = "Rassegna italiana di chirurgia e medicina",
  ["rass ital gastroenterol"] = "Rassegna italiana di gastro-enterologia",
  ["rass ital ottalmol"] = "Rassegna italiana d'ottalmologia",
  ["rass ital sociol"] = "Rassegna italiana di sociologia",
  ["rass lett ital"] = "La Rassegna della letteratura italiana",
  ["rass med ind ig lav"] = "Rassegna di medicina industriale e di igiene del lavoro",
  ["rass med sarda"] = "Rassegna medica sarda",
  ["rass med sper"] = "Rassegna di medicina sperimentale",
  ["rass medica"] = "Rassegna medica",
  ["rass neurol veg"] = "Rassegna di neurologia vegetativa",
  ["rass neuropsichiatr"] = "Rassegna di neuropsichiatria e scienze affini",
  ["rass odontotec"] = "Rassegna odontotecnica",
  ["rass penititenziaria crim"] = "Rassegna penitenziaria e criminologica",
  ["rass stor risorgim"] = "Rassegna storica del Risorgimento : organo della Società nazionale per la storia del Risorgimento italiano",
  ["rass studi psichiatr"] = "Rassegna di studi psichiatrici",
  ["rass trimest odontoiatr"] = "Rassegna trimestrale di odontoiatria",
  ["rassegna econ"] = "Rassegna Economica",
  ["rasy nar sovrem etn ras probl"] = "Rasy i narody",
  ["ratio (ns)"] = "Ratio (New Series)",
  ["ratio (oxf)"] = "Ratio",
  ["ratio juris"] = "Ratio juris",
  ["ration drug ther"] = "Rational drug therapy",
  ["ration soc"] = "Rationality And Society",
  ["raumforsch raumord"] = "Raumforschung und Raumordnung",
  ["raumforsch raumordn"] = "Raumforschung und Raumordnung",
  ["raumplan, infh"] = "Raumplanung, Informationshefte",
  ["rber rev bus econ res"] = "RBER, Review of business and economic research",
  ["rca rev"] = "RCA Review",
  ["rcm midwives"] = "RCM midwives : the official journal of the Royal College of Midwives",
  ["rcm midwives j"] = "RCM midwives journal : official journal of the Royal College of Midwives",
  ["rcn nurs stand"] = "RCN nursing standard",
  ["re dai hai yang xue bao"] = "Re dai hai yang xue bao = Journal of tropical oceanography",
  ["re dai nong ye ke xue"] = "Re Dai Nong Ye Ke Xue",
  ["re dai yi xue za zhi"] = "Re dai yi xue za zhi = Journal of tropical medicine",
  ["re dai zuo wu xue bao"] = "Re dai zuo wu xue bao = Chinese journal of tropical crops",
  ["re:gen open"] = "Re:GEN Open",
  ["reach out"] = "Reaching out",
  ["reach rev hum space explor"] = "REACH - Reviews in Human Space Exploration",
  ["react chem eng"] = "Reaction Chemistry & Engineering",
  ["react funct polym"] = "Reactive and Functional Polymers",
  ["react kinet catal lett"] = "Reaction Kinetics and Catalysis Letters",
  ["react kinet mech catal"] = "Reaction Kinetics Mechanisms and Catalysis",
  ["react kinet, mech catal"] = "Reaction Kinetics, Mechanisms and Catalysis",
  ["react oxyg species (apex)"] = "Reactive oxygen species (Apex, N.C.)",
  ["react polym"] = "Reactive Polymers",
  ["react solids"] = "Reactivity of Solids",
  ["react wkly"] = "Reactions Weekly",
  ["read dig"] = "The Reader's digest",
  ["read math"] = "Readings in Mathematics",
  ["read philos"] = "Readings in Philosophy",
  ["read psychol"] = "Reading psychology",
  ["read res instr"] = "Reading research and instruction : the journal of the College Reading Association",
  ["read res q"] = "Reading research quarterly",
  ["read teach"] = "The Reading teacher",
  ["read writ"] = "Reading and writing",
  ["read writ q"] = "Reading & writing quarterly : overcoming learning difficulties",
  ["real acad farm barcelona sesion inaug curso"] = "Real Academia de Farmacia de Barcelona. Sesión inaugural del curso",
  ["real anal exch"] = "Real Analysis Exchange",
  ["real anal exchange"] = "Real Analysis Exchange",
  ["real clin"] = "Réalités cliniques : revue européenne d'odontologie",
  ["real datos espacio"] = "Realidad, datos y espacio : revista internacional de estadística y geografía",
  ["real estate econ"] = "Real Estate Economics",
  ["real property probate trust j"] = "Real property, probate, and trust journal",
  ["real-time syst"] = "Real-Time Systems",
  ["reanim organes artif"] = "Réanimation et organes artificiels. Wiederbelebung und künstliche Organe. Reanimation and artificial organs",
  ["reason pap"] = "Reason papers",
  ["rec assoc bar city new york"] = "The Record of the Association of the Bar of the City of New York",
  ["rec aust mus"] = "Records of the Australian Museum",
  ["rec bucks"] = "Records of Buckinghamshire, or, Papers and notes on the history, antiquities, and architecture of the county, together with the proceedings of the Architectural and Archaeological Society for the County of Buckingham",
  ["rec canterb mus"] = "Records of the Canterbury Museum",
  ["rec chem prog"] = "Record of chemical progress",
  ["rec columb hist soc"] = "Records of the Columbia Historical Society, Washington, D.C",
  ["rec columbia med soc richland cty sc"] = "The Recorder of the Columbia Medical Society of Richland County, S. C",
  ["rec med vet ec alfort"] = "Recueil de médecine vétérinaire",
  ["rec nat prod"] = "Records of Natural Products",
  ["rec res dev immunol"] = "Recent research developments in immunology",
  ["rec west aust mus"] = "Records of the Western Australian Museum. Western Australian Museum",
  ["rec zool surv india"] = "Records of the Zoological Survey of India",
  ["recens geral popul"] = "Recenseamento geral da populaçao. Instituto Nacional de Estatística (Portugal)",
  ["recent adv biol psychiatry"] = "Recent advances in biological psychiatry",
  ["recent adv cardiovasc drug discov"] = "Recent advances in cardiovascular drug discovery",
  ["recent adv cardiovasc drug discovery"] = "Recent Advances in Cardiovascular Drug Discovery",
  ["recent adv clin nucl med"] = "Recent advances in clinical nuclear medicine",
  ["recent adv comput chem"] = "Recent Advances in Computational Chemistry",
  ["recent adv dna gene seq"] = "Recent advances in DNA & gene sequences",
  ["recent adv dna gene sequences"] = "Recent Advances in DNA & Gene Sequences",
  ["recent adv drug deliv formul"] = "Recent advances in drug delivery and formulation",
  ["recent adv electr electron eng"] = "Recent Advances in Electrical and Electronic Engineering",
  ["recent adv nurs"] = "Recent advances in nursing",
  ["recent adv phytochem"] = "Recent advances in phytochemistry",
  ["recent adv res updat"] = "Recent advances & research updates",
  ["recent adv stud cardiac struct metab"] = "Recent advances in studies on cardiac structure and metabolism",
  ["recent dev alcohol"] = "Recent developments in alcoholism : an official publication of the American Medical Society on Alcoholism, the Research Society on Alcoholism, and the National Council on Alcoholism",
  ["recent innovations chem eng"] = "Recent Innovations in Chemical Engineering",
  ["recent pat anti-cancer drug discovery"] = "Recent Patents on Anti-Cancer Drug Discovery",
  ["recent pat anti-infect drug discovery"] = "Recent Patents on Anti-Infective Drug Discovery",
  ["recent pat anticancer drug discov"] = "Recent patents on anti-cancer drug discovery",
  ["recent pat antiinfect drug discov"] = "Recent patents on anti-infective drug discovery",
  ["recent pat biomark"] = "Recent patents on biomarkers",
  ["recent pat biomed eng"] = "Recent patents on biomedical engineering",
  ["recent pat biotechnol"] = "Recent patents on biotechnology",
  ["recent pat cardiovasc drug discov"] = "Recent patents on cardiovascular drug discovery",
  ["recent pat cns drug discov"] = "Recent patents on CNS drug discovery",
  ["recent pat cns drug discovery"] = "Recent Patents on CNS Drug Discovery",
  ["recent pat dna gene seq"] = "Recent patents on DNA & gene sequences",
  ["recent pat dna gene sequences"] = "Recent Patents on DNA & Gene Sequences",
  ["recent pat drug deliv formul"] = "Recent patents on drug delivery & formulation",
  ["recent pat drug delivery formulation"] = "Recent Patents on Drug Delivery & Formulation",
  ["recent pat endocr metab immune drug discov"] = "Recent patents on endocrine, metabolic & immune drug discovery",
  ["recent pat endocr, metab immune drug discovery"] = "Recent Patents on Endocrine, Metabolic & Immune Drug Discovery",
  ["recent pat eng"] = "Recent Patents on Engineering",
  ["recent pat food nutr agric"] = "Recent patents on food, nutrition & agriculture",
  ["recent pat inflamm allergy drug discov"] = "Recent patents on inflammation & allergy drug discovery",
  ["recent pat inflammation allergy drug discovery"] = "Recent Patents on Inflammation & Allergy Drug Discovery",
  ["recent pat mech eng"] = "Recent Patents on Mechanical Engineering",
  ["recent pat med imaging"] = "Recent patents on medical imaging",
  ["recent pat nanomed"] = "Recent patents on nanomedicine",
  ["recent pat nanotechnol"] = "Recent patents on nanotechnology",
  ["recent pat regen med"] = "Recent patents on regenerative medicine",
  ["recent pat space technol"] = "Recent patents on space technology",
  ["recent patents anticancer drug discov"] = "Recent patents on anti-cancer drug discovery",
  ["recent patents cardiovasc drug discov"] = "Recent patents on cardiovascular drug discovery",
  ["recent prog horm res"] = "Recent progress in hormone research",
  ["recent prog mater"] = "Recent progress in materials",
  ["recent progr psychiatr"] = "Recent progress in psychiatry",
  ["recent publ articles"] = "Recently published articles",
  ["recent res dev neurosci"] = "Recent research developments in neuroscience",
  ["recent res dev org chem"] = "Recent research developments in organic chemistry",
  ["recent res dev virol"] = "Recent research developments in virology",
  ["recent results cancer res"] = "Recent results in cancer research. Fortschritte der Krebsforschung. Progrès dans les recherches sur le cancer",
  ["recenti prog med"] = "Recenti progressi in medicina",
  ["recept channels"] = "Receptors & channels",
  ["recept signal transduct"] = "Receptors & signal transduction",
  ["receptors channels"] = "Receptors and Channels",
  ["receptors clin investig"] = "Receptors & clinical investigation",
  ["rech amerindien que"] = "Recherches amérindiennes au Québec",
  ["rech econ louvain"] = "Recherches Economiques de Louvain",
  ["rech educ"] = "Recherches & éducations",
  ["rech fem"] = "Recherches féministes",
  ["rech geogr strasbg"] = "Recherches géographiques à Strasbourg",
  ["rech math appl"] = "Recherches en Mathematiques Appliquees",
  ["rech philos lang"] = "Recherches sur la Philosophie et le Langage",
  ["rech soc"] = "Recherche sociale",
  ["rech sociogr"] = "Recherches sociographiques",
  ["rech sociol anthropol"] = "Recherches sociologiques et anthropologiques : RS & A",
  ["rech soins infirm"] = "Recherche en soins infirmiers",
  ["recher fed groupes etud recher inst"] = "Recherches",
  ["rechtsmedizin (berl)"] = "Rechtsmedizin (Berlin, Germany)",
  ["recl trav chim pays bas"] = "Recueil des travaux chimiques des Pays-Bas (Leiden, Netherlands : 1920)",
  ["recl trav chim pays-bas"] = "Recueil des Travaux Chimiques des Pays-Bas",
  ["recomb dna tech bull"] = "Recombinant DNA technical bulletin",
  ["reconstr surg traumatol"] = "Reconstruction surgery and traumatology",
  ["record (washington)"] = "Record",
  ["recovery j"] = "Recovery journal",
  ["recreat math mag"] = "Recreational Mathematics Magazine",
  ["recreat park tour public health"] = "Recreation, parks, and tourism in public health",
  ["recruit retain"] = "Recruit & retain : a monthly publication for professionals concerned with nurse recruitment and retention",
  ["recruit retent rep"] = "Recruitment & retention report",
  ["recruit retent restruct rep"] = "Recruitment, retention & restructuring report",
  ["recusant hist"] = "Recusant history",
  ["red cross cour"] = "The Red Cross courier",
  ["red river val hist rev"] = "Red River Valley historical review",
  ["redia (firenze)"] = "Redia (Firenze)",
  ["redox biol"] = "Redox biology",
  ["redox exp biol"] = "Redox Experimental Biology",
  ["redox rep"] = "Redox report : communications in free radical research",
  ["reeduc orthoph"] = "Rééducation orthophonique",
  ["reese rec"] = "The Reese record",
  ["ref guide health care technol ind"] = "Reference guide for the health care technology industry",
  ["ref gynecol obstet"] = "Références en gynécologie obstétrique",
  ["ref libr"] = "The Reference librarian",
  ["ref sci"] = "References Sciences",
  ["ref serv rev"] = "RSR. Reference services review",
  ["ref source sci technol"] = "Reference Sources in Science and Technology",
  ["ref sources social sci humanities"] = "Reference Sources for the Social Sciences and Humanities",
  ["refereed ejournal"] = "Refereed e-journal",
  ["reflect nurs leadersh"] = "Reflections on nursing leadership",
  ["reflections (long beach)"] = "Reflections (Long Beach, Calif.)",
  ["reflets phys"] = "Reflets de la Physique",
  ["reform j"] = "The Reformed journal",
  ["reforma med"] = "La Reforma médica",
  ["refract corneal surg"] = "Refractive & corneal surgery",
  ["refract ind ceram"] = "Refractories and Industrial Ceramics",
  ["refresh courses anesthesiol"] = "Refresher courses in anesthesiology",
  ["refrig sci technol"] = "Refrigeration Science and Technology",
  ["refu hashinayim"] = "Refuat ha-Shinayim",
  ["refu vet"] = "Refuʼah ṿeṭerinarit : rivʻon Histadrut ha-rofʼim ha-ṿeṭerinariyim be-Erets-Yiśraʼel",
  ["refuat hapeh vehashinayim"] = "Refuʼat ha-peh ṿeha-shinayim (Tel Aviv, Israel : 1969)",
  ["refuat hapeh vehashinayim (1993)"] = "Refuʼat ha-peh ṿeha-shinayim (1993)",
  ["refuat hashinayim"] = "Refuʼat ha-shinayim (Tel Aviv, Israel : 1944)",
  ["refug rep"] = "Refugee reports",
  ["reg anaesth"] = "Regional-Anaesthesie",
  ["reg anesth"] = "Regional anesthesia",
  ["reg anesth pain med"] = "Regional anesthesia and pain medicine",
  ["reg basil"] = "Regio Basiliensis",
  ["reg cohes"] = "Regions & cohesion : the journal of the Consortium for Comparative Research on Regional Integration and Social Cohesion",
  ["reg dev dialogue"] = "Regional development dialogue",
  ["reg environ change"] = "Regional environmental change",
  ["reg immunol"] = "Regional immunology",
  ["reg sci persp"] = "Regional Science Perspectives",
  ["reg sci perspect"] = "Regional science perspectives",
  ["reg sci urban econ"] = "Regional science and urban economics",
  ["reg stud"] = "Regional studies",
  ["reg stud mar sci"] = "Regional studies in marine science",
  ["reg stud reg sci"] = "Regional studies, regional science",
  ["reg sustainability"] = "Regional Sustainability",
  ["regan rep nurs law"] = "The Regan report on nursing law",
  ["regards actual"] = "Regards sur l'actualité",
  ["regen biomater"] = "Regenerative biomaterials",
  ["regen eng transl med"] = "Regenerative engineering and translational medicine",
  ["regen med"] = "Regenerative medicine",
  ["regen med front"] = "Regenerative medicine frontiers",
  ["regen med res"] = "Regenerative medicine research",
  ["regen ther"] = "Regenerative therapy",
  ["regener biomater"] = "Regenerative Biomaterials",
  ["regener eng transl med"] = "Regenerative Engineering and Translational Medicine",
  ["regener med"] = "Regenerative Medicine",
  ["regener med res"] = "Regenerative Medicine Research",
  ["regeneration (oxf)"] = "Regeneration (Oxford, England)",
  ["regenerative med"] = "Regenerative medicine",
  ["regensb jahrb arztl fortbild"] = "Regensburger Jahrbuch für ärztliche Fortbildung",
  ["regensburger math schriften"] = "Regensburger Mathematische Schriften",
  ["regist ky hist soc"] = "The Register of the Kentucky Historical Society",
  ["regist nurse"] = "Registered nurse (Toronto, Ont.)",
  ["regnum veg"] = "Regnum vegetabile",
  ["regul aff j"] = "The regulatory affairs journal",
  ["regul anal med waste"] = "Regulatory analyst. Medical waste",
  ["regul chaotic dyn"] = "Regular and Chaotic Dynamics",
  ["regul gov"] = "Regulation & governance",
  ["regul khaoticheskaya din"] = "Regulyarnaya & Khaoticheskaya Dinamika",
  ["regul mech biosyst"] = "Regulatory mechanisms in biosystems",
  ["regul pept"] = "Regulatory Peptides",
  ["regul pept suppl"] = "Regulatory peptides. Supplement",
  ["regul rivers: res manage"] = "Regulated Rivers: Research & Management",
  ["regul toxicol pharm"] = "Regulatory Toxicology and Pharmacology",
  ["regul toxicol pharmacol"] = "Regulatory toxicology and pharmacology : RTP",
  ["rehab manag"] = "Rehab management",
  ["rehabil couns bull"] = "Rehabilitation counseling bulletin",
  ["rehabil couns educ j"] = "Rehabilitation counselors and educators journal",
  ["rehabil fyz lek"] = "Rehabilitace a fyzikální́́ lékařství",
  ["rehabil lit"] = "Rehabilitation literature",
  ["rehabil nurs"] = "Rehabilitation nursing : the official journal of the Association of Rehabilitation Nurses",
  ["rehabil oncol"] = "Rehabilitation oncology (American Physical Therapy Association. Oncology Section)",
  ["rehabil psychol"] = "Rehabilitation psychology",
  ["rehabil rec"] = "Rehabilitation record",
  ["rehabil res pract"] = "Rehabilitation research and practice",
  ["rehabilitacion (madr)"] = "Rehabilitación",
  ["rehabilitation (bonn)"] = "Rehabilitation: Sozialmedizin, physikalische Medizin, Präventivmedizin; internationale Zeitschrift mit Zentralblatt",
  ["rehabilitation (stuttg)"] = "Die Rehabilitation",
  ["rehabilitation (uelzen)"] = "Rehabilitation; internationale Zeitschrift für physikalische Medizin und Rehabilitation mit Zentralblatt",
  ["reihe med forsch"] = "Reihe, Medizinische Forschung",
  ["reihe wiss"] = "Reihe Wissenschaft",
  ["reihe: planung, inform unternehmungsfuhr"] = "Reihe: Planung, Information und Unternehmungsfuhrung",
  ["rein foie"] = "Rein et foie, maladies de la nutrition; actualités",
  ["reinf plast"] = "Reinforced Plastics",
  ["reinf plast compos world"] = "Reinforced Plastics and Composites World",
  ["rejuvenation res"] = "Rejuvenation research",
  ["rekishigaku kenkyu"] = "Rekishigaku kenkyū",
  ["relac int"] = "Relaciones internacionales",
  ["relaciones (zamora)"] = "Relaciones (Colegio de Michoacán)",
  ["reliab comput"] = "Reliable Computing",
  ["reliab eng"] = "Reliability Engineering",
  ["reliab eng syst saf"] = "Reliability Engineering & System Safety",
  ["relig educ"] = "Religious education (Chicago, Ill.)",
  ["relig gend"] = "Religion & gender",
  ["relig humanism"] = "Religious humanism",
  ["relig life"] = "Religion in life (New York, N.Y.)",
  ["relig state soc"] = "Religion, state & society : the Keston journal",
  ["relig stud rev"] = "Religious studies review",
  ["religion brain behav"] = "Religion, brain & behavior",
  ["religions (basel)"] = "Religions",
  ["rem actual"] = "Remèdes-Actualités",
  ["rem j"] = "Remediation Journal",
  ["rem open access"] = "Remedy open access",
  ["remaking econ eminent post-war econ"] = "Remaking Economics: Eminent Post-War Economists",
  ["remedial spec educ"] = "Remedial and special education : RASE",
  ["remediation (n y)"] = "Remediation (New York, N.Y.)",
  ["remote sens"] = "Remote Sensing",
  ["remote sens (basel)"] = "Remote sensing",
  ["remote sens appl"] = "Remote sensing applications : society and environment",
  ["remote sens appl: soc environ"] = "Remote Sensing Applications: Society and Environment",
  ["remote sens earth syst sci"] = "Remote sensing in earth systems sciences",
  ["remote sens ecol conserv"] = "Remote sensing in ecology and conservation",
  ["remote sens environ"] = "Remote Sensing of Environment",
  ["remote sens lett"] = "Remote sensing letters (Print)",
  ["remote sens ser"] = "Remote Sensing Series",
  ["ren fail"] = "Renal failure",
  ["ren kou dong tai"] = "Ren kou dong tai",
  ["ren kou xue kan"] = "Ren kou xue kan (Changchun shi, China)",
  ["ren min ri bao"] = "Jen min jih pao",
  ["ren physiol"] = "Renal physiology",
  ["ren physiol biochem"] = "Renal physiology and biochemistry",
  ["ren replace ther"] = "Renal replacement therapy",
  ["ren wen ji she hui ke xue ji kan"] = "Ren wen ji she hui ke xue ji kan",
  ["renaiss drama"] = "Renaissance drama (Evanston, Ill. : 1964)",
  ["renaiss mod stud"] = "Renaissance and modern studies",
  ["renaiss q"] = "Renaissance quarterly",
  ["renaiss reform"] = "Renaissance and Reformation",
  ["renaiss stud"] = "Renaissance studies : journal of the Society for Renaissance Studies",
  ["renal nutr forum"] = "Renal nutrition forum",
  ["renal physiol"] = "Renal Physiology",
  ["rend accad naz sci xl mem mat appl (5)"] = "Accademia Nazionale delle Scienze detta dei XL",
  ["rend accad naz sci xl mem sci fis natur (5)"] = "Rendiconti della Accademia Nazionale delle Scienze detta dei XL",
  ["rend accad naz xl"] = "Rendiconti / Accademia Nazionale Del Xl",
  ["rend accad sci fis mat napoli (4)"] = "Societa Nazionale di Scienze, Lettere e Arti in Napoli. Rendiconto dell’Accademia delle Scienze Fisiche e Matematiche. Serie IV",
  ["rend circ mat palermo"] = "Rendiconti del Circolo Matematico di Palermo",
  ["rend circ mat palermo (2)"] = "Rendiconti del Circolo Matematico di Palermo. Second Series",
  ["rend circ mat palermo (2) suppl"] = "Rendiconti del Circolo Matematico di Palermo",
  ["rend ist lombardo accad sci lett sez"] = "Rendiconti. B, Scienze biologiche e mediche",
  ["rend ist sup sanit"] = "Rendiconti - Istituto superiore di sanità",
  ["rend istit mat univ trieste"] = "Rendiconti dell’Istituto di Matematica dell’Universita di Trieste. An International Journal of Mathematics",
  ["rend lincei sci fis nat"] = "Rendiconti Lincei. Scienze fisiche e naturali",
  ["rend lincei-mat appl"] = "Rendiconti Lincei-Matematica e Applicazioni",
  ["rend mat appl (7)"] = "Rendiconti di Matematica e delle sue Applicazioni. Serie VII",
  ["rend online soc geol ital"] = "Rendiconti Online Societa Geologica Italiana",
  ["rend sem fac sci univ cagliari"] = "Rendiconti del Seminario della Facoltà di Scienze dell’Università di Cagliari",
  ["rend sem mat fis milano"] = "Rendiconti del Seminario Matematico e Fisico di Milano",
  ["rend sem mat messina ser ii"] = "Rendiconti del Seminario Matematico di Messina",
  ["rend sem mat univ padova"] = "Rendiconti del Seminario Matematico della Università di Padova",
  ["rend sem mat univ politec torino"] = "Università e Politecnico di Torino",
  ["rend semin mat univ padova"] = "Rendiconti del Seminario Matematico della Universita di Padova",
  ["rend semin mat univ politec torino"] = "Rendiconti del Seminario Matematico. Universita e Politecnico Torino",
  ["rendezvous (buffalo)"] = "Rendezvous",
  ["renew energy"] = "Renewable energy",
  ["renew sustain energy rev"] = "Renewable & Sustainable Energy Reviews",
  ["renew wind water sol"] = "Renewables: wind, water, and solar",
  ["renewable agric food syst"] = "Renewable Agriculture and Food Systems",
  ["renewable energy"] = "Renewable Energy",
  ["renewable energy environ sustainability"] = "Renewable Energy and Environmental Sustainability",
  ["renewable energy focus"] = "Renewable Energy Focus",
  ["renewable energy power qual j"] = "Renewable Energy and Power Quality Journal",
  ["renewable resour j"] = "Renewable Resources Journal",
  ["renewable sustainable energy rev"] = "Renewable and Sustainable Energy Reviews",
  ["renewable sustainable energy transition"] = "Renewable and Sustainable Energy Transition",
  ["renewables: wind, water, solar"] = "Renewables: Wind, Water, and Solar",
  ["rengong jingti xuebao"] = "Rengong Jingti Xuebao",
  ["renkou yanjiu"] = "Ren kou yan jiu = Renkou yanjiu",
  ["renkou yu jingji"] = "Ren kou yu jing ji = Renkou yu jingji",
  ["rep adv phys sci"] = "Reports in Advances of Physical Sciences",
  ["rep am univ field staff"] = "Reports - American Universities Field Staff. American Universities Field Staff",
  ["rep assoc med serv inc hannah inst hist med"] = "Report. Associated Medical Services, Inc",
  ["rep biochem mol biol"] = "Reports of biochemistry & molecular biology",
  ["rep can def res chem lab"] = "Report. Canada. Defence Research Chemical Laboratories",
  ["rep can def res kings lab"] = "Report. Canada. Defence Research Kingston Laboratory",
  ["rep carcinog"] = "Report on carcinogens : carcinogen profiles",
  ["rep carcinog backgr doc"] = "Report on carcinogens background document for [substance name]",
  ["rep cases argued determ super court appell div chancery div law div cty courts state n j n j super court"] = "Reports of cases argued and determined in the Superior Court, Appellate Division, Chancery Division, Law Division, and in the county courts of the state of New Jersey. New Jersey. Superior Court",
  ["rep cases argued determ supreme court n j n j supreme court"] = "Reports of cases argued and determined in the Supreme Court of New Jersey. New Jersey. Supreme Court",
  ["rep cases decided appell div supreme court state ny ny state supreme court appell div"] = "Reports of cases decided in the Appellate Division of the Supreme Court, State of New York. New York (State). Supreme Court. Appellate Division",
  ["rep cases determ court appeal state calif"] = "Reports of cases determined in the courts of appeal of the state of California",
  ["rep civ aeromed res inst us"] = "[Report]. Civil Aeromedical Research Institute (U.S.)",
  ["rep comm accredit rehabil facil"] = "Report - Commission on Accreditation of Rehabilitation Facilities",
  ["rep congr (ser)"] = "Report to Congress (United States. Physician Payment Review Commission : Series)",
  ["rep congr eur orthod soc"] = "Report of the congress. European Orthodontic Society",
  ["rep congr physician paym rev comm"] = "Report to Congress. United States. Physician Payment Review Commission",
  ["rep cucurbit genet coop"] = "Report. Cucurbit Genetics Cooperative",
  ["rep del nurses assoc"] = "Reporter (Wilmington, Del.)",
  ["rep dep health scotl"] = "Report. Great Britain. Department of Health for Scotland",
  ["rep emot behav disord youth"] = "Report on emotional & behavioral disorders in youth",
  ["rep enlarged sess semin i vekua appl math"] = "Reports of Enlarged Sessions of Seminar of I. Vekua Institute of Applied Mathematics",
  ["rep fac eng kanagawa univ"] = "Kanagawa University",
  ["rep fac eng oita univ"] = "Oita University",
  ["rep fac sci eng saga univ math"] = "Reports of the Faculty of Science and Engineering",
  ["rep fac sci kagoshima univ"] = "Reports of the Faculty of Science",
  ["rep fac sci kagoshima univ math phys chem"] = "Reports of the Faculty of Science",
  ["rep fac sci shizuoka univ"] = "Reports of Faculty of Science, Shizuoka University",
  ["rep fac sci technol meijo univ"] = "Reports of the Faculty of Science and Technology",
  ["rep group adv psychiatry"] = "Report (Group for the Advancement of Psychiatry)",
  ["rep health care"] = "Report of health care",
  ["rep health soc subj (lond)"] = "Reports on health and social subjects",
  ["rep inst hum values"] = "Report - Institute on Human Values in Medicine",
  ["rep inst math"] = "Reports of the Institute of Mathematics",
  ["rep inst philos public policy"] = "Report from the Institute for Philosophy & Public Policy",
  ["rep int dev res cent can"] = "Reports (International Development Research Centre (Canada))",
  ["rep list memb fac hist phil med pharm worshipf soc apoth london"] = "Report and list of members. Worshipful Society of Apothecaries of London. Faculty of the History and Philosophy of Medicine and Pharmacy",
  ["rep math logic"] = "Reports on Mathematical Logic",
  ["rep math phys"] = "Reports on Mathematical Physics",
  ["rep mech 1983"] = "Report MECH 1983",
  ["rep med guidel outcomes res"] = "Report on medical guidelines & outcomes research",
  ["rep med imaging"] = "Reports in Medical Imaging",
  ["rep n y state vet coll cornell univ"] = "Report of the New York State Veterinary College at Cornell University. New York State Veterinary College",
  ["rep nav med neuropsychiatr res unit"] = "Report - Navy Medical Neuropsychiatric Research Unit",
  ["rep no naec acel united states aerosp crew equip lab phila"] = "[Report no.] NAEC-ACEL-. United States. Aerospace Crew Equipment Laboratory, Philadelphia",
  ["rep no namc acel united states air crew equip lab phila"] = "[Report no.] NAMC-ACEL-. United States. Air Crew Equipment Laboratory, Philadelphia",
  ["rep popul fam plann"] = "Reports on population/family planning",
  ["rep pract oncol radiother"] = "Reports of practical oncology and radiotherapy : journal of Greatpoland Cancer Center in Poznań and Polish Society of Radiation Oncology",
  ["rep proc scott soc hist med"] = "Report of proceedings. Scottish Society of the History of Medicine",
  ["rep prog phys"] = "Reports on progress in physics. Physical Society (Great Britain)",
  ["rep progr phys"] = "Reports on Progress in Physics",
  ["rep public health med subj (lond)"] = "Reports on public health and medical subjects",
  ["rep rheum dis"] = "Reports on rheumatic diseases",
  ["rep scm"] = "Reports SCM",
  ["rep sel cases decided courts state n y court appeals appell div supreme court"] = "Reports of selected cases decided in courts of the State of New York other than the Court of Appeals and the Appellate Division of the Supreme Court : miscellaneous reports",
  ["rep ser theoret phys"] = "Report Series in Theoretical Physics",
  ["rep soc norms"] = "Report on social norms",
  ["rep statist appl res un japan sci engrs"] = "Reports of Statistical Application Research",
  ["rep tomato genet coop tomato genet coop"] = "Report of the Tomato Genetics Cooperative. Tomato Genetics Cooperative",
  ["rep trans devonshire assoc"] = "Report & transactions of the Devonshire Association for the Advancement of Science, Literature and Art. Devonshire Association for the Advancement of Science, Literature and Art",
  ["rep u s nav med res lab"] = "Report (U.S. Naval Medical Research Laboratory)",
  ["rep us army med res lab"] = "Report. Army Medical Research Laboratory (U.S.)",
  ["rep us army med res nutr lab denver"] = "Report. U.S. Army Medical Research and Nutrition Laboratory",
  ["rep us nav med res lab"] = "Report - U. S. Naval Medical Research Laboratory",
  ["rep us nav submar med cent"] = "Report (U.S. Naval Submarine Medical Center)",
  ["rep us navy exp diving unit"] = "Report. United States. Navy Experimental Diving Unit, Washington, D. C",
  ["report hum reprod law"] = "Reporter on human reproduction and the law",
  ["repr theory appl categ"] = "Reprints in Theory and Applications of Categories",
  ["represent mind"] = "Representation and Mind",
  ["represent res soc psychol"] = "Representative research in social psychology",
  ["represent theory"] = "Representation Theory. An Electronic Journal of the American Mathematical Society",
  ["representation (mcdougall trust)"] = "Representation (McDougall Trust)",
  ["representations (berkeley)"] = "Representations (Berkeley, Calif.)",
  ["reprod abstr ser"] = "Reproduction (Cambridge, England). Abstract series",
  ["reprod biol"] = "Reproductive biology",
  ["reprod biol endocrinol"] = "Reproductive biology and endocrinology : RB&E",
  ["reprod biomed online"] = "Reproductive biomedicine online",
  ["reprod biomed soc online"] = "Reproductive biomedicine & society online",
  ["reprod contracept"] = "Reproduction and contraception",
  ["reprod domest anim"] = "Reproduction in domestic animals = Zuchthygiene",
  ["reprod fertil dev"] = "Reproduction, fertility, and development",
  ["reprod freedom news"] = "Reproductive freedom news",
  ["reprod genet eng"] = "Reproductive and genetic engineering",
  ["reprod health"] = "Reproductive health",
  ["reprod health matters"] = "Reproductive health matters",
  ["reprod med (basel)"] = "Reproductive medicine (Basel, Switzerland)",
  ["reprod med biol"] = "Reproductive medicine and biology",
  ["reprod nutr dev"] = "Reproduction, nutrition, development",
  ["reprod sci"] = "Reproductive sciences (Thousand Oaks, Calif.)",
  ["reprod suppl"] = "Reproduction (Cambridge, England) Supplement",
  ["reprod syst sex disord"] = "Reproductive system & sexual disorders : current research",
  ["reprod toxicol"] = "Reproductive toxicology (Elmsford, N.Y.)",
  ["reproduction (bristol, u k)"] = "Reproduction (Bristol, United Kingdom)",
  ["reprowatch (youth ed)"] = "ReproWatch (Youth edition)",
  ["reptile rap"] = "Reptile rap",
  ["reptiles amphib"] = "Reptiles & amphibians : conservation and natural history",
  ["repúb venezuela bol acad cienc fís mat natur"] = "República de Venezuela",
  ["requir eng"] = "Requirements engineering",
  ["res act fac sci eng tokyo denki univ"] = "Research Activities",
  ["res adv psychiatry"] = "Research and advances in psychiatry",
  ["res agenda brief"] = "Research agenda brief",
  ["res aging"] = "Research on aging",
  ["res agric agron"] = "Research in agriculture and agronomy",
  ["res astron astrophys"] = "Research in Astronomy and Astrophysics",
  ["res autism spectr disord"] = "Research in autism spectrum disorders",
  ["res biol"] = "Research in biology",
  ["res biomed eng"] = "Research on Biomedical Engineering",
  ["res biotechnol"] = "Research in biotechnology",
  ["res brief"] = "Research brief",
  ["res bull (sun chiwawitthaya thang thale phuket)"] = "Research bulletin (Sūn Chīwawitthayā thāng Thalē Phūket)",
  ["res bull coll exp for, hokkaido univ"] = "Research Bulletins of the College Experiment Forests, Hokkaido University",
  ["res bull fac ed oita univ"] = "The Research Bulletin of the Faculty of Education",
  ["res bull panjab univ sci"] = "Research Bulletin of the Panjab University",
  ["res bull u s"] = "Research bulletin. United States. Army Air Forces. Training Command. Psychological Section",
  ["res cardiovasc med"] = "Research in cardiovascular medicine",
  ["res chem intermed"] = "Research on Chemical Intermediates",
  ["res child adolesc psychopathol"] = "Research on child and adolescent psychopathology",
  ["res clin stud headache"] = "Research and clinical studies in headache",
  ["res cold arid reg"] = "Research in Cold and Arid Regions",
  ["res commun biochem cell mol biol"] = "Research communications in biochemistry and cell & molecular biology",
  ["res commun chem pathol pharmacol"] = "Research communications in chemical pathology and pharmacology",
  ["res commun inst ferment"] = "Research communications - Institute for Fermentation, Osaka",
  ["res commun mol pathol pharmacol"] = "Research Communications in Molecular Pathology and Pharmacology",
  ["res commun pharmacol toxicol"] = "Research Communications in Pharmacology and Toxicology",
  ["res commun psychol psychiatr behav"] = "Research communications in psychology, psychiatry and behavior",
  ["res commun subst abuse"] = "Research communications in substance abuse",
  ["res community ment health"] = "Research in community and mental health",
  ["res comp int educ"] = "Research in comparative and international education",
  ["res dev disabil"] = "Research in developmental disabilities",
  ["res dev tech rep"] = "Research and development technical report. United States. Naval Radiological Defense Laboratory, San Francisco",
  ["res diagn interventional imaging"] = "Research in Diagnostic and Interventional Imaging",
  ["res discl"] = "Research Disclosure",
  ["res drama educ"] = "Research in drama education",
  ["res econ"] = "Research in economics = Ricerche economiche",
  ["res econ anthropol"] = "Research in economic anthropology",
  ["res econ hist"] = "Research in economic history",
  ["res econ inequal"] = "Research on economic inequality",
  ["res eng des"] = "Research in Engineering Design",
  ["res eng struct mater"] = "Research on Engineering Structures and Materials",
  ["res ethics"] = "Research ethics",
  ["res eval"] = "Research Evaluation",
  ["res exp math"] = "Research and Exposition in Mathematics",
  ["res exp med"] = "Research In Experimental Medicine",
  ["res exp med (berl)"] = "Research in experimental medicine. Zeitschrift für die gesamte experimentelle Medizin einschliesslich experimenteller Chirurgie",
  ["res file"] = "The research file",
  ["res front fertil regul"] = "Research frontiers in fertility regulation : RFFR",
  ["res gerontol nurs"] = "Research in gerontological nursing",
  ["res health econ"] = "Research in health economics",
  ["res health sci"] = "Research in health science",
  ["res high educ"] = "Research in higher education",
  ["res high educ j"] = "Research in higher education journal",
  ["res hum cap dev"] = "Research in human capital and development",
  ["res hum dev"] = "Research in human development",
  ["res ideas outcomes"] = "Research ideas and outcomes",
  ["res immunol"] = "Research in immunology",
  ["res initiat treat action"] = "Research initiative, treatment action : RITA",
  ["res int bus finance"] = "Research in international business and finance",
  ["res integr peer rev"] = "Research integrity and peer review",
  ["res integrity peer rev"] = "Research Integrity and Peer Review",
  ["res intell manuf (rims)"] = "Research on Intelligent Manufacturing (RIMS)",
  ["res involv engagem"] = "Research involvement and engagement",
  ["res ipsa loquitur"] = "Res ipsa loquitur (Washington, D.C.)",
  ["res j agric biol sci"] = "Research journal of agriculture and biological sciences",
  ["res j appl sci"] = "Research journal of applied sciences",
  ["res j biotechnol"] = "Research journal of biotechnology",
  ["res j chem environ"] = "Research journal of chemistry and environment",
  ["res j chem environ sci"] = "Research journal of chemical and environmental sciences",
  ["res j environ sci"] = "Research Journal of Environmental Sciences",
  ["res j health sci"] = "Research journal of health sciences",
  ["res j life sci bioinform pharm chem sci"] = "Research journal of life sciences, bioinformatics, pharmaceutical and chemical sciences",
  ["res j med sci"] = "Research journal of medical sciences",
  ["res j microbiol"] = "Research journal of microbiology",
  ["res j pharm biol chem sci"] = "Research journal of pharmaceutical, biological and chemical sciences",
  ["res j pharm technol"] = "Research journal of pharmacy and technology",
  ["res j recent sci"] = "Research journal of recent sciences",
  ["res j text apparel"] = "Research Journal of Textile and Apparel",
  ["res j vet sci"] = "Research journal of veterinary sciences",
  ["res j womens health"] = "Research journal of women's health",
  ["res knowl"] = "Research & knowledge",
  ["res lang soc interact"] = "Research on language and social interaction",
  ["res law econ"] = "Research in Law and Economics",
  ["res lett biochem"] = "Research letters in biochemistry",
  ["res lett signal process"] = "Research letters in signal processing",
  ["res life sci"] = "Research in Life Sciences",
  ["res manag rev"] = "Research management review",
  ["res manage"] = "Research management",
  ["res math"] = "Research in Mathematics",
  ["res math educ"] = "Research in Mathematics Education",
  ["res math sci"] = "Research in the Mathematical Sciences",
  ["res meas approv"] = "Research—Measurement—Approval",
  ["res methods med health sci"] = "Research Methods in Medicine & Health Sciences",
  ["res microbiol"] = "Research in microbiology",
  ["res newsl"] = "Research newsletter. College of General Practitioners (Great Britain)",
  ["res nondestr eval"] = "Research in Nondestructive Evaluation",
  ["res notes aas"] = "Research Notes of the AAS",
  ["res notes math"] = "Research Notes in Mathematics",
  ["res number theory"] = "Research in Number Theory",
  ["res nurs health"] = "Research in nursing & health",
  ["res opin anim vet sci"] = "Research opinions in animal & veterinary sciences",
  ["res organ behav"] = "Research in organizational behavior",
  ["res pap educ"] = "Research papers in education",
  ["res pap hist med assoc"] = "Research papers (University of Arkansas for Medical Sciences. History of Medicine Associates)",
  ["res pap usda for serv"] = "Research Paper USDA Forest Service",
  ["res perspect ghent anal pde cent"] = "Research Perspectives Ghent Analysis and PDE Center",
  ["res perspect migr"] = "Research perspectives on migration : a joint project of the International Migration Policy Program of the Carnegie Endowment for International Peace and the Urban Institute",
  ["res pharm sci"] = "Research in pharmaceutical sciences",
  ["res plant biol"] = "Research in plant biology",
  ["res pol"] = "Research Policy",
  ["res policy"] = "Research policy",
  ["res popul ecol (kyoto)"] = "Researches on population ecology",
  ["res popul econ"] = "Research in population economics",
  ["res prac sch"] = "Research and Practice in the Schools : The Official Journal of the Texas Association of School Psychologists",
  ["res pract alzheimers dis"] = "Research and practice in Alzheimer's disease",
  ["res pract persons severe disabl"] = "Research and practice for persons with severe disabilities : the journal of TASH",
  ["res pract technol enhanc learn"] = "Research and practice in technology enhanced learning",
  ["res pract technol enhanced learn"] = "Research and Practice in Technology Enhanced Learning",
  ["res pract thromb haemost"] = "Research and practice in thrombosis and haemostasis",
  ["res prog org biol med chem"] = "Research progress in organic, biological and medicinal chemistry",
  ["res program"] = "Research program. United States. Army. Medical Department",
  ["res prostaglandins"] = "Research in prostaglandins",
  ["res psychother"] = "Research in psychotherapy (Milano)",
  ["res publ assoc res nerv ment dis"] = "Research publications - Association for Research in Nervous and Mental Disease",
  ["res publ wellcome unit hist med"] = "Research publications of the Wellcome Unit for the History of Medicine",
  ["res publica"] = "Res publica (Liverpool, England)",
  ["res publica litt"] = "Res publica litterarum",
  ["res q"] = "Research quarterly",
  ["res q exerc sport"] = "Research quarterly for exercise and sport",
  ["res rep"] = "Research Reports",
  ["res rep (montgomery, ala)"] = "Research reports (Montgomery)",
  ["res rep anan college tech"] = "Research Reports of the Anan College of Technology",
  ["res rep biochem"] = "Research and reports in biochemistry",
  ["res rep biol"] = "Research and reports in biology",
  ["res rep brooke army med cent"] = "Research report. Brooke Army Medical Center. Surgical Research Unit",
  ["res rep clin cardiol"] = "Research Reports in Clinical Cardiology",
  ["res rep endocr disord"] = "Research and reports in endocrine disorders",
  ["res rep forensic med sci"] = "Research and Reports in Forensic Medical Science",
  ["res rep health eff inst"] = "Research report (Health Effects Institute)",
  ["res rep inf sci"] = "Research Reports on Information Sciences",
  ["res rep med chem"] = "Research and reports in medicinal chemistry",
  ["res rep miyagi nat college tech"] = "Miyagi National College of Technology",
  ["res rep nav med res inst (us)"] = "Research report. Naval Medical Research Institute (U.S.)",
  ["res rep neonatol"] = "Research and reports in neonatology",
  ["res rep nucl med"] = "Research and reports in nuclear medicine",
  ["res rep trop med"] = "Research and reports in tropical medicine",
  ["res rep u s nav med field res lab"] = "[Research reports]. Naval Medical Field Research Laboratory (Camp Lejeune, N.C.)",
  ["res rep u s nav sch aviat med"] = "Research report. Naval School of Aviation Medicine (U.S.)",
  ["res rep urol"] = "Research and reports in urology",
  ["res reprod"] = "Research in reproduction",
  ["res resour rep"] = "Research resources reporter",
  ["res results pharmacol"] = "Research Results in Pharmacology",
  ["res rev"] = "Research reviews. United States. Office of Naval Research",
  ["res rev biosci"] = "Research and reviews in biosciences",
  ["res rev drug deliv"] = "Research & reviews. Drug delivery",
  ["res rev j eng technol"] = "Research & reviews. Journal of engineering and technology",
  ["res rev j hosp clin pharm"] = "Research & reviews. Journal of hospital and clinical pharmacy",
  ["res rev j microbiol biotechnol"] = "Research & reviews. Journal of microbiology and biotechnology",
  ["res rev j nurs health sci"] = "Research & reviews. Journal of nursing and health sciences",
  ["res rev j pharm pharm sci"] = "Research & reviews. Journal of pharmacy and pharmaceutical sciences",
  ["res rev j stat math sci"] = "Research & reviews. Journal of statistics and mathematical sciences",
  ["res rev polym"] = "Research & Reviews in Polymer",
  ["res rural sociol dev"] = "Research in rural sociology and development",
  ["res sch"] = "Research in the schools : a nationally refereed journal sponsored by the Mid-South Educational Research Association and the University of Alabama",
  ["res sci educ"] = "Research in science education",
  ["res serv med"] = "Research in the service of medicine",
  ["res soc sci study relig"] = "Research in the social scientific study of religion",
  ["res soc stratif mobil"] = "Research in social stratification and mobility",
  ["res soc work pract"] = "Research on social work practice",
  ["res social adm pharm"] = "Research in social & administrative pharmacy : RSAP",
  ["res sociol educ"] = "Research in sociology of education",
  ["res sociol educ social"] = "Research In Sociology Of Education And Socialization",
  ["res sociol health care"] = "Research in the sociology of health care",
  ["res sociol work"] = "Research in the sociology of work",
  ["res sports med"] = "Research in sports medicine (Print)",
  ["res sq"] = "Research square",
  ["res stat"] = "Research in Statistics",
  ["res stat note"] = "Research and statistics note - Social Security Administration, Office of Research and Statistics",
  ["res stat note health care financ adm off policy plan res"] = "Research and statistics note - Health Care Financing Administration, Office of Policy, Planning, and Research",
  ["res steroids (amst)"] = "Research on steroids",
  ["res stud"] = "Research studies (Pullman, Wash.)",
  ["res summ"] = "Research Summaries",
  ["res summ (nav med res inst)"] = "Research summaries",
  ["res synth methods"] = "Research synthesis methods",
  ["res theory nurs pract"] = "Research and theory for nursing practice",
  ["res today"] = "Research today",
  ["res trends phys"] = "Research Trends in Physics",
  ["res urban econ"] = "Research in urban economics",
  ["res vet sci"] = "Research in veterinary science",
  ["res virol"] = "Research in virology",
  ["research & reconstruction|pizhuhish va sazandidi"] = "Pizhuhish va sazandidi",
  ["research (lambertville)"] = "Research (Lambertville, N. J.)",
  ["research (wash d c)"] = "Research (Washington, D.C.)",
  ["research reports of atb|forschungsberichte atb"] = "Forschungsberichte des ATB",
  ["reseaux ciephum"] = "Réseaux - Ciephum",
  ["resen clin cient"] = "Resenha clínico-científica",
  ["resen med"] = "Resenha médica",
  ["resena hist"] = "Reseña histórica (Mexico City, Mexico)",
  ["resid staff physician"] = "Resident and staff physician",
  ["resid treat child youth"] = "Residential treatment for children & youth",
  ["residue rev"] = "Residue reviews",
  ["resilient cities struct"] = "Resilient Cities and Structures",
  ["resour biomed res educ"] = "Resources for biomedical research and education",
  ["resour chem mater"] = "Resources Chemicals and Materials",
  ["resour conserv"] = "Resource Conservation",
  ["resour conserv recycl"] = "Resources, Conservation and Recycling",
  ["resour conserv recycl: x"] = "Resources, Conservation and Recycling: X",
  ["resour conserv recyl adv"] = "Resources, Conservation & Recycling Advances",
  ["resour energy econ"] = "Resource and energy economics",
  ["resour eng technol sustainable world"] = "Resource Engineering and Technology for Sustainable World",
  ["resour environ sustainability"] = "Resources, Environment and Sustainability",
  ["resour fem res"] = "Resources for feminist research : RFR = Documentation sur la recherche féministe : DRF",
  ["resour geol"] = "Resource Geology",
  ["resour policy"] = "Resources Policy",
  ["resour recovery conserv"] = "Resource Recovery and Conservation",
  ["resour-effic technol"] = "Resource-Efficient Technologies",
  ["resource energy econ"] = "Resource and Energy Economics",
  ["resources energy"] = "Resources and Energy",
  ["resources pol"] = "Resources Policy",
  ["respir care"] = "Respiratory care",
  ["respir care clin n am"] = "Respiratory care clinics of North America",
  ["respir investig"] = "Respiratory investigation",
  ["respir manage"] = "Respiratory management",
  ["respir med"] = "Respiratory medicine",
  ["respir med case rep"] = "Respiratory medicine case reports",
  ["respir med cme"] = "Respiratory medicine CME",
  ["respir med res"] = "Respiratory medicine and research",
  ["respir med x"] = "Respiratory medicine: X",
  ["respir physiol"] = "Respiration physiology",
  ["respir physiol neurobiol"] = "Respiratory physiology & neurobiology",
  ["respir res"] = "Respiratory research",
  ["respir ther"] = "Respiratory therapy",
  ["respirol case rep"] = "Respirology case reports",
  ["responsa meridiana"] = "Responsa meridiana",
  ["responsive community"] = "The Responsive community : rights and responsibilities",
  ["restaurants inst"] = "Restaurants & institutions",
  ["reston book"] = "A Reston Book",
  ["restor dent endod"] = "Restorative dentistry & endodontics",
  ["restor ecol"] = "Restoration Ecology",
  ["restor neurol neurosci"] = "Restorative neurology and neuroscience",
  ["restor q"] = "Restoration quarterly",
  ["restorative dent"] = "Restorative dentistry",
  ["results appl math"] = "Results in Applied Mathematics",
  ["results chem"] = "Results in chemistry",
  ["results control optim"] = "Results in Control and Optimization",
  ["results eng"] = "Results in Engineering",
  ["results geophys sci"] = "Results in Geophysical Sciences",
  ["results immunol"] = "Results in immunology",
  ["results in geochem"] = "Results in Geochemistry",
  ["results mater"] = "Results in Materials",
  ["results math"] = "Results in Mathematics",
  ["results opt"] = "Results in Optics",
  ["results pharma sci"] = "Results in pharma sciences",
  ["results phys"] = "Results in Physics",
  ["results probl cell differ"] = "Results and problems in cell differentiation",
  ["results surf interfaces"] = "Results in Surfaces and Interfaces",
  ["resum trab med venez"] = "Resúmenes de trabajos médicos venezolanos",
  ["resusc plus"] = "Resuscitation plus",
  ["rethink hist"] = "Rethinking history",
  ["rethinking ecol"] = "Rethinking Ecology",
  ["retin cases brief rep"] = "Retinal cases & brief reports",
  ["retin physician"] = "Retinal physician",
  ["retrovirology (auckl)"] = "Retrovirology : research and treatment",
  ["reumatol clin"] = "Reumatología clinica",
  ["reumatol pol"] = "Reumatologia polska",
  ["rev abpapal"] = "Revista ABP-APAL",
  ["rev acad arabe damas"] = "Majallat Majmaʻ al-Lughah al-ʻArabīyah bi-Dimashq",
  ["rev acad canaria cienc"] = "Revista de la Academia Canaria de Ciencias",
  ["rev acad cienc zaragoza (2)"] = "Revista de la Academia de Ciencias Exactas, Físicas Químicas y Naturales de Zaragoza",
  ["rev acad colomb cienc exactas fis nat"] = "Revista de la Academia Colombiana de Ciencias Exactas, Fisicas y Naturales",
  ["rev acad colombiana cienc exact fis natur"] = "Academia Colombiana de Ciencias Exactas, Fisicas y Naturales. Revista",
  ["rev acad colombiana cienc exact fís natur"] = "Academia Colombiana de Ciencias Exactas, Físicas y Naturales",
  ["rev acad pol sci"] = "Revue de l'Académie polonaise des sciences. Polska Akademia Nauk",
  ["rev acc stud"] = "Review of Accounting Studies",
  ["rev accel sci technol"] = "Reviews of Accelerator Science and Technology",
  ["rev act metallges ag"] = "Review of the Activities Metallgesellschaft AG",
  ["rev actual estomatol esp"] = "Revista de actualidad estomatológica española",
  ["rev actual odontoestomatol esp"] = "Revista de actualidad odontoestomatológica española : boletín de información, del Ilustre Consejo General de Colegios de Odontólogos y Estomatólogos de España",
  ["rev adiccion cienc"] = "Revista adicción y ciencia",
  ["rev adm"] = "Revista ADM (Asociación Dental Mexicana : 1988)",
  ["rev adm munic"] = "Revista de administração municipal",
  ["rev adm nac agua"] = "Revista de la Administración Nacional del Agua. Argentina. Administración Nacional del Agua",
  ["rev adv chem"] = "Reviews and Advances in Chemistry",
  ["rev adv mater sci"] = "Reviews on Advanced Materials Science",
  ["rev afr polit econ"] = "Review of African political economy",
  ["rev african polit economy"] = "Review of African Political Economy",
  ["rev agr econ"] = "Review of Agricultural Economics",
  ["rev agrar acad"] = "Revista agrária acadêmica",
  ["rev agric, food environ stud"] = "Review of Agricultural, Food and Environmental Studies",
  ["rev agrup odontol cap fed"] = "Revista de la Agrupacion Odontologica de la Capital Federal",
  ["rev alerg"] = "Revista alergia : organo oficial de la Sociedad Mexicana de Alergia e Inmunlogía",
  ["rev alerg mex"] = "Revista alergia Mexico (Tecamachalco, Puebla, Mexico : 1993)",
  ["rev aliment"] = "Revista alimentar",
  ["rev allem"] = "Revue d'Allemagne",
  ["rev allergy"] = "Review of allergy",
  ["rev allergy appl immunol"] = "Review of allergy and applied immunology",
  ["rev alsace"] = "Revue d'Alsace",
  ["rev am hist"] = "Reviews in American history",
  ["rev anal chem"] = "Reviews in Analytical Chemistry",
  ["rev anal econ"] = "Revue D’Analyse Economique",
  ["rev anal numér théor approx"] = "Revue d’Analyse Numérique et de Théorie de l’Approximation",
  ["rev analg"] = "Reviews in analgesia",
  ["rev andal med deport"] = "Revista andaluza de medicina del deporte",
  ["rev annu soc odontostomatol nordest"] = "Revue annuelle - Société odonto-stomatologique du Nord-Est",
  ["rev antopol soc"] = "Revista de antropología social",
  ["rev aquac"] = "Reviews in aquaculture",
  ["rev aquacult"] = "Reviews in Aquaculture",
  ["rev arch bibl mus"] = "Revista de archivos, bibliotecas y museos (Madrid, Spain : 1897)",
  ["rev archeol centre france"] = "Revue archéologique du centre de la France",
  ["rev archeol mod archeol gen"] = "Revue d'archéologie moderne et d'archéologie générale : RAMAGE",
  ["rev archéom"] = "Revue d’Archéometrie",
  ["rev argent alerg"] = "Revista argentina de alergia",
  ["rev argent anest analg"] = "Revista argentina de anestesia y analgesia",
  ["rev argent antropol biol"] = "Revista argentina de antropología biológica",
  ["rev argent broncoesofagol"] = "Revista argentina de bronco-esofagología",
  ["rev argent cardiol"] = "Revista argentina de cardiología",
  ["rev argent cienc comport"] = "Revista Argentina de ciencias del comportamiento",
  ["rev argent cir"] = "Revista argentina de cirugía",
  ["rev argent dermatosifilol"] = "Revista argentina de dermatosifilología",
  ["rev argent endocrinol metab"] = "Revista argentina de endocrinología y metabolismo",
  ["rev argent implantol estomatol"] = "Revista Argentina De Implantologia Estomatologica",
  ["rev argent microbiol"] = "Revista Argentina de microbiología",
  ["rev argent neurol psiquiatr"] = "Revista argentina de neurología y psiquiatría",
  ["rev argent neurol psiquiatr neurocir"] = "Revista argentina de neurología, psiquiatría y neurocirugía",
  ["rev argent pueric neonatol"] = "Revista argentina de puericultura e neonatología",
  ["rev argent reumatol"] = "Revista argentina de reumatología",
  ["rev argent salud publica"] = "Revista argentina de salud pública",
  ["rev argent tuberc enferm pulm"] = "Revista argentina de tuberculosis y enfermedades pulmonares",
  ["rev argent urol"] = "Revista argentina de urología",
  ["rev argent urol nefrol"] = "Revista argentina de urología y nefrología",
  ["rev argent zoonosis enferm infecc emerg"] = "Revista Argentina de zoonosis y enfermedades infecciosas emergentes",
  ["rev arh"] = "Revista arhivelor",
  ["rev arvore"] = "Revista Arvore",
  ["rev asoc argent microbiol"] = "Revista de la Asociacion Argentina de Microbiologia",
  ["rev asoc bioquim argent"] = "Revista de la Asociación Bioquímica Argentina. Asociación Bioquímica Argentina",
  ["rev asoc columb med vet"] = "Revista. Asociación Colombiana de Médicos Veterinarios",
  ["rev asoc geol argent"] = "Revista de la Asociacion Geologica Argentina",
  ["rev asoc med argent"] = "Revista de la Asociacion Medica Argentina",
  ["rev asoc med latinoam"] = "Revista. Asociación Médica Latinoamericana",
  ["rev asoc medica mex"] = "Revista. Asociación Médica Mexicana",
  ["rev asoc mex enferm"] = "Revista de la Asociacion Mexicana de Enfermeras",
  ["rev asoc odontol argent"] = "Revista de la Asociacion Odontologica Argentina",
  ["rev asoc odontol costa rica"] = "Revista de la Asociacion Odontologica de Costa Rica",
  ["rev asoc prof hosp nac odontol"] = "Revista De La Asociacion De Profesionales Del Hospital Nacional De Odontologia",
  ["rev asset pricing stud"] = "Review of asset pricing studies",
  ["rev assoc med bras"] = "Revista da Associacao Medica Brasileira",
  ["rev assoc med bras (1992)"] = "Revista da Associação Médica Brasileira (1992)",
  ["rev assoc med minas gerais"] = "Revista da Associação Médica de Minas Gerais",
  ["rev assoc med rio grande do sul"] = "Revista da Associação Médica do Rio Grande do Sul",
  ["rev assoc paul cir dent"] = "Revista da Associacao Paulista de Cirurgioes Dentistas",
  ["rev assoc paul cir dent reg aracatuba"] = "Revista da Associação Paulista de Cirurgiões Dentistas Regional de Araçatuba",
  ["rev assyriol archeol orient"] = "Revue d'assyriologie et d'archéologie orientale",
  ["rev ateneo argent odontol"] = "Revista del Ateneo Argentino de Odontología : R.A.A.O",
  ["rev ateneo catedra tec oper dent"] = "Revista Del Ateneo De La Catedra De Tecnica De Operatoria Dental",
  ["rev atheroscler"] = "Revue de l'athérosclérose",
  ["rev atheroscler (paris)"] = "Revue de l'athérosclérose et des artériopathies périphériques",
  ["rev austrian econ"] = "Review of Austrian Economics",
  ["rev automot eng"] = "Review of Automotive Engineering",
  ["rev auvergne"] = "Revue d'Auvergne",
  ["rev avranchin pays granville"] = "Revue de l'Avranchin et du pays de Granville",
  ["rev belg homoeopath"] = "Revue belge d'homoeopathie",
  ["rev belg pathol med exp"] = "Revue belge de pathologie et de médecine expérimentale",
  ["rev belge archeol hist art"] = "Revue belge d'archéologie et d'histoire de l'art",
  ["rev belge geogr"] = "Revue belge de géographie",
  ["rev belge hist milit"] = "Revue belge d'histoire militaire = Belgisch tijdschrift voor militaire geschiedenis",
  ["rev belge med dent"] = "Revue Belge de Medecine Dentaire",
  ["rev belge med dent (1984)"] = "Revue belge de médecine dentaire",
  ["rev belge philol hist"] = "Revue belge de philologie et d'histoire. Belgisch tijdschrift voor philologie en geschiedenis",
  ["rev belge sci dent"] = "Revue belge de science dentaire. Belgisch tijdschrift voor tandheelkunde",
  ["rev belge sci med"] = "Revue belge des sciences médicales",
  ["rev belge stomatol"] = "Revue belge de stomatologie. Belgisch tijdschrift voor stomatologie",
  ["rev bibl nac jose marti"] = "Revista de la Biblioteca Nacional José Martí. Biblioteca Nacional José Martí",
  ["rev bibl natl"] = "Revue de la Bibliothèque nationale. Bibliothèque nationale (France)",
  ["rev biodivers neotrop"] = "Revista biodiversidad neotropical",
  ["rev bioet derecho"] = "Revista de bioética y derecho",
  ["rev biol mar oceanogr"] = "Revista de biología marina y oceanografía",
  ["rev biol oral"] = "Revista de biologia oral",
  ["rev biol trop"] = "Revista de Biologia Tropical",
  ["rev black polit econ"] = "The Review of Black political economy",
  ["rev black polit economy"] = "Review of Black Political Economy",
  ["rev bois appl"] = "Revue du bois et de ses applications",
  ["rev boliv quim"] = "Revista Boliviana de Quimica",
  ["rev bras agroecol"] = "Revista Brasileira de agroecologia",
  ["rev bras anal comport"] = "Revista brasileira de análise do comportamento",
  ["rev bras anestesiol"] = "Revista brasileira de anestesiologia",
  ["rev bras apl vacuo"] = "Revista Brasileira de Aplicacoes de Vacuo",
  ["rev bras biocienc"] = "Revista brasileira de biociências = Brazilian journal of biosciences",
  ["rev bras biol"] = "Revista brasileira de biologia",
  ["rev bras bot"] = "Revista brasileira de botânica : Brazilian journal of botany",
  ["rev bras cienc avic"] = "Revista brasileira de ciência avi̧cola",
  ["rev bras cienc solo"] = "Revista brasileira de ciência do solo",
  ["rev bras cir"] = "Revista brasileira de cirurgia",
  ["rev bras cir cardiovasc"] = "Revista brasileira de cirurgia cardiovascular : órgão oficial da Sociedade Brasileira de Cirurgia Cardiovascular",
  ["rev bras crescimento desenvolv hum"] = "Revista brasileira de crescimento e desenvolvimento humano",
  ["rev bras educ med"] = "Revista brasileira de educação médica",
  ["rev bras enferm"] = "Revista brasileira de enfermagem",
  ["rev bras eng agric ambiental"] = "Revista Brasileira de Engenharia Agricola e Ambiental",
  ["rev bras ensino quim"] = "Revista Brasileira de Ensino de Quimica",
  ["rev bras entomol"] = "Revista brasileira de entomologia",
  ["rev bras epidemiol"] = "Revista brasileira de epidemiologia = Brazilian journal of epidemiology",
  ["rev bras estat"] = "Revista brasileira de estatística",
  ["rev bras estud popul"] = "Revista brasileira de estudos de população",
  ["rev bras farm"] = "Revista brasileira de farmácia",
  ["rev bras farmacogn"] = "Revista brasileira de farmacognosia : orgão oficial da Sociedade Brasileira de Farmacognosia",
  ["rev bras fis"] = "Revista Brasileira de Fisica",
  ["rev bras fisioter"] = "Revista brasileira de fisioterapia (São Carlos (São Paulo, Brazil))",
  ["rev bras frutic"] = "Revista Brasileira de fruticultura",
  ["rev bras gastroenterol"] = "Revista brasileira de gastroenterologia",
  ["rev bras genet"] = "Revista brasileira de genética",
  ["rev bras geofis"] = "Revista Brasileira de Geofisica",
  ["rev bras geogr"] = "Revista brasileira de geografia",
  ["rev bras ginecol obstet"] = "Revista brasileira de ginecologia e obstetrícia : revista da Federação Brasileira das Sociedades de Ginecologia e Obstetrícia",
  ["rev bras hematol hemoter"] = "Revista brasileira de hematologia e hemoterapia",
  ["rev bras hist"] = "Revista brasileira de história",
  ["rev bras hist mat"] = "Revista Brasileira de Historia da Matematica. An International Journal on the History of Mathematics",
  ["rev bras hortic ornam"] = "Revista brasileira de horticultura ornamental",
  ["rev bras leprol"] = "Revista brasileira de leprologia",
  ["rev bras malariol doencas trop"] = "Revista brasileira de malariologia e doenças tropicais. Publicações avulsas",
  ["rev bras med"] = "Revista brasileira de medicina",
  ["rev bras med trab"] = "Revista brasileira de medicina do trabalho : publicação oficial da Associação Nacional de Medicina do Trabalho-ANAMT",
  ["rev bras med vet"] = "Revista brasileira de medicina veterinaria",
  ["rev bras milho sorgo"] = "Revista Brasileira de milho e sorgo",
  ["rev bras oceangr"] = "Revista Brasileira de Oceanografia",
  ["rev bras odontol"] = "Revista brasileira de odontologia",
  ["rev bras odontol mil"] = "Revista brasileira de odontologia militar : órgão oficial da Academia Brasileira de Odontologia Militar",
  ["rev bras oftalmol"] = "Revista brasileira de oftalmologia",
  ["rev bras ornitol"] = "Revista brasileira de ornitologia",
  ["rev bras ortop"] = "Revista brasileira de ortopedia",
  ["rev bras ortop (sao paulo)"] = "Revista brasileira de ortopedia",
  ["rev bras otorinolaringol"] = "Revista brasileira de otorrinolaringologia",
  ["rev bras otorrinolaringol"] = "Revista brasileira de oto-rino-laringologia",
  ["rev bras otorrinolaringol (engl ed)"] = "Revista brasileira de otorrinolaringologia (English ed.)",
  ["rev bras paleontol"] = "Revista Brasileira de Paleontologia",
  ["rev bras parasitol vet"] = "Revista brasileira de parasitologia veterinária = Brazilian journal of veterinary parasitology : Órgão Oficial do Colégio Brasileiro de Parasitologia Veterinária",
  ["rev bras pesqui med biol"] = "Revista brasileira de pesquisas médicas e biológicas",
  ["rev bras psiquiatr"] = "Revista brasileira de psiquiatria (Sao Paulo, Brazil : 1999)",
  ["rev bras reumatol"] = "Revista brasileira de reumatologia",
  ["rev bras sementes"] = "Revista brasileira de sementes = Brazilian seed journal",
  ["rev bras ter intensiva"] = "Revista Brasileira de terapia intensiva",
  ["rev bras tuberc doencas torac"] = "Revista brasileira de tuberculose e doenças torácicas",
  ["rev bras zool"] = "Revista brasileira de zoologia",
  ["rev bryol lichénol"] = "Revue bryologique et lichénologique",
  ["rev bull calcutta math soc"] = "Calcutta Mathematical Society. Review Bulletin",
  ["rev bus econ res"] = "Review of Business and Economic Research",
  ["rev calid asist"] = "Revista de calidad asistencial : organo de la Sociedad Española de Calidad Asistencial",
  ["rev camara coner bogota"] = "Revista (Cámara de Comercio de Bogotá)",
  ["rev can biol"] = "Revue canadienne de biologie",
  ["rev can biol exp"] = "Revue canadienne de biologie expérimentale",
  ["rev can etudes dev"] = "Revue canadienne d'études du développement = Canadian journal of development studies",
  ["rev cardiovasc med"] = "Reviews in cardiovascular medicine",
  ["rev catalana micol"] = "Revista Catalana De Micologia",
  ["rev catarinense odontol"] = "Revista catarinense de odontologia",
  ["rev cent am odontol"] = "Revista Centro América odontológica",
  ["rev cent estud demograficos"] = "Revista. Centro de Estudos Demográficos (Portugal)",
  ["rev cent etud pays est"] = "Revue du Centre d'étude des pays de l'Est et du Centre national pour l'étude des États de l'Est. Brussels (Belgium). Université libre. Centre d'étude des pays de l'Est",
  ["rev cent policlin valencia"] = "Revista Del Centro Policlinico Valencia",
  ["rev centro cienc biomed univ fed uberlandia"] = "Revista do Centro de Ciencias Biomedicas da Universidade Federal de Uberlandia",
  ["rev centroam econ"] = "Revista centroamericana de economía",
  ["rev cepae"] = "Revista CEPAE",
  ["rev chapingo ser hortic"] = "Revista Chapingo. Serie: Horticultura",
  ["rev chem eng"] = "Reviews in Chemical Engineering",
  ["rev chil entomol"] = "Revista chilena de entomología : publicación de la Faculdad de Filosofía y Educación, Universidad de Chile, y de la Sociedad Chilena de Entomología",
  ["rev chil hig med prev"] = "Revista chilena de higiene y medicina preventiva",
  ["rev chil hist geogr"] = "Revista chilena de historia y geografía",
  ["rev chil hist nat"] = "Revista chilena de historia natural (Valparaíso, Chile : 1983)",
  ["rev chil neuropsiquiatr"] = "Revista chilena de neuro-psiquiatría",
  ["rev chil nutr"] = "Revista chilena de nutrición : órgano oficial de la Sociedad Chilena de Nutrición, Bromatología y Toxicología",
  ["rev chil obstet ginecol"] = "Revista chilena de obstetricia y ginecología",
  ["rev chil odontoestomatol"] = "Revista chilena de odontoestomatología",
  ["rev chil pediatr"] = "Revista chilena de pediatría",
  ["rev chil urol"] = "Revista chilena de urología",
  ["rev chilena infectol"] = "Revista chilena de infectología : órgano oficial de la Sociedad Chilena de Infectología",
  ["rev chim"] = "Revista de Chimie",
  ["rev chim (bucharest, rom)"] = "Revista de Chimie (Bucharest, Romania)",
  ["rev chir"] = "Revue de chirurgie",
  ["rev chir oncol radiol o r l oftalmol stomatol chir"] = "Revista de chirurgie, oncologie, radiologie, o.r.l., oftalmologie, stomatologie. Chirurgie",
  ["rev chir oncol radiol o r l oftalmol stomatol otorinolaringol"] = "Revista de chirurgie, oncologie, radiologie, o.r.l., oftalmologie, stomatologie. Oto-rino-laringologia",
  ["rev chir oncol radiol o r l oftalmol stomatol ser oftalmol"] = "Revista de chirurgie, oncologie, radiologie, o. r. l., oftalmologie, stomatologie. Seria: Oftalmologie",
  ["rev chir oncol radiol o r l oftalmol stomatol ser stomatol"] = "Revista de chirurgie, oncologie, radiologie, o. r. l., oftalmologie, stomatologie. Seria: Stomatologie",
  ["rev chir oncol radiol orl oftalmol stomatol oncol"] = "Revista de chirurgie, oncologie, radiologie, o. r. l., oftalmologie, stomatologie. Oncologia",
  ["rev chir orthop reparatrice appar mot"] = "Revue de chirurgie orthopédique et réparatrice de l'appareil moteur",
  ["rev cienc soc"] = "Revista de ciencias sociales",
  ["rev cient (maracaibo)"] = "Revista científica (Universidad del Zulia. Facultad de Ciencias Veterinarias. División de Investigación)",
  ["rev cient rural"] = "Revista científica rural",
  ["rev cim migr am lat"] = "Revista del CIM sobre migraciones en América Latina",
  ["rev cinc biomed"] = "Revista de ciências biomédicas",
  ["rev cir (mex)"] = "Revista de cirugía",
  ["rev cir sao paulo"] = "Revista de cirurgia de São Paulo",
  ["rev cir urug"] = "Revista de cirugía del Uruguay",
  ["rev circ argent odontol"] = "Revista. Círculo Argentino de Odontología",
  ["rev circ odontol cordoba"] = "Revista del Circulo Odontologico de Cordoba",
  ["rev circ odontol oeste"] = "Revista del Círculo Odontólogico del Oeste. Círculo Odontológico del Oeste, Buenos Aires",
  ["rev circ odontol sur"] = "Revista Del Circulo Odontologico Del Sur",
  ["rev circul ondontol ros"] = "Revista. Círculo Odontológico de Rosario",
  ["rev clin basic pharm"] = "Reviews in clinical & basic pharmacology",
  ["rev clin esp"] = "Revista Clinica Espanola",
  ["rev clin esp (barc)"] = "Revista clínica espanõla",
  ["rev clin exp hematol"] = "Reviews in clinical and experimental hematology",
  ["rev clin gerontol"] = "Reviews in clinical gerontology",
  ["rev clin inst matern lisb"] = "Revista clínica. Lisbon. Instituto Maternal",
  ["rev clin sao paulo"] = "Revista clinica de São Paulo",
  ["rev col bras cir"] = "Revista do Colégio Brasileiro de Cirurgiões",
  ["rev col estomatol guatem"] = "Revista del Colegio Estomatologico de Guatemala",
  ["rev col med"] = "Revista del Colegio Médico",
  ["rev col med cir guatem"] = "Revista del Colegio de Médicos y Cirujanos de Guatemala",
  ["rev col med estado falcon"] = "Revista del Colegio Médico del Estado Falcón",
  ["rev col med guatem"] = "Revista del Colegio Medico de Guatemala",
  ["rev col nac enferm"] = "Revista del Colegio Nacional de Enfermeras",
  ["rev colomb biotecnol"] = "Revista Colombiana de biotecnología",
  ["rev colomb entomol"] = "Revista Colombiana de Entomologia",
  ["rev colomb estad"] = "Revista Colombiana de estadística",
  ["rev colomb gastroenterol"] = "Revista colombiana de gastroenterología",
  ["rev colomb obstet ginecol"] = "Revista colombiana de obstetricia y ginecología",
  ["rev colomb pediatr pueric"] = "Revista colombiana de pediatría y puericultura",
  ["rev colomb psiquiatr"] = "Revista colombiana de psiquiatría",
  ["rev colomb quim"] = "Revista Colombiana de Quimica",
  ["rev colombiana estadist"] = "Revista Colombiana de Estadistica",
  ["rev colombiana estadíst"] = "Revista Colombiana Estadística",
  ["rev colombiana mat"] = "Revista Colombiana de Matematicas",
  ["rev coloniale med chir"] = "Revue coloniale de médecine & chirurgie",
  ["rev comm"] = "The review of communication",
  ["rev comminges"] = "Revue de Comminges (Pyrénées Centrales)",
  ["rev complut hist am"] = "Revista complutense de historia de América",
  ["rev compos mater av"] = "Revue des Composites et des Materiaux Avances",
  ["rev comput chem"] = "Reviews in Computational Chemistry",
  ["rev conserv"] = "Reviews in Conservation",
  ["rev constr"] = "Revista de la Construccion",
  ["rev coree"] = "Revue de Corée",
  ["rev corps sante armees terre mer air"] = "Revue des corps de santé des armées terre, mer, air",
  ["rev corps sante mil"] = "Revue du corps de santé militaire",
  ["rev costarric cienc med"] = "Revista costarricense de ciencias médicas",
  ["rev criminol polic cient"] = "Revista de criminología y policía científica",
  ["rev crit cienc sociais"] = "Revista crítica de ciencias sociais",
  ["rev cub salud publica"] = "Revista cubana de salud pública",
  ["rev cuba med sov"] = "Revista cubana de medicina soviética",
  ["rev cuba tuberc"] = "Revista cubana de tuberculosis",
  ["rev cubana cardiol"] = "Revista cubana de cardiología",
  ["rev cubana enferm"] = "Revista cubana de enfermería",
  ["rev cubana estomatol"] = "Revista cubana de estomatología",
  ["rev cubana hig epidemiol"] = "Revista cubana de higiene y epidemiología",
  ["rev cubana lab clin"] = "Revista cubana de laboratorio clínico",
  ["rev cubana med"] = "Revista cubana de medicina",
  ["rev cubana med trop"] = "Revista cubana de medicina tropical",
  ["rev cubana pediatr"] = "Revista cubana de pediatría",
  ["rev cuhana adm salud"] = "Revista cubana de administración de salud",
  ["rev cult vozes"] = "Vozes",
  ["rev cytol clin"] = "Revue de cytologie clinique",
  ["rev czech med"] = "Review of Czechoslovak medicine",
  ["rev demogr hist"] = "Revista de demografía histórica",
  ["rev dent (st domingo)"] = "Revista dental",
  ["rev dent chile"] = "Revista dental de Chile",
  ["rev dent liban"] = "Revue dentaire libanaise. Lebanese dental magazine",
  ["rev derecho genoma hum"] = "Revista de derecho y genoma humano = Law and the human genome review",
  ["rev derivatives res"] = "Review of Derivatives Research",
  ["rev desarro int"] = "Revista del desarrollo internacional. Revue du développement international. International development review",
  ["rev deux mondes"] = "Revue des deux mondes (Paris, France : 1948)",
  ["rev dev econ"] = "Review of development economics",
  ["rev devel econ"] = "Review of Development Economics",
  ["rev diabet stud"] = "The review of diabetic studies : RDS",
  ["rev dialoctol tradiciones pop (madrid)"] = "Revista de dialectología y tradiciones populares",
  ["rev diet"] = "Revue de diététique",
  ["rev dimens"] = "Revista dimension",
  ["rev disabil stud"] = "Review of disability studies",
  ["rev divulg tec agric agroind"] = "Revista de divulgación técnica agrícola y agroindustrial",
  ["rev droit"] = "Revue de droit",
  ["rev dromoise"] = "Revue dromoise",
  ["rev drug metab drug interact"] = "Reviews on drug metabolism and drug interactions",
  ["rev east med sci"] = "Review of eastern medical sciences",
  ["rev eaux for"] = "Revue des eaux et forêts",
  ["rev ecol"] = "Revue d'écologie",
  ["rev econ"] = "Revista economică",
  ["rev econ cond italy"] = "Review of economic conditions in Italy",
  ["rev econ des"] = "Review of Economic Design",
  ["rev econ design"] = "Review of Economic Design",
  ["rev econ dev"] = "Revue d'économie du développement",
  ["rev econ dyn"] = "Review of economic dynamics",
  ["rev econ dynam"] = "Review of Economic Dynamics",
  ["rev econ financ"] = "Revue d'économie financière",
  ["rev econ househ"] = "Review of economics of the household",
  ["rev econ nordeste"] = "Revista econômica do Nordeste",
  ["rev econ polit"] = "Revue d'économie politique",
  ["rev econ reg urbaine"] = "Revue d'économie régionale et urbaine (ADICUEER (Association))",
  ["rev econ soc"] = "Revue économique et sociale",
  ["rev econ stat"] = "The review of economics and statistics",
  ["rev econ statist"] = "Review of Economics and Statistics",
  ["rev econ stud"] = "The Review of economic studies",
  ["rev econ sud ouest"] = "Revue économique du Sud-Ouest",
  ["rev econom stud"] = "Review of Economic Studies",
  ["rev ecuat hig med trop"] = "Revista ecuatoriana de higiene y medicina tropical",
  ["rev ecuat med cienc biol"] = "Revista ecuatoriana de medicina y ciencias biológicas",
  ["rev ecuat pediatr"] = "Revista ecuatoriana de pediatría",
  ["rev ecuat pediatr pueric"] = "Revista ecuatoriana de pediatría y puericultura",
  ["rev educ res"] = "Review of educational research",
  ["rev eesp cir traumatol ortop"] = "Revista española de cirugía, traumatología y ortopedia",
  ["rev egyptol"] = "Revue d'égyptologie",
  ["rev electr electron"] = "Revue de l’Electricite et de l’Electronique",
  ["rev electroencephalogr neurophysiol clin"] = "Revue d'électroencéphalographie et de neurophysiologie clinique",
  ["rev electron comun inf inov saude"] = "Revista electrônica de comunicação, informação & inovação em saúde : RECIIS",
  ["rev electron investig psicoeduc psigopedag"] = "Revista electrónica de investigación psicoeducativa : REIPE = Electronic journal of research in educational psychology",
  ["rev elev med vet pays trop"] = "Revue d'élevage et de médecine vétérinaire des pays tropicaux",
  ["rev endocr metab disord"] = "Reviews in endocrine & metabolic disorders",
  ["rev enferm"] = "Revista de enfermería (Barcelona, Spain)",
  ["rev enferm (lisboa)"] = "Revista de enfermagem",
  ["rev enferm nov dimens"] = "Revista enfermagem em novas dimensões",
  ["rev enseign philos"] = "Revue de l'enseignement philosophique",
  ["rev enseign super"] = "Revue de l'enseignement supérieur",
  ["rev environ contam toxicol"] = "Reviews of environmental contamination and toxicology",
  ["rev environ econ policy"] = "Review of environmental economics and policy",
  ["rev environ health"] = "Reviews on environmental health",
  ["rev environ sci bio/technol"] = "Reviews in Environmental Science and Bio/Technology",
  ["rev environ sci biotechnol"] = "Reviews in Environmental Science and Biotechnology",
  ["rev environ toxicol"] = "Reviews in Environmental Toxicology",
  ["rev epidemiol med soc sante publique"] = "Revue d'épidémiologie, médecine sociale et santé publique",
  ["rev epidemiol sante publique"] = "Revue d'épidémiologie et de santé publique",
  ["rev esc enferm usp"] = "Revista da Escola de Enfermagem da U S P",
  ["rev esc odontol tucuman"] = "Revista De La Escuela De Odontologia, Universidad Nacional De Tucuman, Facultad De Medicina",
  ["rev esp anestesiol"] = "Revista española de anestesiología",
  ["rev esp anestesiol reanim"] = "Revista española de anestesiología y reanimación",
  ["rev esp antropol am"] = "Revista española de antropología americana",
  ["rev esp cardiol"] = "Revista Espanola de Cardiologia",
  ["rev esp cardiol (engl ed)"] = "Revista española de cardiología (English ed.)",
  ["rev esp cir (madr 1944)"] = "Revista española de cirugía",
  ["rev esp cir ortop traumatol"] = "Revista española de cirugía ortopédica y traumatología",
  ["rev esp doc cient"] = "Revista española la de documentación científica",
  ["rev esp endodoncia"] = "Revista española de endodoncia",
  ["rev esp enferm apar dig"] = "Revista española de las enfermedades del aparato digestivo",
  ["rev esp enferm apar dig nutr"] = "Revista española de las enfermedades del aparato digestivo y de la nutrición",
  ["rev esp enferm dig"] = "Revista española de enfermedades digestivas : organo oficial de la Sociedad Española de Patología Digestiva",
  ["rev esp escler mult"] = "Revista española de esclerosis múltiple",
  ["rev esp estomatol"] = "Revista española de estomatología",
  ["rev esp fisiol"] = "Revista Espanola de Fisiologia",
  ["rev esp geriatr gerontol"] = "Revista española de geriatría y gerontología",
  ["rev esp herpetol"] = "Revista española de herpetología",
  ["rev esp invest sociol"] = "Reis",
  ["rev esp linguist apl"] = "Revista española de lingüística aplicada",
  ["rev esp med cir guerra"] = "Revista española de medicina y cirugía de guerra",
  ["rev esp med nucl"] = "Revista española de medicina nuclear",
  ["rev esp med nucl imagen mol"] = "Revista española de medicina nuclear e imagen molecular",
  ["rev esp obstet ginecol"] = "Revista española de obstetricia y ginecología",
  ["rev esp oncol"] = "Revista española de oncología",
  ["rev esp otoneurooftalmol neurocir"] = "Revista española de oto-neuro-oftalmología y neurocirugía",
  ["rev esp paleontol"] = "Revista española de paleontología",
  ["rev esp parad"] = "Revista Espanola De Paradoncia",
  ["rev esp patol"] = "Revista española de patología : publicación oficial de la Sociedad Española de Anatomía Patológica y de la Sociedad Española de Citología",
  ["rev esp pediatr"] = "Revista española de pediatría",
  ["rev esp quimioter"] = "Revista española de quimioterapia : publicación oficial de la Sociedad Española de Quimioterapia",
  ["rev esp reum enferm osteoartic"] = "Revista española de reumatismo y enfermedades osteoarticulares",
  ["rev esp reumatol"] = "Revista española de reumatología : órgano oficial de la Sociedad Española de Reumatología",
  ["rev esp salud publica"] = "Revista española de salud pública",
  ["rev esp sanid penit"] = "Revista española de sanidad penitenciaria",
  ["rev esp tuberc"] = "Revista española de tuberculosis",
  ["rev est"] = "Revue de l'Est",
  ["rev estad"] = "Revista estadística (Havana, Cuba)",
  ["rev estomatol cuba"] = "Revista estomatológica de Cuba",
  ["rev estud extremenos"] = "Revista de estudios extremeños",
  ["rev estud vida local"] = "Revista de estudios de la vida local",
  ["rev ethnol que"] = "Revue d'ethnologie du Québec",
  ["rev etud byz"] = "Revue des études byzantines",
  ["rev etud comp est ouest"] = "Revue d'études comparatives est-ouest",
  ["rev etud grec"] = "Revue des études grecques",
  ["rev etud ital"] = "Revue des études italiennes",
  ["rev etud juives"] = "Revue des études juives",
  ["rev etud lat"] = "Revue des études latines",
  ["rev etud palest"] = "Revue d'études palestiniennes : revue trimestrielle publiée par l'Institut des études palestiniennes",
  ["rev etud slaves"] = "Revue des études slaves",
  ["rev etud sud est eur"] = "Revue des études sud-est européennes",
  ["rev eur comp int environ law"] = "Review of European, comparative & international environmental law",
  ["rev eur dermatol mst"] = "Revue européenne de dermatologie et de MST",
  ["rev eur estud latinoam caribe"] = "Revista europea de estudios latinoamericanos y del Caribe = European review of Latin American and Caribbean studies",
  ["rev eur etud clin biol"] = "Revue européenne d'études cliniques et biologiques. European journal of clinical and biological research",
  ["rev eur migr int"] = "Revue européenne des migrations internationales",
  ["rev eur odontoestomatol"] = "Revista europea de odonto-estomatología",
  ["rev eur stud"] = "Review of European studies",
  ["rev européenne élém finis"] = "Revue Européenne des Éléments Finis",
  ["rev fac cien med univ nac cordoba"] = "Revista de la Facultad de Ciencias Médicas (Córdoba, Argentina)",
  ["rev fac cienc agrar"] = "Revista de la Facultad de Ciencias Agrarias",
  ["rev fac cienc med"] = "Revista de la Facultad de Ciencias Médicas",
  ["rev fac cienc med cordoba"] = "Revista de la Facultad de Ciencias Medicas de Cordoba",
  ["rev fac cienc med univ nac cordoba"] = "Revista de la Facultad de Ciencias Médicas, Universidad Nacional de Córdoba",
  ["rev fac cienc quim quim farm"] = "Revista de la Facultad de Ciencias Químicas (Química y Farmacia). La Plata. Universidad Nacional. Facultad de Química y Farmacia",
  ["rev fac farm odontol araraquara"] = "Revista da Faculdade de Farmacia e Odontologia de Araraquara",
  ["rev fac farm odontol ribeiro preto"] = "Revista da Faculdade de Farmacia e Odontologia de Ribeirao Preto",
  ["rev fac farm univ cent venez"] = "Revista de la Facultad de Farmacia, Universidad Central de Venezuela",
  ["rev fac med tucuman"] = "Revista. Facultad de Medicina de Tucumán",
  ["rev fac med univ nac auton mex"] = "Revista de la Facultad de Medicina, Universidad Nacional Autónoma de México",
  ["rev fac med univ nac colomb"] = "Revista de la Facultad de Medicina, Universidad Nacional de Colombia",
  ["rev fac med vet (lima)"] = "Revista de la Facultad de Medicina Veterinaria de la Universidad Nacional Mayor de San Marcos",
  ["rev fac med vet univ sao paulo"] = "Revista da Faculdade de Medicina Veterinária",
  ["rev fac nac agron medellin"] = "Revista Facultad Nacional De Agronomia Medellin",
  ["rev fac odontol aracatuba"] = "Revista da Faculdade de Odontologia de Aracatuba",
  ["rev fac odontol p alegre"] = "Revista da Faculdade de Odontologia, Porto Alegre",
  ["rev fac odontol pernambuco"] = "Revista da Faculdade de Odontologia de Pernambuco",
  ["rev fac odontol ribeiro preto"] = "Revista da Faculdade de Odontologia de Ribeirao Preto",
  ["rev fac odontol sao jose dos campos"] = "Revista da Faculdade de Odontologia de Sao Jose dos Campos",
  ["rev fac odontol sao paulo"] = "Revista da Faculdade de Odontologia da Universidade de Sao Paulo",
  ["rev fac odontol tucuman"] = "Revista de la Facultad de Odontología (San Miguel de Tucumán, Argentina)",
  ["rev fac odontol univ antioq"] = "Revista de la Facultad de Odontología Universidad de Antioquia",
  ["rev fac odontol univ chile"] = "Revista de la Facultad de Odontología (Santiago, Chile)",
  ["rev fac odontol univ fed bahia"] = "Revista da Faculdade de Odontologia da Universidade Federal da Bahia",
  ["rev fac odontol univ nac (cordoba)"] = "Revista de la Facultad de Odontologia",
  ["rev fac odontol univ nac colomb"] = "Revista de la Facultad de Odontologia de la Universidad Nacional de Colombia",
  ["rev faculdade odontol fzl"] = "Revista da Faculdade de Odontologia da FZL",
  ["rev faculdade odontol lins"] = "Revista da Faculdade de Odontologia de Lins",
  ["rev faith int aff"] = "The review of faith & international affairs",
  ["rev farm bioquim univ sao paulo"] = "Revista de farmácia e bioquímica da Universidade de São Paulo",
  ["rev farm cuba"] = "Revista farmacéutica de Cuba",
  ["rev farm odontol"] = "Revista de farmacia e odontologia",
  ["rev farm peru"] = "Revista farmacéutica peruana",
  ["rev farmacol med exp"] = "Revista de farmacología y medicina experimental",
  ["rev fed am health syst"] = "Review (Federation of American Health Systems)",
  ["rev fed am hosp"] = "Review - Federation of American Hospitals",
  ["rev fed odontol colomb"] = "Revista de la Federacion Odontologica Colombiana",
  ["rev fed odontol ecuat"] = "Revista de la Federacion Odontologica Ecuatoriana",
  ["rev finan econ"] = "Review of Financial Economics",
  ["rev finan stud"] = "Review of Financial Studies",
  ["rev financ"] = "Review of finance",
  ["rev financ stud"] = "The review of financial studies",
  ["rev fish biol fish"] = "Reviews in fish biology and fisheries",
  ["rev fish sci aquacult"] = "Reviews in Fisheries Science & Aquaculture",
  ["rev fitotec mex"] = "Revista Fitotecnia Mexicana",
  ["rev flora med"] = "Revista da flora medicinal",
  ["rev fluoresc"] = "Reviews in fluorescence",
  ["rev foa"] = "Revista F O A",
  ["rev foie"] = "La Revue du foie",
  ["rev fom soc"] = "Revista de fomento social",
  ["rev for fr"] = "Revue forestière française",
  ["rev fr aff soc"] = "Revue française des affaires sociales",
  ["rev fr allerg"] = "Revue française d'allergie",
  ["rev fr allergol"] = "Revue française d'allergologie",
  ["rev fr allergol (2009)"] = "Revue française d'allergologie (2009)",
  ["rev fr allergol immunol clin"] = "Revue française d'allergologie et d'immunologie clinique",
  ["rev fr endocrinol clin"] = "La Revue française d'endocrinologie clinique, nutrition, et métabolisme",
  ["rev fr endod"] = "Revue française d'endodontie : publication officielle de la Société française d'endodontie",
  ["rev fr etud am"] = "Revue française d'études américaines",
  ["rev fr etud clin biol"] = "Revue française d'études cliniques et biologiques",
  ["rev fr geotech"] = "Revue Francaise de Geotechnique",
  ["rev fr gerontol"] = "Revue française de gérontologie",
  ["rev fr gynecol obstet"] = "Revue française de gynécologie et d'obstétrique",
  ["rev fr hist livre"] = "Revue française d'histoire du livre",
  ["rev fr hist outre mer"] = "Revue française d'histoire d'outre-mer",
  ["rev fr mal respir"] = "Revue française des maladies respiratoires",
  ["rev fr odontostomatol"] = "Revue française d'odonto-stomatologie",
  ["rev fr oenol"] = "Revue Francaise d’Oenologie",
  ["rev fr prothes dent"] = "Revue française des prothésistes dentaires",
  ["rev fr prothese dent"] = "La Revue française de la prothèse dentaire",
  ["rev fr psychanal"] = "Revue française de psychanalyse",
  ["rev fr sci polit"] = "Revue française de science politique",
  ["rev fr sociol"] = "Revue française de sociologie",
  ["rev fr transfus"] = "Revue française de transfusion",
  ["rev fr transfus hemobiol"] = "Revue française de transfusion et d'hémobiologie : bulletin de la Société nationale de transfusion sanguine",
  ["rev fr transfus immunohematol"] = "Revue française de transfusion et immuno-hématologie",
  ["rev francaise serol chimiother"] = "Revue française de sérologie et de chimiothérapie",
  ["rev francoph lab"] = "Revue francophone des laboratoires : RFL",
  ["rev fuerzas armadas venez"] = "Fuerzas armadas de Venezuela (Caracas, Venezuela : 1971)",
  ["rev gastroenterol"] = "The Review of gastroenterology",
  ["rev gastroenterol disord"] = "Reviews in gastroenterological disorders",
  ["rev gastroenterol mex"] = "Revista de gastroenterología de México",
  ["rev gastroenterol peru"] = "Revista de gastroenterología del Perú : órgano oficial de la Sociedad de Gastroenterología del Perú",
  ["rev gaucha enferm"] = "Revista gaúcha de enfermagem",
  ["rev gaucha odontol"] = "Revista gaúcha de odontologia",
  ["rev gen bot"] = "Revue générale de Botanique",
  ["rev gen caoutch plast"] = "Revue Generale des Caoutchoucs & Plastiques",
  ["rev gen droit"] = "Revue générale de droit",
  ["rev gen mar"] = "Revista general de marina",
  ["rev gen med chir union fr"] = "Revue générale de médecine & de chirurgie de l'Union française",
  ["rev gen nucl"] = "Revue Generale Nucleaire",
  ["rev gen psychol"] = "Review of general psychology : journal of Division 1, of the American Psychological Association",
  ["rev gen sci pures appl bull assoc fr av sci"] = "Revue générale des sciences pures et appliquées et bulletin de l'Association française pour l'avancement des sciences",
  ["rev gen therm"] = "Revue Generale de Thermique",
  ["rev geofis"] = "Revista Geofisica",
  ["rev geogr alp"] = "Revue de géographie alpine",
  ["rev geogr est"] = "Revue géographique de l'Est",
  ["rev geogr inst panam geogr hist"] = "Revista geográfica del Instituto Panamericano de Geografía e Historia",
  ["rev geogr lyon"] = "Revue de géographie de Lyon",
  ["rev geogr montr"] = "La Revue de géographie de Montréal",
  ["rev geogr pyren sud ouest"] = "Revue géographique des Pyrénées et du Sud-Ouest",
  ["rev geol chile"] = "Revista Geologica de Chile",
  ["rev geophys"] = "Reviews of Geophysics",
  ["rev geophys space phys"] = "Reviews of Geophysics and Space Physics",
  ["rev gerontol expr fr"] = "Revue de gérontologic d'expression française",
  ["rev ginecol obstet (sao paulo)"] = "Revista de ginecologia e d'obstetrícia",
  ["rev goiana med"] = "Revista Goiana de medicina",
  ["rev griot"] = "Revista Griot",
  ["rev guatem estomatol"] = "Revista guatemalteca de estomatología",
  ["rev gynecol obstet"] = "Revista De Ginecologia Y Obstetricia",
  ["rev haute auvergne"] = "Revue de la Haute-Auvergne",
  ["rev health care"] = "Reviews in health care",
  ["rev hematol"] = "Revue d'hématologie",
  ["rev high ed"] = "The review of higher education",
  ["rev hist (arg)"] = "Revista de historia",
  ["rev hist (brazil)"] = "Revista de história",
  ["rev hist (costa rica)"] = "Revista de historia",
  ["rev hist (paris)"] = "Revue historique (Paris, France)",
  ["rev hist am argent"] = "Revista de historia americana y argentina",
  ["rev hist am fr"] = "Revue d'histoire de l'Amérique française",
  ["rev hist armees"] = "Revue historique des armées",
  ["rev hist art dent"] = "Revue d'histoire de l'art dentaire",
  ["rev hist bordeaux dep gironde"] = "Revue historique de Bordeaux et du département de la Gironde",
  ["rev hist canaria"] = "Revista de historia canaria",
  ["rev hist demogr"] = "Review of historical demography",
  ["rev hist deuxieme guerre mond"] = "Revue d'histoire de la deuxième guerre mondiale",
  ["rev hist eccles"] = "Revue d'histoire ecclésiastique",
  ["rev hist econ soc"] = "Revue d'histoire économique et sociale",
  ["rev hist eglise fr"] = "Revue d'histoire de l'Église de France",
  ["rev hist jeronimo zurita"] = "Revista de historia Jerónimo Zurita",
  ["rev hist litt fr"] = "Revue d'histoire littéraire de la France",
  ["rev hist maghreb"] = "Revue d'histoire maghrebine",
  ["rev hist med heb"] = "Revue d'histoire de la médecine hébraique",
  ["rev hist mil"] = "Revista de historia militar",
  ["rev hist mod contemp"] = "Revue d'histoire moderne et contemporaine",
  ["rev hist naval"] = "Revista de historia naval",
  ["rev hist pharm (paris)"] = "Revue d'histoire de la pharmacie",
  ["rev hist philos relig"] = "Revue d'histoire et de philosophie religieuses",
  ["rev hist psicol"] = "Revista de historia de la psicología",
  ["rev hist relig"] = "Revue de l'histoire des religions",
  ["rev hist sci"] = "Revue d’Histoire des Sciences",
  ["rev hist sci (paris)"] = "Revue d'histoire des sciences et de leurs applications",
  ["rev hist sci [paris]"] = "Revue d’histoire des sciences",
  ["rev hist sci paris"] = "Revue d'histoire des sciences",
  ["rev hist textes"] = "Revue d'histoire des textes",
  ["rev hist theatre"] = "Revue d'histoire du théâtre",
  ["rev histoire math"] = "Revue d’Histoire des Mathematiques. Journal for the History of Mathematics",
  ["rev histoire sci"] = "Revue d’Histoire des Sciences",
  ["rev hosp clin fac med sao paulo"] = "Revista do Hospital das Clínicas",
  ["rev hosp nino (lima)"] = "Revista del Hospital del Niño",
  ["rev hum factors ergon"] = "Review of human factors and ergonomics",
  ["rev hyg med soc"] = "Revue d'hygiène et de médecine sociale",
  ["rev hyg prof"] = "Revue de l'hygiène professionnelle",
  ["rev iber aracnol"] = "Revista ibérica de aracnología",
  ["rev iber endocrinol"] = "Revista ibérica de endocrinología",
  ["rev iber parasitol"] = "Revista ibérica de parasitología",
  ["rev iberoam autom inf ind"] = "RIAI Revista Iberoamericana de Automatica e Informatica Industrial",
  ["rev iberoam bioet"] = "Revista iberoamericana de bioética",
  ["rev iberoam estud desarro"] = "Revista iberoamericana de estudios de desarrollo = Iberoamerican journal of development studies",
  ["rev iberoam micol"] = "Revista iberoamericana de micología",
  ["rev iberoam ortod"] = "Revista ibero-americana de ortodoncia : publicación oficial de la Asociación Ibero-Americana Ortodoncistas",
  ["rev ibys"] = "Revista IBYS",
  ["rev ieee am lat"] = "Revista IEEE America Latina",
  ["rev ig bacteriol virusol parazitol epidemiol pneumoftiziol bacteriol virusol parazitol epi"] = "Revista de Igiena, Bacteriologie, Virusologie, Parazitologie, Epidemiologie, Pneumoftiziologie. Bacteriologia, Virusologia, Parazitologia, Epidemiologia",
  ["rev ig bacteriol virusol parazitol epidemiol pneumoftiziol bacteriol virusol parazitol epidemiol"] = "Revista de igienă, bacteriologie, virusologie, parazitologie, epidemiologie, pneumoftiziologie. Bacteriologia, virusologia, parazitologia, epidemiologia",
  ["rev ig bacteriol virusol parazitol epidemiol pneumoftiziol pneumoftiziol"] = "Revista de igienă, bacteriologie, virusologie, parazitologie, epidemiologie, pneumoftiziologie. Pneumoftiziologia",
  ["rev ig med muncii med soc bacteriol virusol parazitol epidemiol pneumoftiziol pneumoftiziol"] = "Revista de igienă, medicina muncii, medicină socială, bacteriologie, virusologie, parazitologie, epidemiologie, pneumoftiziologie. Pneumoftiziologie",
  ["rev immunogenet"] = "Reviews in immunogenetics",
  ["rev immunol (paris)"] = "Revue d'immunologie",
  ["rev immunol ther antimicrob"] = "Revue d'immunologie et de thérapie antimicrobienne",
  ["rev income wealth"] = "The Review of income and wealth",
  ["rev ind organ"] = "Review of Industrial Organization",
  ["rev indias"] = "Revista de Indias",
  ["rev inf legis"] = "Revista de informação legislativa",
  ["rev infect"] = "Reviews in infection",
  ["rev infect dis"] = "Reviews of infectious diseases",
  ["rev infirm"] = "Revue de l'infirmière",
  ["rev infirm [inf]"] = "Revue de l’Infirmiere. Informations",
  ["rev infirm assist soc"] = "Revue de l'infirmière et de l'assistante sociale",
  ["rev infirm inf"] = "Revue de l'infirmière. Informations",
  ["rev infirm infirm aux que"] = "La Revue des infirmières et infirmiers auxiliaires du Québec",
  ["rev inform med ter"] = "Revista de información médico-terapéutica",
  ["rev ing biomed"] = "Revista ingeniería biomédica",
  ["rev ing univ medellin"] = "Revista Ingenierias Univsidad de Medellin",
  ["rev inorg chem"] = "Reviews in Inorganic Chemistry",
  ["rev inst adolfo lutz"] = "Revista do Instituto Adolfo Lutz",
  ["rev inst antibiot (recife)"] = "Revista do Instituto de Antibioticos, Universidade Federal de Pernambuco",
  ["rev inst eur"] = "Revista de instituciones europeas",
  ["rev inst hist geogr bras"] = "Revista do Instituto Historico e Geographico Brazileiro (1906)",
  ["rev inst hyg mines (hasselt)"] = "Revue de l'Institut d'hygiène des mines",
  ["rev inst med trop s paulo"] = "Revista do Instituto de Medicina Tropical de São Paulo",
  ["rev inst med trop sao paulo"] = "Revista do Instituto de Medicina Tropical de São Paulo",
  ["rev inst mex ing chim"] = "Revista del Instituto Mexicano de Ingenieros Quimicos",
  ["rev inst mex pet"] = "Revista del Instituto Mexicano del Petroleo",
  ["rev inst nac cancerol (mex)"] = "Revista del Instituto Nacional de Cancerologia",
  ["rev inst napoleon"] = "Revue de l'Institut Napoléon (Paris, France : 1954)",
  ["rev inst salubr enferm trop"] = "Revista del Instituto de Salubridad y Enfermedades Tropicales",
  ["rev inst sociol"] = "Revue de l'Institut de sociologie",
  ["rev int androl"] = "Revista internacional de andrología",
  ["rev int cienc deporte"] = "Revista internacional de ciencias del deporte",
  ["rev int comm jurists"] = "Review (International Commission of Jurists (1952- )",
  ["rev int criminol police tech"] = "Revue internationale de criminologie et de police technique",
  ["rev int croix rouge"] = "Revue internationale de la Croix-rouge",
  ["rev int econ"] = "Review of international economics",
  ["rev int hepatol"] = "Revue internationale d'hépatologie",
  ["rev int hist mil"] = "Revue internationale d'histoire militaire",
  ["rev int hist psychanal"] = "Revue internationale d'histoire de la psychanalyse",
  ["rev int hist psychiatr"] = "Revue internationale d'histoire de la psychiatrie",
  ["rev int metod numer calc diseno ing"] = "Revista Internacional de Metodos Numericos para Calculo y Diseno en Ingenieria",
  ["rev int metodos numer calc diseno ing"] = "Revista Internacional de Metodos Numericos para Calculo y Diseno en Ingenieria",
  ["rev int organ"] = "Review of International Organizations",
  ["rev int pediatr"] = "Revue internationale de pédiatrie",
  ["rev int philos"] = "Revue internationale de philosophie",
  ["rev int polit econ"] = "Review of international political economy",
  ["rev int polit economy"] = "Review of International Political Economy",
  ["rev int psicol ter psicol"] = "Revista internacional de psicología y terapia psicológica = International journal of psychology and psychological therapy",
  ["rev int serv sante armees"] = "Revue international des services de santé des armées de terre, de mer et de l'air",
  ["rev int serv sante forces armees"] = "Revue internationale des services de santé des forces armées : organe du Comité international de médecine et de pharmacie militaires = International review of the armed forces medical services : organ of the International Committee of Military Medicine and Pharmacy",
  ["rev int sociol"] = "Revista internacional de sociología",
  ["rev int stud"] = "Review of international studies",
  ["rev int trach"] = "Revue internationale du trachome. International review of trachoma",
  ["rev int trach pathol ocul trop subtrop"] = "Revue internationale du trachome et de pathologie oculaire tropicale et subtropicale : organe de la Ligue contre le trachome avec la collaboration de l'International Organization against Trachoma et des organisations nationales et internationales de santé publique",
  ["rev int trach pathol ocul trop subtrop sante publique"] = "Revue internationale du trachome et de pathologie oculaire tropicale et subtropicale et de santé publique : organe de la Ligue contre le trachome avec la collaboration de l'International Organization against Trachoma et des organisations nationales et internationales de santé publique",
  ["rev integr temas mat"] = "Revista Integracion. Temas de Matematicas",
  ["rev interam bibliogr"] = "Revista interamericana de bibliografía = Review of inter-American bibliography",
  ["rev interam planif"] = "Revista interamericana de planificación",
  ["rev interam radiol"] = "Revista interamericana de radiología",
  ["rev internac métod numér cálc diseñ ingr"] = "Revista Internacional de Métodos Numéricos para Cálculo y Diseño en Ingeniería",
  ["rev invest (guadalajara)"] = "Revista de investigación",
  ["rev invest clin"] = "Revista de investigación clínica; organo del Hospital de Enfermedades de la Nutrición",
  ["rev invest salud publica"] = "Revista de investigación en salud pública",
  ["rev investig mar"] = "Revista de investigaciones marinas",
  ["rev investig vet peru"] = "Revista de investigaciones veterinarias del Perú (Online)",
  ["rev islenha"] = "Islenha",
  ["rev istor"] = "Revista de istorie",
  ["rev j autism dev disord"] = "Review journal of autism and developmental disorders",
  ["rev j chem"] = "Review journal of chemistry",
  ["rev jurid univ p r"] = "Revista jurídica de la Universidad de Puerto Rico",
  ["rev kairos"] = "Revista Kairós : gerontologia",
  ["rev kinesither"] = "Revue de kinésithérapie",
  ["rev kuba med trop parasitol"] = "Revista Kuba de medicina tropical y parasitología",
  ["rev lang rom"] = "Revue des langues romanes",
  ["rev laryngol otol rhinol (bord)"] = "Revue de laryngologie - otologie - rhinologie",
  ["rev lasallista investig"] = "Revista lasallista de investigacion",
  ["rev lat am enfermagem"] = "Revista latino-americana de enfermagem",
  ["rev lat sociol"] = "Revista latina de sociología",
  ["rev latinoam anat patol"] = "Revista latinoamericana de anatomía patológica",
  ["rev latinoam bioet"] = "Revista latinoamericana de bioética",
  ["rev latinoam cir plast"] = "Revista latinoamericana de cirugía plástica",
  ["rev latinoam etnomat"] = "Revista Latinoamericana de Etnomatematica. Perspectivas Socioculturales de la Educacion Matematica",
  ["rev latinoam ing quim quim apl"] = "Revista Latinoamericana de Ingenieria Quimica y Quimica Aplicada",
  ["rev latinoam microbiol"] = "Revista Latinoamericana de Microbiologia",
  ["rev latinoam microbiol (1958)"] = "Revista latinoamericana de microbiología",
  ["rev latinoam microbiol parasitol (mex)"] = "Revista latinoamericana de microbiología y parasitología",
  ["rev latinoam patol"] = "Revista latinoamericana de patología",
  ["rev latinoam perinatol"] = "Revista latinoamericana de perinatología : órgano oficial de la Federación Latinoamericana de Asociaciones de Medicina [Perinatal] (F.L.A.M.P.)",
  ["rev latinoam psicol"] = "Revista latinoamericana de psicología",
  ["rev latinoam quim"] = "Revista Latinoamericana de Quimica",
  ["rev law soc change"] = "Review of law and social change. New York University",
  ["rev louisiane"] = "Revue de Louisiane. Louisiana review",
  ["rev louvre mus fr"] = "La Revue du Louvre et des musées de France",
  ["rev lyon med"] = "La Revue lyonnaise de médecine",
  ["rev mal respir"] = "Revue des maladies respiratoires",
  ["rev malad respir actual"] = "Revue des maladies respiratoires actualités",
  ["rev mali infect microbiol"] = "Revue malienne d'infectiologie et de microbiologie",
  ["rev mar"] = "Revista de marina (Valparaíso, Chile)",
  ["rev marketing agr econ"] = "Review of Marketing and Agricultural Economics",
  ["rev mat apl"] = "Revista de Matemáticas Aplicadas",
  ["rev mat complut"] = "Revista matemática complutense",
  ["rev mat complutense"] = "Revista Matematica Complutense",
  ["rev mat estatíst"] = "Revista de Matemática e Estatística",
  ["rev mat iberoam"] = "Revista Matematica Iberoamericana",
  ["rev mat iberoamericana"] = "Revista Matemática Iberoamericana",
  ["rev mat teor apl"] = "Revista de Matematica. Teoria y Aplicaciones",
  ["rev mat univ complut madrid"] = "Revista Matemática de la Universidad Complutense de Madrid",
  ["rev math math phys"] = "Reviews in Mathematics and Mathematical Physics",
  ["rev math phys"] = "Reviews in Mathematical Physics",
  ["rev med"] = "Revista médica. Ateneo Ramón y Cajal, Mexico",
  ["rev med (mex)"] = "Revista médica (Instituto Mexicano del Seguro Social)",
  ["rev med (paris)"] = "La Revue de médecine",
  ["rev med (puebla)"] = "Revista médica",
  ["rev med (rio j)"] = "Revista de medicina",
  ["rev med aeron"] = "Revista médica da aeronáutica",
  ["rev med aeron braz"] = "Revista médica da aeronáutica do Brasil",
  ["rev med aeronaut"] = "Revue de médecine aéronautique",
  ["rev med aliment"] = "Revista de medicina y alimentación",
  ["rev med alpes fr"] = "Revue médicale des alpes françaises",
  ["rev med bahia"] = "Revista médica da Bahia",
  ["rev med bras"] = "Revista médica brasileira",
  ["rev med brux"] = "Revue Medicale de Bruxelles",
  ["rev med cancerol"] = "Revista de medicina, cancerología",
  ["rev med chil"] = "Revista médica de Chile",
  ["rev med chile supl"] = "Revista médica de Chile. Suplemento",
  ["rev med chir mal foie"] = "Revue médico-chirurgicale des maladies du foie",
  ["rev med chir soc med nat iasi"] = "Revista medico-chirurgicală̆ a Societă̆ţ̜ii de Medici ş̧i Naturaliş̧ti din Iaş̧i",
  ["rev med cienc afines"] = "Revista de medicina y ciencias afines",
  ["rev med cir"] = "Revista de medicina y cirugía",
  ["rev med cir la habana"] = "Revista de medicina y cirugia de la Habana",
  ["rev med cir sao paulo"] = "Revista de medicina e cirurgia de São Paulo",
  ["rev med cordoba"] = "Revista médica de Córdoba",
  ["rev med costa rica"] = "Revista médica de Costa Rica",
  ["rev med cubana"] = "Revista médica cubana",
  ["rev med dominic"] = "Revista médica dominicana",
  ["rev med fr"] = "Revue médicale de France",
  ["rev med hondur"] = "Revista médica hondureña",
  ["rev med hosp britanico"] = "Revista médica del Hospital Británico",
  ["rev med hosp cent empl"] = "Revista médica del Hospital Central del Empleado",
  ["rev med hosp gen (mex)"] = "Revista médica del Hospital General",
  ["rev med hosp obrero"] = "Revista médica del Hospital Obrero",
  ["rev med hyg outre mer"] = "Revue de médecine et d'hygiène d'Outre-Mer",
  ["rev med inst mex seguro soc"] = "Revista médica del Instituto Mexicano del Seguro Social",
  ["rev med int photo cinema telev"] = "Revue médicale internationale de photo, cinéma, télévision. International journal of medical photography, cinematography & television",
  ["rev med interna neurol psihiatr neurochir dermatovenerol med intern"] = "Revista de Medicina Interna, Neurologie, Psihiatrie, Neurochirurgie, Dermato-Venerologie. Medicina Interna",
  ["rev med interna neurol psihiatr neurochir dermatovenerol med interna"] = "Revista de medicină internă, neurologe, psihiatrie, neurochirurgie, dermato-venerologie. Medicină internă",
  ["rev med interna neurol psihiatr neurochir dermatovenerol neurol psihiatr neurochir"] = "Revista de medicină internă, neurologie, psihiatrie, neurochirurgie, dermato-venerologie. Neurologie, psihiatrie, neurochirurgie",
  ["rev med interne"] = "La Revue de médecine interne",
  ["rev med ipss"] = "Revista médica del IPSS",
  ["rev med latinoam"] = "Revista médica latino-americana",
  ["rev med leg"] = "Revista de medicina legal",
  ["rev med leg colomb"] = "Revista de medicina legal de Colombia",
  ["rev med liege"] = "Revue médicale de Liège",
  ["rev med louvain"] = "Revue médicale de Louvain",
  ["rev med microbiol"] = "Reviews in medical microbiology : a journal of the Pathological Society of Great Britain and Ireland",
  ["rev med mil"] = "Revista de medicina militar",
  ["rev med miniere"] = "Revue médicale minière",
  ["rev med moamb"] = "Revista médica de Moçambique",
  ["rev med moyen orient"] = "Revue médicale du Moyen-Orient",
  ["rev med munic"] = "Revista médica municipal",
  ["rev med nancy"] = "Revue médicale de Nancy",
  ["rev med nav metrop o-m"] = "Revue de médecine navale (métropole et outre-mer); travaux scientifiques des médecins et pharmaciens-chimistes de la marine",
  ["rev med normandes"] = "Les Revues médicales normandes",
  ["rev med panama"] = "Revista Medica de Panama",
  ["rev med parag"] = "Revista médica del Paraguay",
  ["rev med pasteur"] = "Revista médica Pasteur",
  ["rev med pernamb"] = "Revista médica de Pernambuco",
  ["rev med peru"] = "Revista médica peruana",
  ["rev med picardie"] = "La Revue médicale de Picardie",
  ["rev med psychosom"] = "Revue de médecine psychosomatique",
  ["rev med psychosom psychol med"] = "Revue de médecine psychosomatique et de psychologie médicale",
  ["rev med quir oriente"] = "Revista médico-quirúrgica de Oriente",
  ["rev med quir patol fem"] = "Revista médico-quirúrgica de patología femenina",
  ["rev med rio gd sul"] = "Revista de medicina do Rio Grande do Sul",
  ["rev med rosario"] = "Revista médica de Rosario",
  ["rev med suisse"] = "Revue médicale suisse",
  ["rev med suisse romande"] = "Revue Medicale de la Suisse Romande",
  ["rev med toulouse"] = "Revue de médecine de Toulouse",
  ["rev med tours"] = "Revue de médecine de Tours",
  ["rev med trop parasitol bacteriol clin lab"] = "Revista de medicina tropical y parasitología, bacteriología, clínica y laboratorio",
  ["rev med univ navarra"] = "Revista de medicina de la Universidad de Navarra",
  ["rev med urug (montev)"] = "La Revista médica del Uruguay",
  ["rev med valparaiso"] = "Revista médica de Valparaíso",
  ["rev med veracruzana"] = "Revista médica veracruzana",
  ["rev med vet (b aires)"] = "Revista de medicina veterinaria",
  ["rev med vet (bogota)"] = "Revista de medicina veterinaria",
  ["rev med vet (toulouse)"] = "Revue de médecine vétérinaire",
  ["rev med virol"] = "Reviews in medical virology",
  ["rev med xalapena"] = "Revista médica xalapeña",
  ["rev med yucatan"] = "La Revista médica de Yucatán",
  ["rev medica hosp esp"] = "Revista médica del Hospital Español",
  ["rev medica minas"] = "Revista médica do sul de Minas",
  ["rev mediterr sci med"] = "Revue méditerranéenne des sciences médicales",
  ["rev mens mus collect ville genève"] = "Revue mensuelle des Musées et Collections de la ville Genève",
  ["rev metal"] = "Revista de Metalurgia (Madrid, Spain)",
  ["rev metall / cah inf tech"] = "Revue de Metallurgie / Cahiers d’Informations Techniques",
  ["rev metaphys"] = "The Review of metaphysics",
  ["rev metaphys morale"] = "Revue de métaphysique et de morale (Paris, France : 1945)",
  ["rev meteor"] = "Revista Meteorologica",
  ["rev mex anal conducta"] = "Revista mexicana de análisis de la conducta = Mexican journal of behavior analysis",
  ["rev mex astron astrofis"] = "Revista Mexicana de Astronomia y Astrofisica",
  ["rev mex biodivers"] = "Revista Mexicana de biodiversidad",
  ["rev mex cienc farm"] = "Revista Mexicana de Ciencias Farmaceuticas",
  ["rev mex cienc for"] = "Revista mexicana de ciencias forestales",
  ["rev mex cienc med bio"] = "Revista mexicana de ciencias médicas y biológicas",
  ["rev mex cienc pecu"] = "Revista Mexicana de ciencias pecuarias",
  ["rev mex cienc polit soc"] = "Revista mexicana de ciencias políticas y sociales",
  ["rev mex cir ginecol cancer"] = "Revista mexicana de cirugía, ginecología y cáncer",
  ["rev mex de cienc agric"] = "Revista mexicana de ciencias agrícolas",
  ["rev mex fis"] = "Revista Mexicana de Fisica",
  ["rev mex fis e"] = "Revista Mexicana de Fisica E",
  ["rev mex fitopatol"] = "Revista mexicana de fitopatologia : la revista oficial de la Sociedad Mexicana de Fitopatologia",
  ["rev mex ing biomed"] = "Revista Mexicana De Ingenieria Biomedica",
  ["rev mex ing chim"] = "Revista Mexicana de Ingenieria Quimica",
  ["rev mex ing quim"] = "Revista mexicana de ingeniería química",
  ["rev mex micol"] = "Revista mexicana de micología : órgano oficial de la Sociedad Mexicana de Micología",
  ["rev mex sociol"] = "Revista mexicana de sociología",
  ["rev mex trastor aliment"] = "Revista Mexicana de trastornos alimentarios = Mexican journal of eating disorders",
  ["rev mex tuberc enferm apar respir"] = "Revista mexicana de tuberculosis y enfermedades del aparato respiratorio",
  ["rev mex urol"] = "Revista mexicana de urología",
  ["rev mexicana fis"] = "Revista Mexicana de Fisica",
  ["rev mexicana fís"] = "Revista Mexicana de Física",
  ["rev micropaleontol"] = "Revue de Micropaleontologie",
  ["rev micropaleontol/font>"] = "Revue de micropaléontologie",
  ["rev mil remonta vet"] = "Revista militar de remonta e veterinaria",
  ["rev mineral geochem"] = "Reviews in Mineralogy and Geochemistry",
  ["rev mod logic"] = "Review of Modern Logic",
  ["rev mod phys"] = "Reviews of Modern Physics",
  ["rev mod plasma phys"] = "Reviews of modern plasma physics",
  ["rev modern phys"] = "Reviews of Modern Physics",
  ["rev mol biotechnol"] = "Reviews in Molecular Biotechnology",
  ["rev monde musulman mediterr"] = "Revue du monde musulman et de la Méditerranée",
  ["rev mus argent cienc nat nueva ser"] = "Revista del Museo Argentino de Ciencias Naturales Nueva Serie",
  ["rev mus nac"] = "Revista del Museo Nacional. Museo Nacional (Peru)",
  ["rev museo fac odontol b aires"] = "Revista del Museo de la Facultad de Odontologia de Buenos Aires",
  ["rev mvz cordoba"] = "Revista MVZ Córdoba",
  ["rev neurol"] = "Revista de neurologia",
  ["rev neurol (paris)"] = "Revue neurologique",
  ["rev neurol b aires"] = "Revista neurologica de Buenos Aires",
  ["rev neurol dis"] = "Reviews in neurological diseases",
  ["rev neuropsiquiatr"] = "Revista de neuro-psiquiatría",
  ["rev neuropsychiatr infant"] = "Revue de neuropsychiatrie infantile et d'hygiène mentale de l'enfance",
  ["rev neurosci"] = "Reviews in the neurosciences",
  ["rev nonlinear dyn complex"] = "Reviews of Nonlinear Dynamics and Complexity",
  ["rev nord"] = "Revue du Nord",
  ["rev nouv"] = "Revue nouvelle (Brussels, Belgium)",
  ["rev numis"] = "Revue numismatique",
  ["rev obras sanit nac b aires"] = "Revista de Obras Sanitarias de la Nación. Obras Sanitarias de la Nación (Argentina)",
  ["rev obs filos"] = "Revista observaciones filosóficas",
  ["rev obstet ginecol venez"] = "Revista de obstetricia y ginecología de Venezuela",
  ["rev obstet gynecol"] = "Reviews in obstetrics & gynecology",
  ["rev occident musulman mediterr"] = "Revue de l'Occident musulman et de la Méditerranée",
  ["rev occidente"] = "Revista de occidente (Madrid, Spain : 1923)",
  ["rev oculomot res"] = "Reviews of oculomotor research",
  ["rev odontoestomatol"] = "Revista odonto-estomatológica",
  ["rev odontoimplantol"] = "Revue odonto-implantologique",
  ["rev odontol"] = "La Revue odontologique",
  ["rev odontol (b aires)"] = "Revista odontologica",
  ["rev odontol (cordoba)"] = "Revista Odontologica; Facultad De Odontologia, Universidad Nacional De Cordoba",
  ["rev odontol (la paz)"] = "Revista Odontologica",
  ["rev odontol circ odontol parag"] = "Revista odontológica",
  ["rev odontol concepcion"] = "Revista odontológica de Concepción",
  ["rev odontol costa rica"] = "Revista odontológica de Costa Rica",
  ["rev odontol ecuat"] = "Revista odontológica ecuatoriana",
  ["rev odontol p r"] = "Revista odontológica de Puerto Rico",
  ["rev odontol parana"] = "Revista odontológica do Paraná",
  ["rev odontol st catarina"] = "Revista De Odontologia Da Universidade Federal De Santa Catarina",
  ["rev odontol unesp"] = "Revista de odontologia da UNESP",
  ["rev odontol unicid"] = "Revista de odontologia da UNICID",
  ["rev odontol univ sao paulo"] = "Revista de odontologia da Universidade de São Paulo",
  ["rev odontostomatol"] = "Revue d’odonto-stomatologie",
  ["rev odontostomatol (paris)"] = "Revue d'odonto-stomatologie",
  ["rev odontostomatol midi fr"] = "Revue d'odonto-stomatologie du midi de la France",
  ["rev odontostomatol nordest"] = "Revue odonto stomatologique du Nord-Est",
  ["rev oim sobre migr am lat"] = "Revista de la OIM sobre migraciones en América Latina",
  ["rev opt theor instrum"] = "Revue dOptique, Théorique et Instrumentale",
  ["rev ordem med"] = "Revista (Ordem dos Médicos (Portugal))",
  ["rev organon"] = "Revista Organon",
  ["rev orthop chir appar mot"] = "Revue d'orthopédie et de chirurgie de l'appareil moteur",
  ["rev orthop dento faciale"] = "Revue d'orthopédie dento-faciale",
  ["rev ortop traumatol ed lat am"] = "Revista de ortopedia y traumatología. Edición latino americana",
  ["rev oto neuro oftalmol cir neurol sud am"] = "Revista oto-neuro-oftalmológica y cirugía neurologica sudamericana",
  ["rev otoneuroophtalmol"] = "Revue d'oto-neuro-ophtalmologie",
  ["rev otorrinolaringol"] = "Revista de otorrinolaringología",
  ["rev ouie"] = "Revue de l'ouïe",
  ["rev pain"] = "Reviews in pain",
  ["rev palaeobot palynol"] = "Review of palaeobotany and palynology",
  ["rev palais decouv"] = "Revue. Palais de la découverte (Paris, France)",
  ["rev paleobiol"] = "Revue de Paleobiologie",
  ["rev palud med trop"] = "Revue du paludisme et de médecine tropicale",
  ["rev panam med cir torax"] = "Revista panamericana de medicina y cirugía del tórax",
  ["rev panam salud publica"] = "Revista panamericana de salud pública = Pan American journal of public health",
  ["rev panamazonica saude"] = "Revista pan-amazônica de saúde",
  ["rev pathol comp"] = "Revue de pathologie comparée",
  ["rev pathol gen physiol clin"] = "Revue de pathologie générale et de physiologie clinique",
  ["rev patient focus care assoc"] = "Review (Patient Focused Care Association)",
  ["rev pau bearn"] = "Revue de Pau et du Béarn",
  ["rev paul endodontia"] = "Revista Paulista De Endodontia",
  ["rev paul enferm"] = "Revista paulista de enfermagem",
  ["rev paul hosp"] = "Revista paulista de hospitais",
  ["rev paul med"] = "Revista paulista de medicina",
  ["rev paul odontol"] = "Revista paulista de odontologia",
  ["rev paul pediatr"] = "Revista paulista de pediatria : orgão oficial da Sociedade de Pediatria de São Paulo",
  ["rev paul tisiol torax"] = "Revista paulista de tisiologia e do tórax",
  ["rev pediatr"] = "La Revue de pédiatrie",
  ["rev pediatr obstet ginecol obstet ginecol"] = "Revista de pediatrie, obstetrică si ginecologie. Obstetrică si ginecologie",
  ["rev pediatr obstet ginecol pediatr"] = "Revista de pediatrie, obstetrică şi ginecologie. Pediatria",
  ["rev pernambucana odontol"] = "Revista pernambucana de odontologia",
  ["rev peru biol"] = "Revista peruana de biología",
  ["rev peru med exp salud publica"] = "Revista peruana de medicina experimental y salud pública",
  ["rev peru pediatr"] = "Revista peruana de pediatría",
  ["rev peru poblac"] = "Revista peruana de población",
  ["rev peru quim ing quim"] = "Revista Peruana de Quimica e Ingenieria Quimica",
  ["rev philol litt hist anc"] = "Revue de philologie, de littérature et d'histoire anciennes",
  ["rev philos france let"] = "Revue philosophique de la France et de l'étranger",
  ["rev philos louv"] = "Revue philosophique de Louvain",
  ["rev philos psychol"] = "Review of philosophy and psychology",
  ["rev phys"] = "Reviews in Physics",
  ["rev phys chem jpn"] = "Review of Physical Chemistry of Japan",
  ["rev phys technol"] = "Review of Physics in Technology",
  ["rev physiol biochem pharmacol"] = "Reviews of physiology, biochemistry and pharmacology",
  ["rev physiol, biochem pharmacol"] = "Reviews of Physiology Biochemistry and Pharmacology",
  ["rev phytother (paris)"] = "Revue de phytothérapie",
  ["rev planeac desarro"] = "Revista de planeación y desarrollo",
  ["rev plant pathol"] = "Review of Plant Pathology",
  ["rev plast mod"] = "Revista de Plasticos Modernos",
  ["rev pneumol clin"] = "Revue de pneumologie clinique",
  ["rev pol acad sci"] = "The review of the Polish Academy of Sciences. Polska Akademia Nauk",
  ["rev policlin caracas"] = "Revista. Caracas (Venezuela). Policlínica Caracas",
  ["rev policy res"] = "The review of policy research",
  ["rev polit"] = "The Review of politics",
  ["rev polit economy"] = "Review of Political Economy",
  ["rev port cardiol"] = "Revista portuguesa de cardiologia : orgão oficial da Sociedade Portuguesa de Cardiologia = Portuguese journal of cardiology : an official journal of the Portuguese Society of Cardiology",
  ["rev port cir cardiotorac vasc"] = "Revista portuguesa de cirurgia cardio-torácica e vascular : órgão oficial da Sociedade Portuguesa de Cirurgia Cardio-Torácica e Vascular",
  ["rev port estomatol cir maxilofac"] = "Revista portuguesa de estomatologia e cirurgia maxilo-facial",
  ["rev port farm"] = "Revista portuguesa de farmácia",
  ["rev port hist"] = "Revista portuguesa de história",
  ["rev port med mil"] = "Revista portuguesa de medicina militar",
  ["rev port pediatr"] = "Revista portuguesa de pediatria",
  ["rev port pediatr pueric"] = "Revista portuguesa de pediatria e puericultura",
  ["rev port pneumol"] = "Revista portuguesa de pneumologia",
  ["rev port quim"] = "Revista Portuguesa de Quimica",
  ["rev prat"] = "La Revue du praticien",
  ["rev prog color relat top"] = "Review of Progress in Coloration and Related Topics",
  ["rev prot veg"] = "Revista de protección vegetal",
  ["rev psicoanal"] = "Revista de psicoanálisis",
  ["rev psicol gen apl"] = "Revista de psicología general y aplicada",
  ["rev psicol soc"] = "Revista de psicologia social",
  ["rev psiquiatr clín"] = "Revista de psiquiatria clínica",
  ["rev psiquiatr psicol med eur am lat"] = "Revista de psiquiatría y psicología médica de Europa y América latinas",
  ["rev psiquiatr salud ment"] = "Revista de psiquiatrí́a y salud mental",
  ["rev psychoeduc"] = "Revue de psychoéducation",
  ["rev psychol appl"] = "Revue de psychologie appliquée",
  ["rev publ nav"] = "Revista de publicaciones navales; suplemento de Sanidad naval",
  ["rev public data use"] = "Review of public data use",
  ["rev public pers adm"] = "Review of public personnel administration",
  ["rev puertoriq med salud publica"] = "Revista Puertorriqueña de medicina y salud pública",
  ["rev puertorriquena psicol"] = "Revista Puertorriqueña de psicología",
  ["rev pure appl pharmacol sci"] = "Reviews in pure & applied pharmacological sciences",
  ["rev quant finance acc"] = "Review of Quantitative Finance and Accounting",
  ["rev quest sci"] = "Revue des questions scientifiques",
  ["rev questions sci"] = "Revue des Questions Scientifiques",
  ["rev quim farm"] = "Revista de química e farmácia",
  ["rev quim text"] = "Revista de Quimica Textil",
  ["rev quir esp"] = "Revista quirúrgica española : RQE",
  ["rev r acad cienc exactas fis nat (esp)"] = "Revista de la Real Academia de Ciencias Exactas, Fisicas y Naturales (Espana)",
  ["rev r acad cienc exactas fis nat a mat"] = "Revista de la Real Academia de Ciencias Exactas, Físicas y Naturales. Serie A, Matemáticas",
  ["rev r acad cienc exactas fis nat ser a mat racsam"] = "Revista de la Real Academia de Ciencias Exactas, Fisicas y Naturales. Serie A. Matematicas. RACSAM",
  ["rev r acad cienc exactas fis nat ser a math racsam"] = "Revista de la Real Academia de Ciencias Exactas, Fisicas y Naturales. Serie A. Matematicas. RACSAM",
  ["rev r acad cienc exactas fis quim nat zaragoza (2)"] = "Revista de la Real Academia de Ciencias Exactas, Fisicas, Quimicas y Naturales de Zaragoza. Serie 2",
  ["rev r acad cienc exactas fís nat (esp)"] = "Revista de la Real Academia de Ciencias Exactas, Físicas y Naturales (España)",
  ["rev r acad cienc exactas, fis nat ser a mat"] = "Revista de la Real Academia de Ciencias Exactas, Fisicas y Naturales. Serie A. Matematicas",
  ["rev radic polit econ"] = "The Review of radical political economics",
  ["rev radical polit econ"] = "Review of Radical Political Economics",
  ["rev radiol fisioter"] = "Revista de radiología y fisioterapía",
  ["rev real acad cienc exact fís natur madrid"] = "Real Academia de Ciencias Exactas, Físicas y Naturales de Madrid",
  ["rev recent clin trials"] = "Reviews on recent clinical trials",
  ["rev rech serv can for"] = "Revue de recherches du Service canadien des forêts",
  ["rev red cross"] = "Revista. Red Cross. Colombia. Cruz Roja Nacional",
  ["rev reg aracatuba assoc paul cir dent"] = "Revista Regional de Araçatuba, Associação Paulista de Cirurgiões Dentistas",
  ["rev reg res"] = "Review of Regional Research",
  ["rev reg stud"] = "The Review of regional studies",
  ["rev relig"] = "Review for religious",
  ["rev relig res"] = "Review of religious research",
  ["rev reprod"] = "Reviews of reproduction",
  ["rev res med microbiol"] = "Reviews and Research in Medical Microbiology",
  ["rev rev interam"] = "Revista/review interamericana",
  ["rev rhum"] = "Revue du rhumatisme",
  ["rev rhum ed fr"] = "Revue du rhumatisme (Ed. française : 1993)",
  ["rev rhum engl ed"] = "Revue du rhumatisme (English ed.)",
  ["rev rhum mal osteoartic"] = "Revue du rhumatisme et des maladies ostéo-articulaires",
  ["rev rom bioet"] = "Revista românǎ de bioeticǎ",
  ["rev rom mater"] = "Revista Romana de Materiale",
  ["rev rom med lab"] = "Revista română de medicină de laborator",
  ["rev rom stat"] = "Revista română de statistică : organ al Comisiei Naționale pentru Statistică",
  ["rev rouergue"] = "Revue du Rouergue",
  ["rev roum chim"] = "Revue Roumaine de Chimie",
  ["rev roum endocrinol"] = "Revue roumaine d'endocrinologie",
  ["rev roum geophys"] = "Revue Roumaine de Geophysique",
  ["rev roum hist"] = "Revue roumaine d'histoire",
  ["rev roum inframicrobiol"] = "Revue roumaine d'inframicrobiologie",
  ["rev roum med"] = "Revue roumaine de médecine (1974)",
  ["rev roum med intern"] = "Revue roumaine de médecine interne (1964)",
  ["rev roum morphol embryol"] = "Revue roumaine de morphologie et d'embryologie",
  ["rev roum morphol physiol"] = "Revue roumaine de morphologie et de physiologie",
  ["rev roum neurol"] = "Revue roumaine de neurologie",
  ["rev roum neurol psychiatr"] = "Revue roumaine de neurologie et de psychiatrie",
  ["rev roum physiol"] = "Revue roumaine de physiologie",
  ["rev roum physiol  (1990)"] = "Revue roumaine de physiologie (Bucharest, Romania : 1990)",
  ["rev roum virol"] = "Revue roumaine de virologie (Bucharest, Romania : 1990)",
  ["rev roumaine math pures appl"] = "Revue Roumaine de Mathematiques Pures et Appliquees. Romanian Journal of Pure and Applied Mathematics",
  ["rev roumaine sci tech sér méc appl"] = "Revue Roumaine des Sciences Techniques",
  ["rev roumaine sci tech sér électrotech énergét"] = "Revue Roumaine des Sciences Techniques",
  ["rev saintonge aunis"] = "Revue de la Saintonge et de l'Aunis",
  ["rev salud ambient"] = "Revista de salud ambiental",
  ["rev salud front"] = "Revista de salud fronteriza = Journal of border health",
  ["rev salud publica (bogota)"] = "Revista de salud pública (Bogotá, Colombia)",
  ["rev sanid"] = "Revista de sanidad",
  ["rev sanid aeronaut"] = "Revista de sanidad de aeronáutica",
  ["rev sanid asist soc"] = "Revista de sanidad y asistencia social",
  ["rev sanid hig publica (madr)"] = "Revista de sanidad e higiene pública",
  ["rev sanid mil"] = "Revista de la sanidad militar",
  ["rev sanid mil peru"] = "Revista de la sanidad militar del Perú",
  ["rev sanid milit"] = "Revista de sanidad militar",
  ["rev sanid milit argent"] = "Revista de la sanidad militar argentina",
  ["rev sanid polic"] = "Revista de la sanidad de policía",
  ["rev sanid vet"] = "Revista de sanidad veterinaria",
  ["rev saude publica"] = "Revista de saúde pública",
  ["rev saúde pública"] = "Revista de Saúde Pública",
  ["rev sci eau"] = "Revue des Sciences de l’Eau",
  ["rev sci edu"] = "Revue des sciences de l'éducation",
  ["rev sci fondam appl"] = "Revue des sciences fondamentales et appliquées",
  ["rev sci hum"] = "Revue des sciences humaines",
  ["rev sci instrum"] = "Review of Scientific Instruments",
  ["rev sci med"] = "Revue des sciences médicales",
  ["rev sci morales polit"] = "Revue des sciences morales & politiques : travaux de l'Académie des sciences morales et politiques",
  ["rev sci philos theol"] = "Revue des sciences philosophiques et théologiques",
  ["rev sci soc fr est"] = "Revue des sciences sociales de la France de l'Est",
  ["rev sci tech"] = "Revue scientifique et technique (International Office of Epizootics)",
  ["rev sci tech ser sci hum"] = "Revue science et technique. Série Sciences humaines",
  ["rev serv nac salud"] = "Revista. Chile. Servicio Nacional de Salud",
  ["rev serv nac tubers"] = "Revista do Serviço Nacional de Tuberculose",
  ["rev sesda"] = "Revue du SESDA",
  ["rev sifilogr leprolog dermatol"] = "Revista de sifilografía, leprología y dermatología",
  ["rev soc argent biol"] = "Revista de la Sociedad Argentina de Biologia",
  ["rev soc biom hum"] = "Revue de la Société de biométrie humaine",
  ["rev soc bolivar venez"] = "Revista de la Sociedad Bolivariana de Venezuela. Sociedad Bolivariana de Venezuela",
  ["rev soc bras hist cienc"] = "Revista da Sociedade Brasileira de Historia da Ciência",
  ["rev soc bras med trop"] = "Revista da Sociedade Brasileira de Medicina Tropical",
  ["rev soc colomb endocrinol"] = "Revista de la Sociedad Colombiana de Endocrinología",
  ["rev soc colomb ortod"] = "Revista. Sociedad Colombiana de Ortodoncia",
  ["rev soc colomb pediatr pueric"] = "Revista de la Sociedad Colombiana de Pediatría y Puericultura",
  ["rev soc econ"] = "Review of social economy",
  ["rev soc economy"] = "Review of Social Economy",
  ["rev soc entomol argent"] = "Revista de la Sociedad Entomológica Argentina",
  ["rev soc fr hist art dent"] = "Revue de la Société française d'histoire de l'art dentaire",
  ["rev soc fr hist hop"] = "Revue de la Société française d'histoire des hôpitaux",
  ["rev soc ginecol obstet"] = "Revista De La Sociedad De Ginecologia Y Obstetricia",
  ["rev soc med  quir zulia"] = "Revista de la Sociedad Médico-Quirúrgica del Zulia",
  ["rev soc med trujillo"] = "Revista de la Sociedad Médica de Trujillo",
  ["rev soc mex hist nat"] = "Revista de la Sociedad Mexicana de Historia Natural",
  ["rev soc odontol la plata"] = "Revista de la Sociedad Odontológica de La Plata",
  ["rev soc pediatr litoral"] = "Revista de la Sociedad de Pediatría del Litoral",
  ["rev soc peru endocrinol"] = "Revista de la Sociedad Peruana de Endocrinologia",
  ["rev soc peru protes dent maxilofac lima"] = "Revista. Sociedad Peruana de Prótesis Dental y Máxilo-Facial, Lima",
  ["rev soc policy"] = "Review of social policy",
  ["rev soc quim mex"] = "Revista de la Sociedad Quimica de Mexico",
  ["rev soc quim peru"] = "Revista de la Sociedad Quimica del Peru",
  ["rev soc savantes haute normandie"] = "Revue des sociétés savantes de Haute-Normandie",
  ["rev soc venez hist med"] = "Revista de la Sociedad Venezolana de Historia de la Medicina",
  ["rev social law"] = "Review of socialist law",
  ["rev sociol"] = "Revija za sociologiju",
  ["rev socionetwork strategies"] = "Review of Socionetwork Strategies",
  ["rev stat"] = "Revista De Statistica",
  ["rev stat appl"] = "Revue de statistique appliquée",
  ["rev stiint med"] = "Revista ştiinţelor medicale: medicina internă",
  ["rev stomatol chir maxillofac"] = "Revue de stomatologie et de chirurgie maxillo-faciale",
  ["rev stomatol chir maxillofac chir orale"] = "Revue de stomatologie, de chirurgie maxillo-faciale et de chirurgie orale",
  ["rev stomatoodontol nord fr"] = "La Revue stomato-odontologique du nord de la France",
  ["rev sudam morfol"] = "Revista sudamericana de morfología",
  ["rev suisse agric"] = "Revue Suisse d’Agriculture",
  ["rev suisse hydrol"] = "Revue suisse d’Hydrologie",
  ["rev suisse vitic aboric"] = "Revue Suisse de Viticulture et Aboriculture",
  ["rev suisse vitic arboric hortic"] = "Revue suisse de viticulture arboriculture horticulture",
  ["rev suisse zool"] = "Revue Suisse de Zoologie",
  ["rev surg"] = "Review of surgery",
  ["rev symb log"] = "Review of Symbolic Logic",
  ["rev synth"] = "Revue de synthèse",
  ["rev synth (6)"] = "Revue de Synthese. 6e Serie",
  ["rev synth (7)"] = "Revue de Synthese. 7e Serie",
  ["rev tarn"] = "Revue du Tarn",
  ["rev toxicol"] = "Reviews in Toxicology",
  ["rev trav acad sci morales polit c r seances"] = "Revue des travaux de l'Académie des sciences morales et politiques et comptes rendus de ses séances",
  ["rev tuberc"] = "Revue de la tuberculose",
  ["rev tuberc pneumol (paris)"] = "Revue de tuberculose et de pneumologie",
  ["rev tuberc urug"] = "Revista de tuberculosis del Uruguay",
  ["rev tunis etud popul"] = "Revue tunisienne des études de population",
  ["rev tunis geogr"] = "Revue tunisienne de géographie",
  ["rev tunis sci soc"] = "Revue tunisienne de sciences sociales",
  ["rev técn fac ingr univ zulia"] = "Universidad del Zulia",
  ["rev un mat argentina"] = "Revista de la Union Matematica Argentina",
  ["rev uniao odontol bras"] = "Revista. União Odontologica Brasileira",
  ["rev union mat argent"] = "Revista de la Union Matematica Argentina",
  ["rev univ"] = "Revista universitaria",
  ["rev univ brux"] = "Revue de l'Université de Bruxelles",
  ["rev univ nac cordoba"] = "Revista de la Universidad Nacional de Cordoba. Universidad Nacional de Córdoba",
  ["rev univ ottawa"] = "Revue de l'Université d'Ottawa. University of Ottawa",
  ["rev univ san marcos fac farm bioquim"] = "Revista. Universidad Nacional Mayor de San Marcos. Facultad de Farmacia y Bioquimica",
  ["rev urban reg devel stud"] = "Review of Urban and Regional Development Studies",
  ["rev urol"] = "Reviews in urology",
  ["rev urol (mex)"] = "Revista de urología",
  ["rev urol(mexico)"] = "Revista de urologia",
  ["rev urug psicoanal"] = "Revista uruguaya de psicoanálisis",
  ["rev venez sanid asist soc"] = "Revista venezolana de sanidad y asistencia social",
  ["rev venez urol"] = "Revista venezolana de urología",
  ["rev vet mil"] = "Revista de veterinaria militar",
  ["rev vet mil (paris)"] = "Revue vétérinaire militaire",
  ["rev viernes med"] = "Revista de Viernes Médico",
  ["rev virtual estud ling"] = "Revista virtual de estudos da linguagem",
  ["rev vivarais"] = "Revue du Vivarais",
  ["rev world econ"] = "Review of World Economics",
  ["rev zootec (b aires)"] = "Revista zootécnica",
  ["reveal antiq"] = "Revealing Antiquity",
  ["revenue-cycle strateg"] = "Revenue-cycle strategist",
  ["revis biol celular"] = "Revisiones sobre biología celular : RBC",
  ["revis statut la"] = "Revised statutes : Under arrangement of the official Louisiana revised statutes of 1950. Louisiana",
  ["revista admin emprêsas"] = "Revista de Administraçao d Emprêsas",
  ["revista anal econ"] = "Revista de Analisis Economico",
  ["revista brasileira econ"] = "Revista Brasileira de Economia",
  ["revista de investigacion agraria|span j agric res"] = "Spanish journal of agricultural research",
  ["revista econ"] = "Revista de Economía",
  ["revista econ aplicada"] = "Revista de Economia Aplicada",
  ["revista econ inst"] = "Revista de Economía Institucional",
  ["revista econ política"] = "Revista de Economia Política",
  ["revista econ y estadíst"] = "Revista de Economía y Estadística, N.S.",
  ["revista econometria"] = "Revista de Econometria",
  ["revista española econ"] = "Revista Española de Economía",
  ["revista estadíst"] = "Revista de Estadística",
  ["revista estud polít"] = "Revista de Estudios Políticos",
  ["revista fac cienc econ admin montevideo"] = "Revista de la Facultad de Ciencas Económicas y de Administración de  Montevideo",
  ["revista hist econ"] = "Revista de Historia Económica",
  ["revista hist ind"] = "Revista de Historia Industrial",
  ["revista trab soc (santiago)"] = "Revista de trabajo social",
  ["revmatologiia (mosk)"] = "Revmatologiia (Moscow, Russia)",
  ["revstat stat j"] = "Revstat statistical journal",
  ["revue africaine devel"] = "Revue Africaine de Developpement",
  ["revue can sci admin"] = "Revue Canadienne des Sciences de l’Administration",
  ["revue econ"] = "Revue Economique",
  ["revue econ dével"] = "Revue d’Economie du Développement",
  ["revue econ financ"] = "Revue d’Economie Financière",
  ["revue econ ind"] = "Revue d’Economie Industrielle",
  ["revue econ madagascar"] = "Revue Economique de Madagascar",
  ["revue econ politique"] = "Revue d’Economie Politique",
  ["revue econ reg urbaine"] = "Revue d’Economie Regionale et Urbaine",
  ["revue etud compar est-ouest"] = "Revue d’Etudes Comparatives Est-Ouest",
  ["revue finance contrôle stratégie"] = "Revue Finance Contrôle Stratégie",
  ["revue française econ"] = "Revue Française d’Economie",
  ["revue française gestion"] = "Revue Française de Gestion",
  ["revue geriatr"] = "La Revue de gériatrie",
  ["revue int geomatique"] = "Revue Internationale Geomatique",
  ["revue l’ofce"] = "Revue de L’OFCE",
  ["revue roumaine sci soc série sci econ"] = "Revue Roumaine des Sciences Sociales; Série Sciences Economiques",
  ["revue sci financ"] = "Revue de Science Financière",
  ["revue stomatol"] = "Revue de stomatologie",
  ["revue tiers monde"] = "Revue Tiers Monde",
  ["rezultaty issled mezhdunar geofiz proekt"] = "Rezultaty Issledovanii po Mezhdunarodnym Proektam",
  ["rf illus"] = "RF illustrated",
  ["rguhs j pharm sci"] = "RGUHS journal of pharmaceutical sciences",
  ["rhein mus philol"] = "Rheinisches Museum für Philologie",
  ["rhein vierteljahrsbl"] = "Rheinische Vierteljahrsblätter",
  ["rheinische lebensbilder"] = "Rheinische Lebensbilder",
  ["rheinisches arztebl"] = "Rheinisches Ärzteblatt",
  ["rheol acta"] = "Rheologica Acta",
  ["rhetor prof commun glob"] = "Rhetoric, professional communication, and globalization",
  ["rhetor soc q"] = "Rhetoric Society quarterly",
  ["rheum dis"] = "The Rheumatic diseases",
  ["rheum dis clin north am"] = "Rheumatic diseases clinics of North America",
  ["rheuma plus"] = "Rheuma plus",
  ["rheumatol adv pract"] = "Rheumatology advances in practice",
  ["rheumatol balneol allergol"] = "Rheumatologia, balneologia, allergologia",
  ["rheumatol int"] = "Rheumatology international",
  ["rheumatol phys med"] = "Rheumatology and physical medicine",
  ["rheumatol rehabil"] = "Rheumatology and rehabilitation",
  ["rheumatol ther"] = "Rheumatology and therapy",
  ["rheumatology (oxford)"] = "Rheumatology (Oxford, England)",
  ["rheumatology (sunnyvale)"] = "Rheumatology (Sunnyvale, Calif.)",
  ["rhinol online"] = "Rhinology online",
  ["rhinol suppl"] = "Rhinology. Supplement",
  ["rhod nurse"] = "The Rhodesia nurse",
  ["riazi j karachi math assoc"] = "Riazi",
  ["ric clin lab"] = "La Ricerca in clinica e in laboratorio",
  ["ric mat"] = "Ricerche di Matematica",
  ["ric rev"] = "RIC Reviews",
  ["ric sci"] = "Ricerca Scientifica",
  ["ric sci 2 ser pt 1 riv"] = "La Ricerca scientifica. 2. ser., pt.1: Rivista",
  ["ric sci 2 ser pt 2 rend [b]"] = "Ricerca Scientifica. 2 Ser. Parte 2, Rendiconti. Sezione B, Biologica",
  ["ric sci 2 ser pt 2 rend b"] = "La Ricerca scientifica. 2. ser., pt. 2: Rendiconti. Sezione B: Biologica",
  ["rice (n y)"] = "Rice (New York, N.Y.)",
  ["rice biotechnol q"] = "Rice biotechnology quarterly",
  ["rice genet newsl"] = "Rice genetics newsletter",
  ["rice genom genet"] = "Rice genomics and genetics",
  ["rice sci"] = "Rice Science",
  ["ricerche econ"] = "Ricerche Economiche",
  ["ricerche mat"] = "Ricerche di Matematica",
  ["richmond cty hist"] = "Richmond County history",
  ["riforma med"] = "La Riforma medica",
  ["rim monogr math"] = "RIM Monographs in Mathematics",
  ["ringing migr"] = "Ringing & Migration",
  ["rinsho biseibutshu jinsoku shindan kenkyukai shi"] = "Rinshō Biseibutsu Jinsoku Shindan Kenkyūkai shi = JARMAM : Journal of the Association for Rapid Method and Automation in Microbiology",
  ["rinsho byori"] = "Rinsho byori. The Japanese journal of clinical pathology",
  ["rinsho eiyo"] = "[Rinshō eiyō] [Clinical nutrition]",
  ["rinsho fujinka sanka"] = "[Rinshō fujinka sanka] Clinical gynecology and obstetrics",
  ["rinsho ganka"] = "Rinsho ganka. Japanese journal of clinical ophthalmology",
  ["rinsho geka"] = "Rinsho geka. Journal of clinical surgery",
  ["rinsho hoshasen"] = "Rinsho hoshasen. Clinical radiography",
  ["rinsho ketsueki"] = "[Rinshō ketsueki] The Japanese journal of clinical hematology",
  ["rinsho kyobu geka"] = "Rinshō kyōbu geka = Japanese annals of thoracic surgery",
  ["rinsho naika shonika intern med"] = "Rinsho naika shonika. Internal medicine and pediatrics",
  ["rinsho shika"] = "[Rinshō shika] [Clinical dentistry]",
  ["rinsho shinkeigaku"] = "Rinshō shinkeigaku = Clinical neurology",
  ["rinsho shishubyo danwakai kaishi"] = "Rinshō Shishūbyō Danwakai kaishi",
  ["rinsho shokakibyogaku"] = "[Rinshō shōkakibyōgaku] Clinical gastro-enterology",
  ["rinsho to uirusu"] = "Rinsho to uirusu. Clinical virology",
  ["rinsho yakuri"] = "Rinsho Yakuri",
  ["riogrande odontol"] = "Riogrande odontológico",
  ["ripon forum"] = "Ripon forum",
  ["risec: int rev econ bus"] = "RISEC: International Review of Economics and Business",
  ["risk anal"] = "Risk analysis : an official publication of the Society for Risk Analysis",
  ["risk dec pol"] = "Risk Decision and Policy",
  ["risk eng"] = "Risk Engineering",
  ["risk hazards crisis public policy"] = "Risk, hazards & crisis in public policy",
  ["risk issues health saf"] = "Risk, issues in health & safety",
  ["risk manag can health care"] = "Risk management in Canadian health care",
  ["risk manag healthc policy"] = "Risk management and healthcare policy",
  ["risk manage"] = "Risk management (New York, N.Y.)",
  ["riso rep"] = "Risø report",
  ["riv anat patol oncol"] = "Rivista di anatomia patologica e di oncologia",
  ["riv biol"] = "Rivista di biologia",
  ["riv biol coloniale"] = "Rivista di biologia coloniale",
  ["riv chir med"] = "Rivista di chirurgia e medicina",
  ["riv chir pediatr"] = "Rivista di chirurgia pediatrica",
  ["riv clin pediatr"] = "Rivista di clinica pediatrica",
  ["riv combust ind chim"] = "Rivista dei Combustibili e dell’Industia Chimica",
  ["riv crit clin med"] = "Rivista critica di clinica medica",
  ["riv crit stor filos"] = "Rivista critica di storia della filosofia",
  ["riv dif soc"] = "Rivista di difesa sociale",
  ["riv emoter immunoematol"] = "Rivista di emoterapia ed immunoematologia",
  ["riv estet"] = "Rivista di estetica",
  ["riv eur sci med farmacol"] = "Rivista europea per le scienze mediche e farmacologiche = European review for medical and pharmacological sciences = Revue européenne pour les sciences médicales et pharmacologiques",
  ["riv filol istruz classica"] = "Rivista di filologia e d'istruzione classica",
  ["riv filos"] = "Rivista di filosofia",
  ["riv filos neo scolast"] = "Rivista di filosofia neo-scolastica",
  ["riv gastroenterol"] = "Rivista di gastro-enterologia",
  ["riv geogr ital"] = "Rivista geografica italiana",
  ["riv inferm"] = "Rivista dell'infermiere",
  ["riv infort mal prof"] = "Rivista degli infortuni e delle malattie professionali",
  ["riv int psicol ipn"] = "Rivista internazionale di psicologia e ipnosi",
  ["riv int sci sociali"] = "Rivista internazionale di scienze sociali",
  ["riv internazionale filos psicol"] = "Rivista internazionale di filosofia e psicologia",
  ["riv ist sieroter ital"] = "Rivista dell'Istituto sieroterapico italiano",
  ["riv ist vaccinogeno consorzi prov antituberc"] = "Rivista dell'Istituto vaccinogeno e dei consorzi provinciali antitubercolari",
  ["riv istochim norm patol"] = "Rivista di istochimica, normale e patologica",
  ["riv ital econ demogr stat"] = "Rivista italiana di economia, demografia e statistica",
  ["riv ital ginecol"] = "Rivista italiana di ginecologia",
  ["riv ital ig"] = "Rivista italiana d'igiene",
  ["riv ital med lab"] = "Rivista Italiana della Medicina di Laboratorio",
  ["riv ital med leg"] = "Rivista italiana di medicina legale : dottrina, casistica, ricerca sperimentale, giurisprudenza e legislazione",
  ["riv ital odontoiatr infant"] = "Rivista italiana di odontoiatria infantile : organo ufficiale della Società italiana di odontoiatria infantile",
  ["riv ital odontotec"] = "Rivista italiana degli odontotecnici",
  ["riv ital paleontol stratigr"] = "Rivista Italiana Di Paleontologia E Stratigrafia",
  ["riv ital pediatr"] = "Rivista italiana di pediatria = The Italian journal of pediatrics",
  ["riv ital stomatol"] = "Rivista italiana di stomatologia",
  ["riv ital trac patol ocul esotica"] = "Rivista italiana del tracoma e di patologia oculare esotica",
  ["riv linguist"] = "Rivista di linguistica",
  ["riv malariol"] = "Rivista di malariologia",
  ["riv maritt"] = "Rivista marittima",
  ["riv mat pura appl"] = "Rivista di Matematica Pura ed Applicata",
  ["riv mat sci econom social"] = "Rivista di Matematica per le Scienze Economiche e Sociali",
  ["riv mat univ parma (5)"] = "Rivista di Matematica della Università di Parma",
  ["riv math univ parma (ns)"] = "Rivista di Matematica della Universita di Parma. New Series. A Journal of Pure and Applied Mathematics",
  ["riv med aeronaut"] = "Rivista di medicina aeronautica",
  ["riv med aeronaut spaz"] = "Rivista di medicina aeronautica e spaziale",
  ["riv med leg legis sanit"] = "Rivista di medicina legale e legislazione sanitaria",
  ["riv meteorol aeronaut"] = "Rivista di meteorologia aeronautica",
  ["riv micol"] = "Rivista di micologia : bollettino dell'Associazione micologica Bresadola",
  ["riv micol romana"] = "Rivista micologica romana",
  ["riv mil"] = "Rivista militare",
  ["riv neurobiol"] = "Rivista di neurobiologia : organo ufficiale della Società dei neurologi, neuroradiologi e neurochirurghi ospedalieri",
  ["riv neurol"] = "Rivista di neurologia",
  ["riv nuovo cimento"] = "Rivista del Nuovo Cimento",
  ["riv nuovo cimento (3)"] = "La Rivista del Nuovo Cimento",
  ["riv nuovo cimento soc ital fis (4)"] = "La Rivista del Nuovo Cimento della Società Italiana di Fisica",
  ["riv odontoiatr amici brugg"] = "Rivista di odontoiatria degli Amici di Brugg",
  ["riv odontojatria ortognatod"] = "Rivista di odontojatria e ortognatodonzia",
  ["riv odontostomatol implantoprotesi"] = "Rivista di odontostomatologia e implantoprotesi",
  ["riv ostet ginecol"] = "Rivista di ostetricia e ginecologia",
  ["riv ostet ginecol prat"] = "Rivista d'ostetricia e ginecologia pratica",
  ["riv ostet ginecol prat med perinat"] = "Rivista di ostetricia ginecologia pratica e medicina perinatale",
  ["riv otoneurooftalmol"] = "Rivista oto-neuro-oftalmologica",
  ["riv parassitol"] = "Rivista di parassitologia",
  ["riv patol clin"] = "Rivista di patologia e clinica",
  ["riv patol clin sper"] = "Rivista di patologia clinica e sperimentale",
  ["riv patol clin tuberc"] = "Rivista di patologia e clinica della tubercolosi",
  ["riv patol nerv ment"] = "Rivista di patologia nervosa e mentale",
  ["riv pediatr sicil"] = "Rivista pediatrica siciliana",
  ["riv psichiatr"] = "Rivista di psichiatria",
  ["riv psicoanal"] = "Rivista di psicoanalisi",
  ["riv psicol"] = "Rivista di psicologia",
  ["riv psicol (1912)"] = "Rivista di psicologia",
  ["riv psicol (1945)"] = "Rivista di psicologia (1945)",
  ["riv psicol (1955)"] = "Rivista di psicologia (1955)",
  ["riv psicol scr"] = "Rivista di psicologia della scrittura",
  ["riv psicol soc arch ital psicol gen"] = "Rivista di psicologia sociale e Archivio italiano di psicologia generale e del lavoro",
  ["riv psicolinguist appl"] = "Rivista di psicolinguistica applicata",
  ["riv radiol"] = "Rivista di radiologia",
  ["riv rosmin filos cult"] = "Rivista rosminiana di filosofia e di cultura",
  ["riv sicil tuberc"] = "Rivista siciliana della tubercolosi",
  ["riv sicil tuberc mal appar respir"] = "Rivista siciliana della tubercolosi e delle malattie dell'apparato respiratorio",
  ["riv sper freniatr"] = "Rivista Sperimentale di Freniatria e Medicina Legale delle Alienazioni Mentali",
  ["riv sper freniatr med leg alien ment"] = "Rivista sperimentale di freniatria e medicina legale delle alienazioni mentali",
  ["riv stor calabr"] = "Rivista storica calabrese",
  ["riv stor chiesa ital"] = "Rivista di storia della chiesa in Italia",
  ["riv stor filos"] = "Rivista di storia della filosofia (Milan, Italy : 1984)",
  ["riv stor ital"] = "Rivista storica italiana",
  ["riv stor med"] = "Rivista di storia della medicina",
  ["riv stor sci"] = "Rivista di storia della scienza",
  ["riv stor sci (2)"] = "Rivista di Storia della Scienza",
  ["riv stor sci mediche nat"] = "Rivista di storia delle scienze mediche e naturali",
  ["riv studi class"] = "Rivista di studi classici",
  ["riv studi orient"] = "Rivista degli studi orientali",
  ["riv tuberc mal appar respir"] = "Rivista della tubercolosi e delle malattie dell'apparato respiratorio",
  ["river publ ser math eng sci"] = "River Publishers Series in Mathematical and Engineering Sciences",
  ["river publ ser math stat comput model eng"] = "River Publishers Series in Mathematical, Statistical and Computational Modelling for Engineering",
  ["river publ ser softw eng"] = "River Publishers Series in Software Engineering",
  ["river res appl"] = "River Research and Applications",
  ["river syst"] = "River Systems",
  ["rivista int sci econ com"] = "Rivista Internazionale di Scienze Economiche e Commerciali",
  ["rivista int sci soc"] = "Rivista Internazionale di Scienze Sociali",
  ["rivista italiana econ"] = "Rivista Italiana degli Economisti",
  ["rivista politica econ"] = "Rivista di Politica Economica",
  ["rivista statist appl"] = "Rivista di Statistica Applicata",
  ["rivista storia econ"] = "Rivista di Storia Economica, N.S.",
  ["rmd open"] = "RMD open",
  ["rmle online"] = "RMLE online : research in middle level education",
  ["rms res math stat"] = "RMS. Research in Mathematics & Statistics",
  ["rn (for managers)"] = "RN (For managers)",
  ["rn ida"] = "RN Idaho",
  ["rna biol"] = "RNA biology",
  ["rna dis"] = "RNA & disease (Houston, Tex.)",
  ["rna technol"] = "RNA Technologies",
  ["rnabc news"] = "RNABC news",
  ["rnao news"] = "RNAO news",
  ["road mater pavement des"] = "Road Materials and Pavement Design",
  ["rob autom syst"] = "Robotics and Autonomous Systems",
  ["rob auton syst"] = "Robotics and autonomous systems",
  ["rob biomimetics"] = "Robotics and Biomimetics",
  ["rob comput integr manuf"] = "Robotics and Computer-Integrated Manufacturing",
  ["rob intell autom"] = "Robotic Intelligence and Automation",
  ["rob rep"] = "Robotics Reports",
  ["rob surg: res rev"] = "Robotic Surgery: Research and Reviews",
  ["robomech j"] = "ROBOMECH Journal",
  ["robot comput integr manuf"] = "Robotics and computer-integrated manufacturing",
  ["robot mechatron (2015)"] = "Robotics and mechatronics : proceedings of the 4th IFToMM International Symposium on Robotics and Mechatronics. IFToMM International Symposium on Robotics and Mechatronics (4th : 2015 : Poitiers, France)",
  ["robot sci syst"] = "Robotics science and systems : online proceedings",
  ["robot surg"] = "Robotic surgery (Auckland)",
  ["robotics biomim"] = "Robotics and biomimetics",
  ["robototekhn gibk proizvod sist"] = "Robototekhnika i Gibkie Proizvodstvennye Sistemy",
  ["roche rev"] = "Roche review",
  ["rochester hist"] = "Rochester history",
  ["rock mech bull"] = "Rock Mechanics Bulletin",
  ["rock mech rock eng"] = "Rock Mechanics and Rock Engineering",
  ["rocks miner"] = "Rocks and Minerals",
  ["rocky mountain j math"] = "The Rocky Mountain Journal of Mathematics",
  ["rocky mt for range  exp stn, for ser"] = "Rocky Mountain Forest and Range Experiment Station, Forest Series",
  ["rocky mt geol"] = "Rocky Mountain Geology",
  ["rocky mt j math"] = "Rocky Mountain journal of mathematics",
  ["rocky mt med j"] = "Rocky Mountain medical journal",
  ["rocky mt soc sci j"] = "The Rocky Mountain social science journal",
  ["rocz akad med bialymst"] = "Roczniki Akademii Medycznej w Białymstoku (1995)",
  ["rocz akad med bialymst supl"] = "Roczniki Akademii Medycznej w Białymstoku. Suplement = Annales Academiae Medicae Bialostocensis. Supplementum",
  ["rocz akad med im juliana marchlewskiego bialymst"] = "Roczniki Akademii Medycznej im. Juliana Marchlewskiego w Bialymstoku",
  ["rocz akad med im juliana marchlewskiego bialymst suppl"] = "Roczniki Akademii Medycznej im. Juliana Marchlewskiego w Białymstoku. Suplement",
  ["rocz filoz"] = "Roczniki filozoficzne. Annales de philosophie",
  ["rocz humanist"] = "Roczniki humanistyczne",
  ["rocz panstw zakl hig"] = "Roczniki Państwowego Zakładu Higieny",
  ["rocz pomor akad med"] = "Rocznik Pomorskiej Akademii Medycznej im. Gen. Karola Swierczewskiego w Szczecinie",
  ["rocznik nauk-dydakt prace mat"] = "Wyższa Szkoła Pedagogiczna im. Komisji Edukacji Narodowej w Krakowie. Rocznik Naukowo-Dydaktyczny. Prace Matematyczne",
  ["rodo kagaku"] = "Rodo kagaku. The Journal of science of labour",
  ["rodo kagaku kenkyu"] = "[Rōdō kagaku kenkyū] The Journal of science of labour",
  ["roeper rev"] = "Roeper review",
  ["rofo fortschr geb rontgenstr neuen bildgeb verfahr"] = "Rofo; Fortschritte auf dem Gebiete der Rontgenstrahlen und der Neuen Bildgebenden Verfahren",
  ["rofo fortschr geb rontgenstr nuklearmed"] = "RoFo Fortschritte auf dem Gebiete der Rontgenstrahlen und der Nuklearmedizin",
  ["rogerian nurs sci news"] = "Rogerian nursing science news : newsletter of the Society of Rogerian Scholars",
  ["rolf nevanlinna inst res rep a"] = "Rolf Nevanlinna Institute Research Reports A",
  ["roll stone"] = "Rolling stone (San Francisco, Calif.)",
  ["rom agric res"] = "Romanian agricultural research",
  ["rom biotechnol lett"] = "Romanian Biotechnology Letters",
  ["rom hist mitt"] = "Römische historische Mitteilungen",
  ["rom j anaesth intensive care"] = "Romanian journal of anaesthesia and intensive care",
  ["rom j biochem"] = "Romanian Journal of Biochemistry",
  ["rom j biophys"] = "Romanian journal of biophysics",
  ["rom j chem"] = "Romanian Journal of Chemistry",
  ["rom j diabetes nutr metab dis"] = "Romanian journal of diabetes, nutrition, & metabolic diseases",
  ["rom j endocrinol"] = "Romanian journal of endocrinology",
  ["rom j gastroenterol"] = "Romanian journal of gastroenterology",
  ["rom j geogr"] = "Romanian Journal Of Geography",
  ["rom j gerontol geriatr"] = "Romanian journal of gerontology and geriatrics",
  ["rom j inf sci technol"] = "Romanian Journal Of Information Science And Technology",
  ["rom j intern med"] = "Romanian journal of internal medicine = Revue roumaine de médecine interne",
  ["rom j legal med"] = "Romanian Journal of Legal Medicine",
  ["rom j mater"] = "Romanian Journal of Materials",
  ["rom j math comput sci"] = "Romanian Journal of Mathematics and Computer Science",
  ["rom j morphol embryol"] = "Romanian journal of morphology and embryology = Revue roumaine de morphologie et embryologie",
  ["rom j neurol psychiatry"] = "Romanian journal of neurology and psychiatry = Revue roumaine de neurologie et psychiatrie",
  ["rom j ophthalmol"] = "Romanian journal of ophthalmology",
  ["rom j phys"] = "Romanian Journal of Physics",
  ["rom j physiol"] = "Romanian journal of physiology : physiological sciences",
  ["rom j tech sci appl mech"] = "Romanian Journal of Technical Sciences. Applied Mechanics",
  ["rom j virol"] = "Romanian journal of virology",
  ["rom med rev"] = "Romanian medical review",
  ["rom rep phys"] = "Romanian reports in physics",
  ["romai j"] = "ROMAI Journal",
  ["roman linguist 2008 (2008)"] = "Romance linguistics 2008 : interactions in romance : selected papers from the 38th linguistic symposium on romance languages (LSRL), Urbana-Champaign, April 2008. Linguistic Symposium on Romance Languages (38th : 2008 : Urbana-Champaign, Ill.)",
  ["roman notes"] = "Romance notes",
  ["romanian j parasit dis"] = "Romanian journal of parasitic diseases",
  ["romanian j phys"] = "Romanian Journal of Physics",
  ["romanian rep phys"] = "Romanian Reports in Physics",
  ["ronald e mcnair postbac achiev program"] = "Ronald E. McNair postbaccalaureate achievement program : [journal]",
  ["rontgen laborator"] = "Röntgen- und Laboratoriumspraxis",
  ["rontgendiagnostik ergeb"] = "Röntgendiagnostik; Ergebnisse",
  ["rontgeneur radiodiagn clin eur"] = "Roentgen-Europ; radiodiagnostic clinique européen",
  ["rontgenphotogr med photogr med lab prax"] = "Röntgenphotographie; medizinsche Photographie und medizinische Laboratoriumspraxis",
  ["rontgenprax diagn rontgen radium lichtther"] = "Röntgenpraxis; Diagnostik, Röntgen-, Radium-, Lichttherapie",
  ["rorschach res exch j proj tech"] = "Rorschach research exchange and journal of projective techniques",
  ["rose-hulman undergrad math j"] = "Rose-Hulman Undergraduate Math Journal",
  ["ross fiziol zh im i m sechenova"] = "Rossiĭskii fiziologicheskiĭ zhurnal imeni I.M. Sechenova",
  ["ross gastroenterol zh"] = "Rossiĭskiĭ gastroėnterologicheskiĭ zhurnal : ezhekvartalńyĭ nauchno-prakticheskiĭ zhurnal",
  ["ross immunol zhurnal"] = "Rossiĭskiĭ immunologicheskiĭ zhurnal",
  ["ross khim zh"] = "Rossiiskii Khimicheskii Zhurnal",
  ["ross med zh"] = "Rossiĭskiĭ medit︠s︡inskiĭ zhurnal : organ Ministerstva zdravookhranenii︠a︡ RSFSR",
  ["rostock math kolloq"] = "Rostocker Mathematisches Kolloquium",
  ["roum arch microbiol immunol"] = "Roumanian archives of microbiology and immunology",
  ["round table"] = "The Round table",
  ["rout stud philos rel"] = "Routledge Studies in The Philosophy of Religion",
  ["routl eur sociol ass"] = "Routledge European Sociological Association Studies in European Societies",
  ["routl hindu stud ser"] = "Self-surrender (prapatti) to God in Srivaisnavism: Tamil Cats and Sanskrit Monkeys",
  ["routl philos"] = "Routledge Philosophers",
  ["routl stud ecol econ"] = "Routledge Studies in Ecological Economics",
  ["routledge contemp read philos"] = "Routledge Contemporary Readings in Philosophy",
  ["routledge inem adv econ methodol"] = "Routledge INEM Advances in Economic Methodology",
  ["routledge philos"] = "Routledge Philosophers",
  ["routledge res early mod hist"] = "Routledge Research in Early Modern History",
  ["routledge stud philos math phys"] = "Routledge Studies in the Philosophy of Mathematics and Physics",
  ["rouxs arch dev biol"] = "Roux's archives of developmental biology : the official organ of the EDBO",
  ["roy inst philos suppl"] = "Royal Institute of Philosophy Supplement",
  ["roy statist soc lecture note ser"] = "Royal Statistical Society Lecture Note Series",
  ["rozhl chir"] = "Rozhledy v chirurgii : měsíčník Československé chirurgické společnosti",
  ["rozhl tuberk"] = "Rozhledy v tuberkulose",
  ["rozpr ces akad ved umeni trida ii mat prirodoved"] = "Rozpravy České akademie věd a umění. Třída II, Mathematicko-přírodovědecká",
  ["rozpr monogr"] = "Rozprawy Monografie",
  ["rozpr wydz nauk med"] = "Rozprawy Wydzialu Nauk Medycznych",
  ["rozprawy politech poznań"] = "Politechnika Poznańska",
  ["rps pharm pharmacol rep"] = "RPS Pharmacy and Pharmacology Reports",
  ["rsa ser transdiscipl rhetor"] = "The RSA Series in Transdisciplinary Rhetoric",
  ["rsc adv"] = "RSC Advances",
  ["rsc appl interfaces"] = "RSC Applied Interfaces",
  ["rsc appl polym"] = "RSC Applied Polymers",
  ["rsc biomol sci"] = "RSC biomolecular sciences",
  ["rsc chem biol"] = "RSC chemical biology",
  ["rsc med chem"] = "RSC medicinal chemistry",
  ["rsc sustainability"] = "RSC Sustainability",
  ["rsme springer ser"] = "RSME Springer Series",
  ["rsyn chem sci"] = "RSYN Chemical Sciences",
  ["rsyn proc"] = "RSYN Proceedings",
  ["rsyn res j"] = "RSYN Research Journal",
  ["rt-11 tech user’s ser"] = "The RT-11 Technical User’s Series",
  ["rubber chem technol"] = "Rubber Chemistry and Technology",
  ["ruch praw ekon socjol"] = "Ruch prawniczy, ekonomiczny i socjologiczny (1962)",
  ["rudolstaedter heimath"] = "Rudolstädter Heimathefte",
  ["rum med rev"] = "Rumanian medical review",
  ["rumah tangga kesehatan"] = "Rumah Tangga Dan Kesehatan",
  ["ruperto carola"] = "Ruperto-Carola",
  ["ruptures (univ montr groupe rech interdiscip sante)"] = "Ruptures (Université de Montréal. Groupe de recherche interdisciplinaire en santé)",
  ["rural afr"] = "Rural Africana",
  ["rural cond trends"] = "Rural conditions and trends",
  ["rural demogr"] = "Rural demography",
  ["rural hist"] = "Rural history : economy, society, culture",
  ["rural landscapes: soc environ hist"] = "Rural Landscapes: Society, Environment, History",
  ["rural ment health"] = "Rural mental health",
  ["rural policy brief"] = "Rural policy brief",
  ["rural reconstr forum"] = "Rural reconstruction forum",
  ["rural remote health"] = "Rural and remote health",
  ["rural sociol"] = "Rural sociology",
  ["rural sustainability res"] = "Rural Sustainability Research",
  ["rurds rev urb reg dev stud"] = "Review of urban and regional development studies : RURDS : journal of the Applied Regional Conference",
  ["rus ornithol z"] = "Russkij ornithologičeskij žurnal",
  ["rus zhurnal spid rak obshchestvennoe zdor"] = "Russkiĭ zhurnal SPID, rak i obshchestvennoe zdorovʹe = Russian journal AIDS, cancer, and public health",
  ["russ aeronaut"] = "Russian Aeronautics",
  ["russ agric sci"] = "Russian agricultural sciences",
  ["russ chem bull"] = "Russian Chemical Bulletin",
  ["russ chem rev"] = "Russian Chemical Reviews",
  ["russ coiles health trends"] = "Russ Coile's health trends",
  ["russ electr eng"] = "Russian Electrical Engineering",
  ["russ eng res"] = "Russian Engineering Research",
  ["russ entomol j"] = "Russian entomological journal = Russkiĭ ėntomologicheskiĭ zhurnal",
  ["russ geol geophys"] = "Russian Geology and Geophysics",
  ["russ hist (pittsburgh)"] = "Russian history. Histoire russe",
  ["russ j agric socioecon sci"] = "Russian journal of agricultural and socio-economic sciences",
  ["russ j appl chem"] = "Russian Journal of Applied Chemistry",
  ["russ j biol invasions"] = "Russian journal of biological invasions",
  ["russ j bioorg chem"] = "Russian Journal of Bioorganic Chemistry",
  ["russ j coord chem"] = "Russian Journal of Coordination Chemistry",
  ["russ j dev biol"] = "Russian journal of developmental biology",
  ["russ j earth sci"] = "Russian Journal of Earth Sciences",
  ["russ j ecol"] = "Russian journal of ecology",
  ["russ j ecosyst ecol"] = "Russian journal of ecosystem ecology",
  ["russ j electrochem"] = "Russian Journal of Electrochemistry",
  ["russ j gen chem"] = "Russian Journal of General Chemistry",
  ["russ j genet"] = "Russian Journal of Genetics",
  ["russ j genet appl res"] = "Russian journal of genetics. Applied research",
  ["russ j genet: appl res"] = "Russian Journal of Genetics: Applied Research",
  ["russ j herpetol"] = "Russian journal of herpetology",
  ["russ j immunol"] = "Russian journal of immunology : RJI : official journal of Russian Society of Immunology",
  ["russ j inorg chem"] = "Russian Journal of Inorganic Chemistry",
  ["russ j mar biol"] = "Russian Journal of Marine Biology",
  ["russ j math phys"] = "Russian Journal of Mathematical Physics",
  ["russ j nematol"] = "Russian journal of nematology",
  ["russ j non-ferrous met"] = "Russian Journal of Non-Ferrous Metals",
  ["russ j nondestr test"] = "Russian Journal of Nondestructive Testing",
  ["russ j nonlinear dyn"] = "Russian Journal of Nonlinear Dynamics",
  ["russ j numer anal math model"] = "Russian Journal of Numerical Analysis and Mathematical Modelling",
  ["russ j org chem"] = "Russian Journal of Organic Chemistry",
  ["russ j pac geol"] = "Russian Journal of Pacific Geology",
  ["russ j phys chem"] = "Russian Journal of Physical Chemistry",
  ["russ j phys chem a"] = "Russian Journal of Physical Chemistry A",
  ["russ j phys chem b"] = "Russian Journal of Physical Chemistry B, Focus on Physics",
  ["russ j plant physiol"] = "Russian journal of plant physiology: a comprehensive Russian journal on modern phytophysiology",
  ["russ j theriol"] = "Russian journal of theriology",
  ["russ math"] = "Russian Mathematics",
  ["russ math surv"] = "Russian Mathematical Surveys",
  ["russ metall"] = "Russian Metallurgy",
  ["russ meteorol hydrol"] = "Russian Meteorology and Hydrology",
  ["russ microlectron"] = "Russian Microelectronics",
  ["russ phys j"] = "Russian Physics Journal",
  ["russ rev"] = "The Russian review",
  ["russ rev biol"] = "Russian review of biology",
  ["russ soc sci rev"] = "Russian social science review : a journal of translations",
  ["russ stud hist"] = "Russian studies in history",
  ["russian acad sci izv math"] = "Russian Academy of Sciences",
  ["russian e europ finance trade"] = "Russian and East European Finance and Trade",
  ["russian econ trends"] = "Russian Economic Trends",
  ["russian j math phys"] = "Russian Journal of Mathematical Physics",
  ["russian j numer anal math modelling"] = "Russian Journal of Numerical Analysis and Mathematical Modelling",
  ["russian math (iz vuz)"] = "Russian Mathematics (Izvestiya VUZ. Matematika)",
  ["russian math surveys"] = "Russian Mathematical Surveys",
  ["russian phys j"] = "Russian Physics Journal",
  ["russkii entomologicheskii zhurnal|russ entomol j"] = "Russian entomological journal",
  ["rutg self soc id"] = "Rutgers Series On Self and Social Identity",
  ["rutgers camden law j"] = "Rutgers Camden law journal",
  ["rutgers comput technol law j"] = "Rutgers computer & technology law journal",
  ["rutgers j bioeth"] = "The Rutgers journal of bioethics",
  ["rutgers j comput technol law"] = "Rutgers journal--computers, technology, and the law",
  ["rutgers law j"] = "Rutgers law journal",
  ["rutgers law rev"] = "Rutgers law review",
  ["rutgers lect philos"] = "The Rutgers Lectures in Philosophy",
  ["rutland rec"] = "Rutland record : journal of the Rutland Record Society",
  ["ryan advis health serv gov boards"] = "The Ryan advisory for health services governing boards",
  ["ryoikibetsu shokogun shirizu"] = "Ryōikibetsu shōkōgun shirīzu",
  ["ryukyu math j"] = "Ryukyu Mathematical Journal",
  ["région dével"] = "Région et Développement",
  ["s a nurs j"] = "S.A. nursing journal = S.A. verplegingstydskrif",
  ["s afr archaeol bull"] = "South African Archaeological Bulletin",
  ["s afr cancer bull"] = "South African cancer bulletin = Suid-Afrikaanse kankerbulletin",
  ["s afr comput j"] = "South African Computer Journal",
  ["s afr dent j"] = "South African dental journal. Suid Afrikaanse tandarts tydskrif",
  ["s afr fam pract (2004)"] = "South African family practice : official journal of the South African Academy of Family Practice/Primary Care",
  ["s afr geogr j"] = "The South African geographical journal, being a record of the proceedings of the South African Geographical Society",
  ["s afr health rev"] = "South African health review",
  ["s afr hist j"] = "Suid-Afrikaanse historiese joernaal",
  ["s afr humanit"] = "South African Humanities",
  ["s afr j afr aff"] = "The South African journal of African affairs",
  ["s afr j agric ext"] = "South African Journal of Agricultural Extension",
  ["s afr j anim sci"] = "South African Journal Of Animal Science",
  ["s afr j bioeth law"] = "South African journal of bioethics and law",
  ["s afr j bot"] = "South African Journal of Botany",
  ["s afr j chem"] = "South African Journal of Chemistry",
  ["s afr j chem eng"] = "South African Journal of Chemical Engineering",
  ["s afr j clin nutr"] = "South African Journal of Clinical Nutrition",
  ["s afr j clin sci"] = "South African journal of clinical science. Suid-Afrikaanse tydskrif vir kliniese wetenskap",
  ["s afr j commun disord"] = "The South African journal of communication disorders = Die Suid-Afrikaanse tydskrif vir Kommunikasieafwykings",
  ["s afr j econ"] = "The South African journal of economics. Suid-afrikaanse tydskrif vir ekonomie",
  ["s afr j econ hist"] = "The South African journal of economic history : [journal of the Economic History Society of Southern Africa]",
  ["s afr j educ"] = "South African journal of education",
  ["s afr j enol vitic"] = "South African Journal of Enology and Viticulture",
  ["s afr j geol"] = "South African Journal of Geology",
  ["s afr j geomatics"] = "South African Journal of Geomatics",
  ["s afr j hum rights"] = "South African journal on human rights",
  ["s afr j ind eng"] = "South African Journal of Industrial Engineering",
  ["s afr j ind psychol"] = "South African Journal of Industrial Psychology",
  ["s afr j inf manag"] = "South African journal of information management",
  ["s afr j infect dis"] = "Southern African journal of infectious diseases",
  ["s afr j lab clin med"] = "South African journal of laboratory and clinical medicine. Suid-Afrikaanse tydskrif vir laboratorium- en kliniekwerk",
  ["s afr j mar sci"] = "South African Journal of Marine Science",
  ["s afr j med sci"] = "The South African journal of medical sciences",
  ["s afr j obstet gynaecol"] = "South African journal of obstetrics and gynaecology",
  ["s afr j obstet gynaecol (1999)"] = "South African journal of obstetrics and gynaecology (1999)",
  ["s afr j physiother"] = "The South African journal of physiotherapy",
  ["s afr j plant soil"] = "South African Journal of Plant and Soil",
  ["s afr j psychiatr"] = "The South African journal of psychiatry : SAJP : the journal of the Society of Psychiatrists of South Africa",
  ["s afr j psychol"] = "South African journal of psychology = Suid-Afrikaanse tydskrif vir sielkunde",
  ["s afr j radiol"] = "South African Journal of Radiology",
  ["s afr j res sport phys educ recreat"] = "South African Journal for Research in Sport Physical Education and Recreation",
  ["s afr j sci"] = "South African journal of science",
  ["s afr j sports med"] = "South African Journal of Sports Medicine",
  ["s afr j surg"] = "South African journal of surgery. Suid-Afrikaanse tydskrif vir chirurgie",
  ["s afr j wildl res"] = "South African Journal of Wildlife Research",
  ["s afr labour bull"] = "South African labour bulletin",
  ["s afr law j"] = "South African law journal",
  ["s afr law rep"] = "South African law reports. Die Suid-Afrikaanse hofverslae. Beslissings van: Die Hooggeregshof van Suid-Afrika, Die Suidwes-Afrika Afdeling, die Hoërhof van Rhodesië. Decisions of: The Supreme Court of South Africa, the South-West Africa Division, the High Court of Rhodesia",
  ["s afr med j"] = "South African medical journal = Suid-Afrikaanse tydskrif vir geneeskunde",
  ["s afr nursing j"] = "South African nursing journal. Suid-Afrikaanse verplegingstydskrif",
  ["s afr pharm j"] = "South African pharmaceutical journal. Suid-Afrikaanse tydskrif vir apteekwese",
  ["s afr respir j"] = "South African respiratory journal",
  ["s afr sociol rev"] = "South African sociological review",
  ["s afr stat j"] = "South African statistical journal = Suid-Afrikaanse statistiese tydskrif",
  ["s afr tydskr sosiol"] = "Die Suid-afrikaanse tydskrif vir sosiologie. The South African journal of sociology",
  ["s afr tydskr strafr kriminol"] = "Suid-Afrikaanse tydskrif vir strafreg en kriminologie = South African journal of criminal law and criminology",
  ["s c dent j"] = "South Carolina dental journal",
  ["s c hist illus"] = "South Carolina history illustrated",
  ["s c hist mag"] = "South Carolina historical magazine",
  ["s c law rev"] = "South Carolina law review",
  ["s c nurs"] = "South Carolina nursing",
  ["s c nurse"] = "The South Carolina nurse",
  ["s c nurse (1994)"] = "South Carolina nurse (Columbia, S.C. : 1994)",
  ["s d farm home res"] = "South Dakota farm & home research",
  ["s d hist collect"] = "South Dakota historical collections",
  ["s d j med"] = "South Dakota journal of medicine",
  ["s d law rev"] = "South Dakota law review",
  ["s d med"] = "South Dakota medicine : the journal of the South Dakota State Medical Association",
  ["s d nurse"] = "The South Dakota nurse",
  ["s e m"] = "S. E. M",
  ["s ta nu"] = "S & TA & NU. Rivista di scienza e tecnologia degli alimenti e di nutrizione umana",
  ["sa j contin med educ"] = "The S.A. journal of continuing medical education = Die S.A. tydskrif van voortgesette geneeskundige onderrig",
  ["sa j radiol"] = "SA journal of radiology",
  ["sa nurs j"] = "SA nursing journal. SA verplegingstydskrif",
  ["saad dig"] = "SAAD digest",
  ["saalburg jahrb"] = "Saalburg-Jahrbuch",
  ["saarl arztebl"] = "Saarländisches Aerzteblatt",
  ["saas bull biochem biotechnol"] = "SAAS bulletin, biochemistry and biotechnology",
  ["saas-fee adv course"] = "Saas-Fee Advanced Course",
  ["saber (cumana)"] = "Saber (Cumaná, Venezuela)",
  ["sabrao j breed genet"] = "SABRAO journal of breeding and genetics",
  ["sac explor"] = "Sac Explorer",
  ["sacema q"] = "SACEMA quarterly",
  ["sacitra ayurveda"] = "Sacitra Āyurveda",
  ["sae int j adv curr pract mobil"] = "SAE International journal of advances and current practices in mobility",
  ["sae int j adv curr pract mobility"] = "SAE International Journal of Advances and Current Practices in Mobility",
  ["sae int j aerosp"] = "SAE International Journal of Aerospace",
  ["sae int j commer veh"] = "SAE International Journal of Commercial Vehicles",
  ["sae int j connected autom veh"] = "SAE International Journal of Connected and Automated Vehicles",
  ["sae int j electrified veh"] = "SAE International Journal of Electrified Vehicles",
  ["sae int j engines"] = "SAE International Journal of Engines",
  ["sae int j fuels lubr"] = "SAE International Journal of Fuels and Lubricants",
  ["sae int j manuf mater"] = "SAE International Journal of Manufacturing & Materials",
  ["sae int j mater manuf"] = "SAE International Journal of Materials and Manufacturing",
  ["sae int j passeng cars electron electr syst"] = "SAE International journal of passenger cars. Electronic and electrical systems",
  ["sae int j passenger cars electron electr syst"] = "SAE International Journal of Passenger Cars - Electronic and Electrical Systems",
  ["sae int j passenger cars mech syst"] = "SAE International Journal of Passenger Cars - Mechanical Systems",
  ["sae int j sustainable transp energy environ policy"] = "SAE International Journal of Sustainable Transportation, Energy, Environment, & Policy",
  ["sae int j transp cybersecur privacy"] = "SAE International Journal of Transportation Cybersecurity and Privacy",
  ["sae int j transp saf"] = "SAE international journal of transportation safety",
  ["sae int j veh dyn stab nvh"] = "SAE International journal of vehicle dynamics, stability, and NVH",
  ["sae j"] = "SAE Journal",
  ["sae q trans"] = "SAE Quarterly Transactions",
  ["sae trans"] = "SAE Transactions",
  ["saengmyeong gwahag hoeji"] = "Saengmyeong gwahag hoeji",
  ["saengyak hakhoe chi"] = "Saengyak Hakhoe chi",
  ["saf extreme environ"] = "Safety in Extreme Environments",
  ["saf health work"] = "Safety and health at work",
  ["saf reliab"] = "Safety and Reliability",
  ["saf sci"] = "Safety Science",
  ["safaids news"] = "SAfAIDS news : Southern Africa AIDS Information Dissemination Service bulletin",
  ["safe j"] = "Safe journal",
  ["safe mother"] = "Safe motherhood",
  ["safety (basel)"] = "Safety (Basel, Switzerland)",
  ["sagamore army materials res conf proc"] = "Sagamore Army Materials Research Conference Proceedings",
  ["sage benchmarks soc res methods"] = "Sage Benchmarks in Social Research Methods",
  ["sage focus ed"] = "Sage Focus Editions",
  ["sage open"] = "SAGE open",
  ["sage open eng"] = "Sage Open Engineering",
  ["sage open med"] = "SAGE open medicine",
  ["sage open med case rep"] = "SAGE open medical case reports",
  ["sage open nurs"] = "SAGE open nursing",
  ["saggi sci"] = "Saggi Scientifici",
  ["saglik bilim derg"] = "Sağlık bilimleri dergisi",
  ["sago palm"] = "Sago palm",
  ["saguenay med"] = "Le Saguenay médical",
  ["sah bull"] = "SAH Bulletin",
  ["sahara (segrate)"] = "Sahara (Segrate, Italy)",
  ["sahara j"] = "SAHARA J : journal of Social Aspects of HIV/AIDS Research Alliance",
  ["saibo kogaku"] = "Saibo Kogaku",
  ["saiee afr res j"] = "SAIEE Africa Research Journal",
  ["sains indones"] = "Sains Indonesia",
  ["sains malays"] = "Sains Malaysiana",
  ["sairaanh vuosik"] = "Sairaanhoidon vuosikirja",
  ["sais rev"] = "SAIS review (Paul H. Nitze School of Advanced International Studies)",
  ["saishin igaku"] = "Saishin igaku. Modern medicine",
  ["saisons alsace"] = "Saisons d'Alsace",
  ["saitama math j"] = "Saitama Mathematical Journal",
  ["salamandra (frankf)"] = "Salamandra (Frankfurt)",
  ["saline syst"] = "Saline systems",
  ["saline systems"] = "Saline systems",
  ["salisbury med bull"] = "Salisbury medical bulletin",
  ["sallim kwahak yongu"] = "Sallim kwahak yŏn'gu",
  ["salubr asist"] = "Salubridad y asistencia",
  ["salud bucal"] = "Salud bucal",
  ["salud colect"] = "Salud colectiva",
  ["salud drogas"] = "Salud y drogas",
  ["salud ment (mex)"] = "Salud mental (Mexico City, Mexico)",
  ["salud publica mex"] = "Salud pública de México",
  ["salud pública méx"] = "Salud Pública de México",
  ["salud reprod soc"] = "Salud reproductiva y sociedad : órgano informativo del Programa Salud Reproductiva y Sociedad de El Colegio de México",
  ["salud soc"] = "Salud y Sociedad",
  ["salud uis"] = "Salud UIS : revista de la Universidad Industrial de Santander, Facultad de Salud",
  ["salud vida"] = "Salud y vida",
  ["salzburger beitr paracelsusforsch"] = "Salzburger Beiträge zur Paracelsusforschung",
  ["same day surg"] = "Same-day surgery",
  ["samid aliqtisadi"] = "Ṣāmid al-iqtiṣādī = Samed",
  ["samml tusculum"] = "Sammlung Tusculum",
  ["samml zwangl abh geb psychiatr neurol"] = "Sammlung zwangloser Abhandlungen aus dem Gebiete der Psychiatrie und Neurologie",
  ["samoa med j"] = "Samoa medical journal",
  ["sampe j"] = "SAMPE Journal",
  ["sampl theory signal image process"] = "Sampling Theory in Signal and Image Processing. An International Journal",
  ["sampl theory signal process data anal"] = "Sampling Theory, Signal Processing, and Data Analysis",
  ["sample regist bull"] = "Sample registration bulletin",
  ["sampling theory, signal process, data anal"] = "Sampling Theory, Signal Processing, and Data Analysis",
  ["san diego law rev"] = "The San Diego law review",
  ["san fernando val dent soc bull"] = "San Fernando Valley Dental Society bulletin",
  ["san franc med"] = "San Francisco medicine",
  ["san franc state univ ser philos"] = "San Francisco State University Series in Philosophy",
  ["san gabriel val dent soc bull"] = "San Gabriel Valley Dental Society bulletin",
  ["san jose stud"] = "San José studies",
  ["sanfujinka chiryo"] = "[Sanfujinka chiryō] Obstetrical and gynecological therapy",
  ["sanfujinka no jissai"] = "Sanfujinka no jissai. Practice of gynecology and obstetrics",
  ["sanfujinka no shinpo"] = "[Sanfujinka no shinpo] Advances in obstetrics and gynecology",
  ["sangre (barc)"] = "Sangre",
  ["sangyo eiseigaku zasshi"] = "Sangyō eiseigaku zasshi = Journal of occupational health",
  ["sangyo igaku"] = "Sangyō igaku. Japanese journal of industrial health",
  ["sangyo ika daigaku zasshi"] = "Sangyo Ika Daigaku Zasshi",
  ["sanid aeronaut"] = "Sanidad aeronáutica : organo informativo técnico",
  ["sanid benefic munic"] = "Sanidad y beneficencia municipal; revista médico social",
  ["sanit rev j public health"] = "The Sanitary review and journal of public health",
  ["sanita pubblica"] = "Sanità pubblica",
  ["sankhya a"] = "Sankhya A. The Indian Journal of Statistics",
  ["sankhya b"] = "Sankhya B. Applied and Interdisciplinary Statistics",
  ["sankhya b (2008)"] = "Sankhyā. Series B (2008)",
  ["sankhya ser a"] = "Sankhyā. Series A. (2008)",
  ["sankhya ser b"] = "Sankhyā. Series B. [Methodological.]",
  ["sankyo kenkyusho nenpo"] = "Sankyo Kenkyusho Nenpo",
  ["sanop misaengmul hakhoe chi"] = "Sanŏp Misaengmul Hakhoe chi",
  ["sanop misaengmul hakhoechi"] = "Sanop Misaengmul Hakhoechi",
  ["santa clara comput high technol law j"] = "Santa Clara computer and high-technology law journal",
  ["santa clara law rev"] = "Santa Clara law review",
  ["santa clara lawyer"] = "Santa Clara lawyer (1961)",
  ["santa fe inst stud sci complexity lecture notes"] = "Santa Fe Institute Studies in the Sciences of Complexity",
  ["santa fe inst stud sci complexity proc"] = "Santa Fe Institute Studies in the Sciences of Complexity",
  ["sante homme"] = "La Santé de l'homme",
  ["sante ment que"] = "Santé mentale au Québec",
  ["sante publique"] = "Santé publique (Vandoeuvre-lès-Nancy, France)",
  ["sante publique (bucur)"] = "La Santé publique",
  ["sante publique (paris)"] = "Santé publique",
  ["sante que"] = "Santé Québec : revue de la Corporation professionnelle des infirmières et infirmiers auxiliaires du Québec",
  ["sante salud"] = "Santé Salud",
  ["sao palo j math sci"] = "Sao Paulo Journal of Mathematical Sciences",
  ["sao paulo j math sci"] = "Sao Paulo Journal of Mathematical Sciences",
  ["sao paulo med"] = "São Paulo médico",
  ["sao paulo med j"] = "São Paulo medical journal = Revista paulista de medicina",
  ["sapporo igaku zasshi"] = "Sapporo igaku zasshi. The Sapporo medical journal",
  ["sar qsar environ res"] = "SAR and QSAR in environmental research",
  ["sarajevo j math"] = "Sarajevo Journal of Mathematics",
  ["sarcoidosis vasc diffuse lung dis"] = "Sarcoidosis, vasculitis, and diffuse lung diseases : official journal of WASOG",
  ["sas glob forum"] = "SAS global forum",
  ["sas j"] = "SAS journal",
  ["sask hist"] = "Saskatchewan history",
  ["sask law rev"] = "Saskatchewan law review",
  ["satell navig"] = "Satellite Navigation",
  ["saturday rev"] = "Saturday review",
  ["saturday rev sci"] = "Saturday review of the sciences",
  ["saturday rev world"] = "Saturday review/world",
  ["saudi dent j"] = "The Saudi dental journal",
  ["saudi j anaesth"] = "Saudi journal of anaesthesia",
  ["saudi j biol sci"] = "Saudi journal of biological sciences",
  ["saudi j dent res"] = "The Saudi journal for dental research",
  ["saudi j gastroenterol"] = "Saudi journal of gastroenterology : official journal of the Saudi Gastroenterology Association",
  ["saudi j health sci"] = "Saudi journal for health sciences",
  ["saudi j kidney dis transpl"] = "Saudi journal of kidney diseases and transplantation : an official publication of the Saudi Center for Organ Transplantation, Saudi Arabia",
  ["saudi j med med sci"] = "Saudi journal of medicine & medical sciences",
  ["saudi j ophthalmol"] = "Saudi journal of ophthalmology : official journal of the Saudi Ophthalmological Society",
  ["saudi med j"] = "Saudi medical journal",
  ["saudi pharm j"] = "Saudi pharmaceutical journal : SPJ : the official publication of the Saudi Pharmaceutical Society",
  ["sauvegarde enfance (paris, 1946)"] = "Sauvegarde de l'enfance",
  ["savasta erbaslar"] = "Savaşta erbaşlar",
  ["savoir mantice"] = "Le Savoir de Mantice",
  ["savoirs actuels"] = "Savoirs Actuels",
  ["savoirs sci prat enseign"] = "Savoirs Scientifiques & Pratiques d’Enseignement",
  ["sb arch pr"] = "Sborník archivních prací",
  ["sb lek"] = "Sbornik Lekarsky",
  ["sb math"] = "Sbornik. Mathematics",
  ["sb nar muz pr rada c lit hist"] = "Sborník Národního muzea v Praze. Řada C, Literární historie = Acta Musei Nationalis Pragae. Series C, Historia litterarum",
  ["sb nar muz praze rada a hist"] = "Sborník Národního musea v Praze. A, Historický = Acta Musei Nationalis Prague. A, Historia",
  ["sb nar muz praze rada b"] = "Sborník Národního musea v Praze. B, Přírodovědný = Acta Musei Nationalis Pragae. B, Hist. naturalis",
  ["sb pathofysiol traveni vyz gastroenterol bohema"] = "Sborník pro pathofysiologii trávení a výz̆ivy; gastroenterologia bohema",
  ["sb pr filos fak brnenske univ rada hist"] = "Sborník prací Filosofické fakulty brněnské university. C, Řada historická",
  ["sb tr azerbaidzhanskii gos meditsinskii inst n narimanova"] = "Sbornik trudov. Azerbaĭdzhanskiĭ gosudarstvennyĭ medit︠s︡inskiĭ institut",
  ["sb ved pr lek fak karlovy univerzity hradci kralove"] = "Sborník vědeckých prací Lékařské fakulty Karlovy university v Hradci Králové",
  ["sb ved pr lek fak karlovy univerzity hradci kralove suppl"] = "Sborník vědeckých prací Lékařské fakulty Karlovy univerzity v Hradci Králové. Supplementum",
  ["sb ved pr, vys sk chemickotechnol pardubice"] = "Sbornik Vedeckych Praci, Vysoka Skola Chemickotechnologicka Pardubice",
  ["sb věd odb prací"] = "Vysoké učení Technické v Brně",
  ["sbornik math"] = "Sbornik Mathematics",
  ["sbs ces geogr spol"] = "Sborník České geografické společnosti",
  ["sc rep"] = "SC [reports]. U.S. Atomic Energy Commission",
  ["sc trodent"] = "SC Trodent",
  ["scada j"] = "Scada Journal",
  ["scalable comput commun"] = "Scalable Computing and Communications",
  ["scalpel (brux)"] = "Le Scalpel",
  ["scalpel alpha epsil delta"] = "The Scalpel of Alpha Epsilon Delta",
  ["scalpel tongs"] = "Scalpel & tongs : American journal of medical philately",
  ["scan electron microsc"] = "Scanning electron microscopy",
  ["scan j des hist"] = "Scandinavian journal of design history",
  ["scan polit stud"] = "Scandinavian political studies",
  ["scand actuar j"] = "Scandinavian actuarial journal",
  ["scand actuarial j"] = "Scandinavian Actuarial Journal",
  ["scand audiol"] = "Scandinavian audiology",
  ["scand audiol suppl"] = "Scandinavian audiology. Supplementum",
  ["scand cardiovasc j"] = "Scandinavian cardiovascular journal : SCJ",
  ["scand cardiovasc j suppl"] = "Scandinavian cardiovascular journal. Supplement",
  ["scand econ hist rev"] = "The Scandinavian economic history review",
  ["scand j caring sci"] = "Scandinavian journal of caring sciences",
  ["scand j child adolesc psychiatr psychol"] = "Scandinavian journal of child and adolescent psychiatry and psychology",
  ["scand j clin lab invest"] = "Scandinavian journal of clinical and laboratory investigation",
  ["scand j clin lab invest suppl"] = "Scandinavian journal of clinical and laboratory investigation. Supplementum",
  ["scand j dent res"] = "Scandinavian journal of dental research",
  ["scand j dev altern"] = "Scandinavian journal of development alternatives",
  ["scand j disabil res"] = "Scandinavian journal of disability research : SJDR",
  ["scand j econ"] = "The Scandinavian journal of economics",
  ["scand j food nutr"] = "Scandinavian journal of food & nutrition",
  ["scand j for res"] = "Scandinavian Journal of Forest Research",
  ["scand j gastroenterol"] = "Scandinavian journal of gastroenterology",
  ["scand j gastroenterol suppl"] = "Scandinavian journal of gastroenterology. Supplement",
  ["scand j haematol"] = "Scandinavian journal of haematology",
  ["scand j haematol suppl"] = "Scandinavian journal of haematology. Supplementum",
  ["scand j hist"] = "Scandinavian journal of history",
  ["scand j immunol"] = "Scandinavian journal of immunology",
  ["scand j immunol suppl"] = "Scandinavian journal of immunology. Supplement",
  ["scand j infect dis"] = "Scandinavian journal of infectious diseases",
  ["scand j infect dis suppl"] = "Scandinavian journal of infectious diseases. Supplementum",
  ["scand j lab anim sci"] = "Scandinavian journal of laboratory animal science = Scand las nyt : official quarterly journal of the Scandinavian Federation for Laboratory Animal Science",
  ["scand j med sci sports"] = "Scandinavian journal of medicine & science in sports",
  ["scand j metall"] = "Scandinavian Journal of Metallurgy",
  ["scand j occup ther"] = "Scandinavian journal of occupational therapy",
  ["scand j old testam"] = "Scandinavian journal of the Old Testament : SJOT",
  ["scand j pain"] = "Scandinavian journal of pain",
  ["scand j plast reconstr surg"] = "Scandinavian journal of plastic and reconstructive surgery",
  ["scand j plast reconstr surg hand surg"] = "Scandinavian journal of plastic and reconstructive surgery and hand surgery",
  ["scand j plast reconstr surg hand surg suppl"] = "Scandinavian journal of plastic and reconstructive surgery and hand surgery. Supplementum",
  ["scand j plast reconstr surg suppl"] = "Scandinavian journal of plastic and reconstructive surgery. Supplementum",
  ["scand j prim health care"] = "Scandinavian journal of primary health care",
  ["scand j prim health care suppl"] = "Scandinavian journal of primary health care. Supplement",
  ["scand j psychol"] = "Scandinavian journal of psychology",
  ["scand j public health"] = "Scandinavian journal of public health",
  ["scand j public health suppl"] = "Scandinavian journal of public health. Supplement",
  ["scand j rehabil med"] = "Scandinavian journal of rehabilitation medicine",
  ["scand j rehabil med suppl"] = "Scandinavian journal of rehabilitation medicine. Supplement",
  ["scand j respir dis"] = "Scandinavian journal of respiratory diseases",
  ["scand j respir dis suppl"] = "Scandinavian journal of respiratory diseases. Supplementum",
  ["scand j rheumatol"] = "Scandinavian journal of rheumatology",
  ["scand j rheumatol suppl"] = "Scandinavian journal of rheumatology. Supplement",
  ["scand j soc med"] = "Scandinavian journal of social medicine",
  ["scand j soc med suppl"] = "Scandinavian journal of social medicine. Supplementum",
  ["scand j soc welf"] = "Scandinavian journal of social welfare",
  ["scand j stat"] = "Scandinavian Journal of Statistics",
  ["scand j statist"] = "Scandinavian Journal of Statistics",
  ["scand j surg"] = "Scandinavian journal of surgery : SJS : official organ for the Finnish Surgical Society and the Scandinavian Surgical Society",
  ["scand j thorac cardiovasc surg"] = "Scandinavian journal of thoracic and cardiovascular surgery",
  ["scand j thorac cardiovasc surg suppl"] = "Scandinavian journal of thoracic and cardiovascular surgery. Supplementum",
  ["scand j trauma resusc emerg med"] = "Scandinavian journal of trauma, resuscitation and emergency medicine",
  ["scand j urol"] = "Scandinavian journal of urology",
  ["scand j urol nephrol"] = "Scandinavian journal of urology and nephrology",
  ["scand j urol nephrol suppl"] = "Scandinavian journal of urology and nephrology. Supplementum",
  ["scand j work environ health"] = "Scandinavian journal of work, environment & health",
  ["scand popul stud"] = "Scandinavian population studies",
  ["scand soc forensic odontol newsl"] = "Scandinavian Society of Forensic Odontology Newsletter",
  ["scand stat theory appl"] = "Scandinavian journal of statistics, theory and applications",
  ["scand stud"] = "Scandinavian studies : publication of the Society for the Advancement of Scandinavian Study",
  ["scand stud law"] = "Scandinavian studies in law",
  ["scanning microsc"] = "Scanning microscopy",
  ["scanning microsc nanotechnol"] = "Scanning Microscopy for Nanotechnology",
  ["scanning microsc suppl"] = "Scanning microscopy. Supplement",
  ["scapel quill"] = "Scalpel and quill",
  ["scars burn heal"] = "Scars, burns & healing",
  ["sch comm j"] = "School community journal",
  ["sch couns"] = "The School counselor",
  ["sch dent serv gaz n z"] = "School Dental Service gazette, New Zealand",
  ["sch eff sch improv"] = "School effectiveness and school improvement : an international journal of research, policy and practice",
  ["sch health rev"] = "School health review",
  ["sch inq nurs pract"] = "Scholarly inquiry for nursing practice",
  ["sch psychol"] = "School psychology (Washington, D.C.)",
  ["sch psychol dig"] = "The School psychology digest",
  ["sch psychol forum res pract"] = "School Psychology Forum, research in practice",
  ["sch psychol int"] = "School Psychology International",
  ["sch psychol q"] = "School psychology quarterly : the official journal of the Division of School Psychology, American Psychological Association",
  ["sch sci math"] = "School science and mathematics",
  ["sch sci rev"] = "The School science review",
  ["sch soc work j"] = "School social work journal",
  ["scha j"] = "Scha Journal",
  ["schaffhauser beitr vaterl gesch"] = "Schaffhauser Beiträge zur vaterländischen Geschichte",
  ["schaum’s outline ser comput"] = "Schaum’s Outline Series in Computers",
  ["schiff zeit"] = "Schiff und Zeit",
  ["schizophr bull"] = "Schizophrenia bulletin",
  ["schizophr bull open"] = "Schizophrenia bulletin open",
  ["schizophr res"] = "Schizophrenia research",
  ["schizophr res cogn"] = "Schizophrenia research. Cognition",
  ["schizophr res treatment"] = "Schizophrenia research and treatment",
  ["schmalenbachs z betriebswirtsch forsch"] = "Schmalenbachs Zeitschrift fur betriebswirtschaftliche Forschung",
  ["schmollers jahr"] = "Schmollers Jahrbuch für Wirtschafts- und Sozialwissenschaften",
  ["schmollers jahr: z wirtsch sozialwissen"] = "Schmollers Jahrbuch: Zeitschrift für Wirtschafts- und Sozialwissenschaften",
  ["scholarpedia j"] = "Scholarpedia journal",
  ["scholars acad j biosci"] = "Scholars academic journal of biosciences",
  ["scholarsh pract undergrad res"] = "Scholarship and practice of undergraduate research",
  ["scholarsh teach learn psychol"] = "Scholarship of teaching and learning in psychology",
  ["scholast update"] = "Scholastic update (Teachers' ed.)",
  ["school ment health"] = "School mental health",
  ["school nurse news"] = "School nurse news",
  ["school psych rev"] = "School psychology review",
  ["schr carl friedrich von siemens stift"] = "Schriften der Carl Friedrich von Siemens Stiftung",
  ["schr forstl fak univ gött niedersächs  forstl versanst"] = "Schriften aus der Forstlichen Fakultät der Universität Göttingen und der Niedersächsischen forstlichen Versuchsanstalt",
  ["schr genet ressour"] = "Schriften zu genetischen Ressourcen",
  ["schr gesch kult alten orients"] = "Schriften zur Geschichte und Kultur des alten Orients",
  ["schr koenigsb gelehrten ges, naturwiss kl"] = "Schriften der Koenigsberger Gelehrten Gesellschaft, Naturwissenschaftliche Klasse",
  ["schr konigsb gelehrten ges"] = "Schriften der Konigsberger Gelehrten Gesellschaft",
  ["schr malakozool haus nat cismar"] = "Schriften zur Malakozoologie aus dem Haus der Natur, Cismar",
  ["schr naturwissen verbr naturwissen kennt wien"] = "Schriften des Vereines zur Verbreitung Naturwissenschaftlicher Kenntnisse in Wien",
  ["schr wirtschinform"] = "Schriften zur Wirtschaftsinformatik",
  ["schrift adam-ries-bundes annaberg-buchholz"] = "Schriften des Adam-Ries-Bundes Annaberg-Buchholz",
  ["schrift mus univ tubingen (mut)"] = "Schriften des Museums der Universitat Tubingen (MUT)",
  ["schrift wirtschaftspolit nf"] = "Schriften zur Wirtschaftspolitik. Neue Folge",
  ["schriften wirtschaftswiss forsch"] = "Schriften zur Wirtschaftswissenschaftlichen Forschung",
  ["schriftenr erfahr dtsch sanit zweiten weltkrieg"] = "Schriftenreihe Erfahrungen des deutschen Sanitätsdienstes im Zweiten Weltkrieg",
  ["schriftenr geb off gesundheitswes"] = "Schriftenreihe aus dem Gebiete des öffentlichen Gesundheitswesens",
  ["schriftenr geol wissen"] = "Schriftenreihe fur Geologische Wissenschaften",
  ["schriftenr ges freunde med hochsch hannover"] = "Schriftenreihe. Gesellschaft der Freunde der Medizinischen Hochschule Hannover",
  ["schriftenr gesch versamml dtsch naturforsch arzte"] = "Schriftenreihe zur Geschichte der Versammlungen deutscher Naturforscher und Aerzte",
  ["schriftenr inst systdyn (isd) opt syst (ios)"] = "Schriftenreihe der Institute fur Systemdynamik (ISD) und optische Systeme (IOS)",
  ["schriftenr neurol"] = "Schriftenreihe Neurologie",
  ["schriftenr ver wasser boden lufthyg"] = "Schriftenreihe des Vereins für Wasser-, Boden- und Lufthygiene",
  ["schriftenr zentralbl arbeitsmed arbeitsschutz prophyl"] = "Schriftenreihe Zentralblatt für Arbeitsmedizin, Arbeitsschutz und Prophylaxe",
  ["schriftenr zentralbl arbeitsmed arbeitsschutz prophyl ergonomie"] = "Schriftenreihe Zentralblatt fur Arbeitsmedizin, Arbeitsschutz, Prophylaxe, und Ergonomie",
  ["schriftenreihe der institut fur mathematik bei der deutschen akademie der wissenschaften zu berlin, reihe a reine mat"] = "Schriftenreihe der Institut fur Mathematik bei der Deutschen Akademie der Wissenschaften zu Berlin, Reihe A. Reine Mathematik",
  ["schriftenreihe der institut fur mathematik bei der deutschen akademie der wissenschaften zu berlin, reihe b angewandt"] = "Schriftenreihe der Institut fur Mathematik bei der Deutschen Akademie der Wissenschaften zu Berlin, Reihe B. Angewandte Mathematik und Mechanik",
  ["schriftenreihe inst empirische wirtschaftsforsch univ zurich"] = "Schriftenreihe des Instituts fur Empirische Wirtschaftsforschung der Universitat Zurich",
  ["schriftenreihe math inst univ münster 3 ser"] = "Schriftenreihe des Mathematischen Instituts der Universität Münster, 3",
  ["schriftenreihe pädagog hochsch heidelberg"] = "Schriftenreihe der Pädagogischen Hochschule Heidelberg",
  ["schriftenreihe versicher"] = "Schriftenreihe “Die Versicherung”",
  ["schriftenreihe wiss frieden"] = "Schriftenreihe “Wissenschaft und Frieden”",
  ["schrreihe bad forstl versanst, freibg"] = "Schriftenreihe der Badischen Forstlichen Versuchsanstalt Freiburg",
  ["schrreihe bayer landesamt umweltschutz"] = "Schriftenreihe Bayerisches Landesamt für Umweltschutz",
  ["schrreihe bayer staatsminist ernähr  landwirtsch forsten"] = "Schriftenreihe des Bayerischen Staatsministeriums für Ernährung, Landwirtschaft und Forsten",
  ["schrreihe dtsch rat landespfl"] = "Schriftenreihe des Deutschen Rates für Landespflege",
  ["schrreihe forstwiss fak albert-ludwigs-univ  freibg ib"] = "Schriftenreihe der Forstwissenschaftlichen Fakultät der Albert-Ludwigs-Universität Freiburg i. Br.",
  ["schrreihe inst landespfl univ freibg"] = "Schriftenreihe des Institutes für Landespflege der Universität Freiburg",
  ["schrreihe landesanst ökol landschentwickl forstplan nordrh-westfal"] = "Schriftenreihe der Landesanstalt für Ökologie, Landschaftsentwicklung und Forstplanung Nordrhein-Westfalen",
  ["schrreihe landesforstverwalt baden-württ"] = "Schriftenreihe der Landesforstverwaltung Baden-Württemberg",
  ["schrreihe landesstelle natschutz landschpfl  nordrh-westfal"] = "Schriftenreihe der Landesstelle für Naturschutz und Landschaftspflege in Nordrhein-Westfalen",
  ["schrreihe landschpfl natschutz"] = "Schriftenreihe für Landschaftspflege und Naturschutz",
  ["schrreihe lebensraum vorarlberg"] = "Schriftenreihe Lebensraum Vorarlberg",
  ["schrreihe umw"] = "Schriftenreihe Umwelt",
  ["schrreihe vegkd"] = "Schriftenreihe für Vegetationskunde",
  ["schs stud"] = "SCHS studies",
  ["schweiss prueftech"] = "Schweiss- und Prueftechnik",
  ["schweiz aktuarver mitt"] = "Schweizerische Aktuarvereinigung",
  ["schweiz apoth ztg"] = "Schweizerische Apotheker Zeitung",
  ["schweiz arch neurol neurochir psychiatr"] = "Schweizer Archiv für Neurologie, Neurochirurgie und Psychiatrie = Archives suisses de neurologie, neurochirurgie et de psychiatrie",
  ["schweiz arch neurol psychiatr"] = "Schweizer Archiv für Neurologie und Psychiatrie. Archives suisses de neurologie et de psychiatrie. Archivio svizzero di neurologia e psichiatria",
  ["schweiz arch neurol psychiatr (1985)"] = "Schweizer Archiv für Neurologie und Psychiatrie (Zurich, Switzerland : 1985)",
  ["schweiz arch tierheilkd"] = "Schweizer Archiv für Tierheilkunde",
  ["schweiz arzteztg"] = "Schweizerische Ärztezeitung. Bulletin des médecins suisses. Bollettino dei medici svizzeri",
  ["schweiz arzteztg standesfr"] = "Schweizerische Ärztezeitung für Standesfragen. Bulletin professionnel des médecins suisses. Bollettino dei medici svizzeri per interessi professionali",
  ["schweiz beitr dendrol"] = "Schweizerische Beiträge zur Dendrologie",
  ["schweiz bl krankenpfl"] = "Schweizerische Blätter für Krankenpflege. Revue suisse des infirmières",
  ["schweiz drog ztg (1901)"] = "Schweizerische Drogisten Zeitung",
  ["schweiz förster"] = "Schweizer Förster, Der",
  ["schweiz hebamme"] = "Die Schweizer Hebamme",
  ["schweiz holz-börse"] = "Schweizer Holz-Börse",
  ["schweiz holzwirtsch"] = "Schweizer Holzwirtschaft",
  ["schweiz holzztg"] = "Schweizer Holzzeitung",
  ["schweiz ing archit"] = "Schweizer Ingenieur und Architekt",
  ["schweiz interessengem indholz, anleit"] = "Schweizerische Interessengemeinschaft Industrieholz, Anleitung",
  ["schweiz krankenkassen ztg"] = "Schweizerische Krankenkassen-Zeitung",
  ["schweiz lab z"] = "Schweizerische Laboratoriums Zeitschrift",
  ["schweiz landtech"] = "Schweizer Landtechnik",
  ["schweiz landwirtsch forsch"] = "Schweizerische landwirtschaftliche Forschung",
  ["schweiz med jahrb"] = "Schweizerisches medizinisches jahrbuch",
  ["schweiz med wochenschr"] = "Schweizerische medizinische Wochenschrift",
  ["schweiz med wochenschr suppl"] = "Schweizerische medizinische Wochenschrift. Supplementum",
  ["schweiz mineral petrogr mitt"] = "Schweizerische Mineralogische und Petrographische Mitteilungen",
  ["schweiz monatsh"] = "Schweizer Monatshefte",
  ["schweiz monatsschr zahnmed"] = "Schweizer Monatsschrift für Zahnmedizin = Revue mensuelle suisse d'odonto-stomatologie = Rivista mensile svizzera di odontologia e stomatologia",
  ["schweiz monatsschr zahnmed (1984)"] = "Schweizerische Monatsschrift für Zahnmedizin = Revue mensuelle suisse d'odonto-stomatologie = Rivista mensile svizzera di odontologia e stomatologia",
  ["schweiz nationalbank quart"] = "Schweizerische Nationalbank Quartalsheft",
  ["schweiz natschutz"] = "Schweizer Naturschutz",
  ["schweiz opt opt suisse"] = "Der Schweizer Optiker. L'Opticien suisse",
  ["schweiz rundsch med prax"] = "Schweizerische Rundschau für Medizin Praxis = Revue suisse de médecine Praxis",
  ["schweiz schreinerztg"] = "Schweizerische Schreinerzeitung",
  ["schweiz volkskd"] = "Schweizer Volkskunde",
  ["schweiz wald"] = "Schweizer Wald",
  ["schweiz wochenschr chem pharm"] = "Schweizerische Wochenschrift fur Chemie und Pharmacie",
  ["schweiz z forstwes"] = "Schweizerische Zeitschrift für Forstwesen",
  ["schweiz z gesch"] = "Schweizerische Zeitschrift für Geschichte. Revue suisse d'histoire. Rivista storica svizzera",
  ["schweiz z gynakol geburtshilfe"] = "Schweizerische Zeitschrift für Gynäkologie und Geburtshilfe. Revue suisse de gynécologie et d'obstétrique",
  ["schweiz z homoopath"] = "Schweizerische Zeitschrift für Homöopathie. Journal suisse d'homoeopathie",
  ["schweiz z hydrol"] = "Schweizerische Zeitschrift für Hydrologie",
  ["schweiz z med traumatol"] = "Schweizerische Zeitschrift für Medizin und Traumatologie = Revue suisse pour médecine et traumatologie",
  ["schweiz z pathol bakteriol"] = "Schweizerische Zeitschrift für Pathologie und Bakteriologie. Revue suisse de pathologie et de bactériologie",
  ["schweiz z pilzkd"] = "Schweizerische zeitschrift für pilzkunde",
  ["schweiz z psychol anwend"] = "Schweizerische Zeitschrift für Psychologie und ihre Andwendungen. Revue suisse de psychologie, pure et appliquée",
  ["schweiz z soziol"] = "Schweizerische Zeitschrift für Soziologie. Revue suisse de sociologie",
  ["schweiz z sportmed"] = "Schweizerische Zeitschrift fur Sportmedizin",
  ["schweiz z tuberc pneumonol"] = "Schweizerische Zeitschrift für Tuberkulose und Pneumonologie. Revue suisse de la tuberculose et de pneumonologie. Rivista svizzera della tubercolosi e della pneumonologia",
  ["schweiz z tuberk"] = "Schweizerische Zeitschrift für Tuberkulose. Revue suisse de la tuberculose. Rivista svizzera della tubercolosi",
  ["schweiz z volkswirtsch stat"] = "Schweizerische Zeitschrift für Volkswirtschaft und Statistik",
  ["schweiz z volkswirtsch statist"] = "Schweizerische Zeitschrift für Volkswirtschaft und Statistik",
  ["schweiz z zahnheilkd"] = "Schweizerische Zeitschrift für Zahnheilkunde. Journal suisse d'odontologie. Odontologia svizzera. Swiss journal of dental surgery",
  ["schweizer chem ztg tech ind"] = "Schweizer Chemiker-Zeitung und Technik-Industrie",
  ["schweiz  ges phytomed, info"] = "Schweizerische Gesellschaft für Phytomedizin, Info",
  ["schwest rev"] = "Schwestern Revue",
  ["sci act"] = "Science activities",
  ["sci adv"] = "Science advances",
  ["sci adv mater"] = "Science of advanced materials",
  ["sci adv mater science of advanced materials"] = "Sci. Adv. Mater. Science of Advanced Materials",
  ["sci afr"] = "Scientific African",
  ["sci aging knowledge environ"] = "Science of aging knowledge environment : SAGE KE",
  ["sci agrar"] = "Scientia agraria",
  ["sci agric"] = "Scientia agricola",
  ["sci aliments"] = "Sciences des aliments",
  ["sci am"] = "Scientific American",
  ["sci am mind"] = "Scientific American mind",
  ["sci am neurol"] = "Scientific American neurology",
  ["sci amer"] = "Scientific American",
  ["sci amer lib paperback"] = "Scientific American Library Paperback",
  ["sci ann comput sci"] = "Scientific Annals of Computer Science",
  ["sci ann danube delta inst"] = "Scientific Annals of the Danube Delta Institute",
  ["sci asia"] = "ScienceAsia : journal of the Science Society of Thailand",
  ["sci avenir"] = "Sciences et avenir",
  ["sci basis med annu rev"] = "The Scientific basis of medicine annual reviews",
  ["sci bio j"] = "Science Of Biology Journal",
  ["sci bull"] = "Science Bulletin",
  ["sci bull (beijing)"] = "Science bulletin",
  ["sci bull fac ed nagasaki univ"] = "Science Bulletin of the Faculty of Education",
  ["sci bull josai univ"] = "Science Bulletin of Josai University",
  ["sci bull sci found philipp"] = "Science bulletin (Kuo chia kʻo hsüeh wei yüan hui)",
  ["sci can"] = "Scientia canadensis",
  ["sci china a: math"] = "Science in China, Series A: Mathematics",
  ["sci china b"] = "Science in China. Series B, Chemistry, life sciences & earth sciences",
  ["sci china b chem"] = "Science in China. Series B, Chemistry",
  ["sci china c life sci"] = "Science in China. Series C, Life sciences",
  ["sci china c: life sci"] = "Science in China, Series C: Life Sciences",
  ["sci china chem"] = "Science China. Chemistry",
  ["sci china d: earth sci"] = "Science in China, Series D: Earth Sciences",
  ["sci china earth sci"] = "Science China. Earth sciences",
  ["sci china g: phys mech astron"] = "Science in China, Series G: Physics, Mechanics and Astronomy",
  ["sci china inf sci"] = "Science China Information Sciences",
  ["sci china life sci"] = "Science China. Life sciences",
  ["sci china mater"] = "Science China Materials",
  ["sci china math"] = "Science China. Mathematics",
  ["sci china phys mech astron"] = "Science China. Physics, mechanics & astronomy",
  ["sci china phys, mech astron"] = "Science China Physics, Mechanics & Astronomy",
  ["sci china ser a"] = "Science in China (Scientia Sinica)",
  ["sci china ser a math"] = "Science in China. Series A, Mathematics",
  ["sci china ser a math phys astron"] = "Science in China. Series A, Mathematics, physics, astronomy",
  ["sci china ser d earth sci"] = "Science in China. Series D, Earth sciences",
  ["sci china ser e"] = "Science in China (Scientia Sinica)",
  ["sci china ser e: technol sci"] = "Science in China Series E: Technological Sciences",
  ["sci china technol sci"] = "Science China. Technological sciences",
  ["sci china, ser a math"] = "Science in China, Series A: Mathematics",
  ["sci china, ser a: math"] = "Science in China, Series A: Mathematics",
  ["sci china, ser b chem"] = "Science in China, Series B: Chemistry",
  ["sci china, ser b: chem"] = "Science in China, Series B: Chemistry, Life Sciences, & Earth Sciences",
  ["sci china, ser c life sci"] = "Science in China, Series C: Life Sciences",
  ["sci china, ser c: life sci"] = "Science in China Series C: Life Sciences",
  ["sci china, ser d earth sci"] = "Science in China, Series D: Earth Sciences",
  ["sci china, ser d: earth sci"] = "Science in China Series D: Earth Sciences",
  ["sci china, ser e eng mater sci"] = "Science in China, Series E: Engineering & Materials Science",
  ["sci china, ser e: tech sci"] = "Science in China, Series E: Technological Sciences",
  ["sci china, ser f: info sci"] = "Science in China Series F: Information Sciences",
  ["sci china, ser g"] = "Science in China, Series G: Physics Mechanics and Astronomy",
  ["sci china, ser g: phys, mech, astron"] = "Science in China Series G: Physics, Mechanics & Astronomy",
  ["sci china: chem"] = "Science China: Chemistry",
  ["sci christ belief"] = "Science & Christian belief",
  ["sci citizen"] = "Scientist and citizen",
  ["sci cold arid reg"] = "Sciences in cold and arid regions",
  ["sci commun"] = "Science communication",
  ["sci comput"] = "Scientific Computation",
  ["sci comput program"] = "Science of computer programming",
  ["sci comput programming"] = "Science of Computer Programming",
  ["sci concept found"] = "Science and its Conceptual Foundations",
  ["sci concepts probl"] = "Sciences: Concepts et Problemes",
  ["sci conserv"] = "Science for conservation (Wellington)",
  ["sci context"] = "Science in context",
  ["sci cult"] = "Science and culture",
  ["sci cult (lond)"] = "Science as culture",
  ["sci data"] = "Scientific data",
  ["sci dig"] = "Science digest",
  ["sci dig (ny)"] = "Science digest (New York, N.Y.)",
  ["sci diliman"] = "Science Diliman",
  ["sci dimens"] = "Science Dimension",
  ["sci dipl"] = "Science & diplomacy",
  ["sci drill"] = "Scientific Drilling",
  ["sci econom"] = "Sciences Economiques",
  ["sci educ"] = "Science education",
  ["sci educ (arlingt)"] = "Science educator : the National Science Education Leadership Association journal",
  ["sci educ (dordr)"] = "Science & education",
  ["sci educ bull"] = "Scientific And Educational Bulletin",
  ["sci educ civ engagem"] = "Science education & civic engagement : an international journal",
  ["sci educ j"] = "Scientific and educational journal",
  ["sci eng compos mater"] = "Science and Engineering of Composite Materials",
  ["sci eng ethics"] = "Science and engineering ethics",
  ["sci engrg comput"] = "Scientific and Engineering Computation",
  ["sci fict"] = "Science and Fiction",
  ["sci for"] = "Scientia forestalis",
  ["sci forum"] = "Science forum",
  ["sci found china"] = "Science Foundation in China",
  ["sci fungorum"] = "Scientia fungorum",
  ["sci genet"] = "Scientia genetica; periodico di genetica per i paesi latini",
  ["sci geol mem"] = "Sciences géologiques: Mémoire",
  ["sci gov rep"] = "Science & government report",
  ["sci graeco-arabica"] = "Scientia Graeco-Arabica",
  ["sci hortic"] = "Scientia horticulturae",
  ["sci hortic (amsterdam)"] = "Scientia horticulturae",
  ["sci illus"] = "Science illustrated",
  ["sci immunol"] = "Science immunology",
  ["sci int (lahore)"] = "Science international (Lahore)",
  ["sci iran"] = "Scientia Iranica",
  ["sci iran d comput sci eng electr eng"] = "Scientia Iranica. Transactions D, Computer science & engineering, electrical engineering",
  ["sci issues jan dlugosz univ czest math"] = "Scientific Issues. Jan Dlugosz University in Czestochowa. Mathematics",
  ["sci j med res"] = "Scientific journal of medical research",
  ["sci justice"] = "Science & justice : journal of the Forensic Science Society",
  ["sci lett"] = "Science Letters",
  ["sci lett j"] = "Science letters journal",
  ["sci mar"] = "Scientia marina",
  ["sci mat"] = "Scienze Matematiche",
  ["sci math jpn"] = "Scientiae Mathematicae Japonicae",
  ["sci med (phila)"] = "Science & medicine",
  ["sci med (porto alegre)"] = "Scientia medica",
  ["sci med football"] = "Science and Medicine in Football",
  ["sci med ital"] = "Scientia medica italica. English ed",
  ["sci med man"] = "Science, medicine and man",
  ["sci med prat"] = "La Science médicale pratique",
  ["sci med technol east asia"] = "Science, medicine, and technology in East Asia",
  ["sci min j"] = "Scientific Mining Journal",
  ["sci model simul"] = "Scientific Modeling and Simulation",
  ["sci model simul smns"] = "Scientific Modeling and Simulation SMNS",
  ["sci mon"] = "The Scientific monthly",
  ["sci mus group j"] = "Science Museum Group journal",
  ["sci musique ser etudes"] = "Sciences de la Musique. Serie Etudes",
  ["sci nat"] = "Science of Nature",
  ["sci networks hist stud"] = "Science Networks. Historical Studies",
  ["sci new guinea"] = "Science in New Guinea",
  ["sci news"] = "Science news",
  ["sci news lett"] = "Science news letter",
  ["sci nurs"] = "SCI nursing : a publication of the American Association of Spinal Cord Injury Nurses",
  ["sci occident xxe siècle"] = "Les Sciences hors d’Occident au XXe Siècle",
  ["sci online lett atmos"] = "Scientific Online Letters on the Atmosphere",
  ["sci pages pulmonol"] = "Scientific pages of pulmonology",
  ["sci pap univ kansas nat hist mus"] = "Scientific papers (University of Kansas. Natural History Museum)",
  ["sci people"] = "Science for the people",
  ["sci pharm"] = "Scientia Pharmaceutica",
  ["sci philos twent century basic works log empir"] = "Science and Philosophy in the Twentieth Century",
  ["sci phone apps mobile devices"] = "Scientific Phone Apps and Mobile Devices",
  ["sci postprint"] = "Science postprint",
  ["sci pract perspect"] = "Science & practice perspectives",
  ["sci proc cardiff med soc"] = "Scientific proceedings of the Cardiff Medical Society",
  ["sci proc r dublin soc"] = "The scientific proceedings of the Royal Dublin Society. Royal Dublin Society",
  ["sci prog"] = "Science Progress",
  ["sci prog (new haven)"] = "Science in progress",
  ["sci prog twent century"] = "Science progress in the twentieth century",
  ["sci program"] = "Scientific programming",
  ["sci psychol"] = "Scientific Psychology",
  ["sci public policy"] = "Science & public policy",
  ["sci rech odontostomatol"] = "Science & recherche odontostomatologiques",
  ["sci remote sens"] = "Science of Remote Sensing",
  ["sci rep"] = "Scientific reports",
  ["sci rep fac ed fukushima univ"] = "Fukushima University",
  ["sci rep fac ed gifu univ natur sci"] = "Science Reports of the Faculty of Education",
  ["sci rep hirosaki univ"] = "Science Reports of the Hirosaki University",
  ["sci rep ist super sanita"] = "Scientific reports of the Istituto superiore di sanità",
  ["sci rep kagoshima univ"] = "Science Reports of Kagoshima University",
  ["sci rep kanazawa univ"] = "The Science Reports of Kanazawa University",
  ["sci rep res inst eng kanagawa univ"] = "Kanagawa University",
  ["sci rep res inst tohoku univ [med]"] = "Science Reports of the Research Institute, Tohoku University. Ser. C, Medicine",
  ["sci rep res inst tohoku univ med"] = "The science reports of the research institutes, Tohoku University. Ser. C, Medicine. Tōhoku Daigaku",
  ["sci rep tohoku univ ser 7"] = "The Science reports of the Tôhoku University. Seventh series: (Geography). Tōhoku Daigaku",
  ["sci rep yokohama nat univ sect i math phys chem"] = "Science Reports of the Yokohama National University",
  ["sci repub pop roum"] = "La Science dans la République populaire roumaine",
  ["sci rob"] = "Science Robotics",
  ["sci robot"] = "Science robotics",
  ["sci savoirs"] = "Sciences et Savoirs",
  ["sci scope (wash d c)"] = "Science scope (Washington, D.C.)",
  ["sci ser a math sci (ns)"] = "Scientia. Series A. Mathematical Sciences. New Series",
  ["sci signal"] = "Science signaling",
  ["sci signaling"] = "Science Signaling",
  ["sci sin"] = "Scientia Sinica",
  ["sci sin [b]"] = "Scientia Sinica. Series B, Chemical, Biological, Agricultural, Medical and Earth Sciences",
  ["sci sin b"] = "Scientia Sinica. Series B, Chemical, biological, agricultural, medical & earth sciences",
  ["sci sin chim"] = "Scientia Sinica Chimica",
  ["sci sintering"] = "Science of Sintering",
  ["sci soc"] = "Science and society",
  ["sci soc sante"] = "Sciences sociales et santé",
  ["sci society"] = "Science and Society",
  ["sci sol"] = "Science du Sol",
  ["sci sports"] = "Science & sports",
  ["sci stke"] = "Science's STKE : signal transduction knowledge environment",
  ["sci struct matiere"] = "Sciences des Structures et de la Matiere",
  ["sci stud"] = "Science studies (Helsinki, Finland)",
  ["sci stud (st bonaventure)"] = "Science studies",
  ["sci stud read"] = "Scientific studies of reading : the official journal of the Society for the Scientific Study of Reading",
  ["sci stud res ser math inform"] = "Scientific Studies and Research. Series Mathematics and Informatics",
  ["sci stud/wet"] = "Science Studies/Wetenschapsstudies",
  ["sci study lit"] = "Scientific study of literature",
  ["sci sustainability"] = "SCI Sustainability",
  ["sci synth, knowl updates"] = "Science of Synthesis, Knowledge Updates",
  ["sci talks"] = "Science Talks",
  ["sci teach"] = "Science teacher (Normal, Ill.)",
  ["sci tech perspect"] = "Sciences et techniques en perspective",
  ["sci tech perspect (2)"] = "Sciences et Techniques en Perspective. IIe Serie",
  ["sci technol adv mater"] = "Science and technology of advanced materials",
  ["sci technol adv mater: methods"] = "Science and Technology of Advanced Materials: Methods",
  ["sci technol archaeol res"] = "Science and technology of archaeological research",
  ["sci technol archaeolog res"] = "Science and Technology of Archaeological Research",
  ["sci technol arts res"] = "Science, technology and arts research",
  ["sci technol built environ"] = "Science and technology for the built environment",
  ["sci technol china"] = "Science & technology in China",
  ["sci technol educ libr"] = "Science & Technology Education Library",
  ["sci technol energetic mater"] = "Science and Technology of Energetic Materials",
  ["sci technol energy transition"] = "Science and Technology for Energy Transition",
  ["sci technol human values"] = "Science, technology & human values",
  ["sci technol libr (new york, ny)"] = "Science & technology libraries",
  ["sci technol manag ser"] = "Science, Technology, and Management Series",
  ["sci technol mater"] = "Science and Technology of Materials",
  ["sci technol med anc cult"] = "Science, Technology, and Medicine in Ancient Cultures",
  ["sci technol nucl install"] = "Science and Technology of Nuclear Installations",
  ["sci technol rev"] = "Science & technology review (Livermore, Calif.)",
  ["sci technol weld joining"] = "Science and Technology of Welding and Joining",
  ["sci temper j"] = "Scientific temperance journal",
  ["sci times j diabetes"] = "Scientific times journal of diabetes",
  ["sci total environ"] = "The Science of the total environment",
  ["sci trans environ technovation"] = "Scientific transactions in environment and technovation",
  ["sci transl med"] = "Science translational medicine",
  ["sci vis"] = "Science vision",
  ["sci wales"] = "Scientists of Wales",
  ["sci world j"] = "Scientific World Journal",
  ["sci writ anc mediev world"] = "Scientific Writings from the Ancient and Medieval World",
  ["sci, technol hum values"] = "Science, Technology & Human Values",
  ["sci, technol soc"] = "Science, Technology and Society",
  ["science (1979)"] = "Science",
  ["science (80- )"] = "Science",
  ["science (washington, dc, u s)"] = "Science (Washington, DC, United States)",
  ["scienceopen res"] = "ScienceOpen Research",
  ["sciences (new york)"] = "The Sciences",
  ["scientia (bristol)"] = "Scientia (Bristol, England)",
  ["scientia ser phys-math"] = "Scientia. Serie Physico-Mathematique",
  ["scientifica (cairo)"] = "Scientifica",
  ["scifed j cardiol"] = "SciFed journal of cardiology",
  ["scifed virol res j"] = "SciFed virology research journal",
  ["scimed j"] = "SciMedicine journal",
  ["scipost phys"] = "Scipost Physics",
  ["scipost phys proc"] = "Scipost Physics Proceedings",
  ["scm not"] = "SCM Noticies",
  ["scn news"] = "SCN news",
  ["scna newsl"] = "Scna Newsletter",
  ["scoliosis spinal disord"] = "Scoliosis and spinal disorders",
  ["scope (kalamazoo)"] = "Scope",
  ["scot j polit economy"] = "Scottish Journal of Political Economy",
  ["scott aff"] = "Scottish affairs",
  ["scott archaeol j"] = "Scottish Archaeological Journal",
  ["scott econ soc hist"] = "Scottish economic & social history",
  ["scott edu rev"] = "Scottish educational review",
  ["scott for"] = "Scottish Forestry",
  ["scott geogr j"] = "Scottish Geographical Journal",
  ["scott geogr mag"] = "Scottish geographical magazine",
  ["scott grad ser"] = "Scottish Graduate Series",
  ["scott hist rev"] = "The Scottish historical review",
  ["scott j crim justice stud"] = "The Scottish journal of criminal justice studies : the journal of the Scottish Association for the Study of Delinquency",
  ["scott j geol"] = "Scottish Journal of Geology",
  ["scott j polit econ"] = "Scottish journal of political economy",
  ["scott j sociol"] = "The Scottish journal of sociology",
  ["scott med j"] = "Scottish medical journal",
  ["scr ethnol"] = "Scripta ethnologica (Centro Argentino de Etnología Americana)",
  ["scr geobot"] = "Scripta Geobotanica",
  ["scr geol"] = "Scripta geologica",
  ["scr instituti donneriani abo"] = "Scripta Instituti Donneriani Aboensis",
  ["scr mater"] = "Scripta Materialia",
  ["scr med (brno)"] = "Scripta medica",
  ["scr metall"] = "Scripta Metallurgica",
  ["scr metall mater"] = "Scripta Metallugica et Materialia",
  ["scr minora"] = "Scripta minora Regiae Societatis humaniorum litterarum Lundensis = Studier",
  ["screen (london)"] = "Screen",
  ["scripta comput sci appl math"] = "Scripta",
  ["sd j med pharm"] = "The South Dakota journal of medicine and pharmacy",
  ["sda dent"] = "Sda Dentist",
  ["sdc mag"] = "SDC magazine",
  ["sdrp j biomed eng"] = "SDRP journal of biomedical engineering",
  ["sdrp j plant sci"] = "SDRP journal of plant science",
  ["se pu"] = "Se pu = Chinese journal of chromatography",
  ["sea view hosp bull"] = "Sea View Hospital bulletin",
  ["seagull theatre q"] = "Seagull theatre quarterly",
  ["sealing technol"] = "Sealing Technology",
  ["seaprap res rep"] = "SEAPRAP research report : a report of research undertaken with the assistance of an award from the Southeast Asia Population Research Awards Program (SEAPRAP), Institute of Southeast Asian Studies",
  ["seara med"] = "Seara médica",
  ["search (syd)"] = "Search",
  ["search medford n j"] = "Searcher (Medford, N.J.)",
  ["seattle univ law rev"] = "Seattle University law review",
  ["seb exp biol ser"] = "SEB experimental biology series",
  ["secolas ann"] = "SECOLAS annals : journal of the Southeastern Council on Latin American Studies. Southeastern Council on Latin American Studies",
  ["second lang res"] = "Second language research",
  ["second messengers phosphoproteins"] = "Second messengers and phosphoproteins",
  ["second opin"] = "Second opinion (Park Ridge, Ill.)",
  ["second opin (chic)"] = "Second opinion (Chicago, Ill. : 1999)",
  ["second opin health care issues"] = "Second opinions on health care issues",
  ["secur commun netw"] = "Security and Communication Networks",
  ["secur dialogue"] = "Security dialogue",
  ["secur inform"] = "Security informatics",
  ["secur saf"] = "Security and Safety",
  ["securitas (roma)"] = "Securitas",
  ["sediment geol"] = "Sedimentary Geology",
  ["seed sci res"] = "Seed Science Research",
  ["seeing perceiving"] = "Seeing and perceiving",
  ["sefer asya"] = "Sefer asya",
  ["seg discovery"] = "SEG Discovery",
  ["seibutsu kankyo chosetsu"] = "Seibutsu kankyo chosetsu. [Environment control in biology",
  ["seibutsu kogaku kaishi"] = "Seibutsu Kogaku Kaishi",
  ["seibutsu seisangaku kenkyu"] = "Seibutsu seisangaku kenkyū : Hiroshima Daigaku Seibutsu Seisan Gakubu kiyō",
  ["seibutsu shiryo bunseki"] = "Seibutsu shiryō bunseki = Journal of analytical bio-science",
  ["seikatsu eisei"] = "Seikatsu eisei",
  ["seikei geka"] = "Seikeigeka. Orthopedic surgery",
  ["seimei kagaku kenkyujo kiyo"] = "Seimei Kagaku Kenkyujo kiyō",
  ["seiri seital"] = "Seiri seitai. [Physiology and ecology]",
  ["seiroka kango daigaku kiyo"] = "Sei Roka Kango Daigaku kiyō",
  ["seishin igaku"] = "Seishin igaku. Clinical psychiatry",
  ["seishin igaku kenkyusho gyosekishu"] = "Seishin Igaku Kenkyūjo gyōsekishū = Bulletin of the Seishin-Igaku Institute",
  ["seishin shinkeigaku zasshi"] = "Seishin shinkeigaku zasshi = Psychiatria et neurologia Japonica",
  ["seismic instrum"] = "Seismic Instruments",
  ["seismol res lett"] = "Seismological Research Letters",
  ["seitai kagaku"] = "[Seitai no kagaku] [Science of the living body]",
  ["seitai no kagaku"] = "Seitai no Kagaku",
  ["sej surg pain"] = "SEJ surgery and pain",
  ["sel annu rev anal sci"] = "Selected Annual Reviews of the Analytical Sciences",
  ["sel cancer ther"] = "Selective cancer therapeutics",
  ["sel chapters number theory: spec numbers"] = "Selected Chapters of Number Theory: Special Numbers",
  ["sel lectures math"] = "Selected Lectures in Mathematics",
  ["sel mat"] = "Selecciones Matematicas",
  ["sel math new ser"] = "Selecta Mathematica New Series",
  ["sel odontol (sao paulo)"] = "Seleções odontológicas",
  ["sel pap consort revolut eur 1759 1850"] = "Selected papers. Consortium on Revolutionary Europe, 1750-1850",
  ["sel works b b mandelbrot"] = "Selected Works of Benoit B",
  ["sel works probab stat"] = "Selected Works in Probability and Statistics",
  ["selecta math (ns)"] = "Selecta Mathematica. New Series",
  ["self care depend care nurs"] = "Self-care, dependent-care & nursing : the official journal of the International Orem Society",
  ["self identity"] = "Self and identity : the journal of the International Society for Self and Identity",
  ["self nonself"] = "Self/nonself",
  ["self-teach guides math"] = "Self-Teaching Guides on Mathematics",
  ["selskokhoziaĭstvennaia biol"] = "Selʹskokhozi︠a︡ĭstvennai︠a︡ biologii︠a︡",
  ["sem epidemiol"] = "Semana Epidemiologica",
  ["sem hop"] = "La semaine des hôpitaux : organe fondé par l'Association d'enseignement médical des hôpitaux de Paris",
  ["sem hop inf"] = "Semaine des hôpitaux: informations",
  ["sem hop paris"] = "La Semaine des hôpitaux de Paris",
  ["sem hop ther"] = "Semaine des hôpitaux. Therapeutique",
  ["sem hop ther paris"] = "La Semaine des hôpitaux: thérapeutique",
  ["sem lothar combin"] = "Seminaire Lotharingien de Combinatoire",
  ["sem math (n s)"] = "Seminaire de Mathematique (Nouvelle Serie)",
  ["sem math sci"] = "Seminar on Mathematical Sciences",
  ["sem med"] = "La Semana médica",
  ["sem med esp"] = "Semana médica española",
  ["sem med mex"] = "Semana médica de México",
  ["sem med prof med soc"] = "Semaine médicale professionelle et médico-sociale",
  ["sem medicale med soc"] = "La Semaine médicale [médecine sociale. Edition: couverture jaune]",
  ["sem ser math algebra"] = "Seminar Series in Mathematics",
  ["sem ser math anal"] = "Seminar Series in Mathematics",
  ["sem ther"] = "Semaine thérapeutique",
  ["sema j"] = "SeMA Journal",
  ["sema simai springer ser"] = "SEMA SIMAI Springer Series",
  ["semant pragmat"] = "Semantics and pragmatics",
  ["semant web"] = "Semantic web",
  ["semicond phys quantum electron optoelectron"] = "Semiconductor Physics, Quantum Electronics and Optoelectronics",
  ["semicond sci technol"] = "Semiconductor Science and Technology",
  ["semicond today"] = "Semiconductor Today",
  ["semigroup forum"] = "Semigroup Forum",
  ["semin adolesc med"] = "Seminars in adolescent medicine",
  ["semin anesth"] = "Seminars in anesthesia",
  ["semin arthritis rheum"] = "Seminars in arthritis and rheumatism",
  ["semin arthroplasty"] = "Seminars in arthroplasty",
  ["semin breast dis"] = "Seminars in breast disease",
  ["semin cancer biol"] = "Seminars in cancer biology",
  ["semin cardiothorac vasc anesth"] = "Seminars in cardiothoracic and vascular anesthesia",
  ["semin cell biol"] = "Seminars in Cell Biology",
  ["semin cell dev biol"] = "Seminars in cell & developmental biology",
  ["semin cienc agrar"] = "Semina. Ciências agrárias",
  ["semin clin neuropsychiatry"] = "Seminars in clinical neuropsychiatry",
  ["semin colon rectal surg"] = "Seminars in colon & rectal surgery",
  ["semin congr"] = "Seminaires et Congres",
  ["semin cutan med surg"] = "Seminars in cutaneous medicine and surgery",
  ["semin dent hyg"] = "Seminars in dental hygiene",
  ["semin dermatol"] = "Seminars in dermatology",
  ["semin dev biol"] = "Seminars in developmental biology",
  ["semin diagn pathol"] = "Seminars in diagnostic pathology",
  ["semin dial"] = "Seminars in dialysis",
  ["semin drug treat"] = "Seminars in drug treatment",
  ["semin ensen odontopediatr"] = "Seminario De Ensenanza De La Odontopediatria",
  ["semin equ deriv partielles"] = "Seminaire: Equations aux Derivees Partielles",
  ["semin fam med"] = "Seminars in family medicine",
  ["semin fetal neonatal med"] = "Seminars in fetal & neonatal medicine",
  ["semin gastrointest dis"] = "Seminars in gastrointestinal disease",
  ["semin hear"] = "Seminars in hearing",
  ["semin hematol"] = "Seminars in hematology",
  ["semin i vekua inst appl math rep"] = "Seminar of Ilia Vekua Institute of Applied Mathematics. Reports",
  ["semin immunol"] = "Seminars in immunology",
  ["semin immunopathol"] = "Seminars in immunopathology",
  ["semin int"] = "Seminar international",
  ["semin interv cardiol"] = "Seminars in interventional cardiology : SIIC",
  ["semin intervent radiol"] = "Seminars in interventional radiology",
  ["semin laparosc surg"] = "Seminars in laparoscopic surgery",
  ["semin liver dis"] = "Seminars in liver disease",
  ["semin med pract"] = "Seminars in medical practice",
  ["semin musculoskelet radiol"] = "Seminars in musculoskeletal radiology",
  ["semin neonatol"] = "Seminars in neonatology : SN",
  ["semin nephrol"] = "Seminars in nephrology",
  ["semin neurol"] = "Seminars in neurology",
  ["semin nucl med"] = "Seminars in nuclear medicine",
  ["semin nurse manag"] = "Seminars for nurse managers",
  ["semin oncol"] = "Seminars in oncology",
  ["semin oncol nurs"] = "Seminars in oncology nursing",
  ["semin ophthalmol"] = "Seminars in ophthalmology",
  ["semin orthod"] = "Seminars in orthodontics",
  ["semin pediatr infect dis"] = "Seminars in pediatric infectious diseases",
  ["semin pediatr neurol"] = "Seminars in pediatric neurology",
  ["semin pediatr surg"] = "Seminars in pediatric surgery",
  ["semin perinatol"] = "Seminars in perinatology",
  ["semin perioper nurs"] = "Seminars in perioperative nursing",
  ["semin plast surg"] = "Seminars in plastic surgery",
  ["semin probab"] = "Seminaire de Probabilites",
  ["semin psychiatry"] = "Seminars in psychiatry",
  ["semin radiat oncol"] = "Seminars in radiation oncology",
  ["semin rep merck sharp dohme"] = "Seminar report. Merck Sharp & Dohme",
  ["semin reprod endocrinol"] = "Seminars in reproductive endocrinology",
  ["semin reprod med"] = "Seminars in reproductive medicine",
  ["semin respir crit care med"] = "Seminars in respiratory and critical care medicine",
  ["semin respir infect"] = "Seminars in respiratory infections",
  ["semin roentgenol"] = "Seminars in roentgenology",
  ["semin speech lang"] = "Seminars in speech and language",
  ["semin spine surg"] = "Seminars in spine surgery",
  ["semin surg oncol"] = "Seminars in surgical oncology",
  ["semin thorac cardiovasc surg"] = "Seminars in thoracic and cardiovascular surgery",
  ["semin thorac cardiovasc surg pediatr card surg annu"] = "Seminars in thoracic and cardiovascular surgery. Pediatric cardiac surgery annual",
  ["semin thromb hemost"] = "Seminars in thrombosis and hemostasis",
  ["semin ultrasound ct mr"] = "Seminars in ultrasound, CT, and MR",
  ["semin ultrasound ct mri"] = "Seminars in Ultrasound, CT and MRI",
  ["semin urol"] = "Seminars in urology",
  ["semin urol oncol"] = "Seminars in urologic oncology",
  ["semin vasc med"] = "Seminars in vascular medicine",
  ["semin vasc surg"] = "Seminars in vascular surgery",
  ["semin vet med surg (small anim)"] = "Seminars in Veterinary Medicine and Surgery (Small Animal)",
  ["semin vet med surg small anim"] = "Seminars in veterinary medicine and surgery (small animal)",
  ["semin virol"] = "Seminars in virology",
  ["semina cienc agrar"] = "Semina: Ciencias Agrarias",
  ["semstat elem"] = "SemStat Elements",
  ["senckenb biol"] = "Senckenbergiana biologica",
  ["senckenbergiana biol"] = "Senckenbergiana Biologie",
  ["senckenbergiana lethaea"] = "Senckenbergiana Lethaea",
  ["seniors hous care j"] = "Seniors housing & care journal",
  ["sens actuators (warrendale pa)"] = "Sensors and actuators (Warrendale, Pa.)",
  ["sens actuators a phys"] = "Sensors and actuators. A, Physical",
  ["sens actuators b chem"] = "Sensors and actuators. B, Chemical",
  ["sens actuators phys"] = "Sensors and Actuators A: Physical",
  ["sens actuators rep"] = "Sensors and Actuators Reports",
  ["sens actuators, a"] = "Sensors and Actuators A: Physical",
  ["sens actuators, b"] = "Sensors and Actuators B: Chemical",
  ["sens bio-sens res"] = "Sensing and Bio-Sensing Research",
  ["sens biosensing res"] = "Sensing and Bio-Sensing Research",
  ["sens imaging"] = "Sensing and imaging",
  ["sens instrum food qual saf"] = "Sensing and Instrumentation for Food Quality and Safety",
  ["sens int"] = "Sensors International",
  ["sens lett"] = "Sensor letters",
  ["sens mater"] = "Sensors and Materials",
  ["sens neur"] = "Sensory neuron",
  ["sens processes"] = "Sensory processes",
  ["sens rev"] = "Sensor Review",
  ["sens switz"] = "Sensors (Switzerland)",
  ["sens syst"] = "Sensory systems",
  ["sensecam 2013"] = "SenseCam 2013 : proceedings of the 4th SenseCam Conference : SenseCam and Pervasive Imaging 2013 : San Diego, USA, November 18-19, 2013. SenseCam (Conference) (4th : 2013 : San Diego, Calif.)",
  ["sensors (basel)"] = "Sensors (Basel, Switzerland)",
  ["sensors update"] = "Sensors Update",
  ["sentinel event alert"] = "Sentinel event alert",
  ["sen’i gakkaishi"] = "Sen’i Gakkaishi",
  ["seohaean jiyeog baljeon yeongu"] = "Seohae-an ji-yeog baljeon yeon-gu = The west coast area development institute",
  ["seoul j econ"] = "Seoul Journal of Economics",
  ["seoul j med"] = "The Seoul journal of medicine",
  ["sep purif methods"] = "Separation and Purification Methods",
  ["sep purif rev"] = "Separation and Purification Reviews",
  ["sep purif technol"] = "Separation and Purification Technology",
  ["sep sci plus"] = "Separation science plus",
  ["sep sci technol"] = "Separation Science and Technology",
  ["sep sci technol (philadelphia, pa, u s)"] = "Separation Science and Technology (Philadelphia, PA, United States)",
  ["sep technol"] = "Separations Technology",
  ["sepsis (boston)"] = "Sepsis",
  ["seq anal"] = "Sequential analysis",
  ["sequential anal"] = "Sequential Analysis",
  ["ser adv math appl sci"] = "Series on Advances in Mathematics for Applied Sciences",
  ["ser adv statist mech"] = "Series on Advances in Statistical Mechanics",
  ["ser algebr differ geom"] = "Series in Algebraic and Differential Geometry",
  ["ser algebra"] = "Series in Algebra",
  ["ser anal"] = "Series in Analysis",
  ["ser anal appl comput"] = "Series on Analysis, Applications and Computation",
  ["ser appl comput math"] = "Series in Applied and Computational Mathematics",
  ["ser appl math"] = "Series on Applied Mathematics",
  ["ser approx decompos"] = "Series in Approximations and Decompositions",
  ["ser bioeng"] = "Series in BioEngineering",
  ["ser biomech"] = "Series on biomechanics",
  ["ser cient"] = "Serie Cientifica",
  ["ser coding theory cryptol"] = "Series on Coding Theory and Cryptology",
  ["ser complex nonlinearity chaos"] = "Series on Complexity, Nonlinearity and Chaos",
  ["ser complex sci"] = "Series on Complexity Science",
  ["ser comput oper res"] = "Series on Computers and Operations Research",
  ["ser comput phys process mech thermal sci"] = "Series in Computational and Physical Processes in Mechanics and Thermal Sciences",
  ["ser comput vis"] = "Series in Computer Vision",
  ["ser concr appl math"] = "Series on Concrete and Applicable Mathematics",
  ["ser contemp appl math cam"] = "Series in Contemporary Applied Mathematics CAM",
  ["ser contemp math"] = "Series in Contemporary Mathematics",
  ["ser disp sci tech"] = "Series in Display Science and Technology",
  ["ser electr comput eng"] = "Series in Electrical and Computer Engineering",
  ["ser filol ros"] = "Seria filologia rosyjska",
  ["ser fiz (uniw im adama mickiewicza poznaniu)"] = "Seria Fizyka (Uniwersytet im. Adama Mickiewicza w Poznaniu)",
  ["ser found nat sci tech"] = "Series on the Foundations of Natural Science and Technology",
  ["ser haematol"] = "Series haematologica (1968)",
  ["ser intell control intell autom"] = "Series in Intelligent Control and Intelligent Automation",
  ["ser internat bus econom"] = "Series in International Business and Economics",
  ["ser knots everything"] = "Series on Knots and Everything",
  ["ser konf (wroc)"] = "Seria: Konferencje",
  ["ser libr"] = "The Serials librarian",
  ["ser mach percept artif intell"] = "Series in Machine Perception and Artificial Intelligence",
  ["ser mat fis"] = "Serie di Matematica e Fisica",
  ["ser math anal appl"] = "Series in Mathematical Analysis and Applications",
  ["ser math biol med"] = "Series in Mathematical Biology and Medicine",
  ["ser math educ"] = "Series on Mathematics Education",
  ["ser modern condensed matter phys"] = "Series in Modern Condensed Matter Physics",
  ["ser multivariate anal"] = "Series on Multivariate Analysis",
  ["ser new front adv math ist ric base"] = "Series on New Frontiers in Advanced Mathematics of the Istituto per la Ricerca di Base",
  ["ser number theory"] = "Series in Number Theory",
  ["ser number theory appl"] = "Series on Number Theory and its Applications",
  ["ser nurs adm"] = "Series on nursing administration",
  ["ser opt optoelectron"] = "Series in Optics and Optoelectronics",
  ["ser optim"] = "Series on Optimization",
  ["ser optim appl"] = "Series on Optimization and its Applications",
  ["ser paedopsychiatr"] = "Series paedopsychiatrica",
  ["ser pure appl math"] = "Series on the Pure and Applied Mathematics",
  ["ser pure math"] = "Series in Pure Mathematics",
  ["ser quant financ"] = "Series in Quantitative Finance",
  ["ser real anal"] = "Series in Real Analysis",
  ["ser rev med soc"] = "Ser (revista médico-social)",
  ["ser sci"] = "Serie Scientifica",
  ["ser soft condens matter"] = "Series in Soft Condensed Matter",
  ["ser soviet east european math"] = "Series on Soviet and East European Mathematics",
  ["ser stab vib control struct"] = "Series on Stability, Vibration and Control of Structures",
  ["ser stab vib control syst ser a"] = "Series on Stability, Vibration and Control of Systems. Series A: Textbooks, Monographs and Treatises",
  ["ser stab vib control syst ser b"] = "Series on Stability, Vibration and Control of Systems",
  ["ser students texts contemp biology"] = "A Series of Student Texts in Contemporary Biology",
  ["ser syst autom"] = "Serie Systemes Automatises",
  ["ser t"] = "La Serie T",
  ["ser theor comput sci"] = "Series on Theoretical Computer Science",
  ["ser univ math"] = "Series on University Mathematics",
  ["ser work organ"] = "Series on Work and Organization",
  ["serb astron j"] = "Serbian Astronomical Journal",
  ["serdica j comput"] = "Serdica Journal of Computing",
  ["serdica math j"] = "Serdica. Mathematical Journal. Serdika. Matematichesko Spisanie",
  ["serie: ecole nationale de la statistique et de l’administration economique et du centre d’etudes des programmes"] = "Serie: Ecole Nationale de la Statistique et de l’Administration economique et du Centre d’Etudes des Programmes Economiques",
  ["series (berl)"] = "SERIEs : journal of the Spanish Economic Association",
  ["serono symp publ raven press"] = "Serono Symposia publications from Raven Press",
  ["serv migr"] = "Servizio migranti",
  ["serv oriented comput appl"] = "Service Oriented Computing and Applications",
  ["serv rech terres for, qué"] = "Service de la recherche terres et forêts, Québec",
  ["serv sci"] = "Service science",
  ["serv soc que"] = "Service social",
  ["sess laws guam"] = "Session laws. Guam",
  ["sess laws hawaii passed state legis hawaii"] = "Session laws of Hawaii passed by the State Legislature. Hawaii",
  ["sess laws state minn minn"] = "Session laws of the State of Minnesota. Minnesota",
  ["set valued var anal"] = "Set-valued and variational analysis",
  ["set-valued anal"] = "Set-Valued Analysis",
  ["set-valued var anal"] = "Set-Valued and Variational Analysis",
  ["seton hall law rev"] = "Seton Hall law review",
  ["seton hall legis j"] = "Seton Hall legislative journal",
  ["settim med"] = "Settimana medica",
  ["seven days"] = "Seven days",
  ["seventeenth century"] = "The Seventeenth century",
  ["seventeenth century fr stud"] = "Seventeenth-century French studies",
  ["seventh conf health survey res methods (1999)"] = "Seventh Conference on Health Survey Research Methods. Conference on Health Survey Research Methods (7th : 1999 : Williamsburg, Va.)",
  ["sewage ind waste"] = "Sewage and industrial wastes",
  ["sewage work j"] = "Sewage works journal",
  ["sex abuse"] = "Sexual abuse : a journal of research and treatment",
  ["sex addict compulsivity"] = "Sexual addiction & compulsivity",
  ["sex cult"] = "Sexuality & culture",
  ["sex dev"] = "Sexual development : genetics, molecular biology, evolution, endocrinology, embryology, and pathology of sex determination and differentiation",
  ["sex disabil"] = "Sexuality and disability",
  ["sex educ"] = "Sex education",
  ["sex gend policy"] = "Sexuality, gender & policy",
  ["sex health"] = "Sexual health",
  ["sex health exch"] = "Sexual health exchange",
  ["sex marital ther"] = "Sexual and marital therapy : journal of the Association of Sexual and Marital Therapists",
  ["sex med"] = "Sexual medicine",
  ["sex med rev"] = "Sexual medicine reviews",
  ["sex med today"] = "Sexual medicine today",
  ["sex padag fam plan"] = "Sexualpadagogik und familienplanung",
  ["sex planeam fam"] = "Sexualidade e planeamento familiar",
  ["sex plant reprod"] = "Sexual plant reproduction",
  ["sex relation ther"] = "Sexual and relationship therapy : journal of the British Association for Sexual and Relationship Therapy",
  ["sex reprod health matters"] = "Sexual and reproductive health matters",
  ["sex reprod healthc"] = "Sexual & reproductive healthcare : official journal of the Swedish Association of Midwives",
  ["sex reprod menopause"] = "Sexuality, reproduction & menopause",
  ["sex res social policy"] = "Sexuality research & social policy : journal of NSRC : SR & SP",
  ["sex roles"] = "Sex roles",
  ["sex salud soc (rio j)"] = "Sexualidad, salud y sociedad : revista Latinoamericana",
  ["sex transm dis"] = "Sexually transmitted diseases",
  ["sex transm infect"] = "Sexually transmitted infections",
  ["sex wkly plus"] = "Sex weekly plus",
  ["sf j biotechnol biomed eng"] = "SF journal of biotechnology and biomedical engineering",
  ["sf j med oncol cancer"] = "SF journal of medical oncology and cancer",
  ["shaanxi keji daxue xuebao"] = "Shaanxi Keji Daxue xuebao",
  ["shakai keizai shigaku"] = "Shakai keizai shigaku",
  ["shakespeare stud"] = "Shakespeare studies",
  ["shan di nong ye sheng wu xue bao"] = "Shan di nong ye sheng wu xue bao = Journal of mountain agriculture and biology",
  ["shandong da xue xue bao"] = "Shandong Da Xue xue bao. Yi xue ban",
  ["shandong da xue xue bao li xue ban"] = "Shandong da xue xue bao. Li xue ban /Natural science",
  ["shandong daxue xuebao ziran kexue ban"] = "Shandong Daxue Xuebao",
  ["shandong kuangye xueyuan xuebao"] = "Shandong Kuangye Xueyuan Xuebao",
  ["shandong li gong da xue xue bao she hui ke xue ban"] = "Shandong li gong da xue xue bao. She hui ke xue ban",
  ["shandong nongye kexue"] = "Shandong nongye kexue",
  ["shandong yi kan"] = "[Shandong yi kan] [Shantung medical publication]",
  ["shanghai arch psychiatry"] = "Shanghai archives of psychiatry",
  ["shanghai chest"] = "Shanghai chest",
  ["shanghai hai yang da xue xue bao"] = "Shanghai hai yang da xue xue bao",
  ["shanghai huanjing kexue"] = "Shanghai Huanjing Kexue",
  ["shanghai jiao tong da xue xue bao yi xue ban"] = "Shanghai jiao tong da xue xue bao. Yi xue ban",
  ["shanghai jiaotong daxue xuebao nongye kexueban"] = "Shanghai jiaotong daxue xuebao. Nongye kexueban",
  ["shanghai journal of medical laboratory sciences|shanghai yixue jianyan zazhi"] = "Shanghai yixue jianyan zazhi",
  ["shanghai kou qiang yi xue"] = "Shanghai kou qiang yi xue = Shanghai journal of stomatology",
  ["shanghai mien i hsueh tsa chih"] = "Shanghai mian yi xue za zhi",
  ["shanghai nong xue yuan xue bao"] = "Shanghai nong xue yuan xue bao = Journal of Shanghai Agricultural College = Shanghai nongxueyuan xuebao",
  ["shanghai nong ye xue bao"] = "Shanghai nong ye xue bao = Acta agriculturae Shanghai",
  ["shanghai shifan daxue xuebao ziran kexue ban"] = "Shanghai Shifan Daxue Xuebao",
  ["shanghai yixue jianyan zazhi"] = "Shanghai yixue jianyan zazhi = Shanghai journal of medical laboratory sciences",
  ["shanxi da xue xue bao zi ran ke xue ban"] = "Shanxi da xue xue bao. Zi ran ke xue ban = Shanxi University journal",
  ["shanxi nong ye da xue xue bao zi ran ke xue ban"] = "Shanxi nong ye da xue xue bao. Zi ran ke xue ban",
  ["shanxi shi da xue bao zi ran ke xue ban"] = "Shanxi shi da xue bao. Zi ran ke xue ban = Journal of Shaanxi Normal University",
  ["shape med imaging (2018)"] = "Shape in medical imaging : International Workshop, ShapeMI 2018, held in conjunction with MICCAI 2018, Granada, Spain, September 20, 2018 : proceedings. ShapeMI (Workshop) (2018 : Granada, Spain)",
  ["shape med imaging (2020)"] = "Shape in Medical Imaging : International Workshop, ShapeMI 2020, Held in Conjunction with MICCAI 2020, Lima, Peru, October 4, 2020, Proceedings",
  ["shape mem superelasticity"] = "Shape Memory and Superelasticity",
  ["shb12 (2012)"] = "SHB'12 : proceedings of the 2012 ACM International Workshop on Smart Health and Wellbeing : October 29, 2012, Maui, Hawaii, USA. International Workshop on Smart Health and Wellbeing (2012 : Maui, Hawaii)",
  ["she ji: j des econ innovation"] = "She Ji: The Journal of Design, Economics, and Innovation",
  ["she zhi"] = "She zhi = Journal of snake",
  ["shendet pop"] = "Shëndetësia popullore",
  ["sheng chung ping tsa chih"] = "Hsueh Yu Ch",
  ["sheng li hsueh pao"] = "Sheng Li Hsueh Pao [Acta Physiologica Sinica]",
  ["sheng li ke xue jin zhan"] = "Sheng li ke xue jin zhan [Progress in physiology]",
  ["sheng li ko hsueh chin chan"] = "Sheng Li Ko Hsueh Chin Chan [Progress In Physiological Sciences]",
  ["sheng li xue bao"] = "Sheng li xue bao : [Acta physiologica Sinica]",
  ["sheng ming ke xue yan jiu"] = "Sheng ming ke xue yan jiu",
  ["sheng tai huan jing xue bao"] = "Sheng tai huan jing xue bao = Ecology and environmental sciences",
  ["sheng tai xue bao"] = "Sheng tai xue bao = Acta ecologica Sinica",
  ["sheng wu duo yang xing"] = "Sheng wu duo yang xing = Chinese biodiversity",
  ["sheng wu gong cheng xue bao"] = "Sheng wu gong cheng xue bao = Chinese journal of biotechnology",
  ["sheng wu hua hsueh tsa chih"] = "Sheng wu hua xue za zhi = Chinese biochemical journal",
  ["sheng wu hua xue yu sheng wu wu li jin zhan"] = "Sheng wu hua xue yu sheng wu wu li jin zhan",
  ["sheng wu hua xue yu sheng wu wu li xue bao (shanghai)"] = "Sheng wu hua xue yu sheng wu wu li xue bao Acta biochimica et biophysica Sinica",
  ["sheng wu i hsueh kung cheng hsueh tsa chih"] = "Sheng Wu I Hsueh Kung Cheng Hsueh Tsa Chih Journal of Biomedical Engineering (Cheng-Tu",
  ["sheng wu ji shu tong bao"] = "Sheng wu ji shu tong bao = Biotechnology bulletin",
  ["sheng wu jia gong guo cheng"] = "Sheng wu jia gong guo cheng",
  ["sheng wu wu li hsueh bao"] = "Sheng Wu Wu Li Hsueh Bao",
  ["sheng wu xue bao|bioformosa"] = "Bioformosa",
  ["sheng wu xue za zhi"] = "Sheng wu xue za zhi = Journal of biology",
  ["sheng wu yi xue gong cheng xue za zhi"] = "Sheng wu yi xue gong cheng xue za zhi = Journal of biomedical engineering = Shengwu yixue gongchengxue zazhi",
  ["shengtai huanjing"] = "Shengtai huanjing",
  ["shengwu huaxue yu shengwu wuli jinzhan"] = "Shengwu Huaxue Yu Shengwu Wuli Jinzhan",
  ["shengwu huaxue yu shengwu wuli xuebao"] = "Shengwu Huaxue Yu Shengwu Wuli Xuebao",
  ["shengwu huaxue zazhi"] = "Shengwu Huaxue Zazhi",
  ["shengwu jishu"] = "Shengwu jishu = Biotechnology",
  ["shengwu jishu tongxun"] = "Shengwu jishu tongxun",
  ["shengwu xinxixue"] = "Shengwu xinxixue",
  ["shengwu ziyuan"] = "Shengwu ziyuan",
  ["shengzhi yu biyun"] = "Sheng zhi yu bi yun = Reproduction and contraception",
  ["shenyang nongye daxue xuebao"] = "Shenyang nongye daxue xuebao",
  ["shenyang yao ke da xue xue bao"] = "Shenyang yao ke da xue xue bao = Journal of Shenyang Pharmaceutical University",
  ["shenyang yaoxueyuan xuebao"] = "Shenyang Yaoxueyuan Xuebao",
  ["shi da sheng wu xue bao"] = "Shi da sheng wu xue bao",
  ["shi da xue bao"] = "Shi da xue bao",
  ["shi jie hua ren xiao hua za zhi"] = "Shi jie hua ren xiao hua za zhi = World Chinese journal of digestology",
  ["shi jie li shi"] = "Shi jie li shi = Shi jie li shi",
  ["shi pin yu fa xiao gong ye"] = "Shi pin yu fa xiao gong ye = Shipin yu faxiao gongye",
  ["shi xue yue kan"] = "Shi xue yue kan",
  ["shi yan li xue"] = "Shi yan li xue = Journal of experimental mechanics",
  ["shi yan sheng wu xue bao"] = "Shi yan sheng wu xue bao",
  ["shi yan yu jian yan yi xue"] = "Shi yan yu jian yan yi xue",
  ["shi yong jun"] = "Shi yong jun : shuang yue kan = Edible fungi",
  ["shi yong jun xue bao"] = "Shi yong jun xue bao (Online)",
  ["shigaku zasshi"] = "Shigaku zasshi",
  ["shih yen sheng wu hsueh pao"] = "Shih Yen Sheng Wu Hsueh Pao Journal of Experimental Biology",
  ["shihezi daxue xuebao"] = "Shihezi daxue xuebao. Ziran kexue ban = Journal of Shihezi University. Natural science",
  ["shika igaku"] = "[Shika igaku] [Dental medicine]",
  ["shika kiso igakkai zasshi"] = "Shika Kiso Igakkai zasshi = Japanese journal of oral biology",
  ["shika rikogaku zasshi"] = "Shika rikōgaku zasshi. Journal of the Japan Society for Dental Apparatus and Materials",
  ["shika zairyo kikai"] = "Shika zairyō, kikai = Journal of the Japanese Society for Dental Materials and Devices",
  ["shikai tenbo"] = "Shikai tenbo = Dental outlook",
  ["shikwa gakuho"] = "Shika gakuho. Dental science reports",
  ["shilap soc hispano luso am lepid"] = "SHILAP : revista de lepidopterología",
  ["shimane igaku"] = "[Shimane igaku] [Shimane medicine]",
  ["shimane j med sci"] = "Shimane journal of medical science",
  ["shinkei ganka"] = "Shinkei ganka = Neuro-ophthalmology Japan",
  ["shinkei kenkyu no shimpo"] = "Shinkei kenkyū no shimpo. Advances in neurological sciences",
  ["shinrigaku kenkyu"] = "Shinrigaku kenkyu : The Japanese journal of psychology",
  ["ship technol res"] = "Ship Technology Research",
  ["shipin anquan zhiliangjjiance xuebao"] = "Shipin anquan zhiliang jiance xuebao",
  ["shipin kexue (beijing)"] = "Shipin kexue",
  ["shipin yu shengwu jishu xuebao"] = "Shipin yu shengwu jishu xuebao",
  ["ships offshore struct"] = "Ships and Offshore Structures",
  ["shiraz e med j"] = "Shiraz e-medical journal",
  ["shirin (kyoto 1916)"] = "Shirin (Kyoto, Japan)",
  ["shishitsu seikagaku kenkyu"] = "Shishitsu seikagaku kenkyu. [Studies on biochemistry of lipids]",
  ["shiyan wulixue congshu"] = "Shiyan Wulixeu Congshu",
  ["shiyou huagong"] = "Shiyou Huagong",
  ["shiyou xuebao, shiyou jiagong"] = "Shiyou Xuebao, Shiyou Jiagong",
  ["shizhen guo yi guo yao"] = "Shizhen guo yi guo yao = Lishizhen medicine and materia medica research",
  ["shizuoka daigaku nogakubu kenkyu hokoku"] = "Shizuoka Daigaku Nōgakubu kenkyū hōkoku = Bulletin of the Faculty of Agriculture, Shizuoka University",
  ["shock vib"] = "Shock and Vibration",
  ["shock wave high press phenom"] = "Shock Wave and High Pressure Phenomena",
  ["shock waves"] = "Shock waves",
  ["shofar (west lafayette indiana)"] = "Shofar (West Lafayette, Ind.)",
  ["shokubutsu kenkyu zasshi"] = "Shokubutsu kenkyu zasshi. The Journal of Japanese botany",
  ["shokuhin eiseigaku zasshi"] = "Shokuhin eiseigaku zasshi. Journal of the Food Hygienic Society of Japan",
  ["shokuseishi kenkyu"] = "Shokuseishi kenkyu = Japanese journal of historical botany",
  ["shoni shikagaku zasshi"] = "Shōni shikagaku zasshi. The Japanese journal of pedodontics",
  ["shonika kiyo"] = "Shonika kiyo. Annales paediatrici Japonici",
  ["short textb log"] = "Short Textbooks in Logic",
  ["shou lei xue bao"] = "Shòu lèi xuè bào = Acta theriologica sinica",
  ["shoulder elbow"] = "Shoulder & elbow",
  ["showa igakkai zasshi"] = "Shōwa Igakkai zasshi = The Journal of the Showa Medical Association",
  ["showa shigakkai zasshi"] = "Shōwa Shigakkai zasshi = The Journal of Showa University Dental Society",
  ["shui chan yan jiu"] = "Shui chan yan jiu = Journal of Taiwan fisheries research",
  ["shui chu li ji shu"] = "Shui chu li ji shu",
  ["shui li yu ye"] = "Shui li yu ye",
  ["shui sheng sheng wu hsueh bao"] = "Shui sheng sheng wu hsueh bao = Acta hydrobiologica Sinica",
  ["shui sheng tai xue za zhi"] = "Shui sheng tai xue za zhi = Journal of hydroecology",
  ["shuichan xuebao"] = "Shuichan xuebao",
  ["shuichuli jishu"] = "Shuichuli Jishu",
  ["si yu yan"] = "Si yu yan",
  ["siam j appl algebr geom"] = "SIAM journal on applied algebra and geometry",
  ["siam j appl algebra geom"] = "SIAM Journal on Applied Algebra and Geometry",
  ["siam j appl dyn syst"] = "SIAM Journal on Applied Dynamical Systems",
  ["siam j appl math"] = "SIAM Journal on Applied Mathematics",
  ["siam j comput"] = "SIAM Journal on Computing",
  ["siam j control optim"] = "SIAM Journal on Control and Optimization",
  ["siam j discret math"] = "SIAM journal on discrete mathematics",
  ["siam j discrete math"] = "SIAM Journal on Discrete Mathematics",
  ["siam j financ math"] = "SIAM Journal on Financial Mathematics",
  ["siam j financial math"] = "SIAM Journal on Financial Mathematics",
  ["siam j imag sci"] = "SIAM Journal on Imaging Sciences",
  ["siam j imaging sci"] = "SIAM journal on imaging sciences",
  ["siam j math anal"] = "SIAM Journal on Mathematical Analysis",
  ["siam j math data sci"] = "SIAM Journal on Mathematics of Data Science",
  ["siam j matrix anal appl"] = "SIAM Journal on Matrix Analysis and Applications",
  ["siam j numer anal"] = "SIAM Journal on Numerical Analysis",
  ["siam j optim"] = "SIAM Journal on Optimization",
  ["siam j sci comput"] = "SIAM Journal on Scientific Computing",
  ["siam j sci stat comput"] = "SIAM Journal on Scientific and Statistical Computing",
  ["siam monogr math model comput"] = "SIAM Monographs on Mathematical Modeling and Computation",
  ["siam rev"] = "SIAM Review",
  ["siam rev soc ind appl math"] = "SIAM review. Society for Industrial and Applied Mathematics",
  ["siam stud appl math"] = "SIAM Studies in Applied Mathematics",
  ["siam undergrad res online"] = "SIAM undergraduate research online",
  ["siam/asa j uncertain quantif"] = "SIAM/ASA Journal on Uncertainty Quantification",
  ["siam/asa j uncertainty quantif"] = "SIAM/ASA Journal on Uncertainty Quantification",
  ["siauliai math semin"] = "Siauliai Mathematical Seminar",
  ["sib adv math"] = "Siberian Advances in Mathematics",
  ["sib elektron mat izv"] = "Sibirskie Elektronnye Matematicheskie Izvestiya. Siberian Electronic Mathematical Reports",
  ["sib math j"] = "Siberian Mathematical Journal",
  ["sib shk algebry log"] = "\\cyr Sibirskaya Shkola Algebry i Logiki",
  ["sib zh chist prikl mat"] = "Sibirskii Zhurnal Chistoi i Prikladnoi Matematiki",
  ["sib zh ind mat"] = "Sibirskii Zhurnal Industrialnoi Matematiki",
  ["sib zh vychisl mat"] = "Sibirskii Zhurnal Vychislitelnoi Matematiki. Siberian Journal of Numerical Mathematics. Rossiiskaya Akademiya Nauk. Sibirskoe Otdelenie",
  ["siberian adv math"] = "Siberian Advances in Mathematics",
  ["siberian j comput math"] = "Siberian Journal of Computer Mathematics",
  ["siberian j differential equations"] = "Siberian Journal of Differential Equations",
  ["siberian math j"] = "Siberian Mathematical Journal",
  ["siberian school algebra logic"] = "Siberian School of Algebra and Logic",
  ["sibirsk mat zh"] = "Rossiiskaya Akademiya Nauk. Sibirskoe Otdelenie. Institut Matematiki im. S. L. Soboleva. Sibirskii Matematicheskii Zhurnal",
  ["sibirskii ekolog z"] = "Sibirskii Ekologicheskii Zhurnal",
  ["sibirskii vestnik selskokhoziaistvennoi nauki"] = "Sibirskiĭ vestnik selʹskokhozi︠a︡ĭstvennoĭ nauki",
  ["sice j control meas syst integr"] = "SICE Journal of Control, Measurement, and System Integration",
  ["sichuan da xue xue bao"] = "Sichuan da xue xue bao = Sichuandaxue xuebao",
  ["sichuan da xue xue bao yi xue ban"] = "Sichuan da xue xue bao. Yi xue ban = Journal of Sichuan University. Medical science edition",
  ["sichuan da xue xue bao zi ran ke xue ban"] = "Sichuan da xue xue bao. Zi ran ke xue ban = Sichuan daxue xuebao",
  ["sichuan daxue xuebao"] = "Journal of Sichuan University. Natural Science Edition. Sichuan Daxue Xuebao. Ziran Kexue Ban",
  ["sichuan daxue xuebao|sichuan da xue xue bao zi ran ke xue ban"] = "Sichuan da xue xue bao. Zi ran ke xue ban",
  ["sichuan dong wu"] = "Sichuan dong wu = Sichuan dongwu = Sichuan journal of zoology",
  ["sichuan lin ye ke ji"] = "Sichuan lin ye ke ji = Journal of Sichuan forestry science and technology",
  ["sichuan shifan daxue xuebao"] = "Sichuan Shifan Daxue xuebao (Ziran kexue ban)",
  ["sichuan shifan daxue xuebao ziran kexue ban"] = "Sichuan Shifan Daxue Xuebao",
  ["sichuan yi xue yuan xue bao"] = "Sichuan yi xue yuan xue bao = Acta Academiae Medicinae Sichuan",
  ["sicil sanit"] = "Sicilia sanitaria",
  ["sicilia med"] = "Sicilia medica",
  ["sicot j"] = "SICOT-J",
  ["sics software-intensive cyber-phys syst"] = "SICS Software-Intensive Cyber-Physical Systems",
  ["sid symp dig tech pap"] = "SID Symposium Digest of Technical Papers",
  ["sieccan j"] = "Sieccan journal",
  ["siecus rep"] = "SIECUS report",
  ["siemens rev"] = "Siemens Review",
  ["siggraph asia 2017 symp vis (2017)"] = "SIGGRAPH Asia 2017 Symposium on Visualization. SIGGRAPH Asia Symposium on Visualization (2017 : Bangkok, Thailand)",
  ["sight life mag"] = "Sight and life magazine",
  ["sight sav rev"] = "The Sight-saving review",
  ["sightsav rev"] = "Sightsaving Review",
  ["sigkdd explor"] = "SIGKDD explorations : newsletter of the Special Interest Group (SIG) on Knowledge Discovery & Data Mining",
  ["sigma j eng nat sci"] = "Sigma Journal of Engineering and Natural Sciences",
  ["sigma ser appl math"] = "Sigma Series in Applied Mathematics",
  ["sigma ser pure math"] = "Sigma Series in Pure Mathematics",
  ["sigma symmetry integrability geom methods appl"] = "SIGMA. Symmetry, Integrability and Geometry. Methods and Applications",
  ["sigmod rec"] = "SIGMOD record",
  ["sigmul bunryu hag hoeji"] = "Sigmul bunryu hag-hoeji = Korean journal of plant taxonomy",
  ["sigmul saengmyeong gong haghoeji"] = "Sigmul saengmyeong gong haghoeji = Korean journal of plant biotechnology",
  ["sigmulbyeong yeongu"] = "Sigmulbyeong yeon'gu",
  ["sigmund freud house bull"] = "Sigmund Freud House bulletin",
  ["sign lang linguist"] = "Sign language and linguistics",
  ["sign lang stud"] = "Sign language studies",
  ["sign transduct insights"] = "Signal transduction insights",
  ["signal image process"] = "Signal and image processing : an international journal",
  ["signal image video process"] = "Signal, image and video processing",
  ["signal process"] = "Signal Processing",
  ["signal process digit filtering"] = "Signal Processing and Digital Filtering",
  ["signal process image commun"] = "Signal Processing: Image Communication",
  ["signal processing"] = "Signal processing",
  ["signal transduct"] = "Signal transduction",
  ["signal transduct target ther"] = "Signal transduction and targeted therapy",
  ["signal transduction"] = "Signal Transduction",
  ["signal transduction insights"] = "Signal Transduction Insights",
  ["signal transduction targeted ther"] = "Signal Transduction and Targeted Therapy",
  ["signal, image video process"] = "Signal, Image and Video Processing",
  ["signals commun technol"] = "Signals and Communication Technology",
  ["signif (oxf)"] = "Significance (Oxford, England)",
  ["signs (chic)"] = "Signs",
  ["signs soc (chic)"] = "Signs and society (Chicago, Ill.)",
  ["silfs ser"] = "Italian Society for Logic and Philosophy of Science SILFS Series",
  ["silic indus"] = "Silicates Industriels",
  ["silica assoc respirable miner part (2012)"] = "Silica and associated respirable mineral particles",
  ["silicon chem"] = "Silicon Chemistry",
  ["sillul baft"] = "Sillūl va bāft",
  ["silva balcan"] = "Silva Balcanica",
  ["silva fenn"] = "Silva Fennica",
  ["silva fenn (hels)"] = "Silva fennica (Helsinki, Finland : 1967)",
  ["silva lusit"] = "Silva lusitana",
  ["silva lusitana"] = "Silva Lusitana",
  ["silvae genet"] = "Silvae genetica",
  ["silvaecult trop subtrop"] = "Silvaecultura Tropica et Subtropica",
  ["simon wiesenthal cent annu"] = "Simon Wiesenthal Center annual",
  ["simons symp"] = "Simons Symposia",
  ["simul found methods appl"] = "Simulation Foundations, Methods and Applications",
  ["simul games"] = "Simulation & games",
  ["simul gaming"] = "Simulation & gaming",
  ["simul healthc"] = "Simulation in healthcare : journal of the Society for Simulation in Healthcare",
  ["simul model pract theory"] = "Simulation modelling practice and theory",
  ["simul modell pract theory"] = "Simulation Modelling Practice and Theory",
  ["simul pract theory"] = "Simulation Practice and Theory",
  ["simula springerbriefs comput"] = "Simula SpringerBriefs on Computing",
  ["sinai hosp j (balt)"] = "Sinai Hospital journal",
  ["sind med j"] = "The Sind medical journal",
  ["singap j trop geogr"] = "Singapore journal of tropical geography",
  ["singap libr"] = "Singapore libraries",
  ["singap stat bull"] = "Singapore statistical bulletin. Singapore. National Statistical Commission",
  ["singapore dent j"] = "Singapore dental journal",
  ["singapore econ rev"] = "Singapore Economic Review",
  ["singapore fam physician"] = "The Singapore family physician",
  ["singapore j obstet gynecol"] = "Singapore journal of obstetrics & gynaecology",
  ["singapore j trop geogr"] = "Singapore Journal of Tropical Geography",
  ["singapore med j"] = "Singapore medical journal",
  ["singapore public health bull"] = "The Singapore public health bulletin",
  ["single cell biol"] = "Single cell biology",
  ["singmul hakhoe chi"] = "Singmul Hakhoe chi. The Korean journal of botany",
  ["sinh hoc"] = "Sinh học",
  ["sino west cult relat j"] = "Sino-Western cultural relations journal = Zhong Xi wen hua jiao liu shi za zhi",
  ["sint trab cient"] = "Síntesis de trabajos científicos",
  ["sintesi medica"] = "Sintesi medica",
  ["sir lanka j popul stud"] = "Sri Lanka journal of population studies",
  ["siriraj hosp gaz"] = "Siriraj Hospital gazette",
  ["siriraj med j"] = "Siriraj medical journal",
  ["sisli etfal hastan tıp bul"] = "Şişli Etfal Hastanesi tıp bülteni",
  ["sissa springer ser"] = "SISSA Springer Series",
  ["sist nerv"] = "Sistema nervoso",
  ["sist tecnol inf (2017)"] = "Sistemas e Tecnologias de Informação : atas de 12a Conferência Ibérica de Sistemas e Tecnologias de Informação (CISTI'2017) : 21 a 24 de junho de 2017, Lisboa, Portugal = Information systems and technologies : proceedings of the 12th Iberian Conference on Information Systems and Technologies (CISTI'2017) : 21 to 24 of June 2017, Lisbon, Portugal. Iberian Conference on Information Systems and Technologies (12th : 2017 : Lisbon, Portugal)",
  ["sist urbani"] = "Sistemi urbani",
  ["sitzungsber akad wiss ddr math naturwiss tech jahrg 1976"] = "Sitzungsberichte der Akademie der Wissenschaften der DDR, Mathematik—Naturwissenschaften—Technik, Jahrgang 1976",
  ["sitzungsber akad wiss ddr, math, naturwiss, tech"] = "Sitzungsberichte der Akademie der Wissenschaften der DDR, Mathematik, Naturwissenschaften, Technik",
  ["sitzungsber bayer akad wiss math naturwiss kl"] = "Sitzungsberichte Bayerische Akademie der Wissenschaften, Mathematische - Naturwissenschaftliche Klasse",
  ["sitzungsber heidelb akad wiss math naturwiss kl"] = "Sitzungsberichte der Heidelberger Akademie der Wissenschaften, Mathematisch-Naturwissenschaftliche Klasse",
  ["sitzungsber k preuss akad wiss"] = "Sitzungsberichte der Koniglichen Preussischen Akademie der Wissenschaften",
  ["sitzungsber math-naturwiss kl"] = "Sitzungsberichte der Mathematisch-Naturwissenschaftlichen Klasse",
  ["sitzungsber sachsischen akad wissenschaft leipzig math nat sci sect band 116"] = "Sitzungsberichte der Sachsischen Akademie der Wissenschaften zu Leipzig. Mathematisch-naturwissenschaftliche Klasse. Band 116",
  ["sitzungsber wiss ges johann wolfgang goethe-univ frankfurt am main"] = "Sitzungsberichte der Wissenschaftlichen Gesellschaft an der Johann Wolfgang Goethe-Universität Frankfurt am Main",
  ["sixt century j"] = "The Sixteenth century journal",
  ["sixth n am paleontol conv n am paleontol conv 6th 1996 smithson inst washington dc"] = "Sixth North American Paleontological Convention : abstracts and program : Field Museum of Natural History, June 9-June 12, 1996. North American Paleontological Convention (6th : 1996 : Smithsonian Institution, Washington, D.C.)",
  ["sixth-gener comput tech ser"] = "Sixth-Generation Computer Technology Series",
  ["skand arch physiol"] = "Skandinavisches Archiv fur Physiologie",
  ["skand banken quart rev"] = "Skandinaviska Banken Quarterly Review",
  ["skandinavisk vet tidsskr bakteriol patol samt kjottoch mjokhygienen"] = "Skandinavisk veterinär-tidsskrift för bakteriologi, patologi, samt kjöttoch mjölkhygienen",
  ["skelet muscle"] = "Skeletal muscle",
  ["skeletal radiol"] = "Skeletal radiology",
  ["skin (los angeles)"] = "Skin",
  ["skin appendage disord"] = "Skin appendage disorders",
  ["skin pharmacol"] = "Skin pharmacology : the official journal of the Skin Pharmacology Society",
  ["skin pharmacol appl skin physiol"] = "Skin pharmacology and applied skin physiology",
  ["skin pharmacol physiol"] = "Skin pharmacology and physiology",
  ["skin res technol"] = "Skin research and technology : official journal of International Society for Bioengineering and the Skin (ISBS) [and] International Society for Digital Imaging of Skin (ISDIS) [and] International Society for Skin Imaging (ISSI)",
  ["skin therapy lett"] = "Skin therapy letter",
  ["skript math statist"] = "Skripten zur Mathematischen Statistik",
  ["skrypty uniw śląskiego"] = "Skrypty Uniwersytetu Śląskiego",
  ["skuast j res"] = "SKUAST journal of research",
  ["skull base rep"] = "Skull base reports",
  ["skull base surg"] = "Skull base surgery",
  ["sky telescope"] = "Sky and telescope",
  ["slas discov"] = "SLAS discovery : advancing life sciences R & D",
  ["slas discovery"] = "SLAS Discovery",
  ["slas technol"] = "SLAS technology",
  ["slaski kwat hist sabotka"] = "Ślas̨ki kwartalnik historyczny Sobótka",
  ["slav east eur rev"] = "Slavonic and East European review",
  ["slavery abol"] = "Slavery & abolition",
  ["slavic rev"] = "Slavic review",
  ["sleep adv"] = "Sleep advances : a journal of the Sleep Research Society",
  ["sleep biol rhythms"] = "Sleep and biological rhythms",
  ["sleep breath"] = "Sleep & breathing = Schlaf & Atmung",
  ["sleep breathing"] = "Sleep and Breathing",
  ["sleep disord"] = "Sleep disorders",
  ["sleep health"] = "Sleep health",
  ["sleep hypn"] = "Sleep and hypnosis : an international journal of sleep, dream, and hypnosis",
  ["sleep med"] = "Sleep medicine",
  ["sleep med clin"] = "Sleep medicine clinics",
  ["sleep med disord"] = "Sleep medicine and disorders : international journal",
  ["sleep med res"] = "Sleep medicine research",
  ["sleep med rev"] = "Sleep medicine reviews",
  ["sleep med x"] = "Sleep medicine: X",
  ["sleep med: x"] = "Sleep Medicine: X",
  ["sleep res online"] = "Sleep research online : SRO",
  ["sleep sci"] = "Sleep science (São Paulo, Brazil)",
  ["sleep sci pract"] = "Sleep science and practice",
  ["sleep vigil"] = "Sleep and vigilance",
  ["sleep vigilance"] = "Sleep and Vigilance",
  ["slezsky sb"] = "Slezský sborník",
  ["sloan manage rev"] = "Sloan management review",
  ["slov arch"] = "Slovenská archivistika",
  ["slov lek"] = "Slovensky lekár",
  ["slov prehl"] = "Slovanský přehled",
  ["slov vet zb"] = "Slovenski veterinarski zbornik = Slovenian veterinary research",
  ["slown biogr pol med xx wieku"] = "Słownik biograficzny polskich nauk medycznych XX wieku",
  ["sm j clin med imaging"] = "SM journal of clinical and medical imaging",
  ["sm j community med"] = "SM journal of community medicine",
  ["sm j reprod health infertil"] = "SM journal of reproductive health & infertility",
  ["sm radiol j"] = "SM radiology journal",
  ["sm trop med j"] = "SM tropical medicine journal",
  ["smai j comput math"] = "SMAI Journal of Computational Mathematics",
  ["small adv healthcare mater"] = "Small and Advanced Healthcare Materials",
  ["small bus econ"] = "Small Business Economics",
  ["small carniv conserv"] = "Small carnivore conservation : the newsletter and journal of the IUCN/SSC Mustelid, Viverrid & Procynid Specialist Group",
  ["small group res"] = "Small group research",
  ["small gtpases"] = "Small GTPases",
  ["small methods"] = "Small methods",
  ["small mol"] = "Small Molecules",
  ["small rumin res"] = "Small ruminant research : the journal of the International Goat Association",
  ["small ruminant res"] = "Small Ruminant Research",
  ["small sci"] = "Small Science",
  ["small struct"] = "Small structures",
  ["small-scale for"] = "Small-Scale Forestry",
  ["small-scale for econ manage policy"] = "Small-scale Forest Economics, Management and Policy",
  ["smarandache notions j"] = "Smarandache Notions Journal",
  ["smart agric technol"] = "Smart Agricultural Technology",
  ["smart energy"] = "Smart Energy",
  ["smart health"] = "Smart Health",
  ["smart health (amst)"] = "Smart health (Amsterdam, Netherlands)",
  ["smart homecare technol telehealth"] = "Smart homecare technology and telehealth",
  ["smart infrastruct constr"] = "Smart Infrastructure and Construction",
  ["smart innov syst technol"] = "Smart Innovation, Systems and Technologies",
  ["smart learn environ"] = "Smart Learning Environments",
  ["smart manuf"] = "Smart Manufacturing",
  ["smart mater bull"] = "Smart Materials Bulletin",
  ["smart mater manuf"] = "Smart Materials in Manufacturing",
  ["smart mater med"] = "Smart materials in medicine",
  ["smart mater struct"] = "Smart Materials and Structures",
  ["smart med"] = "Smart Medicine",
  ["smart resilient transp"] = "Smart and Resilient Transportation",
  ["smart sci"] = "Smart Science",
  ["smart struct mater nondestruct eval health monit diagn"] = "Smart structures and materials. Nondestructive evaluation for health monitoring and diagnostics",
  ["smart struct syst"] = "Smart structures and systems",
  ["smart sustain manuf syst"] = "Smart and sustainable manufacturing systems",
  ["smart sustainable built environ"] = "Smart and Sustainable Built Environment",
  ["smart sustainable manuf syst"] = "Smart and Sustainable Manufacturing Systems",
  ["smart water"] = "Smart Water",
  ["smf journ annu"] = "SMF Journee Annuelle",
  ["smith coll stud soc work"] = "Smith College studies in social work",
  ["smith-hurd ill annot statut ill"] = "Smith-Hurd Illinois annotated statutes. Illinois",
  ["smithiana bull"] = "Smithiana. Bulletin",
  ["smithson ann flight"] = "Smithsonian annals of flight",
  ["smithson contrib mar sci"] = "Smithsonian contributions to the marine sciences",
  ["smithson contrib zool"] = "Smithsonian contributions to zoology",
  ["smithson stud hist technol"] = "Smithsonian studies in history and technology. Smithsonian Institution",
  ["smpte motion imaging j"] = "SMPTE motion imaging journal",
  ["smu law rev"] = "SMU law review : a publication of Southern Methodist University School of Law",
  ["sn  partial differ equations appl"] = "SN Partial Differential Equations and Applications",
  ["sn appl sci"] = "SN applied sciences",
  ["sn compr clin med"] = "SN comprehensive clinical medicine",
  ["sn comput sci"] = "SN computer science",
  ["sn oper res forum"] = "SN Operations Research Forum",
  ["sn partial differ equ appl"] = "SN partial differential equations and applications",
  ["snakeroot extr"] = "Snakeroot extract",
  ["snudebiller (monchengladb)"] = "Snudebiller : studies on taxonomy, biology, and ecology of Curculionoidea",
  ["soa chongsonyon chongsin uihak"] = "Soa--chʻŏngsonyŏn chŏngsin ŭihak = Journal of child & adolescent psychiatry",
  ["soap perfum cosmet (lond)"] = "Soap, perfumery & cosmetics",
  ["soap sanit chem"] = "Soap and sanitary chemicals",
  ["soap sanit chem blue book cat ed"] = "Soap and sanitary chemicals. Blue book and catalog ed",
  ["soap, cosmet, chem spec"] = "Soap, Cosmetics, Chemical Specialties",
  ["soc action"] = "Social action",
  ["soc afr sida"] = "Sociétés d'Afrique & SIDA : newsletter of the Sociétés d'Afrique & SIDA Network",
  ["soc anal"] = "Social analysis (Adelaide, S.A.)",
  ["soc anc med rev"] = "Society for Ancient Medicine review",
  ["soc anim"] = "Society & animals : social scientific studies of the human experience of other animals",
  ["soc anthropol"] = "Social anthropology : the journal of the European Association of Social Anthropologists = Anthropologie sociale",
  ["soc appl bacteriol symp ser"] = "Society for Applied Bacteriology symposium series",
  ["soc astron lett"] = "Soviet Astronomy Letters [translation of Pisma v Astronomicheskii Zhurnal]",
  ["soc behav"] = "Social behaviour",
  ["soc behav pers"] = "Social behavior and personality",
  ["soc biol"] = "Social biology",
  ["soc biol hum aff"] = "Social biology and human affairs",
  ["soc casework"] = "Social casework",
  ["soc change"] = "Social change",
  ["soc choice welf"] = "Social Choice and Welfare",
  ["soc choice welfare"] = "Social choice and welfare",
  ["soc cogn"] = "Social cognition",
  ["soc cogn affect neurosci"] = "Social cognitive and affective neuroscience",
  ["soc compass"] = "Social compass",
  ["soc comput behav cult model predict (2015)"] = "Social computing, behavioral-cultural modeling, and prediction : 8th International Conference, SBP 2015, Washington, DC, USA, March 31-April 3, 2015. Proceedings. International Conference on Social Computing, Behavioral Modeling and Prediction (8th : 2015 : Washington, D.C.)",
  ["soc cult"] = "Society and culture",
  ["soc cult behav model (2016)"] = "Social, Cultural, and Behavioral Modeling : 9th International Conference, SBP-BRiMS 2016, Washington, DC, USA, June 28-July 1, 2016, Proceedings. International Conference on Social Computing, Behavioral Modeling and Prediction (9th : 2016 : Washington, D.C.)",
  ["soc cult geogr"] = "Social & cultural geography",
  ["soc curr"] = "Social currents",
  ["soc dev"] = "Social development (Oxford, England)",
  ["soc dev issues"] = "Social development issues",
  ["soc dyn"] = "Social dynamics",
  ["soc econ adm (lond)"] = "Social and economic administration",
  ["soc econ stud"] = "Social and economic studies",
  ["soc epistemol"] = "Social epistemology",
  ["soc forces"] = "Social forces; a scientific medium of social study and interpretation",
  ["soc gen physiol ser"] = "Society of General Physiologists series",
  ["soc hist"] = "Social history",
  ["soc hist alcohol drugs"] = "The social history of alcohol and drugs",
  ["soc hist alcohol rev"] = "The Social history of alcohol review",
  ["soc hist med"] = "Social history of medicine : the journal of the Society for the Social History of Medicine",
  ["soc incl"] = "Social inclusion",
  ["soc indic res"] = "Social indicators research",
  ["soc influ"] = "Social influence",
  ["soc inform (2016)"] = "Social informatics : 8th International Conference, SocInfo 2016, Bellevue, WA, USA, November 11-14, 2016, Proceedings. Part II. SocInfo (Conference) (8th : 2016 : Bellevue, Wash.)",
  ["soc issues policy rev"] = "Social issues and policy review",
  ["soc ital dermatol sifilogr sezioni interprov soc ital dermatol sifilogr"] = "Atti della Società italiana di dermatologia e sifilografia e delle sezioni interprovinciali. Società italiana di dermatologia e sifilografia",
  ["soc ital ecol, atti"] = "Società Italiana di Ecologia, Atti",
  ["soc justice"] = "Social justice (San Francisco, Calif.)",
  ["soc justice res"] = "Social justice research",
  ["soc justice rev"] = "Social justice review",
  ["soc labour bull"] = "Social and labour bulletin",
  ["soc leg stud"] = "Social & legal studies",
  ["soc malawi j"] = "The Society of Malaŵi journal. Society of Malawi (Historical and Scientific)",
  ["soc mar q"] = "Social marketing quarterly",
  ["soc mark forum"] = "Social Marketing Forum",
  ["soc mark update"] = "Social marketing update",
  ["soc med (soc med publ group)"] = "Social medicine (Social Medicine Publication Group)",
  ["soc med tidskr"] = "Social-Medicinsk tidskrift",
  ["soc media soc"] = "Social media + society",
  ["soc ment health"] = "Society and mental health",
  ["soc mov stud"] = "Social movement studies",
  ["soc nat resour"] = "Society & natural resources",
  ["soc netw"] = "Social networking",
  ["soc netw anal min"] = "Social network analysis and mining",
  ["soc networks"] = "Social networks",
  ["soc neurosci"] = "Social neuroscience",
  ["soc nurs hist gaz"] = "The Society for Nursing History gazette",
  ["soc personal psychol compass"] = "Social and personality psychology compass",
  ["soc philos policy"] = "Social philosophy & policy",
  ["soc photogr instrum eng j"] = "Society of Photographic Instrumentation Engineers Journal",
  ["soc policy"] = "Social policy",
  ["soc policy adm"] = "Social policy & administration",
  ["soc policy rep"] = "Social policy report",
  ["soc policy soc"] = "Social policy and society : a journal of the Social Policy Association",
  ["soc polit"] = "Social politics",
  ["soc politica"] = "Societate şi politică",
  ["soc prax"] = "Social praxis",
  ["soc probl"] = "Social problems",
  ["soc psychiatry"] = "Social psychiatry. Sozialpsychiatrie. Psychiatrie sociale",
  ["soc psychiatry psychiatr epidemiol"] = "Social psychiatry and psychiatric epidemiology",
  ["soc psychol"] = "Social psychology",
  ["soc psychol educ"] = "Social psychology of education : an international journal",
  ["soc psychol personal sci"] = "Social psychological and personality science",
  ["soc psychol q"] = "Social psychology quarterly",
  ["soc rehabil rec"] = "The Social and rehabilitation record",
  ["soc reprod fertil suppl"] = "Society of Reproduction and Fertility supplement",
  ["soc res"] = "Social Research",
  ["soc res (new york)"] = "Social research",
  ["soc responsib journal law med"] = "Social responsibility, journalism, law, medicine",
  ["soc sci"] = "Social science",
  ["soc sci (basel)"] = "Social sciences (Basel, Switzerland)",
  ["soc sci (new delhi)"] = "Social scientist",
  ["soc sci china"] = "Social sciences in China",
  ["soc sci comput rev"] = "Social science computer review",
  ["soc sci dent"] = "Social science and dentistry",
  ["soc sci diliman"] = "Social science Diliman",
  ["soc sci hist"] = "Social science history",
  ["soc sci inf (paris)"] = "Social sciences information. Information sur les sciences sociales",
  ["soc sci inf stud"] = "Social science information studies : SSIS",
  ["soc sci j"] = "The Social science journal",
  ["soc sci japan j"] = "Social Science Japan Journal",
  ["soc sci jpn j"] = "Social science Japan journal",
  ["soc sci med"] = "Social science & medicine (1982)",
  ["soc sci med [a]"] = "Social Science and Medicine. Part A, Medical Sociology",
  ["soc sci med [b]"] = "Social Science and Medicine. Part B, Medical Anthropology",
  ["soc sci med [c]"] = "Social Science and Medicine. Part C, Medical Economics",
  ["soc sci med [d]"] = "Social Science and Medicine. Part D, Medical Geography",
  ["soc sci med [e]"] = "Social Science and Medicine. Part E, Medical Psychology",
  ["soc sci med [f]"] = "Social Science and Medicine. Part F, Medical and Social Ethics",
  ["soc sci med [med anthropol]"] = "Social Science and Medicine. Medical Anthropology",
  ["soc sci med [med econ]"] = "Social Science and Medicine. Medical Economics",
  ["soc sci med [med geogr]"] = "Social Science and Medicine. Medical Geography",
  ["soc sci med [med psychol med sociol]"] = "Social Science and Medicine. Medical Psychology and Medical Sociology",
  ["soc sci med a"] = "Social science & medicine. Part A, Medical sociology",
  ["soc sci med b"] = "Social science & medicine. Part B, Medical anthropology",
  ["soc sci med c"] = "Social science & medicine. Part C, Medical economics",
  ["soc sci med d"] = "Social science & medicine. Part D, Medical geography",
  ["soc sci med e"] = "Social science & medicine. Part E, Medical psychology",
  ["soc sci med f"] = "Social science & medicine. Part F, Medical & social ethics",
  ["soc sci med med anthropol"] = "Social science & medicine. Medical anthropology",
  ["soc sci med med econ"] = "Social science & medicine. Medical economics",
  ["soc sci med med geogr"] = "Social science & medicine. Medical geography",
  ["soc sci med med psychol med sociol"] = "Social science & medicine. Medical psychology & medical sociology",
  ["soc sci new lett"] = "Social Science News Letter",
  ["soc sci q"] = "Social science quarterly",
  ["soc sci quart"] = "Social Science Quarterly",
  ["soc sci res"] = "Social science research",
  ["soc sec bull"] = "Social Security Bulletin",
  ["soc secur bull"] = "Social security bulletin",
  ["soc secur bull annu stat suppl"] = "Social security bulletin. Annual statistical supplement",
  ["soc serv rev"] = "The Social service review",
  ["soc soc hist med bull (lond)"] = "The Society for the Social History of Medicine bulletin",
  ["soc stor"] = "Società e storia",
  ["soc stud"] = "Social studies (Philadelphia, Pa. : 1953)",
  ["soc stud (maynooth)"] = "Social studies (Maynooth, Ireland)",
  ["soc stud sci"] = "Social studies of science",
  ["soc theory health"] = "Social theory & health : STH",
  ["soc theory pract"] = "Social theory and practice",
  ["soc thought"] = "Social thought",
  ["soc welfare (india)"] = "Social welfare",
  ["soc wetenschap"] = "Sociale wetenschappen",
  ["soc work"] = "Social work",
  ["soc work educ"] = "Social work in education",
  ["soc work educ dev newsl"] = "Social work education and development newsletter",
  ["soc work forum (n y n y)"] = "Social work forum (New York, N.Y.)",
  ["soc work groups"] = "Social work with groups",
  ["soc work health care"] = "Social work in health care",
  ["soc work ment health"] = "Social work in mental health",
  ["soc work pract res"] = "The social work practitioner-researcher",
  ["soc work public health"] = "Social work in public health",
  ["soc work res"] = "Social work research",
  ["soc work res abstr"] = "Social work research & abstracts",
  ["soc work soc"] = "Social work & society",
  ["soc work soc sci rev"] = "Social work & social sciences review",
  ["soccer soc"] = "Soccer and society",
  ["social indic res"] = "Social Indicators Research",
  ["social network anal min"] = "Social Network Analysis and Mining",
  ["social neurosci"] = "Social Neuroscience",
  ["social rev"] = "Socialist review",
  ["social revolut"] = "Socialist revolution",
  ["social sci comput rev"] = "Social Science Computer Review",
  ["social work (stellenbosch)"] = "Social Work",
  ["socialmed tidskr skriftser"] = "Socialmedicinsk tidskrift. Skriftserie",
  ["societies (basel)"] = "Societies (Basel, Switzerland)",
  ["socio-ecol pract res"] = "Socio-Ecological Practice Research",
  ["socioaffect neurosci psychol"] = "Socioaffective neuroscience & psychology",
  ["socioeco prac res"] = "Socio-ecological practice research",
  ["socioecon issues health"] = "Socioeconomic issues of health",
  ["socioecon plann sci"] = "Socio-economic planning sciences",
  ["socioecon rev"] = "Socio-economic review",
  ["sociol anal"] = "SA. Sociological analysis",
  ["sociol anthropol (alhambra)"] = "Sociology and anthropology (Alhambra, Calif.)",
  ["sociol bull"] = "Sociological bulletin",
  ["sociol cas"] = "Sociologický časopis",
  ["sociol compass"] = "Sociology compass",
  ["sociol dev (oakl)"] = "Sociology of development (Oakland, Calif.)",
  ["sociol diritto"] = "Sociologia del diritto",
  ["sociol educ"] = "Sociology of education",
  ["sociol focus"] = "Sociological focus",
  ["sociol forsk"] = "Sociologisk Forskning",
  ["sociol forum"] = "Sociological Forum",
  ["sociol forum (randolph n j)"] = "Sociological forum (Randolph, N.J.)",
  ["sociol gids"] = "Sociologische gids",
  ["sociol health ill"] = "Sociology Of Health & Illness",
  ["sociol health illn"] = "Sociology of health & illness",
  ["sociol inq"] = "Sociological inquiry",
  ["sociol lav"] = "Sociologia del lavoro",
  ["sociol methodol"] = "Sociological methodology",
  ["sociol methods res"] = "Sociological methods & research",
  ["sociol neerl"] = "Sociologia Neerlandica",
  ["sociol perspect"] = "Sociological perspectives : SP : official publication of the Pacific Sociological Association",
  ["sociol q"] = "The Sociological quarterly",
  ["sociol race ethn (thousand oaks)"] = "Sociology of race and ethnicity (Thousand Oaks, Calif.)",
  ["sociol relig"] = "Sociology of religion",
  ["sociol res"] = "Sociological research",
  ["sociol res online"] = "Sociological research online",
  ["sociol rev"] = "The Sociological review",
  ["sociol rev [monogr]"] = "Sociological Review Monograph",
  ["sociol rev monogr"] = "Sociological review monograph",
  ["sociol rom"] = "Sociologie românească",
  ["sociol rural"] = "Sociologia Ruralis",
  ["sociol ruralis"] = "Sociologia ruralis",
  ["sociol sci"] = "Sociological science",
  ["sociol sci yearbook"] = "Sociology of the Sciences. Yearbook",
  ["sociol sela"] = "Sociologija sela",
  ["sociol soc"] = "Sociologie et sociétés",
  ["sociol soc res"] = "Sociology and social research",
  ["sociol spectr"] = "Sociological spectrum : the official journal of the Mid-South Sociological Association",
  ["sociol sport j"] = "Sociology of sport journal",
  ["sociol stud child youth"] = "Sociological studies of children and youth",
  ["sociol symp"] = "Sociological symposium",
  ["sociol theor"] = "Sociological Theory",
  ["sociol theor methods"] = "Sociological Theory And Methods",
  ["sociol theory"] = "Sociological theory",
  ["sociol trav"] = "Sociologie du travail",
  ["sociol urbana rurale (testo stamp)"] = "Sociologia urbana e rurale",
  ["sociol work occup"] = "Sociology of work and occupations",
  ["sociol-j brit sociol assoc"] = "Sociology-The Journal Of The British Sociological Association",
  ["socra source"] = "SOCRA source",
  ["sodinink darzinink"] = "Sodininkyste ir darzininkyste",
  ["soemmering forsch"] = "Soemmerring-Forschungen",
  ["soft biol matter"] = "Soft and Biological Matter",
  ["soft comput"] = "Soft computing",
  ["soft comput lett"] = "Soft Computing Letters",
  ["soft mater"] = "Soft Materials",
  ["soft matter"] = "Soft matter",
  ["soft rob"] = "Soft Robotics",
  ["soft robot"] = "Soft robotics",
  ["softw eng healthc syst sehs ieee acm int workshop"] = "Software Engineering in Healthcare Systems (SEHS), IEEE/ACM International Workshop on",
  ["softw healthc"] = "Software in healthcare",
  ["softw law j"] = "Software law journal",
  ["softw pract exp"] = "Software: practice & experience",
  ["softw syst model"] = "Software and systems modeling",
  ["softw test verif reliab"] = "Software Testing, Verification and Reliability",
  ["softw: pract exper"] = "Software: Practice and Experience",
  ["software environ tools"] = "Software, Environments, and Tools",
  ["software impacts"] = "Software Impacts",
  ["software qual j"] = "Software Quality Journal",
  ["software syst model"] = "Software and Systems Modeling",
  ["sogo igaku"] = "Sogo igaku. Medicine",
  ["sogo kango"] = "Sogo kango. Comprehensive nursing, quarterly",
  ["sogo rinsho"] = "[Sōgō rinshō] Clinic all-round",
  ["soil biol biochem"] = "Soil biology & biochemistry",
  ["soil discuss"] = "Soil Discussions",
  ["soil dyn earthquake eng"] = "Soil Dynamics and Earthquake Engineering",
  ["soil ecol lett"] = "Soil Ecology Letters",
  ["soil horiz"] = "Soil Horizons",
  ["soil mech found eng"] = "Soil Mechanics and Foundation Engineering",
  ["soil org"] = "Soil organisms",
  ["soil res"] = "Soil research (Collingwood, Vic.)",
  ["soil sci"] = "Soil science",
  ["soil sci plant nutr"] = "Soil science and plant nutrition",
  ["soil sci soc am j"] = "Soil Science Society of America Journal",
  ["soil secur"] = "Soil Security",
  ["soil sediment contam"] = "Soil and Sediment Contamination",
  ["soil syst"] = "Soil systems",
  ["soil technol"] = "Soil Technology",
  ["soil tillage res"] = "Soil and Tillage Research",
  ["soil use manag"] = "Soil use and management",
  ["soil use manage"] = "Soil Use and Management",
  ["soil water res"] = "Soil and Water Research",
  ["soils found"] = "Soils and Foundations",
  ["soins cardiol"] = "Soins. Cardiologie",
  ["soins chir"] = "Soins. Chirurgie",
  ["soins chir gen spec"] = "Soins. Chirurgie generale et specialisée",
  ["soins form pedagog encadr"] = "Soins. Formation, pédagogie, encadrement : avec la participation du CEEIEC",
  ["soins gerontol"] = "Soins. Gérontologie",
  ["soins gynecol obstet pueric"] = "Soins. Gynécologie, obstétrique, puériculture",
  ["soins gynecol obstet pueric pediatr"] = "Soins. Gynécologie, obstétrique, puériculture, pédiatrie",
  ["soins pathol trop"] = "Soins. Pathologie tropicale",
  ["soins pediatr pueric"] = "Soins. Pédiatrie, puériculture",
  ["soins psychiatr"] = "Soins. Psychiatrie",
  ["soj anesthesiol pain manag"] = "SOJ anesthesiology & pain management",
  ["soj biochem"] = "SOJ biochemistry",
  ["soj immunol"] = "SOJ immunology",
  ["soj microbiol infect dis"] = "SOJ microbiology & infectious diseases",
  ["soj neurol"] = "SOJ neurology",
  ["soj nurs health care"] = "SOJ nursing & health care",
  ["soj pharm pharm sci"] = "SOJ pharmacy & pharmaceutical sciences",
  ["soj psychol"] = "SOJ psychology",
  ["soj surg"] = "SOJ surgery",
  ["soj urol nephrol open access"] = "SOJ urology and nephrology open access",
  ["sokoto j med lab sci"] = "Sokoto journal of medical laboratory science",
  ["sol compass"] = "Solar Compass",
  ["sol energ mat sol"] = "Solar Energy Materials and Solar Cells",
  ["sol energy"] = "Solar Energy",
  ["sol energy adv"] = "Solar Energy Advances",
  ["sol energy mater"] = "Solar Energy Materials",
  ["sol energy mater sol cells"] = "Solar Energy Materials and Solar Cells",
  ["sol phys"] = "Solar Physics",
  ["sol rrl"] = "Solar RRL",
  ["sol syst res"] = "Solar System Research",
  ["solar cells"] = "Solar Cells",
  ["solar wind technol"] = "Solar & Wind Technology",
  ["soldering surf mount technol"] = "Soldering & Surface Mount Technology",
  ["solicit j"] = "Solicitors' journal (London, England : 1928)",
  ["solid earth"] = "Solid Earth",
  ["solid earth discuss"] = "Solid Earth Discussions",
  ["solid earth sci"] = "Solid Earth Science",
  ["solid freeform fabr symp proc"] = "Solid Freeform Fabrication Symposium proceedings. Solid Freeform Fabrication Symposium",
  ["solid fuel chem"] = "Solid Fuel Chemistry",
  ["solid mech appl"] = "Solid Mechanics and its Applications",
  ["solid state commun"] = "Solid State Communications",
  ["solid state electron"] = "Solid-state electronics",
  ["solid state electron lett"] = "Solid State Electronics Letters",
  ["solid state ion"] = "Solid state ionics",
  ["solid state ionics"] = "Solid State Ionics",
  ["solid state nucl magn reson"] = "Solid State Nuclear Magnetic Resonance",
  ["solid state sci"] = "Solid State Sciences",
  ["solid-state electron"] = "Solid-State Electronics",
  ["solid–state electron"] = "Solid–State Electronics",
  ["solubility data ser"] = "Solubility Data Series",
  ["solutions (burlingt)"] = "Solutions (Burlington, Vt.)",
  ["solvent extr ion exch"] = "Solvent Extraction and Ion Exchange",
  ["solvent extr res dev, jpn"] = "Solvent Extraction Research and Development, Japan",
  ["somat cell mol genet"] = "Somatic cell and molecular genetics",
  ["somatic cell genet"] = "Somatic cell genetics",
  ["somatosens mot res"] = "Somatosensory & motor research",
  ["somatosens res"] = "Somatosensory research",
  ["somnologie (berl)"] = "Somnologie : Schlafforschung und Schlafmedizin = Somnology : sleep research and sleep medicine",
  ["sonderb strahlenther onkol"] = "Sonderbande zur Strahlentherapie und Onkologie",
  ["sonderb z strahlenther onkol"] = "Sonderband der Zeitschrift Strahlentherapie und Onkologie",
  ["sonderdr schr sudetdtsch akad wiss künste forschbeitr natwiss kl"] = "Sonderdruck aus Schriften der Sudetendeutschen Akademie der Wissenschaften und Küntse",
  ["songklanakarin journal / prince of songkla university, hat yai campus|warasan songkhla nakharin"] = "Warasan Songkhla Nakharin / Mahawitthayalai Songkhla Nakharin, Witthayakhet Hat Yai",
  ["soobshch akad nauk gruz ssr"] = "Soobshchenii︠a︡ Akademii nauk Gruzinskoĭ SSR (1963)",
  ["soobshch akad nauk gruzii"] = "Akademiya Nauk Gruzii",
  ["soobshch ob\\cdprime ed inst yadernykh issled dubna"] = "\\cyr Soobshcheniya Ob\\cdprime edinennogo Instituta Yadernykh Issledovaniy",
  ["soobshch prikl mat"] = "\\cyr Soobshcheniya po Prikladnoy Matematike",
  ["soochow j math"] = "Soochow Journal of Mathematics",
  ["sophia stud cross-cult philos tradit cult"] = "Sophia Studies in Cross-Cultural Philosophy of Traditions and Cultures",
  ["sort (barc)"] = "SORT (Barcelona)",
  ["sos laaketiet aikak"] = "Sosiaalilääketieteellinen Aikakauslehti",
  ["sotahist aikak"] = "Sotahistoriallinen aikakauskirja",
  ["sotilaslaak aikak"] = "Sotilaslääketieteellinen aikakauslehti",
  ["sots gig organ zdravookhr istor med"] = "Sotsial'naia gigiena, organizatsiia zdravookhraneniia i istoriia meditsiny",
  ["sotsiol issled"] = "Sot︠s︡iologicheskie issledovanii︠a︡",
  ["sotsiologicheskie issled"] = "Sotsiologicheskie Issledovaniya",
  ["soud lek"] = "Soudní lékarství",
  ["sound vib"] = "Sound & Vibration",
  ["source code biol med"] = "Source code for biology and medicine",
  ["source j genom"] = "Source journal of genomics",
  ["source j obstet gynaecol"] = "Source journal of obstetrics and gynaecology",
  ["source notes hist art"] = "Source (New York, N.Y. : 1981)",
  ["sources hist math phys sci"] = "Sources in the History of Mathematics and Physical Sciences",
  ["sources stud hist math phys sci"] = "Sources and Studies in the History of Mathematics and Physical Sciences",
  ["south afr j clin nutr"] = "The South African journal of clinical nutrition : SAJCN : the official journal of the South African Society of Parenteral and Enteral Nutrition and the Association for Dietetics in South Africa",
  ["south afr j demogr"] = "Southern African journal of demography = Suidelike Afrikaanse tydskrif vir demografie",
  ["south afr j epidemiol infect"] = "The Southern African journal of epidemiology & infection : official journal of the Sexually Transmitted Diseases, Infectious Diseases and Epidemiological Societies of Southern Africa",
  ["south afr j geol"] = "South African journal of geology : being the transactions of the Geological Society of South Africa = Suid-Afrikaanse tydskrif vir geologie : syndie die verhandelinge van die Geologiese Vereniging van Suid-Afrika",
  ["south afr j hiv med"] = "Southern African journal of HIV medicine",
  ["south afr polit econ mon"] = "Southern Africa political & economic monthly",
  ["south african j econ"] = "South African Journal of Economics",
  ["south african j econ manage sci"] = "South African Journal of Economic and Management Sciences, N.S.",
  ["south african statist j"] = "South African Statistical Journal. Suid-Afrikaanse Statistiese Tydskrif",
  ["south am indian stud"] = "South American Indian studies",
  ["south am j herpetol"] = "South american journal of herpetology",
  ["south amer j log"] = "South American Journal of Logic",
  ["south asia"] = "South Asia (Nedlands, W.A.)",
  ["south asia bull"] = "South Asia bulletin",
  ["south asia res"] = "South Asia research",
  ["south asian hist cult"] = "South Asian history and culture",
  ["south asian j cancer"] = "South Asian journal of cancer",
  ["south asian rev"] = "South Asian review",
  ["south atl q"] = "The South Atlantic quarterly",
  ["south calif interdiscip law j"] = "Southern California interdisciplinary law journal",
  ["south calif law rev"] = "Southern California law review",
  ["south calif q"] = "Southern California quarterly",
  ["south calif rev law womens stud"] = "Southern California review of law and women's studies",
  ["south clin istanb eurasia"] = "Southern clinics of Istanbul Eurasia",
  ["south commun j"] = "Southern communication journal",
  ["south east asia res"] = "South East Asia research",
  ["south east asian j math math sci"] = "South East Asian Journal of Mathematics and Mathematical Sciences",
  ["south east eur j public health"] = "South Eastern European journal of public health",
  ["south east report second ser"] = "South eastern reporter. Second series",
  ["south econ j"] = "Southern economic journal",
  ["south eur soc polit"] = "South European society & politics",
  ["south expos"] = "Southern exposure",
  ["south folk q"] = "Southern folklore quarterly",
  ["south for"] = "Southern forests",
  ["south for: j for sci"] = "Southern Forests: a Journal of Forest Science",
  ["south hist"] = "Southern history",
  ["south hosp"] = "Southern hospitals",
  ["south ill univ law j"] = "Southern Illinois University law journal. Southern Illinois University at Carbondale. School of Law",
  ["south indian stud"] = "Southern Indian studies",
  ["south j appl for"] = "Southern Journal of Applied Forestry",
  ["south j educ res"] = "The Southern journal of educational research",
  ["south j optom"] = "Southern journal of optometry",
  ["south j philos"] = "The Southern journal of philosophy",
  ["south med"] = "Southern medicine",
  ["south med j"] = "Southern medical journal",
  ["south med rev"] = "Southern med review",
  ["south med surg"] = "Southern medicine and surgery",
  ["south online j nurs res"] = "Southern online journal of nursing research",
  ["south pac bull"] = "South Pacific bulletin",
  ["south pac j nat appl sci"] = "South Pacific Journal of Natural and Applied Sciences",
  ["south report second ser cases argued determ courts ala fla la miss"] = "Southern reporter. Second series. Cases argued and determined in the courts of Alabama, Florida, Louisiana, Mississippi",
  ["south stud"] = "Southern studies",
  ["south surg"] = "The Southern surgeon",
  ["south tex law j"] = "South Texas law journal",
  ["south tex law rev"] = "South Texas law review",
  ["south univ law rev"] = "Southern University law review. Southern University and A & M College. School of Law",
  ["southampt med j"] = "Southampton medical journal",
  ["southeast archaeol"] = "Southeastern Archaeology",
  ["southeast asian bull math"] = "Southeast Asian Bulletin of Mathematics",
  ["southeast asian j sci"] = "Southeast Asian journal of sciences",
  ["southeast asian j soc sci"] = "Southeast Asian journal of social science",
  ["southeast asian j surg"] = "Southeast Asian journal of surgery",
  ["southeast asian j trop med public health"] = "The Southeast Asian journal of tropical medicine and public health",
  ["southeast eur"] = "Southeastern Europe. L'Europe du Sud-Est",
  ["southeast eur ser"] = "Southeast Europe series",
  ["southeast geogr"] = "Southeastern geographer",
  ["southeast geol"] = "Southeastern Geology",
  ["southeast nat"] = "Southeastern naturalist (Steuben, Me.)",
  ["southern econ j"] = "Southern Economic Journal",
  ["southern j agr econ"] = "Southern Journal of Agricultural Economics",
  ["southern quart"] = "Southern Quarterly",
  ["southwest china journal of agricultural sciences|xi nan nong ye xue bao"] = "Xi nan nong ye xue bao",
  ["southwest entomol"] = "The Southwestern entomologist",
  ["southwest entomolog"] = "Southwestern Entomologist",
  ["southwest hist q"] = "The Southwestern historical quarterly",
  ["southwest j anthropol"] = "Southwestern journal of anthropology",
  ["southwest j philos"] = "The Southwestern journal of philosophy",
  ["southwest j pulm crit care"] = "Southwest journal of pulmonary & critical care",
  ["southwest j pure appl math"] = "Southwest Journal of Pure and Applied Mathematics",
  ["southwest law j"] = "Southwestern law journal",
  ["southwest med"] = "Southwestern medicine",
  ["southwest nat"] = "The Southwestern naturalist",
  ["southwest report"] = "The Southwestern reporter",
  ["southwest rev"] = "Southwest review",
  ["southwest univ law rev"] = "Southwestern University law review",
  ["sov anthropol archeol"] = "Soviet Anthropology And Archeology",
  ["sov arkh"] = "Sovetskai︠a︡ arkhitektura (Kommunisticheskai︠a︡ akademii︠a︡ (Soviet Union). Sekt︠s︡ii︠a︡ sot︠s︡rasseleni︠a︡ i zhilishchno-bytogo stroitelʹstvo)",
  ["sov astron"] = "Soviet Astronomy [translation of Astronomicheskii Zhurnal]",
  ["sov econ"] = "Soviet economy (Silver Spring, Md.)",
  ["sov educ"] = "Soviet education",
  ["sov etnogr"] = "Sovetskai︠a︡ ėtnografii︠a︡",
  ["sov genet"] = "Soviet genetics",
  ["sov geogr"] = "Soviet geography",
  ["sov gos pravo"] = "Sovetskoe gosudarstvo i pravo",
  ["sov j at energy"] = "Soviet Journal of Atomic Energy [translation of Atomnaya Energiya]",
  ["sov j bioorg chem"] = "Soviet journal of bioorganic chemistry",
  ["sov j dev biol"] = "The Soviet journal of developmental biology",
  ["sov j ecol"] = "The Soviet journal of ecology",
  ["sov j low temp phys"] = "Soviet Journal of Low Temperature Physics [translation of Fizika Nizkikh Temperatur]",
  ["sov j nucl phys"] = "Soviet Journal of Nuclear Physics [translation of Yadernaya Fizika]",
  ["sov j opt technol"] = "Soviet Journal of Optical Technology [translation of Optiko–Mekhanicheskaya Promyshlennost]",
  ["sov j part nucl"] = "Soviet Journal of Particles and Nuclei [translation of Fizika Elementarnykh Chastitsi Atomnogo Yadra]",
  ["sov j plasma phys"] = "Soviet Journal of Plasma Physics [translation of Fizika Plazmy]",
  ["sov j quantum electron"] = "Soviet Journal of Quantum Electronics",
  ["sov jew aff"] = "Soviet Jewish affairs",
  ["sov law gov"] = "Soviet law and government",
  ["sov med"] = "Sovetskaia meditsina",
  ["sov phys - solid state"] = "Soviet Physics – Solid State",
  ["sov phys acoust"] = "Soviet Physics–Acoustics [translation of Akusticheskii Zhurnal]",
  ["sov phys crystallogr"] = "Soviet physics, crystallography",
  ["sov phys dokl"] = "Soviet Physics–Doklady [translation of Doklady Akademii Nauk SSSR]",
  ["sov phys j"] = "Soviet Physics Journal [translation of Izvestiya Vysshikh Uchebnykh Zavedenii, Fizika]",
  ["sov phys jetp"] = "Soviet Physics–JETP [translation of Zhurnal Eksperimentalnoi i Teoreticheskoi Fiziki]",
  ["sov phys semicond"] = "Soviet Physics–Semiconductors [translation of Fizika i Tekhnika Poluprovodnikov]",
  ["sov phys solid state"] = "Soviet Physics–Solid State [translation of Fizika Tverdogo Tela (Leningrad)]",
  ["sov phys tech phys"] = "Soviet Physics–Technical Physics [translation of Zhurnal Tekhnicheskoi Fiziki]",
  ["sov phys usp"] = "Soviet Physics–Uspekhi [translation of Uspekhi Fizicheskikh Nauk]",
  ["sov radiophys"] = "Soviet Radiophysics [translation of Izvestiya Vysshikh Uchebnykh Zavedenii, Radiofizika]",
  ["sov rev"] = "The Soviet review",
  ["sov slavanovednie"] = "Sovetskoe slavi︠a︡novedenie (Moscow, Russia : 1965)",
  ["sov soc"] = "Soviet Sociology",
  ["sov stud"] = "Soviet studies",
  ["sov stud hist"] = "Soviet studies in history",
  ["sov stud philos"] = "Soviet studies in philosophy",
  ["sov tech phys lett"] = "Soviet Technical Physics Letters [translation of Pisma v Zhurnal Tekhnicheskoi Fiziki]",
  ["sov union"] = "Soviet Union. Union soviétique",
  ["sov varchebnyii sb"] = "Sovetskiĭ vrachebnyĭ sbornik",
  ["sov zdravookhr"] = "Sovetskoe zdravookhranenie",
  ["sov zdravookhr kirg"] = "Sovetskoe zdravookhranenie Kirgizii",
  ["soviet e europ foreign trade"] = "Soviet and Eastern European Foreign Trade",
  ["soviet economy"] = "Soviet Economy",
  ["soviet sci rev sect a phys rev"] = "Soviet Scientific Reviews/Section A: Physics Reviews",
  ["soviet stud"] = "Soviet Studies",
  ["sovrem fiz-tekh probl"] = "Sovremennye Fiziko-Tekhnicheskie Problemy",
  ["sovrem gerpetol"] = "Sovremennai︠a︡ gerpetologii︠a︡",
  ["sovrem mat fundam napravl"] = "Sovremennaya Matematika. Fundamentalnye Napravleniya",
  ["sovrem mat—stud aspir"] = "\\cyr Sovremennaya Matematika—Studentam i Aspirantam",
  ["sovrem probl fiz"] = "\\cyr Sovremennye Problemy Fiziki",
  ["sovrem probl mat"] = "Sovremennye Problemy Matematiki",
  ["sovrem probl mat fund naprav"] = "\\cyr Sovremennye Problemy Matematiki",
  ["sovrem probl onkol"] = "Sovremennye problemy onkologii",
  ["sovrem probl tuberk"] = "Sovremennye problemy tuberkuleza",
  ["sovrem tekhnologii med"] = "Sovremennye tekhnologii v medit︠s︡ine",
  ["sowjetwiss naturwiss beitr"] = "Sowjetwissenschaft; Naturwissenschaftliche Beiträge",
  ["soyb genet newsl"] = "Soybean genetics newsletter",
  ["soz praventivmed"] = "Sozial- und Praventivmedizin",
  ["soz welt"] = "Soziale Welt : Zeitschrift fuer sozialwissenschaftliche Forschung und Praxis",
  ["soz wiss inf"] = "Sozialwissenschaftliche Informationen",
  ["sozial forstwirtsch"] = "sozialistische Forstwirtschaft, Die",
  ["sozial welt-zeit sozialwiss forsch praxis"] = "Soziale Welt-Zeitschrift für sozialwissenschaftliche Forschung Und Praxis",
  ["sozialgesch med"] = "Sozialgeschichte der Medizin",
  ["sozialmed padagog jugendkd"] = "Sozialmedizinische und pädagogische Jugendkunde",
  ["space commun"] = "Space communications",
  ["space conf expo (2016)"] = "SPACE Conferences and Exposition : AIAA SPACE 2016",
  ["space cult"] = "Space and culture : the journal",
  ["space life sci"] = "Space life sciences",
  ["space med med eng (beijing)"] = "Hang tian yi xue yu yi xue gong cheng = Space medicine & medical engineering",
  ["space policy"] = "Space policy",
  ["space polity"] = "Space & polity",
  ["space res today"] = "Space Research Today",
  ["space sci rev"] = "Space Science Reviews",
  ["space sci technol"] = "Space: Science & Technology",
  ["space technol"] = "Space technology (Oxford, England)",
  ["space technol libr"] = "Space Technology Library",
  ["space technol proc"] = "Space Technology Proceedings",
  ["space weather"] = "Space weather : the international journal of research & applications",
  ["space weather q"] = "Space Weather Quarterly",
  ["space world"] = "Space world",
  ["spacecraft satell"] = "Spacecraft and Satellites",
  ["spaceflight (lond)"] = "Spaceflight",
  ["spaces flows"] = "Spaces & flows : an international journal of urban and extraurban studies",
  ["spagna contemp"] = "Spagna contemporanea",
  ["span j agric res"] = "Spanish journal of agricultural research = Revista de investigación agraria",
  ["span j palaeontol"] = "Spanish Journal of Palaeontology",
  ["span j psychol"] = "The Spanish journal of psychology",
  ["span j soil sci"] = "Spanish Journal of Soil Science",
  ["spanish econ rev"] = "Spanish Economic Review",
  ["spartan med res j"] = "Spartan medical research journal",
  ["spas book ser"] = "SPAS Book Series",
  ["spastic quart"] = "Spastics' quarterly",
  ["spastic rev"] = "The Spastic review",
  ["spat cogn comput"] = "Spatial cognition and computation",
  ["spat demogr"] = "Spatial demography",
  ["spat econ anal"] = "Spatial economic analysis",
  ["spat spatiotemporal epidemiol"] = "Spatial and spatio-temporal epidemiology",
  ["spat stat"] = "Spatial statistics",
  ["spat vis"] = "Spatial vision",
  ["spatial cognit comput"] = "Spatial Cognition & Computation",
  ["spatial demogr"] = "Spatial Demography",
  ["spatial inf res"] = "Spatial Information Research",
  ["spatial res plann"] = "Spatial Research and Planning",
  ["spatial spatio-temporal epidemiol"] = "Spatial and Spatio-temporal Epidemiology",
  ["spatial stat"] = "Spatial Statistics",
  ["spatiotemporal image anal longitud time ser image data (2012)"] = "Spatio-temporal image analysis for longitudinal and time-series image data : Second International Workshop, STIA 2012, held in conjunction with MICCAI 2012, Nice, France, October 1, 2012, proceedings. STIA (Conference) (2nd : 2012 : Nice, France)",
  ["spatiotemporal image anal longitud time ser image data (2014)"] = "Spatio-temporal image analysis for longitudinal and time-series image data : third International Workshop, STIA 2014, held in conjunction with MICCAI 2014, Boston, MA, USA, September 18, 2014 : revised selected papers. STIA (Conference) (3rd : 2014 : Boston, Mass.)",
  ["spe drill complet"] = "SPE Drilling & Completion",
  ["spe form eval"] = "SPE Formation Evaluation",
  ["spe j"] = "SPE Journal",
  ["spe polym"] = "SPE Polymers",
  ["spe prod eng"] = "SPE Production Engineering",
  ["spe prod facil"] = "SPE Production & Facilities",
  ["spe prod oper"] = "SPE Production & Operations",
  ["spe reservoir eng"] = "SPE Reservoir Engineering",
  ["spe reservoir eval eng"] = "SPE Reservoir Evaluation & Engineering",
  ["spec bull jpn soc coleopterol"] = "Special bulletin of the Japanese Society of Coleopterology",
  ["spec care dentist"] = "Special care in dentistry : official publication of the American Association of Hospital Dentists, the Academy of Dentistry for the Handicapped, and the American Society for Geriatric Dentistry",
  ["spec collect"] = "Special collections",
  ["spec discuss faraday soc"] = "Special Discussions of the Faraday Society",
  ["spec educ"] = "Special education",
  ["spec educ forward trends"] = "Special education: forward trends",
  ["spec issue imvi"] = "Special Issue of IMVI",
  ["spec law dig health care (mon)"] = "Specialty law digest. Health care (Monthly)",
  ["spec law dig health care law"] = "Specialty law digest. Health care law",
  ["spec libr"] = "Special libraries",
  ["spec matrices"] = "Special Matrices",
  ["spec pap geol surv finl"] = "Special Paper of the Geological Survey of Finland",
  ["spec publ (eur aquac soc)"] = "Special publication (European Aquaculture Society)",
  ["spec publ - r soc chem"] = "Special Publication - Royal Society of Chemistry",
  ["spec publ am philos soc"] = "Special publication (American Philosophical Society)",
  ["spec publ tex tech univ mus"] = "Special publications (Texas Tech University. Museum)",
  ["spec rep ser indian counc med res"] = "Special report series - Indian Council of Medical Research",
  ["spec rep ser med res counc (g b)"] = "Special report series (Medical Research Council (Great Britain))",
  ["spec top endocrinol metab"] = "Special topics in endocrinology and metabolism",
  ["spec top rev porous media: int j"] = "Special Topics & Reviews in Porous Media: An International Journal",
  ["special oper j"] = "Special Operations Journal",
  ["species divers"] = "Species diversity : an international journal for taxonomy, systematics, speciation, biogeography, and life history research of animals",
  ["specif eng"] = "Specifying engineer",
  ["spectr shape anal med imaging (2016)"] = "Spectral and shape analysis in medical imaging : first International Workshop, SeSAMI 2016, held in conjunction with MICCAI 2016, Athens, Greece, October 21, 2016 : revised selected papers. SeSAMI (Workshop) (1st : 2016 : Athens, Greece)",
  ["spectr slovak"] = "Spectrum Slovakia",
  ["spectra anal"] = "Spectra Analyse",
  ["spectra biol"] = "Spectra Biologie",
  ["spectrochim acta"] = "Spectrochimica Acta",
  ["spectrochim acta [a]"] = "Spectrochimica acta. Part A: Molecular spectroscopy",
  ["spectrochim acta a"] = "Spectrochimica acta. Part A: Molecular spectroscopy",
  ["spectrochim acta a mol biomol spectrosc"] = "Spectrochimica acta. Part A, Molecular and biomolecular spectroscopy",
  ["spectrochim acta part b at spectrosc"] = "Spectrochimica acta. Part B, Atomic spectroscopy",
  ["spectrochim acta, part a"] = "Spectrochimica Acta Part A: Molecular and Biomolecular Spectroscopy",
  ["spectrochim acta, part b"] = "Spectrochimica Acta Part B: Atomic Spectroscopy",
  ["spectrosc asia"] = "Spectroscopy Asia",
  ["spectrosc eur"] = "Spectroscopy Europe",
  ["spectrosc lett"] = "Spectroscopy Letters",
  ["spectrosc prop inorg organomet compd"] = "Spectroscopic Properties of Inorganic and Organometallic Compounds",
  ["spectroscopy (amsterdam, neth)"] = "Spectroscopy (Amsterdam, Netherlands)",
  ["spectroscopy (springf)"] = "Spectroscopy (Springfield, Or.)",
  ["spectrum book"] = "A Spectrum Book",
  ["speculations sci technol"] = "Speculations in science and technology",
  ["speech commun"] = "Speech communication",
  ["speech lang hear"] = "Speech, language and hearing (London, England)",
  ["speech pathol ther"] = "Speech pathology and therapy",
  ["speech prosody"] = "Speech prosody (Urbana, Ill.)",
  ["spektrum augenheilkd"] = "Spektrum der Augenheilkunde : Zeitschrift der Österreichischen Ophthalmologischen Gesellschaft, ÖOG",
  ["spektrum lehrbuch"] = "Spektrum Lehrbuch",
  ["spektrum wiss"] = "Spektrum der Wissenschaft",
  ["spekulation erfahrung abt ii unters"] = "Spekulation und Erfahrung",
  ["speleobiology notes"] = "Speleobiology notes",
  ["spg biomed"] = "SPG biomed",
  ["spie milest ser"] = "SPIE Milestone Series",
  ["spie newsroom"] = "SPIE newsroom",
  ["spie proc"] = "Proceedings of SPIE",
  ["spie rev"] = "SPIE Reviews",
  ["spieg hist"] = "Spiegel historiael",
  ["spill sci technol bull"] = "Spill Science and Technology Bulletin",
  ["spinal cord"] = "Spinal cord",
  ["spinal cord ser cases"] = "Spinal cord series and cases",
  ["spine (phila pa 1976)"] = "Spine",
  ["spine (phila pa 1986)"] = "Spine (Philadelphia, Pa. : 1986)",
  ["spine deform"] = "Spine deformity",
  ["spine j"] = "The spine journal : official journal of the North American Spine Society",
  ["spine surg relat res"] = "Spine surgery and related research",
  ["spiritual clin pract (wash d c )"] = "Spirituality in clinical practice (Washington, D.C.)",
  ["spis blg akad nauk"] = "Spisanie na Bŭlgarskata akademii︠a︡ na naukite. Bŭlgarska akademii︠a︡ na naukite",
  ["spisy ped fak ostrave"] = "Spisy Pedagogicke Fakulty v Ostrave",
  ["splice life"] = "Splice : the splice of life",
  ["sport educ soc"] = "Sport, education and society",
  ["sport ethics philos"] = "Sport, ethics and philosophy",
  ["sport exerc perform psychol"] = "Sport, exercise, and performance psychology",
  ["sport exercise perform psychol"] = "Sport Exercise and Performance Psychology",
  ["sport hist rev"] = "Sport history review",
  ["sport j"] = "The sport journal",
  ["sport psychol"] = "The Sport psychologist",
  ["sport sci health"] = "Sport sciences for health",
  ["sport soc"] = "Sport in society",
  ["sportarzt ver sportmed"] = "Der Sportarzt vereinigt mit Sportmedizin",
  ["sports (basel)"] = "Sports (Basel, Switzerland)",
  ["sports biomech"] = "Sports biomechanics",
  ["sports eng"] = "Sports Engineering",
  ["sports health"] = "Sports health",
  ["sports hist"] = "The sports historian",
  ["sports med"] = "Sports medicine (Auckland, N.Z.)",
  ["sports med - open"] = "Sports Medicine - Open",
  ["sports med arthrosc"] = "Sports medicine and arthroscopy review",
  ["sports med arthrosc rehabil ther technol"] = "Sports medicine, arthroscopy, rehabilitation, therapy & technology : SMARTT",
  ["sports med arthrosc rev"] = "Sports medicine and arthroscopy review",
  ["sports med health sci"] = "Sports Medicine and Health Science",
  ["sports med int open"] = "Sports medicine international open",
  ["sports med open"] = "Sports medicine - open",
  ["sportverletz sportschaden"] = "Sportverletzung Sportschaden : Organ der Gesellschaft für Orthopädisch-Traumatologische Sportmedizin",
  ["spraw czynnosci posiedz lodz tow nauk"] = "Sprawozdania z czynności i posiedzeń - Lódzkie Towarzystwo Naukowe",
  ["springer actuar"] = "Springer Actuarial",
  ["springer actuar textb"] = "Springer Actuarial Textbooks",
  ["springer aerosp technol"] = "Springer Aerospace Technology",
  ["springer biogr"] = "Springer Biographies",
  ["springer briefs bus"] = "Springer Briefs in Business",
  ["springer briefs environ sci"] = "Springer Briefs in Environmental Science",
  ["springer briefs oper res"] = "Springer Briefs in Operations Research",
  ["springer briefs water sci technol"] = "Springer Briefs in Water Science and Technology",
  ["springer clim"] = "Springer Climate",
  ["springer collect works math"] = "Springer Collected Works in Mathematics",
  ["springer finance"] = "Springer Finance",
  ["springer finance lect notes"] = "Springer Finance Lecture Notes",
  ["springer finance textb"] = "Springer Finance Textbooks",
  ["springer geol"] = "Springer Geology",
  ["springer geophys"] = "Springer Geophysics",
  ["springer grad texts philos"] = "Springer Graduate Texts in Philosophy",
  ["springer handb auditory res"] = "Springer handbook of auditory research",
  ["springer handb comput stat"] = "Springer Handbooks of Computational Statistics",
  ["springer indam ser"] = "Springer INdAM Series",
  ["springer int handb educ"] = "Springer International Handbooks of Education",
  ["springer lehrbuch math"] = "Springer Lehrbuch Mathematik",
  ["springer monogr math"] = "Springer Monographs in Mathematics",
  ["springer optim appl"] = "Springer Optimization and Its Applications",
  ["springer proc adv robot"] = "Springer Proceedings in Advanced Robotics",
  ["springer proc bus econ"] = "Springer Proceedings in Business and Economics",
  ["springer proc complexity"] = "Springer Proceedings in Complexity",
  ["springer proc earth environ sci"] = "Springer Proceedings in Earth and Environmental Sciences",
  ["springer proc math"] = "Springer Proceedings in Mathematics",
  ["springer proc math stat"] = "Springer Proceedings in Mathematics & Statistics",
  ["springer proc phys"] = "Springer Proceedings in Physics",
  ["springer sci rev"] = "Springer science reviews",
  ["springer semin immunopathol"] = "Springer seminars in immunopathology",
  ["springer ser adv manuf"] = "Springer Series in Advanced Manufacturing",
  ["springer ser astrophys cosmol"] = "Springer Series in Astrophysics and Cosmology",
  ["springer ser astrostatistics"] = "Springer Series in Astrostatistics",
  ["springer ser at opt plasma phys"] = "Springer Series on Atomic, Optical, and Plasma Physics",
  ["springer ser bio-/neuroinform"] = "Springer Series in Bio-/Neuroinformatics",
  ["springer ser biophys"] = "Springer series in biophysics",
  ["springer ser chem phys"] = "Springer Series in Chemical Physics",
  ["springer ser chem sens biosens"] = "Springer Series on Chemical Sensors and Biosensors",
  ["springer ser comput math"] = "Springer Series in Computational Mathematics",
  ["springer ser cult comput"] = "Springer Series on Cultural Computing",
  ["springer ser data sci"] = "Springer Series in the Data Sciences",
  ["springer ser demogr methods popul anal"] = "The Springer Series on Demographic Methods and Population Analysis",
  ["springer ser discrete math theor comput sci"] = "Springer Series in Discrete Mathematics and Theoretical Computer Science",
  ["springer ser game theory"] = "Springer Series in Game Theory",
  ["springer ser health care soc"] = "Springer series on health care and society",
  ["springer ser inform sci"] = "Springer Series in Information Sciences",
  ["springer ser light scatt"] = "Springer Series in Light Scattering",
  ["springer ser mater sci"] = "Springer Series in Materials Science",
  ["springer ser meas sci technol"] = "Springer Series in Measurement Science and Technology",
  ["springer ser nav archit mar eng shipbuild shipp"] = "Springer Series on Naval Architecture, Marine Engineering, Shipbuilding and Shipping",
  ["springer ser nonlinear dynam"] = "Springer Series in Nonlinear Dynamics",
  ["springer ser oper res"] = "Springer Series in Operations Research",
  ["springer ser oper res financ eng"] = "Springer Series in Operations Research and Financial Engineering",
  ["springer ser opt sci"] = "Springer Series in Opical Sciences",
  ["springer ser optical sci"] = "Springer Series in Optical Sciences",
  ["springer ser pharm stat"] = "Springer Series in Pharmaceutical Statistics",
  ["springer ser plasma sci technol"] = "Springer Series in Plasma Science and Technology",
  ["springer ser reliab eng"] = "Springer Series in Reliability Engineering",
  ["springer ser solid struct mech"] = "Springer Series in Solid and Structural Mechanics",
  ["springer ser solid-state sci"] = "Springer Series in Solid-State Sciences",
  ["springer ser statist"] = "Springer Series in Statistics",
  ["springer ser statist perspect statist"] = "Springer Series in Statistics. Perspectives in Statistics",
  ["springer ser statist statist health sci"] = "Springer Series in Statistics: Statistics in the Health Sciences",
  ["springer ser supply chain manag"] = "Springer Series in Supply Chain Management",
  ["springer ser surface sci"] = "Springer Series in Surface Sciences",
  ["springer ser synergetics"] = "Springer Series in Synergetics",
  ["springer ser wave phenomena"] = "Springer Series on Wave Phenomena",
  ["springer stud hist econ thought"] = "Springer Studies in the History of Economic Thought",
  ["springer stud math master"] = "Springer Studium Mathematik—Master",
  ["springer textb earth sci geogr environ"] = "Springer Textbooks in Earth Sciences, Geography and Environment",
  ["springer texts bus econ"] = "Springer Texts in Business and Economics",
  ["springer texts soc sci"] = "Springer Texts in Social Sciences",
  ["springer texts statist"] = "Springer Texts in Statistics",
  ["springer top signal process"] = "Springer Topics in Signal Processing",
  ["springer tracts adv robot"] = "Springer Tracts in Advanced Robotics",
  ["springer tracts civ eng"] = "Springer Tracts in Civil Engineering",
  ["springer tracts mech eng"] = "Springer Tracts in Mechanical Engineering",
  ["springer tracts modern phys"] = "Springer Tracts in Modern Physics",
  ["springer tracts nat philos"] = "Springer Tracts in Natural Philosophy",
  ["springer tracts nat-inspir comput"] = "Springer Tracts in Nature-Inspired Computing",
  ["springer tracts transp traffic"] = "Springer Tracts on Transportation and Traffic",
  ["springer trans civil environ eng"] = "Springer Transactions in Civil and Environmental Engineering",
  ["springer uncertain res"] = "Springer Uncertainty Research",
  ["springer undergrad math ser"] = "Springer Undergraduate Mathematics Series",
  ["springer undergrad textb stat"] = "Springer Undergraduate Textbooks in Statistics",
  ["springer undergrad texts math technol"] = "Springer Undergraduate Texts in Mathematics and Technology",
  ["springer undergrad texts philos"] = "Springer Undergraduate Texts in Philosophy",
  ["springer-lehrb mastercl"] = "Springer-Lehrbuch Masterclass",
  ["springer-praxis books astron planet sci"] = "Springer-Praxis Books in Astronomy and Planetary Sciences",
  ["springer-praxis books environ sci"] = "Springer-Praxis Books in Environmental Sciences",
  ["springer-praxis books pop sci"] = "Springer-Praxis Books. Popular Science",
  ["springerbriefs adv inf knowl process"] = "SpringerBriefs in Advanced Information and Knowledge Processing",
  ["springerbriefs appl sci technol"] = "SpringerBriefs in Applied Sciences and Technology",
  ["springerbriefs appl stat econom"] = "SpringerBriefs in Applied Statistics and Econometrics",
  ["springerbriefs astron"] = "SpringerBriefs in Astronomy",
  ["springerbriefs bioeng"] = "SpringerBriefs in Bioengineering",
  ["springerbriefs cogn math"] = "SpringerBriefs in Cognitive Mathematics",
  ["springerbriefs complex"] = "SpringerBriefs in Complexity",
  ["springerbriefs comput intell"] = "SpringerBriefs in Computational Intelligence",
  ["springerbriefs comput mech"] = "SpringerBriefs in Computational Mechanics",
  ["springerbriefs comput sci"] = "SpringerBriefs in Computer Science",
  ["springerbriefs contin mech"] = "SpringerBriefs in Continuum Mechanics",
  ["springerbriefs control autom robot"] = "SpringerBriefs in Control, Automation and Robotics",
  ["springerbriefs earth sci"] = "SpringerBriefs in Earth Sciences",
  ["springerbriefs earth syst sci"] = "SpringerBriefs in Earth System Sciences",
  ["springerbriefs ecol"] = "SpringerBriefs in Ecology",
  ["springerbriefs econ"] = "SpringerBriefs in Economics",
  ["springerbriefs electr comput eng"] = "SpringerBriefs in Electrical and Computer Engineering",
  ["springerbriefs energy"] = "SpringerBriefs in Energy",
  ["springerbriefs finance"] = "SpringerBriefs in Finance",
  ["springerbriefs geogr"] = "SpringerBriefs in Geography",
  ["springerbriefs hist sci technol"] = "SpringerBriefs in History of Science and Technology",
  ["springerbriefs inf secur cryptogr"] = "SpringerBriefs in Information Security and Cryptography",
  ["springerbriefs intell syst"] = "SpringerBriefs in Intelligent Systems",
  ["springerbriefs math"] = "SpringerBriefs in Mathematics",
  ["springerbriefs math mater"] = "SpringerBriefs in the Mathematics of Materials",
  ["springerbriefs math methods"] = "SpringerBriefs in Mathematical Methods",
  ["springerbriefs math phys"] = "SpringerBriefs in Mathematical Physics",
  ["springerbriefs math planet earth"] = "SpringerBriefs in Mathematics of Planet Earth",
  ["springerbriefs mol sci"] = "SpringerBriefs in Molecular Science",
  ["springerbriefs nonlinear circuits"] = "SpringerBriefs in Nonlinear Circuits",
  ["springerbriefs optim"] = "SpringerBriefs in Optimization",
  ["springerbriefs pdes data sci"] = "SpringerBriefs on PDEs and Data Science",
  ["springerbriefs philos"] = "SpringerBriefs in Philosophy",
  ["springerbriefs phys"] = "SpringerBriefs in Physics",
  ["springerbriefs popul stud"] = "SpringerBriefs in Population Studies",
  ["springerbriefs probab math stat"] = "SpringerBriefs in Probability and Mathematical Statistics",
  ["springerbriefs quant finance"] = "SpringerBriefs in Quantitative Finance",
  ["springerbriefs signal process"] = "SpringerBriefs in Signal Processing",
  ["springerbriefs stat"] = "SpringerBriefs in Statistics",
  ["springerbriefs stat abe"] = "SpringerBriefs in Statistics—ABE",
  ["springerbriefs struct mech"] = "SpringerBriefs in Structural Mechanics",
  ["springerbriefs therm eng appl sci"] = "SpringerBriefs in Thermal Engineering and Applied Science",
  ["springers kurzlehrb wirtsch"] = "Springers Kurzlehrbucher der Wirtschaftswissenschaften",
  ["spums j"] = "SPUMS journal",
  ["squibb memo"] = "Squibb memoranda",
  ["sr care pharm"] = "The Senior care pharmacist",
  ["sr nurse"] = "Senior nurse",
  ["sr scholast"] = "Senior scholastic",
  ["sr scholast (teach ed)"] = "Senior scholastic (Teacher[s] ed.)",
  ["sra j"] = "SRA journal",
  ["sri lanka j soc sci"] = "Sri Lanka journal of social sciences",
  ["sri lanka popul dig"] = "Sri Lanka population digest",
  ["sri lankan j appl stat"] = "Sri Lankan journal of applied statistics",
  ["srl neurol neurosurg"] = "SRL neurology & neurosurgery",
  ["srp arh celok lek"] = "Srpski arhiv za celokupno lekarstvo",
  ["srs bull"] = "SRS bulletin",
  ["srx pharmacol"] = "SRX pharmacology",
  ["ssm popul health"] = "SSM - population health",
  ["sso schweiz monatsschr zahnheilkd"] = "Schweizerische Monatsschrift für Zahnheilkunde = Revue mensuelle suisse d'odonto-stomatologie",
  ["st anthony messenger"] = "Saint Anthony messenger; a national Catholic family magazine",
  ["st barnabas hosp med bull"] = "St. Barnabas Hospital medical bulletin",
  ["st bartholomews hosp j"] = "St. Bartholomew's Hospital journal",
  ["st georg hosp gaz"] = "The Saint George's Hospital gazette",
  ["st johns law rev"] = "St. John's law review",
  ["st louis u j health law policy"] = "Saint Louis University journal of health law & policy",
  ["st louis univ law j"] = "Saint Louis University law journal",
  ["st louis univ public law rev"] = "Saint Louis University public law review",
  ["st luc med"] = "Saint-Luc médical. Sint-Lucas tijdschrift",
  ["st lukes hosp gaz (guardamangia)"] = "The St. Luke's Hospital gazette",
  ["st lukes hosp staff clin bull"] = "St. Luke's Hospital staff clinics bulletin",
  ["st lukes j theol"] = "St. Luke's journal of theology",
  ["st marys hosp gaz"] = "St. Mary's Hospital gazette",
  ["st marys law j"] = "St. Mary's law journal",
  ["st petersburg math j"] = "St. Petersburg Mathematical Journal",
  ["st petersburg polytech univ j: phys math"] = "St. Petersburg Polytechnical University Journal: Physics and Mathematics",
  ["st thomas hosp gaz"] = "St. Thomas's Hospital gazette",
  ["st thomas law rev"] = "St. Thomas law review",
  ["st tomas j med"] = "Santo Tomas journal of medicine",
  ["st tomas nurs j"] = "Santo Tomas nursing journal",
  ["st vincents hosp med bull"] = "St. Vincent's Hospital medical bulletin",
  ["st vladmirs theol q"] = "St. Vladimir's theological quarterly",
  ["stab oscil optim syst"] = "Stability, Oscillations and Optimization of Systems",
  ["stability control theory methods appl"] = "Stability and Control: Theory, Methods and Applications",
  ["stadion (koln)"] = "Stadion (Cologne, Germany)",
  ["staff bull easton hosp (easton, pa)"] = "Staff bulletin. Easton, Pa. Hospital",
  ["staff pap int monet fund"] = "Staff papers - International Monetary Fund. International Monetary Fund",
  ["staff proc"] = "Staff proceedings. Hillcrest Medical Center (Tulsa, Okla.)",
  ["stain technol"] = "Stain technology",
  ["stal, sci tech anim lab"] = "STAL, Sciences et Techniques de l’Animal de Laboratoire",
  ["stammering res"] = "Stammering research : an on-line journal published by the British Stammering Association",
  ["stand genomic sci"] = "Standards in genomic sciences",
  ["stand news"] = "Standardization news : SN",
  ["stanf environ law j"] = "Stanford environmental law journal",
  ["stanf j law sci policy"] = "Stanford journal of law, science & policy",
  ["stanf technol law rev"] = "Stanford technology law review : STLR : an online high-technology law journal from Stanford Law School",
  ["stanf text technol"] = "Stanford Text Technologies",
  ["stanford fr rev"] = "Stanford French review",
  ["stanford j int law"] = "Stanford journal of international law",
  ["stanford law pol rev"] = "Stanford law & policy review",
  ["stanford law rev"] = "Stanford law review",
  ["stanford lectures"] = "Stanford Lectures",
  ["stanford med bull"] = "Stanford medical bulletin",
  ["stapp car crash j"] = "Stapp car crash journal",
  ["star protoc"] = "STAR protocols",
  ["star: sci technol archaeol res"] = "STAR: Science & Technology of Archaeological Research",
  ["starch - starke"] = "Starch - Starke",
  ["stat (int stat inst)"] = "Stat (International Statistical Institute)",
  ["stat anal data min"] = "Statistical Analysis and Data Mining",
  ["stat appl"] = "Statistics and applications",
  ["stat appl genet mol biol"] = "Statistical applications in genetics and molecular biology",
  ["stat atlases comput models heart"] = "Statistical atlases and computational models of the heart. STACOM (Workshop)",
  ["stat biol health"] = "Statistics for Biology and Health",
  ["stat biopharm res"] = "Statistics in biopharmaceutical research",
  ["stat biosci"] = "Statistics in biosciences",
  ["stat bull czechoslov"] = "Statistical bulletin of Czechoslovakia. Czechoslovak Republic. Státni Urad Statistický",
  ["stat bull metrop insur co"] = "Statistical bulletin (Metropolitan Life Insurance Company : 1984)",
  ["stat bull metrop life found"] = "Statistical bulletin (Metropolitan Life Foundation)",
  ["stat bull metropol life insur co"] = "Statistical bulletin (Metropolitan Life Insurance Company)",
  ["stat commun infec dis"] = "Statistical Communications in Infectious Diseases",
  ["stat commun infect dis"] = "Statistical communications in infectious diseases",
  ["stat comput"] = "Statistics and Computing",
  ["stat eng phys sci"] = "Statistics for Engineering and Physical Science",
  ["stat etud midi pyren"] = "Statistiques & études: Midi-Pyrénees",
  ["stat ind technol"] = "Statistics for Industry and Technology",
  ["stat ind technol eng"] = "Statistics for Industry, Technology, and Engineering",
  ["stat inf bull afr"] = "Statistical information bulletin for Africa = Bulletin d'information statistique pour l'Afrique",
  ["stat inference stoch process"] = "Statistical Inference for Stochastic Processes. An International Journal Devoted to Time Series Analysis and the Statistics of Continuous Time Processes and Dynamical Systems",
  ["stat inference stochastic processes"] = "Statistical Inference for Stochastic Processes",
  ["stat interface"] = "Statistics and its interface",
  ["stat j iaos"] = "Statistical journal of the IAOS",
  ["stat j un econ comm eur"] = "Statistical journal of the United Nations Economic Commission for Europe",
  ["stat jahrb schweiz"] = "Statistisches Jahrbuch der Schweiz",
  ["stat math probab"] = "Statistique Mathématique et Probabilité",
  ["stat med"] = "Statistics in medicine",
  ["stat meth appl"] = "Statistical Methods & Applications",
  ["stat methodol"] = "Statistical Methodology",
  ["stat methods appl"] = "Statistical Methods & Applications. Journal of the Italian Statistical Society",
  ["stat methods appt"] = "Statistical methods & applications",
  ["stat methods med res"] = "Statistical methods in medical research",
  ["stat model"] = "Statistical Modelling. An International Journal",
  ["stat modell"] = "Statistical Modelling",
  ["stat modelling"] = "Statistical modelling",
  ["stat nachr osterr stat zent amt"] = "Statistische Nachrichten. Österreichisches Statistisches Zentralamt",
  ["stat navy med"] = "Statistics of Navy medicine",
  ["stat neerl"] = "Statistica Neerlandica",
  ["stat news pol"] = "Statistical news. Poland. Główny Urzad Statystyczny",
  ["stat notes health plann"] = "Statistical notes for health planners",
  ["stat notes jpn"] = "Statistical Notes Of Japan",
  ["stat optim inf comput"] = "Statistics, Optimization & Information Computing",
  ["stat pap"] = "Statistical Papers",
  ["stat pap (berl)"] = "Statistical papers (Berlin, Germany)",
  ["stat politics policy"] = "Statistics, politics, and policy",
  ["stat probab lett"] = "Statistics & probability letters",
  ["stat public policy (phila)"] = "Statistics and public policy (Philadelphia, Pa.)",
  ["stat report"] = "Statistical Reporter / Office Of Federal Statistical Policy And Standards",
  ["stat risk model"] = "Statistics & Risk Modeling with Applications in Finance and Insurance",
  ["stat sci"] = "Statistical science : a review journal of the Institute of Mathematical Statistics",
  ["stat sci interdiscip res"] = "Statistical Science and Interdisciplinary Research",
  ["stat sin"] = "Statistica Sinica",
  ["stat soc behav sci"] = "Statistics for Social and Behavioral Sciences",
  ["stat soc behav sci ser"] = "Statistics in the Social and Behavioral Sciences Series",
  ["stat surv"] = "Statistics Surveys",
  ["stat szle"] = "Statisztikai szemle : a Magyar Központi Statisztikai Hivatal folyóirata",
  ["stat textb monogr"] = "Statistics: Textbooks and Monographs",
  ["stat theory relat fields"] = "Statistical Theory and Related Fields",
  ["stat tidskr"] = "Statistisk tidskrift. Statistical review. Sweden. Statistiska centralbyrån",
  ["stat zprav"] = "Statistický zpravodaj. Czechoslovak Republic. Státni Urad Statistický",
  ["stata j"] = "The Stata journal",
  ["state art rep"] = "State of the Art Reports",
  ["state arts ser"] = "State-of-the-arts Series",
  ["state black am"] = "The state of Black America. National Urban League",
  ["state coverage initiat issue brief"] = "State Coverage Initiatives issue brief : a national initiative of the Robert Wood Johnson Foundation",
  ["state demogr"] = "Statistiques démographiques. Institut national de statistique (Belgium)",
  ["state gov"] = "State government (Denver, Colo.)",
  ["state health care am"] = "The state of health care in America",
  ["state health legis rep"] = "State health legislation report",
  ["state legis"] = "State legislatures",
  ["state med soc annual rep"] = "Annual report. State Medical Society of Wisconsin",
  ["state nurs legis q"] = "State nursing legislation quarterly",
  ["state polit policy q"] = "State politics & policy quarterly",
  ["state reprod health monit"] = "State reproductive health monitor : legislative proposals and actions",
  ["states health"] = "States of health",
  ["static dyn game theory found appl"] = "Static & Dynamic Game Theory: Foundations & Applications",
  ["statist comput"] = "Statistics and Computing",
  ["statist decisions"] = "Statistics & Decisions",
  ["statist decisions econom"] = "Statistique et Decisions Economiques",
  ["statist ecology ser"] = "Statistical Ecology Series",
  ["statist efterret"] = "Statistiske efterretninger. Befolkning og valg",
  ["statist health sci"] = "Statistics in the Health Sciences",
  ["statist medd ser be befolk val"] = "Statistiska meddelanden. Serie Be",
  ["statist metody ekonom issled"] = "Statisticheskie Metody v Ekonomicheskikh Issledovaniyakh",
  ["statist model decis sci"] = "Statistical Modeling and Decision Science",
  ["statist neerlandica"] = "Statistica Neerlandica",
  ["statist papers"] = "Statistical Papers",
  ["statist practice"] = "Statistics in Practice",
  ["statist probab lett"] = "Statistics & Probability Letters",
  ["statist sci"] = "Statistical Science. A Review Journal of the Institute of Mathematical Statistics",
  ["statist sinica"] = "Statistica Sinica",
  ["statist textbooks monogr"] = "Statistics: Textbooks and Monographs",
  ["statistica (bologna)"] = "Statistica",
  ["statistical j"] = "Statistical Journal",
  ["statistical pap"] = "Statistical Papers",
  ["statistics (ber)"] = "Statistics",
  ["stats (basel)"] = "Stats",
  ["statsokon tidsskr"] = "Statsokonomisk Tidsskrift",
  ["statut calif dig meas calif"] = "Statutes of California and digests of measures. California",
  ["staub - reinhalt luft"] = "Staub - Reinhaltung der Luft",
  ["std bull"] = "STD bulletin",
  ["std news"] = "STD news : a quarterly newsletter of the American Social Health Association",
  ["steam-h: sci technol eng agric math health"] = "STEAM-H: Science, Technology, Engineering, Agriculture, Mathematics & Health",
  ["steel compos struct"] = "Steel and Composite Structures",
  ["steel constr des res"] = "Steel Construction - Design and Research",
  ["steel res"] = "Steel Research",
  ["steel res int"] = "Steel Research International",
  ["steel transl"] = "Steel in Translation",
  ["steklo keram"] = "Steklo i Keramika",
  ["stem cell adv res ther"] = "Stem cell: advanced research and therapy",
  ["stem cell epigenet"] = "Stem cell epigenetics",
  ["stem cell investig"] = "Stem cell investigation",
  ["stem cell rep"] = "Stem Cell Reports",
  ["stem cell reports"] = "Stem cell reports",
  ["stem cell res"] = "Stem cell research",
  ["stem cell res ther"] = "Stem cell research & therapy",
  ["stem cell res ther (walnut)"] = "Stem cell research & therapeutics",
  ["stem cell rev"] = "Stem cell reviews",
  ["stem cell rev rep"] = "Stem cell reviews and reports",
  ["stem cell stud"] = "Stem Cell Studies",
  ["stem cell transl investig"] = "Stem cell and translational investigation",
  ["stem cells"] = "Stem cells (Dayton, Ohio)",
  ["stem cells cloning"] = "Stem cells and cloning : advances and applications",
  ["stem cells cloning: adv appl"] = "Stem Cells and Cloning: Advances and Applications",
  ["stem cells dev"] = "Stem cells and development",
  ["stem cells int"] = "Stem cells international",
  ["stem cells transl med"] = "Stem cells translational medicine",
  ["stem educ"] = "STEM Education",
  ["stem fellowship j"] = "STEM Fellowship Journal",
  ["step perspect"] = "STEP perspective",
  ["stephan mueller spec publ ser"] = "Stephan Mueller Special Publication Series",
  ["stereotact funct neurosurg"] = "Stereotactic and functional neurosurgery",
  ["sterile world"] = "Sterile world",
  ["sterne weltraum"] = "Sterne und Weltraum",
  ["steroids lipids res"] = "Steroids and lipids research",
  ["stetson law rev"] = "Stetson law review",
  ["stift reusstal, jahresber"] = "Stiftung Reusstal, Jahresbericht",
  ["stigma health"] = "Stigma and health",
  ["stigma res action"] = "Stigma research and action",
  ["stimmen zeit"] = "Stimmen der Zeit",
  ["stns j trauma nurs"] = "STN's journal of trauma nursing : the official journal of the Society of Trauma Nurses",
  ["stoch anal appl"] = "Stochastic analysis and applications",
  ["stoch dyn"] = "Stochastics and Dynamics",
  ["stoch environ res risk assess"] = "Stochastic environmental research and risk assessment : research journal",
  ["stoch model"] = "Stochastic models",
  ["stoch model appl probab"] = "Stochastic Modelling and Applied Probability",
  ["stoch models"] = "Stochastic Models",
  ["stoch models surviv anal reliab set"] = "Stochastic Models in Survival Analysis and Reliability Set",
  ["stoch partial differ equ"] = "Stochastic partial differential equations : analysis and computations",
  ["stoch partial differ equ anal comput"] = "Stochastic Partial Differential Equations. Analysis and Computations",
  ["stoch process their appl"] = "Stochastic processes and their applications",
  ["stoch qual control"] = "Stochastics and Quality Control",
  ["stoch syst"] = "Stochastic Systems",
  ["stochastic anal appl"] = "Stochastic Analysis and Applications",
  ["stochastic environ res risk assess"] = "Stochastic Environmental Research and Risk Assessment",
  ["stochastic model"] = "Stochastic Modeling",
  ["stochastic model ser"] = "Stochastic Modeling Series",
  ["stochastic models"] = "Stochastic Models",
  ["stochastic process appl"] = "Stochastic Processes and their Applications",
  ["stochastic processes appl"] = "Stochastic Processes and their Applications",
  ["stochastic syst"] = "Stochastic Systems",
  ["stochastics (abingdon)"] = "Stochastics (Abingdon, England : 2005)",
  ["stochastics dyn"] = "Stochastics and Dynamics",
  ["stochastics monogr"] = "Stochastics Monographs",
  ["stochastics partial differ equations-anal comput"] = "Stochastics and Partial Differential Equations-Analysis and Computations",
  ["stochastics stochastics rep"] = "Stochastics and Stochastics Reports",
  ["stockh stud engl"] = "Acta Universitatis Stockholmiensis. Stockholm studies in English",
  ["stoma (heidelb)"] = "Stoma",
  ["stoma (lisb)"] = "Stoma (Lisbon, Portugal)",
  ["stoma (thessaloniki)"] = "Stoma (Thessalonikē, Greece)",
  ["stomatol chron (athenai)"] = "Stomatologika chronika. Acta stomatologica Hellenica",
  ["stomatol ddr"] = "Stomatologie der DDR",
  ["stomatol glas srb"] = "Stomatolos̆ki glasnik Srbije",
  ["stomatol mediterr"] = "Stomatologia mediterranea : SM",
  ["stomatol vestn"] = "Stomatologický vĕstník",
  ["stomatol vjesn"] = "Stomatolos̆ki vjesnik. Stomatological review",
  ["stomatol zpr"] = "Stomatologické zprávy",
  ["stomatologia (athenai)"] = "Stomatologia",
  ["stomatologia (bucur)"] = "Stomatologia",
  ["stomatologica (genova)"] = "Stomatologica",
  ["stomatologiia (mosk)"] = "Stomatologii︠a︡",
  ["stomatologiia (sofiia)"] = "Stomatologiia. Stomatology",
  ["stor contemp"] = "Storia contemporanea",
  ["stor crit psicol"] = "Storia e critica della psicologia",
  ["stor lomb"] = "Storia in Lombardia",
  ["stor med pop"] = "Storia e medicina popolare",
  ["stor urbana"] = "Storia urbana",
  ["strahlenschutz forsch prax"] = "Strahlenschutz in Forschung und Praxis",
  ["strahlenther onkol"] = "Strahlentherapie und Onkologie",
  ["strahlentherapie [sonderb]"] = "Strahlentherapie. Sonderbande",
  ["strahlentherapie sonderb"] = "Strahlentherapie. Sonderbände",
  ["strasb med"] = "Strasbourg médical",
  ["strateg healthc excell"] = "Strategies for healthcare excellence : organizational productivity, quality and effectiveness",
  ["strateg mol biol"] = "Strategies in molecular biology",
  ["strateg organ"] = "Strategic organization",
  ["strategic plann energy environ"] = "Strategic Planning for Energy and the Environment",
  ["strategies trauma limb reconstr"] = "Strategies in trauma and limb reconstruction",
  ["strategy tactics"] = "Strategy & tactics (Cambria, Calif.)",
  ["stratigr geol correl"] = "Stratigraphy and Geological Correlation",
  ["strength cond j"] = "Strength and conditioning journal",
  ["strength fract complex"] = "Strength, fracture and complexity",
  ["strength fract complexity"] = "Strength, Fracture and Complexity",
  ["strength mater"] = "Strength of Materials",
  ["strenna stor bolognese"] = "Strenna storica bolognese",
  ["stress biol"] = "Stress Biology",
  ["stress health"] = "Stress and health : journal of the International Society for the Investigation of Stress",
  ["stress med"] = "Stress medicine",
  ["stroke res treat"] = "Stroke research and treatment",
  ["stroke vasc neurol"] = "Stroke and vascular neurology",
  ["stroke: vasc interventional neurol"] = "Stroke: Vascular and Interventional Neurology",
  ["struct bond"] = "Structure and Bonding",
  ["struct change econ dynam"] = "Structural Change and Economic Dynamics",
  ["struct chem"] = "Structural Chemistry",
  ["struct concr"] = "Structural Concrete",
  ["struct control health monit"] = "Structural control & health monitoring",
  ["struct des tall build"] = "Structural Design of Tall Buildings",
  ["struct des tall special build"] = "Structural Design of Tall and Special Buildings",
  ["struct durability health monit"] = "Structural Durability & Health Monitoring",
  ["struct dyn"] = "Structural dynamics (Melville, N.Y.)",
  ["struct eng int"] = "Structural Engineering International",
  ["struct eng mech"] = "Structural Engineering and Mechanics",
  ["struct equ model"] = "Structural Equation Modeling. A Multidisciplinary Journal",
  ["struct equ modeling"] = "Structural equation modeling : a multidisciplinary journal",
  ["struct equation modell multidiscip j"] = "Structural Equation Modelling A Multidisciplinary Journal",
  ["struct health monit"] = "Structural health monitoring",
  ["struct heart"] = "Structural heart : the journal of the Heart Team",
  ["struct infrastruct eng"] = "Structure and Infrastructure Engineering",
  ["struct integr"] = "Structural Integrity",
  ["struct multidiscip  optim"] = "Structural and Multidisciplinary Optimization",
  ["struct multidiscip optim"] = "Structural and Multidisciplinary Optimization",
  ["struct multidiscipl optim"] = "Structural and multidisciplinary optimization : journal of the International Society for Structural and Multidisciplinary Optimization",
  ["struct saf"] = "Structural Safety",
  ["struct surv"] = "Structural Survey",
  ["structural topology"] = "Structural Topology",
  ["structure (cambridge, ma, u s)"] = "Structure (Cambridge, MA, United States)",
  ["structure (oxford, u k)"] = "Structure (Oxford, United Kingdom)",
  ["structure fold des"] = "Structure with Folding and Design",
  ["stud adv math"] = "Studies in Advanced Mathematics",
  ["stud am renaiss"] = "Studies in the American renaissance",
  ["stud anc med"] = "Studies in ancient medicine",
  ["stud angew wirtschaftsforsch statist inst statist okonom univ hamburg"] = "Studien zur Angewandten Wirtschaftsforschung und Statistik aus dem Institut fur Statistik und Okonometrie der Universitat Hamburg",
  ["stud appl math"] = "Studies in Applied Mathematics",
  ["stud appl mech"] = "Studies in Applied Mechanics",
  ["stud appl philos epistemol ration ethics"] = "Studies in Applied Philosophy, Epistemology and Rational Ethics",
  ["stud asiat"] = "Studia Asiatica (Bucharest, Romania)",
  ["stud behav adapt"] = "Studies in Behavioural Adaptation",
  ["stud bibliogr"] = "Studies in bibliography",
  ["stud big data"] = "Studies in Big Data",
  ["stud bot hung"] = "Studia botanica Hungarica",
  ["stud burke time"] = "Studies in Burke and his time",
  ["stud bus cycles"] = "Business Cycles, Indicators and Forecasting",
  ["stud cartesiana"] = "Studia Cartesiana",
  ["stud cerc mat"] = "Studii şi Cercetări Matematice",
  ["stud cerc mec apl"] = "Studii şi Cercetări de Mecanică Aplicată",
  ["stud cercet econ"] = "Studii si Cercetari Economice",
  ["stud cercet endocrinol"] = "Studii şi cercetari de endocrinologie",
  ["stud cercet fiziol"] = "Studii şi cercetări de fiziologie",
  ["stud cercet inframicrobiol"] = "Studii şi cercetări de inframicrobiologie",
  ["stud cercet med interna"] = "Studii şi cercetări de medicină internă",
  ["stud cercet neurol"] = "Studii si cercetări de neurologie",
  ["stud cercet virusol"] = "Studii şi cercetări de virusologie",
  ["stud choice welf"] = "Studies in Choice and Welfare",
  ["stud christ ethics"] = "Studies in Christian ethics",
  ["stud chur hist"] = "Studies in church history (Chicago, Ill.)",
  ["stud cl"] = "Studii clasice",
  ["stud cl orient"] = "Studi classici e orientali",
  ["stud classification data anal knowledge organ"] = "Studies in Classification, Data Analysis, and Knowledge Organization",
  ["stud comeniana hist"] = "Studia Comeniana et historica",
  ["stud comp int dev"] = "Studies in comparative international development",
  ["stud comput intell"] = "Studies in Computational Intelligence",
  ["stud comput math"] = "Studies in Computational Mathematics",
  ["stud comput sci artif intell"] = "Studies in Computer Science and Artif. Intell.",
  ["stud conserv"] = "Studies in Conservation",
  ["stud decor arts"] = "Studies in the decorative arts",
  ["stud demogr"] = "Studi di demografia",
  ["stud dipterol"] = "Studia dipterologica",
  ["stud dynam econom sci"] = "Studies in Dynamical Economic Science",
  ["stud econ"] = "Studi Economici",
  ["stud econ des"] = "Studies in Economic Design",
  ["stud econ finance"] = "Studies in Economics and Finance",
  ["stud econom theory"] = "Studies in Economic Theory",
  ["stud eighteenth cent cult"] = "Studies in eighteenth-century culture",
  ["stud emigrazione"] = "Studi emigrazione",
  ["stud empir econom"] = "Studies in Empirical Economics",
  ["stud engl lit"] = "Studies in English literature",
  ["stud ethics law technol"] = "Studies in ethics, law, and technology",
  ["stud ethn natl"] = "Studies in ethnicity and nationalism",
  ["stud eur thought"] = "Studies in European thought",
  ["stud fam plann"] = "Studies in family planning",
  ["stud family planning"] = "Studies in Family Planning",
  ["stud filoz"] = "Studia filozoficzne",
  ["stud for suec"] = "Studia forestalia suecica",
  ["stud fungi"] = "Studies in fungi",
  ["stud fuzziness"] = "Studies in Fuzziness",
  ["stud fuzziness soft comput"] = "Studies in Fuzziness and Soft Computing",
  ["stud gen (berl)"] = "Studium generale; Zeitschrift für die Einheit der Wissenschaften im Zusammenhang ihrer Begriffsbildungen und Forschungsmethoden",
  ["stud gen philos sci"] = "Studies in General Philosophy of Science",
  ["stud gend sex"] = "Studies in gender and sexuality",
  ["stud geogr"] = "Studia geographica",
  ["stud geophys geod"] = "Studia Geophysica et Geodaetica",
  ["stud gesch krankenhauswesens"] = "Studien zur Geschichte des Krankenhauswesens",
  ["stud gesch univ wien"] = "Studien zur Geschichte der Universität Wien",
  ["stud hastings cent"] = "Studies - Hastings Center",
  ["stud health technol inf"] = "Studies in Health Technology and Informatics",
  ["stud health technol inform"] = "Studies in health technology and informatics",
  ["stud hibernica"] = "Studia hibernica",
  ["stud high energy phys cosmol gravit"] = "Studies in High Energy Physics Cosmology and Gravitation",
  ["stud hist"] = "Studies in history",
  ["stud hist (sahibabad)"] = "Studies in history",
  ["stud hist biol"] = "Studies in history of biology",
  ["stud hist gandensia"] = "Studia historica Gandensia",
  ["stud hist gardens des landscapes"] = "Studies in the History of Gardens & Designed Landscapes",
  ["stud hist ha contemp"] = "Studia historica. Ha. contemporánea",
  ["stud hist ha mod"] = "Studia historica. Ha. moderna",
  ["stud hist math enq"] = "Studies in the History of Mathematical Enquiry",
  ["stud hist med"] = "Studies in history of medicine",
  ["stud hist med sci"] = "Studies in history of medicine and science",
  ["stud hist oecon"] = "Studia historiae oeconomicae",
  ["stud hist philos biol biomed sci"] = "Studies in history and philosophy of biological and biomedical sciences",
  ["stud hist philos math"] = "Studies in the History and Philosophy of Mathematics",
  ["stud hist philos sci"] = "Studies in history and philosophy of science",
  ["stud hist philos sci b stud hist philos modern phys"] = "Studies in History and Philosophy of Science. Part B. Studies in History and Philosophy of Modern Physics",
  ["stud hist philos sci part a"] = "Studies in History and Philosophy of Science Part A",
  ["stud hist philos sci part b: stud hist philos mod phys"] = "Studies in History and Philosophy of Science Part B: Studies in History and Philosophy of Modern Physics",
  ["stud hist philos sci part c: stud hist philos biol biomed sci"] = "Studies in History and Philosophy of Science Part C: Studies in History and Philosophy of Biological and Biomedical Sciences",
  ["stud hum ecol"] = "Studies in human ecology",
  ["stud infrastruct control"] = "Studies in Infrastructure and Control",
  ["stud inst divi thomae"] = "Studies. Institutum Divi Thomae",
  ["stud intercult phil"] = "Studies in Intercultural Philosophy",
  ["stud islam"] = "Studia islamica",
  ["stud knowledge"] = "Studies in Knowledge",
  ["stud labour hist"] = "Studies in labour history",
  ["stud lang"] = "Studies in language",
  ["stud language companion ser (slcs)"] = "Studies in Language Companion Series (SLCS)",
  ["stud lat am pop cult"] = "Studies in Latin American popular culture",
  ["stud lawyer"] = "Student lawyer (Chicago, Ill. : 1972)",
  ["stud leibnitiana"] = "Studia Leibnitiana",
  ["stud linguist"] = "Studia linguistica",
  ["stud linguist philos"] = "Studies in Linguistics and Philosophy",
  ["stud log"] = "Studia logica",
  ["stud log (lond)"] = "Studies in Logic (London)",
  ["stud log pract reason"] = "Studies in Logic and Practical Reasoning",
  ["stud logic comput"] = "Studies in Logic and Computation",
  ["stud logic found math"] = "Studies in Logic and the Foundations of Mathematics",
  ["stud logic lang inform"] = "Studies in Logic, Language and Information",
  ["stud logica"] = "Studia Logica",
  ["stud mater dziej nauk pol ser b"] = "Studia i materiały z dziejów nauki polskiej. Seria B: Historia nauk biologicznych i medycznych",
  ["stud mater dziejow nauk pol"] = "Studia i materiały z dziejów nauki polskiej",
  ["stud mater ist medie"] = "Studii și materiale de istorie medie. Institutul de Istorie (Academia Republicii Populare Romîne)",
  ["stud math"] = "Studia Mathematica",
  ["stud math appl"] = "Studies in Mathematics and its Applications",
  ["stud math arts humanit"] = "Studies in Mathematics in the Arts and Humanities",
  ["stud math ed ser"] = "Studies in Mathematics Education Series",
  ["stud math libr"] = "Student Mathematical Library",
  ["stud math managerial econom"] = "Studies in Mathematical and Managerial Economics",
  ["stud math phys"] = "Studies in Mathematical Physics",
  ["stud mechanobiol tissue eng biomater"] = "Studies in Mechanobiology, Tissue Engineering and Biomaterials",
  ["stud med"] = "Student medicine",
  ["stud med popul subj"] = "Studies on medical and population subjects",
  ["stud media commun"] = "Studies in media and communication",
  ["stud mediev cult"] = "Studies in medieval culture",
  ["stud medizingesch neunzehnten jahrhunderts"] = "Studien zur Medizingeschichte des neunzehnten Jahrhunderts",
  ["stud ment ineffic"] = "Studies in mental inefficiency",
  ["stud mitt gesch benediktinerorden"] = "Studien und Mitteilungen zur Geschichte des Benediktiner-Ordens und seiner Zweige",
  ["stud morphol"] = "Studies in Morphology",
  ["stud mycol"] = "Studies in mycology",
  ["stud nat lang linguist theory"] = "Studies in Natural Language and Linguistic Theory",
  ["stud nat prod chem"] = "Studies in Natural Products Chemistry",
  ["stud neophilol"] = "Studia neophilologica",
  ["stud neotrop fauna environ"] = "Studies on neotropical fauna and environment",
  ["stud nepali hist soc"] = "Studies in Nepali history and society",
  ["stud neuroanat"] = "Studies in neuro-anatomy",
  ["stud nonlinear dyn econom"] = "Studies in Nonlinear Dynamics and Econometrics",
  ["stud nonlinear dynam econometrics"] = "Studies in Nonlinear Dynamics and Econometrics",
  ["stud nonlinear phenom life sci"] = "Studies of Nonlinear Phenomena in Life Science",
  ["stud odonatol hung"] = "Studia odonatologica Hungarica",
  ["stud org chem (amsterdam)"] = "Studies in Organic Chemistry (Amsterdam)",
  ["stud papyrol"] = "Studia papyrologica",
  ["stud philol"] = "Studies in philology",
  ["stud philos"] = "Studia Philosophica",
  ["stud philos hist"] = "Studia Philosophica et Historica",
  ["stud philos hist philos"] = "Studies in Philosophy and the History of Philosophy",
  ["stud philos med"] = "Studies in philosophy of medicine",
  ["stud platonism neoplatonism platonic tradit"] = "Studies in Platonism, Neoplatonism, and the Platonic Tradition",
  ["stud polit econ"] = "Studies in political economy : a socialist review",
  ["stud prakt"] = "Student und Praktikant",
  ["stud profertility ser"] = "Studies in profertility series",
  ["stud psychol (bratisl)"] = "Studia psychologica",
  ["stud public choice"] = "Studies in Public Choice",
  ["stud q j"] = "Students’ Quarterly Journal",
  ["stud quat"] = "Studia Quaternaria",
  ["stud relig"] = "Studies in religion. Sciences religieuses",
  ["stud renaissance"] = "Studies in the Renaissance",
  ["stud rockefeller inst med res repr"] = "Studies from the Rockefeller institute for medical research. Reprints. Rockefeller Institute for Medical Research",
  ["stud romanticism"] = "Studies in romanticism",
  ["stud rosenthaliana"] = "Studia Rosenthaliana",
  ["stud russ econ dev"] = "Studies on Russian economic development",
  ["stud sci educ"] = "Studies in science education",
  ["stud sci math hung"] = "Studia Scientiarum Mathematicarum Hungaricae",
  ["stud scott lit"] = "Studies in Scottish literature",
  ["stud second lang acquis"] = "Studies in second language acquisition",
  ["stud semant web"] = "Studies on the Semantic Web",
  ["stud soc sci res"] = "Studies in social science research",
  ["stud soc sci west ga coll"] = "Studies in the social sciences (West Georgia College)",
  ["stud sov thought"] = "Studies in Soviet thought",
  ["stud stor"] = "Studi storici",
  ["stud surf sci catal"] = "Studies in surface science and catalysis",
  ["stud symb interact"] = "Studies In Symbolic Interaction",
  ["stud syst decis control"] = "Studies in Systems, Decision and Control",
  ["stud texts"] = "Studies and Texts",
  ["stud theor appl stat sel papers stat soc"] = "Studies in Theoretical and Applied Statistics. Selected Papers of the Statistical Societies",
  ["stud theor philos"] = "Studies in Theoretical Philosophy",
  ["stud theor psycholinguist"] = "Studies in Theoretical Psycholinguistics",
  ["stud third world soc"] = "Studies in third world societies",
  ["stud tokugawa inst"] = "Studies from the Tokugawa Institute. Tokugawa Seibutsugaku Kenkyūsho, Tokyo",
  ["stud trop andean ecosyst"] = "Studies on Tropical Andean Ecosystems",
  ["stud univ babes bolyai biol"] = "Studia Universitatis Babeș-Bolyai. Biologia",
  ["stud univ babes-bolyai inform"] = "Studia. Universitatis Babes-Bolyai. Informatica",
  ["stud univ babes-bolyai math"] = "Studia. Universitatis Babes-Bolyai Mathematica",
  ["stud univ babes-bolyai, chem"] = "Studia Universitatis Babes-Bolyai, Chemia",
  ["stud univ babeș-bolyai inform"] = "Studia",
  ["stud univ transp commun žilina math-phys ser"] = "Studies of the University of Transport and Communications in Žilina",
  ["stud univ zilina math ser"] = "Studies of the University of Zilina. Mathematical Series",
  ["stud univ žilina math phys ser"] = "Studies of University in Žilina",
  ["stud univers log"] = "Studies in Universal Logic",
  ["stud voltaire 18th century"] = "Studies on Voltaire and the eighteenth century",
  ["stud wirtschmath"] = "Studienbucher Wirtschaftsmathematik",
  ["stud wiss soz bildungsgesch math"] = "Studien zur Wissenschafts-, Sozial- und Bildungsgeschichte der Mathematik",
  ["stud zakresu inz"] = "Studia z Zakresu Inzynierii",
  ["stud österr philos"] = "Studien zur Österreichischen Philosophie",
  ["studenterraad med"] = "Stud. med",
  ["studi cent interuniv stor univ ital"] = "Studi. Centro Interuniversitario per la Storia delle Universita Italiane",
  ["studi emigr"] = "Studi emigrazione : international journal of migration studies",
  ["studi filol ital"] = "Studi di filologia italiana : bulletino dell' Accademia della Crusca",
  ["studi fr"] = "Studi francesi",
  ["studi galileiani"] = "Studi Galileiani",
  ["studi ist gramsci sicil"] = "Studi dell’Istituto Gramsci Siciliano",
  ["studi ital filol cl"] = "Studi italiani di filologia classica",
  ["studi med chir sport"] = "Studi di medicina e chirurgia dello sport",
  ["studi medievali"] = "Studi medievali (Turin, Italy : 1928)",
  ["studi piemont"] = "Studi piemontesi",
  ["studi sassar"] = "Studi sassaresi",
  ["studi sociol"] = "Studi di sociologia",
  ["studi trentini sci nat acta geol"] = "Studi trentini di scienze naturali. Acta geologica",
  ["studi veneziani"] = "Studi veneziani",
  ["studia automat inform"] = "Poznańskie Towarzystwo Przyjaciół Nauk",
  ["studia demogr"] = "Studia demograficzne",
  ["studia leibnitiana"] = "Studia Leibnitiana",
  ["studia logica"] = "Studia Logica. An International Journal for Symbolic Logic",
  ["studia math"] = "Studia Mathematica",
  ["studia sci math hungar"] = "Studia Scientiarum Mathematicarum Hungarica. A Quarterly of the Hungarian Academy of Sciences",
  ["studia univ babeș-bolyai math"] = "Universitatis Babeș-Bolyai",
  ["studia univ babeș-bolyai phys"] = "Universitatis Babeș-Bolyai",
  ["studienbucher math"] = "Studienbucher Mathematik",
  ["studienbucher tech wiss"] = "Studienbucher der Technischen Wissenschaften",
  ["studime hist"] = "Studime historike",
  ["studium (rotterdam)"] = "Studium (Rotterdam, Netherlands)",
  ["study encounter"] = "Study encounter",
  ["study sess can cathol hist assoc"] = "Study sessions - Canadian Catholic Historical Association. Canadian Catholic Historical Association",
  ["stuttg beitr naturkd"] = "Stuttgarter Beiträge zur Naturkunde",
  ["stuttgart beitr naturkd a"] = "Stuttgarter Beitraege zur Naturkunde A",
  ["sub-cell biochem"] = "Sub-cellular Biochemistry",
  ["subacute care"] = "Subacute care",
  ["subcell biochem"] = "Sub-cellular biochemistry",
  ["subnucl ser"] = "The Subnuclear Series",
  ["subser control"] = "Subseries in Control",
  ["subsid med"] = "Subsidia medica",
  ["subst abus"] = "Substance abuse",
  ["subst abuse"] = "Substance abuse : research and treatment",
  ["subst abuse rehabil"] = "Substance abuse and rehabilitation",
  ["subst abuse treat prev policy"] = "Substance abuse treatment, prevention, and policy",
  ["subst alcohol actions misuse"] = "Substance and alcohol actions/misuse",
  ["subst use misuse"] = "Substance use & misuse",
  ["subterr biol"] = "Subterranean biology",
  ["sud med chir"] = "Le Sud médical et chirurgical",
  ["sud med ekspert"] = "Sudebno-meditsinskaia ekspertiza",
  ["sudan j paediatr"] = "Sudanese journal of paediatrics",
  ["sudan j popul stud"] = "Sudan journal of population studies",
  ["sudan j vet res"] = "The Sudan journal of veterinary research",
  ["sudan med j"] = "Sudan medical journal",
  ["sudan notes rec"] = "Sudan notes and records",
  ["sudan nubia"] = "Sudan & Nubia : the Sudan Archaeological Research Society bulletin",
  ["suddeutsch apoth ztg"] = "Süddeutsche Apotheker-Zeitung",
  ["suddtsch optikerztg"] = "Süddeutsche Optikerzeitung",
  ["sudest med"] = "Sud-est médical",
  ["sudhoffs arch"] = "Sudhoffs Archiv",
  ["sudhoffs arch gesch med"] = "Sudhoffs Archiv für Geschichte der Medizin",
  ["sudhoffs arch gesch med naturwiss"] = "Sudhoffs Archiv für Geschichte der Medizin und der Naturwissenschaften",
  ["sudhoffs arch vierteljahrsschr gesch med naturwiss pharm math"] = "Sudhoffs Archiv; Vierteljahrsschrift fur Geschichte der Medizin und der Naturwissenschaften, der Pharmazie und der Mathematik",
  ["sudhoffs arch z wissenschaftsgesch"] = "Sudhoffs Archiv; Zeitschrift fur Wissenschaftsgeschichte",
  ["sudhoffs arch z wissenschaftsgesch beih"] = "Sudhoffs Archiv; Zeitschrift fur Wissenschaftsgeschichte. Beihefte",
  ["sudhoffs arch, vierteljahresschr fur gesch med naturwiss pharm math"] = "Sudhoffs Archiv. Vierteljahresschrift fur Geschichte der Medizin und der Naturwissenschaften, der Pharmazie und der Mathematik",
  ["sudhoffs arch, z wissenschaftsgesch"] = "Sudhoffs Archiv. Zeitschrift für Wissenschaftsgeschichte",
  ["sudhoffs klass med"] = "Sudhoffs Klassiker der Medizin (Leipzig, Germany : 1981)",
  ["sudwestdtsch arztebl"] = "Südwestdeutsches Ärzteblatt",
  ["suffield tech note can suffield exp stn ralston alta"] = "Suffield technical note. Canada. Suffield Experimental Station, Ralston, Alta",
  ["suffield tech pap can suffield exp stn ralston alta"] = "Suffield technical paper. Canada. Suffield Experimental Station, Ralston, Alta",
  ["suffolk transnatl law r"] = "Suffolk transnational law review",
  ["suffolk univ law rev"] = "Suffolk University law review",
  ["sugaku expositions"] = "Sugaku Expositions",
  ["sugar tech"] = "Sugar tech : an international journal of sugar crops & related industries",
  ["suicide life threat behav"] = "Suicide & life-threatening behavior",
  ["suicidol online"] = "Suicidology online",
  ["suid-afrikaanse tydskr ekon bestuurswetenskappe"] = "Suid-Afrikaanse Tydskrif vir Ekonomiese en Bestuurswetenskappe",
  ["suid-afrikaanse tydskr natuurwetenskap tegnol"] = "Die Suid-Afrikaanse Tydskrif vir Natuurwetenskap en Tegnologie",
  ["suid-afrikaanse tydskrif vir plantkunde : amptelike tydskrif van die suid-afrikaanse genootskap van plantkundiges|s afr j bot"] = "South African journal of botany : official journal of the South African Association of Botanists",
  ["suisan ikushu"] = "Suisan ikushu",
  ["suisan sogo kenkyu senta kenkyu hokoku"] = "Suisan Sōgō Kenkyū Sentā kenkyū hōkoku = Bulletin of Fisheries Research Agency",
  ["suisan zoshoku"] = "Suisan zōshoku",
  ["sujil bojeon"] = "Sujil bojeon",
  ["sulfur lett"] = "Sulfur Letters",
  ["sultan qaboos univ med j"] = "Sultan Qaboos University medical journal",
  ["sumar list"] = "Šumarski list",
  ["sumer j med healthc"] = "Sumerianz journal of medical and healthcare",
  ["summ shute inst"] = "The Summary",
  ["summa phytopathol"] = "Summa phytopathológica",
  ["summer comput simul conf (2014)"] = "Summer Computer Simulation Conference : (SCSC 2014) : 2014 Summer Simulation Multi-Conference : Monterey, California, USA, 6-10 July 2014. Summer Computer Simulation Conference (2014 : Monterey, Calif.)",
  ["summer comput simul conf (2015)"] = "Summer Computer Simulation Conference : (SCSC 2015) : 2015 Summer Simulation Multi-Conference (SummerSim'15) : Chicago, Illinois, USA, 26-29 July 2015. Summer Computer Simulation Conference (2015 : Chicago, Illinois)",
  ["summit transl bioinform"] = "Summit on translational bioinformatics",
  ["sums read"] = "SUMS Readings",
  ["sunday times"] = "Sunday times (London, England : 1931)",
  ["sunday times magazine"] = "The Sunday times magazine",
  ["suny ser anc greek philos"] = "SUNY Series in Ancient Greek Philosophy",
  ["suny ser logic lang"] = "SUNY Series in Logic and Language",
  ["suny ser reform math educ"] = "SUNY Series, Reform in Mathematics Education",
  ["suny ser sci tech soc"] = "SUNY Series in Science, Technology, and Society",
  ["suom apteenkkaril"] = "Suomen apteekkarilehti. Finlands apotekartidning",
  ["suom hammaslaak toim"] = "Suomen Hammaslaakariseuran Toimituksia",
  ["suom hammaslaakarilehti"] = "Suomen hammaslääkärilehti = Finlands tandläkartidning",
  ["suom laakaril"] = "Suomen lääkärilehti. Finlands läkartidning",
  ["suom naishammaslaak julk"] = "Suomen Naishammaslaakarit Ryhma Julkaisu",
  ["suom ortoped traumatol"] = "Suomen ortopedia ja traumatologia = Ortopedi och traumatologi i Finland = Finnish journal of orthopaedics and traumatology",
  ["suom tiedeakat toim a5 medica anthropol"] = "Suomalaisen Tiedeakatemian toimituksia. Sar. A.5, Medica-anthropologica",
  ["supercomput front innov"] = "Supercomputing frontiers and innovations",
  ["supercond sci technol"] = "Superconductor Science and Technology",
  ["superlattices microstruct"] = "Superlattices and Microstructures",
  ["superv manage"] = "Supervisory management",
  ["superv nurse"] = "Supervisor nurse",
  ["supl pract med"] = "Practica médica",
  ["supl rev mex fis"] = "Suplemento de la Revista Mexicana de FÃ­sica",
  ["suppl clin neurophysiol"] = "Supplements to Clinical neurophysiology",
  ["suppl eur j neurosci"] = "Supplement to the European Journal Neuroscience",
  ["suppl int j gynecol obstet"] = "Supplement to International journal of gynecology and obstetrics",
  ["suppl thromb haemost"] = "Supplementum ad Thrombosis and Haemostasis",
  ["suppl tumori"] = "Supplementi di Tumori",
  ["support cancer ther"] = "Supportive cancer therapy",
  ["support care cancer"] = "Supportive care in cancer : official journal of the Multinational Association of Supportive Care in Cancer",
  ["supramol chem"] = "Supramolecular Chemistry",
  ["supramol mater"] = "Supramolecular Materials",
  ["supramol photosensit electroact mater"] = "Supramolecular Photosensitive and Electroactive Materials",
  ["supramol sci"] = "Supramolecular Science",
  ["supreme court econ rev"] = "Supreme Court Economic Review",
  ["supreme court rev"] = "The Supreme Court review",
  ["surf coat int"] = "Surface Coatings International",
  ["surf coat int: part a"] = "Surface Coatings International Part A: Coatings Journal",
  ["surf coat int: part b"] = "Surface Coatings International Part B: Coatings Transactions",
  ["surf coat technol"] = "Surface and Coatings Technology",
  ["surf eng"] = "Surface Engineering",
  ["surf eng appl electrochem"] = "Surface Engineering and Applied Electrochemistry",
  ["surf innov"] = "Surface innovations",
  ["surf innovations"] = "Surface Innovations",
  ["surf interface anal"] = "Surface and Interface Analysis",
  ["surf interfaces"] = "Surfaces and Interfaces",
  ["surf rev lett"] = "Surface Review and Letters",
  ["surf sci"] = "Surface Science",
  ["surf sci lett"] = "Surface Science Letters",
  ["surf sci rep"] = "Surface Science Reports",
  ["surf sci spectra"] = "Surface Science Spectra",
  ["surf technol"] = "Surface Technology",
  ["surf topogr"] = "Surface topography : metrology and properties",
  ["surf topogr metrol prop"] = "Surface Topography: Metrology and Properties",
  ["surf topogr: metrol prop"] = "Surface Topography: Metrology and Properties",
  ["surg annu"] = "Surgery annual",
  ["surg bus"] = "Surgical business",
  ["surg case rep"] = "Surgical case reports",
  ["surg case rep (tallinn)"] = "Surgical case reports (Tallinn, Estonia)",
  ["surg circ lett u s army far east command med sect"] = "Surgeon's circular letter. United States. Far East Command. Medical Section",
  ["surg clin north am"] = "The Surgical clinics of North America",
  ["surg endosc"] = "Surgical endoscopy",
  ["surg equip"] = "Surgical equipment",
  ["surg exp pathol"] = "Surgical and Experimental Pathology",
  ["surg forum"] = "Surgical forum",
  ["surg gastroenterol"] = "Surgical gastroenterology",
  ["surg gynecol obstet"] = "Surgery, gynecology & obstetrics",
  ["surg infect (larchmt)"] = "Surgical infections",
  ["surg innov"] = "Surgical innovation",
  ["surg innovation"] = "Surgical Innovation",
  ["surg j (n y)"] = "Surgery journal (New York, N.Y.)",
  ["surg laparosc endosc"] = "Surgical laparoscopy & endoscopy",
  ["surg laparosc endosc percutan tech"] = "Surgical laparoscopy, endoscopy & percutaneous techniques",
  ["surg med technol"] = "Surgery and Medical Technologies",
  ["surg neurol"] = "Surgical neurology",
  ["surg neurol int"] = "Surgical neurology international",
  ["surg obes relat dis"] = "Surgery for obesity and related diseases : official journal of the American Society for Bariatric Surgery",
  ["surg oncol"] = "Surgical oncology",
  ["surg oncol clin n am"] = "Surgical oncology clinics of North America",
  ["surg open sci"] = "Surgery open science",
  ["surg pathol clin"] = "Surgical pathology clinics",
  ["surg pract"] = "Surgical practice",
  ["surg radiol anat"] = "Surgical and radiologic anatomy : SRA",
  ["surg res commun"] = "Surgical research communications",
  ["surg res pract"] = "Surgery research and practice",
  ["surg rounds"] = "Surgical rounds",
  ["surg sci"] = "Surgical science",
  ["surg staff  semin u s veterans adm hosp minneap"] = "Surgical staff seminars. United States. Veterans Administration. Hospital, Minneapolis",
  ["surg technol"] = "The Surgical technologist",
  ["surg technol int"] = "Surgical technology international",
  ["surg today"] = "Surgery today",
  ["surgery (oxf)"] = "Surgery (Oxford, Oxfordshire)",
  ["surgo glasg univ med j"] = "Surgo; Glasgow University medical journal",
  ["surv approx theory"] = "Surveys in Approximation Theory",
  ["surv biol prog"] = "Survey of biological progress",
  ["surv curr bus"] = "Survey of Current Business",
  ["surv differ geom"] = "Surveys in Differential Geometry",
  ["surv geophys"] = "Surveys in Geophysics",
  ["surv graph"] = "Survey graphic",
  ["surv high energy phys"] = "Surveys in High Energy Physics",
  ["surv immunol res"] = "Survey of immunologic research",
  ["surv math appl"] = "Surveys in Mathematics and its Applications",
  ["surv methodol"] = "Survey methodology",
  ["surv methods insights field"] = "Survey methods - insights from the field",
  ["surv mod math"] = "Surveys of Modern Mathematics",
  ["surv oper res manag sci"] = "Surveys in Operations Research and Management Science",
  ["surv ophthalmol"] = "Survey of ophthalmology",
  ["surv pract"] = "Survey practice",
  ["surv res methods"] = "Survey research methods",
  ["surv rev"] = "Survey Review",
  ["surv synth pathol res"] = "Survey and synthesis of pathology research",
  ["surv tutor appl math sci"] = "Surveys and Tutorials in the Applied Mathematical Sciences",
  ["surv world obstet gynecol (jpn)"] = "[Sekai sanfujinka sōran] [Survey of world obstetrics and gynecology]",
  ["surveil effic"] = "Surveillance efficace",
  ["surveill soc"] = "Surveillance & society",
  ["survey (lond)"] = "Survey (London, England : 1961)",
  ["surveys appl math"] = "Surveys in Applied Mathematics",
  ["surveys math indust"] = "Surveys on Mathematics for Industry",
  ["surviv news (atlanta ga)"] = "Survival news (Atlanta, Ga.)",
  ["survival (lond)"] = "Survival",
  ["susan haeyang gyoyuk yeongu"] = "Susan haeyang gyoyuk yeon-gu",
  ["suspop news"] = "SusPop news : the Sustainable Population Society newsletter",
  ["sustain chem pharm"] = "Sustainable chemistry and pharmacy",
  ["sustain cities soc"] = "Sustainable cities and society",
  ["sustain energy fuels"] = "Sustainable energy & fuels",
  ["sustain prod consum"] = "Sustainable production and consumption",
  ["sustain resilient infrastruct"] = "Sustainable and resilient infrastructure",
  ["sustain sci"] = "Sustainability science",
  ["sustain water resour manag"] = "Sustainable water resources management",
  ["sustainability (new rochelle)"] = "Sustainability (New Rochelle, N.Y.)",
  ["sustainability anal model"] = "Sustainability Analytics and Modeling",
  ["sustainability circularity now"] = "Sustainability & Circularity NOW",
  ["sustainability clim change"] = "Sustainability and Climate Change",
  ["sustainability manage forum"] = "Sustainability Management Forum",
  ["sustainability sci"] = "Sustainability Science",
  ["sustainability water qual ecol"] = "Sustainability Water Quality and Ecology",
  ["sustainability: sci, pract policy"] = "Sustainability: Science, Practice and Policy",
  ["sustainable build"] = "Sustainable Buildings",
  ["sustainable chem clim action"] = "Sustainable Chemistry for Climate Action",
  ["sustainable chem pharm"] = "Sustainable Chemistry and Pharmacy",
  ["sustainable chem processes"] = "Sustainable Chemical Processes",
  ["sustainable cities soc"] = "Sustainable Cities and Society",
  ["sustainable comput inf syst"] = "Sustainable Computing Informatics & Systems",
  ["sustainable dev"] = "Sustainable Development",
  ["sustainable earth"] = "Sustainable Earth",
  ["sustainable energy fuels"] = "Sustainable Energy & Fuels",
  ["sustainable energy grids networks"] = "Sustainable Energy, Grids and Networks",
  ["sustainable energy res"] = "Sustainable Energy Research",
  ["sustainable energy technol assess"] = "Sustainable Energy Technologies and Assessments",
  ["sustainable environ"] = "Sustainable Environment",
  ["sustainable environ res"] = "Sustainable Environment Research",
  ["sustainable food proteins"] = "Sustainable Food Proteins",
  ["sustainable food technol"] = "Sustainable Food Technology",
  ["sustainable futures"] = "Sustainable Futures",
  ["sustainable horiz"] = "Sustainable Horizons",
  ["sustainable manuf serv econ"] = "Sustainable Manufacturing and Service Economics",
  ["sustainable matertechnol"] = "Sustainable Materials and Technologies",
  ["sustainable oper comput"] = "Sustainable Operations and Computers",
  ["sustainable prod consumption"] = "Sustainable Production and Consumption",
  ["sustainable resilient infrastruct"] = "Sustainable and Resilient Infrastructure",
  ["sustainable technol entrepreneurship"] = "Sustainable Technology and Entrepreneurship",
  ["sustainable water resour manage"] = "Sustainable Water Resources Management",
  ["sut j math"] = "SUT Journal of Mathematics",
  ["suui kawahak yongu nonmunjip"] = "Suŭi kawahak yon'gu nonmunjip",
  ["suvr med (sofiia)"] = "Sŭvremenna medit︠s︡ina",
  ["svar proizvod"] = "Svarochnoe Proizvodstvo",
  ["sveikatos apsauga"] = "Sveikatos apsauga",
  ["sven bot tidskr"] = "Svensk Botanisk Tidskrift",
  ["sven farm tidskr"] = "Svensk farmaceutisk tidskrift",
  ["sven geogr arsb"] = "Svensk geografisk årsbok",
  ["sven juristtidn"] = "Svensk juristtidning",
  ["sven lakartidn"] = "Svenska läkartidningen",
  ["sven linnesallskap arsskr"] = "Svenska Linnésällskapets årsskrift = Yearbook of the Swedish Linnaeus Society",
  ["sven med tidskr"] = "Svensk medicinhistorisk tidskrift",
  ["sven medicinhistor tidskr"] = "Svensk Medicinhistorisk Tidskrift",
  ["sven mykol tidskr"] = "Svensk mykologisk tidskrift",
  ["sven papperstidn"] = "Svensk Papperstidning",
  ["sven tandlak tidskr"] = "Svensk tandläkare tidskrift. Swedish dental journal",
  ["sven tidskr"] = "Svensk tidskrift (Uppsala, Sweden)",
  ["svensk sjukkasse tidn"] = "Svensk sjukkasse-tidning",
  ["sver tandlakarforb tidn"] = "Tidning. Sveriges Tandläkarförbund",
  ["sverkhprovodimost: fiz, khim, tekh"] = "Sverkhprovodimost: Fizika, Khimiya, Tekhnika",
  ["svob misl"] = "Svobodnai︠a︡ myslʹ (Moscow, Russia : 1991)",
  ["svu int j basic sci"] = "SVU international journal of basic sciences",
  ["svyaz\\cprime vremen"] = "\\cyr Svyaz\\cprime Vremen",
  ["swarm evol comput"] = "Swarm and Evolutionary Computation",
  ["swarm intell"] = "Swarm Intelligence",
  ["swasth hind"] = "Swasth hind",
  ["swed am hist q"] = "The Swedish-American historical quarterly",
  ["swed dent j"] = "Swedish dental journal",
  ["swed dent j suppl"] = "Swedish dental journal. Supplement",
  ["swed pioneer hist q"] = "The Swedish pioneer historical quarterly",
  ["swedish econ pol rev"] = "Swedish Economic Policy Review",
  ["swedish j econ"] = "Swedish Journal of Economics",
  ["swiss biotech"] = "Swiss biotech",
  ["swiss dent"] = "Swiss dent",
  ["swiss dent j"] = "Swiss dental journal",
  ["swiss j econ stat"] = "Swiss journal of economics and statistics",
  ["swiss j econ statist"] = "Swiss Journal of Economics and Statistics",
  ["swiss j geosci"] = "Swiss Journal of Geosciences",
  ["swiss j palaeontol"] = "Swiss journal of palaeontology",
  ["swiss j psychol"] = "Swiss journal of psychology : official publication of the Swiss Psychological Society Schweizerische Zeitschrift für Psychologie = Revue suisse de psychologie",
  ["swiss med wkly"] = "Swiss medical weekly",
  ["swiss polit sci rev"] = "Swiss Political Science Review",
  ["swiss surg"] = "Swiss Surgery",
  ["swiss surg suppl"] = "Swiss surgery. Supplement",
  ["sxi springer innov"] = "Sxi—Springer per l’Innovazione/Sxi—Springer for Innovation",
  ["syd law rev"] = "The Sydney law review",
  ["sydn univ med j"] = "Sydney University medical journal",
  ["sydsven medicinhist sallsk arsskr"] = "Sydsvenska medicinhistoriska sällskapets årsskrift",
  ["sydsven medicinhist sallsk arsskr suppl"] = "Sydsvenska medicinhistoriska sällskapets årsskrift. Supplementum",
  ["sykepl fag"] = "Sykepleien. Fag",
  ["symb interact"] = "Symbolic interaction",
  ["symb oslo"] = "Symbolae Osloenses",
  ["symbol comput comput graph"] = "Symbolic Computation. Computer Graphics",
  ["symbol numer comput ser"] = "Symbolic and Numeric Computation Series",
  ["symmetry (basel)"] = "Symmetry",
  ["symmetry cult sci"] = "Symmetry: Culture and Science",
  ["symp (int) combust"] = "Symposium (International) on Combustion",
  ["symp (int) combust, [proc]"] = "Symposium (International) on Combustion, [Proceedings]",
  ["symp - int astron union"] = "Symposium - International Astronomical Union",
  ["symp faraday soc"] = "Symposium of the Faraday Society",
  ["symp fundam cancer res"] = "Symposium on Fundamental Cancer Research",
  ["symp geom process"] = "Symposium on geometry processing : [proceedings]. Symposium on Geometry Processing",
  ["symp int astron union"] = "Symposium International Astronomical Union",
  ["symp oral sens percept"] = "Symposium on Oral Sensation and Perception",
  ["symp pharmacol ther toxicol group"] = "Symposium Of The Pharmacology, Therapeutics And Toxicology Group, International Association For Dental Research",
  ["symp ser soc appl microbiol"] = "Symposium series (Society for Applied Microbiology)",
  ["symp soc dev biol"] = "Symposium of the Society for Developmental Biology",
  ["symp soc exp biol"] = "Symposia of the Society for Experimental Biology",
  ["symp swed nutr found"] = "Symposia of the Swedish Nutrition Foundation",
  ["symp theory model simul"] = "Symposium on Theory of Modeling & Simulation : DEVS Integrative M&S Symposium (TMS-DEVS). Symposium on Theory of Modeling and Simulation",
  ["symp vlsi circuits"] = "Symposium on VLSI Circuits : [proceedings]. Symposium on VLSI Circuits",
  ["sympos gaussiana"] = "Symposia Gaussiana",
  ["sympos math"] = "Symposia Mathematica",
  ["synapse (hoboken, nj, u s)"] = "Synapse (Hoboken, NJ, United States)",
  ["synapse (n y, ny, u s)"] = "Synapse (New York, NY, United States)",
  ["synchrotron radiat news"] = "Synchrotron radiation news",
  ["synerg syntropie nichtlineare syst"] = "Synergie—Syntropie—Nichtlineare Systeme",
  ["synergist (akron)"] = "Synergist (Akron, Ohio)",
  ["synth biol"] = "Synthetic Biology",
  ["synth biol (oxf)"] = "Synthetic biology (Oxford, England)",
  ["synth commun"] = "Synthetic Communications",
  ["synth fibres"] = "Synthetic Fibers",
  ["synth lect algorithms softw eng"] = "Synthesis Lectures on Algorithms and Software in Engineering",
  ["synth lect artif intell mach learn"] = "Synthesis Lectures on Artificial Intelligence and Machine Learning",
  ["synth lect comput sci"] = "Synthesis Lectures on Computer Science",
  ["synth lect comput vis"] = "Synthesis Lectures on Computer Vision",
  ["synth lect data min knowl discov"] = "Synthesis Lectures on Data Mining and Knowledge Discovery",
  ["synth lect data semant knowl"] = "Synthesis Lectures on Data, Semantics, and Knowledge",
  ["synth lect digit circuits syst"] = "Synthesis Lectures on Digital Circuits and Systems",
  ["synth lect eng sci technol"] = "Synthesis Lectures on Engineering, Science, and Technology",
  ["synth lect games comput intell"] = "Synthesis Lectures on Games and Computational Intelligence",
  ["synth lect hum lang technol"] = "Synthesis Lectures on Human Language Technologies",
  ["synth lect intell technol"] = "Synthesis Lectures on Intelligent Technologies",
  ["synth lect learn netw algorithms"] = "Synthesis Lectures on Learning, Networks, and Algorithms",
  ["synth lect math stat"] = "Synthesis Lectures on Mathematics and Statistics",
  ["synth lect mech eng"] = "Synthesis Lectures on Mechanical Engineering",
  ["synth lect oper res appl"] = "Synthesis Lectures on Operations Research and Applications",
  ["synth lect theor comput sci"] = "Synthesis Lectures on Theoretical Computer Science",
  ["synth lect vis comput"] = "Synthesis Lectures on Visual Computing",
  ["synth lect wave phenom phys sci"] = "Synthesis Lectures on Wave Phenomena in the Physical Sciences",
  ["synth libr"] = "Synthese Library. Studies in Epistemology, Logic, Methodology, and Philosophy of Science",
  ["synth lubr"] = "Synthetic Lubrication",
  ["synth met"] = "Synthetic Metals",
  ["synth philos"] = "Synthesis philosophica",
  ["synth proj res synth rep"] = "The Synthesis project. Research synthesis report",
  ["synth react inorg met-org chem"] = "Synthesis and Reactivity in Inorganic and Metal-Organic Chemistry",
  ["synth react inorg, met-org, nano-met chem"] = "Synthesis and Reactivity in Inorganic, Metal-Organic, and Nano-Metal Chemistry",
  ["synth syst biotechnol"] = "Synthetic and systems biotechnology",
  ["synthese lib"] = "Synthese Library",
  ["synthesis (stuttg)"] = "Synthesis",
  ["syracuse law civ engagem"] = "Syracuse law & civic engagement",
  ["syracuse law rev"] = "Syracuse law review",
  ["syst appl acarol"] = "Systematic and applied acarology",
  ["syst appl microbiol"] = "Systematic and applied microbiology",
  ["syst assoc spec vol"] = "Systematics Association special volume",
  ["syst biodivers"] = "Systematics and Biodiversity",
  ["syst biol"] = "Systematic biology",
  ["syst biol (stevenage)"] = "Systems biology",
  ["syst biol reprod med"] = "Systems biology in reproductive medicine",
  ["syst biomed (austin)"] = "Systems biomedicine (Austin, Tex.)",
  ["syst bot"] = "Systematic botany",
  ["syst bot monogr"] = "Systematic botany monographs : monographic series of the American Society of Plant Taxonomists",
  ["syst control lett"] = "Systems & control letters",
  ["syst dyn rev"] = "System dynamics review",
  ["syst eng"] = "Systems Engineering",
  ["syst eng procedia"] = "Systems Engineering Procedia",
  ["syst eng theory pract"] = "Systems Engineering - Theory & Practice",
  ["syst entomol"] = "Systematic entomology",
  ["syst eval predict decis-mak ser"] = "Systems Evaluation, Prediction, and Decision-Making Series",
  ["syst geogr plants"] = "Systematics and geography of plants",
  ["syst med (new rochelle)"] = "Systems medicine (New Rochelle, N.Y.)",
  ["syst microbiol biomanuf"] = "Systems Microbiology and Biomanufacturing",
  ["syst parasitol"] = "Systematic parasitology",
  ["syst pharmacol"] = "Systems pharmacology",
  ["syst pract action res"] = "Systemic practice and action research",
  ["syst res"] = "Systems research : the official journal of the International Federation for Systems Research",
  ["syst res behav sci"] = "Systems research and behavioral science",
  ["syst res forum"] = "Systems Research Forum",
  ["syst rev"] = "Systematic reviews",
  ["syst rev evid action int cochrane colloq 6th 1998 baltim md"] = "Systematic reviews, evidence for action : abstracts for workshops and scienitific [sic] sessions. International Cochrane Colloquium (6th : 1998 : Baltimore, Md.)",
  ["syst rev pharm"] = "Systematic Reviews in Pharmacy",
  ["syst sci control eng"] = "Systems Science and Control Engineering",
  ["syst soft comput"] = "Systems and Soft Computing",
  ["syst synth biol"] = "Systems and synthetic biology",
  ["syst thinker"] = "The Systems thinker",
  ["syst zool"] = "Systematic zoology",
  ["system develop found benchmark ser"] = "System Development Foundation Benchmark Series",
  ["systema ascomycetum"] = "Systema Ascomycetum",
  ["systems control found appl"] = "Systems & Control: Foundations & Applications",
  ["systems control inform"] = "Systems, Control and Information",
  ["systems control lett"] = "Systems & Control Letters",
  ["systems sci"] = "Technical University of Wrocław",
  ["systems sci math sci"] = "Systems Science and Mathematical Sciences",
  ["szabolcs szatmari szle"] = "Honismereti módszertani füzetek",
  ["szczecin tow nauk wydz nauk lek"] = "Szczecińskie Towarzystwo Naukowe, Wydział Nauk Lekarskich",
  ["szpit pol"] = "Szpitalnictwo polskie",
  ["säugetierkd mitt"] = "Säugetierkundliche Mitteilungen",
  ["sém anal moderne"] = "Séminaire d’Analyse Moderne",
  ["sém lothar combin"] = "Séminaire Lotharingien de Combinatoire",
  ["sém math luxembourg"] = "Séminaire de Mathématique de Luxembourg",
  ["sémin anal univ blaise pascal (clermont ii)"] = "Séminaire d’Analyse",
  ["sémin congr"] = "Séminaires et Congrès",
  ["sémin théor spectr géom"] = "Séminaire de Théorie Spectrale et Géométrie",
  ["sémin équ dériv partielles"] = "Séminaire: Équations aux Dérivées Partielles",
  ["słup prace mat przyr mat fiz"] = "Słupskie Prace Matematyczno-Przyrodnicze",
  ["t\\lasp bilis a razmadzis saxel mat\\lasp inst shromebi"] = "T\\lasp bilisis A",
  ["tabriz univ ser"] = "Tabriz University Series",
  ["tabulae biol"] = "Tabulae biologicae",
  ["taehan chikkwa uisa hyophoe chi"] = "Taehan Ch'ikkwa Uisa Hyŏphoe chi",
  ["taehan imsang misaengmul hakhoe chi"] = "Taehan Imsang Misaengmul Hakhoe chi = Korean journal of clinical microbiology",
  ["taehan kan hakhoe chi"] = "Taehan Kan Hakhoe chi = The Korean journal of hepatology",
  ["taehan kanho"] = "Taehan kanho. The Korean nurse",
  ["taehan kanho hakhoe chi"] = "Taehan Kanho Hakhoe chi",
  ["taehan kumsok hakhoechi"] = "Taehan Kumsok Hakhoechi",
  ["taehan kumsok, chaeryo hakhoechi"] = "Taehan Kumsok, Chaeryo Hakhoechi",
  ["taehan misaengmul hakhoe chi"] = "Taehan Misaengmul Hakhoe chi = The journal of the Korean Society for Microbiology",
  ["taehan naekwa hakhoe chapchi"] = "Taehan Naekwa Hakhoe chapchi = The Korean journal of internal medicine",
  ["taehan oekwa hakhoe chapchi"] = "[Chapchi] Journal. Taehan Oekwa Hakhoe",
  ["taehan pairosu hakhoe chi"] = "Taehan Pairŏsŭ Hakhoe chi",
  ["taehan pinyogikwa hakhoe chapchi"] = "Taehan Pinyogikwa Hakhoe chapchi = The Korean journal of urology",
  ["taehan sanbuinkwa hakhoe chapchi"] = "Taehan Sanbuinkwa Hakhoe chapchi = Korean journal of obstetrics and gynecology",
  ["taehan suui hakhoe chi taehan suui hakhoe"] = "Taehan Suŭi Hakhoe chi. Taehan Suŭi Hakhoe",
  ["taehan uihak hyophoe chi"] = "Taehan Ŭihak Hyŏphoe chi. The Journal of the Korean Medical Association",
  ["taehan uijinkyun hakhoe chi"] = "Taehan Ŭijinkyun Hakhoe chi = Korean journal of medical mycology",
  ["tag ber osterr arztetag"] = "Tagungsbericht. Österreichische Ärztetagung",
  ["tages-anz mag"] = "Tages-Anzeiger Magazin",
  ["tahq eq"] = "Tahqiqat-e eqtesadi (Quarterly Journal of Economic Research)",
  ["tahqiqati zhinitik islahi giyahani martaii jangalii iran"] = "Taḥqīqāt-i zhinitīk va iṣlāḥ-i giyāhān-i martaʻī-i va jangalī-i Īrān",
  ["tai da li shi xue bao"] = "Tai da li shi xue bao",
  ["tai da wen shi zhe xue bao"] = "Tai da wen shi zhe xue bao",
  ["tairyoku eiyo menekigaku zasshi"] = "Tairyoku, eiyō men'ekigaku zasshi = Journal of physical fitness, nutrition and immunology",
  ["taiwan econ rev"] = "Taiwan Economic Review",
  ["taiwan j math"] = "Taiwanese Journal of Mathematics",
  ["taiwan j obstet gynecol"] = "Taiwanese journal of obstetrics & gynecology",
  ["taiwan j ophthalmol"] = "Taiwan journal of ophthalmology",
  ["taiwan j orthod"] = "Taiwanese journal of orthodontics",
  ["taiwan jing ji yu ce"] = "Taiwan jing ji yu ce",
  ["taiwan lin ye ke xue"] = "Taiwan lin ye ke xue",
  ["taiwan nong ye hua xue yu shi pin ke xue"] = "Taiwan nong ye hua xue yu shi pin ke xue = Taiwanese journal of agricultural chemistry and food science",
  ["taiwan nong ye yan jiu"] = "Taiwan nong ye yan jiu",
  ["taiwan sheng wu duo yang xing yan jiu"] = "Taiwan sheng wu duo yang xing yan jiu",
  ["taiwan vet j"] = "Taiwan Veterinary Journal",
  ["taiwan xu mu shou yi xue hui hui bao"] = "Taiwan xu mu shou yi xue hui hui bao. Tʻai-wan sheng hsü mu shou i hsüeh hui",
  ["taiwan yi xue  ren wen xue kan"] = "Taiwan yi xue ren wen xue kan",
  ["taiwan yi xue hui za zhi"] = "Taiwan yi xue hui za zhi. Journal of the Formosan Medical Association",
  ["taiwanese j math"] = "Taiwanese Journal of Mathematics",
  ["taiwanese journal of agricultural chemistry and food science|taiwan nong ye hua xue yu shi pin ke xue"] = "Taiwan nong ye hua xue yu shi pin ke xue",
  ["taiyuan li gong da xue xue bao"] = "Taiyuan li gong da xue xue bao = Journal of Taiyuan University of Technology",
  ["takamine kenkyusho nenpo"] = "Takamine Kenkyūjo nenpō The Annual report of Takamine Laboratory. Sankyō Kabushiki Kaisha. Takamine Kenkyūjo",
  ["taksunumi biyusistimatik"] = "Tāksūnūmī va bīyūsīstimātīk",
  ["talanta open"] = "Talanta Open",
  ["talimu nongken daxue xuebao"] = "Talimu Nongken daxue xuebao",
  ["tamkang j math"] = "Tamkang Journal of Mathematics",
  ["tampa bay hist"] = "Tampa Bay history",
  ["tamsui oxf j inf math sci"] = "Tamsui Oxford Journal of Information and Mathematical Sciences",
  ["tamsui oxf j math sci"] = "Tamsui Oxford Journal of Mathematical Sciences",
  ["tamsui oxford j management sci"] = "Tamsui Oxford Journal of Management Sciences",
  ["tanaguchi symp brain sci"] = "Taniguchi symposia on brain sciences",
  ["tandlaegernes tidsskr"] = "Tandlaegernes nye tidsskrift",
  ["tani girisim radyol"] = "Tanısal ve girişimsel radyoloji : Tıbbi Görüntüleme ve Girişimsel Radyoloji Derneği yayın organı",
  ["tanpakushitsu kakusan koso"] = "Tanpakushitsu kakusan koso. Protein, nucleic acid, enzyme",
  ["tanzan econ rev"] = "Tanzanian economic review",
  ["tanzan health res bull"] = "Tanzania health research bulletin",
  ["tanzan j health res"] = "Tanzania journal of health research",
  ["tanzan notes rec"] = "Tanzania notes and records",
  ["tanzania j health res"] = "Tanzania Journal of Health Research",
  ["tap chi cong nghe sinh hoc"] = "Tạp chí công nghê sinh học",
  ["tap chi hoa hoc"] = "Tạp chí hóa học",
  ["tap chi khoa hoc"] = "Tạp chí khoa học = Journal of science",
  ["tap chi khoa hoc (hue)"] = "Tạp chí khoa học (Hue)",
  ["tap chi khoa hoc giao duc"] = "Tạp chí Khoa học & Giáo dục",
  ["tap chi khoa hoc khoa hoc xa hoi"] = "Tạp chí khoa học, Khoa học xã hội",
  ["tap chi khoa hoc ky thuat nong lam nghiep"] = "Tập chí khoa học kỹ thuật Nông Lâm Nghiệp",
  ["tap chi khoa hoc tu nhien"] = "Tạp chí Khoa học Tự nhiên",
  ["tap chi khoa hoc va cong nghe"] = "Tạp chí Khoa học và Công nghệ",
  ["tap chi khoa hoc va cong nghe nong nghiep viet nam"] = "Tạp chí Khoa học và Công nghệ Nông nghiệp Việt Nam",
  ["tap chi khoa hoc van lang"] = "Tạp chí Khoa học Văn Lang",
  ["tap chi nghien y hoc"] = "Tạp chí nghiên cứu y học",
  ["tap chi sinh hoc"] = "Tạp chí sinh học",
  ["tap chi y hoc du phong"] = "Tạp chí y học dự phòng = Journal of preventive medicine",
  ["tappi j"] = "Tappi Journal",
  ["tar heel nurse"] = "Tar heel nurse",
  ["target oncol"] = "Targeted oncology",
  ["target proteins database"] = "Target proteins database",
  ["targeted diagn ther"] = "Targeted diagnosis and therapy",
  ["targeted oncol"] = "Targeted Oncology",
  ["tartu ül toimetised"] = "Tartu Ülikooli Toimetised",
  ["tata inst fund res lectures on math and phys"] = "Tata Institute of Fundamental Research Lectures on Mathematics and Physics",
  ["tata inst fundam res stud math"] = "Tata Institute of Fundamental Research Studies in Mathematics",
  ["tatra mt math publ"] = "Tatra Mountains Mathematical Publications",
  ["taylor & francis syst control book ser"] = "The Taylor & Francis Systems and Control Book Series",
  ["tạp chi y te cong cong"] = "Tạp chí y tế công cộng",
  ["tb hiv"] = "TB & HIV",
  ["tbilisi math j"] = "Tbilisi Mathematical Journal",
  ["tbiliss gos univ inst prikl mat trudy"] = "Tbilisskiy Gosudarstvennyy Universitet",
  ["tdr news"] = "TDR news",
  ["teach child math"] = "Teaching children mathematics",
  ["teach coll rec"] = "Teachers College record",
  ["teach coll rec (1970)"] = "Teachers College record (1970)",
  ["teach educ spec educ"] = "Teacher education and special education",
  ["teach except child"] = "Teaching exceptional children",
  ["teach in"] = "Teach-in : the journal for junior hospital doctors and senior medical students",
  ["teach learn"] = "Teaching & learning",
  ["teach learn med"] = "Teaching and learning in medicine",
  ["teach learn nurs"] = "Teaching and learning in nursing : official journal of the National Organization for Assciate Degree Nursing",
  ["teach math appl"] = "Teaching Mathematics and its Applications",
  ["teach notes popul"] = "Teaching notes on population",
  ["teach philos"] = "Teaching philosophy",
  ["teach polit sci"] = "Teaching political science",
  ["teach psychol"] = "Teaching of psychology (Columbia, Mo.)",
  ["teach sociol"] = "Teaching Sociology",
  ["teach stat"] = "Teaching statistics",
  ["teach teach educ"] = "Teaching and teacher education",
  ["teaching bus econ"] = "Teaching Business and Economics",
  ["teaching sociology"] = "Teaching Sociology",
  ["tec bull (online)"] = "TEC bulletin",
  ["tech  doc rep rtd tdr"] = "Technical documentary report, RTD TDR. United States. Air Force. Systems Command. Research and Technology Division",
  ["tech anal strategic manage"] = "Technology Analysis and Strategic Management",
  ["tech belge prothese dent"] = "Le Technicien belge en prothèse dentaire",
  ["tech bull dep army"] = "Department of the Army technical bulletin. TB MED",
  ["tech bull regist med technol"] = "Technical bulletin of the Registry of Medical Technologists. American Society of Clinical Pathologists. Registry of Medical Technologists",
  ["tech bull united states veterans admin"] = "Technical bulletin. United States. Veterans Administration",
  ["tech coloproctol"] = "Techniques in coloproctology",
  ["tech commun"] = "Technical communication",
  ["tech commun q"] = "Technical Communication Quarterly",
  ["tech cult"] = "Technology and Culture",
  ["tech dig int electron devices meet"] = "Technical digest. International Electron Devices Meeting",
  ["tech doc hy"] = "Technical Documents in Hydrology",
  ["tech doc rep arct aeromed lab us"] = "Technical documentary report; AAL-TDR. Arctic Aeromedical Laboratory (U.S.)",
  ["tech doc rep arl tdr"] = "Technical documentary report, ARL-TDR. United States. Aeromedical Research Laboratory, Holloman Air Force Base, N. M",
  ["tech doc rep samtdr usaf sch aerosp med"] = "Technical documentary report. SAM-TDR. USAF School of Aerospace Medicine",
  ["tech doc rep u s air force syst command electron syst div"] = "Technical documentary report. United States. Air Force. Systems Command. Electronic Systems Division",
  ["tech eau assainissement"] = "La Technique de l'eau et de l'assainissement",
  ["tech ed ser"] = "Technology in Education Series",
  ["tech foot ankle surg"] = "Techniques in foot & ankle surgery",
  ["tech gastrointest endosc"] = "Techniques in gastrointestinal endoscopy",
  ["tech hand up extrem surg"] = "Techniques in hand & upper extremity surgery",
  ["tech hosp med soc sanit"] = "Techniques hospitalières, médico-sociales et sanitaires",
  ["tech ing constantes phys chim"] = "Techniques de l’Ingenieur, Constantes Physico-Chimiques",
  ["tech ing environ"] = "Techniques de l’Ingenieur, Environnement",
  ["tech ing genie nucl"] = "Techniques de l’Ingenieur, Genie Nuclear",
  ["tech ing genie procedes"] = "Techniques de l’Ingenieur, Genie des Procede",
  ["tech ing mater fond"] = "Techniques de l’Ingenieur, Materiaux Fonctionnels",
  ["tech ing mater met"] = "Techniques de l’Ingenieur, Materiaux Metalliques",
  ["tech ing plast compos"] = "Techniques de l’Ingenieur, Plastiques et Composites",
  ["tech ing sci fondam math ing"] = "Techniques de l’Ingenieur, Sciences Fondamentales: Mathematiques pour l’Ingenieur",
  ["tech ing sci fondam phys chim"] = "Techniques de l’Ingenieur, Sciences Fondamentales: Physique, Chemie",
  ["tech innov gastrointest endosc"] = "Techniques and innovations in gastrointestinal endoscopy",
  ["tech innov patient support radiat oncol"] = "Technical innovations & patient support in radiation oncology",
  ["tech man us army biol lab"] = "Technical manual. U.S. Army Biological Laboratories",
  ["tech manuscr us army biol lab"] = "Technical manuscript. U.S. Army Biological Laboratories",
  ["tech mess"] = "Technisches Messen",
  ["tech mitt"] = "Technische Mitteilungen",
  ["tech mitt aeg-telefunken"] = "Technische Mitteilungen AEG-Telefunken",
  ["tech note afmdc tn"] = "Technical note, AFMDC TN. Air Force Missile Development Center (U.S.)",
  ["tech note arct aeromed lab (us)"] = "Technical note; TN. Arctic Aeromedical Laboratory (U.S.)",
  ["tech note u s natl aeronaut space adm"] = "Technical note. United States. National Aeronautics and Space Administration",
  ["tech orthop"] = "Techniques orthopediques",
  ["tech phys"] = "Technical Physics",
  ["tech phys lett"] = "Technical Physics Letters",
  ["tech q master brew assoc am"] = "Technical Quarterly - Master Brewers Association of the Americas",
  ["tech q mbaa commun"] = "Technical Quarterly & the MBAA Communicator",
  ["tech reg anesth pain manag"] = "Techniques in regional anesthesia & pain management",
  ["tech rep"] = "Technical Report",
  ["tech rep arct aeromed lab us"] = "Technical report.; TR. Arctic Aeromedical Laboratory (U.S.)",
  ["tech rep aust mus"] = "Technical reports of the Australian Museum. Australian Museum",
  ["tech rep brookhaven natl lab"] = "T. Brookhaven National Laboratory",
  ["tech rep cp"] = "Technical report, CP. Quartermaster Research & Development Center (U.S.). Chemicals and Plastics Division",
  ["tech rep crdlr us army chem res dev lab"] = "Technical report. CRDLR. US Army Edgewood Arsenal Chemical Research and Development Laboratories",
  ["tech rep cwlr us army chem warf lab"] = "Technical report. CWLR. U.S. Army Chemical Warfare Laboratories",
  ["tech rep for eng res inst can"] = "Technical Report Forest Engineering Research Institute of Canada",
  ["tech rep hydrol water resour"] = "Technical Reports in Hydrology and Water Resources",
  ["tech rep kansai univ"] = "Kansai University",
  ["tech rep navtradevcen"] = "Technical report: NAVTRADEVCEN. Naval Training Device Center",
  ["tech rep ny nav shipyard mater lab"] = "Technical report. New York Naval Shipyard. Material Laboratory",
  ["tech rep osaka univ"] = "Technology Reports of the Osaka University",
  ["tech rep sam-tr"] = "[Technical report] SAM-TR. USAF School of Aerospace Medicine",
  ["tech rundsch"] = "Technische Rundschau",
  ["tech sanit munic"] = "La technique sanitaire et municipale",
  ["tech sci adv electron"] = "Technical Sciences: Advances in Electronics",
  ["tech shoulder elb surg"] = "Techniques in shoulder & elbow surgery",
  ["tech tips online"] = "Technical tips online",
  ["tech tworczego myslenia"] = "Techniki Tworczego Myslenia",
  ["tech uberwach"] = "Technische Uberwachung",
  ["tech urol"] = "Techniques in urology",
  ["tech vasc interv radiol"] = "Techniques in vascular and interventional radiology",
  ["technikgesch einzeldarst"] = "Technikgeschichte in Einzeldarstellungen",
  ["techniques and equipment for environmental pollution control / zhongguo ke xue yuan sheng tai huan jing yan jiu zhong xin zhu ban|huan jing wu ran zhi li ji shu yu she bei"] = "Huan jing wu ran zhi li ji shu yu she bei",
  ["technoetic arts"] = "Technoetic arts : a journal of speculative research",
  ["technol anal strateg manag"] = "Technology analysis & strategic management",
  ["technol cancer res treat"] = "Technology in cancer research & treatment",
  ["technol cult"] = "Technology and culture",
  ["technol disabil"] = "Technology and disability",
  ["technol econ smart grids sustainable energy"] = "Technology and Economics of Smart Grids and Sustainable Energy",
  ["technol eval cent asses program exec summ"] = "Technology Evaluation Center Assessment Program. Executive summary",
  ["technol eval cent assess program exec summ"] = "Technology Evaluation Center Assessment Program. Executive summary",
  ["technol forecast soc change"] = "Technological forecasting and social change",
  ["technol forecasting social change"] = "Technological Forecasting and Social Change",
  ["technol health care"] = "Technology and health care : official journal of the European Society for Engineering and Medicine",
  ["technol innov"] = "Technology and innovation",
  ["technol invest"] = "Technology and investment",
  ["technol mind behav"] = "Technology, Mind, and Behavior",
  ["technol rev"] = "Technology review",
  ["technol risk soc"] = "Technology, Risk, and Society",
  ["technol sci"] = "Technology science",
  ["technol soc"] = "Technology in society",
  ["technol sustainability"] = "Technological Sustainability",
  ["technol, knowl learn"] = "Technology, Knowledge and Learning",
  ["technologies (basel)"] = "Technologies",
  ["technology (singap world sci)"] = "Technology",
  ["tecnociencia (panama)"] = "Tecnociencia",
  ["tecnol metal mater min"] = "Tecnologia em metalurgia, materiais e mineração",
  ["tecnol metal, mater min"] = "Tecnologia em Metalurgia, Materiais e Mineracao",
  ["tecnologica map suppl"] = "Tecnologica. MAP supplement. Blue Cross and Blue Shield Association. Medical Advisory Panel",
  ["tegen tuberc"] = "Tegen de tuberculose",
  ["teion kogaku"] = "Teion kōgaku = Cryogenic engineering : [official journal of the Cryogenic Association of Japan]",
  ["tel aviver jahrb dtsch gesch"] = "Tel Aviver Jahrbuch für deutsche Geschichte",
  ["telecomm policy"] = "Telecommunications policy",
  ["telecommun radio eng"] = "Telecommunications and Radio Engineering",
  ["telecommun syst"] = "Telecommunication systems",
  ["telecommunications pol"] = "Telecommunications Policy",
  ["telemed e-health"] = "Telemedicine and e-Health",
  ["telemed j"] = "Telemedicine journal : the official journal of the American Telemedicine Association",
  ["telemed j e health"] = "Telemedicine journal and e-health : the official journal of the American Telemedicine Association",
  ["telemed rep"] = "Telemedicine reports",
  ["telemed telehealth netw"] = "Telemedicine and telehealth networks : newsmagazine of distance healthcare",
  ["telemed today"] = "Telemedicine today",
  ["telemed virtual real"] = "Telemedicine and virtual reality",
  ["telindes oper gynecol updates"] = "TeLinde's operative gynecology updates",
  ["tellus a: dyn meteorol oceanogr"] = "Tellus A: Dynamic Meteorology and Oceanography",
  ["tellus b chem phys meteorol"] = "Tellus. Series B, Chemical and physical meteorology",
  ["tellus b: chem phys meteorol"] = "Tellus B: Chemical and Physical Meteorology",
  ["tellus ser a"] = "Tellus Series A",
  ["tellus ser b"] = "Tellus Series B",
  ["telopea (syd)"] = "Telopea",
  ["tema tend mat apl comput"] = "TEMA. Tendencias em Matematica Aplicada e Computacional",
  ["temas agenda publica"] = "Temas de la agenda pública",
  ["temas odontol"] = "Temas odontológicos",
  ["temas poblac"] = "Temas de población",
  ["temida (beogr)"] = "Temida : časopis o viktimizaciji, ljudskim pravima i rodu",
  ["temperature (austin)"] = "Temperature (Austin, Tex.)",
  ["temple dent rev"] = "Temple dental review",
  ["temple int comp law j"] = "Temple international and comparative law journal",
  ["temple law q"] = "Temple law quarterly (Philadelphia, Pa. : 1946)",
  ["temple law rev"] = "Temple law review",
  ["tempo (rio j)"] = "Tempo (Rio de Janeiro, Brazil)",
  ["temps mod"] = "Les temps modernes",
  ["tenn code annot tenn"] = "Tennessee code annotated. Tennessee",
  ["tenn folk soc bull"] = "Tennessee Folklore Society bulletin",
  ["tenn hist q"] = "Tennessee historical quarterly",
  ["tenn law rev"] = "Tennessee law review",
  ["tenn med"] = "Tennessee medicine : journal of the Tennessee Medical Association",
  ["tenn nurse"] = "Tennessee nurse",
  ["tenn publ welfare rec"] = "Tennessee public welfare record",
  ["tenn stud lit"] = "Tennessee studies in literature",
  ["tenside, surfactants, deterg"] = "Tenside, Surfactants, Detergents",
  ["tensor (ns)"] = "The Tensor Society. Tensor. New Series",
  ["tenth int symp infect immunocompromised host 21 24 june 1998 davos switz int symp infect immunocompromised host 10th 1998 davos switz"] = "The Tenth International Symposium on Infections in the Immunocompromised Host, 21-24 June 1998, Davos, Switzerland. International Symposium on Infections in the Immunocompromised Host (10th : 1998 : Davos, Switzerland)",
  ["teor eksp khim"] = "Teoreticheskaya i Eksperimental’naya Khimiya",
  ["teor funktsiy funktsional anal i prilozhen"] = "Teoriya Funktsiy, Funktsional\\cprime nyy Analiz i ikh Prilozheniya",
  ["teor imovir mat stat"] = "Teoriya Imovirnostei ta Matematichna Statistika. Kiivskii Universitet imeni Tarasa Shevchenka",
  ["teor metod sistem anal"] = "\\cyr Teoriya i Metody Sistemnogo Analiza",
  ["teor model"] = "Teorie & modelli",
  ["teor osn khim tekhnol"] = "Teoreticheskie Osnovy Khimicheskoi Tekhnologii",
  ["teor prak fiz kult"] = "Teoriia i praktika fizicheskoĭ kul'tury",
  ["teor prax teles vychovy"] = "Teorie a praxe tĕlesné výchovy a sportu",
  ["teor prikl ekol"] = "Teoreticheskai︠a︡ i prikladnai︠a︡ ėkologii︠a︡",
  ["teor veroyatn primen"] = "Rossiiskaya Akademiya Nauk. Teoriya Veroyatnostei i ee Primeneniya",
  ["teor veroyatnost i primenen"] = "Rossíiskaya Akademiya Nauk",
  ["teor veroyatnost mat statist"] = "\\cyr Teoriya Veroyatnostey i Matematicheskaya Statistika",
  ["teoret elektrotekhn"] = "Teoreticheskaya Elektrotekhnika",
  ["teoret i prikl mekh (donetsk)"] = "Donetskiy Gosudarstvennyy Universitet",
  ["teoret i prikl mekh (minsk)"] = "Belorusskiy Politekhnicheskiy Institut",
  ["teoret mat fiz"] = "Rossiiskaya Akademiya Nauk. Teoreticheskaya i Matematicheskaya Fizika",
  ["teoria (ns)"] = "Teoria",
  ["teploenergetika (moscow)"] = "Teploenergetika (Moscow)",
  ["teplofiz vys temp"] = "Teplofizika Vysokikh Temperatur",
  ["teplofiz vys temp [high temp (ussr)]"] = "Teplofizika Vysokikh Temperatur [High Temperature (USSR)]",
  ["ter arkh"] = "Terapevticheskiĭ arkhiv",
  ["terahertz sci technol"] = "Terahertz Science and Technology",
  ["terapia (quito)"] = "Terapia",
  ["teratog carcinog mutagen"] = "Teratogenesis, carcinogenesis, and mutagenesis",
  ["teratog, carcinog, mutagen"] = "Teratogenesis, Carcinogenesis, and Mutagenesis",
  ["terr arthropod rev"] = "Terrestrial arthropod reviews",
  ["terr incogn"] = "Terrae incognitae",
  ["terr magn"] = "Terrestrial Magnetism",
  ["terr magn atmos electr"] = "Terrestrial Magnetism and Atmospheric Electricity",
  ["terr, oceanic atmos sci"] = "Terrestrial, Atmospheric and Oceanic Sciences",
  ["terra (hels)"] = "Terra (Helsinki, Finland)",
  ["terra ameriga"] = "Terra Ameriga",
  ["terra latinoam"] = "Terra Latinoamericana : organo científico de la Sociedad Mexicana de la Ciencia del Suelo, A.C",
  ["terra nova"] = "Terra Nova",
  ["terramycine inf"] = "Terramycine informations",
  ["terre vie"] = "Terre et la Vie, La",
  ["territ politic gov"] = "Territory, politics, governance",
  ["tesi perfez"] = "Tesi di Perfezionamento",
  ["tesi sc norm super pisa (n s)"] = "Tesi. Scuola Normale Superiore di Pisa (Nuova Series)",
  ["test (madr)"] = "Test (Madrid, Spain)",
  ["tetrahedron asymmetry"] = "Tetrahedron, asymmetry",
  ["tetrahedron chem"] = "Tetrahedron Chem",
  ["tetrahedron green chem"] = "Tetrahedron Green Chem",
  ["tetrahedron lett"] = "Tetrahedron Letters",
  ["tetrahedron: asymmetry"] = "Tetrahedron: Asymmetry",
  ["tetsu to hagane"] = "Tetsu to Hagane",
  ["teubner skr math stochastik"] = "Teubner Skripten zur Mathematischen Stochastik",
  ["teubner skr numer"] = "Teubner Skripten zur Numerik",
  ["teubner studbuch wirtschwiss"] = "Teubner Studienbucher Wirtschaftswissenschaften",
  ["teubner studienbuch phys/chem"] = "Teubner Studienbucher Physik/Chemie",
  ["teubner studienbüch math"] = "Teubner Studienbücher Mathematik",
  ["teubner studienskr math ingen"] = "Teubner Studienskripten: Mathematik, Ingenieurwissenschaften",
  ["teubner-arch math"] = "Teubner-Archiv zur Mathematik",
  ["teubner-arch math suppl"] = "Teubner-Archiv zur Mathematik",
  ["teubner-texte inform"] = "Teubner-Texte zur Informatik",
  ["teubner-texte math"] = "Teubner-Texte zur Mathematik",
  ["tex cancer bull"] = "Texas cancer bulletin",
  ["tex dent assist assoc bull"] = "Texas Dental Assistants Association bulletin",
  ["tex dent j"] = "Texas dental journal",
  ["tex educ rev (austin)"] = "Texas education review (Austin, Tex.)",
  ["tex heart inst j"] = "Texas Heart Institute journal",
  ["tex hosp"] = "Texas hospitals",
  ["tex int law j"] = "Texas international law journal",
  ["tex j health syst pharm"] = "Texas journal of health-system pharmacy",
  ["tex j pharm"] = "Texas journal of pharmacy",
  ["tex j sci"] = "The Texas journal of science",
  ["tex j women law"] = "Texas journal of women and the law",
  ["tex law rev"] = "Texas law review",
  ["tex med"] = "Texas medicine",
  ["tex med j (austin)"] = "Texas medical journal (Austin, Tex.)",
  ["tex mon"] = "Texas monthly (Austin, Tex.)",
  ["tex nurs"] = "Texas nursing",
  ["tex public health j"] = "Texas public health journal",
  ["tex rep biol med"] = "Texas reports on biology and medicine",
  ["tex rev law polit"] = "Texas review of law & politics",
  ["tex state j med"] = "Texas state journal of medicine",
  ["tex tech law rev"] = "Texas Tech law review",
  ["texas bar j"] = "Texas bar journal",
  ["texas j sci"] = "The Texas Journal of Science",
  ["texas stud lit lang"] = "Texas studies in literature and language",
  ["text chem color"] = "Textile Chemist and Colorist",
  ["text chem color am dyest rep"] = "Textile Chemist and Colorist & American Dyestuff Reporter",
  ["text clothing sustainability"] = "Textiles and Clothing Sustainability",
  ["text hist"] = "Textile history",
  ["text prog"] = "Textile Progress",
  ["text rent"] = "Textile rental",
  ["text res j"] = "Textile Research Journal",
  ["text sci tech"] = "Textile Science and Technology",
  ["text speech dialog"] = "Text, speech and dialogue. TSD",
  ["text world"] = "Textile World",
  ["text: j cloth cult"] = "Textile: The Journal of Cloth and Culture",
  ["textb electr electron eng"] = "Textbooks in Electrical and Electronic Engineering",
  ["textb math"] = "Textbooks in Mathematics",
  ["textb ser phys sci"] = "Textbook Series in Physical Sciences",
  ["textbooks math sci"] = "Textbooks in Mathematical Sciences",
  ["texte didakt math"] = "Texte zur Didaktik der Mathematik",
  ["texte gesch praventivmed"] = "Texte zur Geschichte der Präventivmedizin : TGP",
  ["textes etudes moyen age"] = "Textes et Etudes du Moyen Age",
  ["textos bronce"] = "Textos del Bronce",
  ["textos man ensen"] = "Textos y Manuales de Ensenanza",
  ["textos mat"] = "Textos de Matematica",
  ["textos mat sér b"] = "Textos de Matemática",
  ["textos mat/math texts"] = "Textos de Matematica/Mathematics Texts",
  ["textos nepo"] = "Textos NEPO",
  ["texts algorithmics"] = "Texts in Algorithmics",
  ["texts appl math"] = "Texts in Applied Mathematics",
  ["texts comput"] = "Texts in Computing",
  ["texts comput sci"] = "Texts in Computer Science",
  ["texts comput sci eng"] = "Texts in Computational Science and Engineering",
  ["texts log reason"] = "Texts in Logic and Reasoning",
  ["texts math"] = "Texts in Mathematics",
  ["texts monogr phys"] = "Texts and Monographs in Physics",
  ["texts monogr symbol comput"] = "Texts and Monographs in Symbolic Computation",
  ["texts monogr theor phys"] = "Texts and Monographs in Theoretical Physics",
  ["texts philos"] = "Texts in Philosophy",
  ["texts quant crit think"] = "Texts for Quantitative Critical Thinking",
  ["texts read math"] = "Texts and Readings in Mathematics",
  ["texts read phys sci"] = "Texts and Readings in Physical Sciences",
  ["texts statist sci ser"] = "Texts in Statistical Science Series",
  ["texts theoret comput sci eatcs ser"] = "Texts in Theoretical Computer Science. An EATCS Series",
  ["textual pract"] = "Textual practice",
  ["teyou shengwu yanjiu"] = "Tèyou shengwù yánjiu",
  ["tezisy dokl mezhrespub konf"] = "Tezisy dokladov. Mezhrespublikanskaia konferentsiia po voprosam istorii estestvoznaniia i tekhniki v Pribaltike",
  ["th open"] = "TH open : companion journal to thrombosis and haemostasis",
  ["thai for bull"] = "Thai forest bulletin",
  ["thai j agric sci"] = "Thai journal of agricultural science",
  ["thai j genet"] = "Thai journal of genetics",
  ["thai j math"] = "Thai Journal of Mathematics",
  ["thai j nurs"] = "Thai journal of nursing",
  ["thai j nurs res"] = "Thai journal of nursing research",
  ["thai j pharm sci"] = "Thai Journal of Pharmaceutical Sciences",
  ["thai j sci technol"] = "Thai journal of science and technology",
  ["thaiszia (kosice)"] = "Thaiszia",
  ["thalamus relat syst"] = "Thalamus & related systems",
  ["thalassas int j mar sci"] = "Thalassas: International Journal of Marine Sciences",
  ["thalassia salentina"] = "Thalassia Salentina",
  ["thammasat int j sci tech"] = "Thammasat international journal of science and technology : TIJSAT",
  ["thang long j sci math math sci"] = "Thang Long Journal of Science. Mathematics and Mathematical Sciences",
  ["the auk: ornithol adv"] = "The Auk: Ornithological Advances",
  ["the chinese journal of clinical pharmacology / [bian ji, zhongguo lin chuang yao li xue za zhi bian ji bu]|zhongguo lin chuang yao li xue za zhi"] = "Zhongguo lin chuang yao li xue za zhi",
  ["the condor: ornithol appl"] = "The Condor: Ornithological Applications",
  ["the journal of scientific research / faculty of science, chulalongkorn university|j sci res chulalongkorn univ"] = "Warsan wichai witthayasat / Khana Witthayasat Chulalongkon Mahawitthayalai",
  ["the journal of sericultural science|nihon sanshigaku zasshi"] = "Nihon sanshigaku zasshi",
  ["the journal of the korean society for microbiology|taehan misaengmul hakhoe chi"] = "Taehan Misaengmul Hakhoe chi",
  ["the kasetsart journal|witthayasan kasetsat"] = "Witthayasan Kasetsat",
  ["the microorganisms and industry|misaengmul gwa san-eob"] = "Misaengmul kwa sanop",
  ["theatre surv"] = "Theatre survey",
  ["themes econ"] = "Themes in Economics",
  ["themes mod econom"] = "Themes in Modern Econometrics",
  ["theol bibl topelmann"] = "Theologische Bibliothek Topelmann",
  ["theol dig"] = "Theology digest",
  ["theol stud"] = "Theological studies",
  ["theol sz"] = "Theologiai szemle",
  ["theol today"] = "Theology today (Princeton, N.J.)",
  ["theor appl climatol"] = "Theoretical and Applied Climatology",
  ["theor appl fract mech"] = "Theoretical and Applied Fracture Mechanics",
  ["theor appl genet"] = "TAG. Theoretical and applied genetics. Theoretische und angewandte Genetik",
  ["theor appl mech lett"] = "Theoretical and Applied Mechanics Letters",
  ["theor biol"] = "Theoretical Biology",
  ["theor biol forum"] = "Theoretical biology forum",
  ["theor biol med model"] = "Theoretical biology & medical modelling",
  ["theor biol med modell"] = "Theoretical Biology and Medical Modelling",
  ["theor chem acc"] = "Theoretical Chemistry Accounts",
  ["theor chim acta"] = "Theoretica Chimica Acta",
  ["theor comput fluid dyn"] = "Theoretical and Computational Fluid Dynamics",
  ["theor comput sci"] = "Theoretical computer science",
  ["theor criminol"] = "Theoretical criminology",
  ["theor ecol"] = "Theoretical ecology",
  ["theor econ"] = "Theoretical Economics",
  ["theor exp chem"] = "Theoretical and Experimental Chemistry",
  ["theor exp plant physiol"] = "Theoretical and experimental plant physiology",
  ["theor found chem eng"] = "Theoretical Foundations of Chemical Engineering",
  ["theor issues ergon sci"] = "Theoretical issues in ergonomics science",
  ["theor linguist"] = "Theoretical linguistics",
  ["theor math phys"] = "Theoretical and Mathematical Physics",
  ["theor med"] = "Theoretical medicine",
  ["theor med bioeth"] = "Theoretical medicine and bioethics",
  ["theor med bioethics"] = "Theoretical Medicine and Bioethics",
  ["theor popul biol"] = "Theoretical population biology",
  ["theor prax korperkult"] = "Theorie und Praxis der Körperkultur",
  ["theor surg"] = "Theoretical surgery",
  ["theoret and math phys"] = "Theoretical and Mathematical Physics",
  ["theoret appl mech"] = "Theoretical and Applied Mechanics",
  ["theoret comput sci"] = "Theoretical Computer Science",
  ["theoret linguist"] = "Theoretical Linguistics",
  ["theoret math phys"] = "Theoretical and Mathematical Physics",
  ["theoria (san sebastian) (2)"] = "Theoria. Revista de Teoria, Historia y Fundamentos de la Ciencia. Segunda Epoca. An International Journal for Theory, History and Foundations of Science. Second Series",
  ["theoria (san sebastián) (2)"] = "Theoria",
  ["theoria hist sci"] = "Theoria et Historia Scientiarum",
  ["theory and decision"] = "Theory and Decision. An International Journal for Multidisciplinary Advances in Decision Science",
  ["theory appl categ"] = "Theory and Applications of Categories",
  ["theory appl comput"] = "Theory and Applications of Computability",
  ["theory appl graphs"] = "Theory and Applications of Graphs",
  ["theory appl math comput sci"] = "Theory and Applications of Mathematics & Computer Science",
  ["theory appl nat lang process"] = "Theory and Applications of Natural Language Processing",
  ["theory appl transp porous media"] = "Theory and Applications of Transport in Porous Media",
  ["theory biosci"] = "Theory in biosciences = Theorie in den Biowissenschaften",
  ["theory comput"] = "Theory of Computing. An Open Access Journal",
  ["theory comput syst"] = "Theory of Computing Systems",
  ["theory cult soc"] = "Theory, culture & society",
  ["theory dec"] = "Theory and Decision",
  ["theory decis"] = "Theory and decision",
  ["theory decis lib ser b math statist methods"] = "Theory and Decision Library",
  ["theory decis lib ser d system theory knowledge eng probl solving"] = "Theory and Decision Library",
  ["theory decis libr c, game theory soc choice decis theory optim"] = "Theory and Decision Library C. Game Theory, Social Choice, Decision Theory, and Optimization",
  ["theory decis libr ser c game theory math program oper res"] = "Theory and Decision Library. Series C: Game Theory, Mathematical Programming and Operations Research",
  ["theory eng appl comput methods"] = "Theory and Engineering Applications of Computational Methods",
  ["theory pract"] = "Theory into practice",
  ["theory pract log program"] = "Theory and Practice of Logic Programming",
  ["theory pract logic program"] = "Theory and Practice of Logic Programming",
  ["theory probab appl"] = "Theory of Probability and its Applications",
  ["theory probab math statist"] = "Theory of Probability and Mathematical Statistics",
  ["theory psychol"] = "Theory & psychology",
  ["theory res educ"] = "Theory and research in education : TRE",
  ["theory res soc educ"] = "Theory and research in social education",
  ["theory soc"] = "Theory and society",
  ["theory stoch process"] = "Theory of Stochastic Processes",
  ["ther adv cardiovasc dis"] = "Therapeutic advances in cardiovascular disease",
  ["ther adv chronic dis"] = "Therapeutic advances in chronic disease",
  ["ther adv drug saf"] = "Therapeutic advances in drug safety",
  ["ther adv endocrinol metab"] = "Therapeutic advances in endocrinology and metabolism",
  ["ther adv gastrointest endosc"] = "Therapeutic advances in gastrointestinal endoscopy",
  ["ther adv hematol"] = "Therapeutic advances in hematology",
  ["ther adv infect dis"] = "Therapeutic advances in infectious disease",
  ["ther adv med oncol"] = "Therapeutic advances in medical oncology",
  ["ther adv musculoskelet dis"] = "Therapeutic advances in musculoskeletal disease",
  ["ther adv neurol disord"] = "Therapeutic advances in neurological disorders",
  ["ther adv ophthalmol"] = "Therapeutic advances in ophthalmology",
  ["ther adv psychopharmacol"] = "Therapeutic advances in psychopharmacology",
  ["ther adv reprod health"] = "Therapeutic advances in reproductive health",
  ["ther adv respir dis"] = "Therapeutic advances in respiratory disease",
  ["ther adv urol"] = "Therapeutic advances in urology",
  ["ther adv vaccines"] = "Therapeutic advances in vaccines",
  ["ther adv vaccines immunother"] = "Therapeutic advances in vaccines and immunotherapy",
  ["ther apher"] = "Therapeutic apheresis : official journal of the International Society for Apheresis and the Japanese Society for Apheresis",
  ["ther apher dial"] = "Therapeutic apheresis and dialysis : official peer-reviewed journal of the International Society for Apheresis, the Japanese Society for Apheresis, the Japanese Society for Dialysis Therapy",
  ["ther ber"] = "Therapeutische Berichte",
  ["ther clin risk manag"] = "Therapeutics and clinical risk management",
  ["ther clin risk manage"] = "Therapeutics and Clinical Risk Management",
  ["ther communities"] = "Therapeutic communities",
  ["ther deliv"] = "Therapeutic delivery",
  ["ther delivery"] = "Therapeutic Delivery",
  ["ther drug monit"] = "Therapeutic drug monitoring",
  ["ther ggw"] = "Therapie der Gegenwart",
  ["ther hung"] = "Therapia Hungarica",
  ["ther hypothermia temp manag"] = "Therapeutic hypothermia and temperature management",
  ["ther hypothermia temp manage"] = "Therapeutic Hypothermia and Temperature Management",
  ["ther immunol"] = "Therapeutic immunology",
  ["ther innov regul sci"] = "Therapeutic innovation & regulatory science",
  ["ther innovation regul sci"] = "Therapeutic Innovation & Regulatory Science",
  ["ther notes"] = "Therapeutic notes",
  ["ther nova"] = "Therapeutica nova",
  ["ther patient educ"] = "Therapeutic patient education",
  ["ther probl today"] = "Therapeutic problems of today",
  ["ther radiol oncol"] = "Therapeutic radiology and oncology",
  ["ther recreation j"] = "Therapeutic recreation journal",
  ["ther targets neurol dis"] = "Therapeutic targets for neurological diseases",
  ["ther umsch"] = "Therapeutische Umschau. Revue thérapeutique",
  ["ther umsch med bibliogr"] = "Therapeutische Umschau und medizinische Bibliographie. Revue thérapeutique et bibliographie médicale",
  ["therap adv gastroenterol"] = "Therapeutic advances in gastroenterology",
  ["therm eng"] = "Thermal Engineering",
  ["therm sci eng prog"] = "Thermal Science and Engineering Progress",
  ["thermochim acta"] = "Thermochimica Acta",
  ["thermophys aeromech"] = "Thermophysics and Aeromechanics",
  ["theta ser adv math"] = "Theta Series in Advanced Mathematics",
  ["thin solid films"] = "Thin Solid Films",
  ["thin walled struct"] = "Thin Walled Structures",
  ["thin-walled struct"] = "Thin-Walled Structures",
  ["think (lond)"] = "Think (London, England)",
  ["think reason"] = "Thinking & reasoning",
  ["think skills creat"] = "Thinking skills and creativity",
  ["third ieee int conf big data secur cloud ieee bigdatasecur 2017 (2017)"] = "The Third IEEE International Conference on Big Data Security on Cloud, IEEE BigDataSecurity 2017 : The Third IEEE International Conference on High Performance and Smart Computing, IEEE HPSC 2017 ; the Second IEEE International Conference on Intelligent Data and Security, IEEE IDS 2017 : proceedings : 26-28 May 2017, Beijing, China. IEEE International Conference on Big Data Security on Cloud (3rd : 2017 : Beijing, China)",
  ["third repub"] = "Third Republic = Troisième République",
  ["third world plann rev"] = "Third world planning review",
  ["third world q"] = "Third world quarterly",
  ["thomas jefferson law rev"] = "Thomas Jefferson law review",
  ["thorac cancer"] = "Thoracic cancer",
  ["thorac cardiovasc surg"] = "The Thoracic and cardiovascular surgeon",
  ["thorac cardiovasc surg rep"] = "The Thoracic and cardiovascular surgeon reports",
  ["thorac surg clin"] = "Thoracic surgery clinics",
  ["thorac surg sci"] = "Thoracic surgical science",
  ["thoraxchir vask chir"] = "Thoraxchirurgie und vaskuläre Chirurgie",
  ["thought (hoboken)"] = "Thought (Hoboken, N.J.)",
  ["thromb diath haemorrh"] = "Thrombosis et diathesis haemorrhagica",
  ["thromb diath haemorrh suppl"] = "Thrombosis et diathesis haemorrhagica. Supplementum",
  ["thromb haemost"] = "Thrombosis and haemostasis",
  ["thromb j"] = "Thrombosis journal",
  ["thromb res"] = "Thrombosis research",
  ["thromb res suppl"] = "Thrombosis research. Supplement",
  ["thurg jahrb"] = "Thurgauer Jahrbuch",
  ["thurgood marshall law rev"] = "Thurgood Marshall law review",
  ["thymos (harriman)"] = "Thymos (Harriman, Tenn.)",
  ["thyroid res"] = "Thyroid research",
  ["ti erh chun i ta hsueh hsueh pao"] = "Di 2 jun yi da xue xue bao = Dier junyi daxue xuebao = Academic journal of Second Military Medical College",
  ["tianjin daxue xuebao"] = "Tianjin da xue xue bao = Journal of Tianjin University",
  ["tianjin yi ke da xue xue bao"] = "Tianjin yi ke da xue xue bao = Journal of Tianjin Medical University",
  ["tianjin yi yao"] = "Tianjin yi yao = Tianjin medical journal",
  ["tianran chanwu yanjiu yu kaifa"] = "Tianran Chanwu Yanjiu Yu Kaifa",
  ["tibet news rev"] = "Tibet news review",
  ["ticks tick borne dis"] = "Ticks and tick-borne diseases",
  ["ticks tick-borne dis"] = "Ticks and Tick-borne Diseases",
  ["tid rep"] = "TID [reports]. U.S. Atomic Energy Commission",
  ["tid tann"] = "Tidens tann",
  ["tidskr mil halsov"] = "Tidskrift i militär hälsovård",
  ["tidskr sjukvardspedagog"] = "Tidskrift för sjukvårdspedagoger",
  ["tidskr sver sjukskot"] = "Tidskrift för Sveriges sjuksköterskor",
  ["tidsskr kjemi bergv metall"] = "Tidsskrift for Kjemi, Bergvesen, og Metallurgi",
  ["tidsskr nor laegeforen"] = "Tidsskrift for den Norske lægeforening : tidsskrift for praktisk medicin, ny række",
  ["tidsskr prakt tandlaeg"] = "Tidsskrift for praktiserende tandlaeger",
  ["tidsskr rettsvitenskap"] = "Tidsskrift for rettsvidenskap",
  ["tidsskr samfunnsforsk"] = "Tidsskrift for samfunnsforskning",
  ["tidsskr skogbruk"] = "Tidsskrift for Skogbruk",
  ["tidsskr sygepl"] = "Tidsskrift for sygeplejersker",
  ["tidsskr sykepl"] = "Tidsskriftet sykepleien",
  ["tidsskr tandlaeger"] = "Tidsskrift for tandlaeger (Copenhagen, Denmark : 1981)",
  ["tierarztl prax"] = "Tierarztliche Praxis",
  ["tierarztl prax ausg g grosstiere nutztiere"] = "Tierärztliche Praxis. Ausgabe G, Grosstiere/Nutztiere",
  ["tierarztl prax ausg k kleintiere heimtiere"] = "Tierärztliche Praxis. Ausgabe K, Kleintiere/Heimtiere",
  ["tierarztl prax ausg k klientiere heimtiere"] = "Tierarztliche Praxis. Ausgabe K, Kleintiere/Heimtiere",
  ["tierarztl prax suppl"] = "Tierarztliche Praxis. Supplement",
  ["tierarztl umsch"] = "Tierarztliche Umschau",
  ["tiers monde"] = "Revue tiers-monde",
  ["tijd klei glas keram"] = "Tijdschrift vor Klei, Glas en Keramiek",
  ["tijdschr bedr verzekeringsgeneeskd"] = "Tijdschrift voor bedrijfs- en verzekeringsgeneeskunde",
  ["tijdschr bejaarden kraam ziekenverzorging"] = "Tijdschrift voor bejaarden-, kraam- en ziekenversorging : bkz",
  ["tijdschr diergeneeskd"] = "Tijdschrift voor diergeneeskunde",
  ["tijdschr econ"] = "Tijdschrift voor Economie",
  ["tijdschr econ manage"] = "Tijdschrift voor Economie en Management",
  ["tijdschr econ soc geogr"] = "Tijdschrift voor economische en sociale geografie = Journal of economic and social geography = Revue de géographie économique et humaine = Zeitschrift für ökonomische und soziale Geographie = Revista de geografía económica y social",
  ["tijdschr entomol"] = "Tijdschrift voor entomologie",
  ["tijdschr filos"] = "Tijdschrift voor philosophie",
  ["tijdschr gastroenterol"] = "Tijdschrift voor gastro-enterologie",
  ["tijdschr geneeskd"] = "Tijdschrift voor geneeskunde",
  ["tijdschr gerontol geriatr"] = "Tijdschrift voor gerontologie en geriatrie",
  ["tijdschr geschied"] = "Tijdschrift voor geschiedenis (1920)",
  ["tijdschr geschied geneeskd natuurwet wiskd tech"] = "Tijdschrift voor de geschiedenis der geneeskunde, natuurwetenschappen, wiskunde en techniek",
  ["tijdschr gezondh ethiek"] = "Tijdschrift voor gezondheidszorg & ethiek : TGE",
  ["tijdschr kindergeneeskd"] = "Tijdschrift voor kindergeneeskunde",
  ["tijdschr nucl geneeskd"] = "Tijdschrift voor nu[c]leaire geneeskunde",
  ["tijdschr psychiatr"] = "Tijdschrift voor psychiatrie",
  ["tijdschr soc geneeskd"] = "Tijdschrift voor sociale geneeskunde",
  ["tijdschr soc geschied"] = "Tijdschrift voor sociale geschiedenis",
  ["tijdschr soc gezondheidsz"] = "Tijdschrift voor sociale gezondheidszorg : TSG : 14-daags blad van de Algemene Nederlandse Vereniging voor Sociale Gezondheidszorg",
  ["tijdschr soc wet"] = "Tijdschrift voor sociale wetenschappen",
  ["tijdschr sociol"] = "Tijdschrift voor sociologie",
  ["tijdschr tandheelkd"] = "Tijdschrift Voor Tandheelkunde",
  ["tijdschr verzorg"] = "Tijdschrift voor verzorgenden",
  ["tijdschr voor tandheelkd"] = "Tijdschrift Voor Tandheelkunde",
  ["tijdschr voor ziekenverpl"] = "Tijdschrift voor ziekenverpleging",
  ["tijdschr ziekenverpl"] = "Tijdschrift voor ziekenverpleging",
  ["tijeretazos malar"] = "Tijeretazos sobre malaria",
  ["tile till"] = "Tile and till",
  ["timarit hjukrunarfel isl"] = "Timarit Hjukrunarfelags Islands",
  ["timber bull eur"] = "Timber Bulletin for Europe",
  ["time soc"] = "Time & society",
  ["timely top med cardiovasc dis"] = "Timely topics in medicine. Cardiovascular diseases",
  ["times (lond)"] = "Times (London, England : 1788)",
  ["timing time percept"] = "Timing & time perception (Leiden, Netherlands)",
  ["timisoara med"] = "Timişoara medicală",
  ["tip fak mecm"] = "Tip Fakültesi mecmuasi",
  ["tip tarihi arastirmalari"] = "Tıp tarihi araştırmaları = History of medicine studies",
  ["tire sci technol"] = "Tire Science and Technology",
  ["tirol heim"] = "Tiroler Heimat",
  ["tirol heimatbl"] = "Tiroler heimatblätter : Zeitschrift für Geschichte, Natur und Volkskunde",
  ["tissue antigens"] = "Tissue Antigens",
  ["tissue barriers"] = "Tissue barriers",
  ["tissue cell"] = "Tissue & cell",
  ["tissue eng"] = "Tissue Engineering",
  ["tissue eng part a"] = "Tissue engineering. Part A",
  ["tissue eng part b"] = "Tissue Engineering Part B",
  ["tissue eng part b rev"] = "Tissue engineering. Part B, Reviews",
  ["tissue eng part c"] = "Tissue Engineering Part C",
  ["tissue eng part c methods"] = "Tissue engineering. Part C, Methods",
  ["tissue eng regen med"] = "Tissue engineering and regenerative medicine",
  ["tissue eng regener med"] = "Tissue Engineering and Regenerative Medicine",
  ["tit j life sci"] = "T.-I.-T. journal of life sciences",
  ["tm - tech mess"] = "tm - Technisches Messen",
  ["tmj update"] = "TMJ update",
  ["tob control"] = "Tobacco control",
  ["tob induc dis"] = "Tobacco induced diseases",
  ["tob induced dis"] = "Tobacco Induced Diseases",
  ["tob prev cessat"] = "Tobacco prevention & cessation",
  ["tob regul sci"] = "Tobacco regulatory science",
  ["tob use insights"] = "Tobacco use insights",
  ["todays cathol teach"] = "Today's Catholic teacher",
  ["todays christ dr"] = "Today's Christian doctor : the journal of the Christian Medical & Dental Society",
  ["todays fda"] = "Today's FDA : official monthly journal of the Florida Dental Association",
  ["todays health"] = "Today's health",
  ["todays nurs home"] = "Today's nursing home",
  ["todays or nurse"] = "Today's OR nurse",
  ["todays surg nurse"] = "Today's surgical nurse",
  ["today’s engineer"] = "Today’s Engineer",
  ["todo soc"] = "Todo social",
  ["toho igakkai zasshi"] = "Toho Igakkai zasshi. Journal of the Medical Society of Toho University",
  ["tohoku igaku zasshi"] = "Tohoku igaku zasshi",
  ["tohoku j exp med"] = "The Tohoku journal of experimental medicine",
  ["tohoku math j"] = "Tohoku Mathematical Journal",
  ["tohoku math j (2)"] = "The Tohoku Mathematical Journal. Second Series",
  ["tohoku math publ"] = "Tohoku Mathematical Publications",
  ["tohoku shika daigaku gakkai shi"] = "Tōhoku Shika Daigaku Gakkai shi",
  ["tokai j exp clin med"] = "The Tokai journal of experimental and clinical medicine",
  ["tokoginecol pract"] = "Toko-ginecología práctica",
  ["tokushima j exp med"] = "The Tokushima journal of experimental medicine",
  ["tokyo igaku"] = "Tokyo igaku. The Tokyo journal of medical sciences",
  ["tokyo igaku zasshi"] = "Tokyo igaku zasshi. The Tokyo journal of medical sciences",
  ["tokyo iji shinshi"] = "[Tokyo iji shinshi] The Tokyo medical journal",
  ["tokyo ika daigaku zasshi"] = "Tokyo Ika Daigaku Zasshi (Journal of Tokyo Medical College)",
  ["tokyo ika shika daigaku iyo kizai kenkyusho hokoku"] = "Iyō Kizai Kenkyūjo hōkoku. Reports of the Institute for Medical and Dental Engineering, Tokyo Medical and Dental University",
  ["tokyo j math"] = "Tokyo Journal of Mathematics",
  ["tokyo jikeikai ika daigaku zasshi"] = "[Zasshi] Tokyo Jikeikai medical journal. Tōkyō Jikeikai Ika Daigaku",
  ["tokyo nogyo daigaku nogaku shuho"] = "Tōkyō Nōgyō Daigaku nōgaku shūhō",
  ["tomogr mater struct"] = "Tomography of Materials and Structures",
  ["tomorrows technol transf"] = "Tomorrow's technology transfer",
  ["tonan ajia kenkyu"] = "Tōnan Ajia kenkyū",
  ["tong ji yi ke da xue xue bao"] = "Tong ji yi ke da xue xue bao = Acta Universitatis Medicinae Tongji",
  ["tonga munhwa"] = "Tonga munhwa",
  ["tongji daxue xuebao ziran kexue ban"] = "Tongji Daxue Xuebao",
  ["tongmul hakhoe chi"] = "Tongmul Hakhoe chi",
  ["tonkie khim tekhnol"] = "Tonkie Khimicheskie Tekhnologii",
  ["tools algorithms constr anal syst i (2020)"] = "Tools and algorithms for the construction and analysis of systems : 26th International Conference, TACAS 2020, held as part of the European Joint Conferences on Theory and Practice of Software, ETAPS 2020, Dublin, Ireland, April 25-30, 2020, Proceedings. Part I. TACAS (Conference) (26th : 2020 : Dublin, Ireland)",
  ["top adv geoinformatics"] = "Topics in Advanced Geoinformatics",
  ["top anticancer res"] = "Topics in anti-cancer research",
  ["top antivir med"] = "Topics in antiviral medicine",
  ["top appl phys"] = "Topics in Applied Physics",
  ["top biodivers conserv"] = "Topics in Biodiversity and Conservation",
  ["top biol inorg chem"] = "Topics in Biological and Inorganic Chemistry",
  ["top catal"] = "Topics in Catalysis",
  ["top clin nurs"] = "Topics in clinical nursing",
  ["top clin nutr"] = "Topics in clinical nutrition",
  ["top cogn sci"] = "Topics in cognitive science",
  ["top companion anim med"] = "Topics in companion animal medicine",
  ["top curr chem"] = "Topics in Current Chemistry",
  ["top curr chem (cham)"] = "Topics in current chemistry (Cham)",
  ["top curr genet"] = "Topics in current genetics",
  ["top econ anal policy"] = "Topics in economic analysis & policy",
  ["top emerg med"] = "Topics in emergency medicine",
  ["top fluoresc spectrosc"] = "Topics in Fluorescence Spectroscopy",
  ["top geriatr rehabil"] = "Topics in geriatric rehabilitation",
  ["top health care financ"] = "Topics in health care financing",
  ["top health inf manage"] = "Topics in health information management",
  ["top health rec manage"] = "Topics in health record management",
  ["top heterocycl chem"] = "Topics in heterocyclic chemistry",
  ["top hiv med"] = "Topics in HIV medicine : a publication of the International AIDS Society, USA",
  ["top hosp law"] = "Topics in hospital law",
  ["top hosp pharm manage"] = "Topics in hospital pharmacy management",
  ["top inorg chem"] = "Topics in Inorganic Chemistry",
  ["top intell eng inform"] = "Topics in Intelligent Engineering and Informatics",
  ["top lang disord"] = "Topics in language disorders",
  ["top magn reson imaging"] = "Topics in magnetic resonance imaging : TMRI",
  ["top organomet chem"] = "Topics in Organometallic Chemistry",
  ["top phys chem"] = "Topics in Physical Chemistry",
  ["top probl psychother"] = "Topical problems of psychotherapy",
  ["top saf risk reliab qual"] = "Topics in Safety, Risk, Reliability and Quality",
  ["top spinal cord inj rehabil"] = "Topics in spinal cord injury rehabilitation",
  ["top stereochem"] = "Topics in stereochemistry",
  ["top stroke rehabil"] = "Topics in stroke rehabilitation",
  ["topic (wash)"] = "Topic",
  ["topics appl phys"] = "Topics in Applied Physics",
  ["topics discrete math"] = "Topics in Discrete Mathematics",
  ["topics early child spec educ"] = "Topics in early childhood special education",
  ["topics eng"] = "Topics in Engineering",
  ["topics in comput math"] = "Topics in Computer Mathematics",
  ["topics photogr preserv"] = "Topics in photographic preservation",
  ["toplum bilim"] = "Toplum ve bilim",
  ["topoi (dordr)"] = "Topoi : an international review of philosophy",
  ["topol algebra appl"] = "Topological Algebra and its Applications",
  ["topol appl"] = "Topology and its applications",
  ["topol methods nonlinear anal"] = "Topological Methods in Nonlinear Analysis",
  ["topology appl"] = "Topology and its Applications",
  ["topology proc"] = "Topology Proceedings",
  ["tor j theol"] = "Toronto journal of theology",
  ["tor star"] = "Toronto star",
  ["torah u madda j"] = "The Torah u-madda journal",
  ["torreon med"] = "Torreon médico",
  ["tort insur law j"] = "Tort & insurance law journal",
  ["tort stat tanulm"] = "Történeti statisztikai tanulmányok",
  ["tort trial insur pract law j"] = "Tort trial & insurance practice law journal",
  ["tortenelmi sz"] = "Történelmi szemle",
  ["total patient care hiv hcv"] = "Total patient care in HIV & HCV",
  ["tottels j prof neglig"] = "Tottel's journal of professional negligence",
  ["tottori rinsho kagaku kenkyukai shi"] = "Tottori Rinshō Kagaku Kenkyūkai shi",
  ["touchstone (nashv)"] = "Touchstone (Nashville, Tenn.)",
  ["toulouse med"] = "Toulouse médical",
  ["toung pao"] = "Tʻung pao. Tʻoung pao",
  ["tour manag"] = "Tourism management",
  ["tour manag perspect"] = "Tourism management perspectives",
  ["tour stud"] = "Tourist studies",
  ["tourbe philos"] = "La tourbe des philosophes : revue d'études alchimiques",
  ["tourism econ"] = "Tourism Economics",
  ["tourism hospitality manage"] = "Tourism and Hospitality Management",
  ["touro law rev"] = "Touro law review",
  ["town ctry plann"] = "Town and country planning : the quarterly review of the Town and Country Planning Association",
  ["town hall j"] = "Town hall journal",
  ["town plan rev"] = "The Town planning review",
  ["toxic rep ser"] = "Toxicity report series",
  ["toxic subst mech"] = "Toxic substance mechanisms",
  ["toxicol anal clin"] = "Toxicologie Analytique et Clinique",
  ["toxicol appl pharmacol"] = "Toxicology and Applied Pharmacology",
  ["toxicol commun"] = "Toxicology communications",
  ["toxicol environ chem"] = "Toxicological and Environmental Chemistry",
  ["toxicol environ health sci"] = "Toxicology and environmental health sciences",
  ["toxicol eur res"] = "Toxicological European research. Recherche européenne en toxicologie",
  ["toxicol forensic med"] = "Toxicology and forensic medicine : open journal",
  ["toxicol in vitro"] = "Toxicology in vitro : an international journal published in association with BIBRA",
  ["toxicol ind health"] = "Toxicology and industrial health",
  ["toxicol int"] = "Toxicology international",
  ["toxicol lett"] = "Toxicology Letters",
  ["toxicol mech methods"] = "Toxicology mechanisms and methods",
  ["toxicol open access"] = "Toxicology: open access",
  ["toxicol pathol"] = "Toxicologic pathology",
  ["toxicol rep"] = "Toxicology reports",
  ["toxicol res"] = "Toxicological research",
  ["toxicol res (camb)"] = "Toxicology research",
  ["toxicol res (cambridge, u k)"] = "Toxicology Research (Cambridge, United Kingdom)",
  ["toxicol res appl"] = "Toxicology Research and Application",
  ["toxicol rev"] = "Toxicological reviews",
  ["toxicol sci"] = "Toxicological Sciences",
  ["toxicon x"] = "Toxicon: X",
  ["toxicon: x"] = "Toxicon: X",
  ["toxin rev"] = "Toxin reviews",
  ["toxins (basel)"] = "Toxins",
  ["toyama math j"] = "Toyama Mathematical Journal",
  ["toyoshi kenkyu"] = "Tōyōshi kenkyū",
  ["tōhō igakkai zasshi"] = "Tōhō Igakkai zasshi = Journal of the Medical Society of Toho University",
  ["tpm test psychom methodol appl psychol"] = "TPM. Testing, Psychometrics, Methodology in Applied Psychology",
  ["tr akad med nauk sssr mosc inst grudn khir"] = "Trudy. Institut grudnoĭ khirurgii (Akademii︠a︡ medit︠s︡inskikh nauk SSSR)",
  ["tr akad nauk sssr inst genet"] = "Trudy. Institut genetiki (Akademii︠a︡ nauk SSSR)",
  ["tr beloruss gos univ ser fiziol biohim mol osn funkc sist"] = "Trudy Belorusskogo gosudarstvennogo universiteta. Serii︠a︡ fiziologicheskie, biohimicheskie i molekuli︠a︡rnye osnovy funkt︠s︡ionirovanii︠a︡ sistem",
  ["tr fiziol lab akad i p pavlova"] = "Trudy fiziologicheskikh laboratoriĭ akademika I.P. Pavlova",
  ["tr gelmintol lab"] = "Trudy Gelʹmintologicheskoĭ laboratorii",
  ["tr geofiz inst, akad nauk sssr"] = "Trudy Geofizicheskogo Instituta, Akademiya Nauk SSSR",
  ["tr gl geofiz obs"] = "Trudy Glavnoi Geofizicheskoi Observatorii",
  ["tr inst biol"] = "Trudy Instituta biologii",
  ["tr inst fiz akad nauk gruz ssr"] = "Trudy Instituta fiziologii, Akademiia nauk Gruzinskoĭ SSR",
  ["tr inst fiziol im i p pavlova"] = "Trudy - Institut fiziologii imeni I. P. Pavlova",
  ["tr inst im pastera"] = "Trudy Instituta imeni Pastera",
  ["tr inst mat"] = "Trudy Instituta Matematiki",
  ["tr inst mat mekh"] = "Trudy Instituta Matematiki i Mekhaniki",
  ["tr inst morfol zhivotn an severtsova"] = "Trudy Instituta morfologii zhivotnykh im. A. N. Severtsova",
  ["tr inst norm patol fiziol"] = "Trudy - Institut normal'noĭ i patologicheskoĭ fiziologii",
  ["tr inst prikl mat mekh"] = "Trudy Instituta Prikladnoi Matematiki i Mekhaniki",
  ["tr inst tuberk"] = "Trudy. Institut tuberkuleza (Akademii︠a︡ medit︠s︡inskikh nauk SSSR)",
  ["tr latv padomju soc repub zinat akad mikrobiol inst"] = "Trudy. Mikrobiologijas institūts (Latvijas PSR Zinātnu akadēmija)",
  ["tr leningr inst epidemiol mikrobiol"] = "Trudy Leningradskogo instituta ėpidemiologii i mikrobiologii im. Pastera",
  ["tr leningr nauchnoissled inst epidemiol mikrobiol"] = "Trudy Leningradskogo nauchno-issledovatelʹskogo instituta ėpidemiologii i mikrobiologii imeni Pastera",
  ["tr leningr sanitarnogig med inst"] = "Trudy Leningradskogo sanitarno-gigienicheskogo meditsinskogo instituta",
  ["tr mat inst steklova"] = "Trudy Matematicheskogo Instituta Imeni V. A. Steklova",
  ["tr mosk mat obs"] = "Trudy Moskovskogo Matematicheskogo Obshchestva",
  ["tr mosk nauchnoissled inst epidmiol mikrobiol"] = "Trudy Moskovskogo nauchno-issledovatel'skogo instituta epidemiologii i mikrobiologii",
  ["tr petrozavodsk gos univ ser mat"] = "Trudy Petrozavodskogo Gosudarstvennogo Universiteta",
  ["tr russ entomol obs (2000)"] = "Trudy Russkogo èntomologičeskogo obŝestva",
  ["tr zool inst"] = "Trudy Zoologicheskogo instituta",
  ["trab catedra hist med"] = "Trabajos de la Cátedra de Historia de la Medicina",
  ["trab inst cajal"] = "Trabajos del Instituto Cajal",
  ["trab inst cajal invest biol"] = "Trabajos del Instituto Cajal de investigaciones biológicas",
  ["trab invest"] = "Trabalhos de Investigação",
  ["trab memorias"] = "Trabajos y memorias. Sociedad Nacional de Cirugía, Havana",
  ["trac, trends anal chem"] = "TRAC Trends in Analytical Chemistry",
  ["trace elem electrolytes"] = "Trace Elements and Electrolytes",
  ["trace elem med"] = "Trace elements in medicine",
  ["track rep"] = "Tracking report",
  ["tradit med mod med"] = "Traditional Medicine and Modern Medicine",
  ["traffic (oxford, u k)"] = "Traffic (Oxford, United Kingdom)",
  ["traffic inj prev"] = "Traffic injury prevention",
  ["train dev j"] = "Training and development journal",
  ["train educ prof psychol"] = "Training and education in professional psychology",
  ["train sch bull (vinel)"] = "The Training school bulletin",
  ["trained nurse hosp rev"] = "The Trained nurse and hospital review",
  ["traité nouvelles tech"] = "Traité des Nouvelles Technologies",
  ["traject cienc"] = "Trajectos Ciencia",
  ["trajecta (leuven)"] = "Trajecta",
  ["trak univ j nat sci"] = "Trakya University journal of natural sciences",
  ["trans a razmadze math inst"] = "Transactions of A. Razmadze Mathematical Institute",
  ["trans action"] = "Trans-action",
  ["trans am acad insur med"] = "Transactions of the American Academy of Insurance Medicine : Annual Meeting. American Academy of Insurance Medicine. Meeting",
  ["trans am acad ophthalmol otolaryngol"] = "Transactions - American Academy of Ophthalmology and Otolaryngology. American Academy of Ophthalmology and Otolaryngology",
  ["trans am assoc genitourin surg"] = "Transactions of the American Association of Genito-Urinary Surgeons",
  ["trans am climatol clin assoc"] = "Transactions of the American Climatological and Clinical Association. American Climatological and Clinical Association",
  ["trans am clin climatol assoc"] = "Transactions of the American Clinical and Climatological Association",
  ["trans am coll cardiol"] = "Transactions. American College of Cardiology",
  ["trans am crystallogr assoc"] = "Transactions of the American Crystallographic Association",
  ["trans am entomol soc"] = "Transactions of the American Entomological Society",
  ["trans am fish soc"] = "Transactions of the American Fisheries Society",
  ["trans am geophys union"] = "Transactions of the American Geophysical Union",
  ["trans am gynecol soc"] = "Transactions of the American Gynecological Society",
  ["trans am inst electr eng"] = "Transactions of the American Institute of Electrical Engineers",
  ["trans am inst electr eng part 1"] = "Transactions of the American Institute of Electrical Engineers Part 1: Communication and Electronics",
  ["trans am inst electr eng part 2"] = "Transactions of the American Institute of Electrical Engineers Part 2: Applications and Industry",
  ["trans am inst electr eng part 3"] = "Transactions of the American Institute of Electrical Engineers Part 3: Power Apparatus and Systems",
  ["trans am inst met"] = "Transactions of the American Institute of Metals",
  ["trans am inst min metall pet eng"] = "Transactions of the American Institute of Mining, Metallurgical and Petroleum Engineers",
  ["trans am inst min metall pet eng soc min eng aime"] = "Transactions of the American Institute of Mining, Metallurgical and Petroleum Engineers, Society of Mining Engineers of AIME",
  ["trans am laryngol assoc"] = "Transactions of the Annual Meeting of the American Laryngological Association",
  ["trans am laryngol rhinol otol soc"] = "Transactions of the American Laryngological, Rhinological and Otological Society, Inc",
  ["trans am math soc"] = "Transactions of the American mathematical society",
  ["trans am microsc soc"] = "Transactions of the American Microscopical Society",
  ["trans am neurol assoc"] = "Transactions of the American Neurological Association",
  ["trans am nucl soc"] = "Transactions of the American Nuclear Society",
  ["trans am ophthalmol soc"] = "Transactions of the American Ophthalmological Society",
  ["trans am otol soc"] = "Transactions of the American Otological Society",
  ["trans am proctol soc"] = "Transactions. American Proctologic Society",
  ["trans am soc artif intern organs"] = "Transactions - American Society for Artificial Internal Organs",
  ["trans am soc mech eng"] = "Transactions of the American Society of Mechanical Engineers",
  ["trans am soc met"] = "Transactions of the American Society for Metals",
  ["trans am soc neurochem"] = "Transactions of the American Society for Neurochemistry",
  ["trans am soc ophthalmol otolaryngol allergy"] = "Transactions - American Society of Ophthalmologic and Otolaryngologic Allergy",
  ["trans amer fish soc"] = "Transactions of the American Fisheries Society",
  ["trans amer math soc"] = "Transactions of the American Mathematical Society",
  ["trans amer math soc ser b"] = "Transactions of the American Mathematical Society. Series B",
  ["trans amer philos soc"] = "Transactions of the American Philosophical Society",
  ["trans annu meet am bronchoesophagol assoc"] = "Transactions of the Annual Meeting of the American Broncho-Esophagological Association",
  ["trans annu meet natl tuberc assoc"] = "Transactions of the annual meeting. National Tuberculosis Association",
  ["trans annu meet orthop res soc"] = "Transactions of the annual meeting of the Orthopaedic Research Society. Orthopaedic Research Society",
  ["trans asabe"] = "Transactions of the ASABE",
  ["trans asae"] = "Transactions of the ASAE. American Society of Agricultural Engineers",
  ["trans asme j appl mech"] = "American Society of Mechanical Engineers",
  ["trans assoc am physicians"] = "Transactions of the Association of American Physicians",
  ["trans assoc comput linguist"] = "Transactions of the Association for Computational Linguistics",
  ["trans assoc ind med off"] = "The Transactions of the Association of Industrial Medical Officers",
  ["trans assoc life insur med dir am"] = "Transactions of the Association of Life Insurance Medical Directors of America",
  ["trans aust coll ophthalmol"] = "Transactions of the Australian College of Ophthalmologists",
  ["trans aust med congr"] = "Transactions. Australasian Medical Congress (British Medical Association)",
  ["trans azerb natl acad sci ser phys-tech math sci phys astron"] = "Transactions of Azerbaijan National Academy of Sciences. Series of Physical-Technical and Mathematical Sciences. Physics and Astronomy",
  ["trans beijing inst tech"] = "Transactions of Beijing Institute of Technology. Beijing Ligong Daxue Xuebao",
  ["trans br ceram soc"] = "Transactions of the British Ceramic Society",
  ["trans br mycol soc"] = "Transactions of the British Mycological Society",
  ["trans br soc hist pharm"] = "Transactions of the British Society for the History of Pharmacy",
  ["trans br soc study orthod"] = "Transactions of the British Society for the Study of Orthodontics",
  ["trans bristol glos archaeol soc"] = "Transactions - Bristol and Gloucestershire Archaeological Society. Bristol and Gloucestershire Archaeological Society",
  ["trans bull ind hyg found am"] = "Transactions bulletin. Industrial Hygiene Foundation of America",
  ["trans camb bibliogr soc"] = "Transactions of the Cambridge Bibliographical Society",
  ["trans can opthalmolog soc"] = "Transactions of the Canadian Ophthalmological Society",
  ["trans can soc mech eng"] = "Transactions of the Canadian Society for Mechanical Engineering",
  ["trans comb"] = "Transactions on Combinatorics",
  ["trans conf liver inj"] = "Transactions. Conference on Liver Injury",
  ["trans conf metab asp conval"] = "Transactions. Conference on Metabolic Aspects of Convalescence",
  ["trans data priv"] = "Transactions on data privacy",
  ["trans earth, environ, sustainability"] = "Transactions in Earth, Environment, and Sustainability",
  ["trans edinb obstet soc"] = "Transactions. Edinburgh Obstetrical Society",
  ["trans egypt soc chem eng"] = "Transactions of the Egyptian Society of Chemical Engineers",
  ["trans electr electron mater"] = "Transactions on Electrical and Electronic Materials",
  ["trans emerging telecommun technol"] = "Transactions on Emerging Telecommunications Technologies",
  ["trans engl ceram circle"] = "Transactions (English Ceramic Circle)",
  ["trans entomol soc am"] = "Transactions of the Entomological Society of America",
  ["trans epidemiol soc lond"] = "Transactions. Epidemiological Society of London",
  ["trans essex archaeol soc"] = "Essex archaeology and history : the transactions of the Essex Archaeological Society",
  ["trans eur orthod soc"] = "Transactions. European Orthodontic Society",
  ["trans famena"] = "Transactions of FAMENA",
  ["trans faraday soc"] = "Transactions of the Faraday Society",
  ["trans gis"] = "Transactions in GIS : TG",
  ["trans hist"] = "Transport history",
  ["trans hunter soc"] = "The Transactions of the Hunterian Society",
  ["trans ill state acad sci"] = "Transactions of the Illinois State Academy of Science. Illinois State Academy of Science",
  ["trans imf"] = "Transactions of the IMF",
  ["trans indian ceram soc"] = "Transactions of the Indian Ceramic Society",
  ["trans indian inst met"] = "Transactions of the Indian Institute of Metals",
  ["trans indian natl acad eng"] = "Transactions of Indian National Academy of Engineering",
  ["trans indiana acad ophthalmol otolaryngol"] = "Transactions - Indiana Academy of Ophthalmology and Otolaryngology",
  ["trans inform process soc japan"] = "Information Processing Society of Japan",
  ["trans inst br geogr"] = "Transactions (Institute of British Geographers : 1965)",
  ["trans inst brew"] = "Transactions of the Institute of Brewing",
  ["trans inst civ eng"] = "Transactions of the Institution of Civil Engineers",
  ["trans inst indian geogr"] = "Transactions of the Institute of Indian Geographers",
  ["trans inst meas control"] = "Transactions of the Institute of Measurement and Control",
  ["trans inst met finish"] = "Transactions of the Institute of Metal Finishing",
  ["trans inst min metall, sect a"] = "Transactions of the Institution of Mining and Metallurgy Section A",
  ["trans inst min metall, sect b"] = "Transactions of the Institution of Mining and Metallurgy Section B",
  ["trans inst min metall, sect c"] = "Transactions of the Institution of Mining and Metallurgy Section C: Mineral Processing and Extractive Metallurgy",
  ["trans inst systems control inform engrs"] = "Institute of Systems, Control and Information Engineers",
  ["trans int astron union"] = "Transactions of the International Astronomical Union",
  ["trans int conf endod"] = "Transaction of the International Conference on Endodontics. International Conference on Endodontics",
  ["trans int conf oral surg"] = "Transactions of the International Conference on Oral Surgery",
  ["trans int soc music inf retr"] = "Transactions of the International Society for Music Information Retrieva",
  ["trans japan acad"] = "Transactions of the Japan Academy",
  ["trans jew hist soc engl"] = "Jewish historical studies : transactions of the Jewish Historical Society of England",
  ["trans jpn inst electron packag"] = "Transactions of the Japan Institute of Electronics Packaging",
  ["trans jpn soc aeronaut space sci"] = "Transactions of the Japan Society for Aeronautical and Space Sciences",
  ["trans jpn soc artif intell"] = "Transactions of the Japanese Society for Artificial Intelligence = Jinkō Chinō Gakkai ronbunshi",
  ["trans kans acad sci"] = "Transactions of the Kansas Academy of Science. Kansas Academy of Science",
  ["trans kansas acad sci"] = "Transactions of the Kansas Academy of Science",
  ["trans korean inst electr eng"] = "Transactions of the Korean Institute of Electrical Engineers",
  ["trans korean soc mech eng a"] = "Transactions of the Korean Society of Mechanical Engineers, A",
  ["trans korean soc mech eng b"] = "Transactions of the Korean Society of Mechanical Engineers, B",
  ["trans lab club"] = "Transactions of the Laboratory Club",
  ["trans lancaster cheshire antiq soc"] = "Transactions of the Lancashire and Cheshire Antiquarian Society. Lancashire and Cheshire Antiquarian Society",
  ["trans leicester archaeol hist soc"] = "Transactions - The Leicestershire Archaeological and Historical Society. Leicestershire Archaeological and Historical Society",
  ["trans lepidopterol soc jpn"] = "Transactions of the Lepidopterological Society of Japan",
  ["trans linn soc london"] = "Transactions of the Linnean Society of London",
  ["trans london math soc"] = "Transactions of the London Mathematical Society",
  ["trans mater res soc jpn"] = "Transactions of the Materials Research Society of Japan",
  ["trans math appl"] = "Transactions of Mathematics and its Applications",
  ["trans med chir soc edinb"] = "Transactions. Medico-Chirurgical Society of Edinburgh",
  ["trans med soc lond"] = "Transactions of the Medical Society of London",
  ["trans metall soc aime"] = "Transactions of the Metallurgical Society of AIME",
  ["trans microsc soc london"] = "Transactions of the Microscopical Society of London",
  ["trans mo acad sci"] = "Transactions of the Missouri Academy of Science. Missouri Academy of Science",
  ["trans moscow math soc"] = "Transactions of the Moscow Mathematical Society",
  ["trans n engl obstet gynecol soc"] = "Transactions of the New England Obstetrical and Gynecological Society",
  ["trans n y acad sci"] = "Transactions of the New York Academy of Sciences",
  ["trans natl acad sci azerb ser phys-tech math sci"] = "Transactions of National Academy of Sciences of Azerbaijan. Series of Physical-Technical and Mathematical Sciences",
  ["trans natl saf congr"] = "Transactions. National Safety Council. National Safety Congress",
  ["trans nebr acad sci affil soc"] = "Transactions of the Nebraska Academy of Sciences and affiliated societies",
  ["trans new orleans acad ophthalmol"] = "Transactions of the New Orleans Academy of Ophthalmology",
  ["trans new york acad sci ser ii"] = "Transactions of the New York Academy of Sciences, Series II",
  ["trans nonferrous met soc china"] = "Transactions of the Nonferrous Metals Society of China",
  ["trans ophthalmol soc aust"] = "Transactions of the Ophthalmological Society of Australia",
  ["trans ophthalmol soc n z"] = "Transactions of the Ophthalmological Society of New Zealand",
  ["trans ophthalmol soc u k"] = "Transactions of the ophthalmological societies of the United Kingdom",
  ["trans opt soc"] = "Transactions of the Optical Society",
  ["trans opthal soc u k"] = "Transactions. Ophthalmological Society of the United Kingdom",
  ["trans pa acad ophthalmol otolaryngol"] = "Transactions - Pennsylvania Academy of Ophthalmology and Otolaryngology",
  ["trans pac coast obstet gynecol soc"] = "Transactions of the Pacific Coast Obstetrical and Gynecological Society",
  ["trans pac coast otoophthalmol soc annu meet"] = "Transactions of the Pacific Coast Oto-Ophthalmological Society annual meeting",
  ["trans philol soc"] = "Transactions of the Philological Society. Philological Society (Great Britain)",
  ["trans plann urban res"] = "Transactions in Planning and Urban Research",
  ["trans proc am philol assoc"] = "Transactions and proceedings of the American Philological Association. American Philological Association",
  ["trans r entomol soc london"] = "Transactions of the Royal Entomological Society of London",
  ["trans r hist soc"] = "Transactions of the Royal Historical Society",
  ["trans r sch dent stockh umea"] = "Transactions of the Royal Schools of Dentistry, Stockholm and Umeå. Stockholm. Tandläkarhögskolan",
  ["trans r soc can"] = "Transactions of the Royal Society of Canada",
  ["trans r soc can (1963)"] = "Transactions of the Royal Society of Canada. Sections 1, 2, and 3 = Mémoires de la Société royale du Canada. Sections 1, 2, et 3",
  ["trans r soc can 5 biol sci"] = "Transactions of the Royal Society of Canada. Section 5, Biological sciences. Royal Society of Canada",
  ["trans r soc edinb earth sci"] = "Transactions of the Royal Society of Edinburgh: earth sciences",
  ["trans r soc edinb, earth sci"] = "Transactions of the Royal Society of Edinburgh, Earth Sciences",
  ["trans r soc edinburgh: earth sci"] = "Transactions of the Royal Society of Edinburgh Earth Science",
  ["trans r soc s afr"] = "Transactions of the Royal Society of South Africa",
  ["trans r soc s aust"] = "Transactions of the Royal Society of South Australia, Incorporated. Royal Society of South Australia",
  ["trans r soc trop med hyg"] = "Transactions of the Royal Society of Tropical Medicine and Hygiene",
  ["trans rep liverp med inst"] = "Transactions and report. Liverpool Medical Institution",
  ["trans saest"] = "Transactions of the SAEST",
  ["trans sect ophthalmol am acad ophthalmol otolaryngol"] = "Transactions. Section on Ophthalmology. American Academy of Ophthalmology and Otolaryngology",
  ["trans sect otolaryngol am acad ophthalmol otolaryngol"] = "Transactions. Section on Otolaryngology. American Academy of Ophthalmology and Otolaryngology",
  ["trans soc adv electrochem sci technol"] = "Transactions of the Society for Advancement of Electrochemical Science and Technology",
  ["trans soc min eng"] = "Transactions of Society of Mining Engineers",
  ["trans soc min eng aime"] = "Transactions of the Society of Mining Engineers of AIME",
  ["trans soc min metall explor"] = "Transactions of the Society for Mining, Metallurgy, and Exploration",
  ["trans soc min metall explor inc"] = "Transactions of Society for Mining, Metallurgy, and Exploration, Inc",
  ["trans soc occup med"] = "The Transactions of the Society of Occupational Medicine",
  ["trans soc rheol"] = "Transactions of the Society of Rheology",
  ["trans south afr inst electr eng"] = "Transactions of the South African Institute of Electrical Engineering",
  ["trans south cent sect am urol assoc"] = "Transactions. American Urological Association. South Central Section",
  ["trans south surg assoc"] = "Transactions of the Southern Surgical Association. Southern Surgical Association (U.S.)",
  ["trans southeast sect am urol assoc"] = "Transactions. American Urological Association. Southeastern Section",
  ["trans st johns hosp dermatol soc"] = "Transactions of the St. John's Hospital Dermatological Society",
  ["trans stud coll physicians phila"] = "Transactions & studies of the College of Physicians of Philadelphia",
  ["trans tianjin univ"] = "Transactions of Tianjin University",
  ["trans urban data, sci, technol"] = "Transactions in Urban Data, Science, and Technology",
  ["trans west sect am urol assoc"] = "Transactions. American Urological Association. Western Section",
  ["trans west surg assoc"] = "Transactions. Western Surgical Association",
  ["trans zool soc london"] = "Transactions of the Zoological Society of London",
  ["trans-am geophys union"] = "Transactions of the American Geophysical Union",
  ["transbound emerg dis"] = "Transboundary and emerging diseases",
  ["transboundary emerging dis"] = "Transboundary and Emerging Diseases",
  ["transcr open access"] = "Transcriptomics: open access",
  ["transcult psychiatry"] = "Transcultural psychiatry",
  ["transcult res heidelb stud asia europe glob context"] = "Transcultural Research—Heidelberg Studies on Asia and Europe in a Global Context",
  ["transdiscipl j eng sci"] = "Transdisciplinary journal of engineering & science",
  ["transfer (bruss)"] = "Transfer (Brussels, Belgium)",
  ["transform digit worlds (2018)"] = "Transforming digital worlds : 13th International Conference, iConference 2018, Sheffield, UK, March 25-28, 2018, Proceedings. International Conference on Transforming Digital Worlds (13th : 2018 : Sheffield, England)",
  ["transform ekosistem"] = "Transformat︠s︡ii︠a︡ ėkosistem",
  ["transform groups"] = "Transformation Groups",
  ["transfus altern transfus med"] = "Transfusion alternatives in transfusion medicine : TATM",
  ["transfus apher sci"] = "Transfusion and apheresis science : official journal of the World Apheresis Association : official journal of the European Society for Haemapheresis",
  ["transfus apheresis sci"] = "Transfusion and Apheresis Science",
  ["transfus clin biol"] = "Transfusion clinique et biologique : journal de la Société française de transfusion sanguine",
  ["transfus med"] = "Transfusion medicine (Oxford, England)",
  ["transfus med hemother"] = "Transfusion medicine and hemotherapy : offizielles Organ der Deutschen Gesellschaft fur̈ Transfusionsmedizin und Immunham̈atologie",
  ["transfus med rev"] = "Transfusion medicine reviews",
  ["transfus sci"] = "Transfusion science",
  ["transfusion (paris)"] = "Transfusion",
  ["transgend health"] = "Transgender health",
  ["transgenic res"] = "Transgenic research",
  ["transit water bull"] = "Transitional waters bulletin",
  ["transition met chem"] = "Transition Metal Chemistry",
  ["transition met chem (dordrecht, neth)"] = "Transition Metal Chemistry (Dordrecht, Netherlands)",
  ["transition met chem (london)"] = "Transition Metal Chemistry (London)",
  ["transl  psychiatry"] = "Translational Psychiatry",
  ["transl androl urol"] = "Translational andrology and urology",
  ["transl anim sci"] = "Translational animal science",
  ["transl behav med"] = "Translational behavioral medicine",
  ["transl biomed"] = "Translational biomedicine",
  ["transl biophotonics"] = "Translational biophotonics",
  ["transl brain rhythm"] = "Translational brain rhythmicity",
  ["transl cancer res"] = "Translational cancer research",
  ["transl clin pharmacol"] = "Translational and clinical pharmacology",
  ["transl dev psychiatry"] = "Translational developmental psychiatry",
  ["transl endocrinol metab"] = "Translational endocrinology & metabolism",
  ["transl gastroenterol hepatol"] = "Translational gastroenterology and hepatology",
  ["transl gastrointest cancer"] = "Translational gastrointestinal cancer",
  ["transl issues psychol sci"] = "Translational issues in psychological science",
  ["transl j am coll sports med"] = "Translational journal of the American College of Sports Medicine",
  ["transl lung cancer res"] = "Translational lung cancer research",
  ["transl mater res"] = "Translational Materials Research",
  ["transl math monogr"] = "Translations of Mathematical Monographs",
  ["transl med (sunnyvale)"] = "Translational medicine (Sunnyvale, Calif.)",
  ["transl med aging"] = "Translational medicine of aging",
  ["transl med commun"] = "Translational medicine communications",
  ["transl med unisa"] = "Translational medicine @ UniSa",
  ["transl metab syndr res"] = "Translational Metabolic Syndrome Research",
  ["transl neurodegener"] = "Translational neurodegeneration",
  ["transl neurosci"] = "Translational neuroscience",
  ["transl oncogenomics"] = "Translational oncogenomics",
  ["transl oncol"] = "Translational oncology",
  ["transl pediatr"] = "Translational pediatrics",
  ["transl perioper pain med"] = "Translational perioperative and pain medicine",
  ["transl proteom"] = "Translational proteomics",
  ["transl proteomics"] = "Translational Proteomics",
  ["transl psychiatry"] = "Translational psychiatry",
  ["transl res"] = "Translational research : the journal of laboratory and clinical medicine",
  ["transl respir med"] = "Translational respiratory medicine",
  ["transl sci"] = "Translational science",
  ["transl sci rare dis"] = "Translational science of rare diseases",
  ["transl sports med"] = "Translational sports medicine",
  ["transl stroke res"] = "Translational stroke research",
  ["transl syst sci"] = "Translational Systems Sciences",
  ["transl vis sci technol"] = "Translational vision science & technology",
  ["transl vision sci technol"] = "Translational Vision Science & Technology",
  ["translation (austin)"] = "Translation (Austin, Tex.)",
  ["translation (st barbara, calif)"] = "Translation (Santa Barbara, Calif.)",
  ["transnat corporations"] = "Transnational Corporations",
  ["transp cybersecur privacy"] = "Transportation Cybersecurity and Privacy",
  ["transp dev econ"] = "Transportation in Developing Economies",
  ["transp eng"] = "Transportation Engineering",
  ["transp geotech"] = "Transportation Geotechnics",
  ["transp infrastruct geotechnol"] = "Transportation Infrastructure Geotechnology",
  ["transp j"] = "Transportation Journal",
  ["transp lett"] = "Transportation Letters",
  ["transp plann technol"] = "Transportation Planning and Technology",
  ["transp policy (oxf)"] = "Transport policy",
  ["transp porous media"] = "Transport in Porous Media",
  ["transp res"] = "Transportation Research",
  ["transp res d transp environ"] = "Transportation research. Part D, Transport and environment",
  ["transp res e logist transp rev"] = "Transportation research. Part E, Logistics and transportation review",
  ["transp res interdiscip perspect"] = "Transportation research interdisciplinary perspectives",
  ["transp res part a policy pract"] = "Transportation research. Part A, Policy and practice",
  ["transp res part b methodol"] = "Transportation Research Part B Methodological",
  ["transp res part c emerg technol"] = "Transportation research. Part C, Emerging technologies",
  ["transp res part c emerging technol"] = "Transportation Research Part C Emerging Technologies",
  ["transp res part d transp environ"] = "Transportation Research Part D Transport and Environment",
  ["transp res part e logist transp rev"] = "Transportation Research Part E Logistics and Transportation Review",
  ["transp res part f psychol behav"] = "Transportation Research Part F Traffic Psychology and Behaviour",
  ["transp res part f traffic psychol behav"] = "Transportation research. Part F, Traffic psychology and behaviour",
  ["transp res procedia"] = "Transportation Research Procedia",
  ["transp res rec"] = "Transportation Research Record",
  ["transp res: part a: pol practice"] = "Transportation Research: Part A: Policy and Practice",
  ["transp res: part b: methodological"] = "Transportation Research: Part B: Methodological",
  ["transp res: part d: transport environ"] = "Transportation Research: Part D: Transport and Environment",
  ["transp res: part e: logist transp rev"] = "Transportation Research: Part E: Logistics and Transporation Review",
  ["transp rev"] = "Transport reviews",
  ["transp saf  environ"] = "Transportation Safety and Environment",
  ["transp theory stat phys"] = "Transport theory and statistical physics",
  ["transpers psychol rev"] = "Transpersonal psychology review",
  ["transpl immunol"] = "Transplant immunology",
  ["transpl infect dis"] = "Transplant infectious disease : an official journal of the Transplantation Society",
  ["transpl int"] = "Transplant international : official journal of the European Society for Organ Transplantation",
  ["transplant bull"] = "Transplantation bulletin",
  ["transplant cell ther"] = "Transplantation and cellular therapy",
  ["transplant direct"] = "Transplantation direct",
  ["transplant int"] = "Transplant International",
  ["transplant proc"] = "Transplantation proceedings",
  ["transplant res"] = "Transplantation research",
  ["transplant res risk manage"] = "Transplant Research and Risk Management",
  ["transplant rev"] = "Transplantation reviews",
  ["transplant rev (orlando)"] = "Transplantation reviews (Orlando, Fla.)",
  ["transplant sci"] = "Transplantation science",
  ["transport rev"] = "Transport Reviews",
  ["transport theory statist phys"] = "Transport Theory and Statistical Physics",
  ["transportation (amst)"] = "Transportation",
  ["transportmetrica a: transport sci"] = "Transportmetrica A: Transport Science",
  ["transportmetrica b: transport dyn"] = "Transportmetrica B: Transport Dynamics",
  ["trastor adict"] = "Trastornos adictivos",
  ["trauma acute care"] = "Trauma & acute care",
  ["trauma berufskrankh"] = "Trauma und Berufskrankheit",
  ["trauma case rep"] = "Trauma case reports",
  ["trauma emerg care"] = "Trauma and emergency care",
  ["trauma mon"] = "Trauma monthly",
  ["trauma surg acute care open"] = "Trauma surgery & acute care open",
  ["trauma violence abuse"] = "Trauma, violence & abuse",
  ["traumatology (tallahass fla)"] = "Traumatology",
  ["trav hum"] = "Le Travail humain",
  ["trav lab matiere med pharm galenique fac pharm paris"] = "Travaux des Laboratoires de matière médicale et de pharmacie galénique de la Faculté de pharmacie de Paris",
  ["trav math"] = "Travaux Mathematiques",
  ["trav sci parc natl vanoise"] = "Travaux Scientifiques du Parc National de la Vanoise",
  ["trav soc pharm montp"] = "Travaux de la Société de pharmacie de Montpellier",
  ["trav stn rech eaux for"] = "Travaux Station de recherches des eaux et forêts",
  ["travail emploi"] = "Travail et Emploi",
  ["travaux centre rech sémiol"] = "Travaux du Centre de Recherches Sémiologiques",
  ["travaux en cours"] = "Travaux en Cours",
  ["travaux log"] = "Travaux de Logique",
  ["travel behav soc"] = "Travel behaviour & society",
  ["travel med infect dis"] = "Travel medicine and infectious disease",
  ["travel stand"] = "The Travelers standard",
  ["trazos (bilbao)"] = "Trazos; entretenimientos de clínica y terapéutica",
  ["treat endocrinol"] = "Treatments in endocrinology",
  ["treat guidel med lett"] = "Treatment guidelines from the Medical Letter",
  ["treat respir med"] = "Treatments in respiratory medicine",
  ["treat rev"] = "Treatment review",
  ["treat serv bull"] = "Treatment services bulletin. Canada. Department of Veterans' Affairs",
  ["treat strateg oncol"] = "Treatment strategies. Oncology",
  ["treat strategies diabetes"] = "Treatment strategies. Diabetes",
  ["treat strategies hematol"] = "Treatment strategies. Hematology",
  ["tree genet genomes"] = "Tree genetics & genomes",
  ["tree physiol"] = "Tree physiology",
  ["tree ring res"] = "Tree-ring research",
  ["tree-ring bull"] = "Tree-ring Bulletin",
  ["tree-ring res"] = "Tree-Ring Research",
  ["trees (berl west)"] = "Trees (Berlin, Germany : West)",
  ["trees, forests people"] = "Trees, Forests and People",
  ["trees-struct funct"] = "Trees-Structure and Function",
  ["tremor other hyperkinet mov (n y)"] = "Tremor and other hyperkinetic movements (New York, N.Y.)",
  ["tren reprod bio"] = "Trends in reproductive biology",
  ["trends abstr appl anal"] = "Trends in Abstract and Applied Analysis",
  ["trends amplif"] = "Trends in amplification",
  ["trends anal chem"] = "Trends in Analytical Chemistry",
  ["trends analyt chem"] = "Trends in analytical chemistry : TRAC",
  ["trends appl sci res"] = "Trends in applied sciences research",
  ["trends appl spectrosc"] = "Trends in applied spectroscopy",
  ["trends biochem sci"] = "Trends in biochemical sciences",
  ["trends biomater artif organs"] = "Trends in Biomaterials and Artificial Organs",
  ["trends biosci"] = "Trends in Biosciences",
  ["trends biotechnol"] = "Trends in biotechnology",
  ["trends biotechnol res"] = "Trends in biotechnology research (Solapur)",
  ["trends cancer"] = "Trends in cancer",
  ["trends cancer res"] = "Trends in cancer research",
  ["trends carbohydr res"] = "Trends in carbohydrate research",
  ["trends cardiovasc med"] = "Trends in cardiovascular medicine",
  ["trends cell biol"] = "Trends in cell biology",
  ["trends cell mol biol"] = "Trends in cell & molecular biology",
  ["trends chem"] = "Trends in chemistry",
  ["trends class suppl vol"] = "Trends in Classics—Supplementary Volumes",
  ["trends cogn sci"] = "Trends in cognitive sciences",
  ["trends cognit sci"] = "Trends in Cognitive Sciences",
  ["trends comput appl math"] = "Trends in Computational and Applied Mathematics",
  ["trends dev biol"] = "Trends in developmental biology",
  ["trends diabetes metab"] = "Trends in diabetes and metabolism",
  ["trends ecol evol"] = "Trends in ecology & evolution",
  ["trends endocrinol metab"] = "Trends in endocrinology and metabolism: TEM",
  ["trends entomol"] = "Trends in entomology",
  ["trends environ anal chem"] = "Trends in Environmental Analytical Chemistry",
  ["trends evol biol"] = "Trends in evolutionary biology",
  ["trends food sci technol"] = "Trends in Food Science and Technology",
  ["trends genet"] = "Trends in genetics : TIG",
  ["trends glycosci glycotechnol"] = "Trends in glycoscience and glycotechnology : TIGG",
  ["trends health care law ethics"] = "Trends in health care, law & ethics",
  ["trends hear"] = "Trends in hearing",
  ["trends hist"] = "Trends in history",
  ["trends hist sci"] = "Trends in the History of Science",
  ["trends immunol"] = "Trends in immunology",
  ["trends log stud log libr"] = "Trends in Logic—Studia Logica Library",
  ["trends math"] = "Trends in Mathematics",
  ["trends math res perspect"] = "Trends in Mathematics. Research Perspectives",
  ["trends math res perspect crm barc"] = "Trends in Mathematics. Research Perspectives CRM Barcelona",
  ["trends microbiol"] = "Trends in microbiology",
  ["trends mol med"] = "Trends in molecular medicine",
  ["trends neurosci"] = "Trends in neurosciences",
  ["trends neurosci educ"] = "Trends in neuroscience and education",
  ["trends org chem"] = "Trends in organic chemistry",
  ["trends organ crime"] = "Trends in organized crime",
  ["trends parasitol"] = "Trends in parasitology",
  ["trends pharmacol sci"] = "Trends in pharmacological sciences",
  ["trends plant sci"] = "Trends in plant science",
  ["trends polym sci"] = "Trends in polymer science",
  ["trends psychiatry psychother"] = "Trends in psychiatry and psychotherapy",
  ["trends res"] = "Trends in research",
  ["trends sport sci"] = "Trends in sport sciences",
  ["trends tech contemp dent lab"] = "Trends & techniques in the contemporary dental laboratory",
  ["trenie iznos"] = "Trenie i Iznos",
  ["trial lawyers q"] = "Trial lawyers quarterly",
  ["trials vaccinol"] = "Trials in vaccinology",
  ["triangulo rev sandoz cienc med"] = "Triangulo",
  ["trib farm"] = "Tribuna farmacêutica",
  ["trib int womens trib cent"] = "Tribune (International Women's Tribune Centre : English ed.)",
  ["trib med soc"] = "Tribuna médico-social",
  ["trib odontol (b aires)"] = "La Tribuna odontológica",
  ["trib odontol (guanabara)"] = "Tribuna Odontologica Do Sindicato Dos Odontologistas Do Estado Da Guanabara",
  ["tribol int"] = "Tribology International",
  ["tribol interface eng ser"] = "Tribology and Interface Engineering Series",
  ["tribol lett"] = "Tribology Letters",
  ["tribol mater surf interfaces"] = "Tribology - Materials, Surfaces and & Interfaces",
  ["tribol schmierungstech"] = "Tribologie und Schmierungstechnik",
  ["tribol trans"] = "Tribology Transactions",
  ["tribulus (abu dhabi)"] = "Tribulus : bulletin of the Emirates Natural History Group",
  ["trim econ"] = "El Trimestre Economico",
  ["trimest econ"] = "El Trimestre económico",
  ["triple helix"] = "Triple Helix",
  ["triple helix (heidelb)"] = "Triple helix (Heidelberg)",
  ["trop agric"] = "Tropical agriculturist",
  ["trop anim health prod"] = "Tropical animal health and production",
  ["trop biomed"] = "Tropical biomedicine",
  ["trop bryol"] = "Tropical bryology",
  ["trop conserv sci"] = "Tropical conservation science",
  ["trop cyclone res rev"] = "Tropical Cyclone Research and Review",
  ["trop dis bull"] = "Tropical diseases bulletin",
  ["trop dis travel med vaccines"] = "Tropical diseases, travel medicine and vaccines",
  ["trop dis, travel med vaccines"] = "Tropical Diseases, Travel Medicine and Vaccines",
  ["trop doct"] = "Tropical doctor",
  ["trop ecol"] = "Tropical ecology",
  ["trop gastroenterol"] = "Tropical gastroenterology : official journal of the Digestive Diseases Foundation",
  ["trop geogr med"] = "Tropical and geographical medicine",
  ["trop grasslands"] = "Tropical Grasslands",
  ["trop heal"] = "Tropical Health",
  ["trop j nat prod res"] = "Tropical Journal of Natural Product Research",
  ["trop j obstet gynaecol"] = "Tropical journal of obstetrics and gynaecology",
  ["trop j pharm res"] = "Tropical journal of pharmaceutical research : TJPR",
  ["trop lepid res"] = "Tropical lepidoptera research",
  ["trop life sci res"] = "Tropical life sciences research",
  ["trop med health"] = "Tropical medicine and health",
  ["trop med infect dis"] = "Tropical medicine and infectious disease",
  ["trop med int health"] = "Tropical medicine & international health : TM & IH",
  ["trop med news"] = "Tropical medicine news",
  ["trop med parasitol"] = "Tropical medicine and parasitology : official organ of Deutsche Tropenmedizinische Gesellschaft and of Deutsche Gesellschaft für Technische Zusammenarbeit (GTZ)",
  ["trop med surg"] = "Tropical medicine & surgery",
  ["trop nat hist"] = "Tropical natural history",
  ["trop parasitol"] = "Tropical parasitology",
  ["trop plant biol"] = "Tropical plant biology",
  ["trop plant pathol"] = "Tropical plant pathology",
  ["trop subtrop agroecosyt"] = "Tropical and Subtropical Agroecosystems",
  ["trop zool"] = "Tropical Zoology",
  ["tropenmed parasitol"] = "Tropenmedizin und Parasitologie",
  ["trudy (inst biol vnutr vod im id papanina)"] = "Trudy (Institut biologii vnutrennikh vod im. I.D. Papanina)",
  ["trudy bgtu"] = "Trudy BGTU",
  ["trudy fiz inst lebedev"] = "Rossíiskaya Akademiya Nauk",
  ["trudy geom sem kazan univ"] = "Kazanskiy Universitet",
  ["trudy inst mat"] = "\\cyr Trudy Instituta Matematiki",
  ["trudy inst mikrobiol"] = "Trudy Instituta mikrobiologii",
  ["trudy iofan"] = "\\cyr Trudy Instituta Obshchey Fiziki",
  ["trudy leningr inst epidemiol mikrobiol gig im pastera"] = "Trudy Leningradskogo instituta ėpidemiologii, mikrobiologii, i gigieny im. Pastera",
  ["trudy mat inst steklov"] = "Rossíiskaya Akademiya Nauk",
  ["trudy mosc russ vsesoiuznyi nauchno issled vitam inst"] = "Trudy. Moscow (Russia). Vsesoiuznyĭ nauchno-issledovatel'skiĭ vitaminnyĭ institut",
  ["trudy s-peterburg mat obshch"] = "Trudy Sankt-Peterburgskogo Matematicheskogo Obshchestva",
  ["trudy sem kraev zadacham"] = "Trudy Seminara po Kraevym Zadacham",
  ["trudy sem petrovsk"] = "Moskovskiy Universitet",
  ["trudy sem s l soboleva"] = "Trudy Seminara S. L. Soboleva",
  ["trudy sem vektor tenzor anal"] = "Trudy Seminara po Vektornomu i Tenzornomu Analizu s ikh Prilozheniyami k Geometrii, Mekhanike i Fizike",
  ["trudy tbiliss univ mat mekh astronom"] = "Tbilisskiy Universitet",
  ["trudy tsentralnyi naucnoissled inst protez protezostr (mosc)"] = "Trudy. TSentral'nyĭ nauchno-issledovatel'skiĭ institut protezirovaniia i protezostroeniia (Moscow, R.S.F.S.R.)",
  ["trudy vsesoiuznoe fiziol obshchestvo im i p pavlov orenb otd"] = "Trudy. Vsesoiuznoe fiziologicheskoe obshchestvo imeni I. P. Pavlova. Orenburgskoe otdelenie",
  ["trust eternal syst via evol softw data knowl (2012)"] = "Trustworthy eternal systems via evolving software, data and knowledge : second international workshop, EternalS 2012, Montpellier, France, August 28, 2012, revised selected papers. EternalS (Workshop) (2nd : 2012 : Montpellier, France)",
  ["trusts estates"] = "Trusts and estates",
  ["tsa chih"] = "Chung-Hua Min Kuo Hsiao Erh Ko I Hsueh Hui Tsa Chih Chung Hua Min KuoHsiao Erh Ko I Hsueh Hu",
  ["tsa chih gaoxiong yi xue yuan tong xue hui"] = "[Tsa chih] [Journal of the] Alumni of the University of Takau. Gaoxiong yi xue yuan. Tong xue hui",
  ["tschermaks mineral petrogr mitt"] = "Tschermaks Mineralogische und Petrographische Mitteilungen",
  ["tsetse trypanos inf"] = "Tsetse and trypanosomiasis information",
  ["tsetse trypanos inf q"] = "Tsetse and trypanosomiasis information quarterly",
  ["tsinghua sci technol"] = "Tsinghua Science and Technology",
  ["tsitol genet"] = "T︠S︡itologii︠a︡ i genetika",
  ["tsuchi to biseibutsu"] = "Tsuchi to biseibutsu = Soil microorganisms",
  ["tsukuba daigaku jinbun chirigaku kenkyu"] = "Tsukuba Daigaku jinbun chirigaku kenkyū",
  ["tsukuba j math"] = "Tsukuba Journal of Mathematics",
  ["tsurumi shigaku"] = "Tsurumi shigaku. Tsurumi University dental journal",
  ["tsvetn met (moscow)"] = "Tsvetnaya Metally",
  ["tu nhien va cong nghe"] = "Tự nhiên và Công nghệ : Khoa học",
  ["tu rang xue bao"] = "Tu rang xue bao = Acta pedologica Sinica",
  ["tuber lung dis"] = "Tubercle and lung disease : the official journal of the International Union against Tuberculosis and Lung Disease",
  ["tuberc airborne dis wkly"] = "Tuberculosis And Airborne Disease Weekly",
  ["tuberc res treat"] = "Tuberculosis research and treatment",
  ["tuberc respir dis (seoul)"] = "Tuberculosis and respiratory diseases",
  ["tuberculol thorac dis"] = "Tuberculology and thoracic diseases",
  ["tuberculosis (edinb)"] = "Tuberculosis (Edinburgh, Scotland)",
  ["tuberk biolezni legkih"] = "Tuberkulëz i bolezni lëgkikh",
  ["tuberk grenzgeb einzeldarst"] = "Die Tuberkulose und ihre Grenzgebiete in Einzeldarstellungen",
  ["tuberk kerdesei"] = "Tuberkulózis kérdései",
  ["tuberk toraks"] = "Tüberküloz ve toraks",
  ["tuberk tudobetegsegek"] = "Tuberkulózis és tüdöbetegségek",
  ["tufts dent outlook"] = "Tufts dental outlook",
  ["tufts folia med"] = "Tufts folia medica",
  ["tufts health sci rev"] = "Tufts health science review",
  ["tufts med j"] = "Tufts medical journal",
  ["tulane law rev"] = "Tulane law review",
  ["tulsa law j"] = "Tulsa law journal",
  ["tumor biol"] = "Tumor Biology",
  ["tumor microenviron"] = "Tumor & microenvironment",
  ["tumor microenviron ther"] = "Tumor microenvironment and therapy",
  ["tumor res"] = "Tumor research",
  ["tumour biol"] = "Tumour biology : the journal of the International Society for Oncodevelopmental Biology and Medicine",
  ["tumour virus res"] = "Tumour virus research",
  ["tunis j math"] = "Tunisian Journal of Mathematics",
  ["tunis j plant prot"] = "Tunisian journal of plant protection",
  ["tunis med"] = "La Tunisie médicale",
  ["tunnelling underground space technol"] = "Tunnelling and Underground Space Technology",
  ["turcica (paris)"] = "Turcica",
  ["turk arch otorhinolaryngol"] = "Turkish archives of otorhinolaryngology",
  ["turk biyokim derg"] = "Türk biyokimya dergisi = Turkish journal of biochemistry",
  ["turk comput theor chem"] = "Turkish Computational And Theoretical Chemistry",
  ["turk entomol derg"] = "Türkiye entomoloji dergisi",
  ["turk geriatri derg"] = "Türk geriatri dergisi",
  ["turk gogus kalp damar cerrahisi derg"] = "Türk göğüs kalp damar cerrahisi dergisi",
  ["turk hemsire derg"] = "Türk hemşireler dergisi",
  ["turk hij deney biyol derg"] = "Türk hijiyen ve deneysel biyoloji dergisi. Turkish bulletin of hygiene and experimental biology",
  ["turk hij tecr biyol derg"] = "Türk hijiyen ve tecrübi biyoloji dergisi",
  ["turk j agric for"] = "Turkish Journal of Agriculture and Forestry",
  ["turk j anaesthesiol reanim"] = "Turkish journal of anaesthesiology and reanimation",
  ["turk j biochem"] = "Turkish Journal of Biochemistry",
  ["turk j biol"] = "Turkish journal of biology = Türk biyoloji dergisi",
  ["turk j bot"] = "Turkish Journal of Botany",
  ["turk j botany"] = "Turkish journal of botany",
  ["turk j chem"] = "Turkish Journal of Chemistry",
  ["turk j civ eng"] = "Turkish Journal of Civil Engineering",
  ["turk j comput math educ"] = "Turkish Journal of Computer And Mathematics Education",
  ["turk j earth sci"] = "Turkish Journal of Earth Sciences",
  ["turk j electr eng comput sci"] = "Turkish Journal of Electrical Engineering and Computer Sciences",
  ["turk j emerg med"] = "Turkish journal of emergency medicine",
  ["turk j endocrinol metab"] = "Turkishi Journal of Endocrinology and Metabolism",
  ["turk j eng environ sci"] = "Turkish Journal of Engineering and Environmental Science",
  ["turk j field crops"] = "Turkish Journal of Field Crops",
  ["turk j fish aquat sci"] = "Turkish journal of fisheries and aquatic sciences",
  ["turk j gastroenterol"] = "The Turkish journal of gastroenterology : the official journal of Turkish Society of Gastroenterology",
  ["turk j haematol"] = "Turkish journal of haematology : official journal of Turkish Society of Haematology",
  ["turk j hematol"] = "Turkish Journal of Hematology",
  ["turk j immunol"] = "Turkish Journal of Immunology",
  ["turk j math"] = "Turkish Journal of Mathematics",
  ["turk j med sci"] = "Turkish journal of medical sciences",
  ["turk j obstet gynecol"] = "Turkish journal of obstetrics and gynecology",
  ["turk j ophthalmol"] = "Turkish journal of ophthalmology",
  ["turk j orthod"] = "Turkish journal of orthodontics",
  ["turk j pathol"] = "Turkish Journal of Pathology",
  ["turk j pediatr"] = "The Turkish journal of pediatrics",
  ["turk j pharm sci"] = "Turkish journal of pharmaceutical sciences",
  ["turk j phys"] = "Turkisch Journal of Physics",
  ["turk j phys med rehabil"] = "Turkish journal of physical medicine and rehabilitation",
  ["turk j physiother rehabil"] = "Turkish Journal of Physiotherapy and Rehabilitation",
  ["turk j surg"] = "Turkish journal of surgery",
  ["turk j urol"] = "Turkish journal of urology",
  ["turk j vet anim sci"] = "Turkish journal of veterinary and animal sciences",
  ["turk j zool"] = "Turkish Journal of Zoology",
  ["turk kardiyol dern ars"] = "Türk Kardiyoloji Derneği arşivi : Türk Kardiyoloji Derneğinin yayın organıdır",
  ["turk klin immunol alerji"] = "Türkiye klinikleri. Immünoloji alerji",
  ["turk klin tip etigi hukuku tarihi"] = "Türkiye klinikleri. Tıp etiği-hukuku-tarihi = Türkiye klinikleri journal of medical ethics, law, and history",
  ["turk mikrobiyol cemiy derg"] = "Türk Mikrobiyoloji Cemiyeti dergisi",
  ["turk neurosurg"] = "Turkish neurosurgery",
  ["turk noroloji dergisi"] = "Türk nöroloji dergisi = Turkish journal of neurology",
  ["turk oftalmol derg"] = "Türk oftalmolojı dergısı",
  ["turk ortodonti derg"] = "Türk ortodonti dergisi : Ortodonti Derneğ'nin resmi yayin organidir = Turkish journal of orthodontics",
  ["turk patoloji derg"] = "Türk patoloji dergisi",
  ["turk pediatri ars"] = "Türk pediatri arşivi",
  ["turk psikiyatri derg"] = "Türk psikiyatri dergisi = Turkish journal of psychiatry",
  ["turk silahlı kuvvetleri koruyucu hekim bul"] = "Türk Silahlı Kuvvetleri koruyucu hekimlik bülteni",
  ["turk tarim arast derg"] = "Türkiye tarımsal araştırmalar dergisi",
  ["turk tarim gida bilim teknol derg"] = "Türk tarım gıda bilim ve teknoloji dergisi",
  ["turk thorac j"] = "Turkish thoracic journal",
  ["turk tip cemiy mecm"] = "Türk Tip Cemīyetī mecmuasi",
  ["turk tip dernegi derg"] = "Türk Tip Derneğī dergīsī",
  ["turk world math soc j app eng math"] = "Turkic World Mathematical Society Journal of Applied and Engineering Mathematics",
  ["turk zool derg"] = "Türk zooloji dergisi = Turkish journal of zoology",
  ["turkish j math"] = "Turkish Journal of Mathematics",
  ["turkish journal of botany|turk j bot"] = "Turk botanik dergisi",
  ["turkiye parazitol derg"] = "Türkiye parazitolojii dergisi",
  ["turnbull libr rec"] = "The Turnbull Library record",
  ["turt isl j indig health"] = "Turtle Island journal of indigenous health",
  ["turtox news"] = "Turtox news",
  ["turun hist ark (finl)"] = "Turun historiallinen arkisto",
  ["tutkimuksia—res rep"] = "Tutkimuksia—Research Reports",
  ["tutmondaj sciencoj tek"] = "Tutmondaj sciencoj kaj teknikoj",
  ["tutor quant methods psychol"] = "Tutorials in quantitative methods for psychology",
  ["tutor sch workshops math sci"] = "Tutorials, Schools, and Workshops in the Mathematical Sciences",
  ["twent century china"] = "Twentieth-century China = Er shi shi ji Zhongguo",
  ["twin res"] = "Twin research : the official journal of the International Society for Twin Studies",
  ["twin res hum genet"] = "Twin research and human genetics : the official journal of the International Society for Twin Studies",
  ["twms j app eng math"] = "TWMS Journal of Applied and Engineering Mathematics",
  ["twms j pure appl math"] = "TWMS Journal of Pure and Applied Mathematics",
  ["tydskr rasse aangel"] = "Tydskrif vir rasse-aangeleenthede. Journal of racial affairs",
  ["tydskr stud ekon ekon"] = "Tydskrif vir studies in ekonomie en ekonometrie = Journal for studies in economics and econometrics",
  ["tydskrif hedendaagse romeins holandse reg"] = "Tydskrif vir hedendaagse Romeins-Holandse Reg",
  ["tzu chi med j"] = "Tzu chi medical journal",
  ["tätigkber natforsch ges baselland"] = "Tätigkeitsbericht der Naturforschenden Gesellschaft Baselland",
  ["tüb geogr stud"] = "Tübinger Geographische Studien",
  ["tōhoku math j (2)"] = "The Tōhoku Mathematical Journal",
  ["u m c cordoba"] = "U M C",
  ["u s armed forces med j"] = "United States Armed Forces medical journal",
  ["u s chin j lymphology oncol"] = "U.S. Chinese journal of lymphology and oncology",
  ["u s code annot u s"] = "United States code annotated. United States",
  ["u s nav med bull"] = "United States naval medical bulletin",
  ["ubersich tierernahr"] = "Ubersichten zur Tierernahrung",
  ["ubiquitous comput electron mob commun conf (uemcon) ieee annu"] = "Ubiquitous Computing, Electronics & Mobile Communication Conference (UEMCON), IEEE Annual",
  ["uc davis law rev"] = "U.C. Davis law review. University of California, Davis. School of Law",
  ["uc zap tavriceskogo nac univ im v i vemadskogo"] = "Uchenye zapiski Tavricheskogo nat︠s︡ionalʹnogo universiteta im. V.I. Vernadskogo",
  ["ucd rep"] = "UCD [reports]. U.S. Atomic Energy Commission",
  ["uch zap artsakh gos univ"] = "Uchenye Zapiski. Artsakhskii Gosudarstvennyi Universitet",
  ["uch zap inst farmakol khimioter amn sssr"] = "Uchenye zapiski Instituta farmakologii i khimioterapii AMN SSSR",
  ["uch zap kazan univ ser fiz-mat nauki"] = "Uchenye Zapiski Kazanskogo Universiteta. Seriya Fiziko-Matematicheskie Nauki",
  ["uchen zap tsagi"] = "Uchenye Zapiski Tsentralnogo Aero-Gidrodinamicheskogo Instituta (TsAGI)",
  ["uchenye zap (mosc)"] = "Uchenye zapiski. Moscow (Russia). Moskovskiĭ gorodskoĭ pedagogicheskiĭ institut",
  ["uchenye zap fiz fak mgu"] = "Uchenye zapiski Fizicheskogo fakulʹteta MGU",
  ["uchenye zap mosk meditsinskii inst"] = "Uchenye zapiski. Vtoroĭ Moskovskiĭ ordena Lenina gosudarstvennyĭ medit︠s︡inskiĭ institut im N.I. Pirogova",
  ["uchenye zap seriia biolog leningr univ"] = "Uchenye zapiski. Seriia biologicheskaia. Leningrad. Universitet",
  ["uchu koku kankyo igaku"] = "Uchū kōkū kankyō igaku",
  ["ucl open"] = "UCL Open",
  ["ucl open environ"] = "UCL open environment",
  ["ucla forum med sci"] = "UCLA forum in medical sciences",
  ["ucla law rev"] = "UCLA law review. University of California, Los Angeles. School of Law",
  ["ucla pac basin law j"] = "UCLA Pacific Basin law journal",
  ["ucla rep"] = "UCLA [reports]. U.S. Atomic Energy Commission",
  ["ucla symp mol cell bio, new ser"] = "UCLA Symposia on Molecular and Cellular Biology, New Series",
  ["ucla symp mol cell biol"] = "UCLA symposia on molecular and cellular biology",
  ["ucla womens law j"] = "UCLA women's law journal",
  ["ucrl us at energy comm"] = "UCRL [reports]. U.S. Atomic Energy Commission",
  ["ufa math j"] = "Ufa Mathematical Journal",
  ["ufsi rep"] = "UFSI reports",
  ["ug ross ekol razvit"] = "I︠U︡g Rossii - ėkologii︠a︡, razvitie",
  ["uganda med j"] = "Uganda medical journal",
  ["ugeskr laeger"] = "Ugeskrift for laeger",
  ["ui j"] = "UI journal",
  ["ujcd union jeunes chir dent"] = "Ujcd. Union Des Jeunes Chirurgiens-Dentistes",
  ["uk energy res cent"] = "UK Energy Research Centre",
  ["ukr antarct j"] = "Ukrainian Antarctic Journal",
  ["ukr biochem j"] = "Ukrainian Biochemical Journal",
  ["ukr biokhim zh"] = "Ukrainskii Biokhimicheskii Zhurnal",
  ["ukr biokhim zh (1978)"] = "Ukrainskiĭ biokhimicheskiĭ zhurnal (1978)",
  ["ukr biokhim zh (1999)"] = "Ukraïnsʹkyĭ biokhimichnyĭ z︠h︡urnal (1999 )",
  ["ukr bioorg acta"] = "Ukrainica Bioorganica Acta",
  ["ukr bot z"] = "Ukraïns'kyǐ botanichnyǐ z︠h︡urnal (Kiev, U.R.S.R. : 1956)",
  ["ukr bot z ukr bot tovarystvo"] = "Ukraïnsʹkyĭ botanichnyĭ z︠h︡urnal. Ukrainian botanical review. Ukraïnsʹke botanichne tovarystvo",
  ["ukr botan j"] = "Ukrainian Botanical Journal",
  ["ukr fiz zh"] = "Ukrainskii Fizicheskii Zhurnal",
  ["ukr fiz zh (russ ed)"] = "Ukrainskii Fizicheskii Zhurnal (Russian Edition)",
  ["ukr food j"] = "Ukrainian Food Journal",
  ["ukr istor z"] = "Ukraïnsʹkyĭ istorychnyĭ zhurnal",
  ["ukr j ecol"] = "Ukrainian journal of ecology",
  ["ukr j food sci"] = "Ukrainian Journal of Food Science",
  ["ukr j for wood sci"] = "Ukrainian Journal of Forest and Wood Science",
  ["ukr j phys"] = "Ukrainian Journal of Physics",
  ["ukr j phys opt"] = "Ukrainian Journal Of Physical Optics",
  ["ukr j radiol oncol"] = "Ukrainian Journal Of Radiology And Oncology",
  ["ukr khim zh"] = "Ukrainskii Khimicheskii Zhurnal",
  ["ukr khim zh (russ ed)"] = "Ukrainskii Khimicheskii Zhurnal (Russian Edition)",
  ["ukr mat visn"] = "Ukrainskii Matematichnii Visnik",
  ["ukr math j"] = "Ukrainian Mathematical Journal",
  ["ukr phys j"] = "Ukrainian Physics Journal [translation of Ukrainskii Fizicheskii Zhurnal (Russian Edition)]",
  ["ukr q"] = "The Ukrainian quarterly",
  ["ukr rev (lond)"] = "Ukrainian review (London, England)",
  ["ukrain geom sb"] = "Ukrainskiy Geometricheskiy Sbornik",
  ["ukrain mat zh"] = "Natsionalna Akademiya Nauk Ukraini. Institut Matematiki. Ukrainskii Matematichnii Zhurnal",
  ["ukrainian econ rev"] = "Ukrainian Economic Review",
  ["ukrainian math j"] = "Ukrainian Mathematical Journal",
  ["ulam quart"] = "Ulam Quarterly",
  ["ulster folklife"] = "Ulster folklife",
  ["ulster med j"] = "The Ulster medical journal",
  ["ultra sci phys sci"] = "Ultra Scientist of Physical Sciences",
  ["ultrafast sci"] = "Ultrafast Science",
  ["ultraschall med"] = "Ultraschall in der Medizin (Stuttgart, Germany : 1980)",
  ["ultrason imaging"] = "Ultrasonic imaging",
  ["ultrason sonochem"] = "Ultrasonics Sonochemistry",
  ["ultrasound clin"] = "Ultrasound clinics",
  ["ultrasound int open"] = "Ultrasound international open",
  ["ultrasound j"] = "The ultrasound journal",
  ["ultrasound med biol"] = "Ultrasound in medicine & biology",
  ["ultrasound obstet gynecol"] = "Ultrasound in obstetrics & gynecology : the official journal of the International Society of Ultrasound in Obstetrics and Gynecology",
  ["ultrasound q"] = "Ultrasound quarterly",
  ["ultrastruct pathol"] = "Ultrastructural pathology",
  ["ulum va tiknuluzhii khaliji fars"] = "ʻUlūm va tiknuluzhī-i khalīj-i fars",
  ["ulus cerrahi derg"] = "Ulusal cerrahi dergisi",
  ["ulus travma acil cerrahi derg"] = "Ulusal travma ve acil cerrahi dergisi = Turkish journal of trauma & emergency surgery : TJTES",
  ["ulus travma derg"] = "Ulusal travma dergisi = Turkish journal of trauma & emergency surgery : TJTES",
  ["umap j"] = "The UMAP Journal",
  ["umkc law rev"] = "UMKC law review",
  ["umsch wiss tech"] = "Umschau in Wissenschaft und Technik",
  ["umw gesundh online"] = "Umwelt und gesundheit online",
  ["umweltmed forsch prax"] = "Umweltmedizine in Forschung und Praxis",
  ["umweltrecht prax"] = "Umweltrecht in der Praxis",
  ["umweltschutz schweiz"] = "Umweltschutz in der Schweiz",
  ["umweltwiss schadst forsch"] = "Umweltwissenschaften und Schadstoff-Forschung",
  ["un chron"] = "UN chronicle",
  ["un mon chron"] = "UN monthly chronicle",
  ["una commun"] = "UNA communiqué",
  ["una nurs j"] = "UNA nursing journal",
  ["uncertain comput tech decis intell"] = "Uncertainty, Computational Techniques, and Decision Intelligence",
  ["uncertain oper res"] = "Uncertainty and Operations Research",
  ["uncertain safe util mach learn med imaging graph biomed image anal (2020)"] = "Uncertainty for safe utilization of machine learning in medical imaging, and graphs in biomedical image analysis : Second International Workshop, UNSURE 2020, and Third International Workshop, GRAIL 2020, Held in Conjunction with MICCAI 2020, Lima, Peru, October 8, 2020, Proceedings",
  ["uncertain safe util machine learn med imaging clin image based proced (2019)"] = "Uncertainty for Safe Utilization of Machine Learning in Medical Imaging and Clinical Image-based Procedures : first International Workshop, UNSURE 2019, and 8th International Workshop, CLIP 2019, held in conjunction with MICCAI 2019, Shenzhen, China, October 17, 2019, Proceedings. UNSURE (Workshop) (1st : 2019 : Shenzhen Shi, China)",
  ["unconv resour"] = "Unconventional Resources",
  ["unctad rev"] = "UNCTAD Review",
  ["undergrad j philos"] = "The Undergraduate journal of philosophy",
  ["undergrad lect notes phys"] = "Undergraduate Lecture Notes in Physics",
  ["undergrad texts comput sci"] = "Undergraduate Texts in Computer Science",
  ["undergrad texts math"] = "Undergraduate Texts in Mathematics",
  ["undergrad texts math read math"] = "Undergraduate Texts in Mathematics. Readings in Mathematics",
  ["undergrad texts phys"] = "Undergraduate Texts in Physics",
  ["undergrad top comput sci"] = "Undergraduate Topics in Computer Science",
  ["underground space"] = "Underground Space",
  ["unders interv broaden particip res careers"] = "Understanding interventions that broaden participation in research careers",
  ["undersea biomed res"] = "Undersea biomedical research",
  ["undersea hyperb med"] = "Undersea & hyperbaric medicine : journal of the Undersea and Hyperbaric Medical Society, Inc",
  ["underst chem react"] = "Understanding Chemical Reactivity",
  ["underst child"] = "Understanding the child; a magazine for teachers",
  ["underst complex syst"] = "Understanding complex systems",
  ["underst interpret mach learn med image comput appl (2018)"] = "Understanding and interpreting machine learning in medical image computing applications : first International Workshops, MLCN 2018, DLF 2018, and iMIMIC 2018, held in conjunction with MICCAI 2018, Granada, Spain, September 16-20, 2018 : proceedings. MLCN (Workshop) (1st : 2018 : Granada, Spain)",
  ["undp news"] = "UNDP news : networking publication of UNDP staff worldwide. United Nations Development Programme",
  ["unesco bull libr"] = "Unesco bulletin for libraries",
  ["unesco cour"] = "The Unesco courier",
  ["unesco sources"] = "Unesco sources",
  ["ungyong tonggye yongu"] = "Ŭngyong tʻonggye yŏn'gu = The Korean journal of applied statistics",
  ["unicef news"] = "UNICEF news",
  ["unidad sanit"] = "Unidad sanitaria",
  ["unif distrib theory"] = "Uniform Distribution Theory",
  ["union burma j life sci"] = "Union of Burma journal of life sciences",
  ["union med can"] = "L'unión médicale du Canada",
  ["union med prat francais"] = "L'Union médicale des praticiens français",
  ["unipa springer ser"] = "UNIPA Springer Series",
  ["united european gastroenterol j"] = "United European gastroenterology journal",
  ["united evangelical action"] = "United evangelical action",
  ["united synag rev"] = "United Synagogue review",
  ["unitext phys"] = "Unitext for Physics",
  ["univ ariz bull"] = "University of Arizona Bulletin",
  ["univ ark little rock law j"] = "University of Arkansas at Little Rock law journal",
  ["univ b c law rev"] = "University of British Columbia law review. University of British Columbia",
  ["univ baltimore law rev"] = "University of Baltimore law review. University of Baltimore. School of Law",
  ["univ beograd publ elektrotehn fak ser mat"] = "Univerzitet u Beogradu",
  ["univ calif davis law rev"] = "University of California, Davis law review",
  ["univ calif publ pharmacol"] = "University of California publications in pharmacology",
  ["univ calif publ physiol"] = "University of California publications in physiology",
  ["univ calif publ zool"] = "University of California publications in zoology",
  ["univ chem educ"] = "University Chemistry Education",
  ["univ chic law rev"] = "The University of Chicago law review. University of Chicago. Law School",
  ["univ chic leg forum"] = "The University of Chicago legal forum",
  ["univ cincinnati law rev"] = "University of Cincinnati law review. University of Cincinnati. College of Law",
  ["univ coll hosp mag"] = "The University College Hospital magazine",
  ["univ colo law rev"] = "University of Colorado law review. University of Colorado (Boulder campus). School of Law",
  ["univ dayton law rev"] = "University of Dayton law review. University of Dayton. Law School",
  ["univ detroit j urban law"] = "University of Detroit journal of urban law",
  ["univ detroit law rev"] = "University of Detroit law review",
  ["univ detroit mercy law rev"] = "University of Detroit Mercy law review",
  ["univ durh med gaz"] = "University of Durham medical gazette",
  ["univ edinb j"] = "University of Edinburgh journal",
  ["univ fla j law public policy"] = "University of Florida journal of law and public policy : JL & PP",
  ["univ fla law rev"] = "University of Florida law review",
  ["univ ghana law j"] = "University of Ghana law journal",
  ["univ hosp bull"] = "University Hospital bulletin",
  ["univ iagel acta math"] = "Universitatis Iagellonicae. Acta Mathematica",
  ["univ ill law forum"] = "The University of Illinois law forum",
  ["univ ill law rev"] = "University of Illinois law review",
  ["univ joensuu dep phys math rep ser"] = "University of Joensuu, Department of Physics and Mathematics. Report Series",
  ["univ joensuu dept math rep ser"] = "University of Joensuu, Department of Mathematics Report Series",
  ["univ jordan publ"] = "University of Jordan Publications",
  ["univ kans law rev"] = "University of Kansas law review. University of Kansas. School of Law",
  ["univ lecture ser"] = "University Lecture Series",
  ["univ lehr stud naturwissenschaft"] = "Universitats-Lehr- und Studienbucher. Naturwissenschaften",
  ["univ louisv j fam law"] = "University of Louisville journal of family law",
  ["univ louisville j respir infect"] = "The University of Louisville journal of respiratory infections",
  ["univ manit med j"] = "The University of Manitoba medical journal",
  ["univ md law forum"] = "The University of Maryland law forum",
  ["univ memphis law rev"] = "The University of Memphis law review",
  ["univ miami entertain sports law rev"] = "University of Miami entertainment & sports law review",
  ["univ miami law rev"] = "University of Miami law review",
  ["univ mich j law reform"] = "University of Michigan journal of law reform. University of Michigan. Law School",
  ["univ mich med cent j"] = "University of Michigan Medical Center journal",
  ["univ minn med bull"] = "University of Minnesota medical bulletin",
  ["univ missouri bus govt rev"] = "University of Missouri Business and Government Review",
  ["univ n c news lett"] = "University Of North Carolina News Letter",
  ["univ n m orthop res j"] = "University of New Mexico orthopaedics research journal",
  ["univ n s w law j"] = "The University of New South Wales law journal",
  ["univ nac tucumán rev ser a"] = "Universidad Nacional de Tucumán",
  ["univ newcastle tyne med gaz"] = "University of Newcastle upon Tyne medical gazette",
  ["univ odontol"] = "Universitas odontológica : revista cientifica de la Facultad de Odontológica",
  ["univ oulu rep ser phys sci"] = "University of Oulu Report Series in Physical Sciences",
  ["univ pa j const law"] = "University of Pennsylvania journal of constitutional law",
  ["univ pa law rev"] = "University of Pennsylvania law review",
  ["univ pa med bull"] = "University of Pennsylvania medical bulletin",
  ["univ pittsbg law rev"] = "University of Pittsburgh law review. University of Pittsburgh. School of Law",
  ["univ politeh bucharest sci bull ser a"] = "University Politehnica of Bucharest Scientific Bulletin\tSeries A Applied Mathematics and Physics",
  ["univ politeh bucharest sci bull ser b"] = "University Politehnica of Bucharest Scientific Bulletin\tSeries B Chemistry and Materials Science",
  ["univ politeh bucharest sci bull ser c"] = "University Politehnica of Bucharest Scientific Bulletin\tSeries C Electrical Engineering and Computer Science",
  ["univ politeh bucharest sci bull ser d"] = "University Politehnica of Bucharest Scientific Bulletin\tSeries D Mechanical Engineering",
  ["univ puget sound law rev"] = "University of Puget Sound law review. University of Puget Sound. School of Law",
  ["univ qld law j"] = "The University of Queensland law journal",
  ["univ richmond law rev"] = "University of Richmond law review. University of Richmond",
  ["univ salud"] = "Universidad y salud",
  ["univ san fernando valley law review"] = "University of San Fernando Valley law review",
  ["univ san francisco law rev"] = "University of San Francisco law review. University of San Francisco. School of Law",
  ["univ sci (bogota)"] = "Universitas scientiarum",
  ["univ ser higher math"] = "The University Series in Higher Mathematics",
  ["univ ser math"] = "The University Series in Mathematics",
  ["univ south calif bull sch med"] = "School of Medicine. University of Southern California. School of Medicine",
  ["univ tasman law rev"] = "University of Tasmania law review",
  ["univ tehran sci humanit ser"] = "University of Tehran Science and Humanities Series",
  ["univ texts math sci"] = "University Texts in the Mathematical Sciences",
  ["univ toledo law rev"] = "University of Toledo law review. University of Toledo. College of Law",
  ["univ tor dent j"] = "University of Toronto dental journal",
  ["univ tor fac law rev"] = "University of Toronto Faculty of Law review",
  ["univ tor law j"] = "The University of Toronto law journal",
  ["univ tor stud biol ser"] = "University of Toronto studies. Biological series",
  ["univ toronto med j"] = "University of Toronto medical journal",
  ["univ toronto undergrad dent j"] = "University of Toronto undergraduate dental journal",
  ["univ wash bus rev"] = "University of Washington Business Review",
  ["univ west los angel law rev"] = "University of West Los Angeles law review",
  ["univ west ont law rev"] = "University of Western Ontario law review",
  ["univ west ont med j"] = "University of Western Ontario medical journal",
  ["univ western ontario ser philos sci"] = "University of Western Ontario Series in Philosophy of Science",
  ["univ wyo publ"] = "University of Wyoming publications (Laramie, Wyo. : 1935)",
  ["univer leeds med mag"] = "University of Leeds medical magazine",
  ["univers access hum comput interact (2009)"] = "Universal access in human-computer interaction : 5th international conference, UAHCI 2009, held as part of HCI International 2009, San Diego, CA, USA, July 19-24, 2009 : proceedings. International Conference on Universal Access in Human-Computer Interaction (5th : 2009 : San Diego, CA)",
  ["univers access hum comput interact (2017)"] = "Universal access in human-computer interaction : designing novel interactions : 11th International Conference, UAHCI 2017, held as part of HCI International 2017, Vancouver, BC, Canada, July 9-14, 2017, Proceedings. Part II. International Conference on Universal Access in Human-Computer Interaction (11th : 2017 : Vancouver, B.C.)",
  ["univers access inf soc"] = "Universal access in the information society",
  ["univers j agric res"] = "Universal journal of agricultural research",
  ["univers j electr electron eng"] = "Universal Journal of Electrical and Electronic Engineering",
  ["univers j microbiol res"] = "Universal journal of microbiology research",
  ["univers j public health"] = "Universal journal of public health",
  ["universale econom"] = "Universale Economica",
  ["universitas (ghana)"] = "Universitas (University of Ghana)",
  ["universitas (stuttg)"] = "Universitas",
  ["unmanned syst"] = "Unmanned Systems",
  ["unos update"] = "UNOS update",
  ["unscn nutr"] = "UNSCN nutrition",
  ["unser wald"] = "Unser Wald",
  ["unsere heim"] = "Unsere Heimat",
  ["unsolved probl in intuitive math"] = "Unsolved Problems in Intuitive Mathematics",
  ["unsupervised and semi-supervised learn"] = "Unsupervised and Semi-Supervised Learning",
  ["untertoggenburg neujahrsbl"] = "Untertoggenburger Neujahrsblätter",
  ["untimely medit"] = "Untimely Meditations",
  ["upb sci bull ser a"] = "U.P.B. Scientific Bulletin Series A Applied Mathematics and Physics",
  ["upb sci bull ser b"] = "U.P.B. Scientific Bulletin Series B Chemistry and Materials Science",
  ["upb sci bull ser c"] = "U.P.B. Scientific Bulletin Series C Electrical Engineering and Computer Science",
  ["upb sci bull ser d"] = "U.P.B. Scientific Bulletin Series D Mechanical Engineering",
  ["update cancer ther"] = "Update on cancer therapeutics",
  ["update counc accredit occup hearing conserv"] = "Update (Council for Accreditation in Occupational Hearing Conservation)",
  ["update ethics"] = "Update On Ethics",
  ["update natl minor aids counc"] = "Update (National Minority AIDS Council)",
  ["update pediatr dent"] = "Update in pediatric dentistry",
  ["update univ s c dep music"] = "Update (Music Educators National Conference (U.S.))",
  ["updates surg"] = "Updates in surgery",
  ["upr bolsh sist"] = "Upravlenie Bolshimi Sistemami",
  ["ups j med sci"] = "Upsala journal of medical sciences",
  ["ups j med sci suppl"] = "Upsala journal of medical sciences. Supplement",
  ["upsala lakareforen forh"] = "Upsala läkareförenings förhandlingar",
  ["upstream oil gas technol"] = "Upstream Oil and Gas Technology",
  ["upv j nat sci"] = "UPV journal of natural sciences",
  ["ur rep"] = "UR [reports]. U.S. Atomic Energy Commission",
  ["ural gos univ mat zap"] = "Ural\\cprime skiy Gosudarstvennyy Universitet im",
  ["ural math j"] = "Ural Mathematical Journal",
  ["urban aff q"] = "Urban affairs quarterly",
  ["urban aff rev thousand oaks calif"] = "Urban affairs review (Thousand Oaks, Calif.)",
  ["urban agric reg food syst"] = "Urban Agriculture & Regional Food Systems",
  ["urban anthropol"] = "Urban anthropology",
  ["urban anthropol stud cult syst world econ dev"] = "Urban anthropology and studies of cultural systems and world economic development",
  ["urban book ser"] = "The Urban Book Series",
  ["urban clim"] = "Urban Climate",
  ["urban ecol"] = "Urban Ecology",
  ["urban ecosyst"] = "Urban ecosystems",
  ["urban educ (beverly hills calif)"] = "Urban education",
  ["urban for urban green"] = "Urban forestry & urban greening",
  ["urban for urban greening"] = "Urban Forestry & Urban Greening",
  ["urban geogr"] = "Urban geography",
  ["urban health"] = "Urban health",
  ["urban health newsl"] = "Urbanisation and health newsletter",
  ["urban hist rev"] = "Urban history review. Revue d'histoire urbaine",
  ["urban hist yearb"] = "Urban history yearbook",
  ["urban history"] = "Urban history",
  ["urban inf"] = "Urban Informatics",
  ["urban lawyer"] = "The urban lawyer",
  ["urban leag rev"] = "The Urban League review. National Urban League. Research Dept",
  ["urban plan"] = "Urban planning",
  ["urban plan transp res"] = "Urban, planning and transport research",
  ["urban plann transport res"] = "Urban, Planning and Transport Research",
  ["urban rail transit"] = "Urban rail transit",
  ["urban res pract"] = "Urban research & practice",
  ["urban rev"] = "The Urban review",
  ["urban sci"] = "Urban science (Basel, Switzerland)",
  ["urban soc change review"] = "The Urban & social change review",
  ["urban soc work"] = "Urban social work",
  ["urban stud"] = "Urban studies (Edinburgh, Scotland)",
  ["urban transform"] = "Urban transformations",
  ["urban water j"] = "Urban water journal",
  ["uremia invest"] = "Uremia investigation",
  ["urine (amst)"] = "Urine (Amsterdam, Netherlands)",
  ["urol ann"] = "Urology annals",
  ["urol cancer care"] = "Urologists in cancer care",
  ["urol case rep"] = "Urology case reports",
  ["urol clin north am"] = "The Urologic clinics of North America",
  ["urol cutaneous rev"] = "The Urologic and cutaneous review",
  ["urol int"] = "Urologia internationalis",
  ["urol j"] = "Urology journal",
  ["urol listy"] = "Urologicke listy",
  ["urol mosc"] = "Urologii︠a︡ (Moscow, Russia : 1923)",
  ["urol nefrol (mosk)"] = "Urologii︠a︡ i nefrologii︠a︡",
  ["urol nurs"] = "Urologic nursing",
  ["urol oncol"] = "Urologic oncology",
  ["urol pol"] = "Urologia polska",
  ["urol pract"] = "Urology practice",
  ["urol radiol"] = "Urologic radiology",
  ["urol res"] = "Urological Research",
  ["urol sci"] = "Urological science",
  ["urol surv"] = "Urological survey",
  ["urol times"] = "Urology times",
  ["urol video j"] = "Urology video journal",
  ["urologe a"] = "Der Urologe. Ausg. A",
  ["urologe b"] = "Der Urologe. Ausg. B",
  ["ursi radio sci bull"] = "URSI Radio Science Bulletin",
  ["us army med dep j"] = "U.S. Army Medical Department journal",
  ["us cathol"] = "U.S. Catholic",
  ["us census bur cent econ stud res pap ser"] = "U.S. Census Bureau Center for Economic Studies research paper series",
  ["us clivar rep"] = "US CLIVAR reports",
  ["us dep agric bull"] = "United States Department of Agriculture Bulletin",
  ["us dep agric misc publ"] = "United States Department of Agriculture Miscellaneous Publication",
  ["us endocrinol"] = "US endocrinology",
  ["us gastroenterol hepatol rev"] = "US gastroenterology & hepatology review",
  ["us geol surv bull"] = "U.S. Geological Survey Bulletin",
  ["us geol surv prof pap"] = "U.S. Geological Survey Professional Paper",
  ["us geol surv sci invest rep"] = "U.S. Geological Survey Scientific Investigations Report",
  ["us geol surv water resour invest rep"] = "U.S. Geological Survery Water Resources  Investigations Report",
  ["us healthc"] = "U.S. healthcare",
  ["us jpn womens j engl suppl"] = "U.S.-Japan women's journal. English supplement = Nichi-Bei josei jānaru. English supplement",
  ["us long term rev"] = "U.S. long-term review",
  ["us med"] = "U. S. medicine",
  ["us navy med"] = "U. S. Navy medicine",
  ["us navy med news lett"] = "United States Navy medical news letter",
  ["us neurol"] = "US neurology",
  ["us news world rep"] = "U.S. news & world report",
  ["us obstet gynecol"] = "US obstetrics & gynaecology",
  ["us oncol"] = "US oncology",
  ["us oncol hematol"] = "US oncology & hematology",
  ["us ophthalmic rev"] = "US ophthalmic review",
  ["us pharm"] = "U. S. pharmacist",
  ["us psyc"] = "US psychiatry",
  ["us respir pulm dis"] = "US respiratory & pulmonary diseases",
  ["us statut large"] = "United States statutes at large",
  ["us supreme court rep"] = "United States Supreme Court reports. United States. Supreme Court",
  ["usa today"] = "USA today",
  ["usa weekend"] = "USA weekend",
  ["usaf med serv dig"] = "USAF Medical Service digest",
  ["usaid dev"] = "USAID developments",
  ["usaid highlights"] = "USAID highlights",
  ["user model user-adapt interact"] = "User modeling and user-adapted interaction",
  ["usp biol khim"] = "Uspekhi biologicheskoĭ khimii",
  ["usp fiz nauk [sov phys usp]"] = "Uspekhi Fizicheskikh Nauk [Soviet Physics–Uspekhi]",
  ["usp fiziol nauk"] = "Uspekhi fiziologicheskikh nauk",
  ["usp khim"] = "Uspekhi khimii",
  ["usp sovrem biol"] = "Uspekhi sovremennoĭ biologii",
  ["uspekhi mat nauk"] = "Rossiiskaya Akademiya Nauk. Moskovskoe Matematicheskoe Obshchestvo. Uspekhi Matematicheskikh Nauk",
  ["usqr union semin q rev"] = "Union Seminary quarterly review",
  ["utah code annot 1953 utah"] = "Utah code annotated 1953. Utah",
  ["utah hist q"] = "Utah historical quarterly",
  ["utah law rev"] = "Utah law review",
  ["utah nurse"] = "Utah nurse",
  ["utah sci"] = "Utah science",
  ["utah womens health rev"] = "The Utah women's health review",
  ["util math"] = "Utilitas Mathematica",
  ["util policy"] = "Utilities policy",
  ["utilitas math"] = "Utilitas Mathematica",
  ["utne read"] = "The Utne reader",
  ["utokyo eng course/basic math"] = "UTokyo Engineering Course/Basic Mathematics",
  ["utop stud"] = "Utopian studies",
  ["uttar pradesh state dent j"] = "The Uttar Pradesh State dental journal : an official publication of the State U.P. Dental Branch of the Indian Dental Association",
  ["uwfl rep"] = "UWFL [reports]. U.S. Atomic Energy Commission",
  ["uzb khim zh"] = "Uzbekskii Khimicheskii Zhurnal",
  ["uzbek mat zh"] = "Akademiya Nauk Respubliki Uzbekistan",
  ["uzbek math j"] = "Uzbek Mathematical Journal",
  ["va bar news"] = "Virginia bar news",
  ["va cavalcade"] = "Virginia cavalcade",
  ["va dent j"] = "Virginia dental journal",
  ["va explor"] = "The Virginia explorer",
  ["va health bull"] = "Virginia health bulletin",
  ["va j int law"] = "Virginia journal of international law",
  ["va j sci"] = "Virginia journal of science",
  ["va j soc policy law"] = "Virginia journal of social policy & the law",
  ["va law rev"] = "Virginia law review",
  ["va mag hist biogr"] = "The Virginia magazine of history and biography",
  ["va med"] = "Virginia medical",
  ["va med mon"] = "Virginia Medical Monthly",
  ["va med mon (1918)"] = "Virginia medical monthly",
  ["va med q"] = "Virginia medical quarterly : VMQ",
  ["va nurse"] = "Virginia nurse",
  ["va nurse q"] = "Virginia nurse quarterly",
  ["va q rev"] = "The Virginia quarterly review",
  ["va rev sociol"] = "Virginia review of sociology",
  ["vaccin res"] = "Vaccination research : open journal",
  ["vaccine (auckl)"] = "Vaccine (Auckland, N.Z.)",
  ["vaccine immun news"] = "Vaccine & immunization news : the newsletter of the Global Programme for Vaccines and Immunization",
  ["vaccine rep"] = "Vaccine reports",
  ["vaccine wkly"] = "Vaccine weekly",
  ["vaccine x"] = "Vaccine: X",
  ["vaccines (basel)"] = "Vaccines",
  ["vad j"] = "The VAD journal : the journal of mechanical assisted circulation and heart failure",
  ["vadose zone j"] = "Vadose Zone Journal",
  ["vak forsch prax"] = "Vakuum in Forschung und Praxis",
  ["vakbl biol"] = "Vakblad voor biologen",
  ["valparaiso univ law rev"] = "Valparaiso University law review. Valparaiso University. School of Law",
  ["value health"] = "Value in health : the journal of the International Society for Pharmacoeconomics and Outcomes Research",
  ["value health reg issues"] = "Value in health regional issues",
  ["van nostrand reinhold math stud"] = "Van Nostrand Reinhold Mathematics Studies",
  ["vanc sun"] = "Vancouver sun (1986)",
  ["vancouver stud cogn sci"] = "Vancouver Studies in Cognitive Science",
  ["vanderbilt j entertain technol law"] = "Vanderbilt journal of entertainment and technology law",
  ["vanderbilt law rev"] = "Vanderbilt law review",
  ["varahmihir j math sci"] = "Varahmihir Journal of Mathematical Sciences",
  ["vard nord utveckl forsk"] = "Vard i Norden",
  ["variorum collect stud ser"] = "Variorum Collected Studies Series",
  ["various publ ser"] = "Various Publications Series",
  ["vasa suppl"] = "VASA. Supplementum",
  ["vasc biol"] = "Vascular biology (Bristol, England)",
  ["vasc cell"] = "Vascular cell",
  ["vasc dis"] = "Vascular diseases",
  ["vasc dis prev"] = "Vascular disease prevention",
  ["vasc endovascular surg"] = "Vascular and endovascular surgery",
  ["vasc healt risk manage"] = "Vascular Health and Risk Management",
  ["vasc health risk manag"] = "Vascular health and risk management",
  ["vasc investig ther"] = "Vascular investigation and therapy",
  ["vasc med"] = "Vascular medicine (London, England)",
  ["vasc pharmacol"] = "Vascular Pharmacology",
  ["vasc specialist int"] = "Vascular specialist international",
  ["vasc surg"] = "Vascular surgery",
  ["vascul pharmacol"] = "Vascular pharmacology",
  ["vavilovskii zhurnal genet selektsii"] = "Vavilovskiĭ zhurnal genetiki i selekt︠s︡ii",
  ["vb news"] = "VB news",
  ["vdi ber"] = "VDI-Berichte",
  ["vdi bericht"] = "35 Vdi Jahrestagung Schadensanalyse:produktverbesserung Durch Schadensanalyse",
  ["vecteur environ"] = "Vecteur Environnement",
  ["vector borne zoonotic dis"] = "Vector borne and zoonotic diseases (Larchmont, N.Y.)",
  ["vector optim"] = "Vector Optimization",
  ["vector-borne zoonotic dis"] = "Vector-Borne and Zoonotic Diseases",
  ["veg bode"] = "Vegetarische bode",
  ["veg classif surv"] = "Vegetation Classification and Survey",
  ["veg crops  res bull"] = "Vegetable Crops Research Bulletin",
  ["veg crops res bull"] = "Vegetable Crops Research Bulletin",
  ["veg hist archaeobot"] = "Vegetation history and archaeobotany",
  ["veh commun"] = "Vehicular Communications",
  ["veh syst dyn"] = "Vehicle System Dynamics",
  ["venez odontol"] = "Venezuela odontológica",
  ["venoms toxins"] = "Venoms and Toxins",
  ["venture capital"] = "Venture Capital",
  ["verbum (nancy)"] = "Verbum (Nancy, France)",
  ["verh - int ver theor angew limnol"] = "Verhandlungen - Internationale Vereinigung für Theoretische und Angewandte Limnologie",
  ["verh anat ges"] = "Verhandlungen der Anatomischen Gesellschaft",
  ["verh dtsch ges herz kreislaufforsch"] = "Verhandlungen der Deutschen Gesellschaft fur Herz- und Kreislaufforschung",
  ["verh dtsch ges inn med"] = "Verhandlungen der Deutschen Gesellschaft fur Innere Medizin",
  ["verh dtsch ges kreislaufforsch"] = "Verhandlungen der Deutschen Gesellschaft fur Kreislaufforschung",
  ["verh dtsch ges pathol"] = "Verhandlungen der Deutschen Gesellschaft fur Pathologie",
  ["verh dtsch ges rheumatol"] = "Verhandlungen der Deutschen Gesellschaft fur Rheumatologie",
  ["verh dtsch phys ges"] = "Verhandlungen der Deutschen Physikalischen Gesellschaft",
  ["verh ges ökol"] = "Verhandlungen der Gesellschaft für Ökologie",
  ["verh int ver theor angew limnol"] = "Verhandlungen - Internationale Vereinigung fur Theoretische und Angewandte Limnologie",
  ["verh int ver theoret angew limnol"] = "Verhandlungen der Internationalen Vereinigung für Theoretische und Angewandte Limnologie",
  ["verh k acad geneeskd belg"] = "Verhandelingen - Koninklijke Academie voor Geneeskunde van België",
  ["verh k vlaam acad geneeskd belg"] = "Verhandelingen - Koninklijke Vlaamse Academie voor Geneeskunde van België",
  ["verh natforsch ges basel"] = "Verhandlungen der Naturforschenden Gesellschaft in Basel",
  ["verh naturwiss ver hambg (1980)"] = "Verhandlungen des Naturwissenschaftlichen Vereins in Hamburg (1980)",
  ["verh schweiz natforsch ges"] = "Verhandlungen der Schweizerischen naturforschenden Gesellschaft",
  ["verh zool bot ges osterr"] = "Verhandlungen der Zoologisch-Botanischen Gesellschaft in Österreich",
  ["verh zool bot ges wien"] = "Verhandlungen der Zoologisch-Botanischen Gesellschaft in Wien. Zoologisch-Botanische Gesellschaft in Wien",
  ["veridian ej sci technol silpakorn univ"] = "Veridian e-journal science and technology Silpakorn University",
  ["vermess photogramm kulttech"] = "Vermessung, Photogrammetrie, Kulturtechnik",
  ["vernac archit"] = "Vernacular Architecture",
  ["vernons annot revis civ statut state tex tex"] = "Vernon's annotated Revised civil statutes of the state of Texas. Texas",
  ["veroff dtsch hist inst london"] = "Veroffentlichungen des Deutschen Historischen Instituts London",
  ["veroff gesch naturwiss math med"] = "Veroffentlichungen zur Geschichte der Naturwissenschaften, Mathematik und Medizin",
  ["veroff int ges gesch pharm"] = "Veröffentlichungen der Internationalen Gesellschaft für Geschichte der Pharmazie e. V",
  ["veroff morphol pathol"] = "Veröffentlichungen aus der morphologischen Pathologie",
  ["veroff natkdmus (erfurt)"] = "Veröffentlichungen des Naturkundemuseums Erfurt",
  ["veroff pathol"] = "Veroffentlichungen aus der Pathologie",
  ["veroff schweiz ges gesch med naturwiss"] = "Veröffentlichungen der Schweizerischen Gesellschaft für Geschichte der Medizin und der Naturwissenschaften",
  ["veroff schweiz ges gesch pharm"] = "Veröffentlichungen der Schweizerischen Gesellschaft für Geschichte der Pharmazie",
  ["veroffentl inst bibliothekswissen bibliothekarausbild freien univ berlin"] = "Veroffentlichungen des Instituts fur Bibliothekswissenschaft und Bibliothekarausbildung der Freien Universitat Berlin",
  ["veroffentlichungen des forschungsinstituts des deutschen museums fur die geschichte der naturwissenschaften und der"] = "Veroffentlichungen des Forschungsinstituts des Deutschen Museums fur die Geschichte der Naturwissenschaften und der Technik, Reihe A: Kleine Mitteilungen",
  ["verpleegkd gem"] = "Verpleegkundigen en gemeenschapszorg : tijdschrift van de het Nationaal Verbond der Katholieke Vlaamse Verplegenden",
  ["versanst wasserbau hydrol glaziol"] = "Versuchsanstalt für Wasserbau, Hydrologie und Glaziologie",
  ["versl meded k viaam acad taal lett"] = "Verslagen en mededelingen - Koninklijke Academie voor Nederlandse Taal- en Letterkunde. Koninklijke Academie voor Nederlandse Taal- en Letterkunde",
  ["versl meded k vlaam acad taal lett"] = "Verslagen en mededelingen van de Koninklijke Vlaamse Academie voor Taal- en Letterkunde. Koninklijke Vlaamse Academie voor Taal- en Letterkunde",
  ["verslag meded betreff volksgezond"] = "Verslagen en mededelingen betreffende de volksgezondheid",
  ["verständl forsch"] = "Verständliche Forschung",
  ["vertebr anat morphol paleontol"] = "Vertebrate Anatomy Morphology Palaeontology",
  ["vertebr zool"] = "Vertebrate zoology",
  ["very short introd"] = "Very Short Introductions",
  ["veröff geobot inst eidgenöss tech hochsch, stift rübel  zür"] = "Veröffentlichungen des Geobotanischen Institutes der Eidg. Technischen Hochschule, Stiftung Rübel in Zürich",
  ["veröff landesstelle natschutz landschpfl baden-württ"] = "Veröffentlichungen der Landesstelle für Naturschutz und Landschaftspflege in Baden-Württemberg",
  ["veröff nat-mus luzern"] = "Veröffentlichungen Natur-Museum Luzern",
  ["veröffentl leibniz-archivs"] = "Veröffentlichungen des Leibniz-Archivs",
  ["vessel plus"] = "Vessel plus",
  ["vesta co ser pure appl math"] = "Vesta Company Series in Pure and Applied Mathematics",
  ["vestis lat psr zinat akad"] = "Vestis. Izvestiia. Latvijas PSR Zinātnu akadēmija",
  ["vestn akad med nauk sssr"] = "Vestnik Akademii meditsinskikh nauk SSSR",
  ["vestn akad nauk sssr"] = "Vestnik Akademii nauk SSSR",
  ["vestn beloruss gos univ im vi lenina ser 2 him biol geol geogr"] = "Vestnik Belorusskogo gosudarstvennogo universiteta imeni V.I. Lenina. Seriíà 2, Khimiíà, biologiíà, geografiíà",
  ["vestn beloruss gos univ ser 1 fiz mat inform"] = "Vestnik Belorusskogo Gosudarstvennogo Universiteta",
  ["vestn bgu ser 2 him biol geogr"] = "Vestnik BGU. Seriâ 2, himiâ, biologiâ, geografiâ",
  ["vestn ceskoslov lekaru"] = "Vĕstník c̆eskoslovenských lékar̆ů",
  ["vestn dermatol venerol"] = "Vestnik dermatologii i venerologii",
  ["vestn drevnej istor (moskva)"] = "Vestnik drevneĭ istorii. Journal of ancient history",
  ["vestn irkutsk gos teh univ"] = "Vestnik Irkutskogo gosudarstvennogo tehničeskogo universiteta",
  ["vestn kamcatsk gos teh univ"] = "Vestnik Kamčatskogo gosudarstvennogo tehničeskogo universiteta",
  ["vestn khir im i i grek"] = "Vestnik khirurgii imeni I. I. Grekova",
  ["vestn kraunts fiz-mat nauki"] = "Vestnik KRAUNTS. Fiziko-Matematicheskie Nauki",
  ["vestn leningr univ [biol]"] = "Vestnik Leningradskogo Universiteta. Biologiia",
  ["vestn leningr univ biol"] = "Vestnik Leningradskogo universiteta. Biologiia",
  ["vestn leningr univ geol geogr"] = "Vestnik - Leningradskogo Universiteta. Geologii︠a︡, geografii︠a︡. Leningradskiĭ gosudarstvennyĭ universitet",
  ["vestn leningr univ ser biol"] = "Vestnik Leningradskogo universiteta. Seriia biologii",
  ["vestn minister nauki akad nauk resp kaz"] = "Vestnik Ministerstva Nauki–Akademiya Nauk Respubliki Kazakhstan",
  ["vestn mord univ"] = "Vestnik Mordovskogo Universiteta",
  ["vestn mosk univ ser 8 istor"] = "Vestnik Moskovskogo universiteta. Serii︠a︡ VIII: Istorii︠a︡. Moskovskiĭ gosudarstvennyĭ universitet im. M.V. Lomonosova",
  ["vestn mosk univ ser v"] = "Vestnik Moskovskogo universiteta. Serii︠a︡ V, Geografii︠a︡",
  ["vestn mosk univ, ser 2: khim"] = "Vestnik Moskovskogo Universiteta, Seriya 2: Khimiya",
  ["vestn mosk univ, ser 3: fiz, astron"] = "Vestnik Moskovskogo Universiteta, Seriya 3: Fizika, Astronomiya",
  ["vestn oftalmol"] = "Vestnik oftalmologii",
  ["vestn ohotoved"] = "Vestnik ohotovedeniâ",
  ["vestn otorinolaringol"] = "Vestnik otorinolaringologii",
  ["vestn permsk natsionalnogo issledovatelskogo politekh univ prikl ekol urban"] = "Vestnik Permskogo nat︠s︡ionalʹnogo issledovatelʹskogo politekhnicheskogo universiteta. Prikladnai︠a︡ ėkologii︠a︡, urbanistika",
  ["vestn permsk univ ser biol"] = "Vestnik permskogo universiteta : Seriâ biologiâ",
  ["vestn rentgenol radiol"] = "Vestnik rentgenologii i radiologii",
  ["vestn ross akad med nauk"] = "Vestnik Rossiiskoi Akademii Meditsinskikh Nauk",
  ["vestn ross univ druhzby narodov ser agron zhivotnovod"] = "Vestnik Rossiĭskogo universiteta druzhby narodov. Serii︠a︡ agronomii︠a︡ i zhivotnovodstvo",
  ["vestn s-peterb univ, ser 4: fiz, khim"] = "Vestnik Sankt-Peterburgskogo Universiteta, Seriya 4: Fizika, Khimiya",
  ["vestn samar univ estestvennonauchn ser"] = "Vestnik Samarskogo Universiteta. Estestvennonauchnaya Seriya",
  ["vestn sarat gos akad prava"] = "Vestnik Saratovskoj gosudarstvennoj akademii prava",
  ["vestn sev mezdunar univ"] = "Vestnik severnogo meždunarodnogo universiteta",
  ["vestn severo vost nauchnogo tsentra dalnevost otd ran"] = "Vestnik Severo-Vostochnogo nauchnogo t︠s︡entra Dalʹnevostochnogo otdelenii︠a︡ RAN",
  ["vestn slov kem drus"] = "Vestnik Slovenskega Kemijskega Drustva",
  ["vestn st peterbg univ ser istor azykozn literaturoved"] = "Vestnik Sankt-Peterburgskogo universiteta. Serii︠a︡ 2, Istorii︠a︡, i︠a︡zykoznanie, literaturovedenie",
  ["vestn st-peterbg univ mat mekh astron"] = "Vestnik Sankt-Peterburgskogo Universiteta. Matematika. Mekhanika. Astronomiya",
  ["vestn st-peterbg univ prikl mat inform protsessy upr"] = "Vestnik Sankt-Peterburgskogo Universiteta. Prikladnaya Matematika. Informatika. Protsessy Upravleniya",
  ["vestn statistiki"] = "Vestnik statistiki (Moscow, Russia : 1949)",
  ["vestn stpetersburg univ, math"] = "Vestnik St. Petersburg University, Mathematics",
  ["vestn tomsk gos univ"] = "Vestnik Tomskogo gosudarstvennogo universiteta",
  ["vestn tomsk gos univ biol"] = "Vestnik Tomskogo gosudarstvennogo universiteta. Biologiâ",
  ["vestn tomsk gos univ mat mech"] = "Vestnik Tomskogo Gosudarstvennogo Universiteta. Matematika i Mechanika",
  ["vestn udmurt univ mat mekh kompyut nauki"] = "Vestnik Udmurtskogo Universiteta. Matematika. Mekhanika. Kompyuternye Nauki",
  ["vestn venerol dermatol"] = "Vestnik venerologii i dermatologii",
  ["vestn vogis"] = "Vestnik VOGiS",
  ["vestn voronezh gos univ ser fiz, mat"] = "Vestnik Voronezhskogo Gosudarstvennogo Universiteta. Seriya Fizika, Matematika",
  ["vestn zas rast"] = "Vestnik zashchity rasteniĭ = Plant protection news",
  ["vestn zool"] = "Vestnik zoologii",
  ["vestnik akad nauk respub kazakhstan"] = "Akademiya Nauk Respubliki Kazakhstan",
  ["vestnik chelyabinsk univ ser 3 mat mekh"] = "Chelyabinskiy Universitet",
  ["vestnik khar\\cprime kov univ"] = "Khar\\cprime kovskiy Universitet",
  ["vestnik kiev univ model optim slozhn sist"] = "Kievskiy Universitet",
  ["vestnik l\\cprime vov politekhn inst"] = "Vestnik L\\cprime vovskogo Politekhnicheskogo Instituta",
  ["vestnik moskov univ ser i mat mekh"] = "Vestnik Moskovskogo Universiteta. Seriya I. Matematika, Mekhanika",
  ["vestnik moskov univ ser iii fiz astronom"] = "Vestnik Moskovskogo Universiteta. Seriya III. Fizika, Astronomiya",
  ["vestnik moskov univ ser xv vychisl mat kibernet"] = "Vestnik Moskovskogo Universiteta. Seriya XV. Vychislitelnaya Matematika i Kibernetika",
  ["vestnik ross akad nauk"] = "Rossíiskaya Akademiya Nauk",
  ["vestnik rossiiskoi akademii selskokhoziaistvennykh nauk"] = "Vestnik Rossiĭskoĭ akademii sel'skokhozi︠a︡ĭstvennykh nauk : dvukhmesi︠a︡chnyĭ nauchno-teoreticheskiĭ zhurnal",
  ["vestnik s-peterburg univ fiz khim"] = "Vestnik Sankt-Peterburgskogo Universiteta",
  ["vestnik s-peterburg univ mat mekh astronom"] = "Vestnik Sankt-Peterburgskogo Universiteta",
  ["vestnik saratovskogo gosudarstvennogo universiteta"] = "Vestnik Saratovskogo gosudarstvennogo universiteta",
  ["vestnik st petersburg univ math"] = "Vestnik St. Petersburg University. Mathematics",
  ["vestsi akad navuk bssr, ser biyal navuk"] = "Vestsi Akademii Navuk Belarusi, Seryya Biyalagichnykh Navuk",
  ["vestsi akad navuk bssr, ser fiz-energ navuk"] = "Vestsi Akademii Navuk Belarusi, Seryya Fizika-Energetychnykh Navuk",
  ["vestsi akad navuk bssr, ser khim navuk"] = "Vestsi Akademii Navuk Belarusi, Seryya Khimichnykh Navuk",
  ["vestsi nats akad navuk belarusi ser fiz-mat navuk"] = "Vestsi Natsyyanalnai Akademii Navuk Belarusi. Seryya Fizika-Matematychnykh Navuk. Izvestiya Natsionalnoi Akademii Nauk Belarusi. Seriya Fiziko-Matematicheskikh Nauk. Proceedings of the National Academy of Sciences of Belarus. Physics and Mathematics Series",
  ["vestsi nats akad navuk belarusi, ser khim navuk"] = "Vestsi Natsyyanal’nai Akademii Navuk Belarusi, Seryya Khimichnykh Navuk",
  ["vestsi natsyianalnai akademii navuk belarusi"] = "Vestsi Natsyi︠a︡nalʹnaĭ akadėmii navuk Belarusi. Seryi︠a︡ bii︠a︡lahichnykh navuk",
  ["vet (kiev)"] = "Veterinarii︠a︡ (Ukraine. Ministerstvo silʹsʹkoho hospodarstva)",
  ["vet anaesth analg"] = "Veterinary anaesthesia and analgesia",
  ["vet anim sci"] = "Veterinary and animal science",
  ["vet arh"] = "Veterinarski arhiv",
  ["vet bull"] = "The Veterinary bulletin",
  ["vet clin n am"] = "Veterinary Clinics of North America",
  ["vet clin n am: equine pract"] = "Veterinary Clinics of North America: Equine Practice",
  ["vet clin n am: exot anim pract"] = "Veterinary Clinics of North America: Exotic Animal Practice",
  ["vet clin n am: food anim pract"] = "Veterinary Clinics of North America: Food Animal Practice",
  ["vet clin n am: large anim pract"] = "Veterinary Clinics of North America: Large Animal Practice",
  ["vet clin n am: small anim pract"] = "Veterinary Clinics of North America: Small Animal Practice",
  ["vet clin north am"] = "The Veterinary clinics of North America",
  ["vet clin north am equine pract"] = "The Veterinary clinics of North America. Equine practice",
  ["vet clin north am exot anim pract"] = "The veterinary clinics of North America. Exotic animal practice",
  ["vet clin north am food anim pract"] = "The Veterinary clinics of North America. Food animal practice",
  ["vet clin north am large anim pract"] = "The Veterinary clinics of North America. Large animal practice",
  ["vet clin north am small anim pract"] = "The Veterinary clinics of North America. Small animal practice",
  ["vet clin pathol"] = "Veterinary clinical pathology",
  ["vet comp oncol"] = "Veterinary and comparative oncology",
  ["vet comp orthop traumatol"] = "Veterinary and comparative orthopaedics and traumatology : V.C.O.T",
  ["vet dermatol"] = "Veterinary dermatology",
  ["vet ext q"] = "Veterinary extension quarterly",
  ["vet fak derg"] = "Veteriner Fakültesi dergisi",
  ["vet glas"] = "Veterinarski glasnik",
  ["vet herit"] = "Veterinary heritage : bulletin of the American Veterinary History Society",
  ["vet hist"] = "Veterinary history",
  ["vet historisch genoot cah"] = "Veterinair Historisch Genootschap cahier",
  ["vet hum toxicol"] = "Veterinary and Human Toxicology",
  ["vet immunol immunopathol"] = "Veterinary immunology and immunopathology",
  ["vet ital"] = "Veterinaria italiana",
  ["vet j"] = "Veterinary journal (London, England : 1997)",
  ["vet med"] = "Veterinary medicine",
  ["vet med (auckl)"] = "Veterinary medicine (Auckland, N.Z.)",
  ["vet med (praha)"] = "Veterinární medicína",
  ["vet med int"] = "Veterinary medicine international",
  ["vet med j"] = "Veterinary medical journal (Jīzah, Egypt : 1995)",
  ["vet med nachr"] = "Veterinär-medizinische Nachrichten",
  ["vet med nauki"] = "Veterinarno-meditsinski nauki",
  ["vet med rev"] = "Veterinary medical review",
  ["vet med sci"] = "Veterinary medicine and science",
  ["vet med small anim clin"] = "Veterinary medicine, small animal clinician : VM, SAC",
  ["vet med: res rep"] = "Veterinary Medicine: Research and Reports",
  ["vet microbiol"] = "Veterinary microbiology",
  ["vet ophthalmol"] = "Veterinary ophthalmology",
  ["vet parasitol"] = "Veterinary parasitology",
  ["vet parasitol reg stud reports"] = "Veterinary parasitology, regional studies and reports",
  ["vet parasitol x"] = "Veterinary parasitology: X",
  ["vet pathol"] = "Veterinary pathology",
  ["vet pathol suppl"] = "Veterinary pathology. Supplement",
  ["vet q"] = "The veterinary quarterly",
  ["vet radiol ultrasound"] = "Veterinary radiology & ultrasound : the official journal of the American College of Veterinary Radiology and the International Veterinary Radiology Association",
  ["vet rec"] = "The Veterinary record",
  ["vet rec case rep"] = "Veterinary record case reports",
  ["vet rec open"] = "Veterinary record open",
  ["vet res"] = "Veterinary research",
  ["vet res commun"] = "Veterinary research communications",
  ["vet res forum"] = "Veterinary research forum : an international quarterly journal",
  ["vet rev annot"] = "Veterinary reviews and annotations",
  ["vet sci"] = "Veterinary sciences",
  ["vet sci china"] = "Veterinary science in China",
  ["vet sci commun"] = "Veterinary science communications",
  ["vet segodnia"] = "Veterinarii︠a︡ segodni︠a︡ = Veterinary science today",
  ["vet surg"] = "Veterinary surgery : VS",
  ["vet ther"] = "Veterinary therapeutics : research in applied veterinary medicine",
  ["vet world"] = "Veterinary world",
  ["veterinaria abaj"] = "Veterinariâ (Abaj)",
  ["veterinary clin north am exot anim pract"] = "Veterinary Clinics of North America. Exotic Animal Practice",
  ["veterynama medytsyna"] = "Veterynarna medyt︠s︡yna",
  ["vgb kraftwerkstech"] = "VGB Kraftwerkstechnik",
  ["viata med"] = "Viaţa medicală",
  ["viata med rev inf prof stiint cadrelor medii sanit"] = "Viaţa medicală; revistă de informare profesională şi ştiinţifică a cadrelor medii sanitare",
  ["vib spectrosc"] = "Vibrational Spectroscopy",
  ["vibrant (brasilia)"] = "Vibrant : virtual Brazilian anthropology",
  ["vic hist j"] = "The Victorian historical journal",
  ["vic hist mag"] = "The Victorian historical magazine",
  ["vic lit cult"] = "Victorian literature and culture",
  ["vic newsl"] = "The Victorian newsletter",
  ["vic period newsl"] = "Victorian periodicals newsletter",
  ["vic period rev"] = "Victorian periodicals review",
  ["vic stud"] = "Victorian studies",
  ["vic vet proc"] = "Victorian veterinary proceedings",
  ["vichinfektsiia immunosuppr"] = "Vich-infekt︠s︡ii︠a︡ i immunosupresii",
  ["vict offender"] = "Victims & offenders",
  ["victor bostrom fund rep"] = "Report. Victor-Bostrom Fund",
  ["vida etica"] = "Vida y ética",
  ["vida med"] = "Vida médica",
  ["vida nueva"] = "Vida nueva",
  ["vida odontol"] = "Vida Odontologica",
  ["video assist thorac surg"] = "Video-assisted thoracic surgery",
  ["video j prosthet urol"] = "Video journal of prosthetic urology",
  ["videourology (new rochelle)"] = "Videourology (New Rochelle, N.Y.)",
  ["vie lang"] = "Vie et langage",
  ["vie med"] = "Vie médicale (Paris, France : 1920)",
  ["vie med can fr"] = "La Vie médicale au Canada français",
  ["vie med evolut med ther"] = "La Vie médicale. Evolution médicale et thérapeutique",
  ["vie milieu"] = "Vie et milieu (Paris, France : 1980)",
  ["vie milieu paris"] = "Vie et milieu (Paris, France : 1980)",
  ["vie sante"] = "Vie et santé",
  ["vienna circ inst libr"] = "Vienna Circle Institute Library",
  ["vienna circ inst yearb"] = "Vienna Circle Institute Yearbook",
  ["vienna yearb popul res"] = "Vienna yearbook of population research",
  ["vierteljahresschr dermatol syph"] = "Vierteljahresschrift fur Dermatologie und Syphilis",
  ["vierteljahrs wirtschaftsforschung"] = "Vierteljahrsheftze zur Wirtschaftsforschung",
  ["vierteljahrschr soz wirtschaftsgesch"] = "Vierteljahrschrift für Sozial- und Wirtschaftsgeschichte",
  ["vierteljahrsh zeitgesch"] = "Vierteljahrshefte für Zeitgeschichte",
  ["vierteljahrsschr natforsch ges zür"] = "Vierteljahrsschrift der Naturforschenden Gesellschaft in Zürich",
  ["vierteljahrsschr schweiz sanitatsoff"] = "Vierteljahrsschrift für schweizerische Sanitätsoffiziere. Journal trimestriel des officiers suisses du service de santé. Rivista trimestrale degli ufficiali sanitari svizzeri",
  ["vietnam j chem"] = "Vietnam Journal of Chemistry",
  ["vietnam j comput sci"] = "Vietnam Journal of Computer Science",
  ["vietnam j math"] = "Vietnam Journal of Mathematics",
  ["vietnam j sci technol"] = "Vietnam journal of science and technology",
  ["vietnam stud"] = "Vietnamese studies (Hanoi, Vietnam : 1966)",
  ["vieweg math"] = "Vieweg Mathematik",
  ["vieweg programmbibl taschenrechner"] = "Vieweg Programmbibliothek. Taschenrechner",
  ["vieweg stud aufbaukurs math"] = "Vieweg Studium: Aufbaukurs Mathematik",
  ["vieweg stud grundkurs math"] = "Vieweg Studium: Grundkurs Mathematik",
  ["vieweg textbook math"] = "Vieweg Textbook Mathematics",
  ["vieweg tracts pure appl phys"] = "Vieweg Tracts in Pure and Applied Physics",
  ["viewp dig dis"] = "Viewpoints on digestive diseases",
  ["vigo int j appl linguist"] = "Vigo international journal of applied linguistics",
  ["viitorul soc"] = "Viitorul social",
  ["vijnana parishad anusandhan patrika"] = "Vijnana Parishad Anusandhan Patrika. The Research Journal of the Hindi Science Academy",
  ["vikram math j"] = "The Vikram Mathematical Journal",
  ["villaclara med"] = "Villaclara médica",
  ["village voice"] = "Village voice (Greenwich Village, New York, N.Y.)",
  ["villanova law rev"] = "Villanova law review",
  ["vina q"] = "Vina quarterly",
  ["vingtieme siecle"] = "Vingtième siècle (Paris, France : 1984)",
  ["violence against women"] = "Violence against women",
  ["violence gend"] = "Violence and gender",
  ["violence update"] = "Violence Update",
  ["violence vict"] = "Violence and victims",
  ["vipimage 2017 (2017)"] = "VipIMAGE 2017 : proceedings of the VI ECCOMAS Thematic Conference on Computational Vision and Medical Image Processing Porto, Portugal, October 18-20, 2017. VipIMAGE (Conference) (2017 : Porto, Portugal)",
  ["vipimage 2019 (2019)"] = "VipIMAGE 2019 : proceedings of the VII ECCOMAS Thematic Conference on Computational Vision and Medical Image Processing, October 16-18, 2019, Porto, Portugal. VipIMAGE (Conference) (2019 : Porto, Portugal)",
  ["viral hepat pract"] = "Viral hepatitis in practice",
  ["viral immunol"] = "Viral immunology",
  ["virchows arch"] = "Virchows Archiv : an international journal of pathology",
  ["virchows arch a pathol anat histol"] = "Virchows Archiv. A, Pathological anatomy and histology",
  ["virchows arch a pathol anat histopathol"] = "Virchows Archiv. A, Pathological anatomy and histopathology",
  ["virchows arch a pathol pathol anat"] = "Virchows Archiv. A, Pathology. Pathologische Anatomie",
  ["virchows arch b cell pathol"] = "Virchows Archiv. B, Cell pathology",
  ["virchows arch b cell pathol incl mol pathol"] = "Virchows Archiv. B, Cell pathology including molecular pathology",
  ["virchows arch pathol anat physiol klin med"] = "Virchows Archiv für pathologische Anatomie und Physiologie und für klinische Medizin",
  ["virol j"] = "Virology journal",
  ["virol monogr"] = "Virology monographs. Die Virusforschung in Einzeldarstellungen",
  ["virol mycol"] = "Virology & mycology : infectious diseases",
  ["virol rep"] = "Virology Reports",
  ["virol sin"] = "Virologica Sinica",
  ["virologica sinica|ping tu hsueh tsa chih"] = "Bing du xue za zhi",
  ["virologica sinica|zhongguo bing du xue"] = "Zhongguo bing du xue",
  ["virologie (montrouge)"] = "Virologie (Montrouge, France)",
  ["virology (auckl)"] = "Virology : research and treatment",
  ["virology (hyderabad)"] = "Virology (Hyderabad)",
  ["virtual augment mixed real (2020)"] = "Virtual, augmented and mixed reality : industrial and everyday life applications : 12th International Conference, VAMR 2020, held as part of the 22nd HCI International Conference, HCII 2020, Copenhagen, Denmark, July 19-24, 2020, Proceedings. Part II. VAMR (Conference) (12th : 2020 : Online)",
  ["virtual j at quantum fluids"] = "Virtual Journal of Atomic Quantum Fluids",
  ["virtual lab"] = "The Virtual Laboratory",
  ["virtual mentor"] = "The virtual mentor : VM",
  ["virtual phys prototyp"] = "Virtual and physical prototyping",
  ["virtual phys prototyping"] = "Virtual and Physical Prototyping",
  ["virtual real"] = "Virtual reality",
  ["virtual reality"] = "Virtual Reality",
  ["virtual ser symplectic geom"] = "Virtual Series on Symplectic Geometry",
  ["virus evol"] = "Virus evolution",
  ["virus genes"] = "Virus genes",
  ["virus inf exch newsl south east asia west pac"] = "Virus information exchange newsletter for South-East Asia and the Western Pacific",
  ["virus res"] = "Virus research",
  ["virus res suppl"] = "Virus research. Supplement",
  ["virus rev res"] = "Virus reviews & research : journal of the Brazilian Society for Virology",
  ["vis anthropol"] = "Visual anthropology",
  ["vis anthropol rev"] = "Visual anthropology review : journal of the Society for Visual Anthropology",
  ["vis cogn"] = "Visual cognition",
  ["vis comput"] = "The Visual computer",
  ["vis comput ind biomed art"] = "Visual computing for industry, biomedicine, and art",
  ["vis cult br"] = "Visual culture in Britain",
  ["vis data anal"] = "Visualization and data analysis",
  ["vis dev rehabil"] = "Vision development and rehabilitation",
  ["vis impair res"] = "Visual impairment research",
  ["vis j emerg med"] = "Visual journal of emergency medicine",
  ["vis neurosci"] = "Visual neuroscience",
  ["vis stud"] = "Visual studies",
  ["vis ukr tov genet sel"] = "Visnyk Ukraïnsʹkoho tovarystva henetykiv i selekt︠s︡ioneriv",
  ["visc med"] = "Visceral medicine",
  ["vishwa int j graph theory"] = "Vishwa International Journal of Graph Theory",
  ["visible lang"] = "Visible language",
  ["vision (basel)"] = "Vision (Basel, Switzerland)",
  ["vision res"] = "Vision research",
  ["vision syst des"] = "Vision Systems Design",
  ["visnik harkivskogo nacionalnogo universitetu imeni v n karazina seria biologia"] = "Visnyk Kharkivsʹkoho nat︠s︡ionalʹnoho universytetu imeni V.N. Karazina. Serii︠a︡ biolohii︠a︡",
  ["visnyk kyivskoho natsionalnoho universytetu imeni tarasa shevchenka biolohiia"] = "Visnyk Kyïvsʹkoho nat︠s︡ionalʹnoho universytetu imeni Tarasa Shevchenka. Biolohii︠a︡",
  ["visnyk lvivskoho universytetu seriia biolohichna"] = "Visnyk L'vivs'koho universytetu. Seriia biolohichna",
  ["vistas astron"] = "Vistas in Astronomy",
  ["vistas astronom"] = "Vistas in Astronomy",
  ["vistas in astron"] = "Vistas in Astronomy",
  ["visual cognit"] = "Visual Cognition",
  ["visual comput"] = "Visual Computer",
  ["visual comput ind, biomed, art"] = "Visual Computing for Industry, Biomedicine, and Art",
  ["visual geosci"] = "Visual Geosciences",
  ["visual neurosci"] = "Visual Neuroscience",
  ["visualization eng"] = "Visualization in Engineering",
  ["visualization image proc comput biomed"] = "Visualization, Image Processing and Computation in Biomedicine",
  ["visualization mech processes: int online j"] = "Visualization of Mechanical Processes: An International Online Journal",
  ["visualized cancer med"] = "Visualized Cancer Medicine",
  ["vita hum int z lebensalterforsch"] = "Vita humana. Internationale Zeitschrift für Lebensaltersforschung. International journal of human development. Journal international de développement humain",
  ["vita malacol"] = "Vita malacologica",
  ["vita math"] = "Vita Mathematica",
  ["vitae scholasticae"] = "Vitae scholasticae",
  ["vital health stat 1"] = "Vital and health statistics. Ser. 1, Programs and collection procedures",
  ["vital health stat 10"] = "Vital and health statistics. Series 10, Data from the National Health Survey",
  ["vital health stat 11"] = "Vital and health statistics. Series 11, Data from the National Health Survey",
  ["vital health stat 13"] = "Vital and health statistics. Series 13, Data from the National Health Survey",
  ["vital health stat 14"] = "Vital and health statistics. Series 14, Data on national health resources",
  ["vital health stat 2"] = "Vital and health statistics. Series 2, Data evaluation and methods research",
  ["vital health stat 20"] = "Vital and health statistics. Series 20, Data from the National Vital Statistics System",
  ["vital health stat 20 data mortal"] = "Vital and health statistics. Series 20, Data on mortality",
  ["vital health stat 20 data natl vital stat syst"] = "Vital and Health Statistics. Series 20: Data from the National Vital Statistics System",
  ["vital health stat 21"] = "Vital and health statistics. Series 21, Data from the National Vital Statistics System",
  ["vital health stat 23"] = "Vital and health statistics. Series 23, Data from the National Survey of Family Growth",
  ["vital health stat 3"] = "Vital & health statistics. Series 3, Analytical and epidemiological studies",
  ["vital health stat 4"] = "Vital and health statistics. Ser. 4, Documents and committee reports",
  ["vital health stat 5"] = "Vital and health statistics. Series 5, Comparative international vital and health statistics reports",
  ["vital health stat ser 16"] = "Vital and health statistics. Series 16, Compilations of advance data from vital and health statistics",
  ["vital health stat ser 24"] = "Vital and health statistics. Series 24, Compilations of data on natality, mortality, marriage, divorce, and induced terminations of pregnancy",
  ["vital signs"] = "Vital Signs",
  ["vital speeches day"] = "Vital speeches of the day",
  ["vital stat (ott annu ed)"] = "Vital statistics",
  ["vital stat spec rep"] = "Vital statistics, special reports",
  ["vital stat u s (hyattsville)"] = "Vital statistics of the United States",
  ["vitalst zivilisationskr"] = "Vitalstoffe-Zivilisationskrankheiten",
  ["vitam d dig"] = "Vitamin D digest",
  ["vitam horm"] = "Vitamins and hormones",
  ["vitam horm (ny)"] = "Vitamins and Hormones (New York)",
  ["vitam horm leipzig"] = "Vitamine und Hormone",
  ["vitic data j"] = "Viticulture Data Journal",
  ["viva orig"] = "Viva origino. [Origin of life]",
  ["vivechan int j res"] = "Vivechan international journal of research",
  ["vivre autrement"] = "Vivre autrement",
  ["vivre val d"] = "Vivre en Val-d'Oise",
  ["vlaams diergeneeskd tijdschr"] = "Vlaams diergeneeskundig tijdschrift",
  ["vladikavkaz mat zh"] = "Vladikavkazskii Matematicheskii Zhurnal",
  ["vldb j"] = "The VLDB journal : very large data bases : a publication of the VLDB Endowment",
  ["vlinders (wagening)"] = "Vlinders",
  ["vlsi des"] = "VLSI Design",
  ["vnitr lek"] = "Vnitr̆ní lékar̆ství",
  ["vnr new math library"] = "VNR New Mathematics Library",
  ["voces desde trab soc"] = "Voces desde el trabajo social",
  ["voen istor zh"] = "Voenno-istoricheskii zhurnal",
  ["voen med delo"] = "Voenno meditsinsko delo",
  ["voen med zh"] = "Voenno-Meditsinskii Zhurnal",
  ["voen meditsinskii sb"] = "Voenno-meditsinskiĭ sbornik",
  ["voice speech rev"] = "Voice & speech review",
  ["voix dent"] = "La Voix dentaire",
  ["voix femme"] = "La voix de la femme : magazine trimestriel d'information, d'éducation et de mobilisation de l'APDF",
  ["vojen zdrav listy"] = "Vojenské zdravotnické listy",
  ["vojenskozdrav knih"] = "Vojenskozdravotnická knihovna",
  ["vojnois glas"] = "Vojno-istoriski glasnik",
  ["vojnosanit pregl"] = "Vojnosanitetski Pregled",
  ["vol graph"] = "Volume graphics. International Symposium on Volume Graphics",
  ["vol homenaje"] = "Volúmenes de Homenaje",
  ["volksgesundheit bad hombg"] = "Volksgesundheit (Bern, Switzerland)",
  ["volta rev"] = "The Volta review",
  ["volta voices"] = "Volta voices",
  ["volunt action leadersh"] = "Voluntary action leadership",
  ["volunt adm"] = "Volunteer administration",
  ["volunt leader"] = "The Volunteer leader",
  ["vom zahlstein zum comput"] = "Vom Zahlstein zum Computer",
  ["vopr biokhim mozga"] = "Voprosy biokhimii mozga",
  ["vopr biokhimii"] = "Voprosy biokhimii. [Problems of biochemistry]",
  ["vopr ekon"] = "Voprosy ėkonomiki",
  ["vopr eksper klinicheskoi urol"] = "Voprosy Eksperimentalnoi I Klinicheskoi Urologii",
  ["vopr elektropatol elektrotravmatizma elektrobezop"] = "Voprosy e̊lektropatologii, e̊lektrotravmatizma i e̊lektrobezopasnosti",
  ["vopr filos"] = "Voprosy filosofii",
  ["vopr fiziol"] = "Voprosy fiziologii",
  ["vopr geogr"] = "Voprosy geografii",
  ["vopr istor"] = "Voprosy istorii (Moscow, Russia : 1945)",
  ["vopr istor estestvozn tekh"] = "Voprosy istorii estestvoznanii︠a︡ i tekhniki (Institut istorii estestvoznanii︠a︡ i tekhniki (Akademii︠a︡ nauk SSSR))",
  ["vopr istor kpss"] = "Voprosy istorii KPSS (Moscow, Russia)",
  ["vopr klin eksp oftalmol"] = "Voprosy klinicheskoĭ i eksperimental'noĭ oftalmologiĭ",
  ["vopr klin eksp urol"] = "Voprosy klinicheskoĭ i e̊ksperimental'noĭ urologii; sbornik trudov",
  ["vopr klin lecheniia zlokachestvennykh novoobraz"] = "Voprosy kliniki i lecheniia zlokachestvennykh novoobrazovaniĭ",
  ["vopr kurortol fizioter lech fiz kult"] = "Voprosy kurortologii, fizioterapii, i lechebnoĭ fizicheskoĭ kultury",
  ["vopr med khim"] = "Voprosy Meditsinskoi Khimii",
  ["vopr med khimii"] = "Voprosy meditsinskoĭ khimii",
  ["vopr med virusol"] = "Voprosy meditsinskof virusologii",
  ["vopr neirokhir"] = "Voprosy neĭrokhirurgii",
  ["vopr okhr materin det"] = "Voprosy okhrany materinstva i detstva",
  ["vopr onkol"] = "Voprosy onkologii",
  ["vopr patol serdechno sosud sist"] = "Voprosy patologii serdechno-sosudistoĭ sistemy",
  ["vopr pediatrii"] = "Voprosy pediatrii",
  ["vopr pitan"] = "Voprosy pitaniia",
  ["vopr psikhiatr nevropatol"] = "Voprosy psikhiatrii i nevropatologii; sbornik trudov",
  ["vopr psikhiatrii nevrol"] = "Voprosy psikhīatrīi i nevrologīi",
  ["vopr psikhol"] = "Voprosy psikhologii",
  ["vopr revm"] = "Voprosy revmatizma",
  ["vopr sots gig organ zdravookhr istor med"] = "Voprosy sotsial'noĭ gigieny, organizatsii zdravookhraneniia i istorii meditsiny",
  ["vopr virusol"] = "Voprosy virusologii",
  ["vopr vychisl prikl mat"] = "Voprosy Vychislitelnoi i Prikladnoi Matematiki. Hisoblash va Amaliy Matematika Masalalari",
  ["vopros stat"] = "Voprosy statistiki (Moscow, Russia : 1994)",
  ["voprosy dinamiki i prochnosti"] = "Rizhskiy Politekhnicheskiy Institut",
  ["voprosy istor estestvoznan i tekhn"] = "Rossíiskaya Akademiya Nauk",
  ["voprosy mekh protsess upravl"] = "\\cyr Voprosy Mekhaniki i Protsessov Upravleniya",
  ["voprosy optim plan upravl sots ekonom"] = "Voprosy Optimalnogo Planirovaniya i Upravleniya Sotsialisticheskoi Ekonomikoi",
  ["voprosy teor sistem avtomat upravleniya"] = "Leningradskiy Universitet",
  ["voprosy vychisl i prikl mat"] = "Akademiya Nauk Respubliki Uzbekistan",
  ["vorarlb naturschau"] = "Vorarlberger naturschau",
  ["vorlesungen fachbereich math univ essen"] = "Vorlesungen aus dem Fachbereich Mathematik der Universität GH Essen",
  ["vorlesungen math inst giessen"] = "Vorlesungen aus dem Mathematischen Institut Giessen",
  ["vortr nordrh-westfäl akad wiss"] = "Vorträge",
  ["vorträge nordrhein-westfälische akad wiss"] = "Nordrhein-Westfälische Akademie der Wissenschaften",
  ["vostok (mosk)"] = "Vostok (Moscow, Russia : 1991)",
  ["vox sang"] = "Vox sanguinis",
  ["voyag math"] = "Voyages en Mathematiques",
  ["vrach delo"] = "Vrachebnoe delo",
  ["vt hist"] = "Vermont history",
  ["vt hist news"] = "Vermont history news",
  ["vt law rev"] = "Vermont law review",
  ["vt regist nurse"] = "Vermont registered nurse",
  ["vt statut annot vt"] = "Vermont statutes annotated. Vermont",
  ["vtt symp"] = "VTT Symposium",
  ["vulnerable child youth stud"] = "Vulnerable children and youth studies",
  ["vutr boles"] = "Vŭtreshni bolesti",
  ["vychisl prikl mat (kiev)"] = "Vychislitel\\cprime naya i Prikladnaya Matematika",
  ["vychisl sistemy"] = "Rossíiskaya Akademiya Nauk",
  ["vychisl tekhn sistemy upravlenie"] = "Vychislitel\\cprime naya Tekhnika",
  ["vychisl tekhn sots stran"] = "Vychislitelnaya Tekhnika Sotsialisticheskikh Stran",
  ["vychisl tekhn vopr kibern"] = "Vychislitel\\cprime naya Tekhnika i Voprosy Kibernetiki",
  ["vychisl tekhn voprosy kibernet"] = "Leningradskiy Gosudarstvennyy Universitet",
  ["vychisl tekhnol"] = "Vychislitelnye Tekhnologii. Computational Technologies",
  ["vychislitelnyi tsentr moskovskogo gosudarstvennogo universiteta seriya: statistika i stokhasticheskie"] = "Vychislitelnyi Tsentr Moskovskogo Gosudarstvennogo Universiteta. Seriya: Statistika i Stokhasticheskie Sistemi,",
  ["vysokochist veshchestva"] = "Vysokochistye Veshchestva",
  ["vysokomol soedin, ser a"] = "Vysokomolekulyarnye Soedineniya, Seriya A",
  ["vysokomol soedin, ser a ser b"] = "Vysokomolekulyarnye Soedineniya, Seriya A i Seriya B",
  ["vysokomol soedin, ser b"] = "Vysokomolekulyarnye Soedineniya, Seriya B: Kratkie Soobshcheniya",
  ["vyziva lidu"] = "Výz̆iva lidu",
  ["vögel heim"] = "Vögel der Heimat",
  ["w v dent j"] = "West Virginia dental journal",
  ["w v med j"] = "The West Virginia medical journal",
  ["w va hist"] = "West Virginia history",
  ["wachira wechasan"] = "[Wachira wēchasān] Vajīra medical journal",
  ["wadc tech note united states air force wright air dev cen day ohio"] = "WADC technical note. Wright Air Development Center",
  ["wadc tech rep united states air force wright air dev cent day ohio"] = "WADC technical report. Wright Air Development Center",
  ["waimm j earth resour eng"] = "WAIMM Journal of Earth Resources Engineering",
  ["wakayama med rep"] = "Wakayama medical reports",
  ["wake forest law rev"] = "Wake Forest law review",
  ["waking sleeping"] = "Waking and sleeping",
  ["walailak j sci technol"] = "Walailak journal of science and technology",
  ["wald holz"] = "Wald und Holz",
  ["wald- holzwirtsch"] = "Wald- und Holzwirtschaft, Die",
  ["walkerana (ann arbor mich)"] = "Walkerana : transactions of the POETS Society",
  ["wall st j (east ed)"] = "Wall Street journal (Eastern ed.)",
  ["wall st j [midwest ed]"] = "Wall Street Journal. Midwest Edition",
  ["wall st j midwest ed"] = "Wall Street journal (Midwest ed.)",
  ["wall str j"] = "Wall Street journal (Southwest ed.)",
  ["walter eucken inst wirtschaftswissensch wirtschaftsrechtl unters"] = "Walter Eucken Institut. Wirtschaftswissenschaftliche und Wirtschaftsrechtliche Untersuchungen",
  ["war dep tech bull tb med"] = "War Department technical bulletin. TB MED",
  ["war hist"] = "War in history",
  ["war hung"] = "War on hunger; a report from the Agency for International Development",
  ["war med (chic 1941)"] = "War medicine",
  ["war rep"] = "Warreport : bulletin of the Institute for War & Peace Reporting",
  ["war soc"] = "War & society",
  ["warasan kanpramong"] = "Wārasān kānpramong = Thai fisheries gazette",
  ["warasan khana witthayasat maha witthayalai chiang mai"] = "Wārasān Khana Witthayāsāt, Mahāwitthayālai Chīang Mai = Journal of the Science Faculty of Chiang Mai University",
  ["warasan phesatchasat"] = "Wārasān phēsatchasāt = Mahidol University journal of pharmaceutical sciences",
  ["warasan phrueksasat thai"] = "Warasan phrueksasat Thai",
  ["warasan phuetsat songkhlanakharin"] = "Warasan phuetsat songkhlanakharin",
  ["warasan prachakon lae sangkhom"] = "Wārasān prachākō̜n læ sangkhom = Journal of population and social studies",
  ["warasan rok phut"] = "Wārasān rōk phư̄t",
  ["warasan songkhla nakharin"] = "Wārasān Songkhlā Nakharin",
  ["warasan technol suranaree"] = "Warasan technology suranaree",
  ["warasan wichai lae songsoem wichakan kaset"] = "Warasan wichai lae songsoem wichakan kaset",
  ["warasan wichai mo kho"] = "Warasān wičhai Mō̧.Khō̧. = KKU research journal",
  ["warasan wichai mo kho (chobap bandit sueksa)"] = "Warasan wichai Mo Kho. (Chobap Bandit Sueksa)",
  ["warasan wichai witthayasat kanphaet"] = "Wārasān wičhai witthayāsāt kānphǣt = Thai journal of health research",
  ["warasan wichakan kaset"] = "Wārasān wichākān kasēt : Thai agricultural research journal",
  ["warasan wichakan witthayasat lae the noloyi mahawithayalai ratchaphat nakhon sawan"] = "Warasan wichakan witthayasat lae the Noloyi Mahawithayalai Ratchaphat Nakhon Sawan",
  ["warasan witthayasart burapha"] = "Warasan Witthayasart Burapha",
  ["warasan witthayasat mahawitthayalai naresuan"] = "Warasan Witthayasat Mahawitthayalai Naresuan",
  ["warburg inst stud texts"] = "Warburg Institute Studies and Texts",
  ["warta demografi"] = "Warta Demografi",
  ["warwickshire hist"] = "Warwickshire history",
  ["wash drug device lett"] = "Washington drug & device letter",
  ["wash j environ law policy"] = "Washington journal of environmental law & policy",
  ["wash law rev"] = "Washington law review (Seattle, Wash. : 1962)",
  ["wash lee law rev"] = "Washington and Lee law review",
  ["wash memo alan guttmacher inst"] = "Washington memo",
  ["wash mon"] = "The Washington monthly",
  ["wash nurse"] = "The Washington nurse",
  ["wash q"] = "The Washington quarterly",
  ["wash rep"] = "WASH [reports]. U.S. Atomic Energy Commission",
  ["wash rep med health"] = "Washington report on medicine & health",
  ["wash state dent j"] = "Washington State dental journal",
  ["wash state j nurs"] = "Washington State journal of nursing",
  ["wash state j public health pract"] = "Washington State journal of public health practice",
  ["wash times"] = "Washington times (Washington, D.C. : 1982)",
  ["wash univ dent j"] = "Washington University dental journal",
  ["wash univ j am indian alsk native health"] = "Washington University journal of American Indian & Alaska Native health",
  ["wash univ j urban contemp law"] = "Washington University journal of urban and contemporary law",
  ["wash univ law q"] = "Washington University law quarterly. Washington University (Saint Louis, Mo.). School of Law",
  ["wash univ med alumni q"] = "The Washington University medical alumni quarterly",
  ["washburn law j"] = "Washburn law journal",
  ["washington post"] = "The Washington post",
  ["wasser boden"] = "Wasser und Boden",
  ["wasser energ luft"] = "wasser, energie, luft",
  ["wasser- energwirtsch"] = "Wasser- und Energiewirtschaft",
  ["wasserwirtsch wassertech"] = "Wasserwirtschaft Wassertechnik",
  ["waste biomass valorization"] = "Waste and Biomass Valorization",
  ["waste dispos sustain energy"] = "Waste disposal & sustainable energy",
  ["waste disposal sustainable energy"] = "Waste Disposal & Sustainable Energy",
  ["waste manag"] = "Waste management (New York, N.Y.)",
  ["waste manag res"] = "Waste management & research : the journal of the International Solid Wastes and Public Cleansing Association, ISWA",
  ["waste manage (oxford)"] = "Waste Management",
  ["waste manage (tucson, ariz)"] = "Waste Management (Tucson, Arizona)",
  ["waste manage res"] = "Waste Management and Research",
  ["waste treat recovery"] = "Waste Treatment and Recovery",
  ["water (basel)"] = "Water",
  ["water air soil pollut"] = "Water, air, and soil pollution",
  ["water air soil pollut focus"] = "Water, Air, & Soil Pollution: Focus",
  ["water altern"] = "Water Alternatives",
  ["water asset manage int"] = "Water Asset Management International",
  ["water biol secur"] = "Water Biology and Security",
  ["water bodem lucht"] = "Water, bodem, lucht",
  ["water conserv manage"] = "Water Conservation and Management",
  ["water conserv sci eng"] = "Water Conservation Science and Engineering",
  ["water cycle"] = "Water Cycle",
  ["water econ policy"] = "Water Economics And Policy",
  ["water energy int"] = "Water and Energy International",
  ["water environ j"] = "Water Environment Journal",
  ["water environ manage"] = "Water and Environmental Management",
  ["water environ res"] = "Water environment research : a research publication of the Water Environment Federation",
  ["water hist"] = "Water history",
  ["water int"] = "Water International",
  ["water intell online"] = "Water Intelligence Online",
  ["water policy"] = "Water Policy",
  ["water pract technol"] = "Water Practice Technology",
  ["water qual ecosyst model"] = "Water Quality and Ecosystems Modeling",
  ["water qual expo health"] = "Water quality, exposure, and health",
  ["water qual exposure health"] = "Water Quality, Exposure and Health",
  ["water qual res j"] = "Water Quality Research Journal",
  ["water qual res j can"] = "Water Quality Research Journal of Canada",
  ["water res"] = "Water research",
  ["water res x"] = "Water research X",
  ["water resour"] = "Water Resources",
  ["water resour econ"] = "Water Resources and Economics",
  ["water resour ind"] = "Water Resources and Industry",
  ["water resour invest rep us geol surv"] = "Water Resources Investigations Report (united States Geological Survey)",
  ["water resour manage"] = "Water Resources Management",
  ["water resour prot"] = "Water Resources Protection",
  ["water resour res"] = "Water Resources Research",
  ["water resour rural dev"] = "Water Resources and Rural Development",
  ["water resources res"] = "Water Resources Research",
  ["water reuse"] = "Water Reuse",
  ["water sa"] = "Water SA",
  ["water sci"] = "Water Science",
  ["water sci eng"] = "Water Science and Engineering",
  ["water sci technol"] = "Water science and technology : a journal of the International Association on Water Pollution Research",
  ["water sci technol lib"] = "Water Science and Technology Library",
  ["water sci technol libr"] = "Water Science and Technology Library",
  ["water sci technol water supply"] = "Water Science and Technology Water Supply",
  ["water secur"] = "Water security",
  ["water sew works"] = "Water & sewage works",
  ["water sewage effluent"] = "Water Sewage and Effluent",
  ["water supply"] = "Water Supply",
  ["water util manage int"] = "Water Utilities Management International",
  ["water waves"] = "Water Waves",
  ["water works sewerage"] = "Water works and sewerage",
  ["water, air, soil pollut"] = "Water, Air, & Soil Pollution",
  ["water-energy nexus"] = "Water-Energy Nexus",
  ["watermark (arch libr hist health sci)"] = "Watermark (Archivists and Librarians in the History of the Health Sciences)",
  ["watershed ecol environ"] = "Watershed Ecology and the Environment",
  ["wave motion"] = "Wave Motion",
  ["wavelet anal appl"] = "Wavelet Analysis and its Applications",
  ["waves random complex media"] = "Waves in Random and Complex Media. Propagation, Scattering and Imaging",
  ["waves random complex medium"] = "Waves in Random and Complex Media",
  ["waves random media"] = "Waves in Random Media",
  ["waves wavelets fractals"] = "Waves, Wavelets and Fractals",
  ["wayne law rev"] = "Wayne law review",
  ["wda j"] = "WDA journal",
  ["we blind"] = "We the blind",
  ["we int"] = "WE international",
  ["wea forecasting"] = "Weather and Forecasting",
  ["wearable technol"] = "Wearable Technologies",
  ["weather clim dyn"] = "Weather and Climate Dynamics",
  ["weather clim extrem"] = "Weather and climate extremes",
  ["weather clim extremes"] = "Weather and Climate Extremes",
  ["weather clim soc"] = "Weather, Climate, and Society",
  ["weather forecast"] = "Weather and forecasting",
  ["weather forecasting"] = "Weather and Forecasting",
  ["weather vane"] = "The Weather vane",
  ["web ecol"] = "Web Ecology",
  ["web semant"] = "Web semantics (Online)",
  ["weed biol manag"] = "Weed biology and management",
  ["weed biol manage"] = "Weed Biology and Management",
  ["weed res"] = "Weed research",
  ["weed sci"] = "Weed Science",
  ["weed technol"] = "Weed technology : a journal of the Weed Science Society of America",
  ["weed turfgrass sci"] = "Weed & turfgrass science",
  ["weeds world"] = "Weeds world : the international electronic arabidopsis newsletter",
  ["weekend argus"] = "Weekend argus",
  ["weekend mag"] = "Weekend magazine",
  ["weekly mail"] = "Weekly mail (Johannesburg, South Africa)",
  ["weevil news"] = "Weevil news",
  ["wehrmed mitt"] = "Wehrmedizinische Mitteilungen",
  ["wehrmed monatsschr"] = "Wehrmedizinische Monatsschrift",
  ["wei sheng wu hsueh pao"] = "Wei Sheng Wu Hsueh Pao [Acta Microbiologica Sinica]",
  ["wei sheng wu xue bao"] = "Wei sheng wu xue bao = Acta microbiologica Sinica",
  ["wei sheng wu xue mian yi xue jin zhan"] = "Wei sheng wu xue mian yi xue jin zhan",
  ["wei sheng wu xue tong bao"] = "Wei sheng wu xue tong bao",
  ["wei sheng wu xue za zhi"] = "Wei sheng wu xue za zhi",
  ["wei sheng wu yu gan ran"] = "Wei sheng wu yu gan ran",
  ["wei sheng yan jiu"] = "Wei sheng yan jiu = Journal of hygiene research",
  ["wei sheng yen chiu"] = "Wei Sheng Yen Chiu Journal of Hygiene Research (Pei-Ching Shih",
  ["wei ti gu sheng wu xue bao"] = "Wei ti gu sheng wu xue bao = Acta micropalaeontologica Sinica",
  ["weichangbingxue he ganbingxue zazhi"] = "Weichangbingxue he ganbingxue zazhi = Chinese journal of gastroenterology and hepatology",
  ["wein wiss"] = "Wein Wissenschaft",
  ["weinwirtsch tech"] = "Weinwirtschaft Technik",
  ["weld int"] = "Welding International",
  ["weld j"] = "Welding journal",
  ["weld world"] = "Welding in the World",
  ["welding j"] = "Welding Journal",
  ["weleda korrespondenzbl aerzte"] = "Weleda Korrespondenzblätter für Aerzte",
  ["welf bull"] = "The Welfare bulletin",
  ["welf rev"] = "Welfare in review",
  ["wellbeing space soc"] = "Wellbeing, Space and Society",
  ["wellcome open res"] = "Wellcome open research",
  ["welsh hist rev"] = "The Welsh history review. Cylchgrawn hanes cymru",
  ["weltall bildgeschmuckte z astron verw geb"] = "Das Weltall. Bildgeschmuckte Zeitschrift fur Astronomie und verwandte Gebiete",
  ["weltwirtsch arch"] = "Weltwirtschaftliches Archiv",
  ["wen wu"] = "Wen wu (1959)",
  ["wendepunkt leben leiden"] = "Der Wendepunkt im Leben und im Leiden",
  ["wereld zending"] = "Wereld en Zending",
  ["werkst korros"] = "Werkstoffe und Korrosion",
  ["werkstatt bion evol"] = "Werkstatt Bionik und Evolutionstechnik",
  ["werkstatt gesch"] = "Werkstatt geschichte",
  ["wesleyan theol j"] = "Wesleyan theological journal",
  ["west afr"] = "West Africa",
  ["west afr j appl ecol"] = "West African Journal of Applied Ecology",
  ["west afr j med"] = "West African journal of medicine",
  ["west afr j pharmacol drug res"] = "West African Journal of Pharmacology and Drug Research",
  ["west afr j radiol"] = "West African journal of radiology",
  ["west afr j ultrasound"] = "West African journal of ultrasound",
  ["west afr med j"] = "The West African medical journal",
  ["west afr med j niger med dent pract"] = "The West African medical journal and Nigerian medical & dental practitioner",
  ["west afr med j niger pract"] = "The West African medical journal and Nigerian practitioner",
  ["west afr pharm"] = "West African pharmacist",
  ["west aust nat"] = "The Western Australian naturalist",
  ["west birds"] = "Western birds",
  ["west can j anthropol"] = "The Western Canadian journal of anthropology",
  ["west crim rev"] = "Western criminology review",
  ["west dent soc bull"] = "Bulletin (Western Dental Society)",
  ["west drug (vanc)"] = "The Western druggist",
  ["west engl med j"] = "West of England medical journal",
  ["west eur polit"] = "West European politics",
  ["west folk"] = "Western folklore",
  ["west hist q"] = "The Western historical quarterly",
  ["west hum rev"] = "The Western humanities review",
  ["west ill reg stud"] = "Western Illinois regional studies",
  ["west indian med j"] = "The West Indian medical journal",
  ["west indian ocean j mar sci"] = "Western Indian Ocean Journal of Marine Science",
  ["west j appl for"] = "Western Journal of Applied Forestry",
  ["west j black stud"] = "The Western journal of black studies",
  ["west j commun"] = "Western journal of communication",
  ["west j emerg med"] = "The western journal of emergency medicine",
  ["west j med"] = "The Western journal of medicine",
  ["west j med surg"] = "Western journal of medicine and surgery",
  ["west j nurs res"] = "Western journal of nursing research",
  ["west j speech commun"] = "Western journal of speech communication : WJSC",
  ["west j surg obstet gynecol"] = "Western journal of surgery, obstetrics, and gynecology",
  ["west lond med j"] = "West London medical journal",
  ["west med med j west"] = "Western medicine; the medical journal of the west",
  ["west n am nat"] = "Western North American naturalist",
  ["west new engl law rev"] = "Western New England law review",
  ["west ont law rev"] = "Western Ontario law review",
  ["west ont ser philos sci"] = "Western Ontario Series in Philosophy of Science",
  ["west pa hist mag"] = "The Western Pennsylvania historical magazine",
  ["west philos ser"] = "Western Philosophy Series",
  ["west polit q"] = "The Western political quarterly",
  ["west state univ law rev"] = "Western State University law review",
  ["west states jew hist"] = "Western states Jewish history",
  ["west states jew hist q"] = "Western states Jewish historical quarterly",
  ["west tenn hist soc pap"] = "The West Tennessee Historical Society papers. West Tennessee Historical Society",
  ["west tex hist asso year b"] = "The West Texas Historical Association year book. West Texas Historical Association",
  ["west va law rev"] = "West Virginia law review",
  ["west wildl"] = "Western wildlife (Albany, Calif.)",
  ["west's atl report"] = "West's Atlantic reporter",
  ["westchest med bull"] = "Westchester medical bulletin",
  ["western econ j"] = "Western Economic Journal",
  ["western j agr econ"] = "Western Journal of Agricultural Economics",
  ["western pac surveill response j"] = "Western Pacific surveillance and response journal : WPSAR",
  ["westminst inst rev"] = "Westminster Institute review",
  ["wests annot calif codes calif"] = "West's annotated California codes. California",
  ["wests annot indiana code"] = "West's annotated Indiana code : under arrangement of the official Indiana code. Indiana",
  ["wests calif report"] = "West's California reporter",
  ["wests fed rep"] = "West's federal reporter : cases argued and determined in the United States courts of appeals and Temporary Emergency Court of Appeals",
  ["wests fed rules decis"] = "West's federal rules decisions : opinions, decisions and rulings involving the Federal rules of civil procedure and Federal rules of criminal procedure",
  ["wests fed suppl"] = "West's federal supplement",
  ["wests fed suppl (perm ed)"] = "West's federal supplement (Permanent edition)",
  ["wests fla statut annot fla"] = "West's Florida statutes annotated : under arrangement of the official Florida statutes. Florida",
  ["wests n y suppl"] = "West's New York supplement",
  ["wests north east rep"] = "West's north eastern reporter",
  ["wests north west rep"] = "West's north western reporter",
  ["wests pac report"] = "West's Pacific reporter",
  ["wests revis code wash annot wash state"] = "West's revised code of Washington annotated. Washington (State)",
  ["wests south east report"] = "West's south eastern reporter",
  ["wests south report"] = "West's southern reporter",
  ["wests south west report"] = "West's south western reporter",
  ["wests supreme court report"] = "West's Supreme Court reporter",
  ["wests wis statut annot wis"] = "West's Wisconsin statutes annotated. Wisconsin",
  ["west’s atl report"] = "West’s Atlantic reporter",
  ["wet tijd"] = "Wetenschappelijke tijdingen",
  ["wetchasan sattawaphaet"] = "Wētchasān sattawaphaet = The Thai journal of veterinary medicine",
  ["wetl ecol manag"] = "Wetlands ecology and management",
  ["wetland sci"] = "Wetland Science",
  ["wetland sci pract"] = "Wetland Science and Practice",
  ["wetlands (wilmington)"] = "Wetlands (Wilmington, N.C.)",
  ["wetlands ecol manage"] = "Wetlands Ecology and Management",
  ["wetter leben"] = "Wetter und Leben",
  ["whats new"] = "What's new",
  ["whats new ind hyg"] = "What's new in industrial hygiene",
  ["whats new plant physiol"] = "What's new in plant physiology",
  ["what’s new for res"] = "What’s New in Forest Research",
  ["whittier law rev"] = "Whittier law review",
  ["who aids tech bull"] = "World Health Organization AIDS technical bulletin",
  ["who chron"] = "WHO chronicle",
  ["who feature"] = "WHO features",
  ["who offset publ"] = "WHO offset publication",
  ["who reg publ eur ser"] = "WHO regional publications. European series",
  ["who south east asia j public health"] = "WHO South-East Asia journal of public health",
  ["whole earth rev"] = "Whole earth review",
  ["whos who am art"] = "Who's who in American art; a biographical directory of contemporary artists, editors, critics, executives, etc",
  ["why sci anc world matter"] = "Why the Sciences of the Ancient World Matter",
  ["wiad lek"] = "Wiadomości lekarskie (Warsaw, Poland : 1960)",
  ["wiad lek wars pol 1948"] = "Wiadomości lekarskie (Warsaw, Poland : 1948)",
  ["wiad mat"] = "Wiadomosci Matematyczne",
  ["wiad parazytol"] = "Wiadomosci Parazytologiczne",
  ["wiad stat (warsaw)"] = "Wiadomości statystyczne (Warsaw, Poland : 1956)",
  ["wiadom mat"] = "Roczniki Polskiego Towarzystwa Matematycznego",
  ["widening partic lifelong learn"] = "Widening participation and lifelong learning : the journal of the Institute for Access Studies and the European Access Network",
  ["wideochir inne tech maloinwazyjne"] = "Wideochirurgia i inne techniki małoinwazyjne = Videosurgery and other miniinvasive techniques",
  ["wiederherstellungschir traumatol"] = "Wiederherstellungschirurgie und Traumatologie. Reconstruction surgery and traumatology",
  ["wien arb philos reihe a univstud"] = "Wiener Arbeiten zur Philosophie",
  ["wien arch psychol psychiatr neurologie"] = "Wiener Archiv für Psychologie, Psychiatrie und Neurologie",
  ["wien beitr chir"] = "Wiener Beiträge zur Chirurgie",
  ["wien beitr dermatol"] = "Wiener Beiträge zur Dermatologie",
  ["wien beitr geburtshilfe gynakol"] = "Wiener Beiträge zur Geburtshilfe und Gynäkologie",
  ["wien beitr gesch med"] = "Wiener Beiträge zur Geschichte der Medizin (Universitaẗ Wien. Institut für Geschichte der Medizin)",
  ["wien beitr gesch neuzeit"] = "Wiener Beiträge zur Geschichte der Neuzeit",
  ["wien beitr hals nasen ohrenheilkd"] = "Wiener Beiträge zur Hals-, Nasen- und Ohrenheilkunde",
  ["wien beitr hyg"] = "Wiener Beiträge zur Hygiene",
  ["wien beitr inneren med"] = "Wiener Beiträge zur inneren Medizin",
  ["wien beitr kinderheilkd"] = "Wiener Beiträge zur Kinderheilkunde",
  ["wien beitr zahnheilkd"] = "Wiener Beiträge zur Zahnheilkunde",
  ["wien geschichtsbl"] = "Wiener Geschichtsblätter",
  ["wien klin mag"] = "Wiener klinisches Magazin : Beilage zur Wiener klinischen Wochenschrift",
  ["wien klin wochenschr"] = "Wiener klinische Wochenschrift",
  ["wien klin wochenschr educ"] = "Wiener klinische Wochenschrift. Education",
  ["wien klin wochenschr suppl"] = "Wiener klinische Wochenschrift. Supplementum",
  ["wien med wochenschr"] = "Wiener medizinische Wochenschrift (1946)",
  ["wien med wochenschr beih"] = "Wiener medizinische Wochenschrift. Beihefte",
  ["wien med wochenschr suppl"] = "Wiener medizinische Wochenschrift. Supplement",
  ["wien reihe"] = "Wiener Reihe",
  ["wien stud med gesch philos"] = "Wiener Studien zur Medizin, Geschichte und Philosophie",
  ["wien tierarztl monatsschr"] = "Wiener Tierarztliche Monatsschrift",
  ["wien z gesch neuzeit"] = "Wiener Zeitschrift zur Geschichte der Neuzeit",
  ["wien z inn med"] = "Wiener Zeitschrift für innere Medizin und ihre Grenzgebiete",
  ["wien z kunde morgenl"] = "Wiener Zeitschrift für die Kunde des Morgenlandes",
  ["wien z nervenheilkd grenzgeb"] = "Wiener Zeitschrift fur Nervenheilkunde und deren Grenzgebiete",
  ["wien z prakt psychol"] = "Wiener Zeitschrift für praktische Psychologie",
  ["wies wspolcz"] = "Wieś współczesna; pismo ruchu ludowego",
  ["wijsgerig perspect maatsch wet"] = "Wijsgerig perspectief op maatschappij en wetenschap",
  ["wildbach- lawinenverbau"] = "Wildbach- und Lawinenverbau",
  ["wildern environ med"] = "Wilderness & Environmental Medicine",
  ["wilderness environ med"] = "Wilderness & environmental medicine",
  ["wildfowl (slimbr)"] = "Wildfowl. Wildfowl Trust",
  ["wildl biol"] = "Wildlife Biology",
  ["wildl biol pract"] = "Wildlife biology in practice (Online)",
  ["wildl dis"] = "Wildlife disease",
  ["wildl lett"] = "Wildlife Letters",
  ["wildl monogr"] = "Wildlife Monographs",
  ["wildl res"] = "Wildlife research (East Melbourne, Melbourne, Vic.)",
  ["wildl soc bull"] = "Wildlife Society bulletin",
  ["wildlife biol"] = "Wildlife biology",
  ["wilehm roux arch dev biol"] = "Wilhelm Roux's archives of developmental biology",
  ["wiley classics lib"] = "Wiley Classics Library",
  ["wiley finance ed"] = "Wiley Finance Editions",
  ["wiley finance ser"] = "Wiley Finance Series",
  ["wiley handb finance eng econom"] = "Wiley Handbook in Financial Engineering and Econometrics",
  ["wiley interdiscip rev clim change"] = "Wiley Interdisciplinary Reviews: Climate Change",
  ["wiley interdiscip rev cogn sci"] = "Wiley interdisciplinary reviews. Cognitive science",
  ["wiley interdiscip rev cognit sci"] = "Wiley Interdisciplinary Reviews: Cognitive Science",
  ["wiley interdiscip rev comput mol sci"] = "Wiley interdisciplinary reviews. Computational molecular science",
  ["wiley interdiscip rev comput stat"] = "Wiley Interdisciplinary Reviews: Computational Statistics",
  ["wiley interdiscip rev data min knowl discov"] = "Wiley interdisciplinary reviews. Data mining and knowledge discovery",
  ["wiley interdiscip rev dev biol"] = "Wiley interdisciplinary reviews. Developmental biology",
  ["wiley interdiscip rev energy environ"] = "Wiley interdisciplinary reviews. Energy and environment",
  ["wiley interdiscip rev membr transp signal"] = "Wiley interdisciplinary reviews. Membrane transport and signaling",
  ["wiley interdiscip rev nanomed nanobiotechnol"] = "Wiley interdisciplinary reviews. Nanomedicine and nanobiotechnology",
  ["wiley interdiscip rev rna"] = "Wiley interdisciplinary reviews. RNA",
  ["wiley interdiscip rev syst biol med"] = "Wiley interdisciplinary reviews. Systems biology and medicine",
  ["wiley interdiscip rev: comput mol sci"] = "Wiley Interdisciplinary Reviews: Computational Molecular Science",
  ["wiley interdiscip rev: data min knowl discovery"] = "Wiley Interdisciplinary Reviews: Data Mining and Knowledge Discovery",
  ["wiley interdiscip rev: dev biol"] = "Wiley Interdisciplinary Reviews: Developmental Biology",
  ["wiley interdiscip rev: energy environ"] = "Wiley Interdisciplinary Reviews: Energy and Environment",
  ["wiley interdiscip rev: forensic sci"] = "Wiley Interdisciplinary Reviews: Forensic Science",
  ["wiley interdiscip rev: membr transp signaling"] = "Wiley Interdisciplinary Reviews: Membrane Transport and Signaling",
  ["wiley interdiscip rev: rna"] = "Wiley Interdisciplinary Reviews: RNA",
  ["wiley interdiscip rev: water"] = "Wiley Interdisciplinary Reviews: Water",
  ["wiley press book"] = "A Wiley Press Book",
  ["wiley prof paperb ser"] = "Wiley Professional Paperback Series",
  ["wiley ser adapt cogn dyn syst"] = "Wiley Series on Adaptive and Cognitive Dynamic Systems",
  ["wiley ser beam phys accel tech"] = "Wiley Series in Beam Physics and Accelerator Technology",
  ["wiley ser bioinform comput tech eng"] = "Wiley Series on Bioinformatics: Computational Techniques and Engineering",
  ["wiley ser comput quant soc sci"] = "Wiley Series in Computational and Quantitative Social Science",
  ["wiley ser comput stat"] = "Wiley Series in Computational Statistics",
  ["wiley ser discrete math optim"] = "Wiley Series in Discrete Mathematics and Optimization",
  ["wiley ser math methods pract"] = "Wiley Series in Mathematical Methods in Practice",
  ["wiley ser methods appl statist"] = "Wiley series in Methods and Applications of Statistics",
  ["wiley ser model simul"] = "Wiley Series in Modeling and Simulation",
  ["wiley ser nonlinear sci"] = "Wiley Series in Nonlinear Science",
  ["wiley ser oper res manag sci"] = "Wiley Series in Operations Research and Management Science",
  ["wiley ser parallel distrib comput"] = "Wiley Series on Parallel and Distributed Computing",
  ["wiley ser probab math statist"] = "Wiley Series in Probability and Mathematical Statistics",
  ["wiley ser probab stat"] = "Wiley Series in Probability and Statistics",
  ["wiley ser probab stat texts ref sect"] = "Wiley Series in Probability and Statistics: Texts and References Section",
  ["wiley ser probab statist appl probab statist"] = "Wiley Series in Probability and Statistics: Applied Probability and Statistics",
  ["wiley ser probab statist probab statist"] = "Wiley Series in Probability and Statistics: Probability and Statistics",
  ["wiley ser surv methodol"] = "Wiley Series in Survey Methodology",
  ["wiley ser syst eng"] = "Wiley Series in Systems Engineering",
  ["wiley-asme press ser"] = "Wiley-ASME Press Series",
  ["wiley-intersci publ"] = "A Wiley-Interscience Publication",
  ["wiley-intersci ser discrete math optim"] = "Wiley-Interscience Series in Discrete Mathematics and Optimization",
  ["wiley-intersci ser systems optim"] = "Wiley-Interscience Series in Systems and Optimization",
  ["wiley-teubner comput"] = "Wiley-Teubner Computing",
  ["wiley-teubner ser adv numer math"] = "Wiley-Teubner Series Advances in Numerical Mathematics",
  ["wilhelm roux arch entwickl mech org"] = "Wilhelm Roux' Archiv für Entwicklungsmechanik der Organismen",
  ["wilhelm roux’ arch entwicklungsmech org"] = "Wilhelm Roux’ Archiv fuer Entwicklungsmechanik der Organismen",
  ["willamette law j"] = "Willamette law journal",
  ["willamette law rev"] = "Willamette law review",
  ["william mary bill rights j"] = "The William and Mary Bill of Rights journal : a student publication of the Marshall-Wythe School of Law",
  ["william mary j women law"] = "William & Mary journal of women and the law",
  ["william mary law rev"] = "William and Mary law review",
  ["william mary policy rev"] = "William & Mary policy review",
  ["william mary q"] = "The William and Mary quarterly",
  ["william mitchell law rev"] = "William Mitchell law review",
  ["wilson bull"] = "The Wilson bulletin",
  ["wilson j ornith"] = "Wilson Journal Of Ornithology",
  ["wilson j ornithol"] = "The Wilson journal of ornithology",
  ["wilson libr bull"] = "Wilson library bulletin",
  ["wilson q"] = "The Wilson quarterly",
  ["wilson wilsons compr anal chem"] = "Wilson and Wilson's comprehensive analytical chemistry",
  ["win news"] = "News - Women's International Network. Women's International Network",
  ["wind energy"] = "Wind Energy",
  ["wind energy (chichester)"] = "Wind energy (Chichester, England)",
  ["wind energy sci"] = "Wind Energy Science",
  ["wind eng"] = "Wind Engineering",
  ["wind struct"] = "Wind and Structures",
  ["windows time"] = "Windows in time",
  ["wine econ policy"] = "Wine Economics and Policy",
  ["winnipeg clin q"] = "Winnipeg Clinic quarterly",
  ["winter braids lect notes"] = "Winter Braids Lecture Notes",
  ["winterber eidgenöss inst schnee- lawinenforsch"] = "Winterberichte des Eidg. Institutes für Schnee- und Lawinenforschung",
  ["winterthur portf"] = "Winterthur portfolio",
  ["wirel commun mob comput"] = "Wireless communications & mobile computing",
  ["wirel netw"] = "Wireless Networks",
  ["wirel pers commun"] = "Wireless personal communications",
  ["wireless commun mobile comput"] = "Wireless Communications and Mobile Computing",
  ["wireless netw"] = "Wireless Networks",
  ["wireless pers commun"] = "Wireless Personal Communications",
  ["wireless power transfer"] = "Wireless Power Transfer",
  ["wires clim change"] = "WIREs Climate Change",
  ["wires cognit sci"] = "WIREs Cognitive Science",
  ["wires comput mol sci"] = "WIREs Computational Molecular Science",
  ["wires comput stat"] = "WIREs Computational Statistics",
  ["wires data min knowl discovery"] = "WIREs Data Mining and Knowledge Discovery",
  ["wires dev biol"] = "WIREs Developmental Biology",
  ["wires energy environ"] = "WIREs Energy and Environment",
  ["wires forensic sci"] = "WIREs Forensic Science",
  ["wires mech dis"] = "WIREs Mechanisms of Disease",
  ["wires membr transp signaling"] = "WIREs Membrane Transport and Signaling",
  ["wires membr transport signaling"] = "WIREs Membrane Transport and Signaling",
  ["wires nanomed nanobiotechnol"] = "WIREs Nanomedicine and Nanobiotechnology",
  ["wires rna"] = "WIREs RNA",
  ["wires syst biol med"] = "WIREs Systems Biology and Medicine",
  ["wires water"] = "WIREs Water",
  ["wiring install supplies"] = "Wiring Installations and Supplies",
  ["wirtsch recht"] = "Wirtschaft und Recht",
  ["wirtsch stat"] = "Wirtschaft und Statistik",
  ["wirtschaftspolitische blätter"] = "Wirtschaftspolitische Blätter",
  ["wirtschaftswissenschaftliches stud"] = "Wirtschaftswissenschaftliches Studium",
  ["wis acad rev"] = "Wisconsin academy review",
  ["wis bar bull"] = "The Wisconsin bar bulletin",
  ["wis dent assoc j"] = "Wisconsin Dental Association journal (Milwaukee, Wis. : 1992)",
  ["wis l rev"] = "Wisconsin law review",
  ["wis mag hist"] = "Wisconsin magazine of history",
  ["wis med j"] = "Wisconsin medical journal",
  ["wis publ hist sci med"] = "Wisconsin publications in the history of science and medicine",
  ["wis sess laws wis"] = "Wisconsin session laws. Wisconsin",
  ["wis welf"] = "Wisconsin welfare",
  ["wis womens law j"] = "Wisconsin women's law journal",
  ["wisc sociol"] = "The Wisconsin sociologist",
  ["wiss 20 jhd transdiszip reflex"] = "Wissenschaft im 20",
  ["wiss alpverheft"] = "Wissenschaftliche Alpenvereinshefte",
  ["wiss beitr gesch seelenheilkd"] = "Wissenschaftliche Beiträge zur Geschichte der Seelenheilkunde",
  ["wiss beitr martin luther univ halle wittenberg"] = "Wissenschaftliche Beiträge der Martin-Luther-Universität Halle-Wittenberg",
  ["wiss bucherei"] = "Wissenschaftliche Bucherei",
  ["wiss dienst sudosteur"] = "Wissenschaftlicher Dienst Südosteuropa",
  ["wiss glaube"] = "Wissenschaft und Glaube",
  ["wiss jahrb tirol landesmuseen"] = "Wissenschaftliches jahrbuch der Tiroler Landesmuseen",
  ["wiss mitt"] = "Wissenschaftliche Mitteilungen. Österreichischer Apothekerverein, Vienna. Pharmazeutisches Forschungsinstitut",
  ["wiss schriftenr math"] = "Wissenschaftliche Schriftenreihe Mathematik",
  ["wiss veroff dtsch ges ernahr"] = "Wissenschaftliche Veröffentlichungen. Deutsche Gesellschaft für Ernährung",
  ["wiss z ernst moritz arndt univ [ges sprachwiss]"] = "Wissenschaftliche Zeitschrift der Ernst-Moritz-Arndt-Universitat, Greifswald. Gesellschaftswissenschaftliche Reihe",
  ["wiss z ernst moritz arndt univ [math]"] = "Wissenschaftliche Zeitschrift. Mathematisch-naturwissenschaftliche Reihe. Ernst-Moritz-Arndt-Universitat Greifswald",
  ["wiss z ernst moritz arndt univ ges sprachwiss"] = "Wissenschaftliche Zeitschrift der Ernst-Moritz-Arndt-Universität, Greifswald. Gesellschaftswissenschaftliche Reihe",
  ["wiss z ernst moritz arndt univ math"] = "Wissenschaftliche Zeitschrift. Mathematisch-naturwissenschaftliche Reihe. Ernst-Moritz-Arndt-Universität Greifswald",
  ["wiss z friedrich schiller univ jena ges sprachwiss"] = "Wissenschaftliche Zeitschrift der Friedrich-Schiller-Universität Jena. Gesellschafts- und sprachwissenschaftliche Reihe",
  ["wiss z humboldt univ berl (ges sprachwiss)"] = "Wissenschaftliche Zeitschrift der Humboldt-Universität zu Berlin. Gesellschafts- und sprachwissenschaftliche Reihe",
  ["wiss z humboldt univ berl [math naturwiss]"] = "Wissenschaftliche Zeitschrift der Humboldt-Universitat zu Berlin. Mathematisch-Naturwissenschaftliche Reihe",
  ["wiss z humboldt univ berl ges sprachwiss reihe"] = "Wissenschaftliche Zeitschrift der Humboldt-Universität zu Berlin. Gesellschaftswissenschaftliche Reihe",
  ["wiss z humboldt univ berl math naturwiss"] = "Wissenschaftliche Zeitschrift der Humboldt-Universität zu Berlin. Mathematisch-Naturwissenschaftliche Reihe",
  ["wiss z karl marx univ leipzig [ges sprachwiss]"] = "Wissenschaftliche Zeitschrift. Gesellschafts- und sprachwissenschaftliche Reihe (Karl-Marx-Universitat Leipzig)",
  ["wiss z karl marx univ leipzig ges sprachwiss"] = "Wissenschaftliche Zeitschrift. Gesellschafts- und sprachwissenschaftliche Reihe (Karl-Marx-Universität Leipzig)",
  ["wiss z karl marx univ math naturwiss"] = "Wissenschaftliche Zeitschrift der Karl-Marx-Universität Leipzig. Mathematisch-naturwissenschaftliche Reihe",
  ["wiss z martin-luther univ halle-wittenberg"] = "Wissenshaftliche Zeitschrift der Martin Luther Universitat Halle-Wittenberg",
  ["wiss z tech univ dresd, sepr 5"] = "Wissenschaftliche Zeitschrift der Technischen Universität Dresden, Separatreihe 5",
  ["wiss z tech univ dresden"] = "Wissenschaftliche Zeitschrift der Technischen Universität Dresden",
  ["wiss z wilhelm pieck univ rostock [ges sprachwiss]"] = "Wissenschaftliche Zeitschrift der Wilhelm-Pieck-Universitat Rostock. Gesellschafts- und sprachwissenschaftliche Reihe",
  ["wiss z wilhelm pieck univ rostock ges sprachwiss"] = "Wissenschaftliche Zeitschrift der Wilhelm-Pieck-Universität Rostock. Gesellschafts- und sprachwissenschaftliche Reihe",
  ["wissen kompakt (heidelb)"] = "Wissen kompakt : Fortbildung für Zahnärzte",
  ["wissenschaftliche meeresuntersuchungen"] = "Wiss. Meeresunters.",
  ["wistar inst symp monogr"] = "The Wistar Institute symposium monograph",
  ["wit trans built environ"] = "WIT Transactions on the Built Environment",
  ["wit trans ecol environ"] = "WIT Transactions on Ecology and the Environment",
  ["wit trans eng sci"] = "WIT Transactions on Engineering Sciences",
  ["wit trans model simul"] = "WIT Transactions on Modelling and Simulation",
  ["wit trans modelling simul"] = "WIT transactions on modelling and simulation",
  ["wits j clin med"] = "Wits journal of clinical medicine",
  ["witthayasan kasetsat"] = "Witthayāsān Kasētsāt = The Kasetsart journal",
  ["witthayasat mo kho"] = "Witthayasat Mo Kho",
  ["wkly abstr sanit rep"] = "Weekly abstract of sanitary reports. United States. Marine Hospital Service",
  ["wkly bull epidemiol inf receiv"] = "Weekly bulletin; epidemiological information received. Interim Commission of the World Health Organization. Epidemiological Intelligence Station, Singapore",
  ["wkly epidemiol rec"] = "Weekly Epidemiological Record",
  ["wkly law rep"] = "The weekly law reports",
  ["wkly rec conv dis ports other localities home abroad"] = "Weekly record of convention diseases at ports and other localities at home and abroad. Great Britain. Ministry of Health",
  ["wkly rep ill"] = "Weekly report. Illinois. Division of Communicable Diseases",
  ["wmf press bull"] = "WMF press bulletins",
  ["wmo bull"] = "World Meteorological Organization Bulletin",
  ["wochenber dtsch inst wirtschaftsforsch"] = "Wochenbericht. Deutsches Institut für Wirtschaftsforschung",
  ["wochenbl papierfabr"] = "Wochenblatt fur Papierfabrikation",
  ["wochenschr brau"] = "Wochenschrift fur Brauerei",
  ["wolfenbutteler renaiss mitt"] = "Wolfenbütteler Renaissance Mitteilungen",
  ["woman physician"] = "The Woman physician",
  ["women 2000"] = "Women 2000",
  ["women action"] = "Women in action (Rome, Italy)",
  ["women aids"] = "Women And Aids",
  ["women alive"] = "Women alive (Los Angeles, Calif.)",
  ["women antiq"] = "Women in Antiquity",
  ["women birth"] = "Women and birth : journal of the Australian College of Midwives",
  ["women crim justice"] = "Women & criminal justice",
  ["women environ"] = "Women and environments",
  ["women envis"] = "Women envision",
  ["women health"] = "Women & health",
  ["women health care issues"] = "Women health care and issues",
  ["women hist philos sci"] = "Women in the History of Philosophy and Sciences",
  ["women lawyers j"] = "Women lawyers' journal",
  ["women med"] = "Women in medicine; official quarterly of American Medical Women's Association, Inc",
  ["women polit"] = "Women & politics",
  ["women sport phys act j"] = "Women in sport & physical activity journal",
  ["women ther"] = "Women & therapy",
  ["women vietnam"] = "Women of Vietnam",
  ["women's rights law report"] = "Women's rights law reporter",
  ["womens annu"] = "The Women's annual",
  ["womens health"] = "Women's health (Hillsdale, N.J.)",
  ["womens health (lond)"] = "Women's health (London, England)",
  ["womens health bull"] = "Women's health bulletin",
  ["womens health care"] = "Women's health care : a practical journal for nurse practitioners",
  ["womens health data book"] = "The women's health data book : a profile of women's health in the United States",
  ["womens health issues"] = "Women's health issues : official publication of the Jacobs Institute of Women's Health",
  ["womens health j"] = "Women's health journal",
  ["womens health newsl"] = "Womens Health Newsletter",
  ["womens health rep (new rochelle)"] = "Women's health reports (New Rochelle, N.Y.)",
  ["womens health update"] = "Women's health update",
  ["womens health urban life"] = "Women's health & urban life : an international and interdisciplinary journal",
  ["womens health wkly"] = "Women's health weekly",
  ["womens healthc (doylestown)"] = "Women's healthcare (Doylestown, Pa.)",
  ["womens hist"] = "Women's history",
  ["womens hist rev"] = "Women's history review",
  ["womens midlife health"] = "Women's midlife health",
  ["womens reprod health (phila)"] = "Women's reproductive health (Philadelphia, Pa.)",
  ["womens stud"] = "Women's studies",
  ["womens stud forum"] = "Women's studies forum",
  ["womens stud int forum"] = "Women's studies international forum",
  ["womens stud int q"] = "Women's studies international quarterly",
  ["womens watch"] = "The Women's watch",
  ["womens world"] = "Women's world (Kampala, Uganda)",
  ["womens writ"] = "Women's writing : the Elizabethan to Victorian period",
  ["women’s health"] = "Women’s Health",
  ["women’s rights law report"] = "Women’s rights law reporter",
  ["wood des focus"] = "Wood design focus",
  ["wood fiber sci"] = "Wood and Fiber Science",
  ["wood mater sci eng"] = "Wood Material Science & Engineering",
  ["wood res"] = "Wood Research",
  ["wood sci technol"] = "Wood Science and Technology",
  ["woodhead publ mech eng"] = "Woodhead Publishing in Mechanical Engineering",
  ["woodhead publ ser biomater"] = "Woodhead publishing series in biomaterials",
  ["woodhead publ ser biomed"] = "Woodhead Publishing Series in Biomedicine",
  ["woodhead publ ser compos sci eng"] = "Woodhead Publishing Series in Composites Science and Engineering",
  ["worcester med news"] = "Worcester medical news",
  ["word (n y : 1945)"] = "Word (New York, N.Y. : 1945)",
  ["word image (lond)"] = "Word & image",
  ["word world"] = "Word & world",
  ["work aging retire"] = "Work, aging and retirement",
  ["work employ soc"] = "Work, employment & society : a journal of the British Sociological Association",
  ["work environ health"] = "Work, environment, health",
  ["work health"] = "Work and Health",
  ["work life policies"] = "Work-life policies. National Symposium on Family Issues (15th : 2007 : Pennsylvania State University)",
  ["work occup"] = "Work and occupations",
  ["work older people"] = "Working with older people (Brighton, England)",
  ["work pap"] = "Working paper (Université catholique de Louvain (1970- ). Département de démographie)",
  ["work pap mag"] = "Working papers magazine",
  ["work pap new soc"] = "Working papers for a new society",
  ["work paper (cent community dev invest)"] = "Working paper (Center for Community Development Investments)",
  ["work stress"] = "Work and stress",
  ["work woman"] = "Working woman",
  ["workplace health saf"] = "Workplace health & safety",
  ["workshop counter space adapt exerc"] = "Workshop on Countering Space Adaptation with Exercise : Current Issues",
  ["workshop speech lang process assist technol"] = "Workshop on Speech and Language Processing for Assistive Technologies",
  ["workshops comput"] = "Workshops in Computing",
  ["workshops poster abstr"] = "Workshops & poster abstracts : [proceedings]. IEEE Computational Systems Bioinformatics Conference",
  ["world acad sci eng technol"] = "World academy of science, engineering and technology",
  ["world acad sci j"] = "World Academy of Sciences journal",
  ["world aids day features"] = "World AIDS day features",
  ["world aids day newsl"] = "World AIDS day newsletter",
  ["world allergy organ j"] = "The World Allergy Organization journal",
  ["world appl sci j"] = "World applied sciences journal",
  ["world archaeol"] = "World archaeology",
  ["world bank econ rev"] = "The World Bank economic review",
  ["world bank res news"] = "World Bank Research News",
  ["world bank res obs"] = "The World Bank research observer",
  ["world bank res observer"] = "World Bank Research Observer",
  ["world chinese journal of digestology|shi jie hua ren xiao hua za zhi"] = "Shi jie hua ren xiao hua za zhi",
  ["world citz news"] = "World citizen news : newsletter of the World Government of World Citizens",
  ["world conserv"] = "World conservation",
  ["world cult psychiatry res rev"] = "World cultural psychiatry research review : official journal of World Association of Cultural Psychiatry",
  ["world dev"] = "World development",
  ["world dev forum"] = "World development forum",
  ["world dev perspect"] = "World development perspectives",
  ["world devel"] = "World Development",
  ["world econ"] = "The World economy",
  ["world economy"] = "World Economy",
  ["world educ rep"] = "World education reports",
  ["world electr veh j"] = "World Electric Vehicle Journal",
  ["world environ rep"] = "World environment report",
  ["world futures"] = "World futures",
  ["world futures rev"] = "World futures review",
  ["world haptics 2007 (2007)"] = "World Haptics 2007 : Second Joint EuroHaptics Conference and Symposium on Haptic Interfaces for Virtual Environment and Teleoperator Systems : 22-24 March, 2007, Tsukuba, Japan. World Haptics Conference (2nd : 2007 : Tsukuba-shi, Japan)",
  ["world health"] = "World health",
  ["world health des"] = "World health design",
  ["world health forum"] = "World health forum",
  ["world health organ monogr ser"] = "World Health Organization Monograph Series",
  ["world health organ tech rep ser"] = "World Health Organization technical report series",
  ["world health popul"] = "World health & population",
  ["world health stat q"] = "World health statistics quarterly. Rapport trimestriel de statistiques sanitaires mondiales",
  ["world health stat rep"] = "World health statistics report. Rapport de statistiques sanitaires mondiales",
  ["world heart j"] = "World heart journal",
  ["world hosp"] = "World hospitals",
  ["world hosp health serv"] = "World hospitals and health services : the official journal of the International Hospital Federation",
  ["world ir nurs"] = "World of Irish nursing (Dublin, Ireland : 1995)",
  ["world j acupunct moxibustion"] = "World journal of acupuncture-moxibustion",
  ["world j agric res"] = "World journal of agricultural research",
  ["world j aids"] = "World journal of AIDS",
  ["world j biol chem"] = "World journal of biological chemistry",
  ["world j biol psychiatry"] = "The world journal of biological psychiatry : the official journal of the World Federation of Societies of Biological Psychiatry",
  ["world j cardiol"] = "World journal of cardiology",
  ["world j cardiovasc dis"] = "World Journal of Cardiovascular Diseases",
  ["world j cardiovasc surg"] = "World journal of cardiovascular surgery",
  ["world j clin cases"] = "World journal of clinical cases",
  ["world j clin infect dis"] = "World journal of clinical infectious diseases",
  ["world j clin oncol"] = "World journal of clinical oncology",
  ["world j clin pediatr"] = "World journal of clinical pediatrics",
  ["world j clin urol"] = "World journal of clinical urology",
  ["world j colorectal surg"] = "World journal of colorectal surgery",
  ["world j condens matter phys"] = "World Journal of Condensed Matter Physics",
  ["world j crit care med"] = "World journal of critical care medicine",
  ["world j diabetes"] = "World journal of diabetes",
  ["world j educ res"] = "World journal of educational research (Los Angeles, Calif.)",
  ["world j emerg med"] = "World journal of emergency medicine",
  ["world j emerg surg"] = "World journal of emergency surgery : WJES",
  ["world j eng"] = "World Journal of Engineering",
  ["world j exp med"] = "World journal of experimental medicine",
  ["world j gastroenterol"] = "World journal of gastroenterology",
  ["world j gastrointest endosc"] = "World journal of gastrointestinal endoscopy",
  ["world j gastrointest oncol"] = "World journal of gastrointestinal oncology",
  ["world j gastrointest pathophysiol"] = "World journal of gastrointestinal pathophysiology",
  ["world j gastrointest pharmacol ther"] = "World journal of gastrointestinal pharmacology and therapeutics",
  ["world j gastrointest surg"] = "World journal of gastrointestinal surgery",
  ["world j gynecol womens health"] = "World journal of gynecology & womens health",
  ["world j hepatol"] = "World journal of hepatology",
  ["world j hypertens"] = "World journal of hypertension",
  ["world j immunol"] = "World journal of immunology",
  ["world j med genet"] = "World journal of medical genetics",
  ["world j mens health"] = "The world journal of men's health",
  ["world j metaanal"] = "World journal of meta-analysis",
  ["world j methodol"] = "World journal of methodology",
  ["world j microbiol biotechnol"] = "World journal of microbiology & biotechnology",
  ["world j nephrol"] = "World journal of nephrology",
  ["world j nephrol urol"] = "World journal of nephrology and urology",
  ["world j neurol"] = "World journal of neurology",
  ["world j neurosci"] = "World journal of neuroscience",
  ["world j nucl med"] = "World journal of nuclear medicine",
  ["world j obstet gynecol"] = "World journal of obstetrics and gynecology",
  ["world j oncol"] = "World journal of oncology",
  ["world j ophthalmol"] = "World journal of ophthalmology",
  ["world j orthod"] = "World journal of orthodontics",
  ["world j orthop"] = "World journal of orthopedics",
  ["world j otorhinolaryngol"] = "World journal of otorhinolaryngology",
  ["world j otorhinolaryngol head neck surg"] = "World journal of otorhinolaryngology - head and neck surgery",
  ["world j pediatr"] = "World journal of pediatrics : WJP",
  ["world j pediatr congenit heart surg"] = "World journal for pediatric & congenital heart surgery",
  ["world j pharm pharm sci"] = "World journal of pharmacy and pharmaceutical sciences",
  ["world j pharm res"] = "World journal of pharmaceutical research",
  ["world j pharmacol"] = "World journal of pharmacology",
  ["world j plast surg"] = "World journal of plastic surgery",
  ["world j psychiatry"] = "World journal of psychiatry",
  ["world j psychiatry ment health res"] = "World journal of psychiatry and mental health research",
  ["world j radiol"] = "World journal of radiology",
  ["world j res rev"] = "World journal of research and review",
  ["world j respirol"] = "World journal of respirology",
  ["world j rheumatol"] = "World journal of rheumatology",
  ["world j sci technol sustainable dev"] = "World Journal of Science, Technology and Sustainable Development",
  ["world j stem cells"] = "World journal of stem cells",
  ["world j surg"] = "World journal of surgery",
  ["world j surg oncol"] = "World journal of surgical oncology",
  ["world j surg proced"] = "World journal of surgical procedures",
  ["world j tradit chin med"] = "World journal of traditional Chinese medicine",
  ["world j transl med"] = "World journal of translational medicine",
  ["world j transplant"] = "World journal of transplantation",
  ["world j urol"] = "World journal of urology",
  ["world j vaccines"] = "World journal of vaccines",
  ["world j virol"] = "World journal of virology",
  ["world learn"] = "The World of Learning",
  ["world learn (lond)"] = "The world of learning",
  ["world leis j"] = "World leisure journal",
  ["world med"] = "World medicine",
  ["world med health policy"] = "World medical & health policy",
  ["world med j"] = "World medical journal",
  ["world min surf underground"] = "World of Mining: Surface and Underground",
  ["world mycotoxin j"] = "World mycotoxin journal",
  ["world neighb action"] = "World Neighbors in action",
  ["world neurol"] = "World neurology",
  ["world neurosurg"] = "World neurosurgery",
  ["world neurosurg x"] = "World neurosurgery: X",
  ["world news maxillofac radiol"] = "World news on maxillofacial radiology",
  ["world policy j"] = "World policy journal",
  ["world polit"] = "World politics",
  ["world psychiatry"] = "World psychiatry : official journal of the World Psychiatric Association (WPA)",
  ["world rabbit sci"] = "World Rabbit Science",
  ["world res j agric biotechnol"] = "World research journal of agricultural biotechnology",
  ["world resour rev"] = "World Resource Review",
  ["world rev nutr diet"] = "World review of nutrition and dietetics",
  ["world rev sci technol sustainable dev"] = "World Review of Science Technology and Sustainable Development",
  ["world review"] = "World review (Brisbane, Qld.)",
  ["world sci annu rev artif intell"] = "World Scientific Annual Review of Artificial Intelligence",
  ["world sci annu rev biomech"] = "World Scientific Annual Review of Biomechanics",
  ["world sci annu rev cancer immunol"] = "World Scientific Annual Review of Cancer Immunology",
  ["world sci annu rev data sci"] = "World Scientific Annual Review of Data Science",
  ["world sci annu rev fintech"] = "World Scientific Annual Review of FinTech",
  ["world sci annu rev funct mate"] = "World Scientific Annual Review of Functional Materials",
  ["world sci annu rev part phys"] = "World Scientific Annual Review of Particle Physics",
  ["world sci annu rev vaccine des"] = "World Scientific Annual Review of Vaccine Design",
  ["world sci handb financ econ ser"] = "World Scientific Handbook in Financial Economics Series",
  ["world sci lect notes econ policy"] = "World Scientific Lecture Notes in Economics and Policy",
  ["world sci lect notes finance"] = "World Scientific Lecture Notes in Finance",
  ["world sci lecture notes phys"] = "World Scientific Lecture Notes in Physics",
  ["world sci news"] = "World scientific news",
  ["world sci now publ ser bus"] = "World Scientific—Now Publishers Series in Business",
  ["world sci ser 20th century math"] = "World Scientific Series in 20th Century Mathematics",
  ["world sci ser 20th century phys"] = "World Scientific Series in 20th Century Physics",
  ["world sci ser 21st century math"] = "World Scientific Series in 21st Century Mathematics",
  ["world sci ser appl anal"] = "World Scientific Series in Applicable Analysis",
  ["world sci ser econ theory"] = "World Scientific Series in Economic Theory",
  ["world sci ser finance"] = "World Scientific Series in Finance",
  ["world sci ser inf stud"] = "World Scientific Series in Information Studies",
  ["world sci ser nonlinear sci ser a monogr treatises"] = "World Scientific Series on Nonlinear Science. Series A: Monographs and Treatises",
  ["world sci ser nonlinear sci ser b spec theme issues proc"] = "World Scientific Series on Nonlinear Science. Series B: Special Theme Issues and Proceedings",
  ["world sci ser probab theory appl"] = "World Scientific Series on Probability Theory and Its Applications",
  ["world sci ser quantum algorithms inf learn"] = "World Scientific Series on Quantum Algorithms, Information, and Learning",
  ["world sci ser robot intell systems"] = "World Scientific Series in Robotics and Intelligent Systems",
  ["world smoking health"] = "World smoking & health",
  ["world tob"] = "World tobacco",
  ["world today"] = "The World today",
  ["world watch"] = "World watch",
  ["world water policy"] = "World Water Policy",
  ["world wide abstr gen med"] = "World-wide abstracts of general medicine",
  ["world wide web"] = "World wide web",
  ["world wide web j biol"] = "The World Wide Web journal of biology",
  ["world wirel mob multimed netw"] = "World of wireless mobile and multimedia networks. IEEE International Symposium on a World of Wireless Mobile and Multimedia Networks",
  ["world wood"] = "World Wood",
  ["world work"] = "World of work : the magazine of the ILO",
  ["worlds poult sci j"] = "World's poultry science journal",
  ["worldviews evid based nurs"] = "Worldviews on evidence-based nursing",
  ["worldwide waste: j interdiscip stud"] = "Worldwide Waste: Journal of Interdisciplinary Studies",
  ["world’s poult sci j"] = "World’s Poultry Science Journal",
  ["worm breed gaz"] = "The worm breeder's gazette",
  ["wosp 2017 (2017)"] = "WOSP 2017 : proceedings of the 6th International Workshop on Mining Scientific Publications. International Workshop on Mining Scientific Publications (6th : 2017 : Toronto, Ont.)",
  ["wound manag prev"] = "Wound management & prevention",
  ["wound repair regen"] = "Wound repair and regeneration : official publication of the Wound Healing Society [and] the European Tissue Repair Society",
  ["wound repair regener"] = "Wound Repair and Regeneration",
  ["wri issues ideas"] = "WRI issues and ideas",
  ["writ am hist"] = "Writings on American history: a subject bibliography of articles",
  ["writ commun"] = "Written communication",
  ["writ lang lit"] = "Written language and literacy",
  ["writ syst res"] = "Writing systems research",
  ["wseas trans math"] = "WSEAS transactions on mathematics",
  ["wu li hua xue xue bao"] = "Wu li hua xue xue bao = Acta physico-chimica sinica",
  ["wuhan botanical research|wuhan zhi wu xue yan jiu"] = "Wuhan zhi wu xue yan jiu",
  ["wuhan da xue xue bao li xue ban"] = "Wuhan da xue xue bao. Li xue ban = Journal of Wuhan University. Natural science edition",
  ["wuhan da xue xue bao zi ran ke xue ban"] = "Wuhan da xue xue bao. Zi ran ke xue ban = Journal of Wuhan University. Natural sciences edition",
  ["wuhan univ j nat sci"] = "Wuhan University Journal of Natural Sciences",
  ["wuhan zhi wu xue yan jiu"] = "Wuhan zhi wu xue yan jiu = Wuhan botanical research",
  ["wuji huaxue xuebao"] = "Wuji Huaxue Xuebao",
  ["wuli huaxue xuebao"] = "Wuli Huaxue Xuebao",
  ["wuli xuebao"] = "Wuli Xuebao",
  ["wuli xuebao (acta phys sin) [chin j phys]"] = "Wuli Xuebao (Acta Physica Sinica) [Chinese Journal of Physics]",
  ["wurzbg medizinhist forsch"] = "Würzburger medizinhistorische Forschungen",
  ["wurzbg medizinhist mitt"] = "Würzburger medizinhistorische Mitteilungen",
  ["wyo nurses newsl"] = "Wyoming nurses' newsletter",
  ["wyo statut annot wyo"] = "Wyoming statutes, annotated. Wyoming",
  ["wythe cty hist rev"] = "Wythe County historical review",
  ["x-ray spectrom"] = "X-Ray Spectrometry",
  ["x-ray struct anal online"] = "X-Ray Structure Analysis Online",
  ["xi bao sheng wu xue za zhi"] = "Xi bao sheng wu xue za zhi",
  ["xi bao yu fen zi mian yi xue za zhi"] = "Xi bao yu fen zi mian yi xue za zhi = Chinese journal of cellular and molecular immunology",
  ["xi bei nong lin ke ji da xue xue bao zi ran ke xue ban"] = "Xi bei nong lin ke ji da xue xue bao. Zi ran ke xue ban = Journal of Northwest Sci-Tech University of Agriculture and Forestry. Natural science edition",
  ["xi bei nong ye xue bao"] = "Xi bei nong ye xue bao = Acta agriculturae Boreali-Occidentalis Sinica",
  ["xi bei ren kou"] = "Xi bei ren kou = Xibeirenkou",
  ["xi nan nong ye da xue xue bao"] = "Xi nan nong ye da xue xue bao = Journal of Southwest Agricultural University",
  ["xi nan nong ye xue bao"] = "Xi nan nong ye xue bao = Southwest China journal of agricultural sciences",
  ["xi psi phi q"] = "The Xi Psi Phi quarterly",
  ["xiamen da xue xue bao zi ran ke xue ban"] = "Xiamen da xue xue bao. Zi ran ke xue ban = Journal of Xiamen University. Natural science",
  ["xiamen daxue xuebao ziran kexue ban"] = "Xiamen Daxue Xuebao. Ziran Kexue Ban. Journal of Xiamen University. Natural Science",
  ["xian dai hua gong"] = "Xian dai hua gong = Modern chemical industry",
  ["xian dai sheng wu yi xue jin zhan"] = "Xian dai sheng wu yi xue jin zhan",
  ["xian dai shi yong yi xue"] = "Xian dai shi yong yi xue",
  ["xian dai yu fang yi xue"] = "Xian dai yu fang yi xue = Modern preventive medicine",
  ["xiandai shipin keji"] = "Xiandai shipin keji = Modern food science and technology",
  ["xianggang hu li za zhi"] = "Xianggang hu li za zhi. The Hong Kong nursing journal",
  ["xianggang she hui ke xue xue bao"] = "Xianggang she hui ke xue xue bao = Hong Kong journal of social sciences",
  ["xibei zhiwu xuebao"] = "Xi bei zhi wu xue bao = Acta botanica Boreali-Occidentalia Sinica",
  ["xibu linye kexue"] = "Xibu Linye Kexue = Journal of West China Forestry Science",
  ["xin li ke xue jin zhan"] = "Xin li ke xue jin zhan",
  ["xin li xue bao"] = "Xin li xue bao. Acta psychologica Sinica",
  ["xin zhong hua yi yao yue kan"] = "[Xin Zhong hua yi yao yue kan] [New China medical monthly]",
  ["xinan minzu daxue xuebao ziran kexue ban"] = "Xi'nan Minzu Daxue xuebao. Ziran kexue ban",
  ["xinan shifan daxue xuebao"] = "Xinan Shifan Daxue Xuebao",
  ["xinan shifan daxue xuebao ziran kexue ban"] = "Xinan Shifan Daxue Xuebao",
  ["xinjiang nong ye da xue xue bao"] = "Xinjiang nong ye da xue xue bao = Journal of Xinjiang Agricultural University",
  ["xinjiang nong ye ke xue"] = "Xinjiang nong ye ke xue",
  ["xi’an jiaotong daxue xuebao"] = "Xi’an Jiaotong Daxue Xuebao. Journal of Xi’an Jiaotong University",
  ["xray spectrom"] = "X-ray spectrometry : XRS",
  ["xray tech"] = "The X-ray technician",
  ["xrds: crossroads, acm mag stud"] = "XRDS: Crossroads, The ACM Magazine for Students",
  ["xu mu shou yi xue bao"] = "Xu mu shou yi xue bao = Acta veterinaria et zootechnica Sinica",
  ["xumu yu shouyi"] = "Xumu yu shouyi",
  ["xviie siecle"] = "XVIIe siècle",
  ["y hoc thanh pho ho chi minh"] = "Y học Thành phó̂ Hò̂ Chí Minh",
  ["y hoc viet nam"] = "Y hoc Viêt Nam",
  ["y rep"] = "Y [report]. U.S. Atomic Energy Commission",
  ["yad fiz"] = "Yadernaya Fizika",
  ["yad fiz [sov j nucl phys]"] = "Yadernaya Fizika [Soviet Journal of Nuclear Physics]",
  ["yad vashem stud eur jew catastrophe resist"] = "Yad Vashem studies",
  ["yakhak hoechi"] = "Yakhak Hoechi",
  ["yakhakhoe chi"] = "Yakhakhoe chi. Journal of the Pharmaceutical Society of Korea",
  ["yakubutsu seishin kodo"] = "Yakubutsu, seishin, kōdō = Japanese journal of psychopharmacology",
  ["yakugaku kenkyu"] = "[Yakugaku kenkyū] Japanese journal of pharmacy and chemistry",
  ["yakugaku zasshi"] = "Yakugaku zasshi : Journal of the Pharmaceutical Society of Japan",
  ["yakuri to chiryo"] = "Yakuri to Chiryo",
  ["yakushigaku zasshi"] = "Yakushigaku zasshi",
  ["yale econ essays"] = "Yale Economic Essays",
  ["yale j biol med"] = "The Yale journal of biology and medicine",
  ["yale j health policy law ethics"] = "Yale journal of health policy, law, and ethics",
  ["yale j law fem"] = "Yale journal of law and feminism",
  ["yale j law humanit"] = "Yale journal of law & the humanities",
  ["yale j regul"] = "Yale journal on regulation",
  ["yale law j"] = "The Yale law journal",
  ["yale law policy rev"] = "Yale law & policy review",
  ["yale leibniz"] = "The Yale Leibniz",
  ["yale math monogr"] = "Yale Mathematical Monographs",
  ["yale sci mag"] = "Yale scientific magazine",
  ["yale univ libr gaz"] = "The Yale University library gazette. Yale University. Library",
  ["yalkut moreshet"] = "Yalḳuṭ moreshet",
  ["yamashina chorui kenkyujo kenkyu hokoku"] = "Yamashina Chōrui Kenkyūjo kenkyū hōkōku",
  ["yamashina choruigaku zasshi"] = "Yamashina chōruigaku zasshi",
  ["yan jiu bao gao"] = "[Yan jiu bao gao] Science reports of the National Taiwan University. Guo li Taiwan da xue. Xin li xue xi",
  ["yan jiu bao gao (taiwan)"] = "Yan jiu bao gao (Taiwan)",
  ["yan ke xue bao"] = "Yan ke xue bao = Eye science",
  ["yan ke xue bao (hong kong)"] = "Yan ke xue bao (2016)",
  ["yan ke xue shu hui kan"] = "[Yan ke xue shu hui kan] = Ophthalmological series",
  ["yancao keji"] = "Yancao keji",
  ["yang zhou da xue xue bao nong ye yu sheng ming ke xue ban"] = "Yang zhou da xue xue bao. Nong ye yu sheng ming ke xue ban",
  ["yao hsueh hsueh pao"] = "Yao Hsueh Hsueh Pao [Acta Pharmaceutica Sinica]",
  ["yao wu fen xi za zhi"] = "Yao wu fen xi za zhi = Yaowu fenxi zazhi",
  ["yao wu sheng wu ji shu"] = "Yao wu sheng wu ji shu",
  ["yao wu shi pin fen xi"] = "Yao wu shi pin fen xi = Journal of food and drug analysis",
  ["yao xue xue bao"] = "Yao xue xue bao = Acta pharmaceutica Sinica",
  ["yaowu fenxi zazhi"] = "Yaowu Fenxi Zazhi",
  ["yaoxue xuebao"] = "Yaoxue Xuebao",
  ["yapi kredi econ rev"] = "Yapi Kredi Economic Review",
  ["year b carnegie inst wash"] = "Year book - Carnegie Institution of Washington",
  ["year b leo baeck inst"] = "Year book",
  ["year b r soc edinb"] = "Year book - Royal Society of Edinburgh",
  ["year immunol"] = "The Year in immunology",
  ["yearb acad appl osteopathy"] = "Year book. Academy of Applied Osteopathy",
  ["yearb am philos soc"] = "Year book - American Philosophical Society",
  ["yearb conf lat am geogr"] = "Yearbook (Conference of Latin Americanist Geographers)",
  ["yearb eston nat soc"] = "Yearbook of the Estonian Naturalist Society",
  ["yearb ger am stud"] = "Yearbook of German-American studies",
  ["yearb med inform"] = "Yearbook of medical informatics",
  ["yearb phys anthropol"] = "Yearbook of physical anthropology",
  ["yearb popul res finl"] = "Yearbook of population research in Finland = Väestöntutkimuksen vuosikirja",
  ["yearb res cent ger austrian exile stud"] = "The yearbook of the Research Centre for German and Austrian Exile Studies",
  ["yejin fenxi"] = "Yejin Fenxi",
  ["yen ko hsueh pao"] = "Yen Ko Hsueh Pao [Eye Science] (Chung-Kuo Kuang-Chou",
  ["yeni tip tarihi arastirmalari"] = "Yeni tıp tarihi araştırmaları = The new history of medicine studies",
  ["yenisei ser pure appl math"] = "Yenisei Series in Pure and Applied Mathematics",
  ["yesca (maliano muriedas)"] = "Yesca",
  ["yeungnam univ j med"] = "Yeungnam University journal of medicine",
  ["yi chuan"] = "Yi chuan = Hereditas",
  ["yi chuan xue bao"] = "Yi chuan xue bao = Acta genetica Sinica",
  ["yi yong sheng wu li xue"] = "Yi yong sheng wu li xue",
  ["ying yang xue bao"] = "[Ying yang xue bao] Acta nutrimenta Sinica",
  ["ying yong kun chong xue bao"] = "Ying yong kun chong xue bao = Chinese journal of applied entomology",
  ["ying yong sheng tai xue bao"] = "Ying yong sheng tai xue bao = The journal of applied ecology",
  ["ying yong yu huan jing sheng wu xue bao"] = "Ying yong yu huan jing sheng wu xue bao = Chinese journal of applied and environmental biology",
  ["yingxiang kexue yu guang huaxue"] = "Yingxiang Kexue Yu Guang Huaxue",
  ["yingyong huaxue"] = "Yingyong Huaxue",
  ["yngre laeger"] = "Yngre laeger",
  ["yokohama math j"] = "Yokohama Mathematical Journal",
  ["yokohama med bull"] = "Yokohama medical bulletin",
  ["yokufuen chosa kenkyu kiyo"] = "Yokufūen chōsa kenkyū kiyō. Acta gerontologica Japonica",
  ["yonago acta med"] = "Yonago acta medica",
  ["yonago igaku zasshi"] = "Yonago igaku zasshi. The Journal of the Yonago Medical Association",
  ["yonsei med j"] = "Yonsei medical journal",
  ["york state trad"] = "York state tradition",
  ["yorkshire archaeol j"] = "Yorkshire Archaeological Journal",
  ["yorkshire bull econ soc res"] = "Yorkshire Bull. Econ. Soc. Res.",
  ["yosetsu gakkai ronbunshu"] = "Yosetsu Gakkai Ronbunshu",
  ["youji huaxue"] = "Youji Huaxue",
  ["young child"] = "Young children",
  ["young minds"] = "Young Minds",
  ["your child patient"] = "Your Child Patient",
  ["your health (salt lake city)"] = "Your health",
  ["your okla dent assoc j"] = "Your Oklahoma Dental Association journal",
  ["youth soc"] = "Youth & society",
  ["youth violence juv justice"] = "Youth violence and juvenile justice",
  ["yu mi ke xue"] = "Yu mi ke xue",
  ["yuan yi xue bao"] = "Yuan yi xue bao",
  ["yugosl j oper res"] = "Yugoslav Journal of Operations Research. An International Journal dealing with Theoretical and Computational Aspects of Operations Research, Systems Science, and Management Science",
  ["yugosl surv"] = "Yugoslav survey",
  ["yuki gosei kagaku kyokaishi"] = "Yuki Gosei Kagaku Kyokaishi",
  ["yunnan nong ye da xue xue bao"] = "Yunnan nong ye da xue xue bao = Journal of Yunnan Agricultural University",
  ["yunnan zhi wu yan jiu"] = "Yunnan zhi wu yan jiu = Yun nan zhi wu yan jiu",
  ["z aachener geschichtsver"] = "Zeitschrift des Aachener Geschichtsvereins. Aachener Geschichtsverein",
  ["z aegypt sprach altertumskd"] = "Zeitschrift für ägyptische sprache und altertumskunde",
  ["z agrarges agrarsoziol"] = "Zeitschrift für Agrargeschichte und Agrarsoziologie",
  ["z agrargesch agrarsoziol"] = "Zeitschrift für Agrargeschichte und Agrarsoziologie",
  ["z allg mikrobiol"] = "Zeitschrift fur allgemeine Mikrobiologie",
  ["z allg textile marktwirtsch"] = "Zeitschrift für Allgemaine und Textile Marktwirts schaft",
  ["z allg wissenschaftstheor"] = "Zeitschrift für allgemeine Wissenschaftstheorie. Journal for general philosophy of science",
  ["z allgemeinmed"] = "Zeitschrift fur Allgemeinmedizin",
  ["z alternsforsch"] = "Zeitschrift fur Alternsforschung",
  ["z anal anwend"] = "Zeitschrift fur Analysis und ihre Anwendungen",
  ["z anal anwendungen"] = "Zeitschrift für Analysis und ihre Anwendungen",
  ["z anal chem"] = "Zeitschrift fur Analytische Chemie",
  ["z anat entwicklungsgesch"] = "Zeitschrift fur Anatomie und Entwicklungsgeschichte",
  ["z angew bader klimaheilkd"] = "Zeitschrift für angewandte Bäder- und Klimaheilkunde",
  ["z angew entomol"] = "Zeitschrift fur Angewandte Entomologie",
  ["z angew geol"] = "Zeitschrift fur Angewandte Geologie",
  ["z angew math mech"] = "Zeitschrift für angewandte Mathematik und Mechanik",
  ["z angew math phys"] = "Zeitschrift fuer angewandte mathematik und physik",
  ["z angew photogr wiss tech"] = "Zeitschrift fur Angewandte Photographie in Wissenschaft und Technik",
  ["z angew phys"] = "Zeitschrift fur Physik",
  ["z anorg allg chem"] = "Zeitschrift für anorganische und allgemeine Chemie",
  ["z anorg chem"] = "Zeitschrift fur Anorganische Chemie",
  ["z arbeitswiss"] = "Zeitschrift für Arbeitswissenschaft",
  ["z arbgem österr entomol"] = "Zeitschrift der Arbeitsgemeinschaft österreichischer Entomologen",
  ["z arznei gewurzpflanzen"] = "Zeitschrift fur Arznei und Gewurzpflanzen",
  ["z arztl fortbild"] = "Zeitschrift fur Arztliche Fortbildung",
  ["z arztl fortbild (berl)"] = "Zeitschrift für ärztliche Fortbildung (Berlin)",
  ["z arztl fortbild (jena)"] = "Zeitschrift für ärztliche Fortbildung",
  ["z arztl fortbild beih (jena)"] = "Zeitschrift für ärztliche Fortbildung. Beiheft",
  ["z arztl fortbild qualitatssich"] = "Zeitschrift für ärztliche Fortbildung und Qualitätssicherung",
  ["z astrophys"] = "Zeitschrift für Astrophysik",
  ["z bayer landesgesch"] = "Zeitschrift für bayerische Landesgeschichte",
  ["z betriebswirtschaft"] = "Zeitschrift für Betriebswirtschaft",
  ["z bevolkerungswiss"] = "Zeitschrift für Bevölkerungswissenschaft",
  ["z bibliothekswes bibliogr"] = "Zeitschrift für Bibliothekswesen und Bibliographie",
  ["z bienenforsch"] = "Zeitschrift fur Bienenforschung",
  ["z biol"] = "Zeitschrift fur Biologie",
  ["z chem"] = "Zeitschrift fur Chemie",
  ["z didakt naturwiss"] = "Zeitschrift fur Didaktik der Naturwissenschaften",
  ["z dtsch altert dtsch lit"] = "Zeitschrift für deutsches Alterthum und deutsche Litteratur",
  ["z dtsch geol ges"] = "Zeitschrift der Deutschen Geologischen Gesellschaft",
  ["z dtsch ges geowiss"] = "Zeitschrift der Deutschen Gesellschaft fur Geowissenschaften",
  ["z dtsch morgenl ges"] = "Zeitschrift der Deutschen Morgenländischen Gesellschaft",
  ["z elektrochem"] = "Zeitschrift für Elektrochemie",
  ["z elektrochem angew phys chem"] = "Zeitschrift fur Angewandte Physik und Chemie",
  ["z elektrotech elektrochem"] = "Zeitschrift fur Elektrotechnik und Elektrochemie",
  ["z entwicklungspsychol padagog psychol"] = "Zeitschrift für Entwicklungspsychologie und Pädagogische Psychologie",
  ["z erdkundeunterr"] = "Zeitschrift für den Erdkundeunterricht",
  ["z erkr atmungsorgane"] = "Zeitschrift fur Erkrankungen der Atmungsorgane",
  ["z erkr atmungsorgane folia bronchol"] = "Zeitschrift fũr Erkrankungen der Atmungsorgane mit Folia bronchologica",
  ["z ernahrungswiss"] = "Zeitschrift fur Ernahrungswissenschaft",
  ["z ernahrungswiss suppl"] = "Zeitschrift für Ernährungswissenschaft. Journal of nutritional sciences. Supplementa",
  ["z erzbergbau metallhuttenwes"] = "Zeitschrift fur Erzbergbau und Metallhuttenwesen",
  ["z erziehwiss"] = "Zeitschrift für Erziehungswissenschaft : ZfE",
  ["z ethnol"] = "Zeitschrift für Ethnologie",
  ["z evid fortbild qual gesundhwes"] = "Zeitschrift für Evidenz, Fortbildung und Qualität im Gesundheitswesen",
  ["z exp angew psychol"] = "Zeitschrift für experimentelle und angewandte Psychologie",
  ["z exp chir"] = "Zeitschrift für experimentelle Chirurgie",
  ["z exp chir transplant kunstliche organe"] = "Zeitschrift für experimentelle Chirurgie, Transplantation, und künstliche Organe : Organ der Sektion Experimentelle Chirurgie der Gesellschaft für Chirurgie der DDR",
  ["z exp pathol ther"] = "Zeitschrift fur Experimentelle Pathologie und Therapie",
  ["z exp psychol"] = "Zeitschrift für experimentelle Psychologie : Organ der Deutschen Gesellschaft für Psychologie",
  ["z fam forsch"] = "Zeitschrift fur Familienforschung",
  ["z feldherpetol"] = "Zeitschrift fur Feldherpetologie",
  ["z flugtech motorluftschiffahrt"] = "Zeitschrift fur Flugtechnik und Motorluftschiffahrt",
  ["z flugwiss weltraumforsch"] = "Zeitschrift fur Flugwissenschaften und Weltraumforschung",
  ["z forst- jagdwes"] = "Zeitschrift für das Forst- und Jagdwesen",
  ["z forstgenet forstpflanzenzücht"] = "Zeitschrift für Forstgenetik und Forstpflanzenzüchtung",
  ["z fr spr lit"] = "ZFSL, Zeitschrift für französische Sprache und Literatur",
  ["z für krist"] = "Zeitschrift für Kristallographie",
  ["z für phys"] = "Zeitschrift für Physik",
  ["z gastroenterol"] = "Zeitschrift für Gastroenterologie",
  ["z gastroenterol verh"] = "Zeitschrift fur Gastroenterologie. Verhandlungsband",
  ["z geburtshilfe gynakol"] = "Zeitschrift für Geburtshilfe und Gynäkologie",
  ["z geburtshilfe neonatol"] = "Zeitschrift für Geburtshilfe und Neonatologie",
  ["z geburtshilfe perinatol"] = "Zeitschrift fur Geburtshilfe und Perinatologie",
  ["z geol wiss"] = "Zeitschrift fur Geologische Wissenschaften",
  ["z geomorphol"] = "Zeitschrift fur Geomorphologie",
  ["z geomorphol suppl"] = "Zeitschrift fur Geomorphologie Supplementband",
  ["z geophys"] = "Zeitschrift fur Geophysik",
  ["z gerontol"] = "Zeitschrift fur Gerontologie",
  ["z gerontol geriatr"] = "Zeitschrift für Gerontologie und Geriatrie",
  ["z ges schleswig holstein gesch"] = "Zeitschrift der Gesellschaft für Schleswig-Holsteinische Geschichte. Gesellschaft für Schleswig-Holsteinische Geschichte",
  ["z ges staatswissen (jite)"] = "Zeitschrift für die gesamte Staatswissenschaft (JITE)",
  ["z gesamte exp med"] = "Zeitschrift für die gesamte experimentelle Medizin",
  ["z gesamte hyg"] = "Zeitschrift für die gesamte Hygiene und ihre Grenzgebiete",
  ["z gesamte inn med"] = "Zeitschrift für die gesamte innere Medizin und ihre Grenzgebiete",
  ["z gesamte schiess sprengstoffwes"] = "Zeitschrift fur das Gesamte Schiess und Sprengstoffwesen",
  ["z gesamte staatswiss"] = "Zeitschrift für die gesamte Staatswissenschaft = Journal of institutional and theoretical economics",
  ["z gesch arab islam wiss"] = "Zeitschrift für Geschichte der Arabisch-Islamischen Wissenschaften",
  ["z gesch arab-islam wiss"] = "Zeitschrift fur Geschichte der Arabisch-Islamischen Wissenschaften",
  ["z geschichtswiss"] = "Zeitschrift für Geschichtswissenschaft",
  ["z geschiebeforsch"] = "Zeitschrift fur Geschiebeforschung",
  ["z gesundh wiss"] = "Zeitschrift für Gesundheitswissenschaften = Journal of public health",
  ["z gletscherkd glazialgeol"] = "Zeitschrift fur Gletscherkunde und Glazialgeologie",
  ["z gletschkd glazialgeol"] = "Zeitschrift für Gletscherkunde und Glazialgeologie",
  ["z haut geschlechtskr"] = "Zeitschrift fur Haut- und Geschlechtskrankheiten",
  ["z hautkr"] = "Zeitschrift fur Hautkrankheiten",
  ["z heereskd"] = "Zeitschrift für Heereskunde",
  ["z herz thorax gefasschir"] = "Zeitschrift für Herz-, Thorax- und Gefässchirurgie",
  ["z hist forsch"] = "Zeitschrift für historische Forschung",
  ["z hist ver schwaben neuburg"] = "Zeitschrift des Historischen Vereins für Schwaben. Historischer Verein für Schwaben",
  ["z hist ver steiermark"] = "Zeitschrift des Historischen Vereines für Steiermark",
  ["z hochfrequenztech"] = "Zeitschrift fur Hochfrequenztechnik",
  ["z hyg"] = "Zeitschrift fur Hygiene",
  ["z hyg infektionskr"] = "Zeitschrift für Hygiene und Infektionskrankheiten; medizinische Mikrobiologie, Immunologie und Virologie",
  ["z hyg infektionskrankh"] = "Zeitschrift fur Hygiene und Infektionskrankheiten",
  ["z immun exp ther"] = "Zeitschrift für Immunitätsforschung und experimentelle Therapie",
  ["z immunitats allergieforsch"] = "Zeitschrift für Immunitats- und Allergieforschung",
  ["z immunitatsforsch allerg klin immunol"] = "Zeitschrift für Immunitätsforschung, Allergie und klinische Immunologie",
  ["z immunitatsforsch exp klin immunol"] = "Zeitschrift für Immunitätsforschung, experimentelle und klinische Immunologie",
  ["z immunitatsforsch exp ther"] = "Zeitschrift fur Immunitatsforschung und Experimentalle Therapie",
  ["z immunitatsforsch immunobiol"] = "Zeitschrift für Immunitätsforschung. Immunobiology",
  ["z immunitatsforsch immunobiol suppl"] = "Zeitschrift für Immunitätsforschung. Immunobiology. Supplemente",
  ["z indukt abstamm vererbungsl"] = "Zeitschrift für induktive Abstammungs- und Vererbungslehre",
  ["z infektol"] = "Zhurnal infektologii",
  ["z instrumentenkd"] = "Zeitschrift fur Instrumentenkunde",
  ["z jagdwiss"] = "Zeitschrift fur Jagdwissenschaft",
  ["z k preuss stat bur"] = "Zeitschrift des KoÌˆniglichen Preussischen Statistischen Bureaus",
  ["z kardiol"] = "Zeitschrift fur Kardiologie",
  ["z kardiol suppl"] = "Zeitschrift fur Kardiologie. Supplementum",
  ["z kinder jugendpsychiatr"] = "Zeitschrift für Kinder- und Jugendpsychiatrie",
  ["z kinder jugendpsychiatr psychother"] = "Zeitschrift für Kinder- und Jugendpsychiatrie und Psychotherapie",
  ["z kinder jugenpsychiatr"] = "Zeitschrift fur Kinder- und Jugendpsychiatrie",
  ["z kinderchir"] = "Zeitschrift für Kinderchirurgie : organ der Deutschen, der Schweizerischen und der Osterreichischen Gesellschaft für Kinderchirurgie = Surgery in infancy and childhood",
  ["z kinderchir grenzgeb"] = "Zeitschrift fur Kinderchirurgie und Grenzgebiete",
  ["z kinderheilkd"] = "Zeitschrift fur Kinderheilkunde",
  ["z kinderpsychiatr"] = "Zeitschrift für Kinderpsychiatrie. Revue de psychiatrie infantile",
  ["z kirchengesch"] = "Zeitschrift für Kirchengeschichte",
  ["z klin chem klin biochem"] = "Zeitschrift fur Klinische Chemie und Klinische Biochemie",
  ["z klin med"] = "Zeitschrift fur Klinische Medicin",
  ["z klin psychol psychiatr psychother"] = "Zeitschrift für klinische Psychologie, Psychiatrie und Psychotherapie",
  ["z klin psychol psychopathol psychother"] = "Zeitschrift für klinische Psychologie, Psychopathologie und Psychotherapie",
  ["z klin psychol psychother"] = "Zeitschrift für klinische Psychologie und Psychotherapie",
  ["z klin psychol psychother (gott)"] = "Zeitschrift für klinische Psychologie und Psychotherapie (Göttingen, Germany)",
  ["z krankenpfl"] = "Zeitschrift für Krankenpflege. Revue suisse des infirmières",
  ["z krebsforsch"] = "Zeitschrift fur Krebsforschung",
  ["z krebsforsch klin onkol cancer res clin oncol"] = "Zeitschrift für Krebsforschung und klinische Onkologie. Cancer research and clinical oncology",
  ["z kreislaufforsch"] = "Zeitschrift fur Kreislaufforschung",
  ["z krist"] = "Zeitschrift für Kristallographie",
  ["z kristallogr"] = "Zeitschrift fur Kristallograhie",
  ["z kristallogr - cryst mater"] = "Zeitschrift fur Kristallographie -  Crystalline Materials",
  ["z kristallogr - new cryst struct"] = "Zeitschrift fur Kristallographie - New Crystral Stuctures",
  ["z kristallogr cryst mater"] = "Zeitschrift für Kristallographie. Crystalline materials",
  ["z kristallogr kristallgeom kristallphys kristallchem"] = "Zeitschrift fur Kristallographie, Kristallgeometrie, Kristallphysik, Kristallchemie",
  ["z kristallogr mineral petrogr"] = "Zeitschrift fur Kristallographie, Mineralogie und Petrographie",
  ["z kulttech flurbereinig"] = "Zeitschrift für Kulturtechnik und Flurbereinigung",
  ["z kunstgesch"] = "Zeitschrift für Kunstgeschichte",
  ["z laryngol rhinol otol"] = "Zeitschrift für Laryngologie, Rhinologie, Otologie und ihre Grenzgebiete",
  ["z lebensm unters forsch"] = "Zeitschrift fur Lebensmittel-Untersuchung und -Forschung",
  ["z lebensm unters forsch a food res technol"] = "Zeitschrift für Lebensmittel-Untersuchung und -Forschung. A, Food research and technology",
  ["z lebensm-unters forsch"] = "Zeitschrift für Lebensmittel-Untersuchung und -Forschung",
  ["z lebensm-unters forsch a"] = "Zeitschrift fuer Lebensmittel-Untersuchung und -Forschung A: European Food Research and Technology",
  ["z lymphol"] = "Zeitschrift für Lymphologie. Journal of lymphology",
  ["z mag"] = "Z magazine (Boston, Mass.)",
  ["z med ethik"] = "Zeitschrift für medizinische Ethik : Wissenschaft, Kultur, Religion",
  ["z med isotopenforsch deren grenzgeb"] = "Zeitschrift fur Medizinische Isotopenforschung und Deren Grenzgebiete",
  ["z med lab diagn"] = "Zeitschrift für medizinische Laboratoriumsdiagnostik",
  ["z med laboratoriumsdiagn"] = "Zeitschrift fur Medizinische Laboratoriumsdiagnostik",
  ["z med labortech"] = "Zeitschrift fur Medizinische Labortechnik",
  ["z med mikrobiol immunol"] = "Zeitschrift für medizinische Mikrobiologie und Immunologie",
  ["z med phys"] = "Zeitschrift für medizinische Physik",
  ["z med psychol"] = "Zeitschrift für medizinische Psychologie : ZMP",
  ["z mensch vererb konstitutionsl"] = "Zeitschrift für menschliche Vererbungs- und Konstitutionslehre",
  ["z metailkd"] = "Zeitschrift fur MetaIlkunde",
  ["z metallkd"] = "Zeitschrift für Metallkunde",
  ["z meteor"] = "Zeitschrift für Meteorologie",
  ["z mikrosk anat forsch"] = "Zeitschrift für mikroskopisch-anatomische Forschung",
  ["z morphol anthropol"] = "Zeitschrift fur Morphologie und Anthropologie",
  ["z morphol okol tiere"] = "Zeitschrift für Morphologie und Oekologie der Tiere",
  ["z mykol"] = "Zeitschrift für mykologie",
  ["z nat forsch a j phys sci"] = "Zeitschrift für Naturforschung. A, Journal of physical sciences",
  ["z nationalökon"] = "Zeitschrift für Nationalökonomie",
  ["z naturforsch"] = "Zeitschrift für Naturforschung",
  ["z naturforsch [b]"] = "Zeitschrift fur Naturforschung. Teil B, Anorganische Chemie, Organische Chemie, Biochemie, Biophysik, Biologie",
  ["z naturforsch [c]"] = "Zeitschrift fur Naturforschung. Section C, Biosciences",
  ["z naturforsch b"] = "Zeitschrift für Naturforschung. Teil B. Anorganische Chemie, organische Chemie, Biochemie, Biophysik, Biologie",
  ["z naturforsch b j chem sci"] = "Zeitschrift für Naturforschung. B, Journal of chemical sciences",
  ["z naturforsch c"] = "Zeitschrift für Naturforschung. Teil C: Biochemie, Biophysik, Biologie, Virologie",
  ["z naturforsch c biosci"] = "Zeitschrift für Naturforschung. Section C, Biosciences",
  ["z naturforsch c j biosci"] = "Zeitschrift für Naturforschung. C, Journal of biosciences",
  ["z naturforsch teil a"] = "Zeitschrift für Naturforschung, Teil A: Physik, Physikalische Chemie, Kosmophysik",
  ["z naturforsch, a: phys sci"] = "Zeitschrift fur Naturforschung A: Journal of Physical Sciences",
  ["z naturforsch, b: chem sci"] = "Zeitschrift fur Naturforschung B: Journal of Chemical Sciences",
  ["z naturforsch, b: j chem sci"] = "Zeitschrift fuer Naturforschung, B: A Journal of Chemical Sciences",
  ["z naturforsch, c: biosci"] = "Zeitschrift fur Naturforschung C: Journal of Biosciences",
  ["z naturforsch, c: j biosci"] = "Zeitschrift für Naturforschung, C: Journal of Biosciences",
  ["z naturwiss med grundlagenforsch"] = "Zeitschrift für naturwissenschaftlichmedizinische Grundlagenforschung",
  ["z neurol"] = "Zeitschrift fur Neurologie",
  ["z orthop ihre grenzgeb"] = "Zeitschrift für Orthopädie und ihre Grenzgebiete",
  ["z orthop unfall"] = "Zeitschrift für Orthopädie und Unfallchirurgie",
  ["z ostforsch"] = "Zeitschrift für Ostforschung",
  ["z ostpreuss meschkinnes forsch"] = "Zeitschrift fur Ostpreussische Meschkinnes Forschung",
  ["z papyrologie epigraphik"] = "Zeitschrift für Papyrologie und Epigraphik",
  ["z parasitenkd"] = "Zeitschrift fur Parasitenkunde",
  ["z pflanzenernahr bodenkd"] = "Zeitschrift fur Pflanzenernahrung und Bodenkunde",
  ["z pflanzenernahr dung bodenkd"] = "Zeitschrift fur Pflanzenernahrung Dungung Bodenkunde",
  ["z pflanzenernähr bodenkd"] = "Zeitschrift für Pflanzenernährung und Bodenkunde",
  ["z pflanzenkr pflanzenschutz"] = "Zeitschrift für Pflanzenkrankheiten und Pflanzenschutz",
  ["z pflanzenkrankh pflanzenschutz"] = "Zeitschrift fur Pflanzenkrankheiten und Pflanzenschutz",
  ["z pflanzenphysiol"] = "Zeitschrift fur Pflanzenphysiologie",
  ["z philos forsch"] = "Zeitschrift für philosophische Forschung",
  ["z photogramm fernerkund"] = "Zeitschrift für Photogrammetrie und Fernerkundung",
  ["z phys"] = "Zeitschrift für Physik",
  ["z phys a"] = "Zeitschrift für Physik A",
  ["z phys a at nucl"] = "Zeitschrift für Physik. A, Atomic nuclei",
  ["z phys a hadrons nucl"] = "Zeitschrift für Physik A Hadrons and Nuclei",
  ["z phys a: at nucl"] = "Zeitschrift fur Physik A: Atomes and Nuclei",
  ["z phys a: hadrons nucl"] = "Zeitschrift fur Physik A: Hadrons and Nuclei",
  ["z phys b"] = "Zeitschrift für Physik B",
  ["z phys b condens matter"] = "Zeitschrift für Physik B Condensed Matter",
  ["z phys b condens matter quanta"] = "Zeitschrift für Physik B Condensed Matter and Quanta",
  ["z phys b: condens matter"] = "Zeitschrift fur Physik B: Condensed Matter",
  ["z phys b: condens matter quanta"] = "Zeitschrift fur Physik B: Condensed Matter and Quanta",
  ["z phys c"] = "Zeitschrift für Physik C",
  ["z phys c: part fields"] = "Zeitschrift fur Physik C: Particle Fields",
  ["z phys chem"] = "Zeitschrift fur Physikalische Chemie",
  ["z phys chem (frankurt am main)"] = "Zeitschrift für Physikalische Chemie (Frankfurt am Main)",
  ["z phys chem (leipzig)"] = "Zeitschrift für Physikalische Chemie (Leipzig)",
  ["z phys chem (muenchen, ger)"] = "Zeitschrift für Physikalische Chemie (Muenchen, Germany)",
  ["z phys chem (n f)"] = "Zeitschrift für physikalische Chemie (Frankfurt am Main, Germany)",
  ["z phys chem abt a"] = "Zeitschrift für Physikalische Chemie, Abteilung A: Chemische Thermodynamik, Kinetik, Elektrochemie, Eigenschaftslehre",
  ["z phys chem abt a:"] = "Zeitschrift fur Physikalische Chemie, Abteilung A",
  ["z phys chem abt b"] = "Zeitschrift für Physikalische Chemie, Abteilung B: Chemie der Elementarprozesse, Aufbau der Materie",
  ["z phys chem abt b:"] = "Zeitschrift fur Physikalische Chemie, Abteilung B",
  ["z phys chem materialforsch"] = "Zeitschrift für Physikalisch–Chemische Materialforschung",
  ["z phys chem stoechiom verwandtschafts"] = "Zeitschrift fur Physikalische Chemie, Stoeechiometrie und Verwandschaftslehre",
  ["z phys d"] = "Zeitschrift für Physik D",
  ["z phys d: at mol clust"] = "Zeitschrift für Physik D: Atoms, Molecules and Clusters",
  ["z phys d: at mol clusters"] = "Zeitschrift fur Physik D: Atoms, Molecules and Clusters",
  ["z phys ther bader klimanheikd"] = "Zeitschrift für physikalische Therapie, Bäder- und Klimaheilkunde",
  ["z physiol chem"] = "Zeitschrift fur Physiologische Chemie",
  ["z physiother"] = "Zeitschrift fur Physiotherapie",
  ["z phytother"] = "Zeitschrift für Phytotherapie : offizielles Organ der Ges. f. Phytotherapie e.V",
  ["z pilzkd"] = "Zeitschrift für Pilzkunde",
  ["z plast chir"] = "Zeitschrift für plastische Chirurgie",
  ["z pola walki"] = "Z pola walki",
  ["z polit"] = "Zeitschrift für Politik",
  ["z prakt anasth"] = "Zeitschrift für praktische Anästhesie und Wiederbelebung",
  ["z prakt anasth wiederbeleb intensivther"] = "Zeitschrift für praktische Anästhesie, Wiederbelebung und Intensivtherapie",
  ["z praventivmed"] = "Zeitschrift für Präventivmedizin. Revue de médecine préventive",
  ["z preuss gesch landeskd"] = "Zeitschrift fur preussische Geschichte und Landeskunde",
  ["z psychol"] = "Zeitschrift für Psychologie",
  ["z psychol z angew psychol"] = "Zeitschrift für Psychologie mit Zeitschrift für angewandte Psychologie",
  ["z psychosom med"] = "Zeitschrift für Psycho-somatische Medizin",
  ["z psychosom med psychoanal"] = "Zeitschrift fur Psychosomatische Medizin und Psychoanalyse",
  ["z psychosom med psychother"] = "Zeitschrift für Psychosomatische Medizin und Psychotherapie",
  ["z psychother med psychol"] = "Zeitschrift für Psychotherapie und medizinische Psychologie",
  ["z rechtsmed"] = "Zeitschrift fur Rechtsmedizin",
  ["z relig geistesgesch"] = "Zeitschrift für Religions- und Geistesgeschichte",
  ["z rheumaforsch"] = "Zeitschrift fur Rheumaforschung",
  ["z rheumatol"] = "Zeitschrift für Rheumatologie",
  ["z rom philol"] = "Zeitschrift für romanische Philologie",
  ["z saugetierkd"] = "Zeitschrift für Säugetierkunde",
  ["z savigny stift rechtsgesch ger abt"] = "Zeitschrift der Savigny-Stiftung für Rechtsgeschichte. Germanistische Abteilung",
  ["z savigny stift rechtsgesch kanon abt"] = "Zeitschrift der Savigny-Stiftung für Rechtsgeschichte. Kanonistische Abteilung",
  ["z sex forsch"] = "Zeitschrift für Sexualforschung (Stuttgart, Germany)",
  ["z sozialreform"] = "Zeitschrift für Sozialreform",
  ["z soziol"] = "Zeitschrift für Soziologie",
  ["z sportpsychol"] = "Zeitschrift für sportpsychologie",
  ["z stomatol"] = "Zeitschrift für Stomatologie (1984)",
  ["z stomatol (1921)"] = "Zeitschrift für Stomatologie",
  ["z säugetierkd"] = "Zeitschrift für Säugetierkunde",
  ["z tech phys"] = "Zeitschrift fur Technische Physik",
  ["z theor angew genet"] = "Zeitschrift fur Theoretische und Angewandte Genetik",
  ["z tierernahr futtermittelkd"] = "Zeitschrift fur Tierernahrung und Futtermittelkunde",
  ["z tierphysiol tierernahr futtermittelkd"] = "Zeitschrift fur Tierphysiologie, Tierernahrung und Futtermittelkunde",
  ["z tierpsychol"] = "Zeitschrift fur Tierpsychologie",
  ["z tierz zuchtungsbiol"] = "Zeitschrift fur Tierzuchtung und Zuchtungsbiologie",
  ["z tierzuecht zuechtungsbiol"] = "Zeitschrift für Tierzüchtung und Züchtungsbiologie : Organ der Reichsarbeitsgemeinschaft Tierzucht im Forschungsdienst",
  ["z tropenmed parasitol"] = "Zeitschrift fur Tropenmedizin und Parasitologie",
  ["z tuberk"] = "Zeitschrift für Tuberkulose",
  ["z tuberk erkr thoraxorg"] = "Zeitschrift fur Tuberkulose und Erkrankungen der Thoraxorgane",
  ["z unfallchir versicherungsmed"] = "Zeitschrift für Unfallchirurgie und Versicherungsmedizin : offizielles Organ der Schweizerischen Gesellschaft für Unfallmedizin und Berufskrankheiten = Revue de traumatologie et d'assicurologie : organe officiel de la Société suisse de médecine des accidents et des maladies professionnelles",
  ["z unfallchir versicherungsmed berufskr"] = "Zeitschrift für Unfallchirurgie, Versicherungsmedizin und Berufskrankheiten : offizielles Organ der Schweizerischen Gesellschaft für Unfallmedizin und Berufskrankheiten = Revue de traumatologie, d'assicurologie et des maladies professionnelles : organe officiel de la Société suisse de médecine des accidents et des maladies professionnelles",
  ["z unfallmed berufskr"] = "Zeitschrift für Unfallmedizin und Berufskrankheiten. Revue de médecine des accidents et des maladies professionelles",
  ["z unfug ges"] = "Zeitschrift der Unfug Gesellschaft",
  ["z urol"] = "Zeitschrift für Urologie",
  ["z urol nephrol"] = "Zeitschrift fur Urologie und Nephrologie",
  ["z ver dtsch ing"] = "Zeitschrift des Vereins Deutscher Ingenieure",
  ["z ver dtsch zucker ind"] = "Zeitschrift des Vereines der Deutschen Zucker Industrie",
  ["z ver lubeck gesch altertumskd"] = "Zeitschrift des Vereins für Lübeckische Geschichte und Altertumskunde. Verein für Lübeckische Geschichte und Altertumskunde",
  ["z vererbungsl"] = "Zeitschrift fur Vererbungslehre",
  ["z vermessungswes"] = "Zeitschrift fur Vermessungswesen",
  ["z versuchstierkd"] = "Zeitschrift fur Versuchstierkunde",
  ["z vgl physiol"] = "Zeitschrift für vergleichende Physiologie",
  ["z vitam forsch"] = "Zeitschrift für Vitaminforschung",
  ["z vitam horm fermentforsch"] = "Zeitschrift für Vitamin-, Hormon- und Fermentforschung",
  ["z wasser abwasser-forsch"] = "Zeitschrift fur Wasser- und Abwasserforschung",
  ["z wasser wasser-forsch"] = "Zeitschrift für Wasser- und Abwasserforschung",
  ["z weltforstwirtsch"] = "Zeitschrift für Weltforstwirtschaft",
  ["z wirtsch sozialwiss"] = "Zeitschrift für Wirtschafts- und Sozialwissenschaften",
  ["z wirtsch sozialwissen"] = "Zeitschrift für Wirtschafts- und Sozialwissenschaften",
  ["z wirtschaftspolitik"] = "Zeitschrift für Wirtschaftspolitik",
  ["z wirtschgeogr"] = "Zeitschrift für Wirtschaftsgeographie",
  ["z wiss mikrosk"] = "Zeitschrift für wissenschaftliche Mikroskopie und mikroskopische Technik",
  ["z wiss mikrosk mikrosk tech"] = "Zeitschrift fur Wissenschaftliche Mikroskopie und fur Mikroskopische Technik",
  ["z wiss photogr photophys photochem"] = "Zeitschrift fur Wissenschaftliche Photographie Photophysik und Photochemie",
  ["z wiss zool"] = "Zeitschrift für wissenschartliche Zoologie",
  ["z wurttemb landesgesch"] = "Zeitschrift für Württembergische Landesgeschichte",
  ["z zellforch microsk anat histochem"] = "Zeitschrift für Zellforschung und Mikroskopische Anatomie. Abteilung Histochemie",
  ["z zellforsch"] = "Zeitschrift für Zellforschung",
  ["z zellforsch mikrosk anat"] = "Zeitschrift für Zellforschung und mikroskopische Anatomie (Vienna, Austria : 1948)",
  ["z zucht reihe b"] = "Zeitschrift fur Zuchtung Reihe B",
  ["z ökol natschutz"] = "Zeitschrift für Ökologie und Naturschutz",
  ["zagazig vet j"] = "Zagazig veterinary journal",
  ["zagreb int rev econ bus"] = "Zagreb International Review of Economics and Business",
  ["zahn mund kieferheilkd zentralbl"] = "Zahn-, Mund-, und Kieferheilkunde mit Zentralblatt",
  ["zahnarztebl baden wurttemb"] = "Zahnarzteblatt Baden-Wurttemberg",
  ["zahnarztl gesundheitsdienst"] = "Zahnärztlicher Gesundheitsdienst : offizielles Organ des Bundesverbandes der Zahnärzte des Öffentlichen Gesundheitsdienstes E.V",
  ["zahnarztl mitt"] = "Zahnarztliche Mitteilungen",
  ["zahnarztl nachr sudwurttemb"] = "Zahnarztliche Nachrichten Sudwurttemburg-Hohenzollern",
  ["zahnarztl prax"] = "Zahnarztliche Praxis",
  ["zahnarztl praxisfuhr"] = "Zahnarztliche Praxisfuhrung",
  ["zahnarztl rundsch"] = "Zahnarztliche Rundschau",
  ["zahnarztl welt"] = "Zahnärztliche Welt",
  ["zahnarztl welt zahnarztl reform zwr"] = "Zahnärztliche Welt, zahnärztliche Reform, ZWR",
  ["zahnarztl welt zahnarztl rundsch zwr zahnarztl reform"] = "Zahnãrztliche Welt, zahnãrztliche Rundschau, ZWR, vereinigt mit Zahnãrztliche Reform",
  ["zahnheilkd mundheilkd kieferheilkd vortr"] = "Zahn-, Mund- und Kieferheilkunde in Vorträgen",
  ["zahntech (basel)"] = "Der Zahntechniker. Le Mécanicien-dentiste",
  ["zahntechnik (berl)"] = "Zahntechnik; Zeitschrift für Theorie und Praxis der wissenschaftlichen Zahntechnik",
  ["zahntechnik (zur)"] = "Zahntechnik (Zurich, Switzerland)",
  ["zahntechniker (basel)"] = "Der Zahntechniker. Le Mecanicien-dentiste",
  ["zahradnictvi (praha)"] = "Zahradnictví (Prague, Czech Republic : 1992)",
  ["zaire afr"] = "Zaïre-Afrique",
  ["zambia nurse"] = "Zambia nurse (Kitwe, Zambia : 1978)",
  ["zambia nurse j"] = "The Zambia nurse journal",
  ["zamm z angew math mech"] = "ZAMM. Zeitschrift fur Angewandte Mathematik und Mechanik. Journal of Applied Mathematics and Mechanics",
  ["zanco j med sci"] = "Zanco journal of medical sciences",
  ["zanco j pure appl sci"] = "Zanco journal of pure and applied sciences",
  ["zap nauchn sem s-peterburg otdel mat inst steklov (pomi)"] = "Rossiiskaya Akademiya Nauk. Sankt-Peterburgskoe Otdelenie. Matematicheskii Institut im. V. A. Steklova. Zapiski Nauchnykh Seminarov (POMI)",
  ["zap vses mineral o-va"] = "Zapiski Vsesoyuznogo Mineralogicheskogo Obshchestva",
  ["zas karantin rast"] = "Zashchita i karantin rasteniĭ",
  ["zashch met"] = "Zashchita Metallov",
  ["zasshi fukuoka ika daigaku"] = "[Zasshi] [Journal]. Fukuoka Ika Daigaku",
  ["zasshi kanazawa daigaku igakubu juzen igakkai"] = "[Zasshi] Journal. Kanazawa Daigaku. Igakubu. Jūzen Igakkai",
  ["zasshi tokyo ika daigaku"] = "Zasshi. Tokyo Ika Daigaku",
  ["zast bilja"] = "Zaštita bilja",
  ["zb bioteh fak univ ljubl kmet supl"] = "Zbornik Biotehniške fakultete Univerze v Ljubljani. Kmetijstvo (1990)",
  ["zb istor"] = "Zbornik za istoriju",
  ["zb istor bosne herceg"] = "Zbornik za istoriju Bosne i Hercegovine = Recueil de l'histoire de Bosnie et Herzegovine",
  ["zb lek fak kosice"] = "Zborník prác. Univerzita Pavla Jozefa Safárika v Kosiciach. Lekárska fakulta",
  ["zb matice srp istor"] = "Zbornik Matice srpske za istoriju",
  ["zb matice srp prir nauk"] = "Zbornik Matice srpske za prirodne nauke",
  ["zb odsjeka povij znan zovoda povij drus znan hrvat akad znan umjet"] = "Zbornik Odsjeka za povijesne znanosti Zavoda za povijesne i društvene znanosti Hrvatske akademije znanosti i umjetnosti",
  ["zb pr inst mat nan ukr"] = "Zbirnik Prats Institutu Matematiki NAN Ukraini",
  ["zb rad (beogr)"] = "Zbornik Radova. (Beograd)",
  ["zb rad (kragujevac)"] = "Univerzitet u Kragujevcu",
  ["zb rad (varaždin)"] = "Zbornik Radova",
  ["zb rad naunog sastanka"] = "Zbornik radova nauc̆nog sastanka",
  ["zb vojnomed akad"] = "Zbornik. Vojnomedicinska akademija (Yugoslavia)",
  ["zb zgodovino naravoslovja teh"] = "Zbornik za zgodovino naravoslovja in tehnike = History reviews, science and technology",
  ["zbirka izbran poglav fiz"] = "Zbirka Izbranih Poglavij iz Fizike",
  ["zbirka izbran poglav mat"] = "Zbirka Izbranih Poglavij iz Matematike",
  ["zbornik radova ekon fakul rijeka"] = "Zbornik Radova Ekonomskog Fakulteta Rijeka",
  ["zbornik radova ekon fakul u rijeci"] = "Zbornik Radova Ekonomskog Fakulteta U. Rijeci",
  ["zdm math educ"] = "ZDM Mathematics Education",
  ["zdr delo"] = "Zdravno delo",
  ["zdr varst"] = "Zdravstveno varstvo",
  ["zdrav aktual"] = "Zdravotnické aktuality",
  ["zdrav prac"] = "Zdravotnická pracovnice",
  ["zdrav radn"] = "Zdravstveni radnik",
  ["zdrav rev vestn minist zdrav"] = "Zdravotnická revue; vĕstník Ministerstva Zdravotnictví",
  ["zdrav vestn"] = "Zdravstveni vestnik",
  ["zdravookhr beloruss"] = "Zdravookhranenie Belorussii",
  ["zdravookhr kirg"] = "Zdravookhranenie Kirgizii",
  ["zdravookhr ross fed"] = "Zdravookhranenie Rossiĭskoĭ Federatsii",
  ["zdravookhr turkmenistana"] = "Zdravookhranenie Turkmenistana",
  ["zdravookhranenie (minsk)"] = "Zdravookhranenie (Minsk, Belarus)",
  ["zdravookhranenie belarusi"] = "Zdravookhranenie Belarusi (Belorusskiĭ tsentr nauchnoĭ meditsinskoĭ informatsii)",
  ["zdravookhranenie kazakhstana"] = "Zdravookhranenie Kazakhstana",
  ["zdrow publiczne"] = "Zdrowie publiczne",
  ["zeiss – inf"] = "Zeiss – Informationen",
  ["zeitschrift fur angewandte entomologie|j appl entomol"] = "Journal of applied entomology",
  ["zeitschrift fur lebensmittel-untersuchung und -forschung a|eur food res technol"] = "European food research and technology",
  ["zeitschrift fur pflanzenzuchtung|plant breed"] = "Plant breeding",
  ["zeitschrift fur saugetierkunde|mamm biol"] = "Mammalian biology",
  ["zeitschrift fur zoologische systematik und evolutionsforschung|j zoolog syst evol res"] = "Journal of zoological systematics and evolutionary research",
  ["zellst papier"] = "Zellstoff und Papier",
  ["zem kalk gips"] = "Zement Kalk Gips",
  ["zent arb arb beih"] = "Zentralblatt für Arbeitsmedizin und Arbeitsschutz. Beihefte",
  ["zent arbeitswissenschaft"] = "Zentralblatt für Arbeitswissenschaft",
  ["zent bibliothekswesen"] = "Zentralblatt für Bibliothekswesen",
  ["zent-ztg opt mech"] = "Zentral-Zeitung fur Optik und Mechanik",
  ["zentralasiat stud"] = "Zentralasiatische Studien",
  ["zentralbl allg pathol"] = "Zentralblatt für allgemeine Pathologie u. pathologische Anatomie",
  ["zentralbl arbeitsmed"] = "Zentralblatt fur Arbeitsmedizin und Arbeitsschutz",
  ["zentralbl arbeitsmed arbeitsschutz ergon"] = "Zentralblatt fur Arbeitsmedizin, Arbeitsschutz und Ergonomie",
  ["zentralbl arbeitsmed arbeitsschutz prophyl"] = "Zentralblatt fur Arbeitsmedizin, Arbeitsschutz und Prophylaxe",
  ["zentralbl arbeitsmed arbeitsschutz prophyl ergonomie"] = "Zentralblatt fur Arbeitsmedizin, Arbeitsschutz, Prophylaxe und Ergonomie",
  ["zentralbl bakteriol"] = "Zentralblatt für Bakteriologie : international journal of medical microbiology",
  ["zentralbl bakteriol [b]"] = "Zentralblatt fur Bakteriologie, Parasitenkunde, Infektionskrankheiten und Hygiene. Erste Abteilung, Originale. Reihe B, Hygiene, Betriebshygiene, Praventive Medizin",
  ["zentralbl bakteriol [orig a]"] = "Zentralblatt fur Bakteriologie, Parasitenkunde, Infektionskrankheiten und Hygiene. Erste Abteilung Originale. Reihe A: Medizinische Mikrobiologie und Parasitologie",
  ["zentralbl bakteriol [orig b]"] = "Zentralblatt fur Bakteriologie, Parasitenkunde, Infektionskrankheiten und Hygiene. Erste",
  ["zentralbl bakteriol [orig]"] = "Zentralblatt fur Bakteriologie, Parasitenkunde, Infektionskrankheiten und Hygiene. 1. Abt.,",
  ["zentralbl bakteriol a"] = "Zentralblatt für Bakteriologie. 1. Abt. Originale. A: Medizinische Mikrobiologie, Infektionskrankheiten und Parasitologie",
  ["zentralbl bakteriol b"] = "Zentralblatt für Bakteriologie. 1. Abt. Originale B, Hygiene, Krankenhaushygiene, Betriebshygiene, präventive Medizin",
  ["zentralbl bakteriol mikrobiol hyg [a]"] = "Zentralblatt fur Bakteriologie, Mikrobiologie, und Hygiene. Series A, Medical Microbiology, Infectious Diseases, Virology, Parasitology",
  ["zentralbl bakteriol mikrobiol hyg [b]"] = "Zentralblatt fur Bakteriologie, Mikrobiologie und Hygiene. 1. Abt. Originale B, Hygiene",
  ["zentralbl bakteriol mikrobiol hyg a"] = "Zentralblatt für Bakteriologie, Mikrobiologie und Hygiene. 1. Abt. Originale A, Medizinische Mikrobiologie, Infektionskrankheiten und Parasitologie = International journal of microbiology and hygiene. A, Medical microbiology, infectiousdiseases, parasitology",
  ["zentralbl bakteriol mikrobiol hyg b"] = "Zentralblatt für Bakteriologie, Mikrobiologie und Hygiene. 1. Abt. Originale B, Hygiene",
  ["zentralbl bakteriol naturwiss"] = "Zentralblatt für Bakteriologie, Parasitenkunde, Infektionskrankheiten und Hygiene. Zweite naturwissenschaftliche Abteilung: Mikrobiologie der Landwirtschaft der Technologie und des Umweltschutzes",
  ["zentralbl bakteriol orig"] = "Zentralblatt für Bakteriologie, Parasitenkunde, Infektionskrankheiten und Hygiene. 1. Abt. Medizinisch-hygienische Bakteriologie, Virusforschung und Parasitologie. Originale",
  ["zentralbl bakteriol orig a"] = "Zentralblatt für Bakteriologie, Parasitenkunde, Infektionskrankheiten und Hygiene. Erste Abteilung Originale. Reihe A: Medizinische Mikrobiologie und Parasitologie",
  ["zentralbl bakteriol orig b"] = "Zentralblatt für Bakteriologie, Parasitenkunde, Infektionskrankheiten und Hygiene. Erste Abteilung Originale. Reihe B: Hygiene, präventive Medizin",
  ["zentralbl bakteriol parasitenkd infektionskr hyg"] = "Zentralblatt für Bakteriologie, Parasitenkunde, Infektionskrankheiten und Hygiene. Zweite naturwissenschaftliche Abt.: Allgemeine, landwirtschaftliche und technische Mikrobiologie",
  ["zentralbl bakteriol, suppl"] = "Zentralblatt für Bakteriologie, Supplement",
  ["zentralbl bauverwaltung"] = "Zentralblatt fur Bauverwaltung",
  ["zentralbl biochem biophys"] = "Zentralblatt fur Biochemie und Biophysik",
  ["zentralbl biol aerosolforsch"] = "Zentralblatt für biologische Aerosolforschung",
  ["zentralbl chir"] = "Zentralblatt für Chirurgie",
  ["zentralbl geol palaeontol teil 1"] = "Zentralblatt fur Geologie und Palaeontologie Teil 1: Allgemeine, Angewandte, Regionale, und Historische Geologie",
  ["zentralbl geol palaeontol teil 2"] = "Zentralblatt fur Geologie und Palaeontologie Teil 2: Palaeontologie",
  ["zentralbl gesamte rechtsmed grenzgeb"] = "Zentralblatt für die gesamte Rechtsmedizin und ihre Grenzgebiete",
  ["zentralbl gynakol"] = "Zentralblatt fur Gynakologie",
  ["zentralbl haut geschlechtskr grenzgeb"] = "Zentralblatt für Haut- und Geschlechtskrankheiten, sowie deren Grenzgebiete",
  ["zentralbl haut und geschlechtskr"] = "Zentralblatt Haut- und Geschlechtskrankheiten",
  ["zentralbl hyg umweltmed"] = "Zentralblatt fur Hygiene und Umweltmedizin",
  ["zentralbl mikrobiol"] = "Zentralblatt fur Mikrobiologie",
  ["zentralbl mineral geol palaeontol"] = "Zentralblatt fur Mineralogie, Geologie und Palaeontologie",
  ["zentralbl mineral teil 1"] = "Zentralblatt fur Mineralogie Teil 1: Kristallographie und Mineralogie",
  ["zentralbl mineral teil 2"] = "Zentralblatt fur Mineralogie Teil 2: Petrographie, Technische Mineralogie, Geochemie und Lagerstattenkunde",
  ["zentralbl neurochir"] = "Zentralblatt fur Neurochirurgie",
  ["zentralbl pathol"] = "Zentralblatt fur Pathologie",
  ["zentralbl pharm"] = "Zentralblatt fur Pharmazie",
  ["zentralbl pharm pharmakother laboratoriumsdiagn"] = "Zentralblatt fur Pharmazie, Pharmakotherapie und Laboratoriumsdiagnostik",
  ["zentralbl phlebol"] = "Zentralblatt fur Phlebologie",
  ["zentralbl verkehrsmed verkehrspsychol luft raumfahrtmed"] = "Zentralblatt für Verkehrs-Medizin, Verkehrs-Psychologie, Luft- und Raumfahrt-Medizin",
  ["zentralbl veterinarmed"] = "Zentralblatt fur Veterinarmedizin",
  ["zentralbl veterinarmed [b]"] = "Zentralblatt fur Veterinarmedizin. Reihe B",
  ["zentralbl veterinarmed [c]"] = "Zentralblatt fur Veterinarmedizin. Reihe C, Anatomia, Histologia, Embryologia",
  ["zentralbl veterinarmed a"] = "Zentralblatt fur Veterinarmedizin. Reihe A",
  ["zentralbl veterinarmed b"] = "Zentralblatt für Veterinärmedizin. Reihe B. Journal of veterinary medicine. Series B",
  ["zentralbl veterinarmed c"] = "Zentralblatt für Veterinärmedizin. Reihe C: Anatomie, Histologie, Embryologie",
  ["zentralblatt arbeitsmedizin arbeitsschutz erogon"] = "Zentralblatt fur Arbeitsmedizin, Arbeitsschutz und Ergonomie",
  ["zentralblatt bakteriol parasitenkd infekt hyg"] = "Zentralblatt für Bakteriologie, Parasitenkunde, Infektionskrankheiten und Hygiene. 1. Abt. Medizinisch-hygienische Bakteriologie, Virusforschung und Parasitologie. Referate",
  ["zentralsterilisation (wiesb)"] = "Zentralsterilisation (Wiesbaden)",
  ["zero popul growth natl rep"] = "Zero Population Growth national reporter",
  ["zero three"] = "Zero to three",
  ["zest hist"] = "Zeszyty historyczne (Paris, France)",
  ["zesz nauk uniw jagiell pr geogr"] = "Zeszyty Naukowe Uniwersytetu Jagiellonskiego. Prace Geograficzne",
  ["zeszyty nauk geom"] = "Zeszyty Naukowe Geometria",
  ["zeszyty nauk politech rzeszowskiej mat"] = "Zeszyty Naukowe Politechniki Rzeszowskiej",
  ["zeszyty nauk wyż szkoły ped powstańców śl opolu mat"] = "Zeszyty Naukowe Wyższa Szkoła Pedagogiczna im. Powstańców Śl\\polhk askich w Opolu. Matematyka",
  ["zfa (stuttgart)"] = "ZFA. Zeitschrift für Allgemeinmedizin",
  ["zfa - z allgemeinmed"] = "ZFA - Zeitschrift für Allgemeinmedizin",
  ["zh beloruss gos univ mat inform"] = "Zhurnal Belorusskogo Gosudarstvennogo Universiteta. Matematika. Informatika",
  ["zh eksp klin med"] = "Zhurnal eksperimental'noĭ i klinicheskoĭ meditsiny",
  ["zh eksp teor fiz"] = "Zhurnal Éksperimentalnoi i Teoreticheskoi Fiziki",
  ["zh eksp teor fiz [sov phys jetp]"] = "Zhurnal Eksperimentalnoi i Teoreticheskoi Fiziki [Soviet Physics–JETP]",
  ["zh evol biokhim fiziol"] = "Zhurnal evoliutsionnoĭ biokhimii i fiziologii",
  ["zh fiz khim"] = "Zhurnal Fizicheskoi Khimii",
  ["zh fiz khim [russ j phys chem]"] = "Zhurnal Fizicheskoi Khimii [Russian Journal of Physical Chemistry]",
  ["zh mat fiz anal geom"] = "Zhurnal Matematicheskoi Fiziki, Analiza, Geometrii. Journal of Mathematical Physics, Analysis, Geometry",
  ["zh mikrobiol epidemiol immunobiol"] = "Zhurnal mikrobiologii, ėpidemiologii i immunobiologii",
  ["zh neorg khim"] = "Zhurnal Neorganicheskoi Khimii",
  ["zh nevrol psikhiatr im s s korsakova"] = "Zhurnal nevrologii i psikhiatrii imeni S.S. Korsakova",
  ["zh nevropatol psikhiatr im s s korsakova"] = "Zhurnal nevropatologii i psikhiatrii imeni S.S. Korsakova (Moscow, Russia : 1952)",
  ["zh obshch biol"] = "Zhurnal Obshchei Biologii",
  ["zh obshch khim"] = "Zhurnal Obshchei Khimii",
  ["zh org farm khim"] = "Zhurnal Organichnoi ta Farmatsevtichnoi Khimii",
  ["zh org khim"] = "Zhurnal Organicheskoi Khimii",
  ["zh prikl khim"] = "Zhurnal Prikladnoi Khimii",
  ["zh prikl khim (leningrad)"] = "Zhurnal Prikladnoi Khimii (S. -Peterburg)",
  ["zh prikl spektrosk"] = "Zhurnal Prikladnoi Spektroskopii",
  ["zh prikl spektrosk [j appl spectrosc (ussr)]"] = "Zhurnal Prikladnoi Spektroskopii [Journal of Applied Spectroscopy (USSR)]",
  ["zh sib fed univ mat fiz"] = "Zhurnal Sibirskogo Federalnogo Universiteta. Matematika i Fizika",
  ["zh tekh fiz [sov phys tech phys]"] = "Zhurnal Tekhnicheskoi Fiziki [Soviet Physics–Technical Physics]",
  ["zh ushn nos gorl bolezn"] = "Zhurnal ushnykh, nosovykh i gorlovykh bolezneĭ = The journal of otology, rhinology, and laryngologie [sic]",
  ["zh vopr neirokhir im n n burdenko"] = "Zhurnal voprosy neĭrokhirurgii imeni N. N. Burdenko",
  ["zh vses khim o-va im d i mendeleeva"] = "Zhurnal Vsesoyuznogo Khimicheskogo Obshchestva im. D. I. Mendeleeva",
  ["zh vychisl mat mat fiz"] = "Zhurnal Vychislitelnoi Matematiki i Matematicheskoi Fiziki. Rossiiskaya Akademiya Nauk",
  ["zh vyssh nerv deiat im i p pavlova"] = "Zhurnal Vysshei Nervnoi Deiatelnosti Imeni I. P. Pavlova",
  ["zhangzhou shiyuan xuebao (ziran kexue ban)"] = "Zhangzhou Shiyuan Xuebao (Ziran Kexue Ban)",
  ["zhe jiang yi ke da xue xue bao"] = "Zhejiang yi ke da xue xue bao = Journal of Zhejiang Medical University",
  ["zhejiang da xue xue bao li xue ban"] = "Zhejiang da xue xue bao. Li xue ban = Journal of Zhejiang University. Sciences edition",
  ["zhejiang da xue xue bao yi xue ban"] = "Zhejiang da xue xue bao. Yi xue ban = Journal of Zhejiang University. Medical sciences",
  ["zhejiang daxue xue bao nong ye yu sheng ming ke xue ban"] = "Zhejiang daxue xue bao. Nong ye yu sheng ming ke xue ban = Journal of Zhejiang University. Agriculture and life sciences",
  ["zhejiang daxue xuebao ziran kexue ban"] = "Zhejiang Daxue Xuebao",
  ["zhejiang haiyang xueyuan xuebao ziran kexue ban"] = "Zhejiang haiyang xueyuan xuebao ziran kexue ban",
  ["zhejiang nong ye da xue xue bao"] = "Zhejiang nong ye da xue xue bao = Acta Agriculturae Universitatis Chekianensis",
  ["zhejiang nong ye xue bao"] = "Zhejiang nong ye xue bao = Acta agriculturae Zhejiangensis",
  ["zhejiang yu fang yi xue"] = "Zhejiang yu fang yi xue",
  ["zhen ci yan jiu"] = "Zhen ci yan jiu = Acupuncture research",
  ["zhengzhou da xue xue bao yi xue ban"] = "Zhengzhou da xue xue bao. Yi xue ban = Journal of Zhengzhou University. Medical sciences = Zhengzhou daxue xuebao. Yixueban",
  ["zhi wu bao hu"] = "Zhi wu bao hu = Plant protection",
  ["zhi wu bao hu xue bao"] = "Zhi wu bao hu xue bao = Acta phytophylacica sinica",
  ["zhi wu bao hu xue hui hui kan"] = "Zhi wu bao hu xue hui hui kan Plant protection bulletin",
  ["zhi wu bing li xue bao"] = "Zhi wu bing li xue bao = Acta phytopathologica sinica",
  ["zhi wu bing li xue hui kan"] = "Zhi wu bing li xue hui kan = Plant pathology bulletin",
  ["zhi wu fen lei xue bao"] = "Zhi wu fen lei xue bao = Acta phytotaxonomica Sinica",
  ["zhi wu jian yi"] = "Zhi wu jian yi",
  ["zhi wu sheng li xue bao"] = "Zhi wu sheng li xue bao = Acta phytophysiologica Sinica",
  ["zhi wu sheng li xue bao (2011)"] = "Zhi wu sheng li xue bao (2011)",
  ["zhi wu sheng li xue tong xun"] = "Zhi wu sheng li xue tong xun",
  ["zhi wu sheng li yu fen zi sheng wu xue xue bao"] = "Zhi wu sheng li yu fen zi sheng wu xue xue bao = Journal of plant physiology and molecular biology",
  ["zhi wu yan jiu"] = "Zhi wu yan jiu (2001)",
  ["zhi wu yi xue qi kan"] = "Zhi wu yi xue qi kan",
  ["zhinitik hizarahi sivvum"] = "Zhinitīk dar hizārah-i sivvum",
  ["zhinitik i nuvin"] = "Zhinitīk-i nuvīn",
  ["zhipu xuebao"] = "Zhipu Xuebao",
  ["zhivoy yazyk"] = "\\cyr Zhivoy Yazyk",
  ["zhiwu xuebao"] = "Zhiwu xuebao = Chinese bulletin of botany",
  ["zhiwu yichuan ziyuan xuebao"] = "Zhiwu yichuan ziyuan xuebao",
  ["zhiwu ziyuan yu huanjing xuebao"] = "Zhiwu ziyuan yu huanjing xuebao",
  ["zhiwuxue tongbao"] = "Zhiwuxue tongbao",
  ["zhong cao yao"] = "Zhong cao yao = Chinese traditional and herbal drugs",
  ["zhong gong yan jiu"] = "Zhong gong yan jiu = Studies on Chinese communism",
  ["zhong guo di fang bing xue za zhi"] = "Zhongguo di fang bing xue za zhi = Zhongguo difangbingxue zazhi = Chinese journal of endemiology",
  ["zhong guo ke ji lun wen zai xian"] = "Zhong guo ke ji lun wen zai xian",
  ["zhong liu fang zhi za zhi"] = "Zhong liu fang zhi za zhi = China journal of cancer prevention and treatment",
  ["zhong liu za zhi"] = "Zhong liu za zhi",
  ["zhong nan da xue xue bao yi xue ban"] = "Zhong nan da xue xue bao. Yi xue ban = Journal of Central South University. Medical sciences",
  ["zhong xi yi jie he xue bao"] = "Zhong xi yi jie he xue bao = Journal of Chinese integrative medicine",
  ["zhong xi yi jie he za zhi"] = "Zhong xi yi jie he za zhi = Chinese journal of modern developments in traditional medicine",
  ["zhong yao cai"] = "Zhong yao cai = Zhongyaocai = Journal of Chinese medicinal materials",
  ["zhong yao tong bao"] = "Zhong yao tong bao (Beijing, China : 1981)",
  ["zhonggue rupin gongye"] = "Zhongguo rupin gongye = China dairy industry",
  ["zhongguo baozhuang keji bolan"] = "Zhongguo baozhuang ke-ji bolan",
  ["zhongguo bing du bing za zhi"] = "Zhongguo bing du bing za zhi",
  ["zhongguo bing du xue"] = "Zhongguo bing du xue = Virologica Sinica",
  ["zhongguo bing yuan sheng wu xue za zhi"] = "Zhongguo bing yuan sheng wu xue za zhi = Journal of pathogen biology",
  ["zhongguo cao shi dong wu"] = "Zhongguo cao shi dong wu = China herbivores",
  ["zhongguo caodi xuebao"] = "Zhongguo caodi xuebao = Chinese journal of grassland",
  ["zhongguo chao sheng yixue za zhi"] = "Zhongguo chao sheng yi xue za zhi = Zhongguo chaosheng yixue zazhi = Ultrasonic medical journal of China",
  ["zhongguo dang dai er ke za zhi"] = "Zhongguo dang dai er ke za zhi = Chinese journal of contemporary pediatrics",
  ["zhongguo dongwu jianyi"] = "Zhongguo dongwu jianyi = Chinese journal of animal quaratine",
  ["zhongguo fei ai za zhi"] = "Zhongguo fei ai za zhi = Chinese journal of lung cancer",
  ["zhongguo gong gong wei sheng"] = "Zhongguo gong gong wei sheng = China public health",
  ["zhongguo gu shang"] = "Zhongguo gu shang = China journal of orthopaedics and traumatology",
  ["zhongguo hai yang hu zhao xue bao / edited by the chinese society of oceanology and limnology|chin j oceanol limnol"] = "Chinese journal of oceanology and limnology",
  ["zhongguo hai yang yao wu"] = "Zhongguo hai yang yao wu = Chinese journal of marine drugs",
  ["zhongguo haiyang daxue xuebao ziran kexue ban"] = "Zhongguo hai yang da xue xue bao. Zi ran ke xue ban = Zhongguo haiyang daxue xuebao. Ziran kexue ban",
  ["zhongguo hsu mu za zhi"] = "Zhongguo Hsu Mu Za Zhi",
  ["zhongguo ji sheng chong bing fang zhi za zhi"] = "Zhongguo ji sheng chong bing fang zhi za zhi = Chinese journal of parasitic disease control",
  ["zhongguo ji sheng chong xue yu ji sheng chong bing za zhi"] = "Zhongguo ji sheng chong xue yu ji sheng chong bing za zhi = Chinese journal of parasitology & parasitic diseases",
  ["zhongguo jiaqin"] = "Zhongguo jiaqin",
  ["zhongguo jiguang"] = "Zhongguo Jiguang",
  ["zhongguo kang sheng su za zhi"] = "Zhongguo kang sheng su za zhi = Chinese journal of antibiotics",
  ["zhongguo kangshengsu zazhi"] = "Zhongguo Kangshengsu Zazhi",
  ["zhongguo ke ji shi liao"] = "Zhongguo ke ji shi liao = China historical materials of science and technology",
  ["zhongguo ke xue"] = "Zhongguo ke xue : Sheng ming ke xue = Scientia sinica vitae",
  ["zhongguo kun chong xue|entomol sin"] = "Entomologia Sinica",
  ["zhongguo lin chuang yao li xue za zhi"] = "Zhongguo lin chuang yao li xue za zhi = The Chinese journal of clinical pharmacology",
  ["zhongguo mei jie sheng wu xue ji kong zhi za zhi"] = "Zhongguo mei jie sheng wu xue ji kong zhi za zhi : Zhongguo yu fang yi xue hui xi lie za zhi = Chinese journal of vector biology and control : serial journal of China Preventive Medical Association = Zhongguo meijieshengwuxue ji kongzhi zazhi",
  ["zhongguo mian yi xue za zhi"] = "Zhongguo mian yi xue za zhi = Chinese journal of immunology",
  ["zhongguo niangzao"] = "Zhongguo niang zao = Zhongguo niangzao",
  ["zhongguo nong xue tong bao"] = "Zhongguo nong xue tong bao = Chinese agricultural science bulletin",
  ["zhongguo nong ye da xue xue bao"] = "Zhongguo nong ye da xue xue bao = Journal of China Agricultural University",
  ["zhongguo nong ye ke xue"] = "Zhongguo nong ye ke xue = Zhongguo nongye kexue",
  ["zhongguo nongye kexue|zhongguo nong ye ke xue"] = "Zhongguo nong ye ke xue",
  ["zhongguo ren kou ke xue"] = "Zhongguo ren kou ke xue = Population science of China",
  ["zhongguo ren shou gong huan bing za zhi"] = "Zhongguo ren shou gong huan bing za zhi = Chinese journal of zoonoses",
  ["zhongguo she hui xue kan"] = "Zhongguo she hui xue kan",
  ["zhongguo sheng hua yao wu za zhi"] = "Zhongguo sheng hua yao wu za zhi",
  ["zhongguo sheng tai nong ye xue bao"] = "Zhongguo sheng tai nong ye xue bao",
  ["zhongguo sheng wu fang zhi"] = "Zhongguo sheng wu fang zhi",
  ["zhongguo sheng wu gong cheng za zhi"] = "Zhongguo sheng wu gong cheng za zhi = Journal of Chinese biotechnology",
  ["zhongguo sheng wu hua xue yu fen zi sheng wu xue bao"] = "Zhongguo sheng wu hua xue yu fen zi sheng wu xue bao = Chinese journal of biochemistry and molecular biology",
  ["zhongguo sheng wu zhi pin xue za zhi"] = "Zhongguo sheng wu zhi pin xue za zhi : Zhonghua yu fang yi xue hui xi lie za zhi = Zhongguo shengwu zhipin xue zazhi = Chinese journal of biologicals : serial journal of China Preventive Medical Association",
  ["zhongguo shengwu huaxue yu fenzi shengwu xuebao"] = "Zhongguo Shengwu Huaxue Yu Fenzi Shengwu Xuebao",
  ["zhongguo shi yan dong wu xue bao"] = "Zhongguo shi yan dong wu xue bao = Acta laboratorium animalis scientia sinica",
  ["zhongguo shi yan fang ji xue za zhi"] = "Zhongguo shi yan fang ji xue za zhi = Chinese journal of experimental traditional medical formulae",
  ["zhongguo shi yan jiu"] = "Zhongguo shi yan jiu = Zhongguoshi yanjiu",
  ["zhongguo shi yan xue ye xue za zhi"] = "Zhongguo shi yan xue ye xue za zhi",
  ["zhongguo shipin xuebao"] = "Zhongguo shipin xuebao = Journal of Chinese Institute of Food Science and Technology",
  ["zhongguo shiyan zhenduanxue"] = "Zhongguo shiyan zhenduanxue",
  ["zhongguo shou yao zazhi"] = "Zhongguo shou yao za zhi",
  ["zhongguo shou yi za zhi"] = "Zhongguo shou yi za zhi = Chinese journal of veterinary medicine",
  ["zhongguo shouyi ke ji"] = "Zhongguo shou yi ke-ji = Chinese journal of veterinary science and technology",
  ["zhongguo shu cai"] = "Zhongguo shu cai = Chinese vegetables",
  ["zhongguo shui chan ke xue"] = "Zhongguo shui chan ke xue = Journal of fishery sciences of China",
  ["zhongguo shuidao kexue"] = "Zhongguo shuidao kexue",
  ["zhongguo tu rang yu fei liao"] = "Zhongguo tu rang yu fei liao = Soil and fertilizer sciences in China",
  ["zhongguo wei sheng jian yan"] = "Zhongguo wei sheng jian yan",
  ["zhongguo wei zhong bing ji jiu yi xue"] = "Zhongguo wei zhong bing ji jiu yi xue = Chinese critical care medicine = Zhongguo weizhongbing jijiuyixue",
  ["zhongguo weishengtaixue zazhi"] = "Zhongguo wei sheng tai xue za zhi",
  ["zhongguo wuji fenxi huaxue"] = "Zhongguo Wuji Fenxi Huaxue",
  ["zhongguo xian dai shen jing ji bing za zhi"] = "Zhongguo xian dai shen jing ji bing za zhi",
  ["zhongguo xin xiao hua bing xue za zhi|china natl j new gastroenterol"] = "China national journal of new gastroenterology",
  ["zhongguo xitu xuebao"] = "Zhongguo Xitu Xuebao",
  ["zhongguo xiu fu chong jian wai ke za zhi"] = "Zhongguo xiu fu chong jian wai ke za zhi = Zhongguo xiufu chongjian waike zazhi = Chinese journal of reparative and reconstructive surgery",
  ["zhongguo xu mu shou yi"] = "Zhongguo xu mu shou yi = Zhongguo xumu shouyi = China animal husbandry & veterinary medicine",
  ["zhongguo xu qin chuan ran bing"] = "Zhongguo xu qin chuan ran bing",
  ["zhongguo xue xi chong bing fang zhi za zhi"] = "Zhongguo xue xi chong bing fang zhi za zhi = Chinese journal of schistosomiasis control",
  ["zhongguo yan cao xue bao"] = "Zhongguo yan cao xue bao = Acta Tabacaria sinica = Zhongguo yancao xuebao",
  ["zhongguo yao li xue bao"] = "Zhongguo yao li xue bao = Acta pharmacologica Sinica",
  ["zhongguo yao shi"] = "Zhongguo yao shi",
  ["zhongguo yao xue za zhi"] = "Zhongguo yao xue za zhi (Zhongguo yao xue hui : 1989)",
  ["zhongguo yaoli xuebao"] = "Zhongguo Yaoli Xuebao",
  ["zhongguo yi liao qi xie za zhi"] = "Zhongguo yi liao qi xie za zhi = Chinese journal of medical instrumentation",
  ["zhongguo yi miao he mian yi"] = "Zhongguo yi miao he mian yi",
  ["zhongguo yi shi za zhi"] = "Zhongguo yi shi za zhi",
  ["zhongguo yi xue ke xue yuan xue bao"] = "Zhongguo yi xue ke xue yuan xue bao. Acta Academiae Medicinae Sinicae",
  ["zhongguo ying yong sheng li xue za zhi"] = "Zhongguo ying yong sheng li xue za zhi = Zhongguo yingyong shenglixue zazhi = Chinese journal of applied physiology",
  ["zhongguo yiyao gongye zazhi"] = "Zhongguo Yiyao Gongye Zazhi",
  ["zhongguo yiyao shengwu jishu"] = "Zhongguo yiyao shengwu jishu",
  ["zhongguo you liao zuo wu xue bao"] = "Zhongguo you liao zuo wu xue bao = Chinese journal of oil crop sciences",
  ["zhongguo yuanyi wenzhai"] = "Zhongguo yuanyi wenzhai",
  ["zhongguo zhen jiu"] = "Zhongguo zhen jiu = Chinese acupuncture & moxibustion",
  ["zhongguo zhen jun xue za zhi"] = "Zhongguo zhen jun xue za zhi",
  ["zhongguo zhi bao dao kan"] = "Zhongguo zhi bao dao kan = China plant protection",
  ["zhongguo zhong xi yi jie he za zhi"] = "Zhongguo Zhong xi yi jie he za zhi Zhongguo Zhongxiyi jiehe zazhi = Chinese journal of integrated traditional and Western medicine",
  ["zhongguo zhong yao za zhi"] = "Zhongguo Zhong yao za zhi = Zhongguo zhongyao zazhi = China journal of Chinese materia medica",
  ["zhonghua bing li xue za zhi"] = "Zhonghua bing li xue za zhi = Chinese journal of pathology",
  ["zhonghua chuan ran bing za zhi"] = "Zhonghua chuan ran bing za zhi = Zhonghua chuanranbing zazhi = Chinese journal of infectious diseases",
  ["zhonghua er bi yan hou ke za zhi"] = "Zhonghua er bi yan hou ke za zhi",
  ["zhonghua er bi yan hou tou jing wai ke za zhi"] = "Zhonghua er bi yan hou tou jing wai ke za zhi = Chinese journal of otorhinolaryngology head and neck surgery",
  ["zhonghua er ke za zhi"] = "Zhonghua er ke za zhi = Chinese journal of pediatrics",
  ["zhonghua fang she xue za zhi"] = "Zhonghua fang she xue za zhi Chinese journal of radiology",
  ["zhonghua fang she yi xue yu fang hu za zhi"] = "Zhonghua fang she yi xue yu fang hu za zhi = Chinese journal of radiological medicine and protection",
  ["zhonghua fangsh yixue yu fanghu zazhi"] = "Zhonghua Fangsh Yixue Yu Fanghu Zazhi",
  ["zhonghua fu chan ke za zhi"] = "Zhonghua fu chan ke za zhi",
  ["zhonghua gan zang bing za zhi"] = "Zhonghua gan zang bing za zhi = Zhonghua ganzangbing zazhi = Chinese journal of hepatology",
  ["zhonghua hu li za zhi"] = "Zhonghua hu li za zhi = Chinese journal of nursing",
  ["zhonghua jian yan yi xue za zhi"] = "Zhonghua jian yan yi xue za zhi = Chinese journal of laboratory medicine",
  ["zhonghua jie he he hu xi xi ji bing za zhi"] = "Zhonghua jie he he hu xi xi ji bing za zhi = Chinese journal of tuberculosis and respiratory diseases",
  ["zhonghua jie he he hu xi za zhi"] = "Zhonghua jie he he hu xi za zhi = Zhonghua jiehe he huxi zazhi = Chinese journal of tuberculosis and respiratory diseases",
  ["zhonghua kou qiang ke za zhi"] = "Zhonghua kou qiang ke za zhi [Chinese journal of stomatology]",
  ["zhonghua kou qiang yi xue za zhi"] = "Zhonghua kou qiang yi xue za zhi = Zhonghua kouqiang yixue zazhi = Chinese journal of stomatology",
  ["zhonghua lao dong wei sheng zhi ye bing za zhi"] = "Zhonghua lao dong wei sheng zhi ye bing za zhi = Zhonghua laodong weisheng zhiyebing zazhi = Chinese journal of industrial hygiene and occupational diseases",
  ["zhonghua linxue jikan"] = "Zhonghua lin xue ji kan = Quarterly journal of Chinese forestry",
  ["zhonghua liu xing bing xue za zhi"] = "Zhonghua liu xing bing xue za zhi = Zhonghua liuxingbingxue zazhi",
  ["zhonghua min guo shou yi xue hui za zhi"] = "Zhonghua Minguo shou yi xue hui za zhi Journal of the Chinese Society of Veterinary Science",
  ["zhonghua min guo taiwan jen kou tong ji ji kan"] = "Zhonghua min guo Taiwan ren kou tong ji ji kan. China (Republic : 1949- ). Nei zheng bu",
  ["zhonghua min guo wei sheng wu ji mian yi xue za zhi"] = "Zhonghua Minguo wei sheng wu ji mian yi xue za zhi = Chinese journal of microbiology and immunology",
  ["zhonghua min guo wei sheng wu xue za zhi"] = "Zhonghua Minguo wei sheng wu xue za zhi = Chinese journal of microbiology",
  ["zhonghua min guo xiao er ke yi xue hui za zhi"] = "Zhonghua Minguo xiao er ke yi xue hui za zhi [Journal]. Zhonghua Minguo xiao er ke yi xue hui",
  ["zhonghua minguo xin zang xue hui za zhi"] = "Zhonghua Minguo xin zang xue hui za zhi = Acta Cardiologica Sinica",
  ["zhonghua nan ke xue"] = "Zhonghua nan ke xue = National journal of andrology",
  ["zhonghua nei ke za zhi"] = "Zhonghua nei ke za zhi",
  ["zhonghua shao shang za zhi"] = "Zhonghua shao shang za zhi = Zhonghua shaoshang zazhi = Chinese journal of burns",
  ["zhonghua shen jing ge za zhi"] = "Zhonghua shen jing ke za zhi = Chinese journal of neurology",
  ["zhonghua shen jing jing shen ke za zhi"] = "Zhonghua shen jing jing shen ke za zhi = Chinese journal of neurology and psychiatry",
  ["zhonghua sheng wu yi xue gong cheng za zhi"] = "Zhonghua sheng wu yi xue gong cheng za zhi",
  ["zhonghua shi yan he lin chuang bing du xue za zhi"] = "Zhonghua shi yan he lin chuang bing du xue za zhi = Zhonghua shiyan he linchuang bingduxue zazhi = Chinese journal of experimental and clinical virology",
  ["zhonghua shi yan wai ke za zhi"] = "Zhonghua shi yan wai ke za zhi = Zhonghua shiyan waike zazhi = Chinese journal of experimental surgery",
  ["zhonghua wai ke za zhi"] = "Zhonghua wai ke za zhi [Chinese journal of surgery]",
  ["zhonghua wei chang wai ke za zhi"] = "Zhonghua wei chang wai ke za zhi = Chinese journal of gastrointestinal surgery",
  ["zhonghua wei sheng wu xue he mian yi xue za zhi"] = "Zhonghua wei sheng wu xue he mian yi xue za zhi = Zhonghua weishengwuxue he mianyixue zazhi",
  ["zhonghua wei zhong bing ji jiu yi xue"] = "Zhonghua wei zhong bing ji jiu yi xue",
  ["zhonghua xin xue guan bing za zhi"] = "Zhonghua xin xue guan bing za zhi",
  ["zhonghua xue ye xue za zhi"] = "Zhonghua xue ye xue za zhi = Zhonghua xueyexue zazhi",
  ["zhonghua ya yi xue hui za zhi"] = "Zhonghua ya yi xue hui za zhi",
  ["zhonghua yan ke za zhi"] = "[Zhonghua yan ke za zhi] Chinese journal of ophthalmology",
  ["zhonghua yi shi za zhi"] = "Zhonghua yi shi za zhi (Beijing, China : 1980)",
  ["zhonghua yi xue yi chuan xue za zhi"] = "Zhonghua yi xue yi chuan xue za zhi = Zhonghua yixue yichuanxue zazhi = Chinese journal of medical genetics",
  ["zhonghua yi xue za zhi"] = "Zhonghua yi xue za zhi",
  ["zhonghua yi xue za zhi (taipei)"] = "Zhonghua yi xue za zhi = Chinese medical journal; Free China ed",
  ["zhonghua yi yuan gan ran xue za zhi"] = "Zhonghua yi yuan gan ran xue za zhi = Chinese journal of nosocomiology",
  ["zhonghua yu fang yi xue za zhi"] = "Zhonghua yu fang yi xue za zhi [Chinese journal of preventive medicine]",
  ["zhonghua zheng xing shao shang wai ke za zhi"] = "Zhonghua zheng xing shao shang wai ke za zhi = Zhonghua zheng xing shao shang waikf [i.e. waike] zazhi = Chinese journal of plastic surgery and burns",
  ["zhonghua zheng xing wai ke za zhi"] = "Zhonghua zheng xing wai ke za zhi = Zhonghua zhengxing waike zazhi = Chinese journal of plastic surgery",
  ["zhonghua zhong liu za zhi"] = "Zhonghua zhong liu za zhi [Chinese journal of oncology]",
  ["zhongnan minzu daxue xuebao ziran kexue ban"] = "Zhongnan Minzu Daxue xuebao. Ziran kexue ban = Journal of South-Central University for Nationalities. Natural Science Edition",
  ["zhongshan da xue xue bao zi ran ke xue ban"] = "Zhongshan da xue xue bao. Zi ran ke xue ban = Acta scientiarum naturalium Universitatis Sunyatseni",
  ["zhongshan yi ke da xue xue bao"] = "Zhongshan yi ke da xue xue bao = Zhongshan yike daxue xuebao = Academic journal of Sun Yat-sen University of Medical Sciences",
  ["zhonguo nongshi"] = "Zhongguo nong shi",
  ["zhurnal prikl him"] = "Zhurnal prikladnoĭ khimii (Leningrad, R.S.F.S.R.)",
  ["zhuxing xuebao"] = "Zhuxing xuebao",
  ["zi ran ke xue"] = "Zi ran ke xue (Irvine, Calif.)",
  ["zi ran ke xue jin zhan"] = "Zi ran ke xue jin zhan",
  ["zi ran ke xue shi yan"] = "Zi ran ke xue shi yan = Studies in the history of natural sciences",
  ["zi ran za zhi"] = "Zi ran za zhi. Ziran zazhi",
  ["zidonghua xuebao"] = "Zi dong hua xue bao = Acta automatica Sinica",
  ["zimb j econ"] = "The Zimbabwe journal of economics",
  ["zimbabwe nurse"] = "The Zimbabwe nurse",
  ["zimbabwe rhod nurse"] = "Zimbabwe Rhodesia Nurse",
  ["zist fanavarii giyahani zirai"] = "Zīst/fanāvarī-i giyāhān-i zirāī",
  ["zistshinasii mikruurganismha"] = "Zīstʹshināsī-i mīkrūʹurgānīsmʹhā : Biological journal of microorganism",
  ["ziva antika"] = "Živa antika. Antiquité vivante",
  ["zkg int"] = "ZKG International",
  ["znam daty istor zdravookhr med azerb"] = "Znamenatel'nye daty istorii zdravookhraneniia i meditsiny Azerbaĭdzhana",
  ["zobozdrav vestn"] = "Zobozdravstveni vestnik",
  ["zoo biol"] = "Zoo biology",
  ["zool abh"] = "Zoologische Abhandlungen",
  ["zool anz"] = "Zoologischer Anzeiger",
  ["zool bespozvon"] = "Zoologiâ bespozvonoc̆nyh",
  ["zool caboverdiana"] = "Zoologia caboverdiana",
  ["zool ecol"] = "Zoology and ecology (Print)",
  ["zool gart"] = "Der zoologische Garten",
  ["zool gart nf"] = "Zoologischer Garten N.F.",
  ["zool garten"] = "Zoologische Garten",
  ["zool j linn soc"] = "Zoological journal of the Linnean Society",
  ["zool jahrb abt anat ontogenie tiere"] = "Zoologische Jahrbücher. Abteilung für Anatomie und Ontogenie der Tiere",
  ["zool jahrb abt syst"] = "Zoologisches Jahrbuch, Abteilung für Systematik",
  ["zool lett"] = "Zoological Letters",
  ["zool meded"] = "Zoologische mededeelingen",
  ["zool middle east"] = "Zoology in the Middle East",
  ["zool pol"] = "Zoologica Poloniae",
  ["zool res"] = "Zoological research",
  ["zool sci"] = "Zoological Science",
  ["zool sci contrib n y zool soc"] = "Zoologica; scientific contributions of the New York Zoological Society",
  ["zool scr"] = "Zoologica scripta",
  ["zool stud"] = "Zoological studies",
  ["zool syst"] = "Zoological systematics",
  ["zool zh"] = "Zoologicheskii Zhurnal",
  ["zool zhurnal"] = "Zoologicheskiĭ zhurnal",
  ["zoolog sci"] = "Zoological science",
  ["zoologia (curitiba)"] = "Zoologia (Curitiba, Brazil)",
  ["zoological lett"] = "Zoological letters",
  ['zoological research / "dong wu xue yan jiu" bian ji we yuan hui bian ji|zool res'] = "Dong wu xue yan jiu",
  ["zoology (jena)"] = "Zoology (Jena, Germany)",
  ["zoonoses public health"] = "Zoonoses and public health",
  ["zoonoses res"] = "Zoonoses research",
  ["zoos print j"] = "Zoos' print journal",
  ["zoosyst evol"] = "Zoosystematics and Evolution",
  ["zoosyst ross"] = "Zoosystematica rossica",
  ["zootec trop"] = "Zootecnia tropical",
  ["zorg welz"] = "Zorg + welzijn",
  ["zpg report"] = "ZPG reporter",
  ["zpr geogr ust csav"] = "Zprávy Geografického ústavu ČSAV",
  ["zu guo"] = "Zu guo",
  ["zubolekarski pregl"] = "Zubolekarski pregled",
  ["zuger nachr"] = "Zuger Nachrichten",
  ["zuger neujahrsbl"] = "Zuger Neujahrsblatt",
  ["zuo wu xue bao"] = "Zuo wu xue bao",
  ["zur lect adv math"] = "Zurich Lectures in Advanced Mathematics",
  ["zur lect multiph flow"] = "Zurich Lectures on Multiphase Flow",
  ["zur medizingesch abh"] = "Zürcher medizingeschichtliche Abhandlungen",
  ["zvesti cerv kriza"] = "Zvestí c̆erveného kríz̆a",
  ["zür geogr schr"] = "Zürcher Geographische Schriften",
  ["âne d’or"] = "L’Âne d’Or",
  ["éc prat ht études iv$\\sp e$ sect sci hist philol v ht études médiév mod"] = "École Pratique des Hautes études",
  ["ökol landbau"] = "Ökologie und Landbau",
  ["österr bot z"] = "Österreichische botanische Zeitschrift",
  ["österr forstztg"] = "Österreichische Forstzeitung",
  ["österr wasser- abfallwirtsch"] = "Österreichische Wasser- und Abfallwirtschaft",
  ["österreich akad wiss math-natur kl sitzungsber ii"] = "Österreichische Akademie der Wissenschaften Mathematisch-Naturwissenschaftliche Klasse",
  ["überbl math"] = "Überblicke Mathematik",
}

return {
  abbrevs = abbrevs,
  unabbrevs = unabbrevs,
}
